(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[   7297560,     188767]*)
(*NotebookOutlinePosition[   7331752,     189999]*)
(*  CellTagsIndexPosition[   7331659,     189993]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Basic Plotting", "Title",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "from ",
  StyleBox[ButtonBox["Essential Mathematica for Students of Science",
    Active->False,
    ButtonData:>{
      URL[ 
      "http://www.physics.umd.edu/courses/CourseWare/EssentialMathematica/"], 
      None},
    Active->False,
    ButtonStyle->"Hyperlink"],
    Active->False],
  "\n\[Copyright] James J. Kelly, 1998\nLast revised: January, 2006"
}], "Author",
  Active->True,
  TextAlignment->Center,
  TextJustification->0],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell[TextData[{
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " provides a collection of plotting tools that rival those of many packages \
dedicated to that single purpose.  Although many plotting packages provide \
friendlier user interfaces, ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " is unique in its integration of graphics with numerical, symbolic, and \
typesetting capabilities.  Some of the awkward features of the plotting \
commands can be attributed to the integration of all of these capabilities \
within a uniform structure.  Hence, although it is almost trivial to obtain \
quick plots, producing publication-quality figures requires choosing many \
rather verbose options and may take a lot of time.  Nevertheless, it is worth \
mastering some of the most important options and techniques early in one's ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " career, particularly for science and engineering applications for which \
visualization is an important tool for solving problems and developing \
insights.  In this notebook we introduce the plotting commands and options \
that I have found to be most useful for routine applications (at least in \
physics).  "
}], "Text"],

Cell[TextData[{
  "The following table lists selected built-in plotting functions with \
schematic syntax and brief descriptions.  In addition there are animation \
functions, such as ",
  StyleBox["MoviePlot", "InlineInput",
    FontWeight->"Bold"],
  ", which we will not discuss here.  The behavior of each of these functions \
can be controlled by a vast array of plotting options \[LongDash] we cannot \
hope to illustrate all functions and options without exhausting the reader \
(and the writer)!  Plotting functions can also be combined with ",
  StyleBox["graphics primitives",
    FontSlant->"Italic"],
  " to add text, line drawings, or other visual information not usually \
described as a plot.  By using the ",
  StyleBox["Reference Guide",
    FontSlant->"Italic"],
  " or ",
  StyleBox["Help Browser",
    FontSlant->"Italic"],
  " you should be able to find a command suitable for practically any type of \
plot or figure that you might need and the use of the available options \
should, with a little practice, be almost self-explanatory."
}], "Text"],

Cell[BoxData[
    StyleBox[GridBox[{
          {
            StyleBox[\(Selected\ built - in\ plotting\ functions\),
              FontFamily->"Times New Roman",
              FontWeight->"Bold"]},
          {
            StyleBox[GridBox[{
                  {
                    StyleBox["function",
                      "Text"], 
                    StyleBox[\(purpose/format\),
                      "Text"]},
                  {
                    StyleBox[
                      RowBox[{"Plot", "[", 
                        RowBox[{
                          StyleBox["y",
                            "Text"], 
                          StyleBox[",",
                            "Text"], 
                          RowBox[{
                            StyleBox["{",
                              "Text"], 
                            RowBox[{
                              StyleBox["x",
                                "Text"], 
                              StyleBox[",",
                                "Text"], 
                              SubscriptBox[
                                StyleBox["x",
                                  "Text"], "min"], ",", \(x\_max\)}], "}"}]}],
                         "]"}],
                      "Text"], 
                    
                    StyleBox[\(simple\ plot\ of\ y[x] 
                        or\ list\ of\ functions\),
                      "Text"]},
                  {
                    StyleBox[\(ListPlot[list]\),
                      "Text"], 
                    StyleBox[\(plot\ list\ of {x, y}\ points\),
                      "Text"]},
                  {
                    StyleBox[\(ParametricPlot[{x, y}, {t, t\_min, t\_max}]\),
                      "Text"], 
                    StyleBox[\(parametric\ plot\ of\ y[t]\ vs . \ x[t]\),
                      "Text"]},
                  {
                    StyleBox[
                      RowBox[{"DensityPlot", "[", 
                        RowBox[{"f", ",", 
                          RowBox[{"{", 
                            RowBox[{"x", ",", 
                              SubscriptBox[
                                StyleBox["x",
                                  "Text"], "min"], ",", \(x\_max\)}], "}"}], 
                          ",", \({y, y\_min, y\_max}\)}], "]"}],
                      "Text"], 
                    
                    StyleBox[\(2 - 
                        d\ grayscale\ or\ color\ density\ plot\ of\ f[x, y]\),
                      
                      "Text"]},
                  {
                    StyleBox[
                      RowBox[{"ContourPlot", "[", 
                        RowBox[{"f", ",", 
                          RowBox[{"{", 
                            RowBox[{"x", ",", 
                              SubscriptBox[
                                StyleBox["x",
                                  "Text"], "min"], ",", \(x\_max\)}], "}"}], 
                          ",", \({y, y\_min, y\_max}\)}], "]"}],
                      "Text"], 
                    StyleBox[\(2 - d\ \ contour\ plot\ of\ f[x, y]\),
                      "Text"]},
                  {
                    StyleBox[\(ListDensityPlot[list]\),
                      "Text"], 
                    
                    StyleBox[\(density\ plot\ for\ 2\[Times]2\ array\ of\ \
values\),
                      "Text"]},
                  {
                    StyleBox[\(ListContourPlot[list]\),
                      "Text"], 
                    
                    StyleBox[\(contour\ plot\ for\ 2\[Times]2\ array\ of\ \
values\),
                      "Text"]},
                  {
                    StyleBox[
                      RowBox[{"Plot3D", "[", 
                        RowBox[{"z", ",", 
                          RowBox[{"{", 
                            RowBox[{"x", ",", 
                              SubscriptBox[
                                StyleBox["x",
                                  "Text"], "min"], ",", \(x\_max\)}], "}"}], 
                          ",", \({y, y\_min, y\_max}\)}], "]"}],
                      "Text"], 
                    
                    StyleBox[\(3 - 
                        d\ plot\ of\ z[x, y] or\ list\ of\ functions\),
                      "Text"]},
                  {
                    
                    StyleBox[\(ParametricPlot3D[{x, y, z}, {t, t\_min, 
                          t\_max}]\),
                      "Text"], 
                    StyleBox[\(parametric\ plot\ of\ {x[t], y[t]\ , z[t]}\),
                      "Text"]},
                  {
                    StyleBox[\(ListPlot3D[list]\),
                      "Text"], 
                    StyleBox[\(plot\ list\ of {x, y, z}\ points\),
                      "Text"]}
                  },
                ColumnAlignments->{Left},
                GridFrame->True,
                RowLines->{True, False}],
              FontFamily->"Times New Roman",
              FontWeight->"Bold"]}
          }],
      Background->RGBColor[0.500008, 0.996109, 0.996109]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "In addition to the built-in functions, there is also a large number of \
standard packages which either enhance the features of built-in functions or \
provide more specialized plotting formats.  We list some of the more useful \
packages in the table below, but if you do not find what you need here be \
sure to check the appropriate directories for other standard or extra \
packages or ",
  StyleBox["MathSource ",
    FontSlant->"Italic"],
  "for user-written packages.  Each of these packages is located in the \
Graphics directory and includes that directory as a prefix to its name.  For \
example, to use ",
  StyleBox["ImplicitPlot", "InlineInput",
    FontWeight->"Bold"],
  " one would need the command ",
  StyleBox["Needs[\"Graphics`ImplicitPlot`\"]", "InlineInput",
    FontWeight->"Bold"],
  " or to use ",
  StyleBox["Graphics", "InlineInput",
    FontWeight->"Bold"],
  " the repetitive command ",
  StyleBox["Needs[\"Graphics`Graphics`\"]", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[BoxData[
    StyleBox[GridBox[{
          {\(Selected\ Graphics\ Packages\)},
          {GridBox[{
                {
                  StyleBox["package",
                    "Text"], 
                  StyleBox[\(purpose/format\),
                    "Text"]},
                {
                  StyleBox["Graphics",
                    "Text"], 
                  StyleBox[\(large\ collection\ of\ common\ 2 - d\ formats\),
                    "Text"]},
                {
                  StyleBox["Graphics3D",
                    "Text"], 
                  StyleBox[\(large\ collection\ of\ common\ 3 - d\ formats\),
                    "Text"]},
                {
                  StyleBox["ImplicitPlot",
                    "Text"], 
                  
                  StyleBox[\(plot\ equations, \ related\ to\ ParametricPlot\),
                    
                    "Text"]},
                {
                  StyleBox["MultipleListPlot",
                    "Text"], 
                  StyleBox[\(several\ lists\ with\ varying\ formats\),
                    "Text"]},
                {
                  StyleBox["PlotField",
                    "Text"], 
                  StyleBox[\(vector, \ gradient, \ hamiltonian\ fields\),
                    "Text"]},
                {
                  StyleBox["Legend",
                    "Text"], 
                  StyleBox[\(fancy\ legend\ formats\),
                    "Text"]},
                {
                  StyleBox["ParametricPlot3D",
                    "Text"], 
                  StyleBox[\(enhances\ built - in\ function\),
                    "Text"]},
                {
                  StyleBox["ContourPlot3D",
                    "Text"], 
                  StyleBox[\(3 - d\ contour\ plots\),
                    "Text"]},
                {
                  StyleBox["Shapes",
                    "Text"], 
                  StyleBox[\(3 - d\ shapes\ and\ transformations\),
                    "Text"]},
                {
                  StyleBox["Polyhedra",
                    "Text"], 
                  StyleBox[\(create\ and\ manipulate\ polyhedra\),
                    "Text"]}
                },
              ColumnAlignments->{Left},
              GridFrame->True,
              RowLines->{True, False}]}
          }],
      Background->RGBColor[0.500008, 0.996109, 0.996109]]], "Text",
  TextAlignment->Center,
  TextJustification->0,
  FontFamily->"Times New Roman",
  FontWeight->"Bold"],

Cell[TextData[{
  "The easiest way to ensure that the required graphics package is available \
when needed, while avoiding shadows, is to load the graphics master package.  \
Before Version 5.1 one entered ",
  StyleBox["Needs[\"", "InlineInput"],
  StyleBox["Graphics`Master`]", "InlineInput",
    FontWeight->"Bold"],
  " but only ",
  StyleBox["Needs[\"Graphics`\"]", "InlineInput"],
  " thereafter",
  ".  Recall that a master package creates symbols for all important \
functions in any of the packages in its directory.  Thus, it is not necessary \
to remember in which package a particular symbol is found or to load that \
package before using any of the symbols in its jurisdiction.  Nor does this \
method consume resources unnecessarily.  Therefore, we employ this method in \
the initialization section below."
}], "Text"],

Cell["\<\
In this notebook we discuss some of the basic features common to many of the \
plotting functions and provide examples of some of the more common functions, \
but do not attempt to cover everything.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization", "Section"],

Cell[BoxData[{
    \(\(ClearAll["\<Global`*\>"];\)\), "\n", 
    \(\(Off[General::spell, \ General::spell1];\)\), "\n", 
    \(\($TextStyle = {FontFamily \[Rule] "\<Times\>", 
          FontSize \[Rule] 12};\)\)}], "Input"],

Cell["List all symbols with Plot in their name.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?*Plot*\)\)], "Input"],

Cell[BoxData[GridBox[{
        {
          StyleBox["System`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ArrayPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812434-6929725", "System`ArrayPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListContourPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812434-6929725", 
                    "System`ListContourPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812434-6929725", "System`ListPlot3D"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ParametricPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812434-6929725", 
                    "System`ParametricPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot3Matrix",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812434-6929725", 
                    "System`Plot3Matrix"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotLabel",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812434-6929725", "System`PlotLabel"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotRegion",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812434-6929725", "System`PlotRegion"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"]},
              {
                ButtonBox[
                  StyleBox["ContourPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812434-6929725", 
                    "System`ContourPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListDensityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812434-6929725", 
                    "System`ListDensityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["MaxPlotPoints",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812434-6929725", 
                    "System`MaxPlotPoints"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812434-6929725", "System`Plot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotDivision",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812434-6929725", 
                    "System`PlotDivision"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotPoints",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812434-6929725", "System`PlotPoints"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotStyle",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812434-6929725", "System`PlotStyle"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"]},
              {
                ButtonBox[
                  StyleBox["DensityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812434-6929725", 
                    "System`DensityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812434-6929725", "System`ListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ParametricPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812434-6929725", 
                    "System`ParametricPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812434-6929725", "System`Plot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotJoined",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812434-6929725", "System`PlotJoined"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotRange",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812434-6929725", "System`PlotRange"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], ""}
              },
            RowMinHeight->{1, 1, 1.05}]}
        },
      RowSpacings->{2, 3},
      ColumnsEqual->True]], "Print",
  CellMargins->{{20, Inherited}, {Inherited, Inherited}},
  ShowCellLabel->False,
  CellFrameMargins->{{Inherited, Inherited}, {14, 14}},
  Background->RGBColor[0.964706, 0.929412, 0.839216],
  ButtonBoxOptions->{Active->True},
  GridBoxOptions->{ColumnAlignments->{Left}}]
}, Closed]],

Cell["Now load the master graphics package", "Text"],

Cell[BoxData[
    \(Needs["\<Graphics`\>"]\)], "Input"],

Cell["and see that many more symbols related to Plot are found.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?*Plot*\)\)], "Input"],

Cell[BoxData[GridBox[{
        {
          StyleBox["System`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ArrayPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812443-5219848", "System`ArrayPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListContourPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "System`ListContourPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812443-5219848", "System`ListPlot3D"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ParametricPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "System`ParametricPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot3Matrix",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "System`Plot3Matrix"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotLabel",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812443-5219848", "System`PlotLabel"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotRegion",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812443-5219848", "System`PlotRegion"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"]},
              {
                ButtonBox[
                  StyleBox["ContourPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "System`ContourPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListDensityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "System`ListDensityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["MaxPlotPoints",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "System`MaxPlotPoints"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812443-5219848", "System`Plot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotDivision",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "System`PlotDivision"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotPoints",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812443-5219848", "System`PlotPoints"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotStyle",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812443-5219848", "System`PlotStyle"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"]},
              {
                ButtonBox[
                  StyleBox["DensityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "System`DensityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812443-5219848", "System`ListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ParametricPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "System`ParametricPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812443-5219848", "System`Plot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotJoined",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812443-5219848", "System`PlotJoined"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotRange",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812443-5219848", "System`PlotRange"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], ""}
              },
            RowMinHeight->{1, 1, 1.05}]},
        {
          StyleBox["Graphics`Animation`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["MovieContourPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Animation`MovieContourPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Animation`"], 
                ButtonBox[
                  StyleBox["MovieDensityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Animation`MovieDensityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Animation`"], 
                ButtonBox[
                  StyleBox["MovieParametricPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Animation`MovieParametricPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Animation`"], 
                ButtonBox[
                  StyleBox["MoviePlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Animation`MoviePlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Animation`"], 
                ButtonBox[
                  StyleBox["MoviePlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Animation`MoviePlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Animation`"], ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`ContourPlot3D`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ContourPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`ContourPlot3D`ContourPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`ContourPlot3D`"], 
                ButtonBox[
                  StyleBox["ListContourPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`ContourPlot3D`ListContourPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`ContourPlot3D`"], "", "", "", "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`FilledPlot`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["FilledListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`FilledPlot`FilledListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`FilledPlot`"], 
                ButtonBox[
                  StyleBox["FilledPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`FilledPlot`FilledPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`FilledPlot`"], 
                ButtonBox[
                  StyleBox["ListFilledPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`FilledPlot`ListFilledPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`FilledPlot`"], "", "", "", "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`Graphics`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ErrorListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`ErrorListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["LinearLogPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`LinearLogPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["LogLinearPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`LogLinearPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["LogLogPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`LogLogPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["PolarPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`PolarPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["TextListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`TextListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], ""},
              {
                ButtonBox[
                  StyleBox["LabeledListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`LabeledListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["ListAndCurvePlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`ListAndCurvePlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["LogListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`LogListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["LogPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`LogPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["ScaledListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`ScaledListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], "", ""},
              {
                ButtonBox[
                  StyleBox["LinearLogListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`LinearLogListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["LogLinearListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`LogLinearListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["LogLogListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`LogLogListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["PolarListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`PolarListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["ScaledPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics`ScaledPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], "", ""}
              },
            RowMinHeight->{1, 1, 1.05}]},
        {
          StyleBox["Graphics`Graphics3D`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ListShadowPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics3D`ListShadowPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics3D`"], 
                ButtonBox[
                  StyleBox["ListSurfacePlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics3D`ListSurfacePlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics3D`"], 
                ButtonBox[
                  StyleBox["ScatterPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics3D`ScatterPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics3D`"], 
                ButtonBox[
                  StyleBox["ShadowPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Graphics3D`ShadowPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics3D`"], "", "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`ImplicitPlot`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ImplicitPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`ImplicitPlot`ImplicitPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`ImplicitPlot`"], "", "", "", "", "", 
                "", "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`InequalityGraphics`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ComplexInequalityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`InequalityGraphics`ComplexInequalityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`InequalityGraphics`"], 
                ButtonBox[
                  StyleBox["InequalityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`InequalityGraphics`InequalityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`InequalityGraphics`"], 
                ButtonBox[
                  StyleBox["InequalityPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`InequalityGraphics`InequalityPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`InequalityGraphics`"], "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`Legend`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["PlotLegend",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`Legend`PlotLegend"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Legend`"], "", "", "", "", "", "", "",
                 "", "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`MultipleListPlot`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["MultipleListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`MultipleListPlot`MultipleListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`MultipleListPlot`"], 
                ButtonBox[
                  StyleBox["PlotSymbol",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`MultipleListPlot`PlotSymbol"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`MultipleListPlot`"], "", "", "", "", 
                ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`ParametricPlot3D`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["CylindricalPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`ParametricPlot3D`CylindricalPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`ParametricPlot3D`"], 
                ButtonBox[
                  StyleBox["PointParametricPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`ParametricPlot3D`PointParametricPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`ParametricPlot3D`"], 
                ButtonBox[
                  StyleBox["SphericalPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`ParametricPlot3D`SphericalPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`ParametricPlot3D`"], "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`PlotField`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ListPlotVectorField",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`PlotField`ListPlotVectorField"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField`"], 
                ButtonBox[
                  StyleBox["PlotGradientField",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`PlotField`PlotGradientField"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField`"], 
                ButtonBox[
                  StyleBox["PlotHamiltonianField",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`PlotField`PlotHamiltonianField"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField`"], 
                ButtonBox[
                  StyleBox["PlotPolyaField",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`PlotField`PlotPolyaField"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField`"], 
                ButtonBox[
                  StyleBox["PlotVectorField",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`PlotField`PlotVectorField"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField`"], ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`PlotField3D`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ListPlotVectorField3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`PlotField3D`ListPlotVectorField3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField3D`"], 
                ButtonBox[
                  StyleBox["PlotGradientField3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`PlotField3D`PlotGradientField3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField3D`"], 
                ButtonBox[
                  StyleBox["PlotVectorField3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350812443-5219848", 
                    "Graphics`PlotField3D`PlotVectorField3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField3D`"], "", ""}
              },
            RowMinHeight->1.05]}
        },
      RowSpacings->{
        2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 
          3, 2, 3},
      ColumnsEqual->True]], "Print",
  CellMargins->{{20, Inherited}, {Inherited, Inherited}},
  ShowCellLabel->False,
  CellFrameMargins->{{Inherited, Inherited}, {14, 14}},
  Background->RGBColor[0.964706, 0.929412, 0.839216],
  ButtonBoxOptions->{Active->True},
  GridBoxOptions->{ColumnAlignments->{Left}}]
}, Closed]],

Cell["\<\
Many of these symbols are just stubs, but will be fleshed out as needed \
below.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Plotting a function of one variable", "Section"],

Cell[CellGroupData[{

Cell["Basic syntax", "Subsection"],

Cell[TextData[{
  "The basic plotting function is, of course, ",
  ButtonBox["Plot",
    ButtonStyle->"RefGuideLink"],
  " with syntax ",
  StyleBox["Plot[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["f,{x,xmin,xmax},opts", "InlineInput"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " where ",
  StyleBox["f", "InlineInput"],
  StyleBox[" ",
    FontSlant->"Italic"],
  " is an explicit function of the independent variable ",
  StyleBox["x", "InlineInput"],
  " and where ",
  StyleBox["xmin", "InlineInput"],
  StyleBox[" ",
    FontSlant->"Italic"],
  "and ",
  StyleBox["xmax", "InlineInput"],
  " represent the limits of the interval to be plotted.  For example, the \
following command plots the inverse hyperbolic cosine.  Note that we use a \
range ",
  Cell[BoxData[
      \(TraditionalForm\`x \[GreaterEqual] 1\)]],
  " where that function produces real values."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Plot[\ ArcCosh[x], \ {x, 1, 10}]\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0952381 0.0147151 0.196646 [
[.21429 .00222 -5 -15.6875 ]
[.21429 .00222 5 0 ]
[.40476 .00222 -5 -15.6875 ]
[.40476 .00222 5 0 ]
[.59524 .00222 -5 -15.6875 ]
[.59524 .00222 5 0 ]
[.78571 .00222 -5 -15.6875 ]
[.78571 .00222 5 0 ]
[.97619 .00222 -8 -15.6875 ]
[.97619 .00222 8 0 ]
[.01131 .11304 -19 -7.84375 ]
[.01131 .11304 0 7.84375 ]
[.01131 .21136 -10 -7.84375 ]
[.01131 .21136 0 7.84375 ]
[.01131 .30968 -19 -7.84375 ]
[.01131 .30968 0 7.84375 ]
[.01131 .40801 -10 -7.84375 ]
[.01131 .40801 0 7.84375 ]
[.01131 .50633 -19 -7.84375 ]
[.01131 .50633 0 7.84375 ]
[.01131 .60465 -10 -7.84375 ]
[.01131 .60465 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .01472 m
.21429 .02097 L
s
gsave
.21429 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.40476 .01472 m
.40476 .02097 L
s
gsave
.40476 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.59524 .01472 m
.59524 .02097 L
s
gsave
.59524 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.78571 .01472 m
.78571 .02097 L
s
gsave
.78571 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(8) show
1.000 setlinewidth
grestore
.97619 .01472 m
.97619 .02097 L
s
gsave
.97619 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .01472 m
.07143 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.16667 .01472 m
.16667 .01847 L
s
.2619 .01472 m
.2619 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.35714 .01472 m
.35714 .01847 L
s
.45238 .01472 m
.45238 .01847 L
s
.5 .01472 m
.5 .01847 L
s
.54762 .01472 m
.54762 .01847 L
s
.64286 .01472 m
.64286 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.7381 .01472 m
.7381 .01847 L
s
.83333 .01472 m
.83333 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.92857 .01472 m
.92857 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .11304 m
.03006 .11304 L
s
gsave
.01131 .11304 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .21136 m
.03006 .21136 L
s
gsave
.01131 .21136 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .30968 m
.03006 .30968 L
s
gsave
.01131 .30968 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.5) show
1.000 setlinewidth
grestore
.02381 .40801 m
.03006 .40801 L
s
gsave
.01131 .40801 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.02381 .50633 m
.03006 .50633 L
s
gsave
.01131 .50633 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2.5) show
1.000 setlinewidth
grestore
.02381 .60465 m
.03006 .60465 L
s
gsave
.01131 .60465 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .03438 m
.02756 .03438 L
s
.02381 .05404 m
.02756 .05404 L
s
.02381 .07371 m
.02756 .07371 L
s
.02381 .09337 m
.02756 .09337 L
s
.02381 .1327 m
.02756 .1327 L
s
.02381 .15237 m
.02756 .15237 L
s
.02381 .17203 m
.02756 .17203 L
s
.02381 .1917 m
.02756 .1917 L
s
.02381 .23103 m
.02756 .23103 L
s
.02381 .25069 m
.02756 .25069 L
s
.02381 .27035 m
.02756 .27035 L
s
.02381 .29002 m
.02756 .29002 L
s
.02381 .32935 m
.02756 .32935 L
s
.02381 .34901 m
.02756 .34901 L
s
.02381 .36868 m
.02756 .36868 L
s
.02381 .38834 m
.02756 .38834 L
s
.02381 .42767 m
.02756 .42767 L
s
.02381 .44734 m
.02756 .44734 L
s
.02381 .467 m
.02756 .467 L
s
.02381 .48666 m
.02756 .48666 L
s
.02381 .52599 m
.02756 .52599 L
s
.02381 .54566 m
.02756 .54566 L
s
.02381 .56532 m
.02756 .56532 L
s
.02381 .58499 m
.02756 .58499 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.11905 .01489 m
.12011 .044 L
.12107 .05515 L
.12323 .07278 L
.12523 .08521 L
.12713 .09515 L
.13597 .13029 L
.14453 .15554 L
.15382 .17802 L
.19174 .2444 L
.22852 .28987 L
.26394 .32464 L
.29799 .35273 L
.33425 .37858 L
.36914 .40047 L
.40624 .42127 L
.44198 .43936 L
.47635 .45528 L
.51294 .47089 L
.54815 .48481 L
.58558 .49858 L
.62164 .51098 L
.65633 .52221 L
.69323 .53348 L
.72877 .54375 L
.76295 .55315 L
.79933 .56267 L
.83435 .57142 L
.87157 .58031 L
.90743 .5885 L
.94193 .59608 L
.97619 .60332 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00L03oool00`2oM0000000000000040000000304RLg`3oool0oooo04<0oooo00@0oooO09a8
0000B9`0gooo@`3oool01@3ogi`0B000000000000000M;oo04<0oooo00D0gia80000000000000000
04RLg`100?ooo`060;md00000000000000000000B9`0gooo0`3oool01@3ogi`0B0000000000007@0
_ooo0140oooo001a0?ooo`030=nLB0000000000000<0000000<004RL0=ooo`3oool0@`3oool0103o
oml0W4P00018W03Ooom20?ooo`070?oOW0180000B9cO0?ooo`3ookl0M00007Boo`110?ooo`070=nL
B0000000M;oo0?ooo`3ooml0W4P004RLg`100?ooo`030?oOW0180000M;oo00@0oooo00L0oono07@0
001d_ol0oooo0?noM00007@0_ooo0100oooo001b0?ooo`030=nLB018W=l0oooo0080oooo00<0giad
09cOo`3oool0@`3oool0103ooml0W4P00018W03Ooom10?ooo`030?oog`2LB00007Bo00<0oooo00<0
okmd0018W03Oool0?`3oool00`3ooml0W4P0001d_`030?ooo`030?noM000B9`0gooo03l0oooo00<0
omnL04P0001d_ol0103oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00l0oooo001c
0?ooo`030=nLB01d_ol0oooo04<0oooo00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`0n
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0@03oool00`2oM00007Bo0?ooo`02
0?ooo`030;md0000B9`0gooo03l0oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M00004RLg`03
0?ooo`030=nLB00007@0_ooo00l0oooo001d0?ooo`030;mdB02Lgol0oooo0480oooo00P0omnL07BL
g`3oool0oooo0?oog`2LB00004RL0=oood40oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`100?ooo`070?oOW01807@0_ooo0?ooo`2oM000000007Boo`100?ooo`030?oOW0180000M;oo
00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0?0?ooo`00M03oool00`3ooml0
W4Qd0;ooo`130?ooo`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`110?ooo`030?oOW0180000
M;oo00<0oooo00<0_g@00000B02Lgol0@@3oool01@3ookl0M4Q804P000000000B9cO0440oooo00<0
omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00l0oooo001e
0?ooo`030?noM000M;l0oooo0480oooo00L0oooO09a8M02oool0oooO09a80000B9`0gooo0440oooo
00P0oono07@000000000B9cO0?ooo`3o_g@00000001d_d80oooo00@0omnL04P00000000007Bo@P3o
ool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol03`3o
ool007D0oooo00<0oono07@0B02Lgol0@`3oool01P3ogi`0B7Bo0?oog`2LB00004RL0=oood80oooo
00<0_g@00018W02oW4P00P0000000`00B9`0gooo0?ooo`100?ooo`060?oOW0180000001d0;ooo`3o
_g@007Bo@@3oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol03`3oool007D0oooo00<0oono07@00018W=l0A03oool01@3OW4P0M;oO09a80000B9`0gooo
0480oooo00<0omnL04P0001d_ol0A@3oool00`2oM000001d0;ooo`020?ooo`030;md0018W=l0oooo
03l0oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo
00l0oooo001`0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l0A03oool01@3ooml0
W4Q804Q80000B9`0gooo04<0oooo00<0gia80000B02Lgol0A03oool00`2oM00004RL0=ooo`020?oo
o`030=nLB000M;l0oooo03l0oooo00<0omnL04P0001d_ol0103oool00`3OW4P004RL0=ooo`020?oo
o`030;md0000M;l0oooo00l0oooo001`0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02o
oom50?ooo`040?oOW018000004RL0=oood@0oooo00<0okmd00000000M;l0@`3oool01`3o_g@00000
07Boo`3oool0oono07@0001d_ol0?P3oool01@3ooml0W4P0000000000000M;oo00@0oooo00L0oono
07@0001d_ol0oooo0?noM00007@0_ooo0100oooo001a0?ooo`060?oog`2LB00000000000000007@0
_oooA`3oool00`3OW4P004RL0=ooo`150?ooo`050?oog`2LB00000000000M02oool0@@3oool01@3o
_g@00000000000000000B9cO0440oooo00<0okmd0000001d_ol01@3oool01@3ookl0M00000000000
07@0_ooo0140oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
000[0?ooo`030000003oool0oooo0?l0ooooP`3oool002/0oooo00<000000?ooo`3oool0o`3ooon3
0?ooo`00:`3oool00`000000oooo0?ooo`3o0?oooh<0oooo000[0?ooo`030000003oool0oooo0?l0
ooooP`3oool002/0oooo00<000000?ooo`3oool0o`3ooon30?ooo`00:`3oool00`000000oooo0?oo
o`3o0?oooh<0oooo000R0?ooool00000P@00000>0?ooo`00:`3oool00`000000oooo0?ooo`0@0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo0100oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0@0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo0100oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0@0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool05@3oool002/0oooo00<000000?ooo`3o
ool08P3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3o
ool0AP3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3o
ool05@3oool002/0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`3o0?oooeh0oooo
000[0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0o`3ooomN0?ooo`00:`3oool0
0`000000oooo0?ooo`0R0?ooo`030000003oool0oooo0?l0ooooGP3oool002/0oooo00<000000?oo
o`3oool08P3oool00`000000oooo0?ooo`3o0?oooeh0oooo000[0?ooo`<000008P3oool00`000000
oooo0?ooo`3o0?oooeh0oooo000[0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0
o`3ooomN0?ooo`00:`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo0?l0ooooGP3o
ool002/0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`3o0?oooeh0oooo000[0?oo
o`030000003oool0oooo0280oooo00<000000?ooo`3oool0o`3ooomN0?ooo`00:`3oool00`000000
oooo0?ooo`0R0?ooo`030000003oool0oooo0?l0ooooGP3oool002/0oooo00<000000?ooo`3oool0
8P3oool00`000000oooo0?ooo`3o0?oooeh0oooo000[0?ooo`030000003oool0oooo02<0oooo00<0
00000?ooo`3oool0o`3ooomM0?ooo`00:`3oool3000002<0oooo00<000000?ooo`3oool0o`3ooomM
0?ooo`00:`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0?l0ooooG@3oool002/0
oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`3o0?oooed0oooo000[0?ooo`030000
003oool0oooo02<0oooo00<000000?ooo`3oool0o`3ooomM0?ooo`00:`3oool00`000000oooo0?oo
o`0S0?ooo`030000003oool0oooo0?l0ooooG@3oool002/0oooo00<000000?ooo`3oool08`3oool0
0`000000oooo0?ooo`3o0?oooed0oooo000[0?ooo`030000003oool0oooo02<0oooo00<000000?oo
o`3oool0o`3ooomM0?ooo`00:`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0?l0
ooooG@3oool002/0oooo0`00000S0?ooo`030000003oool0oooo0?l0ooooG@3oool002/0oooo00<0
00000?ooo`3oool08`3oool00`000000oooo0?ooo`3o0?oooed0oooo000[0?ooo`030000003oool0
oooo02@0oooo00<000000?ooo`3oool0o`3ooomL0?ooo`00:`3oool00`000000oooo0?ooo`0T0?oo
o`030000003oool0oooo0?l0ooooG03oool002/0oooo00<000000?ooo`3oool0903oool00`000000
oooo0?ooo`3o0?oooe`0oooo000[0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0
o`3ooomL0?ooo`00:`3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo0?l0ooooG03o
ool002/0oooo0`00000U0?ooo`030000003oool0oooo0?l0ooooF`3oool002/0oooo00<000000?oo
o`3oool09@3oool00`000000oooo0?ooo`3o0?oooe/0oooo000[0?ooo`030000003oool0oooo02D0
oooo00<000000?ooo`3oool0o`3ooomK0?ooo`003`3oool01@3ogi`0B0000000000007@0_ooo00<0
oooo00H0okmd0000001d_ol0oooo0?oog`2LB0030000000307Boo`3oool0oooo00P0oooo00<00000
0?ooo`3oool09@3oool00`000000oooo0?ooo`3o0?oooe/0oooo000>0?ooo`070?oo_`1d0000M;oo
0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo0?ooo`3ooml0W4Qd0;ooo`020?ooo`03
0;md001d_ol0oooo00P0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`3o0?oooeX0
oooo000>0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd
0;ooo`080?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0o`3ooomJ0?ooo`003@3o
ool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00/0oooo00<0_g@007Boo`3oool01`3o
ool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0?l0ooooFP3oool000d0oooo00<0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030?oog`2LB000B9cO00P0oooo00<00000
0?ooo`3oool09P3oool00`000000oooo0?ooo`3o0?oooeX0oooo000=0?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol02P3oool00`3OW4P0000007Boo`080?ooo`@000009P3oool00`00
0000oooo0?ooo`3o0?oooeT0oooo000=0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol01P3oool01`3Ogml0oooo0?oo_`1d000000000000M02oool0203oool00`000000oooo0?ooo`0W
0?ooo`030000003oool0oooo0?l0ooooF@3oool000d0oooo00<0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`060?ooo`060=nLB000000000000000000004P0W=oo2@3oool00`000000oooo0?oo
o`0W0?ooo`030000003oool0oooo0?l0ooooF@3oool000d0oooo00<0oono07@00018W=l00`3oool0
0`3OW4P0001d0;ooo`060?ooo`040?oo_`1d0000000004RLg`/0oooo00<000000?ooo`3oool09`3o
ool00`000000oooo0?ooo`3o0?oooeT0oooo000>0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0
001d_`3oool01`3oool00`2oM000M;oo0?ooo`0;0?ooo`030000003oool0oooo02P0oooo00<00000
0?ooo`3oool0o`3ooomH0?ooo`003P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0203o
ool01P3ogi`0B000000000000000001809cOo`P0oooo00<000000?ooo`3oool0:03oool00`000000
oooo0?ooo`3o0?oooeP0oooo000?0?ooo`050?oo_`1d000000000000M02oool02P3oool01@2oM000
0000000000000000B9cO00P0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`3o0?oo
oeP0oooo000[0?ooo`<00000:03oool00`000000oooo0?ooo`3o0?oooeP0oooo000[0?ooo`030000
003oool0oooo02T0oooo00<000000?ooo`3oool0o`3ooomG0?ooo`00:`3oool00`000000oooo0?oo
o`0Y0?ooo`030000003oool0oooo0?l0ooooE`3oool002/0oooo00<000000?ooo`3oool0:P3oool0
0`000000oooo0?ooo`3o0?oooeH0oooo000[0?ooo`030000003oool0oooo02X0oooo00<000000?oo
o`3oool0o`3ooomF0?ooo`00:`3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo0?l0
ooooEP3oool002/0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`3o0?oooeD0oooo
000[0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0o`3ooomE0?ooo`00:`3oool3
000002`0oooo00<000000?ooo`3oool0o`3ooomD0?ooo`00:`3oool00`000000oooo0?ooo`0/0?oo
o`030000003oool0oooo0?l0ooooE03oool002/0oooo00<000000?ooo`3oool0;03oool00`000000
oooo0?ooo`3o0?oooe@0oooo000[0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0
o`3ooomC0?ooo`00:`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo0?l0ooooD`3o
ool002/0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`3o0?oooe<0oooo000[0?oo
o`030000003oool0oooo02h0oooo00<000000?ooo`3oool0o`3ooomB0?ooo`00:`3oool3000002h0
oooo00<000000?ooo`3oool0o`3ooomB0?ooo`00:`3oool00`000000oooo0?ooo`0_0?ooo`030000
003oool0oooo0?l0ooooD@3oool002/0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?oo
o`3o0?oooe40oooo000[0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0o`3ooom@
0?ooo`00:`3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo0?l0ooooD03oool002/0
oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`3o0?ooodl0oooo000[0?ooo`030000
003oool0oooo0340oooo00<000000?ooo`3oool0o`3ooom?0?ooo`00:`3oool00`000000oooo0?oo
o`0b0?ooo`030000003oool0oooo0?l0ooooCP3oool002/0oooo0`00000b0?ooo`030000003oool0
oooo0?l0ooooCP3oool002/0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`3o0?oo
odd0oooo000M0?ooo`060;md00000000000000000000B9`0gooo203oool00`000000oooo0?ooo`0d
0?ooo`030000003oool0oooo0?l0ooooC03oool001h0oooo00<0omnL04P0001d_ol02P3oool00`00
0000oooo0?ooo`0d0?ooo`030000003oool0oooo0?l0ooooC03oool001h0oooo00<0omnL04P0001d
_ol02P3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo0?l0ooooB`3oool001h0oooo
00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo0?l0oooo
B`3oool001h0oooo00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`0f0?ooo`030000003o
ool0oooo0?l0ooooBP3oool001h0oooo00<0omnL04P0001d_ol02P3oool4000003D0oooo00<00000
0?ooo`3oool0o`3ooom:0?ooo`007P3oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo
03L0oooo00<000000?ooo`3oool0o`3ooom90?ooo`007P3oool00`3ogi`0B00007Boo`0:0?ooo`03
0000003oool0oooo03P0oooo00<000000?ooo`3oool0o`3ooom80?ooo`007P3oool00`3ogi`0B000
07Boo`0:0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool0o`3ooom80?ooo`007P3o
ool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0o`3o
oom70?ooo`00703oool01@3ooml0W4P0000000000000M;oo00X0oooo00<000000?ooo`3oool0>@3o
ool00`000000oooo0?ooo`3o0?ooodL0oooo000N0?ooo`030?noM0000000M;oo00X0oooo00<00000
0?ooo`3oool0>P3oool00`000000oooo0?ooo`3o0?ooodH0oooo000[0?ooo`030000003oool0oooo
03/0oooo00<000000?ooo`3oool0o`3ooom50?ooo`00:`3oool3000003/0oooo00<000000?ooo`3o
ool0o`3ooom50?ooo`00:`3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo0?l0oooo
A03oool002/0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`3o0?oood@0oooo000[
0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool0o`3ooom30?ooo`00:`3oool00`00
0000oooo0?ooo`0m0?ooo`030000003oool0oooo0?l0oooo@`3oool002/0oooo00<000000?ooo`3o
ool0?P3oool00`000000oooo0?ooo`3o0?oood80oooo000[0?ooo`030000003oool0oooo03l0oooo
00<000000?ooo`3oool0o`3ooom10?ooo`00:`3oool300000400oooo00<000000?ooo`3oool0o`3o
oom00?ooo`00:`3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo0?l0oooo@03oool0
02/0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`3o0?ooocl0oooo000[0?ooo`03
0000003oool0oooo0480oooo00<000000?ooo`3oool0o`3oooln0?ooo`00:`3oool00`000000oooo
0?ooo`130?ooo`030000003oool0oooo0?l0oooo?@3oool002/0oooo00<000000?ooo`3oool0A03o
ool00`000000oooo0?ooo`3o0?oooc`0oooo000[0?ooo`030000003oool0oooo04D0oooo00<00000
0?ooo`3oool0o`3ooolk0?ooo`00:`3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo
0?l0oooo>`3oool002/0oooo0`0000160?ooo`030000003oool0oooo0?l0oooo>P3oool002/0oooo
00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`3o0?ooocT0oooo000[0?ooo`030000003o
ool0oooo04P0oooo00<000000?ooo`3oool0o`3ooolh0?ooo`00:`3oool00`000000oooo0?ooo`19
0?ooo`030000003oool0oooo0?l0oooo=`3oool002/0oooo00<000000?ooo`3oool0BP3oool00`00
0000oooo0?ooo`3o0?ooocH0oooo000[0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3o
ool0o`3ooolf0?ooo`00:`3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo0?l0oooo
=@3oool002/0oooo0`00001<0?ooo`030000003oool0oooo0?l0oooo=03oool002/0oooo00<00000
0?ooo`3oool0C@3oool00`000000oooo0?ooo`3o0?oooc<0oooo000[0?ooo`030000003oool0oooo
04h0oooo00<000000?ooo`3oool0o`3ooolb0?ooo`003`3oool01P2oM000000000000000000004RL
0=ooo`80oooo00H0okmd0000001d_ol0oooo0?oog`2LB0030000000307Boo`3oool0oooo00P0oooo
00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`3o0?oooc40oooo000@0?ooo`030?oOW018
0000M;oo00@0oooo00L0okmd0000001d_ol0oooo0?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3o
ool0203oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo0?l0oooo<03oool00100oooo
00<0omnL04P0001d_ol03@3oool00`3ooml0W4Qd0;ooo`080?ooo`030000003oool0oooo0540oooo
00<000000?ooo`3oool0o`3oool_0?ooo`00403oool00`3ogi`0B00007Boo`0>0?ooo`030;md001d
_ol0oooo00L0oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`3o0?ooobh0oooo000@
0?ooo`030?oOW0180000M;oo00d0oooo00<0oooO09a80018W=l0203oool00`000000oooo0?ooo`1B
0?ooo`030000003oool0oooo0?l0oooo;P3oool00100oooo00<0omnL04P0001d_ol03@3oool00`3O
W4P0000007Boo`080?ooo`@00000DP3oool00`000000oooo0?ooo`3o0?ooobd0oooo000@0?ooo`03
0?oOW0180000M;oo00T0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00P0oooo00<00000
0?ooo`3oool0E03oool00`000000oooo0?ooo`3o0?ooob`0oooo000@0?ooo`030?oOW0180000M;oo
00T0oooo00H0gia800000000000000000000B02Lgol90?ooo`030000003oool0oooo05D0oooo00<0
00000?ooo`3oool0o`3oool[0?ooo`00403oool00`3ogi`0B00007Boo`090?ooo`040?oo_`1d0000
000004RLg`/0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`3o0?ooobX0oooo000@
0?ooo`030?oOW0180000M;oo00X0oooo00<0_g@007Boo`3oool02`3oool00`000000oooo0?ooo`1G
0?ooo`030000003oool0oooo0?l0oooo:@3oool000h0oooo00D0oooO09a800000000000007Boo`0:
0?ooo`060?oOW018000000000000000004P0W=oo203oool00`000000oooo0?ooo`1H0?ooo`030000
003oool0oooo0?l0oooo:03oool00100oooo00<0okmd0000001d_ol02`3oool01@2oM00000000000
00000000B9cO00P0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`3o0?ooobL0oooo
000[0?ooo`<00000FP3oool00`000000oooo0?ooo`3o0?ooobH0oooo000[0?ooo`030000003oool0
oooo05/0oooo0P00003o0?ooobH0oooo000[0?ooo`030000003oool0oooo05d0oooo00<000000?oo
o`3oool0o`3ooolS0?ooo`00:`3oool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo0?l0
oooo8P3oool002/0oooo00<000000?ooo`3oool0G`3oool200000?l0oooo8P3oool002/0oooo00<0
00000?ooo`3oool0H@3oool00`000000oooo0?ooo`3o0?oooal0oooo000[0?ooo`030000003oool0
oooo0680oooo00<000000?ooo`3oool0o`3ooolN0?ooo`00:`3oool00`000000oooo0?ooo`1S0?oo
o`800000o`3ooolN0?ooo`00:`3oool3000006D0oooo00<000000?ooo`3oool0o`3ooolK0?ooo`00
:`3oool00`000000oooo0?ooo`1V0?ooo`030000003oool0oooo0?l0oooo6P3oool002/0oooo00<0
00000?ooo`3oool0I`3oool200000?l0oooo6P3oool002/0oooo00<000000?ooo`3oool0J@3oool0
0`000000oooo0?ooo`3o0?oooaL0oooo000[0?ooo`030000003oool0oooo06X0oooo00<000000?oo
o`3oool0o`3ooolF0?ooo`00:`3oool00`000000oooo0?ooo`1[0?ooo`800000o`3ooolF0?ooo`00
:`3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo0?l0oooo4`3oool002/0oooo0`00
001^0?ooo`800000o`3ooolC0?ooo`00:`3oool00`000000oooo0?ooo`1`0?ooo`030000003oool0
oooo0?l0oooo403oool002/0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`3o0?oo
o`l0oooo000[0?ooo`030000003oool0oooo0780oooo0P00003o0?ooo`l0oooo000[0?ooo`030000
003oool0oooo07@0oooo00<000000?ooo`3oool0o`3oool<0?ooo`00:`3oool00`000000oooo0?oo
o`1e0?ooo`800000o`3oool<0?ooo`00:`3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0
oooo0?l0oooo2@3oool002/0oooo00<000000?ooo`3oool0N03oool200000?l0oooo2@3oool002/0
oooo0`00001j0?ooo`030000003oool0oooo0?l0oooo1P3oool002/0oooo00<000000?ooo`3oool0
N`3oool200000?l0oooo1P3oool002/0oooo00<000000?ooo`3oool0O@3oool200000?l0oooo103o
ool001/0oooo00<0_g@0000000000000100000000`18W=l0oooo0?ooo`060?ooo`030000003oool0
oooo07l0oooo00<000000?ooo`3oool0o`3oool10?ooo`00703oool00`3OW4P00000000000030000
00030018W03Oool0oooo00H0oooo00<000000?ooo`3oool0P03oool200000?l0oooo0@3oool001d0
oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`060?ooo`030000003oool0oooo0880
oooo00<000000?ooo`3oool0o@3oool001h0oooo00<0gia807Boo`3oool02P3oool00`000000oooo
0?ooo`230?ooo`800000o@3oool001l0oooo00<0_gA809cOo`3oool02@3oool4000008@0oooo0P00
003k0?ooo`007`3oool00`3ooml0W4Qd0;ooo`090?ooo`030000003oool0oooo08L0oooo0P00003i
0?ooo`00803oool00`3o_g@007Bo0?ooo`080?ooo`030000003oool0oooo08T0oooo00<000000?oo
o`3oool0mP3oool00200oooo00<0oono07@0B02Lgol0203oool00`000000oooo0?ooo`2:0?ooo`80
0000mP3oool00200oooo00<0oono07@00018W=l0203oool00`000000oooo0?ooo`2<0?ooo`800000
m03oool001/0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`080?ooo`030000003o
ool0oooo08h0oooo0P00003b0?ooo`006`3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0
_ooo203oool00`000000oooo0?ooo`2@0?ooo`800000l03oool001`0oooo00H0oooO09a800000000
00000000M02oool90?ooo`030000003oool0oooo0980oooo0P00003^0?ooo`00:`3oool3000009@0
oooo0P00003/0?ooo`00:`3oool00`000000oooo0?ooo`2F0?ooo`800000jP3oool002/0oooo00<0
00000?ooo`3oool0V03oool200000>P0oooo000[0?ooo`030000003oool0oooo09X0oooo0P00003V
0?ooo`00:`3oool00`000000oooo0?ooo`2L0?ooo`800000i03oool002/0oooo00<000000?ooo`3o
ool0WP3oool200000>80oooo000[0?ooo`030000003oool0oooo0:00oooo0P00003P0?ooo`00:`3o
ool00`000000oooo0?ooo`2R0?ooo`800000gP3oool002/0oooo0`00002T0?ooo`800000g03oool0
02/0oooo00<000000?ooo`3oool0YP3oool200000=X0oooo000[0?ooo`030000003oool0oooo0:P0
oooo0P00003H0?ooo`00:`3oool00`000000oooo0?ooo`2Z0?ooo`<00000e@3oool002/0oooo00<0
00000?ooo`3oool0[@3oool200000=<0oooo000[0?ooo`030000003oool0oooo0:l0oooo0P00003A
0?ooo`00:`3oool00`000000oooo0?ooo`2a0?ooo`<00000cP3oool002/0oooo0`00002d0?ooo`80
0000c03oool002/0oooo00<000000?ooo`3oool0]P3oool200000<X0oooo000[0?ooo`030000003o
ool0oooo0;P0oooo0`0000370?ooo`00:`3oool00`000000oooo0?ooo`2k0?ooo`800000a@3oool0
02/0oooo00<000000?ooo`3oool0_@3oool300000<80oooo000[0?ooo`030000003oool0oooo0<00
oooo0`00002o0?ooo`00:`3oool00`000000oooo0?ooo`330?ooo`800000_@3oool002/0oooo00<0
00000?ooo`3oool0a@3oool300000;X0oooo000[0?ooo`<00000b03oool300000;L0oooo000[0?oo
o`030000003oool0oooo0</0oooo0P00002e0?ooo`003@3oool00`2oM00000000000000400000009
04RLg`3oool0oooo0?noM0000000M;oo0?ooo`3ooml0W4P000<0000000<0M;oo0?ooo`3oool0203o
ool00`000000oooo0?ooo`3=0?ooo`<00000/P3oool000h0oooo00<0gia80000000000000`000000
2P00B9`0gooo0?ooo`3o_g@0000007Boo`3oool0oooO09a8M02oool20?ooo`030;md001d_ol0oooo
00P0oooo00<000000?ooo`3oool0d03oool300000:l0oooo000?0?ooo`030=nLB018W=l0oooo0080
oooo00<0giad09cOo`3oool02@3oool00`3ooml0W4Qd0;ooo`080?ooo`030000003oool0oooo0=<0
oooo0`00002/0?ooo`00403oool00`3OW4P0M;oo0?ooo`0>0?ooo`030;md001d_ol0oooo00L0oooo
00<000000?ooo`3oool0eP3oool300000:T0oooo000A0?ooo`030;mdB02Lgol0oooo00`0oooo00<0
oooO09a80018W=l0203oool00`000000oooo0?ooo`3I0?ooo`800000Y`3oool00140oooo00<0oooO
09a8M02oool0303oool00`3OW4P0000007Boo`080?ooo`@00000fP3oool300000:@0oooo000B0?oo
o`030?noM000M;l0oooo00L0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00P0oooo00<0
00000?ooo`3oool0gP3oool300000:40oooo000B0?ooo`030?oo_`1d04P0W=oo00L0oooo00H0gia8
00000000000000000000B02Lgol90?ooo`030000003oool0oooo0>40oooo0`00002N0?ooo`004P3o
ool00`3ookl0M00004RLg`070?ooo`040?oo_`1d0000000004RLg`/0oooo00<000000?ooo`3oool0
i03oool3000009/0oooo000=0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l0203o
ool00`2oM000M;oo0?ooo`0;0?ooo`030000003oool0oooo0>L0oooo1000002G0?ooo`003@3oool0
203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo203oool01P3ogi`0B0000000000000000018
09cOo`P0oooo00<000000?ooo`3oool0j`3oool3000009@0oooo000>0?ooo`060?oog`2LB0000000
0000000007@0_ooo2P3oool01@2oM0000000000000000000B9cO00P0oooo00<000000?ooo`3oool0
kP3oool300000940oooo000[0?ooo`030000003oool0oooo0?40oooo1000002=0?ooo`00:`3oool3
00000?D0oooo100000290?ooo`00:`3oool00`000000oooo0?ooo`3i0?ooo`@00000Q@3oool002/0
oooo00<000000?ooo`3oool0o@3oool300000880oooo000[0?ooo`030000003oool0oooo0?l0oooo
0@3oool3000007l0oooo000[0?ooo`030000003oool0oooo0?l0oooo103oool4000007/0oooo000[
0?ooo`030000003oool0oooo0?l0oooo203oool3000007P0oooo000[0?ooo`030000003oool0oooo
0?l0oooo2`3oool3000007D0oooo000[0?ooo`<00000o`3oool>0?ooo`<00000LP3oool002/0oooo
00<000000?ooo`3oool0o`3ooolA0?ooo`@00000KP3oool002/0oooo00<000000?ooo`3oool0o`3o
oolE0?ooo`<00000J`3oool002/0oooo00<000000?ooo`3oool0o`3ooolH0?ooo`<00000J03oool0
02/0oooo00<000000?ooo`3oool0o`3ooolK0?ooo`@00000I03oool002/0oooo00<000000?ooo`3o
ool0o`3ooolO0?ooo`<00000H@3oool002/0oooo00<000000?ooo`3oool0o`3ooolR0?ooo`@00000
G@3oool002/0oooo00<000000?ooo`3oool0o`3ooolV0?ooo`@00000F@3oool002/0oooo0`00003o
0?ooobX0oooo1000001E0?ooo`00:`3oool00`000000oooo0?ooo`3o0?ooobh0oooo1@00001@0?oo
o`00:`3oool00`000000oooo0?ooo`3o0?oooc<0oooo1@00001;0?ooo`00:`3oool00`000000oooo
0?ooo`3o0?ooocP0oooo100000170?ooo`00:`3oool00`000000oooo0?ooo`3o0?oooc`0oooo1@00
00120?ooo`00:`3oool00`000000oooo0?ooo`3o0?oood40oooo1000000n0?ooo`00:`3oool00`00
0000oooo0?ooo`3o0?ooodD0oooo1000000j0?ooo`00:`3oool300000?l0ooooB@3oool3000003L0
oooo000[0?ooo`030000003oool0oooo0?l0ooooC03oool4000003<0oooo000[0?ooo`030000003o
ool0oooo0?l0ooooD03oool5000002h0oooo000K0?ooo`030?oo_`1d000000000080000000<0M;oo
0?ooo`3oool0203oool00`000000oooo0?ooo`3o0?oooeD0oooo1P00000X0?ooo`006`3oool00`3o
gi`0B00004RLg`020?ooo`030;md0018W=l0oooo00P0oooo00<000000?ooo`3oool0o`3ooomK0?oo
o`H000008P3oool00200oooo00<0oono07@0B02Lgol0203oool00`000000oooo0?ooo`3o0?ooof40
oooo1000000N0?ooo`00803oool00`3ookl0M00007Boo`080?ooo`030000003oool0oooo0?l0oooo
I@3oool4000001X0oooo000P0?ooo`030?noM0000000M;oo00P0oooo00<000000?ooo`3oool0o`3o
oomY0?ooo`<000005`3oool001l0oooo00@0omnL04P0000004P0W=oo203oool400000?l0ooooPP3o
ool001d0oooo00D0okmd000000000000001809cOo`090?ooo`030000003oool0oooo0?l0ooooP`3o
ool001l0oooo00<0_g@0001d_`3oool02@3oool00`000000oooo0?ooo`3o0?oooh<0oooo000P0?oo
o`030;md0018W=l0oooo00P0oooo00<000000?ooo`3oool0o`3ooon30?ooo`006`3oool00`3ooml0
W9bo0?ooo`020?ooo`030=nLB00007@0_ooo00P0oooo00<000000?ooo`3oool0o`3ooon30?ooo`00
703oool01P3o_g@0B9cO0?ooo`3ookl0M000001d_`T0oooo00<000000?ooo`3oool0o`3ooon30?oo
o`007@3oool01@3OW4P000000000000007@0_ooo0?l0ooooS`3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.20388, -0.376202, \
0.0411503, 0.0199296}}],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Closed]],

Cell[TextData[{
  "Notice that the ",
  StyleBox["Plot", "InlineInput"],
  " command produced two output cells.  The first contains the plot itself \
and is displayed on the monitor.  The second carries an output label and \
contains the relatively useless message ",
  StyleBox["\[SkeletonIndicator]Graphics\[SkeletonIndicator]", 
    "InlineInput"],
  ".  The purpose of this message is to report that Plot has produced a ",
  StyleBox["Graphics", "InlineInput"],
  " object.  If we had assigned that object a name, we could then apply \
suitable transformation or display functions to that graphics object.  We \
will do this with other plots shortly.  For the moment, it is sufficient to \
recognize that the output cell can be suppressed by terminating the plot \
command with a semi-colon.  The graphics object is still produced, but the \
annoying message is no longer printed.  The choice of suppressing or printing \
this message is, of course, a matter of personal taste \[LongDash] I almost \
always suppress it."
}], "Text"],

Cell[TextData[{
  "Plot ",
  Cell[BoxData[
      \(TraditionalForm\`x\^3\/\(Exp[x] + 1\)\)]],
  " for positive x."
}], "Exercise"],

Cell["\<\
Plot a) BesselJ[5,x] and b) BesselJ[6,x] for a range in x that is symmetric \
around 0.\
\>", "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Function evaluation", "Subsection"],

Cell["\<\
It is critically important that the function to be plotted produces a real \
number when its variable is replaced by a numerical value.  If your function \
contains unevaluated parameters, it cannot be plotted.  Perhaps the most \
common error encountered using plotting commands is to leave some constant in \
symbolic form.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Cos[k\ x], \ {x, \(-1\), 1}];\)\)], "Input"],

Cell[BoxData[
    RowBox[{\(Plot::"plnr"\), \(\(:\)\(\ \)\), "\<\"\\!\\(Cos[\\(\\(k\\\\ \
x\\)\\)]\\) is not a machine-size real number at \\!\\(x\\) = \
\\!\\(-0.9999999166666667`\\). \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"Plot::plnr\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(Plot::"plnr"\), \(\(:\)\(\ \)\), "\<\"\\!\\(Cos[\\(\\(k\\\\ \
x\\)\\)]\\) is not a machine-size real number at \\!\\(x\\) = \
\\!\\(-0.9188660168541684`\\). \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"Plot::plnr\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(Plot::"plnr"\), \(\(:\)\(\ \)\), "\<\"\\!\\(Cos[\\(\\(k\\\\ \
x\\)\\)]\\) is not a machine-size real number at \\!\\(x\\) = \
\\!\\(-0.8303824002812525`\\). \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"Plot::plnr\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(General::"stop"\), \(\(:\)\(\ \)\), "\<\"Further output of \
\\!\\(Plot :: \\\"plnr\\\"\\) will be suppressed during this calculation. \\!\
\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"General::stop\\\"]\\)\"\>"}]], "Message"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0147151 0.588604 [
[.21429 .00222 -9.5 -15.6875 ]
[.21429 .00222 9.5 0 ]
[.40476 .00222 -9.5 -15.6875 ]
[.40476 .00222 9.5 0 ]
[.59524 .00222 -9.5 -15.6875 ]
[.59524 .00222 9.5 0 ]
[.78571 .00222 -9.5 -15.6875 ]
[.78571 .00222 9.5 0 ]
[.97619 .00222 -5 -15.6875 ]
[.97619 .00222 5 0 ]
[.01131 .13244 -19 -7.84375 ]
[.01131 .13244 0 7.84375 ]
[.01131 .25016 -19 -7.84375 ]
[.01131 .25016 0 7.84375 ]
[.01131 .36788 -19 -7.84375 ]
[.01131 .36788 0 7.84375 ]
[.01131 .4856 -19 -7.84375 ]
[.01131 .4856 0 7.84375 ]
[.01131 .60332 -10 -7.84375 ]
[.01131 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .01472 m
.21429 .02097 L
s
gsave
.21429 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.40476 .01472 m
.40476 .02097 L
s
gsave
.40476 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.59524 .01472 m
.59524 .02097 L
s
gsave
.59524 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.78571 .01472 m
.78571 .02097 L
s
gsave
.78571 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.97619 .01472 m
.97619 .02097 L
s
gsave
.97619 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .01472 m
.07143 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.16667 .01472 m
.16667 .01847 L
s
.2619 .01472 m
.2619 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.35714 .01472 m
.35714 .01847 L
s
.45238 .01472 m
.45238 .01847 L
s
.5 .01472 m
.5 .01847 L
s
.54762 .01472 m
.54762 .01847 L
s
.64286 .01472 m
.64286 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.7381 .01472 m
.7381 .01847 L
s
.83333 .01472 m
.83333 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.92857 .01472 m
.92857 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .13244 m
.03006 .13244 L
s
gsave
.01131 .13244 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.02381 .25016 m
.03006 .25016 L
s
gsave
.01131 .25016 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.02381 .36788 m
.03006 .36788 L
s
gsave
.01131 .36788 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.02381 .4856 m
.03006 .4856 L
s
gsave
.01131 .4856 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.02381 .60332 m
.03006 .60332 L
s
gsave
.01131 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04415 m
.02756 .04415 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .10301 m
.02756 .10301 L
s
.02381 .16187 m
.02756 .16187 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .30902 m
.02756 .30902 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .45617 m
.02756 .45617 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00K03oool01@3ogi`0B0000000000007@0_ooo00<0oooo00D0okmd0000001d_ol0oooo0;md
00060000000304RLg`3oool0oooo03<0oooo00D0omnL04P000000000001d0;ooo`030?ooo`030?no
M0000000M;oo00D0oooo00@0oooO09a80000B9`0gooo=@3oool01@3ogi`0B0000000000007@0_ooo
00<0oooo00<0okmd0000001d_ol00`3oool01@3ogi`0B000000000000000M;oo03H0oooo00D0omnL
04P000000000001d0;ooo`030?ooo`030?noM0000000M;oo00<0oooo00D0gia80000000000000000
04RLg`0m0?ooo`060;md00000000000000000000B9`0gooo4`3oool006/0oooo00L0oono07@0001d
_ol0oooo0?noM00007@0_ooo0080oooo00<0okmd0000001d_ol00P3oool00`3OW4P0000000000003
000000030018W03Oool0oooo0380oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo
00<0okmd0000001d_ol01@3oool0103ooml0W4P00018W03Ooold0?ooo`070?oo_`1d0000M;oo0?oo
o`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo0080oooo00L0omnL04P00018W=l0oooo0?oo
_`1d0000M;oo03@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00<0okmd0000
001d_ol00P3oool01`3OW4P0000007Boo`3oool0oooO09a80018W=l0?@3oool00`3ogi`0B00007Bo
o`0E0?ooo`00J`3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0gia8
04RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`0b0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0
001d_`3oool02@3oool0103ooml0W4P00018W03Ooold0?ooo`030;md0000B9`0gooo0080oooo00<0
_g@0001d_`3oool01@3oool00`3ooml0W4P0001d_`030?ooo`030?noM000B9`0gooo03<0oooo00<0
_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?oog`2LB00007Bo00<0oooo00<0
okmd0018W03Oool0?03oool00`3ogi`0B00007Boo`0E0?ooo`00JP3oool00`3ookl0M00004RLg`03
0?ooo`030=nLB00007@0_ooo00P0oooo00<0gia807Boo`3oool0=@3oool00`3ookl0M00004RLg`03
0?ooo`030=nLB00007@0_ooo00D0oooo00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`0`
0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01@3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo0380oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`06
0?ooo`030;md0000M;l0oooo0080oooo00<0_g@00018W03Oool0?03oool00`3ogi`0B00007Boo`0E
0?ooo`00JP3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00<0_gA809cO
o`3oool0=03oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00D0oooo00P0omnL07BL
g`3oool0oooo0?oog`2LB00004RL0=oooc<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`050?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0<P3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0omnL04P0M02oool0oooo0;md00000000M;oo
03d0oooo00<0omnL04P0001d_ol05@3oool006X0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`090?ooo`030?oog`2LB7@0_ooo03@0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`060?ooo`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`0c0?ooo`030?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol01@3oool00`3ogi`0B00007Boo`030?ooo`030;md000004P0
W=oo0380oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`050?oo_`1dB4P0
B00000000018W=l0?P3oool00`3ogi`0B00007Boo`0E0?ooo`00JP3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo00X0oooo00<0okmd001d_`3oool0<`3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo00H0oooo00L0oooO09a8M02oool0oooO09a80000B9`0gooo03<0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`050?ooo`080?oo_`1d0000000004RLg`3o
ool0okmd00000000M;lc0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool0
103ogi`0B00000000000M;lo0?ooo`030?oOW0180000M;oo01D0oooo001Z0?ooo`030?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3ookl0M01809cOo`0c0?ooo`030?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol01`3oool01P3ogi`0B7Bo0?oog`2LB00004RL0=oooc<0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`030;md0000B9`0_ia800800000
00<004RL0=ooo`3oool0<P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo
00H0omnL04P0000007@0_ooo0?noM000M;ln0?ooo`030?oOW0180000M;oo01D0oooo001Z0?ooo`03
0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02P3oool00`3ookl0M00004RLg`0c0?ooo`03
0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool01@3OW4P0M;oO09a80000B9`0gooo
03<0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`030?oOW0180000M;oo
03L0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`030;md000007@0_ooo
0080oooo00<0_g@004RLg`3oool0?03oool00`3ogi`0B00007Boo`0E0?ooo`00J`3oool00`3OW4P0
04RL0=ooo`020?ooo`030;md0000M;l0oooo00D0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0
B00004RLg`0d0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0203oool01@3ooml0
W4Q804Q80000B9`0gooo03@0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?oo
o`030=nLB00004P0W=oo03L0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`060?oo
o`030;md0000B9`0gooo0080oooo00<0gia8001d_`3oool0?03oool00`3ogi`0B00007Boo`0E0?oo
o`00J`3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool01P3oool0203ooml0W4Q809cOo`3o
ool0oono07@0000007@0_ooo=03oool01`3ookl0M00007Boo`3oool0okmd0000M02oool02P3oool0
103ogi`0B0000018W03Ooold0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`03
0?noM000000007Bo03H0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00L0oooo00L0okmd
0000001d_ol0oooo0?oo_`1d0000M;oo03/0oooo00D0oooO09a800000000000007Boo`0E0?ooo`00
K03oool01@3ookl0M0000000000007@0_ooo00P0oooo00H0oooO09a80000000000000000M02ooolf
0?ooo`050?oo_`1d000000000000M02oool0303oool00`3OW4P004RL0=ooo`0e0?ooo`050?oo_`1d
000000000000M02oool02`3oool01@3ooml0W4P00000000007@0_ooo03@0oooo00D0oono07@00000
0000001d0;ooo`090?ooo`050?noM0000000000000000018W=l0?P3oool00`3o_g@0000007Boo`0E
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00;03oool0
0`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0
oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo
000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon2
0?ooo`008`3ooooo000008800000303oool002`0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0@0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0@0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0@0?ooo`030000003oool0oooo01<0oooo000/0?ooo`030000003oool0oooo04H0oooo
00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo04H0oooo
00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0C0?ooo`00;03oool00`000000oooo0?oo
o`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?oo
o`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000
003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool0
0`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0
oooo0`00003o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<0
00000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?oo
o`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00
;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3o
ool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oo
oh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0
o`3ooon20?ooo`00;03oool300000?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon2
0?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0
ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?oo
o`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?oo
o`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000
003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool3
00000?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000
oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<0
00000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000@0?oo
o`050?oOW018000000000000M02oool00`3oool01@3o_g@0000007Boo`3oool0_g@000H0000000<0
B9cO0?ooo`3oool01P3oool00`000000oooo0?ooo`3o0?oooh80oooo000?0?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo0080oooo00<0gia8000000000000
0`0000000`00B9`0gooo0?ooo`060?ooo`030000003oool0oooo0?l0ooooPP3oool000l0oooo00<0
_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030=nLB018W=l0oooo0080oooo00<0
giad09cOo`3oool01P3oool00`000000oooo0?ooo`3o0?oooh80oooo000>0?ooo`030?oo_`1d0000
B9cO00<0oooo00<0gia80000M02oool0203oool00`3OW4P0M;oo0?ooo`0:0?ooo`030000003oool0
oooo0?l0ooooPP3oool000h0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`090?oo
o`030;mdB02Lgol0oooo00T0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003P3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00<0oooO09a8M02oool02@3oool400000?l0
ooooP@3oool000h0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030?no
M000M;l0oooo00P0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003P3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00X0oooo00<0oono07@0B02Lgol0203oool00`000000oooo0?oo
o`3o0?oooh80oooo000>0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02P3oool0
0`3ookl0M00004RLg`080?ooo`030000003oool0oooo0?l0ooooPP3oool000l0oooo00<0gia80018
W03Oool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P0
0018W=l0203oool00`000000oooo0?ooo`3o0?oooh80oooo000?0?ooo`070?oo_`1d0000M;oo0?oo
o`3o_g@0001d0;ooo`060?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool80?ooo`03
0000003oool0oooo0?l0ooooPP3oool00100oooo00D0oono07@000000000001d0;ooo`080?ooo`06
0?oog`2LB00000000000000007@0_ooo2@3oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?oo
o`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00
;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3o
ool002`0oooo0`00003o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0
oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo
000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon2
0?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0
ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?oo
o`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?oo
o`3oool0o`3ooon20?ooo`00;03oool300000?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0
o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0
oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000
oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<0
00000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?oo
o`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00
;03oool300000?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool0
0`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0
oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo
000@0?ooo`050?oOW018000000000000M02oool00`3oool00`3o_g@0000007Boo`050?ooo`040?oo
g`2LB00004RL0=ooo`P0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003`3oool01`3ookl0M000
07Boo`3oool0okmd0000M02oool00P3oool00`3o_g@0000007Boo`050?ooo`040?oog`2LB00004RL
0=ooo`P0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003`3oool00`2oM00004RL0=ooo`020?oo
o`030;md0000M;l0oooo00T0oooo00@0oooO09a80000B9`0gooo203oool00`000000oooo0?ooo`3o
0?oooh80oooo000>0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01@3oool00`2o
M0000000000000050000000307Boo`3oool0oooo00D0oooo00<000000?ooo`3oool0o`3ooon20?oo
o`003P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00D0oooo00P0omnL07BLg`3o
ool0oooo0?oog`2LB00004RL0=ooo`P0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003P3oool0
0`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0gia807Boo`3oool0oooO09a8
0000B9`0gooo00P0oooo1000003o0?oooh40oooo000>0?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol01P3oool01`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool0203oool00`000000
oooo0?ooo`3o0?oooh80oooo000>0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
1`3oool01P3ogi`0B7Bo0?oog`2LB00004RL0=ooo`P0oooo00<000000?ooo`3oool0o`3ooon20?oo
o`003P3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00P0oooo00D0gia807Bog`2L
B00004RL0=ooo`080?ooo`030000003oool0oooo0?l0ooooPP3oool000l0oooo00<0gia80018W03O
ool00P3oool00`2oM00007Bo0?ooo`080?ooo`050?oog`2LB4P0B4P00018W03Oool0203oool00`00
0000oooo0?ooo`3o0?oooh80oooo000?0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0:
0?ooo`040?oOW018000004RL0=ooo`P0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00403oool0
1@3ookl0M0000000000007@0_ooo00`0oooo00<0gia80018W03Oool0203oool00`000000oooo0?oo
o`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?oo
o`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000
003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool3
00000?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000
oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<0
00000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?oo
o`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00
;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3o
ool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool300000?l0ooooPP3oool002`0
oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo
000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon2
0?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0
ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?oo
o`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?oo
o`3oool0o`3ooon20?ooo`00;03oool300000?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0
o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0
oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000
oooo0?ooo`3o0?oooh80oooo000@0?ooo`050?oOW018000000000000M02oool00`3oool00`3o_g@0
000007Boo`030?ooo`050?oOW018000000000000001d_ol02@3oool00`000000oooo0?ooo`3o0?oo
oh80oooo000?0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000
M;oo0080oooo00L0omnL04P00018W=l0oooo0?oo_`1d0000M;oo00P0oooo00<000000?ooo`3oool0
o`3ooon20?ooo`003`3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00D0oooo00<0
oooO09a80000M;l00`3oool00`3o_g@004RL0=ooo`070?ooo`030000003oool0oooo0?l0ooooPP3o
ool000h0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`050?ooo`030?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`3o0?oooh80oooo000>0?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01@3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003P3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00D0oooo00<0omnL04P0001d_ol00`3oool00`2oM000
001809cOo`070?ooo`@00000o`3ooon10?ooo`003P3oool00`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00D0oooo00P0oono07@000000000B9cO0?ooo`3o_g@00000001d_`P0oooo00<00000
0?ooo`3oool0o`3ooon20?ooo`003P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00H0oooo00<0_g@00018W02oW4P00P0000000`00B9`0gooo0?ooo`070?ooo`030000003oool0oooo
0?l0ooooPP3oool000h0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`03
0?oOW0180000M;oo00`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003`3oool00`3OW4P004RL
0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0gia80000B02Lgol02`3oool00`000000oooo
0?ooo`3o0?oooh80oooo000?0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`03
0?noM000000007Bo00X0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00403oool01@3ookl0M000
0000000007@0_ooo00/0oooo00D0oooO09a800000000001d0;ooo`070?ooo`030000003oool0oooo
0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo
0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<00000
0?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`<0
0000o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003o
ool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`00
0000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo
00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/
0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?oo
o`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`<00000o`3ooon20?ooo`00;03o
ool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool0
02`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80
oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3o
oon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo
0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo
0?ooo`3o0?oooh80oooo000/0?ooo`<00000o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o
0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3o
ool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003o
ool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00403oool01@3o
gi`0B0000000000007@0_ooo00<0oooo00<0okmd0000001d_ol00`3oool01@3OW4P0000000000000
0000B9cO00T0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003`3oool01`3ookl0M00007Boo`3o
ool0okmd0000M02oool00P3oool00`3o_g@0000007Boo`020?ooo`070=nLB0000000M;oo0?ooo`3o
oml0W4P004RLg`080?ooo`030000003oool0oooo0?l0ooooPP3oool000l0oooo00<0_g@00018W03O
ool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?oog`2LB00007Bo00<0oooo00<0okmd0018W03O
ool01`3oool00`000000oooo0?ooo`3o0?oooh80oooo000>0?ooo`030?oo_`1d0000B9cO00<0oooo
00<0gia80000M02oool01P3oool00`2oM00007Bo0?ooo`020?ooo`030;md0000B9`0gooo00L0oooo
00<000000?ooo`3oool0o`3ooon20?ooo`003P3oool00`3ogi`0B00007Boo`030?ooo`030?noM000
04P0W=oo00H0oooo00L0omnL04P0M02oool0oooo0;md00000000M;oo00P0oooo00<000000?ooo`3o
ool0o`3ooon20?ooo`003P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo
00D0oono07A8B0180000000004RLg`090?ooo`@00000o`3ooon10?ooo`003P3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00L0oooo00@0omnL04P00000000007Bo2P3oool00`000000
oooo0?ooo`3o0?oooh80oooo000>0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
1P3oool01P3ogi`0B0000000M02oool0okmd001d_`T0oooo00<000000?ooo`3oool0o`3ooon20?oo
o`003P3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00<0_g@00000M02o
ool00P3oool00`2oM000B9cO0?ooo`070?ooo`030000003oool0oooo0?l0ooooPP3oool000l0oooo
00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`060?ooo`030;md0000B9`0gooo0080oooo
00<0gia8001d_`3oool01`3oool00`000000oooo0?ooo`3o0?oooh80oooo000?0?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`070?ooo`070?noM0000000M;oo0?ooo`3ookl0M00007Boo`08
0?ooo`030000003oool0oooo0?l0ooooPP3oool00100oooo00D0oono07@000000000001d0;ooo`09
0?ooo`050?noM0000000000000000018W=l02@3oool00`000000oooo0?ooo`3o0?oooh80oooo000/
0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?oo
o`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0oooo
PP3oool002`0oooo0`00003o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool0
02`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80
oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3o
oon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo
0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo
0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo0`00003o
0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3o
ool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003o
ool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`00
0000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo
00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/
0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo0`00003o0?oooh80oooo000/0?ooo`03
0000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03o
ool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool0
02`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80
oooo000N0?ooo`060;md00000000000000000000B9`0gooo203oool00`000000oooo0?ooo`3o0?oo
oh80oooo000O0?ooo`030?oOW0180000M;oo00X0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00
7`3oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo0?l0ooooPP3oool001l0oooo00<0
omnL04P0001d_ol02P3oool00`000000oooo0?ooo`3o0?oooh80oooo000O0?ooo`030?oOW0180000
M;oo00X0oooo00<000000?ooo`3oool0o`3ooon20?ooo`007`3oool00`3ogi`0B00007Boo`0:0?oo
o`@00000o`3ooon10?ooo`007`3oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo0?l0
ooooPP3oool001l0oooo00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`3o0?oooh80oooo
000O0?ooo`030?oOW0180000M;oo00X0oooo00<000000?ooo`3oool0o`3ooon20?ooo`007`3oool0
0`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo0?l0ooooPP3oool001d0oooo00D0oooO09a8
00000000000007Boo`0:0?ooo`030000003oool0oooo0?l0ooooPP3oool001l0oooo00<0okmd0000
001d_ol0o`3ooon?0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.122957, -0.125454, \
0.00410595, 0.00664356}}]
}, Closed]],

Cell[TextData[{
  "In the example above, because we failed to specify the value of ",
  Cell[BoxData[
      \(TraditionalForm\`k\)]],
  " the argument does not evaluate to a number.  Although ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " complains that the function is not a machine-size real number, it isn't \
even a definite number of any kind.  Nevertheless, ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " will plot whatever it can, useful or not.  In this case, it simply \
produces a set of axes with default ranges and increments because that is the \
only part of the command it can evaluate.  It is much happier with the \
versions below, in which the function is specified completely."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Cos[\[Pi]\ x], \ {x, \(-1\), 1}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.309018 0.294303 [
[.02381 .29652 -9.15625 -15.6875 ]
[.02381 .29652 9.15625 0 ]
[.2619 .29652 -13.6563 -15.6875 ]
[.2619 .29652 13.6563 0 ]
[.7381 .29652 -9.5 -15.6875 ]
[.7381 .29652 9.5 0 ]
[.97619 .29652 -5 -15.6875 ]
[.97619 .29652 5 0 ]
[.4875 .01472 -18.3125 -7.84375 ]
[.4875 .01472 0 7.84375 ]
[.4875 .16187 -27.3125 -7.84375 ]
[.4875 .16187 0 7.84375 ]
[.4875 .45617 -19 -7.84375 ]
[.4875 .45617 0 7.84375 ]
[.4875 .60332 -10 -7.84375 ]
[.4875 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 .30902 m
.02381 .31527 L
s
gsave
.02381 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 .30902 m
.2619 .31527 L
s
gsave
.2619 .29652 -74.6563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.7381 .30902 m
.7381 .31527 L
s
gsave
.7381 .29652 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 .30902 m
.97619 .31527 L
s
gsave
.97619 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .30902 m
.07143 .31277 L
s
.11905 .30902 m
.11905 .31277 L
s
.16667 .30902 m
.16667 .31277 L
s
.21429 .30902 m
.21429 .31277 L
s
.30952 .30902 m
.30952 .31277 L
s
.35714 .30902 m
.35714 .31277 L
s
.40476 .30902 m
.40476 .31277 L
s
.45238 .30902 m
.45238 .31277 L
s
.54762 .30902 m
.54762 .31277 L
s
.59524 .30902 m
.59524 .31277 L
s
.64286 .30902 m
.64286 .31277 L
s
.69048 .30902 m
.69048 .31277 L
s
.78571 .30902 m
.78571 .31277 L
s
.83333 .30902 m
.83333 .31277 L
s
.88095 .30902 m
.88095 .31277 L
s
.92857 .30902 m
.92857 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5 .01472 m
.50625 .01472 L
s
gsave
.4875 .01472 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .16187 m
.50625 .16187 L
s
gsave
.4875 .16187 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .45617 m
.50625 .45617 L
s
gsave
.4875 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .60332 m
.50625 .60332 L
s
gsave
.4875 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .04415 m
.50375 .04415 L
s
.5 .07358 m
.50375 .07358 L
s
.5 .10301 m
.50375 .10301 L
s
.5 .13244 m
.50375 .13244 L
s
.5 .1913 m
.50375 .1913 L
s
.5 .22073 m
.50375 .22073 L
s
.5 .25016 m
.50375 .25016 L
s
.5 .27959 m
.50375 .27959 L
s
.5 .33845 m
.50375 .33845 L
s
.5 .36788 m
.50375 .36788 L
s
.5 .39731 m
.50375 .39731 L
s
.5 .42674 m
.50375 .42674 L
s
.5 .4856 m
.50375 .4856 L
s
.5 .51503 m
.50375 .51503 L
s
.5 .54446 m
.50375 .54446 L
s
.5 .57389 m
.50375 .57389 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01472 m
.02499 .01472 L
.02605 .01475 L
.02729 .01479 L
.02846 .01485 L
.03053 .015 L
.03279 .01523 L
.03527 .01556 L
.0379 .01599 L
.04262 .01698 L
.04749 .0183 L
.05205 .01981 L
.06244 .02422 L
.07305 .03011 L
.08274 .03668 L
.10458 .05552 L
.14429 .10289 L
.18248 .16177 L
.22313 .23454 L
.26226 .3097 L
.30384 .3894 L
.34391 .46058 L
.38246 .51918 L
.402 .54393 L
.42346 .56659 L
.44388 .58338 L
.45301 .58929 L
.46295 .59457 L
.46838 .59694 L
.4734 .5988 L
.4781 .60026 L
.4833 .60154 L
.48553 .60198 L
.48794 .60239 L
.49002 .60268 L
.4923 .60294 L
.49473 .60314 L
.49605 .60322 L
.49728 .60327 L
.49844 .6033 L
.49949 .60332 L
.50071 .60332 L
.50186 .6033 L
.50301 .60326 L
.50425 .60321 L
.50542 .60313 L
.50648 .60305 L
.5091 .60279 L
.51158 .60246 L
.5165 .60158 L
Mistroke
.52187 .60026 L
.53144 .59701 L
.54032 .59297 L
.56088 .5799 L
.5795 .56376 L
.61962 .51633 L
.65822 .45702 L
.69928 .38356 L
.73882 .30761 L
.77684 .2346 L
.81732 .1621 L
.85628 .1021 L
.87613 .07654 L
.89769 .0533 L
.91843 .03582 L
.92832 .02927 L
.93759 .02421 L
.94643 .02037 L
.95595 .01733 L
.9611 .01617 L
.96369 .01572 L
.96652 .01531 L
.96894 .01505 L
.9703 .01494 L
.97153 .01485 L
.97268 .01479 L
.97393 .01475 L
.97511 .01472 L
.97619 .01472 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00f@3oool0
0`000000oooo0?ooo`3D0?ooo`00bP3oool01P2oM000000000000000000004RL0=ooo`T0oooo00<0
00000?ooo`3oool0e03oool00</0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`3D
0?ooo`00b`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0=@0oooo003;0?ooo`03
0?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0e03oool00;/0oooo00<0oooO09a800000000
1`0000000`18W=l0oooo0?ooo`030?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0
e03oool001D0oooo0`00002c0?ooo`030?oOW0180000M;oo00/0oooo1000002n0?ooo`<000004P3o
ool001P0oooo1P00002]0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0^@3oool6
000001D0oooo000N0?ooo`@00000Z@3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo
0;L0oooo0P00000K0?ooo`008P3oool200000:L0oooo00<0omnL04P0001d_ol02`3oool00`000000
oooo0?ooo`2d0?ooo`<000007@3oool002@0oooo0P00002U0?ooo`030?oOW0180000M;oo00/0oooo
00<000000?ooo`3oool0/@3oool300000200oooo000V0?ooo`030000003oool0oooo0:00oooo00D0
oooO09a800000000000007Boo`0;0?ooo`030000003oool0oooo0:l0oooo0P00000S0?ooo`009`3o
ool200000:80oooo00<0okmd0000001d_ol02`3oool00`000000oooo0?ooo`2]0?ooo`8000009@3o
ool002T0oooo00<000000?ooo`3oool0[@3oool00`000000oooo0?ooo`2/0?ooo`030000003oool0
oooo02D0oooo000Z0?ooo`800000[@3oool00`000000oooo0?ooo`2Z0?ooo`800000:03oool002`0
oooo0P00002[0?ooo`030000003oool0oooo0:T0oooo00<000000?ooo`3oool0:03oool002h0oooo
00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo02T0oooo
000_0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?oo
o`0Z0?ooo`00<03oool00`000000oooo0?ooo`2V0?ooo`030000003oool0oooo0:H0oooo00<00000
0?ooo`3oool0:`3oool00340oooo0P00002V0?ooo`<00000Y03oool2000002h0oooo000c0?ooo`03
0000003oool0oooo0:<0oooo00<000000?ooo`3oool0X`3oool00`000000oooo0?ooo`0^0?ooo`00
=03oool00`000000oooo0?ooo`2R0?ooo`030000003oool0oooo0:80oooo00<000000?ooo`3oool0
;`3oool003D0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo0300oooo000f0?ooo`030000003oool0oooo0:00oooo00<000000?ooo`3oool0X03oool0
0`000000oooo0?ooo`0a0?ooo`00=`3oool00`000000oooo0?ooo`2O0?ooo`030000003oool0oooo
09l0oooo00<000000?ooo`3oool0<P3oool003P0oooo00<000000?ooo`3oool0WP3oool00`000000
oooo0?ooo`2N0?ooo`030000003oool0oooo03<0oooo000i0?ooo`030000003oool0oooo09d0oooo
00<000000?ooo`3oool0W@3oool00`000000oooo0?ooo`0d0?ooo`00>@3oool00`000000oooo0?oo
o`2M0?ooo`030000003oool0oooo09`0oooo00<000000?ooo`3oool0=@3oool003X0oooo00<00000
0?ooo`3oool0W03oool00`000000oooo0?ooo`2L0?ooo`030000003oool0oooo03D0oooo000k0?oo
o`030000003oool0oooo09/0oooo00<000000?ooo`3oool0V`3oool00`000000oooo0?ooo`0f0?oo
o`00?03oool00`000000oooo0?ooo`2J0?ooo`030000003oool0oooo09X0oooo00<000000?ooo`3o
ool0=`3oool003d0oooo00<000000?ooo`3oool0V@3oool3000009T0oooo00<000000?ooo`3oool0
>03oool003h0oooo00<000000?ooo`3oool0V03oool00`000000oooo0?ooo`2H0?ooo`030000003o
ool0oooo03T0oooo000n0?ooo`030000003oool0oooo09P0oooo00<000000?ooo`3oool0U`3oool0
0`000000oooo0?ooo`0j0?ooo`00?`3oool00`000000oooo0?ooo`2G0?ooo`030000003oool0oooo
09H0oooo00<000000?ooo`3oool0>`3oool00400oooo00<000000?ooo`3oool0UP3oool00`000000
oooo0?ooo`2F0?ooo`030000003oool0oooo03/0oooo00110?ooo`030000003oool0oooo09D0oooo
00<000000?ooo`3oool0U@3oool00`000000oooo0?ooo`0l0?ooo`00@P3oool00`000000oooo0?oo
o`2D0?ooo`030000003oool0oooo09@0oooo00<000000?ooo`3oool0?@3oool004<0oooo00<00000
0?ooo`3oool0T`3oool00`000000oooo0?ooo`2C0?ooo`030000003oool0oooo03h0oooo00140?oo
o`030000003oool0oooo0980oooo00<000000?ooo`3oool0TP3oool00`000000oooo0?ooo`0o0?oo
o`00A03oool00`000000oooo0?ooo`2B0?ooo`030000003oool0oooo0980oooo00<000000?ooo`3o
ool0?`3oool004D0oooo00<000000?ooo`3oool0T@3oool00`000000oooo0?ooo`2A0?ooo`030000
003oool0oooo0400oooo00160?ooo`030000003oool0oooo0900oooo00<000000?ooo`3oool0T03o
ool00`000000oooo0?ooo`110?ooo`00A`3oool00`000000oooo0?ooo`2?0?ooo`<00000S`3oool0
0`000000oooo0?ooo`120?ooo`00B03oool00`000000oooo0?ooo`2>0?ooo`030000003oool0oooo
08l0oooo00<000000?ooo`3oool0@P3oool004P0oooo00<000000?ooo`3oool0SP3oool00`000000
oooo0?ooo`2>0?ooo`030000003oool0oooo04<0oooo00190?ooo`030000003oool0oooo08d0oooo
00<000000?ooo`3oool0S@3oool00`000000oooo0?ooo`140?ooo`00B@3oool00`000000oooo0?oo
o`2=0?ooo`030000003oool0oooo08d0oooo00<000000?ooo`3oool0A03oool004X0oooo00<00000
0?ooo`3oool0S03oool00`000000oooo0?ooo`2<0?ooo`030000003oool0oooo04D0oooo001;0?oo
o`030000003oool0oooo08/0oooo00<000000?ooo`3oool0S03oool00`000000oooo0?ooo`150?oo
o`00B`3oool00`000000oooo0?ooo`2;0?ooo`030000003oool0oooo08/0oooo00<000000?ooo`3o
ool0AP3oool004`0oooo00<000000?ooo`3oool0RP3oool00`000000oooo0?ooo`2:0?ooo`030000
003oool0oooo04L0oooo001=0?ooo`030000003oool0oooo08T0oooo00<000000?ooo`3oool0RP3o
ool00`000000oooo0?ooo`170?ooo`00C@3oool00`000000oooo0?ooo`290?ooo`030000003oool0
oooo08T0oooo00<000000?ooo`3oool0B03oool004h0oooo00<000000?ooo`3oool0R03oool00`00
0000oooo0?ooo`280?ooo`030000003oool0oooo04T0oooo001>0?ooo`030000003oool0oooo08P0
oooo0`0000280?ooo`030000003oool0oooo04T0oooo001?0?ooo`030000003oool0oooo08L0oooo
00<000000?ooo`3oool0Q`3oool00`000000oooo0?ooo`1:0?ooo`00D03oool00`000000oooo0?oo
o`260?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3oool0B`3oool00500oooo00<00000
0?ooo`3oool0QP3oool00`000000oooo0?ooo`260?ooo`030000003oool0oooo04/0oooo001A0?oo
o`030000003oool0oooo08D0oooo00<000000?ooo`3oool0Q@3oool00`000000oooo0?ooo`1<0?oo
o`00DP3oool00`000000oooo0?ooo`240?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3o
ool0C@3oool00580oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`240?ooo`030000
003oool0oooo04d0oooo001C0?ooo`030000003oool0oooo06H0oooo00D0omnL04P000000000001d
0;ooo`030?ooo`060?noM0000000M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`090?oo
o`030000003oool0oooo08<0oooo00<000000?ooo`3oool0CP3oool005<0oooo00<000000?ooo`3o
ool0I@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@0000007Boo`3o
ool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`090?ooo`030000003oool0oooo08<0oooo
00<000000?ooo`3oool0CP3oool005@0oooo00<000000?ooo`3oool0I03oool00`2oM00004RL0=oo
o`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO09a8M02oool02@3oool00`000000oooo0?oo
o`220?ooo`030000003oool0oooo04l0oooo001E0?ooo`030000003oool0oooo0680oooo00<0oono
07@00018W=l00`3oool00`3OW4P0001d0;ooo`0;0?ooo`030;md001d_ol0oooo00P0oooo00<00000
0?ooo`3oool0P@3oool00`000000oooo0?ooo`1@0?ooo`00E@3oool00`000000oooo0?ooo`1F0?oo
o`030?oog`2LB000000000L0000000D0B9cO0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo00X0oooo00<0oooO09a80018W=l02@3oool00`000000oooo0?ooo`210?ooo`030000003oool0
oooo0500oooo001F0?ooo`030000003oool0oooo0640oooo00<0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`0:0?ooo`030=nLB0000000M;oo00T0oooo1000001o0?ooo`030000003oool0oooo
0540oooo001G0?ooo`030000003oool0oooo0600oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`060?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`090?ooo`030000003oool0
oooo07l0oooo00<000000?ooo`3oool0DP3oool005L0oooo00<000000?ooo`3oool0H03oool00`3o
gi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00H0gia800000000000000000000B02L
gol:0?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool0DP3oool005P0oooo00<00000
0?ooo`3oool0G`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono
07@000000000B9cO303oool00`000000oooo0?ooo`1n0?ooo`030000003oool0oooo05<0oooo001H
0?ooo`030000003oool0oooo0600oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`07
0?ooo`030;md001d_ol0oooo00`0oooo00<000000?ooo`3oool0OP3oool00`000000oooo0?ooo`1C
0?ooo`00F@3oool00`000000oooo0?ooo`1O0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oo
o`080?ooo`060?oOW018000000000000000004P0W=oo2@3oool00`000000oooo0?ooo`1m0?ooo`03
0000003oool0oooo05@0oooo001I0?ooo`030000003oool0oooo0600oooo00D0oono07@000000000
001d0;ooo`0:0?ooo`050;md00000000000000000018W=l02@3oool00`000000oooo0?ooo`1m0?oo
o`030000003oool0oooo05@0oooo001J0?ooo`030000003oool0oooo07`0oooo00<000000?ooo`3o
ool0O03oool00`000000oooo0?ooo`1E0?ooo`00F`3oool00`000000oooo0?ooo`1k0?ooo`030000
003oool0oooo07/0oooo00<000000?ooo`3oool0EP3oool005/0oooo00<000000?ooo`3oool0N`3o
ool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo05H0oooo001L0?ooo`030000003oool0
oooo07X0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`1G0?ooo`00G03oool00`00
0000oooo0?ooo`1j0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0E`3oool005d0
oooo00<000000?ooo`3oool0N@3oool3000007T0oooo00<000000?ooo`3oool0F03oool005d0oooo
00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo05P0oooo
001N0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0N03oool00`000000oooo0?oo
o`1I0?ooo`00GP3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo07L0oooo00<00000
0?ooo`3oool0FP3oool005l0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`1g0?oo
o`030000003oool0oooo05X0oooo001P0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3o
ool0MP3oool00`000000oooo0?ooo`1K0?ooo`00H03oool00`000000oooo0?ooo`1f0?ooo`030000
003oool0oooo07H0oooo00<000000?ooo`3oool0F`3oool00640oooo00<000000?ooo`3oool0M@3o
ool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo05`0oooo001Q0?ooo`030000003oool0
oooo07D0oooo00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`1L0?ooo`00HP3oool00`00
0000oooo0?ooo`1d0?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3oool0G@3oool00680
oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`1d0?ooo`030000003oool0oooo05d0
oooo001S0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool0L`3oool00`000000oooo
0?ooo`1N0?ooo`00I03oool00`000000oooo0?ooo`1b0?ooo`<00000LP3oool00`000000oooo0?oo
o`1O0?ooo`00I03oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo0780oooo00<00000
0?ooo`3oool0G`3oool006D0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`1a0?oo
o`030000003oool0oooo0600oooo001U0?ooo`030000003oool0oooo0740oooo00<000000?ooo`3o
ool0L@3oool00`000000oooo0?ooo`1P0?ooo`00IP3oool00`000000oooo0?ooo`1`0?ooo`030000
003oool0oooo0700oooo00<000000?ooo`3oool0H@3oool006H0oooo00<000000?ooo`3oool0L03o
ool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo0640oooo001W0?ooo`030000003oool0
oooo06l0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`1R0?ooo`00J03oool00`00
0000oooo0?ooo`1^0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool0H`3oool006P0
oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`1^0?ooo`030000003oool0oooo06<0
oooo001Y0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0K@3oool00`000000oooo
0?ooo`1T0?ooo`00J@3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo06d0oooo00<0
00000?ooo`3oool0I03oool006X0oooo00<000000?ooo`3oool0K03oool00`000000oooo0?ooo`1/
0?ooo`030000003oool0oooo06D0oooo001Z0?ooo`030000003oool0oooo06`0oooo0`00001/0?oo
o`030000003oool0oooo06D0oooo001[0?ooo`030000003oool0oooo06/0oooo00<000000?ooo`3o
ool0J`3oool00`000000oooo0?ooo`1V0?ooo`00J`3oool00`000000oooo0?ooo`1[0?ooo`030000
003oool0oooo06/0oooo00<000000?ooo`3oool0IP3oool001T0oooo00H0_g@00000000000000000
0018W03Ooom=0?ooo`030000003oool0oooo00D0oooo00D0omnL04P000000000001d0;ooo`030?oo
o`060?noM0000000M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`1A0?ooo`030000003o
ool0oooo05H0oooo00D0omnL04P000000000001d0;ooo`030?ooo`060?noM0000000M;oo0?ooo`3o
oml0W4P00`000000101d_ol0oooo0?ooo`00001D0?ooo`060;md00000000000000000000B9`0gooo
3`3oool001X0oooo00<0omnL04P0001d_ol0C`3oool00`000000oooo0?ooo`040?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo0?ooo`3ooml0W4Qd0;ooo`02
0?ooo`030;md001d_ol0oooo0540oooo00<000000?ooo`3oool0E@3oool01`3ookl0M00007Boo`3o
ool0okmd0000M02oool00P3oool01`3o_g@0000007Boo`3oool0oooO09a8M02oool00P3oool0102o
M000M;oo0?ooo`00001E0?ooo`030?oOW0180000M;oo0140oooo000J0?ooo`030?oOW0180000M;oo
0500oooo00<000000?ooo`3oool00`3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo
00X0oooo00<0oooO09a8M02oool0D@3oool00`000000oooo0?ooo`1E0?ooo`030;md0000B9`0gooo
0080oooo00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd0000001F0?ooo`030?oOW0180000M;oo
0140oooo000J0?ooo`030?oOW0180000M;oo0500oooo00<000000?ooo`3oool00P3oool00`3ookl0
M00004RLg`030?ooo`030=nLB00007@0_ooo00/0oooo00<0_g@007Boo`3oool0D03oool00`000000
oooo0?ooo`1D0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02`3oool00`2oM000
00000?ooo`1E0?ooo`030?oOW0180000M;oo0140oooo000:0?ooo`030?oog`2LB000000000L00000
00<0B9cO0?ooo`3oool00`3oool00`3ogi`0B00007Boo`180?ooo`030?oog`2LB000000000L00000
00H0B9cO0?ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo00X0oooo00<0oooO09a8
0018W=l0D@3oool00`000000oooo0?ooo`1D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol02P3oool00`3ooml0000004RLg`1F0?ooo`030?oOW0180000M;oo0140oooo000J0?ooo`03
0?oOW0180000M;oo0540oooo00L000000?ooo`3oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol02P3oool00`3OW4P0000007Boo`1A0?ooo`030000003oool0oooo05@0oooo00<0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030=nLB0000000M;oo05H0oooo00<0omnL
04P0001d_ol04@3oool001X0oooo00<0omnL04P0001d_ol0DP3oool01P000000oooo0?ooo`3ogi`0
B00007Boo`<0oooo00<0okmd0000B02Lgol01P3oool01`3Ogml0oooo0?oo_`1d000000000000M02o
ool0D@3oool00`000000oooo0?ooo`1D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol01P3oool01`3Ogml0oooo0?oo_`1d000000000000M02oool0EP3oool00`3ogi`0B00007Boo`0A
0?ooo`006P3oool00`3ogi`0B00007Boo`1B0?ooo`060000003oool0oooo0?oOW0180000M;oo0`3o
ool00`3o_g@0001809cOo`060?ooo`060=nLB000000000000000000004P0W=ooDP3oool00`000000
oooo0?ooo`1D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3OW4P0
0000000000000000001809cOoeL0oooo00<0omnL04P0001d_ol04@3oool001X0oooo00<0omnL04P0
001d_ol0D`3oool01@000000oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool0
103ookl0M00000000018W=mD0?ooo`030000003oool0oooo05@0oooo00<0oono07@00018W=l00`3o
ool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d00000000000005T0oooo00<0omnL04P0001d_ol0
4@3oool001X0oooo00<0omnL04P0001d_ol0D`3oool01P000000oooo0?ooo`3OW4P004RL0=ooo`80
oooo00<0_g@0001d_`3oool01`3oool00`2oM000M;oo0?ooo`1D0?ooo`<00000E@3oool00`3OW4P0
04RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0_g@00000003oool0F@3oool00`3ogi`0
B00007Boo`0A0?ooo`00603oool01@3ooml0W4P0000000000000M;oo05@0oooo00T000000?ooo`3o
okl0M00007Boo`3oool0okmd0000M02oool0203oool01P3ogi`0B000000000000000001809cOoe40
oooo00<000000?ooo`3oool0E@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0203oool0
1P3ogi`00000000000000000001809cOoe@0oooo00D0oooO09a800000000000007Boo`0A0?ooo`00
6P3oool00`3o_g@0000007Boo`1D0?ooo`080000003oool0oooo0?oo_`1d000000000000M02oool:
0?ooo`050;md00000000000000000018W=l0D@3oool00`000000oooo0?ooo`1F0?ooo`050?oo_`1d
000000000000M02oool02@3oool00`000000_g@0000000020000000304RLg`3oool0oooo05@0oooo
00<0okmd0000001d_ol04@3oool00780oooo00<000000?ooo`3oool0I03oool00`000000oooo0?oo
o`1T0?ooo`030000003oool0oooo06d0oooo001b0?ooo`030000003oool0oooo06@0oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`1^0?ooo`00L`3oool00`000000oooo0?ooo`1S0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool0KP3oool007<0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo06l0oooo001d0?ooo`030000
003oool0oooo0680oooo00<000000?ooo`3oool0HP3oool00`000000oooo0?ooo`1_0?ooo`00M03o
ool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool0L03o
ool007D0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0
oooo0700oooo001e0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool0H03oool00`00
0000oooo0?ooo`1a0?ooo`00MP3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo0600
oooo00<000000?ooo`3oool0L@3oool007H0oooo00<000000?ooo`3oool0H03oool00`000000oooo
0?ooo`1O0?ooo`030000003oool0oooo0780oooo000;0?ooool00000WP0000080?ooo`005@3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0140oooo0P00000B0?ooo`030000003oool0
oooo0100oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`8000004`3oool0
0`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0100oooo000E0?ooo`030000003oool0oooo
05l0oooo0P00001P0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`000000H@3oool00`00
0000oooo0?ooo`0@0?ooo`00N@3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo05d0
oooo00<000000?ooo`3oool0M03oool007T0oooo00<000000?ooo`3oool0G@3oool00`000000oooo
0?ooo`1L0?ooo`030000003oool0oooo07D0oooo001j0?ooo`030000003oool0oooo05`0oooo00<0
00000?ooo`3oool0G03oool00`000000oooo0?ooo`1e0?ooo`00NP3oool00`000000oooo0?ooo`1L
0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0MP3oool007/0oooo00<000000?oo
o`3oool0F`3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo07H0oooo001k0?ooo`03
0000003oool0oooo05/0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`1g0?ooo`00
O03oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool0
M`3oool007`0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`1I0?ooo`030000003o
ool0oooo07P0oooo001m0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3oool0F@3oool0
0`000000oooo0?ooo`1h0?ooo`00O@3oool00`000000oooo0?ooo`1I0?ooo`<00000F03oool00`00
0000oooo0?ooo`1i0?ooo`00OP3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo05P0
oooo00<000000?ooo`3oool0N@3oool007h0oooo00<000000?ooo`3oool0F03oool00`000000oooo
0?ooo`1G0?ooo`030000003oool0oooo07X0oooo001o0?ooo`030000003oool0oooo05L0oooo00<0
00000?ooo`3oool0E`3oool00`000000oooo0?ooo`1j0?ooo`00O`3oool00`000000oooo0?ooo`1G
0?ooo`030000003oool0oooo05H0oooo00<000000?ooo`3oool0N`3oool00800oooo00<000000?oo
o`3oool0EP3oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo07/0oooo00200?ooo`03
0000003oool0oooo05H0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`1l0?ooo`00
P@3oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0
O03oool00840oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`1D0?ooo`030000003o
ool0oooo07d0oooo00220?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0E03oool0
0`000000oooo0?ooo`1m0?ooo`00PP3oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo
05<0oooo00<000000?ooo`3oool0OP3oool008<0oooo00<000000?ooo`3oool0D`3oool00`000000
oooo0?ooo`1C0?ooo`030000003oool0oooo07h0oooo00230?ooo`030000003oool0oooo05<0oooo
0`00001B0?ooo`030000003oool0oooo07l0oooo00240?ooo`030000003oool0oooo0580oooo00<0
00000?ooo`3oool0DP3oool00`000000oooo0?ooo`1o0?ooo`00Q03oool00`000000oooo0?ooo`1B
0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool0P03oool008D0oooo00<000000?oo
o`3oool0D@3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo0800oooo00250?ooo`03
0000003oool0oooo0540oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`210?ooo`00
QP3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0
P@3oool008H0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`1?0?ooo`030000003o
ool0oooo0880oooo00270?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0CP3oool0
0`000000oooo0?ooo`230?ooo`00Q`3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo
04h0oooo00<000000?ooo`3oool0P`3oool008P0oooo00<000000?ooo`3oool0CP3oool00`000000
oooo0?ooo`1=0?ooo`030000003oool0oooo08@0oooo00290?ooo`030000003oool0oooo04d0oooo
00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`240?ooo`00R@3oool00`000000oooo0?oo
o`1=0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0Q@3oool008X0oooo00<00000
0?ooo`3oool0C03oool3000004`0oooo00<000000?ooo`3oool0Q@3oool008X0oooo00<000000?oo
o`3oool0C03oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo08H0oooo002;0?ooo`03
0000003oool0oooo04/0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`270?ooo`00
S03oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0
Q`3oool008`0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`190?ooo`030000003o
ool0oooo08P0oooo002=0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0B@3oool0
0`000000oooo0?ooo`280?ooo`00S@3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo
04P0oooo00<000000?ooo`3oool0R@3oool008h0oooo00<000000?ooo`3oool0B03oool00`000000
oooo0?ooo`180?ooo`030000003oool0oooo08T0oooo002>0?ooo`030000003oool0oooo04P0oooo
00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`2:0?ooo`00S`3oool00`000000oooo0?oo
o`170?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0R`3oool00900oooo00<00000
0?ooo`3oool0AP3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo08/0oooo002@0?oo
o`030000003oool0oooo04H0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`2<0?oo
o`00T@3oool00`000000oooo0?ooo`150?ooo`<00000A@3oool00`000000oooo0?ooo`2<0?ooo`00
T@3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0
S@3oool00980oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`140?ooo`030000003o
ool0oooo08d0oooo002C0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0@`3oool0
0`000000oooo0?ooo`2>0?ooo`00T`3oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo
04<0oooo00<000000?ooo`3oool0SP3oool009@0oooo00<000000?ooo`3oool0@P3oool00`000000
oooo0?ooo`120?ooo`030000003oool0oooo08l0oooo002D0?ooo`030000003oool0oooo0480oooo
00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`2@0?ooo`00U@3oool00`000000oooo0?oo
o`0T0?ooo`050?oOW018000000000000M02oool00`3oool01P3o_g@0000007Boo`3oool0oooO09a8
00<0000000<0M;oo0?ooo`3oool02@3oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo
0900oooo002E0?ooo`030000003oool0oooo02<0oooo00L0oono07@0001d_ol0oooo0?noM00007@0
_ooo0080oooo00L0okmd0000001d_ol0oooo0?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool0
2@3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo0940oooo002F0?ooo`030000003o
ool0oooo0280oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`0:0?ooo`030?oog`2L
B7@0_ooo00T0oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`2A0?ooo`00U`3oool0
0`000000oooo0?ooo`0P0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02`3oool0
0`2oM000M;oo0?ooo`080?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0TP3oool0
09L0oooo00<000000?ooo`3oool0803oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00X0oooo00<0oooO09a80018W=l02@3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo
0980oooo002H0?ooo`030000003oool0oooo01l0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`0:0?ooo`030=nLB0000000M;oo00T0oooo1000000m0?ooo`030000003oool0oooo09<0
oooo002H0?ooo`030000003oool0oooo01l0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`060?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`090?ooo`030000003oool0oooo
03d0oooo00<000000?ooo`3oool0U03oool009T0oooo00<000000?ooo`3oool07P3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00H0gia800000000000000000000B02Lgol:
0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool0U03oool009X0oooo00<000000?oo
o`3oool07@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono07@0
00000000B9cO303oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo09D0oooo002J0?oo
o`030000003oool0oooo01h0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?oo
o`030;md001d_ol0oooo00`0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`2F0?oo
o`00V`3oool00`000000oooo0?ooo`0M0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`08
0?ooo`060?oOW018000000000000000004P0W=oo2@3oool00`000000oooo0?ooo`0k0?ooo`030000
003oool0oooo09H0oooo002L0?ooo`030000003oool0oooo01d0oooo00D0oono07@000000000001d
0;ooo`0:0?ooo`050;md00000000000000000018W=l02@3oool00`000000oooo0?ooo`0j0?ooo`03
0000003oool0oooo09L0oooo002L0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0
>P3oool00`000000oooo0?ooo`2G0?ooo`00W@3oool00`000000oooo0?ooo`0i0?ooo`030000003o
ool0oooo03T0oooo00<000000?ooo`3oool0V03oool009h0oooo00<000000?ooo`3oool0>03oool0
0`000000oooo0?ooo`0h0?ooo`030000003oool0oooo09T0oooo002N0?ooo`030000003oool0oooo
03P0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`2I0?ooo`00W`3oool00`000000
oooo0?ooo`0g0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0VP3oool00:00oooo
00<000000?ooo`3oool0=P3oool3000003H0oooo00<000000?ooo`3oool0V`3oool00:00oooo00<0
00000?ooo`3oool0=P3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo09/0oooo002Q
0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`2L
0?ooo`00XP3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo03@0oooo00<000000?oo
o`3oool0W@3oool00:80oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0d0?ooo`03
0000003oool0oooo09d0oooo002S0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0
<`3oool00`000000oooo0?ooo`2N0?ooo`00Y03oool00`000000oooo0?ooo`0b0?ooo`030000003o
ool0oooo0380oooo00<000000?ooo`3oool0W`3oool00:@0oooo00<000000?ooo`3oool0<P3oool0
0`000000oooo0?ooo`0b0?ooo`030000003oool0oooo09l0oooo002U0?ooo`030000003oool0oooo
0340oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`2P0?ooo`00YP3oool00`000000
oooo0?ooo`0`0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0X03oool00:H0oooo
00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo0:40oooo
002W0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?oo
o`2R0?ooo`00Z03oool00`000000oooo0?ooo`0^0?ooo`<00000;`3oool00`000000oooo0?ooo`2R
0?ooo`00Z03oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo02h0oooo00<000000?oo
o`3oool0X`3oool00:T0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0]0?ooo`03
0000003oool0oooo0:@0oooo002Z0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0
;03oool00`000000oooo0?ooo`2U0?ooo`00Z`3oool00`000000oooo0?ooo`0[0?ooo`030000003o
ool0oooo02/0oooo00<000000?ooo`3oool0YP3oool00:/0oooo00<000000?ooo`3oool0:`3oool0
0`000000oooo0?ooo`0[0?ooo`030000003oool0oooo0:H0oooo002/0?ooo`030000003oool0oooo
02X0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`2W0?ooo`00[@3oool00`000000
oooo0?ooo`0Y0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0Z03oool00:h0oooo
00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo0:T0oooo
002_0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?oo
o`2Z0?ooo`00[`3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo02H0oooo00<00000
0?ooo`3oool0Z`3oool00;00oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0V0?oo
o`030000003oool0oooo0:/0oooo002a0?ooo`030000003oool0oooo02D0oooo0`00000U0?ooo`03
0000003oool0oooo0:`0oooo002b0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0
903oool00`000000oooo0?ooo`2]0?ooo`00/`3oool00`000000oooo0?ooo`0S0?ooo`030000003o
ool0oooo02<0oooo00<000000?ooo`3oool0[P3oool00;<0oooo00<000000?ooo`3oool08`3oool0
0`000000oooo0?ooo`0R0?ooo`030000003oool0oooo0:l0oooo002d0?ooo`030000003oool0oooo
0280oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`2`0?ooo`00]@3oool00`000000
oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0/03oool00;H0oooo
00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0;40oooo
002g0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?oo
o`2b0?ooo`00]`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01h0oooo00<00000
0?ooo`3oool0/`3oool00;P0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0M0?oo
o`030000003oool0oooo0;@0oooo002i0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3o
ool0703oool00`000000oooo0?ooo`2e0?ooo`00^P3oool00`000000oooo0?ooo`0L0?ooo`030000
003oool0oooo01/0oooo00<000000?ooo`3oool0]P3oool00;/0oooo0P00000L0?ooo`030000003o
ool0oooo01T0oooo0P00002i0?ooo`00_@3oool00`000000oooo0?ooo`0I0?ooo`<00000603oool0
0`000000oooo0?ooo`2i0?ooo`00_P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo
01L0oooo00<000000?ooo`3oool0^P3oool00;l0oooo0P00000H0?ooo`030000003oool0oooo01H0
oooo00<000000?ooo`3oool0^`3oool00<40oooo00<000000?ooo`3oool05@3oool00`000000oooo
0?ooo`0D0?ooo`800000_P3oool00<80oooo00<000000?ooo`3oool0503oool00`000000oooo0?oo
o`0C0?ooo`030000003oool0oooo0;h0oooo00330?ooo`800000503oool00`000000oooo0?ooo`0A
0?ooo`800000`@3oool00<D0oooo0P00000B0?ooo`030000003oool0oooo00l0oooo0P0000330?oo
o`00a`3oool2000000030?ooo`2oM00000000080000000<004RL0=ooo`3oool0203oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo0<<0oooo00390?ooo`80000000<0omnL04P0001d_ol0
2`3oool00`000000oooo0?ooo`0<0?ooo`800000aP3oool00</0oooo0P0000000`1d_ol0oooo0?oo
o`090?ooo`030000003oool0oooo00T0oooo0`0000380?ooo`00b`3oool00`3ogi`0B00000000003
000000P0oooo00<000000?ooo`3oool01P3oool300000</0oooo003;0?ooo`030?oOW0180000M;oo
00<0oooo1@0000030?ooo`030000003oool0oooo0080oooo1000003>0?ooo`00b`3oool00`3ogi`0
B00007Boo`080?ooo`P00000dP3oool00</0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo
0?ooo`3D0?ooo`00b`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0=@0oooo003;
0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0e03oool00</0oooo00<0omnL04P0
001d_ol02`3oool00`000000oooo0?ooo`3D0?ooo`00b@3oool01@3ooml0W4P0000000000000M;oo
00/0oooo00<000000?ooo`3oool0e03oool00</0oooo00<0okmd0000001d_ol02`3oool00`000000
oooo0?ooo`3D0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?oo
o`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.10918, -1.09709, \
0.00765073, 0.0123791}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(k\  = \ \[Pi];\)\), "\n", 
    \(\(Plot[Cos[k\ x], \ {x, \(-1\), 1}];\)\)}], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.309018 0.294303 [
[.02381 .29652 -9.15625 -15.6875 ]
[.02381 .29652 9.15625 0 ]
[.2619 .29652 -13.6563 -15.6875 ]
[.2619 .29652 13.6563 0 ]
[.7381 .29652 -9.5 -15.6875 ]
[.7381 .29652 9.5 0 ]
[.97619 .29652 -5 -15.6875 ]
[.97619 .29652 5 0 ]
[.4875 .01472 -18.3125 -7.84375 ]
[.4875 .01472 0 7.84375 ]
[.4875 .16187 -27.3125 -7.84375 ]
[.4875 .16187 0 7.84375 ]
[.4875 .45617 -19 -7.84375 ]
[.4875 .45617 0 7.84375 ]
[.4875 .60332 -10 -7.84375 ]
[.4875 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 .30902 m
.02381 .31527 L
s
gsave
.02381 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 .30902 m
.2619 .31527 L
s
gsave
.2619 .29652 -74.6563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.7381 .30902 m
.7381 .31527 L
s
gsave
.7381 .29652 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 .30902 m
.97619 .31527 L
s
gsave
.97619 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .30902 m
.07143 .31277 L
s
.11905 .30902 m
.11905 .31277 L
s
.16667 .30902 m
.16667 .31277 L
s
.21429 .30902 m
.21429 .31277 L
s
.30952 .30902 m
.30952 .31277 L
s
.35714 .30902 m
.35714 .31277 L
s
.40476 .30902 m
.40476 .31277 L
s
.45238 .30902 m
.45238 .31277 L
s
.54762 .30902 m
.54762 .31277 L
s
.59524 .30902 m
.59524 .31277 L
s
.64286 .30902 m
.64286 .31277 L
s
.69048 .30902 m
.69048 .31277 L
s
.78571 .30902 m
.78571 .31277 L
s
.83333 .30902 m
.83333 .31277 L
s
.88095 .30902 m
.88095 .31277 L
s
.92857 .30902 m
.92857 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5 .01472 m
.50625 .01472 L
s
gsave
.4875 .01472 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .16187 m
.50625 .16187 L
s
gsave
.4875 .16187 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .45617 m
.50625 .45617 L
s
gsave
.4875 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .60332 m
.50625 .60332 L
s
gsave
.4875 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .04415 m
.50375 .04415 L
s
.5 .07358 m
.50375 .07358 L
s
.5 .10301 m
.50375 .10301 L
s
.5 .13244 m
.50375 .13244 L
s
.5 .1913 m
.50375 .1913 L
s
.5 .22073 m
.50375 .22073 L
s
.5 .25016 m
.50375 .25016 L
s
.5 .27959 m
.50375 .27959 L
s
.5 .33845 m
.50375 .33845 L
s
.5 .36788 m
.50375 .36788 L
s
.5 .39731 m
.50375 .39731 L
s
.5 .42674 m
.50375 .42674 L
s
.5 .4856 m
.50375 .4856 L
s
.5 .51503 m
.50375 .51503 L
s
.5 .54446 m
.50375 .54446 L
s
.5 .57389 m
.50375 .57389 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01472 m
.02499 .01472 L
.02605 .01475 L
.02729 .01479 L
.02846 .01485 L
.03053 .015 L
.03279 .01523 L
.03527 .01556 L
.0379 .01599 L
.04262 .01698 L
.04749 .0183 L
.05205 .01981 L
.06244 .02422 L
.07305 .03011 L
.08274 .03668 L
.10458 .05552 L
.14429 .10289 L
.18248 .16177 L
.22313 .23454 L
.26226 .3097 L
.30384 .3894 L
.34391 .46058 L
.38246 .51918 L
.402 .54393 L
.42346 .56659 L
.44388 .58338 L
.45301 .58929 L
.46295 .59457 L
.46838 .59694 L
.4734 .5988 L
.4781 .60026 L
.4833 .60154 L
.48553 .60198 L
.48794 .60239 L
.49002 .60268 L
.4923 .60294 L
.49473 .60314 L
.49605 .60322 L
.49728 .60327 L
.49844 .6033 L
.49949 .60332 L
.50071 .60332 L
.50186 .6033 L
.50301 .60326 L
.50425 .60321 L
.50542 .60313 L
.50648 .60305 L
.5091 .60279 L
.51158 .60246 L
.5165 .60158 L
Mistroke
.52187 .60026 L
.53144 .59701 L
.54032 .59297 L
.56088 .5799 L
.5795 .56376 L
.61962 .51633 L
.65822 .45702 L
.69928 .38356 L
.73882 .30761 L
.77684 .2346 L
.81732 .1621 L
.85628 .1021 L
.87613 .07654 L
.89769 .0533 L
.91843 .03582 L
.92832 .02927 L
.93759 .02421 L
.94643 .02037 L
.95595 .01733 L
.9611 .01617 L
.96369 .01572 L
.96652 .01531 L
.96894 .01505 L
.9703 .01494 L
.97153 .01485 L
.97268 .01479 L
.97393 .01475 L
.97511 .01472 L
.97619 .01472 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00f@3oool0
0`000000oooo0?ooo`3D0?ooo`00bP3oool01P2oM000000000000000000004RL0=ooo`T0oooo00<0
00000?ooo`3oool0e03oool00</0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`3D
0?ooo`00b`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0=@0oooo003;0?ooo`03
0?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0e03oool00;/0oooo00<0oooO09a800000000
1`0000000`18W=l0oooo0?ooo`030?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0
e03oool001D0oooo0`00002c0?ooo`030?oOW0180000M;oo00/0oooo1000002n0?ooo`<000004P3o
ool001P0oooo1P00002]0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0^@3oool6
000001D0oooo000N0?ooo`@00000Z@3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo
0;L0oooo0P00000K0?ooo`008P3oool200000:L0oooo00<0omnL04P0001d_ol02`3oool00`000000
oooo0?ooo`2d0?ooo`<000007@3oool002@0oooo0P00002U0?ooo`030?oOW0180000M;oo00/0oooo
00<000000?ooo`3oool0/@3oool300000200oooo000V0?ooo`030000003oool0oooo0:00oooo00D0
oooO09a800000000000007Boo`0;0?ooo`030000003oool0oooo0:l0oooo0P00000S0?ooo`009`3o
ool200000:80oooo00<0okmd0000001d_ol02`3oool00`000000oooo0?ooo`2]0?ooo`8000009@3o
ool002T0oooo00<000000?ooo`3oool0[@3oool00`000000oooo0?ooo`2/0?ooo`030000003oool0
oooo02D0oooo000Z0?ooo`800000[@3oool00`000000oooo0?ooo`2Z0?ooo`800000:03oool002`0
oooo0P00002[0?ooo`030000003oool0oooo0:T0oooo00<000000?ooo`3oool0:03oool002h0oooo
00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo02T0oooo
000_0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?oo
o`0Z0?ooo`00<03oool00`000000oooo0?ooo`2V0?ooo`030000003oool0oooo0:H0oooo00<00000
0?ooo`3oool0:`3oool00340oooo0P00002V0?ooo`<00000Y03oool2000002h0oooo000c0?ooo`03
0000003oool0oooo0:<0oooo00<000000?ooo`3oool0X`3oool00`000000oooo0?ooo`0^0?ooo`00
=03oool00`000000oooo0?ooo`2R0?ooo`030000003oool0oooo0:80oooo00<000000?ooo`3oool0
;`3oool003D0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo0300oooo000f0?ooo`030000003oool0oooo0:00oooo00<000000?ooo`3oool0X03oool0
0`000000oooo0?ooo`0a0?ooo`00=`3oool00`000000oooo0?ooo`2O0?ooo`030000003oool0oooo
09l0oooo00<000000?ooo`3oool0<P3oool003P0oooo00<000000?ooo`3oool0WP3oool00`000000
oooo0?ooo`2N0?ooo`030000003oool0oooo03<0oooo000i0?ooo`030000003oool0oooo09d0oooo
00<000000?ooo`3oool0W@3oool00`000000oooo0?ooo`0d0?ooo`00>@3oool00`000000oooo0?oo
o`2M0?ooo`030000003oool0oooo09`0oooo00<000000?ooo`3oool0=@3oool003X0oooo00<00000
0?ooo`3oool0W03oool00`000000oooo0?ooo`2L0?ooo`030000003oool0oooo03D0oooo000k0?oo
o`030000003oool0oooo09/0oooo00<000000?ooo`3oool0V`3oool00`000000oooo0?ooo`0f0?oo
o`00?03oool00`000000oooo0?ooo`2J0?ooo`030000003oool0oooo09X0oooo00<000000?ooo`3o
ool0=`3oool003d0oooo00<000000?ooo`3oool0V@3oool3000009T0oooo00<000000?ooo`3oool0
>03oool003h0oooo00<000000?ooo`3oool0V03oool00`000000oooo0?ooo`2H0?ooo`030000003o
ool0oooo03T0oooo000n0?ooo`030000003oool0oooo09P0oooo00<000000?ooo`3oool0U`3oool0
0`000000oooo0?ooo`0j0?ooo`00?`3oool00`000000oooo0?ooo`2G0?ooo`030000003oool0oooo
09H0oooo00<000000?ooo`3oool0>`3oool00400oooo00<000000?ooo`3oool0UP3oool00`000000
oooo0?ooo`2F0?ooo`030000003oool0oooo03/0oooo00110?ooo`030000003oool0oooo09D0oooo
00<000000?ooo`3oool0U@3oool00`000000oooo0?ooo`0l0?ooo`00@P3oool00`000000oooo0?oo
o`2D0?ooo`030000003oool0oooo09@0oooo00<000000?ooo`3oool0?@3oool004<0oooo00<00000
0?ooo`3oool0T`3oool00`000000oooo0?ooo`2C0?ooo`030000003oool0oooo03h0oooo00140?oo
o`030000003oool0oooo0980oooo00<000000?ooo`3oool0TP3oool00`000000oooo0?ooo`0o0?oo
o`00A03oool00`000000oooo0?ooo`2B0?ooo`030000003oool0oooo0980oooo00<000000?ooo`3o
ool0?`3oool004D0oooo00<000000?ooo`3oool0T@3oool00`000000oooo0?ooo`2A0?ooo`030000
003oool0oooo0400oooo00160?ooo`030000003oool0oooo0900oooo00<000000?ooo`3oool0T03o
ool00`000000oooo0?ooo`110?ooo`00A`3oool00`000000oooo0?ooo`2?0?ooo`<00000S`3oool0
0`000000oooo0?ooo`120?ooo`00B03oool00`000000oooo0?ooo`2>0?ooo`030000003oool0oooo
08l0oooo00<000000?ooo`3oool0@P3oool004P0oooo00<000000?ooo`3oool0SP3oool00`000000
oooo0?ooo`2>0?ooo`030000003oool0oooo04<0oooo00190?ooo`030000003oool0oooo08d0oooo
00<000000?ooo`3oool0S@3oool00`000000oooo0?ooo`140?ooo`00B@3oool00`000000oooo0?oo
o`2=0?ooo`030000003oool0oooo08d0oooo00<000000?ooo`3oool0A03oool004X0oooo00<00000
0?ooo`3oool0S03oool00`000000oooo0?ooo`2<0?ooo`030000003oool0oooo04D0oooo001;0?oo
o`030000003oool0oooo08/0oooo00<000000?ooo`3oool0S03oool00`000000oooo0?ooo`150?oo
o`00B`3oool00`000000oooo0?ooo`2;0?ooo`030000003oool0oooo08/0oooo00<000000?ooo`3o
ool0AP3oool004`0oooo00<000000?ooo`3oool0RP3oool00`000000oooo0?ooo`2:0?ooo`030000
003oool0oooo04L0oooo001=0?ooo`030000003oool0oooo08T0oooo00<000000?ooo`3oool0RP3o
ool00`000000oooo0?ooo`170?ooo`00C@3oool00`000000oooo0?ooo`290?ooo`030000003oool0
oooo08T0oooo00<000000?ooo`3oool0B03oool004h0oooo00<000000?ooo`3oool0R03oool00`00
0000oooo0?ooo`280?ooo`030000003oool0oooo04T0oooo001>0?ooo`030000003oool0oooo08P0
oooo0`0000280?ooo`030000003oool0oooo04T0oooo001?0?ooo`030000003oool0oooo08L0oooo
00<000000?ooo`3oool0Q`3oool00`000000oooo0?ooo`1:0?ooo`00D03oool00`000000oooo0?oo
o`260?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3oool0B`3oool00500oooo00<00000
0?ooo`3oool0QP3oool00`000000oooo0?ooo`260?ooo`030000003oool0oooo04/0oooo001A0?oo
o`030000003oool0oooo08D0oooo00<000000?ooo`3oool0Q@3oool00`000000oooo0?ooo`1<0?oo
o`00DP3oool00`000000oooo0?ooo`240?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3o
ool0C@3oool00580oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`240?ooo`030000
003oool0oooo04d0oooo001C0?ooo`030000003oool0oooo06H0oooo00D0omnL04P000000000001d
0;ooo`030?ooo`060?noM0000000M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`090?oo
o`030000003oool0oooo08<0oooo00<000000?ooo`3oool0CP3oool005<0oooo00<000000?ooo`3o
ool0I@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@0000007Boo`3o
ool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`090?ooo`030000003oool0oooo08<0oooo
00<000000?ooo`3oool0CP3oool005@0oooo00<000000?ooo`3oool0I03oool00`2oM00004RL0=oo
o`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO09a8M02oool02@3oool00`000000oooo0?oo
o`220?ooo`030000003oool0oooo04l0oooo001E0?ooo`030000003oool0oooo0680oooo00<0oono
07@00018W=l00`3oool00`3OW4P0001d0;ooo`0;0?ooo`030;md001d_ol0oooo00P0oooo00<00000
0?ooo`3oool0P@3oool00`000000oooo0?ooo`1@0?ooo`00E@3oool00`000000oooo0?ooo`1F0?oo
o`030?oog`2LB000000000L0000000D0B9cO0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo00X0oooo00<0oooO09a80018W=l02@3oool00`000000oooo0?ooo`210?ooo`030000003oool0
oooo0500oooo001F0?ooo`030000003oool0oooo0640oooo00<0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`0:0?ooo`030=nLB0000000M;oo00T0oooo1000001o0?ooo`030000003oool0oooo
0540oooo001G0?ooo`030000003oool0oooo0600oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`060?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`090?ooo`030000003oool0
oooo07l0oooo00<000000?ooo`3oool0DP3oool005L0oooo00<000000?ooo`3oool0H03oool00`3o
gi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00H0gia800000000000000000000B02L
gol:0?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool0DP3oool005P0oooo00<00000
0?ooo`3oool0G`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono
07@000000000B9cO303oool00`000000oooo0?ooo`1n0?ooo`030000003oool0oooo05<0oooo001H
0?ooo`030000003oool0oooo0600oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`07
0?ooo`030;md001d_ol0oooo00`0oooo00<000000?ooo`3oool0OP3oool00`000000oooo0?ooo`1C
0?ooo`00F@3oool00`000000oooo0?ooo`1O0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oo
o`080?ooo`060?oOW018000000000000000004P0W=oo2@3oool00`000000oooo0?ooo`1m0?ooo`03
0000003oool0oooo05@0oooo001I0?ooo`030000003oool0oooo0600oooo00D0oono07@000000000
001d0;ooo`0:0?ooo`050;md00000000000000000018W=l02@3oool00`000000oooo0?ooo`1m0?oo
o`030000003oool0oooo05@0oooo001J0?ooo`030000003oool0oooo07`0oooo00<000000?ooo`3o
ool0O03oool00`000000oooo0?ooo`1E0?ooo`00F`3oool00`000000oooo0?ooo`1k0?ooo`030000
003oool0oooo07/0oooo00<000000?ooo`3oool0EP3oool005/0oooo00<000000?ooo`3oool0N`3o
ool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo05H0oooo001L0?ooo`030000003oool0
oooo07X0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`1G0?ooo`00G03oool00`00
0000oooo0?ooo`1j0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0E`3oool005d0
oooo00<000000?ooo`3oool0N@3oool3000007T0oooo00<000000?ooo`3oool0F03oool005d0oooo
00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo05P0oooo
001N0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0N03oool00`000000oooo0?oo
o`1I0?ooo`00GP3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo07L0oooo00<00000
0?ooo`3oool0FP3oool005l0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`1g0?oo
o`030000003oool0oooo05X0oooo001P0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3o
ool0MP3oool00`000000oooo0?ooo`1K0?ooo`00H03oool00`000000oooo0?ooo`1f0?ooo`030000
003oool0oooo07H0oooo00<000000?ooo`3oool0F`3oool00640oooo00<000000?ooo`3oool0M@3o
ool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo05`0oooo001Q0?ooo`030000003oool0
oooo07D0oooo00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`1L0?ooo`00HP3oool00`00
0000oooo0?ooo`1d0?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3oool0G@3oool00680
oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`1d0?ooo`030000003oool0oooo05d0
oooo001S0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool0L`3oool00`000000oooo
0?ooo`1N0?ooo`00I03oool00`000000oooo0?ooo`1b0?ooo`<00000LP3oool00`000000oooo0?oo
o`1O0?ooo`00I03oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo0780oooo00<00000
0?ooo`3oool0G`3oool006D0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`1a0?oo
o`030000003oool0oooo0600oooo001U0?ooo`030000003oool0oooo0740oooo00<000000?ooo`3o
ool0L@3oool00`000000oooo0?ooo`1P0?ooo`00IP3oool00`000000oooo0?ooo`1`0?ooo`030000
003oool0oooo0700oooo00<000000?ooo`3oool0H@3oool006H0oooo00<000000?ooo`3oool0L03o
ool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo0640oooo001W0?ooo`030000003oool0
oooo06l0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`1R0?ooo`00J03oool00`00
0000oooo0?ooo`1^0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool0H`3oool006P0
oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`1^0?ooo`030000003oool0oooo06<0
oooo001Y0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0K@3oool00`000000oooo
0?ooo`1T0?ooo`00J@3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo06d0oooo00<0
00000?ooo`3oool0I03oool006X0oooo00<000000?ooo`3oool0K03oool00`000000oooo0?ooo`1/
0?ooo`030000003oool0oooo06D0oooo001Z0?ooo`030000003oool0oooo06`0oooo0`00001/0?oo
o`030000003oool0oooo06D0oooo001[0?ooo`030000003oool0oooo06/0oooo00<000000?ooo`3o
ool0J`3oool00`000000oooo0?ooo`1V0?ooo`00J`3oool00`000000oooo0?ooo`1[0?ooo`030000
003oool0oooo06/0oooo00<000000?ooo`3oool0IP3oool001T0oooo00H0_g@00000000000000000
0018W03Ooom=0?ooo`030000003oool0oooo00D0oooo00D0omnL04P000000000001d0;ooo`030?oo
o`060?noM0000000M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`1A0?ooo`030000003o
ool0oooo05H0oooo00D0omnL04P000000000001d0;ooo`030?ooo`060?noM0000000M;oo0?ooo`3o
oml0W4P00`000000101d_ol0oooo0?ooo`00001D0?ooo`060;md00000000000000000000B9`0gooo
3`3oool001X0oooo00<0omnL04P0001d_ol0C`3oool00`000000oooo0?ooo`040?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo0?ooo`3ooml0W4Qd0;ooo`02
0?ooo`030;md001d_ol0oooo0540oooo00<000000?ooo`3oool0E@3oool01`3ookl0M00007Boo`3o
ool0okmd0000M02oool00P3oool01`3o_g@0000007Boo`3oool0oooO09a8M02oool00P3oool0102o
M000M;oo0?ooo`00001E0?ooo`030?oOW0180000M;oo0140oooo000J0?ooo`030?oOW0180000M;oo
0500oooo00<000000?ooo`3oool00`3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo
00X0oooo00<0oooO09a8M02oool0D@3oool00`000000oooo0?ooo`1E0?ooo`030;md0000B9`0gooo
0080oooo00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd0000001F0?ooo`030?oOW0180000M;oo
0140oooo000J0?ooo`030?oOW0180000M;oo0500oooo00<000000?ooo`3oool00P3oool00`3ookl0
M00004RLg`030?ooo`030=nLB00007@0_ooo00/0oooo00<0_g@007Boo`3oool0D03oool00`000000
oooo0?ooo`1D0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02`3oool00`2oM000
00000?ooo`1E0?ooo`030?oOW0180000M;oo0140oooo000:0?ooo`030?oog`2LB000000000L00000
00<0B9cO0?ooo`3oool00`3oool00`3ogi`0B00007Boo`180?ooo`030?oog`2LB000000000L00000
00H0B9cO0?ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo00X0oooo00<0oooO09a8
0018W=l0D@3oool00`000000oooo0?ooo`1D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol02P3oool00`3ooml0000004RLg`1F0?ooo`030?oOW0180000M;oo0140oooo000J0?ooo`03
0?oOW0180000M;oo0540oooo00L000000?ooo`3oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol02P3oool00`3OW4P0000007Boo`1A0?ooo`030000003oool0oooo05@0oooo00<0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030=nLB0000000M;oo05H0oooo00<0omnL
04P0001d_ol04@3oool001X0oooo00<0omnL04P0001d_ol0DP3oool01P000000oooo0?ooo`3ogi`0
B00007Boo`<0oooo00<0okmd0000B02Lgol01P3oool01`3Ogml0oooo0?oo_`1d000000000000M02o
ool0D@3oool00`000000oooo0?ooo`1D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol01P3oool01`3Ogml0oooo0?oo_`1d000000000000M02oool0EP3oool00`3ogi`0B00007Boo`0A
0?ooo`006P3oool00`3ogi`0B00007Boo`1B0?ooo`060000003oool0oooo0?oOW0180000M;oo0`3o
ool00`3o_g@0001809cOo`060?ooo`060=nLB000000000000000000004P0W=ooDP3oool00`000000
oooo0?ooo`1D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3OW4P0
0000000000000000001809cOoeL0oooo00<0omnL04P0001d_ol04@3oool001X0oooo00<0omnL04P0
001d_ol0D`3oool01@000000oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool0
103ookl0M00000000018W=mD0?ooo`030000003oool0oooo05@0oooo00<0oono07@00018W=l00`3o
ool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d00000000000005T0oooo00<0omnL04P0001d_ol0
4@3oool001X0oooo00<0omnL04P0001d_ol0D`3oool01P000000oooo0?ooo`3OW4P004RL0=ooo`80
oooo00<0_g@0001d_`3oool01`3oool00`2oM000M;oo0?ooo`1D0?ooo`<00000E@3oool00`3OW4P0
04RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0_g@00000003oool0F@3oool00`3ogi`0
B00007Boo`0A0?ooo`00603oool01@3ooml0W4P0000000000000M;oo05@0oooo00T000000?ooo`3o
okl0M00007Boo`3oool0okmd0000M02oool0203oool01P3ogi`0B000000000000000001809cOoe40
oooo00<000000?ooo`3oool0E@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0203oool0
1P3ogi`00000000000000000001809cOoe@0oooo00D0oooO09a800000000000007Boo`0A0?ooo`00
6P3oool00`3o_g@0000007Boo`1D0?ooo`080000003oool0oooo0?oo_`1d000000000000M02oool:
0?ooo`050;md00000000000000000018W=l0D@3oool00`000000oooo0?ooo`1F0?ooo`050?oo_`1d
000000000000M02oool02@3oool00`000000_g@0000000020000000304RLg`3oool0oooo05@0oooo
00<0okmd0000001d_ol04@3oool00780oooo00<000000?ooo`3oool0I03oool00`000000oooo0?oo
o`1T0?ooo`030000003oool0oooo06d0oooo001b0?ooo`030000003oool0oooo06@0oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`1^0?ooo`00L`3oool00`000000oooo0?ooo`1S0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool0KP3oool007<0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo06l0oooo001d0?ooo`030000
003oool0oooo0680oooo00<000000?ooo`3oool0HP3oool00`000000oooo0?ooo`1_0?ooo`00M03o
ool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool0L03o
ool007D0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0
oooo0700oooo001e0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool0H03oool00`00
0000oooo0?ooo`1a0?ooo`00MP3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo0600
oooo00<000000?ooo`3oool0L@3oool007H0oooo00<000000?ooo`3oool0H03oool00`000000oooo
0?ooo`1O0?ooo`030000003oool0oooo0780oooo000;0?ooool00000WP0000080?ooo`005@3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0140oooo0P00000B0?ooo`030000003oool0
oooo0100oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`8000004`3oool0
0`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0100oooo000E0?ooo`030000003oool0oooo
05l0oooo0P00001P0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`000000H@3oool00`00
0000oooo0?ooo`0@0?ooo`00N@3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo05d0
oooo00<000000?ooo`3oool0M03oool007T0oooo00<000000?ooo`3oool0G@3oool00`000000oooo
0?ooo`1L0?ooo`030000003oool0oooo07D0oooo001j0?ooo`030000003oool0oooo05`0oooo00<0
00000?ooo`3oool0G03oool00`000000oooo0?ooo`1e0?ooo`00NP3oool00`000000oooo0?ooo`1L
0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0MP3oool007/0oooo00<000000?oo
o`3oool0F`3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo07H0oooo001k0?ooo`03
0000003oool0oooo05/0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`1g0?ooo`00
O03oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool0
M`3oool007`0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`1I0?ooo`030000003o
ool0oooo07P0oooo001m0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3oool0F@3oool0
0`000000oooo0?ooo`1h0?ooo`00O@3oool00`000000oooo0?ooo`1I0?ooo`<00000F03oool00`00
0000oooo0?ooo`1i0?ooo`00OP3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo05P0
oooo00<000000?ooo`3oool0N@3oool007h0oooo00<000000?ooo`3oool0F03oool00`000000oooo
0?ooo`1G0?ooo`030000003oool0oooo07X0oooo001o0?ooo`030000003oool0oooo05L0oooo00<0
00000?ooo`3oool0E`3oool00`000000oooo0?ooo`1j0?ooo`00O`3oool00`000000oooo0?ooo`1G
0?ooo`030000003oool0oooo05H0oooo00<000000?ooo`3oool0N`3oool00800oooo00<000000?oo
o`3oool0EP3oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo07/0oooo00200?ooo`03
0000003oool0oooo05H0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`1l0?ooo`00
P@3oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0
O03oool00840oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`1D0?ooo`030000003o
ool0oooo07d0oooo00220?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0E03oool0
0`000000oooo0?ooo`1m0?ooo`00PP3oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo
05<0oooo00<000000?ooo`3oool0OP3oool008<0oooo00<000000?ooo`3oool0D`3oool00`000000
oooo0?ooo`1C0?ooo`030000003oool0oooo07h0oooo00230?ooo`030000003oool0oooo05<0oooo
0`00001B0?ooo`030000003oool0oooo07l0oooo00240?ooo`030000003oool0oooo0580oooo00<0
00000?ooo`3oool0DP3oool00`000000oooo0?ooo`1o0?ooo`00Q03oool00`000000oooo0?ooo`1B
0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool0P03oool008D0oooo00<000000?oo
o`3oool0D@3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo0800oooo00250?ooo`03
0000003oool0oooo0540oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`210?ooo`00
QP3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0
P@3oool008H0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`1?0?ooo`030000003o
ool0oooo0880oooo00270?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0CP3oool0
0`000000oooo0?ooo`230?ooo`00Q`3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo
04h0oooo00<000000?ooo`3oool0P`3oool008P0oooo00<000000?ooo`3oool0CP3oool00`000000
oooo0?ooo`1=0?ooo`030000003oool0oooo08@0oooo00290?ooo`030000003oool0oooo04d0oooo
00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`240?ooo`00R@3oool00`000000oooo0?oo
o`1=0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0Q@3oool008X0oooo00<00000
0?ooo`3oool0C03oool3000004`0oooo00<000000?ooo`3oool0Q@3oool008X0oooo00<000000?oo
o`3oool0C03oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo08H0oooo002;0?ooo`03
0000003oool0oooo04/0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`270?ooo`00
S03oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0
Q`3oool008`0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`190?ooo`030000003o
ool0oooo08P0oooo002=0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0B@3oool0
0`000000oooo0?ooo`280?ooo`00S@3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo
04P0oooo00<000000?ooo`3oool0R@3oool008h0oooo00<000000?ooo`3oool0B03oool00`000000
oooo0?ooo`180?ooo`030000003oool0oooo08T0oooo002>0?ooo`030000003oool0oooo04P0oooo
00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`2:0?ooo`00S`3oool00`000000oooo0?oo
o`170?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0R`3oool00900oooo00<00000
0?ooo`3oool0AP3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo08/0oooo002@0?oo
o`030000003oool0oooo04H0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`2<0?oo
o`00T@3oool00`000000oooo0?ooo`150?ooo`<00000A@3oool00`000000oooo0?ooo`2<0?ooo`00
T@3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0
S@3oool00980oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`140?ooo`030000003o
ool0oooo08d0oooo002C0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0@`3oool0
0`000000oooo0?ooo`2>0?ooo`00T`3oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo
04<0oooo00<000000?ooo`3oool0SP3oool009@0oooo00<000000?ooo`3oool0@P3oool00`000000
oooo0?ooo`120?ooo`030000003oool0oooo08l0oooo002D0?ooo`030000003oool0oooo0480oooo
00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`2@0?ooo`00U@3oool00`000000oooo0?oo
o`0T0?ooo`050?oOW018000000000000M02oool00`3oool01P3o_g@0000007Boo`3oool0oooO09a8
00<0000000<0M;oo0?ooo`3oool02@3oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo
0900oooo002E0?ooo`030000003oool0oooo02<0oooo00L0oono07@0001d_ol0oooo0?noM00007@0
_ooo0080oooo00L0okmd0000001d_ol0oooo0?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool0
2@3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo0940oooo002F0?ooo`030000003o
ool0oooo0280oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`0:0?ooo`030?oog`2L
B7@0_ooo00T0oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`2A0?ooo`00U`3oool0
0`000000oooo0?ooo`0P0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02`3oool0
0`2oM000M;oo0?ooo`080?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0TP3oool0
09L0oooo00<000000?ooo`3oool0803oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00X0oooo00<0oooO09a80018W=l02@3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo
0980oooo002H0?ooo`030000003oool0oooo01l0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`0:0?ooo`030=nLB0000000M;oo00T0oooo1000000m0?ooo`030000003oool0oooo09<0
oooo002H0?ooo`030000003oool0oooo01l0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`060?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`090?ooo`030000003oool0oooo
03d0oooo00<000000?ooo`3oool0U03oool009T0oooo00<000000?ooo`3oool07P3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00H0gia800000000000000000000B02Lgol:
0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool0U03oool009X0oooo00<000000?oo
o`3oool07@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono07@0
00000000B9cO303oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo09D0oooo002J0?oo
o`030000003oool0oooo01h0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?oo
o`030;md001d_ol0oooo00`0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`2F0?oo
o`00V`3oool00`000000oooo0?ooo`0M0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`08
0?ooo`060?oOW018000000000000000004P0W=oo2@3oool00`000000oooo0?ooo`0k0?ooo`030000
003oool0oooo09H0oooo002L0?ooo`030000003oool0oooo01d0oooo00D0oono07@000000000001d
0;ooo`0:0?ooo`050;md00000000000000000018W=l02@3oool00`000000oooo0?ooo`0j0?ooo`03
0000003oool0oooo09L0oooo002L0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0
>P3oool00`000000oooo0?ooo`2G0?ooo`00W@3oool00`000000oooo0?ooo`0i0?ooo`030000003o
ool0oooo03T0oooo00<000000?ooo`3oool0V03oool009h0oooo00<000000?ooo`3oool0>03oool0
0`000000oooo0?ooo`0h0?ooo`030000003oool0oooo09T0oooo002N0?ooo`030000003oool0oooo
03P0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`2I0?ooo`00W`3oool00`000000
oooo0?ooo`0g0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0VP3oool00:00oooo
00<000000?ooo`3oool0=P3oool3000003H0oooo00<000000?ooo`3oool0V`3oool00:00oooo00<0
00000?ooo`3oool0=P3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo09/0oooo002Q
0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`2L
0?ooo`00XP3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo03@0oooo00<000000?oo
o`3oool0W@3oool00:80oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0d0?ooo`03
0000003oool0oooo09d0oooo002S0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0
<`3oool00`000000oooo0?ooo`2N0?ooo`00Y03oool00`000000oooo0?ooo`0b0?ooo`030000003o
ool0oooo0380oooo00<000000?ooo`3oool0W`3oool00:@0oooo00<000000?ooo`3oool0<P3oool0
0`000000oooo0?ooo`0b0?ooo`030000003oool0oooo09l0oooo002U0?ooo`030000003oool0oooo
0340oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`2P0?ooo`00YP3oool00`000000
oooo0?ooo`0`0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0X03oool00:H0oooo
00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo0:40oooo
002W0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?oo
o`2R0?ooo`00Z03oool00`000000oooo0?ooo`0^0?ooo`<00000;`3oool00`000000oooo0?ooo`2R
0?ooo`00Z03oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo02h0oooo00<000000?oo
o`3oool0X`3oool00:T0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0]0?ooo`03
0000003oool0oooo0:@0oooo002Z0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0
;03oool00`000000oooo0?ooo`2U0?ooo`00Z`3oool00`000000oooo0?ooo`0[0?ooo`030000003o
ool0oooo02/0oooo00<000000?ooo`3oool0YP3oool00:/0oooo00<000000?ooo`3oool0:`3oool0
0`000000oooo0?ooo`0[0?ooo`030000003oool0oooo0:H0oooo002/0?ooo`030000003oool0oooo
02X0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`2W0?ooo`00[@3oool00`000000
oooo0?ooo`0Y0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0Z03oool00:h0oooo
00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo0:T0oooo
002_0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?oo
o`2Z0?ooo`00[`3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo02H0oooo00<00000
0?ooo`3oool0Z`3oool00;00oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0V0?oo
o`030000003oool0oooo0:/0oooo002a0?ooo`030000003oool0oooo02D0oooo0`00000U0?ooo`03
0000003oool0oooo0:`0oooo002b0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0
903oool00`000000oooo0?ooo`2]0?ooo`00/`3oool00`000000oooo0?ooo`0S0?ooo`030000003o
ool0oooo02<0oooo00<000000?ooo`3oool0[P3oool00;<0oooo00<000000?ooo`3oool08`3oool0
0`000000oooo0?ooo`0R0?ooo`030000003oool0oooo0:l0oooo002d0?ooo`030000003oool0oooo
0280oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`2`0?ooo`00]@3oool00`000000
oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0/03oool00;H0oooo
00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0;40oooo
002g0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?oo
o`2b0?ooo`00]`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01h0oooo00<00000
0?ooo`3oool0/`3oool00;P0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0M0?oo
o`030000003oool0oooo0;@0oooo002i0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3o
ool0703oool00`000000oooo0?ooo`2e0?ooo`00^P3oool00`000000oooo0?ooo`0L0?ooo`030000
003oool0oooo01/0oooo00<000000?ooo`3oool0]P3oool00;/0oooo0P00000L0?ooo`030000003o
ool0oooo01T0oooo0P00002i0?ooo`00_@3oool00`000000oooo0?ooo`0I0?ooo`<00000603oool0
0`000000oooo0?ooo`2i0?ooo`00_P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo
01L0oooo00<000000?ooo`3oool0^P3oool00;l0oooo0P00000H0?ooo`030000003oool0oooo01H0
oooo00<000000?ooo`3oool0^`3oool00<40oooo00<000000?ooo`3oool05@3oool00`000000oooo
0?ooo`0D0?ooo`800000_P3oool00<80oooo00<000000?ooo`3oool0503oool00`000000oooo0?oo
o`0C0?ooo`030000003oool0oooo0;h0oooo00330?ooo`800000503oool00`000000oooo0?ooo`0A
0?ooo`800000`@3oool00<D0oooo0P00000B0?ooo`030000003oool0oooo00l0oooo0P0000330?oo
o`00a`3oool2000000030?ooo`2oM00000000080000000<004RL0=ooo`3oool0203oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo0<<0oooo00390?ooo`80000000<0omnL04P0001d_ol0
2`3oool00`000000oooo0?ooo`0<0?ooo`800000aP3oool00</0oooo0P0000000`1d_ol0oooo0?oo
o`090?ooo`030000003oool0oooo00T0oooo0`0000380?ooo`00b`3oool00`3ogi`0B00000000003
000000P0oooo00<000000?ooo`3oool01P3oool300000</0oooo003;0?ooo`030?oOW0180000M;oo
00<0oooo1@0000030?ooo`030000003oool0oooo0080oooo1000003>0?ooo`00b`3oool00`3ogi`0
B00007Boo`080?ooo`P00000dP3oool00</0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo
0?ooo`3D0?ooo`00b`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0=@0oooo003;
0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0e03oool00</0oooo00<0omnL04P0
001d_ol02`3oool00`000000oooo0?ooo`3D0?ooo`00b@3oool01@3ooml0W4P0000000000000M;oo
00/0oooo00<000000?ooo`3oool0e03oool00</0oooo00<0okmd0000001d_ol02`3oool00`000000
oooo0?ooo`3D0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?oo
o`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.10918, -1.09709, \
0.00765073, 0.0123791}}]
}, Closed]],

Cell[TextData[{
  "Now that ",
  StyleBox["k", "InlineInput"],
  " has been set, it can be used freely.  However, we should remember to \
clear it if we ever want to use it as a symbol."
}], "Text"],

Cell[BoxData[
    \(f[x_]\  := \ \(\(x\)\(\ \)\(Cos[\[Pi]\ x]\)\(\ \)\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[f[x], \ {x, \(-1\), 1}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.309017 0.294302 [
[.02381 .29652 -9.15625 -15.6875 ]
[.02381 .29652 9.15625 0 ]
[.2619 .29652 -13.6563 -15.6875 ]
[.2619 .29652 13.6563 0 ]
[.7381 .29652 -9.5 -15.6875 ]
[.7381 .29652 9.5 0 ]
[.97619 .29652 -5 -15.6875 ]
[.97619 .29652 5 0 ]
[.4875 .01472 -18.3125 -7.84375 ]
[.4875 .01472 0 7.84375 ]
[.4875 .16187 -27.3125 -7.84375 ]
[.4875 .16187 0 7.84375 ]
[.4875 .45617 -19 -7.84375 ]
[.4875 .45617 0 7.84375 ]
[.4875 .60332 -10 -7.84375 ]
[.4875 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 .30902 m
.02381 .31527 L
s
gsave
.02381 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 .30902 m
.2619 .31527 L
s
gsave
.2619 .29652 -74.6563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.7381 .30902 m
.7381 .31527 L
s
gsave
.7381 .29652 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 .30902 m
.97619 .31527 L
s
gsave
.97619 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .30902 m
.07143 .31277 L
s
.11905 .30902 m
.11905 .31277 L
s
.16667 .30902 m
.16667 .31277 L
s
.21429 .30902 m
.21429 .31277 L
s
.30952 .30902 m
.30952 .31277 L
s
.35714 .30902 m
.35714 .31277 L
s
.40476 .30902 m
.40476 .31277 L
s
.45238 .30902 m
.45238 .31277 L
s
.54762 .30902 m
.54762 .31277 L
s
.59524 .30902 m
.59524 .31277 L
s
.64286 .30902 m
.64286 .31277 L
s
.69048 .30902 m
.69048 .31277 L
s
.78571 .30902 m
.78571 .31277 L
s
.83333 .30902 m
.83333 .31277 L
s
.88095 .30902 m
.88095 .31277 L
s
.92857 .30902 m
.92857 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5 .01472 m
.50625 .01472 L
s
gsave
.4875 .01472 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .16187 m
.50625 .16187 L
s
gsave
.4875 .16187 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .45617 m
.50625 .45617 L
s
gsave
.4875 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .60332 m
.50625 .60332 L
s
gsave
.4875 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .04415 m
.50375 .04415 L
s
.5 .07358 m
.50375 .07358 L
s
.5 .10301 m
.50375 .10301 L
s
.5 .13244 m
.50375 .13244 L
s
.5 .1913 m
.50375 .1913 L
s
.5 .22073 m
.50375 .22073 L
s
.5 .25016 m
.50375 .25016 L
s
.5 .27959 m
.50375 .27959 L
s
.5 .33845 m
.50375 .33845 L
s
.5 .36788 m
.50375 .36788 L
s
.5 .39731 m
.50375 .39731 L
s
.5 .42674 m
.50375 .42674 L
s
.5 .4856 m
.50375 .4856 L
s
.5 .51503 m
.50375 .51503 L
s
.5 .54446 m
.50375 .54446 L
s
.5 .57389 m
.50375 .57389 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .60332 m
.04262 .58952 L
.06244 .5707 L
.10458 .51952 L
.14264 .46542 L
.18314 .40625 L
.22214 .35356 L
.24201 .32988 L
.26358 .3074 L
.28258 .29074 L
.30351 .2761 L
.32335 .26596 L
.33421 .26197 L
.34438 .25923 L
.34934 .25823 L
.35196 .2578 L
.35476 .2574 L
.35744 .25709 L
.35988 .25686 L
.36232 .25668 L
.36342 .25662 L
.36463 .25656 L
.36579 .25652 L
.36688 .25649 L
.36803 .25646 L
.36926 .25645 L
.37041 .25646 L
.37149 .25647 L
.37247 .25649 L
.37353 .25652 L
.37468 .25656 L
.37595 .25662 L
.37853 .25679 L
.38091 .25699 L
.38315 .25722 L
.38785 .25783 L
.39298 .25869 L
.40333 .26102 L
.41292 .26384 L
.42184 .26699 L
.46088 .28564 L
.50238 .31049 L
.54236 .33418 L
.56068 .34355 L
.58083 .35204 L
.59041 .35525 L
.60071 .35802 L
.60639 .35923 L
.61158 .36012 L
.61407 .36048 L
Mistroke
.6168 .36081 L
.61938 .36107 L
.62175 .36127 L
.62431 .36142 L
.62566 .36149 L
.62634 .36151 L
.62708 .36153 L
.62832 .36156 L
.62947 .36158 L
.63018 .36158 L
.63084 .36158 L
.63209 .36157 L
.63323 .36155 L
.63443 .36151 L
.63545 .36147 L
.63658 .36142 L
.6391 .36126 L
.64146 .36106 L
.64674 .36043 L
.65157 .35963 L
.65687 .35852 L
.66245 .35707 L
.67202 .35391 L
.68223 .34959 L
.70061 .33937 L
.72199 .32356 L
.74216 .30501 L
.78068 .26093 L
.82165 .20491 L
.8611 .14714 L
.89904 .09366 L
.92016 .06688 L
.93943 .04538 L
.97619 .01472 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00f@3oool0
0`000000oooo0?ooo`3D0?ooo`00bP3oool01P2oM000000000000000000004RL0=ooo`T0oooo00<0
00000?ooo`3oool0e03oool00</0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`3D
0?ooo`00b`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0=@0oooo003;0?ooo`03
0?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0e03oool00;/0oooo00<0oooO09a800000000
1`0000000`18W=l0oooo0?ooo`030?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0
e03oool00</0oooo00<0omnL04P0001d_ol02`3oool400000<00oooo00<000000?ooo`3oool0403o
ool00</0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`300?ooo`030000003oool0
oooo0140oooo003;0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0_`3oool00`00
0000oooo0?ooo`0B0?ooo`00b`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0;d0
oooo0P00000E0?ooo`00b`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0;`0oooo
00<000000?ooo`3oool05@3oool00<T0oooo00D0oooO09a800000000000007Boo`0;0?ooo`030000
003oool0oooo0;/0oooo00<000000?ooo`3oool05P3oool00</0oooo00<0okmd0000001d_ol02`3o
ool00`000000oooo0?ooo`2j0?ooo`030000003oool0oooo01L0oooo003I0?ooo`030000003oool0
oooo0;T0oooo00<000000?ooo`3oool0603oool00=T0oooo00<000000?ooo`3oool0^03oool00`00
0000oooo0?ooo`0I0?ooo`00f@3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo01X0
oooo003I0?ooo`030000003oool0oooo0;D0oooo0P00000M0?ooo`00f@3oool00`000000oooo0?oo
o`2d0?ooo`030000003oool0oooo01d0oooo003I0?ooo`030000003oool0oooo0;<0oooo00<00000
0?ooo`3oool07P3oool00=T0oooo0`00002b0?ooo`030000003oool0oooo01l0oooo003I0?ooo`03
0000003oool0oooo0;40oooo00<000000?ooo`3oool0803oool00=T0oooo00<000000?ooo`3oool0
/03oool00`000000oooo0?ooo`0Q0?ooo`00f@3oool00`000000oooo0?ooo`2_0?ooo`030000003o
ool0oooo0280oooo003I0?ooo`030000003oool0oooo0:h0oooo00<000000?ooo`3oool08`3oool0
0=T0oooo00<000000?ooo`3oool0[P3oool00`000000oooo0?ooo`0S0?ooo`00f@3oool00`000000
oooo0?ooo`2]0?ooo`030000003oool0oooo02@0oooo003I0?ooo`030000003oool0oooo0:`0oooo
00<000000?ooo`3oool09@3oool00=T0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?oo
o`0V0?ooo`00f@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo02L0oooo003I0?oo
o`030000003oool0oooo0:T0oooo00<000000?ooo`3oool0:03oool00=T0oooo00<000000?ooo`3o
ool0Z03oool00`000000oooo0?ooo`0Y0?ooo`00f@3oool300000:P0oooo00<000000?ooo`3oool0
:@3oool00=T0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?ooo`0Z0?ooo`00f@3oool0
0`000000oooo0?ooo`2V0?ooo`030000003oool0oooo02/0oooo003I0?ooo`030000003oool0oooo
0:D0oooo00<000000?ooo`3oool0;03oool00=T0oooo00<000000?ooo`3oool0Y03oool00`000000
oooo0?ooo`0]0?ooo`00f@3oool00`000000oooo0?ooo`2S0?ooo`030000003oool0oooo02h0oooo
003I0?ooo`030000003oool0oooo0:<0oooo00<000000?ooo`3oool0;P3oool00=T0oooo00<00000
0?ooo`3oool0XP3oool00`000000oooo0?ooo`0_0?ooo`00f@3oool00`000000oooo0?ooo`2Q0?oo
o`030000003oool0oooo0300oooo003I0?ooo`030000003oool0oooo0:00oooo00<000000?ooo`3o
ool0<@3oool00=T0oooo00<000000?ooo`3oool0X03oool00`000000oooo0?ooo`0a0?ooo`00f@3o
ool00`000000oooo0?ooo`2O0?ooo`030000003oool0oooo0380oooo003I0?ooo`<00000WP3oool0
0`000000oooo0?ooo`0c0?ooo`00f@3oool00`000000oooo0?ooo`2N0?ooo`030000003oool0oooo
03<0oooo003I0?ooo`030000003oool0oooo09d0oooo00<000000?ooo`3oool0=03oool00=T0oooo
00<000000?ooo`3oool0W03oool00`000000oooo0?ooo`0e0?ooo`00f@3oool00`000000oooo0?oo
o`2L0?ooo`030000003oool0oooo03D0oooo003I0?ooo`030000003oool0oooo09/0oooo00<00000
0?ooo`3oool0=P3oool00=T0oooo00<000000?ooo`3oool0VP3oool00`000000oooo0?ooo`0g0?oo
o`00f@3oool00`000000oooo0?ooo`2I0?ooo`030000003oool0oooo03P0oooo003I0?ooo`030000
003oool0oooo09T0oooo00<000000?ooo`3oool0>03oool00=T0oooo00<000000?ooo`3oool0V03o
ool00`000000oooo0?ooo`0i0?ooo`00f@3oool00`000000oooo0?ooo`2G0?ooo`030000003oool0
oooo03X0oooo003I0?ooo`030000003oool0oooo09L0oooo00<000000?ooo`3oool0>P3oool00=T0
oooo0`00002F0?ooo`030000003oool0oooo03/0oooo003I0?ooo`030000003oool0oooo09D0oooo
00<000000?ooo`3oool0?03oool00=T0oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?oo
o`0l0?ooo`00f@3oool00`000000oooo0?ooo`2D0?ooo`030000003oool0oooo03d0oooo003I0?oo
o`030000003oool0oooo09<0oooo00<000000?ooo`3oool0?P3oool00=T0oooo00<000000?ooo`3o
ool0T`3oool00`000000oooo0?ooo`0n0?ooo`00f@3oool00`000000oooo0?ooo`2B0?ooo`030000
003oool0oooo03l0oooo002l0?ooo`050?oOW018000000000000M02oool00`3oool01P3o_g@00000
07Boo`3oool0oooO09a800<0000000<0M;oo0?ooo`3oool02@3oool00`000000oooo0?ooo`2A0?oo
o`030000003oool0oooo0400oooo002k0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`02
0?ooo`070?noM0000000M;oo0?ooo`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00T0oooo
00<000000?ooo`3oool0T@3oool00`000000oooo0?ooo`100?ooo`00^`3oool00`2oM00004RL0=oo
o`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO09a8M02oool02@3oool00`000000oooo0?oo
o`2@0?ooo`030000003oool0oooo0440oooo002j0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia8
0000M02oool02`3oool00`2oM000M;oo0?ooo`080?ooo`030000003oool0oooo08l0oooo00<00000
0?ooo`3oool0@P3oool00:h0oooo00<0oooO09a8000000001`0000001@18W=l0oooo0?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3ooml0W4P004RLg`090?ooo`030000003oool0
oooo08h0oooo00<000000?ooo`3oool0@`3oool00;X0oooo00<0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`0:0?ooo`030=nLB0000000M;oo00T0oooo1000002=0?ooo`030000003oool0oooo
04<0oooo002j0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01`3Ogml0
oooo0?oo_`1d000000000000M02oool02@3oool00`000000oooo0?ooo`2=0?ooo`030000003oool0
oooo04@0oooo002j0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3O
W4P00000000000000000001809cOo`X0oooo00<000000?ooo`3oool0S03oool00`000000oooo0?oo
o`150?ooo`00^P3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono
07@000000000B9cO303oool00`000000oooo0?ooo`2<0?ooo`030000003oool0oooo04D0oooo002k
0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`2oM000M;oo0?ooo`0<
0?ooo`030000003oool0oooo08/0oooo00<000000?ooo`3oool0AP3oool00;/0oooo00L0oono07@0
001d_ol0oooo0?noM00007@0_ooo00P0oooo00H0omnL04P00000000000000000B02Lgol90?ooo`03
0000003oool0oooo08X0oooo00<000000?ooo`3oool0A`3oool00;`0oooo00D0oono07@000000000
001d0;ooo`0:0?ooo`050;md00000000000000000018W=l02@3oool00`000000oooo0?ooo`290?oo
o`030000003oool0oooo04P0oooo003I0?ooo`030000003oool0oooo08T0oooo00<000000?ooo`3o
ool0B03oool00=T0oooo00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`190?ooo`00f@3o
ool00`000000oooo0?ooo`270?ooo`030000003oool0oooo04X0oooo003I0?ooo`030000003oool0
oooo08L0oooo00<000000?ooo`3oool0BP3oool00=T0oooo00<000000?ooo`3oool0QP3oool00`00
0000oooo0?ooo`1;0?ooo`00f@3oool3000008D0oooo00<000000?ooo`3oool0C03oool00=T0oooo
00<000000?ooo`3oool0Q@3oool00`000000oooo0?ooo`1<0?ooo`00f@3oool00`000000oooo0?oo
o`240?ooo`030000003oool0oooo04d0oooo003I0?ooo`030000003oool0oooo08<0oooo00<00000
0?ooo`3oool0CP3oool00=T0oooo00<000000?ooo`3oool0PP3oool00`000000oooo0?ooo`1?0?oo
o`00f@3oool00`000000oooo0?ooo`220?ooo`030000003oool0oooo04l0oooo003I0?ooo`030000
003oool0oooo0840oooo00<000000?ooo`3oool0D03oool00=T0oooo00<000000?ooo`3oool0P03o
ool00`000000oooo0?ooo`1A0?ooo`00f@3oool00`000000oooo0?ooo`200?ooo`030000003oool0
oooo0540oooo003I0?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool0DP3oool00=T0
oooo00<000000?ooo`3oool0OP3oool00`000000oooo0?ooo`1C0?ooo`00f@3oool00`000000oooo
0?ooo`1n0?ooo`030000003oool0oooo05<0oooo003I0?ooo`<00000O@3oool00`000000oooo0?oo
o`1D0?ooo`00f@3oool00`000000oooo0?ooo`1l0?ooo`030000003oool0oooo05D0oooo003I0?oo
o`030000003oool0oooo07/0oooo00<000000?ooo`3oool0EP3oool00=T0oooo00<000000?ooo`3o
ool0N`3oool00`000000oooo0?ooo`1F0?ooo`00f@3oool00`000000oooo0?ooo`1j0?ooo`030000
003oool0oooo05L0oooo003I0?ooo`030000003oool0oooo07T0oooo00<000000?ooo`3oool0F03o
ool00=T0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`1H0?ooo`00f@3oool00`00
0000oooo0?ooo`1h0?ooo`030000003oool0oooo05T0oooo003I0?ooo`030000003oool0oooo07L0
oooo00<000000?ooo`3oool0FP3oool00=T0oooo00<000000?ooo`3oool0M`3oool00`000000oooo
0?ooo`1J0?ooo`00f@3oool00`000000oooo0?ooo`1f0?ooo`030000003oool0oooo05/0oooo003I
0?ooo`030000003oool0oooo07D0oooo00<000000?ooo`3oool0G03oool00=T0oooo0`00001d0?oo
o`030000003oool0oooo05d0oooo003I0?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3o
ool0G@3oool00=T0oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`1N0?ooo`006@3o
ool01P2oM000000000000000000004RL0=oooeD0oooo00D0omnL04P000000000001d0;ooo`030?oo
o`060?noM0000000M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`0D0?oooa000000;@3o
ool00`000000oooo0?ooo`1F0?ooo`050?oOW018000000000000M02oool00`3oool01P3o_g@00000
07Boo`3oool0oooO09a800<0000000<0M;oo0?ooo`3oool0203oool00`000000oooo0?ooo`1:0?oo
o`060;md00000000000000000000B9`0gooo3`3oool001X0oooo00<0omnL04P0001d_ol0EP3oool0
1`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@0000007Boo`3oool0oooO09a8
M02oool00P3oool00`2oM000M;oo0?ooo`0@0?ooo`@00000403oool3000002X0oooo00<000000?oo
o`3oool0E@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@0000007Bo
o`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`080?ooo`030000003oool0oooo04/0
oooo00<0omnL04P0001d_ol04@3oool001X0oooo00<0omnL04P0001d_ol0EP3oool00`2oM00004RL
0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO09a8M02oool0303oool4000001L0oooo
1000000V0?ooo`030000003oool0oooo05D0oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo
0?ooo`0:0?ooo`030?oog`2LB7@0_ooo00L0oooo00<000000?ooo`3oool0C03oool00`3ogi`0B000
07Boo`0A0?ooo`006P3oool00`3ogi`0B00007Boo`1E0?ooo`030?oo_`1d0000B9cO00<0oooo00<0
gia80000M02oool02`3oool00`2oM000M;oo0?ooo`090?ooo`8000007`3oool400000280oooo00<0
00000?ooo`3oool0E03oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00/0oooo00<0
_g@007Boo`3oool01@3oool00`000000oooo0?ooo`1=0?ooo`030?oOW0180000M;oo0140oooo000:
0?ooo`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool00`3oool00`3ogi`0B00007Boo`18
0?ooo`030?oog`2LB000000000L0000000H0B9cO0?ooo`3oool0omnL04P0001d_ol30?ooo`030?no
M00004P0W=oo00X0oooo00<0oooO09a80018W=l0203oool2000002D0oooo0`00000O0?ooo`030000
003oool0oooo05@0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030?oo
g`2LB000B9cO00D0oooo00<000000?ooo`3oool0CP3oool00`3ogi`0B00007Boo`0A0?ooo`006P3o
ool00`3ogi`0B00007Boo`1E0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3o
ool00`3OW4P0000007Boo`060?ooo`800000:P3oool2000001d0oooo00<000000?ooo`3oool0E03o
ool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0gia80000001d_ol0103o
ool00`000000oooo0?ooo`1?0?ooo`030?oOW0180000M;oo0140oooo000J0?ooo`030?oOW0180000
M;oo05D0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`070=oOg`3oool0
oono07@000000000001d0;ooo`040?ooo`800000;P3oool2000001/0oooo00<000000?ooo`3oool0
E03oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0
M0000000000007@0_ooo00<0oooo00<000000?ooo`3oool0D03oool00`3ogi`0B00007Boo`0A0?oo
o`006P3oool00`3ogi`0B00007Boo`1E0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol01P3oool01P3OW4P00000000000000000001809cOo`<0oooo0P00000b0?ooo`8000006@3oool0
0`000000oooo0?ooo`1D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool0
1P3OW4P00000000000000000001809cOo`@0oooo00<000000?ooo`3oool0D03oool00`3ogi`0B000
07Boo`0A0?ooo`006P3oool00`3ogi`0B00007Boo`1E0?ooo`030?oo_`1d0000B9cO00<0oooo00<0
gia80000M02oool01P3oool0103ookl0M00000000018W=l30?ooo`800000=P3oool2000001L0oooo
00<000000?ooo`3oool0E03oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo
00@0oono07@000000000B9cO1@3oool00`000000oooo0?ooo`1A0?ooo`030?oOW0180000M;oo0140
oooo000J0?ooo`030?oOW0180000M;oo05H0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo
0?ooo`070?ooo`040;md001d_ol0oooo0?ooo`800000>P3oool2000001D0oooo0`00001E0?ooo`03
0=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`2oM000M;oo0?ooo`040?ooo`03
0000003oool0oooo0580oooo00<0omnL04P0001d_ol04@3oool001P0oooo00D0oooO09a800000000
000007Boo`1F0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`060?oOW0180000
00000000000004P0W=oo?03oool2000001<0oooo00<000000?ooo`3oool0E@3oool01`3ookl0M000
07Boo`3oool0okmd0000M02oool0203oool01`3ogi`0B000000000000000001809cOo`000000D`3o
ool01@3ooml0W4P0000000000000M;oo0140oooo000J0?ooo`030?noM0000000M;oo05L0oooo00D0
oono07@000000000001d0;ooo`0:0?ooo`@0000000<0B9cO0?ooo`3oool0?03oool200000140oooo
00<000000?ooo`3oool0EP3oool01@3ookl0M0000000000007@0_ooo00X0oooo00<0_g@000000000
00000P00001F0?ooo`030?noM0000000M;oo0140oooo00220?ooo`030000003oool0oooo04<0oooo
0P00000?0?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool0J@3oool00800oooo0P00
00180?ooo`8000003@3oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo06X0oooo001o
0?ooo`030000003oool0oooo04X0oooo0P00000;0?ooo`030000003oool0oooo06H0oooo00<00000
0?ooo`3oool0J`3oool007h0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`080?oo
o`030000003oool0oooo06D0oooo00<000000?ooo`3oool0K03oool007d0oooo00<000000?ooo`3o
ool0C`3oool2000000P0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`1/0?ooo`00
O03oool00`000000oooo0?ooo`1B0?ooo`8000001P3oool00`000000oooo0?ooo`1T0?ooo`030000
003oool0oooo06d0oooo001k0?ooo`030000003oool0oooo05D0oooo0P0000040?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`3oool0KP3oool007X0oooo00<000000?ooo`3oool0F03oool0
1@000000oooo0?ooo`3oool0000006@0oooo00<000000?ooo`3oool0K`3oool007T0oooo00<00000
0?ooo`3oool0FP3oool2000000030?ooo`000000oooo0680oooo00<000000?ooo`3oool0L03oool0
07P0oooo00<000000?ooo`3oool0G@3oool200000680oooo00<000000?ooo`3oool0L@3oool000/0
ooooo`00002N000000P0oooo000E0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0
403oool2000001<0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0A0?ooo`030000
003oool0oooo0100oooo00<000000?ooo`3oool04@3oool010000000oooo00000000000@0?ooo`03
0000003oool0oooo0100oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0oooo0100oooo0P00000C0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3o
ool0403oool001D0oooo00<000000?ooo`3oool0G@3oool00`000000oooo0000001Q0?ooo`050000
003oool0oooo0?ooo`000000FP3oool2000000030?ooo`000000oooo0600oooo00<000000?ooo`3o
ool0403oool007@0oooo00<000000?ooo`3oool0HP3oool00`000000oooo0?ooo`020?ooo`800000
E`3oool00`000000oooo0?ooo`1f0?ooo`00L`3oool00`000000oooo0?ooo`1S0?ooo`030000003o
ool0oooo00@0oooo0P00001D0?ooo`030000003oool0oooo07L0oooo001b0?ooo`030000003oool0
oooo06@0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0
oooo07P0oooo001a0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3oool01`3oool20000
04l0oooo00<000000?ooo`3oool0N@3oool00700oooo00<000000?ooo`3oool0IP3oool00`000000
oooo0?ooo`090?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0NP3oool006l0oooo
00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`0:0?ooo`800000B@3oool2000007d0oooo
001^0?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool0303oool2000004H0oooo00<0
00000?ooo`3oool0O@3oool006d0oooo00<000000?ooo`3oool0J@3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0OP3oool006d0oooo00<000000?oo
o`3oool0J@3oool00`000000oooo0?ooo`0?0?ooo`800000@03oool200000840oooo001/0?ooo`03
0000003oool0oooo06X0oooo0`00000A0?ooo`800000?@3oool00`000000oooo0?ooo`210?ooo`00
J`3oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo01<0oooo0P00000i0?ooo`800000
Q03oool006X0oooo00<000000?ooo`3oool0K03oool00`000000oooo0?ooo`0E0?ooo`<00000=03o
ool2000008H0oooo001Y0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0603oool2
00000300oooo0P0000280?ooo`00J@3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo
01X0oooo0`00000[0?ooo`800000RP3oool006P0oooo00<000000?ooo`3oool0KP3oool00`000000
oooo0?ooo`0M0?ooo`8000009`3oool2000008`0oooo001W0?ooo`030000003oool0oooo06l0oooo
00<000000?ooo`3oool07`3oool200000280oooo0`00002>0?ooo`00IP3oool00`000000oooo0?oo
o`1`0?ooo`030000003oool0oooo0240oooo0`00000L0?ooo`<00000T@3oool006H0oooo00<00000
0?ooo`3oool0L03oool00`000000oooo0?ooo`0T0?ooo`H000004@3oool5000009@0oooo001U0?oo
o`030000003oool0oooo0740oooo00<000000?ooo`3oool0:P3ooolA000009T0oooo001T0?ooo`03
0000003oool0oooo0780oooo00<000000?ooo`3oool0e03oool006<0oooo00<000000?ooo`3oool0
L`3oool00`000000oooo0?ooo`3D0?ooo`00H`3oool00`000000oooo0?ooo`1c0?ooo`<00000e03o
ool00680oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`3D0?ooo`00H@3oool00`00
0000oooo0?ooo`1e0?ooo`030000003oool0oooo0=@0oooo001P0?ooo`030000003oool0oooo07H0
oooo00<000000?ooo`3oool0e03oool00600oooo00<000000?ooo`3oool0MP3oool00`000000oooo
0?ooo`3D0?ooo`00G`3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo0=@0oooo001N
0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0e03oool005h0oooo00<000000?oo
o`3oool0N03oool00`000000oooo0?ooo`3D0?ooo`00G@3oool00`000000oooo0?ooo`1i0?ooo`03
0000003oool0oooo0=@0oooo001L0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0
e03oool005/0oooo00<000000?ooo`3oool0N`3oool00`000000oooo0?ooo`3D0?ooo`00F`3oool0
0`000000oooo0?ooo`1k0?ooo`030000003oool0oooo0=@0oooo001J0?ooo`030000003oool0oooo
07`0oooo0`00003D0?ooo`00F@3oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo0=@0
oooo001H0?ooo`030000003oool0oooo07h0oooo00<000000?ooo`3oool0e03oool005P0oooo00<0
00000?ooo`3oool0OP3oool00`000000oooo0?ooo`3D0?ooo`00E`3oool00`000000oooo0?ooo`1o
0?ooo`030000003oool0oooo0=@0oooo001F0?ooo`030000003oool0oooo0800oooo00<000000?oo
o`3oool0e03oool005H0oooo00<000000?ooo`3oool0P03oool00`000000oooo0?ooo`3D0?ooo`00
E@3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo0=@0oooo001D0?ooo`030000003o
ool0oooo0880oooo00<000000?ooo`3oool0e03oool005<0oooo00<000000?ooo`3oool0P`3oool0
0`000000oooo0?ooo`3D0?ooo`00D`3oool00`000000oooo0?ooo`230?ooo`030000003oool0oooo
0=@0oooo001B0?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool0e03oool00540oooo
00<000000?ooo`3oool0Q@3oool300000=@0oooo001A0?ooo`030000003oool0oooo08D0oooo00<0
00000?ooo`3oool0e03oool00500oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`3D
0?ooo`00C`3oool00`000000oooo0?ooo`270?ooo`030000003oool0oooo0=@0oooo001>0?ooo`03
0000003oool0oooo08P0oooo00<000000?ooo`3oool0e03oool004h0oooo00<000000?ooo`3oool0
R03oool00`000000oooo0?ooo`3D0?ooo`00C@3oool00`000000oooo0?ooo`290?ooo`030000003o
ool0oooo0=@0oooo001<0?ooo`030000003oool0oooo06d0oooo00D0omnL04P000000000001d0;oo
o`030?ooo`060?noM0000000M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`090?ooo`03
0000003oool0oooo0=@0oooo001<0?ooo`030000003oool0oooo06`0oooo00L0oono07@0001d_ol0
oooo0?noM00007@0_ooo0080oooo00L0okmd0000001d_ol0oooo0?oog`2LB7@0_ooo0080oooo00<0
_g@007Boo`3oool02@3oool00`000000oooo0?ooo`3D0?ooo`00B`3oool00`000000oooo0?ooo`1]
0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd0;ooo`09
0?ooo`030000003oool0oooo0=@0oooo001:0?ooo`030000003oool0oooo06d0oooo00<0oono07@0
0018W=l00`3oool00`3OW4P0001d0;ooo`0;0?ooo`030;md001d_ol0oooo00P0oooo00<000000?oo
o`3oool0e03oool004X0oooo00<000000?ooo`3oool0K@3oool00`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo00X0oooo00<0oooO09a80018W=l02@3oool00`000000oooo0?ooo`3D0?ooo`00
B@3oool00`000000oooo0?ooo`1^0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
2P3oool00`3OW4P0000007Boo`090?ooo`@00000d`3oool004P0oooo00<000000?ooo`3oool0K`3o
ool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M000
0000000007@0_ooo00T0oooo00<000000?ooo`3oool0e03oool004L0oooo00<000000?ooo`3oool0
L03oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00H0gia8000000000000
00000000B02Lgol:0?ooo`030000003oool0oooo0=@0oooo00170?ooo`030000003oool0oooo0700
oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d0000000004RL
g``0oooo00<000000?ooo`3oool0e03oool004H0oooo00<000000?ooo`3oool0LP3oool00`3OW4P0
04RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0_g@007Boo`3oool0303oool00`000000
oooo0?ooo`3D0?ooo`00A@3oool00`000000oooo0?ooo`1c0?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`080?ooo`060?oOW018000000000000000004P0W=oo2@3oool00`000000oooo0?oo
o`3D0?ooo`00A@3oool00`000000oooo0?ooo`1d0?ooo`050?oo_`1d000000000000M02oool02P3o
ool01@2oM0000000000000000000B9cO00T0oooo00<000000?ooo`3oool0e03oool004@0oooo00<0
00000?ooo`3oool0TP3oool00`000000oooo0?ooo`3D0?ooo`00@`3oool00`000000oooo0?ooo`2C
0?ooo`030000003oool0oooo0=@0oooo00120?ooo`030000003oool0oooo09@0oooo00<000000?oo
o`3oool0e03oool00480oooo00<000000?ooo`3oool0U03oool00`000000oooo0?ooo`3D0?ooo`00
@@3oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo0=@0oooo00100?ooo`030000003o
ool0oooo09H0oooo0`00003D0?ooo`00?`3oool00`000000oooo0?ooo`2G0?ooo`030000003oool0
oooo0=@0oooo000o0?ooo`030000003oool0oooo09L0oooo00<000000?ooo`3oool0e03oool003h0
oooo00<000000?ooo`3oool0V03oool00`000000oooo0?ooo`3D0?ooo`00?@3oool00`000000oooo
0?ooo`2I0?ooo`030000003oool0oooo0=@0oooo000m0?ooo`030000003oool0oooo09T0oooo00<0
00000?ooo`3oool0e03oool003`0oooo00<000000?ooo`3oool0VP3oool00`000000oooo0?ooo`3D
0?ooo`00>`3oool00`000000oooo0?ooo`2K0?ooo`030000003oool0oooo0=@0oooo000j0?ooo`03
0000003oool0oooo09`0oooo00<000000?ooo`3oool0e03oool003X0oooo00<000000?ooo`3oool0
W03oool00`000000oooo0?ooo`3D0?ooo`00>@3oool00`000000oooo0?ooo`2M0?ooo`030000003o
ool0oooo0=@0oooo000h0?ooo`030000003oool0oooo09h0oooo00<000000?ooo`3oool0e03oool0
03L0oooo00<000000?ooo`3oool0W`3oool300000=@0oooo000g0?ooo`030000003oool0oooo09l0
oooo00<000000?ooo`3oool0e03oool003H0oooo00<000000?ooo`3oool0X03oool00`000000oooo
0?ooo`3D0?ooo`00=@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0=@0oooo000d
0?ooo`030000003oool0oooo0:80oooo00<000000?ooo`3oool0e03oool003@0oooo00<000000?oo
o`3oool0XP3oool00`000000oooo0?ooo`3D0?ooo`00<`3oool00`000000oooo0?ooo`2S0?ooo`03
0000003oool0oooo0=@0oooo000b0?ooo`030000003oool0oooo0:@0oooo00<000000?ooo`3oool0
e03oool00340oooo00<000000?ooo`3oool0Y@3oool00`000000oooo0?ooo`3D0?ooo`00<03oool0
0`000000oooo0?ooo`2V0?ooo`030000003oool0oooo0=@0oooo000`0?ooo`030000003oool0oooo
0:H0oooo00<000000?ooo`3oool0e03oool002l0oooo00<000000?ooo`3oool0Y`3oool00`000000
oooo0?ooo`3D0?ooo`00;P3oool00`000000oooo0?ooo`2X0?ooo`<00000e03oool002d0oooo00<0
00000?ooo`3oool0Z@3oool00`000000oooo0?ooo`3D0?ooo`00;03oool00`000000oooo0?ooo`2Z
0?ooo`030000003oool0oooo0=@0oooo000[0?ooo`030000003oool0oooo0:/0oooo00<000000?oo
o`3oool0e03oool002/0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`3D0?ooo`00
:P3oool00`000000oooo0?ooo`2/0?ooo`030000003oool0oooo0=@0oooo000Y0?ooo`030000003o
ool0oooo0:d0oooo00<000000?ooo`3oool0e03oool002P0oooo00<000000?ooo`3oool0[P3oool0
0`000000oooo0?ooo`3D0?ooo`009`3oool00`000000oooo0?ooo`2_0?ooo`030000003oool0oooo
0=@0oooo000W0?ooo`030000003oool0oooo0:l0oooo00<000000?ooo`3oool0e03oool002H0oooo
00<000000?ooo`3oool0/03oool00`000000oooo0?ooo`3D0?ooo`009@3oool00`000000oooo0?oo
o`2a0?ooo`030000003oool0oooo0=@0oooo000T0?ooo`030000003oool0oooo0;80oooo00<00000
0?ooo`3oool0e03oool002<0oooo00<000000?ooo`3oool0/`3oool300000=@0oooo000R0?ooo`03
0000003oool0oooo0;@0oooo00<000000?ooo`3oool0e03oool00240oooo00<000000?ooo`3oool0
]@3oool00`000000oooo0?ooo`3D0?ooo`00803oool00`000000oooo0?ooo`2f0?ooo`030000003o
ool0oooo0=@0oooo000O0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool0e03oool0
01h0oooo00<000000?ooo`3oool0^03oool00`000000oooo0?ooo`3D0?ooo`007@3oool00`000000
oooo0?ooo`2i0?ooo`030000003oool0oooo0=@0oooo000L0?ooo`030000003oool0oooo0:/0oooo
00H0_g@000000000000000000018W03Oool90?ooo`030000003oool0oooo0=@0oooo000J0?ooo`80
0000[`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0=@0oooo000I0?ooo`030000
003oool0oooo0:l0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`3D0?ooo`00603o
ool00`000000oooo0?ooo`2`0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0e03o
ool001H0oooo0P00002c0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0e03oool0
01D0oooo00<000000?ooo`3oool0/`3oool00`3ogi`0B00007Boo`0;0?ooo`@00000d`3oool00</0
oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`3D0?ooo`00b`3oool00`3ogi`0B000
07Boo`0;0?ooo`030000003oool0oooo0=@0oooo003;0?ooo`030?oOW0180000M;oo00/0oooo00<0
00000?ooo`3oool0e03oool00</0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`3D
0?ooo`00b@3oool01@3ooml0W4P0000000000000M;oo00/0oooo00<000000?ooo`3oool0e03oool0
0</0oooo00<0okmd0000001d_ol02`3oool00`000000oooo0?ooo`3D0?ooo`00o`3ooona0?ooo`00
o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.10918, -1.09709, \
0.00765073, 0.0123791}}]
}, Closed]],

Cell[TextData[{
  "Now suppose that we create a function by assigning a name to a numerical \
solution of some differential equation.  [We discuss the details of ",
  StyleBox["NDSolve", "InlineInput",
    FontWeight->"Bold"],
  " in ",
  ButtonBox["diffeq",
    ButtonData:>{"Primary topics/diffeq.nb", None},
    ButtonStyle->"Hyperlink"],
  ".]"
}], "Text"],

Cell[BoxData[
    \(g[x_]\  := \ 
      y[x] /. NDSolve[{\(y\^\[DoublePrime]\)[x] \[Equal] x\ Sin[x], \ 
            y[0] \[Equal] 0, \ \(y\^\[Prime]\)[0] \[Equal] 1}, 
          y[x], {x, 0, 10}]\)], "Input"],

Cell[TextData[{
  "If we try to plot this function directly, we get several errors messages \
which arise because the function is not fully evaluated before we try to \
plot.  Similar messages are often encountered when the function argument \
supplied to ",
  StyleBox["Plot", "InlineInput",
    FontWeight->"Bold"],
  " is actually a command rather than an explicit function."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[g[x], \ {x, 0, 5}];\)\)], "Input"],

Cell[BoxData[
    RowBox[{\(NDSolve::"dsvar"\), \(\(:\)\(\ \)\), \
"\<\"\\!\\(2.0833333333333333`*^-7\\) cannot be used as a variable. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"NDSolve::dsvar\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(ReplaceAll::"reps"\), \(\(:\)\(\ \)\), \
"\<\"\\!\\({\\(NDSolve[\\(\\(\\(\\({\\(\\(\\(\\(\\(\\(\\*SuperscriptBox[\\\"y\
\\\", \\\"\[Prime]\[Prime]\\\", Rule[MultilineFunction, \
None]]\\)\\)[2.0833333333333333`*^-7]\\)\\) \[Equal] \
4.340277777777746`*^-14\\)\\), \\(\\(\\(\\(y[0]\\)\\) \[Equal] 0\\)\\), \
\\(\\(\\(\\(\\(\\(\\*SuperscriptBox[\\\"y\\\", \\\"\[Prime]\\\", \
Rule[MultilineFunction, None]]\\)\\)[0]\\)\\) \[Equal] 1\\)\\)}\\)\\), \\(\\(\
\[LeftSkeleton] 1 \[RightSkeleton]\\)\\), \\(\\({2.0833333333333333`*^-7, 0, \
10}\\)\\)\\)\\)]\\)}\\) is neither a list of replacement rules nor a valid \
dispatch table, and so cannot be used for replacing. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"ReplaceAll::reps\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(NDSolve::"dsvar"\), \(\(:\)\(\ \)\), \
"\<\"\\!\\(2.0833333333333333`*^-7\\) cannot be used as a variable. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"NDSolve::dsvar\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(ReplaceAll::"reps"\), \(\(:\)\(\ \)\), \
"\<\"\\!\\({\\(NDSolve[\\(\\(\\(\\({\\(\\(\\(\\(\\(\\(\\*SuperscriptBox[\\\"y\
\\\", \\\"\[Prime]\[Prime]\\\", Rule[MultilineFunction, \
None]]\\)\\)[2.0833333333333333`*^-7]\\)\\) \[Equal] \
4.340277777777746`*^-14\\)\\), \\(\\(\\(\\(y[0.`]\\)\\) \[Equal] 0.`\\)\\), \
\\(\\(\\(\\(\\(\\(\\*SuperscriptBox[\\\"y\\\", \\\"\[Prime]\\\", \
Rule[MultilineFunction, None]]\\)\\)[0.`]\\)\\) \[Equal] 1.`\\)\\)}\\)\\), \
\\(\\(y[\\(\\(\[LeftSkeleton] 23 \[RightSkeleton]\\)\\)]\\)\\), \\(\\({\\(\\(\
\[LeftSkeleton] 23 \[RightSkeleton]\\)\\), 0.`, 10.`}\\)\\)\\)\\)]\\)}\\) is \
neither a list of replacement rules nor a valid dispatch table, and so cannot \
be used for replacing. \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"ReplaceAll::reps\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(NDSolve::"dsvar"\), \(\(:\)\(\ \)\), \
"\<\"\\!\\(2.0833333333333333`*^-7\\) cannot be used as a variable. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"NDSolve::dsvar\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(General::"stop"\), \(\(:\)\(\ \)\), "\<\"Further output of \
\\!\\(NDSolve :: \\\"dsvar\\\"\\) will be suppressed during this calculation. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"General::stop\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(ReplaceAll::"reps"\), \(\(:\)\(\ \)\), \
"\<\"\\!\\({\\(NDSolve[\\(\\(\\(\\({\\(\\(\\(\\(\\(\\(\\*SuperscriptBox[\\\"y\
\\\", \\\"\[Prime]\[Prime]\\\", Rule[MultilineFunction, \
None]]\\)\\)[2.0833333333333333`*^-7]\\)\\) \[Equal] \
4.340277777777746`*^-14\\)\\), \\(\\(\\(\\(y[0]\\)\\) \[Equal] 0\\)\\), \
\\(\\(\\(\\(\\(\\(\\*SuperscriptBox[\\\"y\\\", \\\"\[Prime]\\\", \
Rule[MultilineFunction, None]]\\)\\)[0]\\)\\) \[Equal] 1\\)\\)}\\)\\), \\(\\(\
\[LeftSkeleton] 1 \[RightSkeleton]\\)\\), \\(\\({2.0833333333333333`*^-7, 0, \
10}\\)\\)\\)\\)]\\)}\\) is neither a list of replacement rules nor a valid \
dispatch table, and so cannot be used for replacing. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"ReplaceAll::reps\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(General::"stop"\), \(\(:\)\(\ \)\), "\<\"Further output of \
\\!\\(ReplaceAll :: \\\"reps\\\"\\) will be suppressed during this \
calculation. \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"General::stop\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(Plot::"plnr"\), \(\(:\)\(\ \)\), "\<\"\\!\\(g[x]\\) is not a \
machine-size real number at \\!\\(x\\) = \\!\\(2.0833333333333333`*^-7\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Plot::plnr\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(Plot::"plnr"\), \(\(:\)\(\ \)\), "\<\"\\!\\(g[x]\\) is not a \
machine-size real number at \\!\\(x\\) = \\!\\(0.20283495786457897`\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Plot::plnr\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(Plot::"plnr"\), \(\(:\)\(\ \)\), "\<\"\\!\\(g[x]\\) is not a \
machine-size real number at \\!\\(x\\) = \\!\\(0.4240439992968685`\\). \\!\\(\
\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\\\", \
ButtonFrame->None, ButtonData:>\\\"Plot::plnr\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(General::"stop"\), \(\(:\)\(\ \)\), "\<\"Further output of \
\\!\\(Plot :: \\\"plnr\\\"\\) will be suppressed during this calculation. \\!\
\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"General::stop\\\"]\\)\"\>"}]], "Message"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0147151 0.588604 [
[.21429 .00222 -9.5 -15.6875 ]
[.21429 .00222 9.5 0 ]
[.40476 .00222 -9.5 -15.6875 ]
[.40476 .00222 9.5 0 ]
[.59524 .00222 -9.5 -15.6875 ]
[.59524 .00222 9.5 0 ]
[.78571 .00222 -9.5 -15.6875 ]
[.78571 .00222 9.5 0 ]
[.97619 .00222 -5 -15.6875 ]
[.97619 .00222 5 0 ]
[.01131 .13244 -19 -7.84375 ]
[.01131 .13244 0 7.84375 ]
[.01131 .25016 -19 -7.84375 ]
[.01131 .25016 0 7.84375 ]
[.01131 .36788 -19 -7.84375 ]
[.01131 .36788 0 7.84375 ]
[.01131 .4856 -19 -7.84375 ]
[.01131 .4856 0 7.84375 ]
[.01131 .60332 -10 -7.84375 ]
[.01131 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .01472 m
.21429 .02097 L
s
gsave
.21429 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.40476 .01472 m
.40476 .02097 L
s
gsave
.40476 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.59524 .01472 m
.59524 .02097 L
s
gsave
.59524 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.78571 .01472 m
.78571 .02097 L
s
gsave
.78571 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.97619 .01472 m
.97619 .02097 L
s
gsave
.97619 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .01472 m
.07143 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.16667 .01472 m
.16667 .01847 L
s
.2619 .01472 m
.2619 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.35714 .01472 m
.35714 .01847 L
s
.45238 .01472 m
.45238 .01847 L
s
.5 .01472 m
.5 .01847 L
s
.54762 .01472 m
.54762 .01847 L
s
.64286 .01472 m
.64286 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.7381 .01472 m
.7381 .01847 L
s
.83333 .01472 m
.83333 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.92857 .01472 m
.92857 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .13244 m
.03006 .13244 L
s
gsave
.01131 .13244 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.02381 .25016 m
.03006 .25016 L
s
gsave
.01131 .25016 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.02381 .36788 m
.03006 .36788 L
s
gsave
.01131 .36788 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.02381 .4856 m
.03006 .4856 L
s
gsave
.01131 .4856 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.02381 .60332 m
.03006 .60332 L
s
gsave
.01131 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04415 m
.02756 .04415 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .10301 m
.02756 .10301 L
s
.02381 .16187 m
.02756 .16187 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .30902 m
.02756 .30902 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .45617 m
.02756 .45617 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00K03oool01@3ogi`0B0000000000007@0_ooo00<0oooo00D0okmd0000001d_ol0oooo0;md
00060000000304RLg`3oool0oooo03<0oooo00D0omnL04P000000000001d0;ooo`030?ooo`030?no
M0000000M;oo00D0oooo00@0oooO09a80000B9`0gooo=@3oool01@3ogi`0B0000000000007@0_ooo
00<0oooo00<0okmd0000001d_ol00`3oool01@3ogi`0B000000000000000M;oo03H0oooo00D0omnL
04P000000000001d0;ooo`030?ooo`030?noM0000000M;oo00<0oooo00D0gia80000000000000000
04RLg`0m0?ooo`060;md00000000000000000000B9`0gooo4`3oool006/0oooo00L0oono07@0001d
_ol0oooo0?noM00007@0_ooo0080oooo00<0okmd0000001d_ol00P3oool00`3OW4P0000000000003
000000030018W03Oool0oooo0380oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo
00<0okmd0000001d_ol01@3oool0103ooml0W4P00018W03Ooold0?ooo`070?oo_`1d0000M;oo0?oo
o`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo0080oooo00L0omnL04P00018W=l0oooo0?oo
_`1d0000M;oo03@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00<0okmd0000
001d_ol00P3oool01`3OW4P0000007Boo`3oool0oooO09a80018W=l0?@3oool00`3ogi`0B00007Bo
o`0E0?ooo`00J`3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0gia8
04RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`0b0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0
001d_`3oool02@3oool0103ooml0W4P00018W03Ooold0?ooo`030;md0000B9`0gooo0080oooo00<0
_g@0001d_`3oool01@3oool00`3ooml0W4P0001d_`030?ooo`030?noM000B9`0gooo03<0oooo00<0
_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?oog`2LB00007Bo00<0oooo00<0
okmd0018W03Oool0?03oool00`3ogi`0B00007Boo`0E0?ooo`00JP3oool00`3ookl0M00004RLg`03
0?ooo`030=nLB00007@0_ooo00P0oooo00<0gia807Boo`3oool0=@3oool00`3ookl0M00004RLg`03
0?ooo`030=nLB00007@0_ooo00D0oooo00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`0`
0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01@3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo0380oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`06
0?ooo`030;md0000M;l0oooo0080oooo00<0_g@00018W03Oool0?03oool00`3ogi`0B00007Boo`0E
0?ooo`00JP3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00<0_gA809cO
o`3oool0=03oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00D0oooo00P0omnL07BL
g`3oool0oooo0?oog`2LB00004RL0=oooc<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`050?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0<P3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0omnL04P0M02oool0oooo0;md00000000M;oo
03d0oooo00<0omnL04P0001d_ol05@3oool006X0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`090?ooo`030?oog`2LB7@0_ooo03@0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`060?ooo`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`0c0?ooo`030?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol01@3oool00`3ogi`0B00007Boo`030?ooo`030;md000004P0
W=oo0380oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`050?oo_`1dB4P0
B00000000018W=l0?P3oool00`3ogi`0B00007Boo`0E0?ooo`00JP3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo00X0oooo00<0okmd001d_`3oool0<`3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo00H0oooo00L0oooO09a8M02oool0oooO09a80000B9`0gooo03<0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`050?ooo`080?oo_`1d0000000004RLg`3o
ool0okmd00000000M;lc0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool0
103ogi`0B00000000000M;lo0?ooo`030?oOW0180000M;oo01D0oooo001Z0?ooo`030?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3ookl0M01809cOo`0c0?ooo`030?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol01`3oool01P3ogi`0B7Bo0?oog`2LB00004RL0=oooc<0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`030;md0000B9`0_ia800800000
00<004RL0=ooo`3oool0<P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo
00H0omnL04P0000007@0_ooo0?noM000M;ln0?ooo`030?oOW0180000M;oo01D0oooo001Z0?ooo`03
0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02P3oool00`3ookl0M00004RLg`0c0?ooo`03
0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool01@3OW4P0M;oO09a80000B9`0gooo
03<0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`030?oOW0180000M;oo
03L0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`030;md000007@0_ooo
0080oooo00<0_g@004RLg`3oool0?03oool00`3ogi`0B00007Boo`0E0?ooo`00J`3oool00`3OW4P0
04RL0=ooo`020?ooo`030;md0000M;l0oooo00D0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0
B00004RLg`0d0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0203oool01@3ooml0
W4Q804Q80000B9`0gooo03@0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?oo
o`030=nLB00004P0W=oo03L0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`060?oo
o`030;md0000B9`0gooo0080oooo00<0gia8001d_`3oool0?03oool00`3ogi`0B00007Boo`0E0?oo
o`00J`3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool01P3oool0203ooml0W4Q809cOo`3o
ool0oono07@0000007@0_ooo=03oool01`3ookl0M00007Boo`3oool0okmd0000M02oool02P3oool0
103ogi`0B0000018W03Ooold0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`03
0?noM000000007Bo03H0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00L0oooo00L0okmd
0000001d_ol0oooo0?oo_`1d0000M;oo03/0oooo00D0oooO09a800000000000007Boo`0E0?ooo`00
K03oool01@3ookl0M0000000000007@0_ooo00P0oooo00H0oooO09a80000000000000000M02ooolf
0?ooo`050?oo_`1d000000000000M02oool0303oool00`3OW4P004RL0=ooo`0e0?ooo`050?oo_`1d
000000000000M02oool02`3oool01@3ooml0W4P00000000007@0_ooo03@0oooo00D0oono07@00000
0000001d0;ooo`090?ooo`050?noM0000000000000000018W=l0?P3oool00`3o_g@0000007Boo`0E
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00;03oool0
0`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0
oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo
000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon2
0?ooo`008`3ooooo000008800000303oool002`0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0@0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0@0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0@0?ooo`030000003oool0oooo01<0oooo000/0?ooo`030000003oool0oooo04H0oooo
00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo04H0oooo
00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0C0?ooo`00;03oool00`000000oooo0?oo
o`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?oo
o`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000
003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool0
0`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0
oooo0`00003o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<0
00000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?oo
o`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00
;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3o
ool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oo
oh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0
o`3ooon20?ooo`00;03oool300000?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon2
0?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0
ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?oo
o`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?oo
o`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000
003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool3
00000?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000
oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<0
00000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000@0?oo
o`050?oOW018000000000000M02oool00`3oool01@3o_g@0000007Boo`3oool0_g@000H0000000<0
B9cO0?ooo`3oool01P3oool00`000000oooo0?ooo`3o0?oooh80oooo000?0?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo0080oooo00<0gia8000000000000
0`0000000`00B9`0gooo0?ooo`060?ooo`030000003oool0oooo0?l0ooooPP3oool000l0oooo00<0
_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030=nLB018W=l0oooo0080oooo00<0
giad09cOo`3oool01P3oool00`000000oooo0?ooo`3o0?oooh80oooo000>0?ooo`030?oo_`1d0000
B9cO00<0oooo00<0gia80000M02oool0203oool00`3OW4P0M;oo0?ooo`0:0?ooo`030000003oool0
oooo0?l0ooooPP3oool000h0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`090?oo
o`030;mdB02Lgol0oooo00T0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003P3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00<0oooO09a8M02oool02@3oool400000?l0
ooooP@3oool000h0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030?no
M000M;l0oooo00P0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003P3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00X0oooo00<0oono07@0B02Lgol0203oool00`000000oooo0?oo
o`3o0?oooh80oooo000>0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02P3oool0
0`3ookl0M00004RLg`080?ooo`030000003oool0oooo0?l0ooooPP3oool000l0oooo00<0gia80018
W03Oool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P0
0018W=l0203oool00`000000oooo0?ooo`3o0?oooh80oooo000?0?ooo`070?oo_`1d0000M;oo0?oo
o`3o_g@0001d0;ooo`060?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool80?ooo`03
0000003oool0oooo0?l0ooooPP3oool00100oooo00D0oono07@000000000001d0;ooo`080?ooo`06
0?oog`2LB00000000000000007@0_ooo2@3oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?oo
o`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00
;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3o
ool002`0oooo0`00003o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0
oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo
000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon2
0?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0
ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?oo
o`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?oo
o`3oool0o`3ooon20?ooo`00;03oool300000?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0
o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0
oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000
oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<0
00000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?oo
o`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00
;03oool300000?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool0
0`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0
oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo
000@0?ooo`050?oOW018000000000000M02oool00`3oool00`3o_g@0000007Boo`050?ooo`040?oo
g`2LB00004RL0=ooo`P0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003`3oool01`3ookl0M000
07Boo`3oool0okmd0000M02oool00P3oool00`3o_g@0000007Boo`050?ooo`040?oog`2LB00004RL
0=ooo`P0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003`3oool00`2oM00004RL0=ooo`020?oo
o`030;md0000M;l0oooo00T0oooo00@0oooO09a80000B9`0gooo203oool00`000000oooo0?ooo`3o
0?oooh80oooo000>0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01@3oool00`2o
M0000000000000050000000307Boo`3oool0oooo00D0oooo00<000000?ooo`3oool0o`3ooon20?oo
o`003P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00D0oooo00P0omnL07BLg`3o
ool0oooo0?oog`2LB00004RL0=ooo`P0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003P3oool0
0`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0gia807Boo`3oool0oooO09a8
0000B9`0gooo00P0oooo1000003o0?oooh40oooo000>0?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol01P3oool01`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool0203oool00`000000
oooo0?ooo`3o0?oooh80oooo000>0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
1`3oool01P3ogi`0B7Bo0?oog`2LB00004RL0=ooo`P0oooo00<000000?ooo`3oool0o`3ooon20?oo
o`003P3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00P0oooo00D0gia807Bog`2L
B00004RL0=ooo`080?ooo`030000003oool0oooo0?l0ooooPP3oool000l0oooo00<0gia80018W03O
ool00P3oool00`2oM00007Bo0?ooo`080?ooo`050?oog`2LB4P0B4P00018W03Oool0203oool00`00
0000oooo0?ooo`3o0?oooh80oooo000?0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0:
0?ooo`040?oOW018000004RL0=ooo`P0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00403oool0
1@3ookl0M0000000000007@0_ooo00`0oooo00<0gia80018W03Oool0203oool00`000000oooo0?oo
o`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?oo
o`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000
003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool3
00000?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000
oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<0
00000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?oo
o`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00
;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3o
ool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool300000?l0ooooPP3oool002`0
oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo
000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon2
0?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0
ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?oo
o`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?oo
o`3oool0o`3ooon20?ooo`00;03oool300000?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0
o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0
oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000
oooo0?ooo`3o0?oooh80oooo000@0?ooo`050?oOW018000000000000M02oool00`3oool00`3o_g@0
000007Boo`030?ooo`050?oOW018000000000000001d_ol02@3oool00`000000oooo0?ooo`3o0?oo
oh80oooo000?0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000
M;oo0080oooo00L0omnL04P00018W=l0oooo0?oo_`1d0000M;oo00P0oooo00<000000?ooo`3oool0
o`3ooon20?ooo`003`3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00D0oooo00<0
oooO09a80000M;l00`3oool00`3o_g@004RL0=ooo`070?ooo`030000003oool0oooo0?l0ooooPP3o
ool000h0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`050?ooo`030?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`3o0?oooh80oooo000>0?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01@3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003P3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00D0oooo00<0omnL04P0001d_ol00`3oool00`2oM000
001809cOo`070?ooo`@00000o`3ooon10?ooo`003P3oool00`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00D0oooo00P0oono07@000000000B9cO0?ooo`3o_g@00000001d_`P0oooo00<00000
0?ooo`3oool0o`3ooon20?ooo`003P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00H0oooo00<0_g@00018W02oW4P00P0000000`00B9`0gooo0?ooo`070?ooo`030000003oool0oooo
0?l0ooooPP3oool000h0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`03
0?oOW0180000M;oo00`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003`3oool00`3OW4P004RL
0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0gia80000B02Lgol02`3oool00`000000oooo
0?ooo`3o0?oooh80oooo000?0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`03
0?noM000000007Bo00X0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00403oool01@3ookl0M000
0000000007@0_ooo00/0oooo00D0oooO09a800000000001d0;ooo`070?ooo`030000003oool0oooo
0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo
0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<00000
0?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`<0
0000o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003o
ool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`00
0000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo
00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/
0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?oo
o`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`<00000o`3ooon20?ooo`00;03o
ool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool0
02`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80
oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3o
oon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo
0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo
0?ooo`3o0?oooh80oooo000/0?ooo`<00000o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o
0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3o
ool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003o
ool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00403oool01@3o
gi`0B0000000000007@0_ooo00<0oooo00<0okmd0000001d_ol00`3oool01@3OW4P0000000000000
0000B9cO00T0oooo00<000000?ooo`3oool0o`3ooon20?ooo`003`3oool01`3ookl0M00007Boo`3o
ool0okmd0000M02oool00P3oool00`3o_g@0000007Boo`020?ooo`070=nLB0000000M;oo0?ooo`3o
oml0W4P004RLg`080?ooo`030000003oool0oooo0?l0ooooPP3oool000l0oooo00<0_g@00018W03O
ool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?oog`2LB00007Bo00<0oooo00<0okmd0018W03O
ool01`3oool00`000000oooo0?ooo`3o0?oooh80oooo000>0?ooo`030?oo_`1d0000B9cO00<0oooo
00<0gia80000M02oool01P3oool00`2oM00007Bo0?ooo`020?ooo`030;md0000B9`0gooo00L0oooo
00<000000?ooo`3oool0o`3ooon20?ooo`003P3oool00`3ogi`0B00007Boo`030?ooo`030?noM000
04P0W=oo00H0oooo00L0omnL04P0M02oool0oooo0;md00000000M;oo00P0oooo00<000000?ooo`3o
ool0o`3ooon20?ooo`003P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo
00D0oono07A8B0180000000004RLg`090?ooo`@00000o`3ooon10?ooo`003P3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00L0oooo00@0omnL04P00000000007Bo2P3oool00`000000
oooo0?ooo`3o0?oooh80oooo000>0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
1P3oool01P3ogi`0B0000000M02oool0okmd001d_`T0oooo00<000000?ooo`3oool0o`3ooon20?oo
o`003P3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00<0_g@00000M02o
ool00P3oool00`2oM000B9cO0?ooo`070?ooo`030000003oool0oooo0?l0ooooPP3oool000l0oooo
00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`060?ooo`030;md0000B9`0gooo0080oooo
00<0gia8001d_`3oool01`3oool00`000000oooo0?ooo`3o0?oooh80oooo000?0?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`070?ooo`070?noM0000000M;oo0?ooo`3ookl0M00007Boo`08
0?ooo`030000003oool0oooo0?l0ooooPP3oool00100oooo00D0oono07@000000000001d0;ooo`09
0?ooo`050?noM0000000000000000018W=l02@3oool00`000000oooo0?ooo`3o0?oooh80oooo000/
0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?oo
o`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0oooo
PP3oool002`0oooo0`00003o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool0
02`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80
oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3o
oon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo
0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo
0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo0`00003o
0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3o
ool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003o
ool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`00
0000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo
00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/
0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo0`00003o0?oooh80oooo000/0?ooo`03
0000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03o
ool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool0
02`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80
oooo000N0?ooo`060;md00000000000000000000B9`0gooo203oool00`000000oooo0?ooo`3o0?oo
oh80oooo000O0?ooo`030?oOW0180000M;oo00X0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00
7`3oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo0?l0ooooPP3oool001l0oooo00<0
omnL04P0001d_ol02P3oool00`000000oooo0?ooo`3o0?oooh80oooo000O0?ooo`030?oOW0180000
M;oo00X0oooo00<000000?ooo`3oool0o`3ooon20?ooo`007`3oool00`3ogi`0B00007Boo`0:0?oo
o`@00000o`3ooon10?ooo`007`3oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo0?l0
ooooPP3oool001l0oooo00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`3o0?oooh80oooo
000O0?ooo`030?oOW0180000M;oo00X0oooo00<000000?ooo`3oool0o`3ooon20?ooo`007`3oool0
0`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo0?l0ooooPP3oool001d0oooo00D0oooO09a8
00000000000007Boo`0:0?ooo`030000003oool0oooo0?l0ooooPP3oool001l0oooo00<0okmd0000
001d_ol0o`3ooon?0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.122957, -0.125454, \
0.00410595, 0.00664356}}]
}, Closed]],

Cell[TextData[{
  "The solution to this difficulty is to wrap our function inside the ",
  ButtonBox["Evaluate",
    ButtonStyle->"RefGuideLink"],
  " function.  ",
  StyleBox["Evaluate", "InlineInput",
    FontWeight->"Bold"],
  " instructs ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " to evaluate an expression and the commands it contains first before \
passing the contents as an argument for processing by another command.  Thus, \
if we evaluate ",
  StyleBox["g[x]", "InlineInput"],
  " first, the argument becomes a well-defined function which can be plotted \
directly."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Evaluate[g[x]], \ {x, 0, 10}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0952381 0.0147151 0.0307874 [
[.21429 .00222 -5 -15.6875 ]
[.21429 .00222 5 0 ]
[.40476 .00222 -5 -15.6875 ]
[.40476 .00222 5 0 ]
[.59524 .00222 -5 -15.6875 ]
[.59524 .00222 5 0 ]
[.78571 .00222 -5 -15.6875 ]
[.78571 .00222 5 0 ]
[.97619 .00222 -8 -15.6875 ]
[.97619 .00222 8 0 ]
[.01131 .16865 -10 -7.84375 ]
[.01131 .16865 0 7.84375 ]
[.01131 .32259 -16 -7.84375 ]
[.01131 .32259 0 7.84375 ]
[.01131 .47653 -16 -7.84375 ]
[.01131 .47653 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .01472 m
.21429 .02097 L
s
gsave
.21429 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.40476 .01472 m
.40476 .02097 L
s
gsave
.40476 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.59524 .01472 m
.59524 .02097 L
s
gsave
.59524 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.78571 .01472 m
.78571 .02097 L
s
gsave
.78571 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(8) show
1.000 setlinewidth
grestore
.97619 .01472 m
.97619 .02097 L
s
gsave
.97619 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .01472 m
.07143 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.16667 .01472 m
.16667 .01847 L
s
.2619 .01472 m
.2619 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.35714 .01472 m
.35714 .01847 L
s
.45238 .01472 m
.45238 .01847 L
s
.5 .01472 m
.5 .01847 L
s
.54762 .01472 m
.54762 .01847 L
s
.64286 .01472 m
.64286 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.7381 .01472 m
.7381 .01847 L
s
.83333 .01472 m
.83333 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.92857 .01472 m
.92857 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .16865 m
.03006 .16865 L
s
gsave
.01131 .16865 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.02381 .32259 m
.03006 .32259 L
s
gsave
.01131 .32259 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.02381 .47653 m
.03006 .47653 L
s
gsave
.01131 .47653 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(15) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .0455 m
.02756 .0455 L
s
.02381 .07629 m
.02756 .07629 L
s
.02381 .10708 m
.02756 .10708 L
s
.02381 .13786 m
.02756 .13786 L
s
.02381 .19944 m
.02756 .19944 L
s
.02381 .23023 m
.02756 .23023 L
s
.02381 .26101 m
.02756 .26101 L
s
.02381 .2918 m
.02756 .2918 L
s
.02381 .35338 m
.02756 .35338 L
s
.02381 .38416 m
.02756 .38416 L
s
.02381 .41495 m
.02756 .41495 L
s
.02381 .44574 m
.02756 .44574 L
s
.02381 .50731 m
.02756 .50731 L
s
.02381 .5381 m
.02756 .5381 L
s
.02381 .56889 m
.02756 .56889 L
s
.02381 .59968 m
.02756 .59968 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01472 m
.06244 .02727 L
.10458 .04209 L
.14415 .05949 L
.18221 .08219 L
.22272 .11519 L
.26171 .1563 L
.30316 .20815 L
.34309 .26131 L
.3815 .30887 L
.40095 .32929 L
.41127 .33867 L
.42237 .34745 L
.43298 .35443 L
.44268 .35951 L
.44735 .36148 L
.45244 .36326 L
.45725 .36459 L
.46172 .36551 L
.46436 .36591 L
.46679 .36618 L
.46807 .36629 L
.46946 .36638 L
.47017 .36641 L
.47092 .36643 L
.47163 .36645 L
.47228 .36645 L
.47356 .36645 L
.47473 .36642 L
.47602 .36636 L
.47668 .36632 L
.4774 .36627 L
.47989 .36603 L
.48224 .36571 L
.48691 .3648 L
.49119 .36366 L
.4959 .36206 L
.50085 .35999 L
.51039 .35487 L
.52059 .3478 L
.53895 .33104 L
.5603 .30566 L
.58045 .27687 L
.61892 .21377 L
.65984 .14593 L
.68019 .11661 L
.69924 .09416 L
.70861 .08534 L
.71751 .07853 L
.72241 .07547 L
Mistroke
.72776 .07272 L
.73262 .07079 L
.73713 .06948 L
.73842 .06919 L
.73978 .06894 L
.74096 .06875 L
.74222 .06858 L
.74293 .06851 L
.74358 .06845 L
.74485 .06837 L
.74555 .06834 L
.74632 .06832 L
.74705 .06832 L
.74772 .06833 L
.74891 .06837 L
.75002 .06844 L
.75124 .06856 L
.75256 .06872 L
.75524 .0692 L
.75769 .0698 L
.76203 .07124 L
.76677 .07337 L
.77631 .07947 L
.78661 .08879 L
.79596 .09972 L
.81692 .13261 L
.83615 .17247 L
.85696 .22496 L
.89793 .34962 L
.93738 .48164 L
.97532 .60083 L
.97619 .60332 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00J`3oool00`2oM0000000000000040000000304RLg`3oool0oooo04@0oooo00@0oooO09a8
0000B9`0goooA@3oool01@3ogi`0B000000000000000M;oo04H0oooo00D0gia80000000000000000
04RLg`120?ooo`060;md00000000000000000000B9`0gooo0`3oool01@3ogi`0B0000000000007@0
_ooo00h0oooo001/0?ooo`030=nLB0000000000000<0000000<004RL0=ooo`3oool0A03oool0103o
oml0W4P00018W03Ooom40?ooo`070?oOW0180000B9cO0?ooo`3ookl0M00007Boo`140?ooo`070=nL
B0000000M;oo0?ooo`3ooml0W4P004RLg`120?ooo`030?oOW0180000M;oo00@0oooo00L0oono07@0
001d_ol0oooo0?noM00007@0_ooo00d0oooo001]0?ooo`030=nLB018W=l0oooo0080oooo00<0giad
09cOo`3oool0A03oool0103ooml0W4P00018W03Ooom30?ooo`030?oog`2LB00007Bo00<0oooo00<0
okmd0018W03Oool0@P3oool00`3ooml0W4P0001d_`030?ooo`030?noM000B9`0gooo0440oooo00<0
omnL04P0001d_ol0103oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00`0oooo001^
0?ooo`030=nLB01d_ol0oooo04@0oooo00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`10
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0@`3oool00`2oM00007Bo0?ooo`02
0?ooo`030;md0000B9`0gooo0440oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M00004RLg`03
0?ooo`030=nLB00007@0_ooo00`0oooo001_0?ooo`030;mdB02Lgol0oooo04<0oooo00P0omnL07BL
g`3oool0oooo0?oog`2LB00004RL0=oood<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`130?ooo`070?oOW01807@0_ooo0?ooo`2oM000000007Boo`120?ooo`030?oOW0180000M;oo
00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0<0?ooo`00K`3oool00`3ooml0
W4Qd0;ooo`140?ooo`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`130?ooo`030?oOW0180000
M;oo00<0oooo00<0_g@00000B02Lgol0A03oool01@3ookl0M4Q804P000000000B9cO04<0oooo00<0
omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00`0oooo001`
0?ooo`030?noM000M;l0oooo04<0oooo00L0oooO09a8M02oool0oooO09a80000B9`0gooo04<0oooo
00P0oono07@000000000B9cO0?ooo`3o_g@00000001d_dD0oooo00@0omnL04P00000000007BoA03o
ool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0303o
ool00700oooo00<0oono07@0B02Lgol0A03oool01P3ogi`0B7Bo0?oog`2LB00004RL0=oood@0oooo
00<0_g@00018W02oW4P00P0000000`00B9`0gooo0?ooo`130?ooo`060?oOW0180000001d0;ooo`3o
_g@007Bo@`3oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol0303oool00700oooo00<0oono07@00018W=l0A@3oool01@3OW4P0M;oO09a80000B9`0gooo
04@0oooo00<0omnL04P0001d_ol0B03oool00`2oM000001d0;ooo`020?ooo`030;md0018W=l0oooo
0440oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo
00`0oooo001[0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l0A@3oool01@3ooml0
W4Q804Q80000B9`0gooo04D0oooo00<0gia80000B02Lgol0A`3oool00`2oM00004RL0=ooo`020?oo
o`030=nLB000M;l0oooo0440oooo00<0omnL04P0001d_ol0103oool00`3OW4P004RL0=ooo`020?oo
o`030;md0000M;l0oooo00`0oooo001[0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02o
oom60?ooo`040?oOW018000004RL0=ooodH0oooo00<0okmd00000000M;l0AP3oool01`3o_g@00000
07Boo`3oool0oono07@0001d_ol0@03oool01@3ooml0W4P0000000000000M;oo00@0oooo00L0oono
07@0001d_ol0oooo0?noM00007@0_ooo00d0oooo001/0?ooo`060?oog`2LB00000000000000007@0
_oooB03oool00`3OW4P004RL0=ooo`170?ooo`050?oog`2LB00000000000M02oool0A03oool01@3o
_g@00000000000000000B9cO04<0oooo00<0okmd0000001d_ol01@3oool01@3ookl0M00000000000
07@0_ooo00h0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
000T0?ooo`030000003oool0oooo0?l0ooooRP3oool002@0oooo00<000000?ooo`3oool0o`3ooon:
0?ooo`00903oool00`000000oooo0?ooo`3o0?ooohX0oooo000T0?ooo`030000003oool0oooo0?l0
ooooRP3oool002@0oooo00<000000?ooo`3oool0o`3ooon:0?ooo`00903oool00`000000oooo0?oo
o`3o0?ooohX0oooo000K0?ooool00000R`00000;0?ooo`00903oool00`000000oooo000000020000
00h0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0B0?ooo`00903oool00`000000
oooo0?ooo`020?ooo`<000002`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0180
oooo000T0?ooo`030000003oool0oooo00D0oooo0`0000100?ooo`030000003oool0oooo04P0oooo
00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo04P0oooo
00<000000?ooo`3oool04P3oool002@0oooo00<000000?ooo`3oool0203oool300000?l0ooooO`3o
ool002@0oooo00<000000?ooo`3oool02`3oool300000?l0ooooO03oool002@0oooo00<000000?oo
o`3oool03P3oool300000?l0ooooN@3oool002@0oooo00<000000?ooo`3oool04@3oool300000?l0
ooooMP3oool002@0oooo00<000000?ooo`3oool0503oool200000?l0ooooM03oool002@0oooo00<0
00000?ooo`3oool05P3oool300000?l0ooooL@3oool002@0oooo00<000000?ooo`3oool06@3oool3
00000?l0ooooKP3oool002@0oooo00<000000?ooo`3oool0703oool300000?l0ooooJ`3oool002@0
oooo0`00000O0?ooo`800000o`3ooomY0?ooo`00903oool00`000000oooo0?ooo`0Q0?ooo`800000
o`3ooomW0?ooo`00903oool00`000000oooo0?ooo`0S0?ooo`800000o`3ooomU0?ooo`00903oool0
0`000000oooo0?ooo`0U0?ooo`800000o`3ooomS0?ooo`00903oool00`000000oooo0?ooo`0W0?oo
o`800000o`3ooomQ0?ooo`00903oool00`000000oooo0?ooo`0Y0?ooo`800000o`3ooomO0?ooo`00
903oool00`000000oooo0?ooo`0[0?ooo`800000o`3ooomM0?ooo`00903oool00`000000oooo0?oo
o`0]0?ooo`800000o`3ooomK0?ooo`00903oool00`000000oooo0?ooo`0_0?ooo`800000o`3ooomI
0?ooo`00903oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo0><0oooo1P00001/0?oo
o`00903oool00`000000oooo0?ooo`0b0?ooo`800000h03oool3000000H0oooo1000001X0?ooo`00
903oool00`000000oooo0?ooo`0d0?ooo`800000g03oool2000000d0oooo00<000000?ooo`3oool0
I@3oool002@0oooo0`00000f0?ooo`030000003oool0oooo0=L0oooo0P00000@0?ooo`030000003o
ool0oooo06@0oooo000T0?ooo`030000003oool0oooo03L0oooo0P00003E0?ooo`8000004`3oool2
000006@0oooo000T0?ooo`030000003oool0oooo03T0oooo0P00003B0?ooo`030000003oool0oooo
01D0oooo00<000000?ooo`3oool0H@3oool002@0oooo00<000000?ooo`3oool0>`3oool00`000000
oooo0?ooo`3>0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0H03oool002@0oooo
00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`3<0?ooo`030000003oool0oooo01T0oooo
0P00001P0?ooo`00903oool00`000000oooo0?ooo`0m0?ooo`800000b`3oool00`000000oooo0?oo
o`0L0?ooo`030000003oool0oooo05d0oooo000T0?ooo`030000003oool0oooo03l0oooo00<00000
0?ooo`3oool0aP3oool200000200oooo00<000000?ooo`3oool0G03oool002@0oooo00<000000?oo
o`3oool0@03oool00`000000oooo0?ooo`340?ooo`030000003oool0oooo0240oooo00<000000?oo
o`3oool0F`3oool002@0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`320?ooo`03
0000003oool0oooo0280oooo00<000000?ooo`3oool0F`3oool002@0oooo00<000000?ooo`3oool0
@P3oool00`000000oooo0?ooo`300?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0
FP3oool002@0oooo00<000000?ooo`3oool0@`3oool200000;l0oooo00<000000?ooo`3oool09P3o
ool00`000000oooo0?ooo`1I0?ooo`00903oool00`000000oooo0?ooo`150?ooo`030000003oool0
oooo0;/0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`1H0?ooo`00903oool30000
04H0oooo00<000000?ooo`3oool0^P3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo
05P0oooo000T0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0^03oool00`000000
oooo0?ooo`0Z0?ooo`030000003oool0oooo05L0oooo000T0?ooo`030000003oool0oooo04P0oooo
0P00002g0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0EP3oool002@0oooo00<0
00000?ooo`3oool0BP3oool00`000000oooo0?ooo`2c0?ooo`030000003oool0oooo02d0oooo00<0
00000?ooo`3oool0EP3oool002@0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`2a
0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0E@3oool002@0oooo00<000000?oo
o`3oool0C03oool00`000000oooo0?ooo`2_0?ooo`030000003oool0oooo0340oooo00<000000?oo
o`3oool0E03oool002@0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`2^0?ooo`03
0000003oool0oooo0340oooo00<000000?ooo`3oool0E03oool002@0oooo00<000000?ooo`3oool0
CP3oool00`000000oooo0?ooo`2/0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0
D`3oool002@0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`2Z0?ooo`030000003o
ool0oooo03D0oooo00<000000?ooo`3oool0DP3oool002@0oooo00<000000?ooo`3oool0D03oool0
0`000000oooo0?ooo`2Y0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0DP3oool0
02@0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo
03L0oooo00<000000?ooo`3oool0D@3oool002@0oooo00<000000?ooo`3oool0DP3oool00`000000
oooo0?ooo`2U0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool0D@3oool002@0oooo
00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`2T0?ooo`030000003oool0oooo03T0oooo
00<000000?ooo`3oool0D03oool002@0oooo0`00001D0?ooo`030000003oool0oooo0:80oooo00<0
00000?ooo`3oool0>P3oool00`000000oooo0?ooo`1@0?ooo`00903oool00`000000oooo0?ooo`1E
0?ooo`030000003oool0oooo0:00oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`1?
0?ooo`00903oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo09l0oooo00<000000?oo
o`3oool0?03oool00`000000oooo0?ooo`1?0?ooo`00903oool00`000000oooo0?ooo`1G0?ooo`03
0000003oool0oooo09d0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`1>0?ooo`00
903oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo09/0oooo00<000000?ooo`3oool0
?`3oool00`000000oooo0?ooo`1>0?ooo`00903oool00`000000oooo0?ooo`1I0?ooo`030000003o
ool0oooo09X0oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`1=0?ooo`00903oool0
0`000000oooo0?ooo`1J0?ooo`030000003oool0oooo09P0oooo00<000000?ooo`3oool0@@3oool0
0`000000oooo0?ooo`1=0?ooo`004`3oool00`3ooml0W4P0000000020000000307Boo`3oool0oooo
00T0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`2G0?ooo`030000003oool0oooo
0480oooo00<000000?ooo`3oool0C03oool001<0oooo00<0oooO09a8M02oool00P3oool00`2oM000
M;oo0?ooo`090?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3oool0U@3oool00`000000
oooo0?ooo`130?ooo`030000003oool0oooo04`0oooo000H0?ooo`030?oog`2LB7@0_ooo00T0oooo
00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`2C0?ooo`030000003oool0oooo04D0oooo
00<000000?ooo`3oool0B`3oool001T0oooo00<0_g@007Boo`3oool0203oool00`000000oooo0?oo
o`1M0?ooo`030000003oool0oooo09<0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?oo
o`1;0?ooo`00603oool00`3ooml0W4P004RLg`090?ooo`030000003oool0oooo05h0oooo00<00000
0?ooo`3oool0T@3oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo04X0oooo000H0?oo
o`030=nLB0000000M;oo00T0oooo1000001N0?ooo`030000003oool0oooo08l0oooo00<000000?oo
o`3oool0B03oool00`000000oooo0?ooo`1:0?ooo`00503oool01`3Ogml0oooo0?oo_`1d00000000
0000M02oool02@3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo08h0oooo00<00000
0?ooo`3oool0B@3oool00`000000oooo0?ooo`190?ooo`00503oool01P3OW4P00000000000000000
001809cOo`X0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`2<0?ooo`030000003o
ool0oooo04X0oooo00<000000?ooo`3oool0B@3oool001@0oooo00@0oono07@000000000B9cO303o
ool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool0B`3o
ool00`000000oooo0?ooo`180?ooo`005@3oool00`2oM000M;oo0?ooo`0<0?ooo`030000003oool0
oooo0680oooo00<000000?ooo`3oool0RP3oool00`000000oooo0?ooo`1<0?ooo`030000003oool0
oooo04P0oooo000E0?ooo`060?oOW018000000000000000004P0W=oo2@3oool00`000000oooo0?oo
o`1S0?ooo`030000003oool0oooo08P0oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?oo
o`170?ooo`005P3oool01@2oM0000000000000000000B9cO00T0oooo00<000000?ooo`3oool0I03o
ool00`000000oooo0?ooo`270?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool0A`3o
ool002@0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`250?ooo`030000003oool0
oooo04l0oooo00<000000?ooo`3oool0A`3oool002@0oooo00<000000?ooo`3oool0I@3oool00`00
0000oooo0?ooo`250?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0AP3oool002@0
oooo00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`230?ooo`030000003oool0oooo0540
oooo00<000000?ooo`3oool0AP3oool002@0oooo00<000000?ooo`3oool0I`3oool00`000000oooo
0?ooo`210?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0AP3oool002@0oooo00<0
00000?ooo`3oool0J03oool00`000000oooo0?ooo`200?ooo`030000003oool0oooo05<0oooo00<0
00000?ooo`3oool0A@3oool002@0oooo0`00001Y0?ooo`030000003oool0oooo07h0oooo00<00000
0?ooo`3oool0E03oool00`000000oooo0?ooo`150?ooo`00903oool00`000000oooo0?ooo`1Y0?oo
o`030000003oool0oooo07d0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`140?oo
o`00903oool00`000000oooo0?ooo`1Z0?ooo`030000003oool0oooo07`0oooo00<000000?ooo`3o
ool0EP3oool00`000000oooo0?ooo`140?ooo`00903oool00`000000oooo0?ooo`1[0?ooo`030000
003oool0oooo07X0oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`140?ooo`00903o
ool00`000000oooo0?ooo`1/0?ooo`030000003oool0oooo07T0oooo00<000000?ooo`3oool0F03o
ool00`000000oooo0?ooo`130?ooo`00903oool00`000000oooo0?ooo`1]0?ooo`030000003oool0
oooo07L0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`130?ooo`00903oool00`00
0000oooo0?ooo`1]0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3oool0FP3oool00`00
0000oooo0?ooo`130?ooo`00903oool00`000000oooo0?ooo`1^0?ooo`030000003oool0oooo07D0
oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`120?ooo`00903oool00`000000oooo
0?ooo`1_0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool0G03oool00`000000oooo
0?ooo`120?ooo`00903oool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo0780oooo00<0
00000?ooo`3oool0G@3oool00`000000oooo0?ooo`110?ooo`00903oool00`000000oooo0?ooo`1`
0?ooo`030000003oool0oooo0740oooo00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`11
0?ooo`00903oool00`000000oooo0?ooo`1a0?ooo`030000003oool0oooo06l0oooo00<000000?oo
o`3oool0G`3oool00`000000oooo0?ooo`110?ooo`00903oool300000780oooo00<000000?ooo`3o
ool0KP3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo0400oooo000T0?ooo`030000
003oool0oooo07<0oooo00<000000?ooo`3oool0K03oool00`000000oooo0?ooo`1Q0?ooo`030000
003oool0oooo0400oooo000T0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool0K03o
ool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo0400oooo000T0?ooo`030000003oool0
oooo07@0oooo00<000000?ooo`3oool0JP3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
oooo03l0oooo000T0?ooo`030000003oool0oooo07D0oooo00<000000?ooo`3oool0J03oool00`00
0000oooo0?ooo`1T0?ooo`030000003oool0oooo03l0oooo000T0?ooo`030000003oool0oooo07H0
oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo03l0
oooo000T0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3oool0IP3oool00`000000oooo
0?ooo`1V0?ooo`030000003oool0oooo03h0oooo000T0?ooo`030000003oool0oooo07L0oooo00<0
00000?ooo`3oool0I@3oool00`000000oooo0?ooo`1V0?ooo`030000003oool0oooo03h0oooo000T
0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1W
0?ooo`030000003oool0oooo03h0oooo000T0?ooo`030000003oool0oooo07T0oooo00<000000?oo
o`3oool0H@3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo03d0oooo000T0?ooo`03
0000003oool0oooo07T0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`1Y0?ooo`03
0000003oool0oooo03d0oooo000T0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0
G`3oool00`000000oooo0?ooo`1Z0?ooo`030000003oool0oooo03d0oooo000T0?ooo`<00000N`3o
ool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo06/0oooo00<000000?ooo`3oool0?03o
ool002@0oooo00<000000?ooo`3oool0O03oool00`000000oooo0?ooo`1L0?ooo`030000003oool0
oooo06`0oooo00<000000?ooo`3oool0?03oool002@0oooo00<000000?ooo`3oool0O@3oool00`00
0000oooo0?ooo`1J0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0?03oool002@0
oooo00<000000?ooo`3oool0O@3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo06h0
oooo00<000000?ooo`3oool0>`3oool002@0oooo00<000000?ooo`3oool0OP3oool00`000000oooo
0?ooo`1H0?ooo`030000003oool0oooo06l0oooo00<000000?ooo`3oool0>`3oool002@0oooo00<0
00000?ooo`3oool0O`3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo06l0oooo00<0
00000?ooo`3oool0>`3oool002@0oooo00<000000?ooo`3oool0P03oool00`000000oooo0?ooo`1E
0?ooo`030000003oool0oooo0740oooo00<000000?ooo`3oool0>P3oool002@0oooo00<000000?oo
o`3oool0P@3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo0780oooo00<000000?oo
o`3oool0>P3oool002@0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`1C0?ooo`03
0000003oool0oooo0780oooo00<000000?ooo`3oool0>P3oool002@0oooo00<000000?ooo`3oool0
PP3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3oool0
>@3oool002@0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`1?0?ooo`030000003o
ool0oooo07D0oooo00<000000?ooo`3oool0>@3oool002@0oooo00<000000?ooo`3oool0Q03oool0
0`000000oooo0?ooo`1>0?ooo`030000003oool0oooo07D0oooo00<000000?ooo`3oool0>@3oool0
02@0oooo0`0000240?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3oool0MP3oool00`00
0000oooo0?ooo`0i0?ooo`00903oool00`000000oooo0?ooo`250?ooo`030000003oool0oooo04/0
oooo00<000000?ooo`3oool0N03oool00`000000oooo0?ooo`0h0?ooo`00903oool00`000000oooo
0?ooo`260?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0N03oool00`000000oooo
0?ooo`0h0?ooo`00903oool00`000000oooo0?ooo`270?ooo`030000003oool0oooo04P0oooo00<0
00000?ooo`3oool0N@3oool00`000000oooo0?ooo`0h0?ooo`00903oool00`000000oooo0?ooo`28
0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0N`3oool00`000000oooo0?ooo`0g
0?ooo`00903oool00`000000oooo0?ooo`280?ooo`030000003oool0oooo04H0oooo00<000000?oo
o`3oool0N`3oool00`000000oooo0?ooo`0g0?ooo`00903oool00`000000oooo0?ooo`290?ooo`03
0000003oool0oooo04@0oooo00<000000?ooo`3oool0O03oool00`000000oooo0?ooo`0g0?ooo`00
303oool01P2oM000000000000000000004RL0=ooo`<0oooo00D0omnL04P000000000001d0;ooo`0:
0?ooo`030000003oool0oooo08X0oooo00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`1n
0?ooo`030000003oool0oooo03H0oooo000=0?ooo`030?oOW0180000M;oo00@0oooo00L0oono07@0
001d_ol0oooo0?noM00007@0_ooo00T0oooo00<000000?ooo`3oool0R`3oool00`000000oooo0?oo
o`100?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool0=P3oool000d0oooo00<0omnL
04P0001d_ol0103oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00P0oooo00<00000
0?ooo`3oool0S03oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo07l0oooo00<00000
0?ooo`3oool0=P3oool000d0oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M00004RLg`030?oo
o`030=nLB00007@0_ooo00P0oooo00<000000?ooo`3oool0S@3oool00`000000oooo0?ooo`0m0?oo
o`030000003oool0oooo0840oooo00<000000?ooo`3oool0=@3oool000d0oooo00<0omnL04P0001d
_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00P0oooo00<000000?ooo`3o
ool0S@3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo0880oooo00<000000?ooo`3o
ool0=@3oool000d0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00P0oooo1000002=0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0
P`3oool00`000000oooo0?ooo`0e0?ooo`003@3oool00`3ogi`0B00007Boo`030?ooo`030?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol0203oool00`000000oooo0?ooo`2?0?ooo`030000003o
ool0oooo03P0oooo00<000000?ooo`3oool0Q@3oool00`000000oooo0?ooo`0d0?ooo`003@3oool0
0`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0203oool0
0`000000oooo0?ooo`2@0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0Q@3oool0
0`000000oooo0?ooo`0d0?ooo`003@3oool00`3ogi`0B00007Boo`030?ooo`030?oo_`1d0000B9cO
00<0oooo00<0gia80000M02oool0203oool00`000000oooo0?ooo`2A0?ooo`030000003oool0oooo
03D0oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`0d0?ooo`003@3oool00`3ogi`0
B00007Boo`040?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0203oool00`000000
oooo0?ooo`2B0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0R03oool00`000000
oooo0?ooo`0c0?ooo`002`3oool01@3ooml0W4P0000000000000M;oo00@0oooo00L0oono07@0001d
_ol0oooo0?noM00007@0_ooo00T0oooo00<000000?ooo`3oool0T`3oool200000380oooo00<00000
0?ooo`3oool0R@3oool00`000000oooo0?ooo`0c0?ooo`003@3oool00`3o_g@0000007Boo`050?oo
o`050?oo_`1d000000000000M02oool02P3oool00`000000oooo0?ooo`2E0?ooo`030000003oool0
oooo02h0oooo00<000000?ooo`3oool0RP3oool00`000000oooo0?ooo`0c0?ooo`00903oool00`00
0000oooo0?ooo`2F0?ooo`030000003oool0oooo02/0oooo0P00002>0?ooo`030000003oool0oooo
0380oooo000T0?ooo`030000003oool0oooo09L0oooo00<000000?ooo`3oool0:@3oool00`000000
oooo0?ooo`2>0?ooo`030000003oool0oooo0380oooo000T0?ooo`030000003oool0oooo09P0oooo
0P00000X0?ooo`030000003oool0oooo08l0oooo00<000000?ooo`3oool0<P3oool002@0oooo00<0
00000?ooo`3oool0VP3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo0940oooo00<0
00000?ooo`3oool0<@3oool002@0oooo00<000000?ooo`3oool0V`3oool00`000000oooo0?ooo`0R
0?ooo`030000003oool0oooo0980oooo00<000000?ooo`3oool0<@3oool002@0oooo0`00002L0?oo
o`800000803oool2000009D0oooo00<000000?ooo`3oool0<@3oool002@0oooo00<000000?ooo`3o
ool0WP3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo09H0oooo00<000000?ooo`3o
ool0<03oool002@0oooo00<000000?ooo`3oool0W`3oool2000001X0oooo0P00002I0?ooo`030000
003oool0oooo0300oooo000T0?ooo`030000003oool0oooo0:40oooo0P00000F0?ooo`800000V`3o
ool00`000000oooo0?ooo`0`0?ooo`00903oool00`000000oooo0?ooo`2S0?ooo`@000003P3oool4
000009h0oooo00<000000?ooo`3oool0;`3oool002@0oooo00<000000?ooo`3oool0Y`3oool>0000
0:80oooo00<000000?ooo`3oool0;`3oool002@0oooo00<000000?ooo`3oool0o`3ooomH0?ooo`03
0000003oool0oooo02l0oooo000T0?ooo`030000003oool0oooo0?l0ooooF03oool00`000000oooo
0?ooo`0_0?ooo`00903oool00`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3oool0;P3o
ool002@0oooo00<000000?ooo`3oool0o`3ooomI0?ooo`030000003oool0oooo02h0oooo000T0?oo
o`030000003oool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0^0?ooo`00903oool00`000000
oooo0?ooo`3o0?oooeX0oooo00<000000?ooo`3oool0;@3oool002@0oooo0`00003o0?oooeX0oooo
00<000000?ooo`3oool0;@3oool002@0oooo00<000000?ooo`3oool0o`3ooomJ0?ooo`030000003o
ool0oooo02d0oooo000T0?ooo`030000003oool0oooo0?l0ooooF`3oool00`000000oooo0?ooo`0/
0?ooo`00903oool00`000000oooo0?ooo`3o0?oooe/0oooo00<000000?ooo`3oool0;03oool002@0
oooo00<000000?ooo`3oool0o`3ooomK0?ooo`030000003oool0oooo02`0oooo000T0?ooo`030000
003oool0oooo0?l0ooooG03oool00`000000oooo0?ooo`0[0?ooo`00903oool00`000000oooo0?oo
o`3o0?oooe`0oooo00<000000?ooo`3oool0:`3oool002@0oooo00<000000?ooo`3oool0o`3ooomL
0?ooo`030000003oool0oooo02/0oooo000T0?ooo`030000003oool0oooo0?l0ooooG03oool00`00
0000oooo0?ooo`0[0?ooo`00903oool00`000000oooo0?ooo`3o0?oooed0oooo00<000000?ooo`3o
ool0:P3oool002@0oooo00<000000?ooo`3oool0o`3ooomM0?ooo`030000003oool0oooo02X0oooo
000T0?ooo`030000003oool0oooo0?l0ooooG@3oool00`000000oooo0?ooo`0Z0?ooo`00903oool0
0`000000oooo0?ooo`3o0?oooeh0oooo00<000000?ooo`3oool0:@3oool002@0oooo0`00003o0?oo
oeh0oooo00<000000?ooo`3oool0:@3oool002@0oooo00<000000?ooo`3oool0o`3ooomN0?ooo`03
0000003oool0oooo02T0oooo000T0?ooo`030000003oool0oooo0?l0ooooG`3oool00`000000oooo
0?ooo`0X0?ooo`00903oool00`000000oooo0?ooo`3o0?oooel0oooo00<000000?ooo`3oool0:03o
ool002@0oooo00<000000?ooo`3oool0o`3ooomO0?ooo`030000003oool0oooo02P0oooo000T0?oo
o`030000003oool0oooo0?l0ooooH03oool00`000000oooo0?ooo`0W0?ooo`00903oool00`000000
oooo0?ooo`3o0?ooof00oooo00<000000?ooo`3oool09`3oool002@0oooo00<000000?ooo`3oool0
o`3ooomP0?ooo`030000003oool0oooo02L0oooo000T0?ooo`030000003oool0oooo0?l0ooooH03o
ool00`000000oooo0?ooo`0W0?ooo`00903oool00`000000oooo0?ooo`3o0?ooof40oooo00<00000
0?ooo`3oool09P3oool002@0oooo00<000000?ooo`3oool0o`3ooomQ0?ooo`030000003oool0oooo
02H0oooo000T0?ooo`030000003oool0oooo0?l0ooooH@3oool00`000000oooo0?ooo`0V0?ooo`00
903oool300000?l0ooooHP3oool00`000000oooo0?ooo`0U0?ooo`00903oool00`000000oooo0?oo
o`3o0?ooof80oooo00<000000?ooo`3oool09@3oool002@0oooo00<000000?ooo`3oool0o`3ooomR
0?ooo`030000003oool0oooo02D0oooo000T0?ooo`030000003oool0oooo0?l0ooooH`3oool00`00
0000oooo0?ooo`0T0?ooo`00903oool00`000000oooo0?ooo`3o0?ooof<0oooo00<000000?ooo`3o
ool0903oool002@0oooo00<000000?ooo`3oool0o`3ooomS0?ooo`030000003oool0oooo02@0oooo
000T0?ooo`030000003oool0oooo0?l0ooooI03oool00`000000oooo0?ooo`0S0?ooo`00303oool0
2@2oM000000000000000000004RL0=ooo`3oool0oooO09a800030000000307Boo`3oool0oooo00T0
oooo00<000000?ooo`3oool0o`3ooomT0?ooo`030000003oool0oooo02<0oooo000=0?ooo`030?oO
W0180000M;oo00<0oooo00<0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`090?ooo`030000
003oool0oooo0?l0ooooI03oool00`000000oooo0?ooo`0S0?ooo`003@3oool00`3ogi`0B00007Bo
o`080?ooo`030?oog`2LB7@0_ooo00T0oooo00<000000?ooo`3oool0o`3ooomT0?ooo`030000003o
ool0oooo02<0oooo000=0?ooo`030?oOW0180000M;oo00T0oooo00<0_g@007Boo`3oool0203oool0
0`000000oooo0?ooo`3o0?ooofD0oooo00<000000?ooo`3oool08P3oool000d0oooo00<0omnL04P0
001d_ol0203oool00`3ooml0W4P004RLg`090?ooo`030000003oool0oooo0?l0ooooI@3oool00`00
0000oooo0?ooo`0R0?ooo`003@3oool00`3ogi`0B00007Boo`080?ooo`030=nLB0000000M;oo00T0
oooo1000003o0?ooof@0oooo00<000000?ooo`3oool08P3oool000d0oooo00<0omnL04P0001d_ol0
103oool01`3Ogml0oooo0?oo_`1d000000000000M02oool02@3oool00`000000oooo0?ooo`3o0?oo
ofH0oooo00<000000?ooo`3oool08@3oool000d0oooo00<0omnL04P0001d_ol0103oool01P3OW4P0
0000000000000000001809cOo`X0oooo00<000000?ooo`3oool0o`3ooomV0?ooo`030000003oool0
oooo0240oooo000=0?ooo`030?oOW0180000M;oo00@0oooo00@0oono07@000000000B9cO303oool0
0`000000oooo0?ooo`3o0?ooofH0oooo00<000000?ooo`3oool08@3oool000d0oooo00<0omnL04P0
001d_ol01@3oool00`2oM000M;oo0?ooo`0<0?ooo`030000003oool0oooo0?l0ooooI`3oool00`00
0000oooo0?ooo`0P0?ooo`002`3oool01@3ooml0W4P0000000000000M;oo00D0oooo00H0omnL04P0
0000000000000000B02Lgol90?ooo`030000003oool0oooo0?l0ooooI`3oool00`000000oooo0?oo
o`0P0?ooo`003@3oool00`3o_g@0000007Boo`060?ooo`050;md00000000000000000018W=l02@3o
ool00`000000oooo0?ooo`3o0?ooofL0oooo00<000000?ooo`3oool0803oool002@0oooo00<00000
0?ooo`3oool0o`3ooomX0?ooo`030000003oool0oooo01l0oooo000T0?ooo`030000003oool0oooo
0?l0ooooJ03oool00`000000oooo0?ooo`0O0?ooo`00903oool00`000000oooo0?ooo`3o0?ooofP0
oooo00<000000?ooo`3oool07`3oool002@0oooo00<000000?ooo`3oool0o`3ooomY0?ooo`030000
003oool0oooo01h0oooo000T0?ooo`030000003oool0oooo0?l0ooooJ@3oool00`000000oooo0?oo
o`0N0?ooo`00903oool300000?l0ooooJ@3oool00`000000oooo0?ooo`0N0?ooo`00903oool00`00
0000oooo0?ooo`3o0?ooofX0oooo00<000000?ooo`3oool07@3oool002@0oooo00<000000?ooo`3o
ool0o`3ooomZ0?ooo`030000003oool0oooo01d0oooo000T0?ooo`030000003oool0oooo0?l0oooo
JP3oool00`000000oooo0?ooo`0M0?ooo`00903oool00`000000oooo0?ooo`3o0?ooofX0oooo00<0
00000?ooo`3oool07@3oool002@0oooo00<000000?ooo`3oool0o`3ooom[0?ooo`030000003oool0
oooo01`0oooo000T0?ooo`030000003oool0oooo0?l0ooooJ`3oool00`000000oooo0?ooo`0L0?oo
o`00903oool00`000000oooo0?ooo`3o0?ooof/0oooo00<000000?ooo`3oool0703oool002@0oooo
00<000000?ooo`3oool0o`3ooom/0?ooo`030000003oool0oooo01/0oooo000T0?ooo`030000003o
ool0oooo0?l0ooooK03oool00`000000oooo0?ooo`0K0?ooo`00903oool00`000000oooo0?ooo`3o
0?ooof`0oooo00<000000?ooo`3oool06`3oool002@0oooo00<000000?ooo`3oool0o`3ooom]0?oo
o`030000003oool0oooo01X0oooo000T0?ooo`<00000o`3ooom]0?ooo`030000003oool0oooo01X0
oooo000T0?ooo`030000003oool0oooo0?l0ooooK@3oool00`000000oooo0?ooo`0J0?ooo`00903o
ool00`000000oooo0?ooo`3o0?ooofh0oooo00<000000?ooo`3oool06@3oool002@0oooo00<00000
0?ooo`3oool0o`3ooom^0?ooo`030000003oool0oooo01T0oooo000T0?ooo`030000003oool0oooo
0?l0ooooKP3oool00`000000oooo0?ooo`0I0?ooo`00903oool00`000000oooo0?ooo`3o0?ooofl0
oooo00<000000?ooo`3oool0603oool002@0oooo00<000000?ooo`3oool0o`3ooom_0?ooo`030000
003oool0oooo01P0oooo000T0?ooo`030000003oool0oooo0?l0ooooK`3oool00`000000oooo0?oo
o`0H0?ooo`00903oool00`000000oooo0?ooo`3o0?ooog00oooo00<000000?ooo`3oool05`3oool0
02@0oooo00<000000?ooo`3oool0o`3ooom`0?ooo`030000003oool0oooo01L0oooo000T0?ooo`03
0000003oool0oooo0?l0ooooL03oool00`000000oooo0?ooo`0G0?ooo`00903oool00`000000oooo
0?ooo`3o0?ooog40oooo00<000000?ooo`3oool05P3oool002@0oooo0`00003o0?ooog40oooo00<0
00000?ooo`3oool05P3oool002@0oooo00<000000?ooo`3oool0o`3oooma0?ooo`030000003oool0
oooo01H0oooo000T0?ooo`030000003oool0oooo0?l0ooooL@3oool00`000000oooo0?ooo`0F0?oo
o`00903oool00`000000oooo0?ooo`3o0?ooog80oooo00<000000?ooo`3oool05@3oool002@0oooo
00<000000?ooo`3oool0o`3ooomb0?ooo`030000003oool0oooo01D0oooo000T0?ooo`030000003o
ool0oooo0?l0ooooLP3oool00`000000oooo0?ooo`0E0?ooo`00903oool00`000000oooo0?ooo`3o
0?ooog<0oooo00<000000?ooo`3oool0503oool002@0oooo00<000000?ooo`3oool0o`3ooomc0?oo
o`030000003oool0oooo01@0oooo000T0?ooo`030000003oool0oooo0?l0ooooL`3oool00`000000
oooo0?ooo`0D0?ooo`00903oool00`000000oooo0?ooo`3o0?ooog@0oooo00<000000?ooo`3oool0
4`3oool002@0oooo00<000000?ooo`3oool0o`3ooomd0?ooo`030000003oool0oooo01<0oooo000T
0?ooo`030000003oool0oooo0?l0ooooM03oool00`000000oooo0?ooo`0C0?ooo`00903oool30000
0?l0ooooM@3oool00`000000oooo0?ooo`0B0?ooo`00903oool00`000000oooo0?ooo`3o0?ooogD0
oooo00<000000?ooo`3oool04P3oool002@0oooo00<000000?ooo`3oool0o`3ooome0?ooo`030000
003oool0oooo0180oooo000T0?ooo`030000003oool0oooo0?l0ooooRP3oool002@0oooo00<00000
0?ooo`3oool0o`3ooon:0?ooo`00903oool00`000000oooo0?ooo`3o0?ooohX0oooo000T0?ooo`03
0000003oool0oooo0?l0ooooRP3oool002@0oooo00<000000?ooo`3oool0o`3ooon:0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.975554, -2.34814, \
0.0400221, 0.123805}}]
}, Closed]],

Cell["\<\
It is good practice to routinely wrap the evaluation function around the \
function we wish to plot.  It will do no harm for functions that don't need \
prior evaluation, and sometimes saves computation time, but will avoid many \
annoying messages.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Plotting function lists", "Subsection"],

Cell["\<\
The plot command can place several curves on the same graph if given a list \
of functions as its first argument.  A couple of simple examples follow.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[{LaguerreL[1, x], LaguerreL[2, x], LaguerreL[3, x]}, \ {x, 0, 
          10}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0952381 0.277504 0.0386421 [
[.21429 .265 -5 -15.6875 ]
[.21429 .265 5 0 ]
[.40476 .265 -5 -15.6875 ]
[.40476 .265 5 0 ]
[.59524 .265 -5 -15.6875 ]
[.59524 .265 5 0 ]
[.78571 .265 -5 -15.6875 ]
[.78571 .265 5 0 ]
[.97619 .265 -8 -15.6875 ]
[.97619 .265 8 0 ]
[.01131 .04565 -18.3125 -7.84375 ]
[.01131 .04565 0 7.84375 ]
[.01131 .12294 -18.3125 -7.84375 ]
[.01131 .12294 0 7.84375 ]
[.01131 .20022 -18.3125 -7.84375 ]
[.01131 .20022 0 7.84375 ]
[.01131 .35479 -10 -7.84375 ]
[.01131 .35479 0 7.84375 ]
[.01131 .43207 -10 -7.84375 ]
[.01131 .43207 0 7.84375 ]
[.01131 .50936 -10 -7.84375 ]
[.01131 .50936 0 7.84375 ]
[.01131 .58664 -10 -7.84375 ]
[.01131 .58664 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .2775 m
.21429 .28375 L
s
gsave
.21429 .265 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.40476 .2775 m
.40476 .28375 L
s
gsave
.40476 .265 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.59524 .2775 m
.59524 .28375 L
s
gsave
.59524 .265 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.78571 .2775 m
.78571 .28375 L
s
gsave
.78571 .265 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(8) show
1.000 setlinewidth
grestore
.97619 .2775 m
.97619 .28375 L
s
gsave
.97619 .265 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .2775 m
.07143 .28125 L
s
.11905 .2775 m
.11905 .28125 L
s
.16667 .2775 m
.16667 .28125 L
s
.2619 .2775 m
.2619 .28125 L
s
.30952 .2775 m
.30952 .28125 L
s
.35714 .2775 m
.35714 .28125 L
s
.45238 .2775 m
.45238 .28125 L
s
.5 .2775 m
.5 .28125 L
s
.54762 .2775 m
.54762 .28125 L
s
.64286 .2775 m
.64286 .28125 L
s
.69048 .2775 m
.69048 .28125 L
s
.7381 .2775 m
.7381 .28125 L
s
.83333 .2775 m
.83333 .28125 L
s
.88095 .2775 m
.88095 .28125 L
s
.92857 .2775 m
.92857 .28125 L
s
.25 Mabswid
0 .2775 m
1 .2775 L
s
.02381 .04565 m
.03006 .04565 L
s
gsave
.01131 .04565 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.02381 .12294 m
.03006 .12294 L
s
gsave
.01131 .12294 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.02381 .20022 m
.03006 .20022 L
s
gsave
.01131 .20022 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.02381 .35479 m
.03006 .35479 L
s
gsave
.01131 .35479 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.02381 .43207 m
.03006 .43207 L
s
gsave
.01131 .43207 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.02381 .50936 m
.03006 .50936 L
s
gsave
.01131 .50936 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.02381 .58664 m
.03006 .58664 L
s
gsave
.01131 .58664 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(8) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .06497 m
.02756 .06497 L
s
.02381 .08429 m
.02756 .08429 L
s
.02381 .10362 m
.02756 .10362 L
s
.02381 .14226 m
.02756 .14226 L
s
.02381 .16158 m
.02756 .16158 L
s
.02381 .1809 m
.02756 .1809 L
s
.02381 .21954 m
.02756 .21954 L
s
.02381 .23886 m
.02756 .23886 L
s
.02381 .25818 m
.02756 .25818 L
s
.02381 .29683 m
.02756 .29683 L
s
.02381 .31615 m
.02756 .31615 L
s
.02381 .33547 m
.02756 .33547 L
s
.02381 .37411 m
.02756 .37411 L
s
.02381 .39343 m
.02756 .39343 L
s
.02381 .41275 m
.02756 .41275 L
s
.02381 .45139 m
.02756 .45139 L
s
.02381 .47071 m
.02756 .47071 L
s
.02381 .49004 m
.02756 .49004 L
s
.02381 .52868 m
.02756 .52868 L
s
.02381 .548 m
.02756 .548 L
s
.02381 .56732 m
.02756 .56732 L
s
.02381 .02633 m
.02756 .02633 L
s
.02381 .00701 m
.02756 .00701 L
s
.02381 .60596 m
.02756 .60596 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .31615 m
.06244 .30047 L
.10458 .28337 L
.14415 .26732 L
.18221 .25188 L
.22272 .23544 L
.26171 .21962 L
.30316 .2028 L
.34309 .1866 L
.3815 .17102 L
.42237 .15444 L
.46172 .13847 L
.49955 .12312 L
.53984 .10677 L
.57861 .09104 L
.61984 .07431 L
.65954 .0582 L
.69774 .04271 L
.73838 .02621 L
.77751 .01034 L
s
.77751 .01034 m
.80299 0 L
s
.02381 .31615 m
.06244 .28797 L
.08255 .27583 L
.10458 .2645 L
.1253 .25573 L
.14415 .24934 L
.16372 .24431 L
.17449 .24224 L
.18466 .24073 L
.19436 .23971 L
.19925 .23934 L
.20206 .23918 L
.20463 .23906 L
.20699 .23898 L
.20922 .23892 L
.21044 .23889 L
.21158 .23888 L
.21283 .23887 L
.21352 .23886 L
.21416 .23886 L
.2153 .23886 L
.21636 .23887 L
.21752 .23888 L
.21875 .2389 L
.21981 .23893 L
.22081 .23895 L
.22307 .23903 L
.22567 .23914 L
.22808 .23927 L
.23353 .23965 L
.23833 .24009 L
.24343 .24067 L
.25225 .24193 L
.26198 .24371 L
.28133 .24844 L
.30219 .25532 L
.34333 .27433 L
.38295 .29946 L
.42106 .32994 L
.46162 .36917 L
.50066 .41356 L
.54216 .46786 L
.58214 .52711 L
.62061 .59055 L
s
.62061 .59055 m
.63573 .61803 L
s
.02381 .31615 m
.04262 .29546 L
.06244 .27823 L
.08255 .26518 L
.09396 .25963 L
.10458 .25559 L
.10961 .25404 L
.11508 .25262 L
.11989 .25157 L
.125 .25069 L
.12959 .25007 L
.13203 .24981 L
.13466 .24959 L
.13706 .24943 L
.13926 .24932 L
.14032 .24928 L
.14147 .24925 L
.14256 .24923 L
.14357 .24922 L
.14486 .24922 L
.14551 .24922 L
.14622 .24923 L
.14747 .24925 L
.14865 .24928 L
.15004 .24933 L
.1513 .24938 L
.15414 .24955 L
.15659 .24974 L
.15925 .24999 L
.1641 .25057 L
.17376 .25218 L
.1827 .25417 L
.20137 .25975 L
.22197 .26786 L
.26217 .28812 L
.30086 .31088 L
.342 .33566 L
.38162 .35723 L
.40159 .36636 L
.41973 .37324 L
.42914 .37619 L
.43934 .37884 L
.44495 .38005 L
.45012 .38099 L
.45502 .38173 L
.46029 .38236 L
.46289 .38259 L
.46536 .38278 L
.46757 .3829 L
.46884 .38296 L
Mistroke
.46999 .383 L
.47116 .38304 L
.47227 .38306 L
.47343 .38307 L
.47469 .38308 L
.4758 .38307 L
.47697 .38305 L
.47796 .38303 L
.47906 .383 L
.48044 .38294 L
.48176 .38288 L
.48421 .38272 L
.48686 .3825 L
.48974 .38219 L
.49491 .38147 L
.4997 .38061 L
.50935 .37827 L
.5184 .37533 L
.52833 .37122 L
.53901 .36571 L
.55909 .35215 L
.58021 .33302 L
.61837 .28448 L
.63802 .2518 L
.65898 .21066 L
.70053 .10843 L
Mfstroke
.70053 .10843 m
.7347 0 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00803oool00`000000oooo0?ooo`3o0?oooal0oooo00<000000?ooo`3o
ool05`3oool2000005<0oooo000P0?ooo`030000003oool0oooo0?l0oooo7`3oool00`000000oooo
0?ooo`0E0?ooo`800000E@3oool00200oooo00<000000?ooo`3oool0o`3ooolN0?ooo`030000003o
ool0oooo01<0oooo0`00001G0?ooo`00803oool300000?l0oooo7P3oool00`000000oooo0?ooo`0A
0?ooo`800000FP3oool00200oooo00<000000?ooo`3oool0o`3ooolN0?ooo`030000003oool0oooo
00h0oooo0`00001L0?ooo`00803oool00`000000oooo0?ooo`3o0?oooad0oooo00<000000?ooo`3o
ool03@3oool2000005l0oooo000P0?ooo`030000003oool0oooo0?l0oooo7@3oool00`000000oooo
0?ooo`0;0?ooo`800000H@3oool00200oooo00<000000?ooo`3oool0o`3ooolM0?ooo`030000003o
ool0oooo00T0oooo0P00001S0?ooo`00803oool00`000000oooo0?ooo`3o0?oooa`0oooo00<00000
0?ooo`3oool01`3oool3000006D0oooo000P0?ooo`030000003oool0oooo0?l0oooo703oool00`00
0000oooo0?ooo`050?ooo`800000J03oool00200oooo00<000000?ooo`3oool0o`3ooolL0?ooo`03
0000003oool0oooo00<0oooo0P00001Z0?ooo`00803oool300000?l0oooo6`3oool010000000oooo
0?ooo`3oool3000006`0oooo000P0?ooo`030000003oool0oooo0?l0oooo6`3oool010000000oooo
00000000001_0?ooo`00803oool00`000000oooo0?ooo`3o0?oooaX0oooo0`00001a0?ooo`004@3o
ool01@3ogi`0B000000000000000M;oo00X0oooo00<000000?ooo`3oool0o`3ooolG0?ooo`<00000
00<0oooo0000003oool0L@3oool00100oooo00L0omnL04P00018W=l0oooo0?oo_`1d0000M;oo00T0
oooo00<000000?ooo`3oool0o`3ooolE0?ooo`8000000`3oool00`000000oooo0?ooo`1a0?ooo`00
3`3oool00`3ooml0W4P0001d_`030?ooo`030?noM000B9`0gooo00P0oooo00<000000?ooo`3oool0
o`3ooolB0?ooo`<000001@3oool00`000000oooo0?ooo`1a0?ooo`003`3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00P0oooo00<000000?ooo`3oool0o`3oool?0?ooo`<00000203o
ool00`000000oooo0?ooo`1a0?ooo`000`3oool00`3ooml0W4P0000000070000000504RLg`3oool0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`080?ooo`030000003oool0oooo0?l0oooo3@3o
ool2000000X0oooo00<000000?ooo`3oool0LP3oool000l0oooo00<0omnL04P0001d_ol00`3oool0
0`2oM000001809cOo`080?ooo`@00000o`3oool:0?ooo`800000303oool00`000000oooo0?ooo`1b
0?ooo`003`3oool0203ookl0M00000000018W=l0oooo0?noM000000007Bo2@3oool00`000000oooo
0?ooo`3o0?ooo`T0oooo0P00000>0?ooo`030000003oool0oooo0780oooo000@0?ooo`030;md0000
B9`0_ia80080000000<004RL0=ooo`3oool0203oool00`000000oooo0?ooo`3o0?ooo`H0oooo0`00
000?0?ooo`030000003oool0oooo07<0oooo000@0?ooo`030?oOW0180000M;oo00d0oooo00<00000
0?ooo`3oool0o`3oool40?ooo`8000004P3oool00`000000oooo0?ooo`1c0?ooo`004@3oool00`3O
W4P0001809cOo`0<0?ooo`030000003oool0oooo0?l0oooo0P3oool2000001@0oooo00<000000?oo
o`3oool0L`3oool00180oooo00<0okmd00000000M;l02`3oool00`000000oooo0?ooo`3n0?ooo`<0
00005@3oool00`000000oooo0?ooo`1d0?ooo`004`3oool01@3ooml0W4P00000000007@0_ooo00P0
oooo00<000000?ooo`3oool0o03oool2000001P0oooo00<000000?ooo`3oool0M03oool00200oooo
00<000000?ooo`3oool0n@3oool3000001X0oooo00<000000?ooo`3oool0M03oool00200oooo0`00
003f0?ooo`<00000703oool00`000000oooo0?ooo`1e0?ooo`00803oool00`000000oooo0?ooo`3d
0?ooo`8000007`3oool00`000000oooo0?ooo`1e0?ooo`00803oool00`000000oooo0?ooo`3a0?oo
o`<000008@3oool00`000000oooo0?ooo`1e0?ooo`00803oool00`000000oooo0?ooo`3^0?ooo`<0
00008`3oool00`000000oooo0?ooo`1f0?ooo`00803oool00`000000oooo0?ooo`3/0?ooo`800000
9P3oool00`000000oooo0?ooo`1f0?ooo`00803oool00`000000oooo0?ooo`3Y0?ooo`<00000:03o
ool00`000000oooo0?ooo`1f0?ooo`00803oool00`000000oooo0?ooo`3W0?ooo`800000:P3oool0
0`000000oooo0?ooo`1g0?ooo`00803oool300000>D0oooo0P00000/0?ooo`030000003oool0oooo
07L0oooo000P0?ooo`030000003oool0oooo0>80oooo0`00000^0?ooo`030000003oool0oooo07L0
oooo000P0?ooo`030000003oool0oooo0>00oooo0P00000a0?ooo`030000003oool0oooo07L0oooo
000P0?ooo`030000003oool0oooo0=d0oooo0`00000b0?ooo`030000003oool0oooo07P0oooo000P
0?ooo`030000003oool0oooo0=/0oooo0P00000e0?ooo`030000003oool0oooo07P0oooo000P0?oo
o`030000003oool0oooo0=P0oooo0`00000g0?ooo`030000003oool0oooo07P0oooo000P0?ooo`03
0000003oool0oooo0=D0oooo0`00000i0?ooo`030000003oool0oooo07T0oooo000P0?ooo`030000
003oool0oooo0=<0oooo0P00000l0?ooo`030000003oool0oooo07T0oooo000P0?ooo`<00000d03o
ool3000003h0oooo00<000000?ooo`3oool0N@3oool00200oooo00<000000?ooo`3oool0c@3oool3
00000400oooo00<000000?ooo`3oool0NP3oool00200oooo00<000000?ooo`3oool0b`3oool20000
04<0oooo00<000000?ooo`3oool0NP3oool001<0oooo00@0oooO09a80000B9`0gooo2@3oool00`00
0000oooo0?ooo`390?ooo`800000A@3oool00`000000oooo0?ooo`1j0?ooo`004`3oool0103ooml0
W4P00018W03Oool90?ooo`030000003oool0oooo0<L0oooo0P0000160?ooo`030000003oool0oooo
07/0oooo000C0?ooo`040?oog`2LB00004RL0=ooo`T0oooo00<000000?ooo`3oool0a03oool30000
04P0oooo00<000000?ooo`3oool0N`3oool000l0oooo00<0_g@00000000000001@0000000`1d_ol0
oooo0?ooo`060?ooo`030000003oool0oooo0<80oooo0P00001:0?ooo`030000003oool0oooo07`0
oooo00030?ooo`030?oog`2LB000000000L0000000@0B9cO0?ooo`3ogi`0M9cO0P3oool0103ooml0
W4P00018W03Oool90?ooo`030000003oool0oooo0<00oooo0P00001<0?ooo`030000003oool0oooo
07`0oooo000@0?ooo`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`090?ooo`@00000_03oool3
000004h0oooo00<000000?ooo`3oool0O03oool00100oooo00L0oooO09a8M02oool0oooO09a80000
B9`0gooo00T0oooo00<000000?ooo`3oool0^`3oool200000500oooo00<000000?ooo`3oool0O@3o
ool00140oooo00H0omnL04Qd_`3ooml0W4P00018W03Oool90?ooo`030000003oool0oooo0;P0oooo
0`00001B0?ooo`030000003oool0oooo07d0oooo000B0?ooo`050=nLB01d_ml0W4P00018W03Oool0
2@3oool00`000000oooo0?ooo`2f0?ooo`800000E03oool00`000000oooo0?ooo`1n0?ooo`004P3o
ool01@3ooml0W4Q804Q80000B9`0gooo00T0oooo00<000000?ooo`3oool0/`3oool3000005H0oooo
00<000000?ooo`3oool0OP3oool001<0oooo00@0omnL04P00000B9`0gooo2@3oool00`000000oooo
0?ooo`2a0?ooo`800000F@3oool00`000000oooo0?ooo`1n0?ooo`00503oool00`3OW4P004RL0=oo
o`090?ooo`030000003oool0oooo0:h0oooo0`00001J0?ooo`030000003oool0oooo07l0oooo000P
0?ooo`030000003oool0oooo0:`0oooo0P00001M0?ooo`030000003oool0oooo07l0oooo000P0?oo
o`<00000ZP3oool2000005h0oooo00<000000?ooo`3oool0P03oool00200oooo00<000000?ooo`3o
ool0Z03oool200000600oooo00<000000?ooo`3oool0P03oool00200oooo00<000000?ooo`3oool0
Y@3oool300000680oooo00<000000?ooo`3oool0P03oool00200oooo00<000000?ooo`3oool0X`3o
ool2000006@0oooo00<000000?ooo`3oool0P@3oool00200oooo00<000000?ooo`3oool0X@3oool2
000006H0oooo00<000000?ooo`3oool0P@3oool00200oooo00<000000?ooo`3oool0WP3oool30000
06L0oooo00<000000?ooo`3oool0PP3oool00200oooo00<000000?ooo`3oool0W03oool2000006X0
oooo00<000000?ooo`3oool0PP3oool00200oooo00<000000?ooo`3oool0V@3oool3000006/0oooo
00<000000?ooo`3oool0P`3oool00200oooo0`00002G0?ooo`800000KP3oool00`000000oooo0?oo
o`230?ooo`00803oool00`000000oooo0?ooo`2E0?ooo`800000L03oool00`000000oooo0?ooo`23
0?ooo`00803oool00`000000oooo0?ooo`2B0?ooo`<00000L@3oool00`000000oooo0?ooo`240?oo
o`00803oool00`000000oooo0?ooo`2@0?ooo`800000M03oool00`000000oooo0?ooo`240?ooo`00
803oool00`000000oooo0?ooo`2=0?ooo`<00000M@3oool00`000000oooo0?ooo`250?ooo`00803o
ool00`000000oooo0?ooo`2;0?ooo`800000N03oool00`000000oooo0?ooo`250?ooo`00803oool0
0`000000oooo0?ooo`280?ooo`<00000NP3oool00`000000oooo0?ooo`250?ooo`00803oool00`00
0000oooo0?ooo`260?ooo`800000O03oool00`000000oooo0?ooo`260?ooo`00803oool3000008<0
oooo0`00001n0?ooo`030000003oool0oooo08H0oooo000P0?ooo`030000003oool0oooo0840oooo
0P0000200?ooo`030000003oool0oooo08L0oooo000?0?ooo`030;md00000000000000@0000000<0
B9cO0?ooo`3oool01`3oool00`000000oooo0?ooo`1n0?ooo`<00000PP3oool00`000000oooo0?oo
o`270?ooo`00403oool00`3OW4P0000000000003000000030018W03Oool0oooo00L0oooo00<00000
0?ooo`3oool0N`3oool3000008D0oooo00<000000?ooo`3oool0Q`3oool00140oooo00<0gia804RL
g`3oool00P3oool00`3OW7@0W=oo0?ooo`070?ooo`030000003oool0oooo07P0oooo0`0000270?oo
o`030000003oool0oooo08P0oooo000B0?ooo`030=nLB01d_ol0oooo00/0oooo00<000000?ooo`3o
ool0MP3oool2000008X0oooo00<000000?ooo`3oool0R03oool000<0oooo00<0oooO09a800000000
1`0000000`18W=l0oooo0?ooo`030?ooo`030;mdB02Lgol0oooo00X0oooo00<000000?ooo`3oool0
L`3oool3000008/0oooo00<000000?ooo`3oool0R@3oool001<0oooo00<0oooO09a8M02oool02P3o
ool4000006l0oooo0`00002>0?ooo`030000003oool0oooo08T0oooo000D0?ooo`030?noM000M;l0
oooo00T0oooo00<000000?ooo`3oool0K@3oool300000940oooo00<000000?ooo`3oool0R@3oool0
01@0oooo00<0oono07@0B02Lgol02@3oool00`000000oooo0?ooo`1[0?ooo`800000T`3oool00`00
0000oooo0?ooo`2:0?ooo`00503oool00`3ookl0M00004RLg`090?ooo`030000003oool0oooo06T0
oooo0P00002E0?ooo`030000003oool0oooo08X0oooo000?0?ooo`030?noW02Lgol0oooo0080oooo
00<0omnL04P00018W=l02@3oool00`000000oooo0?ooo`1W0?ooo`800000UP3oool00`000000oooo
0?ooo`2;0?ooo`003`3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo2@3oool00`00
0000oooo0?ooo`1T0?ooo`<00000V03oool00`000000oooo0?ooo`2;0?ooo`00403oool01P3ooml0
W4P0000000000000001d0;ooo`X0oooo00<000000?ooo`3oool0HP3oool2000009X0oooo00<00000
0?ooo`3oool0S03oool00200oooo00<000000?ooo`3oool0H03oool2000009`0oooo00<000000?oo
o`3oool0S03oool00200oooo0`00001M0?ooo`<00000W@3oool00`000000oooo0?ooo`2=0?ooo`00
803oool00`000000oooo0?ooo`1K0?ooo`800000X03oool00`000000oooo0?ooo`2=0?ooo`00803o
ool00`000000oooo0?ooo`160?ooo`030;md00000000000000@0000000<0B9cO0?ooo`3oool02@3o
ool2000003`0oooo00@0oooO09a80000B9`0goooA`3oool01@3ogi`0B000000000000000M;oo01D0
oooo00<000000?ooo`3oool0<@3oool01@3OW4P00000000000000000B9cO04<0oooo00H0_g@00000
0000000000000018W03Oool30?ooo`050?oOW018000000000000M02oool01`3oool00200oooo00<0
00000?ooo`3oool0A`3oool00`3OW4P0000000000003000000030018W03Oool0oooo00L0oooo0P00
000n0?ooo`040?oog`2LB00004RL0=ooodH0oooo00L0omnL04P00018W=l0oooo0?oo_`1d0000M;oo
01@0oooo00<000000?ooo`3oool0<03oool01`3OW4P0000007Boo`3oool0oooO09a80018W=l0@`3o
ool00`3ogi`0B00007Boo`040?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`060?ooo`00
803oool00`000000oooo0?ooo`180?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0
103oool300000400oooo00@0oooO09a80000B9`0goooA@3oool00`3ooml0W4P0001d_`030?ooo`03
0?noM000B9`0gooo0180oooo00<000000?ooo`3oool0<03oool00`3ooml0W4P0001d_`030?ooo`03
0?noM000B9`0gooo0480oooo00<0omnL04P0001d_ol0103oool00`2oM00004RL0=ooo`020?ooo`03
0;md0000M;l0oooo00D0oooo000P0?ooo`030000003oool0oooo04T0oooo00<0gia807Boo`3oool0
1P3oool2000003l0oooo00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`120?ooo`030?oO
W0180000M;oo00<0oooo00<0okmd0000B02Lgol04@3oool00`000000oooo0?ooo`0b0?ooo`030;md
0000M;l0oooo0080oooo00<0_g@00018W03Oool0@P3oool00`3ogi`0B00007Boo`030?ooo`030?oo
_`1d0000B9cO00<0oooo00<0gia80000M02oool01@3oool00200oooo00<000000?ooo`3oool0BP3o
ool00`2oM4P0W=oo0?ooo`030?ooo`800000@@3oool0203ogi`0M9cO0?ooo`3oool0oooO09a80000
B9`0goooA@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo0140oooo00<000000?oo
o`3oool0<P3oool01`3ogi`0B01d0;ooo`3oool0_g@00000001d_ol0@`3oool00`3ogi`0B00007Bo
o`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01@3oool00200oooo00<00000
0?ooo`3oool0BP3oool00`3ooml0W4Qd0;ooo`03000004@0oooo00L0gia807Boo`3oool0oooO09a8
0000B9`0gooo04D0oooo00<0omnL04P0001d_ol00`3oool00`2oM000001809cOo`0@0?ooo`030000
003oool0oooo03@0oooo00D0oono07A8B0180000000004RLg`140?ooo`030?oOW0180000M;oo00<0
oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`050?ooo`00803oool300000440oooo
4P0000110?ooo`070?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`150?ooo`080?oo_`1d00000000
04RLg`3oool0okmd00000000M;lA0?ooo`030000003oool0oooo03@0oooo00@0omnL04P000000000
07BoA@3oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol01@3oool00200oooo00<000000?ooo`3oool0>@3oool8000000H0oooo0`000000103oool0oono
07@0B02Lgol50?ooo`T00000>@3oool01P3ogi`0B7Bo0?oog`2LB00004RL0=ooodH0oooo00<0_g@0
0018W02oW4P00P0000000`00B9`0gooo0?ooo`0?0?ooo`030000003oool0oooo03@0oooo00H0omnL
04P0000007@0_ooo0?noM000M;m40?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol0
0`3oool00`3o_g@0001809cOo`050?ooo`00803oool00`000000oooo0?ooo`0e0?ooo`@00000303o
ool2000000@0oooo00<0oono07@00018W=l03P3oool4000003H0oooo00D0gia807Bog`2LB00004RL
0=ooo`160?ooo`030?oOW0180000M;oo01@0oooo00<000000?ooo`3oool0=03oool00`2oM000001d
0;ooo`020?ooo`030;md0018W=l0oooo0480oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M000
04RLg`030?ooo`030=nLB00007@0_ooo00D0oooo000P0?ooo`030000003oool0oooo0340oooo1000
000=0?ooo`<0000000<0oooo0?noW02Lgol00`3oool00`3ogi`0B00004RLg`0B0?ooo`@00000<P3o
ool01@3ooml0W4Q804Q80000B9`0gooo04L0oooo00<0gia80000B02Lgol04P3oool00`000000oooo
0?ooo`0e0?ooo`030;md0000B9`0gooo0080oooo00<0gia8001d_`3oool0@P3oool00`3ogi`0B000
07Boo`040?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01@3oool00200oooo00<0
00000?ooo`3oool0:P3oool9000000`0oooo0`0000040?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0
M0000000M02ooolF0?ooo`@00000;`3oool0103ogi`0B0000018W03Ooom80?ooo`030?noM0000000
07Bo0140oooo00<000000?ooo`3oool0=@3oool01`3o_g@0000007Boo`3oool0oono07@0001d_ol0
@@3oool01@3ooml0W4P0000000000000M;oo00@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0
_ooo00H0oooo000P0?ooo`030000003oool0oooo02@0oooo1P0000000`3oool00000000000060?oo
o``00000203oool01P3ooml0W4P0000000000000001d0;oooa/0oooo0P00000^0?ooo`030=nLB000
B9`0gooo04T0oooo00D0oooO09a800000000001d0;ooo`0=0?ooo`030000003oool0oooo03L0oooo
00D0okmd000000000000000004RLg`140?ooo`030?noM0000000M;oo00D0oooo00D0oono07@00000
0000001d0;ooo`070?ooo`00803oool00`000000oooo0?ooo`0P0?ooo`@00000103oool3000000l0
oooo1P00000Z0?ooo`<00000R03oool00`000000oooo0?ooo`2E0?ooo`00803oool00`000000oooo
0?ooo`0N0?ooo`8000001@3oool300000100oooo0P0000060?ooo`800000:`3oool3000008D0oooo
00<000000?ooo`3oool0U@3oool00200oooo0`00000L0?ooo`8000001@3oool200000140oooo0P00
000:0?ooo`800000;03oool200000880oooo00<000000?ooo`3oool0UP3oool00200oooo00<00000
0?ooo`3oool06@3oool3000000@0oooo0`00000A0?ooo`8000003P3oool3000002/0oooo0`00001o
0?ooo`030000003oool0oooo09H0oooo000P0?ooo`030000003oool0oooo01L0oooo0P0000050?oo
o`8000004P3oool2000001<0oooo0P00000/0?ooo`800000O03oool00`000000oooo0?ooo`2G0?oo
o`00803oool00`000000oooo0?ooo`0E0?ooo`8000001@3oool200000180oooo0P00000G0?ooo`<0
0000:`3oool2000007X0oooo00<000000?ooo`3oool0U`3oool00200oooo00<000000?ooo`3oool0
4`3oool2000000D0oooo0P00000A0?ooo`<00000703oool3000002X0oooo0`00001f0?ooo`030000
003oool0oooo09P0oooo000P0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool00`3o
ool200000140oooo0P00000R0?ooo`800000:`3oool2000007<0oooo00<000000?ooo`3oool0V@3o
ool00200oooo00<000000?ooo`3oool0403oool2000000@0oooo0P00000@0?ooo`<000009P3oool2
000002/0oooo0P00001a0?ooo`030000003oool0oooo09T0oooo000P0?ooo`030000003oool0oooo
00h0oooo0P0000050?ooo`030000003oool0oooo00d0oooo0`00000[0?ooo`800000:`3oool20000
06h0oooo00<000000?ooo`3oool0VP3oool001H0ooooo`00002H000000<0oooo000P0?ooo`030000
003oool0oooo00`0oooo00L000000?ooo`3oool0oooo0000003oool0000000d0oooo0`0000000`3o
ool000000?ooo`0A0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool02P3oool20000
00D0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0?0?ooo`8000004`3oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0403oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0403oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo00/0oooo000P0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool00P3oool2000000/0oooo0`00000Z0?ooo`030000003oool0oooo00`0
oooo0P00000Z0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0BP3oool00`000000
oooo0?ooo`070?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool0BP3oool00`000000
oooo0?ooo`0;0?ooo`00803oool00`000000oooo0?ooo`0:0?ooo`040000003oool0oooo0?ooo`80
00002`3oool2000003h0oooo0P00000Y0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3o
ool0W03oool00200oooo00<000000?ooo`3oool0203oool2000000<0oooo00<000000?ooo`3oool0
203oool300000480oooo0P00000X0?ooo`800000I03oool00`000000oooo0?ooo`2M0?ooo`00803o
ool00`000000oooo0?ooo`070?ooo`040000003oool0oooo0?ooo`8000002@3oool2000004L0oooo
0P00000X0?ooo`030000003oool0oooo0600oooo00<000000?ooo`3oool0WP3oool00200oooo00<0
00000?ooo`3oool01P3oool01@000000oooo0?ooo`3oool0000000T0oooo0P00001;0?ooo`800000
9`3oool200000600oooo00<000000?ooo`3oool0WP3oool00200oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`02000000L0oooo0`00001?0?ooo`030000003oool0oooo02H0oooo00<0
00000?ooo`3oool0G03oool00`000000oooo0?ooo`2O0?ooo`00803oool3000000@0oooo00@00000
0?ooo`3oool000001`3oool2000005<0oooo0P00000W0?ooo`800000F`3oool00`000000oooo0?oo
o`2P0?ooo`00803oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo000000D0oooo0`00
001G0?ooo`8000009`3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo0:40oooo000P
0?ooo`030000003oool0oooo0080oooo00@000000?ooo`0000000000103oool2000005`0oooo00<0
00000?ooo`3oool09@3oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo0:80oooo000P
0?ooo`070000003oool0oooo0?ooo`000000oooo000000030?ooo`<00000G`3oool2000002H0oooo
0P00001E0?ooo`030000003oool0oooo0:80oooo000P0?ooo`030000003oool0oooo00<0000000@0
oooo0000000000000000I03oool2000002H0oooo00<000000?ooo`3oool0D@3oool00`000000oooo
0?ooo`2S0?ooo`00803oool01@000000oooo000000000000oooo00800000J@3oool00`000000oooo
0?ooo`0T0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0Y03oool00200oooo1@00
001/0?ooo`8000009@3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo0:D0oooo000P
0?ooo`<00000L03oool2000002@0oooo0P00001=0?ooo`030000003oool0oooo0:D0oooo000P0?oo
o`030000003oool0oooo0780oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`190?oo
o`030000003oool0oooo0:H0oooo000P0?ooo`030000003oool0oooo07<0oooo0P00000T0?ooo`03
0000003oool0oooo04L0oooo00<000000?ooo`3oool0Y`3oool00200oooo00<000000?ooo`3oool0
M@3oool2000002<0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`2X0?ooo`00803o
ool00`000000oooo0?ooo`1g0?ooo`8000008P3oool2000004@0oooo00<000000?ooo`3oool0Z@3o
ool00200oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0
oooo0440oooo00<000000?ooo`3oool0Z@3oool00200oooo00<000000?ooo`3oool0NP3oool20000
0280oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`2Z0?ooo`00803oool00`000000
oooo0?ooo`1l0?ooo`8000008@3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo0:/0
oooo000P0?ooo`<00000OP3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo03/0oooo
00<000000?ooo`3oool0[03oool00200oooo00<000000?ooo`3oool0O`3oool200000200oooo00<0
00000?ooo`3oool0>@3oool00`000000oooo0?ooo`2]0?ooo`00803oool00`000000oooo0?ooo`21
0?ooo`8000007`3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo0:h0oooo000?0?oo
o`030;md00000000000000@0000000<0B9cO0?ooo`3oool01`3oool00`000000oooo0?ooo`230?oo
o`8000007P3oool00`000000oooo0?ooo`0d0?ooo`800000/@3oool00100oooo00<0gia800000000
00000`0000000`00B9`0gooo0?ooo`070?ooo`030000003oool0oooo08D0oooo00<000000?ooo`3o
ool0703oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0;40oooo000A0?ooo`030=nL
B018W=l0oooo0080oooo00<0giad09cOo`3oool01`3oool00`000000oooo0?ooo`260?ooo`800000
7@3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo0;80oooo000B0?ooo`030=nLB01d
_ol0oooo00/0oooo00<000000?ooo`3oool0R03oool2000001`0oooo0P00000_0?ooo`030000003o
ool0oooo0;<0oooo000C0?ooo`030;mdB02Lgol0oooo00X0oooo00<000000?ooo`3oool0RP3oool2
000001`0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`2d0?ooo`004`3oool00`3o
oml0W4Qd0;ooo`0:0?ooo`@00000R`3oool2000001/0oooo00<000000?ooo`3oool0:@3oool00`00
0000oooo0?ooo`2e0?ooo`00503oool00`3o_g@007Bo0?ooo`090?ooo`030000003oool0oooo08h0
oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0V0?ooo`800000^03oool001@0oooo
00<0oono07@0B02Lgol02@3oool00`000000oooo0?ooo`2?0?ooo`8000006P3oool00`000000oooo
0?ooo`0T0?ooo`030000003oool0oooo0;P0oooo000D0?ooo`030?oo_`1d0000B9cO00T0oooo00<0
00000?ooo`3oool0T@3oool2000001T0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?oo
o`2i0?ooo`003`3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO00T0oooo00<00000
0?ooo`3oool0T`3oool2000001P0oooo00<000000?ooo`3oool07`3oool200000;`0oooo000?0?oo
o`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool90?ooo`030000003oool0oooo09D0oooo
0`00000F0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0_03oool00100oooo00H0
oooO09a80000000000000000M02oool:0?ooo`030000003oool0oooo09P0oooo0P00000E0?ooo`03
0000003oool0oooo01X0oooo0P00002o0?ooo`00803oool00`000000oooo0?ooo`2J0?ooo`<00000
4`3oool00`000000oooo0?ooo`0F0?ooo`<00000`@3oool00200oooo0`00002M0?ooo`<000004@3o
ool00`000000oooo0?ooo`0B0?ooo`<00000a03oool00200oooo00<000000?ooo`3oool0X03oool4
000000h0oooo00<000000?ooo`3oool03`3oool200000<L0oooo000P0?ooo`030000003oool0oooo
0:@0oooo1@00000:0?ooo`030000003oool0oooo00/0oooo0`0000390?ooo`00803oool00`000000
oooo0?ooo`2Y0?ooo`P000000P3oool00`000000oooo0?ooo`020?ooo`T00000c03oool00200oooo
00<000000?ooo`3oool0/@3oool700000=D0oooo000P0?ooo`030000003oool0oooo0;D0oooo00<0
00000?ooo`3oool0e@3oool00200oooo00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`3D
0?ooo`00803oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo0=<0oooo000P0?ooo`<0
0000^03oool00`000000oooo0?ooo`3B0?ooo`00803oool00`000000oooo0?ooo`2i0?ooo`030000
003oool0oooo0=40oooo000P0?ooo`030000003oool0oooo0;X0oooo00<000000?ooo`3oool0d03o
ool00200oooo00<000000?ooo`3oool0^`3oool00`000000oooo0?ooo`3?0?ooo`00803oool00`00
0000oooo0?ooo`2k0?ooo`030000003oool0oooo0<l0oooo000P0?ooo`030000003oool0oooo0;`0
oooo00<000000?ooo`3oool0cP3oool00200oooo00<000000?ooo`3oool0_@3oool00`000000oooo
0?ooo`3=0?ooo`00803oool00`000000oooo0?ooo`2n0?ooo`030000003oool0oooo0<`0oooo000P
0?ooo`<00000_`3oool00`000000oooo0?ooo`3;0?ooo`00803oool00`000000oooo0?ooo`300?oo
o`030000003oool0oooo0<X0oooo000P0?ooo`030000003oool0oooo0<00oooo00<000000?ooo`3o
ool0bP3oool001<0oooo00@0oooO09a80000B9`0gooo2@3oool00`000000oooo0?ooo`310?ooo`03
0000003oool0oooo0<T0oooo000C0?ooo`040?oog`2LB00004RL0=ooo`T0oooo00<000000?ooo`3o
ool0`P3oool00`000000oooo0?ooo`380?ooo`004`3oool0103ooml0W4P00018W03Oool90?ooo`03
0000003oool0oooo0<<0oooo00<000000?ooo`3oool0a`3oool000l0oooo00<0_g@0000000000000
1@0000000`1d_ol0oooo0?ooo`060?ooo`030000003oool0oooo0<<0oooo00<000000?ooo`3oool0
a`3oool000l0oooo00P0omnL07BLg`3oool0oooo0?oog`2LB00004RL0=ooo`T0oooo100000330?oo
o`030000003oool0oooo0<H0oooo000@0?ooo`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`09
0?ooo`030000003oool0oooo0<D0oooo00<000000?ooo`3oool0a@3oool00100oooo00L0oooO09a8
M02oool0oooO09a80000B9`0gooo00T0oooo00<000000?ooo`3oool0aP3oool00`000000oooo0?oo
o`340?ooo`004@3oool01P3ogi`0B7Bo0?oog`2LB00004RL0=ooo`T0oooo00<000000?ooo`3oool0
aP3oool00`000000oooo0?ooo`340?ooo`004P3oool01@3OW4P0M;oO09a80000B9`0gooo00T0oooo
00<000000?ooo`3oool0a`3oool00`000000oooo0?ooo`330?ooo`004P3oool01@3ooml0W4Q804Q8
0000B9`0gooo00T0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`320?ooo`004`3o
ool0103ogi`0B0000018W03Oool90?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3oool0
`P3oool001@0oooo00<0gia80018W03Oool02@3oool00`000000oooo0?ooo`390?ooo`030000003o
ool0oooo0<40oooo000P0?ooo`<00000bP3oool00`000000oooo0?ooo`300?ooo`00803oool00`00
0000oooo0?ooo`3;0?ooo`030000003oool0oooo0;l0oooo000P0?ooo`030000003oool0oooo0</0
oooo00<000000?ooo`3oool0_`3oool00200oooo00<000000?ooo`3oool0c03oool00`000000oooo
0?ooo`2n0?ooo`00803oool00`000000oooo0?ooo`3=0?ooo`030000003oool0oooo0;d0oooo000P
0?ooo`030000003oool0oooo0<h0oooo00<000000?ooo`3oool0_03oool00200oooo00<000000?oo
o`3oool0cP3oool00`000000oooo0?ooo`2l0?ooo`00803oool00`000000oooo0?ooo`3?0?ooo`03
0000003oool0oooo0;/0oooo000P0?ooo`<00000d03oool00`000000oooo0?ooo`2j0?ooo`00803o
ool00`000000oooo0?ooo`3@0?ooo`030000003oool0oooo0;X0oooo000P0?ooo`030000003oool0
oooo0=40oooo00<000000?ooo`3oool0^@3oool00200oooo00<000000?ooo`3oool0dP3oool00`00
0000oooo0?ooo`2h0?ooo`00803oool00`000000oooo0?ooo`3C0?ooo`030000003oool0oooo0;L0
oooo000P0?ooo`030000003oool0oooo0=<0oooo00<000000?ooo`3oool0]`3oool00200oooo00<0
00000?ooo`3oool0e03oool00`000000oooo0?ooo`2f0?ooo`00803oool00`000000oooo0?ooo`3E
0?ooo`030000003oool0oooo0;D0oooo000P0?ooo`<00000e@3oool00`000000oooo0?ooo`2e0?oo
o`00803oool00`000000oooo0?ooo`3F0?ooo`030000003oool0oooo0;@0oooo000P0?ooo`030000
003oool0oooo0=L0oooo00<000000?ooo`3oool0/`3oool00140oooo00D0omnL04P0000000000000
07Boo`0:0?ooo`030000003oool0oooo0=L0oooo00<000000?ooo`3oool0/`3oool00100oooo00L0
omnL04P00018W=l0oooo0?oo_`1d0000M;oo00T0oooo00<000000?ooo`3oool0f03oool00`000000
oooo0?ooo`2b0?ooo`003`3oool00`3ooml0W4P0001d_`030?ooo`030?noM000B9`0gooo00P0oooo
00<000000?ooo`3oool0f@3oool00`000000oooo0?ooo`2a0?ooo`003`3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00P0oooo00<000000?ooo`3oool0fP3oool00`000000oooo0?oo
o`2`0?ooo`003`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00P0oooo00<00000
0?ooo`3oool0fP3oool00`000000oooo0?ooo`2`0?ooo`003`3oool00`3ogi`0B00007Boo`030?oo
o`030;md000004P0W=oo00P0oooo1000003J0?ooo`030000003oool0oooo0:l0oooo000?0?ooo`08
0?oo_`1d0000000004RLg`3oool0okmd00000000M;l90?ooo`030000003oool0oooo0=`0oooo00<0
00000?ooo`3oool0[P3oool00100oooo00<0_g@00018W02oW4P00P0000000`00B9`0gooo0?ooo`08
0?ooo`030000003oool0oooo0=`0oooo00<000000?ooo`3oool0[P3oool00100oooo00<0omnL04P0
001d_ol03@3oool00`000000oooo0?ooo`3M0?ooo`030000003oool0oooo0:d0oooo000A0?ooo`03
0=nLB00004P0W=oo00`0oooo00<000000?ooo`3oool0gP3oool00`000000oooo0?ooo`2/0?ooo`00
4P3oool00`3o_g@00000001d_`0;0?ooo`030000003oool0oooo0=l0oooo00<000000?ooo`3oool0
Z`3oool001<0oooo00D0oooO09a800000000001d0;ooo`080?ooo`030000003oool0oooo0=l0oooo
00<000000?ooo`3oool0Z`3oool00200oooo00<000000?ooo`3oool0h03oool00`000000oooo0?oo
o`2Z0?ooo`00803oool300000>40oooo00<000000?ooo`3oool0Z@3oool00200oooo00<000000?oo
o`3oool0h@3oool00`000000oooo0?ooo`2Y0?ooo`00803oool00`000000oooo0?ooo`3R0?ooo`03
0000003oool0oooo0:P0oooo000P0?ooo`030000003oool0oooo0>80oooo00<000000?ooo`3oool0
Z03oool00200oooo00<000000?ooo`3oool0h`3oool00`000000oooo0?ooo`2W0?ooo`00803oool0
0`000000oooo0?ooo`3S0?ooo`030000003oool0oooo0:L0oooo000P0?ooo`030000003oool0oooo
0>@0oooo00<000000?ooo`3oool0YP3oool00200oooo00<000000?ooo`3oool0i@3oool00`000000
oooo0?ooo`2U0?ooo`00803oool300000>D0oooo00<000000?ooo`3oool0Y@3oool00200oooo00<0
00000?ooo`3oool0iP3oool00`000000oooo0?ooo`2T0?ooo`00803oool00`000000oooo0?ooo`3V
0?ooo`030000003oool0oooo0:@0oooo000P0?ooo`030000003oool0oooo0>L0oooo00<000000?oo
o`3oool0X`3oool00200oooo00<000000?ooo`3oool0i`3oool00`000000oooo0?ooo`2S0?ooo`00
803oool00`000000oooo0?ooo`3X0?ooo`030000003oool0oooo0:80oooo000P0?ooo`030000003o
ool0oooo0>T0oooo00<000000?ooo`3oool0X@3oool00200oooo0`00003Y0?ooo`030000003oool0
oooo0:40oooo000P0?ooo`030000003oool0oooo0>X0oooo00<000000?ooo`3oool0X03oool00200
oooo00<000000?ooo`3oool0jP3oool00`000000oooo0?ooo`2P0?ooo`004@3oool01@3OW4P00000
000000000000B9cO00X0oooo00<000000?ooo`3oool0j`3oool00`000000oooo0?ooo`2O0?ooo`00
403oool01`3OW4P0000007Boo`3oool0oooO09a80018W=l02@3oool00`000000oooo0?ooo`3/0?oo
o`030000003oool0oooo09h0oooo000?0?ooo`030?oog`2LB00007Bo00<0oooo00<0okmd0018W03O
ool0203oool00`000000oooo0?ooo`3/0?ooo`030000003oool0oooo09h0oooo000@0?ooo`030;md
0000M;l0oooo0080oooo00<0_g@00018W03Oool0203oool00`000000oooo0?ooo`3]0?ooo`030000
003oool0oooo09d0oooo000@0?ooo`070?oOW01807@0_ooo0?ooo`2oM000000007Boo`090?ooo`03
0000003oool0oooo0>d0oooo00<000000?ooo`3oool0W@3oool00140oooo00D0oono07A8B0180000
000004RLg`0:0?ooo`@00000k@3oool00`000000oooo0?ooo`2L0?ooo`004@3oool0103ogi`0B000
00000000M;l;0?ooo`030000003oool0oooo0>h0oooo00<000000?ooo`3oool0W03oool00100oooo
00H0omnL04P0000007@0_ooo0?noM000M;l:0?ooo`030000003oool0oooo0>l0oooo00<000000?oo
o`3oool0V`3oool00100oooo00<0_g@00000M02oool00P3oool00`2oM000B9cO0?ooo`080?ooo`03
0000003oool0oooo0?00oooo00<000000?ooo`3oool0VP3oool00100oooo00<0_g@00018W03Oool0
0P3oool00`3OW4P007Bo0?ooo`080?ooo`030000003oool0oooo0?00oooo00<000000?ooo`3oool0
VP3oool00100oooo00L0okmd0000001d_ol0oooo0?oo_`1d0000M;oo00T0oooo00<000000?ooo`3o
ool0l@3oool00`000000oooo0?ooo`2I0?ooo`004@3oool01@3o_g@00000000000000000B9cO00X0
oooo00<000000?ooo`3oool0l@3oool00`000000oooo0?ooo`2I0?ooo`00803oool00`000000oooo
0?ooo`3b0?ooo`030000003oool0oooo09P0oooo000P0?ooo`<00000lP3oool00`000000oooo0?oo
o`2H0?ooo`00803oool00`000000oooo0?ooo`3c0?ooo`030000003oool0oooo09L0oooo000P0?oo
o`030000003oool0oooo0?<0oooo00<000000?ooo`3oool0U`3oool00200oooo00<000000?ooo`3o
ool0m03oool00`000000oooo0?ooo`2F0?ooo`00803oool00`000000oooo0?ooo`3d0?ooo`030000
003oool0oooo09H0oooo000P0?ooo`030000003oool0oooo0?D0oooo00<000000?ooo`3oool0U@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.842082, -7.66263, \
0.0388143, 0.0956625}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[\ 
        Evaluate[\ Table[LegendreP[n, x], {n, 0, 10}]], {x, \(-1\), 
          1}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.309017 0.294302 [
[.02381 .29652 -9.15625 -15.6875 ]
[.02381 .29652 9.15625 0 ]
[.2619 .29652 -13.6563 -15.6875 ]
[.2619 .29652 13.6563 0 ]
[.7381 .29652 -9.5 -15.6875 ]
[.7381 .29652 9.5 0 ]
[.97619 .29652 -5 -15.6875 ]
[.97619 .29652 5 0 ]
[.4875 .01472 -18.3125 -7.84375 ]
[.4875 .01472 0 7.84375 ]
[.4875 .16187 -27.3125 -7.84375 ]
[.4875 .16187 0 7.84375 ]
[.4875 .45617 -19 -7.84375 ]
[.4875 .45617 0 7.84375 ]
[.4875 .60332 -10 -7.84375 ]
[.4875 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 .30902 m
.02381 .31527 L
s
gsave
.02381 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 .30902 m
.2619 .31527 L
s
gsave
.2619 .29652 -74.6563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.7381 .30902 m
.7381 .31527 L
s
gsave
.7381 .29652 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 .30902 m
.97619 .31527 L
s
gsave
.97619 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .30902 m
.07143 .31277 L
s
.11905 .30902 m
.11905 .31277 L
s
.16667 .30902 m
.16667 .31277 L
s
.21429 .30902 m
.21429 .31277 L
s
.30952 .30902 m
.30952 .31277 L
s
.35714 .30902 m
.35714 .31277 L
s
.40476 .30902 m
.40476 .31277 L
s
.45238 .30902 m
.45238 .31277 L
s
.54762 .30902 m
.54762 .31277 L
s
.59524 .30902 m
.59524 .31277 L
s
.64286 .30902 m
.64286 .31277 L
s
.69048 .30902 m
.69048 .31277 L
s
.78571 .30902 m
.78571 .31277 L
s
.83333 .30902 m
.83333 .31277 L
s
.88095 .30902 m
.88095 .31277 L
s
.92857 .30902 m
.92857 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5 .01472 m
.50625 .01472 L
s
gsave
.4875 .01472 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .16187 m
.50625 .16187 L
s
gsave
.4875 .16187 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .45617 m
.50625 .45617 L
s
gsave
.4875 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .60332 m
.50625 .60332 L
s
gsave
.4875 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .04415 m
.50375 .04415 L
s
.5 .07358 m
.50375 .07358 L
s
.5 .10301 m
.50375 .10301 L
s
.5 .13244 m
.50375 .13244 L
s
.5 .1913 m
.50375 .1913 L
s
.5 .22073 m
.50375 .22073 L
s
.5 .25016 m
.50375 .25016 L
s
.5 .27959 m
.50375 .27959 L
s
.5 .33845 m
.50375 .33845 L
s
.5 .36788 m
.50375 .36788 L
s
.5 .39731 m
.50375 .39731 L
s
.5 .42674 m
.50375 .42674 L
s
.5 .4856 m
.50375 .4856 L
s
.5 .51503 m
.50375 .51503 L
s
.5 .54446 m
.50375 .54446 L
s
.5 .57389 m
.50375 .57389 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .60332 m
.06244 .60332 L
.10458 .60332 L
.14415 .60332 L
.18221 .60332 L
.22272 .60332 L
.26171 .60332 L
.30316 .60332 L
.34309 .60332 L
.3815 .60332 L
.42237 .60332 L
.46172 .60332 L
.49955 .60332 L
.53984 .60332 L
.57861 .60332 L
.61984 .60332 L
.65954 .60332 L
.69774 .60332 L
.73838 .60332 L
.77751 .60332 L
.81909 .60332 L
.85916 .60332 L
.89771 .60332 L
.93871 .60332 L
.97619 .60332 L
s
.02381 .01472 m
.06244 .03859 L
.10458 .06463 L
.14415 .08909 L
.18221 .11261 L
.22272 .13765 L
.26171 .16175 L
.30316 .18736 L
.34309 .21204 L
.3815 .23578 L
.42237 .26104 L
.46172 .28536 L
.49955 .30874 L
.53984 .33364 L
.57861 .3576 L
.61984 .38308 L
.65954 .40762 L
.69774 .43123 L
.73838 .45635 L
.77751 .48053 L
.81909 .50623 L
.85916 .53099 L
.89771 .55482 L
.93871 .58016 L
.97619 .60332 L
s
.02381 .60332 m
.06244 .53459 L
.10458 .46626 L
.14415 .40839 L
.18221 .35848 L
.22272 .31155 L
.26171 .27241 L
.30316 .2373 L
.34309 .2098 L
.3815 .1892 L
.40095 .18097 L
.42237 .1736 L
.44268 .16826 L
.45178 .16639 L
.46172 .16472 L
.4671 .16397 L
.4721 .16338 L
.47727 .16287 L
.48196 .1625 L
.48658 .16222 L
.4887 .16211 L
.49093 .16203 L
.49332 .16195 L
.49438 .16193 L
.49552 .16191 L
.49675 .16189 L
.49789 .16187 L
.49859 .16187 L
.49925 .16187 L
.50049 .16187 L
.50163 .16187 L
.50286 .16188 L
.50401 .1619 L
.50508 .16192 L
.50754 .16198 L
.51014 .16207 L
.51268 .16218 L
.51504 .16231 L
.5204 .16268 L
.5293 .16354 L
.53882 .1648 L
.54906 .16655 L
.56016 .16891 L
.58032 .17442 L
.60019 .18141 L
.62123 .19048 L
.65912 .21116 L
.69946 .23932 L
.73829 .27241 L
.77956 .31402 L
Mistroke
.81932 .36038 L
.85757 .41078 L
.89827 .47066 L
.93745 .53441 L
.97619 .60332 L
Mfstroke
.02381 .01472 m
.06244 .14385 L
.10458 .25431 L
.14415 .33187 L
.18221 .38494 L
.20178 .40476 L
.22272 .42081 L
.23293 .42681 L
.24402 .43203 L
.25449 .43579 L
.26416 .43827 L
.26882 .43915 L
.27139 .43954 L
.27383 .43986 L
.27609 .4401 L
.27856 .44032 L
.27965 .44039 L
.2808 .44046 L
.28182 .44051 L
.28289 .44056 L
.28355 .44058 L
.28428 .4406 L
.28556 .44062 L
.28629 .44063 L
.28706 .44063 L
.28779 .44063 L
.28846 .44062 L
.28975 .4406 L
.29048 .44058 L
.29115 .44056 L
.29247 .44051 L
.29371 .44044 L
.29597 .44029 L
.29846 .44008 L
.30351 .43948 L
.30821 .43875 L
.31334 .43774 L
.3237 .43512 L
.3422 .42853 L
.36181 .41914 L
.38276 .40672 L
.42426 .37627 L
.46424 .34186 L
.50271 .30651 L
.54363 .26914 L
.58303 .23594 L
.62092 .20897 L
.642 .19689 L
.66126 .1881 L
.67079 .18463 L
Mistroke
.68117 .18158 L
.68592 .18045 L
.69098 .17943 L
.69541 .1787 L
.70008 .17811 L
.70264 .17786 L
.704 .17775 L
.70545 .17764 L
.70679 .17757 L
.70801 .17751 L
.70923 .17746 L
.71039 .17743 L
.71103 .17742 L
.71174 .17741 L
.71248 .1774 L
.71317 .1774 L
.7144 .17741 L
.7151 .17742 L
.71573 .17743 L
.71708 .17748 L
.71856 .17754 L
.7201 .17763 L
.72153 .17773 L
.72417 .17796 L
.72665 .17823 L
.7313 .17891 L
.7363 .17986 L
.74171 .18116 L
.7512 .18415 L
.76163 .1885 L
.78269 .20088 L
.80217 .21689 L
.82308 .23929 L
.86441 .30092 L
.90422 .38432 L
.94252 .48923 L
.97619 .60332 L
Mfstroke
.02381 .60332 m
.04262 .4971 L
.06244 .40543 L
.08255 .33166 L
.10458 .27057 L
.12415 .23154 L
.13511 .21527 L
.14509 .20363 L
.15504 .1948 L
.16424 .18895 L
.16904 .18671 L
.17432 .18486 L
.17698 .18417 L
.17985 .1836 L
.18123 .18338 L
.18255 .18321 L
.18373 .18309 L
.18502 .18299 L
.18566 .18295 L
.18635 .18292 L
.18757 .18289 L
.18889 .18289 L
.18963 .18291 L
.19033 .18293 L
.19159 .18299 L
.19275 .18308 L
.19408 .18321 L
.19534 .18336 L
.19766 .18372 L
.19982 .18414 L
.20469 .18538 L
.20953 .18699 L
.21486 .1892 L
.22566 .19492 L
.24501 .20879 L
.26587 .22783 L
.30701 .27284 L
.34663 .31875 L
.38474 .35914 L
.40599 .37832 L
.4253 .393 L
.44538 .40509 L
.45441 .40937 L
.46435 .41323 L
.47407 .41612 L
.47938 .41732 L
.48436 .41819 L
.48674 .41853 L
.48901 .41879 L
.49106 .41899 L
Mistroke
.49326 .41916 L
.4945 .41923 L
.49563 .41929 L
.49676 .41933 L
.49783 .41936 L
.49908 .41938 L
.50022 .41938 L
.50148 .41937 L
.50212 .41936 L
.50281 .41934 L
.50412 .4193 L
.50533 .41924 L
.50807 .41906 L
.51048 .41885 L
.51304 .41855 L
.51747 .4179 L
.52236 .41695 L
.53316 .41406 L
.54329 .41035 L
.56351 .40016 L
.58203 .38777 L
.6217 .35278 L
.65986 .31135 L
.70048 .26422 L
.73958 .22252 L
.75953 .20517 L
.77003 .19764 L
.78113 .19114 L
.79106 .18678 L
.7967 .18499 L
.7994 .18431 L
.80197 .18379 L
.80429 .18342 L
.80551 .18326 L
.80681 .18312 L
.8079 .18303 L
.8091 .18295 L
.8102 .18291 L
.81123 .18289 L
.81244 .18289 L
.81372 .18293 L
.81481 .18298 L
.816 .18307 L
.81729 .18319 L
.81867 .18337 L
.82116 .18378 L
.82374 .18434 L
.82648 .1851 L
.83139 .18689 L
.83667 .18943 L
Mistroke
.84248 .19302 L
.85288 .20161 L
.86263 .21237 L
.87302 .22697 L
.88289 .24405 L
.90142 .28531 L
.92124 .34419 L
.93963 .41411 L
.97619 .60332 L
Mfstroke
.02381 .01472 m
.04262 .16621 L
.06244 .28168 L
.07286 .32672 L
.08255 .36023 L
.0932 .38858 L
.09912 .40089 L
.10458 .41024 L
.10944 .41704 L
.1148 .42297 L
.12018 .42741 L
.12313 .42922 L
.12589 .43054 L
.12713 .43103 L
.12848 .43147 L
.12968 .4318 L
.13095 .43208 L
.13221 .43229 L
.13336 .43243 L
.13451 .43251 L
.13559 .43253 L
.13686 .43251 L
.1375 .43247 L
.1382 .43241 L
.13944 .43226 L
.1406 .43207 L
.14322 .43146 L
.14603 .43054 L
.15131 .4281 L
.1562 .42509 L
.16727 .41592 L
.18739 .39274 L
.22665 .33489 L
.2644 .27916 L
.28538 .25297 L
.30461 .23354 L
.32345 .21943 L
.33293 .21431 L
.33826 .21202 L
.34329 .21025 L
.34822 .20889 L
.35082 .20833 L
.35357 .20784 L
.35588 .20751 L
.35711 .20737 L
.35843 .20725 L
.35963 .20716 L
.36074 .20709 L
.36185 .20704 L
.3629 .20702 L
Mistroke
.36423 .207 L
.36542 .20702 L
.36609 .20703 L
.36681 .20705 L
.36811 .20712 L
.36933 .2072 L
.37062 .20731 L
.37293 .20756 L
.37563 .20796 L
.37814 .20842 L
.38385 .20979 L
.38888 .21137 L
.39433 .21346 L
.40425 .21823 L
.42282 .23028 L
.44263 .24698 L
.46365 .26803 L
.50145 .3107 L
.54171 .35563 L
.56022 .37367 L
.58045 .39014 L
.59016 .39658 L
.60053 .40225 L
.60626 .40482 L
.61147 .40679 L
.61641 .40832 L
.62164 .40957 L
.62394 .41 L
.62641 .41038 L
.62782 .41056 L
.62912 .4107 L
.63043 .41082 L
.63165 .4109 L
.63275 .41096 L
.63391 .411 L
.63508 .41103 L
.63638 .41103 L
.63749 .41101 L
.63869 .41097 L
.63982 .41091 L
.64086 .41084 L
.64337 .41061 L
.6457 .41031 L
.64826 .40988 L
.651 .40932 L
.65659 .40782 L
.66167 .40604 L
.67164 .40143 L
.68084 .39585 L
.70162 .37873 L
Mistroke
.72036 .35836 L
.74099 .33143 L
.7813 .27086 L
.79985 .24298 L
.82009 .21587 L
.83106 .20375 L
.8364 .19874 L
.84124 .19478 L
.84592 .19155 L
.85097 .18877 L
.85375 .18758 L
.85631 .18671 L
.85876 .18609 L
.85999 .18586 L
.86064 .18576 L
.86134 .18567 L
.86264 .18555 L
.86385 .1855 L
.86515 .18551 L
.86589 .18555 L
.86659 .1856 L
.86783 .18575 L
.86898 .18594 L
.87023 .18621 L
.87155 .18657 L
.87386 .18738 L
.876 .18836 L
.88086 .19137 L
.88616 .19601 L
.89098 .20153 L
.89623 .20906 L
.90178 .21886 L
.91177 .24166 L
.92098 .26913 L
.93089 .30639 L
.94178 .35756 L
.95838 .45873 L
.97619 .60332 L
Mfstroke
.02381 .60332 m
.03279 .49739 L
.04262 .40375 L
.05288 .3279 L
.06244 .27471 L
.06792 .25093 L
.0731 .23248 L
.07781 .21883 L
.0828 .20734 L
.08539 .20251 L
.08823 .19803 L
.09085 .19463 L
.09324 .19211 L
.09606 .18982 L
.09762 .18885 L
.0991 .18813 L
.10046 .18762 L
.10172 .18728 L
.10244 .18714 L
.1031 .18704 L
.10386 .18698 L
.10458 .18696 L
.10574 .187 L
.10699 .18715 L
.10816 .18739 L
.10923 .18769 L
.11049 .18814 L
.11186 .18874 L
.11434 .19013 L
.1197 .1943 L
.12466 .19945 L
.13361 .21136 L
.14313 .22686 L
.18088 .30068 L
.20185 .33974 L
.22108 .36926 L
.23085 .38125 L
.23993 .39039 L
.24941 .39779 L
.25442 .4008 L
.25977 .40333 L
.26236 .40431 L
.2651 .40516 L
.26744 .40574 L
.27001 .40623 L
.27146 .40643 L
.27278 .40657 L
.27351 .40663 L
.2743 .40669 L
.27505 .40672 L
Mistroke
.27575 .40675 L
.27645 .40676 L
.27711 .40675 L
.27835 .40673 L
.27904 .40669 L
.27968 .40666 L
.28112 .40654 L
.28229 .4064 L
.28354 .40623 L
.2858 .40582 L
.28848 .40519 L
.29092 .40448 L
.29624 .40251 L
.30127 .40012 L
.31095 .39416 L
.32122 .38606 L
.33965 .36774 L
.37745 .32078 L
.41771 .26967 L
.43622 .24985 L
.44671 .24034 L
.45645 .2328 L
.46698 .2262 L
.47653 .22171 L
.48179 .21986 L
.48674 .21854 L
.48947 .21799 L
.49198 .2176 L
.4934 .21742 L
.49473 .21728 L
.49612 .21718 L
.49685 .21713 L
.49764 .2171 L
.49886 .21706 L
.49955 .21705 L
.5002 .21705 L
.5015 .21707 L
.50224 .21709 L
.50291 .21712 L
.50405 .21719 L
.50529 .21729 L
.5078 .21757 L
.51057 .218 L
.51352 .2186 L
.51886 .22006 L
.52427 .22203 L
.52939 .22432 L
.53897 .22972 L
.55896 .24531 L
.5774 .26411 L
Mistroke
.61677 .31316 L
.65463 .36126 L
.67568 .3833 L
.68567 .39168 L
.69494 .39799 L
.70445 .4028 L
.70948 .40461 L
.71228 .40539 L
.71483 .40594 L
.71601 .40615 L
.71727 .40634 L
.71835 .40648 L
.71951 .40659 L
.72075 .40668 L
.72192 .40674 L
.72263 .40675 L
.7233 .40676 L
.72456 .40674 L
.72567 .40669 L
.72684 .40661 L
.72785 .40651 L
.72895 .40638 L
.73142 .40597 L
.73373 .40546 L
.73843 .40403 L
.74357 .40184 L
.74894 .39885 L
.75393 .39544 L
.77244 .37746 L
.78297 .36366 L
.79291 .34848 L
.8115 .31548 L
.84999 .23938 L
.86027 .22105 L
.86988 .20636 L
.87512 .19971 L
.8799 .19467 L
.88522 .19041 L
.88662 .18955 L
.88815 .18874 L
.8896 .18811 L
.89093 .18764 L
.89218 .18731 L
.89349 .18708 L
.89481 .18697 L
.89556 .18696 L
.89626 .18699 L
.89753 .18713 L
.89869 .18737 L
.89933 .18755 L
Mistroke
.90002 .18778 L
.90128 .1883 L
.90365 .18962 L
.90615 .19156 L
.91065 .1965 L
.91314 .20008 L
.91581 .20466 L
.92064 .21495 L
.92639 .23086 L
.93164 .24925 L
.94148 .29467 L
.95062 .35145 L
.96065 .43243 L
.97134 .54351 L
.97619 .60332 L
Mfstroke
.02381 .01472 m
.03279 .15129 L
.04262 .26243 L
.04758 .30538 L
.05288 .34282 L
.05785 .37074 L
.06244 .39098 L
.06788 .40884 L
.07075 .41583 L
.07379 .4216 L
.07642 .42532 L
.07889 .4278 L
.0803 .42881 L
.08159 .42948 L
.08231 .42976 L
.08309 .42998 L
.08382 .43011 L
.08451 .43017 L
.08577 .43013 L
.08648 .43003 L
.08713 .42988 L
.08833 .42947 L
.0896 .42886 L
.0918 .4274 L
.09422 .42525 L
.09958 .41862 L
.10458 .41049 L
.12493 .36502 L
.14349 .31749 L
.16419 .26978 L
.17329 .25263 L
.18333 .23713 L
.18795 .23131 L
.19292 .22596 L
.19722 .22213 L
.20192 .21878 L
.20694 .21616 L
.20967 .21514 L
.21098 .21476 L
.21222 .21446 L
.21337 .21422 L
.21461 .21403 L
.21574 .2139 L
.21679 .21383 L
.21791 .21379 L
.21915 .21381 L
.22045 .21388 L
.22166 .214 L
.22291 .21418 L
.22425 .21443 L
Mistroke
.22664 .21504 L
.22949 .21601 L
.23207 .21712 L
.23716 .21991 L
.24195 .22324 L
.25086 .23108 L
.26049 .24163 L
.30154 .30094 L
.322 .33203 L
.34108 .35764 L
.35945 .37706 L
.36974 .38514 L
.37911 .39056 L
.38393 .39259 L
.38657 .39349 L
.38905 .39419 L
.39133 .39471 L
.39341 .39508 L
.39452 .39524 L
.39573 .39538 L
.39695 .39549 L
.39823 .39557 L
.3989 .3956 L
.39964 .39561 L
.40033 .39562 L
.40096 .39561 L
.40217 .39558 L
.40344 .39551 L
.4041 .39546 L
.40481 .39539 L
.40609 .39525 L
.409 .39478 L
.41168 .39419 L
.41423 .39348 L
.419 .39179 L
.42864 .38694 L
.43889 .37983 L
.4573 .36264 L
.49502 .31574 L
.51597 .28765 L
.53519 .26375 L
.54486 .25311 L
.55521 .24314 L
.56479 .23543 L
.57385 .22966 L
.57859 .22728 L
.58374 .22521 L
.58602 .22447 L
.58843 .22382 L
.59068 .22331 L
Mistroke
.59275 .22295 L
.59392 .22278 L
.5952 .22264 L
.5959 .22257 L
.59655 .22252 L
.59783 .22245 L
.59904 .22242 L
.60014 .22242 L
.6014 .22245 L
.60259 .22251 L
.60391 .22262 L
.60531 .22277 L
.6078 .22316 L
.61049 .22373 L
.61344 .22455 L
.61868 .22648 L
.62365 .22887 L
.63296 .2348 L
.64298 .24319 L
.65388 .25449 L
.69468 .31122 L
.71495 .34257 L
.73397 .3695 L
.74283 .38026 L
.75215 .38983 L
.75703 .394 L
.76237 .39783 L
.76723 .40058 L
.77174 .40248 L
.77437 .40328 L
.7768 .4038 L
.77806 .404 L
.77878 .40408 L
.77946 .40415 L
.78017 .4042 L
.78082 .40423 L
.78156 .40424 L
.78227 .40424 L
.78347 .40419 L
.78474 .40408 L
.78594 .40392 L
.78703 .40374 L
.7897 .40308 L
.79221 .40222 L
.79687 .39997 L
.80115 .39716 L
.81081 .38816 L
.82058 .37532 L
.82968 .36012 L
.85037 .31571 L
Mistroke
.86905 .26838 L
.88935 .21943 L
.89472 .20881 L
.90042 .19941 L
.90528 .1933 L
.90805 .19075 L
.90939 .18978 L
.91064 .18904 L
.91188 .18848 L
.91254 .18825 L
.91324 .18807 L
.91453 .18787 L
.91571 .18787 L
.91685 .18804 L
.91809 .18842 L
.91926 .18897 L
.92034 .18965 L
.92276 .19179 L
.92413 .1934 L
.92542 .19519 L
.9282 .20002 L
.93078 .20573 L
.93578 .22053 L
.94122 .24283 L
.94602 .26849 L
.95112 .30269 L
.95994 .38102 L
.96968 .5006 L
.97619 .60332 L
Mfstroke
.02381 .60332 m
.02846 .50845 L
.03279 .43435 L
.03793 .36224 L
.04262 .31002 L
.04754 .26721 L
.0522 .23673 L
.05711 .2137 L
.05988 .20435 L
.06244 .19782 L
.06487 .19334 L
.06623 .19149 L
.06747 .1902 L
.06862 .18934 L
.06988 .18873 L
.07107 .18847 L
.07218 .18848 L
.07285 .18861 L
.07358 .18885 L
.0749 .18953 L
.07618 .19048 L
.07737 .1916 L
.08006 .19494 L
.0829 .19957 L
.08824 .2108 L
.09323 .22362 L
.10458 .25783 L
.12322 .31683 L
.13305 .34432 L
.14349 .36846 L
.14873 .37828 L
.15452 .38721 L
.15979 .39355 L
.16473 .39796 L
.16719 .3996 L
.16978 .40094 L
.17199 .40177 L
.17325 .40212 L
.17441 .40236 L
.17506 .40246 L
.17577 .40255 L
.17706 .40262 L
.17831 .40261 L
.17946 .40253 L
.1807 .40236 L
.18141 .40222 L
.18208 .40207 L
.18342 .4017 L
.18485 .4012 L
.18743 .40004 L
Mistroke
.18983 .39868 L
.19526 .39461 L
.20514 .38414 L
.22367 .35638 L
.24369 .32021 L
.26531 .2816 L
.27548 .26561 L
.28622 .25115 L
.29627 .24029 L
.30543 .23292 L
.31031 .23002 L
.31493 .22796 L
.31746 .2271 L
.31981 .22649 L
.32113 .22622 L
.32238 .22602 L
.32367 .22586 L
.32436 .22579 L
.3251 .22574 L
.32626 .22569 L
.32734 .22568 L
.32851 .22571 L
.32977 .22579 L
.33041 .22585 L
.3311 .22592 L
.33235 .2261 L
.33468 .22655 L
.3369 .22713 L
.33925 .22789 L
.34345 .22966 L
.3484 .23235 L
.35291 .23537 L
.36306 .24396 L
.38139 .26471 L
.4212 .32155 L
.4394 .34663 L
.44976 .35918 L
.4595 .36936 L
.4698 .37809 L
.47925 .38405 L
.48417 .38631 L
.48935 .38805 L
.49205 .38868 L
.49455 .38911 L
.496 .38929 L
.49731 .3894 L
.49803 .38944 L
.49882 .38947 L
.49956 .38949 L
.50025 .38949 L
Mistroke
.5015 .38946 L
.50269 .3894 L
.50375 .38931 L
.50491 .38918 L
.50617 .389 L
.50751 .38877 L
.50993 .38823 L
.51468 .38675 L
.51899 .38492 L
.52962 .37852 L
.53962 .37019 L
.55971 .34777 L
.57819 .32243 L
.59863 .2924 L
.6177 .26588 L
.63604 .24483 L
.64632 .23593 L
.65078 .23286 L
.65569 .23008 L
.6605 .22798 L
.66314 .22712 L
.66561 .22648 L
.66674 .22626 L
.66792 .22606 L
.66894 .22592 L
.67003 .22581 L
.67121 .22572 L
.6723 .22569 L
.67361 .22569 L
.67479 .22573 L
.67609 .22583 L
.67674 .2259 L
.67746 .226 L
.67992 .22643 L
.6813 .22676 L
.68279 .22718 L
.68547 .22811 L
.69074 .23058 L
.69556 .23359 L
.70544 .24194 L
.71459 .25214 L
.73534 .28267 L
.77454 .35331 L
.78486 .37021 L
.79601 .38553 L
.80113 .39121 L
.80653 .39609 L
.81114 .39926 L
.81376 .40061 L
.81619 .40157 L
Mistroke
.81746 .40195 L
.81881 .40227 L
.81996 .40246 L
.82122 .40259 L
.82193 .40262 L
.82258 .40263 L
.82333 .40261 L
.82404 .40256 L
.8253 .40241 L
.82602 .40228 L
.82668 .40214 L
.82798 .40178 L
.8292 .40136 L
.83148 .40034 L
.83394 .39889 L
.8366 .39692 L
.8413 .39237 L
.84627 .3861 L
.85517 .37112 L
.86465 .35016 L
.8748 .32268 L
.89313 .26516 L
.9033 .2335 L
.90866 .21853 L
.91436 .20495 L
.91933 .19585 L
.92069 .19391 L
.92216 .19211 L
.92342 .19082 L
.92481 .18972 L
.92608 .18899 L
.92727 .18858 L
.92852 .18844 L
.92922 .1885 L
.92986 .18864 L
.931 .18912 L
.93224 .18995 L
.9334 .19107 L
.93448 .19239 L
.93692 .19649 L
.93915 .20166 L
.94173 .20947 L
.94417 .21887 L
.94851 .24077 L
.95324 .27322 L
.95842 .32058 L
.96394 .38663 L
.97387 .5539 L
.97619 .60332 L
Mfstroke
.02381 .01472 m
.02846 .1307 L
.03279 .21694 L
.03754 .2906 L
.04262 .3486 L
.04518 .37076 L
.04795 .3901 L
.05058 .4043 L
.053 .41423 L
.05518 .42083 L
.05642 .42366 L
.05758 .42574 L
.05881 .42738 L
.05993 .4284 L
.06122 .42905 L
.06244 .42916 L
.0631 .42903 L
.06381 .42875 L
.0651 .42788 L
.06635 .42661 L
.0675 .4251 L
.07012 .42055 L
.07289 .41427 L
.08259 .38374 L
.10458 .29911 L
.11406 .26763 L
.11928 .25315 L
.12422 .24156 L
.12881 .23274 L
.13381 .22534 L
.13601 .2228 L
.13835 .22059 L
.14035 .21908 L
.14255 .21783 L
.14377 .21731 L
.1449 .21695 L
.14614 .21668 L
.14684 .21658 L
.14747 .21653 L
.14875 .21652 L
.14947 .21657 L
.15013 .21665 L
.15135 .2169 L
.15264 .21728 L
.15374 .21771 L
.15492 .21827 L
.15705 .21953 L
.1619 .22356 L
.16749 .23001 L
.1728 .23769 L
Mistroke
.18275 .25533 L
.221 .33616 L
.23116 .35458 L
.24071 .3689 L
.24601 .37537 L
.25172 .38109 L
.25692 .3851 L
.25942 .38662 L
.2617 .38777 L
.26412 .38875 L
.26641 .38944 L
.26767 .38973 L
.26883 .38993 L
.27011 .39009 L
.27075 .39014 L
.27146 .39018 L
.27264 .3902 L
.27372 .39016 L
.27491 .39007 L
.27616 .38991 L
.27725 .38972 L
.27826 .38949 L
.28056 .38884 L
.28326 .3878 L
.2858 .38657 L
.29056 .38362 L
.29569 .37955 L
.30124 .3742 L
.32115 .34841 L
.34221 .31462 L
.36205 .28267 L
.38015 .25792 L
.39015 .24728 L
.39963 .23966 L
.40483 .2366 L
.40964 .23449 L
.41215 .23367 L
.41492 .233 L
.41632 .23275 L
.41705 .23264 L
.41782 .23255 L
.4185 .23248 L
.41924 .23243 L
.42054 .23237 L
.42186 .23236 L
.42307 .23241 L
.4237 .23245 L
.4244 .23251 L
.42581 .23268 L
.42705 .23288 L
Mistroke
.42839 .23314 L
.4308 .23376 L
.4333 .23459 L
.43567 .23555 L
.44013 .23779 L
.45014 .24486 L
.46107 .25546 L
.48076 .28044 L
.50173 .31165 L
.52172 .34106 L
.53224 .35481 L
.54334 .36718 L
.55329 .3759 L
.55895 .3797 L
.56422 .38244 L
.56671 .38345 L
.56907 .38423 L
.57119 .3848 L
.57349 .38525 L
.57468 .38542 L
.57599 .38556 L
.57665 .38561 L
.57736 .38565 L
.57865 .38567 L
.57986 .38565 L
.58114 .38558 L
.58223 .38548 L
.58342 .38533 L
.58474 .3851 L
.58596 .38485 L
.58871 .38411 L
.59114 .38325 L
.59371 .38215 L
.60305 .37647 L
.60821 .37225 L
.61303 .36764 L
.624 .355 L
.66285 .29504 L
.68261 .26412 L
.69304 .25029 L
.69876 .24383 L
.70414 .23863 L
.70935 .23447 L
.71209 .23267 L
.71498 .23106 L
.71742 .22994 L
.72009 .22899 L
.72127 .22866 L
.72253 .22836 L
.72371 .22815 L
Mistroke
.72479 .228 L
.72598 .22789 L
.72724 .22784 L
.72832 .22785 L
.72951 .22791 L
.73078 .22805 L
.73149 .22816 L
.73215 .22827 L
.73463 .22888 L
.7369 .22967 L
.73939 .23079 L
.74392 .23351 L
.74923 .23781 L
.75424 .24294 L
.76359 .25518 L
.78456 .29305 L
.80293 .33277 L
.81336 .35495 L
.8231 .37351 L
.83271 .38829 L
.83811 .39448 L
.84049 .39666 L
.84303 .39855 L
.8443 .39934 L
.84572 .40008 L
.84701 .40063 L
.8482 .40102 L
.84889 .40119 L
.8496 .40134 L
.85032 .40144 L
.85111 .40151 L
.85179 .40153 L
.85252 .40151 L
.85384 .40136 L
.85453 .40122 L
.85516 .40106 L
.8566 .40058 L
.85787 .4 L
.85921 .39924 L
.86153 .39755 L
.86409 .39513 L
.8693 .38836 L
.87424 .37965 L
.88353 .35739 L
.89352 .32571 L
.90441 .28469 L
.91456 .24488 L
.92412 .21197 L
.92669 .20482 L
.92941 .19844 L
Mistroke
.93083 .19566 L
.93239 .19309 L
.93381 .19124 L
.93513 .18997 L
.93642 .18917 L
.93765 .18886 L
.93875 .18898 L
.93994 .18954 L
.94124 .19071 L
.9419 .19153 L
.94262 .19262 L
.94391 .19506 L
.9451 .19794 L
.94692 .20351 L
.94891 .21138 L
.95291 .23347 L
.95637 .26018 L
.96007 .29763 L
.96442 .35527 L
.96852 .42505 L
.97619 .60332 L
Mfstroke
.02381 .60332 m
.02846 .46503 L
.03279 .36786 L
.03754 .2906 L
.0402 .25872 L
.04262 .23587 L
.04516 .2175 L
.0475 .20508 L
.04873 .20016 L
.05008 .19588 L
.05076 .19416 L
.05149 .19261 L
.05281 .19057 L
.05404 .18949 L
.05474 .18921 L
.05539 .18916 L
.05658 .18955 L
.05784 .19062 L
.05908 .19227 L
.06026 .19433 L
.06244 .19934 L
.06757 .21591 L
.07299 .23842 L
.08269 .28402 L
.09325 .33138 L
.09907 .35343 L
.10458 .37072 L
.1095 .38291 L
.11209 .38805 L
.11485 .39254 L
.11716 .39554 L
.11839 .39684 L
.11972 .39804 L
.12093 .39894 L
.12204 .3996 L
.1231 .40009 L
.12422 .40045 L
.12537 .40068 L
.12641 .40074 L
.12762 .40066 L
.12875 .40044 L
.12981 .4001 L
.13078 .39968 L
.13297 .39837 L
.13519 .39655 L
.13763 .39401 L
.14255 .3873 L
.15271 .36804 L
.162 .34612 L
.18289 .29305 L
Mistroke
.19211 .27211 L
.20191 .25359 L
.20756 .2451 L
.21268 .23894 L
.21515 .2365 L
.21782 .23426 L
.22035 .23252 L
.22265 .23127 L
.22503 .23031 L
.22634 .22992 L
.22758 .22964 L
.22873 .22947 L
.22979 .22937 L
.23095 .22934 L
.23218 .22939 L
.23342 .22953 L
.23411 .22964 L
.23475 .22977 L
.23715 .23045 L
.23843 .23094 L
.23982 .23157 L
.24263 .23315 L
.24792 .23718 L
.25356 .24288 L
.26371 .25625 L
.28366 .29037 L
.30526 .33021 L
.31539 .34696 L
.3261 .36193 L
.33525 .37187 L
.34044 .37619 L
.34529 .3793 L
.34789 .38059 L
.35029 .38155 L
.35292 .38234 L
.35426 .38263 L
.35497 .38276 L
.35573 .38287 L
.35701 .38302 L
.35818 .38309 L
.35944 .38311 L
.36016 .38309 L
.36083 .38305 L
.36204 .38294 L
.36314 .3828 L
.36442 .38257 L
.36562 .38229 L
.36787 .38163 L
.37025 .38072 L
.37452 .37857 L
Mistroke
.37911 .37554 L
.38417 .37137 L
.40359 .34865 L
.42448 .31655 L
.4442 .28504 L
.45462 .27 L
.46572 .25632 L
.47129 .25062 L
.47653 .24607 L
.48163 .24244 L
.48633 .23985 L
.4908 .23807 L
.49305 .23744 L
.49431 .23716 L
.49551 .23694 L
.49681 .23677 L
.498 .23666 L
.49927 .2366 L
.49999 .23659 L
.50065 .2366 L
.50176 .23665 L
.50298 .23675 L
.50425 .23691 L
.50545 .23711 L
.50788 .23768 L
.51015 .23839 L
.51529 .24066 L
.5203 .24372 L
.52565 .24786 L
.53535 .25751 L
.54416 .26835 L
.58416 .33037 L
.60249 .35676 L
.6129 .3686 L
.618 .37326 L
.62264 .37678 L
.62757 .3797 L
.62983 .38076 L
.63224 .38167 L
.63358 .38208 L
.63479 .38239 L
.63603 .38265 L
.63716 .38284 L
.6378 .38293 L
.6385 .383 L
.63919 .38305 L
.63993 .38309 L
.64058 .38311 L
.64128 .3831 L
.64253 .38305 L
Mistroke
.64322 .383 L
.64395 .38292 L
.64525 .38272 L
.64665 .38244 L
.6482 .38204 L
.65096 .38108 L
.65355 .37991 L
.65838 .37703 L
.66358 .37293 L
.67368 .36221 L
.68471 .34682 L
.70479 .31197 L
.7261 .27276 L
.73617 .25644 L
.74543 .24405 L
.75004 .23909 L
.7551 .23472 L
.75756 .23303 L
.76019 .23157 L
.76164 .23091 L
.76297 .23041 L
.76423 .23002 L
.76557 .2297 L
.76672 .22951 L
.76799 .22938 L
.76915 .22934 L
.77023 .22937 L
.77151 .2295 L
.77268 .22969 L
.77396 .23 L
.77531 .23043 L
.7776 .23139 L
.77972 .23256 L
.78455 .23622 L
.78959 .24149 L
.79506 .24883 L
.80499 .26617 L
.82358 .30914 L
.84305 .35841 L
.84839 .37042 L
.85412 .38172 L
.8595 .39034 L
.86448 .39624 L
.86693 .3983 L
.8683 .39919 L
.86953 .39983 L
.8702 .4001 L
.8709 .40034 L
.87167 .40054 L
.87237 .40066 L
Mistroke
.87364 .40074 L
.87437 .40071 L
.87504 .40062 L
.87634 .40029 L
.87757 .39979 L
.87868 .39919 L
.87988 .39836 L
.88257 .39583 L
.88506 .39266 L
.8894 .38519 L
.89419 .37406 L
.90388 .34271 L
.91411 .29906 L
.92488 .24815 L
.92933 .22835 L
.93407 .20997 L
.9368 .20139 L
.93928 .19527 L
.94047 .19299 L
.94177 .19107 L
.94299 .18984 L
.94412 .18925 L
.94509 .18918 L
.94597 .1895 L
.94699 .19034 L
.94796 .19163 L
.94885 .19329 L
.94967 .19523 L
.95154 .20117 L
.95342 .2095 L
.9555 .22173 L
.95767 .23827 L
.95971 .25768 L
.96384 .31012 L
.96763 .37608 L
.9718 .47168 L
.97619 .60332 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00f@3oool0
0`000000oooo0?ooo`3D0?ooo`00bP3oool01P2oM000000000000000000004RL0=ooo`T0oooo00<0
00000?ooo`3oool0e03oool00</0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`3D
0?ooo`00b`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0=@0oooo003;0?ooo`03
0?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0e03oool00;/0oooo00<0oooO09a800000000
1`0000000`18W=l0oooo0?ooo`030?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0
e03oool001D0oooo00<000000?ooo`3oool0/`3oool00`3ogi`0B00007Boo`0;0?ooo`@00000d`3o
ool001D0oooo0`00002c0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0e03oool0
01D0oooo0P0000000`3oool000000?ooo`2a0?ooo`030?oOW0180000M;oo00/0oooo00<000000?oo
o`3oool0e03oool001D0oooo0P0000020?ooo`800000/03oool00`3ogi`0B00007Boo`0;0?ooo`03
0000003oool0oooo0=@0oooo000E0?ooo`800000103oool200000:h0oooo00<0omnL04P0001d_ol0
2`3oool00`000000oooo0?ooo`3D0?ooo`005@3oool2000000H0oooo00<000000?ooo`3oool0Z@3o
ool01@3ooml0W4P0000000000000M;oo00/0oooo00<000000?ooo`3oool0e03oool001D0oooo0`00
00060?ooo`800000Z`3oool00`3o_g@0000007Boo`0;0?ooo`030000003oool0oooo0=@0oooo000E
0?ooo`<00000203oool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo0=@0oooo000E0?oo
o`<000002@3oool200000;H0oooo00<000000?ooo`3oool0e03oool001D0oooo0P0000000`3oool0
00000?ooo`090?ooo`800000]03oool00`000000oooo0?ooo`3D0?ooo`005@3oool2000000030?oo
o`000000oooo00/0oooo00<000000?ooo`3oool0/@3oool00`000000oooo0?ooo`3D0?ooo`005@3o
ool2000000030?ooo`000000oooo00`0oooo0P00002a0?ooo`030000003oool0oooo0=@0oooo000E
0?ooo`<0000000<0oooo0000003oool03@3oool00`000000oooo0?ooo`2^0?ooo`030000003oool0
oooo0=@0oooo000F0?ooo`80000000<0oooo0000003oool03P3oool200000:h0oooo0`00003D0?oo
o`005P3oool2000000030?ooo`000000oooo0100oooo00<000000?ooo`3oool0Z`3oool00`000000
oooo0?ooo`3D0?ooo`005P3oool2000000030?ooo`000000oooo0140oooo0P00002[0?ooo`030000
003oool0oooo0=@0oooo000F0?ooo`8000000P3oool00`000000oooo0?ooo`0A0?ooo`800000Z@3o
ool00`000000oooo0?ooo`3D0?ooo`005P3oool200000080oooo00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`2V0?ooo`030000003oool0oooo0=@0oooo000F0?ooo`8000000P3oool00`00
0000oooo0?ooo`0D0?ooo`800000YP3oool00`000000oooo0?ooo`3D0?ooo`005P3oool2000000<0
oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`2S0?ooo`030000003oool0oooo0=@0
oooo000F0?ooo`030000003oool000000080oooo00<000000?ooo`3oool05P3oool200000:<0oooo
00<000000?ooo`3oool0e03oool001H0oooo00<000000?ooo`0000000P3oool00`000000oooo0?oo
o`0H0?ooo`030000003oool0oooo0:00oooo00<000000?ooo`3oool0e03oool001H0oooo0`000003
0?ooo`030000003oool0oooo01P0oooo0P00002P0?ooo`030000003oool0oooo0=@0oooo000F0?oo
o`<000000`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo09d0oooo00<000000?oo
o`3oool0e03oool001H0oooo0`0000030?ooo`030000003oool0oooo01/0oooo0P00002M0?ooo`03
0000003oool0oooo0=@0oooo000F0?ooo`<000000`3oool00`000000oooo0?ooo`0M0?ooo`800000
V`3oool300000=@0oooo000F0?ooo`<00000103oool00`000000oooo0?ooo`0N0?ooo`030000003o
ool0oooo09P0oooo00<000000?ooo`3oool0e03oool001H0oooo0`0000040?ooo`030000003oool0
oooo01l0oooo0P00002H0?ooo`030000003oool0oooo0=@0oooo000F0?ooo`80000000<0oooo0000
003oool00P3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo09D0oooo00<000000?oo
o`3oool0e03oool001H0oooo0P0000000`3oool000000?ooo`030?ooo`030000003oool0oooo0240
oooo0P00002E0?ooo`030000003oool0oooo0=@0oooo000F0?ooo`80000000<0oooo0000003oool0
0`3oool00`000000oooo0?ooo`0S0?ooo`800000T`3oool00`000000oooo0?ooo`3D0?ooo`005P3o
ool2000000030?ooo`000000oooo00<0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?oo
o`2@0?ooo`030000003oool0oooo0=@0oooo000F0?ooo`80000000<0oooo0000003oool00`3oool0
0`000000oooo0?ooo`0V0?ooo`800000T03oool00`000000oooo0?ooo`3D0?ooo`005P3oool20000
00030?ooo`000000oooo00@0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`2=0?oo
o`030000003oool0oooo0=@0oooo000F0?ooo`80000000<0oooo0000003oool0103oool00`000000
oooo0?ooo`0X0?ooo`800000S@3oool00`000000oooo0?ooo`3D0?ooo`005P3oool2000000030?oo
o`000000oooo00@0oooo00<000000?ooo`3oool0:P3oool2000008/0oooo00<000000?ooo`3oool0
e03oool001H0oooo00D000000?ooo`000000oooo000000050?ooo`030000003oool0oooo02/0oooo
00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`3D0?ooo`005`3oool2000000030?ooo`00
0000oooo00@0oooo00<000000?ooo`3oool0;03oool2000008P0oooo0`00003D0?ooo`005`3oool2
000000030?ooo`000000oooo00@0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`25
0?ooo`030000003oool0oooo0=@0oooo000G0?ooo`80000000<0oooo0000003oool01@3oool00`00
0000oooo0?ooo`0^0?ooo`800000Q@3oool00`000000oooo0?ooo`3D0?ooo`005`3oool200000003
0?ooo`000000oooo00D0oooo00<000000?ooo`3oool0<03oool2000008<0oooo00<000000?ooo`3o
ool0e03oool001L0oooo0P0000000`3oool000000?ooo`050?ooo`030000003oool0oooo0380oooo
00<000000?ooo`3oool0P03oool00`000000oooo0?ooo`3D0?ooo`005`3oool2000000030?ooo`00
0000oooo00D0oooo00<000000?ooo`3oool0<`3oool200000800oooo00<000000?ooo`3oool0e03o
ool001L0oooo0P0000000`3oool000000?ooo`060?ooo`030000003oool0oooo03@0oooo0P00001n
0?ooo`030000003oool0oooo0=@0oooo000G0?ooo`8000000P3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo03H0oooo00<000000?ooo`3oool0N`3oool00`000000oooo0?ooo`3D0?oo
o`005`3oool200000080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0g0?ooo`80
0000N`3oool00`000000oooo0?ooo`3D0?ooo`005`3oool200000080oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`0h0?ooo`800000N@3oool00`000000oooo0?ooo`3D0?ooo`005`3o
ool200000080oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0j0?ooo`800000M`3o
ool00`000000oooo0?ooo`3D0?ooo`005`3oool200000080oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0l0?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3oool0e03oool0
01L0oooo0P0000020?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0?03oool20000
07@0oooo0`00003D0?ooo`005`3oool01@000000oooo0000003oool0000000P0oooo00<000000?oo
o`3oool0?P3oool200000780oooo00<000000?ooo`3oool0e03oool001L0oooo00D000000?ooo`00
0000oooo000000080?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool0K`3oool00`00
0000oooo0?ooo`3D0?ooo`005`3oool00`000000oooo000000020?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool0@@3oool2000006l0oooo00<000000?ooo`3oool0e03oool001L0oooo
00<000000?ooo`0000000P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0480oooo
00<000000?ooo`3oool0K03oool00`000000oooo0?ooo`3D0?ooo`005`3oool00`000000oooo0000
00020?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0@`3oool2000006`0oooo00<0
00000?ooo`3oool0e03oool001L0oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`06
0?ooo`030000003oool0oooo04D0oooo0P00001Z0?ooo`030000003oool0oooo0=@0oooo000G0?oo
o`030000003oool000000080oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`160?oo
o`030000003oool0oooo04X0oooo00D0omnL04P000000000001d0;ooo`030?ooo`060?noM0000000
M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`090?ooo`030000003oool0oooo0=@0oooo
000G0?ooo`030000003oool000000080oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`170?ooo`800000B@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@0
000007Boo`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`090?ooo`030000003oool0
oooo0=@0oooo000H0?ooo`8000000P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo
04P0oooo00<000000?ooo`3oool0AP3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo
00X0oooo00<0oooO09a8M02oool02@3oool00`000000oooo0?ooo`3D0?ooo`00603oool200000080
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`190?ooo`800000A@3oool00`3ookl0
M00004RLg`030?ooo`030=nLB00007@0_ooo00/0oooo00<0_g@007Boo`3oool0203oool00`000000
oooo0?ooo`3D0?ooo`00603oool2000000<0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`1;0?ooo`800000=`3oool00`3ooml0W4P0000000070000000504RLg`3oool0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030?oog`2LB000B9cO00T0oooo00<000000?ooo`3o
ool0e03oool001P0oooo0P0000030?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
C03oool00`000000oooo0?ooo`100?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
2P3oool00`3OW4P000000000000C00000<d0oooo000H0?ooo`8000000`3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo04d0oooo0P0000100?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol01@3oool7000000030;ooo`3oool0oooo00L0oooo00<000000?ooo`3oool01`3o
ool800000<D0oooo000H0?ooo`030000003oool000000080oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`1?0?ooo`800000?P3oool00`3ogi`0B00007Boo`030?ooo`040?noM00004P0
W=oo0?ooo`@0000000<0oooo0=nLB00000000P0000000`0004P0W=oo0?ooo`090?ooo`030000003o
ool0oooo00l0oooo1@0000300?ooo`00603oool00`000000oooo000000020?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`0k0?ooo`030?oo_`1d0000
B9cO00<0oooo100000050?ooo`040?oo_`1d0000000004RLg``0oooo00<000000?ooo`3oool0503o
ool400000;`0oooo000H0?ooo`030000003oool000000080oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`1A0?ooo`800000?03oool00`3OW4P0000000000002000000030;md0000M;l0
oooo00L0oooo00<0_g@007Boo`3oool0303oool00`000000oooo0?ooo`0H0?ooo`@00000^03oool0
01P0oooo00<000000?ooo`0000000`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo
0580oooo0P00000g0?ooo`@0000000H0M00007Boo`3oool0okmd0000M02oool80?ooo`060?oOW018
000000000000000004P0W=oo2@3oool00`000000oooo0?ooo`0L0?ooo`@00000]03oool001P0oooo
00<000000?ooo`0000000`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo05@0oooo
0P00000b0?ooo`<00000103oool01@3ookl0M0000000000007@0_ooo00X0oooo00D0_g@000000000
0000000004RLg`090?ooo`030000003oool0oooo0200oooo0P00000`0?ooo`L00000N`3oool001P0
oooo00<000000?ooo`0000000`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo05H0
oooo00<000000?ooo`3oool0;03oool3000002@0oooo00<000000?ooo`3oool08P3oool3000002L0
oooo1P0000060?ooo`L00000M@3oool001P0oooo00<000000?ooo`0000000`3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo05H0oooo0P00000Y0?ooo`<000009`3oool00`000000oooo
0?ooo`0U0?ooo`<000007P3oool6000001<0oooo0`00001b0?ooo`00603oool00`000000oooo0000
00030?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool09@3oool;000002P0oooo0P00
000U0?ooo`800000:P3oool00`000000oooo0?ooo`0X0?ooo`8000006@3oool3000001`0oooo1000
000A0?ooo``00000D@3oool001P0oooo00<000000?ooo`0000000`3oool00`000000oooo0?ooo`0:
0?ooo`030000003oool0oooo00L0oooo1000000H0?ooo`8000002`3oool5000002D0oooo00<00000
0?ooo`3oool07`3oool3000002`0oooo00<000000?ooo`3oool0:P3oool2000001D0oooo0P00000S
0?ooo`8000003@3oool2000000`0oooo100000090?ooo`P000001P3oool400000380oooo000H0?oo
o`030000003oool0000000@0oooo00@000000?ooo`0000000000103oool4000000030?ooo`000000
oooo00D0oooo0P0000030?ooo`8000005@3oool200000180oooo0P00000T0?ooo`8000007@3oool2
000002l0oooo00<000000?ooo`3oool0;03oool200000100oooo0`00000W0?ooo`8000002@3oool2
00000180oooo0P0000040?ooo`<00000203oool010000000oooo0?ooo`3oool300000080oooo0`00
00020?ooo`D0000000@0oooo00000000000000000P3oool5000001h0oooo000H0?ooo`030000003o
ool0000000@0oooo00D000000?ooo`000000oooo000000020?ooo`8000000P3oool3000000@0oooo
0P0000070?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0503oool00`000000oooo
0?ooo`0S0?ooo`030000003oool0oooo01P0oooo0P00000a0?ooo`<00000;P3oool2000000`0oooo
0P00000/0?ooo`8000001@3oool2000001H0oooo00@000000?ooo`0000000000303oool4000000L0
oooo0P0000050?ooo`040000003oool0oooo0?ooo`8000000P3oool00`000000oooo0000000N0?oo
o`00603oool010000000oooo0?ooo`0000030?ooo`8000000`3oool00`000000oooo000000040?oo
o`030000003oool0000000<0oooo00<000000?ooo`3oool01`3oool2000000l0oooo0P00000H0?oo
o`800000903oool2000001H0oooo0P00000c0?ooo`030000003oool0oooo0300oooo0P0000070?oo
o`<00000<03oool00`000000oooo0?ooo`02000001T0oooo00<000000?ooo`3oool0303oool30000
00P0oooo0P0000040?ooo`030000003oool000000080oooo0P0000000`3oool00000000000020?oo
o`030000003oool0oooo01/0oooo000H0?ooo`040000003oool0oooo000000<0oooo0P0000030?oo
o`8000001P3oool200000080oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0A0?oo
o`800000=@3oool00`000000oooo0?ooo`0b0?ooo`8000000`3oool2000003@0oooo0P00000J0?oo
o`030000003oool0000000`0oooo00D000000?ooo`3oool0oooo000000060?ooo`040000003oool0
oooo000000<0oooo00D000000?ooo`000000oooo000000020?ooo`80000000@0oooo0000003oool0
00007@3oool001T0oooo00<000000?ooo`0000000`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00D0oooo00<000000?ooo`0000003P3oool00`000000oooo0?ooo`090?ooo`030000
003oool0oooo01`0oooo0P00000T0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
=@3oool00`000000oooo0?ooo`0d0?ooo`<00000=@3oool00`000000oooo0?ooo`02000001H0oooo
0P0000030?ooo`8000002P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo
00D000000?ooo`3oool0oooo000000020?ooo`060000003oool0oooo0000003oool000000P3oool0
1P000000oooo0?ooo`000000oooo000001d0oooo000I0?ooo`030000003oool0000000<0oooo00<0
00000?ooo`3oool00P3oool2000000L0oooo0P00000@0?ooo`030000003oool0oooo00H0oooo0P00
000Q0?ooo`030000003oool0oooo0280oooo0P00000=0?ooo`800000>03oool00`000000oooo0?oo
o`0d0?ooo`040000003oool00000000003<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0B0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`040?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`0000000`3oool200000080
oooo0P0000001@3oool000000?ooo`3oool0000001`0oooo000I0?ooo`030000003oool0000000<0
oooo0P0000030?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0280oooo0P00000T0?ooo`030000003oool0oooo00P0
oooo0P00000j0?ooo`030000003oool0oooo0380oooo0P0000040?ooo`800000;`3oool2000000L0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool01P3oool2000000H0oooo0P0000040?ooo`040000003oool0oooo0?oo
o`8000000P3oool00`000000oooo0000000L0?ooo`006@3oool00`000000oooo000000030?ooo`80
00000`3oool00`000000oooo000000060?ooo`8000004@3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo02D0oooo0P00000S0?ooo`8000001P3oool2000003`0oooo00<000000?ooo`3o
ool0<03oool2000000P0oooo0P00000/0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool01`3oool2000000L0oooo00<000000?ooo`3oool00P3oool2000000040?ooo`000000oooo0000
0080oooo00<000000?ooo`000000703oool001T0oooo00<000000?ooo`0000000`3oool200000080
oooo00@000000?ooo`3oool000001P3oool00`000000oooo0000000A0?ooo`050000003oool0oooo
0?ooo`000000:P3oool00`000000oooo0?ooo`0R0?ooo`040000003oool0oooo0?ooo`l00000<@3o
ool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo00X0oooo0P00000Y0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0:0?ooo`050000003o
ool0oooo0?ooo`0000002P3oool2000000H0oooo0P0000040?ooo`80000000D0oooo0000003oool0
oooo000000020?ooo`030000003oool0000001/0oooo000I0?ooo`030000003oool0000000<0oooo
0P0000020?ooo`040000003oool0oooo000000D0oooo00@000000?ooo`3oool000004P3oool01000
0000oooo0?ooo`00000[0?ooo`030000003oool0oooo0280oooo1000000>0?ooo`@00000;@3oool0
0`000000oooo0?ooo`0]0?ooo`8000003`3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0
oooo00`0oooo0P00000<0?ooo`030000003oool0oooo00/0oooo00@000000?ooo`3oool000002P3o
ool2000000H0oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`0200000080oooo00H0
00000?ooo`3oool000000?ooo`00000K0?ooo`006@3oool00`000000oooo000000020?ooo`030000
003oool000000080oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo0080
00004P3oool2000001h0oooo1000000;0?ooo`030000003oool0oooo0240oooo0`00000C0?ooo`<0
0000:P3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo0100oooo0P00000T0?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0=0?ooo`80
00002P3oool010000000oooo0?ooo`0000040?ooo`040000003oool0oooo00000080oooo00<00000
0?ooo`3oool00P0000020?ooo`060000003oool0oooo0000003oool000006`3oool001T0oooo00<0
00000?ooo`0000000P3oool00`000000oooo000000020?ooo`050000003oool0oooo0?ooo`000000
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0
0P3oool2000001H0oooo100000030?ooo`@00000203oool00`000000oooo0?ooo`0N0?ooo`800000
0`3oool2000001@0oooo0P00000X0?ooo`030000003oool0oooo02X0oooo0P00000E0?ooo`030000
003oool0oooo01l0oooo0P00000C0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
3P3oool2000000X0oooo00@000000?ooo`3oool00000103oool01`000000oooo0?ooo`3oool00000
0?ooo`0000000P3oool00`000000oooo0?ooo`020?ooo`050000003oool000000?ooo`0000006`3o
ool001T0oooo00<000000?ooo`0000000P3oool01P000000oooo0?ooo`000000oooo000000<0oooo
00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo00l0oooo00@000000?ooo`00
0000oooo1P00000B0?ooo`8000002`3oool00`000000oooo0?ooo`060?ooo`800000703oool30000
00H0oooo0P00000D0?ooo`800000803oool=000002D0oooo00<000000?ooo`3oool05P3oool00`00
0000oooo0?ooo`0M0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0000000P0oooo00<000000?ooo`3oool00P3oool01000
0000oooo0?ooo`0000040?ooo`8000000`3oool2000000<0oooo00<000000?ooo`0000000P3oool0
0`000000oooo0?ooo`0H0?ooo`006@3oool00`000000oooo000000020?ooo`030000003oool0oooo
00800000103oool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool03P3oool0
10000000oooo0000000000060?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0303o
ool2000000P0oooo00<000000?ooo`3oool05`3oool3000000X0oooo0P00000D0?ooo`800000703o
ool2000000H0oooo0`0000040?ooo`8000008P3oool00`000000oooo0?ooo`0H0?ooo`800000703o
ool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3o
ool010000000oooo0?ooo`0000080?ooo`030000003oool0oooo0080oooo00@000000?ooo`3oool0
00001@3oool010000000oooo0?ooo`3oool2000000<0oooo00<000000?ooo`0000000P3oool00`00
0000oooo0?ooo`0H0?ooo`006@3oool01P000000oooo0?ooo`000000oooo00000080oooo0P000005
0?ooo`030000003oool0000000H0oooo00<000000?ooo`3oool03@3oool01@000000oooo0000003o
ool0000000L0oooo0P00000=0?ooo`030000003oool0oooo00l0oooo0P0000070?ooo`030000003o
ool0oooo01@0oooo0P0000000`3oool000000?ooo`0<0?ooo`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool05P3oool3000000P0oooo00<000000?ooo`3oool01P3oool2000001T0oooo2P00
000J0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool01@000000oooo0?oo
o`3oool000000100oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo00<0oooo0P0000060?ooo`040000003oool0oooo0?ooo`8000000`3oool01P000000
oooo0?ooo`000000oooo000001X0oooo000I0?ooo`060000003oool0oooo0000003oool000000P3o
ool2000000D0oooo00<000000?ooo`0000001P3oool00`000000oooo0?ooo`0<0?ooo`030000003o
ool000000080oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo0180oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0B0?ooo`030000003o
ool0000000D000002`3oool2000001@0oooo00<000000?ooo`3oool04P3oool3000000/0oooo00<0
00000?ooo`3oool0203oool2000001D0oooo0P0000040?ooo`030000003oool0oooo00<0oooo0P00
000E0?ooo`H000005`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool000000140oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00@0oooo0P00
00050?ooo`050000003oool000000?ooo`0000000P3oool02@000000oooo0?ooo`3oool000000?oo
o`000000oooo0000000J0?ooo`006@3oool00`000000oooo0?ooo`02000000<0oooo0P0000050?oo
o`030000003oool0000000L0oooo00<000000?ooo`3oool02P3oool00`000000oooo000000040?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0D0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool0403oool4000000D0oooo0`00000:0?oo
o`8000004`3oool200000140oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0:0?oo
o`8000004@3oool2000000@0oooo0P0000080?ooo`800000403oool3000000D0oooo0`00000>0?oo
o`L000007P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo00D000000?oo
o`000000oooo000000020?ooo`090000003oool0oooo0?ooo`000000oooo0000003oool0000001X0
oooo000I0?ooo`030000003oool0oooo008000000`3oool2000000D0oooo0P0000080?ooo`030000
003oool0oooo00X0oooo00<000000?ooo`0000001@3oool00`000000oooo0?ooo`060?ooo`030000
003oool0oooo00H0oooo00<000000?ooo`3oool0303oool6000000@0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`0=0?ooo`@000002P3oool2000000X0oooo0P00000C0?ooo`030000
003oool0oooo00`0oooo0P00000?0?ooo`030000003oool0oooo00`0oooo0P00000=0?ooo`800000
1@3oool00`000000oooo0?ooo`0:0?ooo`800000303oool2000000X0oooo0`0000090?ooo`@00000
103oool4000000P0oooo1P00000=0?ooo`030000003oool000000100oooo00<000000?ooo`3oool0
1P3oool010000000oooo0?ooo`0000070?ooo`8000001@3oool00`000000oooo0?ooo`0200000080
oooo00T000000?ooo`3oool0oooo0000003oool000000?ooo`0000006P3oool001T0oooo00<00000
0?ooo`3oool00P0000030?ooo`8000001P3oool00`000000oooo0?ooo`060?ooo`8000002P3oool0
0`000000oooo000000060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`0;0?ooo`8000001P3oool2000000<0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0;0?ooo`<000003P3oool2000000X0oooo0P00000>0?ooo`H000002`3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0203oool2
000000D0oooo0P00000?0?ooo`800000203oool2000000h0oooo0P0000070?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`8000001P3oool2
000000X0oooo00D000000?ooo`3oool0oooo0000000>0?ooo`030000003oool0oooo00L0oooo00@0
00000?ooo`3oool000001`3oool00`000000oooo000000040?ooo`040000003oool0oooo000000@0
oooo00P000000?ooo`3oool000000?ooo`000000oooo000001X0oooo000I0?ooo`030000003oool0
oooo008000000`3oool2000000H0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0000000P0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0203oool01@000000oooo0?ooo`3o
ool0000000L0oooo00<000000?ooo`3oool02@3oool300000140oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`090?ooo`8000001@3oool4000000L0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`0?0?ooo`8000001`3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0140oooo0P0000040?ooo`8000004@3oool2000000D0oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`040?ooo`8000000P3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`0000001`3oool00`000000oooo000000030?ooo`050000003o
ool0oooo0?ooo`000000103oool010000000oooo0?ooo`3oool2000000030?ooo`000000oooo01T0
oooo000J0?ooo`040000003oool00000000000<0oooo0P0000050?ooo`8000002@3oool00`000000
oooo0?ooo`070?ooo`030000003oool0000000P0oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool010000000
oooo0?ooo`0000080?ooo`030000003oool0oooo00H0oooo0`00000D0?ooo`030000003oool0oooo
00T0oooo0P0000070?ooo`800000203oool00`000000oooo0?ooo`02000000<0oooo0P00000?0?oo
o`T000003`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0503oool010000000oooo0?ooo`00000D0?ooo`<000000P3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo00L0oooo0P00000<0?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo0P000007
0?ooo`050000003oool0oooo0?ooo`0000000P3oool010000000oooo0?ooo`3oool2000000<0oooo
00@000000?ooo`3oool0oooo0P0000000`3oool000000?ooo`0I0?ooo`006P3oool010000000oooo
0000000000030?ooo`8000001@3oool2000000T0oooo0P0000070?ooo`030000003oool0000000X0
oooo00<000000?ooo`3oool01`3oool010000000oooo0?ooo`00000<0?ooo`030000003oool0oooo
00`0oooo0P0000000`3oool000000?ooo`060?ooo`030000003oool0oooo00D0oooo0`00000F0?oo
o`8000002`3oool00`000000oooo0?ooo`020?ooo`8000002`3oool200000080oooo00<000000?oo
o`0000003`3oool2000000@0oooo00<000000?ooo`3oool00P3oool2000000h0oooo00<000000?oo
o`3oool00P0000060?ooo`030000003oool0oooo01H0oooo0P00000G0?ooo`<000001`3oool00`00
0000oooo0?ooo`080?ooo`8000003@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00H0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
00D0oooo00D000000?ooo`3oool0oooo000000020?ooo`040000003oool0oooo0?ooo`800000103o
ool01@000000oooo0?ooo`000000oooo008000006P3oool001X0oooo00@000000?ooo`0000000000
0`3oool2000000D0oooo0P00000:0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000
2P3oool00`000000oooo0?ooo`070?ooo`030000003oool0000000`0oooo00<000000?ooo`3oool0
3`3oool00`000000oooo000000070?ooo`030000003oool0oooo00<0oooo0`00000I0?ooo`030000
003oool0oooo00T0oooo0P0000000`3oool000000?ooo`0>0?ooo`030000003oool0oooo00800000
3@3oool2000000H0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0<0?ooo`800000
1`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0000001L0oooo00<000000?ooo`000000
1@3oool00`000000oooo0?ooo`080?ooo`040000003oool0oooo000000d0oooo00D000000?ooo`3o
ool0oooo0000000:0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool02P3oool20000
00L0oooo00<000000?ooo`3oool00P3oool2000000<0oooo00<000000?ooo`000000103oool01@00
0000oooo0?ooo`000000oooo008000006P3oool001X0oooo00@000000?ooo`00000000000P3oool0
0`000000oooo000000050?ooo`030000003oool0000000T0oooo00<000000?ooo`3oool0103oool0
10000000oooo0?ooo`00000;0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`0A0?ooo`030000003oool0000000L0oooo00@000000?ooo`3oool0oooo
0P00000L0?ooo`030000003oool0oooo00T0oooo0P00000A0?ooo`8000000P3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01@3oool2000000`0oooo
00<000000?ooo`000000103oool2000001T0oooo00D000000?ooo`3oool0oooo0000000E0?ooo`03
0000003oool0000000800000103oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool02`3oool00`000000oooo0000000;0?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool02P3oool00`000000oooo000000060?ooo`030000003oool0oooo0080
oooo0P0000030?ooo`030000003oool0000000@0oooo00D000000?ooo`3oool000000?ooo`020000
01X0oooo000J0?ooo`030000003oool0000000<0oooo00@000000?ooo`3oool000000`3oool01000
0000oooo0?ooo`00000:0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`000000303oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool04P3oool0
0`000000oooo000000080?ooo`040000003oool0oooo000001h0oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`02000000l0oooo0P0000030?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`8000002@3oool01@000000oooo0?oo
o`3oool000000080oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0100oooo0P0000040?ooo`050000003oool0oooo0?ooo`0000002@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00/0oooo00<000000?ooo`000000303oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00/0oooo00<000000?ooo`0000001@3oool00`000000
oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000000P3oool01@000000oooo0?ooo`3oool0
00000080oooo00@000000?ooo`00000000006P3oool001X0oooo00@000000?ooo`00000000000P3o
ool010000000oooo0?ooo`0000030?ooo`040000003oool0oooo000000X0oooo00<000000?ooo`3o
ool00P3oool010000000oooo0?ooo`00000=0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`0000002@3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0000000P0oooo0P00000P0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool2000000`0oooo00@000000?oo
o`3oool000000`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00P0oooo0`000009
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool2000001X0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`0000002@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00/0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00X0oooo00D00000
0?ooo`3oool0oooo000000040?ooo`030000003oool0oooo00<0oooo0P0000030?ooo`040000003o
ool0oooo000000@0oooo00<000000?ooo`3oool00P0000000`3oool000000?ooo`0H0?ooo`006P3o
ool010000000oooo0000000000020?ooo`040000003oool0oooo000000<0oooo00@000000?ooo`3o
ool000002P3oool2000000<0oooo00<000000?ooo`0000003P3oool00`000000oooo0?ooo`040?oo
o`040000003oool0oooo000000T0oooo00<000000?ooo`3oool05@3oool00`000000oooo00000007
0?ooo`8000008@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000001@3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`040?ooo`8000006@3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool02`3oool2000000d0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0;0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo
0?ooo`030?ooo`8000000`3oool010000000oooo0?ooo`0000040?ooo`030000003oool0oooo0080
000000<0oooo0000003oool0603oool001X0oooo00@000000?ooo`00000000000P3oool010000000
oooo0?ooo`0000030?ooo`040000003oool0oooo000000/0oooo00@000000?ooo`3oool000000P3o
ool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo
000000070?ooo`030000003oool0oooo01H0oooo00@000000?ooo`3oool00000103oool200000080
oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00L0
oooo0P0000070?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo000000030?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool01@000000oooo0?ooo`3oool0
000000H0oooo00@000000?ooo`3oool000005`3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool01P3oool00`000000oooo000000080?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0000000>0?ooo`050000003o
ool0oooo0?ooo`000000303oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo000000H0
oooo0P0000020?ooo`050000003oool0oooo0?ooo`000000103oool01@000000oooo0?ooo`000000
oooo008000006@3oool001T0oooo00<0_g@00000000000000P0000000`3Oool0oooo000000020?oo
o`050000003oool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`080?ooo`040000003oool0
oooo00000080oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0D0?ooo`090?oOW018
000000000000M0000000oooo0000003oool0okmd00@0000000D0W4P0000000000000000007Boo`0O
0?ooo`050000003oool0oooo0?ooo`000000303oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00H0oooo00L000000?ooo`3oool0oooo0000003oool0000000d0oooo00<000000?ooo`3o
ool0303oool00`000000oooo000000060?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0503oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`03000000030000M02oool0oooo0080oooo0P000000101d_ol0
oooo0?oog`2LB0030000000407Boo`3oool0oooo000000T0oooo00D000000?ooo`3oool0oooo0000
000>0?ooo`030000003oool0000000d0oooo00<000000?ooo`3oool00`3oool010000000oooo0?oo
o`0000060?ooo`050000003oool000000?ooo`000000103oool01@000000oooo0?ooo`3oool00000
0080oooo00@000000?ooo`0000000000103oool01P2oM000000000000000000004RL0=ooo`l0oooo
000J0?ooo`0600000018000000000000003oool000000`3oool010000000oooo0?ooo`0000040?oo
o`030000003oool0oooo00P0oooo0`0000020?ooo`030000003oool0oooo00l0oooo00D000000?oo
o`3oool0oooo000000050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool04`3oool0
403ookl0M00007Boo`3oool0okmd0000M0000000oooo0000003o_g@0000007Boo`000000oooO09a8
M02oool20?ooo`030000001d_ol0oooo01h0oooo00<000000?ooo`0000003P3oool200000080oooo
00<000000?ooo`3oool0203oool01@000000oooo0?ooo`3oool0000000h0oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool04P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool01P3oool04`000000oooo0?ooo`3ookl0M0000000003oool000000000M02oool0oooo0000003o
_g@0000007Boo`3oool0oooO09a8M02oool00P3oool0102oM000M;oo0?ooo`0000090?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool02`3oool00`000000oooo0000000=0?ooo`030000
003oool0oooo00<0oooo00@000000?ooo`3oool000001@3oool01P000000oooo0?ooo`000000oooo
000000@0oooo00D000000?ooo`3oool0oooo000000020?ooo`040000003oool00000000000D0oooo
00<0omnL04P0001d_ol04@3oool001X0oooo00H0000004P00000000000000?ooo`0000040?ooo`03
0000003oool0000000@0oooo00<000000?ooo`3oool02@3oool200000080oooo00<000000?ooo`3o
ool03`3oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0503oool00`2oM00004RL0=ooo`020?ooo`070;md00000000oooo0?ooo`000000oooo0000
00060?ooo`030?oog`000000_ooo01l0oooo00<000000?ooo`3oool03`3oool2000000`0oooo00D0
00000?ooo`000000oooo0000000=0?ooo`030000003oool0oooo00`0oooo0P0000040?ooo`800000
203oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`050?ooo`0=0000003oool0oooo0?ooo`2oM00004RL0=ooo`000000
oooo00000000M;l0oooo000000090?ooo`050?oog`2LB7@0_ooo0?ooo`0000001`3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`050?ooo`030000003oool0000000D0oooo00H000000?ooo`3oool000000?ooo`000004
0?ooo`050000003oool0oooo0?ooo`0000000P3oool010000000oooo0000000000050?ooo`030?oO
W0180000M;oo0140oooo000J0?ooo`0600000018000000000000003oool00000103oool00`000000
oooo000000040?ooo`030000003oool0oooo00T0oooo0P0000020?ooo`030000003oool0oooo0100
oooo00<000000?ooo`0000001`3oool01@000000oooo0?ooo`3oool0000001D0oooo00<0oono07@0
0018W=l00`3oool00`3OW4P0001d00000004000000L0oooo00<0000007Boo`3oool07@3oool00`00
0000oooo0000000@0?ooo`<00000303oool010000000oooo0?ooo`00000=0?ooo`030000003oool0
oooo00/0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00/000000?ooo`3oool0oooo0?oo_`1d0000B9cO0?ooo`000000oooo0=nLB002
000000/0oooo00@0_g@007Boo`3oool000001`3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo00/0oooo0P00000<0?ooo`030000003oool0oooo00D0oooo0P0000060?ooo`030000003o
ool0oooo008000001@3oool00`000000oooo0?ooo`020?ooo`040000003oool000000?ooo`800000
1@3oool00`3ogi`0B00007Boo`0A0?ooo`002P3oool00`3ooml0W4P0000000070000000304RLg`3o
ool0oooo00<0oooo00H0000004P00000000000000?ooo`0000040?ooo`030000003oool0000000D0
oooo00<000000?ooo`3oool0203oool2000000030?ooo`000000oooo01<0oooo00<000000?ooo`3o
ool01P3oool010000000oooo0?ooo`0000090?ooo`030?oog`2LB000000000L0000000H0B9cO0?oo
o`3oool0omnL04P0001d_ol30?ooo`040?noM000000000000?ooo`8000001`3oool00`3ooml0W4P0
0000000N0?ooo`040000003oool0oooo000000h0oooo00D000000?ooo`3oool0oooo0000000:0?oo
o`030000003oool000000080oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?oo
o`070000003oool0oooo0?ooo`000000oooo0000000=0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool0403oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool00P3oool00`3ogi`0B00007Boo`020?ooo`030000003o_g@00018008000002@3o
ool00`3ooml0W4P004RLg`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01P3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0000000/0oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<000001@3oool00`3ogi`0B00007Bo
o`0A0?ooo`006P3oool01P000000B000000000000000oooo000000@0oooo0P0000060?ooo`030000
003oool0oooo00L0oooo00<000000?ooo`0000000P00000D0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`0000005P3oool00`3ogi`0B00007Boo`030?ooo`030000000004P0W=oo00800000
00<0oooo0000003oool01@3oool0103OW4P0000007Boo`00000L0?ooo`050000003oool0oooo0?oo
o`0000003@3oool00`000000oooo0?ooo`030?ooo`8000001`3oool01@000000oooo0?ooo`3oool0
00000080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo00<0oooo00<0omnL04P0001d_ol00`3oool20000000309cOo`3oool0000000P0oooo00<0gia8
0000001d_ol00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool02@3oool010000000oooo0?ooo`00000:0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00D0
00000?ooo`3oool0oooo000000020?ooo`<000001@3oool00`3ogi`0B00007Boo`0A0?ooo`006P3o
ool01P000000B000000000000000oooo000000@0oooo0P0000060?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`0000000P00000D0?ooo`800000203oool2000001L0oooo00<0omnL04P0001d
_ol00P3oool00`000000okmd0000B00200000080oooo00`000000?ooo`3oool0gmoO0?ooo`3ookl0
M0000000000007@0_ooo0?ooo`00000J0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool02@3oool00`000000oooo0?ooo`060?ooo`800000103oool00`000000oooo0?ooo`030?ooo`04
0000003oool0oooo000000X0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`020?oo
o`030000003oool0000000h0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0B0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000002Lgol00P3oool02`000000oooo0?ooo`3oool0
gmoO0?ooo`3ookl0M0000000000007@0_ooo00<0oooo00D000000?ooo`3oool0oooo0000000:0?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`000000103oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo000000020?ooo`80000000<0oooo
0000003oool00`3oool00`3ogi`0B00007Boo`0A0?ooo`006P3oool0103ogi`000000000003oool2
000000@0oooo0P0000060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`0000000P00000C
0?ooo`030000003oool0000000T0oooo00<000000?ooo`3oool05@3oool02`3ogi`0B00007Boo`3o
ool000000?ooo`3o_g@000180000003oool000000080oooo00<000000?ooo`3OW4P00`0000000`00
04P0W=oo0?ooo`020?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo00T0oooo00D000000?ooo`3oool0oooo00000006
0?ooo`040000003oool0oooo000000T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`020?ooo`040000003oool0oooo000000l0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`0D0?ooo`050000003oool0oooo0?ooo`0000001P3oool00`000000oooo0?ooo`040?ooo`03
0?oOW0180000M;oo00<0oooo0P0000001@2Lgol0oooo0?ooo`3oool000000080oooo00H0gia80000
0000000000000000B02Lgol40?ooo`040000003oool0oooo000000/0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo000000030?ooo`030000003oool0oooo0080oooo00<000000?ooo`000000
1@3oool01@000000oooo0?ooo`3oool000000080oooo0P0000000`3oool000000?ooo`030?ooo`03
0?oOW0180000M;oo0140oooo000J0?ooo`040?oOW0000000M;oo0?ooo`8000001@3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00H0oooo00@000000?ooo`3oool0oooo0P00000C0?oo
o`040000003oool0oooo000000L0oooo0P00000G0?ooo`040?oo_`1d0000B9cO00000080oooo00<0
gia80000000000000P3oool020000000oooo0?ooo`000000oono07@000000000B9cO1P3oool00`00
0000oooo0?ooo`0D0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`0:0?ooo`<00000203oool00`000000oooo000000090?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool00000
00H0oooo00<000000?ooo`3oool01@3oool00`3ookl0M00004RLg`020?ooo`040000003OW4P0001d
000000@0oooo00H000000?ooo`3ookl0M00000000018W=l70?ooo`030000003oool0000000`0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool01`3oool00`000000oooo000000030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`0000001@3oool01@000000oooo0?ooo`3oool000000080oooo0P0000000`3oool0
00000?ooo`030?ooo`030?oOW0180000M;oo0140oooo000J0?ooo`050?oOW0000000M;oo0?ooo`00
00001P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00L000000?ooo`3o
ool0oooo0000003oool000000140oooo00D000000?ooo`3oool0oooo000000070?ooo`030000003o
ool0000001L0oooo00<0gia80000003Oool00P3oool2000000<0oooo00L000000?ooo`3oool0oooo
0000002oM000M;oo00P0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`070?ooo`03
0000003oool0000000P0oooo0`0000070?ooo`050000003oool0oooo0?ooo`0000001P3oool00`00
0000oooo0?ooo`0=0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool05P3oool20000
00H0oooo00<000000?ooo`3oool01`3oool02@3OW4P004RL0=ooo`3oool000000;md0000M;l0oooo
000000040?ooo`040000003oool0_g@007Boo`T0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`070?ooo`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`020?oo
o`030000003oool0000000H0oooo00H000000?ooo`3oool0oooo0000003oool2000000030?ooo`00
0000oooo00<0oooo00<0omnL04P0001d_ol04@3oool001P0oooo00L0oooO09a800000000000007Bo
o`3oool0000000D0oooo0P0000070?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
0P0000000`3oool000000?ooo`0@0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool010000000oooo0?ooo`00000G0?ooo`040000001d0000M;oo0?ooo`80000000<0_ooo0?oo
o`3oool00P3oool01P000000oooo0?ooo`000000omnL04P00080000000<0001809cOo`3oool0103o
ool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0=0?ooo`040000003oool00000000000P0oooo00<000000?ooo`000000
1`3oool00`000000oooo0?ooo`060?ooo`050000003oool0oooo0?ooo`0000001`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0603oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00P0oooo00T0oono07@0001d_ol000000?noM00007@0
_ooo0?ooo`0000001@3oool00`000000omnL04P00002000000030000B02Lgol0oooo00<0oooo0P00
000=0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00L0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000
003oool0oooo0080oooo00<000000?ooo`0000001P3oool020000000oooo0?ooo`3oool000000?oo
o`000000oooo0P0000020?ooo`050?oog`2LB00000000000001d_ol04@3oool001X0oooo00D0okmd
0000001d_ol0oooo000000050?ooo`8000001`3oool00`000000oooo0?ooo`040?ooo`090000003o
ool0oooo0?ooo`000000oooo0000003oool0000000l0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000005@3oool010000000oooo0?oo_`1d
0002000000030;ooo`3oool0oooo00<0oooo00H000000?ooo`3oool0oooo0000002oM00300000003
04RLg`3oool0oooo00<0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0;0?ooo`05
0000003oool0oooo0?ooo`000000403oool010000000oooo0?ooo`3oool2000000L0oooo00<00000
0?ooo`0000001P3oool00`000000oooo0?ooo`050?ooo`050000003oool0oooo0?ooo`0000002@3o
ool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool05`3o
ool00`000000oooo000000030?ooo`030000003oool0oooo00X0oooo00D0oono07@000000000001d
0;ooo`030?ooo`030000003oool0oooo00<0oooo00<000000;md000000000P0000000`18W=l0oooo
0?ooo`020?ooo`030000003oool0000000d0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`070?ooo`050000003oool0oooo0?ooo`0000002P3oool00`000000oooo0?ooo`020?ooo`80
0000103oool01@000000oooo0?ooo`3oool0000000D0oooo00P000000?ooo`3oool0oooo0000003o
ool000000?ooo`800000103oool00`3o_g@0000007Boo`0A0?ooo`006`3oool010000000oooo0?oo
o`0000050?ooo`8000001`3oool00`000000oooo0?ooo`040?ooo`090000003oool0oooo0?ooo`00
0000oooo0000003oool0000000l0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3o
ool0000000@0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`020?ooo`8000001`3o
ool01@000000oooo0?ooo`3oool0000000X0oooo00<000000?ooo`3oool03P3oool00`000000oooo
0?ooo`0<0?ooo`050000003oool0oooo0?ooo`0000003`3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo00D0oooo0P0000060?ooo`030000003oool0oooo00D0oooo00@000000?ooo`3o
ool000002`3oool00`000000oooo0?ooo`0=0?ooo`050000003oool0oooo0?ooo`0000006@3oool0
1`000000oooo0?ooo`3oool000000?ooo`0000003P3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool010000000oooo0?ooo`00000=0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool02@3oool010000000oooo0?ooo`0000090?oo
o`030000003oool0oooo00<0oooo0P0000040?ooo`050000003oool0oooo0?ooo`0000001@3oool0
1@000000oooo0?ooo`3oool000000080oooo0`00000H0?ooo`006`3oool010000000oooo0?ooo`00
00050?ooo`030000003oool0000000H0oooo00<000000?ooo`3oool00`3oool01@000000oooo0?oo
o`3oool000000080oooo00@000000?ooo`3oool000003@3oool00`000000oooo0?ooo`050?ooo`05
0000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
0080oooo0P0000080?ooo`050000003oool0oooo0?ooo`0000002`3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`0000003`3oool00`000000oooo0?ooo`06
0?ooo`8000001P3oool2000000D0oooo00<000000?ooo`3oool0103oool010000000oooo00000000
000<0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`0000006P3oool010000000oooo0?oo
o`3oool2000000l0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0:0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool02P3oool00`000000oooo000000090?ooo`030000
003oool0oooo00<0oooo0P0000040?ooo`050000003oool0oooo0?ooo`0000001@3oool01@000000
oooo0?ooo`3oool000000080oooo0`00000H0?ooo`006`3oool010000000oooo0?ooo`0000050?oo
o`030000003oool0000000L0oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0
00000080oooo00@000000?ooo`3oool000003@3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0000000H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`030?ooo`8000002@3o
ool01@000000oooo0?ooo`3oool0000000/0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0?0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0203oool2000000@0oooo
00<000000?ooo`000000103oool00`000000oooo0?ooo`030?ooo`030000003oool0000000l0oooo
00<000000?ooo`3oool03P3oool2000001T0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool02`3oool010000000oooo
0?ooo`00000=0?ooo`8000002P3oool00`000000oooo0?ooo`030?ooo`800000103oool01@000000
oooo0?ooo`3oool0000000H0oooo00@000000?ooo`3oool000000P3oool3000001P0oooo000K0?oo
o`040000003oool0oooo000000D0oooo00<000000?ooo`0000001`3oool00`000000oooo0?ooo`02
0?ooo`050000003oool0oooo0?ooo`0000000`3oool010000000oooo0?ooo`00000;0?ooo`030000
003oool0oooo00L0oooo00<000000?ooo`0000001`3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo0080oooo0P00000;0?ooo`050000003oool0oooo0?ooo`0000002`3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo00h0oooo0P00000?0?ooo`030000003oool0oooo00/0
oooo0P0000030?ooo`800000103oool00`000000oooo0?ooo`020?ooo`030000003oool000000140
oooo00<000000?ooo`3oool03@3oool2000001T0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0000000<0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0103oool01@000000
oooo0?ooo`3oool0000000H0oooo00<000000?ooo`3oool02P3oool010000000oooo0?ooo`00000>
0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0103oool2000000<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo00000080oooo0`00
000H0?ooo`006`3oool00`000000oooo0?ooo`02000000<0oooo00@000000?ooo`3oool000001`3o
ool01@000000oooo0?ooo`3oool0000000<0oooo00<000000?ooo`3oool00P3oool010000000oooo
0?ooo`00000;0?ooo`030000003oool0oooo00P0oooo0P0000070?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0000000;0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`0000003P3oool00`000000oooo0?ooo`0=0?ooo`040000003oool0oooo0?oo
o`8000000`3oool00`000000oooo0?ooo`020?ooo`8000004P3oool00`000000oooo0?ooo`0<0?oo
o`030000003oool0000001P0oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0
000000/0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`040?ooo`030000003oool0
000000L0oooo00<000000?ooo`3oool02`3oool2000000h0oooo00<000000?ooo`000000203oool0
0`000000oooo0?ooo`040?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00H000000?ooo`3oool0oooo0000003oool3000001P0oooo000K0?ooo`030000003o
ool0oooo008000000`3oool010000000oooo0?ooo`0000070?ooo`050000003oool0oooo0?ooo`00
00000`3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000002P3oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0303oool00`00
0000oooo0?ooo`020?ooo`030000003oool0000000d0oooo00D000000?ooo`3oool0oooo0000000;
0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool03`3oool01@000000oooo0?ooo`3o
ool0000000`0oooo00<000000?ooo`3oool03`3oool200000080oooo0P0000020?ooo`040000003o
ool0oooo0?ooo`800000503oool00`000000oooo0?ooo`0:0?ooo`050000003oool0oooo0?ooo`00
00005P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00D0oooo0P0000080?ooo`03
0000003oool0oooo00X0oooo0P00000>0?ooo`030000003oool0000000L0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo000000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool01P000000oooo0?ooo`3oool000000?ooo`80000000<0oooo0000003oool05P3oool0
01/0oooo00D000000?ooo`000000oooo000000030?ooo`050000003oool0oooo0?ooo`0000001P3o
ool01@000000oooo0?ooo`3oool0000000<0oooo00<000000?ooo`3oool00`3oool010000000oooo
0?ooo`0000090?ooo`030000003oool0oooo00T0oooo0P0000080?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool00`3oool2000000h0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`090?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool04P3oool200000005
0?ooo`00000000000?ooo`0000000P3oool2000001H0oooo00<000000?ooo`3oool02@3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo01<0oooo00D000000?ooo`3oool0oooo00000006
0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`04
0?ooo`800000203oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00d0oooo00@00000
0?ooo`3oool000001P3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo00000080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`060000003oool0oooo0?ooo`00
0000oooo0P0000000`3oool000000?ooo`0F0?ooo`006`3oool01@000000oooo0000003oool00000
00<0oooo00D000000?ooo`3oool0oooo000000070?ooo`030000003oool0000000<0oooo00<00000
0?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000P0oooo00<000000?ooo`3oool02@3o
ool2000000P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0000000l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool0503oool3000000030?ooo`000000oooo00800000603o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool04@3o
ool01@000000oooo0?ooo`3oool0000000P0oooo00<000000?ooo`3oool01@3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool00000203oool00`000000
oooo0?ooo`090?ooo`8000003@3oool01@000000oooo0?ooo`3oool0000000H0oooo00<000000?oo
o`3oool0103oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool01P000000oooo0?ooo`3oool000000?ooo`80000000<0oooo0000003oool0
5P3oool001/0oooo00D000000?ooo`000000oooo000000030?ooo`050000003oool0oooo0?ooo`00
00001`3oool00`000000oooo000000030?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3o
ool0oooo000000070?ooo`030000003oool0oooo00T0oooo00@000000?ooo`3oool00000203oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00<0oooo00<000000?ooo`000000403oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`0G0?ooo`D000006@3oool00`000000oooo0?ooo`060?ooo`030000003oool0
oooo00H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0=0?ooo`050000003oool0
oooo0?ooo`000000103oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`0000003@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool01@3oool010000000oooo0?ooo`3oool2000000L0oooo00<000000?ooo`3oool0
0`3oool01@000000oooo0?ooo`000000oooo0080000000<0oooo0000003oool05P3oool001/0oooo
0`0000001@3oool000000?ooo`3oool0000000@0oooo00<000000?ooo`3oool01@3oool00`000000
oooo000000030?ooo`030000003oool0oooo00D0oooo00D000000?ooo`3oool0oooo000000060?oo
o`030000003oool0oooo00T0oooo00@000000?ooo`3oool00000203oool00`000000oooo0?ooo`06
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`0000004@3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00X0oooo00D000000?ooo`3oool0oooo0000000@0?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0J0?ooo`8000006`3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool03`3o
ool01@000000oooo0?ooo`3oool0000000X0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0?0?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`070?ooo`040000003oool0oooo000000/0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00D0oooo00@000000?ooo`3oool0oooo0P000007
0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool000000?ooo`02000000030?ooo`00
0000oooo01H0oooo000;0?ooool00000WP0000080?ooo`005@3oool00`000000oooo0?ooo`020?oo
o`060000003oool000000000003oool000000P3oool00`000000oooo0?ooo`030?ooo`8000001@3o
ool2000000<0oooo00<000000?ooo`3oool01@3oool2000000@0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`060?ooo`040000003oool0oooo000000@0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo000000030?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool0
00001@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0080oooo0P00000;0?ooo`03
0000003oool0000000@0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`05
0000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
0100oooo1@00000@0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool00P3oool01@00
0000oooo0?ooo`3oool0000000P0oooo00<000000?ooo`3oool0203oool010000000oooo0?ooo`00
00030?ooo`030000003oool0oooo00T0oooo00@000000?ooo`3oool000000P3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01`3oool01@000000oooo
0?ooo`3oool000000080oooo00<000000?ooo`3oool01@3oool01@000000oooo0?ooo`3oool00000
00<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`050?ooo`060000003oool0oooo
0000003oool000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`000000103oool010000000oooo0?ooo`0000020?ooo`030000003oool0000000@0oooo
00<000000?ooo`3oool0403oool001D0oooo00<000000?ooo`3oool00P3oool010000000oooo0000
000000020?ooo`030000003oool0000000D0oooo00<000000?ooo`3oool01@3oool01@000000oooo
0?ooo`3oool0000000T0oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`060?ooo`050000003oool0oooo
0?ooo`0000001`3oool00`000000oooo000000060?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool02`3oool01@000000oooo0?ooo`3oool0000001/0oooo00@000000?ooo`000000oooo
1000000I0?ooo`040000003oool0oooo000000h0oooo00<000000?ooo`3oool02P3oool010000000
oooo0?ooo`00000?0?ooo`030000003oool0000000h0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo000000070?ooo`030000003oool0oooo0080oooo0P0000080?ooo`030000003oool0oooo
0080oooo00@000000?ooo`3oool000000P3oool3000000@0oooo00<000000?ooo`3oool0403oool0
01X0oooo00@000000?ooo`00000000000P3oool00`000000oooo000000050?ooo`030000003oool0
oooo00@0oooo0P0000030?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool00P3oool0
10000000oooo0?ooo`00000;0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool01P3o
ool010000000oooo0?ooo`0000070?ooo`040000003oool0oooo000001D0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`080?ooo`030000003oool0000000h0oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0000000J0?ooo`8000000P3oool01`000000oooo0000003oool00000
0?ooo`0000006@3oool200000100oooo00<000000?ooo`3oool0203oool010000000oooo0?ooo`00
000A0?ooo`8000004`3oool00`000000oooo0000000:0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01P3oool00`000000oooo000000070?ooo`030000003oool0oooo0080oooo0P00
00080?ooo`030000003oool0oooo0080oooo00H000000?ooo`3oool0oooo0000003oool3000001L0
oooo000J0?ooo`040000003oool0000000000080oooo00<000000?ooo`0000001@3oool00`000000
oooo0?ooo`040?ooo`8000000`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0080
oooo00@000000?ooo`3oool000002`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
00H0oooo0P0000070?ooo`040000003oool0oooo000001L0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`070?ooo`040000003oool0oooo000000`0oooo00<000000?ooo`3oool03P3o
ool2000001X0oooo0P0000030?ooo`030000003oool0oooo008000000P3oool00`000000oooo0?oo
o`0F0?ooo`800000403oool00`000000oooo0?ooo`070?ooo`040000003oool0oooo00000180oooo
00<000000?ooo`3oool04@3oool010000000oooo0?ooo`00000;0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`0000001`3oool00`000000oooo0?ooo`020?ooo`060000003oool0oooo0?oo
o`000000oooo0`00000G0?ooo`006P3oool010000000oooo0000000000020?ooo`8000001`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool000000080oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`020?ooo`800000303oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo00H0oooo00<000000?ooo`3oool01@3oool010000000oooo0?ooo`00000H0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0000000<0oooo00@000000?ooo`3oool0oooo0P0000020?ooo`8000005P3oool200000140oooo
00<000000?ooo`3oool01P3oool00`000000oooo0000000C0?ooo`8000004P3oool01@000000oooo
0?ooo`3oool0000000/0oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`0000080?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`000000203oool01@000000oooo0?ooo`3oool0
000000<0oooo00<000000?ooo`0000000P00000G0?ooo`006P3oool00`000000oooo000000030?oo
o`8000001`3oool00`000000oooo0?ooo`020?ooo`030000003oool000000080oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`020?ooo`8000002`3oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo00D0oooo0P0000060?ooo`040000003oool0oooo000001T0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0000001L0oooo00<00000
0?ooo`000000103oool01`000000oooo0?ooo`3oool000000?ooo`0000000`3oool00`000000oooo
0?ooo`0B0?ooo`040000003oool0oooo00000140oooo00<000000?ooo`3oool0103oool00`000000
oooo0000000C0?ooo`040000003oool0oooo00000100oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00@000000?ooo`3oool00000203oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo00T0oooo0P0000060?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`000000203oool00`000000oooo0?ooo`020?ooo`050000003oool0
oooo0000003oool00`00000G0?ooo`006P3oool00`000000oooo000000030?ooo`8000001`3oool0
0`000000oooo0?ooo`020?ooo`060000003oool0oooo0000003oool000002`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`040?ooo`030000003oool0000000D0oooo00@000000?ooo`3oool000006P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`050?ooo`030000003oool0oooo00l0oooo00@000000?ooo`3oool000005P3oool00`00
0000oooo000000050?ooo`030000003oool0oooo0080oooo00<000000?ooo`0000000`3oool20000
0180oooo00@000000?ooo`3oool000004P3oool00`000000oooo0?ooo`020?ooo`030000003oool0
000001@0oooo00@000000?ooo`3oool000003`3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool00P3oool2000000X0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`080?ooo`030000003oool0000000H0oooo00<000000?ooo`3oool00P3o
ool01@000000oooo0?ooo`3oool0000000L0oooo00<000000?ooo`3oool00P3oool01@000000oooo
0?ooo`000000oooo0080000000<0oooo0000003oool05@3oool001X0oooo00<000000?ooo`000000
0`3oool2000000L0oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0000000`0
oooo00@000000?ooo`3oool0oooo0P00000;0?ooo`030000003oool0oooo00P0oooo00<000000?oo
o`3oool00P3oool01@000000oooo0?ooo`3oool0000000<0oooo00D000000?ooo`3oool0oooo0000
000K0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0403oool01@000000oooo0?oo
o`3oool0000001@0oooo00<000000?ooo`0000001P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0000000@0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo0100oooo00@000000?ooo`3oool0oooo0P00000D0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00X0
oooo00@000000?ooo`3oool0oooo0P00000:0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool02@3oool00`000000oooo000000050?ooo`030000003oool0oooo00<0oooo00D000000?oo
o`3oool0oooo000000070?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool000000?oo
o`02000000030?ooo`000000oooo01D0oooo000J0?ooo`030000003oool0000000<0oooo0P000008
0?ooo`040000003oool0oooo000000<0oooo0P00000=0?ooo`070000003oool0oooo0?ooo`000000
oooo000000090?ooo`030000003oool0oooo00T0oooo00D000000?ooo`3oool0oooo000000040?oo
o`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0
oooo00<0oooo00@000000?ooo`3oool000002`3oool01@000000oooo0?ooo`3oool000000180oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0000000L0oooo
00<000000?ooo`3oool00`3oool010000000oooo0?ooo`0000040?ooo`8000003@3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P00000D0?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool00P00000;0?ooo`050000003oool0oooo0?oo
o`0000002`3oool010000000oooo0?ooo`0000040?ooo`030000003oool0oooo00<0oooo00D00000
0?ooo`3oool0oooo000000070?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool00000
0?ooo`02000000030?ooo`000000oooo01D0oooo000J0?ooo`030000003oool0000000@0oooo00<0
00000?ooo`3oool01P3oool010000000oooo0?ooo`0000030?ooo`8000003P3oool010000000oooo
0?ooo`0000020?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool02P3oool00`000000
oooo000000060?ooo`030000003oool0000000<0oooo00<000000?ooo`3oool06`3oool00`000000
oooo0?ooo`040?ooo`030000003oool0000000`0oooo00<000000?ooo`0000004P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00h0oooo0P0000020?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool0103oool00`000000oooo000000060?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0@0?ooo`80000000<0oooo0000003o
ool04`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`0:0?ooo`040000003oool00000000000/0oooo00D000000?oo
o`3oool0oooo0000000:0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01@3oool01@000000oooo0?oo
o`3oool000000080oooo00H000000?ooo`00000000000?ooo`00000F0?ooo`006P3oool010000000
oooo0000000000020?ooo`800000203oool00`000000oooo000000050?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`0000000`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
00/0oooo0P0000070?ooo`8000000`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo
00<0oooo0P00000>0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`0<0?ooo`050000003oool0oooo0?ooo`000000203oool00`000000oooo0?ooo`05
0?ooo`030000003oool0000000H0oooo0P0000090?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool0403oool2000001@0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02`3oool010000000oooo0?ooo`00
000:0?ooo`040000003oool0oooo000000/0oooo00<000000?ooo`3oool00P3oool010000000oooo
0?ooo`0000050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool01@000000
oooo0?ooo`3oool000000080oooo00H000000?ooo`00000000000?ooo`00000F0?ooo`006P3oool0
10000000oooo0000000000020?ooo`800000203oool00`000000oooo000000050?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`000000103oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo00/0oooo0P0000080?ooo`050000003oool0oooo0?ooo`0000007`3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo00`0oooo0P00000A0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool02P3oool01@000000oooo0?ooo`3oool0000000T0oooo0`0000050?ooo`04
0000003oool0oooo000000L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo00h0oooo0`00000D0?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00@000000?oo
o`3oool000002`3oool00`000000oooo0000000:0?ooo`030000003oool0oooo00<0oooo00@00000
0?ooo`3oool000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo00D0
00000?ooo`3oool0oooo000000020?ooo`060000003oool0oooo0000003oool000005P3oool001X0
oooo00@000000?ooo`00000000000P3oool2000000T0oooo0P0000040?ooo`8000003`3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02P3oool01000
0000oooo0?ooo`0000060?ooo`8000000P3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`000000303oool010000000oooo0?ooo`00000?0?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool0203oool01@000000oooo0?ooo`3oool0000000X0oooo
00<000000?ooo`3oool01P3oool010000000oooo0?ooo`0000070?ooo`8000001@3oool00`000000
oooo0?ooo`080?ooo`030000003oool0oooo00d0oooo0P0000020?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool02@3oool00`000000oooo000000020?ooo`030000003oool0oooo00P0
oooo0P00000;0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`0000001@3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo000000020?oo
o`060000003oool0oooo0000003oool000005P3oool001X0oooo00@000000?ooo`00000000000P3o
ool2000000T0oooo0P0000040?ooo`030000003oool0000000h0oooo00<000000?ooo`3oool01@3o
ool01@000000oooo0?ooo`3oool0000000`0oooo00D000000?ooo`3oool0oooo000000050?ooo`06
0000003oool0oooo0000003oool000008@3oool01@000000oooo0?ooo`3oool0000000<0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`070?ooo`050000003oool0oooo0?ooo`000000
2`3oool00`000000oooo0?ooo`070?ooo`040000003oool0oooo000000P0oooo0P0000020?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool0303oool01@000000oooo0?ooo`3oool00000
0180oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool0203oool01P000000oooo0?ooo`000000oooo000000/0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`040?ooo`030000003oool0000000D0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`050?ooo`050000003oool0oooo0?ooo`000000
0P3oool00`000000oooo00000002000001H0oooo000J0?ooo`040000003oool0000000000080oooo
0P0000090?ooo`030000003oool0oooo00<0oooo00<000000?ooo`0000003P3oool00`000000oooo
0?ooo`050?ooo`050000003oool0oooo0?ooo`0000002`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P00000S0?ooo`040000003oool0oooo0000
00<0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`040000003oool0oooo
000000T0oooo00<000000?ooo`000000303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo0080oooo00<000000?ooo`0000002P3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo00D0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool00000008000005P3o
ool001T0oooo00D000000?ooo`000000oooo000000020?ooo`8000002@3oool2000000@0oooo00<0
00000?ooo`0000003@3oool00`000000oooo000000070?ooo`040000003oool0oooo000000/0oooo
00<000000?ooo`3oool0103oool010000000oooo0?ooo`0000040?ooo`030000003oool0oooo0280
oooo0P0000050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000
oooo0?ooo`080?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0203oool010000000
oooo0?ooo`00000:0?ooo`8000003@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0<0?ooo`050000003oool0oooo
0?ooo`0000004@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00<0oooo0P000009
0?ooo`8000002P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000000P3o
ool00`000000oooo00000002000001H0oooo000I0?ooo`050000003oool000000?ooo`0000000P3o
ool00`000000oooo000000080?ooo`8000000`3oool01@000000oooo0?ooo`3oool0000000`0oooo
00<000000?ooo`000000203oool2000000/0oooo00<000000?ooo`3oool01@3oool00`000000oooo
000000040?ooo`800000903oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool00P0000040?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
2@3oool010000000oooo0?ooo`0000080?ooo`040000003oool0oooo000000d0oooo00<000000?oo
o`3oool01`3oool2000000P0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool000000180oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?oo
o`800000203oool00`000000oooo000000090?ooo`030000003oool0oooo00H0oooo0P0000040?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0
00000080oooo00<000000?ooo`0000000P00000F0?ooo`006@3oool01@000000oooo0000003oool0
00000080oooo00<000000?ooo`000000203oool2000000<0oooo00D000000?ooo`3oool0oooo0000
000<0?ooo`040000003oool0oooo000000P0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`070?ooo`800000103oool00`000000oooo0000000R0?ooo`030000003oool0000000L0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool0403oool00`000000oooo000000050?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool02P3oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo0080
oooo0P00000;0?ooo`030000003oool0oooo00H0oooo0P00000:0?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool01P3oool2000000L0oooo00@000000?ooo`3oool00000203oool00`00
0000oooo0?ooo`060?ooo`030000003oool0000000<0oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`050?ooo`040000003oool0oooo00000080oooo00<000000?ooo`0000000P00000F
0?ooo`006@3oool01`000000oooo0000003oool000000?ooo`0000000P3oool00`000000oooo0?oo
o`050?ooo`030000003oool0000000<0oooo00D000000?ooo`3oool0oooo0000000;0?ooo`050000
003oool0oooo0?ooo`0000001`3oool2000000T0oooo00<000000?ooo`3oool0203oool2000000<0
oooo00D000000?ooo`3oool0oooo0000000Q0?ooo`040000003oool0oooo000000H0oooo00D00000
0?ooo`3oool0oooo0000000=0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool04P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool02`3o
ool010000000oooo0?ooo`0000040?ooo`030000003oool0oooo00D0oooo0P00000:0?ooo`030000
003oool0oooo00@0oooo0P00000;0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`0C0?ooo`050000003oool0oooo0?ooo`0000002P3oool00`000000
oooo0?ooo`050?ooo`050000003oool0oooo0?ooo`0000001`3oool00`000000oooo0?ooo`060?oo
o`040000003oool0oooo00000080oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`05
0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`0000000P00000F0?ooo`006@3oool0
1`000000oooo0000003oool000000?ooo`0000000P3oool00`000000oooo0?ooo`050?ooo`040000
003oool0oooo00000080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`080?ooo`05
0000003oool0oooo0?ooo`0000001`3oool00`000000oooo000000070?ooo`030000003oool0oooo
00P0oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
01h0oooo00D000000?ooo`3oool0oooo000000070?ooo`030000003oool0000000l0oooo00@00000
0?ooo`3oool00000503oool00`000000oooo000000040?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool0303oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`020?ooo`030000003oool0000000`0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0000001@0oooo
00D000000?ooo`3oool0oooo0000000;0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00L0oooo00@000000?ooo`3o
ool000000P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00@0oooo00@000000?oo
o`3oool000000P3oool00`000000oooo00000002000001H0oooo000I0?ooo`030000003oool00000
0080oooo0P0000020?ooo`030000003oool0oooo00D0oooo00H000000?ooo`3oool000000?ooo`00
00050?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`090?ooo`060000
003oool0oooo0000003oool000001@3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool00000
01<0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool0303oool010000000oooo0?ooo`3oool2000000`0oooo0P0000090?ooo`060000003o
ool0oooo0000003oool000003P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0100
oooo00D000000?ooo`3oool0oooo0000000D0?ooo`030000003oool0000000`0oooo0P0000040?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`060?oo
o`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00@0oooo00P000000?ooo`3oool0oooo0000003oool000000?ooo`80000000<0oooo0000003o
ool0503oool001T0oooo00<000000?ooo`0000000P3oool200000080oooo00<000000?ooo`3oool0
1@3oool01P000000oooo0?ooo`000000oooo000000D0oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo00000004
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`04
0?ooo`8000004@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00<0oooo0P00000A
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool00P00000>0?ooo`8000001`3oool3
000000030?ooo`000000oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool04@3oool2000000h0oooo0P000002
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo0080oooo0P00000;0?ooo`030000003oool0oooo00@0oooo00P00000
0?ooo`3oool0oooo0000003oool000000?ooo`80000000<0oooo0000003oool0503oool001T0oooo
00<000000?ooo`0000000P3oool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`8000001@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00@0oooo
00D000000?ooo`3oool0oooo000000050?ooo`050000003oool0oooo0?ooo`0000002P3oool00`00
0000oooo0?ooo`020?ooo`8000001`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`00000?0?ooo`030000003oool0
00000140oooo00<000000?ooo`3oool0103oool200000140oooo00<000000?ooo`3oool03P3oool2
00000080oooo0P00000>0?ooo`030000003oool0oooo00@0oooo0P0000020?ooo`030000003oool0
oooo00l0oooo00@000000?ooo`3oool000004P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0140oooo0P00000>0?ooo`050000003oool000000?ooo`0000001P3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool00`3oool2000000/0oooo
00<000000?ooo`3oool01@3oool01`000000oooo0?ooo`000000oooo0000003oool00P0000000`3o
ool000000?ooo`0D0?ooo`006@3oool00`000000oooo000000020?ooo`8000000`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo0P0000060?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool01@3oool01@000000oooo0?ooo`3oool0000000H0oooo00<000000?oo
o`0000002`3oool00`000000oooo0?ooo`020?ooo`030000003oool0000000L0oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`080?ooo`040000003oool0oooo000000<0oooo00<00000
0?ooo`3oool0303oool01@000000oooo0?ooo`3oool0000000h0oooo0P0000060?ooo`040000003o
ool0oooo00000100oooo0`00000>0?ooo`800000103oool00`000000oooo0?ooo`0<0?ooo`800000
0`3oool00`000000oooo0?ooo`02000001<0oooo00<000000?ooo`0000004@3oool00`000000oooo
0?ooo`060?ooo`030000003oool0oooo0100oooo0P00000?0?ooo`<00000203oool010000000oooo
0?ooo`0000090?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`040?ooo`040000003oool0oooo00000080oooo0`0000000`3oool000000?ooo`0D0?oo
o`006@3oool00`000000oooo000000020?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00@000000?ooo`3oool00000203oool00`000000oooo0?ooo`090?oo
o`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo000000070?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool02P3oool2000000D0oooo00<000000?ooo`3oool02P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool0
10000000oooo0?ooo`0000040?ooo`030000003oool0oooo00d0oooo0`0000030?ooo`<000004`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool03P3o
ool010000000oooo0?ooo`00000>0?ooo`<00000203oool010000000oooo0?ooo`0000090?ooo`03
0000003oool0oooo00<0oooo0P00000<0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3o
ool000000P3oool3000000030?ooo`000000oooo01@0oooo000I0?ooo`030000003oool000000080
oooo0P0000030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool2000000H0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`070?ooo`030000003oool0000000L0
oooo0P00000:0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool02`3oool2000000H0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`040?ooo`8000003@3oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo0080oooo0P0000030?ooo`030000003oool0oooo00d0
oooo1@0000030?ooo`030000003oool0oooo00l0oooo0P00000@0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00`0
oooo00<000000?ooo`000000203oool2000000T0oooo00<000000?ooo`3oool0103oool00`000000
oooo0000000;0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000000P3oool01@00
0000oooo0000003oool0000001D0oooo000I0?ooo`030000003oool000000080oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo0P0000070?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool01`3oool2000000L0oooo00@000000?oo
o`3oool000002@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00/0oooo00@00000
0?ooo`3oool000001P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`8000001`3oool01@000000oooo0?oo
o`3oool0000000d0oooo00<000000?ooo`0000000P0000050?ooo`030000003oool0oooo00d0oooo
00@000000?ooo`3oool000003P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00`0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0:0?ooo`050000003oool000000?oo
o`0000001`3oool2000000T0oooo00<000000?ooo`3oool0103oool00`000000oooo0000000<0?oo
o`030000003oool0oooo00<0oooo00P000000?ooo`3oool0oooo0000003oool000000?ooo`<00000
5@3oool001T0oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`030?ooo`050000003o
ool0oooo0?ooo`0000001@3oool00`000000oooo000000060?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`050?ooo`050000003oool0oooo0?ooo`00
00001`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool01@3oool2000001<0oooo00<000000?ooo`3oool02@3oool2000000@0oooo0P0000050?ooo`03
0000003oool0oooo00<0oooo0P00000;0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool02@3oool00`000000oooo0?ooo`080?ooo`050000003oool0oooo0?ooo`0000002`3oool00`00
0000oooo000000030?ooo`800000103oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`090?ooo`040000003oool0oooo
00000080oooo00<000000?ooo`3oool0103oool2000000P0oooo00<000000?ooo`3oool0103oool0
1@000000oooo0?ooo`3oool0000000/0oooo00<000000?ooo`3oool00`3oool020000000oooo0?oo
o`3oool000000?ooo`000000oooo0`00000E0?ooo`006@3oool2000000<0oooo00<000000?ooo`3o
ool00`3oool01@000000oooo0?ooo`3oool0000000@0oooo00@000000?ooo`3oool000001`3oool0
1@000000oooo0?ooo`3oool0000000X0oooo00<000000?ooo`0000001@3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00H0oooo0P0000040?ooo`030000003oool0oooo008000003P3o
ool010000000oooo0?ooo`3oool2000000L0oooo00<000000?ooo`3oool02`3oool2000000T0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0:0?ooo`050000003oool0oooo0?ooo`00
0000203oool2000000030?ooo`000000oooo00D0oooo0P0000030?ooo`<000001`3oool3000000H0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0080
oooo00<000000?ooo`0000001P3oool00`000000oooo000000070?ooo`030000003oool0oooo00@0
oooo00D000000?ooo`3oool0oooo0000000;0?ooo`030000003oool0oooo00@0oooo00L000000?oo
o`3oool000000?ooo`000000oooo00<000005@3oool001T0oooo0P0000030?ooo`800000103oool0
10000000oooo0?ooo`0000050?ooo`040000003oool0oooo000000L0oooo00@000000?ooo`3oool0
00002`3oool00`000000oooo000000040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`0<0?ooo`8000002`3oool00`000000oooo0?ooo`090?ooo`L0
00004@3oool3000000P0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo00H0oooo0P00000>0?ooo`<0000000<0oooo0000000000001@3oool4000000T0
oooo00<000000?ooo`3oool00`3oool7000000X0oooo00<000000?ooo`3oool01@3oool2000001<0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool00`3oool00`000000oooo000000040?ooo`040000003oool0oooo0000
00H0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00P0oooo00<000000?ooo`3oool0103oool01`000000oooo0?ooo`000000oooo0000003oool00`00
000E0?ooo`006@3oool2000000<0oooo0P0000040?ooo`040000003oool0oooo000000D0oooo00D0
00000?ooo`3oool0oooo000000060?ooo`040000003oool0oooo000000X0oooo00D000000?ooo`3o
ool0oooo000000020?ooo`030000003oool0oooo00H0oooo00D000000?ooo`3oool0oooo00000004
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`09
0?ooo`800000303oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00l0oooo0P000006
0?ooo`8000004P3oool3000000D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0A
0?ooo`P000003P3oool2000001D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0D
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000000`3oool01@000000oooo0?oo
o`3oool0000000D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00T0oooo00<000000?ooo`3oool00`3oool01`000000oooo0?ooo`000000oooo0000
003oool00`00000E0?ooo`006@3oool2000000<0oooo0P0000050?ooo`030000003oool0000000@0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`050?ooo`8000002`3oool00`000000
oooo0?ooo`020?ooo`8000002P3oool00`000000oooo000000040?ooo`030000003oool0oooo00`0
oooo00<000000?ooo`3oool01P3oool3000000H0oooo0P00000=0?ooo`030000003oool0oooo01<0
oooo0P00000=0?ooo`8000000P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01D0
oooo0`0000020?ooo`030000003oool0oooo00D00000603oool00`000000oooo0000000A0?ooo`80
0000503oool01@000000oooo0?ooo`3oool0000001P0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0<0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01P3oool01P00
0000oooo0?ooo`000000oooo000000@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`030?ooo`070000003oool0oooo0000003oool000000?ooo`03000001D0oooo000I0?ooo`800000
0`3oool2000000D0oooo00<000000?ooo`000000103oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00@0oooo0P00000:0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
2@3oool2000000<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`080?ooo`H00000
3P3oool00`000000oooo0?ooo`0F0?ooo`8000002P3oool00`000000oooo0?ooo`030?ooo`@00000
6`3oool5000001d0oooo0P00000D0?ooo`030000003oool0oooo0180oooo0`00000J0?ooo`040000
003oool0oooo00000100oooo00@000000?ooo`3oool000002P3oool010000000oooo0?ooo`000006
0?ooo`040000003oool0oooo000000L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo00<0oooo00L000000?ooo`3oool000000?ooo`000000oooo00<0
00005@3oool001P0oooo00<000000?ooo`0000000`3oool2000000D0oooo0P0000050?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool0103oool2000000X0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0000000:0?ooo`8000000P3oool00`000000oooo0?ooo`0@0?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool06@3oool3000000H0oooo00<000000?ooo`3oool0
0P3oool5000001h0oooo00<000000?ooo`3oool0703oool010000000oooo0?ooo`00000D0?ooo`80
00004P3oool2000001`0oooo0P00000B0?ooo`8000002`3oool010000000oooo0000000000070?oo
o`030000003oool0000000L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`090?oo
o`030000003oool0oooo0080oooo00@000000?ooo`3oool0oooo0P000000103oool0000000000000
000E0?ooo`00603oool00`000000oooo000000020?ooo`030000003oool0000000D0oooo0P000004
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool00`000000oooo00000008
0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo000000080?ooo`030000003o
ool0oooo00800000503oool2000001T0oooo00<000000?ooo`3oool07@3oool;000000D0oooo0P00
000L0?ooo`030000003oool0oooo01X0oooo0P0000040?ooo`800000503oool2000000l0oooo00<0
00000?ooo`3oool00P00000J0?ooo`8000004P3oool2000000`0oooo00@000000?ooo`3oool00000
1`3oool00`000000oooo0?ooo`040?ooo`800000203oool00`000000oooo0?ooo`090?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P000000103oool0000000000000000E0?ooo`00
603oool00`000000oooo000000020?ooo`030000003oool0000000H0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0
oooo000000050?ooo`8000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool00`00000F0?ooo`030000003oool0oooo01@0oooo0P00000S0?ooo`03
0000003oool0oooo00`0oooo00<000000?ooo`3oool06@3oool3000001P0oooo0P0000080?ooo`03
0000003oool0oooo01<0oooo0P00000<0?ooo`030000003oool0oooo00<0oooo0P00000F0?ooo`80
00000P3oool2000000l0oooo00@000000?ooo`3oool000002`3oool3000000030?ooo`000000oooo
00@0oooo00<000000?ooo`000000103oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo
00P0oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`000000oooo00@000005@3oool0
01P0oooo00<000000?ooo`0000000P3oool010000000oooo0?ooo`0000050?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool01P3oool01@000000oooo0?ooo`3oool0000000@0oooo1@00
00060?ooo`030000003oool0oooo00D0oooo0P0000040?ooo`030000003oool0000000<0oooo00<0
00000?ooo`3oool0503oool200000180oooo0P00000T0?ooo`030000003oool0oooo00h0oooo0P00
000I0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool02@3oool2000001D0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`8000004`3oool00`000000oooo0?oo
o`040?ooo`8000002`3oool2000000@0oooo0P0000060?ooo`<000000`3oool3000000040?ooo`00
00000000000000<0oooo1@00000;0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
0`3oool01@000000oooo0?ooo`000000oooo00@000005@3oool001P0oooo00<000000?ooo`000000
0P3oool010000000oooo0?ooo`0000040?ooo`800000103oool00`000000oooo0?ooo`060?ooo`04
0000003oool0oooo00000100oooo00<000000?ooo`3oool01`3oool6000000D0oooo00<000000?oo
o`3oool05@3oool2000000h0oooo0P00000T0?ooo`8000004`3oool2000001L0oooo00<000000?oo
o`3oool05@3oool2000000h0oooo0`00000C0?ooo`8000001`3oool00`000000oooo0?ooo`090?oo
o`8000003`3oool2000000T0oooo0`0000050?ooo`<00000203oool7000000L0oooo0`00000F0?oo
o`030000003oool0oooo00P0oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`000000
oooo00@000005@3oool001P0oooo00<000000?ooo`0000000P3oool010000000oooo0?ooo`000004
0?ooo`8000000`3oool00`000000oooo0?ooo`080?ooo`030000003oool0000000l0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo01H0oooo0`000007
0?ooo`@000009@3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo01@0oooo00<00000
0?ooo`3oool04`3oool2000001<0oooo0P00000C0?ooo`8000000`3oool2000000h0oooo10000006
0?ooo`D000003P3oool5000001X0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool000000?ooo`04000001D0oooo000H
0?ooo`030000003oool000000080oooo00@000000?ooo`3oool00000103oool00`000000oooo0000
00020?ooo`030000003oool0oooo00P0oooo0P00000?0?ooo`030000003oool0oooo00X0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`0H0?ooo`L00000:03oool00`000000oooo0?oo
o`0G0?ooo`800000503oool00`000000oooo0?ooo`0A0?ooo`8000005`3oool4000000l0oooo1@00
000D0?ooo`H00000<`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool00P3oool01@000000oooo0?ooo`000000oooo00@000005@3oool001P0oooo00<0
00000?ooo`0000000P3oool010000000oooo0?ooo`0000040?ooo`030000003oool000000080oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool02`3oool2000004D0oooo0P00000L0?ooo`<000004@3oool00`000000oooo
0?ooo`0?0?ooo`8000007@3oool6000000<0oooo1P0000030?ooo`800000CP3oool00`000000oooo
0?ooo`0D0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`02000000030?ooo`000000000001D0oooo000H0?ooo`030000003oool000000080oooo00D0
00000?ooo`3oool0oooo000000020?ooo`060000003oool0oooo0000003oool000002`3oool00`00
0000oooo0000000=0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`100?ooo`8000008@3oool3000000h0oooo00<000000?ooo`3oool03@3oool20000
02@0oooo1@00000:0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool04`3oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P0000000`3o
ool000000000000E0?ooo`00603oool00`000000oooo000000020?ooo`050000003oool0oooo0?oo
o`0000000P3oool010000000oooo0?ooo`3oool2000000/0oooo00<000000?ooo`000000303oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0?@3oool2
000002H0oooo0P00000<0?ooo`030000003oool0oooo00T0oooo1000000f0?ooo`800000C03oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`02000000030?ooo`000000000001D0oooo000H0?ooo`8000000P3oool00`00
0000oooo0?ooo`020?ooo`040000003oool0oooo000000<0oooo00<000000?ooo`3oool02@3oool0
1@000000oooo0?ooo`3oool0000000X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0A0?ooo`800000>`3oool2000002X0oooo1@0000070?ooo`030000003oool0oooo00@0oooo1@00
000l0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P0000000`3oool000000000
000E0?ooo`00603oool200000080oooo00<000000?ooo`3oool00`3oool00`000000oooo00000003
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo00P0oooo0P00000F0?ooo`030000003oool0oooo03L0oooo00<00000
0?ooo`3oool0;`3oool>00000480oooo0P00001;0?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool01`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080000000<0oooo
0000000000005@3oool001P0oooo0P0000020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`0000000P3oool00`000000oooo000000090?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`070?ooo`8000006@3oool00`000000oooo0?ooo`0d0?oo
o`800000>@3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo04P0oooo00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`070?ooo`050000003oool0oooo0?ooo`0000000P3oool2
000000030?ooo`000000000001D0oooo000H0?ooo`8000000P3oool00`000000oooo0?ooo`030?oo
o`8000000`3oool00`000000oooo0?ooo`02000000H0oooo00<000000?ooo`3oool01P3oool20000
00@0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0J0?ooo`800000<P3oool20000
03/0oooo00<000000?ooo`3oool0B@3oool2000004T0oooo00<000000?ooo`3oool04P3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool000000?ooo`02
000001D0oooo000H0?ooo`8000000P3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo
000000D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`090?ooo`050000003oool0
oooo0?ooo`000000203oool00`000000oooo0?ooo`0M0?ooo`<00000;@3oool2000003d0oooo0`00
001;0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01@00000D0?ooo`00603oool2
00000080oooo00<000000?ooo`3oool0103oool3000000H0oooo1@00000=0?ooo`030000003oool0
000000P0oooo00<000000?ooo`3oool08@3oool2000002P0oooo0`00000o0?ooo`030000003oool0
oooo04`0oooo0P0000170?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00D00000503oool001P0oooo0P0000020?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool06@3oool:000002H0oooo0P00000R0?oo
o`@00000@P3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo04D0oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool01@00000D0?ooo`00603oool200000080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0J0?ooo`030000003oool0oooo02l0oooo0P00000M0?ooo`<00000AP3oool00`000000
oooo0?ooo`1?0?ooo`800000AP3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00L0
oooo00D000000?ooo`3oool0oooo000000020?ooo`D00000503oool001P0oooo0P0000020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0b0?ooo`<0
0000603oool2000004T0oooo00<000000?ooo`3oool0D@3oool2000004@0oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`060?ooo`050000003oool0oooo0?ooo`0000000P3oool50000
01@0oooo000H0?ooo`8000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01P0
oooo00<000000?ooo`3oool0=P3oool7000000/0oooo1P00001;0?ooo`030000003oool0oooo05<0
oooo00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool00P3oool00`000000oooo00000004000001@0oooo000G0?ooo`050000
003oool000000?ooo`0000001P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo03d0
oooo2`00001A0?ooo`030000003oool0oooo05@0oooo0P0000130?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`020?ooo`030000003oool0000000@0
0000503oool001L0oooo00D000000?ooo`000000oooo000000060?ooo`030000003oool0oooo01L0
oooo00<000000?ooo`3oool0O@3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00H0okmd0000
001d_ol0oooo0?oog`2LB0030000000307Boo`3oool0oooo00T0oooo00<000000?ooo`3oool0EP3o
ool200000480oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`0000001000000D0?ooo`005`3oool01@000000oooo0000003o
ool0000000D0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`1m0?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo0?ooo`3ooml0W4Qd0;ooo`02
0?ooo`030;md001d_ol0oooo00T0oooo00<000000?ooo`3oool0F03oool200000400oooo00<00000
0?ooo`3oool03`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool01000000D0?ooo`005`3oool01@000000oooo0000003oool0000000D0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`1m0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0
001d_`3oool02P3oool00`3ooml0W4Qd0;ooo`090?ooo`030000003oool0oooo05X0oooo00<00000
0?ooo`3oool0?P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00D0oooo00D00000
0?ooo`3oool0oooo000000020?ooo`@00000503oool001L0oooo00D000000?ooo`000000oooo0000
00050?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0O@3oool00`3ookl0M00004RL
g`030?ooo`030=nLB00007@0_ooo00/0oooo00<0_g@007Boo`3oool0203oool00`000000oooo0?oo
o`1K0?ooo`800000?`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00D0oooo00D0
00000?ooo`3oool0oooo000000020?ooo`040000003oool00000000001@0oooo000G0?ooo`800000
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0
OP3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0oooO09a80018W=l0
2@3oool00`000000oooo0?ooo`1M0?ooo`800000?@3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo00D0oooo00<000000?ooo`3oool00P3oool010000000oooo0000003oool2000001@0
oooo000G0?ooo`8000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01D0oooo
00<000000?ooo`3oool0OP3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo
00<0gia80000001d_ol02@3oool4000005h0oooo00<000000?ooo`3oool0>`3oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo
00000003000001@0oooo000G0?ooo`8000000P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo01D0oooo00<000000?ooo`3oool0O`3oool00`3ogi`0B00007Boo`030?ooo`030?noM000
04P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00T0oooo00<000000?ooo`3o
ool0H03oool2000003`0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`0000000`00000D0?ooo`005`3oool200000080oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0800oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`060=nLB0000000000000000000
04P0W=oo2P3oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo03T0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`040?ooo`060000003oool0oooo0?ooo`000000oooo1000
000D0?ooo`005`3oool2000000030?ooo`000000oooo00@0oooo00<000000?ooo`3oool0503oool0
0`000000oooo0?ooo`200?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool0
103ookl0M00000000018W=l<0?ooo`030000003oool0oooo06<0oooo0P00000j0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool0103oool01P000000oooo0?ooo`3oool000000?ooo`@0
0000503oool001L0oooo0P0000000`3oool000000?ooo`040?ooo`030000003oool0oooo01<0oooo
00<000000?ooo`3oool0PP3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo
00<0_g@007Boo`3oool0303oool00`000000oooo0?ooo`1U0?ooo`800000>@3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00@0oooo00H000000?ooo`3oool0oooo0000003oool40000
01@0oooo000G0?ooo`80000000<0oooo0000003oool0103oool00`000000oooo0?ooo`0B0?ooo`03
0000003oool0oooo08<0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00H0omnL
04P00000000000000000B02Lgol90?ooo`030000003oool0oooo06L0oooo00<000000?ooo`3oool0
=P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00<0oooo00H000000?ooo`3oool0
oooo0000003oool4000001@0oooo000G0?ooo`80000000<0oooo0000003oool00`3oool00`000000
oooo0?ooo`0C0?ooo`030000003oool0oooo08@0oooo00D0oono07@000000000001d0;ooo`0:0?oo
o`050;md00000000000000000018W=l02@3oool00`000000oooo0?ooo`1X0?ooo`800000=`3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool5
000001@0oooo000G0?ooo`80000000<0oooo0000003oool00`3oool00`000000oooo0?ooo`0B0?oo
o`030000003oool0oooo0:80oooo00<000000?ooo`3oool0JP3oool00`000000oooo0?ooo`0e0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`3oool5
000001@0oooo000G0?ooo`80000000<0oooo0000003oool00`3oool00`000000oooo0?ooo`0B0?oo
o`030000003oool0oooo0:80oooo00<000000?ooo`3oool0J`3oool2000003D0oooo00<000000?oo
o`3oool0303oool00`000000oooo0?ooo`030?ooo`060000003oool0oooo0?ooo`000000oooo0`00
000D0?ooo`005`3oool2000000030?ooo`000000oooo00<0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`2S0?ooo`030000003oool0oooo06d0oooo0P00000d0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool00`3oool01P000000oooo0?ooo`3oool000000?ooo`<00000
503oool001L0oooo0P0000000`3oool000000?ooo`020?ooo`030000003oool0oooo0140oooo00<0
00000?ooo`3oool0Y03oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo0340oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`020?ooo`060000003oool0oooo0?ooo`000000
oooo0`00000D0?ooo`005`3oool2000000030?ooo`000000oooo0080oooo00<000000?ooo`3oool0
4@3oool00`000000oooo0?ooo`2T0?ooo`030000003oool0oooo0700oooo0P00000b0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool00P3oool01P000000oooo0?ooo`3oool000000?oo
o`<00000503oool001L0oooo0P0000000`3oool000000?ooo`020?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool0Y@3oool300000780oooo00<000000?ooo`3oool0<03oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool000000?ooo`030000
01@0oooo000G0?ooo`<00000103oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0:D0
oooo00<000000?ooo`3oool0L`3oool200000300oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`020?ooo`050000003oool0oooo0000003oool00`00000D0?ooo`005`3oool3000000@0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`2V0?ooo`030000003oool0oooo07D0
oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0080
oooo00@000000?ooo`3oool0oooo1000000D0?ooo`005`3oool00`000000oooo000000030?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?ooo`1f0?ooo`80
0000;`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00@000000?ooo`3o
ool0oooo1000000D0?ooo`005`3oool00`000000oooo000000030?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?ooo`1h0?ooo`800000;@3oool00`000000
oooo0?ooo`0:0?ooo`050000003oool0oooo0?ooo`0000000`3oool4000001@0oooo000G0?ooo`03
0000003oool0000000<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`2X0?ooo`03
0000003oool0oooo07X0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool01000000D0?ooo`005`3oool00`000000oooo
000000030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0Z@3oool00`000000oooo
0?ooo`1k0?ooo`800000:`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool01000000D0?ooo`005P3oool2000000030?ooo`000000oooo0080oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`2Y0?ooo`030000003oool0oooo07d0oooo00<0
00000?ooo`3oool0:@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo0080oooo00D0
00000?ooo`3oool000000?ooo`02000001@0oooo000F0?ooo`80000000D0oooo0000003oool0oooo
0000000?0?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0OP3oool2000002X0oooo
00<000000?ooo`3oool0203oool01@000000oooo0?ooo`3oool000000080oooo00@000000?ooo`00
00000000503oool001H0oooo0P0000001@3oool000000?ooo`3oool0000000l0oooo00<000000?oo
o`3oool0ZP3oool00`000000oooo0?ooo`200?ooo`030000003oool0oooo02L0oooo00<000000?oo
o`3oool0203oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`0000000P00000C
0?ooo`005P3oool2000000050?ooo`000000oooo0?ooo`0000003P3oool00`000000oooo0?ooo`2[
0?ooo`030000003oool0oooo0840oooo0P00000X0?ooo`030000003oool0oooo00L0oooo00D00000
0?ooo`3oool0oooo000000020?ooo`030000003oool00000008000004`3oool001H0oooo0`000003
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0[03oool00`000000oooo0?ooo`23
0?ooo`8000009P3oool00`000000oooo0?ooo`080?ooo`050000003oool0oooo0?ooo`0000000P3o
ool4000001<0oooo000F0?ooo`<000000`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo0:`0oooo0`0000250?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool01`3oool0
1@000000oooo0?ooo`3oool000000080oooo1000000C0?ooo`005P3oool3000000<0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`2]0?ooo`030000003oool0oooo08H0oooo0P00000U
0?ooo`030000003oool0oooo00H0oooo00D000000?ooo`3oool0oooo000000020?ooo`@000004`3o
ool001H0oooo0`0000020?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0[@3oool0
0`000000oooo0?ooo`280?ooo`8000008`3oool00`000000oooo0?ooo`070?ooo`040000003oool0
oooo00000080oooo1000000C0?ooo`005P3oool300000080oooo00<000000?ooo`3oool02P3oool0
0`000000oooo0?ooo`2^0?ooo`030000003oool0oooo08X0oooo0P00000R0?ooo`030000003oool0
oooo00H0oooo00@000000?ooo`3oool000000P3oool4000001<0oooo000F0?ooo`<000000P3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo0:l0oooo00<000000?ooo`3oool0S03oool0
0`000000oooo0?ooo`0P0?ooo`030000003oool0oooo00D0oooo00H000000?ooo`3oool0oooo0000
003oool4000001<0oooo000F0?ooo`<000000P3oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo0:l0oooo00<000000?ooo`3oool0S@3oool200000200oooo00<000000?ooo`3oool01@3o
ool01P000000oooo0?ooo`3oool000000?ooo`@000004`3oool001H0oooo0`0000020?ooo`030000
003oool0oooo00P0oooo00<000000?ooo`3oool0/03oool00`000000oooo0?ooo`2?0?ooo`800000
7`3oool00`000000oooo0?ooo`050?ooo`050000003oool0oooo0000003oool01000000C0?ooo`00
5P3oool3000000030?ooo`000000oooo00X0oooo00<000000?ooo`3oool0/03oool00`000000oooo
0?ooo`2A0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool01@3oool01`000000oooo
0?ooo`000000oooo0000003oool00P00000C0?ooo`005P3oool01@000000oooo0000003oool00000
00X0oooo00<000000?ooo`3oool0/@3oool00`000000oooo0?ooo`2B0?ooo`8000007@3oool00`00
0000oooo0?ooo`040?ooo`040000003oool0oooo00000080oooo0`00000C0?ooo`005P3oool01@00
0000oooo0000003oool0000000T0oooo00<000000?ooo`3oool0/P3oool00`000000oooo0?ooo`2D
0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`00
0000oooo00<000004`3oool001H0oooo0P0000020?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool0/P3oool00`000000oooo0?ooo`2E0?ooo`8000006`3oool00`000000oooo0?ooo`04
0?ooo`050000003oool0oooo0000003oool00`00000C0?ooo`005P3oool200000080oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`2c0?ooo`030000003oool0oooo09L0oooo0P00000J
0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool000000?ooo`03000001<0oooo000F
0?ooo`8000000P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0;<0oooo0`00002I
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0103oool010000000oooo0000003o
ool3000001<0oooo000F0?ooo`80000000<0oooo0000003oool01`3oool00`000000oooo0?ooo`2d
0?ooo`030000003oool0oooo09X0oooo0P00000H0?ooo`030000003oool0oooo00<0oooo00@00000
0?ooo`000000oooo0`00000C0?ooo`005P3oool2000000030?ooo`000000oooo00L0oooo00<00000
0?ooo`3oool0]03oool00`000000oooo0?ooo`2L0?ooo`030000003oool0oooo01D0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`04000001<0oooo000F0?ooo`80000000<0oooo0000
003oool01P3oool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo09d0oooo0P00000F0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool01000000C0?ooo`005P3oool200000003
0?ooo`000000oooo00H0oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`2O0?ooo`80
0000503oool00`000000oooo0?ooo`030?ooo`040000003oool000000?ooo`8000004`3oool001H0
oooo0P0000000`3oool000000?ooo`050?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0080oooo00@000000?ooo`00
0000oooo0P00000C0?ooo`005P3oool3000000H0oooo00<000000?ooo`3oool0]`3oool00`000000
oooo0?ooo`2R0?ooo`8000004`3oool020000000oooo0?ooo`3oool000000?ooo`000000oooo0P00
000C0?ooo`005P3oool3000000H0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`2T
0?ooo`8000004@3oool00`000000oooo0?ooo`020?ooo`030000003oool00000008000004`3oool0
01H0oooo0`0000050?ooo`030000003oool0oooo0;P0oooo00<000000?ooo`3oool0YP3oool00`00
0000oooo0?ooo`0?0?ooo`060000003oool0oooo0?ooo`000000oooo0`00000C0?ooo`005P3oool0
0`000000oooo000000050?ooo`030000003oool0oooo0;P0oooo00<000000?ooo`3oool0Y`3oool2
000000l0oooo00H000000?ooo`3oool0oooo0000003oool3000001<0oooo000F0?ooo`030000003o
ool0000000@0oooo00<000000?ooo`3oool0^@3oool00`000000oooo0?ooo`2Y0?ooo`8000003P3o
ool010000000oooo0?ooo`3oool4000001<0oooo000E0?ooo`80000000<0oooo0000003oool00`3o
ool00`000000oooo0?ooo`2i0?ooo`030000003oool0oooo0:/0oooo0P00000<0?ooo`040000003o
ool0oooo0?ooo`@000004`3oool001D0oooo0`0000040?ooo`030000003oool0oooo0;X0oooo00<0
00000?ooo`3oool0[@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00D000004P3o
ool001D0oooo0`0000040?ooo`030000003oool0oooo0;X0oooo0`00002^0?ooo`8000002P3oool0
1@000000oooo0?ooo`000000oooo00<000004P3oool001D0oooo0`0000030?ooo`030000003oool0
oooo0;/0oooo00<000000?ooo`3oool0/03oool2000000T0oooo00<000000?ooo`3oool01000000B
0?ooo`005@3oool3000000<0oooo00<000000?ooo`3oool0^`3oool00`000000oooo0?ooo`2b0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool01000000B0?ooo`005@3oool300000080
oooo00<000000?ooo`3oool0_03oool00`000000oooo0?ooo`2c0?ooo`8000001`3oool00`000000
oooo0000000300000180oooo000E0?ooo`8000000P3oool00`000000oooo0?ooo`2m0?ooo`030000
003oool0oooo0;D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0000000200000180oooo
000E0?ooo`8000000P3oool00`000000oooo0?ooo`2m0?ooo`030000003oool0oooo0;H0oooo0P00
00050?ooo`030000003oool00000008000004P3oool001D0oooo0P0000000`3oool000000?ooo`2o
0?ooo`030000003oool0oooo0;P0oooo00<000000?ooo`3oool00`3oool400000180oooo000E0?oo
o`80000000<0oooo0000003oool0/03oool01P2oM000000000000000000004RL0=ooo`T0oooo00<0
00000?ooo`3oool0^@3oool2000000<0oooo00@000000?ooo`00000000004P3oool001D0oooo0`00
002c0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0^`3oool010000000oooo0?oo
o`3oool300000180oooo000E0?ooo`<00000/`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003o
ool0oooo0;`0oooo0P000000103oool0000000000000000B0?ooo`005@3oool200000;@0oooo00<0
omnL04P0001d_ol02`3oool00`000000oooo0?ooo`2n0?ooo`040000003oool0000000000180oooo
000E0?ooo`800000]03oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0;l0oooo0`00
000B0?ooo`005@3ooooo000008X000004P3oool00</0oooo00<0omnL04P0001d_ol02`3oool00`00
0000oooo0?ooo`3D0?ooo`00b`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0=@0
oooo003;0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0e03oool00</0oooo00<0
omnL04P0001d_ol02`3oool00`000000oooo0?ooo`3D0?ooo`00b@3oool01@3ooml0W4P000000000
0000M;oo00/0oooo00<000000?ooo`3oool0e03oool00</0oooo00<0okmd0000001d_ol02`3oool0
0`000000oooo0?ooo`3D0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3o
oona0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.10918, -1.09709, \
0.00765073, 0.0123791}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Useful embellishments", "Subsection"],

Cell[CellGroupData[{

Cell["Options available to Plot", "Subsubsection"],

Cell[TextData[{
  "Although one often neglects to label plots properly while developing a \
solution to a problem, treating the notebook as glorified scratch paper, you \
should always label figures as the final solution is approached.  It may also \
be necessary to override the default choices for plotting range or \
resolution.  The options available to ",
  StyleBox["Plot", "InlineInput",
    FontWeight->"Bold"],
  " are easily printed, and most are common to other similar plotting \
commands."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Options[Plot]\)], "Input"],

Cell[BoxData[
    \({AspectRatio \[Rule] 1\/GoldenRatio, Axes \[Rule] Automatic, 
      AxesLabel \[Rule] None, AxesOrigin \[Rule] Automatic, 
      AxesStyle \[Rule] Automatic, Background \[Rule] Automatic, 
      ColorOutput \[Rule] Automatic, Compiled \[Rule] True, 
      DefaultColor \[Rule] Automatic, 
      DefaultFont \[RuleDelayed] $DefaultFont, 
      DisplayFunction \[RuleDelayed] $DisplayFunction, Epilog \[Rule] {}, 
      FormatType \[RuleDelayed] $FormatType, Frame \[Rule] False, 
      FrameLabel \[Rule] None, FrameStyle \[Rule] Automatic, 
      FrameTicks \[Rule] Automatic, GridLines \[Rule] None, 
      ImageSize \[Rule] Automatic, MaxBend \[Rule] 10.`, 
      PlotDivision \[Rule] 30.`, PlotLabel \[Rule] None, 
      PlotPoints \[Rule] 25, PlotRange \[Rule] Automatic, 
      PlotRegion \[Rule] Automatic, PlotStyle \[Rule] Automatic, 
      Prolog \[Rule] {}, RotateLabel \[Rule] True, 
      TextStyle \[RuleDelayed] $TextStyle, 
      Ticks \[Rule] Automatic}\)], "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Axis labels and titles", "Subsubsection"],

Cell[TextData[{
  "The simplest labeling uses ",
  StyleBox["AxesLabel", "InlineInput",
    FontWeight->"Bold"],
  " for the axes and ",
  StyleBox["PlotLabel", "InlineInput",
    FontWeight->"Bold"],
  " for the entire figure.  One simply replaces the defaults by a list of \
strings for ",
  StyleBox["AxesLabel", "InlineInput",
    FontWeight->"Bold"],
  ", where the first label is for the x-axis and the second for the y-axis, \
or by a single string for ",
  StyleBox["PlotLabel", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Gamma[x], {x, \(-5\), 5}, \n\t
        AxesLabel \[Rule] {"\<x\>", \ "\<\[CapitalGamma][x]\>"}, \ 
        PlotLabel \[Rule] "\<Gamma Function\>"];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.0952381 0.316178 0.0220537 [
[.11905 .30368 -9.15625 -15.6875 ]
[.11905 .30368 9.15625 0 ]
[.30952 .30368 -9.15625 -15.6875 ]
[.30952 .30368 9.15625 0 ]
[.69048 .30368 -5 -15.6875 ]
[.69048 .30368 5 0 ]
[.88095 .30368 -5 -15.6875 ]
[.88095 .30368 5 0 ]
[1.025 .31618 0 -7.84375 ]
[1.025 .31618 10.125 7.84375 ]
[.4875 .09564 -24.3125 -7.84375 ]
[.4875 .09564 0 7.84375 ]
[.4875 .20591 -18.3125 -7.84375 ]
[.4875 .20591 0 7.84375 ]
[.4875 .42645 -10 -7.84375 ]
[.4875 .42645 0 7.84375 ]
[.4875 .53672 -16 -7.84375 ]
[.4875 .53672 0 7.84375 ]
[.5 .64303 -12.5313 0 ]
[.5 .64303 12.5313 15.6875 ]
[.5 .63053 -43.7188 0 ]
[.5 .63053 43.7188 15.6875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.11905 .31618 m
.11905 .32243 L
s
gsave
.11905 .30368 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.30952 .31618 m
.30952 .32243 L
s
gsave
.30952 .30368 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.69048 .31618 m
.69048 .32243 L
s
gsave
.69048 .30368 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.88095 .31618 m
.88095 .32243 L
s
gsave
.88095 .30368 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.16667 .31618 m
.16667 .31993 L
s
.21429 .31618 m
.21429 .31993 L
s
.2619 .31618 m
.2619 .31993 L
s
.35714 .31618 m
.35714 .31993 L
s
.40476 .31618 m
.40476 .31993 L
s
.45238 .31618 m
.45238 .31993 L
s
.54762 .31618 m
.54762 .31993 L
s
.59524 .31618 m
.59524 .31993 L
s
.64286 .31618 m
.64286 .31993 L
s
.7381 .31618 m
.7381 .31993 L
s
.78571 .31618 m
.78571 .31993 L
s
.83333 .31618 m
.83333 .31993 L
s
.07143 .31618 m
.07143 .31993 L
s
.02381 .31618 m
.02381 .31993 L
s
.92857 .31618 m
.92857 .31993 L
s
.97619 .31618 m
.97619 .31993 L
s
.25 Mabswid
0 .31618 m
1 .31618 L
s
gsave
1.025 .31618 -61 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
69.125 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.5 .09564 m
.50625 .09564 L
s
gsave
.4875 .09564 -85.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.5 .20591 m
.50625 .20591 L
s
gsave
.4875 .20591 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.5 .42645 m
.50625 .42645 L
s
gsave
.4875 .42645 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.5 .53672 m
.50625 .53672 L
s
gsave
.4875 .53672 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .1177 m
.50375 .1177 L
s
.5 .13975 m
.50375 .13975 L
s
.5 .1618 m
.50375 .1618 L
s
.5 .18386 m
.50375 .18386 L
s
.5 .22796 m
.50375 .22796 L
s
.5 .25002 m
.50375 .25002 L
s
.5 .27207 m
.50375 .27207 L
s
.5 .29412 m
.50375 .29412 L
s
.5 .33823 m
.50375 .33823 L
s
.5 .36029 m
.50375 .36029 L
s
.5 .38234 m
.50375 .38234 L
s
.5 .40439 m
.50375 .40439 L
s
.5 .4485 m
.50375 .4485 L
s
.5 .47055 m
.50375 .47055 L
s
.5 .49261 m
.50375 .49261 L
s
.5 .51466 m
.50375 .51466 L
s
.5 .07359 m
.50375 .07359 L
s
.5 .05153 m
.50375 .05153 L
s
.5 .02948 m
.50375 .02948 L
s
.5 .00743 m
.50375 .00743 L
s
.5 .55877 m
.50375 .55877 L
s
.5 .58082 m
.50375 .58082 L
s
.5 .60288 m
.50375 .60288 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
gsave
.5 .64303 -73.5313 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(G) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
70.063 15.813 moveto
(@) show
74.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
80.125 15.813 moveto
(D) show
84.063 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.5 .63053 -104.719 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Gamma) show
104.000 15.813 moveto
(Function) show
146.438 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02498 0 m
.02499 .30097 L
.02605 .30806 L
.02729 .31081 L
.02846 .31207 L
.02954 .31277 L
.03053 .31322 L
.03163 .31358 L
.03279 .31386 L
.03395 .31407 L
.0352 .31425 L
.03746 .31449 L
.03884 .31459 L
.04016 .31468 L
.04262 .31479 L
.045 .31487 L
.0463 .31491 L
.04753 .31493 L
.04868 .31495 L
.04993 .31497 L
.05105 .31499 L
.0521 .315 L
.05335 .315 L
.05401 .31501 L
.05472 .31501 L
.05546 .31501 L
.05617 .31501 L
.05749 .31501 L
.05872 .31501 L
.05986 .31501 L
.06111 .315 L
.06244 .31499 L
.06504 .31496 L
.06742 .31493 L
.07003 .31488 L
.07277 .31482 L
.07741 .31469 L
.08247 .31447 L
.08537 .31431 L
.08801 .31413 L
.09085 .3139 L
.09392 .31357 L
.09678 .31317 L
.09941 .31269 L
.10192 .31208 L
.10319 .31169 L
.10458 .31119 L
.10589 .31061 L
.10711 .30996 L
.10837 .30912 L
Mistroke
.10951 .30816 L
.11064 .30696 L
.11169 .30549 L
.11281 .30338 L
.11402 .30005 L
.11524 .29447 L
.11594 .28934 L
.11659 .28193 L
.1179 .24147 L
.11863 .10742 L
Mfstroke
.11863 .10742 m
.11924 .61803 L
s
.1195 .61803 m
.1205 .37806 L
.12179 .3495 L
.12302 .3397 L
.12415 .33484 L
.1252 .33195 L
.12634 .32978 L
.1274 .32827 L
.12838 .32721 L
.13065 .32545 L
.13179 .32481 L
.13304 .32425 L
.1354 .32343 L
.13793 .32282 L
.14031 .32241 L
.14247 .32213 L
.14479 .32191 L
.1461 .32181 L
.14733 .32174 L
.14863 .32168 L
.14981 .32164 L
.15047 .32162 L
.15118 .32161 L
.15247 .32159 L
.15358 .32158 L
.15479 .32159 L
.15607 .3216 L
.15726 .32163 L
.15837 .32166 L
.15939 .32169 L
.16169 .32179 L
.1642 .32194 L
.16653 .32212 L
.1693 .32239 L
.17182 .32269 L
.17698 .32351 L
.17982 .32411 L
.1825 .32481 L
.18712 .3264 L
.18975 .32761 L
.19221 .32905 L
.1948 .33099 L
.19716 .33331 L
.19847 .33491 L
.19989 .33701 L
.20123 .3394 L
.20249 .34218 L
.20366 .34535 L
.20492 .3497 L
.20607 .35483 L
Mistroke
.20712 .36102 L
.20827 .37022 L
.20952 .38546 L
.21083 .41327 L
.21206 .46886 L
Mfstroke
.21206 .46886 m
.21308 .61803 L
s
.21321 .61803 m
.21324 0 L
s
.21552 0 m
.21634 .14124 L
.2175 .20237 L
.21873 .23237 L
.21978 .24731 L
.22094 .25829 L
.22215 .26631 L
.22329 .27179 L
.22453 .27638 L
.22586 .28015 L
.22725 .28324 L
.22857 .28557 L
.23104 .28886 L
.23335 .29108 L
.23548 .29262 L
.23782 .2939 L
.2403 .2949 L
.24277 .29562 L
.24416 .29593 L
.24543 .29615 L
.24661 .29631 L
.2479 .29645 L
.24911 .29653 L
.25023 .29658 L
.25092 .29659 L
.25167 .29659 L
.25237 .29658 L
.25301 .29656 L
.25424 .2965 L
.25553 .29639 L
.25692 .29624 L
.25821 .29606 L
.26116 .2955 L
.26377 .29484 L
.26619 .29406 L
.26879 .29303 L
.27165 .29162 L
.27646 .28846 L
.2791 .28618 L
.28157 .28356 L
.28396 .28046 L
.28624 .27684 L
.28831 .27281 L
.29051 .26745 L
.2929 .25995 L
.29511 .25064 L
.29647 .24331 L
.29775 .23481 L
.299 .22445 L
.30014 .21255 L
Mistroke
.3013 .19698 L
.30254 .17446 L
.3037 .14491 L
.30479 .10353 L
.30606 .02245 L
Mfstroke
.30606 .02245 m
.30623 0 L
s
.30951 0 m
.30959 .61803 L
s
.31325 .61803 m
.31377 .5745 L
.31452 .53772 L
.31521 .51221 L
.31648 .47896 L
.31763 .45783 L
.31895 .43994 L
.32021 .4272 L
.32245 .41078 L
.32358 .40457 L
.32482 .39886 L
.32701 .39091 L
.32907 .38525 L
.33149 .38014 L
.3337 .37656 L
.33625 .3734 L
.33868 .37113 L
.33993 .3702 L
.34127 .36935 L
.34246 .36872 L
.34372 .36817 L
.34485 .36777 L
.3461 .36742 L
.34723 .3672 L
.34828 .36705 L
.34892 .367 L
.34961 .36696 L
.35087 .36697 L
.35209 .36706 L
.35322 .36722 L
.35444 .36746 L
.35579 .36783 L
.35851 .36888 L
.3609 .37016 L
.36308 .37164 L
.36544 .37362 L
.36801 .37629 L
.37235 .38225 L
.37471 .38649 L
.37691 .39123 L
.37937 .39769 L
.38197 .40627 L
.38483 .41853 L
.38745 .43369 L
.3899 .45292 L
.39219 .47794 L
.3935 .49701 L
.3947 .51886 L
.39609 .55172 L
.3974 .59444 L
Mistroke
Mfstroke
.3974 .59444 m
.39792 .61803 L
s
.40444 .61803 m
.40452 0 L
s
.41171 0 m
.41218 .0211 L
.41288 .04545 L
.41353 .0644 L
.4148 .09408 L
.416 .116 L
.41716 .13311 L
.41843 .14837 L
.42066 .16903 L
.42325 .18655 L
.42567 .19862 L
.42797 .20755 L
.43005 .21403 L
.43233 .21976 L
.43473 .22461 L
.43738 .22883 L
.43983 .23186 L
.44246 .23434 L
.44397 .23544 L
.44535 .23626 L
.44655 .23684 L
.44785 .23734 L
.44856 .23755 L
.44922 .23771 L
.45052 .23792 L
.45164 .238 L
.45287 .23798 L
.45415 .23783 L
.45535 .23757 L
.45653 .23721 L
.45779 .2367 L
.46005 .23544 L
.46274 .23334 L
.46519 .2308 L
.46774 .22737 L
.47052 .22261 L
.47313 .21685 L
.47554 .21016 L
.47761 .20305 L
.47982 .19357 L
.48226 .18014 L
.48453 .16348 L
.48575 .15223 L
.48708 .13745 L
.48826 .12155 L
.48951 .10063 L
.49074 .07422 L
.49186 .04329 L
.49299 .00213 L
s
.49299 .00213 m
.49303 0 L
s
.50068 0 m
.50068 .61803 L
s
.50675 .61803 m
.50762 .58072 L
.50876 .54517 L
.51008 .51386 L
.51153 .48796 L
.51289 .46901 L
.51417 .45451 L
.5166 .43324 L
.51881 .41878 L
.52119 .40665 L
.52376 .39633 L
.52839 .38259 L
.53266 .37354 L
.53735 .3661 L
.54238 .36007 L
.54713 .35566 L
.55147 .35243 L
.55654 .34934 L
.56207 .34665 L
.56725 .34459 L
.57212 .34299 L
.58151 .34059 L
.58692 .33953 L
.59188 .33871 L
.60198 .33743 L
.60729 .33694 L
.61288 .33652 L
.61787 .33623 L
.62241 .33602 L
.62493 .33593 L
.62765 .33585 L
.63023 .3358 L
.63137 .33577 L
.6326 .33576 L
.63375 .33574 L
.63479 .33573 L
.63585 .33572 L
.63685 .33572 L
.63801 .33571 L
.63908 .33571 L
.64026 .33571 L
.64152 .33571 L
.64274 .33572 L
.64344 .33573 L
.64407 .33573 L
.64647 .33576 L
.64866 .3358 L
.65096 .33585 L
.65593 .33599 L
.66045 .33616 L
Mistroke
.67062 .33667 L
.68004 .33732 L
.68896 .33809 L
.70889 .34039 L
.72732 .34334 L
.746 .34731 L
.76661 .35313 L
.78598 .3604 L
.80684 .37082 L
.82691 .38426 L
.84801 .40339 L
.86777 .4277 L
.88615 .45795 L
.90663 .50371 L
.92522 .56051 L
Mfstroke
.92522 .56051 m
.93903 .61803 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003o603oool0
0`000000oooo0?ooo`140?ooo`030000003oool0oooo0200oooo0P00000S0?ooo`030000003oool0
000001h0oooo00@000000?ooo`3oool00000i03oool001P0oooo00<000000?ooo`3oool0A03oool0
0`000000oooo0?ooo`0P0?ooo`8000008`3oool00`000000oooo0000000N0?ooo`040000003oool0
oooo00000>@0oooo000H0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0803oool2
000002<0oooo00<000000?ooo`0000007P3oool00`000000oooo0?ooo`0200000><0oooo000H0?oo
o`030000003oool0oooo04@0oooo00<000000?ooo`3oool0803oool2000002<0oooo00<000000?oo
o`0000007P3oool010000000oooo0?ooo`00003T0?ooo`00603oool00`000000oooo0?ooo`140?oo
o`030000003oool0oooo0200oooo0P00000S0?ooo`030000003oool0000001h0oooo00@000000?oo
o`3oool00000i03oool001P0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`0P0?oo
o`8000008`3oool010000000oooo0?ooo`00000M0?ooo`040000003oool0oooo00000>@0oooo000H
0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0803oool2000002<0oooo00@00000
0?ooo`3oool000007@3oool010000000oooo0?ooo`00003T0?ooo`00603oool00`000000oooo0?oo
o`140?ooo`030000003oool0oooo0200oooo0P00000S0?ooo`040000003oool0oooo000001d0oooo
00@000000?ooo`3oool00000i03oool001P0oooo00<000000?ooo`3oool0A03oool00`000000oooo
0?ooo`0P0?ooo`8000008`3oool010000000oooo0?ooo`00000M0?ooo`040000003oool0oooo0000
0>@0oooo000H0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0803oool2000002<0
oooo00@000000?ooo`3oool000007@3oool010000000oooo0?ooo`00003T0?ooo`00603oool00`00
0000oooo0?ooo`140?ooo`030000003oool0oooo0200oooo0P00000S0?ooo`040000003oool0oooo
000001d0oooo00@000000?ooo`3oool00000i03oool001P0oooo00<000000?ooo`3oool0A03oool0
0`000000oooo0?ooo`0P0?ooo`8000008`3oool010000000oooo0?ooo`00000M0?ooo`030000003o
ool0oooo00800000h`3oool001P0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`0P
0?ooo`8000008`3oool010000000oooo0?ooo`00000M0?ooo`040000003oool0oooo00000>@0oooo
000H0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0803oool2000002<0oooo00@0
00000?ooo`3oool000007@3oool010000000oooo0?ooo`00003T0?ooo`00603oool00`000000oooo
0?ooo`140?ooo`030000003oool0oooo0200oooo0P00000S0?ooo`040000003oool0oooo000001d0
oooo00@000000?ooo`3oool00000i03oool001P0oooo00<000000?ooo`3oool0A03oool00`000000
oooo0?ooo`0P0?ooo`8000008`3oool010000000oooo0?ooo`00000M0?ooo`040000003oool0oooo
00000>@0oooo000H0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0803oool20000
02<0oooo00@000000?ooo`3oool000007@3oool010000000oooo0?ooo`00003T0?ooo`00603oool0
0`000000oooo0?ooo`140?ooo`030000003oool0oooo0200oooo0P00000S0?ooo`040000003oool0
oooo000001d0oooo00@000000?ooo`3oool00000i03oool001P0oooo00<000000?ooo`3oool0A03o
ool00`000000oooo0?ooo`0P0?ooo`8000008`3oool010000000oooo0?ooo`00000M0?ooo`040000
003oool0oooo00000>@0oooo000H0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0
803oool2000002<0oooo00@000000?ooo`3oool000007@3oool00`000000oooo0?ooo`0200000><0
oooo000H0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0803oool2000002<0oooo
00@000000?ooo`3oool000007@3oool010000000oooo0?ooo`00003T0?ooo`00603oool00`000000
oooo0?ooo`140?ooo`030000003oool0oooo0200oooo0P00000S0?ooo`040000003oool0oooo0000
01d0oooo00@000000?ooo`3oool00000i03oool001P0oooo00<000000?ooo`3oool0A03oool00`00
0000oooo0?ooo`0P0?ooo`8000008`3oool010000000oooo0?ooo`00000L0?ooo`050000003oool0
oooo0?ooo`000000i03oool001P0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`0P
0?ooo`8000008`3oool010000000oooo0?ooo`00000L0?ooo`050000003oool0oooo0?ooo`000000
i03oool001P0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`0O0?ooo`030000003o
ool0000002<0oooo00@000000?ooo`3oool00000703oool01@000000oooo0?ooo`3oool000000>@0
oooo000H0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool07`3oool00`000000oooo
0000000S0?ooo`040000003oool0oooo000001`0oooo00D000000?ooo`3oool0oooo0000003T0?oo
o`00603oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo01l0oooo00<000000?ooo`00
00008`3oool010000000oooo0?ooo`00000L0?ooo`050000003oool0oooo0?ooo`000000i03oool0
01P0oooo00<000000?ooo`3oool0A03oool200000200oooo00<000000?ooo`0000008`3oool01000
0000oooo0?ooo`00000L0?ooo`040000003oool0oooo0?ooo`800000h`3oool001P0oooo00<00000
0?ooo`3oool0A03oool200000200oooo00<000000?ooo`0000008`3oool010000000oooo0?ooo`00
000L0?ooo`050000003oool0oooo0?ooo`000000i03oool001P0oooo00<000000?ooo`3oool0A03o
ool200000200oooo00<000000?ooo`0000008`3oool010000000oooo0?ooo`00000L0?ooo`050000
003oool0oooo0?ooo`000000i03oool001P0oooo00<000000?ooo`3oool0A03oool200000200oooo
00<000000?ooo`0000008`3oool01@000000oooo0?ooo`3oool0000001/0oooo00D000000?ooo`3o
ool0oooo0000003T0?ooo`00603oool00`000000oooo0?ooo`140?ooo`800000803oool00`000000
oooo0000000S0?ooo`050000003oool0oooo0?ooo`0000001`3oool01P2oM0000000000000000000
04RL0=ooo`<0oooo00D0omnL04P000000000001d0;ooo`060?ooo`050000003oool0oooo0?ooo`00
0000i03oool001P0oooo00<000000?ooo`3oool0A03oool200000200oooo00<000000?ooo`000000
8`3oool01@000000oooo0?ooo`3oool0000000P0oooo00<0omnL04P0001d_ol0103oool01`3ookl0
M00007Boo`3oool0okmd0000M02oool01@3oool01@000000oooo0?ooo`3oool000000>@0oooo000H
0?ooo`030000003oool0oooo04@0oooo0P00000P0?ooo`030000003oool0000002<0oooo00D00000
0?ooo`3oool0oooo000000080?ooo`030?oOW0180000M;oo00@0oooo00<0_g@00018W03Oool00P3o
ool00`2oM00007Bo0?ooo`040?ooo`050000003oool0oooo0?ooo`000000i03oool001P0oooo00<0
00000?ooo`3oool0A03oool200000200oooo00<000000?ooo`0000008`3oool01@000000oooo0?oo
o`3oool0000000P0oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M00004RLg`030?ooo`030=nL
B00007@0_ooo00@0oooo00D000000?ooo`3oool0oooo0000003T0?ooo`00603oool00`000000oooo
0?ooo`140?ooo`800000803oool00`000000oooo0000000Q0?ooo`030?oog`2LB000000000L00000
00<0B9cO0?ooo`3oool00P3oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol0103oool01@000000oooo0?ooo`3oool000000>@0oooo000H0?ooo`030000
003oool0oooo04@0oooo0P00000P0?ooo`030000003oool0000002<0oooo00D000000?ooo`3oool0
oooo000000080?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`040?ooo`040000003oool0oooo0?ooo`<00000hP3oool001P0oooo00<000000?ooo`3o
ool0A03oool200000200oooo00<000000?ooo`0000008`3oool01@000000oooo0?ooo`3oool00000
00P0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00@0oooo00D000000?ooo`3oool0oooo0000003T0?ooo`00603oool00`000000oooo0?ooo`140?oo
o`800000803oool00`000000oooo0000000S0?ooo`050000003oool0oooo0?ooo`000000203oool0
0`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0103oool0
1@000000oooo0?ooo`3oool000000>@0oooo000H0?ooo`030000003oool0oooo04@0oooo0P00000P
0?ooo`030000003oool0000002<0oooo00D000000?ooo`3oool0oooo000000080?ooo`030?oOW018
0000M;oo00<0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`040?ooo`050000003o
ool0oooo0?ooo`000000i03oool001P0oooo00<000000?ooo`3oool0803oool00`000000oooo0?oo
o`0Q0?ooo`800000803oool00`000000oooo0000000S0?ooo`050000003oool0oooo0?ooo`000000
203oool00`3ogi`0B00007Boo`040?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0
103oool01@000000oooo0?ooo`3oool000000>@0oooo000H0?ooo`030000003oool0oooo0200oooo
00<000000?ooo`3oool08@3oool200000200oooo00<000000?ooo`0000008`3oool01@000000oooo
0?ooo`3oool0000000H0oooo00D0oooO09a800000000000007Boo`040?ooo`070?oo_`1d0000M;oo
0?ooo`3o_g@0001d0;ooo`050?ooo`050000003oool0oooo0?ooo`000000i03oool001P0oooo00<0
00000?ooo`3oool0803oool00`000000oooo0?ooo`0Q0?ooo`800000803oool00`000000oooo0000
000S0?ooo`050000003oool0oooo0?ooo`000000203oool00`3o_g@0000007Boo`050?ooo`050?oo
_`1d000000000000M02oool01@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0>80
oooo000H0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool08@3oool200000200oooo
00<000000?ooo`0000008`3oool01@000000oooo0?ooo`3oool0000001X0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`3R0?ooo`00603oool00`000000oooo0?ooo`0P0?ooo`030000
003oool0oooo0240oooo0P00000P0?ooo`030000003oool0000002<0oooo00D000000?ooo`3oool0
oooo0000000J0?ooo`030000003oool0oooo0080oooo0P00003S0?ooo`00603oool00`000000oooo
0?ooo`0P0?ooo`030000003oool0oooo0240oooo0P00000P0?ooo`030000003oool0000002<0oooo
00D000000?ooo`3oool0oooo0000000J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0hP3oool001P0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0Q0?ooo`800000
803oool00`000000oooo0000000S0?ooo`050000003oool0oooo0?ooo`0000006P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0>80oooo000H0?ooo`030000003oool0oooo0200oooo
00<000000?ooo`3oool08@3oool200000200oooo00<000000?ooo`0000008`3oool01@000000oooo
0?ooo`3oool0000001X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`3R0?ooo`00
603oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0240oooo0P00000P0?ooo`030000
003oool0000002<0oooo00D000000?ooo`3oool0oooo0000000J0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0hP3oool001P0oooo00<000000?ooo`3oool0803oool00`000000oooo
0?ooo`0Q0?ooo`800000803oool00`000000oooo0000000S0?ooo`050000003oool0oooo0?ooo`00
00006P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0>80oooo000H0?ooo`030000
003oool0oooo0200oooo00<000000?ooo`3oool08@3oool200000200oooo00<000000?ooo`000000
8`3oool01@000000oooo0?ooo`3oool0000001X0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`3R0?ooo`00603oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0240oooo
0P00000P0?ooo`030000003oool0000002<0oooo00D000000?ooo`3oool0oooo0000000J0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0hP3oool001P0oooo00<000000?ooo`3oool0
803oool00`000000oooo0?ooo`0Q0?ooo`800000803oool00`000000oooo0000000S0?ooo`050000
003oool0oooo0?ooo`0000006P3oool00`000000oooo0?ooo`020?ooo`800000h`3oool001P0oooo
00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0Q0?ooo`800000803oool00`000000oooo
0000000S0?ooo`050000003oool0oooo0?ooo`0000006P3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo0>80oooo000H0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0
8@3oool200000200oooo00<000000?ooo`0000008`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo01H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`3R0?ooo`00603o
ool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0240oooo0P00000P0?ooo`030000003o
ool0000002<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0F0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool0hP3oool001P0oooo00<000000?ooo`3oool0803oool0
0`000000oooo0?ooo`0Q0?ooo`800000803oool00`000000oooo0000000S0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0>80oooo000H0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool08@3oool20000
0200oooo00<000000?ooo`0000008`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
01H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`3R0?ooo`00603oool00`000000
oooo0?ooo`0P0?ooo`030000003oool0oooo0240oooo0P00000P0?ooo`030000003oool0000002<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0hP3oool001P0oooo00<000000?ooo`3oool0803oool00`000000oooo
0?ooo`0Q0?ooo`800000803oool00`000000oooo0000000S0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0>80oooo
000H0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool08@3oool2000001l0oooo00@0
00000?ooo`3oool000008`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01D0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`3R0?ooo`00603oool00`000000oooo0?oo
o`0P0?ooo`030000003oool0oooo0240oooo0P00000O0?ooo`040000003oool0oooo000002<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00<0oooo
0P00003S0?ooo`00603oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0240oooo0P00
000O0?ooo`040000003oool0oooo000002<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0E0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0hP3oool001P0oooo00<0
00000?ooo`3oool0803oool00`000000oooo0?ooo`0Q0?ooo`8000007`3oool010000000oooo0?oo
o`00000S0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool05@3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0>80oooo000H0?ooo`030000003oool0oooo0200oooo00<0
00000?ooo`3oool08@3oool2000001l0oooo00@000000?ooo`3oool000008`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`3R0?ooo`00603oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0240oooo0P00
000O0?ooo`040000003oool0oooo000002<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0D0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0hP3oool001P0oooo00<0
00000?ooo`3oool0803oool00`000000oooo0?ooo`0Q0?ooo`8000007`3oool010000000oooo0?oo
o`00000S0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0503oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo0>80oooo000H0?ooo`030000003oool0oooo0200oooo00<0
00000?ooo`3oool08@3oool2000001l0oooo00@000000?ooo`3oool000008`3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`3R0?ooo`00603oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0240oooo0P00
000O0?ooo`040000003oool0oooo000002<0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`0C0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0hP3oool001P0oooo00<0
00000?ooo`3oool0803oool00`000000oooo0?ooo`0Q0?ooo`8000007`3oool010000000oooo0?oo
o`00000S0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool04`3oool00`000000oooo
0?ooo`040?ooo`800000h`3oool001P0oooo00<000000?ooo`3oool0803oool00`000000oooo0?oo
o`0Q0?ooo`8000007`3oool010000000oooo0?ooo`00000S0?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0>80oooo
000H0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool08@3oool2000001l0oooo00@0
00000?ooo`3oool000008`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`3R0?ooo`00603oool00`000000oooo0?oo
o`0P0?ooo`030000003oool0oooo0240oooo0P00000O0?ooo`040000003oool0oooo000002<0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`090?ooo`030?oog`2LB000000000800000
00D0M;oo0?ooo`3oool0oooo000000070?ooo`030000003oool0oooo0>80oooo000H0?ooo`030000
003oool0oooo0200oooo00<000000?ooo`3oool08@3oool2000001l0oooo00@000000?ooo`3oool0
00008`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00P0oooo00<0oooO09a8M02o
ool00P3oool01@2oM000M;oo0?ooo`3oool0000000L0oooo00<000000?ooo`3oool0hP3oool001P0
oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0Q0?ooo`8000007`3oool010000000
oooo0?ooo`00000S0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool03@3oool01@3o
oml0W4Qd0;ooo`3oool0000000L0oooo00<000000?ooo`3oool0hP3oool001P0oooo00<000000?oo
o`3oool0803oool00`000000oooo0?ooo`0Q0?ooo`8000007`3oool010000000oooo0?ooo`00000S
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool03P3oool00`2oM000M;oo00000008
0?ooo`030000003oool0oooo0>80oooo000H0?ooo`030000003oool0oooo0200oooo00<000000?oo
o`3oool08@3oool2000001l0oooo00@000000?ooo`3oool000008`3oool00`000000oooo0?ooo`04
0?ooo`030?oog`000000000000L0000000<0B9cO0?ooo`3oool0103oool0103ooml0W4P004RLg`00
00080?ooo`030000003oool0oooo0>80oooo000H0?ooo`030000003oool0oooo0200oooo00<00000
0?ooo`3oool08@3oool2000001h0oooo00D000000?ooo`3oool0oooo0000000S0?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool0303oool0103OW4P0000007Boo`0000080?ooo`<00000
hP3oool001P0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0Q0?ooo`8000007P3o
ool01@000000oooo0?ooo`3oool0000002<0oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`080?ooo`070=oOg`3oool0oono07@000000000001d000000090?ooo`030000003oool0oooo
0>80oooo000H0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool08@3oool2000001h0
oooo00D000000?ooo`3oool0oooo0000000S0?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool0203oool01`3OW4P00000000000000000001809cOo`0000002@3oool00`000000oooo0?oo
o`3R0?ooo`00603oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0240oooo0P00000N
0?ooo`050000003oool0oooo0?ooo`0000008`3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00P0oooo00@0oono07@000000000B9cO0P3oool00`000000oooo0?ooo`070?ooo`030000
003oool0oooo0>80oooo000H0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool08@3o
ool2000001h0oooo00D000000?ooo`3oool0oooo0000000S0?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool02@3oool00`2oM000M;oo0?ooo`020?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool0hP3oool001P0oooo00<000000?ooo`3oool0803oool00`000000oooo0?oo
o`0Q0?ooo`030000003oool0000001d0oooo00D000000?ooo`3oool0oooo0000000S0?ooo`030000
003oool0oooo00L0oooo00<000000?ooo`3oool0203oool01P3ogi`0B00000000000000000180000
00T0oooo00<000000?ooo`3oool0hP3oool001P0oooo00<000000?ooo`3oool0803oool00`000000
oooo0?ooo`0Q0?ooo`030000003oool0000001d0oooo00D000000?ooo`3oool0oooo0000000S0?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool02@3oool01@2oM0000000000000000000
B9cO00T0oooo00<000000?ooo`3oool0hP3oool001P0oooo00<000000?ooo`3oool0803oool00`00
0000oooo0?ooo`0Q0?ooo`030000003oool0000001d0oooo00D000000?ooo`3oool0oooo0000000S
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo0>80oooo000H0?ooo`030000003oool0oooo0200oooo00<000000?oo
o`3oool08@3oool00`000000oooo0000000M0?ooo`050000003oool0oooo0?ooo`0000008`3oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`3R0?ooo`00603oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo
0240oooo00<000000?ooo`000000703oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
0240oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo
00X0oooo0P00003S0?ooo`00603oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0240
oooo00<000000?ooo`000000703oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0240
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool0hP3oool001P0oooo00<000000?ooo`3oool0803oool00`000000oooo
0?ooo`0Q0?ooo`030000003oool0000001`0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0Q0?ooo`030000003oool0oooo00/0oooo0P0000050?ooo`8000003@3oool00`000000oooo
0?ooo`3R0?ooo`00603oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0240oooo00<0
00000?ooo`000000703oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0240oooo00<0
00000?ooo`3oool0303oool6000000l0oooo00<000000?ooo`3oool0hP3oool001P0oooo00<00000
0?ooo`3oool0803oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0000001`0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<00000
0?ooo`3oool0hP3oool001P0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0Q0?oo
o`030000003oool0000001`0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0Q0?oo
o`030000003oool0oooo0240oooo00<000000?ooo`3oool0hP3oool001P0oooo00<000000?ooo`3o
ool0803oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0000001`0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3o
ool0hP3oool001P0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0Q0?ooo`030000
003oool0000001/0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0Q0?ooo`030000
003oool0oooo0240oooo00<000000?ooo`3oool0hP3oool001P0oooo00<000000?ooo`3oool0803o
ool00`000000oooo0?ooo`0Q0?ooo`030000003oool0000001/0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo0P00003S0?ooo`00603oool0
0`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0240oooo00<000000?ooo`0000006`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool08@3oool0
0`000000oooo0?ooo`3R0?ooo`00603oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo
0240oooo00@000000?ooo`3oool000006P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0240oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`3R0?ooo`00603oool00`00
0000oooo0?ooo`0P0?ooo`030000003oool0oooo00<0oooo00@0oooO09a80000B9`0gooo6P3oool0
10000000oooo0?ooo`00000J0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`2oM0001P00
00000`18W=l0oooo0?ooo`0H0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0@03o
ool00`2oM0000000000000040000000304RLg`3oool0oooo0480oooo00@0oooO09a80000B9`0gooo
DP3oool001P0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`030?ooo`040?oog`2L
B00004RL0=oooaX0oooo00@000000?ooo`3oool000006P3oool00`000000oooo0?ooo`030?ooo`04
0000003oool0oooo0=nLB0D0000000<004RL0=ooo`3oool0603oool00`000000oooo0?ooo`0Q0?oo
o`030000003oool0oooo0440oooo00<0gia80000000000000`0000000`00B9`0gooo0?ooo`120?oo
o`040?oog`2LB00004RL0=oooe80oooo000H0?ooo`030000003oool0oooo01l0oooo0P0000050?oo
o`040?oog`2LB00004RL0=oooaX0oooo00@000000?ooo`3oool000006@3oool00`000000oooo0?oo
o`040?ooo`060000003oool0oooo0?ooo`3OW4P0B9cO0`3oool00`3OW7@0W=oo0?ooo`0H0?ooo`03
0000003oool0oooo0240oooo00<000000?ooo`3oool0@P3oool00`3OW4P0B9cO0?ooo`020?ooo`03
0=nLM02Lgol0oooo0480oooo00@0oooO09a80000B9`0goooDP3oool001P0oooo00<000000?ooo`3o
ool07`3oool2000000030?ooo`2oM000000000H0000000<0M;oo0?ooo`3oool05`3oool010000000
oooo0?ooo`00000I0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`3O
W4P0M;oo0?ooo`0L0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0@`3oool00`3O
W4P0M;oo0?ooo`120?ooo`030;md00000000000000D0000000<0M;oo0?ooo`3oool0C`3oool001P0
oooo00<000000?ooo`3oool05@3oool00`3ooml0W4P000000009000000030?ooo`3ogi`0M9cO0080
oooo00@0oooO09a80000B9`0gooo6P3oool010000000oooo0?ooo`00000E0?ooo`030?oog`2LB000
000000L0000000<0B9cO0000003oool0103oool00`2oM4P0W=oo0?ooo`0K0?ooo`030000003oool0
oooo0240oooo00<000000?ooo`3oool0A03oool00`2oM4P0W=oo0?ooo`110?ooo`080?oOW01dW=l0
oooo0?ooo`3ooml0W4P00018W03OoomB0?ooo`00603oool00`000000oooo0?ooo`0O0?ooo`800000
0P3oool01`3OW4P0M;oo0?ooo`3ooml0W4P00018W03Oool06P3oool010000000oooo0?ooo`00000H
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool00`3ooml0W4Qd0;ooo`0K
0?ooo`030000003oool0oooo0240oooo0P0000150?ooo`030?oog`2LB7@0_ooo0480oooo00L0gia8
07Boo`3oool0oooO09a80000B9`0gooo0580oooo000H0?ooo`030000003oool0oooo01l0oooo0P00
00020?ooo`070?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`0J0?ooo`040000003oool0oooo0000
01L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`040?ooo`030?noM000M;l0oooo
01X0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`150?ooo`030?noM000M;l0oooo
0440oooo00L0oooO09a8M02oool0oooO09a80000B9`0gooo0580oooo000H0?ooo`030000003oool0
oooo01l0oooo0P0000030?ooo`060?oOW018M;l0oooO09a80000B9`0gooo6P3oool01@000000oooo
0?ooo`3oool0000001H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`040?ooo`03
0?oo_`1d04P0W=oo01X0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`150?ooo`03
0?oo_`1d04P0W=oo0480oooo00H0omnL04Qd_`3ooml0W4P00018W03OoomB0?ooo`00603oool00`00
0000oooo0?ooo`0O0?ooo`800000103oool01@3OW4P0M;oO09a80000B9`0gooo01X0oooo00D00000
0?ooo`3oool0oooo0000000E0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0103o
ool00`3ookl0M00004RLg`0J0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0A@3o
ool00`3ookl0M00004RLg`130?ooo`050=nLB01d_ml0W4P00018W03Oool0DP3oool001P0oooo00<0
00000?ooo`3oool07`3oool2000000@0oooo00D0oooO09a8B018B00004RL0=ooo`0J0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`080?ooo`040000
003oool0oknL09cOo`<0oooo00<0omnL04P00018W=l06P3oool00`000000oooo0?ooo`0Q0?ooo`03
0000003oool0oooo0400oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`130?ooo`05
0?oog`2LB4P0B4P00018W03Oool0DP3oool001P0oooo00<000000?ooo`3oool07`3oool2000000D0
oooo00@0omnL04P00000B9`0gooo6P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool02@3oool02P000000oooo0?oog`2LB4P0W=oo0?ooo`3ookl0M000
0000M02ooolJ0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0@03oool0203ooml0
W4Q809cOo`3oool0oono07@0000007@0_oooA03oool0103ogi`0B0000018W03OoomB0?ooo`00603o
ool00`000000oooo0?ooo`0O0?ooo`8000001P3oool00`3OW4P004RL0=ooo`0J0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0:0?ooo`050000003o
ool0oooo0?oog`2LB0000P0000000`0007@0_ooo0?ooo`0J0?ooo`030000003oool0oooo0240oooo
00<000000?ooo`3oool0@@3oool01P3ooml0W4P0000000000000001d0;ooodH0oooo00<0gia80018
W03Oool0DP3oool001P0oooo00<000000?ooo`3oool07`3oool2000002<0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0>80oooo000H0?ooo`030000
003oool0oooo01l0oooo0P00000S0?ooo`030000003oool0oooo00D0oooo0P00000:0?ooo`<00000
3@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo0P00003S0?ooo`00603o
ool00`000000oooo0?ooo`0O0?ooo`8000008`3oool00`000000oooo0?ooo`070?ooo`X00000403o
ool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0hP3o
ool001P0oooo00<000000?ooo`3oool07P3oool00`000000oooo0000000S0?ooo`030000003oool0
oooo0240oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0
oooo0>80oooo000H0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`0000008`3oool00`00
0000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool08@3oool00`00
0000oooo0?ooo`3R0?ooo`00603oool00`000000oooo0?ooo`0N0?ooo`030000003oool0000002<0
oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240
oooo00<000000?ooo`3oool0aP3oool01@2oM0000000001d_`3oool0gia80080000000<004RL0=oo
o`3oool04P3oool001P0oooo00<000000?ooo`3oool07@3oool010000000oooo0?ooo`00000S0?oo
o`030000003oool0oooo0240oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?oo
o`030000003oool0oooo0<H0oooo00P0oooO09a8M02oool0oooo0?oog`2LB00004RL0=oooa@0oooo
000H0?ooo`030000003oool0oooo01`0oooo0P0000020?ooo`030000003oool0oooo0240oooo00<0
00000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<0
00000?ooo`3oool0a`3oool01P3ookl0M4RL0=ooo`2oM000001d0;oooaD0oooo000I0?ooo`030000
003oool0oooo01P0oooo0`0000040?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0
8@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0
b03oool0103ogi`0M4Q80000001d_olF0?ooo`006P3ooolJ000000L0oooo00<000000?ooo`3oool0
8@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0
8@3oool00`000000oooo0?ooo`380?ooo`030?oo_`1d0000B9cO01L0oooo000>0?ooool00000N`00
000=0?ooo`050?oog`2LB00004RL09adB02Lgol05P3oool001L0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01@0oooo00H0_g@00000B02Lgol0oooO
09a8M02ooolE0?ooo`00>`3oool00`000000oooo0?ooo`040?ooo`h000003`3oool00`000000oooo
0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool08@3oool00`000000oooo
0?ooo`140?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0=P3oool02@3ooml0W4P0
00000000000007Bo0?noM0000000001d0;ooo`0D0?ooo`00>`3oool010000000oooo0?ooo`3oool3
000000h0oooo1000000;0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool08@3oool0
0`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0>80oooo000k0?ooo`030000003oool0oooo
00800000503oool3000000P0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?oo
o`030000003oool0oooo0240oooo00<000000?ooo`3oool0hP3oool003/0oooo00@000000?ooo`3o
ool00000603oool2000000H0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?oo
o`030000003oool0oooo0240oooo00<000000?ooo`3oool0hP3oool003/0oooo00<000000?ooo`00
00006`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0240oooo00<000000?ooo`3o
ool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0>80oooo000k0?ooo`030000
003oool0000001/0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0Q0?ooo`030000
003oool0oooo0240oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0W0?oooa@00000
Y`3oool003/0oooo00<000000?ooo`000000703oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo0240oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`800000803o
ool8000001@0oooo3@00002J0?ooo`00>`3oool00`000000oooo0000000L0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0
oooo0240oooo00<000000?ooo`3oool06P3oool5000002T0oooo1`00002C0?ooo`00>`3oool00`00
0000oooo0000000M0?ooo`050000003oool0oooo0?ooo`0000008`3oool00`000000oooo0?ooo`0Q
0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool05P3oool4000003D0oooo1@00002>
0?ooo`00>`3oool2000001h0oooo00D000000?ooo`3oool0oooo0000000S0?ooo`030000003oool0
oooo0240oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0D0?ooo`800000?P3oool4
000008X0oooo000k0?ooo`8000007P3oool01@000000oooo0?ooo`3oool0000002<0oooo00<00000
0?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0180oooo0P000014
0?ooo`<00000Q`3oool003/0oooo0P00000N0?ooo`050000003oool0oooo0?ooo`0000008`3oool0
0`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0403oool2
000004T0oooo100000230?ooo`00>`3oool2000001l0oooo00@000000?ooo`3oool000008`3oool0
0`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`1=0?ooo`@00000O`3oool003/0oooo0P00000O0?ooo`040000003oool0oooo
000002<0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool0DP3oool4000007/0oooo000k0?ooo`8000007`3oool01000
0000oooo0?ooo`00000S0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool08@3oool2
000000h0oooo00<000000?ooo`3oool0E`3oool2000007T0oooo000k0?ooo`8000007`3oool01000
0000oooo0?ooo`00000S0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool08@3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo05X0oooo0P00001g0?ooo`00>`3oool20000
01l0oooo00@000000?ooo`3oool000008`3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0
oooo0240oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1L0?ooo`800000M@3oool0
03/0oooo0P00000P0?ooo`030000003oool0000002<0oooo00<000000?ooo`3oool02P3oool50000
0180oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
05l0oooo0P00001c0?ooo`00>`3oool200000200oooo00<000000?ooo`0000008`3oool00`000000
oooo0?ooo`080?ooo`8000001@3oool200000100oooo00<000000?ooo`3oool08@3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo0680oooo0P00001a0?ooo`00>`3oool200000200oooo
00<000000?ooo`0000008`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo
0P00000>0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`1U0?ooo`800000K`3oool003/0oooo0P00000P0?ooo`030000003oool0000002<0oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool08@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo06P0oooo00<0
00000?ooo`3oool0K03oool003/0oooo0P00000P0?ooo`030000003oool0000002<0oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool08@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo06T0oooo0P00001/
0?ooo`00>`3oool200000200oooo00<000000?ooo`0000008`3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0Q0?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool0K03oool2000006X0oooo000k0?ooo`80
0000803oool00`000000oooo0000000S0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0240oooo0P0000080?ooo`03
0000003oool0oooo06h0oooo00<000000?ooo`3oool0I`3oool003/0oooo0P00000P0?ooo`030000
003oool0000002<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo00T0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`070?ooo`030000
003oool0oooo06l0oooo00<000000?ooo`3oool0IP3oool003/0oooo0P00000P0?ooo`030000003o
ool0000002<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo0700oooo00<000000?ooo`3oool0I@3oool003/0oooo0P00000P0?ooo`030000003oool0
000002<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`060?ooo`030000003oool0
oooo0780oooo00<000000?ooo`3oool0I03oool003/0oooo0P00000P0?ooo`030000003oool00000
02<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
00P0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo
07<0oooo00<000000?ooo`3oool0H`3oool003/0oooo0P00000P0?ooo`030000003oool0000002<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo07@0
oooo00<000000?ooo`3oool0HP3oool003/0oooo0P00000P0?ooo`030000003oool0000002<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo07D0oooo
00<000000?ooo`3oool0H@3oool003/0oooo0P00000Q0?ooo`8000008`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`0Q0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?oo
o`1P0?ooo`00>`3oool200000240oooo0P00000S0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool04`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0240oooo0P000006
0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0G`3oool003/0oooo0P00000Q0?oo
o`8000008`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01<0oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool0N@3oool00`000000oooo0?ooo`1N0?ooo`00>`3oool200000240oooo0P00000S0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo0240oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`1j0?ooo`03
0000003oool0oooo05d0oooo000k0?ooo`8000008@3oool2000002<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
403oool00`3ooml0W4P0000000020000000307Boo`3oool0oooo00T0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo05d0oooo000k0?ooo`8000008@3o
ool2000002<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0D0?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool0403oool00`3ooml0W4Qd0;ooo`020?ooo`030;md001d
_ol0oooo00T0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`1l0?ooo`030000003o
ool0oooo05`0oooo000k0?ooo`8000008@3oool2000002<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool05@3oool0
0`3ooml0W4Qd0;ooo`090?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0O@3oool0
0`000000oooo0?ooo`1K0?ooo`00>`3oool200000240oooo0P00000S0?ooo`050000003oool0oooo
0?ooo`0000005`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01H0oooo00<0_g@0
07Boo`3oool0203oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo07h0oooo00<00000
0?ooo`3oool0FP3oool003/0oooo0P00000Q0?ooo`8000008`3oool01@000000oooo0?ooo`3oool0
000001L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0E0?ooo`030?oog`2LB000
B9cO00T0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`1n0?ooo`030000003oool0
oooo05X0oooo000k0?ooo`8000008@3oool2000002<0oooo00D000000?ooo`3oool0oooo0000000G
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool05@3oool00`3OW4P0000007Boo`09
0?ooo`<000000`3oool00`000000oooo0?ooo`200?ooo`030000003oool0oooo05T0oooo000k0?oo
o`8000008@3oool2000002<0oooo00D000000?ooo`3oool0oooo0000000H0?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool04@3oool01`3Ogml0oooo0?oo_`1d000000000000M02oool0
2@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool0
F03oool003/0oooo0P00000Q0?ooo`8000008`3oool01@000000oooo0?ooo`3oool0000001P0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0A0?ooo`060=nLB0000000000000000000
04P0W=oo2P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0880oooo00<000000?oo
o`3oool0E`3oool003/0oooo0P00000Q0?ooo`8000008`3oool010000000oooo0?ooo`00000I0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool04@3oool0103ookl0M00000000018W=l<
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0PP3oool00`000000oooo0?ooo`1G
0?ooo`00>`3oool200000240oooo0P00000S0?ooo`040000003oool0oooo000001T0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`0B0?ooo`030;md001d_ol0oooo00`0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`230?ooo`030000003oool0oooo05H0oooo000k0?oo
o`8000008@3oool2000002<0oooo00@000000?ooo`3oool000006@3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0180oooo00H0omnL04P00000000000000000B02Lgol90?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`1E0?ooo`00>`3o
ool200000240oooo0P00000S0?ooo`040000003oool0oooo000001T0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0C0?ooo`050;md00000000000000000018W=l02@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool0E@3oool003/0oooo
0P00000Q0?ooo`8000008`3oool010000000oooo0?ooo`00000J0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo08H0
oooo00<000000?ooo`3oool0E03oool003/0oooo0P00000Q0?ooo`8000008`3oool010000000oooo
0?ooo`00000J0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool08@3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3oool0E03oool003/0oooo
0P00000Q0?ooo`8000008`3oool010000000oooo0?ooo`00000J0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool08@3oool2000000<0oooo00<000000?ooo`3oool0Q`3oool00`000000
oooo0?ooo`1C0?ooo`00>`3oool200000240oooo0P00000S0?ooo`040000003oool0oooo000001X0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`1B0?ooo`00>`3oool200000240oooo
0P00000S0?ooo`040000003oool0oooo000001X0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0Q0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0R03oool00`000000
oooo0?ooo`1B0?ooo`00>`3oool200000240oooo0P00000S0?ooo`040000003oool0oooo000001X0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0R@3oool00`000000oooo0?ooo`1A0?ooo`00>`3oool200000240oooo
0P00000S0?ooo`040000003oool0oooo000001X0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0Q0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0R@3oool00`000000
oooo0?ooo`1A0?ooo`00>`3oool200000240oooo0P00000S0?ooo`040000003oool0oooo000001X0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0RP3oool00`000000oooo0?ooo`1@0?ooo`00>`3oool200000240oooo
0P00000S0?ooo`040000003oool0oooo000001X0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0Q0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0RP3oool00`000000
oooo0?ooo`1@0?ooo`00>`3oool200000240oooo0P00000S0?ooo`040000003oool0oooo000001/0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0R`3oool00`000000oooo0?ooo`1?0?ooo`00>`3oool200000240oooo
0P00000S0?ooo`040000003oool0oooo000001/0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0Q0?ooo`8000000`3oool00`000000oooo0?ooo`2;0?ooo`030000003oool0oooo04l0
oooo000k0?ooo`8000008@3oool2000002<0oooo00@000000?ooo`3oool000006`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`2<0?ooo`030000003oool0oooo04h0oooo000k0?ooo`8000008@3oool2000002<0oooo
00@000000?ooo`3oool000006`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0240
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`2<0?ooo`030000003oool0oooo04h0
oooo000k0?ooo`8000008@3oool2000002<0oooo00@000000?ooo`3oool000006`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0240oooo00D000000?ooo`3oool0oooo0000002@0?oo
o`030000003oool0oooo04d0oooo000k0?ooo`8000008@3oool2000002<0oooo00@000000?ooo`3o
ool000006`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0240oooo00D000000?oo
o`3oool0oooo0000002@0?ooo`030000003oool0oooo04d0oooo000k0?ooo`8000008@3oool20000
02<0oooo00@000000?ooo`3oool000006`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo0240oooo00D000000?ooo`3oool0oooo0000002A0?ooo`030000003oool0oooo04`0oooo000k
0?ooo`8000008@3oool2000002<0oooo00@000000?ooo`3oool000006`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo0240oooo00D000000?ooo`3oool0oooo0000002A0?ooo`030000
003oool0oooo04`0oooo000k0?ooo`8000008@3oool2000002<0oooo00@000000?ooo`3oool00000
703oool01@000000oooo0?ooo`3oool0000002<0oooo00D000000?ooo`3oool0oooo0000002B0?oo
o`030000003oool0oooo04/0oooo000k0?ooo`8000008@3oool2000002<0oooo00@000000?ooo`3o
ool00000703oool01@000000oooo0?ooo`3oool0000002<0oooo0P0000020?ooo`030000003oool0
oooo0900oooo00<000000?ooo`3oool0B`3oool003/0oooo0P00000Q0?ooo`8000008`3oool01000
0000oooo0?ooo`00000L0?ooo`050000003oool0oooo0?ooo`0000008`3oool01@000000oooo0?oo
o`3oool0000009<0oooo00<000000?ooo`3oool0BP3oool003/0oooo0P00000Q0?ooo`8000008`3o
ool00`000000oooo0000000M0?ooo`050000003oool0oooo0?ooo`0000008`3oool01@000000oooo
0?ooo`3oool0000009<0oooo00<000000?ooo`3oool0BP3oool003/0oooo00<000000?ooo`3oool0
803oool2000002<0oooo00<000000?ooo`0000007@3oool01@000000oooo0?ooo`3oool0000002<0
oooo00D000000?ooo`3oool0oooo0000002D0?ooo`030000003oool0oooo04T0oooo000k0?ooo`03
0000003oool0oooo0200oooo0P00000S0?ooo`030000003oool0000001d0oooo00D000000?ooo`3o
ool0oooo0000000S0?ooo`040000003oool0oooo000009D0oooo00<000000?ooo`3oool0B@3oool0
03/0oooo00<000000?ooo`3oool0803oool2000002<0oooo00<000000?ooo`0000007@3oool01@00
0000oooo0?ooo`3oool0000002<0oooo00@000000?ooo`3oool00000U@3oool00`000000oooo0?oo
o`190?ooo`00>`3oool00`000000oooo0?ooo`0P0?ooo`8000008`3oool00`000000oooo0000000M
0?ooo`050000003oool0oooo0?ooo`0000008`3oool010000000oooo0?ooo`00002F0?ooo`030000
003oool0oooo04P0oooo000k0?ooo`030000003oool0oooo0200oooo0P00000S0?ooo`030000003o
ool0000001d0oooo00D000000?ooo`3oool0oooo0000000S0?ooo`040000003oool0oooo000009H0
oooo00<000000?ooo`3oool0B03oool003/0oooo00<000000?ooo`3oool0803oool2000002<0oooo
00<000000?ooo`0000007@3oool01@000000oooo0?ooo`3oool0000002<0oooo00@000000?ooo`3o
ool00000UP3oool00`000000oooo0?ooo`180?ooo`00>`3oool00`000000oooo0?ooo`0P0?ooo`80
00008`3oool00`000000oooo0000000M0?ooo`050000003oool0oooo0?ooo`0000008`3oool20000
00030?ooo`000000oooo09H0oooo00<000000?ooo`3oool0A`3oool003/0oooo00<000000?ooo`3o
ool0803oool2000002<0oooo00<000000?ooo`0000007@3oool01@000000oooo0?ooo`3oool00000
02<0oooo00@000000?ooo`3oool00000U`3oool00`000000oooo0?ooo`170?ooo`00>`3oool00`00
0000oooo0?ooo`0P0?ooo`8000008`3oool00`000000oooo0000000M0?ooo`050000003oool0oooo
0?ooo`0000008`3oool010000000oooo0?ooo`00002G0?ooo`030000003oool0oooo04L0oooo000k
0?ooo`030000003oool0oooo0200oooo0P00000S0?ooo`030000003oool0000001d0oooo00D00000
0?ooo`3oool0oooo0000000;0?ooo`060;md00000000000000000000B9`0gooo0`3oool01@3ogi`0
B0000000000007@0_ooo00X0oooo00@000000?ooo`3oool00000V03oool00`000000oooo0?ooo`16
0?ooo`00>`3oool00`000000oooo0?ooo`0P0?ooo`8000008`3oool00`000000oooo0000000M0?oo
o`050000003oool0oooo0?ooo`000000303oool00`3ogi`0B00007Boo`040?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`040000003oool0oooo000009P0oooo00<000000?ooo`3o
ool0AP3oool003/0oooo00<000000?ooo`3oool0803oool2000002<0oooo00<000000?ooo`000000
7@3oool01@000000oooo0?ooo`3oool0000000`0oooo00<0omnL04P0001d_ol0103oool00`2oM000
04RL0=ooo`020?ooo`030;md0000M;l0oooo00P0oooo00@000000?ooo`3oool00000V03oool00`00
0000oooo0?ooo`160?ooo`00>`3oool00`000000oooo0?ooo`0P0?ooo`8000008`3oool00`000000
oooo0000000M0?ooo`050000003oool0oooo0?ooo`000000303oool00`3ogi`0B00007Boo`030?oo
o`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool010000000oooo0?ooo`00002H
0?ooo`030000003oool0oooo04H0oooo000k0?ooo`030000003oool0oooo0200oooo0P00000S0?oo
o`030000003oool0000001d0oooo00D000000?ooo`3oool0oooo0000000<0?ooo`030?oOW0180000
M;oo00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`080?ooo`040000003oool0
oooo000009T0oooo00<000000?ooo`3oool0A@3oool003/0oooo00<000000?ooo`3oool0803oool2
000002<0oooo00<000000?ooo`0000007P3oool010000000oooo0?ooo`00000<0?ooo`030?oOW018
0000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`080?ooo`@00000V@3o
ool00`000000oooo0?ooo`150?ooo`00>`3oool00`000000oooo0?ooo`0P0?ooo`8000008`3oool0
0`000000oooo0000000N0?ooo`040000003oool0oooo000000`0oooo00<0omnL04P0001d_ol00`3o
ool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00P0oooo00@000000?ooo`3oool00000
V@3oool00`000000oooo0?ooo`150?ooo`00>`3oool00`000000oooo0?ooo`0P0?ooo`8000008`3o
ool00`000000oooo0000000N0?ooo`040000003oool0oooo000000`0oooo00<0omnL04P0001d_ol0
0`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00P0oooo00@000000?ooo`3oool0
0000VP3oool00`000000oooo0?ooo`140?ooo`00>`3oool00`000000oooo0?ooo`0Q0?ooo`030000
003oool0oooo0240oooo00<000000?ooo`0000007P3oool010000000oooo0?ooo`00000<0?ooo`03
0?oOW0180000M;oo00<0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`080?ooo`04
0000003oool0oooo000009X0oooo00<000000?ooo`3oool0A03oool003/0oooo00<000000?ooo`3o
ool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0000001h0oooo00@000000?ooo`3o
ool00000303oool00`3ogi`0B00007Boo`040?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d
_`3oool0203oool010000000oooo0?ooo`00002J0?ooo`030000003oool0oooo04@0oooo000k0?oo
o`030000003oool0oooo0240oooo00<000000?ooo`3oool08@3oool00`000000oooo0000000N0?oo
o`040000003oool0oooo000000X0oooo00D0oooO09a800000000000007Boo`040?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`040000003oool0oooo000009/0oooo00<000000?oo
o`3oool0@`3oool003/0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`03
0000003oool0000001h0oooo00@000000?ooo`3oool00000303oool00`3o_g@0000007Boo`050?oo
o`050?oo_`1d000000000000M02oool02P3oool010000000oooo0?ooo`00002K0?ooo`030000003o
ool0oooo04<0oooo000k0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool08@3oool0
0`000000oooo0000000N0?ooo`040000003oool0oooo000002<0oooo00@000000?ooo`3oool00000
V`3oool00`000000oooo0?ooo`130?ooo`00>`3oool00`000000oooo0?ooo`0Q0?ooo`030000003o
ool0oooo0240oooo00<000000?ooo`0000007P3oool010000000oooo0?ooo`00000S0?ooo`800000
00<0oooo0000003oool0V`3oool00`000000oooo0?ooo`120?ooo`00>`3oool00`000000oooo0?oo
o`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`0000007P3oool010000000oooo0?oo
o`00000S0?ooo`040000003oool0oooo000009`0oooo00<000000?ooo`3oool0@P3oool003/0oooo
00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0000001h0oooo
00@000000?ooo`3oool000008`3oool010000000oooo0?ooo`00002L0?ooo`030000003oool0oooo
0480oooo000k0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool08@3oool00`000000
oooo0000000N0?ooo`040000003oool0oooo000002<0oooo00@000000?ooo`3oool00000W03oool0
0`000000oooo0?ooo`120?ooo`00>`3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo
0240oooo00<000000?ooo`0000007P3oool010000000oooo0?ooo`00000S0?ooo`040000003oool0
oooo000009d0oooo00<000000?ooo`3oool0@@3oool003/0oooo00<000000?ooo`3oool08@3oool0
0`000000oooo0?ooo`0Q0?ooo`030000003oool0000001h0oooo00@000000?ooo`3oool000008`3o
ool010000000oooo0?ooo`00002M0?ooo`030000003oool0oooo0440oooo000k0?ooo`030000003o
ool0oooo0240oooo00<000000?ooo`3oool08@3oool00`000000oooo0000000N0?ooo`040000003o
ool0oooo000002<0oooo00@000000?ooo`3oool00000W@3oool00`000000oooo0?ooo`110?ooo`00
>`3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`000000
7P3oool010000000oooo0?ooo`00000S0?ooo`040000003oool0oooo000009d0oooo00<000000?oo
o`3oool0@@3oool003/0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`03
0000003oool0000001h0oooo00@000000?ooo`3oool000008`3oool010000000oooo0?ooo`00002N
0?ooo`030000003oool0oooo0400oooo000k0?ooo`030000003oool0oooo0240oooo00<000000?oo
o`3oool08@3oool00`000000oooo0000000N0?ooo`040000003oool0oooo000002<0oooo0P000000
0`3oool000000?ooo`2M0?ooo`030000003oool0oooo0400oooo000k0?ooo`030000003oool0oooo
0240oooo00<000000?ooo`3oool08@3oool00`000000oooo0000000N0?ooo`040000003oool0oooo
000002<0oooo00@000000?ooo`3oool00000WP3oool00`000000oooo0?ooo`100?ooo`00>`3oool0
0`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`0000007P3oool0
10000000oooo0?ooo`00000S0?ooo`040000003oool0oooo000009h0oooo00<000000?ooo`3oool0
@03oool003/0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003o
ool0000001h0oooo00@000000?ooo`3oool000008`3oool010000000oooo0?ooo`00002N0?ooo`03
0000003oool0oooo0400oooo000k0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0
8@3oool00`000000oooo0000000N0?ooo`040000003oool0oooo000002<0oooo00@000000?ooo`3o
ool00000W`3oool00`000000oooo0?ooo`0o0?ooo`00>`3oool00`000000oooo0?ooo`0Q0?ooo`03
0000003oool0oooo0240oooo00<000000?ooo`0000007P3oool010000000oooo0?ooo`00000S0?oo
o`040000003oool0oooo000009l0oooo00<000000?ooo`3oool0?`3oool003/0oooo00<000000?oo
o`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`8000007`3oool010000000oooo0?ooo`00000S
0?ooo`040000003oool0oooo000009l0oooo00<000000?ooo`3oool0?`3oool003/0oooo00<00000
0?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`8000007`3oool010000000oooo0?ooo`00
000S0?ooo`030000003oool000000:00oooo00<000000?ooo`3oool0?`3oool003/0oooo00<00000
0?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`8000007`3oool010000000oooo0?ooo`00
000S0?ooo`<00000X@3oool00`000000oooo0?ooo`0n0?ooo`00>`3oool00`000000oooo0?ooo`0Q
0?ooo`030000003oool0oooo0240oooo0P00000O0?ooo`040000003oool0oooo000002<0oooo00<0
00000?ooo`000000X@3oool00`000000oooo0?ooo`0n0?ooo`00>`3oool00`000000oooo0?ooo`0Q
0?ooo`030000003oool0oooo0240oooo0P00000O0?ooo`040000003oool0oooo000002<0oooo00<0
00000?ooo`000000X@3oool00`000000oooo0?ooo`0n0?ooo`00>`3oool00`000000oooo0?ooo`0Q
0?ooo`030000003oool0oooo0240oooo0P00000O0?ooo`040000003oool0oooo000002<0oooo00<0
00000?ooo`000000X@3oool00`000000oooo0?ooo`0n0?ooo`00>`3oool00`000000oooo0?ooo`0Q
0?ooo`030000003oool0oooo0240oooo0P00000O0?ooo`040000003oool0oooo000002<0oooo00<0
00000?ooo`000000XP3oool00`000000oooo0?ooo`0m0?ooo`00>`3oool00`000000oooo0?ooo`0Q
0?ooo`030000003oool0oooo0240oooo0P00000O0?ooo`040000003oool0oooo000002<0oooo00<0
00000?ooo`000000XP3oool00`000000oooo0?ooo`0m0?ooo`00>`3oool00`000000oooo0?ooo`0Q
0?ooo`030000003oool0oooo0240oooo0P00000O0?ooo`040000003oool0oooo000002<0oooo00<0
00000?ooo`000000XP3oool00`000000oooo0?ooo`0m0?ooo`00o`3ooona0?ooo`00o`3ooona0?oo
o`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?oo
o`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00S`3oool00`3ooml0W4P000000003
000000060000M02oool0oooo0?ooo`3ooml0W4P00P0000001@00B9`0W7@000000000M;l0gia80080
000000<0001809cOg`2LB0000P0000000`0007@0_ooO09a8000200000007001d_`3OW4P000000000
000004P0W=oO09a80002000000030000M02ooml0W4P00080000000<007Bo0?oog`2LB0000P000000
1000B9`0W7@000000000M;l40?ooo`030?oog`2LB00000000080000000<0001809cOo`3oool0103o
ool02@3ookl0M000000000000000B9bo09a80000B9`0gooo0;md0002000000030000M02ooml0W4P0
0080000000@004RL0=ooo`3oool0_g@00P0000002@0004P0W=oo0?ooo`3oool0omnL04P000000000
B9cO0=nLB002000000040000M02oool0oooo0=nLB0<0000000@0M;oo0?ooo`3oool0_g@00P000000
0`0007@0_ooO09a80002000000030018W03Oool0oooo0:D0oooo002>0?ooo`040?noM00000000000
07Boo`<0oooo0100gia80000001d_ol0oooo0?noM0000000M;oo0?ooo`3ogi`0B000001d_`3o_kl0
_ooo0=nLB00007@0_ooo0P3oool01`2oM00004RL0=ooo`3oool0oooO09a80000M;l00P3oool00`3O
W4P0001d0;ooo`020?ooo`0B0;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0okmd0000001d
_ol0oooo0?oOW018000007Bo0?no_`2oool0gia80`000000101d_ol0omnL04P00018W=l70?ooo`05
0<AO5000000000000000001VM4P00P0000002@0007@0_ooo0=nLB000B9`0gooo0?ooo`3ooml0W4P0
001d_`020?ooo`0H0;md0000000004RL0=ooo`3ooml0W4Qd0;ooo`3oool0gia80000M02oool0gkno
0=ooo`3OW4P0001d0;ooo`3oool0gia800000018W=l0oooo0?oog`2LB000B9cO0P3oool01`3OW4P0
04RL0=ooo`3oool0oooO09a80000M;l0Z03oool008d0oooo00@0okmd0000000007@0_ooo103oool0
2`3o_g@0000007Boo`3oool0okmd0000B02Lgol0oooo0?oog`2LB00007Bo0080oooo00<0gia80000
M02oool00P3oool01`2oM00004RL0=ooo`3oool0oooO09a80000M;l00P3oool00`3OW4P0001d0;oo
o`020?ooo`0?0;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0okmd0000B02Lgol0oooo0?oo
g`2LB00007Bo0080oooo00<0gia80000M02oool00P3oool00`3ookl0M00004RLg`070?ooo`070;md
0000B9`0gono07@0001d_ml0W4P0001d_`020?ooo`0;0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d
_`3oool0omnL04P00000M;l00`3oool01P3ogi`0W;oo0?ooo`3OW4P004RL0=ooo`80oooo00H0gia8
0000M02oool0omnL04P00000M;l30?ooo`0;0?noM00007@0_ooo0?ooo`3OW4P004RL0=ooo`3oool0
oooO09a80000M;l0Z03oool008`0oooo00@0oooO09a800000000M;oo1@3oool00`3o_g@0000007Bo
o`020?ooo`060?noM000000007Bo0?oog`2LB00007Bo0P3oool00`3OW4P0001d0;ooo`020?ooo`07
0;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`030=nLB00007@0_ooo0080oooo00L0_g@0
0018W03Oool0oooo0?oog`2LB00007Bo0080oooo00H0okmd00000000M;l0oooO09a80000M;l20?oo
o`030=nLB000B9`0gooo0080oooo00<0oono07@00018W=l01`3oool01`2oM00004RL0=oo_`1d0000
M;oO09a80000M;l00P3oool02`3OW4P004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM00004P0
W=oo00H0oooo00<0gia80018W03Oool00P3oool01P3OW4P0001d0;ooo`3o_g@0000007Boo`<0oooo
00/0okmd0000001d_ol0oooo0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`2X0?ooo`00S03oool0
103ogi`0B0000018W03Oool50?ooo`030?noM0000000M;oo00@0oooo00@0okmd00000000000007Bo
0P3oool00`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`020?oo
o`030=nLB00007@0_ooo0080oooo00L0_g@00018W03Oool0oooo0?oog`2LB00007Bo00@0oooo00@0
okmd00000000000007Bo0P3oool00`3OW4P004RL0=ooo`020?ooo`030?oo_`1d0000B9cO0080oooo
00<0omnL07BLg`3oool00P3oool01`2oM00004RL0=oo_`1d0000M;oO09a80000M;l00P3oool02`3O
W4P004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM00007@0_ooo00H0oooo00<0gia80018W03O
ool00P3oool01P3OW4P0001d0;ooo`3o_g@0000007Boo`<0oooo00/0okmd0000001d_ol0oooo0=nL
B000B9`0gooo0?ooo`3ooml0W4P0001d_`2X0?ooo`00S03oool00`3o_g@00000001d_`060?ooo`0;
0?noM0000000M;oo0?ooo`3ookl0M00007Boo`3oool0oooO09a80000M;l00P3oool00`3OW4P00018
09cOo`020?ooo`070;md000007@0_ooo0?ooo`3ooml0W4P0001d_`020?ooo`030=nLB00004P0W=oo
0080oooo00L0_g@00000M02oool0oooo0?oog`2LB00005MG00@0000000@007Bo0?oog`2LB00007Bo
0P3oool00`3OW4P004RL0=ooo`020?ooo`050;md0000000004RX0?ooo`3OW4P00P0000002P00B9`0
gooo0?ooo`2oM00004RL0=oo_`1d0000M;oO09a80000M;l20?ooo`0;0;md000007@0_ooo0?ooo`3o
oml0W4P0001d_`3oool0omnL04P0B02Lgol00P3oool01`3o_g@0001809cOo`3oool0gia80018W03O
ool00P3oool01P3OW4P0001d0;ooo`3ookl0M00007Boo`<0oooo00/0_g@00000B02Lgol0oooo0;md
000007@0_ooo0?ooo`3ooml0W4P0001d_`2X0?ooo`00S03oool00`3o_g@00000001d_`040?ooo`03
0?oo_`1d000000000080000002`007Bo0?oo_`1d0000B9cO0?ooo`3ookl0M000001d_`3ooml0W4P0
0000000004P0B4Q809cOo`3ookl0M000001dM01d07@0_ooo0?oo_`1d0000B9cO0?oog`2LB0000000
0000B018B4P0W=oo0?oo_`1d000007Ad07@0M02oool0oono07@00018W=l0oooO09a80000000087CO
0?ooo`3ookl0M000001d_`80oooo00<0gia80018W03Oool00P3oool00`3oojL0B000000000030000
000D000/R@3Oool0oooo0?ooo`2oM00004RL0=oo_`1d0000M;oO09a80000M;l0oooO09a800000000
04Q804P0B02Lgol0oono07@00018W=l20?ooo`0Q0;md0000B9`0gooo0?oog`2LB000001d0;ooo`3o
ool0gia80018W03Oool0oooO09a800000000001d0;ooo`3oool0gia80018W03Oool0oooo0?noM000
000007Bo0?oog`2LB00000000018B01804P0W=oo0?oo_`1d0000B9cO0:P0oooo002<0?ooo`040?oo
_`1d000004RL0=ooo`X0oooo00H0oono07@00000000000000000M02oool20?ooo`0<0;md000007@0
_ooO09a800000000001809cOo`3ookl0M0000000000007@0_ooo0P3oool0402oM000001d0;oog`2L
B00000000000B02Lgol0oono07@000000000001d0;ooo`3oool0_g@00000=P180002000000070000
M02oool0oooo0?ooo`3OW4P004RL0=ooo`020?ooo`0A0?oo_`1d000082b90=ooo`2oM000000e03FL
g`3oool0oono07@00000000004RL0=oo_`1d0000800000000000M;l00P3oool01`2oM00004RL0=oo
g`2LB00000000000M02oool00`3oool01`3OW4P000000000000004P0W=oo0?ooo`2oM0000`000000
1@00B9`0gooo0;md000007@0_ooo0080oooo00D0okmd000000000000000004RLg`030?ooo`070;md
0000B9`0gooO09a800000000001d0;ooo`2X0?ooo`00S@3oool00`2oM000001809cOo`060?ooo`03
0=nLM02Lgol0oooo02@0oooo00<0_g@00000B02Lgol01P3oool00`3OW4P004RL0=ooo`020?ooo`07
0?oo_`1d0000B8R907A8B0000000M:RL07BLg`050?ooo`030?oo_`1d0000M;oo01D0oooo00<0_g@0
0018W03Oool0_P3oool008d0oooo00<0omnL04P00000M;l01@3oool00`3ookl0M01d0;ooo`0U0?oo
o`030;md000004P0W=oo00H0oooo00<0gia80018W03Oool00P3oool01@3ookl0M00004RLZ01d0000
B9cO0080oooo00<0giad09cOo`3oool00P3oool00`3ookl0M00007Boo`0E0?ooo`030?oo_`1dB9`0
gooo0;h0oooo002>0?ooo`040?oOW018000004RL0=ooo`80oooo00@0omnL04P0000007@0_ooo9@3o
ool00`2oM000001809cOo`060?ooo`030=nLB000B9`0gooo0080oooo00T0omnL04P0000[;00004RL
09adB02Lgol0oono07@0M02oool00`3oool00`3ookl0M00007Boo`0E0?ooo`030?oog`2LM9`0gooo
0;h0oooo002?0?ooo`030?oog`2LB00000000080000000@004RL0=ooo`2oW9`0gooo9@3oool00`2o
M000001809cOo`060?ooo`050=nLB000B9`0gooo0?oog`2LB0001`0000000`0007@0_ooo0?ooo`02
0?ooo`030?oo_`1d0000M;oo01X0oooo00<0gia80000B02Lgol0^@3oool00;d0oooo00<0_g@00000
B02Lgol01P3oool01P3OW4P004RL0=ooo`3oool0oooO09a80080000000D007Bo0?noM0000000001d
0;ooo`040?ooo`030?oo_`1d0000M;oo01X0oooo00<0gia80000B02Lgol0^@3oool00;d0oooo00<0
_g@00000B02Lgol01P3oool00`3OW4P004RL0=ooo`0>0?ooo`030?oo_`1d0000M;oo0=H0oooo002m
0?ooo`030;md000004P0W=oo00H0oooo00<0gia80018W03Oool03P3oool00`3ookl0M00007Boo`3F
0?ooo`00_@3oool00`2oM000001809cOo`040?ooo`050?oOW02L_ol0gia80018W03Oool03P3oool0
0`3ookl0M00007Boo`3F0?ooo`00_03oool0103ooml0W4P00000B02Lgol30?ooo`060?oog`2LB000
M;oo0=nLB00007@0_ooo3P3oool00`3ookl0M00007Boo`3F0?ooo`00^`3oool00`3o_g@000000000
00070000000307Boo`3OW4P000000080000000<0M;oo0?ooo`3oool0203oool01@3OW4P000000000
00000000M;oo0=H0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40
oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-5.66118, -14.3369, \
0.0417911, 0.180474}}]
}, Closed]],

Cell[TextData[{
  "Unfortunately, ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " is not always terribly smart about where it places those labels.  \
Although we can place labels manually, I prefer to use ",
  StyleBox["Frame\[Rule]True", "InlineInput",
    FontWeight->"Bold"],
  " to draw a frame around the figure and ",
  StyleBox["FrameLabel\[Rule]{", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["bottom, left, top, right", "InlineInput"],
  StyleBox["}", "InlineInput",
    FontWeight->"Bold"],
  " to place labels outside the edges of the figure.  Note that ",
  StyleBox["FrameLabel", "InlineInput",
    FontWeight->"Bold"],
  " expects lists with either 2 or 4 elements.  Thus, if we do not want to \
label the right side we can include a blank string with the 3 labels we do \
want, or we can give 2 labels to ",
  StyleBox["FrameLabel", "InlineInput",
    FontWeight->"Bold"],
  " and use ",
  StyleBox["PlotLabel", "InlineInput",
    FontWeight->"Bold"],
  " for the top."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Gamma[x], {x, \(-5\), 5}, \ Frame \[Rule] True, 
        FrameLabel \[Rule] {"\<x\>", \ "\<\[CapitalGamma][x]\>", "\<Gamma \
Function\>", "\< \>"}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.0952381 0.316178 0.0220537 [
[.11905 -0.0125 -9.15625 -15.6875 ]
[.11905 -0.0125 9.15625 0 ]
[.30952 -0.0125 -9.15625 -15.6875 ]
[.30952 -0.0125 9.15625 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.69048 -0.0125 -5 -15.6875 ]
[.69048 -0.0125 5 0 ]
[.88095 -0.0125 -5 -15.6875 ]
[.88095 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -5.0625 -31.375 ]
[.5 -0.0125 5.0625 -15.6875 ]
[-0.0125 .09564 -24.3125 -7.84375 ]
[-0.0125 .09564 0 7.84375 ]
[-0.0125 .20591 -18.3125 -7.84375 ]
[-0.0125 .20591 0 7.84375 ]
[-0.0125 .31618 -10 -7.84375 ]
[-0.0125 .31618 0 7.84375 ]
[-0.0125 .42645 -10 -7.84375 ]
[-0.0125 .42645 0 7.84375 ]
[-0.0125 .53672 -16 -7.84375 ]
[-0.0125 .53672 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -40 -12.5313 ]
[-0.01875 .30902 -24.3125 12.5313 ]
[ 0 .61803 .125 0 ]
[.5 .63053 -43.7188 0 ]
[.5 .63053 43.7188 15.6875 ]
[ 1 0 .125 0 ]
[1.0125 .30902 0 -3.5 ]
[1.0125 .30902 15.6875 3.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.11905 0 m
.11905 .00625 L
s
gsave
.11905 -0.0125 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.30952 0 m
.30952 .00625 L
s
gsave
.30952 -0.0125 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.69048 0 m
.69048 .00625 L
s
gsave
.69048 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.88095 0 m
.88095 .00625 L
s
gsave
.88095 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.2619 0 m
.2619 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.7381 0 m
.7381 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.07143 0 m
.07143 .00375 L
s
.02381 0 m
.02381 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.97619 0 m
.97619 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -66.0625 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
69.125 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .09564 m
.00625 .09564 L
s
gsave
-0.0125 .09564 -85.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
0 .20591 m
.00625 .20591 L
s
gsave
-0.0125 .20591 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .31618 m
.00625 .31618 L
s
gsave
-0.0125 .31618 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .42645 m
.00625 .42645 L
s
gsave
-0.0125 .42645 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .53672 m
.00625 .53672 L
s
gsave
-0.0125 .53672 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .1177 m
.00375 .1177 L
s
0 .13975 m
.00375 .13975 L
s
0 .1618 m
.00375 .1618 L
s
0 .18386 m
.00375 .18386 L
s
0 .22796 m
.00375 .22796 L
s
0 .25002 m
.00375 .25002 L
s
0 .27207 m
.00375 .27207 L
s
0 .29412 m
.00375 .29412 L
s
0 .33823 m
.00375 .33823 L
s
0 .36029 m
.00375 .36029 L
s
0 .38234 m
.00375 .38234 L
s
0 .40439 m
.00375 .40439 L
s
0 .4485 m
.00375 .4485 L
s
0 .47055 m
.00375 .47055 L
s
0 .49261 m
.00375 .49261 L
s
0 .51466 m
.00375 .51466 L
s
0 .07359 m
.00375 .07359 L
s
0 .05153 m
.00375 .05153 L
s
0 .02948 m
.00375 .02948 L
s
0 .00743 m
.00375 .00743 L
s
0 .55877 m
.00375 .55877 L
s
0 .58082 m
.00375 .58082 L
s
0 .60288 m
.00375 .60288 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -105.688 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
73.5313 11.8438 translate
90 rotate
-73.5313 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(G) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
70.063 15.813 moveto
(@) show
74.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
80.125 15.813 moveto
(D) show
84.063 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.11905 .61178 m
.11905 .61803 L
s
.30952 .61178 m
.30952 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.69048 .61178 m
.69048 .61803 L
s
.88095 .61178 m
.88095 .61803 L
s
.125 Mabswid
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.2619 .61428 m
.2619 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.7381 .61428 m
.7381 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.07143 .61428 m
.07143 .61803 L
s
.02381 .61428 m
.02381 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.97619 .61428 m
.97619 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
gsave
.5 .63053 -104.719 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Gamma) show
104.000 15.813 moveto
(Function) show
146.438 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.99375 .09564 m
1 .09564 L
s
.99375 .20591 m
1 .20591 L
s
.99375 .31618 m
1 .31618 L
s
.99375 .42645 m
1 .42645 L
s
.99375 .53672 m
1 .53672 L
s
.125 Mabswid
.99625 .1177 m
1 .1177 L
s
.99625 .13975 m
1 .13975 L
s
.99625 .1618 m
1 .1618 L
s
.99625 .18386 m
1 .18386 L
s
.99625 .22796 m
1 .22796 L
s
.99625 .25002 m
1 .25002 L
s
.99625 .27207 m
1 .27207 L
s
.99625 .29412 m
1 .29412 L
s
.99625 .33823 m
1 .33823 L
s
.99625 .36029 m
1 .36029 L
s
.99625 .38234 m
1 .38234 L
s
.99625 .40439 m
1 .40439 L
s
.99625 .4485 m
1 .4485 L
s
.99625 .47055 m
1 .47055 L
s
.99625 .49261 m
1 .49261 L
s
.99625 .51466 m
1 .51466 L
s
.99625 .07359 m
1 .07359 L
s
.99625 .05153 m
1 .05153 L
s
.99625 .02948 m
1 .02948 L
s
.99625 .00743 m
1 .00743 L
s
.99625 .55877 m
1 .55877 L
s
.99625 .58082 m
1 .58082 L
s
.99625 .60288 m
1 .60288 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
gsave
1.0125 .30902 -56.6563 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
64.5 11.8438 translate
90 rotate
-64.5 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
66.000 15.813 moveto
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .31618 m
1 .31618 L
s
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02498 0 m
.02499 .30097 L
.02605 .30806 L
.02729 .31081 L
.02846 .31207 L
.02954 .31277 L
.03053 .31322 L
.03163 .31358 L
.03279 .31386 L
.03395 .31407 L
.0352 .31425 L
.03746 .31449 L
.03884 .31459 L
.04016 .31468 L
.04262 .31479 L
.045 .31487 L
.0463 .31491 L
.04753 .31493 L
.04868 .31495 L
.04993 .31497 L
.05105 .31499 L
.0521 .315 L
.05335 .315 L
.05401 .31501 L
.05472 .31501 L
.05546 .31501 L
.05617 .31501 L
.05749 .31501 L
.05872 .31501 L
.05986 .31501 L
.06111 .315 L
.06244 .31499 L
.06504 .31496 L
.06742 .31493 L
.07003 .31488 L
.07277 .31482 L
.07741 .31469 L
.08247 .31447 L
.08537 .31431 L
.08801 .31413 L
.09085 .3139 L
.09392 .31357 L
.09678 .31317 L
.09941 .31269 L
.10192 .31208 L
.10319 .31169 L
.10458 .31119 L
.10589 .31061 L
.10711 .30996 L
.10837 .30912 L
Mistroke
.10951 .30816 L
.11064 .30696 L
.11169 .30549 L
.11281 .30338 L
.11402 .30005 L
.11524 .29447 L
.11594 .28934 L
.11659 .28193 L
.1179 .24147 L
.11863 .10742 L
Mfstroke
.11863 .10742 m
.11924 .61803 L
s
.1195 .61803 m
.1205 .37806 L
.12179 .3495 L
.12302 .3397 L
.12415 .33484 L
.1252 .33195 L
.12634 .32978 L
.1274 .32827 L
.12838 .32721 L
.13065 .32545 L
.13179 .32481 L
.13304 .32425 L
.1354 .32343 L
.13793 .32282 L
.14031 .32241 L
.14247 .32213 L
.14479 .32191 L
.1461 .32181 L
.14733 .32174 L
.14863 .32168 L
.14981 .32164 L
.15047 .32162 L
.15118 .32161 L
.15247 .32159 L
.15358 .32158 L
.15479 .32159 L
.15607 .3216 L
.15726 .32163 L
.15837 .32166 L
.15939 .32169 L
.16169 .32179 L
.1642 .32194 L
.16653 .32212 L
.1693 .32239 L
.17182 .32269 L
.17698 .32351 L
.17982 .32411 L
.1825 .32481 L
.18712 .3264 L
.18975 .32761 L
.19221 .32905 L
.1948 .33099 L
.19716 .33331 L
.19847 .33491 L
.19989 .33701 L
.20123 .3394 L
.20249 .34218 L
.20366 .34535 L
.20492 .3497 L
.20607 .35483 L
Mistroke
.20712 .36102 L
.20827 .37022 L
.20952 .38546 L
.21083 .41327 L
.21206 .46886 L
Mfstroke
.21206 .46886 m
.21308 .61803 L
s
.21321 .61803 m
.21324 0 L
s
.21552 0 m
.21634 .14124 L
.2175 .20237 L
.21873 .23237 L
.21978 .24731 L
.22094 .25829 L
.22215 .26631 L
.22329 .27179 L
.22453 .27638 L
.22586 .28015 L
.22725 .28324 L
.22857 .28557 L
.23104 .28886 L
.23335 .29108 L
.23548 .29262 L
.23782 .2939 L
.2403 .2949 L
.24277 .29562 L
.24416 .29593 L
.24543 .29615 L
.24661 .29631 L
.2479 .29645 L
.24911 .29653 L
.25023 .29658 L
.25092 .29659 L
.25167 .29659 L
.25237 .29658 L
.25301 .29656 L
.25424 .2965 L
.25553 .29639 L
.25692 .29624 L
.25821 .29606 L
.26116 .2955 L
.26377 .29484 L
.26619 .29406 L
.26879 .29303 L
.27165 .29162 L
.27646 .28846 L
.2791 .28618 L
.28157 .28356 L
.28396 .28046 L
.28624 .27684 L
.28831 .27281 L
.29051 .26745 L
.2929 .25995 L
.29511 .25064 L
.29647 .24331 L
.29775 .23481 L
.299 .22445 L
.30014 .21255 L
Mistroke
.3013 .19698 L
.30254 .17446 L
.3037 .14491 L
.30479 .10353 L
.30606 .02245 L
Mfstroke
.30606 .02245 m
.30623 0 L
s
.30951 0 m
.30959 .61803 L
s
.31325 .61803 m
.31377 .5745 L
.31452 .53772 L
.31521 .51221 L
.31648 .47896 L
.31763 .45783 L
.31895 .43994 L
.32021 .4272 L
.32245 .41078 L
.32358 .40457 L
.32482 .39886 L
.32701 .39091 L
.32907 .38525 L
.33149 .38014 L
.3337 .37656 L
.33625 .3734 L
.33868 .37113 L
.33993 .3702 L
.34127 .36935 L
.34246 .36872 L
.34372 .36817 L
.34485 .36777 L
.3461 .36742 L
.34723 .3672 L
.34828 .36705 L
.34892 .367 L
.34961 .36696 L
.35087 .36697 L
.35209 .36706 L
.35322 .36722 L
.35444 .36746 L
.35579 .36783 L
.35851 .36888 L
.3609 .37016 L
.36308 .37164 L
.36544 .37362 L
.36801 .37629 L
.37235 .38225 L
.37471 .38649 L
.37691 .39123 L
.37937 .39769 L
.38197 .40627 L
.38483 .41853 L
.38745 .43369 L
.3899 .45292 L
.39219 .47794 L
.3935 .49701 L
.3947 .51886 L
.39609 .55172 L
.3974 .59444 L
Mistroke
Mfstroke
.3974 .59444 m
.39792 .61803 L
s
.40444 .61803 m
.40452 0 L
s
.41171 0 m
.41218 .0211 L
.41288 .04545 L
.41353 .0644 L
.4148 .09408 L
.416 .116 L
.41716 .13311 L
.41843 .14837 L
.42066 .16903 L
.42325 .18655 L
.42567 .19862 L
.42797 .20755 L
.43005 .21403 L
.43233 .21976 L
.43473 .22461 L
.43738 .22883 L
.43983 .23186 L
.44246 .23434 L
.44397 .23544 L
.44535 .23626 L
.44655 .23684 L
.44785 .23734 L
.44856 .23755 L
.44922 .23771 L
.45052 .23792 L
.45164 .238 L
.45287 .23798 L
.45415 .23783 L
.45535 .23757 L
.45653 .23721 L
.45779 .2367 L
.46005 .23544 L
.46274 .23334 L
.46519 .2308 L
.46774 .22737 L
.47052 .22261 L
.47313 .21685 L
.47554 .21016 L
.47761 .20305 L
.47982 .19357 L
.48226 .18014 L
.48453 .16348 L
.48575 .15223 L
.48708 .13745 L
.48826 .12155 L
.48951 .10063 L
.49074 .07422 L
.49186 .04329 L
.49299 .00213 L
s
.49299 .00213 m
.49303 0 L
s
.50068 0 m
.50068 .61803 L
s
.50675 .61803 m
.50762 .58072 L
.50876 .54517 L
.51008 .51386 L
.51153 .48796 L
.51289 .46901 L
.51417 .45451 L
.5166 .43324 L
.51881 .41878 L
.52119 .40665 L
.52376 .39633 L
.52839 .38259 L
.53266 .37354 L
.53735 .3661 L
.54238 .36007 L
.54713 .35566 L
.55147 .35243 L
.55654 .34934 L
.56207 .34665 L
.56725 .34459 L
.57212 .34299 L
.58151 .34059 L
.58692 .33953 L
.59188 .33871 L
.60198 .33743 L
.60729 .33694 L
.61288 .33652 L
.61787 .33623 L
.62241 .33602 L
.62493 .33593 L
.62765 .33585 L
.63023 .3358 L
.63137 .33577 L
.6326 .33576 L
.63375 .33574 L
.63479 .33573 L
.63585 .33572 L
.63685 .33572 L
.63801 .33571 L
.63908 .33571 L
.64026 .33571 L
.64152 .33571 L
.64274 .33572 L
.64344 .33573 L
.64407 .33573 L
.64647 .33576 L
.64866 .3358 L
.65096 .33585 L
.65593 .33599 L
.66045 .33616 L
Mistroke
.67062 .33667 L
.68004 .33732 L
.68896 .33809 L
.70889 .34039 L
.72732 .34334 L
.746 .34731 L
.76661 .35313 L
.78598 .3604 L
.80684 .37082 L
.82691 .38426 L
.84801 .40339 L
.86777 .4277 L
.88615 .45795 L
.90663 .50371 L
.92522 .56051 L
Mfstroke
.92522 .56051 m
.93903 .61803 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00iP3oool01@2oM0000000001d_`3oool0gia80080000000<004RL0=ooo`3oool0`03oool0
0>H0oooo00P0oooO09a8M02oool0oooo0?oog`2LB00004RL0=oool80oooo003W0?ooo`060?oo_`1d
B9`0gooo0;md000007@0_ooo``3oool00>P0oooo00@0omnL07A8B0000000M;ooa03oool00>P0oooo
00<0oono07@00018W=l0a@3oool00>L0oooo00D0oooO09a80000B9`0W7A809cOo`340?ooo`00i`3o
ool01P2oM000001809cOo`3ooml0W4Qd0;oool<0oooo003U0?ooo`090?oog`2LB000000000000000
M;l0okmd0000000007@0_ooo0<80oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo001l0?ooo`040?oog`2LB00004RL0=oo
oc80oooo00<0_g@0000000000000100000000`18W=l0oooo0?ooo`0/0?ooo`050?oOW01800000000
0000M02oool0<`3oool00`2oM0000000000000040000000304RLg`3oool0oooo03<0oooo00@0oooO
09a80000B9`0goooC`3oool007`0oooo00@0oooO09a80000B9`0gooo<`3oool00`3OW4P000000000
0003000000030018W03Oool0oooo02/0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo03<0
oooo00<0gia80000000000000`0000000`00B9`0gooo0?ooo`0c0?ooo`040?oog`2LB00004RL0=oo
odl0oooo001l0?ooo`040?oog`2LB00004RL0=oooc@0oooo00<0gia804RLg`3oool00P3oool00`3O
W7@0W=oo0?ooo`0[0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool0<`3oool00`3O
W4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo03<0oooo00@0oooO09a80000B9`0goooC`3oool0
07P0oooo00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`0b0?ooo`030=nLB01d_ol0oooo
02h0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0d0?ooo`030=nLB01d_ol0oooo
03<0oooo00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`1<0?ooo`00K03oool00`3ooml0
W4P0000000070000000404RLg`3oool0omnL07BLg`80oooo00@0oooO09a80000B9`0gooo9P3oool0
0`3ooml0W4P0000000070000000304RLg`3oool0oooo00<0oooo00<0_gA809cOo`3oool0;@3oool0
0`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo03D0oooo00<0_gA809cOo`3oool0<P3oool0
203ogi`0M9cO0?ooo`3oool0oooO09a80000B9`0goooC`3oool007T0oooo00L0gia807Boo`3oool0
oooO09a80000B9`0gooo03H0oooo00<0oooO09a8M02oool0;@3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo03D0oooo00<0oooO09a8M02oool0<`3oool01`3OW4P0M;oo0?ooo`3ooml0
W4P00018W03Oool0C`3oool007T0oooo00L0oooO09a8M02oool0oooO09a80000B9`0gooo03L0oooo
00<0okmd001d_`3oool0;03oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo03H0oooo
00<0okmd001d_`3oool0<P3oool01`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool0C`3oool007X0
oooo00H0omnL04Qd_`3ooml0W4P00018W03Ooolg0?ooo`030?oo_`1d04P0W=oo02`0oooo00<0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`0f0?ooo`030?oo_`1d04P0W=oo03<0oooo00H0omnL
04Qd_`3ooml0W4P00018W03Ooom?0?ooo`00N`3oool01@3OW4P0M;oO09a80000B9`0gooo03L0oooo
00<0oono07@00018W=l0;03oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo03H0oooo
00<0oono07@00018W=l0=03oool01@3OW4P0M;oO09a80000B9`0gooo04l0oooo001k0?ooo`050?oo
g`2LB4P0B4P00018W03Oool0<P3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO02d0
oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`0a0?ooo`030?noW02Lgol0oooo0080
oooo00<0omnL04P00018W=l0=03oool01@3ooml0W4Q804Q80000B9`0gooo04l0oooo001l0?ooo`04
0?oOW018000004RL0=oooc80oooo00P0oooO09a8B02Lgol0oooo0?oo_`1d0000001d0;ooobd0oooo
00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0380oooo00P0oooO09a8B02Lgol0oooo0?oo_`1d
0000001d0;ooocD0oooo00@0omnL04P00000B9`0goooC`3oool007d0oooo00<0gia80018W03Oool0
<`3oool01P3ooml0W4P0000000000000001d0;ooobl0oooo00D0oono07@000000000001d0;ooo`0d
0?ooo`060?oog`2LB00000000000000007@0_ooo=`3oool00`3OW4P004RL0=ooo`1?0?ooo`00o`3o
oona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3o
oona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00DP3ooooo000003800000;@3oool00580oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`8000003@3oool00`000000oooo
0?ooo`0:0?ooo`8000003P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0000000`0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0000000>0?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool0:`3oool00580oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0;0?ooo`8000003@3oool00`000000oooo0?ooo`0:0?ooo`8000003P3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0000000`0oooo00<000000?ooo`3oool02@3oool00`000000oooo0000
000>0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0:`3oool00580oooo0P000006
0?ooo`030000003oool0oooo03H0oooo0P00000J0?ooo`800000703oool00`000000oooo0000000H
0?ooo`030000003oool0000009H0oooo0P00000]0?ooo`00DP3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo03H0oooo0P00000J0?ooo`800000703oool010000000oooo0?ooo`00000G
0?ooo`030000003oool0000009L0oooo00<000000?ooo`3oool0:`3oool00580oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`0f0?ooo`8000006P3oool2000001`0oooo00@000000?oo
o`3oool000005`3oool00`000000oooo0000002G0?ooo`030000003oool0oooo02/0oooo001B0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool0=P3oool2000001X0oooo0P00000L0?oo
o`040000003oool0oooo000001L0oooo00<000000?ooo`000000U`3oool00`000000oooo0?ooo`0[
0?ooo`00DP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo03H0oooo0P00000J0?oo
o`800000703oool010000000oooo0?ooo`00000G0?ooo`030000003oool0000009L0oooo00<00000
0?ooo`3oool0:`3oool00580oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0f0?oo
o`8000006P3oool2000001`0oooo00@000000?ooo`3oool000005`3oool00`000000oooo0000002G
0?ooo`030000003oool0oooo02/0oooo001B0?ooo`8000001P3oool00`000000oooo0?ooo`0f0?oo
o`8000006P3oool2000001`0oooo00@000000?ooo`3oool000005`3oool00`000000oooo0000002F
0?ooo`800000;@3oool00580oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0f0?oo
o`8000006P3oool2000001`0oooo00@000000?ooo`3oool000005`3oool00`000000oooo0000002G
0?ooo`030000003oool0oooo02/0oooo001B0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool0=P3oool2000001X0oooo0P00000L0?ooo`040000003oool0oooo000001L0oooo00<00000
0?ooo`000000U`3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo03H0oooo0P00000J0?ooo`800000703oool010000000oooo0?ooo`00000G
0?ooo`030000003oool0000009L0oooo00<000000?ooo`3oool0:`3oool00580oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`0f0?ooo`8000006P3oool2000001`0oooo00@000000?oo
o`3oool000005`3oool00`000000oooo0000002G0?ooo`030000003oool0oooo02/0oooo001B0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool0=P3oool2000001X0oooo0P00000L0?oo
o`040000003oool0oooo000001L0oooo00<000000?ooo`000000U`3oool00`000000oooo0?ooo`0[
0?ooo`00DP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo03H0oooo0P00000J0?oo
o`800000703oool010000000oooo0?ooo`00000G0?ooo`030000003oool0000009L0oooo00<00000
0?ooo`3oool0:`3oool00580oooo0P0000060?ooo`030000003oool0oooo03H0oooo0P00000J0?oo
o`800000703oool010000000oooo0?ooo`00000G0?ooo`030000003oool0000009H0oooo0P00000]
0?ooo`00DP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo03H0oooo0P00000J0?oo
o`800000703oool010000000oooo0?ooo`00000G0?ooo`030000003oool0000009L0oooo00<00000
0?ooo`3oool0:`3oool00580oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0f0?oo
o`8000006P3oool2000001`0oooo00@000000?ooo`3oool000005`3oool00`000000oooo0000002G
0?ooo`030000003oool0oooo02/0oooo001B0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool0=P3oool2000001X0oooo0P00000L0?ooo`040000003oool0oooo000001H0oooo00@00000
0?ooo`3oool00000U`3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo03H0oooo0P00000J0?ooo`800000703oool010000000oooo0?ooo`00
000F0?ooo`040000003oool0oooo000009L0oooo00<000000?ooo`3oool0:`3oool00580oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`0f0?ooo`8000006P3oool2000001`0oooo00@0
00000?ooo`3oool000005P3oool010000000oooo0?ooo`00002G0?ooo`030000003oool0oooo02/0
oooo001B0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0=P3oool2000001X0oooo
0P00000L0?ooo`040000003oool0oooo000001H0oooo00@000000?ooo`3oool00000U`3oool00`00
0000oooo0?ooo`0[0?ooo`00DP3oool2000000H0oooo00<000000?ooo`3oool0=P3oool2000001X0
oooo0P00000L0?ooo`040000003oool0oooo000001H0oooo00@000000?ooo`3oool00000UP3oool2
000002d0oooo001B0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0=P3oool20000
01X0oooo0P00000L0?ooo`040000003oool0oooo000001H0oooo00@000000?ooo`3oool00000U`3o
ool00`000000oooo0?ooo`0[0?ooo`00?03oool01P2oM000000000000000000004RL0=ooo`<0oooo
00D0omnL04P000000000001d0;ooo`080?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool0=P3oool2000001X0oooo0P00000L0?ooo`040000003oool0oooo000001H0oooo00@000000?oo
o`3oool00000U`3oool00`000000oooo0?ooo`0[0?ooo`00?@3oool00`3ogi`0B00007Boo`040?oo
o`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`070?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool0=P3oool2000001X0oooo0P00000L0?ooo`040000003oool0oooo000001H0oooo
00@000000?ooo`3oool00000U`3oool00`000000oooo0?ooo`0[0?ooo`00?@3oool00`3ogi`0B000
07Boo`040?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool01P3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo03H0oooo0P00000J0?ooo`800000703oool010000000oooo
0?ooo`00000F0?ooo`040000003oool0oooo000009L0oooo00<000000?ooo`3oool0:`3oool003d0
oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0f0?ooo`8000006P3oool2000001`0
oooo00@000000?ooo`3oool000005P3oool010000000oooo0?ooo`00002G0?ooo`030000003oool0
oooo02/0oooo000]0?ooo`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool00`3oool00`3o
gi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo03H0oooo0P00000J0?ooo`800000703oool01000
0000oooo0?ooo`00000F0?ooo`040000003oool0oooo000009L0oooo00<000000?ooo`3oool0:`3o
ool003d0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo00H0oooo0`0000050?ooo`030000003oool0oooo03H0oooo0P00000J0?ooo`800000703oool0
10000000oooo0?ooo`00000F0?ooo`040000003oool0oooo000009D0oooo0`00000]0?ooo`00?@3o
ool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo03H0oooo0P00000J0?ooo`800000703o
ool010000000oooo0?ooo`00000F0?ooo`040000003oool0oooo000009L0oooo00<000000?ooo`3o
ool0:`3oool003d0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0f0?ooo`800000
6P3oool2000001`0oooo00@000000?ooo`3oool000005P3oool010000000oooo0?ooo`00002G0?oo
o`030000003oool0oooo02/0oooo000m0?ooo`030?oOW0180000M;oo00<0oooo00<0oono07@00018
W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool06@3oool00`000000oooo0?ooo`0J0?ooo`8000006P3oool2000001`0oooo00D000000?ooo`3o
ool0oooo0000000E0?ooo`040000003oool0oooo000009L0oooo00<000000?ooo`3oool0:`3oool0
03d0oooo00<0omnL04P0001d_ol0103oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo
00H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
01X0oooo0P00000J0?ooo`800000703oool01@000000oooo0?ooo`3oool0000001D0oooo00@00000
0?ooo`3oool00000U`3oool00`000000oooo0?ooo`0[0?ooo`00>`3oool01@3ooml0W4P000000000
0000M;oo00@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00L0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00000J0?ooo`80
0000703oool01@000000oooo0?ooo`3oool0000001D0oooo00@000000?ooo`3oool00000U`3oool0
0`000000oooo0?ooo`0[0?ooo`00?@3oool00`3o_g@0000007Boo`050?ooo`050?oo_`1d00000000
0000M02oool0203oool2000000H0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J
0?ooo`8000006P3oool2000001`0oooo00D000000?ooo`3oool0oooo0000000E0?ooo`040000003o
ool0oooo000009H0oooo0P00000]0?ooo`00DP3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool06P3oool2000001X0oooo0P00000L0?ooo`050000003o
ool0oooo0?ooo`0000005@3oool010000000oooo0?ooo`00002G0?ooo`030000003oool0oooo02/0
oooo001B0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06@3oool00`000000oooo
0?ooo`0J0?ooo`8000006P3oool2000001`0oooo00D000000?ooo`3oool0oooo0000000E0?ooo`04
0000003oool0oooo000009L0oooo00<000000?ooo`3oool0:`3oool00580oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00000J0?ooo`80
0000703oool01@000000oooo0?ooo`3oool0000001D0oooo00@000000?ooo`3oool00000U`3oool0
0`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`3oool06P3oool2000001X0oooo0P00000L0?ooo`050000003oool0oooo
0?ooo`000000503oool01@000000oooo0?ooo`3oool0000009L0oooo00<000000?ooo`3oool0:`3o
ool00580oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo01X0oooo0P00000J0?ooo`800000703oool01@000000oooo0?ooo`3oool0000001@0oooo00D0
00000?ooo`3oool0oooo0000002G0?ooo`030000003oool0oooo02/0oooo001B0?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`8000006P3o
ool2000001`0oooo00D000000?ooo`3oool0oooo0000000D0?ooo`050000003oool0oooo0?ooo`00
0000U`3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool2000000H0oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`0J0?ooo`8000006P3oool2000001`0oooo00D000000?ooo`3oool0
oooo0000000D0?ooo`050000003oool0oooo0?ooo`000000UP3oool2000002d0oooo001B0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`80
00006P3oool2000001`0oooo00D000000?ooo`3oool0oooo0000000D0?ooo`050000003oool0oooo
0?ooo`000000U`3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool2000001X0oooo0P00000L0?oo
o`050000003oool0oooo0?ooo`000000503oool01@000000oooo0?ooo`3oool0000009L0oooo00<0
00000?ooo`3oool0:`3oool00580oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I
0?ooo`030000003oool0oooo01X0oooo0P00000J0?ooo`800000703oool01@000000oooo0?ooo`3o
ool0000001@0oooo00D000000?ooo`3oool0oooo0000002G0?ooo`030000003oool0oooo02/0oooo
001B0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`0J0?ooo`8000006P3oool2000001`0oooo00D000000?ooo`3oool0oooo0000000D0?ooo`050000
003oool0oooo0?ooo`000000U`3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool2000001X0oooo
0P00000L0?ooo`050000003oool0oooo0?ooo`000000503oool01@000000oooo0?ooo`3oool00000
09L0oooo00<000000?ooo`3oool0:`3oool00580oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00000I0?ooo`030000003oool0000001`0
oooo00D000000?ooo`3oool0oooo0000000C0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0U@3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool2000000H0oooo00<000000?ooo`3o
ool06@3oool00`000000oooo0?ooo`0J0?ooo`8000006@3oool00`000000oooo0000000L0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo09@0oooo0P00000]0?ooo`00DP3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool2000001T0oooo00<000000?ooo`000000
703oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo02/0oooo001B0?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`8000006@3o
ool00`000000oooo0000000L0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo09D0oooo00<000000?ooo`3oool0:`3o
ool00580oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo01X0oooo0P00000I0?ooo`030000003oool0000001`0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0U@3oool0
0`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`3oool06P3oool2000001T0oooo00<000000?ooo`000000703oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`2E0?ooo`030000003oool0oooo02/0oooo001B0?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`8000006@3oool00`000000
oooo0000000L0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo09D0oooo00<000000?ooo`3oool0:`3oool00580oooo
0P0000060?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool2000001T0oooo
00<000000?ooo`000000703oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`2D0?ooo`800000;@3oool00580oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00
000I0?ooo`030000003oool0000001`0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?oo
o`0[0?ooo`00@`3oool00`3ooml0W4P0000000020000000307Boo`3oool0oooo00L0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00000I
0?ooo`030000003oool0000001`0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?ooo`0[
0?ooo`00@`3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00L0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00000I0?oo
o`030000003oool0000001`0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?ooo`0[0?oo
o`00B03oool00`3ooml0W4Qd0;ooo`070?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool06@3oool00`000000oooo0?ooo`0J0?ooo`8000006@3oool00`000000oooo0000000L0?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo09D0oooo00<000000?ooo`3oool0:`3oool004T0oooo00<0_g@007Boo`3oool0
1P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0
6P3oool2000001T0oooo00<000000?ooo`000000703oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`2E0?ooo`030000
003oool0oooo02/0oooo000f0?ooo`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool01@3o
ool00`3ooml0W4P004RLg`070?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06@3o
ool00`000000oooo0?ooo`0J0?ooo`8000006@3oool00`000000oooo0000000L0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo09D0oooo00<000000?ooo`3oool0:`3oool004P0oooo00<0gia80000001d_ol01`3oool3
000000D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`800000603oool0
10000000oooo0?ooo`00000L0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo09<0oooo0`00000]0?ooo`00A03oool0
1`3Ogml0oooo0?oo_`1d000000000000M02oool01`3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool2000001P0oooo00@000000?ooo`3oool0
0000703oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo02/0oooo00140?ooo`060=nL
B000000000000000000004P0W=oo203oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`3oool06P3oool2000001P0oooo00@000000?ooo`3oool00000703oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`2E0?ooo`030000003oool0oooo02/0oooo00140?ooo`040?oo_`1d00000000
04RLg`X0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo01X0oooo0P00000H0?ooo`040000003oool0oooo000001`0oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0U@3o
ool00`000000oooo0?ooo`0[0?ooo`00A@3oool00`2oM000M;oo0?ooo`0:0?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
000001L0oooo00@000000?ooo`3oool00000703oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`2E0?ooo`030000003o
ool0oooo02/0oooo00150?ooo`060?oOW018000000000000000004P0W=oo1`3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0000000G0?ooo`040000003oool0oooo000001`0oooo00<000000?ooo`3oool01P3oool00`000000
oooo0?ooo`080?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0U@3oool00`000000
oooo0?ooo`0[0?ooo`00AP3oool01@2oM0000000000000000000B9cO00L0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`00
00005`3oool010000000oooo0?ooo`00000L0?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo09D0oooo00<000000?oo
o`3oool0:`3oool00580oooo0P0000060?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0000000G0?ooo`040000003oool0oooo000001`0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool0U03oool2000002d0oooo001B0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0000001L0oooo00@000000?ooo`3o
ool00000703oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00@0oooo0P0000090?oo
o`030000003oool0oooo09D0oooo00<000000?ooo`3oool0:`3oool00580oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`00
00005`3oool010000000oooo0?ooo`00000L0?ooo`030000003oool0oooo00T0oooo0P0000020?oo
o`8000002`3oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo02/0oooo001B0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0000001H0oooo00D000000?ooo`3oool0oooo0000000L0?ooo`030000003oool0oooo
00X0oooo1000000<0?ooo`030000003oool0oooo09D0oooo00<000000?ooo`3oool0:`3oool00580
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`0000005P3oool01@000000oooo0?ooo`3oool0000001`0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo02/0oooo001B0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0000001H0oooo00D000000?ooo`3oool0oooo0000000L0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool2000000H0
oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0000001H0
oooo00D000000?ooo`3oool0oooo0000000L0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool0U03oool2000002d0oooo001B0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0000001H0oooo00D000000?ooo`3o
ool0oooo0000000L0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0U@3oool00`00
0000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01T0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000E0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo09D0
oooo00<000000?ooo`3oool0:`3oool00580oooo00<000000?ooo`3oool01@3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`0000005@3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`2E0?ooo`030000003oool0oooo02/0oooo000D0?ooo`030?oo_`1dM;l0oooo00X0oooo00<0
gibL0;ooo`3oool0;P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01T0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0000000E0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo09D0oooo00<0
00000?ooo`3oool0:`3oool001@0oooo00<0oono07A8W03Oool02@3oool00`3ooml0W4Qd0;ooo`0_
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J
0?ooo`040000003oool0oooo000001@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?oo
o`0[0?ooo`00503oool00`3ookl0M0000000000:000000030000M02oool0oooo02h0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00@00000
0?ooo`3oool00000503oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01X0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo02/0oooo000D
0?ooo`030?oo_`1d0000000000X0000000<0001d0;ooo`3oool0;P3oool2000000H0oooo00<00000
0?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`040000003oool0oooo000001<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<0
00000?ooo`3oool0U03oool2000002d0oooo000D0?ooo`030?oo_`1dM;l0oooo00T0oooo00<0oooO
09a8M02oool0;`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01T0oooo00<00000
0?ooo`3oool06P3oool01@000000oooo0?ooo`3oool000000140oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0U@3o
ool00`000000oooo0?ooo`0[0?ooo`00503oool00`3ookl0M7Bo0?ooo`0:0?ooo`030=nLW02oool0
oooo02h0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo01X0oooo00D000000?ooo`3oool0oooo0000000A0?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo09D0oooo00<0
00000?ooo`3oool0:`3oool001@0oooo00<0oono07Ad_`3oool0>`3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool01@000000oooo0?ooo`3o
ool000000140oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?ooo`0[0?ooo`00503oool0
0`3ookl0M7Bo0?ooo`0k0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06@3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0303oool2
000000L0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo09D0oooo00<000000?ooo`3oool0:`3oool001@0oooo00<0oono07@0B02Lgol0>`3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0U@3oool00`00
0000oooo0?ooo`0[0?ooo`00503oool0103ookl0M00000000018W=lj0?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool0603oool2000001`0oooo00<000000?ooo`3oool00`3oool20000
00T0oooo0P00000:0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`2E0?ooo`030000003oool0oooo02/0oooo001B0?ooo`8000001P3oool00`000000
oooo0?ooo`0H0?ooo`800000703oool00`000000oooo0?ooo`050?ooo`<000000P3oool4000000`0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo09@0
oooo0P00000]0?ooo`00503oool00`3ogi`0B0000000000>000000030000M02oool0oooo02X0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0H0?ooo`800000703oool00`000000oooo
0?ooo`080?ooo`<000003`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool0U@3oool00`000000oooo0?ooo`0[0?ooo`00503oool00`3ogi`0M9cO0?oo
o`0>0?ooo`030;mdM02oool0oooo01d0oooo00D0omnL04P000000000001d0;ooo`080?ooo`030000
003oool0oooo00D0oooo00<000000?ooo`3oool0603oool2000001`0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
U@3oool00`000000oooo0?ooo`0[0?ooo`00503oool00`3ogi`0W;oo0?ooo`0>0?ooo`030=nLW02o
ool0oooo01`0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00L0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0G0?ooo`<00000703oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`2E0?ooo`03
0000003oool0oooo02/0oooo000D0?ooo`030?oOW02L_ol0oooo00h0oooo00<0gibL0;ooo`3oool0
703oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00H0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`0F0?ooo`80000000<0oooo0000003oool06`3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`2E0?ooo`030000003oool0oooo02/0oooo00130?ooo`030?oo_`1d0000B9cO00<0oooo00<0
gia80000M02oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo01<0oooo0`00
00020?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?oo
o`0[0?ooo`006P3oool00`2oM7@0_ooo0?ooo`040?ooo`030;mdM02oool0oooo01l0oooo00<0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`030000003oool0oooo00L0oooo5@000005
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?ooo`0[
0?ooo`006P3oool00`2oM000B9cO0?ooo`030?ooo`030=nLB00007@0_ooo0200oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`060?ooool00000<P00000]0?ooo`006P3oool02@2oM000
00000000M02oool0oooO09a80018W=l0gibL0;ooo`0P0?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol01P3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo00@0oooo2`00
000;0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo09D0oooo00<000000?ooo`3oool0:`3oool001X0oooo00H0_ibL
0=oOW018000000000000B02LgolS0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
1P3oool00`000000oooo0?ooo`0Q0?ooo`040000003oool0oooo0?ooo`<000002`3oool3000000P0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?ooo`0[0?ooo`006P3oool00`2oM7@0_mnL
04P00002000000040000B02Lgol0gibL0;ooob00oooo00<0oono07@00018W=l00`3oool00`3OW4P0
001d0;ooo`060?ooo`030000003oool0oooo0240oooo00@000000?ooo`3oool000004@3oool30000
00D0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?ooo`0[0?ooo`006P3oool00`2oM000
04RL0=ooo`020?ooo`040;md00000000001d0;ooob40oooo00<0gia80018W03Oool00P3oool00`2o
M00007Bo0?ooo`060?ooo`030000003oool0oooo0240oooo00@000000?ooo`0000000000503oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo09D0oooo00<000000?ooo`3oool0:`3oool0
01X0oooo00<0_gA809cOo`3oool00`3oool00`3o_g@0001d0;ooo`0Q0?ooo`070?oo_`1d0000M;oo
0?ooo`3o_g@0001d0;ooo`070?ooo`030000003oool0oooo0240oooo00<000000?ooo`0000005@3o
ool2000000<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool08`3oool9000006T0oooo00<000000?ooo`3oool0:`3o
ool001X0oooo00<0_ibL0=ooo`3oool0103oool00`3OW9`0_ooo0?ooo`0Q0?ooo`050?oo_`1d0000
00000000M02oool0203oool200000280oooo00<000000?ooo`0000005P3oool01@000000oooo0?oo
o`3oool0000001`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000
003oool0oooo01T0oooo2P0000090?ooo`h00000FP3oool2000002d0oooo001B0?ooo`030000003o
ool0oooo0240oooo00<000000?ooo`0000005P3oool200000080oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool05@3o
ool400000240oooo1P00001E0?ooo`030000003oool0oooo02/0oooo001B0?ooo`030000003oool0
oooo0240oooo00<000000?ooo`0000005`3oool010000000oooo0?ooo`00000L0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0B0?ooo`<00000:`3o
ool500000500oooo00<000000?ooo`3oool0:`3oool001@0oooo00<0omnL09boo`3oool03P3oool0
0`3OW9`0_ooo0?ooo`0Z0?ooo`030000003oool0oooo0240oooo0P00000H0?ooo`040000003oool0
oooo000001`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo0100oooo0P00000c0?ooo`D00000B`3oool00`000000oooo0?ooo`0[0?ooo`00503oool0
0`3ogi`0W;oo0?ooo`0>0?ooo`030=nLW02oool0oooo02X0oooo00<000000?ooo`3oool08@3oool2
000001T0oooo00<000000?ooo`000000703oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool03P3oool2000003X0oooo0`0000180?ooo`030000003oool0
oooo02/0oooo000D0?ooo`030?oOW01dW=l0oooo00h0oooo00<0_gAd0;ooo`3oool0:P3oool00`00
0000oooo0?ooo`0Q0?ooo`8000006@3oool00`000000oooo0000000L0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0;0?ooo`<00000?`3oool30000
04D0oooo00<000000?ooo`3oool0:`3oool001@0oooo00<0omnL04P0000000003P0000000`0007@0
_ooo0?ooo`0Z0?ooo`030000003oool0oooo0240oooo0P00000I0?ooo`030000003oool0000001`0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool0@`3oool300000480oooo00<000000?ooo`3oool0:`3oool00580oooo
0P00000R0?ooo`8000006@3oool00`000000oooo0000000L0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo04H0oooo
0P00000o0?ooo`800000;@3oool00580oooo00<000000?ooo`3oool08P3oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0000001`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0B@3oool2000003h0oooo00<0
00000?ooo`3oool0:`3oool00580oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0H
0?ooo`800000703oool00`000000oooo0?ooo`070?ooo`D000003P3oool00`000000oooo0?ooo`0J
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0B`3oool2000003`0oooo00<00000
0?ooo`3oool0:`3oool00580oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0H0?oo
o`800000703oool00`000000oooo0?ooo`060?ooo`800000103oool3000000/0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo04h0oooo0P00000j0?oo
o`030000003oool0oooo02/0oooo001B0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3o
ool0603oool2000001`0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo0540oooo0P00000h0?ooo`030000003oool0oooo02/0oooo001B0?ooo`030000
003oool0oooo0280oooo00<000000?ooo`3oool0603oool2000001`0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool0
=@3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo0?ooo`0R0?ooo`030000003o
ool0oooo01P0oooo0P00000L0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`080?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool01P3o
ool00`000000oooo0?ooo`1E0?ooo`800000=@3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool2
000002<0oooo00<000000?ooo`3oool0603oool2000001`0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0<@3oool2
000002d0oooo001B0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0603oool20000
01`0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
00H0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
05T0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000
oooo0?ooo`0R0?ooo`030000003oool0oooo01P0oooo0P00000L0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo0300
oooo00<000000?ooo`3oool0:`3oool00580oooo00<000000?ooo`3oool08P3oool00`000000oooo
0?ooo`0H0?ooo`800000703oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool0G03oool200000300oooo00<000000?ooo`3oool0:`3oool00580oooo00<0
00000?ooo`3oool08P3oool00`000000oooo0?ooo`0H0?ooo`800000703oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0GP3oool00`000000oooo0?oo
o`0]0?ooo`030000003oool0oooo02/0oooo001B0?ooo`030000003oool0oooo0280oooo00<00000
0?ooo`3oool0603oool2000001`0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0[
0?ooo`00DP3oool2000002<0oooo00<000000?ooo`3oool0603oool2000001`0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0600oooo00<000000?oo
o`3oool0:P3oool2000002d0oooo001B0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3o
ool0603oool2000001`0oooo00D000000?ooo`3oool0oooo0000000A0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0640oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0[0?ooo`00@`3oool00`3ooml0
W4P0000000020000000307Boo`3oool0oooo00L0oooo00<000000?ooo`3oool08P3oool00`000000
oooo0?ooo`0H0?ooo`800000703oool01@000000oooo0?ooo`3oool000000140oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0HP3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo02/0oooo00130?ooo`03
0?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool01`3oool00`000000oooo0?ooo`0R0?ooo`03
0000003oool0oooo01P0oooo0P00000L0?ooo`050000003oool0oooo0?ooo`0000004@3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`1T0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0:`3oool004P0
oooo00<0oooO09a8M02oool01`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01P0
oooo0P00000L0?ooo`050000003oool0oooo0?ooo`0000004P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`1T0?oo
o`030000003oool0oooo02P0oooo00<000000?ooo`3oool0:`3oool004T0oooo00<0_g@007Boo`3o
ool01P3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01P0oooo0P00000L0?ooo`05
0000003oool0oooo0?ooo`0000004P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo
02L0oooo00<000000?ooo`3oool0:`3oool004P0oooo00<0oooO09a80018W=l01`3oool00`000000
oooo0?ooo`0R0?ooo`030000003oool0oooo01P0oooo0P00000L0?ooo`050000003oool0oooo0?oo
o`0000004P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`1V0?ooo`030000003oool0oooo02H0oooo00<000000?oo
o`3oool0:`3oool004P0oooo00<0gia80000001d_ol01`3oool300000280oooo00<000000?ooo`3o
ool0603oool2000001`0oooo00D000000?ooo`3oool0oooo0000000B0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
06P0oooo00<000000?ooo`3oool08`3oool3000002d0oooo00140?ooo`070=oOg`3oool0oono07@0
00000000001d0;ooo`070?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0603oool2
000001`0oooo00D000000?ooo`3oool0oooo0000000C0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo06P0oooo00<0
00000?ooo`3oool09@3oool00`000000oooo0?ooo`0[0?ooo`00A03oool01P3OW4P0000000000000
0000001809cOo`P0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0H0?ooo`800000
703oool010000000oooo0?ooo`00000D0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo06T0oooo00<000000?ooo`3o
ool0903oool00`000000oooo0?ooo`0[0?ooo`00A03oool0103ookl0M00000000018W=l:0?ooo`03
0000003oool0oooo0280oooo00<000000?ooo`3oool0603oool2000001`0oooo00@000000?ooo`3o
ool00000503oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo02@0oooo00<000000?oo
o`3oool0:`3oool004D0oooo00<0_g@007Boo`3oool02P3oool00`000000oooo0?ooo`0R0?ooo`03
0000003oool0oooo01P0oooo0P00000L0?ooo`040000003oool0oooo000001@0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0JP3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo02/0oooo00150?ooo`06
0?oOW018000000000000000004P0W=oo1`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0
oooo01T0oooo00<000000?ooo`3oool06P3oool010000000oooo0?ooo`00000D0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo06X0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0[0?ooo`00AP3oool0
1@2oM0000000000000000000B9cO00L0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?oo
o`0I0?ooo`030000003oool0oooo01X0oooo00@000000?ooo`3oool000005@3oool01@000000oooo
0?ooo`3oool0000001`0oooo00D000000?ooo`3oool0oooo0000001^0?ooo`030000003oool0oooo
0280oooo00<000000?ooo`3oool0:`3oool00580oooo0P00000S0?ooo`030000003oool0oooo01T0
oooo00<000000?ooo`3oool06P3oool010000000oooo0?ooo`00000E0?ooo`050000003oool0oooo
0?ooo`000000703oool01@000000oooo0?ooo`3oool0000006h0oooo00<000000?ooo`3oool08@3o
ool2000002d0oooo001B0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool06@3oool0
0`000000oooo0?ooo`0J0?ooo`040000003oool0oooo000001D0oooo00D000000?ooo`3oool0oooo
0000000L0?ooo`050000003oool0oooo0?ooo`000000K`3oool00`000000oooo0?ooo`0Q0?ooo`03
0000003oool0oooo02/0oooo001B0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`0J0?ooo`040000003oool0oooo000001D0oooo00D000000?ooo`3o
ool0oooo0000000L0?ooo`050000003oool0oooo0?ooo`000000K`3oool00`000000oooo0?ooo`0Q
0?ooo`030000003oool0oooo02/0oooo001B0?ooo`030000003oool0oooo0280oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`040000003oool0oooo000001D0oooo00D00000
0?ooo`3oool0oooo0000000L0?ooo`050000003oool0oooo0?ooo`000000L03oool00`000000oooo
0?ooo`0P0?ooo`030000003oool0oooo02/0oooo001B0?ooo`030000003oool0oooo0280oooo00<0
00000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`040000003oool0oooo000001D0oooo
00D000000?ooo`3oool0oooo0000000L0?ooo`050000003oool0oooo0?ooo`000000L@3oool00`00
0000oooo0?ooo`0O0?ooo`030000003oool0oooo02/0oooo001B0?ooo`030000003oool0oooo0280
oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`040000003oool0oooo0000
01D0oooo00D000000?ooo`3oool0oooo0000000L0?ooo`050000003oool0oooo0?ooo`000000L@3o
ool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo02/0oooo001B0?ooo`8000008`3oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`0000005P3oool0
1@000000oooo0?ooo`3oool0000001`0oooo00D000000?ooo`3oool0oooo0000001b0?ooo`030000
003oool0oooo01d0oooo0P00000]0?ooo`00DP3oool00`000000oooo0?ooo`0R0?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000F0?ooo`050000003o
ool0oooo0?ooo`000000703oool01@000000oooo0?ooo`3oool000000780oooo00<000000?ooo`3o
ool07P3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo0?ooo`0R0?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000F0?ooo`050000
003oool0oooo0?ooo`000000703oool01@000000oooo0?ooo`3oool0000007<0oooo00<000000?oo
o`3oool07@3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo0?ooo`0R0?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000F0?ooo`05
0000003oool0oooo0?ooo`000000703oool01@000000oooo0?ooo`3oool0000007<0oooo00<00000
0?ooo`3oool07@3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo0?ooo`0R0?oo
o`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000F0?oo
o`050000003oool0oooo0?ooo`000000703oool01@000000oooo0?ooo`3oool0000007@0oooo00<0
00000?ooo`3oool0703oool00`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo0?ooo`0R
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000F
0?ooo`050000003oool0oooo0?ooo`000000703oool01@000000oooo0?ooo`3oool0000007@0oooo
00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo0?oo
o`0R0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000
000G0?ooo`040000003oool0oooo000001`0oooo00D000000?ooo`3oool0oooo0000001e0?ooo`03
0000003oool0oooo01/0oooo00<000000?ooo`3oool0:`3oool00580oooo0P00000S0?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000G0?ooo`040000
003oool0oooo000001`0oooo00D000000?ooo`3oool0oooo0000001e0?ooo`030000003oool0oooo
01X0oooo0P00000]0?ooo`00DP3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01T0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000G0?ooo`040000003oool0oooo0000
01`0oooo00D000000?ooo`3oool0oooo0000001f0?ooo`030000003oool0oooo01X0oooo00<00000
0?ooo`3oool0:`3oool00580oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0I0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`0000005`3oool010000000oooo0?ooo`00000L
0?ooo`050000003oool0oooo0?ooo`000000MP3oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo02/0oooo001B0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool06@3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0000001L0oooo00@000000?ooo`3oool00000703o
ool010000000oooo0?ooo`00001h0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0
:`3oool00580oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`0000005`3oool010000000oooo0?ooo`00000L0?ooo`040000
003oool0oooo000007P0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0[0?ooo`00
DP3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0000000G0?ooo`040000003oool0oooo000001`0oooo00@000000?ooo`3o
ool00000N@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo02/0oooo001B0?ooo`03
0000003oool0oooo0280oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0000001L0oooo00@000000?ooo`3oool00000703oool010000000oooo0?ooo`00001i
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0:`3oool00580oooo0P00000S0?oo
o`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000G0?oo
o`040000003oool0oooo000001`0oooo00@000000?ooo`3oool00000N@3oool00`000000oooo0?oo
o`0G0?ooo`800000;@3oool00580oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0I
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`0000005`3oool010000000oooo0?ooo`00
000L0?ooo`040000003oool0oooo000007T0oooo00<000000?ooo`3oool0603oool00`000000oooo
0?ooo`0[0?ooo`00?03oool01P2oM000000000000000000004RL0=ooo`<0oooo00D0omnL04P00000
0000001d0;ooo`080?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool06@3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0000001L0oooo00@000000?ooo`3oool00000703oool0
10000000oooo0?ooo`00001j0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0:`3o
ool003d0oooo00<0omnL04P0001d_ol0103oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0
1`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0000000G0?ooo`040000003oool0oooo000001`0oooo00@000000?ooo`3o
ool00000NP3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo02/0oooo000m0?ooo`03
0?oOW0180000M;oo00@0oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`060?ooo`03
0000003oool0oooo0280oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0000001L0oooo00@000000?ooo`3oool00000703oool010000000oooo0?ooo`00001j
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0:`3oool003d0oooo00<0omnL04P0
001d_ol00`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00<000000?oo
o`3oool08P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`0000005`3oool010000000oooo0?ooo`00000L0?ooo`040000003oool0oooo000007X0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`0[0?ooo`00?@3oool00`3ogi`0B00007Boo`03
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool300000280oooo00<00000
0?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0000001L0oooo00@00000
0?ooo`3oool00000703oool010000000oooo0?ooo`00001k0?ooo`030000003oool0oooo01@0oooo
0`00000]0?ooo`00?@3oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol01P3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01T0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0000000H0?ooo`030000003oool0000001`0oooo00@0
00000?ooo`3oool00000N`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo02/0oooo
000m0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`060?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0000001P0oooo00<000000?ooo`000000703oool010000000oooo0?oo
o`00001k0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0:`3oool003d0oooo00<0
omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00<0
00000?ooo`3oool08P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<0
00000?ooo`000000603oool00`000000oooo0000000L0?ooo`040000003oool0oooo000007`0oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0[0?ooo`00?@3oool00`3ogi`0B00007Bo
o`030?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool00`000000oooo0?oo
o`0R0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000
000H0?ooo`030000003oool0000001`0oooo00@000000?ooo`3oool00000O03oool00`000000oooo
0?ooo`0E0?ooo`030000003oool0oooo02/0oooo000m0?ooo`030?oOW0180000M;oo00@0oooo00<0
gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`060?ooo`030000003oool0oooo0280oooo00<0
00000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0000001P0oooo00<0
00000?ooo`000000703oool010000000oooo0?ooo`00001l0?ooo`030000003oool0oooo01D0oooo
00<000000?ooo`3oool0:`3oool003/0oooo00D0oooO09a800000000000007Boo`040?ooo`070?oo
_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`070?ooo`030000003oool0oooo0280oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0000001P0oooo00<000000?oo
o`000000703oool010000000oooo0?ooo`00001l0?ooo`030000003oool0oooo01D0oooo00<00000
0?ooo`3oool0:`3oool003d0oooo00<0okmd0000001d_ol01@3oool01@3ookl0M0000000000007@0
_ooo00P0oooo0P00000S0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0000000H0?ooo`030000003oool0000001`0oooo00@000000?ooo`3oool00000O@3o
ool00`000000oooo0?ooo`0C0?ooo`800000;@3oool00580oooo00<000000?ooo`3oool08P3oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`000000603oool0
0`000000oooo0000000L0?ooo`040000003oool0oooo000007d0oooo00<000000?ooo`3oool0503o
ool00`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0
oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000H0?ooo`030000003oool0
000001`0oooo00<000000?ooo`000000OP3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0
oooo02/0oooo001B0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool06@3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0000001P0oooo00<000000?ooo`000000703oool00`00
0000oooo0000001o0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0:`3oool00580
oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`000000603oool00`000000oooo0000000L0?ooo`030000003oool0000007l0
oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo
0?ooo`0R0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0000000H0?ooo`030000003oool0000001`0oooo00<000000?ooo`000000O`3oool00`000000oooo
0?ooo`0C0?ooo`030000003oool0oooo02/0oooo001B0?ooo`030000003oool0oooo0280oooo00<0
00000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0000001P0oooo00<0
00000?ooo`000000703oool00`000000oooo0000001o0?ooo`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool0:`3oool00580oooo0P00000S0?ooo`030000003oool0oooo01T0oooo00<00000
0?ooo`3oool06P3oool00`000000oooo0000000H0?ooo`030000003oool0000001`0oooo00<00000
0?ooo`000000P03oool00`000000oooo0?ooo`0A0?ooo`800000;@3oool00580oooo00<000000?oo
o`3oool08P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`000000603oool00`000000oooo0000000L0?ooo`030000003oool000000800oooo00<000000?oo
o`3oool04P3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool00`000000oooo0?ooo`0R0?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000H0?ooo`03
0000003oool0000001`0oooo00<000000?ooo`000000P03oool00`000000oooo0?ooo`0B0?ooo`03
0000003oool0oooo02/0oooo001B0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0000001P0oooo00<000000?ooo`000000
703oool00`000000oooo000000210?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
:`3oool00580oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`000000603oool00`000000oooo0000000L0?ooo`030000003o
ool000000840oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0[0?ooo`00DP3oool0
0`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool2
000001T0oooo00<000000?ooo`000000703oool00`000000oooo000000210?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool0:`3oool00580oooo00<000000?ooo`3oool08P3oool00`00
0000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00000I0?ooo`030000003oool00000
01`0oooo00<000000?ooo`000000PP3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
02/0oooo001B0?ooo`8000008`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0
oooo0P00000I0?ooo`030000003oool0000001`0oooo00<000000?ooo`000000PP3oool00`000000
oooo0?ooo`0?0?ooo`800000;@3oool00580oooo00<000000?ooo`3oool08P3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00000I0?ooo`030000003oool0000001`0oooo
00<000000?ooo`000000PP3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo02/0oooo
001B0?ooo`030000003oool0oooo0240oooo0P00000K0?ooo`030000003oool0oooo01X0oooo0P00
000I0?ooo`030000003oool0000001`0oooo00<000000?ooo`000000=`3oool00`000000oooo0?oo
o`0g0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`0[0?ooo`00DP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00/0oooo00<00000
0?ooo`3oool02`3oool2000000d0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo00/0oooo0P00000=0?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`0000003P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0000000`0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`050000003oool0oooo0?ooo`0000002P3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo02/0oooo001B0?ooool00000<P00000]
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00[`3oool00`3ooml0W4P000000003000000060000M02oool0oooo0?ooo`3ooml0W4P00P00
00001@00B9`0W7@000000000M;l0gia80080000000<0001809cOg`2LB0000P0000000`0007@0_ooO
09a8000200000007001d_`3OW4P000000000000004P0W=oO09a80002000000030000M02ooml0W4P0
0080000000<007Bo0?oog`2LB0000P0000001000B9`0W7@000000000M;l30?ooo`030?oog`2LB000
00000080000000<0001809cOo`3oool0103oool02@3ookl0M000000000000000B9bo09a80000B9`0
gooo0;md0002000000030000M02ooml0W4P00080000000@004RL0=ooo`3oool0_g@00P0000002@00
04P0W=oo0?ooo`3oool0omnL04P000000000B9cO0=nLB002000000040000M02oool0oooo0=nLB0<0
000000@0M;oo0?ooo`3oool0_g@00P0000000`0007@0_ooO09a80002000000030018W03Oool0oooo
08H0oooo002^0?ooo`040?noM0000000000007Boo`<0oooo0100gia80000001d_ol0oooo0?noM000
0000M;oo0?ooo`3ogi`0B000001d_`3o_kl0_ooo0=nLB00007@0_ooo0P3oool01`2oM00004RL0=oo
o`3oool0oooO09a80000M;l00P3oool00`3OW4P0001d0;ooo`020?ooo`0A0;md0000B9`0gooo0?oo
o`3ooml0W4P0001d_`3oool0okmd0000001d_ol0oooo0?oOW018000007Bo0?no_`2oool0103oool0
0`3ogi`0B00004RLg`070?ooo`050=nLB00007@0_ooo0?ooo`3OW4P00P0000002@0007@0_ooo0=nL
B000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`0H0;md0000000004RL0=ooo`3ooml0W4Qd0;oo
o`3oool0gia80000M02oool0gkno0=ooo`3OW4P0001d0;ooo`3oool0gia800000018W=l0oooo0?oo
g`2LB000B9cO0P3oool01`3OW4P004RL0=ooo`3oool0oooO09a80000M;l0R@3oool00:d0oooo00@0
okmd0000000007@0_ooo103oool02`3o_g@0000007Boo`3oool0okmd0000B02Lgol0oooo0?oog`2L
B00007Bo0080oooo00<0gia80000M02oool00P3oool01`2oM00004RL0=ooo`3oool0oooO09a80000
M;l00P3oool00`3OW4P0001d0;ooo`020?ooo`0?0;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`3o
ool0okmd0000B02Lgol0oooo0?oog`2LB00007Bo00H0oooo00<0oono07@00018W=l01`3oool01`2o
M00004RL0=ooo`3oool0oooO09a80000M;l00P3oool02`3OW4P004RL0=ooo`3oool0oooO09a80000
M;l0oooo0?oOW018000007Bo00<0oooo00H0omnL09boo`3oool0gia80018W03Oool20?ooo`060=nL
B00007@0_ooo0?oOW018000007Bo0`3oool02`3o_g@0001d0;ooo`3oool0gia80018W03Oool0oooo
0?oog`2LB00007Bo08T0oooo002/0?ooo`040?oog`2LB000000007Boo`D0oooo00<0okmd0000001d
_ol00P3oool01P3o_g@00000001d_`3ooml0W4P0001d_`80oooo00<0gia80000M02oool00P3oool0
1`2oM00004RL0=ooo`3oool0oooO09a80000M;l00P3oool00`3OW4P0001d0;ooo`020?ooo`070;md
0000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`060?noM000000007Bo0?oog`2LB00007Bo1P3o
ool00`3ookl0M00004RLg`070?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`0;
0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0okmd0000B02Lgol01P3oool00`3OW4P004RL
0=ooo`020?ooo`060=nLB00007@0_ooo0?noM0000000M;oo0`3oool02`3o_g@0000007Boo`3oool0
gia80018W03Oool0oooo0?oog`2LB00007Bo08T0oooo002/0?ooo`040?oOW018000004RL0=ooo`D0
oooo00<0okmd0000001d_ol0103oool0103o_g@0000000000000M;l20?ooo`030=nLB00007@0_ooo
0080oooo00L0_g@00018W03Oool0oooo0?oog`2LB00007Bo0080oooo00<0gia80000M02oool00P3o
ool01`2oM00004RL0=ooo`3oool0oooO09a80000M;l0103oool0103o_g@0000000000000M;l60?oo
o`030?oo_`1d0000B9cO0080oooo00<0omnL07BLg`3oool00P3oool01`2oM00004RL0=ooo`3oool0
oooO09a80000M;l00P3oool02`3OW4P004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM00007@0
_ooo00H0oooo00<0gia80018W03Oool00P3oool01P3OW4P0001d0;ooo`3o_g@0000007Boo`<0oooo
00/0okmd0000001d_ol0oooo0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`290?ooo`00[03oool0
0`3o_g@00000001d_`060?ooo`0;0?noM0000000M;oo0?ooo`3ookl0M00007Boo`3oool0oooO09a8
0000M;l00P3oool00`3OW4P0001809cOo`020?ooo`070;md000007@0_ooo0?ooo`3ooml0W4P0001d
_`020?ooo`030=nLB00004P0W=oo0080oooo00l0_g@00000M02oool0oooo0?oog`2LB00007Bo0?oo
o`3ookl0M00007Boo`3oool0oooO09a80000M;l01P3oool00`3ookl0M00004RLg`020?ooo`030?no
M018W=l0oooo0080oooo00L0_g@00018W03Oool0oooo0?oog`2LB00007Bo0080oooo00/0_g@00000
M02oool0oooo0?oog`2LB00007Bo0?ooo`3ogi`0B01809cOo`020?ooo`070?noM00004P0W=oo0?oo
o`3OW4P004RL0=ooo`020?ooo`060=nLB00007@0_ooo0?oo_`1d0000M;oo0`3oool02`2oM0000018
09cOo`3oool0_g@00000M02oool0oooo0?oog`2LB00007Bo08T0oooo002/0?ooo`030?noM0000000
07Bo00@0oooo00<0oono07@0000000000P000000;000M;l0oono07@00018W=l0oooo0?oo_`1d0000
07Bo0?oog`2LB00000000000B018B4P0W=oo0?oo_`1d000007Ad07@0M02oool0oono07@00018W=l0
oooO09a800000000001804Q8B02Lgol0oono07@00000M7@0M01d0;ooo`3ookl0M00004RLg`3oool0
oono07@00018W=l0oooo0?oo_`1d000007Bo1P3oool00`3ookl0M000000000030000000D04RLg`3o
ool0oooo0?ooo`2oM00004RL0=ooo`3oool0oooO09a80000M;l0oooO09a80000000004Q804P0B02L
gol0oono07@00018W=l20?ooo`0Q0;md0000B9`0gooo0?oog`2LB000001d0;ooo`3oool0gia80018
W03Oool0oooO09a800000000001d0;ooo`3oool0gia80018W03Oool0oooo0?noM000000007Bo0?oo
g`2LB00000000018B01804P0W=oo0?oo_`1d0000B9cO08T0oooo002/0?ooo`040?oo_`1d000004RL
0=ooo`X0oooo00H0oono07@00000000000000000M02oool20?ooo`0<0;md000007@0_ooO09a80000
0000001809cOo`3ookl0M0000000000007@0_ooo0P3oool0302oM000001d0;oog`2LB00000000000
B02Lgol0oono07@000000000001d0;ooo`80oooo00H0oono07@00000000000000000M02oool60?oo
o`030?oo_`1d0000B9cO0080oooo00`0okmd04RLg`3oool0oono07@00000000004RL0=ooo`3ogi`0
B00000000000M;l20?ooo`070;md0000B9`0gooO09a800000000001d0;ooo`030?ooo`070=nLB000
000000000000B02Lgol0oooo0;md0003000000050018W03Oool0_g@00000M02oool00P3oool01@3o
_g@00000000000000000B9cO00<0oooo00L0_g@00018W03Ooml0W4P00000000007@0_ooo08T0oooo
002]0?ooo`030;md000004P0W=oo00H0oooo00<0giad09cOo`3oool0<@3oool00`3ookl0M00004RL
g`020?ooo`030?oOW01dW=l0oooo01`0oooo00<0_g@00018W03Oool0W`3oool00:d0oooo00<0omnL
04P00000M;l01@3oool00`3ookl0M01d0;ooo`0b0?ooo`030?oo_`1d0000B9cO00@0oooo00<0giad
09cOo`3oool06P3oool00`3ookl0M4RL0=ooo`2O0?ooo`00[P3oool0103ogi`0B0000018W03Oool2
0?ooo`040?oOW0180000001d0;oooc80oooo00<0omnL04P00018W=l00`3oool00`3ookl0M01d0;oo
o`0K0?ooo`030?oog`2LM9`0gooo09l0oooo002_0?ooo`030?oog`2LB00000000080000000@004RL
0=ooo`2oW9`0gooo<03oool00`3ooml0W4P000000006000000030000M02oool0oooo01l0oooo00<0
gia80000B02Lgol0VP3oool00?l0oooo503oool00`3OW4P0001809cOo`2J0?ooo`00o`3ooona0?oo
o`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-8.11422, -21.9307, \
0.0518628, 0.223967}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Grid lines", "Subsubsection"],

Cell[TextData[{
  "It is often nice to add grid lines for ease in reading values.  The \
simplest method uses ",
  StyleBox["GridLines\[Rule]Automatic", "InlineInput",
    FontWeight->"Bold"],
  ", although it is possible to use more sophisticated options."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Gamma[x], {x, \(-5\), 5}, \ GridLines \[Rule] Automatic, 
        Frame \[Rule] True, \ 
        FrameLabel \[Rule] {"\<x\>", \ "\<\[CapitalGamma][x]\>"}, \ 
        PlotLabel \[Rule] "\<Gamma Function\>"];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.0952381 0.316178 0.0220537 [
[.11905 -0.0125 -9.15625 -15.6875 ]
[.11905 -0.0125 9.15625 0 ]
[.30952 -0.0125 -9.15625 -15.6875 ]
[.30952 -0.0125 9.15625 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.69048 -0.0125 -5 -15.6875 ]
[.69048 -0.0125 5 0 ]
[.88095 -0.0125 -5 -15.6875 ]
[.88095 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -5.0625 -31.375 ]
[.5 -0.0125 5.0625 -15.6875 ]
[-0.0125 .09564 -24.3125 -7.84375 ]
[-0.0125 .09564 0 7.84375 ]
[-0.0125 .20591 -18.3125 -7.84375 ]
[-0.0125 .20591 0 7.84375 ]
[-0.0125 .31618 -10 -7.84375 ]
[-0.0125 .31618 0 7.84375 ]
[-0.0125 .42645 -10 -7.84375 ]
[-0.0125 .42645 0 7.84375 ]
[-0.0125 .53672 -16 -7.84375 ]
[-0.0125 .53672 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -40 -12.5313 ]
[-0.01875 .30902 -24.3125 12.5313 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -43.7188 0 ]
[.5 .63053 43.7188 15.6875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.11905 0 m
.11905 .61803 L
s
.30952 0 m
.30952 .61803 L
s
.5 0 m
.5 .61803 L
s
.69048 0 m
.69048 .61803 L
s
.88095 0 m
.88095 .61803 L
s
0 .09564 m
1 .09564 L
s
0 .20591 m
1 .20591 L
s
0 .31618 m
1 .31618 L
s
0 .42645 m
1 .42645 L
s
0 .53672 m
1 .53672 L
s
0 g
.11905 0 m
.11905 .00625 L
s
gsave
.11905 -0.0125 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.30952 0 m
.30952 .00625 L
s
gsave
.30952 -0.0125 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.69048 0 m
.69048 .00625 L
s
gsave
.69048 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.88095 0 m
.88095 .00625 L
s
gsave
.88095 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.2619 0 m
.2619 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.7381 0 m
.7381 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.07143 0 m
.07143 .00375 L
s
.02381 0 m
.02381 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.97619 0 m
.97619 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -66.0625 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
69.125 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .09564 m
.00625 .09564 L
s
gsave
-0.0125 .09564 -85.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
0 .20591 m
.00625 .20591 L
s
gsave
-0.0125 .20591 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .31618 m
.00625 .31618 L
s
gsave
-0.0125 .31618 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .42645 m
.00625 .42645 L
s
gsave
-0.0125 .42645 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .53672 m
.00625 .53672 L
s
gsave
-0.0125 .53672 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .1177 m
.00375 .1177 L
s
0 .13975 m
.00375 .13975 L
s
0 .1618 m
.00375 .1618 L
s
0 .18386 m
.00375 .18386 L
s
0 .22796 m
.00375 .22796 L
s
0 .25002 m
.00375 .25002 L
s
0 .27207 m
.00375 .27207 L
s
0 .29412 m
.00375 .29412 L
s
0 .33823 m
.00375 .33823 L
s
0 .36029 m
.00375 .36029 L
s
0 .38234 m
.00375 .38234 L
s
0 .40439 m
.00375 .40439 L
s
0 .4485 m
.00375 .4485 L
s
0 .47055 m
.00375 .47055 L
s
0 .49261 m
.00375 .49261 L
s
0 .51466 m
.00375 .51466 L
s
0 .07359 m
.00375 .07359 L
s
0 .05153 m
.00375 .05153 L
s
0 .02948 m
.00375 .02948 L
s
0 .00743 m
.00375 .00743 L
s
0 .55877 m
.00375 .55877 L
s
0 .58082 m
.00375 .58082 L
s
0 .60288 m
.00375 .60288 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -105.688 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
73.5313 11.8438 translate
90 rotate
-73.5313 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(G) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
70.063 15.813 moveto
(@) show
74.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
80.125 15.813 moveto
(D) show
84.063 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.11905 .61178 m
.11905 .61803 L
s
.30952 .61178 m
.30952 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.69048 .61178 m
.69048 .61803 L
s
.88095 .61178 m
.88095 .61803 L
s
.125 Mabswid
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.2619 .61428 m
.2619 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.7381 .61428 m
.7381 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.07143 .61428 m
.07143 .61803 L
s
.02381 .61428 m
.02381 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.97619 .61428 m
.97619 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .09564 m
1 .09564 L
s
.99375 .20591 m
1 .20591 L
s
.99375 .31618 m
1 .31618 L
s
.99375 .42645 m
1 .42645 L
s
.99375 .53672 m
1 .53672 L
s
.125 Mabswid
.99625 .1177 m
1 .1177 L
s
.99625 .13975 m
1 .13975 L
s
.99625 .1618 m
1 .1618 L
s
.99625 .18386 m
1 .18386 L
s
.99625 .22796 m
1 .22796 L
s
.99625 .25002 m
1 .25002 L
s
.99625 .27207 m
1 .27207 L
s
.99625 .29412 m
1 .29412 L
s
.99625 .33823 m
1 .33823 L
s
.99625 .36029 m
1 .36029 L
s
.99625 .38234 m
1 .38234 L
s
.99625 .40439 m
1 .40439 L
s
.99625 .4485 m
1 .4485 L
s
.99625 .47055 m
1 .47055 L
s
.99625 .49261 m
1 .49261 L
s
.99625 .51466 m
1 .51466 L
s
.99625 .07359 m
1 .07359 L
s
.99625 .05153 m
1 .05153 L
s
.99625 .02948 m
1 .02948 L
s
.99625 .00743 m
1 .00743 L
s
.99625 .55877 m
1 .55877 L
s
.99625 .58082 m
1 .58082 L
s
.99625 .60288 m
1 .60288 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .31618 m
1 .31618 L
s
.5 0 m
.5 .61803 L
s
gsave
.5 .63053 -104.719 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Gamma) show
104.000 15.813 moveto
(Function) show
146.438 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02498 0 m
.02499 .30097 L
.02605 .30806 L
.02729 .31081 L
.02846 .31207 L
.02954 .31277 L
.03053 .31322 L
.03163 .31358 L
.03279 .31386 L
.03395 .31407 L
.0352 .31425 L
.03746 .31449 L
.03884 .31459 L
.04016 .31468 L
.04262 .31479 L
.045 .31487 L
.0463 .31491 L
.04753 .31493 L
.04868 .31495 L
.04993 .31497 L
.05105 .31499 L
.0521 .315 L
.05335 .315 L
.05401 .31501 L
.05472 .31501 L
.05546 .31501 L
.05617 .31501 L
.05749 .31501 L
.05872 .31501 L
.05986 .31501 L
.06111 .315 L
.06244 .31499 L
.06504 .31496 L
.06742 .31493 L
.07003 .31488 L
.07277 .31482 L
.07741 .31469 L
.08247 .31447 L
.08537 .31431 L
.08801 .31413 L
.09085 .3139 L
.09392 .31357 L
.09678 .31317 L
.09941 .31269 L
.10192 .31208 L
.10319 .31169 L
.10458 .31119 L
.10589 .31061 L
.10711 .30996 L
.10837 .30912 L
Mistroke
.10951 .30816 L
.11064 .30696 L
.11169 .30549 L
.11281 .30338 L
.11402 .30005 L
.11524 .29447 L
.11594 .28934 L
.11659 .28193 L
.1179 .24147 L
.11863 .10742 L
Mfstroke
.11863 .10742 m
.11924 .61803 L
s
.1195 .61803 m
.1205 .37806 L
.12179 .3495 L
.12302 .3397 L
.12415 .33484 L
.1252 .33195 L
.12634 .32978 L
.1274 .32827 L
.12838 .32721 L
.13065 .32545 L
.13179 .32481 L
.13304 .32425 L
.1354 .32343 L
.13793 .32282 L
.14031 .32241 L
.14247 .32213 L
.14479 .32191 L
.1461 .32181 L
.14733 .32174 L
.14863 .32168 L
.14981 .32164 L
.15047 .32162 L
.15118 .32161 L
.15247 .32159 L
.15358 .32158 L
.15479 .32159 L
.15607 .3216 L
.15726 .32163 L
.15837 .32166 L
.15939 .32169 L
.16169 .32179 L
.1642 .32194 L
.16653 .32212 L
.1693 .32239 L
.17182 .32269 L
.17698 .32351 L
.17982 .32411 L
.1825 .32481 L
.18712 .3264 L
.18975 .32761 L
.19221 .32905 L
.1948 .33099 L
.19716 .33331 L
.19847 .33491 L
.19989 .33701 L
.20123 .3394 L
.20249 .34218 L
.20366 .34535 L
.20492 .3497 L
.20607 .35483 L
Mistroke
.20712 .36102 L
.20827 .37022 L
.20952 .38546 L
.21083 .41327 L
.21206 .46886 L
Mfstroke
.21206 .46886 m
.21308 .61803 L
s
.21321 .61803 m
.21324 0 L
s
.21552 0 m
.21634 .14124 L
.2175 .20237 L
.21873 .23237 L
.21978 .24731 L
.22094 .25829 L
.22215 .26631 L
.22329 .27179 L
.22453 .27638 L
.22586 .28015 L
.22725 .28324 L
.22857 .28557 L
.23104 .28886 L
.23335 .29108 L
.23548 .29262 L
.23782 .2939 L
.2403 .2949 L
.24277 .29562 L
.24416 .29593 L
.24543 .29615 L
.24661 .29631 L
.2479 .29645 L
.24911 .29653 L
.25023 .29658 L
.25092 .29659 L
.25167 .29659 L
.25237 .29658 L
.25301 .29656 L
.25424 .2965 L
.25553 .29639 L
.25692 .29624 L
.25821 .29606 L
.26116 .2955 L
.26377 .29484 L
.26619 .29406 L
.26879 .29303 L
.27165 .29162 L
.27646 .28846 L
.2791 .28618 L
.28157 .28356 L
.28396 .28046 L
.28624 .27684 L
.28831 .27281 L
.29051 .26745 L
.2929 .25995 L
.29511 .25064 L
.29647 .24331 L
.29775 .23481 L
.299 .22445 L
.30014 .21255 L
Mistroke
.3013 .19698 L
.30254 .17446 L
.3037 .14491 L
.30479 .10353 L
.30606 .02245 L
Mfstroke
.30606 .02245 m
.30623 0 L
s
.30951 0 m
.30959 .61803 L
s
.31325 .61803 m
.31377 .5745 L
.31452 .53772 L
.31521 .51221 L
.31648 .47896 L
.31763 .45783 L
.31895 .43994 L
.32021 .4272 L
.32245 .41078 L
.32358 .40457 L
.32482 .39886 L
.32701 .39091 L
.32907 .38525 L
.33149 .38014 L
.3337 .37656 L
.33625 .3734 L
.33868 .37113 L
.33993 .3702 L
.34127 .36935 L
.34246 .36872 L
.34372 .36817 L
.34485 .36777 L
.3461 .36742 L
.34723 .3672 L
.34828 .36705 L
.34892 .367 L
.34961 .36696 L
.35087 .36697 L
.35209 .36706 L
.35322 .36722 L
.35444 .36746 L
.35579 .36783 L
.35851 .36888 L
.3609 .37016 L
.36308 .37164 L
.36544 .37362 L
.36801 .37629 L
.37235 .38225 L
.37471 .38649 L
.37691 .39123 L
.37937 .39769 L
.38197 .40627 L
.38483 .41853 L
.38745 .43369 L
.3899 .45292 L
.39219 .47794 L
.3935 .49701 L
.3947 .51886 L
.39609 .55172 L
.3974 .59444 L
Mistroke
Mfstroke
.3974 .59444 m
.39792 .61803 L
s
.40444 .61803 m
.40452 0 L
s
.41171 0 m
.41218 .0211 L
.41288 .04545 L
.41353 .0644 L
.4148 .09408 L
.416 .116 L
.41716 .13311 L
.41843 .14837 L
.42066 .16903 L
.42325 .18655 L
.42567 .19862 L
.42797 .20755 L
.43005 .21403 L
.43233 .21976 L
.43473 .22461 L
.43738 .22883 L
.43983 .23186 L
.44246 .23434 L
.44397 .23544 L
.44535 .23626 L
.44655 .23684 L
.44785 .23734 L
.44856 .23755 L
.44922 .23771 L
.45052 .23792 L
.45164 .238 L
.45287 .23798 L
.45415 .23783 L
.45535 .23757 L
.45653 .23721 L
.45779 .2367 L
.46005 .23544 L
.46274 .23334 L
.46519 .2308 L
.46774 .22737 L
.47052 .22261 L
.47313 .21685 L
.47554 .21016 L
.47761 .20305 L
.47982 .19357 L
.48226 .18014 L
.48453 .16348 L
.48575 .15223 L
.48708 .13745 L
.48826 .12155 L
.48951 .10063 L
.49074 .07422 L
.49186 .04329 L
.49299 .00213 L
s
.49299 .00213 m
.49303 0 L
s
.50068 0 m
.50068 .61803 L
s
.50675 .61803 m
.50762 .58072 L
.50876 .54517 L
.51008 .51386 L
.51153 .48796 L
.51289 .46901 L
.51417 .45451 L
.5166 .43324 L
.51881 .41878 L
.52119 .40665 L
.52376 .39633 L
.52839 .38259 L
.53266 .37354 L
.53735 .3661 L
.54238 .36007 L
.54713 .35566 L
.55147 .35243 L
.55654 .34934 L
.56207 .34665 L
.56725 .34459 L
.57212 .34299 L
.58151 .34059 L
.58692 .33953 L
.59188 .33871 L
.60198 .33743 L
.60729 .33694 L
.61288 .33652 L
.61787 .33623 L
.62241 .33602 L
.62493 .33593 L
.62765 .33585 L
.63023 .3358 L
.63137 .33577 L
.6326 .33576 L
.63375 .33574 L
.63479 .33573 L
.63585 .33572 L
.63685 .33572 L
.63801 .33571 L
.63908 .33571 L
.64026 .33571 L
.64152 .33571 L
.64274 .33572 L
.64344 .33573 L
.64407 .33573 L
.64647 .33576 L
.64866 .3358 L
.65096 .33585 L
.65593 .33599 L
.66045 .33616 L
Mistroke
.67062 .33667 L
.68004 .33732 L
.68896 .33809 L
.70889 .34039 L
.72732 .34334 L
.746 .34731 L
.76661 .35313 L
.78598 .3604 L
.80684 .37082 L
.82691 .38426 L
.84801 .40339 L
.86777 .4277 L
.88615 .45795 L
.90663 .50371 L
.92522 .56051 L
Mfstroke
.92522 .56051 m
.93903 .61803 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00l`3oool01@2oM0000000001d_`3oool0gia80080000000<004RL0=ooo`3oool0/`3oool0
0?<0oooo00P0oooO09a8M02oool0oooo0?oog`2LB00004RL0=oookD0oooo003d0?ooo`060?oo_`1d
B9`0gooo0;md000007@0_ooo]P3oool00?D0oooo00@0omnL07A8B0000000M;oo]`3oool00?D0oooo
00<0oono07@00018W=l0^03oool00?@0oooo00D0oooO09a80000B9`0W7A809cOo`2g0?ooo`00m03o
ool01P2oM000001809cOo`3ooml0W4Qd0;oookH0oooo003b0?ooo`090?oog`2LB000000000000000
M;l0okmd0000000007@0_ooo0;D0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo002:0?ooo`040?oog`2LB00004RL0=oo
oc80oooo00<0_g@0000000000000100000000`18W=l0oooo0?ooo`0[0?ooo`050?oOW01800000000
0000M02oool0<`3oool00`2oM0000000000000040000000304RLg`3oool0oooo03@0oooo00@0oooO
09a80000B9`0gooo@@3oool008X0oooo00@0oooO09a80000B9`0gooo<`3oool00`3OW4P000000000
0003000000030018W03Oool0oooo02X0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo03<0
oooo00<0gia80000000000000`0000000`00B9`0gooo0?ooo`0d0?ooo`040?oog`2LB00004RL0=oo
od40oooo002:0?ooo`040?oog`2LB00004RL0=oooc@0oooo00<0gia804RLg`3oool00P3oool00`3O
W7@0W=oo0?ooo`0Z0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool0<`3oool00`3O
W4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo03@0oooo00@0oooO09a80000B9`0gooo@@3oool0
08H0oooo00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`0b0?ooo`030=nLB01d_ol0oooo
02d0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0d0?ooo`030=nLB01d_ol0oooo
03@0oooo00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`0n0?ooo`00N@3oool00`3ooml0
W4P0000000070000000504RLg`3oool0oooo0?oOW01dW=l00P3oool0103ooml0W4P00018W03OoolU
0?ooo`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool0103oool00`2oM4P0W=oo0?ooo`0/
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0=@3oool00`2oM4P0W=oo0?ooo`0c
0?ooo`080?oOW01dW=l0oooo0?ooo`3ooml0W4P00018W03Ooom10?ooo`00Q`3oool01`3OW4P0M;oo
0?ooo`3ooml0W4P00018W03Oool0=P3oool00`3ooml0W4Qd0;ooo`0/0?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol0=@3oool00`3ooml0W4Qd0;ooo`0d0?ooo`070=nLB01d_ol0oooo
0?oog`2LB00004RL0=ooo`110?ooo`00Q`3oool01`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool0
=`3oool00`3o_g@007Bo0?ooo`0[0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
=P3oool00`3o_g@007Bo0?ooo`0c0?ooo`070?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`110?oo
o`00R03oool01P3ogi`0B7Bo0?oog`2LB00004RL0=ooocL0oooo00<0oono07@0B02Lgol0:`3oool0
0`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo03H0oooo00<0oono07@0B02Lgol0=03oool0
1P3ogi`0B7Bo0?oog`2LB00004RL0=oood40oooo00290?ooo`050=nLB01d_ml0W4P00018W03Oool0
=`3oool00`3ookl0M00004RLg`0[0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0
=P3oool00`3ookl0M00004RLg`0e0?ooo`050=nLB01d_ml0W4P00018W03Oool0@@3oool008T0oooo
00D0oooO09a8B018B00004RL0=ooo`0b0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P00018
W=l0;03oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo0340oooo00<0oknL09cOo`3o
ool00P3oool00`3ogi`0B00004RLg`0e0?ooo`050?oog`2LB4P0B4P00018W03Oool0@@3oool008X0
oooo00@0omnL04P00000B9`0gooo<P3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo
;03oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0<P3oool0203ooml0W4Q809cOo`3oool0
oono07@0000007@0_ooo=P3oool0103ogi`0B0000018W03Ooom10?ooo`00R`3oool00`3OW4P004RL
0=ooo`0c0?ooo`060?oog`2LB00000000000000007@0_ooo;P3oool01@3ookl0M0000000000007@0
_ooo03@0oooo00H0oooO09a80000000000000000M02ooolh0?ooo`030=nLB000B9`0gooo0440oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo001P0?ooool00000<P00000O0?ooo`00
H03oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`0;0?ooo`8000003@3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0000000/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0000000=0?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo00D0oooo00<000000?ooo`3oool07@3oool00600oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3o
ool2000000d0oooo00<000000?ooo`3oool0303oool00`000000oooo0000000;0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`0000003@3oool00`000000oooo0?ooo`0<0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo01d0oooo001P0?ooo`8000001@3oool00`000000oooo0?ooo`0J0?ooo`030000O`3oool0
oooo01X0oooo00<000000?ooo`3oool06@3oool2000001`0oooo00<000000?ooo`000000603oool0
0`000000oooo0000000i0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0803oool3
000001l0oooo001P0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06P3oool00`00
07l0oooo0?ooo`0J0?ooo`030000003oool0oooo01T0oooo0P00000L0?ooo`030000003oool00000
01P0oooo00<000000?ooo`000000>@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo
0280oooo00<000000?ooo`3oool07@3oool00600oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`0J0?ooo`030000O`3oool0oooo01X0oooo00<000000?ooo`3oool06@3oool2000001`0
oooo00<000000?ooo`000000603oool00`000000oooo0000000i0?ooo`030000O`3oool0oooo03H0
oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo01X0oooo00<0001o0?ooo`3oool06P3oool00`000000oooo
0?ooo`0I0?ooo`800000703oool00`000000oooo0000000H0?ooo`030000003oool0000003T0oooo
00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo
001P0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06P3oool00`0007l0oooo0?oo
o`0J0?ooo`030000003oool0oooo01T0oooo0P00000L0?ooo`030000003oool0000001P0oooo00<0
00000?ooo`000000>@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<0
00000?ooo`3oool07@3oool00600oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0J
0?ooo`030000O`3oool0oooo01X0oooo00<000000?ooo`3oool06@3oool2000001`0oooo00<00000
0?ooo`0000005`3oool010000000oooo0?ooo`00000i0?ooo`030000O`3oool0oooo03H0oooo00<0
001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool2000000D0oooo00<00000
0?ooo`3oool06P3oool00`0007l0oooo0?ooo`0J0?ooo`030000003oool0oooo01T0oooo0P00000L
0?ooo`030000003oool0000001L0oooo00@000000?ooo`3oool00000>@3oool00`0007l0oooo0?oo
o`0f0?ooo`030000O`3oool0oooo0200oooo0`00000O0?ooo`00H03oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo01X0oooo00<0001o0?ooo`3oool06P3oool00`000000oooo0?ooo`0I
0?ooo`800000703oool00`000000oooo0000000G0?ooo`040000003oool0oooo000003T0oooo00<0
001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo001P
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06P3oool00`0007l0oooo0?ooo`0J
0?ooo`030000003oool0oooo01T0oooo0P00000L0?ooo`030000003oool0000001L0oooo00@00000
0?ooo`3oool00000>@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<0
00000?ooo`3oool07@3oool00600oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0J
0?ooo`030000O`3oool0oooo01X0oooo00<000000?ooo`3oool06@3oool2000001`0oooo00<00000
0?ooo`0000005`3oool010000000oooo0?ooo`00000i0?ooo`030000O`3oool0oooo03H0oooo00<0
001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo01X0oooo00<0001o0?ooo`3oool06P3oool00`000000oooo0?ooo`0I
0?ooo`800000703oool00`000000oooo0000000G0?ooo`040000003oool0oooo000003T0oooo00<0
001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo001P
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06P3oool00`0007l0oooo0?ooo`0J
0?ooo`030000003oool0oooo01T0oooo0P00000L0?ooo`030000003oool0000001L0oooo00@00000
0?ooo`3oool00000>@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<0
00000?ooo`3oool07@3oool00600oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0J
0?ooo`030000O`3oool0oooo01X0oooo00<000000?ooo`3oool06@3oool2000001`0oooo00<00000
0?ooo`0000005`3oool010000000oooo0?ooo`00000i0?ooo`030000O`3oool0oooo03H0oooo00<0
001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool2000000D0oooo00<00000
0?ooo`3oool06P3oool00`0007l0oooo0?ooo`0J0?ooo`030000003oool0oooo01T0oooo0P00000L
0?ooo`030000003oool0000001L0oooo00@000000?ooo`3oool00000>@3oool00`0007l0oooo0?oo
o`0f0?ooo`030000O`3oool0oooo0200oooo0`00000O0?ooo`00H03oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo01X0oooo00<0001o0?ooo`3oool06P3oool00`000000oooo0?ooo`0I
0?ooo`800000703oool00`000000oooo0000000G0?ooo`040000003oool0oooo000003T0oooo00<0
001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo001P
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06P3oool00`0007l0oooo0?ooo`0J
0?ooo`030000003oool0oooo01T0oooo0P00000L0?ooo`030000003oool0000001L0oooo00@00000
0?ooo`3oool00000>@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<0
00000?ooo`3oool07@3oool00600oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0J
0?ooo`030000O`3oool0oooo01X0oooo00<000000?ooo`3oool06@3oool2000001`0oooo00<00000
0?ooo`0000005`3oool010000000oooo0?ooo`00000i0?ooo`030000O`3oool0oooo03H0oooo00<0
001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo01X0oooo00<0001o0?ooo`3oool06P3oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0000001`0oooo00<000000?ooo`0000005`3oool010000000oooo0?ooo`00
000i0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?oo
o`0M0?ooo`00H03oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01X0oooo00<0001o
0?ooo`3oool06P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0000001`0oooo00<00000
0?ooo`0000005`3oool010000000oooo0?ooo`00000i0?ooo`030000O`3oool0oooo03H0oooo00<0
001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo01X0oooo00<0001o0?ooo`3oool06P3oool2000001T0oooo00<00000
0?ooo`000000703oool00`000000oooo0000000G0?ooo`040000003oool0oooo000003T0oooo00<0
001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo001P
0?ooo`8000001@3oool00`000000oooo0?ooo`0J0?ooo`030000O`3oool0oooo01X0oooo0P00000I
0?ooo`030000003oool0000001`0oooo00<000000?ooo`0000005`3oool010000000oooo0?ooo`00
000i0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0803oool3000001l0oooo001P
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06P3oool00`0007l0oooo0?ooo`0J
0?ooo`8000006@3oool00`000000oooo0000000L0?ooo`030000003oool0000001L0oooo00@00000
0?ooo`3oool00000>@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<0
00000?ooo`3oool07@3oool004T0oooo00H0_g@000000000000000000018W03Oool30?ooo`050?oO
W018000000000000M02oool02@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01X0
oooo00<0001o0?ooo`3oool06P3oool2000001T0oooo00<000000?ooo`000000703oool010000000
oooo0?ooo`00000F0?ooo`040000003oool0oooo000003T0oooo00<0001o0?ooo`3oool0=P3oool0
0`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo001:0?ooo`030?oOW0180000M;oo
00@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`0J0?ooo`030000O`3oool0oooo01X0oooo0P00000I0?ooo`030000003o
ool0000001`0oooo00@000000?ooo`3oool000005P3oool010000000oooo0?ooo`00000i0?ooo`03
0000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00
BP3oool00`3ogi`0B00007Boo`040?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool0
1`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01X0oooo00<0001o0?ooo`3oool0
6P3oool2000001T0oooo00<000000?ooo`000000703oool010000000oooo0?ooo`00000F0?ooo`04
0000003oool0oooo000003T0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0R0?oo
o`030000003oool0oooo01d0oooo001:0?ooo`030?oOW0180000M;oo00<0oooo00<0oono07@00018
W=l00`3oool00`3OW4P0001d0;ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool06P3oool00`0007l0oooo0?ooo`0J0?ooo`8000006@3oool00`000000oooo0000000L0?ooo`04
0000003oool0oooo000001H0oooo00@000000?ooo`3oool00000>@3oool00`0007l0oooo0?ooo`0f
0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool003/0oooo00<0oooO09a8
000000001`0000000`18W=l0oooo0?ooo`020?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool06P3oool00`0007l0oooo0?ooo`0J0?ooo`8000006@3oool00`000000oooo0000000L0?oo
o`040000003oool0oooo000001H0oooo00@000000?ooo`3oool00000>@3oool00`0007l0oooo0?oo
o`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool004X0oooo00<0omnL
04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo0`000004
0000O`030000000007l0001o03L0001o0P00000I0000O`030000000007l0000001`0001o00@00000
0000O`0007l000005P0007l010000000001o0000O`00002E0000O`<000007`3oool004X0oooo00<0
omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`0J0?ooo`030000O`3oool0oooo01X0oooo0P00
000I0?ooo`030000003oool0000001`0oooo00@000000?ooo`3oool000005P3oool010000000oooo
0?ooo`00000i0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08P3oool00`000000
oooo0?ooo`0M0?ooo`00BP3oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01X0oooo
00<0001o0?ooo`3oool06P3oool2000001T0oooo00<000000?ooo`000000703oool010000000oooo
0?ooo`00000F0?ooo`040000003oool0oooo000003T0oooo00<0001o0?ooo`3oool0=P3oool00`00
07l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo001:0?ooo`030?oOW0180000M;oo00<0
oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`070?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`8000006@3oool00`000000
oooo0000000L0?ooo`040000003oool0oooo000001H0oooo00@000000?ooo`3oool00000>@3oool0
0`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool0
04X0oooo00<0omnL04P0001d_ol0103oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo
00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo
01X0oooo0P00000I0?ooo`030000003oool0000001`0oooo00@000000?ooo`3oool000005@3oool0
1@000000oooo0?ooo`3oool0000003T0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?oo
o`0R0?ooo`030000003oool0oooo01d0oooo00180?ooo`050?oog`2LB00000000000001d_ol0103o
ool01`3ookl0M00007Boo`3oool0okmd0000M02oool0203oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool2000001T0oooo00<000000?ooo`00
0000703oool010000000oooo0?ooo`00000E0?ooo`050000003oool0oooo0?ooo`000000>@3oool0
0`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool0
04X0oooo00<0okmd0000001d_ol01@3oool01@3ookl0M0000000000007@0_ooo00T0oooo0P000005
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool2000001T0oooo00<00000
0?ooo`000000703oool010000000oooo0?ooo`00000E0?ooo`050000003oool0oooo0?ooo`000000
>@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0200oooo0`00000O0?ooo`00H03o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3o
ool2000001T0oooo00<000000?ooo`000000703oool010000000oooo0?ooo`00000E0?ooo`050000
003oool0oooo0?ooo`000000>@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280
oooo00<000000?ooo`3oool07@3oool00600oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo01X0oooo0P00000I0?ooo`030000003oool0000001`0oooo
00@000000?ooo`3oool000005@3oool01@000000oooo0?ooo`3oool0000003T0oooo00<0001o0?oo
o`3oool0=P3oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo001P0?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`80
00006@3oool00`000000oooo0000000L0?ooo`050000003oool0oooo0?ooo`000000503oool01@00
0000oooo0?ooo`3oool0000003T0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0R
0?ooo`030000003oool0oooo01d0oooo001P0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`8000006@3oool00`000000oooo0000000L0?oo
o`050000003oool0oooo0?ooo`000000503oool01@000000oooo0?ooo`3oool0000003T0oooo00<0
001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo001P
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J
0?ooo`8000006@3oool00`000000oooo0000000L0?ooo`050000003oool0oooo0?ooo`000000503o
ool01@000000oooo0?ooo`3oool0000003T0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo
0?ooo`0R0?ooo`030000003oool0oooo01d0oooo001P0?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`8000006@3oool00`000000oooo0000
000L0?ooo`050000003oool0oooo0?ooo`000000503oool01@000000oooo0?ooo`3oool0000003T0
oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0
oooo001P0?ooo`8000001@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo
0P00000I0?ooo`030000003oool0000001`0oooo00D000000?ooo`3oool0oooo0000000D0?ooo`05
0000003oool0oooo0?ooo`000000>@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo
0200oooo0`00000O0?ooo`00H03oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool06P3oool2000001T0oooo00<000000?ooo`000000703oool01@000000
oooo0?ooo`3oool0000001@0oooo00D000000?ooo`3oool0oooo0000000i0?ooo`030000O`3oool0
oooo03H0oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool20000
01T0oooo00<000000?ooo`000000703oool01@000000oooo0?ooo`3oool0000001<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0g0?ooo`030000O`3oool0oooo03H0oooo00<0001o
0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool2000001T0oooo00<000000?oo
o`000000703oool01@000000oooo0?ooo`3oool0000001<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0g0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08P3oool0
0`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`3oool06P3oool2000001T0oooo00<000000?ooo`000000703oool01@00
0000oooo0?ooo`3oool0000001<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0g
0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M
0?ooo`00H03oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01T0oooo00<000000?oo
o`3oool06P3oool2000001T0oooo00<000000?ooo`000000703oool01@000000oooo0?ooo`3oool0
000001<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0g0?ooo`030000O`3oool0
oooo03H0oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool20000
01T0oooo00<000000?ooo`000000703oool01@000000oooo0?ooo`3oool0000001<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0g0?ooo`030000O`3oool0oooo03H0oooo00<0001o
0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool2000000H0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`8000006@3oool00`000000oooo0000000L0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo03L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0P0?oo
o`<000007`3oool00600oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`03
0000003oool0oooo01X0oooo0P00000I0?ooo`030000003oool0000001`0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3o
ool07@3oool00600oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo01X0oooo0P00000I0?ooo`030000003oool0000001`0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool0
7@3oool00600oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo01X0oooo0P00000I0?ooo`030000003oool0000001`0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0=`3o
ool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3o
ool00600oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo01X0oooo0P00000I0?ooo`030000003oool0000001`0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0=`3oool0
0`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool0
0600oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
01X0oooo0P00000I0?ooo`030000003oool0000001`0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0=`3oool00`00
07l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool00600
oooo0P0000060?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool2000001T0
oooo00<000000?ooo`000000703oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0g0?ooo`030000O`3oool0oooo03H0
oooo00<0001o0?ooo`3oool0803oool3000001l0oooo001P0?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`800000603oool010000000oooo
0?ooo`00000L0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo03L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0
oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo001@0?ooo`030?oog`2LB000000000800000
00<0M;oo0?ooo`3oool0203oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool06P3oool2000001P0oooo00@000000?ooo`3oool00000703oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0g0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08P3oool00`000000
oooo0?ooo`0M0?ooo`00D03oool00`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00P0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo
0P00000H0?ooo`040000003oool0oooo000001`0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0=`3oool00`0007l0
oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool005D0oooo
00<0oooO09a8M02oool0203oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool06P3oool2000001P0oooo00@000000?ooo`3oool00000703oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0g0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08P3oool00`000000
oooo0?ooo`0M0?ooo`00EP3oool00`2oM000M;oo0?ooo`070?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`800000603oool010000000oooo
0?ooo`00000L0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo03L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0
oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo00140?ooo`030?oog`2LB000000000L00000
00<0B9cO0?ooo`3oool0103oool00`3ooml0W4P004RLg`080?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`800000603oool010000000oooo
0?ooo`00000L0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo03L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0
oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo001E0?ooo`030=nLB0000000M;oo00P0oooo
0`0000050000O`030000000007l0001o01T0001o00<000000000O`0007l06P0007l2000001P0001o
00@000000000O`0007l00000700007l00`000000001o0000O`040000O`030000000007l0001o00/0
001o00<000000000O`0007l01@0007l00`000000001o0000O`2C0000O`<000007`3oool00540oooo
00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00P0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00000H0?ooo`040000003oool0oooo
000001`0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool0=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0
oooo0280oooo00<000000?ooo`3oool07@3oool00540oooo00H0gia800000000000000000000B02L
gol90?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`0J0?ooo`800000603oool010000000oooo0?ooo`00000L0?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo03L0oooo
00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo
001A0?ooo`040?oo_`1d0000000004RLg`/0oooo00<000000?ooo`3oool01@3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00000H0?ooo`040000003oool0oooo000001`0
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool0=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280
oooo00<000000?ooo`3oool07@3oool00580oooo00<0_g@007Boo`3oool02`3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool2000001P0oooo
00@000000?ooo`3oool00000703oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0g0?ooo`030000O`3oool0oooo03H0
oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00DP3oool01P3ogi`0B000
000000000000001809cOo`P0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?oo
o`030000003oool0oooo01X0oooo0P00000H0?ooo`040000003oool0oooo000001`0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool0=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<00000
0?ooo`3oool07@3oool005<0oooo00D0_g@0000000000000000004RLg`080?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`800000603oool0
10000000oooo0?ooo`00000L0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo03L0oooo00<0001o0?ooo`3oool0=P3o
ool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo001P0?ooo`8000001P3oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00000H0?ooo`040000003oool0
oooo000001`0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool0=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3o
ool0oooo0200oooo0`00000O0?ooo`00H03oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo01T0oooo00<000000?ooo`3oool06P3oool2000001L0oooo00D000000?ooo`3oool0oooo0000
000L0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0103oool2000000X0oooo00<0
00000?ooo`3oool0=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<0
00000?ooo`3oool07@3oool00600oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I
0?ooo`030000003oool0oooo01X0oooo0P00000G0?ooo`050000003oool0oooo0?ooo`000000703o
ool00`000000oooo0?ooo`080?ooo`<000000P3oool2000000/0oooo00<000000?ooo`3oool0=`3o
ool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3o
ool00600oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo01X0oooo0P00000G0?ooo`050000003oool0oooo0?ooo`000000703oool00`000000oooo0?oo
o`0:0?ooo`@00000303oool00`000000oooo0?ooo`0g0?ooo`030000O`3oool0oooo03H0oooo00<0
001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000F
0?ooo`050000003oool0oooo0?ooo`000000703oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo03L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0R0?ooo`030000003o
ool0oooo01d0oooo001P0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06@3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0000001H0oooo00D000000?ooo`3oool0oooo0000
000L0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0=`3oool00`0007l0oooo0?oo
o`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool00600oooo0P000006
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000F
0?ooo`050000003oool0oooo0?ooo`000000703oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo03L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0P0?ooo`<000007`3o
ool00600oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`0000005P3oool01@000000oooo0?ooo`3oool0000001`0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`0g0?ooo`030000O`3oool0oooo03H0oooo00<0
001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000E
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J
0?ooo`030000003oool0oooo03L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0R
0?ooo`030000003oool0oooo01d0oooo001P0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0000001D0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool0=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?oo
o`3oool07@3oool00280oooo00<0oono07Ad_`3oool02P3oool00`3OW9`0_ooo0?ooo`0^0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool0603oool2000001`0oooo00<000000?ooo`00
00005@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0g0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3o
ool08P3oool00`000000oooo0?ooo`0M0?ooo`008P3oool00`3ookl0M4RL0=ooo`090?ooo`030?oo
g`2LB7@0_ooo02l0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0H0?ooo`800000
703oool00`000000oooo0000000D0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo03L0oooo00<0001o0?ooo`3oool0
=P3oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo000R0?ooo`030?oo_`1d
0000000000X0000000<0001d0;ooo`3oool0;P3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo01P0oooo0P00000L0?ooo`030000003oool0000001@0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0=`3o
ool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3o
ool00280oooo00<0oono07@0000000002P0000000`0007@0_ooo0?ooo`0^0?ooo`8000001P3oool0
0`000000oooo0?ooo`0H0?ooo`800000703oool010000000oooo0?ooo`00000B0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo03L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0P0?ooo`<000007`3o
ool00280oooo00<0oono07Ad_`3oool02@3oool00`3ooml0W4Qd0;ooo`0_0?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool0603oool2000001`0oooo00@000000?ooo`3oool000004P3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0g0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08P3o
ool00`000000oooo0?ooo`0M0?ooo`008P3oool00`3ookl0M7Bo0?ooo`0:0?ooo`030=nLW02oool0
oooo02h0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0H0?ooo`800000703oool0
10000000oooo0?ooo`00000B0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo03L0oooo00<0001o0?ooo`3oool0=P3o
ool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo000R0?ooo`030?oo_`1dM;l0
oooo03/0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0H0?ooo`800000703oool0
1@000000oooo0?ooo`3oool000000100oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0=`3oool00`0007l0oooo0?oo
o`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool00280oooo00<0oono
07Ad_`3oool0>`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01P0oooo0P00000L
0?ooo`050000003oool0oooo0?ooo`0000003`3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0g0?ooo`030000O`3o
ool0oooo03H0oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`008P3oool0
0`3ookl0M01809cOo`0k0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0603oool2
000001`0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo03L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0
oooo01d0oooo000R0?ooo`040?oo_`1d0000000004RLgcX0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0H0?ooo`800000703oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00P0oooo0P00000:0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`0g0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08P3oool0
0`000000oooo0?ooo`0M0?ooo`00H03oool2000000H0oooo00<000000?ooo`3oool0603oool20000
01`0oooo00<000000?ooo`3oool0103oool400000080oooo1000000<0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0g0?ooo`030000O`3oool0oooo
03H0oooo00<0001o0?ooo`3oool0803oool3000001l0oooo000R0?ooo`030?oOW0180000000000h0
000000<0001d0;ooo`3oool0:P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01P0
oooo0P00000L0?ooo`030000003oool0oooo00L0oooo0`00000@0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0g0?ooo`030000O`3oool0oooo03H0
oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`008P3oool00`3ogi`0M9cO
0?ooo`0>0?ooo`030;mdM02oool0oooo01`0oooo00D0omnL04P000000000001d0;ooo`090?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool0603oool2000001`0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3o
ool0=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3o
ool07@3oool00280oooo00<0omnL09boo`3oool03P3oool00`3OW9`0_ooo0?ooo`0K0?ooo`070?oo
_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool05`3oool00`000000oooo0000000L0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo03L0oooo00<0001o0?oo
o`3oool0=P3oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo000R0?ooo`03
0?oOW02L_ol0oooo00h0oooo00<0gibL0;ooo`3oool06`3oool00`2oM00004RL0=ooo`020?ooo`03
0;md0000M;l0oooo00L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0F0?ooo`80
000000<0oooo0000003oool06`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0g0?ooo`030000O`3oool0oooo03H0
oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00D03oool00`3ookl0M000
04RLg`030?ooo`030=nLB00007@0_ooo00L0oooo00<000000?ooo`3oool01@3oool2000001@0oooo
0`0000030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0=`3oool00`0007l0oooo
0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool002P0oooo00<0
_gAd0;ooo`3oool0103oool00`2oM7@0_ooo0?ooo`0N0?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol01`3oool00`000000oooo0?ooo`060?oooaD000001P3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo03L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?oo
o`0R0?ooo`030000003oool0oooo01d0oooo000X0?ooo`030;md0018W=l0oooo00<0oooo00<0gia8
0000M02oool07`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0ooooo`00000b
000001l0oooo000X0?ooo`090;md00000000001d0;ooo`3ooml0W4P004RLg`3OW9`0_ooo01l0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo0240oooo
00<000000?ooo`3oool0103oool:000000`0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0=`3oool00`0007l0oooo
0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool002P0oooo00H0
_ibL0=oOW018000000000000B02LgolR0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol01`3oool00`000000oooo0?ooo`0Q0?ooo`040000003oool0oooo0?ooo`<000002P3oool30000
00T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool0=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo
0280oooo00<000000?ooo`3oool07@3oool002P0oooo00<0_gAd0;oOW01800000P000000100004P0
W=oo0=nLW02ooolO0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01`3oool00`00
0000oooo0?ooo`0Q0?ooo`040000003oool0oooo00000100oooo0`0000060?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo03L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0
oooo01d0oooo000X0?ooo`030;md0000B9`0gooo0080oooo00@0_g@00000000007@0_ooo803oool0
0`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<000000?ooo`3oool08@3oool0
0`000000oooo0000000D0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0=`3oool0
0`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool0
02P0oooo00<0_gA809cOo`3oool00`3oool00`3o_g@0001d0;ooo`0P0?ooo`070?oo_`1d0000M;oo
0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo0240oooo00<000000?ooo`0000005@3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo02<0oooo2@00000;0?ooo`030000O`3o
ool0oooo03H0oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00:03oool0
0`2oW9`0gooo0?ooo`040?ooo`030=nLW02oool0oooo0200oooo00D0oono07@000000000001d0;oo
o`090?ooo`8000008P3oool2000001L0oooo00D000000?ooo`3oool0oooo0000000L0?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0I0?ooo`X00000
2@3oool=000003L0oooo00<0001o0?ooo`3oool0803oool3000001l0oooo001P0?ooo`030000003o
ool0oooo0240oooo0P00000G0?ooo`050000003oool0oooo0?ooo`000000703oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool05@3oool4000001h0oooo
00<0001o0?ooo`0000001@00000a0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool0
7@3oool00600oooo00<000000?ooo`3oool08@3oool2000001P0oooo00@000000?ooo`3oool00000
703oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
4@3oool400000280oooo00<0001o0?ooo`3oool01@3oool6000002/0oooo00<0001o0?ooo`3oool0
8P3oool00`000000oooo0?ooo`0M0?ooo`008P3oool00`3ogi`0W;oo0?ooo`0>0?ooo`030=nLW02o
ool0oooo02X0oooo00<000000?ooo`3oool08@3oool2000001P0oooo00@000000?ooo`3oool00000
703oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`0T0?ooo`030000O`3oool0oooo00/0oooo1000000W0?ooo`030000
O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool00280oooo00<0omnL09boo`3oool03P3o
ool00`3OW9`0_ooo0?ooo`0Z0?ooo`030000003oool0oooo0240oooo0P00000H0?ooo`040000003o
ool0oooo000001`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000
003oool0oooo00h0oooo0P00000W0?ooo`030000O`3oool0oooo00l0oooo1000000S0?ooo`030000
O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool00280oooo00<0omnL07BLg`3oool03P3o
ool00`2oM7@0_ooo0?ooo`0Z0?ooo`030000003oool0oooo0240oooo0P00000H0?ooo`040000003o
ool0oooo000001`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000
003oool0oooo00/0oooo0`00000Y0?ooo`030000O`3oool0oooo01<0oooo0`00000P0?ooo`030000
O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool00280oooo00<0omnL04P0000000003P00
00000`0007@0_ooo0?ooo`0Z0?ooo`030000003oool0oooo0240oooo0P00000H0?ooo`040000003o
ool0oooo000001`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000
003oool0oooo00X0oooo00<000000?ooo`3oool0:P3oool00`0007l0oooo0?ooo`0F0?ooo`<00000
7@3oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01d0oooo001P0?ooo`8000008P3o
ool2000001T0oooo00<000000?ooo`000000703oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0Z0?ooo`030000O`3o
ool0oooo01T0oooo0P00000K0?ooo`030000O`3oool0oooo0200oooo0`00000O0?ooo`00H03oool0
0`000000oooo0?ooo`0Q0?ooo`8000006@3oool00`000000oooo0000000L0?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo02/0oooo00<0001o0?ooo`3oool06`3oool2000001T0oooo00<0001o0?ooo`3oool08P3oool0
0`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo0?ooo`0Q0?ooo`8000006@3oool00`00
0000oooo0000000L0?ooo`030000003oool0oooo00L0oooo1@00000>0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0/0?ooo`030000O`3oool0oooo
01d0oooo0P00000G0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool00600
oooo00<000000?ooo`3oool08@3oool2000001T0oooo00<000000?ooo`000000703oool00`000000
oooo0?ooo`060?ooo`800000103oool2000000`0oooo00<000000?ooo`3oool06P3oool00`000000
oooo0?ooo`070?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool07`3oool2000001D0
oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo
0?ooo`0Q0?ooo`8000006@3oool00`000000oooo0000000L0?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo0240oooo
0P00000C0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool00600oooo00<0
00000?ooo`3oool08@3oool2000001T0oooo00<000000?ooo`000000703oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0;P3oool00`0007l0oooo0?oo
o`0S0?ooo`030000003oool0oooo0100oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?oo
o`0M0?ooo`00H03oool00`000000oooo0?ooo`0Q0?ooo`8000006P3oool2000001`0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool06P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo02h0oooo00<0001o
0?ooo`3oool0903oool200000100oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M
0?ooo`00H03oool200000280oooo0P00000J0?ooo`800000703oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0J0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool0;`3oool00`0007l0oooo0?ooo`0V0?oo
o`030000003oool0oooo00d0oooo00<0001o0?ooo`3oool0803oool3000001l0oooo001P0?ooo`03
0000003oool0oooo0240oooo0P00000J0?ooo`800000703oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00`0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool0;`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000003oool0oooo00`0oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00
H03oool00`000000oooo0?ooo`0Q0?ooo`8000006P3oool2000001`0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0300oooo00<0001o0?ooo`3oool0
:03oool00`000000oooo0?ooo`0;0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool0
7@3oool00600oooo00<000000?ooo`3oool08@3oool2000001X0oooo0P00000L0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0`0?ooo`030000O`3o
ool0oooo02T0oooo00<000000?ooo`3oool02P3oool00`0007l0oooo0?ooo`0R0?ooo`030000003o
ool0oooo01d0oooo001P0?ooo`030000003oool0oooo0240oooo0P00000J0?ooo`800000703oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0<03oool0
0`0007l0oooo0?ooo`0Z0?ooo`030000003oool0oooo00T0oooo00<0001o0?ooo`3oool08P3oool0
0`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo0?ooo`0Q0?ooo`8000006P3oool20000
01`0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0300oooo00<0001o0?ooo`3oool0:`3oool00`000000oooo0?ooo`080?ooo`030000O`3oool0oooo
0280oooo00<000000?ooo`3oool07@3oool00600oooo0P00000R0?ooo`8000006P3oool2000001`0
oooo00D000000?ooo`3oool0oooo0000000A0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0340oooo00<0001o0?oo
o`3oool0;03oool00`000000oooo0?ooo`070?ooo`030000O`3oool0oooo0200oooo0`00000O0?oo
o`00H03oool00`000000oooo0?ooo`0Q0?ooo`8000006P3oool2000001`0oooo00D000000?ooo`3o
ool0oooo0000000A0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0;@3oool00`00
0000oooo0?ooo`060?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool00500
oooo00<0oooO09a8000000000P0000000`1d_ol0oooo0?ooo`080?ooo`030000003oool0oooo0240
oooo0P00000J0?ooo`800000703oool01@000000oooo0?ooo`3oool000000140oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool0<@3oool00`0007l0oooo0?ooo`0^0?ooo`030000003oool0oooo00D0oooo00<0001o0?oo
o`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00D03oool00`3ooml0W4Qd0;ooo`020?ooo`03
0;md001d_ol0oooo00P0oooo00<000000?ooo`3oool08@3oool2000001X0oooo0P00000L0?ooo`05
0000003oool0oooo0?ooo`0000004@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0b0?ooo`030000O`3oool0oooo
02l0oooo00<000000?ooo`3oool0103oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo
01d0oooo001E0?ooo`030?oog`2LB7@0_ooo00P0oooo00<000000?ooo`3oool08@3oool2000001X0
oooo0P00000L0?ooo`050000003oool0oooo0?ooo`0000004P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0b0?oo
o`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0103oool00`0007l0oooo0?ooo`0R0?oo
o`030000003oool0oooo01d0oooo001F0?ooo`030;md001d_ol0oooo00L0oooo00<000000?ooo`3o
ool08@3oool2000001X0oooo0P00000L0?ooo`040000003oool0oooo000001<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0<P3oool00`0007l0oooo0?ooo`0`0?ooo`030000003oool0oooo00<0oooo00<0001o0?oo
o`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00E@3oool00`3ooml0W4P004RLg`080?ooo`03
0000003oool0oooo0240oooo0P00000J0?ooo`800000703oool010000000oooo0?ooo`00000C0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo0380oooo00<0001o0?ooo`3oool0<@3oool00`000000oooo0?ooo`020?oo
o`030000O`3oool0oooo0280oooo00<000000?ooo`3oool07@3oool005D0oooo00<0gia80000001d
_ol0203oool300000240001o0P00000J0000O`800000700007l010000000001o0000O`00000C0000
O`030000000007l0001o00<0001o00<000000000O`0007l06P0007l00`000000001o0000O`020000
O`030000000007l0001o06L0001o00<000000000O`0007l0900007l3000001l0oooo001A0?ooo`07
0=oOg`3oool0oono07@000000000001d0;ooo`080?ooo`030000003oool0oooo0240oooo0P00000J
0?ooo`800000703oool010000000oooo0?ooo`00000D0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0380oooo00<0
001o0?ooo`3oool0<`3oool010000000oooo0?ooo`0007lT0?ooo`030000003oool0oooo01d0oooo
001A0?ooo`060=nLB000000000000000000004P0W=oo2@3oool00`000000oooo0?ooo`0Q0?ooo`80
00006P3oool2000001`0oooo00@000000?ooo`3oool00000503oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0b0?oo
o`030000O`3oool0oooo03<0oooo00@000000?ooo`3oool0001o903oool00`000000oooo0?ooo`0M
0?ooo`00D@3oool0103ookl0M00000000018W=l;0?ooo`030000003oool0oooo0240oooo0P00000J
0?ooo`800000703oool010000000oooo0?ooo`00000D0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0380oooo00<0
001o0?ooo`3oool0=03oool00`000000oooo0000O`0T0?ooo`030000003oool0oooo01d0oooo001B
0?ooo`030;md001d_ol0oooo00/0oooo00<000000?ooo`3oool08@3oool2000001X0oooo0P00000L
0?ooo`040000003oool0oooo000001@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0<P3oool00`0007l0oooo0?oo
o`0d0?ooo`030000003oool0001o02@0oooo00<000000?ooo`3oool07@3oool00580oooo00H0omnL
04P00000000000000000B02Lgol80?ooo`030000003oool0oooo0240oooo0P00000J0?ooo`800000
703oool010000000oooo0?ooo`00000D0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0380oooo00<0001o0?ooo`3o
ool0=@3oool00`000000001o0?ooo`0S0?ooo`030000003oool0oooo01d0oooo001C0?ooo`050;md
00000000000000000018W=l0203oool00`000000oooo0?ooo`0Q0?ooo`8000006P3oool2000001`0
oooo00@000000?ooo`3oool00000503oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
01X0oooo00D000000?ooo`3oool0oooo0000000e0?ooo`030000O`3oool0oooo03H0oooo00<00000
0?ooo`3oool08P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool200000280oooo0P00000J0?oo
o`800000703oool00`000000oooo0000000E0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool06P3oool01@000000oooo0?ooo`3oool0000003D0oooo00<0001o0?ooo`3oool0=P3oool0
0`000000oooo0?ooo`0P0?ooo`<000007`3oool00600oooo00<000000?ooo`3oool08@3oool20000
01X0oooo0P00000L0?ooo`030000003oool0000001D0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0J0?ooo`050000003oool0oooo0?ooo`000000=@3oool00`0007l0oooo0?ooo`0f
0?ooo`030000O`000000oooo0280oooo00<000000?ooo`3oool07@3oool00600oooo00<000000?oo
o`3oool08@3oool2000001X0oooo0P00000L0?ooo`030000003oool0000001D0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0J0?ooo`050000003oool0oooo0?ooo`000000=@3oool0
0`0007l0oooo0?ooo`0f0?ooo`030000O`000000oooo0280oooo00<000000?ooo`3oool07@3oool0
0600oooo00<000000?ooo`3oool08@3oool2000001X0oooo0P00000L0?ooo`030000003oool00000
01D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0J0?ooo`050000003oool0oooo
0?ooo`000000=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool000000280oooo00<00000
0?ooo`3oool07@3oool00600oooo00<000000?ooo`3oool08@3oool2000001X0oooo0P00000L0?oo
o`030000003oool0000001D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0J0?oo
o`050000003oool0oooo0?ooo`000000=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0
00000280oooo00<000000?ooo`3oool07@3oool00600oooo00<000000?ooo`3oool08@3oool20000
01X0oooo0P00000L0?ooo`030000003oool0000001D0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0J0?ooo`050000003oool0oooo0?ooo`000000=@3oool00`0007l0oooo0?ooo`0f
0?ooo`040000O`3oool0oooo00000240oooo00<000000?ooo`3oool07@3oool00600oooo0P00000R
0?ooo`8000006P3oool2000001`0oooo00<000000?ooo`0000005P3oool01@000000oooo0?ooo`3o
ool0000001`0oooo00D000000?ooo`3oool0oooo0000000e0?ooo`030000O`3oool0oooo03H0oooo
00@0001o0?ooo`3oool000007`3oool3000001l0oooo001P0?ooo`030000003oool0oooo0240oooo
0P00000J0?ooo`800000703oool00`000000oooo0000000F0?ooo`050000003oool0oooo0?ooo`00
0000703oool01@000000oooo0?ooo`3oool0000003D0oooo00<0001o0?ooo`3oool0=P3oool01@00
07l0oooo0?ooo`3oool000000200oooo00<000000?ooo`3oool07@3oool00600oooo00<000000?oo
o`3oool08@3oool2000001X0oooo0P00000L0?ooo`030000003oool0000001H0oooo00D000000?oo
o`3oool0oooo0000000L0?ooo`050000003oool0oooo0?ooo`000000=@3oool00`0007l0oooo0?oo
o`0f0?ooo`050000O`3oool0oooo0?ooo`000000803oool00`000000oooo0?ooo`0M0?ooo`00H03o
ool00`000000oooo0?ooo`0Q0?ooo`8000006P3oool2000001`0oooo00<000000?ooo`0000005P3o
ool01@000000oooo0?ooo`3oool0000001`0oooo00@000000?ooo`3oool00000=P3oool00`0007l0
oooo0?ooo`0f0?ooo`030000O`3oool0oooo0080oooo00<000000?ooo`3oool07@3oool00`000000
oooo0?ooo`0M0?ooo`00H03oool00`000000oooo0?ooo`0Q0?ooo`8000006P3oool2000001`0oooo
00<000000?ooo`0000005P3oool01@000000oooo0?ooo`3oool0000001`0oooo00@000000?ooo`3o
ool00000=P3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0080oooo00<000000?oo
o`3oool07@3oool00`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo0?ooo`0Q0?ooo`80
00006P3oool2000001`0oooo00<000000?ooo`0000005P3oool01@000000oooo0?ooo`3oool00000
01`0oooo00@000000?ooo`3oool00000=P3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0
oooo00<0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0M0?ooo`00H03oool00`00
0000oooo0?ooo`0Q0?ooo`8000006P3oool2000001`0oooo00<000000?ooo`0000005P3oool01@00
0000oooo0?ooo`3oool0000001`0oooo00@000000?ooo`3oool00000=P3oool00`0007l0oooo0?oo
o`0f0?ooo`030000O`3oool0oooo00<0oooo00<000000?ooo`3oool0703oool00`000000oooo0?oo
o`0M0?ooo`00H03oool200000280oooo0P00000J0?ooo`800000703oool00`000000oooo0000000F
0?ooo`050000003oool0oooo0?ooo`000000703oool010000000oooo0?ooo`00000f0?ooo`030000
O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0103oool00`000000oooo0?ooo`0I0?ooo`<00000
7`3oool00600oooo00<000000?ooo`3oool08@3oool2000001X0oooo0P00000L0?ooo`030000003o
ool0000001H0oooo00D000000?ooo`3oool0oooo0000000L0?ooo`040000003oool0oooo000003H0
oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`040?ooo`030000003oool0oooo01/0
oooo00<000000?ooo`3oool07@3oool00600oooo00<000000?ooo`3oool08@3oool00`000000oooo
0?ooo`0I0?ooo`800000703oool00`000000oooo0000000F0?ooo`050000003oool0oooo0?ooo`00
0000703oool010000000oooo0?ooo`00000f0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?oo
o`3oool01@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01d0oooo001P0?ooo`03
0000003oool0oooo0240oooo00<000000?ooo`3oool06@3oool2000001`0oooo00<000000?ooo`00
00005P3oool01@000000oooo0?ooo`3oool0000001`0oooo00@000000?ooo`3oool00000=P3oool0
0`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo00D0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo
01T0oooo0P00000L0?ooo`030000003oool0000001H0oooo00D000000?ooo`3oool0oooo0000000L
0?ooo`040000003oool0oooo000003H0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?oo
o`050?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool07@3oool00600oooo00<00000
0?ooo`3oool08@3oool00`000000oooo0?ooo`0I0?ooo`800000703oool00`000000oooo0000000G
0?ooo`040000003oool0oooo000001`0oooo00@000000?ooo`3oool00000=P3oool00`0007l0oooo
0?ooo`0f0?ooo`030000O`3oool0oooo00H0oooo00<000000?ooo`3oool06@3oool00`000000oooo
0?ooo`0M0?ooo`00H03oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo01T0oooo0P00
000L0?ooo`030000003oool0000001L0oooo00@000000?ooo`3oool00000703oool010000000oooo
0?ooo`00000f0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool01P3oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo01d0oooo001P0?ooo`8000008P3oool00`000000oooo
0?ooo`0I0?ooo`800000703oool00`000000oooo0000000G0?ooo`040000003oool0oooo000001`0
oooo00@000000?ooo`3oool00000=P3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo
00H0oooo00<000000?ooo`3oool05`3oool3000001l0oooo001P0?ooo`030000003oool0oooo0240
oooo00<000000?ooo`3oool06@3oool2000001`0oooo00<000000?ooo`0000005`3oool010000000
oooo0?ooo`00000L0?ooo`040000003oool0oooo000003H0oooo00<0001o0?ooo`3oool0=P3oool0
0`0007l0oooo0?ooo`070?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool07@3oool0
04T0oooo00H0_g@000000000000000000018W03Oool30?ooo`050?oOW018000000000000M02oool0
2@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo01T0oooo0P00000L0?ooo`030000
003oool0000001L0oooo00@000000?ooo`3oool00000703oool010000000oooo0?ooo`00000f0?oo
o`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool01`3oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01d0oooo001:0?ooo`030?oOW0180000M;oo00@0oooo00L0oono07@0001d
_ol0oooo0?noM00007@0_ooo00P0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0I
0?ooo`800000703oool2000001P0oooo00@000000?ooo`3oool00000703oool010000000oooo0?oo
o`00000f0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool01`3oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo01d0oooo001:0?ooo`030?oOW0180000M;oo00@0oooo00<0
_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030000003oool0oooo0240oooo00<0
00000?ooo`3oool06@3oool2000001`0oooo0P00000H0?ooo`040000003oool0oooo000001`0oooo
00@000000?ooo`3oool00000=P3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo00P0
oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0M0?ooo`00BP3oool00`3ogi`0B000
07Boo`030?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01`3oool00`000000oooo
0?ooo`0Q0?ooo`030000003oool0oooo01T0oooo0P00000L0?ooo`800000603oool010000000oooo
0?ooo`00000L0?ooo`030000003oool0000003L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0
oooo0?ooo`080?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool07@3oool004X0oooo
00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo
0`00000Q0000O`030000000007l0001o01T0001o0P00000L0000O`800000600007l010000000001o
0000O`00000L0000O`030000000007l0000007/0001o00<000000000O`0007l05@0007l3000001l0
oooo001:0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`070?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool06@3oool2000001`0oooo
0P00000H0?ooo`040000003oool0oooo000001`0oooo00<000000?ooo`000000=`3oool00`0007l0
oooo0?ooo`0f0?ooo`030000O`3oool0oooo00T0oooo00<000000?ooo`3oool05P3oool00`000000
oooo0?ooo`0M0?ooo`00BP3oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo01T0oooo
0P00000L0?ooo`800000603oool010000000oooo0?ooo`00000L0?ooo`030000003oool0000003L0
oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`090?ooo`030000003oool0oooo01H0
oooo00<000000?ooo`3oool07@3oool004X0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool08@3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo01X0oooo0P00000H0?ooo`040000003oool0oooo000001`0
oooo00<000000?ooo`000000=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo00T0
oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0M0?ooo`00BP3oool00`3ogi`0B000
07Boo`030?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01`3oool00`000000oooo
0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool2000001P0oooo
00@000000?ooo`3oool00000703oool00`000000oooo0000000g0?ooo`030000O`3oool0oooo03H0
oooo00<0001o0?ooo`3oool02P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo01d0
oooo001:0?ooo`030?oOW0180000M;oo00@0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo
0?ooo`070?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0J0?ooo`800000603oool010000000oooo0?ooo`00000L0?ooo`030000003oool0000003L0
oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0:0?ooo`030000003oool0oooo01D0
oooo00<000000?ooo`3oool07@3oool004P0oooo00D0oooO09a800000000000007Boo`040?ooo`07
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo0240oooo00<00000
0?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`800000603oool010000000oooo0?ooo`00
000L0?ooo`030000003oool0000003L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?oo
o`0:0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool07@3oool004X0oooo00<0okmd
0000001d_ol01@3oool01@3ookl0M0000000000007@0_ooo00T0oooo0P00000R0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool06P3oool2000001P0oooo00@000000?ooo`3oool00000
703oool00`000000oooo0000000g0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0
2`3oool00`000000oooo0?ooo`0B0?ooo`<000007`3oool00600oooo00<000000?ooo`3oool08@3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo0P00000H0?ooo`040000003o
ool0oooo000001`0oooo00<000000?ooo`000000=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000
O`3oool0oooo00/0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0M0?ooo`00H03o
ool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3o
ool2000001P0oooo00@000000?ooo`3oool00000703oool00`000000oooo0000000g0?ooo`030000
O`3oool0oooo03H0oooo00<0001o0?ooo`3oool02`3oool00`000000oooo0?ooo`0D0?ooo`030000
003oool0oooo01d0oooo001P0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0J0?ooo`800000603oool010000000oooo0?ooo`00000L0?ooo`030000
003oool0000003L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0;0?ooo`030000
003oool0oooo01@0oooo00<000000?ooo`3oool07@3oool00600oooo00<000000?ooo`3oool08@3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo0P00000H0?ooo`040000003o
ool0oooo000001`0oooo00<000000?ooo`000000=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000
O`3oool0oooo00`0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0M0?ooo`00H03o
ool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3o
ool2000001T0oooo00<000000?ooo`000000703oool00`000000oooo0000000g0?ooo`030000O`3o
ool0oooo03H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0C0?ooo`030000003o
ool0oooo01d0oooo001P0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`0J0?ooo`8000006@3oool00`000000oooo0000000L0?ooo`030000003oool0
000003L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0
oooo01<0oooo00<000000?ooo`3oool07@3oool00600oooo0P00000R0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool06P3oool2000001T0oooo00<000000?ooo`000000703oool00`00
0000oooo0000000g0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0303oool00`00
0000oooo0?ooo`0A0?ooo`<000007`3oool00600oooo00<000000?ooo`3oool08@3oool00`000000
oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo0P00000I0?ooo`030000003oool0000001`0
oooo00<000000?ooo`000000=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo00d0
oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo
0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool2000001T0oooo
00<000000?ooo`000000703oool00`000000oooo0000000g0?ooo`030000O`3oool0oooo03H0oooo
00<0001o0?ooo`3oool03@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01d0oooo
001P0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0J0?ooo`8000006@3oool00`000000oooo0000000L0?ooo`030000003oool0000003L0oooo00<0
001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0=0?ooo`030000003oool0oooo0180oooo00<0
00000?ooo`3oool07@3oool00600oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0J
0?ooo`030000003oool0oooo01X0oooo0P00000I0?ooo`030000003oool0000001`0oooo00<00000
0?ooo`000000=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo00d0oooo00<00000
0?ooo`3oool04P3oool00`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo0?ooo`0Q0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool2000001T0oooo00<000000?oo
o`000000703oool00`000000oooo0000000g0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?oo
o`3oool03P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo01d0oooo001P0?ooo`03
0000003oool0oooo0240oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`80
00006@3oool00`000000oooo0000000L0?ooo`030000003oool0000003L0oooo00<0001o0?ooo`3o
ool0=P3oool00`0007l0oooo0?ooo`0>0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3o
ool07@3oool00600oooo0P00000R0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
6P3oool2000001T0oooo00<000000?ooo`000000703oool00`000000oooo0000000g0?ooo`030000
O`3oool0oooo03H0oooo00<0001o0?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`<00000
7`3oool00600oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo01X0oooo0P00000I0?ooo`030000003oool0000001`0oooo00<000000?ooo`000000=`3o
ool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo00h0oooo00<000000?ooo`3oool04@3o
ool00`000000oooo0?ooo`0M0?ooo`00H03oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool06P3oool2000001T0oooo00<000000?ooo`000000703oool0
0`000000oooo0000000g0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01d0oooo001P0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
00`0oooo0P00000<0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`0000003@3oool00`00
0000oooo0?ooo`0<0?ooo`030000003oool0000000/0oooo00<000000?ooo`3oool0303oool00`00
0000oooo0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`0<0?ooo`040000003oool0oooo000000X0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0M0?ooo`00H03ooooo0000038000007`3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00;`0oooo00<0oooO09a80000
00000`0000001P0007@0_ooo0?ooo`3oool0oooO09a80080000000D004RL09ad0000000007Bo0=nL
B002000000030000B02Lgml0W4P00080000000<0001d0;oog`2LB0000P0000001`00M;l0gia80000
00000000001809cOg`2LB0000P0000000`0007@0_ooO09a8000200000003001d_`3ooml0W4P00080
000000@004RL09ad0000000007Bo103oool00`3ooml0W4P000000002000000030000B02Lgol0oooo
00@0oooo00T0oono07@000000000000004RL_`2LB00004RL0=ooo`2oM0000P0000000`0007@0_ooO
09a80002000000040018W03Oool0oooo0;md0080000000T0001809cOo`3oool0oooo0?oOW0180000
000004RLg`3OW4P00P000000100007@0_ooo0?ooo`3OW4P30000000407Boo`3oool0oooo0;md0080
000000<0001d0;oog`2LB0000P0000000`00B9`0gooo0?ooo`1h0?ooo`00^`3oool0103o_g@00000
0000001d_ol30?ooo`0@0=nLB0000000M;oo0?ooo`3o_g@0000007Boo`3oool0omnL04P00000M;l0
okno0;ooo`3OW4P0001d0;ooo`80oooo00L0_g@00018W03Oool0oooo0?oog`2LB00007Bo0080oooo
00<0gia80000M02oool00P3oool04@2oM00004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM000
0000M;oo0?ooo`3ogi`0B000001d_`3o_kl0_ooo00D0oooo00<0omnL04P00018W=l01`3oool01@3O
W4P0001d0;ooo`3oool0gia80080000000T0001d0;ooo`3OW4P004RL0=ooo`3oool0oooO09a80000
M;l00P3oool0602oM00000000018W03Oool0oooO09a8M02oool0oooo0=nLB00007@0_ooo0=no_`3O
ool0gia80000M02oool0oooo0=nLB0000000B9cO0?ooo`3ooml0W4P004RLg`80oooo00L0gia80018
W03Oool0oooo0?oog`2LB00007Bo07/0oooo002j0?ooo`040?noM0000000001d0;ooo`@0oooo00/0
okmd0000001d_ol0oooo0?noM00004P0W=oo0?ooo`3ooml0W4P0001d_`020?ooo`030=nLB00007@0
_ooo0080oooo00L0_g@00018W03Oool0oooo0?oog`2LB00007Bo0080oooo00<0gia80000M02oool0
0P3oool03`2oM00004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM00004P0W=oo0?ooo`3ooml0
W4P0001d_`070?ooo`030?oo_`1d0000B9cO00L0oooo00L0_g@00018W03Oool0oooo0?oog`2LB000
07Bo0080oooo00/0gia80018W03Oool0oooo0?oog`2LB00007Bo0?ooo`3ogi`0B000001d_`030?oo
o`060?oOW02L_ol0oooo0=nLB000B9`0gooo0P3oool01P3OW4P0001d0;ooo`3ogi`0B000001d_`<0
oooo00/0okmd0000M02oool0oooo0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`1k0?ooo`00^@3o
ool0103ooml0W4P00000001d_ol50?ooo`030?noM0000000M;oo0080oooo00H0okmd00000000M;l0
oooO09a80000M;l20?ooo`030=nLB00007@0_ooo0080oooo00L0_g@00018W03Oool0oooo0?oog`2L
B00007Bo0080oooo00<0gia80000M02oool00P3oool01`2oM00004RL0=ooo`3oool0oooO09a80000
M;l00P3oool01P3o_g@00000001d_`3ooml0W4P0001d_`L0oooo00<0oono07@00018W=l01`3oool0
1`2oM00004RL0=ooo`3oool0oooO09a80000M;l00P3oool02`3OW4P004RL0=ooo`3oool0oooO09a8
0000M;l0oooo0?noM00004P0W=oo00H0oooo00<0gia80018W03Oool00P3oool01P3OW4P0001d0;oo
o`3o_g@0000007Boo`<0oooo00/0okmd0000001d_ol0oooo0=nLB000B9`0gooo0?ooo`3ooml0W4P0
001d_`1k0?ooo`00^@3oool0103ogi`0B0000018W03Oool50?ooo`030?noM0000000M;oo00@0oooo
00@0okmd00000000000007Bo0P3oool00`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?oo
o`3ooml0W4P0001d_`020?ooo`030=nLB00007@0_ooo0080oooo00L0_g@00018W03Oool0oooo0?oo
g`2LB00007Bo00@0oooo00@0okmd00000000000007Bo1`3oool00`3ookl0M00004RLg`020?ooo`03
0?oOW01dW=l0oooo0080oooo00L0_g@00018W03Oool0oooo0?oog`2LB00007Bo0080oooo00/0gia8
0018W03Oool0oooo0?oog`2LB00007Bo0?ooo`3o_g@0001d0;ooo`060?ooo`030=nLB000B9`0gooo
0080oooo00H0gia80000M02oool0okmd0000001d_ol30?ooo`0;0?noM0000000M;oo0?ooo`3OW4P0
04RL0=ooo`3oool0oooO09a80000M;l0N`3oool00;T0oooo00<0okmd00000000M;l01P3oool02`3o
_g@0000007Boo`3oool0oono07@0001d_ol0oooo0?oog`2LB00007Bo0080oooo00<0gia80000B02L
gol00P3oool01`2oM000001d0;ooo`3oool0oooO09a80000M;l00P3oool00`3OW4P0001809cOo`02
0?ooo`0?0;md000007@0_ooo0?ooo`3ooml0W4P0001d_`3oool0oono07@0001d_ol0oooo0?oog`2L
B00007Bo00L0oooo00<0oono07@00018W=l00P3oool00`3o_g@0B9cO0?ooo`020?ooo`070;md0000
B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`0;0;md000007@0_ooo0?ooo`3ooml0W4P0001d_`3o
ool0omnL04P0B02Lgol00P3oool01`3o_g@0001809cOo`3oool0gia80018W03Oool00P3oool01P3O
W4P0001d0;ooo`3ookl0M00007Boo`<0oooo00/0_g@00000B02Lgol0oooo0;md000007@0_ooo0?oo
o`3ooml0W4P0001d_`1k0?ooo`00^@3oool00`3o_g@00000001d_`040?ooo`030?oo_`1d00000000
0080000002`007Bo0?oo_`1d0000B9cO0?ooo`3ookl0M000001d_`3ooml0W4P00000000004P0B4Q8
09cOo`3ookl0M000001dM01d07@0_ooo0?oo_`1d0000B9cO0?oog`2LB00000000000B018B4P0W=oo
0?oo_`1d000007Ad07@0M02oool0oono07@00018W=l0oooo0?oo_`1d0000B9cO0?ooo`3ookl0M000
001d_`L0oooo00<0oono07@0000000000`0000005018W=l0oooo0?ooo`3oool0_g@00018W03Oool0
oooo0?oog`2LB00007Bo0?oog`2LB00000000018B01804P0W=oo0?oo_`1d0000B9cO0P3oool08@2o
M00004RL0=ooo`3ooml0W4P00000M02oool0oooo0=nLB000B9`0gooo0?oog`2LB00000000000M02o
ool0oooo0=nLB000B9`0gooo0?ooo`3o_g@00000001d_`3ooml0W4P000000000B4P0B01809cOo`3o
okl0M00004RLg`1k0?ooo`00^@3oool0103ookl0M0000018W03Oool:0?ooo`060?oo_`1d00000000
0000000007@0_ooo0P3oool0302oM000001d0;oog`2LB00000000000B02Lgol0oono07@000000000
001d0;ooo`80oooo00`0_g@00000M02ooml0W4P00000000004P0W=oo0?oo_`1d000000000000M02o
ool20?ooo`060?oo_`1d000000000000000007@0_ooo1`3oool00`3ookl0M00004RLg`020?ooo`0<
0?noM018W=l0oooo0?oo_`1d000000000018W03Oool0omnL04P00000000007Bo0P3oool01`2oM000
04RL0=oog`2LB00000000000M02oool00`3oool01`3OW4P000000000000004P0W=oo0?ooo`2oM000
0`0000001@00B9`0gooo0;md000007@0_ooo0080oooo00D0okmd000000000000000004RLg`030?oo
o`070;md0000B9`0gooO09a800000000001d0;ooo`1k0?ooo`00^P3oool00`2oM000001809cOo`06
0?ooo`030=nLM02Lgol0oooo0380oooo00<0oono07@00018W=l00P3oool00`3ogi`0M9cO0?ooo`0L
0?ooo`030;md0000B9`0gooo0940oooo002j0?ooo`030?oOW018000007Bo00D0oooo00<0oono07@0
M02oool0<`3oool00`3ookl0M00004RLg`040?ooo`030=nLM02Lgol0oooo01X0oooo00<0oono07A8
W03Oool0T@3oool00;/0oooo00@0omnL04P00000B9`0gooo0P3oool0103ogi`0B0000000M02ooolc
0?ooo`030?oOW0180000B9cO00<0oooo00<0oono07@0M02oool06`3oool00`3ooml0W7BL0=ooo`2A
0?ooo`00_03oool00`3ooml0W4P000000002000000040018W03Oool0_ibL0=oooc40oooo00<0oooO
09a8000000001P0000000`0007@0_ooo0?ooo`0O0?ooo`030=nLB00004P0W=oo08`0oooo003o0?oo
ob80oooo00<0gia80000B02Lgol0S03oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-8.5834, -21.9307, \
0.0518628, 0.223967}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Plot range", "Subsubsection"],

Cell["\<\
The default plot ranges may not be suitable for functions which vary strongly \
over parts of their domain.  For example, if there is a singularity, the \
default choices may emphasize the divergence while you would prefer to see \
the behavior for smaller values found away from the singularity.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Tan[x], \ {x, \(-\[Pi]\), \ \[Pi]}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.151576 0.304543 0.00648486 [
[.04527 .29204 -9.15625 -15.6875 ]
[.04527 .29204 9.15625 0 ]
[.19685 .29204 -9.15625 -15.6875 ]
[.19685 .29204 9.15625 0 ]
[.34842 .29204 -9.15625 -15.6875 ]
[.34842 .29204 9.15625 0 ]
[.65158 .29204 -5 -15.6875 ]
[.65158 .29204 5 0 ]
[.80315 .29204 -5 -15.6875 ]
[.80315 .29204 5 0 ]
[.95473 .29204 -5 -15.6875 ]
[.95473 .29204 5 0 ]
[.4875 .04515 -24.3125 -7.84375 ]
[.4875 .04515 0 7.84375 ]
[.4875 .17485 -24.3125 -7.84375 ]
[.4875 .17485 0 7.84375 ]
[.4875 .43424 -16 -7.84375 ]
[.4875 .43424 0 7.84375 ]
[.4875 .56394 -16 -7.84375 ]
[.4875 .56394 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.04527 .30454 m
.04527 .31079 L
s
gsave
.04527 .29204 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.19685 .30454 m
.19685 .31079 L
s
gsave
.19685 .29204 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.34842 .30454 m
.34842 .31079 L
s
gsave
.34842 .29204 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.65158 .30454 m
.65158 .31079 L
s
gsave
.65158 .29204 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.80315 .30454 m
.80315 .31079 L
s
gsave
.80315 .29204 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.95473 .30454 m
.95473 .31079 L
s
gsave
.95473 .29204 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.07559 .30454 m
.07559 .30829 L
s
.1059 .30454 m
.1059 .30829 L
s
.13622 .30454 m
.13622 .30829 L
s
.16653 .30454 m
.16653 .30829 L
s
.22716 .30454 m
.22716 .30829 L
s
.25748 .30454 m
.25748 .30829 L
s
.28779 .30454 m
.28779 .30829 L
s
.31811 .30454 m
.31811 .30829 L
s
.37874 .30454 m
.37874 .30829 L
s
.40905 .30454 m
.40905 .30829 L
s
.43937 .30454 m
.43937 .30829 L
s
.46968 .30454 m
.46968 .30829 L
s
.53032 .30454 m
.53032 .30829 L
s
.56063 .30454 m
.56063 .30829 L
s
.59095 .30454 m
.59095 .30829 L
s
.62126 .30454 m
.62126 .30829 L
s
.68189 .30454 m
.68189 .30829 L
s
.71221 .30454 m
.71221 .30829 L
s
.74252 .30454 m
.74252 .30829 L
s
.77284 .30454 m
.77284 .30829 L
s
.83347 .30454 m
.83347 .30829 L
s
.86378 .30454 m
.86378 .30829 L
s
.8941 .30454 m
.8941 .30829 L
s
.92441 .30454 m
.92441 .30829 L
s
.01496 .30454 m
.01496 .30829 L
s
.98504 .30454 m
.98504 .30829 L
s
.25 Mabswid
0 .30454 m
1 .30454 L
s
.5 .04515 m
.50625 .04515 L
s
gsave
.4875 .04515 -85.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
.5 .17485 m
.50625 .17485 L
s
gsave
.4875 .17485 -85.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.5 .43424 m
.50625 .43424 L
s
gsave
.4875 .43424 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.5 .56394 m
.50625 .56394 L
s
gsave
.4875 .56394 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .07757 m
.50375 .07757 L
s
.5 .11 m
.50375 .11 L
s
.5 .14242 m
.50375 .14242 L
s
.5 .20727 m
.50375 .20727 L
s
.5 .23969 m
.50375 .23969 L
s
.5 .27212 m
.50375 .27212 L
s
.5 .33697 m
.50375 .33697 L
s
.5 .36939 m
.50375 .36939 L
s
.5 .40182 m
.50375 .40182 L
s
.5 .46666 m
.50375 .46666 L
s
.5 .49909 m
.50375 .49909 L
s
.5 .53151 m
.50375 .53151 L
s
.5 .01272 m
.50375 .01272 L
s
.5 .59636 m
.50375 .59636 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .30454 m
.06244 .30623 L
.10458 .30837 L
.12507 .30966 L
.14415 .31114 L
.16254 .31298 L
.18221 .31572 L
.19242 .31769 L
.19781 .31895 L
.20351 .32054 L
.2085 .32218 L
.21399 .32437 L
.21861 .32663 L
.22365 .32969 L
.22885 .33381 L
.23379 .3391 L
.23612 .34229 L
.2386 .34638 L
.24092 .35109 L
.24306 .35644 L
.24577 .36524 L
.2472 .37116 L
.2487 .3788 L
.24992 .38641 L
.25123 .3965 L
.25267 .41086 L
.25401 .42898 L
.25529 .45305 L
.25601 .47126 L
.25668 .49259 L
.2579 .55021 L
s
.2579 .55021 m
.25865 .61803 L
s
.2621 .61803 m
.26231 0 L
s
.26516 0 m
.2652 .00603 L
.26634 .0829 L
.26766 .1337 L
.2689 .16411 L
.27005 .18399 L
.27111 .19792 L
.27228 .20994 L
.27351 .22003 L
.27475 .22823 L
.27609 .23544 L
.2785 .24554 L
.28136 .25431 L
.28398 .26033 L
.28916 .26887 L
.29209 .27241 L
.29481 .27514 L
.30016 .2794 L
.30508 .2824 L
.31017 .28487 L
.31568 .28704 L
.32564 .29004 L
.3347 .2921 L
.34426 .29381 L
.36394 .29641 L
.38229 .29817 L
.42125 .30083 L
.46266 .30291 L
.50256 .30465 L
.54094 .30634 L
.58177 .30843 L
.60207 .30972 L
.62109 .31121 L
.6393 .31304 L
.6589 .3158 L
.66849 .31766 L
.67898 .32032 L
.68427 .32203 L
.68992 .32425 L
.69486 .32666 L
.70009 .32986 L
.70543 .33417 L
.71036 .33958 L
.7131 .3435 L
.71565 .34803 L
.71846 .35433 L
.72147 .36345 L
.72407 .3744 L
.72553 .38259 L
.72689 .39208 L
Mistroke
.72807 .40243 L
.72937 .41712 L
.73003 .42629 L
.73074 .43801 L
.73202 .46618 L
.73318 .5046 L
.73446 .5749 L
Mfstroke
.73446 .5749 m
.73494 .61803 L
s
.73715 .61803 m
.73727 0 L
s
.7414 0 m
.74164 .02714 L
.74287 .09859 L
.74399 .13786 L
.74529 .16795 L
.74652 .18796 L
.74765 .20182 L
.7487 .21203 L
.75109 .22907 L
.75245 .23626 L
.75373 .24189 L
.75612 .25028 L
.7587 .25714 L
.76149 .26286 L
.76682 .27073 L
.7695 .2737 L
.77245 .27642 L
.77728 .28002 L
.78252 .28306 L
.78777 .28547 L
.79274 .28734 L
.80205 .2901 L
.81292 .29249 L
.82297 .2942 L
.84165 .29658 L
.86225 .29848 L
.90248 .30112 L
.94119 .30302 L
.97619 .30454 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oL@3oool2
000006@0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0000001`0?ooo`00L@3oool20000
06@0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0000001`0?ooo`00L@3oool2000006@0
oooo00<000000?ooo`3oool0H`3oool00`000000oooo0000001`0?ooo`00L@3oool2000006@0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0000001`0?ooo`00L@3oool2000006@0oooo00<0
00000?ooo`3oool0H`3oool00`000000oooo0000001`0?ooo`00L@3oool2000006@0oooo0`00001S
0?ooo`030000003oool000000700oooo001a0?ooo`800000I03oool00`000000oooo0?ooo`1S0?oo
o`030000003oool000000700oooo001a0?ooo`800000I03oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool000000700oooo001a0?ooo`800000I03oool00`000000oooo0?ooo`1S0?ooo`030000
003oool000000700oooo001a0?ooo`800000I03oool00`000000oooo0?ooo`1S0?ooo`030000003o
ool000000700oooo001a0?ooo`800000I03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
00000700oooo001a0?ooo`800000I03oool00`000000oooo0?ooo`1S0?ooo`030000003oool00000
0700oooo001a0?ooo`800000I03oool00`000000oooo0?ooo`1S0?ooo`030000003oool000000700
oooo001a0?ooo`800000I03oool00`000000oooo0?ooo`1S0?ooo`030000003oool000000700oooo
001a0?ooo`800000CP3oool0103ooml0W4P00018W03Oool30?ooo`050?oOW018000000000000M02o
ool02P3oool00`000000oooo0?ooo`1S0?ooo`030000003oool000000700oooo001a0?ooo`800000
CP3oool0103ooml0W4P00018W03Oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`09
0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`000000L03oool00740oooo0P00001>0?oo
o`040?oog`2LB00004RL0=ooo`80oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`08
0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`000000L03oool00740oooo0P00001:0?oo
o`030;md00000000000000D0000000@0M;oo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`08
0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`000000L03oool00740oooo0P00000m0?oo
o`030?oog`2LB000000000L0000000D0B9cO0?ooo`3oool0omnL07BLg`020?ooo`080?oog`2LB000
04RL0=ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo00P0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0000001`0?ooo`00L@3oool2000004/0oooo00/0gia807Boo`3oool0
oooO09a80000B9`0gooo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00P0oooo1000
001R0?ooo`030000003oool000000700oooo001a0?ooo`800000B`3oool02`3ooml0W4Qd0;ooo`3o
oml0W4P00018W03Oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0203oool00`00
0000oooo0?ooo`1S0?ooo`030000003oool000000700oooo001a0?ooo`800000C03oool02P3ogi`0
B7Bo0?oog`2LB00004RL0=ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo00P0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0000001`0?ooo`00L@3oool2000004d0oooo00T0
gia807Bog`2LB00004RL0=ooo`3oool0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`080?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`000000L03oool00740oooo0P00001=0?ooo`05
0?oog`2LB4P0B4P00018W03Oool00P3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo
00P0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0000001`0?ooo`00L@3oool2000004h0
oooo00@0omnL04P00000B9`0gooo0P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool02@3o
ool00`000000oooo0?ooo`1S0?ooo`030000003oool000000700oooo001a0?ooo`800000C`3oool0
0`3OW4P004RL0=ooo`030?ooo`050?oo_`1d000000000000M02oool02P3oool00`000000oooo0?oo
o`1S0?ooo`030000003oool000000700oooo001a0?ooo`800000I03oool00`000000oooo0?ooo`1S
0?ooo`030000003oool000000700oooo001a0?ooo`800000I03oool00`000000oooo0?ooo`1S0?oo
o`030000003oool000000700oooo001a0?ooo`800000I03oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool000000700oooo001a0?ooo`800000I03oool00`000000oooo0?ooo`1S0?ooo`030000
003oool000000700oooo001a0?ooo`800000I03oool00`000000oooo0?ooo`1S0?ooo`030000003o
ool000000700oooo001a0?ooo`800000I03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
00000700oooo001a0?ooo`800000I03oool00`000000oooo0?ooo`1S0?ooo`030000003oool00000
0700oooo001a0?ooo`800000I03oool3000006<0oooo00<000000?ooo`000000L03oool00740oooo
0P00001T0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`000000L03oool00740oooo0P00
001T0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`000000L03oool00740oooo0P00001T
0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`000000L03oool00740oooo0P00001T0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`000000L03oool00740oooo0P00001T0?ooo`03
0000003oool0oooo06<0oooo00<000000?ooo`000000L03oool00740oooo0P00001T0?ooo`030000
003oool0oooo06<0oooo00<000000?ooo`000000L03oool00740oooo0P00001T0?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`000000L03oool00740oooo0P00001T0?ooo`030000003oool0
oooo06<0oooo00<000000?ooo`000000L03oool00740oooo0P00001T0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`000000L03oool00740oooo0P00001T0?ooo`030000003oool0oooo06<0
oooo00<000000?ooo`000000L03oool00740oooo0P00001T0?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`000000L03oool00740oooo0P00001T0?ooo`030000003oool0oooo06<0oooo00<0
00000?ooo`000000L03oool00740oooo0P00001T0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`000000L03oool00740oooo00<000000?ooo`000000H`3oool3000006<0oooo00<000000?oo
o`000000L03oool00740oooo00<000000?ooo`000000H`3oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool000000700oooo001a0?ooo`030000003oool0000006<0oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0000001`0?ooo`00L@3oool00`000000oooo0000001S0?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`000000L03oool00740oooo00<000000?ooo`000000H`3oool0
0`000000oooo0?ooo`1S0?ooo`040000003oool0oooo000006l0oooo001a0?ooo`030000003oool0
000006<0oooo00<000000?ooo`3oool0H`3oool010000000oooo0?ooo`00001_0?ooo`00L@3oool0
0`000000oooo0000001S0?ooo`030000003oool0oooo06<0oooo00@000000?ooo`3oool00000K`3o
ool00740oooo00<000000?ooo`000000H`3oool00`000000oooo0?ooo`1S0?ooo`040000003oool0
oooo000006l0oooo001a0?ooo`030000003oool0000006<0oooo00<000000?ooo`3oool0H`3oool0
10000000oooo0?ooo`00001_0?ooo`00L@3oool00`000000oooo0000001S0?ooo`030000003oool0
oooo06<0oooo00@000000?ooo`3oool00000K`3oool00740oooo00<000000?ooo`000000H`3oool0
0`000000oooo0?ooo`1S0?ooo`040000003oool0oooo000006l0oooo001a0?ooo`030000003oool0
000006<0oooo00<000000?ooo`3oool0H`3oool010000000oooo0?ooo`00001_0?ooo`00L@3oool0
0`000000oooo0000001S0?ooo`030000003oool0oooo06<0oooo00@000000?ooo`3oool00000K`3o
ool00740oooo00<000000?ooo`000000H`3oool00`000000oooo0?ooo`1S0?ooo`040000003oool0
oooo000006l0oooo001a0?ooo`030000003oool0000006<0oooo0`00001S0?ooo`040000003oool0
oooo000006l0oooo001a0?ooo`030000003oool0000006<0oooo00<000000?ooo`3oool0H`3oool0
10000000oooo0?ooo`00001_0?ooo`00L@3oool00`000000oooo0000001S0?ooo`030000003oool0
oooo06<0oooo00@000000?ooo`3oool00000K`3oool00740oooo00@000000?ooo`3oool00000HP3o
ool00`000000oooo0?ooo`1S0?ooo`040000003oool0oooo000006l0oooo001a0?ooo`040000003o
ool0oooo00000680oooo00<000000?ooo`3oool0H`3oool010000000oooo0?ooo`00001_0?ooo`00
L@3oool010000000oooo0?ooo`00001R0?ooo`030000003oool0oooo06<0oooo00D000000?ooo`3o
ool0oooo0000001^0?ooo`00L@3oool010000000oooo0?ooo`00001R0?ooo`030000003oool0oooo
06<0oooo00D000000?ooo`3oool0oooo0000001^0?ooo`00L@3oool010000000oooo0?ooo`00001R
0?ooo`030000003oool0oooo06<0oooo00D000000?ooo`3oool0oooo0000001^0?ooo`00L@3oool0
10000000oooo0?ooo`00001R0?ooo`030000003oool0oooo06<0oooo00D000000?ooo`3oool0oooo
0000001^0?ooo`00L@3oool010000000oooo0?ooo`0000180?ooo`030;md00000000000000@00000
00T0B9cO0?ooo`3oool0oooo0?oOW018000000000000M02oool02P3oool00`000000oooo0?ooo`1S
0?ooo`050000003oool0oooo0?ooo`000000KP3oool00740oooo00@000000?ooo`3oool00000B@3o
ool00`3OW4P00000000000030000000:0018W03Oool0oooo0?oo_`1d0000M;oo0?ooo`3o_g@0001d
0;ooo`T0oooo00<000000?ooo`3oool0H`3oool01@000000oooo0?ooo`3oool0000006h0oooo001a
0?ooo`040000003oool0oooo000004X0oooo00<0gia804RLg`3oool00P3oool01P3OW7@0W=oo0?oo
o`2oM00004RL0=ooo`80oooo00<0_g@0001d_`3oool0203oool00`000000oooo0?ooo`1S0?ooo`05
0000003oool0oooo0?ooo`000000KP3oool00740oooo00@000000?ooo`3oool00000B`3oool00`3O
W4P0M;oo0?ooo`030?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool00`00
0000oooo0?ooo`1S0?ooo`050000003oool0oooo0?ooo`000000KP3oool00740oooo00@000000?oo
o`3oool00000>`3oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00@0oooo00<0_gA8
09cOo`3oool00P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00P0oooo00<00000
0?ooo`3oool0H`3oool01@000000oooo0?ooo`3oool0000006h0oooo001a0?ooo`040000003oool0
oooo000004`0oooo00<0oooO09a8M02oool00P3oool00`3ogi`0B00007Boo`030?ooo`030?noM000
04P0W=oo00P0oooo1000001R0?ooo`050000003oool0oooo0?ooo`000000KP3oool00740oooo00@0
00000?ooo`3oool00000C@3oool01`3o_g@007Bo0?ooo`3oool0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`080?ooo`030000003oool0oooo06<0oooo00D000000?ooo`3oool0oooo0000001^
0?ooo`00L@3oool010000000oooo0?ooo`00001=0?ooo`070?oo_`1d04P0W=oo0?ooo`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00P0oooo00<000000?ooo`3oool0H`3oool01@000000oooo
0?ooo`3oool0000006h0oooo001a0?ooo`040000003oool0oooo000004d0oooo00L0oono07@00018
W=l0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool00`000000oooo0?ooo`1S
0?ooo`050000003oool0oooo0?ooo`000000KP3oool00740oooo00@000000?ooo`3oool00000B03o
ool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO0080oooo00<0gia80018W03Oool00P3o
ool00`2oM00007Bo0?ooo`080?ooo`030000003oool0oooo06<0oooo00D000000?ooo`3oool0oooo
0000001^0?ooo`00L@3oool010000000oooo0?ooo`0000180?ooo`080?oog`2LB4P0W=oo0?ooo`3o
okl0M0000000M02oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`030000
003oool0oooo06<0oooo00D000000?ooo`3oool0oooo0000001^0?ooo`00L@3oool010000000oooo
0?ooo`0000190?ooo`060?oog`2LB00000000000000007@0_ooo103oool01@3ookl0M00000000000
07@0_ooo00X0oooo00<000000?ooo`3oool0H`3oool01@000000oooo0?ooo`3oool0000006h0oooo
001a0?ooo`040000003oool0oooo00000680oooo00<000000?ooo`3oool0H`3oool01@000000oooo
0?ooo`3oool0000006h0oooo001a0?ooo`050000003oool0oooo0?ooo`000000H@3oool00`000000
oooo0?ooo`1S0?ooo`050000003oool0oooo0?ooo`000000KP3oool00740oooo00D000000?ooo`3o
ool0oooo0000001Q0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`1[0?ooo`00L@3oool01@000000oooo0?ooo`3oool000000640oooo00<000000?oo
o`3oool0H`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo06/0oooo001a0?ooo`05
0000003oool0oooo0?ooo`000000H@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0J`3oool00740oooo00D000000?ooo`3oool0oooo0000001Q0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1[0?oo
o`00L@3oool01@000000oooo0?ooo`3oool000000640oooo00<000000?ooo`3oool0H`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo06/0oooo001a0?ooo`050000003oool0oooo0?oo
o`000000H@3oool3000006<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1[0?oo
o`00L@3oool01@000000oooo0?ooo`3oool000000640oooo00<000000?ooo`3oool0H`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo06/0oooo001a0?ooo`050000003oool0oooo0?oo
o`000000H@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0J`3oool00740oooo00D000000?ooo`3oool0oooo0000001Q0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1[0?ooo`00L@3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool0H`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo06/0oooo001a0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0J`3oool00740oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`1N0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`1Z0?ooo`00L@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo05h0oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo06X0oooo001a0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`1S0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool0JP3oool00740oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`1Z0?ooo`00L@3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo05h0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo06X0oooo001a0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0G@3o
ool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0J@3o
ool00740oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0
oooo06<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`1Y0?ooo`00L@3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool0H`3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo06T0oooo001a0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0G@3oool3000006<0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`1Y0?ooo`00L@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo05`0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo06T0oooo
001a0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0G03oool00`000000oooo0?oo
o`1S0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0J@3oool00740oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`1X0?ooo`00L@3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo05`0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo06P0oooo001a0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool0F`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool0J03oool001D0oooo00<0oono07@0000000000P0000000`1d_ol0oooo0?ooo`0i0?ooo`030;md
00000000000000@0000000<0B9cO0?ooo`3oool04@3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo01d0oooo00H0_g@000000000000000000018W03Ooolh0?ooo`030000003oool0oooo
03`0oooo00H0_g@000000000000000000018W03OoolQ0?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool0303oool00`2oM0000000000000040000000304RLg`3oool0oooo03L0oooo00<0
oono07@0000000000P0000000`1d_ol0oooo0?ooo`0B0?ooo`005@3oool00`3ogi`0B00004RLg`02
0?ooo`030;md0018W=l0oooo03X0oooo00<0gia80000000000000`0000000`00B9`0gooo0?ooo`0A
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool07P3oool00`3ogi`0B00007Boo`0j
0?ooo`030000003oool0oooo03d0oooo00<0omnL04P0001d_ol08`3oool00`000000oooo0?ooo`06
0?ooo`030000003oool0oooo00d0oooo00<0gia80000000000000`0000000`00B9`0gooo0?ooo`0g
0?ooo`030?oOW0180000B9cO0080oooo00<0_g@004RLg`3oool04P3oool001X0oooo00<0oono07@0
B02Lgol0>`3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo0140oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`0M0?ooo`030?oOW0180000M;oo03X0oooo00<000000?oo
o`3oool0?@3oool00`3ogi`0B00007Boo`0S0?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool03P3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo03`0oooo00<0oono07@0
B02Lgol04P3oool001X0oooo00<0oono07@0001d_ol0?03oool00`3OW4P0M;oo0?ooo`0E0?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool07@3oool00`3ogi`0B00007Boo`0j0?ooo`03
0000003oool0oooo03d0oooo00<0omnL04P0001d_ol08`3oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo00h0oooo00<0gia807Boo`3oool0@03oool00`3ookl0M00007Boo`0B0?ooo`00
2@3oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00@0oooo00<0okmd0000001d_ol0
;@3oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00<0oooo00<0_gA809cOo`3oool0
503oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00d0oooo00<0oooO09a800000000
1`0000000`18W=l0oooo0?ooo`020?ooo`030?oOW0180000M;oo03X0oooo00<000000?ooo`3oool0
?@3oool00`3ogi`0B00007Boo`0S0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
3`3oool00`2oM4P0W=oo0?ooo`0o0?ooo`030?noM0000000M;oo0180oooo000I0?ooo`040?oOW018
0000001809cOocd0oooo00<0oooO09a8M02oool0503oool00`000000oooo0?ooo`070?ooo`030000
003oool0oooo01`0oooo00<0omnL04P0001d_ol0>P3oool00`000000oooo0?ooo`0m0?ooo`030?oO
W0180000M;oo02<0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0>0?ooo`030?oo
g`2LB7@0_ooo03h0oooo00@0omnL04P0000004P0W=oo4P3oool001L0oooo00D0okmd000000000000
001809cOo`0o0?ooo`030?noM000M;l0oooo01<0oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`0K0?ooo`030?oOW0180000M;oo03X0oooo00<000000?ooo`3oool0?@3oool00`3ogi`0
B00007Boo`0S0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool03P3oool00`3o_g@0
07Bo0?ooo`0k0?ooo`050?noM000000000000000B02Lgol04`3oool001T0oooo00<0_g@0001d_`3o
ool0?`3oool00`3ookl0M01809cOo`0C0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool06P3oool00`3ogi`0B00007Boo`0j0?ooo`030000003oool0oooo03d0oooo00<0omnL04P0001d
_ol08`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00d0oooo00<0oono07@0B02L
gol0?@3oool00`2oM00007Bo0?ooo`0C0?ooo`006P3oool00`2oM000B9cO0?ooo`0n0?ooo`030?oo
_`1d0000B9cO01<0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0I0?ooo`030?oO
W0180000M;oo03X0oooo0`00000m0?ooo`030?oOW0180000M;oo02<0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`0<0?ooo`030?oo_`1d0000B9cO03h0oooo00<0_g@004RLg`3oool0
4P3oool001D0oooo00<0oooO09bL_`3oool00P3oool00`3OW4P0001d0;ooo`0i0?ooo`030?noW02L
gol0oooo0080oooo00<0omnL04P00018W=l04`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo01P0oooo00<0omnL04P0001d_ol0>P3oool00`000000oooo0?ooo`0m0?ooo`030?oOW018
0000M;oo02<0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`070?ooo`030?noW02L
gol0oooo0080oooo00<0omnL04P00018W=l0>@3oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB000
07@0_ooo0180oooo000F0?ooo`060?noM018W=l0oooo0?oo_`1d000007Bo>P3oool0203ooml0W4Q8
09cOo`3oool0oono07@0000007@0_ooo4`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo01D0oooo00D0oooO09a800000000000007Boo`0j0?ooo`030000003oool0oooo03/0oooo00D0
oooO09a800000000000007Boo`0S0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
1P3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo>P3oool01P3o_g@0B9cO0?ooo`3o
okl0M000001d_a<0oooo000G0?ooo`050=nLB000000000000000M02oool0>`3oool01P3ooml0W4P0
000000000000001d0;oooa@0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0F0?oo
o`030?noM0000000M;oo03X0oooo00<000000?ooo`3oool0?@3oool00`3o_g@0000007Boo`0S0?oo
o`030000003oool0oooo00d0oooo0`0000060?ooo`060?oog`2LB00000000000000007@0_ooo?03o
ool01@3OW4P000000000000007@0_ooo01<0oooo001a0?ooo`030000003oool0oooo00h0oooo0P00
001C0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0403oool2000005h0oooo001a
0?ooo`030000003oool0oooo0100oooo0`00001@0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool04P3oool2000005`0oooo001a0?ooo`030000003oool0oooo01<0oooo0`00001=0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool0503oool3000005T0oooo001a0?ooo`03
0000003oool0oooo01H0oooo100000190?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3o
ool05`3oool5000005@0oooo001a0?ooo`030000003oool0oooo01X0oooo100000150?ooo`030000
003oool0oooo06<0oooo00<000000?ooo`3oool0703oool400000500oooo001a0?ooo`030000003o
ool0oooo01h0oooo1P00000o0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0803o
ool6000004X0oooo001a0?ooo`030000003oool0oooo02@0oooo2000000g0?ooo`030000003oool0
oooo06<0oooo00<000000?ooo`3oool09P3oool900000440oooo001a0?ooo`030000003oool0oooo
02`0oooo3@00000Z0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0;`3oool=0000
03@0oooo001a0?ooo`030000003oool0oooo03T0oooo4@00000I0?ooo`030000003oool0oooo06<0
oooo00<000000?ooo`3oool0?03ooolA000002<0oooo001a0?ooo`030000003oool0oooo04X0oooo
4P0000070?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0C@3oool@000001<0oooo
003o00000;0000000@3oool000H0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:
0?ooo`040000003oool0oooo0?ooo`h000002@3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0000000X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo3P00000<0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`000000303o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00D0oooo000C0?ooo`030000003oool0
oooo01`0oooo4000000C0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool08P3oool0
0`000000oooo0?ooo`0n0?ooo`030000003oool0oooo02D0oooo2000000A0?ooo`030000003oool0
oooo0280oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0
oooo0180oooo00120?ooo`P000009`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
02d0oooo4000000V0?ooo`030000003oool0oooo0700oooo001:0?ooo`L00000803oool00`000000
oooo0?ooo`1S0?ooo`030000003oool0oooo03d0oooo1P00000P0?ooo`030000003oool0oooo0700
oooo001A0?ooo`<000007@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo04<0oooo
1@00000K0?ooo`030000003oool0oooo0700oooo001D0?ooo`8000006`3oool00`000000oooo0?oo
o`1S0?ooo`030000003oool0oooo04P0oooo0`00000H0?ooo`030000003oool0oooo0700oooo001F
0?ooo`D000005P3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo04/0oooo0`00000E
0?ooo`030000003oool0oooo0700oooo001K0?ooo`800000503oool00`000000oooo0?ooo`1S0?oo
o`030000003oool0oooo04h0oooo0P00000C0?ooo`030000003oool0oooo0700oooo001M0?ooo`80
00004P3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0500oooo0P00000A0?ooo`03
0000003oool0oooo0700oooo001O0?ooo`800000403oool00`000000oooo0?ooo`1S0?ooo`030000
003oool0oooo0580oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1`0?ooo`00H@3o
ool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0D`3o
ool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0700oooo001R0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1D0?ooo`8000003@3oool0
0`000000oooo0?ooo`1`0?ooo`00H`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
0700oooo001S0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0H`3oool3000005H0
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`1`0?ooo`00I03oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0E`3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo0700oooo001U0?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool0H`3oool00`000000oooo0?ooo`1H0?ooo`8000002@3oool00`000000oooo0?oo
o`1`0?ooo`00IP3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool0F@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0700oooo001W0?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1J0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool0L03oool006L0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`1`0?ooo`00J03oool00`000000oooo0?ooo`060?ooo`030000
003oool0oooo06<0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo0700oooo001X0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0H`3o
ool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0L03o
ool006P0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
oooo05/0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`1`0?ooo`00J03oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0F`3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo0700oooo001Y0?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool0L03oool006T0oooo00<000000?ooo`3oool01@3oool00`000000oooo
0?ooo`1S0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`1`0?ooo`00JP3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo06<0oooo00<0
00000?ooo`3oool0G03oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0700oooo001Z
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1L
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0L03oool006X0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`1S0?ooo`<00000G@3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0700oooo001Z0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool0L03oool006X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`1S0?ooo`030000
003oool0oooo05d0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`1`0?ooo`00J`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0G@3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0700oooo001[0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1N0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0L03oool006/0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`1S0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`1`0?ooo`00J`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo06<0
oooo00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0700
oooo001[0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0H`3oool00`000000oooo
0?ooo`1N0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0L03oool006/0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo05l0oooo00D0
00000?ooo`3oool0oooo0000001b0?ooo`00J`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`3oool0G`3oool01@000000oooo0?ooo`3oool000000780oooo
001/0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?oo
o`1O0?ooo`050000003oool0oooo0?ooo`000000LP3oool006`0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo05l0oooo00D000000?ooo`3oool0oooo
0000001b0?ooo`00K03oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo06<0oooo00<0
00000?ooo`3oool0G`3oool01@000000oooo0?ooo`3oool000000780oooo001/0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0H`3oool3000005l0oooo00D000000?ooo`3oool0oooo
0000001b0?ooo`00K03oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo06<0oooo00<0
00000?ooo`3oool0G`3oool01@000000oooo0?ooo`3oool000000780oooo001]0?ooo`050000003o
ool0oooo0?ooo`000000I@3oool00`000000oooo0?ooo`1O0?ooo`050000003oool0oooo0?ooo`00
0000LP3oool006d0oooo00D000000?ooo`3oool0oooo0000001U0?ooo`030000003oool0oooo05l0
oooo00D000000?ooo`3oool0oooo0000001b0?ooo`00K@3oool01@000000oooo0?ooo`3oool00000
06D0oooo00<000000?ooo`3oool0G`3oool01@000000oooo0?ooo`3oool000000780oooo001]0?oo
o`050000003oool0oooo0?ooo`000000I@3oool00`000000oooo0?ooo`1P0?ooo`040000003oool0
oooo00000780oooo001]0?ooo`050000003oool0oooo0?ooo`000000I@3oool00`000000oooo0?oo
o`1P0?ooo`040000003oool0oooo00000780oooo001]0?ooo`050000003oool0oooo0?ooo`000000
I@3oool00`000000oooo0?ooo`1P0?ooo`040000003oool0oooo00000780oooo001]0?ooo`050000
003oool0oooo0?ooo`000000I@3oool00`000000oooo0?ooo`1P0?ooo`040000003oool0oooo0000
0780oooo001]0?ooo`050000003oool0oooo0?ooo`000000B`3oool00`2oM0000000000000040000
000904RLg`3oool0oooo0?ooo`3ogi`0B0000000000007@0_ooo00X0oooo00<000000?ooo`3oool0
H03oool010000000oooo0?ooo`00001b0?ooo`00K@3oool01@000000oooo0?ooo`3oool0000004`0
oooo00<0gia80000000000000`0000002P00B9`0gooo0?ooo`3ookl0M00007Boo`3oool0okmd0000
M02oool90?ooo`030000003oool0oooo0600oooo00@000000?ooo`3oool00000LP3oool006d0oooo
00D000000?ooo`3oool0oooo0000001=0?ooo`030=nLB018W=l0oooo0080oooo00H0giad09cOo`3o
ool0_g@00018W03Oool20?ooo`030;md0000M;l0oooo00P0oooo00<000000?ooo`3oool0H03oool0
10000000oooo0?ooo`00001b0?ooo`00K@3oool01@000000oooo0?ooo`3oool0000004h0oooo00<0
gia807Boo`3oool00`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00P0oooo00<0
00000?ooo`3oool0H03oool010000000oooo0?ooo`00001b0?ooo`00K@3oool01@000000oooo0?oo
o`3oool0000004l0oooo00<0_gA809cOo`3oool00P3oool00`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00P0oooo00<000000?ooo`3oool0H03oool010000000oooo0?ooo`00001b0?ooo`00
K@3oool01@000000oooo0?ooo`3oool0000004l0oooo00<0oooO09a8M02oool00P3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00P0oooo1000001O0?ooo`040000003oool0oooo0000
0780oooo001]0?ooo`050000003oool0oooo0?ooo`000000D03oool01`3o_g@007Bo0?ooo`3oool0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`080?ooo`030000003oool0oooo0600oooo00@0
00000?ooo`3oool00000LP3oool006d0oooo00D000000?ooo`3oool0oooo0000001@0?ooo`070?oo
_`1d04P0W=oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00P0oooo00<000000?oo
o`3oool0H03oool010000000oooo0?ooo`00001b0?ooo`00K@3oool01@000000oooo0?ooo`3oool0
00000500oooo00L0oono07@00018W=l0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0
203oool00`000000oooo0?ooo`1P0?ooo`040000003oool0oooo00000780oooo001^0?ooo`040000
003oool0oooo000004/0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`020?ooo`03
0=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0203oool00`000000oooo0?ooo`1P0?ooo`04
0000003oool0oooo00000780oooo001^0?ooo`040000003oool0oooo000004/0oooo00P0oooO09a8
B02Lgol0oooo0?oo_`1d0000001d0;ooo`80oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo
00T0oooo00<000000?ooo`3oool0H03oool010000000oooo0?ooo`00001b0?ooo`00KP3oool01000
0000oooo0?ooo`00001<0?ooo`060?oog`2LB00000000000000007@0_ooo103oool01@3ookl0M000
0000000007@0_ooo00X0oooo00<000000?ooo`3oool0H03oool010000000oooo0?ooo`00001b0?oo
o`00KP3oool010000000oooo0?ooo`00001U0?ooo`030000003oool0oooo0600oooo00@000000?oo
o`3oool00000LP3oool006h0oooo00@000000?ooo`3oool00000I@3oool00`000000oooo0?ooo`1P
0?ooo`040000003oool0oooo00000780oooo001^0?ooo`040000003oool0oooo000006D0oooo00<0
00000?ooo`3oool0H@3oool00`000000oooo0000001b0?ooo`00KP3oool010000000oooo0?ooo`00
001U0?ooo`030000003oool0oooo0640oooo00<000000?ooo`000000LP3oool006h0oooo00@00000
0?ooo`3oool00000I@3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool000000780oooo001^
0?ooo`040000003oool0oooo000006D0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0000
001b0?ooo`00KP3oool010000000oooo0?ooo`00001U0?ooo`030000003oool0oooo0640oooo00<0
00000?ooo`000000LP3oool006h0oooo00@000000?ooo`3oool00000I@3oool300000640oooo00<0
00000?ooo`000000LP3oool006h0oooo00@000000?ooo`3oool00000I@3oool00`000000oooo0?oo
o`1Q0?ooo`030000003oool000000780oooo001^0?ooo`040000003oool0oooo000006D0oooo00<0
00000?ooo`3oool0H@3oool00`000000oooo0000001b0?ooo`00KP3oool010000000oooo0?ooo`00
001U0?ooo`030000003oool0oooo0640oooo00<000000?ooo`000000LP3oool006h0oooo00@00000
0?ooo`3oool00000I@3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool000000780oooo001^
0?ooo`040000003oool0oooo000006D0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0000
001b0?ooo`00KP3oool010000000oooo0?ooo`00001U0?ooo`030000003oool0oooo0640oooo00<0
00000?ooo`000000LP3oool006h0oooo00@000000?ooo`3oool00000I@3oool00`000000oooo0?oo
o`1Q0?ooo`030000003oool000000780oooo001^0?ooo`040000003oool0oooo000006D0oooo00<0
00000?ooo`3oool0H@3oool00`000000oooo0000001b0?ooo`00KP3oool010000000oooo0?ooo`00
001U0?ooo`030000003oool0oooo0640oooo00<000000?ooo`000000LP3oool006h0oooo00@00000
0?ooo`3oool00000I@3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool000000780oooo001^
0?ooo`040000003oool0oooo000006D0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0000
001b0?ooo`00KP3oool010000000oooo0?ooo`00001U0?ooo`030000003oool0oooo0640oooo00<0
00000?ooo`000000LP3oool006h0oooo00@000000?ooo`3oool00000I@3oool00`000000oooo0?oo
o`1Q0?ooo`030000003oool000000780oooo001^0?ooo`040000003oool0oooo000006D0oooo0`00
001Q0?ooo`030000003oool000000780oooo001^0?ooo`040000003oool0oooo000006D0oooo00<0
00000?ooo`3oool0H@3oool00`000000oooo0000001b0?ooo`00KP3oool010000000oooo0?ooo`00
001U0?ooo`030000003oool0oooo0640oooo00<000000?ooo`000000LP3oool006h0oooo00@00000
0?ooo`3oool00000I@3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool000000780oooo001^
0?ooo`040000003oool0oooo000006D0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0000
001b0?ooo`00KP3oool010000000oooo0?ooo`00001U0?ooo`030000003oool0oooo0640oooo00<0
00000?ooo`000000LP3oool006h0oooo00@000000?ooo`3oool00000I@3oool00`000000oooo0?oo
o`1Q0?ooo`030000003oool000000780oooo001^0?ooo`040000003oool0oooo000006D0oooo00<0
00000?ooo`3oool0H@3oool00`000000oooo0000001b0?ooo`00KP3oool010000000oooo0?ooo`00
001U0?ooo`030000003oool0oooo0640oooo00<000000?ooo`000000LP3oool006h0oooo00@00000
0?ooo`3oool00000I@3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool000000780oooo001^
0?ooo`040000003oool0oooo000006D0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0000
001b0?ooo`00K`3oool00`000000oooo0000001U0?ooo`030000003oool0oooo0640oooo00<00000
0?ooo`000000LP3oool006l0oooo00<000000?ooo`000000I@3oool00`000000oooo0?ooo`1Q0?oo
o`030000003oool000000780oooo001_0?ooo`030000003oool0000006D0oooo00<000000?ooo`3o
ool0H@3oool00`000000oooo0000001b0?ooo`00K`3oool00`000000oooo0000001U0?ooo`<00000
H@3oool00`000000oooo0000001b0?ooo`00K`3oool00`000000oooo0000001U0?ooo`030000003o
ool0oooo0640oooo00<000000?ooo`000000LP3oool006l0oooo00<000000?ooo`000000I@3oool0
0`000000oooo0?ooo`1Q0?ooo`030000003oool000000780oooo001_0?ooo`030000003oool00000
06D0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0000001b0?ooo`00K`3oool00`000000
oooo0000001U0?ooo`030000003oool0oooo0640oooo00<000000?ooo`000000LP3oool006l0oooo
00<000000?ooo`000000I@3oool00`000000oooo0?ooo`1R0?ooo`800000LP3oool006l0oooo00<0
00000?ooo`000000I@3oool00`000000oooo0?ooo`1R0?ooo`800000LP3oool006l0oooo00<00000
0?ooo`000000I@3oool00`000000oooo0?ooo`1R0?ooo`800000LP3oool006l0oooo00<000000?oo
o`000000I@3oool00`000000oooo0?ooo`1R0?ooo`800000LP3oool006l0oooo00<000000?ooo`00
0000C`3oool0103ooml0W4P00018W03Oool30?ooo`050?oOW018000000000000M02oool02P3oool0
0`000000oooo0?ooo`1R0?ooo`800000LP3oool006l0oooo00<000000?ooo`000000C`3oool0103o
oml0W4P00018W03Oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`030000
003oool0oooo0680oooo0P00001b0?ooo`00K`3oool00`000000oooo0000001?0?ooo`040?oog`2L
B00004RL0=ooo`80oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`080?ooo`030000
003oool0oooo0680oooo0P00001b0?ooo`00K`3oool00`000000oooo0000001;0?ooo`030;md0000
0000000000D0000000@0M;oo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`080?ooo`030000
003oool0oooo0680oooo0P00001b0?ooo`00K`3oool00`000000oooo0000001;0?ooo`0<0?oOW01d
W=l0oooo0?ooo`3ooml0W4P00018W03Oool0oooo0?oOW0180000M;oo0`3oool00`3o_g@0001809cO
o`080?ooo`030000003oool0oooo0680oooo0P00001b0?ooo`00K`3oool00`000000oooo0000001<
0?ooo`0;0=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`3oool0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`080?ooo`@00000H@3oool200000780oooo001_0?ooo`030000003oool0000004`0
oooo00/0oooO09a8M02oool0oooO09a80000B9`0gooo0?ooo`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00P0oooo00<000000?ooo`3oool0HP3oool200000780oooo001_0?ooo`030000003o
ool0000004d0oooo00X0omnL04Qd_`3ooml0W4P00018W03Oool0oooo0?oOW0180000M;oo0`3oool0
0`3o_g@0001809cOo`080?ooo`030000003oool0oooo0680oooo0P00001b0?ooo`00K`3oool00`00
0000oooo0000001>0?ooo`090=nLB01d_ml0W4P00018W03Oool0oooo0?oo_`1d0000B9cO00<0oooo
00<0gia80000M02oool0203oool00`000000oooo0?ooo`1R0?ooo`800000LP3oool006l0oooo00<0
00000?ooo`000000CP3oool01@3ooml0W4Q804Q80000B9`0gooo0080oooo00<0gia80018W03Oool0
0P3oool00`2oM00007Bo0?ooo`080?ooo`030000003oool0oooo0680oooo0P00001b0?ooo`00K`3o
ool00`000000oooo0000001?0?ooo`040?oOW018000004RL0=ooo`80oooo00L0oono07@0001d_ol0
oooo0?noM00007@0_ooo00T0oooo00<000000?ooo`3oool0HP3oool200000780oooo001_0?ooo`03
0000003oool000000500oooo00<0gia80018W03Oool00`3oool01@3ookl0M0000000000007@0_ooo
00X0oooo00<000000?ooo`3oool0HP3oool200000780oooo001_0?ooo`030000003oool0000006D0
oooo00<000000?ooo`3oool0HP3oool200000780oooo001_0?ooo`030000003oool0000006D0oooo
00<000000?ooo`3oool0HP3oool200000780oooo001_0?ooo`030000003oool0000006D0oooo00<0
00000?ooo`3oool0HP3oool200000780oooo001_0?ooo`030000003oool0000006D0oooo00<00000
0?ooo`3oool0HP3oool200000780oooo001_0?ooo`030000003oool0000006D0oooo00<000000?oo
o`3oool0HP3oool200000780oooo001_0?ooo`030000003oool0000006D0oooo00<000000?ooo`3o
ool0HP3oool200000780oooo001_0?ooo`030000003oool0000006D0oooo00<000000?ooo`3oool0
HP3oool200000780oooo001_0?ooo`030000003oool0000006D0oooo0`00001R0?ooo`800000LP3o
ool006l0oooo00<000000?ooo`000000I@3oool00`000000oooo0?ooo`1R0?ooo`800000LP3oool0
06l0oooo00<000000?ooo`000000I@3oool00`000000oooo0?ooo`1R0?ooo`800000LP3oool006l0
oooo00<000000?ooo`000000I@3oool00`000000oooo0?ooo`1R0?ooo`800000LP3oool006l0oooo
00<000000?ooo`000000I@3oool00`000000oooo0?ooo`1R0?ooo`800000LP3oool006l0oooo00<0
00000?ooo`000000I@3oool00`000000oooo0?ooo`1R0?ooo`800000LP3oool006l0oooo00<00000
0?ooo`000000I@3oool00`000000oooo0?ooo`1R0?ooo`800000LP3oool006l0oooo00<000000?oo
o`000000I@3oool00`000000oooo0?ooo`1R0?ooo`800000LP3oool006l0oooo00<000000?ooo`00
0000I@3oool00`000000oooo0?ooo`1R0?ooo`800000LP3oool006l0oooo00<000000?ooo`000000
I@3oool00`000000oooo0?ooo`1R0?ooo`800000LP3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-3.30486, -46.9626, \
0.0230037, 0.537684}}]
}, Closed]],

Cell[TextData[{
  "The ",
  ButtonBox["PlotRange",
    ButtonStyle->"RefGuideLink"],
  " option can be used to adjust these ranges.  If we replace the default \
with a list of the form ",
  StyleBox["{{xmin,xmax},{ymin,ymax}}", "InlineInput"],
  " we can limit the plot to specified ranges."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Tan[x], \ {x, \(-\[Pi]\), \ \[Pi]}, 
        PlotRange \[Rule] {Automatic, {\(-5\), 5}}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.151576 0.309017 0.0618034 [
[.04527 .29652 -9.15625 -15.6875 ]
[.04527 .29652 9.15625 0 ]
[.19685 .29652 -9.15625 -15.6875 ]
[.19685 .29652 9.15625 0 ]
[.34842 .29652 -9.15625 -15.6875 ]
[.34842 .29652 9.15625 0 ]
[.65158 .29652 -5 -15.6875 ]
[.65158 .29652 5 0 ]
[.80315 .29652 -5 -15.6875 ]
[.80315 .29652 5 0 ]
[.95473 .29652 -5 -15.6875 ]
[.95473 .29652 5 0 ]
[.4875 .0618 -18.3125 -7.84375 ]
[.4875 .0618 0 7.84375 ]
[.4875 .18541 -18.3125 -7.84375 ]
[.4875 .18541 0 7.84375 ]
[.4875 .43262 -10 -7.84375 ]
[.4875 .43262 0 7.84375 ]
[.4875 .55623 -10 -7.84375 ]
[.4875 .55623 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.04527 .30902 m
.04527 .31527 L
s
gsave
.04527 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.19685 .30902 m
.19685 .31527 L
s
gsave
.19685 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.34842 .30902 m
.34842 .31527 L
s
gsave
.34842 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.65158 .30902 m
.65158 .31527 L
s
gsave
.65158 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.80315 .30902 m
.80315 .31527 L
s
gsave
.80315 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.95473 .30902 m
.95473 .31527 L
s
gsave
.95473 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.07559 .30902 m
.07559 .31277 L
s
.1059 .30902 m
.1059 .31277 L
s
.13622 .30902 m
.13622 .31277 L
s
.16653 .30902 m
.16653 .31277 L
s
.22716 .30902 m
.22716 .31277 L
s
.25748 .30902 m
.25748 .31277 L
s
.28779 .30902 m
.28779 .31277 L
s
.31811 .30902 m
.31811 .31277 L
s
.37874 .30902 m
.37874 .31277 L
s
.40905 .30902 m
.40905 .31277 L
s
.43937 .30902 m
.43937 .31277 L
s
.46968 .30902 m
.46968 .31277 L
s
.53032 .30902 m
.53032 .31277 L
s
.56063 .30902 m
.56063 .31277 L
s
.59095 .30902 m
.59095 .31277 L
s
.62126 .30902 m
.62126 .31277 L
s
.68189 .30902 m
.68189 .31277 L
s
.71221 .30902 m
.71221 .31277 L
s
.74252 .30902 m
.74252 .31277 L
s
.77284 .30902 m
.77284 .31277 L
s
.83347 .30902 m
.83347 .31277 L
s
.86378 .30902 m
.86378 .31277 L
s
.8941 .30902 m
.8941 .31277 L
s
.92441 .30902 m
.92441 .31277 L
s
.01496 .30902 m
.01496 .31277 L
s
.98504 .30902 m
.98504 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5 .0618 m
.50625 .0618 L
s
gsave
.4875 .0618 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.5 .18541 m
.50625 .18541 L
s
gsave
.4875 .18541 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 .43262 m
.50625 .43262 L
s
gsave
.4875 .43262 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 .55623 m
.50625 .55623 L
s
gsave
.4875 .55623 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .09271 m
.50375 .09271 L
s
.5 .12361 m
.50375 .12361 L
s
.5 .15451 m
.50375 .15451 L
s
.5 .21631 m
.50375 .21631 L
s
.5 .24721 m
.50375 .24721 L
s
.5 .27812 m
.50375 .27812 L
s
.5 .33992 m
.50375 .33992 L
s
.5 .37082 m
.50375 .37082 L
s
.5 .40172 m
.50375 .40172 L
s
.5 .46353 m
.50375 .46353 L
s
.5 .49443 m
.50375 .49443 L
s
.5 .52533 m
.50375 .52533 L
s
.5 .0309 m
.50375 .0309 L
s
.5 .58713 m
.50375 .58713 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .30902 m
.06244 .32512 L
.10458 .34547 L
.12507 .35778 L
.14415 .37189 L
.16254 .38939 L
.18221 .41552 L
.19242 .43427 L
.19781 .44635 L
.20351 .46143 L
.2085 .4771 L
.21399 .49795 L
.21861 .51947 L
.22365 .5487 L
.22885 .58791 L
s
.22885 .58791 m
.2318 .61803 L
s
.26219 .61803 m
.26222 0 L
s
.29185 0 m
.29209 .00278 L
.29481 .02877 L
.30016 .06936 L
.30508 .09795 L
.31017 .12155 L
.31568 .14217 L
.32564 .17081 L
.3347 .19038 L
.34426 .20669 L
.36394 .23151 L
.38229 .24829 L
.42125 .27367 L
.46266 .29348 L
.50256 .31006 L
.54094 .32613 L
.58177 .34602 L
.60207 .35832 L
.62109 .37251 L
.6393 .39001 L
.6589 .41633 L
.66849 .43401 L
.67898 .45936 L
.68427 .47568 L
.68992 .49688 L
.69486 .51979 L
.70009 .55031 L
.70543 .59139 L
s
.70543 .59139 m
.70798 .61803 L
s
.73721 .61803 m
.73722 0 L
s
.76807 0 m
.7695 .01503 L
.77245 .041 L
.77728 .07527 L
.78252 .10424 L
.78777 .12724 L
.79274 .14507 L
.80205 .17133 L
.81292 .19415 L
.82297 .21043 L
.84165 .23308 L
.86225 .25125 L
.90248 .27634 L
.94119 .29448 L
.97619 .30902 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oL@3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo06<0oooo001a0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool0H`3oool00740oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`1F0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`1S0?ooo`00L@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo05H0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo06<0oooo
001a0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?oo
o`1S0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0HP3oool00740oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`1R0?ooo`00L@3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo05H0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo0680oooo001a0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool0E@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool0HP3oool00740oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1E0?ooo`030000
003oool0oooo06<0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1R0?ooo`00L@3o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0H`3o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0680oooo001a0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool0HP3oool00740oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`1E0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`1R0?ooo`00L@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo05D0
oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0640
oooo001a0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0E@3oool3000006<0oooo
00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1Q0?ooo`00L@3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo0640oooo001a0?ooo`030000003oool0oooo00/0oooo00<00000
0?ooo`3oool0E@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0H@3oool00740oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1E0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1Q0?oo
o`00L@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0640oooo001a0?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`1S0?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool0H@3oool00740oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`1Q0?ooo`00L@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo05@0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo0640oooo001a0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0A`3oool0103o
oml0W4P00018W03Oool90?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`1P0?ooo`00L@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo
04L0oooo00@0oooO09a80000B9`0gooo2@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool0H03oool00740oooo00<000000?ooo`3oool0303oool00`00
0000oooo0?ooo`170?ooo`040?oog`2LB00004RL0=ooo`T0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0600oooo001a0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool0@`3oool00`2oM0000000000000050000000307Boo`3oool0oooo
00H0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
0600oooo001a0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0=@3oool00`3ooml0
W4P0000000070000000504RLg`3oool0oooo0?oOW01dW=l00P3oool0103ooml0W4P00018W03Oool9
0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1P
0?ooo`00L@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo04<0oooo00L0gia807Bo
o`3oool0oooO09a80000B9`0gooo00T0oooo1000001R0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool0H03oool00740oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`13
0?ooo`070?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`090?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1P0?ooo`00L@3oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo04@0oooo00H0omnL04Qd_`3ooml0W4P00018W03Oool90?ooo`03
0000003oool0oooo06<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1P0?ooo`00
L@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo04D0oooo00D0gia807Bog`2LB000
04RL0=ooo`090?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`1O0?ooo`00L@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo04D0oooo
00D0oooO09a8B018B00004RL0=ooo`090?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`1O0?ooo`00L@3oool00`000000oooo0?ooo`0=0?ooo`030000
003oool0oooo04H0oooo00@0omnL04P00000B9`0gooo2@3oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool0G`3oool00740oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`170?ooo`030=nLB000B9`0gooo00T0oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo05l0oooo001a0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`1S0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool0G`3oool00740oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`1B0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`1O0?ooo`00L@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
0580oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
05h0oooo001a0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0DP3oool00`000000
oooo0?ooo`1S0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0GP3oool00740oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1N0?ooo`00L@3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo05h0oooo001a0?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool0D@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool0G@3oool00740oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1A0?oo
o`<00000H`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo05d0oooo001a0?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool0oooo0100oooo00<000000?ooo`3oool0G@3oool00740oooo00<000000?ooo`3oool0
3`3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`1M0?ooo`00L@3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo0540oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0A0?ooo`030000003o
ool0oooo05`0oooo001a0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0D@3oool0
0`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0G03oool0
0740oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`1L0?ooo`00L@3oool00`000000
oooo0?ooo`0@0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0H`3oool00`000000
oooo0?ooo`0A0?ooo`030000003oool0oooo05`0oooo001a0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool0F`3oool00740oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`1@0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`1K0?ooo`00L@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0500oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo05/0oooo001a0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`1S0?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool0F`3oool00740oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3o
ool04P3oool00`000000oooo0?ooo`1K0?ooo`00L@3oool00`000000oooo0?ooo`0A0?ooo`030000
003oool0oooo04l0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0C0?ooo`030000
003oool0oooo05X0oooo001a0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0C`3o
ool3000006<0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`1J0?ooo`00L@3oool0
0`000000oooo0?ooo`0B0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo05X0oooo001a0?ooo`030000003oool0oooo
0180oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
01<0oooo00<000000?ooo`3oool0FP3oool00740oooo00<000000?ooo`3oool04P3oool00`000000
oooo0?ooo`1>0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool04`3oool00`000000
oooo0?ooo`1J0?ooo`00L@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo04d0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo05T0oooo
001a0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?oo
o`1S0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0F@3oool00740oooo00<00000
0?ooo`3oool04`3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool0503oool00`000000oooo0?ooo`1I0?ooo`00L@3oool00`000000oooo0?ooo`0D0?oo
o`030000003oool0oooo04`0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0D0?oo
o`030000003oool0oooo05T0oooo001a0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3o
ool0C03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3o
ool0F03oool00740oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`1<0?ooo`030000
003oool0oooo06<0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`1H0?ooo`00L@3o
ool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0H`3o
ool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo05P0oooo001a0?ooo`030000003oool0
oooo01D0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool0E`3oool00740oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`1;0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool05P3oool00`00
0000oooo0?ooo`1G0?ooo`00L@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo04X0
oooo0`00001S0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0E`3oool00740oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`1F0?ooo`00L@3oool00`000000oooo0?oo
o`0F0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0oooo05H0oooo001a0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool0B@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo01P0oooo00<00000
0?ooo`3oool0E@3oool00740oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`190?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1E0?oo
o`00L@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo05D0oooo001a0?ooo`030000
003oool0oooo01P0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`1S0?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool0E03oool00740oooo00<000000?ooo`3oool0603o
ool00`000000oooo0?ooo`180?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool06@3o
ool00`000000oooo0?ooo`1D0?ooo`00L@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0
oooo03L0oooo00<0_g@0000000000000100000000`18W=l0oooo0?ooo`070?ooo`030000003oool0
oooo06<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`1D0?ooo`00L@3oool00`00
0000oooo0?ooo`0I0?ooo`030000003oool0oooo03L0oooo00<0gia80000000000000`0000000`00
B9`0gooo0?ooo`070?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`1C0?ooo`00L@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo03P0
oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`070?ooo`030000003oool0oooo06<0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`1C0?ooo`00L@3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo03P0oooo00<0gia807Boo`3oool02`3oool00`000000oooo
0?ooo`1S0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0DP3oool00740oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`0X0?ooo`030?oog`2LB000000000L0000000<0
B9cO0?ooo`3oool0103oool00`2oM4P0W=oo0?ooo`0:0?ooo`030000003oool0oooo06<0oooo00<0
00000?ooo`3oool06`3oool00`000000oooo0?ooo`1B0?ooo`00L@3oool00`000000oooo0?ooo`0K
0?ooo`030000003oool0oooo03P0oooo00<0oooO09a8M02oool02P3oool400000680oooo00<00000
0?ooo`3oool0703oool00`000000oooo0?ooo`1A0?ooo`00L@3oool00`000000oooo0?ooo`0K0?oo
o`030000003oool0oooo03T0oooo00<0okmd001d_`3oool02@3oool00`000000oooo0?ooo`1S0?oo
o`030000003oool0oooo01`0oooo00<000000?ooo`3oool0D@3oool00740oooo00<000000?ooo`3o
ool0703oool00`000000oooo0?ooo`0h0?ooo`030?oo_`1d04P0W=oo00T0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0500oooo001a0?ooo`030000
003oool0oooo01`0oooo00<000000?ooo`3oool0>03oool00`3ookl0M00004RLg`090?ooo`030000
003oool0oooo06<0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`1@0?ooo`00L@3o
ool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0380oooo00<0oknL09cOo`3oool00P3o
ool00`3ogi`0B00004RLg`090?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool07P3o
ool00`000000oooo0?ooo`1?0?ooo`00L@3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0
oooo0380oooo00P0oooO09a8B02Lgol0oooo0?oo_`1d0000001d0;ooo`T0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo04h0oooo001a0?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool0<P3oool01P3ooml0W4P0000000000000001d0;oo
o`X0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo
04h0oooo001a0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0@P3oool00`000000
oooo0?ooo`1S0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0C@3oool00740oooo
00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool0803oool00`000000oooo0?ooo`1=0?ooo`00L@3oool00`000000oooo0?oo
o`0O0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?oo
o`0Q0?ooo`030000003oool0oooo04`0oooo001a0?ooo`030000003oool0oooo0200oooo00<00000
0?ooo`3oool0@03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0240oooo00<00000
0?ooo`3oool0C03oool00740oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0o0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`1;0?oo
o`00L@3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo04X0oooo001a0?ooo`030000
003oool0oooo0280oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`1S0?ooo`030000
003oool0oooo02@0oooo00<000000?ooo`3oool0B@3oool00740oooo00<000000?ooo`3oool08`3o
ool00`000000oooo0?ooo`0m0?ooo`<00000H`3oool00`000000oooo0?ooo`0T0?ooo`030000003o
ool0oooo04T0oooo001a0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0?03oool0
0`000000oooo0?ooo`1S0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0B03oool0
0740oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`170?ooo`00L@3oool00`000000
oooo0?ooo`0V0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0H`3oool00`000000
oooo0?ooo`0W0?ooo`030000003oool0oooo04H0oooo001a0?ooo`030000003oool0oooo02H0oooo
00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo02P0oooo
00<000000?ooo`3oool0A@3oool00740oooo00<000000?ooo`3oool09`3oool00`000000oooo0?oo
o`0i0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?oo
o`150?ooo`00L@3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo03P0oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo04@0oooo001a0?oo
o`030000003oool0oooo02T0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`1S0?oo
o`030000003oool0oooo02X0oooo00<000000?ooo`3oool0@`3oool00740oooo00<000000?ooo`3o
ool0:P3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3o
ool0:`3oool00`000000oooo0?ooo`120?ooo`00L@3oool00`000000oooo0?ooo`0[0?ooo`030000
003oool0oooo03D0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0/0?ooo`030000
003oool0oooo0440oooo001a0?ooo`030000003oool0oooo02`0oooo0P00000e0?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`100?ooo`00L@3oool0
0`000000oooo0?ooo`0^0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`0^0?ooo`800000@03oool00740oooo00<000000?ooo`3oool0;`3oool00`00
0000oooo0?ooo`0a0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0<03oool00`00
0000oooo0?ooo`0m0?ooo`00L@3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo0300
oooo0`00001S0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0?03oool00740oooo
00<000000?ooo`3oool0<@3oool200000300oooo00<000000?ooo`3oool0H`3oool00`000000oooo
0?ooo`0b0?ooo`030000003oool0oooo03/0oooo001a0?ooo`030000003oool0oooo03<0oooo00<0
00000?ooo`3oool0;@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo03<0oooo00<0
00000?ooo`3oool0>P3oool00740oooo00<000000?ooo`3oool0=03oool2000002d0oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`0d0?ooo`800000>P3oool001D0oooo00<0oono07@0
000000000P0000000`1d_ol0oooo0?ooo`0i0?ooo`030;md00000000000000@0000000<0B9cO0?oo
o`3oool04@3oool00`000000oooo0?ooo`0U0?ooo`060;md00000000000000000000B9`0gooo2`3o
ool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo03`0oooo00H0_g@00000000000000000
0018W03OoolQ0?ooo`030000003oool0oooo01D0oooo00<0_g@0000000000000100000000`18W=l0
oooo0?ooo`0G0?ooo`030000003oool0oooo01d0oooo00<0oono07@0000000000P0000000`1d_ol0
oooo0?ooo`0B0?ooo`005@3oool00`3ogi`0B00004RLg`020?ooo`030;md0018W=l0oooo03X0oooo
00<0gia80000000000000`0000000`00B9`0gooo0?ooo`0A0?ooo`030000003oool0oooo02H0oooo
00<0omnL04P0001d_ol03P3oool2000002X0oooo00<000000?ooo`3oool0?@3oool00`3ogi`0B000
07Boo`0S0?ooo`030000003oool0oooo01H0oooo00<0gia80000000000000`0000000`00B9`0gooo
0?ooo`0H0?ooo`8000007@3oool00`3ogi`0B00004RLg`020?ooo`030;md0018W=l0oooo0180oooo
000J0?ooo`030?oo_`1d04P0W=oo03/0oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?oo
o`0A0?ooo`030000003oool0oooo02H0oooo00<0omnL04P0001d_ol0403oool2000002P0oooo00<0
00000?ooo`3oool0?@3oool00`3ogi`0B00007Boo`0S0?ooo`030000003oool0oooo01L0oooo00<0
gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`0J0?ooo`030000003oool0oooo01l0oooo00<0
oono07@0B02Lgol04P3oool001X0oooo00<0oono07@0001d_ol0?03oool00`3OW4P0M;oo0?ooo`0E
0?ooo`030000003oool0oooo02H0oooo00<0omnL04P0001d_ol04P3oool00`000000oooo0?ooo`0U
0?ooo`030000003oool0oooo03d0oooo00<0omnL04P0001d_ol08`3oool00`000000oooo0?ooo`0H
0?ooo`030=nLB01d_ol0oooo01l0oooo0P00000O0?ooo`030?oo_`1d0000M;oo0180oooo00090?oo
o`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool0103oool00`3o_g@0000007Boo`0]0?oo
o`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool00`3oool00`2oM4P0W=oo0?ooo`0D0?oo
o`030000003oool0oooo01L0oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`020?oo
o`030?oOW0180000M;oo01<0oooo0P00000U0?ooo`030000003oool0oooo03d0oooo00<0omnL04P0
001d_ol08`3oool00`000000oooo0?ooo`0I0?ooo`030;mdB02Lgol0oooo0200oooo0P00000M0?oo
o`030?noM0000000M;oo0180oooo000I0?ooo`040?oOW0180000001809cOocd0oooo00<0oooO09a8
M02oool0503oool00`000000oooo0?ooo`0V0?ooo`030?oOW0180000M;oo01D0oooo00<000000?oo
o`3oool08P3oool00`000000oooo0?ooo`0m0?ooo`030?oOW0180000M;oo02<0oooo00<000000?oo
o`3oool06@3oool00`3ooml0W4Qd0;ooo`0R0?ooo`030000003oool0oooo01T0oooo00@0omnL04P0
000004P0W=oo4P3oool001L0oooo00D0okmd000000000000001809cOo`0o0?ooo`030?noM000M;l0
oooo01<0oooo00<000000?ooo`3oool09P3oool00`3ogi`0B00007Boo`0F0?ooo`8000008P3oool0
0`000000oooo0?ooo`0m0?ooo`030?oOW0180000M;oo02<0oooo00<000000?ooo`3oool06P3oool0
0`3o_g@007Bo0?ooo`0R0?ooo`8000005`3oool01@3o_g@000000000000004P0W=oo01<0oooo000I
0?ooo`030;md0000M;l0oooo03l0oooo00<0oono07@0B02Lgol04`3oool00`000000oooo0?ooo`0V
0?ooo`030?oOW0180000M;oo01P0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0m
0?ooo`030?oOW0180000M;oo02<0oooo00<000000?ooo`3oool06P3oool00`3ookl0M01809cOo`0T
0?ooo`030000003oool0oooo01H0oooo00<0_g@0001d_`3oool04`3oool001X0oooo00<0_g@004RL
g`3oool0?P3oool00`3ookl0M00004RLg`0C0?ooo`030000003oool0oooo02H0oooo00<0omnL04P0
001d_ol06@3oool2000001l0oooo00<000000?ooo`3oool0?@3oool00`3ogi`0B00007Boo`0S0?oo
o`030000003oool0oooo01X0oooo00<0oono07@00018W=l09@3oool2000001L0oooo00<0_g@004RL
g`3oool04P3oool001D0oooo00<0oooO09bL_`3oool00P3oool00`3OW4P0001d0;ooo`0i0?ooo`03
0?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l04`3oool00`000000oooo0?ooo`0V0?ooo`03
0?oOW0180000M;oo01/0oooo0P00000M0?ooo`<00000?@3oool00`3ogi`0B00007Boo`0S0?ooo`03
0000003oool0oooo01D0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`0W0?ooo`80
0000403oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo0180oooo000F0?ooo`060?no
M018W=l0oooo0?oo_`1d000007Bo>P3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo
4`3oool00`000000oooo0?ooo`0T0?ooo`050?oog`2LB00000000000001d_ol07@3oool2000001/0
oooo00<000000?ooo`3oool0>`3oool01@3ooml0W4P0000000000000M;oo02<0oooo00<000000?oo
o`3oool05@3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo:@3oool2000000l0oooo
00H0okmd04RLg`3oool0oono07@00000M;lC0?ooo`005`3oool01@3OW4P000000000000007@0_ooo
03/0oooo00H0oooO09a80000000000000000M02ooolD0?ooo`030000003oool0oooo02H0oooo00<0
okmd0000001d_ol07`3oool2000001T0oooo00<000000?ooo`3oool0?@3oool00`3o_g@0000007Bo
o`0S0?ooo`030000003oool0oooo01H0oooo00H0oooO09a80000000000000000M02oool/0?ooo`<0
00003@3oool01@3OW4P000000000000007@0_ooo01<0oooo001a0?ooo`030000003oool0oooo04X0
oooo0P00000G0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0B`3oool2000002<0
oooo001a0?ooo`030000003oool0oooo04`0oooo0P00000E0?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool0C@3oool200000240oooo001a0?ooo`030000003oool0oooo04h0oooo0P00
000C0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0C`3oool3000001h0oooo001a
0?ooo`030000003oool0oooo0500oooo0P00000A0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool0DP3oool2000001`0oooo001a0?ooo`030000003oool0oooo0580oooo0`00000>0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool0E03oool3000001T0oooo001a0?ooo`03
0000003oool0oooo05D0oooo0P00000<0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3o
ool0E`3oool2000001L0oooo001a0?ooo`030000003oool0oooo05L0oooo0`0000090?ooo`030000
003oool0oooo06<0oooo00<000000?ooo`3oool0F@3oool2000001D0oooo001a0?ooo`030000003o
ool0oooo05X0oooo0P0000070?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0F`3o
ool2000001<0oooo001a0?ooo`030000003oool0oooo05`0oooo0P0000050?ooo`030000003oool0
oooo06<0oooo00<000000?ooo`3oool0G@3oool200000140oooo001a0?ooo`030000003oool0oooo
05h0oooo0`0000020?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0G`3oool20000
00l0oooo001a0?ooo`030000003oool0oooo0640oooo0`00001U0?ooo`030000003oool0oooo0640
oooo0P00000=0?ooo`00o`00002`00000040oooo00060?ooo`030000003oool0oooo00<0oooo0`00
00040?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`0000002P3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00800000203oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0000000`0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`050?ooo`003`3oool3
000000030?ooo`000000oooo0400oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0R
0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool00P3oool3000003T0oooo00<00000
0?ooo`3oool08P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo03l0oooo00<00000
0?ooo`3oool04P3oool00180oooo0P00001M0?ooo`030000003oool0oooo06<0oooo00<000000?oo
o`3oool01@3oool2000005`0oooo00<000000?ooo`3oool0L03oool001@0oooo0`00001J0?ooo`03
0000003oool0oooo06<0oooo00<000000?ooo`3oool01`3oool2000005X0oooo00<000000?ooo`3o
ool0L03oool001L0oooo0`00001G0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0
2@3oool3000005L0oooo00<000000?ooo`3oool0L03oool001X0oooo0P00001E0?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`3oool0303oool2000005D0oooo00<000000?ooo`3oool0L03o
ool001`0oooo0P00001C0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool03P3oool3
00000580oooo00<000000?ooo`3oool0L03oool001h0oooo0P00001A0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool04@3oool200000500oooo00<000000?ooo`3oool0L03oool00200
oooo0P00001?0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool04`3oool2000004h0
oooo00<000000?ooo`3oool0L03oool00280oooo0P00001=0?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool05@3oool2000004`0oooo00<000000?ooo`3oool0L03oool002@0oooo0P00
001;0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool05`3oool2000004X0oooo00<0
00000?ooo`3oool0L03oool002H0oooo0P0000190?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool06@3oool2000004P0oooo00<000000?ooo`3oool0L03oool002P0oooo0P0000170?oo
o`030000003oool0oooo06<0oooo0`00000K0?ooo`800000AP3oool00`000000oooo0?ooo`1`0?oo
o`00:P3oool2000004D0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0M0?ooo`80
0000A03oool00`000000oooo0?ooo`1`0?ooo`00;03oool2000004<0oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`0O0?ooo`800000@P3oool00`000000oooo0?ooo`1`0?ooo`00;P3o
ool200000440oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0Q0?ooo`800000@03o
ool00`000000oooo0?ooo`1`0?ooo`00<03oool00`000000oooo0?ooo`0n0?ooo`030000003oool0
oooo06<0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0
oooo0700oooo000a0?ooo`800000?P3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
02@0oooo0P00000m0?ooo`030000003oool0oooo0700oooo000c0?ooo`030000003oool0oooo03/0
oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo03X0
oooo00<000000?ooo`3oool0L03oool003@0oooo0P00000k0?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool09`3oool2000003X0oooo00<000000?ooo`3oool0L03oool003H0oooo00<0
00000?ooo`3oool0>03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo02T0oooo00<0
00000?ooo`3oool0=`3oool00`000000oooo0?ooo`1`0?ooo`00=`3oool00`000000oooo0?ooo`0g
0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0f
0?ooo`030000003oool0oooo0700oooo000h0?ooo`800000=`3oool00`000000oooo0?ooo`1S0?oo
o`030000003oool0oooo02/0oooo0P00000f0?ooo`030000003oool0oooo0700oooo000j0?ooo`03
0000003oool0oooo03@0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0]0?ooo`03
0000003oool0oooo03<0oooo00<000000?ooo`3oool0L03oool003/0oooo00<000000?ooo`3oool0
<`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0
<P3oool00`000000oooo0?ooo`1`0?ooo`00?03oool2000003<0oooo00<000000?ooo`3oool0H`3o
ool3000002l0oooo0P00000b0?ooo`030000003oool0oooo0700oooo000n0?ooo`030000003oool0
oooo0300oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0
oooo02l0oooo00<000000?ooo`3oool0L03oool003l0oooo00<000000?ooo`3oool0;`3oool00`00
0000oooo0?ooo`1S0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0;P3oool00`00
0000oooo0?ooo`1`0?ooo`00@03oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo06<0
oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo0700
oooo00110?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0H`3oool00`000000oooo
0?ooo`0d0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0L03oool00480oooo0P00
000]0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0=@3oool2000002`0oooo00<0
00000?ooo`3oool0L03oool004@0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`1S
0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`1`
0?ooo`00A@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo06<0oooo00<000000?oo
o`3oool0>03oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo0700oooo00160?ooo`03
0000003oool0oooo02P0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0i0?ooo`03
0000003oool0oooo02L0oooo00<000000?ooo`3oool0L03oool004L0oooo00<000000?ooo`3oool0
9`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0
9P3oool00`000000oooo0?ooo`1`0?ooo`00A`3oool00`000000oooo0?ooo`0W0?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0V0?ooo`030000003o
ool0oooo0700oooo00180?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`0k0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0L03oool0
04T0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
03`0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`1`0?ooo`00BP3oool00`000000
oooo0?ooo`0T0?ooo`030000003oool0oooo06<0oooo0`00000l0?ooo`030000003oool0oooo02@0
oooo00<000000?ooo`3oool0L03oool004X0oooo00<000000?ooo`3oool0903oool00`000000oooo
0?ooo`1S0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool08`3oool00`000000oooo
0?ooo`1`0?ooo`00B`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo06<0oooo00<0
00000?ooo`3oool0?P3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo0700oooo001<
0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0n
0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0L03oool004d0oooo00<000000?oo
o`3oool08@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo03l0oooo00<000000?oo
o`3oool08@3oool00`000000oooo0?ooo`1`0?ooo`00C@3oool00`000000oooo0?ooo`0Q0?ooo`03
0000003oool0oooo06<0oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`0P0?ooo`03
0000003oool0oooo0700oooo001>0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0
L03oool004l0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`1S0?ooo`030000003o
ool0oooo0440oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`1`0?ooo`00C`3oool0
0`000000oooo0?ooo`0O0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0@P3oool0
0`000000oooo0?ooo`0N0?ooo`030000003oool0oooo0700oooo001@0?ooo`030000003oool0oooo
01h0oooo00<000000?ooo`3oool0DP3oool00`2oM0000000000000040000000304RLg`3oool0oooo
00L0oooo00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo
0700oooo001@0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0D`3oool00`3OW4P0
000000000003000000030018W03Oool0oooo00L0oooo00<000000?ooo`3oool0@`3oool00`000000
oooo0?ooo`0M0?ooo`030000003oool0oooo0700oooo001A0?ooo`030000003oool0oooo01d0oooo
00<000000?ooo`3oool0E03oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo00L0oooo
00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0700oooo
001B0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0E@3oool00`3OW4P0M;oo0?oo
o`0;0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0703oool00`000000oooo0?oo
o`1`0?ooo`00DP3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo05H0oooo00<0_gA8
09cOo`3oool02P3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo01`0oooo00<00000
0?ooo`3oool0L03oool005<0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`1F0?oo
o`030?oog`2LB7@0_ooo00X0oooo100000140?ooo`030000003oool0oooo01/0oooo00<000000?oo
o`3oool0L03oool005<0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`1G0?ooo`03
0?noM000M;l0oooo00T0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0K0?ooo`03
0000003oool0oooo0700oooo001D0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
E`3oool00`3ookl0M01809cOo`090?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`1`0?ooo`00E03oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo05L0oooo00<0oono07@00018W=l02@3oool00`000000oooo0?ooo`160?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool0L03oool005@0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`1B0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l02@3oool0
0`000000oooo0?ooo`170?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0L03oool0
05D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`1B0?ooo`080?oog`2LB4P0W=oo
0?ooo`3ookl0M0000000M02oool90?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`1`0?ooo`00E@3oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo05<0oooo00H0oooO09a80000000000000000M02oool:0?ooo`030000003oool0oooo04P0
oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1`0?ooo`00E@3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0B03oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo0700oooo001F0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool0H`3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool0L03oool005H0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1S
0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`1`
0?ooo`00EP3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo06<0oooo00<000000?oo
o`3oool0BP3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0700oooo001G0?ooo`03
0000003oool0oooo01L0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1:0?ooo`03
0000003oool0oooo01H0oooo00<000000?ooo`3oool0L03oool005L0oooo00<000000?ooo`3oool0
5`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0
5P3oool00`000000oooo0?ooo`1`0?ooo`00E`3oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo06<0oooo0`00001;0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0L03o
ool005P0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
oooo04/0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`1`0?ooo`00F03oool00`00
0000oooo0?ooo`0F0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0B`3oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo0700oooo001I0?ooo`030000003oool0oooo01D0
oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo01D0
oooo00<000000?ooo`3oool0L03oool005T0oooo00<000000?ooo`3oool05@3oool00`000000oooo
0?ooo`1S0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0503oool00`000000oooo
0?ooo`1`0?ooo`00FP3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo06<0oooo00<0
00000?ooo`3oool0C03oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0700oooo001J
0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1<
0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0L03oool005X0oooo00<000000?oo
o`3oool0503oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo04d0oooo00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`1`0?ooo`00FP3oool00`000000oooo0?ooo`0D0?ooo`03
0000003oool0oooo06<0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`0C0?ooo`03
0000003oool0oooo0700oooo001K0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0
L03oool005/0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`1S0?ooo`030000003o
ool0oooo04h0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`1`0?ooo`00F`3oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0CP3oool0
0`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0700oooo001K0?ooo`030000003oool0oooo
01<0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo
0180oooo00<000000?ooo`3oool0L03oool005`0oooo00<000000?ooo`3oool04P3oool00`000000
oooo0?ooo`1S0?ooo`<00000C`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0700
oooo001L0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0H`3oool00`000000oooo
0?ooo`1?0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0L03oool005`0oooo00<0
00000?ooo`3oool04P3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo04l0oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`1`0?ooo`00G03oool00`000000oooo0?ooo`0B
0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo0700oooo001L0?ooo`030000003oool0oooo0180oooo00<000000?oo
o`3oool0H`3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0L03oool005d0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool0oooo0500oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1`0?ooo`00
G@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0
D03oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0700oooo001M0?ooo`030000003o
ool0oooo0140oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1@0?ooo`030000003o
ool0oooo0100oooo00<000000?ooo`3oool0L03oool005d0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`1`0?ooo`00GP3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
0700oooo001N0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0H`3oool00`000000
oooo0?ooo`1A0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0L03oool005h0oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0540oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1`0?ooo`00GP3oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo0700oooo001N0?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool0L03oool005h0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1S0?oo
o`<00000D@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0700oooo001O0?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1B0?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool0L03oool005l0oooo00<000000?ooo`3oool0
3`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`1`0?ooo`00G`3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo0700oooo001O0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`1B0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0L03oool0
05l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
0580oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1`0?ooo`00G`3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0DP3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo0700oooo001P0?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool0L03oool00600oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`1F0?ooo`040?oog`2LB00004RL0=ooo`T0oooo00<000000?ooo`3oool0D`3oool00`000000oooo
0?ooo`0=0?ooo`030000003oool0oooo0700oooo001P0?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool0EP3oool0103ooml0W4P00018W03Oool90?ooo`030000003oool0oooo05<0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1`0?ooo`00H03oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo05H0oooo00@0oooO09a80000B9`0gooo2@3oool00`000000oooo
0?ooo`1C0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0L03oool00600oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`1B0?ooo`030;md00000000000000D0000000<0
M;oo0?ooo`3oool01P3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool0L03oool00600oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1B
0?ooo`080?oOW01dW=l0oooo0?ooo`3ooml0W4P00018W03Oool90?ooo`030000003oool0oooo05<0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1`0?ooo`00H03oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo05<0oooo00L0gia807Boo`3oool0oooO09a80000B9`0gooo
00T0oooo1000001B0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0L03oool00600
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1C0?ooo`070?oog`2LB7@0_ooo0?oo
g`2LB00004RL0=ooo`090?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`1`0?ooo`00H@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
05@0oooo00H0omnL04Qd_`3ooml0W4P00018W03Oool90?ooo`030000003oool0oooo05@0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`1`0?ooo`00H@3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo05D0oooo00D0gia807Bog`2LB00004RL0=ooo`090?ooo`030000003o
ool0oooo05@0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1`0?ooo`00H@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo05D0oooo00D0oooO09a8B018B00004RL0=oo
o`090?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`1`0?ooo`00H@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo05H0oooo00@0omnL
04P00000B9`0gooo2@3oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool0L03oool00640oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1G
0?ooo`030=nLB000B9`0gooo00T0oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo0700oooo001Q0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool0H`3oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool0L03oool00640oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool0oooo05@0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1`0?ooo`00
H@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0
E03oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0700oooo001R0?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1D0?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool0L03oool00680oooo00<000000?ooo`3oool0303oool0
0`000000oooo0?ooo`1S0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`1`0?ooo`00HP3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
0700oooo001R0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0H`3oool3000005D0
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1`0?ooo`00HP3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0E@3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo0700oooo001R0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool0H`3oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool0L03oool00680oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1S
0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1`
0?ooo`00HP3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo06<0oooo00<000000?oo
o`3oool0E@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0700oooo001S0?ooo`03
0000003oool0oooo00/0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1E0?ooo`03
0000003oool0oooo00/0oooo00<000000?ooo`3oool0L03oool006<0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`1`0?ooo`00H`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo0700oooo001S0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`1F0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0L03oool0
06<0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
05H0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`1`0?ooo`00H`3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0EP3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo0700oooo001T0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool0L03oool006@0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`1S0?ooo`030000003oool0oooo05H0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`1`0?ooo`00I03oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool0EP3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0700oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-3.30486, -5.00005, \
0.0230037, 0.0564177}}]
}, Closed]],

Cell[TextData[{
  "Note that we can replace either of the range specifications by ",
  StyleBox["All", "InlineInput",
    FontWeight->"Bold"],
  ",",
  " which includes all points, or ",
  StyleBox["Automatic", "InlineInput",
    FontWeight->"Bold"],
  ", which drops outlying points.  For a two-dimensional plot we could have \
given the y-range only because the x-range was already supplied by the second \
argument to ",
  StyleBox["Plot", "InlineInput",
    FontWeight->"Bold"],
  ".  Similarly, for a three-dimensional plot a single range would be applied \
to the z-axis.  However, if more than one range is given, the all ranges (2 \
or 3) must be given."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(tanPlot\  = 
        Plot[Tan[x], \ {x, \(-\[Pi]\), \ \[Pi]}, 
          PlotRange \[Rule] {\(-5\), 5}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.151576 0.309017 0.0618034 [
[.04527 .29652 -9.15625 -15.6875 ]
[.04527 .29652 9.15625 0 ]
[.19685 .29652 -9.15625 -15.6875 ]
[.19685 .29652 9.15625 0 ]
[.34842 .29652 -9.15625 -15.6875 ]
[.34842 .29652 9.15625 0 ]
[.65158 .29652 -5 -15.6875 ]
[.65158 .29652 5 0 ]
[.80315 .29652 -5 -15.6875 ]
[.80315 .29652 5 0 ]
[.95473 .29652 -5 -15.6875 ]
[.95473 .29652 5 0 ]
[.4875 .0618 -18.3125 -7.84375 ]
[.4875 .0618 0 7.84375 ]
[.4875 .18541 -18.3125 -7.84375 ]
[.4875 .18541 0 7.84375 ]
[.4875 .43262 -10 -7.84375 ]
[.4875 .43262 0 7.84375 ]
[.4875 .55623 -10 -7.84375 ]
[.4875 .55623 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.04527 .30902 m
.04527 .31527 L
s
gsave
.04527 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.19685 .30902 m
.19685 .31527 L
s
gsave
.19685 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.34842 .30902 m
.34842 .31527 L
s
gsave
.34842 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.65158 .30902 m
.65158 .31527 L
s
gsave
.65158 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.80315 .30902 m
.80315 .31527 L
s
gsave
.80315 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.95473 .30902 m
.95473 .31527 L
s
gsave
.95473 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.07559 .30902 m
.07559 .31277 L
s
.1059 .30902 m
.1059 .31277 L
s
.13622 .30902 m
.13622 .31277 L
s
.16653 .30902 m
.16653 .31277 L
s
.22716 .30902 m
.22716 .31277 L
s
.25748 .30902 m
.25748 .31277 L
s
.28779 .30902 m
.28779 .31277 L
s
.31811 .30902 m
.31811 .31277 L
s
.37874 .30902 m
.37874 .31277 L
s
.40905 .30902 m
.40905 .31277 L
s
.43937 .30902 m
.43937 .31277 L
s
.46968 .30902 m
.46968 .31277 L
s
.53032 .30902 m
.53032 .31277 L
s
.56063 .30902 m
.56063 .31277 L
s
.59095 .30902 m
.59095 .31277 L
s
.62126 .30902 m
.62126 .31277 L
s
.68189 .30902 m
.68189 .31277 L
s
.71221 .30902 m
.71221 .31277 L
s
.74252 .30902 m
.74252 .31277 L
s
.77284 .30902 m
.77284 .31277 L
s
.83347 .30902 m
.83347 .31277 L
s
.86378 .30902 m
.86378 .31277 L
s
.8941 .30902 m
.8941 .31277 L
s
.92441 .30902 m
.92441 .31277 L
s
.01496 .30902 m
.01496 .31277 L
s
.98504 .30902 m
.98504 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5 .0618 m
.50625 .0618 L
s
gsave
.4875 .0618 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.5 .18541 m
.50625 .18541 L
s
gsave
.4875 .18541 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 .43262 m
.50625 .43262 L
s
gsave
.4875 .43262 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 .55623 m
.50625 .55623 L
s
gsave
.4875 .55623 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .09271 m
.50375 .09271 L
s
.5 .12361 m
.50375 .12361 L
s
.5 .15451 m
.50375 .15451 L
s
.5 .21631 m
.50375 .21631 L
s
.5 .24721 m
.50375 .24721 L
s
.5 .27812 m
.50375 .27812 L
s
.5 .33992 m
.50375 .33992 L
s
.5 .37082 m
.50375 .37082 L
s
.5 .40172 m
.50375 .40172 L
s
.5 .46353 m
.50375 .46353 L
s
.5 .49443 m
.50375 .49443 L
s
.5 .52533 m
.50375 .52533 L
s
.5 .0309 m
.50375 .0309 L
s
.5 .58713 m
.50375 .58713 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .30902 m
.06244 .32512 L
.10458 .34547 L
.12507 .35778 L
.14415 .37189 L
.16254 .38939 L
.18221 .41552 L
.19242 .43427 L
.19781 .44635 L
.20351 .46143 L
.2085 .4771 L
.21399 .49795 L
.21861 .51947 L
.22365 .5487 L
.22885 .58791 L
s
.22885 .58791 m
.2318 .61803 L
s
.26219 .61803 m
.26222 0 L
s
.29185 0 m
.29209 .00278 L
.29481 .02877 L
.30016 .06936 L
.30508 .09795 L
.31017 .12155 L
.31568 .14217 L
.32564 .17081 L
.3347 .19038 L
.34426 .20669 L
.36394 .23151 L
.38229 .24829 L
.42125 .27367 L
.46266 .29348 L
.50256 .31006 L
.54094 .32613 L
.58177 .34602 L
.60207 .35832 L
.62109 .37251 L
.6393 .39001 L
.6589 .41633 L
.66849 .43401 L
.67898 .45936 L
.68427 .47568 L
.68992 .49688 L
.69486 .51979 L
.70009 .55031 L
.70543 .59139 L
s
.70543 .59139 m
.70798 .61803 L
s
.73721 .61803 m
.73722 0 L
s
.76807 0 m
.7695 .01503 L
.77245 .041 L
.77728 .07527 L
.78252 .10424 L
.78777 .12724 L
.79274 .14507 L
.80205 .17133 L
.81292 .19415 L
.82297 .21043 L
.84165 .23308 L
.86225 .25125 L
.90248 .27634 L
.94119 .29448 L
.97619 .30902 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oL@3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo06<0oooo001a0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool0H`3oool00740oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`1F0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`1S0?ooo`00L@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo05H0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo06<0oooo
001a0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?oo
o`1S0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0HP3oool00740oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`1R0?ooo`00L@3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo05H0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo0680oooo001a0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool0E@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool0HP3oool00740oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1E0?ooo`030000
003oool0oooo06<0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1R0?ooo`00L@3o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0H`3o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0680oooo001a0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool0HP3oool00740oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`1E0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`1R0?ooo`00L@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo05D0
oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0640
oooo001a0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0E@3oool3000006<0oooo
00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1Q0?ooo`00L@3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo0640oooo001a0?ooo`030000003oool0oooo00/0oooo00<00000
0?ooo`3oool0E@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0H@3oool00740oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1E0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1Q0?oo
o`00L@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0640oooo001a0?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`1S0?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool0H@3oool00740oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`1Q0?ooo`00L@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo05@0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo0640oooo001a0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0A`3oool0103o
oml0W4P00018W03Oool90?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`1P0?ooo`00L@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo
04L0oooo00@0oooO09a80000B9`0gooo2@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool0H03oool00740oooo00<000000?ooo`3oool0303oool00`00
0000oooo0?ooo`170?ooo`040?oog`2LB00004RL0=ooo`T0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0600oooo001a0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool0@`3oool00`2oM0000000000000050000000307Boo`3oool0oooo
00H0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
0600oooo001a0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0=@3oool00`3ooml0
W4P0000000070000000504RLg`3oool0oooo0?oOW01dW=l00P3oool0103ooml0W4P00018W03Oool9
0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1P
0?ooo`00L@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo04<0oooo00L0gia807Bo
o`3oool0oooO09a80000B9`0gooo00T0oooo1000001R0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool0H03oool00740oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`13
0?ooo`070?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`090?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1P0?ooo`00L@3oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo04@0oooo00H0omnL04Qd_`3ooml0W4P00018W03Oool90?ooo`03
0000003oool0oooo06<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1P0?ooo`00
L@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo04D0oooo00D0gia807Bog`2LB000
04RL0=ooo`090?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`1O0?ooo`00L@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo04D0oooo
00D0oooO09a8B018B00004RL0=ooo`090?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`1O0?ooo`00L@3oool00`000000oooo0?ooo`0=0?ooo`030000
003oool0oooo04H0oooo00@0omnL04P00000B9`0gooo2@3oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool0G`3oool00740oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`170?ooo`030=nLB000B9`0gooo00T0oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo05l0oooo001a0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`1S0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool0G`3oool00740oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`1B0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`1O0?ooo`00L@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
0580oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
05h0oooo001a0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0DP3oool00`000000
oooo0?ooo`1S0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0GP3oool00740oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1N0?ooo`00L@3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo05h0oooo001a0?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool0D@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool0G@3oool00740oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1A0?oo
o`<00000H`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo05d0oooo001a0?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool0oooo0100oooo00<000000?ooo`3oool0G@3oool00740oooo00<000000?ooo`3oool0
3`3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`1M0?ooo`00L@3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo0540oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0A0?ooo`030000003o
ool0oooo05`0oooo001a0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0D@3oool0
0`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0G03oool0
0740oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`1L0?ooo`00L@3oool00`000000
oooo0?ooo`0@0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0H`3oool00`000000
oooo0?ooo`0A0?ooo`030000003oool0oooo05`0oooo001a0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool0F`3oool00740oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`1@0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`1K0?ooo`00L@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0500oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo05/0oooo001a0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`1S0?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool0F`3oool00740oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3o
ool04P3oool00`000000oooo0?ooo`1K0?ooo`00L@3oool00`000000oooo0?ooo`0A0?ooo`030000
003oool0oooo04l0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0C0?ooo`030000
003oool0oooo05X0oooo001a0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0C`3o
ool3000006<0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`1J0?ooo`00L@3oool0
0`000000oooo0?ooo`0B0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo05X0oooo001a0?ooo`030000003oool0oooo
0180oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
01<0oooo00<000000?ooo`3oool0FP3oool00740oooo00<000000?ooo`3oool04P3oool00`000000
oooo0?ooo`1>0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool04`3oool00`000000
oooo0?ooo`1J0?ooo`00L@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo04d0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo05T0oooo
001a0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?oo
o`1S0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0F@3oool00740oooo00<00000
0?ooo`3oool04`3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool0503oool00`000000oooo0?ooo`1I0?ooo`00L@3oool00`000000oooo0?ooo`0D0?oo
o`030000003oool0oooo04`0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0D0?oo
o`030000003oool0oooo05T0oooo001a0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3o
ool0C03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3o
ool0F03oool00740oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`1<0?ooo`030000
003oool0oooo06<0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`1H0?ooo`00L@3o
ool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0H`3o
ool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo05P0oooo001a0?ooo`030000003oool0
oooo01D0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool0E`3oool00740oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`1;0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool05P3oool00`00
0000oooo0?ooo`1G0?ooo`00L@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo04X0
oooo0`00001S0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0E`3oool00740oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`1F0?ooo`00L@3oool00`000000oooo0?oo
o`0F0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0oooo05H0oooo001a0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool0B@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo01P0oooo00<00000
0?ooo`3oool0E@3oool00740oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`190?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1E0?oo
o`00L@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo05D0oooo001a0?ooo`030000
003oool0oooo01P0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`1S0?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool0E03oool00740oooo00<000000?ooo`3oool0603o
ool00`000000oooo0?ooo`180?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool06@3o
ool00`000000oooo0?ooo`1D0?ooo`00L@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0
oooo03L0oooo00<0_g@0000000000000100000000`18W=l0oooo0?ooo`070?ooo`030000003oool0
oooo06<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`1D0?ooo`00L@3oool00`00
0000oooo0?ooo`0I0?ooo`030000003oool0oooo03L0oooo00<0gia80000000000000`0000000`00
B9`0gooo0?ooo`070?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`1C0?ooo`00L@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo03P0
oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`070?ooo`030000003oool0oooo06<0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`1C0?ooo`00L@3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo03P0oooo00<0gia807Boo`3oool02`3oool00`000000oooo
0?ooo`1S0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0DP3oool00740oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`0X0?ooo`030?oog`2LB000000000L0000000<0
B9cO0?ooo`3oool0103oool00`2oM4P0W=oo0?ooo`0:0?ooo`030000003oool0oooo06<0oooo00<0
00000?ooo`3oool06`3oool00`000000oooo0?ooo`1B0?ooo`00L@3oool00`000000oooo0?ooo`0K
0?ooo`030000003oool0oooo03P0oooo00<0oooO09a8M02oool02P3oool400000680oooo00<00000
0?ooo`3oool0703oool00`000000oooo0?ooo`1A0?ooo`00L@3oool00`000000oooo0?ooo`0K0?oo
o`030000003oool0oooo03T0oooo00<0okmd001d_`3oool02@3oool00`000000oooo0?ooo`1S0?oo
o`030000003oool0oooo01`0oooo00<000000?ooo`3oool0D@3oool00740oooo00<000000?ooo`3o
ool0703oool00`000000oooo0?ooo`0h0?ooo`030?oo_`1d04P0W=oo00T0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0500oooo001a0?ooo`030000
003oool0oooo01`0oooo00<000000?ooo`3oool0>03oool00`3ookl0M00004RLg`090?ooo`030000
003oool0oooo06<0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`1@0?ooo`00L@3o
ool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0380oooo00<0oknL09cOo`3oool00P3o
ool00`3ogi`0B00004RLg`090?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool07P3o
ool00`000000oooo0?ooo`1?0?ooo`00L@3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0
oooo0380oooo00P0oooO09a8B02Lgol0oooo0?oo_`1d0000001d0;ooo`T0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo04h0oooo001a0?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool0<P3oool01P3ooml0W4P0000000000000001d0;oo
o`X0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo
04h0oooo001a0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0@P3oool00`000000
oooo0?ooo`1S0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0C@3oool00740oooo
00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool0803oool00`000000oooo0?ooo`1=0?ooo`00L@3oool00`000000oooo0?oo
o`0O0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?oo
o`0Q0?ooo`030000003oool0oooo04`0oooo001a0?ooo`030000003oool0oooo0200oooo00<00000
0?ooo`3oool0@03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0240oooo00<00000
0?ooo`3oool0C03oool00740oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0o0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`1;0?oo
o`00L@3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo04X0oooo001a0?ooo`030000
003oool0oooo0280oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`1S0?ooo`030000
003oool0oooo02@0oooo00<000000?ooo`3oool0B@3oool00740oooo00<000000?ooo`3oool08`3o
ool00`000000oooo0?ooo`0m0?ooo`<00000H`3oool00`000000oooo0?ooo`0T0?ooo`030000003o
ool0oooo04T0oooo001a0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0?03oool0
0`000000oooo0?ooo`1S0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0B03oool0
0740oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`170?ooo`00L@3oool00`000000
oooo0?ooo`0V0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0H`3oool00`000000
oooo0?ooo`0W0?ooo`030000003oool0oooo04H0oooo001a0?ooo`030000003oool0oooo02H0oooo
00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo02P0oooo
00<000000?ooo`3oool0A@3oool00740oooo00<000000?ooo`3oool09`3oool00`000000oooo0?oo
o`0i0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?oo
o`150?ooo`00L@3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo03P0oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo04@0oooo001a0?oo
o`030000003oool0oooo02T0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`1S0?oo
o`030000003oool0oooo02X0oooo00<000000?ooo`3oool0@`3oool00740oooo00<000000?ooo`3o
ool0:P3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3o
ool0:`3oool00`000000oooo0?ooo`120?ooo`00L@3oool00`000000oooo0?ooo`0[0?ooo`030000
003oool0oooo03D0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0/0?ooo`030000
003oool0oooo0440oooo001a0?ooo`030000003oool0oooo02`0oooo0P00000e0?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`100?ooo`00L@3oool0
0`000000oooo0?ooo`0^0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`0^0?ooo`800000@03oool00740oooo00<000000?ooo`3oool0;`3oool00`00
0000oooo0?ooo`0a0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0<03oool00`00
0000oooo0?ooo`0m0?ooo`00L@3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo0300
oooo0`00001S0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0?03oool00740oooo
00<000000?ooo`3oool0<@3oool200000300oooo00<000000?ooo`3oool0H`3oool00`000000oooo
0?ooo`0b0?ooo`030000003oool0oooo03/0oooo001a0?ooo`030000003oool0oooo03<0oooo00<0
00000?ooo`3oool0;@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo03<0oooo00<0
00000?ooo`3oool0>P3oool00740oooo00<000000?ooo`3oool0=03oool2000002d0oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`0d0?ooo`800000>P3oool001D0oooo00<0oono07@0
000000000P0000000`1d_ol0oooo0?ooo`0i0?ooo`030;md00000000000000@0000000<0B9cO0?oo
o`3oool04@3oool00`000000oooo0?ooo`0U0?ooo`060;md00000000000000000000B9`0gooo2`3o
ool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo03`0oooo00H0_g@00000000000000000
0018W03OoolQ0?ooo`030000003oool0oooo01D0oooo00<0_g@0000000000000100000000`18W=l0
oooo0?ooo`0G0?ooo`030000003oool0oooo01d0oooo00<0oono07@0000000000P0000000`1d_ol0
oooo0?ooo`0B0?ooo`005@3oool00`3ogi`0B00004RLg`020?ooo`030;md0018W=l0oooo03X0oooo
00<0gia80000000000000`0000000`00B9`0gooo0?ooo`0A0?ooo`030000003oool0oooo02H0oooo
00<0omnL04P0001d_ol03P3oool2000002X0oooo00<000000?ooo`3oool0?@3oool00`3ogi`0B000
07Boo`0S0?ooo`030000003oool0oooo01H0oooo00<0gia80000000000000`0000000`00B9`0gooo
0?ooo`0H0?ooo`8000007@3oool00`3ogi`0B00004RLg`020?ooo`030;md0018W=l0oooo0180oooo
000J0?ooo`030?oo_`1d04P0W=oo03/0oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?oo
o`0A0?ooo`030000003oool0oooo02H0oooo00<0omnL04P0001d_ol0403oool2000002P0oooo00<0
00000?ooo`3oool0?@3oool00`3ogi`0B00007Boo`0S0?ooo`030000003oool0oooo01L0oooo00<0
gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`0J0?ooo`030000003oool0oooo01l0oooo00<0
oono07@0B02Lgol04P3oool001X0oooo00<0oono07@0001d_ol0?03oool00`3OW4P0M;oo0?ooo`0E
0?ooo`030000003oool0oooo02H0oooo00<0omnL04P0001d_ol04P3oool00`000000oooo0?ooo`0U
0?ooo`030000003oool0oooo03d0oooo00<0omnL04P0001d_ol08`3oool00`000000oooo0?ooo`0H
0?ooo`030=nLB01d_ol0oooo01l0oooo0P00000O0?ooo`030?oo_`1d0000M;oo0180oooo00090?oo
o`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool0103oool00`3o_g@0000007Boo`0]0?oo
o`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool00`3oool00`2oM4P0W=oo0?ooo`0D0?oo
o`030000003oool0oooo01L0oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`020?oo
o`030?oOW0180000M;oo01<0oooo0P00000U0?ooo`030000003oool0oooo03d0oooo00<0omnL04P0
001d_ol08`3oool00`000000oooo0?ooo`0I0?ooo`030;mdB02Lgol0oooo0200oooo0P00000M0?oo
o`030?noM0000000M;oo0180oooo000I0?ooo`040?oOW0180000001809cOocd0oooo00<0oooO09a8
M02oool0503oool00`000000oooo0?ooo`0V0?ooo`030?oOW0180000M;oo01D0oooo00<000000?oo
o`3oool08P3oool00`000000oooo0?ooo`0m0?ooo`030?oOW0180000M;oo02<0oooo00<000000?oo
o`3oool06@3oool00`3ooml0W4Qd0;ooo`0R0?ooo`030000003oool0oooo01T0oooo00@0omnL04P0
000004P0W=oo4P3oool001L0oooo00D0okmd000000000000001809cOo`0o0?ooo`030?noM000M;l0
oooo01<0oooo00<000000?ooo`3oool09P3oool00`3ogi`0B00007Boo`0F0?ooo`8000008P3oool0
0`000000oooo0?ooo`0m0?ooo`030?oOW0180000M;oo02<0oooo00<000000?ooo`3oool06P3oool0
0`3o_g@007Bo0?ooo`0R0?ooo`8000005`3oool01@3o_g@000000000000004P0W=oo01<0oooo000I
0?ooo`030;md0000M;l0oooo03l0oooo00<0oono07@0B02Lgol04`3oool00`000000oooo0?ooo`0V
0?ooo`030?oOW0180000M;oo01P0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0m
0?ooo`030?oOW0180000M;oo02<0oooo00<000000?ooo`3oool06P3oool00`3ookl0M01809cOo`0T
0?ooo`030000003oool0oooo01H0oooo00<0_g@0001d_`3oool04`3oool001X0oooo00<0_g@004RL
g`3oool0?P3oool00`3ookl0M00004RLg`0C0?ooo`030000003oool0oooo02H0oooo00<0omnL04P0
001d_ol06@3oool2000001l0oooo00<000000?ooo`3oool0?@3oool00`3ogi`0B00007Boo`0S0?oo
o`030000003oool0oooo01X0oooo00<0oono07@00018W=l09@3oool2000001L0oooo00<0_g@004RL
g`3oool04P3oool001D0oooo00<0oooO09bL_`3oool00P3oool00`3OW4P0001d0;ooo`0i0?ooo`03
0?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l04`3oool00`000000oooo0?ooo`0V0?ooo`03
0?oOW0180000M;oo01/0oooo0P00000M0?ooo`<00000?@3oool00`3ogi`0B00007Boo`0S0?ooo`03
0000003oool0oooo01D0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`0W0?ooo`80
0000403oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo0180oooo000F0?ooo`060?no
M018W=l0oooo0?oo_`1d000007Bo>P3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo
4`3oool00`000000oooo0?ooo`0T0?ooo`050?oog`2LB00000000000001d_ol07@3oool2000001/0
oooo00<000000?ooo`3oool0>`3oool01@3ooml0W4P0000000000000M;oo02<0oooo00<000000?oo
o`3oool05@3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo:@3oool2000000l0oooo
00H0okmd04RLg`3oool0oono07@00000M;lC0?ooo`005`3oool01@3OW4P000000000000007@0_ooo
03/0oooo00H0oooO09a80000000000000000M02ooolD0?ooo`030000003oool0oooo02H0oooo00<0
okmd0000001d_ol07`3oool2000001T0oooo00<000000?ooo`3oool0?@3oool00`3o_g@0000007Bo
o`0S0?ooo`030000003oool0oooo01H0oooo00H0oooO09a80000000000000000M02oool/0?ooo`<0
00003@3oool01@3OW4P000000000000007@0_ooo01<0oooo001a0?ooo`030000003oool0oooo04X0
oooo0P00000G0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0B`3oool2000002<0
oooo001a0?ooo`030000003oool0oooo04`0oooo0P00000E0?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool0C@3oool200000240oooo001a0?ooo`030000003oool0oooo04h0oooo0P00
000C0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0C`3oool3000001h0oooo001a
0?ooo`030000003oool0oooo0500oooo0P00000A0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool0DP3oool2000001`0oooo001a0?ooo`030000003oool0oooo0580oooo0`00000>0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool0E03oool3000001T0oooo001a0?ooo`03
0000003oool0oooo05D0oooo0P00000<0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3o
ool0E`3oool2000001L0oooo001a0?ooo`030000003oool0oooo05L0oooo0`0000090?ooo`030000
003oool0oooo06<0oooo00<000000?ooo`3oool0F@3oool2000001D0oooo001a0?ooo`030000003o
ool0oooo05X0oooo0P0000070?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0F`3o
ool2000001<0oooo001a0?ooo`030000003oool0oooo05`0oooo0P0000050?ooo`030000003oool0
oooo06<0oooo00<000000?ooo`3oool0G@3oool200000140oooo001a0?ooo`030000003oool0oooo
05h0oooo0`0000020?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0G`3oool20000
00l0oooo001a0?ooo`030000003oool0oooo0640oooo0`00001U0?ooo`030000003oool0oooo0640
oooo0P00000=0?ooo`00o`00002`00000040oooo00060?ooo`030000003oool0oooo00<0oooo0`00
00040?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`0000002P3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00800000203oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0000000`0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`050?ooo`003`3oool3
000000030?ooo`000000oooo0400oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0R
0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool00P3oool3000003T0oooo00<00000
0?ooo`3oool08P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo03l0oooo00<00000
0?ooo`3oool04P3oool00180oooo0P00001M0?ooo`030000003oool0oooo06<0oooo00<000000?oo
o`3oool01@3oool2000005`0oooo00<000000?ooo`3oool0L03oool001@0oooo0`00001J0?ooo`03
0000003oool0oooo06<0oooo00<000000?ooo`3oool01`3oool2000005X0oooo00<000000?ooo`3o
ool0L03oool001L0oooo0`00001G0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0
2@3oool3000005L0oooo00<000000?ooo`3oool0L03oool001X0oooo0P00001E0?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`3oool0303oool2000005D0oooo00<000000?ooo`3oool0L03o
ool001`0oooo0P00001C0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool03P3oool3
00000580oooo00<000000?ooo`3oool0L03oool001h0oooo0P00001A0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool04@3oool200000500oooo00<000000?ooo`3oool0L03oool00200
oooo0P00001?0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool04`3oool2000004h0
oooo00<000000?ooo`3oool0L03oool00280oooo0P00001=0?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool05@3oool2000004`0oooo00<000000?ooo`3oool0L03oool002@0oooo0P00
001;0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool05`3oool2000004X0oooo00<0
00000?ooo`3oool0L03oool002H0oooo0P0000190?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool06@3oool2000004P0oooo00<000000?ooo`3oool0L03oool002P0oooo0P0000170?oo
o`030000003oool0oooo06<0oooo0`00000K0?ooo`800000AP3oool00`000000oooo0?ooo`1`0?oo
o`00:P3oool2000004D0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0M0?ooo`80
0000A03oool00`000000oooo0?ooo`1`0?ooo`00;03oool2000004<0oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`0O0?ooo`800000@P3oool00`000000oooo0?ooo`1`0?ooo`00;P3o
ool200000440oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0Q0?ooo`800000@03o
ool00`000000oooo0?ooo`1`0?ooo`00<03oool00`000000oooo0?ooo`0n0?ooo`030000003oool0
oooo06<0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0
oooo0700oooo000a0?ooo`800000?P3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
02@0oooo0P00000m0?ooo`030000003oool0oooo0700oooo000c0?ooo`030000003oool0oooo03/0
oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo03X0
oooo00<000000?ooo`3oool0L03oool003@0oooo0P00000k0?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool09`3oool2000003X0oooo00<000000?ooo`3oool0L03oool003H0oooo00<0
00000?ooo`3oool0>03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo02T0oooo00<0
00000?ooo`3oool0=`3oool00`000000oooo0?ooo`1`0?ooo`00=`3oool00`000000oooo0?ooo`0g
0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0f
0?ooo`030000003oool0oooo0700oooo000h0?ooo`800000=`3oool00`000000oooo0?ooo`1S0?oo
o`030000003oool0oooo02/0oooo0P00000f0?ooo`030000003oool0oooo0700oooo000j0?ooo`03
0000003oool0oooo03@0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0]0?ooo`03
0000003oool0oooo03<0oooo00<000000?ooo`3oool0L03oool003/0oooo00<000000?ooo`3oool0
<`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0
<P3oool00`000000oooo0?ooo`1`0?ooo`00?03oool2000003<0oooo00<000000?ooo`3oool0H`3o
ool3000002l0oooo0P00000b0?ooo`030000003oool0oooo0700oooo000n0?ooo`030000003oool0
oooo0300oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0
oooo02l0oooo00<000000?ooo`3oool0L03oool003l0oooo00<000000?ooo`3oool0;`3oool00`00
0000oooo0?ooo`1S0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0;P3oool00`00
0000oooo0?ooo`1`0?ooo`00@03oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo06<0
oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo0700
oooo00110?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0H`3oool00`000000oooo
0?ooo`0d0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0L03oool00480oooo0P00
000]0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0=@3oool2000002`0oooo00<0
00000?ooo`3oool0L03oool004@0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`1S
0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`1`
0?ooo`00A@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo06<0oooo00<000000?oo
o`3oool0>03oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo0700oooo00160?ooo`03
0000003oool0oooo02P0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0i0?ooo`03
0000003oool0oooo02L0oooo00<000000?ooo`3oool0L03oool004L0oooo00<000000?ooo`3oool0
9`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0
9P3oool00`000000oooo0?ooo`1`0?ooo`00A`3oool00`000000oooo0?ooo`0W0?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0V0?ooo`030000003o
ool0oooo0700oooo00180?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`0k0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0L03oool0
04T0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
03`0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`1`0?ooo`00BP3oool00`000000
oooo0?ooo`0T0?ooo`030000003oool0oooo06<0oooo0`00000l0?ooo`030000003oool0oooo02@0
oooo00<000000?ooo`3oool0L03oool004X0oooo00<000000?ooo`3oool0903oool00`000000oooo
0?ooo`1S0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool08`3oool00`000000oooo
0?ooo`1`0?ooo`00B`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo06<0oooo00<0
00000?ooo`3oool0?P3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo0700oooo001<
0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0n
0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0L03oool004d0oooo00<000000?oo
o`3oool08@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo03l0oooo00<000000?oo
o`3oool08@3oool00`000000oooo0?ooo`1`0?ooo`00C@3oool00`000000oooo0?ooo`0Q0?ooo`03
0000003oool0oooo06<0oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`0P0?ooo`03
0000003oool0oooo0700oooo001>0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0
L03oool004l0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`1S0?ooo`030000003o
ool0oooo0440oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`1`0?ooo`00C`3oool0
0`000000oooo0?ooo`0O0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0@P3oool0
0`000000oooo0?ooo`0N0?ooo`030000003oool0oooo0700oooo001@0?ooo`030000003oool0oooo
01h0oooo00<000000?ooo`3oool0DP3oool00`2oM0000000000000040000000304RLg`3oool0oooo
00L0oooo00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo
0700oooo001@0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0D`3oool00`3OW4P0
000000000003000000030018W03Oool0oooo00L0oooo00<000000?ooo`3oool0@`3oool00`000000
oooo0?ooo`0M0?ooo`030000003oool0oooo0700oooo001A0?ooo`030000003oool0oooo01d0oooo
00<000000?ooo`3oool0E03oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo00L0oooo
00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0700oooo
001B0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0E@3oool00`3OW4P0M;oo0?oo
o`0;0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0703oool00`000000oooo0?oo
o`1`0?ooo`00DP3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo05H0oooo00<0_gA8
09cOo`3oool02P3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo01`0oooo00<00000
0?ooo`3oool0L03oool005<0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`1F0?oo
o`030?oog`2LB7@0_ooo00X0oooo100000140?ooo`030000003oool0oooo01/0oooo00<000000?oo
o`3oool0L03oool005<0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`1G0?ooo`03
0?noM000M;l0oooo00T0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0K0?ooo`03
0000003oool0oooo0700oooo001D0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
E`3oool00`3ookl0M01809cOo`090?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`1`0?ooo`00E03oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo05L0oooo00<0oono07@00018W=l02@3oool00`000000oooo0?ooo`160?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool0L03oool005@0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`1B0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l02@3oool0
0`000000oooo0?ooo`170?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0L03oool0
05D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`1B0?ooo`080?oog`2LB4P0W=oo
0?ooo`3ookl0M0000000M02oool90?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`1`0?ooo`00E@3oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo05<0oooo00H0oooO09a80000000000000000M02oool:0?ooo`030000003oool0oooo04P0
oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1`0?ooo`00E@3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0B03oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo0700oooo001F0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool0H`3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool0L03oool005H0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1S
0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`1`
0?ooo`00EP3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo06<0oooo00<000000?oo
o`3oool0BP3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0700oooo001G0?ooo`03
0000003oool0oooo01L0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1:0?ooo`03
0000003oool0oooo01H0oooo00<000000?ooo`3oool0L03oool005L0oooo00<000000?ooo`3oool0
5`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0
5P3oool00`000000oooo0?ooo`1`0?ooo`00E`3oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo06<0oooo0`00001;0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0L03o
ool005P0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
oooo04/0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`1`0?ooo`00F03oool00`00
0000oooo0?ooo`0F0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0B`3oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo0700oooo001I0?ooo`030000003oool0oooo01D0
oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo01D0
oooo00<000000?ooo`3oool0L03oool005T0oooo00<000000?ooo`3oool05@3oool00`000000oooo
0?ooo`1S0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0503oool00`000000oooo
0?ooo`1`0?ooo`00FP3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo06<0oooo00<0
00000?ooo`3oool0C03oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0700oooo001J
0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1<
0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0L03oool005X0oooo00<000000?oo
o`3oool0503oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo04d0oooo00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`1`0?ooo`00FP3oool00`000000oooo0?ooo`0D0?ooo`03
0000003oool0oooo06<0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`0C0?ooo`03
0000003oool0oooo0700oooo001K0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0
L03oool005/0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`1S0?ooo`030000003o
ool0oooo04h0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`1`0?ooo`00F`3oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0CP3oool0
0`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0700oooo001K0?ooo`030000003oool0oooo
01<0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo
0180oooo00<000000?ooo`3oool0L03oool005`0oooo00<000000?ooo`3oool04P3oool00`000000
oooo0?ooo`1S0?ooo`<00000C`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0700
oooo001L0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0H`3oool00`000000oooo
0?ooo`1?0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0L03oool005`0oooo00<0
00000?ooo`3oool04P3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo04l0oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`1`0?ooo`00G03oool00`000000oooo0?ooo`0B
0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo0700oooo001L0?ooo`030000003oool0oooo0180oooo00<000000?oo
o`3oool0H`3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0L03oool005d0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool0oooo0500oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1`0?ooo`00
G@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0
D03oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0700oooo001M0?ooo`030000003o
ool0oooo0140oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1@0?ooo`030000003o
ool0oooo0100oooo00<000000?ooo`3oool0L03oool005d0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`1`0?ooo`00GP3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
0700oooo001N0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0H`3oool00`000000
oooo0?ooo`1A0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0L03oool005h0oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0540oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1`0?ooo`00GP3oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo0700oooo001N0?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool0L03oool005h0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1S0?oo
o`<00000D@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0700oooo001O0?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1B0?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool0L03oool005l0oooo00<000000?ooo`3oool0
3`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`1`0?ooo`00G`3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo0700oooo001O0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`1B0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0L03oool0
05l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
0580oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1`0?ooo`00G`3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0DP3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo0700oooo001P0?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool0L03oool00600oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`1F0?ooo`040?oog`2LB00004RL0=ooo`T0oooo00<000000?ooo`3oool0D`3oool00`000000oooo
0?ooo`0=0?ooo`030000003oool0oooo0700oooo001P0?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool0EP3oool0103ooml0W4P00018W03Oool90?ooo`030000003oool0oooo05<0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1`0?ooo`00H03oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo05H0oooo00@0oooO09a80000B9`0gooo2@3oool00`000000oooo
0?ooo`1C0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0L03oool00600oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`1B0?ooo`030;md00000000000000D0000000<0
M;oo0?ooo`3oool01P3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool0L03oool00600oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1B
0?ooo`080?oOW01dW=l0oooo0?ooo`3ooml0W4P00018W03Oool90?ooo`030000003oool0oooo05<0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1`0?ooo`00H03oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo05<0oooo00L0gia807Boo`3oool0oooO09a80000B9`0gooo
00T0oooo1000001B0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0L03oool00600
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1C0?ooo`070?oog`2LB7@0_ooo0?oo
g`2LB00004RL0=ooo`090?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`1`0?ooo`00H@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
05@0oooo00H0omnL04Qd_`3ooml0W4P00018W03Oool90?ooo`030000003oool0oooo05@0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`1`0?ooo`00H@3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo05D0oooo00D0gia807Bog`2LB00004RL0=ooo`090?ooo`030000003o
ool0oooo05@0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1`0?ooo`00H@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo05D0oooo00D0oooO09a8B018B00004RL0=oo
o`090?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`1`0?ooo`00H@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo05H0oooo00@0omnL
04P00000B9`0gooo2@3oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool0L03oool00640oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1G
0?ooo`030=nLB000B9`0gooo00T0oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo0700oooo001Q0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool0H`3oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool0L03oool00640oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool0oooo05@0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1`0?ooo`00
H@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0
E03oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0700oooo001R0?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1D0?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool0L03oool00680oooo00<000000?ooo`3oool0303oool0
0`000000oooo0?ooo`1S0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`1`0?ooo`00HP3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
0700oooo001R0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0H`3oool3000005D0
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1`0?ooo`00HP3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0E@3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo0700oooo001R0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool0H`3oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool0L03oool00680oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1S
0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1`
0?ooo`00HP3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo06<0oooo00<000000?oo
o`3oool0E@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0700oooo001S0?ooo`03
0000003oool0oooo00/0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1E0?ooo`03
0000003oool0oooo00/0oooo00<000000?ooo`3oool0L03oool006<0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`1`0?ooo`00H`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo0700oooo001S0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`1F0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0L03oool0
06<0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
05H0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`1`0?ooo`00H`3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0EP3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo0700oooo001T0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool0L03oool006@0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`1S0?ooo`030000003oool0oooo05H0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`1`0?ooo`00I03oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool0EP3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0700oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-3.30486, -5.00005, \
0.0230037, 0.0564177}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Resolution and smoothness options", "Subsubsection"],

Cell[TextData[{
  "Obviously ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " must evaluate the function for a finite number of points if it is to \
generate a plot.  According to the options listed above, the default value of \
",
  StyleBox["PlotPoints", "InlineInput",
    FontWeight->"Bold"],
  " is 25 for its initial evaluation.  Because ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " uses an adaptive sampling algorithm, this parameter represents the \
minimum number of points.  If a strong kink is detected, it may automatically \
increase the density of points in that region in an attempt to improve the \
resolution of the plot.  However, sometimes a kink will not be noticed and we \
might get a somewhat spurious result, or perhaps we are fussy about segmented \
lines.  In either case, we can increase ",
  StyleBox["PlotPoints", "InlineInput",
    FontWeight->"Bold"],
  " to improve either the accuracy or appearance of the plot.  Other relevant \
parameters are ",
  StyleBox["MaxBend", "InlineInput",
    FontWeight->"Bold"],
  ", which is the maximum angle permitted between successive line segments, \
and ",
  StyleBox["PlotDivision", "InlineInput",
    FontWeight->"Bold"],
  ", the maximum number of bisections permitted to reduce the bend angle.  \
These parameters are obviously interrelated."
}], "Text"],

Cell[TextData[{
  "For example, a smoother plot of the tangent function can be obtained by \
reducing ",
  StyleBox["MaxBend", "InlineInput",
    FontWeight->"Bold"],
  " from its default value of 10.  Even so, a purist might object to the \
vertical lines connecting \[PlusMinus]\[Infinity]."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Tan[x], \ {x, \(-\[Pi]\), \ \[Pi]}, 
        PlotRange \[Rule] {\(-5\), 5}, MaxBend \[Rule] 1];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.151576 0.309017 0.0618034 [
[.04527 .29652 -9.15625 -15.6875 ]
[.04527 .29652 9.15625 0 ]
[.19685 .29652 -9.15625 -15.6875 ]
[.19685 .29652 9.15625 0 ]
[.34842 .29652 -9.15625 -15.6875 ]
[.34842 .29652 9.15625 0 ]
[.65158 .29652 -5 -15.6875 ]
[.65158 .29652 5 0 ]
[.80315 .29652 -5 -15.6875 ]
[.80315 .29652 5 0 ]
[.95473 .29652 -5 -15.6875 ]
[.95473 .29652 5 0 ]
[.4875 .0618 -18.3125 -7.84375 ]
[.4875 .0618 0 7.84375 ]
[.4875 .18541 -18.3125 -7.84375 ]
[.4875 .18541 0 7.84375 ]
[.4875 .43262 -10 -7.84375 ]
[.4875 .43262 0 7.84375 ]
[.4875 .55623 -10 -7.84375 ]
[.4875 .55623 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.04527 .30902 m
.04527 .31527 L
s
gsave
.04527 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.19685 .30902 m
.19685 .31527 L
s
gsave
.19685 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.34842 .30902 m
.34842 .31527 L
s
gsave
.34842 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.65158 .30902 m
.65158 .31527 L
s
gsave
.65158 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.80315 .30902 m
.80315 .31527 L
s
gsave
.80315 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.95473 .30902 m
.95473 .31527 L
s
gsave
.95473 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.07559 .30902 m
.07559 .31277 L
s
.1059 .30902 m
.1059 .31277 L
s
.13622 .30902 m
.13622 .31277 L
s
.16653 .30902 m
.16653 .31277 L
s
.22716 .30902 m
.22716 .31277 L
s
.25748 .30902 m
.25748 .31277 L
s
.28779 .30902 m
.28779 .31277 L
s
.31811 .30902 m
.31811 .31277 L
s
.37874 .30902 m
.37874 .31277 L
s
.40905 .30902 m
.40905 .31277 L
s
.43937 .30902 m
.43937 .31277 L
s
.46968 .30902 m
.46968 .31277 L
s
.53032 .30902 m
.53032 .31277 L
s
.56063 .30902 m
.56063 .31277 L
s
.59095 .30902 m
.59095 .31277 L
s
.62126 .30902 m
.62126 .31277 L
s
.68189 .30902 m
.68189 .31277 L
s
.71221 .30902 m
.71221 .31277 L
s
.74252 .30902 m
.74252 .31277 L
s
.77284 .30902 m
.77284 .31277 L
s
.83347 .30902 m
.83347 .31277 L
s
.86378 .30902 m
.86378 .31277 L
s
.8941 .30902 m
.8941 .31277 L
s
.92441 .30902 m
.92441 .31277 L
s
.01496 .30902 m
.01496 .31277 L
s
.98504 .30902 m
.98504 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5 .0618 m
.50625 .0618 L
s
gsave
.4875 .0618 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.5 .18541 m
.50625 .18541 L
s
gsave
.4875 .18541 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 .43262 m
.50625 .43262 L
s
gsave
.4875 .43262 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 .55623 m
.50625 .55623 L
s
gsave
.4875 .55623 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .09271 m
.50375 .09271 L
s
.5 .12361 m
.50375 .12361 L
s
.5 .15451 m
.50375 .15451 L
s
.5 .21631 m
.50375 .21631 L
s
.5 .24721 m
.50375 .24721 L
s
.5 .27812 m
.50375 .27812 L
s
.5 .33992 m
.50375 .33992 L
s
.5 .37082 m
.50375 .37082 L
s
.5 .40172 m
.50375 .40172 L
s
.5 .46353 m
.50375 .46353 L
s
.5 .49443 m
.50375 .49443 L
s
.5 .52533 m
.50375 .52533 L
s
.5 .0309 m
.50375 .0309 L
s
.5 .58713 m
.50375 .58713 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .30902 m
.04262 .31673 L
.0522 .32073 L
.06244 .32512 L
.0679 .32752 L
.07286 .32974 L
.0779 .33206 L
.08255 .33425 L
.08798 .33687 L
.09079 .33825 L
.09388 .33981 L
.09907 .3425 L
.10192 .34402 L
.10458 .34547 L
.10954 .34825 L
.11235 .34988 L
.11496 .35142 L
.11742 .35291 L
.11974 .35435 L
.12241 .35605 L
.12483 .35762 L
.12708 .35913 L
.12946 .36076 L
.13149 .36219 L
.13373 .36379 L
.13613 .36556 L
.1387 .36752 L
.14114 .36943 L
.14335 .37122 L
.14564 .37313 L
.14816 .3753 L
.14943 .37642 L
.15079 .37766 L
.15197 .37874 L
.15326 .37995 L
.1556 .38219 L
.15687 .38345 L
.15805 .38464 L
.16017 .38684 L
.16136 .3881 L
.16244 .38927 L
.16373 .39071 L
.1651 .39226 L
.16629 .39364 L
.16756 .39515 L
.16893 .39681 L
.17021 .3984 L
.17173 .40034 L
.17312 .40217 L
.1744 .40391 L
Mistroke
.17559 .40556 L
.1769 .40741 L
.17829 .40945 L
.17961 .41145 L
.18086 .41338 L
.18199 .41517 L
.1832 .41714 L
.18444 .41924 L
.18558 .4212 L
.18689 .42353 L
.18813 .42581 L
.18927 .42797 L
.19033 .43003 L
.19149 .43235 L
.19273 .4349 L
.19396 .43754 L
.19466 .43906 L
.19529 .44048 L
.19646 .44316 L
.1977 .44608 L
.19835 .44768 L
.19905 .44942 L
.20031 .45265 L
.20109 .45469 L
.20181 .45666 L
.20252 .4586 L
.20317 .46046 L
.20439 .46402 L
.20504 .46596 L
.20574 .46809 L
.20646 .47038 L
.20715 .4726 L
.20844 .47692 L
.20965 .48113 L
.21077 .4852 L
.21199 .48987 L
.2133 .49508 L
.21459 .50052 L
.21531 .50371 L
.21598 .50672 L
.21721 .51249 L
.2185 .51892 L
.21919 .52249 L
.21992 .52642 L
.22062 .53027 L
.22125 .53387 L
.22199 .53825 L
.22279 .54319 L
.22355 .54804 L
.22425 .55268 L
Mistroke
.2255 .56139 L
.22669 .57022 L
.22775 .57867 L
.22892 .58849 L
.23018 .59994 L
.23088 .60673 L
.23152 .61316 L
Mfstroke
.23152 .61316 m
.23197 .61803 L
s
.26153 .61803 m
.26156 0 L
s
.29183 0 m
.29247 .00673 L
.29317 .01367 L
.29381 .01973 L
.29454 .02644 L
.29524 .03253 L
.29641 .04219 L
.29765 .05179 L
.29871 .05948 L
.29987 .06748 L
.30113 .07553 L
.30183 .07984 L
.30248 .08366 L
.30367 .0904 L
.30492 .09711 L
.30615 .10335 L
.30685 .10672 L
.30749 .10975 L
.30867 .11509 L
.30992 .12047 L
.31103 .12506 L
.31208 .12918 L
.31323 .13351 L
.31446 .13796 L
.31577 .14247 L
.3165 .14493 L
.31719 .14715 L
.31845 .15109 L
.31915 .15322 L
.31978 .15511 L
.32099 .15858 L
.32229 .16218 L
.32352 .16545 L
.32465 .16835 L
.32572 .17099 L
.32686 .17373 L
.32793 .17623 L
.32891 .17844 L
.33006 .18097 L
.33131 .18362 L
.33249 .18604 L
.33359 .18822 L
.33471 .19039 L
.33576 .19238 L
.33692 .1945 L
.33816 .19671 L
.3394 .19885 L
.34069 .20103 L
.3418 .20284 L
.34302 .20476 L
Mistroke
.34434 .20681 L
.34556 .20864 L
.3469 .2106 L
.34831 .21261 L
.34954 .21431 L
.35069 .21586 L
.35196 .21753 L
.35331 .21927 L
.35451 .22077 L
.35582 .22237 L
.35697 .22375 L
.3582 .22518 L
.3593 .22644 L
.36051 .22781 L
.36266 .23015 L
.36398 .23155 L
.36537 .23299 L
.36785 .23549 L
.36924 .23684 L
.37053 .23807 L
.37204 .23948 L
.37346 .24078 L
.37603 .24306 L
.37837 .24507 L
.38089 .24716 L
.38362 .24935 L
.38597 .25117 L
.38849 .25308 L
.39088 .25482 L
.39306 .25637 L
.39561 .25814 L
.39839 .26002 L
.40102 .26174 L
.40346 .2633 L
.40621 .26501 L
.4088 .26658 L
.41364 .26943 L
.41636 .27097 L
.41885 .27236 L
.4218 .27397 L
.42449 .2754 L
.42919 .27784 L
.43423 .28038 L
.4389 .28266 L
.44332 .28477 L
.44846 .28715 L
.45322 .28931 L
.46399 .29405 L
.47362 .29815 L
.48266 .30191 L
Mistroke
.50326 .31035 L
.52174 .31794 L
.53221 .32235 L
.53732 .32455 L
.54195 .32657 L
.55165 .33093 L
.55697 .33341 L
.562 .33581 L
.56727 .3384 L
.57213 .34087 L
.57746 .34367 L
.58015 .34513 L
.5831 .34676 L
.5856 .34817 L
.58832 .34974 L
.59072 .35116 L
.59327 .3527 L
.5958 .35427 L
.59809 .35572 L
.60026 .35713 L
.60256 .35865 L
.60525 .36048 L
.60773 .36222 L
.6104 .36414 L
.61332 .36632 L
.61585 .36827 L
.61856 .37043 L
.62112 .37254 L
.62345 .37452 L
.62575 .37654 L
.62827 .37884 L
.62955 .38004 L
.63092 .38134 L
.6321 .3825 L
.63339 .38378 L
.63467 .38508 L
.63585 .3863 L
.63708 .3876 L
.63819 .38879 L
.64031 .39115 L
.64148 .39248 L
.64258 .39375 L
.64388 .39531 L
.6451 .39679 L
.64638 .39837 L
.64777 .40016 L
.64898 .40174 L
.65008 .40321 L
.65135 .40496 L
.65256 .40665 L
Mistroke
.65386 .40853 L
.65524 .41059 L
.65642 .4124 L
.65773 .41445 L
.65912 .41671 L
.66063 .41923 L
.66205 .42169 L
.66336 .42403 L
.66463 .42639 L
.66583 .42868 L
.66691 .43081 L
.66809 .43318 L
.66936 .43584 L
.67001 .43723 L
.67072 .43877 L
.67197 .44159 L
.67315 .44431 L
.67432 .44714 L
.67544 .44992 L
.67644 .45249 L
.67754 .4554 L
.67872 .45865 L
.67999 .46226 L
.68118 .4658 L
.68226 .46914 L
.68356 .47331 L
.68421 .47549 L
.68493 .47793 L
.6862 .48242 L
.68738 .4868 L
.6881 .48955 L
.68878 .49223 L
.69006 .49745 L
.69077 .50049 L
.69143 .50337 L
.69215 .50664 L
.69292 .51021 L
.69411 .51596 L
.69538 .5225 L
.69659 .52904 L
.69769 .53536 L
.69892 .54278 L
.69961 .54719 L
.70026 .55143 L
.70156 .56045 L
.70229 .56578 L
.70295 .57081 L
.70415 .58036 L
.70527 .58991 L
.70649 .60116 L
Mistroke
.70781 .61423 L
Mfstroke
.70781 .61423 m
.70817 .61803 L
s
.73687 .61803 m
.73687 0 L
s
.76803 0 m
.76861 .00621 L
.76935 .01357 L
.77004 .02009 L
.77129 .03135 L
.77199 .03728 L
.77263 .04245 L
.7733 .04775 L
.77402 .05312 L
.77529 .06224 L
.776 .06706 L
.77666 .0714 L
.77739 .07599 L
.77817 .08073 L
.77884 .08467 L
.77957 .08882 L
.78026 .09261 L
.7809 .09601 L
.7821 .10217 L
.78338 .10836 L
.78453 .11361 L
.78577 .11906 L
.78708 .12446 L
.78829 .12925 L
.78943 .13354 L
.79046 .13729 L
.79165 .14144 L
.79278 .1452 L
.79401 .14913 L
.7953 .15312 L
.79641 .15639 L
.79763 .15985 L
.79894 .16342 L
.79968 .16536 L
.80035 .1671 L
.80165 .17037 L
.80289 .17336 L
.80394 .1758 L
.80509 .17842 L
.8063 .18107 L
.80744 .18348 L
.80849 .18565 L
.80946 .1876 L
.81053 .18968 L
.81165 .19181 L
.81279 .19391 L
.814 .1961 L
.81515 .19811 L
.8162 .19988 L
.81743 .20192 L
Mistroke
.81876 .20406 L
.82002 .20603 L
.82118 .20779 L
.82243 .20964 L
.82362 .21136 L
.82468 .21286 L
.82584 .21446 L
.8271 .21615 L
.82844 .21791 L
.82971 .21953 L
.83086 .22096 L
.8331 .22368 L
.83436 .22516 L
.83552 .22649 L
.83766 .22887 L
.83874 .23003 L
.83992 .23129 L
.8413 .23273 L
.84261 .23406 L
.84507 .23649 L
.84644 .23781 L
.84771 .239 L
.8492 .24037 L
.85059 .24162 L
.85289 .24364 L
.85542 .24579 L
.85789 .24782 L
.86055 .24993 L
.86295 .25177 L
.86551 .25369 L
.86793 .25544 L
.87013 .25699 L
.87271 .25876 L
.87551 .26063 L
.87816 .26235 L
.8806 .26389 L
.88548 .26688 L
.88826 .26851 L
.89085 .27001 L
.89608 .27293 L
.89902 .27452 L
.9017 .27594 L
.90649 .27841 L
.91158 .28095 L
.91639 .28328 L
.92074 .28534 L
.92588 .28771 L
.93143 .29022 L
.94148 .29461 L
.94954 .29804 L
Mistroke
.95837 .30172 L
.97619 .30902 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oL03oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo06<0oooo001`0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool0H`3oool00700oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`1F0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`1R0?ooo`00L03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo05H0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0680oooo
001`0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?oo
o`1S0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0HP3oool00700oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`1R0?ooo`00L03oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo05H0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo0680oooo001`0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool0EP3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool0HP3oool00700oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1F0?ooo`030000
003oool0oooo06<0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1R0?ooo`00L03o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo05H0oooo00<000000?ooo`3oool0H`3o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0680oooo001`0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool0HP3oool00700oooo00<000000?ooo`3oool0303oool00`00
0000oooo0?ooo`1E0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0303oool00`00
0000oooo0?ooo`1Q0?ooo`00L03oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo05D0
oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0640
oooo001`0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0E@3oool3000006<0oooo
00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1Q0?ooo`00L03oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo0640oooo001`0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0E@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0H@3oool00700oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1E0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1Q0?oo
o`00L03oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0640oooo001`0?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`1S0?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool0H@3oool00700oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`1Q0?ooo`00L03oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo05D0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo0600oooo001`0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0A`3oool0103o
oml0W4P00018W03Oool90?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`1P0?ooo`00L03oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
04L0oooo00@0oooO09a80000B9`0gooo2@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool0H03oool00700oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`170?ooo`040?oog`2LB00004RL0=ooo`T0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0600oooo001`0?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool0@`3oool00`2oM0000000000000050000000307Boo`3oool0oooo
00H0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
0600oooo001`0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0=P3oool00`3ooml0
W4P0000000070000000504RLg`3oool0oooo0?oOW01dW=l00P3oool0103ooml0W4P00018W03Oool9
0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1P
0?ooo`00L03oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo04@0oooo00L0gia807Bo
o`3oool0oooO09a80000B9`0gooo00T0oooo1000001R0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool0H03oool00700oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`14
0?ooo`070?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`090?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1P0?ooo`00L03oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo04@0oooo00H0omnL04Qd_`3ooml0W4P00018W03Oool90?ooo`03
0000003oool0oooo06<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1O0?ooo`00
L03oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo04D0oooo00D0gia807Bog`2LB000
04RL0=ooo`090?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`1O0?ooo`00L03oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo04D0oooo
00D0oooO09a8B018B00004RL0=ooo`090?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`1O0?ooo`00L03oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo04H0oooo00@0omnL04P00000B9`0gooo2@3oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool0G`3oool00700oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`170?ooo`030=nLB000B9`0gooo00T0oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo05l0oooo001`0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`1S0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool0G`3oool00700oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`1B0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`1N0?ooo`00L03oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
0580oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
05h0oooo001`0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0DP3oool00`000000
oooo0?ooo`1S0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0GP3oool00700oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1N0?ooo`00L03oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo05h0oooo001`0?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool0D@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool0G@3oool00700oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1A0?oo
o`<00000H`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo05d0oooo001`0?ooo`03
0000003oool0oooo0100oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool0oooo0100oooo00<000000?ooo`3oool0G@3oool00700oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`1M0?ooo`00L03oool00`000000oooo0?ooo`0A0?ooo`030000003o
ool0oooo0500oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo05d0oooo001`0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0D03oool0
0`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0G@3oool0
0700oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1M0?ooo`00L03oool00`000000
oooo0?ooo`0A0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0H`3oool00`000000
oooo0?ooo`0A0?ooo`030000003oool0oooo05`0oooo001`0?ooo`030000003oool0oooo0140oooo
00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0140oooo
00<000000?ooo`3oool0G03oool00700oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`1@0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`1L0?ooo`00L03oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo04l0oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo05`0oooo001`0?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`1S0?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool0F`3oool00700oooo00<000000?ooo`3o
ool04P3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3o
ool04P3oool00`000000oooo0?ooo`1K0?ooo`00L03oool00`000000oooo0?ooo`0B0?ooo`030000
003oool0oooo04l0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0B0?ooo`030000
003oool0oooo05/0oooo001`0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0CP3o
ool3000006<0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`1J0?ooo`00L03oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo05X0oooo001`0?ooo`030000003oool0oooo
01<0oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
01<0oooo00<000000?ooo`3oool0FP3oool00700oooo00<000000?ooo`3oool04`3oool00`000000
oooo0?ooo`1>0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool04`3oool00`000000
oooo0?ooo`1J0?ooo`00L03oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo04d0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo05T0oooo
001`0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?oo
o`1S0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0F@3oool00700oooo00<00000
0?ooo`3oool0503oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool0503oool00`000000oooo0?ooo`1I0?ooo`00L03oool00`000000oooo0?ooo`0D0?oo
o`030000003oool0oooo04d0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0D0?oo
o`030000003oool0oooo05T0oooo001`0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3o
ool0C03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3o
ool0F@3oool00700oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`1<0?ooo`030000
003oool0oooo06<0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`1H0?ooo`00L03o
ool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0H`3o
ool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo05P0oooo001`0?ooo`030000003oool0
oooo01D0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
oooo01D0oooo00<000000?ooo`3oool0F03oool00700oooo00<000000?ooo`3oool05P3oool00`00
0000oooo0?ooo`1;0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool05P3oool00`00
0000oooo0?ooo`1G0?ooo`00L03oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo04/0
oooo0`00001S0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0E`3oool00700oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`1F0?ooo`00L03oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0oooo05H0oooo001`0?ooo`030000003oool0oooo01P0oooo00<00000
0?ooo`3oool0B@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool0EP3oool00700oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`190?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1E0?oo
o`00L03oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo05D0oooo001`0?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`1S0?ooo`030000
003oool0oooo01P0oooo00<000000?ooo`3oool0E@3oool00700oooo00<000000?ooo`3oool06@3o
ool00`000000oooo0?ooo`180?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool06@3o
ool00`000000oooo0?ooo`1D0?ooo`00L03oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo03H0oooo00<0_g@0000000000000100000000`18W=l0oooo0?ooo`070?ooo`030000003oool0
oooo06<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`1D0?ooo`00L03oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo03L0oooo00<0gia80000000000000`0000000`00
B9`0gooo0?ooo`070?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`1C0?ooo`00L03oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo03P0
oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`070?ooo`030000003oool0oooo06<0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`1C0?ooo`00L03oool00`000000oooo
0?ooo`0K0?ooo`030000003oool0oooo03P0oooo00<0gia807Boo`3oool02`3oool00`000000oooo
0?ooo`1S0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0D`3oool00700oooo00<0
00000?ooo`3oool06`3oool00`000000oooo0?ooo`0X0?ooo`030?oog`2LB000000000L0000000<0
B9cO0?ooo`3oool0103oool00`2oM4P0W=oo0?ooo`0:0?ooo`030000003oool0oooo06<0oooo00<0
00000?ooo`3oool06`3oool00`000000oooo0?ooo`1B0?ooo`00L03oool00`000000oooo0?ooo`0K
0?ooo`030000003oool0oooo03T0oooo00<0oooO09a8M02oool02P3oool400000680oooo00<00000
0?ooo`3oool06`3oool2000005<0oooo001`0?ooo`030000003oool0oooo01`0oooo00<000000?oo
o`3oool0>@3oool00`3o_g@007Bo0?ooo`090?ooo`030000003oool0oooo06<0oooo00<000000?oo
o`3oool0703oool00`000000oooo0?ooo`1A0?ooo`00L03oool00`000000oooo0?ooo`0L0?ooo`03
0000003oool0oooo03T0oooo00<0oono07@0B02Lgol02@3oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool0oooo01d0oooo00<000000?ooo`3oool0D03oool00700oooo00<000000?ooo`3oool0
7@3oool00`000000oooo0?ooo`0h0?ooo`030?oo_`1d0000B9cO00T0oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0500oooo001`0?ooo`030000003o
ool0oooo01h0oooo00<000000?ooo`3oool0<P3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW018
0000B9cO00T0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0N0?ooo`030000003o
ool0oooo04l0oooo001`0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0<P3oool0
203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo2@3oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool0oooo01h0oooo00<000000?ooo`3oool0C`3oool00700oooo00<000000?ooo`3oool0
7`3oool00`000000oooo0?ooo`0b0?ooo`060?oog`2LB00000000000000007@0_ooo2P3oool00`00
0000oooo0?ooo`1S0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0CP3oool00700
oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo06<0
oooo00<000000?ooo`3oool07`3oool2000004l0oooo001`0?ooo`030000003oool0oooo0200oooo
00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo0200oooo
00<000000?ooo`3oool0C@3oool00700oooo00<000000?ooo`3oool0803oool00`000000oooo0?oo
o`110?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?oo
o`1<0?ooo`00L03oool00`000000oooo0?ooo`0Q0?ooo`800000@@3oool00`000000oooo0?ooo`1S
0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0C03oool00700oooo00<000000?oo
o`3oool08P3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo06<0oooo00<000000?oo
o`3oool08P3oool00`000000oooo0?ooo`1;0?ooo`00L03oool00`000000oooo0?ooo`0S0?ooo`03
0000003oool0oooo03h0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0R0?ooo`80
0000C03oool00700oooo00<000000?ooo`3oool08`3oool2000003l0oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo04X0oooo001`0?ooo`030000003o
ool0oooo02@0oooo00<000000?ooo`3oool0?@3oool3000006<0oooo00<000000?ooo`3oool0903o
ool00`000000oooo0?ooo`190?ooo`00L03oool00`000000oooo0?ooo`0U0?ooo`030000003oool0
oooo03`0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0
oooo04P0oooo001`0?ooo`030000003oool0oooo02D0oooo0P00000m0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`180?ooo`00L03oool00`000000
oooo0?ooo`0V0?ooo`800000?03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo02H0
oooo0P0000180?ooo`00L03oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo03X0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo04H0oooo
001`0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?oo
o`1S0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0A@3oool00700oooo00<00000
0?ooo`3oool0:@3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool0:@3oool00`000000oooo0?ooo`140?ooo`00L03oool00`000000oooo0?ooo`0Z0?oo
o`030000003oool0oooo03L0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0Z0?oo
o`030000003oool0oooo04<0oooo001`0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3o
ool0=P3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3o
ool0@P3oool00700oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0e0?ooo`030000
003oool0oooo06<0oooo00<000000?ooo`3oool0;03oool200000480oooo001`0?ooo`030000003o
ool0oooo02d0oooo0P00000e0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0;@3o
ool00`000000oooo0?ooo`100?ooo`00L03oool00`000000oooo0?ooo`0_0?ooo`030000003oool0
oooo0380oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0^0?ooo`800000@03oool0
0700oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0m0?ooo`00L03oool00`000000
oooo0?ooo`0a0?ooo`030000003oool0oooo0300oooo0`00001S0?ooo`030000003oool0oooo0340
oooo00<000000?ooo`3oool0?03oool00700oooo00<000000?ooo`3oool0<P3oool00`000000oooo
0?ooo`0_0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0<P3oool00`000000oooo
0?ooo`0k0?ooo`00L03oool00`000000oooo0?ooo`0c0?ooo`800000;`3oool00`000000oooo0?oo
o`1S0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0>P3oool00700oooo00<00000
0?ooo`3oool0=@3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool0=03oool00`000000oooo0?ooo`0i0?ooo`005@3oool00`3ookl0M000000000020000
000307Boo`3oool0oooo03T0oooo00<0_g@0000000000000100000000`18W=l0oooo0?ooo`0@0?oo
o`030000003oool0oooo02H0oooo00H0_g@000000000000000000018W03Oool:0?ooo`030000003o
ool0oooo02/0oooo00<000000?ooo`3oool0?03oool01P2oM000000000000000000004RL0=ooob40
oooo00<000000?ooo`3oool05@3oool00`2oM0000000000000040000000304RLg`3oool0oooo01H0
oooo0P00000O0?ooo`030?oo_`1d000000000080000000<0M;oo0?ooo`3oool04P3oool001D0oooo
00<0omnL04P00018W=l00P3oool00`2oM000B9cO0?ooo`0j0?ooo`030=nLB0000000000000<00000
00<004RL0=ooo`3oool0403oool00`000000oooo0?ooo`0W0?ooo`030?oOW0180000M;oo00d0oooo
0P00000[0?ooo`030000003oool0oooo03d0oooo00<0omnL04P0001d_ol08`3oool00`000000oooo
0?ooo`0F0?ooo`030=nLB0000000000000<0000000<004RL0=ooo`3oool0603oool00`000000oooo
0?ooo`0L0?ooo`030?oOW0180000B9cO0080oooo00<0_g@004RLg`3oool04P3oool001X0oooo00<0
oono07@0B02Lgol0>`3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo0100oooo00<0
00000?ooo`3oool09`3oool00`3ogi`0B00007Boo`0?0?ooo`030000003oool0oooo02P0oooo00<0
00000?ooo`3oool0?@3oool00`3ogi`0B00007Boo`0S0?ooo`030000003oool0oooo01L0oooo00<0
gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`0I0?ooo`<00000803oool00`3ookl0M01809cO
o`0B0?ooo`006P3oool00`3ookl0M00007Boo`0l0?ooo`030=nLB01d_ol0oooo01@0oooo00<00000
0?ooo`3oool09`3oool00`3ogi`0B00007Boo`0@0?ooo`030000003oool0oooo02L0oooo00<00000
0?ooo`3oool0?@3oool00`3ogi`0B00007Boo`0S0?ooo`030000003oool0oooo01P0oooo00<0gia8
07Boo`3oool0803oool00`000000oooo0?ooo`0M0?ooo`030?oo_`1d0000M;oo0180oooo00090?oo
o`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool0103oool00`3o_g@0000007Boo`0]0?oo
o`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool00`3oool00`2oM4P0W=oo0?ooo`0C0?oo
o`030000003oool0oooo01P0oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`020?oo
o`030?oOW0180000M;oo0140oooo0`00000V0?ooo`030000003oool0oooo03d0oooo00<0omnL04P0
001d_ol08`3oool00`000000oooo0?ooo`0I0?ooo`030;mdB02Lgol0oooo0200oooo0P00000M0?oo
o`030?noM0000000M;oo0180oooo000I0?ooo`040?oOW0180000001809cOocd0oooo00<0oooO09a8
M02oool04`3oool00`000000oooo0?ooo`0W0?ooo`030?oOW0180000M;oo01@0oooo00<000000?oo
o`3oool08`3oool00`000000oooo0?ooo`0m0?ooo`030?oOW0180000M;oo02<0oooo00<000000?oo
o`3oool06@3oool00`3ooml0W4Qd0;ooo`0R0?ooo`030000003oool0oooo01T0oooo00@0omnL04P0
000004P0W=oo4P3oool001L0oooo00D0okmd000000000000001809cOo`0o0?ooo`030?noM000M;l0
oooo0180oooo00<000000?ooo`3oool09`3oool00`3ogi`0B00007Boo`0E0?ooo`8000008`3oool0
0`000000oooo0?ooo`0m0?ooo`030?oOW0180000M;oo02<0oooo00<000000?ooo`3oool06P3oool0
0`3o_g@007Bo0?ooo`0R0?ooo`030000003oool0oooo01H0oooo00D0okmd000000000000001809cO
o`0C0?ooo`006@3oool00`2oM00007Bo0?ooo`0o0?ooo`030?oo_`1d04P0W=oo0180oooo00<00000
0?ooo`3oool09`3oool00`3ogi`0B00007Boo`0G0?ooo`8000008@3oool00`000000oooo0?ooo`0m
0?ooo`030?oOW0180000M;oo02<0oooo00<000000?ooo`3oool06P3oool00`3ookl0M01809cOo`0S
0?ooo`800000603oool00`2oM00007Bo0?ooo`0C0?ooo`006P3oool00`2oM000B9cO0?ooo`0n0?oo
o`030?oo_`1d0000B9cO0180oooo00<000000?ooo`3oool09`3oool00`3ogi`0B00007Boo`0I0?oo
o`8000007`3oool00`000000oooo0?ooo`0m0?ooo`030?oOW0180000M;oo02<0oooo00<000000?oo
o`3oool06P3oool00`3ookl0M00004RLg`0U0?ooo`<000005P3oool00`2oM000B9cO0?ooo`0B0?oo
o`005@3oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo03T0oooo00<0oknL09cOo`3o
ool00P3oool00`3ogi`0B00004RLg`0B0?ooo`030000003oool0oooo02L0oooo00<0omnL04P0001d
_ol06`3oool2000001d0oooo0`00000m0?ooo`030?oOW0180000M;oo02<0oooo00<000000?ooo`3o
ool05@3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO02P0oooo0P00000?0?ooo`03
0?oog`2LW;l0oooo0080oooo00<0gia80000M02oool04P3oool001H0oooo00H0okmd04RLg`3oool0
oono07@00000M;lj0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02ooolB0?ooo`030000
003oool0oooo02D0oooo00D0oooO09a800000000000007Boo`0M0?ooo`8000006`3oool00`000000
oooo0?ooo`0k0?ooo`050?oog`2LB00000000000001d_ol08`3oool00`000000oooo0?ooo`0E0?oo
o`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02ooolZ0?ooo`8000003P3oool01P3o_g@0B9cO
0?ooo`3ookl0M000001d_a<0oooo000G0?ooo`050=nLB000000000000000M02oool0>`3oool01P3o
oml0W4P0000000000000001d0;oooa<0oooo00<000000?ooo`3oool09`3oool00`3o_g@0000007Bo
o`0O0?ooo`8000006@3oool00`000000oooo0?ooo`0m0?ooo`030?noM0000000M;oo02<0oooo00<0
00000?ooo`3oool05P3oool01P3ooml0W4P0000000000000001d0;ooobd0oooo0P00000=0?ooo`05
0=nLB000000000000000M02oool04`3oool00700oooo00<000000?ooo`3oool0B`3oool2000001L0
oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1;0?ooo`8000008`3oool00700oooo
00<000000?ooo`3oool0C@3oool2000001D0oooo00<000000?ooo`3oool0H`3oool00`000000oooo
0?ooo`1=0?ooo`<00000803oool00700oooo00<000000?ooo`3oool0C`3oool300000180oooo00<0
00000?ooo`3oool0H`3oool00`000000oooo0?ooo`1@0?ooo`8000007P3oool00700oooo00<00000
0?ooo`3oool0DP3oool200000100oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1B
0?ooo`800000703oool00700oooo00<000000?ooo`3oool0E03oool2000000h0oooo00<000000?oo
o`3oool0H`3oool00`000000oooo0?ooo`1D0?ooo`8000006P3oool00700oooo00<000000?ooo`3o
ool0EP3oool2000000`0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1F0?ooo`80
0000603oool00700oooo00<000000?ooo`3oool0F03oool2000000X0oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`1H0?ooo`8000005P3oool00700oooo00<000000?ooo`3oool0FP3o
ool2000000P0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1J0?ooo`<000004`3o
ool00700oooo00<000000?ooo`3oool0G03oool3000000D0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`1M0?ooo`8000004@3oool00700oooo00<000000?ooo`3oool0G`3oool20000
00<0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1O0?ooo`8000003`3oool00700
oooo00<000000?ooo`3oool0H@3oool4000006D0oooo00<000000?ooo`3oool0H@3oool2000000d0
oooo003o00000;0000000@3oool000H0oooo00<000000?ooo`3oool00`3oool2000000D0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<0
00000?ooo`3oool02P3oool2000000/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`02000000P0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool0203oool00`000000oooo0000000<0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool01@3oool000h0oooo0`0000020?ooo`030000003oool0
oooo03l0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0S0?ooo`030000003oool0
oooo03h0oooo00<000000?ooo`3oool00P3oool3000003T0oooo00<000000?ooo`3oool08P3oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool04P3oool0
0140oooo0P00001M0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool01@3oool30000
05/0oooo00<000000?ooo`3oool0L03oool001<0oooo0P00001K0?ooo`030000003oool0oooo06@0
oooo00<000000?ooo`3oool0203oool2000005T0oooo00<000000?ooo`3oool0L03oool001D0oooo
1000001G0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool02P3oool2000005L0oooo
00<000000?ooo`3oool0L03oool001T0oooo0`00001D0?ooo`030000003oool0oooo06@0oooo00<0
00000?ooo`3oool0303oool2000005D0oooo00<000000?ooo`3oool0L03oool001`0oooo00<00000
0?ooo`3oool0D@3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo00h0oooo0P00001C
0?ooo`030000003oool0oooo0700oooo000M0?ooo`<00000D03oool00`000000oooo0?ooo`1T0?oo
o`030000003oool0oooo0100oooo0P00001A0?ooo`030000003oool0oooo0700oooo000P0?ooo`80
0000CP3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo0180oooo0`00001>0?ooo`03
0000003oool0oooo0700oooo000R0?ooo`800000C03oool00`000000oooo0?ooo`1T0?ooo`030000
003oool0oooo01D0oooo0P00001<0?ooo`030000003oool0oooo0700oooo000T0?ooo`800000BP3o
ool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo01L0oooo0P00001:0?ooo`030000003o
ool0oooo0700oooo000V0?ooo`800000B03oool00`000000oooo0?ooo`1T0?ooo`030000003oool0
oooo01T0oooo0P0000180?ooo`030000003oool0oooo0700oooo000X0?ooo`800000AP3oool00`00
0000oooo0?ooo`1T0?ooo`<000006`3oool2000004H0oooo00<000000?ooo`3oool0L03oool002X0
oooo0P0000140?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool07@3oool2000004@0
oooo00<000000?ooo`3oool0L03oool002`0oooo0P0000120?ooo`030000003oool0oooo06@0oooo
00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo0700oooo
000^0?ooo`800000@03oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo0200oooo0`00
00100?ooo`030000003oool0oooo0700oooo000`0?ooo`030000003oool0oooo03d0oooo00<00000
0?ooo`3oool0I03oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo03d0oooo00<00000
0?ooo`3oool0L03oool00340oooo0P00000m0?ooo`030000003oool0oooo06@0oooo00<000000?oo
o`3oool0903oool2000003d0oooo00<000000?ooo`3oool0L03oool003<0oooo00<000000?ooo`3o
ool0>P3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3o
ool0>P3oool00`000000oooo0?ooo`1`0?ooo`00=03oool3000003T0oooo00<000000?ooo`3oool0
I03oool00`000000oooo0?ooo`0W0?ooo`800000>P3oool00`000000oooo0?ooo`1`0?ooo`00=`3o
ool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0:@3o
ool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo0700oooo000h0?ooo`030000003oool0
oooo03D0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`0Z0?ooo`800000=`3oool0
0`000000oooo0?ooo`1`0?ooo`00>@3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo
06@0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo
0700oooo000j0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0I03oool00`000000
oooo0?ooo`0]0?ooo`800000=03oool00`000000oooo0?ooo`1`0?ooo`00>`3oool2000003<0oooo
00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo0340oooo
00<000000?ooo`3oool0L03oool003d0oooo0P00000a0?ooo`030000003oool0oooo06@0oooo0`00
000`0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0L03oool003l0oooo00<00000
0?ooo`3oool0;P3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo0340oooo00<00000
0?ooo`3oool0;`3oool00`000000oooo0?ooo`1`0?ooo`00@03oool00`000000oooo0?ooo`0]0?oo
o`030000003oool0oooo06@0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0^0?oo
o`030000003oool0oooo0700oooo00100?ooo`800000;P3oool00`000000oooo0?ooo`1T0?ooo`03
0000003oool0oooo03<0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`1`0?ooo`00
@@3oool2000002d0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`0d0?ooo`800000
;@3oool00`000000oooo0?ooo`1`0?ooo`00@`3oool00`000000oooo0?ooo`0Z0?ooo`030000003o
ool0oooo06@0oooo00<000000?ooo`3oool0=@3oool2000002`0oooo00<000000?ooo`3oool0L03o
ool004<0oooo0P00000[0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0=P3oool2
000002/0oooo00<000000?ooo`3oool0L03oool004D0oooo00<000000?ooo`3oool0:03oool00`00
0000oooo0?ooo`1T0?ooo`030000003oool0oooo03L0oooo0P00000Z0?ooo`030000003oool0oooo
0700oooo00150?ooo`800000:@3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo03P0
oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`1`0?ooo`00AP3oool2000002P0oooo
00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`0i0?ooo`800000:03oool00`000000oooo
0?ooo`1`0?ooo`00A`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo06@0oooo00<0
00000?ooo`3oool0>P3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0700oooo0018
0?ooo`8000009P3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo03/0oooo00<00000
0?ooo`3oool09@3oool00`000000oooo0?ooo`1`0?ooo`00B@3oool00`000000oooo0?ooo`0T0?oo
o`030000003oool0oooo06@0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`0T0?oo
o`030000003oool0oooo0700oooo001:0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3o
ool0I03oool3000003d0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`1`0?ooo`00
BP3oool2000002@0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`0n0?ooo`030000
003oool0oooo0280oooo00<000000?ooo`3oool0L03oool004/0oooo00<000000?ooo`3oool08P3o
ool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool08P3o
ool00`000000oooo0?ooo`1`0?ooo`00C03oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0
oooo06@0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0
oooo0700oooo001=0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0I03oool00`00
0000oooo0?ooo`100?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0L03oool004d0
oooo0P00000Q0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0@03oool00`000000
oooo0?ooo`0P0?ooo`030000003oool0oooo0700oooo001>0?ooo`030000003oool0oooo01l0oooo
00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo01l0oooo
00<000000?ooo`3oool0L03oool004l0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?oo
o`1T0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool07`3oool00`000000oooo0?oo
o`1`0?ooo`00C`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo06@0oooo00<00000
0?ooo`3oool0@P3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo0700oooo001@0?oo
o`030000003oool0oooo01d0oooo00<000000?ooo`3oool0D`3oool00`2oM0000000000000040000
000304RLg`3oool0oooo00L0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`0M0?oo
o`030000003oool0oooo0700oooo001@0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3o
ool0E03oool00`3OW4P0000000000003000000030018W03Oool0oooo00L0oooo00<000000?ooo`3o
ool0@`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0700oooo001A0?ooo`030000
003oool0oooo01`0oooo00<000000?ooo`3oool0E@3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nL
M02Lgol0oooo00L0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`0L0?ooo`030000
003oool0oooo0700oooo001A0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0EP3o
ool00`3OW4P0M;oo0?ooo`0;0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0703o
ool00`000000oooo0?ooo`1`0?ooo`00DP3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0
oooo05L0oooo00<0_gA809cOo`3oool02P3oool00`000000oooo0?ooo`150?ooo`030000003oool0
oooo01/0oooo00<000000?ooo`3oool0L03oool00580oooo00<000000?ooo`3oool06`3oool00`00
0000oooo0?ooo`1G0?ooo`030?oog`2LB7@0_ooo00X0oooo100000140?ooo`030000003oool0oooo
01/0oooo00<000000?ooo`3oool0L03oool005<0oooo00<000000?ooo`3oool06P3oool00`000000
oooo0?ooo`1H0?ooo`030?noM000M;l0oooo00T0oooo00<000000?ooo`3oool0AP3oool00`000000
oooo0?ooo`0J0?ooo`030000003oool0oooo0700oooo001C0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool0F03oool00`3ookl0M01809cOo`090?ooo`030000003oool0oooo04H0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`1`0?ooo`00E03oool00`000000oooo0?oo
o`0I0?ooo`030000003oool0oooo05P0oooo00<0oono07@00018W=l02@3oool00`000000oooo0?oo
o`160?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0L03oool005@0oooo00<00000
0?ooo`3oool06@3oool00`000000oooo0?ooo`1C0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL
04P00018W=l02@3oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo01T0oooo00<00000
0?ooo`3oool0L03oool005@0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`1C0?oo
o`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool90?ooo`030000003oool0oooo04L0oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`1`0?ooo`00E@3oool00`000000oooo0?oo
o`0H0?ooo`030000003oool0oooo05@0oooo00H0oooO09a80000000000000000M02oool:0?ooo`03
0000003oool0oooo04P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1`0?ooo`00
EP3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0
B03oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo0700oooo001F0?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`190?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool0L03oool005H0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`1T0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`1`0?ooo`00E`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo
06@0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo
0700oooo001G0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0I03oool00`000000
oooo0?ooo`1:0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0L03oool005L0oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo04X0oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`1`0?ooo`00F03oool00`000000oooo0?oo
o`0E0?ooo`030000003oool0oooo06@0oooo0`00001:0?ooo`030000003oool0oooo01H0oooo00<0
00000?ooo`3oool0L03oool005P0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`1T
0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`1`
0?ooo`00F03oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo06@0oooo00<000000?oo
o`3oool0B`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0700oooo001I0?ooo`03
0000003oool0oooo01@0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`1;0?ooo`03
0000003oool0oooo01D0oooo00<000000?ooo`3oool0L03oool005T0oooo00<000000?ooo`3oool0
503oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0
503oool00`000000oooo0?ooo`1`0?ooo`00F@3oool00`000000oooo0?ooo`0D0?ooo`030000003o
ool0oooo06@0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`0D0?ooo`030000003o
ool0oooo0700oooo001J0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0I03oool0
0`000000oooo0?ooo`1=0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0L03oool0
05X0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo
04d0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`1`0?ooo`00FP3oool00`000000
oooo0?ooo`0C0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0C@3oool00`000000
oooo0?ooo`0C0?ooo`030000003oool0oooo0700oooo001K0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo01<0oooo
00<000000?ooo`3oool0L03oool005/0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`1T0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`1`0?ooo`00F`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo06@0oooo00<00000
0?ooo`3oool0CP3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0700oooo001K0?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`1>0?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool0L03oool005`0oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`1T0?ooo`<00000CP3oool00`000000oooo0?ooo`0B0?ooo`03
0000003oool0oooo0700oooo001L0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
I03oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
L03oool005`0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`1T0?ooo`030000003o
ool0oooo04l0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`1`0?ooo`00G03oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0C`3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0700oooo001M0?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool0L03oool005d0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`1T0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool04@3oool00`000000
oooo0?ooo`1`0?ooo`00G@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo06@0oooo
00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0700oooo
001M0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0I03oool00`000000oooo0?oo
o`1@0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0L03oool005d0oooo00<00000
0?ooo`3oool0403oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo0500oooo00<00000
0?ooo`3oool0403oool00`000000oooo0?ooo`1`0?ooo`00G@3oool00`000000oooo0?ooo`0@0?oo
o`030000003oool0oooo06@0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo0700oooo001N0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool0I03oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool0L03oool005h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1T0?ooo`030000
003oool0oooo0540oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1`0?ooo`00GP3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0D@3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0700oooo001N0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`1A0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool0L03oool005l0oooo00<000000?ooo`3oool03P3oool00`00
0000oooo0?ooo`1T0?ooo`<00000DP3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
0700oooo001O0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0I03oool00`000000
oooo0?ooo`1B0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0L03oool005l0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo0580oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1`0?ooo`00G`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo0700oooo001O0?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool0I03oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool0L03oool005l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1T0?oo
o`030000003oool0oooo0580oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1`0?oo
o`00G`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3o
ool0D`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0700oooo001P0?ooo`030000
003oool0oooo00d0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`1C0?ooo`030000
003oool0oooo00d0oooo00<000000?ooo`3oool0L03oool00600oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`1G0?ooo`040?oog`2LB00004RL0=ooo`T0oooo00<000000?ooo`3oool0
D`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0700oooo001P0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool0E`3oool0103ooml0W4P00018W03Oool90?ooo`030000
003oool0oooo05<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1`0?ooo`00H03o
ool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo05L0oooo00@0oooO09a80000B9`0gooo
2@3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
L03oool00600oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1C0?ooo`030;md0000
0000000000D0000000<0M;oo0?ooo`3oool01P3oool00`000000oooo0?ooo`1C0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool0L03oool00600oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`1C0?ooo`080?oOW01dW=l0oooo0?ooo`3ooml0W4P00018W03Oool90?ooo`03
0000003oool0oooo05<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1`0?ooo`00
H03oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo05@0oooo00L0gia807Boo`3oool0
oooO09a80000B9`0gooo00T0oooo1000001B0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool0L03oool00640oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1D0?ooo`07
0?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`090?ooo`030000003oool0oooo05@0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`1`0?ooo`00H@3oool00`000000oooo0?ooo`0<0?oo
o`030000003oool0oooo05D0oooo00H0omnL04Qd_`3ooml0W4P00018W03Oool90?ooo`030000003o
ool0oooo05@0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1`0?ooo`00H@3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo05H0oooo00D0gia807Bog`2LB00004RL0=oo
o`090?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`1`0?ooo`00H@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo05H0oooo00D0oooO
09a8B018B00004RL0=ooo`090?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`1`0?ooo`00H@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo05L0oooo00@0omnL04P00000B9`0gooo2@3oool00`000000oooo0?ooo`1D0?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool0L03oool00640oooo00<000000?ooo`3oool0303oool0
0`000000oooo0?ooo`1H0?ooo`030=nLB000B9`0gooo00T0oooo00<000000?ooo`3oool0E03oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0700oooo001Q0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool0L03oool00640oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`1T0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`1`0?ooo`00HP3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo06@0oooo
00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0700oooo
001R0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?oo
o`1E0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0L03oool00680oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo05D0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`1`0?ooo`00HP3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo06@0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo0700oooo001R0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool0I03oool3000005D0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1`0?ooo`00
HP3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0
E@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0700oooo001R0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`1E0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool0L03oool00680oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`1T0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`1`0?ooo`00H`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
06@0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
0700oooo001S0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0I03oool00`000000
oooo0?ooo`1F0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0L03oool006<0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo05H0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`1`0?ooo`00H`3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo0700oooo001S0?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool0I03oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool0L03oool006<0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`1T0?oo
o`030000003oool0oooo05H0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`1`0?oo
o`00H`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3o
ool0EP3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0700oooo001T0?ooo`030000
003oool0oooo00T0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`1F0?ooo`030000
003oool0oooo00X0oooo00<000000?ooo`3oool0L03oool006@0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo05H0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`1`0?ooo`00I03oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo06@0oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo0700oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-3.30486, -5.00005, \
0.0230037, 0.0564177}}]
}, Closed]],

Cell["\<\
This technique will usually work for most \"normal\" functions, but may not \
be sufficient for highly oscillatory functions or functions which resemble \
pulse trains.  Here, for example, is a function for which the unguided \
sampling algorithm fails miserably.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[\((1 + Tanh[x\^2\ Sin[2  x]])\)/2, {x, 0, 100}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00952381 0.0147151 0.588604 [
[.21429 .00222 -8 -15.6875 ]
[.21429 .00222 8 0 ]
[.40476 .00222 -8 -15.6875 ]
[.40476 .00222 8 0 ]
[.59524 .00222 -8 -15.6875 ]
[.59524 .00222 8 0 ]
[.78571 .00222 -8 -15.6875 ]
[.78571 .00222 8 0 ]
[.97619 .00222 -11 -15.6875 ]
[.97619 .00222 11 0 ]
[.01131 .13244 -19 -7.84375 ]
[.01131 .13244 0 7.84375 ]
[.01131 .25016 -19 -7.84375 ]
[.01131 .25016 0 7.84375 ]
[.01131 .36788 -19 -7.84375 ]
[.01131 .36788 0 7.84375 ]
[.01131 .4856 -19 -7.84375 ]
[.01131 .4856 0 7.84375 ]
[.01131 .60332 -10 -7.84375 ]
[.01131 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .01472 m
.21429 .02097 L
s
gsave
.21429 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.40476 .01472 m
.40476 .02097 L
s
gsave
.40476 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
.59524 .01472 m
.59524 .02097 L
s
gsave
.59524 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(60) show
1.000 setlinewidth
grestore
.78571 .01472 m
.78571 .02097 L
s
gsave
.78571 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(80) show
1.000 setlinewidth
grestore
.97619 .01472 m
.97619 .02097 L
s
gsave
.97619 .00222 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .01472 m
.07143 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.16667 .01472 m
.16667 .01847 L
s
.2619 .01472 m
.2619 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.35714 .01472 m
.35714 .01847 L
s
.45238 .01472 m
.45238 .01847 L
s
.5 .01472 m
.5 .01847 L
s
.54762 .01472 m
.54762 .01847 L
s
.64286 .01472 m
.64286 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.7381 .01472 m
.7381 .01847 L
s
.83333 .01472 m
.83333 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.92857 .01472 m
.92857 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .13244 m
.03006 .13244 L
s
gsave
.01131 .13244 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.02381 .25016 m
.03006 .25016 L
s
gsave
.01131 .25016 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.02381 .36788 m
.03006 .36788 L
s
gsave
.01131 .36788 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.02381 .4856 m
.03006 .4856 L
s
gsave
.01131 .4856 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.02381 .60332 m
.03006 .60332 L
s
gsave
.01131 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04415 m
.02756 .04415 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .10301 m
.02756 .10301 L
s
.02381 .16187 m
.02756 .16187 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .30902 m
.02756 .30902 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .45617 m
.02756 .45617 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .30902 m
.02499 .31011 L
.02605 .31645 L
.02729 .33517 L
.02846 .36631 L
.03056 .44413 L
.03171 .48409 L
.03279 .5116 L
.03395 .52825 L
.03524 .52971 L
.03633 .51204 L
.0375 .45508 L
.03816 .39258 L
.03887 .2935 L
.04017 .10455 L
.04144 .02969 L
.04262 .0168 L
.04386 .01496 L
.04456 .01479 L
.04521 .01475 L
.0464 .01472 L
.04765 .01472 L
.04878 .01472 L
.04983 .01473 L
.05098 .0148 L
.05221 .01696 L
.05353 .19544 L
.05427 .55139 L
.05496 .60087 L
.05622 .60331 L
.05686 .60332 L
.05757 .60332 L
.05878 .60332 L
.06008 .60332 L
.06131 .60332 L
.06244 .60332 L
.06373 .60332 L
.0649 .60332 L
.06555 .60332 L
.06625 .60332 L
.06752 .6033 L
.06872 .27158 L
.06998 .01472 L
.07106 .01472 L
.07225 .01472 L
.07296 .01472 L
.07364 .01472 L
.07492 .01472 L
.07563 .01472 L
.07629 .01472 L
Mistroke
.07702 .01472 L
.07779 .01472 L
.07846 .01472 L
.07919 .01472 L
.08051 .01472 L
.08179 .01472 L
.08297 .01473 L
.08422 .60328 L
.08493 .60332 L
.08557 .60332 L
.08681 .60332 L
.08799 .60332 L
.09084 .60332 L
.09349 .60332 L
.09482 .60332 L
.09607 .60332 L
.09737 .60332 L
.09806 .60332 L
.0988 .01849 L
.09951 .01472 L
.10027 .01472 L
.10107 .01472 L
.10182 .01472 L
.10252 .01472 L
.10315 .01472 L
.10389 .01472 L
.10458 .01472 L
.10574 .01472 L
.10698 .01472 L
.10815 .01472 L
.10922 .01472 L
.11048 .01472 L
.11119 .01472 L
.11184 .01472 L
.11305 .01472 L
.11433 .60332 L
.11554 .60332 L
.11685 .60332 L
.11922 .60332 L
.12181 .60332 L
.12463 .60332 L
.12581 .60332 L
.12706 .60332 L
.12825 .60332 L
.12932 .01472 L
.13054 .01472 L
.13186 .01472 L
.13311 .01472 L
.13426 .01472 L
.13532 .01472 L
Mistroke
.1363 .01472 L
.13739 .01472 L
.13855 .01472 L
.13977 .01472 L
.14093 .01472 L
.14197 .01472 L
.1431 .01472 L
.14375 .60332 L
.14443 .60332 L
.14565 .60332 L
.14842 .60332 L
.15343 .60332 L
.15558 .60332 L
.15677 .60332 L
.15789 .60332 L
.15852 .01577 L
.15921 .01472 L
.16042 .01472 L
.16165 .01472 L
.16279 .01472 L
.16399 .01472 L
.16528 .01472 L
.1665 .01472 L
.16762 .01472 L
.16887 .01472 L
.16957 .01472 L
.17022 .01472 L
.17092 .01472 L
.17155 .01472 L
.17227 .01472 L
.17296 .01472 L
.17427 .60332 L
.17493 .60332 L
.17566 .60332 L
.17815 .60332 L
.18083 .60332 L
.18376 .60332 L
.18508 .60332 L
.18647 .60332 L
.18779 .60332 L
.18898 .01472 L
.19021 .01472 L
.1909 .01472 L
.19154 .01472 L
.1927 .01472 L
.19394 .01472 L
.19504 .01472 L
.19623 .01472 L
.19735 .01472 L
.19837 .01472 L
Mistroke
.19957 .01472 L
.20087 .01472 L
.20209 .01472 L
.20323 .01472 L
.20443 .60332 L
.20572 .60332 L
.20805 .60332 L
.21337 .60332 L
.21474 .60332 L
.21625 .60332 L
.21693 .60332 L
.21767 .60332 L
.21898 .01472 L
.21965 .01472 L
.22037 .01472 L
.22168 .01472 L
.22295 .01472 L
.22412 .01472 L
.22538 .01472 L
.22658 .01472 L
.22765 .01472 L
.22882 .01472 L
.23009 .01472 L
.23073 .01472 L
.23143 .01472 L
.23268 .01472 L
.23385 .60332 L
.23496 .60332 L
.23613 .60332 L
.23822 .60332 L
.24294 .60332 L
.2636 .60332 L
.26852 .60332 L
.27112 .60332 L
.27386 .60332 L
.27626 .60332 L
.27692 .60332 L
.27763 .60332 L
.2789 .01472 L
.28015 .01472 L
.28133 .01472 L
.2824 .01472 L
.28356 .01472 L
.28479 .01472 L
.28595 .01472 L
.28699 .01472 L
.28811 .01472 L
.28934 .01472 L
.29064 .01472 L
.29175 .01472 L
Mistroke
.29297 .01472 L
.29411 .60332 L
.29515 .60332 L
.29748 .60332 L
.30169 .60332 L
.30404 .60332 L
.30652 .60332 L
.30782 .60332 L
.30855 .01472 L
.30925 .01472 L
.31046 .01472 L
.31177 .01472 L
.31292 .01472 L
.31419 .01472 L
.31551 .01472 L
.31674 .01472 L
.3179 .01472 L
.31894 .01472 L
.32015 .01472 L
.3213 .01472 L
.32251 .01472 L
.3232 .60332 L
.32384 .60332 L
.32653 .60332 L
.33135 .60332 L
.33405 .60332 L
.33538 .60332 L
.33661 .60332 L
.33729 .60332 L
.33802 .01472 L
.33875 .01472 L
.33955 .01472 L
.34022 .01472 L
.34094 .01472 L
.34224 .01472 L
.34349 .01472 L
.34463 .01472 L
.34595 .01472 L
.34719 .01472 L
.34834 .01472 L
.3494 .01472 L
.35056 .01472 L
.35179 .01472 L
.35303 .60332 L
.35373 .60332 L
.35436 .60332 L
.35677 .60332 L
.36225 .60332 L
.38334 .60332 L
.38847 .60332 L
Mistroke
.39331 .60332 L
.39599 .60332 L
.39672 .60332 L
.3975 .60332 L
.39816 .01472 L
.39888 .01472 L
.40014 .01472 L
.40085 .01472 L
.4015 .01472 L
.40275 .01472 L
.40394 .01472 L
.40503 .01472 L
.40624 .01472 L
.40751 .01472 L
.40869 .01472 L
.4098 .01472 L
.41081 .01472 L
.41192 .01472 L
.41309 .60332 L
.41429 .60332 L
.41556 .60332 L
.41785 .60332 L
.42305 .60332 L
.42546 .60332 L
.42667 .60332 L
.42735 .60332 L
.428 .01472 L
.4287 .01472 L
.42934 .01472 L
.43008 .01472 L
.43078 .01472 L
.43201 .01472 L
.43271 .01472 L
.43335 .01472 L
.43463 .01472 L
.43582 .01472 L
.4369 .01472 L
.43806 .01472 L
.43933 .01472 L
.43997 .01472 L
.44067 .01472 L
.44191 .01472 L
.44307 .60332 L
.4442 .60332 L
.44543 .60332 L
.44767 .60332 L
.45185 .60332 L
.45423 .60332 L
.45552 .60332 L
.45674 .60332 L
Mistroke
.45778 .01472 L
.45889 .01472 L
.46012 .01472 L
.46126 .01472 L
.46254 .01472 L
.46375 .01472 L
.46484 .01472 L
.46601 .01472 L
.4673 .01472 L
.46794 .01472 L
.46866 .01472 L
.46991 .01472 L
.47109 .01472 L
.47227 .01472 L
.47339 .60332 L
.4744 .60332 L
.4755 .60332 L
.48022 .60332 L
.48291 .60332 L
.48544 .60332 L
.4867 .60332 L
.48742 .60332 L
.48807 .01472 L
.48875 .01472 L
.48949 .01472 L
.49026 .01472 L
.49098 .01472 L
.49228 .01472 L
.4935 .01472 L
.4946 .01472 L
.49579 .01472 L
.49709 .01472 L
.49775 .01472 L
.49847 .01472 L
.49975 .01472 L
.50094 .01472 L
.50219 .01472 L
.50289 .60332 L
.50352 .60332 L
.50591 .60332 L
.51133 .60332 L
.51378 .60332 L
.51506 .60332 L
.51576 .60332 L
.51642 .60332 L
.51761 .01472 L
.51891 .01472 L
.52002 .01472 L
.52121 .01472 L
.52245 .01472 L
Mistroke
.52359 .01472 L
.52485 .01472 L
.52549 .01472 L
.52619 .01472 L
.52736 .01472 L
.52862 .01472 L
.52981 .01472 L
.5309 .01472 L
.53197 .01472 L
.53296 .60332 L
.53405 .60332 L
.53521 .60332 L
.53974 .60332 L
.54233 .60332 L
.54375 .60332 L
.54508 .60332 L
.54621 .60332 L
.54742 .01472 L
.54811 .01472 L
.54877 .01472 L
.55 .01472 L
.5507 .01472 L
.55133 .01472 L
.55207 .01472 L
.55276 .01472 L
.55399 .01472 L
.55468 .01472 L
.55531 .01472 L
.55601 .01472 L
.55666 .01472 L
.55737 .01472 L
.55813 .01472 L
.55893 .01472 L
.55969 .01472 L
.56037 .01472 L
.56111 .01472 L
.56236 .01472 L
.56353 .60332 L
.56481 .60332 L
.56619 .60332 L
.57159 .60332 L
.57418 .60332 L
.57545 .60332 L
.57664 .60332 L
.57771 .01472 L
.57885 .01472 L
.58012 .01472 L
.58126 .01472 L
.58247 .01472 L
.58359 .01472 L
Mistroke
.58482 .01472 L
.58614 .01472 L
.5868 .01472 L
.58752 .01472 L
.58883 .01472 L
.59011 .01472 L
.59129 .01472 L
.59238 .60332 L
.59341 .60332 L
.59576 .60332 L
.60051 .60332 L
.60305 .60332 L
.60444 .60332 L
.60573 .60332 L
.60684 .60332 L
.60802 .01472 L
.60933 .01472 L
.61053 .01472 L
.6112 .01472 L
.61191 .01472 L
.61319 .01472 L
.6139 .01472 L
.61457 .01472 L
.6153 .01472 L
.61608 .01472 L
.61675 .01472 L
.61749 .01472 L
.61818 .01472 L
.61882 .01472 L
.62003 .01472 L
.62133 .01472 L
.62257 .60332 L
.62328 .60332 L
.62394 .60332 L
.6267 .60332 L
.63164 .60332 L
.6344 .60332 L
.63576 .60332 L
.63701 .60332 L
.63772 .01472 L
.63846 .01472 L
.63921 .01472 L
.64003 .01472 L
.64071 .01472 L
.64145 .01472 L
.64214 .01472 L
.64277 .01472 L
.64395 .01472 L
.64521 .01472 L
.6464 .01472 L
Mistroke
.64748 .01472 L
.64874 .01472 L
.64946 .01472 L
.65012 .01472 L
.65133 .01472 L
.65261 .60332 L
.65384 .60332 L
.65516 .60332 L
.65754 .60332 L
.66296 .60332 L
.66418 .60332 L
.66549 .60332 L
.66674 .60332 L
.66788 .01472 L
.66909 .01472 L
.67036 .01472 L
.67167 .01472 L
.67242 .01472 L
.67311 .01472 L
.67386 .01472 L
.67456 .01472 L
.67588 .01472 L
.67662 .01472 L
.67731 .01472 L
.67806 .01472 L
.67885 .01472 L
.68016 .01472 L
.68138 .01472 L
.68208 .60332 L
.68272 .60332 L
.68417 .60332 L
.6891 .60332 L
.6917 .60332 L
.69313 .60332 L
.69448 .60332 L
.69576 .60332 L
.69648 .60332 L
.69715 .01472 L
.69832 .01472 L
.69956 .01472 L
.70073 .01472 L
.70181 .01472 L
.703 .01472 L
.70428 .01472 L
.7055 .01472 L
.70619 .01472 L
.70684 .01472 L
.70816 .01472 L
.70889 .01472 L
.70957 .01472 L
Mistroke
.71071 .01472 L
.71195 .01472 L
.71325 .60332 L
.71446 .60332 L
.72553 .60332 L
.74566 .60332 L
.78472 .60332 L
.78596 .60332 L
.78665 .60332 L
.78729 .01472 L
.78852 .01472 L
.7897 .01472 L
.7909 .01472 L
.79221 .01472 L
.7929 .01472 L
.79364 .01472 L
.79434 .01472 L
.79497 .01472 L
.79571 .01472 L
.7964 .01472 L
.79715 .01472 L
.79795 .01472 L
.79864 .01472 L
.79928 .01472 L
.79999 .01472 L
.80075 .01472 L
.80143 .01472 L
.80218 .60332 L
.80288 .60332 L
.80354 .60332 L
.80607 .60332 L
.81112 .60332 L
.81359 .60332 L
.81468 .60332 L
.81586 .60332 L
.81707 .01472 L
.81823 .01472 L
.81887 .01472 L
.81956 .01472 L
.82079 .01472 L
.8215 .01472 L
.82226 .01472 L
.82298 .01472 L
.82364 .01472 L
.8249 .01472 L
.82553 .01472 L
.82623 .01472 L
.82753 .01472 L
.82874 .01472 L
.83006 .01472 L
Mistroke
.83073 .01472 L
.83146 .01472 L
.8327 .60332 L
.83339 .60332 L
.83402 .60332 L
.83643 .60332 L
.84573 .60332 L
.86664 .60332 L
.86908 .60332 L
.87174 .60332 L
.87295 .60332 L
.87424 .60332 L
.87534 .60332 L
.87655 .01472 L
.87727 .01472 L
.87793 .01472 L
.8787 .01472 L
.87942 .01472 L
.8807 .01472 L
.88143 .01472 L
.8821 .01472 L
.88339 .01472 L
.88455 .01472 L
.88519 .01472 L
.88589 .01472 L
.88715 .01472 L
.88834 .01472 L
.88946 .01472 L
.89068 .01472 L
.89199 .60332 L
.89265 .60332 L
.89336 .60332 L
.89466 .60332 L
.8971 .60332 L
.90624 .60332 L
.90741 .01472 L
.90863 .01472 L
.90927 .01472 L
.90997 .01472 L
.9112 .01472 L
.91236 .01472 L
.91361 .01472 L
.91478 .01472 L
.91585 .01472 L
.91704 .01472 L
.91834 .01472 L
.91956 .01472 L
.92069 .01472 L
.92193 .60332 L
.92263 .60332 L
Mistroke
.92329 .60332 L
.92604 .60332 L
.93057 .60332 L
.9328 .60332 L
.93486 .60332 L
.93601 .60332 L
.93723 .01472 L
.93846 .01472 L
.93916 .01472 L
.93981 .01472 L
.94095 .01472 L
.94199 .01472 L
.94313 .01472 L
.94434 .01472 L
.94536 .01472 L
.94648 .01472 L
.94753 .01472 L
.94851 .01472 L
.94938 .01472 L
.95035 .01472 L
.95136 .60332 L
.9523 .60332 L
.95676 .60332 L
.96083 .60332 L
.96264 .60332 L
.9636 .60332 L
.96462 .60332 L
.9656 .60332 L
.96649 .01472 L
.96739 .01472 L
.96823 .01472 L
.9692 .01472 L
.97022 .01472 L
.97109 .01472 L
.97205 .01472 L
.97309 .01472 L
.97419 .01472 L
.97524 .01472 L
.97619 .01472 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00J@3oool00`2oM0000000000000040000000904RLg`3oool0oooo0?ooo`3ogi`0B0000000
000007@0_ooo03d0oooo00@0oooO09a80000B9`0gooo0`3oool01@3ogi`0B0000000000007@0_ooo
03/0oooo00D0omnL04P000000000000007Boo`040?ooo`050?oOW018000000000000M02oool0>`3o
ool01@3OW4P00000000000000000B9cO00@0oooo00D0omnL04P000000000001d0;ooo`0g0?ooo`06
0;md00000000000000000000B9`0gooo0`3oool01@3ogi`0B0000000000007@0_ooo00@0oooo00D0
omnL04P000000000001d0;ooo`0>0?ooo`00JP3oool00`3OW4P00000000000030000000:0018W03O
ool0oooo0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oooc`0oooo00@0oooO09a80000B9`0gooo0P3o
ool01`3ookl0M00007Boo`3oool0okmd0000M02oool0>@3oool01`3ogi`0B00004RLg`3oool0oono
07@0001d_ol00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0>@3oool01`3OW4P00000
07Boo`3oool0oooO09a80018W=l00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0=`3o
ool00`3ogi`0B00007Boo`040?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`07
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0=0?ooo`00J`3oool00`3OW4P0B9cO0?ooo`020?oo
o`060=nLM02Lgol0oooo0;md0000B9`0gooo0P3oool00`2oM00007Bo0?ooo`0k0?ooo`040?oog`2L
B00004RL0=ooo`80oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`0g0?ooo`030?oo
g`2LB00007Bo00<0oooo00L0okmd0018W03Oool0oooo0;md0000B9`0gooo0080oooo00<0_g@0001d
_`3oool0=`3oool00`3ooml0W4P0001d_`030?ooo`070?noM000B9`0gooo0?ooo`2oM00004RL0=oo
o`020?ooo`030;md0000M;l0oooo03H0oooo00<0omnL04P0001d_ol0103oool00`2oM00004RL0=oo
o`020?ooo`070;md0000M;l0oooo0?ooo`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00`0
oooo001/0?ooo`030=nLB01d_ol0oooo00<0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d
0;ooo`0g0?ooo`030;md00000000000000D0000000@0M;oo0?oo_`1d0000B9cO0`3oool00`3OW4P0
001d0;ooo`0g0?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0oono07@00018W=l3
0?ooo`030=nLB00007@0_ooo03P0oooo00<0_g@0001d_`3oool00P3oool01P2oM00004RL0=ooo`3o
okl0M00004RLg`<0oooo00<0gia80000M02oool0=P3oool00`3ogi`0B00007Boo`030?ooo`030?oo
_`1d0000B9cO00<0oooo00H0gia80000M02oool0oono07@00018W=l30?ooo`030=nLB00007@0_ooo
00`0oooo001]0?ooo`030;mdB02Lgol0oooo0080oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`0g0?ooo`0<0?oOW01dW=l0oooo0?ooo`3ooml0W4P00018W03Oool0oooo0?oOW0180000
M;oo0`3oool00`3o_g@0001809cOo`0g0?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02L
gol0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo03P0oooo00/0omnL04P0M02oool0oooo0;md
00000000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo03H0oooo00<0omnL04P0
001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3o
ool00`3o_g@0001809cOo`0<0?ooo`00K@3oool00`3ooml0W4Qd0;ooo`020?ooo`030?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol0>03oool02`3OW4P0M;oo0?ooo`3ooml0W4P00018W03Oool0
oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0=`3oool00`3ogi`0B00007Boo`030?oo
o`060;md000004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0i0?ooo`050?oo_`1d
B4P0B00000000018W=l00P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo03H0oooo
00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo0?oOW018
0000M;oo0`3oool00`3o_g@0001809cOo`0<0?ooo`00KP3oool01`3o_g@007Bo0?ooo`3oool0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`0h0?ooo`0;0?oog`2LB7@0_ooo0?oog`2LB00004RL
0=ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0g0?ooo`0<0?oo_`1d00000000
04RLg`3oool0okmd00000000M;l0oooo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0i0?oo
o`040?oOW01800000000001d_`<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0f
0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool01P3o_g@0001809cOo`3o
gi`0B00007Boo`<0oooo00<0okmd0000B02Lgol0303oool006h0oooo00L0oono07@0B02Lgol0oooo
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0>@3oool02P3ogi`0B7Bo0?oog`2LB00004RL
0=ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo03P0oooo00<0_g@00018W02oW4P0
0P0000001P00B9`0gooo0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol0>03oool01P3o
gi`0B0000000M02oool0okmd001d_`80oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`0f0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool01P3o_g@0001809cO
o`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol0303oool006h0oooo00L0oono07@00018W=l0
oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0>P3oool02@3OW4P0M;oO09a80000B9`0
gooo0?ooo`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo03P0oooo00<0omnL04P0001d_ol0
1@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo03P0oooo00<0_g@00000M02oool0
0P3oool01P2oM000B9cO0?ooo`3ookl0M00004RLg`<0oooo00<0gia80000M02oool0=P3oool00`3o
gi`0B00007Boo`030?ooo`030?oo_`1d0000B9cO00<0oooo00H0gia80000M02oool0oono07@00018
W=l30?ooo`030=nLB00007@0_ooo00`0oooo001Y0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL
04P00018W=l00P3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo03X0oooo00D0oooO
09a8B018B00004RL0=ooo`020?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0>@3o
ool00`3OW4P0001809cOo`050?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0>03o
ool00`2oM00004RL0=ooo`020?ooo`070=nLB000M;l0oooo0?ooo`3OW4P004RL0=ooo`020?ooo`03
0;md0000M;l0oooo03H0oooo00<0omnL04P0001d_ol0103oool00`3OW4P004RL0=ooo`020?ooo`07
0;md0000M;l0oooo0?ooo`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00`0oooo001Y0?oo
o`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`0l0?ooo`040?oOW018000004RL0=ooo`80oooo00L0oono07@0001d_ol0oooo0?no
M00007@0_ooo03/0oooo00<0okmd00000000M;l0103oool01`3ookl0M00007Boo`3oool0okmd0000
M02oool0>@3oool01`3o_g@0000007Boo`3oool0oono07@0001d_ol00P3oool01`3ookl0M00007Bo
o`3oool0okmd0000M02oool0=@3oool01@3ooml0W4P0000000000000M;oo00@0oooo00L0oono07@0
001d_ol0oooo0?noM00007@0_ooo0080oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00d0
oooo001Z0?ooo`060?oog`2LB00000000000000007@0_ooo103oool01@3ookl0M0000000000007@0
_ooo03h0oooo00<0gia80018W03Oool00`3oool01@3ookl0M0000000000007@0_ooo03d0oooo00D0
oooO09a800000000001d0;ooo`020?ooo`050?oo_`1d000000000000M02oool0>`3oool01@3o_g@0
0000000000000000B9cO00@0oooo00D0oono07@000000000001d0;ooo`0h0?ooo`030?noM0000000
M;oo00D0oooo00D0oono07@000000000001d0;ooo`040?ooo`050?oo_`1d000000000000M02oool0
3P3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool002T0oooo
00<000000?ooo`3oool0o`3ooon50?ooo`00:@3oool00`000000oooo0?ooo`3o0?ooohD0oooo000Y
0?ooo`030000003oool0oooo0?l0ooooQ@3oool002T0oooo00<000000?ooo`3oool0o`3ooon50?oo
o`00:@3oool00`000000oooo0?ooo`3o0?ooohD0oooo000Y0?ooo`030000003oool0oooo0?l0oooo
Q@3oool00200ooooo`00002100000100oooo000Y0?ooo`030000003oool0oooo00@0oooo00D00000
0?ooo`3oool0oooo000000050?ooo`8000000`3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`0000000`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00L000000?ooo`3oool0oooo0000003oool0000000D0oooo00<000000?ooo`3o
ool00`3oool01`000000oooo0?ooo`3oool000000?ooo`000000103oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0000000h0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00L000000?ooo`3oool0oooo0000003oool0000000D0
oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`3oool2000000D0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0000000@0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool000000080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo00000080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00l0oooo0P000004
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool000000080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`050000003oool0oooo0?ooo`0000006@3oool002T0oooo00<000000?ooo`3oool0103o
ool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00L000000?ooo`3oool0oooo0000003oool00000
00@0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool010000000oooo0?ooo`0000020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
000000@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`800000103oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
1@000000oooo0?ooo`3oool0000001T0oooo000Y0?ooo`030000003oool0oooo00@0oooo00D00000
0?ooo`3oool0oooo000000050?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`040?ooo`050000003oool0
oooo0?ooo`0000001@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3o
ool0000000D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo0000
00050?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0K0?ooo`00:@3oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000001@3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3o
ool002T0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000D0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y
0?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool0
0`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo0P000005
0?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool0103oool0
1@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00@0oooo00D00000
0?ooo`3oool0oooo000000050?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`040?ooo`050000003oool0
oooo0?ooo`0000001@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3o
ool0000000D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo0000
00050?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0K0?ooo`00:@3oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000001@3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3o
ool002T0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000D0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y
0?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool0
02T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0K0?ooo`00:@3oool2000000@0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool0
02T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool06`3oool002T0oooo0P0000040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`003@3oool0
1@3ogi`0B0000000000007@0_ooo00<0oooo00D0okmd0000001d_ol0oooo0;md00060000000304RL
g`3oool0oooo00H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo01/0oooo000<0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`03
0?noM0000000M;oo0080oooo00<0gia80000000000000`0000000`00B9`0gooo0?ooo`060?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00
303oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0gia804RLg`3oool0
0P3oool00`3OW7@0W=oo0?ooo`060?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0K0?ooo`002`3oool00`3ookl0M00004RLg`030?ooo`030=nLB000
07@0_ooo00P0oooo00<0gia807Boo`3oool02P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool06`3oool000/0oooo00<0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`090?ooo`030;mdB02Lgol0oooo00T0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000;0?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol02@3oool00`3ooml0W4Qd0;ooo`090?ooo`<000000`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000;0?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol02P3oool00`3o_g@007Bo0?ooo`080?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`002`3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0oono07@0B02Lgol0203oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool000/0oooo00<0
oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0:0?ooo`030?oo_`1d0000B9cO00P0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000<
0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01@3oool00`3o_i`0W=oo0?ooo`02
0?ooo`030?oOW0180000B9cO00P0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo01/0oooo000<0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oo
o`060?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool80?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`003@3oool01@3ookl0
M0000000000007@0_ooo00P0oooo00H0oooO09a80000000000000000M02oool90?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool0
02T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool06`3oool002T0oooo0P0000040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool0
02T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool2000000@0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool0
02T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo0P0000040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool0
02T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0K0?ooo`003@3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00<0okmd0000001d
_ol01@3oool0103ooml0W4P00018W03Oool80?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00303oool01`3ookl0M00007Boo`3oool0okmd
0000M02oool00P3oool00`3o_g@0000007Boo`050?ooo`040?oog`2LB00004RL0=ooo`P0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000<
0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02@3oool0103ooml0W4P00018W03O
ool80?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0K0?ooo`002`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00D0oooo00<0_g@0
0000000000001@0000000`1d_ol0oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`002`3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00D0oooo00P0omnL07BLg`3oool0oooo0?oog`2LB00004RL0=ooo`P0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo
000;0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01`3OW4P0M;oo0?oo
o`3ooml0W4P00018W03Oool0203oool3000000<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0K0?ooo`002`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo
00L0oooO09a8M02oool0oooO09a80000B9`0gooo00P0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000;0?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol01`3oool01P3ogi`0B7Bo0?oog`2LB00004RL0=ooo`P0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000;0?oo
o`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool01@3OW4P0M;oO09a80000B9`0
gooo00P0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo01/0oooo000<0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0203oool01@3o
oml0W4Q804Q80000B9`0gooo00P0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo01/0oooo000<0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oo
o`0:0?ooo`040?oOW018000004RL0=ooo`P0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo01/0oooo000=0?ooo`050?oo_`1d000000000000M02oool0
303oool00`3OW4P004RL0=ooo`080?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo
000Y0?ooo`800000103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K
0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo
000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0K0?ooo`00:@3oool2000000@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K
0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool06`3oool002T0oooo0P0000040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool06`3oool002T0oooo0P0000040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
6`3oool002T0oooo0P0000040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3o
ool002T0oooo0P0000040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool0
02T0oooo0P0000040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0
oooo0P0000040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo
0P0000040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo0P00
00040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo0P000004
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo0P0000040?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo0P0000040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo0P0000030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo0P0000030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo0P0000030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo0P0000030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool06`3oool000d0oooo00D0omnL04P000000000001d0;ooo`030?oo
o`030?noM0000000M;oo00<0oooo00D0omnL04P000000000000007Boo`090?ooo`030000003oool0
00000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00303oool01`3o
okl0M00007Boo`3oool0okmd0000M02oool00P3oool00`3o_g@0000007Boo`020?ooo`070?oOW018
0000B9cO0?ooo`3ookl0M00007Boo`080?ooo`030000003oool000000080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`0K0?ooo`00303oool00`2oM00004RL0=ooo`020?ooo`030;md
0000M;l0oooo00D0oooo00<0oooO09a80000M;l00`3oool00`3o_g@004RL0=ooo`070?ooo`030000
003oool000000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`002`3o
ool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00D0oooo00<0omnL04P0001d_ol00`3o
ool00`3o_g@0001809cOo`070?ooo`030000003oool000000080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0K0?ooo`002`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo00D0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0
00000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`002`3oool00`3o
gi`0B00007Boo`030?ooo`030?noM00004P0W=oo00D0oooo00<0omnL04P0001d_ol00`3oool00`2o
M000001809cOo`070?ooo`<000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
01/0oooo000;0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01@3oool0203ookl0
M00000000018W=l0oooo0?noM000000007Bo203oool00`000000oooo000000020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool06`3oool000/0oooo00<0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`060?ooo`030;md0000B9`0_ia80080000000<004RL0=ooo`3oool01`3oool0
0`000000oooo000000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool0
00/0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`030?oOW0180000M;oo
00`0oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
01/0oooo000<0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`3OW4P0
001809cOo`0;0?ooo`030000003oool000000080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0K0?ooo`00303oool01`3ookl0M00007Boo`3oool0okmd0000M02oool02@3oool00`3o
_g@00000001d_`0:0?ooo`030000003oool000000080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0K0?ooo`003@3oool01@3ookl0M0000000000007@0_ooo00/0oooo00D0oooO09a8
00000000001d0;ooo`070?ooo`030000003oool000000080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo000000020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`0000000P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool000000080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0000
00020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<00000
0?ooo`0000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?oo
o`<000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`03
0000003oool000000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00
:@3oool00`000000oooo000000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
6`3oool002T0oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo01/0oooo000Y0?ooo`030000003oool000000080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo000000020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`0000000P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool000000080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0000
00020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<00000
0?ooo`0000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?oo
o`030000003oool000000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?oo
o`00:@3oool300000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00
:@3oool00`000000oooo000000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
6`3oool002T0oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo01/0oooo000Y0?ooo`030000003oool000000080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo000000020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`0000000P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool000000080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0000
00020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<00000
0?ooo`0000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?oo
o`030000003oool000000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?oo
o`00:@3oool00`000000oooo000000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool06`3oool002T0oooo0`0000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
6`3oool002T0oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo01/0oooo000Y0?ooo`030000003oool000000080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo000000020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool06`3oool002T0oooo00H000000?ooo`3oool000000?ooo`000005
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo01/0oooo000Y0?ooo`060000003oool0oooo0000003oool000001@3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0K0?ooo`00:@3oool01P000000oooo0?ooo`000000oooo000000D0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool06`3oool000d0oooo00D0omnL04P000000000001d0;ooo`030?ooo`030?no
M0000000M;oo00<0oooo00D0gia8000000000000000004RLg`090?ooo`060000003oool0oooo0000
003oool000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00303oool01`3ookl0M00007Boo`3oool0
okmd0000M02oool00P3oool00`3o_g@0000007Boo`020?ooo`070=nLB0000000M;oo0?ooo`3ooml0
W4P004RLg`080?ooo`060000003oool0oooo0000003oool000001@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K
0?ooo`00303oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00D0oooo00<0oooO09a8
0000M;l00`3oool00`3o_g@004RL0=ooo`070?ooo`060000003oool0oooo0000003oool000001@3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0K0?ooo`002`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0
_ooo00H0oooo00<0_g@0001d_`3oool00P3oool00`2oM00004RL0=ooo`070?ooo`060000003oool0
oooo0000003oool000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`002`3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00H0oooo00L0omnL04P0M02oool0oooo0;md00000000M;oo00P0
oooo00H000000?ooo`3oool000000?ooo`0000050?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000;0?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool01@3ookl0M4Q804P000000000
B9cO00T0oooo100000000`3oool000000?ooo`040?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000;0?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool0103ogi`0B00000000000M;l:
0?ooo`060000003oool0oooo0000003oool000001@3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`002`3o
ool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00H0omnL04P0000007@0_ooo
0?noM000M;l90?ooo`060000003oool0oooo0000003oool000001@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K
0?ooo`002`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00<0_g@00000
M02oool00P3oool00`2oM000B9cO0?ooo`070?ooo`060000003oool0oooo0000003oool000001@3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0K0?ooo`00303oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0
oooo00H0oooo00<0_g@00018W03Oool00P3oool00`3OW4P007Bo0?ooo`070?ooo`060000003oool0
oooo0000003oool000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00303oool01`3ookl0M00007Bo
o`3oool0okmd0000M02oool01`3oool01`3o_g@0000007Boo`3oool0oono07@0001d_ol0203oool0
1P000000oooo0?ooo`000000oooo000000D0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool000d0oooo00D0
oono07@000000000001d0;ooo`090?ooo`050?noM0000000000000000018W=l02@3oool01P000000
oooo0?ooo`000000oooo000000D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00H000000?oo
o`3oool000000?ooo`0000050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06@3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`060000003oool0
oooo0000003oool000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool01P000000oooo0?oo
o`000000oooo000000D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00H000000?ooo`3oool0
00000?ooo`0000050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`80000000@0oooo0000003o
ool000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool01P000000oooo0?ooo`000000oooo
000000D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00P0000060?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo01/0oooo000Y0?ooo`050000003oool0oooo0?ooo`0000001P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0K0?ooo`00:@3oool01@000000oooo0?ooo`3oool0000000H0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool06`3oool002T0oooo00D000000?ooo`3oool0oooo000000060?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
01/0oooo000Y0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?oo
o`00:@3oool2000000T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool0
02T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000
oooo0?ooo`080?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0K0?ooo`00:@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?oo
o`030000003oool0oooo00P0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool2000000T0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
6`3oool002T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool02@3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo
000Y0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool001X0oooo00H0_g@0
00000000000000000018W03Oool90?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
6`3oool001/0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo01/0oooo000K0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0K0?ooo`006`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool06`3oool001/0oooo00<0omnL04P0001d_ol02`3oool00`000000
oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000K0?ooo`030?oOW0180000M;oo00/0oooo
0`0000090?ooo`H000001@3oool6000000D0oooo1`0000050?ooo`H000001@3oool7000000D0oooo
1P0000050?oooa8000001@3oool7000000D0oooo1P0000050?oooa8000001@3oool7000000D0oooo
1P0000050?ooo`L000001@3oool6000000D0oooo1P0000060?ooo`H000001@3oool6000000D0oooo
1`0000050?ooo`H000001@3oool7000000D0oooo7@0000050?ooo`L000001@3ooolA000000H0oooo
1P0000050?ooo`L000001@3oool6000001d0oooo000K0?ooo`030?oOW0180000M;oo00/0oooo00<0
00000?ooo`3oool0o`3ooon50?ooo`006`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0
oooo0?l0ooooQ@3oool001/0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`3o0?oo
ohD0oooo000K0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0o`3ooon50?ooo`00
6@3oool01@3ooml0W4P0000000000000M;oo00/0oooo00<000000?ooo`3oool0o`3ooon50?ooo`00
6`3oool00`3o_g@0000007Boo`3o0?oooi<0oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-11.2874, -0.125454, \
0.410595, 0.00664356}}]
}, Closed]],

Cell[TextData[{
  "With a larger number value of ",
  StyleBox["PlotPoints", "InlineInput",
    FontWeight->"Bold"],
  " the periodic behavior of this function is revealed.  Except perhaps for \
the first pulse, this function provides a reasonable analytic facsimile of a \
train of unit pulses."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[\((1 + Tanh[x\^2\ Sin[2  x]])\)/2, {x, 0, 100}, 
        PlotPoints \[Rule] 100];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00952381 0.0147151 0.588604 [
[.21429 .00222 -8 -15.6875 ]
[.21429 .00222 8 0 ]
[.40476 .00222 -8 -15.6875 ]
[.40476 .00222 8 0 ]
[.59524 .00222 -8 -15.6875 ]
[.59524 .00222 8 0 ]
[.78571 .00222 -8 -15.6875 ]
[.78571 .00222 8 0 ]
[.97619 .00222 -11 -15.6875 ]
[.97619 .00222 11 0 ]
[.01131 .13244 -19 -7.84375 ]
[.01131 .13244 0 7.84375 ]
[.01131 .25016 -19 -7.84375 ]
[.01131 .25016 0 7.84375 ]
[.01131 .36788 -19 -7.84375 ]
[.01131 .36788 0 7.84375 ]
[.01131 .4856 -19 -7.84375 ]
[.01131 .4856 0 7.84375 ]
[.01131 .60332 -10 -7.84375 ]
[.01131 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .01472 m
.21429 .02097 L
s
gsave
.21429 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.40476 .01472 m
.40476 .02097 L
s
gsave
.40476 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
.59524 .01472 m
.59524 .02097 L
s
gsave
.59524 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(60) show
1.000 setlinewidth
grestore
.78571 .01472 m
.78571 .02097 L
s
gsave
.78571 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(80) show
1.000 setlinewidth
grestore
.97619 .01472 m
.97619 .02097 L
s
gsave
.97619 .00222 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .01472 m
.07143 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.16667 .01472 m
.16667 .01847 L
s
.2619 .01472 m
.2619 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.35714 .01472 m
.35714 .01847 L
s
.45238 .01472 m
.45238 .01847 L
s
.5 .01472 m
.5 .01847 L
s
.54762 .01472 m
.54762 .01847 L
s
.64286 .01472 m
.64286 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.7381 .01472 m
.7381 .01847 L
s
.83333 .01472 m
.83333 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.92857 .01472 m
.92857 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .13244 m
.03006 .13244 L
s
gsave
.01131 .13244 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.02381 .25016 m
.03006 .25016 L
s
gsave
.01131 .25016 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.02381 .36788 m
.03006 .36788 L
s
gsave
.01131 .36788 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.02381 .4856 m
.03006 .4856 L
s
gsave
.01131 .4856 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.02381 .60332 m
.03006 .60332 L
s
gsave
.01131 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04415 m
.02756 .04415 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .10301 m
.02756 .10301 L
s
.02381 .16187 m
.02756 .16187 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .30902 m
.02756 .30902 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .45617 m
.02756 .45617 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .30902 m
.02409 .30903 L
.02435 .30913 L
.02465 .30942 L
.02494 .30998 L
.0252 .31081 L
.02544 .31191 L
.0257 .31353 L
.02599 .31579 L
.02627 .31869 L
.02657 .32257 L
.02712 .33174 L
.02777 .34648 L
.02837 .36354 L
.03086 .45545 L
.03149 .4771 L
.03208 .49471 L
.03265 .50873 L
.03318 .51871 L
.03346 .5229 L
.03376 .52644 L
.03406 .52907 L
.03423 .53014 L
.03439 .53085 L
.03455 .53131 L
.03471 .53145 L
.03501 .5309 L
.03533 .52901 L
.03551 .52729 L
.03568 .52522 L
.036 .51975 L
.03631 .51254 L
.03664 .50178 L
.037 .486 L
.03736 .46492 L
.03774 .43544 L
.03806 .40332 L
.03841 .36034 L
.03966 .16971 L
.04002 .12196 L
.04036 .0864 L
.04066 .06267 L
.04083 .05238 L
.04099 .04462 L
.04128 .03405 L
.04155 .02736 L
.04184 .02253 L
.04216 .01931 L
.04231 .01822 L
.04249 .01733 L
Mistroke
.0428 .01623 L
.04311 .01561 L
.04339 .01526 L
.0437 .01504 L
.04386 .01496 L
.04403 .0149 L
.04434 .01483 L
.04462 .01479 L
.0448 .01477 L
.04496 .01476 L
.04527 .01474 L
.04544 .01474 L
.0456 .01473 L
.04577 .01473 L
.04596 .01473 L
.04624 .01472 L
.04655 .01472 L
.04684 .01472 L
.04711 .01472 L
.0474 .01472 L
.04757 .01472 L
.04773 .01472 L
.04804 .01472 L
.04822 .01472 L
.04838 .01472 L
.04866 .01472 L
.04896 .01472 L
.04921 .01472 L
.04949 .01472 L
.04979 .01473 L
.04996 .01473 L
.05011 .01473 L
.05041 .01474 L
.0507 .01476 L
.05096 .01479 L
.05124 .01486 L
.05148 .01499 L
.05175 .01527 L
.05203 .01599 L
.05234 .01802 L
.05262 .02331 L
.05289 .03617 L
.05318 .07325 L
.05348 .17168 L
.05364 .25557 L
.05382 .36122 L
.05412 .51058 L
.05441 .57432 L
.05473 .59605 L
.05502 .60146 L
Mistroke
.05529 .6028 L
.05546 .60309 L
.05562 .60321 L
.05579 .60327 L
.05598 .6033 L
.05616 .60331 L
.05632 .60332 L
.05664 .60332 L
.05694 .60332 L
.05722 .60332 L
.05753 .60332 L
.05769 .60332 L
.05786 .60332 L
.05817 .60332 L
.05846 .60332 L
.05873 .60332 L
.05901 .60332 L
.05933 .60332 L
.05948 .60332 L
.05966 .60332 L
.05996 .60332 L
.06025 .60332 L
.06053 .60332 L
.06081 .60332 L
.06106 .60332 L
.06133 .60332 L
.06162 .60332 L
.06193 .60332 L
.06222 .60332 L
.06249 .60332 L
.06276 .60332 L
.06306 .60332 L
.06338 .60332 L
.06367 .60332 L
.06395 .60332 L
.0642 .60332 L
.06449 .60332 L
.06477 .60332 L
.06506 .60332 L
.06538 .60332 L
.06568 .60332 L
.06595 .60332 L
.06627 .60332 L
.06645 .60332 L
.06661 .60332 L
.06692 .60332 L
.06724 .60332 L
.06752 .6033 L
.06777 .60315 L
.06806 .60125 L
Mistroke
.06834 .57978 L
.0686 .42454 L
.06884 .13204 L
.0691 .02666 L
.06937 .01554 L
.06965 .01477 L
.06995 .01472 L
.07023 .01472 L
.07049 .01472 L
.07079 .01472 L
.07096 .01472 L
.07112 .01472 L
.07141 .01472 L
.07172 .01472 L
.07202 .01472 L
.07219 .01472 L
.07235 .01472 L
.07252 .01472 L
.07267 .01472 L
.07285 .01472 L
.07302 .01472 L
.07331 .01472 L
.07359 .01472 L
.07389 .01472 L
.07421 .01472 L
.07453 .01472 L
.07471 .01472 L
.07487 .01472 L
.07517 .01472 L
.07549 .01472 L
.07566 .01472 L
.07584 .01472 L
.07601 .01472 L
.07617 .01472 L
.07635 .01472 L
.07655 .01472 L
.07673 .01472 L
.07691 .01472 L
.07719 .01472 L
.07749 .01472 L
.07778 .01472 L
.07804 .01472 L
.07834 .01472 L
.07851 .01472 L
.07867 .01472 L
.07897 .01472 L
.07928 .01472 L
.07945 .01472 L
.07961 .01472 L
.07978 .01472 L
Mistroke
.07996 .01472 L
.08027 .01472 L
.08044 .01472 L
.0806 .01472 L
.0809 .01472 L
.08118 .01472 L
.08149 .01472 L
.0818 .01472 L
.08207 .01472 L
.08237 .01472 L
.08265 .01472 L
.08291 .01472 L
.0832 .01508 L
.08351 .06807 L
.08367 .35341 L
.08384 .57841 L
.08414 .60317 L
.08431 .60331 L
.08446 .60332 L
.08463 .60332 L
.08481 .60332 L
.08511 .60332 L
.08528 .60332 L
.08544 .60332 L
.08574 .60332 L
.08602 .60332 L
.08633 .60332 L
.08663 .60332 L
.0911 .60332 L
.09343 .60332 L
.09471 .60332 L
.09592 .60332 L
.09641 .60332 L
.09668 .60332 L
.09693 .60332 L
.09721 .60332 L
.09751 .60332 L
.09776 .60332 L
.09804 .60332 L
.09834 .60278 L
.09851 .56307 L
.09867 .1252 L
.09896 .01479 L
.09926 .01472 L
.09954 .01472 L
.09979 .01472 L
.10007 .01472 L
.10038 .01472 L
.10054 .01472 L
.1007 .01472 L
Mistroke
.10099 .01472 L
.10131 .01472 L
.10149 .01472 L
.10165 .01472 L
.10192 .01472 L
.10222 .01472 L
.10254 .01472 L
.10283 .01472 L
.10299 .01472 L
.10318 .01472 L
.10335 .01472 L
.1035 .01472 L
.1038 .01472 L
.10412 .01472 L
.10428 .01472 L
.10446 .01472 L
.10478 .01472 L
.10496 .01472 L
.10516 .01472 L
.10532 .01472 L
.1055 .01472 L
.10566 .01472 L
.10583 .01472 L
.10614 .01472 L
.10644 .01472 L
.10672 .01472 L
.10702 .01472 L
.1072 .01472 L
.10736 .01472 L
.10752 .01472 L
.10768 .01472 L
.10786 .01472 L
.10803 .01472 L
.10831 .01472 L
.10861 .01472 L
.10889 .01472 L
.10915 .01472 L
.10945 .01472 L
.10962 .01472 L
.10978 .01472 L
.11007 .01472 L
.11037 .01472 L
.11067 .01472 L
.11084 .01472 L
.11099 .01472 L
.11128 .01472 L
.11158 .01472 L
.11185 .01472 L
.1121 .01472 L
.11238 .01472 L
Mistroke
.11268 .01472 L
.11299 .01472 L
.11317 .01472 L
.11334 .01483 L
.11364 .56696 L
.11381 .60325 L
.11397 .60332 L
.11426 .60332 L
.11457 .60332 L
.11487 .60332 L
.11515 .60332 L
.11959 .60332 L
.12415 .60332 L
.1254 .60332 L
.12609 .60332 L
.12674 .60332 L
.12735 .60332 L
.12751 .60332 L
.12769 .60332 L
.12801 .60332 L
.12829 .60332 L
.12855 .14263 L
.12885 .01472 L
.12916 .01472 L
.12946 .01472 L
.12974 .01472 L
.13006 .01472 L
.13036 .01472 L
.13064 .01472 L
.1309 .01472 L
.13118 .01472 L
.13148 .01472 L
.13178 .01472 L
.13195 .01472 L
.1321 .01472 L
.1324 .01472 L
.13268 .01472 L
.13284 .01472 L
.13301 .01472 L
.13332 .01472 L
.13351 .01472 L
.13368 .01472 L
.13384 .01472 L
.13401 .01472 L
.13433 .01472 L
.13463 .01472 L
.1348 .01472 L
.13496 .01472 L
.13514 .01472 L
.13531 .01472 L
Mistroke
.13548 .01472 L
.13565 .01472 L
.13597 .01472 L
.13614 .01472 L
.1363 .01472 L
.13648 .01472 L
.13668 .01472 L
.13698 .01472 L
.13714 .01472 L
.1373 .01472 L
.13762 .01472 L
.1378 .01472 L
.13796 .01472 L
.13825 .01472 L
.13856 .01472 L
.13885 .01472 L
.13913 .01472 L
.13943 .01472 L
.13972 .01472 L
.13998 .01472 L
.14026 .01472 L
.14057 .01472 L
.14074 .01472 L
.14089 .01472 L
.1412 .01472 L
.14148 .01472 L
.14175 .01472 L
.14203 .01472 L
.14227 .01472 L
.14254 .01472 L
.14282 .01472 L
.14313 .01472 L
.14342 .02008 L
.14368 .60332 L
.14399 .60332 L
.14415 .60332 L
.14432 .60332 L
.14491 .60332 L
.14625 .60332 L
.14869 .60332 L
.15339 .60332 L
.15475 .60332 L
.15603 .60332 L
.15719 .60332 L
.15751 .60332 L
.15768 .60332 L
.15784 .60332 L
.15816 .60332 L
.15845 .27607 L
.15873 .01472 L
Mistroke
.15904 .01472 L
.15932 .01472 L
.15958 .01472 L
.15988 .01472 L
.16006 .01472 L
.16022 .01472 L
.16051 .01472 L
.16082 .01472 L
.16111 .01472 L
.16143 .01472 L
.16173 .01472 L
.16201 .01472 L
.16231 .01472 L
.16248 .01472 L
.16265 .01472 L
.16282 .01472 L
.16297 .01472 L
.16315 .01472 L
.16332 .01472 L
.16358 .01472 L
.16386 .01472 L
.16416 .01472 L
.16444 .01472 L
.1647 .01472 L
.16493 .01472 L
.16521 .01472 L
.16547 .01472 L
.16575 .01472 L
.16605 .01472 L
.16633 .01472 L
.16658 .01472 L
.16688 .01472 L
.16705 .01472 L
.16721 .01472 L
.16749 .01472 L
.1678 .01472 L
.16809 .01472 L
.16826 .01472 L
.16842 .01472 L
.1687 .01472 L
.169 .01472 L
.16927 .01472 L
.16952 .01472 L
.1698 .01472 L
.1701 .01472 L
.1704 .01472 L
.17057 .01472 L
.17073 .01472 L
.17105 .01472 L
.17122 .01472 L
Mistroke
.17139 .01472 L
.17168 .01472 L
.17199 .01472 L
.17229 .01472 L
.17257 .01472 L
.17282 .01472 L
.1731 .01472 L
.17336 .01776 L
.1736 .60332 L
.17387 .60332 L
.17418 .60332 L
.17473 .60332 L
.17701 .60332 L
.18171 .60332 L
.1842 .60332 L
.18551 .60332 L
.18622 .60332 L
.18689 .60332 L
.18748 .60332 L
.18765 .60332 L
.18781 .60332 L
.18811 .60332 L
.18827 .60332 L
.18844 .01476 L
.18875 .01472 L
.18892 .01472 L
.18908 .01472 L
.18926 .01472 L
.18945 .01472 L
.18975 .01472 L
.18992 .01472 L
.19008 .01472 L
.19036 .01472 L
.19067 .01472 L
.19098 .01472 L
.19126 .01472 L
.19152 .01472 L
.19179 .01472 L
.19208 .01472 L
.19235 .01472 L
.19265 .01472 L
.19296 .01472 L
.19312 .01472 L
.19329 .01472 L
.19361 .01472 L
.19391 .01472 L
.19419 .01472 L
.19446 .01472 L
.1947 .01472 L
.19497 .01472 L
Mistroke
.19526 .01472 L
.19555 .01472 L
.19586 .01472 L
.19612 .01472 L
.1964 .01472 L
.19672 .01472 L
.19702 .01472 L
.19717 .01472 L
.19734 .01472 L
.19752 .01472 L
.19768 .01472 L
.19799 .01472 L
.19827 .01472 L
.19842 .01472 L
.19859 .01472 L
.19889 .01472 L
.19919 .01472 L
.19946 .01472 L
.19978 .01472 L
.20007 .01472 L
.20034 .01472 L
.20059 .01472 L
.20087 .01472 L
.20116 .01472 L
.20147 .01472 L
.20176 .01472 L
.20192 .01472 L
.2021 .01472 L
.20241 .01472 L
.20271 .01472 L
.20302 .01472 L
.20329 .01586 L
.20358 .60332 L
.20376 .60332 L
.20393 .60332 L
.20424 .60332 L
.20495 .60332 L
.20624 .60332 L
.21109 .60332 L
.21549 .60332 L
.21667 .60332 L
.21729 .60332 L
.21763 .60332 L
.21795 .60332 L
.21822 .60331 L
.21852 .01472 L
.21869 .01472 L
.21885 .01472 L
.21915 .01472 L
.21943 .01472 L
Mistroke
.21968 .01472 L
.21996 .01472 L
.22026 .01472 L
.22057 .01472 L
.22086 .01472 L
.22117 .01472 L
.22133 .01472 L
.2215 .01472 L
.2218 .01472 L
.22211 .01472 L
.22241 .01472 L
.22267 .01472 L
.22284 .01472 L
.22302 .01472 L
.22333 .01472 L
.22351 .01472 L
.22367 .01472 L
.22385 .01472 L
.22405 .01472 L
.22421 .01472 L
.22439 .01472 L
.22456 .01472 L
.22472 .01472 L
.22502 .01472 L
.22534 .01472 L
.22564 .01472 L
.22581 .01472 L
.22597 .01472 L
.22614 .01472 L
.2263 .01472 L
.22648 .01472 L
.22664 .01472 L
.22694 .01472 L
.22722 .01472 L
.22752 .01472 L
.22767 .01472 L
.22784 .01472 L
.22801 .01472 L
.22819 .01472 L
.22851 .01472 L
.22869 .01472 L
.22886 .01472 L
.22904 .01472 L
.22923 .01472 L
.2294 .01472 L
.22958 .01472 L
.22976 .01472 L
.22992 .01472 L
.23022 .01472 L
.23054 .01472 L
Mistroke
.23084 .01472 L
.23112 .01472 L
.23144 .01472 L
.23174 .01472 L
.23202 .01472 L
.23228 .01472 L
.23257 .01472 L
.23287 .01472 L
.23317 .01472 L
.23334 .60332 L
.2335 .60332 L
.23378 .60332 L
.23409 .60332 L
.23543 .60332 L
.24026 .60332 L
.24466 .60332 L
.24583 .60332 L
.24644 .60332 L
.2471 .60332 L
.24767 .60332 L
.24782 .60332 L
.24799 .60332 L
.2483 .01472 L
.24859 .01472 L
.24887 .01472 L
.24913 .01472 L
.2494 .01472 L
.24957 .01472 L
.24975 .01472 L
.24992 .01472 L
.25007 .01472 L
.25037 .01472 L
.25068 .01472 L
.25084 .01472 L
.25101 .01472 L
.25132 .01472 L
.25151 .01472 L
.2517 .01472 L
.25186 .01472 L
.25203 .01472 L
.25233 .01472 L
.25249 .01472 L
.25265 .01472 L
.25296 .01472 L
.25314 .01472 L
.2533 .01472 L
.25358 .01472 L
.25388 .01472 L
.25419 .01472 L
.25448 .01472 L
Mistroke
.25479 .01472 L
.25494 .01472 L
.25512 .01472 L
.25542 .01472 L
.2557 .01472 L
.25587 .01472 L
.25604 .01472 L
.25634 .01472 L
.25651 .01472 L
.25667 .01472 L
.25684 .01472 L
.25703 .01472 L
.25734 .01472 L
.25752 .01472 L
.25768 .01472 L
.25798 .01472 L
.25829 .01472 L
.25858 .01472 L
.25884 .01472 L
.25913 .01472 L
.25944 .01472 L
.25974 .01472 L
.26002 .01472 L
.26033 .01472 L
.26048 .01472 L
.26065 .01472 L
.26094 .01472 L
.26124 .01472 L
.26153 .01472 L
.2618 .01472 L
.26206 .01472 L
.2623 .01472 L
.26256 .01472 L
.26284 .01472 L
.26314 .01504 L
.26342 .60332 L
.26367 .60332 L
.26394 .60332 L
.26868 .60332 L
.27119 .60332 L
.27388 .60332 L
.2752 .60332 L
.27646 .60332 L
.27701 .60332 L
.27759 .60332 L
.2779 .60332 L
.27807 .60332 L
.27823 .01472 L
.27852 .01472 L
.27882 .01472 L
Mistroke
.27911 .01472 L
.27938 .01472 L
.27968 .01472 L
.28 .01472 L
.28016 .01472 L
.28033 .01472 L
.28064 .01472 L
.28095 .01472 L
.28123 .01472 L
.2815 .01472 L
.28174 .01472 L
.28201 .01472 L
.28231 .01472 L
.28261 .01472 L
.2829 .01472 L
.28316 .01472 L
.28344 .01472 L
.2836 .01472 L
.28377 .01472 L
.28406 .01472 L
.28438 .01472 L
.28456 .01472 L
.28473 .01472 L
.28504 .01472 L
.28532 .01472 L
.28547 .01472 L
.28564 .01472 L
.28595 .01472 L
.28622 .01472 L
.28652 .01472 L
.28683 .01472 L
.28713 .01472 L
.2874 .01472 L
.28765 .01472 L
.28793 .01472 L
.28822 .01472 L
.28854 .01472 L
.28883 .01472 L
.28899 .01472 L
.28916 .01472 L
.28948 .01472 L
.28977 .01472 L
.29008 .01472 L
.29035 .01472 L
.29065 .01472 L
.29082 .01472 L
.29099 .01472 L
.29131 .01472 L
.29148 .01472 L
.29165 .01472 L
Mistroke
.29183 .01472 L
.29202 .01472 L
.29219 .01472 L
.29237 .01472 L
.29254 .01472 L
.29269 .01472 L
.29299 .01472 L
.2933 .60332 L
.2936 .60332 L
.29392 .60332 L
.29451 .60332 L
.29561 .60332 L
.29808 .60332 L
.30253 .60332 L
.30501 .60332 L
.30622 .60332 L
.30733 .60332 L
.30762 .60332 L
.30793 .60332 L
.30823 .01472 L
.3085 .01472 L
.30882 .01472 L
.30911 .01472 L
.30927 .01472 L
.30944 .01472 L
.30976 .01472 L
.30992 .01472 L
.31008 .01472 L
.31025 .01472 L
.31043 .01472 L
.31063 .01472 L
.31081 .01472 L
.31097 .01472 L
.31115 .01472 L
.3113 .01472 L
.31147 .01472 L
.31165 .01472 L
.31181 .01472 L
.31211 .01472 L
.31243 .01472 L
.31273 .01472 L
.31301 .01472 L
.31333 .01472 L
.31363 .01472 L
.31391 .01472 L
.31417 .01472 L
.31445 .01472 L
.31475 .01472 L
.31505 .01472 L
.31522 .01472 L
Mistroke
.31537 .01472 L
.31567 .01472 L
.31596 .01472 L
.31612 .01472 L
.31629 .01472 L
.31659 .01472 L
.31678 .01472 L
.31696 .01472 L
.31712 .01472 L
.31729 .01472 L
.31761 .01472 L
.31791 .01472 L
.31808 .01472 L
.31823 .01472 L
.31841 .01472 L
.31858 .01472 L
.31875 .01472 L
.31892 .01472 L
.31924 .01472 L
.31941 .01472 L
.31958 .01472 L
.31976 .01472 L
.31995 .01472 L
.32025 .01472 L
.32042 .01472 L
.32058 .01472 L
.32089 .01472 L
.32107 .01472 L
.32123 .01472 L
.32152 .01472 L
.32183 .01472 L
.32213 .01472 L
.3224 .01472 L
.32271 .01472 L
.323 .0168 L
.32325 .60332 L
.32354 .60332 L
.32476 .60332 L
.32696 .60332 L
.33196 .60332 L
.33443 .60332 L
.33564 .60332 L
.33675 .60332 L
.33732 .60332 L
.3376 .60332 L
.33791 .60332 L
.33823 .01472 L
.3384 .01472 L
.33856 .01472 L
.33887 .01472 L
Mistroke
.33917 .01472 L
.33932 .01472 L
.33949 .01472 L
.33967 .01472 L
.33983 .01472 L
.34013 .01472 L
.34045 .01472 L
.34063 .01472 L
.34082 .01472 L
.34101 .01472 L
.34118 .01472 L
.34134 .01472 L
.3415 .01472 L
.34167 .01472 L
.34184 .01472 L
.34212 .01472 L
.34242 .01472 L
.34271 .01472 L
.34297 .01472 L
.34327 .01472 L
.34345 .01472 L
.34361 .01472 L
.3439 .01472 L
.34421 .01472 L
.3445 .01472 L
.34482 .01472 L
.34512 .01472 L
.3454 .01472 L
.3457 .01472 L
.34588 .01472 L
.34604 .01472 L
.34621 .01472 L
.34636 .01472 L
.34654 .01472 L
.34671 .01472 L
.34697 .01472 L
.34725 .01472 L
.34755 .01472 L
.34783 .01472 L
.34809 .01472 L
.34833 .01472 L
.3486 .01472 L
.34886 .01472 L
.34914 .01472 L
.34944 .01472 L
.34972 .01472 L
.34997 .01472 L
.35027 .01472 L
.35044 .01472 L
.3506 .01472 L
Mistroke
.35089 .01472 L
.35119 .01472 L
.35149 .01472 L
.35165 .01472 L
.35181 .01472 L
.35209 .01472 L
.3524 .01472 L
.35266 .01472 L
.35292 .01706 L
.3532 .60332 L
.35349 .60332 L
.35596 .60332 L
.36041 .60332 L
.36286 .60332 L
.36415 .60332 L
.36553 .60332 L
.36621 .60332 L
.36685 .60332 L
.36743 .60332 L
.36759 .60332 L
.36775 .60332 L
.36805 .01472 L
.36833 .01472 L
.36863 .01472 L
.36891 .01472 L
.36917 .01472 L
.36947 .01472 L
.36964 .01472 L
.3698 .01472 L
.37009 .01472 L
.3704 .01472 L
.37067 .01472 L
.37097 .01472 L
.37129 .01472 L
.37158 .01472 L
.37186 .01472 L
.37211 .01472 L
.37239 .01472 L
.37267 .01472 L
.37296 .01472 L
.37328 .01472 L
.37355 .01472 L
.37384 .01472 L
.37416 .01472 L
.37434 .01472 L
.3745 .01472 L
.37481 .01472 L
.37511 .01472 L
.37539 .01472 L
.37565 .01472 L
Mistroke
.37594 .01472 L
.37621 .01472 L
.37647 .01472 L
.3767 .01472 L
.37696 .01472 L
.37723 .01472 L
.37751 .01472 L
.3778 .01472 L
.37808 .01472 L
.37834 .01472 L
.37863 .01472 L
.3788 .01472 L
.37896 .01472 L
.37924 .01472 L
.37954 .01472 L
.37982 .01472 L
.38012 .01472 L
.38044 .01472 L
.38073 .01472 L
.38101 .01472 L
.38126 .01472 L
.38155 .01472 L
.38182 .01472 L
.38212 .01472 L
.38243 .01472 L
.3827 .01472 L
.383 .60332 L
.38331 .60332 L
.38365 .60332 L
.38427 .60332 L
.3887 .60332 L
.39116 .60332 L
.39384 .60332 L
.39507 .60332 L
.39637 .60332 L
.39701 .60332 L
.39732 .60332 L
.3976 .60332 L
.39786 .01472 L
.39814 .01472 L
.39841 .01472 L
.39872 .01472 L
.39904 .01472 L
.39933 .01472 L
.39948 .01472 L
.39965 .01472 L
.39983 .01472 L
.39999 .01472 L
.4003 .01472 L
.40058 .01472 L
Mistroke
.40073 .01472 L
.4009 .01472 L
.4012 .01472 L
.4015 .01472 L
.40177 .01472 L
.40208 .01472 L
.40237 .01472 L
.40265 .01472 L
.4029 .01472 L
.40317 .01472 L
.40346 .01472 L
.40378 .01472 L
.40407 .01472 L
.40422 .01472 L
.4044 .01472 L
.40471 .01472 L
.40501 .01472 L
.40532 .01472 L
.40559 .01472 L
.40588 .01472 L
.40605 .01472 L
.40622 .01472 L
.40654 .01472 L
.40672 .01472 L
.40688 .01472 L
.40706 .01472 L
.40725 .01472 L
.40742 .01472 L
.4076 .01472 L
.40777 .01472 L
.40792 .01472 L
.40822 .01472 L
.40854 .01472 L
.40883 .01472 L
.40915 .01472 L
.40945 .01472 L
.40973 .01472 L
.41001 .01472 L
.41026 .01472 L
.41055 .01472 L
.41082 .01472 L
.41112 .01472 L
.41144 .01472 L
.41171 .01472 L
.41201 .01472 L
.41232 .01472 L
.4125 .01472 L
.41267 .01472 L
.41298 .60332 L
.41328 .60332 L
Mistroke
.41772 .60332 L
.42215 .60332 L
.42466 .60332 L
.4259 .60332 L
.42644 .60332 L
.42703 .60332 L
.42731 .60332 L
.42761 .60332 L
.42789 .01472 L
.42815 .01472 L
.42845 .01472 L
.42876 .01472 L
.42906 .01472 L
.42934 .01472 L
.42964 .01472 L
.42991 .01472 L
.43022 .01472 L
.43039 .01472 L
.43055 .01472 L
.43086 .01472 L
.43104 .01472 L
.43121 .01472 L
.43151 .01472 L
.43166 .01472 L
.43183 .01472 L
.43198 .01472 L
.43215 .01472 L
.43232 .01472 L
.43249 .01472 L
.43278 .01472 L
.43309 .01472 L
.43327 .01472 L
.43346 .01472 L
.43364 .01472 L
.43381 .01472 L
.43412 .01472 L
.43428 .01472 L
.43446 .01472 L
.43462 .01472 L
.43478 .01472 L
.43507 .01472 L
.43539 .01472 L
.43556 .01472 L
.43573 .01472 L
.43603 .01472 L
.43631 .01472 L
.43663 .01472 L
.43692 .01472 L
.43721 .01472 L
.43751 .01472 L
Mistroke
.43779 .01472 L
.43805 .01472 L
.43836 .01472 L
.43853 .01472 L
.43869 .01472 L
.43898 .01472 L
.43929 .01472 L
.43959 .01472 L
.43991 .01472 L
.44021 .01472 L
.44048 .01472 L
.44079 .01472 L
.44096 .01472 L
.44112 .01472 L
.44129 .01472 L
.44145 .01472 L
.44163 .01472 L
.44179 .01472 L
.44205 .01472 L
.44234 .01472 L
.44264 .01472 L
.44291 .60332 L
.44317 .60332 L
.44341 .60332 L
.44394 .60332 L
.44627 .60332 L
.45128 .60332 L
.45371 .60332 L
.45592 .60332 L
.45646 .60332 L
.45703 .60332 L
.45733 .60332 L
.4575 .60332 L
.45766 .01489 L
.45794 .01472 L
.45823 .01472 L
.45851 .01472 L
.45882 .01472 L
.45911 .01472 L
.45937 .01472 L
.45961 .01472 L
.45988 .01472 L
.46015 .01472 L
.46041 .01472 L
.46071 .01472 L
.46087 .01472 L
.46103 .01472 L
.46131 .01472 L
.46161 .01472 L
.46191 .01472 L
Mistroke
.46219 .01472 L
.46244 .01472 L
.46271 .01472 L
.46301 .01472 L
.46318 .01472 L
.46334 .01472 L
.46366 .01472 L
.46384 .01472 L
.46401 .01472 L
.4643 .01472 L
.46462 .01472 L
.46492 .01472 L
.46519 .01472 L
.46548 .01472 L
.46578 .01472 L
.46604 .01472 L
.46632 .01472 L
.46662 .01472 L
.46679 .01472 L
.46694 .01472 L
.46724 .01472 L
.46753 .01472 L
.46779 .01472 L
.46807 .01472 L
.46831 .01472 L
.46858 .01472 L
.46886 .01472 L
.46916 .01472 L
.46945 .01472 L
.46972 .01472 L
.47002 .01472 L
.4702 .01472 L
.47035 .01472 L
.47066 .01472 L
.47094 .01472 L
.4711 .01472 L
.47128 .01472 L
.47158 .01472 L
.47175 .01472 L
.47191 .01472 L
.47209 .01472 L
.47228 .01472 L
.47259 .01472 L
.47277 .60332 L
.47293 .60332 L
.47355 .60332 L
.47471 .60332 L
.47963 .60332 L
.48203 .60332 L
.48424 .60332 L
Mistroke
.48639 .60332 L
.48691 .60332 L
.4872 .60332 L
.48747 .60332 L
.48762 .01472 L
.48779 .01472 L
.48808 .01472 L
.48838 .01472 L
.48865 .01472 L
.48892 .01472 L
.48922 .01472 L
.48949 .01472 L
.48975 .01472 L
.49004 .01472 L
.49021 .01472 L
.49037 .01472 L
.49065 .01472 L
.49096 .01472 L
.49124 .01472 L
.49155 .01472 L
.49185 .01472 L
.49212 .01472 L
.49241 .01472 L
.49258 .01472 L
.49274 .01472 L
.49306 .01472 L
.49324 .01472 L
.4934 .01472 L
.49369 .01472 L
.49396 .01472 L
.49425 .01472 L
.49457 .01472 L
.49473 .01472 L
.4949 .01472 L
.49522 .01472 L
.49553 .01472 L
.49581 .01472 L
.4961 .01472 L
.49638 .01472 L
.49662 .01472 L
.49689 .01472 L
.49718 .01472 L
.49749 .01472 L
.49775 .01472 L
.49804 .01472 L
.4982 .01472 L
.49838 .01472 L
.4987 .01472 L
.49901 .01472 L
.4993 .01472 L
Mistroke
.49961 .01472 L
.49979 .01472 L
.49995 .01472 L
.50013 .01472 L
.50029 .01472 L
.50046 .01472 L
.50065 .01472 L
.50094 .01472 L
.50125 .01472 L
.50154 .01472 L
.50181 .01472 L
.50212 .01472 L
.5023 .01472 L
.50246 .01472 L
.50276 .60332 L
.50308 .60332 L
.51227 .60332 L
.51469 .60332 L
.51588 .60332 L
.51646 .60332 L
.51698 .60332 L
.5173 .60332 L
.5176 .01472 L
.51791 .01472 L
.51809 .01472 L
.51825 .01472 L
.51857 .01472 L
.51875 .01472 L
.51892 .01472 L
.5191 .01472 L
.51926 .01472 L
.51945 .01472 L
.51963 .01472 L
.51992 .01472 L
.52023 .01472 L
.5205 .01472 L
.52079 .01472 L
.5211 .01472 L
.52128 .01472 L
.52144 .01472 L
.52173 .01472 L
.52205 .01472 L
.52235 .01472 L
.52252 .01472 L
.52267 .01472 L
.52296 .01472 L
.52326 .01472 L
.52354 .01472 L
.52379 .01472 L
.52407 .01472 L
Mistroke
.52437 .01472 L
.52469 .01472 L
.52487 .01472 L
.52503 .01472 L
.52534 .01472 L
.52549 .01472 L
.52566 .01472 L
.52596 .01472 L
.52627 .01472 L
.52657 .01472 L
.52684 .01472 L
.5271 .01472 L
.52738 .01472 L
.52764 .01472 L
.52788 .01472 L
.52816 .01472 L
.52846 .01472 L
.52875 .01472 L
.52901 .01472 L
.52929 .01472 L
.52954 .01472 L
.52982 .01472 L
.53012 .01472 L
.53044 .01472 L
.53074 .01472 L
.53101 .01472 L
.5313 .01472 L
.53146 .01472 L
.53162 .01472 L
.53192 .01472 L
.53224 .01472 L
.53242 .01472 L
.53259 .60332 L
.5329 .60332 L
.53318 .60332 L
.5338 .60332 L
.53607 .60332 L
.54116 .60332 L
.54341 .60332 L
.54579 .60332 L
.54639 .60332 L
.5467 .60332 L
.54703 .60332 L
.54721 .60332 L
.5474 .60227 L
.54758 .01472 L
.54774 .01472 L
.54791 .01472 L
.54806 .01472 L
.54824 .01472 L
Mistroke
.54841 .01472 L
.5487 .01472 L
.54901 .01472 L
.54928 .01472 L
.54957 .01472 L
.54988 .01472 L
.55006 .01472 L
.55022 .01472 L
.55053 .01472 L
.55082 .01472 L
.55113 .01472 L
.5513 .01472 L
.55146 .01472 L
.55174 .01472 L
.55205 .01472 L
.55232 .01472 L
.55258 .01472 L
.55286 .01472 L
.55316 .01472 L
.55348 .01472 L
.55366 .01472 L
.55383 .01472 L
.55413 .01472 L
.55429 .01472 L
.55446 .01472 L
.55475 .01472 L
.55507 .01472 L
.55537 .01472 L
.55564 .01472 L
.55591 .01472 L
.55621 .01472 L
.55648 .01472 L
.55673 .01472 L
.55703 .01472 L
.55735 .01472 L
.55765 .01472 L
.55793 .01472 L
.55819 .01472 L
.55843 .01472 L
.5587 .01472 L
.55899 .01472 L
.55929 .01472 L
.55957 .01472 L
.55982 .01472 L
.5601 .01472 L
.5604 .01472 L
.56067 .01472 L
.56096 .01472 L
.56128 .01472 L
.56155 .01472 L
Mistroke
.5618 .01472 L
.56208 .01472 L
.56238 .60332 L
.56255 .60332 L
.56271 .60332 L
.56306 .60332 L
.5637 .60332 L
.56489 .60332 L
.56999 .60332 L
.57453 .60332 L
.57578 .60332 L
.57647 .60332 L
.57681 .60332 L
.57712 .60332 L
.57741 .01472 L
.57772 .01472 L
.5779 .01472 L
.57807 .01472 L
.57838 .01472 L
.57867 .01472 L
.57894 .01472 L
.57923 .01472 L
.57954 .01472 L
.57983 .01472 L
.58 .01472 L
.58016 .01472 L
.58044 .01472 L
.58074 .01472 L
.58101 .01472 L
.58127 .01472 L
.58154 .01472 L
.58184 .01472 L
.58214 .01472 L
.58231 .01472 L
.58247 .01472 L
.58279 .01472 L
.58297 .01472 L
.58313 .01472 L
.58342 .01472 L
.58374 .01472 L
.58403 .01472 L
.58431 .01472 L
.58456 .01472 L
.58484 .01472 L
.5851 .01472 L
.58534 .01472 L
.58562 .01472 L
.58592 .01472 L
.5862 .01472 L
.58647 .01472 L
Mistroke
.58674 .01472 L
.58699 .01472 L
.58727 .01472 L
.58758 .01472 L
.58788 .01472 L
.58819 .01472 L
.58846 .01472 L
.58875 .01472 L
.58907 .01472 L
.58937 .01472 L
.58952 .01472 L
.58969 .01472 L
.58987 .01472 L
.59003 .01472 L
.59034 .01472 L
.59061 .01472 L
.59077 .01472 L
.59094 .01472 L
.59124 .01472 L
.59154 .01472 L
.59181 .01472 L
.59212 .01472 L
.59241 .60332 L
.59269 .60332 L
.59294 .60332 L
.5935 .60332 L
.59858 .60332 L
.60314 .60332 L
.60439 .60332 L
.60574 .60332 L
.60639 .60332 L
.6067 .60332 L
.607 .60332 L
.60727 .01472 L
.60755 .01472 L
.60784 .01472 L
.60815 .01472 L
.60845 .01472 L
.60872 .01472 L
.60903 .01472 L
.60919 .01472 L
.60935 .01472 L
.60966 .01472 L
.60983 .01472 L
.60999 .01472 L
.61029 .01472 L
.6106 .01472 L
.61087 .01472 L
.61115 .01472 L
.61142 .01472 L
Mistroke
.61167 .01472 L
.61196 .01472 L
.61227 .01472 L
.61257 .01472 L
.61285 .01472 L
.61317 .01472 L
.61334 .01472 L
.61351 .01472 L
.61381 .01472 L
.61413 .01472 L
.61442 .01472 L
.61469 .01472 L
.61498 .01472 L
.6153 .01472 L
.61547 .01472 L
.61563 .01472 L
.61581 .01472 L
.616 .01472 L
.61631 .01472 L
.61649 .01472 L
.61666 .01472 L
.61697 .01472 L
.61726 .01472 L
.61742 .01472 L
.61759 .01472 L
.6179 .01472 L
.61806 .01472 L
.61823 .01472 L
.61852 .01472 L
.61884 .01472 L
.61902 .01472 L
.61919 .01472 L
.6195 .01472 L
.61977 .01472 L
.61993 .01472 L
.6201 .01472 L
.6204 .01472 L
.62067 .01472 L
.62097 .01472 L
.62129 .01472 L
.62158 .01472 L
.62185 .01472 L
.6221 .01472 L
.62238 .60332 L
.62267 .60332 L
.62775 .60332 L
.63259 .60332 L
.6347 .60332 L
.63581 .60332 L
.63642 .60332 L
Mistroke
.63671 .60332 L
.63699 .60332 L
.63728 .01472 L
.63759 .01472 L
.63775 .01472 L
.63793 .01472 L
.6381 .01472 L
.63826 .01472 L
.63855 .01472 L
.63882 .01472 L
.63914 .01472 L
.63943 .01472 L
.63971 .01472 L
.64001 .01472 L
.64026 .01472 L
.64054 .01472 L
.64084 .01472 L
.64101 .01472 L
.64116 .01472 L
.64146 .01472 L
.64174 .01472 L
.64203 .01472 L
.64219 .01472 L
.64235 .01472 L
.64266 .01472 L
.64284 .01472 L
.643 .01472 L
.64327 .01472 L
.64357 .01472 L
.64388 .01472 L
.64417 .01472 L
.64434 .01472 L
.64452 .01472 L
.64468 .01472 L
.64484 .01472 L
.64514 .01472 L
.64545 .01472 L
.64561 .01472 L
.64579 .01472 L
.6461 .01472 L
.64629 .01472 L
.64647 .01472 L
.64664 .01472 L
.64681 .01472 L
.64711 .01472 L
.64738 .01472 L
.64767 .01472 L
.64799 .01472 L
.64826 .01472 L
.64851 .01472 L
Mistroke
.64879 .01472 L
.64909 .01472 L
.64926 .01472 L
.64941 .01472 L
.64958 .01472 L
.64976 .01472 L
.65007 .01472 L
.65024 .01472 L
.6504 .01472 L
.6507 .01472 L
.65098 .01472 L
.6513 .01472 L
.6516 .01472 L
.65192 .01472 L
.65221 .60332 L
.65237 .60332 L
.65255 .60332 L
.65286 .60332 L
.65403 .60332 L
.65669 .60332 L
.66143 .60332 L
.66394 .60332 L
.66532 .60332 L
.666 .60332 L
.6663 .60332 L
.66663 .60332 L
.66694 .60332 L
.66723 .01472 L
.66754 .01472 L
.66771 .01472 L
.66787 .01472 L
.66818 .01472 L
.66836 .01472 L
.66853 .01472 L
.6687 .01472 L
.66886 .01472 L
.66905 .01472 L
.66922 .01472 L
.66951 .01472 L
.66981 .01472 L
.67007 .01472 L
.67035 .01472 L
.67066 .01472 L
.67083 .01472 L
.67099 .01472 L
.67129 .01472 L
.67158 .01472 L
.67187 .01472 L
.67214 .01472 L
.67244 .01472 L
Mistroke
.67273 .01472 L
.67299 .01472 L
.67324 .01472 L
.67351 .01472 L
.67379 .01472 L
.67408 .01472 L
.67439 .01472 L
.67468 .01472 L
.67494 .01472 L
.67525 .01472 L
.67543 .01472 L
.67559 .01472 L
.67589 .01472 L
.6762 .01472 L
.67648 .01472 L
.67678 .01472 L
.6771 .01472 L
.67739 .01472 L
.67767 .01472 L
.67792 .01472 L
.67821 .01472 L
.67848 .01472 L
.67878 .01472 L
.67909 .01472 L
.67936 .01472 L
.67966 .01472 L
.67998 .01472 L
.68015 .01472 L
.68032 .01472 L
.68063 .01472 L
.68093 .01472 L
.68119 .01472 L
.68147 .01472 L
.68176 .01472 L
.68203 .01472 L
.68229 .60332 L
.68253 .60332 L
.68306 .60332 L
.68537 .60332 L
.69034 .60332 L
.69276 .60332 L
.69497 .60332 L
.69607 .60332 L
.69637 .60332 L
.69669 .60332 L
.69696 .60332 L
.69726 .01472 L
.69754 .01472 L
.6978 .01472 L
.6981 .01472 L
Mistroke
.69838 .01472 L
.69864 .01472 L
.69888 .01472 L
.69914 .01472 L
.69943 .01472 L
.69958 .01472 L
.69975 .01472 L
.70005 .01472 L
.70036 .01472 L
.70055 .01472 L
.70071 .01472 L
.70102 .01472 L
.7013 .01472 L
.70146 .01472 L
.70162 .01472 L
.70193 .01472 L
.7022 .01472 L
.7025 .01472 L
.70281 .01472 L
.70311 .01472 L
.70338 .01472 L
.70363 .01472 L
.70391 .01472 L
.7042 .01472 L
.7045 .01472 L
.70466 .01472 L
.70482 .01472 L
.70511 .01472 L
.70541 .01472 L
.70568 .01472 L
.70593 .01472 L
.70621 .01472 L
.70651 .01472 L
.70683 .01472 L
.70701 .01472 L
.70718 .01472 L
.70748 .01472 L
.70765 .01472 L
.7078 .01472 L
.7081 .01472 L
.70841 .01472 L
.70871 .01472 L
.70899 .01472 L
.70924 .01472 L
.70952 .01472 L
.70978 .01472 L
.71002 .01472 L
.7103 .01472 L
.7106 .01472 L
.71089 .01472 L
Mistroke
.71115 .01472 L
.71143 .01472 L
.71168 .01472 L
.71196 .01472 L
.71226 .60332 L
.71258 .60332 L
.71288 .60332 L
.71344 .60332 L
.71829 .60332 L
.7227 .60332 L
.72387 .60332 L
.72515 .60332 L
.72574 .60332 L
.72635 .60332 L
.72661 .60332 L
.72688 .60332 L
.72718 .01472 L
.72746 .01472 L
.72763 .01472 L
.7278 .01472 L
.72812 .01472 L
.72829 .01472 L
.72845 .01472 L
.72863 .01472 L
.72883 .01472 L
.72898 .01472 L
.72916 .01472 L
.72934 .01472 L
.7295 .01472 L
.7298 .01472 L
.73012 .01472 L
.73043 .01472 L
.73061 .01472 L
.73077 .01472 L
.73107 .01472 L
.73139 .01472 L
.73168 .01472 L
.73194 .01472 L
.73223 .01472 L
.73255 .01472 L
.73271 .01472 L
.73287 .01472 L
.73316 .01472 L
.73348 .01472 L
.73366 .01472 L
.73382 .01472 L
.73409 .01472 L
.73439 .01472 L
.73471 .01472 L
.735 .01472 L
Mistroke
.73517 .01472 L
.73535 .01472 L
.73552 .01472 L
.73568 .01472 L
.73598 .01472 L
.73629 .01472 L
.73646 .01472 L
.73663 .01472 L
.73695 .01472 L
.73714 .01472 L
.73733 .01472 L
.7375 .01472 L
.73768 .01472 L
.73784 .01472 L
.73801 .01472 L
.73832 .01472 L
.73862 .01472 L
.7389 .01472 L
.7392 .01472 L
.73937 .01472 L
.73953 .01472 L
.7397 .01472 L
.73986 .01472 L
.74003 .01472 L
.7402 .01472 L
.74049 .01472 L
.74079 .01472 L
.74104 .01472 L
.74133 .01472 L
.74163 .01472 L
.7418 .01472 L
.74196 .60332 L
.74225 .60332 L
.74255 .60332 L
.75178 .60332 L
.75422 .60332 L
.75542 .60332 L
.75594 .60332 L
.75622 .60332 L
.75652 .60332 L
.75684 .6033 L
.75714 .01472 L
.75732 .01472 L
.75749 .01472 L
.7578 .01472 L
.75797 .01472 L
.75814 .01472 L
.75846 .01472 L
.75866 .01472 L
.75884 .01472 L
Mistroke
.759 .01472 L
.75918 .01472 L
.75947 .01472 L
.75979 .01472 L
.75996 .01472 L
.76013 .01472 L
.76044 .01472 L
.76075 .01472 L
.76105 .01472 L
.76131 .01472 L
.7616 .01472 L
.7619 .01472 L
.76217 .01472 L
.76246 .01472 L
.76278 .01472 L
.76305 .01472 L
.7633 .01472 L
.76358 .01472 L
.76388 .01472 L
.76405 .01472 L
.76421 .01472 L
.76438 .01472 L
.76456 .01472 L
.76486 .01472 L
.76504 .01472 L
.7652 .01472 L
.7655 .01472 L
.76578 .01472 L
.76609 .01472 L
.7664 .01472 L
.76656 .01472 L
.76672 .01472 L
.76701 .01472 L
.76732 .01472 L
.7675 .01472 L
.76766 .01472 L
.76793 .01472 L
.76823 .01472 L
.76854 .01472 L
.76883 .01472 L
.769 .01472 L
.76918 .01472 L
.76935 .01472 L
.7695 .01472 L
.7698 .01472 L
.77012 .01472 L
.77028 .01472 L
.77045 .01472 L
.77077 .01472 L
.77096 .01472 L
Mistroke
.77114 .01472 L
.77131 .01472 L
.77149 .01472 L
.77165 .01472 L
.77181 .60288 L
.77211 .60332 L
.77243 .60332 L
.77278 .60332 L
.77399 .60332 L
.77625 .60332 L
.78134 .60332 L
.78376 .60332 L
.78496 .60332 L
.78606 .60332 L
.78638 .60332 L
.78668 .60332 L
.78699 .01472 L
.78715 .01472 L
.78733 .01472 L
.7875 .01472 L
.78768 .01472 L
.78787 .01472 L
.78805 .01472 L
.78837 .01472 L
.78855 .01472 L
.78871 .01472 L
.78902 .01472 L
.78931 .01472 L
.78962 .01472 L
.78978 .01472 L
.78996 .01472 L
.79025 .01472 L
.79056 .01472 L
.79086 .01472 L
.79113 .01472 L
.79144 .01472 L
.79174 .01472 L
.792 .01472 L
.79229 .01472 L
.7926 .01472 L
.79276 .01472 L
.79293 .01472 L
.79324 .01472 L
.79352 .01472 L
.79381 .01472 L
.79408 .01472 L
.79433 .01472 L
.79459 .01472 L
.79488 .01472 L
.79519 .01472 L
Mistroke
.79545 .01472 L
.79574 .01472 L
.79606 .01472 L
.79622 .01472 L
.79639 .01472 L
.7967 .01472 L
.79699 .01472 L
.79716 .01472 L
.79733 .01472 L
.79764 .01472 L
.79781 .01472 L
.79797 .01472 L
.79815 .01472 L
.79833 .01472 L
.79862 .01472 L
.79893 .01472 L
.79922 .01472 L
.79949 .01472 L
.7998 .01472 L
.79998 .01472 L
.80014 .01472 L
.80044 .01472 L
.80076 .01472 L
.80107 .01472 L
.80124 .01472 L
.8014 .01472 L
.80157 .01472 L
.80172 .01472 L
.8019 .60332 L
.80207 .60332 L
.80327 .60332 L
.80597 .60332 L
.81086 .60332 L
.81321 .60332 L
.8145 .60332 L
.81513 .60332 L
.81571 .60332 L
.81596 .60332 L
.81623 .60332 L
.81649 .60332 L
.81673 .01472 L
.81701 .01472 L
.81726 .01472 L
.81756 .01472 L
.81783 .01472 L
.81812 .01472 L
.81842 .01472 L
.81868 .01472 L
.81896 .01472 L
.81926 .01472 L
Mistroke
.81943 .01472 L
.81959 .01472 L
.81989 .01472 L
.82018 .01472 L
.82045 .01472 L
.82075 .01472 L
.82103 .01472 L
.82129 .01472 L
.82157 .01472 L
.82189 .01472 L
.82204 .01472 L
.82221 .01472 L
.82252 .01472 L
.82268 .01472 L
.82284 .01472 L
.82301 .01472 L
.82319 .01472 L
.82349 .01472 L
.82366 .01472 L
.82382 .01472 L
.82412 .01472 L
.82439 .01472 L
.8247 .01472 L
.825 .01472 L
.82528 .01472 L
.82554 .01472 L
.82583 .01472 L
.82611 .01472 L
.82637 .01472 L
.8266 .01472 L
.82686 .01472 L
.82714 .01472 L
.82742 .01472 L
.82771 .01472 L
.82799 .01472 L
.82825 .01472 L
.82855 .01472 L
.82872 .01472 L
.82887 .01472 L
.82916 .01472 L
.82947 .01472 L
.82974 .01472 L
.83005 .01472 L
.83036 .01472 L
.83066 .01472 L
.83093 .01472 L
.83118 .01472 L
.83147 .01472 L
.83174 .60332 L
.83204 .60332 L
Mistroke
.83236 .60332 L
.83292 .60332 L
.8342 .60332 L
.83863 .60332 L
.84092 .60332 L
.84338 .60332 L
.84458 .60332 L
.84516 .60332 L
.84571 .60332 L
.8462 .60332 L
.84645 .60332 L
.84672 .01472 L
.847 .01472 L
.84729 .01472 L
.84757 .01472 L
.84783 .01472 L
.84814 .01472 L
.84843 .01472 L
.84869 .01472 L
.84897 .01472 L
.84928 .01472 L
.84944 .01472 L
.84961 .01472 L
.84991 .01472 L
.8502 .01472 L
.85048 .01472 L
.85076 .01472 L
.851 .01472 L
.85126 .01472 L
.85155 .01472 L
.85186 .01472 L
.85215 .01472 L
.85241 .01472 L
.85272 .01472 L
.85288 .01472 L
.85305 .01472 L
.85336 .01472 L
.85365 .01472 L
.85382 .01472 L
.85399 .01472 L
.8543 .01472 L
.85447 .01472 L
.85463 .01472 L
.85481 .01472 L
.85499 .01472 L
.85528 .01472 L
.85559 .01472 L
.85588 .01472 L
.85615 .01472 L
.85645 .01472 L
Mistroke
.85661 .01472 L
.85677 .01472 L
.85709 .01472 L
.85726 .01472 L
.85742 .01472 L
.85771 .01472 L
.85798 .01472 L
.85828 .01472 L
.8586 .01472 L
.85876 .01472 L
.85893 .01472 L
.85925 .01472 L
.85956 .01472 L
.85985 .01472 L
.86014 .01472 L
.86041 .01472 L
.86066 .01472 L
.86093 .01472 L
.86122 .01472 L
.86153 .01472 L
.86179 .60332 L
.86208 .60332 L
.86712 .60332 L
.87186 .60332 L
.87323 .60332 L
.87452 .60332 L
.87568 .60332 L
.876 .60332 L
.87617 .60332 L
.87633 .60332 L
.87662 .01472 L
.87694 .01472 L
.87724 .01472 L
.87741 .01472 L
.87756 .01472 L
.87786 .01472 L
.87814 .01472 L
.8783 .01472 L
.87847 .01472 L
.87877 .01472 L
.87894 .01472 L
.87909 .01472 L
.87927 .01472 L
.87945 .01472 L
.87976 .01472 L
.87993 .01472 L
.88009 .01472 L
.88039 .01472 L
.8807 .01472 L
.88098 .01472 L
Mistroke
.88124 .01472 L
.88153 .01472 L
.88184 .01472 L
.88212 .01472 L
.88239 .01472 L
.88268 .01472 L
.88298 .01472 L
.88314 .01472 L
.88331 .01472 L
.88361 .01472 L
.88391 .01472 L
.88418 .01472 L
.88444 .01472 L
.88468 .01472 L
.88494 .01472 L
.88523 .01472 L
.88552 .01472 L
.8858 .01472 L
.88606 .01472 L
.88633 .01472 L
.88649 .01472 L
.88666 .01472 L
.88695 .01472 L
.88727 .01472 L
.88745 .01472 L
.88762 .01472 L
.88793 .01472 L
.88821 .01472 L
.88836 .01472 L
.88853 .01472 L
.88884 .01472 L
.88911 .01472 L
.88941 .01472 L
.88972 .01472 L
.89002 .01472 L
.89029 .01472 L
.89054 .01472 L
.89082 .01472 L
.89111 .01472 L
.89143 .01472 L
.89172 .60332 L
.89188 .60332 L
.89205 .60332 L
.89237 .60332 L
.89354 .60332 L
.89619 .60332 L
.90092 .60332 L
.90343 .60332 L
.90481 .60332 L
.90549 .60332 L
Mistroke
.90579 .60332 L
.90612 .60332 L
.90643 .60332 L
.90672 .01472 L
.90703 .01472 L
.9072 .01472 L
.90736 .01472 L
.90767 .01472 L
.90785 .01472 L
.90802 .01472 L
.90819 .01472 L
.90835 .01472 L
.90853 .01472 L
.90871 .01472 L
.90899 .01472 L
.90929 .01472 L
.90955 .01472 L
.90984 .01472 L
.91014 .01472 L
.91031 .01472 L
.91047 .01472 L
.91078 .01472 L
.91107 .01472 L
.91136 .01472 L
.91162 .01472 L
.91192 .01472 L
.91221 .01472 L
.91248 .01472 L
.91272 .01472 L
.91299 .01472 L
.91327 .01472 L
.91356 .01472 L
.91387 .01472 L
.91416 .01472 L
.91442 .01472 L
.91473 .01472 L
.91491 .01472 L
.91507 .01472 L
.91537 .01472 L
.91568 .01472 L
.91596 .01472 L
.91626 .01472 L
.91658 .01472 L
.91687 .01472 L
.91715 .01472 L
.9174 .01472 L
.91768 .01472 L
.91796 .01472 L
.91825 .01472 L
.91857 .01472 L
Mistroke
.91884 .01472 L
.91913 .01472 L
.91945 .01472 L
.91963 .01472 L
.91979 .01472 L
.92011 .01472 L
.92041 .01472 L
.92069 .01472 L
.92094 .01472 L
.92123 .01472 L
.92151 .60332 L
.92176 .60332 L
.922 .60332 L
.92253 .60332 L
.92484 .60332 L
.9298 .60332 L
.93222 .60332 L
.93442 .60332 L
.93552 .60332 L
.93582 .60332 L
.93614 .60332 L
.93641 .01472 L
.93671 .01472 L
.93699 .01472 L
.93725 .01472 L
.93753 .01472 L
.93783 .01472 L
.93809 .01472 L
.93833 .01472 L
.93859 .01472 L
.93888 .01472 L
.93903 .01472 L
.9392 .01472 L
.93949 .01472 L
.93981 .01472 L
.93999 .01472 L
.94016 .01472 L
.94047 .01472 L
.94075 .01472 L
.9409 .01472 L
.94107 .01472 L
.94137 .01472 L
.94165 .01472 L
.94195 .01472 L
.94226 .01472 L
.94255 .01472 L
.94283 .01472 L
.94308 .01472 L
.94335 .01472 L
.94364 .01472 L
Mistroke
.94394 .01472 L
.94411 .01472 L
.94426 .01472 L
.94455 .01472 L
.94485 .01472 L
.94512 .01472 L
.94538 .01472 L
.94565 .01472 L
.94595 .01472 L
.94627 .01472 L
.94645 .01472 L
.94661 .01472 L
.94692 .01472 L
.94709 .01472 L
.94724 .01472 L
.94754 .01472 L
.94785 .01472 L
.94815 .01472 L
.94842 .01472 L
.94868 .01472 L
.94896 .01472 L
.94922 .01472 L
.94946 .01472 L
.94973 .01472 L
.95004 .01472 L
.95032 .01472 L
.95059 .01472 L
.95086 .01472 L
.95112 .01472 L
.9514 .60332 L
.9517 .60332 L
.95287 .60332 L
.95764 .60332 L
.96016 .60332 L
.96285 .60332 L
.9642 .60332 L
.96486 .60332 L
.96546 .60332 L
.96575 .60332 L
.96602 .60332 L
.96629 .01472 L
.9666 .01472 L
.96691 .01472 L
.9672 .01472 L
.96735 .01472 L
.96752 .01472 L
.96783 .01472 L
.96815 .01472 L
.96844 .01472 L
.96871 .01472 L
Mistroke
.969 .01472 L
.96931 .01472 L
.96947 .01472 L
.96964 .01472 L
.96995 .01472 L
.97024 .01472 L
.97053 .01472 L
.9708 .01472 L
.97105 .01472 L
.97131 .01472 L
.9716 .01472 L
.97191 .01472 L
.97217 .01472 L
.97246 .01472 L
.97269 .01472 L
.97295 .01472 L
.97318 .01472 L
.9734 .01472 L
.97362 .01472 L
.97382 .01472 L
.97404 .01472 L
.97426 .01472 L
.97449 .01472 L
.97474 .01472 L
.97495 .01472 L
.97519 .01472 L
.97544 .01472 L
.97571 .01472 L
.97596 .01472 L
.97619 .01472 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00J@3oool00`2oM0000000000000040000000904RLg`3oool0oooo0?ooo`3ogi`0B0000000
000007@0_ooo03d0oooo00@0oooO09a80000B9`0gooo0`3oool01@3ogi`0B0000000000007@0_ooo
03/0oooo00D0omnL04P000000000000007Boo`040?ooo`050?oOW018000000000000M02oool0>`3o
ool01@3OW4P00000000000000000B9cO00@0oooo00D0omnL04P000000000001d0;ooo`0g0?ooo`06
0;md00000000000000000000B9`0gooo0`3oool01@3ogi`0B0000000000007@0_ooo00@0oooo00D0
omnL04P000000000001d0;ooo`0>0?ooo`00JP3oool00`3OW4P00000000000030000000:0018W03O
ool0oooo0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oooc`0oooo00@0oooO09a80000B9`0gooo0P3o
ool01`3ookl0M00007Boo`3oool0okmd0000M02oool0>@3oool01`3ogi`0B00004RLg`3oool0oono
07@0001d_ol00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0>@3oool01`3OW4P00000
07Boo`3oool0oooO09a80018W=l00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0=`3o
ool00`3ogi`0B00007Boo`040?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`07
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0=0?ooo`00J`3oool00`3OW4P0B9cO0?ooo`020?oo
o`060=nLM02Lgol0oooo0;md0000B9`0gooo0P3oool00`2oM00007Bo0?ooo`0k0?ooo`040?oog`2L
B00004RL0=ooo`80oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`0g0?ooo`030?oo
g`2LB00007Bo00<0oooo00L0okmd0018W03Oool0oooo0;md0000B9`0gooo0080oooo00<0_g@0001d
_`3oool0=`3oool00`3ooml0W4P0001d_`030?ooo`070?noM000B9`0gooo0?ooo`2oM00004RL0=oo
o`020?ooo`030;md0000M;l0oooo03H0oooo00<0omnL04P0001d_ol0103oool00`2oM00004RL0=oo
o`020?ooo`070;md0000M;l0oooo0?ooo`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00`0
oooo001/0?ooo`030=nLB01d_ol0oooo00<0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d
0;ooo`0g0?ooo`030;md00000000000000D0000000@0M;oo0?oo_`1d0000B9cO0`3oool00`3OW4P0
001d0;ooo`0g0?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0oono07@00018W=l3
0?ooo`030=nLB00007@0_ooo03P0oooo00<0_g@0001d_`3oool00P3oool01P2oM00004RL0=ooo`3o
okl0M00004RLg`<0oooo00<0gia80000M02oool0=P3oool00`3ogi`0B00007Boo`030?ooo`030?oo
_`1d0000B9cO00<0oooo00H0gia80000M02oool0oono07@00018W=l30?ooo`030=nLB00007@0_ooo
00`0oooo001]0?ooo`030;mdB02Lgol0oooo0080oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`0g0?ooo`0<0?oOW01dW=l0oooo0?ooo`3ooml0W4P00018W03Oool0oooo0?oOW0180000
M;oo0`3oool00`3o_g@0001809cOo`0g0?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02L
gol0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo03P0oooo00/0omnL04P0M02oool0oooo0;md
00000000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo03H0oooo00<0omnL04P0
001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3o
ool00`3o_g@0001809cOo`0<0?ooo`00K@3oool00`3ooml0W4Qd0;ooo`020?ooo`030?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol0>03oool02`3OW4P0M;oo0?ooo`3ooml0W4P00018W03Oool0
oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0=`3oool00`3ogi`0B00007Boo`030?oo
o`060;md000004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0i0?ooo`050?oo_`1d
B4P0B00000000018W=l00P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo03H0oooo
00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo0?oOW018
0000M;oo0`3oool00`3o_g@0001809cOo`0<0?ooo`00KP3oool01`3o_g@007Bo0?ooo`3oool0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`0h0?ooo`0;0?oog`2LB7@0_ooo0?oog`2LB00004RL
0=ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0g0?ooo`0<0?oo_`1d00000000
04RLg`3oool0okmd00000000M;l0oooo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0i0?oo
o`040?oOW01800000000001d_`<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0f
0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool01P3o_g@0001809cOo`3o
gi`0B00007Boo`<0oooo00<0okmd0000B02Lgol0303oool006h0oooo00L0oono07@0B02Lgol0oooo
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0>@3oool02P3ogi`0B7Bo0?oog`2LB00004RL
0=ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo03P0oooo00<0_g@00018W02oW4P0
0P0000001P00B9`0gooo0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol0>03oool01P3o
gi`0B0000000M02oool0okmd001d_`80oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`0f0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool01P3o_g@0001809cO
o`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol0303oool006h0oooo00L0oono07@00018W=l0
oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0>P3oool02@3OW4P0M;oO09a80000B9`0
gooo0?ooo`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo03P0oooo00<0omnL04P0001d_ol0
1@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo03P0oooo00<0_g@00000M02oool0
0P3oool01P2oM000B9cO0?ooo`3ookl0M00004RLg`<0oooo00<0gia80000M02oool0=P3oool00`3o
gi`0B00007Boo`030?ooo`030?oo_`1d0000B9cO00<0oooo00H0gia80000M02oool0oono07@00018
W=l30?ooo`030=nLB00007@0_ooo00`0oooo001Y0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL
04P00018W=l00P3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo03X0oooo00D0oooO
09a8B018B00004RL0=ooo`020?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0>@3o
ool00`3OW4P0001809cOo`050?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0>03o
ool00`2oM00004RL0=ooo`020?ooo`070=nLB000M;l0oooo0?ooo`3OW4P004RL0=ooo`020?ooo`03
0;md0000M;l0oooo03H0oooo00<0omnL04P0001d_ol0103oool00`3OW4P004RL0=ooo`020?ooo`07
0;md0000M;l0oooo0?ooo`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00`0oooo001Y0?oo
o`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`0l0?ooo`040?oOW018000004RL0=ooo`80oooo00L0oono07@0001d_ol0oooo0?no
M00007@0_ooo03/0oooo00<0okmd00000000M;l0103oool01`3ookl0M00007Boo`3oool0okmd0000
M02oool0>@3oool01`3o_g@0000007Boo`3oool0oono07@0001d_ol00P3oool01`3ookl0M00007Bo
o`3oool0okmd0000M02oool0=@3oool01@3ooml0W4P0000000000000M;oo00@0oooo00L0oono07@0
001d_ol0oooo0?noM00007@0_ooo0080oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00d0
oooo001Z0?ooo`060?oog`2LB00000000000000007@0_ooo103oool01@3ookl0M0000000000007@0
_ooo03h0oooo00<0gia80018W03Oool00`3oool01@3ookl0M0000000000007@0_ooo03d0oooo00D0
oooO09a800000000001d0;ooo`020?ooo`050?oo_`1d000000000000M02oool0>`3oool01@3o_g@0
0000000000000000B9cO00@0oooo00D0oono07@000000000001d0;ooo`0h0?ooo`030?noM0000000
M;oo00D0oooo00D0oono07@000000000001d0;ooo`040?ooo`050?oo_`1d000000000000M02oool0
3P3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool002T0oooo
00<000000?ooo`3oool0o`3ooon50?ooo`00:@3oool00`000000oooo0?ooo`3o0?ooohD0oooo000Y
0?ooo`030000003oool0oooo0?l0ooooQ@3oool002T0oooo00<000000?ooo`3oool0o`3ooon50?oo
o`00:@3oool00`000000oooo0?ooo`3o0?ooohD0oooo000Y0?ooo`030000003oool0oooo0?l0oooo
Q@3oool00200ooooo`00002100000100oooo000Y0?ooo`030000003oool0oooo00@0oooo00D00000
0?ooo`3oool0oooo000000050?ooo`800000103oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`0000000`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00L000000?ooo`3oool0oooo0000003oool0000000D0oooo00<000000?ooo`3o
ool00`3oool010000000oooo0?ooo`3oool2000000D0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`800000103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`0000000`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00L000000?ooo`3oool0oooo0000003oool0000000D0oooo00<000000?ooo`3oool0
0`3oool010000000oooo0?ooo`3oool2000000D0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`800000103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`0000000`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00@000000?ooo`3oool0oooo0P0000050?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo000000030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool0000001T0oooo000Y0?ooo`030000003o
ool0oooo00@0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo0?ooo`80
00001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3o
ool000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool2000000@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo0000000I0?ooo`00:@3oool00`00
0000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K
0?ooo`00:@3oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000001@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo
0?ooo`0000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?oo
o`00:@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool20000
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo
000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0
oooo000Y0?ooo`800000103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool06`3oool002T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3o
ool002T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo0P00
00040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?oo
o`00:@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0K0?ooo`003@3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00D0
okmd0000001d_ol0oooo0;md00060000000304RLg`3oool0oooo00H0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
6`3oool000`0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00<0okmd0000001d
_ol00P3oool00`3OW4P0000000000003000000030018W03Oool0oooo00H0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool06`3oool000`0oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030=nL
B018W=l0oooo0080oooo00<0giad09cOo`3oool01P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`002`3o
ool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00P0oooo00<0gia807Boo`3oool02P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0K0?ooo`002`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo00T0oooo00<0_gA809cOo`3oool02@3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`002`3oool00`3o
gi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00<0oooO09a8M02oool02@3oool30000
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01/0oooo000;0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3o_g@0
07Bo0?ooo`080?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000;0?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol02P3oool00`3ookl0M01809cOo`080?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo
000;0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02P3oool00`3ookl0M00004RL
g`080?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo01/0oooo000<0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0
001d_`3oool01@3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO00P0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool06`3oool000`0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00H0oooo00P0
oooO09a8B02Lgol0oooo0?oo_`1d0000001d0;ooo`P0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool000d0
oooo00D0oono07@000000000001d0;ooo`080?ooo`060?oog`2LB00000000000000007@0_ooo2@3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0K0?ooo`00:@3oool2000000@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0
oooo000Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo01/0oooo000Y0?ooo`800000103oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool06`3oool002T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3o
ool002T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool06`3oool002T0oooo0P0000040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`003@3oool01@3ogi`0B000
0000000007@0_ooo00<0oooo00<0okmd0000001d_ol01@3oool0103ooml0W4P00018W03Oool80?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo01/0oooo000<0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`02
0?ooo`030?noM0000000M;oo00D0oooo00@0oooO09a80000B9`0gooo203oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0K0?ooo`00303oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00T0oooo00@0oooO
09a80000B9`0gooo203oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`002`3oool00`3ookl0M00004RLg`03
0?ooo`030=nLB00007@0_ooo00D0oooo00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`05
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo01/0oooo000;0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol01@3oool0203ogi`0M9cO0?ooo`3oool0oooO09a80000B9`0gooo203oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0K0?ooo`002`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0
gia807Boo`3oool0oooO09a80000B9`0gooo00P0oooo0`0000030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`002`3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0oooO09a8M02oool0oooO09a80000B9`0gooo
00P0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool06`3oool000/0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`070?ooo`060?oOW018M;l0oooO09a80000B9`0gooo203oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K
0?ooo`002`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00P0oooo00D0gia807Bo
g`2LB00004RL0=ooo`080?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000<0?ooo`030=nLB000B9`0gooo
0080oooo00<0_g@0001d_`3oool0203oool01@3ooml0W4Q804Q80000B9`0gooo00P0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool06`3oool000`0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00X0oooo00@0
omnL04P00000B9`0gooo203oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`003@3oool01@3ookl0M0000000
000007@0_ooo00`0oooo00<0gia80018W03Oool0203oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0K0?ooo`00:@3oool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo01/0oooo000Y0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0
oooo000Y0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000Y
0?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo0P0000030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00
:@3oool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo01/0oooo000Y0?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo0P0000030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool20000
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
01/0oooo000Y0?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool06`3oool002T0oooo0P0000030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool2000000<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y
0?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool06`3oool002T0oooo0P0000030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool2000000<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`800000
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
6`3oool002T0oooo0P0000030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0K0?ooo`00:@3oool2000000<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`8000000`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool000d0
oooo00D0omnL04P000000000001d0;ooo`030?ooo`030?noM0000000M;oo00<0oooo00D0omnL04P0
00000000000007Boo`090?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool06`3oool000`0oooo00L0oono07@0001d_ol0oooo0?noM000
07@0_ooo0080oooo00<0okmd0000001d_ol00P3oool01`3ogi`0B00004RLg`3oool0oono07@0001d
_ol0203oool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo01/0oooo000<0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool0
1@3oool00`3ooml0W4P0001d_`030?ooo`030?noM000B9`0gooo00L0oooo0P0000030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`002`3o
ool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00D0oooo00<0omnL04P0001d_ol00`3o
ool00`3o_g@0001809cOo`070?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool06`3oool000/0oooo00<0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`050?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool2
000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo01/0oooo000;0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01@3oool00`3o
gi`0B00007Boo`030?ooo`030;md000004P0W=oo00L0oooo0`0000020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`002`3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00D0oooo00P0oono07@000000000B9cO0?ooo`3o_g@0
0000001d_`P0oooo0P0000030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0K0?ooo`002`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo00H0oooo00<0_g@00018W02oW4P00P0000000`00B9`0gooo0?ooo`070?ooo`8000000`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool0
00/0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`030?oOW0180000M;oo
00`0oooo0P0000030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0K0?ooo`00303oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0
oooo00<0gia80000B02Lgol02`3oool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000<0?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`090?ooo`030?noM000000007Bo00X0oooo0P0000030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`003@3oool01@3ookl0
M0000000000007@0_ooo00/0oooo00D0oooO09a800000000001d0;ooo`070?ooo`8000000`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool0
02T0oooo0P0000030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0K0?ooo`00:@3oool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`8000000`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo0P00
00030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0K0?ooo`00:@3oool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo01/0oooo000Y0?ooo`8000000`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo0P0000030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00
:@3oool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo01/0oooo000Y0?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo0P0000030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`00
0000oooo000000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo000000020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo
000000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0K0?ooo`00:@3oool00`000000oooo000000020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo00000002
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K
0?ooo`00:@3oool00`000000oooo000000020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool300000080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000
003oool000000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo01/0oooo000Y0?ooo`030000003oool000000080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0
00000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo01/0oooo000Y0?ooo`030000003oool000000080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool000000080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0
oooo000Y0?ooo`030000003oool000000080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool000000080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y
0?ooo`030000003oool000000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool000000080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`03
0000003oool000000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo01/0oooo000Y0?ooo`<000000P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`0000000P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3o
ool002T0oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`0000000P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0
oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool06`3oool002T0oooo00H000000?ooo`3oool000000?ooo`0000050?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00H000000?ooo`3o
ool000000?ooo`0000050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool0
00d0oooo00D0omnL04P000000000001d0;ooo`030?ooo`030?noM0000000M;oo00<0oooo00D0gia8
000000000000000004RLg`090?ooo`060000003oool0oooo0000003oool000001@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000<0?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`020?ooo`030?noM0000000M;oo0080oooo00L0gia80000001d_ol0oooo0?oog`2L
B000B9cO00P0oooo00H000000?ooo`3oool000000?ooo`0000050?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool06`3oool000`0oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo
0?ooo`050?ooo`030?oog`2LB00007Bo00<0oooo00<0okmd0018W03Oool01`3oool01P000000oooo
0?ooo`000000oooo000000D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?oo
o`002`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00<0_g@0001d_`3o
ool00P3oool00`2oM00004RL0=ooo`070?ooo`060000003oool0oooo0000003oool000001@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000;0?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol01P3oool01`3ogi`0B01d0;ooo`3oool0_g@00000001d_ol0203o
ool01P000000oooo0?ooo`000000oooo000000D0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0K0?ooo`002`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo
00D0oono07A8B0180000000004RLg`090?ooo`@0000000<0oooo0000003oool0103oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000;0?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol01`3oool0103ogi`0B00000000000M;l:0?ooo`060000003oool0oooo0000
003oool000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000;0?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3ogi`0B0000000M02oool0
okmd001d_`T0oooo00H000000?ooo`3oool000000?ooo`0000050?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool06`3oool000/0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d
0;ooo`060?ooo`030;md000007@0_ooo0080oooo00<0_g@004RLg`3oool01`3oool01P000000oooo
0?ooo`000000oooo000000D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?oo
o`00303oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00H0oooo00<0_g@00018W03O
ool00P3oool00`3OW4P007Bo0?ooo`070?ooo`060000003oool0oooo0000003oool000001@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000<0?ooo`070?oo_`1d0000M;oo
0?ooo`3o_g@0001d0;ooo`070?ooo`070?noM0000000M;oo0?ooo`3ookl0M00007Boo`080?ooo`06
0000003oool0oooo0000003oool000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo01/0oooo000=0?ooo`050?oo_`1d000000000000M02oool02@3oool01@3o_g@0000000000000
0000B9cO00T0oooo00H000000?ooo`3oool000000?ooo`0000050?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool06`3oool002T0oooo00H000000?ooo`3oool000000?ooo`0000050?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00H000000?ooo`3o
ool000000?ooo`0000050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool0
02T0oooo00H000000?ooo`3oool000000?ooo`0000050?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool06`3oool002T0oooo00H000000?ooo`3oool000000?ooo`0000050?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo0P000000103oool000000?oo
o`0000050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<0
00000?ooo`3oool00P0000060?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3o
ool002T0oooo00<000000?ooo`3oool00P0000060?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool00P0000060?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool06`3oool002T0oooo00D000000?ooo`3oool0oooo000000060?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00D000000?ooo`3o
ool0oooo000000060?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0
oooo00D000000?ooo`3oool0oooo000000060?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool06`3oool002T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`8000002@3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K
0?ooo`00:@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool06`3oool002T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo
00P0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool2000000T0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`00:@3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool002T0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000Y
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K
0?ooo`00:@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool06`3oool002T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo01/0oooo000Y0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0K0?ooo`006P3oool01P2oM000000000000000000004RL0=ooo`T0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0
oooo000K0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo01/0oooo000K0?ooo`030?oOW0180000M;oo00/0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000K
0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo01/0oooo000K0?ooo`030?oOW0180000M;oo00/0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo000K0?ooo`03
0?oOW0180000M;oo00/0oooo0`0000090?ooo`H000001@3oool6000000D0oooo1`0000050?ooo`H0
00001@3oool7000000D0oooo1P0000050?ooo`L000001@3oool6000000D0oooo1`0000050?ooo`H0
00001@3oool7000000D0oooo1P0000050?ooo`L000001@3oool6000000D0oooo1`0000040?ooo`L0
00001@3oool7000000@0oooo1`0000050?ooo`H000001@3oool7000000D0oooo1P0000050?ooo`L0
00001@3oool6000000D0oooo1`0000050?ooo`H000001@3oool7000000D0oooo1P0000050?ooo`L0
00001@3oool6000000D0oooo1`0000050?ooo`H000007@3oool001/0oooo00<0omnL04P0001d_ol0
2`3oool00`000000oooo0?ooo`3o0?ooohD0oooo000K0?ooo`030?oOW0180000M;oo00/0oooo00<0
00000?ooo`3oool0o`3ooon50?ooo`006`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0
oooo0?l0ooooQ@3oool001/0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`3o0?oo
ohD0oooo000I0?ooo`050?oog`2LB00000000000001d_ol02`3oool00`000000oooo0?ooo`3o0?oo
ohD0oooo000K0?ooo`030?noM0000000M;oo0?l0ooooT`3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-11.2874, -0.125454, \
0.410595, 0.00664356}}]
}, Closed]],

Cell["\<\
Although our pulse train may appear to be a rather artificial example, you \
will undoubtedly soon encounter some peculiar kinks that are often merely \
artifacts of the sampling algorithm and which disappear upon closer \
examination.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Locating the points", "Subsubsection"],

Cell["\<\
A graphics object contains a list of the points that uses to draw curves.  \
Examination of the placement of those points should offer some insight into \
the adaptive sampling algorithm.  Examining the full form of a graphics \
object, \
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(FullForm[tanPlot]\)], "Input"],

Cell[BoxData[
    TagBox[
      StyleBox[\(Graphics[
          List[List[
              Line[List[
                  List[\(-3.141592391790405`\), 2.617993880685933`*^-7], 
                  List[\(-2.88670272818237`\), 0.26055722071078874`], 
                  List[\(-2.608723248393842`\), 0.5897779827871114`], 
                  List[\(-2.473566025588546`\), 0.789047192556016`], 
                  List[\(-2.347653339586964`\), 1.0172298832208286`], 
                  List[\(-2.2263473457511984`\), 1.3003890667404432`], 
                  List[\(-2.0965832775706224`\), 1.7233311407461886`], 
                  List[\(-2.0291840485828976`\), 2.0265792893102845`], 
                  List[\(-1.9936715556914684`\), 2.2220957266140617`], 
                  List[\(-1.95602428980909`\), 2.4661676071711613`], 
                  List[\(-1.923148129219481`\), 2.7196382582062806`], 
                  List[\(-1.886936557492664`\), 3.0570642830012074`], 
                  List[\(-1.8564406049348516`\), 3.40512106419145`], 
                  List[\(-1.823152878366059`\), 3.8781691049013034`], 
                  List[\(-1.7888717445048734`\), 4.51264625154833`], 
                  List[\(-1.756284317105562`\), 5.3292130014326755`], 
                  List[\(-1.7409285200854718`\), 5.820961866702547`], 
                  List[\(-1.7245585777233723`\), 6.452211780519368`], 
                  List[\(-1.709218336070457`\), 7.178084956348073`], 
                  List[\(-1.695104581966425`\), 8.003039195793262`], 
                  List[\(-1.6772355012547118`\), 9.359530615173025`], 
                  List[\(-1.6678361055984043`\), 10.272685426042909`], 
                  List[\(-1.6579059690394728`\), 11.45073416190837`], 
                  List[\(-1.649843340860473`\), 12.624339243120078`], 
                  List[\(-1.6412001449368085`\), 14.180299415281594`], 
                  List[\(-1.6317165975178773`\), 16.394585723776384`], 
                  List[\(-1.6228628603627002`\), 19.188836415186103`], 
                  List[\(-1.614435552759284`\), 22.900615290335047`], 
                  List[\(-1.609673908912667`\), 25.7088037025339`], 
                  List[\(-1.6052684472840189`\), 28.99745781235199`], 
                  List[\(-1.5971866978572158`\), 37.883811345358446`], 
                  List[\(-1.5886008551306916`\), 56.15955176302964`], 
                  List[\(-1.5809421029723305`\), 98.55980163750921`], 
                  List[\(-1.5737519737922305`\), 338.3344117043103`], 
                  List[\(-1.5658644641708015`\), \(-202.76150583452485`\)], 
                  List[\(-1.5573616823448824`\), \(-74.42994414721348`\)], 
                  List[\(-1.5490756329954618`\), \(-46.03180457491175`\)], 
                  List[\(-1.541546670051891`\), \(-34.17868495723749`\)], 
                  List[\(-1.5328558783232087`\), \(-26.344446845857227`\)], 
                  List[\(-1.5246512206412997`\), \(-21.65538650228559`\)], 
                  List[\(-1.5170535069655857`\), \(-18.589219065508`\)], 
                  List[\(-1.5100483272249412`\), \(-16.441193090270204`\)], 
                  List[\(-1.5023590234790893`\), \(-14.589094578983481`\)], 
                  List[\(-1.494218216058976`\), \(-13.033025888234452`\)], 
                  List[\(-1.486026118460995`\), \(-11.76832689094084`\)], 
                  List[\(-1.4772253935693824`\), \(-10.655870943642753`\)], 
                  List[\(-1.4613272346776642`\), \(-9.098479688197376`\)], 
                  List[\(-1.442421565099917`\), \(-7.746854460853755`\)], 
                  List[\(-1.4251470769172228`\), \(-6.817190925811454`\)], 
                  List[\(-1.3909821200098462`\), \(-5.501228171904876`\)], 
                  List[\(-1.3716566572023905`\), \(-4.955045132223053`\)], 
                  List[\(-1.3537386285511752`\), \(-4.534489109217866`\)], 
                  List[\(-1.3184093374950097`\), \(-3.8776809354034287`\)], 
                  List[\(-1.2859475134868004`\), \(-3.4151670371852627`\)], 
                  List[\(-1.2523478981821852`\), \(-3.0333517743241707`\)], 
                  List[\(-1.216050020733974`\), \(-2.6996630149149436`\)], 
                  List[\(-1.1502894436783027`\), \(-2.236232700615572`\)], 
                  List[\(-1.0905162928522247`\), \(-1.91950800685314`\)], 
                  List[\(-1.0274511073883414`\), \(-1.655667552412321`\)], 
                  List[\(-0.8976320450977752`\), \(-1.254048174198956`\)], 
                  List[\(-0.776593904473269`\), \(-0.9825447120089745`\)], 
                  List[\(-0.5195564563593538`\), \(-0.5719730304987136`\)], 
                  List[\(-0.24633891625605964`\), \(-0.25144576149186065`\)], 
                  List[0.016878777056698428`, 0.01688038012183686`], 
                  List[0.27009662357892045`, 0.2768621631254877`], 
                  List[0.5394945620905214`, 0.598742781426269`], 
                  List[0.6733721559872652`, 0.7977576446162298`], 
                  List[0.7988926538115863`, 1.0273598488204982`], 
                  List[0.9189927430311354`, 1.3105228895920926`], 
                  List[1.0482908987421151`, 1.7364324972494483`], 
                  List[1.1116007459554136`, 2.02246027676789`], 
                  List[1.1807626577888573`, 2.432532059330034`], 
                  List[1.215694550229486`, 2.6967196357405054`], 
                  List[1.2529637764293162`, 3.039646245627995`], 
                  List[1.285571989101508`, 3.41041772516031`], 
                  List[1.3200494808608656`, 3.9041513200825477`], 
                  List[1.3553204011095208`, 4.5688410620144015`], 
                  List[1.3877977710467568`, 5.403387861049481`], 
                  List[1.4058660283294324`, 6.008090622177044`], 
                  List[1.4227480942257071`, 6.705133803925026`], 
                  List[1.4412705248518454`, 7.67724583951819`], 
                  List[1.4611428755475742`, 9.083059515670163`], 
                  List[1.478237148610103`, 10.773028168120183`], 
                  List[1.487899936694879`, 12.035606676059901`], 
                  List[1.496847207988422`, 13.49815288091565`], 
                  List[1.5046424792581787`, 15.094221035002203`], 
                  List[1.5132539662437217`, 17.359316492551994`], 
                  List[1.5175802659522275`, 18.77357742883139`], 
                  List[1.5222478414125558`, 20.581779572365654`], 
                  List[1.530698274916435`, 24.925499979825517`], 
                  List[1.5383920122484598`, 30.84928585243641`], 
                  List[1.5468148683425746`, 41.69088761584124`], 
                  List[1.551064713297856`, 50.67351526646596`], 
                  List[1.5556085927316556`, 65.83754400526755`], 
                  List[1.5637967197347893`, 142.8628292788277`], 
                  List[1.571975727259716`, \(-847.8880296959725`\)], 
                  List[1.5797035139915492`, \(-112.26591872250522`\)], 
                  List[1.5866415139437733`, \(-63.10536437461966`\)], 
                  List[1.5941687065867658`, \(-42.777753608861`\)], 
                  List[1.6022723827579917`, \(-31.759688443527878`\)], 
                  List[1.6096829085721307`, \(-25.70284782267368`\)], 
                  List[1.618236449598304`, \(-21.06338769701428`\)], 
                  List[1.6263633037737193`, \(-17.977774079956642`\)], 
                  List[1.6338411045948853`, \(-15.840721504917026`\)], 
                  List[1.6407799597900201`, \(-14.265719773865088`\)], 
                  List[1.656508478342084`, \(-11.63837243765051`\)], 
                  List[1.6654859397319222`, \(-10.529238355448692`\)], 
                  List[1.673938942211847`, \(-9.660908300325243`\)], 
                  List[1.689735090752876`, \(-8.368004149461955`\)], 
                  List[1.7067461526031569`, \(-7.310282338388057`\)], 
                  List[1.725138280973073`, \(-6.427590519734413`\)], 
                  List[1.7603014309435272`, \(-5.213582601010507`\)], 
                  List[1.778006775182255`, \(-4.756742798418793`\)], 
                  List[1.797431406789406`, \(-4.33657460185153`\)], 
                  List[1.8292819385720092`, \(-3.78213916457858`\)], 
                  List[1.863911058359035`, \(-3.313364213252067`\)], 
                  List[1.8985275899183633`, \(-2.9412463771281763`\)], 
                  List[1.9313041398598`, \(-2.652641464415601`\)], 
                  List[1.9927203992093079`, \(-2.227755381146867`\)], 
                  List[2.0644265304382423`, \(-1.8585280256101193`\)], 
                  List[2.130735494162217`, \(-1.595240327558034`\)], 
                  List[2.253966283337834`, \(-1.228625353594322`\)], 
                  List[2.389920726782013`, \(-0.9347242798395093`\)], 
                  List[2.6552862046006513`, \(-0.5286530766087659`\)], 
                  List[2.910651835628754`, \(-0.2351359778354019`\)], 
                  List[3.141592391790405`, \(-2.617993880685933`*^-7\)]]]]], 
          List[Rule[PlotRange, List[\(-5\), 5]], 
            Rule[AspectRatio, Power[GoldenRatio, \(-1\)]], 
            RuleDelayed[DisplayFunction, $DisplayFunction], 
            Rule[ColorOutput, Automatic], Rule[Axes, Automatic], 
            Rule[AxesOrigin, Automatic], Rule[PlotLabel, None], 
            Rule[AxesLabel, None], Rule[Ticks, Automatic], 
            Rule[GridLines, None], Rule[Prolog, List[]], 
            Rule[Epilog, List[]], Rule[AxesStyle, Automatic], 
            Rule[Background, Automatic], Rule[DefaultColor, Automatic], 
            RuleDelayed[DefaultFont, $DefaultFont], Rule[RotateLabel, True], 
            Rule[Frame, False], Rule[FrameStyle, Automatic], 
            Rule[FrameTicks, Automatic], Rule[FrameLabel, None], 
            Rule[PlotRegion, Automatic], Rule[ImageSize, Automatic], 
            RuleDelayed[TextStyle, $TextStyle], 
            RuleDelayed[FormatType, $FormatType]]]\),
        ShowSpecialCharacters->False,
        ShowStringCharacters->True,
        NumberMarks->True],
      FullForm]], "Output"]
}, Closed]],

Cell[TextData[{
  "we find that the data are contained within a nested structure of the form \
",
  StyleBox["List[List[Line[List[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["pt1,pt2,\[CenterEllipsis]", "InlineInput"],
  StyleBox["]]]]", "InlineInput",
    FontWeight->"Bold"],
  " where each point is itself a list containing its coordinates.  The points \
can be extracted by peeling off 4 layers to produce a list of the underlying \
points.  Each layer is peeled by successive applications of ",
  StyleBox["First", "InlineInput",
    FontWeight->"Bold"],
  "; hence, ",
  StyleBox["First", "InlineInput",
    FontWeight->"Bold"],
  " must be nested 4 times using ",
  StyleBox["Nest", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[BoxData[
    \(getPoints[graphics_]\  := \ Nest[First, graphics, 4]\)], "Input"],

Cell[BoxData[
    \(\(tanPoints\  = \ getPoints[tanPlot];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ListPlot[tanPoints, PlotRange \[Rule] {\(-10\), 10}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.151576 0.309017 0.0309017 [
[.04527 .29652 -9.15625 -15.6875 ]
[.04527 .29652 9.15625 0 ]
[.19685 .29652 -9.15625 -15.6875 ]
[.19685 .29652 9.15625 0 ]
[.34842 .29652 -9.15625 -15.6875 ]
[.34842 .29652 9.15625 0 ]
[.65158 .29652 -5 -15.6875 ]
[.65158 .29652 5 0 ]
[.80315 .29652 -5 -15.6875 ]
[.80315 .29652 5 0 ]
[.95473 .29652 -5 -15.6875 ]
[.95473 .29652 5 0 ]
[.4875 0 -24.3125 -7.84375 ]
[.4875 0 0 7.84375 ]
[.4875 .07725 -27.3125 -7.84375 ]
[.4875 .07725 0 7.84375 ]
[.4875 .15451 -18.3125 -7.84375 ]
[.4875 .15451 0 7.84375 ]
[.4875 .23176 -27.3125 -7.84375 ]
[.4875 .23176 0 7.84375 ]
[.4875 .38627 -19 -7.84375 ]
[.4875 .38627 0 7.84375 ]
[.4875 .46353 -10 -7.84375 ]
[.4875 .46353 0 7.84375 ]
[.4875 .54078 -19 -7.84375 ]
[.4875 .54078 0 7.84375 ]
[.4875 .61803 -16 -7.84375 ]
[.4875 .61803 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.04527 .30902 m
.04527 .31527 L
s
gsave
.04527 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.19685 .30902 m
.19685 .31527 L
s
gsave
.19685 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.34842 .30902 m
.34842 .31527 L
s
gsave
.34842 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.65158 .30902 m
.65158 .31527 L
s
gsave
.65158 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.80315 .30902 m
.80315 .31527 L
s
gsave
.80315 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.95473 .30902 m
.95473 .31527 L
s
gsave
.95473 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.07559 .30902 m
.07559 .31277 L
s
.1059 .30902 m
.1059 .31277 L
s
.13622 .30902 m
.13622 .31277 L
s
.16653 .30902 m
.16653 .31277 L
s
.22716 .30902 m
.22716 .31277 L
s
.25748 .30902 m
.25748 .31277 L
s
.28779 .30902 m
.28779 .31277 L
s
.31811 .30902 m
.31811 .31277 L
s
.37874 .30902 m
.37874 .31277 L
s
.40905 .30902 m
.40905 .31277 L
s
.43937 .30902 m
.43937 .31277 L
s
.46968 .30902 m
.46968 .31277 L
s
.53032 .30902 m
.53032 .31277 L
s
.56063 .30902 m
.56063 .31277 L
s
.59095 .30902 m
.59095 .31277 L
s
.62126 .30902 m
.62126 .31277 L
s
.68189 .30902 m
.68189 .31277 L
s
.71221 .30902 m
.71221 .31277 L
s
.74252 .30902 m
.74252 .31277 L
s
.77284 .30902 m
.77284 .31277 L
s
.83347 .30902 m
.83347 .31277 L
s
.86378 .30902 m
.86378 .31277 L
s
.8941 .30902 m
.8941 .31277 L
s
.92441 .30902 m
.92441 .31277 L
s
.01496 .30902 m
.01496 .31277 L
s
.98504 .30902 m
.98504 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5 0 m
.50625 0 L
s
gsave
.4875 0 -85.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.5 .07725 m
.50625 .07725 L
s
gsave
.4875 .07725 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(7.5) show
1.000 setlinewidth
grestore
.5 .15451 m
.50625 .15451 L
s
gsave
.4875 .15451 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.5 .23176 m
.50625 .23176 L
s
gsave
.4875 .23176 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2.5) show
1.000 setlinewidth
grestore
.5 .38627 m
.50625 .38627 L
s
gsave
.4875 .38627 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2.5) show
1.000 setlinewidth
grestore
.5 .46353 m
.50625 .46353 L
s
gsave
.4875 .46353 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.5 .54078 m
.50625 .54078 L
s
gsave
.4875 .54078 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(7.5) show
1.000 setlinewidth
grestore
.5 .61803 m
.50625 .61803 L
s
gsave
.4875 .61803 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .01545 m
.50375 .01545 L
s
.5 .0309 m
.50375 .0309 L
s
.5 .04635 m
.50375 .04635 L
s
.5 .0618 m
.50375 .0618 L
s
.5 .09271 m
.50375 .09271 L
s
.5 .10816 m
.50375 .10816 L
s
.5 .12361 m
.50375 .12361 L
s
.5 .13906 m
.50375 .13906 L
s
.5 .16996 m
.50375 .16996 L
s
.5 .18541 m
.50375 .18541 L
s
.5 .20086 m
.50375 .20086 L
s
.5 .21631 m
.50375 .21631 L
s
.5 .24721 m
.50375 .24721 L
s
.5 .26266 m
.50375 .26266 L
s
.5 .27812 m
.50375 .27812 L
s
.5 .29357 m
.50375 .29357 L
s
.5 .32447 m
.50375 .32447 L
s
.5 .33992 m
.50375 .33992 L
s
.5 .35537 m
.50375 .35537 L
s
.5 .37082 m
.50375 .37082 L
s
.5 .40172 m
.50375 .40172 L
s
.5 .41717 m
.50375 .41717 L
s
.5 .43262 m
.50375 .43262 L
s
.5 .44807 m
.50375 .44807 L
s
.5 .47898 m
.50375 .47898 L
s
.5 .49443 m
.50375 .49443 L
s
.5 .50988 m
.50375 .50988 L
s
.5 .52533 m
.50375 .52533 L
s
.5 .55623 m
.50375 .55623 L
s
.5 .57168 m
.50375 .57168 L
s
.5 .58713 m
.50375 .58713 L
s
.5 .60258 m
.50375 .60258 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.008 w
.02381 .30902 Mdot
.06244 .31707 Mdot
.10458 .32724 Mdot
.12507 .3334 Mdot
.14415 .34045 Mdot
.16254 .3492 Mdot
.18221 .36227 Mdot
.19242 .37164 Mdot
.19781 .37768 Mdot
.20351 .38523 Mdot
.2085 .39306 Mdot
.21399 .40349 Mdot
.21861 .41424 Mdot
.22365 .42886 Mdot
.22885 .44847 Mdot
.23379 .4737 Mdot
.23612 .48889 Mdot
.2386 .5084 Mdot
.24092 .53083 Mdot
.24306 .55632 Mdot
.24577 .59824 Mdot
.2785 .02786 Mdot
.28136 .06963 Mdot
.28398 .09835 Mdot
.28916 .13902 Mdot
.29209 .1559 Mdot
.29481 .16889 Mdot
.30016 .18919 Mdot
.30508 .20348 Mdot
.31017 .21528 Mdot
.31568 .22559 Mdot
.32564 .23991 Mdot
.3347 .2497 Mdot
.34426 .25785 Mdot
.36394 .27026 Mdot
.38229 .27865 Mdot
.42125 .29134 Mdot
.46266 .30125 Mdot
.50256 .30954 Mdot
.54094 .31757 Mdot
.58177 .32752 Mdot
.60207 .33367 Mdot
.62109 .34076 Mdot
.6393 .34951 Mdot
.6589 .36268 Mdot
.66849 .37151 Mdot
.67898 .38419 Mdot
.68427 .39235 Mdot
.68992 .40295 Mdot
.69486 .4144 Mdot
.70009 .42966 Mdot
.70543 .4502 Mdot
.71036 .47599 Mdot
.7131 .49468 Mdot
.71565 .51622 Mdot
.71846 .54626 Mdot
.72147 .5897 Mdot
.75373 .01048 Mdot
.75612 .05043 Mdot
.7587 .08312 Mdot
.76149 .11039 Mdot
.76682 .14791 Mdot
.7695 .16203 Mdot
.77245 .17501 Mdot
.77728 .19214 Mdot
.78252 .20663 Mdot
.78777 .21813 Mdot
.79274 .22705 Mdot
.80205 .24018 Mdot
.81292 .25159 Mdot
.82297 .25972 Mdot
.84165 .27105 Mdot
.86225 .28013 Mdot
.90248 .29268 Mdot
.94119 .30175 Mdot
.97619 .30902 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00_`3oool01P2oM000000000000000000004RL0=ooo`<0oooo00D0omnL04P000000000001d
0;ooo`3S0?ooo`00`03oool00`3ogi`0B00007Boo`040?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0
001d0;ooo`3R0?ooo`00`03oool00`3ogi`0B00007Boo`040?ooo`030;md0000B9`0gooo0080oooo
00<0_g@0001d_`3oool0h@3oool00<00oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M00004RL
g`030?ooo`030=nLB00007@0_ooo0>40oooo002a0?ooo`030?oog`2LB000000000L0000000<0B9cO
0?ooo`3oool00P3oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol0h@3oool00<00oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00P0oooo0`00003F0?ooo`00`03oool00`3ogi`0B00007Boo`030?ooo`03
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0203oool00`000000oooo0?ooo`3F0?ooo`00
`03oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
203oool00`000000oooo0?ooo`3F0?ooo`00`03oool00`3ogi`0B00007Boo`030?ooo`030?oo_`1d
0000B9cO00<0oooo00<0gia80000M02oool0203oool00`000000oooo0?ooo`3F0?ooo`00`03oool0
0`3ogi`0B00007Boo`040?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0203oool0
0`000000oooo0?ooo`1P0?ooo`030000003oool0oooo07<0oooo002n0?ooo`050?oog`2LB0000000
0000001d_ol0103oool01`3ookl0M00007Boo`3oool0okmd0000M02oool02@3oool00`000000oooo
0?ooo`1O0?ooo`<00000M03oool00<00oooo00<0okmd0000001d_ol01@3oool01@3ookl0M0000000
000007@0_ooo00X0oooo0P00001Q0?ooo`030000003oool0oooo07<0oooo003G0?ooo`030000003o
ool0oooo0=H0oooo003G0?ooo`030000003oool0oooo0=H0oooo003G0?ooo`030000003oool0oooo
0=H0oooo00200?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0eP3oool007l0oooo
0`00001E0?ooo`030000003oool0oooo0=H0oooo00200?ooo`030000003oool0oooo05@0oooo00<0
00000?ooo`3oool0eP3oool00=L0oooo0P00003G0?ooo`00e`3oool00`000000oooo0?ooo`3F0?oo
o`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3o
ool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool20000
0680oooo00<000000?ooo`3oool0LP3oool00=L0oooo00<000000?ooo`3oool0H03oool3000007<0
oooo003G0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool0LP3oool00=L0oooo00<0
00000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?oo
o`3oool0eP3oool00=L0oooo0P00003G0?ooo`00_03oool00`3OW4P0B9cO0?ooo`040?ooo`060?no
M0000000M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`080?ooo`030000003oool0oooo
0=H0oooo00210?ooo`030000003oool0oooo03P0oooo00<0omnL04Q8W03Oool0103oool01`3o_g@0
000007Boo`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`080?ooo`030000003oool0
oooo0=H0oooo00200?ooo`<00000>@3oool00`3ooml0W4Q809cOo`0=0?ooo`030?oog`2LB7@0_ooo
00P0oooo00<000000?ooo`3oool0eP3oool00840oooo00<000000?ooo`3oool0>@3oool00`3OW4P0
B9cO0?ooo`0=0?ooo`030;md001d_ol0oooo00L0oooo00<000000?ooo`3oool0eP3oool00:`0oooo
00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`040?ooo`030?oOW018B9`0gooo00`0oooo
00<0oooO09a80018W=l0203oool00`000000oooo0?ooo`3F0?ooo`00_@3oool00`3ooml0W4Q809cO
o`0<0?ooo`030=nLB0000000M;oo00P0oooo0`00003F0?ooo`00_P3oool00`3OW4P007Bo0?ooo`07
0?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`080?ooo`030000003oool0oooo0680oooo
00<000000?ooo`3oool0L@3oool00;h0oooo00<0oono07@0M02oool01`3oool01P3OW4P000000000
00000000001809cOo`T0oooo00<000000?ooo`3oool0H@3oool300000780oooo002i0?ooo`030?oO
g`3Oool0oooo00<0oooo00<0_g@007Boo`3oool01P3oool0103ookl0M00000000018W=l;0?ooo`03
0000003oool0oooo0680oooo00<000000?ooo`3oool0L@3oool00;T0oooo00<0oono07A8W03Oool0
0`3oool00`3o_g@007Bo0?ooo`070?ooo`030;md001d_ol0oooo00/0oooo00<000000?ooo`3oool0
eP3oool00;X0oooo00<0_g@00000000000000`0000000`0007@0_ooo0?ooo`060?ooo`060?oOW018
000000000000000004P0W=oo203oool00`000000oooo0?ooo`3F0?ooo`00^P3oool00`3o_g@00000
000000040000000307Boo`3oool0oooo00H0oooo00D0_g@0000000000000000004RLg`080?ooo`80
0000e`3oool00880oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`3F0?ooo`00P@3o
ool3000005<0oooo00<000000?ooo`3oool0eP3oool00880oooo00<000000?ooo`3oool0DP3oool0
0`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000
oooo0?ooo`3F0?ooo`00e`3oool2000006@0oooo00<000000?ooo`3oool0L03oool00=L0oooo00<0
00000?ooo`3oool0HP3oool300000740oooo003G0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool0L03oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3o
ool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo0P00003G0?ooo`00e`3oool0
0`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000
oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00Q03oool00`000000oooo0?oo
o`1@0?ooo`030000003oool0oooo0=H0oooo00230?ooo`<00000D@3oool200000=L0oooo00240?oo
o`030000003oool0oooo03l0oooo00<0oooO09a8000000000P0000000`1d_ol0oooo0?ooo`090?oo
o`030000003oool0oooo0=H0oooo00360?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3o
ool02@3oool00`000000oooo0?ooo`1V0?ooo`030000003oool0oooo06d0oooo003;0?ooo`030?oo
g`2LB7@0_ooo00T0oooo00<000000?ooo`3oool0I@3oool3000006h0oooo003<0?ooo`030;md001d
_ol0oooo00P0oooo00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`1]0?ooo`00^P3oool0
0`3ooml0W4P0000000070000000304RLg`3oool0oooo00@0oooo00<0oooO09a80018W=l02@3oool0
0`000000oooo0?ooo`3F0?ooo`00Q@3oool00`000000oooo0?ooo`130?ooo`030=nLB0000000M;oo
00T0oooo0`00003F0?ooo`00Q03oool300000400oooo00L0gmoO0?ooo`3ookl0M0000000000007@0
_ooo00T0oooo00<000000?ooo`3oool0eP3oool008D0oooo00<000000?ooo`3oool0?`3oool01P3O
W4P00000000000000000001809cOo`X0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?oo
o`1/0?ooo`00a`3oool0103ookl0M00000000018W=l<0?ooo`030000003oool0oooo06H0oooo0`00
001]0?ooo`00b03oool00`2oM000M;oo0?ooo`0<0?ooo`030000003oool0oooo06L0oooo00<00000
0?ooo`3oool0K03oool008H0oooo00<000000?ooo`3oool0?`3oool01P3ogi`0B000000000000000
001809cOo`T0oooo00<000000?ooo`3oool0eP3oool008D0oooo0`0000110?ooo`050;md00000000
000000000018W=l02@3oool200000=L0oooo00260?ooo`030000003oool0oooo04h0oooo00<00000
0?ooo`3oool0J03oool00`000000oooo0?ooo`1[0?ooo`00e`3oool00`000000oooo0?ooo`1W0?oo
o`<00000K03oool00=L0oooo00<000000?ooo`3oool0J03oool00`000000oooo0?ooo`1[0?ooo`00
e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool2
00000=L0oooo00290?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0eP3oool008P0
oooo0`00001<0?ooo`030000003oool0oooo06X0oooo00<000000?ooo`3oool0J@3oool008T0oooo
00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`1Y0?ooo`<00000JP3oool00=L0oooo00<0
00000?ooo`3oool0JP3oool00`000000oooo0?ooo`1Y0?ooo`00e`3oool00`000000oooo0?ooo`3F
0?ooo`00R`3oool00`000000oooo0?ooo`190?ooo`800000e`3oool008X0oooo0`00001:0?ooo`03
0000003oool0oooo06`0oooo00<000000?ooo`3oool0I`3oool008/0oooo00<000000?ooo`3oool0
B@3oool00`000000oooo0?ooo`1[0?ooo`<00000J03oool00=L0oooo00<000000?ooo`3oool0K03o
ool00`000000oooo0?ooo`1W0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00S@3oool00`00
0000oooo0?ooo`170?ooo`030000003oool0oooo0=H0oooo002<0?ooo`<00000B03oool2000006l0
oooo00<000000?ooo`3oool0I@3oool008d0oooo00<000000?ooo`3oool0:@3oool00`2oM0000000
000000040000000904RLg`3oool0oooo0?noM0000000M;oo0?ooo`3ooml0W4P000<0000000<0M;oo
0?ooo`3oool0203oool00`000000oooo0?ooo`1]0?ooo`<00000IP3oool00;X0oooo00<0gia80000
000000000`0000002P00B9`0gooo0?ooo`3o_g@0000007Boo`3oool0oooO09a8M02oool20?ooo`03
0;md001d_ol0oooo00P0oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`1U0?ooo`00
S`3oool00`000000oooo0?ooo`0Y0?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0
2@3oool00`3ooml0W4Qd0;ooo`080?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0
H`3oool008h0oooo0`00000[0?ooo`030=nLB01d_ol0oooo00h0oooo00<0_g@007Boo`3oool01`3o
ool00`000000oooo0?ooo`1_0?ooo`<00000I03oool008l0oooo00<000000?ooo`3oool06P3oool0
0`3ooml0W4P0000000070000000304RLg`3oool0oooo00@0oooo00<0_gA809cOo`3oool0303oool0
0`3ooml0W4P004RLg`080?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0H`3oool0
0;d0oooo00<0oooO09a8M02oool0303oool00`3OW4P0000007Boo`080?ooo`<00000eP3oool00;h0
oooo00<0okmd001d_`3oool01`3oool01`3Ogml0oooo0?oo_`1d000000000000M02oool0203oool0
0`000000oooo0?ooo`3F0?ooo`00_P3oool00`3ookl0M01809cOo`070?ooo`060=nLB00000000000
0000000004P0W=oo2@3oool00`000000oooo0?ooo`3F0?ooo`00T`3oool00`000000oooo0?ooo`0X
0?ooo`030?oo_`1d0000B9cO00L0oooo00@0oono07@000000000B9cO2`3oool00`000000oooo0?oo
o`1c0?ooo`030000003oool0oooo0600oooo002B0?ooo`<00000903oool00`3o_i`0W=oo0?ooo`02
0?ooo`030?oOW0180000B9cO00P0oooo00<0_g@007Boo`3oool02`3oool00`000000oooo0?ooo`1b
0?ooo`<00000H@3oool009<0oooo00<000000?ooo`3oool08`3oool0203ooml0W4Q809cOo`3oool0
oono07@0000007@0_ooo203oool01P3ogi`0B000000000000000001809cOo`P0oooo00<000000?oo
o`3oool0L`3oool00`000000oooo0?ooo`1P0?ooo`00UP3oool00`000000oooo0?ooo`0Q0?ooo`06
0?oog`2LB00000000000000007@0_ooo2P3oool01@2oM0000000000000000000B9cO00P0oooo0P00
003G0?ooo`00U@3oool3000003l0oooo00<000000?ooo`3oool0N03oool00`000000oooo0?ooo`1K
0?ooo`00UP3oool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo07L0oooo0`00001L0?oo
o`009P3oool00`3ookl0M000000000020000000307Boo`3oool0oooo03@0oooo00<0_g@000000000
0000100000000`18W=l0oooo0?ooo`0c0?ooo`060;md00000000000000000000B9`0gooo<P3oool0
0`000000oooo0?ooo`0f0?ooo`060;md00000000000000000000B9`0gooo<`3oool00`2oM0000000
000000040000000304RLg`3oool000000380oooo00<0oono07@0000000000P0000000`1d_ol0oooo
0?ooo`0S0?ooo`009P3oool00`3ogi`0B00004RLg`020?ooo`030;md0018W=l0oooo03D0oooo00<0
gia80000000000000`0000000`00B9`0gooo0?ooo`0^0?ooo`030000003oool0oooo00<0oooo00<0
omnL04P0001d_ol0=03oool00`000000oooo0?ooo`0g0?ooo`030?oOW0180000M;oo03H0oooo00<0
gia80000000000000`0000000`00B9`0gooo0?ooo`030?ooo`030000003oool0oooo02`0oooo00<0
omnL04P00018W=l00P3oool00`2oM000B9cO0?ooo`0S0?ooo`00:`3oool00`3ookl0M01809cOo`0f
0?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0;@3oool3000000@0oooo00<0omnL
04P0001d_ol0=03oool00`000000oooo0?ooo`0g0?ooo`030?oOW0180000M;oo03L0oooo00<0gia8
04RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`020?ooo`<00000<P3oool00`3ookl0M01809cOo`0S
0?ooo`00:`3oool00`3ookl0M00007Boo`0g0?ooo`030=nLB01d_ol0oooo0380oooo00<000000?oo
o`3oool00`3oool00`3ogi`0B00007Boo`0d0?ooo`800000>03oool00`3ogi`0B00007Boo`0h0?oo
o`030=nLB01d_ol0oooo00L0oooo00<000000?ooo`3oool0<@3oool00`3ookl0M00007Boo`0S0?oo
o`006P3oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00@0oooo00<0okmd0000001d
_ol09`3oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00@0oooo00<0_gA809cOo`3o
ool0:03oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo0080oooo00<0omnL04P0001d
_ol0=03oool00`000000oooo0?ooo`0g0?ooo`030?oOW0180000M;oo03T0oooo00<0_gA809cOo`3o
ool0>P3oool00`3o_g@0000007Boo`0S0?ooo`00:P3oool0103ogi`0B0000000B02Lgolh0?ooo`03
0?oog`2LB7@0_ooo03L0oooo00<0omnL04P000000000=03oool00`000000oooo0?ooo`0g0?ooo`03
0?oOW0180000M;oo03T0oooo00<0oooO09a8M02oool0>@3oool0103ogi`0B0000000B02LgolS0?oo
o`00:03oool01@3o_g@000000000000004P0W=oo03X0oooo00<0okmd001d_`3oool0=P3oool0103o
gi`0000000000000000c0?ooo`030000003oool0oooo03L0oooo00<0omnL04P0001d_ol0>P3oool0
0`3o_g@007Bo0?ooo`0<0?ooo`030000003oool0oooo02L0oooo00D0okmd000000000000001809cO
o`0T0?ooo`00:P3oool00`2oM00007Bo0?ooo`0j0?ooo`030?oo_`1d04P0W=oo03H0oooo00<0omnL
04P000000000=03oool00`000000oooo0?ooo`0g0?ooo`030?oOW0180000M;oo03X0oooo00<0oono
07@0B02Lgol02`3oool3000002X0oooo00<0_g@0001d_`3oool0903oool002/0oooo00<0_g@004RL
g`3oool0>@3oool00`3ookl0M00004RLg`0f0?ooo`030?oOW0180000M;oo03@0oooo00<000000?oo
o`3oool0=`3oool00`3ogi`0B00007Boo`0j0?ooo`030?oo_`1d0000B9cO00`0oooo00<000000?oo
o`3oool0:P3oool00`2oM000B9cO0?ooo`0S0?ooo`009P3oool00`3ooml0W9bo0?ooo`020?ooo`03
0=nLB00007@0_ooo03@0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`0f0?ooo`03
0?oOW0180000M;oo00H0oooo00<000000?ooo`3oool0:`3oool2000003P0oooo00<0omnL04P0001d
_ol0=@3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO01@0oooo00<000000?ooo`3o
ool07@3oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo02<0oooo000W0?ooo`060?no
M018W=l0oooo0?oo_`1d000007Bo=@3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo
=03oool01@3ooml0W4P0000000000000M;oo00D0oooo0`00000/0?ooo`030000003oool0oooo03D0
oooo00D0oooO09a800000000000007Boo`0e0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000
M02ooolC0?ooo`<000007`3oool01P3o_g@0B9cO0?ooo`3ookl0M000001d_b@0oooo000X0?ooo`05
0=nLB000000000000000M02oool0=P3oool01P3ooml0W4P0000000000000001d0;ooocL0oooo00<0
okmd0000001d_ol01P3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo03L0oooo00<0
okmd0000001d_ol0=P3oool01P3ooml0W4P0000000000000001d0;oooaD0oooo00<000000?ooo`3o
ool07`3oool01@3OW4P000000000000007@0_ooo02@0oooo003G0?ooo`030000003oool0oooo0=H0
oooo003G0?ooo`030000003oool0oooo0=H0oooo002h0?ooo`030000003oool0oooo01`0oooo00<0
00000?ooo`3oool0V`3oool00`000000oooo0?ooo`0h0?ooo`00]`3oool3000001d0oooo00<00000
0?ooo`3oool0VP3oool3000003T0oooo002h0?ooo`030000003oool0oooo01`0oooo0P00002L0?oo
o`030000003oool0oooo03P0oooo003G0?ooo`030000003oool0oooo0=H0oooo00380?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0Y0?ooo`00a`3o
ool3000000d0oooo00<000000?ooo`3oool0Z@3oool3000002X0oooo00380?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0Y0?ooo`00703oool00`00
0000oooo0?ooo`2h0?ooo`800000^@3oool00`000000oooo0?ooo`0K0?ooo`004`3ooooo000008X0
0000503oool001T0oooo00@000000?ooo`3oool00000203oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool2000000X0oooo00<000000?ooo`3o
ool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00H0oooo00@000000?ooo`3o
ool000006P3oool002D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0b0?ooo`03
0000003oool0oooo03T0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo02P0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0h0?ooo`03
0000003oool0oooo02@0oooo000Z0?ooo`<00000ZP3oool00`000000oooo0?ooo`0<0?ooo`<00000
a`3oool002/0oooo00<000000?ooo`3oool0Z@3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo0<H0oooo003G0?ooo`030000003oool0oooo0=H0oooo000l0?ooo`030000003oool0oooo
09P0oooo0P00000N0?ooo`030000003oool0oooo0;H0oooo000k0?ooo`<00000V@3oool00`000000
oooo0?ooo`0L0?ooo`<00000]`3oool003`0oooo00<000000?ooo`3oool01@3oool00`000000oooo
0?ooo`2@0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool01@3oool00`000000oooo
0?ooo`2^0?ooo`00@`3oool300000940oooo00<000000?ooo`3oool0903oool300000:l0oooo0014
0?ooo`030000003oool0oooo0900oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`2^
0?ooo`00B`3oool00`000000oooo0?ooo`290?ooo`030000003oool0oooo02`0oooo00<000000?oo
o`3oool0Y`3oool004X0oooo0`00002:0?ooo`800000;03oool300000:P0oooo001;0?ooo`030000
003oool0oooo08T0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`2W0?ooo`00D`3o
ool00`000000oooo0?ooo`210?ooo`030000003oool0oooo0=H0oooo001B0?ooo`<00000PP3oool0
0`000000oooo0?ooo`0d0?ooo`030000003oool0oooo09l0oooo001C0?ooo`030000003oool0oooo
0840oooo00<000000?ooo`3oool0<`3oool300000:00oooo003G0?ooo`030000003oool0oooo03@0
oooo00<000000?ooo`3oool0W`3oool00=L0oooo0P00003G0?ooo`00e`3oool00`000000oooo0?oo
o`3F0?ooo`00FP3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo03/0oooo00<00000
0?ooo`3oool0V03oool005T0oooo0`00001k0?ooo`030000003oool0oooo03X0oooo0`00002I0?oo
o`00FP3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3o
ool0V03oool005h0oooo00<000000?ooo`3oool0MP3oool00`000000oooo0?ooo`0o0?ooo`030000
003oool0oooo09@0oooo001M0?ooo`<00000M`3oool2000003l0oooo0`00002E0?ooo`00GP3oool0
0`000000oooo0?ooo`1H0?ooo`030;md00000000000000@0000000T0B9cO0?ooo`3oool0okmd0000
001d_ol0oooo0?oog`2LB0000`0000000`1d_ol0oooo0?ooo`080?ooo`030000003oool0oooo03l0
oooo00<000000?ooo`3oool0U03oool00600oooo00<000000?ooo`3oool0E`3oool00`3OW4P00000
000000030000000:0018W03Oool0oooo0?noM0000000M;oo0?ooo`3ooml0W4Qd0;ooo`80oooo00<0
_g@007Boo`3oool0203oool00`000000oooo0?ooo`3F0?ooo`00G`3oool3000005T0oooo00<0gia8
04RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`090?ooo`030?oog`2LB7@0_ooo00P0oooo00<00000
0?ooo`3oool0eP3oool00600oooo00<000000?ooo`3oool0F@3oool00`3OW4P0M;oo0?ooo`0>0?oo
o`030;md001d_ol0oooo00L0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`2@0?oo
o`00H`3oool00`000000oooo0?ooo`1G0?ooo`030;mdB02Lgol0oooo00`0oooo00<0oooO09a80018
W=l0203oool00`000000oooo0?ooo`120?ooo`<00000T@3oool00680oooo0`00001H0?ooo`030?oo
g`2LB7@0_ooo00`0oooo00<0gia80000001d_ol0203oool3000004<0oooo00<000000?ooo`3oool0
T03oool006<0oooo00<000000?ooo`3oool0F03oool00`3o_g@007Bo0?ooo`070?ooo`070=oOg`3o
ool0oono07@000000000001d0;ooo`080?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3o
ool0SP3oool006D0oooo00<000000?ooo`3oool0EP3oool00`3ookl0M01809cOo`070?ooo`060=nL
B000000000000000000004P0W=oo2@3oool00`000000oooo0?ooo`140?ooo`<00000S`3oool006@0
oooo0`00001G0?ooo`030?oo_`1d0000B9cO00L0oooo00@0oono07@000000000B9cO2`3oool00`00
0000oooo0?ooo`150?ooo`030000003oool0oooo08h0oooo001U0?ooo`030000003oool0oooo0540
oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`080?ooo`030;md001d_ol0oooo00/0
oooo00<000000?ooo`3oool0eP3oool00;T0oooo00P0oooO09a8B02Lgol0oooo0?oo_`1d0000001d
0;ooo`P0oooo00H0omnL04P00000000000000000B02Lgol80?ooo`030000003oool0oooo04L0oooo
00<000000?ooo`3oool0S03oool006L0oooo00<000000?ooo`3oool0D03oool01P3ooml0W4P00000
00000000001d0;ooo`X0oooo00D0_g@0000000000000000004RLg`080?ooo`800000A`3oool30000
08d0oooo001V0?ooo`<00000KP3oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo08`0
oooo001W0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<0
00000?ooo`3oool0eP3oool006T0oooo00<000000?ooo`3oool0J`3oool00`000000oooo0?ooo`19
0?ooo`030000003oool0oooo08X0oooo001X0?ooo`<00000K03oool00`000000oooo0?ooo`180?oo
o`<00000R`3oool006T0oooo00<000000?ooo`3oool0J`3oool2000004X0oooo00<000000?ooo`3o
ool0RP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3o
ool00=L0oooo00<000000?ooo`3oool0eP3oool006/0oooo00<000000?ooo`3oool0J@3oool00`00
0000oooo0?ooo`1;0?ooo`030000003oool0oooo08P0oooo001Z0?ooo`<00000JP3oool00`000000
oooo0?ooo`1:0?ooo`<00000R@3oool006/0oooo00<000000?ooo`3oool0J@3oool2000004`0oooo
00<000000?ooo`3oool0R03oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<00000
0?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3o
ool0eP3oool006d0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`3F0?ooo`00K03o
ool3000006P0oooo0P00001>0?ooo`030000003oool0oooo08H0oooo001]0?ooo`030000003oool0
oooo05H0oooo00<0oooO09a8000000000P0000000`1d_ol0oooo0?ooo`090?ooo`030000003oool0
oooo04`0oooo0`0000270?ooo`00aP3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo
00T0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`260?ooo`00b`3oool00`3ooml0
W4Qd0;ooo`090?ooo`030000003oool0oooo0=H0oooo003<0?ooo`030;md001d_ol0oooo00P0oooo
00<000000?ooo`3oool0eP3oool00</0oooo00<0oooO09a80018W=l02@3oool00`000000oooo0?oo
o`3F0?ooo`00b`3oool00`3OW4P0000007Boo`090?ooo`<00000eP3oool00<L0oooo00L0gmoO0?oo
o`3ookl0M0000000000007@0_ooo00T0oooo00<000000?ooo`3oool0eP3oool00<L0oooo00H0gia8
00000000000000000000B02Lgol:0?ooo`030000003oool0oooo0=H0oooo001_0?ooo`030000003o
ool0oooo05D0oooo00@0oono07@000000000B9cO303oool00`000000oooo0?ooo`3F0?ooo`00KP3o
ool3000005L0oooo00<0_g@007Boo`3oool0303oool00`000000oooo0?ooo`1?0?ooo`030000003o
ool0oooo08@0oooo001_0?ooo`030000003oool0oooo05H0oooo00H0omnL04P00000000000000000
B02Lgol90?ooo`030000003oool0oooo04h0oooo0`0000250?ooo`00b@3oool01@2oM00000000000
00000000B9cO00T0oooo0P00001@0?ooo`030000003oool0oooo08@0oooo003G0?ooo`030000003o
ool0oooo0=H0oooo003G0?ooo`030000003oool0oooo0=H0oooo001_0?ooo`030000003oool0oooo
06D0oooo00<000000?ooo`3oool0eP3oool006h0oooo0`00001V0?ooo`030000003oool0oooo0=H0
oooo001_0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3oool0D@3oool00`000000oooo
0?ooo`220?ooo`00e`3oool200000540oooo0`0000230?ooo`00e`3oool00`000000oooo0?ooo`1A
0?ooo`030000003oool0oooo0880oooo003G0?ooo`030000003oool0oooo0=H0oooo003G0?ooo`03
0000003oool0oooo0=H0oooo003G0?ooo`030000003oool0oooo0=H0oooo001`0?ooo`030000003o
ool0oooo06@0oooo00<000000?ooo`3oool0eP3oool006l0oooo0`00001U0?ooo`800000e`3oool0
0700oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000
oooo0?ooo`1A0?ooo`030000003oool0oooo0880oooo003G0?ooo`030000003oool0oooo0500oooo
0`0000230?ooo`00e`3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo0880oooo003G
0?ooo`030000003oool0oooo0=H0oooo003G0?ooo`800000e`3oool00;`0oooo00<0gia804RLg`3o
ool0103oool01P3o_g@0000007Boo`3oool0oooO09a800<0000000<0M;oo0?ooo`3oool0203oool0
0`000000oooo0?ooo`3F0?ooo`00L@3oool00`000000oooo0?ooo`180?ooo`030?oOW018B9`0gooo
00@0oooo00L0okmd0000001d_ol0oooo0?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool0203o
ool00`000000oooo0?ooo`3F0?ooo`00L03oool3000004T0oooo00<0oooO09a8B02Lgol03@3oool0
0`3ooml0W4Qd0;ooo`080?ooo`030000003oool0oooo0=H0oooo001a0?ooo`030000003oool0oooo
04T0oooo00<0gia804RLg`3oool03@3oool00`2oM000M;oo0?ooo`070?ooo`030000003oool0oooo
0=H0oooo002m0?ooo`030?oOW018B9`0gooo00`0oooo00<0oooO09a80018W=l0203oool00`000000
oooo0?ooo`3F0?ooo`00_@3oool00`3ooml0W4Q809cOo`0<0?ooo`030=nLB0000000M;oo00P0oooo
0`00003F0?ooo`00_P3oool00`3OW4P007Bo0?ooo`070?ooo`070=oOg`3oool0oono07@000000000
001d0;ooo`080?ooo`030000003oool0oooo0=H0oooo002n0?ooo`030?oo_`1d07@0_ooo00L0oooo
00H0gia800000000000000000000B02Lgol90?ooo`030000003oool0oooo05<0oooo00<000000?oo
o`3oool0P03oool00;T0oooo00<0omoO0=ooo`3oool00`3oool00`2oM000M;oo0?ooo`060?ooo`04
0?oo_`1d0000000004RLg`/0oooo00<000000?ooo`3oool0DP3oool300000840oooo002i0?ooo`03
0?oo_`1dB9`0gooo00<0oooo00<0okmd001d_`3oool01`3oool00`2oM000M;oo0?ooo`0;0?ooo`03
0000003oool0oooo05<0oooo00<000000?ooo`3oool0P03oool00;X0oooo00<0_g@0000000000000
0`0000000`0007@0_ooo0?ooo`060?ooo`060?oOW018000000000000000004P0W=oo203oool00`00
0000oooo0?ooo`3F0?ooo`00LP3oool00`000000oooo0?ooo`150?ooo`030?noM0000000000000@0
000000<0M;oo0?ooo`3oool01P3oool01@2oM0000000000000000000B9cO00P0oooo0P00003G0?oo
o`00L@3oool3000006<0oooo00<000000?ooo`3oool0eP3oool00780oooo00<000000?ooo`3oool0
HP3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool0
0`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000
oooo0?ooo`3F0?ooo`00e`3oool200000=L0oooo003G0?ooo`030000003oool0oooo0=H0oooo003G
0?ooo`030000003oool0oooo0=H0oooo003G0?ooo`030000003oool0oooo0=H0oooo003G0?ooo`03
0000003oool0oooo0=H0oooo003G0?ooo`030000003oool0oooo0=H0oooo003G0?ooo`800000E@3o
ool00`000000oooo0?ooo`1o0?ooo`00e`3oool00`000000oooo0?ooo`1C0?ooo`<00000P03oool0
0=L0oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`1o0?ooo`00L`3oool00`000000
oooo0?ooo`1Q0?ooo`030000003oool0oooo0=H0oooo001b0?ooo`<00000HP3oool00`000000oooo
0?ooo`3F0?ooo`00L`3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo0=H0oooo003G
0?ooo`800000e`3oool00;l0oooo00H0_g@000000000000000000018W03Oool30?ooo`050?oOW018
000000000000M02oool02P3oool00`000000oooo0?ooo`3F0?ooo`00`03oool00`3ogi`0B00007Bo
o`040?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`030000003oool0oooo0=H0
oooo00300?ooo`030?oOW0180000M;oo00@0oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo
0?ooo`080?ooo`030000003oool0oooo0=H0oooo00300?ooo`030?oOW0180000M;oo00<0oooo00<0
oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`080?ooo`030000003oool0oooo0=H0oooo0030
0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`08
0?ooo`030000003oool0oooo0=H0oooo00300?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`080?ooo`<00000eP3oool00<00oooo00<0omnL04P0001d
_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo0>40oooo00300?ooo`030?oO
W0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`3Q0?ooo`00`03o
ool00`3ogi`0B00007Boo`030?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0h@3o
ool00<00oooo00<0omnL04P0001d_ol0103oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0
oooo0>40oooo002n0?ooo`050?oog`2LB00000000000001d_ol0103oool01`3ookl0M00007Boo`3o
ool0okmd0000M02oool0hP3oool00<00oooo00<0okmd0000001d_ol01@3oool01@3ookl0M0000000
000007@0_ooo0><0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40
oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-3.62576, -10.9711, \
0.0252373, 0.123792}}]
}, Closed]],

Cell["\<\
Clearly the points are spaced more densely where the curvature is large.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Aspect ratio", "Subsubsection"],

Cell["\<\
The appearance of a plotted curves depends upon the length units employed for \
the axes.  Although these units often do not affect one's appreciation of the \
curve, it is sometimes necessary to control the aspect ratio between \
horizontal and vertical length units.  Consider, for example, a unit circle \
expressed as two branches of a square-root function.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[{\@\(1 - x\^2\), \(-\@\(1 - x\^2\)\)}, {x, \(-1\), 
          1}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.309017 0.294302 [
[.02381 .29652 -9.15625 -15.6875 ]
[.02381 .29652 9.15625 0 ]
[.2619 .29652 -13.6563 -15.6875 ]
[.2619 .29652 13.6563 0 ]
[.7381 .29652 -9.5 -15.6875 ]
[.7381 .29652 9.5 0 ]
[.97619 .29652 -5 -15.6875 ]
[.97619 .29652 5 0 ]
[.4875 .01472 -18.3125 -7.84375 ]
[.4875 .01472 0 7.84375 ]
[.4875 .16187 -27.3125 -7.84375 ]
[.4875 .16187 0 7.84375 ]
[.4875 .45617 -19 -7.84375 ]
[.4875 .45617 0 7.84375 ]
[.4875 .60332 -10 -7.84375 ]
[.4875 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 .30902 m
.02381 .31527 L
s
gsave
.02381 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 .30902 m
.2619 .31527 L
s
gsave
.2619 .29652 -74.6563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.7381 .30902 m
.7381 .31527 L
s
gsave
.7381 .29652 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 .30902 m
.97619 .31527 L
s
gsave
.97619 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .30902 m
.07143 .31277 L
s
.11905 .30902 m
.11905 .31277 L
s
.16667 .30902 m
.16667 .31277 L
s
.21429 .30902 m
.21429 .31277 L
s
.30952 .30902 m
.30952 .31277 L
s
.35714 .30902 m
.35714 .31277 L
s
.40476 .30902 m
.40476 .31277 L
s
.45238 .30902 m
.45238 .31277 L
s
.54762 .30902 m
.54762 .31277 L
s
.59524 .30902 m
.59524 .31277 L
s
.64286 .30902 m
.64286 .31277 L
s
.69048 .30902 m
.69048 .31277 L
s
.78571 .30902 m
.78571 .31277 L
s
.83333 .30902 m
.83333 .31277 L
s
.88095 .30902 m
.88095 .31277 L
s
.92857 .30902 m
.92857 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5 .01472 m
.50625 .01472 L
s
gsave
.4875 .01472 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .16187 m
.50625 .16187 L
s
gsave
.4875 .16187 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .45617 m
.50625 .45617 L
s
gsave
.4875 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .60332 m
.50625 .60332 L
s
gsave
.4875 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .04415 m
.50375 .04415 L
s
.5 .07358 m
.50375 .07358 L
s
.5 .10301 m
.50375 .10301 L
s
.5 .13244 m
.50375 .13244 L
s
.5 .1913 m
.50375 .1913 L
s
.5 .22073 m
.50375 .22073 L
s
.5 .25016 m
.50375 .25016 L
s
.5 .27959 m
.50375 .27959 L
s
.5 .33845 m
.50375 .33845 L
s
.5 .36788 m
.50375 .36788 L
s
.5 .39731 m
.50375 .39731 L
s
.5 .42674 m
.50375 .42674 L
s
.5 .4856 m
.50375 .4856 L
s
.5 .51503 m
.50375 .51503 L
s
.5 .54446 m
.50375 .54446 L
s
.5 .57389 m
.50375 .57389 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .30914 m
.02499 .32968 L
.02605 .33756 L
.02846 .35003 L
.03279 .36589 L
.03784 .37993 L
.04262 .39091 L
.06244 .42514 L
.08265 .45073 L
.10458 .473 L
.14545 .50548 L
.1848 .52962 L
.22263 .54824 L
.26292 .56425 L
.30169 .57658 L
.34292 .58685 L
.36351 .59097 L
.38262 .59424 L
.40229 .59706 L
.42082 .59922 L
.44134 .60108 L
.45995 .60228 L
.46971 .60272 L
.47505 .60291 L
.48003 .60306 L
.48428 .60316 L
.48883 .60324 L
.49143 .60327 L
.49271 .60328 L
.49389 .60329 L
.49494 .6033 L
.4961 .60331 L
.49726 .60331 L
.4985 .60332 L
.49971 .60332 L
.50085 .60332 L
.50209 .60332 L
.50278 .60331 L
.50341 .60331 L
.50469 .6033 L
.50608 .60329 L
.50859 .60327 L
.51093 .60324 L
.51341 .6032 L
.51785 .60311 L
.52298 .60298 L
.52778 .60282 L
.5387 .60235 L
.54911 .60175 L
.55899 .60105 L
Mistroke
.57753 .59939 L
.59737 .5971 L
.61577 .59449 L
.65495 .5873 L
.69659 .57707 L
.73671 .56438 L
.77531 .54915 L
.81636 .52898 L
.8559 .50454 L
.89393 .47437 L
.91511 .45322 L
.9344 .42957 L
.944 .41539 L
.95442 .39698 L
.9592 .38693 L
.96427 .37446 L
.9686 .36137 L
.97089 .3528 L
.97336 .34103 L
.97408 .33667 L
.97484 .33117 L
.97549 .32503 L
.97619 .30914 L
Mfstroke
.02381 .3089 m
.02499 .28835 L
.02605 .28047 L
.02846 .268 L
.03279 .25214 L
.03784 .2381 L
.04262 .22712 L
.06244 .19289 L
.08265 .16731 L
.10458 .14503 L
.14545 .11255 L
.1848 .08842 L
.22263 .06979 L
.26292 .05378 L
.30169 .04145 L
.34292 .03119 L
.36351 .02706 L
.38262 .0238 L
.40229 .02098 L
.42082 .01881 L
.44134 .01696 L
.45995 .01576 L
.46971 .01531 L
.47505 .01512 L
.48003 .01497 L
.48428 .01488 L
.48883 .0148 L
.49143 .01476 L
.49271 .01475 L
.49389 .01474 L
.49494 .01473 L
.4961 .01472 L
.49726 .01472 L
.4985 .01472 L
.49971 .01472 L
.50085 .01472 L
.50209 .01472 L
.50278 .01472 L
.50341 .01472 L
.50469 .01473 L
.50608 .01474 L
.50859 .01476 L
.51093 .01479 L
.51341 .01483 L
.51785 .01492 L
.52298 .01506 L
.52778 .01522 L
.5387 .01569 L
.54911 .01628 L
.55899 .01698 L
Mistroke
.57753 .01864 L
.59737 .02093 L
.61577 .02355 L
.65495 .03073 L
.69659 .04096 L
.73671 .05365 L
.77531 .06889 L
.81636 .08905 L
.8559 .11349 L
.89393 .14367 L
.91511 .16482 L
.9344 .18846 L
.944 .20265 L
.95442 .22106 L
.9592 .23111 L
.96427 .24358 L
.9686 .25666 L
.97089 .26524 L
.97336 .277 L
.97408 .28136 L
.97484 .28687 L
.97549 .29301 L
.97619 .3089 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00f@3oool0
0`000000oooo0?ooo`3D0?ooo`00bP3oool01P2oM000000000000000000004RL0=ooo`T0oooo00<0
00000?ooo`3oool0e03oool00</0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`3D
0?ooo`00b`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0=@0oooo003;0?ooo`03
0?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0e03oool00;/0oooo00<0oooO09a800000000
1`0000000`18W=l0oooo0?ooo`030?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0
e03oool00</0oooo00<0omnL04P0001d_ol00`3oool@00000<l0oooo002m0?oooa@00000203oool0
0`000000oooo0?ooo`050?oooa@00000^`3oool00;D0oooo2000000>0?ooo`030?oOW0180000M;oo
00/0oooo00<000000?ooo`3oool06@3oool800000;<0oooo002^0?ooo`L000005P3oool00`3ogi`0
B00007Boo`0;0?ooo`030000003oool0oooo0240oooo2000002[0?ooo`00YP3oool8000001d0oooo
00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`0Y0?ooo`L00000Y03oool009l0oooo1`00
000S0?ooo`050?oog`2LB00000000000001d_ol02`3oool00`000000oooo0?ooo`0`0?ooo`D00000
W`3oool009/0oooo1000000/0?ooo`030?noM0000000M;oo00/0oooo00<000000?ooo`3oool0=@3o
ool6000009T0oooo002F0?ooo`D00000?P3oool00`000000oooo0?ooo`0k0?ooo`D00000U03oool0
0980oooo100000130?ooo`030000003oool0oooo0400oooo1000002@0?ooo`00SP3oool4000004L0
oooo00<000000?ooo`3oool0A03oool4000008`0oooo002:0?ooo`@00000B`3oool00`000000oooo
0?ooo`180?ooo`@00000R03oool008H0oooo1000001?0?ooo`030000003oool0oooo04`0oooo1000
00240?ooo`00P`3oool3000005<0oooo00<000000?ooo`3oool0D03oool300000840oooo00200?oo
o`<00000EP3oool3000005<0oooo0`00001n0?ooo`00O03oool4000005T0oooo00<000000?ooo`3o
ool0EP3oool4000007X0oooo001i0?ooo`<00000G@3oool00`000000oooo0?ooo`1J0?ooo`<00000
M`3oool007H0oooo0`00001P0?ooo`030000003oool0oooo05d0oooo0`00001d0?ooo`00M03oool2
000006<0oooo00<000000?ooo`3oool0H03oool200000780oooo001b0?ooo`800000I@3oool00`00
0000oooo0?ooo`1R0?ooo`800000L03oool00700oooo0P00001W0?ooo`030000003oool0oooo06@0
oooo0P00001^0?ooo`00K@3oool3000006T0oooo00<000000?ooo`3oool0IP3oool3000006/0oooo
001[0?ooo`800000K03oool00`000000oooo0?ooo`1Y0?ooo`800000J@3oool006T0oooo0P00001^
0?ooo`030000003oool0oooo06/0oooo0P00001W0?ooo`00I`3oool200000700oooo00<000000?oo
o`3oool0K@3oool3000006@0oooo001U0?ooo`800000LP3oool00`000000oooo0?ooo`1`0?ooo`80
0000HP3oool006<0oooo0P00001d0?ooo`<00000LP3oool200000600oooo001Q0?ooo`800000MP3o
ool00`000000oooo0?ooo`1d0?ooo`800000GP3oool005l0oooo0P00001h0?ooo`030000003oool0
oooo07H0oooo0P00001L0?ooo`00G@3oool2000007X0oooo00<000000?ooo`3oool0N03oool20000
05X0oooo001K0?ooo`800000O03oool00`000000oooo0?ooo`1j0?ooo`800000F03oool005T0oooo
0P00001n0?ooo`030000003oool0oooo07`0oooo0P00001F0?ooo`00E`3oool200000800oooo00<0
00000?ooo`3oool0OP3oool2000005@0oooo001E0?ooo`800000PP3oool00`000000oooo0?ooo`20
0?ooo`800000DP3oool005@0oooo00<000000?ooo`3oool0PP3oool00`000000oooo0?ooo`220?oo
o`800000D03oool00580oooo0P0000250?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3o
ool0C@3oool00500oooo0P0000270?ooo`030000003oool0oooo08D0oooo0P00001=0?ooo`00C`3o
ool00`000000oooo0?ooo`270?ooo`030000003oool0oooo08L0oooo0P00001;0?ooo`00C@3oool2
000008X0oooo0`0000290?ooo`800000B@3oool004`0oooo00<000000?ooo`3oool0RP3oool00`00
0000oooo0?ooo`2;0?ooo`030000003oool0oooo04H0oooo001:0?ooo`800000S@3oool00`000000
oooo0?ooo`2<0?ooo`800000AP3oool004P0oooo0P00002?0?ooo`030000003oool0oooo08h0oooo
0P0000140?ooo`00A`3oool00`000000oooo0?ooo`2?0?ooo`030000003oool0oooo0900oooo00<0
00000?ooo`3oool0@@3oool004H0oooo00<000000?ooo`3oool0T03oool00`000000oooo0?ooo`2A
0?ooo`030000003oool0oooo0400oooo00140?ooo`800000T`3oool00`000000oooo0?ooo`2B0?oo
o`800000@03oool004<0oooo00<000000?ooo`3oool0T`3oool00`000000oooo0?ooo`2D0?ooo`03
0000003oool0oooo03d0oooo00120?ooo`030000003oool0oooo09@0oooo00<000000?ooo`3oool0
U@3oool00`000000oooo0?ooo`0l0?ooo`00@03oool2000009L0oooo00<000000?ooo`3oool0UP3o
ool00`000000oooo0?ooo`0k0?ooo`00?`3oool00`000000oooo0?ooo`2G0?ooo`030000003oool0
oooo09L0oooo0P00000k0?ooo`00?P3oool00`000000oooo0?ooo`2H0?ooo`030000003oool0oooo
09T0oooo00<000000?ooo`3oool0>03oool003`0oooo0P00002K0?ooo`<00000VP3oool00`000000
oooo0?ooo`0g0?ooo`00>`3oool00`000000oooo0?ooo`2K0?ooo`030000003oool0oooo09/0oooo
00<000000?ooo`3oool0=P3oool003X0oooo00<000000?ooo`3oool0W03oool00`000000oooo0?oo
o`2L0?ooo`800000=P3oool003P0oooo0P00002O0?ooo`030000003oool0oooo09h0oooo00<00000
0?ooo`3oool0<`3oool003L0oooo00<000000?ooo`3oool0W`3oool00`000000oooo0?ooo`2O0?oo
o`030000003oool0oooo0380oooo000f0?ooo`030000003oool0oooo0:00oooo00<000000?ooo`3o
ool0X03oool00`000000oooo0?ooo`0a0?ooo`00=@3oool00`000000oooo0?ooo`2Q0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool0<03oool003@0oooo00<000000?ooo`3oool0Q@3o
ool01@3ogi`0B0000000000007@0_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oog`2LB0030000
000307Boo`3oool0oooo00T0oooo00<000000?ooo`3oool0XP3oool00`000000oooo0?ooo`0_0?oo
o`00<`3oool00`000000oooo0?ooo`250?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`02
0?ooo`070?noM0000000M;oo0?ooo`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00T0oooo
00<000000?ooo`3oool0X`3oool00`000000oooo0?ooo`0^0?ooo`00<P3oool00`000000oooo0?oo
o`260?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd0;oo
o`090?ooo`030000003oool0oooo0:@0oooo00<000000?ooo`3oool0;@3oool00340oooo00<00000
0?ooo`3oool0QP3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00/0oooo00<0_g@0
07Boo`3oool0203oool00`000000oooo0?ooo`2U0?ooo`030000003oool0oooo02`0oooo000`0?oo
o`030000003oool0oooo07/0oooo00<0oooO09a8000000001`0000001@18W=l0oooo0?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3ooml0W4P004RLg`090?ooo`030000003oool0
oooo0:H0oooo00<000000?ooo`3oool0:`3oool002l0oooo00<000000?ooo`3oool0R03oool00`3o
gi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0gia80000001d_ol02@3oool40000
0:H0oooo00<000000?ooo`3oool0:P3oool002h0oooo00<000000?ooo`3oool0R@3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0
_ooo00T0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0Y0?ooo`00;@3oool00`00
0000oooo0?ooo`2:0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3O
W4P00000000000000000001809cOo`X0oooo00<000000?ooo`3oool0Z@3oool00`000000oooo0?oo
o`0X0?ooo`00;03oool00`000000oooo0?ooo`2;0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia8
0000M02oool01P3oool0103ookl0M00000000018W=l<0?ooo`030000003oool0oooo0:X0oooo00<0
00000?ooo`3oool09`3oool002`0oooo00<000000?ooo`3oool0S03oool00`3OW4P004RL0=ooo`02
0?ooo`030;md0000M;l0oooo00L0oooo00<0_g@007Boo`3oool0303oool00`000000oooo0?ooo`2Z
0?ooo`030000003oool0oooo02L0oooo000[0?ooo`030000003oool0oooo08d0oooo00L0oono07@0
001d_ol0oooo0?noM00007@0_ooo00P0oooo00H0omnL04P00000000000000000B02Lgol90?ooo`03
0000003oool0oooo0:/0oooo00<000000?ooo`3oool09P3oool002X0oooo00<000000?ooo`3oool0
S`3oool01@3ookl0M0000000000007@0_ooo00X0oooo00D0_g@0000000000000000004RLg`090?oo
o`030000003oool0oooo0:`0oooo00<000000?ooo`3oool09@3oool002T0oooo00<000000?ooo`3o
ool0[@3oool00`000000oooo0?ooo`2]0?ooo`030000003oool0oooo02@0oooo000Y0?ooo`030000
003oool0oooo0:d0oooo00<000000?ooo`3oool0[P3oool00`000000oooo0?ooo`0S0?ooo`00:03o
ool00`000000oooo0?ooo`2^0?ooo`030000003oool0oooo0:h0oooo00<000000?ooo`3oool08`3o
ool002L0oooo00<000000?ooo`3oool0[`3oool00`000000oooo0?ooo`2_0?ooo`030000003oool0
oooo0280oooo000V0?ooo`030000003oool0oooo0;00oooo00<000000?ooo`3oool0/03oool00`00
0000oooo0?ooo`0Q0?ooo`009P3oool00`000000oooo0?ooo`2`0?ooo`<00000/@3oool00`000000
oooo0?ooo`0P0?ooo`009@3oool00`000000oooo0?ooo`2a0?ooo`030000003oool0oooo0;40oooo
00<000000?ooo`3oool0803oool002@0oooo00<000000?ooo`3oool0/P3oool00`000000oooo0?oo
o`2b0?ooo`030000003oool0oooo01l0oooo000T0?ooo`030000003oool0oooo0;80oooo00<00000
0?ooo`3oool0/`3oool00`000000oooo0?ooo`0N0?ooo`008`3oool00`000000oooo0?ooo`2c0?oo
o`030000003oool0oooo0;<0oooo00<000000?ooo`3oool07P3oool002<0oooo00<000000?ooo`3o
ool0/`3oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo01d0oooo000R0?ooo`030000
003oool0oooo0;@0oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`0L0?ooo`008P3o
ool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3oool0703o
ool00240oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`2f0?ooo`030000003oool0
oooo01/0oooo000P0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0]P3oool00`00
0000oooo0?ooo`0K0?ooo`00803oool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo0;L0
oooo00<000000?ooo`3oool06P3oool001l0oooo00<000000?ooo`3oool0]`3oool00`000000oooo
0?ooo`2g0?ooo`030000003oool0oooo01X0oooo000O0?ooo`030000003oool0oooo0;L0oooo0`00
002h0?ooo`030000003oool0oooo01T0oooo000N0?ooo`030000003oool0oooo0;P0oooo00<00000
0?ooo`3oool0^03oool00`000000oooo0?ooo`0I0?ooo`007P3oool00`000000oooo0?ooo`2h0?oo
o`030000003oool0oooo0;T0oooo00<000000?ooo`3oool0603oool001d0oooo00<000000?ooo`3o
ool0^@3oool00`000000oooo0?ooo`2i0?ooo`030000003oool0oooo01P0oooo000L0?ooo`030000
003oool0oooo0;X0oooo00<000000?ooo`3oool0^P3oool00`000000oooo0?ooo`0G0?ooo`00703o
ool00`000000oooo0?ooo`2j0?ooo`030000003oool0oooo0;X0oooo00<000000?ooo`3oool05`3o
ool001/0oooo00<000000?ooo`3oool0^`3oool00`000000oooo0?ooo`2k0?ooo`030000003oool0
oooo01H0oooo000K0?ooo`030000003oool0oooo0;/0oooo00<000000?ooo`3oool0^`3oool00`00
0000oooo0?ooo`0F0?ooo`006`3oool00`000000oooo0?ooo`2k0?ooo`030000003oool0oooo0;/0
oooo00<000000?ooo`3oool05P3oool001X0oooo00<000000?ooo`3oool0_03oool00`000000oooo
0?ooo`2l0?ooo`030000003oool0oooo01D0oooo000J0?ooo`030000003oool0oooo0;`0oooo00<0
00000?ooo`3oool0_03oool00`000000oooo0?ooo`0E0?ooo`006P3oool00`000000oooo0?ooo`2l
0?ooo`030000003oool0oooo0;`0oooo00<000000?ooo`3oool05@3oool001T0oooo00<000000?oo
o`3oool0_@3oool300000;d0oooo00<000000?ooo`3oool0503oool001T0oooo00<000000?ooo`3o
ool0_@3oool00`000000oooo0?ooo`2m0?ooo`030000003oool0oooo01@0oooo000I0?ooo`030000
003oool0oooo0;d0oooo00<000000?ooo`3oool0_P3oool00`000000oooo0?ooo`0C0?ooo`00603o
ool00`000000_g@000000002000000030018W03Oool0oooo05@0oooo00D0omnL04P000000000001d
0;ooo`030?ooo`060?noM0000000M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`1A0?oo
o`030000003oool0oooo05H0oooo00D0omnL04P000000000001d0;ooo`030?ooo`060?noM0000000
M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`1D0?ooo`030000002oM000000000800000
00<004RL0=ooo`3oool03P3oool001P0oooo00D000000?ooo`3ogi`0B00007Boo`1F0?ooo`070?oo
_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo0?ooo`3ooml0W4Qd0;oo
o`020?ooo`030;md001d_ol0oooo0540oooo00<000000?ooo`3oool0E@3oool01`3ookl0M00007Bo
o`3oool0okmd0000M02oool00P3oool01`3o_g@0000007Boo`3oool0oooO09a8M02oool00P3oool0
0`2oM000M;oo0?ooo`1D0?ooo`050000003oool0omnL04P0001d_ol04@3oool001P0oooo00D00000
0?ooo`3ogi`0B00007Boo`1F0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3o
ool00`3ooml0W4Qd0;ooo`1A0?ooo`030000003oool0oooo05D0oooo00<0_g@00018W03Oool00P3o
ool00`2oM00007Bo0?ooo`0:0?ooo`030?oog`2LB7@0_ooo05@0oooo00D000000?ooo`3ogi`0B000
07Boo`0A0?ooo`00603oool01@000000oooo0?oOW0180000M;oo05D0oooo00<0oono07@00018W=l0
0`3oool00`3OW4P0001d0;ooo`0;0?ooo`030;md001d_ol0oooo0500oooo00<000000?ooo`3oool0
E03oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00/0oooo00<0_g@007Boo`3oool0
E03oool010000000omnL04P0001d_olA0?ooo`002P3oool00`3ooml0W4P0000000070000000404RL
g`3oool0oooo00000080oooo00<0omnL04P0001d_ol0B03oool00`3ooml0W4P00000000700000006
04RLg`3oool0oooo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0:0?ooo`030?oog`2LB000
B9cO0540oooo00<000000?ooo`3oool0E03oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo00X0oooo00<0oooO09a80018W=l0E@3oool010000000omnL04P0001d_olA0?ooo`005`3oool0
1P000000oooo0?ooo`3ogi`0B00007BooeD0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`0:0?ooo`030=nLB0000000M;oo0540oooo00<000000?ooo`3oool0E03oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0gia80000001d_ol0E@3oool010000000omnL
04P0001d_olA0?ooo`005`3oool01P000000oooo0?ooo`3ogi`0B00007BooeD0oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`070=oOg`3oool0oono07@000000000001d0;oo
o`1A0?ooo`030000003oool0oooo05@0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`060?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`1E0?ooo`040000003ogi`0B00007Bo
oa40oooo000G0?ooo`060000003oool0oooo0?oOW0180000M;ooE@3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo00H0oooo00H0gia800000000000000000000B02LgomB0?ooo`030000
003oool0oooo05@0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`060=nL
B000000000000000000004P0W=ooE`3oool00`000000B00007Boo`0A0?ooo`005P3oool01`000000
oooo0?ooo`3oool0omnL04P0001d_ol0E@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0
_ooo00H0oooo00@0oono07@000000000B9cOE03oool00`000000oooo0?ooo`1D0?ooo`030?oo_`1d
0000B9cO00<0oooo00<0gia80000M02oool01P3oool0103ookl0M00000000018W=mI0?ooo`030000
00180000M;oo0140oooo000F0?ooo`070000003oool0oooo0?ooo`3ogi`0B00007Boo`1F0?ooo`03
0=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`2oM000M;oo0?ooo`1D0?ooo`<0
0000E@3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0_g@007Boo`3o
ool0F@3oool00`000000B00007Boo`0A0?ooo`005P3oool010000000oooo0?oog`2LB00200000003
07Boo`3oool0oooo05@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00H0omnL
04P00000000000000000B02LgomA0?ooo`030000003oool0oooo05D0oooo00L0oono07@0001d_ol0
oooo0?noM00007@0_ooo00P0oooo00H0omnL04P00000000000000000B02LgomD0?ooo`050?oog`2L
B00000000000001d_ol04@3oool001H0oooo00L000000?ooo`3oool0oooo0?noM0000000M;oo05L0
oooo00D0oono07@000000000001d0;ooo`0:0?ooo`050;md00000000000000000018W=l0D@3oool0
0`000000oooo0?ooo`1F0?ooo`050?oo_`1d000000000000M02oool02P3oool01@2oM00000000000
00000000B9cO05H0oooo0P0000000`1d_ol0oooo0?ooo`0?0?ooo`005@3oool00`000000oooo0?oo
o`310?ooo`030000003oool0oooo0<00oooo00<000000?ooo`3oool04@3oool001D0oooo00<00000
0?ooo`3oool0`@3oool00`000000oooo0?ooo`300?ooo`030000003oool0oooo0140oooo000E0?oo
o`030000003oool0oooo0<40oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`0@0?oo
o`005@3oool00`000000oooo0?ooo`310?ooo`030000003oool0oooo0<40oooo00<000000?ooo`3o
ool0403oool001D0oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`310?ooo`030000
003oool0oooo0100oooo000E0?ooo`030000003oool0oooo0<40oooo00<000000?ooo`3oool0`@3o
ool00`000000oooo0?ooo`0@0?ooo`005@3oool00`000000oooo0?ooo`310?ooo`030000003oool0
oooo0<40oooo00<000000?ooo`3oool0403oool001D0oooo00<000000?ooo`3oool0`@3oool00`00
0000oooo0?ooo`310?ooo`030000003oool0oooo0100oooo000E0?ooo`030000003oool0oooo0<40
oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`0@0?ooo`005@3oool00`000000oooo
0?ooo`310?ooo`030000003oool0oooo0<40oooo00<000000?ooo`3oool0403oool000/0ooooo`00
002N000000P0oooo000E0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0@0?ooo`005@3oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo
05l0oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo
0100oooo000E0?ooo`030000003oool0oooo0<40oooo00<000000?ooo`3oool0`@3oool00`000000
oooo0?ooo`0@0?ooo`005@3oool00`000000oooo0?ooo`310?ooo`030000003oool0oooo0<40oooo
00<000000?ooo`3oool0403oool001D0oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?oo
o`310?ooo`030000003oool0oooo0100oooo000E0?ooo`030000003oool0oooo0<40oooo00<00000
0?ooo`3oool0`@3oool00`000000oooo0?ooo`0@0?ooo`005@3oool00`000000oooo0?ooo`310?oo
o`030000003oool0oooo0<40oooo00<000000?ooo`3oool0403oool001D0oooo00<000000?ooo`3o
ool0`@3oool00`000000oooo0?ooo`300?ooo`030000003oool0oooo0140oooo000E0?ooo`030000
003oool0oooo0<40oooo00<000000?ooo`3oool0`03oool00`000000oooo0?ooo`0A0?ooo`005P3o
ool00`000000oooo0?ooo`300?ooo`030000003oool0oooo0<00oooo00<000000?ooo`3oool04@3o
ool001H0oooo00<000000?ooo`3oool0`03oool00`000000oooo0?ooo`300?ooo`030000003oool0
oooo0140oooo000F0?ooo`030000003oool0oooo0<00oooo0`0000300?ooo`030000003oool0oooo
0140oooo000F0?ooo`030000003oool0oooo0<00oooo00<000000?ooo`3oool0`03oool00`000000
oooo0?ooo`0A0?ooo`005`3oool00`000000oooo0?ooo`2o0?ooo`030000003oool0oooo0<00oooo
00<000000?ooo`3oool04@3oool001L0oooo00<000000?ooo`3oool0_`3oool00`000000oooo0?oo
o`300?ooo`030000003oool0oooo0140oooo000G0?ooo`030000003oool0oooo0;l0oooo00<00000
0?ooo`3oool0_`3oool00`000000oooo0?ooo`0B0?ooo`005`3oool00`000000oooo0?ooo`2o0?oo
o`030000003oool0oooo0;l0oooo00<000000?ooo`3oool04P3oool001P0oooo00<000000?ooo`3o
ool0_P3oool00`000000oooo0?ooo`2o0?ooo`030000003oool0oooo0180oooo000H0?ooo`030000
003oool0oooo0;h0oooo00<000000?ooo`3oool0_`3oool00`000000oooo0?ooo`0B0?ooo`00603o
ool00`000000oooo0?ooo`2n0?ooo`030000003oool0oooo0;h0oooo00<000000?ooo`3oool04`3o
ool001P0oooo00<000000?ooo`3oool0_P3oool00`000000oooo0?ooo`2n0?ooo`030000003oool0
oooo01<0oooo000I0?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0_P3oool00`00
0000oooo0?ooo`0C0?ooo`006@3oool00`000000oooo0?ooo`2m0?ooo`030000003oool0oooo0;d0
oooo00<000000?ooo`3oool0503oool001T0oooo00<000000?ooo`3oool0_@3oool300000;d0oooo
00<000000?ooo`3oool0503oool001X0oooo00<000000?ooo`3oool0_03oool00`000000oooo0?oo
o`2m0?ooo`030000003oool0oooo01@0oooo000J0?ooo`030000003oool0oooo0;`0oooo00<00000
0?ooo`3oool0_03oool00`000000oooo0?ooo`0E0?ooo`006P3oool00`000000oooo0?ooo`2l0?oo
o`030000003oool0oooo0;`0oooo00<000000?ooo`3oool05@3oool001/0oooo00<000000?ooo`3o
ool0^`3oool00`000000oooo0?ooo`2l0?ooo`030000003oool0oooo01D0oooo000K0?ooo`030000
003oool0oooo0;/0oooo00<000000?ooo`3oool0^`3oool00`000000oooo0?ooo`0F0?ooo`006`3o
ool00`000000oooo0?ooo`2k0?ooo`030000003oool0oooo0;/0oooo00<000000?ooo`3oool05P3o
ool001`0oooo00<000000?ooo`3oool0^P3oool00`000000oooo0?ooo`2j0?ooo`030000003oool0
oooo01L0oooo000L0?ooo`030000003oool0oooo0;X0oooo00<000000?ooo`3oool0^P3oool00`00
0000oooo0?ooo`0G0?ooo`007@3oool00`000000oooo0?ooo`2i0?ooo`030000003oool0oooo0;T0
oooo00<000000?ooo`3oool0603oool001h0oooo00<000000?ooo`3oool0^03oool00`000000oooo
0?ooo`2i0?ooo`030000003oool0oooo01P0oooo000N0?ooo`030000003oool0oooo0;P0oooo00<0
00000?ooo`3oool0^03oool00`000000oooo0?ooo`0I0?ooo`007`3oool00`000000oooo0?ooo`2g
0?ooo`<00000^03oool00`000000oooo0?ooo`0I0?ooo`007`3oool00`000000oooo0?ooo`2g0?oo
o`030000003oool0oooo0;L0oooo00<000000?ooo`3oool06P3oool00200oooo00<000000?ooo`3o
ool0]P3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo01X0oooo000P0?ooo`030000
003oool0oooo0;H0oooo00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`0K0?ooo`008@3o
ool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool06`3o
ool00280oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`2e0?ooo`030000003oool0
oooo01`0oooo000R0?ooo`030000003oool0oooo0;@0oooo00<000000?ooo`3oool0]@3oool00`00
0000oooo0?ooo`0L0?ooo`008`3oool00`000000oooo0?ooo`2c0?ooo`030000003oool0oooo0;@0
oooo00<000000?ooo`3oool07@3oool002<0oooo00<000000?ooo`3oool0/`3oool00`000000oooo
0?ooo`2c0?ooo`030000003oool0oooo01h0oooo000T0?ooo`030000003oool0oooo0;80oooo00<0
00000?ooo`3oool0/`3oool00`000000oooo0?ooo`0N0?ooo`00903oool00`000000oooo0?ooo`2b
0?ooo`030000003oool0oooo0;80oooo00<000000?ooo`3oool07`3oool002D0oooo00<000000?oo
o`3oool0/@3oool00`000000oooo0?ooo`2a0?ooo`030000003oool0oooo0200oooo000V0?ooo`03
0000003oool0oooo0;00oooo0`00002a0?ooo`030000003oool0oooo0200oooo000V0?ooo`030000
003oool0oooo0;00oooo00<000000?ooo`3oool0/03oool00`000000oooo0?ooo`0Q0?ooo`009`3o
ool00`000000oooo0?ooo`2_0?ooo`030000003oool0oooo0:l0oooo00<000000?ooo`3oool08P3o
ool002P0oooo00<000000?ooo`3oool0[P3oool00`000000oooo0?ooo`2^0?ooo`030000003oool0
oooo02<0oooo000Y0?ooo`030000003oool0oooo0:d0oooo00<000000?ooo`3oool0[P3oool00`00
0000oooo0?ooo`0S0?ooo`00:@3oool00`000000oooo0?ooo`2]0?ooo`030000003oool0oooo0:d0
oooo00<000000?ooo`3oool0903oool002X0oooo00<000000?ooo`3oool0[03oool00`000000oooo
0?ooo`2/0?ooo`030000003oool0oooo02D0oooo000[0?ooo`030000003oool0oooo08h0oooo00D0
omnL04P000000000001d0;ooo`030?ooo`060?noM0000000M;oo0?ooo`3ooml0W4P00`0000000`1d
_ol0oooo0?ooo`090?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool09P3oool002`0
oooo00<000000?ooo`3oool0S03oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool0
1`3o_g@0000007Boo`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`090?ooo`030000
003oool0oooo0:X0oooo00<000000?ooo`3oool09`3oool002`0oooo00<000000?ooo`3oool0S03o
ool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO09a8M02oool02@3o
ool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo02L0oooo000]0?ooo`030000003oool0
oooo08X0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0;0?ooo`030;md001d_ol0
oooo00P0oooo00<000000?ooo`3oool0Z@3oool00`000000oooo0?ooo`0X0?ooo`00;P3oool00`00
0000oooo0?ooo`290?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3o
oml0W4P004RLg`090?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0:@3oool002l0
oooo00<000000?ooo`3oool0R03oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0
oooo00<0gia80000001d_ol02@3oool400000:H0oooo00<000000?ooo`3oool0:P3oool00300oooo
00<000000?ooo`3oool0Q`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo
00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00T0oooo00<000000?ooo`3oool0YP3oool00`00
0000oooo0?ooo`0[0?ooo`00<@3oool00`000000oooo0?ooo`260?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol01P3oool01P3OW4P00000000000000000001809cOo`X0oooo00<00000
0?ooo`3oool0Y@3oool00`000000oooo0?ooo`0/0?ooo`00<P3oool00`000000oooo0?ooo`250?oo
o`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool0103ookl0M00000000018W=l<
0?ooo`030000003oool0oooo0:@0oooo00<000000?ooo`3oool0;@3oool003<0oooo00<000000?oo
o`3oool0Q@3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0_g@007Bo
o`3oool0303oool00`000000oooo0?ooo`2S0?ooo`030000003oool0oooo02h0oooo000d0?ooo`03
0000003oool0oooo08@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00H0omnL
04P00000000000000000B02Lgol90?ooo`030000003oool0oooo0:80oooo00<000000?ooo`3oool0
;`3oool003D0oooo00<000000?ooo`3oool0Q03oool01@3ookl0M0000000000007@0_ooo00X0oooo
00D0_g@0000000000000000004RLg`090?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool0<03oool003H0oooo00<000000?ooo`3oool0X03oool00`000000oooo0?ooo`2P0?ooo`030000
003oool0oooo0340oooo000g0?ooo`030000003oool0oooo09l0oooo00<000000?ooo`3oool0W`3o
ool00`000000oooo0?ooo`0b0?ooo`00>03oool2000009l0oooo00<000000?ooo`3oool0WP3oool0
0`000000oooo0?ooo`0c0?ooo`00>P3oool00`000000oooo0?ooo`2L0?ooo`030000003oool0oooo
09`0oooo0P00000f0?ooo`00>`3oool00`000000oooo0?ooo`2K0?ooo`030000003oool0oooo09/0
oooo00<000000?ooo`3oool0=P3oool003`0oooo0P00002K0?ooo`<00000VP3oool00`000000oooo
0?ooo`0g0?ooo`00?P3oool00`000000oooo0?ooo`2H0?ooo`030000003oool0oooo09T0oooo00<0
00000?ooo`3oool0>03oool003l0oooo00<000000?ooo`3oool0U`3oool00`000000oooo0?ooo`2G
0?ooo`800000>`3oool00400oooo0P00002G0?ooo`030000003oool0oooo09H0oooo00<000000?oo
o`3oool0>`3oool00480oooo00<000000?ooo`3oool0U03oool00`000000oooo0?ooo`2E0?ooo`03
0000003oool0oooo03`0oooo00130?ooo`030000003oool0oooo09<0oooo00<000000?ooo`3oool0
U03oool00`000000oooo0?ooo`0m0?ooo`00A03oool2000009<0oooo00<000000?ooo`3oool0TP3o
ool200000400oooo00160?ooo`030000003oool0oooo0900oooo00<000000?ooo`3oool0T@3oool0
0`000000oooo0?ooo`100?ooo`00A`3oool00`000000oooo0?ooo`2?0?ooo`030000003oool0oooo
0900oooo00<000000?ooo`3oool0@@3oool004P0oooo0P00002?0?ooo`030000003oool0oooo08h0
oooo0P0000140?ooo`00BP3oool00`000000oooo0?ooo`2<0?ooo`030000003oool0oooo08`0oooo
0P0000160?ooo`00B`3oool2000008`0oooo00<000000?ooo`3oool0R`3oool00`000000oooo0?oo
o`160?ooo`00C@3oool2000008X0oooo0`0000290?ooo`800000B@3oool004l0oooo00<000000?oo
o`3oool0Q`3oool00`000000oooo0?ooo`270?ooo`800000B`3oool00500oooo0P0000270?ooo`03
0000003oool0oooo08D0oooo0P00001=0?ooo`00DP3oool00`000000oooo0?ooo`240?ooo`030000
003oool0oooo08@0oooo00<000000?ooo`3oool0C@3oool005<0oooo0P0000240?ooo`030000003o
ool0oooo0880oooo0P00001@0?ooo`00E@3oool200000880oooo00<000000?ooo`3oool0P03oool2
00000580oooo001G0?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool0OP3oool20000
05@0oooo001H0?ooo`800000O`3oool00`000000oooo0?ooo`1l0?ooo`800000EP3oool005X0oooo
0P00001m0?ooo`030000003oool0oooo07X0oooo0P00001H0?ooo`00G03oool2000007/0oooo00<0
00000?ooo`3oool0N03oool2000005X0oooo001N0?ooo`800000N@3oool00`000000oooo0?ooo`1f
0?ooo`800000G03oool00600oooo0P00001g0?ooo`030000003oool0oooo07@0oooo0P00001N0?oo
o`00HP3oool2000007D0oooo0`00001b0?ooo`800000H03oool006@0oooo0P00001c0?ooo`030000
003oool0oooo0700oooo0P00001R0?ooo`00IP3oool300000700oooo00<000000?ooo`3oool0K@3o
ool3000006@0oooo001Y0?ooo`800000KP3oool00`000000oooo0?ooo`1[0?ooo`800000I`3oool0
06/0oooo0P00001/0?ooo`030000003oool0oooo06T0oooo0P00001Y0?ooo`00K@3oool2000006X0
oooo00<000000?ooo`3oool0I`3oool2000006/0oooo001_0?ooo`<00000I`3oool00`000000oooo
0?ooo`1T0?ooo`<00000K@3oool00780oooo0P00001U0?ooo`030000003oool0oooo0680oooo0P00
001`0?ooo`00M03oool2000006<0oooo00<000000?ooo`3oool0H03oool200000780oooo001f0?oo
o`<00000H03oool00`000000oooo0?ooo`1M0?ooo`<00000M03oool007T0oooo0`00001M0?ooo`03
0000003oool0oooo05X0oooo0`00001g0?ooo`00O03oool3000005X0oooo00<000000?ooo`3oool0
E`3oool3000007X0oooo001o0?ooo`@00000EP3oool00`000000oooo0?ooo`1C0?ooo`@00000O@3o
ool008<0oooo0`00001C0?ooo`<00000D03oool300000840oooo00260?ooo`@00000C`3oool00`00
0000oooo0?ooo`1<0?ooo`@00000Q03oool008X0oooo1000001;0?ooo`030000003oool0oooo04P0
oooo100000280?ooo`00SP3oool4000004L0oooo00<000000?ooo`3oool0A03oool4000008`0oooo
002B0?ooo`@00000@`3oool00`000000oooo0?ooo`100?ooo`@00000T03oool009H0oooo1@00000n
0?ooo`030000003oool0oooo03/0oooo1@00002D0?ooo`00V`3oool4000003X0oooo00<000000?oo
o`3oool0=P3oool5000009T0oooo002O0?ooo`H000009@3oool01P2oM000000000000000000004RL
0=ooo`T0oooo00<000000?ooo`3oool0<03oool6000009h0oooo002U0?ooo`P000007P3oool00`3o
gi`0B00007Boo`0;0?ooo`030000003oool0oooo02X0oooo1P00002T0?ooo`00[@3oool8000001H0
oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`0R0?ooo`P00000ZP3oool00;D0oooo
2000000>0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool06P3oool800000;80oooo
002m0?oooa4000002`3oool00`000000oooo0?ooo`0;0?ooo`l00000^P3oool00</0oooo00<0omnL
04P0001d_ol06@0000390?ooo`00b`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo
0=@0oooo003;0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0e03oool00</0oooo
00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`3D0?ooo`00b`3oool00`3ogi`0B00007Bo
o`0;0?ooo`030000003oool0oooo0=@0oooo00390?ooo`050?oog`2LB00000000000001d_ol02`3o
ool00`000000oooo0?ooo`3D0?ooo`00b`3oool00`3o_g@0000007Boo`0;0?ooo`030000003oool0
oooo0=@0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o
0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.10918, -1.09709, \
0.00765073, 0.0123791}}]
}, Closed]],

Cell[TextData[{
  "With default options the circle appears deformed into an ellipse.  A more \
accurate impression is obtained by choosing ",
  StyleBox["AspectRatio\[Rule]1", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[{\@\(1 - x\^2\), \(-\@\(1 - x\^2\)\)}, {x, \(-1\), 1}, 
        AspectRatio \[Rule] 1];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.5 0.476191 [
[.02381 .4875 -9.15625 -15.6875 ]
[.02381 .4875 9.15625 0 ]
[.2619 .4875 -13.6563 -15.6875 ]
[.2619 .4875 13.6563 0 ]
[.7381 .4875 -9.5 -15.6875 ]
[.7381 .4875 9.5 0 ]
[.97619 .4875 -5 -15.6875 ]
[.97619 .4875 5 0 ]
[.4875 .02381 -18.3125 -7.84375 ]
[.4875 .02381 0 7.84375 ]
[.4875 .2619 -27.3125 -7.84375 ]
[.4875 .2619 0 7.84375 ]
[.4875 .7381 -19 -7.84375 ]
[.4875 .7381 0 7.84375 ]
[.4875 .97619 -10 -7.84375 ]
[.4875 .97619 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 .5 m
.02381 .50625 L
s
gsave
.02381 .4875 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 .5 m
.2619 .50625 L
s
gsave
.2619 .4875 -74.6563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.7381 .5 m
.7381 .50625 L
s
gsave
.7381 .4875 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 .5 m
.97619 .50625 L
s
gsave
.97619 .4875 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .5 m
.07143 .50375 L
s
.11905 .5 m
.11905 .50375 L
s
.16667 .5 m
.16667 .50375 L
s
.21429 .5 m
.21429 .50375 L
s
.30952 .5 m
.30952 .50375 L
s
.35714 .5 m
.35714 .50375 L
s
.40476 .5 m
.40476 .50375 L
s
.45238 .5 m
.45238 .50375 L
s
.54762 .5 m
.54762 .50375 L
s
.59524 .5 m
.59524 .50375 L
s
.64286 .5 m
.64286 .50375 L
s
.69048 .5 m
.69048 .50375 L
s
.78571 .5 m
.78571 .50375 L
s
.83333 .5 m
.83333 .50375 L
s
.88095 .5 m
.88095 .50375 L
s
.92857 .5 m
.92857 .50375 L
s
.25 Mabswid
0 .5 m
1 .5 L
s
.5 .02381 m
.50625 .02381 L
s
gsave
.4875 .02381 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .2619 m
.50625 .2619 L
s
gsave
.4875 .2619 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .7381 m
.50625 .7381 L
s
gsave
.4875 .7381 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .97619 m
.50625 .97619 L
s
gsave
.4875 .97619 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .07143 m
.50375 .07143 L
s
.5 .11905 m
.50375 .11905 L
s
.5 .16667 m
.50375 .16667 L
s
.5 .21429 m
.50375 .21429 L
s
.5 .30952 m
.50375 .30952 L
s
.5 .35714 m
.50375 .35714 L
s
.5 .40476 m
.50375 .40476 L
s
.5 .45238 m
.50375 .45238 L
s
.5 .54762 m
.50375 .54762 L
s
.5 .59524 m
.50375 .59524 L
s
.5 .64286 m
.50375 .64286 L
s
.5 .69048 m
.50375 .69048 L
s
.5 .78571 m
.50375 .78571 L
s
.5 .83333 m
.50375 .83333 L
s
.5 .88095 m
.50375 .88095 L
s
.5 .92857 m
.50375 .92857 L
s
.25 Mabswid
.5 0 m
.5 1 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
.5 Mabswid
.02381 .50019 m
.02499 .53344 L
.02605 .54619 L
.02846 .56636 L
.03279 .59202 L
.03784 .61474 L
.04262 .63251 L
.06244 .68789 L
.08265 .72929 L
.10458 .76533 L
.14545 .81788 L
.1848 .85694 L
.22263 .88707 L
.26292 .91298 L
.30169 .93293 L
.34292 .94954 L
.36351 .95621 L
.38262 .9615 L
.40229 .96606 L
.42082 .96956 L
.44134 .97256 L
.45995 .9745 L
.46971 .97523 L
.47505 .97554 L
.48003 .97577 L
.48428 .97593 L
.48883 .97606 L
.49143 .97611 L
.49271 .97613 L
.49389 .97615 L
.49494 .97616 L
.4961 .97617 L
.49726 .97618 L
.4985 .97619 L
.49971 .97619 L
.50085 .97619 L
.50209 .97619 L
.50278 .97618 L
.50341 .97618 L
.50469 .97617 L
.50608 .97615 L
.50859 .97611 L
.51093 .97607 L
.51341 .976 L
.51785 .97586 L
.52298 .97564 L
.52778 .97538 L
.5387 .97462 L
.54911 .97365 L
.55899 .97252 L
Mistroke
.57753 .96984 L
.59737 .96613 L
.61577 .9619 L
.65495 .95027 L
.69659 .93372 L
.73671 .91319 L
.77531 .88854 L
.81636 .85591 L
.8559 .81637 L
.89393 .76754 L
.91511 .73332 L
.9344 .69507 L
.944 .67211 L
.95442 .64232 L
.9592 .62606 L
.96427 .60589 L
.9686 .58471 L
.97089 .57084 L
.97336 .5518 L
.97408 .54474 L
.97484 .53584 L
.97549 .52591 L
.97619 .50019 L
Mfstroke
.02381 .49981 m
.02499 .46656 L
.02605 .45381 L
.02846 .43364 L
.03279 .40798 L
.03784 .38526 L
.04262 .36749 L
.06244 .31211 L
.08265 .27071 L
.10458 .23467 L
.14545 .18212 L
.1848 .14306 L
.22263 .11293 L
.26292 .08702 L
.30169 .06707 L
.34292 .05046 L
.36351 .04379 L
.38262 .0385 L
.40229 .03394 L
.42082 .03044 L
.44134 .02744 L
.45995 .0255 L
.46971 .02477 L
.47505 .02446 L
.48003 .02423 L
.48428 .02407 L
.48883 .02394 L
.49143 .02389 L
.49271 .02387 L
.49389 .02385 L
.49494 .02384 L
.4961 .02383 L
.49726 .02382 L
.4985 .02381 L
.49971 .02381 L
.50085 .02381 L
.50209 .02381 L
.50278 .02382 L
.50341 .02382 L
.50469 .02383 L
.50608 .02385 L
.50859 .02389 L
.51093 .02393 L
.51341 .024 L
.51785 .02414 L
.52298 .02436 L
.52778 .02462 L
.5387 .02538 L
.54911 .02635 L
.55899 .02748 L
Mistroke
.57753 .03016 L
.59737 .03387 L
.61577 .0381 L
.65495 .04973 L
.69659 .06628 L
.73671 .08681 L
.77531 .11146 L
.81636 .14409 L
.8559 .18363 L
.89393 .23246 L
.91511 .26668 L
.9344 .30493 L
.944 .32789 L
.95442 .35768 L
.9592 .37394 L
.96427 .39411 L
.9686 .41529 L
.97089 .42916 L
.97336 .4482 L
.97408 .45526 L
.97484 .46416 L
.97549 .47409 L
.97619 .49981 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo
0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D
0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00
b@3oool01P2oM000000000000000000004RL0=ooo`X0oooo00<000000?ooo`3oool0e03oool00<X0
oooo00<0omnL04P0001d_ol0303oool00`000000oooo0?ooo`3D0?ooo`00bP3oool00`3ogi`0B000
07Boo`0<0?ooo`030000003oool0oooo0=@0oooo003:0?ooo`030?oOW0180000M;oo00`0oooo00<0
00000?ooo`3oool0e03oool00;/0oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`02
0?ooo`030?oOW0180000M;oo00P0oooo2000003C0?ooo`00a@3oool@000000@0oooo00@000000?oo
o`3oool0oooo3`0000340?ooo`00_03oool9000000D0oooo00<0omnL04P0001d_ol0303oool00`00
0000oooo0?ooo`0@0?ooo`X00000^P3oool00;H0oooo1P00000>0?ooo`030?oOW0180000M;oo00`0
oooo00<000000?ooo`3oool06P3oool600000;@0oooo002b0?ooo`@00000503oool00`3ogi`0B000
07Boo`0<0?ooo`030000003oool0oooo0200oooo1000002`0?ooo`00[P3oool4000001P0oooo00<0
omnL04P0001d_ol0303oool00`000000oooo0?ooo`0T0?ooo`@00000[03oool00:X0oooo1000000J
0?ooo`050?oog`2LB00000000000001d_ol0303oool00`000000oooo0?ooo`0X0?ooo`@00000Z03o
ool00:D0oooo1@00000P0?ooo`030?noM0000000M;oo00`0oooo00<000000?ooo`3oool0;03oool4
00000:@0oooo002Q0?ooo`@00000=03oool00`000000oooo0?ooo`0`0?ooo`@00000X03oool009d0
oooo1000000h0?ooo`030000003oool0oooo03@0oooo0`00002M0?ooo`00V`3oool2000003`0oooo
00<000000?ooo`3oool0=`3oool3000009X0oooo002H0?ooo`<00000?P3oool00`000000oooo0?oo
o`0j0?ooo`@00000UP3oool009D0oooo0`0000110?ooo`030000003oool0oooo03h0oooo0`00002C
0?ooo`00T`3oool2000004@0oooo00<000000?ooo`3oool0@@3oool200000940oooo002@0?ooo`<0
0000AP3oool00`000000oooo0?ooo`130?ooo`<00000SP3oool008h0oooo0P0000190?ooo`030000
003oool0oooo04H0oooo0P00002<0?ooo`00R`3oool3000004/0oooo00<000000?ooo`3oool0B03o
ool3000008T0oooo00280?ooo`<00000CP3oool00`000000oooo0?ooo`1;0?ooo`<00000QP3oool0
08H0oooo0P00001A0?ooo`030000003oool0oooo04h0oooo0P0000240?ooo`00P`3oool3000005<0
oooo00<000000?ooo`3oool0D03oool300000840oooo00210?ooo`800000EP3oool00`000000oooo
0?ooo`1C0?ooo`800000O`3oool007l0oooo0P00001H0?ooo`800000EP3oool2000007d0oooo001m
0?ooo`800000FP3oool00`000000oooo0?ooo`1G0?ooo`800000N`3oool007/0oooo0P00001L0?oo
o`030000003oool0oooo05T0oooo0P00001i0?ooo`00N@3oool2000005h0oooo00<000000?ooo`3o
ool0F`3oool2000007L0oooo001g0?ooo`800000H03oool00`000000oooo0?ooo`1M0?ooo`800000
M@3oool007D0oooo0P00001R0?ooo`030000003oool0oooo05l0oooo0P00001c0?ooo`00L`3oool2
000006@0oooo00<000000?ooo`3oool0H@3oool200000740oooo001a0?ooo`800000IP3oool00`00
0000oooo0?ooo`1S0?ooo`800000K`3oool00700oooo00<000000?ooo`3oool0IP3oool00`000000
oooo0?ooo`1U0?ooo`030000003oool0oooo06`0oooo001_0?ooo`030000003oool0oooo06L0oooo
00<000000?ooo`3oool0IP3oool2000006`0oooo001]0?ooo`800000JP3oool00`000000oooo0?oo
o`1X0?ooo`030000003oool0oooo06T0oooo001/0?ooo`030000003oool0oooo06X0oooo00<00000
0?ooo`3oool0J@3oool00`000000oooo0?ooo`1X0?ooo`00JP3oool2000006d0oooo00<000000?oo
o`3oool0JP3oool2000006P0oooo001Y0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3o
ool0K03oool00`000000oooo0?ooo`1U0?ooo`00I`3oool200000700oooo00<000000?ooo`3oool0
K@3oool2000006D0oooo001V0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0K`3o
ool00`000000oooo0?ooo`1R0?ooo`00I@3oool00`000000oooo0?ooo`1a0?ooo`030000003oool0
oooo0700oooo0P00001R0?ooo`00H`3oool2000007@0oooo00<000000?ooo`3oool0LP3oool00`00
0000oooo0?ooo`1O0?ooo`00HP3oool00`000000oooo0?ooo`1d0?ooo`030000003oool0oooo07<0
oooo00<000000?ooo`3oool0GP3oool00640oooo00<000000?ooo`3oool0M@3oool00`000000oooo
0?ooo`1d0?ooo`800000GP3oool005l0oooo0P00001h0?ooo`800000M`3oool00`000000oooo0?oo
o`1K0?ooo`00GP3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo07L0oooo00<00000
0?ooo`3oool0FP3oool005d0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`1h0?oo
o`800000FP3oool005/0oooo0P00001l0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3o
ool0E`3oool005X0oooo00<000000?ooo`3oool0O03oool00`000000oooo0?ooo`1k0?ooo`030000
003oool0oooo05H0oooo001I0?ooo`030000003oool0oooo07d0oooo00<000000?ooo`3oool0O03o
ool00`000000oooo0?ooo`1E0?ooo`00E`3oool200000800oooo00<000000?ooo`3oool0O@3oool2
000005D0oooo001F0?ooo`030000003oool0oooo0800oooo00<000000?ooo`3oool0O`3oool00`00
0000oooo0?ooo`1B0?ooo`00E@3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo0800
oooo00<000000?ooo`3oool0D@3oool005<0oooo0P0000240?ooo`030000003oool0oooo0840oooo
0P00001A0?ooo`00DP3oool00`000000oooo0?ooo`240?ooo`030000003oool0oooo08<0oooo00<0
00000?ooo`3oool0CP3oool00540oooo00<000000?ooo`3oool0Q@3oool00`000000oooo0?ooo`24
0?ooo`030000003oool0oooo04d0oooo001@0?ooo`030000003oool0oooo08H0oooo00<000000?oo
o`3oool0Q@3oool00`000000oooo0?ooo`1<0?ooo`00C`3oool00`000000oooo0?ooo`270?ooo`03
0000003oool0oooo08H0oooo00<000000?ooo`3oool0B`3oool004h0oooo00<000000?ooo`3oool0
R03oool00`000000oooo0?ooo`270?ooo`030000003oool0oooo04X0oooo001=0?ooo`030000003o
ool0oooo08T0oooo00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`190?ooo`00C03oool0
0`000000oooo0?ooo`2:0?ooo`030000003oool0oooo08T0oooo00<000000?ooo`3oool0B03oool0
04/0oooo00<000000?ooo`3oool0R`3oool00`000000oooo0?ooo`2:0?ooo`030000003oool0oooo
04L0oooo001:0?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool0R`3oool00`000000
oooo0?ooo`160?ooo`00B@3oool00`000000oooo0?ooo`2=0?ooo`030000003oool0oooo08`0oooo
00<000000?ooo`3oool0A@3oool004P0oooo00<000000?ooo`3oool0SP3oool00`000000oooo0?oo
o`2=0?ooo`030000003oool0oooo04@0oooo00170?ooo`030000003oool0oooo08l0oooo0P00002?
0?ooo`030000003oool0oooo04<0oooo00160?ooo`030000003oool0oooo0900oooo00<000000?oo
o`3oool0S`3oool00`000000oooo0?ooo`120?ooo`00A@3oool00`000000oooo0?ooo`2A0?ooo`03
0000003oool0oooo0900oooo00<000000?ooo`3oool0@@3oool004@0oooo00<000000?ooo`3oool0
TP3oool00`000000oooo0?ooo`2A0?ooo`030000003oool0oooo0400oooo00130?ooo`030000003o
ool0oooo09<0oooo00<000000?ooo`3oool0TP3oool00`000000oooo0?ooo`0o0?ooo`00@P3oool0
0`000000oooo0?ooo`2D0?ooo`030000003oool0oooo09<0oooo00<000000?ooo`3oool0?P3oool0
0440oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?ooo`2D0?ooo`030000003oool0oooo
03d0oooo00100?ooo`030000003oool0oooo09H0oooo00<000000?ooo`3oool0U@3oool00`000000
oooo0?ooo`0l0?ooo`00@03oool00`000000oooo0?ooo`2F0?ooo`030000003oool0oooo09H0oooo
00<000000?ooo`3oool0>`3oool003l0oooo00<000000?ooo`3oool0U`3oool00`000000oooo0?oo
o`2G0?ooo`030000003oool0oooo03X0oooo000n0?ooo`030000003oool0oooo09P0oooo00<00000
0?ooo`3oool0U`3oool00`000000oooo0?ooo`0j0?ooo`00?@3oool00`000000oooo0?ooo`2I0?oo
o`030000003oool0oooo09P0oooo00<000000?ooo`3oool0>@3oool003d0oooo00<000000?ooo`3o
ool0V@3oool00`000000oooo0?ooo`2I0?ooo`030000003oool0oooo03P0oooo000l0?ooo`030000
003oool0oooo09X0oooo00<000000?ooo`3oool0VP3oool00`000000oooo0?ooo`0g0?ooo`00>`3o
ool00`000000oooo0?ooo`2K0?ooo`030000003oool0oooo09X0oooo00<000000?ooo`3oool0=`3o
ool003X0oooo00<000000?ooo`3oool0W03oool00`000000oooo0?ooo`2K0?ooo`030000003oool0
oooo03H0oooo000j0?ooo`030000003oool0oooo09`0oooo00<000000?ooo`3oool0W03oool00`00
0000oooo0?ooo`0e0?ooo`00>@3oool00`000000oooo0?ooo`2M0?ooo`030000003oool0oooo09d0
oooo00<000000?ooo`3oool0=03oool003P0oooo00<000000?ooo`3oool0WP3oool00`000000oooo
0?ooo`2M0?ooo`030000003oool0oooo03@0oooo000g0?ooo`030000003oool0oooo09l0oooo00<0
00000?ooo`3oool0WP3oool00`000000oooo0?ooo`0c0?ooo`00=`3oool00`000000oooo0?ooo`2O
0?ooo`800000X03oool00`000000oooo0?ooo`0b0?ooo`00=P3oool00`000000oooo0?ooo`2P0?oo
o`030000003oool0oooo0:00oooo00<000000?ooo`3oool0<@3oool003D0oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`2P0?ooo`030000003oool0oooo0340oooo000d0?ooo`030000
003oool0oooo0:80oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0`0?ooo`00=03o
ool00`000000oooo0?ooo`2R0?ooo`030000003oool0oooo0:80oooo00<000000?ooo`3oool0;`3o
ool003<0oooo00<000000?ooo`3oool0X`3oool00`000000oooo0?ooo`2S0?ooo`030000003oool0
oooo02h0oooo000b0?ooo`030000003oool0oooo0:@0oooo00<000000?ooo`3oool0X`3oool00`00
0000oooo0?ooo`0^0?ooo`00<@3oool00`000000oooo0?ooo`2U0?ooo`030000003oool0oooo0:@0
oooo00<000000?ooo`3oool0;@3oool00340oooo00<000000?ooo`3oool0Y@3oool00`000000oooo
0?ooo`2U0?ooo`030000003oool0oooo02`0oooo000`0?ooo`030000003oool0oooo0:H0oooo00<0
00000?ooo`3oool0YP3oool00`000000oooo0?ooo`0[0?ooo`00;`3oool00`000000oooo0?ooo`2W
0?ooo`030000003oool0oooo0:H0oooo00<000000?ooo`3oool0:`3oool002l0oooo00<000000?oo
o`3oool0Y`3oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo02X0oooo000^0?ooo`03
0000003oool0oooo0:P0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0Y0?ooo`00
;@3oool00`000000oooo0?ooo`2Y0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0
:@3oool002d0oooo00<000000?ooo`3oool0Z@3oool00`000000oooo0?ooo`2Y0?ooo`030000003o
ool0oooo02P0oooo000/0?ooo`030000003oool0oooo08d0oooo00D0omnL04P000000000001d0;oo
o`030?ooo`060?noM0000000M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`090?ooo`03
0000003oool0oooo0:T0oooo00<000000?ooo`3oool0:03oool002/0oooo00<000000?ooo`3oool0
S@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@0000007Boo`3oool0
oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`090?ooo`030000003oool0oooo0:X0oooo00<0
00000?ooo`3oool09`3oool002/0oooo00<000000?ooo`3oool0S@3oool00`2oM00004RL0=ooo`02
0?ooo`030;md0000M;l0oooo00X0oooo00<0oooO09a8M02oool02@3oool00`000000oooo0?ooo`2[
0?ooo`030000003oool0oooo02H0oooo000Z0?ooo`030000003oool0oooo08d0oooo00<0oono07@0
0018W=l00`3oool00`3OW4P0001d0;ooo`0;0?ooo`030;md001d_ol0oooo00P0oooo00<000000?oo
o`3oool0Z`3oool00`000000oooo0?ooo`0V0?ooo`00:@3oool00`000000oooo0?ooo`210?ooo`03
0?oog`2LB000000000L0000000H0B9cO0?ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0
W=oo00X0oooo00<0oooO09a80018W=l02@3oool00`000000oooo0?ooo`2/0?ooo`030000003oool0
oooo02D0oooo000Y0?ooo`030000003oool0oooo08h0oooo00<0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`0:0?ooo`030=nLB0000000M;oo00T0oooo0`00002]0?ooo`030000003oool0oooo
02@0oooo000X0?ooo`030000003oool0oooo08l0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`060?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`090?ooo`030000003oool0
oooo0:d0oooo00<000000?ooo`3oool0903oool002L0oooo00<000000?ooo`3oool0T03oool00`3o
gi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00H0gia800000000000000000000B02L
gol:0?ooo`030000003oool0oooo0:h0oooo00<000000?ooo`3oool08`3oool002L0oooo00<00000
0?ooo`3oool0T03oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono
07@000000000B9cO303oool00`000000oooo0?ooo`2^0?ooo`030000003oool0oooo02<0oooo000V
0?ooo`030000003oool0oooo0980oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`07
0?ooo`030;md001d_ol0oooo00`0oooo00<000000?ooo`3oool0[`3oool00`000000oooo0?ooo`0R
0?ooo`009P3oool00`000000oooo0?ooo`2B0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oo
o`080?ooo`060?oOW018000000000000000004P0W=oo2@3oool00`000000oooo0?ooo`2_0?ooo`03
0000003oool0oooo0280oooo000U0?ooo`030000003oool0oooo09@0oooo00D0oono07@000000000
001d0;ooo`0:0?ooo`050;md00000000000000000018W=l02@3oool00`000000oooo0?ooo`2`0?oo
o`030000003oool0oooo0240oooo000U0?ooo`030000003oool0oooo0;40oooo00<000000?ooo`3o
ool0/03oool00`000000oooo0?ooo`0Q0?ooo`00903oool00`000000oooo0?ooo`2b0?ooo`030000
003oool0oooo0;40oooo00<000000?ooo`3oool0803oool002@0oooo00<000000?ooo`3oool0/P3o
ool00`000000oooo0?ooo`2a0?ooo`030000003oool0oooo0200oooo000S0?ooo`030000003oool0
oooo0;<0oooo00<000000?ooo`3oool0/P3oool00`000000oooo0?ooo`0O0?ooo`008`3oool00`00
0000oooo0?ooo`2c0?ooo`030000003oool0oooo0;80oooo00<000000?ooo`3oool07`3oool00280
oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`2c0?ooo`030000003oool0oooo01h0
oooo000R0?ooo`030000003oool0oooo0;@0oooo00<000000?ooo`3oool0/`3oool00`000000oooo
0?ooo`0N0?ooo`008P3oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo0;@0oooo00<0
00000?ooo`3oool07@3oool00240oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`2d
0?ooo`030000003oool0oooo01d0oooo000Q0?ooo`030000003oool0oooo0;D0oooo00<000000?oo
o`3oool0]@3oool00`000000oooo0?ooo`0L0?ooo`00803oool00`000000oooo0?ooo`2f0?ooo`03
0000003oool0oooo0;D0oooo00<000000?ooo`3oool0703oool00200oooo00<000000?ooo`3oool0
]P3oool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo01/0oooo000O0?ooo`030000003o
ool0oooo0;L0oooo00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`0K0?ooo`007`3oool0
0`000000oooo0?ooo`2g0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool06P3oool0
01h0oooo00<000000?ooo`3oool0^03oool200000;P0oooo00<000000?ooo`3oool06P3oool001h0
oooo00<000000?ooo`3oool0^03oool00`000000oooo0?ooo`2h0?ooo`030000003oool0oooo01T0
oooo000N0?ooo`030000003oool0oooo0;P0oooo00<000000?ooo`3oool0^03oool00`000000oooo
0?ooo`0I0?ooo`007@3oool00`000000oooo0?ooo`2i0?ooo`030000003oool0oooo0;T0oooo00<0
00000?ooo`3oool0603oool001d0oooo00<000000?ooo`3oool0^@3oool00`000000oooo0?ooo`2i
0?ooo`030000003oool0oooo01P0oooo000L0?ooo`030000003oool0oooo0;X0oooo00<000000?oo
o`3oool0^P3oool00`000000oooo0?ooo`0G0?ooo`00703oool00`000000oooo0?ooo`2j0?ooo`03
0000003oool0oooo0;X0oooo00<000000?ooo`3oool05`3oool001`0oooo00<000000?ooo`3oool0
^P3oool00`000000oooo0?ooo`2k0?ooo`030000003oool0oooo01H0oooo000K0?ooo`030000003o
ool0oooo0;/0oooo00<000000?ooo`3oool0^`3oool00`000000oooo0?ooo`0F0?ooo`006`3oool0
0`000000oooo0?ooo`2k0?ooo`030000003oool0oooo0;`0oooo00<000000?ooo`3oool05@3oool0
01X0oooo00<000000?ooo`3oool0_03oool00`000000oooo0?ooo`2l0?ooo`030000003oool0oooo
01D0oooo000J0?ooo`030000003oool0oooo0;`0oooo00<000000?ooo`3oool0_03oool00`000000
oooo0?ooo`0E0?ooo`006P3oool00`000000oooo0?ooo`2l0?ooo`030000003oool0oooo0;d0oooo
00<000000?ooo`3oool0503oool001T0oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?oo
o`2m0?ooo`030000003oool0oooo01@0oooo000I0?ooo`030000003oool0oooo0;d0oooo00<00000
0?ooo`3oool0_@3oool00`000000oooo0?ooo`0D0?ooo`006@3oool00`000000oooo0?ooo`2m0?oo
o`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0503oool001P0oooo00<000000?ooo`3o
ool0_P3oool00`000000oooo0?ooo`2n0?ooo`030000003oool0oooo01<0oooo000H0?ooo`030000
003oool0oooo0;h0oooo00<000000?ooo`3oool0_P3oool00`000000oooo0?ooo`0C0?ooo`005`3o
ool00`000000oooo0?ooo`2o0?ooo`030000003oool0oooo0;h0oooo00<000000?ooo`3oool04`3o
ool001L0oooo00<000000?ooo`3oool0_`3oool00`000000oooo0?ooo`2o0?ooo`030000003oool0
oooo0180oooo000G0?ooo`030000003oool0oooo0;l0oooo0P0000300?ooo`030000003oool0oooo
0180oooo000F0?ooo`030000003oool0oooo0<00oooo00<000000?ooo`3oool0_`3oool00`000000
oooo0?ooo`0B0?ooo`005P3oool00`000000oooo0?ooo`300?ooo`030000003oool0oooo0<00oooo
00<000000?ooo`3oool04@3oool001D0oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?oo
o`300?ooo`030000003oool0oooo0140oooo000E0?ooo`030000003oool0oooo0<40oooo00<00000
0?ooo`3oool0`03oool00`000000oooo0?ooo`0A0?ooo`005@3oool00`000000oooo0?ooo`310?oo
o`030000003oool0oooo0<00oooo00<000000?ooo`3oool04@3oool001@0oooo00<000000?ooo`3o
ool0`P3oool00`000000oooo0?ooo`310?ooo`030000003oool0oooo0100oooo000D0?ooo`030000
003oool0oooo0<80oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`0@0?ooo`00503o
ool00`000000oooo0?ooo`320?ooo`030000003oool0oooo0<40oooo00<000000?ooo`3oool0403o
ool001@0oooo00<000000?ooo`3oool0`P3oool00`000000oooo0?ooo`310?ooo`030000003oool0
oooo0100oooo000C0?ooo`030000003oool0oooo0<<0oooo00<000000?ooo`3oool0`P3oool00`00
0000oooo0?ooo`0?0?ooo`004`3oool00`000000oooo0?ooo`330?ooo`030000003oool0oooo0<80
oooo00<000000?ooo`3oool03`3oool001<0oooo00<000000?ooo`3oool0``3oool00`000000oooo
0?ooo`320?ooo`030000003oool0oooo00l0oooo000C0?ooo`030000003oool0oooo0<<0oooo00<0
00000?ooo`3oool0`P3oool00`000000oooo0?ooo`0?0?ooo`004`3oool00`000000oooo0?ooo`33
0?ooo`030000003oool0oooo0<<0oooo00<000000?ooo`3oool03P3oool00180oooo00<000000?oo
o`3oool0a03oool00`000000oooo0?ooo`330?ooo`030000003oool0oooo00h0oooo000B0?ooo`03
0000003oool0oooo0<@0oooo00<000000?ooo`3oool0``3oool00`000000oooo0?ooo`0>0?ooo`00
4P3oool00`000000oooo0?ooo`340?ooo`030000003oool0oooo0<<0oooo00<000000?ooo`3oool0
3P3oool00180oooo00<000000?ooo`3oool0a03oool00`000000oooo0?ooo`330?ooo`030000003o
ool0oooo00h0oooo000B0?ooo`030000003oool0oooo0<@0oooo00<000000?ooo`3oool0a03oool0
0`000000oooo0?ooo`0=0?ooo`004@3oool00`000000oooo0?ooo`350?ooo`800000a@3oool00`00
0000oooo0?ooo`0=0?ooo`004@3oool00`000000oooo0?ooo`350?ooo`030000003oool0oooo0<@0
oooo00<000000?ooo`3oool03@3oool00140oooo00<000000?ooo`3oool0a@3oool00`000000oooo
0?ooo`340?ooo`030000003oool0oooo00d0oooo000A0?ooo`030000003oool0oooo0<D0oooo00<0
00000?ooo`3oool0a@3oool00`000000oooo0?ooo`0<0?ooo`004@3oool00`000000oooo0?ooo`35
0?ooo`030000003oool0oooo0<D0oooo00<000000?ooo`3oool0303oool00100oooo00<000000?oo
o`3oool0aP3oool00`000000oooo0?ooo`350?ooo`030000003oool0oooo00`0oooo000@0?ooo`03
0000003oool0oooo0<H0oooo00<000000?ooo`3oool0a@3oool00`000000oooo0?ooo`0<0?ooo`00
403oool00`000000oooo0?ooo`360?ooo`030000003oool0oooo0<D0oooo00<000000?ooo`3oool0
303oool00100oooo00<000000?ooo`3oool0aP3oool00`000000oooo0?ooo`350?ooo`030000003o
ool0oooo00`0oooo000@0?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3oool0aP3oool0
0`000000oooo0?ooo`0;0?ooo`00403oool00`000000oooo0?ooo`360?ooo`030000003oool0oooo
0<H0oooo00<000000?ooo`3oool02`3oool000l0oooo00<000000?ooo`3oool0a`3oool00`000000
oooo0?ooo`360?ooo`030000003oool0oooo00/0oooo000?0?ooo`030000003oool0oooo0<L0oooo
00<000000?ooo`3oool0aP3oool00`000000oooo0?ooo`0;0?ooo`003`3oool00`000000oooo0?oo
o`370?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3oool02`3oool000l0oooo00<00000
0?ooo`3oool0a`3oool00`000000oooo0?ooo`360?ooo`030000003oool0oooo00/0oooo000?0?oo
o`030000003oool0oooo0<L0oooo00<000000?ooo`3oool0aP3oool00`000000oooo0?ooo`0;0?oo
o`003`3oool00`000000oooo0?ooo`370?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3o
ool02P3oool000h0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`370?ooo`030000
003oool0oooo00X0oooo000>0?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3oool0a`3o
ool00`000000oooo0?ooo`0:0?ooo`003P3oool010000000oooo0?ooo`2oM003000000030018W03O
ool0oooo05P0oooo00D0omnL04P000000000001d0;ooo`030?ooo`060?noM0000000M;oo0?ooo`3o
oml0W4P00`0000000`1d_ol0oooo0?ooo`1E0?ooo`030000003oool0oooo05T0oooo00D0omnL04P0
00000000001d0;ooo`030?ooo`060?noM0000000M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo
0?ooo`1I0?ooo`060;md00000000000000000000B9`0gooo203oool000h0oooo00L000000?ooo`3o
ool0oooo0?oOW0180000M;oo05X0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo
00L0okmd0000001d_ol0oooo0?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool0E@3oool00`00
0000oooo0?ooo`1H0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM000
0000M;oo0?ooo`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo05X0oooo00<0000004P0001d
_ol02P3oool000h0oooo00L000000?ooo`3oool0oooo0?oOW0180000M;oo05X0oooo00<0_g@00018
W03Oool00P3oool00`2oM00007Bo0?ooo`0:0?ooo`030?oog`2LB7@0_ooo05D0oooo0P00001I0?oo
o`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd0;ooo`1J0?oo
o`03000000180000M;oo00X0oooo000>0?ooo`070000003oool0oooo0?ooo`3ogi`0B00007Boo`1I
0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02`3oool00`2oM000M;oo0?ooo`1D
0?ooo`030000003oool0oooo05L0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0;
0?ooo`030;md001d_ol0oooo05T0oooo00<0000004P0001d_ol02P3oool000<0oooo00<0oooO09a8
000000001`0000000`18W=l000000?ooo`020?ooo`030?oOW0180000M;oo04d0oooo00<0oooO09a8
000000001`0000001@18W=l0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool0
0`3ooml0W4P004RLg`1E0?ooo`030000003oool0oooo05L0oooo00<0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`0:0?ooo`030?oog`2LB000B9cO05X0oooo00<0000004P0001d_ol02P3oool0
00h0oooo00L000000?ooo`3oool0oooo0?oOW0180000M;oo05T0oooo00<0omnL04P0001d_ol00`3o
ool00`3o_g@0001809cOo`0:0?ooo`030=nLB0000000M;oo05D0oooo00<000000?ooo`3oool0E`3o
ool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0gia80000001d_ol0FP3o
ool00`000000B00007Boo`0:0?ooo`003P3oool01`000000oooo0?ooo`3oool0omnL04P0001d_ol0
F@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0
M0000000000007@0_ooo05D0oooo00<000000?ooo`3oool0E`3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo05X0oooo00<0
omnL0000001d_ol02P3oool000h0oooo00L000000?ooo`3oool0oooo0?oOW0180000M;oo05T0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`060=nLB0000000000000000000
04P0W=ooEP3oool00`000000oooo0?ooo`1G0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol01P3oool01P3OW4P00000000000000000001809cOoe/0oooo00<0omnL0000001d_ol02P3o
ool000h0oooo00L000000?ooo`3oool0oooo0?oOW0180000M;oo05T0oooo00<0oono07@00018W=l0
0`3oool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d0000000004RLgeP0oooo00<000000?ooo`3o
ool0E`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono07@00000
0000B9cOG@3oool00`3ogi`0000007Boo`0:0?ooo`003P3oool01`000000oooo0?ooo`3oool0omnL
04P0001d_ol0FP3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0_g@0
07Boo`3oool0F03oool00`000000oooo0?ooo`1H0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0
001d_`3oool01`3oool00`2oM000M;oo0?ooo`1M0?ooo`030?oOW0000000M;oo00X0oooo000>0?oo
o`040000003oool0oooO09a80080000000<0M;oo0?ooo`3oool0F03oool01`3ookl0M00007Boo`3o
ool0okmd0000M02oool0203oool01P3ogi`0B000000000000000001809cOoeD0oooo00<000000?oo
o`3oool0F03oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0203oool01P3ogi`0B0000000
00000000001809cOoeP0oooo00D0oooO09a800000000000007Boo`0:0?ooo`003P3oool01`000000
oooo0?ooo`3oool0okmd0000001d_ol0F`3oool01@3ookl0M0000000000007@0_ooo00X0oooo00D0
_g@0000000000000000004RLg`1E0?ooo`030000003oool0oooo05T0oooo00D0oono07@000000000
001d0;ooo`0:0?ooo`050;md00000000000000000018W=l0FP3oool00`3o_g@0000007Boo`0:0?oo
o`003P3oool00`000000oooo0?ooo`380?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3o
ool02@3oool000h0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`380?ooo`030000
003oool0oooo00T0oooo000>0?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3oool0b03o
ool00`000000oooo0?ooo`090?ooo`003@3oool00`000000oooo0?ooo`390?ooo`030000003oool0
oooo0<P0oooo00<000000?ooo`3oool02@3oool000d0oooo00<000000?ooo`3oool0b@3oool00`00
0000oooo0?ooo`380?ooo`030000003oool0oooo00T0oooo000=0?ooo`030000003oool0oooo0<T0
oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`090?ooo`003@3oool00`000000oooo
0?ooo`390?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3oool02@3oool000d0oooo00<0
00000?ooo`3oool0b@3oool00`000000oooo0?ooo`380?ooo`030000003oool0oooo00T0oooo000=
0?ooo`030000003oool0oooo0<T0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`09
0?ooo`003@3oool00`000000oooo0?ooo`390?ooo`030000003oool0oooo0<P0oooo00<000000?oo
o`3oool02@3oool000<0ooooo`00002]00000040oooo000=0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0140oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0140oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`090?ooo`003@3oool00`000000oooo0?oo
o`0B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`0B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`0B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`0B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`0B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool02@3oool000d0oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool0HP3oool00`000000oooo0?ooo`090?ooo`003@3oool00`000000oooo0?ooo`390?oo
o`030000003oool0oooo0<P0oooo00<000000?ooo`3oool02@3oool000d0oooo00<000000?ooo`3o
ool0b@3oool00`000000oooo0?ooo`380?ooo`030000003oool0oooo00T0oooo000=0?ooo`030000
003oool0oooo0<T0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`090?ooo`003@3o
ool00`000000oooo0?ooo`390?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3oool02@3o
ool000h0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`380?ooo`030000003oool0
oooo00T0oooo000>0?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3oool0b03oool00`00
0000oooo0?ooo`090?ooo`003P3oool00`000000oooo0?ooo`380?ooo`030000003oool0oooo0<P0
oooo00<000000?ooo`3oool02@3oool000h0oooo00<000000?ooo`3oool0b03oool00`000000oooo
0?ooo`380?ooo`030000003oool0oooo00T0oooo000>0?ooo`030000003oool0oooo0<P0oooo00<0
00000?ooo`3oool0b03oool00`000000oooo0?ooo`090?ooo`003P3oool00`000000oooo0?ooo`38
0?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3oool02@3oool000h0oooo00<000000?oo
o`3oool0b03oool00`000000oooo0?ooo`380?ooo`030000003oool0oooo00T0oooo000>0?ooo`03
0000003oool0oooo0<P0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`090?ooo`00
3P3oool00`000000oooo0?ooo`380?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3oool0
2@3oool000h0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`370?ooo`030000003o
ool0oooo00X0oooo000>0?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3oool0a`3oool0
0`000000oooo0?ooo`0:0?ooo`003P3oool00`000000oooo0?ooo`380?ooo`030000003oool0oooo
0<L0oooo00<000000?ooo`3oool02P3oool000h0oooo00<000000?ooo`3oool0b03oool200000<P0
oooo00<000000?ooo`3oool02P3oool000h0oooo00<000000?ooo`3oool0b03oool00`000000oooo
0?ooo`370?ooo`030000003oool0oooo00X0oooo000>0?ooo`030000003oool0oooo0<P0oooo00<0
00000?ooo`3oool0a`3oool00`000000oooo0?ooo`0:0?ooo`003P3oool00`000000oooo0?ooo`38
0?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3oool02P3oool000h0oooo00<000000?oo
o`3oool0b03oool00`000000oooo0?ooo`370?ooo`030000003oool0oooo00X0oooo000?0?ooo`03
0000003oool0oooo0<L0oooo00<000000?ooo`3oool0a`3oool00`000000oooo0?ooo`0:0?ooo`00
3`3oool00`000000oooo0?ooo`370?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3oool0
2`3oool000l0oooo00<000000?ooo`3oool0a`3oool00`000000oooo0?ooo`360?ooo`030000003o
ool0oooo00/0oooo000?0?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3oool0aP3oool0
0`000000oooo0?ooo`0;0?ooo`003`3oool00`000000oooo0?ooo`370?ooo`030000003oool0oooo
0<H0oooo00<000000?ooo`3oool02`3oool000l0oooo00<000000?ooo`3oool0a`3oool00`000000
oooo0?ooo`360?ooo`030000003oool0oooo00/0oooo000@0?ooo`030000003oool0oooo0<H0oooo
00<000000?ooo`3oool0aP3oool00`000000oooo0?ooo`0;0?ooo`00403oool00`000000oooo0?oo
o`360?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3oool02`3oool00100oooo00<00000
0?ooo`3oool0aP3oool00`000000oooo0?ooo`350?ooo`030000003oool0oooo00`0oooo000@0?oo
o`030000003oool0oooo0<H0oooo00<000000?ooo`3oool0a@3oool00`000000oooo0?ooo`0<0?oo
o`00403oool00`000000oooo0?ooo`360?ooo`030000003oool0oooo0<D0oooo00<000000?ooo`3o
ool0303oool00100oooo00<000000?ooo`3oool0aP3oool00`000000oooo0?ooo`350?ooo`030000
003oool0oooo00`0oooo000A0?ooo`030000003oool0oooo0<D0oooo00<000000?ooo`3oool0a@3o
ool00`000000oooo0?ooo`0<0?ooo`004@3oool00`000000oooo0?ooo`350?ooo`030000003oool0
oooo0<D0oooo00<000000?ooo`3oool0303oool00140oooo00<000000?ooo`3oool0a@3oool00`00
0000oooo0?ooo`340?ooo`030000003oool0oooo00d0oooo000A0?ooo`030000003oool0oooo0<D0
oooo00<000000?ooo`3oool0a03oool00`000000oooo0?ooo`0=0?ooo`004@3oool00`000000oooo
0?ooo`350?ooo`800000a@3oool00`000000oooo0?ooo`0=0?ooo`004P3oool00`000000oooo0?oo
o`340?ooo`030000003oool0oooo0<@0oooo00<000000?ooo`3oool03@3oool00180oooo00<00000
0?ooo`3oool0a03oool00`000000oooo0?ooo`330?ooo`030000003oool0oooo00h0oooo000B0?oo
o`030000003oool0oooo0<@0oooo00<000000?ooo`3oool0``3oool00`000000oooo0?ooo`0>0?oo
o`004P3oool00`000000oooo0?ooo`340?ooo`030000003oool0oooo0<<0oooo00<000000?ooo`3o
ool03P3oool00180oooo00<000000?ooo`3oool0a03oool00`000000oooo0?ooo`330?ooo`030000
003oool0oooo00h0oooo000C0?ooo`030000003oool0oooo0<<0oooo00<000000?ooo`3oool0``3o
ool00`000000oooo0?ooo`0>0?ooo`004`3oool00`000000oooo0?ooo`330?ooo`030000003oool0
oooo0<80oooo00<000000?ooo`3oool03`3oool001<0oooo00<000000?ooo`3oool0``3oool00`00
0000oooo0?ooo`320?ooo`030000003oool0oooo00l0oooo000C0?ooo`030000003oool0oooo0<<0
oooo00<000000?ooo`3oool0`P3oool00`000000oooo0?ooo`0?0?ooo`004`3oool00`000000oooo
0?ooo`330?ooo`030000003oool0oooo0<80oooo00<000000?ooo`3oool03`3oool001@0oooo00<0
00000?ooo`3oool0`P3oool00`000000oooo0?ooo`310?ooo`030000003oool0oooo0100oooo000D
0?ooo`030000003oool0oooo0<80oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`0@
0?ooo`00503oool00`000000oooo0?ooo`320?ooo`030000003oool0oooo0<40oooo00<000000?oo
o`3oool0403oool001@0oooo00<000000?ooo`3oool0`P3oool00`000000oooo0?ooo`310?ooo`03
0000003oool0oooo0100oooo000E0?ooo`030000003oool0oooo0<40oooo00<000000?ooo`3oool0
`03oool00`000000oooo0?ooo`0A0?ooo`005@3oool00`000000oooo0?ooo`310?ooo`030000003o
ool0oooo0<00oooo00<000000?ooo`3oool04@3oool001D0oooo00<000000?ooo`3oool0`@3oool0
0`000000oooo0?ooo`300?ooo`030000003oool0oooo0140oooo000F0?ooo`030000003oool0oooo
0<00oooo00<000000?ooo`3oool0`03oool00`000000oooo0?ooo`0A0?ooo`005P3oool00`000000
oooo0?ooo`300?ooo`030000003oool0oooo0;l0oooo00<000000?ooo`3oool04P3oool001L0oooo
00<000000?ooo`3oool0_`3oool200000<00oooo00<000000?ooo`3oool04P3oool001L0oooo00<0
00000?ooo`3oool0_`3oool00`000000oooo0?ooo`2o0?ooo`030000003oool0oooo0180oooo000G
0?ooo`030000003oool0oooo0;l0oooo00<000000?ooo`3oool0_P3oool00`000000oooo0?ooo`0C
0?ooo`00603oool00`000000oooo0?ooo`2n0?ooo`030000003oool0oooo0;h0oooo00<000000?oo
o`3oool04`3oool001P0oooo00<000000?ooo`3oool0_P3oool00`000000oooo0?ooo`2n0?ooo`03
0000003oool0oooo01<0oooo000I0?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0
_@3oool00`000000oooo0?ooo`0D0?ooo`006@3oool00`000000oooo0?ooo`2m0?ooo`030000003o
ool0oooo0;d0oooo00<000000?ooo`3oool0503oool001T0oooo00<000000?ooo`3oool0_@3oool0
0`000000oooo0?ooo`2m0?ooo`030000003oool0oooo01@0oooo000J0?ooo`030000003oool0oooo
0;`0oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`0D0?ooo`006P3oool00`000000
oooo0?ooo`2l0?ooo`030000003oool0oooo0;`0oooo00<000000?ooo`3oool05@3oool001X0oooo
00<000000?ooo`3oool0_03oool00`000000oooo0?ooo`2l0?ooo`030000003oool0oooo01D0oooo
000K0?ooo`030000003oool0oooo0;/0oooo00<000000?ooo`3oool0_03oool00`000000oooo0?oo
o`0E0?ooo`006`3oool00`000000oooo0?ooo`2k0?ooo`030000003oool0oooo0;/0oooo00<00000
0?ooo`3oool05P3oool001`0oooo00<000000?ooo`3oool0^P3oool00`000000oooo0?ooo`2k0?oo
o`030000003oool0oooo01H0oooo000L0?ooo`030000003oool0oooo0;X0oooo00<000000?ooo`3o
ool0^P3oool00`000000oooo0?ooo`0G0?ooo`00703oool00`000000oooo0?ooo`2j0?ooo`030000
003oool0oooo0;X0oooo00<000000?ooo`3oool05`3oool001d0oooo00<000000?ooo`3oool0^@3o
ool00`000000oooo0?ooo`2i0?ooo`030000003oool0oooo01P0oooo000M0?ooo`030000003oool0
oooo0;T0oooo00<000000?ooo`3oool0^@3oool00`000000oooo0?ooo`0H0?ooo`007P3oool00`00
0000oooo0?ooo`2h0?ooo`030000003oool0oooo0;P0oooo00<000000?ooo`3oool06@3oool001h0
oooo00<000000?ooo`3oool0^03oool00`000000oooo0?ooo`2h0?ooo`030000003oool0oooo01T0
oooo000N0?ooo`030000003oool0oooo0;P0oooo0P00002h0?ooo`030000003oool0oooo01X0oooo
000O0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?oo
o`0J0?ooo`007`3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo0;H0oooo00<00000
0?ooo`3oool06`3oool00200oooo00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`2f0?oo
o`030000003oool0oooo01/0oooo000P0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3o
ool0]@3oool00`000000oooo0?ooo`0L0?ooo`008@3oool00`000000oooo0?ooo`2e0?ooo`030000
003oool0oooo0;D0oooo00<000000?ooo`3oool0703oool00240oooo00<000000?ooo`3oool0]@3o
ool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo01d0oooo000R0?ooo`030000003oool0
oooo0;@0oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`0M0?ooo`008P3oool00`00
0000oooo0?ooo`2d0?ooo`030000003oool0oooo0;<0oooo00<000000?ooo`3oool07P3oool00280
oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`2c0?ooo`030000003oool0oooo01h0
oooo000S0?ooo`030000003oool0oooo0;<0oooo00<000000?ooo`3oool0/P3oool00`000000oooo
0?ooo`0O0?ooo`008`3oool00`000000oooo0?ooo`2c0?ooo`030000003oool0oooo0;80oooo00<0
00000?ooo`3oool07`3oool002@0oooo00<000000?ooo`3oool0/P3oool00`000000oooo0?ooo`2a
0?ooo`030000003oool0oooo0200oooo000T0?ooo`030000003oool0oooo0;80oooo00<000000?oo
o`3oool0/@3oool00`000000oooo0?ooo`0P0?ooo`009@3oool00`000000oooo0?ooo`2a0?ooo`03
0000003oool0oooo0;00oooo00<000000?ooo`3oool08@3oool002D0oooo00<000000?ooo`3oool0
/@3oool00`000000oooo0?ooo`2`0?ooo`030000003oool0oooo0240oooo000V0?ooo`030000003o
ool0oooo09<0oooo00D0omnL04P000000000001d0;ooo`030?ooo`060?noM0000000M;oo0?ooo`3o
oml0W4P00`0000000`1d_ol0oooo0?ooo`090?ooo`030000003oool0oooo0:l0oooo00<000000?oo
o`3oool08P3oool002H0oooo00<000000?ooo`3oool0TP3oool01`3ookl0M00007Boo`3oool0okmd
0000M02oool00P3oool01`3o_g@0000007Boo`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo
0?ooo`090?ooo`030000003oool0oooo0:l0oooo00<000000?ooo`3oool08P3oool002L0oooo00<0
00000?ooo`3oool0T@3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo00<0
oooO09a8M02oool02@3oool00`000000oooo0?ooo`2^0?ooo`030000003oool0oooo02<0oooo000W
0?ooo`030000003oool0oooo0900oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0;
0?ooo`030;md001d_ol0oooo00P0oooo00<000000?ooo`3oool0[P3oool00`000000oooo0?ooo`0S
0?ooo`00:03oool00`000000oooo0?ooo`2?0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol02P3oool00`3ooml0W4P004RLg`090?ooo`030000003oool0oooo0:d0oooo00<000000?oo
o`3oool0903oool002T0oooo00<000000?ooo`3oool0SP3oool00`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo00X0oooo00<0gia80000001d_ol02@3oool300000:d0oooo00<000000?ooo`3o
ool0903oool002T0oooo00<000000?ooo`3oool0SP3oool00`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00T0oooo00<000000?oo
o`3oool0[03oool00`000000oooo0?ooo`0U0?ooo`00:P3oool00`000000oooo0?ooo`2=0?ooo`03
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3OW4P000000000000000000018
09cOo`X0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`0V0?ooo`00:`3oool00`00
0000oooo0?ooo`2<0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool0103o
okl0M00000000018W=l<0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool09P3oool0
02/0oooo00<000000?ooo`3oool0S@3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo
00L0oooo00<0_g@007Boo`3oool0303oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo
02L0oooo000/0?ooo`030000003oool0oooo08`0oooo00L0oono07@0001d_ol0oooo0?noM00007@0
_ooo00P0oooo00H0omnL04P00000000000000000B02Lgol90?ooo`030000003oool0oooo0:T0oooo
00<000000?ooo`3oool0:03oool002d0oooo00<000000?ooo`3oool0S03oool01@3ookl0M0000000
000007@0_ooo00X0oooo00D0_g@0000000000000000004RLg`090?ooo`030000003oool0oooo0:T0
oooo00<000000?ooo`3oool0:03oool002d0oooo00<000000?ooo`3oool0Z@3oool00`000000oooo
0?ooo`2X0?ooo`030000003oool0oooo02T0oooo000^0?ooo`030000003oool0oooo0:P0oooo00<0
00000?ooo`3oool0Z03oool00`000000oooo0?ooo`0Y0?ooo`00;`3oool00`000000oooo0?ooo`2W
0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0:P3oool002l0oooo00<000000?oo
o`3oool0Y`3oool00`000000oooo0?ooo`2V0?ooo`030000003oool0oooo02/0oooo000`0?ooo`03
0000003oool0oooo0:H0oooo00<000000?ooo`3oool0YP3oool00`000000oooo0?ooo`0[0?ooo`00
<@3oool00`000000oooo0?ooo`2U0?ooo`030000003oool0oooo0:D0oooo00<000000?ooo`3oool0
;03oool00340oooo00<000000?ooo`3oool0Y@3oool00`000000oooo0?ooo`2T0?ooo`030000003o
ool0oooo02d0oooo000b0?ooo`030000003oool0oooo0:@0oooo00<000000?ooo`3oool0X`3oool0
0`000000oooo0?ooo`0^0?ooo`00<`3oool00`000000oooo0?ooo`2S0?ooo`030000003oool0oooo
0:<0oooo00<000000?ooo`3oool0;P3oool003@0oooo00<000000?ooo`3oool0XP3oool00`000000
oooo0?ooo`2R0?ooo`030000003oool0oooo02l0oooo000d0?ooo`030000003oool0oooo0:80oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0`0?ooo`00=@3oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0:00oooo00<000000?ooo`3oool0<@3oool003H0oooo00<00000
0?ooo`3oool0X03oool00`000000oooo0?ooo`2P0?ooo`030000003oool0oooo0340oooo000g0?oo
o`030000003oool0oooo09l0oooo0P00002P0?ooo`030000003oool0oooo0380oooo000g0?ooo`03
0000003oool0oooo09l0oooo00<000000?ooo`3oool0WP3oool00`000000oooo0?ooo`0c0?ooo`00
>03oool00`000000oooo0?ooo`2N0?ooo`030000003oool0oooo09d0oooo00<000000?ooo`3oool0
=03oool003T0oooo00<000000?ooo`3oool0W@3oool00`000000oooo0?ooo`2M0?ooo`030000003o
ool0oooo03@0oooo000j0?ooo`030000003oool0oooo09`0oooo00<000000?ooo`3oool0W03oool0
0`000000oooo0?ooo`0e0?ooo`00>P3oool00`000000oooo0?ooo`2L0?ooo`030000003oool0oooo
09/0oooo00<000000?ooo`3oool0=P3oool003/0oooo00<000000?ooo`3oool0V`3oool00`000000
oooo0?ooo`2J0?ooo`030000003oool0oooo03L0oooo000l0?ooo`030000003oool0oooo09X0oooo
00<000000?ooo`3oool0VP3oool00`000000oooo0?ooo`0g0?ooo`00?@3oool00`000000oooo0?oo
o`2I0?ooo`030000003oool0oooo09T0oooo00<000000?ooo`3oool0>03oool003d0oooo00<00000
0?ooo`3oool0V@3oool00`000000oooo0?ooo`2H0?ooo`030000003oool0oooo03T0oooo000n0?oo
o`030000003oool0oooo09P0oooo00<000000?ooo`3oool0U`3oool00`000000oooo0?ooo`0j0?oo
o`00?`3oool00`000000oooo0?ooo`2G0?ooo`030000003oool0oooo09L0oooo00<000000?ooo`3o
ool0>P3oool00400oooo00<000000?ooo`3oool0UP3oool00`000000oooo0?ooo`2F0?ooo`030000
003oool0oooo03/0oooo00100?ooo`030000003oool0oooo09H0oooo00<000000?ooo`3oool0U@3o
ool00`000000oooo0?ooo`0l0?ooo`00@@3oool00`000000oooo0?ooo`2E0?ooo`030000003oool0
oooo09@0oooo00<000000?ooo`3oool0?@3oool00480oooo00<000000?ooo`3oool0U03oool00`00
0000oooo0?ooo`2C0?ooo`030000003oool0oooo03h0oooo00130?ooo`030000003oool0oooo09<0
oooo00<000000?ooo`3oool0TP3oool00`000000oooo0?ooo`0o0?ooo`00A03oool00`000000oooo
0?ooo`2B0?ooo`030000003oool0oooo0940oooo00<000000?ooo`3oool0@03oool004D0oooo00<0
00000?ooo`3oool0T@3oool00`000000oooo0?ooo`2@0?ooo`030000003oool0oooo0440oooo0016
0?ooo`030000003oool0oooo0900oooo00<000000?ooo`3oool0S`3oool00`000000oooo0?ooo`12
0?ooo`00A`3oool00`000000oooo0?ooo`2?0?ooo`800000S`3oool00`000000oooo0?ooo`130?oo
o`00B03oool00`000000oooo0?ooo`2>0?ooo`030000003oool0oooo08d0oooo00<000000?ooo`3o
ool0A03oool004T0oooo00<000000?ooo`3oool0S@3oool00`000000oooo0?ooo`2<0?ooo`030000
003oool0oooo04D0oooo001:0?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool0R`3o
ool00`000000oooo0?ooo`160?ooo`00B`3oool00`000000oooo0?ooo`2;0?ooo`030000003oool0
oooo08X0oooo00<000000?ooo`3oool0A`3oool004`0oooo00<000000?ooo`3oool0RP3oool00`00
0000oooo0?ooo`290?ooo`030000003oool0oooo04P0oooo001=0?ooo`030000003oool0oooo08T0
oooo00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`190?ooo`00CP3oool00`000000oooo
0?ooo`280?ooo`030000003oool0oooo08L0oooo00<000000?ooo`3oool0BP3oool004l0oooo00<0
00000?ooo`3oool0Q`3oool00`000000oooo0?ooo`260?ooo`030000003oool0oooo04/0oooo001@
0?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3oool0Q@3oool00`000000oooo0?ooo`1<
0?ooo`00D@3oool00`000000oooo0?ooo`250?ooo`030000003oool0oooo08@0oooo00<000000?oo
o`3oool0C@3oool00580oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`230?ooo`03
0000003oool0oooo04h0oooo001C0?ooo`030000003oool0oooo08<0oooo00<000000?ooo`3oool0
P@3oool200000540oooo001D0?ooo`800000P`3oool00`000000oooo0?ooo`200?ooo`030000003o
ool0oooo0540oooo001F0?ooo`030000003oool0oooo0800oooo00<000000?ooo`3oool0O`3oool0
0`000000oooo0?ooo`1B0?ooo`00E`3oool00`000000oooo0?ooo`1o0?ooo`030000003oool0oooo
07d0oooo0P00001E0?ooo`00F03oool2000007l0oooo00<000000?ooo`3oool0O03oool00`000000
oooo0?ooo`1E0?ooo`00FP3oool00`000000oooo0?ooo`1l0?ooo`030000003oool0oooo07/0oooo
00<000000?ooo`3oool0EP3oool005/0oooo00<000000?ooo`3oool0N`3oool00`000000oooo0?oo
o`1j0?ooo`030000003oool0oooo05L0oooo001L0?ooo`800000N`3oool00`000000oooo0?ooo`1h
0?ooo`800000FP3oool005h0oooo00<000000?ooo`3oool0N03oool00`000000oooo0?ooo`1g0?oo
o`030000003oool0oooo05X0oooo001O0?ooo`030000003oool0oooo07L0oooo0P00001g0?ooo`03
0000003oool0oooo05/0oooo001P0?ooo`800000M`3oool00`000000oooo0?ooo`1d0?ooo`800000
GP3oool00680oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`1c0?ooo`030000003o
ool0oooo05h0oooo001S0?ooo`800000M03oool00`000000oooo0?ooo`1b0?ooo`030000003oool0
oooo05l0oooo001U0?ooo`030000003oool0oooo0740oooo00<000000?ooo`3oool0L03oool20000
0680oooo001V0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0K`3oool00`000000
oooo0?ooo`1R0?ooo`00I`3oool200000700oooo00<000000?ooo`3oool0K@3oool2000006D0oooo
001Y0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0K03oool00`000000oooo0?oo
o`1U0?ooo`00JP3oool2000006d0oooo00<000000?ooo`3oool0J`3oool00`000000oooo0?ooo`1V
0?ooo`00K03oool00`000000oooo0?ooo`1Z0?ooo`030000003oool0oooo06T0oooo0P00001Y0?oo
o`00K@3oool2000006X0oooo00<000000?ooo`3oool0J03oool00`000000oooo0?ooo`1Y0?ooo`00
K`3oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo06H0oooo0P00001/0?ooo`00L03o
ool00`000000oooo0?ooo`1V0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3oool0K03o
ool00740oooo0P00001V0?ooo`030000003oool0oooo06<0oooo0P00001_0?ooo`00L`3oool20000
06@0oooo00<000000?ooo`3oool0H@3oool200000740oooo001e0?ooo`800000HP3oool00`000000
oooo0?ooo`1O0?ooo`800000L`3oool007L0oooo0P00001P0?ooo`030000003oool0oooo05d0oooo
0P00001e0?ooo`00N@3oool2000005h0oooo00<000000?ooo`3oool0F`3oool2000007L0oooo001k
0?ooo`800000G03oool00`000000oooo0?ooo`1I0?ooo`800000N@3oool007d0oooo0P00001J0?oo
o`030000003oool0oooo05L0oooo0P00001k0?ooo`00O`3oool2000005P0oooo0P00001F0?ooo`80
0000O@3oool00840oooo0P00001F0?ooo`030000003oool0oooo05<0oooo0P00001o0?ooo`00P`3o
ool3000005<0oooo00<000000?ooo`3oool0D03oool300000840oooo00260?ooo`800000D@3oool0
0`000000oooo0?ooo`1>0?ooo`800000Q03oool008P0oooo0`00001>0?ooo`030000003oool0oooo
04/0oooo0`0000260?ooo`00R`3oool2000004`0oooo00<000000?ooo`3oool0B@3oool2000008T0
oooo002=0?ooo`<00000B@3oool00`000000oooo0?ooo`160?ooo`<00000R`3oool00900oooo0`00
00160?ooo`030000003oool0oooo04<0oooo0`00002>0?ooo`00T`3oool2000004@0oooo00<00000
0?ooo`3oool0@@3oool200000940oooo002E0?ooo`<00000@@3oool00`000000oooo0?ooo`0n0?oo
o`<00000T`3oool009P0oooo0P00000o0?ooo`030000003oool0oooo03X0oooo1000002F0?ooo`00
VP3oool3000003`0oooo00<000000?ooo`3oool0=`3oool3000009X0oooo002M0?ooo`@00000>03o
ool00`000000oooo0?ooo`0d0?ooo`<00000W@3oool00:40oooo1000000d0?ooo`030000003oool0
oooo0300oooo1000002P0?ooo`00Y@3oool400000300oooo00<000000?ooo`3oool0;@3oool30000
0:@0oooo002Y0?ooo`@00000;03oool00`000000oooo0?ooo`0Y0?ooo`@00000Y`3oool00:d0oooo
1000000H0?ooo`060;md00000000000000000000B9`0gooo2P3oool00`000000oooo0?ooo`0U0?oo
o`@00000Z`3oool00;40oooo1000000E0?ooo`030?oOW0180000M;oo00`0oooo00<000000?ooo`3o
ool08@3oool400000:l0oooo002e0?ooo`L000003P3oool00`3ogi`0B00007Boo`0<0?ooo`030000
003oool0oooo01/0oooo1P00002c0?ooo`00_03oool8000000H0oooo00<0omnL04P0001d_ol0303o
ool00`000000oooo0?ooo`0@0?ooo`/00000^@3oool00<@0oooo4@0000040?ooo`040000003oool0
oooo0?ooo`l00000a03oool00<X0oooo00<0omnL04P0001d_ol0203oool800000=<0oooo003:0?oo
o`030?oOW0180000M;oo00`0oooo00<000000?ooo`3oool0e03oool00<X0oooo00<0omnL04P0001d
_ol0303oool00`000000oooo0?ooo`3D0?ooo`00bP3oool00`3ogi`0B00007Boo`0<0?ooo`030000
003oool0oooo0=@0oooo003:0?ooo`030?oOW0180000M;oo00`0oooo00<000000?ooo`3oool0e03o
ool00<P0oooo00D0oooO09a800000000000007Boo`0<0?ooo`030000003oool0oooo0=@0oooo003:
0?ooo`030?noM0000000M;oo00`0oooo00<000000?ooo`3oool0e03oool00=T0oooo00<000000?oo
o`3oool0e03oool00=T0oooo00<000000?ooo`3oool0e03oool00=T0oooo00<000000?ooo`3oool0
e03oool00=T0oooo00<000000?ooo`3oool0e03oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-1.06748, -1.05875, 0.00736947, \
0.00736947}}]
}, Closed]],

Cell["\<\
Some nerd decided that aesthetically pleasing figures would be generated \
using a default aspect ratio equal to the reciprocal of the so-called golden \
ratio, and some choice had to be made, but you should not feel obligated to \
share his/her sensibilities!\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?GoldenRatio\)\)], "Input"],

Cell[BoxData[
    RowBox[{"\<\"GoldenRatio is the golden ratio (1 + Sqrt[5])/2, with \
numerical value approximately equal to 1.61803.\"\>", " ", 
      ButtonBox[
        StyleBox["More\[Ellipsis]",
          "SR"],
        ButtonData:>"GoldenRatio",
        Active->True,
        ButtonStyle->"RefGuideLink"]}]], "Print",
  CellTags->"Info3350813126-5278694"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Line styles", "Subsubsection"],

Cell[TextData[{
  "When plotting several curves together, it is often useful to distinguish \
them either by dashing pattern, color, or thickness and options are provided \
to control any of these attributes.  Your choices should be bundled together \
in the replacement list for the ",
  ButtonBox["PlotStyle",
    ButtonStyle->"RefGuideLink"],
  " option.  The basic syntax is ",
  StyleBox["PlotStyle", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\[Rule]{", "InlineInput"],
  StyleBox["style", "InlineInput"],
  StyleBox["}", "InlineInput"],
  " or ",
  StyleBox["PlotStyle", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\[Rule]{{", "InlineInput"],
  StyleBox["style1", "InlineInput"],
  StyleBox["},{", "InlineInput"],
  StyleBox["style2", "InlineInput"],
  StyleBox["},...}", "InlineInput"],
  " where each ",
  StyleBox["{", "InlineInput"],
  StyleBox["style", "InlineInput"],
  StyleBox["}", "InlineInput"],
  StyleBox[" ",
    FontSlant->"Italic"],
  "is a list of graphics directives, including ",
  StyleBox["Dashing", "InlineInput",
    FontWeight->"Bold"],
  ", ",
  StyleBox["RGBColor", "InlineInput",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Hue", "InlineInput",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Thickness", "InlineInput",
    FontWeight->"Bold"],
  ", among others.  If a figure contains many lines, the styles are applied \
cyclically.  Because style specifications tend to be verbose and may be used \
for several related figures, it is convenient to collect them into list \
symbols.  Several examples follow."
}], "Text"],

Cell[BoxData[{
    \(\(line[1]\  = \ {Dashing[{0.02, 0.02}], \ RGBColor[1, 0, 0], \ 
          Thickness[0.01]};\)\), "\n", 
    \(\(line[2]\  = \ {Dashing[{0.02, 0.02}], \ RGBColor[0, 1, 0], \ 
          Thickness[0.01]};\)\), "\n", 
    \(\(line[3]\  = \ {Dashing[{0.02, 0.02}], \ RGBColor[0, 0, 1], \ 
          Thickness[0.01]};\)\)}], "Input"],

Cell[BoxData[
    \(\(lines\  = \ Table[line[n], {n, 3}];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[\ Evaluate[\ Table[LegendreP[n, x], {n, 0, 4}]], {x, \(-1\), 1}, 
        PlotStyle \[Rule] lines];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.309017 0.294302 [
[.02381 .29652 -9.15625 -15.6875 ]
[.02381 .29652 9.15625 0 ]
[.2619 .29652 -13.6563 -15.6875 ]
[.2619 .29652 13.6563 0 ]
[.7381 .29652 -9.5 -15.6875 ]
[.7381 .29652 9.5 0 ]
[.97619 .29652 -5 -15.6875 ]
[.97619 .29652 5 0 ]
[.4875 .01472 -18.3125 -7.84375 ]
[.4875 .01472 0 7.84375 ]
[.4875 .16187 -27.3125 -7.84375 ]
[.4875 .16187 0 7.84375 ]
[.4875 .45617 -19 -7.84375 ]
[.4875 .45617 0 7.84375 ]
[.4875 .60332 -10 -7.84375 ]
[.4875 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 .30902 m
.02381 .31527 L
s
gsave
.02381 .29652 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 .30902 m
.2619 .31527 L
s
gsave
.2619 .29652 -74.6563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.7381 .30902 m
.7381 .31527 L
s
gsave
.7381 .29652 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 .30902 m
.97619 .31527 L
s
gsave
.97619 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .30902 m
.07143 .31277 L
s
.11905 .30902 m
.11905 .31277 L
s
.16667 .30902 m
.16667 .31277 L
s
.21429 .30902 m
.21429 .31277 L
s
.30952 .30902 m
.30952 .31277 L
s
.35714 .30902 m
.35714 .31277 L
s
.40476 .30902 m
.40476 .31277 L
s
.45238 .30902 m
.45238 .31277 L
s
.54762 .30902 m
.54762 .31277 L
s
.59524 .30902 m
.59524 .31277 L
s
.64286 .30902 m
.64286 .31277 L
s
.69048 .30902 m
.69048 .31277 L
s
.78571 .30902 m
.78571 .31277 L
s
.83333 .30902 m
.83333 .31277 L
s
.88095 .30902 m
.88095 .31277 L
s
.92857 .30902 m
.92857 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5 .01472 m
.50625 .01472 L
s
gsave
.4875 .01472 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .16187 m
.50625 .16187 L
s
gsave
.4875 .16187 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .45617 m
.50625 .45617 L
s
gsave
.4875 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .60332 m
.50625 .60332 L
s
gsave
.4875 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .04415 m
.50375 .04415 L
s
.5 .07358 m
.50375 .07358 L
s
.5 .10301 m
.50375 .10301 L
s
.5 .13244 m
.50375 .13244 L
s
.5 .1913 m
.50375 .1913 L
s
.5 .22073 m
.50375 .22073 L
s
.5 .25016 m
.50375 .25016 L
s
.5 .27959 m
.50375 .27959 L
s
.5 .33845 m
.50375 .33845 L
s
.5 .36788 m
.50375 .36788 L
s
.5 .39731 m
.50375 .39731 L
s
.5 .42674 m
.50375 .42674 L
s
.5 .4856 m
.50375 .4856 L
s
.5 .51503 m
.50375 .51503 L
s
.5 .54446 m
.50375 .54446 L
s
.5 .57389 m
.50375 .57389 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
1 0 0 r
.01 w
[ .02 .02 ] 0 setdash
.02381 .60332 m
.06244 .60332 L
.10458 .60332 L
.14415 .60332 L
.18221 .60332 L
.22272 .60332 L
.26171 .60332 L
.30316 .60332 L
.34309 .60332 L
.3815 .60332 L
.42237 .60332 L
.46172 .60332 L
.49955 .60332 L
.53984 .60332 L
.57861 .60332 L
.61984 .60332 L
.65954 .60332 L
.69774 .60332 L
.73838 .60332 L
.77751 .60332 L
.81909 .60332 L
.85916 .60332 L
.89771 .60332 L
.93871 .60332 L
.97619 .60332 L
s
0 1 0 r
.02381 .01472 m
.06244 .03859 L
.10458 .06463 L
.14415 .08909 L
.18221 .11261 L
.22272 .13765 L
.26171 .16175 L
.30316 .18736 L
.34309 .21204 L
.3815 .23578 L
.42237 .26104 L
.46172 .28536 L
.49955 .30874 L
.53984 .33364 L
.57861 .3576 L
.61984 .38308 L
.65954 .40762 L
.69774 .43123 L
.73838 .45635 L
.77751 .48053 L
.81909 .50623 L
.85916 .53099 L
.89771 .55482 L
.93871 .58016 L
.97619 .60332 L
s
0 0 1 r
.02381 .60332 m
.06244 .53459 L
.10458 .46626 L
.14415 .40839 L
.18221 .35848 L
.22272 .31155 L
.26171 .27241 L
.30316 .2373 L
.34309 .2098 L
.3815 .1892 L
.40095 .18097 L
.42237 .1736 L
.44268 .16826 L
.45178 .16639 L
.46172 .16472 L
.4671 .16397 L
.4721 .16338 L
.47727 .16287 L
.48196 .1625 L
.48658 .16222 L
.4887 .16211 L
.49093 .16203 L
.49332 .16195 L
.49438 .16193 L
.49552 .16191 L
.49675 .16189 L
.49789 .16187 L
.49859 .16187 L
.49925 .16187 L
.50049 .16187 L
.50163 .16187 L
.50286 .16188 L
.50401 .1619 L
.50508 .16192 L
.50754 .16198 L
.51014 .16207 L
.51268 .16218 L
.51504 .16231 L
.5204 .16268 L
.5293 .16354 L
.53882 .1648 L
.54906 .16655 L
.56016 .16891 L
.58032 .17442 L
.60019 .18141 L
.62123 .19048 L
.65912 .21116 L
.69946 .23932 L
.73829 .27241 L
.77956 .31402 L
Mistroke
.81932 .36038 L
.85757 .41078 L
.89827 .47066 L
.93745 .53441 L
.97619 .60332 L
Mfstroke
1 0 0 r
.02381 .01472 m
.06244 .14385 L
.10458 .25431 L
.14415 .33187 L
.18221 .38494 L
.20178 .40476 L
.22272 .42081 L
.23293 .42681 L
.24402 .43203 L
.25449 .43579 L
.26416 .43827 L
.26882 .43915 L
.27139 .43954 L
.27383 .43986 L
.27609 .4401 L
.27856 .44032 L
.27965 .44039 L
.2808 .44046 L
.28182 .44051 L
.28289 .44056 L
.28355 .44058 L
.28428 .4406 L
.28556 .44062 L
.28629 .44063 L
.28706 .44063 L
.28779 .44063 L
.28846 .44062 L
.28975 .4406 L
.29048 .44058 L
.29115 .44056 L
.29247 .44051 L
.29371 .44044 L
.29597 .44029 L
.29846 .44008 L
.30351 .43948 L
.30821 .43875 L
.31334 .43774 L
.3237 .43512 L
.3422 .42853 L
.36181 .41914 L
.38276 .40672 L
.42426 .37627 L
.46424 .34186 L
.50271 .30651 L
.54363 .26914 L
.58303 .23594 L
.62092 .20897 L
.642 .19689 L
.66126 .1881 L
.67079 .18463 L
Mistroke
.68117 .18158 L
.68592 .18045 L
.69098 .17943 L
.69541 .1787 L
.70008 .17811 L
.70264 .17786 L
.704 .17775 L
.70545 .17764 L
.70679 .17757 L
.70801 .17751 L
.70923 .17746 L
.71039 .17743 L
.71103 .17742 L
.71174 .17741 L
.71248 .1774 L
.71317 .1774 L
.7144 .17741 L
.7151 .17742 L
.71573 .17743 L
.71708 .17748 L
.71856 .17754 L
.7201 .17763 L
.72153 .17773 L
.72417 .17796 L
.72665 .17823 L
.7313 .17891 L
.7363 .17986 L
.74171 .18116 L
.7512 .18415 L
.76163 .1885 L
.78269 .20088 L
.80217 .21689 L
.82308 .23929 L
.86441 .30092 L
.90422 .38432 L
.94252 .48923 L
.97619 .60332 L
Mfstroke
0 1 0 r
.02381 .60332 m
.04262 .4971 L
.06244 .40543 L
.08255 .33166 L
.10458 .27057 L
.12415 .23154 L
.13511 .21527 L
.14509 .20363 L
.15504 .1948 L
.16424 .18895 L
.16904 .18671 L
.17432 .18486 L
.17698 .18417 L
.17985 .1836 L
.18123 .18338 L
.18255 .18321 L
.18373 .18309 L
.18502 .18299 L
.18566 .18295 L
.18635 .18292 L
.18757 .18289 L
.18889 .18289 L
.18963 .18291 L
.19033 .18293 L
.19159 .18299 L
.19275 .18308 L
.19408 .18321 L
.19534 .18336 L
.19766 .18372 L
.19982 .18414 L
.20469 .18538 L
.20953 .18699 L
.21486 .1892 L
.22566 .19492 L
.24501 .20879 L
.26587 .22783 L
.30701 .27284 L
.34663 .31875 L
.38474 .35914 L
.40599 .37832 L
.4253 .393 L
.44538 .40509 L
.45441 .40937 L
.46435 .41323 L
.47407 .41612 L
.47938 .41732 L
.48436 .41819 L
.48674 .41853 L
.48901 .41879 L
.49106 .41899 L
Mistroke
.49326 .41916 L
.4945 .41923 L
.49563 .41929 L
.49676 .41933 L
.49783 .41936 L
.49908 .41938 L
.50022 .41938 L
.50148 .41937 L
.50212 .41936 L
.50281 .41934 L
.50412 .4193 L
.50533 .41924 L
.50807 .41906 L
.51048 .41885 L
.51304 .41855 L
.51747 .4179 L
.52236 .41695 L
.53316 .41406 L
.54329 .41035 L
.56351 .40016 L
.58203 .38777 L
.6217 .35278 L
.65986 .31135 L
.70048 .26422 L
.73958 .22252 L
.75953 .20517 L
.77003 .19764 L
.78113 .19114 L
.79106 .18678 L
.7967 .18499 L
.7994 .18431 L
.80197 .18379 L
.80429 .18342 L
.80551 .18326 L
.80681 .18312 L
.8079 .18303 L
.8091 .18295 L
.8102 .18291 L
.81123 .18289 L
.81244 .18289 L
.81372 .18293 L
.81481 .18298 L
.816 .18307 L
.81729 .18319 L
.81867 .18337 L
.82116 .18378 L
.82374 .18434 L
.82648 .1851 L
.83139 .18689 L
.83667 .18943 L
Mistroke
.84248 .19302 L
.85288 .20161 L
.86263 .21237 L
.87302 .22697 L
.88289 .24405 L
.90142 .28531 L
.92124 .34419 L
.93963 .41411 L
.97619 .60332 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00f@3oool0
0`000000oooo0?ooo`3D0?ooo`00bP3oool01P2oM000000000000000000004RL0=ooo`T0oooo00<0
00000?ooo`3oool0e03oool00</0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`3D
0?ooo`00b`3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0=@0oooo000D0?ooo`80
0?l0]@3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0=@0oooo000D0?ooo`<00?l0
Y03oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00<0oooo00<0omnL04P0001d_ol0
2`3oool00`000000oooo0?ooo`3D0?ooo`004`3oool2003o0080o`000P00o`2b0?ooo`030?oOW018
0000M;oo00/0oooo1000003C0?ooo`004`3oool50?l000800?l0/@3oool00`3ogi`0B00007Boo`0;
0?ooo`030000003oool0oooo0=@0oooo000C0?ooo`D0o`000`00o`2`0?ooo`030?oOW0180000M;oo
00/0oooo00<000000?ooo`3oool0e03oool001@0oooo103o0005003o0:h0oooo00<0omnL04P0001d
_ol02`3oool00`000000oooo0?ooo`3D0?ooo`00503oool50?l000D00?l0[@3oool00`3ogi`0B000
07Boo`0;0?ooo`030000003oool0oooo0=@0oooo000D0?ooo`D0o`001P00o`2Z0?ooo`050?oog`2L
B00000000000001d_ol02`3oool00`000000oooo0?ooo`3D0?ooo`005@3oool40?l00080oooo0`00
o`2]0?ooo`030?noM0000000M;oo00/0oooo00<000000?ooo`3oool0e03oool001D0oooo1@3o0002
0?ooo`800?l0103oool2003o0;D0oooo00<000000?ooo`3oool0e03oool001D0oooo1@3o00080?oo
o`D00?l0/P3oool00`000000oooo0?ooo`3D0?ooo`005P3oool40?l000L0oooo1`00o`2a0?ooo`03
0000003oool0oooo0=@0oooo000F0?ooo`D0o`001P3oool8003o0;00oooo00<000000?ooo`3oool0
e03oool001H0oooo1@3o00070?ooo`T00?l0[P3oool00`000000oooo0?ooo`3D0?ooo`005`3oool2
0?l000`0oooo1`00o`2]0?ooo`030000003oool0oooo0=@0oooo000W0?ooo`H00?l0[03oool30000
0=@0oooo000Y0?ooo`<00?l0[@3oool00`000000oooo0?ooo`3D0?ooo`00:P3oool2003o0:d0oooo
00<000000?ooo`3oool0e03oool001X0oooo0P3o000D0?ooo`800?l0Y`3oool00`000000oooo0?oo
o`3D0?ooo`00603oool40?l001@0oooo0`00o`2V0?ooo`030000003oool0oooo0=@0oooo000H0?oo
o`D0o`004P3oool6003o0:@0oooo00<000000?ooo`3oool0e03oool001T0oooo103o000C0?ooo`L0
0?l0XP3oool00`000000oooo0?ooo`3D0?ooo`006@3oool40?l001@0oooo2@00o`2O0?ooo`030000
003oool0oooo0=@0oooo000I0?ooo`@0o`005P3oool8003o09h0oooo00<000000?ooo`3oool0e03o
ool001T0oooo1@3o000F0?ooo`L00?l0WP3oool00`000000oooo0?ooo`3D0?ooo`006P3oool40?l0
01L0oooo1@00o`2O0?ooo`030000003oool0oooo0=@0oooo000J0?ooo`@0o`006P3oool2003o00@0
oooo0P00o`2I0?ooo`030000003oool0oooo0=@0oooo000J0?ooo`@0o`00803oool3003o09P0oooo
0`00003D0?ooo`006P3oool50?l001h0oooo1P00o`2F0?ooo`030000003oool0oooo0=@0oooo000K
0?ooo`@0o`007`3oool6003o09D0oooo00<000000?ooo`3oool0e03oool001/0oooo0P3o000R0?oo
o`H00?l0U03oool00`000000oooo0?ooo`3D0?ooo`00@@3oool6003o0980oooo00<000000?ooo`3o
ool0e03oool00480oooo1P00o`2A0?ooo`030000003oool0oooo0=@0oooo000O0?ooo`80o`008P3o
ool6003o0900oooo00<000000?ooo`3oool0e03oool001d0oooo103o000T0?ooo`<00?l0T@3oool0
0`000000oooo0?ooo`3D0?ooo`007@3oool50?l002@0oooo0P00o`040?ooo`800?l0R`3oool00`00
0000oooo0?ooo`3D0?ooo`007P3oool40?l002X0oooo0`00o`2:0?ooo`030000003oool0oooo0=@0
oooo000N0?ooo`@0o`00:@3oool6003o08P0oooo00<000000?ooo`3oool0e03oool001h0oooo103o
000Z0?ooo`L00?l0QP3oool00`000000oooo0?ooo`3D0?ooo`007P3oool50?l002X0oooo2000o`24
0?ooo`<00000e03oool001l0oooo103o000/0?ooo`L00?l0P`3oool00`000000oooo0?ooo`3D0?oo
o`007`3oool40?l002h0oooo1P00o`220?ooo`030000003oool0oooo0=@0oooo000O0?ooo`@0o`00
<03oool3003o08<0oooo00<000000?ooo`3oool0e03oool001l0oooo1@3o000`0?ooo`800?l0P`3o
ool00`000000oooo0?ooo`3D0?ooo`00803oool40?l003H0oooo0P00o`1m0?ooo`030000003oool0
oooo0=@0oooo000P0?ooo`80o`00>03oool3003o07`0oooo00<000000?ooo`3oool0e03oool005T0
oooo1P00o`1j0?ooo`030000003oool0oooo0=@0oooo001J0?ooo`L00?l0N03oool00`000000oooo
0?ooo`3D0?ooo`00F`3oool8003o07H0oooo00<000000?ooo`3oool0e03oool002@0oooo0P3o000g
0?ooo`L00?l0M@3oool00`000000oooo0?ooo`3D0?ooo`008P3oool40?l003T0oooo1P00o`1d0?oo
o`030000003oool0oooo0=@0oooo000R0?ooo`D0o`00>P3oool3003o07D0oooo0`00003D0?ooo`00
8`3oool40?l003/0oooo0P00o`040?ooo`800?l0K`3oool00`000000oooo0?ooo`3D0?ooo`008`3o
ool40?l00440oooo0`00o`1^0?ooo`030000003oool0oooo0=@0oooo000S0?ooo`@0o`00@03oool6
003o06`0oooo00<000000?ooo`3oool0e03oool002<0oooo103o00110?ooo`H00?l0J`3oool00`00
0000oooo0?ooo`3D0?ooo`008`3oool50?l00440oooo1P00o`1Z0?ooo`030000003oool0oooo0=@0
oooo000T0?ooo`@0o`00@`3oool6003o06P0oooo00<000000?ooo`3oool0e03oool002@0oooo103o
00140?ooo`H00?l0BP3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00H0okmd0000001d_ol0
oooo0?oog`2LB0030000000307Boo`3oool0oooo00T0oooo00<000000?ooo`3oool0e03oool002@0
oooo1@3o00140?ooo`H00?l0B03oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool0
1`3o_g@0000007Boo`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`090?ooo`030000
003oool0oooo0=@0oooo000U0?ooo`@0o`00AP3oool3003o04T0oooo00<0_g@00018W03Oool00P3o
ool00`2oM00007Bo0?ooo`0:0?ooo`030?oog`2LB7@0_ooo00T0oooo00<000000?ooo`3oool0e03o
ool002D0oooo0P3o00190?ooo`800?l0103oool2003o0480oooo00<0oono07@00018W=l00`3oool0
0`3OW4P0001d0;ooo`0;0?ooo`030;md00000?l0003o00P0oooo00<000000?ooo`3oool0e03oool0
07H0oooo0`00o`0e0?ooo`030?oog`2LB000000000L0000000D0B9cO0?ooo`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo00H0oooo0P000?l00`3oool0003o0000o`030000o`@0oooo3@000?l6
0?ooo`80003oaP3oool007D0oooo1P00o`0o0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol00`3oool<0000o`<0oooo3@000?l40?ooo`P0003o`P3oool007H0oooo1`00o`0m0?ooo`03
0?oOW0180000M;oo0080oooo0P000?l0100004P0W=oo0?ooo`3oool=0000o`<0oooo3@000?l40?oo
o``0003o_P3oool002T0oooo0P3o001<0?ooo`P00?l0>`3oool00`3ogi`0B0000000o`040000o`04
0000B02Lgol0oooo0?ooo``0003o103oool=0000o`@0oooo30000?nn0?ooo`009`3oool40?l004h0
oooo1`00o`0i0?ooo`T0003o00@0_ooo0?ooo`3oool0oooo2@000?l:0?ooo`030000003oool0oooo
00X0oooo30000?l70?ooo`80003o]@3oool002L0oooo1@3o001?0?ooo`H00?l0=@3oool<0000o`@0
oooo0P000?l20?ooo`030;md001d_ol0oooo00`0oooo00<000000?ooo`3oool03`3oool70000o`@0
oooo20000?nb0?ooo`009`3oool50?l00540oooo0`00o`0e0?ooo`/0003o00<0001d0;ooo`3oool0
1`3oool01P3ogi`0B000000000000000001809cOo`T0oooo00<000000?ooo`3oool06P3oool;0000
obL0oooo0P3o00040?ooo`P0o`00NP3oool002P0oooo1@3o001A0?ooo`800?l0;P3oool30000o`@0
oooo20000?l0101d000000000000M02oool:0?ooo`050;md00000000000000000018W=l02@3oool0
0`000000oooo0?ooo`0J0?ooo`d0003o8@3oool70?l000<0oooo3@3o001e0?ooo`00:03oool50?l0
02D0oooo0P00o`0`0?ooo`800?l09P3oool50000o`D0oooo1@000?lN0?ooo`030000003oool0oooo
01X0oooo3@000?lN0?ooo`X0o`000`3oool=0?l000D0oooo0P3o000A0?ooo`800?l02@3oool00`00
o`00oooo0?ooo`1?0?ooo`00:@3oool50?l002<0oooo3000o`0W0?ooo`D00?l08@3oool80000o`@0
oooo0P000?lQ0?ooo`030000003oool0oooo0240oooo1@000?l40?ooo`80003o5`3oool<0?l000<0
oooo3@3o00050?ooo`@0o`00303oool6003o00@0oooo2000o`1>0?ooo`00:@3oool50?l00280oooo
3@00o`070?ooo`03003o003oool0oooo01`0oooo2000o`0M0?ooo`X0003o9`3oool00`000000oooo
0?ooo`0T0?ooo`80003o103oool40000o`l0oooo0`3o00030?ooo``0o`001`3oool90?l000@0oooo
203o00070?ooo`P00?l0103oool;003o04/0oooo000Z0?ooo`D0o`008@3oool=003o00@0oooo1P00
o`0L0?ooo`T00?l06P3oool:0000obT0oooo00<000000?ooo`3oool0:@3oool80000o`X0oooo1@3o
00040?ooo`P0o`005`3oool90?l000@0oooo2P00o`040?ooo`d00?l0B@3oool002X0oooo1@3o000R
0?ooo``00?l0103oool7003o01d0oooo2000o`0I0?ooo`P0003o:`3oool00`000000oooo0?ooo`0Y
0?ooo`T0003o1`3oool80?l000<0oooo0P3o000N0?ooo`T0o`0000<0oooo003o0000o`002P00o`04
0?ooo`d00?l0B@3oool002/0oooo103o000J0?ooo`@00?l01@3oool00`00o`00oooo0?ooo`0<0?oo
o`P00?l07`3oool7003o01L0oooo1P000?l]0?ooo`<00000:P3oool;0000o`80oooo2P3o000U0?oo
o`P0o`002@00o`0:0?ooo`T00?l0BP3oool002/0oooo0P3o000L0?ooo`@00?l0503oool;003o01d0
oooo1P00o`0A0?ooo`80003o103oool40000obl0oooo00<000000?ooo`3oool0;03oool90000o`X0
o`00:P3oool50?l000H00?l0503oool2003o04X0oooo00180?ooo`H00?l05`3oool8003o01d0oooo
1000o`0A0?ooo`<0003o103oool40000obl0oooo00<000000?ooo`3oool0;P3oool70000o`P0o`00
;P3oool30?l000D00?l06`3oool2003o04@0oooo00160?ooo`P00?l0603oool7003o01l0oooo00<0
0?l00?ooo`3oool0103oool00`00o`00oooo0?ooo`070?ooo`H0003o=P3oool00`000000oooo0?oo
o`0`0?ooo`D0003o1P3o000]0?ooo`800?l00P3oool20?l001d0oooo1@00o`140?ooo`00A@3oool8
003o01X0oooo1@00o`0V0?ooo`@00?l01@3oool70000ocL0oooo00<000000?ooo`3oool0<`3oool2
0000o`@0o`000P000?l/0?ooo`@00?l0203oool00`3o0000oooo0?ooo`0E0?ooo`H00?l0@`3oool0
02l0oooo0P3o000C0?ooo`L00?l07`3oool2003o02D0oooo1P00o`020?ooo`P0003o>03oool00`00
0000oooo0?ooo`0_0?ooo`80o`00103oool40?l000@0003o:@3oool6003o00H0oooo103o000E0?oo
o`L00?l0@P3oool002d0oooo1@3o000A0?ooo`L00?l09P3oool00`00o`00oooo0?ooo`0N0?ooo`L0
0?l01`000?lj0?ooo`030000003oool0oooo02d0oooo1@3o00060?ooo`P0003o9@3oool6003o00H0
oooo1P3o000D0?ooo`P00?l0@@3oool002d0oooo1@3o000A0?ooo`D00?l09`3oool4003o0200oooo
1000o`060000oc`0oooo00<000000?ooo`3oool0;03oool60?l000H0oooo2@000?lS0?ooo`L00?l0
1P3oool70?l001H0oooo1P00o`100?ooo`00;P3oool40?l00180oooo0`00o`0W0?ooo`H00?l0803o
ool4003o00<0003o00<00?l00?ooo`3oool0>`3oool00`000000oooo0?ooo`0[0?ooo`L0o`001`3o
ool90000ob40oooo1P00o`0:0?ooo`H0o`005P3oool5003o0400oooo000^0?ooo`D0o`004P3oool0
0`00o`00oooo0?ooo`0V0?ooo`L00?l0803oool3003o0080003o0`00o`0l0?ooo`030000003oool0
oooo02T0oooo2@3o00090?ooo`P0003o7P3oool7003o00`0oooo1P3o000F0?ooo`D00?l0?`3oool0
02h0oooo1@3o000m0?ooo`H00?l0703oool010000?l0oooo0?ooo`3oool7003o03`0oooo00<00000
0?ooo`3oool0:03oool60?l00100oooo1P000?lM0?ooo`H00?l03P3oool70?l001@0oooo1P00o`0n
0?ooo`00;`3oool40?l003h0oooo1P00o`0I0?ooo`@0003o0`3oool5003o03d0oooo00<000000?oo
o`3oool09`3oool60?l001<0oooo0`000?lO0?ooo`@00?l0403oool60?l001D0oooo1000o`0o0?oo
o`00;`3oool50?l000`0oooo00<00?l00?ooo`3oool0;`3oool7003o01H0oooo1P000?l50?ooo`80
0?l0103oool2003o03L0oooo0`00000X0?ooo`<0o`005P3oool20000ob00oooo00<00?l00?ooo`3o
ool04@3oool50?l001H0oooo0P00o`100?ooo`00;`3oool50?l000/0oooo1000o`0`0?ooo`H00?l0
5@3oool70000o`/0oooo0`00o`0f0?ooo`030000003oool0oooo02P0oooo0P3o000M0?ooo`80003o
;P3oool50?l005P0oooo000`0?ooo`@0o`002P3oool5003o0340oooo1000o`0E0?ooo`H0003o303o
ool6003o03@0oooo00<000000?ooo`3oool08P3oool20?l002<0oooo0`000?lB0?ooo`800?l0MP3o
ool00300oooo1@3o00080?ooo`D00?l0=03oool00`00o`00oooo0?ooo`0C0?ooo`H0003o3P3oool7
003o0380oooo00<000000?ooo`3oool08@3oool30?l00280oooo1P000?l@0?ooo`<00?l0803oool0
0`3o0000oooo0?ooo`0F0?ooo`03003o003oool0oooo03T0oooo000`0?ooo`D0o`00203oool4003o
04T0oooo1`000?l@0?ooo`P00?l0<03oool00`000000oooo0?ooo`0N0?ooo`L0o`008P3oool60000
o`h0oooo1@00o`0N0?ooo`<0o`005@3oool4003o03X0oooo000a0?ooo`80o`002@3oool5003o03/0
oooo00<00?l00?ooo`3oool02`3oool60000oa<0oooo1`00o`0_0?ooo`030000003oool0oooo01d0
oooo1`3o000T0?ooo`L0003o2`3oool5003o01h0oooo1@3o000D0?ooo`D00?l0>@3oool003`0oooo
1000o`0k0?ooo`<00?l03@3oool40000oaH0oooo1P00o`0^0?ooo`030000003oool0oooo01`0oooo
1`3o000W0?ooo`H0003o2@3oool6003o01h0oooo1P3o000D0?ooo`@00?l0>@3oool003`0oooo0`00
o`0k0?ooo`D00?l03@3oool00`000?l0oooo0?ooo`0H0?ooo`<00?l0;`3oool00`000000oooo0?oo
o`0K0?ooo`H0o`00:P3oool60000o`L0oooo1P00o`0P0?ooo`H0o`00503oool4003o03P0oooo000k
0?ooo`@00?l0>`3oool6003o00D0oooo00<0003o0?ooo`3oool0803oool2003o02l0oooo00<00000
0?ooo`3oool06P3oool60?l002d0oooo0`000?l70?ooo`H00?l08P3oool60?l001<0oooo1@00o`0g
0?ooo`00=@3oool20?l000@0oooo1000o`0l0?ooo`H00?l00P3oool40000obL0oooo0P00o`0Y0?oo
o`030000003oool0oooo01X0oooo1@3o000_0?ooo`80003o1`3oool5003o02@0oooo1@3o000D0?oo
o`@00?l0=`3oool003<0oooo103o00030?ooo`@00?l0?P3oool6003o00H0003o9P3oool3003o02P0
oooo00<000000?ooo`3oool06`3oool30?l003P0oooo1@00o`0V0?ooo`D0o`004`3oool5003o03H0
oooo000c0?ooo`D0o`00103oool2003o03l0oooo1P00o`050000obD0oooo1P00o`0V0?ooo`030000
003oool0oooo01`0oooo00<0o`000?ooo`3oool0=P3oool00`000?l0oooo003o0002003o02P0oooo
1@3o000C0?ooo`@00?l0=P3oool003@0oooo103o00160?ooo`H00?l00P000?lX0?ooo`L00?l0903o
ool3000005@0oooo10000?l2003o02T0oooo103o000C0?ooo`@00?l0=P3oool003@0oooo103o0017
0?ooo`H00?l0:P3oool8003o0280oooo00<000000?ooo`3oool05P3oool20?l003/0oooo1P000?l[
0?ooo`030?l0003oool0oooo01@0oooo1000o`0e0?ooo`00=03oool30?l000800?l0A03oool30000
o`@00?l0;@3oool7003o0240oooo00<000000?ooo`3oool05@3oool30?l003T0oooo0P00o`070000
od40oooo0P00o`0g0?ooo`006@3oool01P2oM000000000000000000004RL0=oooaH0oooo00<0o`00
003o0000o`000`00o`0i0?ooo`050?oOW018000000000000M02oool00`3oool00`3o_g@0003o0000
o`020000o`800?l00P0000000`1d_ol0oooo0?ooo`0[0?ooo`H00?l0803oool00`000000oooo0?oo
o`0C0?ooo`H0o`00=`3oool4003o00030?ooo`000?l0003o00@0003o0`3oool01P3o_g@0000007Bo
o`3oool0oooO09a800<0000000<0M;oo0?ooo`3oool0E@3oool01P2oM000000000000000000004RL
0=ooo`l0oooo000J0?ooo`030?oOW0180000M;oo01P0oooo00<0o`00003o0000o`000`00o`0h0?oo
o`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000003o00<0003o00D0
_ooo0?ooo`3oool0_g@007Boo`0^0?ooo`<00?l08@3oool00`000000oooo0?ooo`0B0?ooo`H0o`00
=`3oool6003o00030?oo_`000?l0003o00@0003o0P3oool01`3o_g@0000007Boo`3oool0oooO09a8
M02oool00P3oool00`2oM000M;oo0?ooo`0O0?ooo`030?l0003oool0oooo03@0oooo00<0omnL04P0
001d_ol04@3oool001X0oooo00<0omnL04P0001d_ol06@3oool4003o03T0oooo00<0_g@00018W03O
ool00P3oool00`2oM00007Bo0?ooo`040?ooo`030000o`3oool0oooo00<0oooo00<0oooO09a8M000
o`00;P3oool2003o00@0oooo0P00o`0K0?ooo`030000003oool0oooo0140oooo1P3o000g0?ooo`H0
0?l000<0oooo0;md0000B9`01`000?l90?ooo`030?oog`2LB7@0_ooo01d0oooo103o000e0?ooo`03
0?oOW0180000M;oo0140oooo000J0?ooo`030?oOW0180000M;oo01P0oooo1@00o`0h0?ooo`030?oo
_`1d0000B9cO0080oooo0P000?l00`0007@0_ooo0?ooo`0:0?ooo`<00?l0<`3oool3003o01X0oooo
00<000000?ooo`3oool03`3oool60?l003P0oooo1P00o`001@3oool0oono07@00018W=l0oooo00H0
003o2P3oool00`2oM000M;oo0?ooo`0L0?ooo`D0o`004`3oool2003o01l0oooo00<0omnL04P0001d
_ol04@3oool000X0oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`030?ooo`030?oO
W0180000M;oo01P0oooo1@00o`000`3o0000oooo0?ooo`0X0?ooo`030?oog`2LB000000000L00000
00L0B9cO0?ooo`3oool0omnL04P0001d_ol0oooo00@0003o00<0W=oo0?ooo`3oool0203oool5003o
0340oooo1P00o`0H0?ooo`030000003oool0oooo00h0oooo1P3o000h0?ooo`H00?l00P3oool00`3o
gi`0B00007Boo`020?ooo`@0003o2P3oool00`3ooml0W4P004RLg`0N0?ooo`@0o`004@3oool4003o
01l0oooo00<0omnL04P0001d_ol04@3oool001X0oooo00<0omnL04P0001d_ol05`3oool5003o03T0
oooo00<0omnL04P0001d_ol01P000?l:0?ooo`H00?l0<@3oool6003o01L0oooo00<000000?ooo`3o
ool03@3oool60?l003P0oooo1P00o`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000o`2L
gol02P3oool00`3OW4P0000007Boo`0O0?ooo`@0o`00403oool5003o01h0oooo00<0omnL04P0001d
_ol04@3oool001X0oooo00<0omnL04P0001d_ol05`3oool5003o03T0oooo00<0omnL04P000000?l0
1@000?l00`2Lgol0oooo0?ooo`040?ooo`050=oOg`3oool0oono07@0000000001P00o`0a0?ooo`H0
0?l05P3oool00`000000oooo0?ooo`0>0?ooo`<0o`00>@3oool6003o00@0oooo00<0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`060?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`0O
0?ooo`@0o`00403oool5003o01h0oooo00<0omnL04P0001d_ol04@3oool001X0oooo00<0omnL04P0
001d_ol05`3oool4003o03X0oooo00<0omnL0000o`000?l010000?l00`0004P0W=oo0?ooo`050?oo
o`060=nLB000000000000000000004P0W=oo1P00o`0b0?ooo`H00?l0503oool00`000000oooo0?oo
o`0>0?ooo`80o`00>P3oool5003o00D0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`040?ooo`030000o`3oool0gia800<0000000<0001809cOo`3oool0803oool40?l00100oooo1@00
o`0M0?ooo`030?oOW0180000M;oo0140oooo000J0?ooo`030?oOW0180000M;oo01L0oooo1000o`04
0?ooo`80o`00=03oool60000o`030=nLB00007@0_ooo00H0oooo00@0oono07@000000000B9cO0`3o
ool6003o0380oooo1P00o`0C0?ooo`030000003oool0oooo00P0oooo0P3o00110?ooo`<00?l01P3o
ool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00<0oooo10000?l00`1d0000000004RL
g`0S0?ooo`@0o`00403oool5003o01d0oooo00<0omnL04P0001d_ol04@3oool001X0oooo00<0omnL
04P0001d_ol05P3oool5003o0080oooo103o000c0?ooo`H0003o00<0oooo0;md0000M;l0203oool0
0`2oM000M;oo0?ooo`040?ooo`H00?l0<P3oool6003o0180oooo0`0000070?ooo`@0o`00@@3oool0
0`00o`00oooo0?ooo`060?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool00P3oool6
0000o`0307Boo`3oool0oooo02<0oooo103o000@0?ooo`D00?l0703oool00`3ogi`0B00007Boo`0A
0?ooo`00603oool01@3ooml0W4P0000000000000M;oo01H0oooo1000o`040?ooo`@0o`00<@3oool6
0000o`040?ooo`3o_g@0001d0;ooo`P0oooo00H0omnL04P00000000000000000B02Lgol20?ooo`H0
0?l0<`3oool3003o01<0oooo00<000000?ooo`3oool01P3oool60?l004T0oooo00L0oono07@0001d
_ol0oooo0?noM00007@0_ooo00<0oooo1`000?l2000000030000B02Lgol0oooo0200oooo1@3o000?
0?ooo`D00?l06P3oool01@3ooml0W4P0000000000000M;oo0140oooo000J0?ooo`030?noM0000000
M;oo01P0oooo0P00o`040?ooo`@0o`00<@3oool50000o`0407@000000000001d0;ooo`X0oooo00D0
_g@0000000000000000004RLg`030?ooo`@00?l0=@3oool2003o00@0oooo0P00o`0=0?ooo`030000
003oool0oooo00D0oooo1P3o000k0?ooo`800?l03P3oool01@3ookl0M0000000000007@0_ooo00H0
oooo1P000?l20000000304RLg`3oool0oooo0200oooo103o000?0?ooo`H00?l06`3oool00`3o_g@0
000007Boo`0A0?ooo`00?03oool40?l00340oooo0`000?lH0?ooo`800?l0?03oool3003o00`0oooo
00<000000?ooo`3oool0103oool60?l003`0oooo0`00o`0I0?ooo`H0003o9@3oool00`3o0000oooo
0?ooo`0?0?ooo`H00?l0;`3oool003`0oooo103o000b0?ooo`030000o`3oool0oooo05@0oooo1P00
o`0:0?ooo`030000003oool0oooo00<0oooo1P3o000l0?ooo`D00?l06@3oool70000ocD0oooo1P00
o`0_0?ooo`00<@3oool2003o00X0oooo0`3o002:0?ooo`L00?l0203oool00`000000oooo0?ooo`02
0?ooo`H0o`00?03oool5003o01/0oooo1P000?le0?ooo`D00?l0<03oool00340oooo1000o`080?oo
o`@0o`00:`3oool20000ob@0oooo00<00?l00?ooo`3oool0=P3oool8003o00H0oooo00@000000?oo
o`3oool0oooo1P3o000l0?ooo`H00?l0703oool40000obT0oooo00<0o`000?ooo`3oool0?`3oool0
0300oooo1@00o`080?ooo`@0o`00:P3oool40000ob80oooo0`00o`0i0?ooo`L00?l01@3oool01000
0000oooo0?ooo`3oool50?l003`0oooo1P00o`0O0?ooo`030000o`3oool0oooo02L0oooo103o000o
0?ooo`00<03oool4003o00X0oooo103o000X0?ooo`H0003o803oool5003o03X0oooo1P00o`040?oo
o`030000003oool0oooo0080oooo0`3o000l0?ooo`H00?l0B@3oool50?l003l0oooo000`0?ooo`@0
0?l02P3oool40?l002L0oooo1P000?lQ0?ooo`H00?l0>`3oool3003o00D0oooo00<000000?ooo`3o
ool00`3oool00`3o0000oooo0?ooo`0k0?ooo`D00?l09`3oool20000ob40oooo1P3o000?0?ooo`80
0?l0;@3oool00300oooo1000o`0;0?ooo`@0o`009@3oool60000ob<0oooo1P00o`0k0?ooo`800?l0
1@3oool00`000000oooo0?ooo`100?ooo`D00?l09`3oool30000ob80oooo1@3o000=0?ooo`D00?l0
;03oool002l0oooo1@00o`0;0?ooo`80o`009P3oool60000obD0oooo1P00o`100?ooo`03003o003o
0000oooo0480oooo0`00o`0W0?ooo`D0003o8P3oool40?l000d0oooo1@00o`0/0?ooo`00;`3oool4
003o03<0oooo1P000?lW0?ooo`H00?l0?P3oool40?l004<0oooo0P00o`0X0?ooo`D0003o8P3oool4
0?l000d0oooo1000o`0/0?ooo`002`3ooolT000000@00?l0<P0000060000obT000001P00o`0l0000
00H0o`0000<00?l0000000000000J@0000060000ob400000103o000=000000D00?l08`0000080?oo
o`005@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00<0oooo1000o`090?ooo`03
0000003oool0oooo00@0oooo0P3o000;0?ooo`030000003oool0oooo0100oooo00<000000?ooo`00
0?l010000?l=0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01`3oool6003o00<0
oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00d0
oooo1`3o0003003o00h0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0A0?ooo`03
0000003oool0oooo00800?l03P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00/0
oooo1@000?l00`3oool000000?ooo`0A0?ooo`030000003oool0oooo00/0oooo103o00020?ooo`03
0000003oool0oooo00L0oooo1@00o`040?ooo`030000003oool0oooo0140oooo00<000000?ooo`3o
ool0403oool001D0oooo00<000000?ooo`3oool05P3oool5003o00h0oooo103o000Q0?ooo`<0003o
3P3oool00`000000oooo0?ooo`0L0?ooo`@00?l0>P3oool60?l000L00?l0=P3oool3003o0240oooo
00<000000?ooo`3oool0303oool50000ob40oooo103o000=0?ooo`@00?l0603oool00`000000oooo
0?ooo`0@0?ooo`00;P3oool4003o0100oooo103o000Q0?ooo`030000o`3oool0oooo02d0oooo0P00
o`0j0?ooo`H0o`0000<000000?ooo`00o`001P00o`0d0?ooo`D00?l0;`3oool60000ob40oooo103o
000<0?ooo`D00?l0:P3oool00300oooo0P00o`0@0?ooo`@0o`00R`3oool70?l000030?ooo`000000
oooo0080oooo1P00o`0b0?ooo`D00?l0<@3oool60000ob00oooo0P3o000>0?ooo`D00?l0:P3oool0
04<0oooo0`3o000J0?ooo`80003o>`3oool00`00o`00oooo0?ooo`0a0?ooo`H0o`000P3oool00`00
0000oooo0?ooo`030?ooo`<00?l0<P3oool6003o0380oooo1P000?l`0?ooo`@00?l0:P3oool004<0
oooo103o000I0?ooo`<0003o>@3oool3003o03<0oooo103o00030?ooo`030000003oool0oooo00@0
oooo0P00o`050?ooo`03003o003oool0oooo02T0oooo1P00o`0d0?ooo`H0003o;`3oool5003o02T0
oooo00130?ooo`@0o`00603oool50000ocL0oooo1@00o`0c0?ooo`030?l0003oool0oooo00<0oooo
00<000000?ooo`3oool02P3oool3003o02T0oooo1P00o`0f0?ooo`@0003o<03oool5003o02T0oooo
000/0?ooo`800?l05P3oool40?l001H0oooo1@000?lh0?ooo`H00?l0>03oool00`000000oooo0?oo
o`090?ooo`D00?l0:03oool5003o03P0oooo0P000?lT0?ooo`030?l0003oool0oooo00/0oooo0P00
o`0[0?ooo`00;03oool4003o01@0oooo1@3o000D0?ooo`H0003o>@3oool6003o02/0oooo0P3o000:
0?ooo`030000003oool0oooo00T0oooo1P00o`0V0?ooo`D00?l0G@3oool40?l003T0oooo000[0?oo
o`D00?l0503oool60?l00180oooo1P000?lk0?ooo`H00?l0:@3oool30?l000X0oooo00<000000?oo
o`3oool02P3oool7003o02D0oooo0`00o`100?ooo`80003o703oool50?l003P0oooo000[0?ooo`@0
0?l05P3oool40?l00180oooo1P000?lm0?ooo`H00?l09@3oool70?l000T0oooo00<000000?ooo`3o
ool02`3oool8003o02@0oooo0P00o`0o0?ooo`<0003o7@3oool40?l003P0oooo000[0?ooo`@00?l0
5`3oool20?l001<0oooo1@000?lo0?ooo`H00?l08`3oool70?l000X0oooo0`00000<0?ooo`P00?l0
H`3oool50000oad0oooo103o000=0?ooo`800?l0:03oool002/0oooo1000o`0[0?ooo`D0003o@@3o
ool6003o0240oooo1`3o000;0?ooo`030000003oool0oooo00l0oooo1P00o`0L0?ooo`03003o003o
ool0oooo04@0oooo1@000?lL0?ooo`@0o`002P3oool5003o02P0oooo000Z0?ooo`D00?l0;03oool3
0000od<0oooo1000o`0Q0?ooo`H0o`003@3oool00`000000oooo0?ooo`0A0?ooo`<00?l0703oool4
003o04@0oooo1P000?lL0?ooo`@0o`002@3oool6003o02L0oooo000Z0?ooo`@00?l07P3oool20?l0
00h0oooo0P000?m40?ooo`800?l08@3oool60?l000h0oooo00<000000?ooo`3oool04P3oool2003o
00D0oooo00<00?l00?ooo`3oool04`3oool5003o04D0oooo1P000?lK0?ooo`@0o`002@3oool6003o
02L0oooo000Z0?ooo`@00?l07@3oool30?l007L0oooo1@3o000?0?ooo`030000003oool0oooo01P0
oooo1000o`0C0?ooo`@00?l0A`3oool60000oa/0oooo103o00080?ooo`H00?l09`3oool002X0oooo
1000o`0L0?ooo`D0o`001`3oool20000odl0oooo0P00o`0M0?ooo`<0o`00403oool00`000000oooo
0?ooo`0G0?ooo`H00?l0403oool5003o04T0oooo1P000?lJ0?ooo`D0o`002@3oool4003o02L0oooo
000Y0?ooo`D00?l07@3oool50?l000D0oooo10000?m>0?ooo`D00?l06`3oool00`3o0000oooo0?oo
o`0?0?ooo`030000003oool0oooo01L0oooo1`00o`0>0?ooo`H00?l0BP3oool50000oa/0oooo103o
00090?ooo`D00?l09P3oool002T0oooo1000o`0N0?ooo`D0o`00103oool60000od`0oooo1`00o`0/
0?ooo`030000003oool0oooo01T0oooo2000o`0:0?ooo`L00?l0B`3oool50000oa/0oooo00<0o`00
0?ooo`3oool02P3oool4003o02H0oooo000[0?ooo`800?l07`3oool50?l000<0oooo1@000?m=0?oo
o`L00?l04`3oool00`3o0000oooo0?ooo`0F0?ooo`030000003oool0oooo01X0oooo2@00o`060?oo
o`L00?l0CP3oool30000obT0oooo1000o`0V0?ooo`00C@3oool50?l00080oooo10000?m?0?ooo`L0
0?l0403oool40?l001L0oooo00<000000?ooo`3oool06`3oool8003o00H0oooo1P00o`1?0?ooo`80
003o:P3oool4003o02H0oooo001=0?ooo`H0o`0010000?mB0?ooo`H00?l03P3oool60?l001H0oooo
00<000000?ooo`3oool07@3oool5003o00P0oooo1000o`1b0?ooo`80o`00203oool5003o02D0oooo
001>0?ooo`D0o`0010000?mC0?ooo`H00?l0303oool70?l001H0oooo00<000000?ooo`3oool0803o
ool2003o00T0oooo00<00?l00?ooo`3oool0EP3oool20000oaP0oooo103o00090?ooo`@00?l09@3o
ool002P0oooo0P00o`0U0?ooo`D0o`000P000?mE0?ooo`D00?l02`3oool60?l001P0oooo0`00000V
0?ooo`800?l0F`3oool30000oaP0oooo1@3o00080?ooo`800?l09`3oool002L0oooo1@00o`0T0?oo
o`<0o`000P000?mG0?ooo`D00?l02@3oool60?l001T0oooo00<000000?ooo`3oool09@3oool4003o
05T0oooo1@000?lG0?ooo`D0o`00<@3oool002L0oooo1@00o`0T0?ooo`80o`000`000?mH0?ooo`<0
0?l0203oool70?l001X0oooo00<000000?ooo`3oool08`3oool8003o05P0oooo1@000?lG0?ooo`D0
o`00<03oool002L0oooo1000o`0U0?ooo`@0003oF@3oool2003o00T0oooo1P3o000K0?ooo`030000
003oool0oooo0280oooo2`00o`1F0?ooo`D0003o5`3oool50?l000X0oooo0P00o`0T0?ooo`009P3o
ool5003o02L0oooo00<0003o0?ooo`3oool0H03oool01000o`00oooo0?ooo`3oool40?l001`0oooo
00<000000?ooo`3oool08@3oool>003o05D0oooo1@000?lG0?ooo`D0o`001`3oool5003o02<0oooo
000V0?ooo`D00?l0:@3oool40?l005`0oooo1000o`020?ooo`030?l0003oool0oooo01`0oooo00<0
00000?ooo`3oool0803oool6003o00<0oooo1`00o`1E0?ooo`D0003o5P3oool50?l000L0oooo1@00
o`0S0?ooo`009P3oool4003o0280oooo0P000?l60?ooo`@0o`00F`3oool6003o0200oooo00<00000
0?ooo`3oool07`3oool6003o00H0oooo1P00o`1D0?ooo`H0003o5P3oool50?l000L0oooo1000o`0S
0?ooo`009@3oool5003o0280oooo0`000?l50?ooo`D0o`00FP3oool7003o01l0oooo00<000000?oo
o`3oool07@3oool7003o00T0oooo0`00o`1F0?ooo`D0003o5P3oool50?l000L0oooo1@00o`0R0?oo
o`009@3oool5003o0240oooo1@000?l40?ooo`H0o`00E`3oool40?l000H00?l07P3oool00`000000
oooo0?ooo`0M0?ooo`H00?l02`3oool2003o00@0oooo0P00o`1A0?ooo`D0003o5@3oool60?l000H0
oooo1@00o`0R0?ooo`009@3oool4003o0240oooo1@000?l50?ooo`L0o`00E@3oool60?l000L00?l0
703oool00`000000oooo0?ooo`0N0?ooo`@00?l04P3oool3003o0540oooo0`000?lG0?ooo`@0o`00
203oool4003o0280oooo000T0?ooo`D00?l0803oool60000o`L0oooo1P3o001C0?ooo`H0o`000P3o
ool6003o01`0oooo00<000000?ooo`3oool07`3oool00`00o`00oooo0?ooo`0A0?ooo`H00?l0C`3o
ool20000oaT0oooo0P3o00090?ooo`D00?l08@3oool002@0oooo1@00o`0P0?ooo`D0003o2@3oool6
0?l00500oooo1P3o00050?ooo`H00?l06`3oool00`000000oooo0?ooo`0G0?ooo`800?l06`3oool6
003o07@0oooo1@00o`0Q0?ooo`009P3oool2003o0200oooo1@000?l;0?ooo`H0o`00CP3oool60?l0
00L0oooo1000o`0L0?ooo`<000005@3oool5003o01/0oooo1P00o`1C0?ooo`030000o`3oool0oooo
01h0oooo1000o`0Q0?ooo`00A`3oool50000o`d0oooo103o001>0?ooo`H0o`002@3oool2003o00D0
oooo0P00o`0F0?ooo`030000003oool0oooo0180oooo2000o`0M0?ooo`H00?l0C`3oool40000oal0
oooo1@00o`0P0?ooo`00A`3oool50000o`h0oooo0P3o001@0?ooo`<0o`00403oool6003o01@0oooo
00<000000?ooo`3oool0403oool:003o01h0oooo1P00o`1>0?ooo`D0003o5P3oool20?l000H0oooo
1@00o`0P0?ooo`00AP3oool50000oe/0oooo0P3o00040?ooo`80o`004@3oool9003o0140oooo00<0
00000?ooo`3oool03`3oool;003o01l0oooo1P00o`1>0?ooo`D0003o4`3oool40?l000L0oooo0P00
o`0R0?ooo`008`3oool2003o0280oooo0`000?lH0?ooo`030?l0003oool0oooo0400oooo0`3o000G
0?ooo``00?l03P3oool00`000000oooo0?ooo`0?0?ooo`P00?l0903oool3003o0500oooo10000?lC
0?ooo`D0o`00:P3oool002<0oooo1000o`0Q0?ooo`80003o5`3oool40?l003h0oooo1P3o000F0?oo
o`h00?l0303oool00`000000oooo0?ooo`090?ooo`800?l01@3oool4003o02P0oooo0P00o`040?oo
o`800?l0BP3oool50000oa<0oooo103o000Z0?ooo`008P3oool5003o03T0oooo1P3o000l0?ooo`H0
o`006`3oool:003o00D0oooo00<00?l00?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`L0
0?l01@3oool2003o0300oooo0`00o`1:0?ooo`@0003o4`3oool40?l002X0oooo000R0?ooo`@00?l0
>P3oool70?l003T0oooo1`3o000O0?ooo`H00?l01@3oool3003o00D0oooo00@000000?ooo`3oool0
oooo2`00o`0e0?ooo`H00?l0B@3oool30000oa<0oooo103o00080?ooo`800?l0803oool00280oooo
1000o`0L0?ooo`80003o7P3oool60?l003D0oooo203o000U0?ooo`800?l0103oool7003o0080oooo
00<00000003o0000o`003000o`0f0?ooo`L00?l0A`3oool40000oa80oooo1@3o00050?ooo`@00?l0
803oool00280oooo1000o`0L0?ooo`<0003o7P3oool70?l003<0oooo1`3o000/0?ooo`L00?l00P3o
ool00`0000000?l0003o000:003o03T0oooo2000o`150?ooo`@0003o4`3oool40?l000D0oooo1@00
o`0O0?ooo`008P3oool4003o01/0oooo1@000?lN0?ooo`H0o`00<`3oool50?l002l0oooo1P00o`02
0?ooo`0300000000o`000?l000P00?l0?@3oool7003o04D0oooo10000?lB0?ooo`@0o`001P3oool4
003o01l0oooo000R0?ooo`@00?l06P3oool50000ob00oooo1@3o000^0?ooo`80o`000`3oool50?l0
02l0oooo1P00o`020?ooo`0300000000o`000?l000@00?l0@`3oool6003o04@0oooo0P000?lD0?oo
o`D0o`001@3oool4003o01l0oooo000R0?ooo`@00?l06@3oool60000ob00oooo1@3o00050?ooo`80
o`008`3oool60?l003l0oooo0`0000190?ooo`<00?l0G03oool40?l000D0oooo1000o`0O0?ooo`00
8P3oool4003o01P0oooo1P000?lY0?ooo`H0o`007@3oool;0?l003h0oooo00<000000?ooo`3oool0
BP3oool2003o05`0oooo0P3o00070?ooo`@00?l07`3oool00240oooo1@00o`0G0?ooo`H0003o:P3o
ool90?l001P0oooo3@3o000n0?ooo`030000003oool0oooo0500oooo0P00o`150?ooo`80003o603o
ool5003o01h0oooo000Q0?ooo`@00?l0603oool50000ob/0oooo2`3o000@0?ooo`80o`00103oool;
0?l00400oooo00<000000?ooo`3oool0D03oool3003o04<0oooo0`000?lI0?ooo`@00?l07P3oool0
02<0oooo0P00o`0G0?ooo`D0003o;03oool=0?l000/0oooo1@3o00040?ooo`L0o`00A03oool00`00
0000oooo0?ooo`1?0?ooo`H00?l0@03oool50000oaP0oooo1000o`0N0?ooo`00?@3oool30000oc40
oooo2@3o00030?ooo`h0o`000`3oool40?l004L0oooo00<000000?ooo`3oool0D03oool7003o03l0
oooo1@000?lA0?ooo`80o`00103oool4003o01h0oooo000n0?ooo`80003o=03oool50?l000@0oooo
3P3o00030?ooo`@0o`00A`3oool00`000000oooo0?ooo`1A0?ooo`P00?l0?@3oool60000o`h0oooo
1@3o00030?ooo`D00?l07@3oool007L0oooo0P3o00040?ooo`d0o`00<P3oool01@3ogi`0B0000000
000007@0_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oog`2LB0030000000307Boo`3oool0oooo
00T0oooo00<000000?ooo`3oool0D`3oool7003o03d0oooo1P000?l=0?ooo`D0o`00103oool4003o
01d0oooo000P0?ooo`800?l0F`3oool;0?l003<0oooo00L0oono07@0001d_ol0oooo0?noM00007@0
_ooo0080oooo00L0okmd0000001d_ol0oooo0?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool0
2@3oool00`000000oooo0?ooo`1E0?ooo`H00?l0?@3oool60000o`d0oooo103o00040?ooo`800?l0
7`3oool00200oooo1000o`0E0?ooo`80003oP03oool00`2oM00004RL0=ooo`020?ooo`030;md0000
M;l0oooo00X0oooo00<0oooO09a8M02oool02@3oool00`000000oooo0?ooo`1G0?ooo`<00?l0?`3o
ool50000o`d0oooo1@3o000T0?ooo`007`3oool5003o01D0oooo0`000?mn0?ooo`030?oo_`1d0000
B9cO00<0oooo00<0gia80000M02oool02`3oool00`2oM000M;oo0?ooo`080?ooo`030000003oool0
oooo05P0oooo0P00o`040?ooo`800?l0>P3oool50000o``0oooo1@3o000T0?ooo`007`3oool4003o
01D0oooo1@000?mm0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3o
oml0W4P004RLg`090?ooo`030000003oool0oooo05h0oooo0`00o`0j0?ooo`<0003o3P3oool40?l0
02@0oooo000O0?ooo`@00?l0503oool50000ogh0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`0:0?ooo`030=nLB0000000M;oo00T0oooo1000001L0?ooo`H00?l0>03oool20000o`l0
oooo1@3o00050?ooo`800?l0703oool001l0oooo1000o`0C0?ooo`H0003oOP3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo
00T0oooo00<000000?ooo`3oool0GP3oool6003o04P0oooo1@3o00030?ooo`@00?l0703oool001h0
oooo1@00o`0B0?ooo`H0003oO`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0
oooo00H0gia800000000000000000000B02Lgol:0?ooo`030000003oool0oooo05l0oooo1P00o`18
0?ooo`@0o`000`3oool5003o01/0oooo000N0?ooo`@00?l04`3oool50000oh00oooo00<0oono07@0
0018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d0000000004RLg``0oooo00<00000
0?ooo`3oool0H@3oool6003o03T0oooo00<0003o0?ooo`3oool02P3oool50?l000<0oooo1000o`0K
0?ooo`007P3oool4003o0180oooo1@000?n20?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d
_`3oool01`3oool00`2oM000M;oo0?ooo`0<0?ooo`030000003oool0oooo0680oooo1P00o`0g0?oo
o`<0003o2`3oool50?l000<0oooo1000o`0K0?ooo`007P3oool4003o0180oooo10000?n30?ooo`07
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`060?oOW018000000000000000004P0W=oo
2@3oool00`000000oooo0?ooo`1S0?ooo`H00?l0=@3oool50000o`/0oooo0P3o00050?ooo`@00?l0
6`3oool001d0oooo1@00o`0B0?ooo`@0003oQ03oool01@3ookl0M0000000000007@0_ooo00X0oooo
00D0_g@0000000000000000004RLg`090?ooo`030000003oool0oooo06D0oooo0`00o`0f0?ooo`H0
003o4@3oool4003o01/0oooo000M0?ooo`@00?l04P3oool40000oj80oooo00<000000?ooo`3oool0
IP3oool2003o00@0oooo0P00o`0a0?ooo`D0003o4@3oool4003o01/0oooo000O0?ooo`800?l0503o
ool20000oj80oooo00<000000?ooo`3oool0K03oool5003o02l0oooo1@000?l@0?ooo`@00?l06`3o
ool00=T0oooo00<000000?ooo`3oool0J`3oool7003o02l0oooo1@000?l<0?ooo`80o`0000<0oooo
003o0000o`000P00o`0K0?ooo`00f@3oool00`000000oooo0?ooo`1[0?ooo`P00?l0;`3oool40000
o`X0oooo103o00000`3oool00?l0003o0003003o01X0oooo000`0?ooo`80003oY`3oool00`000000
oooo0?ooo`1/0?ooo`T00?l0;P3oool40000o`T0oooo1@3o00000`3oool00?l0003o0002003o01X0
oooo000`0?ooo`<0003oYP3oool3000006l0oooo1`00o`0]0?ooo`D0003o2@3oool40?l000030?oo
o`00o`000?l001`0oooo000K0?ooo`@00?l0403oool50000ojD0oooo00<000000?ooo`3oool0L@3o
ool6003o02d0oooo10000?l90?ooo`@0o`007`3oool001/0oooo1000o`0?0?ooo`D0003oYP3oool0
0`000000oooo0?ooo`1c0?ooo`<00?l0;`3oool00`000?l0oooo0?ooo`090?ooo`D0o`007P3oool0
01/0oooo1000o`0>0?ooo`H0003oYP3oool00`000000oooo0?ooo`1d0?ooo`800?l0?03oool40?l0
01h0oooo000K0?ooo`D00?l03@3oool50000ojL0oooo00<000000?ooo`3oool0NP3oool2003o03H0
oooo103o00030?ooo`800?l06@3oool001/0oooo1@00o`0<0?ooo`D0003oZ03oool00`000000oooo
0?ooo`1j0?ooo`<00?l0;03oool00`000?l0oooo0?ooo`060?ooo`@0o`0000<0oooo003o0000o`00
0P00o`0I0?ooo`006`3oool4003o00`0oooo1@000?nY0?ooo`030000003oool0oooo07T0oooo1P00
o`0X0?ooo`@0003o1`3oool50?l000D00?l0603oool001/0oooo1000o`0<0?ooo`D0003oZ@3oool0
0`000000oooo0?ooo`1j0?ooo`L00?l09P3oool50000o`L0oooo103o00000`3oool00?l0003o0002
003o01P0oooo000K0?ooo`@00?l02`3oool50000ojX0oooo00<000000?ooo`3oool0N`3oool9003o
02@0oooo10000?l70?ooo`80o`000`3oool4003o01P0oooo000K0?ooo`@00?l0303oool30000oj/0
oooo00<000000?ooo`3oool0O@3oool8003o02@0oooo10000?l;0?ooo`@00?l0603oool001X0oooo
1@00o`0=0?ooo`80003oZ`3oool00`000000oooo0?ooo`1n0?ooo`L00?l0903oool40000o`/0oooo
1000o`0H0?ooo`006P3oool4003o0;/0oooo00<000000?ooo`3oool0O`3oool5003o02H0oooo1000
0?l:0?ooo`@00?l0603oool001`0oooo0P00o`2k0?ooo`<00000PP3oool2003o00@0oooo0P00o`0P
0?ooo`@0003o2@3oool00`3o00000?l0003o0002003o01P0oooo000X0?ooo`030000o`3oool0oooo
0:h0oooo00<000000?ooo`3oool0R03oool3003o0200oooo10000?l60?ooo`<0o`001000o`0H0?oo
o`009`3oool40000ojh0oooo00<000000?ooo`3oool0Q`3oool6003o01h0oooo1@000?l50?ooo`<0
o`001@00o`0G0?ooo`009P3oool50000ojh0oooo00<000000?ooo`3oool0R03oool6003o01h0oooo
10000?l60?ooo`<0o`001000o`0G0?ooo`006@3oool2003o00/0oooo10000?n_0?ooo`030000003o
ool0oooo08T0oooo1P00o`0N0?ooo`030000o`3oool0oooo00H0oooo0`3o0002003o01T0oooo000I
0?ooo`@00?l0203oool40000ok00oooo00<000000?ooo`3oool0R`3oool6003o02D0oooo103o000J
0?ooo`00603oool5003o00L0oooo1@000?n`0?ooo`030000003oool0oooo08`0oooo1P00o`0T0?oo
o`D0o`006@3oool001P0oooo1000o`080?ooo`@0003o/@3oool00`000000oooo0?ooo`2=0?ooo`H0
0?l07`3oool20000o`<0oooo103o000I0?ooo`00603oool4003o00L0oooo1@000?na0?ooo`030000
003oool0oooo08l0oooo0`00o`0N0?ooo`@0003o0`3oool40?l000030?ooo`00o`000?l001H0oooo
000H0?ooo`@00?l01`3oool40000ok80oooo00<000000?ooo`3oool0T03oool2003o00@0oooo0P00
o`0I0?ooo`@0003o0P3oool30?l000@00?l05P3oool001P0oooo1000o`070?ooo`<0003o/`3oool0
0`000000oooo0?ooo`2F0?ooo`<00?l0603oool40000o`80oooo0`3o0005003o01D0oooo000H0?oo
o`@00?l01P3oool40000ok<0oooo00<000000?ooo`3oool0U@3oool6003o01L0oooo10000?l20?oo
o`<0o`001000o`0E0?ooo`00603oool4003o00P0oooo0P000?nc0?ooo`<00000UP3oool7003o01D0
oooo10000?l20?ooo`80o`0000<0oooo003o0000o`000P00o`0E0?ooo`00603oool4003o0;d0oooo
00<000000?ooo`3oool0U`3oool8003o01@0oooo0`000?l50?ooo`@00?l05@3oool001L0oooo1@00
o`2m0?ooo`030000003oool0oooo09T0oooo1`00o`0C0?ooo`@0003o103oool4003o01D0oooo000G
0?ooo`@00?l01@3oool00`000?l0oooo0?ooo`2f0?ooo`030000003oool0oooo09/0oooo1P00o`0B
0?ooo`@0003o103oool4003o01D0oooo000I0?ooo`800?l0103oool40000okH0oooo00<000000?oo
o`3oool0W@3oool3003o01@0oooo10000?l30?ooo`@00?l05@3oool001h0oooo1@000?nf0?ooo`03
0000003oool0oooo09h0oooo0P00o`0D0?ooo`@0003o00<0oooo0?l0003o00001000o`0E0?ooo`00
7P3oool40000okL0oooo00<000000?ooo`3oool0Y03oool2003o00l0oooo10000?l20?l000D00?l0
503oool001d0oooo10000?nh0?ooo`030000003oool0oooo0:@0oooo0`00o`0>0?ooo`80003o0`3o
ool20?l000@00?l0503oool001H0oooo0P00o`050?ooo`@0003o^03oool00`000000oooo0?ooo`2S
0?ooo`H00?l04@3oool20?l000800?l000<0o`000?ooo`3oool04`3oool001H0oooo1000o`020?oo
o`@0003o^@3oool00`000000oooo0?ooo`2T0?ooo`L00?l03`3oool50?l001D0oooo000E0?ooo`D0
0?l00P3oool40000okT0oooo00<000000?ooo`3oool0Y@3oool8003o00d0oooo00<0003o0?l0003o
00000P3o000E0?ooo`005@3oool4003o0080oooo10000?nj0?ooo`030000003oool0oooo0:L0oooo
1`00o`0:0?ooo`<0003o1@3o000D0?ooo`005@3oool4003o00030?ooo`000?l0003o00<0003o^P3o
ool00`000000oooo0?ooo`2Y0?ooo`H00?l02@3oool30000o`@0o`000P00o`0C0?ooo`005@3oool4
003o00030?ooo`000?l0003o0080003o^`3oool300000:/0oooo0`00o`0;0?ooo`<0003o00<0o`00
003o0000o`000P00o`0C0?ooo`005@3oool4003o00<0oooo00<0003o0?ooo`3oool0^P3oool00`00
0000oooo0?ooo`2/0?ooo`800?l0103oool2003o00H0oooo0P000?l00`3o00000?l0003o0003003o
0180oooo000E0?ooo`@00?l0`03oool00`000000oooo0?ooo`2b0?ooo`<00?l01@3oool20000o`80
o`001000o`0B0?ooo`005@3oool4003o0<00oooo00<000000?ooo`3oool0/@3oool6003o00@0oooo
0P000?l00`3o00000?l0003o0002003o0180oooo000E0?ooo`@00?l0`03oool00`000000oooo0?oo
o`2b0?ooo`H00?l00`3oool30000o`@00?l04P3oool001@0oooo1@00o`300?ooo`030000003oool0
oooo0;<0oooo1P00o`030?ooo`80003o1@00o`0A0?ooo`00503oool4003o00<0003o_P3oool00`00
0000oooo0?ooo`2e0?ooo`H00?l000@0oooo0000o`000?l0003o1000o`0A0?ooo`005P3oool2003o
00<0003o[`3oool01P2oM000000000000000000004RL0=ooo`T0oooo00<000000?ooo`3oool0]P3o
ool6003o00030?ooo`000?l0003o00@00?l04@3oool001H0oooo10000?na0?ooo`030?oOW0180000
M;oo00/0oooo00<000000?ooo`3oool0]`3oool6003o00030?ooo`3o00000?l000<00?l04@3oool0
01D0oooo10000?nb0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0^@3oool3003o
0080oooo00<0o`00003o0000o`000`00o`0@0?ooo`005@3oool40000ok80oooo00<0omnL04P0001d
_ol02`3oool00`000000oooo0?ooo`2j0?ooo`800?l00`3oool00`3o00000?l0003o0002003o0100
oooo000C0?ooo`030?l000000?l0003o0080003o1`3o00040?ooo`d0o`000`3oool=0?l000<0oooo
3P3o00040?ooo``0o`00103oool<0?l000D0oooo303o00040?ooo``0o`001@3oool<0?l000@0oooo
303o00050?ooo``0o`00103oool<0?l000D0oooo303o00040?ooo``0o`00103oool=0?l000@0oooo
303o00040?ooo`d0o`00103oool<0?l000@0oooo3@3o00040?ooo``0o`00103oool<0?l000D0oooo
303o00040?ooo``0o`001@3oool<0?l000050?ooo`3o00000?l0003o003o00004@3oool001<0oooo
1@000?l70?l000@0oooo3@3o00030?ooo`d0o`000`3oool>0?l000@0oooo303o00040?ooo``0o`00
1@3oool<0?l000@0oooo303o00050?ooo``0o`00103oool<0?l000D0oooo303o00040?ooo``0o`00
1@3oool<0?l000@0oooo303o00040?ooo`d0o`00103oool<0?l000@0oooo3@3o00040?ooo``0o`00
103oool=0?l000@0oooo303o00040?ooo``0o`001@3oool<0?l000@0oooo303o00050?ooo``0o`00
00<0oooo0?l0003o00000`3o000@0?ooo`004`3oool40000o`P0o`00103oool=0?l000<0oooo3@3o
00030?ooo`h0o`00103oool<0?l000@0oooo303o00050?ooo``0o`00103oool<0?l000D0oooo303o
00040?ooo``0o`001@3oool<0?l000@0oooo303o00050?ooo``0o`00103oool<0?l000@0oooo3@3o
00040?ooo``0o`00103oool=0?l000@0oooo303o00040?ooo`d0o`00103oool<0?l000@0oooo303o
00050?ooo``0o`00103oool<0?l000D0oooo303o00020?ooo`@0o`00403oool001<0oooo0P3o0000
0`000?l0o`000?l000070?l000@0oooo3@3o00030?ooo`d0o`000`3oool>0?l000@0oooo303o0004
0?ooo``0o`001@3oool<0?l000@0oooo303o00050?ooo``0o`00103oool<0?l000D0oooo303o0004
0?ooo``0o`001@3oool<0?l000@0oooo303o00040?ooo`d0o`00103oool<0?l000@0oooo3@3o0004
0?ooo``0o`00103oool=0?l000@0oooo303o00040?ooo``0o`001@3oool<0?l000@0oooo303o0005
0?ooo``0o`000P3oool20?l00180oooo003;0?ooo`030?oOW0180000M;oo00/0oooo00<000000?oo
o`3oool0e03oool00</0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`3D0?ooo`00
b@3oool01@3ooml0W4P0000000000000M;oo00/0oooo00<000000?ooo`3oool0e03oool00</0oooo
00<0okmd0000001d_ol02`3oool00`000000oooo0?ooo`3D0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.10918, -1.09709, \
0.00765073, 0.0123791}}]
}, Closed]],

Cell[TextData[{
  "Here we provide very brief explanations of these graphics directives; \
please refer to the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " ",
  StyleBox["Book",
    FontSlant->"Italic"],
  " or to the Help Browser for more detailed information about these or \
related directives."
}], "Text"],

Cell[TextData[{
  ButtonBox["Thickness",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[r]", "InlineInput",
    FontWeight->"Bold"],
  " specifies line thickness as a fraction ",
  StyleBox["r",
    FontSlant->"Italic"],
  " of the width of the entire plot",
  StyleBox[" ",
    FontWeight->"Bold"]
}], "Text",
  CellDingbat->"\[EmptySmallSquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
  ButtonBox["AbsoluteThickness",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[d]", "InlineInput",
    FontWeight->"Bold"],
  " specifies line thickness in units of printer's points (",
  Cell[BoxData[
      \(TraditionalForm\`\(\(\[TildeTilde]\)\(\ \)\(1\/72\)\)\)]],
  "inch)",
  StyleBox[" ",
    FontWeight->"Bold"]
}], "Text",
  CellDingbat->"\[EmptySmallSquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
  ButtonBox["Dashing",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[{", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[Cell[BoxData[
      \(TraditionalForm\`r\_1, \ r\_2, \  ... \)], "InlineInput"], 
    "InlineInput"],
  StyleBox["}]", "InlineInput",
    FontWeight->"Bold"],
  " specifies a sequence of segments and spaces as sizes relative to figure \
width"
}], "Text",
  CellDingbat->"\[EmptySmallSquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
  ButtonBox["AbsoluteDashing",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[{", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[Cell[BoxData[
      \(TraditionalForm\`r\_1, \ r\_2, \  ... \)], "InlineInput"], 
    "InlineInput"],
  StyleBox["}]", "InlineInput",
    FontWeight->"Bold"],
  " specifies a sequence of segments and spaces in units of printer's points \
(",
  Cell[BoxData[
      \(TraditionalForm\`\(\(\[TildeTilde]\)\(\ \)\(1\/72\)\)\)]],
  "inch)",
  StyleBox[" ",
    FontWeight->"Bold"]
}], "Text",
  CellDingbat->"\[EmptySmallSquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
  ButtonBox["RGBColor",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[r,g,b]", "InlineInput",
    FontWeight->"Bold"],
  " specifies color values between 0 and 1 for red, green, and blue"
}], "Text",
  CellDingbat->"\[EmptySmallSquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}},
  FontWeight->"Plain"],

Cell[TextData[{
  ButtonBox["Hue",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[h]", "InlineInput",
    FontWeight->"Bold"],
  " specifies hue values between 0 and 1"
}], "Text",
  CellDingbat->"\[EmptySmallSquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}},
  FontWeight->"Plain"]
}, Closed]],

Cell[CellGroupData[{

Cell["Text styles", "Subsubsection"],

Cell[TextData[{
  "The font and style parameters used to format text and labels can be chosen \
by replacing the default ",
  ButtonBox["TextStyle",
    ButtonStyle->"RefGuideLink"],
  ".  The default value is ",
  StyleBox["$TextStyle", "InlineInput",
    FontWeight->"Bold"],
  " and depends upon installation, but is usually empty.  However, I usually \
prefer to use somewhat larger labels than produced by the default settings \
and usually include the following statement near the beginning of my \
notebooks.  Then, providing that the cells are executed in order, one should \
find a slightly improved version of an earlier plot."
}], "Text"],

Cell[BoxData[
    \(\($TextStyle = {FontFamily \[Rule] "\<Times\>", 
          FontSize \[Rule] 12};\)\)], "Input"],

Cell[TextData[{
  "Now we define a plot for the gamma function and play with ",
  StyleBox["TextStyle", "InlineInput",
    FontWeight->"Bold"],
  " a little."
}], "Text"],

Cell[BoxData[
    \(\(plotGamma\  := \ 
        Plot[Gamma[x], {x, \(-5\), 5}, \ GridLines \[Rule] Automatic, 
          Frame \[Rule] True, \ 
          FrameLabel \[Rule] {"\<x\>", \ "\<\[CapitalGamma][x]\>"}, \ 
          PlotLabel \[Rule] "\<Gamma Function\>"];\)\)], "Input"],

Cell[BoxData[
    \(\($TextStyle = {FontFamily \[Rule] "\<Courier\>", 
          FontSize \[Rule] 12};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plotGamma;\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.0952381 0.316178 0.0220537 [
[.11905 -0.0125 -9.5625 -14.875 ]
[.11905 -0.0125 9.5625 0 ]
[.30952 -0.0125 -9.5625 -14.875 ]
[.30952 -0.0125 9.5625 0 ]
[.5 -0.0125 -5.59375 -14.875 ]
[.5 -0.0125 5.59375 0 ]
[.69048 -0.0125 -5.59375 -14.875 ]
[.69048 -0.0125 5.59375 0 ]
[.88095 -0.0125 -5.59375 -14.875 ]
[.88095 -0.0125 5.59375 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -5.59375 -29.75 ]
[.5 -0.0125 5.59375 -14.875 ]
[-0.0125 .09564 -26.3125 -7.4375 ]
[-0.0125 .09564 0 7.4375 ]
[-0.0125 .20591 -19.125 -7.4375 ]
[-0.0125 .20591 0 7.4375 ]
[-0.0125 .31618 -11.1875 -7.4375 ]
[-0.0125 .31618 0 7.4375 ]
[-0.0125 .42645 -11.1875 -7.4375 ]
[-0.0125 .42645 0 7.4375 ]
[-0.0125 .53672 -18.375 -7.4375 ]
[-0.0125 .53672 0 7.4375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -41.1875 -16.375 ]
[-0.01875 .30902 -26.3125 16.375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -52.3125 0 ]
[.5 .63053 52.3125 14.875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.11905 0 m
.11905 .61803 L
s
.30952 0 m
.30952 .61803 L
s
.5 0 m
.5 .61803 L
s
.69048 0 m
.69048 .61803 L
s
.88095 0 m
.88095 .61803 L
s
0 .09564 m
1 .09564 L
s
0 .20591 m
1 .20591 L
s
0 .31618 m
1 .31618 L
s
0 .42645 m
1 .42645 L
s
0 .53672 m
1 .53672 L
s
0 g
.11905 0 m
.11905 .00625 L
s
gsave
.11905 -0.0125 -70.5625 -18.875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 14.625 moveto
%%IncludeResource: font Mathematica1Mono
%%IncludeFont: Mathematica1Mono
/Mathematica1Mono findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
70.938 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.30952 0 m
.30952 .00625 L
s
gsave
.30952 -0.0125 -70.5625 -18.875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 14.625 moveto
%%IncludeResource: font Mathematica1Mono
%%IncludeFont: Mathematica1Mono
/Mathematica1Mono findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
70.938 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66.5938 -18.875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.69048 0 m
.69048 .00625 L
s
gsave
.69048 -0.0125 -66.5938 -18.875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.88095 0 m
.88095 .00625 L
s
gsave
.88095 -0.0125 -66.5938 -18.875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.2619 0 m
.2619 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.7381 0 m
.7381 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.07143 0 m
.07143 .00375 L
s
.02381 0 m
.02381 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.97619 0 m
.97619 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -66.5938 -33.75 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
70.188 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .09564 m
.00625 .09564 L
s
gsave
-0.0125 .09564 -87.3125 -11.4375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 14.625 moveto
%%IncludeResource: font Mathematica1Mono
%%IncludeFont: Mathematica1Mono
/Mathematica1Mono findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
70.938 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
0 .20591 m
.00625 .20591 L
s
gsave
-0.0125 .20591 -80.125 -11.4375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 14.625 moveto
%%IncludeResource: font Mathematica1Mono
%%IncludeFont: Mathematica1Mono
/Mathematica1Mono findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
70.938 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .31618 m
.00625 .31618 L
s
gsave
-0.0125 .31618 -72.1875 -11.4375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .42645 m
.00625 .42645 L
s
gsave
-0.0125 .42645 -72.1875 -11.4375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .53672 m
.00625 .53672 L
s
gsave
-0.0125 .53672 -79.375 -11.4375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .1177 m
.00375 .1177 L
s
0 .13975 m
.00375 .13975 L
s
0 .1618 m
.00375 .1618 L
s
0 .18386 m
.00375 .18386 L
s
0 .22796 m
.00375 .22796 L
s
0 .25002 m
.00375 .25002 L
s
0 .27207 m
.00375 .27207 L
s
0 .29412 m
.00375 .29412 L
s
0 .33823 m
.00375 .33823 L
s
0 .36029 m
.00375 .36029 L
s
0 .38234 m
.00375 .38234 L
s
0 .40439 m
.00375 .40439 L
s
0 .4485 m
.00375 .4485 L
s
0 .47055 m
.00375 .47055 L
s
0 .49261 m
.00375 .49261 L
s
0 .51466 m
.00375 .51466 L
s
0 .07359 m
.00375 .07359 L
s
0 .05153 m
.00375 .05153 L
s
0 .02948 m
.00375 .02948 L
s
0 .00743 m
.00375 .00743 L
s
0 .55877 m
.00375 .55877 L
s
0 .58082 m
.00375 .58082 L
s
0 .60288 m
.00375 .60288 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -111.125 -11.4375 Mabsadd m
1 1 Mabs scale
currentpoint translate
77.375 11.4375 translate
90 rotate
-77.375 -11.4375 translate
0 22.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 14.625 moveto
%%IncludeResource: font Mathematica1Mono
%%IncludeFont: Mathematica1Mono
/Mathematica1Mono findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(G) show
%%IncludeResource: font Mathematica2Mono
%%IncludeFont: Mathematica2Mono
/Mathematica2Mono findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
70.188 14.625 moveto
(@) show
77.375 14.625 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
%%IncludeResource: font Mathematica2Mono
%%IncludeFont: Mathematica2Mono
/Mathematica2Mono findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
84.563 14.625 moveto
(D) show
91.750 14.625 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.11905 .61178 m
.11905 .61803 L
s
.30952 .61178 m
.30952 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.69048 .61178 m
.69048 .61803 L
s
.88095 .61178 m
.88095 .61803 L
s
.125 Mabswid
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.2619 .61428 m
.2619 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.7381 .61428 m
.7381 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.07143 .61428 m
.07143 .61803 L
s
.02381 .61428 m
.02381 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.97619 .61428 m
.97619 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .09564 m
1 .09564 L
s
.99375 .20591 m
1 .20591 L
s
.99375 .31618 m
1 .31618 L
s
.99375 .42645 m
1 .42645 L
s
.99375 .53672 m
1 .53672 L
s
.125 Mabswid
.99625 .1177 m
1 .1177 L
s
.99625 .13975 m
1 .13975 L
s
.99625 .1618 m
1 .1618 L
s
.99625 .18386 m
1 .18386 L
s
.99625 .22796 m
1 .22796 L
s
.99625 .25002 m
1 .25002 L
s
.99625 .27207 m
1 .27207 L
s
.99625 .29412 m
1 .29412 L
s
.99625 .33823 m
1 .33823 L
s
.99625 .36029 m
1 .36029 L
s
.99625 .38234 m
1 .38234 L
s
.99625 .40439 m
1 .40439 L
s
.99625 .4485 m
1 .4485 L
s
.99625 .47055 m
1 .47055 L
s
.99625 .49261 m
1 .49261 L
s
.99625 .51466 m
1 .51466 L
s
.99625 .07359 m
1 .07359 L
s
.99625 .05153 m
1 .05153 L
s
.99625 .02948 m
1 .02948 L
s
.99625 .00743 m
1 .00743 L
s
.99625 .55877 m
1 .55877 L
s
.99625 .58082 m
1 .58082 L
s
.99625 .60288 m
1 .60288 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .31618 m
1 .31618 L
s
.5 0 m
.5 .61803 L
s
gsave
.5 .63053 -113.313 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Gamma) show
106.125 14.625 moveto
(Function) show
163.625 14.625 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02498 0 m
.02499 .30097 L
.02605 .30806 L
.02729 .31081 L
.02846 .31207 L
.02954 .31277 L
.03053 .31322 L
.03163 .31358 L
.03279 .31386 L
.03395 .31407 L
.0352 .31425 L
.03746 .31449 L
.03884 .31459 L
.04016 .31468 L
.04262 .31479 L
.045 .31487 L
.0463 .31491 L
.04753 .31493 L
.04868 .31495 L
.04993 .31497 L
.05105 .31499 L
.0521 .315 L
.05335 .315 L
.05401 .31501 L
.05472 .31501 L
.05546 .31501 L
.05617 .31501 L
.05749 .31501 L
.05872 .31501 L
.05986 .31501 L
.06111 .315 L
.06244 .31499 L
.06504 .31496 L
.06742 .31493 L
.07003 .31488 L
.07277 .31482 L
.07741 .31469 L
.08247 .31447 L
.08537 .31431 L
.08801 .31413 L
.09085 .3139 L
.09392 .31357 L
.09678 .31317 L
.09941 .31269 L
.10192 .31208 L
.10319 .31169 L
.10458 .31119 L
.10589 .31061 L
.10711 .30996 L
.10837 .30912 L
Mistroke
.10951 .30816 L
.11064 .30696 L
.11169 .30549 L
.11281 .30338 L
.11402 .30005 L
.11524 .29447 L
.11594 .28934 L
.11659 .28193 L
.1179 .24147 L
.11863 .10742 L
Mfstroke
.11863 .10742 m
.11924 .61803 L
s
.1195 .61803 m
.1205 .37806 L
.12179 .3495 L
.12302 .3397 L
.12415 .33484 L
.1252 .33195 L
.12634 .32978 L
.1274 .32827 L
.12838 .32721 L
.13065 .32545 L
.13179 .32481 L
.13304 .32425 L
.1354 .32343 L
.13793 .32282 L
.14031 .32241 L
.14247 .32213 L
.14479 .32191 L
.1461 .32181 L
.14733 .32174 L
.14863 .32168 L
.14981 .32164 L
.15047 .32162 L
.15118 .32161 L
.15247 .32159 L
.15358 .32158 L
.15479 .32159 L
.15607 .3216 L
.15726 .32163 L
.15837 .32166 L
.15939 .32169 L
.16169 .32179 L
.1642 .32194 L
.16653 .32212 L
.1693 .32239 L
.17182 .32269 L
.17698 .32351 L
.17982 .32411 L
.1825 .32481 L
.18712 .3264 L
.18975 .32761 L
.19221 .32905 L
.1948 .33099 L
.19716 .33331 L
.19847 .33491 L
.19989 .33701 L
.20123 .3394 L
.20249 .34218 L
.20366 .34535 L
.20492 .3497 L
.20607 .35483 L
Mistroke
.20712 .36102 L
.20827 .37022 L
.20952 .38546 L
.21083 .41327 L
.21206 .46886 L
Mfstroke
.21206 .46886 m
.21308 .61803 L
s
.21321 .61803 m
.21324 0 L
s
.21552 0 m
.21634 .14124 L
.2175 .20237 L
.21873 .23237 L
.21978 .24731 L
.22094 .25829 L
.22215 .26631 L
.22329 .27179 L
.22453 .27638 L
.22586 .28015 L
.22725 .28324 L
.22857 .28557 L
.23104 .28886 L
.23335 .29108 L
.23548 .29262 L
.23782 .2939 L
.2403 .2949 L
.24277 .29562 L
.24416 .29593 L
.24543 .29615 L
.24661 .29631 L
.2479 .29645 L
.24911 .29653 L
.25023 .29658 L
.25092 .29659 L
.25167 .29659 L
.25237 .29658 L
.25301 .29656 L
.25424 .2965 L
.25553 .29639 L
.25692 .29624 L
.25821 .29606 L
.26116 .2955 L
.26377 .29484 L
.26619 .29406 L
.26879 .29303 L
.27165 .29162 L
.27646 .28846 L
.2791 .28618 L
.28157 .28356 L
.28396 .28046 L
.28624 .27684 L
.28831 .27281 L
.29051 .26745 L
.2929 .25995 L
.29511 .25064 L
.29647 .24331 L
.29775 .23481 L
.299 .22445 L
.30014 .21255 L
Mistroke
.3013 .19698 L
.30254 .17446 L
.3037 .14491 L
.30479 .10353 L
.30606 .02245 L
Mfstroke
.30606 .02245 m
.30623 0 L
s
.30951 0 m
.30959 .61803 L
s
.31325 .61803 m
.31377 .5745 L
.31452 .53772 L
.31521 .51221 L
.31648 .47896 L
.31763 .45783 L
.31895 .43994 L
.32021 .4272 L
.32245 .41078 L
.32358 .40457 L
.32482 .39886 L
.32701 .39091 L
.32907 .38525 L
.33149 .38014 L
.3337 .37656 L
.33625 .3734 L
.33868 .37113 L
.33993 .3702 L
.34127 .36935 L
.34246 .36872 L
.34372 .36817 L
.34485 .36777 L
.3461 .36742 L
.34723 .3672 L
.34828 .36705 L
.34892 .367 L
.34961 .36696 L
.35087 .36697 L
.35209 .36706 L
.35322 .36722 L
.35444 .36746 L
.35579 .36783 L
.35851 .36888 L
.3609 .37016 L
.36308 .37164 L
.36544 .37362 L
.36801 .37629 L
.37235 .38225 L
.37471 .38649 L
.37691 .39123 L
.37937 .39769 L
.38197 .40627 L
.38483 .41853 L
.38745 .43369 L
.3899 .45292 L
.39219 .47794 L
.3935 .49701 L
.3947 .51886 L
.39609 .55172 L
.3974 .59444 L
Mistroke
Mfstroke
.3974 .59444 m
.39792 .61803 L
s
.40444 .61803 m
.40452 0 L
s
.41171 0 m
.41218 .0211 L
.41288 .04545 L
.41353 .0644 L
.4148 .09408 L
.416 .116 L
.41716 .13311 L
.41843 .14837 L
.42066 .16903 L
.42325 .18655 L
.42567 .19862 L
.42797 .20755 L
.43005 .21403 L
.43233 .21976 L
.43473 .22461 L
.43738 .22883 L
.43983 .23186 L
.44246 .23434 L
.44397 .23544 L
.44535 .23626 L
.44655 .23684 L
.44785 .23734 L
.44856 .23755 L
.44922 .23771 L
.45052 .23792 L
.45164 .238 L
.45287 .23798 L
.45415 .23783 L
.45535 .23757 L
.45653 .23721 L
.45779 .2367 L
.46005 .23544 L
.46274 .23334 L
.46519 .2308 L
.46774 .22737 L
.47052 .22261 L
.47313 .21685 L
.47554 .21016 L
.47761 .20305 L
.47982 .19357 L
.48226 .18014 L
.48453 .16348 L
.48575 .15223 L
.48708 .13745 L
.48826 .12155 L
.48951 .10063 L
.49074 .07422 L
.49186 .04329 L
.49299 .00213 L
s
.49299 .00213 m
.49303 0 L
s
.50068 0 m
.50068 .61803 L
s
.50675 .61803 m
.50762 .58072 L
.50876 .54517 L
.51008 .51386 L
.51153 .48796 L
.51289 .46901 L
.51417 .45451 L
.5166 .43324 L
.51881 .41878 L
.52119 .40665 L
.52376 .39633 L
.52839 .38259 L
.53266 .37354 L
.53735 .3661 L
.54238 .36007 L
.54713 .35566 L
.55147 .35243 L
.55654 .34934 L
.56207 .34665 L
.56725 .34459 L
.57212 .34299 L
.58151 .34059 L
.58692 .33953 L
.59188 .33871 L
.60198 .33743 L
.60729 .33694 L
.61288 .33652 L
.61787 .33623 L
.62241 .33602 L
.62493 .33593 L
.62765 .33585 L
.63023 .3358 L
.63137 .33577 L
.6326 .33576 L
.63375 .33574 L
.63479 .33573 L
.63585 .33572 L
.63685 .33572 L
.63801 .33571 L
.63908 .33571 L
.64026 .33571 L
.64152 .33571 L
.64274 .33572 L
.64344 .33573 L
.64407 .33573 L
.64647 .33576 L
.64866 .3358 L
.65096 .33585 L
.65593 .33599 L
.66045 .33616 L
Mistroke
.67062 .33667 L
.68004 .33732 L
.68896 .33809 L
.70889 .34039 L
.72732 .34334 L
.746 .34731 L
.76661 .35313 L
.78598 .3604 L
.80684 .37082 L
.82691 .38426 L
.84801 .40339 L
.86777 .4277 L
.88615 .45795 L
.90663 .50371 L
.92522 .56051 L
Mfstroke
.92522 .56051 m
.93903 .61803 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00l`3oool01`3o_g@00000000000000000B9cO0?ooo`3OW4P00`000000
0`1d_ol0oooo0?ooo`2`0?ooo`00m03oool00`3ookl0M00007Boo`030?ooo`030?noM00007@0_ooo
0;<0oooo003e0?ooo`070?oo_`1d0000B9cO0?ooo`3OW4P0001d0;ooo`2d0?ooo`00m`3oool00`2L
B000B4Q80018W02f0?ooo`00m`3oool00`3OW4P00000001d_`2f0?ooo`00mP3oool01@2oM00004RL
0?ooo`2oM00007Bo0;D0oooo003e0?ooo`030;md0000M;l0oooo0080oooo00<0_g@00018W03oool0
/`3oool00?<0oooo00L0oono07@000000000000004RLg`3oool0gia80080000000<004RL0?ooo`3o
ool0/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0
oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0
oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool008L0
oooo00D0okmd00000000000000000018W00c0?ooo`030;md00000000000000@0000000<004RL0?oo
o`3oool0;@3oool01@3o_g@00000000000000000B9cO03@0oooo00<0_g@000000000000010000000
0`00B9`0oooo0?ooo`0d0?ooo`050?noM0000000000000000000B9`0?P3oool008T0oooo00<0okmd
001d_`3oool0<`3oool00`3OW4P0001d0;ooo`030?ooo`030?oOW018B9`0oooo02d0oooo00<0omnL
04P0M02oool00P3oool00`2LB000M;oo0?ooo`0b0?ooo`030=nLB00007@0_ooo00<0oooo00<0omnL
04Q8W03oool0=`3oool00`3o_g@007Bo0?ooo`0n0?ooo`00R@3oool00`3o_g@007Bo0?ooo`0d0?oo
o`030?noM00004P0W=oo0380oooo00<0_g@007Boo`3oool00P3oool00`3ogi`0B4RL0?ooo`0c0?oo
o`030?noM00004P0W=oo03`0oooo00<0okmd001d_`3oool0?P3oool007T0oooo00<0omnL04P00000
0000100000001P0007@0_ooo0?ooo`3oool0omnL04P000D0000000<004RL0?ooo`3oool09@3oool0
0`3ogi`0B00000000004000000030000M02oool0oooo00@0oooo00<0omnL04P0001d_ol0<03oool0
0`3ookl0M01d0;ooo`030?ooo`030?oo_`1d04P0W=oo03@0oooo00<0omnL04P0001d_ol0=P3oool0
0`3ogi`0B00000000004000000030018W03oool0oooo03`0oooo00240?ooo`030?oo_`1d0000M;oo
0080oooo00<0okmd001d_`3oool0=P3oool00`3ookl0M00004RLg`0_0?ooo`030?oOW01807@0_ooo
00@0oooo00<0W4P007Boo`3oool0=03oool00`3ookl0M00004RLg`0e0?ooo`030?oo_`1d0000M;oo
0080oooo00<0okmd001d_`3oool0?P3oool008D0oooo00H0okmd0000M02oool0oooo0?noM000M;li
0?ooo`0309a80018W=l0oooo02d0oooo00<0omnL04P0M02oool0103oool00`2LB000M;oo0?ooo`0f
0?ooo`0309a80018W=l0oooo03@0oooo00H0okmd0000M02oool0oooo0?noM000M;lo0?ooo`00QP3o
ool01@2LB000B9cO0?ooo`3o_g@007Bo03X0oooo00<0W4Q809cOo`3oool0;03oool00`3ogi`0B01d
0;ooo`040?ooo`0309a8001d_ol0oooo03L0oooo00<0W4Q809cOo`3oool0=03oool01@2LB000B9cO
0?ooo`3o_g@007Bo03l0oooo00260?ooo`050?oOW01804P0W=oo0?noM000M;l0>P3oool00`3o_g@0
07Bo0?ooo`0/0?ooo`030?oo_`1d07@0_ooo00@0oooo00<0W4Q809cOo`3oool0=`3oool00`3o_g@0
07Bo0?ooo`0d0?ooo`050?oOW01804P0W=oo0?noM000M;l0?`3oool008L0oooo00@0gia80018W03o
_g@007Bo=03oool00`3o_g@004RL0?ooo`030?ooo`030=nLB000M;l0oooo02d0oooo00<0W4Q809cO
o`3oool00P3oool00`3ookl0M4RL0?ooo`0b0?ooo`030?noM000B9`0oooo00<0oooo00<0gia8001d
_`3oool0=@3oool0103OW4P004RL0?noM000M;lo0?ooo`00R03oool00`2LB000B4Q8001d_`0e0?oo
o`0709a80000M;l0oooo0?ooo`3o_g@0000007Boo`0^0?ooo`030?oOW01804P0W=oo0080oooo00<0
W4P007Boo`3oool0<`3oool01`2LB00007Bo0?ooo`3oool0okmd0000001d_ol0=`3oool00`2LB000
B4Q8001d_`0o0?ooo`00R03oool00`3ogi`0B000001d_`0f0?ooo`050;md0000000000000000001d
_ol0<03oool01@3ogi`0B000000000000000B9cO03H0oooo00D0_g@0000000000000000007Boo`0h
0?ooo`030?oOW018000007Bo03l0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
001N0?ooool00000=`00000L0?ooo`00GP3oool00`000000oooo0?ooo`040?ooo`8000003@3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool2
000000d0oooo00<000000?ooo`3oool02`3oool2000000h0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0000000<0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`0000003@3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0303oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0303oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo01X0oooo001N0?ooo`8000001P3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo0P00000K0?ooo`800000703oool00`00
0000oooo0000000I0?ooo`030000003oool0000003X0oooo00<000000?ooo`3oool0>03oool00`00
0000oooo0?ooo`0P0?ooo`800000703oool005h0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`0J0?ooo`030000O`3oool0oooo01X0oooo0P00000K0?ooo`800000703oool00`000000
oooo0000000I0?ooo`030000003oool0000003X0oooo00<0001o0?ooo`3oool0>03oool00`0007l0
oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001N0?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool06P3oool00`0007l0oooo0?ooo`0J0?ooo`8000006`3oool2000001`0oooo
00@000000?ooo`3oool00000603oool00`000000oooo0000000j0?ooo`030000O`3oool0oooo03P0
oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo01X0oooo00<0001o0?ooo`3oool06P3oool2000001/0oooo
0P00000L0?ooo`040000003oool0oooo000001P0oooo00<000000?ooo`000000>P3oool00`0007l0
oooo0?ooo`0h0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool06P3oool005h0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0J0?ooo`030000O`3oool0oooo01X0oooo
0P00000K0?ooo`800000703oool010000000oooo0?ooo`00000H0?ooo`030000003oool0000003X0
oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0
oooo001N0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06P3oool00`0007l0oooo
0?ooo`0J0?ooo`8000006`3oool2000001`0oooo00@000000?ooo`3oool00000603oool00`000000
oooo0000000j0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3oool00`000000
oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01X0oooo
00<0001o0?ooo`3oool06P3oool2000001/0oooo0P00000L0?ooo`040000003oool0oooo000001L0
oooo00@000000?ooo`3oool00000>P3oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo
0240oooo00<000000?ooo`3oool06P3oool005h0oooo0P0000060?ooo`030000003oool0oooo01X0
oooo00<0001o0?ooo`3oool06P3oool2000001/0oooo0P00000L0?ooo`040000003oool0oooo0000
01L0oooo00@000000?ooo`3oool00000>P3oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0
oooo0200oooo0P00000L0?ooo`00GP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
01X0oooo00<0001o0?ooo`3oool06P3oool2000001/0oooo0P00000L0?ooo`040000003oool0oooo
000001L0oooo00@000000?ooo`3oool00000>P3oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3o
ool0oooo0240oooo00<000000?ooo`3oool06P3oool005h0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0J0?ooo`030000O`3oool0oooo01X0oooo0P00000K0?ooo`800000703oool0
10000000oooo0?ooo`00000G0?ooo`040000003oool0oooo000003X0oooo00<0001o0?ooo`3oool0
>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001N0?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool06P3oool00`0007l0oooo0?ooo`0J0?ooo`8000006`3o
ool2000001`0oooo00@000000?ooo`3oool000005`3oool010000000oooo0?ooo`00000j0?ooo`03
0000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00
GP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01X0oooo00<0001o0?ooo`3oool0
6P3oool2000001/0oooo0P00000L0?ooo`040000003oool0oooo000001L0oooo00@000000?ooo`3o
ool00000>P3oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo0240oooo00<000000?oo
o`3oool06P3oool005h0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0J0?ooo`03
0000O`3oool0oooo01X0oooo0P00000K0?ooo`800000703oool010000000oooo0?ooo`00000G0?oo
o`040000003oool0oooo000003X0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q
0?ooo`030000003oool0oooo01X0oooo001N0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool06P3oool00`0007l0oooo0?ooo`0J0?ooo`8000006`3oool2000001`0oooo00@000000?oo
o`3oool000005`3oool010000000oooo0?ooo`00000j0?ooo`030000O`3oool0oooo03P0oooo00<0
001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool2000000H0oooo00<00000
0?ooo`3oool06P3oool00`0007l0oooo0?ooo`0J0?ooo`8000006`3oool2000001`0oooo00@00000
0?ooo`3oool000005`3oool010000000oooo0?ooo`00000j0?ooo`030000O`3oool0oooo03P0oooo
00<0001o0?ooo`3oool0803oool2000001`0oooo001N0?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool06P3oool00`0007l0oooo0?ooo`0J0?ooo`8000006`3oool2000001`0oooo00@0
00000?ooo`3oool000005`3oool010000000oooo0?ooo`00000j0?ooo`030000O`3oool0oooo03P0
oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo01X0oooo00<0001o0?ooo`3oool06P3oool2000001/0oooo
0P00000L0?ooo`040000003oool0oooo000001L0oooo00@000000?ooo`3oool00000>P3oool00`00
07l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool06P3oool005h0
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0J0?ooo`030000O`3oool0oooo01X0
oooo0P00000K0?ooo`800000703oool010000000oooo0?ooo`00000G0?ooo`040000003oool0oooo
000003X0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0
oooo01X0oooo001N0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06P3oool00`00
07l0oooo0?ooo`0J0?ooo`8000006P3oool00`000000oooo0000000L0?ooo`040000003oool0oooo
000001L0oooo00@000000?ooo`3oool00000>P3oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3o
ool0oooo0240oooo00<000000?ooo`3oool06P3oool005h0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0J0?ooo`030000O`3oool0oooo01X0oooo0P00000J0?ooo`030000003oool0
000001`0oooo00@000000?ooo`3oool000005`3oool010000000oooo0?ooo`00000j0?ooo`030000
O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01X0oooo00<0001o0?ooo`3oool06P3o
ool2000001X0oooo00<000000?ooo`000000703oool010000000oooo0?ooo`00000G0?ooo`040000
003oool0oooo000003X0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`03
0000003oool0oooo01X0oooo001N0?ooo`8000001P3oool00`000000oooo0?ooo`0J0?ooo`030000
O`3oool0oooo01X0oooo0P00000J0?ooo`030000003oool0000001`0oooo00@000000?ooo`3oool0
00005`3oool010000000oooo0?ooo`00000j0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?oo
o`3oool0803oool2000001`0oooo001N0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool06P3oool00`0007l0oooo0?ooo`0J0?ooo`8000006P3oool00`000000oooo0000000L0?ooo`04
0000003oool0oooo000001L0oooo00@000000?ooo`3oool00000>P3oool00`0007l0oooo0?ooo`0h
0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool06P3oool005h0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`0J0?ooo`030000O`3oool0oooo01X0oooo0P00000J0?oo
o`030000003oool0000001`0oooo00@000000?ooo`3oool000005`3oool010000000oooo0?ooo`00
000j0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?oo
o`0J0?ooo`00@`3oool00`3ookl0M000000000050000000307Boo`3oool0oooo0080oooo00D0okmd
000000000000000004RLg`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06P3o
ool00`0007l0oooo0?ooo`0J0?ooo`8000006P3oool00`000000oooo0000000L0?ooo`040000003o
ool0oooo000001L0oooo00@000000?ooo`3oool00000>P3oool00`0007l0oooo0?ooo`0h0?ooo`03
0000O`3oool0oooo0240oooo00<000000?ooo`3oool06P3oool004H0oooo00<0oono07@0B02Lgol0
1P3oool00`3ogi`0B01d0;ooo`020?ooo`0309a8001d_ol0oooo00L0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`0J0?ooo`030000O`3oool0oooo01X0oooo0P00000J0?ooo`030000
003oool0000001`0oooo00@000000?ooo`3oool000005P3oool01@000000oooo0?ooo`3oool00000
03X0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo
01X0oooo00160?ooo`030?oo_`1d04P0W=oo00H0oooo00<0_g@007Boo`3oool00P3oool00`3ogi`0
B4RL0?ooo`070?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06P3oool00`0007l0
oooo0?ooo`0J0?ooo`8000006P3oool00`000000oooo0000000L0?ooo`040000003oool0oooo0000
01H0oooo00D000000?ooo`3oool0oooo0000000j0?ooo`030000O`3oool0oooo03P0oooo00<0001o
0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00=`3oool00`3ogi`0B000000000040000
00030000M02oool0oooo00D0oooo00<0oono07@0B02Lgol01@3oool00`3ookl0M01d0;ooo`030?oo
o`030?oo_`1d04P0W=oo00L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0J0?oo
o`030000O`3oool0oooo01X0oooo0P00000J0?ooo`030000003oool0000001`0oooo00@000000?oo
o`3oool000005P3oool01@000000oooo0?ooo`3oool0000003X0oooo00<0001o0?ooo`3oool0>03o
ool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo00160?ooo`030?oo_`1d04P0
W=oo00D0oooo00<0omnL04P0M02oool0103oool00`2LB000M;oo0?ooo`060?ooo`<000001@0007l0
0`000000001o0000O`0g0000O`8000006P0007l00`000000001o0000000L0000O`040000000007l0
001o000001H0001o00D000000000O`0007l0001o0000002G0000O`<00000703oool004H0oooo00<0
oono07@0B02Lgol01@3oool00`3ogi`0B01d0;ooo`040?ooo`0309a8001d_ol0oooo00H0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`0J0?ooo`030000O`3oool0oooo01X0oooo0P00
000J0?ooo`030000003oool0000001`0oooo00@000000?ooo`3oool000005P3oool01@000000oooo
0?ooo`3oool0000003X0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`03
0000003oool0oooo01X0oooo00160?ooo`030?oo_`1d04P0W=oo00D0oooo00<0omnL04P0M02oool0
103oool00`2LB000M;oo0?ooo`060?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
6P3oool00`0007l0oooo0?ooo`0J0?ooo`8000006P3oool00`000000oooo0000000L0?ooo`040000
003oool0oooo000001H0oooo00D000000?ooo`3oool0oooo0000000j0?ooo`030000O`3oool0oooo
03P0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00AP3oool00`3ookl0
M01809cOo`050?ooo`030?oo_`1d07@0_ooo00@0oooo00<0W4Q809cOo`3oool01P3oool00`000000
oooo0?ooo`050?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool2000001X0
oooo00<000000?ooo`000000703oool01@000000oooo0?ooo`3oool0000001D0oooo00D000000?oo
o`3oool0oooo0000000j0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3oool0
0`000000oooo0?ooo`0J0?ooo`00AP3oool00`3ookl0M01809cOo`060?ooo`0309a8B02Lgol0oooo
0080oooo00<0oono07A8W03oool01`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool06P3oool2000001X0oooo00<000000?ooo`000000703oool01@00
0000oooo0?ooo`3oool0000001D0oooo00D000000?ooo`3oool0oooo0000000j0?ooo`030000O`3o
ool0oooo03P0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00@`3oool0
1P3ogi`0B00000000018W9`0M01809cOo`H0oooo00<0omnL04P0B02Lgol00P3oool00`2LB000M;oo
0?ooo`070?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0J0?ooo`8000006P3oool00`000000oooo0000000L0?ooo`050000003oool0oooo0?ooo`00
00005@3oool01@000000oooo0?ooo`3oool0000003X0oooo00<0001o0?ooo`3oool0>03oool00`00
07l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo00160?ooo`0309a8000004P0W=oo00L0
oooo00D0omnL04P000000000000004RLg`090?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`8000006P3oool00`000000oooo0000000L0?oo
o`050000003oool0oooo0?ooo`0000005@3oool01@000000oooo0?ooo`3oool0000003X0oooo00<0
001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001N
0?ooo`8000001P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo0P00000J
0?ooo`030000003oool0000001`0oooo00D000000?ooo`3oool0oooo0000000E0?ooo`050000003o
ool0oooo0?ooo`000000>P3oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo0200oooo
0P00000L0?ooo`00GP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01X0oooo00<0
00000?ooo`3oool06P3oool2000001X0oooo00<000000?ooo`000000703oool01@000000oooo0?oo
o`3oool0000001D0oooo00D000000?ooo`3oool0oooo0000000j0?ooo`030000O`3oool0oooo03P0
oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool2000001X0oooo
00<000000?ooo`000000703oool01@000000oooo0?ooo`3oool0000001D0oooo00D000000?ooo`3o
ool0oooo0000000j0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3oool00`00
0000oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool06P3oool2000001X0oooo00<000000?ooo`000000703oool01@000000
oooo0?ooo`3oool0000001D0oooo00D000000?ooo`3oool0oooo0000000j0?ooo`030000O`3oool0
oooo03P0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool20000
01X0oooo00<000000?ooo`000000703oool01@000000oooo0?ooo`3oool0000001D0oooo00D00000
0?ooo`3oool0oooo0000000j0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3o
ool00`000000oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool06P3oool2000001X0oooo00<000000?ooo`000000703oool0
1@000000oooo0?ooo`3oool0000001D0oooo00D000000?ooo`3oool0oooo0000000j0?ooo`030000
O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3o
ool2000000H0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`8000006P3o
ool00`000000oooo0000000L0?ooo`050000003oool0oooo0?ooo`0000005@3oool01@000000oooo
0?ooo`3oool0000003X0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0P0?ooo`80
0000703oool005h0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0J0?ooo`030000
003oool0oooo01X0oooo0P00000J0?ooo`030000003oool0000001`0oooo00D000000?ooo`3oool0
oooo0000000E0?ooo`050000003oool0oooo0?ooo`000000>P3oool00`0007l0oooo0?ooo`0h0?oo
o`030000O`3oool0oooo0240oooo00<000000?ooo`3oool06P3oool005h0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo0P00000J0?ooo`03
0000003oool0000001`0oooo00D000000?ooo`3oool0oooo0000000D0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0>03oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo
0240oooo00<000000?ooo`3oool06P3oool005h0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo0P00000J0?ooo`030000003oool0000001`0
oooo00D000000?ooo`3oool0oooo0000000D0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0>03oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo0240oooo00<000000?oo
o`3oool06P3oool005h0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo01X0oooo0P00000J0?ooo`030000003oool0000001`0oooo00D000000?ooo`3o
ool0oooo0000000D0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0>03oool00`00
07l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool06P3oool005h0
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0
oooo0P00000J0?ooo`030000003oool0000001`0oooo00D000000?ooo`3oool0oooo0000000D0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0>03oool00`0007l0oooo0?ooo`0h0?oo
o`030000O`3oool0oooo0240oooo00<000000?ooo`3oool06P3oool005h0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo0P00000J0?ooo`03
0000003oool0000001`0oooo00D000000?ooo`3oool0oooo0000000D0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0>03oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo
0240oooo00<000000?ooo`3oool06P3oool005h0oooo0P0000060?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool06P3oool2000001X0oooo00<000000?ooo`000000703oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0h0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool0803oool2000001`0
oooo001N0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0J0?ooo`8000006P3oool00`000000oooo0000000L0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo03P0oooo
00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo
001N0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0J0?ooo`8000006P3oool00`000000oooo0000000L0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo03P0oooo00<0
001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001N
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J
0?ooo`8000006P3oool00`000000oooo0000000L0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool04@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo03P0oooo00<0001o
0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001N0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool06@3oool2000001`0oooo0P00000J0?oo
o`030000003oool0000001`0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0@0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool0>03oool00`0007l0oooo0?ooo`0h0?oo
o`030000O`3oool0oooo0240oooo00<000000?ooo`3oool06P3oool005h0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0I0?ooo`800000703oool2000001X0oooo00<000000?ooo`00
0000703oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0h0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3o
ool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo01T0oooo0P00000L0?ooo`8000006P3oool00`000000oooo0000000L0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo03P0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000
003oool0oooo01X0oooo001N0?ooo`8000001P3oool00`000000oooo0?ooo`0I0?ooo`800000703o
ool2000001X0oooo00<000000?ooo`000000703oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00l0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0h0?ooo`030000O`3o
ool0oooo03P0oooo00<0001o0?ooo`3oool0803oool2000001`0oooo001N0?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool06@3oool2000001`0oooo0P00000I0?ooo`040000003oool0
oooo000001`0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool0>03oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3o
ool0oooo0240oooo00<000000?ooo`3oool06P3oool005h0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0I0?ooo`800000703oool2000001T0oooo00@000000?ooo`3oool00000703o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`0h0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3o
ool00`000000oooo0?ooo`0J0?ooo`00CP3oool00`3ookl0M0000000000200000003001d_`3oool0
oooo00P0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`800000703oool2
000001T0oooo00@000000?ooo`3oool00000703oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0h0?ooo`030000O`3o
ool0oooo03P0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00C@3oool0
0`3o_g@0000007Boo`030?ooo`0309a80000M;l0oooo00P0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0I0?ooo`800000703oool2000001T0oooo00@000000?ooo`3oool00000703o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`0h0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3o
ool00`000000oooo0?ooo`0J0?ooo`00D`3oool00`3ookl0M01809cOo`080?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool06@3oool2000001`0oooo0P00000I0?ooo`040000003oool0
oooo000001`0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool0>03oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3o
ool0oooo0240oooo00<000000?ooo`3oool06P3oool00480oooo00<0omnL04P00000000010000000
0`0007@0_ooo0?ooo`080?ooo`030;md001d_ol0oooo00L0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0I0?ooo`800000703oool2000001T0oooo00@000000?ooo`3oool00000703o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`0h0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3o
ool00`000000oooo0?ooo`0J0?ooo`00E03oool00`2LB000M;oo0?ooo`070?ooo`<000001@0007l0
0`000000001o0000O`0I0000O`800000700007l2000001T0001o00@000000000O`0007l000007000
07l00`000000001o0000O`040000O`030000000007l0001o00`0001o00<000000000O`0007l01@00
07l00`000000001o0000O`2E0000O`<00000703oool005<0oooo00<0okmd0018W03oool0203oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo01T0oooo0P00000L0?ooo`8000006@3oool0
10000000oooo0?ooo`00000L0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo03P0oooo00<0001o0?ooo`3oool0>03o
ool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001>0?ooo`030=nLB0000000
00000080000000<0001d0;ooo`3oool0203oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo01T0oooo0P00000L0?ooo`8000006@3oool010000000oooo0?ooo`00000L0?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo03P0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000003o
ool0oooo01X0oooo001>0?ooo`030=nLB000M;l0oooo00d0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0I0?ooo`800000703oool2000001T0oooo00@000000?ooo`3oool00000703o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool01P3o
ool00`000000oooo0?ooo`0h0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3o
ool00`000000oooo0?ooo`0J0?ooo`00CP3oool00`3OW4P007Bo0?ooo`0=0?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool06@3oool2000001`0oooo00<000000?ooo`000000603oool0
10000000oooo0?ooo`00000L0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo03P0oooo00<0001o0?ooo`3oool0>03o
ool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001>0?ooo`030=nLB000M;l0
oooo00d0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`800000703oool0
0`000000oooo0000000H0?ooo`040000003oool0oooo000001`0oooo00<000000?ooo`3oool01P3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0>03o
ool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool06P3o
ool004h0oooo00<0gia80000000000000`0000000`18W=l0oooo0?ooo`070?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool06@3oool2000001`0oooo00<000000?ooo`000000603oool0
10000000oooo0?ooo`00000L0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo03P0oooo00<0001o0?ooo`3oool0>03o
ool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001N0?ooo`8000001P3oool0
0`000000oooo0?ooo`0I0?ooo`800000703oool00`000000oooo0000000H0?ooo`040000003oool0
oooo000001`0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool0>03oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3o
ool0oooo0200oooo0P00000L0?ooo`00GP3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo01T0oooo0P00000L0?ooo`030000003oool0000001L0oooo00D000000?ooo`3oool0oooo0000
000L0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0103oool2000000X0oooo00<0
00000?ooo`3oool0>03oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo0240oooo00<0
00000?ooo`3oool06P3oool005h0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I
0?ooo`800000703oool00`000000oooo0000000G0?ooo`050000003oool0oooo0?ooo`000000703o
ool00`000000oooo0?ooo`090?ooo`8000000`3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo03P0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000003o
ool0oooo01X0oooo001N0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06@3oool2
000001`0oooo00<000000?ooo`0000005`3oool01@000000oooo0?ooo`3oool0000001`0oooo00<0
00000?ooo`3oool02`3oool3000000d0oooo00<000000?ooo`3oool0>03oool00`0007l0oooo0?oo
o`0h0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool06P3oool005h0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`800000703oool00`000000oooo0000000G
0?ooo`050000003oool0oooo0?ooo`000000703oool00`000000oooo0?ooo`0K0?ooo`030000003o
ool0oooo03P0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000003o
ool0oooo01X0oooo000O0?ooo`030?oOW01dW=l0oooo00L0oooo00<0_gAd0;ooo`3oool0<P3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo01T0oooo0P00000L0?ooo`030000003oool0
000001L0oooo00D000000?ooo`3oool0oooo0000000L0?ooo`030000003oool0oooo01/0oooo00<0
00000?ooo`3oool0>03oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo0240oooo00<0
00000?ooo`3oool06P3oool001l0oooo00<0omnL07BLg`3oool01`3oool00`2oM7@0_ooo0?ooo`0b
0?ooo`8000001P3oool00`000000oooo0?ooo`0I0?ooo`800000703oool00`000000oooo0000000F
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K
0?ooo`030000003oool0oooo03P0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0P
0?ooo`800000703oool001l0oooo00<0omnL04P0000000001`0000000`0007@0_ooo0?ooo`0b0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool06@3oool2000001`0oooo00<000000?oo
o`0000005P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool06`3oool00`000000oooo0?ooo`0h0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?oo
o`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`007`3oool00`3ogi`0M9cO0?ooo`070?ooo`03
0;mdM02oool0oooo0380oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`80
0000703oool00`000000oooo0000000F0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo03P0oooo00<0001o0?ooo`3o
ool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo000O0?ooo`030?oO
W01dW=l0oooo00L0oooo00<0_gAd0;ooo`3oool0<P3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo01T0oooo0P00000L0?ooo`030000003oool0000001H0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0
>03oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0
6P3oool001l0oooo00<0omnL07BLg`3oool0?03oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo01T0oooo0P00000L0?ooo`030000003oool0000001H0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0>03o
ool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool06P3o
ool001l0oooo00<0omnL07BLg`3oool0?03oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo01T0oooo0P00000L0?ooo`040000003oool0oooo000001@0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0>03o
ool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool06P3o
ool001l0oooo00<0omnL07BLg`3oool0?03oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo01T0oooo0P00000L0?ooo`040000003oool0oooo000001@0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0>03o
ool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool06P3o
ool001l0oooo00D0omnL04P000000000001809cOo`0j0?ooo`8000001P3oool00`000000oooo0?oo
o`0I0?ooo`800000703oool010000000oooo0?ooo`00000C0?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo03P0oooo
00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0P0?ooo`800000703oool005h0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`800000703oool010000000oooo0?oo
o`00000C0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0K0?ooo`030000003oool0oooo03P0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo
0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001N0?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool06@3oool2000001`0oooo00@000000?ooo`3oool000004`3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool00`000000oooo
0?ooo`0h0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo
0?ooo`0J0?ooo`00GP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01T0oooo0P00
000L0?ooo`050000003oool0oooo0?ooo`0000004@3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0h0?ooo`030000
O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01T0oooo0P00000L0?ooo`040000003o
ool0oooo0?ooo`8000003`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0h0?ooo`030000O`3oool0oooo03P0oooo
00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo01T0oooo0P00000L0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo01X0oooo00<0
00000?ooo`3oool06`3oool00`000000oooo0?ooo`0h0?ooo`030000O`3oool0oooo03P0oooo00<0
001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo01T0oooo0P00000L0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool02@3oool2000000X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K
0?ooo`030000003oool0oooo03P0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q
0?ooo`030000003oool0oooo01X0oooo000N0?ooo`030?noM0000000000000`0000000<0M;oo0?oo
o`3oool0;P3oool2000000H0oooo00<000000?ooo`3oool06@3oool2000001`0oooo00<000000?oo
o`3oool0103oool2000000L0oooo0P00000<0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool06`3oool00`000000oooo0?ooo`0h0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?oo
o`3oool0803oool2000001`0oooo000N0?ooo`030?noM018W=l0oooo00/0oooo00<0gia807Boo`3o
ool0;`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01P0oooo00<000000?ooo`00
0000703oool00`000000oooo0?ooo`060?ooo`L000003P3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo01/0oooo00<000000?ooo`3oool0>03oool00`0007l0oooo0?ooo`0h0?ooo`03
0000O`3oool0oooo0240oooo00<000000?ooo`3oool06P3oool001h0oooo00<0okmd04RLg`3oool0
2`3oool00`3OW4P0M;oo0?ooo`0_0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
603oool00`000000oooo0000000L0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo03P0oooo00<0001o0?ooo`3oool0
>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001?0?ooo`050?noM000
0000000000000018W=l02P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`000000703oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0h0?ooo`030000O`3oool0oooo03P0oooo
00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00CP3oool00`3ogi`0B01d0;oo
o`020?ooo`0309a8001d_ol0oooo00P0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`0G0?ooo`80000000<0oooo0000003oool06`3oool00`000000oooo0?ooo`0K0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0h0?ooo`030000O`3o
ool0oooo03P0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00CP3oool0
0`2oM000M;oo0?ooo`020?ooo`030?oOW018B9`0oooo00P0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0E0?ooo`<000000P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo01/0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0
oooo03P0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0
oooo01X0oooo001=0?ooo`030?oo_`1d07@0_ooo00<0oooo00<0oono07@0B02Lgol0203oool00`00
0000oooo0?ooo`060?ooo`@000002`3oool9000000@0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`0K0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool00`00
0000oooo0?ooo`0h0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3oool00`00
0000oooo0?ooo`0J0?ooo`008@3oool00`3ooml0W7BL0=ooo`040?ooo`030?oog`2LB7@0_ooo0280
oooo00<0omnL04P0M02oool0103oool00`2LB000M;oo0?ooo`070?ooool00000=`00000L0?ooo`00
8@3oool0103ooml0W4P00018W03Oool30?ooo`030;md000007@0_ooo0280oooo00<0omnL04P0M02o
ool0103oool00`2LB000M;oo0?ooo`070?ooo`030000003oool0oooo0280oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3o
ool06`3oool00`000000oooo0?ooo`0h0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3o
ool08@3oool00`000000oooo0?ooo`0J0?ooo`008@3oool0103ooml0W7BL0=noM000M;l20?ooo`04
0=nLB01d_ml0W4Qd0;ooob80oooo00<0omnL04P0M02oool0103oool00`2LB000M;oo0?ooo`070?oo
o`030000003oool0oooo0280oooo00@000000?ooo`3oool0oooo400000090?ooo`030000003oool0
oooo01/0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0
oooo03P0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0
oooo01X0oooo000T0?ooo`040?noM000M;l0gia804RLgbD0oooo00<0oono07@0M02oool0103oool0
0`2LB4P0W=oo0?ooo`070?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool00P00000>
0?ooo`@000001P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01X0oooo00<00000
0?ooo`3oool06`3oool00`000000oooo0?ooo`0h0?ooo`030000O`3oool0oooo03P0oooo00<0001o
0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`009@3oool00`3OW4P0B9cO0?ooo`0V0?oo
o`0309a8B02Lgol0oooo0080oooo00<0oono07A8W03oool0203oool00`000000oooo0?ooo`0R0?oo
o`030000003oool0000001<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0K0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0h0?oo
o`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?oo
o`00903oool0103o_g@007Bo0=nLB018W=lV0?ooo`030?oOW01804P0W=oo0080oooo00<0W4P007Bo
o`3oool0203oool00`000000oooo0?ooo`0R0?ooo`030000003oool0000001@0oooo0P0000040?oo
o`030000003oool0oooo01/0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?oo
o`030000003oool0oooo03P0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0Q0?oo
o`030000003oool0oooo01X0oooo000Q0?ooo`040?oog`2LM9`0gmnL04Qd_`80oooo00@0_g@007Bo
g`2LB7@0_ooo903oool01@3ogi`0B000000000000000B9cO00X0oooo00<000000?ooo`3oool08P3o
ool2000001H0oooo0P0000030?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01h0oooo600000020?ooo`030000O`3o
ool0oooo03P0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`008@3oool0
103ooml0W4P00018W03Oool30?ooo`030;md000007@0_ooo03<0oooo0P00000S0?ooo`8000005`3o
ool01@000000oooo0?ooo`3oool0000001d0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0K0?ooo`030000003oool0oooo01D0oooo2@00000H0?ooo`X00000<`3oool00`0007l0oooo
0?ooo`0P0?ooo`800000703oool00240oooo00<0oooO09adW03Oool0103oool00`3ooml0W4Qd0;oo
o`0c0?ooo`030000003oool0oooo0280oooo0P00000H0?ooo`040000003oool0oooo000001d0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01<0oooo
0P00000S0?ooo`030000O`3oool0oooo00D0oooo1@00000^0?ooo`030000O`3oool0oooo0240oooo
00<000000?ooo`3oool06P3oool005h0oooo00<000000?ooo`3oool08P3oool2000001P0oooo00@0
00000?ooo`3oool000007@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01/0oooo
00<000000?ooo`3oool0403oool3000002D0oooo00<0001o0?ooo`3oool02P3oool3000002/0oooo
00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo0?oo
o`0R0?ooo`800000603oool010000000oooo0?ooo`00000M0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0>0?ooo`800000:03oool00`0007l0oooo
0?ooo`0=0?ooo`@000009`3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo
001N0?ooo`030000003oool0oooo0280oooo0P00000H0?ooo`040000003oool0oooo000001d0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool0:03oool00`0007l0oooo0?ooo`0A0?ooo`@000008`3oool00`0007l0oooo
0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo000N0?ooo`030?noM018W=l0oooo00/0oooo00<0
gia807Boo`3oool0;`3oool00`000000oooo0?ooo`0R0?ooo`8000006@3oool00`000000oooo0000
000M0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?oo
o`0;0?ooo`800000:`3oool00`0007l0oooo0?ooo`0E0?ooo`8000008@3oool00`0007l0oooo0?oo
o`0Q0?ooo`030000003oool0oooo01X0oooo000N0?ooo`030?noM018W=l0oooo00/0oooo00<0gia8
07Boo`3oool0;`3oool00`000000oooo0?ooo`0R0?ooo`8000006@3oool00`000000oooo0000000M
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0:
0?ooo`030000003oool0oooo02/0oooo00<0001o0?ooo`3oool05`3oool2000001l0oooo00<0001o
0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`007P3oool00`3o_g@000000000000<0000
000307Boo`3oool0oooo02h0oooo0P00000S0?ooo`8000006@3oool00`000000oooo0000000M0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo02/0oooo00<0001o0?ooo`3oool06@3oool3000001`0oooo00<0001o0?oo
o`3oool0803oool2000001`0oooo001N0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3o
ool0603oool00`000000oooo0000000M0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3o
ool06`3oool00`000000oooo0?ooo`080?ooo`800000;P3oool00`0007l0oooo0?ooo`0L0?ooo`80
00006P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001N0?ooo`030000
003oool0oooo0280oooo00<000000?ooo`3oool0603oool00`000000oooo0000000M0?ooo`030000
003oool0oooo00L0oooo1P00000=0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo01h0oooo0P00000H0?ooo`030000
O`3oool0oooo0240oooo00<000000?ooo`3oool06P3oool005h0oooo00<000000?ooo`3oool08P3o
ool00`000000oooo0?ooo`0H0?ooo`030000003oool0000001d0oooo00<000000?ooo`3oool01P3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool06`3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo02h0oooo00<0001o0?ooo`3oool0803o
ool2000001H0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool0
0`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`0000007@3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0;P3oool0
0`0007l0oooo0?ooo`0R0?ooo`800000503oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0
oooo01X0oooo001N0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0603oool00`00
0000oooo0000000M0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`0_0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool04@3oool00`00
07l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001N0?ooo`8000008`3oool00`000000
oooo0?ooo`0I0?ooo`8000007@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool0<03oool00`0007l0oooo0?ooo`0U0?ooo`8000004@3oool00`0007l0
oooo0?ooo`0P0?ooo`800000703oool005h0oooo00<000000?ooo`3oool08P3oool00`000000oooo
0?ooo`0I0?ooo`8000007@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool0<03oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo00h0oooo
00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo0?oo
o`0R0?ooo`030000003oool0oooo01T0oooo0P00000M0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo01/0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`0`0?ooo`030000O`3oool0oooo02P0oooo00<0
00000?ooo`3oool03@3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001N
0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool06@3oool2000001d0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool06`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0340oooo00<0001o
0?ooo`3oool0:@3oool00`000000oooo0?ooo`0<0?ooo`030000O`3oool0oooo0240oooo00<00000
0?ooo`3oool06P3oool005h0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0I0?oo
o`8000007@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0<@3oool00`0007l0oooo0?ooo`0Z0?ooo`030000003oool0oooo00/0oooo00<0001o0?oo
o`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo0?ooo`0R0?ooo`03
0000003oool0oooo01T0oooo0P00000M0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo02/0oooo00<000000?ooo`3o
ool02P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001N0?ooo`030000
003oool0oooo0280oooo00<000000?ooo`3oool06@3oool2000001d0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
6`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0
;03oool00`000000oooo0?ooo`090?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0
6P3oool005h0oooo0P00000S0?ooo`030000003oool0oooo01T0oooo0P00000M0?ooo`050000003o
ool0oooo0?ooo`0000004@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01/0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0b0?ooo`030000O`3oool0oooo02d0oooo
00<000000?ooo`3oool0203oool00`0007l0oooo0?ooo`0P0?ooo`800000703oool005h0oooo00<0
00000?ooo`3oool08P3oool00`000000oooo0?ooo`0I0?ooo`8000007@3oool01@000000oooo0?oo
o`3oool000000140oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0K0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0^0?ooo`030000
003oool0oooo00L0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3o
ool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01T0oooo0P00000M0?ooo`050000003o
ool0oooo0?ooo`0000004@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01/0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0b0?ooo`030000O`3oool0oooo02l0oooo
00<000000?ooo`3oool01P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo
001>0?ooo`030?oo_`1d000000000080000000<007Bo0?ooo`3oool0203oool00`000000oooo0?oo
o`0R0?ooo`030000003oool0oooo01T0oooo0P00000M0?ooo`050000003oool0oooo0?ooo`000000
4@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0b0?ooo`030000O`3oool0oooo0300oooo00<000000?ooo`3oool0
1@3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001=0?ooo`030?noM000
0000M;oo00<0oooo00<0W4P0001d_`3oool0203oool00`000000oooo0?ooo`0R0?ooo`030000003o
ool0oooo01T0oooo0P00000M0?ooo`050000003oool0oooo0?ooo`0000004P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0b0?ooo`030000O`3oool0oooo0300oooo00<000000?ooo`3oool01@3oool00`0007l0oooo
0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001C0?ooo`030?oo_`1d04P0W=oo00P0oooo00<0
00000?ooo`3oool08P3oool00`000000oooo0?ooo`0I0?ooo`8000007@3oool010000000oooo0?oo
o`00000C0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0380oooo00<0001o0?ooo`3oool0<@3oool00`000000oooo
0?ooo`040?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool06P3oool005@0oooo00<0
_g@007Boo`3oool01`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01T0oooo0P00
000M0?ooo`040000003oool0oooo000001<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0K0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0<P3oool00`0007l0oooo
0?ooo`0b0?ooo`030000003oool0oooo00<0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo
0?ooo`0J0?ooo`00E03oool00`2LB000M;oo0?ooo`070?ooo`<000008P0007l00`000000001o0000
O`0I0000O`8000007@0007l010000000001o0000O`00000C0000O`030000000007l0001o00<0001o
00<000000000O`0007l06`0007l00`000000001o0000O`020000O`030000000007l0001o06T0001o
00<000000000O`0007l0900007l3000001`0oooo001C0?ooo`030?noM000B9`0oooo00P0oooo00<0
00000?ooo`3oool08P3oool00`000000oooo0?ooo`0I0?ooo`8000007@3oool010000000oooo0?oo
o`00000D0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo03<0oooo00<0001o0?ooo`3oool0=03oool01@000000oooo
0?ooo`3oool0001o02<0oooo00<000000?ooo`3oool06P3oool004h0oooo00<0gia8000000000000
0P0000000`0007@0_ooo0?ooo`080?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0
6@3oool2000001d0oooo00@000000?ooo`3oool00000503oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo01/0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0c0?ooo`03
0000O`3oool0oooo03@0oooo00D000000?ooo`3oool0oooo0000O`0S0?ooo`030000003oool0oooo
01X0oooo001>0?ooo`030=nLB000M;l0oooo00d0oooo00<000000?ooo`3oool08P3oool00`000000
oooo0?ooo`0I0?ooo`8000007@3oool010000000oooo0?ooo`00000D0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
03<0oooo00<0001o0?ooo`3oool0=@3oool010000000oooo0?ooo`0007lS0?ooo`030000003oool0
oooo01X0oooo001>0?ooo`030=nLB000M;l0oooo00d0oooo00<000000?ooo`3oool08P3oool00`00
0000oooo0?ooo`0I0?ooo`8000007@3oool010000000oooo0?ooo`00000D0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo03<0oooo00<0001o0?ooo`3oool0=@3oool010000000oooo0?ooo`0007lS0?ooo`030000003o
ool0oooo01X0oooo001>0?ooo`030=nLB000M;l0oooo00d0oooo00<000000?ooo`3oool08P3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01/0oooo00@000000?ooo`3oool00000503o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0c0?ooo`030000O`3oool0oooo03H0oooo00<000000?ooo`0007l08`3o
ool00`000000oooo0?ooo`0J0?ooo`00CP3oool00`3OW4P00000000000030000000304RLg`3oool0
oooo00L0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo01/0oooo00@000000?ooo`3oool000005@3oool01@000000oooo0?ooo`3oool0000001d0oooo
00D000000?ooo`3oool0oooo0000000f0?ooo`030000O`3oool0oooo03L0oooo00<000000000O`3o
ool08P3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool2000002<0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0000001H0oooo00D000000?ooo`3oool0
oooo0000000M0?ooo`050000003oool0oooo0?ooo`000000=P3oool00`0007l0oooo0?ooo`0g0?oo
o`030000000007l0oooo0240oooo0P00000L0?ooo`00GP3oool00`000000oooo0?ooo`0R0?ooo`03
0000003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool00`000000oooo0000000F0?ooo`05
0000003oool0oooo0?ooo`0000007@3oool01@000000oooo0?ooo`3oool0000003H0oooo00<0001o
0?ooo`3oool0>03oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo01X0oooo001N0?oo
o`030000003oool0oooo0280oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?oo
o`030000003oool0000001H0oooo00D000000?ooo`3oool0oooo0000000M0?ooo`050000003oool0
oooo0?ooo`000000=P3oool00`0007l0oooo0?ooo`0h0?ooo`030000003oool0oooo0240oooo00<0
00000?ooo`3oool06P3oool005h0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0J
0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`0000005P3oool01@000000oooo0?ooo`3o
ool0000001d0oooo00D000000?ooo`3oool0oooo0000000f0?ooo`030000O`3oool0oooo03P0oooo
00<0001o0000003oool08@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo0?oo
o`0R0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool00`000000oooo0000
000F0?ooo`050000003oool0oooo0?ooo`0000007@3oool01@000000oooo0?ooo`3oool0000003H0
oooo00<0001o0?ooo`3oool0>03oool00`0007l000000?ooo`0Q0?ooo`030000003oool0oooo01X0
oooo001N0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0K0?ooo`030000003oool0000001H0oooo00D000000?ooo`3oool0oooo0000000M0?ooo`05
0000003oool0oooo0?ooo`000000=P3oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool00000
0240oooo00<000000?ooo`3oool06P3oool005h0oooo00<000000?ooo`3oool08P3oool00`000000
oooo0?ooo`0J0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`0000005P3oool01@000000
oooo0?ooo`3oool0000001d0oooo00D000000?ooo`3oool0oooo0000000f0?ooo`030000O`3oool0
oooo03P0oooo00<0001o0?ooo`0000008@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool20000
02<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool00000
01H0oooo00D000000?ooo`3oool0oooo0000000M0?ooo`050000003oool0oooo0?ooo`000000=P3o
ool00`0007l0oooo0?ooo`0h0?ooo`040000O`3oool0oooo000001l0oooo0P00000L0?ooo`00GP3o
ool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06`3o
ool00`000000oooo0000000F0?ooo`050000003oool0oooo0?ooo`0000007@3oool01@000000oooo
0?ooo`3oool0000003H0oooo00<0001o0?ooo`3oool0>03oool0100007l0oooo0?ooo`00000P0?oo
o`030000003oool0oooo01X0oooo001N0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0000001H0oooo00D000000?ooo`3o
ool0oooo0000000M0?ooo`040000003oool0oooo000003L0oooo00<0001o0?ooo`3oool0>03oool0
1@0007l0oooo0?ooo`3oool0000001l0oooo00<000000?ooo`3oool06P3oool005h0oooo00<00000
0?ooo`3oool08P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01/0oooo00<00000
0?ooo`0000005P3oool01@000000oooo0?ooo`3oool0000001d0oooo00@000000?ooo`3oool00000
=`3oool00`0007l0oooo0?ooo`0h0?ooo`050000O`3oool0oooo0?ooo`0000007`3oool00`000000
oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool06`3oool00`000000oooo0000000F0?ooo`050000003oool0oooo0?ooo`00
00007@3oool010000000oooo0?ooo`00000g0?ooo`030000O`3oool0oooo03P0oooo00D0001o0?oo
o`3oool0oooo0000000O0?ooo`030000003oool0oooo01X0oooo001N0?ooo`030000003oool0oooo
0280oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool00000
01H0oooo00D000000?ooo`3oool0oooo0000000M0?ooo`040000003oool0oooo000003L0oooo00<0
001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`020?ooo`030000003oool0oooo01`0oooo00<0
00000?ooo`3oool06P3oool005h0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0J
0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`0000005`3oool010000000oooo0?ooo`00
000M0?ooo`040000003oool0oooo000003L0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo
0?ooo`020?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool06P3oool005h0oooo0P00
000S0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool00`000000oooo0000
000G0?ooo`040000003oool0oooo000001d0oooo00@000000?ooo`3oool00000=`3oool00`0007l0
oooo0?ooo`0h0?ooo`030000O`3oool0oooo00<0oooo00<000000?ooo`3oool06P3oool2000001`0
oooo001N0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0K0?ooo`030000003oool0000001L0oooo00@000000?ooo`3oool000007@3oool010000000
oooo0?ooo`00000g0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool00`3oool00`00
0000oooo0?ooo`0K0?ooo`030000003oool0oooo01X0oooo001N0?ooo`030000003oool0oooo0280
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0000001L0
oooo00@000000?ooo`3oool000007@3oool010000000oooo0?ooo`00000g0?ooo`030000O`3oool0
oooo03P0oooo00<0001o0?ooo`3oool0103oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo01X0oooo001N0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`0K0?ooo`030000003oool0000001L0oooo00@000000?ooo`3oool000007@3oool0
10000000oooo0?ooo`00000g0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool0103o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo001N0?ooo`030000003oool0
oooo0280oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0
000001L0oooo00@000000?ooo`3oool000007@3oool010000000oooo0?ooo`00000g0?ooo`030000
O`3oool0oooo03P0oooo00<0001o0?ooo`3oool0103oool00`000000oooo0?ooo`0J0?ooo`030000
003oool0oooo01X0oooo001N0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0K0?ooo`030000003oool0000001L0oooo00@000000?ooo`3oool00000
7@3oool010000000oooo0?ooo`00000g0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3o
ool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo001N0?ooo`800000
8`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`000000
5`3oool010000000oooo0?ooo`00000M0?ooo`040000003oool0oooo000003L0oooo00<0001o0?oo
o`3oool0>03oool00`0007l0oooo0?ooo`050?ooo`030000003oool0oooo01P0oooo0P00000L0?oo
o`00GP3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3o
ool06`3oool00`000000oooo0000000G0?ooo`040000003oool0oooo000001d0oooo00@000000?oo
o`3oool00000=`3oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo00D0oooo00<00000
0?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo0?ooo`0R0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool00`000000oooo0000000G0?oo
o`040000003oool0oooo000001d0oooo00@000000?ooo`3oool00000=`3oool00`0007l0oooo0?oo
o`0h0?ooo`030000O`3oool0oooo00H0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`0J0?ooo`00@`3oool00`3ookl0M000000000050000000307Boo`3oool0oooo0080oooo00D0okmd
000000000000000004RLg`090?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0K0?ooo`030000003oool0000001L0oooo00@000000?ooo`3oool00000
7@3oool010000000oooo0?ooo`00000g0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3o
ool01P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01X0oooo00160?ooo`030?oo
_`1d04P0W=oo00H0oooo00<0omnL04P0M02oool00P3oool00`2LB000M;oo0?ooo`070?ooo`030000
003oool0oooo0280oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`800000
603oool010000000oooo0?ooo`00000M0?ooo`040000003oool0oooo000003L0oooo00<0001o0?oo
o`3oool0>03oool00`0007l0oooo0?ooo`060?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool06P3oool004H0oooo00<0oono07@0B02Lgol01P3oool00`2oM000M;oo0?ooo`020?ooo`03
0?oOW018B9`0oooo00L0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo01/0oooo0P00000H0?ooo`040000003oool0oooo000001d0oooo00<000000?oo
o`000000>03oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo00L0oooo00<000000?oo
o`3oool05`3oool00`000000oooo0?ooo`0J0?ooo`00AP3oool00`3ookl0M01809cOo`050?ooo`03
0?oo_`1d07@0_ooo00<0oooo00<0oono07@0B02Lgol01`3oool00`000000oooo0?ooo`0R0?ooo`03
0000003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool2000001P0oooo00@000000?ooo`3o
ool000007@3oool00`000000oooo0000000h0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?oo
o`3oool01`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01X0oooo00160?ooo`03
0?oo_`1d04P0W=oo00D0oooo00<0omnL04P0M02oool0103oool00`2LB000M;oo0?ooo`060?ooo`<0
00008P0007l00`000000001o0000O`0J0000O`030000000007l0001o01/0001o0P00000H0000O`04
0000000007l0001o000001d0001o00<000000000O`000000O@0007l00`000000001o0000O`0E0000
O`<00000703oool004H0oooo00<0oono07@0B02Lgol01@3oool00`3ogi`0B01d0;ooo`040?ooo`03
09a8001d_ol0oooo00H0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo01/0oooo0P00000I0?ooo`030000003oool0000001d0oooo00<000000?ooo`00
0000>03oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo00P0oooo00<000000?ooo`3o
ool05P3oool00`000000oooo0?ooo`0J0?ooo`00AP3oool00`3ookl0M01809cOo`050?ooo`030?oO
W01807@0_ooo00@0oooo00<0W4P007Boo`3oool01P3oool00`000000oooo0?ooo`0R0?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool2000001T0oooo00<000000?ooo`000000
7@3oool00`000000oooo0000000h0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool0
203oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01X0oooo00160?ooo`030?oo_`1d
04P0W=oo00D0oooo00<0oono07@0M02oool0103oool00`2LB4P0W=oo0?ooo`060?ooo`030000003o
ool0oooo0280oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`8000006@3o
ool00`000000oooo0000000M0?ooo`030000003oool0000003P0oooo00<0001o0?ooo`3oool0>03o
ool00`0007l0oooo0?ooo`080?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool06P3o
ool004H0oooo00<0oono07@0B02Lgol01P3oool00`2LB4P0W=oo0?ooo`020?ooo`030?oo_`1dB9`0
oooo00L0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo01/0oooo0P00000I0?ooo`030000003oool0000001d0oooo00<000000?ooo`000000>03oool0
0`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo00T0oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`0J0?ooo`00@`3oool01P3ogi`0B00000000018W9`0M01809cOo`H0oooo00<0
omnL04P0B02Lgol00P3oool00`2LB000M;oo0?ooo`070?ooo`030000003oool0oooo0280oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`8000006@3oool00`000000oooo0000
000M0?ooo`030000003oool0000003P0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?oo
o`090?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool06P3oool004H0oooo00<0W4P0
0000B02Lgol01`3oool01@3ogi`0B000000000000000B9cO00T0oooo00<000000?ooo`3oool08P3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01/0oooo0P00000I0?ooo`030000003o
ool0000001d0oooo00<000000?ooo`000000>03oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3o
ool0oooo00T0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool2
000002<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`8000006@3oool0
0`000000oooo0000000M0?ooo`030000003oool0000003P0oooo00<0001o0?ooo`3oool0>03oool0
0`0007l0oooo0?ooo`0:0?ooo`030000003oool0oooo01<0oooo0P00000L0?ooo`00GP3oool00`00
0000oooo0?ooo`0R0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool20000
01T0oooo00<000000?ooo`0000007@3oool00`000000oooo0000000h0?ooo`030000O`3oool0oooo
03P0oooo00<0001o0?ooo`3oool02P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo
01X0oooo001N0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool06P3oool00`000000
oooo0?ooo`0K0?ooo`8000006@3oool00`000000oooo0000000M0?ooo`030000003oool0000003P0
oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0:0?ooo`030000003oool0oooo01@0
oooo00<000000?ooo`3oool06P3oool005h0oooo00<000000?ooo`3oool08P3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo01/0oooo0P00000I0?ooo`030000003oool0000001d0oooo
00<000000?ooo`000000>03oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo00/0oooo
00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo0?oo
o`0R0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool2000001T0oooo00<0
00000?ooo`0000007@3oool00`000000oooo0000000h0?ooo`030000O`3oool0oooo03P0oooo00<0
001o0?ooo`3oool02`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01X0oooo001N
0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K
0?ooo`8000006@3oool00`000000oooo0000000M0?ooo`030000003oool0000003P0oooo00<0001o
0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0;0?ooo`030000003oool0oooo01<0oooo00<00000
0?ooo`3oool06P3oool005h0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0J0?oo
o`030000003oool0oooo01/0oooo0P00000I0?ooo`030000003oool0000001d0oooo00<000000?oo
o`000000>03oool00`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo00/0oooo00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`0J0?ooo`00GP3oool2000002<0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0K0?ooo`8000006@3oool00`000000oooo0000000M0?ooo`03
0000003oool0000003P0oooo00<0001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0<0?ooo`03
0000003oool0oooo0140oooo0P00000L0?ooo`00GP3oool00`000000oooo0?ooo`0R0?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool2000001T0oooo00<000000?ooo`000000
7@3oool00`000000oooo0000000h0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool0
303oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01X0oooo001N0?ooo`030000003o
ool0oooo0280oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`8000006@3o
ool00`000000oooo0000000M0?ooo`030000003oool0000003P0oooo00<0001o0?ooo`3oool0>03o
ool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool06P3o
ool005h0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo01/0oooo0P00000I0?ooo`030000003oool0000001d0oooo00<000000?ooo`000000>03oool0
0`0007l0oooo0?ooo`0h0?ooo`030000O`3oool0oooo00d0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0J0?ooo`00GP3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool06`3oool2000001T0oooo00<000000?ooo`0000007@3oool00`00
0000oooo0000000h0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool03@3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo01X0oooo001N0?ooo`030000003oool0oooo0280
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`8000006@3oool00`000000
oooo0000000M0?ooo`030000003oool0000003P0oooo00<0001o0?ooo`3oool0>03oool00`0007l0
oooo0?ooo`0=0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool06P3oool005h0oooo
00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01/0oooo
0P00000I0?ooo`030000003oool0000001d0oooo00<000000?ooo`000000>03oool00`0007l0oooo
0?ooo`0h0?ooo`030000O`3oool0oooo00h0oooo00<000000?ooo`3oool0403oool00`000000oooo
0?ooo`0J0?ooo`00GP3oool2000002<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0K0?ooo`8000006@3oool00`000000oooo0000000M0?ooo`030000003oool0000003P0oooo00<0
001o0?ooo`3oool0>03oool00`0007l0oooo0?ooo`0>0?ooo`030000003oool0oooo00l0oooo0P00
000L0?ooo`00GP3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01X0oooo00<00000
0?ooo`3oool06`3oool2000001T0oooo00<000000?ooo`0000007@3oool00`000000oooo0000000h
0?ooo`030000O`3oool0oooo03P0oooo00<0001o0?ooo`3oool03P3oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0oooo01X0oooo001N0?ooo`030000003oool0oooo0280oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0K0?ooo`8000006@3oool00`000000oooo0000000M0?oo
o`030000003oool0000003P0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo0100oooo00<000000?ooo`3oool06P3oool005h0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3o
ool0303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3o
ool0303oool2000000d0oooo00<000000?ooo`3oool02@3oool00`000000oooo0000000>0?ooo`03
0000003oool0oooo00`0oooo00<000000?ooo`0000002`3oool00`000000oooo0?ooo`0<0?ooo`03
0000003oool0oooo00`0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo00`0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0<0?ooo`03
0000003oool0oooo00/0oooo00D000000?ooo`3oool0oooo0000000:0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool06P3oool005h0ooooo`00000g000001`0oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
002_0?ooo`030?oo_`1d0000000000<0000000<004RL0?ooo`3oool00P3oool02P2LB00000000000
00000000B9cO0=nLB000000004RL0?oo_`1d000200000007001d_`3oool0W4P000000018W=l0oono
07@000040000000;001d_`3oool0W4P000000018W=l0oono07@000000000M;oo0?ooo`2LB0000`00
00001018W=l0gia800000000B9`<0?ooo`030?noM000000000000080000000<0001809cOo`3oool0
103oool02`3ogi`0B00000000000000004RL0?oo_`1d0000000007Boo`3ookl0M0000080000000<0
04RL0?ooo`3OW4P00`0000001@1d_ol0oooo0?ooo`3oool0_g@000@0000000<007Bo0?ooo`3oool0
0`3oool01P3ogi`0B000000000000000001809cOo`<0oooo00<0gia80000000000001@0000001P18
W=l0oooo0?ooo`3oool0oono07@000<0000000H0001d0;ooo`3oool0oooo0?oo_`1d000200000003
0018W03oool0gia800<0000000<0M;oo0?ooo`3oool0J03oool00:h0oooo00<0oono07@0001d_ol0
103oool01P3OW4P004RL0?ooo`3oool0gia80018W0<0oooo00<0W4P000000018W=l00`3oool03@2o
M000M;oo0?ooo`3oool0W4Q809cOo`3oool0oono07@0M02oool0oooo0;md001d_ol00P3oool02@2L
B4P0W=oo0?ooo`3ookl0M01d0;ooo`3oool0gia80018W0030?ooo`0309a800000000B9cO00h0oooo
00<0oono07@0M02oool0203oool0202LB000B9cO0?ooo`3oool0gia80000000007@0_ooo0`3oool0
0`2oM000B9cO0?ooo`020?ooo`030?oo_`1d07@0_ooo00<0oooo00<0W4P0001d_`3oool00`3oool0
0`2oM000001d0;ooo`030?ooo`0709a80018W=l0oooo0?ooo`3ogi`0B000001d_`050?ooo`030?oo
_`1d04P0W=oo00D0oooo00<0oono07@0001d_ol00`3oool00`3o_g@0001d0;ooo`030?ooo`030;md
0018W=l0oooo0080oooo00<0oono07@0M02oool0J`3oool00:h0oooo00<0_g@004RLg`3oool0103o
ool01P3o_g@004RL0?ooo`3oool0_g@007Boo`@0oooo00<0gia804RLg`3oool00P3oool03@2oM000
M;oo0?ooo`3oool0W4Q809cOo`3oool0oono07@0M02oool0oooo0;md001d_ol00P3oool02@2LB4P0
W=oo0?ooo`3ookl0M01d0;ooo`3oool0_g@007Boo`040?ooo`030=nLB018W=l0oooo00d0oooo00<0
oono07@0M02oool01`3oool00`3ookl0M01809cOo`030?ooo`030?oOW01807@0_ooo00<0oooo00<0
_g@004RLg`3oool00P3oool00`3ookl0M01d0;ooo`020?ooo`030?noM000B9`0oooo00T0oooo00<0
oono07@0B02Lgol02P3oool00`3ookl0M01809cOo`050?ooo`030;md0018W=l0oooo00@0oooo00<0
gia8001d_`3oool00P3oool00`2oM000B9cO0?ooo`020?ooo`030?oo_`1d07@0_ooo06/0oooo002]
0?ooo`030?oo_`1d04P0W=oo00D0oooo00L0okmd0018W03oool0oooo0?noM00004P0W=oo00<0oooo
00<0gia804RLg`3oool00P3oool03@2oM000M;oo0?ooo`3oool0W4Q809cOo`3oool0oono07@0M02o
ool0oooo0;md001d_ol00P3oool02P2LB4P0W=oo0?ooo`3ookl0M01d0;ooo`3oool0okmd0000B02L
gol30?ooo`030=nLB018W=l0oooo00d0oooo00<0oono07@0M02oool01`3oool00`3ookl0M01809cO
o`030?ooo`030?oo_`1d07@0_ooo00<0oooo00<0_g@004RLg`3oool00P3oool00`3ookl0M01d0;oo
o`020?ooo`030;md0018W=l0oooo00T0oooo00<0oono07@0B02Lgol02P3oool00`3ookl0M01809cO
o`050?ooo`0309a8B02Lgol0oooo00@0oooo00<0omnL04Q8W03oool00P3oool00`2oM000B9cO0?oo
o`020?ooo`030?oo_`1d07@0_ooo06/0oooo002]0?ooo`030?oo_`1d07@0_ooo0080oooo00<0gia8
0000000000000P0000001018W=l0oooo0?ooo`3o_g@50000000604RLg`3oool0oooo0?ooo`2oM000
M;oo0P3oool02@2LB4P0W=oo0?ooo`3ookl0M01d0;ooo`3oool0_g@007Boo`020?ooo`0609a8B02L
gol0oooo0?oo_`1d07@0_ooo0P3oool00`3o_g@00000000000030000000304RLg`3oool0oooo00`0
oooo00H0oono07@0M02oool0oooo0;md001d_ol40?ooo`030?oo_`1d04P0W=oo00<0oooo00<0oono
07@0M02oool00`3oool00`2oM000B9cO0?ooo`020?ooo`030?oo_`1d07@0_ooo0080oooo00<0_g@0
04RLg`3oool02@3oool00`3ookl0M01809cOo`0:0?ooo`030?oo_`1d04P0W=oo00D0oooo00<0W4Q8
09cOo`3oool0103oool00`3ogi`0B4RL0?ooo`020?ooo`030;md0018W=l0oooo0080oooo00<0oono
07@0M02oool0J`3oool00:d0oooo00<0oono07@0M02oool03`3oool00`3OW4P0B9cO0?ooo`020?oo
o`0=0;md0018W=l0oooo0?ooo`2LB4P0W=oo0?ooo`3ookl0M01d0;ooo`3oool0_g@004RLg`020?oo
o`0609a8B02Lgol0oooo0?oo_`1d07@0_ooo1`3oool00`3OW4P0B9cO0?ooo`0=0?ooo`030?oo_`1d
000000000080000000<0B9cO0?ooo`3oool00P3oool00`3ookl0M01809cOo`030?ooo`030?oo_`1d
07@0_ooo00<0oooo00<0_g@00000M02oool00P3oool00`3ogi`0B01d0;ooo`020?ooo`030?noM000
M;l0oooo00@0oooo00<0okmd001d_`3oool00P3oool00`3ookl0M01809cOo`0:0?ooo`030?oo_`1d
04P0W=oo00D0oooo00<0_g@004RLg`3oool0103oool00`3OW4P007Bo0?ooo`020?ooo`030;md0000
07@0_ooo0080oooo00<0omnL04P0M02oool0J`3oool00:h0oooo00<0W4Q809cOo`3oool02@3oool0
1`2o_kl0oooo0?ooo`3oool0oono07@0001d_ol00`3oool05@2oM000000007Boo`3ogi`0B0000000
M02oool0okmd0000M02oool0oooo0;md00000000M;oo0?oOW0180000001d0;ooo`3o_g@0001d0;oo
o`020?ooo`070;no_`3oool0oooo0?ooo`3ookl0M00007Boo`0>0?ooo`060?oo_`1d07@0_ooo0?oo
o`2oM000B9cO103oool00`3ookl0M01809cOo`030?ooo`030?oo_`1d07@0_ooo00<0oooo00<0_g@0
04Q8B000B9`00P3oool00`2oM00004RL0?ooo`030?ooo`030;md0000B9`0oooo0080oooo00<0okmd
00000000M;l00`3oool00`3ookl0M01809cOo`0:0?ooo`030?oo_`1d04P0W=oo00D0oooo00<0oono
07@00000M;l00`3oool00`2oM000001d0;ooo`030?ooo`030;md0018B4P004RL0080oooo00<0_g@0
0018W03oool0J`3oool00:h0oooo00<0gia804RLg`3oool0103oool00`3o_g@004RL0?ooo`020?oo
o`030=nLB000000000000080000000/0M;oo0?ooo`3oool0oono07@000000000M;no07@000000000
B9cO0=nLB0020000000907Boo`3ookl0M0000000001d_kl0M00000000018W=l0gia80080000000D0
M;oo0?ooo`3oool0oooo0=nLB0040000000307Boo`3oool0oooo00d0oooo00<0oono07@0M02oool0
103oool04P3OW4P0B9cO0?noM0000000001809cOo`3oool0oono07@000000000001d0;ooo`3oool0
oono07@000000000B9cO0=nLB080000000<0001d0;ooo`3oool0103oool02P3OW4P0000000000000
0000001d0;noM000M;l0oooo0?oo_`1d000600000003001d_`3oool0oooo0080oooo00D0_g@00000
00000000001809cOo`070?ooo`050;md00000000000000000000M;l00`3oool01@3ookl0M0000000
0018W=l0gia80080000000<0001d0;ooo`3oool0J`3oool00:l0oooo00<0W4P0001d_`3oool00P3o
ool00`3OW4P000000018W00i0?ooo`030?oo_`1d07@0_ooo00@0oooo00<0gia804RLg`3oool08P3o
ool00`3ookl0M01809cOo`2A0?ooo`00/03oool01`2oM000000000000000000004RL0?noM000B9`0
>03oool00`3o_g@00000000000060000000304RLg`3oool0oooo0240oooo00<0oono07@0B02Lgol0
T@3oool00?l0oooo:P3oool00`3o_g@0001d0;ooo`240?ooo`00o`3ooolZ0?ooo`030?noM00007@0
_ooo08@0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-8.45768, -21.4442, \
0.0509096, 0.219851}}]
}, Closed]],

Cell[TextData[{
  "Among the options available to ",
  StyleBox["TextStyle", "InlineInput",
    FontWeight->"Bold"],
  " are:"
}], "Text"],

Cell[TextData[{
  " ",
  StyleBox["FontFamily", "InlineInput",
    FontWeight->"Bold"],
  " ",
  StyleBox["\[LongDash] ",
    FontWeight->"Bold"],
  "type of font to apply.  Common choices are \"Courier\", \"Times\", and \
\"Helvetica\".  More specific font names can be given also, with blanks \
replaced by dashes, such as \"Times-New-Roman\"."
}], "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
  StyleBox["FontWeight", "InlineInput",
    FontWeight->"Bold"],
  " \[LongDash] \"Plain\", \"Bold\", or \"ExtraBold\""
}], "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
  StyleBox["FontSlant", "InlineInput",
    FontWeight->"Bold"],
  " \[LongDash] \"Plain\" or \"Italic\" "
}], "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
  StyleBox["FontSize", "InlineInput",
    FontWeight->"Bold"],
  " \[LongDash] in printer's points"
}], "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
  StyleBox["FontColor", "InlineInput",
    FontWeight->"Bold"],
  " \[LongDash] specified using ",
  ButtonBox["GreyLevel",
    ButtonStyle->"RefGuideLink"],
  ",  ",
  ButtonBox["RBGColor",
    ButtonStyle->"RefGuideLink"],
  " , ",
  ButtonBox["CMYKColor",
    ButtonStyle->"RefGuideLink"],
  ", or ",
  ButtonBox["Hue",
    ButtonStyle->"RefGuideLink"]
}], "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
  StyleBox["Background", "InlineInput",
    FontWeight->"Bold"],
  " \[LongDash] ",
  StyleBox["specified using ",
    FontWeight->"Plain"],
  StyleBox["GreyLevel", "InlineInput",
    FontWeight->"Bold"],
  ",",
  StyleBox[" ",
    FontWeight->"Plain"],
  StyleBox["RGBColor", "InlineInput",
    FontWeight->"Bold"],
  ", ",
  StyleBox["CMYKColor", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[",",
    FontWeight->"Bold"],
  StyleBox[" or ",
    FontWeight->"Plain"],
  StyleBox["Hue", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[" ",
    FontWeight->"Plain"]
}], "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell[BoxData[
    \(\($TextStyle = {FontFamily \[Rule] "\<Courier\>", 
          FontSize \[Rule] 16, \ FontSlant \[Rule] "\<Italic\>", \ 
          FontWeight \[Rule] "\<Bold\>"};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plotGamma;\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.0952381 0.316178 0.0220537 [
[.11905 -0.0125 -12.0313 -20.25 ]
[.11905 -0.0125 12.0313 0 ]
[.30952 -0.0125 -12.0938 -20.25 ]
[.30952 -0.0125 12.0938 0 ]
[.5 -0.0125 -6.8125 -20.25 ]
[.5 -0.0125 6.8125 0 ]
[.69048 -0.0125 -6.84375 -20.25 ]
[.69048 -0.0125 6.84375 0 ]
[.88095 -0.0125 -6.78125 -20.25 ]
[.88095 -0.0125 6.78125 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -7.09375 -40.5 ]
[.5 -0.0125 7.09375 -20.25 ]
[-0.0125 .09564 -33.6875 -10.125 ]
[-0.0125 .09564 0 10.125 ]
[-0.0125 .20591 -24.4375 -10.125 ]
[-0.0125 .20591 0 10.125 ]
[-0.0125 .31618 -13.625 -10.125 ]
[-0.0125 .31618 0 10.125 ]
[-0.0125 .42645 -13.9375 -10.125 ]
[-0.0125 .42645 0 10.125 ]
[-0.0125 .53672 -23.1875 -10.125 ]
[-0.0125 .53672 0 10.125 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -53.9375 -22.4688 ]
[-0.01875 .30902 -33.6875 22.4688 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -69.0625 0 ]
[.5 .63053 69.0625 20.25 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.11905 0 m
.11905 .61803 L
s
.30952 0 m
.30952 .61803 L
s
.5 0 m
.5 .61803 L
s
.69048 0 m
.69048 .61803 L
s
.88095 0 m
.88095 .61803 L
s
0 .09564 m
1 .09564 L
s
0 .20591 m
1 .20591 L
s
0 .31618 m
1 .31618 L
s
0 .42645 m
1 .42645 L
s
0 .53672 m
1 .53672 L
s
0 g
.11905 0 m
.11905 .00625 L
s
gsave
.11905 -0.0125 -73.0313 -24.25 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 28.25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 18.438 moveto
%%IncludeResource: font Mathematica1Mono-Bold
%%IncludeFont: Mathematica1Mono-Bold
/Mathematica1Mono-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
73.500 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.30952 0 m
.30952 .00625 L
s
gsave
.30952 -0.0125 -73.0938 -24.25 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 28.25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 18.438 moveto
%%IncludeResource: font Mathematica1Mono-Bold
%%IncludeFont: Mathematica1Mono-Bold
/Mathematica1Mono-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
73.500 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -67.8125 -24.25 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 28.25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.69048 0 m
.69048 .00625 L
s
gsave
.69048 -0.0125 -67.8438 -24.25 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 28.25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.88095 0 m
.88095 .00625 L
s
gsave
.88095 -0.0125 -67.7813 -24.25 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 28.25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.2619 0 m
.2619 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.7381 0 m
.7381 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.07143 0 m
.07143 .00375 L
s
.02381 0 m
.02381 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.97619 0 m
.97619 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -68.0938 -44.5 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 28.25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
73.188 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .09564 m
.00625 .09564 L
s
gsave
-0.0125 .09564 -94.6875 -14.125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 28.25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 18.438 moveto
%%IncludeResource: font Mathematica1Mono-Bold
%%IncludeFont: Mathematica1Mono-Bold
/Mathematica1Mono-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
73.500 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
0 .20591 m
.00625 .20591 L
s
gsave
-0.0125 .20591 -85.4375 -14.125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 28.25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 18.438 moveto
%%IncludeResource: font Mathematica1Mono-Bold
%%IncludeFont: Mathematica1Mono-Bold
/Mathematica1Mono-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
73.500 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .31618 m
.00625 .31618 L
s
gsave
-0.0125 .31618 -74.625 -14.125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 28.25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .42645 m
.00625 .42645 L
s
gsave
-0.0125 .42645 -74.9375 -14.125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 28.25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .53672 m
.00625 .53672 L
s
gsave
-0.0125 .53672 -84.1875 -14.125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 28.25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .1177 m
.00375 .1177 L
s
0 .13975 m
.00375 .13975 L
s
0 .1618 m
.00375 .1618 L
s
0 .18386 m
.00375 .18386 L
s
0 .22796 m
.00375 .22796 L
s
0 .25002 m
.00375 .25002 L
s
0 .27207 m
.00375 .27207 L
s
0 .29412 m
.00375 .29412 L
s
0 .33823 m
.00375 .33823 L
s
0 .36029 m
.00375 .36029 L
s
0 .38234 m
.00375 .38234 L
s
0 .40439 m
.00375 .40439 L
s
0 .4485 m
.00375 .4485 L
s
0 .47055 m
.00375 .47055 L
s
0 .49261 m
.00375 .49261 L
s
0 .51466 m
.00375 .51466 L
s
0 .07359 m
.00375 .07359 L
s
0 .05153 m
.00375 .05153 L
s
0 .02948 m
.00375 .02948 L
s
0 .00743 m
.00375 .00743 L
s
0 .55877 m
.00375 .55877 L
s
0 .58082 m
.00375 .58082 L
s
0 .60288 m
.00375 .60288 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -127.281 -14.125 Mabsadd m
1 1 Mabs scale
currentpoint translate
83.4688 14.125 translate
90 rotate
-83.4688 -14.125 translate
0 28.25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
%%BeginResource: font Courier-BoldItalic-MISO
%%BeginFont: Courier-BoldItalic-MISO
/Courier-BoldItalic /Courier-BoldItalic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-BoldItalic-MISO
%%IncludeFont: Courier-BoldItalic-MISO
/Courier-BoldItalic-MISO findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 18.438 moveto
%%IncludeResource: font Mathematica1Mono-Bold
%%IncludeFont: Mathematica1Mono-Bold
/Mathematica1Mono-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(G) show
%%IncludeResource: font Mathematica2Mono-Bold
%%IncludeFont: Mathematica2Mono-Bold
/Mathematica2Mono-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
74.625 18.438 moveto
(@) show
84.188 18.438 moveto
%%IncludeResource: font Courier-BoldItalic-MISO
%%IncludeFont: Courier-BoldItalic-MISO
/Courier-BoldItalic-MISO findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
%%IncludeResource: font Mathematica2Mono-Bold
%%IncludeFont: Mathematica2Mono-Bold
/Mathematica2Mono-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
94.375 18.438 moveto
(D) show
103.938 18.438 moveto
%%IncludeResource: font Courier-BoldItalic-MISO
%%IncludeFont: Courier-BoldItalic-MISO
/Courier-BoldItalic-MISO findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.11905 .61178 m
.11905 .61803 L
s
.30952 .61178 m
.30952 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.69048 .61178 m
.69048 .61803 L
s
.88095 .61178 m
.88095 .61803 L
s
.125 Mabswid
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.2619 .61428 m
.2619 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.7381 .61428 m
.7381 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.07143 .61428 m
.07143 .61803 L
s
.02381 .61428 m
.02381 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.97619 .61428 m
.97619 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .09564 m
1 .09564 L
s
.99375 .20591 m
1 .20591 L
s
.99375 .31618 m
1 .31618 L
s
.99375 .42645 m
1 .42645 L
s
.99375 .53672 m
1 .53672 L
s
.125 Mabswid
.99625 .1177 m
1 .1177 L
s
.99625 .13975 m
1 .13975 L
s
.99625 .1618 m
1 .1618 L
s
.99625 .18386 m
1 .18386 L
s
.99625 .22796 m
1 .22796 L
s
.99625 .25002 m
1 .25002 L
s
.99625 .27207 m
1 .27207 L
s
.99625 .29412 m
1 .29412 L
s
.99625 .33823 m
1 .33823 L
s
.99625 .36029 m
1 .36029 L
s
.99625 .38234 m
1 .38234 L
s
.99625 .40439 m
1 .40439 L
s
.99625 .4485 m
1 .4485 L
s
.99625 .47055 m
1 .47055 L
s
.99625 .49261 m
1 .49261 L
s
.99625 .51466 m
1 .51466 L
s
.99625 .07359 m
1 .07359 L
s
.99625 .05153 m
1 .05153 L
s
.99625 .02948 m
1 .02948 L
s
.99625 .00743 m
1 .00743 L
s
.99625 .55877 m
1 .55877 L
s
.99625 .58082 m
1 .58082 L
s
.99625 .60288 m
1 .60288 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .31618 m
1 .31618 L
s
.5 0 m
.5 .61803 L
s
gsave
.5 .63053 -130.063 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 28.25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Gamma) show
120.375 18.438 moveto
(Function) show
197.125 18.438 moveto
%%IncludeResource: font Courier-BoldItalic
%%IncludeFont: Courier-BoldItalic
/Courier-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02498 0 m
.02499 .30097 L
.02605 .30806 L
.02729 .31081 L
.02846 .31207 L
.02954 .31277 L
.03053 .31322 L
.03163 .31358 L
.03279 .31386 L
.03395 .31407 L
.0352 .31425 L
.03746 .31449 L
.03884 .31459 L
.04016 .31468 L
.04262 .31479 L
.045 .31487 L
.0463 .31491 L
.04753 .31493 L
.04868 .31495 L
.04993 .31497 L
.05105 .31499 L
.0521 .315 L
.05335 .315 L
.05401 .31501 L
.05472 .31501 L
.05546 .31501 L
.05617 .31501 L
.05749 .31501 L
.05872 .31501 L
.05986 .31501 L
.06111 .315 L
.06244 .31499 L
.06504 .31496 L
.06742 .31493 L
.07003 .31488 L
.07277 .31482 L
.07741 .31469 L
.08247 .31447 L
.08537 .31431 L
.08801 .31413 L
.09085 .3139 L
.09392 .31357 L
.09678 .31317 L
.09941 .31269 L
.10192 .31208 L
.10319 .31169 L
.10458 .31119 L
.10589 .31061 L
.10711 .30996 L
.10837 .30912 L
Mistroke
.10951 .30816 L
.11064 .30696 L
.11169 .30549 L
.11281 .30338 L
.11402 .30005 L
.11524 .29447 L
.11594 .28934 L
.11659 .28193 L
.1179 .24147 L
.11863 .10742 L
Mfstroke
.11863 .10742 m
.11924 .61803 L
s
.1195 .61803 m
.1205 .37806 L
.12179 .3495 L
.12302 .3397 L
.12415 .33484 L
.1252 .33195 L
.12634 .32978 L
.1274 .32827 L
.12838 .32721 L
.13065 .32545 L
.13179 .32481 L
.13304 .32425 L
.1354 .32343 L
.13793 .32282 L
.14031 .32241 L
.14247 .32213 L
.14479 .32191 L
.1461 .32181 L
.14733 .32174 L
.14863 .32168 L
.14981 .32164 L
.15047 .32162 L
.15118 .32161 L
.15247 .32159 L
.15358 .32158 L
.15479 .32159 L
.15607 .3216 L
.15726 .32163 L
.15837 .32166 L
.15939 .32169 L
.16169 .32179 L
.1642 .32194 L
.16653 .32212 L
.1693 .32239 L
.17182 .32269 L
.17698 .32351 L
.17982 .32411 L
.1825 .32481 L
.18712 .3264 L
.18975 .32761 L
.19221 .32905 L
.1948 .33099 L
.19716 .33331 L
.19847 .33491 L
.19989 .33701 L
.20123 .3394 L
.20249 .34218 L
.20366 .34535 L
.20492 .3497 L
.20607 .35483 L
Mistroke
.20712 .36102 L
.20827 .37022 L
.20952 .38546 L
.21083 .41327 L
.21206 .46886 L
Mfstroke
.21206 .46886 m
.21308 .61803 L
s
.21321 .61803 m
.21324 0 L
s
.21552 0 m
.21634 .14124 L
.2175 .20237 L
.21873 .23237 L
.21978 .24731 L
.22094 .25829 L
.22215 .26631 L
.22329 .27179 L
.22453 .27638 L
.22586 .28015 L
.22725 .28324 L
.22857 .28557 L
.23104 .28886 L
.23335 .29108 L
.23548 .29262 L
.23782 .2939 L
.2403 .2949 L
.24277 .29562 L
.24416 .29593 L
.24543 .29615 L
.24661 .29631 L
.2479 .29645 L
.24911 .29653 L
.25023 .29658 L
.25092 .29659 L
.25167 .29659 L
.25237 .29658 L
.25301 .29656 L
.25424 .2965 L
.25553 .29639 L
.25692 .29624 L
.25821 .29606 L
.26116 .2955 L
.26377 .29484 L
.26619 .29406 L
.26879 .29303 L
.27165 .29162 L
.27646 .28846 L
.2791 .28618 L
.28157 .28356 L
.28396 .28046 L
.28624 .27684 L
.28831 .27281 L
.29051 .26745 L
.2929 .25995 L
.29511 .25064 L
.29647 .24331 L
.29775 .23481 L
.299 .22445 L
.30014 .21255 L
Mistroke
.3013 .19698 L
.30254 .17446 L
.3037 .14491 L
.30479 .10353 L
.30606 .02245 L
Mfstroke
.30606 .02245 m
.30623 0 L
s
.30951 0 m
.30959 .61803 L
s
.31325 .61803 m
.31377 .5745 L
.31452 .53772 L
.31521 .51221 L
.31648 .47896 L
.31763 .45783 L
.31895 .43994 L
.32021 .4272 L
.32245 .41078 L
.32358 .40457 L
.32482 .39886 L
.32701 .39091 L
.32907 .38525 L
.33149 .38014 L
.3337 .37656 L
.33625 .3734 L
.33868 .37113 L
.33993 .3702 L
.34127 .36935 L
.34246 .36872 L
.34372 .36817 L
.34485 .36777 L
.3461 .36742 L
.34723 .3672 L
.34828 .36705 L
.34892 .367 L
.34961 .36696 L
.35087 .36697 L
.35209 .36706 L
.35322 .36722 L
.35444 .36746 L
.35579 .36783 L
.35851 .36888 L
.3609 .37016 L
.36308 .37164 L
.36544 .37362 L
.36801 .37629 L
.37235 .38225 L
.37471 .38649 L
.37691 .39123 L
.37937 .39769 L
.38197 .40627 L
.38483 .41853 L
.38745 .43369 L
.3899 .45292 L
.39219 .47794 L
.3935 .49701 L
.3947 .51886 L
.39609 .55172 L
.3974 .59444 L
Mistroke
Mfstroke
.3974 .59444 m
.39792 .61803 L
s
.40444 .61803 m
.40452 0 L
s
.41171 0 m
.41218 .0211 L
.41288 .04545 L
.41353 .0644 L
.4148 .09408 L
.416 .116 L
.41716 .13311 L
.41843 .14837 L
.42066 .16903 L
.42325 .18655 L
.42567 .19862 L
.42797 .20755 L
.43005 .21403 L
.43233 .21976 L
.43473 .22461 L
.43738 .22883 L
.43983 .23186 L
.44246 .23434 L
.44397 .23544 L
.44535 .23626 L
.44655 .23684 L
.44785 .23734 L
.44856 .23755 L
.44922 .23771 L
.45052 .23792 L
.45164 .238 L
.45287 .23798 L
.45415 .23783 L
.45535 .23757 L
.45653 .23721 L
.45779 .2367 L
.46005 .23544 L
.46274 .23334 L
.46519 .2308 L
.46774 .22737 L
.47052 .22261 L
.47313 .21685 L
.47554 .21016 L
.47761 .20305 L
.47982 .19357 L
.48226 .18014 L
.48453 .16348 L
.48575 .15223 L
.48708 .13745 L
.48826 .12155 L
.48951 .10063 L
.49074 .07422 L
.49186 .04329 L
.49299 .00213 L
s
.49299 .00213 m
.49303 0 L
s
.50068 0 m
.50068 .61803 L
s
.50675 .61803 m
.50762 .58072 L
.50876 .54517 L
.51008 .51386 L
.51153 .48796 L
.51289 .46901 L
.51417 .45451 L
.5166 .43324 L
.51881 .41878 L
.52119 .40665 L
.52376 .39633 L
.52839 .38259 L
.53266 .37354 L
.53735 .3661 L
.54238 .36007 L
.54713 .35566 L
.55147 .35243 L
.55654 .34934 L
.56207 .34665 L
.56725 .34459 L
.57212 .34299 L
.58151 .34059 L
.58692 .33953 L
.59188 .33871 L
.60198 .33743 L
.60729 .33694 L
.61288 .33652 L
.61787 .33623 L
.62241 .33602 L
.62493 .33593 L
.62765 .33585 L
.63023 .3358 L
.63137 .33577 L
.6326 .33576 L
.63375 .33574 L
.63479 .33573 L
.63585 .33572 L
.63685 .33572 L
.63801 .33571 L
.63908 .33571 L
.64026 .33571 L
.64152 .33571 L
.64274 .33572 L
.64344 .33573 L
.64407 .33573 L
.64647 .33576 L
.64866 .3358 L
.65096 .33585 L
.65593 .33599 L
.66045 .33616 L
Mistroke
.67062 .33667 L
.68004 .33732 L
.68896 .33809 L
.70889 .34039 L
.72732 .34334 L
.746 .34731 L
.76661 .35313 L
.78598 .3604 L
.80684 .37082 L
.82691 .38426 L
.84801 .40339 L
.86777 .4277 L
.88615 .45795 L
.90663 .50371 L
.92522 .56051 L
Mfstroke
.92522 .56051 m
.93903 .61803 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00n@3oool00`3ookl0M0000000
0003000000040000B02Lgol0oooO09a800D0000000<0001809cOo`3oool0Y@3oool00?T0oooo00<0
oono07@000000000100000000`00M;l0oooO09a80005000000030000B02Lgol0oooo0:D0oooo003l
0?ooo`050;md0000000000000000M02oool00P3oool01@2oM000000000000000B9`0gooo0:P0oooo
003m0?ooo`060?noM000000000000000001d_ol0gia80P0000000`0007@0_ooo0?ooo`2X0?ooo`00
oP3oool00`3ookl0M00000000003000000030000M02oool0oooo0:T0oooo003o0?ooo`40oooo00D0
_g@0000000000000001d0;ooo`2[0?ooo`00o`3oool00`3o_g@00000000000030000000307Boo`3o
ool0oooo0:P0oooo003n0?ooo`060?noM000000000000000B02Lgkl0M0000P0000000`00M;l0oooo
0?ooo`2W0?ooo`00o@3oool0203o_g@00000000000000000M;oo0?ooo`3ooml0W4P00P0000000`00
07@0_ooo0?ooo`2V0?ooo`00n`3oool00`3OW4P0000000000003000000040000B02Lgol0omnL04P0
00@0000000<0001809cOo`3oool0Y03oool00?/0oooo00<0gia80000000000000`000000100004P0
W=oo0?ooo`2oM004000000030000B02Lgol0oooo0:@0oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo002P0?ooo`030?oOW0180000000000@0
000000<0B9cO0?ooo`3oool09P3oool00`2oM000000000000007000000030000B02Lgol0oooo0280
oooo00H0omnL04P00000000000000000B02LgolZ0?ooo`030;md00000000000000L0000000<00018
09cOo`3oool0:`3oool00`3ogi`0B000000000040000000304RLg`3oool0oooo03l0oooo002P0?oo
o`030?oOW0180000000000@0000000<0B9cO0?ooo`3oool09P3oool00`3o_g@00000000000080000
000307Boo`3oool0oooo0200oooo00<0omnL04P000000000100000000`00M;l0oooo0?ooo`0W0?oo
o`030?noM0000000000000P0000000<0M;oo0?ooo`3oool0:P3oool00`3ogi`0B000000000040000
000304RLg`3oool0oooo03l0oooo002S0?ooo`040?noM0000000001d0;ooobT0oooo00<0oooO09a8
000000000P0000001`1d_ol0oooo0?ooo`3oool0oono07@0001d_ol08P3oool02@2oM00000000000
B02Lgol0oooo0?oOW0180000000004RLg`0X0?ooo`030?oog`2LB00000000080000000L0M;oo0?oo
o`3oool0oooo0?oo_`1d0000M;oo02l0oooo00@0okmd0000000007@0_ooo@P3oool009d0oooo00<0
gia80000000000001`0000000`1d_ol0oooo0?ooo`0X0?ooo`050=nLB0000000000000000000M;l0
9P3oool0103ooml0W4P000000018W=l30?ooo`040?oOW0180000000007BoobT0oooo00D0gia80000
000000000000001d_`0^0?ooo`030=nLB0000000000000L0000000<0M;oo0?ooo`3oool0?`3oool0
08d0oooo00<0gia80000000000001P0000000`0007@0_ooo0?ooo`040?ooo`030?oOW01800000000
00L0000000<007Bo0?ooo`3oool05`3oool00`3OW4P0000000000006000000030000M02oool0oooo
00H0oooo00H0omnL04P00000000000000000M02ooolT0?ooo`040?oo_`1d0000000007Boo`@0oooo
00@0gia800000000B9`0gooo:@3oool01P3ogi`0B000000000000000001d0;ooob`0oooo00<0omnL
04P0000000001`0000000`00M;l0oooo0?ooo`0o0?ooo`00S@3oool00`3OW4P00000000000060000
00030000M02oool0oooo00D0oooo00T0gia80000000004P0W=oo0?ooo`3ooml0W4P000000018W=l0
6P3oool00`3OW4P0000000000006000000030000M02oool0oooo00L0oooo00<0oooO09a800000000
0P0000000`1d_ol0oooo0?ooo`0Q0?ooo`040?oo_`1d0000000007Boo`@0oooo00@0omnL04P00000
04P0W=oo:P3oool00`3ooml0W4P0000000020000000307Boo`3oool0oooo02X0oooo00T0gia80000
000004P0W=oo0?ooo`3ooml0W4P000000018W=l0@P3oool009h0oooo00X0oooO09a80000000004RL
0=ooo`3oool0gia800000000B9`0gooo;P3oool01@3OW4P00000000000000000B9cO0280oooo00@0
oooO09a800000000B9cO103oool0103ooml0W4P000000018W=l/0?ooo`050=nLB000000000000000
0018W=l0:`3oool02P3ooml0W4P000000000B9`0gooo0?ooo`3OW4P000000018W03Ooom10?ooo`00
W`3oool02@3ogi`0B00000000018W=l0oooo0?noM0000000001d0;ooo`0_0?ooo`050?noM0000000
000000000000M;l08P3oool0103OW4P000000018W03Oool40?ooo`030;md0000000007Bo02d0oooo
00D0okmd0000000000000000001d_`0[0?ooo`090?oOW0180000000004RLg`3oool0okmd00000000
07@0_ooo0440oooo002P0?ooo`080=nLB0000000000007Boo`3ogi`0B0000000B02Lgol`0?ooo`05
0?oo_`1d0000000000000018W=l08@3oool0103o_g@000000000M02oool40?ooo`040=nLB0000000
04RL0=ooobd0oooo00D0oono07@000000000000004RLg`0[0?ooo`080=nLB0000000000007Boo`3o
gi`0B0000000B02Lgom10?ooo`00X@3oool01`2oM00000000000M02ookl0M0000000001d_ol0<@3o
ool01@3ooml0W4P00000000004P0W=oo0200oooo00@0oono07@0000004P0W=oo103oool0103o_g@0
00000000B02Lgol^0?ooo`050?oog`2LB00000000000B02Lgol0:`3oool01`2oM00000000000M02o
okl0M0000000001d_ol0@@3oool00:40oooo00L0oono07@00000000007Bo0;md0000000007Bo02/0
oooo00<0_g@00000B02Lgol0103oool0103ogi`0B00000000018W=lQ0?ooo`030;md0000000007Bo
00@0oooo00@0okmd0000000004P0W=oo:03oool00`2oM000001809cOo`040?ooo`040?oOW0180000
000004RLgb/0oooo00L0oono07@00000000007Bo0;md0000000007Bo0440oooo002R0?ooo`030?no
M000000000000080000000<004RL0=ooo`3oool0:@3oool0102oM00000000018W03Oool30?ooo`04
0?oo_`1d0000000004RLgb40oooo00@0omnL04P0000004P0W=oo0`3oool0103OW4P000000000M02o
oolX0?ooo`040;md0000000004RL0=ooo`<0oooo00@0oono07@000000000B9cO;03oool00`3o_g@0
000000000002000000030018W03Oool0oooo03l0oooo002S0?ooo`060=nLB0000000000000000000
07@0_ooo:P3oool01@3ookl0M00000000000B9`0gooo0080oooo00@0_g@0000000000000M;oo8P3o
ool02P3OW4P000000000B02Lgol0oooo0?oo_`1d000000000018W03OoolX0?ooo`050?oo_`1d0000
00000018W03Oool00P3oool0102oM00000000000001d_ol]0?ooo`060=nLB0000000000000000000
07@0_ooo@03oool00:<0oooo00H0oooO09a80000000000000000B02Lgol[0?ooo`030?oo_`1d0000
000000D0000000<004RL0=ooo`3oool08P3oool00`3OW4P00000000000040000000307Boo`3oool0
oooo02P0oooo00<0oono07@0000000001@0000000`00B9`0gooo0?ooo`0/0?ooo`060?oog`2LB000
00000000000004P0W=oo@03oool00:@0oooo00D0omnL04P000000000000004RLg`0]0?ooo`030?no
M000000000000080000000<0B9cO0?ooo`3oool0903oool01P3ookl0M000000000000000001809cO
obd0oooo00<0okmd0000000000000P0000000`18W=l0oooo0?ooo`0^0?ooo`050?oOW01800000000
00000018W=l0@03oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool007X0ooooo`00000B000002D0oooo001j0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool02P3oool2000000/0oooo00<000000?ooo`3oool02@3oool2000000`0oooo00<000000?oo
o`3oool02P3oool00`000000oooo0000000:0?ooo`030000003oool0oooo00P0oooo00<000000?oo
o`000000303oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool02P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo02<0oooo001j0?ooo`80
00001@3oool00`000000oooo0?ooo`0F0?ooo`030000O`3oool0oooo01L0oooo0P00000G0?ooo`80
00006@3oool00`000000oooo0000000E0?ooo`030000003oool0000003<0oooo00<0001o0?ooo`3o
ool0<03oool00`0007l0oooo0?ooo`0M0?ooo`8000009@3oool007X0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0F0?ooo`030000O`3oool0oooo01L0oooo0P00000G0?ooo`800000
6@3oool00`000000oooo0000000E0?ooo`030000003oool0000003<0oooo00<0001o0?ooo`3oool0
<03oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001j0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool05P3oool00`0007l0oooo0?ooo`0G0?ooo`8000005`3o
ool2000001T0oooo00<000000?ooo`0000005@3oool00`000000oooo0000000c0?ooo`030000O`3o
ool0oooo0300oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00NP3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo01H0oooo00<0001o0?ooo`3oool05`3oool2
000001L0oooo0P00000I0?ooo`030000003oool0000001D0oooo00<000000?ooo`000000<`3oool0
0`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool0
07X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0F0?ooo`030000O`3oool0oooo
01L0oooo0P00000G0?ooo`8000006@3oool00`000000oooo0000000E0?ooo`030000003oool00000
03<0oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo
02<0oooo001j0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool05P3oool00`0007l0
oooo0?ooo`0G0?ooo`8000005`3oool2000001T0oooo00<000000?ooo`0000005@3oool00`000000
oooo0000000c0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07P3oool00`000000
oooo0?ooo`0S0?ooo`00NP3oool2000000D0oooo00<000000?ooo`3oool05P3oool00`0007l0oooo
0?ooo`0G0?ooo`8000005`3oool2000001T0oooo00<000000?ooo`0000005@3oool00`000000oooo
0000000c0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07@3oool2000002D0oooo
001j0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool05P3oool00`0007l0oooo0?oo
o`0G0?ooo`8000005`3oool2000001T0oooo00<000000?ooo`0000005@3oool00`000000oooo0000
000c0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?oo
o`0S0?ooo`00NP3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01H0oooo00<0001o
0?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`030000003oool0000001D0oooo00<00000
0?ooo`000000<`3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo01h0oooo00<00000
0?ooo`3oool08`3oool007X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0F0?oo
o`030000O`3oool0oooo01L0oooo0P00000G0?ooo`8000006@3oool00`000000oooo0000000E0?oo
o`030000003oool0000003<0oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0N0?oo
o`030000003oool0oooo02<0oooo001j0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool05P3oool00`0007l0oooo0?ooo`0G0?ooo`8000005`3oool2000001T0oooo00<000000?ooo`00
00005@3oool00`000000oooo0000000c0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3o
ool07P3oool00`000000oooo0?ooo`0S0?ooo`00NP3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo01H0oooo00<0001o0?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`030000
003oool0000001D0oooo00<000000?ooo`000000<`3oool00`0007l0oooo0?ooo`0`0?ooo`030000
O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool007X0oooo0P0000050?ooo`030000003o
ool0oooo01H0oooo00<0001o0?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`030000003o
ool0000001D0oooo00<000000?ooo`000000<`3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3o
ool0oooo01d0oooo0P00000U0?ooo`00NP3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo01H0oooo00<0001o0?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`030000003oool0
000001D0oooo00<000000?ooo`000000<`3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0
oooo01h0oooo00<000000?ooo`3oool08`3oool007X0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`0F0?ooo`030000O`3oool0oooo01L0oooo0P00000G0?ooo`8000006@3oool00`00
0000oooo0000000D0?ooo`040000003oool0oooo000003<0oooo00<0001o0?ooo`3oool0<03oool0
0`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001j0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool05P3oool00`0007l0oooo0?ooo`0G0?ooo`8000005`3oool20000
01T0oooo00<000000?ooo`000000503oool010000000oooo0?ooo`00000c0?ooo`030000O`3oool0
oooo0300oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00NP3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo01H0oooo00<0001o0?ooo`3oool05`3oool20000
01L0oooo0P00000I0?ooo`030000003oool0000001@0oooo00@000000?ooo`3oool00000<`3oool0
0`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool0
07X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0F0?ooo`030000O`3oool0oooo
01L0oooo0P00000G0?ooo`8000006@3oool00`000000oooo0000000D0?ooo`040000003oool0oooo
000003<0oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0
oooo02<0oooo001I0?ooo`030?noM0000000000000P0000000<0M;oo0?ooo`3oool00P3oool01P3o
gi`0B000000000000000001809cOo`/0oooo0P0000050?ooo`030000003oool0oooo01H0oooo00<0
001o0?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`030000003oool0000001@0oooo00@0
00000?ooo`3oool00000<`3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo01d0oooo
0P00000U0?ooo`00F@3oool00`3o_g@00000000000080000000607Boo`3oool0oooo0?ooo`3ogi`0
B0001@0000000`00M;l0oooo0?ooo`080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool05P3oool00`0007l0oooo0?ooo`0G0?ooo`8000005`3oool2000001T0oooo00<000000?ooo`00
0000503oool010000000oooo0?ooo`00000c0?ooo`030000O`3oool0oooo0300oooo00<0001o0?oo
o`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00G@3oool0103ogi`0B0000000001d_ol70?oo
o`090;md00000000001809cOo`3oool0omnL04P000000000B9cO00T0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0F0?ooo`030000O`3oool0oooo01L0oooo0P00000G0?ooo`800000
6@3oool010000000oooo0?ooo`00000C0?ooo`040000003oool0oooo000003<0oooo00<0001o0?oo
o`3oool0<03oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001M0?ooo`04
0?oog`2LB000000004RLg`H0oooo00@0oooO09a800000000B9cO0`3oool0103ogi`0B0000000001d
_ol80?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool05P3oool00`0007l0oooo0?oo
o`0G0?ooo`8000005`3oool2000001T0oooo00@000000?ooo`3oool000004`3oool010000000oooo
0?ooo`00000c0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07P3oool00`000000
oooo0?ooo`0S0?ooo`00B@3oool00`3OW4P0000000000006000000030000M02oool0oooo00T0oooo
00<0_g@000000000M;l01P3oool0103ookl0M0000000001d_ol40?ooo`040=nLB000000004RL0=oo
o`L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0F0?ooo`030000O`3oool0oooo
01L0oooo0P00000G0?ooo`8000006@3oool010000000oooo0?ooo`00000C0?ooo`040000003oool0
oooo000003<0oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0N0?ooo`030000003o
ool0oooo02<0oooo00190?ooo`030=nLB0000000000000H0000000<0001d0;ooo`3oool02@3oool0
103OW4P000000000M02oool50?ooo`040?oo_`1d0000000007Boo`@0oooo00@0omnL04P0000004P0
W=oo1`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01H0oooo00<0001o0?ooo`3o
ool05`3oool2000001L0oooo0P00000I0?ooo`040000003oool0oooo000001<0oooo00@000000?oo
o`3oool00000<`3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo01h0oooo00<00000
0?ooo`3oool08`3oool005h0oooo00@0omnL04P0000004P0W=oo1@3oool0103ooml0W4P000000018
W=l40?ooo`040?oog`2LB000000004RLg`L0oooo0`0000040000O`030000000007l0001o0300001o
0P00000G0000O`8000006@0007l010000000001o0000O`00000C0000O`040000000007l0001o0000
08D0001o0`00000U0?ooo`00GP3oool0103ookl0M0000000001d_ol60?ooo`040=nLB000000004RL
0=ooo`@0oooo00<0_g@000000000M;l01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo01H0oooo00<0001o0?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`040000003oool0
oooo000001<0oooo00@000000?ooo`3oool00000<`3oool00`0007l0oooo0?ooo`0`0?ooo`030000
O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool005h0oooo00@0oooO09a80000000007Bo
1P3oool0103o_g@000000000M02oool40?ooo`040=nLB000000004RL0=ooo`H0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0F0?ooo`030000O`3oool0oooo01L0oooo0P00000G0?oo
o`8000006@3oool010000000oooo0?ooo`00000C0?ooo`040000003oool0oooo000003<0oooo00<0
001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001O
0?ooo`040=nLB000000004RL0=ooo`D0oooo00@0oono07@0000004P0W=oo103oool0103o_g@00000
0000B02Lgol60?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool05P3oool00`000000
oooo0?ooo`0G0?ooo`8000005`3oool2000001T0oooo00@000000?ooo`3oool000004`3oool01000
0000oooo0?ooo`00000c0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07P3oool0
0`000000oooo0?ooo`0S0?ooo`00G`3oool0103o_g@000000000M02oool60?ooo`030;md00000000
07Bo00@0oooo00@0okmd0000000004P0W=oo1P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo01H0oooo00<000000?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`040000003o
ool0oooo000001<0oooo00@000000?ooo`3oool00000<`3oool00`0007l0oooo0?ooo`0`0?ooo`03
0000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool005/0oooo00P0okmd000000000000
001d0;oo_`1d0000000007Boo`H0oooo00@0omnL04P0000004P0W=oo0`3oool0103OW4P000000000
M02oool60?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool05P3oool00`000000oooo
0?ooo`0G0?ooo`8000005`3oool2000001T0oooo00@000000?ooo`3oool000004`3oool010000000
oooo0?ooo`00000c0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07P3oool00`00
0000oooo0?ooo`0S0?ooo`00F`3oool00`3ogi`0B000000000040000000304RLg`3oool0oooo00D0
oooo00X0gia80000000004P0W=oo0?ooo`3ookl0M00000000000B9`0gooo1P3oool2000000D0oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`8000005`3oool2000001T0oooo
00@000000?ooo`3oool000004`3oool010000000oooo0?ooo`00000c0?ooo`030000O`3oool0oooo
0300oooo00<0001o0?ooo`3oool07@3oool2000002D0oooo001N0?ooo`050=nLB000000000000000
0000M;l0203oool00`3OW4P00000000000040000000307Boo`3oool0oooo00D0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo0P00000G0?oo
o`8000006@3oool010000000oooo0?ooo`00000C0?ooo`040000003oool0oooo000003<0oooo00<0
001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001P
0?ooo`040?oog`2LB000001d0;ooo`P0oooo00H0oono07@00000000000000000B02Lgol80?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`80
00005`3oool2000001T0oooo00@000000?ooo`3oool000004`3oool010000000oooo0?ooo`00000c
0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S
0?ooo`00NP3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01H0oooo00<000000?oo
o`3oool05`3oool2000001L0oooo0P00000I0?ooo`040000003oool0oooo00000180oooo00D00000
0?ooo`3oool0oooo0000000c0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07P3o
ool00`000000oooo0?ooo`0S0?ooo`00NP3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`040000003oool0
oooo00000180oooo00D000000?ooo`3oool0oooo0000000c0?ooo`030000O`3oool0oooo0300oooo
00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00NP3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool05`3oool2000001L0oooo0P00
000I0?ooo`040000003oool0oooo00000180oooo00D000000?ooo`3oool0oooo0000000c0?ooo`03
0000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00
NP3oool2000000D0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`800000
5`3oool2000001T0oooo00@000000?ooo`3oool000004P3oool01@000000oooo0?ooo`3oool00000
03<0oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0M0?ooo`8000009@3oool007X0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0
oooo0P00000G0?ooo`8000006@3oool01@000000oooo0?ooo`3oool000000140oooo00D000000?oo
o`3oool0oooo0000000c0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07P3oool0
0`000000oooo0?ooo`0S0?ooo`00NP3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
01H0oooo00<000000?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`050000003oool0oooo
0?ooo`0000004@3oool01@000000oooo0?ooo`3oool0000003<0oooo00<0001o0?ooo`3oool0<03o
ool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001j0?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`8000005`3oool2
000001T0oooo00D000000?ooo`3oool0oooo0000000A0?ooo`050000003oool0oooo0?ooo`000000
<`3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool0
8`3oool007X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0F0?ooo`030000003o
ool0oooo01L0oooo0P00000G0?ooo`8000006@3oool01@000000oooo0?ooo`3oool000000140oooo
00D000000?ooo`3oool0oooo0000000c0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3o
ool07P3oool00`000000oooo0?ooo`0S0?ooo`00NP3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo01H0oooo00<000000?ooo`3oool05`3oool2000001H0oooo00<000000?ooo`000000
6@3oool01@000000oooo0?ooo`3oool000000140oooo00D000000?ooo`3oool0oooo0000000c0?oo
o`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?oo
o`00NP3oool2000000D0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`80
00005P3oool00`000000oooo0000000I0?ooo`050000003oool0oooo0?ooo`0000004@3oool01@00
0000oooo0?ooo`3oool0000003<0oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0M
0?ooo`8000009@3oool007X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0F0?oo
o`030000003oool0oooo01L0oooo0P00000F0?ooo`030000003oool0000001T0oooo00D000000?oo
o`3oool0oooo0000000A0?ooo`050000003oool0oooo0?ooo`000000<`3oool00`0007l0oooo0?oo
o`0`0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool007X0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo0P00000F
0?ooo`030000003oool0000001T0oooo00D000000?ooo`3oool0oooo0000000A0?ooo`050000003o
ool0oooo0?ooo`000000<`3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo01h0oooo
00<000000?ooo`3oool08`3oool007X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`0F0?ooo`030000003oool0oooo01L0oooo0P00000F0?ooo`030000003oool0000001T0oooo00D0
00000?ooo`3oool0oooo0000000@0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
<@3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool0
8`3oool007X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0F0?ooo`030000003o
ool0oooo01L0oooo0P00000F0?ooo`030000003oool0000001T0oooo00D000000?ooo`3oool0oooo
0000000@0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0<@3oool00`0007l0oooo
0?ooo`0`0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool007X0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo0P00
000F0?ooo`030000003oool0000001T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0<@3oool00`0007l0oooo0?oo
o`0`0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool006P0oooo00<0_g@0
0000000000000`0000000`00M;l0oooo0?ooo`090?ooo`8000001@3oool00`000000oooo0?ooo`0F
0?ooo`030000003oool0oooo01L0oooo0P00000F0?ooo`030000003oool0000001T0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo01d0oooo0P00000U
0?ooo`00IP3oool00`3ookl0M000000000070000000304RLg`3oool0oooo00L0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo0P00000F0?oo
o`030000003oool0000001T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0<0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0`0?oo
o`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool006L0oooo00<0_g@000000018
W=l0103oool0103OW4P000000000001d_ol80?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`8000005P3oool00`000000oooo0000000I0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0N0?oo
o`030000003oool0oooo02<0oooo001_0?ooo`040?noM0000000001d0;ooo`L0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo0P00000F0?oo
o`030000003oool0000001T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0`0?oo
o`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool005L0oooo00<0gia800000000
00001P0000000`0007@0_ooo0?ooo`0<0?ooo`040?oo_`1d0000000007Boo`L0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo0P00000F0?oo
o`030000003oool0000001T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0`0?oo
o`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool002/0oooo00<0gia800000018
W=l02@3oool0103ookl0M0000000M02ooolL0?ooo`030=nLB0000000000000H0000000<0001d0;oo
o`3oool03@3oool00`2oM0000000001d_`070?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`8000005P3oool00`000000oooo0000000I0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0N0?oo
o`030000003oool0oooo02<0oooo000[0?ooo`030=nLB0000000B9cO00T0oooo00@0oono07@00000
07@0_ooo=@3oool0102oM00000000018W03Oool60?ooo`<00000100007l00`000000001o0000O`0F
0000O`030000000007l0001o01L0001o0P00000F0000O`030000000007l0000001T0001o00<00000
0000O`0007l00`0007l00`000000001o0000O`0:0000O`030000000007l0001o00@0001o00<00000
0000O`0007l0P`0007l3000002D0oooo000[0?ooo`030=nLB0000000B9cO00T0oooo00@0oono07@0
000007@0_ooo=03oool0103ooml0W4P000000000M;l70?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`8000005P3oool00`000000oooo0000
000I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?oo
o`0N0?ooo`030000003oool0oooo02<0oooo000[0?ooo`030=nLB0000000000000/0000000<0001d
0;ooo`3oool0;@3oool0103o_g@000000018W03Oool20?ooo`040?noM0000000000004RLg`L0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo
0P00000F0?ooo`030000003oool0000001T0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0<@3oool00`0007l0oooo
0?ooo`0`0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool002/0oooo00<0
gia80000000000002`0000000`0007@0_ooo0?ooo`0]0?ooo`030?oOW0180000000000D0000000<0
001d0;ooo`3oool01P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01H0oooo00<0
00000?ooo`3oool05`3oool2000001H0oooo00<000000?ooo`0000006@3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`0a0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?oo
o`0S0?ooo`00:`3oool00`3OW4P0000004RLg`090?ooo`040?oo_`1d0000001d0;ooobh0oooo00<0
oono07@000000000100000000`00B9`0gooo0?ooo`070?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`8000005@3oool010000000oooo0?oo
o`00000I0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo
0?ooo`0N0?ooo`030000003oool0oooo02<0oooo000[0?ooo`030=nLB0000000B9cO00T0oooo00@0
oono07@0000007@0_ooo;P3oool0103ooml0W4P000000018W=l=0?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`8000005@3oool010000000
oooo0?ooo`00000I0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<03oool00`00
07l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo000[0?ooo`030=nLB0000000B9cO00T0
oooo00@0oono07@0000007@0_ooo;`3oool0103OW4P000000018W03Oool<0?ooo`8000001@3oool0
0`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo0P00000E0?ooo`040000003oool0
oooo000001T0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`040?ooo`800000203o
ool00`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07@3o
ool2000002D0oooo000[0?ooo`030=nLB0000000B9cO03`0oooo00<0okmd0000000000001P000000
0`1d_ol0oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool05P3oool0
0`000000oooo0?ooo`0G0?ooo`8000005@3oool010000000oooo0?ooo`00000I0?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo0340oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0N0?ooo`030000003o
ool0oooo02<0oooo000[0?ooo`030=nLB0000000B9cO03`0oooo00<0omnL04P0000000001P000000
0`00M;l0oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool05P3oool0
0`000000oooo0?ooo`0G0?ooo`8000005@3oool010000000oooo0?ooo`00000I0?ooo`030000003o
ool0oooo00P0oooo1@00000:0?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<03o
ool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo000[0?ooo`030=nLB0000000
00000080000000<007Bo0?ooo`3oool0A`3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool05`3oool2000001D0oooo00@000000?ooo`3oool000006@3o
ool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<03o
ool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo000[0?ooo`030=nLB0000000
00000080000000<007Bo0?ooo`3oool0A`3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000D0?ooo`040000003oool0
oooo000001T0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0a0?ooo`030000O`3o
ool0oooo0300oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00NP3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0000000C0?ooo`050000003oool0oooo0?ooo`0000006@3oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?oo
o`0N0?ooo`030000003oool0oooo02<0oooo001j0?ooo`8000001@3oool00`000000oooo0?ooo`0F
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`0000004`3oool01@000000oooo0?ooo`3o
ool0000001T0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0a0?ooo`030000O`3o
ool0oooo0300oooo00<0001o0?ooo`3oool07@3oool2000002D0oooo001j0?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
000001<0oooo00D000000?ooo`3oool0oooo0000000I0?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo01h0oooo00<0
00000?ooo`3oool08`3oool007X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0F
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`0000004`3oool01@000000oooo0?ooo`3o
ool0000001T0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0a0?ooo`030000O`3o
ool0oooo0300oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00NP3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0000000B0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`0G0?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<03oool0
0`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001j0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool00000
0180oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo
01L0oooo00<000000?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo
01h0oooo00<000000?ooo`3oool08`3oool007X0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`0000004@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000
oooo0?ooo`0a0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07P3oool00`000000
oooo0?ooo`0S0?ooo`00NP3oool2000000D0oooo00<000000?ooo`3oool05P3oool00`000000oooo
0?ooo`0G0?ooo`040000003oool0oooo00000100oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0<@3oool00`0007l0
oooo0?ooo`0`0?ooo`030000O`3oool0oooo01d0oooo0P00000U0?ooo`00NP3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool05`3oool010000000oooo
0?ooo`00000@0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool05`3oool00`000000
oooo0?ooo`0G0?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<03oool00`0007l0
oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001j0?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`040000003oool0oooo000000l0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0
oooo00<000000?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo01h0
oooo00<000000?ooo`3oool08`3oool002X0oooo00<0oono07@0000000004@0000000`18W=l0oooo
0?ooo`0i0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool05P3oool00`000000oooo
0?ooo`0G0?ooo`050000003oool0oooo0?ooo`0000003P3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0a0?ooo`03
0000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00
:P3oool00`3ookl0M0000000000A0000000304RLg`3oool0oooo03T0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`090?ooo`8000001`3oool00`000000oooo0?ooo`0G0?ooo`030000
003oool0oooo01L0oooo00<000000?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0`0?ooo`030000
O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool002X0oooo00@0oono07@0000007@0_ooo
3P3oool00`3OW4P0000004RLg`0k0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
5@3oool2000001T0oooo00<000000?ooo`3oool00P3oool2000000P0oooo0P0000090?ooo`030000
003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0a0?ooo`030000
O`3oool0oooo0300oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00:P3o
ool0103ookl0M0000000M02oool>0?ooo`030=nLB0000000B9cO02X0oooo00H0omnL04P000000000
00000000B02Lgol;0?ooo`8000001@3oool00`000000oooo0?ooo`0E0?ooo`8000006@3oool00`00
0000oooo0?ooo`040?ooo`P000002`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo
01L0oooo00<000000?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo
01d0oooo0P00000U0?ooo`00:P3oool0103ookl0M0000000M02oool>0?ooo`030=nLB0000000B9cO
02T0oooo00<0omnL04P000000000100000000`00M;l0oooo0?ooo`080?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool05@3oool2000001T0oooo00<000000?ooo`3oool05`3oool00`00
0000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0<@3oool00`00
07l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool006P0
oooo00T0_g@00000000004P0W=oo0?ooo`3ogi`0B00000000018W=l02@3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo01D0oooo0P00000I0?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo0340oooo00<0
001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001W
0?ooo`040?oog`2LB000000004RLg`<0oooo00@0omnL04P000000000M;oo203oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo01@0oooo00<000000?ooo`0000006@3oool00`000000oooo
0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo
0?ooo`0a0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool07P3oool00`000000oooo
0?ooo`0S0?ooo`00I`3oool0103ookl0M0000000001d_ol40?ooo`040=nLB000000004RL0=ooo`L0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0B0?ooo`<0000000<0oooo0000003o
ool0603oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3o
ool05`3oool00`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3o
ool07P3oool00`000000oooo0?ooo`0S0?ooo`00I`3oool0103ookl0M0000000001d_ol40?ooo`04
0?oOW0180000001809cOo`L0oooo00<000000?ooo`3oool01@3ooolE000000<0oooo00<000000?oo
o`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool05`3oool00`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo0300oooo00<0001o0?oo
o`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00=`3oool0103ooml0W4P00000M02oool/0?oo
o`040?oog`2LB000000004RLg`@0oooo00@0oooO09a800000000B9cO1`3ooooo0000018000009@3o
ool003L0oooo00@0okmd0000000007@0_ooo;@3oool0103OW4P000000018W03Oool40?ooo`030;md
0000000007Bo00L0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`030?ooo`X00000
2P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
5`3oool00`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool0
7P3oool00`000000oooo0?ooo`0S0?ooo`00;`3oool00`3ogi`0B000001d_`040?ooo`050?oog`2L
B00000000000M02oool0;@3oool0103o_g@000000000M02oool40?ooo`040=nLB000000004RL0=oo
o`H0oooo00<000000?ooo`3oool07@3oool010000000oooo0?ooo`3oool2000000X0oooo10000006
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0N
0?ooo`030000003oool0oooo02<0oooo000_0?ooo`040;md0000000004RL0=ooo`<0oooo00D0gia8
000000000000001d0;ooo`0]0?ooo`040?oo_`1d0000001809cOo`@0oooo00@0okmd0000000004P0
W=oo1P3oool00`000000oooo0?ooo`0M0?ooo`040000003oool0000000000100oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool05`3oool00`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo0300oooo00<0001o0?oo
o`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00;`3oool0102oM00000000000001d_ol20?oo
o`060?oOW018000000000000000007@0_ooo;P3oool00`2oM0000000001d_`040?ooo`040?noM000
0000001809cOo`H0oooo00<000000?ooo`3oool07@3oool00`000000oooo0000000B0?ooo`800000
0`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
5`3oool00`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool0
7P3oool00`000000oooo0?ooo`0S0?ooo`00;`3oool01`2oM0000000000000000000M;oo0?oog`2L
B0000`0000000`00B9`0gooo0?ooo`0]0?ooo`040?oOW0180000001809cOo`<0oooo00@0gia80000
000007@0_ooo1P3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0000001<0oooo00D00000
0?ooo`3oool0oooo0000000I0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3o
ool00`000000oooo0?ooo`0H0?oooaL000000P3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3o
ool0oooo01h0oooo00<000000?ooo`3oool08`3oool002l0oooo00<0_g@00000000000000P000000
1P18B4P000000000M02ogi`0B01d0;oooc00oooo00X0gia80000000004P0W=oo0?ooo`3ookl0M000
00000000B9`0gooo1P3oool2000001h0oooo0P00000E0?ooo`040000003oool0oooo000001T0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01<0oooo
1@00000G0?ooo`X00000:`3oool00`0007l0oooo0?ooo`0M0?ooo`8000009@3oool002l0oooo00<0
_g@0000000000000100000000`1d_ol0oooo0?ooo`0b0?ooo`030=nLB0000000000000@0000000<0
M;oo0?ooo`3oool01@3oool00`000000oooo0?ooo`0M0?ooo`8000005@3oool010000000oooo0?oo
o`00000I0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo
0?ooo`0?0?ooo`@000007P3oool00`0007l0oooo0?ooo`050?ooo`D000009P3oool00`0007l0oooo
0?ooo`0N0?ooo`030000003oool0oooo02<0oooo000b0?ooo`080?noM0000000000000000018W=l0
oooo0;md0000M;lb0?ooo`060?oo_`1d000000000000000004P0W=oo203oool00`000000oooo0?oo
o`0M0?ooo`8000005@3oool010000000oooo0?ooo`00000I0?ooo`030000003oool0oooo01L0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0200oooo
00<0001o0?ooo`3oool02P3oool400000280oooo00<0001o0?ooo`3oool07P3oool00`000000oooo
0?ooo`0S0?ooo`00<P3oool00`2oM0000000000000020000000404Q8B0000000001d0;ooocl0oooo
00<000000?ooo`3oool07@3oool2000001H0oooo00<000000?ooo`0000006@3oool00`000000oooo
0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0303oool2000002<0oooo
00<0001o0?ooo`3oool03P3oool200000200oooo00<0001o0?ooo`3oool07P3oool00`000000oooo
0?ooo`0S0?ooo`00;`3oool01P3ogi`0B000000000000000001809bLM0@0000000<0001d0;ooo`3o
ool0?P3oool00`000000oooo0?ooo`0M0?ooo`8000005P3oool00`000000oooo0000000I0?ooo`03
0000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0:0?ooo`80
00009@3oool00`0007l0oooo0?ooo`0@0?ooo`<000007@3oool00`0007l0oooo0?ooo`0N0?ooo`03
0000003oool0oooo02<0oooo000_0?ooo`080;md00000000000000000018W=l0oooo0?oOW0180002
000000030000M02oool0oooo03h0oooo00<000000?ooo`3oool07@3oool2000001H0oooo00<00000
0?ooo`0000006@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`0U0?ooo`030000O`3oool0oooo01<0oooo0`00000J
0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool002l0oooo00D0_g@00000
0000000004RL0=ooo`020?ooo`050?oOW018000000000000M02oool0?`3oool2000001h0oooo0P00
000F0?ooo`030000003oool0000001T0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool09P3oool00`0007l0oooo0?oo
o`0F0?ooo`800000603oool00`0007l0oooo0?ooo`0M0?ooo`8000009@3oool002l0oooo00@0_g@0
0000000004P0W=oo103oool0103o_g@000000000M02ooolo0?ooo`030000003oool0oooo01d0oooo
0P00000F0?ooo`030000003oool0000001T0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`0W0?ooo`030000O`3oool0oooo01P0oooo0P00000F0?ooo`030000O`3oool0oooo01h0oooo
00<000000?ooo`3oool08`3oool002l0oooo00@0_g@000000000B9`0gooo103oool00`3o_g@00000
04RLg`100?ooo`030000003oool0oooo01d0oooo0P00000G0?ooo`8000006@3oool00`000000oooo
0?ooo`050?ooo`L000002`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool0:03oool00`0007l0oooo0?ooo`0J0?ooo`800000503oool00`0007l0oooo
0?ooo`0N0?ooo`030000003oool0oooo02<0oooo000_0?ooo`030=nLB0000000M;oo04P0oooo00<0
00000?ooo`3oool07@3oool2000001L0oooo0P00000I0?ooo`030000003oool0oooo00@0oooo0P00
00060?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo02P0oooo00<0001o0?ooo`3oool0703oool00`000000oooo0?oo
o`0A0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool007X0oooo00<00000
0?ooo`3oool07@3oool2000001L0oooo0P00000I0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`0X0?ooo`030000O`3oool0oooo01d0oooo0P00000A
0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool007X0oooo00<000000?oo
o`3oool07@3oool2000001L0oooo0P00000I0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`0Y0?ooo`030000O`3oool0oooo01l0oooo0P00000?0?oo
o`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool002X0oooo00@0oono07@00000
07@0_ooo3P3oool00`3OW4P0000004RLg`0k0?ooo`8000007P3oool2000001L0oooo0P00000I0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0Y0?oo
o`030000O`3oool0oooo0240oooo0P00000=0?ooo`030000O`3oool0oooo01d0oooo0P00000U0?oo
o`00:P3oool0103ookl0M0000000M02oool>0?ooo`030=nLB0000000B9cO03/0oooo00<000000?oo
o`3oool07@3oool2000001L0oooo0P00000I0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool02P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0Z0?ooo`030000O`3oool0oooo02<0oooo00<000000?oo
o`3oool02P3oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo000Z0?ooo`04
0?oo_`1d0000001d0;ooo`h0oooo00<0gia800000018W=l0>`3oool00`000000oooo0?ooo`0M0?oo
o`8000005`3oool2000001T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo02X0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`090?oo
o`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool002X0oooo00<0oono07@00000
00004@0000000`18W=l0oooo0?ooo`0i0?ooo`030000003oool0oooo01d0oooo0P00000G0?ooo`80
00006@3oool01@000000oooo0?ooo`3oool0000000d0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`0G0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0:`3oool00`00
07l0oooo0?ooo`0U0?ooo`030000003oool0oooo00P0oooo00<0001o0?ooo`3oool07P3oool00`00
0000oooo0?ooo`0S0?ooo`00:P3oool00`3ookl0M0000000000A0000000304RLg`3oool0oooo03T0
oooo00<000000?ooo`3oool07@3oool2000001L0oooo0P00000I0?ooo`050000003oool0oooo0?oo
o`0000003P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0[0?ooo`030000O`3oool0oooo02H0oooo00<000000?oo
o`3oool01`3oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001j0?ooo`03
0000003oool0oooo01d0oooo0P00000G0?ooo`8000006@3oool01@000000oooo0?ooo`3oool00000
00h0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool0:`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo
00H0oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00J03oool00`2oM000
00000000000300000003001d_`3oool0oooo00T0oooo0P00000N0?ooo`8000005`3oool2000001T0
oooo00D000000?ooo`3oool0oooo0000000?0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool05`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo02/0oooo00<0001o0?oo
o`3oool0:03oool00`000000oooo0?ooo`050?ooo`030000O`3oool0oooo01d0oooo0P00000U0?oo
o`00IP3oool00`3ookl0M000000000070000000304RLg`3oool0oooo00L0oooo00<000000?ooo`3o
ool07@3oool2000001L0oooo0P00000I0?ooo`050000003oool0oooo0?ooo`0000003`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0[0?ooo`030000O`3oool0oooo02T0oooo00<000000?ooo`3oool0103oool00`00
07l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001W0?ooo`030;md00000000B9cO00@0
oooo00@0gia8000000000000M;oo203oool00`000000oooo0?ooo`0M0?ooo`8000005`3oool20000
01T0oooo00D000000?ooo`3oool0oooo0000000?0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo02`0oooo00<0001o
0?ooo`3oool0:P3oool00`000000oooo0?ooo`030?ooo`030000O`3oool0oooo01h0oooo00<00000
0?ooo`3oool08`3oool006l0oooo00@0okmd0000000007@0_ooo1`3oool00`000000oooo0?ooo`0M
0?ooo`8000005`3oool2000001T0oooo00D000000?ooo`3oool0oooo0000000@0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo02`0oooo00<0001o0?ooo`3oool0:P3oool00`000000oooo0?ooo`030?ooo`030000O`3o
ool0oooo01h0oooo00<000000?ooo`3oool08`3oool006l0oooo00@0oono07@000000000M;oo1`3o
ool00`000000oooo0?ooo`0M0?ooo`8000005`3oool2000001T0oooo00@000000?ooo`3oool00000
4@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0/0?ooo`030000O`3oool0oooo02/0oooo00<000000?ooo`3oool0
0P3oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001`0?ooo`030;md0000
000007Bo00L0oooo00<000000?ooo`3oool07@3oool2000001L0oooo0P00000I0?ooo`040000003o
ool0oooo00000140oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0G0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0;03oool00`0007l0oooo0?ooo`0/0?ooo`050000
003oool0oooo0?ooo`0007l0803oool00`000000oooo0?ooo`0S0?ooo`00L03oool0102oM0000000
0018W03Oool60?ooo`<000007@0007l2000001L0001o0P00000I0000O`040000000007l0001o0000
0140001o00<000000000O`0007l00P0007l00`000000001o0000O`0G0000O`050000000007l0001o
0000O`000000G`0007l00`000000001o0000O`0O0000O`<000009@3oool006l0oooo00@0oooO09a8
0000000007Bo1`3oool00`000000oooo0?ooo`0M0?ooo`8000005`3oool2000001T0oooo00@00000
0?ooo`3oool000004@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01L0oooo00D0
00000?ooo`3oool0oooo0000000_0?ooo`030000O`3oool0oooo02h0oooo00<000000?ooo`0007l0
803oool00`000000oooo0?ooo`0S0?ooo`00J@3oool0103o_g@000000018W03Oool20?ooo`040?no
M0000000000004RLg`L0oooo00<000000?ooo`3oool07@3oool2000001L0oooo0P00000I0?ooo`04
0000003oool0oooo00000140oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0G0?oo
o`050000003oool0oooo0?ooo`000000;`3oool00`0007l0oooo0?ooo`0^0?ooo`030000003oool0
001o0200oooo00<000000?ooo`3oool08`3oool006T0oooo00<0omnL04P0000000001@0000000`00
07@0_ooo0?ooo`060?ooo`030000003oool0oooo01d0oooo0P00000G0?ooo`8000006@3oool01000
0000oooo0?ooo`00000A0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool05`3oool0
1@000000oooo0?ooo`3oool0000002l0oooo00<0001o0?ooo`3oool0;`3oool00`000000001o0?oo
o`0O0?ooo`030000003oool0oooo02<0oooo001Y0?ooo`030?oo_`1d0000000000@0000000<004RL
0=ooo`3oool01`3oool00`000000oooo0?ooo`0M0?ooo`800000603oool00`000000oooo0?ooo`0G
0?ooo`040000003oool0oooo00000140oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0G0?ooo`050000003oool0oooo0?ooo`000000;`3oool00`0007l0oooo0?ooo`0_0?ooo`030000
000007l0oooo01l0oooo00<000000?ooo`3oool08`3oool006T0oooo00@0oooO09a800000000B9cO
3@3oool00`000000oooo0?ooo`0M0?ooo`800000603oool00`000000oooo0?ooo`0G0?ooo`040000
003oool0oooo00000180oooo00D000000?ooo`3oool0oooo0000000I0?ooo`050000003oool0oooo
0?ooo`000000;`3oool00`0007l0oooo0?ooo`0`0?ooo`030000003oool0oooo01h0oooo00<00000
0?ooo`3oool08`3oool006X0oooo00@0gia800000000B9`0gooo303oool2000001h0oooo0P00000H
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`0000004`3oool01@000000oooo0?ooo`3o
ool0000001T0oooo00D000000?ooo`3oool0oooo0000000_0?ooo`030000O`3oool0oooo0300oooo
00<0001o0000003oool07@3oool2000002D0oooo001Z0?ooo`030?noM0000000000000H0000000<0
M;oo0?ooo`3oool0103oool00`000000oooo0?ooo`0M0?ooo`800000603oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0000001<0oooo00D000000?ooo`3oool0oooo0000000I0?ooo`050000
003oool0oooo0?ooo`000000;`3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`000000oooo01h0
oooo00<000000?ooo`3oool08`3oool006X0oooo00<0omnL04P0000000001P0000000`00M;l0oooo
0?ooo`040?ooo`030000003oool0oooo01d0oooo0P00000H0?ooo`030000003oool0oooo01L0oooo
00<000000?ooo`0000004`3oool01@000000oooo0?ooo`3oool0000001T0oooo00D000000?ooo`3o
ool0oooo0000000_0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`0000007P3oool00`00
0000oooo0?ooo`0S0?ooo`00NP3oool00`000000oooo0?ooo`0M0?ooo`800000603oool00`000000
oooo0?ooo`0G0?ooo`030000003oool0000001<0oooo00D000000?ooo`3oool0oooo0000000I0?oo
o`040000003oool0oooo00000300oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0000000N
0?ooo`030000003oool0oooo02<0oooo001j0?ooo`030000003oool0oooo01d0oooo0P00000H0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`0000004`3oool01@000000oooo0?ooo`3oool0
000001T0oooo00@000000?ooo`3oool00000<03oool00`0007l0oooo0?ooo`0`0?ooo`040000O`3o
ool0oooo000001d0oooo00<000000?ooo`3oool08`3oool007X0oooo00<000000?ooo`3oool07@3o
ool2000001P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000D0?ooo`040000003o
ool0oooo000001T0oooo00@000000?ooo`3oool00000<03oool00`0007l0oooo0?ooo`0`0?ooo`04
0000O`3oool0oooo000001d0oooo00<000000?ooo`3oool08`3oool007X0oooo0P00000N0?ooo`80
0000603oool00`000000oooo0?ooo`0G0?ooo`030000003oool0000001@0oooo00@000000?ooo`3o
ool000006@3oool010000000oooo0?ooo`00000`0?ooo`030000O`3oool0oooo0300oooo00D0001o
0?ooo`3oool0oooo0000000K0?ooo`8000009@3oool007X0oooo00<000000?ooo`3oool07@3oool2
000001P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000D0?ooo`040000003oool0
oooo000001T0oooo00@000000?ooo`3oool00000<03oool00`0007l0oooo0?ooo`0`0?ooo`050000
O`3oool0oooo0?ooo`000000703oool00`000000oooo0?ooo`0S0?ooo`00NP3oool00`000000oooo
0?ooo`0M0?ooo`800000603oool00`000000oooo0?ooo`0G0?ooo`030000003oool0000001@0oooo
00@000000?ooo`3oool000006@3oool010000000oooo0?ooo`00000`0?ooo`030000O`3oool0oooo
0300oooo00D0001o0?ooo`3oool0oooo0000000L0?ooo`030000003oool0oooo02<0oooo001j0?oo
o`030000003oool0oooo01d0oooo0P00000H0?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`000000503oool010000000oooo0?ooo`00000I0?ooo`040000003oool0oooo00000300oooo00<0
001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`020?ooo`030000003oool0oooo01T0oooo00<0
00000?ooo`3oool08`3oool007X0oooo00<000000?ooo`3oool07@3oool2000001P0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0000000D0?ooo`040000003oool0oooo000001T0oooo00@0
00000?ooo`3oool00000<03oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo0080oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0S0?ooo`00NP3oool00`000000oooo0?oo
o`0M0?ooo`800000603oool00`000000oooo0?ooo`0G0?ooo`030000003oool0000001@0oooo00@0
00000?ooo`3oool000006@3oool010000000oooo0?ooo`00000`0?ooo`030000O`3oool0oooo0300
oooo00<0001o0?ooo`3oool00P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo02<0
oooo001j0?ooo`8000007P3oool2000001P0oooo00<000000?ooo`3oool05`3oool00`000000oooo
0000000D0?ooo`040000003oool0oooo000001T0oooo00@000000?ooo`3oool00000<03oool00`00
07l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo00<0oooo00<000000?ooo`3oool05`3oool20000
02D0oooo001j0?ooo`030000003oool0oooo01d0oooo0P00000H0?ooo`030000003oool0oooo01L0
oooo00<000000?ooo`000000503oool010000000oooo0?ooo`00000I0?ooo`040000003oool0oooo
00000300oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`030?ooo`030000003oool0
oooo01P0oooo00<000000?ooo`3oool08`3oool007X0oooo00<000000?ooo`3oool07@3oool20000
01P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000D0?ooo`040000003oool0oooo
000001T0oooo00@000000?ooo`3oool00000<03oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3o
ool0oooo00@0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0S0?ooo`00NP3oool0
0`000000oooo0?ooo`0M0?ooo`800000603oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
000001@0oooo00@000000?ooo`3oool000006@3oool010000000oooo0?ooo`00000`0?ooo`030000
O`3oool0oooo0300oooo00<0001o0?ooo`3oool0103oool00`000000oooo0?ooo`0G0?ooo`030000
003oool0oooo02<0oooo001j0?ooo`030000003oool0oooo01d0oooo0P00000H0?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`000000503oool010000000oooo0?ooo`00000I0?ooo`040000
003oool0oooo00000300oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`040?ooo`03
0000003oool0oooo01L0oooo00<000000?ooo`3oool08`3oool007X0oooo00<000000?ooo`3oool0
7@3oool2000001P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000D0?ooo`040000
003oool0oooo000001T0oooo00@000000?ooo`3oool00000<03oool00`0007l0oooo0?ooo`0`0?oo
o`030000O`3oool0oooo00D0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0S0?oo
o`00F@3oool00`3o_g@00000000000080000000307Boo`3oool0oooo0080oooo00H0omnL04P00000
000000000000B02Lgol;0?ooo`8000007P3oool2000001P0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0000000D0?ooo`040000003oool0oooo000001T0oooo00@000000?ooo`3oool00000
<03oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo00D0oooo00<000000?ooo`3oool0
5@3oool2000002D0oooo001I0?ooo`030?noM0000000000000P0000000H0M;oo0?ooo`3oool0oooo
0?oOW018000500000003001d_`3oool0oooo00P0oooo00<000000?ooo`3oool07@3oool2000001P0
oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000D0?ooo`040000003oool0oooo0000
01T0oooo00@000000?ooo`3oool00000<03oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0
oooo00D0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0S0?ooo`00G@3oool0103o
gi`0B0000000001d_ol70?ooo`090;md00000000001809cOo`3oool0omnL04P000000000B9cO00T0
oooo00<000000?ooo`3oool07@3oool2000001P0oooo00<000000?ooo`3oool05`3oool00`000000
oooo0000000D0?ooo`040000003oool0oooo000001T0oooo00@000000?ooo`3oool00000<03oool0
0`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo00H0oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`0S0?ooo`00G@3oool0103ooml0W4P000000018W=l60?ooo`040?oog`2LB000
000004RLg`<0oooo00@0omnL04P000000000M;oo203oool00`000000oooo0?ooo`0M0?ooo`800000
603oool00`000000oooo0?ooo`0G0?ooo`030000003oool0000001@0oooo00@000000?ooo`3oool0
00006@3oool010000000oooo0?ooo`00000`0?ooo`030000O`3oool0oooo0300oooo00<0001o0?oo
o`3oool01P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo02<0oooo001N0?ooo`03
0;md0000000007Bo00H0oooo00@0oono07@000000000M;oo103oool0103OW4P000000018W03Oool7
0?ooo`030000003oool0oooo01d0oooo0P00000H0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`000000503oool010000000oooo0?ooo`00000I0?ooo`030000003oool000000340oooo00<0
001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`060?ooo`030000003oool0oooo01D0oooo00<0
00000?ooo`3oool08`3oool005h0oooo00@0gia80000000007@0_ooo1@3oool0103ookl0M0000000
001d_ol40?ooo`040?oOW0180000001809cOo`L0oooo00<000000?ooo`3oool07@3oool2000001P0
oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000D0?ooo`040000003oool0oooo0000
01T0oooo00<000000?ooo`000000<@3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo
00L0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0S0?ooo`00GP3oool0103ogi`0
B0000000B02Lgol50?ooo`040?oog`2LB000000004RLg`@0oooo00@0oooO09a800000000B9cO1`3o
ool3000001d0001o0P00000H0000O`030000000007l0001o01L0001o00<000000000O`0000005@00
07l00`000000001o0000000I0000O`030000000007l0000006h0001o00<000000000O`0007l04P00
07l3000002D0oooo001N0?ooo`040?oo_`1d0000000007Boo`H0oooo00@0gia800000000B9`0gooo
103oool00`2oM0000000001d_`070?ooo`030000003oool0oooo01d0oooo0P00000H0?ooo`030000
003oool0oooo01L0oooo00<000000?ooo`0000005@3oool00`000000oooo0000000I0?ooo`030000
003oool000000340oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`070?ooo`030000
003oool0oooo01@0oooo00<000000?ooo`3oool08`3oool005h0oooo00@0oooO09a80000000007Bo
1P3oool0103o_g@000000000M02oool40?ooo`040=nLB000000004RL0=ooo`H0oooo00<000000?oo
o`3oool07@3oool2000001P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000E0?oo
o`030000003oool0000001T0oooo00<000000?ooo`000000<@3oool00`0007l0oooo0?ooo`0`0?oo
o`030000O`3oool0oooo00P0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0S0?oo
o`00G`3oool0103OW4P000000018W03Oool50?ooo`040?oo_`1d0000001809cOo`@0oooo00@0okmd
0000000004P0W=oo1P3oool00`000000oooo0?ooo`0M0?ooo`800000603oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0000001D0oooo00<000000?ooo`0000006@3oool00`000000oooo0000
000a0?ooo`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool0203oool00`000000oooo0?oo
o`0C0?ooo`030000003oool0oooo02<0oooo001O0?ooo`040?noM0000000001d0;ooo`H0oooo00<0
_g@000000000M;l0103oool0103o_g@000000000B02Lgol60?ooo`030000003oool0oooo01d0oooo
0P00000H0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`0000005@3oool00`000000oooo
0000000I0?ooo`030000003oool000000340oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo
0?ooo`080?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool08`3oool005/0oooo00P0
okmd000000000000001d0;oo_`1d0000000007Boo`H0oooo00@0omnL04P0000004P0W=oo0`3oool0
103OW4P000000000M02oool60?ooo`030000003oool0oooo01d0oooo0P00000H0?ooo`030000003o
ool0oooo01L0oooo0P00000F0?ooo`030000003oool0000001T0oooo00<000000?ooo`000000<@3o
ool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo00T0oooo00<000000?ooo`3oool04P3o
ool00`000000oooo0?ooo`0S0?ooo`00F`3oool00`3ogi`0B000000000040000000304RLg`3oool0
oooo00D0oooo00X0gia80000000004P0W=oo0?ooo`3ookl0M00000000000B9`0gooo1P3oool20000
01h0oooo0P00000H0?ooo`030000003oool0oooo01L0oooo0P00000F0?ooo`030000003oool00000
01T0oooo00<000000?ooo`000000<@3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo
00T0oooo00<000000?ooo`3oool04@3oool2000002D0oooo001N0?ooo`050=nLB000000000000000
0000M;l0203oool00`3OW4P00000000000040000000307Boo`3oool0oooo00D0oooo00<000000?oo
o`3oool07@3oool2000001P0oooo00<000000?ooo`3oool05`3oool2000001H0oooo00<000000?oo
o`0000006@3oool00`000000oooo0000000a0?ooo`030000O`3oool0oooo0300oooo00<0001o0?oo
o`3oool02@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo02<0oooo001P0?ooo`04
0?oog`2LB000001d0;ooo`P0oooo00H0oono07@00000000000000000B02Lgol80?ooo`030000003o
ool0oooo01d0oooo0P00000H0?ooo`030000003oool0oooo01L0oooo0P00000F0?ooo`030000003o
ool0000001T0oooo00<000000?ooo`000000<@3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3o
ool0oooo00T0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0S0?ooo`00NP3oool0
0`000000oooo0?ooo`0M0?ooo`800000603oool00`000000oooo0?ooo`0G0?ooo`8000005P3oool0
0`000000oooo0000000I0?ooo`030000003oool000000340oooo00<0001o0?ooo`3oool0<03oool0
0`0007l0oooo0?ooo`0:0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool08`3oool0
07X0oooo00<000000?ooo`3oool07@3oool2000001P0oooo00<000000?ooo`3oool05`3oool20000
01H0oooo00<000000?ooo`0000006@3oool00`000000oooo0000000a0?ooo`030000O`3oool0oooo
0300oooo00<0001o0?ooo`3oool02P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
02<0oooo001j0?ooo`030000003oool0oooo01d0oooo0P00000H0?ooo`030000003oool0oooo01L0
oooo0P00000F0?ooo`030000003oool0000001T0oooo00<000000?ooo`000000<@3oool00`0007l0
oooo0?ooo`0`0?ooo`030000O`3oool0oooo00X0oooo00<000000?ooo`3oool04@3oool00`000000
oooo0?ooo`0S0?ooo`00NP3oool2000001h0oooo0P00000H0?ooo`030000003oool0oooo01L0oooo
0P00000F0?ooo`030000003oool0000001T0oooo00<000000?ooo`000000<@3oool00`0007l0oooo
0?ooo`0`0?ooo`030000O`3oool0oooo00X0oooo00<000000?ooo`3oool0403oool2000002D0oooo
001j0?ooo`030000003oool0oooo01d0oooo0P00000H0?ooo`030000003oool0oooo01L0oooo0P00
000F0?ooo`030000003oool0000001T0oooo00<000000?ooo`000000<@3oool00`0007l0oooo0?oo
o`0`0?ooo`030000O`3oool0oooo00/0oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`0S0?ooo`00NP3oool00`000000oooo0?ooo`0M0?ooo`800000603oool00`000000oooo0?ooo`0G
0?ooo`8000005P3oool00`000000oooo0000000I0?ooo`030000003oool000000340oooo00<0001o
0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0;0?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool08`3oool007X0oooo00<000000?ooo`3oool07@3oool2000001P0oooo00<000000?oo
o`3oool05`3oool2000001H0oooo00<000000?ooo`0000006@3oool00`000000oooo0000000a0?oo
o`030000O`3oool0oooo0300oooo00<0001o0?ooo`3oool02`3oool00`000000oooo0?ooo`0@0?oo
o`030000003oool0oooo02<0oooo001j0?ooo`030000003oool0oooo01d0oooo0P00000H0?ooo`03
0000003oool0oooo01L0oooo0P00000F0?ooo`030000003oool0000001T0oooo00<000000?ooo`00
0000<@3oool00`0007l0oooo0?ooo`0`0?ooo`030000O`3oool0oooo00/0oooo00<000000?ooo`3o
ool0403oool00`000000oooo0?ooo`0S0?ooo`00NP3oool00`000000oooo0?ooo`0M0?ooo`800000
603oool00`000000oooo0?ooo`0G0?ooo`8000005P3oool00`000000oooo0000000I0?ooo`030000
003oool000000340oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0<0?ooo`030000
003oool0oooo00l0oooo00<000000?ooo`3oool08`3oool007X0oooo0P00000N0?ooo`800000603o
ool00`000000oooo0?ooo`0G0?ooo`8000005P3oool00`000000oooo0000000I0?ooo`030000003o
ool000000340oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0<0?ooo`030000003o
ool0oooo00h0oooo0P00000U0?ooo`00NP3oool00`000000oooo0?ooo`0M0?ooo`800000603oool0
0`000000oooo0?ooo`0G0?ooo`8000005P3oool00`000000oooo0000000I0?ooo`030000003oool0
00000340oooo00<0001o0?ooo`3oool0<03oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool08`3oool007X0oooo00<000000?ooo`3oool07@3oool20000
01P0oooo00<000000?ooo`3oool05`3oool2000001H0oooo00<000000?ooo`0000006@3oool00`00
0000oooo0000000a0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0303oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo02<0oooo001j0?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0:0?ooo`8000002`3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool2000000/0
oooo00<000000?ooo`3oool0203oool00`000000oooo0000000<0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`0000002P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0
oooo00@000000?ooo`3oool000002@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
02<0oooo001j0?ooool000004P00000U0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00X@3oool0
0`3OW4P0000000000003000000030000B02Lgol0oooo00<0oooo00<0oono07@0000000000P000000
0`1d_ol0oooO09a8000P000000040000B02Lgol0oono07@000<0000000<0M;oo0?oog`2LB0001000
00000`00B9`0gooo0?ooo`0=0?ooo`030?oog`2LB000000000H0000000<0001d0;ooo`3oool01@3o
ool00`3ookl0M000000000020000000307Boo`3ooml0W4P000<0000000<0B9cO0?noM00000001000
00000`0007@0W9a8000000040000000504RLg`3oool0oooo0?ooo`2oM0001@0000000`0007@0_ooo
0?ooo`060?ooo`030?noM0000000000000<0000000H0B9cO0?ooo`3oool0oooo0?oOW01800090000
00030000B02Lgol0oooo00<0oooo00<0gia80000000000000P0000000`00M;l0oooo0?ooo`030?oo
o`030?noM0000000000000<0000000<0001d09bLB0000000100000000`18W=l0oooo0?ooo`1:0?oo
o`00W`3oool00`3ogi`0B000000000080000000407Boo`3oool0omnL04P000D0000000<004RL09ad
000000003@0000000`1dM7@000000000000;0000000707AdM000000000000000000004P0W;nL04P0
0005000000030018W02LM000000000<0000000<004RL0=ooo`3oool03@3oool00`3ooml0W4P00000
0006000000030000M02oool0oooo00@0oooo00<0oono07@0000000002@0000000`18W=l0okmd0000
0004000000030000M02LW4P0000000@0000000@0B9cO0?ooo`3ookl0M000200000000`0007@0_ooo
0?ooo`030?ooo`030?oOW0180000000000H0000000@007Bo0?ooo`3ogi`0B0002@0000001P0004P0
W=oo0?ooo`3oool0oono07@000L0000000D0B9cO0?ooo`3oool0oooo0?noM005000000030000M02L
W4P0000000@0000000<0B9cO0?ooo`3oool0BP3oool009h0oooo00D0oooO09a800000000000004RL
g`040?ooo`090=nLB0000000000004RLg`3oool0gia80000000004P0W=oo0080oooo00H0oooO09a8
0000000000000000M02oool20?ooo`0P0?noM0000000001d0;ooo`3ookl0M0000000001d_ol0oooo
0?oOW0180000000007Boo`3oool0okmd0000000007@0_ooo0?oo_`1d0000000007Boo`3oool0omnL
04P000000000M;oo0?ooo`3OW4P000000000B02Lgol20?ooo`060?oog`2LB00000000000000007@0
_ooo503oool0103o_g@000000000M02oool90?ooo`040?oOW0180000000007Boo`80oooo00D0okmd
000000000000000004RLg`030?ooo`040?oOW0180000001809cOo`@0oooo00<0_g@000000000M;l0
0`3oool0102oM000000000000018W=l40?ooo`050=nLB000000000000018W03Oool00`3oool0103o
_g@000000000B02Lgol30?ooo`040?noM000000004RL0=ooo`D0oooo00@0gia800000000B9`0gooo
1P3oool0102oM000000000000000M;l30?ooo`040=nLB00000000000001d_`<0oooo00@0omnL04P0
000004P0W=oo103oool00`2oM0000000001d_`1>0?ooo`00WP3oool0103ogi`0B0000000001d_ol6
0?ooo`080;md0000000007Bo0?ooo`3o_g@000000000B02Lgol40?ooo`040?oog`2LB000001809cO
o`80oooo0200omnL04P0000004P0W=oo0?oog`2LB0000000001d_`3oool0oooO09a800000000B9cO
0?ooo`3ogi`0B0000000B02Lgol0oooO09a80000000007Bo0?ooo`3ooml0W4P000000018W=l0oooo
0?noM0000000001809cOo`@0oooo00@0oooO09a8000004P0W=oo503oool0103ogi`0B0000000B02L
gol90?ooo`040?oo_`1d0000000007Boo`@0oooo00<0_g@000000000M;l00`3oool0103ookl0M000
0000001d_ol40?ooo`090=nLB000000004RL0=ooo`3oool0oono07@00000000007Bo00H0oooo00<0
omnL04P00000M;l0103oool0103ogi`0B0000000B02Lgol<0?ooo`040?noM0000000001d0;ooo`D0
oooo00@0oono07@00000000007Bo1@3oool0103ogi`0B00000000018W=l20?ooo`040?oo_`1d0000
000007Boo`@0oooo00@0gia800000000B9`0goooC@3oool009h0oooo00@0okmd0000000004P0W=oo
1P3oool01P3OW4P000000018W03Oool0oooo0;md0080000000P0B9cO0?ooo`3oool0oooo0?oo_`1d
0000000007Boo`80oooo01/0oono07@000000000M;oo0?ooo`3OW4P000000018W03Oool0oooo0;md
0000000007Bo0?ooo`3ookl0M0000000001d_ol0oooo0=nLB000000004RL0=ooo`3oool0_g@00000
0000M;l00P3oool0102oM000000000000018W=l30?ooo`040?oo_`1d0000000007Booa@0oooo00P0
oono07@000000000M;oo0?ooo`3o_g@0000004RLg`D0oooo00@0oooO09a800000000B9cO103oool0
103o_g@000000000M02oool20?ooo`040?oog`2LB000000004RLg`@0oooo00T0okmd0000000007@0
_ooo0?ooo`3ookl0M00000000018W=l03@3oool0103ookl0M0000000001d_ol<0?ooo`040?oOW018
0000001809cOo`D0oooo00@0omnL04P000000000M;oo1P3oool02@3OW4P000000000M02oool0oooo
0?oog`2LB000000004RLg`040?ooo`040?noM0000000001d0;ooodd0oooo002N0?ooo`040?oOW018
0000001809cOo`80oooo00<0omnL04P000000000100000001000B9`0gooo0?ooo`2oM0080000000O
001d_`3oool0oooo0?oog`2LB000000004RLg`3oool0okmd0000000007@0_ooo0?ooo`3OW4P00000
0018W03Oool0oooO09a800000000B9cO0?ooo`3o_g@000000000M02oool0oooo0=nLB000000004RL
0=ooo`020?ooo`030;md00000000000000H0000000<007Bo0?ooo`3oool04P3oool02@3ooml0W4P0
00000018W=l0oooo0?noM0000000001d0;ooo`050?ooo`030;md0000000007Bo00@0oooo00@0omnL
04P0000004P0W=oo0`3oool00`3OW4P00000001d_`040?ooo`090?oOW0180000001809cOo`3oool0
oooO09a80000000007Bo00d0oooo00@0oooO09a80000000007Bo303oool0103ookl0M0000000001d
_ol50?ooo`040?oo_`1d0000000007Boo`H0oooo00@0omnL04P0000004P0W=oo0P3oool00`3OW4P0
0000001d_`040?ooo`040?oOW0180000001809cOodd0oooo002N0?ooo`040?oo_`1d0000000007Bo
o`80oooo00<0oooO09a800000000100000001P0007@0_ooo0?ooo`3oool0oooO09a800H0000001l0
04RL0=ooo`3oool0oooo0=nLB000000004RL0=ooo`3ogi`0B0000000B02Lgol0oooo0?noM0000000
001809cOo`3oool0gia800000000B9`0gooo0?oOW0180000001809cOo`3oool0okmd0000000004P0
W=oo00<0oooo00<0oooO09a8000000001@0000000`00B9`0gooo0?ooo`0C0?ooo`030;md00000000
000000<0000000<0001809cOo`3oool0103oool0103o_g@000000000M02oool30?ooo`040?oo_`1d
0000000007Boo`<0oooo00@0okmd00000000B9`0gooo0`3oool0103ookl0M0000000001d_ol20?oo
o`040=nLB0000000001d0;ooo`D0oooo00<0okmd00000000M;l01@3oool0103OW4P000000018W03O
ool;0?ooo`040?oog`2LB000000004RLg`H0oooo00@0_g@000000000B9`0gooo1@3oool0103ogi`0
B0000000001d_ol20?ooo`040?noM000000004RL0=ooo`<0oooo00@0oono07@000000000M;ooC@3o
ool009h0oooo00@0oooO09a800000000B9cO4`3oool0103o_g@000000000M02oool20?ooo`0K0?no
M0000000001d0;ooo`3ookl0M00000000018W=l0oooo0?oOW0180000000007Boo`3oool0okmd0000
000007@0_ooo0?oo_`1d0000000004RLg`3oool0omnL04P000000000M;oo00T0oooo00@0okmd0000
000007@0_ooo503oool00`3o_g@00000000000040000000307Boo`3oool0oooo00<0oooo00@0omnL
04P0000004P0W=oo0`3oool0103ooml0W4P000000018W=l30?ooo`040?oOW0180000000004RLg`<0
oooo00@0oooO09a800000000B9cO0P3oool0103ookl0M00000000000M;l50?ooo`040=nLB0000000
001d0;ooo`@0oooo00@0okmd0000000007@0_ooo303oool00`2oM0000000001d_`060?ooo`040?oo
_`1d0000000004RLg`D0oooo00@0gia80000000004P0W=oo0P3oool0103ogi`0B00000000018W=l3
0?ooo`040?oog`2LB000000004RLgdd0oooo002O0?ooo`040=nLB000000004RL0=oooa80oooo00@0
gia80000000004P0W=oo0P3oool01P3ogi`0B00000000000M;l0oooO09a80080000000`0M;oo0?oo
_`1d0000000004RLg`3oool0omnL04P00000000007Bo0?oog`2LB0020000000507Boo`3ookl0M000
00000018W=l02@3oool0103OW4P000000000B02LgolD0?ooo`080?oOW0180000001809cOo`3oool0
_g@000000000M;l50?ooo`040?oo_`1d0000000007Boo`@0oooo00@0_g@000000000B9`0gooo0P3o
ool01@3ookl0M00000000000000007Bo0080oooo00@0oono07@000000000B9cO0`3oool01@3ogi`0
B00000000000B9`0gooo0080oooo00D0oono07@000000000000007Boo`040?ooo`040?oOW0180000
000007Boo``0oooo00@0gia800000000B9`0gooo1P3oool01@3o_g@0000000000000B9`0gooo0080
oooo00D0okmd00000000000004RL0=ooo`020?ooo`050?oo_`1d0000000000000000M;l00P3oool0
103ookl0M00000000018W=m=0?ooo`00W`3oool0103ookl0M0000000001d_ol60?ooo`030?oo_`1d
000007Bo0080oooo00<0oono07@0000000001P0000001000B9`0gooo0?oo_`1d000<0000000307Bo
_`1d0000000000/0000000D0M;oo0?ooo`3oool0oono07@00007000000030018W03Oool0oooo01<0
oooo00P0oono07@000000000M;oo0?ooo`3o_g@0000004RLg`<0oooo00<0omnL04P0000000000P00
00000`00M;l0oooo0?noM004000000040000M02oool0oooO09a800<0000000<004Q804P000000000
0`0000000`0004P0W=oo0?ooo`030?ooo`030?oOW0180000000000L0000000@0B9cO0?ooo`3oool0
gia82@0000001P00B9`0gooo0?ooo`3oool0omnL04P000@0000000<0001d0;ooo`3oool01P3oool0
0`3ogi`0B00000000005000000050000B02Lgol0oooo0?oog`2LB0000`0000000`00B4P0B0000000
0003000000030000B02Lgol0oooo04`0oooo002P0?ooo`040?noM0000000001809cOo`D0oooo00@0
omnL04P0000004P0W=oo0P3oool00`3ookl0M00000000004000000050018W03Oool0oooo0?oo_`1d
00000`0000001P18W;l0W4P00000000007@0_ooO09a80080000000<007Bo0?oo_`1d00000`000000
1P18W;l0W4P00000000007@0_ooO09a80080000000<007Bo0?ooo`3oool00P3oool00`3ookl0M000
00000004000000030018W03Oool0oooo01@0oooo00@0oooO09a800000000B9cO1@3oool0103OW4P0
00000018B0180003000000030018W03Oool0okmd00@0000000@0001809cOo`3ooml0W4P00`000000
0`00B9`0gooO09a80002000000030000B02Lgol0oooo00H0oooo00<0okmd0000000000000P000000
100007@0_ono07@00018W=l20?ooo`030?oOW0180000000000L0000000H0001d0;ooo`3oool0oooo
0?oOW0180004000000030000B02Lgol0oooo00P0oooo00<0okmd0000000000000P0000001P00M;l0
oooo0?ooo`3oool0oooO09a800<0000000<004RL0=oog`2LB0000P0000000`0004P0W=oo0?ooo`1=
0?ooo`00X@3oool0103OW4P0000000000018W=l30?ooo`050?oog`2LB00000000000001d_ol0B`3o
ool00`2oM0000000001d_`050?ooo`040?noM0000000001d0;ooobd0oooo00@0gia800000000B9`0
goooN`3oool00:80oooo00<0okmd0000000000001`0000000`18W=l0oooo0?ooo`170?ooo`030;md
00000000000000X0000000<0M;oo0?ooo`3oool0:`3oool0103o_g@000000000B02Lgomk0?ooo`00
Y03oool00`3OW4P000000000000200000004001d_`3ooml0W4P004RLgdT0oooo00<0_g@000000000
00002P0000000`00M;l0oooo0?ooo`0[0?ooo`040?oOW0180000000007Boo`/0oooo00@0omnL04P0
0000000007BoK03oool00?l0oooo<`3oool00`2oM000001809cOo`0;0?ooo`050?oo_`1d00000000
0000M02oool0J`3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-9.98417, -25.0401, \
0.0579561, 0.250281}}]
}, Closed]],

Cell[BoxData[
    \(\($TextStyle = {FontFamily \[Rule] "\<Times\>", 
          FontSize \[Rule] 12};\)\)], "Input"],

Cell[TextData[{
  "Rather than reset the default ",
  StyleBox["$TextStyle", "InlineInput",
    FontWeight->"Bold"],
  " globally, it is often preferred to modify this variable locally instead.  \
This can be done easily using ",
  StyleBox["Block", "InlineInput",
    FontWeight->"Bold"],
  " but, as previously noted in ",
  StyleBox["functions.nb",
    FontWeight->"Bold",
    FontSlant->"Italic"],
  ", does not work in ",
  StyleBox["Module", "InlineInput",
    FontWeight->"Bold"],
  " or ",
  StyleBox["With", "InlineInput",
    FontWeight->"Bold"],
  ".  Of course, one can still embed a ",
  StyleBox["Module", "InlineInput",
    FontWeight->"Bold"],
  " or ",
  StyleBox["With", "InlineInput",
    FontWeight->"Bold"],
  " statement within ",
  StyleBox["Block", "InlineInput",
    FontWeight->"Bold"],
  ".  We use ",
  StyleBox["Block", "InlineInput",
    FontWeight->"Bold"],
  " to temporarily modify system variables or defaults.  "
}], "Text"],

Cell[TextData[{
  StyleBox[
  "Verify that the following statement does not affect the default value for \
"],
  StyleBox["$TextStyle",
    FontWeight->"Bold"],
  StyleBox[" or the appearance of other plots."]
}], "Exercise"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Block[{$TextStyle = {FontFamily \[Rule] "\<Times\>", 
              FontSize \[Rule] 16, FontSlant \[Rule] "\<Italic\>", \ 
              FontWeight \[Rule] "\<Bold\>", 
              FontColor \[Rule] RGBColor[0, 1, 0]}}, 
        Plot[Gamma[x], {x, \(-5\), 5}, \ GridLines \[Rule] Automatic, 
          Frame \[Rule] True, \ 
          FrameLabel \[Rule] {"\<x\>", \ "\<\[CapitalGamma][x]\>"}, \ 
          PlotLabel \[Rule] "\<Gamma Function\>"]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.0952381 0.316178 0.0220537 [
[.11905 -0.0125 -12.1563 -19.875 ]
[.11905 -0.0125 12.1563 0 ]
[.30952 -0.0125 -12.1563 -19.875 ]
[.30952 -0.0125 12.1563 0 ]
[.5 -0.0125 -6 -19.875 ]
[.5 -0.0125 6 0 ]
[.69048 -0.0125 -6 -19.875 ]
[.69048 -0.0125 6 0 ]
[.88095 -0.0125 -6 -19.875 ]
[.88095 -0.0125 6 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -6.46875 -39.75 ]
[.5 -0.0125 6.46875 -19.875 ]
[-0.0125 .09564 -32.3125 -9.9375 ]
[-0.0125 .09564 0 9.9375 ]
[-0.0125 .20591 -24.4375 -9.9375 ]
[-0.0125 .20591 0 9.9375 ]
[-0.0125 .31618 -12 -9.9375 ]
[-0.0125 .31618 0 9.9375 ]
[-0.0125 .42645 -12.125 -9.9375 ]
[-0.0125 .42645 0 9.9375 ]
[-0.0125 .53672 -20 -9.9375 ]
[-0.0125 .53672 0 9.9375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -52.1875 -18.4688 ]
[-0.01875 .30902 -32.3125 18.4688 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -60.7188 0 ]
[.5 .63053 60.7188 19.875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.11905 0 m
.11905 .61803 L
s
.30952 0 m
.30952 .61803 L
s
.5 0 m
.5 .61803 L
s
.69048 0 m
.69048 .61803 L
s
.88095 0 m
.88095 .61803 L
s
0 .09564 m
1 .09564 L
s
0 .20591 m
1 .20591 L
s
0 .31618 m
1 .31618 L
s
0 .42645 m
1 .42645 L
s
0 .53672 m
1 .53672 L
s
0 g
.11905 0 m
.11905 .00625 L
s
gsave
.11905 -0.0125 -73.1563 -23.875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 27.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 19.438 moveto
%%IncludeResource: font Mathematica1-Bold
%%IncludeFont: Mathematica1-Bold
/Mathematica1-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(-) show
75.313 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.30952 0 m
.30952 .00625 L
s
gsave
.30952 -0.0125 -73.1563 -23.875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 27.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 19.438 moveto
%%IncludeResource: font Mathematica1-Bold
%%IncludeFont: Mathematica1-Bold
/Mathematica1-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(-) show
75.313 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -67 -23.875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 27.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.69048 0 m
.69048 .00625 L
s
gsave
.69048 -0.0125 -67 -23.875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 27.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.88095 0 m
.88095 .00625 L
s
gsave
.88095 -0.0125 -67 -23.875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 27.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.2619 0 m
.2619 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.7381 0 m
.7381 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.07143 0 m
.07143 .00375 L
s
.02381 0 m
.02381 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.97619 0 m
.97619 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -67.4688 -43.75 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 27.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.813 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(x) show
71.938 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .09564 m
.00625 .09564 L
s
gsave
-0.0125 .09564 -93.3125 -13.9375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 27.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 19.438 moveto
%%IncludeResource: font Mathematica1-Bold
%%IncludeFont: Mathematica1-Bold
/Mathematica1-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(-) show
75.313 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
0 .20591 m
.00625 .20591 L
s
gsave
-0.0125 .20591 -85.4375 -13.9375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 27.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 19.438 moveto
%%IncludeResource: font Mathematica1-Bold
%%IncludeFont: Mathematica1-Bold
/Mathematica1-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(-) show
75.313 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .31618 m
.00625 .31618 L
s
gsave
-0.0125 .31618 -73 -13.9375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 27.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .42645 m
.00625 .42645 L
s
gsave
-0.0125 .42645 -73.125 -13.9375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 27.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .53672 m
.00625 .53672 L
s
gsave
-0.0125 .53672 -81 -13.9375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 27.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .1177 m
.00375 .1177 L
s
0 .13975 m
.00375 .13975 L
s
0 .1618 m
.00375 .1618 L
s
0 .18386 m
.00375 .18386 L
s
0 .22796 m
.00375 .22796 L
s
0 .25002 m
.00375 .25002 L
s
0 .27207 m
.00375 .27207 L
s
0 .29412 m
.00375 .29412 L
s
0 .33823 m
.00375 .33823 L
s
0 .36029 m
.00375 .36029 L
s
0 .38234 m
.00375 .38234 L
s
0 .40439 m
.00375 .40439 L
s
0 .4485 m
.00375 .4485 L
s
0 .47055 m
.00375 .47055 L
s
0 .49261 m
.00375 .49261 L
s
0 .51466 m
.00375 .51466 L
s
0 .07359 m
.00375 .07359 L
s
0 .05153 m
.00375 .05153 L
s
0 .02948 m
.00375 .02948 L
s
0 .00743 m
.00375 .00743 L
s
0 .55877 m
.00375 .55877 L
s
0 .58082 m
.00375 .58082 L
s
0 .60288 m
.00375 .60288 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -121.719 -13.9375 Mabsadd m
1 1 Mabs scale
currentpoint translate
79.4688 13.9375 translate
90 rotate
-79.4688 -13.9375 translate
0 27.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
%%BeginResource: font Times-BoldItalic-MISO
%%BeginFont: Times-BoldItalic-MISO
/Times-BoldItalic /Times-BoldItalic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-BoldItalic-MISO
%%IncludeFont: Times-BoldItalic-MISO
/Times-BoldItalic-MISO findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 19.438 moveto
%%IncludeResource: font Mathematica1-Bold
%%IncludeFont: Mathematica1-Bold
/Mathematica1-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(G) show
%%IncludeResource: font Mathematica2-Bold
%%IncludeFont: Mathematica2-Bold
/Mathematica2-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
75.500 19.438 moveto
(@) show
82.063 19.438 moveto
%%IncludeResource: font Times-BoldItalic-MISO
%%IncludeFont: Times-BoldItalic-MISO
/Times-BoldItalic-MISO findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(x) show
%%IncludeResource: font Mathematica2-Bold
%%IncludeFont: Mathematica2-Bold
/Mathematica2-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
90.188 19.438 moveto
(D) show
95.938 19.438 moveto
%%IncludeResource: font Times-BoldItalic-MISO
%%IncludeFont: Times-BoldItalic-MISO
/Times-BoldItalic-MISO findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.11905 .61178 m
.11905 .61803 L
s
.30952 .61178 m
.30952 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.69048 .61178 m
.69048 .61803 L
s
.88095 .61178 m
.88095 .61803 L
s
.125 Mabswid
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.2619 .61428 m
.2619 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.7381 .61428 m
.7381 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.07143 .61428 m
.07143 .61803 L
s
.02381 .61428 m
.02381 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.97619 .61428 m
.97619 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .09564 m
1 .09564 L
s
.99375 .20591 m
1 .20591 L
s
.99375 .31618 m
1 .31618 L
s
.99375 .42645 m
1 .42645 L
s
.99375 .53672 m
1 .53672 L
s
.125 Mabswid
.99625 .1177 m
1 .1177 L
s
.99625 .13975 m
1 .13975 L
s
.99625 .1618 m
1 .1618 L
s
.99625 .18386 m
1 .18386 L
s
.99625 .22796 m
1 .22796 L
s
.99625 .25002 m
1 .25002 L
s
.99625 .27207 m
1 .27207 L
s
.99625 .29412 m
1 .29412 L
s
.99625 .33823 m
1 .33823 L
s
.99625 .36029 m
1 .36029 L
s
.99625 .38234 m
1 .38234 L
s
.99625 .40439 m
1 .40439 L
s
.99625 .4485 m
1 .4485 L
s
.99625 .47055 m
1 .47055 L
s
.99625 .49261 m
1 .49261 L
s
.99625 .51466 m
1 .51466 L
s
.99625 .07359 m
1 .07359 L
s
.99625 .05153 m
1 .05153 L
s
.99625 .02948 m
1 .02948 L
s
.99625 .00743 m
1 .00743 L
s
.99625 .55877 m
1 .55877 L
s
.99625 .58082 m
1 .58082 L
s
.99625 .60288 m
1 .60288 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .31618 m
1 .31618 L
s
.5 0 m
.5 .61803 L
s
gsave
.5 .63053 -121.719 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 27.875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(Gamma) show
119.188 19.438 moveto
(Function) show
180.438 19.438 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02498 0 m
.02499 .30097 L
.02605 .30806 L
.02729 .31081 L
.02846 .31207 L
.02954 .31277 L
.03053 .31322 L
.03163 .31358 L
.03279 .31386 L
.03395 .31407 L
.0352 .31425 L
.03746 .31449 L
.03884 .31459 L
.04016 .31468 L
.04262 .31479 L
.045 .31487 L
.0463 .31491 L
.04753 .31493 L
.04868 .31495 L
.04993 .31497 L
.05105 .31499 L
.0521 .315 L
.05335 .315 L
.05401 .31501 L
.05472 .31501 L
.05546 .31501 L
.05617 .31501 L
.05749 .31501 L
.05872 .31501 L
.05986 .31501 L
.06111 .315 L
.06244 .31499 L
.06504 .31496 L
.06742 .31493 L
.07003 .31488 L
.07277 .31482 L
.07741 .31469 L
.08247 .31447 L
.08537 .31431 L
.08801 .31413 L
.09085 .3139 L
.09392 .31357 L
.09678 .31317 L
.09941 .31269 L
.10192 .31208 L
.10319 .31169 L
.10458 .31119 L
.10589 .31061 L
.10711 .30996 L
.10837 .30912 L
Mistroke
.10951 .30816 L
.11064 .30696 L
.11169 .30549 L
.11281 .30338 L
.11402 .30005 L
.11524 .29447 L
.11594 .28934 L
.11659 .28193 L
.1179 .24147 L
.11863 .10742 L
Mfstroke
.11863 .10742 m
.11924 .61803 L
s
.1195 .61803 m
.1205 .37806 L
.12179 .3495 L
.12302 .3397 L
.12415 .33484 L
.1252 .33195 L
.12634 .32978 L
.1274 .32827 L
.12838 .32721 L
.13065 .32545 L
.13179 .32481 L
.13304 .32425 L
.1354 .32343 L
.13793 .32282 L
.14031 .32241 L
.14247 .32213 L
.14479 .32191 L
.1461 .32181 L
.14733 .32174 L
.14863 .32168 L
.14981 .32164 L
.15047 .32162 L
.15118 .32161 L
.15247 .32159 L
.15358 .32158 L
.15479 .32159 L
.15607 .3216 L
.15726 .32163 L
.15837 .32166 L
.15939 .32169 L
.16169 .32179 L
.1642 .32194 L
.16653 .32212 L
.1693 .32239 L
.17182 .32269 L
.17698 .32351 L
.17982 .32411 L
.1825 .32481 L
.18712 .3264 L
.18975 .32761 L
.19221 .32905 L
.1948 .33099 L
.19716 .33331 L
.19847 .33491 L
.19989 .33701 L
.20123 .3394 L
.20249 .34218 L
.20366 .34535 L
.20492 .3497 L
.20607 .35483 L
Mistroke
.20712 .36102 L
.20827 .37022 L
.20952 .38546 L
.21083 .41327 L
.21206 .46886 L
Mfstroke
.21206 .46886 m
.21308 .61803 L
s
.21321 .61803 m
.21324 0 L
s
.21552 0 m
.21634 .14124 L
.2175 .20237 L
.21873 .23237 L
.21978 .24731 L
.22094 .25829 L
.22215 .26631 L
.22329 .27179 L
.22453 .27638 L
.22586 .28015 L
.22725 .28324 L
.22857 .28557 L
.23104 .28886 L
.23335 .29108 L
.23548 .29262 L
.23782 .2939 L
.2403 .2949 L
.24277 .29562 L
.24416 .29593 L
.24543 .29615 L
.24661 .29631 L
.2479 .29645 L
.24911 .29653 L
.25023 .29658 L
.25092 .29659 L
.25167 .29659 L
.25237 .29658 L
.25301 .29656 L
.25424 .2965 L
.25553 .29639 L
.25692 .29624 L
.25821 .29606 L
.26116 .2955 L
.26377 .29484 L
.26619 .29406 L
.26879 .29303 L
.27165 .29162 L
.27646 .28846 L
.2791 .28618 L
.28157 .28356 L
.28396 .28046 L
.28624 .27684 L
.28831 .27281 L
.29051 .26745 L
.2929 .25995 L
.29511 .25064 L
.29647 .24331 L
.29775 .23481 L
.299 .22445 L
.30014 .21255 L
Mistroke
.3013 .19698 L
.30254 .17446 L
.3037 .14491 L
.30479 .10353 L
.30606 .02245 L
Mfstroke
.30606 .02245 m
.30623 0 L
s
.30951 0 m
.30959 .61803 L
s
.31325 .61803 m
.31377 .5745 L
.31452 .53772 L
.31521 .51221 L
.31648 .47896 L
.31763 .45783 L
.31895 .43994 L
.32021 .4272 L
.32245 .41078 L
.32358 .40457 L
.32482 .39886 L
.32701 .39091 L
.32907 .38525 L
.33149 .38014 L
.3337 .37656 L
.33625 .3734 L
.33868 .37113 L
.33993 .3702 L
.34127 .36935 L
.34246 .36872 L
.34372 .36817 L
.34485 .36777 L
.3461 .36742 L
.34723 .3672 L
.34828 .36705 L
.34892 .367 L
.34961 .36696 L
.35087 .36697 L
.35209 .36706 L
.35322 .36722 L
.35444 .36746 L
.35579 .36783 L
.35851 .36888 L
.3609 .37016 L
.36308 .37164 L
.36544 .37362 L
.36801 .37629 L
.37235 .38225 L
.37471 .38649 L
.37691 .39123 L
.37937 .39769 L
.38197 .40627 L
.38483 .41853 L
.38745 .43369 L
.3899 .45292 L
.39219 .47794 L
.3935 .49701 L
.3947 .51886 L
.39609 .55172 L
.3974 .59444 L
Mistroke
Mfstroke
.3974 .59444 m
.39792 .61803 L
s
.40444 .61803 m
.40452 0 L
s
.41171 0 m
.41218 .0211 L
.41288 .04545 L
.41353 .0644 L
.4148 .09408 L
.416 .116 L
.41716 .13311 L
.41843 .14837 L
.42066 .16903 L
.42325 .18655 L
.42567 .19862 L
.42797 .20755 L
.43005 .21403 L
.43233 .21976 L
.43473 .22461 L
.43738 .22883 L
.43983 .23186 L
.44246 .23434 L
.44397 .23544 L
.44535 .23626 L
.44655 .23684 L
.44785 .23734 L
.44856 .23755 L
.44922 .23771 L
.45052 .23792 L
.45164 .238 L
.45287 .23798 L
.45415 .23783 L
.45535 .23757 L
.45653 .23721 L
.45779 .2367 L
.46005 .23544 L
.46274 .23334 L
.46519 .2308 L
.46774 .22737 L
.47052 .22261 L
.47313 .21685 L
.47554 .21016 L
.47761 .20305 L
.47982 .19357 L
.48226 .18014 L
.48453 .16348 L
.48575 .15223 L
.48708 .13745 L
.48826 .12155 L
.48951 .10063 L
.49074 .07422 L
.49186 .04329 L
.49299 .00213 L
s
.49299 .00213 m
.49303 0 L
s
.50068 0 m
.50068 .61803 L
s
.50675 .61803 m
.50762 .58072 L
.50876 .54517 L
.51008 .51386 L
.51153 .48796 L
.51289 .46901 L
.51417 .45451 L
.5166 .43324 L
.51881 .41878 L
.52119 .40665 L
.52376 .39633 L
.52839 .38259 L
.53266 .37354 L
.53735 .3661 L
.54238 .36007 L
.54713 .35566 L
.55147 .35243 L
.55654 .34934 L
.56207 .34665 L
.56725 .34459 L
.57212 .34299 L
.58151 .34059 L
.58692 .33953 L
.59188 .33871 L
.60198 .33743 L
.60729 .33694 L
.61288 .33652 L
.61787 .33623 L
.62241 .33602 L
.62493 .33593 L
.62765 .33585 L
.63023 .3358 L
.63137 .33577 L
.6326 .33576 L
.63375 .33574 L
.63479 .33573 L
.63585 .33572 L
.63685 .33572 L
.63801 .33571 L
.63908 .33571 L
.64026 .33571 L
.64152 .33571 L
.64274 .33572 L
.64344 .33573 L
.64407 .33573 L
.64647 .33576 L
.64866 .3358 L
.65096 .33585 L
.65593 .33599 L
.66045 .33616 L
Mistroke
.67062 .33667 L
.68004 .33732 L
.68896 .33809 L
.70889 .34039 L
.72732 .34334 L
.746 .34731 L
.76661 .35313 L
.78598 .3604 L
.80684 .37082 L
.82691 .38426 L
.84801 .40339 L
.86777 .4277 L
.88615 .45795 L
.90663 .50371 L
.92522 .56051 L
Mfstroke
.92522 .56051 m
.93903 .61803 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00n@3oool0103ookl0M?l0003o0000okl40?ooo`040;oo0000o`000?m8
09cooj/0oooo003i0?ooo`080?ooW018o`000?l0003o0000oi`0gooo0?ooo`3oog@3003o0003003o
B02Lool0oooo0:T0oooo003i0?ooo`080?oog`2Lo`000?md0;ooo`3ooi`0B?nL0=oog`2Lo`02003o
000404Sog`3ooml0W?nL0=ooojT0oooo003n0?ooo`090?ooW01dog@0B?l0003o0000og@0_ooo0?oo
o`3oog@0M?oo0:T0oooo003o0?ooo`040=ooB000o`000?l0003o_jd0oooo003n0?ooo`050?oog`2L
o`000?l0003o001dool0[@3oool00?h0oooo00D0oonL04So0000o`000?md0;ooo`2]0?ooo`00oP3o
ool01@2oo`000?l0003o0000og@0M?no0:d0oooo003m0?ooo`0;0?ooW018o`000?l0003o001dool0
oomd07Coo`3ooml0W?l0003oW03Oool0Z03oool00?/0oooo00<0oono07Co0000o`000P00o`001000
og@0_ooo0?ooo`2oo`02003o0003003oB02Lool0oooo0:L0oooo003m0?ooo`040?oog`2Lo`000?l0
04Sog`<0oooo00@0gom8003o0000oi`0goooZ03oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0
oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0
oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0
oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0
oooo/@3oool00?l0oooo/@3oool009l0oooo00@0oomd003o0000o`00M?oo;03oool00`3ooml0W?l0
003o0006003o0003003o_`3oool0oooo0280oooo00D0oono07Co0000o`000?l0003o_`0/0?ooo`03
0?oog`2Lo`000?l000H00?l000<00?no0?ooo`3oool0;@3oool0103oog@00?l0003o001doom50?oo
o`00W`3oool0103ookl0M?l0003o0000okl]0?ooo`030?oog`2Lo`000?l000D00?l000<00?m809co
o`3oool08@3oool0203ookl0M?l0003oW03Oool0oonL04So0000og@0_ooo:`3oool00`3ooml0W?l0
003o0005003o0003003oB02Lool0oooo02d0oooo00@0oono07Co0000o`000?noA@3oool009l0oooo
00D0oooO09co0000o`000?nL0=ooo`0]0?ooo`030?oog`2Lo`000?l000D00?l000<0B?oO0?ooo`3o
ool0803oool00`3oog@00?l0003o_`020?ooo`040;oo0000o`000?md0;ooob/0oooo00<0oooO09co
0000o`001@00o`000`18oml0oooo0?ooo`0/0?ooo`050?oog`2Lo`000?l0003oW03Oool0A03oool0
09/0oooo00<0gom8003o0000o`001P00o`000`00og@0_ooo0?ooo`0[0?ooo`030?oog`2Lo`00B?oO
00<0oooo00<0oonL04SoW03Oool08@3oool01P2oo`000?l0003oW03Oool0oooo0?ooM0800?l000<0
0?no0?ooo`3oool0:P3oool00`3ooml0W?l004Sog`030?ooo`030?ooW018oi`0gooo02T0oooo00<0
gom8003o0000o`001P00o`000`00og@0_ooo0?ooo`110?ooo`00V`3oool00`3oog@00?l0003o0007
003o000307Coo`3oool0oooo02/0oooo00<0oooO09co0000okl00`3oool00`3ooi`0W?oo0?ooo`0P
0?ooo`0:0;oo0000o`000?md0;ooo`3oool0oono07Co0000o`000?m809coob`0oooo00<0oooO09co
0000okl00`3oool00`3ooi`0W?oo0?ooo`0X0?ooo`030?ooM000o`000?l000L00?l000<0M?oo0?oo
o`3oool0@03oool008X0oooo00<0gom8003o0000o`002P00o`000`00oi`0gooo0?ooo`020?ooo`09
0;ooB02Lool0oooo0?ooo`3ooml0W?l0003o0000oi`0gooo01T0oooo00<0gom8003o0000o`002P00
o`000`00oi`0gooo0?ooo`060?ooo`040?oog`2Lo`000?nL0=ooob@0oooo00@0_ol0003o0000o`00
M?oo0P3oool0102oo`000?l0003o0000okl]0?ooo`040?oog`2Lo`000?nL0=ooobd0oooo00T0_om8
09coo`3oool0oooo0?oog`2Lo`000?l0003oW03Oool0@`3oool008X0oooo00<0gom8003o0000o`00
2P00o`000`00oi`0gooo0?ooo`020?ooo`030?oog`2Log@0_ooo0080oooo00@0_ol0003o0000og@0
_ooo6@3oool00`3OodP00?l0003o000:003o0003003oW03Oool0oooo00L0oooo00@0oooO09co0000
og@0_ooo8`3oool0103OodP00?l0003o0018oml20?ooo`050?ooM000o`000?l0003oM02oool0;@3o
ool0103ooml0W?l0003oM02oool/0?ooo`030?oog`2Log@0_ooo0080oooo00@0_ol0003o0000og@0
_ooo@`3oool009d0oooo00D0oono07CoM02oool0oooo0?ooM002003o000307Coo`3oool0oooo02l0
oooo00@0oono07Co0000og@0_ooo8P3oool0203oog@00?l0003o0000oi`0gooo0?ooo`3ookl0M?l0
0P00o`000`1dool0oooo0?ooo`0/0?ooo`040?oo_`1do`000?md0;ooob`0oooo00D0oono07CoM02o
ool0oooo0?ooM002003o000307Coo`3oool0oooo0440oooo002N0?ooo`070?oo_`1doi`0gooo0?oo
_`1do`000?l0003o_`0b0?ooo`040?ooW018o`000?nL0=ooob40oooo00D0oono07Co0000o`000?m8
09coo`020?ooo`040;oo0000o`000?l0003o_bl0oooo00@0oonL04So0000oi`0gooo;03oool01`3o
okl0M?nL0=ooo`3ookl0M?l0003o0000okl0@`3oool009l0oooo00L0oonL04So_`3oool0_ol0003o
0000og@0_ooo0380oooo00<0_ol0003o0000okl08@3oool01@3ooml0W?l0003o0000o`00B?oO0080
oooo00D0oomd003o0000o`000?nL0=ooo`0_0?ooo`030;oo0000o`000?no02d0oooo00L0oonL04So
_`3oool0_ol0003o0000og@0_ooo0480oooo002P0?ooo`030?ooM018oml0gom800800?l000<0M?oo
0?ooo`3oool0<03oool0103OodP00?l0003oB02LoolQ0?ooo`0;0?ooM000o`000?l0003oW03Oool0
oooo0?oo_`1do`000?l0003oW03Oool0;`3oool0103OodP00?l0003oB02Lool]0?ooo`030?ooM018
oml0gom800800?l000<0M?oo0?ooo`3oool0@03oool00:40oooo00D0gom807CoW018o`000?l004So
g`0b0?ooo`040?ooM000o`000?l004Sogb40oooo00D0oooO09co0000o`000?m809coo`020?ooo`04
0;oo0000o`000?md0;ooobl0oooo00@0oomd003o0000o`00B?oO;P3oool01@3OodP0M?nL04So0000
o`00B?oO0480oooo002R0?ooo`050;oo0000o`000?l0003oW03Oool0:`3oool00`3ookl0M?no0?oo
o`030?ooo`040=ooB000o`000?l004Sogb80oooo00@0oomd003o0000o`00B?oO0P3oool0103OodP0
0?l0003oW03OoolY0?ooo`030?oo_`1dokl0oooo00<0oooo00@0gom8003o0000o`00B?oO;`3oool0
1@2oo`000?l0003o0000oi`0gooo0440oooo002S0?ooo`040;oo0000o`000?m809coob`0oooo00H0
gom8003o_`3oool0oooo0?ooW018o`02003o000304Sog`3oool0oooo0240oooo00P0_ol0003o0000
og@0_ooo0?ooo`3oog@00?l0003o_b/0oooo00H0gom8003o_`3oool0oooo0?ooW018o`02003o0003
04Sog`3oool0oooo02h0oooo00@0_ol0003o0000odP0W?oo@@3oool00:<0oooo00@0oooO09co0000
o`00B?oO;03oool00`3ooml0W?l0003o0004003o0003003oM02oool0oooo02<0oooo00L0_ol0003o
0018oml0oooo0=ooB000odP0W?oo02/0oooo00<0oooO09co0000o`001000o`000`00og@0_ooo0?oo
o`0_0?ooo`040?oog`2Lo`000?l004Sogd40oooo002T0?ooo`030?oo_`1do`000?no02h0oooo00H0
gom8003o0000o`000?l0003oW03OoolV0?ooo`050=ooB000o`000?l0003oB02Lool0;P3oool01P3O
odP00?l0003o0000o`000?nL0=oooc80oooo00<0oono07Co0000okl0@@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool007L0ooooo`00000E
000002D0oooo001g0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool20000
00/0oooo00<000000?ooo`3oool02@3oool2000000`0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0000000;0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`000000303oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo02<0oooo001g0?ooo`8000001@3oool00`000000
oooo0?ooo`0G0?ooo`030000O`3oool0oooo01L0oooo0P00000G0?ooo`8000006@3oool00`000000
oooo0000000F0?ooo`030000003oool0000003<0oooo00<0001o0?ooo`3oool0<@3oool00`0007l0
oooo0?ooo`0M0?ooo`8000009@3oool007L0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`0G0?ooo`030000O`3oool0oooo01L0oooo0P00000G0?ooo`8000006@3oool010000000oooo
0?ooo`00000E0?ooo`030000003oool0000003<0oooo00<0001o0?ooo`3oool0<@3oool00`0007l0
oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001g0?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool05`3oool00`0007l0oooo0?ooo`0G0?ooo`8000005`3oool2000001T0oooo
00@000000?ooo`3oool000005@3oool00`000000oooo0000000c0?ooo`030000O`3oool0oooo0340
oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00M`3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo01L0oooo00<0001o0?ooo`3oool05`3oool2000001L0oooo
0P00000I0?ooo`040000003oool0oooo000001D0oooo00<000000?ooo`000000<`3oool00`0007l0
oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool007L0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0G0?ooo`030000O`3oool0oooo01L0oooo
0P00000G0?ooo`8000006@3oool010000000oooo0?ooo`00000D0?ooo`040000003oool0oooo0000
03<0oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo
02<0oooo001g0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool05`3oool00`0007l0
oooo0?ooo`0G0?ooo`8000005`3oool2000001T0oooo00@000000?ooo`3oool00000503oool01000
0000oooo0?ooo`00000c0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3oool0
0`000000oooo0?ooo`0S0?ooo`00M`3oool2000000D0oooo00<000000?ooo`3oool05`3oool00`00
07l0oooo0?ooo`0G0?ooo`8000005`3oool2000001T0oooo00@000000?ooo`3oool00000503oool0
10000000oooo0?ooo`00000c0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07@3o
ool2000002D0oooo001g0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool05`3oool0
0`0007l0oooo0?ooo`0G0?ooo`8000005`3oool2000001T0oooo00@000000?ooo`3oool00000503o
ool010000000oooo0?ooo`00000c0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool0
7P3oool00`000000oooo0?ooo`0S0?ooo`00M`3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo01L0oooo00<0001o0?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`040000003o
ool0oooo000001@0oooo00@000000?ooo`3oool00000<`3oool00`0007l0oooo0?ooo`0a0?ooo`03
0000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool007L0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0G0?ooo`030000O`3oool0oooo01L0oooo0P00000G0?ooo`800000
6@3oool010000000oooo0?ooo`00000D0?ooo`040000003oool0oooo000003<0oooo00<0001o0?oo
o`3oool0<@3oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001g0?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool05`3oool00`0007l0oooo0?ooo`0G0?ooo`80
00005`3oool2000001T0oooo00@000000?ooo`3oool00000503oool010000000oooo0?ooo`00000c
0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S
0?ooo`00M`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<0001o0?oo
o`3oool05`3oool2000001L0oooo0P00000I0?ooo`040000003oool0oooo000001@0oooo00@00000
0?ooo`3oool00000<`3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0oooo00<0
00000?ooo`3oool08`3oool007L0oooo0P0000050?ooo`030000003oool0oooo01L0oooo00<0001o
0?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`040000003oool0oooo000001@0oooo00@0
00000?ooo`3oool00000<`3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01d0oooo
0P00000U0?ooo`00M`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<0
001o0?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`040000003oool0oooo000001@0oooo
00@000000?ooo`3oool00000<`3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0
oooo00<000000?ooo`3oool08`3oool007L0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`0G0?ooo`030000O`3oool0oooo01L0oooo0P00000G0?ooo`8000006@3oool010000000oooo
0?ooo`00000D0?ooo`040000003oool0oooo000003<0oooo00<0001o0?ooo`3oool0<@3oool00`00
07l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001g0?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool05`3oool00`0007l0oooo0?ooo`0G0?ooo`8000005`3oool2000001T0
oooo00@000000?ooo`3oool00000503oool010000000oooo0?ooo`00000c0?ooo`030000O`3oool0
oooo0340oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00M`3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<0001o0?ooo`3oool05`3oool20000
01L0oooo0P00000I0?ooo`040000003oool0oooo000001@0oooo00@000000?ooo`3oool00000<`3o
ool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3o
ool005T0oooo00<0oomd003o0000o`001@00o`000`00oi`0gooo0?ooo`030?ooo`050?oo_`1do`00
0?l0003o0000okl02`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<0
001o0?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`040000003oool0oooo000001@0oooo
00@000000?ooo`3oool00000<`3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0
oooo00<000000?ooo`3oool08`3oool005/0oooo00H0oooO09co0000o`000?l0003oW03Oool50?oo
o`080?oo_`1do`000?nL0=ooo`3ooi`0B?l0003oM02oool90?ooo`8000001@3oool00`000000oooo
0?ooo`0G0?ooo`030000O`3oool0oooo01L0oooo0P00000G0?ooo`8000006@3oool010000000oooo
0?ooo`00000D0?ooo`040000003oool0oooo000003<0oooo00<0001o0?ooo`3oool0<@3oool00`00
07l0oooo0?ooo`0M0?ooo`8000009@3oool005`0oooo00D0oonL04So0000o`000?nL0=ooo`050?oo
o`030?ooM000o`000?no0080oooo00@0_ol0003o0000og@0_ooo203oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo01L0oooo00<0001o0?ooo`3oool05`3oool2000001L0oooo0P00000I
0?ooo`040000003oool0oooo000001@0oooo00@000000?ooo`3oool00000<`3oool00`0007l0oooo
0?ooo`0a0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool005`0oooo00D0
oooO09co0000o`000?m809coo`050?ooo`060;oo0000o`000?nL0=ooo`3oool0oomd0P00o`000`00
okl0oooo0?ooo`060?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool05`3oool00`00
07l0oooo0?ooo`0G0?ooo`8000005`3oool2000001T0oooo00@000000?ooo`3oool00000503oool0
10000000oooo0?ooo`00000c0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3o
ool00`000000oooo0?ooo`0S0?ooo`00G@3oool0103OodP00?l0003o0018oml50?ooo`0:0;oo0000
o`000?md0;ooo`3oool0oono07Co0000o`000?m809coo`L0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`0G0?ooo`030000O`3oool0oooo01L0oooo0P00000G0?ooo`8000006@3oool0
10000000oooo0?ooo`00000D0?ooo`040000003oool0oooo000003<0oooo00<0001o0?ooo`3oool0
<@3oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo00180?ooo`030=ooB000
o`000?l000X00?l000<00?nL0=ooo`3oool01@3oool01@3oog@00?l0003o0000oi`0gooo00@0oooo
00@0_ol0003o0000o`00M?oo0P3oool0102oo`000?l0003o0000okl70?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool05`3oool00`0007l0oooo0?ooo`0G0?ooo`8000005`3oool20000
01T0oooo00@000000?ooo`3oool00000503oool010000000oooo0?ooo`00000c0?ooo`030000O`3o
ool0oooo0340oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00B03oool0
0`3OodP00?l0003o000:003o0003003oW03Oool0oooo00D0oooo00D0oono07Co0000o`000?m809co
o`040?ooo`040=ooB000o`000?l004Sog`80oooo00D0oomd003o0000o`000?md0;ooo`060?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool05`3oool00`0007l0oooo0?ooo`0G0?ooo`80
00005`3oool2000001T0oooo00@000000?ooo`3oool00000503oool010000000oooo0?ooo`00000c
0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S
0?ooo`00GP3oool0102oo`000?l0003o0018oml40?ooo`080?ooM000o`000?l0003oW03Oool0oooo
0?oo_`1do`02003o000307Coo`3oool0oooo00@0oooo0`0000040000O`030000000007l0001o0340
001o0P00000G0000O`8000006@0007l010000000001o0000O`00000D0000O`040000000007l0001o
000008H0001o0`00000U0?ooo`00GP3oool0103oog@00?l0003o0000okl40?ooo`050?oo_`1do`00
0?l0003oB02Lool00P3oool0102oo`000?l0003o0000okl60?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool05`3oool00`0007l0oooo0?ooo`0G0?ooo`8000005`3oool2000001T0oooo
00@000000?ooo`3oool00000503oool010000000oooo0?ooo`00000c0?ooo`030000O`3oool0oooo
0340oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00GP3oool01@3ookl0
M?l0003o0000og@0_ooo00<0oooo00D0oooO09co0000o`000?l004Sog`020?ooo`050?ooM000o`00
0?l0003oW03Oool01@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<0
001o0?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`040000003oool0oooo000001@0oooo
00@000000?ooo`3oool00000<`3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0
oooo00<000000?ooo`3oool08`3oool005l0oooo00@0_ol0003o0000o`00M?oo103oool02`3oog@0
0?l0003o0000oi`0gooo0?ooo`3ookl0M?l0003o0000oi`0gooo00D0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo0P00000G0?ooo`800000
6@3oool01@000000oooo0?ooo`3oool0000001<0oooo00@000000?ooo`3oool00000<`3oool00`00
07l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool005l0
oooo00@0oomd003o0000o`000?no103oool01@3ooml0W?l0003o0000odP0W?oo0080oooo00@0_ol0
003o0000og@0_ooo1@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`050000003oool0oooo0?ooo`000000
4P3oool01@000000oooo0?ooo`3oool0000003<0oooo00<0001o0?ooo`3oool0<@3oool00`0007l0
oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001O0?ooo`050?oo_`1do`000?l0003oM02o
ool0103oool0103oog@00?l0003o0018oml20?ooo`040=ooB000o`000?nL0=ooo`D0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo0P00000G
0?ooo`8000006@3oool01@000000oooo0?ooo`3oool000000180oooo00D000000?ooo`3oool0oooo
0000000c0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3oool00`000000oooo
0?ooo`0S0?ooo`00G`3oool01@3ookl0M?l0003o0000o`00M?oo00D0oooo00P0_ol0003o0000og@0
_ooo0?ooo`3oog@00?l0003o_`H0oooo0P0000050?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`050000003oool0oooo0?ooo`0000004P3o
ool01@000000oooo0?ooo`3oool0000003<0oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo
0?ooo`0M0?ooo`8000009@3oool005d0oooo00<0oomd003o0000o`000`00o`000`00okl0oooo0?oo
o`040?ooo`070;oo0000o`00B?oO0?ooo`3OodP00?m809coo`060?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`8000005`3oool2000001T0
oooo00D000000?ooo`3oool0oooo0000000B0?ooo`050000003oool0oooo0?ooo`000000<`3oool0
0`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool0
0600oooo00D0oomd003o0000o`000?md0;ooo`060?ooo`050=ooB000o`000?l0003oB02Lool01`3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3o
ool2000001L0oooo0P00000I0?ooo`050000003oool0oooo0?ooo`0000004P3oool01@000000oooo
0?ooo`3oool0000003<0oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0N0?ooo`03
0000003oool0oooo02<0oooo001g0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
5`3oool00`000000oooo0?ooo`0G0?ooo`8000005`3oool2000001T0oooo00D000000?ooo`3oool0
oooo0000000B0?ooo`050000003oool0oooo0?ooo`000000<`3oool00`0007l0oooo0?ooo`0a0?oo
o`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool007L0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo0P00000G0?ooo`80
00006@3oool01@000000oooo0?ooo`3oool000000180oooo00D000000?ooo`3oool0oooo0000000c
0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S
0?ooo`00M`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool05`3oool2000001L0oooo0P00000I0?ooo`050000003oool0oooo0?ooo`0000004P3oool0
1@000000oooo0?ooo`3oool0000003<0oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?oo
o`0N0?ooo`030000003oool0oooo02<0oooo001g0?ooo`8000001@3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo01L0oooo0P00000G0?ooo`8000006@3oool01@000000oooo0?ooo`3o
ool000000180oooo00D000000?ooo`3oool0oooo0000000c0?ooo`030000O`3oool0oooo0340oooo
00<0001o0?ooo`3oool07@3oool2000002D0oooo001g0?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`8000005`3oool2000001T0oooo00D0
00000?ooo`3oool0oooo0000000B0?ooo`050000003oool0oooo0?ooo`000000<`3oool00`0007l0
oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool007L0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo
0P00000G0?ooo`8000006@3oool01@000000oooo0?ooo`3oool000000180oooo00D000000?ooo`3o
ool0oooo0000000c0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3oool00`00
0000oooo0?ooo`0S0?ooo`00M`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01L0
oooo00<000000?ooo`3oool05`3oool2000001L0oooo0P00000I0?ooo`050000003oool0oooo0?oo
o`0000004P3oool01@000000oooo0?ooo`3oool0000003<0oooo00<0001o0?ooo`3oool0<@3oool0
0`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001g0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`8000005`3oool20000
01T0oooo00D000000?ooo`3oool0oooo0000000B0?ooo`050000003oool0oooo0?ooo`000000<`3o
ool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3o
ool007L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
oooo01L0oooo0P00000G0?ooo`8000006@3oool01@000000oooo0?ooo`3oool000000140oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo0340oooo00<0
001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00M`3oool2000000D0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`8000005P3oool00`000000oooo0000000I
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0M
0?ooo`8000009@3oool007L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0G0?oo
o`030000003oool0oooo01L0oooo0P00000F0?ooo`030000003oool0000001T0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0<@3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0oooo00<000000?oo
o`3oool08`3oool007L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0G0?ooo`03
0000003oool0oooo01L0oooo0P00000F0?ooo`030000003oool0000001T0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0<@3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3o
ool08`3oool007L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0G0?ooo`030000
003oool0oooo01L0oooo0P00000F0?ooo`030000003oool0000001T0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
<@3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool0
8`3oool007L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo01L0oooo0P00000F0?ooo`030000003oool0000001T0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0<@3o
ool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3o
ool006H0oooo00<0gom8003o0000o`000P00o`000`00okl0oooo0?ooo`090?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`8000005P3oool0
0`000000oooo0000000I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<@3oool0
0`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001U0?ooo`090?ooW018o`000?l0
003o001dool0oooo0?ooM000oi`0gooo00T0oooo0P0000050?ooo`030000003oool0oooo01L0oooo
00<000000?ooo`3oool05`3oool2000001H0oooo00<000000?ooo`0000006@3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0a0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07@3oool2000002D0oooo
001U0?ooo`040?oog`2Lo`000?l007Coo`<0oooo00<0_ol0003oW03Oool0203oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool2000001H0oooo
00<000000?ooo`0000006@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo0340oooo
00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00K03oool00`3oog@00?l004So
g`080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`0G0?ooo`8000005P3oool00`000000oooo0000000I0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0340oooo00<0
001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001/
0?ooo`040?ooM000o`000?md0;ooo`L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0oooo01L0oooo0P00000F0?ooo`030000003oool0000001T0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0oooo00<0
00000?ooo`3oool08`3oool005<0oooo00<0gom8003o0000o`002P00o`000`00oi`0gooo0?ooo`09
0?ooo`040?ooM000o`000?m809coo`L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0oooo01L0oooo0P00000F0?ooo`030000003oool0000001T0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0oooo00<0
00000?ooo`3oool08`3oool005<0oooo00<0gom8003o0000o`002P00o`000`00oi`0gooo0?ooo`08
0?ooo`050?oog`2Lo`000?l0003o001dool01`3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool05`3oool2000001H0oooo00<000000?ooo`0000006@3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3o
ool00`000000oooo0?ooo`0S0?ooo`00J`3oool01@3OodP00?l0003o0000odP0W?oo00L0oooo0`00
00040000O`030000000007l0001o01L0001o00<000000000O`0007l05`0007l2000001H0001o00<0
00000000O`0000006@0007l00`000000001o0000O`040000O`030000000007l0001o00X0001o00<0
00000000O`0007l0100007l00`000000001o0000O`240000O`<000009@3oool006L0oooo00@0oooO
0=ooo`3ooml0W?l00`00o`000`00oi`0gooo0?ooo`060?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`8000005P3oool00`000000oooo0000
000I0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?oo
o`0N0?ooo`030000003oool0oooo02<0oooo001W0?ooo`030?oog`2Lo`000?l000@00?l000<0M?oo
0?ooo`3oool01P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool05`3oool2000001H0oooo00<000000?ooo`0000006@3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0a
0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S
0?ooo`00J03oool00`3OodP00?l0003o0002003o000307Coo`3oool0oooo00L0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo0P00000F0?oo
o`030000003oool0000001T0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`090?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0a0?oo
o`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool006P0oooo00@0oono07Co0000
og@0_ooo2`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool05`3oool2000001D0oooo00@000000?ooo`3oool000006@3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0a
0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S
0?ooo`00:@3oool00`3ooml0W?no0?ooo`0>0?ooo`030;ooM02oool0oooo02`0oooo00<0gom804So
g`3oool02`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool05`3oool2000001D0oooo00@000000?ooo`3oool000006@3oool00`000000oooo0?ooo`06
0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0a
0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S
0?ooo`00:@3oool00`3ooml0W?nL0=ooo`0=0?ooo`030?oo_`1dog@0_ooo02d0oooo00<0oono07Co
0000o`001000o`000`18oml0oooo0?ooo`040?ooo`8000001@3oool00`000000oooo0?ooo`0G0?oo
o`030000003oool0oooo01L0oooo0P00000E0?ooo`040000003oool0oooo000001T0oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01d0oooo0P00000U
0?ooo`00:@3oool00`3ooml0W?l0003o000>003o0003003oM02oool0oooo02d0oooo00<0_ol0003o
0000o`000`00o`000`00okl0oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`8000005@3oool010000000oooo0?ooo`00000I
0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0N
0?ooo`030000003oool0oooo02<0oooo000Y0?ooo`030?oog`2Lo`000?l000h00?l000<00?md0;oo
o`3oool0;@3oool00`3ooi`0B?l0003o0003003o0003003oM02oool0oooo00@0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo0P00000E0?oo
o`040000003oool0oooo000001T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0a
0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool002T0oooo00<0oooO09co
0000o`003P00o`000`00og@0_ooo0?ooo`0j0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`8000005@3oool010000000oooo0?ooo`00000I
0?ooo`030000003oool0oooo00T0oooo1000000;0?ooo`030000003oool0oooo0340oooo00<0001o
0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo000Y0?oo
o`030?oog`2Lo`00B?oO00d0oooo00<0oono07CoM02oool0>`3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000D0?oo
o`040000003oool0oooo000001T0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0a
0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S
0?ooo`00:@3oool00`3ooml0W?l004Sog`0>0?ooo`030;ooM02oool0oooo03X0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`0000004`3oool01@000000oooo0?ooo`3oool0000001T0oooo00<000000?ooo`3oool0603oool0
0`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3oool0
0`000000oooo0?ooo`0S0?ooo`00:@3oool00`3ooml0W?l004Sog`1;0?ooo`8000001@3oool00`00
0000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`0000004`3oool01@00
0000oooo0?ooo`3oool0000001T0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0a
0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07@3oool2000002D0oooo000Y0?oo
o`030?oog`2Lo`00B?oO04/0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0G0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`0000004`3oool01@000000oooo0?ooo`3oool0
000001T0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0a0?ooo`030000O`3oool0
oooo0340oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00:@3oool00`3o
oml0W?l004Sog`1;0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool05`3oool00`00
0000oooo0?ooo`0G0?ooo`030000003oool0000001<0oooo00D000000?ooo`3oool0oooo0000000I
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0a
0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool002T0oooo00<0oooO09co
0018oml0B`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool05`3oool00`000000oooo0000000C0?ooo`050000003oool0oooo0?ooo`0000006@3oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<@3oool0
0`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo000Y0?ooo`040?oog`2Lo`000?l0
07CoodX0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0F0?ooo`8000006@3oool0
0`000000oooo0000000B0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<@3oool0
0`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo000Y0?ooo`050?oog`2Lo`000?l0
003o0000okl0B@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01H0oooo0P00000I
0?ooo`030000003oool000000180oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0a
0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool007L0oooo0P0000050?oo
o`030000003oool0oooo01H0oooo0P00000I0?ooo`040000003oool0oooo00000140oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01P0oooo00<00000
0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01d0oooo0P00000U
0?ooo`00M`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01H0oooo0P00000I0?oo
o`040000003oool0oooo00000140oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0a
0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool007L0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0F0?ooo`8000006@3oool010000000oooo0?ooo`00000@
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0N
0?ooo`030000003oool0oooo02<0oooo001g0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool05P3oool2000001T0oooo00<000000?ooo`3oool00P00000>0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo
0340oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo
02<0oooo001g0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool05P3oool2000001T0
oooo00D000000?ooo`3oool0oooo0000000=0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool05`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo0340oooo00<0001o0?oo
o`3oool0<@3oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo001W0?ooo`05
0?oo_`1do`000?l0003o0000okl02`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
01H0oooo0P00000I0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool02@3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0603oool00`00
0000oooo0?ooo`0a0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3oool00`00
0000oooo0?ooo`0S0?ooo`00IP3oool0203ookl0M?l0003oW03Oool0oonL04So0000og@0_ooo2@3o
ool2000000D0oooo00<000000?ooo`3oool05P3oool2000001T0oooo00<000000?ooo`3oool00P3o
ool3000000H0oooo0`0000090?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0603o
ool00`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07@3o
ool2000002D0oooo001V0?ooo`030?ooM000o`000?no0080oooo00@0_ol0003o0000og@0_ooo203o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01H0oooo0P00000I0?ooo`030000003o
ool0oooo00@0oooo2000000;0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0603o
ool00`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3o
ool00`000000oooo0?ooo`0S0?ooo`00:P3oool00`3oog@00?l0003o000A003o0003003oW03Oool0
oooo02D0oooo00H0_ol0003o0000oi`0gooo0?ooo`3oog@2003o0003003o_`3oool0oooo00H0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0F0?ooo`8000006@3oool00`000000oooo
0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0603oool00`000000oooo
0?ooo`0a0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3oool00`000000oooo
0?ooo`0S0?ooo`00:P3oool00`3oog@00?l0003o000A003o0003003oW03Oool0oooo02D0oooo00X0
_ol0003o0000og@0_ooo0?ooo`3ookl0M?l0003o0000odP0W?oo1`3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`0000006@3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0a
0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S
0?ooo`00:P3oool00`3oog@00?l0003o000A003o0003003oW03Oool0oooo02D0oooo00@0_ol0003o
0000o`00M?oo0P3oool0102oo`000?l0003o0000okl70?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool04`3oool2000000030?ooo`000000oooo01P0oooo00<000000?ooo`3oool05`3o
ool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0<@3o
ool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3o
ool002X0oooo00<0oomd04Sog`3oool0403oool00`3ookl0M?nL0=ooo`0V0?ooo`040=ooB000o`00
0?l004Sog`80oooo00D0oomd003o0000o`000?md0;ooo`060?ooo`030000003oool0oooo00D0oooo
0P00000B0?ooo`<000000P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo0340oooo
00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo
000Z0?ooo`030?ooM018oml0oooo0100oooo00<0oono07CoW03Oool09P3oool0203oog@00?l0003o
0000oi`0gooo0?ooo`3ookl0M?l00P00o`000`1dool0oooo0?ooo`040?ooool000005@00000U0?oo
o`00:P3oool00`3oog@0B?oO0?ooo`0@0?ooo`030?oo_`1doi`0gooo02H0oooo00D0oono07Co0000
o`000?m809coo`020?ooo`040;oo0000o`000?l0003o_`H0oooo00<000000?ooo`3oool07P3oool0
0`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<@3oool0
0`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo000h0?ooo`040?ooM000o`000?nL
0=ooobX0oooo00D0oooO09co0000o`000?l004Sog`020?ooo`050?ooM000o`000?l0003oW03Oool0
1@3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo0080oooo300000090?ooo`030000
003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0H0?ooo`030000
003oool0oooo0340oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0N0?ooo`030000
003oool0oooo02<0oooo000h0?ooo`040;oo0000o`000?md0;ooob/0oooo00/0oomd003o0000o`00
0?nL0=ooo`3oool0oono07Co0000o`000?nL0=ooo`050?ooo`030000003oool0oooo01h0oooo00<0
00000?ooo`3oool00P00000<0?ooo`<000001P3oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0a0?ooo`030000O`3o
ool0oooo0340oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00<03oool0
0`2oog@0_ooo0?ooo`050?ooo`030?ooW018o`000?no02`0oooo00D0oooO09co0000o`000?m809co
o`020?ooo`040;oo0000o`000?md0;ooo`D0oooo00<000000?ooo`3oool07P3oool00`000000oooo
0000000@0?ooo`<00000103oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo
00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo0340oooo
00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00;`3oool00`3ooml0W?m809co
o`060?ooo`030?ooM000oi`0gooo02d0oooo00@0oomd003o0000o`00B?oO0P3oool0103OodP00?l0
003oW03Oool50?ooo`030000003oool0oooo01h0oooo00<000000?ooo`0000004`3oool01@000000
oooo0?ooo`3oool0000001T0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0N0?oo
o`030000003oool0oooo02<0oooo000_0?ooo`050?oo_`1do`000?l0003o001dool00`3oool00`2o
o`000?no0?ooo`0_0?ooo`080;oo0000o`000?md0;ooo`3oool0oomd003o0000okl60?ooo`800000
7`3oool2000001@0oooo00D000000?ooo`3oool0oooo0000000I0?ooo`030000003oool0oooo01L0
oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?oooa/00000<P3oool00`0007l0
oooo0?ooo`0M0?ooo`8000009@3oool002l0oooo00<0oonL04So0000o`001@00o`000`1dool0oooo
0?ooo`0`0?ooo`070;oo0000o`00B?oO0?ooo`3OodP00?m809coo`060?ooo`030000003oool0oooo
01h0oooo0P00000E0?ooo`040000003oool0oooo000001T0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo0180oooo1P00000I0?ooo`030000O`3oool0
000000D00000;03oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo000`0?oo
o`030?ooM000o`000?l000H00?l000<0B?oO0?ooo`3oool0;`3oool01@3OodP00?l0003o0000odP0
W?oo00L0oooo00<000000?ooo`3oool07P3oool2000001D0oooo00@000000?ooo`3oool000006@3o
ool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool03`3o
ool3000001l0oooo00<0001o0?ooo`3oool01@3oool5000002L0oooo00<0001o0?ooo`3oool07P3o
ool00`000000oooo0?ooo`0S0?ooo`00<P3oool00`3ooml0W?l0003o0005003o0003003o_`3oool0
oooo03X0oooo00<000000?ooo`3oool07P3oool2000001D0oooo00@000000?ooo`3oool000006@3o
ool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`0P0?ooo`030000O`3oool0oooo00X0oooo1000000S0?ooo`030000O`3o
ool0oooo01h0oooo00<000000?ooo`3oool08`3oool00340oooo00<0oooO09co001dool00P3oool0
1P3ookl0M?l0003o0000o`000?md0;oooc/0oooo00<000000?ooo`3oool07P3oool2000001H0oooo
00<000000?ooo`0000006@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool02`3oool3000002<0oooo00<0001o0?ooo`3oool03P3oool300000200oooo
00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00<03oool00`3ookl0M?m809co
o`050?ooo`040?oog`2Lo`000?nL0=oooc/0oooo00<000000?ooo`3oool07P3oool2000001H0oooo
00<000000?ooo`0000006@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0T0?ooo`030000O`3oool0oooo0140oooo
0`00000M0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool002l0oooo00@0
oooO09co0000og@0_ooo1@3oool00`3oog@00?md0;ooo`0l0?ooo`030000003oool0oooo01h0oooo
0P00000F0?ooo`030000003oool0000001T0oooo00<000000?ooo`3oool05`3oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo00P0oooo0P00000W0?ooo`030000O`3oool0oooo01@0oooo
0P00000K0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool002l0oooo00@0
oonL04So0000o`00M?oo103oool00`3ooi`0B?nL0=ooo`0m0?ooo`8000007`3oool2000001H0oooo
00<000000?ooo`0000006@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo01H0oooo
0`00000H0?ooo`030000O`3oool0oooo01d0oooo0P00000U0?ooo`00;`3oool0103ookl0M?l0003o
B02Loom40?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool05@3oool00`000000oooo
0000000I0?ooo`030000003oool0oooo00L0oooo0`00000=0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo01T0oooo
0`00000E0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool007L0oooo00<0
00000?ooo`3oool07P3oool00`000000oooo0?ooo`0F0?ooo`8000006@3oool00`000000oooo0?oo
o`060?ooo`8000000P3oool2000000/0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo02P0oooo00<0001o0?ooo`3oool0703oool2000001<0oooo00<0
001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00M`3oool00`000000oooo0?ooo`0N
0?ooo`030000003oool0oooo01H0oooo0P00000I0?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool0103oool2000000T0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`06
0?ooo`030000003oool0oooo02P0oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0@
0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool002X0oooo00<0oomd04So
g`3oool0403oool00`3ookl0M?nL0=ooo`0g0?ooo`030000003oool0oooo01h0oooo00<000000?oo
o`3oool05P3oool2000001T0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo02T0oooo00<0001o0?ooo`3oool07`3oool200000100oooo00<0001o0?oo
o`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00:P3oool00`3oog@0B?oO0?ooo`0@0?ooo`03
0?oo_`1doi`0gooo03L0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0F0?ooo`80
00006@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool0:@3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo00d0oooo00<0001o0?ooo`3o
ool07P3oool00`000000oooo0?ooo`0S0?ooo`00:P3oool00`3oog@0B?oO0?ooo`0@0?ooo`030?oo
_`1doi`0gooo03L0oooo0P00000O0?ooo`030000003oool0oooo01H0oooo0P00000I0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo01P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0Z0?ooo`030000
O`3oool0oooo0280oooo00<000000?ooo`3oool0303oool00`0007l0oooo0?ooo`0M0?ooo`800000
9@3oool002X0oooo00<0oomd003o0000o`004@00o`000`00oi`0gooo0?ooo`0f0?ooo`030000003o
ool0oooo01h0oooo00<000000?ooo`3oool05P3oool2000001T0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0603o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo02X0oooo00<0001o0?ooo`3oool08`3o
ool00`000000oooo0?ooo`0;0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3o
ool002X0oooo00<0oomd003o0000o`004@00o`000`00oi`0gooo0?ooo`0f0?ooo`030000003oool0
oooo01h0oooo00<000000?ooo`3oool05P3oool2000001T0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0603oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo02/0oooo00<0001o0?ooo`3oool0903oool0
0`000000oooo0?ooo`0:0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool0
02X0oooo00<0oomd003o0000o`004@00o`000`00oi`0gooo0?ooo`0f0?ooo`030000003oool0oooo
01h0oooo00<000000?ooo`3oool05P3oool2000001T0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0;0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0603oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo02/0oooo00<0001o0?ooo`3oool09@3oool20000
00X0oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00IP3oool00`3OodP0
0?l0003o0002003o0003003o_`3oool0oooo00T0oooo00<000000?ooo`3oool07P3oool00`000000
oooo0?ooo`0F0?ooo`8000006@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00`0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0:`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo00L0
oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00I@3oool02@3ooi`0B?l0
003o0000o`00M?oo0?ooo`3oog@00?nL0=ooo`090?ooo`030000003oool0oooo01h0oooo00<00000
0?ooo`3oool05P3oool2000001T0oooo00D000000?ooo`3oool0oooo0000000?0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo02`0oooo00<0001o0?ooo`3oool0:03oool00`000000oooo0?ooo`060?ooo`030000O`3o
ool0oooo01h0oooo00<000000?ooo`3oool08`3oool006D0oooo00@0oooO09co0000o`00M?oo0`3o
ool00`2oo`000?nL0=ooo`080?ooo`8000007`3oool00`000000oooo0?ooo`0F0?ooo`8000006@3o
ool01@000000oooo0?ooo`3oool0000000l0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0;03oool00`0007l0oooo
0?ooo`0Y0?ooo`030000003oool0oooo00D0oooo00<0001o0?ooo`3oool07@3oool2000002D0oooo
001/0?ooo`030?ooM000o`00B?oO00P0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?oo
o`0F0?ooo`8000006@3oool01@000000oooo0?ooo`3oool0000000l0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
;03oool00`0007l0oooo0?ooo`0Z0?ooo`030000003oool0oooo00@0oooo00<0001o0?ooo`3oool0
7P3oool00`000000oooo0?ooo`0S0?ooo`00K03oool0103oog@00?l0003oM02oool70?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool05P3oool2000001T0oooo00D000000?ooo`3oool0
oooo0000000?0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0603oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo02`0oooo00<0001o0?ooo`3oool0:`3oool00`000000
oooo0?ooo`030?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool08`3oool006`0oooo
00@0oomd003o0000odP0W?oo1`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01H0
oooo0P00000I0?ooo`050000003oool0oooo0?ooo`000000403oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0/0?oo
o`030000O`3oool0oooo02/0oooo00<000000?ooo`3oool00`3oool00`0007l0oooo0?ooo`0N0?oo
o`030000003oool0oooo02<0oooo001[0?ooo`050?oog`2Lo`000?l0003o001dool01`3oool00`00
0000oooo0?ooo`0N0?ooo`030000003oool0oooo01H0oooo0P00000I0?ooo`040000003oool0oooo
00000140oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0;03oool00`0007l0oooo0?ooo`0/0?ooo`030000003oool0
oooo0080oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0S0?ooo`00J`3oool01@3O
odP00?l0003o0000odP0W?oo00L0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0F
0?ooo`8000006@3oool010000000oooo0?ooo`00000A0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo02`0oooo00<0
001o0?ooo`3oool0;@3oool01@000000oooo0?ooo`3oool0001o0200oooo00<000000?ooo`3oool0
8`3oool006L0oooo00@0oooO0=ooo`3ooml0W?l00`00o`000`00oi`0gooo0?ooo`060?ooo`<00000
7P0007l00`000000001o0000O`0F0000O`8000006@0007l010000000001o0000O`00000A0000O`03
0000000007l0001o0080001o00<000000000O`0007l0600007l01@000000001o0000O`0007l00000
0600001o00<000000000O`0007l07`0007l3000002D0oooo001W0?ooo`030?oog`2Lo`000?l000@0
0?l000<0M?oo0?ooo`3oool01P3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01H0
oooo0P00000I0?ooo`040000003oool0oooo00000180oooo00D000000?ooo`3oool0oooo0000000J
0?ooo`050000003oool0oooo0?ooo`000000;`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003o
ool0001o0200oooo00<000000?ooo`3oool08`3oool006P0oooo00<0gom8003o0000o`000P00o`00
0`1dool0oooo0?ooo`070?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool05P3oool2
000001T0oooo00@000000?ooo`3oool000004P3oool01@000000oooo0?ooo`3oool0000001X0oooo
00D000000?ooo`3oool0oooo0000000_0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`00
07l0803oool00`000000oooo0?ooo`0S0?ooo`00J03oool0103ookl0M?l0003oM02oool;0?ooo`03
0000003oool0oooo01h0oooo00<000000?ooo`3oool05P3oool2000001T0oooo00@000000?ooo`3o
ool000004P3oool01@000000oooo0?ooo`3oool0000001X0oooo00D000000?ooo`3oool0oooo0000
000_0?ooo`030000O`3oool0oooo0300oooo00<000000000O`3oool07`3oool00`000000oooo0?oo
o`0S0?ooo`00J@3oool00`3OodP0B?oO0?ooo`0;0?ooo`030000003oool0oooo01h0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`040000003oool0oooo00000180oooo00D0
00000?ooo`3oool0oooo0000000J0?ooo`050000003oool0oooo0?ooo`000000;`3oool00`0007l0
oooo0?ooo`0`0?ooo`030000000007l0oooo01l0oooo00<000000?ooo`3oool08`3oool006T0oooo
00<0oono07Co0000o`001000o`000`18oml0oooo0?ooo`040?ooo`030000003oool0oooo01h0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`040000003oool0oooo00000180
oooo00D000000?ooo`3oool0oooo0000000J0?ooo`050000003oool0oooo0?ooo`000000;`3oool0
0`0007l0oooo0?ooo`0a0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool08`3oool0
06X0oooo00<0_ol0003o0000o`000`00o`000`00okl0oooo0?ooo`040?ooo`8000007`3oool00`00
0000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00@000000?ooo`3oool000004P3oool0
1@000000oooo0?ooo`3oool0000001X0oooo00D000000?ooo`3oool0oooo0000000_0?ooo`030000
O`3oool0oooo0340oooo00<0001o0000003oool07@3oool2000002D0oooo001Z0?ooo`030?ooW018
o`000?l000<00?l000<00?md0;ooo`3oool0103oool00`000000oooo0?ooo`0N0?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool05`3oool010000000oooo0?ooo`00000B0?ooo`050000
003oool0oooo0?ooo`0000006P3oool01@000000oooo0?ooo`3oool0000002l0oooo00<0001o0?oo
o`3oool0<@3oool00`0007l000000?ooo`0N0?ooo`030000003oool0oooo02<0oooo001g0?ooo`03
0000003oool0oooo01h0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`04
0000003oool0oooo00000180oooo00D000000?ooo`3oool0oooo0000000J0?ooo`050000003oool0
oooo0?ooo`000000;`3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0000001h0oooo00<0
00000?ooo`3oool08`3oool007L0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo01L0oooo00@000000?ooo`3oool000004P3oool01@000000oooo0?oo
o`3oool0000001X0oooo00@000000?ooo`3oool00000<03oool00`0007l0oooo0?ooo`0a0?ooo`03
0000O`3oool0000001h0oooo00<000000?ooo`3oool08`3oool007L0oooo00<000000?ooo`3oool0
7P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00@000000?ooo`3oool0
00004P3oool01@000000oooo0?ooo`3oool0000001X0oooo00@000000?ooo`3oool00000<03oool0
0`0007l0oooo0?ooo`0a0?ooo`040000O`3oool0oooo000001d0oooo00<000000?ooo`3oool08`3o
ool007L0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
oooo01L0oooo00<000000?ooo`000000503oool010000000oooo0?ooo`00000J0?ooo`040000003o
ool0oooo00000300oooo00<0001o0?ooo`3oool0<@3oool0100007l0oooo0?ooo`00000M0?ooo`03
0000003oool0oooo02<0oooo001g0?ooo`8000007`3oool00`000000oooo0?ooo`0G0?ooo`030000
003oool0oooo01L0oooo00<000000?ooo`000000503oool010000000oooo0?ooo`00000J0?ooo`04
0000003oool0oooo00000300oooo00<0001o0?ooo`3oool0<@3oool01@0007l0oooo0?ooo`3oool0
000001/0oooo0P00000U0?ooo`00M`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo
01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000D0?ooo`040000003oool0oooo
000001X0oooo00@000000?ooo`3oool00000<03oool00`0007l0oooo0?ooo`0a0?ooo`050000O`3o
ool0oooo0?ooo`000000703oool00`000000oooo0?ooo`0S0?ooo`00M`3oool00`000000oooo0?oo
o`0N0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0000
000D0?ooo`040000003oool0oooo000001X0oooo00@000000?ooo`3oool00000<03oool00`0007l0
oooo0?ooo`0a0?ooo`050000O`3oool0oooo0?ooo`000000703oool00`000000oooo0?ooo`0S0?oo
o`00M`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3o
ool05`3oool00`000000oooo0000000D0?ooo`040000003oool0oooo000001X0oooo00@000000?oo
o`3oool00000<03oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo0080oooo00<00000
0?ooo`3oool06@3oool00`000000oooo0?ooo`0S0?ooo`00M`3oool00`000000oooo0?ooo`0N0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000D0?oo
o`040000003oool0oooo000001X0oooo00@000000?ooo`3oool00000<03oool00`0007l0oooo0?oo
o`0a0?ooo`030000O`3oool0oooo0080oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`0S0?ooo`00M`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0000000D0?ooo`040000003oool0oooo000001X0oooo00@0
00000?ooo`3oool00000<03oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo0080oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0S0?ooo`00M`3oool2000001l0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0000001@0oooo00@0
00000?ooo`3oool000006P3oool010000000oooo0?ooo`00000`0?ooo`030000O`3oool0oooo0340
oooo00<0001o0?ooo`3oool00`3oool00`000000oooo0?ooo`0G0?ooo`8000009@3oool007L0oooo
00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo
00<000000?ooo`000000503oool010000000oooo0?ooo`00000J0?ooo`040000003oool0oooo0000
0300oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`030?ooo`030000003oool0oooo
01P0oooo00<000000?ooo`3oool08`3oool007L0oooo00<000000?ooo`3oool07P3oool00`000000
oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`000000503oool010000000
oooo0?ooo`00000J0?ooo`040000003oool0oooo00000300oooo00<0001o0?ooo`3oool0<@3oool0
0`0007l0oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool08`3oool0
07L0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo
01L0oooo00<000000?ooo`000000503oool010000000oooo0?ooo`00000J0?ooo`030000003oool0
00000340oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`040?ooo`030000003oool0
oooo01L0oooo00<000000?ooo`3oool08`3oool007L0oooo00<000000?ooo`3oool07P3oool00`00
0000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`000000503oool01000
0000oooo0?ooo`00000J0?ooo`030000003oool000000340oooo00<0001o0?ooo`3oool0<@3oool0
0`0007l0oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool08`3oool0
05T0oooo00<0oomd003o0000o`001@00o`000`00oi`0gooo0?ooo`030?ooo`050?oo_`1do`000?l0
003o0000okl02`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0000000E0?ooo`030000003oool0000001X0oooo00<00000
0?ooo`000000<@3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo00D0oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`0S0?ooo`00F`3oool01P3ooml0W?l0003o0000o`00
0?nL0=ooo`D0oooo00P0oono07Co0000oi`0gooo0?ooW018o`000?md0;ooo`T0oooo0P00000O0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000E0?oo
o`030000003oool0000001X0oooo00<000000?ooo`000000<@3oool00`0007l0oooo0?ooo`0a0?oo
o`030000O`3oool0oooo00D0oooo00<000000?ooo`3oool05@3oool2000002D0oooo001L0?ooo`05
0?ooW018o`000?l0003oW03Oool01@3oool00`3oog@00?l0003o_`020?ooo`040;oo0000o`000?md
0;ooo`P0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
oooo01L0oooo00<000000?ooo`0000005@3oool00`000000oooo0000000J0?ooo`030000003oool0
00000340oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`050?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool08`3oool005`0oooo00D0oooO09co0000o`000?m809coo`05
0?ooo`060;oo0000o`000?nL0=ooo`3oool0oomd0P00o`000`00okl0oooo0?ooo`060?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`030000
003oool0000001D0oooo00<000000?ooo`0000006P3oool00`000000oooo0000000a0?ooo`030000
O`3oool0oooo0340oooo00<0001o0?ooo`3oool01P3oool00`000000oooo0?ooo`0E0?ooo`030000
003oool0oooo02<0oooo001M0?ooo`040=ooB000o`000?l004Sog`D0oooo00X0_ol0003o0000og@0
_ooo0?ooo`3ookl0M?l0003o0000odP0W?oo1`3oool00`000000oooo0?ooo`0N0?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000E0?ooo`030000003o
ool0000001X0oooo00<000000?ooo`000000<@3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3o
ool0oooo00H0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0S0?ooo`00G@3oool0
1@3oog@00?l0003o0000oi`0gooo00@0oooo00@0_ol0003o0000o`00M?oo0P3oool0102oo`000?l0
003o0000okl70?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool05`3oool00`000000
oooo0?ooo`0G0?ooo`030000003oool0000001D0oooo00<000000?ooo`0000006P3oool00`000000
oooo0000000a0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool01P3oool00`000000
oooo0?ooo`0E0?ooo`030000003oool0oooo02<0oooo001M0?ooo`050?oo_`1do`000?l0003oB02L
ool0103oool0103OodP00?l0003o0018oml20?ooo`050?ooM000o`000?l0003oM02oool01P3oool0
0`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0000000E0?ooo`030000003oool0000001X0oooo00<000000?ooo`000000<@3oool0
0`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo00L0oooo00<000000?ooo`3oool0503oool0
0`000000oooo0?ooo`0S0?ooo`00GP3oool0102oo`000?l0003o0018oml40?ooo`080?ooM000o`00
0?l0003oW03Oool0oooo0?oo_`1do`02003o000307Coo`3oool0oooo00@0oooo0`00000N0000O`03
0000000007l0001o01L0001o00<000000000O`0007l05`0007l00`000000001o0000000E0000O`03
0000000007l0000001X0001o00<000000000O`000000K`0007l00`000000001o0000O`0B0000O`<0
00009@3oool005h0oooo00@0oomd003o0000o`000?no103oool01@3ookl0M?l0003o0000odP0W?oo
0080oooo00@0_ol0003o0000o`000?no1P3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0
oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000E0?ooo`030000003oool0
000001X0oooo00<000000?ooo`000000<@3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0
oooo00L0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0S0?ooo`00GP3oool01@3o
okl0M?l0003o0000og@0_ooo00<0oooo00D0oooO09co0000o`000?l004Sog`020?ooo`050?ooM000
o`000?l0003oW03Oool01@3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01L0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0000000E0?ooo`030000003oool0000001X0oooo
00<000000?ooo`000000<@3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo00P0oooo
00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0S0?ooo`00G`3oool0102oo`000?l0003o
001dool40?ooo`0;0?ooM000o`000?l0003oW03Oool0oooo0?oo_`1do`000?l0003oW03Oool01@3o
ool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3o
ool00`000000oooo0000000E0?ooo`030000003oool0000001X0oooo00<000000?ooo`000000<@3o
ool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo00P0oooo00<000000?ooo`3oool04`3o
ool00`000000oooo0?ooo`0S0?ooo`00G`3oool0103oog@00?l0003o0000okl40?ooo`050?oog`2L
o`000?l0003oB02Lool00P3oool0102oo`000?l0003oM02oool50?ooo`030000003oool0oooo01h0
oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0000001D0
oooo00<000000?ooo`0000006P3oool00`000000oooo0000000a0?ooo`030000O`3oool0oooo0340
oooo00<0001o0?ooo`3oool0203oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo02<0
oooo001O0?ooo`050?oo_`1do`000?l0003oM02oool0103oool0103oog@00?l0003o0018oml20?oo
o`040=ooB000o`000?nL0=ooo`D0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`0000005@3oool00`000000oooo0000000J
0?ooo`030000003oool000000340oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`09
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool08`3oool005l0oooo00D0oono07Co
0000o`000?l007Coo`050?ooo`080;oo0000o`000?md0;ooo`3oool0oomd003o0000okl60?ooo`80
00007`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`00
00005@3oool00`000000oooo0000000J0?ooo`030000003oool000000340oooo00<0001o0?ooo`3o
ool0<@3oool00`0007l0oooo0?ooo`090?ooo`030000003oool0oooo0140oooo0P00000U0?ooo`00
G@3oool00`3oog@00?l0003o0003003o0003003o_`3oool0oooo00@0oooo00L0_ol0003o0018oml0
oooo0=ooB000odP0W?oo00H0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0G0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`0000005@3oool00`000000oooo0000000J0?oo
o`030000003oool000000340oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`090?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool08`3oool00600oooo00D0oomd003o0000
o`000?md0;ooo`060?ooo`050=ooB000o`000?l0003oB02Lool01`3oool00`000000oooo0?ooo`0N
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000E
0?ooo`030000003oool0000001X0oooo00<000000?ooo`000000<@3oool00`0007l0oooo0?ooo`0a
0?ooo`030000O`3oool0oooo00T0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0S
0?ooo`00M`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool05`3oool00`000000oooo0000000E0?ooo`030000003oool0000001X0oooo00<000000?oo
o`000000<@3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo00X0oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`0S0?ooo`00M`3oool00`000000oooo0?ooo`0N0?ooo`03
0000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0000000E0?ooo`03
0000003oool0000001X0oooo00<000000?ooo`000000<@3oool00`0007l0oooo0?ooo`0a0?ooo`03
0000O`3oool0oooo00X0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0S0?ooo`00
M`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
5`3oool00`000000oooo0000000E0?ooo`030000003oool0000001X0oooo00<000000?ooo`000000
<@3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo00X0oooo00<000000?ooo`3oool0
4@3oool00`000000oooo0?ooo`0S0?ooo`00M`3oool2000001l0oooo00<000000?ooo`3oool05`3o
ool00`000000oooo0?ooo`0G0?ooo`030000003oool0000001D0oooo00<000000?ooo`0000006P3o
ool00`000000oooo0000000a0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool02P3o
ool00`000000oooo0?ooo`0@0?ooo`8000009@3oool007L0oooo00<000000?ooo`3oool07P3oool0
0`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`0000005@3oool0
0`000000oooo0000000J0?ooo`030000003oool000000340oooo00<0001o0?ooo`3oool0<@3oool0
0`0007l0oooo0?ooo`0;0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool08`3oool0
07L0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo
01L0oooo00<000000?ooo`0000005@3oool00`000000oooo0000000J0?ooo`030000003oool00000
0340oooo00<0001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0;0?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool08`3oool007L0oooo00<000000?ooo`3oool07P3oool00`000000
oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`0000005@3oool00`000000
oooo0000000J0?ooo`030000003oool000000340oooo00<0001o0?ooo`3oool0<@3oool00`0007l0
oooo0?ooo`0;0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool08`3oool007L0oooo
00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo
0P00000F0?ooo`030000003oool0000001X0oooo00<000000?ooo`000000<@3oool00`0007l0oooo
0?ooo`0a0?ooo`030000O`3oool0oooo00/0oooo00<000000?ooo`3oool0403oool00`000000oooo
0?ooo`0S0?ooo`00M`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool05`3oool2000001H0oooo00<000000?ooo`0000006P3oool200000380oooo00<0
001o0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool08`3oool007L0oooo0P00000O0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool05`3oool2000001H0oooo00<000000?ooo`0000006P3oool200000380oooo00<0001o
0?ooo`3oool0<@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo00h0oooo0P00000U
0?ooo`00M`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool05`3oool2000001H0oooo00<000000?ooo`0000006P3oool200000380oooo00<0001o0?oo
o`3oool0<@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo00l0oooo00<000000?oo
o`3oool08`3oool007L0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0G0?ooo`03
0000003oool0oooo01L0oooo0P00000G0?ooo`8000006P3oool200000380oooo00<0001o0?ooo`3o
ool0<@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool08`3oool007L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0:0?ooo`030000
003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000
003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool2000000/0oooo00<000000?ooo`3oool0
2@3oool2000000d0oooo00<000000?ooo`3oool02P3oool2000000/0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`0:0?ooo`040000003oool0oooo000000T0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`0S0?ooo`00M`3ooooo000001D000009@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00:h0oooo00<0
gom8003o0000o`001000o`000`00oi`0gooo0?ooo`030?ooo`090;oo0000o`000?l007Coo`3ooml0
W?l0003o0000odP0W?oo0080oooo00X0_ol0003o0000o`00M?oo0?ooo`3ookl0M?l0003o0000og@0
_ooo0P3oool0103OodP00?l0003o0018oml20?ooo`0:0;oo0000o`000?l007Coo`3oool0oono07Co
0000o`000?md0;ooo`80oooo00@0gom8003o0000o`00B?oO0`3oool02@2oo`000?l0003o001dool0
oooO09co0000o`000?m809coo`080?ooo`030?oog`2Lo`000?l000D00?l000<00?md0;ooo`3oool0
1`3oool02P3OodP00?l0003oB02Lool0oooo0?oog`2Lo`000?l0003oB02Lool20?ooo`040;oo0000
o`000?l004Sog`<0oooo00@0_ol0003o0000odP0W?oo103oool01P2oo`000?l0003o0000o`000?nL
0=ooo`<0oooo00D0oonL04So0000o`000?l007Coo`020?ooo`050?ooW018o`000?l0003o001dool0
0`3oool00`3ookl0M?l0003o0002003o000307Coo`3oool0oooo00<0oooo00@0_ol0003o0000o`00
B?oO0`3oool0102oo`000?l0003oB02LoomI0?ooo`00[03oool01@3ooi`0B?l0003o0000odP0W?oo
0080oooo00D0gom8003o0000o`000?md0;ooo`020?ooo`030?ooW018o`000?l000800?l000<0M?oO
09co0000o`000P00o`000`18oml0oooo0?ooM002003o0007003o_`3oool0oooo0;oo0000o`000?m8
09coo`020?ooo`070;oo0000o`000?l0003o0000okl0oooo0?ooM002003o0007003o_`3oool0oooo
0;oo0000o`000?m809coo`020?ooo`080;oo0000o`000?l0003o0000okl0oooo0?ooW018o`03003o
000307Cog`2Lo`000?l000800?l000<0B?oO0?ooo`3oool01`3oool00`3ooml0W?l0003o0002003o
000307Coo`3oool0oooo00L0oooo00<0oooO09co0000o`000P00o`000`18oml0oono07Co0003003o
000307Coo`3oool0oomd00800?l000D00?nL0=ooo`3oool0oono07Co0003003o000404Sog`3oool0
oooO09co00H00?l000@0B?oO0?ooo`3oool0oomd0P00o`001@00okl0_ol004Sog`3oool0gom80080
0?l000d00?nL09co0018oml0oooo0?oo_`1do`000?l0003o_`3oool0oomd003o0000og@0_ooo00<0
oooo00P0oomd003o0000o`000?nL0=ooo`3oool0oono07Co00<00?l000<0B?oO0?ooo`3oool0EP3o
ool00:/0oooo00D0oonL04So0000o`000?md0;ooo`030?ooo`050?ooM000o`000?l0003oB02Lool0
0P3oool04`3oog@00?l0003o0000okl0oooO09coB01dodP00?l0003oB02Lool0gomd09coo`3ookl0
M?l0003o0000og@0_ooo0?ooo`3OodP00P00o`001018oml0oooo0?ooo`3OodP2003o000:04Sog`3o
og@0M?oo0?oo_`1do`000?l0003oM02oool0oooo0=ooB0800?l000@0B?oO0?ooo`3oool0gom80P00
o`001018oml0oomd07Coo`3oog@2003o0009003o_`3ooml0W?m807CoB000o`000?m809coo`3Oog@0
W?oo00T0oooo00D0oonL04So0000o`000?m809coo`0:0?ooo`0A0;oo0000o`000?l007Coo`3OodP0
B?no09co0000o`000?m809coo`2oog@0_ooo0?oo_`1do`000?l0003oM02oool00P3oool0202oo`00
0?l0003o001dool0gomd09coo`3ooi`0B?l00P00o`005P1dool0oooo0?ooo`3oool0_ol0003o_`3o
ool0oonL04So0000o`000?nL0=ooo`3Oog@0W?oo0?ooW018o`000?l0003oW03Oool0gom807Coo`3o
og@2003o000704Sog`3oool0oooo0;oo0000o`000?m809coo`020?ooo`050?oo_`1do`000?l0003o
M02oool00P3oool01P2oo`000?l0003o001dool0gomd09cooeL0oooo002Z0?ooo`050?oog`2Lo`00
0?l0003o0018oml0103oool01@3ookl0M?l0003o0000o`00B?oO0080oooo00X0oomd003o0000o`00
B?oO0?ooo`3ookl0M?m804So0000o`00B?oO0`3oool0403OodP00?l0003o001dool0oooo0?ooW018
o`000?l0003oW03Oool0oooo0?ooW018o`000?l0003oW03Oool20?ooo`0B0=ooB000o`000?l007Co
o`3oool0oonL04So0000o`000?nL0=ooo`3oool0oonL04So0000o`000?nL0=ooo`3oool0oomd0P00
o`001`18oml0oooo0?oo_`1dodP0B?l0003o0018oml02`3oool01@3ooml0W?l0003o0000o`00B?oO
00X0oooo00L0oonL04So0000o`000?no0?ooo`3Oog@0M?m800800?l000D0B?oO0?ooo`3oool0oooo
0;oo0002003o000407Coo`3oool0oooo0?ooM0800?l000@00?no0?ooo`3oool0oomd0P00o`000`00
okl0oooo0?ooo`050?ooo`050?oo_`1do`000?l0003oB02Lool00P3oool01@3ooml0W?l0003o0000
odP0W?oo0080oooo00@0oomd003o0000o`000?no0P3oool01@3ookl0M?l0003o0000oi`0gooo0080
oooo00@0_ol0003o0000o`00M?oo0P3oool0103oog@00?l0003o0000okmI0?ooo`00ZP3oool01@3o
oi`0B?l0003o0000o`00M?oo00D0oooo00X0_ol0003o0000o`000?nL0=ooo`3oool0oonL04So0000
o`000?no0P3oool01@3oog@00?l0003o0000oi`0gooo0080oooo00L0oonL04So0000o`000?nL0=oo
o`3ooml0W?l000800?l000@0M?oo0?ooo`3ooml0W?l00P00o`002P1dool0oooo0?ooo`3ooi`0B?l0
003o0000oi`0gooo0?oog`2Lo`02003o000407Coo`3oool0oooO09co00800?l000H0M?oo0?ooo`3o
oi`0B?l0003o0000okl20?ooo`050?ooM000o`000?l0003oW03Oool02`3oool01@3OodP00?l0003o
0000oi`0gooo00T0oooo00/0oooO09co0000o`000?md0;ooo`3ooml0W?m804So0000o`000?nL0=oo
o`020?ooo`0@0?ooM000o`000?l0003oM02oool0oooo0?oo_`1do`000?l0003oM02oool0oooo0?oo
W018o`000?l004Sog`P0oooo00@0_ol0003o0000o`00B?oO0`3oool0103oog@00?l0003o0018oml2
0?ooo`050?ooM000o`000?l0003oW03Oool00P3oool0102oo`000?l0003o0018oml20?ooo`0;0?oo
M000o`000?l0003oM02oool0oooo0?oo_`1do`000?l0003oM02oool0F03oool00:X0oooo00D0oomd
003o0000o`000?m809coo`050?ooo`0A0?ooM000o`000?l0003oB02Lool0oooo0?oog`2Lo`000?l0
003oM02oool0oooo0?oog`2Lo`000?l0003oM02oool00P3oool01`3ooml0W?l0003o0000o`00B?oO
0?ooo`3OodP00P00o`001000odP0W?oo0?ooo`3OodP2003o0005003o_`3oool0oooo0?oog`2Lo`00
0P00o`000`18oml0oooo0=ooB002003o0004003oB02Lool0oooo0=ooB0800?l000d00?no0?ooo`3o
oml0W?l0003o0000og@0_ooo0?ooo`3ooml0W?l0003o0000og@0_ooo00/0oooo00D0oonL04So0000
o`000?m809coo`030?ooo`030=ooM02Lool0oooo00@0oooo00H0gom8003o0000o`00M?oo0?ooo`3o
og@2003o0006003oB02Lool0oooo0?ooo`3ookl0M?l00P00o`001018oml0oooo0?ooo`2oo`02003o
000607Coo`3oool0oono07Co0000o`000?no203oool01@3oog@00?l0003o0000oi`0gooo0080oooo
00/0oono07Co0000o`000?nL0=ooo`3oool0oono07Co0000o`000?m809coo`020?ooo`080?ooW018
o`000?l0003oW03Oool0oooo0?oo_`1do`02003o000404Sog`3oool0oooo0;oo00800?l000<0M?oo
0?ooo`3oool0EP3oool00:X0oooo00D0gom8003o0000o`000?l007Coo`050?ooo`050?ooM000o`00
0?l0003o0000okl00P3oool0103oog@00?l0003o001dool20?ooo`040?ooM000o`000?l007Coo`<0
oooo0100gom8003o0000o`00B?md09coo`3ooi`0B?l0003o0000og@0M?nL0=ooo`3ooi`0B?l0003o
0000og@0_ooo0P3oool04P3OodP00?l0003o0018og@0W?oo0?ooW018o`000?l0003oM01doi`0gooo
0?ooW018o`000?l0003oM02oool0oooo0?ooM0800?l000@0M?oo0?ooo`3oool0oomd0P00o`000`1d
ool0oooo0?ooo`090?ooo`050?oo_`1do`000?l0003o0018oml00`3oool00`3OodP0M?oo0?ooo`04
0?ooo`070?ooW018o`000?l0003oW03Oool0oooo0;oo0002003o000504Sog`3oool0oooo0?ooo`2o
o`000P00o`001018odP0M?oo0?ooo`3oog@2003o0007003o_`3oool0oooo0=ooB000o`000?md0;oo
o`070?ooo`050?oo_`1do`000?l0003oM02oool00`3oool0102oo`000?l0003oB02Lool20?ooo`04
0;oo0000o`000?l004Sog`80oooo00D0oono07Co0000o`000?m809coo`020?ooo`070;oo0000o`00
0?l004SoB01dool0oooo0?ooM002003o0003003o_`3oool0oooo05H0oooo002Z0?ooo`050=ooB000
o`000?l0003o001dool00`3oool00`3ooml0W?l0003o0004003o000<003oM02oool0oooO09co0000
o`000?nL0=ooo`3oool0oonL04So0000o`000?no0`3oool03@3ooi`0B?l0003o0000oi`0W?nL0=oo
g`2Lo`000?l0003oB02Loi`0M?oO0?oog`2Lo`000P00o`00401dool0oooo0?ooo`3ooi`0B?l0003o
0000oi`0W?nL0=oog`2Lo`000?l0003oB02Loi`0M?oO0?oog`2Lo`02003o000<07Coo`3oool0oooO
09co0000o`000?nL0=ooo`3oool0oonL04So0000o`000?no303oool02@2oo`000?l0003o0000oi`0
gooo0?ooo`3ookl0M?l0003o_`050?ooo`070?oog`2Lo`000?l0003oB02Lool0oooo0?ooM002003o
0005003o_`3oool0oooo0?ooo`3oog@00P00o`003@00okl0_omd0;ooo`3ookl0M?l0003o0000odP0
W?oo0?ooo`3ookl0M?l0003o0018oml00P3oool0103ooml0W?l0003o001dool20?ooo`040;oo0000
o`000?l007Coo`<0oooo00@0oonL04So0000o`00B?oO0P3oool01@3ookl0M?l0003o0000og@0_ooo
0080oooo00@0_ol0003o0000odP0W?oo0P3oool02`3oog@00?l0003o0000okl0_omd0;ooo`3ookl0
M?l0003o0000odP0W?oo05L0oooo002Z0?ooo`050?ooM000o`000?l0003o0000okl03@3oool02P3o
oi`0B?l0003o001dool0oooo0?oo_`1do`000?l0003oM02oool20?ooo`070?oo_`1do`000?l0003o
M02ooi`0B?no0=ooB002003o000404Sog`3OodP0B?oO0=ooB0800?l000X00?no0?ooo`3oool0oono
07Co0000o`000?md0;ooW018okl0gom80P00o`001018oml0gom804Sog`3OodP2003o000=003o_`3o
ool0oooo0?ooW018o`000?l007Coo`3oool0oono07Co0000o`000?md0;ooo`0;0?ooo`030?ooM000
o`000?l000D00?l000<00?md0;ooo`3oool0103oool02P3OodP00?l0003o0018oml0oooo0?oo_`1d
o`000?l0003oM02oool20?ooo`080?oo_`1do`000?l0003oM02ookl0M?md0;ooo`2oo`02003o000?
07Coo`3oool0oooo0?ooW018o`000?l007Coo`3oool0oooO09co0000o`000?nL0=ooo`3oool0oomd
00800?l000T00?no0?ooo`3oool0oooo0?oo_`1do`000?l0003oW03Oool00P3oool0103ooi`0B?l0
003o0018oml20?ooo`040;oo0000o`000?m809coo`80oooo00P0oono07Co0000o`000?md0;oo_`1d
og@0_ooo0;oo00800?l000<0M?oo0?ooo`3oool0E@3oool00:X0oooo00H0oono07Co0000o`000?l0
003oM02oool=0?ooo`060?ooW018o`000?md0;ooo`3ooi`0M?m80P00o`000`1dool0oooo0?ooM004
003o000307Coo`3oog@00?l000800?l000<00?no0?ooo`3OodP00`00o`000`00okl0oooo0?ooM004
003o000307Coo`3oog@00?l000800?l000<00?no0?ooo`3OodP00`00o`002P00okl0oooo0?ooo`3o
ool0oonL04So0000og@0_ooo0?ooW01dodP2003o000307Coo`3oool0oooo00T0oooo00D0oono07Co
0000o`000?l004Sog`020?ooo`030=ooB000odP0W?oo00@0oooo00P0gom8003o0000o`000?l0003o
W03Oool0oooo0;oo00800?l000<0M?oo0?ooo`3oog@01000o`000`1dool0oono07Co0003003o000E
07Coo`3oool0oooo0?ooo`3ookl0M?l0003o0018oml0oooo0?oo_`1do`000?nL0=ooo`3oool0oono
07Co0000o`000?md0;ooo`3oool0oomd00<00?l000<00?m809coo`3oool00P3oool02`3ooi`0B?l0
003oB02Lool0oooO09co0000o`000?nL0=ooo`3oool0oomd00@00?l000<0M?oo0?oo_`1do`000`00
o`000`1dool0oooo0?ooo`1E0?ooo`00Z`3oool01@2oo`000?l0003o0000o`00M?oo00h0oooo00P0
oooO09co0000o`000?m809coW018o`000?l0003o_`<0oooo00H0oonL04So0000o`000?no0?ooo`3o
og@2003o000707Coo`3oool0oooo0?ooM000o`000?m809coo`030?ooo`060?ooW018o`000?l0003o
_`3oool0oomd0P00o`001`1dool0oooo0?ooo`3oog@00?l0003oB02Lool0103oool0203ooml0W?l0
003o0000odP0W?nL04So0000o`000?no303oool01@2oo`000?l0003o0000oi`0gooo0080oooo00<0
oomd04Sog`3oool01@3oool01P3OodP00?l0003oB02Lool0oooo0?ooM0800?l000h00?no0?ooo`3o
ool0oooo0?ooW018o`000?l0003o_`3oool0oooO09co0000o`000?nL0=ooo`D0oooo00P0oomd003o
0000o`000?l0003oM02oool0oooo0?ooM0D00?l000L00?nL0=ooo`3oool0oonL04So0000o`00B?oO
00D0oooo00H0gom8003o0000o`000?l0003oW03Oool40?ooo`0:0?ooW018o`000?l0003o_`3oool0
oooO09co0000o`000?nL0=oooeL0oooo002[0?ooo`060?oo_`1do`000?l0003o0000og@0_ooo1`3o
ool00`3ooi`0M?oO0?ooo`170?ooo`050?ooM000o`000?l0003oB02Lool00P3oool00`3ooml0W?nL
0=ooo`0Y0?ooo`050?ooM000o`000?l0003oW03Oool0MP3oool00:`0oooo00D0oonL04So0000o`00
0?l0003o_`070?ooo`030?ooM000oi`0gooo04L0oooo00D0oono07Co0000o`000?l004Sog`050?oo
o`030?ooM01dool0oooo02L0oooo00@0oooO09co0000odP0W?ooMP3oool00:d0oooo00D0oono07Co
0000o`000?l0003o_`060?ooo`030;oo0000og@0_ooo04P0oooo00@0_ol0003o0000o`000?no1@3o
ool00`3OodP00?no0?ooo`0Y0?ooo`030=ooB018oml0oooo0080oooo00@0_ol0003o0000oi`0gooo
K`3oool00:l0oooo00D0gom8003o0000o`000?l007Coo`030?ooo`040=ooB000o`000?m809coodL0
oooo00H0oooO09co0000o`000?l0003oM02oool30?ooo`040?ooM000o`000?nL0=ooobd0oooo00D0
oonL04So0000o`000?l007Coo`1_0?ooo`00/@3oool00`3ooi`0B?l0003o0004003o000307Coo`3O
odP0M?oo04H0oooo00<0_ol0003o0000o`002@00o`000`00odP0W?oo0?ooo`0/0?ooo`050?oog`2L
o`000?l0003oW03Oool0K`3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0
oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-9.8394, -24.7572, \
0.0574018, 0.247887}}]
}, Closed]],

Cell[TextData[{
  "Alternatively, you can use ",
  ButtonBox["StyleForm",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["string,options", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " to select the formatting options for any text string."
}], "Text"],

Cell[BoxData[
    \(xlabel = 
      StyleForm["\<x\>", FontFamily \[Rule] "\<Times\>", FontSize \[Rule] 12, 
        FontSlant \[Rule] "\<Italic\>", \ FontWeight \[Rule] "\<Bold\>", 
        FontColor \[Rule] RGBColor[0, 1, 0]]; 
    ylabel = StyleForm["\<\[CapitalGamma][x]\>", 
        FontFamily \[Rule] "\<Times\>", FontSize \[Rule] 16, 
        FontSlant \[Rule] "\<Italic\>", \ FontWeight \[Rule] "\<Bold\>", 
        FontColor \[Rule] RGBColor[1, 0, 0]]; 
    plotLabel = 
      StyleForm["\<Gamma Function\>", 
        FontFamily \[Rule] "\<Lucida-Calligraphy\>", FontSize \[Rule] 18, 
        FontSlant \[Rule] "\<Italic\>", \ FontWeight \[Rule] "\<Bold\>", 
        FontColor \[Rule] RGBColor[0, 0.5, 0.5]];\)], "Input"],

Cell["These formatted strings can be printed", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Print[xlabel, "\<    \>", ylabel, "\<    \>", plotLabel]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      RowBox[{
        TagBox[
          StyleBox["\<\"x\"\>",
            FontFamily->"Times",
            FontSize->12,
            FontWeight->"Bold",
            FontSlant->"Italic",
            FontColor->RGBColor[0, 1, 0]],
          (
          StyleForm[ #, FontFamily -> "Times", FontSize -> 12, FontSlant -> 
            "Italic", FontWeight -> "Bold", FontColor -> RGBColor[ 
            0, 1, 0]]&)], "\[InvisibleSpace]", "\<\"    \"\>", 
        "\[InvisibleSpace]", 
        TagBox[
          StyleBox["\<\"\[CapitalGamma][x]\"\>",
            FontFamily->"Times",
            FontSize->16,
            FontWeight->"Bold",
            FontSlant->"Italic",
            FontColor->RGBColor[1, 0, 0]],
          (
          StyleForm[ #, FontFamily -> "Times", FontSize -> 16, FontSlant -> 
            "Italic", FontWeight -> "Bold", FontColor -> RGBColor[ 
            1, 0, 0]]&)], "\[InvisibleSpace]", "\<\"    \"\>", 
        "\[InvisibleSpace]", 
        TagBox[
          StyleBox["\<\"Gamma Function\"\>",
            FontFamily->"Lucida-Calligraphy",
            FontSize->18,
            FontWeight->"Bold",
            FontSlant->"Italic",
            FontColor->RGBColor[0, 0.500008, 0.500008]],
          (
          StyleForm[ #, FontFamily -> "Lucida-Calligraphy", FontSize -> 18, 
            FontSlant -> "Italic", FontWeight -> "Bold", FontColor -> 
            RGBColor[ 0, .5, .5]]&)]}],
      SequenceForm[ 
        StyleForm[ 
        "x", FontFamily -> "Times", FontSize -> 12, FontSlant -> "Italic", 
          FontWeight -> "Bold", FontColor -> RGBColor[ 0, 1, 0]], "    ", 
        StyleForm[ 
        "\[CapitalGamma][x]", FontFamily -> "Times", FontSize -> 16, 
          FontSlant -> "Italic", FontWeight -> "Bold", FontColor -> RGBColor[ 
          1, 0, 0]], "    ", 
        StyleForm[ 
        "Gamma Function", FontFamily -> "Lucida-Calligraphy", FontSize -> 18, 
          FontSlant -> "Italic", FontWeight -> "Bold", FontColor -> RGBColor[ 
          0, .5, .5]]],
      Editable->False]], "Print"]
}, Closed]],

Cell["or used to label plots.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Gamma[x], {x, \(-5\), 5}, \ GridLines \[Rule] Automatic, 
        Frame \[Rule] True, FrameLabel \[Rule] {xlabel, \ ylabel}, \ 
        PlotLabel \[Rule] plotLabel];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.0952381 0.316178 0.0220537 [
[.11905 -0.0125 -9.15625 -15.6875 ]
[.11905 -0.0125 9.15625 0 ]
[.30952 -0.0125 -9.15625 -15.6875 ]
[.30952 -0.0125 9.15625 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.69048 -0.0125 -5 -15.6875 ]
[.69048 -0.0125 5 0 ]
[.88095 -0.0125 -5 -15.6875 ]
[.88095 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -5.28125 -31.375 ]
[.5 -0.0125 5.28125 -15.6875 ]
[-0.0125 .09564 -24.3125 -7.84375 ]
[-0.0125 .09564 0 7.84375 ]
[-0.0125 .20591 -18.3125 -7.84375 ]
[-0.0125 .20591 0 7.84375 ]
[-0.0125 .31618 -10 -7.84375 ]
[-0.0125 .31618 0 7.84375 ]
[-0.0125 .42645 -10 -7.84375 ]
[-0.0125 .42645 0 7.84375 ]
[-0.0125 .53672 -16 -7.84375 ]
[-0.0125 .53672 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -40 -18.4688 ]
[-0.01875 .30902 -24.3125 18.4688 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -75.5938 0 ]
[.5 .63053 75.5938 18.0625 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.11905 0 m
.11905 .61803 L
s
.30952 0 m
.30952 .61803 L
s
.5 0 m
.5 .61803 L
s
.69048 0 m
.69048 .61803 L
s
.88095 0 m
.88095 .61803 L
s
0 .09564 m
1 .09564 L
s
0 .20591 m
1 .20591 L
s
0 .31618 m
1 .31618 L
s
0 .42645 m
1 .42645 L
s
0 .53672 m
1 .53672 L
s
0 g
.11905 0 m
.11905 .00625 L
s
gsave
.11905 -0.0125 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.30952 0 m
.30952 .00625 L
s
gsave
.30952 -0.0125 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.69048 0 m
.69048 .00625 L
s
gsave
.69048 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.88095 0 m
.88095 .00625 L
s
gsave
.88095 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.2619 0 m
.2619 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.7381 0 m
.7381 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.07143 0 m
.07143 .00375 L
s
.02381 0 m
.02381 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.97619 0 m
.97619 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -66.2813 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.563 15.813 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(x) show
69.563 15.813 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .09564 m
.00625 .09564 L
s
gsave
-0.0125 .09564 -85.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
0 .20591 m
.00625 .20591 L
s
gsave
-0.0125 .20591 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .31618 m
.00625 .31618 L
s
gsave
-0.0125 .31618 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .42645 m
.00625 .42645 L
s
gsave
-0.0125 .42645 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .53672 m
.00625 .53672 L
s
gsave
-0.0125 .53672 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .1177 m
.00375 .1177 L
s
0 .13975 m
.00375 .13975 L
s
0 .1618 m
.00375 .1618 L
s
0 .18386 m
.00375 .18386 L
s
0 .22796 m
.00375 .22796 L
s
0 .25002 m
.00375 .25002 L
s
0 .27207 m
.00375 .27207 L
s
0 .29412 m
.00375 .29412 L
s
0 .33823 m
.00375 .33823 L
s
0 .36029 m
.00375 .36029 L
s
0 .38234 m
.00375 .38234 L
s
0 .40439 m
.00375 .40439 L
s
0 .4485 m
.00375 .4485 L
s
0 .47055 m
.00375 .47055 L
s
0 .49261 m
.00375 .49261 L
s
0 .51466 m
.00375 .51466 L
s
0 .07359 m
.00375 .07359 L
s
0 .05153 m
.00375 .05153 L
s
0 .02948 m
.00375 .02948 L
s
0 .00743 m
.00375 .00743 L
s
0 .55877 m
.00375 .55877 L
s
0 .58082 m
.00375 .58082 L
s
0 .60288 m
.00375 .60288 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -111.625 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
79.4688 11.8438 translate
90 rotate
-79.4688 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
%%BeginResource: font Times-BoldItalic-MISO
%%BeginFont: Times-BoldItalic-MISO
/Times-BoldItalic /Times-BoldItalic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-BoldItalic-MISO
%%IncludeFont: Times-BoldItalic-MISO
/Times-BoldItalic-MISO findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1-Bold
%%IncludeFont: Mathematica1-Bold
/Mathematica1-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(G) show
%%IncludeResource: font Mathematica2-Bold
%%IncludeFont: Mathematica2-Bold
/Mathematica2-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
75.500 15.813 moveto
(@) show
82.063 15.813 moveto
%%IncludeResource: font Times-BoldItalic-MISO
%%IncludeFont: Times-BoldItalic-MISO
/Times-BoldItalic-MISO findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(x) show
%%IncludeResource: font Mathematica2-Bold
%%IncludeFont: Mathematica2-Bold
/Mathematica2-Bold findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
90.188 15.813 moveto
(D) show
95.938 15.813 moveto
%%IncludeResource: font Times-BoldItalic-MISO
%%IncludeFont: Times-BoldItalic-MISO
/Times-BoldItalic-MISO findfont 16.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.11905 .61178 m
.11905 .61803 L
s
.30952 .61178 m
.30952 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.69048 .61178 m
.69048 .61803 L
s
.88095 .61178 m
.88095 .61803 L
s
.125 Mabswid
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.2619 .61428 m
.2619 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.7381 .61428 m
.7381 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.07143 .61428 m
.07143 .61803 L
s
.02381 .61428 m
.02381 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.97619 .61428 m
.97619 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .09564 m
1 .09564 L
s
.99375 .20591 m
1 .20591 L
s
.99375 .31618 m
1 .31618 L
s
.99375 .42645 m
1 .42645 L
s
.99375 .53672 m
1 .53672 L
s
.125 Mabswid
.99625 .1177 m
1 .1177 L
s
.99625 .13975 m
1 .13975 L
s
.99625 .1618 m
1 .1618 L
s
.99625 .18386 m
1 .18386 L
s
.99625 .22796 m
1 .22796 L
s
.99625 .25002 m
1 .25002 L
s
.99625 .27207 m
1 .27207 L
s
.99625 .29412 m
1 .29412 L
s
.99625 .33823 m
1 .33823 L
s
.99625 .36029 m
1 .36029 L
s
.99625 .38234 m
1 .38234 L
s
.99625 .40439 m
1 .40439 L
s
.99625 .4485 m
1 .4485 L
s
.99625 .47055 m
1 .47055 L
s
.99625 .49261 m
1 .49261 L
s
.99625 .51466 m
1 .51466 L
s
.99625 .07359 m
1 .07359 L
s
.99625 .05153 m
1 .05153 L
s
.99625 .02948 m
1 .02948 L
s
.99625 .00743 m
1 .00743 L
s
.99625 .55877 m
1 .55877 L
s
.99625 .58082 m
1 .58082 L
s
.99625 .60288 m
1 .60288 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .31618 m
1 .31618 L
s
.5 0 m
.5 .61803 L
s
gsave
.5 .63053 -136.594 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 26.0625 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 18.188 moveto
%%IncludeResource: font Lucida-Calligraphy-BoldItalic
%%IncludeFont: Lucida-Calligraphy-BoldItalic
/Lucida-Calligraphy-BoldItalic findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.500 0.500 setrgbcolor
0.000 0.000 rmoveto
63.000 18.188 moveto
%%IncludeResource: font Lucida-Calligraphy-BoldItalic
%%IncludeFont: Lucida-Calligraphy-BoldItalic
/Lucida-Calligraphy-BoldItalic findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.500 0.500 setrgbcolor
(Gamma) show
133.938 18.188 moveto
(Function) show
210.188 18.188 moveto
%%IncludeResource: font Lucida-Calligraphy-BoldItalic
%%IncludeFont: Lucida-Calligraphy-BoldItalic
/Lucida-Calligraphy-BoldItalic findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.500 0.500 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02498 0 m
.02499 .30097 L
.02605 .30806 L
.02729 .31081 L
.02846 .31207 L
.02954 .31277 L
.03053 .31322 L
.03163 .31358 L
.03279 .31386 L
.03395 .31407 L
.0352 .31425 L
.03746 .31449 L
.03884 .31459 L
.04016 .31468 L
.04262 .31479 L
.045 .31487 L
.0463 .31491 L
.04753 .31493 L
.04868 .31495 L
.04993 .31497 L
.05105 .31499 L
.0521 .315 L
.05335 .315 L
.05401 .31501 L
.05472 .31501 L
.05546 .31501 L
.05617 .31501 L
.05749 .31501 L
.05872 .31501 L
.05986 .31501 L
.06111 .315 L
.06244 .31499 L
.06504 .31496 L
.06742 .31493 L
.07003 .31488 L
.07277 .31482 L
.07741 .31469 L
.08247 .31447 L
.08537 .31431 L
.08801 .31413 L
.09085 .3139 L
.09392 .31357 L
.09678 .31317 L
.09941 .31269 L
.10192 .31208 L
.10319 .31169 L
.10458 .31119 L
.10589 .31061 L
.10711 .30996 L
.10837 .30912 L
Mistroke
.10951 .30816 L
.11064 .30696 L
.11169 .30549 L
.11281 .30338 L
.11402 .30005 L
.11524 .29447 L
.11594 .28934 L
.11659 .28193 L
.1179 .24147 L
.11863 .10742 L
Mfstroke
.11863 .10742 m
.11924 .61803 L
s
.1195 .61803 m
.1205 .37806 L
.12179 .3495 L
.12302 .3397 L
.12415 .33484 L
.1252 .33195 L
.12634 .32978 L
.1274 .32827 L
.12838 .32721 L
.13065 .32545 L
.13179 .32481 L
.13304 .32425 L
.1354 .32343 L
.13793 .32282 L
.14031 .32241 L
.14247 .32213 L
.14479 .32191 L
.1461 .32181 L
.14733 .32174 L
.14863 .32168 L
.14981 .32164 L
.15047 .32162 L
.15118 .32161 L
.15247 .32159 L
.15358 .32158 L
.15479 .32159 L
.15607 .3216 L
.15726 .32163 L
.15837 .32166 L
.15939 .32169 L
.16169 .32179 L
.1642 .32194 L
.16653 .32212 L
.1693 .32239 L
.17182 .32269 L
.17698 .32351 L
.17982 .32411 L
.1825 .32481 L
.18712 .3264 L
.18975 .32761 L
.19221 .32905 L
.1948 .33099 L
.19716 .33331 L
.19847 .33491 L
.19989 .33701 L
.20123 .3394 L
.20249 .34218 L
.20366 .34535 L
.20492 .3497 L
.20607 .35483 L
Mistroke
.20712 .36102 L
.20827 .37022 L
.20952 .38546 L
.21083 .41327 L
.21206 .46886 L
Mfstroke
.21206 .46886 m
.21308 .61803 L
s
.21321 .61803 m
.21324 0 L
s
.21552 0 m
.21634 .14124 L
.2175 .20237 L
.21873 .23237 L
.21978 .24731 L
.22094 .25829 L
.22215 .26631 L
.22329 .27179 L
.22453 .27638 L
.22586 .28015 L
.22725 .28324 L
.22857 .28557 L
.23104 .28886 L
.23335 .29108 L
.23548 .29262 L
.23782 .2939 L
.2403 .2949 L
.24277 .29562 L
.24416 .29593 L
.24543 .29615 L
.24661 .29631 L
.2479 .29645 L
.24911 .29653 L
.25023 .29658 L
.25092 .29659 L
.25167 .29659 L
.25237 .29658 L
.25301 .29656 L
.25424 .2965 L
.25553 .29639 L
.25692 .29624 L
.25821 .29606 L
.26116 .2955 L
.26377 .29484 L
.26619 .29406 L
.26879 .29303 L
.27165 .29162 L
.27646 .28846 L
.2791 .28618 L
.28157 .28356 L
.28396 .28046 L
.28624 .27684 L
.28831 .27281 L
.29051 .26745 L
.2929 .25995 L
.29511 .25064 L
.29647 .24331 L
.29775 .23481 L
.299 .22445 L
.30014 .21255 L
Mistroke
.3013 .19698 L
.30254 .17446 L
.3037 .14491 L
.30479 .10353 L
.30606 .02245 L
Mfstroke
.30606 .02245 m
.30623 0 L
s
.30951 0 m
.30959 .61803 L
s
.31325 .61803 m
.31377 .5745 L
.31452 .53772 L
.31521 .51221 L
.31648 .47896 L
.31763 .45783 L
.31895 .43994 L
.32021 .4272 L
.32245 .41078 L
.32358 .40457 L
.32482 .39886 L
.32701 .39091 L
.32907 .38525 L
.33149 .38014 L
.3337 .37656 L
.33625 .3734 L
.33868 .37113 L
.33993 .3702 L
.34127 .36935 L
.34246 .36872 L
.34372 .36817 L
.34485 .36777 L
.3461 .36742 L
.34723 .3672 L
.34828 .36705 L
.34892 .367 L
.34961 .36696 L
.35087 .36697 L
.35209 .36706 L
.35322 .36722 L
.35444 .36746 L
.35579 .36783 L
.35851 .36888 L
.3609 .37016 L
.36308 .37164 L
.36544 .37362 L
.36801 .37629 L
.37235 .38225 L
.37471 .38649 L
.37691 .39123 L
.37937 .39769 L
.38197 .40627 L
.38483 .41853 L
.38745 .43369 L
.3899 .45292 L
.39219 .47794 L
.3935 .49701 L
.3947 .51886 L
.39609 .55172 L
.3974 .59444 L
Mistroke
Mfstroke
.3974 .59444 m
.39792 .61803 L
s
.40444 .61803 m
.40452 0 L
s
.41171 0 m
.41218 .0211 L
.41288 .04545 L
.41353 .0644 L
.4148 .09408 L
.416 .116 L
.41716 .13311 L
.41843 .14837 L
.42066 .16903 L
.42325 .18655 L
.42567 .19862 L
.42797 .20755 L
.43005 .21403 L
.43233 .21976 L
.43473 .22461 L
.43738 .22883 L
.43983 .23186 L
.44246 .23434 L
.44397 .23544 L
.44535 .23626 L
.44655 .23684 L
.44785 .23734 L
.44856 .23755 L
.44922 .23771 L
.45052 .23792 L
.45164 .238 L
.45287 .23798 L
.45415 .23783 L
.45535 .23757 L
.45653 .23721 L
.45779 .2367 L
.46005 .23544 L
.46274 .23334 L
.46519 .2308 L
.46774 .22737 L
.47052 .22261 L
.47313 .21685 L
.47554 .21016 L
.47761 .20305 L
.47982 .19357 L
.48226 .18014 L
.48453 .16348 L
.48575 .15223 L
.48708 .13745 L
.48826 .12155 L
.48951 .10063 L
.49074 .07422 L
.49186 .04329 L
.49299 .00213 L
s
.49299 .00213 m
.49303 0 L
s
.50068 0 m
.50068 .61803 L
s
.50675 .61803 m
.50762 .58072 L
.50876 .54517 L
.51008 .51386 L
.51153 .48796 L
.51289 .46901 L
.51417 .45451 L
.5166 .43324 L
.51881 .41878 L
.52119 .40665 L
.52376 .39633 L
.52839 .38259 L
.53266 .37354 L
.53735 .3661 L
.54238 .36007 L
.54713 .35566 L
.55147 .35243 L
.55654 .34934 L
.56207 .34665 L
.56725 .34459 L
.57212 .34299 L
.58151 .34059 L
.58692 .33953 L
.59188 .33871 L
.60198 .33743 L
.60729 .33694 L
.61288 .33652 L
.61787 .33623 L
.62241 .33602 L
.62493 .33593 L
.62765 .33585 L
.63023 .3358 L
.63137 .33577 L
.6326 .33576 L
.63375 .33574 L
.63479 .33573 L
.63585 .33572 L
.63685 .33572 L
.63801 .33571 L
.63908 .33571 L
.64026 .33571 L
.64152 .33571 L
.64274 .33572 L
.64344 .33573 L
.64407 .33573 L
.64647 .33576 L
.64866 .3358 L
.65096 .33585 L
.65593 .33599 L
.66045 .33616 L
Mistroke
.67062 .33667 L
.68004 .33732 L
.68896 .33809 L
.70889 .34039 L
.72732 .34334 L
.746 .34731 L
.76661 .35313 L
.78598 .3604 L
.80684 .37082 L
.82691 .38426 L
.84801 .40339 L
.86777 .4277 L
.88615 .45795 L
.90663 .50371 L
.92522 .56051 L
Mfstroke
.92522 .56051 m
.93903 .61803 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00lP3oool0103ooml0W?l0003o0018oml20?ooo`040?ooM000o`000?l0003o_k@0oooo003c
0?ooo`0:0;oo0000o`000?l0003o_`3ooml0W?l0003o001dool0oomd04Sogk<0oooo003f0?ooo`05
0?ooM018odP00?l0003oM02oool0]@3oool00?L0oooo00<0_ol0003o0000okl0]P3oool00?H0oooo
00@0oono07Co0000o`00M?oo]P3oool00?H0oooo00D0_ol0003o0000og@0M?md0;ooo`2e0?ooo`00
m03oool02P3ooi`0B?l0003o0000o`000?no0?oo_`1do`000?l0003oB02Loonb0?ooo`00mP3oool0
202oo`000?m809coo`3oool0oono07Co0000odP0W?oo/P3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool008`0oooo00@0
oooO09a80000B9`0gooo<@3oool00`2oM0000000000000040000000304RLg`3oool0oooo02X0oooo
00D0omnL04P000000000001d0;ooo`0b0?ooo`030;md00000000000000@0000000<0B9cO0?ooo`3o
ool0<`3oool0103ooml0W4P00018W03Ooom30?ooo`00S03oool0103ooml0W4P00018W03Ooolb0?oo
o`030=nLB0000000000000<0000000<004RL0=ooo`3oool0:@3oool01`3ookl0M00007Boo`3oool0
okmd0000M02oool0<P3oool00`3OW4P0000000000003000000030018W03Oool0oooo03<0oooo00@0
oooO09a80000B9`0gooo@`3oool008`0oooo00@0oooO09a80000B9`0gooo<`3oool00`3OW4P0B9cO
0?ooo`020?ooo`030=nLM02Lgol0oooo02T0oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo
0?ooo`0b0?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0<`3oool0103ooml0W4P0
0018W03Ooom30?ooo`00R03oool00`2oM0000000000000050000000307Boo`3oool0oooo0340oooo
00<0gia807Boo`3oool0;03oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo03<0oooo
00<0gia807Boo`3oool0<`3oool00`2oM0000000000000050000000307Boo`3oool0oooo0400oooo
001k0?ooo`030?oog`2LB000000000L0000000D0B9cO0?ooo`3oool0omnL07BLg`020?ooo`040?oo
g`2LB00004RL0=ooob@0oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`040?ooo`03
0;mdB02Lgol0oooo02/0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0d0?ooo`03
0;mdB02Lgol0oooo0380oooo00P0omnL07BLg`3oool0oooo0?oog`2LB00004RL0=oood<0oooo0029
0?ooo`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`0e0?ooo`030?oog`2LB7@0_ooo02/0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0d0?ooo`030?oog`2LB7@0_ooo03<0oooo
00L0gia807Boo`3oool0oooO09a80000B9`0gooo04<0oooo00290?ooo`070?oog`2LB7@0_ooo0?oo
g`2LB00004RL0=ooo`0f0?ooo`030?noM000M;l0oooo02X0oooo00<0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`0e0?ooo`030?noM000M;l0oooo0380oooo00L0oooO09a8M02oool0oooO09a8
0000B9`0gooo04<0oooo002:0?ooo`060?oOW018M;l0oooO09a80000B9`0gooo=P3oool00`3ookl0
M01809cOo`0Z0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0=@3oool00`3ookl0
M01809cOo`0c0?ooo`060?oOW018M;l0oooO09a80000B9`0gooo@`3oool008/0oooo00D0gia807Bo
g`2LB00004RL0=ooo`0f0?ooo`030?oo_`1d0000B9cO02X0oooo00<0oono07@00018W=l00`3oool0
0`3OW4P0001d0;ooo`0e0?ooo`030?oo_`1d0000B9cO03@0oooo00D0gia807Bog`2LB00004RL0=oo
o`130?ooo`00R`3oool01@3ooml0W4Q804Q80000B9`0gooo0340oooo00<0oknL09cOo`3oool00P3o
ool00`3ogi`0B00004RLg`0[0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0<03o
ool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO03@0oooo00D0oooO09a8B018B00004RL
0=ooo`130?ooo`00S03oool0103ogi`0B0000018W03Ooola0?ooo`080?oog`2LB4P0W=oo0?ooo`3o
okl0M0000000M02oool[0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0a0?ooo`080?oo
g`2LB4P0W=oo0?ooo`3ookl0M0000000M02ooole0?ooo`040?oOW018000004RL0=oood<0oooo002=
0?ooo`030=nLB000B9`0gooo0380oooo00H0oooO09a80000000000000000M02oool]0?ooo`050?oo
_`1d000000000000M02oool0<`3oool01P3ooml0W4P0000000000000001d0;ooocL0oooo00<0gia8
0018W03Oool0@`3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool006<0ooooo`00
000/00000280oooo001S0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool2
000000d0oooo00<000000?ooo`3oool02P3oool2000000d0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0000000<0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`0000003@3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0200oooo001S0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool02`3oool2000000d0oooo00<000000?ooo`3oool02P3oool20000
00d0oooo00<000000?ooo`3oool02`3oool00`000000oooo0000000<0?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`0000003@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0200oooo001S0?ooo`8000001@3oool00`000000oooo0?ooo`0I0?ooo`030000O`3oool0oooo01T0
oooo0P00000J0?ooo`8000006`3oool00`000000oooo0000000H0?ooo`030000003oool0000003L0
oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0P0?ooo`8000008P3oool006<0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0I0?ooo`030000O`3oool0oooo01T0oooo
0P00000J0?ooo`8000006`3oool010000000oooo0?ooo`00000G0?ooo`030000003oool0000003L0
oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200
oooo001S0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06@3oool00`0007l0oooo
0?ooo`0I0?ooo`8000006P3oool2000001/0oooo00@000000?ooo`3oool000005`3oool00`000000
oooo0000000g0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`000000
oooo0?ooo`0P0?ooo`00H`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01T0oooo
00<0001o0?ooo`3oool06@3oool2000001X0oooo0P00000K0?ooo`040000003oool0oooo000001L0
oooo00<000000?ooo`000000=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0240
oooo00<000000?ooo`3oool0803oool006<0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`0I0?ooo`030000O`3oool0oooo01T0oooo0P00000J0?ooo`8000006`3oool010000000oooo
0?ooo`00000F0?ooo`040000003oool0oooo000003L0oooo00<0001o0?ooo`3oool0=P3oool00`00
07l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo001S0?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool06@3oool00`0007l0oooo0?ooo`0I0?ooo`8000006P3oool2000001/0
oooo00@000000?ooo`3oool000005P3oool010000000oooo0?ooo`00000g0?ooo`030000O`3oool0
oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`00H`3oool20000
00D0oooo00<000000?ooo`3oool06@3oool00`0007l0oooo0?ooo`0I0?ooo`8000006P3oool20000
01/0oooo00@000000?ooo`3oool000005P3oool010000000oooo0?ooo`00000g0?ooo`030000O`3o
ool0oooo03H0oooo00<0001o0?ooo`3oool0803oool200000280oooo001S0?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool06@3oool00`0007l0oooo0?ooo`0I0?ooo`8000006P3oool2
000001/0oooo00@000000?ooo`3oool000005P3oool010000000oooo0?ooo`00000g0?ooo`030000
O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`00H`3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01T0oooo00<0001o0?ooo`3oool06@3o
ool2000001X0oooo0P00000K0?ooo`040000003oool0oooo000001H0oooo00@000000?ooo`3oool0
0000=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3o
ool0803oool006<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0I0?ooo`030000
O`3oool0oooo01T0oooo0P00000J0?ooo`8000006`3oool010000000oooo0?ooo`00000F0?ooo`04
0000003oool0oooo000003L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?oo
o`030000003oool0oooo0200oooo001S0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool06@3oool00`0007l0oooo0?ooo`0I0?ooo`8000006P3oool2000001/0oooo00@000000?ooo`3o
ool000005P3oool010000000oooo0?ooo`00000g0?ooo`030000O`3oool0oooo03H0oooo00<0001o
0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`00H`3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo01T0oooo00<0001o0?ooo`3oool06@3oool2000001X0oooo0P00000K0?oo
o`040000003oool0oooo000001H0oooo00@000000?ooo`3oool00000=`3oool00`0007l0oooo0?oo
o`0f0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool006<0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`0I0?ooo`030000O`3oool0oooo01T0oooo0P00000J
0?ooo`8000006`3oool010000000oooo0?ooo`00000F0?ooo`040000003oool0oooo000003L0oooo
00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo
001S0?ooo`8000001@3oool00`000000oooo0?ooo`0I0?ooo`030000O`3oool0oooo01T0oooo0P00
000J0?ooo`8000006`3oool010000000oooo0?ooo`00000F0?ooo`040000003oool0oooo000003L0
oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0P0?ooo`8000008P3oool006<0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0I0?ooo`030000O`3oool0oooo01T0oooo
0P00000J0?ooo`8000006`3oool010000000oooo0?ooo`00000F0?ooo`040000003oool0oooo0000
03L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo
0200oooo001S0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06@3oool00`0007l0
oooo0?ooo`0I0?ooo`8000006P3oool2000001/0oooo00@000000?ooo`3oool000005P3oool01000
0000oooo0?ooo`00000g0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08@3oool0
0`000000oooo0?ooo`0P0?ooo`00H`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
01T0oooo00<0001o0?ooo`3oool06@3oool2000001X0oooo0P00000K0?ooo`040000003oool0oooo
000001H0oooo00@000000?ooo`3oool00000=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3o
ool0oooo0240oooo00<000000?ooo`3oool0803oool006<0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`0I0?ooo`030000O`3oool0oooo01T0oooo0P00000J0?ooo`8000006`3oool0
10000000oooo0?ooo`00000F0?ooo`040000003oool0oooo000003L0oooo00<0001o0?ooo`3oool0
=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo001S0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool06@3oool00`0007l0oooo0?ooo`0I0?ooo`8000006P3o
ool2000001/0oooo00@000000?ooo`3oool000005P3oool010000000oooo0?ooo`00000g0?ooo`03
0000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`00
H`3oool2000000D0oooo00<000000?ooo`3oool06@3oool00`0007l0oooo0?ooo`0I0?ooo`800000
6P3oool2000001/0oooo00@000000?ooo`3oool000005P3oool010000000oooo0?ooo`00000g0?oo
o`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0803oool200000280oooo001S0?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool06@3oool00`0007l0oooo0?ooo`0I0?ooo`80
00006P3oool2000001/0oooo00@000000?ooo`3oool000005P3oool010000000oooo0?ooo`00000g
0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P
0?ooo`00C03oool01P2oM000000000000000000004RL0=ooo`<0oooo00D0omnL04P000000000001d
0;ooo`090?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06@3oool00`0007l0oooo
0?ooo`0I0?ooo`8000006P3oool2000001/0oooo00@000000?ooo`3oool000005P3oool010000000
oooo0?ooo`00000g0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`00
0000oooo0?ooo`0P0?ooo`00C@3oool00`3ogi`0B00007Boo`040?ooo`070?oo_`1d0000M;oo0?oo
o`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06@3oool0
0`0007l0oooo0?ooo`0I0?ooo`8000006P3oool2000001/0oooo00@000000?ooo`3oool000005P3o
ool010000000oooo0?ooo`00000g0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0
8@3oool00`000000oooo0?ooo`0P0?ooo`00C@3oool00`3ogi`0B00007Boo`040?ooo`030;md0000
B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo01T0oooo00<0001o0?ooo`3oool06@3oool2000001X0oooo0P00000K0?ooo`040000003o
ool0oooo000001H0oooo00@000000?ooo`3oool00000=`3oool00`0007l0oooo0?ooo`0f0?ooo`03
0000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool004d0oooo00<0omnL04P0001d_ol0
0`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00L0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0I0?ooo`030000O`3oool0oooo01T0oooo0P00000J0?ooo`800000
6`3oool010000000oooo0?ooo`00000F0?ooo`040000003oool0oooo000003L0oooo00<0001o0?oo
o`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo000m0?ooo`03
0?oog`2LB000000000L0000000<0B9cO0?ooo`3oool00`3oool00`3ogi`0B00007Boo`030?ooo`03
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo01T0oooo00<0001o0?ooo`3oool06@3oool2000001X0oooo0P00000K0?ooo`04
0000003oool0oooo000001H0oooo00@000000?ooo`3oool00000=`3oool00`0007l0oooo0?ooo`0f
0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool004d0oooo00<0omnL04P0
001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo0P0000050000
O`030000000007l0001o03D0001o0P00000J0000O`8000006`0007l010000000001o0000O`00000F
0000O`040000000007l0001o00000980001o0`00000R0?ooo`00C@3oool00`3ogi`0B00007Boo`03
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo01T0oooo00<0001o0?ooo`3oool06@3oool2000001X0oooo0P00000K
0?ooo`040000003oool0oooo000001H0oooo00@000000?ooo`3oool00000=`3oool00`0007l0oooo
0?ooo`0f0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool004d0oooo00<0
omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`0I0?ooo`030000O`3oool0oooo01T0oooo0P00
000J0?ooo`8000006`3oool010000000oooo0?ooo`00000F0?ooo`040000003oool0oooo000003L0
oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200
oooo001=0?ooo`030?oOW0180000M;oo00<0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d
0;ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06@3oool00`000000oooo
0?ooo`0I0?ooo`8000006P3oool2000001/0oooo00D000000?ooo`3oool0oooo0000000E0?ooo`04
0000003oool0oooo000003L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?oo
o`030000003oool0oooo0200oooo001=0?ooo`030?oOW0180000M;oo00@0oooo00<0gia80018W03O
ool00P3oool00`2oM00007Bo0?ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool06@3oool00`000000oooo0?ooo`0I0?ooo`8000006P3oool2000001/0oooo00D000000?ooo`3o
ool0oooo0000000E0?ooo`040000003oool0oooo000003L0oooo00<0001o0?ooo`3oool0=P3oool0
0`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo001;0?ooo`050?oog`2LB0000000
0000001d_ol0103oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0203oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06@3oool2000001X0oooo
0P00000K0?ooo`050000003oool0oooo0?ooo`000000503oool01@000000oooo0?ooo`3oool00000
03L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo
0200oooo001=0?ooo`030?noM0000000M;oo00D0oooo00D0oono07@000000000001d0;ooo`090?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0I0?oo
o`8000006P3oool2000001/0oooo00D000000?ooo`3oool0oooo0000000D0?ooo`050000003oool0
oooo0?ooo`000000=`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0240oooo00<0
00000?ooo`3oool0803oool006<0oooo0P0000050?ooo`030000003oool0oooo01T0oooo00<00000
0?ooo`3oool06@3oool2000001X0oooo0P00000K0?ooo`050000003oool0oooo0?ooo`000000503o
ool01@000000oooo0?ooo`3oool0000003L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo
0?ooo`0P0?ooo`8000008P3oool006<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`0I0?ooo`030000003oool0oooo01T0oooo0P00000J0?ooo`8000006`3oool01@000000oooo0?oo
o`3oool0000001@0oooo00D000000?ooo`3oool0oooo0000000g0?ooo`030000O`3oool0oooo03H0
oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`00H`3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06@3oool2000001T0oooo
00<000000?ooo`0000006`3oool01@000000oooo0?ooo`3oool0000001@0oooo00D000000?ooo`3o
ool0oooo0000000g0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`00
0000oooo0?ooo`0P0?ooo`00H`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01T0
oooo00<000000?ooo`3oool06@3oool2000001T0oooo00<000000?ooo`0000006`3oool01@000000
oooo0?ooo`3oool0000001@0oooo00D000000?ooo`3oool0oooo0000000g0?ooo`030000O`3oool0
oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`00H`3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06@3oool20000
01T0oooo00<000000?ooo`0000006`3oool01@000000oooo0?ooo`3oool0000001@0oooo00D00000
0?ooo`3oool0oooo0000000g0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08@3o
ool00`000000oooo0?ooo`0P0?ooo`00H`3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo01T0oooo00<000000?ooo`3oool06@3oool2000001T0oooo00<000000?ooo`0000006`3oool0
1@000000oooo0?ooo`3oool0000001@0oooo00D000000?ooo`3oool0oooo0000000g0?ooo`030000
O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`00H`3o
ool2000000D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0I0?ooo`8000006@3o
ool00`000000oooo0000000K0?ooo`050000003oool0oooo0?ooo`000000503oool01@000000oooo
0?ooo`3oool0000003L0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0P0?ooo`80
00008P3oool006<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo01T0oooo0P00000I0?ooo`030000003oool0000001/0oooo00D000000?ooo`3oool0
oooo0000000D0?ooo`050000003oool0oooo0?ooo`000000=`3oool00`0007l0oooo0?ooo`0f0?oo
o`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool006<0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01T0oooo0P00000I0?ooo`03
0000003oool0000001/0oooo00D000000?ooo`3oool0oooo0000000C0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo
0240oooo00<000000?ooo`3oool0803oool006<0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo01T0oooo0P00000I0?ooo`030000003oool0000001/0
oooo00D000000?ooo`3oool0oooo0000000C0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0240oooo00<000000?oo
o`3oool0803oool006<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0I0?ooo`03
0000003oool0oooo01T0oooo0P00000I0?ooo`030000003oool0000001/0oooo00D000000?ooo`3o
ool0oooo0000000C0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0=@3oool00`00
07l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool006<0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01T0
oooo0P00000I0?ooo`030000003oool0000001/0oooo00D000000?ooo`3oool0oooo0000000C0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0f0?oo
o`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool006<0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01T0oooo0P00000I0?ooo`03
0000003oool0000001/0oooo00D000000?ooo`3oool0oooo0000000C0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo
0240oooo00<000000?ooo`3oool0803oool006<0oooo0P0000050?ooo`030000003oool0oooo01T0
oooo00<000000?ooo`3oool06@3oool2000001T0oooo00<000000?ooo`0000006`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0e0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0803oool200000280
oooo001S0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06@3oool00`000000oooo
0?ooo`0I0?ooo`8000006@3oool00`000000oooo0000000K0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo03D0oooo
00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo
001S0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`0I0?ooo`8000006@3oool00`000000oooo0000000K0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0403oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo03D0oooo00<0
001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo001S
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0I
0?ooo`8000006@3oool00`000000oooo0000000K0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool03`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo03D0oooo00<0001o
0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo001S0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0I0?oo
o`8000006@3oool00`000000oooo0000000K0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo03D0oooo00<0001o0?oo
o`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo001S0?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0I0?ooo`80
00006@3oool00`000000oooo0000000K0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3o
ool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo001S0?ooo`800000
1@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01T0oooo0P00000I0?ooo`030000
003oool0000001/0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000
O`3oool0oooo0200oooo0P00000R0?ooo`00H`3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool06@3oool2000001P0oooo00@000000?ooo`3oool00000
6`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0
8@3oool00`000000oooo0?ooo`0P0?ooo`00D`3oool00`3ooml0W4P0000000020000000307Boo`3o
ool0oooo00P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo01T0oooo0P00000H0?ooo`040000003oool0oooo000001/0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0
803oool005<0oooo00<0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`080?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0I0?ooo`800000603o
ool010000000oooo0?ooo`00000K0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0
=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo001H0?ooo`030?oog`2L
B7@0_ooo00P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo01T0oooo0P00000H0?ooo`040000003oool0oooo000001/0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0
803oool005T0oooo00<0_g@007Boo`3oool01`3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool06@3oool2000001P0oooo00@000000?ooo`3oool00000
6`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0
8@3oool00`000000oooo0?ooo`0P0?ooo`00AP3oool00`3ooml0W4P0000000070000000304RLg`3o
ool0oooo00D0oooo00<0oooO09a80018W=l0203oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool06@3oool2000001P0oooo00@000000?ooo`3oool00000
6`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0
8@3oool00`000000oooo0?ooo`0P0?ooo`00F03oool00`3OW4P0000007Boo`080?ooo`8000001@00
07l00`000000001o0000O`0I0000O`030000000007l0001o01T0001o0P00000H0000O`0400000000
07l0001o000001/0001o00<000000000O`0007l0100007l00`000000001o0000O`0;0000O`030000
000007l0001o00D0001o00<000000000O`0007l0T00007l300000280oooo001D0?ooo`070=oOg`3o
ool0oono07@000000000001d0;ooo`080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool06@3oool00`000000oooo0?ooo`0I0?ooo`800000603oool010000000oooo0?ooo`00000K0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?oo
o`030000003oool0oooo0200oooo001D0?ooo`060=nLB000000000000000000004P0W=oo2@3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06@3oool2
000001P0oooo00@000000?ooo`3oool000006`3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0e0?ooo`030000O`3o
ool0oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`00E03oool0
103ookl0M00000000018W=l;0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06@3o
ool00`000000oooo0?ooo`0I0?ooo`800000603oool010000000oooo0?ooo`00000K0?ooo`030000
003oool0oooo00D0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo03D0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000
003oool0oooo0200oooo001E0?ooo`030;md001d_ol0oooo00/0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`0000005`3o
ool010000000oooo0?ooo`00000K0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0
=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo001E0?ooo`060?oOW018
000000000000000004P0W=oo203oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01T0
oooo00<000000?ooo`3oool06@3oool00`000000oooo0000000G0?ooo`040000003oool0oooo0000
01/0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo
00H0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo
0240oooo00<000000?ooo`3oool0803oool005H0oooo00D0_g@0000000000000000004RLg`080?oo
o`8000001@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01T0oooo00<000000?oo
o`0000005`3oool010000000oooo0?ooo`00000K0?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo03D0oooo00<0001o
0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0P0?ooo`8000008P3oool00240oooo00<0oooO0?nL
_`3oool03P3oool00`3oM7@0oooo0?ooo`0^0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0000001H0oooo00D000000?oo
o`3oool0oooo0000000K0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0=P3oool0
0`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo000Q0?ooo`030?oog`3oM9`0oooo
00d0oooo00<0oono0?l0M03oool0;`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`3oool06@3oool00`000000oooo0000000F0?ooo`050000003oool0oooo
0?ooo`0000006`3oool00`000000oooo0?ooo`080?ooo`8000001@3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0Q
0?ooo`030000003oool0oooo0200oooo000Q0?ooo`030?oog`3oB000o`0000h0o`0000<0o`1d0?oo
o`3oool0;P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01T0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0000000F0?ooo`050000003oool0oooo0?ooo`0000006`3oool0
0`000000oooo0?ooo`090?ooo`H000002`3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0
oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`008@3oool00`3o
oml0odP00?l0000>0?l000030?l0M03oool0oooo02h0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`0I0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`0000005P3oool01@00
0000oooo0?ooo`3oool0000001/0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0e
0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P
0?ooo`008@3oool00`3ooml0odP00?l0000>0?l000030?l0M03oool0oooo02h0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01T0oooo00<000000?oo
o`0000005P3oool01@000000oooo0?ooo`3oool0000001/0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08@3oool0
0`000000oooo0?ooo`0P0?ooo`008@3oool00`3ooml0odP00?nLg`0=0?ooo`030?oo_`3o07@0oooo
02l0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`0000005P3oool01@000000oooo0?ooo`3oool0000001/0oooo00<00000
0?ooo`3oool06P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo03H0oooo00<0001o
0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`008@3oool00`3ooml0odP00?nLg`0>0?oo
o`030?mdM03oool0oooo02h0oooo0P0000050?ooo`030000003oool0oooo01T0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0000000F0?ooo`050000003oool0oooo0?ooo`0000006`3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0=P3oool0
0`0007l0oooo0?ooo`0P0?ooo`8000008P3oool00240oooo00<0oooO0?m8003oW=l0?`3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06@3oool01000
0000oooo0?ooo`00000E0?ooo`050000003oool0oooo0?ooo`0000006`3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?oo
o`0Q0?ooo`030000003oool0oooo0200oooo000Q0?ooo`030?oog`3oB000oicO03l0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01T0oooo00@00000
0?ooo`3oool00000503oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01T0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo03H0oooo00<0
001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`008@3oool00`3ooml0odP00?nLg`0o
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0603oool2000001/0oooo00@00000
0?ooo`3oool00000503oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01T0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo03H0oooo00<0
001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`008@3oool00`3ooml0odP00?nLg`0o
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0603oool2000001/0oooo00@00000
0?ooo`3oool000004`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo03H0oooo00<0
001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`008@3oool0103ooml0odP00?l0003o
_oln0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0603oool2000001/0oooo00@0
00000?ooo`3oool000004`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo03H0oooo
00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`008@3oool01@3ooml0odP00?l0
003o0000ogBo03d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0H0?ooo`800000
6`3oool010000000oooo0?ooo`00000C0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3o
ool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo001S0?ooo`800000
1@3oool00`000000oooo0?ooo`0H0?ooo`8000006`3oool010000000oooo0?ooo`00000C0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo03D0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0P0?ooo`80
00008P3oool006<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0H0?ooo`800000
6`3oool01@000000oooo0?ooo`3oool000000140oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0=@3oool00`0007l0
oooo0?ooo`0f0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool006<0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0H0?ooo`8000006`3oool01@000000oooo
0?ooo`3oool000000100oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`03
0000003oool0oooo01X0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0f0?ooo`03
0000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool006<0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0H0?ooo`8000006`3oool01@000000oooo0?ooo`3oool000000100
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0240
oooo00<000000?ooo`3oool0803oool006<0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`0H0?ooo`8000006`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00/0oooo
0P0000080?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0e0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo
0?ooo`0P0?ooo`00H`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01P0oooo0P00
000K0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0e0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?oo
o`0P0?ooo`00H`3oool2000000D0oooo00<000000?ooo`3oool0603oool2000001/0oooo00<00000
0?ooo`3oool0103oool2000000H0oooo0`00000;0?ooo`030000003oool0oooo01T0oooo00<00000
0?ooo`3oool06P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo03H0oooo00<0001o
0?ooo`3oool0803oool200000280oooo001S0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0603oool2000001/0oooo00<000000?ooo`3oool01P3oool6000000h0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo03D0oooo00<0001o0?oo
o`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo000R0?ooo`03
0?noM03o0000o`000140o`0000<0odRL0?ooo`3oool0703oool01@3ogi`0B0000000000007@0_ooo
00T0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0H0?ooo`8000006`3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`0e0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`00
0000oooo0?ooo`0P0?ooo`008P3oool00`3o_g@0o`000?l0000A0?l000030?m8W03oool0oooo01/0
oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`0G0?ooo`030000003oool0000001/0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0=@3oool0
0`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool0
0280oooo00<0okmd0?l0003o00004@3o00000`3oB9`0oooo0?ooo`0K0?ooo`030;md0000B9`0gooo
0080oooo00<0_g@0001d_`3oool01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
01H0oooo0P0000000`3oool000000?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo03D0oooo00<0001o0?oo
o`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo000R0?ooo`03
0?noM03oW=l0oooo0100oooo00<0oono0?m8W03oool06`3oool00`3ookl0M00004RLg`030?ooo`03
0=nLB00007@0_ooo00L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0C0?ooo`<0
00000P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3o
ool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3o
ool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo000R0?ooo`030?no
M03oW=l0oooo0100oooo00<0oono0?m8W03oool06`3oool00`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00L0oooo00<000000?ooo`3oool01@3ooolF000000D0oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0
8@3oool00`000000oooo0?ooo`0P0?ooo`008P3oool00`3o_g@0oicO0?ooo`0@0?ooo`030?oo_`3o
B9`0oooo01/0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooool00000;000
000R0?ooo`00<03oool00`3o_g@0o`000?m8W00P0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol01`3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo00@0oooo2000000=
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J
0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0Q
0?ooo`030000003oool0oooo0200oooo000`0?ooo`030?md003o0000o`1d0200oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo0200oooo00@000000?oo
o`3oool0oooo0`0000080?ooo`D00000203oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0
oooo03H0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`00:03oool00`3o
M7@0oooo0?ooo`050?ooo`030?oOW03o0000ogBo0200oooo00<0oono07@00018W=l00`3oool00`3O
W4P0001d0;ooo`070?ooo`030000003oool0oooo0200oooo00@000000?ooo`3oool00000403oool2
000000H0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0
oooo0240oooo00<000000?ooo`3oool0803oool002L0oooo00<0oooO0?m8B03ogol01P3oool00`3o
_g@0odRL0?ooo`0Q0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`00
0000oooo0?ooo`0P0?ooo`030000003oool0000001<0oooo0P0000040?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo
03D0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo
0200oooo000W0?ooo`050?oo_`3o0000o`000?l0003o_ol00`3oool00`3oM000ogBo0?ooo`0R0?oo
o`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo0200oooo00<0
00000?ooo`000000503oool2000000<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0I0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?oo
o`0f0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool002L0oooo00<0omnL
0?l0003o00001@3o00000`3o_ol0oooo0?ooo`0S0?ooo`050?oo_`1d000000000000M02oool02@3o
ool200000240oooo0P00000F0?ooo`050000003oool0oooo0?ooo`000000703oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0603ooolP000003H0oooo
00<0001o0?ooo`3oool0803oool200000280oooo000X0?ooo`030?noM03o0000o`0000H0o`0000<0
oicO0?ooo`3oool0;`3oool00`000000oooo0?ooo`0P0?ooo`8000005`3oool010000000oooo0?oo
o`00000L0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0D0?ooo`@000007@3oool00`0007l0oooo0?ooo`0600000300oooo00<0001o0?ooo`3oool0
8@3oool00`000000oooo0?ooo`0P0?ooo`00:P3oool00`3ooml0odP00?l000050?l000030?md_`3o
ool0oooo02h0oooo00<000000?ooo`3oool0803oool2000001L0oooo00@000000?ooo`3oool00000
703oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
4@3oool300000240oooo00<0001o0?ooo`3oool01P3oool5000002/0oooo00<0001o0?ooo`3oool0
8@3oool00`000000oooo0?ooo`0P0?ooo`00:@3oool00`3ooml0odP00?noo`020?ooo`050?oo_`3o
0000o`000?l0003o07@0<03oool00`000000oooo0?ooo`0P0?ooo`8000005`3oool010000000oooo
0?ooo`00000L0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000
oooo0?ooo`0?0?ooo`800000903oool00`0007l0oooo0?ooo`0;0?ooo`D000009P3oool00`0007l0
oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo000X0?ooo`030?oo_`3o04P0omoo00D0oooo
00<0oooO0?m8003oB9`0<03oool00`000000oooo0?ooo`0P0?ooo`800000603oool00`000000oooo
0000000L0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0<0?ooo`<000009P3oool00`0007l0oooo0?ooo`0@0?ooo`<000008`3oool00`0007l0oooo
0?ooo`0Q0?ooo`030000003oool0oooo0200oooo000W0?ooo`030?oog`3oB000o`1d00H0oooo00<0
okmd0?l0M03oool0<03oool00`000000oooo0?ooo`0P0?ooo`800000603oool00`000000oooo0000
000L0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool04`3oool300000200oooo00<0
001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`009`3oool0103ogi`0o`000?l0003o
_ol40?ooo`030?oOW03oB9`0oooo0340oooo00<000000?ooo`3oool0803oool2000001P0oooo00<0
00000?ooo`000000703oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`0X0?ooo`030000O`3oool0oooo01H0oooo0`00
000M0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool002L0oooo00@0oono
0?l0003o04P0omoo>03oool200000240oooo0P00000H0?ooo`030000003oool0000001`0oooo00<0
00000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool0:@3oool00`0007l0oooo0?ooo`0I0?ooo`8000006`3oool00`0007l0oooo0?oo
o`0P0?ooo`8000008P3oool006<0oooo00<000000?ooo`3oool0803oool2000001P0oooo00<00000
0?ooo`000000703oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`0Z0?ooo`030000O`3oool0oooo01/0oooo0P00000I
0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool006<0oooo00<000000?oo
o`3oool0803oool2000001T0oooo0P00000L0?ooo`030000003oool0oooo00L0oooo1P00000<0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0Z0?oo
o`030000O`3oool0oooo01d0oooo0P00000G0?ooo`030000O`3oool0oooo0240oooo00<000000?oo
o`3oool0803oool00280oooo00<0okmd0?nLg`3oool0403oool00`3ookl0odRL0?ooo`0[0?ooo`03
0000003oool0oooo0200oooo0P00000I0?ooo`800000703oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo00<0oooo0P00000;0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`0[0?ooo`030000O`3oool0oooo01l0oooo0P00000E0?ooo`03
0000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool00280oooo00<0okmd0?nLg`3oool0
403oool00`3ookl0odRL0?ooo`0[0?ooo`030000003oool0oooo0200oooo0P00000I0?ooo`800000
703oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
;03oool00`0007l0oooo0?ooo`0Q0?ooo`8000004`3oool00`0007l0oooo0?ooo`0Q0?ooo`030000
003oool0oooo0200oooo000R0?ooo`030?noM03oW=l0oooo0100oooo00<0oono0?m8W03oool0:`3o
ool00`000000oooo0?ooo`0P0?ooo`8000006@3oool2000001`0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool08`3o
ool00`000000oooo0?ooo`0@0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0803o
ool00280oooo00<0okmd0?l0003o00004@3o00000`3oB9`0oooo0?ooo`0Z0?ooo`8000008@3oool2
000001T0oooo0P00000L0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool2
000000P0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo02d0oooo00<0001o0?ooo`3oool0903oool200000100oooo00<0001o0?ooo`3oool0803oool2
00000280oooo000R0?ooo`030?noM03o0000o`000140o`0000<0odRL0?ooo`3oool0:P3oool00`00
0000oooo0?ooo`0P0?ooo`8000006@3oool2000001`0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09P3oool00`00
0000oooo0?ooo`0=0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool00280
oooo00<0okmd0?l0003o00004@3o00000`3oB9`0oooo0?ooo`0Z0?ooo`030000003oool0oooo0200
oooo0P00000I0?ooo`800000703oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo00`0
oooo00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`00H`3oool00`000000oooo
0?ooo`0P0?ooo`8000006@3oool2000001`0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo02h0oooo00<0001o0?ooo`3oool0:03oool00`000000oooo
0?ooo`0;0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool006<0oooo00<0
00000?ooo`3oool0803oool2000001T0oooo0P00000L0?ooo`050000003oool0oooo0?ooo`000000
3`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0^0?ooo`030000O`3oool0oooo02T0oooo00<000000?ooo`3oool0
2P3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0200oooo001S0?ooo`030000003o
ool0oooo0200oooo0P00000I0?ooo`800000703oool01@000000oooo0?ooo`3oool000000100oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0Z0?ooo`030000003oool0oooo00T0oooo
00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`00H`3oool00`000000oooo0?oo
o`0P0?ooo`8000006@3oool2000001`0oooo00D000000?ooo`3oool0oooo0000000@0?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo02h0oooo00<0001o0?ooo`3oool0:`3oool00`000000oooo0?ooo`080?ooo`030000
O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool006<0oooo0P00000Q0?ooo`8000006@3o
ool2000001`0oooo00D000000?ooo`3oool0oooo0000000A0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo02l0oooo
00<0001o0?ooo`3oool0;03oool00`000000oooo0?ooo`070?ooo`030000O`3oool0oooo0200oooo
0P00000R0?ooo`00H`3oool00`000000oooo0?ooo`0P0?ooo`8000006@3oool2000001`0oooo00D0
00000?ooo`3oool0oooo0000000A0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo02l0oooo00<0001o0?ooo`3oool0
;@3oool00`000000oooo0?ooo`060?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0
803oool005<0oooo00<0oooO09a8000000000P0000000`1d_ol0oooo0?ooo`080?ooo`030000003o
ool0oooo0200oooo0P00000I0?ooo`800000703oool01@000000oooo0?ooo`3oool000000140oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool0;`3oool00`0007l0oooo0?ooo`0^0?ooo`030000003oool0oooo00D0oooo
00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`00D`3oool00`3ooml0W4Qd0;oo
o`020?ooo`030;md001d_ol0oooo00P0oooo00<000000?ooo`3oool0803oool2000001T0oooo0P00
000L0?ooo`050000003oool0oooo0?ooo`0000004@3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0`0?ooo`030000
O`3oool0oooo02l0oooo00<000000?ooo`3oool0103oool00`0007l0oooo0?ooo`0Q0?ooo`030000
003oool0oooo0200oooo001H0?ooo`030?oog`2LB7@0_ooo00P0oooo00<000000?ooo`3oool0803o
ool2000001T0oooo0P00000L0?ooo`050000003oool0oooo0?ooo`0000004@3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0`0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0103oool00`0007l0oooo
0?ooo`0Q0?ooo`030000003oool0oooo0200oooo001I0?ooo`030;md001d_ol0oooo00L0oooo00<0
00000?ooo`3oool0803oool2000001T0oooo0P00000L0?ooo`040000003oool0oooo00000180oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0<03oool00`0007l0oooo0?ooo`0`0?ooo`030000003oool0oooo00<0oooo
00<0001o0?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`00F03oool00`3ooml0W4P004RL
g`080?ooo`030000003oool0oooo0200oooo0P00000I0?ooo`800000703oool010000000oooo0?oo
o`00000B0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo0300oooo00<0001o0?ooo`3oool0<@3oool00`000000oooo
0?ooo`020?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool0803oool005P0oooo00<0
gia80000001d_ol0203oool200000240001o0P00000I0000O`800000700007l010000000001o0000
O`00000C0000O`030000000007l0001o0080001o00<000000000O`0007l06P0007l00`000000001o
0000O`020000O`030000000007l0001o06D0001o00<000000000O`0007l08`0007l300000280oooo
001D0?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`080?ooo`030000003oool0oooo0200
oooo0P00000I0?ooo`800000703oool010000000oooo0?ooo`00000C0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
0300oooo00<0001o0?ooo`3oool0<`3oool010000000oooo0?ooo`0007lS0?ooo`030000003oool0
oooo0200oooo001D0?ooo`060=nLB000000000000000000004P0W=oo2@3oool00`000000oooo0?oo
o`0P0?ooo`8000006@3oool2000001`0oooo00@000000?ooo`3oool000004`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0`0?ooo`030000O`3oool0oooo03<0oooo00@000000?ooo`3oool0001o8`3oool00`000000
oooo0?ooo`0P0?ooo`00E03oool0103ookl0M00000000018W=l;0?ooo`030000003oool0oooo0200
oooo0P00000I0?ooo`800000703oool010000000oooo0?ooo`00000C0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
0300oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0000O`0S0?ooo`030000003oool0oooo
0200oooo001E0?ooo`030;md001d_ol0oooo00/0oooo00<000000?ooo`3oool0803oool2000001T0
oooo0P00000L0?ooo`040000003oool0oooo000001<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0<03oool00`00
07l0oooo0?ooo`0e0?ooo`030000000007l0oooo0280oooo00<000000?ooo`3oool0803oool005D0
oooo00H0omnL04P00000000000000000B02Lgol80?ooo`030000003oool0oooo0200oooo0P00000J
0?ooo`030000003oool0oooo01X0oooo00@000000?ooo`3oool00000503oool01@000000oooo0?oo
o`3oool0000001`0oooo00D000000?ooo`3oool0oooo0000000c0?ooo`030000O`3oool0oooo03D0
oooo00<000000000O`3oool08P3oool00`000000oooo0?ooo`0P0?ooo`00EP3oool01@2oM0000000
000000000000B9cO00P0oooo0P00000Q0?ooo`8000006P3oool00`000000oooo0?ooo`0J0?ooo`04
0000003oool0oooo000001@0oooo00D000000?ooo`3oool0oooo0000000L0?ooo`050000003oool0
oooo0?ooo`000000<`3oool00`0007l0oooo0?ooo`0f0?ooo`030000003oool0oooo0200oooo0P00
000R0?ooo`00H`3oool00`000000oooo0?ooo`0P0?ooo`8000006P3oool00`000000oooo0?ooo`0J
0?ooo`030000003oool0000001D0oooo00D000000?ooo`3oool0oooo0000000L0?ooo`050000003o
ool0oooo0?ooo`000000<`3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`000000oooo0240oooo
00<000000?ooo`3oool0803oool006<0oooo00<000000?ooo`3oool0803oool2000001X0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0000000E0?ooo`050000003oool0oooo0?ooo`000000
703oool01@000000oooo0?ooo`3oool0000003<0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0
00000?ooo`0Q0?ooo`030000003oool0oooo0200oooo001S0?ooo`030000003oool0oooo0200oooo
0P00000J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`0000005@3oool01@000000oooo
0?ooo`3oool0000001`0oooo00D000000?ooo`3oool0oooo0000000c0?ooo`030000O`3oool0oooo
03H0oooo00<0001o0?ooo`0000008@3oool00`000000oooo0?ooo`0P0?ooo`00H`3oool00`000000
oooo0?ooo`0P0?ooo`8000006P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0000001D0
oooo00D000000?ooo`3oool0oooo0000000L0?ooo`050000003oool0oooo0?ooo`000000<`3oool0
0`0007l0oooo0?ooo`0f0?ooo`030000O`3oool000000240oooo00<000000?ooo`3oool0803oool0
06<0oooo00<000000?ooo`3oool0803oool2000001X0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0000000E0?ooo`050000003oool0oooo0?ooo`000000703oool01@000000oooo0?ooo`3o
ool0000003<0oooo00<0001o0?ooo`3oool0=P3oool0100007l0oooo0?ooo`00000P0?ooo`030000
003oool0oooo0200oooo001S0?ooo`030000003oool0oooo0200oooo0P00000J0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`0000005P3oool010000000oooo0?ooo`00000L0?ooo`050000
003oool0oooo0?ooo`000000<`3oool00`0007l0oooo0?ooo`0f0?ooo`040000O`3oool0oooo0000
0200oooo00<000000?ooo`3oool0803oool006<0oooo0P00000Q0?ooo`8000006P3oool00`000000
oooo0?ooo`0J0?ooo`030000003oool0000001H0oooo00@000000?ooo`3oool00000703oool01@00
0000oooo0?ooo`3oool0000003<0oooo00<0001o0?ooo`3oool0=P3oool01@0007l0oooo0?ooo`3o
ool0000001h0oooo0P00000R0?ooo`00H`3oool00`000000oooo0?ooo`0P0?ooo`8000006P3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0000001H0oooo00@000000?ooo`3oool00000703o
ool01@000000oooo0?ooo`3oool0000003<0oooo00<0001o0?ooo`3oool0=P3oool01@0007l0oooo
0?ooo`3oool0000001l0oooo00<000000?ooo`3oool0803oool006<0oooo00<000000?ooo`3oool0
803oool2000001X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000F0?ooo`040000
003oool0oooo000001`0oooo00@000000?ooo`3oool00000=03oool00`0007l0oooo0?ooo`0f0?oo
o`030000O`3oool0oooo0080oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0P0?oo
o`00H`3oool00`000000oooo0?ooo`0P0?ooo`8000006P3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0000001H0oooo00@000000?ooo`3oool00000703oool010000000oooo0?ooo`00000d
0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool00P3oool00`000000oooo0?ooo`0L
0?ooo`030000003oool0oooo0200oooo001S0?ooo`030000003oool0oooo0200oooo0P00000J0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`0000005P3oool010000000oooo0?ooo`00000L
0?ooo`040000003oool0oooo000003@0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?oo
o`020?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0803oool006<0oooo00<00000
0?ooo`3oool0803oool2000001X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000F
0?ooo`040000003oool0oooo000001`0oooo00@000000?ooo`3oool00000=03oool00`0007l0oooo
0?ooo`0f0?ooo`030000O`3oool0oooo00<0oooo00<000000?ooo`3oool06`3oool00`000000oooo
0?ooo`0P0?ooo`00H`3oool200000240oooo0P00000J0?ooo`030000003oool0oooo01X0oooo00<0
00000?ooo`0000005P3oool010000000oooo0?ooo`00000L0?ooo`040000003oool0oooo000003@0
oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`030?ooo`030000003oool0oooo01X0
oooo0P00000R0?ooo`00H`3oool00`000000oooo0?ooo`0P0?ooo`8000006P3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0000001H0oooo00@000000?ooo`3oool00000703oool010000000
oooo0?ooo`00000d0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0103oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo0200oooo001S0?ooo`030000003oool0oooo0200
oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0000001H0
oooo00@000000?ooo`3oool00000703oool010000000oooo0?ooo`00000d0?ooo`030000O`3oool0
oooo03H0oooo00<0001o0?ooo`3oool0103oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo0200oooo001S0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool06@3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0000001H0oooo00@000000?ooo`3oool00000703oool0
10000000oooo0?ooo`00000d0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool01@3o
ool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0200oooo001S0?ooo`030000003oool0
oooo0200oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
000001H0oooo00@000000?ooo`3oool00000703oool010000000oooo0?ooo`00000d0?ooo`030000
O`3oool0oooo03H0oooo00<0001o0?ooo`3oool01@3oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo0200oooo001S0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool06@3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0000001H0oooo00@000000?ooo`3oool00000
703oool010000000oooo0?ooo`00000d0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3o
ool01@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0200oooo001S0?ooo`030000
003oool0oooo0200oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`030000
003oool0000001H0oooo00@000000?ooo`3oool00000703oool010000000oooo0?ooo`00000d0?oo
o`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool01P3oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo0200oooo001S0?ooo`8000008@3oool00`000000oooo0?ooo`0I0?ooo`03
0000003oool0oooo01X0oooo00<000000?ooo`0000005P3oool010000000oooo0?ooo`00000L0?oo
o`040000003oool0oooo000003@0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`06
0?ooo`030000003oool0oooo01L0oooo0P00000R0?ooo`00C03oool01P2oM0000000000000000000
04RL0=ooo`<0oooo00D0omnL04P000000000001d0;ooo`090?ooo`030000003oool0oooo0200oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0000001H0oooo
00@000000?ooo`3oool00000703oool010000000oooo0?ooo`00000d0?ooo`030000O`3oool0oooo
03H0oooo00<0001o0?ooo`3oool01P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo
0200oooo001=0?ooo`030?oOW0180000M;oo00@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0
_ooo00P0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`0000005P3oool010000000oooo0?ooo`00000L0?ooo`040000003o
ool0oooo000003@0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`060?ooo`030000
003oool0oooo01P0oooo00<000000?ooo`3oool0803oool004d0oooo00<0omnL04P0001d_ol0103o
ool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<000000?ooo`3oool0803o
ool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00000G0?ooo`040000003o
ool0oooo000001`0oooo00@000000?ooo`3oool00000=03oool00`0007l0oooo0?ooo`0f0?ooo`03
0000O`3oool0oooo00L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0P0?ooo`00
C@3oool00`3ogi`0B00007Boo`030?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0
1`3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0
6P3oool2000001L0oooo00@000000?ooo`3oool00000703oool010000000oooo0?ooo`00000d0?oo
o`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool01`3oool00`000000oooo0?ooo`0G0?oo
o`030000003oool0oooo0200oooo001=0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo0200oooo00<000000?ooo`3o
ool06@3oool00`000000oooo0?ooo`0J0?ooo`8000005`3oool010000000oooo0?ooo`00000L0?oo
o`030000003oool0000003D0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`070?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool0803oool004d0oooo00<0omnL04P0001d
_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo0P00000Q0000O`03
0000000007l0001o01T0001o00<000000000O`0007l06P0007l2000001L0001o00@000000000O`00
07l00000700007l00`000000001o0000001i0000O`030000000007l0001o01@0001o0`00000R0?oo
o`00C@3oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol01`3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3o
ool06P3oool2000001P0oooo00<000000?ooo`000000703oool00`000000oooo0000000e0?ooo`03
0000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0203oool00`000000oooo0?ooo`0F0?ooo`03
0000003oool0oooo0200oooo001=0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol0
0`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`0J0?ooo`800000603oool00`000000oooo0000000L0?ooo`030000
003oool0000003D0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`080?ooo`030000
003oool0oooo01H0oooo00<000000?ooo`3oool0803oool004d0oooo00<0omnL04P0001d_ol00`3o
ool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00L0oooo00<000000?ooo`3oool0803o
ool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00000H0?ooo`030000003o
ool0000001`0oooo00<000000?ooo`000000=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3o
ool0oooo00T0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0P0?ooo`00C@3oool0
0`3ogi`0B00007Boo`040?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool0
0`000000oooo0?ooo`0P0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool2
000001P0oooo00<000000?ooo`000000703oool00`000000oooo0000000e0?ooo`030000O`3oool0
oooo03H0oooo00<0001o0?ooo`3oool02@3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo0200oooo001;0?ooo`050?oog`2LB00000000000001d_ol0103oool01`3ookl0M00007Boo`3o
ool0okmd0000M02oool0203oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool06P3oool2000001P0oooo00<000000?ooo`000000703oool00`000000oooo
0000000e0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool02@3oool00`000000oooo
0?ooo`0E0?ooo`030000003oool0oooo0200oooo001=0?ooo`030?noM0000000M;oo00D0oooo00D0
oono07@000000000001d0;ooo`090?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`0J0?ooo`800000603oool00`000000oooo0000000L0?ooo`030000
003oool0000003D0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`090?ooo`030000
003oool0oooo01D0oooo00<000000?ooo`3oool0803oool006<0oooo0P00000Q0?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool06P3oool2000001P0oooo00<000000?ooo`000000703o
ool00`000000oooo0000000e0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool02P3o
ool00`000000oooo0?ooo`0C0?ooo`8000008P3oool006<0oooo00<000000?ooo`3oool0803oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00000H0?ooo`030000003oool0
000001`0oooo00<000000?ooo`000000=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0
oooo00X0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0P0?ooo`00H`3oool00`00
0000oooo0?ooo`0P0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool20000
01P0oooo00<000000?ooo`000000703oool00`000000oooo0000000e0?ooo`030000O`3oool0oooo
03H0oooo00<0001o0?ooo`3oool02P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo
0200oooo001S0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool06@3oool00`000000
oooo0?ooo`0J0?ooo`800000603oool00`000000oooo0000000L0?ooo`030000003oool0000003D0
oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0;0?ooo`030000003oool0oooo01<0
oooo00<000000?ooo`3oool0803oool006<0oooo00<000000?ooo`3oool0803oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00000H0?ooo`030000003oool0000001`0oooo
00<000000?ooo`000000=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo00/0oooo
00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0P0?ooo`00H`3oool00`000000oooo0?oo
o`0P0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool2000001P0oooo00<0
00000?ooo`000000703oool00`000000oooo0000000e0?ooo`030000O`3oool0oooo03H0oooo00<0
001o0?ooo`3oool02`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0200oooo001S
0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J
0?ooo`800000603oool00`000000oooo0000000L0?ooo`030000003oool0000003D0oooo00<0001o
0?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0;0?ooo`030000003oool0oooo01<0oooo00<00000
0?ooo`3oool0803oool006<0oooo0P00000Q0?ooo`030000003oool0oooo01T0oooo00<000000?oo
o`3oool06P3oool2000001P0oooo00<000000?ooo`000000703oool00`000000oooo0000000e0?oo
o`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0A0?oo
o`8000008P3oool006<0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0I0?ooo`03
0000003oool0oooo01X0oooo0P00000H0?ooo`030000003oool0000001`0oooo00<000000?ooo`00
0000=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3o
ool04P3oool00`000000oooo0?ooo`0P0?ooo`00H`3oool00`000000oooo0?ooo`0P0?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool2000001P0oooo00<000000?ooo`000000
703oool00`000000oooo0000000e0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool0
303oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0200oooo001S0?ooo`030000003o
ool0oooo0200oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`800000603o
ool00`000000oooo0000000L0?ooo`030000003oool0000003D0oooo00<0001o0?ooo`3oool0=P3o
ool00`0007l0oooo0?ooo`0=0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0803o
ool006<0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo01X0oooo0P00000H0?ooo`030000003oool0000001`0oooo00<000000?ooo`000000=@3oool0
0`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo00d0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0P0?ooo`00H`3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`3oool06P3oool2000001P0oooo00<000000?ooo`000000703oool00`00
0000oooo0000000e0?ooo`030000O`3oool0oooo03H0oooo00<0001o0?ooo`3oool03@3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo0200oooo001S0?ooo`8000008@3oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00000H0?ooo`030000003oool0000001`0
oooo00<000000?ooo`000000=@3oool00`0007l0oooo0?ooo`0f0?ooo`030000O`3oool0oooo00h0
oooo00<000000?ooo`3oool03`3oool200000280oooo001S0?ooo`030000003oool0oooo0200oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0J0?ooo`800000603oool00`000000oooo
0000000L0?ooo`030000003oool0000003D0oooo00<0001o0?ooo`3oool0=P3oool00`0007l0oooo
0?ooo`0>0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0803oool006<0oooo00<0
00000?ooo`3oool0803oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo0P00
000H0?ooo`030000003oool0000001`0oooo00<000000?ooo`000000=@3oool00`0007l0oooo0?oo
o`0f0?ooo`030000O`3oool0oooo00h0oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`0P0?ooo`00H`3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo01T0oooo00<00000
0?ooo`3oool06P3oool2000001P0oooo00<000000?ooo`000000703oool00`000000oooo0000000e
0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0oooo0200oooo001S0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?oo
o`3oool02`3oool2000000d0oooo00<000000?ooo`3oool02`3oool2000000`0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0000000>0?ooo`030000003oool0oooo00/0oooo00<000000?oo
o`0000002`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00D000000?oo
o`3oool0oooo0000000:0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0803oool0
06<0ooooo`00000/00000280oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o
0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o
0?oook40oooo002?0?ooo`030?oof01dOgl007mo00@007mo00<007n^0;ooo`3oool01`3oool00`3o
on/0W9Mo001oO`02001oO`03002Ga03Oool0glBG00<007mo00@0M=So0?ooo`3oon/0W9Mo0P00Ogl0
0`00Ojh0_ooo0?ooo`020?ooo`050?o[a018Ogl007mo001oO`00[]P0103oool01@3Oa9L007mo001o
O`00Ogl0M=So0080oooo00H0ooo[09bGO`00Ogl007mo001o[P2oool30?ooo`050?o[a018Ogl007mo
001oO`00[]P0103oool01@3Oa9L007mo001oO`00Ogl0M=So00@0oooo00<0ooo[09bGO`00Ogl00P00
Ogl00`00Ul@0gooo0=o4U`03001oO`0307CHo`3oool0oooo00D0oooo00H0ooo[09bGO`00Ogl007mo
001oU`2Ljol<0?ooo`080?oof01dOgl007mo001oO`00OiL0W>_o0?ooo`2o[Wl2001oO`06002Ga03O
ool0oooo0?ooo`3oomP0M7mo0P00Ogl00`00Ul@0gooo0?ooo`020?ooo`050?oH[P00Ogl007mo001o
O`18a>/01P3oool00`3ojl@0B7mo001oO`02001oO`03001oU`2Ljol0oooo00H0oooo00P0_jio001o
O`00Ogl007mo001oU`2Ljol0on_404QoO`8007mo00<00:kH0?ooo`3oool0103oool00`3oon/0W9Mo
001oO`03001oO`03002^f03oool0oooo00<0oooo00H0oooH07AoO`00Ogl007mo002Ga03Oool30?oo
o`050?oH[P00Ogl007mo001oO`18a>/0C`3oool008d0oooo00<0ooo[09bGO`00Ogl02@00Ogl00`00
[]P0oooo0?ooo`030?ooo`030?oof01dOgl007mo00P007mo00D007nG09c[o`3oool0oooo0;n^O`02
001oO`03001oU`2Ljol0oooo0080oooo00H0oooH07AoO`00Ogl007mo002Ga03Oool30?ooo`050?oH
[P00Ogl007mo001oO`18a>/00`3oool01@2o[Wl007mo001oO`00OiL0W>_o00<0oooo00H0oooH07Ao
O`00Ogl007mo002Ga03Oool30?ooo`050?oH[P00Ogl007mo001oO`18a>/00`3oool00`3oomP0M7mo
001oO`08001oO`03001oU`2Ljol0oooo00L0oooo00D0_jio001oO`00Ogl007mo07CHo`0;0?ooo`03
0?oof01dOgl007mo00@007mo00H00:kH0?oH[P00Ogl007mo001oU`2Ljol20?ooo`060?ooj`2LUgl0
07mo001oO`00Ojh0_ooo0`3oool01@3oomP0M7mo001oO`00Ogl00:kH00D0oooo00<0_jio001oO`00
Ogl01@00Ogl00`00Ul@0gooo0?ooo`030?ooo`030;n^O`00Ogl007mo00<007mo00<0B<C[0?ooj`2L
Ugl00P00Ogl00`00Ul@0gooo0?ooo`030?ooo`030=o4U`00Ogl007mo00H007mo00H0B<C[0?ooo`3o
ool0oooo0?ooj`2LUgl2001oO`03001o[P2oool0oooo0080oooo00D0oooH07AoO`00Ogl007mo002^
f01?0?ooo`00S03oool00`3oon/0W9Mo001oO`0<001oO`0507CHo`3oool0oooo0?ooo`2o[Wl02P00
Ogl01@00OiL0W>_o0?ooo`3oool0omR^00<007mo00<0M=So0?ooo`3oool00P3oool01@2o[Wl007mo
001oO`00Ojh0_ooo00<0oooo00D0oooH07AoO`00Ogl007mo002^f0030?ooo`050?oH[P00Ogl007mo
001oO`1df?l0103oool01@2o[Wl007mo001oO`00Ojh0_ooo00<0oooo00D0oooH07AoO`00Ogl007mo
002^f0030?ooo`030;n^O`00Ogl007mo00P007mo00<007nG09c[o`3oool01`3oool01@3Oa9L007mo
001oO`00Ogl0B<C[00/0oooo00<0omR^001oO`00Ogl01@00Ogl01@00UiL0B7mo001oO`00Ogl0M=So
00<0oooo00D0_jio001oO`00Ogl007nG09c[o`030?ooo`060?ooj`2LUgl007mo001oO`00Ojh0_ooo
0`3oool00`3Oa9L007mo001oO`07001oO`06002Ga03Oool0oooo0?ooo`3oomP0M7mo1@00Ogl00`00
[]P0oooo0;n^O`02001oO`03001o[P2oool0oooo0080oooo00<0omR^001oO`00Ogl02000Ogl01@18
a>/0oooo0?ooo`3oool0_jio008007mo00<007nG09c[o`3oool00P3oool01P3oon/0W9Mo001oO`00
Ogl007n^0;ooodh0oooo002<0?ooo`030=o4U`00Ogl007mo00`007mo00D009O40=ooo`3oool0ooo[
09bGO`03001oO`0407CHo`3oool0oooo0;n^O`@007mo00D0M=So0?ooo`3oool0on_404QoO`02001o
O`0304S4j`3oool0oooo0080oooo00D0glBG001oO`00Ogl007mo07CHo`030?ooo`060?ooj`2LUgl0
07mo001oO`00Ojh0_ooo0P3oool01@3ojl@0B7mo001oO`00Ogl0B<C[00@0oooo00D0glBG001oO`00
Ogl007mo07CHo`030?ooo`090?ooj`2LUgl007mo001oO`00Ojh0_ooo0?ooo`3oon/0W9Mo00<007mo
00@0M=So0?ooo`3oool0_jio1000Ogl00`1df?l0oooo0?ooo`060?ooo`050?oH[P00Ogl007mo001o
O`00[]P02`3oool0203of:h007mo001oO`00Ogl009O40=ooo`3oool0_jio1000Ogl01@18a>/0oooo
0?ooo`3oool0glBG00<007mo00<0M=So0?ooo`3oool00P3oool01@2o[Wl007mo001oO`00OiL0W>_o
0080oooo00T0ooo[09bGO`00Ogl007mo001oU`2Ljol0oooo0?oof01dOgl00`00Ogl01@18a>/0oooo
0?ooo`3oon/0W9Mo008007mo00<007n^0;ooo`3oool00P3oool01@3Oa9L007mo001oO`00Ogl0M=So
0080oooo00<0ooo[09bGO`00Ogl00P00Ogl01018a>/0oooo0?ooo`3of:h3001oO`05001oU`2Ljol0
oooo0?ooo`3Oa9L00`00Ogl00`1df?l0oooo0?ooo`020?ooo`050;n^O`00Ogl007mo001oU`2Ljol0
CP3oool008/0oooo00<0oooH07AoO`00Ogl00P00Ogl00`00Ojh0_ooo0?ooo`040?ooo`090;n^O`00
Ogl007mo001oO`00OiL0W>_o0?ooo`3oon/0W9Mo008007mo00<009O40=ooo`3oool00P3oool01@2o
[Wl007mo001oO`00Ogl0B<C[0080oooo00D0oooH07AoO`00Ogl007mo002^f0040?ooo`050?oH[P00
Ogl007mo001oO`18a>/0103oool01@2o[Wl007mo001oO`00OiL0W>_o0080oooo00D0oooH07AoO`00
Ogl007mo002^f0040?ooo`050?oH[P00Ogl007mo001oO`18a>/0103oool0202o[Wl007mo001oO`00
OiL0W>_o0?ooo`3oon/0W9Mo0P00Ogl00`00Ul@0gooo0?ooo`020?ooo`050;n^O`00Ogl007mo001o
O`18a>/0203oool01P3ojl@0B7mo001oO`00Ogl009O40=ooo`X0oooo00D0omR^001oO`00Ogl007mo
07CHo`030?ooo`050=o4U`00Ogl007mo001oO`00[]P00`3oool01@3of:h007mo001oO`00Ogl0B<C[
00@0oooo00D0glBG001oO`00Ogl007mo07CHo`020?ooo`050?o[a018Ogl007mo001oO`18a>/00`3o
ool01P3oomP0M7mo001oO`00Ogl007n^0;ooo`80oooo00D0_jio001oO`00Ogl007nG09c[o`030?oo
o`050?oH[P00Ogl007mo001oO`18a>/00P3oool01@3ojl@0B7mo001oO`00Ogl00:kH00@0oooo00D0
glBG001oO`00Ogl007mo002^f0020?ooo`050?oH[P00Ogl007mo001oO`18a>/0103oool01@3Oa9L0
07mo001oO`00Ogl0M=So04h0oooo002;0?ooo`060?oH[P00Ogl007mo001oO`00Ul@0gooo1`3oool0
1@3of:h007mo001oO`00Ogl0M=So0080oooo00D0_jio001oO`00Ogl007nG09c[o`030?ooo`050?o[
a018Ogl007mo001oO`00[]P00P3oool01P3oon/0W9Mo001oO`00Ogl009O40=ooo`<0oooo00D0on_4
04QoO`00Ogl007mo002^f0040?ooo`050=o4U`00Ogl007mo001oO`1df?l00P3oool01P3oon/0W9Mo
001oO`00Ogl009O40=ooo`<0oooo00D0on_404QoO`00Ogl007mo002^f0040?ooo`050=o4U`00Ogl0
07mo001oO`1df?l00P3oool01@2o[Wl007mo001oO`00OiL0W>_o00<0oooo00D0on_404QoO`00Ogl0
07mo002^f0080?ooo`060?ooj`2LUgl007mo001oO`00Ojh0_ooo2P3oool01@3ojl@0B7mo001oO`00
Ogl0B<C[00<0oooo00H0oooH07AoO`00Ogl007mo002Ga03Oool20?ooo`050?oof01dOgl007mo001o
O`00[]P0103oool01@3ojl@0B7mo001oO`00Ogl0B<C[0080oooo00D0omR^001oO`00Ogl007mo04S4
j`0;0?ooo`050?oH[P00Ogl007mo001oO`1df?l00`3oool01@3ojl@0B7mo001oO`00Ogl00:kH0080
oooo00D0omR^001oO`00Ogl007mo07CHo`040?ooo`090?ooj`2LUgl007mo001oO`00OiL0W>_o0?oo
o`3oomP0M7mo008007mo00<00:kH0?ooo`3oool00P3oool01@3ojl@0B7mo001oO`00Ogl0B<C[04h0
oooo002;0?ooo`050=o4U`00Ogl007mo001oO`1df?l0203oool01@3ojl@0B7mo001oO`00Ogl0B<C[
0080oooo00<0on_404QoO`00Ogl00P00Ogl01@00OiL0W>_o0?ooo`3oomP0M7mo008007mo00D009O4
0=ooo`3oool0oooo0;n^O`02001oO`03001oU`2Ljol0oooo0080oooo00H0oooH07AoO`00Ogl007mo
002Ga03Oool30?ooo`050?oH[P00Ogl007mo001oO`18a>/00`3oool01@2o[Wl007mo001oO`00OiL0
W>_o00<0oooo00H0oooH07AoO`00Ogl007mo002Ga03Oool30?ooo`050?oH[P00Ogl007mo001oO`18
a>/00P3oool00`3ojl@0B7mo001oO`02001oO`05001oU`2Ljol0oooo0?oof01dOgl00P00Ogl00`00
Ul@0gooo0?ooo`070?ooo`050;n^O`00Ogl007mo001oO`1df?l02P3oool01P3oon/0W9Mo001oO`00
Ogl009O40=ooo`<0oooo00D0_jio001oO`00Ogl007nG09c[o`020?ooo`060?ooj`2LUgl007mo001o
O`00Ul@0gooo0`3oool01@3oomP0M7mo001oO`00Ogl00:kH0080oooo00D0omR^001oO`00Ogl007mo
04S4j`0;0?ooo`050?o[a018Ogl007mo001oO`00[]P00`3oool0203oomP0M7mo001oO`00Ogl009O4
0=ooo`3oool0glBG0`00Ogl00`1df?l0oooo0?ooo`030?ooo`080=o4U`00Ogl007mo001oO`1df?l0
oooo0?ooj`2LUgl2001oO`03002Ga03Oool0oooo0080oooo00D0oooH07AoO`00Ogl007mo002^f01>
0?ooo`00R`3oool01@2o[Wl007mo001oO`00OiL0W>_o00<0oooo00<0oooH07AoO`00Ogl01P00Ogl0
1P00[]P0oooo0?ooo`3oool0oooH07AoO`T007mo00D007nG09c[o`3oool0oooo0=o4U`03001oO`06
07CHo`3oool0oooo0?ooo`3oon/0W9Mo0P00Ogl00`00OiL0W>_o0?ooo`020?ooo`060?oof01dOgl0
07mo001oO`00Ul@0gooo0P3oool01@3Oa9L007mo001oO`00Ogl0M=So00<0oooo00H0ooo[09bGO`00
Ogl007mo001oU`2Ljol30?ooo`060?oof01dOgl007mo001oO`00Ul@0gooo0P3oool00`3oomP0M7mo
001oO`08001oO`03001oU`2Ljol0oooo00L0oooo00D0glBG001oO`00Ogl007mo04S4j`0;0?ooo`05
0;n^O`00Ogl007mo001o[P2oool00`3oool01@3Oa9L007mo001oO`00Ogl0M=So00<0oooo00D0_jio
001oO`00Ogl007nG09c[o`030?ooo`090?ooj`2LUgl007mo001oO`00Ojh0_ooo0?ooo`3ojl@0B7mo
008007mo00<00:kH0?ooo`3oool02@3oool01P3oomP0M7mo001oO`00Ogl009O40=ooo`<0oooo00L0
_jio001oO`00Ogl007n^0;ooo`3oool0omR^00<007mo00<0B<C[0?ooo`3oool00`3oool01@3of:h0
07mo001oO`00Ogl0B<C[0080oooo00D0_jio001oO`00Ogl007nG09c[o`030?ooo`060?ooj`2LUgl0
07mo001oO`00Ojh0_oooC@3oool008/0oooo00D0_jio001oO`00Ogl007nG09c[o`030?ooo`030?oo
j`2LUgl007mo00H007mo00<007n^0;ooo`3oool01P3oool00`2o[Wl007mo001oO`04001oO`0407CH
o`3oool0oooo0?oH[P<007mo00<00:kH0?ooo`3oool00P3oool01@3Oa9L007mo001oO`00Ogl0M=So
00<0oooo00H0ooo[09bGO`00Ogl007mo001o[P2oool20?ooo`050?oH[P00Ogl007mo001oO`00[]P0
103oool01@3Oa9L007mo001oO`00Ogl0M=So00<0oooo00H0ooo[09bGO`00Ogl007mo001o[P2oool7
0?ooo`030;n^O`00Ogl007mo00@007mo00<0M=So0?ooo`3oool01P3oool00`3of:h007mo001oO`0:
001oO`04002^f03oool0oooo0=o4U`8007mo00<007nG09c[o`3oool00P3oool01@3of:h007mo001o
O`00Ogl0B<C[00<0oooo00D0glBG001oO`00Ogl007mo07CHo`040?ooo`080;n^O`00Ogl007mo001o
U`2Ljol0oooo0?oof01dOgl2001oO`03002Ga03Oool0oooo00X0oooo00D0_jio001oO`00Ogl007n^
0;ooo`030?ooo`080=o4U`00Ogl007mo001oU`2Ljol0oooo0?o[a018Ogl2001oO`03002^f03oool0
oooo00<0oooo00D0on_404QoO`00Ogl007mo04S4j`020?ooo`050=o4U`00Ogl007mo001oO`1df?l0
103oool01@2o[Wl007mo001oO`00OiL0W>_o04d0oooo002;0?ooo`050=o4U`00Ogl007mo001oU`2L
jol0103oool00`2o[Wl007mo001oO`05001oO`03001oU`2Ljol0oooo00T0oooo00D0on_404QoO`00
Ogl007mo04S4j`020?ooo`060?o[a018Ogl007mo001oO`00Ojh0_ooo0`3oool01P3of:h007mo001o
O`00Ogl009O40=ooo`<0oooo00D0_jio001oO`00Ogl007nG09c[o`020?ooo`060?o[a018Ogl007mo
001oO`00Ojh0_ooo0`3oool01P3of:h007mo001oO`00Ogl009O40=ooo`<0oooo00D0_jio001oO`00
Ogl007nG09c[o`0:0?ooo`050?o[a018Ogl007mo001oO`18a>/0203oool00`3ojl@0B7mo001oO`0:
001oO`05002Ga03Oool0oooo0?o[a018Ogl00P00Ogl01P18a>/0oooo0?ooo`3oool0oooH07AoO`80
07mo00D00:kH0?ooo`3oool0oooo0?oH[P03001oO`03002Ga03Oool0oooo0080oooo00P0glBG001o
O`00Ogl007mo07CHo`3oool0ooo[09bGO`8007mo00<007nG09c[o`3oool02P3oool01@3Oa9L007mo
001oO`00OiL0W>_o00<0oooo00P0omR^001oO`00Ogl007mo04S4j`3oool0oooH07AoO`8007mo00<0
07n^0;ooo`3oool00`3oool01@3ojl@0B7mo001oO`00Ogl0B<C[0080oooo00H0omR^001oO`00Ogl0
07mo002Ga03Oool30?ooo`050=o4U`00Ogl007mo001oO`1df?l0C@3oool008/0oooo00D0omR^001o
O`00Ogl007mo07CHo`040?ooo`030?oH[P00Ogl007mo00H007mo00@0M=So0?ooo`3oool0omR^0`00
Ogl01P00[]P0oooo0?ooo`3oool0oooH07AoO`8007mo00D00:kH0?ooo`3oool0ooo[09bGO`03001o
O`06002^f03oool0oooo0?ooo`3ojl@0B7mo0`00Ogl01@1df?l0oooo0?ooo`3oool0glBG00<007mo
00D0M=So0?ooo`3oool0ooo[09bGO`03001oO`06002^f03oool0oooo0?ooo`3ojl@0B7mo0`00Ogl0
1@1df?l0oooo0?ooo`3oool0glBG00<007mo00@0M=So0?ooo`3oool0omR^0`00Ogl01P00[]P0oooo
0?ooo`3oool0oooH07AoO`8007mo00<00:kH0?ooo`3oool01P3oool00`3oomP0M7mo001oO`0:001o
O`05001o[P2oool0oooo0?oof01dOgl00P00Ogl01P00[]P0oooo0?ooo`3oool0ooo[09bGO`8007mo
00H009O40=ooo`3oool0oooo0?o[a018Ogl3001oO`0504S4j`3oool0oooo0?ooo`3of:h00`00Ogl0
1018a>/0oooo0?ooo`3of:h3001oO`06002^f03oool0oooo0?ooo`3oon/0W9Mo0P00Ogl01@00Ojh0
_ooo0?ooo`3oool0omR^00<007mo00H0B<C[0?ooo`3oool0oooo0?o[a018Ogl2001oO`04002^f03o
ool0oooo0;n^O`<007mo00<0B<C[0?ooo`3oool00P3oool01@3Oa9L007mo001oO`00Ogl0M=So0080
oooo00<0on_404QoO`00Ogl00P00Ogl01@18a>/0oooo0?ooo`3oool0omR^00<007mo00<0B<C[0?oo
o`3oool0B`3oool008/0oooo00D0on_404QoO`00Ogl007mo002^f00@0?ooo`030?ooj`2LUgl007mo
008007mo00@0M=So0?ooo`3oool0glBG0`00Ogl01@00[]P0oooo0?ooo`3oool0_jio00@007mo00@0
0:kH0?ooo`3oool0glBG1@00Ogl01018a>/0oooo0?ooj`2LUgl3001oO`0507CHo`3oool0oooo0?oo
o`2o[Wl01000Ogl01000[]P0oooo0?ooo`3Oa9L5001oO`0404S4j`3oool0ooo[09bGO`<007mo00D0
M=So0?ooo`3oool0ooo[09bGO`03001oO`0407CHo`3oool0oooo0=o4U`<007mo00<00:kH0?ooo`3o
ool01`3oool00`2o[Wl007mo001oO`0:001oO`0407CHo`3oool0ooo[09bGO`8007mo00<007n^0;oo
o`3oool00P3oool01@2o[Wl007mo001oO`00OiL0W>_o0080oooo00<0oooH07AoO`00Ogl00`00Ogl0
1000[]P0oooo0?ooo`2o[Wl3001oO`0504S4j`3oool0oooo0?ooj`2LUgl00`00Ogl01000[]P0oooo
0?ooo`3of:h3001oO`06002Ga03Oool0oooo0?ooo`3oomP0M7mo0P00Ogl01P00[]P0oooo0?ooo`3o
ool0oooH07AoO`8007mo00D009O40=ooo`3oool0on_404QoO`03001oO`0504S4j`3oool0oooo0?o[
a018Ogl00P00Ogl01P00Ojh0_ooo0?ooo`3oool0oooH07AoO`@007mo00@00:kH0?ooo`3oool0_jio
0`00Ogl00`18a>/0oooo0?ooo`1;0?ooo`00R`3oool01P3oon/0W9Mo001oO`00Ogl007nG09c[oa00
oooo00<0omR^001oO`00Ogl02000Ogl01@1df?l0oooo0?ooo`3oool0glBG008007mo00<007n^07B^
O`00Ogl01P00Ogl00`18UiL007mo001oO`05001oO`03001oU`2Ljol0oooo0080oooo00D0glBG001o
O`00Ogl007n^07B^O`07001oO`0304RGU`00Ogl007mo00D007mo00<007nG09c[o`3oool00P3oool0
0`3of:h007mo001oO`08001oO`0307CHo`3oool0oooo00L0oooo00D0glBG001oO`00Ogl007mo04S4
j`0;0?ooo`050=o4U`00Ogl007mo001oU`2Ljol00`3oool01@3Oa9L007mo001oO`00Ogl0M=So0080
oooo00H0ooo[09bGO`00Ogl007mo002GU`18Ogl7001oO`0607CHo`3oool0oooo0?ooo`3ojl@0B7mo
2@00Ogl0101df?l0oooo0?o[a018Ogl6001oO`04001oU`2Ljol0ooo[09bGO`8007mo00D007n^0;oo
o`3oool0oooo0=o4U`0:001oO`0604S4j`3oool0oooo0?ooo`3oon/0W9Mo0P00Ogl00`00UiL0B7mo
001oO`06001oO`0307CHo`3oool0oooo04/0oooo002<0?ooo`050=o4U`00Ogl007mo001oO`00[]P0
203oool01P3oon/0W9Mo001oO`00Ogl007nG09c[o`<0oooo00<0omR^001oO`00Ogl01P00Ogl00`00
[]P0oooo0?ooo`020?ooo`060?oH[P00Ogl007mo001oO`1df>/0W9Mo1@00Ogl00`00Ojh0_ooo0=o4
U`06001oO`03002^f03oool0oooo0080oooo00H0omR^001oO`00Ogl007mo07CHj`2LUgl5001oO`03
001o[P2oool0glBG00H007mo00<00:kH0?ooo`3oool00`3oool00`3of:h007mo001oO`06001oO`03
002^f03oool0oooo00P0oooo00D0omR^001oO`00Ogl007mo002^f00;0?ooo`050?oH[P00Ogl007mo
001oO`1df?l00`3oool01@3of:h007mo001oO`00Ogl0B<C[00<0oooo00H0glBG001oO`00Ogl007n^
0;oof01dOgl5001oO`03002Ga03Oool0oooo00<0oooo00<0oooH07AoO`00Ogl01P00Ogl01@00[]P0
oooo0?ooo`3oon/0W9Mo00L007mo00<0M=So0?ooo`3Oa9L00P00Ogl00`00OiL0W>_o0?ooo`020?oo
o`030?oH[P00Ogl007mo00H007mo00<00:kH0?ooo`3oool00`3oool01P3Oa9L007mo001oO`00Ojh0
_ooH07AoO`D007mo00<009O40=ooo`3oool0C03oool008`0oooo00<0oooH07AoO`00Ogl00P00Ogl0
0`00[]P0oooo0?ooo`050?ooo`060?oH[P00Ogl007mo001oO`00Ul@0gooo1@3oool00`2o[Wl007mo
001oO`03001oO`0307CHo`3oool0oooo00<0oooo00L0on_404QoO`00Ogl007mo04S4j`3oool0omR^
00<007mo00H007n^0;ooo`3oool0oooo0?ooj`2LUgl3001oO`03002^f03oool0oooo00<0oooo00L0
on_404QoO`00Ogl007mo04S4j`3oool0omR^00<007mo00H007n^0;ooo`3oool0oooo0?ooj`2LUgl3
001oO`03002^f03oool0oooo00H0oooo00<0_jio001oO`00Ogl00`00Ogl00`1df?l0oooo0?ooo`09
0?ooo`060?o[a018Ogl007mo001oO`00Ul@0gooo2P3oool01@3ojl@0B7mo001oO`00Ogl00:kH00<0
oooo00D0oooH07AoO`00Ogl007mo002^f0030?ooo`070?oH[P00Ogl007mo001oU`2Ljol0oooo0=o4
U`03001oO`03002^f03oool0oooo00H0oooo00<0on_404QoO`00Ogl00P00Ogl00`00OiL0W>_o0?oo
o`030?ooo`030;n^O`00Ogl007mo00D007mo00<00:kH0?ooo`3of:h00`00Ogl00`1df?l0oooo0?oo
o`030?ooo`030;n^O`00Ogl007mo008007mo00<007n^0;ooo`3oool01@3oool01`3of:h007mo001o
O`00OiL0W>_o0?ooo`3Oa9L00`00Ogl00`00[]P0oooo0?ooo`1=0?ooo`00S@3oool00`3of:h007mo
001oO`03001oO`0307CHo`3oool0oooo0080oooo00<0glBG001oO`00Ogl00P00Ogl00`18a>/0oooo
0?ooo`1G0?ooo`060?oof01dOgl007mo001oO`00Ojh0_ooo>P3oool01@3oomP0M7mo001oO`00Ogl0
B<C[07@0oooo002>0?ooo`030;n^O`00Ogl007mo00/007mo00<007n^0;ooo`3oool0F03oool00`3o
on/0W9Mo001oO`0;001oO`0307CHo`3oool0oooo0300oooo00@0on_404QoO`00Ogl00:kHM03oool0
08l0oooo00<0_jio001oO`00Ogl02P00Ogl00`1df?l0oooo0?ooo`1I0?ooo`030=o4U`00Ogl007mo
00X007mo00<00:kH0?ooo`3oool0<@3oool0103oon/0W9Mo001o[P2oool20?ooo`060?ooj`2LUgl0
07mo001oO`00Ojh0_oooJ`3oool00900oooo00<0omR^001oO`00Ogl02000Ogl00`1df?l0oooo0?oo
o`1J0?ooo`030?oH[P00Ogl007mo00X007mo00<009O40=ooo`3oool0<`3oool00`3Oa:h0W>_o0?oo
o`020?ooo`050;n^O`00Ogl007mo001oU`2Ljol0J`3oool00980oooo00<0on_404QoO`00Ogl01000
Ogl00`1df?l0oooo0?ooo`1L0?ooo`030?o[a018Ogl007mo00X007mo00<007n^0;ooo`3oool0>03o
ool01@3of:h007mo001oO`00Ogl0M=So06/0oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-8.74107, -22.0612, \
0.0528265, 0.228129}}]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Reading values from a plot", "Subsection"],

Cell[TextData[{
  "Suppose that you need to solve a transcendental equation.  ",
  StyleBox["FindRoot", "InlineInput",
    FontWeight->"Bold"],
  " requires that you specify a starting value or the region in which to \
look, or perhaps you would be satisfied with a visual solution.  In either \
case, you can use the mouse to read values from a plot."
}], "Text"],

Cell[TextData[{
  "For example, consider the equation ",
  Cell[BoxData[
      \(TraditionalForm\`\[Xi]\/\@\(\[Beta]\^2 - \[Xi]\^2\)\  \[Equal] \ 
        Cot[\[Xi]]\)]],
  " which arises in the eigenvalue problem for the square well.  Where should \
solutions be sought?  First, plot the two sides separately for some value of \
\[Beta], obviously limiting the range to ",
  Cell[BoxData[
      \(TraditionalForm\`\[Xi] \[LessEqual] \[Beta]\)]],
  "."
}], "Text"],

Cell[BoxData[
    \(lhs\  = \ \[Xi]\/\@\(\[Beta]\^2 - \ \[Xi]\^2\); \ \ rhs\  = \ 
      Cot[\[Xi]];\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Evaluate[{lhs, rhs} /. \[Beta] \[Rule] 5], \ {\[Xi], 0, \ 5}, 
        PlotRange \[Rule] {Automatic, {0, 5}}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.190476 0 0.123607 [
[.21429 -0.0125 -5 -15.6875 ]
[.21429 -0.0125 5 0 ]
[.40476 -0.0125 -5 -15.6875 ]
[.40476 -0.0125 5 0 ]
[.59524 -0.0125 -5 -15.6875 ]
[.59524 -0.0125 5 0 ]
[.78571 -0.0125 -5 -15.6875 ]
[.78571 -0.0125 5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[.01131 .12361 -10 -7.84375 ]
[.01131 .12361 0 7.84375 ]
[.01131 .24721 -10 -7.84375 ]
[.01131 .24721 0 7.84375 ]
[.01131 .37082 -10 -7.84375 ]
[.01131 .37082 0 7.84375 ]
[.01131 .49443 -10 -7.84375 ]
[.01131 .49443 0 7.84375 ]
[.01131 .61803 -10 -7.84375 ]
[.01131 .61803 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 0 m
.21429 .00625 L
s
gsave
.21429 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.40476 0 m
.40476 .00625 L
s
gsave
.40476 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.59524 0 m
.59524 .00625 L
s
gsave
.59524 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.78571 0 m
.78571 .00625 L
s
gsave
.78571 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.125 Mabswid
.0619 0 m
.0619 .00375 L
s
.1 0 m
.1 .00375 L
s
.1381 0 m
.1381 .00375 L
s
.17619 0 m
.17619 .00375 L
s
.25238 0 m
.25238 .00375 L
s
.29048 0 m
.29048 .00375 L
s
.32857 0 m
.32857 .00375 L
s
.36667 0 m
.36667 .00375 L
s
.44286 0 m
.44286 .00375 L
s
.48095 0 m
.48095 .00375 L
s
.51905 0 m
.51905 .00375 L
s
.55714 0 m
.55714 .00375 L
s
.63333 0 m
.63333 .00375 L
s
.67143 0 m
.67143 .00375 L
s
.70952 0 m
.70952 .00375 L
s
.74762 0 m
.74762 .00375 L
s
.82381 0 m
.82381 .00375 L
s
.8619 0 m
.8619 .00375 L
s
.9 0 m
.9 .00375 L
s
.9381 0 m
.9381 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
.02381 .12361 m
.03006 .12361 L
s
gsave
.01131 .12361 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .24721 m
.03006 .24721 L
s
gsave
.01131 .24721 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.02381 .37082 m
.03006 .37082 L
s
gsave
.01131 .37082 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.02381 .49443 m
.03006 .49443 L
s
gsave
.01131 .49443 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.02381 .61803 m
.03006 .61803 L
s
gsave
.01131 .61803 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .02472 m
.02756 .02472 L
s
.02381 .04944 m
.02756 .04944 L
s
.02381 .07416 m
.02756 .07416 L
s
.02381 .09889 m
.02756 .09889 L
s
.02381 .14833 m
.02756 .14833 L
s
.02381 .17305 m
.02756 .17305 L
s
.02381 .19777 m
.02756 .19777 L
s
.02381 .22249 m
.02756 .22249 L
s
.02381 .27193 m
.02756 .27193 L
s
.02381 .29666 m
.02756 .29666 L
s
.02381 .32138 m
.02756 .32138 L
s
.02381 .3461 m
.02756 .3461 L
s
.02381 .39554 m
.02756 .39554 L
s
.02381 .42026 m
.02756 .42026 L
s
.02381 .44498 m
.02756 .44498 L
s
.02381 .46971 m
.02756 .46971 L
s
.02381 .51915 m
.02756 .51915 L
s
.02381 .54387 m
.02756 .54387 L
s
.02381 .56859 m
.02756 .56859 L
s
.02381 .59331 m
.02756 .59331 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 0 m
.06244 .00502 L
.10458 .01052 L
.14415 .01575 L
.18221 .02085 L
.22272 .0264 L
.26171 .03189 L
.30316 .03792 L
.34309 .04398 L
.3815 .05009 L
.42237 .05695 L
.46172 .064 L
.49955 .07127 L
.53984 .07968 L
.57861 .08859 L
.61984 .09918 L
.65954 .11081 L
.69774 .12379 L
.73838 .14029 L
.77751 .16002 L
.79749 .17219 L
.81909 .18761 L
.85916 .22574 L
.8791 .25235 L
.88927 .26912 L
.90016 .29051 L
.91044 .31516 L
.9198 .34305 L
.92922 .37884 L
.93814 .42403 L
.94269 .45365 L
.94772 .49411 L
.95297 .54949 L
.95537 .58145 L
s
.95537 .58145 m
.95768 .61803 L
s
.06151 .61803 m
.06244 .60102 L
.06787 .52476 L
.07287 .46927 L
.08414 .37711 L
.09408 .31971 L
.10458 .27381 L
.12422 .21235 L
.14255 .1719 L
.1629 .13805 L
.18146 .11358 L
.22131 .07308 L
.25964 .04271 L
.30042 .01473 L
s
.30042 .01473 m
.32303 0 L
s
.62291 0 m
.62292 .61803 L
s
.65985 .61803 m
.66385 .55633 L
.66867 .49665 L
.67394 .44383 L
.68284 .3751 L
.69222 .32103 L
.70247 .27575 L
.7118 .24311 L
.73274 .18853 L
.75191 .15257 L
.77269 .12246 L
.81358 .07855 L
.85295 .04644 L
.89081 .02003 L
s
.89081 .02003 m
.92146 0 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00L@3oool01P2oM000000000000000000004RL0=oooch0oooo00<0_g@00000000000001000
00000`18W=l0oooo0?ooo`0l0?ooo`030?oo_`1d000000000080000000<0M;oo0?ooo`3oool0@@3o
ool0103ooml0W4P00018W03Oooln0?ooo`030?oog`2LB00000000080000000<0M;oo0?ooo`3oool0
8P3oool00780oooo00<0omnL04P0001d_ol0@@3oool00`3OW4P0000000000003000000030018W03O
ool0oooo03`0oooo00<0omnL04P00018W=l00P3oool00`2oM000B9cO0?ooo`110?ooo`040?oog`2L
B00004RL0=oooch0oooo00<0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`0R0?ooo`00LP3o
ool00`3ogi`0B00007Boo`120?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0@@3o
ool00`3ookl0M01809cOo`110?ooo`040?oog`2LB00004RL0=oood<0oooo00<0oooO09a8M02oool0
8P3oool00780oooo00<0omnL04P0001d_ol0@`3oool00`3OW4P0M;oo0?ooo`150?ooo`030?oo_`1d
0000M;oo03d0oooo00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`110?ooo`030;md001d
_ol0oooo0240oooo001b0?ooo`030?oOW0180000M;oo04@0oooo00<0_gA809cOo`3oool0A03oool0
0`3o_g@0000007Boo`0m0?ooo`080?oOW01dW=l0oooo0?ooo`3ooml0W4P00018W03Ooom30?ooo`03
0?oog`2LB000B9cO0280oooo001b0?ooo`030?oOW0180000M;oo04@0oooo00<0oooO09a8M02oool0
@`3oool0103ogi`0B0000000B02Lgoln0?ooo`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`13
0?ooo`030=nLB0000000M;oo0280oooo001b0?ooo`030?oOW0180000M;oo04D0oooo00<0okmd001d
_`3oool0@03oool01@3o_g@000000000000004P0W=oo03l0oooo00L0oooO09a8M02oool0oooO09a8
0000B9`0gooo03l0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo0280oooo001b0?ooo`03
0?oOW0180000M;oo04D0oooo00<0oono07@0B02Lgol0@P3oool00`2oM00007Bo0?ooo`100?ooo`06
0?oOW018M;l0oooO09a80000B9`0gooo?`3oool01P3OW4P00000000000000000001809cOob<0oooo
001b0?ooo`030?oOW0180000M;oo04D0oooo00<0oono07@00018W=l0@`3oool00`2oM000B9cO0?oo
o`100?ooo`050=nLB01d_ml0W4P00018W03Oool0?`3oool0103ookl0M00000000018W=lU0?ooo`00
LP3oool00`3ogi`0B00007Boo`100?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l0
?P3oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo0400oooo00D0oooO09a8B018B000
04RL0=ooo`100?ooo`030;md001d_ol0oooo02D0oooo001`0?ooo`050?oog`2LB00000000000001d
_ol0@03oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo?`3oool01P3o_g@0B9cO0?oo
o`3ookl0M000001d_d80oooo00@0omnL04P00000B9`0gooo@03oool01P3ogi`0B000000000000000
001809cOob80oooo001b0?ooo`030?noM0000000M;oo0440oooo00H0oooO09a80000000000000000
M02ooom10?ooo`050=nLB000000000000000M02oool0@`3oool00`3OW4P004RL0=ooo`110?ooo`05
0;md00000000000000000018W=l08P3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool002H0ooooo`00001_000001`0oooo000^0?ooo`040000
003oool0oooo0?ooo`L000000`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`070?ooo`8000000P3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo00L0oooo00D000000?ooo`3oool0oooo0000000=0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00<0oooo0P0000060?ooo`030000003oool0oooo00/0oooo00<000000?oo
o`3oool08P3oool002h0oooo00<000000?ooo`3oool0203oool7000003@0oooo00<000000?ooo`3o
ool08@3oool200000200oooo00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo03T0oooo00<000000?ooo`3oool0:`3oool2000001H0oooo00<000000?ooo`3o
ool08P3oool002h0oooo00<000000?ooo`3oool03`3oool8000004h0oooo0P00001a0?ooo`030000
003oool0oooo06H0oooo00<000000?ooo`3oool0>`3oool002h0oooo00<000000?ooo`3oool05`3o
ool;00000440oooo0P00001c0?ooo`030000003oool0oooo06@0oooo0P00000n0?ooo`00;P3oool0
0`000000oooo0?ooo`0R0?ooo`/00000=@3oool00`000000oooo0?ooo`1c0?ooo`030000003oool0
oooo0680oooo0P0000100?ooo`00;P3oool00`000000oooo0?ooo`0]0?ooo`L00000;03oool20000
07H0oooo00<000000?ooo`3oool0H03oool200000480oooo000^0?ooo`030000003oool0oooo03@0
oooo1`00000T0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3oool0G`3oool00`000000
oooo0?ooo`120?ooo`00;P3oool00`000000oooo0?ooo`0k0?ooo`P000006`3oool00`000000oooo
0?ooo`1g0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool0@`3oool002h0oooo0`00
00130?ooo`L000004P3oool2000007X0oooo00<000000?ooo`3oool0G03oool2000004H0oooo000^
0?ooo`030000003oool0oooo04X0oooo1`00000:0?ooo`030000003oool0oooo07X0oooo00<00000
0?ooo`3oool0F`3oool00`000000oooo0?ooo`160?ooo`00;P3oool00`000000oooo0?ooo`1A0?oo
o`H000000P3oool2000007d0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`170?oo
o`00;P3oool00`000000oooo0?ooo`1G0?ooo`D00000O03oool00`000000oooo0?ooo`1H0?ooo`80
0000BP3oool002h0oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`020?ooo`D00000
M`3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo04X0oooo000^0?ooo`030000003o
ool0oooo05D0oooo0P00000:0?ooo`H00000L@3oool00`000000oooo0?ooo`1F0?ooo`030000003o
ool0oooo04/0oooo000^0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0403oool8
000006T0oooo00<000000?ooo`3oool0E03oool2000004h0oooo000^0?ooo`030000003oool0oooo
05<0oooo00<000000?ooo`3oool06@3oool700000680oooo00<000000?ooo`3oool0D`3oool00`00
0000oooo0?ooo`1>0?ooo`00;P3oool00`000000oooo0?ooo`1A0?ooo`8000008`3oool7000005/0
oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`1?0?ooo`00;P3oool300000500oooo
00<000000?ooo`3oool0:P3oool6000005D0oooo00<000000?ooo`3oool0D03oool200000580oooo
000^0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0<@3oool500000500oooo00<0
00000?ooo`3oool0C`3oool00`000000oooo0?ooo`1B0?ooo`00;P3oool00`000000oooo0?ooo`1>
0?ooo`030000003oool0oooo03L0oooo1@00001;0?ooo`030000003oool0oooo04h0oooo00<00000
0?ooo`3oool0D`3oool002h0oooo00<000000?ooo`3oool0C03oool2000003l0oooo1P0000150?oo
o`030000003oool0oooo04`0oooo0P00001F0?ooo`00;P3oool00`000000oooo0?ooo`1;0?ooo`03
0000003oool0oooo04D0oooo1`00000n0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3o
ool0EP3oool002h0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`1=0?ooo`H00000
>03oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo05L0oooo000^0?ooo`030000003o
ool0oooo04P0oooo0P00001F0?ooo`@00000=03oool00`000000oooo0?ooo`180?ooo`800000FP3o
ool002h0oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`1J0?ooo`D00000;`3oool0
0`000000oooo0?ooo`170?ooo`030000003oool0oooo05X0oooo000^0?ooo`030000003oool0oooo
04H0oooo00<000000?ooo`3oool0H03oool5000002X0oooo00<000000?ooo`3oool0AP3oool00`00
0000oooo0?ooo`1K0?ooo`00;P3oool3000004D0oooo00<000000?ooo`3oool0IP3oool5000002D0
oooo00<000000?ooo`3oool0A03oool2000005h0oooo000^0?ooo`030000003oool0oooo04@0oooo
00<000000?ooo`3oool0K03oool500000200oooo00<000000?ooo`3oool0@`3oool00`000000oooo
0?ooo`1N0?ooo`00;P3oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo0780oooo1@00
000K0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0G`3oool002h0oooo00<00000
0?ooo`3oool0@P3oool00`000000oooo0?ooo`1h0?ooo`@000005`3oool00`000000oooo0?ooo`11
0?ooo`030000003oool0oooo0600oooo000^0?ooo`030000003oool0oooo0440oooo00<000000?oo
o`3oool0O@3oool500000180oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`1Q0?oo
o`00;P3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo08<0oooo1000000>0?ooo`03
0000003oool0oooo03l0oooo00<000000?ooo`3oool0HP3oool002h0oooo00<000000?ooo`3oool0
?`3oool00`000000oooo0?ooo`280?ooo`@000002P3oool00`000000oooo0?ooo`0o0?ooo`030000
003oool0oooo0680oooo000^0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool0S@3o
ool4000000H0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`1S0?ooo`00;P3oool0
0`000000oooo0?ooo`0m0?ooo`030000003oool0oooo0980oooo100000020?ooo`030000003oool0
oooo03d0oooo00<000000?ooo`3oool0I03oool002h0oooo0`00000l0?ooo`030000003oool0oooo
09L0oooo0`00000n0?ooo`030000003oool0oooo06D0oooo000^0?ooo`030000003oool0oooo03/0
oooo00<000000?ooo`3oool0VP3oool5000003T0oooo00<000000?ooo`3oool0IP3oool002h0oooo
00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`2K0?ooo`030000003oool0oooo0080oooo
1000000d0?ooo`030000003oool0oooo06L0oooo000^0?ooo`030000003oool0oooo03T0oooo00<0
00000?ooo`3oool0W03oool00`000000oooo0?ooo`060?ooo`@00000;`3oool00`000000oooo0?oo
o`1X0?ooo`00803oool01P2oM000000000000000000004RL0=ooo`P0oooo00<000000?ooo`3oool0
>03oool00`000000oooo0?ooo`2M0?ooo`030000003oool0oooo00X0oooo0`00000[0?ooo`030000
003oool0oooo06T0oooo000Q0?ooo`030?oOW0180000M;oo00X0oooo00<000000?ooo`3oool0=`3o
ool00`000000oooo0?ooo`2N0?ooo`030000003oool0oooo00d0oooo0`00000X0?ooo`030000003o
ool0oooo06T0oooo000Q0?ooo`030?oOW0180000M;oo00X0oooo00<000000?ooo`3oool0=P3oool0
0`000000oooo0?ooo`2O0?ooo`030000003oool0oooo0100oooo0P00000U0?ooo`030000003oool0
oooo06X0oooo000Q0?ooo`030?oOW0180000M;oo00X0oooo00<000000?ooo`3oool0=@3oool00`00
0000oooo0?ooo`2P0?ooo`030000003oool0oooo0180oooo0`00000Q0?ooo`030000003oool0oooo
06/0oooo000Q0?ooo`030?oOW0180000M;oo00X0oooo00<000000?ooo`3oool0=@3oool00`000000
oooo0?ooo`2P0?ooo`030000003oool0oooo01D0oooo0`00000M0?ooo`030000003oool0oooo06`0
oooo000Q0?ooo`030?oOW0180000M;oo00X0oooo1000000c0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool0603oool3000001T0oooo00<000000?ooo`3oool0K@3oool00240oooo00<0
omnL04P0001d_ol02P3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo0:80oooo00<0
00000?ooo`3oool06`3oool2000001H0oooo00<000000?ooo`3oool0KP3oool00240oooo00<0omnL
04P0001d_ol02P3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0:<0oooo00<00000
0?ooo`3oool07@3oool3000001<0oooo00<000000?ooo`3oool0KP3oool00240oooo00<0omnL04P0
001d_ol02P3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0:<0oooo00<000000?oo
o`3oool0803oool200000100oooo00<000000?ooo`3oool0K`3oool00240oooo00<0omnL04P0001d
_ol02P3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo0:@0oooo00<000000?ooo`3o
ool08P3oool3000000`0oooo00<000000?ooo`3oool0L03oool001l0oooo00D0oooO09a800000000
000007Boo`0:0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0Y@3oool00`000000
oooo0?ooo`0U0?ooo`8000002P3oool00`000000oooo0?ooo`1`0?ooo`008@3oool00`3o_g@00000
07Boo`0:0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0YP3oool00`000000oooo
0?ooo`0W0?ooo`8000001`3oool00`000000oooo0?ooo`1a0?ooo`00;P3oool00`000000oooo0?oo
o`0_0?ooo`030000003oool0oooo0:H0oooo00<000000?ooo`3oool0:@3oool2000000D0oooo00<0
00000?ooo`3oool0L@3oool002h0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`2W
0?ooo`030000003oool0oooo02/0oooo0P0000020?ooo`030000003oool0oooo0780oooo000^0?oo
o`<00000;P3oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo02d0oooo0P00001e0?oo
o`00;P3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3o
ool0;P3oool3000007<0oooo000^0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0
Z03oool00`000000oooo0?ooo`0]0?ooo`040000003oool0oooo0?ooo`800000L@3oool002h0oooo
00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`2Y0?ooo`030000003oool0oooo02`0oooo
00<000000?ooo`3oool0103oool2000006l0oooo000^0?ooo`030000003oool0oooo02`0oooo00<0
00000?ooo`3oool0Z@3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo00H0oooo0P00
001]0?ooo`00;P3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo0:X0oooo00<00000
0?ooo`3oool0:`3oool00`000000oooo0?ooo`090?ooo`800000J`3oool002h0oooo00<000000?oo
o`3oool0:`3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo02/0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`1X0?ooo`00;P3oool00`000000oooo0?ooo`0Z0?ooo`03
0000003oool0oooo0:/0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo06L0oooo000^0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0
Z`3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo00h0oooo0P00001W0?ooo`00;P3o
ool3000002T0oooo00<000000?ooo`3oool0[03oool00`000000oooo0?ooo`0Y0?ooo`030000003o
ool0oooo0140oooo00<000000?ooo`3oool0I03oool002h0oooo00<000000?ooo`3oool0:03oool0
0`000000oooo0?ooo`2]0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool04P3oool2
000006@0oooo000^0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0[@3oool00`00
0000oooo0?ooo`0X0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0H@3oool002h0
oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`2^0?ooo`030000003oool0oooo02P0
oooo00<000000?ooo`3oool05P3oool200000640oooo000^0?ooo`030000003oool0oooo02L0oooo
00<000000?ooo`3oool0[P3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo01T0oooo
0P00001O0?ooo`00;P3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0:l0oooo00<0
00000?ooo`3oool09`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo05`0oooo000^
0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0[`3oool00`000000oooo0?ooo`0V
0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0F`3oool002h0oooo00<000000?oo
o`3oool09@3oool00`000000oooo0?ooo`2`0?ooo`030000003oool0oooo02H0oooo00<000000?oo
o`3oool07P3oool00`000000oooo0?ooo`1J0?ooo`00;P3oool00`000000oooo0?ooo`0U0?ooo`03
0000003oool0oooo0;00oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0P0?ooo`03
0000003oool0oooo05T0oooo000^0?ooo`<00000903oool00`000000oooo0?ooo`2a0?ooo`030000
003oool0oooo02D0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`1H0?ooo`00;P3o
ool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo0;40oooo00<000000?ooo`3oool0903o
ool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo05L0oooo000^0?ooo`030000003oool0
oooo02<0oooo00<000000?ooo`3oool0/P3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0
oooo02@0oooo00<000000?ooo`3oool0EP3oool002h0oooo00<000000?ooo`3oool08`3oool00`00
0000oooo0?ooo`2b0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool09@3oool20000
05H0oooo000^0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0/`3oool00`000000
oooo0?ooo`0S0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0D`3oool002h0oooo
00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`2c0?ooo`030000003oool0oooo02<0oooo
00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`1B0?ooo`00;P3oool00`000000oooo0?oo
o`0R0?ooo`030000003oool0oooo0;<0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?oo
o`0[0?ooo`030000003oool0oooo0540oooo000^0?ooo`030000003oool0oooo0240oooo00<00000
0?ooo`3oool0]03oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo02`0oooo00<00000
0?ooo`3oool0D03oool002h0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`2d0?oo
o`030000003oool0oooo0280oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`1?0?oo
o`00;P3oool300000240oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`0Q0?ooo`03
0000003oool0oooo02l0oooo00<000000?ooo`3oool0CP3oool002h0oooo00<000000?ooo`3oool0
803oool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0
<03oool00`000000oooo0?ooo`1=0?ooo`00;P3oool00`000000oooo0?ooo`0P0?ooo`030000003o
ool0oooo0;D0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0b0?ooo`030000003o
ool0oooo04`0oooo000^0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0]@3oool0
0`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0C03oool0
01h0oooo00<0_g@0000000000000100000000`18W=l0oooo0?ooo`060?ooo`030000003oool0oooo
0200oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo
03<0oooo00<000000?ooo`3oool0B`3oool001l0oooo00<0gia80000000000000`0000000`00B9`0
gooo0?ooo`060?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0]P3oool00`000000
oooo0?ooo`0O0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0BP3oool00200oooo
00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`060?ooo`030000003oool0oooo01l0oooo
00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo03D0oooo
00<000000?ooo`3oool0BP3oool00240oooo00<0gia807Boo`3oool02P3oool00`000000oooo0?oo
o`0O0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?oo
o`0g0?ooo`030000003oool0oooo04T0oooo000R0?ooo`030;mdB02Lgol0oooo00T0oooo00<00000
0?ooo`3oool07P3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo01h0oooo00<00000
0?ooo`3oool0>03oool00`000000oooo0?ooo`180?ooo`008P3oool00`3ooml0W4Qd0;ooo`090?oo
o`@000007@3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo01h0oooo00<000000?oo
o`3oool0>@3oool00`000000oooo0?ooo`170?ooo`008`3oool00`3o_g@007Bo0?ooo`080?ooo`03
0000003oool0oooo01h0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`0M0?ooo`03
0000003oool0oooo03X0oooo00<000000?ooo`3oool0A`3oool002<0oooo00<0oono07@0B02Lgol0
203oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0;P0oooo00<000000?ooo`3oool0
7@3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo04H0oooo000S0?ooo`030?oo_`1d
0000B9cO00P0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`2h0?ooo`030000003o
ool0oooo01d0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`150?ooo`007P3oool0
0`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO00P0oooo00<000000?ooo`3oool07@3oool0
0`000000oooo0?ooo`2h0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0?@3oool0
0`000000oooo0?ooo`150?ooo`007P3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo
203oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0;P0oooo00<000000?ooo`3oool0
703oool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo04@0oooo000O0?ooo`060?oog`2L
B00000000000000007@0_ooo2@3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo0;T0
oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo04<0
oooo000^0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0^@3oool00`000000oooo
0?ooo`0K0?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool0@`3oool002h0oooo00<0
00000?ooo`3oool0703oool00`000000oooo0?ooo`2i0?ooo`030000003oool0oooo01/0oooo00<0
00000?ooo`3oool0@@3oool00`000000oooo0?ooo`120?ooo`00;P3oool3000001/0oooo00<00000
0?ooo`3oool0^P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo0440oooo00<00000
0?ooo`3oool0@P3oool002h0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`2j0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`110?oo
o`00;P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo0;X0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo0440oooo000^0?ooo`030000
003oool0oooo01/0oooo00<000000?ooo`3oool0^P3oool00`000000oooo0?ooo`0J0?ooo`030000
003oool0oooo04@0oooo00<000000?ooo`3oool0@03oool002h0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`2k0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0A03o
ool00`000000oooo0?ooo`100?ooo`00;P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo0;/0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`160?ooo`030000003oool0
oooo03l0oooo000^0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0^`3oool00`00
0000oooo0?ooo`0I0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0?`3oool002h0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`2k0?ooo`030000003oool0oooo01T0
oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`0n0?ooo`00;P3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo0;`0oooo00<000000?ooo`3oool06@3oool00`000000oooo
0?ooo`170?ooo`030000003oool0oooo03h0oooo000^0?ooo`<000006@3oool00`000000oooo0?oo
o`2l0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?oo
o`0m0?ooo`00;P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0;`0oooo00<00000
0?ooo`3oool0603oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo03d0oooo000^0?oo
o`030000003oool0oooo01T0oooo00<000000?ooo`3oool0_03oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo04X0oooo00<000000?ooo`3oool0?03oool002h0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`2m0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3o
ool0BP3oool00`000000oooo0?ooo`0l0?ooo`00;P3oool00`000000oooo0?ooo`0H0?ooo`030000
003oool0oooo0;d0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1;0?ooo`030000
003oool0oooo03/0oooo000^0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0_@3o
ool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0>`3o
ool002h0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`2m0?ooo`030000003oool0
oooo01P0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`0j0?ooo`00;P3oool00`00
0000oooo0?ooo`0G0?ooo`030000003oool0oooo0;h0oooo00<000000?ooo`3oool05`3oool00`00
0000oooo0?ooo`1=0?ooo`030000003oool0oooo03X0oooo000^0?ooo`030000003oool0oooo01L0
oooo00<000000?ooo`3oool0_P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo04d0
oooo00<000000?ooo`3oool0>P3oool002h0oooo0`00000G0?ooo`030000003oool0oooo0;h0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo03T0oooo
000^0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0_P3oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool0>@3oool002h0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`2n0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool0CP3oool00`000000oooo0?ooo`0i0?ooo`00;P3oool00`000000oooo0?ooo`0F0?oo
o`030000003oool0oooo0;l0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`1?0?oo
o`030000003oool0oooo03T0oooo000^0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3o
ool0_`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3o
ool0>03oool002h0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`2o0?ooo`030000
003oool0oooo01H0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`0h0?ooo`00;P3o
ool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0;l0oooo00<000000?ooo`3oool05P3o
ool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo03P0oooo000^0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool0_`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0
oooo0540oooo00<000000?ooo`3oool0=`3oool002h0oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`300?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0DP3oool00`00
0000oooo0?ooo`0g0?ooo`00;P3oool3000001D0oooo00<000000?ooo`3oool0`03oool00`000000
oooo0?ooo`0E0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0=`3oool002h0oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`300?ooo`030000003oool0oooo01D0oooo
00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`0g0?ooo`00;P3oool00`000000oooo0?oo
o`0E0?ooo`030000003oool0oooo0<00oooo00<000000?ooo`3oool05@3oool00`000000oooo0?oo
o`1C0?ooo`030000003oool0oooo03H0oooo000^0?ooo`030000003oool0oooo01D0oooo00<00000
0?ooo`3oool0`03oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo05<0oooo00<00000
0?ooo`3oool0=P3oool001h0oooo00<0oono07@0000000000P0000000`1d_ol0oooo0?ooo`080?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool0`03oool00`000000oooo0?ooo`0D0?oo
o`030000003oool0oooo05@0oooo00<000000?ooo`3oool0=P3oool001h0oooo00<0omnL04P00018
W=l00P3oool00`2oM000B9cO0?ooo`080?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3o
ool0`@3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3o
ool0=P3oool002<0oooo00<0oono07@0B02Lgol0203oool00`000000oooo0?ooo`0D0?ooo`030000
003oool0oooo0<40oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`1E0?ooo`030000
003oool0oooo03D0oooo000S0?ooo`030?oo_`1d0000M;oo00P0oooo00<000000?ooo`3oool0503o
ool00`000000oooo0?ooo`310?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0E@3o
ool00`000000oooo0?ooo`0e0?ooo`008`3oool00`3o_g@0000007Boo`080?ooo`030000003oool0
oooo01@0oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0
oooo05D0oooo00<000000?ooo`3oool0=@3oool00280oooo00@0omnL04P0000004P0W=oo203oool4
000001<0oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0
oooo05H0oooo00<000000?ooo`3oool0=@3oool00200oooo00D0okmd000000000000001809cOo`09
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0`P3oool00`000000oooo0?ooo`0C
0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0=03oool00280oooo00<0_g@0001d
_`3oool02@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0<80oooo00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo03@0oooo000S0?ooo`03
0;md0018W=l0oooo00P0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`320?ooo`03
0000003oool0oooo01<0oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`0d0?ooo`00
7P3oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo00P0oooo00<000000?ooo`3oool0
4`3oool00`000000oooo0?ooo`320?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0
E`3oool00`000000oooo0?ooo`0d0?ooo`007`3oool01P3o_g@0B9cO0?ooo`3ookl0M000001d_`T0
oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`320?ooo`030000003oool0oooo01<0
oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`0d0?ooo`00803oool01@3OW4P00000
0000000007@0_ooo00T0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`320?ooo`03
0000003oool0oooo01<0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`0c0?ooo`00
;P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0<80oooo00<000000?ooo`3oool0
4P3oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo03<0oooo000^0?ooo`030000003o
ool0oooo0180oooo00<000000?ooo`3oool0``3oool00`000000oooo0?ooo`0B0?ooo`030000003o
ool0oooo05T0oooo00<000000?ooo`3oool0<`3oool002h0oooo0`00000B0?ooo`030000003oool0
oooo0<<0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`1I0?ooo`030000003oool0
oooo03<0oooo000^0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0``3oool00`00
0000oooo0?ooo`0B0?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool0<P3oool002h0
oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`330?ooo`030000003oool0oooo0180
oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`0b0?ooo`00;P3oool00`000000oooo
0?ooo`0B0?ooo`030000003oool0oooo0<<0oooo00<000000?ooo`3oool04P3oool00`000000oooo
0?ooo`1J0?ooo`030000003oool0oooo0380oooo000^0?ooo`030000003oool0oooo0180oooo00<0
00000?ooo`3oool0``3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo05X0oooo00<0
00000?ooo`3oool0<P3oool002h0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`33
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0a
0?ooo`00;P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0<<0oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo0340oooo000^0?ooo`03
0000003oool0oooo0140oooo00<000000?ooo`3oool0a03oool00`000000oooo0?ooo`0A0?ooo`03
0000003oool0oooo05`0oooo00<000000?ooo`3oool0<@3oool002h0oooo00<000000?ooo`3oool0
4@3oool00`000000oooo0?ooo`340?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
G03oool00`000000oooo0?ooo`0a0?ooo`00;P3oool300000140oooo00<000000?ooo`3oool0a03o
ool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool0<03o
ool002h0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`340?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`0`0?ooo`00;P3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo0<@0oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`1M0?ooo`030000003oool0oooo0300oooo000^0?ooo`030000003oool0oooo0140
oooo00<000000?ooo`3oool0a03oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo05d0
oooo00<000000?ooo`3oool0<03oool002h0oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`340?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0G@3oool00`000000oooo
0?ooo`0`0?ooo`00;P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0<@0oooo00<0
00000?ooo`3oool0403oool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo0300oooo000^
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0a@3oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool0<03oool002h0oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`350?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0GP3oool00`000000oooo0?ooo`0`0?ooo`00;P3oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0oooo0<D0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1O0?ooo`03
0000003oool0oooo02l0oooo000^0?ooo`<00000403oool00`000000oooo0?ooo`350?ooo`030000
003oool0oooo0100oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?ooo`0_0?ooo`00;P3o
ool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0<D0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo02l0oooo000^0?ooo`030000003oool0
oooo0100oooo00<000000?ooo`3oool0a@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0
oooo05l0oooo00<000000?ooo`3oool0;`3oool002h0oooo00<000000?ooo`3oool0403oool00`00
0000oooo0?ooo`350?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0G`3oool00`00
0000oooo0?ooo`0_0?ooo`00;P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0<D0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo02l0
oooo000^0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0aP3oool00`000000oooo
0?ooo`0?0?ooo`030000003oool0oooo0600oooo00<000000?ooo`3oool0;`3oool002h0oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`360?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool0H03oool00`000000oooo0?ooo`0_0?ooo`00;P3oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1P
0?ooo`030000003oool0oooo02l0oooo000^0?ooo`030000003oool0oooo00l0oooo00<000000?oo
o`3oool0aP3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0640oooo00<000000?oo
o`3oool0;P3oool002h0oooo0`00000?0?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3o
ool03`3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo02h0oooo000^0?ooo`030000
003oool0oooo00l0oooo00<000000?ooo`3oool0aP3oool00`000000oooo0?ooo`0?0?ooo`030000
003oool0oooo0640oooo00<000000?ooo`3oool0;P3oool002h0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`360?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0H@3o
ool00`000000oooo0?ooo`0^0?ooo`00;P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo0<H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0
oooo02h0oooo000^0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0aP3oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool0;P3oool00280
oooo00@0oooO09a80000B9`0gooo203oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
0<H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo
02h0oooo000R0?ooo`040?oog`2LB00004RL0=ooo`P0oooo00<000000?ooo`3oool03P3oool00`00
0000oooo0?ooo`370?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0H`3oool00`00
0000oooo0?ooo`0]0?ooo`008P3oool0103ooml0W4P00018W03Oool80?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool0a`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool0;@3oool001h0oooo00<0_g@00000000000001@0000000`1d_ol0
oooo0?ooo`050?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0a`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0;@3oool001h0oooo
00P0omnL07BLg`3oool0oooo0?oog`2LB00004RL0=ooo`P0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`370?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0H`3oool0
0`000000oooo0?ooo`0]0?ooo`007`3oool01`3OW4P0M;oo0?ooo`3ooml0W4P00018W03Oool0203o
ool4000000d0oooo00<000000?ooo`3oool0a`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`3oool0;@3oool001l0oooo00L0oooO09a8M02oool0oooO09a8
0000B9`0gooo00P0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`370?ooo`030000
003oool0oooo00h0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0]0?ooo`00803o
ool01P3ogi`0B7Bo0?oog`2LB00004RL0=ooo`P0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`370?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0H`3oool00`000000
oooo0?ooo`0]0?ooo`008@3oool01@3OW4P0M;oO09a80000B9`0gooo00P0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`370?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`0]0?ooo`008@3oool01@3ooml0W4Q804Q80000B9`0gooo00P0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`370?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0]0?ooo`008P3oool0103ogi`0B000
0018W03Oool80?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0a`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0;03oool002<0oooo
00<0gia80018W03Oool0203oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0<P0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo02`0oooo
000^0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3oool0;03oool002h0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`380?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool0I@3oool00`000000oooo0?ooo`0/0?ooo`00;P3oool3000000d0oooo00<000000?oo
o`3oool0b03oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo06D0oooo00<000000?oo
o`3oool0;03oool002h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`380?ooo`03
0000003oool0oooo00d0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`0/0?ooo`00
;P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo02`0oooo000^0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo06D0oooo00<000000?ooo`3oool0;03oool002h0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`380?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0I@3oool0
0`000000oooo0?ooo`0/0?ooo`00;P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
0<P0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo
02`0oooo000^0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0b03oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo06H0oooo00<000000?ooo`3oool0:`3oool002h0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`380?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`0[0?ooo`00;P3oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`1W0?ooo`030000003oool0oooo02/0oooo000^0?ooo`<00000303oool00`000000oooo0?ooo`39
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`0[
0?ooo`00;P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0<T0oooo00<000000?oo
o`3oool0303oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo02/0oooo000^0?ooo`03
0000003oool0oooo00`0oooo00<000000?ooo`3oool0b@3oool00`000000oooo0?ooo`0<0?ooo`03
0000003oool0oooo06L0oooo00<000000?ooo`3oool0:`3oool002h0oooo00<000000?ooo`3oool0
303oool00`000000oooo0?ooo`390?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
I`3oool00`000000oooo0?ooo`0[0?ooo`00;P3oool00`000000oooo0?ooo`0<0?ooo`030000003o
ool0oooo0<T0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1W0?ooo`030000003o
ool0oooo02/0oooo000^0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0b@3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo06L0oooo00<000000?ooo`3oool0:`3oool0
02h0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`390?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`0[0?ooo`00;P3oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo0<T0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`1W0?ooo`030000003oool0oooo02/0oooo000^0?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool0b@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo06P0oooo
00<000000?ooo`3oool0:P3oool002h0oooo0`00000<0?ooo`030000003oool0oooo0<T0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`1X0?ooo`030000003oool0oooo02X0oooo000^
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0b@3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool0:P3oool002h0oooo00<000000?oo
o`3oool0303oool00`000000oooo0?ooo`390?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool0J03oool00`000000oooo0?ooo`0Z0?ooo`00;P3oool00`000000oooo0?ooo`0<0?ooo`03
0000003oool0oooo0<T0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1X0?ooo`03
0000003oool0oooo02X0oooo000^0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
b@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool0
:P3oool002h0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`3:0?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool0J03oool00`000000oooo0?ooo`0Z0?ooo`00;P3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0<X0oooo00<000000?ooo`3oool0303oool0
0`000000oooo0?ooo`1X0?ooo`030000003oool0oooo02X0oooo000^0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool0bP3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
06T0oooo00<000000?ooo`3oool0:P3oool002h0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`3:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0J@3oool00`000000
oooo0?ooo`0Z0?ooo`00;P3oool3000000/0oooo00<000000?ooo`3oool0bP3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo06T0oooo00<000000?ooo`3oool0:P3oool002h0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`3:0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool0J@3oool00`000000oooo0?ooo`0Z0?ooo`00;P3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo0<X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1Y
0?ooo`030000003oool0oooo02X0oooo000^0?ooo`030000003oool0oooo00/0oooo00<000000?oo
o`3oool0bP3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo06X0oooo00<000000?oo
o`3oool0:@3oool001h0oooo00<0oooO09a8000000000P0000000`1d_ol0oooo0?ooo`080?ooo`03
0000003oool0oooo00/0oooo00<000000?ooo`3oool0bP3oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo06X0oooo00<000000?ooo`3oool0:@3oool001h0oooo00<0oooO09a8M02oool0
0P3oool00`2oM000M;oo0?ooo`080?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
bP3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo06X0oooo00<000000?ooo`3oool0
:@3oool002<0oooo00<0oooO09a8M02oool0203oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo0<X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1Z0?ooo`030000003o
ool0oooo02T0oooo000T0?ooo`030;md001d_ol0oooo00L0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`3:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0JP3oool0
0`000000oooo0?ooo`0Y0?ooo`008`3oool00`3ooml0W4P004RLg`080?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool0bP3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
06X0oooo00<000000?ooo`3oool0:@3oool002<0oooo00<0gia80000001d_ol0203oool4000000X0
oooo00<000000?ooo`3oool0bP3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo06X0
oooo00<000000?ooo`3oool0:@3oool001l0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo
0?l0ooooR`3oool001l0oooo00H0gia800000000000000000000B02Lgooo0?oooh`0oooo000O0?oo
o`040?oo_`1d0000000004RLgol0ooooSP3oool00200oooo00<0_g@007Boo`3oool0o`3ooon>0?oo
o`00803oool01P3ogi`0B000000000000000001809cOool0ooooR`3oool00240oooo00D0_g@00000
00000000000004RLg`3o0?oooh/0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.671831, -0.62174, \
0.0215349, 0.0331849}}]
}, Closed]],

Cell[TextData[
"The vertical line represents the divergence of the cotangent function at \
\[Pi] and should be disregarded, while the other two intersections are the \
desired solutions.  If you now click on the figure you will see a bounding \
box.  The size of the figure can be changed by grabbing one of the buttons \
and dragging the border with the mouse.  More importantly for us, the cursor \
appears as a cross-hair which can be placed over the point of interest.  To \
read values from the graph follow these steps."], "Text"],

Cell["\<\
First press and hold the control key.  If you then move the mouse you should \
see the cursor coordinates somewhere in the border of the notebook display \
(lower left for me).   \
\>", "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell["\<\
Clicking on the desired point, while the control key is held, then places a \
point on the figure.  \
\>", "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell["\<\
Next, the editor copy function (Ctrl-C) copies the coordinates of that point \
into the paste buffer.  \
\>", "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell["\<\
Finally, clicking somewhere else in the notebook and using the paste editing \
function (Ctrl-V) pastes those coordinates into the notebook.  \
\>", "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell["\<\
The approximate solutions that I found in this manner are listed below.\
\>", "Text",
  CellDingbat->None],

Cell[BoxData[
    \({1.3035, \ 0.325631}; \ {3.82772, \ 1.24741};\)], "Input"],

Cell["More precise values are now obtained easily.", "Text"],

Cell[BoxData[
    \(\(eq\  = \ \((lhs == rhs)\) /. \[Beta] \[Rule] 5;\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(FindRoot[eq, {\[Xi], 1.3}]\)], "Input"],

Cell[BoxData[
    \({\[Xi] \[Rule] 1.306440008369511`}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FindRoot[eq, {\[Xi], 3.8}]\)], "Input"],

Cell[BoxData[
    \({\[Xi] \[Rule] 3.837467106499049`}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Graphically find the first three solutions to the equation ",
  Cell[BoxData[
      \(TraditionalForm\`Sin[x] \[Equal] Log[x\/\(x - 1\)]\)]],
  "; then find more precise numerical solutions. "
}], "Exercise"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Manipulating graphics objects", "Section"],

Cell[CellGroupData[{

Cell["Disabling or enabling display", "Subsection"],

Cell[TextData[{
  "Plots are created as graphics objects, which can be named or placed into \
lists or arrays for further processing.  For example, one often collects \
several figures into a graphics array to produce a combined display.  These \
figures could be superimposed or could be displayed as a matrix of separate \
figures.  It is then useful to suppress display of intermediate results.  \
This can be done by using the option ",
  StyleBox["DisplayFunction", "InlineInput",
    FontWeight->"Bold"],
  " whose default value ",
  StyleBox["$DisplayFunction", "InlineInput",
    FontWeight->"Bold"],
  " generally instructs ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " to display graphical output on a video monitor.  If one chooses ",
  StyleBox["$DisplayFunction\[Rule]Identity", "InlineInput",
    FontWeight->"Bold"],
  " as one of the plot options, the identity function does not produce a \
display.  One can then display the graphics object later using ",
  ButtonBox["Show",
    ButtonStyle->"RefGuideLink"],
  "."
}], "Text"],

Cell["\<\
For example, suppose that we create several plots with disabled display.\
\>", "Text"],

Cell[BoxData[
    \(\(\(\(Clear[x, v, a, t];\)\n
    x[t]\  = \ Cos[t]; \ v[t]\  = \ D[x[t], t]; \ a[t]\  = \ D[v[t], t];\n
    \(plot[x]\  = \ 
        Plot[Evaluate[x[t]], {t, 0, 2  \[Pi]}, 
          DisplayFunction \[Rule] Identity];\)\ \n
    \(plot[v]\  = \ 
        Plot[Evaluate[v[t]], {t, 0, 2  \[Pi]}, 
          DisplayFunction \[Rule] Identity];\)\ \n
    \(plot[a]\  = \ 
        Plot[Evaluate[a[t]], {t, 0, 2  \[Pi]}, 
          DisplayFunction \[Rule] Identity];\)\)\(\ \)\)\)], "Input"],

Cell[TextData[{
  "Any one of these can be displayed using ",
  StyleBox["Show", "InlineInput",
    FontWeight->"Bold"],
  ", but we need to reset the display function."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[plot[v], DisplayFunction \[Rule] $DisplayFunction];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.151576 0.309018 0.294303 [
[.17539 .29652 -5 -15.6875 ]
[.17539 .29652 5 0 ]
[.32696 .29652 -5 -15.6875 ]
[.32696 .29652 5 0 ]
[.47854 .29652 -5 -15.6875 ]
[.47854 .29652 5 0 ]
[.63011 .29652 -5 -15.6875 ]
[.63011 .29652 5 0 ]
[.78169 .29652 -5 -15.6875 ]
[.78169 .29652 5 0 ]
[.93327 .29652 -5 -15.6875 ]
[.93327 .29652 5 0 ]
[.01131 .01471 -18.3125 -7.84375 ]
[.01131 .01471 0 7.84375 ]
[.01131 .16187 -27.3125 -7.84375 ]
[.01131 .16187 0 7.84375 ]
[.01131 .45617 -19 -7.84375 ]
[.01131 .45617 0 7.84375 ]
[.01131 .60332 -10 -7.84375 ]
[.01131 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.17539 .30902 m
.17539 .31527 L
s
gsave
.17539 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.32696 .30902 m
.32696 .31527 L
s
gsave
.32696 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.47854 .30902 m
.47854 .31527 L
s
gsave
.47854 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.63011 .30902 m
.63011 .31527 L
s
gsave
.63011 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.78169 .30902 m
.78169 .31527 L
s
gsave
.78169 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.93327 .30902 m
.93327 .31527 L
s
gsave
.93327 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.125 Mabswid
.05412 .30902 m
.05412 .31277 L
s
.08444 .30902 m
.08444 .31277 L
s
.11476 .30902 m
.11476 .31277 L
s
.14507 .30902 m
.14507 .31277 L
s
.2057 .30902 m
.2057 .31277 L
s
.23602 .30902 m
.23602 .31277 L
s
.26633 .30902 m
.26633 .31277 L
s
.29665 .30902 m
.29665 .31277 L
s
.35728 .30902 m
.35728 .31277 L
s
.38759 .30902 m
.38759 .31277 L
s
.41791 .30902 m
.41791 .31277 L
s
.44822 .30902 m
.44822 .31277 L
s
.50885 .30902 m
.50885 .31277 L
s
.53917 .30902 m
.53917 .31277 L
s
.56948 .30902 m
.56948 .31277 L
s
.5998 .30902 m
.5998 .31277 L
s
.66043 .30902 m
.66043 .31277 L
s
.69074 .30902 m
.69074 .31277 L
s
.72106 .30902 m
.72106 .31277 L
s
.75138 .30902 m
.75138 .31277 L
s
.81201 .30902 m
.81201 .31277 L
s
.84232 .30902 m
.84232 .31277 L
s
.87264 .30902 m
.87264 .31277 L
s
.90295 .30902 m
.90295 .31277 L
s
.96358 .30902 m
.96358 .31277 L
s
.9939 .30902 m
.9939 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.02381 .01471 m
.03006 .01471 L
s
gsave
.01131 .01471 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .16187 m
.03006 .16187 L
s
gsave
.01131 .16187 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .45617 m
.03006 .45617 L
s
gsave
.01131 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .60332 m
.03006 .60332 L
s
gsave
.01131 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04415 m
.02756 .04415 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .10301 m
.02756 .10301 L
s
.02381 .13244 m
.02756 .13244 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .25016 m
.02756 .25016 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .36788 m
.02756 .36788 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .4856 m
.02756 .4856 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .30902 m
.06244 .23481 L
.10458 .15951 L
.14415 .09914 L
.18221 .05447 L
.20178 .03757 L
.21267 .03011 L
.22272 .0245 L
.23293 .02007 L
.23832 .01827 L
.24402 .01676 L
.24641 .01625 L
.24897 .01579 L
.25119 .01545 L
.25361 .01516 L
.25507 .01501 L
.25643 .01491 L
.25773 .01483 L
.2584 .01479 L
.25912 .01476 L
.25983 .01474 L
.26048 .01473 L
.26176 .01472 L
.26293 .01472 L
.26416 .01475 L
.26538 .01479 L
.26607 .01483 L
.26671 .01486 L
.26912 .01505 L
.27131 .01528 L
.27364 .0156 L
.27847 .01647 L
.2838 .01778 L
.29326 .02099 L
.30213 .02502 L
.32232 .03778 L
.34082 .05371 L
.38045 .10022 L
.41855 .15838 L
.45912 .23059 L
.49816 .30545 L
.53966 .38514 L
.57964 .45663 L
.61811 .51582 L
.63759 .54096 L
.65902 .56418 L
.67938 .58152 L
.69843 .5933 L
.70884 .59785 L
.71396 .5996 L
Mistroke
.71872 .60092 L
.72301 .60186 L
.72758 .60261 L
.7302 .60292 L
.73148 .60304 L
.73266 .60313 L
.73371 .6032 L
.73487 .60325 L
.73603 .60329 L
.73725 .60332 L
.73856 .60332 L
.73977 .6033 L
.7404 .60329 L
.7411 .60326 L
.74251 .6032 L
.74375 .60312 L
.74508 .60301 L
.74748 .60276 L
.74985 .60244 L
.75235 .60202 L
.7568 .60108 L
.76198 .59967 L
.7668 .59806 L
.77773 .59332 L
.78828 .58734 L
.79823 .58046 L
.81683 .5645 L
.85534 .51958 L
.89632 .45703 L
.93577 .38657 L
.97371 .31383 L
.97619 .30902 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000N0?ooo`060;md0000000000000000
0000B9`0gooo2@3oool00`000000oooo0?ooo`3o0?oooh40oooo000O0?ooo`030?oOW0180000M;oo
00/0oooo00<000000?ooo`3oool0o`3ooon10?ooo`007`3oool00`3ogi`0B00007Boo`0;0?ooo`03
0000003oool0oooo0?l0ooooP@3oool001l0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo
0?ooo`3o0?oooh40oooo000@0?ooo`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool00P3o
ool00`3ogi`0B00007Boo`0;0?ooo`@00000F@3oool400000?l0oooo8`3oool001l0oooo00<0omnL
04P0001d_ol02`3oool00`000000oooo0?ooo`1D0?ooo`L000000P3oool600000?l0oooo7P3oool0
01l0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`1B0?ooo`8000003`3oool30000
0?l0oooo6`3oool001l0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`1?0?ooo`<0
0000503oool300000?l0oooo603oool001l0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo
0?ooo`1<0?ooo`<000006P3oool200000?l0oooo5P3oool001l0oooo00<0omnL04P0001d_ol02`3o
ool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0o`3o
oolC0?ooo`007@3oool01@3ooml0W4P0000000000000M;oo00/0oooo00<000000?ooo`3oool0B@3o
ool200000240oooo0P00003o0?oooa<0oooo000O0?ooo`030?noM0000000M;oo00/0oooo00<00000
0?ooo`3oool0B03oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0?l0oooo403oool0
02d0oooo00<000000?ooo`3oool0AP3oool2000002L0oooo0P00003o0?oooa00oooo000]0?ooo`03
0000003oool0oooo04@0oooo0P00000[0?ooo`800000o`3oool>0?ooo`00;@3oool00`000000oooo
0?ooo`130?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0o`3oool;0?ooo`00;@3o
ool00`000000oooo0?ooo`120?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0o`3o
ool:0?ooo`00;@3oool300000440oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`3o
0?ooo`T0oooo000]0?ooo`030000003oool0oooo03l0oooo0P00000e0?ooo`800000o`3oool90?oo
o`00;@3oool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3o
ool0o`3oool60?ooo`00;@3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo03T0oooo
00<000000?ooo`3oool0o`3oool50?ooo`00;@3oool00`000000oooo0?ooo`0l0?ooo`030000003o
ool0oooo03/0oooo00<000000?ooo`3oool0o`3oool40?ooo`00;@3oool00`000000oooo0?ooo`0k
0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool0o`3oool30?ooo`00;@3oool00`00
0000oooo0?ooo`0j0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0o`3oool20?oo
o`00;@3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3o
ool0o`3oool10?ooo`00;@3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo0480oooo
00<000000?ooo`3oool0o`3oool002d0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?oo
o`130?ooo`030000003oool0oooo0?l0oooo000]0?ooo`030000003oool0oooo03L0oooo00<00000
0?ooo`3oool0A@3oool00`000000oooo0?ooo`3n0?ooo`00;@3oool00`000000oooo0?ooo`0f0?oo
o`030000003oool0oooo04L0oooo00<000000?ooo`3oool0o@3oool002d0oooo0`00000e0?ooo`03
0000003oool0oooo04T0oooo00<000000?ooo`3oool0o03oool002d0oooo00<000000?ooo`3oool0
=03oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo0?/0oooo000]0?ooo`030000003o
ool0oooo03@0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`3j0?ooo`00;@3oool0
0`000000oooo0?ooo`0c0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool0n@3oool0
02d0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo
0?P0oooo000]0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0DP3oool00`000000
oooo0?ooo`3g0?ooo`00;@3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo05<0oooo
00<000000?ooo`3oool0m`3oool002d0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?oo
o`1E0?ooo`030000003oool0oooo0?H0oooo000]0?ooo`030000003oool0oooo02l0oooo00<00000
0?ooo`3oool0EP3oool00`000000oooo0?ooo`3e0?ooo`00;@3oool00`000000oooo0?ooo`0^0?oo
o`030000003oool0oooo05P0oooo00<000000?ooo`3oool0m03oool002d0oooo00<000000?ooo`3o
ool0;@3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo0?<0oooo000]0?ooo`<00000
;03oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo0?80oooo000]0?ooo`030000003o
ool0oooo02`0oooo00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`3b0?ooo`00;@3oool0
0`000000oooo0?ooo`0[0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool0l@3oool0
02d0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo
0?00oooo000]0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0H03oool00`000000
oooo0?ooo`3`0?ooo`00;@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0680oooo
00<000000?ooo`3oool0k`3oool002d0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?oo
o`1T0?ooo`030000003oool0oooo0>h0oooo000]0?ooo`030000003oool0oooo02P0oooo00<00000
0?ooo`3oool0I03oool00`000000oooo0?ooo`3^0?ooo`00;@3oool00`000000oooo0?ooo`0W0?oo
o`030000003oool0oooo06H0oooo00<000000?ooo`3oool0k@3oool002d0oooo00<000000?ooo`3o
ool09P3oool00`000000oooo0?ooo`1X0?ooo`030000003oool0oooo0>`0oooo000]0?ooo`030000
003oool0oooo02H0oooo00<000000?ooo`3oool0J03oool00`000000oooo0?ooo`3/0?ooo`00;@3o
ool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo06X0oooo00<000000?ooo`3oool0j`3o
ool002d0oooo0`00000T0?ooo`030000003oool0oooo06/0oooo00<000000?ooo`3oool0j`3oool0
02d0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`1/0?ooo`030000003oool0oooo
0>X0oooo000]0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0KP3oool00`000000
oooo0?ooo`3Y0?ooo`00;@3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo06l0oooo
00<000000?ooo`3oool0j@3oool002d0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?oo
o`1`0?ooo`030000003oool0oooo0>P0oooo000]0?ooo`030000003oool0oooo0240oooo00<00000
0?ooo`3oool0LP3oool00`000000oooo0?ooo`3W0?ooo`00;@3oool00`000000oooo0?ooo`0P0?oo
o`030000003oool0oooo07<0oooo00<000000?ooo`3oool0i`3oool00140oooo00D0omnL04P00000
0000001d0;ooo`030?ooo`060?noM0000000M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?oo
o`080?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0M03oool00`000000oooo0?oo
o`3V0?ooo`00403oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@00000
07Boo`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`080?ooo`030000003oool0oooo
01l0oooo00<000000?ooo`3oool0MP3oool00`000000oooo0?ooo`3U0?ooo`00403oool00`2oM000
04RL0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO09a8M02oool0203oool00`000000
oooo0?ooo`0N0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool0i@3oool000l0oooo
00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0;0?ooo`030;md001d_ol0oooo00L0oooo
00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo0>@0oooo
00030?ooo`030?oog`2LB000000000L0000000D0B9cO0?ooo`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00X0oooo00<0oooO09a80018W=l0203oool00`000000oooo0?ooo`0M0?ooo`030000
003oool0oooo07X0oooo00<000000?ooo`3oool0h`3oool000l0oooo00<0omnL04P0001d_ol00`3o
ool00`3o_g@0001809cOo`0:0?ooo`030=nLB0000000M;oo00P0oooo1000000K0?ooo`030000003o
ool0oooo07/0oooo00<000000?ooo`3oool0h`3oool000l0oooo00<0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`060?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`080?ooo`030000
003oool0oooo01`0oooo00<000000?ooo`3oool0O03oool00`000000oooo0?ooo`3R0?ooo`003`3o
ool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00H0gia80000000000000000
0000B02Lgol90?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0O@3oool00`000000
oooo0?ooo`3R0?ooo`003`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo
00@0oono07@000000000B9cO2`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo07h0
oooo00<000000?ooo`3oool0h@3oool00100oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo
0?ooo`070?ooo`030;md001d_ol0oooo00/0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`1o0?ooo`030000003oool0oooo0>40oooo000@0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0
001d0;ooo`080?ooo`060?oOW018000000000000000004P0W=oo203oool00`000000oooo0?ooo`0J
0?ooo`030000003oool0oooo0800oooo00<000000?ooo`3oool0h03oool00140oooo00D0oono07@0
00000000001d0;ooo`0:0?ooo`050;md00000000000000000018W=l0203oool00`000000oooo0?oo
o`0I0?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool0h03oool002d0oooo00<00000
0?ooo`3oool06@3oool00`000000oooo0?ooo`220?ooo`030000003oool0oooo0=l0oooo000]0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`3N0?oo
o`00;@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo08D0oooo00<000000?ooo`3o
ool0gP3oool002d0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`260?ooo`030000
003oool0oooo0=d0oooo000]0?ooo`<000005P3oool00`000000oooo0?ooo`270?ooo`030000003o
ool0oooo0=d0oooo000]0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0R03oool0
0`000000oooo0?ooo`3L0?ooo`00;@3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo
08T0oooo00<000000?ooo`3oool0g03oool002d0oooo00<000000?ooo`3oool05@3oool00`000000
oooo0?ooo`2:0?ooo`030000003oool0oooo0=/0oooo000]0?ooo`030000003oool0oooo01@0oooo
00<000000?ooo`3oool0R`3oool00`000000oooo0?ooo`3K0?ooo`00;@3oool00`000000oooo0?oo
o`0D0?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool0fP3oool002d0oooo00<00000
0?ooo`3oool04`3oool00`000000oooo0?ooo`2=0?ooo`030000003oool0oooo0=X0oooo000]0?oo
o`030000003oool0oooo01<0oooo00<000000?ooo`3oool0SP3oool00`000000oooo0?ooo`3I0?oo
o`00;@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0900oooo00<000000?ooo`3o
ool0f03oool002d0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`2A0?ooo`030000
003oool0oooo0=P0oooo000]0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0TP3o
ool00`000000oooo0?ooo`3G0?ooo`00;@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0
oooo09<0oooo00<000000?ooo`3oool0e`3oool002d0oooo0`00000@0?ooo`030000003oool0oooo
09@0oooo00<000000?ooo`3oool0eP3oool002d0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`2E0?ooo`030000003oool0oooo0=H0oooo000]0?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool0UP3oool00`000000oooo0?ooo`3E0?ooo`00;@3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo09L0oooo00<000000?ooo`3oool0e@3oool002d0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`2H0?ooo`030000003oool0oooo0=@0oooo000]0?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool0VP3oool00`000000oooo0?ooo`3C0?oo
o`00;@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo09/0oooo00<000000?ooo`3o
ool0d`3oool002d0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`2L0?ooo`030000
003oool0oooo0=80oooo000]0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0W@3o
ool00`000000oooo0?ooo`3B0?ooo`00;@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo09h0oooo00<000000?ooo`3oool0d@3oool002d0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`2O0?ooo`030000003oool0oooo0=40oooo000]0?ooo`<000002P3oool00`000000
oooo0?ooo`2P0?ooo`030000003oool0oooo0=00oooo000]0?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`3@0?ooo`00;@3oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo02/0oooo00H0_g@000000000000000000018W03Ooolc0?ooo`03
0;md00000000000000@0000000<0B9cO0?ooo`3oool0<P3oool00`3ookl0M0000000000200000003
07Boo`3oool0oooo03P0oooo00@0oooO09a80000B9`0gooo=03oool00`3ooml0W4P0000000020000
000307Boo`3oool0oooo03H0oooo00D0omnL04P000000000000007Boo`0I0?ooo`00;@3oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo02d0oooo00<0omnL04P0001d_ol0=P3oool00`3O
W4P0000000000003000000030018W03Oool0oooo0380oooo00L0omnL04P00018W=l000000?ooo`2o
M000B9cO03T0oooo00@0oooO09a80000B9`0gooo=03oool00`3ooml0W4Qd0;ooo`020?ooo`030;md
001d_ol0oooo03D0oooo00L0omnL04P00018W=l0oooo0?oo_`1d0000M;oo01P0oooo000]0?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool0;@3oool00`3ogi`0B00007Boo`0g0?ooo`03
0=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0=@3oool01@000000oooo0?oo_`1d04P0W=oo
03P0oooo00@0oooO09a80000B9`0gooo>@3oool00`3ooml0W4Qd0;ooo`0d0?ooo`030?oog`2LB000
07Bo00<0oooo00<0okmd0018W03Oool05`3oool002d0oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`0^0?ooo`030?oOW0180000M;oo03P0oooo00<0gia807Boo`3oool0>P3oool01000
0000oono07@0001d_old0?ooo`030;md00000000000000D0000000<0M;oo0?ooo`3oool0=`3oool0
0`2oM000M;oo0?ooo`0c0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol05`3oool0
02d0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0^0?ooo`030?oOW0180000M;oo
03T0oooo00<0_gA809cOo`3oool0>@3oool010000000okmd0000001d_old0?ooo`080?oOW01dW=l0
oooo0?ooo`3ooml0W4P00018W03Oooli0?ooo`030?oog`2LB000B9cO03@0oooo00<0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`0G0?ooo`00;@3oool00`000000oooo0?ooo`060?ooo`030000
003oool0oooo02l0oooo00<0omnL04P0001d_ol0>@3oool00`3ooml0W4Qd0;ooo`0i0?ooo`040?oO
W0000000001809cOocD0oooo00L0gia807Boo`3oool0oooO09a80000B9`0gooo03T0oooo00<0gia8
0000001d_ol0=03oool00`3ogi`0B00007Boo`030?ooo`030;md000004P0W=oo01L0oooo000]0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool0;`3oool00`3ogi`0B00007Boo`0j0?oo
o`030?noM000M;l0oooo03H0oooo00D0okmd000000000000000009cOo`0f0?ooo`070?oog`2LB7@0
_ooo0?oog`2LB00004RL0=ooo`0e0?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`0d0?oo
o`080?oo_`1d0000000004RLg`3oool0okmd00000000M;lH0?ooo`00;@3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo0300oooo00<0omnL04P0001d_ol0>P3oool00`3ookl0M01809cO
o`0h0?ooo`030;md0000M;l0000003L0oooo00H0omnL04Qd_`3ooml0W4P00018W03Ooole0?ooo`06
0=nLB000000000000000000004P0W=oo=P3oool00`2oM00004RL0;nLB002000000030018W03Oool0
oooo01L0oooo000]0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0<@3oool00`3o
gi`0B00007Boo`0j0?ooo`030?oo_`1d0000B9cO03T0oooo00<0_g@00000003oool0=`3oool01@3O
W4P0M;oO09a80000B9`0gooo03D0oooo00@0oono07@000000000B9cO>03oool00`3ogi`0B00007Bo
o`0L0?ooo`00;@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0340oooo00<0omnL
04P0001d_ol0=@3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO03@0oooo00<0oooO
09bL_`3oool00P3oool00`3OW4P0001d0000000g0?ooo`050?oog`2LB4P0B4P00018W03Oool0=P3o
ool00`2oM000M;oo0?ooo`0i0?ooo`030=nLB00004P0W=oo01/0oooo000]0?ooo`<000000`3oool0
0`000000oooo0?ooo`0`0?ooo`050?oog`2LB00000000000001d_ol0=@3oool0203ooml0W4Q809cO
o`3oool0oono07@0000007@0_ooo=@3oool01`3o_g@0B9cO0?ooo`3ookl0M000001d_`000000>03o
ool0103ogi`0B0000018W03Ooolf0?ooo`060?oOW018000000000000000004P0W=oo=`3oool00`3o
_g@00000001d_`0J0?ooo`00;@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0380
oooo00<0okmd0000001d_ol0=P3oool01P3ooml0W4P0000000000000001d0;ooocL0oooo00L0gia8
000000000000001d0;ooo`3oool0000003P0oooo00<0gia80018W03Oool0=`3oool01@2oM0000000
000000000000B9cO03P0oooo00D0oooO09a800000000001d0;ooo`0G0?ooo`00;@3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0;00oooo00<000000?ooo`3oool0b03oool002d0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`2`0?ooo`030000003oool0oooo0<P0oooo
000]0?ooo`050000003oool0oooo0?ooo`000000]03oool00`000000oooo0?ooo`370?ooo`00;@3o
ool01@000000oooo0?ooo`3oool000000;@0oooo00<000000?ooo`3oool0a`3oool002d0oooo00@0
00000?ooo`3oool00000]P3oool00`000000oooo0?ooo`360?ooo`00;@3oool010000000oooo0?oo
o`00002f0?ooo`030000003oool0oooo0<H0oooo000]0?ooo`030000003oool000000;P0oooo00<0
00000?ooo`3oool0a@3oool002d0oooo00<000000?ooo`000000^03oool00`000000oooo0?ooo`35
0?ooo`00;@3oool200000;X0oooo00<000000?ooo`3oool0a03oool002d0oooo0P00002j0?ooo`03
0000003oool0oooo0<@0oooo000T0?ooool00000S@00000002d0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool01P3o
ool010000000oooo0?ooo`00000;0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2@3oool01@000000oooo0?ooo`3oool0000000L0oooo00<000000?ooo`3oool00@3oool002d0oooo
00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo03T0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo03P0oooo
00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00X0oooo
000]0?ooo`030000003oool0oooo0;/0oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?oo
o`0:0?ooo`00;@3oool00`000000oooo0?ooo`2l0?ooo`030000003oool0oooo0;<0oooo00<00000
0?ooo`3oool02`3oool002d0oooo00<000000?ooo`3oool0_03oool00`000000oooo0?ooo`2c0?oo
o`030000003oool0oooo00/0oooo000]0?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3o
ool0/@3oool00`000000oooo0?ooo`0<0?ooo`00;@3oool00`000000oooo0?ooo`2m0?ooo`030000
003oool0oooo0;40oooo00<000000?ooo`3oool0303oool002d0oooo00<000000?ooo`3oool0_P3o
ool00`000000oooo0?ooo`2_0?ooo`030000003oool0oooo00d0oooo000]0?ooo`030000003oool0
oooo0;h0oooo00<000000?ooo`3oool0[`3oool00`000000oooo0?ooo`0=0?ooo`00;@3oool00`00
0000oooo0?ooo`2o0?ooo`030000003oool0oooo0:d0oooo00<000000?ooo`3oool03P3oool002d0
oooo0`00002o0?ooo`030000003oool0oooo0:d0oooo00<000000?ooo`3oool03P3oool002d0oooo
00<000000?ooo`3oool0`03oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00l0oooo
000]0?ooo`030000003oool0oooo0<00oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?oo
o`0?0?ooo`00;@3oool00`000000oooo0?ooo`310?ooo`030000003oool0oooo0:T0oooo00<00000
0?ooo`3oool0403oool002d0oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`2Y0?oo
o`030000003oool0oooo0100oooo000]0?ooo`030000003oool0oooo0<80oooo00<000000?ooo`3o
ool0Y`3oool00`000000oooo0?ooo`0A0?ooo`00;@3oool00`000000oooo0?ooo`320?ooo`030000
003oool0oooo0:L0oooo00<000000?ooo`3oool04@3oool002d0oooo00<000000?ooo`3oool0``3o
ool00`000000oooo0?ooo`2U0?ooo`030000003oool0oooo0180oooo000]0?ooo`030000003oool0
oooo0<<0oooo00<000000?ooo`3oool0Y@3oool00`000000oooo0?ooo`0B0?ooo`00;@3oool00`00
0000oooo0?ooo`340?ooo`030000003oool0oooo0:<0oooo00<000000?ooo`3oool04`3oool002d0
oooo00<000000?ooo`3oool0a03oool00`000000oooo0?ooo`2S0?ooo`030000003oool0oooo01<0
oooo000]0?ooo`030000003oool0oooo0<D0oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`0D0?ooo`00;@3oool300000<D0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`0D0?ooo`00;@3oool00`000000oooo0?ooo`360?ooo`030000003oool0oooo09l0oooo00<00000
0?ooo`3oool05@3oool002d0oooo00<000000?ooo`3oool0aP3oool00`000000oooo0?ooo`2O0?oo
o`030000003oool0oooo01D0oooo000]0?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3o
ool0W@3oool00`000000oooo0?ooo`0F0?ooo`00;@3oool00`000000oooo0?ooo`370?ooo`030000
003oool0oooo09d0oooo00<000000?ooo`3oool05P3oool002d0oooo00<000000?ooo`3oool0b03o
ool00`000000oooo0?ooo`2K0?ooo`030000003oool0oooo01L0oooo000]0?ooo`030000003oool0
oooo0<P0oooo00<000000?ooo`3oool0V`3oool00`000000oooo0?ooo`0G0?ooo`00;@3oool00`00
0000oooo0?ooo`390?ooo`030000003oool0oooo09T0oooo00<000000?ooo`3oool0603oool002d0
oooo00<000000?ooo`3oool0bP3oool00`000000oooo0?ooo`2G0?ooo`030000003oool0oooo01T0
oooo000]0?ooo`030000003oool0oooo0<X0oooo00<000000?ooo`3oool0U`3oool00`000000oooo
0?ooo`0I0?ooo`00;@3oool00`000000oooo0?ooo`3;0?ooo`030000003oool0oooo09D0oooo00<0
00000?ooo`3oool06P3oool002d0oooo0`00003;0?ooo`030000003oool0oooo09D0oooo00<00000
0?ooo`3oool06P3oool002d0oooo00<000000?ooo`3oool0c03oool00`000000oooo0?ooo`2C0?oo
o`030000003oool0oooo01/0oooo000]0?ooo`030000003oool0oooo0<`0oooo00<000000?ooo`3o
ool0T`3oool00`000000oooo0?ooo`0K0?ooo`00;@3oool00`000000oooo0?ooo`3=0?ooo`030000
003oool0oooo0940oooo00<000000?ooo`3oool0703oool002d0oooo00<000000?ooo`3oool0cP3o
ool00`000000oooo0?ooo`2@0?ooo`030000003oool0oooo01`0oooo000]0?ooo`030000003oool0
oooo0<h0oooo00<000000?ooo`3oool0S`3oool00`000000oooo0?ooo`0M0?ooo`00;@3oool00`00
0000oooo0?ooo`3?0?ooo`030000003oool0oooo08h0oooo00<000000?ooo`3oool07@3oool002d0
oooo00<000000?ooo`3oool0c`3oool00`000000oooo0?ooo`2=0?ooo`030000003oool0oooo01h0
oooo000]0?ooo`030000003oool0oooo0=00oooo00<000000?ooo`3oool0S03oool00`000000oooo
0?ooo`0N0?ooo`00;@3oool00`000000oooo0?ooo`3@0?ooo`030000003oool0oooo08/0oooo00<0
00000?ooo`3oool07`3oool002d0oooo00<000000?ooo`3oool0d@3oool00`000000oooo0?ooo`29
0?ooo`030000003oool0oooo0200oooo000]0?ooo`030000003oool0oooo0=80oooo00<000000?oo
o`3oool0R03oool00`000000oooo0?ooo`0P0?ooo`00;@3oool300000=80oooo00<000000?ooo`3o
ool0Q`3oool00`000000oooo0?ooo`0Q0?ooo`00;@3oool00`000000oooo0?ooo`3C0?ooo`030000
003oool0oooo08H0oooo00<000000?ooo`3oool08@3oool002d0oooo00<000000?ooo`3oool0d`3o
ool00`000000oooo0?ooo`250?ooo`030000003oool0oooo0280oooo000]0?ooo`030000003oool0
oooo0=@0oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`0R0?ooo`00;@3oool00`00
0000oooo0?ooo`3D0?ooo`030000003oool0oooo08<0oooo00<000000?ooo`3oool08`3oool002d0
oooo00<000000?ooo`3oool0e@3oool00`000000oooo0?ooo`220?ooo`030000003oool0oooo02<0
oooo000]0?ooo`030000003oool0oooo0=H0oooo00<000000?ooo`3oool0P03oool00`000000oooo
0?ooo`0T0?ooo`004@3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00H0okmd0000001d_ol0
oooo0?oog`2LB0030000000307Boo`3oool0oooo00P0oooo00<000000?ooo`3oool0eP3oool00`00
0000oooo0?ooo`200?ooo`030000003oool0oooo02@0oooo000@0?ooo`070?oo_`1d0000M;oo0?oo
o`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo0?ooo`3ooml0W4Qd0;ooo`020?ooo`030;md
001d_ol0oooo00P0oooo00<000000?ooo`3oool0e`3oool00`000000oooo0?ooo`1n0?ooo`030000
003oool0oooo02D0oooo000@0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3o
ool00`3ooml0W4Qd0;ooo`080?ooo`030000003oool0oooo0=L0oooo00<000000?ooo`3oool0OP3o
ool00`000000oooo0?ooo`0U0?ooo`003`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0
_ooo00/0oooo00<0_g@007Boo`3oool01`3oool00`000000oooo0?ooo`3H0?ooo`030000003oool0
oooo07`0oooo00<000000?ooo`3oool09P3oool000l0oooo00<0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`0:0?ooo`030?oog`2LB000B9cO00P0oooo00<000000?ooo`3oool0f03oool00`00
0000oooo0?ooo`1l0?ooo`030000003oool0oooo02H0oooo000?0?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol02P3oool00`3OW4P0000007Boo`080?ooo`@00000f03oool00`000000
oooo0?ooo`1j0?ooo`030000003oool0oooo02L0oooo000?0?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol01P3oool01`3Ogml0oooo0?oo_`1d000000000000M02oool0203oool00`00
0000oooo0?ooo`3J0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0:03oool000l0
oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`060=nLB000000000000000
000004P0W=oo2@3oool00`000000oooo0?ooo`3J0?ooo`030000003oool0oooo07P0oooo00<00000
0?ooo`3oool0:03oool000l0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?oo
o`040?oo_`1d0000000004RLg`/0oooo00<000000?ooo`3oool0f`3oool00`000000oooo0?ooo`1f
0?ooo`030000003oool0oooo02T0oooo000@0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d
_`3oool01`3oool00`2oM000M;oo0?ooo`0;0?ooo`030000003oool0oooo0=`0oooo00<000000?oo
o`3oool0M03oool00`000000oooo0?ooo`0Z0?ooo`00403oool01`3ookl0M00007Boo`3oool0okmd
0000M02oool0203oool01P3ogi`0B000000000000000001809cOo`P0oooo00<000000?ooo`3oool0
g03oool00`000000oooo0?ooo`1d0?ooo`030000003oool0oooo02X0oooo000A0?ooo`050?oo_`1d
000000000000M02oool02P3oool01@2oM0000000000000000000B9cO00P0oooo00<000000?ooo`3o
ool0g@3oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo02/0oooo000]0?ooo`030000
003oool0oooo0=h0oooo00<000000?ooo`3oool0L03oool00`000000oooo0?ooo`0/0?ooo`00;@3o
ool00`000000oooo0?ooo`3N0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0;03o
ool002d0oooo00<000000?ooo`3oool0g`3oool00`000000oooo0?ooo`1^0?ooo`030000003oool0
oooo02d0oooo000]0?ooo`030000003oool0oooo0>00oooo00<000000?ooo`3oool0K03oool00`00
0000oooo0?ooo`0^0?ooo`00;@3oool300000>00oooo00<000000?ooo`3oool0K03oool00`000000
oooo0?ooo`0^0?ooo`00;@3oool00`000000oooo0?ooo`3Q0?ooo`030000003oool0oooo06X0oooo
00<000000?ooo`3oool0;`3oool002d0oooo00<000000?ooo`3oool0h@3oool00`000000oooo0?oo
o`1Y0?ooo`030000003oool0oooo0300oooo000]0?ooo`030000003oool0oooo0>80oooo00<00000
0?ooo`3oool0I`3oool00`000000oooo0?ooo`0a0?ooo`00;@3oool00`000000oooo0?ooo`3S0?oo
o`030000003oool0oooo06H0oooo00<000000?ooo`3oool0<@3oool002d0oooo00<000000?ooo`3o
ool0h`3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo0380oooo000]0?ooo`030000
003oool0oooo0>@0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0c0?ooo`00;@3o
ool00`000000oooo0?ooo`3U0?ooo`030000003oool0oooo0680oooo00<000000?ooo`3oool0<`3o
ool002d0oooo00<000000?ooo`3oool0i@3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0
oooo03@0oooo000]0?ooo`030000003oool0oooo0>H0oooo00<000000?ooo`3oool0G`3oool00`00
0000oooo0?ooo`0e0?ooo`00;@3oool00`000000oooo0?ooo`3W0?ooo`030000003oool0oooo05h0
oooo00<000000?ooo`3oool0=@3oool002d0oooo00<000000?ooo`3oool0i`3oool00`000000oooo
0?ooo`1M0?ooo`030000003oool0oooo03H0oooo000]0?ooo`<00000j03oool00`000000oooo0?oo
o`1K0?ooo`030000003oool0oooo03L0oooo000]0?ooo`030000003oool0oooo0>T0oooo00<00000
0?ooo`3oool0FP3oool00`000000oooo0?ooo`0g0?ooo`00;@3oool00`000000oooo0?ooo`3Z0?oo
o`030000003oool0oooo05P0oooo00<000000?ooo`3oool0>03oool002d0oooo00<000000?ooo`3o
ool0jP3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo03T0oooo000]0?ooo`030000
003oool0oooo0>/0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`0j0?ooo`00;@3o
ool00`000000oooo0?ooo`3/0?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool0>`3o
ool002d0oooo00<000000?ooo`3oool0k@3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0
oooo03`0oooo000]0?ooo`030000003oool0oooo0>h0oooo00<000000?ooo`3oool0D03oool00`00
0000oooo0?ooo`0l0?ooo`00;@3oool00`000000oooo0?ooo`3^0?ooo`030000003oool0oooo04l0
oooo00<000000?ooo`3oool0?@3oool002d0oooo00<000000?ooo`3oool0k`3oool00`000000oooo
0?ooo`1=0?ooo`030000003oool0oooo03h0oooo000]0?ooo`030000003oool0oooo0?00oooo00<0
00000?ooo`3oool0B`3oool00`000000oooo0?ooo`0o0?ooo`00;@3oool300000?40oooo00<00000
0?ooo`3oool0B@3oool00`000000oooo0?ooo`100?ooo`00;@3oool00`000000oooo0?ooo`3b0?oo
o`030000003oool0oooo04L0oooo00<000000?ooo`3oool0@@3oool002d0oooo00<000000?ooo`3o
ool0l`3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo0480oooo000]0?ooo`030000
003oool0oooo0?@0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`130?ooo`00;@3o
ool00`000000oooo0?ooo`3d0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0@`3o
ool002d0oooo00<000000?ooo`3oool0m@3oool00`000000oooo0?ooo`110?ooo`030000003oool0
oooo04@0oooo000]0?ooo`030000003oool0oooo0?H0oooo00<000000?ooo`3oool0?`3oool00`00
0000oooo0?ooo`150?ooo`00;@3oool00`000000oooo0?ooo`3g0?ooo`030000003oool0oooo03d0
oooo00<000000?ooo`3oool0AP3oool002d0oooo00<000000?ooo`3oool0n03oool00`000000oooo
0?ooo`0k0?ooo`030000003oool0oooo04L0oooo000]0?ooo`030000003oool0oooo0?T0oooo00<0
00000?ooo`3oool0>@3oool00`000000oooo0?ooo`180?ooo`00;@3oool00`000000oooo0?ooo`3j
0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0B@3oool002d0oooo00<000000?oo
o`3oool0n`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo04X0oooo000]0?ooo`<0
0000o03oool2000003@0oooo00<000000?ooo`3oool0B`3oool002d0oooo00<000000?ooo`3oool0
oP3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo04`0oooo000]0?ooo`030000003o
ool0oooo0?l0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`1=0?ooo`00;@3oool0
0`000000oooo0?ooo`3o0?ooo`40oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`1>
0?ooo`00;@3oool00`000000oooo0?ooo`3o0?ooo`80oooo0P00000Z0?ooo`800000D@3oool002d0
oooo00<000000?ooo`3oool0o`3oool40?ooo`030000003oool0oooo02D0oooo0P00001C0?ooo`00
;@3oool00`000000oooo0?ooo`3o0?ooo`D0oooo0P00000T0?ooo`030000003oool0oooo05<0oooo
000O0?ooo`060;md00000000000000000000B9`0gooo203oool00`000000oooo0?ooo`3o0?ooo`L0
oooo0P00000P0?ooo`800000EP3oool00200oooo00<0omnL04P0001d_ol02P3oool00`000000oooo
0?ooo`3o0?ooo`T0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`1F0?ooo`00803o
ool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo0?l0oooo2P3oool2000001X0oooo0P00
001I0?ooo`00803oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo0?l0oooo303oool4
00000180oooo1000001K0?ooo`00803oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo
0?l0oooo403oool5000000L0oooo1P00001O0?ooo`00803oool00`3ogi`0B00007Boo`0:0?ooo`@0
0000o`3ooolD0?ooo`L00000I@3oool00200oooo00<0omnL04P0001d_ol02P3oool00`000000oooo
0?ooo`3o0?oooh40oooo000P0?ooo`030?oOW0180000M;oo00X0oooo00<000000?ooo`3oool0o`3o
oon10?ooo`00803oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo0?l0ooooP@3oool0
0200oooo00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`3o0?oooh40oooo000N0?ooo`05
0?oog`2LB00000000000001d_ol02P3oool00`000000oooo0?ooo`3o0?oooh40oooo000P0?ooo`03
0?noM0000000M;oo0?l0ooooSP3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.767649, -1.14504, \
0.0250857, 0.01292}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Superimposing graphics objects", "Subsection"],

Cell[TextData[{
  "One can superimpose several graphics objects by supplying ",
  StyleBox["Show", "InlineInput",
    FontWeight->"Bold"],
  " with a list.  If ",
  StyleBox["DisplayFunction\[Rule]Identity", "InlineInput",
    FontWeight->"Bold"],
  " was used in one or more of the graphics objects input, we need to reset \
",
  StyleBox["DisplayFunction\[Rule]$DisplayFunction", "InlineInput",
    FontWeight->"Bold"],
  ". Note that the options chosen for the first plot in the list prevail.  \
Unfortunately, there is no simple way to change ",
  StyleBox["PlotStyle", "InlineInput",
    FontWeight->"Bold"],
  " for various elements of the list."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[{plot[x], plot[v], plot[a]}, 
        DisplayFunction \[Rule] $DisplayFunction];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.151576 0.309017 0.294302 [
[.17539 .29652 -5 -15.6875 ]
[.17539 .29652 5 0 ]
[.32696 .29652 -5 -15.6875 ]
[.32696 .29652 5 0 ]
[.47854 .29652 -5 -15.6875 ]
[.47854 .29652 5 0 ]
[.63011 .29652 -5 -15.6875 ]
[.63011 .29652 5 0 ]
[.78169 .29652 -5 -15.6875 ]
[.78169 .29652 5 0 ]
[.93327 .29652 -5 -15.6875 ]
[.93327 .29652 5 0 ]
[.01131 .01472 -18.3125 -7.84375 ]
[.01131 .01472 0 7.84375 ]
[.01131 .16187 -27.3125 -7.84375 ]
[.01131 .16187 0 7.84375 ]
[.01131 .45617 -19 -7.84375 ]
[.01131 .45617 0 7.84375 ]
[.01131 .60332 -10 -7.84375 ]
[.01131 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.17539 .30902 m
.17539 .31527 L
s
gsave
.17539 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.32696 .30902 m
.32696 .31527 L
s
gsave
.32696 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.47854 .30902 m
.47854 .31527 L
s
gsave
.47854 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.63011 .30902 m
.63011 .31527 L
s
gsave
.63011 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.78169 .30902 m
.78169 .31527 L
s
gsave
.78169 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.93327 .30902 m
.93327 .31527 L
s
gsave
.93327 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.125 Mabswid
.05412 .30902 m
.05412 .31277 L
s
.08444 .30902 m
.08444 .31277 L
s
.11476 .30902 m
.11476 .31277 L
s
.14507 .30902 m
.14507 .31277 L
s
.2057 .30902 m
.2057 .31277 L
s
.23602 .30902 m
.23602 .31277 L
s
.26633 .30902 m
.26633 .31277 L
s
.29665 .30902 m
.29665 .31277 L
s
.35728 .30902 m
.35728 .31277 L
s
.38759 .30902 m
.38759 .31277 L
s
.41791 .30902 m
.41791 .31277 L
s
.44822 .30902 m
.44822 .31277 L
s
.50885 .30902 m
.50885 .31277 L
s
.53917 .30902 m
.53917 .31277 L
s
.56948 .30902 m
.56948 .31277 L
s
.5998 .30902 m
.5998 .31277 L
s
.66043 .30902 m
.66043 .31277 L
s
.69074 .30902 m
.69074 .31277 L
s
.72106 .30902 m
.72106 .31277 L
s
.75138 .30902 m
.75138 .31277 L
s
.81201 .30902 m
.81201 .31277 L
s
.84232 .30902 m
.84232 .31277 L
s
.87264 .30902 m
.87264 .31277 L
s
.90295 .30902 m
.90295 .31277 L
s
.96358 .30902 m
.96358 .31277 L
s
.9939 .30902 m
.9939 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.02381 .01472 m
.03006 .01472 L
s
gsave
.01131 .01472 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .16187 m
.03006 .16187 L
s
gsave
.01131 .16187 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .45617 m
.03006 .45617 L
s
gsave
.01131 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .60332 m
.03006 .60332 L
s
gsave
.01131 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04415 m
.02756 .04415 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .10301 m
.02756 .10301 L
s
.02381 .13244 m
.02756 .13244 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .25016 m
.02756 .25016 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .36788 m
.02756 .36788 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .4856 m
.02756 .4856 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .60332 m
.02499 .60331 L
.02605 .60329 L
.02729 .60324 L
.02846 .60318 L
.03053 .60303 L
.03279 .6028 L
.03527 .60248 L
.0379 .60205 L
.04262 .60106 L
.04749 .59974 L
.05205 .59822 L
.06244 .59381 L
.07305 .58793 L
.08274 .58135 L
.10458 .56251 L
.14429 .51515 L
.18248 .45627 L
.22313 .38349 L
.26226 .30833 L
.30384 .22863 L
.34391 .15746 L
.38246 .09886 L
.402 .07411 L
.42346 .05145 L
.44388 .03466 L
.45301 .02874 L
.46295 .02346 L
.46838 .0211 L
.4734 .01924 L
.4781 .01778 L
.4833 .0165 L
.48553 .01605 L
.48794 .01565 L
.49002 .01535 L
.4923 .0151 L
.49473 .01489 L
.49605 .01482 L
.49728 .01476 L
.49844 .01473 L
.49949 .01472 L
.50071 .01472 L
.50186 .01474 L
.50301 .01477 L
.50425 .01483 L
.50542 .0149 L
.50648 .01498 L
.5091 .01524 L
.51158 .01557 L
.5165 .01646 L
Mistroke
.52187 .01777 L
.53144 .02102 L
.54032 .02507 L
.56088 .03814 L
.5795 .05428 L
.61962 .1017 L
.65822 .16102 L
.69928 .23447 L
.73882 .31042 L
.77684 .38344 L
.81732 .45593 L
.85628 .51593 L
.87613 .54149 L
.89769 .56473 L
.91843 .58221 L
.92832 .58876 L
.93759 .59383 L
.94643 .59767 L
.95595 .6007 L
.9611 .60186 L
.96369 .60232 L
.96652 .60272 L
.96894 .60298 L
.9703 .6031 L
.97153 .60318 L
.97268 .60324 L
.97393 .60329 L
.97511 .60331 L
.97619 .60332 L
Mfstroke
.02381 .30902 m
.06244 .23481 L
.10458 .15951 L
.14415 .09914 L
.18221 .05447 L
.20178 .03757 L
.21267 .03011 L
.22272 .0245 L
.23293 .02007 L
.23832 .01827 L
.24402 .01676 L
.24641 .01625 L
.24897 .01579 L
.25119 .01545 L
.25361 .01516 L
.25507 .01501 L
.25643 .01491 L
.25773 .01483 L
.2584 .01479 L
.25912 .01476 L
.25983 .01474 L
.26048 .01473 L
.26176 .01472 L
.26293 .01472 L
.26416 .01475 L
.26538 .01479 L
.26607 .01483 L
.26671 .01486 L
.26912 .01505 L
.27131 .01528 L
.27364 .0156 L
.27847 .01647 L
.2838 .01778 L
.29326 .02099 L
.30213 .02502 L
.32232 .03778 L
.34082 .05371 L
.38045 .10022 L
.41855 .15838 L
.45912 .23059 L
.49816 .30544 L
.53966 .38514 L
.57964 .45663 L
.61811 .51582 L
.63759 .54096 L
.65902 .56418 L
.67938 .58151 L
.69843 .5933 L
.70884 .59785 L
.71396 .5996 L
Mistroke
.71872 .60092 L
.72301 .60186 L
.72758 .60261 L
.7302 .60292 L
.73148 .60304 L
.73266 .60313 L
.73371 .6032 L
.73487 .60325 L
.73603 .60329 L
.73725 .60331 L
.73856 .60332 L
.73977 .6033 L
.7404 .60328 L
.7411 .60326 L
.74251 .60319 L
.74375 .60311 L
.74508 .60301 L
.74748 .60275 L
.74985 .60243 L
.75235 .60202 L
.7568 .60108 L
.76198 .59967 L
.7668 .59806 L
.77773 .59332 L
.78828 .58734 L
.79823 .58046 L
.81683 .5645 L
.85534 .51957 L
.89632 .45702 L
.93577 .38657 L
.97371 .31383 L
.97619 .30902 L
Mfstroke
.02381 .01472 m
.02499 .01472 L
.02605 .01475 L
.02729 .01479 L
.02846 .01485 L
.03053 .015 L
.03279 .01523 L
.03527 .01556 L
.0379 .01599 L
.04262 .01698 L
.04749 .0183 L
.05205 .01981 L
.06244 .02422 L
.07305 .03011 L
.08274 .03668 L
.10458 .05552 L
.14429 .10289 L
.18248 .16177 L
.22313 .23454 L
.26226 .3097 L
.30384 .3894 L
.34391 .46058 L
.38246 .51917 L
.402 .54392 L
.42346 .56659 L
.44388 .58338 L
.45301 .58929 L
.46295 .59457 L
.46838 .59694 L
.4734 .5988 L
.4781 .60025 L
.4833 .60153 L
.48553 .60198 L
.48794 .60239 L
.49002 .60268 L
.4923 .60294 L
.49473 .60314 L
.49605 .60322 L
.49728 .60327 L
.49844 .6033 L
.49949 .60332 L
.50071 .60332 L
.50186 .6033 L
.50301 .60326 L
.50425 .6032 L
.50542 .60313 L
.50648 .60305 L
.5091 .60279 L
.51158 .60246 L
.5165 .60158 L
Mistroke
.52187 .60026 L
.53144 .59701 L
.54032 .59297 L
.56088 .57989 L
.5795 .56376 L
.61962 .51633 L
.65822 .45702 L
.69928 .38356 L
.73882 .30761 L
.77684 .2346 L
.81732 .1621 L
.85628 .1021 L
.87613 .07654 L
.89769 .0533 L
.91843 .03582 L
.92832 .02927 L
.93759 .02421 L
.94643 .02037 L
.95595 .01733 L
.9611 .01617 L
.96369 .01572 L
.96652 .01531 L
.96894 .01505 L
.9703 .01494 L
.97153 .01485 L
.97268 .01479 L
.97393 .01475 L
.97511 .01472 L
.97619 .01472 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000N0?ooo`060;md0000000000000000
0000B9`0gooo2@3oool00`000000oooo0?ooo`3o0?oooh40oooo000O0?ooo`030?oOW0180000M;oo
00/0oooo00<000000?ooo`3oool0o`3ooon10?ooo`007`3oool00`3ogi`0B00007Boo`0;0?ooo`03
0000003oool0oooo0?l0ooooP@3oool001l0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo
0?ooo`3o0?oooh40oooo000@0?ooo`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool00P3o
ool00`3ogi`0B00007Boo`0;0?ooo`@00000F@3oool4000005X0oooo0`00002i0?ooo`8000002P3o
ool001l0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo00000006000004h0oooo1`000002
0?ooo`H00000D03oool600000080oooo1P00002/0?ooo`L00000303oool001l0oooo00<0omnL04P0
001d_ol02`3oool00`000000oooo0?ooo`060?ooo`@00000B03oool2000000l0oooo0`0000190?oo
o`@000003P3oool300000:L0oooo0P00000C0?ooo`007`3oool00`3ogi`0B00007Boo`0;0?ooo`03
0000003oool0oooo00X0oooo0P0000130?ooo`<00000503oool3000004@0oooo0P00000E0?ooo`<0
0000X@3oool3000001D0oooo000O0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0
303oool2000003h0oooo0`00000J0?ooo`800000@03oool2000001X0oooo0P00002M0?ooo`800000
603oool001l0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo03X0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0k0?ooo`8000007P3o
ool00`000000oooo0?ooo`2H0?ooo`8000006P3oool001d0oooo00D0oooO09a800000000000007Bo
o`0;0?ooo`030000003oool0oooo00l0oooo0P00000h0?ooo`8000008@3oool2000003T0oooo0P00
000Q0?ooo`800000UP3oool2000001`0oooo000O0?ooo`030?noM0000000M;oo00/0oooo00<00000
0?ooo`3oool04@3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo02<0oooo00<00000
0?ooo`3oool0=03oool2000002D0oooo00<000000?ooo`3oool0TP3oool00`000000oooo0?ooo`0L
0?ooo`00;@3oool00`000000oooo0?ooo`0B0?ooo`800000<P3oool2000002L0oooo0P00000b0?oo
o`800000:03oool200000900oooo0P00000O0?ooo`00;@3oool00`000000oooo0?ooo`0D0?ooo`80
0000;P3oool2000002/0oooo0P00000_0?ooo`030000003oool0oooo02X0oooo0P00002=0?ooo`03
0000003oool0oooo01l0oooo000]0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0
:P3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0
;@3oool00`000000oooo0?ooo`290?ooo`030000003oool0oooo0200oooo000]0?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0_0?ooo`030000003o
ool0oooo02T0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`270?ooo`030000003o
ool0oooo0240oooo000]0?ooo`<00000603oool00`000000oooo0?ooo`0V0?ooo`030000003oool0
oooo0340oooo00<000000?ooo`3oool09P3oool2000003<0oooo0P0000260?ooo`030000003oool0
oooo0280oooo000]0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool08`3oool20000
03D0oooo0P00000U0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0P@3oool20000
02D0oooo000]0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool08@3oool00`000000
oooo0?ooo`0g0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0=`3oool00`000000
oooo0?ooo`1o0?ooo`030000003oool0oooo02D0oooo000]0?ooo`030000003oool0oooo01/0oooo
00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo01l0oooo
00<000000?ooo`3oool0>@3oool2000007h0oooo00<000000?ooo`3oool09P3oool002d0oooo00<0
00000?ooo`3oool0703oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo03/0oooo00<0
00000?ooo`3oool07@3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo07X0oooo00<0
00000?ooo`3oool09`3oool002d0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0K
0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0n
0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0:03oool002d0oooo00<000000?oo
o`3oool07P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo03l0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo07H0oooo00<000000?oo
o`3oool0:@3oool002d0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0G0?ooo`03
0000003oool0oooo0440oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`110?ooo`03
0000003oool0oooo07D0oooo00<000000?ooo`3oool0:P3oool002d0oooo00<000000?ooo`3oool0
803oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0
5P3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool0
:`3oool002d0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0D0?ooo`030000003o
ool0oooo04<0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`140?ooo`030000003o
ool0oooo0780oooo00<000000?ooo`3oool0:`3oool002d0oooo00<000000?ooo`3oool08@3oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`160?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0;03oool0
02d0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
04L0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo
06h0oooo00<000000?ooo`3oool0;@3oool002d0oooo0`00000S0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo04T0
oooo00<000000?ooo`3oool0K@3oool00`000000oooo0?ooo`0^0?ooo`00;@3oool00`000000oooo
0?ooo`0T0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0B`3oool00`000000oooo
0?ooo`0=0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0J`3oool00`000000oooo
0?ooo`0_0?ooo`00;@3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool0C03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo04d0oooo00<0
00000?ooo`3oool0J@3oool00`000000oooo0?ooo`0`0?ooo`00;@3oool00`000000oooo0?ooo`0V
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`0:
0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`0a
0?ooo`00;@3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo00P0oooo00<000000?oo
o`3oool0D03oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo04l0oooo00<000000?oo
o`3oool0I`3oool00`000000oooo0?ooo`0a0?ooo`00;@3oool00`000000oooo0?ooo`0X0?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo0540oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`0b0?ooo`00
;@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
D`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`0c0?ooo`00;@3oool00`000000oooo0?ooo`0Y0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo05D0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`0d0?ooo`00;@3oool0
0`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0EP3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo05H0oooo00<000000?ooo`3oool0G`3oool0
0`000000oooo0?ooo`0e0?ooo`00;@3oool00`000000oooo0?ooo`0[0?ooo`040000003oool0oooo
000005X0oooo00@000000?ooo`3oool00000F@3oool00`000000oooo0?ooo`1O0?ooo`030000003o
ool0oooo03D0oooo000]0?ooo`030000003oool0oooo02`0oooo0P00001L0?ooo`800000F`3oool0
0`000000oooo0?ooo`1M0?ooo`030000003oool0oooo03H0oooo000]0?ooo`<00000;03oool20000
05`0oooo0P00001L0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0=`3oool002d0
oooo00<000000?ooo`3oool0;03oool00`000000oooo0000001K0?ooo`800000G@3oool00`000000
oooo0?ooo`1I0?ooo`030000003oool0oooo03P0oooo000]0?ooo`030000003oool0oooo02/0oooo
00@000000?ooo`3oool00000FP3oool010000000oooo0?ooo`00001L0?ooo`030000003oool0oooo
05T0oooo00<000000?ooo`3oool0>03oool002d0oooo00<000000?ooo`3oool0:P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo05H0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`1J0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0>@3oool002d0oooo
00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo05D0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo05D0oooo
00<000000?ooo`3oool0>P3oool002d0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`1J0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0>P3oool002d0oooo00<00000
0?ooo`3oool0:03oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0580oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo05<0oooo00<00000
0?ooo`3oool0>`3oool002d0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo0540oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`1K0?oo
o`030000003oool0oooo0540oooo00<000000?ooo`3oool0?03oool002d0oooo00<000000?ooo`3o
ool09`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo0500oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3o
ool0?03oool002d0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0:0?ooo`030000
003oool0oooo04h0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`1J0?ooo`030000
003oool0oooo04l0oooo00<000000?ooo`3oool0?@3oool002d0oooo00<000000?ooo`3oool09P3o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3oool0?P3o
ool002d0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo04`0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1K0?ooo`030000003oool0
oooo04`0oooo00<000000?ooo`3oool0?P3oool002d0oooo0`00000T0?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo
05/0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`0o0?ooo`00;@3oool00`000000
oooo0?ooo`0T0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0BP3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0B@3oool00`000000
oooo0?ooo`100?ooo`00;@3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo05/0oooo
00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`100?ooo`00;@3oool00`000000oooo0?oo
o`0R0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?oo
o`110?ooo`00;@3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool0AP3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo05/0oooo00<00000
0?ooo`3oool0A@3oool00`000000oooo0?ooo`120?ooo`00;@3oool00`000000oooo0?ooo`0Q0?oo
o`030000003oool0oooo01@0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`0D0?oo
o`030000003oool0oooo05/0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`120?oo
o`00;@3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3o
ool0@`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3o
ool0@P3oool00`000000oooo0?ooo`130?ooo`004@3oool01@3ogi`0B0000000000007@0_ooo00<0
oooo00H0okmd0000001d_ol0oooo0?oog`2LB0030000000307Boo`3oool0oooo00P0oooo00<00000
0?ooo`3oool0803oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0480oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo0440oooo00<00000
0?ooo`3oool0A03oool00100oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00L0
okmd0000001d_ol0oooo0?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool0203oool00`000000
oooo0?ooo`0O0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0@03oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0@03oool00`000000
oooo0?ooo`140?ooo`00403oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo
00<0oooO09a8M02oool0203oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo05`0oooo
00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`150?ooo`003`3oool00`3ookl0M00004RL
g`030?ooo`030=nLB00007@0_ooo00/0oooo00<0_g@007Boo`3oool01`3oool00`000000oooo0?oo
o`0N0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?oo
o`160?ooo`000`3oool00`3ooml0W4P0000000070000000504RLg`3oool0omnL04P0001d_ol00`3o
ool00`3o_g@0001809cOo`0:0?ooo`030?oog`2LB000B9cO00P0oooo00<000000?ooo`3oool07@3o
ool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0703o
ool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0AP3o
ool000l0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030=nLB0000000
M;oo00P0oooo1000000K0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0>`3oool0
0`000000oooo0?ooo`0L0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3oool0>P3oool0
0`000000oooo0?ooo`170?ooo`003`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00P0oooo00<000000?ooo`3oool0703o
ool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool07P3o
ool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0B03o
ool000l0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`060=nLB0000000
00000000000004P0W=oo2@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo0200oooo
00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo05`0oooo
00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`180?ooo`003`3oool00`3ookl0M00004RL
g`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono07@000000000B9cO2`3oool00`000000oooo
0?ooo`0K0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0>03oool00`000000oooo
0?ooo`0P0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0=`3oool00`000000oooo
0?ooo`190?ooo`00403oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0
_g@007Boo`3oool02`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0280oooo00<0
00000?ooo`3oool0=`3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo05`0oooo00<0
00000?ooo`3oool0=P3oool00`000000oooo0?ooo`190?ooo`00403oool01`3ookl0M00007Boo`3o
ool0okmd0000M02oool0203oool01P3ogi`0B000000000000000001809cOo`P0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo03H0oooo00<000000?oo
o`3oool08P3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo03D0oooo00<000000?oo
o`3oool0BP3oool00140oooo00D0oono07@000000000001d0;ooo`0:0?ooo`050;md000000000000
00000018W=l0203oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo02@0oooo00<00000
0?ooo`3oool0=03oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo05`0oooo00<00000
0?ooo`3oool0=03oool00`000000oooo0?ooo`1:0?ooo`00;@3oool00`000000oooo0?ooo`0I0?oo
o`030000003oool0oooo02@0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?oo
o`030000003oool0oooo05/0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`1;0?oo
o`00;@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3o
ool0<P3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3o
ool0<@3oool00`000000oooo0?ooo`1<0?ooo`00;@3oool00`000000oooo0?ooo`0G0?ooo`030000
003oool0oooo02P0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0V0?ooo`030000
003oool0oooo05`0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`1<0?ooo`00;@3o
ool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0<03o
ool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0;`3o
ool00`000000oooo0?ooo`1=0?ooo`00;@3oool3000001H0oooo00<000000?ooo`3oool0:P3oool0
0`000000oooo0?ooo`0_0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0G03oool0
0`000000oooo0?ooo`0^0?ooo`030000003oool0oooo04d0oooo000]0?ooo`030000003oool0oooo
01H0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo
02X0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo
04h0oooo000]0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0;03oool00`000000
oooo0?ooo`0/0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0G03oool00`000000
oooo0?ooo`0/0?ooo`030000003oool0oooo04h0oooo000]0?ooo`030000003oool0oooo01D0oooo
00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo02`0oooo
00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo04l0oooo
000]0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?oo
o`0Z0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0G03oool00`000000oooo0?oo
o`0Y0?ooo`030000003oool0oooo0500oooo000]0?ooo`030000003oool0oooo01@0oooo00<00000
0?ooo`3oool0;P3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo02h0oooo00<00000
0?ooo`3oool0F`3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0500oooo000]0?oo
o`030000003oool0oooo01<0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0X0?oo
o`030000003oool0oooo02l0oooo00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`0W0?oo
o`030000003oool0oooo0540oooo000]0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3o
ool0<03oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3o
ool0F`3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo0540oooo000]0?ooo`030000
003oool0oooo0180oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0V0?ooo`030000
003oool0oooo0380oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0U0?ooo`030000
003oool0oooo0580oooo000]0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0=03o
ool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0G03o
ool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo0580oooo000]0?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`030000003oool0
oooo03@0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0
oooo05<0oooo000]0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0=P3oool00`00
0000oooo0?ooo`0R0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0G03oool00`00
0000oooo0?ooo`0Q0?ooo`030000003oool0oooo05@0oooo000]0?ooo`<00000403oool00`000000
oooo0?ooo`0f0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0=P3oool00`000000
oooo0?ooo`1K0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0E03oool002d0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo0200oooo
00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo01l0oooo
00<000000?ooo`3oool0E@3oool002d0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`0h0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0>03oool00`000000oooo0?oo
o`1K0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0E@3oool002d0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo01h0oooo00<00000
0?ooo`3oool0>@3oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo01d0oooo00<00000
0?ooo`3oool0EP3oool002d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0j0?oo
o`030000003oool0oooo01d0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`1K0?oo
o`030000003oool0oooo01d0oooo00<000000?ooo`3oool0EP3oool002d0oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3o
ool0?03oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3o
ool0E`3oool002d0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0m0?ooo`030000
003oool0oooo01/0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`1K0?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool0F03oool002d0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0?P3o
ool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0F03o
ool002d0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0
oooo01T0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0
oooo01P0oooo00<000000?ooo`3oool0F@3oool002d0oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`100?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0@03oool00`00
0000oooo0?ooo`1K0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0F@3oool002d0
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo01L0
oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo01H0
oooo00<000000?ooo`3oool0FP3oool002d0oooo0`00000:0?ooo`030000003oool0oooo0480oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo05/0oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`1J0?ooo`00;@3oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?oo
o`130?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0503oool00`000000oooo0?oo
o`1K0?ooo`00;@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo02/0oooo00H0_g@0
00000000000000000018W03OoolC0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0
1P3oool00`2oM0000000000000040000000304RLg`3oool0oooo0380oooo00<0oono07@000000000
0P0000000`1d_ol0oooo0?ooo`0h0?ooo`040?oog`2LB00004RL0=oooa`0oooo00<000000?ooo`3o
ool04`3oool010000000oooo0?oog`2LB0030000000307Boo`3oool0oooo03H0oooo00D0omnL04P0
00000000000007Boo`0I0?ooo`00;@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo
02d0oooo00<0omnL04P0001d_ol05@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo
00P0oooo00<0gia80000000000000`0000000`00B9`0gooo0?ooo`0b0?ooo`070?oOW0180000B9cO
0000003oool0_g@004RLg`0i0?ooo`040?oog`2LB00004RL0=oooa`0oooo00<000000?ooo`3oool0
4P3oool01P000000oooo0?ooo`3ooml0W4Qd0;ooo`80oooo00<0_g@007Boo`3oool0=@3oool01`3o
gi`0B00004RLg`3oool0oono07@0001d_ol0603oool002d0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`0]0?ooo`030?oOW0180000M;oo01H0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0:0?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0=@3oool0
1@000000oooo0?oo_`1d04P0W=oo03P0oooo00@0oooO09a80000B9`0gooo7@3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo00D0oooo00<0oooO09a8M02oool0=03oool00`3ooml0W4P0
001d_`030?ooo`030?noM000B9`0gooo01L0oooo000]0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0;P3oool00`3ogi`0B00007Boo`0F0?ooo`030000003oool0oooo0140oooo00<0
00000?ooo`3oool02`3oool00`3OW4P0M;oo0?ooo`0j0?ooo`040000003ookl0M00007Booc@0oooo
00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`0J0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool01`3oool00`2oM000M;oo0?ooo`0c0?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol05`3oool002d0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`0^0?ooo`030?oOW0180000M;oo01L0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`0=0?ooo`030;mdB02Lgol0oooo03T0oooo00@000000?noM0000000M;oo=03oool0203ogi`0M9cO
0?ooo`3oool0oooO09a80000B9`0gooo7P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00H0oooo00<0oooO09a80018W=l0=03oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo01L0oooo000]0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0;`3oool00`3o
gi`0B00007Boo`0G0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03@3oool00`3o
oml0W4Qd0;ooo`0i0?ooo`040?oOW0000000001809cOocD0oooo00L0gia807Boo`3oool0oooO09a8
0000B9`0gooo01h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`070?ooo`030=nL
B0000000M;oo03@0oooo00<0omnL04P0001d_ol00`3oool00`2oM000001809cOo`0G0?ooo`00;@3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo02l0oooo00<0omnL04P0001d_ol0603o
ool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00l0oooo00<0okmd001d_`3oool0=P3o
ool01@3o_g@00000000000000000W=oo03H0oooo00L0oooO09a8M02oool0oooO09a80000B9`0gooo
01l0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`030?ooo`070=oOg`3oool0oono
07@000000000001d0;ooo`0d0?ooo`080?oo_`1d0000000004RLg`3oool0okmd00000000M;lH0?oo
o`00;@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0300oooo00<0omnL04P0001d
_ol0603oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00l0oooo00<0oono07@0B02L
gol0>03oool00`2oM00007Bo0000000g0?ooo`060?oOW018M;l0oooO09a80000B9`0gooo7`3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00@0oooo00H0gia800000000000000000000
B02Lgolf0?ooo`030;md0000B9`0_ia80080000000<004RL0=ooo`3oool05`3oool002d0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`0a0?ooo`030?oOW0180000M;oo01T0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0@0?ooo`030?oo_`1d0000B9cO03T0oooo00<0
_g@00000003oool0=`3oool01@3OW4P0M;oO09a80000B9`0gooo0200oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`040?ooo`040?oo_`1d0000000004RLgcP0oooo00<0omnL04P0001d
_ol0703oool002d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0a0?ooo`030?oO
W0180000M;oo01T0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030?no
W02Lgol0oooo0080oooo00<0omnL04P00018W=l0=03oool00`3ooml0W9bo0?ooo`020?ooo`030=nL
B00007@0000003L0oooo00D0oooO09a8B018B00004RL0=ooo`0P0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool01P3oool00`2oM000M;oo0?ooo`0i0?ooo`030=nLB00004P0W=oo01/0
oooo000]0?ooo`<000000`3oool00`000000oooo0?ooo`0`0?ooo`050?oog`2LB00000000000001d
_ol06P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00`0oooo00P0oooO09a8B02L
gol0oooo0?oo_`1d0000001d0;ooocD0oooo00L0okmd04RLg`3oool0oono07@00000M;l0000003P0
oooo00@0omnL04P00000B9`0gooo8@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo
00H0oooo00H0omnL04P00000000000000000B02Lgolg0?ooo`030?noM000000007Bo01X0oooo000]
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0<P3oool00`3o_g@0000007Boo`0J
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool03P3oool01P3ooml0W4P000000000
0000001d0;ooocL0oooo00L0gia8000000000000001d0;ooo`3oool0000003P0oooo00<0gia80018
W03Oool08@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00D0_g@00000
00000000000004RLg`0h0?ooo`050?oog`2LB00000000000M02oool05`3oool002d0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool0DP3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool0H@3oool002d0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1A0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`1J0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool0HP3oool002d0oooo00D000000?ooo`3o
ool0oooo0000001E0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0E03oool00`00
0000oooo0?ooo`1J0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0HP3oool002d0
oooo00D000000?ooo`3oool0oooo0000001E0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0E@3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0H`3oool002d0oooo00@000000?ooo`3oool00000E`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo05H0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo06<0oooo000]0?ooo`040000003oool0oooo000005L0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo05X0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`1T0?ooo`00;@3oool00`000000oooo0000001I0?ooo`05
0000003oool0oooo0?ooo`000000FP3oool00`000000oooo0?ooo`1J0?ooo`050000003oool0oooo
0?ooo`000000IP3oool002d0oooo00<000000?ooo`000000F@3oool010000000oooo0?ooo`00001K
0?ooo`030000003oool0oooo05X0oooo00@000000?ooo`3oool00000I`3oool002d0oooo0P00001K
0?ooo`030000003oool0000005`0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0000001W
0?ooo`00;@3oool2000005/0oooo0P00001M0?ooo`030000003oool0oooo05X0oooo0P00001X0?oo
o`00903ooooo000008d00000000]0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
1P3oool010000000oooo00000000000;0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00H0oooo00@000000?ooo`3o
ool000002`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`030?ooo`030000003oool0000000<0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool02@3oool01@000000oooo0?ooo`3oool0000000L0oooo00<000000?ooo`3oool00@3oool0
02d0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0M0?ooo`040000003oool0oooo
000001P0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo02l0oooo00<000000?ooo`3oool09@3oool010000000oooo0?ooo`00000?0?ooo`030000003o
ool0oooo03T0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0:0?ooo`00;@3oool0
0`000000oooo0?ooo`1I0?ooo`050000003oool0oooo0?ooo`000000G@3oool00`000000oooo0?oo
o`1G0?ooo`050000003oool0oooo0?ooo`000000F@3oool00`000000oooo0?ooo`0:0?ooo`00;@3o
ool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0G03o
ool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0EP3o
ool00`000000oooo0?ooo`0;0?ooo`00;@3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`1E0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`0;0?ooo`00;@3oool00`00
0000oooo0?ooo`1G0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0G03oool00`00
0000oooo0?ooo`1C0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0E03oool00`00
0000oooo0?ooo`0<0?ooo`00;@3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`0<0?ooo`00;@3oool00`000000oooo
0?ooo`1F0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0G03oool00`000000oooo
0?ooo`1A0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0DP3oool00`000000oooo
0?ooo`0=0?ooo`00;@3oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0F`3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0D@3oool00`000000oooo0?ooo`0=0?ooo`00;@3oool00`000000oooo0?ooo`1E
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`1?
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`0>
0?ooo`00;@3oool3000005D0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`1K0?oo
o`030000003oool0oooo04l0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`1?0?oo
o`030000003oool0oooo00h0oooo000]0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo00l0oooo000]0?ooo`030000
003oool0oooo05@0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1K0?ooo`030000
003oool0oooo04d0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1=0?ooo`030000
003oool0oooo00l0oooo000]0?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo0100oooo000]0?ooo`030000003oool0
oooo05<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0
oooo04/0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0
oooo0100oooo000]0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool03`3oool00`00
0000oooo0?ooo`1K0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool03P3oool00`00
0000oooo0?ooo`1:0?ooo`030000003oool0oooo0140oooo000]0?ooo`030000003oool0oooo0580
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo04T0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo0140
oooo000]0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`1K0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0403oool00`000000oooo
0?ooo`180?ooo`030000003oool0oooo0180oooo000]0?ooo`030000003oool0oooo0540oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo04L0oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo0180oooo000]
0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`1K
0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`16
0?ooo`030000003oool0oooo01<0oooo000]0?ooo`030000003oool0oooo0500oooo00<000000?oo
o`3oool0503oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo04D0oooo00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo01<0oooo000]0?ooo`03
0000003oool0oooo04l0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`1K0?ooo`03
0000003oool0oooo04<0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`140?ooo`03
0000003oool0oooo01@0oooo000]0?ooo`<00000C`3oool00`000000oooo0?ooo`0F0?ooo`030000
003oool0oooo05X0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`0E0?ooo`030000
003oool0oooo04<0oooo00<000000?ooo`3oool0503oool002d0oooo00<000000?ooo`3oool0CP3o
ool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0@@3o
ool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool05@3o
ool002d0oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0
oooo05X0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
oooo0440oooo00<000000?ooo`3oool05@3oool002d0oooo00<000000?ooo`3oool0C@3oool00`00
0000oooo0?ooo`0I0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0?`3oool00`00
0000oooo0?ooo`0H0?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool05P3oool002d0
oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo05X0
oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo03l0
oooo00<000000?ooo`3oool05P3oool002d0oooo00<000000?ooo`3oool0C03oool00`000000oooo
0?ooo`0K0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0?@3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool05`3oool002d0oooo00<0
00000?ooo`3oool0C03oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo05X0oooo00<0
00000?ooo`3oool0?@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo03d0oooo00<0
00000?ooo`3oool05`3oool002d0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`0M
0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`0M
0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0603oool002d0oooo00<000000?oo
o`3oool0B`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo05/0oooo00<000000?oo
o`3oool0>P3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo03X0oooo00<000000?oo
o`3oool06@3oool002d0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`03
0000003oool0oooo05X0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0O0?ooo`03
0000003oool0oooo03T0oooo00<000000?ooo`3oool06@3oool002d0oooo00<000000?ooo`3oool0
BP3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0
>03oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool0
6P3oool002d0oooo0`0000190?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0FP3o
ool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0=`3o
ool00`000000oooo0?ooo`0J0?ooo`00;@3oool00`000000oooo0?ooo`190?ooo`030000003oool0
oooo0280oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0
oooo0240oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0K0?ooo`00;@3oool00`00
0000oooo0?ooo`180?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0FP3oool00`00
0000oooo0?ooo`0e0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0=@3oool00`00
0000oooo0?ooo`0K0?ooo`00;@3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo02@0
oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo02<0
oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0L0?ooo`00;@3oool00`000000oooo
0?ooo`170?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0F`3oool00`000000oooo
0?ooo`0b0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0<`3oool00`000000oooo
0?ooo`0L0?ooo`00;@3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo02P0oooo00<0
00000?ooo`3oool0FP3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo02L0oooo00<0
00000?ooo`3oool0<@3oool00`000000oooo0?ooo`0M0?ooo`00;@3oool00`000000oooo0?ooo`16
0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0`
0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0M
0?ooo`00;@3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo02X0oooo00<000000?oo
o`3oool0FP3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo02T0oooo00<000000?oo
o`3oool0;`3oool00`000000oooo0?ooo`0N0?ooo`00;@3oool00`000000oooo0?ooo`150?ooo`03
0000003oool0oooo02X0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0^0?ooo`03
0000003oool0oooo02T0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0N0?ooo`00
;@3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0
FP3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0
;@3oool00`000000oooo0?ooo`0O0?ooo`00;@3oool00`000000oooo0?ooo`140?ooo`030000003o
ool0oooo02`0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0/0?ooo`030000003o
ool0oooo02/0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0P0?ooo`00;@3oool0
0`000000oooo0?ooo`130?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0F`3oool0
0`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0:`3oool0
0`000000oooo0?ooo`0P0?ooo`00;@3oool3000004<0oooo00<000000?ooo`3oool0;`3oool00`00
0000oooo0?ooo`1J0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0;@3oool00`00
0000oooo0?ooo`0Z0?ooo`030000003oool0oooo0240oooo000]0?ooo`030000003oool0oooo0480
oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo02P0
oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0240
oooo000]0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0<@3oool00`000000oooo
0?ooo`1J0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0;`3oool00`000000oooo
0?ooo`0X0?ooo`030000003oool0oooo0280oooo000]0?ooo`030000003oool0oooo0440oooo00<0
00000?ooo`3oool0<P3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo02H0oooo00<0
00000?ooo`3oool0<@3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo0280oooo000]
0?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`1J
0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0U
0?ooo`030000003oool0oooo02<0oooo000]0?ooo`030000003oool0oooo0400oooo00<000000?oo
o`3oool0=03oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo02@0oooo00<000000?oo
o`3oool0<`3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo02<0oooo000]0?ooo`03
0000003oool0oooo03l0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`1K0?ooo`03
0000003oool0oooo0280oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0S0?ooo`03
0000003oool0oooo02@0oooo000A0?ooo`050?oOW018000000000000M02oool00`3oool01P3o_g@0
000007Boo`3oool0oooO09a800<0000000<0M;oo0?ooo`3oool0203oool00`000000oooo0?ooo`0o
0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`0R
0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0T
0?ooo`00403oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@0000007Bo
o`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`080?ooo`030000003oool0oooo03h0
oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo0200
oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo02D0
oooo000@0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd
0;ooo`080?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool0>@3oool00`000000oooo
0?ooo`1J0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0=`3oool00`000000oooo
0?ooo`0Q0?ooo`030000003oool0oooo02D0oooo000?0?ooo`030?oo_`1d0000B9cO00<0oooo00<0
gia80000M02oool02`3oool00`2oM000M;oo0?ooo`070?ooo`030000003oool0oooo03d0oooo00<0
00000?ooo`3oool0>P3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo01h0oooo00<0
00000?ooo`3oool0>@3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo02H0oooo000?
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3ooml0W4P004RLg`08
0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`1J
0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0O
0?ooo`030000003oool0oooo02H0oooo000?0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol02P3oool00`3OW4P0000007Boo`080?ooo`@00000>`3oool00`000000oooo0?ooo`0l0?oo
o`030000003oool0oooo05/0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0k0?oo
o`030000003oool0oooo01d0oooo00<000000?ooo`3oool09`3oool000l0oooo00<0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`060?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`08
0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`1K
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`0K
0?ooo`030000003oool0oooo02P0oooo000?0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol01P3oool01P3OW4P00000000000000000001809cOo`T0oooo00<000000?ooo`3oool0>`3o
ool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0:03o
ool000l0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d0000
000004RLg`/0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`100?ooo`030000003o
ool0oooo05/0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0o0?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool0:@3oool00100oooo00<0gia80018W03Oool00P3oool0
0`2oM00007Bo0?ooo`070?ooo`030;md001d_ol0oooo00/0oooo00<000000?ooo`3oool0>@3oool0
0`000000oooo0?ooo`120?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool05P3oool0
0`000000oooo0?ooo`110?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0:P3oool0
0100oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00H0omnL04P0000000000000
0000B02Lgol80?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0@`3oool00`000000
oooo0?ooo`1J0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0@@3oool00`000000
oooo0?ooo`0G0?ooo`030000003oool0oooo02X0oooo000A0?ooo`050?oo_`1d000000000000M02o
ool02P3oool01@2oM0000000000000000000B9cO00P0oooo00<000000?ooo`3oool0>03oool00`00
0000oooo0?ooo`140?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`130?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0:`3oool002d0
oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo05/0
oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo01<0
oooo00<000000?ooo`3oool0;03oool002d0oooo00<000000?ooo`3oool0=`3oool00`000000oooo
0?ooo`170?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`170?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0;03oool002d0oooo00<0
00000?ooo`3oool0=P3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo05/0oooo00<0
00000?ooo`3oool0403oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo0140oooo00<0
00000?ooo`3oool0;@3oool002d0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`1:
0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`19
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0;P3oool002d0oooo0`00000e0?oo
o`030000003oool0oooo04/0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`0=0?oo
o`030000003oool0oooo04/0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0^0?oo
o`00;@3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3o
ool0F`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0?ooo`0_0?ooo`00;@3oool00`000000oooo0?ooo`0d0?ooo`030000
003oool0oooo04d0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo04d0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0`0?ooo`00;@3o
ool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool0F`3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`0a0?ooo`00;@3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0
oooo0500oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo04l0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0a0?ooo`00;@3oool00`00
0000oooo0?ooo`0b0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool0FP3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`0b0?ooo`00;@3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo0580
oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo05<0
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0c0?ooo`00;@3oool00`000000oooo
0?ooo`0`0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0F`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`0c0?ooo`00;@3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo05D0oooo00<0
00000?ooo`3oool0FP3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo05D0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0d0?ooo`00;@3oool00`000000oooo0?ooo`0_
0?ooo`030000003oool0oooo05H0oooo00<000000?ooo`3oool0F`3oool01@000000oooo0?ooo`3o
ool0000005T0oooo00D000000?ooo`3oool0oooo0000000g0?ooo`00;@3oool00`000000oooo0?oo
o`0^0?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0000
001K0?ooo`040000003oool0oooo000003L0oooo000]0?ooo`030000003oool0oooo02h0oooo00<0
00000?ooo`3oool0F@3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0000005/0oooo00<0
00000?ooo`000000>03oool002d0oooo0`00000]0?ooo`030000003oool0oooo05X0oooo00<00000
0?ooo`3oool0F`3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo03L0oooo000]0?oo
o`030000003oool0oooo02`0oooo00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`1I0?oo
o`030000003oool0000005`0oooo0P00000h0?ooo`00;@3oool00`000000oooo0?ooo`0[0?ooo`03
0000003oool0oooo05h0oooo00<000000?ooo`3oool0E`3oool01@000000oooo0?ooo`3oool00000
05X0oooo00@000000?ooo`3oool00000=`3oool002d0oooo00<000000?ooo`3oool0:P3oool00`00
0000oooo0?ooo`1P0?ooo`030000003oool0oooo05H0oooo00D000000?ooo`3oool0oooo0000001I
0?ooo`050000003oool0oooo0?ooo`000000=`3oool002d0oooo00<000000?ooo`3oool0:P3oool0
0`000000oooo0?ooo`1P0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`1E0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0=03oool0
02d0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo
05<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool0<`3oool002d0oooo00<000000?ooo`3oool0:03oool00`000000
oooo0?ooo`1T0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`1A0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0<P3oool002d0oooo
00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`1V0?ooo`030000003oool0oooo04l0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool0<@3oool002d0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?oo
o`1X0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`1?0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0<@3oool002d0oooo00<00000
0?ooo`3oool09@3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo04d0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo00/0oooo00<00000
0?ooo`3oool0<03oool002d0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`1Z0?oo
o`030000003oool0oooo04/0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1;0?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool0;`3oool002d0oooo0`00000T0?ooo`03
0000003oool0oooo06`0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0?0?ooo`03
0000003oool0oooo04T0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0^0?ooo`00
;@3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool0
B03oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0
4@3oool00`000000oooo0?ooo`0]0?ooo`00;@3oool00`000000oooo0?ooo`0R0?ooo`030000003o
ool0oooo0700oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0B0?ooo`030000003o
ool0oooo04D0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0/0?ooo`00;@3oool0
0`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0780oooo00<000000?ooo`3oool0A03oool0
0`000000oooo0?ooo`0D0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`0[0?ooo`00;@3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo
07@0oooo00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo
04<0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0[0?ooo`00;@3oool00`000000
oooo0?ooo`0P0?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3oool0@@3oool00`000000
oooo0?ooo`0G0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool05`3oool00`000000
oooo0?ooo`0Z0?ooo`00;@3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo07H0oooo
00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo03l0oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0Y0?ooo`00;@3oool00`000000oooo0?oo
o`0N0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?oo
o`0X0?ooo`00;@3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo07X0oooo00<00000
0?ooo`3oool0?03oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo03/0oooo00<00000
0?ooo`3oool07@3oool00`000000oooo0?ooo`0W0?ooo`00;@3oool00`000000oooo0?ooo`0L0?oo
o`030000003oool0oooo07`0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0N0?oo
o`030000003oool0oooo03T0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0V0?oo
o`00;@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo07h0oooo00<000000?ooo`3o
ool0=`3oool200000280oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0Q0?ooo`03
0000003oool0oooo02D0oooo000]0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
P03oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0
=@3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo02@0oooo000]0?ooo`<00000603o
ool2000008@0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0U0?ooo`800000=03o
ool00`000000oooo0?ooo`0U0?ooo`800000903oool002d0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`250?ooo`800000<@3oool2000002X0oooo00<000000?ooo`3oool0<03oool0
0`000000oooo0?ooo`0X0?ooo`030000003oool0oooo0240oooo000]0?ooo`030000003oool0oooo
01H0oooo00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo
02/0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo
0200oooo000]0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0RP3oool00`000000
oooo0?ooo`0[0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0;03oool00`000000
oooo0?ooo`0/0?ooo`030000003oool0oooo01l0oooo000]0?ooo`030000003oool0oooo01@0oooo
00<000000?ooo`3oool0S03oool00`000000oooo0?ooo`0X0?ooo`800000<@3oool2000002X0oooo
0P00000`0?ooo`030000003oool0oooo01h0oooo000]0?ooo`030000003oool0oooo0180oooo0P00
002@0?ooo`8000009`3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo02D0oooo0P00
000c0?ooo`8000007P3oool002d0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`2B
0?ooo`800000903oool00`000000oooo0?ooo`0e0?ooo`800000903oool00`000000oooo0?ooo`0e
0?ooo`030000003oool0oooo01/0oooo000O0?ooo`060;md00000000000000000000B9`0gooo203o
ool00`000000oooo0?ooo`0?0?ooo`800000U`3oool200000200oooo0P00000j0?ooo`800000803o
ool2000003T0oooo0P00000K0?ooo`00803oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0
oooo00d0oooo0P00002K0?ooo`800000703oool2000003h0oooo00<000000?ooo`3oool0703oool0
0`000000oooo0?ooo`0k0?ooo`<00000603oool00200oooo00<0omnL04P0001d_ol02P3oool00`00
0000oooo0?ooo`0;0?ooo`800000W`3oool2000001L0oooo0`0000110?ooo`8000006P3oool20000
0440oooo0P00000F0?ooo`00803oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo00L0
oooo1000002S0?ooo`@00000403oool3000004H0oooo1000000B0?ooo`@00000A@3oool2000001@0
oooo000P0?ooo`030?oOW0180000M;oo00X0oooo00@000000?ooo`3oool0oooo1P00002[0?ooo`@0
00001`3oool5000004d0oooo1@0000070?ooo`H00000B`3oool6000000h0oooo000P0?ooo`030?oO
W0180000M;oo00X0oooo1000002e0?ooo`L00000E`3oool7000005L0oooo1000000:0?ooo`00803o
ool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo0?l0ooooP@3oool00200oooo00<0omnL
04P0001d_ol02P3oool00`000000oooo0?ooo`3o0?oooh40oooo000P0?ooo`030?oOW0180000M;oo
00X0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00803oool00`3ogi`0B00007Boo`0:0?ooo`03
0000003oool0oooo0?l0ooooP@3oool001h0oooo00D0oooO09a800000000000007Boo`0:0?ooo`03
0000003oool0oooo0?l0ooooP@3oool00200oooo00<0okmd0000001d_ol0o`3ooon>0?ooo`00o`3o
oona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3o
oona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.767649, -1.14503, \
0.0250857, 0.01292}}]
}, Closed]],

Cell[TextData[{
  "An alternative method for superimposing plots which does not require \
manipulation of ",
  StyleBox["DisplayFunction", "InlineInput",
    FontWeight->"Bold"],
  " is provided by the function ",
  StyleBox["DisplayTogether", "InlineInput",
    FontWeight->"Bold"],
  " found in the ",
  StyleBox["Graphics`Graphics`", "InlineInput",
    FontWeight->"Bold"],
  " package.  In the example below, only the final plot is displayed."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(\(Clear[x, v, a, t];\)\n
    x[t]\  = \ Cos[t]; \ v[t]\  = \ D[x[t], t]; \ a[t]\  = \ D[v[t], t];\n
    \(DisplayTogether[Plot[Evaluate[x[t]], {t, 0, 2  \[Pi]}], 
        Plot[Evaluate[v[t]], {t, 0, 2  \[Pi]}], 
        Plot[Evaluate[a[t]], {t, 0, 2  \[Pi]}], Frame \[Rule] True, 
        FrameLabel \[Rule] {"\<time\>", "\<x,v,a\>"}, \ 
        PlotLabel \[Rule] "\<simple harmonic motion\>"];\)\)\(\ \)\)\)], \
"Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.151576 0.309017 0.294302 [
[.02381 -0.0125 -5 -15.6875 ]
[.02381 -0.0125 5 0 ]
[.17539 -0.0125 -5 -15.6875 ]
[.17539 -0.0125 5 0 ]
[.32696 -0.0125 -5 -15.6875 ]
[.32696 -0.0125 5 0 ]
[.47854 -0.0125 -5 -15.6875 ]
[.47854 -0.0125 5 0 ]
[.63011 -0.0125 -5 -15.6875 ]
[.63011 -0.0125 5 0 ]
[.78169 -0.0125 -5 -15.6875 ]
[.78169 -0.0125 5 0 ]
[.93327 -0.0125 -5 -15.6875 ]
[.93327 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -12.625 -31.375 ]
[.5 -0.0125 12.625 -15.6875 ]
[-0.0125 .01472 -18.3125 -7.84375 ]
[-0.0125 .01472 0 7.84375 ]
[-0.0125 .16187 -27.3125 -7.84375 ]
[-0.0125 .16187 0 7.84375 ]
[-0.0125 .30902 -10 -7.84375 ]
[-0.0125 .30902 0 7.84375 ]
[-0.0125 .45617 -19 -7.84375 ]
[-0.0125 .45617 0 7.84375 ]
[-0.0125 .60332 -10 -7.84375 ]
[-0.0125 .60332 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -43 -13.75 ]
[-0.01875 .30902 -27.3125 13.75 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -60.5313 0 ]
[.5 .63053 60.5313 15.6875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.17539 0 m
.17539 .00625 L
s
gsave
.17539 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.32696 0 m
.32696 .00625 L
s
gsave
.32696 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.47854 0 m
.47854 .00625 L
s
gsave
.47854 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.63011 0 m
.63011 .00625 L
s
gsave
.63011 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.78169 0 m
.78169 .00625 L
s
gsave
.78169 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.93327 0 m
.93327 .00625 L
s
gsave
.93327 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.125 Mabswid
.05412 0 m
.05412 .00375 L
s
.08444 0 m
.08444 .00375 L
s
.11476 0 m
.11476 .00375 L
s
.14507 0 m
.14507 .00375 L
s
.2057 0 m
.2057 .00375 L
s
.23602 0 m
.23602 .00375 L
s
.26633 0 m
.26633 .00375 L
s
.29665 0 m
.29665 .00375 L
s
.35728 0 m
.35728 .00375 L
s
.38759 0 m
.38759 .00375 L
s
.41791 0 m
.41791 .00375 L
s
.44822 0 m
.44822 .00375 L
s
.50885 0 m
.50885 .00375 L
s
.53917 0 m
.53917 .00375 L
s
.56948 0 m
.56948 .00375 L
s
.5998 0 m
.5998 .00375 L
s
.66043 0 m
.66043 .00375 L
s
.69074 0 m
.69074 .00375 L
s
.72106 0 m
.72106 .00375 L
s
.75138 0 m
.75138 .00375 L
s
.81201 0 m
.81201 .00375 L
s
.84232 0 m
.84232 .00375 L
s
.87264 0 m
.87264 .00375 L
s
.90295 0 m
.90295 .00375 L
s
.96358 0 m
.96358 .00375 L
s
.9939 0 m
.9939 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -73.625 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(time) show
84.250 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .01472 m
.00625 .01472 L
s
gsave
-0.0125 .01472 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .16187 m
.00625 .16187 L
s
gsave
-0.0125 .16187 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
0 .30902 m
.00625 .30902 L
s
gsave
-0.0125 .30902 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .45617 m
.00625 .45617 L
s
gsave
-0.0125 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
0 .60332 m
.00625 .60332 L
s
gsave
-0.0125 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .04415 m
.00375 .04415 L
s
0 .07358 m
.00375 .07358 L
s
0 .10301 m
.00375 .10301 L
s
0 .13244 m
.00375 .13244 L
s
0 .1913 m
.00375 .1913 L
s
0 .22073 m
.00375 .22073 L
s
0 .25016 m
.00375 .25016 L
s
0 .27959 m
.00375 .27959 L
s
0 .33845 m
.00375 .33845 L
s
0 .36788 m
.00375 .36788 L
s
0 .39731 m
.00375 .39731 L
s
0 .42674 m
.00375 .42674 L
s
0 .4856 m
.00375 .4856 L
s
0 .51503 m
.00375 .51503 L
s
0 .54446 m
.00375 .54446 L
s
0 .57389 m
.00375 .57389 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -109.906 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
74.75 11.8438 translate
90 rotate
-74.75 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
69.125 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(,) show
72.125 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(v) show
78.125 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(,) show
81.125 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(a) show
86.500 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.17539 .61178 m
.17539 .61803 L
s
.32696 .61178 m
.32696 .61803 L
s
.47854 .61178 m
.47854 .61803 L
s
.63011 .61178 m
.63011 .61803 L
s
.78169 .61178 m
.78169 .61803 L
s
.93327 .61178 m
.93327 .61803 L
s
.125 Mabswid
.05412 .61428 m
.05412 .61803 L
s
.08444 .61428 m
.08444 .61803 L
s
.11476 .61428 m
.11476 .61803 L
s
.14507 .61428 m
.14507 .61803 L
s
.2057 .61428 m
.2057 .61803 L
s
.23602 .61428 m
.23602 .61803 L
s
.26633 .61428 m
.26633 .61803 L
s
.29665 .61428 m
.29665 .61803 L
s
.35728 .61428 m
.35728 .61803 L
s
.38759 .61428 m
.38759 .61803 L
s
.41791 .61428 m
.41791 .61803 L
s
.44822 .61428 m
.44822 .61803 L
s
.50885 .61428 m
.50885 .61803 L
s
.53917 .61428 m
.53917 .61803 L
s
.56948 .61428 m
.56948 .61803 L
s
.5998 .61428 m
.5998 .61803 L
s
.66043 .61428 m
.66043 .61803 L
s
.69074 .61428 m
.69074 .61803 L
s
.72106 .61428 m
.72106 .61803 L
s
.75138 .61428 m
.75138 .61803 L
s
.81201 .61428 m
.81201 .61803 L
s
.84232 .61428 m
.84232 .61803 L
s
.87264 .61428 m
.87264 .61803 L
s
.90295 .61428 m
.90295 .61803 L
s
.96358 .61428 m
.96358 .61803 L
s
.9939 .61428 m
.9939 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .01472 m
1 .01472 L
s
.99375 .16187 m
1 .16187 L
s
.99375 .30902 m
1 .30902 L
s
.99375 .45617 m
1 .45617 L
s
.99375 .60332 m
1 .60332 L
s
.125 Mabswid
.99625 .04415 m
1 .04415 L
s
.99625 .07358 m
1 .07358 L
s
.99625 .10301 m
1 .10301 L
s
.99625 .13244 m
1 .13244 L
s
.99625 .1913 m
1 .1913 L
s
.99625 .22073 m
1 .22073 L
s
.99625 .25016 m
1 .25016 L
s
.99625 .27959 m
1 .27959 L
s
.99625 .33845 m
1 .33845 L
s
.99625 .36788 m
1 .36788 L
s
.99625 .39731 m
1 .39731 L
s
.99625 .42674 m
1 .42674 L
s
.99625 .4856 m
1 .4856 L
s
.99625 .51503 m
1 .51503 L
s
.99625 .54446 m
1 .54446 L
s
.99625 .57389 m
1 .57389 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .30902 m
1 .30902 L
s
gsave
.5 .63053 -121.531 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(simple) show
97.938 15.813 moveto
(harmonic) show
146.063 15.813 moveto
(motion) show
180.063 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .60332 m
.02499 .60331 L
.02605 .60329 L
.02729 .60324 L
.02846 .60318 L
.03053 .60303 L
.03279 .6028 L
.03527 .60248 L
.0379 .60205 L
.04262 .60106 L
.04749 .59974 L
.05205 .59822 L
.06244 .59381 L
.07305 .58793 L
.08274 .58135 L
.10458 .56251 L
.14429 .51515 L
.18248 .45627 L
.22313 .38349 L
.26226 .30833 L
.30384 .22863 L
.34391 .15746 L
.38246 .09886 L
.402 .07411 L
.42346 .05145 L
.44388 .03466 L
.45301 .02874 L
.46295 .02346 L
.46838 .0211 L
.4734 .01924 L
.4781 .01778 L
.4833 .0165 L
.48553 .01605 L
.48794 .01565 L
.49002 .01535 L
.4923 .0151 L
.49473 .01489 L
.49605 .01482 L
.49728 .01476 L
.49844 .01473 L
.49949 .01472 L
.50071 .01472 L
.50186 .01474 L
.50301 .01477 L
.50425 .01483 L
.50542 .0149 L
.50648 .01498 L
.5091 .01524 L
.51158 .01557 L
.5165 .01646 L
Mistroke
.52187 .01777 L
.53144 .02102 L
.54032 .02507 L
.56088 .03814 L
.5795 .05428 L
.61962 .1017 L
.65822 .16102 L
.69928 .23447 L
.73882 .31042 L
.77684 .38344 L
.81732 .45593 L
.85628 .51593 L
.87613 .54149 L
.89769 .56473 L
.91843 .58221 L
.92832 .58876 L
.93759 .59383 L
.94643 .59767 L
.95595 .6007 L
.9611 .60186 L
.96369 .60232 L
.96652 .60272 L
.96894 .60298 L
.9703 .6031 L
.97153 .60318 L
.97268 .60324 L
.97393 .60329 L
.97511 .60331 L
.97619 .60332 L
Mfstroke
.02381 .30902 m
.06244 .23481 L
.10458 .15951 L
.14415 .09914 L
.18221 .05447 L
.20178 .03757 L
.21267 .03011 L
.22272 .0245 L
.23293 .02007 L
.23832 .01827 L
.24402 .01676 L
.24641 .01625 L
.24897 .01579 L
.25119 .01545 L
.25361 .01516 L
.25507 .01501 L
.25643 .01491 L
.25773 .01483 L
.2584 .01479 L
.25912 .01476 L
.25983 .01474 L
.26048 .01473 L
.26176 .01472 L
.26293 .01472 L
.26416 .01475 L
.26538 .01479 L
.26607 .01483 L
.26671 .01486 L
.26912 .01505 L
.27131 .01528 L
.27364 .0156 L
.27847 .01647 L
.2838 .01778 L
.29326 .02099 L
.30213 .02502 L
.32232 .03778 L
.34082 .05371 L
.38045 .10022 L
.41855 .15838 L
.45912 .23059 L
.49816 .30544 L
.53966 .38514 L
.57964 .45663 L
.61811 .51582 L
.63759 .54096 L
.65902 .56418 L
.67938 .58151 L
.69843 .5933 L
.70884 .59785 L
.71396 .5996 L
Mistroke
.71872 .60092 L
.72301 .60186 L
.72758 .60261 L
.7302 .60292 L
.73148 .60304 L
.73266 .60313 L
.73371 .6032 L
.73487 .60325 L
.73603 .60329 L
.73725 .60331 L
.73856 .60332 L
.73977 .6033 L
.7404 .60328 L
.7411 .60326 L
.74251 .60319 L
.74375 .60311 L
.74508 .60301 L
.74748 .60275 L
.74985 .60243 L
.75235 .60202 L
.7568 .60108 L
.76198 .59967 L
.7668 .59806 L
.77773 .59332 L
.78828 .58734 L
.79823 .58046 L
.81683 .5645 L
.85534 .51957 L
.89632 .45702 L
.93577 .38657 L
.97371 .31383 L
.97619 .30902 L
Mfstroke
.02381 .01472 m
.02499 .01472 L
.02605 .01475 L
.02729 .01479 L
.02846 .01485 L
.03053 .015 L
.03279 .01523 L
.03527 .01556 L
.0379 .01599 L
.04262 .01698 L
.04749 .0183 L
.05205 .01981 L
.06244 .02422 L
.07305 .03011 L
.08274 .03668 L
.10458 .05552 L
.14429 .10289 L
.18248 .16177 L
.22313 .23454 L
.26226 .3097 L
.30384 .3894 L
.34391 .46058 L
.38246 .51917 L
.402 .54392 L
.42346 .56659 L
.44388 .58338 L
.45301 .58929 L
.46295 .59457 L
.46838 .59694 L
.4734 .5988 L
.4781 .60025 L
.4833 .60153 L
.48553 .60198 L
.48794 .60239 L
.49002 .60268 L
.4923 .60294 L
.49473 .60314 L
.49605 .60322 L
.49728 .60327 L
.49844 .6033 L
.49949 .60332 L
.50071 .60332 L
.50186 .6033 L
.50301 .60326 L
.50425 .6032 L
.50542 .60313 L
.50648 .60305 L
.5091 .60279 L
.51158 .60246 L
.5165 .60158 L
Mistroke
.52187 .60026 L
.53144 .59701 L
.54032 .59297 L
.56088 .57989 L
.5795 .56376 L
.61962 .51633 L
.65822 .45702 L
.69928 .38356 L
.73882 .30761 L
.77684 .2346 L
.81732 .1621 L
.85628 .1021 L
.87613 .07654 L
.89769 .0533 L
.91843 .03582 L
.92832 .02927 L
.93759 .02421 L
.94643 .02037 L
.95595 .01733 L
.9611 .01617 L
.96369 .01572 L
.96652 .01531 L
.96894 .01505 L
.9703 .01494 L
.97153 .01485 L
.97268 .01479 L
.97393 .01475 L
.97511 .01472 L
.97619 .01472 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00j`3oool01@3ogi`0B00000000018W=l0gia80080000000L0001d09bLB000000000000000
B02Lgml0W4P00080000000<0001d0;oog`2LB0000P0000001000M;l0oooo0?ooo`2oM00200000003
0000M02oool0oooo0:L0oooo003[0?ooo`0<0=nLB00007@0_ooo0=no_`3Oool0gia80000M02oool0
oooo0=nLB00007@0_ooo0P3oool01`2oM00004RL0=ooo`3oool0oooO09a80000M;l00P3oool01`2o
M0000000001d_`3oool0omnL04P0M02oool0Y`3oool00>/0oooo00<0gia80018W03Oool00P3oool0
1`3OW4P0001d0;ooo`3oool0gia80000M02oool00P3oool02`2oM00004RL0=ooo`3oool0oooO09a8
0000M;l0oooo0?oOW0180000B9cO00<0oooo00<0omnL07BLg`3oool0YP3oool00>/0oooo00<0gia8
0018W03Oool00P3oool01`3OW4P0001d0;ooo`3oool0gia80000M02oool00P3oool02`2oM00004RL
0=ooo`3oool0oooO09a80000M;l0oooo0?noM00004P0W=oo0:`0oooo003[0?ooo`030=nLB000B9`0
gooo0080oooo00L0gia80000M02oool0oooo0=nLB00007@0_ooo0080oooo00/0_g@00018W03Oool0
oooo0?oog`2LB00007Bo0?ooo`3o_g@0001d0;ooo`2/0?ooo`00j`3oool00`3OW4P004RL0=ooo`02
0?ooo`070=nLB00007@0_ooo0?ooo`3OW4P0001809cOo`020?ooo`0:0;md000007@0_ooo0?ooo`3o
oml0W4P0001d_`3oool0omnL04P000D0000000<0B9cO0?ooo`3oool0Y@3oool00>/0oooo01D0gia8
0018W03Oool0oooO09a800000000001d0;oog`2LB00000000000B018B4P0W=oo0?oo_`1d000007Ad
07@0M02oool0oono07@00018W=l00P3oool01`2oM000B9cO0?ooo`3oool0gia80000B02Lgol0Y`3o
ool00>X0oooo01H0_g@000000000000000000018W03Oool0_g@00000M02oool0oooo0;md000007@0
_ooO09a800000000001809cOo`3ookl0M0000000000007@0_ooo0`3oool01@3OW4P0000000000000
04P0W=oo0:P0oooo003[0?ooo`030;md0000B9`0gooo0<80oooo003[0?ooo`030?oo_`1dB9`0gooo
0<80oooo003[0?ooo`030?oog`2LM9`0gooo0<80oooo003`0?ooo`030=nLB00004P0W=oo0;d0oooo
003`0?ooo`030=nLB00004P0W=oo0;d0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40
oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40
oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo001W0?ooo`050?oOW01800000000
0000M02oool0:@3oool01P2oM000000000000000000004RL0=ooobH0oooo00<0_g@0000000000000
100000000`18W=l0oooo0?ooo`0T0?ooo`030?oo_`1d000000000080000000<0M;oo0?ooo`3oool0
:P3oool0103ooml0W4P00018W03OoolV0?ooo`030?oog`2LB00000000080000000<0M;oo0?ooo`3o
ool0:03oool01@3ogi`0B000000000000000M;oo0300oooo001V0?ooo`070?oo_`1d0000M;oo0?oo
o`3o_g@0001d0;ooo`0Y0?ooo`030?oOW0180000M;oo02T0oooo00<0gia80000000000000`000000
0`00B9`0gooo0?ooo`0T0?ooo`030?oOW0180000B9cO0080oooo00<0_g@004RLg`3oool0:P3oool0
103ooml0W4P00018W03OoolV0?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool09`3o
ool01`3ogi`0B00004RLg`3oool0oono07@0001d_ol0;`3oool006H0oooo00<0_g@00018W03Oool0
0P3oool00`2oM00007Bo0?ooo`0X0?ooo`030?oOW0180000M;oo02X0oooo00<0gia804RLg`3oool0
0P3oool00`3OW7@0W=oo0?ooo`0Y0?ooo`030?oo_`1d04P0W=oo02X0oooo00@0oooO09a80000B9`0
gooo:`3oool00`3ooml0W4Qd0;ooo`0V0?ooo`030?oog`2LB00007Bo00<0oooo00<0okmd0018W03O
ool0;P3oool006D0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0X0?ooo`030?oO
W0180000M;oo02/0oooo00<0gia807Boo`3oool0;@3oool00`3ookl0M00007Boo`0V0?ooo`030;md
00000000000000D0000000<0M;oo0?ooo`3oool0:@3oool00`2oM000M;oo0?ooo`0U0?ooo`030?oO
W0180000M;oo00<0oooo00<0okmd0000B02Lgol0;P3oool006D0oooo00<0omnL04P0001d_ol00`3o
ool00`3o_g@0001809cOo`0X0?ooo`030?oOW0180000M;oo02`0oooo00<0_gA809cOo`3oool0;03o
ool00`3o_g@0000007Boo`0V0?ooo`080?oOW01dW=l0oooo0?ooo`3ooml0W4P00018W03Oool[0?oo
o`030?oog`2LB000B9cO02H0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0^0?oo
o`00I@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo02P0oooo00<0omnL04P0001d
_ol0;03oool00`3ooml0W4Qd0;ooo`0[0?ooo`040?oOW0180000001809cOobL0oooo00L0gia807Bo
o`3oool0oooO09a80000B9`0gooo02/0oooo00<0gia80000001d_ol09P3oool00`3ogi`0B00007Bo
o`030?ooo`030;md000004P0W=oo02h0oooo001U0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol0:03oool00`3ogi`0B00007Boo`0]0?ooo`030?noM000M;l0oooo02P0oooo00D0okmd
000000000000001809cOo`0X0?ooo`070?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`0W0?ooo`07
0=oOg`3oool0oono07@000000000001d0;ooo`0V0?ooo`080?oo_`1d0000000004RLg`3oool0okmd
00000000M;l_0?ooo`00I@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo02P0oooo
00<0omnL04P0001d_ol0;@3oool00`3ookl0M01809cOo`0Z0?ooo`030;md0000M;l0oooo02T0oooo
00H0omnL04Qd_`3ooml0W4P00018W03OoolW0?ooo`060=nLB000000000000000000004P0W=oo:03o
ool00`2oM00004RL0;nLB002000000030018W03Oool0oooo02h0oooo001U0?ooo`030?oo_`1d0000
B9cO00<0oooo00<0gia80000M02oool0:03oool00`3ogi`0B00007Boo`0]0?ooo`030?oo_`1d0000
B9cO02/0oooo00<0_g@004RLg`3oool0:@3oool01@3OW4P0M;oO09a80000B9`0gooo02L0oooo00@0
oono07@000000000B9cO:P3oool00`3ogi`0B00007Boo`0c0?ooo`00IP3oool00`3OW4P004RL0=oo
o`020?ooo`030;md0000M;l0oooo02P0oooo00<0omnL04P0001d_ol0:03oool00`3o_i`0W=oo0?oo
o`020?ooo`030?oOW0180000B9cO02H0oooo00<0oooO09bL_`3oool00P3oool00`3OW4P0001d0;oo
o`0Y0?ooo`050?oog`2LB4P0B4P00018W03Oool0:03oool00`2oM000M;oo0?ooo`0[0?ooo`030=nL
B00004P0W=oo0380oooo001V0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0W0?ooo`05
0?oog`2LB00000000000001d_ol0:03oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo
9`3oool01P3o_g@0B9cO0?ooo`3ookl0M000001d_b/0oooo00@0omnL04P00000B9`0gooo:03oool0
1P3ogi`0B000000000000000001809cOobT0oooo00<0okmd00000000M;l0<@3oool006L0oooo00D0
oono07@000000000001d0;ooo`0Z0?ooo`030?noM0000000M;oo02T0oooo00H0oooO09a800000000
00000000M02ooolY0?ooo`050=nLB000000000000000M02oool0;03oool00`3OW4P004RL0=ooo`0Y
0?ooo`050;md00000000000000000018W=l0:P3oool01@3ooml0W4P00000000007@0_ooo02h0oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo001D0?ooo`060;md0000000000000000
0000B9`0gooo203ooooo0000038000007@3oool005D0oooo00<0omnL04P0001d_ol02P3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0000000M0?ooo`00E@3oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0000001d0
oooo001E0?ooo`030?oOW0180000M;oo00X0oooo00<000000?ooo`3oool0o`3oool^0?ooo`030000
003oool0oooo01/0oooo00160?ooo`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool00P3o
ool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo0?l0oooo;P3oool00`000000oooo0?oo
o`0K0?ooo`00E@3oool00`3ogi`0B00007Boo`0:0?ooo`<00000103oool6000003h0oooo2`00000m
0?ooo`/00000QP3oool6000000@0oooo0`00000M0?ooo`00E@3oool00`3ogi`0B00007Boo`0:0?oo
o`030000003oool0oooo00X0oooo1000000h0?ooo`8000002`3oool2000003P0oooo0`00000;0?oo
o`<00000P@3oool2000000`0oooo00<000000?ooo`3oool06`3oool005D0oooo00<0omnL04P0001d
_ol02P3oool00`000000oooo0?ooo`0>0?ooo`800000<`3oool3000000l0oooo0`00000c0?ooo`80
00004@3oool2000007`0oooo0`00000>0?ooo`030000003oool0oooo01/0oooo001E0?ooo`030?oO
W0180000M;oo00X0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0^0?ooo`800000
5@3oool2000002l0oooo0P00000E0?ooo`800000M`3oool300000140oooo00<000000?ooo`3oool0
6`3oool005D0oooo00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`0A0?ooo`800000;03o
ool2000001T0oooo0P00000[0?ooo`8000006@3oool2000007<0oooo0P00000D0?ooo`030000003o
ool0oooo01/0oooo001C0?ooo`050?oog`2LB00000000000001d_ol02P3oool00`000000oooo0?oo
o`0C0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?oo
o`0V0?ooo`8000007@3oool00`000000oooo0?ooo`1^0?ooo`8000005P3oool00`000000oooo0?oo
o`0K0?ooo`00E@3oool00`3o_g@0000007Boo`0:0?ooo`030000003oool0oooo01@0oooo0P00000V
0?ooo`8000007`3oool2000002D0oooo00<000000?ooo`3oool07P3oool2000006d0oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0F0?oo
o`030000003oool0oooo0240oooo0P00000S0?ooo`030000003oool0oooo0240oooo00<000000?oo
o`3oool08@3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool06`3oool00680oooo00<000000?ooo`3oool05`3oool200000200oooo00<000000?ooo`3o
ool0903oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3o
ool0I`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01/0oooo001R0?ooo`800000
6P3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0
703oool2000002L0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`0H0?ooo`800000
7@3oool00680oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0I0?ooo`800000:P3o
ool2000001/0oooo00<000000?ooo`3oool0:03oool2000006@0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0K0?ooo`800000603oool0
0`000000oooo0?ooo`0/0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0:`3oool0
0`000000oooo0?ooo`1P0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool06`3oool0
0680oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo
02h0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo
05h0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000
oooo0?ooo`0N0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0<03oool00`000000
oooo0?ooo`0C0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0G03oool00`000000
oooo0?ooo`0M0?ooo`030000003oool0oooo01/0oooo001R0?ooo`030000003oool0oooo01l0oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0140oooo
00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo01h0oooo
00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool07`3oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`0c0?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?oo
o`0K0?ooo`00HP3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool0=03oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo03@0oooo00<00000
0?ooo`3oool0E`3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo01/0oooo001R0?oo
o`030000003oool0oooo0240oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0f0?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`1E0?oo
o`030000003oool0oooo0240oooo00<000000?ooo`3oool06`3oool00680oooo0P00000S0?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`0:0?ooo`03
0000003oool0oooo03P0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`0Q0?ooo`80
00007@3oool00680oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`080?ooo`030000
003oool0oooo03X0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0j0?ooo`030000
003oool0oooo0540oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0K0?ooo`00HP3o
ool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0>`3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0C`3o
ool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo01/0oooo001R0?ooo`030000003oool0
oooo02@0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0
oooo02D0oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool09@3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0o0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool09P3oool00`00
0000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0400
oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo01/0
oooo001R0?ooo`030000003oool0oooo02L0oooo00D000000?ooo`3oool0oooo000000130?ooo`04
0000003oool0oooo000004@0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0W0?oo
o`030000003oool0oooo01/0oooo001R0?ooo`030000003oool0oooo02L0oooo00@000000?ooo`3o
ool00000A@3oool00`000000oooo000000150?ooo`030000003oool0oooo04L0oooo00<000000?oo
o`3oool0:03oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0X0?ooo`80
0000A`3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3o
ool0:@3oool00`000000oooo0?ooo`0K0?ooo`00HP3oool2000002T0oooo0P0000160?ooo`030000
003oool0000004L0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`0Y0?ooo`800000
7@3oool00680oooo00<000000?ooo`3oool0:03oool00`000000oooo000000150?ooo`030000003o
ool0000004L0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`0Z0?ooo`030000003o
ool0oooo01/0oooo001R0?ooo`030000003oool0oooo02L0oooo00@000000?ooo`3oool00000A03o
ool01@000000oooo0?ooo`3oool0000004L0oooo00<000000?ooo`3oool0@@3oool00`000000oooo
0?ooo`0[0?ooo`030000003oool0oooo01/0oooo001R0?ooo`030000003oool0oooo02H0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0AP3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo02`0oooo00<0
00000?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`15
0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0K
0?ooo`00HP3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0?P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo04D0oooo00<000000?oo
o`3oool0?@3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo01/0oooo001R0?ooo`03
0000003oool0oooo02@0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0l0?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0l0?ooo`03
0000003oool0oooo02d0oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool0
903oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0
;P3oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0S0?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo04H0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0_0?ooo`030000003o
ool0oooo01/0oooo001R0?ooo`8000008`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo03T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`160?ooo`030000003oool0
oooo03P0oooo00<000000?ooo`3oool0;P3oool2000001d0oooo001R0?ooo`030000003oool0oooo
0280oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo
0300oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool08@3oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`160?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0<@3oool00`000000
oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo04H0oooo
00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo01/0oooo
00170?ooo`050?oOW018000000000000M02oool00`3oool01P3o_g@0000007Boo`3oool0oooO09a8
00<0000000<0M;oo0?ooo`3oool01`3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
04D0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo
01/0oooo00160?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000
M;oo0?ooo`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00L0oooo00<000000?ooo`3oool0
7`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0
<P3oool00`000000oooo0?ooo`0K0?ooo`00AP3oool00`2oM00004RL0=ooo`020?ooo`030;md0000
M;l0oooo00X0oooo00<0oooO09a8M02oool01`3oool00`000000oooo0?ooo`0N0?ooo`030000003o
ool0oooo0180oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0A0?ooo`030000003o
ool0oooo04H0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0c0?ooo`030000003o
ool0oooo01/0oooo00150?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02`3oool0
0`2oM000M;oo0?ooo`060?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`0_0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0AP3oool0
0`000000oooo0?ooo`0_0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool06`3oool0
03P0oooo00<0oooO09a8000000001`0000001P18W=l0oooo0?ooo`3ogi`0B00007Boo`<0oooo00<0
okmd0000B02Lgol02P3oool00`3ooml0W4P004RLg`070?ooo`030000003oool0oooo01d0oooo00<0
00000?ooo`3oool0503oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo01@0oooo00<0
00000?ooo`3oool0AP3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo03@0oooo00<0
00000?ooo`3oool06`3oool004D0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:
0?ooo`030=nLB0000000M;oo00L0oooo0`00000L0?ooo`030000003oool0oooo01H0oooo00<00000
0?ooo`3oool0;@3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo04H0oooo00<00000
0?ooo`3oool0;03oool00`000000oooo0?ooo`0c0?ooo`<000007@3oool004D0oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`070=oOg`3oool0oono07@000000000001d0;oo
o`070?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`0[0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?oo
o`0[0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool06`3oool004D0oooo00<0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`060=nLB000000000000000000004P0W=oo
203oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0
:`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0
:P3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo01/0oooo00150?ooo`030?oo_`1d
0000B9cO00<0oooo00<0gia80000M02oool01P3oool0103ookl0M00000000018W=l:0?ooo`030000
003oool0oooo01/0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0Y0?ooo`030000
003oool0oooo01P0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0Y0?ooo`030000
003oool0oooo03L0oooo00<000000?ooo`3oool06`3oool004H0oooo00<0gia80018W03Oool00P3o
ool00`2oM00007Bo0?ooo`070?ooo`030;md001d_ol0oooo00X0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`160?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0=`3o
ool00`000000oooo0?ooo`0K0?ooo`00AP3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0
203oool01P3ogi`0B000000000000000001809cOo`L0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`0K0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool06`3oool00`00
0000oooo0?ooo`150?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0>03oool00`00
0000oooo0?ooo`0K0?ooo`00A`3oool01@3ookl0M0000000000007@0_ooo00X0oooo00D0_g@00000
00000000000004RLg`070?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0703oool0
0`000000oooo0?ooo`0V0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0AP3oool0
0`000000oooo0?ooo`0V0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool06`3oool0
0680oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo
02D0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo
02@0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0903oool00`000000
oooo0?ooo`0N0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0903oool00`000000
oooo0?ooo`0i0?ooo`030000003oool0oooo01/0oooo001R0?ooo`800000603oool00`000000oooo
0?ooo`0P0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0803oool00`000000oooo
0?ooo`160?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0>@3oool2000001d0oooo
001R0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0803oool00`000000oooo0?oo
o`0R0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?oo
o`0R0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool06`3oool00680oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo0200oooo00<00000
0?ooo`3oool08P3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo0200oooo00<00000
0?ooo`3oool0>`3oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0F0?oo
o`030000003oool0oooo02<0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0S0?oo
o`030000003oool0oooo04D0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0l0?oo
o`030000003oool0oooo01/0oooo001R0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3o
ool0903oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3o
ool0AP3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3o
ool06`3oool00680oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0U0?ooo`030000
003oool0oooo01d0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`150?ooo`030000
003oool0oooo01d0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`0K0?ooo`00HP3o
ool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0703o
ool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0703o
ool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo01/0oooo001R0?ooo`030000003oool0
oooo01<0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo02P0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo03h0oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool04`3oool00`00
0000oooo0?ooo`0X0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0:@3oool00`00
0000oooo0?ooo`150?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0?P3oool00`00
0000oooo0?ooo`0K0?ooo`00HP3oool2000001<0oooo00<000000?ooo`3oool0:P3oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0AP3oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo03h0oooo0P00000M0?ooo`00HP3oool00`000000oooo
0?ooo`0B0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0603oool00`000000oooo
0?ooo`0[0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0603oool00`000000oooo
0?ooo`0o0?ooo`030000003oool0oooo01/0oooo001R0?ooo`030000003oool0oooo0140oooo00<0
00000?ooo`3oool0;03oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo02`0oooo00<0
00000?ooo`3oool0AP3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0400oooo00<0
00000?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0/
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`15
0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`0K
0?ooo`00HP3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo02h0oooo00<000000?oo
o`3oool0503oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo04H0oooo00<000000?oo
o`3oool0503oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo01/0oooo001R0?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0B0?ooo`03
0000003oool0oooo0300oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0B0?ooo`03
0000003oool0oooo0480oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool0
3`3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0
<03oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0
@P3oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo0380oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0b0?ooo`030000003o
ool0oooo04H0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`130?ooo`030000003o
ool0oooo01/0oooo001R0?ooo`8000003`3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0
oooo0100oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`160?ooo`030000003oool0
oooo0100oooo00<000000?ooo`3oool0@P3oool2000001d0oooo001R0?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
03@0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
04@0oooo00<000000?ooo`3oool06`3oool002H0oooo00<0_gAd0;ooo`3oool0103oool00`2oM7@0
_ooo0?ooo`0b0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0=03oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0AP3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool06`3oool002H0oooo
00<0_g@004RLg`3oool00`3oool00`3OW4P0001d0;ooo`0c0?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo03H0oooo
00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo04D0oooo
00<000000?ooo`3oool06`3oool002H0oooo00T0_g@00000000007@0_ooo0?oog`2LB000B9cO0=nL
W02oool0<`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo03H0oooo00<000000?oo
o`3oool0303oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo04H0oooo00<000000?oo
o`3oool0303oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo01/0oooo000V0?ooo`06
0;nLW03Ogi`0B0000000000004P0W=oo=P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo03P0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0
oooo04H0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`160?ooo`030000003oool0
oooo01/0oooo000V0?ooo`030;mdM02ogi`0B0000080000000@0001809cOo`3OW9`0_ooo<`3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool02P3oool0
0`000000oooo0?ooo`0h0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool02P3oool0
0`000000oooo0?ooo`160?ooo`030000003oool0oooo01/0oooo000V0?ooo`030;md0000B9`0gooo
0080oooo00@0_g@00000000007@0_ooo<`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo03X0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0j0?ooo`030000003oool0
oooo04H0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`170?ooo`030000003oool0
oooo01/0oooo000V0?ooo`030;mdB02Lgol0oooo00<0oooo00<0okmd0000M02oool0<`3oool00`00
0000oooo0?ooo`090?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`0j0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`170?ooo`030000003oool0oooo01/0oooo000V0?ooo`030;nLW03Oool0oooo00@0
oooo00<0gibL0;ooo`3oool0<P3oool2000000X0oooo00<000000?ooo`3oool0?03oool00`000000
oooo0?ooo`060?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0AP3oool00`000000
oooo0?ooo`060?ooo`030000003oool0oooo04L0oooo0P00000M0?ooo`00HP3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`0l0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`180?ooo`030000003oool0oooo01/0oooo001R0?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0?P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo03h0oooo00<0
00000?ooo`3oool0AP3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo04T0oooo00<0
00000?ooo`3oool06`3oool002/0oooo00H0oooO09a80000B9`0gooo0?oog`2LW;la0?ooo`030000
003oool0oooo00L0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo03h0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo04T0oooo00<000000?ooo`3oool06`3oool002/0oooo00H0oono07@00000B9`0gooo
0;mdB02LgolS0?ooo`050?oOW018000000000000M02oool02@3oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo0400oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`100?oo
o`030000003oool0oooo04H0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?oo
o`030000003oool0oooo01/0oooo000/0?ooo`040=nLB0000000000007Boob<0oooo00L0oono07@0
001d_ol0oooo0?noM00007@0_ooo00P0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`110?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0@03oool00`000000oooo0?oo
o`160?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?oo
o`0K0?ooo`009P3oool00`2oM7@0_ooo0?ooo`0Z0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0
001d_`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0480oooo00@00000
0?ooo`3oool00000A03oool00`000000oooo0?ooo`160?ooo`040000003oool0oooo000004d0oooo
00<000000?ooo`3oool06`3oool002H0oooo00<0_g@004RLg`3oool0:@3oool00`3ookl0M00004RL
g`030?ooo`030=nLB00007@0_ooo00L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`130?ooo`040000003oool0oooo000004@0oooo00<000000?ooo`3oool0AP3oool010000000oooo
0?ooo`00001=0?ooo`030000003oool0oooo01/0oooo000V0?ooo`060;md00000000000000000000
07@0_ooo9P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`140?ooo`800000AP3oool00`000000oooo0?ooo`160?oo
o`800000CP3oool00`000000oooo0?ooo`0K0?ooo`009P3oool0102oW9`0gooo0?ooo`3o_g@30000
00030000M02oool0oooo0280oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?oo
ool00000<P00000M0?ooo`00:`3oool0103ogi`0B0000018W03OoolS0?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`1<0?ooo`800000A`3oool00`00
0000oooo0?ooo`150?ooo`800000AP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
01/0oooo000Y0?ooo`040?oOW018000004RL0=ooobD0oooo00<0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`070?ooo`030000003oool0oooo04`0oooo0P0000180?ooo`030000003oool0oooo
04@0oooo0P0000160?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06`3oool002H0
oooo00D0_ibL0=oOW018000004RL0=ooo`0W0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000
M02oool01`3oool00`000000oooo0?ooo`1;0?ooo`040000003oool0oooo000004L0oooo00<00000
0?ooo`3oool0@`3oool010000000oooo0?ooo`0000140?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool06`3oool002H0oooo00<0_g@00018W03Oool0:P3oool00`3OW4P004RL0=ooo`02
0?ooo`030;md0000M;l0oooo00L0oooo00<000000?ooo`3oool0B`3oool010000000oooo0?ooo`00
00180?ooo`030000003oool0oooo0480oooo00@000000?ooo`3oool00000A03oool00`000000oooo
0?ooo`060?ooo`030000003oool0oooo01/0oooo000V0?ooo`030;nLW03Oool0oooo02X0oooo00L0
oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00<000000?ooo`3oool0BP3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0@@3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`0K0?ooo`00E03oool01@3ookl0M0000000000007@0_ooo00T0oooo00<000000?ooo`3o
ool0BP3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3o
ool0@03oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0400oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`0K0?ooo`00:`3oool01P3ooml0W4P00018W03Oool0oooO09bL
_c40oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
04D0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
03h0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0K0?ooo`00:`3oool01P3ookl0
M0000018W03Oool0_gA809cOoc40oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0K
0?ooo`00;03oool0103OW4P000000000001d_olb0?ooo`800000B@3oool00`000000oooo0?ooo`06
0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`06
0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0203oool2000001d0oooo001R0?oo
o`030000003oool0oooo04P0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`160?oo
o`030000003oool0oooo03`0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0l0?oo
o`030000003oool0oooo00T0oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3o
ool0A`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3o
ool0>`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0K0?ooo`009`3oool01`2oM000001809cOo`3ookl0M0000000
0018W=l0=03oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo00P0oooo00<000000?oo
o`3oool0AP3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo00P0oooo00<000000?oo
o`3oool0>P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo01/0oooo000V0?ooo`05
0?noM0000000001809cOo`2oM0000P0000000`0007@0_ooo0?ooo`0b0?ooo`030000003oool0oooo
04H0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo
03P0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool06`3oool002H0oooo00T0_gAd0;ooo`3oool0oono07A8W03Oool0
omnL04Q8W03Oool0<`3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo00X0oooo00<0
00000?ooo`3oool0AP3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo00X0oooo00<0
00000?ooo`3oool0=`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01/0oooo000V
0?ooo`080;mdB02Lgol0oooo0=nLM02Lgol0oooo0?noW02Lgold0?ooo`030000003oool0oooo04D0
oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo03H0
oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo00`0
oooo00<000000?ooo`3oool06`3oool002H0oooo00<0gia8000000000000100000000`00B9`0gooo
0?ooo`0b0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`160?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`0e0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool06`3oool002H0oooo00<0
oono07@000000000100000000`00B9`0gooo0?ooo`0b0?ooo`030000003oool0oooo04@0oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo03@0oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool06`3oool002`0oooo00<0_gA809cOo`3oool0<`3oool2000004D0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo03@0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo00d0oooo0P00000M
0?ooo`00HP3oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0AP3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0<P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01/0oooo001R0?ooo`03
0000003oool0oooo04<0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`160?ooo`03
0000003oool0oooo0380oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0a0?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool0
@P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0
<03oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0
3`3oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`120?ooo`030000003o
ool0oooo0180oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0`0?ooo`030000003o
ool0oooo0180oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo01/0oooo001R0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool0503oool0
0`000000oooo0?ooo`160?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0503oool0
0`000000oooo0?ooo`0^0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool06`3oool0
0680oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo
04L0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo
02`0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000
oooo0?ooo`100?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0AP3oool00`000000
oooo0?ooo`0/0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0:`3oool00`000000
oooo0?ooo`0B0?ooo`030000003oool0oooo01/0oooo001R0?ooo`030000003oool0oooo03l0oooo
00<000000?ooo`3oool0603oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo02X0oooo
00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool06`3oool00680oooo0P0000100?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool0AP3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool0:@3oool00`000000oooo0?ooo`0B0?ooo`8000007@3oool00680oooo00<00000
0?ooo`3oool0?P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo04H0oooo00<00000
0?ooo`3oool0:03oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo02P0oooo00<00000
0?ooo`3oool04`3oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0n0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0W0?oo
o`030000003oool0oooo01/0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0D0?oo
o`030000003oool0oooo01/0oooo001R0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3o
ool0703oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3o
ool0703oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3o
ool06`3oool00680oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`0M0?ooo`030000
003oool0oooo04D0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0M0?ooo`030000
003oool0oooo02D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0K0?ooo`00HP3o
ool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0AP3o
ool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool08`3o
ool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01/0oooo001R0?ooo`030000003oool0
oooo03`0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`150?ooo`030000003oool0
oooo02<0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool0>`3oool00`00
0000oooo0?ooo`0Q0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool08@3oool00`00
0000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool05`3oool00`00
0000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo0280
oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0280
oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01/0
oooo001R0?ooo`800000>`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo04D0oooo
00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo01l0oooo
00<000000?ooo`3oool05`3oool2000001d0oooo001R0?ooo`030000003oool0oooo03T0oooo00<0
00000?ooo`3oool0903oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo01h0oooo00<0
00000?ooo`3oool0903oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0U
0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0U
0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0K
0?ooo`00HP3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo02L0oooo00<000000?oo
o`3oool0A@3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo02H0oooo00<000000?oo
o`3oool06`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01/0oooo00170?ooo`05
0?oOW018000000000000M02oool00`3oool01P3o_g@0000007Boo`3oool0oooO09a800<0000000<0
M;oo0?ooo`3oool01`3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo02L0oooo00<0
00000?ooo`3oool0A@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo02L0oooo00<0
00000?ooo`3oool06`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01/0oooo0016
0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo0?ooo`3o
oml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00L0oooo00<000000?ooo`3oool0=`3oool00`00
0000oooo0?ooo`0Y0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool06@3oool00`00
0000oooo0?ooo`0Y0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06`3oool00`00
0000oooo0?ooo`0K0?ooo`00AP3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00X0
oooo00<0oooO09a8M02oool01`3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo02X0
oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo02T0
oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01/0
oooo00150?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02`3oool00`2oM000M;oo
0?ooo`060?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0:`3oool00`000000oooo
0?ooo`150?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0:`3oool00`000000oooo
0?ooo`0G0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool06`3oool004D0oooo00<0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030?oog`2LB000B9cO00L0oooo00<0
00000?ooo`3oool0=P3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo04@0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool0703oool00`000000oooo0?ooo`0K0?ooo`00A@3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo00X0oooo00<0gia80000001d_ol01`3oool3000003D0oooo00<00000
0?ooo`3oool0;@3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo01D0oooo00<00000
0?ooo`3oool0;@3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo01/0oooo0`00000M
0?ooo`00A@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO0?oo
o`3ookl0M0000000000007@0_ooo00L0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?oo
o`0_0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?oo
o`0_0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?oo
o`0K0?ooo`00A@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00H0gia8
00000000000000000000B02Lgol80?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0
<03oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0
;`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0
6`3oool004D0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d
0000000004RLg`X0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0a0?ooo`030000
003oool0oooo04D0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0a0?ooo`030000
003oool0oooo0140oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0K0?ooo`00AP3o
ool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0_g@007Boo`3oool02P3o
ool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0A@3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo01/0oooo00160?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`060?oOW018000000000000000004P0W=oo1`3oool00`00
0000oooo0?ooo`0b0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0A@3oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool03`3oool00`00
0000oooo0?ooo`0P0?ooo`030000003oool0oooo01/0oooo00170?ooo`050?oo_`1d000000000000
M02oool02P3oool01@2oM0000000000000000000B9cO00L0oooo00<000000?ooo`3oool0<@3oool0
0`000000oooo0?ooo`0e0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0e0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool08@3oool0
0`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo
03L0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
03L0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo
01/0oooo001R0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0=`3oool00`000000
oooo0?ooo`150?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0=`3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool06`3oool00680oooo
0P00000`0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0A@3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`0R0?ooo`8000007@3oool00680oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?oo
o`0j0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`0k0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?oo
o`0K0?ooo`00HP3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo03/0oooo00<00000
0?ooo`3oool0A@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo03/0oooo00<00000
0?ooo`3oool01`3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo01/0oooo001R0?oo
o`030000003oool0oooo02d0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`150?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo02D0oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3o
ool0;03oool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool09@3oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0/0?ooo`030000
003oool0oooo03l0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo03l0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0V0?ooo`030000
003oool0oooo01/0oooo001R0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0@@3o
ool00`000000oooo0?ooo`150?ooo`050000003oool0oooo0?ooo`000000@`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool06`3oool00680oooo00<0
00000?ooo`3oool0:P3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo04H0oooo00<0
00000?ooo`000000A@3oool010000000oooo0?ooo`00000Y0?ooo`030000003oool0oooo01/0oooo
001R0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?oo
o`150?ooo`030000003oool0000004D0oooo00<000000?ooo`000000:P3oool00`000000oooo0?oo
o`0K0?ooo`00HP3oool2000002X0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`16
0?ooo`030000003oool0oooo04D0oooo0P00000Y0?ooo`8000007@3oool00680oooo00<000000?oo
o`3oool0:03oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo04@0oooo00<000000?oo
o`000000AP3oool2000002X0oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3o
ool09`3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo0480oooo00@000000?ooo`3o
ool00000A@3oool00`000000oooo0000000Z0?ooo`030000003oool0oooo01/0oooo001R0?ooo`03
0000003oool0oooo02H0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`120?ooo`05
0000003oool0oooo0?ooo`000000@`3oool01@000000oooo0?ooo`3oool0000002T0oooo00<00000
0?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`1:0?oo
o`030000003oool0oooo0400oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0o0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0K0?oo
o`00HP3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3o
ool0?P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo01/0oooo001R0?ooo`030000
003oool0oooo02@0oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`0l0?ooo`030000
003oool0oooo00H0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo02D0oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool08`3o
ool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0903o
ool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0
oooo0540oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo03T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0
oooo01/0oooo001R0?ooo`8000008P3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo
03P0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool08@3oool2000001d0oooo001R0?ooo`030000003oool0oooo0200
oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0240
oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool0803oool00`000000oooo
0?ooo`1F0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool03`3oool00`000000oooo
0?ooo`0c0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0803oool00`000000oooo
0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo05P0oooo00<0
00000?ooo`3oool0<P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo03<0oooo00<0
00000?ooo`3oool0403oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01/0oooo001R
0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`0a
0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0B
0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?oo
o`3oool07@3oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo02l0oooo00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01@0oooo00<000000?oo
o`3oool07@3oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0L0?ooo`03
0000003oool0oooo05h0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0E0?ooo`03
0000003oool0oooo02d0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0L0?ooo`03
0000003oool0oooo01/0oooo001R0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0
H03oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
:`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0
6`3oool00680oooo00<000000?ooo`3oool06@3oool2000006@0oooo00<000000?ooo`3oool0:@3o
ool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01/0oooo001R0?ooo`8000006@3oool0
0`000000oooo0?ooo`1U0?ooo`030000003oool0oooo02H0oooo0P00000M0?ooo`800000:03oool0
0`000000oooo0?ooo`0L0?ooo`8000006@3oool2000001d0oooo001R0?ooo`030000003oool0oooo
01L0oooo00<000000?ooo`3oool0I`3oool2000002D0oooo00<000000?ooo`3oool07`3oool00`00
0000oooo0?ooo`0T0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool05`3oool00`00
0000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo06X0
oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0280
oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01/0
oooo001R0?ooo`030000003oool0oooo01@0oooo0P00001^0?ooo`8000007`3oool2000002D0oooo
0P00000P0?ooo`8000009@3oool2000001H0oooo00<000000?ooo`3oool06`3oool005@0oooo00H0
_g@000000000000000000018W03Oool80?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3o
ool0L03oool2000001`0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0L0?ooo`03
0000003oool0oooo02L0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0K0?ooo`00
E@3oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo0140oooo0P00001e0?ooo`800000
603oool2000002/0oooo0P00000J0?ooo`800000:`3oool2000001<0oooo00<000000?ooo`3oool0
6`3oool005D0oooo00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`0?0?ooo`800000N@3o
ool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo02d0oooo0P00000E0?ooo`<00000;`3o
ool200000140oooo00<000000?ooo`3oool06`3oool005D0oooo00<0omnL04P0001d_ol02P3oool0
0`000000oooo0?ooo`0=0?ooo`800000O03oool300000100oooo0`00000b0?ooo`8000004@3oool2
000003@0oooo0P00000?0?ooo`030000003oool0oooo01/0oooo001E0?ooo`030?oOW0180000M;oo
00X0oooo00<000000?ooo`3oool0203oool500000840oooo100000070?ooo`D00000=`3oool60000
00H0oooo1@00000h0?ooo`D000002P3oool00`000000oooo0?ooo`0K0?ooo`00E@3oool00`3ogi`0
B00007Boo`0:0?ooo`<00000103oool4000008X0oooo1`0000120?ooo`H00000@P3oool4000000@0
oooo0`00000M0?ooo`00E@3oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo0?l0oooo
;P3oool00`000000oooo0?ooo`0K0?ooo`00E@3oool00`3ogi`0B00007Boo`0:0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`0[0?ooo`030000003o
ool0oooo02/0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0[0?ooo`030000003o
ool0oooo02/0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0K0?ooo`00E@3oool0
0`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0000001d0oooo001E0?ooo`030?oOW0180000M;oo
00X0ooooo`00000b000001d0oooo001C0?ooo`050?oog`2LB00000000000001d_ol0o`3ooomI0?oo
o`00E@3oool00`3o_g@0000007Boo`3o0?oooeT0oooo003o0?oook40oooo003o0?oook40oooo003o
0?oook40oooo003o0?oook40oooo002j0?ooo`060?oog`2LB00000000000000007@0_oool03oool0
0;`0oooo00<0_g@00018W03Oool0l@3oool00;`0oooo00<0_g@00018W03Oool0l@3oool00;`0oooo
00<0_g@00018W03Oool0l@3oool00:80oooo00<0oono0;no_`1d00000P0000000`1d_ol0oooo0=nL
B002000000070000M02LW4P000000000000004P0W=oO09a80002000000030000M02ooml0W4P00080
000000D007Bo0?ooo`2oM00004Q804P000020000000404RLg`3oool0oooo0=nLB080000000@0001d
0;ooo`3oool0_g@00P0000000`0007@0_ooo0?ooo`070?ooo`060=nLB000000000000000B02Lgml0
W4P00P0000000`00M;l0oooO09a80002000000050018W02LM0000000001d_`2oM0000P0000001000
04P0W=oo0?ooo`3OW4P2000000030000B02Lgml0W4P00080000000<0001d0;oog`2LB0000P000000
1000M;l0oooo0?ooo`3OW4P30000000407Boo`3oool0oooo0;md0080000000<0001d0;oog`2LB000
0P0000002000B9`0_ia8000000000000001d0;ooo`3oool0_g@00P0000000`0004P0W=oo0?ooo`05
0?ooo`060=nLB000000000000000B02Lgml0W4P00P0000000`0007@0_ooO09a8000200000004001d
_`3oool0oooo0=nLB0<0000000T0M;oo0?ooo`3oool0oooo0?oOW0180000000004RLg`3OW4P00P00
0000100007@0_ooo0?ooo`3OW4P30000000407Boo`3oool0oooo0;md0080000000<0001d0;oog`2L
B0000P0000000`00B9`0gooo0?ooo`1M0?ooo`00XP3oool03`3ookl0M01809cOo`3oool0oono07@0
B02Lgol0oooo0=nLB00007@0_ooo0?ooo`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?oo
o`3ooml0W4P0001d_`020?ooo`070;md00000000M;oo0?ooo`3ookl0M00004RLg`020?ooo`0;0=nL
B00007@0_ooo0?ooo`2oM0000000001d_`3oool0omnL04P0M02oool0203oool0503OW4P0001d0;oo
o`3oool0oooO09a80000M;l0oooo0?noM0000000M;oo0?ooo`3ogi`0B000001d_`3o_kl0_ooo0=nL
B00007@0_ooo0`3oool00`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0
001d_`020?ooo`070=nLB0000000B9cO0?ooo`3ooml0W4P004RLg`020?ooo`070=nLB000B9`0gooo
0?ooo`3ooml0W4P0001d_`020?ooo`0;0=nLB00007@0_ooo0?ooo`2oM00000000018W03Oool0oooO
09a8M02oool01P3oool00`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0
001d_`020?ooo`070=nLB0000000B9cO0?ooo`3ooml0W4P004RLg`020?ooo`0@0=nLB00007@0_ooo
0=no_`3Oool0gia80000M02oool0oooo0=nLB0000000B9cO0?ooo`3ooml0W4P004RLg`80oooo00L0
gia80018W03Oool0oooo0?oog`2LB00007Bo0600oooo002R0?ooo`0?0?oo_`2o_ol0oooo0?ooo`3o
_g@0000007Boo`3oool0gia80000M02oool0oooo0=nLB00007@0_ooo0080oooo00L0_g@00018W03O
ool0oooo0?oog`2LB00007Bo0080oooo00<0_g@00018W03Oool00P3oool02P3o_g@0001d0;ooo`3o
ool0gia80000M02oool0omnL04P00018W=l30?ooo`030?oOW01dW=l0oooo00L0oooo00l0gia80000
M02oool0oooo0?oog`2LB00007Bo0?ooo`3o_g@0001809cOo`3oool0oooO09a80000M;l00P3oool0
0`3OW4P0001d0;ooo`030?ooo`030=nLB00007@0_ooo0080oooo00/0_g@00018W03Oool0oooo0?oo
g`2LB00007Bo0?ooo`3ogi`0B000001d_`030?ooo`0;0?noM00007@0_ooo0?ooo`3OW4P004RL0=oo
o`3oool0oooO09a80000M;l00P3oool01P3OW4P0001d0;ooo`3ogi`0B000001d_`<0oooo00<0omnL
09boo`3oool01@3oool00`3OW4P0001d0;ooo`020?ooo`0;0;md0000B9`0gooo0?ooo`3ooml0W4P0
001d_`3oool0omnL04P00000M;l00`3oool01`3o_g@0001d0;ooo`3oool0gia80018W03Oool00P3o
ool01P3OW4P0001d0;ooo`3ogi`0B000001d_`<0oooo00/0okmd0000M02oool0oooo0=nLB000B9`0
gooo0?ooo`3ooml0W4P0001d_`1P0?ooo`00Y03oool03@3ookl0M0000000000007@0_ooo0?ooo`3O
W4P0001d0;ooo`3oool0gia80000M02oool00P3oool01`2oM00004RL0=ooo`3oool0oooO09a80000
M;l00P3oool00`2oM00004RL0=ooo`020?ooo`0:0?oOW0180000M;oo0?ooo`3OW4P0001d0;ooo`3o
_g@0001809cOo`d0oooo00L0gia80000M02oool0oooo0?oog`2LB00007Bo0080oooo00H0okmd0000
0000M;l0oooO09a80000M;l20?ooo`030=nLB00007@0_ooo00<0oooo00<0gia80000M02oool00P3o
ool02`2oM00004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM0000000M;oo00<0oooo00/0okmd
0000001d_ol0oooo0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`060=nLB00007@0_ooo
0?noM00004P0W=oo2`3oool00`3OW4P0001d0;ooo`020?ooo`0;0;md0000B9`0gooo0?ooo`3ooml0
W4P0001d_`3oool0okmd0000001d_ol00`3oool01`3o_g@0000007Boo`3oool0gia80018W03Oool0
0P3oool01P3OW4P0001d0;ooo`3o_g@0000007Boo`<0oooo00/0okmd0000001d_ol0oooo0=nLB000
B9`0gooo0?ooo`3ooml0W4P0001d_`1P0?ooo`00X`3oool0102oM000000000000018W=l30?ooo`07
0=nLB00007@0_ooo0?ooo`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0
001d_`020?ooo`030;md0000B9`0gooo0080oooo00X0omnL04P0001d_ol0oooo0=nLB00007@0_ooo
0?noM00007@0_ooo3@3oool01`3OW4P0001d0;ooo`3oool0oooO09a80000M;l0103oool0103o_g@0
000000000000M;l20?ooo`030=nLB00007@0_ooo00<0oooo00<0gia80000M02oool00P3oool02`2o
M00004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM0000000M;oo00<0oooo00/0okmd0000001d
_ol0oooo0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`060=nLB00007@0_ooo0?noM000
07@0_ooo2`3oool00`3OW4P0001d0;ooo`020?ooo`0;0;md0000B9`0gooo0?ooo`3ooml0W4P0001d
_`3oool0okmd0000001d_ol00`3oool01`3o_g@0000007Boo`3oool0gia80018W03Oool00P3oool0
1P3OW4P0001d0;ooo`3o_g@0000007Boo`<0oooo00/0okmd0000001d_ol0oooo0=nLB000B9`0gooo
0?ooo`3ooml0W4P0001d_`1P0?ooo`00XP3oool00`3ogi`0B00004RLg`020?ooo`0:0;nLW03Oool0
oooo0=nLB00007@0_ooo0?ooo`3OW4P0001809cOo`80oooo00L0_g@00000M02oool0oooo0?oog`2L
B00007Bo0080oooo00<0_g@00000M02oool00P3oool02@3OW4P0001809cOo`3oool0gia80000M02o
ool0omnL04P000050000000304RLg`3oool0oooo00H0oooo00l0gia80000M02oool0oooo0?oog`2L
B00007Bo0?ooo`3ookl0M00007Boo`3oool0oooO09a80000M;l00P3oool00`3OW4P0001809cOo`03
0?ooo`030=nLB00004P0W=oo0080oooo00/0_g@00000M02oool0oooo0?oog`2LB00007Bo0?ooo`3o
okl0M00007Boo`030?ooo`0;0;md000004P0W=oo0?ooo`2oM000001d0;ooo`3oool0oooO09a80000
M;l00P3oool01P3OW4P0001d0;ooo`3ogi`0B01809cOo`80oooo00<0okmd0000B02Lgol01P3oool0
0`3OW4P0001809cOo`020?ooo`0;0;md000007@0_ooo0?ooo`3ooml0W4P0001d_`3oool0oono07@0
001d_ol00`3oool01`2oM000001809cOo`3oool0gia80018W03Oool00P3oool01P3OW4P0001d0;oo
o`3ookl0M00007Boo`<0oooo00/0_g@00000B02Lgol0oooo0;md000007@0_ooo0?ooo`3ooml0W4P0
001d_`1P0?ooo`00XP3oool0:@3ookl0M01d0;ooo`3oool0okmd0018W03Ooml0W4P00000000007@0
_ooO09a800000000001804Q8B02Lgol0oono07@00000M7@0M01d0;ooo`3ookl0M00004RLg`3ookl0
M00000000000B4P0B4RL0=ooo`3ogi`0B0000018W03Oool0oooo0=nLB00007@0_ooo0?ooo`2oM000
B9cO0080oooo00<0gia80000B02Lgol0203oool08`3OW4P000000000B02Lgol0oono07@00018W=l0
oooo0?oo_`1d0000B9cO0?ooo`3ookl0M000001d_`3ooml0W4P00000000004P0B7BL0=noM00004P0
B4P00000000004P0B4Q809cOo`3ookl0M000001dM01d07@0_ooo0?oo_`1d0000B9cO0080oooo01d0
gia80018W03Oool0oooo0?noM000000007Bo0?oog`2LB00000000018B01804P0W=oo0?oo_`1d0000
B9cO0?oog`2LB00000000000M02oool0oooo0;md0000B9`0gooo0?oog`2LB000001d0;ooo`040?oo
o`0>0?oog`2LB00000000000B018B4P0W=oo0?oo_`1d000007Ad07@0M02oool0oono07@00018W=l2
0?ooo`070=nLB000B9`0gooo0?ooo`3o_g@00000001d_`020?ooo`0I0=nLB000B9`0gooo0?oog`2L
B00000000000M02oool0oooo0=nLB000B9`0gooo0?ooo`3o_g@00000001d_`3ooml0W4P000000000
B4P0B01809cOo`3ookl0M00004RLg`1P0?ooo`00X`3oool0803o_g@0000000000000M;l0_ibL0=oo
o`3oool0_g@00000M02oool0oooo0;md000007@0_ooO09a800000000001809cOo`3ookl0M0000000
000007@0_ooo0?ooo`3ooml0W4P00018W03Ogi`0B0000000000007@0_ooo0P3oool00`3OW4P0001d
0;ooo`020?ooo`050=nLB000000000000000B02Lgol02@3oool01`3OW4P0001d0;oog`2LB0000000
0000M02oool00P3oool01P3ookl0M000000000000000001d0;ooo`80oooo0180_g@00000M02ogi`0
B0000000M02oool0_g@00000M02ooml0W4P00000000004P0W=oo0?oo_`1d000000000000M02oool3
0?ooo`050?noM0000000000000000018W=l00`3oool01`2oM00004RL0=oog`2LB00000000000M02o
ool00P3oool00`2oM000001d0;ooo`020?ooo`050=nLB000000000000000B02Lgol01`3oool0302o
M000001d0;oog`2LB00000000000B02Lgol0oono07@000000000001d0;ooo`<0oooo00D0okmd0000
00000000000004RLg`020?ooo`090;md00000000000000000000B9`0gooo0;md000007@0_ooo0080
oooo00D0okmd000000000000000004RLg`030?ooo`070;md0000B9`0gooO09a800000000001d0;oo
o`1P0?ooo`00a@3oool00`3OW4P0001d0;ooo`0@0?ooo`030=nLB00007@0_ooo05`0oooo00<0_g@0
0018W03Oool0MP3oool00<D0oooo00<0gia80000M02oool0403oool00`3OW4P0001d0;ooo`1L0?oo
o`030?oo_`1dB9`0gooo07H0oooo00350?ooo`030=nLB00007@0_ooo0100oooo00<0gia80000M02o
ool0G03oool00`3ooml0W7BL0=ooo`1f0?ooo`00ZP3oool00`3OW4P0001809cOo`0F0?ooo`050?oo
g`2LB00000000000M02oool03P3oool01@3ooml0W4P00000000007@0_ooo03@0oooo00<0gia80000
B02Lgol0:P3oool00`3OW4P0001809cOo`1a0?ooo`00ZP3oool00`3OW4P0001809cOo`0H0?ooo`03
0;md000007@0_ooo0100oooo00<0_g@00000M02oool0=03oool00`3OW4P0001809cOo`0Z0?ooo`03
0=nLB00004P0W=oo0740oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-2.3004, -1.61905, \
0.0325864, 0.0167831}}]
}, Closed]],

Cell[TextData[{
  StyleBox["DisplayTogether", "InlineInput"],
  StyleBox[" also accepts the formatting options recognized by ",
    FontWeight->"Plain"],
  StyleBox["Show", "InlineInput"],
  StyleBox[".",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell["\<\
Recreate the figure above using Show instead of DisplayTogether.  Note that \
without manipulating DisplayFunction, Show produces intermediate plots that \
you probably do not want.\
\>", "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Creating a graphics array", "Subsection"],

Cell[TextData[{
  "A graphics array is created using ",
  ButtonBox["GraphicsArray",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[list]", "InlineInput",
    FontWeight->"Bold"],
  " where the argument is either a 1-dimensional or 2-dimensional list.  \
Because ",
  StyleBox["GraphicsArray", "InlineInput",
    FontWeight->"Bold"],
  " comes with its own default ",
  StyleBox["DisplayFunction", "InlineInput",
    FontWeight->"Bold"],
  " option, ",
  StyleBox["Show", "InlineInput",
    FontWeight->"Bold"],
  " will then display the plot without resetting the display function.  Note \
that even if the list contains only one element, it must still include curly \
braces so that it is recognized as having the correct type. "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[GraphicsArray[{plot[x]}]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0147151 0.952381 [
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0147151 0.97619 0.603319 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.151576 0.309016 0.294303 [
[.17539 .29652 -5 -15.6875 ]
[.17539 .29652 5 0 ]
[.32696 .29652 -5 -15.6875 ]
[.32696 .29652 5 0 ]
[.47854 .29652 -5 -15.6875 ]
[.47854 .29652 5 0 ]
[.63011 .29652 -5 -15.6875 ]
[.63011 .29652 5 0 ]
[.78169 .29652 -5 -15.6875 ]
[.78169 .29652 5 0 ]
[.93327 .29652 -5 -15.6875 ]
[.93327 .29652 5 0 ]
[.01131 .01471 -18.3125 -7.84375 ]
[.01131 .01471 0 7.84375 ]
[.01131 .16186 -27.3125 -7.84375 ]
[.01131 .16186 0 7.84375 ]
[.01131 .45617 -19 -7.84375 ]
[.01131 .45617 0 7.84375 ]
[.01131 .60332 -10 -7.84375 ]
[.01131 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.17539 .30902 m
.17539 .31527 L
s
gsave
.17539 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.32696 .30902 m
.32696 .31527 L
s
gsave
.32696 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.47854 .30902 m
.47854 .31527 L
s
gsave
.47854 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.63011 .30902 m
.63011 .31527 L
s
gsave
.63011 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.78169 .30902 m
.78169 .31527 L
s
gsave
.78169 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.93327 .30902 m
.93327 .31527 L
s
gsave
.93327 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.125 Mabswid
.05412 .30902 m
.05412 .31277 L
s
.08444 .30902 m
.08444 .31277 L
s
.11476 .30902 m
.11476 .31277 L
s
.14507 .30902 m
.14507 .31277 L
s
.2057 .30902 m
.2057 .31277 L
s
.23602 .30902 m
.23602 .31277 L
s
.26633 .30902 m
.26633 .31277 L
s
.29665 .30902 m
.29665 .31277 L
s
.35728 .30902 m
.35728 .31277 L
s
.38759 .30902 m
.38759 .31277 L
s
.41791 .30902 m
.41791 .31277 L
s
.44822 .30902 m
.44822 .31277 L
s
.50885 .30902 m
.50885 .31277 L
s
.53917 .30902 m
.53917 .31277 L
s
.56948 .30902 m
.56948 .31277 L
s
.5998 .30902 m
.5998 .31277 L
s
.66043 .30902 m
.66043 .31277 L
s
.69074 .30902 m
.69074 .31277 L
s
.72106 .30902 m
.72106 .31277 L
s
.75138 .30902 m
.75138 .31277 L
s
.81201 .30902 m
.81201 .31277 L
s
.84232 .30902 m
.84232 .31277 L
s
.87264 .30902 m
.87264 .31277 L
s
.90295 .30902 m
.90295 .31277 L
s
.96358 .30902 m
.96358 .31277 L
s
.9939 .30902 m
.9939 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.02381 .01471 m
.03006 .01471 L
s
gsave
.01131 .01471 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .16186 m
.03006 .16186 L
s
gsave
.01131 .16186 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .45617 m
.03006 .45617 L
s
gsave
.01131 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .60332 m
.03006 .60332 L
s
gsave
.01131 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04414 m
.02756 .04414 L
s
.02381 .07357 m
.02756 .07357 L
s
.02381 .103 m
.02756 .103 L
s
.02381 .13243 m
.02756 .13243 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .25016 m
.02756 .25016 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .36788 m
.02756 .36788 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .4856 m
.02756 .4856 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .60332 m
.02499 .60331 L
.02605 .60329 L
.02729 .60324 L
.02846 .60318 L
.03053 .60303 L
.03279 .6028 L
.03527 .60248 L
.0379 .60205 L
.04262 .60106 L
.04749 .59974 L
.05205 .59822 L
.06244 .59381 L
.07305 .58793 L
.08274 .58135 L
.10458 .56251 L
.14429 .51515 L
.18248 .45627 L
.22313 .38349 L
.26226 .30833 L
.30384 .22863 L
.34391 .15746 L
.38246 .09886 L
.402 .07411 L
.42346 .05145 L
.44388 .03465 L
.45301 .02874 L
.46295 .02346 L
.46838 .0211 L
.4734 .01923 L
.4781 .01778 L
.4833 .0165 L
.48553 .01605 L
.48794 .01564 L
.49002 .01535 L
.4923 .01509 L
.49473 .01489 L
.49605 .01481 L
.49728 .01476 L
.49844 .01473 L
.49949 .01472 L
.50071 .01472 L
.50186 .01474 L
.50301 .01477 L
.50425 .01483 L
.50542 .0149 L
.50648 .01498 L
.5091 .01524 L
.51158 .01557 L
.5165 .01646 L
Mistroke
.52187 .01777 L
.53144 .02102 L
.54032 .02507 L
.56088 .03814 L
.5795 .05428 L
.61962 .1017 L
.65822 .16102 L
.69928 .23447 L
.73882 .31042 L
.77684 .38343 L
.81732 .45593 L
.85628 .51593 L
.87613 .54149 L
.89769 .56473 L
.91843 .58221 L
.92832 .58876 L
.93759 .59383 L
.94643 .59767 L
.95595 .6007 L
.9611 .60186 L
.96369 .60232 L
.96652 .60272 L
.96894 .60298 L
.9703 .6031 L
.97153 .60318 L
.97268 .60324 L
.97393 .60329 L
.97511 .60331 L
.97619 .60332 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00>03oool00`000000oooo0?ooo`3o0?ooogH0oooo000Y0?ooo`060;md
00000000000000000000B9`0gooo2@3oool00`000000oooo0?ooo`3o0?ooogH0oooo000Z0?ooo`03
0?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0o`3ooomf0?ooo`00:P3oool00`3ogi`0B000
07Boo`0;0?ooo`030000003oool0oooo0?l0ooooMP3oool002X0oooo00<0omnL04P0001d_ol02`3o
ool00`000000oooo0?ooo`3o0?ooogH0oooo000K0?ooo`030?oog`2LB000000000L0000000<0B9cO
0?ooo`3oool00P3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0?l0ooooMP3oool0
02X0oooo00<0omnL04P0001d_ol02`3oool300000:T0oooo3@00002o0?ooo`00:P3oool00`3ogi`0
B00007Boo`0;0?ooo`030000003oool0oooo0:H0oooo0`00000=0?ooo`<00000_03oool002X0oooo
00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`2S0?ooo`<000004`3oool200000;X0oooo
000Z0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0X@3oool2000001P0oooo0P00
002h0?ooo`00:P3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo09l0oooo0P00000L
0?ooo`030000003oool0oooo0;D0oooo000X0?ooo`050?oog`2LB00000000000001d_ol02`3oool0
0`000000oooo0?ooo`2M0?ooo`8000007`3oool200000;D0oooo000Z0?ooo`030?noM0000000M;oo
00/0oooo00<000000?ooo`3oool0V`3oool2000002<0oooo00<000000?ooo`3oool0/P3oool003P0
oooo00<000000?ooo`3oool0V@3oool2000002H0oooo0P00002b0?ooo`00>03oool00`000000oooo
0?ooo`2H0?ooo`030000003oool0oooo02P0oooo0P00002`0?ooo`00>03oool00`000000oooo0?oo
o`2G0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0[@3oool003P0oooo00<00000
0?ooo`3oool0UP3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo0:`0oooo000h0?oo
o`800000UP3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo0:/0oooo000h0?ooo`03
0000003oool0oooo09@0oooo00<000000?ooo`3oool0<@3oool200000:/0oooo000h0?ooo`030000
003oool0oooo09<0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`2X0?ooo`00>03o
ool00`000000oooo0?ooo`2B0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0Y`3o
ool003P0oooo00<000000?ooo`3oool0T@3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0
oooo0:H0oooo000h0?ooo`030000003oool0oooo0900oooo00<000000?ooo`3oool0>P3oool00`00
0000oooo0?ooo`2U0?ooo`00>03oool00`000000oooo0?ooo`2?0?ooo`030000003oool0oooo03`0
oooo00<000000?ooo`3oool0Y03oool003P0oooo00<000000?ooo`3oool0SP3oool00`000000oooo
0?ooo`0n0?ooo`030000003oool0oooo0:<0oooo000h0?ooo`030000003oool0oooo08d0oooo00<0
00000?ooo`3oool0@03oool00`000000oooo0?ooo`2R0?ooo`00>03oool00`000000oooo0?ooo`2<
0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool0XP3oool003P0oooo00<000000?oo
o`3oool0R`3oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo0:40oooo000h0?ooo`80
0000R`3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo0:00oooo000h0?ooo`030000
003oool0oooo08T0oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`2O0?ooo`00>03o
ool00`000000oooo0?ooo`280?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0WP3o
ool003P0oooo00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`1:0?ooo`030000003oool0
oooo09d0oooo000h0?ooo`030000003oool0oooo08L0oooo00<000000?ooo`3oool0C03oool00`00
0000oooo0?ooo`2L0?ooo`00>03oool00`000000oooo0?ooo`260?ooo`030000003oool0oooo04h0
oooo00<000000?ooo`3oool0V`3oool003P0oooo00<000000?ooo`3oool0Q@3oool00`000000oooo
0?ooo`1?0?ooo`030000003oool0oooo09/0oooo000h0?ooo`030000003oool0oooo08D0oooo00<0
00000?ooo`3oool0D03oool00`000000oooo0?ooo`2J0?ooo`00>03oool00`000000oooo0?ooo`24
0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0V@3oool003P0oooo00<000000?oo
o`3oool0P`3oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo09P0oooo000h0?ooo`03
0000003oool0oooo0880oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`2G0?ooo`00
>03oool2000008<0oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`2F0?ooo`00>03o
ool00`000000oooo0?ooo`210?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool0UP3o
ool003P0oooo00<000000?ooo`3oool0P03oool00`000000oooo0?ooo`1J0?ooo`030000003oool0
oooo09D0oooo000h0?ooo`030000003oool0oooo0800oooo00<000000?ooo`3oool0F`3oool00`00
0000oooo0?ooo`2D0?ooo`00>03oool00`000000oooo0?ooo`1o0?ooo`030000003oool0oooo05`0
oooo00<000000?ooo`3oool0U03oool003P0oooo00<000000?ooo`3oool0OP3oool00`000000oooo
0?ooo`1N0?ooo`030000003oool0oooo09<0oooo000h0?ooo`030000003oool0oooo07h0oooo00<0
00000?ooo`3oool0GP3oool00`000000oooo0?ooo`2C0?ooo`00>03oool00`000000oooo0?ooo`1m
0?ooo`030000003oool0oooo0600oooo00<000000?ooo`3oool0TP3oool003P0oooo00<000000?oo
o`3oool0O03oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo0940oooo000h0?ooo`03
0000003oool0oooo07/0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`2A0?ooo`00
>03oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0
T03oool003P0oooo0P00001k0?ooo`030000003oool0oooo06H0oooo00<000000?ooo`3oool0S`3o
ool003P0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`1W0?ooo`030000003oool0
oooo08l0oooo000h0?ooo`030000003oool0oooo07T0oooo00<000000?ooo`3oool0J03oool00`00
0000oooo0?ooo`2>0?ooo`00>03oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo06X0
oooo00<000000?ooo`3oool0S@3oool003P0oooo00<000000?ooo`3oool0M`3oool00`000000oooo
0?ooo`1[0?ooo`030000003oool0oooo08d0oooo000h0?ooo`030000003oool0oooo07L0oooo00<0
00000?ooo`3oool0K03oool00`000000oooo0?ooo`2<0?ooo`00703oool01@3ogi`0B00000000000
07@0_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oog`2LB0030000000307Boo`3oool0oooo00P0
oooo00<000000?ooo`3oool0MP3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo08`0
oooo000K0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo
0?ooo`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00P0oooo00<000000?ooo`3oool0M@3o
ool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo08/0oooo000K0?ooo`030;md0000B9`0
gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd0;ooo`080?ooo`030000003oool0
oooo07D0oooo00<000000?ooo`3oool0L03oool00`000000oooo0?ooo`2:0?ooo`006P3oool00`3o
okl0M00004RLg`030?ooo`030=nLB00007@0_ooo00/0oooo00<0_g@007Boo`3oool01`3oool00`00
0000oooo0?ooo`1d0?ooo`030000003oool0oooo0740oooo00<000000?ooo`3oool0RP3oool000d0
oooo00<0oooO09a8000000001`0000001P18W=l0oooo0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd
0000B02Lgol02P3oool00`3ooml0W4P004RLg`080?ooo`030000003oool0oooo07<0oooo00<00000
0?ooo`3oool0L`3oool00`000000oooo0?ooo`290?ooo`006P3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00X0oooo00<0gia80000001d_ol0203oool3000007<0oooo00<000000?oo
o`3oool0M03oool00`000000oooo0?ooo`280?ooo`006P3oool00`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00P0oooo00<00000
0?ooo`3oool0LP3oool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo08P0oooo000J0?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3OW4P00000000000000000
001809cOo`T0oooo00<000000?ooo`3oool0LP3oool00`000000oooo0?ooo`1f0?ooo`030000003o
ool0oooo08L0oooo000J0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool0
103ookl0M00000000018W=l;0?ooo`030000003oool0oooo0740oooo00<000000?ooo`3oool0M`3o
ool00`000000oooo0?ooo`270?ooo`006`3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0
oooo00L0oooo00<0_g@007Boo`3oool02`3oool00`000000oooo0?ooo`1a0?ooo`030000003oool0
oooo07P0oooo00<000000?ooo`3oool0QP3oool001/0oooo00L0oono07@0001d_ol0oooo0?noM000
07@0_ooo00P0oooo00H0omnL04P00000000000000000B02Lgol80?ooo`030000003oool0oooo0700
oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`260?ooo`00703oool01@3ookl0M000
0000000007@0_ooo00X0oooo00D0_g@0000000000000000004RLg`080?ooo`030000003oool0oooo
0700oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`250?ooo`00>03oool00`000000
oooo0?ooo`1_0?ooo`030000003oool0oooo07`0oooo00<000000?ooo`3oool0Q03oool003P0oooo
00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo08@0oooo
000h0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool0OP3oool00`000000oooo0?oo
o`230?ooo`00>03oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo07l0oooo00<00000
0?ooo`3oool0P`3oool003P0oooo0P00001^0?ooo`030000003oool0oooo0800oooo00<000000?oo
o`3oool0PP3oool003P0oooo00<000000?ooo`3oool0K03oool00`000000oooo0?ooo`210?ooo`03
0000003oool0oooo0880oooo000h0?ooo`030000003oool0oooo06`0oooo00<000000?ooo`3oool0
PP3oool00`000000oooo0?ooo`210?ooo`00>03oool00`000000oooo0?ooo`1[0?ooo`030000003o
ool0oooo08@0oooo00<000000?ooo`3oool0P03oool003P0oooo00<000000?ooo`3oool0J`3oool0
0`000000oooo0?ooo`240?ooo`030000003oool0oooo0800oooo000h0?ooo`030000003oool0oooo
06X0oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`1o0?ooo`00>03oool00`000000
oooo0?ooo`1Y0?ooo`030000003oool0oooo08L0oooo00<000000?ooo`3oool0O`3oool003P0oooo
00<000000?ooo`3oool0J@3oool00`000000oooo0?ooo`280?ooo`030000003oool0oooo07h0oooo
000h0?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool0R@3oool00`000000oooo0?oo
o`1n0?ooo`00>03oool00`000000oooo0?ooo`1X0?ooo`030000003oool0oooo08X0oooo00<00000
0?ooo`3oool0O@3oool003P0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`2<0?oo
o`030000003oool0oooo07`0oooo000h0?ooo`800000J03oool00`000000oooo0?ooo`2<0?ooo`03
0000003oool0oooo07`0oooo000h0?ooo`030000003oool0oooo06H0oooo00<000000?ooo`3oool0
SP3oool00`000000oooo0?ooo`1k0?ooo`00>03oool00`000000oooo0?ooo`1V0?ooo`030000003o
ool0oooo08h0oooo00<000000?ooo`3oool0N`3oool003P0oooo00<000000?ooo`3oool0I@3oool0
0`000000oooo0?ooo`2@0?ooo`030000003oool0oooo07X0oooo000h0?ooo`030000003oool0oooo
06@0oooo00<000000?ooo`3oool0T@3oool00`000000oooo0?ooo`1j0?ooo`00>03oool00`000000
oooo0?ooo`1T0?ooo`030000003oool0oooo0980oooo00<000000?ooo`3oool0N@3oool003P0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`2C0?ooo`030000003oool0oooo07T0oooo
000h0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0U03oool00`000000oooo0?oo
o`1h0?ooo`00>03oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo09D0oooo00<00000
0?ooo`3oool0N03oool003P0oooo00<000000?ooo`3oool0HP3oool00`000000oooo0?ooo`2F0?oo
o`030000003oool0oooo07L0oooo000h0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3o
ool0U`3oool00`000000oooo0?ooo`1g0?ooo`00>03oool2000003@0oooo00H0_g@0000000000000
00000018W03OoolX0?ooo`030000003oool0oooo00D0oooo00<0_g@0000000000000100000000`18
W=l0oooo0?ooo`0_0?ooo`030?oo_`1d000000000080000000<0M;oo0?ooo`3oool0=03oool0103o
oml0W4P00018W03OoolJ0?ooo`030000003oool0oooo01@0oooo00<0oooO09a8000000000P000000
0`1d_ol0oooo0?ooo`0b0?ooo`050?oOW018000000000000001d_ol08`3oool003P0oooo00<00000
0?ooo`3oool0=03oool00`3ogi`0B00007Boo`0Y0?ooo`030000003oool0oooo00L0oooo00<0gia8
0000000000000`0000000`00B9`0gooo0?ooo`0_0?ooo`030?oOW0180000B9cO0080oooo00<0_g@0
04RLg`3oool0=03oool0103ooml0W4P00018W03OoolJ0?ooo`030000003oool0oooo01@0oooo00<0
oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`0a0?ooo`070?oOW0180000B9cO0?ooo`3ookl0
M00007Boo`0R0?ooo`00>03oool00`000000oooo0?ooo`0d0?ooo`030?oOW0180000M;oo02T0oooo
00<000000?ooo`3oool0203oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo03@0oooo
00<0oono07@0B02Lgol0=03oool0103ooml0W4P00018W03OoolK0?ooo`030000003oool0oooo01P0
oooo00<0oooO09a8M02oool0<03oool00`3ooml0W4P0001d_`030?ooo`030?noM000B9`0gooo0240
oooo000h0?ooo`030000003oool0oooo03@0oooo00<0omnL04P0001d_ol0:03oool00`000000oooo
0?ooo`0:0?ooo`030=nLB01d_ol0oooo03P0oooo00<0oono07@0001d_ol0<03oool00`2oM0000000
000000050000000307Boo`3oool0oooo01P0oooo00<000000?ooo`3oool06@3oool00`2oM000M;oo
0?ooo`0_0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol08@3oool003P0oooo00<0
00000?ooo`3oool0=03oool00`3ogi`0B00007Boo`0X0?ooo`030000003oool0oooo00/0oooo00<0
_gA809cOo`3oool0=`3oool00`3o_g@0000007Boo`0`0?ooo`080?oOW01dW=l0oooo0?ooo`3ooml0
W4P00018W03OoolL0?ooo`030000003oool0oooo01L0oooo00<0oooO09a80018W=l0<03oool00`3o
gi`0B00007Boo`030?ooo`030?noM00004P0W=oo0240oooo000h0?ooo`030000003oool0oooo03@0
oooo00<0omnL04P0001d_ol09`3oool00`000000oooo0?ooo`0<0?ooo`030?oog`2LB7@0_ooo03H0
oooo00@0omnL04P0000004P0W=oo<@3oool01`3OW4P0M;oo0?ooo`3ooml0W4P00018W03Oool0703o
ool00`000000oooo0?ooo`0G0?ooo`030=nLB0000000M;oo0300oooo00<0omnL04P0001d_ol00`3o
ool00`2oM000001809cOo`0Q0?ooo`00>03oool00`000000oooo0?ooo`0d0?ooo`030?oOW0180000
M;oo02L0oooo00<000000?ooo`3oool03@3oool00`3o_g@007Bo0?ooo`0c0?ooo`050?noM0000000
00000000B02Lgol0<P3oool01`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool07@3oool00`000000
oooo0?ooo`0B0?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`0`0?ooo`080?oo_`1d0000
000004RLg`3oool0okmd00000000M;lR0?ooo`00>03oool00`000000oooo0?ooo`0d0?ooo`030?oO
W0180000M;oo02H0oooo00<000000?ooo`3oool03P3oool00`3ookl0M01809cOo`0e0?ooo`030;md
0000M;l0oooo03<0oooo00H0omnL04Qd_`3ooml0W4P00018W03OoolM0?ooo`030000003oool0oooo
0180oooo00H0gia800000000000000000000B02Lgolb0?ooo`030;md0000B9`0_ia80080000000<0
04RL0=ooo`3oool08@3oool003P0oooo00<000000?ooo`3oool0=03oool00`3ogi`0B00007Boo`0V
0?ooo`030000003oool0oooo00h0oooo00<0oono07@00018W=l0=P3oool00`2oM000B9cO0?ooo`0c
0?ooo`050=nLB01d_ml0W4P00018W03Oool07P3oool00`000000oooo0?ooo`0A0?ooo`040?oo_`1d
0000000004RLgc@0oooo00<0omnL04P0001d_ol09P3oool003P0oooo00<000000?ooo`3oool0=03o
ool00`3ogi`0B00007Boo`0U0?ooo`030000003oool0oooo00X0oooo00<0oknL09cOo`3oool00P3o
ool00`3ogi`0B00004RLg`0a0?ooo`030?oog`2LW;l0oooo0080oooo00<0gia80000M02oool0<`3o
ool01@3ooml0W4Q804Q80000B9`0gooo01h0oooo00<000000?ooo`3oool04P3oool00`2oM000M;oo
0?ooo`0e0?ooo`030=nLB00004P0W=oo02D0oooo000h0?ooo`030000003oool0oooo0380oooo00D0
oooO09a800000000000007Boo`0U0?ooo`030000003oool0oooo00X0oooo00P0oooO09a8B02Lgol0
oooo0?oo_`1d0000001d0;oooc80oooo00H0okmd04RLg`3oool0oono07@00000M;le0?ooo`040?oO
W018000004RL0=oooal0oooo00<000000?ooo`3oool04@3oool01P3ogi`0B0000000000000000018
09cOoc<0oooo00<0okmd00000000M;l0903oool003P0oooo0P00000e0?ooo`030?noM0000000M;oo
02@0oooo00<000000?ooo`3oool0303oool01P3ooml0W4P0000000000000001d0;oooc@0oooo00D0
gia8000000000000001d0;ooo`0f0?ooo`030=nLB000B9`0gooo01l0oooo00<000000?ooo`3oool0
4P3oool01@2oM0000000000000000000B9cO03@0oooo00D0oooO09a800000000001d0;ooo`0Q0?oo
o`00>03oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo0:@0oooo00<000000?ooo`3o
ool0L03oool003P0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`2U0?ooo`030000
003oool0oooo0700oooo000h0?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool0YP3o
ool00`000000oooo0?ooo`1_0?ooo`00>03oool00`000000oooo0?ooo`1I0?ooo`030000003oool0
oooo0:L0oooo00<000000?ooo`3oool0K`3oool003P0oooo00<000000?ooo`3oool0F@3oool00`00
0000oooo0?ooo`2X0?ooo`030000003oool0oooo06h0oooo000h0?ooo`030000003oool0oooo05P0
oooo00<000000?ooo`3oool0Z@3oool00`000000oooo0?ooo`1^0?ooo`00>03oool00`000000oooo
0?ooo`1H0?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0K@3oool003P0oooo00<0
00000?ooo`3oool0E`3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo06d0oooo000h
0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0[03oool00`000000oooo0?ooo`1/
0?ooo`00>03oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo0:d0oooo00<000000?oo
o`3oool0K03oool002l0ooooo`00001g000000/0oooo000h0?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0000000;0?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`00
00002P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00/0oooo000h0?ooo`030000
003oool0oooo03D0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0G0?ooo`030000
003oool0oooo03D0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0V0?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0R0?ooo`00>03o
ool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo0;80oooo00<000000?ooo`3oool0J@3o
ool003P0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`2c0?ooo`030000003oool0
oooo06T0oooo000h0?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool0]03oool00`00
0000oooo0?ooo`1X0?ooo`00>03oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo0;D0
oooo00<000000?ooo`3oool0J03oool003P0oooo00<000000?ooo`3oool0DP3oool00`000000oooo
0?ooo`2f0?ooo`030000003oool0oooo06L0oooo000h0?ooo`030000003oool0oooo0540oooo00<0
00000?ooo`3oool0]`3oool00`000000oooo0?ooo`1W0?ooo`00>03oool00`000000oooo0?ooo`1A
0?ooo`030000003oool0oooo0;P0oooo00<000000?ooo`3oool0IP3oool003P0oooo00<000000?oo
o`3oool0D03oool00`000000oooo0?ooo`2j0?ooo`030000003oool0oooo06D0oooo000h0?ooo`80
0000D@3oool00`000000oooo0?ooo`2j0?ooo`030000003oool0oooo06D0oooo000h0?ooo`030000
003oool0oooo04l0oooo00<000000?ooo`3oool0_03oool00`000000oooo0?ooo`1T0?ooo`00>03o
ool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0I03o
ool003P0oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`2n0?ooo`030000003oool0
oooo06<0oooo000h0?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3oool0_`3oool00`00
0000oooo0?ooo`1S0?ooo`00>03oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo0<00
oooo00<000000?ooo`3oool0HP3oool003P0oooo00<000000?ooo`3oool0C03oool00`000000oooo
0?ooo`310?ooo`030000003oool0oooo0680oooo000h0?ooo`030000003oool0oooo04`0oooo00<0
00000?ooo`3oool0`P3oool00`000000oooo0?ooo`1Q0?ooo`00>03oool00`000000oooo0?ooo`1;
0?ooo`030000003oool0oooo0<@0oooo00<000000?ooo`3oool0H03oool003P0oooo00<000000?oo
o`3oool0B`3oool00`000000oooo0?ooo`340?ooo`030000003oool0oooo0600oooo000h0?ooo`03
0000003oool0oooo04X0oooo00<000000?ooo`3oool0aP3oool00`000000oooo0?ooo`1O0?ooo`00
>03oool2000004/0oooo00<000000?ooo`3oool0aP3oool00`000000oooo0?ooo`1O0?ooo`00>03o
ool00`000000oooo0?ooo`190?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3oool0GP3o
ool003P0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`380?ooo`030000003oool0
oooo05h0oooo000h0?ooo`030000003oool0oooo04P0oooo00<000000?ooo`3oool0bP3oool00`00
0000oooo0?ooo`1M0?ooo`00>03oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo0<X0
oooo00<000000?ooo`3oool0G@3oool003P0oooo00<000000?ooo`3oool0A`3oool00`000000oooo
0?ooo`3<0?ooo`030000003oool0oooo05`0oooo000h0?ooo`030000003oool0oooo04H0oooo00<0
00000?ooo`3oool0cP3oool00`000000oooo0?ooo`1K0?ooo`00>03oool00`000000oooo0?ooo`16
0?ooo`030000003oool0oooo0<h0oooo00<000000?ooo`3oool0F`3oool003P0oooo00<000000?oo
o`3oool0A@3oool00`000000oooo0?ooo`3@0?ooo`030000003oool0oooo05X0oooo000h0?ooo`03
0000003oool0oooo04D0oooo00<000000?ooo`3oool0d03oool00`000000oooo0?ooo`1J0?ooo`00
>03oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo0=80oooo00<000000?ooo`3oool0
F@3oool003P0oooo0P0000150?ooo`030000003oool0oooo0=80oooo00<000000?ooo`3oool0F@3o
ool003P0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`3D0?ooo`030000003oool0
oooo05P0oooo000h0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0e03oool00`00
0000oooo0?ooo`1H0?ooo`00>03oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo0=H0
oooo00<000000?ooo`3oool0E`3oool003P0oooo00<000000?ooo`3oool0@P3oool00`000000oooo
0?ooo`3G0?ooo`030000003oool0oooo05H0oooo000h0?ooo`030000003oool0oooo0440oooo00<0
00000?ooo`3oool0f03oool00`000000oooo0?ooo`1F0?ooo`00>03oool00`000000oooo0?ooo`11
0?ooo`030000003oool0oooo0=T0oooo00<000000?ooo`3oool0E@3oool003P0oooo00<000000?oo
o`3oool0@03oool00`000000oooo0?ooo`3J0?ooo`030000003oool0oooo05D0oooo000h0?ooo`03
0000003oool0oooo03l0oooo00<000000?ooo`3oool0g03oool00`000000oooo0?ooo`1D0?ooo`00
>03oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo0=`0oooo00<000000?ooo`3oool0
E03oool003P0oooo0P00000o0?ooo`030000003oool0oooo0=h0oooo00<000000?ooo`3oool0D`3o
ool003P0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`3N0?ooo`030000003oool0
oooo05<0oooo000h0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool0h03oool00`00
0000oooo0?ooo`1B0?ooo`00>03oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo0>40
oooo00<000000?ooo`3oool0D@3oool003P0oooo00<000000?ooo`3oool0?03oool00`000000oooo
0?ooo`3R0?ooo`030000003oool0oooo0540oooo000h0?ooo`030000003oool0oooo03`0oooo00<0
00000?ooo`3oool0h`3oool00`000000oooo0?ooo`1@0?ooo`00>03oool00`000000oooo0?ooo`0k
0?ooo`030000003oool0oooo0>@0oooo00<000000?ooo`3oool0D03oool001`0oooo00D0omnL04P0
00000000001d0;ooo`030?ooo`060?noM0000000M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo
0?ooo`080?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0i@3oool00`000000oooo
0?ooo`1?0?ooo`006`3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@0
000007Boo`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`080?ooo`030000003oool0
oooo03X0oooo00<000000?ooo`3oool0iP3oool00`000000oooo0?ooo`1?0?ooo`006`3oool00`2o
M00004RL0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO09a8M02oool0203oool00`00
0000oooo0?ooo`0j0?ooo`030000003oool0oooo0>L0oooo00<000000?ooo`3oool0CP3oool001X0
oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0;0?ooo`030;md001d_ol0oooo00L0
oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`3X0?ooo`030000003oool0oooo04h0
oooo000J0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3ooml0W4P0
04RLg`080?ooo`<00000>@3oool00`000000oooo0?ooo`3Y0?ooo`030000003oool0oooo04d0oooo
000J0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3OW4P0000007Bo
o`080?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool0j`3oool00`000000oooo0?oo
o`1<0?ooo`006P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO
0?ooo`3ookl0M0000000000007@0_ooo00P0oooo00<000000?ooo`3oool0=`3oool00`000000oooo
0?ooo`3/0?ooo`030000003oool0oooo04`0oooo000J0?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol01P3oool01P3OW4P00000000000000000001809cOo`T0oooo00<000000?ooo`3o
ool0=`3oool00`000000oooo0?ooo`3]0?ooo`030000003oool0oooo04/0oooo000J0?ooo`030?oo
_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool0103ookl0M00000000018W=l;0?ooo`03
0000003oool0oooo03H0oooo00<000000?ooo`3oool0kP3oool00`000000oooo0?ooo`1;0?ooo`00
6`3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0_g@007Boo`3oool0
2`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo0?00oooo00<000000?ooo`3oool0
BP3oool001/0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00H0omnL04P00000
000000000000B02Lgol80?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0l@3oool0
0`000000oooo0?ooo`190?ooo`00703oool01@3ookl0M0000000000007@0_ooo00X0oooo00D0_g@0
000000000000000004RLg`080?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0lP3o
ool00`000000oooo0?ooo`190?ooo`00>03oool00`000000oooo0?ooo`0c0?ooo`030000003oool0
oooo0?@0oooo00<000000?ooo`3oool0B03oool003P0oooo00<000000?ooo`3oool0<`3oool00`00
0000oooo0?ooo`3d0?ooo`030000003oool0oooo04P0oooo000h0?ooo`030000003oool0oooo0380
oooo00<000000?ooo`3oool0mP3oool00`000000oooo0?ooo`170?ooo`00>03oool200000380oooo
00<000000?ooo`3oool0n03oool00`000000oooo0?ooo`160?ooo`00>03oool00`000000oooo0?oo
o`0a0?ooo`030000003oool0oooo0?P0oooo00<000000?ooo`3oool0AP3oool003P0oooo00<00000
0?ooo`3oool0<03oool00`000000oooo0?ooo`3j0?ooo`030000003oool0oooo04D0oooo000h0?oo
o`030000003oool0oooo02l0oooo00<000000?ooo`3oool0o03oool00`000000oooo0?ooo`140?oo
o`00>03oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo0?`0oooo00<000000?ooo`3o
ool0A03oool003P0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`3n0?ooo`030000
003oool0oooo04<0oooo000h0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0o`3o
ool00`000000oooo0?ooo`130?ooo`00>03oool00`000000oooo0?ooo`0]0?ooo`030000003oool0
oooo0?l0oooo0@3oool00`000000oooo0?ooo`120?ooo`00>03oool00`000000oooo0?ooo`0/0?oo
o`030000003oool0oooo0?l0oooo0`3oool00`000000oooo0?ooo`110?ooo`00>03oool00`000000
oooo0?ooo`0[0?ooo`030000003oool0oooo0?l0oooo103oool00`000000oooo0?ooo`110?ooo`00
>03oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo0?l0oooo1@3oool00`000000oooo
0?ooo`100?ooo`00>03oool2000002/0oooo00<000000?ooo`3oool0o`3oool60?ooo`030000003o
ool0oooo0400oooo000h0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0o`3oool8
0?ooo`030000003oool0oooo03l0oooo000h0?ooo`030000003oool0oooo02P0oooo00<000000?oo
o`3oool0o`3oool:0?ooo`030000003oool0oooo03h0oooo000h0?ooo`030000003oool0oooo02L0
oooo00<000000?ooo`3oool0o`3oool<0?ooo`030000003oool0oooo03d0oooo000h0?ooo`030000
003oool0oooo02L0oooo00<000000?ooo`3oool0o`3oool=0?ooo`030000003oool0oooo03`0oooo
000h0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0o`3oool?0?ooo`030000003o
ool0oooo03/0oooo000h0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0o`3oool@
0?ooo`030000003oool0oooo03/0oooo000h0?ooo`030000003oool0oooo02@0oooo00<000000?oo
o`3oool0o`3ooolB0?ooo`030000003oool0oooo03X0oooo000h0?ooo`030000003oool0oooo02<0
oooo00<000000?ooo`3oool0o`3ooolD0?ooo`030000003oool0oooo03T0oooo000h0?ooo`030000
003oool0oooo0280oooo00<000000?ooo`3oool0o`3ooolF0?ooo`030000003oool0oooo03P0oooo
000h0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0o`3ooolG0?ooo`030000003o
ool0oooo03L0oooo000h0?ooo`8000008P3oool00`000000oooo0?ooo`3o0?oooaT0oooo00<00000
0?ooo`3oool0=P3oool003P0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`3o0?oo
oa/0oooo00<000000?ooo`3oool0=@3oool003P0oooo00<000000?ooo`3oool07`3oool00`000000
oooo0?ooo`3o0?oooad0oooo00<000000?ooo`3oool0=03oool003P0oooo00<000000?ooo`3oool0
7P3oool00`000000oooo0?ooo`3o0?oooal0oooo00<000000?ooo`3oool0<`3oool003P0oooo00<0
00000?ooo`3oool07@3oool00`000000oooo0?ooo`3o0?ooob00oooo00<000000?ooo`3oool0<`3o
ool003P0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`3o0?ooob40oooo00<00000
0?ooo`3oool0<P3oool003P0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`3o0?oo
ob<0oooo00<000000?ooo`3oool0<@3oool003P0oooo00<000000?ooo`3oool06`3oool00`000000
oooo0?ooo`3o0?ooobD0oooo00<000000?ooo`3oool0<03oool003P0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`3o0?ooobL0oooo00<000000?ooo`3oool0;`3oool003P0oooo00<0
00000?ooo`3oool06@3oool00`000000oooo0?ooo`3o0?ooobT0oooo00<000000?ooo`3oool0;P3o
ool003P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`3o0?ooob/0oooo00<00000
0?ooo`3oool0;@3oool003P0oooo0P00000G0?ooo`800000o`3oool_0?ooo`800000;@3oool003P0
oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`3o0?oooc40oooo00<000000?ooo`3o
ool0:P3oool003P0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`3o0?oooc<0oooo
00<000000?ooo`3oool0:@3oool003P0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?oo
o`3o0?ooocD0oooo00<000000?ooo`3oool0:03oool003P0oooo00<000000?ooo`3oool04P3oool0
0`000000oooo0?ooo`3o0?ooocL0oooo00<000000?ooo`3oool09`3oool003P0oooo00<000000?oo
o`3oool0403oool200000?l0oooo>`3oool2000002L0oooo000Y0?ooo`060;md0000000000000000
0000B9`0gooo2@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0?l0oooo?@3oool0
0`000000oooo0?ooo`0T0?ooo`00:P3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo
00d0oooo0P00003o0?oood40oooo0P00000T0?ooo`00:P3oool00`3ogi`0B00007Boo`0;0?ooo`03
0000003oool0oooo00X0oooo0`00003o0?ooodD0oooo0`00000Q0?ooo`00:P3oool00`3ogi`0B000
07Boo`0;0?ooo`030000003oool0oooo00L0oooo0`00003o0?oood/0oooo0`00000N0?ooo`00:P3o
ool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0o`3o
oom?0?ooo`800000703oool002X0oooo00<0omnL04P0001d_ol02`3oool900000?l0ooooE03oool8
000001@0oooo000Z0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0o`3ooomf0?oo
o`00:P3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo0?l0ooooMP3oool002X0oooo
00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`3o0?ooogH0oooo000Z0?ooo`030?oOW018
0000M;oo00/0oooo00<000000?ooo`3oool0o`3ooomf0?ooo`00:03oool01@3ooml0W4P000000000
0000M;oo00/0oooo00<000000?ooo`3oool0o`3ooomf0?ooo`00:P3oool00`3o_g@0000007Boo`0;
0?ooo`030000003oool0oooo0?l0ooooMP3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.0259845, -0.0154541, \
0.00366114, 0.00366114}, {{10.625, 420.313}, {259.5, 6.3125}} -> {-0.995213, \
-1.21151, 0.0265357, 0.0136668}}]
}, Closed]],

Cell["A 1-dimensional list is displayed as a row.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[GraphicsArray[{plot[x], plot[v], plot[a]}]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .19314 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.31746 0.00459847 0.31746 [
[ 0 0 0 0 ]
[ 1 .19314 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .19314 L
0 .19314 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.00459847 0.321429 0.188537 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.151576 0.309016 0.294303 [
[.17539 .29652 -5 -15.6875 ]
[.17539 .29652 5 0 ]
[.32696 .29652 -5 -15.6875 ]
[.32696 .29652 5 0 ]
[.47854 .29652 -5 -15.6875 ]
[.47854 .29652 5 0 ]
[.63011 .29652 -5 -15.6875 ]
[.63011 .29652 5 0 ]
[.78169 .29652 -5 -15.6875 ]
[.78169 .29652 5 0 ]
[.93327 .29652 -5 -15.6875 ]
[.93327 .29652 5 0 ]
[.01131 .01471 -18.3125 -7.84375 ]
[.01131 .01471 0 7.84375 ]
[.01131 .16186 -27.3125 -7.84375 ]
[.01131 .16186 0 7.84375 ]
[.01131 .45617 -19 -7.84375 ]
[.01131 .45617 0 7.84375 ]
[.01131 .60332 -10 -7.84375 ]
[.01131 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.17539 .30902 m
.17539 .31527 L
s
gsave
.17539 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.32696 .30902 m
.32696 .31527 L
s
gsave
.32696 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.47854 .30902 m
.47854 .31527 L
s
gsave
.47854 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.63011 .30902 m
.63011 .31527 L
s
gsave
.63011 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.78169 .30902 m
.78169 .31527 L
s
gsave
.78169 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.93327 .30902 m
.93327 .31527 L
s
gsave
.93327 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.125 Mabswid
.05412 .30902 m
.05412 .31277 L
s
.08444 .30902 m
.08444 .31277 L
s
.11476 .30902 m
.11476 .31277 L
s
.14507 .30902 m
.14507 .31277 L
s
.2057 .30902 m
.2057 .31277 L
s
.23602 .30902 m
.23602 .31277 L
s
.26633 .30902 m
.26633 .31277 L
s
.29665 .30902 m
.29665 .31277 L
s
.35728 .30902 m
.35728 .31277 L
s
.38759 .30902 m
.38759 .31277 L
s
.41791 .30902 m
.41791 .31277 L
s
.44822 .30902 m
.44822 .31277 L
s
.50885 .30902 m
.50885 .31277 L
s
.53917 .30902 m
.53917 .31277 L
s
.56948 .30902 m
.56948 .31277 L
s
.5998 .30902 m
.5998 .31277 L
s
.66043 .30902 m
.66043 .31277 L
s
.69074 .30902 m
.69074 .31277 L
s
.72106 .30902 m
.72106 .31277 L
s
.75138 .30902 m
.75138 .31277 L
s
.81201 .30902 m
.81201 .31277 L
s
.84232 .30902 m
.84232 .31277 L
s
.87264 .30902 m
.87264 .31277 L
s
.90295 .30902 m
.90295 .31277 L
s
.96358 .30902 m
.96358 .31277 L
s
.9939 .30902 m
.9939 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.02381 .01471 m
.03006 .01471 L
s
gsave
.01131 .01471 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .16186 m
.03006 .16186 L
s
gsave
.01131 .16186 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .45617 m
.03006 .45617 L
s
gsave
.01131 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .60332 m
.03006 .60332 L
s
gsave
.01131 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04414 m
.02756 .04414 L
s
.02381 .07357 m
.02756 .07357 L
s
.02381 .103 m
.02756 .103 L
s
.02381 .13243 m
.02756 .13243 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .25016 m
.02756 .25016 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .36788 m
.02756 .36788 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .4856 m
.02756 .4856 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .60332 m
.02499 .60331 L
.02605 .60329 L
.02729 .60324 L
.02846 .60318 L
.03053 .60303 L
.03279 .6028 L
.03527 .60248 L
.0379 .60205 L
.04262 .60106 L
.04749 .59974 L
.05205 .59822 L
.06244 .59381 L
.07305 .58793 L
.08274 .58135 L
.10458 .56251 L
.14429 .51515 L
.18248 .45627 L
.22313 .38349 L
.26226 .30833 L
.30384 .22863 L
.34391 .15746 L
.38246 .09886 L
.402 .07411 L
.42346 .05145 L
.44388 .03465 L
.45301 .02874 L
.46295 .02346 L
.46838 .0211 L
.4734 .01923 L
.4781 .01778 L
.4833 .0165 L
.48553 .01605 L
.48794 .01564 L
.49002 .01535 L
.4923 .01509 L
.49473 .01489 L
.49605 .01481 L
.49728 .01476 L
.49844 .01473 L
.49949 .01472 L
.50071 .01472 L
.50186 .01474 L
.50301 .01477 L
.50425 .01483 L
.50542 .0149 L
.50648 .01498 L
.5091 .01524 L
.51158 .01557 L
.5165 .01646 L
Mistroke
.52187 .01777 L
.53144 .02102 L
.54032 .02507 L
.56088 .03814 L
.5795 .05428 L
.61962 .1017 L
.65822 .16102 L
.69928 .23447 L
.73882 .31042 L
.77684 .38343 L
.81732 .45593 L
.85628 .51593 L
.87613 .54149 L
.89769 .56473 L
.91843 .58221 L
.92832 .58876 L
.93759 .59383 L
.94643 .59767 L
.95595 .6007 L
.9611 .60186 L
.96369 .60232 L
.96652 .60272 L
.96894 .60298 L
.9703 .6031 L
.97153 .60318 L
.97268 .60324 L
.97393 .60329 L
.97511 .60331 L
.97619 .60332 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.35119 0.00459847 0.64881 0.188537 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.151576 0.309018 0.294303 [
[.17539 .29652 -5 -15.6875 ]
[.17539 .29652 5 0 ]
[.32696 .29652 -5 -15.6875 ]
[.32696 .29652 5 0 ]
[.47854 .29652 -5 -15.6875 ]
[.47854 .29652 5 0 ]
[.63011 .29652 -5 -15.6875 ]
[.63011 .29652 5 0 ]
[.78169 .29652 -5 -15.6875 ]
[.78169 .29652 5 0 ]
[.93327 .29652 -5 -15.6875 ]
[.93327 .29652 5 0 ]
[.01131 .01471 -18.3125 -7.84375 ]
[.01131 .01471 0 7.84375 ]
[.01131 .16187 -27.3125 -7.84375 ]
[.01131 .16187 0 7.84375 ]
[.01131 .45617 -19 -7.84375 ]
[.01131 .45617 0 7.84375 ]
[.01131 .60332 -10 -7.84375 ]
[.01131 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.17539 .30902 m
.17539 .31527 L
s
gsave
.17539 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.32696 .30902 m
.32696 .31527 L
s
gsave
.32696 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.47854 .30902 m
.47854 .31527 L
s
gsave
.47854 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.63011 .30902 m
.63011 .31527 L
s
gsave
.63011 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.78169 .30902 m
.78169 .31527 L
s
gsave
.78169 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.93327 .30902 m
.93327 .31527 L
s
gsave
.93327 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.125 Mabswid
.05412 .30902 m
.05412 .31277 L
s
.08444 .30902 m
.08444 .31277 L
s
.11476 .30902 m
.11476 .31277 L
s
.14507 .30902 m
.14507 .31277 L
s
.2057 .30902 m
.2057 .31277 L
s
.23602 .30902 m
.23602 .31277 L
s
.26633 .30902 m
.26633 .31277 L
s
.29665 .30902 m
.29665 .31277 L
s
.35728 .30902 m
.35728 .31277 L
s
.38759 .30902 m
.38759 .31277 L
s
.41791 .30902 m
.41791 .31277 L
s
.44822 .30902 m
.44822 .31277 L
s
.50885 .30902 m
.50885 .31277 L
s
.53917 .30902 m
.53917 .31277 L
s
.56948 .30902 m
.56948 .31277 L
s
.5998 .30902 m
.5998 .31277 L
s
.66043 .30902 m
.66043 .31277 L
s
.69074 .30902 m
.69074 .31277 L
s
.72106 .30902 m
.72106 .31277 L
s
.75138 .30902 m
.75138 .31277 L
s
.81201 .30902 m
.81201 .31277 L
s
.84232 .30902 m
.84232 .31277 L
s
.87264 .30902 m
.87264 .31277 L
s
.90295 .30902 m
.90295 .31277 L
s
.96358 .30902 m
.96358 .31277 L
s
.9939 .30902 m
.9939 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.02381 .01471 m
.03006 .01471 L
s
gsave
.01131 .01471 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .16187 m
.03006 .16187 L
s
gsave
.01131 .16187 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .45617 m
.03006 .45617 L
s
gsave
.01131 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .60332 m
.03006 .60332 L
s
gsave
.01131 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04415 m
.02756 .04415 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .10301 m
.02756 .10301 L
s
.02381 .13244 m
.02756 .13244 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .25016 m
.02756 .25016 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .36788 m
.02756 .36788 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .4856 m
.02756 .4856 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .30902 m
.06244 .23481 L
.10458 .15951 L
.14415 .09914 L
.18221 .05447 L
.20178 .03757 L
.21267 .03011 L
.22272 .0245 L
.23293 .02007 L
.23832 .01827 L
.24402 .01676 L
.24641 .01625 L
.24897 .01579 L
.25119 .01545 L
.25361 .01516 L
.25507 .01501 L
.25643 .01491 L
.25773 .01483 L
.2584 .01479 L
.25912 .01476 L
.25983 .01474 L
.26048 .01473 L
.26176 .01472 L
.26293 .01472 L
.26416 .01475 L
.26538 .01479 L
.26607 .01483 L
.26671 .01486 L
.26912 .01505 L
.27131 .01528 L
.27364 .0156 L
.27847 .01647 L
.2838 .01778 L
.29326 .02099 L
.30213 .02502 L
.32232 .03778 L
.34082 .05371 L
.38045 .10022 L
.41855 .15838 L
.45912 .23059 L
.49816 .30545 L
.53966 .38514 L
.57964 .45663 L
.61811 .51582 L
.63759 .54096 L
.65902 .56418 L
.67938 .58152 L
.69843 .5933 L
.70884 .59785 L
.71396 .5996 L
Mistroke
.71872 .60092 L
.72301 .60186 L
.72758 .60261 L
.7302 .60292 L
.73148 .60304 L
.73266 .60313 L
.73371 .6032 L
.73487 .60325 L
.73603 .60329 L
.73725 .60332 L
.73856 .60332 L
.73977 .6033 L
.7404 .60329 L
.7411 .60326 L
.74251 .6032 L
.74375 .60312 L
.74508 .60301 L
.74748 .60276 L
.74985 .60244 L
.75235 .60202 L
.7568 .60108 L
.76198 .59967 L
.7668 .59806 L
.77773 .59332 L
.78828 .58734 L
.79823 .58046 L
.81683 .5645 L
.85534 .51958 L
.89632 .45703 L
.93577 .38657 L
.97371 .31383 L
.97619 .30902 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.678571 0.00459847 0.97619 0.188537 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.151576 0.309018 0.294303 [
[.17539 .29652 -5 -15.6875 ]
[.17539 .29652 5 0 ]
[.32696 .29652 -5 -15.6875 ]
[.32696 .29652 5 0 ]
[.47854 .29652 -5 -15.6875 ]
[.47854 .29652 5 0 ]
[.63011 .29652 -5 -15.6875 ]
[.63011 .29652 5 0 ]
[.78169 .29652 -5 -15.6875 ]
[.78169 .29652 5 0 ]
[.93327 .29652 -5 -15.6875 ]
[.93327 .29652 5 0 ]
[.01131 .01472 -18.3125 -7.84375 ]
[.01131 .01472 0 7.84375 ]
[.01131 .16187 -27.3125 -7.84375 ]
[.01131 .16187 0 7.84375 ]
[.01131 .45617 -19 -7.84375 ]
[.01131 .45617 0 7.84375 ]
[.01131 .60332 -10 -7.84375 ]
[.01131 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.17539 .30902 m
.17539 .31527 L
s
gsave
.17539 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.32696 .30902 m
.32696 .31527 L
s
gsave
.32696 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.47854 .30902 m
.47854 .31527 L
s
gsave
.47854 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.63011 .30902 m
.63011 .31527 L
s
gsave
.63011 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.78169 .30902 m
.78169 .31527 L
s
gsave
.78169 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.93327 .30902 m
.93327 .31527 L
s
gsave
.93327 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.125 Mabswid
.05412 .30902 m
.05412 .31277 L
s
.08444 .30902 m
.08444 .31277 L
s
.11476 .30902 m
.11476 .31277 L
s
.14507 .30902 m
.14507 .31277 L
s
.2057 .30902 m
.2057 .31277 L
s
.23602 .30902 m
.23602 .31277 L
s
.26633 .30902 m
.26633 .31277 L
s
.29665 .30902 m
.29665 .31277 L
s
.35728 .30902 m
.35728 .31277 L
s
.38759 .30902 m
.38759 .31277 L
s
.41791 .30902 m
.41791 .31277 L
s
.44822 .30902 m
.44822 .31277 L
s
.50885 .30902 m
.50885 .31277 L
s
.53917 .30902 m
.53917 .31277 L
s
.56948 .30902 m
.56948 .31277 L
s
.5998 .30902 m
.5998 .31277 L
s
.66043 .30902 m
.66043 .31277 L
s
.69074 .30902 m
.69074 .31277 L
s
.72106 .30902 m
.72106 .31277 L
s
.75138 .30902 m
.75138 .31277 L
s
.81201 .30902 m
.81201 .31277 L
s
.84232 .30902 m
.84232 .31277 L
s
.87264 .30902 m
.87264 .31277 L
s
.90295 .30902 m
.90295 .31277 L
s
.96358 .30902 m
.96358 .31277 L
s
.9939 .30902 m
.9939 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.02381 .01472 m
.03006 .01472 L
s
gsave
.01131 .01472 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .16187 m
.03006 .16187 L
s
gsave
.01131 .16187 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .45617 m
.03006 .45617 L
s
gsave
.01131 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .60332 m
.03006 .60332 L
s
gsave
.01131 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04415 m
.02756 .04415 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .10301 m
.02756 .10301 L
s
.02381 .13244 m
.02756 .13244 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .25016 m
.02756 .25016 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .36788 m
.02756 .36788 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .4856 m
.02756 .4856 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01472 m
.02499 .01472 L
.02605 .01475 L
.02729 .01479 L
.02846 .01485 L
.03053 .015 L
.03279 .01523 L
.03527 .01556 L
.0379 .01599 L
.04262 .01698 L
.04749 .0183 L
.05205 .01981 L
.06244 .02422 L
.07305 .03011 L
.08274 .03668 L
.10458 .05552 L
.14429 .10289 L
.18248 .16177 L
.22313 .23454 L
.26226 .3097 L
.30384 .3894 L
.34391 .46058 L
.38246 .51918 L
.402 .54393 L
.42346 .56659 L
.44388 .58338 L
.45301 .58929 L
.46295 .59457 L
.46838 .59694 L
.4734 .5988 L
.4781 .60026 L
.4833 .60154 L
.48553 .60198 L
.48794 .60239 L
.49002 .60268 L
.4923 .60294 L
.49473 .60314 L
.49605 .60322 L
.49728 .60327 L
.49844 .6033 L
.49949 .60332 L
.50071 .60332 L
.50186 .6033 L
.50301 .60326 L
.50425 .60321 L
.50542 .60313 L
.50648 .60305 L
.5091 .60279 L
.51158 .60246 L
.5165 .60158 L
Mistroke
.52187 .60026 L
.53144 .59701 L
.54032 .59297 L
.56088 .5799 L
.5795 .56376 L
.61962 .51633 L
.65822 .45702 L
.69928 .38356 L
.73882 .30761 L
.77684 .2346 L
.81732 .1621 L
.85628 .1021 L
.87613 .07654 L
.89769 .0533 L
.91843 .03582 L
.92832 .02927 L
.93759 .02421 L
.94643 .02037 L
.95595 .01733 L
.9611 .01617 L
.96369 .01572 L
.96652 .01531 L
.96894 .01505 L
.9703 .01494 L
.97153 .01485 L
.97268 .01479 L
.97393 .01475 L
.97511 .01472 L
.97619 .01472 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{460.438, 88.9375},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000:b0000QB000`400?l00000o`00003oo`3ooooo
0?oook@0oooo003o0?ooool0oooo]03oool00?l0ooooo`3ooond0?ooo`00o`3ooooo0?oook@0oooo
003o0?ooool0oooo]03oool00?l0ooooo`3ooond0?ooo`00o`3ooooo0?oook@0oooo003o0?ooool0
oooo]03oool00?l0ooooo`3ooond0?ooo`00o`3ooooo0?oook@0oooo003o0?ooool0oooo]03oool0
0?l0ooooo`3ooond0?ooo`00o`3ooooo0?oook@0oooo000a0?ooo`060;md00000000000000000000
B9`0gooofP3oool01P2oM000000000000000000004RL0=ooom/0oooo00H0_g@00000000000000000
0018W03Ooonj0?ooo`00<P3oool00`3ogi`0B00007Boo`3M0?ooo`030?oOW0180000M;oo0=h0oooo
00<0omnL04P0001d_ol0_03oool00380oooo00<0omnL04P0001d_ol0203oool00`000000oooo0?oo
o`3B0?ooo`030?oOW0180000M;oo00P0oooo00<000000?ooo`3oool0d`3oool00`3ogi`0B00007Bo
o`080?ooo`030000003oool0oooo0;40oooo000b0?ooo`030?oOW0180000M;oo00P0oooo00<00000
0?ooo`3oool0dP3oool00`3ogi`0B00007Boo`080?ooo`030000003oool0oooo0=<0oooo00<0omnL
04P0001d_ol0203oool00`000000oooo0?ooo`2a0?ooo`008P3oool00`3ooml0W4P0000000070000
000304RLg`3oool0oooo00<0oooo00<0omnL04P0001d_ol0203oool00`000000oooo0?ooo`330?oo
o`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool00P3oool00`3ogi`0B00007Boo`080?oo
o`030000003oool0oooo0<<0oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`030?oo
o`030?oOW0180000M;oo00P0oooo00<000000?ooo`3oool0/@3oool00380oooo00<0omnL04P0001d
_ol0203oool2000004T0oooo200000220?ooo`030?oOW0180000M;oo00P0oooo0P00000R0?ooo`P0
0000ZP3oool00`3ogi`0B00007Boo`080?ooo`D00000U03oool5000001H0oooo000b0?ooo`030?oO
W0180000M;oo00P0oooo00<000000?ooo`3oool0A@3oool3000000P0oooo0P0000200?ooo`030?oO
W0180000M;oo00P0oooo00<000000?ooo`3oool07P3oool3000000P0oooo0P00002X0?ooo`030?oO
W0180000M;oo00P0oooo00@000000?ooo`3oool0oooo0`00002?0?ooo`<000006`3oool00380oooo
00<0omnL04P0001d_ol0203oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool0O@3oool00`3ogi`0B00007Boo`080?ooo`030000003oool0oooo01`0oooo
0P00000=0?ooo`030000003oool0oooo0:D0oooo00<0omnL04P0001d_ol0203oool00`000000oooo
0?ooo`040?ooo`800000S03oool00`000000oooo0?ooo`0L0?ooo`00<P3oool00`3ogi`0B00007Bo
o`080?ooo`030000003oool0oooo0480oooo0P00000?0?ooo`800000O@3oool00`3ogi`0B00007Bo
o`080?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool03P3oool200000:D0oooo00<0
omnL04P0001d_ol0203oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo08L0oooo0P00
000O0?ooo`00<P3oool00`3ogi`0B00007Boo`080?ooo`030000003oool0oooo0440oooo00<00000
0?ooo`3oool04@3oool00`000000oooo0?ooo`1j0?ooo`030?oOW0180000M;oo00P0oooo00<00000
0?ooo`3oool06`3oool00`000000oooo0?ooo`0@0?ooo`800000X`3oool00`3ogi`0B00007Boo`08
0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0Q@3oool00`000000oooo0?ooo`0O
0?ooo`00<03oool01@3ooml0W4P0000000000000M;oo00P0oooo0P0000110?ooo`030000003oool0
oooo01<0oooo0P00001h0?ooo`050?oog`2LB00000000000001d_ol0203oool2000001X0oooo0P00
000E0?ooo`800000W`3oool01@3ooml0W4P0000000000000M;oo00P0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`230?ooo`030000003oool0oooo0200oooo000b0?ooo`030?noM000
0000M;oo00P0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`0F0?ooo`800000N03o
ool00`3o_g@0000007Boo`080?ooo`030000003oool0oooo01L0oooo0P00000I0?ooo`030000003o
ool0oooo09h0oooo00<0okmd0000001d_ol0203oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo0840oooo00<000000?ooo`3oool08@3oool003d0oooo00<000000?ooo`3oool0?P3oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0800oooo00<000000?ooo`3oool05P3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02P3oool0
0`000000oooo0?ooo`1o0?ooo`030000003oool0oooo0280oooo000m0?ooo`030000003oool0oooo
03d0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`200?ooo`030000003oool0oooo
01D0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool0O@3oool00`000000oooo0?ooo`0S0?ooo`00?@3oool00`000000
oooo0?ooo`0l0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0O`3oool00`000000
oooo0?ooo`0E0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0Y`3oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool0903oool003d0oooo
0P00000l0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0OP3oool2000001D0oooo
00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`2V0?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`0U0?ooo`00?@3oool00`000000oooo0?oo
o`0j0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0O@3oool00`000000oooo0?oo
o`0C0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0Y@3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool09P3oool003d0oooo00<00000
0?ooo`3oool0>@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo07d0oooo00<00000
0?ooo`3oool04P3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0:D0oooo00<00000
0?ooo`3oool03`3oool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo02L0oooo000m0?oo
o`030000003oool0oooo03T0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`1l0?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`2T0?oo
o`030000003oool0oooo0100oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`0W0?oo
o`00?@3oool2000003T0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`1k0?ooo`80
00004P3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo0:<0oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo02P0oooo000m0?ooo`030000
003oool0oooo03L0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`1j0?ooo`030000
003oool0oooo0100oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`2R0?ooo`030000
003oool0oooo0180oooo00<000000?ooo`3oool0L03oool00`000000oooo0?ooo`0Y0?ooo`00?@3o
ool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0NP3o
ool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0XP3o
ool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0:@3o
ool003d0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0
oooo07T0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0
oooo0:40oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`1^0?ooo`030000003oool0
oooo02X0oooo000m0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0:P3oool00`00
0000oooo0?ooo`1h0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0:03oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0KP3oool00`00
0000oooo0?ooo`0Z0?ooo`008`3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00H0okmd0000
001d_ol0oooo0?oog`2LB0030000000307Boo`3oool0oooo00H0oooo0P00000f0?ooo`030000003o
ool0oooo02X0oooo00<000000?ooo`3oool0G`3oool01@3ogi`0B0000000000007@0_ooo00<0oooo
00H0okmd0000001d_ol0oooo0?oog`2LB0030000000307Boo`3oool0oooo00D0oooo0P00000?0?oo
o`030000003oool0oooo02X0oooo00<000000?ooo`3oool0QP3oool01@3ogi`0B0000000000007@0
_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oog`2LB0030000000307Boo`3oool0oooo00H0oooo
00<000000?ooo`3oool0503oool00`000000oooo0?ooo`1/0?ooo`030000003oool0oooo02/0oooo
000R0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo0?oo
o`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00H0oooo00<000000?ooo`3oool0=03oool0
0`000000oooo0?ooo`0/0?ooo`030000003oool0oooo05d0oooo00L0oono07@0001d_ol0oooo0?no
M00007@0_ooo0080oooo00L0okmd0000001d_ol0oooo0?oog`2LB7@0_ooo0080oooo00<0_g@007Bo
o`3oool01@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo02`0oooo00<000000?oo
o`3oool0Q03oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@0000007Bo
o`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`060?ooo`030000003oool0oooo01D0
oooo00<000000?ooo`3oool0JP3oool00`000000oooo0?ooo`0/0?ooo`008P3oool00`2oM00004RL
0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO09a8M02oool01P3oool00`000000oooo
0?ooo`0c0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0G03oool00`2oM00004RL
0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO09a8M02oool01@3oool00`000000oooo
0?ooo`0=0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0Q03oool00`2oM00004RL
0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO09a8M02oool01P3oool00`000000oooo
0?ooo`0E0?ooo`030000003oool0oooo06X0oooo00<000000?ooo`3oool0;03oool00240oooo00<0
oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0;0?ooo`030;md001d_ol0oooo00D0oooo00<0
00000?ooo`3oool0<`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo05X0oooo00<0
oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0;0?ooo`030;md001d_ol0oooo00@0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo0880oooo00<0
oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0;0?ooo`030;md001d_ol0oooo00D0oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`1X0?ooo`030000003oool0oooo02d0oooo000E
0?ooo`030?oog`2LB000000000L0000000D0B9cO0?ooo`3ogi`0B00007Boo`030?ooo`030?noM000
04P0W=oo00X0oooo00<0oooO09a80018W=l01P3oool00`000000oooo0?ooo`0b0?ooo`030000003o
ool0oooo0300oooo00<000000?ooo`3oool0C@3oool00`3ooml0W4P0000000070000000604RLg`3o
ool0oooo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0:0?ooo`030?oog`2LB000B9cO00D0
oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo07H0
oooo00<0oooO09a8000000001`0000001@18W=l0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol02P3oool00`3ooml0W4P004RLg`060?ooo`030000003oool0oooo01H0oooo00<000000?oo
o`3oool0J03oool00`000000oooo0?ooo`0]0?ooo`008@3oool00`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo00X0oooo00<0gia80000001d_ol01P3oool200000380oooo00<000000?ooo`3o
ool0<P3oool00`000000oooo0?ooo`1I0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol02P3oool00`3OW4P0000007Boo`050?ooo`800000303oool00`000000oooo0?ooo`0`0?ooo`03
0000003oool0oooo0840oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`03
0=nLB0000000M;oo00H0oooo0P00000H0?ooo`030000003oool0oooo06H0oooo00<000000?ooo`3o
ool0;P3oool00240oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`070=oO
g`3oool0oono07@000000000001d0;ooo`060?ooo`030000003oool0oooo0340oooo00<000000?oo
o`3oool0<P3oool00`000000oooo0?ooo`1I0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol01P3oool01`3Ogml0oooo0?oo_`1d000000000000M02oool01@3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0P03oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00H0
oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo02l0
oooo000Q0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3OW4P00000
000000000000001809cOo`L0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0d0?oo
o`030000003oool0oooo05P0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?oo
o`060=nLB000000000000000000004P0W=oo1P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo0380oooo00<000000?ooo`3oool0P03oool00`3ogi`0B00007Boo`030?ooo`030?noM000
04P0W=oo00H0oooo00H0gia800000000000000000000B02Lgol70?ooo`030000003oool0oooo01P0
oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`0_0?ooo`008@3oool00`3ookl0M000
04RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono07@000000000B9cO2@3oool00`000000
oooo0?ooo`0`0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0F03oool00`3ookl0
M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono07@000000000B9cO203oool00`00
0000oooo0?ooo`090?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0O`3oool00`3o
okl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono07@000000000B9cO2@3oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0680oooo00<000000?ooo`3oool0<03oool0
0280oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030;md001d_ol0oooo
00T0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo
05P0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030;md001d_ol0oooo
00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo
07l0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030;md001d_ol0oooo
00T0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo
0300oooo000R0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`060?oOW0180000
00000000000004P0W=oo1P3oool2000002l0oooo00<000000?ooo`3oool0=`3oool00`000000oooo
0?ooo`1H0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`060?oOW01800000000
0000000004P0W=oo1@3oool2000000T0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?oo
o`1o0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`060?oOW018000000000000
000004P0W=oo1P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0600oooo00<00000
0?ooo`3oool0<@3oool002<0oooo00D0oono07@000000000001d0;ooo`0:0?ooo`050;md00000000
000000000018W=l01P3oool00`000000oooo0?ooo`0D0?ooo`060;md00000000000000000000B9`0
gooo403oool00`2oM0000000000000040000000304RLg`3oool0oooo00l0oooo00<0oono07@00000
00000P0000000`1d_ol0oooo0?ooo`0E0?ooo`040?oog`2LB00004RL0=ooo`D0oooo00<000000?oo
o`3oool02@3oool00`3ooml0W4P0000000020000000307Boo`3oool0oooo01<0oooo00D0omnL04P0
00000000000007Boo`0_0?ooo`050?oo_`1d000000000000M02oool02P3oool01@2oM00000000000
00000000B9cO00D0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0:0?ooo`060;md
00000000000000000000B9`0gooo4@3oool00`2oM0000000000000040000000304RLg`3oool0oooo
00d0oooo00@000000?ooo`3ookl0M0000`0000000`1d_ol0oooo0?ooo`0E0?ooo`040?oog`2LB000
04RL0=oooa40oooo00<0oooO09a8000000000P0000000`1d_ol0oooo0?ooo`0C0?ooo`050?oOW018
000000000000001d_ol0;P3oool01@3ookl0M0000000000007@0_ooo00X0oooo00D0_g@000000000
0000000004RLg`060?ooo`030000003oool0oooo01<0oooo00P0_g@000000000000000000018W03O
ool0oooo000000l0oooo00<0_g@0000000000000100000000`18W=l0oooo0?ooo`0?0?ooo`030?oo
_`1d000000000080000000<0M;oo0?ooo`3oool05@3oool0103ooml0W4P00018W03OoolA0?ooo`03
0?oog`2LB00000000080000000<0M;oo0?ooo`0000004`3oool01@3ogi`0B000000000000000M;oo
01`0oooo000m0?ooo`030000003oool0oooo01D0oooo00<0omnL04P0001d_ol04`3oool00`3OW4P0
000000000003000000030018W03Oool0oooo00l0oooo00<0omnL04P00018W=l00P3oool00`2oM000
B9cO0?ooo`0E0?ooo`040?oog`2LB00004RL0=ooo`D0oooo00<000000?ooo`3oool02@3oool00`3o
oml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo0180oooo00L0omnL04P00018W=l0oooo0?oo_`1d
0000M;oo04L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0;0?ooo`030?oOW018
0000M;oo01@0oooo00<0gia80000000000000`0000000`00B9`0gooo0?ooo`0=0?ooo`050000003o
ool0omnL04P00018W=l00P3oool00`2oM000B9cO0?ooo`0E0?ooo`040?oog`2LB00004RL0=oooa40
oooo00<0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`0B0?ooo`070?oOW0180000B9cO0?oo
o`3ookl0M00007Boo`170?ooo`030000003oool0oooo01@0oooo00<0omnL04P0001d_ol0103oool0
0`000000oooo0?ooo`0=0?ooo`030=nLB0000000000000<0000000<004RL0=ooo`3oool03`3oool0
0`3ogi`0B00004RLg`020?ooo`030;md0018W=l0oooo01D0oooo00@0oooO09a80000B9`0gooo4@3o
ool00`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol000000180oooo00L0omnL04P00018W=l0oooo
0?oo_`1d0000M;oo01/0oooo000m0?ooo`030000003oool0oooo01D0oooo00<0omnL04P0001d_ol0
503oool00`3OW4P000000?ooo`020?ooo`030=nLM02Lgol0oooo01@0oooo00<0oono07@0B02Lgol0
5@3oool0103ooml0W4P00018W03Oool60?ooo`030000003oool0oooo00d0oooo00<0oooO09a8M02o
ool04@3oool00`3ooml0W4P0001d_`030?ooo`030?noM000B9`0gooo04H0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`0<0?ooo`030?oOW0180000M;oo01D0oooo00<0gia804RLg`3o
ool00P3oool00`3OW7@0W=oo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<0oono07@0B02L
gol05@3oool0103ooml0W4P00018W03OoolF0?ooo`030?oog`2LB7@0_ooo0140oooo00<0oooO09a8
0000M;l00`3oool00`3o_g@004RL0=ooo`160?ooo`030000003oool0oooo01@0oooo00<0omnL04P0
001d_ol01@3oool00`000000oooo0?ooo`0=0?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cO
o`3oool0503oool00`3ookl0M01809cOo`0E0?ooo`040?oog`2LB00004RL0=oooaH0oooo00<0oooO
0000002oool04@3oool00`3ooml0W4P0001d_`030?ooo`030?noM000B9`0gooo01X0oooo000m0?oo
o`030000003oool0oooo01D0oooo00<0omnL04P0001d_ol0503oool00`000000gia807Boo`0I0?oo
o`030?oo_`1d0000M;oo0140oooo00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`030?oo
o`030000003oool0oooo00h0oooo00<0_g@007Boo`3oool0403oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo04H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0<0?oo
o`030?oOW0180000M;oo01H0oooo00<0gia807Boo`3oool04`3oool00`000000oooo0?ooo`020?oo
o`030?oo_`1d0000M;oo0140oooo00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`0D0?oo
o`030;md001d_ol0oooo0100oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`160?oo
o`030000003oool0oooo01@0oooo00<0omnL04P0001d_ol01@3oool00`000000oooo0?ooo`0>0?oo
o`030=nLB01d_ol0oooo01P0oooo00<0oono07@0001d_ol04@3oool00`2oM0000000000000050000
000307Boo`3oool0oooo01@0oooo00<0000007Boo`3oool0403oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo01X0oooo000m0?ooo`8000005P3oool00`3ogi`0B00007Boo`0D0?ooo`04
0000003oool0_gA809cOoaP0oooo00<0okmd0000001d_ol04@3oool0203ogi`0M9cO0?ooo`3oool0
oooO09a80000B9`0gooo1`3oool00`000000oooo0?ooo`0<0?ooo`030?oog`2LB000B9cO0140oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`160?ooo`8000001P3oool00`000000oooo
0?ooo`0=0?ooo`030?oOW0180000M;oo01L0oooo00<0_gA809cOo`3oool04P3oool00`000000oooo
0?ooo`020?ooo`030?noM0000000M;oo0140oooo00P0omnL07BLg`3oool0oooo0?oog`2LB00004RL
0=oooaH0oooo00<0oooO09a80018W=l04@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo04H0oooo00<000000?ooo`3oool0503oool00`3ogi`0B00007Boo`060?ooo`030000003oool0
oooo00h0oooo00<0_gA809cOo`3oool05`3oool00`3o_g@0000007Boo`0A0?ooo`080?oOW01dW=l0
oooo0?ooo`3ooml0W4P00018W03OoolF0?ooo`030000002LB000B9cO0140oooo00<0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`0J0?ooo`00?@3oool00`000000oooo0?ooo`0E0?ooo`030?oO
W0180000M;oo01<0oooo00H000000?ooo`3oool0oooO09a8M02ooolF0?ooo`040?oOW01800000018
09cOoa80oooo00L0gia807Boo`3oool0oooO09a80000B9`0gooo00L0oooo00<000000?ooo`3oool0
303oool00`3OW4P0000007Boo`0A0?ooo`030?oOW0180000M;oo00<0oooo00<0_g@00000B02Lgol0
AP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00d0oooo00<0omnL04P0001d_ol0
5`3oool00`3ooml0W4Qd0;ooo`0C0?ooo`070000003oool0oooo0?oOW0180000001809cOo`0B0?oo
o`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`0F0?ooo`030=nLB0000000M;oo0140oooo00<0
omnL04P0001d_ol00`3oool00`2oM000001809cOo`160?ooo`030000003oool0oooo01@0oooo00<0
omnL04P0001d_ol01P3oool00`000000oooo0?ooo`0>0?ooo`030?oog`2LB7@0_ooo01H0oooo00@0
omnL04P0000004P0W=oo4P3oool01`3OW4P0M;oo0?ooo`3ooml0W4P00018W03Oool05P3oool20000
000307Boo`3oool0oooo00l0oooo00<0omnL04P0001d_ol00`3oool00`2oM000001809cOo`0J0?oo
o`00?@3oool00`000000oooo0?ooo`0E0?ooo`030?oOW0180000M;oo0180oooo00<000000?ooo`3o
ool00P3oool00`3o_g@007Bo0?ooo`0C0?ooo`050?noM000000000000000B02Lgol04`3oool01`3o
oml0W4Qd0;ooo`3ooml0W4P00018W03Oool0203oool00`000000oooo0?ooo`070?ooo`070=oOg`3o
ool0oono07@000000000001d0;ooo`0A0?ooo`080?oo_`1d0000000004RLg`3oool0okmd00000000
M;m70?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool03P3oool00`3ogi`0B00007Bo
o`0H0?ooo`030?noM000M;l0oooo01<0oooo0`0000000`0004P0W=oo0?ooo`0B0?ooo`070?oog`2L
B7@0_ooo0?oog`2LB00004RL0=ooo`0B0?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`0A
0?ooo`080?oo_`1d0000000004RLg`3oool0okmd00000000M;m70?ooo`030000003oool0oooo01@0
oooo00<0omnL04P0001d_ol01`3oool00`000000oooo0?ooo`0>0?ooo`030?noM000M;l0oooo01<0
oooo00D0okmd000000000000001809cOo`0C0?ooo`070?oog`2LB7@0_ooo0?oog`2LB00004RL0=oo
o`0B0?ooo`030=oOg`3oool0oono0080000000<0001d0;ooo`3oool0403oool0203ookl0M0000000
0018W=l0oooo0?noM000000007Bo6`3oool003d0oooo00<000000?ooo`3oool05@3oool00`3ogi`0
B00007Boo`0B0?ooo`030000003oool0oooo0080oooo00<0oono07@0B02Lgol05@3oool00`2oM000
07Bo0?ooo`0D0?ooo`060?oOW018M;l0oooO09a80000B9`0gooo2@3oool00`000000oooo0?ooo`06
0?ooo`060=nLB000000000000000000004P0W=oo4`3oool00`2oM00004RL0;nLB002000000030018
W03Oool0oooo04H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0?0?ooo`030?oO
W0180000M;oo01P0oooo00<0oono07@0B02Lgol04`3oool010000000oooo0;md0000M;lE0?ooo`06
0?oOW018M;l0oooO09a80000B9`0gooo4P3oool01P3OW4P00000000000000000001809cOoa<0oooo
00<0_g@00018W02oW4P00P0000000`00B9`0gooo0?ooo`160?ooo`030000003oool0oooo01@0oooo
00<0omnL04P0001d_ol01`3oool00`000000oooo0?ooo`0>0?ooo`030?oo_`1d04P0W=oo01D0oooo
00<0_g@0001d_`3oool0503oool01P3ogi`0B7Bo0?oog`2LB00004RL0=oooa80oooo00H0gia80000
0000000000000000B02LgolC0?ooo`030;md0000B9`0_ia80080000000<004RL0=ooo`3oool06P3o
ool003d0oooo00<000000?ooo`3oool05@3oool00`3ogi`0B00007Boo`0A0?ooo`030000003oool0
oooo00<0oooo00<0oono07@00018W=l05P3oool00`2oM000B9cO0?ooo`0D0?ooo`050=nLB01d_ml0
W4P00018W03Oool02@3oool00`000000oooo0?ooo`060?ooo`040?oo_`1d0000000004RLgaD0oooo
00<0omnL04P0001d_ol0B`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00l0oooo
00<0omnL04P0001d_ol0603oool00`3ookl0M00004RLg`0D0?ooo`040000003oool0_g@004RLgaD0
oooo00D0gia807Bog`2LB00004RL0=ooo`0B0?ooo`040?oo_`1d0000000004RLgaD0oooo00<0omnL
04P0001d_ol0B`3oool00`000000oooo0?ooo`0D0?ooo`030?oOW0180000M;oo00P0oooo00<00000
0?ooo`3oool03@3oool00`3ookl0M00004RLg`0F0?ooo`030;md0018W=l0oooo01@0oooo00D0gia8
07Bog`2LB00004RL0=ooo`0B0?ooo`040?oo_`1d0000000004RLgaD0oooo00<0omnL04P0001d_ol0
7`3oool003d0oooo0P00000F0?ooo`030?oOW0180000M;oo0140oooo00<000000?noW02Lgol00`3o
ool00`3ogi`0B00004RLg`0A0?ooo`030?oog`2LW;l0oooo0080oooo00<0gia80000M02oool0503o
ool01@3ooml0W4Q804Q80000B9`0gooo00X0oooo00<000000?ooo`3oool01P3oool00`2oM000M;oo
0?ooo`0F0?ooo`030=nLB00004P0W=oo04X0oooo0P0000030?ooo`030000003oool0oooo0100oooo
00<0omnL04P0001d_ol04`3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO0140oooo
00P0oooO09bL_`3oool000000?ooo`3OW4P0001d0;oooa@0oooo00D0oooO09a8B018B00004RL0=oo
o`0C0?ooo`030;md001d_ol0oooo01H0oooo00<0gia80000B02Lgol0BP3oool00`000000oooo0?oo
o`0D0?ooo`030?oOW0180000M;oo00P0oooo00<000000?ooo`3oool0203oool00`3o_i`0W=oo0?oo
o`020?ooo`030?oOW0180000B9cO0140oooo00<0oooO09bL_`3oool00P3oool00`3OW4P0001d0;oo
o`0D0?ooo`050?oog`2LB4P0B4P00018W03Oool04`3oool00`000000M;oo0?ooo`0F0?ooo`030=nL
B00004P0W=oo01h0oooo000m0?ooo`030000003oool0oooo01<0oooo00D0oooO09a8000000000000
07Boo`0@0?ooo`0:0000003oool0oooO09a8B02Lgol0oooo0?oo_`1d0000001d0;oooa80oooo00H0
okmd04RLg`3oool0oono07@00000M;lF0?ooo`040?oOW018000004RL0=ooo`X0oooo00<000000?oo
o`3oool01P3oool01P3ogi`0B000000000000000001809cOoa@0oooo00<0okmd00000000M;l0B@3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00h0oooo00D0oooO09a8000000000000
07Boo`0C0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02ooolB0?ooo`060?noM018W=l0
oooo0000001d000007Bo5P3oool0103ogi`0B0000018W03OoolC0?ooo`060?oOW018000000000000
000004P0W=oo503oool00`3o_g@00000001d_`190?ooo`030000003oool0oooo0180oooo00D0oooO
09a800000000000007Boo`090?ooo`030000003oool0oooo00L0oooo00P0oooO09a8B02Lgol0oooo
0?oo_`1d0000001d0;oooa80oooo00H0okmd04RLg`3oool0oono07@00000M;lF0?ooo`040?oOW018
000004RL0=oooa<0oooo00H0000004P00000000000000000B02LgolD0?ooo`030?noM000000007Bo
01d0oooo000m0?ooo`030000003oool0oooo01D0oooo00<0okmd0000001d_ol0403oool01@000000
oooo0?ooo`3ooml0W4P00080000000<0001d0;ooo`3oool04`3oool01@3OW4P000000000000007@0
_ooo01L0oooo00<0gia80018W03Oool02`3oool00`000000oooo0?ooo`060?ooo`050;md00000000
000000000018W=l05@3oool01@3ooml0W4P00000000007@0_ooo04H0oooo00D000000?ooo`3oool0
oooo0000000C0?ooo`030?noM0000000M;oo01@0oooo00H0oooO09a80000000000000000M02ooolD
0?ooo`050=nLB000000000000000M02oool05`3oool00`3OW4P004RL0=ooo`0D0?ooo`050;md0000
0000000000000018W=l05@3oool01@3ooml0W4P00000000007@0_ooo04H0oooo00<000000?ooo`3o
ool0503oool00`3o_g@0000007Boo`090?ooo`030000003oool0oooo00P0oooo00H0oooO09a80000
000000000000M02ooolD0?ooo`050=nLB000000000000000M02oool05`3oool00`3OW4P004RL0=oo
o`0B0?ooo`030000003oool0_g@000<0000000<0B9cO0?ooo`3oool04`3oool01@3ooml0W4P00000
000007@0_ooo01X0oooo000m0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0A@3o
ool00`000000oooo0?ooo`1[0?ooo`040000003oool0oooo000004P0oooo00<000000?ooo`3oool0
TP3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0
>03oool003d0oooo0P00000X0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0JP3o
ool2000000030?ooo`000000oooo04L0oooo00<000000?ooo`3oool0TP3oool00`000000oooo0?oo
o`0Q0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool0>@3oool003d0oooo00<00000
0?ooo`3oool09P3oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo06X0oooo00<00000
0?ooo`000000BP3oool00`000000oooo0?ooo`2A0?ooo`030000003oool0oooo0280oooo00<00000
0?ooo`3oool0D03oool00`000000oooo0?ooo`0i0?ooo`00?@3oool00`000000oooo0?ooo`0V0?oo
o`030000003oool0oooo04P0oooo00<000000?ooo`3oool0J@3oool00`000000oooo0000001:0?oo
o`030000003oool0oooo0940oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`1?0?oo
o`030000003oool0oooo03X0oooo000m0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3o
ool0B@3oool00`000000oooo0?ooo`1Y0?ooo`800000C03oool00`000000oooo0?ooo`2@0?ooo`03
0000003oool0oooo02<0oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`0j0?ooo`00
?@3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0
J03oool2000004`0oooo00<000000?ooo`3oool0T03oool00`000000oooo0?ooo`0S0?ooo`030000
003oool0oooo04d0oooo00<000000?ooo`3oool0>`3oool003T0ooooYP00000j0?ooojH00000>`3o
oonV00000180oooo000m0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00@000000?ooo`3oool0oooo0P00
00040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool01P000000oooo0?oo
o`000000oooo000000@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0m0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`04
0000003oool0oooo0?ooo`800000103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
0080oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
0080oooo00@000000?ooo`3oool0oooo0P0000040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00P3oool00`000000oooo000000020?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0A0?ooo`00?@3oool00`000000oooo0?ooo`0S0?oo
o`030000003oool0oooo04d0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`1=0?oo
o`030000003oool0oooo04T0oooo00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`0U0?oo
o`030000003oool0oooo04X0oooo00<000000?ooo`3oool0?03oool003d0oooo00<000000?ooo`3o
ool08`3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo06H0oooo00<000000?ooo`3o
ool0C@3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo0480oooo00<000000?ooo`3o
ool09@3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo03d0oooo000m0?ooo`030000
003oool0oooo0280oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`1V0?ooo`030000
003oool0oooo04h0oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`130?ooo`030000
003oool0oooo02H0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`0m0?ooo`00?@3o
ool2000002<0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`1U0?ooo`800000C`3o
ool00`000000oooo0?ooo`170?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool09P3o
ool00`000000oooo0?ooo`170?ooo`030000003oool0oooo03h0oooo000m0?ooo`030000003oool0
oooo0240oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0
oooo04l0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`140?ooo`030000003oool0
oooo02L0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0n0?ooo`00?@3oool00`00
0000oooo0?ooo`0Q0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0I03oool00`00
0000oooo0?ooo`1?0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0A03oool00`00
0000oooo0?ooo`0W0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0?`3oool003d0
oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo06@0
oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo04D0
oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo03l0
oooo000m0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0E03oool00`000000oooo
0?ooo`1S0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0@`3oool00`000000oooo
0?ooo`150?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0@`3oool00`000000oooo
0?ooo`100?ooo`00?@3oool200000200oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?oo
o`1S0?ooo`800000DP3oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo04H0oooo00<0
00000?ooo`3oool0:@3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo0400oooo000m
0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`1R
0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`16
0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`11
0?ooo`00?@3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo05L0oooo00<000000?oo
o`3oool0HP3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo03l0oooo00<000000?oo
o`3oool0A`3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo0400oooo00<000000?oo
o`3oool0@@3oool003d0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`1H0?ooo`03
0000003oool0oooo0640oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`0o0?ooo`03
0000003oool0oooo04L0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0o0?ooo`03
0000003oool0oooo0480oooo000m0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0
FP3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool0
?@3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0
?@3oool00`000000oooo0?ooo`130?ooo`00?@3oool2000001h0oooo00<000000?ooo`3oool0FP3o
ool00`000000oooo0?ooo`1P0?ooo`800000E03oool00`000000oooo0?ooo`0m0?ooo`030000003o
ool0oooo04P0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0l0?ooo`030000003o
ool0oooo04<0oooo000m0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0G03oool0
0`000000oooo0?ooo`1O0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0>`3oool0
0`000000oooo0?ooo`190?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0>`3oool0
0`000000oooo0?ooo`140?ooo`00?@3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo
05`0oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo
03/0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo
03X0oooo00<000000?ooo`3oool0A03oool003d0oooo00<000000?ooo`3oool06`3oool00`000000
oooo0?ooo`1N0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool0E@3oool00`000000
oooo0?ooo`0i0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0;@3oool00`000000
oooo0?ooo`0i0?ooo`030000003oool0oooo04D0oooo000m0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool0G`3oool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo05D0oooo
00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo02h0oooo
00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`160?ooo`008`3oool01@3ogi`0B0000000
000007@0_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oog`2LB0030000000307Boo`3oool0oooo
00H0oooo0P00000K0?ooo`030000003oool0oooo0600oooo00<000000?ooo`3oool0A03oool01@3o
gi`0B0000000000007@0_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oog`2LB0030000000307Bo
o`3oool0oooo00D0oooo0P00001G0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0
<@3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oog`2LB003
0000000307Boo`3oool0oooo00H0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0g
0?ooo`030000003oool0oooo04H0oooo000R0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oo
o`020?ooo`070?noM0000000M;oo0?ooo`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00H0
oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo0480
oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00L0okmd0000001d_ol0oooo0?oo
g`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool01@3oool00`000000oooo0?ooo`1G0?ooo`030000
003oool0oooo03D0oooo00<000000?ooo`3oool0<@3oool01`3ookl0M00007Boo`3oool0okmd0000
M02oool00P3oool01`3o_g@0000007Boo`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?oo
o`060?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?oo
o`170?ooo`008P3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO
09a8M02oool01P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0680oooo00<00000
0?ooo`3oool0@P3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO
09a8M02oool01@3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo03D0oooo00<00000
0?ooo`3oool0<@3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO
09a8M02oool01P3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo03D0oooo00<00000
0?ooo`3oool0A`3oool00240oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0;0?oo
o`030;md001d_ol0oooo00D0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1T0?oo
o`030000003oool0oooo0400oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0;0?oo
o`030;md001d_ol0oooo00@0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`0c0?oo
o`030000003oool0oooo0340oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0;0?oo
o`030;md001d_ol0oooo00D0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0c0?oo
o`030000003oool0oooo04P0oooo000Q0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol02P3oool00`3ooml0W4P004RLg`060?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3o
ool0I03oool00`000000oooo0?ooo`100?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol02P3oool00`3ooml0W4P004RLg`050?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3o
ool0<`3oool00`000000oooo0?ooo`0a0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol02P3oool00`3ooml0W4P004RLg`060?ooo`030000003oool0oooo0300oooo00<000000?ooo`3o
ool0<`3oool00`000000oooo0?ooo`180?ooo`008@3oool00`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00X0oooo00<0gia80000001d_ol01P3oool2000001P0oooo00<000000?ooo`3oool0
IP3oool00`000000oooo0?ooo`0o0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
2P3oool00`3OW4P0000007Boo`050?ooo`800000FP3oool00`000000oooo0?ooo`0a0?ooo`030000
003oool0oooo0380oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030=nL
B0000000M;oo00H0oooo0P00000b0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0
B@3oool00240oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`070=oOg`3o
ool0oono07@000000000001d0;ooo`060?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3o
ool0J03oool00`000000oooo0?ooo`0n0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol01P3oool01`3Ogml0oooo0?oo_`1d000000000000M02oool01@3oool00`000000oooo0?ooo`1J
0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0<`3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00H0oooo
00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo04X0oooo
000Q0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3OW4P000000000
00000000001809cOo`L0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`1X0?ooo`03
0000003oool0oooo03h0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`06
0=nLB000000000000000000004P0W=oo1P3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0
oooo02l0oooo00<000000?ooo`3oool0<`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo00H0oooo00H0gia800000000000000000000B02Lgol70?ooo`030000003oool0oooo0380oooo
00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`1:0?ooo`008@3oool00`3ookl0M00004RL
g`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono07@000000000B9cO2@3oool00`000000oooo
0?ooo`0E0?ooo`030000003oool0oooo06X0oooo00<000000?ooo`3oool0?@3oool00`3ookl0M000
04RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono07@000000000B9cO203oool00`000000
oooo0?ooo`1K0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0=03oool00`3ookl0
M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono07@000000000B9cO2@3oool00`00
0000oooo0?ooo`0c0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0B`3oool00280
oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030;md001d_ol0oooo00T0
oooo0P00000E0?ooo`030000003oool0oooo06`0oooo00<000000?ooo`3oool0?@3oool00`3OW4P0
04RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0_g@007Boo`3oool0203oool2000005d0
oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`0f0?ooo`030=nLB000B9`0gooo0080
oooo00<0_g@0001d_`3oool01`3oool00`2oM000M;oo0?ooo`090?ooo`030000003oool0oooo03@0
oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`1;0?ooo`008P3oool01`3ookl0M000
07Boo`3oool0okmd0000M02oool0203oool01P3ogi`0B000000000000000001809cOo`H0oooo00<0
00000?ooo`3oool0503oool00`000000oooo0?ooo`1/0?ooo`030000003oool0oooo03d0oooo00L0
oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00H0omnL04P00000000000000000B02Lgol5
0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0g
0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`060?oOW0180000000000000000
04P0W=oo1P3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo02/0oooo00<000000?oo
o`3oool0C03oool002<0oooo00D0oono07@000000000001d0;ooo`0:0?ooo`050;md000000000000
00000018W=l01P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo06h0oooo00<00000
0?ooo`3oool0?@3oool01@3ookl0M0000000000007@0_ooo00X0oooo00D0_g@00000000000000000
04RLg`050?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool0:@3oool00`000000oooo
0?ooo`0h0?ooo`050?oo_`1d000000000000M02oool02P3oool01@2oM0000000000000000000B9cO
00H0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo
04d0oooo000m0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0L03oool00`000000
oooo0?ooo`1E0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool09`3oool00`000000
oooo0?ooo`1C0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0:03oool00`000000
oooo0?ooo`1=0?ooo`00?@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0700oooo
00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo02H0oooo
00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo02H0oooo
00<000000?ooo`3oool0CP3oool003d0oooo0P00000B0?ooo`030000003oool0oooo0780oooo00<0
00000?ooo`3oool0E03oool200000600oooo00<000000?ooo`3oool09@3oool00`000000oooo0?oo
o`1D0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?oo
o`1>0?ooo`00?@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo07@0oooo00<00000
0?ooo`3oool0D`3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo02<0oooo00<00000
0?ooo`3oool0E@3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo02@0oooo00<00000
0?ooo`3oool0C`3oool003d0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1e0?oo
o`030000003oool0oooo05<0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`0Q0?oo
o`030000003oool0oooo05H0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0R0?oo
o`030000003oool0oooo0500oooo000m0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3o
ool0M`3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3o
ool0803oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3o
ool08@3oool00`000000oooo0?ooo`1A0?ooo`00?@3oool00`000000oooo0?ooo`0=0?ooo`030000
003oool0oooo07P0oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`1R0?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`0j0?ooo`030000
003oool0oooo01l0oooo00<000000?ooo`3oool0DP3oool003d0oooo0P00000>0?ooo`030000003o
ool0oooo07T0oooo00<000000?ooo`3oool0D@3oool2000006@0oooo00<000000?ooo`3oool07@3o
ool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool07@3o
ool00`000000oooo0?ooo`1C0?ooo`00?@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo07/0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`1T0?ooo`030000003oool0
oooo01/0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0
oooo01/0oooo00<000000?ooo`3oool0E03oool003d0oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`1m0?ooo`800000D03oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`3oool0E@3oool003d0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`200?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3oool0IP3oool00`000000
oooo0?ooo`0G0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0?P3oool00`000000
oooo0?ooo`0G0?ooo`030000003oool0oooo05H0oooo000m0?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0PP3oool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo06L0oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo03l0oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`1G0?ooo`00<@3oool01P2oM00000000000
0000000004RL0=ooo`H0oooo0P0000090?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3o
ool0?`3oool01P2oM000000000000000000004RL0=ooo`H0oooo0P00001Y0?ooo`030000003oool0
oooo01<0oooo00<000000?ooo`3oool0D@3oool01P2oM000000000000000000004RL0=ooo`H0oooo
00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo05P0oooo
000b0?ooo`030?oOW0180000M;oo00P0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`260?ooo`030000003oool0oooo03l0oooo00<0omnL04P0001d_ol0203oool00`000000oooo0?oo
o`1Y0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0D`3oool00`3ogi`0B00007Bo
o`080?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`1I0?ooo`00<P3oool00`3ogi`0B00007Boo`080?ooo`030000003oool0oooo00D0oooo0P00002:
0?ooo`800000?`3oool00`3ogi`0B00007Boo`080?ooo`030000003oool0oooo06X0oooo0P00000?
0?ooo`800000EP3oool00`3ogi`0B00007Boo`080?ooo`030000003oool0oooo0480oooo00<00000
0?ooo`3oool03P3oool2000005`0oooo000b0?ooo`030?oOW0180000M;oo00P0oooo00<000000?oo
o`3oool00`3oool2000008h0oooo0P00000m0?ooo`030?oOW0180000M;oo00P0oooo00<000000?oo
o`3oool0K03oool2000000`0oooo00<000000?ooo`3oool0EP3oool00`3ogi`0B00007Boo`080?oo
o`030000003oool0oooo04<0oooo0`00000<0?ooo`030000003oool0oooo05`0oooo000b0?ooo`03
0?oOW0180000M;oo00P0oooo00<000000?ooo`3oool00`00002B0?ooo`<00000>P3oool00`3ogi`0
B00007Boo`080?ooo`030000003oool0oooo06h0oooo0`0000050?ooo`@00000F@3oool00`3ogi`0
B00007Boo`080?ooo`030000003oool0oooo04H0oooo0`0000060?ooo`<00000G`3oool00380oooo
00<0omnL04P0001d_ol0203oool3000009L0oooo1000000g0?ooo`030?oOW0180000M;oo00P0oooo
0P00001b0?ooo`D00000G@3oool00`3ogi`0B00007Boo`080?ooo`800000B@3oool700000680oooo
000b0?ooo`030?oOW0180000M;oo00P0oooo00<000000?ooo`3oool0dP3oool00`3ogi`0B00007Bo
o`080?ooo`030000003oool0oooo0=<0oooo00<0omnL04P0001d_ol0203oool00`000000oooo0?oo
o`2a0?ooo`00<P3oool00`3ogi`0B00007Boo`080?ooo`030000003oool0oooo0=80oooo00<0omnL
04P0001d_ol0203oool00`000000oooo0?ooo`3C0?ooo`030?oOW0180000M;oo00P0oooo00<00000
0?ooo`3oool0/@3oool00380oooo00<0omnL04P0001d_ol0g@3oool00`3ogi`0B00007Boo`3N0?oo
o`030?oOW0180000M;oo0;`0oooo000b0?ooo`030?oOW0180000M;oo0=d0oooo00<0omnL04P0001d
_ol0gP3oool00`3ogi`0B00007Boo`2l0?ooo`00<03oool01@3ooml0W4P0000000000000M;oo0=/0
oooo00D0oooO09a800000000000007Boo`3L0?ooo`050?oog`2LB00000000000001d_ol0_03oool0
0380oooo00<0okmd0000001d_ol0g@3oool00`3o_g@0000007Boo`3N0?ooo`030?noM0000000M;oo
0;`0oooo003o0?ooool0oooo]03oool00?l0ooooo`3ooond0?ooo`00o`3ooooo0?oook@0oooo003o
0?ooool0oooo]03oool00?l0ooooo`3ooond0?ooo`00o`3ooooo0?oook@0oooo003o0?ooool0oooo
]03oool00?l0ooooo`3ooond0?ooo`00o`3ooooo0?oook@0oooo003o0?ooool0oooo]03oool00?l0
ooooo`3ooond0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 689.625}, {132.375, 0}} -> {-0.0847624, -0.0144882, \
0.00689403, 0.00689403}, {{18.4375, 222.375}, {129.188, 3.125}} -> {-2.46022, \
-1.36505, 0.0600373, 0.0309212}, {{242.813, 446.75}, {129.188, 3.125}} -> \
{-11.4408, -1.36505, 0.0600373, 0.0309212}, {{467.188, 671.125}, {129.188, \
3.125}} -> {-20.4214, -1.36504, 0.0600373, 0.0309212}}]
}, Closed]],

Cell["\<\
A 2-dimensional array is displayed as a rectangular matrix.  Below is a \
degenerate 2-dimensional list that generates column output.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[GraphicsArray[{{plot[x]}, {plot[v]}, {plot[a]}}]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.97771 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0470883 0.952381 [
[ 0 0 0 0 ]
[ 1 1.97771 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.97771 L
0 1.97771 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0470883 0.97619 0.635692 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.151576 0.309018 0.294303 [
[.17539 .29652 -5 -15.6875 ]
[.17539 .29652 5 0 ]
[.32696 .29652 -5 -15.6875 ]
[.32696 .29652 5 0 ]
[.47854 .29652 -5 -15.6875 ]
[.47854 .29652 5 0 ]
[.63011 .29652 -5 -15.6875 ]
[.63011 .29652 5 0 ]
[.78169 .29652 -5 -15.6875 ]
[.78169 .29652 5 0 ]
[.93327 .29652 -5 -15.6875 ]
[.93327 .29652 5 0 ]
[.01131 .01472 -18.3125 -7.84375 ]
[.01131 .01472 0 7.84375 ]
[.01131 .16187 -27.3125 -7.84375 ]
[.01131 .16187 0 7.84375 ]
[.01131 .45617 -19 -7.84375 ]
[.01131 .45617 0 7.84375 ]
[.01131 .60332 -10 -7.84375 ]
[.01131 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.17539 .30902 m
.17539 .31527 L
s
gsave
.17539 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.32696 .30902 m
.32696 .31527 L
s
gsave
.32696 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.47854 .30902 m
.47854 .31527 L
s
gsave
.47854 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.63011 .30902 m
.63011 .31527 L
s
gsave
.63011 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.78169 .30902 m
.78169 .31527 L
s
gsave
.78169 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.93327 .30902 m
.93327 .31527 L
s
gsave
.93327 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.125 Mabswid
.05412 .30902 m
.05412 .31277 L
s
.08444 .30902 m
.08444 .31277 L
s
.11476 .30902 m
.11476 .31277 L
s
.14507 .30902 m
.14507 .31277 L
s
.2057 .30902 m
.2057 .31277 L
s
.23602 .30902 m
.23602 .31277 L
s
.26633 .30902 m
.26633 .31277 L
s
.29665 .30902 m
.29665 .31277 L
s
.35728 .30902 m
.35728 .31277 L
s
.38759 .30902 m
.38759 .31277 L
s
.41791 .30902 m
.41791 .31277 L
s
.44822 .30902 m
.44822 .31277 L
s
.50885 .30902 m
.50885 .31277 L
s
.53917 .30902 m
.53917 .31277 L
s
.56948 .30902 m
.56948 .31277 L
s
.5998 .30902 m
.5998 .31277 L
s
.66043 .30902 m
.66043 .31277 L
s
.69074 .30902 m
.69074 .31277 L
s
.72106 .30902 m
.72106 .31277 L
s
.75138 .30902 m
.75138 .31277 L
s
.81201 .30902 m
.81201 .31277 L
s
.84232 .30902 m
.84232 .31277 L
s
.87264 .30902 m
.87264 .31277 L
s
.90295 .30902 m
.90295 .31277 L
s
.96358 .30902 m
.96358 .31277 L
s
.9939 .30902 m
.9939 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.02381 .01472 m
.03006 .01472 L
s
gsave
.01131 .01472 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .16187 m
.03006 .16187 L
s
gsave
.01131 .16187 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .45617 m
.03006 .45617 L
s
gsave
.01131 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .60332 m
.03006 .60332 L
s
gsave
.01131 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04415 m
.02756 .04415 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .10301 m
.02756 .10301 L
s
.02381 .13244 m
.02756 .13244 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .25016 m
.02756 .25016 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .36788 m
.02756 .36788 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .4856 m
.02756 .4856 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01472 m
.02499 .01472 L
.02605 .01475 L
.02729 .01479 L
.02846 .01485 L
.03053 .015 L
.03279 .01523 L
.03527 .01556 L
.0379 .01599 L
.04262 .01698 L
.04749 .0183 L
.05205 .01981 L
.06244 .02422 L
.07305 .03011 L
.08274 .03668 L
.10458 .05552 L
.14429 .10289 L
.18248 .16177 L
.22313 .23454 L
.26226 .3097 L
.30384 .3894 L
.34391 .46058 L
.38246 .51918 L
.402 .54393 L
.42346 .56659 L
.44388 .58338 L
.45301 .58929 L
.46295 .59457 L
.46838 .59694 L
.4734 .5988 L
.4781 .60026 L
.4833 .60154 L
.48553 .60198 L
.48794 .60239 L
.49002 .60268 L
.4923 .60294 L
.49473 .60314 L
.49605 .60322 L
.49728 .60327 L
.49844 .6033 L
.49949 .60332 L
.50071 .60332 L
.50186 .6033 L
.50301 .60326 L
.50425 .60321 L
.50542 .60313 L
.50648 .60305 L
.5091 .60279 L
.51158 .60246 L
.5165 .60158 L
Mistroke
.52187 .60026 L
.53144 .59701 L
.54032 .59297 L
.56088 .5799 L
.5795 .56376 L
.61962 .51633 L
.65822 .45702 L
.69928 .38356 L
.73882 .30761 L
.77684 .2346 L
.81732 .1621 L
.85628 .1021 L
.87613 .07654 L
.89769 .0533 L
.91843 .03582 L
.92832 .02927 L
.93759 .02421 L
.94643 .02037 L
.95595 .01733 L
.9611 .01617 L
.96369 .01572 L
.96652 .01531 L
.96894 .01505 L
.9703 .01494 L
.97153 .01485 L
.97268 .01479 L
.97393 .01475 L
.97511 .01472 L
.97619 .01472 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.0238095 0.694552 0.97619 1.28316 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.151576 0.309018 0.294303 [
[.17539 .29652 -5 -15.6875 ]
[.17539 .29652 5 0 ]
[.32696 .29652 -5 -15.6875 ]
[.32696 .29652 5 0 ]
[.47854 .29652 -5 -15.6875 ]
[.47854 .29652 5 0 ]
[.63011 .29652 -5 -15.6875 ]
[.63011 .29652 5 0 ]
[.78169 .29652 -5 -15.6875 ]
[.78169 .29652 5 0 ]
[.93327 .29652 -5 -15.6875 ]
[.93327 .29652 5 0 ]
[.01131 .01471 -18.3125 -7.84375 ]
[.01131 .01471 0 7.84375 ]
[.01131 .16187 -27.3125 -7.84375 ]
[.01131 .16187 0 7.84375 ]
[.01131 .45617 -19 -7.84375 ]
[.01131 .45617 0 7.84375 ]
[.01131 .60332 -10 -7.84375 ]
[.01131 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.17539 .30902 m
.17539 .31527 L
s
gsave
.17539 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.32696 .30902 m
.32696 .31527 L
s
gsave
.32696 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.47854 .30902 m
.47854 .31527 L
s
gsave
.47854 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.63011 .30902 m
.63011 .31527 L
s
gsave
.63011 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.78169 .30902 m
.78169 .31527 L
s
gsave
.78169 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.93327 .30902 m
.93327 .31527 L
s
gsave
.93327 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.125 Mabswid
.05412 .30902 m
.05412 .31277 L
s
.08444 .30902 m
.08444 .31277 L
s
.11476 .30902 m
.11476 .31277 L
s
.14507 .30902 m
.14507 .31277 L
s
.2057 .30902 m
.2057 .31277 L
s
.23602 .30902 m
.23602 .31277 L
s
.26633 .30902 m
.26633 .31277 L
s
.29665 .30902 m
.29665 .31277 L
s
.35728 .30902 m
.35728 .31277 L
s
.38759 .30902 m
.38759 .31277 L
s
.41791 .30902 m
.41791 .31277 L
s
.44822 .30902 m
.44822 .31277 L
s
.50885 .30902 m
.50885 .31277 L
s
.53917 .30902 m
.53917 .31277 L
s
.56948 .30902 m
.56948 .31277 L
s
.5998 .30902 m
.5998 .31277 L
s
.66043 .30902 m
.66043 .31277 L
s
.69074 .30902 m
.69074 .31277 L
s
.72106 .30902 m
.72106 .31277 L
s
.75138 .30902 m
.75138 .31277 L
s
.81201 .30902 m
.81201 .31277 L
s
.84232 .30902 m
.84232 .31277 L
s
.87264 .30902 m
.87264 .31277 L
s
.90295 .30902 m
.90295 .31277 L
s
.96358 .30902 m
.96358 .31277 L
s
.9939 .30902 m
.9939 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.02381 .01471 m
.03006 .01471 L
s
gsave
.01131 .01471 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .16187 m
.03006 .16187 L
s
gsave
.01131 .16187 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .45617 m
.03006 .45617 L
s
gsave
.01131 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .60332 m
.03006 .60332 L
s
gsave
.01131 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04415 m
.02756 .04415 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .10301 m
.02756 .10301 L
s
.02381 .13244 m
.02756 .13244 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .25016 m
.02756 .25016 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .36788 m
.02756 .36788 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .4856 m
.02756 .4856 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .30902 m
.06244 .23481 L
.10458 .15951 L
.14415 .09914 L
.18221 .05447 L
.20178 .03757 L
.21267 .03011 L
.22272 .0245 L
.23293 .02007 L
.23832 .01827 L
.24402 .01676 L
.24641 .01625 L
.24897 .01579 L
.25119 .01545 L
.25361 .01516 L
.25507 .01501 L
.25643 .01491 L
.25773 .01483 L
.2584 .01479 L
.25912 .01476 L
.25983 .01474 L
.26048 .01473 L
.26176 .01472 L
.26293 .01472 L
.26416 .01475 L
.26538 .01479 L
.26607 .01483 L
.26671 .01486 L
.26912 .01505 L
.27131 .01528 L
.27364 .0156 L
.27847 .01647 L
.2838 .01778 L
.29326 .02099 L
.30213 .02502 L
.32232 .03778 L
.34082 .05371 L
.38045 .10022 L
.41855 .15838 L
.45912 .23059 L
.49816 .30545 L
.53966 .38514 L
.57964 .45663 L
.61811 .51582 L
.63759 .54096 L
.65902 .56418 L
.67938 .58152 L
.69843 .5933 L
.70884 .59785 L
.71396 .5996 L
Mistroke
.71872 .60092 L
.72301 .60186 L
.72758 .60261 L
.7302 .60292 L
.73148 .60304 L
.73266 .60313 L
.73371 .6032 L
.73487 .60325 L
.73603 .60329 L
.73725 .60332 L
.73856 .60332 L
.73977 .6033 L
.7404 .60329 L
.7411 .60326 L
.74251 .6032 L
.74375 .60312 L
.74508 .60301 L
.74748 .60276 L
.74985 .60244 L
.75235 .60202 L
.7568 .60108 L
.76198 .59967 L
.7668 .59806 L
.77773 .59332 L
.78828 .58734 L
.79823 .58046 L
.81683 .5645 L
.85534 .51958 L
.89632 .45703 L
.93577 .38657 L
.97371 .31383 L
.97619 .30902 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.0238095 1.34202 0.97619 1.93062 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.151576 0.309016 0.294303 [
[.17539 .29652 -5 -15.6875 ]
[.17539 .29652 5 0 ]
[.32696 .29652 -5 -15.6875 ]
[.32696 .29652 5 0 ]
[.47854 .29652 -5 -15.6875 ]
[.47854 .29652 5 0 ]
[.63011 .29652 -5 -15.6875 ]
[.63011 .29652 5 0 ]
[.78169 .29652 -5 -15.6875 ]
[.78169 .29652 5 0 ]
[.93327 .29652 -5 -15.6875 ]
[.93327 .29652 5 0 ]
[.01131 .01471 -18.3125 -7.84375 ]
[.01131 .01471 0 7.84375 ]
[.01131 .16186 -27.3125 -7.84375 ]
[.01131 .16186 0 7.84375 ]
[.01131 .45617 -19 -7.84375 ]
[.01131 .45617 0 7.84375 ]
[.01131 .60332 -10 -7.84375 ]
[.01131 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.17539 .30902 m
.17539 .31527 L
s
gsave
.17539 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.32696 .30902 m
.32696 .31527 L
s
gsave
.32696 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.47854 .30902 m
.47854 .31527 L
s
gsave
.47854 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.63011 .30902 m
.63011 .31527 L
s
gsave
.63011 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.78169 .30902 m
.78169 .31527 L
s
gsave
.78169 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.93327 .30902 m
.93327 .31527 L
s
gsave
.93327 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.125 Mabswid
.05412 .30902 m
.05412 .31277 L
s
.08444 .30902 m
.08444 .31277 L
s
.11476 .30902 m
.11476 .31277 L
s
.14507 .30902 m
.14507 .31277 L
s
.2057 .30902 m
.2057 .31277 L
s
.23602 .30902 m
.23602 .31277 L
s
.26633 .30902 m
.26633 .31277 L
s
.29665 .30902 m
.29665 .31277 L
s
.35728 .30902 m
.35728 .31277 L
s
.38759 .30902 m
.38759 .31277 L
s
.41791 .30902 m
.41791 .31277 L
s
.44822 .30902 m
.44822 .31277 L
s
.50885 .30902 m
.50885 .31277 L
s
.53917 .30902 m
.53917 .31277 L
s
.56948 .30902 m
.56948 .31277 L
s
.5998 .30902 m
.5998 .31277 L
s
.66043 .30902 m
.66043 .31277 L
s
.69074 .30902 m
.69074 .31277 L
s
.72106 .30902 m
.72106 .31277 L
s
.75138 .30902 m
.75138 .31277 L
s
.81201 .30902 m
.81201 .31277 L
s
.84232 .30902 m
.84232 .31277 L
s
.87264 .30902 m
.87264 .31277 L
s
.90295 .30902 m
.90295 .31277 L
s
.96358 .30902 m
.96358 .31277 L
s
.9939 .30902 m
.9939 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.02381 .01471 m
.03006 .01471 L
s
gsave
.01131 .01471 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .16186 m
.03006 .16186 L
s
gsave
.01131 .16186 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .45617 m
.03006 .45617 L
s
gsave
.01131 .45617 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .60332 m
.03006 .60332 L
s
gsave
.01131 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04414 m
.02756 .04414 L
s
.02381 .07357 m
.02756 .07357 L
s
.02381 .103 m
.02756 .103 L
s
.02381 .13243 m
.02756 .13243 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .25016 m
.02756 .25016 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .36788 m
.02756 .36788 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .4856 m
.02756 .4856 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .60332 m
.02499 .60331 L
.02605 .60329 L
.02729 .60324 L
.02846 .60318 L
.03053 .60303 L
.03279 .6028 L
.03527 .60248 L
.0379 .60205 L
.04262 .60106 L
.04749 .59974 L
.05205 .59822 L
.06244 .59381 L
.07305 .58793 L
.08274 .58135 L
.10458 .56251 L
.14429 .51515 L
.18248 .45627 L
.22313 .38349 L
.26226 .30833 L
.30384 .22863 L
.34391 .15746 L
.38246 .09886 L
.402 .07411 L
.42346 .05145 L
.44388 .03465 L
.45301 .02874 L
.46295 .02346 L
.46838 .0211 L
.4734 .01923 L
.4781 .01778 L
.4833 .0165 L
.48553 .01605 L
.48794 .01564 L
.49002 .01535 L
.4923 .01509 L
.49473 .01489 L
.49605 .01481 L
.49728 .01476 L
.49844 .01473 L
.49949 .01472 L
.50071 .01472 L
.50186 .01474 L
.50301 .01477 L
.50425 .01483 L
.50542 .0149 L
.50648 .01498 L
.5091 .01524 L
.51158 .01557 L
.5165 .01646 L
Mistroke
.52187 .01777 L
.53144 .02102 L
.54032 .02507 L
.56088 .03814 L
.5795 .05428 L
.61962 .1017 L
.65822 .16102 L
.69928 .23447 L
.73882 .31042 L
.77684 .38343 L
.81732 .45593 L
.85628 .51593 L
.87613 .54149 L
.89769 .56473 L
.91843 .58221 L
.92832 .58876 L
.93759 .59383 L
.94643 .59767 L
.95595 .6007 L
.9611 .60186 L
.96369 .60232 L
.96652 .60272 L
.96894 .60298 L
.9703 .6031 L
.97153 .60318 L
.97268 .60324 L
.97393 .60329 L
.97511 .60331 L
.97619 .60332 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{128.5, 254.125},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000300001OB000`400?l00000o`00003o`03oool0
0<00oooo00300?ooo`00`03oool00<00oooo00300?ooo`00`03oool00<00oooo00300?ooo`00`03o
ool00<00oooo00300?ooo`00`03oool00<00oooo00300?ooo`00`03oool00<00oooo00300?ooo`00
`03oool002<0oooo00H0_g@000000000000000000018W03OoonG0?ooo`00903oool00`3ogi`0B000
07Boo`2I0?ooo`00903oool00`3ogi`0B00007Boo`2I0?ooo`00903oool00`3ogi`0B00007Boo`08
0?ooo`030000003oool0oooo08h0oooo000E0?ooo`030?oog`2LB000000000L0000000<0B9cO0?oo
o`3oool00P3oool00`3ogi`0B00007Boo`080?ooo`030000003oool0oooo08h0oooo000T0?ooo`03
0?oOW0180000M;oo00P0oooo100000220?ooo`<00000203oool002@0oooo00<0omnL04P0001d_ol0
203oool00`000000oooo0?ooo`03000007d0oooo0`00000;0?ooo`00903oool00`3ogi`0B00007Bo
o`080?ooo`030000003oool0oooo00<0oooo0P00001i0?ooo`8000003P3oool002@0oooo00<0omnL
04P0001d_ol0203oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo07D0oooo00<00000
0?ooo`3oool03P3oool002@0oooo00<0omnL04P0001d_ol0203oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo07<0oooo00<000000?ooo`3oool03`3oool00280oooo00D0oooO09a80000
0000000007Boo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0L@3oool00`00
0000oooo0?ooo`0@0?ooo`00903oool00`3o_g@0000007Boo`080?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`0A0?ooo`00;`3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool04P3oool002l0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo01<0oooo000_
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0J@3oool00`000000oooo0?ooo`0D
0?ooo`00;`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo06P0oooo00<000000?oo
o`3oool05@3oool002l0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1V0?ooo`03
0000003oool0oooo01H0oooo000_0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
I@3oool00`000000oooo0?ooo`0F0?ooo`00;`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo06<0oooo00<000000?ooo`3oool05`3oool002l0oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo01P0oooo000_0?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool0H03oool00`000000oooo0?ooo`0I0?ooo`005P3oool01@3ogi`0
B0000000000007@0_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oog`2LB0030000000307Boo`3o
ool0oooo00D0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1O0?ooo`030000003o
ool0oooo01T0oooo000E0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?no
M0000000M;oo0?ooo`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00D0oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo01X0oooo000E0?ooo`03
0;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd0;ooo`050?ooo`03
0000003oool0oooo0180oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0K0?ooo`00
503oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00/0oooo00<0_g@007Boo`3oool0
103oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool0
703oool000L0oooo00<0oooO09a8000000001`0000001P18W=l0oooo0?ooo`3ogi`0B00007Boo`<0
oooo00<0okmd0000B02Lgol02P3oool00`3ooml0W4P004RLg`050?ooo`030000003oool0oooo01<0
oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`0L0?ooo`00503oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0gia80000001d_ol01@3oool2000001D0oooo
00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`0M0?ooo`00503oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00D0
oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo01d0
oooo000D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3OW4P00000
000000000000001809cOo`H0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`1E0?oo
o`030000003oool0oooo01h0oooo000D0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02o
ool01P3oool0103ookl0M00000000018W=l80?ooo`030000003oool0oooo0100oooo00H0_g@00000
0000000000000018W000000>0?ooo`030;md00000000000000@0000000<0B9cO0?ooo`3oool02`3o
ool00`3ookl0M000000000020000000307Boo`3oool0oooo0180oooo00@0oooO09a80000B9`0gooo
3P3oool00`3ooml0W4P0000000020000000407Boo`3oool0oooo000000l0oooo00D0omnL04P00000
0000000007Boo`0<0?ooo`005@3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0
oooo00<0_g@007Boo`3oool0203oool00`000000oooo0?ooo`0A0?ooo`030?oOW0180000M;oo0080
oooo00<000000?ooo`3oool0303oool00`3OW4P0000000000003000000030018W03Oool0oooo00/0
oooo00<0omnL04P00018W=l00P3oool00`2oM000B9cO0?ooo`0B0?ooo`040?oog`2LB00004RL0=oo
o`h0oooo00<0oooO09a8M02oool00P3oool00`2oM000M;oo0000000?0?ooo`070?oOW0180000B9cO
0?ooo`3ookl0M00007Boo`0;0?ooo`005@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0
203oool01P3ogi`0B000000000000000001809cOo`D0oooo00<000000?ooo`3oool04@3oool00`3o
gi`0B00007Boo`020?ooo`030000003oool0oooo00d0oooo00<0gia804RLg`3oool00P3oool00`3O
W7@0W=oo0?ooo`0@0?ooo`030?oo_`1d04P0W=oo0180oooo00@0oooO09a80000B9`0gooo4`3oool0
0`3ooml0W4Qd0000000>0?ooo`030?oog`2LB00007Bo00<0oooo00<0okmd0018W03Oool02P3oool0
01H0oooo00D0oono07@000000000001d0;ooo`0:0?ooo`050;md00000000000000000018W=l01@3o
ool00`000000oooo0?ooo`0A0?ooo`030?oOW0180000M;oo00<0oooo00<000000?ooo`3oool03@3o
ool00`3OW4P0M;oo0?ooo`0D0?ooo`030?oo_`1d0000M;oo00h0oooo00<0_g@00000000000001@00
00000`1d_ol0oooo0?ooo`0A0?ooo`030000001d_ol0oooo00d0oooo00<0omnL04P0001d_ol00`3o
ool00`3o_g@0001809cOo`0:0?ooo`00;`3oool00`000000oooo0?ooo`0A0?ooo`030?oOW0180000
M;oo00<0oooo00<000000?ooo`3oool03P3oool00`2oM4P0W=oo0?ooo`0C0?ooo`030?noM0000000
M;oo00h0oooo00P0omnL07BLg`3oool0oooo0?oog`2LB00004RL0=oooa<0oooo00<0oooO00000018
W=l03P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo000_0?ooo`030000
003oool0oooo0140oooo00<0omnL04P0001d_ol0103oool00`000000oooo0?ooo`0=0?ooo`030?oo
g`2LB7@0_ooo0180oooo00@0omnL04P0000004P0W=oo3`3oool01`3OW4P0M;oo0?ooo`3ooml0W4P0
0018W03Oool04`3oool20000000307Boo`3oool0oooo00`0oooo00<0omnL04P0001d_ol00`3oool0
0`2oM000001809cOo`0:0?ooo`00;`3oool00`000000oooo0?ooo`0A0?ooo`030?oOW0180000M;oo
00@0oooo00<000000?ooo`3oool03P3oool00`3o_g@007Bo0?ooo`0?0?ooo`050?noM00000000000
0000B02Lgol0403oool01`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool03`3oool01`3Ogml0oooo
0?oo_`1d000000000000M02oool03P3oool0203ookl0M00000000018W=l0oooo0?noM000000007Bo
2`3oool002l0oooo00<000000?ooo`3oool04@3oool00`3ogi`0B00007Boo`050?ooo`030000003o
ool0oooo00d0oooo00<0oono07@0B02Lgol04@3oool00`2oM00007Bo0?ooo`0A0?ooo`060?oOW018
M;l0oooO09a80000B9`0gooo3`3oool01P3OW4P00000000000000000001809cOoa00oooo00<0_g@0
0018W02oW4P00P0000000`00B9`0gooo0?ooo`0:0?ooo`00;`3oool00`000000oooo0?ooo`0A0?oo
o`030?oOW0180000M;oo00D0oooo00<000000?ooo`3oool03@3oool00`3ookl0M00004RLg`0B0?oo
o`030;md0018W=l0oooo0140oooo00D0gia807Bog`2LB00004RL0=ooo`0?0?ooo`040?oo_`1d0000
000000000180oooo00<0omnL04P0001d_ol03`3oool002l0oooo00<000000?ooo`3oool04@3oool0
0`3ogi`0B00007Boo`060?ooo`030000003oool0oooo00L0oooo00<0oknL09cOo`3oool00P3oool0
0`3ogi`0B00004RLg`0=0?ooo`030?oog`2LW;l0oooo0080oooo00<0gia80000M02oool04@3oool0
1@3ooml0W4Q804Q80000B9`0gooo0100oooo00<0_g@00000003oool04`3oool00`3OW4P0001809cO
o`0>0?ooo`00;`3oool00`000000oooo0?ooo`0?0?ooo`050?oog`2LB00000000000001d_ol01P3o
ool00`000000oooo0?ooo`070?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool>0?oo
o`060?noM018W=l0oooo0?oo_`1d000007Bo4`3oool0103ogi`0B0000018W03Oool@0?ooo`060?oO
W000000000000000000004P0W=oo4@3oool00`3o_g@00000001d_`0=0?ooo`00;`3oool00`000000
oooo0?ooo`0A0?ooo`030?noM0000000M;oo00L0oooo00<000000?ooo`3oool01`3oool01P3ooml0
W4P0000000000000001d0;oooa00oooo00D0gia8000000000000001d0;ooo`0D0?ooo`030=nLB000
B9`0gooo0100oooo00<000000;md000000000P0000000`18W=l0oooo0?ooo`0@0?ooo`050?oog`2L
B00000000000M02oool02P3oool002l0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?oo
o`190?ooo`030000003oool0oooo02@0oooo000_0?ooo`030000003oool0oooo01`0oooo00<00000
0?ooo`3oool0A`3oool00`000000oooo0?ooo`0U0?ooo`00;`3oool00`000000oooo0?ooo`0M0?oo
o`030000003oool0oooo04D0oooo00<000000?ooo`3oool09P3oool002l0oooo00<000000?ooo`3o
ool07@3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo02H0oooo000_0?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`0W0?ooo`00;`3o
ool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool09`3o
ool002/0ooooT@0000040?ooo`00;`3oool01@000000oooo0?ooo`3oool0000000@0oooo00D00000
0?ooo`3oool0oooo000000030?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo
000000030?ooo`040000003oool0oooo0?ooo`8000000`3oool01@000000oooo0?ooo`3oool00000
00@0oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0oooo0080oooo00D00000
0?ooo`3oool0oooo000000030?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo
000000030?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo000000030?ooo`03
0000003oool0oooo0080oooo00D000000?ooo`000000oooo000000030?ooo`030000003oool0oooo
0080oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0oooo0080oooo00D00000
0?ooo`3oool0oooo000000030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool002l0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0o0?ooo`030000003oool0
oooo02T0oooo000_0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0?`3oool00`00
0000oooo0?ooo`0Y0?ooo`00;`3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo03d0
oooo00<000000?ooo`3oool0:P3oool002l0oooo00<000000?ooo`3oool08@3oool00`000000oooo
0?ooo`0m0?ooo`030000003oool0oooo02X0oooo000_0?ooo`030000003oool0oooo0280oooo00<0
00000?ooo`3oool0>`3oool00`000000oooo0?ooo`0[0?ooo`00;`3oool00`000000oooo0?ooo`0R
0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0:`3oool002l0oooo00<000000?oo
o`3oool08`3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo02`0oooo000_0?ooo`03
0000003oool0oooo02<0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0/0?ooo`00
;`3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0
;@3oool002l0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`0g0?ooo`030000003o
ool0oooo02d0oooo000_0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0=@3oool0
0`000000oooo0?ooo`0^0?ooo`00;`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo
03@0oooo00<000000?ooo`3oool0;P3oool002l0oooo00<000000?ooo`3oool09P3oool00`000000
oooo0?ooo`0c0?ooo`030000003oool0oooo02l0oooo000_0?ooo`030000003oool0oooo02L0oooo
00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0_0?ooo`005P3oool01@3ogi`0B0000000
000007@0_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oog`2LB0030000000307Boo`3oool0oooo
00D0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo
0300oooo000E0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000
M;oo0?ooo`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00D0oooo00<000000?ooo`3oool0
:03oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo0340oooo000E0?ooo`030;md0000
B9`0gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd0;ooo`050?ooo`030000003o
ool0oooo02T0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0a0?ooo`00503oool0
0`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00/0oooo00<0_g@007Boo`3oool0103oool0
0`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0<P3oool0
01@0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030?oog`2LB000B9cO
00D0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo
0380oooo000D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3OW4P0
000007Boo`050?ooo`800000:`3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo03<0
oooo000D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01`3Ogml0oooo
0?oo_`1d000000000000M02oool01@3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo
02T0oooo00<000000?ooo`3oool0=03oool001@0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`060?ooo`060=nLB000000000000000000004P0W=oo1P3oool00`000000oooo0?ooo`0/
0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0=03oool001@0oooo00<0oono07@0
0018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d0000000004RLg`P0oooo00<00000
0?ooo`3oool0;03oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo03D0oooo000E0?oo
o`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`2oM000M;oo0?ooo`080?oo
o`030000003oool0oooo02d0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0f0?oo
o`005@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0203oool01P3ogi`0B00000000000
0000001809cOo`D0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0U0?ooo`030000
003oool0oooo03H0oooo000F0?ooo`050?oo_`1d000000000000M02oool02P3oool01@2oM0000000
000000000000B9cO00D0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0S0?ooo`03
0000003oool0oooo03L0oooo000_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0
8@3oool00`000000oooo0?ooo`0h0?ooo`00;`3oool00`000000oooo0?ooo`0_0?ooo`030000003o
ool0oooo0240oooo00<000000?ooo`3oool0>03oool002l0oooo00<000000?ooo`3oool0<03oool0
0`000000oooo0?ooo`0O0?ooo`030000003oool0oooo03T0oooo000_0?ooo`030000003oool0oooo
0340oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0j0?ooo`00;`3oool00`000000
oooo0?ooo`0a0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0>`3oool002l0oooo
00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo03`0oooo
000_0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`0l0?ooo`00;`3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool0?@3oool002l0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo03h0oooo000S0?ooo`060;md00000000000000000000B9`0gooo1P3oool0
0`000000oooo0?ooo`0f0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0?`3oool0
02@0oooo00<0omnL04P0001d_ol0203oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool0@03oool002@0oooo00<0omnL04P0001d_ol0203oool00`000000
oooo0?ooo`0h0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0@@3oool002@0oooo
00<0omnL04P0001d_ol0203oool00`000000oooo0?ooo`0i0?ooo`8000003@3oool2000004@0oooo
000T0?ooo`030?oOW0180000M;oo00P0oooo00<000000?ooo`3oool0>`3oool2000000X0oooo00<0
00000?ooo`3oool0A03oool002@0oooo00<0omnL04P0001d_ol0203oool2000003d0oooo2`000017
0?ooo`00903oool00`3ogi`0B00007Boo`080?ooo`030000003oool0oooo08h0oooo000T0?ooo`03
0?oOW0180000M;oo00P0oooo00<000000?ooo`3oool0SP3oool002@0oooo00<0omnL04P0001d_ol0
203oool00`000000oooo0?ooo`2>0?ooo`00903oool00`3ogi`0B00007Boo`2I0?ooo`008P3oool0
1@3ooml0W4P0000000000000M;oo09T0oooo000T0?ooo`030?noM0000000M;oo09T0oooo00300?oo
o`00`03oool00<00oooo00300?ooo`00`03oool00<00oooo00300?ooo`00`03oool00<00oooo0030
0?ooo`00`03oool00<00oooo00300?ooo`00`03oool00<00oooo00300?ooo`00`03oool00<00oooo
00300?ooo`00`03oool00<00oooo00300?ooo`00`03oool00<00oooo00300?ooo`00`03oool00<00
oooo00300?ooo`008`3oool01P2oM000000000000000000004RL0=oooiL0oooo000T0?ooo`030?oO
W0180000M;oo09T0oooo000T0?ooo`030?oOW0180000M;oo09T0oooo000T0?ooo`030?oOW0180000
M;oo00P0oooo00<000000?ooo`3oool0SP3oool001D0oooo00<0oooO09a8000000001`0000000`18
W=l0oooo0?ooo`020?ooo`030?oOW0180000M;oo00P0oooo00<000000?ooo`3oool0SP3oool002@0
oooo00<0omnL04P0001d_ol0203oool2000001h0oooo1@00001/0?ooo`00903oool00`3ogi`0B000
07Boo`080?ooo`030000003oool0oooo01/0oooo0`0000030?ooo`<00000JP3oool002@0oooo00<0
omnL04P0001d_ol0203oool00`000000oooo0?ooo`0H0?ooo`<000002@3oool2000006P0oooo000T
0?ooo`030?oOW0180000M;oo00P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0<
0?ooo`800000IP3oool002@0oooo00<0omnL04P0001d_ol0203oool00`000000oooo0?ooo`0F0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool0H`3oool00280oooo00D0oooO09a80000
0000000007Boo`080?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`1R0?ooo`00903oool00`3o_g@0000007Boo`080?ooo`030000003oool0oooo01@0
oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`1Q0?ooo`00;`3oool00`000000oooo
0?ooo`0C0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0H03oool002l0oooo00<0
00000?ooo`3oool04P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo05l0oooo000_
0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`1N
0?ooo`00;`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool0GP3oool002l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0L0?ooo`03
0000003oool0oooo05d0oooo000_0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
7P3oool00`000000oooo0?ooo`1L0?ooo`00;`3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo0200oooo00<000000?ooo`3oool0F`3oool002l0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo05X0oooo000_0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`1J0?ooo`005P3oool01@3ogi`0
B0000000000007@0_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oog`2LB0030000000307Boo`3o
ool0oooo00D0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0T0?ooo`030000003o
ool0oooo05T0oooo000E0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?no
M0000000M;oo0?ooo`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00D0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo05T0oooo000E0?ooo`03
0;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd0;ooo`050?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`1H0?ooo`00
503oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00/0oooo00<0_g@007Boo`3oool0
103oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0
E`3oool000L0oooo00<0oooO09a8000000001`0000001P18W=l0oooo0?ooo`3ogi`0B00007Boo`<0
oooo00<0okmd0000B02Lgol02P3oool00`3ooml0W4P004RLg`050?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`1G0?ooo`00503oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0gia80000001d_ol01@3oool2000000T0oooo
00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`1F0?ooo`00503oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00D0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo05D0
oooo000D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3OW4P00000
000000000000001809cOo`H0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0/0?oo
o`030000003oool0oooo05D0oooo000D0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02o
ool01P3oool0103ookl0M00000000018W=l80?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool01`3oool01P2oM000000000000000000004RL0=ooo`h0oooo00<0_g@00000000000001000
00000`18W=l0oooo0?ooo`090?ooo`040000003oool0oono07@000<0000000<0M;oo0?ooo`3oool0
4P3oool0103ooml0W4P00018W03Oool>0?ooo`030?oog`2LB00000000080000000<0M;oo0?ooo`3o
ool0403oool01@3ogi`0B000000000000000M;oo00`0oooo000E0?ooo`030=nLB000B9`0gooo0080
oooo00<0_g@0001d_`3oool01`3oool00`2oM000M;oo0?ooo`080?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool0203oool00`3ogi`0B00007Boo`0A0?ooo`030=nLB0000000000000<0
000000<004RL0=ooo`3oool02@3oool01@000000oooo0?oOW0180000B9cO0080oooo00<0_g@004RL
g`3oool04P3oool0103ooml0W4P00018W03Oool>0?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@0
07Boo`3oool03`3oool01`3ogi`0B00004RLg`3oool0oono07@0001d_ol02`3oool001D0oooo00L0
oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00H0omnL04P00000000000000000B02Lgol5
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool02@3oool00`3ogi`0B00007Boo`0B
0?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool02P3oool00`000000oooo0?ooo`03
0?ooo`030?oo_`1d04P0W=oo0180oooo00@0oooO09a80000B9`0gooo4`3oool00`3ooml0W4Qd0;oo
o`0>0?ooo`030?oog`2LB00007Bo00<0oooo00<0okmd0018W03Oool02P3oool001H0oooo00D0oono
07@000000000001d0;ooo`0:0?ooo`050;md00000000000000000018W=l01@3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo00T0oooo00<0omnL04P0001d_ol04`3oool00`3OW4P0M;oo
0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<0oono07@0001d_ol03P3oool00`2oM0000000
000000050000000307Boo`3oool0oooo0140oooo00<0_g@007Boo`3oool03@3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00X0oooo000_0?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool02P3oool00`3ogi`0B00007Boo`0D0?ooo`030;mdB02Lgol0oooo00h0oooo00<0
00000?ooo`3oool00P3oool00`3o_g@0000007Boo`0>0?ooo`080?oOW01dW=l0oooo0?ooo`3ooml0
W4P00018W03OoolC0?ooo`030?oog`2LB000B9cO00h0oooo00<0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`0:0?ooo`00;`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00X0
oooo00<0omnL04P0001d_ol0503oool00`3ooml0W4Qd0;ooo`0?0?ooo`070000003oool0oooo0?oO
W0180000001809cOo`0?0?ooo`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`0C0?ooo`030=nL
B0000000M;oo00h0oooo00<0omnL04P0001d_ol00`3oool00`2oM000001809cOo`0:0?ooo`00;`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00/0oooo00<0omnL04P0001d_ol05@3o
ool00`3o_g@007Bo0?ooo`0>0?ooo`060000003o_g@000000000000004P0W=oo403oool01`3ooml0
W4Qd0;ooo`3ooml0W4P00018W03Oool03`3oool01`3Ogml0oooo0?oo_`1d000000000000M02oool0
3P3oool0203ookl0M00000000018W=l0oooo0?noM000000007Bo2`3oool002l0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0;0?ooo`030?oOW0180000M;oo01D0oooo00<0oono07@0
B02Lgol03`3oool010000000oooo0;md0000M;lB0?ooo`060?oOW018M;l0oooO09a80000B9`0gooo
3`3oool01P3OW4P00000000000000000001809cOoa00oooo00<0_g@00018W02oW4P00P0000000`00
B9`0gooo0?ooo`0:0?ooo`00;`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00`0
oooo00<0omnL04P0001d_ol05@3oool00`3ookl0M00004RLg`0@0?ooo`040000003oool0_g@004RL
ga80oooo00D0gia807Bog`2LB00004RL0=ooo`0?0?ooo`040?oo_`1d0000000004RLga80oooo00<0
omnL04P0001d_ol03`3oool002l0oooo00D000000?ooo`3oool0oooo0000000?0?ooo`030?oOW018
0000M;oo0100oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`0=0?ooo`080?oog`2L
W;l0oooo0000003oool0gia80000M02ooolA0?ooo`050?oog`2LB4P0B4P00018W03Oool0403oool0
0`2oM000M;oo0?ooo`0C0?ooo`030=nLB00004P0W=oo00h0oooo000_0?ooo`050000003oool0oooo
0?ooo`0000003@3oool01@3ooml0W4P0000000000000M;oo0100oooo00P0oooO09a8B02Lgol0oooo
0?oo_`1d0000001d0;ooo`h0oooo00H0okmd04RLg`3oool0000007@00000M;lC0?ooo`040?oOW018
000004RL0=oooa00oooo00H0omnL04P00000000000000000B02LgolA0?ooo`030?noM000000007Bo
00d0oooo000_0?ooo`040000003oool0oooo00000100oooo00<0okmd0000001d_ol04@3oool01P3o
oml0W4P0000000000000001d0;oooa00oooo00D0gia8000000000000001d0;ooo`0D0?ooo`030=nL
B000B9`0gooo0140oooo00D0_g@0000000000000000004RLg`0B0?ooo`050?oog`2LB00000000000
M02oool02P3oool002l0oooo00@000000?ooo`3oool00000?@3oool00`000000oooo0?ooo`1=0?oo
o`00;`3oool00`000000oooo0000000n0?ooo`030000003oool0oooo04d0oooo000_0?ooo`030000
003oool0000003l0oooo00<000000?ooo`3oool0C03oool002l0oooo0P0000100?ooo`030000003o
ool0oooo04`0oooo000_0?ooo`800000@@3oool00`000000oooo0?ooo`1;0?ooo`00;`3oool00`00
0000oooo0?ooo`100?ooo`030000003oool0oooo04/0oooo000[0?oooi400000103oool002l0oooo
00D000000?ooo`3oool0oooo000000040?ooo`050000003oool0oooo0?ooo`0000000`3oool00`00
0000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`02
0?ooo`050000003oool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`020?ooo`050000003o
ool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`00
00000`3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000000`3oool00`00
0000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`02
0?ooo`050000003oool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`020?ooo`050000003o
ool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`00
00000`3oool00`000000oooo000000020?ooo`030000003oool0oooo00<0oooo000_0?ooo`030000
003oool0oooo0480oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`070?ooo`00;`3o
ool00`000000oooo0?ooo`130?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool01`3o
ool002l0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0
oooo00P0oooo000_0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0?03oool00`00
0000oooo0?ooo`080?ooo`00;`3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo03/0
oooo00<000000?ooo`3oool02@3oool002l0oooo00<000000?ooo`3oool0A@3oool00`000000oooo
0?ooo`0j0?ooo`030000003oool0oooo00T0oooo000_0?ooo`030000003oool0oooo04D0oooo00<0
00000?ooo`3oool0>@3oool00`000000oooo0?ooo`0:0?ooo`00;`3oool00`000000oooo0?ooo`16
0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool02P3oool002l0oooo00<000000?oo
o`3oool0AP3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo00/0oooo000_0?ooo`03
0000003oool0oooo04L0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0;0?ooo`00
;`3oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0
303oool002l0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`0c0?ooo`030000003o
ool0oooo00d0oooo000_0?ooo`030000003oool0oooo04P0oooo00<000000?ooo`3oool0<`3oool0
0`000000oooo0?ooo`0=0?ooo`00;`3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo
0340oooo00<000000?ooo`3oool03P3oool002l0oooo00<000000?ooo`3oool0B@3oool00`000000
oooo0?ooo`0a0?ooo`030000003oool0oooo00h0oooo000F0?ooo`050?oOW018000000000000M02o
ool00`3oool01P3o_g@0000007Boo`3oool0oooO09a800<0000000<0M;oo0?ooo`3oool01@3oool0
0`000000oooo0?ooo`1:0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool03`3oool0
01D0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00L0okmd0000001d_ol0oooo
0?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool01@3oool00`000000oooo0?ooo`1:0?ooo`03
0000003oool0oooo02l0oooo00<000000?ooo`3oool03`3oool001D0oooo00<0_g@00018W03Oool0
0P3oool00`2oM00007Bo0?ooo`0:0?ooo`030?oog`2LB7@0_ooo00D0oooo00<000000?ooo`3oool0
B`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo0100oooo000D0?ooo`030?oo_`1d
0000B9cO00<0oooo00<0gia80000M02oool02`3oool00`2oM000M;oo0?ooo`040?ooo`030000003o
ool0oooo04/0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0@0?ooo`00503oool0
0`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0oooO09a80018W=l01@3oool0
0`000000oooo0?ooo`1<0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool04@3oool0
01@0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030=nLB0000000M;oo
00D0oooo0P00001=0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool04@3oool001@0
oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`070=oOg`3oool0oono07@0
00000000001d0;ooo`050?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3oool0:@3oool0
0`000000oooo0?ooo`0B0?ooo`00503oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00H0oooo00H0gia800000000000000000000B02Lgol60?ooo`030000003oool0oooo04d0oooo00<0
00000?ooo`3oool0:@3oool00`000000oooo0?ooo`0B0?ooo`00503oool00`3ookl0M00004RLg`03
0?ooo`030=nLB00007@0_ooo00H0oooo00@0oono07@000000000B9cO203oool00`000000oooo0?oo
o`1>0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool04`3oool001D0oooo00<0gia8
0018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030;md001d_ol0oooo00P0oooo00<00000
0?ooo`3oool0C`3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo01@0oooo000E0?oo
o`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`060?oOW018000000000000000004P0
W=oo1@3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3o
ool0503oool001H0oooo00D0oono07@000000000001d0;ooo`0:0?ooo`050;md0000000000000000
0018W=l01@3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo02<0oooo00<000000?oo
o`3oool05@3oool002l0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`0Q0?ooo`03
0000003oool0oooo01H0oooo000_0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0
803oool00`000000oooo0?ooo`0F0?ooo`00;`3oool00`000000oooo0?ooo`1C0?ooo`030000003o
ool0oooo01h0oooo00<000000?ooo`3oool05`3oool002l0oooo00<000000?ooo`3oool0D`3oool0
0`000000oooo0?ooo`0M0?ooo`030000003oool0oooo01P0oooo000_0?ooo`030000003oool0oooo
05@0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0I0?ooo`00;`3oool00`000000
oooo0?ooo`1E0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool002l0oooo
00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01/0oooo
000_0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?oo
o`0L0?ooo`00;`3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo01<0oooo00<00000
0?ooo`3oool07@3oool002<0oooo00H0_g@000000000000000000018W03Oool60?ooo`030000003o
ool0oooo05T0oooo0P00000B0?ooo`030000003oool0oooo01h0oooo000T0?ooo`030?oOW0180000
M;oo00P0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo01l0oooo000T0?ooo`030?oOW0180000M;oo00P0oooo00<000000?ooo`3oool0G03oool00`00
0000oooo0?ooo`0;0?ooo`8000008P3oool002@0oooo00<0omnL04P0001d_ol0203oool00`000000
oooo0?ooo`1M0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool08P3oool002@0oooo
00<0omnL04P0001d_ol0203oool00`000000oooo0?ooo`1N0?ooo`D000000P3oool4000002D0oooo
000T0?ooo`030?oOW0180000M;oo00P0oooo0P00001S0?ooo`<00000:@3oool002@0oooo00<0omnL
04P0001d_ol0203oool00`000000oooo0?ooo`2>0?ooo`00903oool00`3ogi`0B00007Boo`080?oo
o`030000003oool0oooo08h0oooo000T0?ooo`030?oOW0180000M;oo09T0oooo000T0?ooo`030?oO
W0180000M;oo09T0oooo000R0?ooo`050?oog`2LB00000000000001d_ol0V@3oool002@0oooo00<0
okmd0000001d_ol0V@3oool00<00oooo00300?ooo`00`03oool00<00oooo00300?ooo`00`03oool0
0<00oooo00300?ooo`00`03oool00<00oooo00300?ooo`00`03oool00<00oooo00300?ooo`00`03o
ool00<00oooo00300?ooo`00`03oool00<00oooo00300?ooo`00`03oool00<00oooo00300?ooo`00
`03oool00<00oooo00300?ooo`00`03oool002<0oooo00H0_g@000000000000000000018W03OoonG
0?ooo`00903oool00`3ogi`0B00007Boo`2I0?ooo`00903oool00`3ogi`0B00007Boo`2I0?ooo`00
903oool00`3ogi`0B00007Boo`080?ooo`030000003oool0oooo08h0oooo000E0?ooo`030?oog`2L
B000000000L0000000<0B9cO0?ooo`3oool00P3oool00`3ogi`0B00007Boo`080?ooo`030000003o
ool0oooo08h0oooo000T0?ooo`030?oOW0180000M;oo00P0oooo0P0000110?ooo`<00000B`3oool0
02@0oooo00<0omnL04P0001d_ol0203oool00`000000oooo0?ooo`0l0?ooo`D000000P3oool40000
04L0oooo000T0?ooo`030?oOW0180000M;oo00P0oooo00<000000?ooo`3oool0>P3oool3000000X0
oooo00<000000?ooo`3oool0A03oool002@0oooo00<0omnL04P0001d_ol0203oool00`000000oooo
0?ooo`0i0?ooo`030000003oool0oooo00`0oooo0P0000140?ooo`00903oool00`3ogi`0B00007Bo
o`080?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`110?ooo`008P3oool01@3ooml0W4P0000000000000M;oo00P0oooo00<000000?ooo`3oool0=`3o
ool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0400oooo000T0?ooo`030?noM0000000
M;oo00P0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0
oooo03l0oooo000_0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`0n0?ooo`00;`3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo01L0
oooo00<000000?ooo`3oool0?@3oool002l0oooo00<000000?ooo`3oool0<`3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo03`0oooo000_0?ooo`030000003oool0oooo0380oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`0l0?ooo`00;`3oool00`000000oooo0?ooo`0a
0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0>`3oool002l0oooo00<000000?oo
o`3oool0<03oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo03X0oooo000_0?ooo`03
0000003oool0oooo02l0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0i0?ooo`00
;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0
>03oool002l0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0R0?ooo`030000003o
ool0oooo03P0oooo000_0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool08`3oool0
0`000000oooo0?ooo`0g0?ooo`005P3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00H0okmd
0000001d_ol0oooo0?oog`2LB0030000000307Boo`3oool0oooo00D0oooo00<000000?ooo`3oool0
;@3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo03H0oooo000E0?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo0?ooo`3ooml0W4Qd0;ooo`02
0?ooo`030;md001d_ol0oooo00D0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0V
0?ooo`030000003oool0oooo03D0oooo000E0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d
_`3oool02P3oool00`3ooml0W4Qd0;ooo`050?ooo`030000003oool0oooo02`0oooo00<000000?oo
o`3oool09`3oool00`000000oooo0?ooo`0e0?ooo`00503oool00`3ookl0M00004RLg`030?ooo`03
0=nLB00007@0_ooo00/0oooo00<0_g@007Boo`3oool0103oool00`000000oooo0?ooo`0/0?ooo`03
0000003oool0oooo02P0oooo00<000000?ooo`3oool0=03oool000L0oooo00<0oooO09a800000000
1`0000001P18W=l0oooo0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol02P3oool00`3o
oml0W4P004RLg`050?ooo`800000;03oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo
03<0oooo000D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3OW4P0
000007Boo`050?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0;03oool00`000000
oooo0?ooo`0b0?ooo`00503oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo
00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00D0oooo00<000000?ooo`3oool0:P3oool00`00
0000oooo0?ooo`0/0?ooo`030000003oool0oooo0380oooo000D0?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol01P3oool01P3OW4P00000000000000000001809cOo`H0oooo00<00000
0?ooo`3oool0403oool01P2oM000000000000000000004RL0=ooo`h0oooo00<0_g@0000000000000
100000000`18W=l0oooo0?ooo`0;0?ooo`030?oo_`1d000000000080000000<0M;oo0?ooo`3oool0
4P3oool0103ooml0W4P00018W03Oool30?ooo`030000003oool0oooo00P0oooo00<0oooO09a80000
00000P0000000`1d_ol0oooo0?ooo`0@0?ooo`050?oOW018000000000000001d_ol0303oool001@0
oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d0000000004RL
g`P0oooo00<000000?ooo`3oool04@3oool00`3ogi`0B00007Boo`0A0?ooo`030=nLB00000000000
00<0000000<004RL0=ooo`3oool02`3oool00`3ogi`0B00004RLg`020?ooo`030;md0018W=l0oooo
0180oooo00@0oooO09a80000B9`0gooo0`3oool00`000000oooo0?ooo`080?ooo`030?oog`2LB7@0
_ooo0080oooo00<0_g@007Boo`3oool03`3oool01`3ogi`0B00004RLg`3oool0oono07@0001d_ol0
2`3oool001D0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030;md001d
_ol0oooo00P0oooo00<000000?ooo`3oool04@3oool00`3ogi`0B00007Boo`0B0?ooo`030=nLB018
W=l000000080oooo00<0giad09cOo`3oool0403oool00`3ookl0M01809cOo`0B0?ooo`040?oog`2L
B00004RL0=ooo`@0oooo00<000000?ooo`3oool0303oool00`3ooml0W4Qd0;ooo`0>0?ooo`030?oo
g`2LB00007Bo00<0oooo00<0okmd0018W03Oool02P3oool001D0oooo00L0oono07@0001d_ol0oooo
0?noM00007@0_ooo00P0oooo00H0omnL04P00000000000000000B02Lgol50?ooo`030000003oool0
oooo0140oooo00<0omnL04P0001d_ol04`3oool00`000000M;oo0?ooo`0D0?ooo`030?oo_`1d0000
M;oo00h0oooo00<0_g@00000000000001@0000001@1d_ol0oooo0?ooo`3oool0000000l0oooo00<0
_g@007Boo`3oool03@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo000F
0?ooo`050?oo_`1d000000000000M02oool02P3oool01@2oM0000000000000000000B9cO00D0oooo
00<000000?ooo`3oool04@3oool00`3ogi`0B00007Boo`0C0?ooo`030000002oM4P0W=oo01@0oooo
00<0okmd0000001d_ol03P3oool0203ogi`0M9cO0?ooo`3oool0oooO09a80000B9`0gooo1@3oool0
0`000000oooo0?ooo`0;0?ooo`030?oog`2LB000B9cO00h0oooo00<0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`0:0?ooo`00;`3oool00`000000oooo0?ooo`0A0?ooo`030?oOW0180000M;oo
0180oooo00D000000?ooo`3ooml0W4Qd0;ooo`0B0?ooo`040?oOW0180000001809cOo`l0oooo00L0
gia807Boo`3oool0oooO09a80000B9`0gooo00D0oooo00<000000?ooo`3oool02`3oool00`3OW4P0
000007Boo`0>0?ooo`030?oOW0180000M;oo00<0oooo00<0_g@00000B02Lgol02P3oool002l0oooo
00<000000?ooo`3oool04@3oool00`3ogi`0B00007Boo`0B0?ooo`050000003oool0oooo0?noM000
M;l0403oool01@3o_g@000000000000004P0W=oo0100oooo00L0oooO09a8M02oool0oooO09a80000
B9`0gooo00H0oooo00<000000?ooo`3oool01P3oool01`3Ogml0oooo0?oo_`1d000000000000M02o
ool03P3oool0203ookl0M00000000018W=l0oooo0?noM000000007Bo2`3oool002l0oooo00<00000
0?ooo`3oool04@3oool00`3ogi`0B00007Boo`0A0?ooo`070000003oool0oooo0?ooo`3ookl0M018
09cOo`0A0?ooo`030;md0000M;l0oooo0140oooo00H0omnL04Qd_`3ooml0W4P00018W03Oool60?oo
o`030000003oool0oooo00H0oooo00H0gia800000000000000000000B02Lgol@0?ooo`030;md0000
B9`0_ia80080000000<004RL0=ooo`3oool02P3oool002l0oooo00<000000?ooo`3oool04@3oool0
0`3ogi`0B00007Boo`0@0?ooo`030000003oool0oooo0080oooo00<0oono07@00018W=l04P3oool0
0`2oM000B9cO0?ooo`0A0?ooo`050=nLB01d_ml0W4P00018W03Oool01`3oool00`000000oooo0?oo
o`050?ooo`040?oo_`1d0000000004RLga80oooo00<0omnL04P0001d_ol03`3oool002l0oooo00<0
00000?ooo`3oool04@3oool00`3ogi`0B00007Boo`0@0?ooo`030000002Lgol0oooo0080oooo00<0
omnL04P00018W=l03@3oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo0140oooo00D0
oooO09a8B018B00004RL0=ooo`070?ooo`030000003oool0oooo00H0oooo00<0_g@007Boo`3oool0
4`3oool00`3OW4P0001809cOo`0>0?ooo`00;`3oool00`000000oooo0?ooo`0?0?ooo`050?oog`2L
B00000000000001d_ol03`3oool02@000000oooO09a8B02Lgol0oooo0?oo_`1d0000001d0;ooo`0>
0?ooo`060?noM018W=l0oooo0?oo_`1d000007Bo4`3oool0103ogi`0B0000018W03Oool80?ooo`03
0000003oool0oooo00D0oooo00H0omnL04P00000000000000000B02LgolA0?ooo`030?noM0000000
07Bo00d0oooo000_0?ooo`030000003oool0oooo0140oooo00<0okmd0000001d_ol03`3oool01000
0000oooo0?oog`2LB002000000030000M02oool0oooo00l0oooo00D0gia8000000000000001d0;oo
o`0D0?ooo`030=nLB000B9`0gooo00P0oooo00<000000?ooo`3oool01P3oool01@2oM00000000000
00000000B9cO0180oooo00D0oooO09a800000000001d0;ooo`0:0?ooo`00;`3oool00`000000oooo
0?ooo`0R0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0:`3oool002l0oooo00<0
00000?ooo`3oool08P3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo02/0oooo000_
0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`0Z
0?ooo`00;`3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo03d0oooo00<000000?oo
o`3oool0:P3oool002l0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0o0?ooo`03
0000003oool0oooo02T0oooo000_0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0
?`3oool00`000000oooo0?ooo`0Y0?ooo`00;`3oool00`000000oooo0?ooo`0O0?ooo`030000003o
ool0oooo0440oooo00<000000?ooo`3oool0:03oool002/0ooooT@0000040?ooo`00;`3oool00`00
0000oooo0?ooo`0N0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool09`3oool002l0
oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo02L0
oooo000_0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0A@3oool00`000000oooo
0?ooo`0V0?ooo`00;`3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo04H0oooo00<0
00000?ooo`3oool09P3oool002l0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`17
0?ooo`030000003oool0oooo02D0oooo000_0?ooo`030000003oool0oooo01/0oooo00<000000?oo
o`3oool0B@3oool00`000000oooo0?ooo`0T0?ooo`00;`3oool00`000000oooo0?ooo`0K0?ooo`03
0000003oool0oooo04T0oooo00<000000?ooo`3oool0903oool002l0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo02<0oooo000_0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`0S0?ooo`00;`3oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3oool08P3oool0
02l0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo
0280oooo000_0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0C`3oool00`000000
oooo0?ooo`0Q0?ooo`00;`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo04l0oooo
00<000000?ooo`3oool08@3oool002l0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`1A0?ooo`030000003oool0oooo0200oooo000_0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool0D@3oool00`000000oooo0?ooo`0P0?ooo`005P3oool01@3ogi`0B0000000000007@0
_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oog`2LB0030000000307Boo`3oool0oooo00D0oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo01l0oooo
000E0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo0?oo
o`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00D0oooo00<000000?ooo`3oool05P3oool0
0`000000oooo0?ooo`1C0?ooo`030000003oool0oooo01l0oooo000E0?ooo`030;md0000B9`0gooo
0080oooo00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd0;ooo`050?ooo`030000003oool0oooo
01D0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`0N0?ooo`00503oool00`3ookl0
M00004RLg`030?ooo`030=nLB00007@0_ooo00/0oooo00<0_g@007Boo`3oool0103oool00`000000
oooo0?ooo`0E0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool07P3oool001@0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030?oog`2LB000B9cO00D0oooo
00<000000?ooo`3oool0503oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo01d0oooo
000D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3OW4P0000007Bo
o`050?ooo`8000005@3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo01d0oooo000D
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01`3Ogml0oooo0?oo_`1d
000000000000M02oool01@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo05T0oooo
00<000000?ooo`3oool0703oool001@0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`060?ooo`060=nLB000000000000000000004P0W=oo1P3oool00`000000oooo0?ooo`0B0?ooo`03
0000003oool0oooo05X0oooo00<000000?ooo`3oool0703oool001@0oooo00<0oono07@00018W=l0
0`3oool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d0000000004RLg`P0oooo00<000000?ooo`3o
ool04P3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo01/0oooo000E0?ooo`030=nL
B000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`2oM000M;oo0?ooo`080?ooo`030000
003oool0oooo0140oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`0J0?ooo`005@3o
ool01`3ookl0M00007Boo`3oool0okmd0000M02oool0203oool01P3ogi`0B0000000000000000018
09cOo`D0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1O0?ooo`030000003oool0
oooo01T0oooo000F0?ooo`050?oo_`1d000000000000M02oool02P3oool01@2oM000000000000000
0000B9cO00D0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1P0?ooo`030000003o
ool0oooo01T0oooo000_0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0H@3oool0
0`000000oooo0?ooo`0H0?ooo`00;`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool05`3oool002l0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`1U0?ooo`030000003oool0oooo01H0oooo000_0?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`0F0?ooo`00;`3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool05@3oool002l0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo01@0oooo000_0?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool0J`3oool00`000000oooo0?ooo`0C0?oo
o`00;`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3o
ool04P3oool002l0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`1_0?ooo`030000
003oool0oooo0140oooo000S0?ooo`060;md00000000000000000000B9`0gooo1P3oool00`000000
oooo0?ooo`070?ooo`030000003oool0oooo0740oooo00<000000?ooo`3oool0403oool002@0oooo
00<0omnL04P0001d_ol0203oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo07<0oooo
00<000000?ooo`3oool03`3oool002@0oooo00<0omnL04P0001d_ol0203oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo07D0oooo00<000000?ooo`3oool03P3oool002@0oooo00<0omnL
04P0001d_ol0203oool00`000000oooo0?ooo`020?ooo`<00000N@3oool3000000d0oooo000T0?oo
o`030?oOW0180000M;oo00P0oooo00<000000?ooo`3oool00P00001o0?ooo`<000002P3oool002@0
oooo00<0omnL04P0001d_ol0203oool3000008<0oooo0`0000080?ooo`00903oool00`3ogi`0B000
07Boo`080?ooo`030000003oool0oooo08h0oooo000T0?ooo`030?oOW0180000M;oo00P0oooo00<0
00000?ooo`3oool0SP3oool002@0oooo00<0omnL04P0001d_ol0V@3oool002@0oooo00<0omnL04P0
001d_ol0V@3oool00280oooo00D0oooO09a800000000000007Boo`2I0?ooo`00903oool00`3o_g@0
000007Boo`2I0?ooo`00`03oool00<00oooo00300?ooo`00`03oool00<00oooo00300?ooo`00`03o
ool00<00oooo00300?ooo`00`03oool00<00oooo00300?ooo`00`03oool00<00oooo00300?ooo`00
`03oool00<00oooo0000\
\>"],
  ImageRangeCache->{{{0, 191.75}, {380.188, 0}} -> {-0.0250052, -0.0520859, \
0.0082139, 0.0082139}, {{4.5625, 187.125}, {370.625, 257.813}} -> {-2.17998, \
-1.56517, 0.0691005, 0.0355891}, {{4.5625, 187.125}, {246.5, 133.625}} -> \
{-2.17998, -4.51093, 0.0691005, 0.0355891}, {{4.5625, 187.125}, {122.375, \
9.5}} -> {-2.17998, -7.45592, 0.0691005, 0.0355891}}]
}, Closed]],

Cell[TextData[{
  "One can also use ",
  StyleBox["DisplayTogetherArray", "InlineInput",
    FontWeight->"Bold"],
  " instead of ",
  StyleBox["GraphicsArray", "InlineInput",
    FontWeight->"Bold"],
  " to avoid manipulating ",
  StyleBox["DisplayFunction", "InlineInput",
    FontWeight->"Bold"],
  ".  This function is found in ",
  StyleBox["Graphics`Graphics`", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell["\<\
Reproduce one of the arrays above using DisplayTogetherArray.\
\>", "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Creating plotting functions", "Subsection"],

Cell["\<\
Often we will wish to use similar commands to plot a function for several \
choices of its parameters, perhaps over different ranges, or with other minor \
variations of the options.  Rather than reproduce the necessary command with \
cut-and-paste and edit the copies, it is better to define commands which \
perform plots on demand, given the necessary parameters.  \
\>", "Text"],

Cell[TextData[{
  "For example, one can show that modified Bessel functions have the \
asymptotic behavior ",
  Cell[BoxData[
      \(TraditionalForm\`I\_n[z]~\(Exp[z]\/\@\(2  \[Pi]\ z\)\) \((1\  + \ 
            O(\ 1\/z))\)\)]],
  ".  The commands below define and illustrate a function that can be used to \
display  the approach to asymptopia.  The arguments of ",
  StyleBox["plotAsymptoticBesselI", "InlineInput",
    FontWeight->"Bold"],
  " are the order, the minimum and maximum arguments to display, and a set of \
options.  Note that the triple underscore indicates that zero or more \
arguments collectively named ",
  StyleBox["opts", "InlineInput"],
  " may be included; these options are expected to be rules.  Of course, we \
can omit optional parameters altogether."
}], "Text"],

Cell[BoxData[
    \(plotAsymptoticBesselI[n_, xmin_, xmax_, opts___Rule]\  := 
      Plot[Evaluate[\@\(2  \[Pi]\ x\)\ Exp[\(-x\)] BesselI[n, x]], {x, xmin, 
          xmax}, opts]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plotAsymptoticBesselI[2, 0, 50];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0190476 0.0147151 0.611326 [
[.21429 .00222 -8 -15.6875 ]
[.21429 .00222 8 0 ]
[.40476 .00222 -8 -15.6875 ]
[.40476 .00222 8 0 ]
[.59524 .00222 -8 -15.6875 ]
[.59524 .00222 8 0 ]
[.78571 .00222 -8 -15.6875 ]
[.78571 .00222 8 0 ]
[.97619 .00222 -8 -15.6875 ]
[.97619 .00222 8 0 ]
[.01131 .13698 -19 -7.84375 ]
[.01131 .13698 0 7.84375 ]
[.01131 .25925 -19 -7.84375 ]
[.01131 .25925 0 7.84375 ]
[.01131 .38151 -19 -7.84375 ]
[.01131 .38151 0 7.84375 ]
[.01131 .50378 -19 -7.84375 ]
[.01131 .50378 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .01472 m
.21429 .02097 L
s
gsave
.21429 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.40476 .01472 m
.40476 .02097 L
s
gsave
.40476 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.59524 .01472 m
.59524 .02097 L
s
gsave
.59524 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(30) show
1.000 setlinewidth
grestore
.78571 .01472 m
.78571 .02097 L
s
gsave
.78571 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
.97619 .01472 m
.97619 .02097 L
s
gsave
.97619 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(50) show
1.000 setlinewidth
grestore
.125 Mabswid
.0619 .01472 m
.0619 .01847 L
s
.1 .01472 m
.1 .01847 L
s
.1381 .01472 m
.1381 .01847 L
s
.17619 .01472 m
.17619 .01847 L
s
.25238 .01472 m
.25238 .01847 L
s
.29048 .01472 m
.29048 .01847 L
s
.32857 .01472 m
.32857 .01847 L
s
.36667 .01472 m
.36667 .01847 L
s
.44286 .01472 m
.44286 .01847 L
s
.48095 .01472 m
.48095 .01847 L
s
.51905 .01472 m
.51905 .01847 L
s
.55714 .01472 m
.55714 .01847 L
s
.63333 .01472 m
.63333 .01847 L
s
.67143 .01472 m
.67143 .01847 L
s
.70952 .01472 m
.70952 .01847 L
s
.74762 .01472 m
.74762 .01847 L
s
.82381 .01472 m
.82381 .01847 L
s
.8619 .01472 m
.8619 .01847 L
s
.9 .01472 m
.9 .01847 L
s
.9381 .01472 m
.9381 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .13698 m
.03006 .13698 L
s
gsave
.01131 .13698 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.02381 .25925 m
.03006 .25925 L
s
gsave
.01131 .25925 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.02381 .38151 m
.03006 .38151 L
s
gsave
.01131 .38151 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.02381 .50378 m
.03006 .50378 L
s
gsave
.01131 .50378 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04528 m
.02756 .04528 L
s
.02381 .07585 m
.02756 .07585 L
s
.02381 .10641 m
.02756 .10641 L
s
.02381 .16755 m
.02756 .16755 L
s
.02381 .19811 m
.02756 .19811 L
s
.02381 .22868 m
.02756 .22868 L
s
.02381 .28981 m
.02756 .28981 L
s
.02381 .32038 m
.02756 .32038 L
s
.02381 .35094 m
.02756 .35094 L
s
.02381 .41208 m
.02756 .41208 L
s
.02381 .44264 m
.02756 .44264 L
s
.02381 .47321 m
.02756 .47321 L
s
.02381 .53434 m
.02756 .53434 L
s
.02381 .56491 m
.02756 .56491 L
s
.02381 .59548 m
.02756 .59548 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01472 m
.06244 .21996 L
.08255 .3178 L
.09396 .35762 L
.10458 .3872 L
.1237 .42742 L
.14415 .4581 L
.16546 .48151 L
.1856 .4984 L
.22401 .52171 L
.24385 .53071 L
.26488 .53871 L
.30271 .55016 L
.343 .55947 L
.38177 .5665 L
.423 .57251 L
.46271 .57726 L
.5009 .58109 L
.54154 .58456 L
.58067 .58743 L
.62226 .59007 L
.66232 .5923 L
.70087 .5942 L
.74188 .59599 L
.78136 .59754 L
.81934 .59889 L
.85976 .60019 L
.89867 .60133 L
.94003 .60243 L
.97619 .60332 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00JP3oool01P2oM000000000000000000004RL0=ooo`<0oooo00D0omnL04P000000000001d
0;ooo`0k0?ooo`030;md00000000000000@0000000T0B9cO0?ooo`3oool0oooo0?oOW01800000000
0000M02oool0>`3oool00`3ookl0M000000000020000000307Boo`3oool0oooo00<0oooo00D0omnL
04P000000000001d0;ooo`0o0?ooo`040?oog`2LB00004RL0=ooo`<0oooo00D0omnL04P000000000
001d0;ooo`0k0?ooo`030?oog`2LB00000000080000000<0M;oo0?ooo`3oool00`3oool01@3ogi`0
B0000000000007@0_ooo00`0oooo001[0?ooo`030?oOW0180000M;oo00@0oooo00L0oono07@0001d
_ol0oooo0?noM00007@0_ooo03/0oooo00<0gia80000000000000`0000002P00B9`0gooo0?ooo`3o
okl0M00007Boo`3oool0okmd0000M02ooolj0?ooo`030?oOW0180000B9cO0080oooo00<0_g@004RL
g`3oool00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0?P3oool0103ooml0W4P00018
W03Oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0j0?ooo`030?oog`2LB7@0_ooo
0080oooo00<0_g@007Boo`3oool00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool02`3o
ool006/0oooo00<0omnL04P0001d_ol0103oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0
oooo03/0oooo00<0gia804RLg`3oool00P3oool01P3OW7@0W=oo0?ooo`2oM00004RL0=ooo`80oooo
00<0_g@0001d_`3oool0?P3oool00`3ookl0M01809cOo`020?ooo`030;md0000B9`0gooo0080oooo
00<0_g@0001d_`3oool0?@3oool0103ooml0W4P00018W03Oool20?ooo`030;md0000B9`0gooo0080
oooo00<0_g@0001d_`3oool0?P3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md0000B9`0gooo0080
oooo00<0_g@0001d_`3oool02P3oool006/0oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M000
04RLg`030?ooo`030=nLB00007@0_ooo03`0oooo00<0gia807Boo`3oool00`3oool00`3ookl0M000
04RLg`030?ooo`030=nLB00007@0_ooo03h0oooo00L0oono07@0001d_ol0oooo0?oo_`1d0000B9cO
00<0oooo00<0gia80000M02oool0>@3oool00`2oM0000000000000050000000407Boo`3ookl0M000
04RLg`<0oooo00<0gia80000M02oool0?`3oool01P2oM000M;oo0?ooo`3ookl0M00004RLg`<0oooo
00<0gia80000M02oool02P3oool006/0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo03d0oooo00<0_gA809cOo`3oool00P3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo03h0oooo00L0okmd0000001d_ol0oooo0?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol0>@3oool0303ogi`0M9cO0?ooo`3oool0oooO09a80000B9`0gooo0?oo
o`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol0?P3oool01`3ooml0W4P004RLg`3oool0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`00J`3oool00`3ogi`0B00007Boo`030?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0?@3oool00`3ooml0W4Qd0;ooo`020?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0?@3oool0203ogi`0B0000000B02Lgol0
oooo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0j0?ooo`0;0=nLB01d_ol0oooo0?oog`2L
B00004RL0=ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0n0?ooo`070=nLB000
0000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo001[0?ooo`030?oO
W0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0n0?ooo`070?no
M000M;l0oooo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo03/0oooo00D0okmd0000
00000000001809cOo`020?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0>P3oool0
2`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol0>P3oool02`3Ogml0oooo0?oo_`1d000000000000M02oool0oooo0?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol02P3oool006/0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo03h0oooo00L0oono07@0B02Lgol0oooo0?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol0?@3oool00`2oM00007Bo0?ooo`020?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol0>`3oool02P3ogi`0B7Bo0?oog`2LB00004RL0=ooo`3oool0omnL
04P0001d_ol30?ooo`030?noM00004P0W=oo03X0oooo00H0gia800000000000000000000B02Lgol2
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool006/0oooo00<0omnL04P0
001d_ol00`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo03h0oooo00L0oono07@0
0018W=l0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0?P3oool01`2oM000B9cO0?oo
o`3oool0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0l0?ooo`090=nLB01d_ml0W4P00018
W03Oool0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0>P3oool0103ookl0M0000000
0018W=l40?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02P3oool006/0oooo00<0
omnL04P0001d_ol0103oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo03T0oooo00<0
oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`020?ooo`030=nLB000B9`0gooo0080oooo00<0
_g@0001d_`3oool0>@3oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo0080oooo00<0
gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`0l0?ooo`050?oog`2LB4P0B4P00018W03Oool0
0P3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo03/0oooo00<0_g@007Boo`3oool0
1@3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo001Y0?ooo`050?oog`2L
B00000000000001d_ol0103oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0>P3oool0203o
oml0W4Q809cOo`3oool0oono07@0000007@0_ooo0P3oool01`3ookl0M00007Boo`3oool0okmd0000
M02oool0>`3oool01P3o_g@0B9cO0?ooo`3ookl0M000001d_`<0oooo00L0oono07@0001d_ol0oooo
0?noM00007@0_ooo03h0oooo00@0omnL04P00000B9`0gooo0P3oool01`3ookl0M00007Boo`3oool0
okmd0000M02oool0?03oool01P3ogi`0B000000000000000001809cOo`80oooo00L0oono07@0001d
_ol0oooo0?noM00007@0_ooo00/0oooo001[0?ooo`030?noM0000000M;oo00D0oooo00D0oono07@0
00000000001d0;ooo`0l0?ooo`060?oog`2LB00000000000000007@0_ooo103oool01@3ookl0M000
0000000007@0_ooo03d0oooo00D0gia8000000000000001d0;ooo`040?ooo`050?oo_`1d00000000
0000M02oool0@03oool00`3OW4P004RL0=ooo`030?ooo`050?oo_`1d000000000000M02oool0?P3o
ool01@2oM0000000000000000000B9cO00<0oooo00D0oono07@000000000001d0;ooo`0<0?ooo`00
o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`009P3oool00`000000
oooo0?ooo`3o0?ooohP0oooo000V0?ooo`030000003oool0oooo0?l0ooooR03oool002H0oooo00<0
00000?ooo`3oool0o`3ooon80?ooo`009P3oool00`000000oooo0?ooo`3o0?ooohP0oooo000V0?oo
o`030000003oool0oooo0?l0ooooR03oool002H0oooo00<000000?ooo`3oool0o`3ooon80?ooo`00
7@3ooooo000008/000002@3oool002H0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0@0?ooo`009P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0100oooo000V0?oo
o`800000B@3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo04P0oooo00<000000?oo
o`3oool0B03oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo0100oooo000V0?ooo`80
0000o`3ooon90?ooo`009P3oool200000?l0ooooR@3oool002H0oooo0P00003o0?ooohT0oooo000V
0?ooo`800000o`3ooon90?ooo`009P3oool00`000000oooo0000003o0?ooohP0oooo000V0?ooo`03
0000003oool000000?l0ooooR03oool002H0oooo00<000000?ooo`000000o`3ooon80?ooo`009P3o
ool00`000000oooo0000003o0?ooohP0oooo000V0?ooo`<00000o`3ooon80?ooo`009P3oool01000
0000oooo0?ooo`00003o0?ooohL0oooo000V0?ooo`040000003oool0oooo00000?l0ooooQ`3oool0
02H0oooo00@000000?ooo`3oool00000o`3ooon70?ooo`009P3oool010000000oooo0?ooo`00003o
0?ooohL0oooo000V0?ooo`040000003oool0oooo00000?l0ooooQ`3oool002H0oooo00D000000?oo
o`3oool0oooo0000003o0?ooohH0oooo000V0?ooo`050000003oool0oooo0?ooo`000000o`3ooon6
0?ooo`009P3oool01@000000oooo0?ooo`3oool000000?l0ooooQP3oool002H0oooo00D000000?oo
o`3oool0oooo0000003o0?ooohH0oooo000V0?ooo`050000003oool0oooo0?ooo`000000o`3ooon6
0?ooo`009P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0?l0ooooP`3oool002H0
oooo0`0000020?ooo`030000003oool0oooo0?l0ooooP`3oool002H0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`3o0?oooh<0oooo000V0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0o`3ooon30?ooo`009P3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo0?l0ooooP`3oool002H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`3o0?oo
oh80oooo000V0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00
9P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0?l0ooooPP3oool002H0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`3o0?oooh80oooo000V0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool0o`3ooon20?ooo`009P3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo0?l0ooooP@3oool002H0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`3o0?oooh40oooo000V0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
o`3ooon10?ooo`009P3oool3000000@0oooo00<000000?ooo`3oool0o`3ooon10?ooo`009P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0?l0ooooP@3oool002H0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`3o0?oooh00oooo000V0?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool0o`3ooon00?ooo`009P3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo0?l0ooooP03oool002H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`3o0?oooh00oooo000V0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0o`3ooon0
0?ooo`002P3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00D0okmd0000001d_ol0oooo0;md
00060000000304RLg`3oool0oooo00H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`3o0?oooh00oooo00090?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?no
M0000000M;oo0080oooo00<0gia80000000000000`0000000`00B9`0gooo0?ooo`060?ooo`030000
003oool0oooo00H0oooo00<000000?ooo`3oool0o`3ooomo0?ooo`002@3oool00`2oM00004RL0=oo
o`020?ooo`030;md0000M;l0oooo00L0oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?oo
o`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0o`3ooomo0?ooo`00203oool0
0`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00P0oooo00<0gia807Boo`3oool02P3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo0?l0ooooO`3oool000P0oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`090?ooo`030;mdB02Lgol0oooo00T0oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`3o0?ooogl0oooo00080?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol02@3oool00`3ooml0W4Qd0;ooo`090?ooo`@000001@3oool00`000000
oooo0?ooo`3o0?ooogl0oooo00080?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
2P3oool00`3o_g@007Bo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
o`3ooomn0?ooo`00203oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0
oono07@0B02Lgol0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0?l0ooooOP3o
ool000P0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0:0?ooo`030?oo_`1d0000
B9cO00P0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`3o0?ooogh0oooo00090?oo
o`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01@3oool00`3o_i`0W=oo0?ooo`020?oo
o`030?oOW0180000B9cO00P0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`3o0?oo
ogh0oooo00090?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`060?ooo`080?oog`2LB4P0
W=oo0?ooo`3ookl0M0000000M02oool80?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool0o`3ooomn0?ooo`002P3oool01@3ookl0M0000000000007@0_ooo00P0oooo00H0oooO09a80000
000000000000M02oool90?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0o`3ooomm
0?ooo`009P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo0?l0ooooO@3oool002H0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`3o0?ooogd0oooo000V0?ooo`030000
003oool0oooo00P0oooo00<000000?ooo`3oool0o`3ooomm0?ooo`009P3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo0?l0ooooO@3oool002H0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`3o0?ooog`0oooo000V0?ooo`<000002@3oool00`000000oooo0?ooo`3o0?oo
og`0oooo000V0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0o`3oooml0?ooo`00
9P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0?l0ooooO03oool002H0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`3o0?ooog`0oooo000V0?ooo`030000003oool0
oooo00X0oooo00<000000?ooo`3oool0o`3ooomk0?ooo`009P3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo0?l0ooooN`3oool002H0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`3o0?ooog/0oooo000V0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
o`3ooomk0?ooo`009P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0?l0ooooN`3o
ool002H0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`3o0?ooogX0oooo000V0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool0o`3ooomj0?ooo`009P3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo0?l0ooooNP3oool002H0oooo0`00000;0?ooo`030000
003oool0oooo0?l0ooooNP3oool002H0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`3o0?ooogX0oooo000V0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0o`3ooomi
0?ooo`009P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0?l0ooooN@3oool002H0
oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`3o0?ooogT0oooo000V0?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool0o`3ooomi0?ooo`009P3oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo0?l0ooooN@3oool002H0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`3o0?ooogP0oooo000V0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool0o`3ooomh0?ooo`009P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0?l0
ooooN03oool002H0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`3o0?ooogP0oooo
000V0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0o`3ooomh0?ooo`009P3oool3
000000h0oooo00<000000?ooo`3oool0o`3ooomg0?ooo`009P3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo0?l0ooooM`3oool002H0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`3o0?ooogL0oooo000V0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
o`3ooomg0?ooo`009P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0?l0ooooM`3o
ool002H0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`3o0?ooogL0oooo000V0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool0o`3ooomf0?ooo`002P3oool01@3ogi`0
B0000000000007@0_ooo00<0oooo00<0okmd0000001d_ol01@3oool0103ooml0W4P00018W03Oool8
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0o`3ooomf0?ooo`002@3oool01`3o
okl0M00007Boo`3oool0okmd0000M02oool00P3oool00`3o_g@0000007Boo`050?ooo`040?oog`2L
B00004RL0=ooo`P0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`3o0?ooogH0oooo
00090?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02@3oool0103ooml0W4P00018
W03Oool80?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0o`3ooomf0?ooo`00203o
ool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00D0oooo00<0_g@00000000000001@00
00000`1d_ol0oooo0?ooo`050?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0o`3o
oomf0?ooo`00203oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00D0oooo00P0omnL
07BLg`3oool0oooo0?oog`2LB00004RL0=ooo`P0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`3o0?ooogD0oooo00080?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
1P3oool01`3OW4P0M;oo0?ooo`3ooml0W4P00018W03Oool0203oool4000000l0oooo00<000000?oo
o`3oool0o`3ooome0?ooo`00203oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0
oooo00L0oooO09a8M02oool0oooO09a80000B9`0gooo00P0oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`3o0?ooogD0oooo00080?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol01`3oool01P3ogi`0B7Bo0?oog`2LB00004RL0=ooo`P0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`3o0?ooogD0oooo00080?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia8
0000M02oool0203oool01@3OW4P0M;oO09a80000B9`0gooo00P0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`3o0?ooogD0oooo00090?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0
001d_`3oool0203oool01@3ooml0W4Q804Q80000B9`0gooo00P0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`3o0?ooogD0oooo00090?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d
0;ooo`0:0?ooo`040?oOW018000004RL0=ooo`P0oooo00<000000?ooo`3oool04@3oool00`000000
oooo0?ooo`3o0?ooog@0oooo000:0?ooo`050?oo_`1d000000000000M02oool0303oool00`3OW4P0
04RL0=ooo`080?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0o`3ooomd0?ooo`00
9P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0?l0ooooM03oool002H0oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`3o0?ooog@0oooo000V0?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool0o`3ooomd0?ooo`009P3oool00`000000oooo0?ooo`0B0?oo
o`030000003oool0oooo0?l0ooooL`3oool002H0oooo00<000000?ooo`3oool04P3oool00`000000
oooo0?ooo`3o0?ooog<0oooo000V0?ooo`<000004P3oool00`000000oooo0?ooo`3o0?ooog<0oooo
000V0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0o`3ooomc0?ooo`009P3oool0
0`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0?l0ooooL`3oool002H0oooo00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`3o0?ooog80oooo000V0?ooo`030000003oool0oooo01<0
oooo00<000000?ooo`3oool0o`3ooomb0?ooo`009P3oool00`000000oooo0?ooo`0C0?ooo`030000
003oool0oooo0?l0ooooLP3oool002H0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?oo
o`3o0?ooog80oooo000V0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0o`3ooomb
0?ooo`009P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0?l0ooooLP3oool002H0
oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`3o0?ooog40oooo000V0?ooo`030000
003oool0oooo01@0oooo00<000000?ooo`3oool0o`3oooma0?ooo`009P3oool00`000000oooo0?oo
o`0D0?ooo`030000003oool0oooo0?l0ooooL@3oool002H0oooo0`00000D0?ooo`030000003oool0
oooo0?l0ooooL@3oool002H0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`3o0?oo
og00oooo000V0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0o`3ooom`0?ooo`00
9P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0?l0ooooL03oool002H0oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`3o0?ooofl0oooo000V0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool0o`3ooom_0?ooo`009P3oool00`000000oooo0?ooo`0F0?oo
o`030000003oool0oooo0?l0ooooK`3oool002H0oooo00<000000?ooo`3oool05P3oool00`000000
oooo0?ooo`3o0?ooofl0oooo000V0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
o`3ooom^0?ooo`009P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo0?l0ooooKP3o
ool002H0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`3o0?ooofh0oooo000V0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0o`3ooom]0?ooo`009P3oool3000001P0
oooo00<000000?ooo`3oool0o`3ooom]0?ooo`009P3oool00`000000oooo0?ooo`0H0?ooo`030000
003oool0oooo0?l0ooooK@3oool002H0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`3o0?ooof`0oooo000V0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0o`3ooom/
0?ooo`009P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0?l0ooooK03oool002H0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`3o0?ooof/0oooo000V0?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool0o`3ooom[0?ooo`002P3oool01@3ogi`0B0000000
000007@0_ooo00<0oooo00<0okmd0000001d_ol00`3oool01@3ogi`0B000000000000000M;oo00T0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`3o0?ooof/0oooo00090?ooo`070?oo
_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo0080oooo00L0omnL04P0
0018W=l0oooo0?oo_`1d0000M;oo00P0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?oo
o`3o0?ooofX0oooo00090?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool01@3oool0
0`3ooml0W4P0001d_`030?ooo`030?noM000B9`0gooo00L0oooo00<000000?ooo`3oool06`3oool0
0`000000oooo0?ooo`3o0?ooofX0oooo00080?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000
M02oool01@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<000000?oo
o`3oool06`3oool00`000000oooo0?ooo`3o0?ooofX0oooo00080?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol01@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0
oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`3o0?ooofT0oooo00080?ooo`030?oO
W0180000M;oo00<0oooo00<0okmd0000B02Lgol01@3oool00`3ogi`0B00007Boo`030?ooo`030;md
000004P0W=oo00L0oooo1000000K0?ooo`030000003oool0oooo0?l0ooooJ@3oool000P0oooo00<0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`050?ooo`080?oo_`1d0000000004RLg`3oool0
okmd00000000M;l80?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0o`3ooomY0?oo
o`00203oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00<0_g@00018W02o
W4P00P0000000`00B9`0gooo0?ooo`070?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3o
ool0o`3ooomX0?ooo`00203oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo
00<0omnL04P0001d_ol0303oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0?l0oooo
J03oool000T0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030=nLB000
04P0W=oo00/0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`3o0?ooofL0oooo0009
0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`030?noM000000007Bo00X0oooo
00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`3o0?ooofL0oooo000:0?ooo`050?oo_`1d
000000000000M02oool02`3oool01@3ooml0W4P00000000007@0_ooo00L0oooo00<000000?ooo`3o
ool07`3oool00`000000oooo0?ooo`3o0?ooofH0oooo000V0?ooo`030000003oool0oooo01l0oooo
00<000000?ooo`3oool0o`3ooomV0?ooo`009P3oool00`000000oooo0?ooo`0P0?ooo`030000003o
ool0oooo0?l0ooooI@3oool002H0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`3o
0?ooofD0oooo000V0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0o`3ooomT0?oo
o`009P3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0?l0ooooI03oool002H0oooo
0`00000R0?ooo`030000003oool0oooo0?l0ooooH`3oool002H0oooo00<000000?ooo`3oool08P3o
ool00`000000oooo0?ooo`3o0?ooof<0oooo000V0?ooo`030000003oool0oooo02<0oooo00<00000
0?ooo`3oool0o`3ooomR0?ooo`009P3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo
0?l0ooooHP3oool002H0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`3o0?ooof40
oooo000V0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0o`3ooomQ0?ooo`009P3o
ool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo0?l0ooooH03oool002H0oooo00<00000
0?ooo`3oool09P3oool00`000000oooo0?ooo`3o0?oooel0oooo000V0?ooo`030000003oool0oooo
02H0oooo00<000000?ooo`3oool0o`3ooomO0?ooo`009P3oool00`000000oooo0?ooo`0W0?ooo`03
0000003oool0oooo0?l0ooooGP3oool002H0oooo00<000000?ooo`3oool09`3oool00`000000oooo
0?ooo`3o0?oooeh0oooo000V0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0o`3o
oomM0?ooo`009P3oool3000002T0oooo00<000000?ooo`3oool0o`3ooomL0?ooo`009P3oool00`00
0000oooo0?ooo`0Y0?ooo`030000003oool0oooo0?l0ooooG03oool002H0oooo00<000000?ooo`3o
ool0:P3oool00`000000oooo0?ooo`3o0?oooe/0oooo000V0?ooo`030000003oool0oooo02/0oooo
00<000000?ooo`3oool0o`3ooomJ0?ooo`009P3oool00`000000oooo0?ooo`0[0?ooo`030000003o
ool0oooo0?l0ooooFP3oool002H0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`3o
0?oooeT0oooo000V0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0o`3ooomI0?oo
o`009P3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo0?l0ooooF03oool002H0oooo
00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`3o0?oooeL0oooo000V0?ooo`030000003o
ool0oooo02l0oooo00<000000?ooo`3oool0o`3ooomF0?ooo`009P3oool00`000000oooo0?ooo`0`
0?ooo`030000003oool0oooo0?l0ooooE@3oool002H0oooo00<000000?ooo`3oool0<@3oool00`00
0000oooo0?ooo`3o0?oooe@0oooo000V0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3o
ool0o`3ooomD0?ooo`009P3oool300000380oooo00<000000?ooo`3oool0o`3ooomC0?ooo`009P3o
ool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo0?l0ooooDP3oool002H0oooo00<00000
0?ooo`3oool0=03oool00`000000oooo0?ooo`3o0?oooe40oooo000V0?ooo`030000003oool0oooo
03D0oooo00<000000?ooo`3oool0o`3ooom@0?ooo`009P3oool00`000000oooo0?ooo`0f0?ooo`03
0000003oool0oooo0?l0ooooC`3oool002H0oooo00<000000?ooo`3oool0=`3oool200000?l0oooo
C`3oool002H0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`3o0?oood`0oooo000:
0?ooo`050?oOW018000000000000M02oool00`3oool00`3o_g@0000007Boo`030?ooo`050=nLB000
0000000000000018W=l02@3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo0?l0oooo
B`3oool000T0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00<0okmd0000001d
_ol00P3oool01`3OW4P0000007Boo`3oool0oooO09a80018W=l0203oool00`000000oooo0?ooo`0k
0?ooo`800000o`3ooom;0?ooo`002@3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo
00D0oooo00<0oooO09a80000M;l00`3oool00`3o_g@004RL0=ooo`070?ooo`030000003oool0oooo
03d0oooo00<000000?ooo`3oool0o`3ooom80?ooo`00203oool00`3ookl0M00004RLg`030?ooo`03
0=nLB00007@0_ooo00H0oooo00<0_g@0001d_`3oool00P3oool00`2oM00004RL0=ooo`070?ooo`03
0000003oool0oooo03h0oooo0P00003o0?ooodP0oooo00080?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol01P3oool01`3ogi`0B01d0;ooo`3oool0_g@00000001d_ol0203oool00`00
0000oooo0?ooo`100?ooo`030000003oool0oooo0?l0ooooA@3oool000P0oooo00<0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`070?ooo`050?oo_`1dB4P0B00000000018W=l02@3oool40000
0400oooo0P00003o0?ooodD0oooo00080?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol01`3oool0103ogi`0B00000000000M;l:0?ooo`030000003oool0oooo04<0oooo00<000000?oo
o`3oool0o`3ooom20?ooo`00203oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0
oooo00H0omnL04P0000007@0_ooo0?noM000M;l90?ooo`030000003oool0oooo04@0oooo0P00003o
0?oood80oooo00080?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool00`2o
M000001d0;ooo`020?ooo`030;md0018W=l0oooo00L0oooo00<000000?ooo`3oool0AP3oool00`00
0000oooo0?ooo`3o0?ooocl0oooo00090?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3o
ool01P3oool00`2oM00004RL0=ooo`020?ooo`030=nLB000M;l0oooo00L0oooo00<000000?ooo`3o
ool0A`3oool200000?l0oooo?`3oool000T0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo
00L0oooo00L0okmd0000001d_ol0oooo0?oo_`1d0000M;oo00P0oooo00<000000?ooo`3oool0B@3o
ool00`000000oooo0?ooo`3o0?oooc`0oooo000:0?ooo`050?oo_`1d000000000000M02oool02@3o
ool01@3o_g@00000000000000000B9cO00T0oooo00<000000?ooo`3oool0BP3oool200000?l0oooo
?03oool002H0oooo00<000000?ooo`3oool0C03oool200000?l0oooo>P3oool002H0oooo00<00000
0?ooo`3oool0CP3oool200000?l0oooo>03oool002H0oooo00<000000?ooo`3oool0D03oool30000
0?l0oooo=@3oool002H0oooo00<000000?ooo`3oool0D`3oool300000?l0oooo<P3oool002H0oooo
00<000000?ooo`3oool0EP3oool200000?l0oooo<03oool002H0oooo0`00001H0?ooo`<00000o`3o
ool]0?ooo`009P3oool00`000000oooo0?ooo`1K0?ooo`<00000o`3ooolZ0?ooo`009P3oool00`00
0000oooo0?ooo`1N0?ooo`<00000o`3ooolW0?ooo`009P3oool00`000000oooo0?ooo`1Q0?ooo`<0
0000o`3ooolT0?ooo`009P3oool00`000000oooo0?ooo`1T0?ooo`<00000o`3ooolQ0?ooo`009P3o
ool00`000000oooo0?ooo`1W0?ooo`<00000o`3ooolN0?ooo`009P3oool00`000000oooo0?ooo`1Z
0?ooo`@00000o`3ooolJ0?ooo`009P3oool00`000000oooo0?ooo`1^0?ooo`D00000o`3ooolE0?oo
o`009P3oool00`000000oooo0?ooo`1c0?ooo`H00000o`3oool?0?ooo`009P3oool00`000000oooo
0?ooo`1i0?ooo`D00000o`3oool:0?ooo`009P3oool00`000000oooo0?ooo`1n0?ooo`D00000o`3o
ool50?ooo`009P3oool00`000000oooo0?ooo`230?ooo`D00000o`3oool002H0oooo0`0000280?oo
o`D00000nP3oool002H0oooo00<000000?ooo`3oool0S@3oool500000?D0oooo000V0?ooo`030000
003oool0oooo0980oooo1P00003_0?ooo`009P3oool00`000000oooo0?ooo`2H0?ooo`X00000i@3o
ool002H0oooo00<000000?ooo`3oool0XP3oool<00000=T0oooo000V0?ooo`030000003oool0oooo
0:h0oooo2000003A0?ooo`009P3oool00`000000oooo0?ooo`2f0?ooo`/00000aP3oool002H0oooo
00<000000?ooo`3oool0`@3oool@00000;H0oooo000V0?ooo`030000003oool0oooo0=40oooo4000
002V0?ooo`009P3oool00`000000oooo0?ooo`3Q0?oooa000000UP3oool002H0oooo00<000000?oo
o`3oool0l@3oool@000008H0oooo000V0?ooo`030000003oool0oooo0?l0oooo0P3oool?000007L0
oooo000V0?ooo`<00000o`3ooolA0?oooal00000F03oool002H0oooo00<000000?ooo`3oool0o`3o
ool`0?oooal00000>@3oool002H0oooo00<000000?ooo`3oool0o`3ooom?0?oooal000006P3oool0
02H0oooo00<000000?ooo`3oool0o`3ooom^0?ooo`P000004P3oool002H0oooo00<000000?ooo`3o
ool0o`3ooon80?ooo`009P3oool00`000000oooo0?ooo`3o0?ooohP0oooo000V0?ooo`030000003o
ool0oooo0?l0ooooR03oool002H0oooo00<000000?ooo`3oool0o`3ooon80?ooo`009P3oool00`00
0000oooo0?ooo`3o0?ooohP0oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-5.17794, -0.118256, \
0.200111, 0.00623502}}]
}, Closed]],

Cell[BoxData[
    \(\(tableAsymptoticBesselI\  = \ 
        Table[plotAsymptoticBesselI[n, 0, 50, 
            DisplayFunction \[Rule] Identity], {n, 0, 5}];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[tableAsymptoticBesselI, 
        DisplayFunction \[Rule] $DisplayFunction];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0190476 0.0147151 0.500874 [
[.21429 .00222 -8 -15.6875 ]
[.21429 .00222 8 0 ]
[.40476 .00222 -8 -15.6875 ]
[.40476 .00222 8 0 ]
[.59524 .00222 -8 -15.6875 ]
[.59524 .00222 8 0 ]
[.78571 .00222 -8 -15.6875 ]
[.78571 .00222 8 0 ]
[.97619 .00222 -8 -15.6875 ]
[.97619 .00222 8 0 ]
[.01131 .11489 -19 -7.84375 ]
[.01131 .11489 0 7.84375 ]
[.01131 .21506 -19 -7.84375 ]
[.01131 .21506 0 7.84375 ]
[.01131 .31524 -19 -7.84375 ]
[.01131 .31524 0 7.84375 ]
[.01131 .41541 -19 -7.84375 ]
[.01131 .41541 0 7.84375 ]
[.01131 .51559 -10 -7.84375 ]
[.01131 .51559 0 7.84375 ]
[.01131 .61576 -19 -7.84375 ]
[.01131 .61576 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .01472 m
.21429 .02097 L
s
gsave
.21429 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.40476 .01472 m
.40476 .02097 L
s
gsave
.40476 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.59524 .01472 m
.59524 .02097 L
s
gsave
.59524 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(30) show
1.000 setlinewidth
grestore
.78571 .01472 m
.78571 .02097 L
s
gsave
.78571 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
.97619 .01472 m
.97619 .02097 L
s
gsave
.97619 .00222 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(50) show
1.000 setlinewidth
grestore
.125 Mabswid
.0619 .01472 m
.0619 .01847 L
s
.1 .01472 m
.1 .01847 L
s
.1381 .01472 m
.1381 .01847 L
s
.17619 .01472 m
.17619 .01847 L
s
.25238 .01472 m
.25238 .01847 L
s
.29048 .01472 m
.29048 .01847 L
s
.32857 .01472 m
.32857 .01847 L
s
.36667 .01472 m
.36667 .01847 L
s
.44286 .01472 m
.44286 .01847 L
s
.48095 .01472 m
.48095 .01847 L
s
.51905 .01472 m
.51905 .01847 L
s
.55714 .01472 m
.55714 .01847 L
s
.63333 .01472 m
.63333 .01847 L
s
.67143 .01472 m
.67143 .01847 L
s
.70952 .01472 m
.70952 .01847 L
s
.74762 .01472 m
.74762 .01847 L
s
.82381 .01472 m
.82381 .01847 L
s
.8619 .01472 m
.8619 .01847 L
s
.9 .01472 m
.9 .01847 L
s
.9381 .01472 m
.9381 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .11489 m
.03006 .11489 L
s
gsave
.01131 .11489 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.02381 .21506 m
.03006 .21506 L
s
gsave
.01131 .21506 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.02381 .31524 m
.03006 .31524 L
s
gsave
.01131 .31524 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.02381 .41541 m
.03006 .41541 L
s
gsave
.01131 .41541 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.02381 .51559 m
.03006 .51559 L
s
gsave
.01131 .51559 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .61576 m
.03006 .61576 L
s
gsave
.01131 .61576 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.2) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .03976 m
.02756 .03976 L
s
.02381 .0648 m
.02756 .0648 L
s
.02381 .08985 m
.02756 .08985 L
s
.02381 .13993 m
.02756 .13993 L
s
.02381 .16498 m
.02756 .16498 L
s
.02381 .19002 m
.02756 .19002 L
s
.02381 .24011 m
.02756 .24011 L
s
.02381 .26515 m
.02756 .26515 L
s
.02381 .2902 m
.02756 .2902 L
s
.02381 .34028 m
.02756 .34028 L
s
.02381 .36533 m
.02756 .36533 L
s
.02381 .39037 m
.02756 .39037 L
s
.02381 .44046 m
.02756 .44046 L
s
.02381 .4655 m
.02756 .4655 L
s
.02381 .49055 m
.02756 .49055 L
s
.02381 .54063 m
.02756 .54063 L
s
.02381 .56568 m
.02756 .56568 L
s
.02381 .59072 m
.02756 .59072 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01653 m
.02499 .30824 L
.02605 .39917 L
.02729 .46548 L
.02846 .50785 L
.02954 .53595 L
.03053 .55524 L
.03163 .57098 L
.03279 .583 L
.03395 .59143 L
.0352 .59743 L
.03638 .60089 L
.03746 .60262 L
.03873 .60332 L
.03944 .60322 L
.04011 .60286 L
.04132 .60169 L
.04262 .59987 L
.05214 .58021 L
.06244 .56169 L
.06837 .55402 L
.07382 .54856 L
.07893 .54451 L
.08454 .54098 L
.09429 .53648 L
.09961 .53462 L
.10458 .53317 L
.11503 .53075 L
.12493 .52901 L
.14349 .52665 L
.15314 .52573 L
.1634 .52491 L
.18182 .52373 L
.20129 .52277 L
.21957 .52206 L
.25826 .52093 L
.2783 .52049 L
.2994 .5201 L
.3375 .51953 L
.37807 .51906 L
.41711 .51871 L
.45861 .5184 L
.49859 .51816 L
.53706 .51796 L
.57798 .51778 L
.61738 .51764 L
.65527 .51751 L
.69561 .51739 L
.73443 .51729 L
.77571 .5172 L
Mistroke
.81547 .51712 L
.85371 .51705 L
.89441 .51698 L
.93359 .51692 L
.97523 .51686 L
.97619 .51686 L
Mfstroke
.02381 .01472 m
.03279 .14503 L
.03793 .2191 L
.04262 .2733 L
.04758 .31813 L
.05288 .35453 L
.0578 .38012 L
.06244 .39889 L
.06754 .41498 L
.07295 .4282 L
.08265 .44524 L
.08804 .45217 L
.09308 .45754 L
.10458 .46689 L
.11496 .47304 L
.12483 .47758 L
.14335 .48389 L
.16316 .48865 L
.18154 .49194 L
.20207 .49477 L
.22067 .49681 L
.26074 .50008 L
.29929 .5023 L
.34029 .50406 L
.37978 .50536 L
.41775 .50636 L
.45818 .50724 L
.49709 .50793 L
.53845 .50855 L
.57829 .50907 L
.61662 .50949 L
.6574 .50989 L
.69667 .51022 L
.73442 .51051 L
.77462 .51079 L
.81331 .51102 L
.85445 .51125 L
.89408 .51145 L
.93218 .51162 L
.97275 .5118 L
.97619 .51181 L
s
.02381 .01472 m
.06244 .18288 L
.08255 .26304 L
.09396 .29566 L
.10458 .3199 L
.1237 .35286 L
.14415 .37799 L
.16546 .39718 L
.1856 .41101 L
.22401 .43011 L
.24385 .43748 L
.26488 .44403 L
.30271 .45342 L
.343 .46104 L
.38177 .4668 L
.423 .47173 L
.46271 .47562 L
.5009 .47876 L
.54154 .4816 L
.58067 .48395 L
.62226 .48612 L
.66232 .48794 L
.70087 .4895 L
.74188 .49097 L
.78136 .49224 L
.81934 .49334 L
.85976 .49441 L
.89867 .49534 L
.94003 .49624 L
.97619 .49697 L
s
.02381 .01472 m
.02499 .01472 L
.02605 .01473 L
.02729 .01477 L
.02846 .01486 L
.02954 .01501 L
.03053 .0152 L
.03163 .01549 L
.03279 .0159 L
.03395 .01644 L
.0352 .01715 L
.03746 .01883 L
.04016 .02151 L
.04262 .02462 L
.04748 .03248 L
.05278 .04332 L
.06244 .06727 L
.10458 .17902 L
.1259 .22332 L
.14523 .25575 L
.18435 .30466 L
.2036 .32282 L
.22442 .33935 L
.26542 .36496 L
.30491 .38354 L
.34288 .39758 L
.38331 .40961 L
.42222 .41909 L
.46358 .42747 L
.50342 .43429 L
.54175 .43992 L
.58253 .44511 L
.6218 .44949 L
.65955 .45321 L
.69975 .45674 L
.73844 .45978 L
.77958 .46268 L
.81921 .46521 L
.85731 .46742 L
.89788 .46957 L
.93692 .47147 L
.97619 .47323 L
s
.02381 .01472 m
.02499 .01472 L
.02605 .01472 L
.02729 .01472 L
.02846 .01472 L
.02954 .01473 L
.03053 .01474 L
.03163 .01475 L
.03279 .01478 L
.03395 .01483 L
.0352 .0149 L
.03638 .01498 L
.03746 .01508 L
.03873 .01523 L
.04011 .01543 L
.04262 .01592 L
.04523 .01663 L
.04759 .01747 L
.0529 .02007 L
.05781 .02343 L
.06244 .02741 L
.07299 .03914 L
.08269 .05251 L
.10458 .08742 L
.14407 .15051 L
.18204 .20112 L
.22246 .24353 L
.26137 .27558 L
.30273 .30267 L
.34258 .32374 L
.38091 .34051 L
.42169 .35547 L
.46096 .36768 L
.49871 .3778 L
.53891 .38718 L
.5776 .39508 L
.61874 .40249 L
.65836 .40882 L
.69647 .41428 L
.73703 .41951 L
.77608 .42406 L
.81757 .42844 L
.85756 .43229 L
.89602 .43569 L
.93694 .43901 L
.97619 .44195 L
s
.02381 .01472 m
.02499 .01472 L
.02605 .01472 L
.02729 .01472 L
.02846 .01472 L
.02954 .01472 L
.03053 .01472 L
.03163 .01472 L
.03279 .01472 L
.03395 .01472 L
.0352 .01473 L
.03638 .01473 L
.03746 .01474 L
.03873 .01476 L
.04011 .01478 L
.04141 .0148 L
.04262 .01483 L
.04386 .01487 L
.04521 .01493 L
.04765 .01506 L
.05003 .01523 L
.05221 .01544 L
.05459 .01573 L
.05719 .01612 L
.06244 .01721 L
.06795 .0188 L
.07296 .02069 L
.08424 .0265 L
.09468 .03369 L
.10458 .04185 L
.14291 .08066 L
.18369 .12447 L
.22296 .16285 L
.26071 .19491 L
.30091 .2241 L
.3396 .24806 L
.38074 .26985 L
.42036 .2879 L
.45847 .30303 L
.49903 .31714 L
.53808 .3291 L
.57958 .34037 L
.61956 .35003 L
.65802 .35839 L
.69894 .36641 L
.73835 .37341 L
.77623 .37955 L
.81657 .38554 L
.8554 .39084 L
.89667 .39601 L
Mistroke
.93643 .40061 L
.97468 .40471 L
.97619 .40487 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00K`3oool01P2oM000000000000000000004RL0=ooo`<0oooo00D0omnL04P000000000001d
0;ooo`0h0?ooo`030;md00000000000000@0000000T0B9cO0?ooo`3oool0oooo0?oOW01800000000
0000M02oool0=`3oool00`3ookl0M000000000020000000307Boo`3oool0oooo00<0oooo00D0omnL
04P000000000001d0;ooo`0l0?ooo`040?oog`2LB00004RL0=ooo`<0oooo00D0omnL04P000000000
001d0;ooo`0h0?ooo`030?oog`2LB00000000080000000<0M;oo0?ooo`3oool00`3oool01@3ogi`0
B0000000000007@0_ooo01@0oooo001`0?ooo`030?oOW0180000M;oo00@0oooo00L0oono07@0001d
_ol0oooo0?noM00007@0_ooo03P0oooo00<0gia80000000000000`0000002P00B9`0gooo0?ooo`3o
okl0M00007Boo`3oool0okmd0000M02ooolf0?ooo`030?oOW0180000B9cO0080oooo00<0_g@004RL
g`3oool00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0>`3oool0103ooml0W4P00018
W03Oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0g0?ooo`030?oog`2LB7@0_ooo
0080oooo00<0_g@007Boo`3oool00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool04`3o
ool00700oooo00<0omnL04P0001d_ol0103oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0
oooo03P0oooo00<0gia804RLg`3oool00P3oool01P3OW7@0W=oo0?ooo`2oM00004RL0=ooo`80oooo
00<0_g@0001d_`3oool0>P3oool00`3ookl0M01809cOo`020?ooo`030;md0000B9`0gooo0080oooo
00<0_g@0001d_`3oool0>P3oool0103ooml0W4P00018W03Oool20?ooo`030;md0000B9`0gooo0080
oooo00<0_g@0001d_`3oool0>`3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md0000B9`0gooo0080
oooo00<0_g@0001d_`3oool04P3oool00700oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M000
04RLg`030?ooo`030=nLB00007@0_ooo03T0oooo00<0gia807Boo`3oool00`3oool00`3ookl0M000
04RLg`030?ooo`030=nLB00007@0_ooo03X0oooo00L0oono07@0001d_ol0oooo0?oo_`1d0000B9cO
00<0oooo00<0gia80000M02oool0=P3oool00`2oM0000000000000050000000407Boo`3ookl0M000
04RLg`<0oooo00<0gia80000M02oool0?03oool01P2oM000M;oo0?ooo`3ookl0M00004RLg`<0oooo
00<0gia80000M02oool04P3oool00700oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo03X0oooo00<0_gA809cOo`3oool00P3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo03X0oooo00L0okmd0000001d_ol0oooo0?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol0=P3oool0303ogi`0M9cO0?ooo`3oool0oooO09a80000B9`0gooo0?oo
o`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol0>`3oool01`3ooml0W4P004RLg`3oool0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`0B0?ooo`00L03oool00`3ogi`0B00007Boo`030?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0>P3oool00`3ooml0W4Qd0;ooo`020?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0>@3oool0203ogi`0B0000000B02Lgol0
oooo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0g0?ooo`0;0=nLB01d_ol0oooo0?oog`2L
B00004RL0=ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0k0?ooo`070=nLB000
0000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo0180oooo001`0?ooo`030?oO
W0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0k0?ooo`070?no
M000M;l0oooo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo03L0oooo00D0okmd0000
00000000001809cOo`020?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0=`3oool0
2`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol0=`3oool02`3Ogml0oooo0?oo_`1d000000000000M02oool0oooo0?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol04P3oool00700oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo03/0oooo00L0oono07@0B02Lgol0oooo0?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol0>@3oool00`2oM00007Bo0?ooo`020?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol0>03oool02P3ogi`0B7Bo0?oog`2LB00004RL0=ooo`3oool0omnL
04P0001d_ol30?ooo`030?noM00004P0W=oo03L0oooo00H0gia800000000000000000000B02Lgol2
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol04P3oool00700oooo00<0omnL04P0
001d_ol00`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo03/0oooo00L0oono07@0
0018W=l0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0>P3oool01`2oM000B9cO0?oo
o`3oool0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0i0?ooo`090=nLB01d_ml0W4P00018
W03Oool0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0=`3oool0103ookl0M0000000
0018W=l40?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool04P3oool00700oooo00<0
omnL04P0001d_ol0103oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo03H0oooo00<0
oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`020?ooo`030=nLB000B9`0gooo0080oooo00<0
_g@0001d_`3oool0=@3oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo0080oooo00<0
gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`0i0?ooo`050?oog`2LB4P0B4P00018W03Oool0
0P3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo03P0oooo00<0_g@007Boo`3oool0
1@3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo0180oooo001^0?ooo`050?oog`2L
B00000000000001d_ol0103oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0=`3oool0203o
oml0W4Q809cOo`3oool0oono07@0000007@0_ooo0P3oool01`3ookl0M00007Boo`3oool0okmd0000
M02oool0=`3oool01P3o_g@0B9cO0?ooo`3ookl0M000001d_`<0oooo00L0oono07@0001d_ol0oooo
0?noM00007@0_ooo03/0oooo00@0omnL04P00000B9`0gooo0P3oool01`3ookl0M00007Boo`3oool0
okmd0000M02oool0>@3oool01P3ogi`0B000000000000000001809cOo`80oooo00L0oono07@0001d
_ol0oooo0?noM00007@0_ooo01<0oooo001`0?ooo`030?noM0000000M;oo00D0oooo00D0oono07@0
00000000001d0;ooo`0i0?ooo`060?oog`2LB00000000000000007@0_ooo103oool01@3ookl0M000
0000000007@0_ooo03T0oooo00D0gia8000000000000001d0;ooo`040?ooo`050?oo_`1d00000000
0000M02oool0?@3oool00`3OW4P004RL0=ooo`030?ooo`050?oo_`1d000000000000M02oool0>`3o
ool01@2oM0000000000000000000B9cO00<0oooo00D0oono07@000000000001d0;ooo`0D0?ooo`00
o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00;`3oool00`000000
oooo0?ooo`3o0?ooogl0oooo000_0?ooo`030000003oool0oooo0?l0ooooO`3oool002l0oooo00<0
00000?ooo`3oool0o`3ooomo0?ooo`00;`3oool00`000000oooo0?ooo`3o0?ooogl0oooo000_0?oo
o`030000003oool0oooo0?l0ooooO`3oool002l0oooo00<000000?ooo`3oool0o`3ooomo0?ooo`00
9P3ooooo000007X000004@3oool002l0oooo00@000000?ooo`3oool0oooo0P0000020?ooo`800000
0`3oool4000000/0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo01P0oooo000_0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool0oooo
0P0000050?ooo`800000=03oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo04@0oooo
00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo01P0oooo
000_0?ooo`800000103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00@0oooo0P00
003o0?ooofd0oooo000_0?ooo`8000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00D0oooo0P00003o0?ooof/0oooo000_0?ooo`8000001P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool0o`3ooomX0?ooo`00;`3oool2000000H0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`060?ooo`800000o`3ooomX0?ooo`00
;`3oool00`000000oooo000000060?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`3o0?ooofD0oooo000_0?ooo`030000003oool0000000H0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0?l0ooooI03o
ool002l0oooo0`0000070?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01`3oool2
00000?l0ooooI03oool002l0oooo0`0000070?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`3o0?ooof40oooo000_0?ooo`<00000203oool00`000000
oooo0?ooo`060?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0o`3ooomP0?ooo`00
;`3oool2000000030?ooo`000000oooo00H0oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo0?l0ooooG`3oool002l0oooo0P0000000`3oool000000?oo
o`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`3o0?oooeh0oooo000_0?ooo`80000000<0oooo0000003oool01`3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0o`3ooomM0?ooo`00;`3oool20000
00030?ooo`000000oooo00L0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?oo
o`030000003oool0oooo0?l0ooooG03oool002l0oooo0P0000020?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0?l0
ooooF`3oool002l0oooo0P0000020?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0?l0ooooFP3oool002l0oooo0P00
00020?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo0?l0ooooFP3oool002l0oooo0P0000020?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo0?l0ooooF@3oool002l0oooo0P0000020?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0?l0ooooF03oool002l0
oooo0P0000030?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool02@3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo0?l0ooooE`3oool002l0oooo0P0000030?ooo`030000
003oool0oooo00L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo0?l0ooooEP3oool002l0oooo0P0000030?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0?l0ooooE@3o
ool002l0oooo0P0000030?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0?l0ooooE03oool002l0oooo00<000000?oo
o`0000000`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool0303oool00`000000oooo0?ooo`3o0?oooe<0oooo000_0?ooo`030000003oool0000000<0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool0o`3ooomB0?ooo`00;`3oool00`000000oooo000000030?ooo`030000
003oool0oooo00P0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0=0?ooo`030000
003oool0oooo0?l0ooooD@3oool002l0oooo0`0000030?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0?l0ooooD03o
ool002l0oooo00<000000?ooo`0000000`3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`3o0?ooodl0oooo000_0?oo
o`030000003oool0000000@0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool0o`3ooom>0?ooo`00;`3oool00`000000
oooo000000040?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo0?l0ooooC@3oool002l0oooo00<000000?ooo`000000
103oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
3`3oool00`000000oooo0?ooo`3o0?oood`0oooo000C0?ooo`050?oOW018000000000000M02oool0
0`3oool01@3o_g@0000007Boo`3oool0_g@000H0000000<0B9cO0?ooo`3oool01P3oool00`000000
oooo000000040?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo0?l0ooooB`3oool00180oooo00L0oono07@0001d_ol0
oooo0?noM00007@0_ooo0080oooo00<0okmd0000001d_ol00P3oool00`3OW4P00000000000030000
00030018W03Oool0oooo00H0oooo00<000000?ooo`0000001@3oool00`000000oooo0?ooo`080?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`3o0?oo
odX0oooo000B0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`3OW4P0
B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo00H0oooo00<000000?ooo`0000001@3oool00`000000
oooo0?ooo`090?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`3o0?ooodT0oooo000A0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0
203oool00`3OW4P0M;oo0?ooo`0:0?ooo`030000003oool0000000D0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0
o`3ooom80?ooo`004@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00<0
_gA809cOo`3oool02@3oool3000000D0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0o`3ooom70?ooo`004@3oool0
0`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00<0oooO09a8M02oool02@3oool0
0`000000oooo000000050?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0?l0ooooAP3oool00140oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030?noM000M;l0oooo00P0oooo00<000000?oo
o`0000001P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool04P3oool00`000000oooo0?ooo`3o0?ooodD0oooo000A0?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol02P3oool00`3ookl0M01809cOo`080?ooo`030000003oool0000000H0
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0180
oooo00<000000?ooo`3oool0o`3ooom40?ooo`004@3oool00`3ookl0M00004RLg`030?ooo`030=nL
B00007@0_ooo00X0oooo00<0oono07@00018W=l0203oool010000000oooo0?ooo`0000050?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0B0?ooo`03
0000003oool0oooo0?l0oooo@`3oool00180oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo
0?ooo`050?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l0203oool010000000oooo
0?ooo`0000050?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0C0?ooo`030000003oool0oooo0?l0oooo@P3oool00180oooo00L0oono07@0001d_ol0
oooo0?noM00007@0_ooo00H0oooo00P0oooO09a8B02Lgol0oooo0?oo_`1d0000001d0;ooo`P0oooo
00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`3o0?oood40oooo000C0?ooo`050?oo
_`1d000000000000M02oool0203oool01P3ooml0W4P0000000000000001d0;ooo`T0oooo00@00000
0?ooo`3oool000001P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool04`3oool00`000000oooo0?ooo`3o0?oood00oooo000_0?ooo`040000003oool0
oooo000000H0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo01@0oooo00<000000?ooo`3oool0o`3ooolo0?ooo`00;`3oool010000000oooo0?ooo`00
00060?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`0D0?ooo`030000003oool0oooo0?l0oooo?P3oool002l0oooo0P0000000`3oool000000?ooo`05
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo0?l0oooo?@3oool002l0oooo00@000000?ooo`3oool000001`3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`3o0?oooc`0oooo000_0?ooo`040000003oool0oooo000000L0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01@0oooo00<00000
0?ooo`3oool0o`3oooll0?ooo`00;`3oool010000000oooo0?ooo`0000070?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo0?l0oooo>`3oool002l0oooo00@000000?ooo`3oool000001`3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`3o
0?ooocX0oooo000_0?ooo`040000003oool0oooo000000P0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0o`3oooli
0?ooo`00;`3oool010000000oooo0?ooo`0000080?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0403oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0?l0oooo>03oool0
02l0oooo00@000000?ooo`3oool00000203oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`3o0?ooocL0oooo000_0?oo
o`040000003oool0oooo000000P0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0B
0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0o`3ooolf0?ooo`00;`3oool20000
00030?ooo`000000oooo00L0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0B0?oo
o`030000003oool0oooo01H0oooo00<000000?ooo`3oool0o`3ooole0?ooo`00;`3oool01@000000
oooo0?ooo`3oool0000000P0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0B0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool0o`3ooold0?ooo`00;`3oool01@000000
oooo0?ooo`3oool0000000P0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0C0?oo
o`030000003oool0oooo01L0oooo0P00003o0?oooc@0oooo000_0?ooo`050000003oool0oooo0?oo
o`000000203oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01<0oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`3o0?oooc40oooo000_0?ooo`050000003oool0oooo0?oo
o`000000203oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01@0oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`3o0?oooc00oooo000_0?ooo`050000003oool0oooo0?oo
o`0000002@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01<0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`3o0?ooobl0oooo000_0?ooo`050000003oool0oooo0?oo
o`0000002@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01@0oooo00<000000?oo
o`3oool06@3oool200000?l0oooo;`3oool002l0oooo00D000000?ooo`3oool0oooo000000090?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0J0?oo
o`030000003oool0oooo0?l0oooo;03oool002l0oooo00D000000?ooo`3oool0oooo000000090?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0J0?oo
o`030000003oool0oooo0?l0oooo:`3oool002l0oooo00D000000?ooo`3oool0oooo0000000:0?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0K0?oo
o`030000003oool0oooo0?l0oooo:P3oool002l0oooo0P0000020?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo01/0
oooo0P00003o0?ooobX0oooo000_0?ooo`050000003oool0oooo0?ooo`0000002P3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0703oool00`000000
oooo0?ooo`3o0?ooobL0oooo000_0?ooo`050000003oool0oooo0?ooo`0000002P3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0703oool00`000000
oooo0?ooo`3o0?ooobH0oooo000_0?ooo`050000003oool0oooo0?ooo`0000002`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool07@3oool200000?l0
oooo9P3oool001<0oooo00D0omnL04P000000000001d0;ooo`030?ooo`030?noM0000000M;oo00D0
oooo00@0oooO09a80000B9`0gooo203oool01@000000oooo0?ooo`3oool0000000/0oooo00<00000
0?ooo`3oool03`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01h0oooo00<00000
0?ooo`3oool0o`3ooolS0?ooo`004P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3o
ool00`3o_g@0000007Boo`050?ooo`040?oog`2LB00004RL0=ooo`P0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
5`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo0?l0oooo8P3oool00180oooo00<0
_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`090?ooo`040?oog`2LB00004RL0=ooo`P0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0N0?ooo`800000o`3ooolR0?ooo`004@3o
ool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00D0oooo00<0_g@00000000000001@00
00000`1d_ol0oooo0?ooo`050?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool07`3o
ool00`000000oooo0?ooo`3o0?oooal0oooo000A0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol01@3oool0203ogi`0M9cO0?ooo`3oool0oooO09a80000B9`0gooo203oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0o`3ooolN0?ooo`00
4@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0gia807Boo`3oool0
oooO09a80000B9`0gooo00P0oooo0`0000020?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01l0oooo0P00003o0?oo
oah0oooo000A0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01`3ooml0
W4Qd0;ooo`3ooml0W4P00018W03Oool0203oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo0200oooo00<000000?ooo`3oool0o`3ooolK0?ooo`004@3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00L0oooo00H0omnL04Qd_`3ooml0W4P00018W03Oool80?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`3o0?oooaX0oooo000A
0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool01@3OW4P0M;oO09a80000
B9`0gooo00P0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo0140oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0Q0?ooo`800000o`3o
oolJ0?ooo`004P3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00P0oooo00D0oooO
09a8B018B00004RL0=ooo`080?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool08P3o
ool00`000000oooo0?ooo`3o0?oooaL0oooo000B0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d
0;ooo`0:0?ooo`040?oOW018000004RL0=ooo`P0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool06P3oool00`000000
oooo0?ooo`0R0?ooo`800000o`3ooolG0?ooo`004`3oool01@3ookl0M0000000000007@0_ooo00`0
oooo00<0gia80018W03Oool0203oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo02<0
oooo00<000000?ooo`3oool0o`3ooolD0?ooo`00;`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0K0?ooo`030000
003oool0oooo02<0oooo0P00003o0?oooa@0oooo000_0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01/0oooo00<0
00000?ooo`3oool0903oool00`000000oooo0?ooo`3o0?oooa40oooo000_0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0
oooo01`0oooo00<000000?ooo`3oool0903oool200000?l0oooo4@3oool002l0oooo0P0000040?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0L0?oo
o`030000003oool0oooo02D0oooo0P00003o0?ooo`l0oooo000_0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo01`0
oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`3o0?ooo``0oooo000_0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0D0?ooo`030000
003oool0oooo01d0oooo00<000000?ooo`3oool09P3oool200000?l0oooo303oool002l0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01@0oooo00<0
00000?ooo`3oool07@3oool2000002P0oooo00<000000?ooo`3oool0o`3oool90?ooo`00;`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0503oool0
0`000000oooo0?ooo`0O0?ooo`030000003oool0oooo02H0oooo0P00003o0?ooo`T0oooo000_0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo01l0oooo00<000000?ooo`3oool09`3oool200000?l0oooo1`3oool002l0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01H0
oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0X0?ooo`800000o`3oool50?ooo`00
;`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
5P3oool00`000000oooo0?ooo`0O0?ooo`800000:P3oool200000?l0oooo0`3oool002l0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01H0oooo00<0
00000?ooo`3oool08@3oool00`000000oooo0?ooo`0Y0?ooo`800000o`3oool10?ooo`00;`3oool2
000000D0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
oooo0240oooo00<000000?ooo`3oool0:P3oool200000?h0oooo000_0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo
0240oooo00<000000?ooo`3oool0:`3oool200000?`0oooo000_0?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo0240
oooo0P00000]0?ooo`<00000n@3oool002l0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`0=0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool08P3oool00`000000oooo
0?ooo`0]0?ooo`800000m`3oool002l0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?oo
o`0^0?ooo`800000m@3oool002l0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool08`3oool200000300oooo0P00003c
0?ooo`00;`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo02l0oooo0P00003a0?oo
o`00;`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0T0?ooo`800000<@3oool200000>l0oooo000_0?ooo`030000
003oool0oooo00D0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0J0?ooo`030000
003oool0oooo02H0oooo00<000000?ooo`3oool0<03oool200000>d0oooo000_0?ooo`8000001P3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool09P3o
ool200000380oooo0`00003Z0?ooo`00;`3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0
oooo0380oooo0`00003W0?ooo`00;`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0W0?ooo`800000=@3oool20000
0>D0oooo000_0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0M0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0=03oool300000>80
oooo000_0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0403oool00`000000oooo
0?ooo`0L0?ooo`030000003oool0oooo02T0oooo0P00000g0?ooo`<00000g`3oool001<0oooo00D0
omnL04P000000000001d0;ooo`030?ooo`030?noM0000000M;oo00<0oooo00D0omnL04P000000000
000007Boo`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`0M0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0=`3oool300000=`0
oooo000B0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo
0080oooo00L0omnL04P00018W=l0oooo0?oo_`1d0000M;oo00P0oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0:P3o
ool2000003X0oooo0P00003J0?ooo`004P3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0
oooo00D0oooo00<0oooO09a80000M;l00`3oool00`3o_g@004RL0=ooo`070?ooo`030000003oool0
oooo00H0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0N0?ooo`030000003oool0
oooo02/0oooo0P00000j0?ooo`<00000e`3oool00140oooo00<0oono07@00018W=l00`3oool00`3O
W4P0001d0;ooo`050?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool07`3oool00`00
0000oooo0?ooo`0/0?ooo`800000>`3oool200000=D0oooo000A0?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol01@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo01l0
oooo00<000000?ooo`3oool0;@3oool2000003/0oooo0`00003B0?ooo`004@3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00D0oooo00<0omnL04P0001d_ol00`3oool00`2oM0000018
09cOo`070?ooo`<000001P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0200oooo
00<000000?ooo`3oool0;P3oool2000003`0oooo0P00003@0?ooo`004@3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00D0oooo00P0oono07@000000000B9cO0?ooo`3o_g@00000001d
_`P0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo
0200oooo00<000000?ooo`3oool0;`3oool2000003`0oooo0`00003=0?ooo`004@3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00<0_g@00018W02oW4P00P0000000`00B9`0
gooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool04P3oool00`000000
oooo0?ooo`0Q0?ooo`030000003oool0oooo0300oooo0P00000m0?ooo`@00000b@3oool00140oooo
00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`030?oOW0180000M;oo00`0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0240oooo
00<000000?ooo`3oool0<@3oool2000003l0oooo100000350?ooo`004P3oool00`3OW4P004RL0=oo
o`020?ooo`030;md0000M;l0oooo00L0oooo00<0gia80000B02Lgol02`3oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?oo
o`0b0?ooo`800000@@3oool400000<40oooo000B0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d
0;ooo`090?ooo`030?noM000000007Bo00X0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`0C0?ooo`030000003oool0oooo0280oooo0P00000d0?ooo`<00000@P3oool300000;h0oooo
000C0?ooo`050?oo_`1d000000000000M02oool02`3oool01@3ooml0W4P00000000007@0_ooo00L0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo02@0
oooo00<000000?ooo`3oool0=03oool2000004<0oooo1000002j0?ooo`00;`3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0903oool00`000000oooo
0?ooo`0e0?ooo`<00000A03oool400000;H0oooo000_0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0503oool00`000000oooo0?ooo`0U0?ooo`800000>03oool2000004H0oooo1000
002b0?ooo`00;`3oool2000000P0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0V
0?ooo`030000003oool0oooo03L0oooo0`0000170?ooo`@00000[P3oool002l0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo02H0oooo0P00000j0?oo
o`800000B@3oool400000:X0oooo000_0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool05P3oool00`000000oooo0?ooo`0X0?ooo`800000>P3oool3000004X0oooo1000002V0?ooo`00
;`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0
:@3oool00`000000oooo0?ooo`0j0?ooo`800000C03oool400000:80oooo000_0?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0Z0?ooo`800000?03o
ool3000004d0oooo1000002N0?ooo`00;`3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo01L0oooo00<000000?ooo`3oool0:`3oool2000003d0oooo0P00001?0?ooo`@00000VP3oool0
02l0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo
02`0oooo00<000000?ooo`3oool0?03oool300000500oooo1@00002E0?ooo`00;`3oool00`000000
oooo0?ooo`080?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0;03oool2000003l0
oooo1000001A0?ooo`D00000T03oool002l0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo02d0oooo0P0000110?ooo`@00000DP3oool5000008/0oooo
000_0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?oo
o`0^0?ooo`800000@`3oool4000005<0oooo1@0000260?ooo`00;`3oool2000000T0oooo00<00000
0?ooo`3oool06`3oool00`000000oooo0?ooo`0`0?ooo`800000A@3oool3000005D0oooo1@000021
0?ooo`00;`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01/0oooo00<000000?oo
o`3oool0<@3oool2000004H0oooo1000001F0?ooo`D00000O03oool002l0oooo00<000000?ooo`3o
ool02@3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo0380oooo0P0000180?ooo`<0
0000F03oool5000007L0oooo000_0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
7@3oool00`000000oooo0?ooo`0c0?ooo`800000B@3oool4000005T0oooo1P00001a0?ooo`00;`3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0=03o
ool2000004/0oooo0`00001L0?ooo`L00000JP3oool002l0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`0O0?ooo`030000003oool0oooo03D0oooo0P00001<0?ooo`@00000G`3oool7
000006<0oooo000_0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0803oool00`00
0000oooo0?ooo`0f0?ooo`<00000C@3oool400000680oooo2000001K0?ooo`00;`3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0>03oool2000004l0
oooo1000001V0?ooo`P00000D`3oool002l0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`0Q0?ooo`030000003oool0oooo03T0oooo0`00001@0?ooo`@00000JP3oool7000004`0oooo
000_0?ooo`8000002`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo03/0oooo0P00
001B0?ooo`D00000K03oool7000004D0oooo000_0?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool08`3oool00`000000oooo0?ooo`0l0?ooo`<00000E03oool5000006h0oooo2000000m
0?ooo`00;`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo02<0oooo00<000000?oo
o`3oool0?P3oool2000005L0oooo1@00001a0?ooo`P00000=@3oool002l0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo03l0oooo0`00001I0?ooo`D0
0000M03oool8000002d0oooo000C0?ooo`050?oOW018000000000000M02oool00`3oool00`3o_g@0
000007Boo`030?ooo`050=nLB0000000000000000018W=l02@3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo02D0oooo0P0000120?ooo`@00000FP3oool5000007L0oooo2P00000S0?oo
o`004P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool00`3o_g@0000007Boo`02
0?ooo`070=nLB0000000M;oo0?ooo`3ooml0W4P004RLg`080?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`130?ooo`@00000F`3oool6000007/0oooo
2@00000J0?ooo`004P3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00D0oooo00<0
oooO09a80000M;l00`3oool00`3o_g@004RL0=ooo`070?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool09`3oool00`000000oooo0?ooo`160?ooo`@00000G@3oool8000009H0oooo000A
0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool00`2oM00007Bo0?ooo`02
0?ooo`030;md0000B9`0gooo00L0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0X
0?ooo`800000BP3oool300000680oooo1`00002?0?ooo`004@3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00H0oooo00L0omnL04P0M02oool0oooo0;md00000000M;oo00P0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0Y0?ooo`800000B`3oool4000006D0oooo1`00
00280?ooo`004@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00D0oono
07A8B0180000000004RLg`090?ooo`<000003@3oool00`000000oooo0?ooo`0[0?ooo`800000C@3o
ool3000006T0oooo1`0000210?ooo`004@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo00L0oooo00@0omnL04P00000000007Bo2P3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo02d0oooo0P00001>0?ooo`@00000K03oool8000007T0oooo000A0?ooo`030?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3ogi`0B0000000M02oool0okmd001d_`T0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0^0?ooo`800000D03oool400000700oooo
2000001a0?ooo`004@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00<0
_g@00000M02oool00P3oool00`2oM000B9cO0?ooo`070?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool0<03oool200000580oooo1@00001c0?ooo`L00000JP3oool00180oooo00<0gia8
0018W03Oool00P3oool00`2oM00007Bo0?ooo`060?ooo`030;md0000B9`0gooo0080oooo00<0gia8
001d_`3oool01`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0340oooo0P00001E
0?ooo`H00000M03oool;000005l0oooo000B0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oo
o`070?ooo`070?noM0000000M;oo0?ooo`3ookl0M00007Boo`080?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool0<`3oool2000005T0oooo1@00001j0?ooo``00000D`3oool001<0oooo
00D0oono07@000000000001d0;ooo`090?ooo`050?noM0000000000000000018W=l02@3oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo03@0oooo0`00001K0?ooo`D00000P@3oool80000
04/0oooo000_0?ooo`8000004@3oool00`000000oooo0?ooo`0g0?ooo`800000GP3oool5000008@0
oooo2P0000110?ooo`00;`3oool200000180oooo00<000000?ooo`3oool0>03oool200000640oooo
1P0000280?ooo`l00000<P3oool002l0oooo0P00000B0?ooo`030000003oool0oooo03X0oooo0`00
001T0?ooo`L00000T03oool@00000280oooo000_0?ooo`8000004`3oool00`000000oooo0?ooo`0l
0?ooo`@00000I`3oool7000009T0oooo2000000J0?ooo`00;`3oool2000001<0oooo00<000000?oo
o`3oool0@03oool4000006X0oooo2000002c0?ooo`00;`3oool2000001@0oooo00<000000?ooo`3o
ool0@`3oool4000006h0oooo2`00002X0?ooo`00;`3oool2000001D0oooo00<000000?ooo`3oool0
AP3oool3000007H0oooo2`00002M0?ooo`00;`3oool2000001H0oooo00<000000?ooo`3oool0B03o
ool4000007d0oooo2000002E0?ooo`00;`3oool2000001L0oooo00<000000?ooo`3oool0B`3oool4
00000840oooo2`00002:0?ooo`00;`3oool2000001P0oooo00<000000?ooo`3oool0CP3oool50000
08L0oooo3P00001l0?ooo`00;`3oool2000001T0oooo00<000000?ooo`3oool0DP3oool6000008l0
oooo3`00001]0?ooo`00;`3oool2000001X0oooo00<000000?ooo`3oool0E`3oool6000009P0oooo
4000001M0?ooo`00;`3oool2000001/0oooo00<000000?ooo`3oool0G03oool700000:40oooo3P00
001?0?ooo`00;`3oool2000001`0oooo00<000000?ooo`3oool0HP3oool700000:P0oooo3`000010
0?ooo`00;`3oool2000001d0oooo0P00001Y0?ooo`P00000[`3oool?00000340oooo000_0?ooo`80
00007`3oool2000006l0oooo3000002b0?oooaL000006P3oool002l0oooo0P00000Q0?ooo`800000
N@3oool>00000=D0oooo000_0?ooo`8000008`3oool3000008@0oooo3`0000360?ooo`00;`3oool2
000002H0oooo1000002?0?ooo`l00000]`3oool002l0oooo0P00000Z0?ooo`<00000V`3oool?0000
0:P0oooo000_0?ooo`800000;@3oool400000:H0oooo4000002H0?ooo`00;`3oool200000340oooo
1@00002a0?oooad00000N`3oool002l0oooo0P00000f0?ooo`L00000a`3ooolN000005d0oooo000_
0?ooo`800000?@3oool800000=d0oooo;000000a0?ooo`00;`3oool2000004D0oooo2`00003n0?oo
oaL000006P3oool002l0oooo0P00001@0?ooo`h00000o`3ooolR0?ooo`00803oool01P2oM0000000
00000000000004RL0=ooo`T0oooo0P00001N0?ooo`l00000o`3ooolC0?ooo`008@3oool00`3ogi`0
B00007Boo`0;0?ooo`800000K@3ooolN00000?@0oooo000Q0?ooo`030?oOW0180000M;oo00/0oooo
0P00002;0?oooc`00000^03oool00240oooo00<0omnL04P0001d_ol02`3oool200000<L0ooooM`00
00110?ooo`008@3oool00`3ogi`0B00007Boo`0;0?ooo`800000o`3ooolo0?ooobL000006P3oool0
0240oooo00<0omnL04P0001d_ol02`3oool300000?l0ooooO`3oool00240oooo00<0omnL04P0001d
_ol02`3oool2000007`0ooooj@00000J0?ooo`008@3oool00`3ogi`0B00007Boo`0;0?ooo`800000
C`3oool]00000?l0oooo103oool00240oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0000
000e0?oooaT00000o`3ooola0?ooo`008@3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0
000002L0oooo3P00003o0?ooodX0oooo000O0?ooo`050?oog`2LB00000000000001d_ol02`3oool0
0`000000oooo0000000R0?ooo`D00000o`3ooomH0?ooo`008@3oool00`3o_g@0000007Boo`0;0?oo
o`030000003oool0000001h0oooo1000003o0?oooed0oooo000_0?ooo`030000003oool0000001T0
oooo1@00003o0?ooof40oooo000_0?ooo`030000003oool0000001L0oooo0P00003o0?ooofH0oooo
000_0?ooo`<000005@3oool200000?l0ooooJ03oool002l0oooo00<000000?ooo`0000004`3oool2
00000?l0ooooJP3oool002l0oooo00<000000?ooo`0000004@3oool200000?l0ooooK03oool002l0
oooo00<000000?ooo`0000003`3oool200000?l0ooooKP3oool002l0oooo00<000000?ooo`000000
3P3oool00`000000oooo0?ooo`3o0?ooofh0oooo000_0?ooo`030000003oool0000000h0oooo00<0
00000?ooo`3oool0o`3ooom^0?ooo`00;`3oool00`000000oooo0000000=0?ooo`030000003oool0
oooo0?l0ooooK`3oool002l0oooo00<000000?ooo`000000303oool00`000000oooo0?ooo`3o0?oo
og00oooo000_0?ooo`030000003oool0000000/0oooo00<000000?ooo`3oool0o`3oooma0?ooo`00
;`3oool00`000000oooo0000000:0?ooo`030000003oool0oooo0?l0ooooLP3oool002l0oooo0P00
00000`3oool000000?ooo`080?ooo`030000003oool0oooo0?l0ooooLP3oool002l0oooo00@00000
0?ooo`3oool00000203oool00`000000oooo0?ooo`3o0?ooog<0oooo000_0?ooo`040000003oool0
oooo000000P0oooo00<000000?ooo`3oool0o`3ooomc0?ooo`00;`3oool010000000oooo0?ooo`00
00070?ooo`030000003oool0oooo0?l0ooooM03oool002l0oooo00@000000?ooo`3oool000001`3o
ool00`000000oooo0?ooo`3o0?ooog@0oooo000_0?ooo`040000003oool0oooo000000H0oooo00<0
00000?ooo`3oool0o`3ooome0?ooo`00;`3oool010000000oooo0?ooo`0000060?ooo`030000003o
ool0oooo0?l0ooooM@3oool002l0oooo00@000000?ooo`3oool000001@3oool00`000000oooo0?oo
o`3o0?ooogH0oooo000_0?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0o`3o
oomf0?ooo`00;`3oool2000000030?ooo`000000oooo00<0oooo00<000000?ooo`3oool0o`3ooomg
0?ooo`00;`3oool010000000oooo0?ooo`0000040?ooo`030000003oool0oooo0?l0ooooM`3oool0
02l0oooo00@000000?ooo`3oool00000103oool00`000000oooo0?ooo`3o0?ooogL0oooo000_0?oo
o`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`3o0?ooogP0oooo000_0?oo
o`060000003oool0oooo0?ooo`000000oooo0P00003o0?ooogX0oooo000C0?ooo`060;md00000000
000000000000B9`0gooo0P3oool01@3o_g@0000007Boo`3oool0_g@000H0000000<0B9cO0?ooo`3o
ool01P3oool00`000000oooo0?ooo`020?ooo`800000o`3ooomk0?ooo`00503oool00`3ogi`0B000
07Boo`040?ooo`030?noM0000000M;oo0080oooo00<0gia80000000000000`0000000`00B9`0gooo
0?ooo`060?ooo`030000003oool0oooo0?l0ooooO`3oool001@0oooo00<0omnL04P0001d_ol02P3o
ool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo00H0oooo00<000000?ooo`3oool0o`3o
oomo0?ooo`00503oool00`3ogi`0B00007Boo`0;0?ooo`030=nLB01d_ol0oooo00X0oooo00<00000
0?ooo`3oool0o`3ooomo0?ooo`00503oool00`3ogi`0B00007Boo`0<0?ooo`030;mdB02Lgol0oooo
00T0oooo00<000000?ooo`3oool0o`3ooomo0?ooo`00503oool00`3ogi`0B00007Boo`0<0?ooo`03
0?oog`2LB7@0_ooo00T0oooo0`00003o0?ooogl0oooo000D0?ooo`030?oOW0180000M;oo00d0oooo
00<0okmd001d_`3oool0203oool00`000000oooo0?ooo`3o0?ooogl0oooo000D0?ooo`030?oOW018
0000M;oo00d0oooo00<0oono07@0B02Lgol0o`3ooon:0?ooo`00503oool00`3ogi`0B00007Boo`0=
0?ooo`030?oo_`1d0000B9cO0?l0ooooRP3oool001@0oooo00<0omnL04P0001d_ol0203oool00`3o
_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO0?l0ooooRP3oool00180oooo00D0oooO09a80000
0000000007Boo`080?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02ooooo0?ooohX0oooo
000D0?ooo`030?noM0000000M;oo00T0oooo00H0oooO09a80000000000000000M02ooooo0?oooh/0
oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40
oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-6.58543, -0.149963, \
0.209546, 0.00796878}}]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Plotting points", "Section"],

Cell[CellGroupData[{

Cell["Basic list plots", "Subsection"],

Cell[TextData[{
  "Not surprisingly, a list of points can be plotted using ",
  ButtonBox["ListPlot",
    ButtonStyle->"RefGuideLink"],
  " for two dimensions or ",
  ButtonBox["ListPlot3D",
    ButtonStyle->"RefGuideLink"],
  " for three dimensions.  Suppose that we generate a list of points that \
might resemble data for some reaction."
}], "Text"],

Cell[BoxData[
    \(bw[x_, 
        x0_, \[CapitalGamma]_] := \((\[CapitalGamma]/2)\)\^2\/\(\((x - \
x0)\)\^2 + \ \((\[CapitalGamma]/2)\)\^2\)\)], "Input"],

Cell[BoxData[
    \(\(bwData2\  = 
        Table[{x, \ bw[x, 1, 1\/2]\ \((0.8 + 0.4\ Random[])\)}, {x, 0, 3, 
            0.1}];\)\)], "Input"],

Cell[TextData[{
  "The options for ",
  StyleBox["ListPlot", "InlineInput",
    FontWeight->"Bold"],
  " are similar to those for ",
  StyleBox["Plot", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[".",
    FontWeight->"Bold"],
  "  The appearance of the points can be adjusted using ",
  StyleBox["PlotStyle", "InlineInput",
    FontWeight->"Bold"],
  ", with the most important adjustment being the point size \[LongDash] the \
defaults size is too small on most monitors."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plot2[bwData2]\  = 
        ListPlot[bwData2, 
          PlotStyle \[Rule] AbsolutePointSize[4]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.31746 0.0147151 0.658091 [
[.18254 .00222 -9.5 -15.6875 ]
[.18254 .00222 9.5 0 ]
[.34127 .00222 -5 -15.6875 ]
[.34127 .00222 5 0 ]
[.5 .00222 -9.5 -15.6875 ]
[.5 .00222 9.5 0 ]
[.65873 .00222 -5 -15.6875 ]
[.65873 .00222 5 0 ]
[.81746 .00222 -9.5 -15.6875 ]
[.81746 .00222 9.5 0 ]
[.97619 .00222 -5 -15.6875 ]
[.97619 .00222 5 0 ]
[.01131 .14633 -19 -7.84375 ]
[.01131 .14633 0 7.84375 ]
[.01131 .27795 -19 -7.84375 ]
[.01131 .27795 0 7.84375 ]
[.01131 .40957 -19 -7.84375 ]
[.01131 .40957 0 7.84375 ]
[.01131 .54119 -19 -7.84375 ]
[.01131 .54119 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.18254 .01472 m
.18254 .02097 L
s
gsave
.18254 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.34127 .01472 m
.34127 .02097 L
s
gsave
.34127 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .01472 m
.5 .02097 L
s
gsave
.5 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.5) show
1.000 setlinewidth
grestore
.65873 .01472 m
.65873 .02097 L
s
gsave
.65873 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.81746 .01472 m
.81746 .02097 L
s
gsave
.81746 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2.5) show
1.000 setlinewidth
grestore
.97619 .01472 m
.97619 .02097 L
s
gsave
.97619 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.05556 .01472 m
.05556 .01847 L
s
.0873 .01472 m
.0873 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.15079 .01472 m
.15079 .01847 L
s
.21429 .01472 m
.21429 .01847 L
s
.24603 .01472 m
.24603 .01847 L
s
.27778 .01472 m
.27778 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.37302 .01472 m
.37302 .01847 L
s
.40476 .01472 m
.40476 .01847 L
s
.43651 .01472 m
.43651 .01847 L
s
.46825 .01472 m
.46825 .01847 L
s
.53175 .01472 m
.53175 .01847 L
s
.56349 .01472 m
.56349 .01847 L
s
.59524 .01472 m
.59524 .01847 L
s
.62698 .01472 m
.62698 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.72222 .01472 m
.72222 .01847 L
s
.75397 .01472 m
.75397 .01847 L
s
.78571 .01472 m
.78571 .01847 L
s
.84921 .01472 m
.84921 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.9127 .01472 m
.9127 .01847 L
s
.94444 .01472 m
.94444 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .14633 m
.03006 .14633 L
s
gsave
.01131 .14633 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.02381 .27795 m
.03006 .27795 L
s
gsave
.01131 .27795 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.02381 .40957 m
.03006 .40957 L
s
gsave
.01131 .40957 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.02381 .54119 m
.03006 .54119 L
s
gsave
.01131 .54119 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04762 m
.02756 .04762 L
s
.02381 .08052 m
.02756 .08052 L
s
.02381 .11343 m
.02756 .11343 L
s
.02381 .17924 m
.02756 .17924 L
s
.02381 .21214 m
.02756 .21214 L
s
.02381 .24505 m
.02756 .24505 L
s
.02381 .31086 m
.02756 .31086 L
s
.02381 .34376 m
.02756 .34376 L
s
.02381 .37667 m
.02756 .37667 L
s
.02381 .44247 m
.02756 .44247 L
s
.02381 .47538 m
.02756 .47538 L
s
.02381 .50828 m
.02756 .50828 L
s
.02381 .57409 m
.02756 .57409 L
s
.02381 .607 m
.02756 .607 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
4 Mabswid
.02381 .05497 Mdot
.05556 .06963 Mdot
.0873 .07614 Mdot
.11905 .07544 Mdot
.15079 .10609 Mdot
.18254 .14166 Mdot
.21429 .21973 Mdot
.24603 .29838 Mdot
.27778 .39206 Mdot
.30952 .55182 Mdot
.34127 .59415 Mdot
.37302 .60332 Mdot
.40476 .45788 Mdot
.43651 .23388 Mdot
.46825 .19816 Mdot
.5 .13872 Mdot
.53175 .1032 Mdot
.56349 .07659 Mdot
.59524 .07404 Mdot
.62698 .05486 Mdot
.65873 .06007 Mdot
.69048 .05023 Mdot
.72222 .04105 Mdot
.75397 .04029 Mdot
.78571 .03366 Mdot
.81746 .03488 Mdot
.84921 .03217 Mdot
.88095 .02967 Mdot
.9127 .02957 Mdot
.94444 .02557 Mdot
.97619 .02285 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00G@3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oo
g`2LB0030000000307Boo`3oool0oooo0340oooo00H0_g@000000000000000000018W03Ooolb0?oo
o`060;md00000000000000000000B9`0gooo0P3oool01P3o_g@0000007Boo`3oool0oooO09a800<0
000000<0M;oo0?ooo`3oool0;`3oool00`2oM0000000000000040000000304RLg`3oool0oooo02h0
oooo00<0_g@0000000000000100000002@18W=l0oooo0?ooo`3o_g@0000007Boo`3oool0oooO09a8
00030000000307Boo`3oool0oooo02l0oooo00<0oono07@0000000000P0000000`1d_ol0oooo0?oo
o`0>0?ooo`00G03oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@00000
07Boo`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`0b0?ooo`030?oOW0180000M;oo
03D0oooo00<0omnL04P0001d_ol0103oool01`3o_g@0000007Boo`3oool0oooO09a8M02oool00P3o
ool00`2oM000M;oo0?ooo`0`0?ooo`030=nLB0000000000000<0000000<004RL0=ooo`3oool0;`3o
ool00`3OW4P00000000000030000000:0018W03Oool0oooo0?noM0000000M;oo0?ooo`3ooml0W4Qd
0;ooo`80oooo00<0_g@007Boo`3oool0;`3oool00`3ogi`0B00004RLg`020?ooo`030;md0018W=l0
oooo00h0oooo001L0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3o
oml0W4Qd0;ooo`0b0?ooo`030?oOW0180000M;oo03D0oooo00<0omnL04P0001d_ol03@3oool00`3o
oml0W4Qd0;ooo`0a0?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0<03oool00`3O
W4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo00T0oooo00<0oooO09a8M02oool0=03oool00`3o
okl0M01809cOo`0>0?ooo`00F`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00/0
oooo00<0_g@007Boo`3oool0<@3oool00`3ogi`0B00007Boo`0e0?ooo`030?oOW0180000M;oo00h0
oooo00<0_g@007Boo`3oool0<@3oool00`3OW4P0M;oo0?ooo`0e0?ooo`030=nLB01d_ol0oooo00h0
oooo00<0_g@007Boo`3oool0<`3oool00`3ookl0M00007Boo`0>0?ooo`00F`3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0oooO09a80018W=l0<P3oool00`3ogi`0B000
07Boo`0e0?ooo`030?oOW0180000M;oo00d0oooo00<0oooO09a80018W=l0<`3oool00`2oM4P0W=oo
0?ooo`0e0?ooo`030;mdB02Lgol0oooo00`0oooo00<0oooO09a80018W=l0=03oool00`3o_g@00000
07Boo`0>0?ooo`00F`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0
gia80000001d_ol0<P3oool00`3ogi`0B00007Boo`0e0?ooo`030?oOW0180000M;oo00d0oooo00<0
gia80000001d_ol0<`3oool00`3ooml0W4Qd0;ooo`0e0?ooo`030?oog`2LB7@0_ooo00`0oooo00<0
gia80000001d_ol0<`3oool0103ogi`0B0000000B02Lgol>0?ooo`00F`3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo0380
oooo00<0omnL04P0001d_ol0=@3oool00`3ogi`0B00007Boo`090?ooo`070=oOg`3oool0oono07@0
00000000001d0;ooo`0d0?ooo`030?noM000M;l0oooo03D0oooo00<0okmd001d_`3oool01`3oool0
1`3Ogml0oooo0?oo_`1d000000000000M02oool0<@3oool01@3o_g@000000000000004P0W=oo00l0
oooo001K0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3OW4P00000
000000000000001809cOoc<0oooo00<0omnL04P0001d_ol0=@3oool00`3ogi`0B00007Boo`090?oo
o`060=nLB000000000000000000004P0W=oo=@3oool00`3ookl0M01809cOo`0e0?ooo`030?oo_`1d
04P0W=oo00L0oooo00H0gia800000000000000000000B02Lgold0?ooo`030;md0000M;l0oooo00l0
oooo001K0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool0103ookl0M000
00000018W=le0?ooo`030?oOW0180000M;oo03D0oooo00<0omnL04P0001d_ol02@3oool0103ookl0
M00000000018W=lg0?ooo`030?oo_`1d0000B9cO03D0oooo00<0oono07@00018W=l01`3oool0103o
okl0M00000000018W=lg0?ooo`030;md0018W=l0oooo00h0oooo001L0?ooo`030=nLB000B9`0gooo
0080oooo00<0_g@0001d_`3oool01`3oool00`2oM000M;oo0?ooo`0e0?ooo`030?oOW0180000M;oo
03D0oooo00<0omnL04P0001d_ol02P3oool00`2oM000M;oo0?ooo`0b0?ooo`030?noW02Lgol0oooo
0080oooo00<0omnL04P00018W=l0<03oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO
00P0oooo00<0_g@007Boo`3oool0<P3oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo
00h0oooo001L0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`060?oOW0180000
00000000000004P0W=oo<03oool01@3ooml0W4P0000000000000M;oo03<0oooo00D0oooO09a80000
0000000007Boo`0:0?ooo`060?oOW018000000000000000004P0W=oo;`3oool0203ooml0W4Q809cO
o`3oool0oono07@0000007@0_ooo<03oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo
203oool01P3ogi`0B000000000000000001809cOoc00oooo00H0okmd04RLg`3oool0oono07@00000
M;l?0?ooo`00G@3oool01@3ookl0M0000000000007@0_ooo00X0oooo00D0_g@00000000000000000
04RLg`0b0?ooo`030?noM0000000M;oo03D0oooo00<0okmd0000001d_ol02`3oool01@2oM0000000
000000000000B9cO0300oooo00H0oooO09a80000000000000000M02ooolb0?ooo`060?oog`2LB000
00000000000007@0_ooo2P3oool01@2oM0000000000000000000B9cO0340oooo00D0gia800000000
0000001d0;ooo`0?0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0
ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?oo
o`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?oo
o`3oool0o`3ooon60?ooo`007P3ooooo000008`000001`3oool002P0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo0P00000@0?ooo`00:03o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo0P00000:0?ooo`@000003`3o
ool002P0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0
oooo03/0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0
oooo01D0oooo0P00000;0?ooo`8000002@3oool4000000P0oooo1P00000>0?ooo`00:03oool00`00
0000oooo0?ooo`3o0?oood80oooo0P0000090?ooo`@000002@3oool4000000L0oooo1P0000070?oo
o`H000003P3oool002P0oooo00<000000?ooo`3oool0o`3ooolY0?ooo`8000005P3oool4000000L0
oooo1P0000070?ooo`H000001P3oool6000000P0oooo1000000?0?ooo`00:03oool00`000000oooo
0?ooo`3o0?ooobP0oooo100000090?ooo`8000002@3oool6000000H0oooo1P0000070?ooo`H00000
1`3oool4000000X0oooo0P00000@0?ooo`00:03oool00`000000oooo0?ooo`3o0?ooobL0oooo1P00
00070?ooo`@00000203oool6000000L0oooo100000090?ooo`@000002@3oool2000001d0oooo000X
0?ooo`030000003oool0oooo0?l0oooo403oool2000000X0oooo0P0000090?ooo`H000001P3oool6
000000P0oooo100000090?ooo`8000002`3oool2000002T0oooo000X0?ooo`030000003oool0oooo
0?l0oooo3`3oool4000000P0oooo100000090?ooo`@000001`3oool6000000T0oooo0P0000120?oo
o`00:03oool00`000000oooo0?ooo`3o0?ooo`h0oooo1P0000060?ooo`H000002@3oool2000000T0
oooo1000001>0?ooo`00:03oool00`000000oooo0?ooo`3o0?ooo`h0oooo1P0000060?ooo`H00000
5@3oool2000004l0oooo000X0?ooo`030000003oool0oooo0?l0oooo0`3oool2000000X0oooo1000
00080?ooo`@00000I`3oool002L0oooo0`00003Z0?ooo`8000005P3oool4000000X0oooo0P00000:
0?ooo`800000J03oool002H0oooo1000003Y0?ooo`@00000503oool6000007l0oooo000U0?ooo`H0
0000i`3oool6000000T0oooo0P0000080?ooo`H00000O`3oool002D0oooo1P00003W0?ooo`H00000
203oool4000000P0oooo100000200?ooo`009P3oool400000>T0oooo100000080?ooo`H00000203o
ool200000840oooo000W0?ooo`800000j`3oool2000000T0oooo1P00002;0?ooo`00:03oool00`00
0000oooo0?ooo`090?ooo`800000jP3oool4000008`0oooo000X0?ooo`030000003oool0oooo00P0
oooo1000003Z0?ooo`800000S@3oool002P0oooo00<000000?ooo`3oool01`3oool6000001D0oooo
0P00002i0?ooo`800000YP3oool002P0oooo00<000000?ooo`3oool01`3oool6000000P0oooo0P00
000:0?ooo`@00000[03oool2000000T0oooo1000002U0?ooo`00:03oool00`000000oooo0?ooo`08
0?ooo`@00000203oool4000000P0oooo1P00002Z0?ooo`@000001`3oool600000:@0oooo000X0?oo
o`030000003oool0oooo00T0oooo0P0000080?ooo`H000001`3oool600000:T0oooo1P0000060?oo
o`H00000Y03oool002P0oooo00<000000?ooo`3oool04`3oool6000000P0oooo1000002Z0?ooo`H0
00001`3oool400000:D0oooo000X0?ooo`8000005@3oool4000000X0oooo0P00002/0?ooo`@00000
2@3oool200000:H0oooo000X0?ooo`030000003oool0oooo01D0oooo0P00002j0?ooo`800000/P3o
ool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?ooo`3o0?oo
ohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0
o`3ooon60?ooo`00:03oool00`000000oooo0?ooo`340?ooo`800000_`3oool002P0oooo00<00000
0?ooo`3oool0``3oool400000;h0oooo000X0?ooo`030000003oool0oooo02h0oooo0P00002B0?oo
o`H00000_@3oool002P0oooo00<000000?ooo`3oool0;@3oool400000940oooo1P00002m0?ooo`00
:03oool00`000000oooo0?ooo`0/0?ooo`H00000T@3oool400000;h0oooo000X0?ooo`030000003o
ool0oooo02`0oooo1P00002B0?ooo`800000_`3oool002P0oooo00<000000?ooo`3oool0;@3oool4
00000?l0ooooE@3oool002P0oooo0P00000_0?ooo`800000o`3ooomF0?ooo`00:03oool00`000000
oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<0
00000?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?oo
o`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00
:03oool00`000000oooo0?ooo`2h0?ooo`800000b`3oool000/0oooo00D0omnL04P000000000001d
0;ooo`030?ooo`050?noM0000000M;oo0?ooo`2oM0001P0000000`18W=l0oooo0?ooo`070?ooo`03
0000003oool0oooo0;L0oooo1000003:0?ooo`002P3oool01`3ookl0M00007Boo`3oool0okmd0000
M02oool00P3oool00`3o_g@0000007Boo`020?ooo`030=nLB0000000000000<0000000<004RL0=oo
o`3oool01`3oool00`000000oooo0?ooo`0k0?ooo`800000N@3oool600000<T0oooo000:0?ooo`03
0;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`3OW4P0B9cO0?ooo`020?ooo`03
0=nLM02Lgol0oooo00L0oooo00<000000?ooo`3oool0>P3oool4000007P0oooo1P0000390?ooo`00
2@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00P0oooo00<0gia807Boo`3oool0
2`3oool00`000000oooo0?ooo`0i0?ooo`H00000N03oool400000<X0oooo00090?ooo`030?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol02@3oool00`2oM4P0W=oo0?ooo`0:0?ooo`030000003o
ool0oooo03T0oooo1P00001i0?ooo`800000b`3oool000T0oooo00<0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`090?ooo`030?oog`2LB7@0_ooo00X0oooo0`00000j0?ooo`@00000o`3ooom8
0?ooo`002@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0okmd001d
_`3oool02@3oool00`000000oooo0?ooo`0k0?ooo`800000o`3ooom90?ooo`002@3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0oono07@0B02Lgol02@3oool00`000000
oooo0?ooo`3o0?ooohH0oooo00090?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0
2P3oool00`3ookl0M00004RLg`090?ooo`030000003oool0oooo0?l0ooooQP3oool000X0oooo00<0
gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?noW02Lgol0oooo0080oooo00<0
omnL04P00018W=l02@3oool00`000000oooo0?ooo`3o0?ooohH0oooo000:0?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`060?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool9
0?ooo`030000003oool0oooo0?l0ooooQP3oool000/0oooo00D0oono07@000000000001d0;ooo`08
0?ooo`060?oog`2LB00000000000000007@0_ooo2P3oool00`000000oooo0?ooo`3o0?ooohH0oooo
000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon6
0?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0
ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?oo
o`3o0?ooohH0oooo000X0?ooo`800000o`3ooon70?ooo`00:03oool00`000000oooo0?ooo`3o0?oo
ohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0
o`3ooon60?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0
oooo0:/0oooo0P00003H0?ooo`00:03oool00`000000oooo0?ooo`2Z0?ooo`@00000e`3oool002P0
oooo00<000000?ooo`3oool0Z@3oool600000=H0oooo000X0?ooo`030000003oool0oooo0:T0oooo
1P00003F0?ooo`00:03oool00`000000oooo0?ooo`2Z0?ooo`@00000e`3oool002P0oooo00<00000
0?ooo`3oool0Z`3oool200000=P0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0
oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool2000004P0oooo0P00003o0?ooocd0oooo
000X0?ooo`030000003oool0oooo04H0oooo1000003o0?oooc`0oooo000X0?ooo`030000003oool0
oooo04D0oooo1P00003o0?oooc/0oooo000X0?ooo`030000003oool0oooo04D0oooo1P00003o0?oo
oc/0oooo000X0?ooo`030000003oool0oooo04H0oooo1000003o0?oooc`0oooo000X0?ooo`030000
003oool0oooo04L0oooo0P00003o0?ooocd0oooo000X0?ooo`030000003oool0oooo09l0oooo0P00
003T0?ooo`00:03oool00`000000oooo0?ooo`2N0?ooo`@00000h`3oool002P0oooo00<000000?oo
o`3oool0W@3oool600000>80oooo000X0?ooo`030000003oool0oooo09d0oooo1P00003R0?ooo`00
:03oool00`000000oooo0?ooo`2N0?ooo`@00000h`3oool002P0oooo00<000000?ooo`3oool0W`3o
ool200000>@0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo0P00003o0?oo
ohL0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0
o`3ooon60?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0
oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000
oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool000/0oooo00D0
omnL04P000000000001d0;ooo`030?ooo`030?noM0000000M;oo00D0oooo00@0oooO09a80000B9`0
gooo2@3oool00`000000oooo0?ooo`3o0?ooohH0oooo000:0?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`020?ooo`030?noM0000000M;oo00D0oooo00@0oooO09a80000B9`0gooo2@3oool0
0`000000oooo0?ooo`3o0?ooohH0oooo000:0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d
_`3oool02@3oool0103ooml0W4P00018W03Oool90?ooo`030000003oool0oooo0?l0ooooQP3oool0
00T0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`050?ooo`030;md000000000000
00D0000000<0M;oo0?ooo`3oool01P3oool00`000000oooo0?ooo`3o0?ooohH0oooo00090?ooo`03
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01@3oool0203ogi`0M9cO0?ooo`3oool0oooO
09a80000B9`0gooo2@3oool00`000000oooo0?ooo`3o0?ooohH0oooo00090?ooo`030?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01`3OW4P0M;oo0?ooo`3ooml0W4P00018W03Oool0
2@3oool300000?l0ooooQP3oool000T0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`060?ooo`070?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`090?ooo`030000003oool0oooo0?l0
ooooQP3oool000T0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`060?oO
W018M;l0oooO09a80000B9`0gooo2@3oool00`000000oooo0?ooo`3o0?ooohH0oooo00090?ooo`03
0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool01@3OW4P0M;oO09a80000B9`0gooo
00T0oooo00<000000?ooo`3oool0o`3ooon60?ooo`002P3oool00`3OW4P004RL0=ooo`020?ooo`03
0;md0000M;l0oooo00P0oooo00D0oooO09a8B018B00004RL0=ooo`090?ooo`030000003oool0oooo
0?l0ooooQP3oool000X0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00X0oooo00@0omnL
04P00000B9`0gooo2@3oool00`000000oooo0?ooo`1D0?ooo`800000o`3oool`0?ooo`002`3oool0
1@3ookl0M0000000000007@0_ooo00`0oooo00<0gia80018W03Oool02@3oool00`000000oooo0?oo
o`1C0?ooo`@00000o`3oool_0?ooo`00:03oool00`000000oooo0?ooo`1B0?ooo`H00000o`3oool^
0?ooo`00:03oool00`000000oooo0?ooo`1B0?ooo`H00000o`3oool^0?ooo`00:03oool00`000000
oooo0?ooo`1C0?ooo`@00000o`3oool_0?ooo`00:03oool00`000000oooo0?ooo`1D0?ooo`800000
o`3oool`0?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0
oooo0?l0ooooQP3oool002P0oooo0P00003o0?ooohL0oooo000X0?ooo`030000003oool0oooo0?l0
ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?oo
o`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?oo
o`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000
003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool0
0`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0
oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo
000X0?ooo`800000o`3ooon70?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?oo
o`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00
:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3o
ool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?ooo`3o0?oo
ohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0
o`3ooon60?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0
oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool200000?l0
ooooQ`3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?oo
o`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0600oooo0P00003o0?ooob@0oooo000X0?oo
o`030000003oool0oooo05l0oooo1000003o0?ooob<0oooo000X0?ooo`030000003oool0oooo05h0
oooo1P00003o0?ooob80oooo000X0?ooo`030000003oool0oooo05h0oooo1P00003o0?ooob80oooo
000X0?ooo`030000003oool0oooo05l0oooo1000003o0?ooob<0oooo000;0?ooo`050?oOW0180000
00000000M02oool00`3oool00`3o_g@0000007Boo`030?ooo`050?oOW018000000000000001d_ol0
2P3oool00`000000oooo0?ooo`1P0?ooo`800000o`3ooolT0?ooo`002P3oool01`3ookl0M00007Bo
o`3oool0okmd0000M02oool00P3oool00`3o_g@0000007Boo`020?ooo`070?oOW0180000B9cO0?oo
o`3ookl0M00007Boo`090?ooo`030000003oool0oooo0?l0ooooQP3oool000X0oooo00<0_g@00018
W03Oool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?oog`2LB00007Bo00<0oooo00<0okmd0018
W03Oool0203oool00`000000oooo0?ooo`3o0?ooohH0oooo00090?ooo`030?oo_`1d0000B9cO00<0
oooo00<0gia80000M02oool01@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00P0
oooo00<000000?ooo`3oool0o`3ooon60?ooo`002@3oool00`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00D0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`080?ooo`<00000
o`3ooon60?ooo`002@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00D0oooo00<0
omnL04P0001d_ol00`3oool00`2oM000001809cOo`080?ooo`030000003oool0oooo0?l0ooooQP3o
ool000T0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`050?ooo`080?oo_`1d0000
000004RLg`3oool0okmd00000000M;l90?ooo`030000003oool0oooo0?l0ooooQP3oool000T0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`030;md0000B9`0_ia800800000
00<004RL0=ooo`3oool0203oool00`000000oooo0?ooo`3o0?ooohH0oooo00090?ooo`030?oo_`1d
0000B9cO00<0oooo00<0gia80000M02oool01P3oool00`3ogi`0B00007Boo`0=0?ooo`030000003o
ool0oooo0?l0ooooQP3oool000X0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`07
0?ooo`030=nLB00004P0W=oo00`0oooo00<000000?ooo`3oool0o`3ooon60?ooo`002P3oool01`3o
okl0M00007Boo`3oool0okmd0000M02oool02@3oool00`3o_g@00000001d_`0;0?ooo`030000003o
ool0oooo0?l0ooooQP3oool000/0oooo00D0oono07@000000000001d0;ooo`0;0?ooo`050?oog`2L
B00000000000M02oool0203oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003o
ool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool00`00
0000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo
00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool200000?l0ooooQ`3oool002P0oooo00<00000
0?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`03
0000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0TP3oool200000?40oooo
000X0?ooo`030000003oool0oooo0940oooo1000003`0?ooo`00:03oool00`000000oooo0?ooo`2@
0?ooo`H00000k`3oool002P0oooo00<000000?ooo`3oool0T03oool600000>l0oooo000X0?ooo`03
0000003oool0oooo0940oooo1000003`0?ooo`00:03oool00`000000oooo0?ooo`2B0?ooo`800000
l@3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?ooo`3o
0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo0P00003o0?ooohL0
oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3o
oon60?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo
0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo
0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<00000
0?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`03
0000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03o
ool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`800000o`3ooon70?ooo`00:03oool00`00
0000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo
00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X
0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?oo
o`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000;0?ooo`050?oOW018000000000000M02o
ool00`3oool00`3o_g@0000007Boo`030?ooo`050=nLB0000000000000000018W=l02P3oool00`00
0000oooo0?ooo`3o0?ooohH0oooo000:0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`02
0?ooo`030?noM0000000M;oo0080oooo00L0gia80000001d_ol0oooo0?oog`2LB000B9cO00T0oooo
00<000000?ooo`3oool0o`3ooon60?ooo`002P3oool00`2oM00004RL0=ooo`020?ooo`030;md0000
M;l0oooo00D0oooo00<0oooO09a80000M;l00`3oool00`3o_g@004RL0=ooo`080?ooo`030000003o
ool0oooo0?l0ooooQP3oool000T0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`06
0?ooo`030;md0000M;l0oooo0080oooo00<0_g@00018W03Oool0203oool00`000000oooo0?ooo`3o
0?ooohH0oooo00090?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01`3o
gi`0B01d0;ooo`3oool0_g@00000001d_ol02@3oool00`000000oooo0?ooo`3o0?ooohH0oooo0009
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool01@3ookl0M4Q804P00000
0000B9cO00X0oooo0`00003o0?ooohH0oooo00090?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol01`3oool0103ogi`0B00000000000M;l;0?ooo`030000003oool0oooo0?l0ooooQP3o
ool000T0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`060?oOW0180000
001d0;ooo`3o_g@007Bo2P3oool00`000000oooo0?ooo`1]0?ooo`800000o`3ooolG0?ooo`002@3o
ool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00<0_g@00000M02oool00P3o
ool00`2oM000B9cO0?ooo`080?ooo`030000003oool0oooo06`0oooo1000003o0?oooaH0oooo000:
0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01P3oool00`2oM00004RL0=ooo`02
0?ooo`030=nLB000M;l0oooo00P0oooo00<000000?ooo`3oool0J`3oool600000?l0oooo5@3oool0
00X0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00L0oooo00L0okmd0000001d_ol0oooo
0?oo_`1d0000M;oo00T0oooo00<000000?ooo`3oool0J`3oool600000?l0oooo5@3oool000/0oooo
00D0oono07@000000000001d0;ooo`090?ooo`050?noM0000000000000000018W=l02P3oool00`00
0000oooo0?ooo`1/0?ooo`@00000o`3ooolF0?ooo`00:03oool00`000000oooo0?ooo`1]0?ooo`80
0000o`3ooolG0?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003o
ool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool00`00
0000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo
0P00003o0?ooohL0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<00000
0?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`03
0000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03o
ool00`000000oooo0?ooo`1i0?ooo`800000o`3oool;0?ooo`00:03oool00`000000oooo0?ooo`1h
0?ooo`@00000o`3oool:0?ooo`00:03oool00`000000oooo0?ooo`1g0?ooo`H00000o`3oool90?oo
o`00:03oool00`000000oooo0?ooo`1g0?ooo`H000002@3oool200000?d0oooo000X0?ooo`030000
003oool0oooo07P0oooo100000090?ooo`@00000o03oool002P0oooo00<000000?ooo`3oool0N@3o
ool2000000T0oooo1P00003k0?ooo`00:03oool00`000000oooo0?ooo`240?ooo`H00000n`3oool0
02P0oooo0P0000260?ooo`@00000o03oool002P0oooo00<000000?ooo`3oool0QP3oool200000?d0
oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3o
oon60?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.321202, -0.109853, \
0.0120066, 0.00579195}}]
}, Closed]],

Cell["It is useful to superimpose the theoretical curve as well. ", "Text"],

Cell[BoxData[
    \(\(plot[bwCurve]\  = \ 
        Plot[bw[x, 1, 1\/2], {x, 0, 3}, 
          DisplayFunction \[Rule] Identity];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[{plot2[bwData2], plot[bwCurve]}, \ 
        DisplayFunction \[Rule] $DisplayFunction];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.31746 0.0147151 0.588631 [
[.18254 .00222 -9.5 -15.6875 ]
[.18254 .00222 9.5 0 ]
[.34127 .00222 -5 -15.6875 ]
[.34127 .00222 5 0 ]
[.5 .00222 -9.5 -15.6875 ]
[.5 .00222 9.5 0 ]
[.65873 .00222 -5 -15.6875 ]
[.65873 .00222 5 0 ]
[.81746 .00222 -9.5 -15.6875 ]
[.81746 .00222 9.5 0 ]
[.97619 .00222 -5 -15.6875 ]
[.97619 .00222 5 0 ]
[.01131 .13244 -19 -7.84375 ]
[.01131 .13244 0 7.84375 ]
[.01131 .25017 -19 -7.84375 ]
[.01131 .25017 0 7.84375 ]
[.01131 .36789 -19 -7.84375 ]
[.01131 .36789 0 7.84375 ]
[.01131 .48562 -19 -7.84375 ]
[.01131 .48562 0 7.84375 ]
[.01131 .60335 -10 -7.84375 ]
[.01131 .60335 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.18254 .01472 m
.18254 .02097 L
s
gsave
.18254 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.34127 .01472 m
.34127 .02097 L
s
gsave
.34127 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .01472 m
.5 .02097 L
s
gsave
.5 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.5) show
1.000 setlinewidth
grestore
.65873 .01472 m
.65873 .02097 L
s
gsave
.65873 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.81746 .01472 m
.81746 .02097 L
s
gsave
.81746 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2.5) show
1.000 setlinewidth
grestore
.97619 .01472 m
.97619 .02097 L
s
gsave
.97619 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.05556 .01472 m
.05556 .01847 L
s
.0873 .01472 m
.0873 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.15079 .01472 m
.15079 .01847 L
s
.21429 .01472 m
.21429 .01847 L
s
.24603 .01472 m
.24603 .01847 L
s
.27778 .01472 m
.27778 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.37302 .01472 m
.37302 .01847 L
s
.40476 .01472 m
.40476 .01847 L
s
.43651 .01472 m
.43651 .01847 L
s
.46825 .01472 m
.46825 .01847 L
s
.53175 .01472 m
.53175 .01847 L
s
.56349 .01472 m
.56349 .01847 L
s
.59524 .01472 m
.59524 .01847 L
s
.62698 .01472 m
.62698 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.72222 .01472 m
.72222 .01847 L
s
.75397 .01472 m
.75397 .01847 L
s
.78571 .01472 m
.78571 .01847 L
s
.84921 .01472 m
.84921 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.9127 .01472 m
.9127 .01847 L
s
.94444 .01472 m
.94444 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .13244 m
.03006 .13244 L
s
gsave
.01131 .13244 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.02381 .25017 m
.03006 .25017 L
s
gsave
.01131 .25017 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.02381 .36789 m
.03006 .36789 L
s
gsave
.01131 .36789 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.02381 .48562 m
.03006 .48562 L
s
gsave
.01131 .48562 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.02381 .60335 m
.03006 .60335 L
s
gsave
.01131 .60335 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04415 m
.02756 .04415 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .10301 m
.02756 .10301 L
s
.02381 .16187 m
.02756 .16187 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22074 m
.02756 .22074 L
s
.02381 .2796 m
.02756 .2796 L
s
.02381 .30903 m
.02756 .30903 L
s
.02381 .33846 m
.02756 .33846 L
s
.02381 .39733 m
.02756 .39733 L
s
.02381 .42676 m
.02756 .42676 L
s
.02381 .45619 m
.02756 .45619 L
s
.02381 .51505 m
.02756 .51505 L
s
.02381 .54448 m
.02756 .54448 L
s
.02381 .57391 m
.02756 .57391 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
4 Mabswid
.02381 .05072 Mdot
.05556 .06383 Mdot
.0873 .06966 Mdot
.11905 .06903 Mdot
.15079 .09645 Mdot
.18254 .12826 Mdot
.21429 .19809 Mdot
.24603 .26844 Mdot
.27778 .35223 Mdot
.30952 .49513 Mdot
.34127 .53299 Mdot
.37302 .54119 Mdot
.40476 .41111 Mdot
.43651 .21075 Mdot
.46825 .1788 Mdot
.5 .12563 Mdot
.53175 .09386 Mdot
.56349 .07006 Mdot
.59524 .06778 Mdot
.62698 .05062 Mdot
.65873 .05528 Mdot
.69048 .04648 Mdot
.72222 .03827 Mdot
.75397 .03759 Mdot
.78571 .03166 Mdot
.81746 .03276 Mdot
.84921 .03033 Mdot
.88095 .0281 Mdot
.9127 .02801 Mdot
.94444 .02443 Mdot
.97619 .02199 Mdot
.5 Mabswid
.02381 .04934 m
.04262 .05354 L
.06244 .05883 L
.08426 .06596 L
.10458 .07421 L
.14264 .09575 L
.16305 .11213 L
.18163 .13137 L
.20078 .15711 L
.22156 .19444 L
.2414 .24255 L
.26243 .31097 L
.30026 .4793 L
.31021 .52517 L
.31967 .56275 L
.32488 .57928 L
.32764 .58647 L
.33057 .59284 L
.33329 .59745 L
.33575 .60051 L
.33693 .60159 L
.33822 .60248 L
.33943 .60303 L
.34055 .6033 L
.34181 .60332 L
.34301 .60306 L
.34408 .60261 L
.34525 .60187 L
.34652 .60079 L
.34786 .59931 L
.35028 .59585 L
.3546 .58719 L
.35937 .57424 L
.37003 .53502 L
.38004 .48996 L
.41872 .31621 L
.43876 .24932 L
.45986 .19679 L
.4796 .16048 L
.49797 .13488 L
.51843 .1131 L
.53702 .09782 L
.55782 .08442 L
.577 .07465 L
.61546 .06022 L
.63423 .05496 L
.65487 .05015 L
.69521 .0429 L
.73403 .03781 L
.77531 .03376 L
Mistroke
.81507 .03078 L
.85331 .02852 L
.89401 .02661 L
.93319 .02511 L
.97483 .02381 L
.97619 .02377 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00H03oool01@3ogi`0B0000000000007@0_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oo
g`2LB0030000000307Boo`3oool0oooo0300oooo00H0_g@000000000000000000018W03Oool`0?oo
o`060;md00000000000000000000B9`0gooo0P3oool01P3o_g@0000007Boo`3oool0oooO09a800<0
000000<0M;oo0?ooo`3oool0;@3oool00`2oM0000000000000040000000304RLg`3oool0oooo02d0
oooo00<0_g@0000000000000100000002@18W=l0oooo0?ooo`3o_g@0000007Boo`3oool0oooO09a8
00030000000307Boo`3oool0oooo02d0oooo00<0oono07@0000000000P0000000`1d_ol0oooo0?oo
o`0C0?ooo`00G`3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@00000
07Boo`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`0a0?ooo`030?oOW0180000M;oo
03<0oooo00<0omnL04P0001d_ol0103oool01`3o_g@0000007Boo`3oool0oooO09a8M02oool00P3o
ool00`2oM000M;oo0?ooo`0^0?ooo`030=nLB0000000000000<0000000<004RL0=ooo`3oool0;P3o
ool00`3OW4P00000000000030000000:0018W03Oool0oooo0?noM0000000M;oo0?ooo`3ooml0W4Qd
0;ooo`80oooo00<0_g@007Boo`3oool0;@3oool00`3ogi`0B00004RLg`020?ooo`030;md0018W=l0
oooo01<0oooo001O0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3o
oml0W4Qd0;ooo`0a0?ooo`030?oOW0180000M;oo03<0oooo00<0omnL04P0001d_ol03@3oool00`3o
oml0W4Qd0;ooo`0_0?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0;`3oool00`3O
W4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo00T0oooo00<0oooO09a8M02oool0<P3oool00`3o
okl0M01809cOo`0C0?ooo`00GP3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00/0
oooo00<0_g@007Boo`3oool0<03oool00`3ogi`0B00007Boo`0c0?ooo`030?oOW0180000M;oo00h0
oooo00<0_g@007Boo`3oool0;`3oool00`3OW4P0M;oo0?ooo`0d0?ooo`030=nLB01d_ol0oooo00h0
oooo00<0_g@007Boo`3oool0<@3oool00`3ookl0M00007Boo`0C0?ooo`00GP3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0oooO09a80018W=l0<@3oool00`3ogi`0B000
07Boo`0c0?ooo`030?oOW0180000M;oo00d0oooo00<0oooO09a80018W=l0<@3oool00`2oM4P0W=oo
0?ooo`0d0?ooo`030;mdB02Lgol0oooo00`0oooo00<0oooO09a80018W=l0<P3oool00`3o_g@00000
07Boo`0C0?ooo`00GP3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0
gia80000001d_ol0<@3oool00`3ogi`0B00007Boo`0c0?ooo`030?oOW0180000M;oo00d0oooo00<0
gia80000001d_ol0<@3oool00`3ooml0W4Qd0;ooo`0d0?ooo`030?oog`2LB7@0_ooo00`0oooo00<0
gia80000001d_ol0<@3oool0103ogi`0B0000000B02LgolC0?ooo`00GP3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo0340
oooo00<0omnL04P0001d_ol0<`3oool00`3ogi`0B00007Boo`090?ooo`070=oOg`3oool0oono07@0
00000000001d0;ooo`0b0?ooo`030?noM000M;l0oooo03@0oooo00<0okmd001d_`3oool01`3oool0
1`3Ogml0oooo0?oo_`1d000000000000M02oool0;`3oool01@3o_g@000000000000004P0W=oo01@0
oooo001N0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3OW4P00000
000000000000001809cOoc80oooo00<0omnL04P0001d_ol0<`3oool00`3ogi`0B00007Boo`090?oo
o`060=nLB000000000000000000004P0W=oo<`3oool00`3ookl0M01809cOo`0d0?ooo`030?oo_`1d
04P0W=oo00L0oooo00H0gia800000000000000000000B02Lgolb0?ooo`030;md0000M;l0oooo01@0
oooo001N0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool0103ookl0M000
00000018W=ld0?ooo`030?oOW0180000M;oo03<0oooo00<0omnL04P0001d_ol02@3oool0103ookl0
M00000000018W=le0?ooo`030?oo_`1d0000B9cO03@0oooo00<0oono07@00018W=l01`3oool0103o
okl0M00000000018W=le0?ooo`030;md0018W=l0oooo01<0oooo001O0?ooo`030=nLB000B9`0gooo
0080oooo00<0_g@0001d_`3oool01`3oool00`2oM000M;oo0?ooo`0d0?ooo`030?oOW0180000M;oo
03<0oooo00<0omnL04P0001d_ol02P3oool00`2oM000M;oo0?ooo`0`0?ooo`030?noW02Lgol0oooo
0080oooo00<0omnL04P00018W=l0;`3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO
00P0oooo00<0_g@007Boo`3oool0<03oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo
01<0oooo001O0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`060?oOW0180000
00000000000004P0W=oo;`3oool01@3ooml0W4P0000000000000M;oo0340oooo00D0oooO09a80000
0000000007Boo`0:0?ooo`060?oOW018000000000000000004P0W=oo;@3oool0203ooml0W4Q809cO
o`3oool0oono07@0000007@0_ooo;`3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo
203oool01P3ogi`0B000000000000000001809cOobh0oooo00H0okmd04RLg`3oool0oono07@00000
M;lD0?ooo`00H03oool01@3ookl0M0000000000007@0_ooo00X0oooo00D0_g@00000000000000000
04RLg`0a0?ooo`030?noM0000000M;oo03<0oooo00<0okmd0000001d_ol02`3oool01@2oM0000000
000000000000B9cO02h0oooo00H0oooO09a80000000000000000M02ooola0?ooo`060?oog`2LB000
00000000000007@0_ooo2P3oool01@2oM0000000000000000000B9cO02l0oooo00D0gia800000000
0000001d0;ooo`0D0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00;03oool00`000000oooo0?ooo`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0
ooooPP3oool002`0oooo00<000000?ooo`3oool0o`3ooon20?ooo`00;03oool00`000000oooo0?oo
o`3o0?oooh80oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002`0oooo00<000000?oo
o`3oool0o`3ooon20?ooo`008`3ooooo000008800000303oool002`0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`8000002@3oool4000001@0oooo000/0?ooo`030000003o
ool0oooo03X0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0j0?ooo`030000003o
ool0oooo03X0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0/0?ooo`@000001`3o
ool6000001<0oooo000/0?ooo`030000003oool0oooo0?l0oooo>`3oool2000000X0oooo0P00000:
0?ooo`800000203oool6000000030?ooo`000000000000T000004`3oool002`0oooo00<000000?oo
o`3oool0o`3ooolR0?ooo`8000002P3oool2000000X0oooo100000080?ooo`@00000203ooolB0000
00H0oooo1000000D0?ooo`00;03oool00`000000oooo0?ooo`3o0?ooob40oooo100000080?ooo`@0
00001@3ooolQ000000L0oooo100000090?ooo`8000005@3oool002`0oooo00<000000?ooo`3oool0
o`3oool:0?ooo`8000002P3oool2000000P0oooo1P0000000`3oool000000000000=000000<0oooo
1P0000060?ooo`H000001P3oool6000000P0oooo0P00000Q0?ooo`00;03oool00`000000oooo0?oo
o`3o0?ooo`T0oooo100000080?ooo`@000000`3oool;000000D0oooo1P0000080?ooo`@00000203o
ool4000000P0oooo1000000/0?ooo`00;03oool00`000000oooo0?ooo`3o0?ooo`P0oooo1P000006
0?ooo`P000001@3oool4000000P0oooo1000000:0?ooo`8000002P3oool2000000X0oooo0P00000]
0?ooo`00;03oool00`000000oooo0?ooo`3o0?ooo`P0oooo4P0000080?ooo`8000002P3oool20000
0580oooo000/0?ooo`030000003oool0oooo0?d0oooo0P0000040?ooo`T00000203oool4000006T0
oooo000/0?ooo`<00000i03oool2000001H0oooo1`0000060?ooo`8000002P3oool2000006X0oooo
000[0?ooo`@00000h`3oool400000140oooo2@0000200?ooo`00:P3oool600000>40oooo1P000008
0?ooo`P000000`3oool600000800oooo000Z0?ooo`P00000g`3oool6000000@0oooo1`0000090?oo
o`@00000P@3oool002/0oooo100000030?ooo`@00000g03oool900000080oooo1P0000090?ooo`80
0000PP3oool002`0oooo0P0000080?ooo`@00000f03oool4000000L0oooo1P00002=0?ooo`00;03o
ool00`000000oooo0?ooo`080?ooo`H00000d@3oool4000000`0oooo1000002>0?ooo`00;03oool0
0`000000oooo0?ooo`070?ooo`H0000000@0oooo0000000000000000403oool200000;D0oooo0P00
00030?ooo`8000004@3oool2000008l0oooo000/0?ooo`030000003oool0oooo00L0oooo1P000004
0?ooo`<0000000<0oooo0000000000002@3oool400000:P0oooo0P0000090?ooo`H00000Y03oool0
02`0oooo00<000000?ooo`3oool0203oool4000000P0oooo100000070?ooo`H00000YP3oool40000
00L0oooo1P00002U0?ooo`00;03oool00`000000oooo0?ooo`090?ooo`800000203oool6000000H0
oooo1P00002U0?ooo`H000001P3oool600000:D0oooo000/0?ooo`030000003oool0oooo01<0oooo
200000050?ooo`@00000YP3oool6000000@0oooo0P0000000`3oool000000000000200000:H0oooo
000/0?ooo`<00000503oool4000000<0oooo0P0000040?ooo`800000Z03oool4000000<0oooo0P00
00040?ooo`800000Y`3oool002`0oooo00<000000?ooo`3oool05@3oool2000000H0oooo0P00002]
0?ooo`8000000P3oool200000:l0oooo000/0?ooo`030000003oool0oooo01l0oooo0P00002]0?oo
o`800000/@3oool002`0oooo00<000000?ooo`3oool08@3oool200000:T0oooo0P00002c0?ooo`00
;03oool00`000000oooo0?ooo`0S0?ooo`800000Y@3oool200000;D0oooo000/0?ooo`030000003o
ool0oooo02D0oooo0P00002I0?ooo`8000001P3oool200000;L0oooo000/0?ooo`030000003oool0
oooo02L0oooo0P0000050?ooo`800000S`3oool4000000@0oooo00<000000?ooo`3oool0]`3oool0
02`0oooo00<000000?ooo`3oool0:@3oool200000080oooo1000002=0?ooo`H0000000<0oooo0000
00000000^P3oool002`0oooo00<000000?ooo`3oool0:`3oool7000008`0oooo1`00002l0?ooo`00
;03oool00`000000oooo0?ooo`0/0?ooo`H00000S@3oool400000;h0oooo000/0?ooo`030000003o
ool0oooo02d0oooo1000002?0?ooo`800000_`3oool002`0oooo0`00000^0?ooo`800000T03oool0
0`000000oooo0?ooo`2n0?ooo`00;03oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo
08/0oooo0P0000310?ooo`00;03oool00`000000oooo0?ooo`0a0?ooo`800000RP3oool00`000000
oooo0?ooo`310?ooo`00;03oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo08H0oooo
00<000000?ooo`3oool0`P3oool002`0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?oo
o`240?ooo`030000003oool0oooo0<<0oooo000/0?ooo`030000003oool0oooo03D0oooo00<00000
0?ooo`3oool0PP3oool00`000000oooo0?ooo`340?ooo`00403oool01@3ogi`0B0000000000007@0
_ooo00<0oooo00D0okmd0000001d_ol0oooo0;md00060000000304RLg`3oool0oooo00H0oooo00<0
00000?ooo`3oool0=P3oool00`000000oooo0?ooo`1k0?ooo`8000000`3oool00`000000oooo0?oo
o`350?ooo`003`3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool00`3o_g@00000
07Boo`020?ooo`030=nLB0000000000000<0000000<004RL0=ooo`3oool01P3oool00`000000oooo
0?ooo`0f0?ooo`040000003oool0oooo0?ooo`800000M`3oool4000000030?ooo`000000oooo0<L0
oooo000?0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`3OW4P0B9cO
0?ooo`020?ooo`030=nLM02Lgol0oooo00H0oooo00<000000?ooo`3oool0=`3oool00`000000oooo
00000003000007D0oooo1P0000390?ooo`003P3oool00`3ookl0M00004RLg`030?ooo`030=nLB000
07@0_ooo00P0oooo00<0gia807Boo`3oool02P3oool00`000000oooo0?ooo`0h0?ooo`H00000M03o
ool600000<T0oooo000>0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02@3oool0
0`2oM4P0W=oo0?ooo`090?ooo`030000003oool0oooo03P0oooo1P00001e0?ooo`@00000bP3oool0
00h0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`090?ooo`030?oog`2LB7@0_ooo
00T0oooo1000000h0?ooo`@00000M`3oool200000</0oooo000>0?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol02P3oool00`3o_g@007Bo0?ooo`080?ooo`030000003oool0oooo03X0
oooo0P00001g0?ooo`030000003oool0oooo0</0oooo000>0?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol02P3oool00`3ookl0M01809cOo`080?ooo`030000003oool0oooo03/0oooo
00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`3<0?ooo`003P3oool00`3ookl0M00004RL
g`030?ooo`030=nLB00007@0_ooo00X0oooo00<0oono07@00018W=l0203oool00`000000oooo0?oo
o`0l0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool0c03oool000l0oooo00<0gia8
0018W03Oool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?noW02Lgol0oooo0080oooo00<0omnL
04P00018W=l0203oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo0740oooo00<00000
0?ooo`3oool0c@3oool000l0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00H0oooo00P0
oooO09a8B02Lgol0oooo0?oo_`1d0000001d0;ooo`P0oooo00<000000?ooo`3oool0?@3oool00`00
0000oooo0?ooo`1`0?ooo`030000003oool0oooo0<h0oooo000@0?ooo`050?oo_`1d000000000000
M02oool0203oool01P3ooml0W4P0000000000000001d0;ooo`T0oooo00<000000?ooo`3oool0?P3o
ool00`000000oooo0?ooo`1^0?ooo`030000003oool0oooo0<l0oooo000/0?ooo`030000003oool0
oooo03l0oooo00<000000?ooo`3oool0K@3oool00`000000oooo0?ooo`3?0?ooo`00;03oool00`00
0000oooo0?ooo`100?ooo`030000003oool0oooo06/0oooo00<000000?ooo`3oool0d03oool002`0
oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`1Z0?ooo`030000003oool0oooo0=40
oooo000/0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool0J@3oool00`000000oooo
0?ooo`3A0?ooo`00;03oool300000480oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?oo
o`3B0?ooo`00;03oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo06H0oooo00<00000
0?ooo`3oool0d`3oool002`0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`1U0?oo
o`030000003oool0oooo0=<0oooo000/0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3o
ool0I03oool00`000000oooo0?ooo`3D0?ooo`00;03oool00`000000oooo0?ooo`140?ooo`030000
003oool0oooo0600oooo0P0000000`3oool000000?ooo`3E0?ooo`00;03oool00`000000oooo0?oo
o`140?ooo`030000003oool0oooo05l0oooo1000003G0?ooo`00;03oool00`000000oooo0?ooo`15
0?ooo`030000003oool0oooo05d0oooo1P00003F0?ooo`00;03oool00`000000oooo0?ooo`160?oo
o`030000003oool0oooo05`0oooo1P00003F0?ooo`00;03oool00`000000oooo0?ooo`160?ooo`03
0000003oool0oooo05d0oooo1000003G0?ooo`00;03oool00`000000oooo0?ooo`170?ooo`030000
003oool0oooo05d0oooo0P00003H0?ooo`00;03oool00`000000oooo0?ooo`170?ooo`030000003o
ool0oooo05d0oooo00<000000?ooo`3oool0e`3oool002`0oooo00<000000?ooo`3oool0B03oool0
0`000000oooo0?ooo`1K0?ooo`030000003oool0oooo0=P0oooo000/0?ooo`<00000AP3oool30000
05d0oooo00<000000?ooo`3oool0f03oool002`0oooo00<000000?ooo`3oool0A@3oool5000005/0
oooo00<000000?ooo`3oool0f@3oool002`0oooo00<000000?ooo`3oool0A03oool6000005/0oooo
00<000000?ooo`3oool0f@3oool002`0oooo00<000000?ooo`3oool0A03oool7000005X0oooo00<0
00000?ooo`3oool0f@3oool002`0oooo00<000000?ooo`3oool0A@3oool4000000030?ooo`000000
oooo05P0oooo00<000000?ooo`3oool0fP3oool002`0oooo00<000000?ooo`3oool0AP3oool20000
00<0oooo00<000000?ooo`3oool0C@3oool2000000L0oooo00<000000?ooo`3oool0fP3oool002`0
oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`1<0?ooo`@000001@3oool00`000000
oooo0?ooo`3K0?ooo`00;03oool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo04X0oooo
1P0000040?ooo`030000003oool0oooo0=/0oooo000/0?ooo`030000003oool0oooo04`0oooo00<0
00000?ooo`3oool0BP3oool6000000@0oooo00<000000?ooo`3oool0f`3oool002`0oooo00<00000
0?ooo`3oool0C@3oool00`000000oooo0?ooo`1:0?ooo`@00000103oool00`000000oooo0?ooo`3L
0?ooo`00;03oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo04/0oooo0P0000050?oo
o`030000003oool0oooo0=`0oooo000/0?ooo`<00000C@3oool00`000000oooo0?ooo`1A0?ooo`03
0000003oool0oooo0=d0oooo000/0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool0
D03oool00`000000oooo0?ooo`3M0?ooo`00;03oool00`000000oooo0?ooo`1>0?ooo`030000003o
ool0oooo0500oooo00<000000?ooo`3oool0g@3oool002`0oooo00<000000?ooo`3oool0C`3oool0
0`000000oooo0?ooo`1>0?ooo`030000003oool0oooo0=h0oooo000/0?ooo`030000003oool0oooo
04l0oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`3N0?ooo`00;03oool00`000000
oooo0?ooo`1@0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0g`3oool00100oooo
00D0omnL04P000000000001d0;ooo`030?ooo`030?noM0000000M;oo00D0oooo00@0oooO09a80000
B9`0gooo203oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo04`0oooo00<000000?oo
o`3oool0g`3oool000l0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00<0okmd
0000001d_ol01@3oool0103ooml0W4P00018W03Oool80?ooo`030000003oool0oooo0540oooo00<0
00000?ooo`3oool0B`3oool00`000000oooo0?ooo`3O0?ooo`003`3oool00`2oM00004RL0=ooo`02
0?ooo`030;md0000M;l0oooo00T0oooo00@0oooO09a80000B9`0gooo203oool00`000000oooo0?oo
o`1A0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0h03oool000h0oooo00<0oono
07@00018W=l00`3oool00`3OW4P0001d0;ooo`050?ooo`030;md00000000000000D0000000<0M;oo
0?ooo`3oool01@3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo04X0oooo00<00000
0?ooo`3oool0h03oool000h0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`050?oo
o`080?oOW01dW=l0oooo0?ooo`3ooml0W4P00018W03Oool80?ooo`030000003oool0oooo0580oooo
00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`3Q0?ooo`003P3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00H0oooo00L0gia807Boo`3oool0oooO09a80000B9`0gooo00P0
oooo1000001A0?ooo`030000003oool0oooo04P0oooo00<000000?ooo`3oool0h@3oool000h0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`070?oog`2LB7@0_ooo0?oog`2L
B00004RL0=ooo`080?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0B03oool00`00
0000oooo0?ooo`3Q0?ooo`003P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0
oooo00H0omnL04Qd_`3ooml0W4P00018W03Oool80?ooo`030000003oool0oooo0580oooo00<00000
0?ooo`3oool0A`3oool00`000000oooo0?ooo`3R0?ooo`003P3oool00`3ookl0M00004RLg`030?oo
o`030=nLB00007@0_ooo00P0oooo00D0gia807Bog`2LB00004RL0=ooo`080?ooo`030000003oool0
oooo05<0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`3R0?ooo`003`3oool00`3O
W4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00P0oooo00D0oooO09a8B018B00004RL0=ooo`08
0?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`3R
0?ooo`003`3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool02P3oool0103ogi`0B0000018
W03Oool80?ooo`030000003oool0oooo0580oooo0P0000170?ooo`030000003oool0oooo0><0oooo
000@0?ooo`050?oo_`1d000000000000M02oool0303oool00`3OW4P004RL0=ooo`080?ooo`030000
003oool0oooo0540oooo100000160?ooo`030000003oool0oooo0><0oooo000/0?ooo`030000003o
ool0oooo0500oooo1P0000150?ooo`030000003oool0oooo0><0oooo000/0?ooo`030000003oool0
oooo0500oooo1P0000150?ooo`030000003oool0oooo0><0oooo000/0?ooo`030000003oool0oooo
0540oooo1@0000140?ooo`030000003oool0oooo0>@0oooo000/0?ooo`030000003oool0oooo0580
oooo0P0000000`3oool000000?ooo`130?ooo`030000003oool0oooo0>@0oooo000/0?ooo`030000
003oool0oooo05D0oooo00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`3T0?ooo`00;03o
ool3000005D0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`3U0?ooo`00;03oool0
0`000000oooo0?ooo`1F0?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool0i@3oool0
02`0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo
0>D0oooo000/0?ooo`030000003oool0oooo05H0oooo00<000000?ooo`3oool0?`3oool00`000000
oooo0?ooo`3V0?ooo`00;03oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo03h0oooo
00<000000?ooo`3oool0iP3oool002`0oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?oo
o`0n0?ooo`030000003oool0oooo0>H0oooo000/0?ooo`030000003oool0oooo05L0oooo00<00000
0?ooo`3oool0?@3oool00`000000oooo0?ooo`3W0?ooo`00;03oool00`000000oooo0?ooo`1H0?oo
o`030000003oool0oooo03`0oooo00<000000?ooo`3oool0i`3oool002`0oooo00<000000?ooo`3o
ool0F03oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo0>L0oooo000/0?ooo`030000
003oool0oooo05P0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`3W0?ooo`00;03o
ool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0j03o
ool002`0oooo0`00001I0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0j03oool0
02`0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo
0>P0oooo000/0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3oool0>@3oool00`000000
oooo0?ooo`3Y0?ooo`00;03oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo03T0oooo
00<000000?ooo`3oool0j@3oool002`0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?oo
o`0h0?ooo`030000003oool0oooo0>T0oooo000/0?ooo`030000003oool0oooo05X0oooo00<00000
0?ooo`3oool0>03oool00`000000oooo0?ooo`3Y0?ooo`00;03oool00`000000oooo0?ooo`1J0?oo
o`030000003oool0oooo03L0oooo00<000000?ooo`3oool0jP3oool002`0oooo00<000000?ooo`3o
ool0FP3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo0>X0oooo000/0?ooo`030000
003oool0oooo05/0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`3Z0?ooo`00;03o
ool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0jP3o
ool002`0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0
oooo0>/0oooo000/0?ooo`<00000F`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo
0>/0oooo000/0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0=@3oool00`000000
oooo0?ooo`3[0?ooo`00;03oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo03@0oooo
00<000000?ooo`3oool0j`3oool002`0oooo00<000000?ooo`3oool0G03oool010000000oooo0000
0000000b0?ooo`030000003oool0oooo0>`0oooo000/0?ooo`030000003oool0oooo05`0oooo1@00
000a0?ooo`030000003oool0oooo0>`0oooo000/0?ooo`030000003oool0oooo05`0oooo1P00000`
0?ooo`030000003oool0oooo0>`0oooo000@0?ooo`050?oOW018000000000000M02oool00`3oool0
0`3o_g@0000007Boo`030?ooo`050?oOW018000000000000001d_ol02@3oool00`000000oooo0?oo
o`1L0?ooo`H00000<03oool00`000000oooo0?ooo`3/0?ooo`003`3oool01`3ookl0M00007Boo`3o
ool0okmd0000M02oool00P3oool00`3o_g@0000007Boo`020?ooo`070?oOW0180000B9cO0?ooo`3o
okl0M00007Boo`080?ooo`030000003oool0oooo05d0oooo1000000a0?ooo`030000003oool0oooo
0>`0oooo000?0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool01@3oool00`3ooml0
W4P0001d_`030?ooo`030?noM000B9`0gooo00L0oooo00<000000?ooo`3oool0G@3oool300000340
oooo00<000000?ooo`3oool0k@3oool000h0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d
0;ooo`050?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo
0?ooo`1M0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0k@3oool000h0oooo00<0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`050?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol01`3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo0340oooo00<0
00000?ooo`3oool0k@3oool000h0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`05
0?ooo`030?oOW0180000M;oo00<0oooo00<0_g@00000B02Lgol01`3oool4000005d0oooo00<00000
0?ooo`3oool0<03oool00`000000oooo0?ooo`3]0?ooo`003P3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00D0oooo00P0oono07@000000000B9cO0?ooo`3o_g@00000001d_`P0oooo
00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo0>h0oooo
000>0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool00`2oM00004RL0;nL
B002000000030018W03Oool0oooo00L0oooo00<000000?ooo`3oool0GP3oool00`000000oooo0?oo
o`0_0?ooo`030000003oool0oooo0>h0oooo000>0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia8
0000M02oool01P3oool00`3ogi`0B00007Boo`0<0?ooo`030000003oool0oooo05h0oooo00<00000
0?ooo`3oool0;`3oool00`000000oooo0?ooo`3^0?ooo`003`3oool00`3OW4P004RL0=ooo`020?oo
o`030;md0000M;l0oooo00L0oooo00<0gia80000B02Lgol02`3oool00`000000oooo0?ooo`1N0?oo
o`030000003oool0oooo02l0oooo00<000000?ooo`3oool0kP3oool000l0oooo00L0oono07@0001d
_ol0oooo0?noM00007@0_ooo00T0oooo00<0okmd00000000M;l02P3oool00`000000oooo0?ooo`1O
0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0kP3oool00100oooo00D0oono07@0
00000000001d0;ooo`0;0?ooo`050?oog`2LB00000000000M02oool01`3oool00`000000oooo0?oo
o`1O0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0k`3oool002`0oooo00<00000
0?ooo`3oool0G`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo0>l0oooo000/0?oo
o`030000003oool0oooo05l0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`3_0?oo
o`00;03oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3o
ool0k`3oool002`0oooo00<000000?ooo`3oool0H03oool00`000000oooo0?ooo`0[0?ooo`030000
003oool0oooo0?00oooo000/0?ooo`030000003oool0oooo0600oooo00<000000?ooo`3oool0:`3o
ool00`000000oooo0?ooo`3`0?ooo`00;03oool300000600oooo00<000000?ooo`3oool0:`3oool0
0`000000oooo0?ooo`3`0?ooo`00;03oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo
02/0oooo00<000000?ooo`3oool0l03oool002`0oooo00<000000?ooo`3oool0H@3oool00`000000
oooo0?ooo`0Y0?ooo`040000003oool0000000000?00oooo000/0?ooo`030000003oool0oooo0640
oooo00<000000?ooo`3oool0:@3oool500000>l0oooo000/0?ooo`030000003oool0oooo0640oooo
00<000000?ooo`3oool0:@3oool600000>h0oooo000/0?ooo`030000003oool0oooo0640oooo00<0
00000?ooo`3oool0:@3oool600000>h0oooo000/0?ooo`030000003oool0oooo0680oooo00<00000
0?ooo`3oool0:03oool500000>l0oooo000/0?ooo`030000003oool0oooo0680oooo00<000000?oo
o`3oool09`3oool00`000000oooo0?ooo`0200000?00oooo000/0?ooo`030000003oool0oooo0680
oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`3b0?ooo`00;03oool00`000000oooo
0?ooo`1R0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0lP3oool002`0oooo00<0
00000?ooo`3oool0HP3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo0?80oooo000/
0?ooo`<00000H`3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo0?<0oooo000/0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`3c0?oo
o`00;03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3o
ool0l`3oool002`0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0U0?ooo`030000
003oool0oooo0?<0oooo000/0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool09@3o
ool00`000000oooo0?ooo`3c0?ooo`00;03oool00`000000oooo0?ooo`1T0?ooo`030000003oool0
oooo02<0oooo00<000000?ooo`3oool0m03oool002`0oooo00<000000?ooo`3oool0I03oool00`00
0000oooo0?ooo`0S0?ooo`030000003oool0oooo0?@0oooo000/0?ooo`030000003oool0oooo06@0
oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`3d0?ooo`00;03oool00`000000oooo
0?ooo`1T0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0m03oool002`0oooo00<0
00000?ooo`3oool0I@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0?D0oooo000/
0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`3e
0?ooo`00;03oool3000006D0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`3e0?oo
o`00;03oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3o
ool0m@3oool002`0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`0P0?ooo`030000
003oool0oooo0?H0oooo000/0?ooo`030000003oool0oooo06H0oooo00<000000?ooo`3oool07`3o
ool00`000000oooo0?ooo`3f0?ooo`00;03oool00`000000oooo0?ooo`1V0?ooo`030000003oool0
oooo01l0oooo00<000000?ooo`3oool0mP3oool002`0oooo00<000000?ooo`3oool0IP3oool00`00
0000oooo0?ooo`0O0?ooo`030000003oool0oooo0?H0oooo000/0?ooo`030000003oool0oooo06H0
oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`3f0?ooo`00403oool01@3ogi`0B000
0000000007@0_ooo00<0oooo00<0okmd0000001d_ol00`3oool01@3OW4P00000000000000000B9cO
00T0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo
0?L0oooo000?0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000
M;oo0080oooo00L0gia80000001d_ol0oooo0?oog`2LB000B9cO00P0oooo00<000000?ooo`3oool0
I`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0?L0oooo000?0?ooo`030;md0000
B9`0gooo0080oooo00<0_g@0001d_`3oool01@3oool00`3ooml0W4P0001d_`030?ooo`030?noM000
B9`0gooo00L0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`0M0?ooo`030000003o
ool0oooo0?L0oooo000>0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool0
0`2oM00007Bo0?ooo`020?ooo`030;md0000B9`0gooo00L0oooo00<000000?ooo`3oool0I`3oool0
0`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0?L0oooo000>0?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol01P3oool01`3ogi`0B01d0;ooo`3oool0_g@00000001d_ol0203o
ool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0n03o
ool000h0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`050?oo_`1dB4P0
B00000000018W=l02@3oool4000006L0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?oo
o`3h0?ooo`003P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00@0omnL
04P00000000007Bo2P3oool00`000000oooo0?ooo`1X0?ooo`030000003oool0oooo008000006@3o
ool00`000000oooo0?ooo`3h0?ooo`003P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo00H0oooo00H0omnL04P0000007@0_ooo0?noM000M;l90?ooo`030000003oool0oooo06P0oooo
00<000000?ooo`0000000`00000H0?ooo`030000003oool0oooo0?P0oooo000>0?ooo`030?oo_`1d
0000B9cO00<0oooo00<0gia80000M02oool01P3oool00`2oM000001d0;ooo`020?ooo`030;md0018
W=l0oooo00L0oooo00<000000?ooo`3oool0J03oool7000001L0oooo00<000000?ooo`3oool0n03o
ool000l0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`060?ooo`030;md0000B9`0
gooo0080oooo00<0gia8001d_`3oool01`3oool00`000000oooo0?ooo`1Y0?ooo`H000005P3oool0
0`000000oooo0?ooo`3i0?ooo`003`3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool01`3o
ool01`3o_g@0000007Boo`3oool0oono07@0001d_ol0203oool00`000000oooo0?ooo`1Y0?ooo`D0
00005`3oool00`000000oooo0?ooo`3i0?ooo`00403oool01@3ookl0M0000000000007@0_ooo00T0
oooo00D0okmd000000000000000004RLg`090?ooo`030000003oool0oooo06T0oooo00@000000?oo
o`0000000000603oool00`000000oooo0?ooo`3i0?ooo`00;03oool00`000000oooo0?ooo`1Y0?oo
o`030000003oool0oooo01T0oooo00<000000?ooo`3oool0n@3oool002`0oooo00<000000?ooo`3o
ool0J@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo0?X0oooo000/0?ooo`030000
003oool0oooo06X0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`3j0?ooo`00;03o
ool00`000000oooo0?ooo`1Z0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0nP3o
ool002`0oooo0`00001Z0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0nP3oool0
02`0oooo00<000000?ooo`3oool0JP3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo
0?X0oooo000/0?ooo`030000003oool0oooo06/0oooo00<000000?ooo`3oool05@3oool00`000000
oooo0?ooo`3k0?ooo`00;03oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo01D0oooo
00<000000?ooo`3oool0n`3oool002`0oooo00<000000?ooo`3oool0J`3oool00`000000oooo0?oo
o`090?ooo`8000002P3oool00`000000oooo0?ooo`3k0?ooo`00;03oool00`000000oooo0?ooo`1[
0?ooo`030000003oool0oooo00P0oooo100000090?ooo`030000003oool0oooo0?/0oooo000/0?oo
o`030000003oool0oooo06`0oooo00<000000?ooo`3oool01P3oool6000000L0oooo00<000000?oo
o`3oool0o03oool002`0oooo00<000000?ooo`3oool0K03oool00`000000oooo0?ooo`060?ooo`H0
00001`3oool300000?`0oooo000/0?ooo`030000003oool0oooo06`0oooo00<000000?ooo`3oool0
1`3oool4000000P0oooo1000003k0?ooo`00;03oool00`000000oooo0?ooo`1/0?ooo`030000003o
ool0oooo00P0oooo0P0000080?ooo`H00000nP3oool002`0oooo00<000000?ooo`3oool0K@3oool0
0`000000oooo0?ooo`0A0?ooo`H00000nP3oool002`0oooo0`00001]0?ooo`030000003oool0oooo
0140oooo1@00003k0?ooo`00;03oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo0140
oooo00@000000?ooo`0000000000o03oool002`0oooo00<000000?ooo`3oool0KP3oool00`000000
oooo0?ooo`0@0?ooo`030000003oool0oooo0?d0oooo000/0?ooo`030000003oool0oooo06h0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`3n0?ooo`00;03oool00`000000oooo0?oo
o`1^0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0oP3oool002`0oooo00<00000
0?ooo`3oool0KP3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0?h0oooo000/0?oo
o`030000003oool0oooo06l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`3n0?oo
o`00;03oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3o
ool0o`3oool002`0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`0=0?ooo`030000
003oool0oooo0?l0oooo000/0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`3o0?ooo`00;03oool00`000000oooo0?ooo`1`0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool0o`3oool002`0oooo0`00001`0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool0o`3oool10?ooo`00;03oool00`000000oooo0?ooo`1a0?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool0o`3oool10?ooo`00;03oool00`000000oooo
0?ooo`1a0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0o`3oool10?ooo`00;03o
ool00`000000oooo0?ooo`1a0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0o`3o
ool20?ooo`00;03oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool0o`3oool20?ooo`00;03oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool0o`3oool30?ooo`00;03oool00`000000oooo0?ooo`1b0?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`3oool0o`3oool30?ooo`007P3oool01P2oM0000000
00000000000004RL0=ooo`P0oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo0?l0oooo0`3oool001l0oooo00<0omnL04P0001d_ol02P3oool00`000000
oooo0?ooo`1c0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0o`3oool40?ooo`00
7`3oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`3o0?ooo`@0oooo000O0?ooo`030?oOW0180000M;oo00X0oooo00<0
00000?ooo`3oool0M03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0?l0oooo1@3o
ool001l0oooo00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`1e0?ooo`80000000<0oooo
000000000000o`3oool80?ooo`007`3oool00`3ogi`0B00007Boo`0:0?ooo`@00000M@3oool30000
0?l0oooo2@3oool001l0oooo00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`3o0?oooh80
oooo000O0?ooo`030?oOW0180000M;oo00X0oooo00<000000?ooo`3oool0o`3ooon20?ooo`007`3o
ool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo0?l0ooooPP3oool001l0oooo00<0omnL
04P0001d_ol02P3oool00`000000oooo0?ooo`3o0?oooh80oooo000M0?ooo`050?oog`2LB0000000
0000001d_ol02P3oool00`000000oooo0?ooo`3o0?oooh80oooo000O0?ooo`030?noM0000000M;oo
0?l0ooooS`3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.368965, -0.125454, \
0.0123185, 0.00664359}}]
}, Closed]],

Cell[TextData[{
  "If we were fitting data, we would probably want to generate figures \
similar to the one above except that we would also want to include error \
bars.  For this we need to load a package that extends the number of plotting \
commands available.  In particular, this package includes ",
  ButtonBox["ErrorListPlot",
    ButtonStyle->"RefGuideLink"],
  " which plots a list of points of the form ",
  Cell[BoxData[
      FormBox[
        RowBox[{"{", 
          RowBox[{\(x\_i\), ",", \(y\_i\), ",", 
            SubscriptBox[
              StyleBox["dy",
                FontSlant->"Italic"], "i"]}], "}"}], TraditionalForm]]],
  "."
}], "Text"],

Cell[BoxData[
    \(\(bwData3\  = 
        Table[{x, \ bw[x, 1, 1\/2]\ \((0.8 + 0.4\ Random[])\), 
            0.1\ bw[x, 1, 1\/2]}, {x, 0, 3, 0.1}];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[{ErrorListPlot[bwData3, DisplayFunction \[Rule] Identity], 
          plot[bwCurve]}, 
        DisplayFunction \[Rule] $DisplayFunction];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.31746 0.0147151 0.460247 [
[.18254 .00222 -9.5 -15.6875 ]
[.18254 .00222 9.5 0 ]
[.34127 .00222 -5 -15.6875 ]
[.34127 .00222 5 0 ]
[.5 .00222 -9.5 -15.6875 ]
[.5 .00222 9.5 0 ]
[.65873 .00222 -5 -15.6875 ]
[.65873 .00222 5 0 ]
[.81746 .00222 -9.5 -15.6875 ]
[.81746 .00222 9.5 0 ]
[.97619 .00222 -5 -15.6875 ]
[.97619 .00222 5 0 ]
[.01131 .10676 -19 -7.84375 ]
[.01131 .10676 0 7.84375 ]
[.01131 .19881 -19 -7.84375 ]
[.01131 .19881 0 7.84375 ]
[.01131 .29086 -19 -7.84375 ]
[.01131 .29086 0 7.84375 ]
[.01131 .38291 -19 -7.84375 ]
[.01131 .38291 0 7.84375 ]
[.01131 .47496 -10 -7.84375 ]
[.01131 .47496 0 7.84375 ]
[.01131 .56701 -19 -7.84375 ]
[.01131 .56701 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.18254 .01472 m
.18254 .02097 L
s
gsave
.18254 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.34127 .01472 m
.34127 .02097 L
s
gsave
.34127 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .01472 m
.5 .02097 L
s
gsave
.5 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.5) show
1.000 setlinewidth
grestore
.65873 .01472 m
.65873 .02097 L
s
gsave
.65873 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.81746 .01472 m
.81746 .02097 L
s
gsave
.81746 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2.5) show
1.000 setlinewidth
grestore
.97619 .01472 m
.97619 .02097 L
s
gsave
.97619 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.05556 .01472 m
.05556 .01847 L
s
.0873 .01472 m
.0873 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.15079 .01472 m
.15079 .01847 L
s
.21429 .01472 m
.21429 .01847 L
s
.24603 .01472 m
.24603 .01847 L
s
.27778 .01472 m
.27778 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.37302 .01472 m
.37302 .01847 L
s
.40476 .01472 m
.40476 .01847 L
s
.43651 .01472 m
.43651 .01847 L
s
.46825 .01472 m
.46825 .01847 L
s
.53175 .01472 m
.53175 .01847 L
s
.56349 .01472 m
.56349 .01847 L
s
.59524 .01472 m
.59524 .01847 L
s
.62698 .01472 m
.62698 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.72222 .01472 m
.72222 .01847 L
s
.75397 .01472 m
.75397 .01847 L
s
.78571 .01472 m
.78571 .01847 L
s
.84921 .01472 m
.84921 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.9127 .01472 m
.9127 .01847 L
s
.94444 .01472 m
.94444 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .10676 m
.03006 .10676 L
s
gsave
.01131 .10676 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.02381 .19881 m
.03006 .19881 L
s
gsave
.01131 .19881 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.02381 .29086 m
.03006 .29086 L
s
gsave
.01131 .29086 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.02381 .38291 m
.03006 .38291 L
s
gsave
.01131 .38291 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.02381 .47496 m
.03006 .47496 L
s
gsave
.01131 .47496 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .56701 m
.03006 .56701 L
s
gsave
.01131 .56701 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.2) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .03773 m
.02756 .03773 L
s
.02381 .06074 m
.02756 .06074 L
s
.02381 .08375 m
.02756 .08375 L
s
.02381 .12978 m
.02756 .12978 L
s
.02381 .15279 m
.02756 .15279 L
s
.02381 .1758 m
.02756 .1758 L
s
.02381 .22183 m
.02756 .22183 L
s
.02381 .24484 m
.02756 .24484 L
s
.02381 .26785 m
.02756 .26785 L
s
.02381 .31388 m
.02756 .31388 L
s
.02381 .33689 m
.02756 .33689 L
s
.02381 .3599 m
.02756 .3599 L
s
.02381 .40593 m
.02756 .40593 L
s
.02381 .42894 m
.02756 .42894 L
s
.02381 .45195 m
.02756 .45195 L
s
.02381 .49797 m
.02756 .49797 L
s
.02381 .52099 m
.02756 .52099 L
s
.02381 .544 m
.02756 .544 L
s
.02381 .59002 m
.02756 .59002 L
s
.02381 .61304 m
.02756 .61304 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.002 w
.02381 .03426 m
.02381 .03968 L
s
.015 w
.02381 .03697 Mdot
.002 w
.05556 .04613 m
.05556 .05272 L
s
.015 w
.05556 .04942 Mdot
.002 w
.0873 .04431 m
.0873 .0525 L
s
.015 w
.0873 .0484 Mdot
.002 w
.11905 .068 m
.11905 .07841 L
s
.015 w
.11905 .0732 Mdot
.002 w
.15079 .07494 m
.15079 .08856 L
s
.015 w
.15079 .08175 Mdot
.002 w
.18254 .11121 m
.18254 .12962 L
s
.015 w
.18254 .12042 Mdot
.002 w
.21429 .10652 m
.21429 .13238 L
s
.015 w
.21429 .11945 Mdot
.002 w
.24603 .17147 m
.24603 .20919 L
s
.015 w
.24603 .19033 Mdot
.002 w
.27778 .22307 m
.27778 .27919 L
s
.015 w
.27778 .25113 Mdot
.002 w
.30952 .38743 m
.30952 .46678 L
s
.015 w
.30952 .42711 Mdot
.002 w
.34127 .51127 m
.34127 .60332 L
s
.015 w
.34127 .55729 Mdot
.002 w
.37302 .41854 m
.37302 .49789 L
s
.015 w
.37302 .45821 Mdot
.002 w
.40476 .32086 m
.40476 .37699 L
s
.015 w
.40476 .34892 Mdot
.002 w
.43651 .19726 m
.43651 .23498 L
s
.015 w
.43651 .21612 Mdot
.002 w
.46825 .11557 m
.46825 .14142 L
s
.015 w
.46825 .1285 Mdot
.002 w
.5 .09349 m
.5 .1119 L
s
.015 w
.5 .1027 Mdot
.002 w
.53175 .06928 m
.53175 .08289 L
s
.015 w
.53175 .07609 Mdot
.002 w
.56349 .05824 m
.56349 .06866 L
s
.015 w
.56349 .06345 Mdot
.002 w
.59524 .04936 m
.59524 .05755 L
s
.015 w
.59524 .05346 Mdot
.002 w
.62698 .0458 m
.62698 .0524 L
s
.015 w
.62698 .0491 Mdot
.002 w
.65873 .03983 m
.65873 .04524 L
s
.015 w
.65873 .04253 Mdot
.002 w
.69048 .03377 m
.69048 .03829 L
s
.015 w
.69048 .03603 Mdot
.002 w
.72222 .03533 m
.72222 .03916 L
s
.015 w
.72222 .03724 Mdot
.002 w
.75397 .03013 m
.75397 .03341 L
s
.015 w
.75397 .03177 Mdot
.002 w
.78571 .02759 m
.78571 .03044 L
s
.015 w
.78571 .02902 Mdot
.002 w
.81746 .02703 m
.81746 .02952 L
s
.015 w
.81746 .02828 Mdot
.002 w
.84921 .02628 m
.84921 .02847 L
s
.015 w
.84921 .02737 Mdot
.002 w
.88095 .02461 m
.88095 .02656 L
s
.015 w
.88095 .02558 Mdot
.002 w
.9127 .02099 m
.9127 .02274 L
s
.015 w
.9127 .02187 Mdot
.002 w
.94444 .02288 m
.94444 .02444 L
s
.015 w
.94444 .02366 Mdot
.002 w
.97619 .02211 m
.97619 .02352 L
s
.015 w
.97619 .02281 Mdot
.5 Mabswid
.02381 .04179 m
.04262 .04507 L
.06244 .04921 L
.08426 .05478 L
.10458 .06123 L
.14264 .07808 L
.16305 .09088 L
.18163 .10592 L
.20078 .12606 L
.22156 .15524 L
.2414 .19286 L
.26243 .24635 L
.30026 .37797 L
.31021 .41384 L
.31967 .44322 L
.32488 .45614 L
.32764 .46177 L
.33057 .46675 L
.33329 .47035 L
.33575 .47274 L
.33693 .47359 L
.33822 .47428 L
.33943 .47472 L
.34055 .47492 L
.34181 .47494 L
.34301 .47474 L
.34408 .47439 L
.34525 .47381 L
.34652 .47296 L
.34786 .47181 L
.35028 .4691 L
.3546 .46233 L
.35937 .4522 L
.37003 .42154 L
.38004 .38631 L
.41872 .25045 L
.43876 .19815 L
.45986 .15708 L
.4796 .12869 L
.49797 .10867 L
.51843 .09164 L
.53702 .07969 L
.55782 .06922 L
.577 .06157 L
.61546 .05029 L
.63423 .04618 L
.65487 .04242 L
.69521 .03675 L
.73403 .03277 L
.77531 .02961 L
Mistroke
.81507 .02728 L
.85331 .02551 L
.89401 .02401 L
.93319 .02284 L
.97483 .02183 L
.97619 .0218 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00G@3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oo
g`2LB0030000000307Boo`3oool0oooo0340oooo00H0_g@000000000000000000018W03Ooolb0?oo
o`060;md00000000000000000000B9`0gooo0P3oool01P3o_g@0000007Boo`3oool0oooO09a800<0
000000<0M;oo0?ooo`3oool0;`3oool00`2oM0000000000000040000000304RLg`3oool0oooo02h0
oooo00<0_g@0000000000000100000002@18W=l0oooo0?ooo`3o_g@0000007Boo`3oool0oooO09a8
00030000000307Boo`3oool0oooo02l0oooo00<0oono07@0000000000P0000000`1d_ol0oooo0?oo
o`0>0?ooo`00G03oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@00000
07Boo`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`0b0?ooo`030?oOW0180000M;oo
03D0oooo00<0omnL04P0001d_ol0103oool01`3o_g@0000007Boo`3oool0oooO09a8M02oool00P3o
ool00`2oM000M;oo0?ooo`0`0?ooo`030=nLB0000000000000<0000000<004RL0=ooo`3oool0;`3o
ool00`3OW4P00000000000030000000:0018W03Oool0oooo0?noM0000000M;oo0?ooo`3ooml0W4Qd
0;ooo`80oooo00<0_g@007Boo`3oool0;`3oool00`3ogi`0B00004RLg`020?ooo`030;md0018W=l0
oooo00h0oooo001L0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3o
oml0W4Qd0;ooo`0b0?ooo`030?oOW0180000M;oo03D0oooo00<0omnL04P0001d_ol03@3oool00`3o
oml0W4Qd0;ooo`0a0?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0<03oool00`3O
W4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo00T0oooo00<0oooO09a8M02oool0=03oool00`3o
okl0M01809cOo`0>0?ooo`00F`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00/0
oooo00<0_g@007Boo`3oool0<@3oool00`3ogi`0B00007Boo`0e0?ooo`030?oOW0180000M;oo00h0
oooo00<0_g@007Boo`3oool0<@3oool00`3OW4P0M;oo0?ooo`0e0?ooo`030=nLB01d_ol0oooo00h0
oooo00<0_g@007Boo`3oool0<`3oool00`3ookl0M00007Boo`0>0?ooo`00F`3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0oooO09a80018W=l0<P3oool00`3ogi`0B000
07Boo`0e0?ooo`030?oOW0180000M;oo00d0oooo00<0oooO09a80018W=l0<`3oool00`2oM4P0W=oo
0?ooo`0e0?ooo`030;mdB02Lgol0oooo00`0oooo00<0oooO09a80018W=l0=03oool00`3o_g@00000
07Boo`0>0?ooo`00F`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0
gia80000001d_ol0<P3oool00`3ogi`0B00007Boo`0e0?ooo`030?oOW0180000M;oo00d0oooo00<0
gia80000001d_ol0<`3oool00`3ooml0W4Qd0;ooo`0e0?ooo`030?oog`2LB7@0_ooo00`0oooo00<0
gia80000001d_ol0<`3oool0103ogi`0B0000000B02Lgol>0?ooo`00F`3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo0380
oooo00<0omnL04P0001d_ol0=@3oool00`3ogi`0B00007Boo`090?ooo`070=oOg`3oool0oono07@0
00000000001d0;ooo`0d0?ooo`030?noM000M;l0oooo03D0oooo00<0okmd001d_`3oool01`3oool0
1`3Ogml0oooo0?oo_`1d000000000000M02oool0<@3oool01@3o_g@000000000000004P0W=oo00l0
oooo001K0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3OW4P00000
000000000000001809cOoc<0oooo00<0omnL04P0001d_ol0=@3oool00`3ogi`0B00007Boo`090?oo
o`060=nLB000000000000000000004P0W=oo=@3oool00`3ookl0M01809cOo`0e0?ooo`030?oo_`1d
04P0W=oo00L0oooo00H0gia800000000000000000000B02Lgold0?ooo`030;md0000M;l0oooo00l0
oooo001K0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool0103ookl0M000
00000018W=le0?ooo`030?oOW0180000M;oo03D0oooo00<0omnL04P0001d_ol02@3oool0103ookl0
M00000000018W=lg0?ooo`030?oo_`1d0000B9cO03D0oooo00<0oono07@00018W=l01`3oool0103o
okl0M00000000018W=lg0?ooo`030;md0018W=l0oooo00h0oooo001L0?ooo`030=nLB000B9`0gooo
0080oooo00<0_g@0001d_`3oool01`3oool00`2oM000M;oo0?ooo`0e0?ooo`030?oOW0180000M;oo
03D0oooo00<0omnL04P0001d_ol02P3oool00`2oM000M;oo0?ooo`0b0?ooo`030?noW02Lgol0oooo
0080oooo00<0omnL04P00018W=l0<03oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO
00P0oooo00<0_g@007Boo`3oool0<P3oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo
00h0oooo001L0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`060?oOW0180000
00000000000004P0W=oo<03oool01@3ooml0W4P0000000000000M;oo03<0oooo00D0oooO09a80000
0000000007Boo`0:0?ooo`060?oOW018000000000000000004P0W=oo;`3oool0203ooml0W4Q809cO
o`3oool0oono07@0000007@0_ooo<03oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo
203oool01P3ogi`0B000000000000000001809cOoc00oooo00H0okmd04RLg`3oool0oono07@00000
M;l?0?ooo`00G@3oool01@3ookl0M0000000000007@0_ooo00X0oooo00D0_g@00000000000000000
04RLg`0b0?ooo`030?noM0000000M;oo03D0oooo00<0okmd0000001d_ol02`3oool01@2oM0000000
000000000000B9cO0300oooo00H0oooO09a80000000000000000M02ooolb0?ooo`060?oog`2LB000
00000000000007@0_ooo2P3oool01@2oM0000000000000000000B9cO0340oooo00D0gia800000000
0000001d0;ooo`0?0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00:03oool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0
ooooQP3oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?oo
o`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?oo
o`3oool0o`3ooon60?ooo`007P3ooooo000008`000001`3oool002P0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
203oool3000000/0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0?0?ooo`00:03o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`070?ooo`D00000203oool3000000X0oooo0`00000@0?ooo`00:03oool0
0`000000oooo0?ooo`0k0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0>`3oool0
0`000000oooo0?ooo`0k0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0203oool3
000000X0oooo0`0000080?oooah000003`3oool002P0oooo00<000000?ooo`3oool0o`3ooolX0?oo
o`<000002P3oool3000000D0oooo8@0000070?ooo`D00000203oool5000000l0oooo000X0?ooo`03
0000003oool0oooo0?l0oooo703oool3000000P0oooo1@0000020?ooo`l000000`3oool5000000P0
oooo1@0000080?ooo`<00000203oool5000000P0oooo1@00000?0?ooo`00:03oool00`000000oooo
0?ooo`3o0?oooa/0oooo4`0000060?ooo`D000001`3oool5000000P0oooo1@00000D0?ooo`<00000
2P3oool300000100oooo000V0?ooo`<00000o`3oool50?ooo`<000002@3ooolA000000L0oooo1@00
00080?ooo`D00000203oool3000000X0oooo0`00000e0?ooo`009@3oool500000?l0oooo0`3oool5
000000<0oooo2@0000080?ooo`D00000203oool3000000X0oooo0`00001>0?ooo`009@3oool50000
0?H0oooo0`0000090?ooo`P00000103oool5000000T0oooo0`00001W0?ooo`009@3oool500000?D0
oooo1@0000000`3oool000000000000:000000L0oooo1@00001c0?ooo`009P3oool600000?40oooo
200000080?ooo`<000002@3oool3000007@0oooo000X0?ooo`040000003oool0oooo0?ooo`D00000
0P3oool3000000T0oooo0`00003B0?ooo`T000000P3oool5000008`0oooo000X0?ooo`030000003o
ool0oooo00H0oooo1P0000070?ooo`D00000d03oool5000000P0oooo0`00002=0?ooo`00:03oool0
0`000000oooo0?ooo`070?ooo`L000001@3oool500000<@0oooo0`0000050?ooo`T00000V03oool0
02P0oooo00<000000?ooo`3oool01`3oool500000080oooo100000000`3oool00000000000030000
0<<0oooo1@000000103oool000000000000000040?ooo`D00000V03oool002P0oooo00<000000?oo
o`3oool0203oool3000000L0oooo1@0000340?ooo`H00000203oool3000009T0oooo000X0?ooo`03
0000003oool0oooo01H0oooo1000002g0?ooo`030000003oool0oooo00L0oooo1@00002U0?ooo`00
:03oool2000001/0oooo1000002b0?ooo`<000001@3oool3000000040?ooo`000000000000000:H0
oooo000X0?ooo`030000003oool0oooo01h0oooo0P00002_0?ooo`D0000000@0oooo000000000000
0000[@3oool002P0oooo00<000000?ooo`3oool0803oool200000:d0oooo1P00002`0?ooo`00:03o
ool00`000000oooo0?ooo`0R0?ooo`800000Z`3oool500000;40oooo000X0?ooo`030000003oool0
oooo0240oooo1@00002O0?ooo`030000003oool0oooo00L0oooo1000002b0?ooo`00:03oool00`00
0000oooo0?ooo`0P0?ooo`D0000000<0oooo000000000000V`3oool3000000L0oooo0P00002f0?oo
o`00:03oool00`000000oooo0?ooo`0P0?ooo`D000000`3oool2000000D0oooo00<000000?ooo`3o
ool0T03oool5000000@0oooo0P00002h0?ooo`00:03oool00`000000oooo0?ooo`0P0?ooo`D00000
1@3oool2000000040?ooo`000000000000000980oooo1@0000020?ooo`800000^P3oool002P0oooo
00<000000?ooo`3oool08@3oool3000000P0oooo1@00002A0?ooo`L00000_03oool002P0oooo0P00
000]0?ooo`D00000TP3oool400000;h0oooo000X0?ooo`030000003oool0oooo02`0oooo1@00002B
0?ooo`800000`03oool002P0oooo00<000000?ooo`3oool0;@3oool3000000030?ooo`0000000000
08h0oooo0P0000320?ooo`00:03oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo08T0
oooo0P0000340?ooo`002`3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00D0okmd0000001d
_ol0oooo0;md00060000000304RLg`3oool0oooo00L0oooo00<000000?ooo`3oool0=03oool00`00
0000oooo0?ooo`210?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0a03oool000X0
oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00<0okmd0000001d_ol00P3oool0
0`3OW4P0000000000003000000030018W03Oool0oooo00L0oooo00<000000?ooo`3oool0=@3oool2
00000840oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`350?ooo`002P3oool00`2o
M00004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0gia804RLg`3oool00P3oool00`3O
W7@0W=oo0?ooo`070?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0O@3oool30000
0080oooo00<000000?ooo`3oool0aP3oool000T0oooo00<0oono07@00018W=l00`3oool00`3OW4P0
001d0;ooo`080?ooo`030=nLB01d_ol0oooo00/0oooo00<000000?ooo`3oool0>03oool00`000000
oooo0?ooo`1k0?ooo`H00000b@3oool000T0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`090?ooo`030;mdB02Lgol0oooo00X0oooo00<000000?ooo`3oool0>@3oool2000007/0oooo
1@00003:0?ooo`002@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00<0
oooO09a8M02oool02P3oool3000003/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`1[0?ooo`D00000bP3oool000T0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:
0?ooo`030?noM000M;l0oooo00T0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo06`0oooo0`00003;0?ooo`002@3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00X0oooo00<0oono07@0B02Lgol02@3oool00`000000oooo0?ooo`0k0?oo
o`030000003oool0000000X0oooo00<000000?ooo`3oool0J`3oool00`000000oooo0000003<0?oo
o`002@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00X0oooo00<0oono07@00018
W=l02@3oool00`000000oooo0?ooo`0k0?ooo`040000003oool0oooo000000T0oooo00<000000?oo
o`3oool0JP3oool00`000000oooo0?ooo`3=0?ooo`002P3oool00`3OW4P004RL0=ooo`020?ooo`03
0;md0000M;l0oooo00D0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`090?ooo`03
0000003oool0oooo03X0oooo0`0000000`3oool000000?ooo`060?ooo`<00000H`3oool00`000000
oooo0?ooo`050?ooo`030000003oool0oooo0<h0oooo000:0?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`060?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool90?ooo`030000
003oool0oooo03T0oooo1@0000000`3oool000000?ooo`040?ooo`D00000HP3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo0<l0oooo000;0?ooo`050?oo_`1d000000000000M02oool0
203oool01P3ooml0W4P0000000000000001d0;ooo`X0oooo00<000000?ooo`3oool0>@3oool50000
0080oooo00<000000?ooo`3oool00P3oool500000680oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`3?0?ooo`00:03oool00`000000oooo0?ooo`0i0?ooo`D000000`3oool010000000
oooo0?ooo`3oool500000600oooo0`0000050?ooo`030000003oool0oooo0=00oooo000X0?ooo`03
0000003oool0oooo03X0oooo0`0000050?ooo`040000003oool0oooo0?ooo`<00000H03oool50000
00<0oooo00<000000?ooo`3oool0d@3oool002P0oooo0P00000l0?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1N0?ooo`D000000P3oool00`000000
oooo0?ooo`3B0?ooo`00:03oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0GP3oool5000000030?ooo`000000oooo0=@0oooo000X0?ooo`030000003o
ool0oooo04@0oooo00<000000?ooo`3oool0H`3oool300000080oooo00<000000?ooo`3oool0d`3o
ool002P0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0
00000=H0oooo000X0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0H`3oool20000
0=L0oooo000X0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0H`3oool00`000000
oooo0?ooo`3F0?ooo`00:03oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo0680oooo
00<000000?ooo`3oool0eP3oool002P0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?oo
o`1P0?ooo`030000003oool0oooo0=L0oooo000X0?ooo`030000003oool0oooo04T0oooo00<00000
0?ooo`3oool0GP3oool00`000000oooo0?ooo`3H0?ooo`00:03oool2000004X0oooo00<000000?oo
o`3oool0G@3oool00`000000oooo0?ooo`3I0?ooo`00:03oool00`000000oooo0?ooo`1:0?ooo`03
0000003oool0oooo05`0oooo00<000000?ooo`3oool0f@3oool002P0oooo00<000000?ooo`3oool0
B`3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo0=X0oooo000X0?ooo`030000003o
ool0oooo04/0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`3K0?ooo`00:03oool0
0`000000oooo0?ooo`1<0?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool0f`3oool0
02P0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo
0=`0oooo000X0?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3oool0EP3oool00`000000
oooo0?ooo`3L0?ooo`00:03oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo05D0oooo
00<000000?ooo`3oool0g@3oool002P0oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool0g@3oool002P0oooo0P00001?
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`3N
0?ooo`00:03oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0C@3oool00`000000oooo0?ooo`3N0?ooo`00:03oool00`000000oooo0?ooo`1?0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`3O0?ooo`00
:03oool00`000000oooo0?ooo`1@0?ooo`050000003oool0oooo0?ooo`000000CP3oool00`000000
oooo0?ooo`3O0?ooo`00:03oool00`000000oooo0?ooo`1@0?ooo`050000003oool0oooo0?ooo`00
0000C@3oool00`000000oooo0?ooo`3P0?ooo`002`3oool01@3ogi`0B0000000000007@0_ooo00<0
oooo00<0okmd0000001d_ol01@3oool0103ooml0W4P00018W03Oool90?ooo`030000003oool0oooo
0540oooo00<000000?ooo`0000000P00001<0?ooo`030000003oool0oooo0>00oooo000:0?ooo`07
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo00D0oooo00@0oooO
09a80000B9`0gooo2@3oool00`000000oooo0?ooo`1A0?ooo`H00000BP3oool00`000000oooo0?oo
o`3Q0?ooo`002P3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00T0oooo00@0oooO
09a80000B9`0gooo2@3oool00`000000oooo0?ooo`1B0?ooo`D00000BP3oool00`000000oooo0?oo
o`3Q0?ooo`002@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00D0oooo00<0_g@0
0000000000001@0000000`1d_ol0oooo0?ooo`060?ooo`030000003oool0oooo0580oooo1@000019
0?ooo`030000003oool0oooo0>80oooo00090?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol01@3oool0203ogi`0M9cO0?ooo`3oool0oooO09a80000B9`0gooo2@3oool00`000000oooo
0?ooo`1C0?ooo`<00000B@3oool200000>@0oooo00090?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol01P3oool01`3OW4P0M;oo0?ooo`3ooml0W4P00018W03Oool02@3oool3000005<0
oooo0P00001:0?ooo`800000i03oool000T0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`060?ooo`070?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`090?ooo`030000003oool0oooo
05@0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`3S0?ooo`002@3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00H0omnL04Qd_`3ooml0W4P00018W03Oool9
0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`3S
0?ooo`002@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00P0oooo00D0gia807Bo
g`2LB00004RL0=ooo`090?ooo`030000003oool0oooo05@0oooo0P0000180?ooo`800000i@3oool0
00X0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`080?ooo`050?oog`2LB4P0B4P0
0018W03Oool02@3oool00`000000oooo0?ooo`1D0?ooo`800000B03oool200000>D0oooo000:0?oo
o`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0:0?ooo`040?oOW018000004RL0=ooo`T0oooo
00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`160?ooo`<00000i03oool000/0oooo00D0
oono07@000000000001d0;ooo`0<0?ooo`030=nLB000B9`0gooo00T0oooo00<000000?ooo`3oool0
EP3oool00`000000oooo0?ooo`140?ooo`D00000h`3oool002P0oooo00<000000?ooo`3oool0EP3o
ool00`000000oooo0?ooo`140?ooo`D00000h`3oool002P0oooo00<000000?ooo`3oool0E`3oool0
0`000000oooo0?ooo`130?ooo`D00000h`3oool002P0oooo0P00001H0?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool0>03oool00`000000oooo0000000200000>@0oooo000X0?ooo`03
0000003oool0oooo05P0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0h0?ooo`04
0000003oool0oooo00000>D0oooo000X0?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`0g0?ooo`050000003oool0oooo0?ooo`000000i@3oool002P0
oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo03L0
oooo00D000000?ooo`3oool0oooo0000003U0?ooo`00:03oool00`000000oooo0?ooo`1I0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool0=`3oool01@000000oooo0?ooo`3oool00000
0>D0oooo000X0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`0f0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0h`3oool002P0oooo
00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo03H0oooo
00<000000?ooo`3oool0j03oool002P0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0j03oool002P0oooo00<00000
0?ooo`3oool0F`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo03D0oooo00<00000
0?ooo`3oool0j@3oool002P0oooo0P00001L0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool0=@3oool00`000000oooo0?ooo`3Y0?ooo`00:03oool00`000000oooo0?ooo`1K0?ooo`04
0000003oool0oooo0?ooo`<00000=P3oool00`000000oooo0?ooo`3Z0?ooo`00:03oool00`000000
oooo0?ooo`1L0?ooo`030000003oool0000000@00000=@3oool00`000000oooo0?ooo`3Z0?ooo`00
:03oool00`000000oooo0?ooo`1L0?ooo`030000003oool0000000@00000=@3oool00`000000oooo
0?ooo`3Z0?ooo`00:03oool00`000000oooo0?ooo`1L0?ooo`030000003oool0000000@00000=03o
ool00`000000oooo0?ooo`3[0?ooo`00:03oool00`000000oooo0?ooo`1L0?ooo`030000003oool0
oooo00<00000=@3oool00`000000oooo0?ooo`3[0?ooo`00:03oool00`000000oooo0?ooo`1M0?oo
o`050000003oool0oooo0?ooo`000000=@3oool00`000000oooo0?ooo`3[0?ooo`00:03oool00`00
0000oooo0?ooo`1M0?ooo`050000003oool0oooo0?ooo`000000=@3oool00`000000oooo0?ooo`3[
0?ooo`00:03oool00`000000oooo0?ooo`1M0?ooo`050000003oool0oooo0?ooo`000000=03oool0
0`000000oooo0?ooo`3/0?ooo`00:03oool2000005h0oooo00D000000?ooo`3oool0oooo0000000d
0?ooo`030000003oool0oooo0>`0oooo000X0?ooo`030000003oool0oooo05h0oooo00@000000?oo
o`3oool00000=03oool00`000000oooo0?ooo`3/0?ooo`00:03oool00`000000oooo0?ooo`1N0?oo
o`040000003oool0oooo000003<0oooo00<000000?ooo`3oool0k@3oool002P0oooo00<000000?oo
o`3oool0GP3oool010000000oooo0?ooo`00000c0?ooo`030000003oool0oooo0>d0oooo000;0?oo
o`050?oOW018000000000000M02oool00`3oool00`3o_g@0000007Boo`030?ooo`050?oOW0180000
00000000001d_ol02P3oool00`000000oooo0?ooo`1N0?ooo`040000003oool0oooo000003<0oooo
00<000000?ooo`3oool0k@3oool000X0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080
oooo00<0okmd0000001d_ol00P3oool01`3ogi`0B00004RLg`3oool0oono07@0001d_ol02@3oool0
0`000000oooo0?ooo`1O0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0k@3oool0
00X0oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?oog`2LB00007Bo
00<0oooo00<0okmd0018W03Oool0203oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo
0380oooo00<000000?ooo`3oool0kP3oool000T0oooo00<0oono07@00018W=l00`3oool00`3OW4P0
001d0;ooo`050?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0203oool00`000000
oooo0?ooo`1O0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0kP3oool000T0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`050?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol0203oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo0340oooo
00<000000?ooo`3oool0kP3oool000T0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`050?ooo`030?oOW0180000M;oo00<0oooo00<0_g@00000B02Lgol0203oool300000600oooo00<0
00000?ooo`3oool0<03oool00`000000oooo0?ooo`3_0?ooo`002@3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo00D0oooo00P0oono07@000000000B9cO0?ooo`3o_g@00000001d_`T0
oooo00<000000?ooo`3oool0H03oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo0>l0
oooo00090?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool00`2oM00004RL
0;nLB002000000030018W03Oool0oooo00P0oooo00<000000?ooo`3oool0H03oool00`000000oooo
0?ooo`0`0?ooo`030000003oool0oooo0>l0oooo00090?ooo`030?oo_`1d0000B9cO00<0oooo00<0
gia80000M02oool01P3oool00`3ogi`0B00007Boo`0=0?ooo`030000003oool0oooo0640oooo00<0
00000?ooo`3oool0;`3oool00`000000oooo0?ooo`3_0?ooo`002P3oool00`3OW4P004RL0=ooo`02
0?ooo`030;md0000M;l0oooo00L0oooo00<0gia80000B02Lgol0303oool00`000000oooo0?ooo`1Q
0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0l03oool000X0oooo00L0oono07@0
001d_ol0oooo0?noM00007@0_ooo00T0oooo00<0okmd00000000M;l02`3oool00`000000oooo0?oo
o`1Q0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0l03oool000/0oooo00D0oono
07@000000000001d0;ooo`0;0?ooo`050?oog`2LB00000000000M02oool0203oool00`000000oooo
0?ooo`1Q0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0l03oool002P0oooo00<0
00000?ooo`3oool0HP3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo0?40oooo000X
0?ooo`030000003oool0oooo0680oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`3a
0?ooo`00:03oool2000006<0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`3a0?oo
o`00:03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3o
ool0l@3oool002P0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0Z0?ooo`030000
003oool0oooo0?80oooo000X0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0:P3o
ool010000000oooo0?ooo`00003a0?ooo`00:03oool00`000000oooo0?ooo`1S0?ooo`030000003o
ool0oooo02X0oooo00@000000?ooo`3oool00000l@3oool002P0oooo00<000000?ooo`3oool0I03o
ool00`000000oooo0?ooo`0Y0?ooo`040000003oool0oooo00000?40oooo000X0?ooo`030000003o
ool0oooo06@0oooo00<000000?ooo`3oool0:03oool01@000000oooo0?ooo`3oool000000?40oooo
000X0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0:03oool01@000000oooo0?oo
o`3oool000000?40oooo000X0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0:03o
ool01@000000oooo0?ooo`3oool000000?40oooo000X0?ooo`800000IP3oool00`000000oooo0?oo
o`0V0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0k`3oool002P0oooo00<00000
0?ooo`3oool0I@3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0k`3oool002P0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`0V0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0k`3oool002P0oooo00<000000?ooo`3o
ool0I@3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo00<00000l@3oool002P0oooo
00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo00D00000
l03oool002P0oooo00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`0T0?ooo`030000003o
ool0oooo00D00000l03oool002P0oooo00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`0T
0?ooo`030000003oool0oooo00D00000l03oool002P0oooo00<000000?ooo`3oool0I`3oool00`00
0000oooo0?ooo`0R0?ooo`030000003oool0oooo0080oooo0`00003a0?ooo`00:03oool00`000000
oooo0?ooo`1W0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`3_0?ooo`00:03oool2000006P0oooo00<000000?ooo`3oool08P3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo0>l0oooo000X0?ooo`030000003oool0oooo06L0oooo00<0
00000?ooo`3oool08P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0>l0oooo000X
0?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo0>l0oooo000X0?ooo`030000003oool0oooo06P0oooo00<000000?oo
o`3oool0803oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0>l0oooo000;0?ooo`05
0?oOW018000000000000M02oool00`3oool00`3o_g@0000007Boo`030?ooo`050=nLB00000000000
00000018W=l02P3oool00`000000oooo0?ooo`1X0?ooo`030000003oool0oooo0200oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`3_0?ooo`002P3oool01`3ookl0M00007Boo`3oool0
okmd0000M02oool00P3oool00`3o_g@0000007Boo`020?ooo`070=nLB0000000M;oo0?ooo`3ooml0
W4P004RLg`090?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool07`3oool00`000000
oooo0?ooo`060?ooo`030000003oool0oooo0>l0oooo000:0?ooo`030;md0000B9`0gooo0080oooo
00<0_g@0001d_`3oool01@3oool00`3ooml0W4P0001d_`030?ooo`030?noM000B9`0gooo00P0oooo
00<000000?ooo`3oool0J@3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool0k`3oool000T0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;oo
o`060?ooo`030;md0000M;l0oooo0080oooo00<0_g@00018W03Oool0203oool00`000000oooo0?oo
o`1Y0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`3_0?ooo`002@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0omnL
04P0M02oool0oooo0;md00000000M;oo00T0oooo00<000000?ooo`3oool0J@3oool00`000000oooo
0?ooo`0N0?ooo`030000003oool0oooo0?P0oooo00090?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol01`3oool01@3ookl0M4Q804P000000000B9cO00X0oooo0`00001Z0?ooo`030000
003oool0oooo01`0oooo00<000000?ooo`3oool0n@3oool000T0oooo00<0omnL04P0001d_ol00`3o
ool00`3o_g@0001809cOo`070?ooo`040?oOW01800000000001d_`/0oooo00<000000?ooo`3oool0
JP3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo0?T0oooo00090?ooo`030?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3ogi`0B0000000M02oool0okmd001d_`X0
oooo00<000000?ooo`3oool0JP3oool010000000oooo0?ooo`00000K0?ooo`030000003oool0oooo
0?T0oooo00090?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool00`2oM000
001d0;ooo`020?ooo`030;md0018W=l0oooo00P0oooo00<000000?ooo`3oool0JP3oool010000000
oooo0?ooo`00000J0?ooo`030000003oool0oooo0?X0oooo000:0?ooo`030=nLB000B9`0gooo0080
oooo00<0_g@0001d_`3oool01P3oool00`2oM00004RL0=ooo`020?ooo`030=nLB000M;l0oooo00P0
oooo00<000000?ooo`3oool0J`3oool00`000000oooo0000000J0?ooo`030000003oool0oooo0?X0
oooo000:0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`070?ooo`070?noM0000000M;oo
0?ooo`3ookl0M00007Boo`090?ooo`030000003oool0oooo06/0oooo00<000000?ooo`0000006P3o
ool00`000000oooo0?ooo`3j0?ooo`002`3oool01@3ookl0M0000000000007@0_ooo00T0oooo00D0
okmd000000000000000004RLg`0:0?ooo`030000003oool0oooo06/0oooo00<000000?ooo`000000
6P3oool00`000000oooo0?ooo`3j0?ooo`00:03oool00`000000oooo0?ooo`1/0?ooo`8000006@3o
ool00`000000oooo0?ooo`3k0?ooo`00:03oool00`000000oooo0?ooo`1/0?ooo`8000006@3oool0
0`000000oooo0?ooo`3k0?ooo`00:03oool2000006d0oooo0P00000I0?ooo`030000003oool0oooo
0?/0oooo000X0?ooo`030000003oool0oooo06`0oooo0P00000H0?ooo`030000003oool0oooo0?`0
oooo000X0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool05P3oool00`000000oooo
0?ooo`3l0?ooo`00:03oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo01H0oooo00<0
00000?ooo`3oool0o03oool002P0oooo00<000000?ooo`3oool0K@3oool00`000000oooo0?ooo`0F
0?ooo`030000003oool0oooo0?`0oooo000X0?ooo`030000003oool0oooo06d0oooo0P00000F0?oo
o`800000oP3oool002P0oooo00<000000?ooo`3oool0K@3oool2000001H0oooo0P00003n0?ooo`00
:03oool00`000000oooo0?ooo`1/0?ooo`<000005P3oool200000?h0oooo000X0?ooo`030000003o
ool0oooo06/0oooo1@00000D0?ooo`030000003oool000000?h0oooo000X0?ooo`800000K03oool5
000001@0oooo00<000000?ooo`000000oP3oool002P0oooo00<000000?ooo`3oool0J`3oool50000
01@0oooo00<000000?ooo`000000oP3oool002P0oooo00<000000?ooo`3oool0K03oool300000003
0?ooo`000000oooo0140oooo00@000000?ooo`3oool00000oP3oool002P0oooo00<000000?ooo`3o
ool0K@3oool010000000oooo0?ooo`00000B0?ooo`040000003oool0oooo00000?h0oooo000X0?oo
o`030000003oool0oooo06d0oooo00@000000?ooo`3oool000004P3oool010000000oooo0?ooo`00
003n0?ooo`00:03oool00`000000oooo0?ooo`1]0?ooo`050000003oool0oooo0?ooo`000000403o
ool01@000000oooo0?ooo`3oool000000?h0oooo000X0?ooo`030000003oool0oooo06d0oooo00D0
00000?ooo`3oool0oooo0000000@0?ooo`050000003oool0oooo0?ooo`000000oP3oool002P0oooo
00<000000?ooo`3oool0K@3oool01@000000oooo0?ooo`3oool000000100oooo00D000000?ooo`3o
ool0oooo0000003n0?ooo`00:03oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0?`0
oooo000X0?ooo`800000KP3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`3l0?ooo`00:03oool00`000000oooo0?oo
o`1]0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`020?ooo`<00000o@3oool002P0oooo00<000000?ooo`3oool0K@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00X0oooo00@000000?ooo`3oool0oooo1@00003l0?ooo`00:03oool0
0`000000oooo0?ooo`1]0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`020?ooo`D00000o03oool001T0oooo00H0_g@000000000000000000018W03O
ool90?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo0080oooo1@00003l0?ooo`006P3oool00`3ogi`0B00007Boo`0;
0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo00@0oooo0`00003m0?ooo`006P3oool00`3ogi`0B00007Boo`0;0?oo
o`030000003oool0oooo06d0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool0o03oool001X0oooo00<0omnL04P0001d
_ol02`3oool00`000000oooo0?ooo`1f0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`3l0?ooo`006P3oool00`3ogi`0B00007Boo`0;0?ooo`030000
003oool0oooo07L0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`070?ooo`030000
003oool0oooo0?`0oooo000J0?ooo`030?oOW0180000M;oo00/0oooo0`00001h0?ooo`@000002P3o
ool00`000000oooo0?ooo`3l0?ooo`006P3oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0
oooo08H0oooo00<000000?ooo`3oool0o03oool001X0oooo00<0omnL04P0001d_ol02`3oool00`00
0000oooo0?ooo`260?ooo`030000003oool0oooo0?`0oooo000J0?ooo`030?oOW0180000M;oo00/0
oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`3l0?ooo`006P3oool00`3ogi`0B000
07Boo`0;0?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3oool0o03oool001P0oooo00D0
oooO09a800000000000007Boo`0;0?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3oool0
o03oool001X0oooo00<0okmd0000001d_ol02`3oool00`000000oooo0?ooo`260?ooo`030000003o
ool0oooo0?`0oooo000X0?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3oool0o03oool0
02P0oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`3l0?ooo`00:03oool2000008L0
oooo00<000000?ooo`3oool0o03oool002P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03o
ool00`000000oooo0?ooo`3o0?ooohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool0
02P0oooo00<000000?ooo`3oool0o`3ooon60?ooo`00:03oool00`000000oooo0?ooo`1j0?ooo`03
0000003oool0oooo0?l0oooo2@3oool002P0oooo00<000000?ooo`3oool0NP3oool00`000000oooo
0?ooo`3o0?ooo`T0oooo000X0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0o`3o
ool90?ooo`00:03oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo0?l0oooo2@3oool0
02P0oooo0P00001k0?ooo`030000003oool0oooo0?l0oooo2@3oool002P0oooo00<000000?ooo`3o
ool0NP3oool00`000000oooo0?ooo`3o0?ooo`T0oooo000X0?ooo`030000003oool0oooo07X0oooo
00<000000?ooo`3oool0o`3oool90?ooo`00:03oool00`000000oooo0?ooo`1j0?ooo`030000003o
ool0oooo0?l0oooo2@3oool002P0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`3o
0?ooo`T0oooo000X0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0o`3oool90?oo
o`00:03oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo0?l0oooo2@3oool002P0oooo
00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`3o0?ooo`T0oooo000X0?ooo`030000003o
ool0oooo07X0oooo00<000000?ooo`3oool0o`3oool90?ooo`00:03oool2000007/0oooo00<00000
0?ooo`3oool0o`3oool90?ooo`00:03oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo
0?l0oooo2@3oool002P0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`3o0?ooo`T0
oooo000X0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0o`3oool90?ooo`002`3o
ool01P2oM000000000000000000004RL0=ooo`80oooo00D0okmd0000001d_ol0oooo0;md00060000
000304RLg`3oool0oooo00L0oooo00<000000?ooo`3oool0N03oool300000?l0oooo2`3oool000`0
oooo00<0omnL04P0001d_ol0103oool00`3o_g@0000007Boo`020?ooo`030=nLB0000000000000<0
000000<004RL0=ooo`3oool01`3oool00`000000oooo0?ooo`1g0?ooo`D00000o`3oool:0?ooo`00
303oool00`3ogi`0B00007Boo`0:0?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0
1`3oool00`000000oooo0?ooo`1g0?ooo`D00000o`3oool:0?ooo`00303oool00`3ogi`0B00007Bo
o`0;0?ooo`030=nLB01d_ol0oooo00/0oooo00<000000?ooo`3oool0M`3oool500000?l0oooo2P3o
ool000`0oooo00<0omnL04P0001d_ol0303oool00`2oM4P0W=oo0?ooo`0:0?ooo`030000003oool0
oooo07P0oooo0`00003o0?ooo`/0oooo000<0?ooo`030?oOW0180000M;oo00`0oooo00<0oooO09a8
M02oool02P3oool3000007X0oooo00<000000?ooo`3oool0o`3oool90?ooo`00303oool00`3ogi`0
B00007Boo`0=0?ooo`030?noM000M;l0oooo00T0oooo00<000000?ooo`3oool0NP3oool00`000000
oooo0?ooo`3o0?ooo`T0oooo000<0?ooo`030?oOW0180000M;oo00d0oooo00<0oono07@0B02Lgol0
2@3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo0?l0oooo2@3oool000`0oooo00<0
omnL04P0001d_ol03@3oool00`3ookl0M00004RLg`090?ooo`030000003oool0oooo07X0oooo00<0
00000?ooo`3oool0o`3oool90?ooo`00303oool00`3ogi`0B00007Boo`080?ooo`030?noW02Lgol0
oooo0080oooo00<0omnL04P00018W=l02@3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0
oooo0?l0oooo2@3oool000X0oooo00D0oooO09a800000000000007Boo`080?ooo`080?oog`2LB4P0
W=oo0?ooo`3ookl0M0000000M02oool90?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3o
ool0o`3oool90?ooo`00303oool00`3o_g@0000007Boo`090?ooo`060?oog`2LB000000000000000
07@0_ooo2P3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo0?l0oooo2@3oool002P0
oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`3o0?ooo`T0oooo000X0?ooo`030000
003oool0oooo07X0oooo00<000000?ooo`3oool0o`3oool90?ooo`00:03oool2000007/0oooo00<0
00000?ooo`3oool0o`3oool90?ooo`00:03oool00`000000oooo0?ooo`1j0?ooo`030000003oool0
oooo0?l0oooo2@3oool002P0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`3o0?oo
o`T0oooo000X0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0o`3oool90?ooo`00
:03oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo0?l0oooo2@3oool002P0oooo00<0
00000?ooo`3oool0NP3oool00`000000oooo0?ooo`3o0?ooo`T0oooo000X0?ooo`030000003oool0
oooo07X0oooo00<000000?ooo`3oool0o`3oool90?ooo`00:03oool00`000000oooo0?ooo`3o0?oo
ohH0oooo000X0?ooo`030000003oool0oooo0?l0ooooQP3oool002P0oooo00<000000?ooo`3oool0
o`3ooon60?ooo`00:03oool200000?l0ooooQ`3oool002P0oooo00<000000?ooo`3oool0o`3ooon6
0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.321202, -0.157075, \
0.0120066, 0.0082817}}]
}, Closed]],

Cell[TextData[{
  "However, you should note that this example is intended to illustrate some \
of the capabilities of ",
  StyleBox["ListPlot", "InlineInput",
    FontWeight->"Bold"],
  " and ",
  StyleBox["ErrorListPlot", "InlineInput",
    FontWeight->"Bold"],
  " and does not accurately represent the statistical properties of a fit.  \
Data of this type would probably be distributed according to either a Poisson \
or Gaussian distribution and the error bars should be derived from the \
variances of the appropriate distribution.  Data analysis will be discussed \
later in the course."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Fancier list plots using MultipleListPlot", "Subsection"],

Cell[TextData[{
  ButtonBox["MultipleListPlot",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput"],
  StyleBox["dataset1,dataset2,\[CenterEllipsis], options", "InlineInput"],
  StyleBox["]", "InlineInput"],
  StyleBox[" can be used to plot several lists together, giving each a \
different symbol and/or line, and can display error bars in vertical and/or \
horizontal directions.  This function is contained in a package by the same \
name, but has already become available by loading the master graphics \
package.  Each data set should have the form ",
    FontWeight->"Plain"],
  StyleBox["{{point,error},\[CenterEllipsis]}", "InlineInput"],
  StyleBox[" where ",
    FontWeight->"Plain"],
  StyleBox["point", "InlineInput"],
  StyleBox[" is an {",
    FontWeight->"Plain"],
  StyleBox["x,y",
    FontWeight->"Plain",
    FontSlant->"Italic"],
  StyleBox["} pair and ",
    FontWeight->"Plain"],
  StyleBox["error", "InlineInput"],
  StyleBox[" is an ",
    FontWeight->"Plain"],
  StyleBox["ErrorBar", "InlineInput"],
  StyleBox[" object of the form ",
    FontWeight->"Plain"],
  StyleBox["ErrorBar[", "InlineInput"],
  StyleBox["yerr", "InlineInput"],
  StyleBox["]", "InlineInput"],
  StyleBox[" or ",
    FontWeight->"Plain"],
  StyleBox["ErrorBar[", "InlineInput"],
  StyleBox["xerr,yerr", "InlineInput"],
  StyleBox["]", "InlineInput"],
  StyleBox[".  Each error is either a single value representing a symmetric \
uncertainty or an interval about zero of the form ",
    FontWeight->"Plain"],
  StyleBox["{negerr,poserr}", "InlineInput"],
  StyleBox[" which can be used for asymmetric error bars.",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[TextData[{
  StyleBox["The options for ",
    FontWeight->"Plain"],
  StyleBox["MultipleListPlot", "InlineInput"],
  StyleBox[" include:",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[TextData[{
  StyleBox["SymbolShape", "InlineInput"],
  StyleBox["\[Rule]{primitives}", "InlineInput"],
  StyleBox[" \[LongDash] gives list symbol shapes to be assigned cyclically \
to data sets",
    FontWeight->"Plain"]
}], "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}},
  FontWeight->"Bold"],

Cell[TextData[{
  StyleBox["SymbolStyle", "InlineInput"],
  StyleBox["\[Rule]{directives}", "InlineInput"],
  StyleBox[" \[LongDash] describes color and other style options",
    FontWeight->"Plain"]
}], "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}},
  FontWeight->"Bold"],

Cell[TextData[{
  StyleBox["SymbolLabel", "InlineInput"],
  StyleBox["\[Rule]{labels}", "InlineInput"],
  StyleBox[" \[LongDash] cyclic list of labels",
    FontWeight->"Plain"]
}], "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}},
  FontWeight->"Bold"],

Cell[TextData[{
  StyleBox["ErrorBarFunction", "InlineInput"],
  StyleBox["\[Rule]function", "InlineInput"],
  StyleBox[" ",
    FontWeight->"Plain",
    FontSlant->"Italic"],
  StyleBox["used to create a graphics object representing the error bar",
    FontWeight->"Plain"]
}], "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}},
  FontWeight->"Bold"],

Cell[TextData[{
  StyleBox["The most important options apply to shape and style.  The shape \
can be specified using ",
    FontWeight->"Plain"],
  StyleBox["PlotSymbol[", "InlineInput"],
  StyleBox["shape,size,options", "InlineInput"],
  StyleBox["]", "InlineInput"],
  StyleBox[" where ",
    FontWeight->"Plain"],
  StyleBox["size", "InlineInput"],
  StyleBox[" is optional and where the available shapes include ",
    FontWeight->"Plain"],
  StyleBox["Diamond", "InlineInput"],
  StyleBox[", ",
    FontWeight->"Plain"],
  StyleBox["Triangle", "InlineInput"],
  StyleBox[", ",
    FontWeight->"Plain"],
  StyleBox["Star", "InlineInput"],
  StyleBox[", ",
    FontWeight->"Plain"],
  StyleBox["Box", "InlineInput"],
  StyleBox[", and others.  The only option presently available for these \
shapes is ",
    FontWeight->"Plain"],
  StyleBox["Filled", "InlineInput"],
  StyleBox[", which defaults to ",
    FontWeight->"Plain"],
  StyleBox["True", "InlineInput",
    FontWeight->"Plain"],
  StyleBox[".  Alternatively, ",
    FontWeight->"Plain"],
  StyleBox["MakeSymbol", "InlineInput"],
  StyleBox[" can be used to create other shapes.  ",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[TextData[{
  StyleBox["Although there are too many options and features to cover here, \
an example should illustrate some of the features of this function.  Below we \
create two data sets with random fluctuations based upon the same basic shape \
but with different error bars, here taken to be symmetric.  The two sets are \
plotted with different symbol formats to illustrate the most important \
features of ",
    FontWeight->"Plain"],
  StyleBox["MultipleListPlot", "InlineInput"],
  StyleBox[".",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[BoxData[{
    \(\(Clear[bwData];\)\), "\n", 
    \(bwData[x0_, \[CapitalGamma]_, \[Delta]x_, \[Delta]y_, xmin_, xmax_, 
        xinc_] := Table[{x, \ bw[x, x0, \[CapitalGamma]]}, {x, xmin, xmax, 
            xinc}]\  /. {{x_, 
              y_} \[RuleDelayed] {{x, 
                y\ Random[Real, {1 - \[Delta]y, 1 + \[Delta]y}]}, 
              ErrorBar[\[Delta]x, \(\@\(1/3\)\) y\ \[Delta]y]}}\)}], "Input"],

Cell[BoxData[{
    \(\(set1 = bwData[1, 1\/2, 0.1, 0.1, 0, 3, 0.2];\)\), "\n", 
    \(\(set2 = bwData[1, 1\/2, 0.05, 0.2, 0.1, 3, 0.2];\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[
        MultipleListPlot[set1, set2, 
          SymbolShape \[Rule] {PlotSymbol[Triangle, 5], 
              PlotSymbol[Diamond, 5, Filled \[Rule] False]}, 
          SymbolStyle \[Rule] {RGBColor[0, 1, 0], RGBColor[1, 0, 0]}], 
        Plot[bw[x, 1, 1\/2], {x, 0, 3}]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0535714 0.297619 0.0147151 0.511153 [
[.20238 .00222 -9.5 -15.6875 ]
[.20238 .00222 9.5 0 ]
[.35119 .00222 -5 -15.6875 ]
[.35119 .00222 5 0 ]
[.5 .00222 -9.5 -15.6875 ]
[.5 .00222 9.5 0 ]
[.64881 .00222 -5 -15.6875 ]
[.64881 .00222 5 0 ]
[.79762 .00222 -9.5 -15.6875 ]
[.79762 .00222 9.5 0 ]
[.94643 .00222 -5 -15.6875 ]
[.94643 .00222 5 0 ]
[.04107 .11695 -19 -7.84375 ]
[.04107 .11695 0 7.84375 ]
[.04107 .21918 -19 -7.84375 ]
[.04107 .21918 0 7.84375 ]
[.04107 .32141 -19 -7.84375 ]
[.04107 .32141 0 7.84375 ]
[.04107 .42364 -19 -7.84375 ]
[.04107 .42364 0 7.84375 ]
[.04107 .52587 -10 -7.84375 ]
[.04107 .52587 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.20238 .01472 m
.20238 .02097 L
s
gsave
.20238 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.35119 .01472 m
.35119 .02097 L
s
gsave
.35119 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .01472 m
.5 .02097 L
s
gsave
.5 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.5) show
1.000 setlinewidth
grestore
.64881 .01472 m
.64881 .02097 L
s
gsave
.64881 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.79762 .01472 m
.79762 .02097 L
s
gsave
.79762 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2.5) show
1.000 setlinewidth
grestore
.94643 .01472 m
.94643 .02097 L
s
gsave
.94643 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.08333 .01472 m
.08333 .01847 L
s
.1131 .01472 m
.1131 .01847 L
s
.14286 .01472 m
.14286 .01847 L
s
.17262 .01472 m
.17262 .01847 L
s
.23214 .01472 m
.23214 .01847 L
s
.2619 .01472 m
.2619 .01847 L
s
.29167 .01472 m
.29167 .01847 L
s
.32143 .01472 m
.32143 .01847 L
s
.38095 .01472 m
.38095 .01847 L
s
.41071 .01472 m
.41071 .01847 L
s
.44048 .01472 m
.44048 .01847 L
s
.47024 .01472 m
.47024 .01847 L
s
.52976 .01472 m
.52976 .01847 L
s
.55952 .01472 m
.55952 .01847 L
s
.58929 .01472 m
.58929 .01847 L
s
.61905 .01472 m
.61905 .01847 L
s
.67857 .01472 m
.67857 .01847 L
s
.70833 .01472 m
.70833 .01847 L
s
.7381 .01472 m
.7381 .01847 L
s
.76786 .01472 m
.76786 .01847 L
s
.82738 .01472 m
.82738 .01847 L
s
.85714 .01472 m
.85714 .01847 L
s
.8869 .01472 m
.8869 .01847 L
s
.91667 .01472 m
.91667 .01847 L
s
.02381 .01472 m
.02381 .01847 L
s
.97619 .01472 m
.97619 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.05357 .11695 m
.05982 .11695 L
s
gsave
.04107 .11695 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.05357 .21918 m
.05982 .21918 L
s
gsave
.04107 .21918 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.05357 .32141 m
.05982 .32141 L
s
gsave
.04107 .32141 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.05357 .42364 m
.05982 .42364 L
s
gsave
.04107 .42364 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.05357 .52587 m
.05982 .52587 L
s
gsave
.04107 .52587 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.05357 .04027 m
.05732 .04027 L
s
.05357 .06583 m
.05732 .06583 L
s
.05357 .09139 m
.05732 .09139 L
s
.05357 .1425 m
.05732 .1425 L
s
.05357 .16806 m
.05732 .16806 L
s
.05357 .19362 m
.05732 .19362 L
s
.05357 .24473 m
.05732 .24473 L
s
.05357 .27029 m
.05732 .27029 L
s
.05357 .29585 m
.05732 .29585 L
s
.05357 .34696 m
.05732 .34696 L
s
.05357 .37252 m
.05732 .37252 L
s
.05357 .39808 m
.05732 .39808 L
s
.05357 .4492 m
.05732 .4492 L
s
.05357 .47475 m
.05732 .47475 L
s
.05357 .50031 m
.05732 .50031 L
s
.05357 .55143 m
.05732 .55143 L
s
.05357 .57698 m
.05732 .57698 L
s
.05357 .60254 m
.05732 .60254 L
s
.25 Mabswid
.05357 0 m
.05357 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.001 w
.08333 .04455 m
.02381 .04455 L
s
.08333 .04455 0 1.5 Mabsadd m
.08333 .04455 0 -1.5 Mabsadd L s
.02381 .04455 0 1.5 Mabsadd m
.02381 .04455 0 -1.5 Mabsadd L s
.05357 .04629 m
.05357 .04282 L
s
.05357 .04629 1.5 0 Mabsadd m
.05357 .04629 -1.5 0 Mabsadd L s
.05357 .04282 1.5 0 Mabsadd m
.05357 .04282 -1.5 0 Mabsadd L s
.14286 .06425 m
.08333 .06425 L
s
.14286 .06425 0 1.5 Mabsadd m
.14286 .06425 0 -1.5 Mabsadd L s
.08333 .06425 0 1.5 Mabsadd m
.08333 .06425 0 -1.5 Mabsadd L s
.1131 .06687 m
.1131 .06162 L
s
.1131 .06687 1.5 0 Mabsadd m
.1131 .06687 -1.5 0 Mabsadd L s
.1131 .06162 1.5 0 Mabsadd m
.1131 .06162 -1.5 0 Mabsadd L s
.20238 .08664 m
.14286 .08664 L
s
.20238 .08664 0 1.5 Mabsadd m
.20238 .08664 0 -1.5 Mabsadd L s
.14286 .08664 0 1.5 Mabsadd m
.14286 .08664 0 -1.5 Mabsadd L s
.17262 .09101 m
.17262 .08228 L
s
.17262 .09101 1.5 0 Mabsadd m
.17262 .09101 -1.5 0 Mabsadd L s
.17262 .08228 1.5 0 Mabsadd m
.17262 .08228 -1.5 0 Mabsadd L s
.2619 .17014 m
.20238 .17014 L
s
.2619 .17014 0 1.5 Mabsadd m
.2619 .17014 0 -1.5 Mabsadd L s
.20238 .17014 0 1.5 Mabsadd m
.20238 .17014 0 -1.5 Mabsadd L s
.23214 .17843 m
.23214 .16185 L
s
.23214 .17843 1.5 0 Mabsadd m
.23214 .17843 -1.5 0 Mabsadd L s
.23214 .16185 1.5 0 Mabsadd m
.23214 .16185 -1.5 0 Mabsadd L s
.32143 .33682 m
.2619 .33682 L
s
.32143 .33682 0 1.5 Mabsadd m
.32143 .33682 0 -1.5 Mabsadd L s
.2619 .33682 0 1.5 Mabsadd m
.2619 .33682 0 -1.5 Mabsadd L s
.29167 .35481 m
.29167 .31882 L
s
.29167 .35481 1.5 0 Mabsadd m
.29167 .35481 -1.5 0 Mabsadd L s
.29167 .31882 1.5 0 Mabsadd m
.29167 .31882 -1.5 0 Mabsadd L s
.38095 .57381 m
.32143 .57381 L
s
.38095 .57381 0 1.5 Mabsadd m
.38095 .57381 0 -1.5 Mabsadd L s
.32143 .57381 0 1.5 Mabsadd m
.32143 .57381 0 -1.5 Mabsadd L s
.35119 .60332 m
.35119 .5443 L
s
.35119 .60332 1.5 0 Mabsadd m
.35119 .60332 -1.5 0 Mabsadd L s
.35119 .5443 1.5 0 Mabsadd m
.35119 .5443 -1.5 0 Mabsadd L s
.44048 .33181 m
.38095 .33181 L
s
.44048 .33181 0 1.5 Mabsadd m
.44048 .33181 0 -1.5 Mabsadd L s
.38095 .33181 0 1.5 Mabsadd m
.38095 .33181 0 -1.5 Mabsadd L s
.41071 .3498 m
.41071 .31381 L
s
.41071 .3498 1.5 0 Mabsadd m
.41071 .3498 -1.5 0 Mabsadd L s
.41071 .31381 1.5 0 Mabsadd m
.41071 .31381 -1.5 0 Mabsadd L s
.5 .16439 m
.44048 .16439 L
s
.5 .16439 0 1.5 Mabsadd m
.5 .16439 0 -1.5 Mabsadd L s
.44048 .16439 0 1.5 Mabsadd m
.44048 .16439 0 -1.5 Mabsadd L s
.47024 .17268 m
.47024 .1561 L
s
.47024 .17268 1.5 0 Mabsadd m
.47024 .17268 -1.5 0 Mabsadd L s
.47024 .1561 1.5 0 Mabsadd m
.47024 .1561 -1.5 0 Mabsadd L s
.55952 .08697 m
.5 .08697 L
s
.55952 .08697 0 1.5 Mabsadd m
.55952 .08697 0 -1.5 Mabsadd L s
.5 .08697 0 1.5 Mabsadd m
.5 .08697 0 -1.5 Mabsadd L s
.52976 .09133 m
.52976 .0826 L
s
.52976 .09133 1.5 0 Mabsadd m
.52976 .09133 -1.5 0 Mabsadd L s
.52976 .0826 1.5 0 Mabsadd m
.52976 .0826 -1.5 0 Mabsadd L s
.61905 .06215 m
.55952 .06215 L
s
.61905 .06215 0 1.5 Mabsadd m
.61905 .06215 0 -1.5 Mabsadd L s
.55952 .06215 0 1.5 Mabsadd m
.55952 .06215 0 -1.5 Mabsadd L s
.58929 .06478 m
.58929 .05953 L
s
.58929 .06478 1.5 0 Mabsadd m
.58929 .06478 -1.5 0 Mabsadd L s
.58929 .05953 1.5 0 Mabsadd m
.58929 .05953 -1.5 0 Mabsadd L s
.67857 .04326 m
.61905 .04326 L
s
.67857 .04326 0 1.5 Mabsadd m
.67857 .04326 0 -1.5 Mabsadd L s
.61905 .04326 0 1.5 Mabsadd m
.61905 .04326 0 -1.5 Mabsadd L s
.64881 .045 m
.64881 .04152 L
s
.64881 .045 1.5 0 Mabsadd m
.64881 .045 -1.5 0 Mabsadd L s
.64881 .04152 1.5 0 Mabsadd m
.64881 .04152 -1.5 0 Mabsadd L s
.7381 .03533 m
.67857 .03533 L
s
.7381 .03533 0 1.5 Mabsadd m
.7381 .03533 0 -1.5 Mabsadd L s
.67857 .03533 0 1.5 Mabsadd m
.67857 .03533 0 -1.5 Mabsadd L s
.70833 .03655 m
.70833 .0341 L
s
.70833 .03655 1.5 0 Mabsadd m
.70833 .03655 -1.5 0 Mabsadd L s
.70833 .0341 1.5 0 Mabsadd m
.70833 .0341 -1.5 0 Mabsadd L s
.79762 .03105 m
.7381 .03105 L
s
.79762 .03105 0 1.5 Mabsadd m
.79762 .03105 0 -1.5 Mabsadd L s
.7381 .03105 0 1.5 Mabsadd m
.7381 .03105 0 -1.5 Mabsadd L s
.76786 .03196 m
.76786 .03014 L
s
.76786 .03196 1.5 0 Mabsadd m
.76786 .03196 -1.5 0 Mabsadd L s
.76786 .03014 1.5 0 Mabsadd m
.76786 .03014 -1.5 0 Mabsadd L s
.85714 .02604 m
.79762 .02604 L
s
.85714 .02604 0 1.5 Mabsadd m
.85714 .02604 0 -1.5 Mabsadd L s
.79762 .02604 0 1.5 Mabsadd m
.79762 .02604 0 -1.5 Mabsadd L s
.82738 .02674 m
.82738 .02533 L
s
.82738 .02674 1.5 0 Mabsadd m
.82738 .02674 -1.5 0 Mabsadd L s
.82738 .02533 1.5 0 Mabsadd m
.82738 .02533 -1.5 0 Mabsadd L s
.91667 .02514 m
.85714 .02514 L
s
.91667 .02514 0 1.5 Mabsadd m
.91667 .02514 0 -1.5 Mabsadd L s
.85714 .02514 0 1.5 Mabsadd m
.85714 .02514 0 -1.5 Mabsadd L s
.8869 .0257 m
.8869 .02458 L
s
.8869 .0257 1.5 0 Mabsadd m
.8869 .0257 -1.5 0 Mabsadd L s
.8869 .02458 1.5 0 Mabsadd m
.8869 .02458 -1.5 0 Mabsadd L s
.97619 .02243 m
.91667 .02243 L
s
.97619 .02243 0 1.5 Mabsadd m
.97619 .02243 0 -1.5 Mabsadd L s
.91667 .02243 0 1.5 Mabsadd m
.91667 .02243 0 -1.5 Mabsadd L s
.94643 .02288 m
.94643 .02198 L
s
.94643 .02288 1.5 0 Mabsadd m
.94643 .02288 -1.5 0 Mabsadd L s
.94643 .02198 1.5 0 Mabsadd m
.94643 .02198 -1.5 0 Mabsadd L s
0 1 0 r
.0001 w
.05357 .04455 0 5 Mabsadd m
.05357 .04455 -4.33 -2.5 Mabsadd L
.05357 .04455 4.33 -2.5 Mabsadd L
.05357 .04455 0 5 Mabsadd L closepath F
.1131 .06425 0 5 Mabsadd m
.1131 .06425 -4.33 -2.5 Mabsadd L
.1131 .06425 4.33 -2.5 Mabsadd L
.1131 .06425 0 5 Mabsadd L closepath F
.17262 .08664 0 5 Mabsadd m
.17262 .08664 -4.33 -2.5 Mabsadd L
.17262 .08664 4.33 -2.5 Mabsadd L
.17262 .08664 0 5 Mabsadd L closepath F
.23214 .17014 0 5 Mabsadd m
.23214 .17014 -4.33 -2.5 Mabsadd L
.23214 .17014 4.33 -2.5 Mabsadd L
.23214 .17014 0 5 Mabsadd L closepath F
.29167 .33682 0 5 Mabsadd m
.29167 .33682 -4.33 -2.5 Mabsadd L
.29167 .33682 4.33 -2.5 Mabsadd L
.29167 .33682 0 5 Mabsadd L closepath F
.35119 .57381 0 5 Mabsadd m
.35119 .57381 -4.33 -2.5 Mabsadd L
.35119 .57381 4.33 -2.5 Mabsadd L
.35119 .57381 0 5 Mabsadd L closepath F
.41071 .33181 0 5 Mabsadd m
.41071 .33181 -4.33 -2.5 Mabsadd L
.41071 .33181 4.33 -2.5 Mabsadd L
.41071 .33181 0 5 Mabsadd L closepath F
.47024 .16439 0 5 Mabsadd m
.47024 .16439 -4.33 -2.5 Mabsadd L
.47024 .16439 4.33 -2.5 Mabsadd L
.47024 .16439 0 5 Mabsadd L closepath F
.52976 .08697 0 5 Mabsadd m
.52976 .08697 -4.33 -2.5 Mabsadd L
.52976 .08697 4.33 -2.5 Mabsadd L
.52976 .08697 0 5 Mabsadd L closepath F
.58929 .06215 0 5 Mabsadd m
.58929 .06215 -4.33 -2.5 Mabsadd L
.58929 .06215 4.33 -2.5 Mabsadd L
.58929 .06215 0 5 Mabsadd L closepath F
.64881 .04326 0 5 Mabsadd m
.64881 .04326 -4.33 -2.5 Mabsadd L
.64881 .04326 4.33 -2.5 Mabsadd L
.64881 .04326 0 5 Mabsadd L closepath F
.70833 .03533 0 5 Mabsadd m
.70833 .03533 -4.33 -2.5 Mabsadd L
.70833 .03533 4.33 -2.5 Mabsadd L
.70833 .03533 0 5 Mabsadd L closepath F
.76786 .03105 0 5 Mabsadd m
.76786 .03105 -4.33 -2.5 Mabsadd L
.76786 .03105 4.33 -2.5 Mabsadd L
.76786 .03105 0 5 Mabsadd L closepath F
.82738 .02604 0 5 Mabsadd m
.82738 .02604 -4.33 -2.5 Mabsadd L
.82738 .02604 4.33 -2.5 Mabsadd L
.82738 .02604 0 5 Mabsadd L closepath F
.8869 .02514 0 5 Mabsadd m
.8869 .02514 -4.33 -2.5 Mabsadd L
.8869 .02514 4.33 -2.5 Mabsadd L
.8869 .02514 0 5 Mabsadd L closepath F
.94643 .02243 0 5 Mabsadd m
.94643 .02243 -4.33 -2.5 Mabsadd L
.94643 .02243 4.33 -2.5 Mabsadd L
.94643 .02243 0 5 Mabsadd L closepath F
0 g
.001 w
.09821 .04507 m
.06845 .04507 L
s
.09821 .04507 0 1.5 Mabsadd m
.09821 .04507 0 -1.5 Mabsadd L s
.06845 .04507 0 1.5 Mabsadd m
.06845 .04507 0 -1.5 Mabsadd L s
.08333 .0493 m
.08333 .04084 L
s
.08333 .0493 1.5 0 Mabsadd m
.08333 .0493 -1.5 0 Mabsadd L s
.08333 .04084 1.5 0 Mabsadd m
.08333 .04084 -1.5 0 Mabsadd L s
.15774 .07561 m
.12798 .07561 L
s
.15774 .07561 0 1.5 Mabsadd m
.15774 .07561 0 -1.5 Mabsadd L s
.12798 .07561 0 1.5 Mabsadd m
.12798 .07561 0 -1.5 Mabsadd L s
.14286 .08229 m
.14286 .06893 L
s
.14286 .08229 1.5 0 Mabsadd m
.14286 .08229 -1.5 0 Mabsadd L s
.14286 .06893 1.5 0 Mabsadd m
.14286 .06893 -1.5 0 Mabsadd L s
.21726 .1032 m
.1875 .1032 L
s
.21726 .1032 0 1.5 Mabsadd m
.21726 .1032 0 -1.5 Mabsadd L s
.1875 .1032 0 1.5 Mabsadd m
.1875 .1032 0 -1.5 Mabsadd L s
.20238 .11501 m
.20238 .0914 L
s
.20238 .11501 1.5 0 Mabsadd m
.20238 .11501 -1.5 0 Mabsadd L s
.20238 .0914 1.5 0 Mabsadd m
.20238 .0914 -1.5 0 Mabsadd L s
.27679 .2259 m
.24702 .2259 L
s
.27679 .2259 0 1.5 Mabsadd m
.27679 .2259 0 -1.5 Mabsadd L s
.24702 .2259 0 1.5 Mabsadd m
.24702 .2259 0 -1.5 Mabsadd L s
.2619 .25009 m
.2619 .20171 L
s
.2619 .25009 1.5 0 Mabsadd m
.2619 .25009 -1.5 0 Mabsadd L s
.2619 .20171 1.5 0 Mabsadd m
.2619 .20171 -1.5 0 Mabsadd L s
.33631 .4172 m
.30655 .4172 L
s
.33631 .4172 0 1.5 Mabsadd m
.33631 .4172 0 -1.5 Mabsadd L s
.30655 .4172 0 1.5 Mabsadd m
.30655 .4172 0 -1.5 Mabsadd L s
.32143 .46808 m
.32143 .36632 L
s
.32143 .46808 1.5 0 Mabsadd m
.32143 .46808 -1.5 0 Mabsadd L s
.32143 .36632 1.5 0 Mabsadd m
.32143 .36632 -1.5 0 Mabsadd L s
.39583 .46958 m
.36607 .46958 L
s
.39583 .46958 0 1.5 Mabsadd m
.39583 .46958 0 -1.5 Mabsadd L s
.36607 .46958 0 1.5 Mabsadd m
.36607 .46958 0 -1.5 Mabsadd L s
.38095 .52047 m
.38095 .4187 L
s
.38095 .52047 1.5 0 Mabsadd m
.38095 .52047 -1.5 0 Mabsadd L s
.38095 .4187 1.5 0 Mabsadd m
.38095 .4187 -1.5 0 Mabsadd L s
.45536 .20822 m
.4256 .20822 L
s
.45536 .20822 0 1.5 Mabsadd m
.45536 .20822 0 -1.5 Mabsadd L s
.4256 .20822 0 1.5 Mabsadd m
.4256 .20822 0 -1.5 Mabsadd L s
.44048 .23241 m
.44048 .18403 L
s
.44048 .23241 1.5 0 Mabsadd m
.44048 .23241 -1.5 0 Mabsadd L s
.44048 .18403 1.5 0 Mabsadd m
.44048 .18403 -1.5 0 Mabsadd L s
.51488 .12351 m
.48512 .12351 L
s
.51488 .12351 0 1.5 Mabsadd m
.51488 .12351 0 -1.5 Mabsadd L s
.48512 .12351 0 1.5 Mabsadd m
.48512 .12351 0 -1.5 Mabsadd L s
.5 .13531 m
.5 .1117 L
s
.5 .13531 1.5 0 Mabsadd m
.5 .13531 -1.5 0 Mabsadd L s
.5 .1117 1.5 0 Mabsadd m
.5 .1117 -1.5 0 Mabsadd L s
.5744 .07998 m
.54464 .07998 L
s
.5744 .07998 0 1.5 Mabsadd m
.5744 .07998 0 -1.5 Mabsadd L s
.54464 .07998 0 1.5 Mabsadd m
.54464 .07998 0 -1.5 Mabsadd L s
.55952 .08666 m
.55952 .0733 L
s
.55952 .08666 1.5 0 Mabsadd m
.55952 .08666 -1.5 0 Mabsadd L s
.55952 .0733 1.5 0 Mabsadd m
.55952 .0733 -1.5 0 Mabsadd L s
.63393 .0533 m
.60417 .0533 L
s
.63393 .0533 0 1.5 Mabsadd m
.63393 .0533 0 -1.5 Mabsadd L s
.60417 .0533 0 1.5 Mabsadd m
.60417 .0533 0 -1.5 Mabsadd L s
.61905 .05753 m
.61905 .04907 L
s
.61905 .05753 1.5 0 Mabsadd m
.61905 .05753 -1.5 0 Mabsadd L s
.61905 .04907 1.5 0 Mabsadd m
.61905 .04907 -1.5 0 Mabsadd L s
.69345 .03533 m
.66369 .03533 L
s
.69345 .03533 0 1.5 Mabsadd m
.69345 .03533 0 -1.5 Mabsadd L s
.66369 .03533 0 1.5 Mabsadd m
.66369 .03533 0 -1.5 Mabsadd L s
.67857 .03823 m
.67857 .03243 L
s
.67857 .03823 1.5 0 Mabsadd m
.67857 .03823 -1.5 0 Mabsadd L s
.67857 .03243 1.5 0 Mabsadd m
.67857 .03243 -1.5 0 Mabsadd L s
.75298 .03475 m
.72321 .03475 L
s
.75298 .03475 0 1.5 Mabsadd m
.75298 .03475 0 -1.5 Mabsadd L s
.72321 .03475 0 1.5 Mabsadd m
.72321 .03475 0 -1.5 Mabsadd L s
.7381 .03686 m
.7381 .03265 L
s
.7381 .03686 1.5 0 Mabsadd m
.7381 .03686 -1.5 0 Mabsadd L s
.7381 .03265 1.5 0 Mabsadd m
.7381 .03265 -1.5 0 Mabsadd L s
.8125 .02662 m
.78274 .02662 L
s
.8125 .02662 0 1.5 Mabsadd m
.8125 .02662 0 -1.5 Mabsadd L s
.78274 .02662 0 1.5 Mabsadd m
.78274 .02662 0 -1.5 Mabsadd L s
.79762 .02822 m
.79762 .02503 L
s
.79762 .02822 1.5 0 Mabsadd m
.79762 .02822 -1.5 0 Mabsadd L s
.79762 .02503 1.5 0 Mabsadd m
.79762 .02503 -1.5 0 Mabsadd L s
.87202 .02625 m
.84226 .02625 L
s
.87202 .02625 0 1.5 Mabsadd m
.87202 .02625 0 -1.5 Mabsadd L s
.84226 .02625 0 1.5 Mabsadd m
.84226 .02625 0 -1.5 Mabsadd L s
.85714 .0275 m
.85714 .025 L
s
.85714 .0275 1.5 0 Mabsadd m
.85714 .0275 -1.5 0 Mabsadd L s
.85714 .025 1.5 0 Mabsadd m
.85714 .025 -1.5 0 Mabsadd L s
.93155 .0233 m
.90179 .0233 L
s
.93155 .0233 0 1.5 Mabsadd m
.93155 .0233 0 -1.5 Mabsadd L s
.90179 .0233 0 1.5 Mabsadd m
.90179 .0233 0 -1.5 Mabsadd L s
.91667 .0243 m
.91667 .02229 L
s
.91667 .0243 1.5 0 Mabsadd m
.91667 .0243 -1.5 0 Mabsadd L s
.91667 .02229 1.5 0 Mabsadd m
.91667 .02229 -1.5 0 Mabsadd L s
1 0 0 r
.0001 w
.08333 .04507 -3.75 0 Mabsadd m
.08333 .04507 0 5 Mabsadd L
.08333 .04507 3.75 0 Mabsadd L
.08333 .04507 0 -5 Mabsadd L
.08333 .04507 -3.75 0 Mabsadd L s
.14286 .07561 -3.75 0 Mabsadd m
.14286 .07561 0 5 Mabsadd L
.14286 .07561 3.75 0 Mabsadd L
.14286 .07561 0 -5 Mabsadd L
.14286 .07561 -3.75 0 Mabsadd L s
.20238 .1032 -3.75 0 Mabsadd m
.20238 .1032 0 5 Mabsadd L
.20238 .1032 3.75 0 Mabsadd L
.20238 .1032 0 -5 Mabsadd L
.20238 .1032 -3.75 0 Mabsadd L s
.2619 .2259 -3.75 0 Mabsadd m
.2619 .2259 0 5 Mabsadd L
.2619 .2259 3.75 0 Mabsadd L
.2619 .2259 0 -5 Mabsadd L
.2619 .2259 -3.75 0 Mabsadd L s
.32143 .4172 -3.75 0 Mabsadd m
.32143 .4172 0 5 Mabsadd L
.32143 .4172 3.75 0 Mabsadd L
.32143 .4172 0 -5 Mabsadd L
.32143 .4172 -3.75 0 Mabsadd L s
.38095 .46958 -3.75 0 Mabsadd m
.38095 .46958 0 5 Mabsadd L
.38095 .46958 3.75 0 Mabsadd L
.38095 .46958 0 -5 Mabsadd L
.38095 .46958 -3.75 0 Mabsadd L s
.44048 .20822 -3.75 0 Mabsadd m
.44048 .20822 0 5 Mabsadd L
.44048 .20822 3.75 0 Mabsadd L
.44048 .20822 0 -5 Mabsadd L
.44048 .20822 -3.75 0 Mabsadd L s
.5 .12351 -3.75 0 Mabsadd m
.5 .12351 0 5 Mabsadd L
.5 .12351 3.75 0 Mabsadd L
.5 .12351 0 -5 Mabsadd L
.5 .12351 -3.75 0 Mabsadd L s
.55952 .07998 -3.75 0 Mabsadd m
.55952 .07998 0 5 Mabsadd L
.55952 .07998 3.75 0 Mabsadd L
.55952 .07998 0 -5 Mabsadd L
.55952 .07998 -3.75 0 Mabsadd L s
.61905 .0533 -3.75 0 Mabsadd m
.61905 .0533 0 5 Mabsadd L
.61905 .0533 3.75 0 Mabsadd L
.61905 .0533 0 -5 Mabsadd L
.61905 .0533 -3.75 0 Mabsadd L s
.67857 .03533 -3.75 0 Mabsadd m
.67857 .03533 0 5 Mabsadd L
.67857 .03533 3.75 0 Mabsadd L
.67857 .03533 0 -5 Mabsadd L
.67857 .03533 -3.75 0 Mabsadd L s
.7381 .03475 -3.75 0 Mabsadd m
.7381 .03475 0 5 Mabsadd L
.7381 .03475 3.75 0 Mabsadd L
.7381 .03475 0 -5 Mabsadd L
.7381 .03475 -3.75 0 Mabsadd L s
.79762 .02662 -3.75 0 Mabsadd m
.79762 .02662 0 5 Mabsadd L
.79762 .02662 3.75 0 Mabsadd L
.79762 .02662 0 -5 Mabsadd L
.79762 .02662 -3.75 0 Mabsadd L s
.85714 .02625 -3.75 0 Mabsadd m
.85714 .02625 0 5 Mabsadd L
.85714 .02625 3.75 0 Mabsadd L
.85714 .02625 0 -5 Mabsadd L
.85714 .02625 -3.75 0 Mabsadd L s
.91667 .0233 -3.75 0 Mabsadd m
.91667 .0233 0 5 Mabsadd L
.91667 .0233 3.75 0 Mabsadd L
.91667 .0233 0 -5 Mabsadd L
.91667 .0233 -3.75 0 Mabsadd L s
0 g
.5 Mabswid
.05357 .04478 m
.0712 .04843 L
.08979 .05303 L
.11024 .05921 L
.12929 .06638 L
.16497 .08508 L
.18411 .09931 L
.20153 .11601 L
.21948 .13837 L
.23896 .17079 L
.25756 .21256 L
.27727 .27197 L
.31274 .41815 L
.32207 .45798 L
.33094 .49062 L
.33583 .50497 L
.33841 .51122 L
.34116 .51675 L
.34371 .52075 L
.34601 .5234 L
.34713 .52435 L
.34833 .52511 L
.34947 .52559 L
.35051 .52583 L
.3517 .52584 L
.35282 .52562 L
.35382 .52523 L
.35492 .52459 L
.35611 .52365 L
.35737 .52237 L
.35964 .51936 L
.36369 .51184 L
.36816 .50059 L
.37815 .46654 L
.38753 .42741 L
.4238 .27652 L
.44259 .21844 L
.46237 .17283 L
.48088 .1413 L
.4981 .11907 L
.51728 .10015 L
.5347 .08688 L
.55421 .07524 L
.57219 .06676 L
.60825 .05423 L
.62584 .04966 L
.64519 .04548 L
.68301 .03919 L
.7194 .03477 L
.7581 .03125 L
Mistroke
.79538 .02867 L
.83123 .02671 L
.86939 .02504 L
.90612 .02374 L
.94515 .02261 L
.94643 .02258 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00G`3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oo
g`2LB0030000000307Boo`3oool0oooo02d0oooo00H0_g@000000000000000000018W03Oool^0?oo
o`060;md00000000000000000000B9`0gooo0P3oool01P3o_g@0000007Boo`3oool0oooO09a800<0
000000<0M;oo0?ooo`3oool0:`3oool00`2oM0000000000000040000000304RLg`3oool0oooo02X0
oooo00<0_g@0000000000000100000002@18W=l0oooo0?ooo`3o_g@0000007Boo`3oool0oooO09a8
00030000000307Boo`3oool0oooo02/0oooo00<0oono07@0000000000P0000000`1d_ol0oooo0?oo
o`0P0?ooo`00GP3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@00000
07Boo`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`0^0?ooo`030?oOW0180000M;oo
0340oooo00<0omnL04P0001d_ol0103oool01`3o_g@0000007Boo`3oool0oooO09a8M02oool00P3o
ool00`2oM000M;oo0?ooo`0/0?ooo`030=nLB0000000000000<0000000<004RL0=ooo`3oool0:`3o
ool00`3OW4P00000000000030000000:0018W03Oool0oooo0?noM0000000M;oo0?ooo`3ooml0W4Qd
0;ooo`80oooo00<0_g@007Boo`3oool0:`3oool00`3ogi`0B00004RLg`020?ooo`030;md0018W=l0
oooo0200oooo001N0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3o
oml0W4Qd0;ooo`0^0?ooo`030?oOW0180000M;oo0340oooo00<0omnL04P0001d_ol03@3oool00`3o
oml0W4Qd0;ooo`0]0?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0;03oool00`3O
W4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo00T0oooo00<0oooO09a8M02oool0<03oool00`3o
okl0M01809cOo`0P0?ooo`00G@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00/0
oooo00<0_g@007Boo`3oool0;@3oool00`3ogi`0B00007Boo`0a0?ooo`030?oOW0180000M;oo00h0
oooo00<0_g@007Boo`3oool0;@3oool00`3OW4P0M;oo0?ooo`0a0?ooo`030=nLB01d_ol0oooo00h0
oooo00<0_g@007Boo`3oool0;`3oool00`3ookl0M00007Boo`0P0?ooo`00G@3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0oooO09a80018W=l0;P3oool00`3ogi`0B000
07Boo`0a0?ooo`030?oOW0180000M;oo00d0oooo00<0oooO09a80018W=l0;`3oool00`2oM4P0W=oo
0?ooo`0a0?ooo`030;mdB02Lgol0oooo00`0oooo00<0oooO09a80018W=l0<03oool00`3o_g@00000
07Boo`0P0?ooo`00G@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0
gia80000001d_ol0;P3oool00`3ogi`0B00007Boo`0a0?ooo`030?oOW0180000M;oo00d0oooo00<0
gia80000001d_ol0;`3oool00`3ooml0W4Qd0;ooo`0a0?ooo`030?oog`2LB7@0_ooo00`0oooo00<0
gia80000001d_ol0;`3oool0103ogi`0B0000000B02LgolP0?ooo`00G@3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo02h0
oooo00<0omnL04P0001d_ol0<@3oool00`3ogi`0B00007Boo`090?ooo`070=oOg`3oool0oono07@0
00000000001d0;ooo`0`0?ooo`030?noM000M;l0oooo0340oooo00<0okmd001d_`3oool01`3oool0
1`3Ogml0oooo0?oo_`1d000000000000M02oool0;@3oool01@3o_g@000000000000004P0W=oo0240
oooo001M0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3OW4P00000
000000000000001809cOobl0oooo00<0omnL04P0001d_ol0<@3oool00`3ogi`0B00007Boo`090?oo
o`060=nLB000000000000000000004P0W=oo<@3oool00`3ookl0M01809cOo`0a0?ooo`030?oo_`1d
04P0W=oo00L0oooo00H0gia800000000000000000000B02Lgol`0?ooo`030;md0000M;l0oooo0240
oooo001M0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool0103ookl0M000
00000018W=la0?ooo`030?oOW0180000M;oo0340oooo00<0omnL04P0001d_ol02@3oool0103ookl0
M00000000018W=lc0?ooo`030?oo_`1d0000B9cO0340oooo00<0oono07@00018W=l01`3oool0103o
okl0M00000000018W=lc0?ooo`030;md0018W=l0oooo0200oooo001N0?ooo`030=nLB000B9`0gooo
0080oooo00<0_g@0001d_`3oool01`3oool00`2oM000M;oo0?ooo`0a0?ooo`030?oOW0180000M;oo
0340oooo00<0omnL04P0001d_ol02P3oool00`2oM000M;oo0?ooo`0^0?ooo`030?noW02Lgol0oooo
0080oooo00<0omnL04P00018W=l0;03oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO
00P0oooo00<0_g@007Boo`3oool0;P3oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo
0200oooo001N0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`060?oOW0180000
00000000000004P0W=oo;03oool01@3ooml0W4P0000000000000M;oo02l0oooo00D0oooO09a80000
0000000007Boo`0:0?ooo`060?oOW018000000000000000004P0W=oo:`3oool0203ooml0W4Q809cO
o`3oool0oono07@0000007@0_ooo;03oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo
203oool01P3ogi`0B000000000000000001809cOob`0oooo00H0okmd04RLg`3oool0oono07@00000
M;lQ0?ooo`00G`3oool01@3ookl0M0000000000007@0_ooo00X0oooo00D0_g@00000000000000000
04RLg`0^0?ooo`030?noM0000000M;oo0340oooo00<0okmd0000001d_ol02`3oool01@2oM0000000
000000000000B9cO02`0oooo00H0oooO09a80000000000000000M02oool^0?ooo`060?oog`2LB000
00000000000007@0_ooo2P3oool01@2oM0000000000000000000B9cO02d0oooo00D0gia800000000
0000001d0;ooo`0Q0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0
ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooomB0?ooo`030?l0003oool0oooo02`0oooo
000]0?ooo`030000003oool0oooo0?l0oooo8`3oool00`3o0000oooo0?ooo`0E0?ooo`030?l0003o
ool0oooo01<0oooo00<0o`000?ooo`3o0000;@3oool002d0oooo00<000000?ooo`3oool0o`3ooolR
0?ooo`030?l0003oool0o`0001D0oooo00<0o`000?ooo`3o0000503oool0103o0000oooo0?ooo`3o
000/0?ooo`00;@3oool00`000000oooo0?ooo`3o0?ooob80oooo00<0o`000?ooo`3o0000503oool0
103o0000oooo0?ooo`3o000C0?ooo`030?l0003oool0oooo0080oooo00<0o`000?ooo`3oool0:@3o
ool001P0ooooo`00000T000000030?l00000000000000180000000D0o`0000000000000000000?l0
000C000000050?l00000000000000000003o00004@0000000`3o0000000000000003000000030?l0
00000000000000`00?l04000000=0?ooo`008P3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030?l0003o
ool0oooo00T0oooo00<000000?ooo`3oool0203oool00`3o000000000?l0000:0?ooo`030000003o
ool0oooo00D0oooo00@0o`000?ooo`3oool000000P3oool00`3o0000oooo0?ooo`060?ooo`030000
003oool0oooo00H0oooo00H0o`000?ooo`3oool000000?ooo`3o00030?ooo`/00?l000<000000?oo
o`3o00000`3oool01`000000oooo0?ooo`3oool0o`000?ooo`0000002P00o`070?ooo`030000003o
ool0oooo01@0oooo000R0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00L0oooo00<0o`000000003o00002P3oool0
0`000000oooo0?ooo`070?ooo`040?l0003oool000000?l000X0oooo00<000000?ooo`3oool01@3o
ool0103o0000oooo0?ooo`0000030?ooo`030?l0003oool0oooo00/00?l000<000000?ooo`3o0000
0`3oool010000000oooo0?ooo`3o00020?ooo`0300000000o`000?l000P00?l000<000000?ooo`3o
00000`3oool00`000000oooo0?ooo`020?ooo`030?l0000000000?l000T00?l01`3oool00`000000
oooo0?ooo`0D0?ooo`00;@3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo03P0oooo
00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo00L0oooo
00@0o`000?ooo`3oool0o`004`3oool01@3o0000oooo0?ooo`3oool0o`0000h0oooo00<000000?oo
o`3o00000`3oool020000000oooo0?ooo`3oool0o`000?ooo`000000oooo2P00o`000`000000o`00
0?ooo`030?ooo`070000003oool0oooo0?ooo`3o0000oooo0000000:003o00030000003o00000000
010000000`00o`09000001H0oooo000]0?ooo`030000003oool0oooo0?40oooo00<0o`000?ooo`3o
ool00P3oool00`3o0000oooo0?ooo`0?0?ooo`060?l0003oool0oooo0000003oool0o`000`3oool;
003o00040000003oool0o`000?ooo`H0000000@0oooo0?l000000000oooo2P00o`000`000000o`00
0?ooo`0N000000040?ooo`3o000000000?ooo`P00?l0203oool00`000000oooo0?ooo`0D0?ooo`00
;@3oool00`000000oooo0?ooo`090?ooo`030?l0003oool0oooo0>@0oooo00<0o`000?ooo`3oool0
0`3oool0103o0000oooo0?ooo`3oool<003o00030?ooo`3o0000oooo0080oooo00@000000?ooo`3o
ool0o`000`3oool:003o00030000003o0000000000T0000000<0o`00000000000000500000000`3o
000000000?ooo`08003o00040?ooo`000000oooo0?l000<0oooo00L000000?ooo`3oool0oooo0?l0
003oool000000080oooo1P00o`090?ooo`030000003oool0oooo01@0oooo000]0?ooo`030000003o
ool0oooo00P0oooo00<0o`000?ooo`3o0000h`3oool00`000000o`000?ooo`020?ooo`070000003o
ool0oooo0?ooo`3o0000oooo0000000;003o00030000003o0000oooo00<0oooo00L000000?ooo`3o
ool0oooo0?l0003oool0000000T00?l03P0000020?ooo`L00?l00P3oool00`000000o`000?ooo`03
0?ooo`080000003oool0oooo0?ooo`3o0000oooo0000003oool7003o0080oooo00@000000?ooo`3o
ool0o`000P3oool01`000000oooo0?ooo`3oool0o`000?ooo`0000000`3oool5003o0200oooo000]
0?ooo`030000003oool0oooo00P0oooo00@0o`000?ooo`3oool0o`00hP3oool00`000000o`000?oo
o`06000000040?ooo`3o000000000?ooo`X00?l000@000000?l0003oool0oooo5@0000000`3o0000
000000000002000000<0oooo00<0o`000?ooo`0000000`3oool6003o0080oooo00<000000?ooo`3o
00000`3oool01`000000oooo0?ooo`3oool0o`000?ooo`0000000P3oool5003o00H0oooo00<0o`00
0?ooo`3oool00`3oool00`3o0000oooo0?ooo`040?ooo`@00?l08@3oool002d0oooo00<000000?oo
o`3oool01`3oool00`3o0000oooo0?ooo`020?ooo`030?l0003oool0oooo0<d0oooo00<0o`000?oo
o`3oool03`3oool00`3o0000000000000008000000030?l0000000000?l000D00?l03`000000103o
000000000?ooo`3oool7003o00@0oooo00@0o`000?ooo`3oool000000P3oool00`3o0000oooo0?oo
o`040?ooo`D00?l01P3oool00`3o0000oooo0?ooo`030?ooo`030?l0003oool0oooo00<0oooo1@00
o`070?ooo`050?l0003oool0oooo0?ooo`3o0000203oool3003o0240oooo000X0?ooo``00?l00P3o
ool00`3o0000oooo0?ooo`030?ooo`030?l0003oool0oooo0<`0oooo00<0o`000000003o0000103o
ool<003o00030000003o0000oooo00H0000000<0oooo0?l0000000001`000003003o00030?ooo`00
0000o`000080oooo1@000000103oool0o`000?ooo`0000020?ooo`H00?l01P3oool00`3o0000oooo
000000020?ooo`030?l0003oool0oooo00D0oooo0`00o`070?ooo`030?l0003oool0oooo0080oooo
00<0o`000?ooo`3oool01@3oool3003o00T0oooo00<0o`000?ooo`3o00002@3oool2003o0280oooo
000R0?ooo`030000003oool0oooo00@0oooo2P00o`0010000000oooo0?l0003oool5000000040?oo
o`3o0000oooo00000<X0oooo00D0o`000?ooo`000000oooo0?l000040?ooo`X00?l000<0oooo0000
003o00000P3oool800000080oooo1`00o`020?ooo`030000003oool0o`0000<0oooo00L000000?oo
o`3oool0oooo0?l0003oool0000000<0oooo1@00o`070?ooo`040?l0003oool0oooo0?l000P0oooo
0`00o`080?ooo`040?l0003oool0oooo0?l000P0oooo0`00o`090?ooo`030?l0003oool0o`0000X0
oooo00<00?l00?ooo`3oool0803oool00280oooo00<000000?ooo`3oool0103oool:003o00030000
003oool0o`0000<0oooo00<000000?ooo`3oool00P3oool00`3o000000000?ooo`380?ooo`030?l0
003oool0oooo00<0000000@0o`0000000000000000001`00o`0:000000<0oooo00<0o`000?ooo`3o
ool00`3oool5003o00H0oooo00<0o`000?ooo`3oool00`3oool00`3o0000oooo0?ooo`050?ooo`<0
0?l0203oool00`3o0000oooo0?l0000:0?ooo`03003o003oool0oooo00P0oooo00<0o`000?ooo`3o
00002@3oool00`00o`00oooo0?ooo`090?ooo`030?l0003oool0oooo02`0oooo000R0?ooo`P00000
0`00o`0400000004003o0000000000000?l000X0000000<0o`000?ooo`3oool0a`3oool0103o0000
oooo0?ooo`0000020?ooo`030?l0003oool0oooo0080oooo00<00?l00000000000000`000002003o
00@0oooo00@0o`000?ooo`3oool000000P3oool00`3o0000oooo0?ooo`040?ooo`D00?l01P3oool0
0`3o0000oooo0?ooo`020?ooo`030?l0003oool0oooo00H0oooo0`00o`090?ooo`030?l0003oool0
oooo01D0oooo00<0o`000?ooo`3oool0@`3oool00280oooo00<000000?ooo`3oool01@3oool7003o
00D000000`3oool00`000000oooo0?ooo`020?ooo`030?l000000000oooo0<D0oooo00<000000?oo
o`3o00000`3oool01@000000oooo0?ooo`3oool0o`0000@000001`00o`050?ooo`050?l0003oool0
00000?ooo`3o0000203oool3003o00P0oooo00@0o`000?ooo`3oool0o`002P3oool00`00o`00oooo
0?ooo`1V0?ooo`008P3oool00`000000oooo0?ooo`060?ooo`H00?l00P3oool00`000000oooo0?l0
0006000000040?ooo`3o0000oooo00000<H0oooo00@000000?l0003oool0oooo1`000000103o0000
00000?ooo`3oool6003o00L0oooo00<0o`000000003o00002@3oool3003o00T0oooo00<0o`000?oo
o`3o0000L`3oool002/0oooo1@00o`060?ooo`040?l0003oool0oooo0?ooo`@00000b03oool00`3o
0000000000000009000000040?l0003oool0oooo0?ooo`D00?l01`3oool00`3o0000oooo0?l0000:
0?ooo`03003o003oool0oooo00T0oooo00<0o`000?ooo`3oool0LP3oool002`0oooo0`00o`070?oo
o`030?l0003oool0oooo00<0oooo00@0o`000?ooo`0000000000]@3oool00`000000oooo0?ooo`03
0?ooo`/00?l0100000020?ooo`030000003oool0oooo0080oooo00<0o`000000003oool00`3oool3
003o00T0oooo00<0o`000?ooo`3oool0RP3oool002`0oooo0`00o`080?ooo`050?l0003oool00000
0?ooo`3o00000`3oool01000o`000000000000000008003o00H0oooo00<0o`000?ooo`3oool0X@3o
ool00`000000oooo0?ooo`040?ooo`P00?l00`0000000`3o0000oooo0?ooo`05000000040?ooo`3o
0000oooo000000@0oooo0`00o`2F0?ooo`00;@3oool00`00o`00oooo0?ooo`080?ooo`030?l00000
0000o`0000D0oooo0`00o`03000000@00?l01P3oool00`3o000000000?l0002R0?ooo`030?l0003o
ool0oooo00@0oooo1P00o`02000000800?l000<000000?ooo`3o00000`3oool01`000000oooo0?oo
o`3oool0o`000?ooo`0000001@3oool00`00o`00oooo0?ooo`2E0?ooo`00;@3oool00`000000oooo
0?ooo`080?ooo`030?l000000000o`0000D0oooo1P00o`04000000D0oooo00@0o`000?ooo`000000
o`00X@3oool00`3o000000000?l00006000000800?l00`000003003o00@0000000@0o`0000000000
000000000P3oool00`3o0000oooo0?ooo`2N0?ooo`00;@3oool3000000T0oooo00<0o`0000000000
00001@000008003o00D0000000H0o`0000000000000000000?ooo`3o002O0?ooo`040?l0003oool0
00000?l000D0oooo0`000005003o00H0oooo00D0o`000?ooo`000000oooo0?l0002Q0?ooo`00;@3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00H0oooo1`00o`030?ooo`80000000<0
o`000000000000000`0000000`3o0000oooo0?ooo`2L0?ooo`060?l0003oool000000?ooo`3o0000
oooo0`0000020?ooo`H00?l01P3oool0103o0000oooo0000003o002R0?ooo`00;@3oool00`000000
oooo0?ooo`090?ooo`030000003oool0oooo00H0oooo1P00o`030?ooo`030000003oool0o`000080
000000<0oooo0000003oool00P3oool00`3o000000000?ooo`2J0?ooo`030?l0003oool0oooo0080
oooo0P0000050?ooo`D00?l0203oool00`3o0000oooo0?l0002R0?ooo`00;@3oool00`000000oooo
0?ooo`0C0?ooo`D00?l00`3oool00`000000o`000?ooo`020?ooo`8000000`3oool00`3o00000000
0?ooo`2I0?ooo`030?l000000000000000@0000000<0o`000?ooo`3oool0103oool3003o00X0oooo
00<0o`000?ooo`3oool0X@3oool002d0oooo00<000000?ooo`3oool0503oool3003o00@0oooo00<0
o`00000000000000200000000`3o0000oooo0?ooo`2F0?ooo`040000003o0000oooo0?ooo`<00000
0`3oool00`3o000000000?ooo`030?ooo`<00?l0[P3oool002d0oooo00<000000?ooo`3oool0503o
ool3003o00@0oooo00<000000?l0003oool00`3oool01P000000oooo000000000000o`00000009P0
oooo00@000000?l00000000000000P3oool01P000000oooo0?ooo`3oool0o`00000000D0oooo00<0
0?l00?ooo`3oool0[@3oool002d0oooo00<000000?ooo`3oool05@3oool00`00o`00oooo0?ooo`03
0?ooo`030000003o0000oooo00<0oooo00@000000?ooo`3oool0oooo0P00000<003o00D0oooo00<0
00000?ooo`3oool0LP3oool00`000000oooo0?ooo`030?ooo``00?l02`0000000`3o0000oooo0?oo
o`2c0?ooo`00;@3oool00`000000oooo0?ooo`0M0?ooo`030?l000000000000000@0000000<0o`00
0?ooo`3oool00P000009003o00H0oooo00<0o`000?ooo`3oool0LP3oool00`000000oooo0?ooo`04
0?ooo`T00?l00`0000000`3o0000oooo0?ooo`020?ooo`060000003oool0oooo0?ooo`3o00000000
]@3oool002d0oooo00<000000?ooo`3oool07P3oool00`3o0000oooo0?ooo`02000000030?l00000
000000000080000000<00?l000000000o`001`00o`05000000030?l000000000o`0007<0oooo00<0
00000?ooo`3oool0103oool7003o0080000000D00?l00?ooo`000000oooo0?l00006000000030?oo
o`3o000000000;D0oooo000]0?ooo`030000003oool0oooo01l0oooo00D0o`000?ooo`000000oooo
0?l000050?ooo`03003o00000000000000D00?l01@3oool0103o0000oooo0000003o001c0?ooo`P0
00001@00o`000`0000000?l0003o0005000000040?l000000000000000000080oooo00<0o`000?oo
o`3oool0]@3oool002d0oooo0`00000O0?ooo`040?l0003oool000000?l000H0oooo0`00o`001000
00000?l0003o0000o`060?ooo`050?l00000000000000000003o0000LP3oool00`000000oooo0?oo
o`060?ooo`<00?l000@00000003o0000o`000?l01@3oool01P3o0000oooo0?ooo`000000oooo0?l0
0;P0oooo000]0?ooo`030000003oool0oooo0200oooo00<0o`000000003o00001`3oool3003o0080
000000<00?l00?ooo`3oool00`3oool0103o0000oooo0?ooo`0000020?ooo`030?l0003oool0oooo
06l0oooo00<000000?ooo`3oool01P3oool00`00o`00000000000003003o00L0oooo00@0o`000?oo
o`000000o`00^@3oool002d0oooo00<000000?ooo`3oool08@3oool00`3o0000oooo0?ooo`060?oo
o`D00?l00P0000030?ooo`050?l0003oool0oooo0?ooo`0000000`3oool00`3o0000oooo0?ooo`1g
0?ooo`0300000000o`000?l000<00?l0203oool00`3o0000oooo0?l0002i0?ooo`00;@3oool00`00
0000oooo0?ooo`0[0?ooo`<00?l00`3oool2000000030?l0003oool0oooo0080oooo00L000000?oo
o`3oool0oooo0?l0003oool0000007H0oooo00<000000?ooo`3oool00`00o`0:0?ooo`030?l0003o
ool0oooo0;P0oooo000]0?ooo`030000003oool0oooo02/0oooo0`00o`040?ooo`030000003o0000
oooo00<0oooo00<000000?ooo`3oool00P3oool00`3o000000000?ooo`1d0?ooo`040000003oool0
oooo0?ooo`<00?l0a@3oool00140oooo00D0omnL04P000000000001d0;ooo`030?ooo`050?noM000
0000M;oo0?ooo`2oM0001P0000000`18W=l0oooo0?ooo`060?ooo`030000003oool0oooo02`0oooo
00<00?l00?ooo`3oool00`3oool00`3o0000000000000009000000030?l0003oool0oooo06`0oooo
00<0o`000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03003o003oool0oooo0<@0oooo
000@0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo0080
oooo00<0gia80000000000000`0000000`00B9`0gooo0?ooo`060?ooo`030000003oool0oooo0380
oooo00<000000?l0000000000`3oool00`000000oooo0?ooo`020?ooo`030?l000000000oooo06`0
oooo00<0o`000?ooo`3o00000`3oool00`000000oooo0?ooo`3;0?ooo`00403oool00`2oM00004RL
0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo
0?ooo`060?ooo`030000003oool0oooo0380oooo00@000000?ooo`3o000000000P3oool01`000000
oooo0?ooo`3oool0o`000?ooo`000000K@3oool00`3o0000oooo0?l000020?ooo`030000003oool0
oooo0<`0oooo000?0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool00`3O
W4P0M;oo0?ooo`0:0?ooo`030000003oool0oooo03D0oooo00@0o`000000003oool000000P3oool0
0`3o0000oooo0?ooo`1]0?ooo`030?l000000000000000<00000c`3oool000l0oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`090?ooo`030;mdB02Lgol0oooo00T0oooo00<000000?oo
o`3oool0=@3oool0103o0000oooo0000000000020?ooo`030?l0003oool0oooo06`0oooo00<0o`00
0?ooo`3oool00P0000000`3oool0o`000?ooo`3>0?ooo`003`3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00T0oooo00<0oooO09a8M02oool02@3oool4000003D0oooo00D0o`000000
0000000000000?l0001_0?ooo`040?l0003oool0oooo000000<0oooo00<0o`000?ooo`3oool0c03o
ool000l0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030?noM000M;l0
oooo00P0oooo00<000000?ooo`3oool0=`3oool00`3o0000oooo0000001]0?ooo`030000003oool0
o`000080oooo0P0000030?ooo`030?l0003oool000000<`0oooo000?0?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol02P3oool00`3ookl0M01809cOo`080?ooo`030000003oool0oooo
03P0oooo00<0o`000?ooo`000000K03oool01`000000oooo0?l0003oool000000?ooo`000000103o
ool00`3o000000000?ooo`3;0?ooo`003`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0
_ooo00X0oooo00<0oono07@00018W=l0203oool00`000000oooo0?ooo`0j0?ooo`030000003oool0
oooo06X0oooo00<000000?l0000000002@0000000`3o0000oooo0?ooo`3:0?ooo`00403oool00`3O
W4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00D0oooo00<0oknL09cOo`3oool00P3oool00`3o
gi`0B00004RLg`080?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0J@3oool01000
0000oooo0?l0000000020?ooo`030000003oool0oooo0080oooo00<0o`000000003oool0b`3oool0
0100oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00H0oooo00P0oooO09a8B02Lgol0oooo
0?oo_`1d0000001d0;ooo`P0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`1X0?oo
o`030000003oool0000000<0oooo00L000000?ooo`3oool0oooo0?l0003oool000000<`0oooo000A
0?ooo`050?oo_`1d000000000000M02oool0203oool01P3ooml0W4P0000000000000001d0;ooo`T0
oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`1Y0?ooo`050000003o0000oooo0?oo
o`0000000P3oool00`3o0000oooo0?ooo`3=0?ooo`00;@3oool00`000000oooo0?ooo`0m0?ooo`03
0000003oool0oooo06P0oooo00H000000?ooo`3oool0o`000?ooo`0000020?ooo`030?l0003oool0
oooo0<d0oooo000]0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool0IP3oool01P00
0000oooo0?ooo`3oool000000?l00080000000<0o`000?ooo`3oool0cP3oool002d0oooo00<00000
0?ooo`3oool0?`3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo0080oooo00<0o`00
0?ooo`3o0000d@3oool002d0oooo0`0000100?ooo`030000003oool0oooo06<0oooo00<000000?oo
o`3oool0103oool00`3o0000oooo0?ooo`3@0?ooo`00;@3oool00`000000oooo0?ooo`100?ooo`03
0000003oool0oooo0680oooo00<000000?ooo`3oool0f03oool002d0oooo00<000000?ooo`3oool0
@@3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo0=P0oooo000]0?ooo`030000003o
ool0oooo0440oooo00<000000?ooo`3oool0H03oool00`000000oooo0?ooo`3I0?ooo`00;@3oool0
0`000000oooo0?ooo`120?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool0fP3oool0
02d0oooo00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo
0=X0oooo000]0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0EP3oool6003o0003
00000000o`000?l000<00?l0f03oool002d0oooo00<000000?ooo`3oool0A03oool00`000000oooo
0?ooo`1?0?ooo`030000003oool0oooo00@0oooo1000o`000`0000000?l0003o0003003o00H0oooo
00<000000?ooo`3oool0d03oool002d0oooo00<000000?ooo`3oool0?P3oool6003o000300000000
o`000?l000<00?l0C03oool00`000000oooo0?ooo`040?ooo`@00?l000<00000003o0000o`000`00
o`060?ooo`030000003oool0oooo0=00oooo000]0?ooo`030000003oool0oooo03P0oooo00<00000
0?ooo`3oool0103oool6003o000400000000o`000?l0003o00H0oooo00<000000?ooo`3oool0A03o
ool8000000800?l000<00000003o0000o`000`00o`0800000=80oooo000]0?ooo`<00000>03oool0
0`000000oooo0?ooo`040?ooo`H00?l000@00000003o0000o`000?l01P3oool00`000000oooo0?oo
o`140?ooo`030000003oool0oooo00D0oooo00<00?l000000000o`001000o`080?ooo`030000003o
ool0oooo0=00oooo000]0?ooo`030000003oool0oooo03P0oooo20000006003o000300000000o`00
000000L00000AP3oool00`000000oooo0?ooo`060?ooo`0300000000o`000?l000<00?l0203oool0
0`000000oooo0?ooo`3@0?ooo`00;@3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo
00D0oooo1P00o`000`0000000?l00?ooo`060?ooo`030000003oool0oooo04`0oooo00<00000003o
0000o`000`00o`000`000000oooo0?ooo`3I0?ooo`00;@3oool00`000000oooo0?ooo`0h0?ooo`03
0000003oool0oooo00H0oooo1P00o`000`000000oooo0?ooo`050?ooo`030000003oool0oooo04`0
oooo00<000000?ooo`00o`000P00o`3M0?ooo`00;@3oool00`000000oooo0?ooo`120?ooo`D00?l0
00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`03003o0=d0oooo000]0?ooo`030000003o
ool0oooo0480oooo1000o`020?ooo`030000003oool0oooo0580oooo00D000000?ooo`3oool0oooo
003o003N0?ooo`00;@3oool00`000000oooo0?ooo`130?ooo`<00?l00P3oool00`000000oooo0?oo
o`1A0?ooo`030000003oool0oooo0>40oooo000]0?ooo`030000003oool0oooo04<0oooo0P00o`04
0?ooo`030000003oool0oooo04L0oooo1P0000030?ooo`030000003oool0oooo0>40oooo000]0?oo
o`030000003oool0oooo04@0oooo00<00?l00?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool0h@3oool002d0oooo00<000000?ooo`3o
ool0BP3oool00`000000oooo0?ooo`190?ooo`030?l0003oool0oooo0080oooo00<000000?ooo`3o
ool0hP3oool002d0oooo0`00001:0?ooo`030000003oool0oooo04P0oooo00<0o`000000003o0000
0`3oool00`000000oooo0?ooo`3R0?ooo`00;@3oool00`000000oooo0?ooo`1;0?ooo`030000003o
ool0oooo04H0oooo00@0o`000?ooo`000000o`000P3oool00`000000oooo0?ooo`3S0?ooo`00;@3o
ool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo04D0oooo00P0o`000?ooo`3oool00000
0?ooo`3o0000oooo00000>D0oooo000]0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3o
ool0A03oool0103o0000oooo0?ooo`0000020?ooo`030?l000000000oooo0>@0oooo000]0?ooo`03
0000003oool0oooo04`0oooo1`00000m0?ooo`030000003oool0o`0000<0oooo00L000000?ooo`3o
ool000000?l0003oool000000><0oooo000]0?ooo`030000003oool0oooo04d0oooo00<000000?oo
o`000000@03oool00`000000o`000?ooo`030?ooo`070000003oool0oooo0000003o0000oooo0000
003S0?ooo`004@3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00<0okmd0000001d_ol01@3o
ool0103ooml0W4P00018W03Oool80?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3o0000
@03oool00`3o0000000000000008000000030?l000000000oooo0>80oooo000@0?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo00D0oooo00@0oooO09a80000
B9`0gooo203oool00`000000oooo0?ooo`1>0?ooo`80000000<0o`000?ooo`3oool0?@3oool00`00
0000o`000?ooo`030?ooo`070000003oool000000?ooo`3o0000oooo0000003S0?ooo`00403oool0
0`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00T0oooo00@0oooO09a80000B9`0gooo203o
ool00`000000oooo0?ooo`1>0?ooo`80000000<0oooo0?l0003oool0?@3oool00`000000o`000?oo
o`030?ooo`070000003oool000000?ooo`3o0000oooo0000003S0?ooo`003`3oool00`3ookl0M000
04RLg`030?ooo`030=nLB00007@0_ooo00D0oooo00<0_g@00000000000001@0000000`1d_ol0oooo
0?ooo`050?ooo`030000003oool0oooo04d0oooo00<0o`000000000000000P3oool00`3o0000oooo
0?ooo`0m0?ooo`040?l0003oool0oooo0?ooo`80000000<0oooo0?l0003oool0i@3oool000l0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`050?ooo`080?oOW01dW=l0oooo0?ooo`3o
oml0W4P00018W03Oool80?ooo`030000003oool0oooo04`0oooo00@0o`000?ooo`3oool000000P3o
ool00`3o0000oooo0?ooo`0n0?ooo`030?l0003oool0oooo0080000000<0o`000?ooo`3oool0i@3o
ool000l0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`070=nLB01d_ol0
oooo0?oog`2LB00004RL0=ooo`080?ooo`@00000B@3oool00`000000o`000?ooo`020?ooo`070000
003oool0oooo0?ooo`3o0000oooo0000000n0?ooo`050?l0003oool000000000003o0000i`3oool0
00l0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`070?oog`2LB7@0_ooo
0?oog`2LB00004RL0=ooo`080?ooo`030000003oool0oooo04X0oooo00<000000?l0003oool00P3o
ool00`000000oooo0?ooo`020?ooo`030?l000000000oooo03d0oooo00@0o`000?ooo`000000o`00
j03oool000l0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`060?oOW018
M;l0oooO09a80000B9`0gooo203oool00`000000oooo0?ooo`1:0?ooo`030?l000000000000000P0
000000<0o`000?ooo`3oool0?@3oool00`3o000000000?l0003X0?ooo`003`3oool00`3ookl0M000
04RLg`030?ooo`030=nLB00007@0_ooo00P0oooo00D0gia807Bog`2LB00004RL0=ooo`080?ooo`03
0000003oool0oooo04X0oooo00<000000?l0003oool00P3oool2000000<0oooo00<0o`000000003o
ool0?`3oool00`000000oooo0?ooo`3W0?ooo`00403oool00`3OW4P004RL0=ooo`020?ooo`030;md
0000M;l0oooo00P0oooo00D0oooO09a8B018B00004RL0=ooo`080?ooo`030000003oool0oooo04X0
oooo00<000000?l0003oool00P3oool200000080oooo00<0o`000?ooo`000000?`3oool200000>T0
oooo000@0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0:0?ooo`040?oOW018000004RL
0=ooo`P0oooo00<000000?ooo`3oool0C03oool0203o0000oooo0?ooo`000000oooo0000003oool0
o`00?`3oool600000>L0oooo000A0?ooo`050?oo_`1d000000000000M02oool0303oool00`3OW4P0
04RL0=ooo`080?ooo`030000003oool0oooo04`0oooo00L0o`000?ooo`3oool000000?ooo`000000
o`000480oooo00<000000?ooo`3oool0j03oool002d0oooo00<000000?ooo`3oool0C@3oool01@3o
0000oooo0000003oool000000480oooo00<000000?ooo`3oool0j@3oool002d0oooo00<000000?oo
o`3oool0CP3oool01@3o000000000?l0003oool000000440oooo00<000000?ooo`3oool0j@3oool0
02d0oooo00<000000?ooo`3oool0CP3oool01@3o000000000?l0003oool000000440oooo00<00000
0?ooo`3oool0j@3oool002d0oooo0`00001?0?ooo`050?l0003oool0oooo0?ooo`000000?`3oool0
0`000000oooo0?ooo`3Z0?ooo`00;@3oool00`000000oooo0?ooo`1?0?ooo`050000003oool0oooo
0?ooo`000000?`3oool00`000000oooo0?ooo`3Z0?ooo`00;@3oool00`000000oooo0?ooo`1=0?oo
o`L00000?P3oool00`000000oooo0?ooo`3[0?ooo`00;@3oool00`000000oooo0?ooo`1D0?ooo`03
0000003oool0oooo03/0oooo00<000000?ooo`3oool0j`3oool002d0oooo00<000000?ooo`3oool0
E03oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo0>/0oooo000]0?ooo`030000003o
ool0oooo05@0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`3/0?ooo`00;@3oool0
0`000000oooo0?ooo`1E0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0k03oool0
02d0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo
0>`0oooo000]0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0>03oool00`000000
oooo0?ooo`3]0?ooo`00;@3oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo03L0oooo
00<000000?ooo`3oool0k@3oool002d0oooo0`00001F0?ooo`030000003oool0oooo03L0oooo00<0
00000?ooo`3oool0k@3oool002d0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`0f
0?ooo`030000003oool0oooo0>h0oooo000]0?ooo`030000003oool0oooo05H0oooo00<000000?oo
o`3oool0=P3oool00`000000oooo0?ooo`3^0?ooo`00;@3oool00`000000oooo0?ooo`1G0?ooo`03
0000003oool0oooo03D0oooo00<000000?ooo`3oool0kP3oool002d0oooo00<000000?ooo`3oool0
E`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo0>h0oooo000]0?ooo`030000003o
ool0oooo05L0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`3_0?ooo`00;@3oool0
0`000000oooo0?ooo`1G0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0k`3oool0
02d0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo
0>l0oooo000]0?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool0<`3oool00`000000
oooo0?ooo`3_0?ooo`00;@3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo0380oooo
00<000000?ooo`3oool0l03oool002d0oooo0`00001H0?ooo`030000003oool0oooo0380oooo00<0
00000?ooo`3oool0l03oool002d0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`0b
0?ooo`030000003oool0oooo0?00oooo000]0?ooo`030000003oool0oooo05T0oooo00<000000?oo
o`3oool0<@3oool00`000000oooo0?ooo`3`0?ooo`00;@3oool00`000000oooo0?ooo`1I0?ooo`03
0000003oool0oooo0300oooo00<000000?ooo`3oool0l@3oool002d0oooo00<000000?ooo`3oool0
F@3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo0?40oooo000A0?ooo`050?oOW018
000000000000M02oool00`3oool00`3o_g@0000007Boo`030?ooo`050?oOW018000000000000001d
_ol02@3oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3o
ool0l@3oool00100oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00<0okmd0000
001d_ol00P3oool01`3ogi`0B00004RLg`3oool0oono07@0001d_ol0203oool00`000000oooo0?oo
o`1J0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0l@3oool00100oooo00<0_g@0
0018W03Oool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?oog`2LB00007Bo00<0oooo00<0okmd
0018W03Oool01`3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo02/0oooo1@00003c
0?ooo`003`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00D0oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo05X0oooo00<000000?oo
o`3oool0;@3oool200000?@0oooo000?0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol01@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3o
ool0F@3oool5000002`0oooo0P00003d0?ooo`003`3oool00`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00D0oooo00<0omnL04P0001d_ol00`3oool00`2oM000001809cOo`070?ooo`@00000
F@3oool2000002h0oooo0P00003d0?ooo`003`3oool00`3ogi`0B00007Boo`030?ooo`030?noM000
04P0W=oo00D0oooo00P0oono07@000000000B9cO0?ooo`3o_g@00000001d_`P0oooo00<000000?oo
o`3oool0F`3oool00`000000oooo0?ooo`0W0?ooo`D00?l000<00000003o0000o`001000o`3_0?oo
o`003`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00<0_g@00018W02o
W4P00P0000000`00B9`0gooo0?ooo`070?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3o
ool0803oool00`000000oooo0?ooo`050?ooo`@00?l000<00000003o0000o`000`00o`060?ooo`03
0000003oool0oooo0>L0oooo000?0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0
1P3oool00`3ogi`0B00007Boo`0<0?ooo`030000003oool0oooo05H0oooo1@00o`000`0000000?l0
003o0004003o01`0oooo00<000000?ooo`3oool01@3oool4003o000300000000o`000?l000<00?l0
1P3oool00`000000oooo0?ooo`3W0?ooo`00403oool00`3OW4P004RL0=ooo`020?ooo`030;md0000
M;l0oooo00L0oooo00<0gia80000B02Lgol02`3oool00`000000oooo0?ooo`1?0?ooo`030000003o
ool0oooo00D0oooo1000o`000`0000000?l0003o0003003o00H0oooo00<000000?ooo`3oool0503o
ool9000000<00?l000<00000003o0000o`000P00o`0800000>T0oooo000@0?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`030?noM000000007Bo00X0oooo00<000000?ooo`3oool0
C`3oool00`000000oooo0?ooo`050?ooo`D00?l000<00000003o0000o`000P00o`060?ooo`030000
003oool0oooo01@0oooo00<000000?ooo`3oool01P3oool3003o000400000000o`000?l0003o00P0
oooo00<000000?ooo`3oool0i`3oool00140oooo00D0oono07@000000000001d0;ooo`0;0?ooo`05
0?oog`2LB00000000000M02oool01`3oool00`000000oooo0?ooo`1?0?ooo`T000001000o`001000
00000?l0003o0000o`08000001H0oooo00<000000?ooo`3oool01`3oool00`00o`000000003o0003
003o00P0oooo00<000000?ooo`3oool0i`3oool002d0oooo00<000000?ooo`3oool0C`3oool00`00
0000oooo0?ooo`060?ooo`@00?l000<00000003o0000o`00203oool00`000000oooo0?ooo`0N0?oo
o`03003o000000000?l000800?l0l`3oool002d0oooo00<000000?ooo`3oool0C`3oool00`000000
oooo0?ooo`070?ooo`<00?l000<00000003o0000o`00203oool00`000000oooo0?ooo`0O0?ooo`03
00000000o`000?l00?@0oooo000]0?ooo`030000003oool0oooo05T0oooo0`00o`000`0000000?l0
0?ooo`0Z0?ooo`0300000000o`000?l00?@0oooo000]0?ooo`<00000FP3oool3003o00030000003o
ool0oooo02P0oooo00<000000?ooo`00o`00m@3oool002d0oooo00<000000?ooo`3oool0FP3oool3
003o00030000003oool0oooo02P0oooo1@00003c0?ooo`00;@3oool00`000000oooo0?ooo`1K0?oo
o`03003o003oool0000002X0oooo00<000000?ooo`3oool0m@3oool002d0oooo00<000000?ooo`3o
ool0F@3oool5000002X0oooo00<000000?ooo`3oool0m@3oool002d0oooo00<000000?ooo`3oool0
GP3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0?H0oooo000]0?ooo`030000003o
ool0oooo05h0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`3f0?ooo`00;@3oool0
0`000000oooo0?ooo`1N0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0mP3oool0
02d0oooo00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`040?ooo`D000007@3oool00`00
0000oooo0?ooo`3f0?ooo`00;@3oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`3g0?ooo`00;@3oool00`000000oooo
0?ooo`1O0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0703oool00`000000oooo
0?ooo`3g0?ooo`00;@3oool3000005l0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`0L0?ooo`030000003oool0oooo0?L0oooo000]0?ooo`030000003oool0oooo05l0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo0?L0oooo000]0?oo
o`030000003oool0oooo05l0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0L0?oo
o`030000003oool0oooo0?L0oooo000]0?ooo`030000003oool0oooo0600oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo0?P0oooo000]0?ooo`030000
003oool0oooo0600oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0K0?ooo`030000
003oool0oooo0?P0oooo000]0?ooo`030000003oool0oooo0600oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo0?P0oooo000]0?ooo`030000003oool0
oooo0600oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0K0?ooo`030000003oool0
oooo0?P0oooo000]0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo0?T0oooo000]0?ooo`030000003oool0oooo0640
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0?T0
oooo000]0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo0?T0oooo000]0?ooo`<00000H@3oool00`000000oooo0?oo
o`030?ooo`030?l0003oool0oooo01X0oooo00<000000?ooo`3oool0n@3oool002d0oooo00<00000
0?ooo`3oool0H@3oool00`000000oooo0?ooo`020?ooo`030?l000000000o`0001X0oooo00<00000
0?ooo`3oool0nP3oool002d0oooo00<000000?ooo`3oool0HP3oool020000000oooo0?ooo`3o0000
oooo0000003oool0o`006@3oool00`000000oooo0?ooo`3j0?ooo`00;@3oool00`000000oooo0?oo
o`1R0?ooo`030000003oool0o`000080oooo00@000000?ooo`3oool0o`00603oool00`000000oooo
0?ooo`3j0?ooo`00;@3oool00`000000oooo0?ooo`1R0?ooo`030000003oool0o`000080oooo00@0
00000?ooo`3oool0o`00603oool00`000000oooo0?ooo`3j0?ooo`004@3oool01@3ogi`0B0000000
000007@0_ooo00<0oooo00<0okmd0000001d_ol00`3oool01@3OW4P00000000000000000B9cO00T0
oooo00<000000?ooo`3oool0H@3oool2000000050?l0003oool0oooo0?ooo`0000000`3oool00`3o
0000oooo0000000D0?ooo`030000003oool0oooo0?/0oooo000@0?ooo`070?oo_`1d0000M;oo0?oo
o`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo0080oooo00L0gia80000001d_ol0oooo0?oo
g`2LB000B9cO00P0oooo00<000000?ooo`3oool0H@3oool00`000000o`00000000030?ooo`030000
003oool0oooo0080oooo00<0o`000000003oool04`3oool00`000000oooo0?ooo`3k0?ooo`00403o
ool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00D0oooo00<0oooO09a80000M;l00`3o
ool00`3o_g@004RL0=ooo`070?ooo`030000003oool0oooo0640oooo00<0o`000000000000002@00
00000`3o0000oooo0?ooo`0B0?ooo`030000003oool0oooo0?/0oooo000?0?ooo`030?oo_`1d0000
B9cO00<0oooo00<0gia80000M02oool01P3oool00`2oM00007Bo0?ooo`020?ooo`030;md0000B9`0
gooo00L0oooo00<000000?ooo`3oool0H@3oool00`000000o`00000000030?ooo`030000003oool0
oooo0080oooo00<0o`000000003oool03@3oool700000?d0oooo000?0?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol01P3oool01`3ogi`0B01d0;ooo`3oool0_g@00000001d_ol0203o
ool00`000000oooo0?ooo`1Q0?ooo`030000003o0000000000<0oooo00L000000?ooo`3oool0oooo
0?l0003oool000000100oooo00@000000?ooo`3oool00000oP3oool000l0oooo00<0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`070?ooo`050?oo_`1dB4P0B00000000018W=l02@3oool40000
0600oooo00<000000?ooo`0000000`3oool01`000000oooo0?ooo`3oool0o`000?ooo`000000403o
ool010000000oooo0?ooo`00003n0?ooo`003`3oool00`3ogi`0B00007Boo`030?ooo`030?noM000
04P0W=oo00L0oooo00@0omnL04P00000000007Bo2P3oool00`000000oooo0?ooo`1T0?ooo`040000
003oool0oooo00000080oooo00<0o`000?ooo`3oool04@3oool010000000oooo0?ooo`00003n0?oo
o`003`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00H0omnL04P00000
07@0_ooo0?noM000M;l90?ooo`030000003oool0oooo06@0oooo00H000000?l0003oool000000?oo
o`3o000D0?ooo`040000003oool0oooo00000?h0oooo000?0?ooo`030?oo_`1d0000B9cO00<0oooo
00<0gia80000M02oool01P3oool00`2oM000001d0;ooo`020?ooo`030;md0018W=l0oooo00L0oooo
00<000000?ooo`3oool0I03oool01@000000o`000?ooo`000000o`0001D0oooo00<000000?ooo`00
0000o`3oool00100oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`060?ooo`030;md
0000B9`0gooo0080oooo00<0gia8001d_`3oool01`3oool00`000000oooo0?ooo`1T0?ooo`050000
003oool0o`000000003o00005@3oool00`000000oooo0000003o0?ooo`00403oool01`3ookl0M000
07Boo`3oool0okmd0000M02oool01`3oool01`3o_g@0000007Boo`3oool0oono07@0001d_ol0203o
ool00`000000oooo0?ooo`1U0?ooo`030000003oool0o`0001H0oooo00<000000?ooo`000000o`3o
ool00140oooo00D0oono07@000000000001d0;ooo`090?ooo`050?noM0000000000000000018W=l0
2@3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0000001H0oooo00<000000?ooo`000000
o`3oool002d0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0000000F0?ooo`800000o`3o
ool10?ooo`00;@3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0000001H0oooo0P00003o
0?ooo`40oooo000]0?ooo`030000003oool0oooo06H0oooo0P00000F0?ooo`800000o`3oool10?oo
o`00;@3oool3000006H0oooo0P00000F0?ooo`800000o`3oool10?ooo`00;@3oool00`000000oooo
0?ooo`1V0?ooo`8000005P3oool00`000000oooo0?ooo`3o0?ooo`00;@3oool00`000000oooo0?oo
o`1V0?ooo`8000005@3oool00`3o000000000?l0003o0?ooo`40oooo000]0?ooo`030000003oool0
oooo06L0oooo00<000000?ooo`3oool04`3oool0103o000000000?ooo`3o003o0?ooo`00;@3oool0
0`000000oooo0?ooo`1W0?ooo`030000003oool0oooo0180oooo00<0o`000?ooo`0000000P3oool0
0`3o0000oooo0?ooo`3l0?ooo`00;@3oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo
0140oooo00@0o`000?ooo`00000000000P3oool00`3o0000oooo0?ooo`3l0?ooo`00;@3oool00`00
0000oooo0?ooo`1W0?ooo`8000003`3oool00`000000oooo0?l000020?ooo`8000000`3oool00`3o
0000oooo0000003k0?ooo`00;@3oool00`000000oooo0?ooo`1U0?ooo`D000003P3oool00`000000
oooo0?l000020?ooo`800000103oool00`3o000000000?ooo`3j0?ooo`00;@3oool00`000000oooo
0?ooo`1X0?ooo`030000003oool0oooo00d0oooo00<000000?l0000000002@0000000`3o0000oooo
0?ooo`3i0?ooo`00;@3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo00`0oooo00L0
00000?ooo`3o0000oooo0000003oool0000000@0oooo00<0o`000000003oool0nP3oool002d0oooo
0`00001Y0?ooo`030000003oool0oooo00`0oooo00L000000?ooo`3o0000oooo0000003oool00000
00<0oooo00<0o`000?ooo`000000n`3oool002d0oooo00<000000?ooo`3oool0J@3oool00`000000
oooo0?ooo`0?0?ooo`040?l000000000oooo000000<0oooo00<0o`000?ooo`3oool0n`3oool002d0
oooo00<000000?ooo`3oool0JP3oool00`000000oooo0?ooo`0>0?ooo`040000003oool0oooo0000
0080oooo00<0o`000?ooo`3oool0o03oool002d0oooo00<000000?ooo`3oool0JP3oool00`000000
oooo0?ooo`0>0?ooo`060000003o0000oooo0000003oool0o`00o`3oool002d0oooo00<000000?oo
o`3oool0JP3oool00`000000oooo0?ooo`0>0?ooo`050000003oool0o`000000003o0000o`3oool1
0?ooo`00;@3oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo00d0oooo00D000000?oo
o`3o000000000?l0003o0?ooo`40oooo000]0?ooo`030000003oool0oooo06/0oooo00<000000?oo
o`3oool0303oool01@000000oooo0?ooo`3oool0o`000?l0oooo0P3oool002d0oooo00<000000?oo
o`3oool0J`3oool00`000000oooo0?ooo`0<0?ooo`050000003oool0oooo0?ooo`000000o`3oool2
0?ooo`00;@3oool00`000000oooo0?ooo`1/0?ooo`030000003oool0oooo00/0oooo00D000000?oo
o`3oool0oooo0000003o0?ooo`80oooo000]0?ooo`030000003oool0oooo06`0oooo00<000000?oo
o`3oool02P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0?l0oooo000]0?ooo`<0
0000K03oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0o`3oool002d0oooo00<000000?ooo`3oool0K@3oool00`000000oooo0?ooo`090?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0o`3oool002d0oooo00<000000?ooo`3oool0K@3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0o`3o
ool002d0oooo00<000000?ooo`3oool0K@3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool0o`3oool002d0oooo00<000000?ooo`3oool0KP3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0o`3oool001l0
oooo00H0_g@000000000000000000018W03Oool80?ooo`030000003oool0oooo06h0oooo00<00000
0?ooo`3oool01`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0?l0oooo000P0?oo
o`030?oOW0180000M;oo00X0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool0o`3oool00200oooo00<0omnL04P0001d
_ol02P3oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`3o0?ooo`00803oool00`3ogi`0B00007Boo`0:0?ooo`030000
003oool0oooo0700oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`H00000
oP3oool00200oooo00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`1a0?ooo`050000003o
ool0oooo0?ooo`000000o`3oool;0?ooo`00803oool00`3ogi`0B00007Boo`0:0?ooo`@00000L03o
ool400000?l0oooo303oool00200oooo00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`3o
0?oooh40oooo000P0?ooo`030?oOW0180000M;oo00X0oooo00<000000?ooo`3oool0o`3ooon10?oo
o`00803oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo0?l0ooooP@3oool00200oooo
00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`3o0?oooh40oooo000N0?ooo`050?oog`2L
B00000000000001d_ol02P3oool00`000000oooo0?ooo`3o0?oooh40oooo000P0?ooo`030?noM000
0000M;oo00X0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`1`
0?ooo`H00000o`3oool;0?ooo`00;@3oool00`000000oooo0?ooo`1c0?ooo`030000003oool0oooo
0?l0oooo2`3oool002d0oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`3o0?ooo`/0
oooo000]0?ooo`<00000L`3oool00`000000oooo0?ooo`3o0?ooo`/0oooo000]0?ooo`030000003o
ool0oooo07<0oooo00<000000?ooo`3oool0o`3oool;0?ooo`00;@3oool00`000000oooo0?ooo`1c
0?ooo`030000003oool0oooo0?l0oooo2`3oool002d0oooo00<000000?ooo`3oool0L`3oool00`00
0000oooo0?ooo`3o0?ooo`/0oooo000]0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3o
ool0o`3oool;0?ooo`00;@3oool00`000000oooo0?ooo`1c0?ooo`030000003oool0oooo0?l0oooo
2`3oool002d0oooo00<000000?ooo`3oool0K@3oool<003o0?l0oooo203oool002d0oooo00<00000
0?ooo`3oool0I`3oool00`000000oooo0?ooo`040?ooo`X00?l01P3oool00`000000oooo0?ooo`3o
0?ooo`00;@3oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo00@0oooo2P00o`060?oo
o`030000003oool0oooo0?l0oooo000]0?ooo`030000003oool0oooo06L0oooo20000008003o00P0
0000o`3oool20?ooo`00;@3oool3000006L0oooo00<000000?ooo`3oool01@3oool8003o00L0oooo
00<000000?ooo`3oool0o`3oool002d0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?oo
o`060?ooo`H00?l0203oool00`000000oooo0?ooo`3o0?ooo`00;@3oool00`000000oooo0?ooo`1a
0?ooo`D00?l0o`3oool;0?ooo`00;@3oool00`000000oooo0?ooo`1a0?ooo`@00?l0o`3oool<0?oo
o`00;@3oool00`000000oooo0?ooo`1b0?ooo`<00?l0o`3oool<0?ooo`00;@3oool00`000000oooo
0?ooo`1b0?ooo`800?l0o`3oool=0?ooo`00;@3oool00`000000oooo0?ooo`1c0?ooo`03003o003o
ool0oooo0?l0oooo2`3oool002d0oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`3o
0?ooo`/0oooo000]0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool0o`3oool;0?oo
o`00;@3oool00`000000oooo0?ooo`1c0?ooo`030000003oool0oooo0?l0oooo2`3oool002d0oooo
0`00001c0?ooo`030000003oool0oooo0?l0oooo2`3oool002d0oooo00<000000?ooo`3oool0L03o
ool600000?l0oooo2`3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`00
0000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo
00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo0000
\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.392619, -0.141431, \
0.0128071, 0.00745692}}]
}, Closed]],

Cell["\<\
Unfortunately, there do not appear to be versions adapted for semi-log or \
log-log formats yet.\
\>", "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Specialized two-dimensional plots", "Section"],

Cell[TextData[{
  "Often one requires more specialized plotting formats.  The standard \
package ",
  StyleBox["Graphics`Graphics`", "InlineInput",
    FontWeight->"Bold"],
  " provides many of the more common plotting formats with syntax and options \
very similar to ",
  StyleBox["Plot", "InlineInput",
    FontWeight->"Bold"],
  ".  In addition, there are several variations of ",
  StyleBox["BarChart", "InlineInput",
    FontWeight->"Bold"],
  " and ",
  StyleBox["PieChart", "InlineInput",
    FontWeight->"Bold"],
  " and variations of ",
  StyleBox["ListPlot", "InlineInput",
    FontWeight->"Bold"],
  " which plot labelled points or use text as points.  The most useful of \
these functions are summarized in the table below and others can be found in \
the Help Browser. "
}], "Text"],

Cell[BoxData[
    StyleBox[GridBox[{
          {\(Selected\ functions\ in\ Graphics`Graphics`\)},
          {GridBox[{
                {
                  StyleBox["function",
                    "Text"], 
                  StyleBox[\(purpose/format\),
                    "Text"]},
                {
                  StyleBox[
                    RowBox[{"LogPlot", "[", 
                      RowBox[{
                        StyleBox["y",
                          "Text"], 
                        StyleBox[",",
                          "Text"], 
                        RowBox[{
                          StyleBox["{",
                            "Text"], 
                          RowBox[{
                            StyleBox["x",
                              "Text"], 
                            StyleBox[",",
                              "Text"], 
                            SubscriptBox[
                              StyleBox["x",
                                "Text"], "min"], ",", \(x\_max\)}], "}"}]}], 
                      "]"}],
                    "Text"], 
                  StyleBox[\(log\ y\ vs . \ linear\ x\ for\ function\),
                    "Text"]},
                {
                  StyleBox[
                    RowBox[{"LogLinearPlot", "[", 
                      RowBox[{
                        StyleBox["y",
                          "Text"], 
                        StyleBox[",",
                          "Text"], 
                        RowBox[{
                          StyleBox["{",
                            "Text"], 
                          RowBox[{
                            StyleBox["x",
                              "Text"], 
                            StyleBox[",",
                              "Text"], 
                            SubscriptBox[
                              StyleBox["x",
                                "Text"], "min"], ",", \(x\_max\)}], "}"}]}], 
                      "]"}],
                    "Text"], 
                  StyleBox[
                    RowBox[{"linear", " ", "y", " ", 
                      RowBox[{"vs", ".", " ", 
                        StyleBox["log",
                          "Text"]}], 
                      StyleBox[" ",
                        "Text"], 
                      StyleBox["x",
                        "Text"]}],
                    "Text"]},
                {
                  StyleBox[
                    RowBox[{
                      StyleBox["LogLogPlot",
                        "Text"], 
                      StyleBox["[",
                        "Text"], 
                      RowBox[{
                        StyleBox["y",
                          "Text"], 
                        StyleBox[",",
                          "Text"], 
                        RowBox[{
                          StyleBox["{",
                            "Text"], 
                          RowBox[{
                            StyleBox["x",
                              "Text"], 
                            StyleBox[",",
                              "Text"], 
                            SubscriptBox[
                              StyleBox["x",
                                "Text"], "min"], ",", \(x\_max\)}], "}"}]}], 
                      "]"}],
                    "Text"], 
                  StyleBox[\(log\ y\ vs . \ log\ x\),
                    "Text"]},
                {
                  StyleBox[\(LogListPlot[{{x, y},  ... }]\),
                    "Text"], 
                  
                  StyleBox[\(log\ y\ vs . \ 
                        linear\ x\ for\ list\ of\ points\),
                    "Text"]},
                {
                  StyleBox[\(LogLinearListPlot[{{x, y},  ... }]\),
                    "Text"], 
                  StyleBox[
                    RowBox[{"linear", " ", "y", " ", 
                      RowBox[{"vs", ".", " ", 
                        StyleBox["log",
                          "Text"]}], 
                      StyleBox[" ",
                        "Text"], 
                      StyleBox["x",
                        "Text"]}],
                    "Text"]},
                {
                  StyleBox[\(LogLogListPlot[{{x, y},  ... }]\),
                    "Text"], 
                  StyleBox[\(log\ y\ vs . \ log\ x\),
                    "Text"]},
                {
                  RowBox[{
                    StyleBox["PolarPlot",
                      "Text"], 
                    StyleBox["[",
                      "Text"], 
                    RowBox[{
                      StyleBox["r",
                        "Text"], 
                      StyleBox[",",
                        FontSlant->"Italic"], 
                      RowBox[{
                        StyleBox["{",
                          FontSlant->"Italic"], 
                        RowBox[{"\[Theta]", 
                          StyleBox[",",
                            FontSlant->"Italic"], 
                          StyleBox[\(\[Theta]\_min\),
                            "Text"], 
                          StyleBox[",",
                            "Text",
                            FontSlant->"Italic"], 
                          StyleBox[\(\[Theta]\_max\),
                            "Text"]}], 
                        StyleBox["}",
                          "Text"]}]}], 
                    StyleBox["]",
                      "Text"]}], 
                  RowBox[{
                    StyleBox["polar",
                      "Text"], 
                    StyleBox[" ",
                      "Text"], 
                    StyleBox["plot",
                      "Text"], 
                    StyleBox[" ",
                      "Text"], 
                    StyleBox["of",
                      "Text"], 
                    StyleBox[" ",
                      "Text",
                      FontSlant->"Italic"], 
                    RowBox[{
                      StyleBox["r",
                        "Text"], 
                      StyleBox["[",
                        FontSlant->"Italic"], 
                      StyleBox["\[Theta]",
                        "Text"], 
                      StyleBox["]",
                        FontSlant->"Italic"]}]}]},
                {
                  StyleBox[\(ErrorListPlot[{{x, y, \[Delta]y},  ... }]\),
                    "Text"], 
                  StyleBox[
                    RowBox[{"points", " ", 
                      RowBox[{
                        StyleBox["(",
                          "Text"], 
                        RowBox[{
                          SubscriptBox[
                            StyleBox["x",
                              "Text"], "i"], ",", 
                          RowBox[{
                            SubscriptBox[
                              StyleBox["y",
                                "Text"], "i"], 
                            StyleBox["\[PlusMinus]",
                              "Text"], 
                            SubscriptBox[
                              StyleBox["\[Delta]y",
                                "Text"], "i"]}]}], ")"}], " ", "with", " ", 
                      "error", " ", "bars"}],
                    "Text"]},
                {
                  StyleBox[\(BarChart[list]\),
                    "Text"], 
                  StyleBox[\(bar\ charts\ or\ histograms\),
                    "Text"]},
                {
                  StyleBox[\(PieChart[list]\),
                    "Text"], 
                  StyleBox[\(pie\ charts\),
                    "Text"]},
                {
                  StyleBox[
                    RowBox[{"PlotVectorField", "[", 
                      RowBox[{
                        RowBox[{
                          StyleBox["{",
                            "Text"], 
                          RowBox[{
                            SubscriptBox[
                              StyleBox["f",
                                "Text"], "x"], ",", \(f\_y\)}], "}"}], 
                        ",", \({x, x\_min, x\_max}\), 
                        ",", \({y, y\_min, y\_max}\)}], "]"}],
                    "Text"], 
                  StyleBox[\(2 - d\ vector\ function\),
                    "Text"]},
                {
                  StyleBox[\(ListPlotVectorField[{{x, y, f[x, y]},  ... }]\),
                    "Text"], 
                  StyleBox[\(2 - d\ vector\ list\),
                    "Text"]},
                {
                  
                  StyleBox[\(SphericalPlot3D[
                      r, {\[Theta], \[Theta]\_min, \[Theta]\_max}, {\[Phi], \
\[Phi]\_min, \[Phi]\_max}]\),
                    "Text"], 
                  StyleBox[\(spherical\ plot\ of\ r[\[Theta], \[Phi]]\),
                    "Text"]}
                },
              ColumnAlignments->{Left},
              GridFrame->True]}
          }],
      Background->RGBColor[0.500008, 0.996109, 0.996109]]], "Text",
  TextAlignment->Center,
  TextJustification->0,
  FontFamily->"Times New Roman",
  FontWeight->"Bold"],

Cell["\<\
If your data or function covers a very wide dynamic range, as might be \
expected from an exponential function, a semilog plot will probably be more \
useful than a linear plot, as simulated below.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[
        LogPlot[\(x\^2\) Exp[\(-x\^2\)], {x, 0, 5}, 
          PlotRange \[Rule] {{0.001, 5}, Automatic}, 
          GridLines \[Rule] Automatic], 
        LogListPlot[
          Table[{x, \(x\^2\) Exp[\(-x\^2\)]\ Random[Real, {0.9, 1.1}]}, {x, 
              0.1, 4, 0.05}], 
          PlotStyle \[Rule] {AbsolutePointSize[4]}]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.00020004 0.20004 0.603319 0.0732227 [
[.19984 -0.0125 -5 -15.6875 ]
[.19984 -0.0125 5 0 ]
[.39988 -0.0125 -5 -15.6875 ]
[.39988 -0.0125 5 0 ]
[.59992 -0.0125 -5 -15.6875 ]
[.59992 -0.0125 5 0 ]
[.79996 -0.0125 -5 -15.6875 ]
[.79996 -0.0125 5 0 ]
[1 -0.0125 -5 -15.6875 ]
[1 -0.0125 5 0 ]
[-0.0125 .01754 -46.625 -7.84375 ]
[-0.0125 .01754 0 7.84375 ]
[-0.0125 .16398 -46.625 -7.84375 ]
[-0.0125 .16398 0 7.84375 ]
[-0.0125 .31043 -37 -7.84375 ]
[-0.0125 .31043 0 7.84375 ]
[-0.0125 .45687 -25 -7.84375 ]
[-0.0125 .45687 0 7.84375 ]
[-0.0125 .60332 -10 -7.84375 ]
[-0.0125 .60332 0 7.84375 ]
[ -0.0005 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.19984 0 m
.19984 .61803 L
s
.39988 0 m
.39988 .61803 L
s
.59992 0 m
.59992 .61803 L
s
.79996 0 m
.79996 .61803 L
s
1 0 m
1 .61803 L
s
0 .01754 m
1 .01754 L
s
0 .16398 m
1 .16398 L
s
0 .31043 m
1 .31043 L
s
0 .45687 m
1 .45687 L
s
0 .60332 m
1 .60332 L
s
0 .09656 m
1 .09656 L
s
0 .11725 m
1 .11725 L
s
0 .12968 m
1 .12968 L
s
0 .13859 m
1 .13859 L
s
0 .14554 m
1 .14554 L
s
0 .15125 m
1 .15125 L
s
0 .15608 m
1 .15608 L
s
0 .16028 m
1 .16028 L
s
0 .243 m
1 .243 L
s
0 .26369 m
1 .26369 L
s
0 .27612 m
1 .27612 L
s
0 .28504 m
1 .28504 L
s
0 .29199 m
1 .29199 L
s
0 .29769 m
1 .29769 L
s
0 .30253 m
1 .30253 L
s
0 .30672 m
1 .30672 L
s
0 .38945 m
1 .38945 L
s
0 .41014 m
1 .41014 L
s
0 .42257 m
1 .42257 L
s
0 .43148 m
1 .43148 L
s
0 .43844 m
1 .43844 L
s
0 .44414 m
1 .44414 L
s
0 .44897 m
1 .44897 L
s
0 .45317 m
1 .45317 L
s
0 .53589 m
1 .53589 L
s
0 .55658 m
1 .55658 L
s
0 .56901 m
1 .56901 L
s
0 .57793 m
1 .57793 L
s
0 .58488 m
1 .58488 L
s
0 .59058 m
1 .59058 L
s
0 .59542 m
1 .59542 L
s
0 .59961 m
1 .59961 L
s
0 g
.19984 0 m
.19984 .00625 L
s
gsave
.19984 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.39988 0 m
.39988 .00625 L
s
gsave
.39988 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.59992 0 m
.59992 .00625 L
s
gsave
.59992 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.79996 0 m
.79996 .00625 L
s
gsave
.79996 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.125 Mabswid
.03981 0 m
.03981 .00375 L
s
.07982 0 m
.07982 .00375 L
s
.11982 0 m
.11982 .00375 L
s
.15983 0 m
.15983 .00375 L
s
.23985 0 m
.23985 .00375 L
s
.27986 0 m
.27986 .00375 L
s
.31986 0 m
.31986 .00375 L
s
.35987 0 m
.35987 .00375 L
s
.43989 0 m
.43989 .00375 L
s
.4799 0 m
.4799 .00375 L
s
.5199 0 m
.5199 .00375 L
s
.55991 0 m
.55991 .00375 L
s
.63993 0 m
.63993 .00375 L
s
.67994 0 m
.67994 .00375 L
s
.71994 0 m
.71994 .00375 L
s
.75995 0 m
.75995 .00375 L
s
.83997 0 m
.83997 .00375 L
s
.87998 0 m
.87998 .00375 L
s
.91998 0 m
.91998 .00375 L
s
.95999 0 m
.95999 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .01754 m
.00625 .01754 L
s
gsave
-0.0125 .01754 -107.625 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.) show
72.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
73.813 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\264) show
83.313 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
95.313 10.938 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
95.313 10.938 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
100.813 10.938 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(8) show
105.125 10.938 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .16398 m
.00625 .16398 L
s
gsave
-0.0125 .16398 -107.625 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.) show
72.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
73.813 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(\\264) show
83.313 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
95.313 10.938 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
95.313 10.938 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
100.813 10.938 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
105.125 10.938 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .31043 m
.00625 .31043 L
s
gsave
-0.0125 .31043 -98 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.0001) show
96.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .45687 m
.00625 .45687 L
s
gsave
-0.0125 .45687 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.01) show
84.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .60332 m
.00625 .60332 L
s
gsave
-0.0125 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
69.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.001 w
0 .09656 m
.00375 .09656 L
s
0 .11725 m
.00375 .11725 L
s
0 .12968 m
.00375 .12968 L
s
0 .13859 m
.00375 .13859 L
s
0 .14554 m
.00375 .14554 L
s
0 .15125 m
.00375 .15125 L
s
0 .15608 m
.00375 .15608 L
s
0 .16028 m
.00375 .16028 L
s
0 .243 m
.00375 .243 L
s
0 .26369 m
.00375 .26369 L
s
0 .27612 m
.00375 .27612 L
s
0 .28504 m
.00375 .28504 L
s
0 .29199 m
.00375 .29199 L
s
0 .29769 m
.00375 .29769 L
s
0 .30253 m
.00375 .30253 L
s
0 .30672 m
.00375 .30672 L
s
0 .38945 m
.00375 .38945 L
s
0 .41014 m
.00375 .41014 L
s
0 .42257 m
.00375 .42257 L
s
0 .43148 m
.00375 .43148 L
s
0 .43844 m
.00375 .43844 L
s
0 .44414 m
.00375 .44414 L
s
0 .44897 m
.00375 .44897 L
s
0 .45317 m
.00375 .45317 L
s
0 .53589 m
.00375 .53589 L
s
0 .55658 m
.00375 .55658 L
s
0 .56901 m
.00375 .56901 L
s
0 .57793 m
.00375 .57793 L
s
0 .58488 m
.00375 .58488 L
s
0 .59058 m
.00375 .59058 L
s
0 .59542 m
.00375 .59542 L
s
0 .59961 m
.00375 .59961 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.00051 0 m
.00103 .27974 L
.00216 .32088 L
.00345 .34875 L
.00468 .36713 L
.00686 .3906 L
.00923 .40895 L
.01183 .42443 L
.0146 .43752 L
.01955 .45576 L
.02467 .47023 L
.02946 .48123 L
.04038 .50054 L
.05151 .51515 L
.06169 .52566 L
.07359 .53556 L
.08463 .54304 L
.10456 .55346 L
.1151 .55771 L
.12633 .56146 L
.13713 .56441 L
.14881 .56694 L
.15985 .56878 L
.17 .57002 L
.17493 .57049 L
.18022 .57089 L
.18522 .57117 L
.18742 .57127 L
.18977 .57135 L
.19109 .5714 L
.19254 .57143 L
.19388 .57146 L
.19511 .57148 L
.19655 .5715 L
.1973 .57151 L
.1981 .57151 L
.19944 .57152 L
.2002 .57152 L
.20091 .57152 L
.20229 .57151 L
.20359 .5715 L
.20475 .57148 L
.20602 .57146 L
.20885 .57139 L
.21147 .57131 L
.21638 .5711 L
.22165 .57079 L
.23117 .57003 L
.2425 .5688 L
.25285 .56738 L
Mistroke
.27283 .56386 L
.29461 .55892 L
.33735 .54605 L
.3785 .52994 L
.42222 .50906 L
.46436 .4854 L
.5049 .45949 L
.54801 .4286 L
.58954 .3957 L
.62947 .36117 L
.67198 .32134 L
.7129 .28006 L
.75639 .23303 L
.79829 .18467 L
.8386 .13536 L
.88148 .07989 L
.92278 .02359 L
Mfstroke
.92278 .02359 m
.93926 0 L
s
4 Mabswid
.0198 .45346 Mdot
.02981 .48426 Mdot
.03981 .50243 Mdot
.04981 .51133 Mdot
.05981 .52506 Mdot
.06981 .53073 Mdot
.07982 .54183 Mdot
.08982 .54499 Mdot
.09982 .54988 Mdot
.10982 .55326 Mdot
.11982 .55722 Mdot
.12983 .56029 Mdot
.13983 .56668 Mdot
.14983 .56929 Mdot
.15983 .56934 Mdot
.16983 .5722 Mdot
.17984 .56823 Mdot
.18984 .56821 Mdot
.19984 .57429 Mdot
.20984 .57324 Mdot
.21984 .56997 Mdot
.22985 .57245 Mdot
.23985 .56776 Mdot
.24985 .56677 Mdot
.25985 .56509 Mdot
.26985 .56746 Mdot
.27986 .56135 Mdot
.28986 .55756 Mdot
.29986 .55838 Mdot
.30986 .55659 Mdot
.31986 .55203 Mdot
.32987 .55037 Mdot
.33987 .54418 Mdot
.34987 .5425 Mdot
.35987 .5368 Mdot
.36987 .53438 Mdot
.37988 .52987 Mdot
.38988 .52676 Mdot
.39988 .52191 Mdot
.40988 .51698 Mdot
.41988 .51015 Mdot
.42989 .50673 Mdot
.43989 .50152 Mdot
.44989 .49671 Mdot
.45989 .4856 Mdot
.46989 .48451 Mdot
.4799 .47602 Mdot
.4899 .47021 Mdot
.4999 .46474 Mdot
.5099 .45856 Mdot
.5199 .44708 Mdot
.52991 .44202 Mdot
.53991 .43254 Mdot
.54991 .42463 Mdot
.55991 .42045 Mdot
.56991 .41299 Mdot
.57992 .40562 Mdot
.58992 .39512 Mdot
.59992 .38561 Mdot
.60992 .37574 Mdot
.61992 .36801 Mdot
.62993 .36172 Mdot
.63993 .35174 Mdot
.64993 .34129 Mdot
.65993 .33459 Mdot
.66993 .32563 Mdot
.67994 .31475 Mdot
.68994 .30273 Mdot
.69994 .29431 Mdot
.70994 .286 Mdot
.71994 .27044 Mdot
.72995 .26352 Mdot
.73995 .25322 Mdot
.74995 .2373 Mdot
.75995 .23191 Mdot
.76995 .21599 Mdot
.77996 .20717 Mdot
.78996 .19725 Mdot
.79996 .18152 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00S@3oool01P2oM00000000000
0000000004RL0=oooch0oooo00<0_g@0000000000000100000000`18W=l0oooo0?ooo`0l0?ooo`03
0?oo_`1d000000000080000000<0M;oo0?ooo`3oool0@P3oool0103ooml0W4P00018W03Oooln0?oo
o`030?oog`2LB00000000080000000<0M;oo0?ooo`3oool01@3oool008h0oooo00<0omnL04P0001d
_ol0@@3oool00`3OW4P0000000000003000000030018W03Oool0oooo03`0oooo00<0omnL04P00018
W=l00P3oool00`2oM000B9cO0?ooo`120?ooo`040?oog`2LB00004RL0=oooch0oooo00<0oooO09a8
M02oool00P3oool00`2oM000M;oo0?ooo`050?ooo`00SP3oool00`3ogi`0B00007Boo`120?ooo`03
0=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0@@3oool00`3ookl0M01809cOo`120?ooo`04
0?oog`2LB00004RL0=oood<0oooo00<0oooO09a8M02oool01@3oool008h0oooo00<0omnL04P0001d
_ol0@`3oool00`3OW4P0M;oo0?ooo`150?ooo`030?oo_`1d0000M;oo03h0oooo00<0_g@000000000
00001@0000000`1d_ol0oooo0?ooo`110?ooo`030;md001d_ol0oooo00@0oooo002>0?ooo`030?oO
W0180000M;oo04@0oooo00<0_gA809cOo`3oool0A03oool00`3o_g@0000007Boo`0n0?ooo`080?oO
W01dW=l0oooo0?ooo`3ooml0W4P00018W03Ooom30?ooo`030?oog`2LB000B9cO00D0oooo002>0?oo
o`030?oOW0180000M;oo04@0oooo00<0oooO09a8M02oool0@`3oool0103ogi`0B0000000B02Lgolo
0?ooo`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`130?ooo`030=nLB0000000M;oo00D0oooo
002>0?ooo`030?oOW0180000M;oo04D0oooo00<0okmd001d_`3oool0@03oool01@3o_g@000000000
000004P0W=oo0400oooo00L0oooO09a8M02oool0oooO09a80000B9`0gooo03l0oooo00L0gmoO0?oo
o`3ookl0M0000000000007@0_ooo00D0oooo002>0?ooo`030?oOW0180000M;oo04D0oooo00<0oono
07@0B02Lgol0@P3oool00`2oM00007Bo0?ooo`110?ooo`060?oOW018M;l0oooO09a80000B9`0gooo
?`3oool01P3OW4P00000000000000000001809cOo`H0oooo002>0?ooo`030?oOW0180000M;oo04D0
oooo00<0oono07@00018W=l0@`3oool00`2oM000B9cO0?ooo`110?ooo`050=nLB01d_ml0W4P00018
W03Oool0?`3oool0103ookl0M00000000018W=l80?ooo`00SP3oool00`3ogi`0B00007Boo`100?oo
o`030?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l0?P3oool00`3ooml0W9bo0?ooo`020?oo
o`030=nLB00007@0_ooo0440oooo00D0oooO09a8B018B00004RL0=ooo`100?ooo`030;md001d_ol0
oooo00P0oooo002<0?ooo`050?oog`2LB00000000000001d_ol0@03oool0203ooml0W4Q809cOo`3o
ool0oono07@0000007@0_ooo?`3oool01P3o_g@0B9cO0?ooo`3ookl0M000001d_d<0oooo00@0omnL
04P00000B9`0gooo@03oool01P3ogi`0B000000000000000001809cOo`D0oooo002>0?ooo`030?no
M0000000M;oo0440oooo00H0oooO09a80000000000000000M02ooom10?ooo`050=nLB00000000000
0000M02oool0A03oool00`3OW4P004RL0=ooo`110?ooo`050;md00000000000000000018W=l01@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool004X0ooooo`00
001O000000P0oooo00050?ooo`060;md00000000000000000000B9`0gooo0P3oool00`3o_g@00000
07Boo`030?ooo`030?oo_`1d0000B9cO00D0oooo00<0oono07@00018W=l01@3oool01P2oM0000000
00000000000004RL0=ooo`<0oooo00D0omnL04P000000000001d0;ooo`0I0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`060?ooo`001P3oool00`3o
gi`0B00007Boo`040?ooo`030?noM0000000M;oo00@0oooo00<0oono07@00018W=l00`3oool00`3o
okl0M00004RLg`070?ooo`030?oOW0180000M;oo00@0oooo00L0oono07@0001d_ol0oooo0?noM000
07@0_ooo01P0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00H0oooo00060?ooo`030?oOW0180000M;oo00`0oooo00L0oono07@00018W=l0oooo0?oo
_`1d0000B9cO00P0oooo00<0omnL04P0001d_ol0103oool00`2oM00004RL0=ooo`020?ooo`030;md
0000M;l0oooo01L0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000
O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0/0?ooo`030000
003oool0oooo01@0oooo00<0001o0?ooo`3oool01P3oool000H0oooo00<0omnL04P0001d_ol03@3o
ool01@3ookl0M00004RLW01d0000B9cO00T0oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M000
04RLg`030?ooo`030=nLB00007@0_ooo01L0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo
0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo
0?ooo`0[0?ooo`030000003oool0oooo01D0oooo00<0001o0?ooo`3oool01P3oool000H0oooo00<0
omnL04P0001d_ol03P3oool00`3ookl0M00004RLg`0:0?ooo`030?oOW0180000M;oo00<0oooo00<0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0G0?ooo`030000003oool0oooo04<0oooo00<0
001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0
001o0?ooo`3oool0:`3oool00`000000oooo0?ooo`0E0?ooo`030000O`3oool0oooo00H0oooo0006
0?ooo`030?oOW0180000M;oo00d0oooo00D0oono07@00018W9`0M00004RLg`090?ooo`030?oOW018
0000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0G0?ooo`<00000o`00
07m20000O`030000000007l0001o01L0001o203oool000H0oooo00<0omnL04P0001d_ol0303oool0
1`3ookl0M00004RLg`3oool0oono07@00018W=l0203oool00`3ogi`0B00007Boo`030?ooo`030?oO
W0180000M;oo00<0oooo00<0okmd0000B02Lgol05`3oool00`000000oooo0?ooo`130?ooo`030000
O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000
O`3oool0oooo02T0oooo00<000000?ooo`3oool05`3oool00`0007l0oooo0?ooo`060?ooo`001P3o
ool00`3ogi`0B00007Boo`0;0?ooo`030?oog`2LB000B9cO00<0oooo00<0oono07@00018W=l01`3o
ool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol05`3o
ool00`000000oooo0?ooo`130?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3o
ool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo02T0oooo00<000000?ooo`3oool05`3o
ool00`0007l0oooo0?ooo`060?ooo`001P3oool00`3ogi`0B00007Boo`0:0?ooo`030?oog`2LB000
B9cO00D0oooo00<0oono07@00018W=l01P3oool00`3ogi`0B00007Boo`030?ooo`030?oo_`1d0000
B9cO00<0oooo00<0gia80000M02oool02@3oool0103OW4P0000000000018W=l:0?ooo`030000003o
ool0oooo04<0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3o
ool0oooo04<0oooo00<0001o0?ooo`3oool0:03oool00`000000oooo0?ooo`0H0?ooo`030000O`3o
ool0oooo00H0oooo00060?ooo`030?oOW0180000M;oo01/0oooo00<0omnL04P0001d_ol0103oool0
0`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00P0oooo00H0omnL04P0M02oool0oono07@0
001d_ol90?ooo`030000003oool0oooo04<0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo
0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo
0?ooo`0I0?ooo`030000O`3oool0oooo00H0oooo00040?ooo`050?oog`2LB00000000000001d_ol0
6@3oool01@3ooml0W4P0000000000000M;oo00@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0
_ooo00T0oooo00H0omnL04Q8W03Oool0oooO09a8001d_ol90?ooo`030000003oool0oooo04<0oooo
00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0oooo
00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0I0?ooo`030000O`3oool0oooo00H0oooo
00060?ooo`030?noM0000000M;oo01/0oooo00<0okmd0000001d_ol01@3oool01@3ookl0M0000000
000007@0_ooo0080oooo00<0omnL04P000000000100000001`18W=l0oooo0=nLM02Lgml0W4P00018
W03Oool02@3oool00`000000oooo0?ooo`130?ooo`030000O`3oool0oooo0480oooo00<0001o0?oo
o`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo02H0oooo00<000000?oo
o`3oool06P3oool00`0007l0oooo0?ooo`060?ooo`00?03oool00`3ookl0M000001d_`0;0?ooo`03
0000003oool0oooo04<0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`03
0000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0K0?ooo`03
0000O`3oool0oooo00H0oooo000k0?ooo`050?oog`2LB00007Bo0?noM018W=l02P3oool00`000000
oooo0?ooo`130?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0
oooo0?ooo`130?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0703oool00`0007l0
oooo0?ooo`060?ooo`00>`3oool01P3ogi`0B01d0;ooo`3ookl0M01809cOo`T0oooo00<000000?oo
o`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?oo
o`3oool0@`3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo01`0oooo00<0001o0?oo
o`3oool01P3oool003/0oooo00H0oono07@0B02Lgol0omnL04P0B02Lgol90?ooo`030000003oool0
oooo04<0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0
oooo04<0oooo00<0001o0?ooo`3oool08`3oool00`000000oooo0?ooo`0M0?ooo`030000O`3oool0
oooo00H0oooo000l0?ooo`040=nLB0000000000004RLg`X0oooo00<000000?ooo`3oool0@`3oool0
0`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool0
0`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo01h0oooo00<0001o0?ooo`3oool01P3oool0
04X0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo
04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo
01l0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3oool0@`3oool00`0007l0
oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0
oooo0?ooo`0Q0?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool01P3oool004X0oooo
00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo
00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0P0?ooo`030000003oool0oooo0200oooo
00<0001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?oo
o`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?oo
o`0O0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<00000
0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o
0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo0280oooo00<0001o
0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?oo
o`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0N0?oo
o`030000003oool0oooo0280oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3o
ool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3o
ool0@`3oool00`0007l0oooo0?ooo`0M0?ooo`030000003oool0oooo02<0oooo00<0001o0?ooo`3o
ool01P3oool004X0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000
O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0L0?ooo`030000
003oool0oooo02@0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3oool0@`3o
ool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3o
ool00`0007l0oooo0?ooo`0K0?ooo`030000003oool0oooo02D0oooo00<0001o0?ooo`3oool01P3o
ool004X0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0
oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0K0?ooo`030000003oool0
oooo02D0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3oool0@`3oool00`00
07l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`00
07l0oooo0?ooo`0J0?ooo`030000003oool0oooo02H0oooo00<0001o0?ooo`3oool01P3oool004X0
oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0
oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0I0?ooo`030000003oool0oooo02L0
oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo
0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo
0?ooo`0H0?ooo`030000003oool0oooo02P0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<0
00000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0
001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo02P0oooo00<0
001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`12
0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0G
0?ooo`030000003oool0oooo02T0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<000000?oo
o`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?oo
o`3oool0@`3oool00`0007l0oooo0?ooo`0F0?ooo`030000003oool0oooo02X0oooo00<0001o0?oo
o`3oool01P3oool004X0oooo0P00003o0000Obh0001o00<000000000O`0007l0;00007l80?ooo`00
BP3oool00`000000oooo0?ooo`130?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0
@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo01D0oooo00<000000?ooo`3oool0
:`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool00`000000oooo0?ooo`130?ooo`030000O`3o
ool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3o
ool0oooo01@0oooo00<000000?ooo`3oool0;03oool00`0007l0oooo0?ooo`060?ooo`00BP3oool0
0`000000oooo0?ooo`130?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool0
0`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo01<0oooo00<000000?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`060?ooo`00BP3oool00`000000oooo0?ooo`130?ooo`030000O`3oool0oooo
0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo
0180oooo00<000000?ooo`3oool0;P3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool00`000000
oooo0?ooo`130?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0
oooo0?ooo`130?ooo`030000O`3oool0oooo0180oooo00<000000?ooo`3oool0;P3oool00`0007l0
oooo0?ooo`060?ooo`00BP3oool00`000000oooo0?ooo`130?ooo`030000O`3oool0oooo0480oooo
00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo0140oooo
00<000000?ooo`3oool0;`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool200000?l0001o:@00
07l00`000000001o0000O`0a0000O`P0oooo001:0?ooo`030000003oool0oooo04<0oooo00<0001o
0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0001o
0?ooo`3oool03`3oool00`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo00H0oooo001:0?oo
o`030000003oool0oooo04<0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?oo
o`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool03`3oool00`000000oooo0?ooo`0a0?oo
o`030000O`3oool0oooo00H0oooo001:0?ooo`030000003oool0oooo04<0oooo00<0001o0?ooo`3o
ool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3o
ool03P3oool00`000000oooo0?ooo`0b0?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`030000
003oool0oooo04<0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000
O`3oool0oooo04<0oooo00<0001o0?ooo`3oool03@3oool00`000000oooo0?ooo`0c0?ooo`030000
O`3oool0oooo00H0oooo001:0?ooo`800000o`0007lU0000O`030000000007l0001o03D0001o203o
ool004X0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0
oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0
oooo03@0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3oool0@`3oool00`00
07l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`00
07l0oooo0?ooo`0;0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool01P3oool004X0
oooo0P00003o0000Ob<0001o00<000000000O`0007l0=`0007l80?ooo`00BP3oool00`000000oooo
0?ooo`130?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo
0?ooo`130?ooo`030000O`3oool0oooo00T0oooo00<000000?ooo`3oool0=`3oool00`0007l0oooo
0?ooo`060?ooo`00BP3oool200000?l0001o8P0007l00`000000001o0000O`0h0000O`P0oooo001:
0?ooo`030000003oool0oooo04<0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`13
0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0203oool00`000000oooo0?ooo`0h
0?ooo`030000O`3oool0oooo00H0oooo00050?ooo`060;md00000000000000000000B9`0gooo0P3o
ool00`3o_g@0000007Boo`030?ooo`030?oo_`1d0000B9cO00D0oooo00<0oono07@00018W=l01@3o
ool01P2oM000000000000000000004RL0=ooo`<0oooo00D0omnL04P000000000001d0;ooo`0I0?oo
o`800000o`0007lP0000O`030000000007l0001o03X0001o203oool000H0oooo00<0omnL04P0001d
_ol0103oool00`3o_g@0000007Boo`040?ooo`030?oo_`1d0000B9cO00<0oooo00<0oono07@00018
W=l01`3oool00`3ogi`0B00007Boo`040?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0H
0?ooo`030000003oool0oooo04<0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`13
0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool00`000000oooo0?ooo`0j
0?ooo`030000O`3oool0oooo00H0oooo00060?ooo`030?oOW0180000M;oo00`0oooo00L0oono07@0
0018W=l0oooo0?oo_`1d0000B9cO00P0oooo00<0omnL04P0001d_ol0103oool00`2oM00004RL0=oo
o`020?ooo`030;md0000M;l0oooo01L0oooo0P00003o0000Oah0001o00<000000000O`0007l0?000
07l80?ooo`001P3oool00`3ogi`0B00007Boo`0=0?ooo`050?oo_`1d0000B9bL07@00018W=l02@3o
ool00`3ogi`0B00007Boo`030?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool05`3o
ool200000?l0001o7@0007l00`000000001o0000O`0m0000O`P0oooo00060?ooo`030?oOW0180000
M;oo00h0oooo00<0oono07@00018W=l02P3oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol05`3oool00`000000oooo0?ooo`130?ooo`030000O`3oool0
oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0
oooo00@0oooo00<000000?ooo`3oool0?03oool00`0007l0oooo0?ooo`060?ooo`001P3oool00`3o
gi`0B00007Boo`0=0?ooo`050?oo_`1d0000B9bL07@00018W=l02@3oool00`3ogi`0B00007Boo`03
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol05`3oool300000?l0001o6`0007l0
0`000000001o0000O`0n0000O`P0oooo00060?ooo`030?oOW0180000M;oo00`0oooo00L0oono07@0
0018W=l0oooo0?oo_`1d0000B9cO00P0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo01L0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?oo
o`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?oo
o`020?ooo`030000003oool0oooo03h0oooo00<0001o0?ooo`3oool01P3oool000H0oooo00<0omnL
04P0001d_ol02`3oool00`3ooml0W4P004RLg`030?ooo`030?oo_`1d0000B9cO00L0oooo00<0omnL
04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00@0okmd
000000000000M;oo2P3oool00`000000oooo0?ooo`130?ooo`030000O`3oool0oooo0480oooo00<0
001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`050000O`3oool0oooo0?ooo`000000
@@3oool00`0007l0oooo0?ooo`060?ooo`001P3oool00`3ogi`0B00007Boo`0:0?ooo`030?oog`2L
B000B9cO00D0oooo00<0oono07@00018W=l01P3oool00`3ogi`0B00007Boo`030?ooo`030?oo_`1d
0000B9cO00<0oooo00<0gia80000M02oool0203oool01P3ogi`0B01809cOo`3ookl0M01809cOo`T0
oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0
oooo00<0001o0?ooo`3oool0@`3oool2000000030?ooo`000000oooo0440oooo00<0001o0?ooo`3o
ool01P3oool000H0oooo00<0omnL04P0001d_ol06`3oool00`3ogi`0B00007Boo`040?ooo`030=nL
B000B9`0gooo0080oooo00<0_g@0001d_`3oool0203oool01P3OW4P007Bo0?ooo`3oool0_g@004RL
g`T0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo
04<0oooo00<0001o0?ooo`3oool0@P3oool4000004<0oooo00<0001o0?ooo`3oool01P3oool000@0
oooo00D0oooO09a800000000000007Boo`0I0?ooo`050?oog`2LB00000000000001d_ol0103oool0
2P3ookl0M00007Boo`3oool0okmd0000M02oool0oooo0?oOW01800050000000304RLg`3OW4P0B9cO
0080oooo00<0_g@004RLg`3oool0203oool00`000000oooo0?ooo`130?ooo`030000O`3oool0oooo
0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`110?ooo`H00000@P3oool00`00
07l0oooo0?ooo`060?ooo`001P3oool00`3o_g@0000007Boo`0K0?ooo`030?noM0000000M;oo00D0
oooo00D0oono07@000000000001d0;ooo`0:0?ooo`060=nLB000M7@0_koo0?oOW0180000M;oo2@3o
ool00`000000oooo0?ooo`130?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3o
ool00`0007l0oooo0?ooo`110?ooo`H00000@P3oool00`0007l0oooo0?ooo`060?ooo`00>`3oool0
1@3ogi`0B01d07Ad0000000007Bo00X0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?oo
o`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@P3oool4000004<0oooo00<0
001o0?ooo`3oool01P3oool003/0oooo00<0oooO09a80018W=l0303oool00`000000oooo0?ooo`13
0?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`12
0?ooo`<00000A03oool00`0007l0oooo0?ooo`060?ooo`00?03oool00`3ookl0M000001d_`0;0?oo
o`030000003oool0oooo04<0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?oo
o`030000O`3oool0oooo03l0oooo0`0000000`3oool0001o0?ooo`140?ooo`030000O`3oool0oooo
00H0oooo000m0?ooo`040?oog`2LB000000007Boo`T0oooo00<000000?ooo`3oool0@`3oool00`00
07l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0?P3oool40000
00030?ooo`0007l0oooo04@0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3o
ool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3o
ool0?@3oool6000000030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo
00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo
00<0001o0?ooo`3oool0?03oool7000000030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0
1P3oool004X0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3o
ool0oooo04<0oooo00<0001o0?ooo`3oool0>`3oool7000000030?ooo`0007l0oooo04@0oooo00<0
001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`12
0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0>P3oool700000080oooo00<0001o
0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool00`000000oooo0?ooo`130?oo
o`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0h0?oo
o`P000000`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`03
0000003oool0oooo04<0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`03
0000O`3oool0oooo03L0oooo200000040?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3o
ool01P3oool004X0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000
O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0=P3oool8000000D0oooo00<0001o0?ooo`3oool0
@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool00`000000oooo0?ooo`130?ooo`030000O`3o
ool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0f0?ooo`H000001`3o
ool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`030000003oool0
oooo04<0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0
oooo03L0oooo100000080?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool0
04X0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo
04<0oooo00<0001o0?ooo`3oool0>03oool2000000T0oooo00<0001o0?ooo`3oool0@`3oool00`00
07l0oooo0?ooo`060?ooo`00BP3oool00`000000oooo0?ooo`130?ooo`030000O`3oool0oooo0480
oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0e0?ooo`<000002`3oool00`0007l0
oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`030000003oool0oooo04<0oooo
00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo03@0oooo
1000000;0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<0
00000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0
001o0?ooo`3oool0<@3oool8000000X0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?oo
o`060?ooo`00BP3oool00`000000oooo0?ooo`130?ooo`030000O`3oool0oooo0480oooo00<0001o
0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0`0?ooo`T000002P3oool00`0007l0oooo0?ooo`13
0?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`030000003oool0oooo04<0oooo00<0001o0?oo
o`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo02l0oooo2@00000;0?oo
o`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3o
ool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3o
ool0;`3oool8000000`0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00
BP3oool200000?l0001o0`0007l4000005H0001o203oool004X0oooo00<000000?ooo`3oool0@`3o
ool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0;P3o
ool500000100oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool0
0`000000oooo0?ooo`130?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool0
0`0007l0oooo0?ooo`0]0?ooo`@000004P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0
oooo00H0oooo001:0?ooo`030000003oool0oooo04<0oooo00<0001o0?ooo`3oool0@P3oool00`00
07l0oooo0?ooo`130?ooo`030000O`3oool0oooo02`0oooo1P00000A0?ooo`030000O`3oool0oooo
04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3oool0@`3oool00`0007l0
oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0;03oool600000140
oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool00`000000oooo
0?ooo`130?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo
0?ooo`0Z0?ooo`L000004P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo
001:0?ooo`030000003oool0oooo04<0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?oo
o`130?ooo`030000O`3oool0oooo02T0oooo100000000`3oool000000000000C0?ooo`030000O`3o
ool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo0P00003i0000O`L00000G00007l8
0?ooo`00BP3oool00`000000oooo0?ooo`130?ooo`030000O`3oool0oooo0480oooo00<0001o0?oo
o`3oool0@`3oool00`0007l0oooo0?ooo`0V0?ooo`P000005@3oool00`0007l0oooo0?ooo`130?oo
o`030000O`3oool0oooo00H0oooo001:0?ooo`030000003oool0oooo04<0oooo00<0001o0?ooo`3o
ool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo02D0oooo2000000F0?ooo`03
0000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3oool0
@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0
9@3oool7000001L0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3o
ool00`000000oooo0?ooo`130?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3o
ool00`0007l0oooo0?ooo`0V0?ooo`@000006@3oool00`0007l0oooo0?ooo`130?ooo`030000O`3o
ool0oooo00H0oooo001:0?ooo`800000n00007l300000640001o203oool004X0oooo00<000000?oo
o`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?oo
o`3oool08`3oool3000001d0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?oo
o`00BP3oool00`000000oooo0?ooo`130?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3o
ool0@`3oool00`0007l0oooo0?ooo`0R0?ooo`@000007@3oool00`0007l0oooo0?ooo`130?ooo`03
0000O`3oool0oooo00H0oooo001:0?ooo`800000l`0007l6000006<0001o203oool004X0oooo00<0
00000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0
001o0?ooo`3oool0803oool7000001`0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?oo
o`060?ooo`00BP3oool200000?40001o1`00001T0000O`P0oooo001:0?ooo`030000003oool0oooo
04<0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo
01`0oooo2@00000N0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool001<0
oooo00D0omnL04P000000000001d0;ooo`030?ooo`030?noM0000000M;oo00<0oooo00D0omnL04P0
00000000001d0;ooo`040?ooo`050?oOW018000000000000M02oool0103oool01@3ogi`0B0000000
000007@0_ooo00@0oooo00H0_g@000000000000000000018W03Oool80?ooo`800000k@0007l90000
06H0001o203oool00180oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00<0okmd
0000001d_ol00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3ookl0M000
07Boo`3oool0okmd0000M02oool00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0103o
ool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo04<0oooo00<0001o0?ooo`3oool0@P3o
ool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo01X0oooo2@00000P0?ooo`030000O`3o
ool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool00180oooo00<0_g@00018W03Oool00P3oool0
0`2oM00007Bo0?ooo`060?ooo`030;md0000B9`0gooo0080oooo00L0_g@0001d_`3oool0oooo0;md
0000B9`0gooo0080oooo00L0_g@0001d_`3oool0oooo0;md0000B9`0gooo0080oooo00<0_g@0001d
_`3oool00`3oool00`3ogi`0B00007Boo`0:0?ooo`800000k00007l8000006P0001o203oool00140
oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`050?ooo`030?oo_`1d0000B9cO00<0
oooo00H0gia80000M02oool0oono07@00018W=l30?ooo`060=nLB00007@0_ooo0?oo_`1d0000B9cO
0`3oool00`3OW4P0001d0;ooo`030?ooo`030?oOW0180000M;oo00X0oooo0P00003]0000O`@00000
J`0007l80?ooo`004@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00D0oooo00<0
omnL04P0001d_ol00`3oool01P3o_g@0001809cOo`3ogi`0B00007Boo`<0oooo00H0okmd0000B02L
gol0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo00<0oooo00<0omnL04P0001d_ol02P3oool2
000004@0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0
oooo01T0oooo0P0000000`3oool000000000000U0?ooo`030000O`3oool0oooo04<0oooo00<0001o
0?ooo`3oool01P3oool00140oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`050?oo
o`030?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0omnL04P0001d_ol30?ooo`060?noM000
04P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`030?ooo`030?oOW0180000M;oo00X0
oooo0`00003Y0000O`@00000KP0007l80?ooo`004@3oool00`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00D0oooo00<0omnL04P0001d_ol00`3oool01P3o_g@0001809cOo`3ogi`0B00007Bo
o`<0oooo00H0okmd0000B02Lgol0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo00<0oooo00<0
omnL04P0001d_ol02P3oool2000004@0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?oo
o`130?ooo`030000O`3oool0oooo01L0oooo1P00000V0?ooo`030000O`3oool0oooo04<0oooo00<0
001o0?ooo`3oool01P3oool00140oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`05
0?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0omnL04P0001d_ol30?ooo`060?no
M00004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`030?ooo`030?oOW0180000M;oo
00X0oooo0P0000140?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`00
07l0oooo0?ooo`0E0?ooo`P000009P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo
00H0oooo000A0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01@3oool00`3ookl0
M00004RLg`030?ooo`060=nLB00007@0_ooo0?oo_`1d0000B9cO0`3oool01P3OW4P0001d0;ooo`3o
okl0M00004RLg`<0oooo00<0gia80000M02oool00`3oool00`3ogi`0B00007Boo`0:0?ooo`800000
A03oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0
503oool8000002L0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`004P3o
ool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00H0oooo00<0gia80018W03Oool00P3o
ool01`2oM00007Bo0?ooo`3oool0gia80018W03Oool00P3oool01`2oM00007Bo0?ooo`3oool0gia8
0018W03Oool00P3oool00`2oM00007Bo0?ooo`030?ooo`030?oOW0180000M;oo00X0oooo0P000014
0?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0C
0?ooo`P00000:03oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo000B0?oo
o`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`070?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0
001d0;ooo`020?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`050?oog`2LB00000000000001d_ol02P3oool2000004@0
oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo01<0
oooo1P00000Z0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool001<0oooo
00D0oono07@000000000001d0;ooo`090?ooo`050?oo_`1d000000000000M02oool0103oool01@3o
okl0M0000000000007@0_ooo00@0oooo00D0oono07@000000000001d0;ooo`050?ooo`030?noM000
0000M;oo00X0oooo0P0000140?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3o
ool00`0007l0oooo0?ooo`0B0?ooo`H00000:`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3o
ool0oooo00H0oooo001:0?ooo`800000A03oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0
oooo04<0oooo00<0001o0?ooo`3oool04@3oool6000002`0oooo00<0001o0?ooo`3oool0@`3oool0
0`0007l0oooo0?ooo`060?ooo`00BP3oool2000004@0oooo00<0001o0?ooo`3oool0@P3oool00`00
07l0oooo0?ooo`130?ooo`030000O`3oool0oooo00h0oooo2000000]0?ooo`030000O`3oool0oooo
04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo0P0000140?ooo`030000O`3oool0oooo0480
oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0=0?ooo`T00000;@3oool00`0007l0
oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`800000A03oool00`0007l0oooo
0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0303oool9000002h0oooo
00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool2000004@0oooo00<0
001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00`0oooo2000
000_0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo0P000014
0?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`0;
0?ooo`H00000<P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo001:0?oo
o`800000A03oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?oo
o`3oool02P3oool6000003<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?oo
o`00BP3oool2000004@0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`03
0000O`3oool0oooo00T0oooo1P00000d0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3o
ool01P3oool004X0oooo0P0000140?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0
@`3oool00`0007l0oooo0?ooo`070?ooo`P00000=03oool00`0007l0oooo0?ooo`130?ooo`030000
O`3oool0oooo00H0oooo001:0?ooo`800000A03oool00`0007l0oooo0?ooo`120?ooo`030000O`3o
ool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool8000003D0oooo00<0001o0?ooo`3oool0@`3o
ool00`0007l0oooo0?ooo`060?ooo`00BP3oool2000004@0oooo00<0001o0?ooo`3oool0@P3oool0
0`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00@0oooo2@00000f0?ooo`030000O`3oool0
oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo0P0000140?ooo`030000O`3oool0oooo
0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`030?ooo`P00000>03oool00`00
07l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`800000A03oool00`0007l0
oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool00P3oool8000003T0
oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool2000004@0oooo
00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00T00000
>P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`800000A03o
ool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`0000001`00
000l0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo0P000014
0?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0000000000006
000003d0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool20000
04@0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`L00000?`3oool00`00
07l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`030000003oool0000004<0
oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`120?ooo`L00000@03oool00`0007l0
oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`030000003oool0000004<0oooo
00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`110?ooo`L00000@@3oool00`0007l0oooo
0?ooo`130?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`030000003oool0000004<0oooo00<0
001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`110?ooo`H00000@P3oool00`0007l0oooo0?oo
o`130?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`<00000bP0007l2000000030000O`000000
000000800000RP0007l80?ooo`00BP3oool00`000000oooo000000130?ooo`030000O`3oool0oooo
0480oooo00<0001o0?ooo`3oool0?P3oool4000000030?ooo`000000000004@0oooo00<0001o0?oo
o`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool00`000000oooo000000130?ooo`03
0000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0?@3oool6000000030000O`3oool0oooo04<0
oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool00`000000oooo
000000130?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0?03oool7000000030000
O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3o
ool00`000000oooo000000130?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0>`3o
ool7000000030?ooo`0007l0oooo04@0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?oo
o`060?ooo`00BP3oool010000000oooo0?ooo`0000120?ooo`030000O`3oool0oooo0480oooo00<0
001o0?ooo`3oool0>P3oool700000080oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?oo
o`130?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`040000003oool0oooo00000480oooo00<0
001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`0h0?ooo`P000000`3oool00`0007l0oooo0?oo
o`130?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00@00000
0?ooo`3oool00000@P3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo03L0oooo2000
00040?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?oo
o`060?ooo`00BP3oool2000000030000O`000000001o0;l0001o2000002B0000O`P0oooo001:0?oo
o`040000003oool0oooo00000480oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`0e
0?ooo`L000001`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0001o
0?ooo`3oool01P3oool004X0oooo00D000000?ooo`3oool0oooo000000110?ooo`030000O`3oool0
oooo0480oooo00<0001o0?ooo`3oool0<@3oool2000000030?ooo`000000000000D00000203oool0
0`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool0
04X0oooo00D000000?ooo`3oool0oooo000000110?ooo`030000O`3oool0oooo0480oooo00<0001o
0?ooo`3oool0<03oool:000000T0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`13
0?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`8000000P0007l00`000000001o0000O`2f0000
O`X00000U`0007l80?ooo`00BP3oool01@000000oooo0?ooo`3oool000000440oooo00<0001o0?oo
o`3oool0@P3oool00`0007l0oooo0?ooo`0^0?ooo`X000002`3oool00`0007l0oooo0?ooo`130?oo
o`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00D000000?ooo`3o
ool0oooo000000110?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0;@3oool70000
00030?ooo`000000000000`0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?oo
o`030000O`3oool0oooo00H0oooo001:0?ooo`8000000`0007l00`000000001o0000O`2b0000O`L0
0000W@0007l80?ooo`00BP3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo03h0oooo
00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`0Z0?ooo`P000004@3oool00`0007l0oooo
0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo0P00
00030000O`030000000007l0001o0:l0001o2000002O0000O`P0oooo001:0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0?P3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0
oooo02L0oooo2@00000C0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool0
0`0007l0oooo0?ooo`060?ooo`009@3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00<0okmd
0000001d_ol00`3oool01@3ogi`0B0000000000007@0_ooo00@0oooo00H0_g@00000000000000000
0018W03Oool80?ooo`800000100007l00`000000001o0000O`2[0000O`P00000XP0007l80?ooo`00
903oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool00`3o_g@0000007Boo`020?oo
o`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`040?ooo`030?oOW0180000M;oo00X0oooo00<0
00000?ooo`3oool00`3oool2000003h0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?oo
o`0U0?ooo`P000005P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0
001o0?ooo`3oool01P3oool002@0oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`06
0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool00`3oool00`3ogi`0B00007Boo`0:
0?ooo`8000000`0007l400000:X0001o1`00002T0000O`P0oooo000S0?ooo`030?oo_`1d0000B9cO
00<0oooo00<0gia80000M02oool01@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo
00<0oooo00<0omnL04P0001d_ol02P3oool010000000oooo0?ooo`3oool6000003`0oooo00<0001o
0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`0S0?ooo`80000000<0oooo0000000000000P00000I
0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`06
0?ooo`008`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00D0oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`030?ooo`030?oOW0180000M;oo00X0oooo0P0000020000
O`H00000YP0007l4000000030000O`00000000000:L0001o203oool002<0oooo00<0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`050?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol00`3oool00`3ogi`0B00007Boo`0:0?ooo`<000000P0007l400000:D0001o1`00002Y0000O`P0
oooo000S0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01@3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00<0oooo00<0omnL04P0001d_ol02P3oool00`000000oooo
0?ooo`030?ooo`800000?P3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo01l0oooo
2000000L0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo
0?ooo`060?ooo`008`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00D0oooo00<0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`030?ooo`030?oOW0180000M;oo00X0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`0k0?ooo`030000O`3oool0oooo0480oooo00<0
001o0?ooo`3oool0703oool:000001d0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?oo
o`130?ooo`030000O`3oool0oooo00H0oooo000S0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia8
0000M02oool01@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00<0oooo00<0omnL
04P0001d_ol02P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo03/0oooo00<0001o
0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`0K0?ooo`X000007P3oool00`0007l0oooo0?ooo`13
0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool002@0oooo00<0gia80018
W03Oool00P3oool00`2oM00007Bo0?ooo`060?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d
_`3oool00`3oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool0>`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo01T0oooo2P00000P0?oo
o`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?oo
o`00903oool01`3ookl0M00007Boo`3oool0okmd0000M02oool01`3oool01`3ookl0M00007Boo`3o
ool0okmd0000M02oool00P3oool01@3ooml0W4P0000000000000M;oo00X0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0k0?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3o
ool0603oool:00000240oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`03
0000O`3oool0oooo00H0oooo000U0?ooo`050?oo_`1d000000000000M02oool02@3oool01@3ookl0
M0000000000007@0_ooo00D0oooo00<0okmd0000001d_ol02P3oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo03X0oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`0G0?oo
o`P00000903oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0001o0?oo
o`3oool01P3oool004X0oooo00<000000?ooo`3oool01P3oool3000003X0oooo00<0001o0?ooo`3o
ool0@P3oool00`0007l0oooo0?ooo`0B0?ooo`80000000<0oooo0000000000001`00000U0?ooo`03
0000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00
BP3oool00`000000oooo0?ooo`060?ooo`@00000>@3oool00`0007l0oooo0?ooo`120?ooo`030000
O`3oool0oooo0140oooo2`00000W0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0
@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool00`000000oooo0?ooo`050?ooo`H00000>03o
ool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo0100oooo2`00000X0?ooo`030000O`3o
ool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool0
0`000000oooo0?ooo`050?ooo`H00000>03oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0
oooo0100oooo2@00000Z0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool0
0`0007l0oooo0?ooo`060?ooo`00BP3oool00`000000oooo0?ooo`060?ooo`@00000>@3oool00`00
07l0oooo0?ooo`120?ooo`030000O`3oool0oooo00h0oooo0P0000000`3oool00000000000050000
02/0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo
00H0oooo001:0?ooo`030000003oool0oooo00L0oooo0`00000i0?ooo`030000O`3oool0oooo0480
oooo00<0001o0?ooo`3oool03@3oool7000000030?ooo`000000000002`0oooo00<0001o0?ooo`3o
ool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`030000
003oool0oooo00X0oooo0P00000g0?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0
2`3oool700000340oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000
O`3oool0oooo00H0oooo001:0?ooo`030000003oool0oooo00T0oooo1000000f0?ooo`030000O`3o
ool0oooo0480oooo00<0001o0?ooo`3oool02P3oool800000340oooo00<0001o0?ooo`3oool0@`3o
ool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`030000003oool0
oooo00P0oooo1P00000e0?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool01`3oool:
00000380oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0
oooo00H0oooo001:0?ooo`030000003oool0oooo00P0oooo2000000c0?ooo`030000O`3oool0oooo
0480oooo00<0001o0?ooo`3oool0103oool<000003<0oooo00<0001o0?ooo`3oool0@`3oool00`00
07l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`030000003oool0oooo00T0
oooo2000000b0?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool00`3oool;000003D0
oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0
oooo001:0?ooo`030000003oool0oooo00X0oooo2000000a0?ooo`030000O`3oool0oooo0480oooo
00<0001o0?ooo`3oool03@00000f0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0
@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool00`000000oooo0?ooo`0<0?ooo`H00000<@3o
ool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0000000X00000>@3oool00`0007l0oooo0?oo
o`130?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<00000
0?ooo`3oool03@3oool400000380oooo00<0001o0?ooo`3oool0@P3oool<000003X0oooo00<0001o
0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo001:0?oo
o`030000003oool0oooo00h0oooo1@00000`0?ooo`030000O`3oool0oooo0440oooo2P00000m0?oo
o`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?oo
o`00BP3oool00`000000oooo0?ooo`0@0?ooo`@00000;`3oool00`0007l0oooo0?ooo`0n0?ooo`X0
0000@03oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3o
ool01P3oool004X0oooo00<000000?ooo`3oool03`3oool8000002`0oooo00<0001o0?ooo`3oool0
>`3oool<00000440oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000
O`3oool0oooo00H0oooo001:0?ooo`030000003oool0oooo00l0oooo2@00000[0?ooo`030000O`3o
ool0oooo03L0oooo0P0000000`3oool0000000000009000004<0oooo00<0001o0?ooo`3oool0@`3o
ool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo001:0?ooo`030000003oool0
oooo0100oooo2@00000Z0?ooo`030000O`3oool0oooo03@0oooo400000140?ooo`030000O`3oool0
oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool00`00
0000oooo0?ooo`0A0?ooo`P00000:P3oool00`0007l0oooo0?ooo`0c0?ooo`h0000000<0oooo0000
O`3oool0A03oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0001o0?oo
o`3oool01P3oool004X0oooo0P00000E0000O`H00000G00007l@00000=D0001o203oool004X0oooo
00<000000?ooo`3oool05@3oool6000000030?ooo`000000000002D0oooo00<0001o0?ooo`3oool0
;@3oool=000000030?ooo`000000000000D0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo
0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<0
00000?ooo`3oool05P3oool9000002@0oooo00<0001o0?ooo`3oool0;03oool=000000T0oooo00<0
001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0
001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3oool05P3oool;00000280oooo00<0001o
0?ooo`3oool0:@3oool=000000`0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`13
0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<000000?oo
o`3oool05`3oool;000000030?ooo`000000000001h0oooo00<0001o0?ooo`3oool09P3oool=0000
00l0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo
04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3oool0603oool200000003
0?ooo`000000000000/000006`3oool00`0007l0oooo0?ooo`0L0?ooo`8000001@3oool?00000100
oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0
oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<000000?ooo`3oool0703oool=000000030?oo
o`000000000001L0oooo00<0001o0?ooo`3oool06`3oool6000000030?ooo`000000000000/00000
4`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0
@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool00`000000oooo0?ooo`0N0?ooo`l000005P3o
ool00`0007l0oooo0?ooo`0H0?oooa@000005P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3o
ool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool2
00000200001o0P0000000`0007l000000000000:000002T0001o0P0000040000Oa@00000jP0007l8
0?ooo`00BP3oool00`000000oooo0?ooo`0S0?ooo``00000303oool200000080oooo0P0000020?oo
o`030000O`3oool0oooo00X0oooo0P0000020?ooo`H0000000<0oooo0000000000004@00000J0?oo
o`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?oo
o`030000O`3oool0oooo00H0oooo001:0?ooo`030000003oool0oooo02H0oooo0P0000000`3oool0
000000000005000000040?ooo`00000000000?ooo`800000103oool8000000030?ooo`0007l0oooo
00@0oooo0P0000040?oooaT0000000<0oooo0000000000007@3oool00`0007l0oooo0?ooo`130?oo
o`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?oo
o`00BP3oool00`000000oooo0?ooo`0Z0?ooo`d0000000<0oooo0000000000002P0000020?ooo`H0
000000<0oooo0000000000004P0000040?ooo`8000008@3oool00`0007l0oooo0?ooo`130?ooo`03
0000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00
BP3oool2000002`0001o=@00003k0000O`P0oooo001:0?ooo`030000003oool0oooo02`0oooo;@00
00000`3oool0000000000002000002X0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?oo
o`130?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo00<00000
0?ooo`3oool0;@3oool2000000030?ooo`000000000000/000000P3ooolC00000080oooo0P000005
0?ooo`800000:`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0001o
0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool200000380001o0P0000001000
07l000000000000007l4000000H0001o2P0000000`0007l000000000000200000?l0001o300007l8
0?ooo`00BP3oool00`000000oooo0?ooo`0h0?ooo`800000203oool200000080oooo0P0000050?oo
o`800000>@3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0oooo00<0001o0?oo
o`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00BP3oool00`000000oooo0?ooo`130?ooo`03
0000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`03
0000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool004X0oooo0P00003o0000Oed0001o
203oool003/0oooo00H0_g@000000000000000000018W03Oool90?ooo`030000003oool0oooo04<0
oooo00<0001o0?ooo`3oool0@P3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo04<0
oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`060?ooo`00?03oool00`3ogi`0B000
07Boo`0;0?ooo`800000o`0007mM0000O`P0oooo000l0?ooo`030?oOW0180000M;oo00/0oooo0P00
003o0000Oed0001o203oool003`0oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`13
0?ooo`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`13
0?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool003`0oooo00<0omnL04P0
001d_ol02`3oool200000?l0001oG@0007l80?ooo`00?03oool00`3ogi`0B00007Boo`0;0?ooo`<0
0000o`0007mL0000O`P0oooo000l0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0
@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0
@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo000l0?ooo`030?oOW018
0000M;oo00/0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3o
ool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3o
ool0oooo00H0oooo000l0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0@`3oool0
0`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool0@`3oool0
0`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo00H0oooo000l0?ooo`030?oOW0180000M;oo
00/0oooo00<000000?ooo`3oool0@`3oool00`0007l0oooo0?ooo`120?ooo`030000O`3oool0oooo
04<0oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?ooo`030000O`3oool0oooo
00H0oooo000j0?ooo`050?oog`2LB00000000000001d_ol02`3oool00`000000oooo0?ooo`130?oo
o`030000O`3oool0oooo0480oooo00<0001o0?ooo`3oool0@`3oool00`0007l0oooo0?ooo`130?oo
o`030000O`3oool0oooo04<0oooo00<0001o0?ooo`3oool01P3oool003`0oooo00<0okmd0000001d
_ol0o`3ooomb0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?oo
o`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?oo
o`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.06273, -9.63436, \
0.0214737, 0.0586649}}]
}, Closed]],

Cell[TextData[{
  "If you suspect that your data follow a power-law function, ",
  Cell[BoxData[
      \(TraditionalForm\`y \[Proportional] x\^\[Alpha]\)]],
  ", a log-log plot should display a straight line, as simulated below."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[
        LogLogPlot[x\^\(\(-1\)/3\), {x, 0, 5}, GridLines \[Rule] Automatic], 
        LogLogListPlot[
          Table[{x, x\^\(\(-1\)/3\)\ Random[Real, {0.9, 1.1}]}, {x, 0.1, 4, 
              0.02}], PlotStyle \[Rule] {AbsolutePointSize[
                4]}]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.612471 0.520365 0.239506 0.96481 [
[.09211 -0.0125 -9.5 -15.6875 ]
[.09211 -0.0125 9.5 0 ]
[.24875 -0.0125 -9.5 -15.6875 ]
[.24875 -0.0125 9.5 0 ]
[.45583 -0.0125 -9.5 -15.6875 ]
[.45583 -0.0125 9.5 0 ]
[.61247 -0.0125 -5 -15.6875 ]
[.61247 -0.0125 5 0 ]
[.76912 -0.0125 -5 -15.6875 ]
[.76912 -0.0125 5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[-0.0125 .23951 -10 -7.84375 ]
[-0.0125 .23951 0 7.84375 ]
[-0.0125 .4094 -19 -7.84375 ]
[-0.0125 .4094 0 7.84375 ]
[-0.0125 .52994 -10 -7.84375 ]
[-0.0125 .52994 0 7.84375 ]
[ -0.0005 -0.0005 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.09211 0 m
.09211 .61803 L
s
.24875 0 m
.24875 .61803 L
s
.45583 0 m
.45583 .61803 L
s
.61247 0 m
.61247 .61803 L
s
.76912 0 m
.76912 .61803 L
s
.97619 0 m
.97619 .61803 L
s
.34038 0 m
.34038 .61803 L
s
.4054 0 m
.4054 .61803 L
s
.49703 0 m
.49703 .61803 L
s
.53187 0 m
.53187 .61803 L
s
.56204 0 m
.56204 .61803 L
s
.58866 0 m
.58866 .61803 L
s
.86075 0 m
.86075 .61803 L
s
.92576 0 m
.92576 .61803 L
s
0 .23951 m
1 .23951 L
s
0 .4094 m
1 .4094 L
s
0 .52994 m
1 .52994 L
s
0 .27944 m
1 .27944 L
s
0 .3159 m
1 .3159 L
s
0 .34944 m
1 .34944 L
s
0 .38049 m
1 .38049 L
s
0 .43644 m
1 .43644 L
s
0 .46185 m
1 .46185 L
s
0 .4858 m
1 .4858 L
s
0 .50845 m
1 .50845 L
s
0 g
.09211 0 m
.09211 .00625 L
s
gsave
.09211 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.1) show
78.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.24875 0 m
.24875 .00625 L
s
gsave
.24875 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
78.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.45583 0 m
.45583 .00625 L
s
gsave
.45583 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
78.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.61247 0 m
.61247 .00625 L
s
gsave
.61247 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
69.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.76912 0 m
.76912 .00625 L
s
gsave
.76912 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
69.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
69.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.001 w
.34038 0 m
.34038 .00375 L
s
.4054 0 m
.4054 .00375 L
s
.49703 0 m
.49703 .00375 L
s
.53187 0 m
.53187 .00375 L
s
.56204 0 m
.56204 .00375 L
s
.58866 0 m
.58866 .00375 L
s
.86075 0 m
.86075 .00375 L
s
.92576 0 m
.92576 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .23951 m
.00625 .23951 L
s
gsave
-0.0125 .23951 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
69.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .4094 m
.00625 .4094 L
s
gsave
-0.0125 .4094 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.5) show
78.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .52994 m
.00625 .52994 L
s
gsave
-0.0125 .52994 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
69.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.001 w
0 .27944 m
.00375 .27944 L
s
0 .3159 m
.00375 .3159 L
s
0 .34944 m
.00375 .34944 L
s
0 .38049 m
.00375 .38049 L
s
0 .43644 m
.00375 .43644 L
s
0 .46185 m
.00375 .46185 L
s
0 .4858 m
.00375 .4858 L
s
0 .50845 m
.00375 .50845 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
0 .61803 m
.25193 .46233 L
.41859 .35933 L
.5087 .30364 L
.5708 .26526 L
.62226 .23346 L
.66271 .20845 L
.69901 .18602 L
.7292 .16736 L
.75488 .15149 L
.77933 .13638 L
.8006 .12323 L
.81933 .11166 L
.8377 .10031 L
.85407 .09019 L
.87027 .08018 L
.88485 .07117 L
.89803 .06302 L
.91127 .05484 L
.92332 .04739 L
.93545 .03989 L
.94656 .03303 L
.95676 .02673 L
.96712 .02032 L
.97619 .01472 L
s
4 Mabswid
.09211 .56287 Mdot
.13331 .5621 Mdot
.16815 .51314 Mdot
.19832 .45425 Mdot
.22494 .47274 Mdot
.24875 .47101 Mdot
.27029 .47816 Mdot
.28995 .44038 Mdot
.30804 .39233 Mdot
.32479 .44304 Mdot
.34038 .4445 Mdot
.35497 .42505 Mdot
.36867 .40882 Mdot
.38159 .35438 Mdot
.39381 .41366 Mdot
.4054 .38081 Mdot
.41642 .39931 Mdot
.42694 .34121 Mdot
.43698 .38201 Mdot
.4466 .35747 Mdot
.45583 .35632 Mdot
.46469 .3467 Mdot
.47322 .3636 Mdot
.48144 .30926 Mdot
.48937 .29174 Mdot
.49703 .34053 Mdot
.50444 .30515 Mdot
.51161 .28788 Mdot
.51857 .32177 Mdot
.52531 .27505 Mdot
.53187 .30184 Mdot
.53823 .31119 Mdot
.54442 .29815 Mdot
.55045 .27526 Mdot
.55632 .29001 Mdot
.56204 .31003 Mdot
.56762 .30522 Mdot
.57307 .24593 Mdot
.57839 .27734 Mdot
.58358 .24294 Mdot
.58866 .29351 Mdot
.59363 .28671 Mdot
.59849 .27106 Mdot
.60325 .25703 Mdot
.60791 .22058 Mdot
.61247 .21802 Mdot
.61695 .26162 Mdot
.62133 .21477 Mdot
.62564 .18943 Mdot
.62986 .21885 Mdot
.63401 .21077 Mdot
.63808 .25896 Mdot
.64208 .1976 Mdot
.64601 .18356 Mdot
.64988 .23021 Mdot
.65367 .18107 Mdot
.65741 .21376 Mdot
.66108 .22002 Mdot
.6647 .2453 Mdot
.66826 .17275 Mdot
.67176 .207 Mdot
.67521 .18684 Mdot
.67861 .17902 Mdot
.68196 .22084 Mdot
.68526 .19936 Mdot
.68851 .18355 Mdot
.69172 .18972 Mdot
.69488 .15864 Mdot
.69799 .1705 Mdot
.70107 .15381 Mdot
.7041 .19806 Mdot
.7071 .21171 Mdot
.71005 .16098 Mdot
.71297 .19872 Mdot
.71585 .16417 Mdot
.71869 .20925 Mdot
.7215 .13363 Mdot
.72427 .18394 Mdot
.72701 .18496 Mdot
.72971 .19296 Mdot
.73239 .16848 Mdot
.73503 .12299 Mdot
.73764 .18035 Mdot
.74023 .17615 Mdot
.74278 .19794 Mdot
.74531 .1738 Mdot
.7478 .15171 Mdot
.75027 .18592 Mdot
.75272 .14685 Mdot
.75513 .13469 Mdot
.75752 .18675 Mdot
.75989 .16773 Mdot
.76223 .11366 Mdot
.76455 .11611 Mdot
.76685 .12487 Mdot
.76912 .17154 Mdot
.77137 .16816 Mdot
.77359 .13147 Mdot
.7758 .17163 Mdot
.77798 .17068 Mdot
.78014 .1579 Mdot
.78228 .15389 Mdot
.78441 .10847 Mdot
.78651 .0967 Mdot
.78859 .10765 Mdot
.79066 .14579 Mdot
.7927 .12736 Mdot
.79473 .11899 Mdot
.79674 .10374 Mdot
.79873 .08027 Mdot
.8007 .08276 Mdot
.80266 .12327 Mdot
.8046 .14647 Mdot
.80652 .13561 Mdot
.80843 .08164 Mdot
.81032 .14001 Mdot
.8122 .13236 Mdot
.81406 .11775 Mdot
.8159 .13854 Mdot
.81773 .14669 Mdot
.81955 .14123 Mdot
.82135 .07854 Mdot
.82313 .14108 Mdot
.8249 .06476 Mdot
.82666 .07197 Mdot
.82841 .09738 Mdot
.83014 .11987 Mdot
.83186 .137 Mdot
.83356 .13236 Mdot
.83526 .11808 Mdot
.83694 .07396 Mdot
.83861 .10007 Mdot
.84026 .10859 Mdot
.8419 .11395 Mdot
.84354 .06627 Mdot
.84516 .13474 Mdot
.84676 .11934 Mdot
.84836 .04981 Mdot
.84995 .05503 Mdot
.85152 .06663 Mdot
.85309 .05627 Mdot
.85464 .08721 Mdot
.85618 .06874 Mdot
.85771 .06958 Mdot
.85924 .06437 Mdot
.86075 .072 Mdot
.86225 .07422 Mdot
.86374 .0652 Mdot
.86522 .05196 Mdot
.8667 .11752 Mdot
.86816 .09716 Mdot
.86961 .06932 Mdot
.87106 .0604 Mdot
.87249 .05638 Mdot
.87392 .07797 Mdot
.87533 .10646 Mdot
.87674 .08904 Mdot
.87814 .0788 Mdot
.87953 .06176 Mdot
.88091 .10393 Mdot
.88229 .10134 Mdot
.88365 .07518 Mdot
.88501 .05232 Mdot
.88636 .08396 Mdot
.8877 .08974 Mdot
.88903 .03095 Mdot
.89036 .02509 Mdot
.89168 .05683 Mdot
.89299 .06709 Mdot
.89429 .08462 Mdot
.89558 .07646 Mdot
.89687 .02903 Mdot
.89815 .05205 Mdot
.89943 .08643 Mdot
.90069 .09777 Mdot
.90195 .07985 Mdot
.9032 .02459 Mdot
.90445 .06396 Mdot
.90569 .05435 Mdot
.90692 .08756 Mdot
.90814 .05179 Mdot
.90936 .01346 Mdot
.91057 .02055 Mdot
.91178 .09417 Mdot
.91298 .06108 Mdot
.91417 .05119 Mdot
.91536 .0784 Mdot
.91654 .03707 Mdot
.91771 .07837 Mdot
.91888 .0423 Mdot
.92004 .07426 Mdot
.9212 .08474 Mdot
.92235 .03331 Mdot
.92349 .06144 Mdot
.92463 .01717 Mdot
.92576 .07396 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00@P3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00<0okmd0000001d_ol00`3oool0
1P2oM000000000000000000004RL0=ooobP0oooo00D0omnL04P000000000001d0;ooo`030?ooo`05
0?noM0000000M;oo0?ooo`2oM0001P0000000`18W=l0oooo0?ooo`0i0?ooo`050?oOW01800000000
0000M02oool00`3oool01P3o_g@0000007Boo`3oool0oooO09a800<0000000<0M;oo0?ooo`3oool0
;`3oool01P2oM000000000000000000004RL0=ooocD0oooo00<0_g@0000000000000100000000`18
W=l0oooo0?ooo`150?ooo`030?oog`2LB00000000080000000<0M;oo0?ooo`3oool03P3oool00440
oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00<0okmd0000001d_ol0103oool0
0`3ogi`0B00007Boo`0Y0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?no
M0000000M;oo0080oooo00<0gia80000000000000`0000000`00B9`0gooo0?ooo`0h0?ooo`070?oo
_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo0?ooo`3ooml0W4Qd0;oo
o`020?ooo`030;md001d_ol0oooo0300oooo00<0omnL04P0001d_ol0>03oool00`3OW4P000000000
0003000000030018W03Oool0oooo04D0oooo00<0oooO09a8M02oool00P3oool00`2oM000M;oo0?oo
o`0>0?ooo`00@@3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00P0oooo00<0omnL
04P0001d_ol0:@3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0gia8
04RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`0h0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0
001d_`3oool02P3oool00`3ooml0W4Qd0;ooo`0`0?ooo`030?oOW0180000M;oo03T0oooo00<0gia8
04RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`1:0?ooo`030?oog`2LB7@0_ooo00h0oooo00100?oo
o`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool00`3ogi`0B00007Boo`0X0?oo
o`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool00`3OW4P0M;oo0?ooo`0k0?oo
o`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02`3oool00`2oM000M;oo0?ooo`0_0?oo
o`030?oOW0180000M;oo03X0oooo00<0gia807Boo`3oool0C`3oool00`2oM000M;oo0?ooo`0=0?oo
o`00@03oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00P0oooo00<0omnL04P0001d
_ol0:03oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00<0_gA809cOo`3o
ool0>P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0oooO09a80018
W=l0<03oool00`3ogi`0B00007Boo`0k0?ooo`030;mdB02Lgol0oooo04d0oooo00<0oooO09a80018
W=l03P3oool00400oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`080?ooo`030?oO
W0180000M;oo02P0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`090?ooo`030?oo
g`2LB7@0_ooo03X0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030=nL
B0000000M;oo0300oooo00<0omnL04P0001d_ol0>`3oool00`3ooml0W4Qd0;ooo`1=0?ooo`030=nL
B0000000M;oo00h0oooo00100?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0203o
ool00`3ogi`0B00007Boo`0X0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3o
ool00`3o_g@007Bo0?ooo`0i0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3o
ool01`3Ogml0oooo0?oo_`1d000000000000M02oool0<03oool00`3ogi`0B00007Boo`0l0?ooo`03
0?noM000M;l0oooo04P0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00h0oooo00100?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0203oool00`3ogi`0B00007Boo`0X0?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3ookl0M01809cOo`0i0?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3OW4P00000000000000000
001809cOoc40oooo00<0omnL04P0001d_ol0?03oool00`3ookl0M01809cOo`180?ooo`060=nLB000
000000000000000004P0W=oo3`3oool00400oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d
0;ooo`080?ooo`030?oOW0180000M;oo02P0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d
0;ooo`0:0?ooo`030?oo_`1d0000B9cO03T0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d
0;ooo`060?ooo`040?oo_`1d0000000004RLgc<0oooo00<0omnL04P0001d_ol0?03oool00`3ookl0
M00004RLg`180?ooo`040?oo_`1d0000000004RLga40oooo00110?ooo`030=nLB000B9`0gooo0080
oooo00<0_g@0001d_`3oool0203oool00`3ogi`0B00007Boo`0Y0?ooo`030=nLB000B9`0gooo0080
oooo00<0_g@0001d_`3oool01@3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO03X0
oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030;md001d_ol0oooo03<0
oooo00<0omnL04P0001d_ol0=`3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO04T0
oooo00<0_g@007Boo`3oool04@3oool00440oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo
00L0oooo00D0oooO09a800000000000007Boo`0Y0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d
0;ooo`060?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02ooolj0?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`060?oOW018000000000000000004P0W=oo;P3oool01@3o
oml0W4P0000000000000M;oo03L0oooo00P0oooO09a8B02Lgol0oooo0?oo_`1d0000001d0;ooodT0
oooo00H0omnL04P00000000000000000B02Lgol>0?ooo`00@P3oool01@3ookl0M0000000000007@0
_ooo00X0oooo00<0okmd0000001d_ol0:P3oool01@3ookl0M0000000000007@0_ooo00P0oooo00H0
oooO09a80000000000000000M02oooll0?ooo`050?oo_`1d000000000000M02oool02P3oool01@2o
M0000000000000000000B9cO0300oooo00<0okmd0000001d_ol0>03oool01P3ooml0W4P000000000
0000001d0;oood/0oooo00D0_g@0000000000000000004RLg`0>0?ooo`00o`3ooona0?ooo`00o`3o
oona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3o
oona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`009`3ooooo000008<00000
1`3oool002L0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0i0?ooo`030000003o
ool0oooo0200oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0A0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0i0?ooo`030000003o
ool0oooo0200oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0A0?ooo`030000003o
ool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0>@3oool0
0`000000oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool0
0`000000oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool0
0`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<000000?ooo`3oool0>@3oool0
0`000000oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool0
0`000000oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo
03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo
0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo
00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo
03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0?0?ooo`8000001@3oool00`00
07l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240
oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0
oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0
oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0
oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo00h0
oooo100000030?ooo`8000004`3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo
0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo
0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo
0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo
0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo
0?ooo`0=0?ooo`H0000000<0oooo0000000000000P00000B0?ooo`030000O`3oool0oooo00h0oooo
000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?oo
o`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?oo
o`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?oo
o`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?oo
o`0P0?ooo`030000O`3oool0oooo00d0oooo3000000A0?ooo`030000003oool0oooo00h0oooo000W
0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P
0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=
0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`08
0?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P
0?ooo`030000O`3oool0oooo00P0oooo0P0000030?ooo``000003`3oool2000000030000O`3oool0
oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`00
07l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`00
07l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`00
07l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`00
07l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo00L0oooo100000000`3oool000000000000:0000
00h0oooo0P0000020?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240
oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0
oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0
oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0
oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo00H0
oooo3@0000020?ooo`8000003@3oool2000000@0oooo00<0001o0?ooo`3oool03P3oool002L0oooo
00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo
00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo
00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo
00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo
00<0001o0?ooo`3oool01P3oool<00000080oooo0P0000000`0007l0oooo0?ooo`090?ooo`800000
1P3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3o
ool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3o
ool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3o
ool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3o
ool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`060?ooo`X000000`3o
ool4000000X0oooo00<000000?ooo`3oool01P3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool0
0`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool0
0`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool0
0`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool0
0`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool0
0`0007l0oooo0?ooo`060?ooo`T000000`3oool6000000L0oooo0P0000090?ooo`030000O`3oool0
oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`00
07l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`00
07l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`00
07l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`00
07l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo00L0oooo1`0000030?ooo`L000001P3oool00`00
0000oooo0?ooo`090?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240
oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0
oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0
oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0
oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo00P0
oooo0P0000000`3oool00000000000030?ooo`L000001@3oool2000000`0oooo00<0001o0?ooo`3o
ool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000
O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000
O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000
O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000
O`3oool0oooo0200oooo00<0001o0?ooo`3oool0403oool6000000040000O`3oool0oooo0?ooo`80
00003P3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000
O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000
O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000
O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000
O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0A0?ooo`H00000
00<0oooo000000000000403oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?oo
o`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?oo
o`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?oo
o`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?oo
o`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool06`3oool2000000<0oooo00@0
001o0?ooo`00000000002P3oool200000080oooo2@00000B0?ooo`030000O`3oool0oooo00h0oooo
000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?oo
o`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?oo
o`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?oo
o`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?oo
o`0J0?ooo`@000000P3oool00`0007l0000000000002000000@0oooo0P0000020?ooo``0000000<0
001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q
0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F
0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:
0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`06
0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool06@3oool6000000030?ooo`000000
000000@000000P3oool@000000030?ooo`0007l0oooo0180oooo00<0001o0?ooo`3oool03P3oool0
02L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo
0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo
00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo
00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo
01T0oooo7`0000000`3oool0001o0?ooo`0B0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`03
0000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`03
0000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`03
0000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`03
0000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0J0?oooad0
000000<0oooo0000000000004`3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo
0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo
0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo
0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo
0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool04P3oool2000000H0oooo
1`0000000`3oool000000000000G00000180oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<0
00000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0
001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0
001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0
001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0140oooo1000
00040?oooaH0000000<0oooo0000000000002P00000A0?ooo`030000O`3oool0oooo00h0oooo000W
0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P
0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=
0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`08
0?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0@
0?ooo`H000000P3ooolT00000140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?oo
o`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?oo
o`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?oo
o`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?oo
o`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0100oooo1P0000000`3o
ool000000000000R00000180oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3o
ool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3o
ool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3o
ool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3o
ool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0140oooo1@0000000`3oool0
00000000000M000000030?ooo`0000000000008000004P3oool00`0007l0oooo0?ooo`0>0?ooo`00
9`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0
803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool0
3@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0
203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0
4@3ooolF000000030?ooo`000000000000D000000P3oool2000000<0oooo1P00000A0?ooo`030000
O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3o
ool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3o
ool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203o
ool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3o
ool00`0007l0oooo0?ooo`080?ooo`8000001`3oool:000000030?ooo`000000000001@000000`3o
ool800000100oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool0
0`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool0
0`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool0
0`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool0
0`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo00L0oooo100000000`3oool0000000000002
0?ooo`/000000`3ooolE000000030?ooo`000000000000L00000403oool00`0007l0oooo0?ooo`0>
0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?oo
o`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?oo
o`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?oo
o`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?oo
o`3oool01P3oool?000000030?ooo`0000000000008000001@3ooolM00000140oooo00<0001o0?oo
o`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`03
0000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`03
0000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`03
0000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`03
0000O`3oool0oooo00H0oooo3`0000020?ooo`800000103oool2000000030?ooo`00000000000080
oooo5`00000B0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo
00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo
00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo
00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo
00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`070?ooo`d00000203oool4000000050000
O`3oool000000000003oool05P00000B0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000
003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000
O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000
O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000
O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`080?ooo`L00000
0P3oool2000000P0oooo1`0000030?oooaH000004P3oool00`0007l0oooo0?ooo`0>0?ooo`009`3o
ool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803o
ool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3o
ool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203o
ool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool02@3o
ool2000000030?ooo`000000000000d0oooo1P0000030?oooaH000004`3oool00`0007l0oooo0?oo
o`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o
0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o
0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o
0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o
0?ooo`3oool00`3oool2000000h0oooo0P0000070?ooo`@0000000<0001o0?ooo`3oool04`000000
103oool000000000000007lC0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0
oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0
oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0
oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0
oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`020?ooo`@00000303oool4
000000030?ooo`000000000000<0oooo0`0000000`3oool0001o0?ooo`0;000000030?ooo`000000
000000L000000P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?oo
o`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?oo
o`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?oo
o`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?oo
o`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool0100007l0oooo0?ooo`3oool8
000000P0oooo2`0000040?ooo`030000O`000000000000T0000000<0oooo00000000000020000002
0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?oo
o`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?oo
o`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?oo
o`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?oo
o`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`040000O`3oool0oooo0?ooo`T000001`3oool:
000000D0oooo00<0001o0000000000002@0000000`3oool0000000000007000000<0oooo00<0001o
0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?oo
o`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?oo
o`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?oo
o`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?oo
o`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool00P3oool9000000L0oooo2@0000050?oo
o`030000O`3oool0000000T000000P3oool4000000030?ooo`000000000000@0oooo00<0001o0?oo
o`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`03
0000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`03
0000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`03
0000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`03
0000O`3oool0oooo03T0oooo00@0001o0?ooo`3oool0oooo2P0000080?ooo`P000001@3oool00`00
07l0oooo0?ooo`09000000<0oooo0P0000080?ooo`030000O`3oool0oooo0140oooo00<0001o0?oo
o`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`03
0000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`03
0000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`03
0000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0f0?ooo`80
000000@0oooo0000O`3oool0oooo2P0000090?ooo`80000000<0oooo000000000000100000040?oo
o`030000O`3oool0oooo00P000003P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo
00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0
oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0
oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0
oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0=@3oool400000003
0000O`3oool0oooo00T00000203oool2000000<0oooo1P0000040?ooo`040000O`3oool0oooo0?oo
o`H000003`3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`03
0000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`03
0000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`03
0000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`03
0000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0=03oool6000000<0oooo1`0000060?ooo`<0
00001@3oool7000000<0oooo00D0001o0?ooo`00000000000?ooo`0200000180oooo00<0001o0?oo
o`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`03
0000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`03
0000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`03
0000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`03
0000O`3oool0oooo03@0oooo1`0000030?ooo`L00000103oool4000000<0oooo2@0000030?ooo`03
0000O`000000000000800000503oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0
oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo
0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo
0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo
0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0:`3oool2000000P0oooo
1P0000050?ooo`L0000000<0oooo000000000000100000000`3oool000000000000900000080oooo
1P00000C0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<0
00000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0
001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0
001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0
001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0Z0?ooo`@00000203oool5000000D0oooo6P00
00020?ooo`H000004`3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W
0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P
0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=
0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`08
0?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0:@3oool6000000L0oooo1P000005
0?ooo``0000000<0oooo0000000000002@0000030?ooo`030000O`000000000000800000503oool0
0`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo
0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo
01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo
00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo
00H0oooo00<0001o0?ooo`3oool0:@3oool6000000L0oooo1P0000040?ooo``000000`3oool50000
0080oooo0P0000040?ooo`040000O`3oool00000000001D0oooo00<0001o0?ooo`3oool04@3oool0
0`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo
03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo
0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo
00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo
02H0oooo0P0000020?ooo`@000001@3oool200000080oooo1P0000030?ooo``00000103oool40000
0080oooo0P0000040?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`00
07l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0
oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140
oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0
oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo02D0
oooo100000020?ooo`8000001@3oool400000080oooo1P0000020?ooo`d000000P3oool:000000<0
oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0
oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo
0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo
0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo
0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0903oool6000000L0oooo
1P0000000`3oool000000000000600000080oooo0P0000020?ooo`L000000P3oool;00000080oooo
00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo
000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?oo
o`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?oo
o`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?oo
o`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0903oool6000000L0oooo1P00
00020?ooo`D000001`3ooolE00000080oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?oo
o`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o
0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o
0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o
0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o
0?ooo`3oool09@3oool4000000T0oooo100000030?ooo`040000O`0000000000000000@0oooo0P00
00020?oooa@000000`3oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0
001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i
0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A
0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`08
0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0V
0?ooo`8000002P3oool3000000@0oooo00<0001o0000003oool0103oool400000080oooo3P000002
0?ooo`800000103oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o
0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?oo
o`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?oo
o`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?oo
o`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0a0?oo
o`@000000`3oool2000000D0oooo500000090?ooo`030000O`3oool0oooo01H0oooo00<0001o0?oo
o`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`03
0000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`03
0000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`03
0000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`03
0000O`3oool0oooo0300oooo1P000000103oool000000?ooo`0007l50?oooa400000303oool00`00
07l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0
oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200
oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0
oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0
oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0N0?ooo`8000003`3oool800000080
oooo00<0001o0?ooo`3oool00P3ooolA000000d0oooo00<0001o0?ooo`3oool05P3oool00`0007l0
oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo
00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo
00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo
00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo
00<0001o0?ooo`3oool0703oool5000000d0oooo1`0000040?ooo`040000O`3oool0oooo0?ooo`H0
00000P3oool4000000030?ooo`0000000000008000003P3oool00`0007l0oooo0?ooo`0F0?ooo`03
0000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool0
8@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool0
5P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool0
2P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0
1P3oool00`0007l0oooo0?ooo`0K0?ooo`T000002P3oool6000000D0oooo00<0001o0?ooo`3oool0
1P0000040?ooo`8000000`3oool2000000l0oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo
0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0
001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0
001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0
001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0
001o0?ooo`3oool06P3oool;000000X0oooo100000060?ooo`030000O`3oool0000000H000006P3o
ool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3o
ool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0
oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0
oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0
oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0J0?ooo``000002@3oool3
000000L0oooo00<0001o0?ooo`0000001@00000K0?ooo`030000O`3oool0oooo01H0oooo00<0001o
0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?oo
o`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?oo
o`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?oo
o`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?oo
o`030000O`3oool0oooo01/0oooo300000030?ooo`80000000<0oooo0000000000001P3oool20000
0080oooo0P0000000`3oool0000000000002000001`0oooo00<0001o0?ooo`3oool05P3oool00`00
07l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240
oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0
oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0
oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0
oooo00<0001o0?ooo`3oool0703oool300000080oooo1`0000000`3oool0000000000002000000L0
oooo2`00000M0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0
oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo
00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo
00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo
00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo0180oooo
0P0000080?ooo`@000000P3oool<000000030?ooo`00000000000080oooo3000000M0?ooo`030000
O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3o
ool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803o
ool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3o
ool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203o
ool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo0140oooo100000060?ooo`H000000P3o
ool4000000030?ooo`000000000000P0000000<0oooo0000000000002`00000L0?ooo`030000O`3o
ool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool0
0`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool0
0`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool0
0`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool0
0`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo0100oooo200000030?ooo`H000000`3oool4
000000030?ooo`000000000000P0000000<0oooo0000000000002P00000L0?ooo`030000O`3oool0
oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`00
0000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`00
07l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`00
07l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`00
07l0oooo0?ooo`060?ooo`030000O`3oool0oooo00`0oooo0P0000020?ooo`T000000`3oool40000
00D0oooo00<000000?ooo`3oool00P3oool900000080oooo0P0000000`3oool00000000000050000
01d0oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo
00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0
oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0
oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0
oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool02@3oool600000080
oooo2`0000000`3oool00000000000040?ooo`8000000P3oool200000080oooo1`0000070?ooo`D0
00007P3oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3o
ool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000
O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000
O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000
O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`080?ooo`P00000
0P3oool;000000@0oooo0P0000030?ooo`d0000000<0oooo0000000000000`3oool00`0007l0oooo
0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo
0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0
001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0
001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0
001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`040000O`3oool0oooo0?ooo`800000
103oool9000000<0oooo2`0000000`3oool00000000000040?oooa8000000P3oool00`0007l0oooo
0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo
0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0
001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0
001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0
001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo00@000000`3o
ool8000000<0oooo3@0000060?oooa<0000000<0oooo0000O`3oool08@3oool00`0007l0oooo0?oo
o`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<00000
0?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o
0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o
0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o
0?ooo`3oool01P3oool00`0007l0oooo00000005000000<0oooo1P0000040?ooo`H0000000<0oooo
000000000000100000070?ooo`P0000000<0oooo0000000000001`0000000`3oool0001o0?ooo`0Q
0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>
0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?oo
o`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?oo
o`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?oo
o`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0000000D000000`3oool30000
00P0oooo100000020?ooo`H00000203oool5000000@0oooo1`0000020?ooo`030000O`3oool0oooo
0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo
00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0
oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0
oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0
oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0100000030?ooo`@0
00002@3oool200000080oooo1P0000020?ooo`80000000<0oooo0000000000000P3oool600000004
0?ooo`00000000000?ooo`80000000<0oooo0000000000000`3oool00`0007l0oooo0?ooo`0P0?oo
o`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?oo
o`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3o
ool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3o
ool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3o
ool0203oool00`0007l0oooo0?ooo`060?ooo`040000O`3oool0oooo0?ooo`8000000`3oool60000
00/0oooo1P0000020?ooo`L0000000<0oooo000000000000200000080?ooo`030000O`3oool0oooo
0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo
00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0
oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0
oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0
oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool01P3oool6000000X0
oooo1P0000020?ooo`T0000000<0oooo000000000000200000070?ooo`030000O`3oool0oooo0200
oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0
oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo
0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo
0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo
0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool01`3oool4000000P0oooo
2@0000020?ooo`T000000P3oool2000000030?ooo`000000000000@000001`3oool00`0007l0oooo
0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo
0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0
001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0
001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0
001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo00D0oooo0P00
00000`3oool00000000000080?ooo`@0000000<0oooo0000000000000P0000040?ooo`P000001P3o
ool4000000P0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3o
ool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool0
0`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool0
0`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool0
0`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool0
0`0007l0oooo0?ooo`040?ooo`@000002@3oool6000000030?ooo`000000000000H0oooo20000006
0?ooo`8000002@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o
0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?oo
o`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?oo
o`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?oo
o`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?oo
o`030000O`3oool0oooo0080000000<0oooo000000000000100000050?ooo`T000002@3oool90000
0100oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo
0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0
oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0
oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0
oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01@3oool200000003
0?ooo`000000000000P00000103oool4000000030?ooo`0000000000008000002@3oool:00000100
oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140
oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo
0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo
0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo
0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0103oool=000000@0oooo
1P0000000`3oool00000000000020?ooo`8000001P3oool900000140oooo00<0001o0?ooo`3oool0
803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool0
3P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3o
ool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3o
ool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3o
ool0oooo00P0oooo00<0001o0?ooo`3oool00`3oool=000000@0oooo1`0000040?ooo`@000001P3o
ool4000000030?ooo`00000000000180oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?oo
o`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<00000
0?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o
0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o
0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o
0?ooo`3oool00P3oool>000000<0oooo00<000000?ooo`3oool01P0000020?ooo`H000001P3oool2
000001H0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0
oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`00
07l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`00
07l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`00
07l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool00P3oool60000
00030?ooo`000000000000D0000000<0oooo0000000000000`3oool600000080oooo1P00000N0?oo
o`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?oo
o`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3o
ool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3o
ool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3o
ool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00<0oooo100000040?ooo`@0
000000<0oooo0000003oool00P3oool2000000030?ooo`000000000000800000103oool4000001l0
oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140
oooo00<0001o0?ooo`3oool03P3oool001T0oooo00H0_g@000000000000000000018W03Oool80?oo
o`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?oo
o`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?oo
o`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?oo
o`030000O`3oool0oooo00@0oooo0P0000000`0007l0oooo0?ooo`030?ooo`@000000`3oool40000
00030?ooo`000000000000H0oooo0P00000P0?ooo`030000O`3oool0oooo0200oooo00<0001o0?oo
o`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000J0?ooo`03
0?oOW0180000M;oo00X0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`03
0000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`03
0000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`03
0000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo1P00000Z0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3o
ool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000J0?ooo`030?oO
W0180000M;oo00X0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000
O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000
O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000
O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`020?ooo`800000
1@3oool6000002X0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000
O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool001X0oooo00<0omnL04P0001d_ol02P3o
ool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803o
ool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3o
ool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203o
ool00`0007l0oooo0?ooo`060?ooo`050000O`3oool0oooo0?ooo`000000203oool4000002/0oooo
00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo
00<0001o0?ooo`3oool03P3oool001X0oooo00<0omnL04P0001d_ol02P3oool00`000000oooo0?oo
o`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?oo
o`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?oo
o`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01@3oool2000000030?ooo`00
07l0oooo00L0oooo00@0001o0?ooo`00000000002P3oool2000002`0oooo00<0001o0?ooo`3oool0
803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool0
3P3oool001X0oooo00<0omnL04P0001d_ol02P3oool400000=P0001o0P0000000`0007l000000000
0002000000T0001o0P00000A0000O`800000P@0007l70?ooo`006P3oool00`3ogi`0B00007Boo`0:
0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P
0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=
0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`09
000000L0oooo00<000000000O`3oool0403oool4000002H0oooo00<0001o0?ooo`3oool0803oool0
0`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool0
01X0oooo00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo
03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo
0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo
00P0oooo00<0001o0?ooo`0000002@0000050?ooo`80000000<0oooo0000O`3oool03`3oool60000
02D0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo
0140oooo00<0001o0?ooo`3oool03P3oool001X0oooo00<0omnL04P0001d_ol02P3oool00`000000
oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0
oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0
oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`00000020000000100007l0
oooo0?ooo`3oool2000000<0oooo00<0001o0?ooo`3oool03P3oool6000002D0oooo00<0001o0?oo
o`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?oo
o`3oool03P3oool001X0oooo00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`0Q0?ooo`03
0000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`03
0000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`03
0000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0100000001@3oool000000000003oool0001o
0080oooo00@000000?ooo`00000000000P3oool00`0007l0oooo0?ooo`0?0?ooo`@000009P3oool0
0`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool0
0`0007l0oooo0?ooo`0>0?ooo`00603oool01@3ooml0W4P0000000000000M;oo00X0oooo00<00000
0?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o
0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o
0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`040000O`3oool0oooo0?ooo`8000001@3o
ool0100007l000000000003oool4000000030?ooo`0007l0oooo00L0oooo0P0000080?ooo`800000
9`3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool0
4@3oool00`0007l0oooo0?ooo`0>0?ooo`006P3oool00`3o_g@0000007Boo`0:0?ooo`030000003o
ool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3o
ool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3o
ool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`070?ooo`8000000P3o
ool6000000030000O`000000000000D0oooo1000000`0?ooo`030000O`3oool0oooo0200oooo00<0
001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W
0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P
0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=
0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`06
0?ooo`030000003oool0001o0080oooo2P0000030?ooo`H00000;`3oool00`0007l0oooo0?ooo`0P
0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>
0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?oo
o`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?oo
o`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?oo
o`3oool0103oool200000080oooo00@0001o0?ooo`3oool0oooo2P0000020?ooo`H00000;`3oool0
0`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool0
0`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo
03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo
0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo
00P0oooo00<0001o0?ooo`3oool00P3oool2000000@0oooo00<0001o0?ooo`3oool00P3oool20000
00030?ooo`000000000000@000000`3oool400000300oooo00<0001o0?ooo`3oool0803oool00`00
07l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0
oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200
oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0
oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`050000O`3oool0oooo0?oo
o`0000001P3oool00`0007l0oooo0?ooo`02000000@0oooo100000050?ooo`800000<@3oool00`00
07l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`00
07l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0
oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140
oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`070?ooo`80000000<0oooo0000O`3o
ool01@3oool200000080oooo00@0001o0?ooo`00000000001`3oool00`0007l0oooo000000030000
00<0oooo00<0001o000000000000>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo
01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000
oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0
oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0
oooo0?ooo`060?ooo`@0000000<0001o0?ooo`3oool00`3oool4000000030?ooo`0007l0000000@0
oooo0P0000030?ooo`030000O`000000000000@000000P3oool00`0007l0oooo0?ooo`0i0?ooo`03
0000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`03
0000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0
>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool0
4@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00D0oooo1P0000040?ooo`L00000
103oool400000080oooo00<0001o000000000000100000020?ooo`030000O`3oool0oooo03T0oooo
00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo
00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?oo
o`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?oo
o`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool01@3oool6000000@0oooo1P00
0000100007l0oooo0?ooo`3oool6000000030?ooo`0007l0oooo00@000000`3oool00`0007l0oooo
0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo
0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`<00000aP0007l4000000H0001o1000
00050000O`H00000100007l2000009X0001o1`3oool002L0oooo00<000000?ooo`3oool08@3oool0
0`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool0
0`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool00P3oool2
000000<0oooo0P0000000`3oool0001o0?ooo`030?ooo`@000000P3oool00`0007l0oooo0?ooo`02
0?ooo`@000000P3oool00`0007l00000000000060?ooo`030000O`3oool0oooo03T0oooo00<0001o
0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o
0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?oo
o`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?oo
o`030000O`3oool0oooo00d0oooo00@0001o0?ooo`3oool0oooo100000050?ooo`050000O`3oool0
oooo0?ooo`000000103oool2000000030000O`3oool0oooo00<0oooo0P0000030?ooo`@000001@3o
ool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3o
ool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0
oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0
oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`090?ooo`8000000P3oool0
0`0007l0oooo0?ooo`06000000@0oooo00@0001o0?ooo`0000000000103oool4000000T0oooo1P00
00040?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?oo
o`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<00000
0?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o
0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00P0oooo10000000
103oool0001o0?ooo`3oool6000000@0oooo0P0000050?ooo`H000001`3oool7000000@0oooo00<0
001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0
001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q
0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F
0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool01`3oool6000000040000O`3oool0
oooo0?ooo`@00000103oool00`000000001o0?ooo`020?ooo`80000000<0oooo0000000000001000
00060?ooo`L000001@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0
001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W
0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P
0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`07
0?ooo`H0000000<0001o0?ooo`3oool00P3oool2000000<0oooo100000020?ooo`@0000000<0oooo
0000000000000P0000070?ooo`H000001P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0
oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0
oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`00
07l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`00
07l0oooo0?ooo`080?ooo`@0000000<0oooo0000O`3oool01`3oool00`000000oooo000000090000
00040?ooo`00000000000000O`P0oooo100000070?ooo`030000O`3oool0oooo03T0oooo00<0001o
0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o
0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?oo
o`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?oo
o`030000O`3oool0oooo00T0oooo0P0000020?ooo`040000O`3oool00000000000<0oooo0P000000
0`3oool0000000000009000000<0oooo00@0001o0?ooo`00000000001P3oool2000000P0oooo00<0
001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0
001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q
0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F
0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool01P3oool2000000D0oooo00<0001o
000000000000100000030?ooo`X00000103oool00`0007l0000000000002000000H0oooo00<0001o
0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o
0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?oo
o`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?oo
o`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`050?oo
o`@00000103oool6000000D0oooo200000050?ooo`H000001@3oool00`0007l0oooo0?ooo`060?oo
o`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?oo
o`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3o
ool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3o
ool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00@0oooo1P0000030?ooo`H0
00001P3oool6000000D0oooo1`0000050?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3o
ool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3o
ool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000
O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000
O`3oool0oooo0140oooo00<0001o0?ooo`3oool0103oool6000000<0oooo1@0000080?ooo`H00000
0`3oool7000000H0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000
O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000
O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3o
ool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3o
ool00`0007l0oooo0?ooo`050?ooo`@000000`3oool00`000000001o0?ooo`02000000T0oooo1P00
00030?ooo`H000001`3oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0
001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0
001o0?ooo`3oool03P3oool002L0oooo0`00002f0000O`8000000P0007l2000000P0001o0P000005
0000O`@000001@0007l400000:L0001o1`3oool002L0oooo00<000000?ooo`3oool08@3oool00`00
07l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`00
07l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00P0oooo0P0000030?ooo`030000O`3oool0oooo
00<0oooo100000030?ooo`040000O`3oool00000000000L0oooo0P0000090?ooo`030000O`3oool0
oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0
oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`00
0000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`00
07l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool01`3oool00`00
0000oooo0?ooo`030?ooo`030000O`3oool0oooo0080oooo1P0000020?ooo`030000O`3oool0oooo
00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo
03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo
0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0
oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0
oooo0?ooo`0A0?ooo`030000O`3oool0oooo00D0oooo0P0000060?ooo`030000O`3oool0oooo0080
oooo1P0000020?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0
oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0
oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo
00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo
00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00<0oooo
0P0000080?ooo`030000O`3oool0oooo00<0oooo100000030?ooo`030000O`3oool0oooo00P0oooo
00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo
00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo
00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?oo
o`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?oo
o`0A0?ooo`030000O`3oool0oooo0080oooo00<000000?ooo`3oool0203oool00`0007l0oooo0?oo
o`040?ooo`800000103oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0
001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0
001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W
0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P
0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`02
000000/0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0
oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0
oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0
oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`00
07l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`00
07l000000000000<0?ooo`800000303oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo
00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo
0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo
00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0
oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool01@3oool2000000X0
oooo00<000000?ooo`3oool02`3oool4000000/0oooo00<0001o0?ooo`3oool0203oool00`0007l0
oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0
oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0
oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo
00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo00@0oooo
100000070?ooo`80000000<0001o0?ooo`3oool02P3oool6000000X0oooo00<0001o0?ooo`3oool0
203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0
>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool0
4@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3o
ool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3o
ool0oooo00<0oooo1P0000040?ooo`8000000P3oool00`0007l0oooo0?ooo`02000000P0oooo1P00
000:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?oo
o`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?oo
o`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<00000
0?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o
0?ooo`3oool05P3oool00`0007l0oooo0?ooo`030?ooo`H000000`3oool00`000000oooo0?ooo`02
0?ooo`030000O`3oool0000000<00000203oool4000000/0oooo00<0001o0?ooo`3oool0203oool0
0`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool0
0`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool0
0`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo
03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo
00@0oooo100000020?ooo`8000001@3oool00`0007l0000000000004000000P0oooo0P00000<0?oo
o`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?oo
o`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?oo
o`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo0`00002@0000O`80
00003`0007l2000000030000O`000000000000H0001o2000003;0000O`L0oooo000W0?ooo`030000
003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000
O`3oool0oooo00`0oooo100000060?ooo`030000O`3oool0oooo00L0oooo00<000000?ooo`3oool0
0`3oool:000000X0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000
O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000
O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000
O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3o
ool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo00/0oooo1P0000050?ooo`030000O`3o
ool0oooo00D0oooo0P0000050?ooo`X000002`3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3o
ool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3o
ool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3o
ool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool0
0`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool02`3oool6
000000D0oooo00<0001o0?ooo`3oool00`3oool2000000H0oooo2@0000030?ooo`800000203oool0
0`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool0
0`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool0
0`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool0
02L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo
0200oooo00<0001o0?ooo`3oool0303oool4000000H0oooo00<0001o0?ooo`3oool00P3oool00`00
0000oooo0?ooo`060?ooo`P000000`3oool4000000L0oooo00<0001o0?ooo`3oool02P3oool00`00
07l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`00
07l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`00
07l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240
oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo00d0
oooo0P0000070?ooo`030000O`3oool0oooo008000002P3oool6000000<0oooo1P0000060?ooo`03
0000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`03
0000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`03
0000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`00
9`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0
803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0000000d0oooo0P0000020?ooo`040000
O`3oool0oooo0?ooo`H000001P3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0
oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0
oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140
oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo
0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool200000180oooo
00<0001o0?ooo`3oool00P3oool4000000L0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo
0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo
0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo
0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0
001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01@0oooo0P00
00000`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00<0oooo0P0000080?ooo`030000O`3o
ool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3o
ool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3o
ool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool0
0`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool0
0`0007l0oooo0?ooo`0C0?ooo`040000003oool0oooo0000Oa<0oooo00<0001o0?ooo`3oool03@3o
ool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203o
ool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803o
ool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3o
ool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0
oooo0200oooo00<0001o0?ooo`3oool04@3oool2000000<0oooo0P00000:0?ooo`8000001P3oool0
0`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool0
0`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool0
0`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool0
0`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo
03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0@0?ooo`030000003oool0oooo
0080oooo100000080?ooo`@000001@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo
00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo
00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo
01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool300000940
001o0P0000040000O`H000001P0007l600000=H0001o1`3oool002L0oooo00<000000?ooo`3oool0
8@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool0
303oool2000000H0oooo1P0000060?ooo`H00000103oool00`0007l0oooo0?ooo`0=0?ooo`030000
O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000
O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000
O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3o
ool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool04`3o
ool2000000/0oooo00<0001o0?ooo`3oool02`3oool00`000000oooo0?ooo`070?ooo`@00000203o
ool4000000D0oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3o
ool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3o
ool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3o
ool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool0
0`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0180oooo1000000:0?ooo`030000O`3oool0
oooo00T0oooo0P00000;0?ooo`8000002P3oool2000000H0oooo00<0001o0?ooo`3oool03@3oool0
0`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool0
0`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool0
0`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool0
02L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo
0140oooo1P0000090?ooo`030000O`3oool0oooo00P0oooo00<000000?ooo`3oool02`3oool00`00
07l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`00
07l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`00
07l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`00
07l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240
oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0A0?ooo`H000002@3oool00`0007l0
oooo0?ooo`060?ooo`8000003P3oool0100007l0oooo0?ooo`3oool2000000h0oooo00<0001o0?oo
o`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?oo
o`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?oo
o`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?oo
o`3oool03P3oool000/0oooo00H0_g@000000000000000000018W03Oool20?ooo`060?noM0000000
M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`080?ooo`030000003oool0oooo0240oooo
00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0B0?ooo`@000002P3oool00`0007l0oooo
0?ooo`040?ooo`800000403oool00`0007l0oooo0?ooo`04000000d0oooo00<0001o0?ooo`3oool0
3@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0
203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0
803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool0
3P3oool000`0oooo00<0omnL04P0001d_ol0103oool01`3o_g@0000007Boo`3oool0oooO09a8M02o
ool00P3oool00`2oM000M;oo0?ooo`080?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3o
ool0>@3oool00`0007l0oooo0?ooo`0C0?ooo`8000002`3oool00`0007l0oooo0?ooo`030?ooo`03
0000003oool0oooo0100oooo00<0001o0?ooo`0000001@00000<0?ooo`030000O`3oool0oooo00d0
oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0
oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200
oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0
oooo000<0?ooo`030?oOW0180000M;oo00d0oooo00<0oooO09a8M02oool0203oool00`000000oooo
0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool0100007l0oooo
0?ooo`3oool2000001<0oooo00<0001o0?ooo`0000001@00000<0?ooo`030000O`3oool0oooo00d0
oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0
oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200
oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0
oooo000<0?ooo`030?oOW0180000M;oo00h0oooo00<0_g@007Boo`3oool01`3oool00`000000oooo
0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool0100007l0oooo
0?ooo`0000070?ooo`800000303oool00`0007l0oooo0?ooo`04000000d0oooo00<0001o0?ooo`3o
ool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3o
ool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3o
ool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3o
ool03P3oool000`0oooo00<0omnL04P0001d_ol03@3oool00`3ooml0W4P004RLg`080?ooo`030000
003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000
O`000000000000L0oooo100000060?ooo`8000000`3oool0100007l0oooo0?ooo`3oool2000000h0
oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0
oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0
oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140
oooo00<0001o0?ooo`3oool03P3oool000`0oooo00<0omnL04P0001d_ol03@3oool00`3OW4P00000
07Boo`080?ooo`@00000OP0007l2000000P0001o1P0000040000O`@00000j00007l70?ooo`00303o
ool00`3ogi`0B00007Boo`090?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`080?ooo`03
0000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0N0?ooo`03
0000003oool0001o00P0oooo1P0000030?ooo`H0000000<0oooo0000O`3oool04P3oool00`0007l0
oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0
oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0
oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0
oooo0?ooo`0>0?ooo`00303oool00`3ogi`0B00007Boo`090?ooo`060=nLB0000000000000000000
04P0W=oo2@3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?oo
o`3oool0703oool200000080oooo00<0001o0?ooo`3oool01`3oool4000000@0oooo1P0000000`3o
ool0001o0?ooo`0B0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`00
07l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`00
07l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`00
07l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000<0?ooo`030?oOW0180000M;oo00T0
oooo00@0oono07@000000000B9cO2`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo
03T0oooo00<0001o0?ooo`3oool06`3oool00`000000oooo0?ooo`020?ooo`030000O`3oool0oooo
00P0oooo0P0000060?ooo`@000000P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo
00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo
00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo
0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo
00h0oooo000<0?ooo`030?oOW0180000M;oo00X0oooo00<0_g@007Boo`3oool02`3oool00`000000
oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool06@3oool2000000D0
oooo00<0001o0?ooo`3oool00P3oool2000000d0oooo0P0000030?ooo`030000O`3oool0oooo0140
oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0
oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0
oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140
oooo00<0001o0?ooo`3oool03P3oool000X0oooo00D0oooO09a800000000000007Boo`0:0?ooo`06
0?oOW018000000000000000004P0W=oo203oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0
oooo03T0oooo00<0001o0?ooo`3oool05`3oool2000000L0oooo00@0001o0?ooo`3oool0oooo1000
000A0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?oo
o`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?oo
o`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?oo
o`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool000`0oooo00<0okmd
0000001d_ol02`3oool01@2oM0000000000000000000B9cO00P0oooo00<000000?ooo`3oool08@3o
ool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo01H0oooo00<000000?ooo`3oool01`3o
ool00`0007l0oooo0?ooo`0600000100oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?oo
o`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?oo
o`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?oo
o`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?oo
o`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o
0?ooo`3oool0503oool2000000X0oooo00<0001o0?ooo`3oool01P00000@0?ooo`030000O`3oool0
oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0
oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0
oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0
oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`00
07l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo01<0oooo00<000000?ooo`3oool02P3oool01000
07l0oooo0?ooo`3oool400000140oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=
0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`08
0?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P
0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>
0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?oo
o`3oool04@3oool2000000d0oooo00<0001o0?ooo`3oool00P3oool200000180oooo00<0001o0?oo
o`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?oo
o`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?oo
o`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?oo
o`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`03
0000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0303oool2000000030?ooo`000000000000l0
oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0
oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0
oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200
oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0
oooo000W0?ooo`<00000J`0007l4000000/0001o0P0000040000O`800000o@0007l70?ooo`009`3o
ool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool02P3o
ool6000000T0oooo100000020?ooo`@000005P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3o
ool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3o
ool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3o
ool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3o
ool0oooo00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0>@3oool0
0`0007l0oooo0?ooo`0:0?ooo`H00000203oool<000001D0oooo00<0001o0?ooo`3oool04@3oool0
0`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool0
0`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool0
0`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool0
0`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo
03T0oooo00<0001o0?ooo`3oool02@3oool6000000T0oooo3000000E0?ooo`030000O`3oool0oooo
0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo
00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo
03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo
0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0
oooo0?ooo`0U0?ooo`8000004P3oool00`0007l0oooo0?ooo`070?ooo`8000000`3oool2000000/0
oooo100000020?ooo`@000005P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0
oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0
oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200
oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0
oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0903oool400000140oooo
00<0001o0?ooo`3oool01P3oool00`000000oooo0?ooo`0A0?ooo`800000103oool2000001L0oooo
00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo
00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo
00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo
00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?oo
o`0Q0?ooo`030000O`3oool0oooo02<0oooo1P00000@0?ooo`030000O`3oool0oooo00@0oooo0P00
000J0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?oo
o`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?oo
o`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?oo
o`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?oo
o`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo02<0oooo1P00000@
0?ooo`030000O`3oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`0007l0oooo0?ooo`0F
0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:
0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`06
0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F
0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?oo
o`3oool08@3oool00`0007l0oooo0?ooo`0T0?ooo`@000004@3oool0100007l0oooo0?ooo`3oool2
000001d0oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0
oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0
oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0
oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0
oooo00h0oooo000W0?ooo`<00000B@0007l2000001@0001o0P00003o0000Oal0001o1`3oool002L0
oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0i0?ooo`8000008@3oool00`0007l0
oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0
oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0
oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0
oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo
00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0_0?ooo`8000001`3oool400000200oooo
00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo
00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo
00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo
00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo
000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0;P3oool4000000@0oooo1`00
000O0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?oo
o`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?oo
o`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?oo
o`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?oo
o`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo02d0oooo1P000002
0?ooo`030000003oool0000000D00000103oool2000001T0oooo00<0001o0?ooo`3oool05P3oool0
0`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool0
0`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool0
0`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool0
0`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo
0240oooo00<0001o0?ooo`3oool0;@3oool8000000<0oooo100000040?ooo`@00000603oool00`00
07l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`00
07l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`00
07l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`00
07l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0
oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0^0?ooo`D000001P3oool2000000@0
oooo1P00000G0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0
oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0
oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0
oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0
oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo02l0oooo
0P0000080?ooo`030000O`3oool0oooo00<0oooo1P00000G0?ooo`030000O`3oool0oooo01H0oooo
00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo
00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo
00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo
00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?oo
o`0Q0?ooo`030000O`3oool0oooo02h0oooo0P0000090?ooo`030000O`3oool0oooo00@0oooo1000
000H0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?oo
o`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?oo
o`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?oo
o`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?oo
o`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo02d0oooo00<00000
0?ooo`3oool02@3oool00`0007l0oooo0?ooo`050?ooo`8000006@3oool00`0007l0oooo0?ooo`0F
0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:
0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`06
0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F
0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo0`00001?0000
O`800000o`0007l_0000O`L0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3o
ool0:@3oool2000000h0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`03
0000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`03
0000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`03
0000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`03
0000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool0
8@3oool00`0007l0oooo0?ooo`0X0?ooo`030000003oool0oooo00h0oooo00<0001o0?ooo`3oool0
803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool0
3@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0
203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0
803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool0
3P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0V0?ooo`8000004@3o
ool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3o
ool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203o
ool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3o
ool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3o
ool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0
oooo02@0oooo0P00000C0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool0
0`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool0
0`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool0
0`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool0
0`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo
0240oooo00<0001o0?ooo`3oool08`3oool00`000000oooo0?ooo`0C0?ooo`030000O`3oool0oooo
0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo
00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo
00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo
0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo
00h0oooo000W0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool08@3oool2000001H0
oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140
oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0
oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0
oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140
oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo
0?ooo`0O0?ooo`800000603oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo
00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo
00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo
00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo
00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool3000003h0001o0P00
00020000O`030000000007l0001o0?l0001o>`0007l70?ooo`009`3oool00`000000oooo0?ooo`0Q
0?ooo`030000O`3oool0oooo01T0oooo1@00000K0?ooo`030000O`3oool0oooo0200oooo00<0001o
0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o
0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o
0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o
0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?oo
o`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0603oool6000001/0oooo00<0001o0?oo
o`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?oo
o`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?oo
o`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?oo
o`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?oo
o`3oool03P3oool002L0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0H0?ooo`H0
00006`3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3o
ool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3o
ool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3o
ool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3o
ool04@3oool00`0007l0oooo0?ooo`0>0?ooo`005`3oool00`2oM0000000000000040000000304RL
g`3oool0oooo00H0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0G0?ooo`H00000
703oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool0
4@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0
203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0
>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool0
4@3oool00`0007l0oooo0?ooo`0>0?ooo`00603oool00`3OW4P0000000000003000000030018W03O
ool0oooo00H0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0F0?ooo`040000003o
ool0oooo0?ooo`8000007@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo
00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo
00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo
00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo
00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`006@3oool00`3OW4P0B9cO0?oo
o`020?ooo`030=nLM02Lgol0oooo00H0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?oo
o`0D0?ooo`8000008`3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0
001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0
001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0
001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0
001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`006P3oool00`3OW4P0M;oo0?ooo`0:
0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool04P3oool2000002D0oooo00<0001o
0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o
0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o
0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o
0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o
0?ooo`3oool03P3oool001/0oooo00<0_gA809cOo`3oool02@3oool00`000000oooo0?ooo`0Q0?oo
o`030000O`3oool0oooo0140oooo00<000000?ooo`3oool09@3oool00`0007l0oooo0?ooo`0P0?oo
o`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?oo
o`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?oo
o`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?oo
o`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?oo
o`006`3oool00`3ooml0W4Qd0;ooo`090?ooo`@00000<P0007l200000?l0001oB`0007l70?ooo`00
703oool00`3o_g@007Bo0?ooo`080?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0
3@3oool2000002X0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000
O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000
O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000
O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000
O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool001`0oooo00<0oono07@0B02Lgol0203o
ool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0:P3o
ool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3o
ool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203o
ool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3o
ool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3o
ool00`0007l0oooo0?ooo`0>0?ooo`00703oool00`3ookl0M00004RLg`080?ooo`030000003oool0
oooo0240oooo00<0001o0?ooo`3oool02P3oool2000002d0oooo00<0001o0?ooo`3oool0803oool0
0`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool0
0`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool0
0`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool0
0`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool0
01L0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`080?ooo`030000003oool0oooo
0240oooo00<0001o0?ooo`3oool0203oool2000002l0oooo00<0001o0?ooo`3oool0803oool00`00
07l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`00
07l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`00
07l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`00
07l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool001L0
oooo00P0oooO09a8B02Lgol0oooo0?oo_`1d0000001d0;ooo`P0oooo00<000000?ooo`3oool08@3o
ool00`0007l0oooo0?ooo`070?ooo`030000003oool0oooo02l0oooo00<0001o0?ooo`3oool0803o
ool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3o
ool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203o
ool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803o
ool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3o
ool001P0oooo00H0oooO09a80000000000000000M02oool90?ooo`030000003oool0oooo0240oooo
00<0001o0?ooo`3oool01@3oool200000380oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo
0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo
0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo
0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo
0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<0
00000?ooo`3oool08@3oool00`0007l0oooo0?ooo`030?ooo`800000=03oool00`0007l0oooo0?oo
o`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?oo
o`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?oo
o`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?oo
o`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?oo
o`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo0080oooo00<00000
0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o
0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o
0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o
0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o
0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?oo
o`030000O`3oool0oooo00800000=`3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo
01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo
00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo
00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo
01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000
oooo0?ooo`0P0?ooo`@00000303oool2000002/0oooo00<0001o0?ooo`3oool0803oool00`0007l0
oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0
oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0
oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0
oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo
00<000000?ooo`3oool07`3oool4000000`0oooo1000000Z0?ooo`030000O`3oool0oooo0200oooo
00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo
00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo
00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo
00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo
000W0?ooo`030000003oool0oooo01h0oooo1P00000:0?ooo`H00000:@3oool00`0007l0oooo0?oo
o`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?oo
o`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?oo
o`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?oo
o`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?oo
o`0>0?ooo`009`3oool00`000000oooo0?ooo`0N0?ooo`H000002P3oool6000002T0oooo00<0001o
0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o
0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o
0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o
0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o
0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool0703oool2000000030?ooo`0000000000
00800000303oool4000002X0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?oo
o`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?oo
o`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?oo
o`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?oo
o`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3o
ool06P3oool2000000@0oooo0P00000>0?ooo`800000:`3oool00`0007l0oooo0?ooo`0P0?ooo`03
0000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`03
0000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`03
0000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`03
0000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`00
9`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00D0oooo00<0001o0?ooo`3oool0
>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool0
4@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0
203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0
>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool0
4@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0G0?ooo`800000203o
ool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3o
ool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3o
ool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3o
ool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3o
ool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0
oooo01D0oooo0P00000:0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool0
0`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool0
0`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool0
0`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool0
0`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool0
02L0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0:0?ooo`030000O`3oool0oooo
03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo
0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0oooo
00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0oooo
03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0oooo
0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool04P3oool2000000d0
oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0
oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0
oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0
oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0
oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo
0?ooo`0@0?ooo`8000003`3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo
00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo
00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo
00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo
00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo
000W0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`0007l0oooo0?oo
o`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?oo
o`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?oo
o`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?oo
o`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?oo
o`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003oool0oooo00d0oooo0P00000B
0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F
0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:
0?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`06
0?ooo`030000O`3oool0oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F
0?ooo`030000O`3oool0oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?oo
o`3oool02`3oool2000001@0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?oo
o`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?oo
o`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?oo
o`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?oo
o`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?oo
o`009`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo01@0oooo00<0001o0?ooo`3o
ool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3o
ool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3o
ool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3o
ool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3o
ool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`080?ooo`800000
5`3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool0
5P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00d0oooo00<0001o0?ooo`3oool0
2P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00P0oooo00<0001o0?ooo`3oool0
1P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3oool0oooo0200oooo00<0001o0?ooo`3oool0
5P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3oool0oooo00h0oooo000W0?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool05`3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3o
ool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3o
ool0oooo00d0oooo00<0001o0?ooo`3oool02P3oool00`0007l0oooo0?ooo`080?ooo`030000O`3o
ool0oooo00P0oooo00<0001o0?ooo`3oool01P3oool00`0007l0oooo0?ooo`0i0?ooo`030000O`3o
ool0oooo0200oooo00<0001o0?ooo`3oool05P3oool00`0007l0oooo0?ooo`0A0?ooo`030000O`3o
ool0oooo00h0oooo000W0?ooo`030000003oool0oooo00D0oooo0P00000J0?ooo`030000O`3oool0
oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0
oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0
oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0
oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0
oooo0140oooo00<0001o0?ooo`3oool03P3oool002L0oooo00<000000?ooo`3oool00`3oool20000
01`0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo
01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo
00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo
00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo
01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo01`0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0
oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0
oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203oool00`0007l0
oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3oool00`0007l0
oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3oool00`0007l0
oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`02000001l0oooo00<0001o0?ooo`3oool0
>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool0
4@3oool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0
203oool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0
>@3oool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool0
4@3oool00`0007l0oooo0?ooo`0>0?ooo`009`3oool300000240oooo00<0001o0?ooo`3oool0>@3o
ool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3o
ool00`0007l0oooo0?ooo`0=0?ooo`030000O`3oool0oooo00X0oooo00<0001o0?ooo`3oool0203o
ool00`0007l0oooo0?ooo`080?ooo`030000O`3oool0oooo00H0oooo00<0001o0?ooo`3oool0>@3o
ool00`0007l0oooo0?ooo`0P0?ooo`030000O`3oool0oooo01H0oooo00<0001o0?ooo`3oool04@3o
ool00`0007l0oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0
oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0
oooo0140oooo00<0001o0?ooo`3oool03@3oool00`0007l0oooo0?ooo`0:0?ooo`030000O`3oool0
oooo00P0oooo00<0001o0?ooo`3oool0203oool00`0007l0oooo0?ooo`060?ooo`030000O`3oool0
oooo03T0oooo00<0001o0?ooo`3oool0803oool00`0007l0oooo0?ooo`0F0?ooo`030000O`3oool0
oooo0140oooo00<0001o0?ooo`3oool03P3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.3769, -0.324658, 0.0075, \
0.00404509}}]
}, Closed]],

Cell[TextData[{
  "Use PolarPlot to produce the following curves: \na) four-leaved rose, ",
  Cell[BoxData[
      \(TraditionalForm\`r = \(cos\^2\) \[Theta] - \(sin\^2\) \[Theta]\)]],
  ";  \nb) cardioid, ",
  Cell[BoxData[
      \(TraditionalForm\`r\  = \ 1\  + \ sin\[VeryThinSpace]\[Theta]\)]],
  "; \nc) lima\[CCedilla]on, ",
  Cell[BoxData[
      \(TraditionalForm\`r = 1\/2 + \ cos\[Theta]\)]],
  "."
}], "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Plotting a function of two variables", "Section"],

Cell[CellGroupData[{

Cell["Plot3D", "Subsection"],

Cell[TextData[{
  "An explicit function of two independent variables can be displayed using \
",
  ButtonBox["Plot3D",
    ButtonStyle->"RefGuideLink"],
  ".  The basic syntax is ",
  StyleBox["Plot3D[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["f,{x,xmin,xmax},{y,ymin,ymax}", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  ".  Note that the default sampling is 15 points in each directions, which \
is often not sufficient."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot3D[
        Evaluate[Re[SphericalHarmonicY[3, 1, \[Theta], \[Phi]]]], {\[Theta], 
          0, \[Pi]}, {\[Phi], 0, 2  \[Pi]}, PlotPoints \[Rule] 50, 
        PlotRange \[Rule] {\(-1\), 1}, 
        AxesLabel \[Rule] {"\<\[Theta]\>", "\<\[Phi]\>", \*"\"\<Re[\!\(Y\_\
\[Lambda]m\)]\>\""}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .81114 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% SurfaceGraphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-1.55041e-017 1.04977 -0.0679587 1.04977 [
[.02757 .24618 -10 -15.1694 ]
[.02757 .24618 0 .51811 ]
[.2187 .17096 -9.68998 -15.6875 ]
[.2187 .17096 .31002 0 ]
[.42503 .08982 -9.02633 -15.6875 ]
[.42503 .08982 .97367 0 ]
[.64842 .00203 -8.36268 -15.6875 ]
[.64842 .00203 1.63732 0 ]
[.29165 .07573 -8.72922 -15.6875 ]
[.29165 .07573 .64578 0 ]
[.70096 -0.00478 0 -10.9151 ]
[.70096 -0.00478 10 4.77238 ]
[.80399 .15298 0 -10.6462 ]
[.80399 .15298 10 5.04128 ]
[.89336 .28984 0 -10.4206 ]
[.89336 .28984 10 5.26688 ]
[.97161 .40968 0 -10.2286 ]
[.97161 .40968 10 5.45887 ]
[.91861 .21225 0 -10.5128 ]
[.91861 .21225 10.5625 5.17466 ]
[.02494 .25992 -18.3125 -4.77238 ]
[.02494 .25992 0 10.9151 ]
[.01611 .31527 -27.3125 -4.85392 ]
[.01611 .31527 0 10.8336 ]
[.00696 .37264 -10 -4.93876 ]
[.00696 .37264 0 10.7487 ]
[-0.00254 .43215 -19 -5.02709 ]
[-0.00254 .43215 0 10.6604 ]
[-0.0124 .49393 -10 -5.11915 ]
[-0.0124 .49393 0 10.5684 ]
[-0.05457 .39542 -45.125 -4.93876 ]
[-0.05457 .39542 0 10.7487 ]
[ 0 0 0 0 ]
[ 1 .81114 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.03716 .25514 m
.68874 0 L
s
.03716 .25514 m
.04196 .25962 L
s
gsave
.02757 .24618 -71 -19.1694 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.22768 .18054 m
.23217 .18532 L
s
gsave
.2187 .17096 -70.69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.43326 .10004 m
.43737 .10515 L
s
gsave
.42503 .08982 -70.0263 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.65575 .01292 m
.65941 .01836 L
s
gsave
.64842 .00203 -69.3627 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.07414 .24066 m
.07698 .24338 L
s
.11167 .22596 m
.11448 .22872 L
s
.14976 .21105 m
.15253 .21384 L
s
.18843 .19591 m
.19116 .19874 L
s
.26753 .16493 m
.27019 .16784 L
s
.30801 .14908 m
.31061 .15203 L
s
.34911 .13299 m
.35167 .13598 L
s
.39085 .11664 m
.39337 .11967 L
s
.47634 .08317 m
.47875 .08628 L
s
.52011 .06603 m
.52247 .06918 L
s
.56459 .04861 m
.5669 .0518 L
s
.60979 .03091 m
.61205 .03414 L
s
gsave
.29165 .07573 -69.7292 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(q) show
68.375 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.25 Mabswid
.68874 0 m
.96935 .42924 L
s
.68874 0 m
.68263 .00239 L
s
gsave
.70096 -0.00478 -61 -14.9151 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.79163 .1574 m
.78546 .15961 L
s
gsave
.80399 .15298 -61 -14.6462 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.88089 .29393 m
.87466 .29598 L
s
gsave
.89336 .28984 -61 -14.4206 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.95906 .4135 m
.95278 .41541 L
s
gsave
.97161 .40968 -61 -14.2286 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.125 Mabswid
.71591 .04156 m
.71223 .04297 L
s
.74208 .08159 m
.73839 .08297 L
s
.7673 .12018 m
.76361 .12153 L
s
.81512 .19332 m
.8114 .19462 L
s
.83779 .22801 m
.83407 .22928 L
s
.85971 .26153 m
.85597 .26278 L
s
.90139 .32529 m
.89764 .32649 L
s
.92123 .35563 m
.91748 .35682 L
s
.94044 .38502 m
.93668 .38619 L
s
gsave
.91861 .21225 -61 -14.5128 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(f) show
69.563 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.25 Mabswid
.03716 .25514 m
0 .48963 L
s
.03716 .25514 m
.04327 .25274 L
s
gsave
.02494 .25992 -79.3125 -8.77238 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02837 .31059 m
.0345 .30825 L
s
gsave
.01611 .31527 -88.3125 -8.85392 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.01926 .36808 m
.02541 .3658 L
s
gsave
.00696 .37264 -71 -8.93876 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.00981 .42772 m
.01599 .4255 L
s
gsave
-0.00254 .43215 -80 -9.02709 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
0 .48963 m
.0062 .48748 L
s
gsave
-0.0124 .49393 -71 -9.11915 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.03543 .26607 m
.0391 .26464 L
s
.03368 .27708 m
.03735 .27566 L
s
.03193 .28817 m
.0356 .28676 L
s
.03016 .29934 m
.03383 .29793 L
s
.02658 .32192 m
.03026 .32053 L
s
.02477 .33334 m
.02845 .33195 L
s
.02295 .34483 m
.02663 .34345 L
s
.02111 .35641 m
.0248 .35504 L
s
.0174 .37983 m
.02109 .37847 L
s
.01552 .39167 m
.01922 .39032 L
s
.01363 .4036 m
.01733 .40225 L
s
.01173 .41561 m
.01543 .41428 L
s
.00788 .43991 m
.01159 .43859 L
s
.00593 .4522 m
.00964 .45089 L
s
.00397 .46458 m
.00768 .46328 L
s
.00199 .47706 m
.00571 .47576 L
s
gsave
-0.05457 .39542 -106.125 -8.93876 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Re) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
76.250 15.813 moveto
(@) show
80.188 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Y) show
88.813 17.688 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(l) show
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(m) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
100.188 15.813 moveto
(D) show
104.125 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.25 Mabswid
.03716 .25514 m
0 .48963 L
s
0 .48963 m
.39787 .81114 L
s
.39787 .81114 m
.40529 .59895 L
s
.40529 .59895 m
.03716 .25514 L
s
.68874 0 m
.96935 .42924 L
s
.96935 .42924 m
1 .6535 L
s
1 .6535 m
.70298 .24544 L
s
.70298 .24544 m
.68874 0 L
s
.03716 .25514 m
0 .48963 L
s
0 .48963 m
.70298 .24544 L
s
.70298 .24544 m
.68874 0 L
s
.68874 0 m
.03716 .25514 L
s
.40529 .59895 m
.96935 .42924 L
s
.96935 .42924 m
1 .6535 L
s
1 .6535 m
.39787 .81114 L
s
.39787 .81114 m
.40529 .59895 L
s
0 0 m
1 0 L
1 .81114 L
0 .81114 L
closepath
clip
newpath
.5 Mabswid
.823 .766 .799 r
.39527 .69625 .40169 .70185 .41272 .6901 .40634 .68453 Metetra
.821 .769 .803 r
.40634 .68453 .41272 .6901 .42372 .67866 .41739 .67312 Metetra
.816 .771 .81 r
.41739 .67312 .42372 .67866 .4347 .66782 .42841 .6623 Metetra
.807 .773 .822 r
.42841 .6623 .4347 .66782 .44566 .65784 .4394 .65234 Metetra
.793 .775 .836 r
.4394 .65234 .44566 .65784 .4566 .64895 .45038 .64347 Metetra
.774 .776 .855 r
.45038 .64347 .4566 .64895 .46752 .64134 .46134 .63586 Metetra
.746 .775 .877 r
.46134 .63586 .46752 .64134 .47844 .63515 .47229 .62966 Metetra
.708 .77 .901 r
.47229 .62966 .47844 .63515 .48937 .63047 .48325 .62497 Metetra
.656 .759 .925 r
.48325 .62497 .48937 .63047 .50032 .62735 .49423 .62182 Metetra
.59 .739 .946 r
.49423 .62182 .50032 .62735 .51131 .62577 .50525 .62021 Metetra
.509 .708 .96 r
.50525 .62021 .51131 .62577 .52235 .62566 .51631 .62006 Metetra
.418 .666 .964 r
.51631 .62006 .52235 .62566 .53346 .62692 .52744 .62126 Metetra
.324 .617 .955 r
.52744 .62126 .53346 .62692 .54465 .62937 .53866 .62366 Metetra
.241 .569 .939 r
.53866 .62366 .54465 .62937 .55593 .63281 .54996 .62704 Metetra
.177 .529 .922 r
.54996 .62704 .55593 .63281 .56731 .637 .56136 .63115 Metetra
.14 .504 .909 r
.56136 .63115 .56731 .637 .57879 .64165 .57286 .63574 Metetra
.133 .496 .906 r
.57286 .63574 .57879 .64165 .59038 .64648 .58446 .64049 Metetra
.156 .508 .912 r
.58446 .64049 .59038 .64648 .60206 .65118 .59616 .64512 Metetra
.204 .535 .925 r
.59616 .64512 .60206 .65118 .61384 .65544 .60796 .64931 Metetra
.273 .574 .939 r
.60796 .64931 .61384 .65544 .62569 .65897 .61983 .65277 Metetra
.355 .617 .95 r
.61983 .65277 .62569 .65897 .63761 .6615 .63177 .65524 Metetra
.441 .659 .953 r
.63177 .65524 .63761 .6615 .64957 .6628 .64375 .65648 Metetra
.522 .696 .948 r
.64375 .65648 .64957 .6628 .66156 .66267 .65577 .65631 Metetra
.594 .724 .935 r
.65577 .65631 .66156 .66267 .67355 .66099 .66779 .65459 Metetra
.654 .746 .917 r
.66779 .65459 .67355 .66099 .68553 .65769 .67981 .65126 Metetra
.702 .761 .898 r
.67981 .65126 .68553 .65769 .69748 .65276 .69179 .64631 Metetra
.74 .771 .878 r
.69179 .64631 .69748 .65276 .70939 .64626 .70374 .63981 Metetra
.769 .778 .86 r
.70374 .63981 .70939 .64626 .72125 .63831 .71564 .63186 Metetra
.791 .783 .845 r
.71564 .63186 .72125 .63831 .73305 .62909 .72749 .62266 Metetra
.807 .786 .833 r
.72749 .62266 .73305 .62909 .74479 .61884 .73928 .61241 Metetra
.818 .789 .825 r
.73928 .61241 .74479 .61884 .75649 .60781 .75103 .6014 Metetra
.825 .791 .82 r
.75103 .6014 .75649 .60781 .76815 .5963 .76274 .58992 Metetra
.828 .793 .819 r
.76274 .58992 .76815 .5963 .77979 .58462 .77443 .57826 Metetra
.827 .795 .821 r
.77443 .57826 .77979 .58462 .79143 .57307 .78613 .56674 Metetra
.822 .796 .827 r
.78613 .56674 .79143 .57307 .80311 .56198 .79786 .55566 Metetra
.813 .797 .837 r
.79786 .55566 .80311 .56198 .81484 .55162 .80964 .54531 Metetra
.798 .796 .851 r
.80964 .54531 .81484 .55162 .82667 .54224 .82152 .53594 Metetra
.776 .793 .867 r
.82152 .53594 .82667 .54224 .83862 .53407 .83352 .52777 Metetra
.747 .787 .886 r
.83352 .52777 .83862 .53407 .85072 .52728 .84567 .52097 Metetra
.709 .777 .905 r
.84567 .52097 .85072 .52728 .86301 .522 .85799 .51568 Metetra
.66 .76 .924 r
.85799 .51568 .86301 .522 .8755 .51831 .87053 .51196 Metetra
.603 .736 .939 r
.87053 .51196 .8755 .51831 .88823 .51623 .88329 .50984 Metetra
.537 .705 .948 r
.88329 .50984 .88823 .51623 .9012 .51573 .8963 .50929 Metetra
.468 .669 .951 r
.8963 .50929 .9012 .51573 .91443 .51672 .90956 .51023 Metetra
.4 .63 .947 r
.90956 .51023 .91443 .51672 .92791 .51907 .92307 .51251 Metetra
.339 .593 .939 r
.92307 .51251 .92791 .51907 .94164 .52258 .93684 .51594 Metetra
.292 .561 .93 r
.93684 .51594 .94164 .52258 .95561 .52703 .95083 .52031 Metetra
.26 .539 .921 r
.95083 .52031 .95561 .52703 .9698 .53214 .96504 .52535 Metetra
.247 .527 .916 r
.96504 .52535 .9698 .53214 .98416 .53763 .97943 .53075 Metetra
.821 .767 .801 r
.3888 .6906 .39527 .69625 .40634 .68453 .39991 .67906 Metetra
.821 .772 .807 r
.39991 .67906 .40634 .68453 .41739 .67312 .41099 .66781 Metetra
.817 .778 .816 r
.41099 .66781 .41739 .67312 .42841 .6623 .42205 .65715 Metetra
.81 .784 .829 r
.42205 .65715 .42841 .6623 .4394 .65234 .43309 .64732 Metetra
.797 .79 .846 r
.43309 .64732 .4394 .65234 .45038 .64347 .4441 .63855 Metetra
.777 .795 .867 r
.4441 .63855 .45038 .64347 .46134 .63586 .45509 .63101 Metetra
.749 .797 .892 r
.45509 .63101 .46134 .63586 .47229 .62966 .46608 .62486 Metetra
.708 .795 .919 r
.46608 .62486 .47229 .62966 .48325 .62497 .47708 .62018 Metetra
.652 .785 .945 r
.47708 .62018 .48325 .62497 .49423 .62182 .48809 .61701 Metetra
.579 .764 .966 r
.48809 .61701 .49423 .62182 .50525 .62021 .49913 .61534 Metetra
.49 .729 .978 r
.49913 .61534 .50525 .62021 .51631 .62006 .51023 .6151 Metetra
.392 .681 .976 r
.51023 .6151 .51631 .62006 .52744 .62126 .52139 .61618 Metetra
.296 .627 .962 r
.52139 .61618 .52744 .62126 .53866 .62366 .53262 .61842 Metetra
.215 .575 .941 r
.53262 .61842 .53866 .62366 .54996 .62704 .54394 .62163 Metetra
.159 .534 .923 r
.54394 .62163 .54996 .62704 .56136 .63115 .55536 .62556 Metetra
.133 .511 .912 r
.55536 .62556 .56136 .63115 .57286 .63574 .56688 .62994 Metetra
.138 .506 .91 r
.56688 .62994 .57286 .63574 .58446 .64049 .5785 .63449 Metetra
.171 .519 .917 r
.5785 .63449 .58446 .64049 .59616 .64512 .59021 .63891 Metetra
.228 .546 .929 r
.59021 .63891 .59616 .64512 .60796 .64931 .60202 .64291 Metetra
.3 .581 .939 r
.60202 .64291 .60796 .64931 .61983 .65277 .61391 .64619 Metetra
.381 .619 .945 r
.61391 .64619 .61983 .65277 .63177 .65524 .62587 .6485 Metetra
.462 .654 .943 r
.62587 .6485 .63177 .65524 .64375 .65648 .63787 .64961 Metetra
.536 .685 .934 r
.63787 .64961 .64375 .65648 .65577 .65631 .64991 .64934 Metetra
.601 .709 .92 r
.64991 .64934 .65577 .65631 .66779 .65459 .66196 .64756 Metetra
.655 .728 .903 r
.66196 .64756 .66779 .65459 .67981 .65126 .67401 .6442 Metetra
.699 .742 .885 r
.67401 .6442 .67981 .65126 .69179 .64631 .68604 .63926 Metetra
.734 .754 .868 r
.68604 .63926 .69179 .64631 .70374 .63981 .69803 .6328 Metetra
.762 .762 .853 r
.69803 .6328 .70374 .63981 .71564 .63186 .70997 .62493 Metetra
.784 .77 .84 r
.70997 .62493 .71564 .63186 .72749 .62266 .72187 .61583 Metetra
.801 .776 .831 r
.72187 .61583 .72749 .62266 .73928 .61241 .73372 .60572 Metetra
.813 .783 .824 r
.73372 .60572 .73928 .61241 .75103 .6014 .74552 .59485 Metetra
.821 .788 .821 r
.74552 .59485 .75103 .6014 .76274 .58992 .75729 .58352 Metetra
.826 .794 .822 r
.75729 .58352 .76274 .58992 .77443 .57826 .76905 .57202 Metetra
.827 .8 .826 r
.76905 .57202 .77443 .57826 .78613 .56674 .78081 .56065 Metetra
.823 .805 .834 r
.78081 .56065 .78613 .56674 .79786 .55566 .7926 .54972 Metetra
.815 .81 .846 r
.7926 .54972 .79786 .55566 .80964 .54531 .80445 .53949 Metetra
.801 .813 .861 r
.80445 .53949 .80964 .54531 .82152 .53594 .81638 .53022 Metetra
.779 .814 .88 r
.81638 .53022 .82152 .53594 .83352 .52777 .82844 .52212 Metetra
.749 .811 .901 r
.82844 .52212 .83352 .52777 .84567 .52097 .84064 .51537 Metetra
.708 .803 .923 r
.84064 .51537 .84567 .52097 .85799 .51568 .85302 .51008 Metetra
.656 .786 .944 r
.85302 .51008 .85799 .51568 .87053 .51196 .8656 .50634 Metetra
.594 .761 .959 r
.8656 .50634 .87053 .51196 .88329 .50984 .8784 .50416 Metetra
.523 .727 .967 r
.8784 .50416 .88329 .50984 .8963 .50929 .89144 .50351 Metetra
.45 .687 .967 r
.89144 .50351 .8963 .50929 .90956 .51023 .90472 .50431 Metetra
.38 .643 .959 r
.90472 .50431 .90956 .51023 .92307 .51251 .91826 .50643 Metetra
.322 .603 .948 r
.91826 .50643 .92307 .51251 .93684 .51594 .93204 .50968 Metetra
.28 .57 .936 r
.93204 .50968 .93684 .51594 .95083 .52031 .94605 .51384 Metetra
.256 .547 .927 r
.94605 .51384 .95083 .52031 .96504 .52535 .96027 .51864 Metetra
.251 .536 .921 r
.96027 .51864 .96504 .52535 .97943 .53075 .97466 .5238 Metetra
.818 .767 .805 r
.38229 .68492 .3888 .6906 .39991 .67906 .39342 .67368 Metetra
.819 .776 .813 r
.39342 .67368 .39991 .67906 .41099 .66781 .40454 .66274 Metetra
.817 .786 .824 r
.40454 .66274 .41099 .66781 .42205 .65715 .41563 .65234 Metetra
.81 .796 .839 r
.41563 .65234 .42205 .65715 .43309 .64732 .4267 .64276 Metetra
.798 .806 .858 r
.4267 .64276 .43309 .64732 .4441 .63855 .43775 .63418 Metetra
.779 .815 .881 r
.43775 .63418 .4441 .63855 .45509 .63101 .44879 .6268 Metetra
.749 .821 .908 r
.44879 .6268 .45509 .63101 .46608 .62486 .45981 .62074 Metetra
.705 .821 .937 r
.45981 .62074 .46608 .62486 .47708 .62018 .47084 .61609 Metetra
.644 .812 .964 r
.47084 .61609 .47708 .62018 .48809 .61701 .48189 .6129 Metetra
.565 .789 .983 r
.48189 .6129 .48809 .61701 .49913 .61534 .49296 .61114 Metetra
.469 .75 .99 r
.49296 .61114 .49913 .61534 .51023 .6151 .50409 .61076 Metetra
.367 .697 .983 r
.50409 .61076 .51023 .6151 .52139 .61618 .51527 .61165 Metetra
.272 .639 .965 r
.51527 .61165 .52139 .61618 .53262 .61842 .52653 .61366 Metetra
.198 .586 .943 r
.52653 .61366 .53262 .61842 .54394 .62163 .53788 .61659 Metetra
.153 .547 .926 r
.53788 .61659 .54394 .62163 .55536 .62556 .54931 .62021 Metetra
.139 .526 .918 r
.54931 .62021 .55536 .62556 .56688 .62994 .56085 .62426 Metetra
.155 .523 .918 r
.56085 .62426 .56688 .62994 .5785 .63449 .57248 .62848 Metetra
.198 .537 .925 r
.57248 .62848 .5785 .63449 .59021 .63891 .58422 .63257 Metetra
.26 .561 .934 r
.58422 .63257 .59021 .63891 .60202 .64291 .59604 .63624 Metetra
.333 .591 .939 r
.59604 .63624 .60202 .64291 .61391 .64619 .60794 .63923 Metetra
.41 .622 .939 r
.60794 .63923 .61391 .64619 .62587 .6485 .61991 .64129 Metetra
.483 .65 .932 r
.61991 .64129 .62587 .6485 .63787 .64961 .63194 .64219 Metetra
.55 .675 .921 r
.63194 .64219 .63787 .64961 .64991 .64934 .644 .64177 Metetra
.607 .695 .906 r
.644 .64177 .64991 .64934 .66196 .64756 .65607 .6399 Metetra
.655 .712 .889 r
.65607 .6399 .66196 .64756 .67401 .6442 .66815 .63652 Metetra
.695 .726 .873 r
.66815 .63652 .67401 .6442 .68604 .63926 .68021 .63161 Metetra
.728 .737 .859 r
.68021 .63161 .68604 .63926 .69803 .6328 .69224 .62525 Metetra
.755 .748 .846 r
.69224 .62525 .69803 .6328 .70997 .62493 .70424 .61753 Metetra
.776 .758 .836 r
.70424 .61753 .70997 .62493 .72187 .61583 .71619 .60863 Metetra
.793 .767 .829 r
.71619 .60863 .72187 .61583 .73372 .60572 .72809 .59876 Metetra
.806 .776 .825 r
.72809 .59876 .73372 .60572 .74552 .59485 .73996 .58816 Metetra
.816 .785 .824 r
.73996 .58816 .74552 .59485 .75729 .58352 .7518 .57711 Metetra
.822 .795 .826 r
.7518 .57711 .75729 .58352 .76905 .57202 .76363 .56589 Metetra
.824 .804 .832 r
.76363 .56589 .76905 .57202 .78081 .56065 .77546 .55481 Metetra
.822 .814 .842 r
.77546 .55481 .78081 .56065 .7926 .54972 .78733 .54414 Metetra
.815 .822 .856 r
.78733 .54414 .7926 .54972 .80445 .53949 .79925 .53414 Metetra
.801 .83 .873 r
.79925 .53414 .80445 .53949 .81638 .53022 .81125 .52507 Metetra
.78 .835 .894 r
.81125 .52507 .81638 .53022 .82844 .52212 .82338 .51712 Metetra
.748 .835 .918 r
.82338 .51712 .82844 .52212 .84064 .51537 .83564 .51045 Metetra
.705 .829 .941 r
.83564 .51045 .84064 .51537 .85302 .51008 .84807 .5052 Metetra
.649 .813 .962 r
.84807 .5052 .85302 .51008 .8656 .50634 .8607 .50143 Metetra
.582 .787 .977 r
.8607 .50143 .8656 .50634 .8784 .50416 .87354 .49917 Metetra
.507 .75 .983 r
.87354 .49917 .8784 .50416 .89144 .50351 .88661 .49837 Metetra
.432 .705 .98 r
.88661 .49837 .89144 .50351 .90472 .50431 .89992 .49897 Metetra
.364 .659 .97 r
.89992 .49897 .90472 .50431 .91826 .50643 .91347 .50083 Metetra
.31 .617 .956 r
.91347 .50083 .91826 .50643 .93204 .50968 .92725 .50378 Metetra
.275 .583 .944 r
.92725 .50378 .93204 .50968 .94605 .51384 .94126 .50759 Metetra
.26 .56 .933 r
.94126 .50759 .94605 .51384 .96027 .51864 .95547 .51203 Metetra
.263 .549 .927 r
.95547 .51203 .96027 .51864 .97466 .5238 .96985 .5168 Metetra
.813 .768 .811 r
.37572 .67919 .38229 .68492 .39342 .67368 .38688 .6684 Metetra
.815 .78 .82 r
.38688 .6684 .39342 .67368 .40454 .66274 .39803 .65788 Metetra
.814 .794 .833 r
.39803 .65788 .40454 .66274 .41563 .65234 .40915 .64789 Metetra
.809 .808 .85 r
.40915 .64789 .41563 .65234 .4267 .64276 .42026 .63864 Metetra
.798 .822 .871 r
.42026 .63864 .4267 .64276 .43775 .63418 .43134 .63035 Metetra
.778 .835 .896 r
.43134 .63035 .43775 .63418 .44879 .6268 .44241 .62319 Metetra
.746 .845 .925 r
.44241 .62319 .44879 .6268 .45981 .62074 .45347 .61727 Metetra
.699 .847 .954 r
.45347 .61727 .45981 .62074 .47084 .61609 .46454 .61268 Metetra
.632 .838 .979 r
.46454 .61268 .47084 .61609 .48189 .6129 .47562 .60946 Metetra
.547 .812 .995 r
.47562 .60946 .48189 .6129 .49296 .61114 .48674 .60759 Metetra
.447 .769 .997 r
.48674 .60759 .49296 .61114 .50409 .61076 .49789 .60703 Metetra
.346 .713 .985 r
.49789 .60703 .50409 .61076 .51527 .61165 .50911 .60766 Metetra
.257 .654 .965 r
.50911 .60766 .51527 .61165 .52653 .61366 .5204 .60934 Metetra
.193 .603 .945 r
.5204 .60934 .52653 .61366 .53788 .61659 .53177 .61189 Metetra
.16 .568 .932 r
.53177 .61189 .53788 .61659 .54931 .62021 .54322 .61509 Metetra
.159 .55 .928 r
.54322 .61509 .54931 .62021 .56085 .62426 .55478 .6187 Metetra
.186 .549 .93 r
.55478 .6187 .56085 .62426 .57248 .62848 .56643 .62245 Metetra
.236 .56 .935 r
.56643 .62245 .57248 .62848 .58422 .63257 .57817 .62608 Metetra
.299 .58 .939 r
.57817 .62608 .58422 .63257 .59604 .63624 .59001 .62932 Metetra
.37 .603 .938 r
.59001 .62932 .59604 .63624 .60794 .63923 .60192 .63192 Metetra
.44 .627 .932 r
.60192 .63192 .60794 .63923 .61991 .64129 .61391 .63363 Metetra
.505 .648 .921 r
.61391 .63363 .61991 .64129 .63194 .64219 .62594 .63425 Metetra
.563 .667 .907 r
.62594 .63425 .63194 .64219 .644 .64177 .63802 .63363 Metetra
.612 .683 .892 r
.63802 .63363 .644 .64177 .65607 .6399 .65012 .63164 Metetra
.655 .697 .877 r
.65012 .63164 .65607 .6399 .66815 .63652 .66222 .62823 Metetra
.691 .71 .863 r
.66222 .62823 .66815 .63652 .68021 .63161 .67432 .62339 Metetra
.721 .722 .851 r
.67432 .62339 .68021 .63161 .69224 .62525 .68639 .61716 Metetra
.746 .734 .841 r
.68639 .61716 .69224 .62525 .70424 .61753 .69844 .60967 Metetra
.767 .746 .834 r
.69844 .60967 .70424 .61753 .71619 .60863 .71044 .60106 Metetra
.784 .758 .829 r
.71044 .60106 .71619 .60863 .72809 .59876 .72242 .59153 Metetra
.798 .77 .827 r
.72242 .59153 .72809 .59876 .73996 .58816 .73435 .58132 Metetra
.809 .782 .828 r
.73435 .58132 .73996 .58816 .7518 .57711 .74627 .57068 Metetra
.816 .795 .832 r
.74627 .57068 .7518 .57711 .76363 .56589 .75818 .55988 Metetra
.82 .808 .84 r
.75818 .55988 .76363 .56589 .77546 .55481 .77009 .5492 Metetra
.819 .822 .852 r
.77009 .5492 .77546 .55481 .78733 .54414 .78204 .53891 Metetra
.813 .835 .867 r
.78204 .53891 .78733 .54414 .79925 .53414 .79405 .52925 Metetra
.799 .847 .887 r
.79405 .52925 .79925 .53414 .81125 .52507 .80613 .52046 Metetra
.777 .856 .909 r
.80613 .52046 .81125 .52507 .82338 .51712 .81833 .51272 Metetra
.744 .859 .934 r
.81833 .51272 .82338 .51712 .83564 .51045 .83066 .5062 Metetra
.697 .855 .958 r
.83066 .5062 .83564 .51045 .84807 .5052 .84315 .50101 Metetra
.637 .839 .978 r
.84315 .50101 .84807 .5052 .8607 .50143 .85583 .49722 Metetra
.566 .811 .991 r
.85583 .49722 .8607 .50143 .87354 .49917 .86871 .49484 Metetra
.49 .772 .994 r
.86871 .49484 .87354 .49917 .88661 .49837 .88181 .49385 Metetra
.415 .725 .989 r
.88181 .49385 .88661 .49837 .89992 .49897 .89513 .49417 Metetra
.352 .677 .978 r
.89513 .49417 .89992 .49897 .91347 .50083 .90869 .49569 Metetra
.306 .635 .964 r
.90869 .49569 .91347 .50083 .92725 .50378 .92247 .49822 Metetra
.279 .602 .952 r
.92247 .49822 .92725 .50378 .94126 .50759 .93646 .50157 Metetra
.273 .58 .942 r
.93646 .50157 .94126 .50759 .95547 .51203 .95065 .5055 Metetra
.285 .568 .935 r
.95065 .5055 .95547 .51203 .96985 .5168 .965 .50975 Metetra
.806 .768 .818 r
.36911 .67342 .37572 .67919 .38688 .6684 .38029 .6632 Metetra
.81 .784 .829 r
.38029 .6632 .38688 .6684 .39803 .65788 .39145 .65324 Metetra
.81 .802 .844 r
.39145 .65324 .39803 .65788 .40915 .64789 .4026 .64375 Metetra
.805 .82 .863 r
.4026 .64375 .40915 .64789 .42026 .63864 .41374 .63495 Metetra
.794 .838 .886 r
.41374 .63495 .42026 .63864 .43134 .63035 .42485 .62703 Metetra
.773 .855 .912 r
.42485 .62703 .43134 .63035 .44241 .62319 .43596 .62015 Metetra
.738 .867 .941 r
.43596 .62015 .44241 .62319 .45347 .61727 .44706 .61441 Metetra
.687 .871 .968 r
.44706 .61441 .45347 .61727 .46454 .61268 .45817 .6099 Metetra
.615 .86 .989 r
.45817 .6099 .46454 .61268 .47562 .60946 .46929 .60666 Metetra
.525 .831 1 r
.46929 .60666 .47562 .60946 .48674 .60759 .48044 .60466 Metetra
.426 .785 .996 r
.48044 .60466 .48674 .60759 .49789 .60703 .49164 .60387 Metetra
.33 .729 .981 r
.49164 .60387 .49789 .60703 .50911 .60766 .50289 .60417 Metetra
.252 .673 .963 r
.50289 .60417 .50911 .60766 .5204 .60934 .51421 .60545 Metetra
.202 .627 .949 r
.51421 .60545 .5204 .60934 .53177 .61189 .5256 .60752 Metetra
.183 .596 .942 r
.5256 .60752 .53177 .61189 .54322 .61509 .53708 .61019 Metetra
.193 .582 .941 r
.53708 .61019 .54322 .61509 .55478 .6187 .54866 .61323 Metetra
.229 .58 .944 r
.54866 .61323 .55478 .6187 .56643 .62245 .56032 .61641 Metetra
.282 .588 .946 r
.56032 .61641 .56643 .62245 .57817 .62608 .57208 .61947 Metetra
.344 .602 .943 r
.57208 .61947 .57817 .62608 .59001 .62932 .58393 .62216 Metetra
.408 .617 .936 r
.58393 .62216 .59001 .62932 .60192 .63192 .59585 .62426 Metetra
.47 .632 .925 r
.59585 .62426 .60192 .63192 .61391 .63363 .60785 .62554 Metetra
.526 .646 .91 r
.60785 .62554 .61391 .63363 .62594 .63425 .6199 .62582 Metetra
.575 .66 .895 r
.6199 .62582 .62594 .63425 .63802 .63363 .63199 .62494 Metetra
.618 .672 .88 r
.63199 .62494 .63802 .63363 .65012 .63164 .64411 .62281 Metetra
.654 .685 .866 r
.64411 .62281 .65012 .63164 .66222 .62823 .65624 .61937 Metetra
.686 .697 .854 r
.65624 .61937 .66222 .62823 .67432 .62339 .66837 .61461 Metetra
.713 .709 .844 r
.66837 .61461 .67432 .62339 .68639 .61716 .68049 .60858 Metetra
.737 .721 .837 r
.68049 .60858 .68639 .61716 .69844 .60967 .69258 .60137 Metetra
.757 .734 .832 r
.69258 .60137 .69844 .60967 .71044 .60106 .70465 .59314 Metetra
.774 .748 .83 r
.70465 .59314 .71044 .60106 .72242 .59153 .71668 .58406 Metetra
.788 .763 .83 r
.71668 .58406 .72242 .59153 .73435 .58132 .7287 .57434 Metetra
.8 .779 .833 r
.7287 .57434 .73435 .58132 .74627 .57068 .7407 .56423 Metetra
.808 .795 .84 r
.7407 .56423 .74627 .57068 .75818 .55988 .75269 .55396 Metetra
.813 .812 .85 r
.75269 .55396 .75818 .55988 .77009 .5492 .7647 .54381 Metetra
.813 .83 .863 r
.7647 .54381 .77009 .5492 .78204 .53891 .77674 .534 Metetra
.807 .847 .88 r
.77674 .534 .78204 .53891 .79405 .52925 .78883 .52479 Metetra
.793 .863 .901 r
.78883 .52479 .79405 .52925 .80613 .52046 .80101 .51636 Metetra
.77 .875 .924 r
.80101 .51636 .80613 .52046 .81833 .51272 .81328 .50891 Metetra
.734 .881 .948 r
.81328 .50891 .81833 .51272 .83066 .5062 .82569 .50258 Metetra
.684 .878 .971 r
.82569 .50258 .83066 .5062 .84315 .50101 .83825 .49748 Metetra
.621 .862 .988 r
.83825 .49748 .84315 .50101 .85583 .49722 .85098 .49366 Metetra
.547 .832 .998 r
.85098 .49366 .85583 .49722 .86871 .49484 .8639 .49115 Metetra
.471 .791 .999 r
.8639 .49115 .86871 .49484 .88181 .49385 .87703 .48992 Metetra
.401 .744 .993 r
.87703 .48992 .88181 .49385 .89513 .49417 .89037 .4899 Metetra
.346 .698 .983 r
.89037 .4899 .89513 .49417 .90869 .49569 .90392 .49097 Metetra
.309 .657 .972 r
.90392 .49097 .90869 .49569 .92247 .49822 .91769 .49298 Metetra
.293 .626 .961 r
.91769 .49298 .92247 .49822 .93646 .50157 .93165 .49575 Metetra
.296 .604 .952 r
.93165 .49575 .93646 .50157 .95065 .5055 .9458 .49904 Metetra
.315 .592 .943 r
.9458 .49904 .95065 .5055 .965 .50975 .96011 .50263 Metetra
.797 .769 .827 r
.36245 .6676 .36911 .67342 .38029 .6632 .37363 .65808 Metetra
.802 .789 .84 r
.37363 .65808 .38029 .6632 .39145 .65324 .38481 .64878 Metetra
.803 .81 .857 r
.38481 .64878 .39145 .65324 .4026 .64375 .39599 .6399 Metetra
.798 .832 .878 r
.39599 .6399 .4026 .64375 .41374 .63495 .40715 .63165 Metetra
.786 .855 .902 r
.40715 .63165 .41374 .63495 .42485 .62703 .41829 .62418 Metetra
.763 .874 .928 r
.41829 .62418 .42485 .62703 .43596 .62015 .42943 .61764 Metetra
.725 .888 .955 r
.42943 .61764 .43596 .62015 .44706 .61441 .44057 .61213 Metetra
.668 .891 .978 r
.44057 .61213 .44706 .61441 .45817 .6099 .45172 .60772 Metetra
.593 .877 .993 r
.45172 .60772 .45817 .6099 .46929 .60666 .46288 .60445 Metetra
.502 .846 .996 r
.46288 .60445 .46929 .60666 .48044 .60466 .47408 .60231 Metetra
.407 .799 .988 r
.47408 .60231 .48044 .60466 .49164 .60387 .48531 .60124 Metetra
.322 .746 .974 r
.48531 .60124 .49164 .60387 .50289 .60417 .4966 .60116 Metetra
.259 .696 .962 r
.4966 .60116 .50289 .60417 .51421 .60545 .50796 .60195 Metetra
.224 .657 .955 r
.50796 .60195 .51421 .60545 .5256 .60752 .51938 .60345 Metetra
.219 .632 .954 r
.51938 .60345 .5256 .60752 .53708 .61019 .53089 .60549 Metetra
.241 .62 .957 r
.53089 .60549 .53708 .61019 .54866 .61323 .54249 .60786 Metetra
.282 .616 .958 r
.54249 .60786 .54866 .61323 .56032 .61641 .55417 .61034 Metetra
.335 .619 .954 r
.55417 .61034 .56032 .61641 .57208 .61947 .56594 .61272 Metetra
.392 .624 .946 r
.56594 .61272 .57208 .61947 .58393 .62216 .5778 .61477 Metetra
.448 .631 .932 r
.5778 .61477 .58393 .62216 .59585 .62426 .58973 .61628 Metetra
.5 .638 .916 r
.58973 .61628 .59585 .62426 .60785 .62554 .60174 .61705 Metetra
.546 .646 .9 r
.60174 .61705 .60785 .62554 .6199 .62582 .6138 .61692 Metetra
.587 .655 .884 r
.6138 .61692 .6199 .62582 .63199 .62494 .62591 .61575 Metetra
.622 .664 .869 r
.62591 .61575 .63199 .62494 .64411 .62281 .63804 .61345 Metetra
.653 .674 .857 r
.63804 .61345 .64411 .62281 .65624 .61937 .6502 .60998 Metetra
.681 .684 .847 r
.6502 .60998 .65624 .61937 .66837 .61461 .66236 .60532 Metetra
.705 .696 .839 r
.66236 .60532 .66837 .61461 .68049 .60858 .67452 .59952 Metetra
.726 .709 .834 r
.67452 .59952 .68049 .60858 .69258 .60137 .68666 .59267 Metetra
.745 .724 .831 r
.68666 .59267 .69258 .60137 .70465 .59314 .69879 .58489 Metetra
.762 .739 .832 r
.69879 .58489 .70465 .59314 .71668 .58406 .7109 .57635 Metetra
.777 .756 .834 r
.7109 .57635 .71668 .58406 .7287 .57434 .72299 .56723 Metetra
.789 .775 .84 r
.72299 .56723 .7287 .57434 .7407 .56423 .73508 .55775 Metetra
.798 .794 .849 r
.73508 .55775 .7407 .56423 .75269 .55396 .74717 .54814 Metetra
.803 .815 .861 r
.74717 .54814 .75269 .55396 .7647 .54381 .75927 .53861 Metetra
.804 .837 .876 r
.75927 .53861 .7647 .54381 .77674 .534 .77141 .5294 Metetra
.798 .859 .895 r
.77141 .5294 .77674 .534 .78883 .52479 .7836 .52072 Metetra
.783 .878 .916 r
.7836 .52072 .78883 .52479 .80101 .51636 .79587 .51274 Metetra
.757 .893 .939 r
.79587 .51274 .80101 .51636 .81328 .50891 .80824 .50564 Metetra
.718 .9 .961 r
.80824 .50564 .81328 .50891 .82569 .50258 .82073 .49954 Metetra
.664 .896 .979 r
.82073 .49954 .82569 .50258 .83825 .49748 .83335 .49454 Metetra
.598 .879 .992 r
.83335 .49454 .83825 .49748 .85098 .49366 .84614 .4907 Metetra
.525 .848 .998 r
.84614 .4907 .85098 .49366 .8639 .49115 .85911 .48804 Metetra
.453 .808 .997 r
.85911 .48804 .8639 .49115 .87703 .48992 .87226 .48653 Metetra
.392 .764 .992 r
.87226 .48653 .87703 .48992 .89037 .4899 .88561 .4861 Metetra
.347 .721 .986 r
.88561 .4861 .89037 .4899 .90392 .49097 .89916 .48665 Metetra
.322 .684 .979 r
.89916 .48665 .90392 .49097 .91769 .49298 .9129 .48805 Metetra
.317 .655 .971 r
.9129 .48805 .91769 .49298 .93165 .49575 .92683 .49012 Metetra
.328 .634 .962 r
.92683 .49012 .93165 .49575 .9458 .49904 .94093 .49266 Metetra
.352 .619 .952 r
.94093 .49266 .9458 .49904 .96011 .50263 .95518 .49545 Metetra
.786 .769 .838 r
.35573 .66174 .36245 .6676 .37363 .65808 .36692 .65302 Metetra
.791 .793 .853 r
.36692 .65302 .37363 .65808 .38481 .64878 .37811 .64449 Metetra
.792 .818 .872 r
.37811 .64449 .38481 .64878 .39599 .6399 .3893 .63633 Metetra
.787 .844 .894 r
.3893 .63633 .39599 .6399 .40715 .63165 .40048 .6287 Metetra
.774 .87 .919 r
.40048 .6287 .40715 .63165 .41829 .62418 .41166 .62175 Metetra
.747 .892 .944 r
.41166 .62175 .41829 .62418 .42943 .61764 .42283 .61561 Metetra
.705 .905 .967 r
.42283 .61561 .42943 .61764 .44057 .61213 .43401 .61036 Metetra
.644 .905 .983 r
.43401 .61036 .44057 .61213 .45172 .60772 .4452 .60608 Metetra
.565 .888 .989 r
.4452 .60608 .45172 .60772 .46288 .60445 .45641 .60278 Metetra
.478 .854 .985 r
.45641 .60278 .46288 .60445 .47408 .60231 .46764 .60046 Metetra
.393 .81 .975 r
.46764 .60046 .47408 .60231 .48531 .60124 .47892 .59908 Metetra
.324 .764 .965 r
.47892 .59908 .48531 .60124 .4966 .60116 .49025 .59856 Metetra
.278 .724 .961 r
.49025 .59856 .4966 .60116 .50796 .60195 .50165 .5988 Metetra
.261 .694 .963 r
.50165 .5988 .50796 .60195 .51938 .60345 .51311 .59965 Metetra
.269 .674 .969 r
.51311 .59965 .51938 .60345 .53089 .60549 .52465 .60096 Metetra
.299 .661 .972 r
.52465 .60096 .53089 .60549 .54249 .60786 .53627 .60256 Metetra
.342 .654 .97 r
.53627 .60256 .54249 .60786 .55417 .61034 .54797 .60426 Metetra
.392 .649 .96 r
.54797 .60426 .55417 .61034 .56594 .61272 .55976 .60586 Metetra
.441 .646 .945 r
.55976 .60586 .56594 .61272 .5778 .61477 .57163 .60716 Metetra
.487 .645 .927 r
.57163 .60716 .5778 .61477 .58973 .61628 .58357 .608 Metetra
.529 .645 .908 r
.58357 .608 .58973 .61628 .60174 .61705 .59558 .60819 Metetra
.566 .647 .889 r
.59558 .60819 .60174 .61705 .6138 .61692 .60765 .60759 Metetra
.598 .651 .873 r
.60765 .60759 .6138 .61692 .62591 .61575 .61977 .60609 Metetra
.627 .657 .86 r
.61977 .60609 .62591 .61575 .63804 .61345 .63193 .60361 Metetra
.653 .664 .849 r
.63193 .60361 .63804 .61345 .6502 .60998 .64411 .60011 Metetra
.676 .674 .84 r
.64411 .60011 .6502 .60998 .66236 .60532 .6563 .59557 Metetra
.697 .685 .835 r
.6563 .59557 .66236 .60532 .67452 .59952 .6685 .59004 Metetra
.716 .698 .832 r
.6685 .59004 .67452 .59952 .68666 .59267 .68069 .58359 Metetra
.733 .713 .832 r
.68069 .58359 .68666 .59267 .69879 .58489 .69288 .57634 Metetra
.749 .73 .834 r
.69288 .57634 .69879 .58489 .7109 .57635 .70506 .56842 Metetra
.763 .749 .84 r
.70506 .56842 .7109 .57635 .72299 .56723 .71724 .55999 Metetra
.775 .77 .848 r
.71724 .55999 .72299 .56723 .73508 .55775 .72942 .55125 Metetra
.785 .793 .859 r
.72942 .55125 .73508 .55775 .74717 .54814 .7416 .54239 Metetra
.79 .818 .874 r
.7416 .54239 .74717 .54814 .75927 .53861 .75381 .5336 Metetra
.791 .843 .891 r
.75381 .5336 .75927 .53861 .77141 .5294 .76606 .52507 Metetra
.784 .868 .911 r
.76606 .52507 .77141 .5294 .7836 .52072 .77835 .517 Metetra
.767 .891 .932 r
.77835 .517 .7836 .52072 .79587 .51274 .79072 .50954 Metetra
.738 .907 .952 r
.79072 .50954 .79587 .51274 .80824 .50564 .80319 .50285 Metetra
.695 .915 .97 r
.80319 .50285 .80824 .50564 .82073 .49954 .81576 .49702 Metetra
.638 .909 .982 r
.81576 .49702 .82073 .49954 .83335 .49454 .82846 .49216 Metetra
.571 .89 .988 r
.82846 .49216 .83335 .49454 .84614 .4907 .84131 .4883 Metetra
.501 .859 .99 r
.84131 .4883 .84614 .4907 .85911 .48804 .85432 .48546 Metetra
.438 .822 .989 r
.85432 .48546 .85911 .48804 .87226 .48653 .8675 .48362 Metetra
.388 .783 .987 r
.8675 .48362 .87226 .48653 .88561 .4861 .88086 .48273 Metetra
.357 .746 .986 r
.88086 .48273 .88561 .4861 .89916 .48665 .89439 .48269 Metetra
.345 .715 .985 r
.89439 .48269 .89916 .48665 .9129 .48805 .90811 .48338 Metetra
.35 .688 .981 r
.90811 .48338 .9129 .48805 .92683 .49012 .92199 .48465 Metetra
.369 .667 .972 r
.92199 .48465 .92683 .49012 .94093 .49266 .93603 .48632 Metetra
.396 .649 .959 r
.93603 .48632 .94093 .49266 .95518 .49545 .95021 .48822 Metetra
.771 .769 .85 r
.34897 .65584 .35573 .66174 .36692 .65302 .36015 .64802 Metetra
.777 .796 .868 r
.36015 .64802 .36692 .65302 .37811 .64449 .37134 .64034 Metetra
.778 .825 .889 r
.37134 .64034 .37811 .64449 .3893 .63633 .38254 .63297 Metetra
.772 .855 .913 r
.38254 .63297 .3893 .63633 .40048 .6287 .39374 .62605 Metetra
.755 .883 .937 r
.39374 .62605 .40048 .6287 .41166 .62175 .40494 .61968 Metetra
.725 .905 .958 r
.40494 .61968 .41166 .62175 .42283 .61561 .41615 .61399 Metetra
.677 .917 .974 r
.41615 .61399 .42283 .61561 .43401 .61036 .42736 .60905 Metetra
.613 .912 .981 r
.42736 .60905 .43401 .61036 .4452 .60608 .4386 .6049 Metetra
.536 .892 .978 r
.4386 .6049 .4452 .60608 .45641 .60278 .44985 .60158 Metetra
.457 .858 .968 r
.44985 .60158 .45641 .60278 .46764 .60046 .46114 .59907 Metetra
.387 .82 .96 r
.46114 .59907 .46764 .60046 .47892 .59908 .47246 .59735 Metetra
.337 .784 .957 r
.47246 .59735 .47892 .59908 .49025 .59856 .48384 .59633 Metetra
.311 .755 .963 r
.48384 .59633 .49025 .59856 .50165 .5988 .49527 .59595 Metetra
.31 .734 .973 r
.49527 .59595 .50165 .5988 .51311 .59965 .50677 .59608 Metetra
.33 .718 .983 r
.50677 .59608 .51311 .59965 .52465 .60096 .51834 .59659 Metetra
.365 .704 .985 r
.51834 .59659 .52465 .60096 .53627 .60256 .52999 .59733 Metetra
.407 .691 .978 r
.52999 .59733 .53627 .60256 .54797 .60426 .54172 .59815 Metetra
.449 .678 .962 r
.54172 .59815 .54797 .60426 .55976 .60586 .55352 .59888 Metetra
.489 .667 .942 r
.55352 .59888 .55976 .60586 .57163 .60716 .56541 .59937 Metetra
.525 .658 .92 r
.56541 .59937 .57163 .60716 .58357 .608 .57736 .59945 Metetra
.556 .652 .899 r
.57736 .59945 .58357 .608 .59558 .60819 .58938 .599 Metetra
.584 .648 .88 r
.58938 .599 .59558 .60819 .60765 .60759 .60146 .59789 Metetra
.609 .648 .864 r
.60146 .59789 .60765 .60759 .61977 .60609 .61359 .59603 Metetra
.631 .651 .851 r
.61359 .59603 .61977 .60609 .63193 .60361 .62576 .59334 Metetra
.652 .656 .842 r
.62576 .59334 .63193 .60361 .64411 .60011 .63796 .5898 Metetra
.67 .664 .835 r
.63796 .5898 .64411 .60011 .6563 .59557 .65019 .58541 Metetra
.688 .675 .832 r
.65019 .58541 .6563 .59557 .6685 .59004 .66242 .58018 Metetra
.705 .688 .831 r
.66242 .58018 .6685 .59004 .68069 .58359 .67467 .57419 Metetra
.72 .703 .833 r
.67467 .57419 .68069 .58359 .69288 .57634 .68692 .56752 Metetra
.735 .721 .838 r
.68692 .56752 .69288 .57634 .70506 .56842 .69918 .56029 Metetra
.748 .741 .846 r
.69918 .56029 .70506 .56842 .71724 .55999 .71144 .55264 Metetra
.759 .765 .857 r
.71144 .55264 .71724 .55999 .72942 .55125 .72371 .54472 Metetra
.768 .791 .871 r
.72371 .54472 .72942 .55125 .7416 .54239 .736 .53669 Metetra
.774 .819 .888 r
.736 .53669 .7416 .54239 .75381 .5336 .74831 .52872 Metetra
.773 .848 .907 r
.74831 .52872 .75381 .5336 .76606 .52507 .76067 .52097 Metetra
.765 .876 .928 r
.76067 .52097 .76606 .52507 .77835 .517 .77308 .51359 Metetra
.746 .9 .947 r
.77308 .51359 .77835 .517 .79072 .50954 .78555 .50672 Metetra
.713 .917 .964 r
.78555 .50672 .79072 .50954 .80319 .50285 .79812 .50048 Metetra
.666 .923 .974 r
.79812 .50048 .80319 .50285 .81576 .49702 .81078 .49496 Metetra
.606 .914 .979 r
.81078 .49496 .81576 .49702 .82846 .49216 .82356 .49025 Metetra
.541 .894 .978 r
.82356 .49025 .82846 .49216 .84131 .4883 .83647 .48637 Metetra
.478 .865 .976 r
.83647 .48637 .84131 .4883 .85432 .48546 .84953 .48334 Metetra
.427 .833 .976 r
.84953 .48334 .85432 .48546 .8675 .48362 .86273 .48114 Metetra
.392 .802 .98 r
.86273 .48114 .8675 .48362 .88086 .48273 .87609 .47973 Metetra
.375 .774 .986 r
.87609 .47973 .88086 .48273 .89439 .48269 .88962 .47903 Metetra
.377 .748 .991 r
.88962 .47903 .89439 .48269 .90811 .48338 .9033 .47893 Metetra
.392 .724 .989 r
.9033 .47893 .90811 .48338 .92199 .48465 .91713 .47931 Metetra
.416 .702 .981 r
.91713 .47931 .92199 .48465 .93603 .48632 .9311 .48003 Metetra
.444 .68 .965 r
.9311 .48003 .93603 .48632 .95021 .48822 .9452 .48092 Metetra
.753 .768 .865 r
.34215 .64989 .34897 .65584 .36015 .64802 .35332 .64304 Metetra
.758 .798 .886 r
.35332 .64304 .36015 .64802 .37134 .64034 .36451 .63631 Metetra
.758 .831 .908 r
.36451 .63631 .37134 .64034 .38254 .63297 .37571 .62981 Metetra
.749 .864 .932 r
.37571 .62981 .38254 .63297 .39374 .62605 .38692 .62365 Metetra
.729 .893 .954 r
.38692 .62365 .39374 .62605 .40494 .61968 .39815 .61793 Metetra
.694 .914 .97 r
.39815 .61793 .40494 .61968 .41615 .61399 .40939 .61273 Metetra
.643 .921 .977 r
.40939 .61273 .41615 .61399 .42736 .60905 .42064 .60811 Metetra
.578 .913 .973 r
.42064 .60811 .42736 .60905 .4386 .6049 .43191 .60412 Metetra
.507 .89 .962 r
.43191 .60412 .4386 .6049 .44985 .60158 .44321 .60077 Metetra
.441 .86 .951 r
.44321 .60077 .44985 .60158 .46114 .59907 .45455 .59806 Metetra
.391 .831 .946 r
.45455 .59806 .46114 .59907 .47246 .59735 .46593 .59596 Metetra
.362 .807 .952 r
.46593 .59596 .47246 .59735 .48384 .59633 .47735 .59441 Metetra
.355 .789 .966 r
.47735 .59441 .48384 .59633 .49527 .59595 .48883 .59335 Metetra
.37 .776 .983 r
.48883 .59335 .49527 .59595 .50677 .59608 .50038 .5927 Metetra
.398 .761 .993 r
.50038 .5927 .50677 .59608 .51834 .59659 .51199 .59233 Metetra
.434 .744 .993 r
.51199 .59233 .51834 .59659 .52999 .59733 .52367 .59214 Metetra
.471 .724 .98 r
.52367 .59214 .52999 .59733 .54172 .59815 .53542 .59201 Metetra
.505 .704 .96 r
.53542 .59201 .54172 .59815 .55352 .59888 .54724 .5918 Metetra
.535 .685 .936 r
.54724 .5918 .55352 .59888 .56541 .59937 .55914 .5914 Metetra
.56 .67 .911 r
.55914 .5914 .56541 .59937 .57736 .59945 .57111 .59067 Metetra
.582 .658 .89 r
.57111 .59067 .57736 .59945 .58938 .599 .58314 .58953 Metetra
.602 .651 .871 r
.58314 .58953 .58938 .599 .60146 .59789 .59523 .58786 Metetra
.619 .647 .856 r
.59523 .58786 .60146 .59789 .61359 .59603 .60737 .58561 Metetra
.636 .647 .844 r
.60737 .58561 .61359 .59603 .62576 .59334 .61955 .58271 Metetra
.651 .65 .836 r
.61955 .58271 .62576 .59334 .63796 .5898 .63177 .57913 Metetra
.665 .656 .832 r
.63177 .57913 .63796 .5898 .65019 .58541 .64403 .57489 Metetra
.679 .666 .83 r
.64403 .57489 .65019 .58541 .66242 .58018 .6563 .56999 Metetra
.693 .678 .831 r
.6563 .56999 .66242 .58018 .67467 .57419 .6686 .56449 Metetra
.706 .693 .835 r
.6686 .56449 .67467 .57419 .68692 .56752 .68091 .55846 Metetra
.718 .712 .843 r
.68091 .55846 .68692 .56752 .69918 .56029 .69324 .55199 Metetra
.73 .733 .853 r
.69324 .55199 .69918 .56029 .71144 .55264 .70559 .54518 Metetra
.741 .759 .867 r
.70559 .54518 .71144 .55264 .72371 .54472 .71796 .53815 Metetra
.749 .787 .884 r
.71796 .53815 .72371 .54472 .736 .53669 .73035 .53104 Metetra
.753 .818 .903 r
.73035 .53104 .736 .53669 .74831 .52872 .74277 .52396 Metetra
.751 .85 .924 r
.74277 .52396 .74831 .52872 .76067 .52097 .75524 .51705 Metetra
.74 .88 .944 r
.75524 .51705 .76067 .52097 .77308 .51359 .76776 .51043 Metetra
.718 .906 .962 r
.76776 .51043 .77308 .51359 .78555 .50672 .78035 .5042 Metetra
.681 .921 .972 r
.78035 .5042 .78555 .50672 .79812 .50048 .79302 .49846 Metetra
.631 .924 .975 r
.79302 .49846 .79812 .50048 .81078 .49496 .80578 .49329 Metetra
.571 .913 .97 r
.80578 .49329 .81078 .49496 .82356 .49025 .81864 .48873 Metetra
.511 .892 .963 r
.81864 .48873 .82356 .49025 .83647 .48637 .83162 .48484 Metetra
.459 .867 .958 r
.83162 .48484 .83647 .48637 .84953 .48334 .84472 .4816 Metetra
.422 .842 .961 r
.84472 .4816 .84953 .48334 .86273 .48114 .85795 .47902 Metetra
.404 .821 .971 r
.85795 .47902 .86273 .48114 .87609 .47973 .87132 .47705 Metetra
.404 .802 .985 r
.87132 .47705 .87609 .47973 .88962 .47903 .88482 .47562 Metetra
.418 .783 .995 r
.88482 .47562 .88962 .47903 .9033 .47893 .89846 .47467 Metetra
.441 .761 .996 r
.89846 .47467 .9033 .47893 .91713 .47931 .91224 .47409 Metetra
.468 .736 .986 r
.91224 .47409 .91713 .47931 .9311 .48003 .92613 .47376 Metetra
.494 .71 .967 r
.92613 .47376 .9311 .48003 .9452 .48092 .94014 .47356 Metetra
.73 .765 .881 r
.33528 .64389 .34215 .64989 .35332 .64304 .34643 .63809 Metetra
.733 .799 .904 r
.34643 .63809 .35332 .64304 .36451 .63631 .35761 .63236 Metetra
.731 .835 .929 r
.35761 .63236 .36451 .63631 .37571 .62981 .36881 .62678 Metetra
.719 .869 .952 r
.36881 .62678 .37571 .62981 .38692 .62365 .38003 .62144 Metetra
.695 .898 .97 r
.38003 .62144 .38692 .62365 .39815 .61793 .39128 .61641 Metetra
.656 .916 .978 r
.39128 .61641 .39815 .61793 .40939 .61273 .40254 .61174 Metetra
.602 .919 .975 r
.40254 .61174 .40939 .61273 .42064 .60811 .41383 .60748 Metetra
.541 .906 .961 r
.41383 .60748 .42064 .60811 .43191 .60412 .42515 .60365 Metetra
.481 .885 .946 r
.42515 .60365 .43191 .60412 .44321 .60077 .4365 .60028 Metetra
.434 .862 .936 r
.4365 .60028 .44321 .60077 .45455 .59806 .44789 .59735 Metetra
.404 .844 .937 r
.44789 .59735 .45455 .59806 .46593 .59596 .45932 .59484 Metetra
.397 .832 .951 r
.45932 .59484 .46593 .59596 .47735 .59441 .47079 .59273 Metetra
.409 .824 .972 r
.47079 .59273 .47735 .59441 .48883 .59335 .48233 .59095 Metetra
.435 .815 .991 r
.48233 .59095 .48883 .59335 .50038 .5927 .49392 .58945 Metetra
.469 .801 .999 r
.49392 .58945 .50038 .5927 .51199 .59233 .50557 .58816 Metetra
.503 .779 .994 r
.50557 .58816 .51199 .59233 .52367 .59214 .51729 .58698 Metetra
.533 .753 .977 r
.51729 .58698 .52367 .59214 .53542 .59201 .52907 .58583 Metetra
.557 .726 .953 r
.52907 .58583 .53542 .59201 .54724 .5918 .54092 .58463 Metetra
.577 .702 .927 r
.54092 .58463 .54724 .5918 .55914 .5914 .55283 .58328 Metetra
.593 .681 .902 r
.55283 .58328 .55914 .5914 .57111 .59067 .56481 .5817 Metetra
.606 .665 .881 r
.56481 .5817 .57111 .59067 .58314 .58953 .57685 .57982 Metetra
.618 .654 .863 r
.57685 .57982 .58314 .58953 .59523 .58786 .58895 .57757 Metetra
.63 .647 .849 r
.58895 .57757 .59523 .58786 .60737 .58561 .6011 .5749 Metetra
.64 .645 .839 r
.6011 .5749 .60737 .58561 .61955 .58271 .6133 .57177 Metetra
.65 .646 .832 r
.6133 .57177 .61955 .58271 .63177 .57913 .62554 .56816 Metetra
.66 .65 .829 r
.62554 .56816 .63177 .57913 .64403 .57489 .63782 .56408 Metetra
.67 .658 .829 r
.63782 .56408 .64403 .57489 .6563 .56999 .65014 .55953 Metetra
.681 .669 .832 r
.65014 .55953 .6563 .56999 .6686 .56449 .66249 .55456 Metetra
.691 .684 .839 r
.66249 .55456 .6686 .56449 .68091 .55846 .67486 .5492 Metetra
.701 .702 .848 r
.67486 .5492 .68091 .55846 .69324 .55199 .68726 .54353 Metetra
.71 .725 .861 r
.68726 .54353 .69324 .55199 .70559 .54518 .69969 .53762 Metetra
.719 .751 .878 r
.69969 .53762 .70559 .54518 .71796 .53815 .71216 .53155 Metetra
.725 .781 .897 r
.71216 .53155 .71796 .53815 .73035 .53104 .72465 .52541 Metetra
.727 .814 .919 r
.72465 .52541 .73035 .53104 .74277 .52396 .73719 .51929 Metetra
.723 .849 .941 r
.73719 .51929 .74277 .52396 .75524 .51705 .74977 .51327 Metetra
.709 .881 .961 r
.74977 .51327 .75524 .51705 .76776 .51043 .76241 .50746 Metetra
.683 .906 .974 r
.76241 .50746 .76776 .51043 .78035 .5042 .77512 .50192 Metetra
.643 .919 .978 r
.77512 .50192 .78035 .5042 .79302 .49846 .78789 .49672 Metetra
.592 .919 .971 r
.78789 .49672 .79302 .49846 .80578 .49329 .80075 .49191 Metetra
.536 .906 .959 r
.80075 .49191 .80578 .49329 .81864 .48873 .81369 .48754 Metetra
.485 .887 .947 r
.81369 .48754 .81864 .48873 .83162 .48484 .82673 .48362 Metetra
.447 .867 .942 r
.82673 .48362 .83162 .48484 .84472 .4816 .83988 .48017 Metetra
.426 .852 .948 r
.83988 .48017 .84472 .4816 .85795 .47902 .85314 .47717 Metetra
.425 .842 .964 r
.85314 .47717 .85795 .47902 .87132 .47705 .86651 .4746 Metetra
.44 .832 .984 r
.86651 .4746 .87132 .47705 .88482 .47562 .88 .47241 Metetra
.465 .818 .997 r
.88 .47241 .88482 .47562 .89846 .47467 .8936 .47054 Metetra
.495 .797 .999 r
.8936 .47054 .89846 .47467 .91224 .47409 .90731 .46893 Metetra
.522 .769 .987 r
.90731 .46893 .91224 .47409 .92613 .47376 .92113 .46749 Metetra
.545 .739 .965 r
.92113 .46749 .92613 .47376 .94014 .47356 .93504 .46614 Metetra
.702 .761 .898 r
.32836 .63785 .33528 .64389 .34643 .63809 .33949 .63313 Metetra
.702 .797 .924 r
.33949 .63313 .34643 .63809 .35761 .63236 .35065 .62845 Metetra
.696 .834 .949 r
.35065 .62845 .35761 .63236 .36881 .62678 .36184 .62384 Metetra
.68 .869 .97 r
.36184 .62384 .36881 .62678 .38003 .62144 .37306 .61937 Metetra
.652 .896 .982 r
.37306 .61937 .38003 .62144 .39128 .61641 .38432 .61506 Metetra
.609 .91 .982 r
.38432 .61506 .39128 .61641 .40254 .61174 .39561 .61094 Metetra
.558 .909 .968 r
.39561 .61094 .40254 .61174 .41383 .60748 .40694 .60705 Metetra
.506 .896 .949 r
.40694 .60705 .41383 .60748 .42515 .60365 .4183 .6034 Metetra
.462 .88 .932 r
.4183 .6034 .42515 .60365 .4365 .60028 .4297 .6 Metetra
.435 .866 .926 r
.4297 .6 .4365 .60028 .44789 .59735 .44114 .59684 Metetra
.428 .86 .934 r
.44114 .59684 .44789 .59735 .45932 .59484 .45263 .59392 Metetra
.441 .859 .953 r
.45263 .59392 .45932 .59484 .47079 .59273 .46416 .59121 Metetra
.468 .857 .977 r
.46416 .59121 .47079 .59273 .48233 .59095 .47575 .58868 Metetra
.503 .85 .994 r
.47575 .58868 .48233 .59095 .49392 .58945 .48739 .5863 Metetra
.538 .833 .999 r
.48739 .5863 .49392 .58945 .50557 .58816 .49909 .58403 Metetra
.567 .806 .989 r
.49909 .58403 .50557 .58816 .51729 .58698 .51085 .58182 Metetra
.589 .775 .968 r
.51085 .58182 .51729 .58698 .52907 .58583 .52266 .57962 Metetra
.604 .744 .943 r
.52266 .57962 .52907 .58583 .54092 .58463 .53454 .57737 Metetra
.615 .715 .917 r
.53454 .57737 .54092 .58463 .55283 .58328 .54648 .57504 Metetra
.623 .691 .893 r
.54648 .57504 .55283 .58328 .56481 .5817 .55847 .57258 Metetra
.629 .672 .873 r
.55847 .57258 .56481 .5817 .57685 .57982 .57053 .56993 Metetra
.634 .658 .856 r
.57053 .56993 .57685 .57982 .58895 .57757 .58264 .56707 Metetra
.639 .649 .843 r
.58264 .56707 .58895 .57757 .6011 .5749 .5948 .56397 Metetra
.644 .644 .835 r
.5948 .56397 .6011 .5749 .6133 .57177 .60701 .56061 Metetra
.65 .642 .829 r
.60701 .56061 .6133 .57177 .62554 .56816 .61927 .55697 Metetra
.655 .645 .828 r
.61927 .55697 .62554 .56816 .63782 .56408 .63158 .55305 Metetra
.661 .651 .829 r
.63158 .55305 .63782 .56408 .65014 .55953 .64393 .54887 Metetra
.668 .661 .834 r
.64393 .54887 .65014 .55953 .66249 .55456 .65633 .54444 Metetra
.675 .675 .843 r
.65633 .54444 .66249 .55456 .67486 .5492 .66876 .53979 Metetra
.681 .693 .854 r
.66876 .53979 .67486 .5492 .68726 .54353 .68123 .53495 Metetra
.688 .716 .87 r
.68123 .53495 .68726 .54353 .69969 .53762 .69375 .52997 Metetra
.693 .742 .889 r
.69375 .52997 .69969 .53762 .71216 .53155 .70631 .5249 Metetra
.697 .774 .911 r
.70631 .5249 .71216 .53155 .72465 .52541 .71891 .51977 Metetra
.696 .808 .934 r
.71891 .51977 .72465 .52541 .73719 .51929 .73156 .51465 Metetra
.689 .843 .957 r
.73156 .51465 .73719 .51929 .74977 .51327 .74426 .50958 Metetra
.672 .876 .975 r
.74426 .50958 .74977 .51327 .76241 .50746 .75702 .50462 Metetra
.642 .9 .984 r
.75702 .50462 .76241 .50746 .77512 .50192 .76983 .4998 Metetra
.601 .911 .98 r
.76983 .4998 .77512 .50192 .78789 .49672 .78272 .49516 Metetra
.552 .908 .966 r
.78272 .49516 .78789 .49672 .80075 .49191 .79567 .49074 Metetra
.503 .896 .947 r
.79567 .49074 .80075 .49191 .81369 .48754 .8087 .48656 Metetra
.465 .881 .933 r
.8087 .48656 .81369 .48754 .82673 .48362 .82182 .48262 Metetra
.442 .869 .929 r
.82182 .48262 .82673 .48362 .83988 .48017 .83501 .47894 Metetra
.44 .864 .939 r
.83501 .47894 .83988 .48017 .85314 .47717 .8483 .47551 Metetra
.455 .863 .96 r
.8483 .47551 .85314 .47717 .86651 .4746 .86167 .47231 Metetra
.483 .86 .982 r
.86167 .47231 .86651 .4746 .88 .47241 .87514 .46931 Metetra
.517 .849 .997 r
.87514 .46931 .88 .47241 .8936 .47054 .8887 .46649 Metetra
.55 .828 .997 r
.8887 .46649 .8936 .47054 .90731 .46893 .90235 .46381 Metetra
.576 .798 .983 r
.90235 .46381 .90731 .46893 .92113 .46749 .91608 .46121 Metetra
.594 .763 .96 r
.91608 .46121 .92113 .46749 .93504 .46614 .9299 .45865 Metetra
.666 .754 .916 r
.32139 .63176 .32836 .63785 .33949 .63313 .33248 .62816 Metetra
.662 .791 .943 r
.33248 .62816 .33949 .63313 .35065 .62845 .34362 .62455 Metetra
.651 .829 .968 r
.34362 .62455 .35065 .62845 .36184 .62384 .3548 .62095 Metetra
.631 .863 .986 r
.3548 .62095 .36184 .62384 .37306 .61937 .36602 .61736 Metetra
.599 .887 .991 r
.36602 .61736 .37306 .61937 .38432 .61506 .37729 .61379 Metetra
.557 .897 .981 r
.37729 .61379 .38432 .61506 .39561 .61094 .3886 .61025 Metetra
.513 .894 .96 r
.3886 .61025 .39561 .61094 .40694 .60705 .39996 .60675 Metetra
.475 .885 .938 r
.39996 .60675 .40694 .60705 .4183 .6034 .41137 .60328 Metetra
.451 .876 .923 r
.41137 .60328 .4183 .6034 .4297 .6 .42282 .59985 Metetra
.447 .874 .922 r
.42282 .59985 .4297 .6 .44114 .59684 .43432 .59645 Metetra
.461 .878 .936 r
.43432 .59645 .44114 .59684 .45263 .59392 .44587 .5931 Metetra
.491 .884 .958 r
.44587 .5931 .45263 .59392 .46416 .59121 .45746 .58977 Metetra
.53 .886 .98 r
.45746 .58977 .46416 .59121 .47575 .58868 .46911 .58647 Metetra
.569 .877 .993 r
.46911 .58647 .47575 .58868 .48739 .5863 .48081 .58319 Metetra
.601 .856 .992 r
.48081 .58319 .48739 .5863 .49909 .58403 .49255 .57991 Metetra
.624 .826 .978 r
.49255 .57991 .49909 .58403 .51085 .58182 .50436 .57664 Metetra
.638 .791 .956 r
.50436 .57664 .51085 .58182 .52266 .57962 .51621 .57336 Metetra
.646 .757 .93 r
.51621 .57336 .52266 .57962 .53454 .57737 .52812 .57005 Metetra
.649 .727 .906 r
.52812 .57005 .53454 .57737 .54648 .57504 .54008 .56671 Metetra
.65 .7 .883 r
.54008 .56671 .54648 .57504 .55847 .57258 .55209 .56334 Metetra
.65 .679 .865 r
.55209 .56334 .55847 .57258 .57053 .56993 .56416 .55992 Metetra
.649 .663 .85 r
.56416 .55992 .57053 .56993 .58264 .56707 .57628 .55644 Metetra
.649 .651 .839 r
.57628 .55644 .58264 .56707 .5948 .56397 .58846 .55289 Metetra
.649 .644 .832 r
.58846 .55289 .5948 .56397 .60701 .56061 .60069 .54929 Metetra
.649 .641 .828 r
.60069 .54929 .60701 .56061 .61927 .55697 .61297 .54561 Metetra
.651 .641 .828 r
.61297 .54561 .61927 .55697 .63158 .55305 .6253 .54187 Metetra
.652 .646 .831 r
.6253 .54187 .63158 .55305 .64393 .54887 .63769 .53806 Metetra
.655 .654 .837 r
.63769 .53806 .64393 .54887 .65633 .54444 .65013 .53419 Metetra
.657 .666 .847 r
.65013 .53419 .65633 .54444 .66876 .53979 .66262 .53026 Metetra
.66 .683 .861 r
.66262 .53026 .66876 .53979 .68123 .53495 .67516 .52628 Metetra
.663 .705 .879 r
.67516 .52628 .68123 .53495 .69375 .52997 .68776 .52225 Metetra
.664 .732 .9 r
.68776 .52225 .69375 .52997 .70631 .5249 .70041 .51819 Metetra
.664 .763 .924 r
.70041 .51819 .70631 .5249 .71891 .51977 .71312 .51411 Metetra
.659 .798 .949 r
.71312 .51411 .71891 .51977 .73156 .51465 .72588 .51002 Metetra
.648 .833 .971 r
.72588 .51002 .73156 .51465 .74426 .50958 .73869 .50592 Metetra
.628 .865 .987 r
.73869 .50592 .74426 .50958 .75702 .50462 .75157 .50183 Metetra
.596 .887 .991 r
.75157 .50183 .75702 .50462 .76983 .4998 .7645 .49776 Metetra
.556 .896 .981 r
.7645 .49776 .76983 .4998 .78272 .49516 .77749 .49371 Metetra
.513 .894 .961 r
.77749 .49371 .78272 .49516 .79567 .49074 .79055 .48969 Metetra
.476 .885 .939 r
.79055 .48969 .79567 .49074 .8087 .48656 .80367 .4857 Metetra
.452 .877 .924 r
.80367 .4857 .8087 .48656 .82182 .48262 .81685 .48174 Metetra
.447 .874 .923 r
.81685 .48174 .82182 .48262 .83501 .47894 .83009 .47783 Metetra
.462 .878 .936 r
.83009 .47783 .83501 .47894 .8483 .47551 .84341 .47395 Metetra
.492 .884 .958 r
.84341 .47395 .8483 .47551 .86167 .47231 .85679 .4701 Metetra
.531 .886 .981 r
.85679 .4701 .86167 .47231 .87514 .46931 .87024 .46628 Metetra
.571 .877 .993 r
.87024 .46628 .87514 .46931 .8887 .46649 .88375 .46247 Metetra
.604 .854 .991 r
.88375 .46247 .8887 .46649 .90235 .46381 .89734 .45868 Metetra
.627 .821 .975 r
.89734 .45868 .90235 .46381 .91608 .46121 .91099 .4549 Metetra
.64 .784 .951 r
.91099 .4549 .91608 .46121 .9299 .45865 .92471 .4511 Metetra
.622 .743 .933 r
.31436 .62563 .32139 .63176 .33248 .62816 .32541 .62315 Metetra
.612 .779 .961 r
.32541 .62315 .33248 .62816 .34362 .62455 .33652 .62063 Metetra
.596 .816 .984 r
.33652 .62063 .34362 .62455 .3548 .62095 .34769 .61804 Metetra
.571 .847 .996 r
.34769 .61804 .3548 .62095 .36602 .61736 .35891 .61535 Metetra
.538 .868 .994 r
.35891 .61535 .36602 .61736 .37729 .61379 .37018 .61253 Metetra
.502 .876 .977 r
.37018 .61253 .37729 .61379 .3886 .61025 .38152 .60958 Metetra
.47 .876 .952 r
.38152 .60958 .3886 .61025 .39996 .60675 .39291 .60646 Metetra
.451 .874 .932 r
.39291 .60646 .39996 .60675 .41137 .60328 .40436 .60318 Metetra
.45 .876 .922 r
.40436 .60318 .41137 .60328 .42282 .59985 .41586 .59972 Metetra
.467 .884 .926 r
.41586 .59972 .42282 .59985 .43432 .59645 .42742 .59609 Metetra
.5 .896 .942 r
.42742 .59609 .43432 .59645 .44587 .5931 .43903 .59229 Metetra
.544 .906 .964 r
.43903 .59229 .44587 .5931 .45746 .58977 .45068 .58834 Metetra
.589 .908 .981 r
.45068 .58834 .45746 .58977 .46911 .58647 .46239 .58426 Metetra
.628 .896 .987 r
.46239 .58426 .46911 .58647 .48081 .58319 .47415 .58006 Metetra
.656 .871 .98 r
.47415 .58006 .48081 .58319 .49255 .57991 .48595 .57577 Metetra
.673 .838 .963 r
.48595 .57577 .49255 .57991 .50436 .57664 .49781 .57142 Metetra
.68 .802 .94 r
.49781 .57142 .50436 .57664 .51621 .57336 .5097 .56704 Metetra
.681 .767 .916 r
.5097 .56704 .51621 .57336 .52812 .57005 .52164 .56267 Metetra
.679 .736 .894 r
.52164 .56267 .52812 .57005 .54008 .56671 .53363 .55832 Metetra
.674 .709 .874 r
.53363 .55832 .54008 .56671 .55209 .56334 .54567 .55403 Metetra
.669 .686 .858 r
.54567 .55403 .55209 .56334 .56416 .55992 .55776 .54982 Metetra
.663 .668 .845 r
.55776 .54982 .56416 .55992 .57628 .55644 .56989 .54572 Metetra
.658 .655 .836 r
.56989 .54572 .57628 .55644 .58846 .55289 .58208 .54174 Metetra
.653 .646 .83 r
.58208 .54174 .58846 .55289 .60069 .54929 .59433 .53789 Metetra
.649 .64 .828 r
.59433 .53789 .60069 .54929 .61297 .54561 .60663 .53418 Metetra
.646 .639 .829 r
.60663 .53418 .61297 .54561 .6253 .54187 .61899 .53061 Metetra
.643 .641 .833 r
.61899 .53061 .6253 .54187 .63769 .53806 .63141 .52717 Metetra
.641 .648 .841 r
.63141 .52717 .63769 .53806 .65013 .53419 .64389 .52386 Metetra
.639 .658 .853 r
.64389 .52386 .65013 .53419 .66262 .53026 .65643 .52065 Metetra
.637 .674 .868 r
.65643 .52065 .66262 .53026 .67516 .52628 .66905 .51753 Metetra
.635 .694 .888 r
.66905 .51753 .67516 .52628 .68776 .52225 .68172 .51447 Metetra
.632 .72 .91 r
.68172 .51447 .68776 .52225 .70041 .51819 .69446 .51144 Metetra
.626 .75 .936 r
.69446 .51144 .70041 .51819 .71312 .51411 .70727 .50841 Metetra
.617 .783 .961 r
.70727 .50841 .71312 .51411 .72588 .51002 .72014 .50536 Metetra
.602 .817 .983 r
.72014 .50536 .72588 .51002 .73869 .50592 .73307 .50224 Metetra
.578 .847 .995 r
.73307 .50224 .73869 .50592 .75157 .50183 .74606 .49904 Metetra
.547 .868 .995 r
.74606 .49904 .75157 .50183 .7645 .49776 .75911 .49572 Metetra
.512 .878 .98 r
.75911 .49572 .7645 .49776 .77749 .49371 .77221 .49226 Metetra
.479 .878 .957 r
.77221 .49226 .77749 .49371 .79055 .48969 .78536 .48864 Metetra
.456 .875 .935 r
.78536 .48864 .79055 .48969 .80367 .4857 .79857 .48485 Metetra
.449 .876 .922 r
.79857 .48485 .80367 .4857 .81685 .48174 .81182 .48088 Metetra
.462 .882 .923 r
.81182 .48088 .81685 .48174 .83009 .47783 .82512 .47673 Metetra
.492 .894 .938 r
.82512 .47673 .83009 .47783 .84341 .47395 .83846 .47239 Metetra
.534 .905 .959 r
.83846 .47239 .84341 .47395 .85679 .4701 .85185 .46788 Metetra
.581 .908 .978 r
.85185 .46788 .85679 .4701 .87024 .46628 .86528 .46322 Metetra
.623 .898 .986 r
.86528 .46322 .87024 .46628 .88375 .46247 .87876 .45843 Metetra
.654 .874 .981 r
.87876 .45843 .88375 .46247 .89734 .45868 .89228 .45352 Metetra
.673 .839 .963 r
.89228 .45352 .89734 .45868 .91099 .4549 .90585 .44853 Metetra
.681 .801 .939 r
.90585 .44853 .91099 .4549 .92471 .4511 .91948 .44349 Metetra
.568 .726 .948 r
.30728 .61945 .31436 .62563 .32541 .62315 .31829 .61808 Metetra
.551 .76 .975 r
.31829 .61808 .32541 .62315 .33652 .62063 .32936 .61664 Metetra
.529 .794 .994 r
.32936 .61664 .33652 .62063 .34769 .61804 .34051 .61506 Metetra
.501 .822 1 r
.34051 .61506 .34769 .61804 .35891 .61535 .35172 .61326 Metetra
.471 .841 .991 r
.35172 .61326 .35891 .61535 .37018 .61253 .363 .6112 Metetra
.447 .851 .97 r
.363 .6112 .37018 .61253 .38152 .60958 .37436 .60882 Metetra
.433 .857 .947 r
.37436 .60882 .38152 .60958 .39291 .60646 .38578 .6061 Metetra
.436 .865 .931 r
.38578 .6061 .39291 .60646 .40436 .60318 .39727 .603 Metetra
.458 .879 .927 r
.39727 .603 .40436 .60318 .41586 .59972 .40882 .59951 Metetra
.495 .896 .936 r
.40882 .59951 .41586 .59972 .42742 .59609 .42044 .59564 Metetra
.544 .913 .952 r
.42044 .59564 .42742 .59609 .43903 .59229 .43211 .59141 Metetra
.596 .923 .968 r
.43211 .59141 .43903 .59229 .45068 .58834 .44384 .58683 Metetra
.643 .921 .977 r
.44384 .58683 .45068 .58834 .46239 .58426 .45561 .58196 Metetra
.679 .905 .975 r
.45561 .58196 .46239 .58426 .47415 .58006 .46743 .57685 Metetra
.701 .878 .964 r
.46743 .57685 .47415 .58006 .48595 .57577 .47929 .57156 Metetra
.712 .844 .945 r
.47929 .57156 .48595 .57577 .49781 .57142 .4912 .56614 Metetra
.715 .809 .923 r
.4912 .56614 .49781 .57142 .5097 .56704 .50314 .56068 Metetra
.711 .775 .902 r
.50314 .56068 .5097 .56704 .52164 .56267 .51512 .55524 Metetra
.704 .743 .882 r
.51512 .55524 .52164 .56267 .53363 .55832 .52714 .54989 Metetra
.696 .716 .865 r
.52714 .54989 .53363 .55832 .54567 .55403 .53921 .54469 Metetra
.686 .693 .851 r
.53921 .54469 .54567 .55403 .55776 .54982 .55131 .53971 Metetra
.676 .675 .841 r
.55131 .53971 .55776 .54982 .56989 .54572 .56347 .53499 Metetra
.667 .66 .833 r
.56347 .53499 .56989 .54572 .58208 .54174 .57567 .53057 Metetra
.658 .649 .829 r
.57567 .53057 .58208 .54174 .59433 .53789 .58793 .52648 Metetra
.65 .641 .828 r
.58793 .52648 .59433 .53789 .60663 .53418 .60025 .52274 Metetra
.642 .638 .831 r
.60025 .52274 .60663 .53418 .61899 .53061 .61264 .51934 Metetra
.634 .638 .837 r
.61264 .51934 .61899 .53061 .63141 .52717 .62509 .51627 Metetra
.627 .642 .846 r
.62509 .51627 .63141 .52717 .64389 .52386 .63761 .51351 Metetra
.619 .651 .859 r
.63761 .51351 .64389 .52386 .65643 .52065 .65021 .51101 Metetra
.612 .664 .876 r
.65021 .51101 .65643 .52065 .66905 .51753 .66289 .50874 Metetra
.604 .682 .896 r
.66289 .50874 .66905 .51753 .68172 .51447 .67564 .50663 Metetra
.595 .705 .92 r
.67564 .50663 .68172 .51447 .69446 .51144 .68846 .50462 Metetra
.584 .733 .945 r
.68846 .50462 .69446 .51144 .70727 .50841 .70137 .50264 Metetra
.57 .764 .97 r
.70137 .50264 .70727 .50841 .72014 .50536 .71434 .50062 Metetra
.55 .796 .99 r
.71434 .50062 .72014 .50536 .73307 .50224 .72738 .49848 Metetra
.526 .824 .999 r
.72738 .49848 .73307 .50224 .74606 .49904 .74048 .49616 Metetra
.498 .844 .995 r
.74048 .49616 .74606 .49904 .75911 .49572 .75364 .49359 Metetra
.471 .856 .979 r
.75364 .49359 .75911 .49572 .77221 .49226 .76685 .49072 Metetra
.451 .863 .957 r
.76685 .49072 .77221 .49226 .78536 .48864 .78011 .48751 Metetra
.445 .868 .937 r
.78011 .48751 .78536 .48864 .79857 .48485 .7934 .48392 Metetra
.456 .878 .927 r
.7934 .48392 .79857 .48485 .81182 .48088 .80672 .47993 Metetra
.485 .892 .93 r
.80672 .47993 .81182 .48088 .82512 .47673 .82007 .47553 Metetra
.528 .909 .944 r
.82007 .47553 .82512 .47673 .83846 .47239 .83345 .47074 Metetra
.579 .922 .961 r
.83345 .47074 .83846 .47239 .85185 .46788 .84685 .46558 Metetra
.629 .925 .974 r
.84685 .46558 .85185 .46788 .86528 .46322 .86026 .46008 Metetra
.671 .913 .976 r
.86026 .46008 .86528 .46322 .87876 .45843 .8737 .45429 Metetra
.698 .887 .967 r
.8737 .45429 .87876 .45843 .89228 .45352 .88717 .44827 Metetra
.713 .852 .948 r
.88717 .44827 .89228 .45352 .90585 .44853 .90067 .44209 Metetra
.717 .813 .925 r
.90067 .44209 .90585 .44853 .91948 .44349 .9142 .43581 Metetra
.504 .702 .959 r
.30014 .61322 .30728 .61945 .31829 .61808 .3111 .61294 Metetra
.479 .733 .982 r
.3111 .61294 .31829 .61808 .32936 .61664 .32214 .61256 Metetra
.451 .762 .996 r
.32214 .61256 .32936 .61664 .34051 .61506 .33326 .61195 Metetra
.424 .787 .996 r
.33326 .61195 .34051 .61506 .35172 .61326 .34446 .61103 Metetra
.404 .806 .983 r
.34446 .61103 .35172 .61326 .363 .6112 .35575 .60971 Metetra
.395 .822 .963 r
.35575 .60971 .363 .6112 .37436 .60882 .36712 .6079 Metetra
.404 .839 .946 r
.36712 .6079 .37436 .60882 .38578 .6061 .37857 .60556 Metetra
.43 .86 .937 r
.37857 .60556 .38578 .6061 .39727 .603 .3901 .60264 Metetra
.474 .884 .939 r
.3901 .60264 .39727 .603 .40882 .59951 .40171 .59912 Metetra
.529 .908 .948 r
.40171 .59912 .40882 .59951 .42044 .59564 .41338 .59502 Metetra
.589 .926 .96 r
.41338 .59502 .42044 .59564 .43211 .59141 .42512 .59035 Metetra
.644 .933 .969 r
.42512 .59035 .43211 .59141 .44384 .58683 .43691 .58516 Metetra
.689 .927 .969 r
.43691 .58516 .44384 .58683 .45561 .58196 .44876 .57953 Metetra
.72 .907 .96 r
.44876 .57953 .45561 .58196 .46743 .57685 .46064 .57352 Metetra
.738 .879 .945 r
.46064 .57352 .46743 .57685 .47929 .57156 .47257 .56723 Metetra
.744 .846 .926 r
.47257 .56723 .47929 .57156 .4912 .56614 .48453 .56077 Metetra
.743 .812 .906 r
.48453 .56077 .4912 .56614 .50314 .56068 .49652 .55425 Metetra
.736 .779 .888 r
.49652 .55425 .50314 .56068 .51512 .55524 .50855 .54777 Metetra
.726 .75 .871 r
.50855 .54777 .51512 .55524 .52714 .54989 .52061 .54144 Metetra
.715 .723 .857 r
.52061 .54144 .52714 .54989 .53921 .54469 .5327 .53537 Metetra
.702 .7 .845 r
.5327 .53537 .53921 .54469 .55131 .53971 .54483 .52963 Metetra
.689 .681 .837 r
.54483 .52963 .55131 .53971 .56347 .53499 .55701 .52431 Metetra
.676 .665 .832 r
.55701 .52431 .56347 .53499 .57567 .53057 .56923 .51947 Metetra
.663 .653 .83 r
.56923 .51947 .57567 .53057 .58793 .52648 .58151 .51515 Metetra
.65 .644 .831 r
.58151 .51515 .58793 .52648 .60025 .52274 .59385 .51137 Metetra
.637 .638 .835 r
.59385 .51137 .60025 .52274 .61264 .51934 .60625 .50813 Metetra
.625 .636 .842 r
.60625 .50813 .61264 .51934 .62509 .51627 .61874 .50542 Metetra
.612 .638 .852 r
.61874 .50542 .62509 .51627 .63761 .51351 .6313 .50319 Metetra
.599 .643 .866 r
.6313 .50319 .63761 .51351 .65021 .51101 .64395 .50139 Metetra
.585 .653 .883 r
.64395 .50139 .65021 .51101 .66289 .50874 .65668 .49994 Metetra
.571 .668 .904 r
.65668 .49994 .66289 .50874 .67564 .50663 .6695 .49875 Metetra
.556 .688 .928 r
.6695 .49875 .67564 .50663 .68846 .50462 .68241 .49774 Metetra
.538 .713 .952 r
.68241 .49774 .68846 .50462 .70137 .50264 .69541 .49677 Metetra
.519 .74 .975 r
.69541 .49677 .70137 .50264 .71434 .50062 .70848 .49576 Metetra
.496 .769 .992 r
.70848 .49576 .71434 .50062 .72738 .49848 .72163 .49457 Metetra
.473 .796 .999 r
.72163 .49457 .72738 .49848 .74048 .49616 .73484 .49312 Metetra
.45 .817 .994 r
.73484 .49312 .74048 .49616 .75364 .49359 .74811 .49129 Metetra
.435 .834 .979 r
.74811 .49129 .75364 .49359 .76685 .49072 .76142 .489 Metetra
.431 .848 .96 r
.76142 .489 .76685 .49072 .78011 .48751 .77478 .48618 Metetra
.442 .864 .945 r
.77478 .48618 .78011 .48751 .7934 .48392 .78815 .48278 Metetra
.471 .882 .939 r
.78815 .48278 .7934 .48392 .80672 .47993 .80155 .47877 Metetra
.514 .903 .942 r
.80155 .47877 .80672 .47993 .82007 .47553 .81495 .47414 Metetra
.568 .923 .952 r
.81495 .47414 .82007 .47553 .83345 .47074 .82836 .46891 Metetra
.624 .935 .962 r
.82836 .46891 .83345 .47074 .84685 .46558 .84177 .4631 Metetra
.674 .935 .968 r
.84177 .4631 .84685 .46558 .86026 .46008 .85518 .45677 Metetra
.713 .921 .964 r
.85518 .45677 .86026 .46008 .8737 .45429 .86858 .45001 Metetra
.736 .894 .951 r
.86858 .45001 .8737 .45429 .88717 .44827 .882 .4429 Metetra
.747 .859 .931 r
.882 .4429 .88717 .44827 .90067 .44209 .89543 .43555 Metetra
.748 .822 .909 r
.89543 .43555 .90067 .44209 .9142 .43581 .90888 .42806 Metetra
.43 .671 .963 r
.29294 .60694 .30014 .61322 .3111 .61294 .30386 .6077 Metetra
.397 .695 .981 r
.30386 .6077 .3111 .61294 .32214 .61256 .31487 .60833 Metetra
.367 .72 .988 r
.31487 .60833 .32214 .61256 .33326 .61195 .32596 .60867 Metetra
.346 .744 .984 r
.32596 .60867 .33326 .61195 .34446 .61103 .33715 .60859 Metetra
.339 .768 .972 r
.33715 .60859 .34446 .61103 .35575 .60971 .34843 .60797 Metetra
.351 .794 .958 r
.34843 .60797 .35575 .60971 .36712 .6079 .35982 .60671 Metetra
.383 .824 .95 r
.35982 .60671 .36712 .6079 .37857 .60556 .3713 .60474 Metetra
.433 .857 .948 r
.3713 .60474 .37857 .60556 .3901 .60264 .38287 .60199 Metetra
.497 .889 .954 r
.38287 .60199 .3901 .60264 .40171 .59912 .39452 .59844 Metetra
.566 .917 .962 r
.39452 .59844 .40171 .59912 .41338 .59502 .40626 .59411 Metetra
.631 .933 .967 r
.40626 .59411 .41338 .59502 .42512 .59035 .41806 .58903 Metetra
.686 .935 .966 r
.41806 .58903 .42512 .59035 .43691 .58516 .42992 .58325 Metetra
.727 .924 .958 r
.42992 .58325 .43691 .58516 .44876 .57953 .44184 .57687 Metetra
.753 .903 .944 r
.44184 .57687 .44876 .57953 .46064 .57352 .45379 .57 Metetra
.766 .875 .926 r
.45379 .57 .46064 .57352 .47257 .56723 .46579 .56276 Metetra
.769 .844 .908 r
.46579 .56276 .47257 .56723 .48453 .56077 .47781 .5553 Metetra
.765 .812 .89 r
.47781 .5553 .48453 .56077 .49652 .55425 .48985 .54776 Metetra
.757 .782 .874 r
.48985 .54776 .49652 .55425 .50855 .54777 .50193 .54028 Metetra
.745 .755 .86 r
.50193 .54028 .50855 .54777 .52061 .54144 .51403 .53302 Metetra
.731 .73 .849 r
.51403 .53302 .52061 .54144 .5327 .53537 .52615 .5261 Metetra
.717 .708 .841 r
.52615 .5261 .5327 .53537 .54483 .52963 .53831 .51964 Metetra
.701 .688 .835 r
.53831 .51964 .54483 .52963 .55701 .52431 .55051 .51375 Metetra
.684 .672 .832 r
.55051 .51375 .55701 .52431 .56923 .51947 .56275 .5085 Metetra
.668 .658 .832 r
.56275 .5085 .56923 .51947 .58151 .51515 .57505 .50395 Metetra
.651 .648 .834 r
.57505 .50395 .58151 .51515 .59385 .51137 .58741 .50014 Metetra
.633 .64 .839 r
.58741 .50014 .59385 .51137 .60625 .50813 .59984 .49705 Metetra
.615 .635 .848 r
.59984 .49705 .60625 .50813 .61874 .50542 .61235 .49467 Metetra
.597 .634 .859 r
.61235 .49467 .61874 .50542 .6313 .50319 .62495 .49295 Metetra
.577 .636 .873 r
.62495 .49295 .6313 .50319 .64395 .50139 .63765 .49181 Metetra
.557 .643 .891 r
.63765 .49181 .64395 .50139 .65668 .49994 .65044 .49115 Metetra
.536 .654 .911 r
.65044 .49115 .65668 .49994 .6695 .49875 .66333 .49085 Metetra
.513 .67 .934 r
.66333 .49085 .6695 .49875 .68241 .49774 .67631 .49078 Metetra
.49 .689 .956 r
.67631 .49078 .68241 .49774 .69541 .49677 .68939 .49079 Metetra
.465 .713 .976 r
.68939 .49079 .69541 .49677 .70848 .49576 .70256 .49074 Metetra
.442 .738 .99 r
.70256 .49074 .70848 .49576 .72163 .49457 .71581 .49047 Metetra
.421 .764 .996 r
.71581 .49047 .72163 .49457 .73484 .49312 .72912 .48985 Metetra
.408 .788 .992 r
.72912 .48985 .73484 .49312 .74811 .49129 .7425 .48873 Metetra
.406 .812 .98 r
.7425 .48873 .74811 .49129 .76142 .489 .75591 .48699 Metetra
.419 .835 .967 r
.75591 .48699 .76142 .489 .77478 .48618 .76936 .48456 Metetra
.448 .861 .957 r
.76936 .48456 .77478 .48618 .78815 .48278 .78282 .48135 Metetra
.492 .887 .954 r
.78282 .48135 .78815 .48278 .80155 .47877 .79628 .47731 Metetra
.548 .913 .956 r
.79628 .47731 .80155 .47877 .81495 .47414 .80974 .47245 Metetra
.609 .933 .96 r
.80974 .47245 .81495 .47414 .82836 .46891 .82318 .46678 Metetra
.666 .943 .962 r
.82318 .46678 .82836 .46891 .84177 .4631 .83661 .46035 Metetra
.714 .939 .959 r
.83661 .46035 .84177 .4631 .85518 .45677 .85001 .45323 Metetra
.748 .922 .949 r
.85001 .45323 .85518 .45677 .86858 .45001 .86339 .44553 Metetra
.768 .895 .933 r
.86339 .44553 .86858 .45001 .882 .4429 .87676 .43737 Metetra
.775 .862 .914 r
.87676 .43737 .882 .4429 .89543 .43555 .89013 .4289 Metetra
.774 .827 .894 r
.89013 .4289 .89543 .43555 .90888 .42806 .90351 .42024 Metetra
.347 .632 .959 r
.28569 .60061 .29294 .60694 .30386 .6077 .29656 .60235 Metetra
.309 .649 .969 r
.29656 .60235 .30386 .6077 .31487 .60833 .30753 .60393 Metetra
.281 .67 .971 r
.30753 .60393 .31487 .60833 .32596 .60867 .31859 .60515 Metetra
.27 .696 .967 r
.31859 .60515 .32596 .60867 .33715 .60859 .32977 .60587 Metetra
.281 .728 .96 r
.32977 .60587 .33715 .60859 .34843 .60797 .34105 .60591 Metetra
.316 .767 .956 r
.34105 .60591 .34843 .60797 .35982 .60671 .35245 .60517 Metetra
.372 .81 .958 r
.35245 .60517 .35982 .60671 .3713 .60474 .36395 .60354 Metetra
.444 .854 .963 r
.36395 .60354 .3713 .60474 .38287 .60199 .37556 .60095 Metetra
.524 .893 .97 r
.37556 .60095 .38287 .60199 .39452 .59844 .38727 .59738 Metetra
.601 .92 .973 r
.38727 .59738 .39452 .59844 .40626 .59411 .39906 .59283 Metetra
.669 .933 .969 r
.39906 .59283 .40626 .59411 .41806 .58903 .41093 .58735 Metetra
.72 .93 .959 r
.41093 .58735 .41806 .58903 .42992 .58325 .42287 .58101 Metetra
.756 .916 .944 r
.42287 .58101 .42992 .58325 .44184 .57687 .43485 .57393 Metetra
.777 .894 .926 r
.43485 .57393 .44184 .57687 .45379 .57 .44688 .56624 Metetra
.787 .867 .908 r
.44688 .56624 .45379 .57 .46579 .56276 .45894 .55811 Metetra
.788 .839 .89 r
.45894 .55811 .46579 .56276 .47781 .5553 .47102 .5497 Metetra
.783 .811 .875 r
.47102 .5497 .47781 .5553 .48985 .54776 .48313 .54119 Metetra
.774 .784 .861 r
.48313 .54119 .48985 .54776 .50193 .54028 .49525 .53278 Metetra
.761 .759 .851 r
.49525 .53278 .50193 .54028 .51403 .53302 .50739 .52463 Metetra
.746 .736 .842 r
.50739 .52463 .51403 .53302 .52615 .5261 .51956 .51692 Metetra
.73 .715 .837 r
.51956 .51692 .52615 .5261 .53831 .51964 .53175 .50979 Metetra
.712 .696 .834 r
.53175 .50979 .53831 .51964 .55051 .51375 .54397 .50335 Metetra
.693 .68 .833 r
.54397 .50335 .55051 .51375 .56275 .5085 .55624 .49773 Metetra
.673 .665 .834 r
.55624 .49773 .56275 .5085 .57505 .50395 .56855 .49297 Metetra
.651 .653 .839 r
.56855 .49297 .57505 .50395 .58741 .50014 .58094 .48911 Metetra
.629 .643 .845 r
.58094 .48911 .58741 .50014 .59984 .49705 .59339 .48617 Metetra
.606 .636 .855 r
.59339 .48617 .59984 .49705 .61235 .49467 .60594 .48411 Metetra
.581 .632 .867 r
.60594 .48411 .61235 .49467 .62495 .49295 .61857 .48286 Metetra
.555 .63 .881 r
.61857 .48286 .62495 .49295 .63765 .49181 .63131 .48233 Metetra
.527 .633 .898 r
.63131 .48233 .63765 .49181 .65044 .49115 .64415 .4824 Metetra
.498 .639 .917 r
.64415 .4824 .65044 .49115 .66333 .49085 .6571 .48293 Metetra
.469 .649 .937 r
.6571 .48293 .66333 .49085 .67631 .49078 .67016 .48375 Metetra
.439 .664 .957 r
.67016 .48375 .67631 .49078 .68939 .49079 .68332 .48467 Metetra
.412 .683 .973 r
.68332 .48467 .68939 .49079 .70256 .49074 .69657 .48553 Metetra
.389 .705 .985 r
.69657 .48553 .70256 .49074 .71581 .49047 .70991 .48612 Metetra
.374 .731 .99 r
.70991 .48612 .71581 .49047 .72912 .48985 .72333 .48628 Metetra
.372 .759 .988 r
.72333 .48628 .72912 .48985 .7425 .48873 .7368 .48582 Metetra
.385 .79 .983 r
.7368 .48582 .7425 .48873 .75591 .48699 .75031 .48461 Metetra
.415 .823 .977 r
.75031 .48461 .75591 .48699 .76936 .48456 .76384 .48254 Metetra
.46 .858 .972 r
.76384 .48254 .76936 .48456 .78282 .48135 .77739 .4795 Metetra
.519 .891 .969 r
.77739 .4795 .78282 .48135 .79628 .47731 .79092 .47544 Metetra
.583 .919 .968 r
.79092 .47544 .79628 .47731 .80974 .47245 .80443 .47036 Metetra
.647 .937 .966 r
.80443 .47036 .80974 .47245 .82318 .46678 .81792 .46428 Metetra
.703 .943 .959 r
.81792 .46428 .82318 .46678 .83661 .46035 .83136 .45725 Metetra
.747 .936 .948 r
.83136 .45725 .83661 .46035 .85001 .45323 .84476 .44938 Metetra
.776 .918 .933 r
.84476 .44938 .85001 .45323 .86339 .44553 .85813 .4408 Metetra
.792 .892 .915 r
.85813 .4408 .86339 .44553 .87676 .43737 .87146 .43165 Metetra
.798 .862 .897 r
.87146 .43165 .87676 .43737 .89013 .4289 .88478 .42211 Metetra
.795 .83 .879 r
.88478 .42211 .89013 .4289 .90351 .42024 .89809 .41236 Metetra
.26 .587 .946 r
.27838 .59423 .28569 .60061 .29656 .60235 .28921 .59687 Metetra
.22 .598 .949 r
.28921 .59687 .29656 .60235 .30753 .60393 .30014 .59932 Metetra
.199 .618 .948 r
.30014 .59932 .30753 .60393 .31859 .60515 .31117 .60136 Metetra
.203 .649 .947 r
.31117 .60136 .31859 .60515 .32977 .60587 .32233 .6028 Metetra
.234 .691 .95 r
.32233 .6028 .32977 .60587 .34105 .60591 .33361 .60346 Metetra
.291 .743 .958 r
.33361 .60346 .34105 .60591 .35245 .60517 .34502 .60319 Metetra
.37 .799 .969 r
.34502 .60319 .35245 .60517 .36395 .60354 .35655 .60187 Metetra
.461 .851 .978 r
.35655 .60187 .36395 .60354 .37556 .60095 .3682 .59943 Metetra
.552 .893 .983 r
.3682 .59943 .37556 .60095 .38727 .59738 .37996 .59583 Metetra
.634 .917 .979 r
.37996 .59583 .38727 .59738 .39906 .59283 .39181 .59108 Metetra
.699 .925 .967 r
.39181 .59108 .39906 .59283 .41093 .58735 .40374 .58523 Metetra
.746 .919 .949 r
.40374 .58523 .41093 .58735 .42287 .58101 .41574 .57837 Metetra
.777 .903 .929 r
.41574 .57837 .42287 .58101 .43485 .57393 .4278 .57064 Metetra
.795 .882 .909 r
.4278 .57064 .43485 .57393 .44688 .56624 .4399 .56219 Metetra
.803 .858 .891 r
.4399 .56219 .44688 .56624 .45894 .55811 .45203 .55323 Metetra
.803 .833 .874 r
.45203 .55323 .45894 .55811 .47102 .5497 .46418 .54394 Metetra
.797 .808 .861 r
.46418 .54394 .47102 .5497 .48313 .54119 .47635 .53455 Metetra
.787 .785 .85 r
.47635 .53455 .48313 .54119 .49525 .53278 .48853 .52528 Metetra
.774 .762 .842 r
.48853 .52528 .49525 .53278 .50739 .52463 .50071 .51632 Metetra
.759 .742 .837 r
.50071 .51632 .50739 .52463 .51956 .51692 .51292 .50787 Metetra
.742 .722 .834 r
.51292 .50787 .51956 .51692 .53175 .50979 .52514 .50012 Metetra
.722 .704 .833 r
.52514 .50012 .53175 .50979 .54397 .50335 .53739 .49319 Metetra
.701 .688 .835 r
.53739 .49319 .54397 .50335 .55624 .49773 .54969 .48721 Metetra
.677 .673 .839 r
.54969 .48721 .55624 .49773 .56855 .49297 .56203 .48225 Metetra
.652 .66 .845 r
.56203 .48225 .56855 .49297 .58094 .48911 .57443 .47836 Metetra
.625 .648 .853 r
.57443 .47836 .58094 .48911 .59339 .48617 .58691 .47555 Metetra
.596 .638 .863 r
.58691 .47555 .59339 .48617 .60594 .48411 .59948 .47376 Metetra
.565 .63 .875 r
.59948 .47376 .60594 .48411 .61857 .48286 .61215 .47294 Metetra
.531 .625 .89 r
.61215 .47294 .61857 .48286 .63131 .48233 .62493 .47298 Metetra
.496 .622 .905 r
.62493 .47298 .63131 .48233 .64415 .4824 .63782 .47372 Metetra
.46 .623 .922 r
.63782 .47372 .64415 .4824 .6571 .48293 .65083 .475 Metetra
.424 .628 .938 r
.65083 .475 .6571 .48293 .67016 .48375 .66395 .47663 Metetra
.39 .637 .954 r
.66395 .47663 .67016 .48375 .68332 .48467 .67718 .4784 Metetra
.361 .652 .967 r
.67718 .4784 .68332 .48467 .69657 .48553 .69052 .48008 Metetra
.342 .672 .976 r
.69052 .48008 .69657 .48553 .70991 .48612 .70394 .48147 Metetra
.335 .699 .982 r
.70394 .48147 .70991 .48612 .72333 .48628 .71745 .48234 Metetra
.344 .732 .985 r
.71745 .48234 .72333 .48628 .7368 .48582 .73101 .48249 Metetra
.372 .77 .986 r
.73101 .48249 .7368 .48582 .75031 .48461 .74461 .48177 Metetra
.418 .811 .986 r
.74461 .48177 .75031 .48461 .76384 .48254 .75823 .48002 Metetra
.478 .853 .985 r
.75823 .48002 .76384 .48254 .77739 .4795 .77185 .47713 Metetra
.547 .89 .983 r
.77185 .47713 .77739 .4795 .79092 .47544 .78546 .47306 Metetra
.617 .918 .977 r
.78546 .47306 .79092 .47544 .80443 .47036 .79902 .46777 Metetra
.681 .934 .968 r
.79902 .46777 .80443 .47036 .81792 .46428 .81255 .4613 Metetra
.733 .937 .954 r
.81255 .4613 .81792 .46428 .83136 .45725 .82601 .45373 Metetra
.772 .928 .936 r
.82601 .45373 .83136 .45725 .84476 .44938 .83942 .44516 Metetra
.798 .91 .917 r
.83942 .44516 .84476 .44938 .85813 .4408 .85278 .43575 Metetra
.811 .886 .899 r
.85278 .43575 .85813 .4408 .87146 .43165 .86609 .42568 Metetra
.815 .859 .881 r
.86609 .42568 .87146 .43165 .88478 .42211 .87936 .41516 Metetra
.812 .831 .865 r
.87936 .41516 .88478 .42211 .89809 .41236 .89262 .40441 Metetra
.174 .537 .925 r
.27102 .5878 .27838 .59423 .28921 .59687 .2818 .59124 Metetra
.138 .545 .922 r
.2818 .59124 .28921 .59687 .30014 .59932 .29269 .59447 Metetra
.127 .567 .923 r
.29269 .59447 .30014 .59932 .31117 .60136 .3037 .59724 Metetra
.147 .605 .929 r
.3037 .59724 .31117 .60136 .32233 .6028 .31484 .59933 Metetra
.199 .659 .943 r
.31484 .59933 .32233 .6028 .33361 .60346 .32612 .60053 Metetra
.278 .723 .962 r
.32612 .60053 .33361 .60346 .34502 .60319 .33754 .60069 Metetra
.377 .789 .98 r
.33754 .60069 .34502 .60319 .35655 .60187 .3491 .59965 Metetra
.482 .846 .991 r
.3491 .59965 .35655 .60187 .3682 .59943 .36078 .59734 Metetra
.58 .886 .991 r
.36078 .59734 .3682 .59943 .37996 .59583 .37258 .59371 Metetra
.661 .907 .979 r
.37258 .59371 .37996 .59583 .39181 .59108 .38449 .58877 Metetra
.722 .911 .96 r
.38449 .58877 .39181 .59108 .40374 .58523 .39649 .58258 Metetra
.765 .903 .937 r
.39649 .58258 .40374 .58523 .41574 .57837 .40856 .57525 Metetra
.792 .888 .914 r
.40856 .57525 .41574 .57837 .4278 .57064 .42069 .56693 Metetra
.807 .868 .893 r
.42069 .56693 .4278 .57064 .4399 .56219 .43286 .55781 Metetra
.813 .847 .875 r
.43286 .55781 .4399 .56219 .45203 .55323 .44506 .54809 Metetra
.813 .826 .86 r
.44506 .54809 .45203 .55323 .46418 .54394 .45728 .53802 Metetra
.807 .805 .849 r
.45728 .53802 .46418 .54394 .47635 .53455 .46951 .52783 Metetra
.798 .785 .84 r
.46951 .52783 .47635 .53455 .48853 .52528 .48174 .51778 Metetra
.786 .766 .835 r
.48174 .51778 .48853 .52528 .50071 .51632 .49398 .50809 Metetra
.77 .747 .832 r
.49398 .50809 .50071 .51632 .51292 .50787 .50623 .49899 Metetra
.752 .73 .832 r
.50623 .49899 .51292 .50787 .52514 .50012 .51849 .49067 Metetra
.732 .713 .834 r
.51849 .49067 .52514 .50012 .53739 .49319 .53078 .4833 Metetra
.708 .698 .838 r
.53078 .4833 .53739 .49319 .54969 .48721 .5431 .477 Metetra
.682 .682 .844 r
.5431 .477 .54969 .48721 .56203 .48225 .55546 .47186 Metetra
.653 .668 .852 r
.55546 .47186 .56203 .48225 .57443 .47836 .56789 .46794 Metetra
.622 .655 .861 r
.56789 .46794 .57443 .47836 .58691 .47555 .5804 .46523 Metetra
.587 .642 .872 r
.5804 .46523 .58691 .47555 .59948 .47376 .593 .4637 Metetra
.548 .63 .885 r
.593 .4637 .59948 .47376 .61215 .47294 .6057 .46326 Metetra
.508 .62 .898 r
.6057 .46326 .61215 .47294 .62493 .47298 .61851 .46378 Metetra
.465 .612 .912 r
.61851 .46378 .62493 .47298 .63782 .47372 .63145 .46512 Metetra
.422 .607 .925 r
.63145 .46512 .63782 .47372 .65083 .475 .6445 .46708 Metetra
.38 .606 .937 r
.6445 .46708 .65083 .475 .66395 .47663 .65768 .46943 Metetra
.344 .61 .948 r
.65768 .46943 .66395 .47663 .67718 .4784 .67098 .47194 Metetra
.316 .621 .958 r
.67098 .47194 .67718 .4784 .69052 .48008 .68439 .47437 Metetra
.301 .641 .966 r
.68439 .47437 .69052 .48008 .70394 .48147 .6979 .47646 Metetra
.303 .669 .974 r
.6979 .47646 .70394 .48147 .71745 .48234 .71148 .47797 Metetra
.325 .706 .981 r
.71148 .47797 .71745 .48234 .73101 .48249 .72513 .47867 Metetra
.367 .751 .988 r
.72513 .47867 .73101 .48249 .74461 .48177 .73881 .47837 Metetra
.427 .799 .994 r
.73881 .47837 .74461 .48177 .75823 .48002 .75252 .47691 Metetra
.498 .845 .995 r
.75252 .47691 .75823 .48002 .77185 .47713 .76621 .47417 Metetra
.574 .884 .991 r
.76621 .47417 .77185 .47713 .78546 .47306 .77988 .47007 Metetra
.646 .911 .981 r
.77988 .47007 .78546 .47306 .79902 .46777 .7935 .46459 Metetra
.708 .924 .965 r
.7935 .46459 .79902 .46777 .81255 .4613 .80707 .45777 Metetra
.756 .925 .945 r
.80707 .45777 .81255 .4613 .82601 .45373 .82056 .4497 Metetra
.791 .915 .923 r
.82056 .4497 .82601 .45373 .83942 .44516 .83399 .4405 Metetra
.813 .899 .902 r
.83399 .4405 .83942 .44516 .85278 .43575 .84734 .43034 Metetra
.825 .878 .883 r
.84734 .43034 .85278 .43575 .86609 .42568 .86063 .41945 Metetra
.828 .854 .866 r
.86063 .41945 .86609 .42568 .87936 .41516 .87388 .40805 Metetra
.825 .83 .852 r
.87388 .40805 .87936 .41516 .89262 .40441 .88711 .39638 Metetra
.094 .488 .899 r
.26359 .58132 .27102 .5878 .2818 .59124 .27434 .58545 Metetra
.066 .495 .894 r
.27434 .58545 .2818 .59124 .29269 .59447 .28519 .58934 Metetra
.069 .521 .898 r
.28519 .58934 .29269 .59447 .3037 .59724 .29618 .59274 Metetra
.107 .568 .914 r
.29618 .59274 .3037 .59724 .31484 .59933 .30731 .59539 Metetra
.178 .632 .939 r
.30731 .59539 .31484 .59933 .32612 .60053 .31859 .59707 Metetra
.276 .706 .967 r
.31859 .59707 .32612 .60053 .33754 .60069 .33002 .59759 Metetra
.39 .778 .989 r
.33002 .59759 .33754 .60069 .3491 .59965 .34159 .5968 Metetra
.504 .836 .999 r
.34159 .5968 .3491 .59965 .36078 .59734 .35331 .59459 Metetra
.603 .873 .992 r
.35331 .59459 .36078 .59734 .37258 .59371 .36516 .59093 Metetra
.682 .891 .974 r
.36516 .59093 .37258 .59371 .38449 .58877 .37713 .58583 Metetra
.738 .893 .949 r
.37713 .58583 .38449 .58877 .39649 .58258 .38919 .57934 Metetra
.777 .884 .923 r
.38919 .57934 .39649 .58258 .40856 .57525 .40133 .5716 Metetra
.801 .87 .899 r
.40133 .5716 .40856 .57525 .42069 .56693 .41353 .56276 Metetra
.814 .853 .878 r
.41353 .56276 .42069 .56693 .43286 .55781 .42577 .55304 Metetra
.82 .836 .861 r
.42577 .55304 .43286 .55781 .44506 .54809 .43804 .54267 Metetra
.82 .818 .848 r
.43804 .54267 .44506 .54809 .45728 .53802 .45033 .53191 Metetra
.815 .801 .838 r
.45033 .53191 .45728 .53802 .46951 .52783 .46262 .52103 Metetra
.806 .784 .832 r
.46262 .52103 .46951 .52783 .48174 .51778 .47491 .5103 Metetra
.795 .768 .829 r
.47491 .5103 .48174 .51778 .49398 .50809 .4872 .49998 Metetra
.78 .753 .829 r
.4872 .49998 .49398 .50809 .50623 .49899 .49949 .4903 Metetra
.762 .738 .831 r
.49949 .4903 .50623 .49899 .51849 .49067 .5118 .4815 Metetra
.74 .723 .835 r
.5118 .4815 .51849 .49067 .53078 .4833 .52412 .47373 Metetra
.716 .708 .842 r
.52412 .47373 .53078 .4833 .5431 .477 .53647 .46715 Metetra
.687 .693 .85 r
.53647 .46715 .5431 .477 .55546 .47186 .54886 .46186 Metetra
.655 .678 .86 r
.54886 .46186 .55546 .47186 .56789 .46794 .56132 .4579 Metetra
.618 .663 .871 r
.56132 .4579 .56789 .46794 .5804 .46523 .57385 .45528 Metetra
.577 .647 .883 r
.57385 .45528 .5804 .46523 .593 .4637 .58648 .45396 Metetra
.532 .632 .895 r
.58648 .45396 .593 .4637 .6057 .46326 .59921 .45384 Metetra
.484 .617 .907 r
.59921 .45384 .6057 .46326 .61851 .46378 .61206 .45479 Metetra
.434 .603 .918 r
.61206 .45479 .61851 .46378 .63145 .46512 .62503 .45663 Metetra
.384 .592 .927 r
.62503 .45663 .63145 .46512 .6445 .46708 .63813 .45916 Metetra
.339 .586 .935 r
.63813 .45916 .6445 .46708 .65768 .46943 .65136 .46213 Metetra
.302 .585 .942 r
.65136 .46213 .65768 .46943 .67098 .47194 .66472 .46529 Metetra
.277 .594 .949 r
.66472 .46529 .67098 .47194 .68439 .47437 .67819 .46835 Metetra
.27 .613 .957 r
.67819 .46835 .68439 .47437 .6979 .47646 .69177 .47105 Metetra
.282 .643 .966 r
.69177 .47105 .6979 .47646 .71148 .47797 .70543 .47311 Metetra
.316 .684 .978 r
.70543 .47311 .71148 .47797 .72513 .47867 .71915 .47428 Metetra
.37 .733 .989 r
.71915 .47428 .72513 .47867 .73881 .47837 .73291 .47436 Metetra
.44 .785 .998 r
.73291 .47436 .73881 .47837 .75252 .47691 .74669 .47314 Metetra
.518 .833 1 r
.74669 .47314 .75252 .47691 .76621 .47417 .76046 .47051 Metetra
.597 .871 .993 r
.76046 .47051 .76621 .47417 .77988 .47007 .77419 .46638 Metetra
.669 .896 .978 r
.77419 .46638 .77988 .47007 .7935 .46459 .78787 .46074 Metetra
.727 .908 .957 r
.78787 .46074 .7935 .46459 .80707 .45777 .80148 .45361 Metetra
.772 .908 .934 r
.80148 .45361 .80707 .45777 .82056 .4497 .815 .44509 Metetra
.804 .899 .91 r
.815 .44509 .82056 .4497 .83399 .4405 .82845 .43534 Metetra
.824 .885 .888 r
.82845 .43534 .83399 .4405 .84734 .43034 .84181 .42453 Metetra
.834 .868 .869 r
.84181 .42453 .84734 .43034 .86063 .41945 .85511 .41292 Metetra
.838 .848 .853 r
.85511 .41292 .86063 .41945 .87388 .40805 .86834 .40074 Metetra
.835 .828 .841 r
.86834 .40074 .87388 .40805 .88711 .39638 .88155 .38829 Metetra
.026 .444 .871 r
.25611 .57479 .26359 .58132 .27434 .58545 .26682 .57948 Metetra
.009 .452 .868 r
.26682 .57948 .27434 .58545 .28519 .58934 .27765 .58393 Metetra
.027 .484 .879 r
.27765 .58393 .28519 .58934 .29618 .59274 .28861 .58784 Metetra
.082 .539 .904 r
.28861 .58784 .29618 .59274 .30731 .59539 .29973 .59095 Metetra
.171 .612 .939 r
.29973 .59095 .30731 .59539 .31859 .59707 .31101 .59302 Metetra
.285 .692 .972 r
.31101 .59302 .31859 .59707 .33002 .59759 .32245 .59383 Metetra
.408 .765 .995 r
.32245 .59383 .33002 .59759 .34159 .5968 .33405 .59323 Metetra
.525 .821 .999 r
.33405 .59323 .34159 .5968 .35331 .59459 .34581 .59111 Metetra
.622 .855 .987 r
.34581 .59111 .35331 .59459 .36516 .59093 .3577 .58742 Metetra
.696 .87 .964 r
.3577 .58742 .36516 .59093 .37713 .58583 .36971 .58217 Metetra
.748 .87 .936 r
.36971 .58217 .37713 .58583 .38919 .57934 .38183 .57544 Metetra
.783 .863 .909 r
.38183 .57544 .38919 .57934 .40133 .5716 .39404 .56734 Metetra
.805 .852 .885 r
.39404 .56734 .40133 .5716 .41353 .56276 .40631 .55807 Metetra
.818 .838 .865 r
.40631 .55807 .41353 .56276 .42577 .55304 .41862 .54785 Metetra
.823 .824 .849 r
.41862 .54785 .42577 .55304 .43804 .54267 .43096 .53693 Metetra
.824 .811 .838 r
.43096 .53693 .43804 .54267 .45033 .53191 .44331 .5256 Metetra
.82 .797 .83 r
.44331 .5256 .45033 .53191 .46262 .52103 .45567 .51414 Metetra
.813 .784 .826 r
.45567 .51414 .46262 .52103 .47491 .5103 .46802 .50284 Metetra
.802 .771 .824 r
.46802 .50284 .47491 .5103 .4872 .49998 .48036 .49199 Metetra
.788 .758 .826 r
.48036 .49199 .4872 .49998 .49949 .4903 .4927 .48184 Metetra
.77 .746 .831 r
.4927 .48184 .49949 .4903 .5118 .4815 .50505 .47262 Metetra
.748 .733 .838 r
.50505 .47262 .5118 .4815 .52412 .47373 .51741 .46453 Metetra
.722 .719 .847 r
.51741 .46453 .52412 .47373 .53647 .46715 .52979 .45771 Metetra
.692 .705 .857 r
.52979 .45771 .53647 .46715 .54886 .46186 .54222 .45228 Metetra
.656 .689 .869 r
.54222 .45228 .54886 .46186 .56132 .4579 .5547 .44829 Metetra
.615 .672 .882 r
.5547 .44829 .56132 .4579 .57385 .45528 .56727 .44573 Metetra
.568 .654 .894 r
.56727 .44573 .57385 .45528 .58648 .45396 .57992 .44458 Metetra
.516 .635 .906 r
.57992 .44458 .58648 .45396 .59921 .45384 .59268 .44472 Metetra
.46 .615 .916 r
.59268 .44472 .59921 .45384 .61206 .45479 .60556 .44601 Metetra
.404 .596 .923 r
.60556 .44601 .61206 .45479 .62503 .45663 .61857 .44827 Metetra
.35 .579 .928 r
.61857 .44827 .62503 .45663 .63813 .45916 .63171 .45127 Metetra
.302 .567 .931 r
.63171 .45127 .63813 .45916 .65136 .46213 .64499 .45475 Metetra
.267 .564 .934 r
.64499 .45475 .65136 .46213 .66472 .46529 .6584 .45843 Metetra
.247 .57 .939 r
.6584 .45843 .66472 .46529 .67819 .46835 .67193 .46201 Metetra
.248 .59 .948 r
.67193 .46201 .67819 .46835 .69177 .47105 .68556 .4652 Metetra
.27 .622 .96 r
.68556 .4652 .69177 .47105 .70543 .47311 .69929 .46771 Metetra
.314 .665 .974 r
.69929 .46771 .70543 .47311 .71915 .47428 .71308 .46927 Metetra
.378 .717 .988 r
.71308 .46927 .71915 .47428 .73291 .47436 .72691 .46964 Metetra
.455 .769 .997 r
.72691 .46964 .73291 .47436 .74669 .47314 .74075 .46863 Metetra
.537 .817 .997 r
.74075 .46863 .74669 .47314 .76046 .47051 .75459 .46609 Metetra
.616 .853 .988 r
.75459 .46609 .76046 .47051 .77419 .46638 .76838 .46193 Metetra
.685 .876 .97 r
.76838 .46193 .77419 .46638 .78787 .46074 .78211 .45613 Metetra
.741 .887 .947 r
.78211 .45613 .78787 .46074 .80148 .45361 .79577 .44874 Metetra
.782 .888 .921 r
.79577 .44874 .80148 .45361 .815 .44509 .80934 .43985 Metetra
.811 .881 .897 r
.80934 .43985 .815 .44509 .82845 .43534 .82281 .42963 Metetra
.83 .87 .875 r
.82281 .42963 .82845 .43534 .84181 .42453 .83619 .41828 Metetra
.84 .857 .857 r
.83619 .41828 .84181 .42453 .85511 .41292 .8495 .40606 Metetra
.844 .842 .843 r
.8495 .40606 .85511 .41292 .86834 .40074 .86274 .39324 Metetra
.842 .826 .832 r
.86274 .39324 .86834 .40074 .88155 .38829 .87594 .38012 Metetra
0 .408 .846 r
.24857 .56821 .25611 .57479 .26682 .57948 .25925 .57333 Metetra
0 .42 .848 r
.25925 .57333 .26682 .57948 .27765 .58393 .27005 .5782 Metetra
.003 .458 .867 r
.27005 .5782 .27765 .58393 .28861 .58784 .28101 .5825 Metetra
.074 .52 .9 r
.28101 .5825 .28861 .58784 .29973 .59095 .29212 .58596 Metetra
.177 .598 .94 r
.29212 .58596 .29973 .59095 .31101 .59302 .3034 .58832 Metetra
.301 .68 .975 r
.3034 .58832 .31101 .59302 .32245 .59383 .31485 .58936 Metetra
.429 .751 .994 r
.31485 .58936 .32245 .59383 .33405 .59323 .32648 .58891 Metetra
.544 .802 .993 r
.32648 .58891 .33405 .59323 .34581 .59111 .33826 .58684 Metetra
.636 .832 .976 r
.33826 .58684 .34581 .59111 .3577 .58742 .35019 .58312 Metetra
.704 .845 .95 r
.35019 .58312 .3577 .58742 .36971 .58217 .36226 .57775 Metetra
.752 .847 .922 r
.36226 .57775 .36971 .58217 .38183 .57544 .37444 .57082 Metetra
.785 .842 .895 r
.37444 .57082 .38183 .57544 .39404 .56734 .3867 .56244 Metetra
.806 .833 .872 r
.3867 .56244 .39404 .56734 .40631 .55807 .39903 .55283 Metetra
.818 .823 .853 r
.39903 .55283 .40631 .55807 .41862 .54785 .41141 .54221 Metetra
.824 .813 .839 r
.41141 .54221 .41862 .54785 .43096 .53693 .42382 .53086 Metetra
.826 .803 .829 r
.42382 .53086 .43096 .53693 .44331 .5256 .43624 .51907 Metetra
.823 .793 .823 r
.43624 .51907 .44331 .5256 .45567 .51414 .44866 .50715 Metetra
.817 .783 .821 r
.44866 .50715 .45567 .51414 .46802 .50284 .46107 .49541 Metetra
.808 .773 .821 r
.46107 .49541 .46802 .50284 .48036 .49199 .47347 .48414 Metetra
.794 .764 .825 r
.47347 .48414 .48036 .49199 .4927 .48184 .48586 .47361 Metetra
.777 .754 .832 r
.48586 .47361 .4927 .48184 .50505 .47262 .49825 .46407 Metetra
.755 .743 .841 r
.49825 .46407 .50505 .47262 .51741 .46453 .51065 .45572 Metetra
.729 .731 .853 r
.51065 .45572 .51741 .46453 .52979 .45771 .52307 .44871 Metetra
.696 .718 .866 r
.52307 .44871 .52979 .45771 .54222 .45228 .53553 .44317 Metetra
.657 .702 .88 r
.53553 .44317 .54222 .45228 .5547 .44829 .54805 .43914 Metetra
.611 .684 .894 r
.54805 .43914 .5547 .44829 .56727 .44573 .56064 .43663 Metetra
.558 .663 .907 r
.56064 .43663 .56727 .44573 .57992 .44458 .57332 .43559 Metetra
.5 .639 .917 r
.57332 .43559 .57992 .44458 .59268 .44472 .58611 .43592 Metetra
.438 .614 .924 r
.58611 .43592 .59268 .44472 .60556 .44601 .59902 .43747 Metetra
.376 .589 .927 r
.59902 .43747 .60556 .44601 .61857 .44827 .61206 .44004 Metetra
.319 .568 .928 r
.61206 .44004 .61857 .44827 .63171 .45127 .62524 .4434 Metetra
.272 .552 .927 r
.62524 .4434 .63171 .45127 .64499 .45475 .63856 .44726 Metetra
.24 .546 .928 r
.63856 .44726 .64499 .45475 .6584 .45843 .65201 .45134 Metetra
.227 .552 .932 r
.65201 .45134 .6584 .45843 .67193 .46201 .66559 .45532 Metetra
.236 .572 .941 r
.66559 .45532 .67193 .46201 .68556 .4652 .67928 .45889 Metetra
.268 .605 .954 r
.67928 .45889 .68556 .4652 .69929 .46771 .69306 .46174 Metetra
.321 .65 .969 r
.69306 .46174 .69929 .46771 .71308 .46927 .70691 .46359 Metetra
.391 .701 .983 r
.70691 .46359 .71308 .46927 .72691 .46964 .7208 .46418 Metetra
.471 .752 .991 r
.7208 .46418 .72691 .46964 .74075 .46863 .73471 .46331 Metetra
.554 .797 .989 r
.73471 .46331 .74075 .46863 .75459 .46609 .7486 .46083 Metetra
.631 .831 .977 r
.7486 .46083 .75459 .46609 .76838 .46193 .76245 .45664 Metetra
.696 .853 .958 r
.76245 .45664 .76838 .46193 .78211 .45613 .77624 .45072 Metetra
.748 .864 .933 r
.77624 .45072 .78211 .45613 .79577 .44874 .78994 .44311 Metetra
.786 .866 .908 r
.78994 .44311 .79577 .44874 .80934 .43985 .80355 .43393 Metetra
.814 .862 .884 r
.80355 .43393 .80934 .43985 .82281 .42963 .81706 .42333 Metetra
.832 .855 .863 r
.81706 .42333 .82281 .42963 .83619 .41828 .83048 .41155 Metetra
.842 .845 .846 r
.83048 .41155 .83619 .41828 .8495 .40606 .8438 .39885 Metetra
.847 .834 .834 r
.8438 .39885 .8495 .40606 .86274 .39324 .85706 .38552 Metetra
.846 .823 .825 r
.85706 .38552 .86274 .39324 .87594 .38012 .87028 .37188 Metetra
0 .382 .828 r
.24096 .56157 .24857 .56821 .25925 .57333 .25163 .56698 Metetra
0 .4 .836 r
.25163 .56698 .25925 .57333 .27005 .5782 .26242 .57213 Metetra
0 .444 .862 r
.26242 .57213 .27005 .5782 .28101 .5825 .27336 .5767 Metetra
.081 .51 .901 r
.27336 .5767 .28101 .5825 .29212 .58596 .28447 .58039 Metetra
.195 .59 .943 r
.28447 .58039 .29212 .58596 .3034 .58832 .29576 .58295 Metetra
.324 .669 .975 r
.29576 .58295 .3034 .58832 .31485 .58936 .30723 .58413 Metetra
.451 .735 .988 r
.30723 .58413 .31485 .58936 .32648 .58891 .31887 .58377 Metetra
.56 .781 .981 r
.31887 .58377 .32648 .58891 .33826 .58684 .33069 .58173 Metetra
.645 .808 .961 r
.33069 .58173 .33826 .58684 .35019 .58312 .34266 .57798 Metetra
.708 .82 .934 r
.34266 .57798 .35019 .58312 .36226 .57775 .35477 .57252 Metetra
.753 .823 .907 r
.35477 .57252 .36226 .57775 .37444 .57082 .367 .56543 Metetra
.783 .82 .882 r
.367 .56543 .37444 .57082 .3867 .56244 .37932 .55686 Metetra
.803 .815 .86 r
.37932 .55686 .3867 .56244 .39903 .55283 .39172 .547 Metetra
.816 .809 .844 r
.39172 .547 .39903 .55283 .41141 .54221 .40416 .53609 Metetra
.823 .802 .831 r
.40416 .53609 .41141 .54221 .42382 .53086 .41663 .52443 Metetra
.826 .795 .823 r
.41663 .52443 .42382 .53086 .43624 .51907 .42911 .51232 Metetra
.825 .789 .818 r
.42911 .51232 .43624 .51907 .44866 .50715 .44159 .50007 Metetra
.82 .782 .817 r
.44159 .50007 .44866 .50715 .46107 .49541 .45406 .48801 Metetra
.812 .776 .82 r
.45406 .48801 .46107 .49541 .47347 .48414 .46651 .47644 Metetra
.799 .769 .825 r
.46651 .47644 .47347 .48414 .48586 .47361 .47896 .46565 Metetra
.783 .762 .834 r
.47896 .46565 .48586 .47361 .49825 .46407 .4914 .45587 Metetra
.761 .754 .846 r
.4914 .45587 .49825 .46407 .51065 .45572 .50384 .44733 Metetra
.734 .744 .859 r
.50384 .44733 .51065 .45572 .52307 .44871 .5163 .44018 Metetra
.7 .731 .875 r
.5163 .44018 .52307 .44871 .53553 .44317 .5288 .43454 Metetra
.658 .716 .891 r
.5288 .43454 .53553 .44317 .54805 .43914 .54134 .43048 Metetra
.608 .697 .907 r
.54134 .43048 .54805 .43914 .56064 .43663 .55397 .42799 Metetra
.55 .673 .92 r
.55397 .42799 .56064 .43663 .57332 .43559 .56668 .42702 Metetra
.485 .646 .929 r
.56668 .42702 .57332 .43559 .58611 .43592 .57949 .42748 Metetra
.417 .616 .933 r
.57949 .42748 .58611 .43592 .59902 .43747 .59243 .42919 Metetra
.351 .586 .932 r
.59243 .42919 .59902 .43747 .61206 .44004 .6055 .43197 Metetra
.293 .56 .928 r
.6055 .43197 .61206 .44004 .62524 .4434 .61871 .43555 Metetra
.249 .541 .924 r
.61871 .43555 .62524 .4434 .63856 .44726 .63207 .43967 Metetra
.222 .534 .923 r
.63207 .43967 .63856 .44726 .65201 .45134 .64556 .44401 Metetra
.217 .54 .926 r
.64556 .44401 .65201 .45134 .66559 .45532 .65918 .44826 Metetra
.234 .56 .935 r
.65918 .44826 .66559 .45532 .67928 .45889 .67291 .45207 Metetra
.274 .594 .949 r
.67291 .45207 .67928 .45889 .69306 .46174 .68674 .45515 Metetra
.333 .637 .964 r
.68674 .45515 .69306 .46174 .70691 .46359 .70064 .45719 Metetra
.407 .686 .975 r
.70064 .45719 .70691 .46359 .7208 .46418 .71458 .45792 Metetra
.488 .734 .98 r
.71458 .45792 .7208 .46418 .73471 .46331 .72854 .45714 Metetra
.568 .775 .976 r
.72854 .45714 .73471 .46331 .7486 .46083 .74249 .45469 Metetra
.64 .806 .963 r
.74249 .45469 .7486 .46083 .76245 .45664 .7564 .45047 Metetra
.701 .827 .942 r
.7564 .45047 .76245 .45664 .77624 .45072 .77024 .44445 Metetra
.75 .838 .919 r
.77024 .44445 .77624 .45072 .78994 .44311 .78399 .43668 Metetra
.786 .843 .895 r
.78399 .43668 .78994 .44311 .80355 .43393 .79765 .42728 Metetra
.813 .843 .872 r
.79765 .42728 .80355 .43393 .81706 .42333 .81121 .41641 Metetra
.831 .839 .853 r
.81121 .41641 .81706 .42333 .83048 .41155 .82466 .40432 Metetra
.842 .833 .838 r
.82466 .40432 .83048 .41155 .8438 .39885 .83803 .39127 Metetra
.848 .827 .827 r
.83803 .39127 .8438 .39885 .85706 .38552 .85132 .37758 Metetra
.849 .819 .82 r
.85132 .37758 .85706 .38552 .87028 .37188 .86456 .36357 Metetra
0 .369 .818 r
.2333 .55488 .24096 .56157 .25163 .56698 .24395 .56043 Metetra
0 .392 .832 r
.24395 .56043 .25163 .56698 .26242 .57213 .25474 .56572 Metetra
.006 .44 .864 r
.25474 .56572 .26242 .57213 .27336 .5767 .26569 .57041 Metetra
.102 .509 .906 r
.26569 .57041 .27336 .5767 .28447 .58039 .2768 .57422 Metetra
.223 .586 .946 r
.2768 .57422 .28447 .58039 .29576 .58295 .2881 .57686 Metetra
.352 .66 .971 r
.2881 .57686 .29576 .58295 .30723 .58413 .29958 .57811 Metetra
.472 .718 .977 r
.29958 .57811 .30723 .58413 .31887 .58377 .31125 .57778 Metetra
.573 .758 .966 r
.31125 .57778 .31887 .58377 .33069 .58173 .3231 .57575 Metetra
.651 .782 .944 r
.3231 .57575 .33069 .58173 .34266 .57798 .3351 .57197 Metetra
.709 .794 .918 r
.3351 .57197 .34266 .57798 .35477 .57252 .34726 .56644 Metetra
.75 .799 .892 r
.34726 .56644 .35477 .57252 .367 .56543 .35953 .55926 Metetra
.779 .799 .869 r
.35953 .55926 .367 .56543 .37932 .55686 .37191 .55056 Metetra
.799 .797 .85 r
.37191 .55056 .37932 .55686 .39172 .547 .38436 .54055 Metetra
.812 .794 .835 r
.38436 .54055 .39172 .547 .40416 .53609 .39685 .52948 Metetra
.821 .791 .824 r
.39685 .52948 .40416 .53609 .41663 .52443 .40938 .51764 Metetra
.824 .788 .818 r
.40938 .51764 .41663 .52443 .42911 .51232 .42192 .50534 Metetra
.824 .785 .815 r
.42192 .50534 .42911 .51232 .44159 .50007 .43446 .4929 Metetra
.821 .781 .815 r
.43446 .4929 .44159 .50007 .45406 .48801 .44699 .48065 Metetra
.814 .778 .819 r
.44699 .48065 .45406 .48801 .46651 .47644 .4595 .46891 Metetra
.803 .774 .827 r
.4595 .46891 .46651 .47644 .47896 .46565 .472 .45795 Metetra
.787 .77 .837 r
.472 .45795 .47896 .46565 .4914 .45587 .48448 .44803 Metetra
.766 .765 .851 r
.48448 .44803 .4914 .45587 .50384 .44733 .49697 .43937 Metetra
.739 .757 .867 r
.49697 .43937 .50384 .44733 .5163 .44018 .50947 .43214 Metetra
.703 .746 .885 r
.50947 .43214 .5163 .44018 .5288 .43454 .52201 .42644 Metetra
.659 .731 .903 r
.52201 .42644 .5288 .43454 .54134 .43048 .53459 .42233 Metetra
.604 .711 .92 r
.53459 .42233 .54134 .43048 .55397 .42799 .54724 .41984 Metetra
.541 .685 .933 r
.54724 .41984 .55397 .42799 .56668 .42702 .55998 .41889 Metetra
.471 .654 .94 r
.55998 .41889 .56668 .42702 .57949 .42748 .57283 .4194 Metetra
.399 .62 .941 r
.57283 .4194 .57949 .42748 .59243 .42919 .5858 .42118 Metetra
.332 .585 .936 r
.5858 .42118 .59243 .42919 .6055 .43197 .5989 .42405 Metetra
.275 .556 .929 r
.5989 .42405 .6055 .43197 .61871 .43555 .61214 .42774 Metetra
.234 .536 .923 r
.61214 .42774 .61871 .43555 .63207 .43967 .62552 .43198 Metetra
.214 .528 .92 r
.62552 .43198 .63207 .43967 .64556 .44401 .63904 .43645 Metetra
.217 .534 .923 r
.63904 .43645 .64556 .44401 .65918 .44826 .65269 .44081 Metetra
.242 .554 .932 r
.65269 .44081 .65918 .44826 .67291 .45207 .66646 .44475 Metetra
.288 .587 .944 r
.66646 .44475 .67291 .45207 .68674 .45515 .68033 .44792 Metetra
.351 .628 .957 r
.68033 .44792 .68674 .45515 .70064 .45719 .69428 .45004 Metetra
.425 .672 .965 r
.69428 .45004 .70064 .45719 .71458 .45792 .70826 .45084 Metetra
.504 .715 .967 r
.70826 .45084 .71458 .45792 .72854 .45714 .72227 .45009 Metetra
.579 .752 .96 r
.72227 .45009 .72854 .45714 .74249 .45469 .73627 .44763 Metetra
.646 .781 .945 r
.73627 .44763 .74249 .45469 .7564 .45047 .75022 .44337 Metetra
.703 .8 .926 r
.75022 .44337 .7564 .45047 .77024 .44445 .76412 .43728 Metetra
.748 .813 .904 r
.76412 .43728 .77024 .44445 .78399 .43668 .77793 .42941 Metetra
.783 .82 .882 r
.77793 .42941 .78399 .43668 .79765 .42728 .79164 .41987 Metetra
.809 .823 .862 r
.79164 .41987 .79765 .42728 .81121 .41641 .80525 .40884 Metetra
.827 .823 .845 r
.80525 .40884 .81121 .41641 .82466 .40432 .81875 .39656 Metetra
.839 .821 .831 r
.81875 .39656 .82466 .40432 .83803 .39127 .83217 .38331 Metetra
.846 .819 .822 r
.83217 .38331 .83803 .39127 .85132 .37758 .84551 .36941 Metetra
.849 .815 .816 r
.84551 .36941 .85132 .37758 .86456 .36357 .8588 .35518 Metetra
0 .369 .818 r
.22557 .54814 .2333 .55488 .24395 .56043 .23623 .55368 Metetra
0 .396 .838 r
.23623 .55368 .24395 .56043 .25474 .56572 .24703 .55895 Metetra
.031 .448 .873 r
.24703 .55895 .25474 .56572 .26569 .57041 .25798 .56363 Metetra
.136 .515 .913 r
.25798 .56363 .26569 .57041 .2768 .57422 .2691 .56742 Metetra
.257 .586 .946 r
.2691 .56742 .2768 .57422 .2881 .57686 .28042 .57005 Metetra
.381 .651 .963 r
.28042 .57005 .2881 .57686 .29958 .57811 .29192 .57128 Metetra
.492 .701 .962 r
.29192 .57128 .29958 .57811 .31125 .57778 .30361 .57092 Metetra
.583 .736 .948 r
.30361 .57092 .31125 .57778 .3231 .57575 .31549 .56887 Metetra
.654 .757 .926 r
.31549 .56887 .3231 .57575 .3351 .57197 .32753 .56505 Metetra
.706 .769 .901 r
.32753 .56505 .3351 .57197 .34726 .56644 .33972 .55949 Metetra
.745 .776 .878 r
.33972 .55949 .34726 .56644 .35953 .55926 .35204 .55227 Metetra
.773 .779 .858 r
.35204 .55227 .35953 .55926 .37191 .55056 .36446 .54353 Metetra
.793 .78 .841 r
.36446 .54353 .37191 .55056 .38436 .54055 .37695 .53347 Metetra
.807 .781 .828 r
.37695 .53347 .38436 .54055 .39685 .52948 .38951 .52236 Metetra
.816 .781 .82 r
.38951 .52236 .39685 .52948 .40938 .51764 .40209 .51046 Metetra
.821 .78 .814 r
.40209 .51046 .40938 .51764 .42192 .50534 .41468 .49811 Metetra
.823 .78 .813 r
.41468 .49811 .42192 .50534 .43446 .4929 .42728 .48563 Metetra
.821 .78 .815 r
.42728 .48563 .43446 .4929 .44699 .48065 .43986 .47333 Metetra
.815 .78 .82 r
.43986 .47333 .44699 .48065 .4595 .46891 .45242 .46154 Metetra
.805 .78 .829 r
.45242 .46154 .4595 .46891 .472 .45795 .46497 .45053 Metetra
.791 .779 .842 r
.46497 .45053 .472 .45795 .48448 .44803 .47751 .44057 Metetra
.77 .776 .858 r
.47751 .44057 .48448 .44803 .49697 .43937 .49004 .43187 Metetra
.742 .771 .876 r
.49004 .43187 .49697 .43937 .50947 .43214 .50259 .42459 Metetra
.705 .762 .896 r
.50259 .42459 .50947 .43214 .52201 .42644 .51516 .41885 Metetra
.659 .748 .916 r
.51516 .41885 .52201 .42644 .53459 .42233 .52778 .41471 Metetra
.601 .727 .933 r
.52778 .41471 .53459 .42233 .54724 .41984 .54047 .41219 Metetra
.534 .699 .946 r
.54047 .41219 .54724 .41984 .55998 .41889 .55324 .41121 Metetra
.46 .665 .951 r
.55324 .41121 .55998 .41889 .57283 .4194 .56612 .41169 Metetra
.386 .626 .949 r
.56612 .41169 .57283 .4194 .5858 .42118 .57911 .41345 Metetra
.318 .589 .941 r
.57911 .41345 .5858 .42118 .5989 .42405 .59224 .4163 Metetra
.264 .557 .931 r
.59224 .4163 .5989 .42405 .61214 .42774 .60551 .41997 Metetra
.229 .536 .923 r
.60551 .41997 .61214 .42774 .62552 .43198 .61891 .42418 Metetra
.216 .528 .92 r
.61891 .42418 .62552 .43198 .63904 .43645 .63246 .42863 Metetra
.226 .534 .923 r
.63246 .42863 .63904 .43645 .65269 .44081 .64614 .43298 Metetra
.258 .553 .93 r
.64614 .43298 .65269 .44081 .66646 .44475 .65994 .43689 Metetra
.308 .584 .94 r
.65994 .43689 .66646 .44475 .68033 .44792 .67384 .44004 Metetra
.372 .621 .948 r
.67384 .44004 .68033 .44792 .69428 .45004 .68782 .44214 Metetra
.445 .66 .953 r
.68782 .44214 .69428 .45004 .70826 .45084 .70184 .4429 Metetra
.518 .698 .951 r
.70184 .4429 .70826 .45084 .72227 .45009 .71589 .44212 Metetra
.588 .73 .942 r
.71589 .44212 .72227 .45009 .73627 .44763 .72993 .43963 Metetra
.649 .756 .927 r
.72993 .43963 .73627 .44763 .75022 .44337 .74393 .43533 Metetra
.701 .775 .909 r
.74393 .43533 .75022 .44337 .76412 .43728 .75788 .4292 Metetra
.744 .788 .889 r
.75788 .4292 .76412 .43728 .77793 .42941 .77174 .42128 Metetra
.777 .798 .87 r
.77174 .42128 .77793 .42941 .79164 .41987 .78551 .41169 Metetra
.802 .803 .852 r
.78551 .41169 .79164 .41987 .80525 .40884 .79918 .40061 Metetra
.821 .807 .838 r
.79918 .40061 .80525 .40884 .81875 .39656 .81275 .38827 Metetra
.835 .809 .826 r
.81275 .38827 .81875 .39656 .83217 .38331 .82622 .37497 Metetra
.843 .811 .819 r
.82622 .37497 .83217 .38331 .84551 .36941 .83963 .361 Metetra
.847 .811 .815 r
.83963 .361 .84551 .36941 .8588 .35518 .85298 .34671 Metetra
0 .382 .828 r
.21778 .54134 .22557 .54814 .23623 .55368 .22846 .54672 Metetra
0 .413 .852 r
.22846 .54672 .23623 .55368 .24703 .55895 .23927 .55183 Metetra
.071 .463 .886 r
.23927 .55183 .24703 .55895 .25798 .56363 .25024 .55636 Metetra
.179 .526 .92 r
.25024 .55636 .25798 .56363 .2691 .56742 .26139 .56001 Metetra
.297 .588 .944 r
.26139 .56001 .2691 .56742 .28042 .57005 .27272 .56251 Metetra
.411 .643 .952 r
.27272 .56251 .28042 .57005 .29192 .57128 .28425 .56363 Metetra
.51 .685 .945 r
.28425 .56363 .29192 .57128 .30361 .57092 .29597 .5632 Metetra
.591 .714 .929 r
.29597 .5632 .30361 .57092 .31549 .56887 .30787 .56108 Metetra
.654 .733 .908 r
.30787 .56108 .31549 .56887 .32753 .56505 .31994 .55724 Metetra
.702 .746 .886 r
.31994 .55724 .32753 .56505 .33972 .55949 .33216 .55167 Metetra
.738 .754 .865 r
.33216 .55167 .33972 .55949 .35204 .55227 .34452 .54446 Metetra
.765 .76 .848 r
.34452 .54446 .35204 .55227 .36446 .54353 .35698 .53576 Metetra
.785 .764 .834 r
.35698 .53576 .36446 .54353 .37695 .53347 .36952 .52577 Metetra
.8 .767 .823 r
.36952 .52577 .37695 .53347 .38951 .52236 .38211 .51472 Metetra
.81 .77 .816 r
.38211 .51472 .38951 .52236 .40209 .51046 .39474 .50291 Metetra
.816 .773 .812 r
.39474 .50291 .40209 .51046 .41468 .49811 .40739 .49065 Metetra
.819 .776 .812 r
.40739 .49065 .41468 .49811 .42728 .48563 .42003 .47826 Metetra
.819 .78 .816 r
.42003 .47826 .42728 .48563 .43986 .47333 .43267 .46605 Metetra
.814 .783 .823 r
.43267 .46605 .43986 .47333 .45242 .46154 .44528 .45433 Metetra
.806 .785 .834 r
.44528 .45433 .45242 .46154 .46497 .45053 .45788 .44339 Metetra
.792 .787 .848 r
.45788 .44339 .46497 .45053 .47751 .44057 .47046 .43348 Metetra
.772 .787 .866 r
.47046 .43348 .47751 .44057 .49004 .43187 .48305 .42481 Metetra
.744 .785 .886 r
.48305 .42481 .49004 .43187 .50259 .42459 .49564 .41754 Metetra
.707 .778 .908 r
.49564 .41754 .50259 .42459 .51516 .41885 .50826 .41179 Metetra
.658 .765 .929 r
.50826 .41179 .51516 .41885 .52778 .41471 .52092 .40762 Metetra
.598 .744 .946 r
.52092 .40762 .52778 .41471 .54047 .41219 .53364 .40504 Metetra
.527 .714 .958 r
.53364 .40504 .54047 .41219 .55324 .41121 .54645 .40398 Metetra
.451 .677 .961 r
.54645 .40398 .55324 .41121 .56612 .41169 .55936 .40435 Metetra
.376 .636 .956 r
.55936 .40435 .56612 .41169 .57911 .41345 .57238 .406 Metetra
.311 .596 .946 r
.57238 .406 .57911 .41345 .59224 .4163 .58553 .4087 Metetra
.262 .563 .935 r
.58553 .4087 .59224 .4163 .60551 .41997 .59882 .41223 Metetra
.234 .542 .926 r
.59882 .41223 .60551 .41997 .61891 .42418 .61225 .41628 Metetra
.228 .534 .922 r
.61225 .41628 .61891 .42418 .63246 .42863 .62582 .42056 Metetra
.245 .539 .924 r
.62582 .42056 .63246 .42863 .64614 .43298 .63952 .42474 Metetra
.281 .557 .929 r
.63952 .42474 .64614 .43298 .65994 .43689 .65334 .4285 Metetra
.333 .584 .935 r
.65334 .4285 .65994 .43689 .67384 .44004 .66726 .4315 Metetra
.396 .616 .939 r
.66726 .4315 .67384 .44004 .68782 .44214 .68127 .43347 Metetra
.464 .649 .939 r
.68127 .43347 .68782 .44214 .70184 .4429 .69532 .43412 Metetra
.532 .681 .934 r
.69532 .43412 .70184 .4429 .71589 .44212 .7094 .43324 Metetra
.595 .709 .923 r
.7094 .43324 .71589 .44212 .72993 .43963 .72348 .43068 Metetra
.65 .732 .909 r
.72348 .43068 .72993 .43963 .74393 .43533 .73753 .42634 Metetra
.698 .75 .892 r
.73753 .42634 .74393 .43533 .75788 .4292 .75152 .4202 Metetra
.737 .765 .875 r
.75152 .4202 .75788 .4292 .77174 .42128 .76544 .41229 Metetra
.769 .776 .859 r
.76544 .41229 .77174 .42128 .78551 .41169 .77927 .40273 Metetra
.794 .785 .844 r
.77927 .40273 .78551 .41169 .79918 .40061 .79301 .39171 Metetra
.814 .792 .832 r
.79301 .39171 .79918 .40061 .81275 .38827 .80664 .37945 Metetra
.828 .798 .823 r
.80664 .37945 .81275 .38827 .82622 .37497 .82019 .36623 Metetra
.838 .803 .817 r
.82019 .36623 .82622 .37497 .83963 .361 .83368 .35236 Metetra
.844 .807 .815 r
.83368 .35236 .83963 .361 .85298 .34671 .84711 .33817 Metetra
0 .408 .846 r
.20992 .53448 .21778 .54134 .22846 .54672 .22064 .53955 Metetra
.034 .439 .871 r
.22064 .53955 .22846 .54672 .23927 .55183 .23148 .54435 Metetra
.123 .486 .901 r
.23148 .54435 .23927 .55183 .25024 .55636 .24248 .54859 Metetra
.229 .54 .926 r
.24248 .54859 .25024 .55636 .26139 .56001 .25366 .55197 Metetra
.338 .592 .939 r
.25366 .55197 .26139 .56001 .27272 .56251 .26502 .55425 Metetra
.44 .636 .939 r
.26502 .55425 .27272 .56251 .28425 .56363 .27657 .55518 Metetra
.527 .67 .928 r
.27657 .55518 .28425 .56363 .29597 .5632 .28832 .5546 Metetra
.598 .694 .91 r
.28832 .5546 .29597 .5632 .30787 .56108 .30024 .5524 Metetra
.654 .712 .89 r
.30024 .5524 .30787 .56108 .31994 .55724 .31234 .54853 Metetra
.697 .724 .871 r
.31234 .54853 .31994 .55724 .33216 .55167 .3246 .54298 Metetra
.731 .734 .854 r
.3246 .54298 .33216 .55167 .34452 .54446 .33698 .53585 Metetra
.757 .742 .839 r
.33698 .53585 .34452 .54446 .35698 .53576 .34947 .52727 Metetra
.777 .748 .827 r
.34947 .52727 .35698 .53576 .36952 .52577 .36205 .51743 Metetra
.792 .755 .819 r
.36205 .51743 .36952 .52577 .38211 .51472 .37468 .50658 Metetra
.803 .76 .814 r
.37468 .50658 .38211 .51472 .39474 .50291 .38735 .49499 Metetra
.81 .766 .812 r
.38735 .49499 .39474 .50291 .40739 .49065 .40004 .48295 Metetra
.815 .773 .814 r
.40004 .48295 .40739 .49065 .42003 .47826 .41273 .47079 Metetra
.815 .779 .819 r
.41273 .47079 .42003 .47826 .43267 .46605 .42541 .4588 Metetra
.812 .785 .827 r
.42541 .4588 .43267 .46605 .44528 .45433 .43807 .44729 Metetra
.805 .791 .839 r
.43807 .44729 .44528 .45433 .45788 .44339 .45072 .43653 Metetra
.792 .796 .855 r
.45072 .43653 .45788 .44339 .47046 .43348 .46335 .42676 Metetra
.773 .799 .874 r
.46335 .42676 .47046 .43348 .48305 .42481 .47598 .41819 Metetra
.745 .799 .896 r
.47598 .41819 .48305 .42481 .49564 .41754 .48862 .41099 Metetra
.706 .794 .919 r
.48862 .41099 .49564 .41754 .50826 .41179 .50128 .40525 Metetra
.656 .782 .941 r
.50128 .40525 .50826 .41179 .52092 .40762 .51399 .40105 Metetra
.594 .761 .959 r
.51399 .40105 .52092 .40762 .53364 .40504 .52675 .39838 Metetra
.522 .73 .969 r
.52675 .39838 .53364 .40504 .54645 .40398 .5396 .39719 Metetra
.446 .691 .97 r
.5396 .39719 .54645 .40398 .55936 .40435 .55254 .39739 Metetra
.373 .649 .963 r
.55254 .39739 .55936 .40435 .57238 .406 .56559 .39881 Metetra
.312 .608 .952 r
.56559 .39881 .57238 .406 .58553 .4087 .57876 .40127 Metetra
.269 .575 .94 r
.57876 .40127 .58553 .4087 .59882 .41223 .59207 .40451 Metetra
.248 .554 .931 r
.59207 .40451 .59882 .41223 .61225 .41628 .60552 .40827 Metetra
.249 .546 .927 r
.60552 .40827 .61225 .41628 .62582 .42056 .61911 .41225 Metetra
.271 .55 .926 r
.61911 .41225 .62582 .42056 .63952 .42474 .63283 .41612 Metetra
.31 .565 .928 r
.63283 .41612 .63952 .42474 .65334 .4285 .64667 .41958 Metetra
.362 .587 .93 r
.64667 .41958 .65334 .4285 .66726 .4315 .66061 .42232 Metetra
.422 .613 .929 r
.66061 .42232 .66726 .4315 .68127 .43347 .67463 .42403 Metetra
.484 .64 .925 r
.67463 .42403 .68127 .43347 .69532 .43412 .68871 .42449 Metetra
.544 .667 .917 r
.68871 .42449 .69532 .43412 .7094 .43324 .70281 .42346 Metetra
.6 .69 .905 r
.70281 .42346 .7094 .43324 .72348 .43068 .71693 .4208 Metetra
.65 .711 .892 r
.71693 .4208 .72348 .43068 .73753 .42634 .73102 .41641 Metetra
.693 .728 .877 r
.73102 .41641 .73753 .42634 .75152 .4202 .74506 .41028 Metetra
.73 .743 .862 r
.74506 .41028 .75152 .4202 .76544 .41229 .75903 .40244 Metetra
.76 .756 .849 r
.75903 .40244 .76544 .41229 .77927 .40273 .77292 .39301 Metetra
.784 .767 .837 r
.77292 .39301 .77927 .40273 .79301 .39171 .78673 .38215 Metetra
.804 .777 .828 r
.78673 .38215 .79301 .39171 .80664 .37945 .80045 .37009 Metetra
.819 .786 .821 r
.80045 .37009 .80664 .37945 .82019 .36623 .81408 .3571 Metetra
.831 .794 .817 r
.81408 .3571 .82019 .36623 .83368 .35236 .82766 .34348 Metetra
.838 .803 .817 r
.82766 .34348 .83368 .35236 .84711 .33817 .84119 .32955 Metetra
.026 .444 .871 r
.202 .52757 .20992 .53448 .22064 .53955 .21276 .53217 Metetra
.094 .473 .893 r
.21276 .53217 .22064 .53955 .23148 .54435 .22365 .53653 Metetra
.183 .513 .914 r
.22365 .53653 .23148 .54435 .24248 .54859 .2347 .54034 Metetra
.282 .557 .928 r
.2347 .54034 .24248 .54859 .25366 .55197 .24591 .54334 Metetra
.38 .597 .931 r
.24591 .54334 .25366 .55197 .26502 .55425 .25731 .54528 Metetra
.468 .63 .924 r
.25731 .54528 .26502 .55425 .27657 .55518 .26889 .54594 Metetra
.542 .657 .91 r
.26889 .54594 .27657 .55518 .28832 .5546 .28066 .54517 Metetra
.603 .677 .893 r
.28066 .54517 .28832 .5546 .30024 .5524 .29262 .54285 Metetra
.652 .692 .875 r
.29262 .54285 .30024 .5524 .31234 .54853 .30474 .53894 Metetra
.691 .705 .858 r
.30474 .53894 .31234 .54853 .3246 .54298 .31702 .53345 Metetra
.722 .716 .843 r
.31702 .53345 .3246 .54298 .33698 .53585 .32943 .52645 Metetra
.747 .725 .832 r
.32943 .52645 .33698 .53585 .34947 .52727 .34195 .51806 Metetra
.767 .734 .822 r
.34195 .51806 .34947 .52727 .36205 .51743 .35455 .50849 Metetra
.782 .742 .816 r
.35455 .50849 .36205 .51743 .37468 .50658 .36721 .49794 Metetra
.794 .751 .813 r
.36721 .49794 .37468 .50658 .38735 .49499 .37991 .48669 Metetra
.803 .76 .813 r
.37991 .48669 .38735 .49499 .40004 .48295 .39264 .47502 Metetra
.808 .769 .816 r
.39264 .47502 .40004 .48295 .41273 .47079 .40536 .46322 Metetra
.81 .778 .823 r
.40536 .46322 .41273 .47079 .42541 .4588 .41808 .45158 Metetra
.809 .787 .833 r
.41808 .45158 .42541 .4588 .43807 .44729 .43079 .4404 Metetra
.802 .796 .846 r
.43079 .4404 .43807 .44729 .45072 .43653 .44348 .42992 Metetra
.79 .804 .863 r
.44348 .42992 .45072 .43653 .46335 .42676 .45617 .4204 Metetra
.771 .81 .884 r
.45617 .4204 .46335 .42676 .47598 .41819 .46884 .41201 Metetra
.743 .813 .907 r
.46884 .41201 .47598 .41819 .48862 .41099 .48153 .40492 Metetra
.705 .81 .931 r
.48153 .40492 .48862 .41099 .50128 .40525 .49424 .39922 Metetra
.654 .8 .953 r
.49424 .39922 .50128 .40525 .51399 .40105 .50699 .39498 Metetra
.591 .779 .97 r
.50699 .39498 .51399 .40105 .52675 .39838 .5198 .3922 Metetra
.519 .747 .978 r
.5198 .3922 .52675 .39838 .5396 .39719 .53268 .39084 Metetra
.445 .708 .978 r
.53268 .39084 .5396 .39719 .55254 .39739 .54565 .39078 Metetra
.376 .665 .971 r
.54565 .39078 .55254 .39739 .56559 .39881 .55874 .3919 Metetra
.322 .625 .959 r
.55874 .3919 .56559 .39881 .57876 .40127 .57194 .39399 Metetra
.286 .592 .947 r
.57194 .39399 .57876 .40127 .59207 .40451 .58527 .39683 Metetra
.272 .571 .938 r
.58527 .39683 .59207 .40451 .60552 .40827 .59874 .40016 Metetra
.279 .562 .932 r
.59874 .40016 .60552 .40827 .61911 .41225 .61234 .40368 Metetra
.304 .564 .929 r
.61234 .40368 .61911 .41225 .63283 .41612 .62607 .40712 Metetra
.344 .575 .927 r
.62607 .40712 .63283 .41612 .64667 .41958 .63992 .41015 Metetra
.393 .592 .924 r
.63992 .41015 .64667 .41958 .66061 .42232 .65387 .41249 Metetra
.447 .612 .919 r
.65387 .41249 .66061 .42232 .67463 .42403 .66791 .41386 Metetra
.503 .633 .911 r
.66791 .41386 .67463 .42403 .68871 .42449 .682 .41403 Metetra
.556 .654 .901 r
.682 .41403 .68871 .42449 .70281 .42346 .69613 .41279 Metetra
.605 .673 .889 r
.69613 .41279 .70281 .42346 .71693 .4208 .71027 .41 Metetra
.649 .691 .876 r
.71027 .41 .71693 .4208 .73102 .41641 .7244 .40557 Metetra
.688 .708 .863 r
.7244 .40557 .73102 .41641 .74506 .41028 .73848 .39947 Metetra
.721 .723 .851 r
.73848 .39947 .74506 .41028 .75903 .40244 .75251 .39176 Metetra
.749 .736 .84 r
.75251 .39176 .75903 .40244 .77292 .39301 .76647 .38252 Metetra
.773 .749 .831 r
.76647 .38252 .77292 .39301 .78673 .38215 .78036 .37194 Metetra
.793 .762 .824 r
.78036 .37194 .78673 .38215 .80045 .37009 .79416 .36021 Metetra
.809 .774 .82 r
.79416 .36021 .80045 .37009 .81408 .3571 .80789 .34759 Metetra
.822 .786 .819 r
.80789 .34759 .81408 .3571 .82766 .34348 .82157 .33437 Metetra
.83 .798 .82 r
.82157 .33437 .82766 .34348 .84119 .32955 .83521 .32084 Metetra
.094 .488 .899 r
.19402 .5206 .202 .52757 .21276 .53217 .20484 .5246 Metetra
.164 .512 .914 r
.20484 .5246 .21276 .53217 .22365 .53653 .21579 .52837 Metetra
.249 .542 .925 r
.21579 .52837 .22365 .53653 .2347 .54034 .22689 .53163 Metetra
.337 .573 .927 r
.22689 .53163 .2347 .54034 .24591 .54334 .23815 .53412 Metetra
.42 .602 .922 r
.23815 .53412 .24591 .54334 .25731 .54528 .24959 .53563 Metetra
.494 .626 .909 r
.24959 .53563 .25731 .54528 .26889 .54594 .26121 .53595 Metetra
.556 .645 .893 r
.26121 .53595 .26889 .54594 .28066 .54517 .27301 .53493 Metetra
.608 .662 .877 r
.27301 .53493 .28066 .54517 .29262 .54285 .28499 .53247 Metetra
.65 .675 .861 r
.28499 .53247 .29262 .54285 .30474 .53894 .29714 .52853 Metetra
.685 .687 .847 r
.29714 .52853 .30474 .53894 .31702 .53345 .30944 .52311 Metetra
.713 .699 .835 r
.30944 .52311 .31702 .53345 .32943 .52645 .32186 .51629 Metetra
.737 .709 .825 r
.32186 .51629 .32943 .52645 .34195 .51806 .3344 .50818 Metetra
.756 .72 .819 r
.3344 .50818 .34195 .51806 .35455 .50849 .34702 .49896 Metetra
.772 .731 .815 r
.34702 .49896 .35455 .50849 .36721 .49794 .3597 .48883 Metetra
.784 .741 .814 r
.3597 .48883 .36721 .49794 .37991 .48669 .37243 .47804 Metetra
.794 .753 .815 r
.37243 .47804 .37991 .48669 .39264 .47502 .38518 .46686 Metetra
.8 .765 .82 r
.38518 .46686 .39264 .47502 .40536 .46322 .39794 .45555 Metetra
.803 .777 .828 r
.39794 .45555 .40536 .46322 .41808 .45158 .41069 .44439 Metetra
.803 .789 .84 r
.41069 .44439 .41808 .45158 .43079 .4404 .42344 .43365 Metetra
.798 .801 .855 r
.42344 .43365 .43079 .4404 .44348 .42992 .43617 .42357 Metetra
.786 .812 .873 r
.43617 .42357 .44348 .42992 .45617 .4204 .4489 .41437 Metetra
.768 .821 .895 r
.4489 .41437 .45617 .4204 .46884 .41201 .46163 .40623 Metetra
.74 .827 .918 r
.46163 .40623 .46884 .41201 .48153 .40492 .47437 .3993 Metetra
.701 .826 .942 r
.47437 .3993 .48153 .40492 .49424 .39922 .48712 .39367 Metetra
.65 .816 .963 r
.48712 .39367 .49424 .39922 .50699 .39498 .49992 .38939 Metetra
.588 .796 .979 r
.49992 .38939 .50699 .39498 .5198 .3922 .51277 .38648 Metetra
.518 .765 .986 r
.51277 .38648 .5198 .3922 .53268 .39084 .5257 .38488 Metetra
.449 .726 .985 r
.5257 .38488 .53268 .39084 .54565 .39078 .53871 .38451 Metetra
.387 .684 .977 r
.53871 .38451 .54565 .39078 .55874 .3919 .55182 .38523 Metetra
.34 .645 .966 r
.55182 .38523 .55874 .3919 .57194 .39399 .56506 .38685 Metetra
.312 .614 .955 r
.56506 .38685 .57194 .39399 .58527 .39683 .57841 .38917 Metetra
.304 .593 .946 r
.57841 .38917 .58527 .39683 .59874 .40016 .5919 .39193 Metetra
.315 .583 .938 r
.5919 .39193 .59874 .40016 .61234 .40368 .60551 .39488 Metetra
.342 .582 .931 r
.60551 .39488 .61234 .40368 .62607 .40712 .61925 .39774 Metetra
.38 .587 .925 r
.61925 .39774 .62607 .40712 .63992 .41015 .6331 .40022 Metetra
.425 .598 .917 r
.6331 .40022 .63992 .41015 .65387 .41249 .64706 .40205 Metetra
.473 .612 .908 r
.64706 .40205 .65387 .41249 .66791 .41386 .6611 .40298 Metetra
.521 .627 .897 r
.6611 .40298 .66791 .41386 .682 .41403 .67521 .40279 Metetra
.567 .643 .885 r
.67521 .40279 .682 .41403 .69613 .41279 .68936 .40128 Metetra
.609 .659 .873 r
.68936 .40128 .69613 .41279 .71027 .41 .70352 .39833 Metetra
.647 .674 .862 r
.70352 .39833 .71027 .41 .7244 .40557 .71768 .39385 Metetra
.682 .689 .851 r
.71768 .39385 .7244 .40557 .73848 .39947 .73181 .38782 Metetra
.712 .704 .841 r
.73181 .38782 .73848 .39947 .75251 .39176 .7459 .38028 Metetra
.738 .718 .833 r
.7459 .38028 .75251 .39176 .76647 .38252 .75992 .37132 Metetra
.761 .733 .827 r
.75992 .37132 .76647 .38252 .78036 .37194 .77389 .36111 Metetra
.781 .748 .822 r
.77389 .36111 .78036 .37194 .79416 .36021 .78778 .34982 Metetra
.797 .762 .821 r
.78778 .34982 .79416 .36021 .80789 .34759 .80162 .33771 Metetra
.811 .777 .822 r
.80162 .33771 .80789 .34759 .82157 .33437 .81541 .32503 Metetra
.821 .792 .826 r
.81541 .32503 .82157 .33437 .83521 .32084 .82917 .31206 Metetra
.174 .537 .925 r
.18597 .51358 .19402 .5206 .20484 .5246 .19687 .51685 Metetra
.241 .552 .931 r
.19687 .51685 .20484 .5246 .21579 .52837 .20789 .5199 Metetra
.316 .571 .931 r
.20789 .5199 .21579 .52837 .22689 .53163 .21906 .52248 Metetra
.39 .59 .923 r
.21906 .52248 .22689 .53163 .23815 .53412 .23038 .52437 Metetra
.458 .607 .91 r
.23038 .52437 .23815 .53412 .24959 .53563 .24186 .52535 Metetra
.518 .623 .895 r
.24186 .52535 .24959 .53563 .26121 .53595 .25353 .52526 Metetra
.569 .636 .878 r
.25353 .52526 .26121 .53595 .27301 .53493 .26536 .52394 Metetra
.612 .649 .862 r
.26536 .52394 .27301 .53493 .28499 .53247 .27737 .52131 Metetra
.648 .66 .848 r
.27737 .52131 .28499 .53247 .29714 .52853 .28954 .51734 Metetra
.678 .672 .837 r
.28954 .51734 .29714 .52853 .30944 .52311 .30185 .51202 Metetra
.704 .683 .827 r
.30185 .51202 .30944 .52311 .32186 .51629 .31429 .50542 Metetra
.726 .695 .82 r
.31429 .50542 .32186 .51629 .3344 .50818 .32683 .49766 Metetra
.744 .707 .816 r
.32683 .49766 .3344 .50818 .34702 .49896 .33946 .48887 Metetra
.76 .719 .814 r
.33946 .48887 .34702 .49896 .3597 .48883 .35216 .47927 Metetra
.773 .732 .815 r
.35216 .47927 .3597 .48883 .37243 .47804 .3649 .46905 Metetra
.783 .746 .819 r
.3649 .46905 .37243 .47804 .38518 .46686 .37766 .45848 Metetra
.79 .76 .826 r
.37766 .45848 .38518 .46686 .39794 .45555 .39045 .44778 Metetra
.795 .775 .836 r
.39045 .44778 .39794 .45555 .41069 .44439 .40323 .43722 Metetra
.795 .791 .848 r
.40323 .43722 .41069 .44439 .42344 .43365 .41601 .42704 Metetra
.79 .806 .865 r
.41601 .42704 .42344 .43365 .43617 .42357 .42879 .41745 Metetra
.78 .82 .884 r
.42879 .41745 .43617 .42357 .4489 .41437 .44156 .40866 Metetra
.762 .832 .906 r
.44156 .40866 .4489 .41437 .46163 .40623 .45433 .40083 Metetra
.734 .839 .929 r
.45433 .40083 .46163 .40623 .47437 .3993 .46712 .3941 Metetra
.695 .84 .952 r
.46712 .3941 .47437 .3993 .48712 .39367 .47993 .38856 Metetra
.645 .831 .972 r
.47993 .38856 .48712 .39367 .49992 .38939 .49278 .38425 Metetra
.585 .812 .986 r
.49278 .38425 .49992 .38939 .51277 .38648 .50567 .38119 Metetra
.52 .782 .992 r
.50567 .38119 .51277 .38648 .5257 .38488 .51864 .37931 Metetra
.457 .745 .991 r
.51864 .37931 .5257 .38488 .53871 .38451 .5317 .37855 Metetra
.404 .706 .984 r
.5317 .37855 .53871 .38451 .55182 .38523 .54485 .37878 Metetra
.366 .67 .974 r
.54485 .37878 .55182 .38523 .56506 .38685 .55811 .37984 Metetra
.346 .64 .963 r
.55811 .37984 .56506 .38685 .57841 .38917 .57149 .38152 Metetra
.344 .619 .953 r
.57149 .38152 .57841 .38917 .5919 .39193 .58499 .38361 Metetra
.358 .606 .943 r
.58499 .38361 .5919 .39193 .60551 .39488 .59862 .38586 Metetra
.384 .601 .933 r
.59862 .38586 .60551 .39488 .61925 .39774 .61236 .38801 Metetra
.419 .601 .921 r
.61236 .38801 .61925 .39774 .6331 .40022 .62622 .38983 Metetra
.458 .606 .91 r
.62622 .38983 .6331 .40022 .64706 .40205 .64018 .39104 Metetra
.499 .614 .897 r
.64018 .39104 .64706 .40205 .6611 .40298 .65423 .39144 Metetra
.539 .624 .884 r
.65423 .39144 .6611 .40298 .67521 .40279 .66834 .3908 Metetra
.577 .635 .872 r
.66834 .3908 .67521 .40279 .68936 .40128 .6825 .38898 Metetra
.613 .647 .86 r
.6825 .38898 .68936 .40128 .70352 .39833 .69668 .38583 Metetra
.646 .66 .849 r
.69668 .38583 .70352 .39833 .71768 .39385 .71087 .3813 Metetra
.675 .673 .84 r
.71087 .3813 .71768 .39385 .73181 .38782 .72504 .37536 Metetra
.702 .687 .832 r
.72504 .37536 .73181 .38782 .7459 .38028 .73919 .36805 Metetra
.727 .702 .827 r
.73919 .36805 .7459 .38028 .75992 .37132 .75328 .35944 Metetra
.748 .717 .823 r
.75328 .35944 .75992 .37132 .77389 .36111 .76733 .34969 Metetra
.767 .734 .822 r
.76733 .34969 .77389 .36111 .78778 .34982 .78133 .33896 Metetra
.784 .751 .822 r
.78133 .33896 .78778 .34982 .80162 .33771 .79528 .32748 Metetra
.798 .768 .826 r
.79528 .32748 .80162 .33771 .81541 .32503 .80919 .31547 Metetra
.809 .787 .832 r
.80919 .31547 .81541 .32503 .82917 .31206 .82308 .3032 Metetra
.26 .587 .946 r
.17785 .50649 .18597 .51358 .19687 .51685 .18885 .50891 Metetra
.319 .591 .942 r
.18885 .50891 .19687 .51685 .20789 .5199 .19996 .51113 Metetra
.381 .597 .931 r
.19996 .51113 .20789 .5199 .21906 .52248 .2112 .51293 Metetra
.44 .605 .916 r
.2112 .51293 .21906 .52248 .23038 .52437 .22259 .51411 Metetra
.494 .613 .899 r
.22259 .51411 .23038 .52437 .24186 .52535 .23413 .51449 Metetra
.541 .621 .881 r
.23413 .51449 .24186 .52535 .25353 .52526 .24584 .51391 Metetra
.581 .629 .864 r
.24584 .51391 .25353 .52526 .26536 .52394 .25772 .51226 Metetra
.616 .638 .85 r
.25772 .51226 .26536 .52394 .27737 .52131 .26975 .50944 Metetra
.646 .648 .838 r
.26975 .50944 .27737 .52131 .28954 .51734 .28194 .50543 Metetra
.672 .658 .828 r
.28194 .50543 .28954 .51734 .30185 .51202 .29426 .50024 Metetra
.695 .669 .821 r
.29426 .50024 .30185 .51202 .31429 .50542 .3067 .49391 Metetra
.714 .681 .817 r
.3067 .49391 .31429 .50542 .32683 .49766 .31925 .48654 Metetra
.732 .694 .815 r
.31925 .48654 .32683 .49766 .33946 .48887 .33188 .47828 Metetra
.747 .708 .815 r
.33188 .47828 .33946 .48887 .35216 .47927 .34457 .46929 Metetra
.76 .723 .818 r
.34457 .46929 .35216 .47927 .3649 .46905 .35732 .45975 Metetra
.77 .739 .824 r
.35732 .45975 .3649 .46905 .37766 .45848 .3701 .44989 Metetra
.778 .756 .833 r
.3701 .44989 .37766 .45848 .39045 .44778 .38289 .43992 Metetra
.783 .774 .844 r
.38289 .43992 .39045 .44778 .40323 .43722 .3957 .43007 Metetra
.785 .792 .859 r
.3957 .43007 .40323 .43722 .41601 .42704 .40851 .42054 Metetra
.781 .81 .876 r
.40851 .42054 .41601 .42704 .42879 .41745 .42132 .41153 Metetra
.771 .827 .896 r
.42132 .41153 .42879 .41745 .44156 .40866 .43413 .40323 Metetra
.753 .841 .918 r
.43413 .40323 .44156 .40866 .45433 .40083 .44695 .39578 Metetra
.725 .85 .941 r
.44695 .39578 .45433 .40083 .46712 .3941 .45979 .3893 Metetra
.687 .852 .962 r
.45979 .3893 .46712 .3941 .47993 .38856 .47265 .38386 Metetra
.639 .845 .979 r
.47265 .38386 .47993 .38856 .49278 .38425 .48555 .37952 Metetra
.584 .827 .991 r
.48555 .37952 .49278 .38425 .50567 .38119 .4985 .37627 Metetra
.526 .799 .996 r
.4985 .37627 .50567 .38119 .51864 .37931 .51152 .37408 Metetra
.472 .766 .995 r
.51152 .37408 .51864 .37931 .5317 .37855 .52461 .37288 Metetra
.428 .73 .989 r
.52461 .37288 .5317 .37855 .54485 .37878 .5378 .37254 Metetra
.399 .697 .981 r
.5378 .37254 .54485 .37878 .55811 .37984 .5511 .37294 Metetra
.386 .668 .971 r
.5511 .37294 .55811 .37984 .57149 .38152 .5645 .37388 Metetra
.389 .646 .959 r
.5645 .37388 .57149 .38152 .58499 .38361 .57802 .37518 Metetra
.404 .63 .946 r
.57802 .37518 .58499 .38361 .59862 .38586 .59166 .37661 Metetra
.428 .621 .932 r
.59166 .37661 .59862 .38586 .61236 .38801 .60541 .37796 Metetra
.458 .615 .917 r
.60541 .37796 .61236 .38801 .62622 .38983 .61928 .379 Metetra
.49 .614 .902 r
.61928 .379 .62622 .38983 .64018 .39104 .63324 .3795 Metetra
.524 .616 .887 r
.63324 .3795 .64018 .39104 .65423 .39144 .64728 .37928 Metetra
.556 .621 .872 r
.64728 .37928 .65423 .39144 .66834 .3908 .6614 .37814 Metetra
.587 .628 .859 r
.6614 .37814 .66834 .3908 .6825 .38898 .67557 .37595 Metetra
.617 .637 .848 r
.67557 .37595 .6825 .38898 .69668 .38583 .68977 .37259 Metetra
.644 .647 .838 r
.68977 .37259 .69668 .38583 .71087 .3813 .70398 .36801 Metetra
.669 .659 .831 r
.70398 .36801 .71087 .3813 .72504 .37536 .71819 .36217 Metetra
.693 .672 .825 r
.71819 .36217 .72504 .37536 .73919 .36805 .73239 .35513 Metetra
.715 .687 .822 r
.73239 .35513 .73919 .36805 .75328 .35944 .74656 .34694 Metetra
.735 .703 .821 r
.74656 .34694 .75328 .35944 .76733 .34969 .76069 .33774 Metetra
.753 .72 .822 r
.76069 .33774 .76733 .34969 .78133 .33896 .77479 .32767 Metetra
.769 .739 .825 r
.77479 .32767 .78133 .33896 .79528 .32748 .78886 .31692 Metetra
.783 .759 .831 r
.78886 .31692 .79528 .32748 .80919 .31547 .8029 .30571 Metetra
.794 .78 .84 r
.8029 .30571 .80919 .31547 .82308 .3032 .81694 .29425 Metetra
.347 .632 .959 r
.16967 .49935 .17785 .50649 .18885 .50891 .18077 .50081 Metetra
.395 .626 .946 r
.18077 .50081 .18885 .50891 .19996 .51113 .19198 .50209 Metetra
.443 .621 .927 r
.19198 .50209 .19996 .51113 .2112 .51293 .20332 .50301 Metetra
.487 .619 .907 r
.20332 .50301 .2112 .51293 .22259 .51411 .21478 .5034 Metetra
.526 .619 .887 r
.21478 .5034 .22259 .51411 .23413 .51449 .2264 .50311 Metetra
.561 .62 .868 r
.2264 .50311 .23413 .51449 .24584 .51391 .23816 .50199 Metetra
.592 .624 .852 r
.23816 .50199 .24584 .51391 .25772 .51226 .25007 .49995 Metetra
.62 .63 .839 r
.25007 .49995 .25772 .51226 .26975 .50944 .26213 .49693 Metetra
.644 .637 .829 r
.26213 .49693 .26975 .50944 .28194 .50543 .27433 .49288 Metetra
.665 .646 .821 r
.27433 .49288 .28194 .50543 .29426 .50024 .28666 .48783 Metetra
.685 .657 .816 r
.28666 .48783 .29426 .50024 .3067 .49391 .2991 .48181 Metetra
.703 .669 .814 r
.2991 .48181 .3067 .49391 .31925 .48654 .31164 .47491 Metetra
.719 .682 .814 r
.31164 .47491 .31925 .48654 .33188 .47828 .32426 .46723 Metetra
.733 .697 .817 r
.32426 .46723 .33188 .47828 .34457 .46929 .33696 .45893 Metetra
.745 .714 .822 r
.33696 .45893 .34457 .46929 .35732 .45975 .3497 .45017 Metetra
.756 .732 .83 r
.3497 .45017 .35732 .45975 .3701 .44989 .36247 .44112 Metetra
.764 .751 .841 r
.36247 .44112 .3701 .44989 .38289 .43992 .37528 .43197 Metetra
.77 .771 .854 r
.37528 .43197 .38289 .43992 .3957 .43007 .3881 .42291 Metetra
.771 .792 .87 r
.3881 .42291 .3957 .43007 .40851 .42054 .40093 .41413 Metetra
.768 .813 .889 r
.40093 .41413 .40851 .42054 .42132 .41153 .41377 .40579 Metetra
.758 .833 .909 r
.41377 .40579 .42132 .41153 .43413 .40323 .42662 .39804 Metetra
.741 .849 .93 r
.42662 .39804 .43413 .40323 .44695 .39578 .43949 .39102 Metetra
.714 .859 .951 r
.43949 .39102 .44695 .39578 .45979 .3893 .45237 .38482 Metetra
.677 .862 .97 r
.45237 .38482 .45979 .3893 .47265 .38386 .46528 .37952 Metetra
.633 .856 .984 r
.46528 .37952 .47265 .38386 .48555 .37952 .47824 .37514 Metetra
.583 .84 .993 r
.47824 .37514 .48555 .37952 .4985 .37627 .49124 .3717 Metetra
.534 .816 .998 r
.49124 .3717 .4985 .37627 .51152 .37408 .50431 .36915 Metetra
.491 .786 .997 r
.50431 .36915 .51152 .37408 .52461 .37288 .51746 .36745 Metetra
.458 .755 .993 r
.51746 .36745 .52461 .37288 .5378 .37254 .53069 .36648 Metetra
.438 .725 .986 r
.53069 .36648 .5378 .37254 .5511 .37294 .54402 .36613 Metetra
.432 .697 .976 r
.54402 .36613 .5511 .37294 .5645 .37388 .55745 .36624 Metetra
.437 .674 .963 r
.55745 .36624 .5645 .37388 .57802 .37518 .57099 .36665 Metetra
.452 .655 .947 r
.57099 .36665 .57802 .37518 .59166 .37661 .58464 .36716 Metetra
.472 .64 .93 r
.58464 .36716 .59166 .37661 .60541 .37796 .59841 .3676 Metetra
.496 .63 .911 r
.59841 .3676 .60541 .37796 .61928 .379 .61227 .36777 Metetra
.522 .623 .893 r
.61227 .36777 .61928 .379 .63324 .3795 .62623 .36748 Metetra
.547 .62 .877 r
.62623 .36748 .63324 .3795 .64728 .37928 .64027 .36657 Metetra
.573 .62 .861 r
.64027 .36657 .64728 .37928 .6614 .37814 .65439 .36487 Metetra
.597 .623 .849 r
.65439 .36487 .6614 .37814 .67557 .37595 .66856 .36227 Metetra
.62 .629 .838 r
.66856 .36227 .67557 .37595 .68977 .37259 .68278 .35868 Metetra
.642 .637 .829 r
.68278 .35868 .68977 .37259 .70398 .36801 .69701 .35404 Metetra
.663 .647 .823 r
.69701 .35404 .70398 .36801 .71819 .36217 .71126 .34833 Metetra
.683 .659 .82 r
.71126 .34833 .71819 .36217 .73239 .35513 .72551 .34159 Metetra
.702 .673 .818 r
.72551 .34159 .73239 .35513 .74656 .34694 .73975 .33388 Metetra
.72 .689 .819 r
.73975 .33388 .74656 .34694 .76069 .33774 .75397 .32529 Metetra
.737 .707 .823 r
.75397 .32529 .76069 .33774 .77479 .32767 .76818 .31597 Metetra
.752 .727 .829 r
.76818 .31597 .77479 .32767 .78886 .31692 .78236 .30606 Metetra
.766 .749 .837 r
.78236 .30606 .78886 .31692 .8029 .30571 .79655 .29575 Metetra
.778 .773 .849 r
.79655 .29575 .8029 .30571 .81694 .29425 .81073 .28522 Metetra
.43 .671 .963 r
.16141 .49214 .16967 .49935 .18077 .50081 .17264 .49255 Metetra
.465 .655 .943 r
.17264 .49255 .18077 .50081 .19198 .50209 .18397 .49281 Metetra
.498 .641 .919 r
.18397 .49281 .19198 .50209 .20332 .50301 .1954 .49277 Metetra
.529 .631 .896 r
.1954 .49277 .20332 .50301 .21478 .5034 .20696 .4923 Metetra
.556 .625 .875 r
.20696 .4923 .21478 .5034 .2264 .50311 .21865 .49127 Metetra
.581 .621 .857 r
.21865 .49127 .2264 .50311 .23816 .50199 .23047 .48956 Metetra
.603 .621 .842 r
.23047 .48956 .23816 .50199 .25007 .49995 .24243 .48711 Metetra
.623 .624 .83 r
.24243 .48711 .25007 .49995 .26213 .49693 .25451 .48386 Metetra
.642 .629 .822 r
.25451 .48386 .26213 .49693 .27433 .49288 .26672 .47979 Metetra
.659 .636 .816 r
.26672 .47979 .27433 .49288 .28666 .48783 .27905 .47489 Metetra
.675 .646 .813 r
.27905 .47489 .28666 .48783 .2991 .48181 .29149 .46921 Metetra
.69 .658 .813 r
.29149 .46921 .2991 .48181 .31164 .47491 .30402 .46281 Metetra
.704 .671 .815 r
.30402 .46281 .31164 .47491 .32426 .46723 .31662 .45578 Metetra
.717 .687 .82 r
.31662 .45578 .32426 .46723 .33696 .45893 .3293 .44824 Metetra
.729 .705 .827 r
.3293 .44824 .33696 .45893 .3497 .45017 .34203 .44032 Metetra
.739 .724 .837 r
.34203 .44032 .3497 .45017 .36247 .44112 .3548 .43217 Metetra
.748 .746 .85 r
.3548 .43217 .36247 .44112 .37528 .43197 .3676 .42393 Metetra
.753 .768 .865 r
.3676 .42393 .37528 .43197 .3881 .42291 .38043 .41575 Metetra
.755 .792 .883 r
.38043 .41575 .3881 .42291 .40093 .41413 .39328 .40779 Metetra
.752 .815 .902 r
.39328 .40779 .40093 .41413 .41377 .40579 .40614 .40019 Metetra
.742 .837 .923 r
.40614 .40019 .41377 .40579 .42662 .39804 .41903 .39306 Metetra
.725 .854 .943 r
.41903 .39306 .42662 .39804 .43949 .39102 .43193 .38652 Metetra
.699 .866 .961 r
.43193 .38652 .43949 .39102 .45237 .38482 .44486 .38063 Metetra
.666 .87 .976 r
.44486 .38063 .45237 .38482 .46528 .37952 .45783 .37547 Metetra
.626 .865 .987 r
.45783 .37547 .46528 .37952 .47824 .37514 .47084 .37106 Metetra
.585 .852 .995 r
.47084 .37106 .47824 .37514 .49124 .3717 .4839 .3674 Metetra
.546 .831 .998 r
.4839 .3674 .49124 .3717 .50431 .36915 .49702 .36447 Metetra
.515 .807 .999 r
.49702 .36447 .50431 .36915 .51746 .36745 .51022 .36221 Metetra
.493 .78 .996 r
.51022 .36221 .51746 .36745 .53069 .36648 .5235 .36055 Metetra
.482 .752 .989 r
.5235 .36055 .53069 .36648 .54402 .36613 .53687 .35938 Metetra
.481 .725 .978 r
.53687 .35938 .54402 .36613 .55745 .36624 .55033 .35858 Metetra
.487 .7 .963 r
.55033 .35858 .55745 .36624 .57099 .36665 .5639 .35802 Metetra
.5 .678 .945 r
.5639 .35802 .57099 .36665 .58464 .36716 .57757 .35753 Metetra
.516 .659 .925 r
.57757 .35753 .58464 .36716 .59841 .3676 .59134 .35698 Metetra
.533 .644 .904 r
.59134 .35698 .59841 .3676 .61227 .36777 .60521 .3562 Metetra
.552 .632 .885 r
.60521 .3562 .61227 .36777 .62623 .36748 .61916 .35504 Metetra
.57 .625 .867 r
.61916 .35504 .62623 .36748 .64027 .36657 .63321 .35337 Metetra
.588 .621 .852 r
.63321 .35337 .64027 .36657 .65439 .36487 .64732 .35107 Metetra
.606 .62 .839 r
.64732 .35107 .65439 .36487 .66856 .36227 .6615 .34803 Metetra
.624 .623 .829 r
.6615 .34803 .66856 .36227 .68278 .35868 .67572 .34418 Metetra
.641 .628 .822 r
.67572 .34418 .68278 .35868 .69701 .35404 .68998 .33948 Metetra
.657 .636 .817 r
.68998 .33948 .69701 .35404 .71126 .34833 .70427 .33392 Metetra
.674 .647 .815 r
.70427 .33392 .71126 .34833 .72551 .34159 .71857 .32752 Metetra
.689 .66 .816 r
.71857 .32752 .72551 .34159 .73975 .33388 .73287 .32033 Metetra
.705 .676 .819 r
.73287 .32033 .73975 .33388 .75397 .32529 .74718 .31242 Metetra
.72 .694 .825 r
.74718 .31242 .75397 .32529 .76818 .31597 .76149 .30391 Metetra
.734 .715 .833 r
.76149 .30391 .76818 .31597 .78236 .30606 .77581 .29493 Metetra
.747 .739 .844 r
.77581 .29493 .78236 .30606 .79655 .29575 .79013 .28561 Metetra
.758 .764 .858 r
.79013 .28561 .79655 .29575 .81073 .28522 .80447 .27611 Metetra
.504 .702 .959 r
.15309 .48488 .16141 .49214 .17264 .49255 .16446 .48416 Metetra
.527 .679 .934 r
.16446 .48416 .17264 .49255 .18397 .49281 .17591 .48333 Metetra
.548 .659 .909 r
.17591 .48333 .18397 .49281 .1954 .49277 .18746 .48226 Metetra
.566 .643 .885 r
.18746 .48226 .1954 .49277 .20696 .4923 .19912 .48086 Metetra
.583 .631 .864 r
.19912 .48086 .20696 .4923 .21865 .49127 .21089 .47904 Metetra
.599 .623 .847 r
.21089 .47904 .21865 .49127 .23047 .48956 .22277 .47672 Metetra
.613 .62 .833 r
.22277 .47672 .23047 .48956 .24243 .48711 .23477 .47383 Metetra
.627 .619 .823 r
.23477 .47383 .24243 .48711 .25451 .48386 .24688 .47034 Metetra
.64 .622 .816 r
.24688 .47034 .25451 .48386 .26672 .47979 .25911 .46623 Metetra
.653 .628 .812 r
.25911 .46623 .26672 .47979 .27905 .47489 .27144 .4615 Metetra
.666 .636 .811 r
.27144 .4615 .27905 .47489 .29149 .46921 .28386 .45618 Metetra
.678 .647 .812 r
.28386 .45618 .29149 .46921 .30402 .46281 .29637 .45032 Metetra
.689 .661 .817 r
.29637 .45032 .30402 .46281 .31662 .45578 .30895 .44399 Metetra
.7 .677 .824 r
.30895 .44399 .31662 .45578 .3293 .44824 .3216 .43727 Metetra
.711 .695 .833 r
.3216 .43727 .3293 .44824 .34203 .44032 .33431 .43026 Metetra
.72 .716 .846 r
.33431 .43026 .34203 .44032 .3548 .43217 .34706 .42306 Metetra
.728 .739 .861 r
.34706 .42306 .3548 .43217 .3676 .42393 .35985 .4158 Metetra
.733 .764 .878 r
.35985 .4158 .3676 .42393 .38043 .41575 .37268 .40857 Metetra
.735 .79 .897 r
.37268 .40857 .38043 .41575 .39328 .40779 .38554 .4015 Metetra
.731 .815 .917 r
.38554 .4015 .39328 .40779 .40614 .40019 .39842 .39469 Metetra
.722 .839 .936 r
.39842 .39469 .40614 .40019 .41903 .39306 .41134 .38823 Metetra
.706 .858 .955 r
.41134 .38823 .41903 .39306 .43193 .38652 .42428 .3822 Metetra
.682 .87 .97 r
.42428 .3822 .43193 .38652 .44486 .38063 .43726 .37667 Metetra
.653 .875 .981 r
.43726 .37667 .44486 .38063 .45783 .37547 .45028 .37166 Metetra
.62 .872 .989 r
.45028 .37166 .45783 .37547 .47084 .37106 .46335 .36722 Metetra
.589 .862 .994 r
.46335 .36722 .47084 .37106 .4839 .3674 .47647 .36333 Metetra
.562 .846 .997 r
.47647 .36333 .4839 .3674 .49702 .36447 .48965 .35998 Metetra
.542 .826 .998 r
.48965 .35998 .49702 .36447 .51022 .36221 .5029 .35713 Metetra
.531 .803 .996 r
.5029 .35713 .51022 .36221 .5235 .36055 .51623 .35472 Metetra
.527 .778 .989 r
.51623 .35472 .5235 .36055 .53687 .35938 .52965 .35268 Metetra
.53 .751 .977 r
.52965 .35268 .53687 .35938 .55033 .35858 .54315 .3509 Metetra
.537 .725 .96 r
.54315 .3509 .55033 .35858 .5639 .35802 .55674 .34929 Metetra
.546 .699 .94 r
.55674 .34929 .5639 .35802 .57757 .35753 .57043 .34774 Metetra
.557 .676 .918 r
.57043 .34774 .57757 .35753 .59134 .35698 .58421 .34612 Metetra
.568 .657 .897 r
.58421 .34612 .59134 .35698 .60521 .3562 .59808 .34433 Metetra
.58 .642 .876 r
.59808 .34433 .60521 .3562 .61916 .35504 .61204 .34225 Metetra
.592 .63 .859 r
.61204 .34225 .61916 .35504 .63321 .35337 .62608 .33977 Metetra
.604 .623 .844 r
.62608 .33977 .63321 .35337 .64732 .35107 .6402 .33682 Metetra
.615 .619 .832 r
.6402 .33682 .64732 .35107 .6615 .34803 .65437 .33331 Metetra
.627 .619 .823 r
.65437 .33331 .6615 .34803 .67572 .34418 .66861 .3292 Metetra
.64 .622 .816 r
.66861 .3292 .67572 .34418 .68998 .33948 .68289 .32444 Metetra
.652 .628 .813 r
.68289 .32444 .68998 .33948 .70427 .33392 .69721 .31904 Metetra
.664 .637 .812 r
.69721 .31904 .70427 .33392 .71857 .32752 .71156 .313 Metetra
.677 .649 .815 r
.71156 .313 .71857 .32752 .73287 .32033 .72593 .30636 Metetra
.689 .664 .82 r
.72593 .30636 .73287 .32033 .74718 .31242 .74033 .29919 Metetra
.702 .682 .828 r
.74033 .29919 .74718 .31242 .76149 .30391 .75474 .29156 Metetra
.714 .703 .838 r
.75474 .29156 .76149 .30391 .77581 .29493 .76918 .28356 Metetra
.725 .727 .852 r
.76918 .28356 .77581 .29493 .79013 .28561 .78365 .2753 Metetra
.735 .755 .868 r
.78365 .2753 .79013 .28561 .80447 .27611 .79815 .2669 Metetra
.568 .726 .948 r
.1447 .47755 .15309 .48488 .16446 .48416 .15621 .47565 Metetra
.58 .697 .922 r
.15621 .47565 .16446 .48416 .17591 .48333 .1678 .47367 Metetra
.591 .673 .897 r
.1678 .47367 .17591 .48333 .18746 .48226 .17948 .47153 Metetra
.6 .653 .874 r
.17948 .47153 .18746 .48226 .19912 .48086 .19124 .46916 Metetra
.608 .637 .854 r
.19124 .46916 .19912 .48086 .21089 .47904 .2031 .46652 Metetra
.616 .626 .839 r
.2031 .46652 .21089 .47904 .22277 .47672 .21505 .46354 Metetra
.623 .62 .826 r
.21505 .46354 .22277 .47672 .23477 .47383 .2271 .4602 Metetra
.631 .617 .817 r
.2271 .4602 .23477 .47383 .24688 .47034 .23924 .45646 Metetra
.639 .617 .812 r
.23924 .45646 .24688 .47034 .25911 .46623 .25148 .45231 Metetra
.647 .621 .809 r
.25148 .45231 .25911 .46623 .27144 .4615 .2638 .44776 Metetra
.656 .628 .81 r
.2638 .44776 .27144 .4615 .28386 .45618 .27621 .44283 Metetra
.665 .638 .813 r
.27621 .44283 .28386 .45618 .29637 .45032 .28869 .43753 Metetra
.673 .651 .82 r
.28869 .43753 .29637 .45032 .30895 .44399 .30125 .43193 Metetra
.682 .667 .829 r
.30125 .43193 .30895 .44399 .3216 .43727 .31386 .42607 Metetra
.69 .686 .84 r
.31386 .42607 .3216 .43727 .33431 .43026 .32654 .42001 Metetra
.698 .707 .855 r
.32654 .42001 .33431 .43026 .34706 .42306 .33927 .41382 Metetra
.705 .732 .872 r
.33927 .41382 .34706 .42306 .35985 .4158 .35204 .40758 Metetra
.709 .758 .891 r
.35204 .40758 .35985 .4158 .37268 .40857 .36486 .40136 Metetra
.71 .786 .911 r
.36486 .40136 .37268 .40857 .38554 .4015 .37772 .39523 Metetra
.707 .813 .931 r
.37772 .39523 .38554 .4015 .39842 .39469 .39062 .38926 Metetra
.698 .838 .95 r
.39062 .38926 .39842 .39469 .41134 .38823 .40356 .38351 Metetra
.683 .858 .966 r
.40356 .38351 .41134 .38823 .42428 .3822 .41654 .37803 Metetra
.663 .872 .978 r
.41654 .37803 .42428 .3822 .43726 .37667 .42957 .37285 Metetra
.639 .878 .985 r
.42957 .37285 .43726 .37667 .45028 .37166 .44264 .36802 Metetra
.616 .877 .99 r
.44264 .36802 .45028 .37166 .46335 .36722 .45577 .36354 Metetra
.595 .87 .993 r
.45577 .36354 .46335 .36722 .47647 .36333 .46895 .35941 Metetra
.58 .859 .995 r
.46895 .35941 .47647 .36333 .48965 .35998 .48219 .35563 Metetra
.572 .843 .996 r
.48219 .35563 .48965 .35998 .5029 .35713 .49551 .35215 Metetra
.57 .824 .993 r
.49551 .35215 .5029 .35713 .51623 .35472 .50889 .34896 Metetra
.572 .8 .985 r
.50889 .34896 .51623 .35472 .52965 .35268 .52235 .34599 Metetra
.577 .773 .972 r
.52235 .34599 .52965 .35268 .54315 .3509 .5359 .34318 Metetra
.583 .745 .954 r
.5359 .34318 .54315 .3509 .55674 .34929 .54952 .34048 Metetra
.589 .718 .933 r
.54952 .34048 .55674 .34929 .57043 .34774 .56323 .3378 Metetra
.595 .692 .91 r
.56323 .3378 .57043 .34774 .58421 .34612 .57703 .33507 Metetra
.601 .67 .888 r
.57703 .33507 .58421 .34612 .59808 .34433 .59091 .33221 Metetra
.606 .651 .868 r
.59091 .33221 .59808 .34433 .61204 .34225 .60487 .32916 Metetra
.612 .637 .851 r
.60487 .32916 .61204 .34225 .62608 .33977 .61891 .32585 Metetra
.618 .626 .837 r
.61891 .32585 .62608 .33977 .6402 .33682 .63302 .32222 Metetra
.625 .62 .825 r
.63302 .32222 .6402 .33682 .65437 .33331 .6472 .31823 Metetra
.631 .617 .817 r
.6472 .31823 .65437 .33331 .66861 .3292 .66144 .31383 Metetra
.639 .617 .812 r
.66144 .31383 .66861 .3292 .68289 .32444 .67574 .30902 Metetra
.646 .621 .81 r
.67574 .30902 .68289 .32444 .69721 .31904 .69009 .30378 Metetra
.655 .628 .811 r
.69009 .30378 .69721 .31904 .71156 .313 .70449 .29812 Metetra
.664 .638 .815 r
.70449 .29812 .71156 .313 .72593 .30636 .71893 .29207 Metetra
.673 .652 .821 r
.71893 .29207 .72593 .30636 .74033 .29919 .73341 .28566 Metetra
.682 .67 .831 r
.73341 .28566 .74033 .29919 .75474 .29156 .74793 .27895 Metetra
.692 .691 .844 r
.74793 .27895 .75474 .29156 .76918 .28356 .7625 .27199 Metetra
.701 .715 .86 r
.7625 .27199 .76918 .28356 .78365 .2753 .77711 .26485 Metetra
.709 .744 .879 r
.77711 .26485 .78365 .2753 .79815 .2669 .79176 .25762 Metetra
.622 .743 .933 r
.13623 .47017 .1447 .47755 .15621 .47565 .14791 .46704 Metetra
.626 .712 .908 r
.14791 .46704 .15621 .47565 .1678 .47367 .15965 .46387 Metetra
.628 .684 .884 r
.15965 .46387 .1678 .47367 .17948 .47153 .17146 .46062 Metetra
.629 .662 .863 r
.17146 .46062 .17948 .47153 .19124 .46916 .18334 .45726 Metetra
.63 .644 .845 r
.18334 .45726 .19124 .46916 .2031 .46652 .19529 .45378 Metetra
.631 .631 .831 r
.19529 .45378 .2031 .46652 .21505 .46354 .20731 .45013 Metetra
.633 .621 .821 r
.20731 .45013 .21505 .46354 .2271 .4602 .21941 .44632 Metetra
.635 .616 .813 r
.21941 .44632 .2271 .4602 .23924 .45646 .23158 .44232 Metetra
.638 .614 .809 r
.23158 .44232 .23924 .45646 .25148 .45231 .24383 .43814 Metetra
.642 .616 .808 r
.24383 .43814 .25148 .45231 .2638 .44776 .25615 .43378 Metetra
.646 .621 .81 r
.25615 .43378 .2638 .44776 .27621 .44283 .26853 .42923 Metetra
.651 .63 .815 r
.26853 .42923 .27621 .44283 .28869 .43753 .28098 .42452 Metetra
.656 .641 .823 r
.28098 .42452 .28869 .43753 .30125 .43193 .2935 .41967 Metetra
.662 .657 .834 r
.2935 .41967 .30125 .43193 .31386 .42607 .30608 .41469 Metetra
.668 .676 .848 r
.30608 .41469 .31386 .42607 .32654 .42001 .31872 .40961 Metetra
.673 .698 .865 r
.31872 .40961 .32654 .42001 .33927 .41382 .33142 .40447 Metetra
.678 .723 .884 r
.33142 .40447 .33927 .41382 .35204 .40758 .34417 .39929 Metetra
.68 .751 .904 r
.34417 .39929 .35204 .40758 .36486 .40136 .35697 .39411 Metetra
.681 .78 .926 r
.35697 .39411 .36486 .40136 .37772 .39523 .36983 .38895 Metetra
.677 .808 .945 r
.36983 .38895 .37772 .39523 .39062 .38926 .38274 .38385 Metetra
.669 .834 .963 r
.38274 .38385 .39062 .38926 .40356 .38351 .3957 .37883 Metetra
.657 .855 .976 r
.3957 .37883 .40356 .38351 .41654 .37803 .40871 .37392 Metetra
.642 .87 .984 r
.40871 .37392 .41654 .37803 .42957 .37285 .42178 .36913 Metetra
.626 .878 .988 r
.42178 .36913 .42957 .37285 .44264 .36802 .43491 .36447 Metetra
.612 .88 .991 r
.43491 .36447 .44264 .36802 .45577 .36354 .44809 .35995 Metetra
.603 .877 .992 r
.44809 .35995 .45577 .36354 .46895 .35941 .46134 .35558 Metetra
.6 .87 .993 r
.46134 .35558 .46895 .35941 .48219 .35563 .47465 .35134 Metetra
.602 .858 .992 r
.47465 .35134 .48219 .35563 .49551 .35215 .48802 .34722 Metetra
.608 .841 .988 r
.48802 .34722 .49551 .35215 .50889 .34896 .50147 .34321 Metetra
.615 .818 .978 r
.50147 .34321 .50889 .34896 .52235 .34599 .51499 .33928 Metetra
.621 .792 .964 r
.51499 .33928 .52235 .34599 .5359 .34318 .52857 .33542 Metetra
.626 .763 .944 r
.52857 .33542 .5359 .34318 .54952 .34048 .54224 .33158 Metetra
.629 .733 .923 r
.54224 .33158 .54952 .34048 .56323 .3378 .55598 .32773 Metetra
.63 .706 .901 r
.55598 .32773 .56323 .3378 .57703 .33507 .56979 .32385 Metetra
.631 .681 .879 r
.56979 .32385 .57703 .33507 .59091 .33221 .58368 .31991 Metetra
.631 .661 .86 r
.58368 .31991 .59091 .33221 .60487 .32916 .59765 .31586 Metetra
.632 .644 .844 r
.59765 .31586 .60487 .32916 .61891 .32585 .61169 .31169 Metetra
.632 .631 .831 r
.61169 .31169 .61891 .32585 .63302 .32222 .6258 .30737 Metetra
.634 .622 .821 r
.6258 .30737 .63302 .32222 .6472 .31823 .63998 .30288 Metetra
.636 .616 .814 r
.63998 .30288 .6472 .31823 .66144 .31383 .65423 .2982 Metetra
.638 .614 .81 r
.65423 .2982 .66144 .31383 .67574 .30902 .66855 .29332 Metetra
.642 .616 .809 r
.66855 .29332 .67574 .30902 .69009 .30378 .68293 .28825 Metetra
.646 .621 .811 r
.68293 .28825 .69009 .30378 .70449 .29812 .69737 .28298 Metetra
.65 .629 .816 r
.69737 .28298 .70449 .29812 .71893 .29207 .71188 .27753 Metetra
.656 .642 .824 r
.71188 .27753 .71893 .29207 .73341 .28566 .72644 .27191 Metetra
.662 .658 .836 r
.72644 .27191 .73341 .28566 .74793 .27895 .74107 .26614 Metetra
.668 .678 .85 r
.74107 .26614 .74793 .27895 .7625 .27199 .75576 .26026 Metetra
.674 .702 .868 r
.75576 .26026 .7625 .27199 .77711 .26485 .7705 .25428 Metetra
.679 .731 .889 r
.7705 .25428 .77711 .26485 .79176 .25762 .78532 .24824 Metetra
.666 .754 .916 r
.12769 .46272 .13623 .47017 .14791 .46704 .13954 .45835 Metetra
.664 .722 .893 r
.13954 .45835 .14791 .46704 .15965 .46387 .15144 .45398 Metetra
.66 .694 .871 r
.15144 .45398 .15965 .46387 .17146 .46062 .16339 .4496 Metetra
.655 .671 .853 r
.16339 .4496 .17146 .46062 .18334 .45726 .17539 .44524 Metetra
.65 .651 .837 r
.17539 .44524 .18334 .45726 .19529 .45378 .18744 .4409 Metetra
.646 .636 .825 r
.18744 .4409 .19529 .45378 .20731 .45013 .19954 .43658 Metetra
.643 .624 .817 r
.19954 .43658 .20731 .45013 .21941 .44632 .21169 .43229 Metetra
.64 .617 .811 r
.21169 .43229 .21941 .44632 .23158 .44232 .22389 .42804 Metetra
.638 .613 .808 r
.22389 .42804 .23158 .44232 .24383 .43814 .23615 .42382 Metetra
.637 .612 .809 r
.23615 .42382 .24383 .43814 .25615 .43378 .24846 .41964 Metetra
.636 .616 .812 r
.24846 .41964 .25615 .43378 .26853 .42923 .26082 .41549 Metetra
.637 .622 .819 r
.26082 .41549 .26853 .42923 .28098 .42452 .27324 .41137 Metetra
.638 .633 .828 r
.27324 .41137 .28098 .42452 .2935 .41967 .28572 .40727 Metetra
.64 .647 .841 r
.28572 .40727 .2935 .41967 .30608 .41469 .29825 .40319 Metetra
.642 .665 .857 r
.29825 .40319 .30608 .41469 .31872 .40961 .31084 .39911 Metetra
.644 .687 .875 r
.31084 .39911 .31872 .40961 .33142 .40447 .3235 .39502 Metetra
.646 .712 .896 r
.3235 .39502 .33142 .40447 .34417 .39929 .33622 .39092 Metetra
.647 .741 .918 r
.33622 .39092 .34417 .39929 .35697 .39411 .349 .38679 Metetra
.646 .771 .939 r
.349 .38679 .35697 .39411 .36983 .38895 .36185 .38263 Metetra
.642 .8 .958 r
.36185 .38263 .36983 .38895 .38274 .38385 .37476 .37842 Metetra
.636 .827 .974 r
.37476 .37842 .38274 .38385 .3957 .37883 .38774 .37415 Metetra
.628 .849 .984 r
.38774 .37415 .3957 .37883 .40871 .37392 .40079 .36982 Metetra
.62 .865 .989 r
.40079 .36982 .40871 .37392 .42178 .36913 .4139 .36542 Metetra
.613 .876 .991 r
.4139 .36542 .42178 .36913 .43491 .36447 .42708 .36094 Metetra
.611 .881 .991 r
.42708 .36094 .43491 .36447 .44809 .35995 .44032 .35639 Metetra
.614 .882 .99 r
.44032 .35639 .44809 .35995 .46134 .35558 .45364 .35176 Metetra
.622 .878 .989 r
.45364 .35176 .46134 .35558 .47465 .35134 .46701 .34705 Metetra
.633 .869 .986 r
.46701 .34705 .47465 .35134 .48802 .34722 .48046 .34228 Metetra
.644 .853 .98 r
.48046 .34228 .48802 .34722 .50147 .34321 .49397 .33743 Metetra
.654 .832 .968 r
.49397 .33743 .50147 .34321 .51499 .33928 .50754 .33253 Metetra
.661 .805 .952 r
.50754 .33253 .51499 .33928 .52857 .33542 .52118 .32758 Metetra
.664 .776 .933 r
.52118 .32758 .52857 .33542 .54224 .33158 .53489 .32259 Metetra
.664 .747 .912 r
.53489 .32259 .54224 .33158 .55598 .32773 .54866 .31757 Metetra
.662 .718 .89 r
.54866 .31757 .55598 .32773 .56979 .32385 .5625 .31252 Metetra
.658 .692 .87 r
.5625 .31252 .56979 .32385 .58368 .31991 .5764 .30747 Metetra
.654 .67 .853 r
.5764 .30747 .58368 .31991 .59765 .31586 .59038 .30242 Metetra
.65 .651 .838 r
.59038 .30242 .59765 .31586 .61169 .31169 .60442 .29738 Metetra
.646 .636 .826 r
.60442 .29738 .61169 .31169 .6258 .30737 .61854 .29236 Metetra
.643 .625 .817 r
.61854 .29236 .6258 .30737 .63998 .30288 .63272 .28736 Metetra
.64 .617 .811 r
.63272 .28736 .63998 .30288 .65423 .2982 .64698 .2824 Metetra
.638 .613 .808 r
.64698 .2824 .65423 .2982 .66855 .29332 .66131 .27746 Metetra
.637 .612 .809 r
.66131 .27746 .66855 .29332 .68293 .28825 .67572 .27256 Metetra
.637 .615 .812 r
.67572 .27256 .68293 .28825 .69737 .28298 .69021 .26768 Metetra
.637 .622 .818 r
.69021 .26768 .69737 .28298 .71188 .27753 .70478 .26284 Metetra
.638 .632 .828 r
.70478 .26284 .71188 .27753 .72644 .27191 .71942 .25801 Metetra
.64 .646 .841 r
.71942 .25801 .72644 .27191 .74107 .26614 .73415 .2532 Metetra
.642 .665 .857 r
.73415 .2532 .74107 .26614 .75576 .26026 .74896 .2484 Metetra
.644 .688 .876 r
.74896 .2484 .75576 .26026 .7705 .25428 .76384 .24359 Metetra
.646 .716 .899 r
.76384 .24359 .7705 .25428 .78532 .24824 .77881 .23877 Metetra
.702 .761 .898 r
.11908 .4552 .12769 .46272 .13954 .45835 .13111 .44959 Metetra
.695 .73 .878 r
.13111 .44959 .13954 .45835 .15144 .45398 .14317 .44402 Metetra
.687 .703 .859 r
.14317 .44402 .15144 .45398 .16339 .4496 .15527 .43853 Metetra
.678 .679 .843 r
.15527 .43853 .16339 .4496 .17539 .44524 .16739 .43317 Metetra
.669 .658 .831 r
.16739 .43317 .17539 .44524 .18744 .4409 .17955 .42797 Metetra
.66 .642 .821 r
.17955 .42797 .18744 .4409 .19954 .43658 .19173 .42298 Metetra
.652 .629 .814 r
.19173 .42298 .19954 .43658 .21169 .43229 .20393 .41823 Metetra
.645 .619 .81 r
.20393 .41823 .21169 .43229 .22389 .42804 .21617 .41372 Metetra
.638 .613 .809 r
.21617 .41372 .22389 .42804 .23615 .42382 .22843 .40947 Metetra
.632 .611 .811 r
.22843 .40947 .23615 .42382 .24846 .41964 .24073 .40547 Metetra
.627 .611 .815 r
.24073 .40547 .24846 .41964 .26082 .41549 .25307 .40171 Metetra
.622 .616 .823 r
.25307 .40171 .26082 .41549 .27324 .41137 .26545 .39817 Metetra
.619 .625 .834 r
.26545 .39817 .27324 .41137 .28572 .40727 .27788 .39482 Metetra
.616 .637 .849 r
.27788 .39482 .28572 .40727 .29825 .40319 .29037 .39163 Metetra
.614 .654 .866 r
.29037 .39163 .29825 .40319 .31084 .39911 .30291 .38854 Metetra
.612 .675 .886 r
.30291 .38854 .31084 .39911 .3235 .39502 .31552 .38551 Metetra
.61 .7 .908 r
.31552 .38551 .3235 .39502 .33622 .39092 .3282 .38248 Metetra
.608 .728 .93 r
.3282 .38248 .33622 .39092 .349 .38679 .34096 .37941 Metetra
.606 .758 .951 r
.34096 .37941 .349 .38679 .36185 .38263 .35379 .37624 Metetra
.603 .787 .969 r
.35379 .37624 .36185 .38263 .37476 .37842 .3667 .37291 Metetra
.599 .815 .983 r
.3667 .37291 .37476 .37842 .38774 .37415 .3797 .3694 Metetra
.597 .839 .99 r
.3797 .3694 .38774 .37415 .40079 .36982 .39277 .36565 Metetra
.597 .858 .993 r
.39277 .36565 .40079 .36982 .4139 .36542 .40592 .36164 Metetra
.602 .872 .993 r
.40592 .36164 .4139 .36542 .42708 .36094 .41915 .35735 Metetra
.611 .881 .991 r
.41915 .35735 .42708 .36094 .44032 .35639 .43246 .35276 Metetra
.626 .885 .988 r
.43246 .35276 .44032 .35639 .45364 .35176 .44584 .34787 Metetra
.643 .884 .985 r
.44584 .34787 .45364 .35176 .46701 .34705 .45929 .3427 Metetra
.661 .877 .979 r
.45929 .3427 .46701 .34705 .48046 .34228 .4728 .33726 Metetra
.677 .862 .97 r
.4728 .33726 .48046 .34228 .49397 .33743 .48638 .33158 Metetra
.689 .841 .956 r
.48638 .33158 .49397 .33743 .50754 .33253 .50002 .3257 Metetra
.696 .815 .939 r
.50002 .3257 .50754 .33253 .52118 .32758 .51372 .31967 Metetra
.697 .786 .92 r
.51372 .31967 .52118 .32758 .53489 .32259 .52747 .31352 Metetra
.695 .757 .899 r
.52747 .31352 .53489 .32259 .54866 .31757 .54128 .30732 Metetra
.69 .729 .88 r
.54128 .30732 .54866 .31757 .5625 .31252 .55515 .30112 Metetra
.683 .703 .862 r
.55515 .30112 .5625 .31252 .5764 .30747 .56908 .29497 Metetra
.676 .679 .846 r
.56908 .29497 .5764 .30747 .59038 .30242 .58306 .28892 Metetra
.667 .659 .833 r
.58306 .28892 .59038 .30242 .60442 .29738 .59711 .28301 Metetra
.659 .643 .822 r
.59711 .28301 .60442 .29738 .61854 .29236 .61123 .2773 Metetra
.652 .63 .815 r
.61123 .2773 .61854 .29236 .63272 .28736 .62543 .2718 Metetra
.645 .62 .81 r
.62543 .2718 .63272 .28736 .64698 .2824 .6397 .26655 Metetra
.638 .614 .809 r
.6397 .26655 .64698 .2824 .66131 .27746 .65405 .26155 Metetra
.633 .611 .81 r
.65405 .26155 .66131 .27746 .67572 .27256 .66848 .25681 Metetra
.628 .611 .814 r
.66848 .25681 .67572 .27256 .69021 .26768 .68301 .25233 Metetra
.624 .615 .822 r
.68301 .25233 .69021 .26768 .70478 .26284 .69764 .24808 Metetra
.62 .623 .832 r
.69764 .24808 .70478 .26284 .71942 .25801 .71236 .24404 Metetra
.617 .635 .846 r
.71236 .24404 .71942 .25801 .73415 .2532 .72718 .24018 Metetra
.615 .652 .864 r
.72718 .24018 .73415 .2532 .74896 .2484 .7421 .23646 Metetra
.613 .673 .884 r
.7421 .23646 .74896 .2484 .76384 .24359 .75712 .23282 Metetra
.61 .699 .907 r
.75712 .23282 .76384 .24359 .77881 .23877 .77224 .22921 Metetra
.73 .765 .881 r
.1104 .44762 .11908 .4552 .13111 .44959 .1226 .44079 Metetra
.721 .736 .863 r
.1226 .44079 .13111 .44959 .14317 .44402 .13483 .43404 Metetra
.71 .71 .848 r
.13483 .43404 .14317 .44402 .15527 .43853 .14708 .42745 Metetra
.698 .686 .835 r
.14708 .42745 .15527 .43853 .16739 .43317 .15934 .42111 Metetra
.686 .666 .825 r
.15934 .42111 .16739 .43317 .17955 .42797 .1716 .41508 Metetra
.674 .649 .817 r
.1716 .41508 .17955 .42797 .19173 .42298 .18386 .40944 Metetra
.661 .634 .812 r
.18386 .40944 .19173 .42298 .20393 .41823 .19612 .40422 Metetra
.65 .623 .81 r
.19612 .40422 .20393 .41823 .21617 .41372 .20839 .39946 Metetra
.638 .615 .811 r
.20839 .39946 .21617 .41372 .22843 .40947 .22067 .39517 Metetra
.627 .61 .814 r
.22067 .39517 .22843 .40947 .24073 .40547 .23295 .39135 Metetra
.617 .609 .82 r
.23295 .39135 .24073 .40547 .25307 .40171 .24527 .38797 Metetra
.607 .611 .829 r
.24527 .38797 .25307 .40171 .26545 .39817 .25761 .38501 Metetra
.598 .617 .842 r
.25761 .38501 .26545 .39817 .27788 .39482 .26999 .38239 Metetra
.59 .627 .857 r
.26999 .38239 .27788 .39482 .29037 .39163 .28242 .38006 Metetra
.582 .642 .875 r
.28242 .38006 .29037 .39163 .30291 .38854 .29492 .37794 Metetra
.575 .661 .896 r
.29492 .37794 .30291 .38854 .31552 .38551 .30748 .37594 Metetra
.569 .684 .918 r
.30748 .37594 .31552 .38551 .3282 .38248 .32012 .37397 Metetra
.564 .711 .94 r
.32012 .37397 .3282 .38248 .34096 .37941 .33284 .37194 Metetra
.56 .741 .961 r
.33284 .37194 .34096 .37941 .35379 .37624 .34566 .36974 Metetra
.559 .771 .977 r
.34566 .36974 .35379 .37624 .3667 .37291 .35856 .36729 Metetra
.56 .8 .989 r
.35856 .36729 .3667 .37291 .3797 .3694 .37157 .36452 Metetra
.565 .826 .994 r
.37157 .36452 .3797 .3694 .39277 .36565 .38466 .36135 Metetra
.576 .848 .996 r
.38466 .36135 .39277 .36565 .40592 .36164 .39786 .35772 Metetra
.592 .866 .994 r
.39786 .35772 .40592 .36164 .41915 .35735 .41114 .35361 Metetra
.614 .879 .99 r
.41114 .35361 .41915 .35735 .43246 .35276 .4245 .34898 Metetra
.639 .886 .986 r
.4245 .34898 .43246 .35276 .44584 .34787 .43795 .34384 Metetra
.664 .887 .979 r
.43795 .34384 .44584 .34787 .45929 .3427 .45147 .33821 Metetra
.688 .88 .97 r
.45147 .33821 .45929 .3427 .4728 .33726 .46506 .33211 Metetra
.706 .866 .958 r
.46506 .33211 .4728 .33726 .48638 .33158 .47871 .32561 Metetra
.719 .846 .942 r
.47871 .32561 .48638 .33158 .50002 .3257 .49242 .31876 Metetra
.725 .821 .924 r
.49242 .31876 .50002 .3257 .51372 .31967 .50618 .31166 Metetra
.726 .794 .906 r
.50618 .31166 .51372 .31967 .52747 .31352 .51999 .30438 Metetra
.722 .765 .887 r
.51999 .30438 .52747 .31352 .54128 .30732 .53384 .29702 Metetra
.715 .738 .869 r
.53384 .29702 .54128 .30732 .55515 .30112 .54774 .28968 Metetra
.706 .712 .853 r
.54774 .28968 .55515 .30112 .56908 .29497 .5617 .28245 Metetra
.695 .689 .84 r
.5617 .28245 .56908 .29497 .58306 .28892 .5757 .27542 Metetra
.684 .668 .829 r
.5757 .27542 .58306 .28892 .59711 .28301 .58977 .26867 Metetra
.672 .651 .82 r
.58977 .26867 .59711 .28301 .61123 .2773 .60389 .26227 Metetra
.661 .636 .814 r
.60389 .26227 .61123 .2773 .62543 .2718 .6181 .25629 Metetra
.65 .624 .811 r
.6181 .25629 .62543 .2718 .6397 .26655 .63238 .25075 Metetra
.639 .616 .811 r
.63238 .25075 .6397 .26655 .65405 .26155 .64675 .24569 Metetra
.629 .61 .813 r
.64675 .24569 .65405 .26155 .66848 .25681 .66121 .24111 Metetra
.619 .608 .818 r
.66121 .24111 .66848 .25681 .68301 .25233 .67578 .23701 Metetra
.61 .61 .826 r
.67578 .23701 .68301 .25233 .69764 .24808 .69046 .23335 Metetra
.601 .615 .838 r
.69046 .23335 .69764 .24808 .71236 .24404 .70526 .23008 Metetra
.593 .625 .852 r
.70526 .23008 .71236 .24404 .72718 .24018 .72017 .22714 Metetra
.586 .639 .87 r
.72017 .22714 .72718 .24018 .7421 .23646 .7352 .22447 Metetra
.578 .658 .891 r
.7352 .22447 .7421 .23646 .75712 .23282 .75035 .22197 Metetra
.571 .681 .914 r
.75035 .22197 .75712 .23282 .77224 .22921 .76561 .21955 Metetra
.753 .768 .865 r
.10164 .43998 .1104 .44762 .1226 .44079 .11403 .43197 Metetra
.742 .741 .85 r
.11403 .43197 .1226 .44079 .13483 .43404 .12643 .42408 Metetra
.73 .716 .838 r
.12643 .42408 .13483 .43404 .14708 .42745 .13883 .41643 Metetra
.716 .694 .828 r
.13883 .41643 .14708 .42745 .15934 .42111 .15122 .40914 Metetra
.701 .674 .82 r
.15122 .40914 .15934 .42111 .1716 .41508 .16358 .40231 Metetra
.686 .656 .815 r
.16358 .40231 .1716 .41508 .18386 .40944 .17593 .39604 Metetra
.671 .641 .812 r
.17593 .39604 .18386 .40944 .19612 .40422 .18825 .39038 Metetra
.655 .629 .812 r
.18825 .39038 .19612 .40422 .20839 .39946 .20055 .38537 Metetra
.639 .619 .814 r
.20055 .38537 .20839 .39946 .22067 .39517 .21284 .38105 Metetra
.623 .612 .819 r
.21284 .38105 .22067 .39517 .23295 .39135 .22512 .3774 Metetra
.607 .608 .826 r
.22512 .3774 .23295 .39135 .24527 .38797 .2374 .37438 Metetra
.592 .607 .837 r
.2374 .37438 .24527 .38797 .25761 .38501 .24971 .37196 Metetra
.576 .61 .85 r
.24971 .37196 .25761 .38501 .26999 .38239 .26204 .37004 Metetra
.561 .617 .866 r
.26204 .37004 .26999 .38239 .28242 .38006 .27442 .36855 Metetra
.547 .629 .884 r
.27442 .36855 .28242 .38006 .29492 .37794 .28685 .36736 Metetra
.534 .645 .905 r
.28685 .36736 .29492 .37794 .30748 .37594 .29936 .36635 Metetra
.523 .666 .927 r
.29936 .36635 .30748 .37594 .32012 .37397 .31196 .3654 Metetra
.515 .691 .948 r
.31196 .3654 .32012 .37397 .33284 .37194 .32465 .36436 Metetra
.511 .72 .967 r
.32465 .36436 .33284 .37194 .34566 .36974 .33744 .36311 Metetra
.512 .75 .982 r
.33744 .36311 .34566 .36974 .35856 .36729 .35034 .36151 Metetra
.519 .781 .992 r
.35034 .36151 .35856 .36729 .37157 .36452 .36335 .35945 Metetra
.534 .81 .997 r
.36335 .35945 .37157 .36452 .38466 .36135 .37647 .35684 Metetra
.556 .836 .997 r
.37647 .35684 .38466 .36135 .39786 .35772 .3897 .35359 Metetra
.585 .858 .995 r
.3897 .35359 .39786 .35772 .41114 .35361 .40303 .34965 Metetra
.618 .875 .99 r
.40303 .34965 .41114 .35361 .4245 .34898 .41646 .34498 Metetra
.652 .885 .983 r
.41646 .34498 .4245 .34898 .43795 .34384 .42997 .33959 Metetra
.684 .887 .973 r
.42997 .33959 .43795 .34384 .45147 .33821 .44357 .33351 Metetra
.711 .881 .96 r
.44357 .33351 .45147 .33821 .46506 .33211 .45724 .32677 Metetra
.731 .867 .945 r
.45724 .32677 .46506 .33211 .47871 .32561 .47096 .31947 Metetra
.744 .848 .928 r
.47096 .31947 .47871 .32561 .49242 .31876 .48474 .31168 Metetra
.75 .824 .91 r
.48474 .31168 .49242 .31876 .50618 .31166 .49857 .30354 Metetra
.75 .798 .892 r
.49857 .30354 .50618 .31166 .51999 .30438 .51243 .29516 Metetra
.745 .772 .875 r
.51243 .29516 .51999 .30438 .53384 .29702 .52634 .28668 Metetra
.737 .746 .859 r
.52634 .28668 .53384 .29702 .54774 .28968 .54028 .27824 Metetra
.726 .721 .846 r
.54028 .27824 .54774 .28968 .5617 .28245 .55427 .26997 Metetra
.713 .698 .834 r
.55427 .26997 .5617 .28245 .5757 .27542 .56829 .26199 Metetra
.699 .677 .825 r
.56829 .26199 .5757 .27542 .58977 .26867 .58238 .25443 Metetra
.685 .659 .819 r
.58238 .25443 .58977 .26867 .60389 .26227 .59652 .24738 Metetra
.67 .643 .815 r
.59652 .24738 .60389 .26227 .6181 .25629 .61073 .24092 Metetra
.655 .63 .813 r
.61073 .24092 .6181 .25629 .63238 .25075 .62503 .23512 Metetra
.64 .62 .814 r
.62503 .23512 .63238 .25075 .64675 .24569 .63942 .23 Metetra
.625 .612 .817 r
.63942 .23 .64675 .24569 .66121 .24111 .65391 .22557 Metetra
.611 .607 .823 r
.65391 .22557 .66121 .24111 .67578 .23701 .66852 .22183 Metetra
.596 .606 .832 r
.66852 .22183 .67578 .23701 .69046 .23335 .68325 .21872 Metetra
.582 .609 .844 r
.68325 .21872 .69046 .23335 .70526 .23008 .69811 .21619 Metetra
.568 .615 .859 r
.69811 .21619 .70526 .23008 .72017 .22714 .71311 .21414 Metetra
.555 .626 .877 r
.71311 .21414 .72017 .22714 .7352 .22447 .72824 .21248 Metetra
.542 .641 .897 r
.72824 .21248 .7352 .22447 .75035 .22197 .74351 .21108 Metetra
.53 .661 .92 r
.74351 .21108 .75035 .22197 .76561 .21955 .75891 .20981 Metetra
.771 .769 .85 r
.0928 .43226 .10164 .43998 .11403 .43197 .10537 .42314 Metetra
.76 .745 .839 r
.10537 .42314 .11403 .43197 .12643 .42408 .11795 .41416 Metetra
.747 .722 .829 r
.11795 .41416 .12643 .42408 .13883 .41643 .1305 .40551 Metetra
.732 .701 .822 r
.1305 .40551 .13883 .41643 .15122 .40914 .14302 .39733 Metetra
.715 .682 .817 r
.14302 .39733 .15122 .40914 .16358 .40231 .15549 .38974 Metetra
.698 .665 .814 r
.15549 .38974 .16358 .40231 .17593 .39604 .16792 .38287 Metetra
.68 .65 .813 r
.16792 .38287 .17593 .39604 .18825 .39038 .1803 .37679 Metetra
.66 .636 .815 r
.1803 .37679 .18825 .39038 .20055 .38537 .19264 .37155 Metetra
.64 .625 .819 r
.19264 .37155 .20055 .38537 .21284 .38105 .20494 .36719 Metetra
.619 .615 .825 r
.20494 .36719 .21284 .38105 .22512 .3774 .21721 .36369 Metetra
.597 .609 .834 r
.21721 .36369 .22512 .3774 .2374 .37438 .22947 .36102 Metetra
.575 .605 .845 r
.22947 .36102 .2374 .37438 .24971 .37196 .24173 .35911 Metetra
.553 .604 .859 r
.24173 .35911 .24971 .37196 .26204 .37004 .25402 .35785 Metetra
.531 .608 .875 r
.25402 .35785 .26204 .37004 .27442 .36855 .26634 .35714 Metetra
.509 .616 .893 r
.26634 .35714 .27442 .36855 .28685 .36736 .27872 .35683 Metetra
.49 .628 .913 r
.27872 .35683 .28685 .36736 .29936 .36635 .29118 .35676 Metetra
.473 .646 .933 r
.29118 .35676 .29936 .36635 .31196 .3654 .30373 .35677 Metetra
.462 .668 .952 r
.30373 .35677 .31196 .3654 .32465 .36436 .31638 .35668 Metetra
.458 .695 .969 r
.31638 .35668 .32465 .36436 .33744 .36311 .32915 .35631 Metetra
.463 .726 .983 r
.32915 .35631 .33744 .36311 .35034 .36151 .34203 .35552 Metetra
.478 .759 .992 r
.34203 .35552 .35034 .36151 .36335 .35945 .35505 .35414 Metetra
.504 .792 .997 r
.35505 .35414 .36335 .35945 .37647 .35684 .36819 .35206 Metetra
.538 .823 .998 r
.36819 .35206 .37647 .35684 .3897 .35359 .38145 .34916 Metetra
.579 .849 .995 r
.38145 .34916 .3897 .35359 .40303 .34965 .39483 .34538 Metetra
.623 .869 .989 r
.39483 .34538 .40303 .34965 .41646 .34498 .40832 .34068 Metetra
.664 .881 .978 r
.40832 .34068 .41646 .34498 .42997 .33959 .42191 .33505 Metetra
.701 .884 .965 r
.42191 .33505 .42997 .33959 .44357 .33351 .43558 .32853 Metetra
.73 .878 .949 r
.43558 .32853 .44357 .33351 .45724 .32677 .44932 .32118 Metetra
.752 .865 .931 r
.44932 .32118 .45724 .32677 .47096 .31947 .46313 .31311 Metetra
.764 .847 .913 r
.46313 .31311 .47096 .31947 .48474 .31168 .47698 .30443 Metetra
.77 .825 .895 r
.47698 .30443 .48474 .31168 .49857 .30354 .49088 .2953 Metetra
.77 .801 .878 r
.49088 .2953 .49857 .30354 .51243 .29516 .50481 .28588 Metetra
.764 .777 .863 r
.50481 .28588 .51243 .29516 .52634 .28668 .51877 .27633 Metetra
.755 .753 .85 r
.51877 .27633 .52634 .28668 .54028 .27824 .53276 .26685 Metetra
.743 .73 .839 r
.53276 .26685 .54028 .27824 .55427 .26997 .54678 .25759 Metetra
.729 .708 .83 r
.54678 .25759 .55427 .26997 .56829 .26199 .56084 .24872 Metetra
.714 .687 .823 r
.56084 .24872 .56829 .26199 .58238 .25443 .57495 .24039 Metetra
.697 .669 .819 r
.57495 .24039 .58238 .25443 .59652 .24738 .58911 .23272 Metetra
.679 .652 .816 r
.58911 .23272 .59652 .24738 .61073 .24092 .60334 .22581 Metetra
.661 .638 .816 r
.60334 .22581 .61073 .24092 .62503 .23512 .61765 .21975 Metetra
.641 .625 .819 r
.61765 .21975 .62503 .23512 .63942 .23 .63206 .21457 Metetra
.622 .615 .823 r
.63206 .21457 .63942 .23 .65391 .22557 .64658 .21028 Metetra
.602 .608 .83 r
.64658 .21028 .65391 .22557 .66852 .22183 .66123 .20687 Metetra
.582 .604 .84 r
.66123 .20687 .66852 .22183 .68325 .21872 .67601 .20428 Metetra
.562 .603 .852 r
.67601 .20428 .68325 .21872 .69811 .21619 .69094 .20244 Metetra
.542 .606 .866 r
.69094 .20244 .69811 .21619 .71311 .21414 .70601 .20123 Metetra
.523 .612 .883 r
.70601 .20123 .71311 .21414 .72824 .21248 .72124 .20052 Metetra
.505 .624 .902 r
.72124 .20052 .72824 .21248 .74351 .21108 .73662 .20015 Metetra
.488 .64 .923 r
.73662 .20015 .74351 .21108 .75891 .20981 .75215 .19996 Metetra
.786 .769 .838 r
.08389 .42449 .0928 .43226 .10537 .42314 .09664 .41431 Metetra
.775 .748 .828 r
.09664 .41431 .10537 .42314 .11795 .41416 .10938 .40434 Metetra
.761 .728 .822 r
.10938 .40434 .11795 .41416 .1305 .40551 .12208 .39475 Metetra
.746 .709 .817 r
.12208 .39475 .1305 .40551 .14302 .39733 .13473 .38573 Metetra
.728 .691 .814 r
.13473 .38573 .14302 .39733 .15549 .38974 .14732 .37745 Metetra
.709 .674 .814 r
.14732 .37745 .15549 .38974 .16792 .38287 .15983 .37002 Metetra
.688 .659 .816 r
.15983 .37002 .16792 .38287 .1803 .37679 .17227 .36354 Metetra
.666 .645 .819 r
.17227 .36354 .1803 .37679 .19264 .37155 .18464 .35809 Metetra
.642 .632 .825 r
.18464 .35809 .19264 .37155 .20494 .36719 .19695 .35369 Metetra
.615 .621 .833 r
.19695 .35369 .20494 .36719 .21721 .36369 .20922 .35034 Metetra
.588 .611 .843 r
.20922 .35034 .21721 .36369 .22947 .36102 .22146 .34797 Metetra
.558 .604 .855 r
.22146 .34797 .22947 .36102 .24173 .35911 .23368 .34652 Metetra
.528 .599 .869 r
.23368 .34652 .24173 .35911 .25402 .35785 .24592 .34588 Metetra
.498 .598 .885 r
.24592 .34588 .25402 .35785 .26634 .35714 .25819 .34589 Metetra
.468 .601 .901 r
.25819 .34589 .26634 .35714 .27872 .35683 .27052 .34639 Metetra
.442 .609 .919 r
.27052 .34639 .27872 .35683 .29118 .35676 .28292 .34718 Metetra
.42 .622 .936 r
.28292 .34718 .29118 .35676 .30373 .35677 .29542 .34808 Metetra
.407 .642 .952 r
.29542 .34808 .30373 .35677 .31638 .35668 .30803 .34886 Metetra
.404 .668 .967 r
.30803 .34886 .31638 .35668 .32915 .35631 .32078 .34933 Metetra
.415 .7 .98 r
.32078 .34933 .32915 .35631 .34203 .35552 .33365 .34928 Metetra
.439 .735 .99 r
.33365 .34928 .34203 .35552 .35505 .35414 .34667 .34853 Metetra
.477 .773 .996 r
.34667 .34853 .35505 .35414 .36819 .35206 .35983 .34694 Metetra
.524 .809 .998 r
.35983 .34694 .36819 .35206 .38145 .34916 .37312 .34437 Metetra
.576 .839 .995 r
.37312 .34437 .38145 .34916 .39483 .34538 .38655 .34073 Metetra
.628 .861 .986 r
.38655 .34073 .39483 .34538 .40832 .34068 .4001 .33599 Metetra
.676 .874 .973 r
.4001 .33599 .40832 .34068 .42191 .33505 .41376 .33013 Metetra
.716 .877 .957 r
.41376 .33013 .42191 .33505 .43558 .32853 .4275 .32321 Metetra
.746 .872 .938 r
.4275 .32321 .43558 .32853 .44932 .32118 .44133 .31529 Metetra
.768 .86 .918 r
.44133 .31529 .44932 .32118 .46313 .31311 .45521 .30649 Metetra
.781 .844 .899 r
.45521 .30649 .46313 .31311 .47698 .30443 .46915 .29698 Metetra
.786 .824 .882 r
.46915 .29698 .47698 .30443 .49088 .2953 .48312 .28693 Metetra
.786 .803 .866 r
.48312 .28693 .49088 .2953 .50481 .28588 .49712 .27653 Metetra
.781 .781 .853 r
.49712 .27653 .50481 .28588 .51877 .27633 .51114 .26599 Metetra
.771 .759 .842 r
.51114 .26599 .51877 .27633 .53276 .26685 .52518 .25553 Metetra
.759 .738 .833 r
.52518 .25553 .53276 .26685 .54678 .25759 .53925 .24536 Metetra
.744 .717 .826 r
.53925 .24536 .54678 .25759 .56084 .24872 .55334 .23566 Metetra
.727 .698 .822 r
.55334 .23566 .56084 .24872 .57495 .24039 .56748 .22661 Metetra
.709 .679 .82 r
.56748 .22661 .57495 .24039 .58911 .23272 .58166 .21836 Metetra
.688 .662 .819 r
.58166 .21836 .58911 .23272 .60334 .22581 .59591 .21104 Metetra
.666 .647 .821 r
.59591 .21104 .60334 .22581 .61765 .21975 .61025 .20474 Metetra
.643 .633 .825 r
.61025 .20474 .61765 .21975 .63206 .21457 .62468 .1995 Metetra
.619 .621 .83 r
.62468 .1995 .63206 .21457 .64658 .21028 .63923 .19533 Metetra
.594 .611 .838 r
.63923 .19533 .64658 .21028 .66123 .20687 .65391 .19223 Metetra
.568 .603 .848 r
.65391 .19223 .66123 .20687 .67601 .20428 .66874 .19012 Metetra
.542 .599 .86 r
.66874 .19012 .67601 .20428 .69094 .20244 .68372 .1889 Metetra
.516 .597 .874 r
.68372 .1889 .69094 .20244 .70601 .20123 .69887 .18846 Metetra
.491 .6 .889 r
.69887 .18846 .70601 .20123 .72124 .20052 .71419 .18862 Metetra
.466 .607 .906 r
.71419 .18862 .72124 .20052 .73662 .20015 .72968 .18921 Metetra
.445 .618 .924 r
.72968 .18921 .73662 .20015 .75215 .19996 .74532 .19002 Metetra
.797 .769 .827 r
.0749 .41664 .08389 .42449 .09664 .41431 .08782 .40552 Metetra
.787 .75 .82 r
.08782 .40552 .09664 .41431 .10938 .40434 .10072 .39463 Metetra
.773 .733 .815 r
.10072 .39463 .10938 .40434 .12208 .39475 .11357 .38419 Metetra
.758 .716 .813 r
.11357 .38419 .12208 .39475 .13473 .38573 .12635 .37442 Metetra
.74 .7 .813 r
.12635 .37442 .13473 .38573 .14732 .37745 .13904 .36549 Metetra
.72 .684 .815 r
.13904 .36549 .14732 .37745 .15983 .37002 .15163 .35756 Metetra
.697 .669 .819 r
.15163 .35756 .15983 .37002 .17227 .36354 .16413 .35073 Metetra
.672 .655 .825 r
.16413 .35073 .17227 .36354 .18464 .35809 .17654 .34508 Metetra
.643 .641 .833 r
.17654 .34508 .18464 .35809 .19695 .35369 .18887 .34064 Metetra
.612 .628 .843 r
.18887 .34064 .19695 .35369 .20922 .35034 .20113 .3374 Metetra
.578 .616 .854 r
.20113 .3374 .20922 .35034 .22146 .34797 .21335 .33531 Metetra
.541 .605 .866 r
.21335 .33531 .22146 .34797 .23368 .34652 .22554 .33428 Metetra
.502 .596 .88 r
.22554 .33428 .23368 .34652 .24592 .34588 .23773 .33418 Metetra
.463 .589 .894 r
.23773 .33418 .24592 .34588 .25819 .34589 .24995 .33484 Metetra
.425 .586 .908 r
.24995 .33484 .25819 .34589 .27052 .34639 .26223 .33606 Metetra
.391 .589 .922 r
.26223 .33606 .27052 .34639 .28292 .34718 .27458 .33764 Metetra
.366 .598 .936 r
.27458 .33764 .28292 .34718 .29542 .34808 .28704 .33934 Metetra
.351 .614 .949 r
.28704 .33934 .29542 .34808 .30803 .34886 .29962 .34091 Metetra
.352 .639 .962 r
.29962 .34091 .30803 .34886 .32078 .34933 .31233 .34212 Metetra
.37 .672 .975 r
.31233 .34212 .32078 .34933 .33365 .34928 .32519 .34275 Metetra
.405 .712 .986 r
.32519 .34275 .33365 .34928 .34667 .34853 .33821 .34257 Metetra
.454 .754 .993 r
.33821 .34257 .34667 .34853 .35983 .34694 .35139 .34142 Metetra
.513 .794 .996 r
.35139 .34142 .35983 .34694 .37312 .34437 .36472 .33914 Metetra
.575 .827 .993 r
.36472 .33914 .37312 .34437 .38655 .34073 .37819 .33563 Metetra
.634 .852 .983 r
.37819 .33563 .38655 .34073 .4001 .33599 .3918 .33085 Metetra
.686 .865 .967 r
.3918 .33085 .4001 .33599 .41376 .33013 .40552 .32478 Metetra
.728 .869 .947 r
.40552 .32478 .41376 .33013 .4275 .32321 .41934 .31748 Metetra
.759 .864 .926 r
.41934 .31748 .4275 .32321 .44133 .31529 .43325 .30903 Metetra
.78 .853 .905 r
.43325 .30903 .44133 .31529 .45521 .30649 .44722 .29958 Metetra
.793 .839 .886 r
.44722 .29958 .45521 .30649 .46915 .29698 .46123 .2893 Metetra
.799 .821 .869 r
.46123 .2893 .46915 .29698 .48312 .28693 .47528 .2784 Metetra
.799 .803 .855 r
.47528 .2784 .48312 .28693 .49712 .27653 .48935 .26711 Metetra
.794 .784 .843 r
.48935 .26711 .49712 .27653 .51114 .26599 .50344 .25568 Metetra
.785 .765 .834 r
.50344 .25568 .51114 .26599 .52518 .25553 .51754 .24433 Metetra
.773 .745 .828 r
.51754 .24433 .52518 .25553 .53925 .24536 .53165 .23332 Metetra
.758 .727 .824 r
.53165 .23332 .53925 .24536 .55334 .23566 .54579 .22287 Metetra
.74 .708 .822 r
.54579 .22287 .55334 .23566 .56748 .22661 .55996 .21317 Metetra
.72 .691 .822 r
.55996 .21317 .56748 .22661 .58166 .21836 .57417 .2044 Metetra
.697 .674 .823 r
.57417 .2044 .58166 .21836 .59591 .21104 .58845 .1967 Metetra
.672 .657 .827 r
.58845 .1967 .59591 .21104 .61025 .20474 .60281 .19017 Metetra
.645 .642 .832 r
.60281 .19017 .61025 .20474 .62468 .1995 .61727 .18487 Metetra
.617 .628 .839 r
.61727 .18487 .62468 .1995 .63923 .19533 .63185 .18081 Metetra
.586 .615 .848 r
.63185 .18081 .63923 .19533 .65391 .19223 .64657 .17797 Metetra
.555 .604 .858 r
.64657 .17797 .65391 .19223 .66874 .19012 .66144 .17628 Metetra
.522 .596 .869 r
.66144 .17628 .66874 .19012 .68372 .1889 .67648 .17563 Metetra
.489 .59 .881 r
.67648 .17563 .68372 .1889 .69887 .18846 .6917 .17588 Metetra
.458 .588 .895 r
.6917 .17588 .69887 .18846 .71419 .18862 .7071 .17683 Metetra
.429 .59 .909 r
.7071 .17683 .71419 .18862 .72968 .18921 .72267 .17827 Metetra
.403 .597 .924 r
.72267 .17827 .72968 .18921 .74532 .19002 .73843 .17999 Metetra
.806 .768 .818 r
.06583 .40872 .0749 .41664 .08782 .40552 .07892 .39676 Metetra
.796 .753 .813 r
.07892 .39676 .08782 .40552 .10072 .39463 .09197 .38506 Metetra
.784 .738 .81 r
.09197 .38506 .10072 .39463 .11357 .38419 .10496 .37387 Metetra
.769 .724 .81 r
.10496 .37387 .11357 .38419 .12635 .37442 .11786 .36344 Metetra
.751 .709 .813 r
.11786 .36344 .12635 .37442 .13904 .36549 .13065 .35394 Metetra
.73 .695 .817 r
.13065 .35394 .13904 .36549 .15163 .35756 .14332 .34556 Metetra
.706 .681 .824 r
.14332 .34556 .15163 .35756 .16413 .35073 .15588 .33841 Metetra
.678 .667 .832 r
.15588 .33841 .16413 .35073 .17654 .34508 .16833 .33258 Metetra
.645 .652 .842 r
.16833 .33258 .17654 .34508 .18887 .34064 .18067 .3281 Metetra
.609 .637 .854 r
.18067 .3281 .18887 .34064 .20113 .3374 .19294 .32497 Metetra
.568 .622 .866 r
.19294 .32497 .20113 .3374 .21335 .33531 .20513 .32312 Metetra
.523 .607 .879 r
.20513 .32312 .21335 .33531 .22554 .33428 .2173 .32245 Metetra
.475 .593 .891 r
.2173 .32245 .22554 .33428 .23773 .33418 .22945 .32281 Metetra
.427 .581 .903 r
.22945 .32281 .23773 .33418 .24995 .33484 .24163 .32403 Metetra
.38 .572 .913 r
.24163 .32403 .24995 .33484 .26223 .33606 .25386 .32589 Metetra
.341 .568 .923 r
.25386 .32589 .26223 .33606 .27458 .33764 .26616 .32815 Metetra
.312 .572 .932 r
.26616 .32815 .27458 .33764 .28704 .33934 .27858 .33054 Metetra
.298 .586 .942 r
.27858 .33054 .28704 .33934 .29962 .34091 .29112 .3328 Metetra
.304 .611 .954 r
.29112 .3328 .29962 .34091 .31233 .34212 .30381 .33467 Metetra
.33 .646 .967 r
.30381 .33467 .31233 .34212 .32519 .34275 .31666 .33588 Metetra
.375 .689 .98 r
.31666 .33588 .32519 .34275 .33821 .34257 .32968 .33621 Metetra
.436 .735 .99 r
.32968 .33621 .33821 .34257 .35139 .34142 .34288 .33544 Metetra
.505 .779 .994 r
.34288 .33544 .35139 .34142 .36472 .33914 .35624 .33341 Metetra
.575 .815 .989 r
.35624 .33341 .36472 .33914 .37819 .33563 .36975 .33001 Metetra
.64 .84 .977 r
.36975 .33001 .37819 .33563 .3918 .33085 .38342 .32519 Metetra
.694 .854 .959 r
.38342 .32519 .3918 .33085 .40552 .32478 .39721 .31893 Metetra
.737 .858 .937 r
.39721 .31893 .40552 .32478 .41934 .31748 .4111 .31129 Metetra
.768 .854 .915 r
.4111 .31129 .41934 .31748 .43325 .30903 .42509 .30237 Metetra
.789 .845 .893 r
.42509 .30237 .43325 .30903 .44722 .29958 .43914 .29233 Metetra
.802 .833 .874 r
.43914 .29233 .44722 .29958 .46123 .2893 .45324 .28137 Metetra
.808 .818 .858 r
.45324 .28137 .46123 .2893 .47528 .2784 .46737 .26972 Metetra
.809 .802 .845 r
.46737 .26972 .47528 .2784 .48935 .26711 .48151 .25764 Metetra
.805 .786 .835 r
.48151 .25764 .48935 .26711 .50344 .25568 .49567 .2454 Metetra
.797 .77 .828 r
.49567 .2454 .50344 .25568 .51754 .24433 .50983 .23328 Metetra
.785 .753 .824 r
.50983 .23328 .51754 .24433 .53165 .23332 .524 .22153 Metetra
.77 .736 .822 r
.524 .22153 .53165 .23332 .54579 .22287 .53819 .2104 Metetra
.752 .72 .822 r
.53819 .2104 .54579 .22287 .55996 .21317 .5524 .20013 Metetra
.73 .703 .825 r
.5524 .20013 .55996 .21317 .57417 .2044 .56665 .19089 Metetra
.706 .686 .829 r
.56665 .19089 .57417 .2044 .58845 .1967 .58095 .18286 Metetra
.678 .67 .835 r
.58095 .18286 .58845 .1967 .60281 .19017 .59534 .17612 Metetra
.648 .653 .842 r
.59534 .17612 .60281 .19017 .61727 .18487 .60983 .17076 Metetra
.615 .637 .85 r
.60983 .17076 .61727 .18487 .63185 .18081 .62444 .16679 Metetra
.579 .622 .859 r
.62444 .16679 .63185 .18081 .64657 .17797 .6392 .16418 Metetra
.541 .607 .869 r
.6392 .16418 .64657 .17797 .66144 .17628 .65411 .16285 Metetra
.502 .594 .879 r
.65411 .16285 .66144 .17628 .67648 .17563 .6692 .16269 Metetra
.463 .584 .889 r
.6692 .16269 .67648 .17563 .6917 .17588 .68448 .16353 Metetra
.426 .576 .9 r
.68448 .16353 .6917 .17588 .7071 .17683 .69995 .16516 Metetra
.392 .574 .911 r
.69995 .16516 .7071 .17683 .72267 .17827 .71561 .16735 Metetra
.364 .577 .922 r
.71561 .16735 .72267 .17827 .73843 .17999 .73146 .16985 Metetra
.813 .768 .811 r
.05668 .40074 .06583 .40872 .07892 .39676 .06992 .38805 Metetra
.804 .755 .807 r
.06992 .38805 .07892 .39676 .09197 .38506 .08312 .37566 Metetra
.793 .743 .807 r
.08312 .37566 .09197 .38506 .10496 .37387 .09623 .36384 Metetra
.778 .731 .809 r
.09623 .36384 .10496 .37387 .11786 .36344 .10924 .35283 Metetra
.761 .719 .814 r
.10924 .35283 .11786 .36344 .13065 .35394 .12213 .34285 Metetra
.739 .707 .821 r
.12213 .34285 .13065 .35394 .14332 .34556 .13489 .33408 Metetra
.714 .694 .83 r
.13489 .33408 .14332 .34556 .15588 .33841 .1475 .32666 Metetra
.684 .68 .841 r
.1475 .32666 .15588 .33841 .16833 .33258 .15999 .32067 Metetra
.648 .665 .853 r
.15999 .32067 .16833 .33258 .18067 .3281 .17235 .31616 Metetra
.606 .649 .866 r
.17235 .31616 .18067 .3281 .19294 .32497 .18462 .3131 Metetra
.558 .631 .88 r
.18462 .3131 .19294 .32497 .20513 .32312 .19681 .31144 Metetra
.505 .612 .892 r
.19681 .31144 .20513 .32312 .2173 .32245 .20895 .31106 Metetra
.448 .592 .903 r
.20895 .31106 .2173 .32245 .22945 .32281 .22107 .31182 Metetra
.39 .573 .911 r
.22107 .31182 .22945 .32281 .24163 .32403 .23321 .31351 Metetra
.336 .557 .917 r
.23321 .31351 .24163 .32403 .25386 .32589 .24539 .3159 Metetra
.291 .548 .921 r
.24539 .3159 .25386 .32589 .26616 .32815 .25766 .31872 Metetra
.261 .548 .926 r
.25766 .31872 .26616 .32815 .27858 .33054 .27004 .3217 Metetra
.251 .56 .934 r
.27004 .3217 .27858 .33054 .29112 .3328 .28255 .32454 Metetra
.262 .585 .945 r
.28255 .32454 .29112 .3328 .30381 .33467 .29522 .32695 Metetra
.297 .622 .96 r
.29522 .32695 .30381 .33467 .31666 .33588 .30806 .32866 Metetra
.353 .668 .974 r
.30806 .32866 .31666 .33588 .32968 .33621 .32109 .32939 Metetra
.423 .717 .986 r
.32109 .32939 .32968 .33621 .34288 .33544 .3343 .32894 Metetra
.501 .763 .99 r
.3343 .32894 .34288 .33544 .35624 .33341 .34769 .32713 Metetra
.577 .801 .984 r
.34769 .32713 .35624 .33341 .36975 .33001 .36125 .32382 Metetra
.644 .826 .97 r
.36125 .32382 .36975 .33001 .38342 .32519 .37496 .31895 Metetra
.7 .841 .95 r
.37496 .31895 .38342 .32519 .39721 .31893 .38882 .31252 Metetra
.743 .845 .927 r
.38882 .31252 .39721 .31893 .4111 .31129 .40279 .30458 Metetra
.774 .843 .904 r
.40279 .30458 .4111 .31129 .42509 .30237 .41685 .29525 Metetra
.795 .836 .883 r
.41685 .29525 .42509 .30237 .43914 .29233 .43099 .2847 Metetra
.809 .826 .864 r
.43099 .2847 .43914 .29233 .45324 .28137 .44517 .27316 Metetra
.815 .814 .849 r
.44517 .27316 .45324 .28137 .46737 .26972 .45937 .26087 Metetra
.817 .801 .837 r
.45937 .26087 .46737 .26972 .48151 .25764 .4736 .24811 Metetra
.813 .788 .829 r
.4736 .24811 .48151 .25764 .49567 .2454 .48783 .23519 Metetra
.806 .774 .823 r
.48783 .23519 .49567 .2454 .50983 .23328 .50206 .22239 Metetra
.795 .76 .821 r
.50206 .22239 .50983 .23328 .524 .22153 .51629 .21001 Metetra
.781 .746 .821 r
.51629 .21001 .524 .22153 .53819 .2104 .53052 .19831 Metetra
.762 .731 .824 r
.53052 .19831 .53819 .2104 .5524 .20013 .54478 .18754 Metetra
.74 .716 .829 r
.54478 .18754 .5524 .20013 .56665 .19089 .55907 .17791 Metetra
.714 .7 .835 r
.55907 .17791 .56665 .19089 .58095 .18286 .57342 .16958 Metetra
.685 .684 .843 r
.57342 .16958 .58095 .18286 .59534 .17612 .58784 .16266 Metetra
.651 .666 .852 r
.58784 .16266 .59534 .17612 .60983 .17076 .60236 .15724 Metetra
.613 .649 .862 r
.60236 .15724 .60983 .17076 .62444 .16679 .617 .15334 Metetra
.571 .63 .871 r
.617 .15334 .62444 .16679 .6392 .16418 .63179 .15091 Metetra
.527 .612 .881 r
.63179 .15091 .6392 .16418 .65411 .16285 .64675 .14988 Metetra
.482 .595 .889 r
.64675 .14988 .65411 .16285 .6692 .16269 .66189 .15012 Metetra
.437 .579 .897 r
.66189 .15012 .6692 .16269 .68448 .16353 .67722 .15144 Metetra
.395 .567 .905 r
.67722 .15144 .68448 .16353 .69995 .16516 .69276 .15364 Metetra
.358 .56 .912 r
.69276 .15364 .69995 .16516 .71561 .16735 .70849 .15646 Metetra
.329 .559 .92 r
.70849 .15646 .71561 .16735 .73146 .16985 .72442 .15961 Metetra
.818 .767 .805 r
.04745 .39268 .05668 .40074 .06992 .38805 .06083 .37941 Metetra
.81 .758 .804 r
.06083 .37941 .06992 .38805 .08312 .37566 .07415 .36646 Metetra
.8 .748 .805 r
.07415 .36646 .08312 .37566 .09623 .36384 .08739 .35411 Metetra
.786 .739 .809 r
.08739 .35411 .09623 .36384 .10924 .35283 .1005 .34264 Metetra
.769 .729 .816 r
.1005 .34264 .10924 .35283 .12213 .34285 .11348 .33226 Metetra
.748 .719 .825 r
.11348 .33226 .12213 .34285 .13489 .33408 .12631 .32318 Metetra
.722 .708 .837 r
.12631 .32318 .13489 .33408 .1475 .32666 .13898 .31553 Metetra
.689 .695 .85 r
.13898 .31553 .1475 .32666 .15999 .32067 .1515 .30941 Metetra
.65 .68 .865 r
.1515 .30941 .15999 .32067 .17235 .31616 .16389 .30485 Metetra
.604 .663 .88 r
.16389 .30485 .17235 .31616 .18462 .3131 .17617 .30185 Metetra
.549 .642 .895 r
.17617 .30185 .18462 .3131 .19681 .31144 .18835 .30033 Metetra
.487 .618 .906 r
.18835 .30033 .19681 .31144 .20895 .31106 .20047 .30019 Metetra
.421 .592 .914 r
.20047 .30019 .20895 .31106 .22107 .31182 .21257 .30124 Metetra
.354 .567 .918 r
.21257 .30124 .22107 .31182 .23321 .31351 .22468 .3033 Metetra
.294 .544 .919 r
.22468 .3033 .23321 .31351 .24539 .3159 .23683 .3061 Metetra
.246 .53 .918 r
.23683 .3061 .24539 .3159 .25766 .31872 .24907 .30937 Metetra
.217 .526 .919 r
.24907 .30937 .25766 .31872 .27004 .3217 .26142 .31281 Metetra
.21 .537 .925 r
.26142 .31281 .27004 .3217 .28255 .32454 .27391 .3161 Metetra
.229 .563 .937 r
.27391 .3161 .28255 .32454 .29522 .32695 .28656 .31895 Metetra
.273 .602 .952 r
.28656 .31895 .29522 .32695 .30806 .32866 .2994 .32104 Metetra
.337 .65 .969 r
.2994 .32104 .30806 .32866 .32109 .32939 .31243 .3221 Metetra
.416 .701 .981 r
.31243 .3221 .32109 .32939 .3343 .32894 .32565 .3219 Metetra
.499 .748 .984 r
.32565 .3219 .3343 .32894 .34769 .32713 .33907 .32024 Metetra
.579 .786 .977 r
.33907 .32024 .34769 .32713 .36125 .32382 .35267 .31699 Metetra
.648 .812 .962 r
.35267 .31699 .36125 .32382 .37496 .31895 .36644 .31208 Metetra
.704 .826 .941 r
.36644 .31208 .37496 .31895 .38882 .31252 .38036 .30549 Metetra
.746 .832 .917 r
.38036 .30549 .38882 .31252 .40279 .30458 .3944 .29731 Metetra
.777 .831 .894 r
.3944 .29731 .40279 .30458 .41685 .29525 .40854 .28764 Metetra
.799 .826 .873 r
.40854 .28764 .41685 .29525 .43099 .2847 .42276 .27668 Metetra
.812 .819 .855 r
.42276 .27668 .43099 .2847 .44517 .27316 .43702 .26465 Metetra
.82 .81 .841 r
.43702 .26465 .44517 .27316 .45937 .26087 .45131 .25183 Metetra
.822 .8 .83 r
.45131 .25183 .45937 .26087 .4736 .24811 .46561 .23853 Metetra
.82 .789 .824 r
.46561 .23853 .4736 .24811 .48783 .23519 .47991 .22504 Metetra
.814 .779 .82 r
.47991 .22504 .48783 .23519 .50206 .22239 .49421 .2117 Metetra
.804 .767 .82 r
.49421 .2117 .50206 .22239 .51629 .21001 .50851 .1988 Metetra
.79 .756 .822 r
.50851 .1988 .51629 .21001 .53052 .19831 .5228 .18663 Metetra
.772 .743 .827 r
.5228 .18663 .53052 .19831 .54478 .18754 .53711 .17546 Metetra
.75 .73 .834 r
.53711 .17546 .54478 .18754 .55907 .17791 .55145 .16549 Metetra
.723 .715 .843 r
.55145 .16549 .55907 .17791 .57342 .16958 .56583 .15692 Metetra
.691 .699 .853 r
.56583 .15692 .57342 .16958 .58784 .16266 .58029 .14985 Metetra
.653 .681 .864 r
.58029 .14985 .58784 .16266 .60236 .15724 .59485 .14437 Metetra
.611 .662 .875 r
.59485 .14437 .60236 .15724 .617 .15334 .60953 .1405 Metetra
.564 .641 .885 r
.60953 .1405 .617 .15334 .63179 .15091 .62436 .13821 Metetra
.514 .619 .894 r
.62436 .13821 .63179 .15091 .64675 .14988 .63935 .13741 Metetra
.463 .597 .9 r
.63935 .13741 .64675 .14988 .66189 .15012 .65454 .13795 Metetra
.412 .577 .905 r
.65454 .13795 .66189 .15012 .67722 .15144 .66992 .13966 Metetra
.366 .56 .909 r
.66992 .13966 .67722 .15144 .69276 .15364 .68551 .1423 Metetra
.327 .548 .913 r
.68551 .1423 .69276 .15364 .70849 .15646 .70131 .1456 Metetra
.299 .544 .917 r
.70131 .1456 .70849 .15646 .72442 .15961 .71732 .14927 Metetra
.821 .767 .801 r
.03814 .38456 .04745 .39268 .06083 .37941 .05164 .37084 Metetra
.815 .76 .801 r
.05164 .37084 .06083 .37941 .07415 .36646 .06507 .35746 Metetra
.806 .753 .804 r
.06507 .35746 .07415 .36646 .08739 .35411 .07841 .34472 Metetra
.793 .747 .81 r
.07841 .34472 .08739 .35411 .1005 .34264 .09162 .33289 Metetra
.777 .74 .819 r
.09162 .33289 .1005 .34264 .11348 .33226 .10468 .32222 Metetra
.756 .732 .83 r
.10468 .32222 .11348 .33226 .12631 .32318 .11758 .31289 Metetra
.729 .723 .845 r
.11758 .31289 .12631 .32318 .13898 .31553 .1303 .30506 Metetra
.695 .711 .861 r
.1303 .30506 .13898 .31553 .1515 .30941 .14286 .29883 Metetra
.653 .697 .878 r
.14286 .29883 .1515 .30941 .16389 .30485 .15528 .29423 Metetra
.601 .679 .896 r
.15528 .29423 .16389 .30485 .17617 .30185 .16757 .29126 Metetra
.539 .655 .91 r
.16757 .29126 .17617 .30185 .18835 .30033 .17975 .28984 Metetra
.469 .627 .921 r
.17975 .28984 .18835 .30033 .20047 .30019 .19187 .28985 Metetra
.394 .595 .926 r
.19187 .28985 .20047 .30019 .21257 .30124 .20395 .29112 Metetra
.32 .563 .924 r
.20395 .29112 .21257 .30124 .22468 .3033 .21604 .29343 Metetra
.255 .534 .919 r
.21604 .29343 .22468 .3033 .23683 .3061 .22817 .29652 Metetra
.206 .515 .914 r
.22817 .29652 .23683 .3061 .24907 .30937 .24039 .30011 Metetra
.18 .509 .912 r
.24039 .30011 .24907 .30937 .26142 .31281 .25271 .30387 Metetra
.18 .519 .917 r
.25271 .30387 .26142 .31281 .27391 .3161 .26519 .30749 Metetra
.206 .546 .929 r
.26519 .30749 .27391 .3161 .28656 .31895 .27783 .31064 Metetra
.258 .587 .946 r
.27783 .31064 .28656 .31895 .2994 .32104 .29067 .313 Metetra
.33 .636 .963 r
.29067 .313 .2994 .32104 .31243 .3221 .3037 .31429 Metetra
.414 .688 .975 r
.3037 .31429 .31243 .3221 .32565 .3219 .31695 .31426 Metetra
.501 .734 .977 r
.31695 .31426 .32565 .3219 .33907 .32024 .3304 .3127 Metetra
.582 .771 .969 r
.3304 .3127 .33907 .32024 .35267 .31699 .34404 .30949 Metetra
.651 .796 .952 r
.34404 .30949 .35267 .31699 .36644 .31208 .35786 .30453 Metetra
.706 .811 .93 r
.35786 .30453 .36644 .31208 .38036 .30549 .37184 .29782 Metetra
.748 .817 .907 r
.37184 .29782 .38036 .30549 .3944 .29731 .38595 .28944 Metetra
.778 .818 .884 r
.38595 .28944 .3944 .29731 .40854 .28764 .40016 .27951 Metetra
.8 .816 .864 r
.40016 .27951 .40854 .28764 .42276 .27668 .41445 .26823 Metetra
.814 .811 .848 r
.41445 .26823 .42276 .27668 .43702 .26465 .42879 .25583 Metetra
.822 .805 .835 r
.42879 .25583 .43702 .26465 .45131 .25183 .44316 .24261 Metetra
.826 .798 .825 r
.44316 .24261 .45131 .25183 .46561 .23853 .45754 .22888 Metetra
.825 .791 .82 r
.45754 .22888 .46561 .23853 .47991 .22504 .47192 .21497 Metetra
.82 .783 .818 r
.47192 .21497 .47991 .22504 .49421 .2117 .48629 .20121 Metetra
.811 .775 .819 r
.48629 .20121 .49421 .2117 .50851 .1988 .50066 .18792 Metetra
.798 .766 .824 r
.50066 .18792 .50851 .1988 .5228 .18663 .51501 .1754 Metetra
.78 .756 .831 r
.51501 .1754 .5228 .18663 .53711 .17546 .52938 .16391 Metetra
.758 .745 .84 r
.52938 .16391 .53711 .17546 .55145 .16549 .54377 .15369 Metetra
.73 .732 .852 r
.54377 .15369 .55145 .16549 .56583 .15692 .5582 .14492 Metetra
.696 .716 .864 r
.5582 .14492 .56583 .15692 .58029 .14985 .5727 .13773 Metetra
.656 .698 .877 r
.5727 .13773 .58029 .14985 .59485 .14437 .5873 .13219 Metetra
.61 .677 .889 r
.5873 .13219 .59485 .14437 .60953 .1405 .60202 .12834 Metetra
.558 .654 .9 r
.60202 .12834 .60953 .1405 .62436 .13821 .61689 .12613 Metetra
.502 .628 .907 r
.61689 .12613 .62436 .13821 .63935 .13741 .63192 .12547 Metetra
.445 .602 .912 r
.63192 .12547 .63935 .13741 .65454 .13795 .64715 .12623 Metetra
.39 .576 .913 r
.64715 .12623 .65454 .13795 .66992 .13966 .66258 .12821 Metetra
.341 .555 .914 r
.66258 .12821 .66992 .13966 .68551 .1423 .67822 .13115 Metetra
.302 .54 .913 r
.67822 .13115 .68551 .1423 .70131 .1456 .69407 .1348 Metetra
.275 .533 .915 r
.69407 .1348 .70131 .1456 .71732 .14927 .71014 .13882 Metetra
.823 .766 .799 r
.02874 .37635 .03814 .38456 .05164 .37084 .04234 .36235 Metetra
.818 .763 .8 r
.04234 .36235 .05164 .37084 .06507 .35746 .05587 .34869 Metetra
.811 .759 .805 r
.05587 .34869 .06507 .35746 .07841 .34472 .0693 .33569 Metetra
.799 .755 .812 r
.0693 .33569 .07841 .34472 .09162 .33289 .08259 .32362 Metetra
.784 .751 .823 r
.08259 .32362 .09162 .33289 .10468 .32222 .09572 .31274 Metetra
.763 .746 .837 r
.09572 .31274 .10468 .32222 .11758 .31289 .10868 .30325 Metetra
.736 .739 .854 r
.10868 .30325 .11758 .31289 .1303 .30506 .12146 .29529 Metetra
.7 .729 .873 r
.12146 .29529 .1303 .30506 .14286 .29883 .13406 .28897 Metetra
.655 .716 .893 r
.13406 .28897 .14286 .29883 .15528 .29423 .1465 .28434 Metetra
.599 .697 .912 r
.1465 .28434 .15528 .29423 .16757 .29126 .15881 .28136 Metetra
.53 .671 .927 r
.15881 .28136 .16757 .29126 .17975 .28984 .17101 .27998 Metetra
.452 .638 .936 r
.17101 .27998 .17975 .28984 .19187 .28985 .18313 .28007 Metetra
.37 .6 .936 r
.18313 .28007 .19187 .28985 .20395 .29112 .19521 .28146 Metetra
.29 .562 .93 r
.19521 .28146 .20395 .29112 .21604 .29343 .20729 .28391 Metetra
.223 .528 .92 r
.20729 .28391 .21604 .29343 .22817 .29652 .21941 .28717 Metetra
.175 .505 .911 r
.21941 .28717 .22817 .29652 .24039 .30011 .23161 .29094 Metetra
.154 .498 .907 r
.23161 .29094 .24039 .30011 .25271 .30387 .24393 .29489 Metetra
.16 .508 .912 r
.24393 .29489 .25271 .30387 .26519 .30749 .25639 .29869 Metetra
.194 .535 .925 r
.25639 .29869 .26519 .30749 .27783 .31064 .26904 .30202 Metetra
.253 .577 .942 r
.26904 .30202 .27783 .31064 .29067 .313 .28188 .30453 Metetra
.33 .626 .959 r
.28188 .30453 .29067 .313 .3037 .31429 .29493 .30595 Metetra
.418 .676 .969 r
.29493 .30595 .3037 .31429 .31695 .31426 .30819 .306 Metetra
.505 .72 .969 r
.30819 .306 .31695 .31426 .3304 .3127 .32167 .30449 Metetra
.585 .755 .959 r
.32167 .30449 .3304 .3127 .34404 .30949 .33536 .30128 Metetra
.652 .779 .942 r
.33536 .30128 .34404 .30949 .35786 .30453 .34923 .29627 Metetra
.706 .794 .92 r
.34923 .29627 .35786 .30453 .37184 .29782 .36326 .28947 Metetra
.747 .802 .897 r
.36326 .28947 .37184 .29782 .38595 .28944 .37743 .28095 Metetra
.777 .805 .876 r
.37743 .28095 .38595 .28944 .40016 .27951 .39171 .27084 Metetra
.799 .805 .857 r
.39171 .27084 .40016 .27951 .41445 .26823 .40608 .25933 Metetra
.813 .803 .842 r
.40608 .25933 .41445 .26823 .42879 .25583 .42049 .24669 Metetra
.823 .8 .83 r
.42049 .24669 .42879 .25583 .44316 .24261 .43494 .2332 Metetra
.827 .796 .822 r
.43494 .2332 .44316 .24261 .45754 .22888 .4494 .21919 Metetra
.827 .792 .818 r
.4494 .21919 .45754 .22888 .47192 .21497 .46386 .20499 Metetra
.824 .787 .817 r
.46386 .20499 .47192 .21497 .48629 .20121 .4783 .19095 Metetra
.816 .782 .82 r
.4783 .19095 .48629 .20121 .50066 .18792 .49273 .1774 Metetra
.804 .776 .827 r
.49273 .1774 .50066 .18792 .51501 .1754 .50715 .16463 Metetra
.787 .769 .836 r
.50715 .16463 .51501 .1754 .52938 .16391 .52158 .15293 Metetra
.765 .76 .848 r
.52158 .15293 .52938 .16391 .54377 .15369 .53602 .14253 Metetra
.737 .749 .862 r
.53602 .14253 .54377 .15369 .5582 .14492 .55051 .13362 Metetra
.701 .735 .877 r
.55051 .13362 .5582 .14492 .5727 .13773 .56506 .12633 Metetra
.659 .717 .891 r
.56506 .12633 .5727 .13773 .5873 .13219 .57971 .12074 Metetra
.608 .695 .905 r
.57971 .12074 .5873 .13219 .60202 .12834 .59447 .11687 Metetra
.552 .669 .915 r
.59447 .11687 .60202 .12834 .61689 .12613 .60937 .11469 Metetra
.491 .64 .921 r
.60937 .11469 .61689 .12613 .63192 .12547 .62445 .11411 Metetra
.429 .609 .923 r
.62445 .11411 .63192 .12547 .64715 .12623 .63971 .11498 Metetra
.371 .579 .922 r
.63971 .11498 .64715 .12623 .66258 .12821 .65518 .11709 Metetra
.321 .554 .918 r
.65518 .11709 .66258 .12821 .67822 .13115 .67087 .12021 Metetra
.282 .535 .915 r
.67087 .12021 .67822 .13115 .69407 .1348 .68677 .12404 Metetra
.258 .526 .914 r
.68677 .12404 .69407 .1348 .71014 .13882 .70289 .12827 Metetra
.824 .766 .798 r
.01926 .36808 .02874 .37635 .04234 .36235 .03294 .35394 Metetra
.82 .765 .801 r
.03294 .35394 .04234 .36235 .05587 .34869 .04655 .34015 Metetra
.814 .764 .807 r
.04655 .34015 .05587 .34869 .0693 .33569 .06005 .32702 Metetra
.804 .764 .816 r
.06005 .32702 .0693 .33569 .08259 .32362 .07341 .31484 Metetra
.789 .762 .829 r
.07341 .31484 .08259 .32362 .09572 .31274 .0866 .30385 Metetra
.769 .76 .845 r
.0866 .30385 .09572 .31274 .10868 .30325 .09961 .29427 Metetra
.741 .756 .864 r
.09961 .29427 .10868 .30325 .12146 .29529 .11243 .28624 Metetra
.705 .748 .885 r
.11243 .28624 .12146 .29529 .13406 .28897 .12508 .27986 Metetra
.657 .736 .908 r
.12508 .27986 .13406 .28897 .1465 .28434 .13755 .27518 Metetra
.596 .716 .928 r
.13755 .27518 .1465 .28434 .15881 .28136 .14988 .27218 Metetra
.522 .688 .943 r
.14988 .27218 .15881 .28136 .17101 .27998 .1621 .27079 Metetra
.437 .651 .95 r
.1621 .27079 .17101 .27998 .18313 .28007 .17423 .27089 Metetra
.349 .608 .946 r
.17423 .27089 .18313 .28007 .19521 .28146 .18632 .27229 Metetra
.266 .564 .935 r
.18632 .27229 .19521 .28146 .20729 .28391 .19841 .27477 Metetra
.199 .527 .921 r
.19841 .27477 .20729 .28391 .21941 .28717 .21053 .27806 Metetra
.155 .502 .909 r
.21053 .27806 .21941 .28717 .23161 .29094 .22273 .28187 Metetra
.139 .494 .905 r
.22273 .28187 .23161 .29094 .24393 .29489 .23505 .28587 Metetra
.152 .504 .91 r
.23505 .28587 .24393 .29489 .25639 .29869 .24753 .28971 Metetra
.193 .532 .923 r
.24753 .28971 .25639 .29869 .26904 .30202 .26018 .29307 Metetra
.257 .572 .94 r
.26018 .29307 .26904 .30202 .28188 .30453 .27303 .29561 Metetra
.338 .619 .954 r
.27303 .29561 .28188 .30453 .29493 .30595 .2861 .29705 Metetra
.425 .666 .962 r
.2861 .29705 .29493 .30595 .30819 .306 .29939 .29711 Metetra
.511 .708 .96 r
.29939 .29711 .30819 .306 .32167 .30449 .3129 .29559 Metetra
.588 .74 .948 r
.3129 .29559 .32167 .30449 .33536 .30128 .32662 .29234 Metetra
.653 .763 .93 r
.32662 .29234 .33536 .30128 .34923 .29627 .34054 .28729 Metetra
.704 .778 .909 r
.34054 .28729 .34923 .29627 .36326 .28947 .35462 .28042 Metetra
.744 .787 .888 r
.35462 .28042 .36326 .28947 .37743 .28095 .36885 .27181 Metetra
.774 .792 .868 r
.36885 .27181 .37743 .28095 .39171 .27084 .3832 .2616 Metetra
.796 .794 .851 r
.3832 .2616 .39171 .27084 .40608 .25933 .39763 .24998 Metetra
.811 .795 .837 r
.39763 .24998 .40608 .25933 .42049 .24669 .41212 .23721 Metetra
.821 .795 .827 r
.41212 .23721 .42049 .24669 .43494 .2332 .42664 .22359 Metetra
.827 .794 .82 r
.42664 .22359 .43494 .2332 .4494 .21919 .44118 .20944 Metetra
.829 .792 .818 r
.44118 .20944 .4494 .21919 .46386 .20499 .45571 .1951 Metetra
.826 .791 .818 r
.45571 .1951 .46386 .20499 .4783 .19095 .47022 .18092 Metetra
.82 .789 .823 r
.47022 .18092 .4783 .19095 .49273 .1774 .48472 .16724 Metetra
.809 .786 .831 r
.48472 .16724 .49273 .1774 .50715 .16463 .49922 .15435 Metetra
.793 .782 .842 r
.49922 .15435 .50715 .16463 .52158 .15293 .51371 .14254 Metetra
.771 .775 .856 r
.51371 .14254 .52158 .15293 .53602 .14253 .52821 .13204 Metetra
.743 .767 .872 r
.52821 .13204 .53602 .14253 .55051 .13362 .54276 .12303 Metetra
.706 .754 .89 r
.54276 .12303 .55051 .13362 .56506 .12633 .55736 .11567 Metetra
.661 .737 .906 r
.55736 .11567 .56506 .12633 .57971 .12074 .57205 .11002 Metetra
.607 .714 .921 r
.57205 .11002 .57971 .12074 .59447 .11687 .58686 .10612 Metetra
.546 .686 .931 r
.58686 .10612 .59447 .11687 .60937 .11469 .60181 .10392 Metetra
.481 .653 .936 r
.60181 .10392 .60937 .11469 .62445 .11411 .61693 .10333 Metetra
.416 .619 .935 r
.61693 .10333 .62445 .11411 .63971 .11498 .63223 .1042 Metetra
.356 .585 .93 r
.63223 .1042 .63971 .11498 .65518 .11709 .64774 .10633 Metetra
.305 .556 .923 r
.64774 .10633 .65518 .11709 .67087 .12021 .66346 .10948 Metetra
.269 .535 .917 r
.66346 .10948 .67087 .12021 .68677 .12404 .67941 .11334 Metetra
.249 .524 .914 r
.67941 .11334 .68677 .12404 .70289 .12827 .69556 .11761 Metetra
0 g
.25 Mabswid
.68874 0 m
.96935 .42924 L
s
.96935 .42924 m
1 .6535 L
s
1 .6535 m
.70298 .24544 L
s
.70298 .24544 m
.68874 0 L
s
.03716 .25514 m
0 .48963 L
s
0 .48963 m
.70298 .24544 L
s
.70298 .24544 m
.68874 0 L
s
.68874 0 m
.03716 .25514 L
s
.03716 .25514 m
.68874 0 L
s
.03716 .25514 m
.04196 .25962 L
s
gsave
.02757 .24618 -71 -19.1694 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.22768 .18054 m
.23217 .18532 L
s
gsave
.2187 .17096 -70.69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.43326 .10004 m
.43737 .10515 L
s
gsave
.42503 .08982 -70.0263 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.65575 .01292 m
.65941 .01836 L
s
gsave
.64842 .00203 -69.3627 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.07414 .24066 m
.07698 .24338 L
s
.11167 .22596 m
.11448 .22872 L
s
.14976 .21105 m
.15253 .21384 L
s
.18843 .19591 m
.19116 .19874 L
s
.26753 .16493 m
.27019 .16784 L
s
.30801 .14908 m
.31061 .15203 L
s
.34911 .13299 m
.35167 .13598 L
s
.39085 .11664 m
.39337 .11967 L
s
.47634 .08317 m
.47875 .08628 L
s
.52011 .06603 m
.52247 .06918 L
s
.56459 .04861 m
.5669 .0518 L
s
.60979 .03091 m
.61205 .03414 L
s
gsave
.29165 .07573 -69.7292 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(q) show
68.375 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 233.563},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001GR000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooolY0?ooo`030?oo_`1d0000000000800000
00<0M;oo0?ooo`3oool0P03oool00?l0oooo:@3oool00`3ogi`0B00004RLg`020?ooo`030;md0018
W=l0oooo0800oooo003o0?ooobh0oooo00<0oono07@0B02Lgol0P03oool00?l0oooo;P3oool00`3o
okl0M00007Boo`200?ooo`00o`3oool^0?ooo`030?noM0000000M;oo01T0oooo00D0omnL04P00000
0000001d0;ooo`1R0?ooo`00o`3oool]0?ooo`040?oOW0180000001809cOoaP0oooo00L0oono07@0
001d_ol0oooo0?noM00007@0_ooo0640oooo003o0?ooob/0oooo00D0okmd000000000000001809cO
o`0I0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool0H03oool00?l0oooo;@3oool0
0`2oM00007Bo0?ooo`0H0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0H03oool0
0?l0oooo;P3oool00`2oM000B9cO0?ooo`0G0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol0H03oool00?l0oooo:@3oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo01L0
oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`1P0?ooo`00o`3ooolZ0?ooo`060?no
M018W=l0oooo0?oo_`1d000007Bo603oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
0600oooo003o0?ooob/0oooo00D0gia8000000000000001d0;ooo`0H0?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol0H03oool00?l0ooooB03oool00`3ookl0M00004RLg`030?ooo`03
0=nLB00007@0_ooo0600oooo003o0?ooodT0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo
0?ooo`1P0?ooo`00o`3ooom90?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`1Q0?ooo`00
o`3ooom:0?ooo`050?oo_`1d000000000000M02oool0HP3oool00?l0oooo?P3oool300000700oooo
003o0?ooocd0oooo0P0000000`3oool000000000001_0?ooo`00o`3ooolj0?ooo`<000000`3oool2
000006l0oooo003o0?ooocP0oooo0P0000060?ooo`030000003oool0000006h0oooo003o0?ooocD0
oooo0`0000080?ooo`040000003oool0oooo000006d0oooo003o0?oooc80oooo0`0000000`3oool0
00000?ooo`080?ooo`040000003oool0oooo000006d0oooo003o0?oooc00oooo0P0000050?ooo`03
0000003oool0oooo00H0oooo00D000000?ooo`3oool0oooo0000001/0?ooo`00o`3oool]0?ooo`<0
0000403oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo06T0oooo003o0?ooob/0oooo
0P00000C0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0J@3oool00:l0oooo00D0
oon?03D000000000000D05o4o`1c0?ooo`<000005P3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo06P0oooo002_0?ooo`0608le0000=Hl0oooo0?o4G`0D01@0GlCoL03oool2000001T0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`1W0?ooo`00[P3oool00`3ool@0Ga@0
03F?o`020?ooo`030<AO500055l0a?oo06`0oooo1000000J0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool0I`3oool00:h0oooo00<0oon?03D0000eSol00P3oool00`3oohl0=@0003F?
o`1Y0?ooo`<000007P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo06H0oooo002^
0?ooo`030?ooS`0e000055o400<0oooo00<0ScD0000eS`3oool0903oool00`2?=@00000000000004
0000000301AOa03oool0oooo03P0oooo0P00000Q0?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool0I@3oool00:h0oooo00<0ooo405lD0000=Hl00`3oool00`34Ga@0000e08ooo`0U0?oo
o`030<AO50000000000000<0000000<001AO0<Coo`3oool0=@3oool3000002<0oooo00<000000?oo
o`3oool0103oool3000006H0oooo002_0?ooo`0308le00000000000000@0000000<0=Hoo0?ooo`3o
ool0903oool00`34Ga@055o40?ooo`020?ooo`030<AO=@1Oa?l0oooo03<0oooo0P00000V0?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool0I03oool00:l0oooo00<0a5lD0000=@2?ool0
0P3oool00`3oohl0=@0001AOa00W0?ooo`030<AO500eSol0oooo03@0oooo0`00000X0?ooo`030000
003oool0oooo00L0oooo00<000000?ooo`3oool0H`3oool00:l0oooo00<0olAO01@0000eSol00`3o
ool00`2?=@0003F?0?ooo`0W0?ooo`0308le501Oa?l0oooo0340oooo0P0000000`3oool000000?oo
o`0X0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0H`3oool00;00oooo00<0ScD0
000DG`34ool00P3oool00`34Ga@003F?0?ooo`0W0?ooo`030?ooa01O53D0Sooo02h0oooo0`00000]
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0HP3oool00;00oooo00<0oon?03D0
000eSol00P3oool00`34Ga@003F?0?ooo`0X0?ooo`030?n?=@00=Hl0oooo02X0oooo0`00000`0?oo
o`030000003oool0oooo00T0oooo00<000000?ooo`3oool0H@3oool00;40oooo00H0ohle0000501O
a?l0oooo08le000eSolY0?ooo`030?ooS`0e01@0GlCo02P0oooo0P00000c0?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0H@3oool00;80oooo00D0olAO01@00000000001AO0<Coo`0Y
0?ooo`030?ooS`0e000055o402D0oooo0`00000e0?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool0H03oool00=/0oooo00<0ohmO05o4o`3oool00P3oool00`3oa5l0500001AOa00S0?oo
o`800000>03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo05l0oooo003K0?ooo`08
0?ooa01O51@0GlCo0?ooo`3oohl0=@000000=@2?oolP0?ooo`<0000000<0oooo0000003oool0=`3o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo05l0oooo003L0?ooo`060?ooa01O5000
00000000000003D0Sooo7`3oool2000003d0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`1N0?ooo`00oP3oool300000400oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`1M0?ooo`00n`3oool3000004<0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1M
0?ooo`00n@3oool2000004H0oooo00<000000?ooo`3oool03@3oool2000005d0oooo003f0?ooo`<0
0000B03oool00`000000oooo0?ooo`0<0?ooo`<00000G@3oool00?@0oooo0P0000020?ooo`030000
003oool0oooo04H0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1K0?ooo`00l@3o
ool3000004d0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1J0?ooo`00k`3oool2
00000500oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1I0?ooo`00k03oool30000
0580oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1I0?ooo`00j@3oool3000005D0
oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`1H0?ooo`00i`3oool3000005L0oooo
00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`1G0?ooo`00i03oool3000000<0oooo00<0
00000?ooo`3oool0<`3oool?00000180oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`1G0?ooo`00hP3oool2000003L0oooo1@00000308^_kP030000001kYnl0NjO_0080NjO_00<00000
06ZNk`1ZW^l00P1ZW^l00`000000FiG^00000005000000`0oooo00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`1F0?ooo`00g`3oool3000003D0oooo1000000309^fj`030000002;[nh0Rjo^
0080Rjo^00<0000007^Wk`1kYnl00`1kYnl01@000000JYk_06ZNk`1ZW^l0000000D0FiG^00<00000
04j>k0000000100000070?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0EP3oool0
0=d0oooo0P00000f0?ooo`80000000<0ZKcW0000002K]^/0102K]^/00`000000Rjo^08^_kP0308^_
kP040000001kYnl0NjO_07^Wk`D0000000<0JYk_0000001KUNh00`1KUNh00`000000CXk/04j>k002
04j>k00300000014R>X0000000<000000`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0
oooo05D0oooo003J0?ooo`<00000=@3oool3000000@0ZKcW00<0000009^fj`2K]^/00P2K]^/:0000
0080OJ?[00<0000006fKk01]Vn`02P0000000`1>S^`0000004B8jP0304B8jPD000005`3oool00`00
0000oooo0?ooo`1D0?ooo`00e`3oool3000000030?ooo`000000oooo03<0oooo0P0000020;C1h`03
0000002Y_>L0ZKcW0080ZKcW00<0000009^fj`2K]^/00P00000209^ai`030000002=ZnX0SJ_Z00<0
SJ_Z00<0000007fSj`1mXn/00P1mXn/00`000000KI_/06fKk00206fKk0030000001NU>`0GYC/0080
GYC/00<0000005:=j`0000000`00000204B8jP040000000oQ^T0?hKY03n6j@8000005P3oool20000
05D0oooo002F0?ooo`0608le0000000000000000000055l0a?oo>@3oool2000003L0oooo0P000005
0;C1h`040000002Y_>L0ZKcW0:Vli`<00000102K/NL00`000000SJ_Z08f[jP0308f[jP030000001m
Xn/0OJ?[0080OJ?[00<0000006fKk01]Vn`00P1]Vn`00`000000GYC/05jDk00205jDk0030000001B
SN/0DXg[0080DXg[1@00000203n6j@030000000oQ^T0000001D0oooo0P00001E0?ooo`00U`3oool0
0`3oa5l0500003F?o`0h0?ooo`<00000=`3oool2000000030;k4g`000000]<7S00D0]<7S0`000002
0:Rgi0030000002K/NL0Vk7W00<0Vk7W00@0000008f[jP2=ZnX0SJ_Z3`00000205jDk0050000001B
SN/0DXg[05:=j`0000000`18R>X6000001L0oooo00<000000?ooo`3oool0303oool00`2oM0000000
000000040000000304RLg`3oool0oooo03`0oooo002G0?ooo`030?o4G`0D0000=Hoo03H0oooo0P00
000h0?ooo`8000000`2na=l00`000000]<7S0;C1h`020;C1h`<000001@2X]n@010000000Vk7W09^a
i`2K/NL5000000<0SZOV00<00000082Pj020X>P00P20X>P00`000000LI[Y076Jj@02076Jj@030000
001STnT0000000D0000000<0DXg[00000018R>X00`18R>X00`000000@XKY00000002000001H0oooo
00<000000?ooo`3oool03@3oool00`3OW4P0000000000003000000030018W03Oool0oooo03`0oooo
002G0?ooo`030?o4G`0D0000=Hoo03<0oooo0`00000i0?ooo`030000002na=l0_/CO00@0_/CO00@0
00000;C1h`00000000000`2c_>000`000000Z;OT0:Rgi0020:Rgi0@000000`2L[N<00`000000SZOV
08jWiP0208jWiP0300000020X>P0P:3X0080P:3X00<00000076Jj@1aV^T00P1aV^T00`000000Hi?Y
06>Cj@0206>Cj@030000001GS^T0EhkY00D0000000@0B8SZ00000012Q^T000000P12Q^T00`000000
oooo0?ooo`0D0?ooo`030000003oool0oooo00d0oooo00<0gia804RLg`3oool00P3oool00`3OW7@0
W=oo0?ooo`0l0?ooo`00U`3oool00`3oa5l0500003F?o`0a0?ooo`800000>`3oool00`000000aLKK
000000060;k4g`030000002c_>00/kcP00<0/kcP00<000000:Rgi02X]n@00P0000020:Nbh0030000
002L[N<0W:gS0080W:gS00<0000008jWiP2>YnH00`2>YnH00`000000P:3X082Pj002082Pj0030000
001aV^T0LI[Y0080LI[Y00<0000006>Cj@1STnT00P1STnT01@000000EhkY05N>j@1GS^T000000080
CH[Y1000000204:6j@030000003oool0oooo01D0oooo00<000000?ooo`3oool03@3oool00`3OW4P0
M;oo0?ooo`100?ooo`00U`3oool00`3oa5l0500003F?o`0^0?ooo`<000000P3oool00`000000oooo
0?ooo`0f0?ooo`8000000P35a]/00`000000_/CO0;k4g`020;k4g`<000001P2c_>02000000@0Yk;P
00<0000009b]h`2L[N<00`2L[N<00`000000SZOV08jWiP0208jWiP0400000020X>P0P:3X082Pj0X0
000000<0Hi?Y0000001GS^T00`1GS^T00`000000CH[Y04f:j@0204f:j@D000005P3oool00`000000
oooo0?ooo`0>0?ooo`030;mdB02Lgol0oooo03l0oooo002G0?ooo`030?o4G`0D0000=Hoo02/0oooo
0`00000m0?ooo`0300000035a]/0aLKK00<0aLKK00@000000;k4g`00000000000`2l_m`010000000
/kcP0;>lh02c_>02000000030;:gg@000000Yk;P00@0Yk;P00<0000009b]h`2L[N<00P2L[N<90000
0080PikT00<0000007JHiP1fV>H00P1fV>H00`000000JI?W06VCi`050000000305N>j@000000CH[Y
00<0CH[Y00@0000004J8jP16R>X0AXSZ0P00000F0?ooo`030000003oool0oooo00d0oooo00<0oooO
09a8M02oool0?`3oool009L0oooo00<0olAO01@0000eSol0:@3oool2000003l0oooo00<000000<G6
f`35a]/01@35a]/00`000000_;oL0;bog0030;bog0040000002c_>000000000000<0/[OM00<00000
0:Nbh02W/^000`2W/^000`000000W:gS09b]h`020000000309bYg`000000T:CR00<0T:CR00<00000
08>Ni023W^@00`23W^@00`000000MYSV07JHiP0207JHiP030000001YTnL0JI?W0080JI?W00<00000
05f?j0000000100000000`1=R^T000000000000404J8jP030000003oool0oooo01@0oooo00<00000
0?ooo`3oool03@3oool00`3o_g@007Bo0?ooo`0n0?ooo`00U`3oool00`3oa5l0500003F?o`0V0?oo
o`<00000@03oool00`000000blSG000000050<G6f`8000001P2l_m`00`000000/[OM0;:gg@040;:g
g@040000002W/^00Yk;P0:Nbh0@00000102LZMl00`000000T:CR092ThP02092ThP0300000023W^@0
PikT0080PikT00<0000007JHiP1fV>H00P1fV>H00`000000JI?W06VCi`0206VCi`030000001MSnP0
GHoX0080GHoX00<0000005><j@0000001@0000000`16R>X000000?ooo`0E0?ooo`030000003oool0
oooo00d0oooo00<0oono07@0B02Lgol0?P3oool009L0oooo00<0olAO01@0000eSol0903oool20000
0440oooo0P0000020<_8e`0400000035a]/0aLKK0<G6f`8000000P33`]T010000000_;oL0;bog02l
_m`2000000030;ZkfP000000/[OM00D0/[OM0`0000020:N^g@030000002LZMl0W:WO00<0W:WO00<0
0000092ThP2@Y>800P2@Y>8010000000PikT08>Ni023W^@500000080MYSV00<0000006VCi`1YTnL0
0P1YTnL01@000000GHoX05f?j01MSnP000000080DhcY00<0000005><j@0000000P1<RnX3000001H0
oooo00<000000?ooo`3oool0303oool00`3ookl0M00004RLg`0n0?ooo`00U@3oool01@3ool@0Ga@0
000000000000=Hoo0240oooo0`0000000`3oool000000?ooo`0o0?ooo`030000003;b=L0blSG00<0
blSG00<000000<G6f`0000001033`]T010000000_;oL0000000000030;ZkfP030000002b]md0/[OM
0080/[OM0`0000050:N^g@030000002LZMl0W:WO00<0W:WO00<00000092ThP2@Y>801@00000308JL
h@030000001kV><0000000X0000000@0GHoX0000001CS>T000000P1CS>T00`000000C8_Z04b;jP02
04b;jP030000003oool0oooo01@0oooo00<000000?ooo`3oool01P3oool00`3o_i`0W=oo0?ooo`02
0?ooo`030?oOW0180000B9cO03h0oooo002G0?ooo`030?n?=@000000=Hoo01l0oooo0P0000140?oo
o`030000003;b=L0blSG00D0blSG00<000000<?2f@33`]T01033`]T00`000000^[_J0;ZkfP040;Zk
fP040000002b]md00000000000<0/;;J00<000000:N^g@2W[]d00P2W[]d010000000W:WO09bYg`2L
ZMl5000000<0TZ7N00<0000008JLh@26W>400P26W>400`000000NiSS07^Hh`0207^Hh`030000001_
U>D0KiCU0080KiCU00<0000006FAi`1UTNL01@0000000`1CS>T0000004b;jP0304b;jP030000003o
ool0oooo01@0oooo00<000000?ooo`3oool01P3oool0203ooml0W4Q809cOo`3oool0oono07@00000
07@0_ooo?P3oool00;H0oooo0`0000150?ooo`030000003?bM@0000000@0blSG0P0000000`39a=H0
00000<?2f@020<?2f@80000000<0`[kG0000002j^mX01@2j^mX00`000000/;;J0;2bfP030;2bfP03
0000002W[]d0YjkM0080YjkM1000000309bVg0030000002BXMh0TZ7N0080TZ7N00<0000008JLh@26
W>400`26W>401@000000NiSS07^Hh`1kV><0000000D0KiCU00@0000006FAi`1UTNL0II7W0P000000
0`1KSnT000000000000300000080C8_Z00<000000?ooo`3oool0503oool00`000000oooo0?ooo`06
0?ooo`060?oog`2LB00000000000000007@0_ooo?`3oool00;<0oooo0`0000160?ooo`8000000`3?
bM@010000000blSG0000000000030<W4eP0400000033`]T00000000000<0`[kG00<000000;ZkfP2j
^mX00P2j^mX3000000H0/;;J00@000000:N^g@00000000000P2VZ]X00`000000W:KL09bVg00209bV
g0030000002BXMh0TZ7N00<0TZ7N00<0000008JLh@26W>400P26W>400`000000NiSS07^Hh`0207^H
h`030000001_U>D0KiCU0080KiCU00<0000006FAi`1UTNL00P1UTNL00`000000FhoY05^?j@0205^?
j@030000001DSn/0000000<000004`3oool2000000030?ooo`000000oooo04/0oooo002a0?ooo`80
0000B03oool00`000000clWD0<o9e0040<o9e00300000039a=H0bLCF00@0bLCF00<000000<:ne`32
_]L01032_]L010000000^[_J0000000000030;Vge`030000002`/]X0/;;J0080/;;J0`0000040:JZ
fP030000002LY]`0W:KL00<0W:KL00@0000009:QgP2BXMh0TZ7N500000000`1UTNL0000005^?j@04
05^?j@030000001DSn/0E8o[0080E8o[00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`1:
0?ooo`00[P3oool4000004P0oooo00<000000<o9e03?bM@0103?bM@00`000000cLKC000000040<W4
eP0300000037`M@0000000H0`[kG00<000000;Vge`2i]mL00`2i]mL00`000000/;;J0;2bfP020000
0080[joG00<000000:JZfP2VZ]X00`2VZ]X00`000000W:KL09bVg00209bVg0@000000P2DWm/00`00
0000RY_N08ZKgP0208ZKgP0300000020V>40P9SQ0080P9SQ00<0000007JEh`1fUN<00P1fUN<00`00
0000KI?V06fCiP0;0000000305B?j`000000oooo01D0oooo00<000000?ooo`3oool0B@3oool00:`0
oooo0P0000030?ooo`030000003oool0oooo04D0oooo00<000000=79dP000000103?bM@010000000
cLKC0<g6d`0000030<W4eP0400000037`M@0al7D000000@0`[kG0`0000060;Vge`800000102_[mL0
0`000000YZ[J0:JZfP030:JZfP030000002LY]`0W:KL0080000000<0WJ?I0000002DWm/0102DWm/0
0`000000RY_N08ZKgP0208ZKgP0300000020V>40P9SQ0080P9SQ00<0000007JEh`1fUN<00P1fUN<0
1P000000KI?V06fCiP000000KI?V000000@0I9;Y00<0000005fCk01MTn`00`00000F0?ooo`030000
003oool0oooo04P0oooo002Y0?ooo`<00000C03oool010000000dLWB0=79dP0000030<o9e0030000
003=a]<0cLKC0080cLKC00<000000<W4eP0000001037`M@010000000`[kG0000000000030;nje@04
0000002i]mL0^KOG0;Vge`80000000<0][?E0000002_[mL0102_[mL010000000YZ[J0:JZfP2VZ]X4
000000@0WJ?I00<0000009BOf`2DWm/00P2DWm/00`000000RY_N08ZKgP0208ZKgP0300000020V>40
P9SQ0080P9SQ00<0000007JEh`1fUN<00P1fUN<01P000000KI?V06fCiP000000KI?V000000D0I9;Y
00D0000005fCk01MTn`0GI?/0000000F0?ooo`030000003oool0oooo04P0oooo002W0?ooo`800000
CP3oool00`000000dLWB0=79dP020=79dP030000003?bM@0000000H0cLKC00<000000<O1e037`M@0
1037`M@00`000000_k[E0;nje@030;nje@040000002i]mL00000000000<0][?E00<000000:n_e`2_
[mL0102_[mL300000080YJOG00<0000009fSf@2MXmT00`2MXmT00`000000U9oK09BOf`0309BOf`03
0000002:Vmh0RY_N0100000000<0KI?V0000001TT^T00`1TT^T00`000000GI?/05fCk00205fCk003
0000003oool0oooo01@0oooo00<000000?ooo`3oool0:P3oool00`3ogi`0W;oo0?ooo`0J0?ooo`00
Y03oool3000004l0oooo00<000000=79dP3AbM80103AbM800`000000cLKC0<g6d`040<g6d`800000
1@37`M@00`000000aKkC000000050;nje@030000002f/mD0][?E00@0][?E00@000000:n_e`2_[mL0
[joG0`0000050:FWe`030000002MXmT0WJ?I00<0WJ?I00<0000009BOf`2DWm/01@00000308jJf`03
00000025V=h0QISN0080QISN00<0000007fGh@1mUn400P1mUn400`000000MIKU0000000:00000003
05fCk0000000oooo01H0oooo00<000000?ooo`3oool0:@3oool00`3ookl0M7Bo0?ooo`0J0?ooo`00
X@3oool300000540oooo00<000000=79dP3AbM80103AbM800`000000d<SA000000030<g6d`800000
0P3<a=800`000000al7D0<O1e00200000080aKkC00<000000;nje@2o^]D00P2o^]D00`000000_KOD
000000050;Jce@040000002_[mL00000000000<0[J_E00<000000:FWe`2UYmL00P2UYmL010000000
WJ?I09fSf@2MXmT5000000<0UYgI00<0000008jJf`2>V]/00P2>V]/00`000000QISN08FHgP0208FH
gP030000001mUn40OIOQ0080OIOQ00<0000007FFi@0000000P1eU^D00`000000KIOX06fGj00206fG
j0030000001WV>`0IiS/00<000005P3oool00`000000oooo0?ooo`0Y0?ooo`030?oog`2LM9`0gooo
01X0oooo002O0?ooo`8000000`3oool00`000000oooo0?ooo`1=0?ooo`030000003Cb]40000000<0
dLWB0P0000030=38d@030000003=a]<0000000@0c<CB00<000000<O1e00000001@35_]<010000000
_k[E0000000000020;fge0030000002f/mD0][?E00<0][?E0P0000050:f[e@030000002UYmL0YJOG
0080YJOG1000000309fPeP030000002FWMT0UYgI0080UYgI00<0000008jJf`2>V]/00`2>V]/00`00
0000QISN08FHgP0208FHgP030000001mUn40OIOQ0080OIOQ0P00000307FFi@030000001]UnP0KIOX
0080KIOX00<0000006NHk01WV>`00P1WV>`00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo
02T0oooo00<0_gA809cOo`3oool06@3oool009`0oooo0`0000060?ooo`030000003oool0oooo04/0
oooo00L000000=?:d@3Cb]40dl[A0000003AbM80000000H0d<SA00<000000<c4dP3<a=80103<a=80
0`000000aKkC0<Fnd`050<Fnd`030000002m]m@0_KOD00<0_KOD00<000000;Jce@2f/mD00P000000
0`2d[m@000000:f[e@050:f[e@040000002UYmL0000000000080YJCD00<0000009fPeP2MX=H00P2M
X=H00`000000UYgI09JMf@0309JMf@030000002>V]/0SY[K0080SY[K00<0000008FHgP25V=h00P25
V=h00`000000OIOQ07fGh@0207fGh@8000000`1eU^D00`000000KIOX06fGj00206fGj0030000001W
V>`0IiS/00<0IiS/00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0V0?ooo`030?noM000
000000000080000000<0M;oo0?ooo`3oool05P3oool009X0oooo0P00001F0?ooo`030000003Cb]40
dl[A00<0dl[A00<000000=38d@3@b=40103@b=42000000D0c<CB00<000000<[1dP0000001@35_]<2
000000H0_KOD0P0000040;B_e0030000002]ZmD0[J_E0080[J_E0`0000040:FTe0030000002MX=H0
WJ3F00<0WJ3F00<0000009JMf@2FWMT00P2FWMT010000000SY[K08jJf`2>V]/M000001D0oooo00<0
00000?ooo`3oool09@3oool0203o_g@0000004RLg`3ogi`0B7Bo0?oo_`1d000007Bo5`3oool009L0
oooo0`00001G0?ooo`030000003Cb]40dl[A00<0dl[A00<000000=;9d@000000103@b=4010000000
clKA0<o6d@0000020<c4dP8000000P3:`M8010000000aKkC0<Fnd`35_]<200000080`k_B00<00000
0;fge02m]m@00P2m]m@00`000000^[?C000000040;B_e0030000002]ZmD0[J_E008000000P2/Z=<0
0`000000YJCD0:FTe0030:FTe0030000002MX=H0WJ3F0080WJ3F00@0000009JMf@2FWMT0UYgI1@00
0002096Jf@030000002:VM`0RYWL0080RYWL00<0000008BHg`24V=l00P24V=l00`000000OIWS07fI
h`0307fIh`030000001gVnL0Mi_W0080Mi_W00<00000076Nk01aW^`00`1aW^`00`000000oooo0?oo
o`0C0?ooo`030000003oool0oooo02@0oooo00T0gia80000M02oool0oono07@0M02oool0oooO09a8
0000M;l05P3oool009D0oooo0P00001I0?ooo`030000003Cb]40dl[A00<0dl[A00@000000=;9d@3B
bM4000000`3@b=401`000000clKA0<o6d@3?a]4000000<c4dP0000001@3:`M800`000000aKkC0000
00040<>kdP030000002m]m@0_KOD008000000P2j/m<00`000000]:oD0;B_e0030;B_e0800000102/
Z=<00`000000YJCD0:FTe0030:FTe0040000002MX=H0WJ3F09fPePD000000P2HW=H00`000000TI[I
096Jf@03096Jf@030000002:VM`0RYWL0080RYWL00@0000008BHg`24V=l0Q9SO0P00000407fIh`03
0000001gVnL0Mi_W0080Mi_W00<00000076Nk01aW^`00`1aW^`00`000000oooo0?ooo`0B0?ooo`03
0000003oool0000002D0oooo00X0gia80000M02oool0oooo0;md001d_ol0oooo0?oOW01804P0W=oo
5@3oool005D0oooo00D0omnL01@000000000000e08ooo`0h0?ooo`<000000P3oool00`000000oooo
0?ooo`1E0?ooo`030000003Db]40000000<0dl[A00<000000=;9d@3BbM400P3BbM400`000000d<SA
000000050<o6d@030000003:`M80b/7B00D0b/7B00<000000<>kdP33^m801033^m82000000D0^[?C
00L000000;B_e02d[m@0]:oD0000002b[=80000000D0[:SC00<000000:FTe02UY=@00P2UY=@40000
0080WYkD00<0000009RLeP2HW=H00`2HW=H00`000000TI[I096Jf@02096Jf@030000002:VM`0RYWL
0080RYWL00@0000008BHg`24V=l0Q9SO0P00000507fIh`030000001gVnL0Mi_W0080Mi_W00@00000
076Nk01aW^`0LIk/0`00000E0?ooo`030000003oool0oooo02<0oooo00X0omnL04P0B02Lgol0oooo
0?noM000M;l0oooo0?oo_`1d0000B9cO5@3oool005@0oooo00L0oono07@0000eSol0oooo0?n?=@00
03D0Sooo03@0oooo0`00001L0?ooo`070000003Db]40e<[A0=C:d@000000dl[A000000060=;9d@03
0000003?a]40clKA00<0clKA00<000000<g4d@0000001@3:`M82000000H0`k_B00<000000;Zcd`2j
/m<0102j/m<010000000]:oD0000000000020;:/dP030000002/Z=<0[:SC00<0[:SC00<000000:FT
e02UY=@00P0000000`2TX]<0000009jNe00409jNe0030000002HW=H0V9cF0080V9cF00<00000096J
f@2AV]T00`2AV]T00`000000RYWL08ZIg00E00000080O:?/00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`0S0?ooo`090;md0000B9`0gooo0?oo_`1d07@0_ooo0?oo_`1d0000B9cO01D0oooo
001D0?ooo`030;md0000B9`0a?oo0080oooo00<0ScD0000eS`3oool0<@3oool2000005h0oooo00<0
00000=C:d@3Db]400`3Db]400`000000d/WA0=;9d@040=;9d@030000003Ab=4000000080clKA0P00
00020<g4d@040000003:`M80b/7B0<[1dP8000000P37_]8010000000`k_B0<>kdP33^m8200000003
0<2hdP000000^[?C00D0^[?C00<000000;:/dP2b[=800`2b[=800`000000[:SC0:bXd`020:bXd`<0
0000102TX]<00`000000WYkD09jNe00209jNe0030000002HW=H0V9cF00<0V9cF00<00000096Jf@2A
V]T01@00000308nJfP030000002:Vmd000000080RY_M00<0000008FLhP25W>800P25W>800`000000
PIoV086OiP03086OiP030000001lXn`0O:?/0080O:?/00<000000?ooo`3oool0503oool00`000000
oooo0?ooo`0R0?ooo`090?oog`2LB00004RL0=oog`2LB000M;oo0=nLB00004P0W=oo01D0oooo001C
0?ooo`030?oo_`1d0000B9cO00<0oooo00<0a5lD0000=@2?ool0;P3oool3000005l0oooo00<00000
0=C:d@3Db]400`3Db]400`000000dl[A000000040=;9d@060000003Ab=40dLSA0000003?a]400000
1@3=a=400`000000b/7B000000040<NndP0400000033^m800000000000<0`;SB00<000000;Zcd`2j
/m<00P2j/m<3000000D0/ZcB00<000000:bXd`2/Z=<00P0000000`2ZYM8000000:BRd`040:BRd`03
0000002NW]@0WYkD00<0WYkD00<0000009RLeP2HW=H01@00000309BJe`030000002?V]X0Si[J0080
Si[J00<0000008ZKg@0000000P2:Vmd00`000000QIcR08FLhP0308FLhP0300000021WnH0PIoV0080
PIoV00<0000007bSk01lXn`00P1lXn`00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo02@0
oooo00<0okmd0000000000000P0000000`1d_ol0oooo0?ooo`0D0?ooo`00D`3oool00`3ogi`0B000
07Boo`030?ooo`030?n?=@0001@0GlCo02`0oooo0P00001Q0?ooo`030000003Db]40e<[A00<0e<[A
00@000000=?:d@3Cb]4000000`3BbM400`000000dLSA0=78d@020=78d@030000003=a=40cLCA00D0
cLCA00<000000<NndP37_]801037_]800`000000`;SB0<2hdP040<2hdP040000002j/m<000000000
00<0^;3B00<000000;:/dP2b[=800`2b[=82000000@0ZZGB00<000000:BRd`2TX]<00`2TX]<00`00
0000WYkD09jNe005000000<0VY_D00<0000009BJe`2DV]L00P2DV]L00`000000Si[J08nJfP0308nJ
fP8000000`2:Vmd00`000000QIcR08FLhP0208FLhP0400000021WnH0PIoV086OiPP000005P3oool0
0`000000oooo0?ooo`0U0?ooo`030?noM000B9`0gooo01L0oooo001C0?ooo`030?oOW0180000M;oo
00<0oooo00<0ohle0000501Oa?l0:@3oool3000000030?ooo`000000oooo05l0oooo00<000000=?;
d@0000000`3Db]400`000000dl[A0=?:d@020=?:d@030000003BbM40000000@0dLSA00<000000=37
d@0000001@3=a=400`000000bl;B000000040<NndP0300000035_=80000000H0`;SB00<000000;R`
dP2h/=800`2h/=8010000000/ZcB0;:/dP2b[=82000000030;2Yd@000000ZZGB00<0ZZGB00<00000
0:BRd`2TX]<00P2TX]<4000000<0WigB00<0000009ZKe02JVm@00P2JVm@00`000000U9[G09BJe`03
09BJe`030000002?V]X0Si[J0080Si[J0P00000208ZKg@X000000P2:Y>D00`000000QjW[08NYj`03
08NYj`030000003oool0oooo01@0oooo00<000000?ooo`3oool0903oool00`3ogi`0B01809cOo`0G
0?ooo`00D`3oool00`3ogi`0B00007Boo`030?ooo`030?n?=@0001@0GlCo02L0oooo0P00001T0?oo
o`070000003Cbm40dl_A0=?;d@000000e<[A000000060=?:d@030000003Ab=40dLSA0080dLSA00@0
00000=37d@3@am4000000`3=a=4200000080bl;B00D000000<NndP37_]80akkB000000020<FldP03
00000030^=80`;SB0080`;SB0`0000060;R`dP<000000`2`ZM400`000000ZZGB0:ZUdP030:ZUdP06
0000002TX]<000000000002TWm4000001@2OWM800`000000VY_D09ZKe00309ZKe0030000002DV]L0
U9[G0080U9[G00@0000008nJfP2?V]X0Si[J1@000002096Mf`030000002>X>00SZ3P0080SZ3P00<0
000008ZTi@2:Y>D00`2:Y>D00`000000QjW[08NYj`0308NYj`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool0903oool00`3ookl0M00004RLg`0G0?ooo`00D`3oool00`3ogi`0B00007Boo`03
0?ooo`030?n?=@0001@0GlCo02@0oooo0`00001U0?ooo`030000003Cbm40dl_A00<0dl_A00<00000
0=?:d@3Cb]40103Cb]400`000000dl[A000000030=78d@070000003@am40d<OA0=37d@000000cLCA
000000050<_2dP0300000037_]80000000@0aKcB00@000000<2hdP00000000000`2m]M800`000000
^;3B0;R`dP030;R`dP030000002`ZM40/:WA00@0/:WA00<000000:ZUdP2ZYM800P2ZYM82000000@0
Y9oA00<0000009nMdP2OWM800`2OWM800`000000VY_D09ZKe00209ZKe0040000002DV]L0U9[G09BJ
e`D000000P2DVmP00`000000TIgK096Mf`03096Mf`030000002>X>00SZ3P0080SZ3P00<0000008ZT
i@2:Y>D00P2:Y>D00`000000QjW[08NYj`0208NYj`<000005@3oool00`000000oooo0?ooo`0S0?oo
o`030?oog`2LB00007Bo01L0oooo001C0?ooo`030?oo_`1d0000B9cO00<0oooo00<0a5lD0000=@2?
ool08@3oool3000006L0oooo00<000000=?;d@3Cbm400`3Cbm400`000000dl_B000000030=?:d@80
00000`3Cb]400`000000dLSA000000050=37d@030000003;`]80bl;B00D0bl;B00<000000<FldP35
_=801035_=800`000000_KGB0;fedP030;fedP040000002h/=80^;3B0;R`dP80000000<0]JgA0000
002`ZM40102`ZM400`000000ZZGB0:ZUdP02000000030:VSd0000000Y9oA00<0Y9oA00<0000009nM
dP2OWM800`2OWM8010000000VY_D09ZKe02JVm@500000080V9_D00<0000009BKf02DVmP00`2DVmP0
0`000000TIgK096Mf`02096Mf`030000002>X>00SZ3P0080SZ3P00@0000008ZTi@2:Y>D0RZCU1P00
000309:^j@030000003oool0oooo01<0oooo00<000000?ooo`3oool0?03oool005@0oooo00<0gia8
0018W034ool00P3oool00`2?=@0003F?0?ooo`0O0?ooo`800000J@3oool2000000@0dl_A00L00000
0=?;dP3Cbm80dl_B0000003Cb]40000000H0dl[A00<000000=37d@3@am400`3@am400`000000c/KB
000000050<_2dP8000001@35_=800`000000`[[B000000050;fedP<000000`2e[M400`000000/:WA
0;2Yd@040;2Yd@8000000`2YXm000`000000Y9oA0:BOd@030:BOd@030000002OWM80WigB0080WigB
1000000209bKdP030000002HVm@0V9_D00<0V9_D00<0000009BKf02DVmP00P2DVmP00`000000TIgK
096Mf`03096Mf`030000002>X>00SZ3P00D000000`2CZ>@00`000000TZkY09:^j@0309:^j@030000
003oool0oooo01<0oooo00<000000?ooo`3oool0?03oool005@0oooo00L0oono07@0000eSol0oooo
0?n?=@0003D0Sooo01d0oooo0`0000020?ooo`030000003oool0oooo06D0oooo00@000000=;;d`3B
bm<000000P3Cbm400`000000dl_B0=?;dP030=?;dP030000003Cb]40dl[A00@0dl[A00<000000=79
dP0000000P3@am4200000080c/KB00@000000<_2dP3;`]80bl;B0P0000020<S0dP0300000035_=80
aKcB008000000P32^]800`000000_KGB0;fedP030;fedP030000002e[M40]JgA00@0]JgA00@00000
0;2Yd@2`ZM40/:WA0`0000050:VSd0030000002TWm40Y9oA0080Y9oA00<0000009nMdP2OWM800P00
00000`2OW=40000009bKdP0409bKdP030000002HVm@0V9_D0080V9_D00<0000009BKf02DVmP00P2D
VmP200000080TIgK1@00000309JSgP030000002CZ>@0TjST0080TjST00<0000009:^j@2B[^T00`2B
[^T2000001@0oooo0P00000m0?ooo`00E@3oool01@3ookl0=@000000000003D0Sooo01`0oooo0P00
001/0?ooo`070000003Bbm<0d/_C0=;;d`000000dl_A000000050=?;dP030000003Cc=<0000000@0
dl[A00H000000=79dP3AbM8000000=37d@0000050<k6dP030000003;`]80000000@0b<3B00<00000
0<FldP0000001@32^]801`000000_KGB0;fedP2m]M8000000;ZbdP0000001@2e[M4010000000/:WA
0000000000030:fVd0030000002YXm00ZJ?@0080ZJ?@00@000000:BOd@2TWm40Y9oA1000000409nL
d@030000002LVm80W9_B0080W9_B00<0000009RKe02HVm@00`2HVm@00`000000U9_H09BKf0050000
00<0UioJ00<0000009JSgP2FXmh00P2FXmh00`000000TjST09>Xi00309>Xi0030000002B[^T0TZkY
0080000000<0W;;X0000003oool05@3oool00`000000oooo0?ooo`0j0?ooo`00L`3oool3000006d0
oooo00<000000=;;d`3Bbm<00`3Bbm<00`000000dl_B0=?;dP030=?;dP040000003Cc=<0dlcC0000
00<0dl[A00<000000=79dP3AbM800P3AbM800`000000c/KB0<k6dP050<k6dP0300000038`=80b<3B
00@0b<3B00<000000<:jdP32^]801032^]8010000000_KGB0000000000020;ZbdP030000002e[M40
]JgA00<0]JgA0P0000050:fVd0030000002YXm00ZJ?@0080ZJ?@100000020:>Nd0030000002OW=40
WicA00<0WicA00<0000009bKdP2LVm800`2LVm800`000000V9_D09RKe0050000000409VMeP000000
VIgF000000@0UioJ00<0000009JSgP2FXmh00`2FXmh010000000TjST09>Xi02CZ>@4000000@0W;;X
00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0j0?ooo`00L@3oool2000006l0oooo00<0
00000=;;d`3Bbm<00`3Bbm<00`000000d/cD000000030=?;dP030000003Cc=<0dlcC0080dlcC00<0
00000=?:d@000000103AbM800`000000d<WD000000050<k6dP8000001@38`=800`000000aKkC0000
00060<:jdP030000002j/]80^[;B00<0^[;B00<000000;F]d@2e[M400P0000000`2bZm4000000:fV
d0050:fVd0060000002YXm0000000000002WXM0000001@2SW]000`000000WicA09nLd@0309nLd@03
0000002LVm80W9_B00D000000`2LW=<01P000000VIgF09VMeP000000VIgF000000D0UioJ00@00000
09JSgP2FXmh0UZ?N1@00000309b/hP030000002L/^P0W;;X0080W;;X00<000000?ooo`3oool0503o
ool00`000000oooo0?ooo`0i0?ooo`00KP3oool300000700oooo00<000000<o;eP0000000P3Bbm<2
000000<0d/cD00<000000=?;dP0000001P3Cc=<00`000000dLWB0=79dP020=79dP040000003@bM@0
d<WD000000<0c/KB0P0000020<c5d`0300000038`=80b<3B008000000P35_]<00`000000`[[B0<:j
dP020<:jdP<000001P2j/]82000000@0/Z_A00<000000:fVd02]Y]000`2]Y]02000000@0Yj7@00<0
00000:>Nd02SW]000P2SW]000`000000WicA09nLd@0209nLd@@000000`2NVm400`000000W9cC09bL
d`0209bLd`040000002IWMH0VIgF00000080VIgF00@0000009NOfP2GWmX0UioJ1@00000209fVg@03
0000002L[>80W:cR00<0W:cR00@0000009bbj02L/^P0W;;X0`00000F0?ooo`030000003oool0oooo
03P0oooo001[0?ooo`<000000P3oool00`000000oooo0?ooo`1/0?ooo`8000000P3?bmH00`000000
d/_C000000060=;<e0030000003Cc=<0dlcC00@0dlcC00<000000=;<e00000000P3AbM800`000000
d<WD0=39e0020=39e0030000003>a]80000000@0c<GC00<000000<S0dP0000001@35_]<010000000
`[[B0000000000030;jgd`040000002j/]80^[;B0;ZbdP<000001@2bZm4010000000[JK@0:fVd02]
Y]03000000D0Yj7@00<000000:>Nd02SW]000`2SW]001P000000WicA000000000000XIc@000000D0
WY_A00<0000009bLd`2LW=<00`2LW=<00`000000VIgF0000000209VMeP@000000P2MX]P00`000000
WJKM09fVg@0309fVg@030000002L[>80W:cR0080W:cR100000030:FgiP030000003oool0oooo01<0
oooo00<000000?ooo`3oool0>03oool006T0oooo0P00001c0?ooo`030000003?bmH0cl_F00<0cl_F
00<000000=;<e03Bc=@0103Bc=@00`000000d/kE000000030=?<d`8000000P3Bc=@00`000000dLWB
000000060=39e0030000003<aM<0c<GC00@0c<GC00<000000<Fnd`35_]<01@35_]<00`000000_[OC
0;jgd`030;jgd`040000002j/]800000000000<0]ZoB00<000000;:[d@2bZm400`2bZm43000000<0
ZjC@00<000000:NQd02WXM000P2WXM000`000000Xik@0:>Nd0020:>Nd0<00000102QW=000`000000
WY_A09jKd@0309jKd@030000002LW=<0W9cC0080W9cC1@0000000`2NWm@0000009fRf00409fRf003
0000002MY]d0WJKM0080WJKM00<0000009b/hP2L[>800P0000000`2T/>0000000:FgiP040:FgiP03
0000003oool0oooo01@0oooo00<000000?ooo`3oool0=`3oool006H0oooo0`00001d0?ooo`030000
003?bmH0cl_F00<0cl_F00<000000=3=e`0000000`3Bc=@2000000<0d/kE00<000000=?<d`000000
1@3Bc=@00`000000d<WD0=39e0040=39e0030000003>bMD0000000@0c<GC00<000000<S3e@000000
1035_]<3000000H0_[OC00<000000;J_dP2f[m800`2f[m800`000000/Z_A0;:[d@020;:[d@800000
1@2[Y=000`000000Yj7@0:NQd0030:NQd0040000002SW]00000000000080Xig?00<000000:6Ld02Q
W=000P2QW=000`000000WY_A09jKd@0309jKd@030000002LW=<0W9cC00H000000`2NWm@00`000000
WJ;H09fRf00209fRf0030000002MY]d0WJKM0080WJKM0`0000040:B`h0030000002U]nH0YKOV0080
YKOV0P00000F0?ooo`030000003oool0oooo03H0oooo001T0?ooo`800000MP3oool00`000000bl_I
000000030<o;eP070000003@cML0d<gG0=3=e`000000d/cD000000060=;>e@030000003Bc=@0d/cD
00<0d/cD00<000000=7=eP0000000`3@bM@2000000<0c/WE00@000000<c5d`00000000000`38`mD0
10000000aKkC0000000000030<:le@040000002n]m<0_[OC0;jgd`80000000<0^[GD0000002f[m80
102f[m8010000000/Z_A0000000000020:nXd@030000002[Y=00ZjC@00<0ZjC@00<000000:NQd02W
XM000P2WXM02000000@0Xig?00<000000:6Ld02QW=000`2QW=000`000000WY_A09jKd@0209jKd@@0
00000P2OWM800`000000WYoD0000000309jOe0030000002MX]P0WJ;H00<0WJ;H00H0000009fVg@00
000000000:BZf`0000050:B`h0030000002U]nH0YKOV008000000P2]^n<00`000000oooo0?ooo`0C
0?ooo`030000003oool0oooo03H0oooo001R0?ooo`800000MP3oool200000080bl_I00D000000<o;
eP3?bmH0cl_F000000040=3=e`030000003Bc]D0d/kE00@0d/kE00<000000=7?e`0000000`3Bc=@0
1`000000dLgF0=7=eP3AcMH000000=39e00000001P3>bMD00`000000b<?E0<S3e@040<S3e@030000
0032_=D0`[cE00<0`[cE00@000000;jgd`00000000000`2j]M@00`000000]ZoB0;J_dP030;J_dP80
0000102_Z=400`000000ZjC@0:^Td0030:^Td0030000002WXM00Yj7@0080000000<0YYo?0000002S
WLl0102SWLl00`000000XIc@0:6Ld0020:6Ld0030000002NVm4000000080000000<0X9c@0000002O
WM80102OWM800`000000WYoD0000000309jOe0030000002MX]P0WJ;H00@00000102TZ]/00`000000
Y;3P0:B`h0020:B`h0<00000102]^n<00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo03D0
oooo001D0?ooo`060;md00000000000000000000B9`0gooo203oool2000007D0oooo00<000000<_;
f@3;bmT00`3;bmT00`000000cl_F000000040=3=e`030000003@cmT0000000<0d/kE0P0000030=7?
e`030000003Bc=@0000000D0dLgF00<000000<k9e@3>bMD0103>bMD2000000D0b<?E00<000000<G2
e`0000001@32_=D00`000000^[GD0;Zee0040;Zee0070000002f[m80]ZoB0;J_dP000000/ZgC0000
00050:nXd@030000002[Y=00ZjC@0080ZjC@0`0000030:JOc`030000002SWLl0Xig?00<0Xig?00@0
00000:6Ld02QW=00XIc@100000040:2Ld0030000002OWM80WigB0080WigB00D0000009jOe02NWm@0
000009jOe00400000080XjGG00<000000:BZf`2TZ]/00`2TZ]/00`000000Y;3P0:B`h00200000080
[;?N00<000000:fkh`2]^n<00P2]^n<2000001H0oooo00<000000?ooo`3oool0=03oool005D0oooo
00<0omnL04P0001d_ol02P3oool00`000000oooo0000001c0?ooo`030000003;bmT0bl_I00D0bl_I
00<000000=3=e`3@cML00P3@cML01`000000d<oI0=3?f@3@cmT000000=;>e@0000001P3AcmL00`00
0000dLgF0=7=eP030=7=eP030000003?cMP0000000<0c/WE0P0000020<_8e`0300000038`mD0b<?E
008000000P35`]L00`000000`[cE0<:le@020<:le@030000002n^]H0000000D0^[GD00@000000;J_
dP00000000000P2b[M<00`000000[jSA0:nXd@030:nXd@030000002[Y=00ZjC@0080000000<0ZJ;@
0000002VWll0102VWll00`000000Xig?0:>Mc`020:>Mc`D0000000<0XYc?0000002PW=00102PW=00
0`000000WigB09nMdP0309nMdP060000002NWm@000000000002SXM<000001@2SYML00`000000Y:[K
0:BZf`020:BZf`<00000102//mh00`000000[K_S0:fkh`0200000080]KkQ00<000000?ooo`3oool0
4`3oool2000003D0oooo001E0?ooo`030?oOW0180000M;oo00X0oooo0P0000000`3oool000000?oo
o`1`0?ooo`0300000036b]h0000000D0bl_I00<000000<c>f`0000000P3@cML00`000000d<oI0=3?
f@030=3?f@030000003AcmL0dLoG00@0dLoG00<000000=3@f@0000000P3AcMH2000000<0clgH00<0
00000<k9e@000000103;b=L00`000000b<?E000000050<G2e`0400000032_=D0000000000080_[[F
00<000000;Zee02j]M@00`2j]M@2000000D0/ZgC00<000000:nXd@2_Z=400P2_Z=43000000@0ZJ;@
00<000000:JOc`2VWll00P2VWll00`000000Xig?0:>Mc`04000000D0XYc?00<000000:2Ld02PW=00
0`2PW=000`000000WigB09nMdP04000000050:>Qd`000000Xj7C0:>Qd`000000102SYML00`000000
Y:[K0:BZf`0200000080ZjgI00<000000:bcgP2//mh00`2//mh2000000D0]KkQ00<000000?ooo`3o
ool04@3oool300000140oooo00@0oooO09a80000B9`0gooo803oool005D0oooo00<0omnL04P0001d
_ol02@3oool2000000<0oooo00<000000?ooo`3oool0K03oool200000080a/[N00@000000<_;f@3;
bmT0bl_I0P0000020<c>f`030000003@cML0000000D0d<oI00<000000<oAfP0000000`3AcmL20000
0080d=3I00<000000=7=eP0000001P3?cMP00`000000blSG0<_8e`040<_8e`0300000035`]L0aL;G
00D0aL;G00<000000;jjeP2n^]H00`2n^]H00`000000^[GD0;Zee002000000030;Jbe@000000/ZgC
00@0/ZgC00<000000:nXd@2_Z=400P0000000`2/Y]4000000:VRd0040:VRd0030000002VWll0YYo?
0080YYo?100000030:>Mc`030000002RW<l0XYc?0080XYc?00<000000:2Ld02PW=000P2PW=040000
0080XioA00L000000:>Qd`2SXM<0Xj7C0000002SXM<0000000@0XjGG0`0000040:^]f@030000002/
/mh0[;?N0080[;?N00<000000;>gg00000001P2e_^400`000000oooo0?ooo`0C0?ooo`030000003o
ool0oooo00h0oooo00@0oooO09a80000B9`0gooo803oool004H0oooo00<0oooO09a8000000001`00
00000`18W=l0oooo0?ooo`020?ooo`030?oOW0180000M;oo00T0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`1Z0?ooo`0300000036b]h0a/[N00<0a/[N00<000000<_;f@0000001@3<
c]/00`000000d<oI0=3?f@030=3?f@070000003?dMX0cm7J0<oAfP000000dLoG000000050=3@f@03
0000003?cMP0clgH00@0clgH00<000000<c=fP0000000`3;b=L2000000030<O7fP000000aL;G00<0
aL;G0`0000060;jjeP800000102f/]D00`000000/ZgC0;:]d`020;:]d`<00000102/Y]400`000000
ZJ;@0:VRd0030:VRd0030000002VWll0000000<000001@2SWLl00`000000XYc?0::Lc`030::Lc`06
0000002PW=0000000000002SWLl000001@2SWm4020000000Xj7C0:>Qd`2SXM<000000:>Qd`000000
XjGG0P0000020:VXeP030000002[[MT0ZjgI00<0ZjgI00@000000:bcgP00000000000P2c]m`00`00
0000]KkQ0;Fnh@020;Fnh@80000000<0_<7N0000003oool04`3oool00`000000oooo0?ooo`0>0?oo
o`040?oog`2LB00004RL0=ooob00oooo001E0?ooo`030?oOW0180000M;oo00T0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0L0?oooa000000?03oool00`000000a/[N0<K:gP050<K:
gP030000003<c]/0c<kK00<0c<kK00<000000<c@g@0000000`3@cmT00`000000cm7J0<oAfP030<oA
fP030000003@d=T0d=3I00<0d=3I00<000000<kAf`0000000`3?cMP200000080c<gJ00@000000<_8
e`00000000001037amX010000000aL;G0000000000030<70f@040000002n^]H0_[[F0;jjeP<00000
1@2f/]D00`000000/ZgC0;:]d`02000000030:n[e0000000[:KA00@0[:KA00<000000:VRd02YX]00
0P2YX]03000000@0YIk?00<000000:>Mc`2SWLl00P2SWLl00`000000XYc?0::Lc`020::Lc`<00000
102SWLl00`000000XioA0:>Od@020:>Od@030000002SXM<0Xj7C0080Xj7C0`0000040:VXeP030000
002[[MT0ZjgI00<0ZjgI0P0000050;>gg0040000002e_^400000000000<0_<7N00<000000?ooo`3o
ool04`3oool00`000000oooo0?ooo`090?ooo`030;md00000000000000D0000000<0M;oo0?ooo`3o
ool07@3oool005D0oooo00<0omnL04P0001d_ol02@3oool2000000H0oooo0P00000I0?ooo`<00000
0`2H]nd01@000000QK3a08F`l@25/?40000000<0KjKc00@0000005VKlP1IVo80FI_b1@00000f0?oo
o`80000000<0_/WS00000036b]h00`36b]h2000000040<K>h000000000000<c>f`8000000`3<d=d0
0`000000d<oI000000050<oAfP030000003?dmd0000000<0d=3I00L000000<kAf`3>dM/0c]7K0000
003?cMP0000000D0c<gJ00<000000<O7fP37amX01@37amX00`000000`L3I0<70f@030<70f@040000
002n^]H00000000000<0^KSH00<000000;Jbe@2f/]D00P2f/]D3000000@0[j_D00<000000:bVd@2/
Y]400`2/Y]44000000030:JQd0000000YIk?00@0YIk?00<000000:>Mc`2SWLl00`2SWLl010000000
XYc?0000000000020:>Lc`030000002SWLl0Xig?0080Xig?00<000000:>Od@2SWm400`2SWm400`00
0000Xj7C00000002000000030:NTdP000000ZJSF00@0ZJSF00<000000:^]f@2[[MT00P0000000`2a
/=P000000;>gg0050;>gg0030000002l`Mh0_<7N00@0_<7N00<000000?ooo`3oool04`3oool00`00
0000oooo0?ooo`080?ooo`080?oOW01dW=l0oooo0?ooo`3ooml0W4P00018W03OoolP0?ooo`00E@3o
ool00`3ogi`0B00007Boo`090?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0503o
ool2000000030:Rlj0000000V;O]00@0V;O]00D0000008F`l@25/?40QK3a0000000306nVl`050000
001IVo80FI_b05VKlP0000000P14T>l500000300oooo0P0000040;k9h`0400000036b]h000000000
00D0a/kP00<000000<c@g@3<d=d0103<d=d00`000000cm7J0<oAfP020<oAfP8000000`3?dmd00`00
0000d=3I000000050<kAf`030000003<cMX0c<gJ00<0c<gJ00<000000<W=g@0000001@37amX20000
00H0`L3I00<000000;Vhf02i^=P00`2i^=P00`000000][;E0;Jbe@0200000080/K3G00<000000:n[
e02_Zm@00P2_Zm@00`000000[:KA0:bVd@020:bVd@<00000102VXM000`000000YIk?0:FNc`030:FN
c`030000002SWLl0Xig?00@00000102SW<l00`000000Xig?0:>Mc`020:>Mc`030000002SWm40XioA
0080XioA0`0000030:NTdP8000001P2YZ=H2000000@0/K3H00@000000;>gg02c]m`0/kOL0`000005
0;c1gP0300000032`m/0000001@0oooo00<000000?ooo`3oool02@3oool01`3OW4P0M;oo0?ooo`3o
oml0W4P00018W03Oool0803oool005D0oooo00<0omnL04P0001d_ol02@3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo0140oooo0P0000040:Rlj0040000002H]nd0V;O]09RgkA000000
00@0A93_0000000bQ^/0<XK[1000000[0?ooo`8000001`2nbN<00`000000a/kP0<K>h0040<K>h003
00000037dN80000000<0c=3M0P0000001@3<dml000000000003?dMX0000000H0cm?M00<000000<kA
f`3>dM/00`3>dM/00`000000c=;N000000030<c=fP0400000039cMd0bLgM000000<0alOJ0P000002
0<?6g00400000031`=T0`L3I0<70f@80000000<0_;kK0000002i^=P0102i^=P2000000@0/K3G00<0
00000:n[e02_Zm@00`2_Zm@400000080Z:CB00<000000:JQd02VXM000P2VXM000`000000YIk?0:FN
c`020:FNc`@000000`2SW<h00`000000Xic?0:>Lc`030:>Lc`030000002SWLl0Xig?0080Xig?00@0
00000:>Od@00000000000P2VXM000`000000YjCB0:NTdP020:NTdP8000000`2YZ=H3000000D0/K3H
00@000000;>gg000000000000`2i^]X00`000000_<7N0;c1gP0200000080`/?K00<000000?ooo`3o
ool04`3oool00`000000oooo0?ooo`080?ooo`070?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`0P
0?ooo`00E@3oool00`3ogi`0B00007Boo`080?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool03P3oool2000000030;BohP000000Z;cX00@0Z;cX0`00000309Vbj@0500000027Znd0Qj_]
08N[k@0000000`1cXnl010000000GYW_05jIk`1NVNl50000000403:6j`0000009h3X02N0j0800000
9`3oool200000080]<OX00<000000;k9h`2nbN<00P2nbN<2000000030;o=i@000000a/kP0080a/kP
0P0000020<OAhP040000003<d=d00000000000D0c=?O00<000000<oCg@3?dmd0103?dmd00`000000
cMGO000000020<kAf`8000000`3<d]h00`000000c<gJ000000040<W=g@0300000037amX0000000@0
`lKL00@000000<70f@00000000000`2l_]/00`000000^KSH0;Vhf0020;Vhf0030000002d]]X00000
00D0/K3G00<000000:n[e02_Zm@0100000040:RTdP030000002VXM00YZ7@00<0YZ7@00H000000:FN
c`00000000000:>Mc`0000050:>LcP030000002SW<l0Xic?00<0Xic?00@000000:>Mc`2SWLl0Xig?
0`0000040:JQd0030000002WY=80YjCB00<0YjCB00@000000:VXeP00000000000`2^Zm@00`000000
/K3H0;6`f0030;6`f0030000002i^]X0^K[J00<0^K[J00<000000;c1gP0000001@32`m/00`000000
oooo0?ooo`0C0?ooo`030000003oool0oooo00P0oooo00H0omnL04Qd_`3ooml0W4P00018W03OoolP
0?ooo`00D`3oool01@3ooml0W4P0000000000000M;oo00P0oooo0`00000;0?ooo`030000003oool0
oooo00/0oooo0P0000040;BohP030000002X_>P0Z;cX0080000000<0YkOT0000002I/^T0102I/^T0
1@000000Qj_]08N[k@27Znd0000000<0Lj?_00D0000005jIk`1NVNl0GYW_0000000204Z?kPD00000
00<09h3X0000000000000`00000Q0?ooo`8000001@2danP010000000_/WS0000000000040;o=i@03
00000036c^00000000D0am7R00<000000<cCg`3<dml0103<dml00`000000bmKR000000030<oCg@80
00000P3=eMl00`000000c]7K000000060<cBgP0300000039cMd0bLgM00@0bLgM00<000000<?6g033
a]`01033a]`00`000000_;kK0;bnf`040;bnf`040000002i^=P0000000000080];KJ00<000000;6`
e`2a/=L00`2a/=L5000000030:ZXe@000000Z:CB00@0Z:CB00<000000:JQd02VXM000P2VXM020000
00@0Xig?00<000000:>LcP2SW<h00`2SW<h00`000000Xic?0:>Lc`020:>Lc`@000000P2UW/l00`00
0000YZ7@0:JQd0030:JQd0030000002WY=80YjCB0080YjCB0P0000050:j[e0040000002a/=P0/K3H
0;6`f0<000001P2i^]X00`000000`/?K0<;3f`030<;3f`0300000037aMT0000001@0oooo00<00000
0?ooo`3oool02@3oool01@3OW4P0M;oO09a80000B9`0gooo0200oooo001E0?ooo`030?noM0000000
M;oo00P0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`080?ooo`80000000<0_L7M
0000002d_n80102d_n82000000@0YkOT00@0000009Vbj@2I/^T0VK;Y40000000101:Snh0000003V7
j`0iQn/60000000302J1j00000000000008000006`3oool2000000030:O3kP000000]<OX00@0]<OX
0P0000070;o=i@0300000037dN80am7R00<0am7R00<000000<KDi00000000`3<dml2000000<0bmKR
00<000000<oCg@0000001@3=eMl00`000000c=;N0<cBgP030<cBgP80000000<0bM;Q00000039cMd0
0P39cMd2000000030<G=h0000000`lKL00<0`lKL00<000000;k5g`0000001@2l_]/2000000D0];KJ
00<000000;6`e`0000001@0000030:ZXe@030000002XY=80Z:CB00<0Z:CB00<000000:JQd02VXM00
0P0000000`2TWm4000000:>Mc`030:>Mc`030000002SW<h0Xic>00<0Xic>00<000000:>Lc`2SW<l0
0P0000000`2TWLl000000:FNc`030:FNc`030000002VXM00YZ7@00<0YZ7@00<000000:NTdP2WY=80
0P0000000`2/Y]8000000:j[e0050:j[e0<000000`2f/mH010000000^K[J0;VjfP2i^]X200000004
0;nmf0000000`/?K0<;3f`8000000P37aMT00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo
00P0oooo00D0oooO09a8B018B00004RL0=ooo`0P0?ooo`00H03oool00`000000oooo0?ooo`0=0?oo
o`030000003oool0oooo00D0oooo0P0000040;g1g@030000002d_n80];oR0080000000<0/k[O0000
002W]n@0102W]n@3000000<0VJgU00D0000008VWj029YnP0RJOX0000000307RPj`040000001TV>d0
I9S]06BHk@D0000000@0>HO[0000000/PNT0;87Y1000000202J1j0D000004`3oool4000000@0Yl?^
00<000000;C7j02danP00P0000020;C;j`030000002ocND0_lgU0080_lgU0P0000000`2odNP00000
0<OAhP020<OAhP0700000036e>@0a]CT0<KDi0000000c=?O000000060<_FhP030000003=eMl0cMGO
00<0cMGO00D000000<_FhP00000000000<cBgP02000000<0bM;Q00@000000<W=g@00000000001035
cN0010000000`lKL0000000000030;k5g`030000002l_]/0_;kK0080000000<0]kcN0000002d]]X0
0`2d]]X3000000050;6`e`00000000000:f]f00000001@2ZZ=D00`000000Z:CB0:RTdP020:RTdP<0
00000`2TWm400`000000Xig?0:>Mc`030:>Mc`030000002SW<h0Xic>0080Xic>0`0000040:BMc`03
0000002UW/l0YIk?0080YIk?00<000000:JQd02VXM000P2VXM03000000@0[:KB00<000000:j[e02^
Zm@00P2^Zm@2000000D0][?F00@000000;VjfP00000000000`2o_MP00`000000`/?K000000050<O5
f@030000003oool0oooo01<0oooo00<000000?ooo`3oool0203oool0103ogi`0B0000018W03OoolP
0?ooo`00H03oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000<C2
f00000001@2m`Md2000000@0/k[O00<000000:Ngi02W]n@00P0000000`2W/^00000009V]i@0309V]
i@0500000029YnP0RJOX08VWj0000000101hX>/010000000I9S]06BHk@000003056@k0D0000000<0
;87Y0000000WOnP00`000000100VPNP000000368k00aR>`D000000<0U[gb00<000000:O3kP2W`nh0
0`2W`nh2000000D0]<_[00@000000;o=i@0000000000102odNP00`000000am7R000000050<KDi003
0000003;e^80bmKR00<0bmKR0P0000001039f>@000000000003=eMl2000000@0bmKR00<000000<WB
h@39d^401039d^400`000000aLgP0<G=h0050<G=h0030000002naMl0_/GO00@0_/GO0P0000040;Nl
gP@0000000@0];KJ0000000000000000102][MP00`000000ZZSE0:ZXe@020:ZXe@030000002XY=80
Z:CB0080000000<0YJ;C0000002TWm40102TWm400`000000Xig?0:>Mc`020:>Mc`030000002SW<h0
Xic>0080000000<0Xig?0000002TWLl0102TWLl00`000000YIk?0:FNc`020:FNc`030000002VXM00
YZ7@0080000000<0ZJ?@0000002/Y]80102/Y]800`000000[Z_D0:j[e002000000030;>^d`000000
][?F00D0][?F00<000000;nmf02o_MP0102o_MP00`000000alGI0<O5f@030<O5f@8000004`3oool3
000000/0oooo00<0gia80018W03Oool0803oool00600oooo0P00000@0?ooo`030000003oool0oooo
008000000P34`]P00`000000_L7M0;g1g@020;g1g@030000002l_MX0000000D0/k[O0P0000030:Nb
h0030000002I[ND0VJgU00<0VJgU00D0000008VWj029YnP0RJOX0000000307RPj`<0000000D0I9S]
0000001AT>`0DI3/000000030468j`@000000P0WOnP5000000050368k0000000@I;`046Bl0000000
101FW_@01@000000K:[f06bZmP1/Z_H0000000@0P[Ge00<0000009JmlP2F_O800`2F_O800`000000
Yl?^0:O3kP02000000030:K7l@000000]<_[00@0]<_[0P0000060;oAj08000001036e>@200000004
0<KGi`00000000000<_FhP8000001@39f>@00`000000bmKR0<_FhP030<_FhP0300000037f>H00000
00@0bM;Q00<000000<GCi@0000001035cN02000000030;o<i0000000_/GO0080_/GO0`0000030;Nl
gP<000000P2d]]X2000000030:jcg0000000[JgH00@0[JgH00@000000:ZXe@2ZZ=D0ZZSE10000004
0:FRd`030000002TWm40Y9oA0080Y9oA00@000000:>Mc`2SWLl0Xig?100000040:>Mc`030000002T
WLl0Y9g?0080Y9g?00<000000:FNc`2UW/l00P2UW/l3000000@0ZJ?@00<000000:bVdP2/Y]800`2/
Y]82000000@0/jkC00@000000;JceP2f/mH0][?F0P0000000`2k]mD000000;nmf0020;nmf0800000
00@0a;oF00000037aMT0alGI0P0000020<_7e`030000003oool0oooo01<0oooo00<000000?ooo`3o
ool0:`3oool00600oooo00<000000?ooo`3oool0403oool2000000D0a<;H00@000000;g1g@000000
00000`2l_MX00`000000/k[O0;>jg`02000000030;6fg0000000Yk;P00<0Yk;P00@0000009V]i@2I
[ND0VJgU2@00000207bNi`D0000000D0DI3/00000011R>/0@HS[0000000303B3j@<000000`0XP^T5
00000003046Bl0000000EYkd00<0EYkd00D0000006bZmP1/Z_H0K:[f0000000408:em@040000002F
_O80U[gb09JmlP@00000102Vao400`000000]<_[0;C;j`0200000080]<o^0P0000020;oAj0800000
0P2ne>X010000000a]CT0000000000050<KGi`0300000039f>@0bMST00<0bMST0P0000001037f^L0
00000<_FhP3;e^82000000<0amSV00@000000<WBh@00000000000`35dnD010000000aLgP00000000
00030;o<i0040000002naMl00000000000<0^<?S0`0000020;NlgP<00000102^/m`00`000000[JgH
0:f]f0020:f]f0@000000P2VY]H00`000000YJ;C0:FRd`030:FRd`030000002TWm40Y9oA0080Y9oA
100000020::Md0030000002SWLl0Xig?00<0Xig?00<000000:BMc`2TWLl00`2TWLl01P000000YIk?
000000000000Yj3?000000D0ZJ?@00@000000:bVdP2/Y]80[:KB100000040;>^d`040000002f/mH0
0000000000<0^kOE00@000000;nmf000000000000`34_mH00`000000alGI000000040<_7e`030000
003oool0oooo01@0oooo00<000000?ooo`3oool0:P3oool005l0oooo00<000000?ooo`3oool04@3o
ool00`000000a<;H0<C2f0050<C2f0030000002l_MX0_;gJ00@0_;gJ0P0000030;6fg0030000002W
/^00Yk;P00<0Yk;P0`00000209ZYh@030000002<Y>D0S:CU0080S:CU00D0000007bNi`1lW^L0O9kW
0000000306^Gj@D0000000H0@HS[0000000dPnT0=8?Y0000000^P^T40000000402R2j@000000<HS/
0368k1@000000`2E`OD00`000000Y/Oa0:K7l@020:K7l@<000001P2dcnh2000000D0_]CZ00<00000
0<KGi`36enL00`36enL2000000040<CJjP00000000000<WHi08000000`37f^L00`000000bmKR0000
00060<OHiP0300000035dnD0aM?U00<0aM?U0P0000060;o<i0040000002h`n<0^<?S0;S3h`<00000
0P2g_=h200000080/;[Q00<000000:jcg02^/m`00P2^/m`00`000000[JgH0:f]f00200000080Yj_J
00<000000:JVeP2VY]H00P2VY]H00`000000YJ;C0:FRd`030:FRd`060000002TWm4000000000002R
Wm8000001@2RWM000`000000Xig?0:>Mc`030:>Mc`030000002TWLl0Y9g?00@00000102WX<l00`00
0000ZJ?@0:VSd0030:VSd0<000000`2`ZM42000000@0/jkC00<000000;^ge@2k]mD0102k]mD00`00
0000a;oF0<BoeP040<BoeP030000003;amL0blOG00@0blOG00<000000?ooo`3oool04`3oool00`00
0000oooo0?ooo`0Z0?ooo`00G`3oool00`000000oooo0000000@0?ooo`0300000039`m@0000000@0
a<;H0P0000000`33_]H000000;bmfP030;bmfP030000002j^MP0000000@0/KKL00<000000:Nbh02W
/^000P0000000`2V[]d0000009ZYh@0309ZYh@030000002<Y>D0S:CU0080S:CU00D0000007bNi`1l
W^L0O9kW0000000306^Gj@030000001JTN/0FY7[00D0000000H0=8?Y0000000^P^T0;X;Y0000000^
Q>X4000000050368k0000000@9?a042Cl@000000101DX?D01@000000Jjgh06^]n01[[OP0000000@0
PKSh00<0000009G1m@2E`OD00P2E`OD00`000000Y/Oa0:K7l@0200000080Y/_c00<000000;C?kP2d
cnh00P2dcnh2000000030;CCl0000000_]CZ00<0_]CZ00D000000;kHk@00000000000<KGi`020000
00D0a=[Z00<000000<OJi`37f^L01037f^L00`000000amSV0<OHiP030<OHiP80000000@0`mSY0000
0035dnD0aM?U0P0000020;oBj08000000P2oc>@2000000040;W;i`0000000000000000<0^<?S0P00
00040;2jh@030000002^/m`0[[?L00<0[[?L0P0000040:N[fP030000002VY]H0YZKF0080YZKF00@0
00000:FRd`2UX]<0YJ;C100000040::OdP030000002RWM00XYg@0080XYg@00<000000:>Mc`2SWLl0
0P2SWLl400000080YIk?00<000000:NPc`2WX<l00`2WX<l00`000000ZJ?@0:VSd0020:VSd0800000
1@2`ZM4200000080/jkC0`0000050;^ge@8000001@34_mH2000000D0blOG00<000000<o9e0000000
5@3oool00`000000oooo0?ooo`0Y0?ooo`00G`3oool2000000l0oooo0P0000030<W3e00400000034
`]P00000000000@0`kkF00@000000;bmfP00000000000P2j^MP00`000000/KKL0;6fg0030;6fg080
00000`2V[]d00`000000VZWQ09ZYh@0209ZYh@040000002<Y>D0S:CU08bTi@`0000000@0FY7[0000
001;Rn/0Bh_[1@0000000`0^P^T0000002j4jP0202j4jPD0000000<0@9?a0000001DX?D00`1DX?D0
1@000000Jjgh06^]n01[[OP0000000@0PKSh00@0000009G1m@2E`OD0UL7e0`0000050:K;l`040000
002dcnh00000000000@0]=?`00@000000;kDjP0000000000102nf>d00`000000a=[Z0<CJjP030<CJ
jP80000000<0`]g]00000037f^L00`37f^L01@000000a=c[000000000000amSV008000001033f>T2
000000H0_m;X1@000000102ibnL000000;S3h`2h`n<2000000030;71iP000000/;[Q00@0/;[Q00<0
00000:jcg02^/m`00P0000000`2X/Mh000000:N[fP040:N[fP030000002VY]H0YZKF0080YZKF0`00
00030:6Qe0030000002RWm80XYoB0080XYoB00<000000::Md02RWM000`2RWM001P000000Xig?0000
00000000Xig?000000D0YIk?00<000000:NPc`2WX<l00`2WX<l01P000000ZJ?@000000000000[JK@
000000H0/:WA0`0000030;Rad`030000002k]mD0^kOE008000000P30^]@00`000000a;oF0<BoeP02
00000080b<;D00<000000<_7e`3;amL00P0000020<o9e0030000003oool0oooo01@0oooo00<00000
0?ooo`3oool0:03oool004@0oooo00D0omnL04P000000000001d0;ooo`030?ooo`060?noM0000000
M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`070?ooo`030000003oool0oooo00d0oooo
00<000000<W3e039`m@01039`m@00`000000`kkF0<>neP050<>neP030000002j^MP0^[WH00<0^[WH
00<000000;6fg02a]]`00P0000000`2`/MT000000:J^g@030:J^g@040000002JZN40VZWQ09ZYh@@0
00000P2>XN401@000000PIcT086Li021W>@0000000@0LYOW1@000000101;Rn/0000003j7j`0nQn/6
0000000402j4jP000000=8_]03B;kA@000000P2DaOP2000000<0Y/_c0`0000060;CCl08000001@2n
f>d01@000000_=__000000000000a=[Z008000001032gNd010000000am[W0000000000040<CLj`03
00000033f>T0`mSY00<0`mSY00<000000;kIk@000000102od^P4000000@0^L_W0`0000040;71iP04
0000002`^^40/;[Q0;2jh@@000000`2X/Mh00`000000Yj_J0:N[fP030:N[fP030000002VY]H0YZKF
0080000000<0XJCH0000002QXM@00`2QXM@00`000000XYoB0::OdP020::OdP030000002RWM00XYg@
0080XYg@0`0000040:>Mc`030000002UW/l0YIk?0080YIk?00<000000:NPc`2WX<l00P2WX<l30000
00@0[JK@00<000000;2Yd@2`ZM400P2`ZM42000000030;Rad`000000^;7C0080^;7C00<000000;^g
e@0000001030^]@00`000000a;oF000000040<S2e0030000003;amL0000000D0clWD00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`0X0?ooo`00@`3oool01`3ookl0M00007Boo`3oool0okmd
0000M02oool00P3oool01`3o_g@0000007Boo`3oool0oooO09a8M02oool00P3oool00`2oM000M;oo
0?ooo`070?ooo`030000003oool0oooo00`0oooo00<000000<g3d00000001039`m@00`000000b<3B
000000040<>neP80000000<0`K_D0000002j^MP0102j^MP2000000@0/;7I00@000000:J^g@2V[]d0
YZkM1000000209^Vg@030000002>XN40SZ7Q0080SZ7Q00D00000086Li021W>@0PIcT0000000307:G
i`040000001ST^T0Hi;Y06>Bj@D0000000H0?XO[0000000gQ^/0=hK[0000000eRN`40000000503B;
k@000000@YKb04:FlP000000101DX_H01@000000JK3i06V`n@1Y/?T0000000@0P;_j0P00000309C5
n0@000000`2UcoH200000080]=?`0P0000020;?Gl`80000000<0_]S]000000000000102lfnl00`00
0000`]g]0<;Mk@040<;Mk@8000001034g>/2000000040<3MkP00000000000<?Hj@8000000`2nfNd0
0`000000_m;X0000000200000080^M;/00<000000;W;i`2ibnL00P0000020;;9j`030000002a`NH0
/L7V0080/L7V100000020:Vgh`030000002X/Mh0Z;7N00<0Z;7N00@000000:N[fP2WZmX0Yj_J1000
00030:6Tf0030000002QXM@0XJ7D00<0XJ7D00<000000::OdP2RWm800P2RWm8010000000XYg@0000
000000020:6Md@030000002SWLl0Xig?0080Xig?00<000000:FNc`2UW/l00`2UW/l010000000Yj3?
0000000000020:VSc`030000002]Y]00[JK@0080[JK@00<000000;2Yd@2`ZM400P0000020;B]d@03
0000002h/M<0000000<0^;7C00<000000<2je030^]@01030^]@00`000000b<;D0<S2e0040<S2e003
0000003?bM@0clWD00<0clWD00<000000=;:d`0000005@3oool00`000000oooo0?ooo`0W0?ooo`00
@`3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO09a8M02oool0
1`3oool2000000`0oooo00H000000<g3d03=`m00cL?@00000039`m@2000000<0b<3B00@000000<>n
eP00000000001031^m@00`000000^[WH0;Zif002000000030;ReeP000000/;7I00@0/;7I0`000002
0:FZfP050000002KY]d0VjKM09^Vg@000000102>XN401@000000PIcT086Li021W>@0000000@0LYOW
00@0000006>Bj@1ST^T000000`1FS^X50000000303N6j`000000=HW/0080=HW/1@0000000`12U_80
000005BRmP0305BRmP050000001Y/?T0JK3i06V`n@0000000`20^oX500000080U<Wj00<000000:G?
mP2UcoH00`2UcoH3000000D0/mOc0P0000040;cKk`80000000<0^]kb0000000000000P32gNd20000
0080_]o`00@000000<CLj`00000000001@30gNh00`000000_]W]0;kIk@020;kIk@<000001@2id^`3
000000@0//W[00<000000;71iP2a`NH00P0000000`2Y_nT000000:Vgh`040:Vgh`030000002X/Mh0
Z;7N0080Z;7N100000020:6Yg0030000002QY=P0XJCH00<0XJCH00<000000:6Qe02QXM@00P2QXM@0
10000000XYoB0::OdP2RWm83000000@0XIgA00<000000:>Mc`2SWLl00`2SWLl00`000000YIk?0:FN
c`020:FNc`800000102YXll00`000000[JK@0:fVd0030:fVd0800000102d[M400`000000^;7C0000
00020;Rad`030000002l]=80000000@0`;[D00<000000<Bmd`0000001038`]@00`000000c<CC0000
00020<o9e08000000P3Bb]<00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo02H0oooo0012
0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02`3oool00`2oM000M;oo0?ooo`06
0?ooo`030000003oool0oooo00X0oooo00<000000<g3d03=`m000`3=`m000`000000b<3B0<S0dP04
0<S0dP0300000031^m@0`K_D00D0`K_D0P0000040;ReeP030000002`/MT0/;7I0080000000<0[jkG
0000002UZ]X00P2UZ]X00`000000VjKM09^Vg@0209^Vg@040000002>XN40SZ7Q08jQh@`0000000<0
Hi;Y0000001FS^X00P1FS^X00`000000BXc[04Z<j`060000000403F9k0000000>Y3_03Z@k`l00000
0`1o_o/00`000000U<Wj09C9nP0209C9nP030000002UcoH0YLof008000000P2TdoP200000080/mOc
0P0000020;7Jm@040000002lfnl00000000000D0^]kb0P0000050;kOl00300000030gNh0`=g^00<0
`=g^0P0000000`2kgO4000000;kIk@0300000080^=Sa00<000000;WBk02id^`00P0000020;;@l003
0000002bbN/0//W[00<0//W[0P0000040:Voj@030000002Y]n<0ZKOS0080ZKOS00<000000:RagP2X
/Mh00P0000000`2Q[^4000000:6Yg0040:6Yg0030000002QY=P0XJCH0080XJCH00@000000:6Qe02Q
XM@0XJ7D1@00000209nNd`030000002QWM40XIgA00<0XIgA00<000000:>Mc`2SWLl00P2SWLl00`00
0000YIk?0:FNc`02000000030:JPd0000000ZJ??00@0ZJ??00L000000:fVd02]Y]00[JK@0000002`
Z=00000000D0]:gA00@000000;Rad`00000000000P2l]=801@000000`;[D0<2je030^]@000000080
a;gC00D000000<S2e038`]@0b<;D000000020<c4d`030000003?bM@0000000D0d/[C00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`0V0?ooo`00=@3oool00`3ooml0W4P00000000700000006
04RLg`3oool0oooo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0:0?ooo`030?oog`2LB000
B9cO00H0oooo00<000000?ooo`3oool02P3oool00`000000cL?@0<g3d0030<g3d0030000003<`Ll0
000000@0b<3B00<000000<Jmd@0000001031^m@3000000H0^;GF0P0000030:n^e`030000002UZ]X0
YJ[J0080YJ[J00@0000009^Vg@2KY]d0VjKM10000002096Og@0500000025Vn40QI_Q08FKh@000000
0`1iUn@00`000000KICW000000070000000604Z<j`000000@Xc]04:<k@000000@8o_10000000100j
T>l0000004FJl`15V_<2000000<0EZKh0P00000206Zcn`800000101o_o/010000000U<Wj09C9nP2D
bOX3000000D0Y=?h0`0000050;7Jm@030000002jg_80^]kb00<0^]kb0P0000000`2gh?@000000000
00020;kOl080000000@0^^3b000000000000`=g^0P0000020;_Ml@@000001@2hf?42000000D0/]3`
00<000000;;9j`2bbN/00P0000000`2Ya^h000000:Voj@030:Voj@040000002Y]n<0ZKOS0:Vgh`@0
0000102Q[^400`000000XJWL0:6Yg0020:6Yg0040000002QY=P0XJCH0:6Tf0D000000P2NX=H00`00
0000WikC09nNd`0209nNd`030000002QWM40XIgA00<0XIgA00@000000:>Mc`2SWLl0Xig?10000003
0:JPd0030000002YXll0ZJ??00<0ZJ??00@000000:fVd000000000000P2`Z=000`000000]:gA0;B]
d@040;B]d@030000002l]=80000000<0_;CB00<000000<2je00000001034_M<00`000000b<;D0000
00040<c4d`030000003Bb]<0d/[C00<0d/[C00<000000=C<d@0000004`3oool3000002L0oooo0012
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3OW4P0000007Boo`06
0?ooo`030000003oool0oooo00T0oooo00<000000<g3d03=`m000`3=`m0010000000c<7?0<c1c`00
00030<S0dP0600000036_M40a[gA0<Jmd@000000`K_D0P0000030;nhd`040000002h]MH0^;GF0;Re
eP80000000<0][7D0000002_[]L00`2_[]L010000000YJ[J0:FZfP2UZ]X400000080Vj?J00D00000
096Og@2AWmd0TIoM0000000308FKh@030000001iUn@0NIOT0080NIOT00D0000006fDi`1]U>L0KICW
00000003066Bj@D0000000<0@Xc]00000010Snl00P10SnlH00000080Tlgk0P0000020:CCn0800000
0P2SeoX200000080/M[e0P000000102_gOL000000000002jg_82000000@0]n3d0`0000040;[PlPP0
000000<0]=ge0000000000000P2hf?42000000030;3Gm0000000000000<0/]3`100000040:W6kP04
0000002Y_nT0ZKoY0:Voj@@000000`2Q]>H00`000000XJkQ0:6^h@020:6^h@040000002QZM`0XJWL
0:6Yg0D000000P2LXmT00`000000WZ3F09jPeP0309jPeP030000002OW]<0WikC0080WikC00<00000
0:6Md@2QWM400P2QWM4400000080Xio@00<000000:JPd02VX=000`2VX=000`000000ZJ??0:VSc`02
0:VSc`8000001@2`Z=0010000000]:gA0;B]d@2d[M43000000030;bddP000000_;CB00<0_;CB00<0
00000<Bmd`34_M<01034_M<00`000000c<CC0<c4d`020<c4d`030000003?a]80000000<0d/[C00D0
00000=C<d@3Dc=40e<cA0000000D0?ooo`030000003oool0oooo02D0oooo00120?ooo`030?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01`3Ogml0oooo0?oo_`1d000000000000M02o
ool01P3oool2000000T0oooo00<000000=33cP0000000`3=`m000`000000c<7?0<c1c`020<c1c`03
00000038`=80000000D0a[gA00<000000;nhd`2o^=<00`2o^=<010000000^;GF0000000000030;Ja
e0030000002_[]L0[jkG00<0[jkG0`0000020:FVe`050000002KXmX0Vj?J09^SfP000000102AWmd0
1@000000QI_Q08FKh@25Vn40000000<0NIOT00<0000006fDi`1]U>L00P1]U>L010000000HI;Y066B
j@00000305NAk0H0000000@0@8o_00000013UO40@iGa0P00000204bOm@030000001JZoT0FZ_i0080
FZ_i00@0000006bgo01/]o`0K;Ol0P0000030833o@8000000`2CcO/3000000D0XmOj0`0000050:oM
m`8000000`2gh?@200000080/n;g00<000000;[PlP2jh?80200000040;CMm@<00000102`eo@40000
00<0ZLkc00@000000:W6kP2Ya^h0ZLK^1@0000020:6kk0030000002Q]>H0XKCV00<0XKCV00@00000
0:6^h@2Q[^40XJkQ1000000209^VgP030000002LXmT0W:?I00<0W:?I00<0000009jPeP2NX=H00P2N
X=H00`000000WikC09nNd`0209nNd`030000002QWM40XIgA0080000000<0X9kB0000002SWm00102S
Wm000`000000YZ3@0:JPd0020:JPd0030000002YXll0ZJ??0080000000<0[:G@0000002`Z=00102`
Z=0010000000]:gA0000000000030;R`d@040000002l]=8000000;bddP<000001@34_M<00`000000
b<3B000000020<c4d`070000003?a]80clKB0<o6dP000000d/[C000000030=C<d@030000003FcM40
000001@0oooo00<000000?ooo`3oool0903oool00480oooo00<0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`060?ooo`060=nLB000000000000000000004P0W=oo1`3oool00`000000oooo0?oo
o`070?ooo`070000003@`lh0d<?>0=33cP000000cL?@000000060<c1c`0300000036_M40a[gA00<0
a[gA00<000000<Bjd00000001@2o^=<00`000000][7D0;Jae0040;Jae0030000002_[]L0[jkG0080
000000<0[J[D0000002UY]L00P2UY]L00`000000Vj?J09^SfP0209^SfP040000002AWmd0TIoM096O
gA00000000<0HI;Y0000001GTN`00P1GTN`01@000000Ci;^04nBkP000000C9Ka00P0000000<0C9oe
0000001JZoT00`1JZoT010000000K;Ol06bgo01/]o`900000080Tm3m0P0000020:?GnP8000000P2R
f_/200000080[mgg0P0000020:gOn@<000001@2ch_L4000000030;KQm@00000000000080/>7h00<0
00000;CMm@2dgOD00P0000020:gLn0030000002`eo@0/=Od008000000P2We?L00`000000ZLkc0:W>
l`030:W>l`<000000`2P`o400`000000XK_/0:6kk0020:6kk0040000002Q]>H0XKCV0:6diPD00000
0P2IZn<00`000000VjKN09^VgP0209^VgP030000002LXmT0W:?I0080W:?I00<0000009jPeP2NX=H0
0`2NX=H00`000000WikC09nNd`04000000@0X9kB00<000000:>Od02SWm000P2SWm000`000000YZ3@
0:JPd0020:JPd0<00000102/YM000`000000/:S@0;2Xd0030;2Xd0030000002h/=40^;3A00<0^;3A
00@000000;bddP00000000000`30^=800`000000a;gC0<Bmd`0200000080b<3B00<000000<c4d`00
00001@3?a]801@000000e<cA0=C<d@3Dc=4000000080e/gA00<000000?ooo`3oool04`3oool00`00
0000oooo0?ooo`0S0?ooo`00@P3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0
oooo00@0oono07@000000000B9cO2@3oool00`000000oooo0?ooo`060?ooo`030000003@`lh0d<?>
00<0d<?>00<000000<c1c`3<`Ll0103<`Ll00`000000bko?000000030<Jmd@0400000034^]00a;[@
000000@0_kSC00<000000;fddP0000001P2f/M@2000000<0[J[D00<000000:FVe`2UY]L00P2UY]L0
10000000Vj?J09^SfP2KXmX400000080TigJ00D0000008ZJg@2:V]d0RY[M00000003082Hh@060000
001fU^@0MYKT07JFi0000000K9KX200000000`1?T^h0000004bFl@0204bFl@040000001=W?@0CIcd
04fLm0/000000P1__?d2000000<0PLOn0P00000309?@o@<00000102Rf_/400000080[Moi0P000002
0:WQnP8000000P2ch_L3000000040:oSn@0000000000000000D0/>7h0P0000050:gLn08000001@2W
e?L00`000000ZLkc0:W>l`020000000309k:mP000000X<?a00<0X<?a00@000000:6kk02Q^n`0XK_/
1@00000209Raj0030000002IZn<0VJ_S0080VJ_S00<0000009^VgP2KY]h00P2KY]h00`000000W:?I
09bSf@0309bSf@030000002NX=H0WZ3F00D000000P2MW]@00`000000X9kB0:2NdP030:2NdP030000
002SWm00Xio@00<0Xio@00H000000:JPd000000000000:RRd@0000050:bUd0040000002`Z=00/:S@
0;2Xd0<000001P2h/=4010000000`;SB0<2hdP0000020<2hdP0300000034_M<0000000D0b<3B00<0
00000<o6dP3?a]800`3?a]800`000000d/WA000000020=C<d@050000003FcM40e/gA0=K=d@000000
503oool00`000000oooo0?ooo`0S0?ooo`00@`3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000
M;l0oooo00L0oooo00<0_g@007Boo`3oool02@3oool00`000000oooo0?ooo`050?ooo`030000003@
`lh0d<?>00<0d<?>00<000000<o2cP0000000`3<`Ll2000000<0bko?00<000000<Jmd@0000001034
^]0010000000_kSC0000000000020;fddP030000002f/M@0][7D0080][7D0`0000050:fZe0030000
002UY]L0YJKG0080YJKG0`00000209bPe`050000002CWMX0TigJ09>MfP0000000`2:V]d01@000000
P9SQ082Hh@20V>40000000@0MYKT00D0000006bFj01/U^P0K9KX0000000306BFk0X0000000@0CIcd
0000001DYOL0E:Gg0P00000305n`nP8000000`1__?d900000080Tm?n1@0000030:3Lo0<000001@2Y
hOX300000080[n?i0P0000020:WSnP8000000P2`hOP2000000030:WPnP00000000000080[Mch0P00
00000`2Tf_X00000000000030:ODm`@00000102Nb_H010000000X<?a0:33l@2P`o4400000080U[O^
00<0000009Raj02H/NP00`2H/NP00`000000VJ_S09V[h`0209V[h`030000002KY]h0VjKN0080VjKN
00<0000009bSf@2LXmT01@00000209ZOf0030000002MW]@0WIkD00<0WIkD00<000000:2NdP2PW]80
0`2PW]800`000000Xio@0:>Od004000000@0Z:;A00<000000:bUd02/YM000`2/YM03000000<0/jcA
00@000000;R`d@2h/=40^;3A0`0000020<2hdP0500000030^=80`;SB0<2hdP0000001@38`=820000
00@0clKB00H000000=;9d@3BbM4000000=C<d@0000030=K=d@030000003Hc]00000001@0oooo00<0
00000?ooo`3oool08P3oool004<0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo
00H0omnL04P00000000000000000B02Lgol50?ooo`030000003oool0000000D0oooo00<000000=73
c@0000000`3@`lh01`000000cl;>0<o2cP3?`/h000000<c1c`0000001P3;_ll00`000000a;[@0<Bj
d0040<Bjd0030000002m]=80_KCB00<0_KCB00@000000;Jae000000000000`2d[]800`000000[J[D
0:fZe0020:fZe0030000002UY]L0YJKG0080000000<0Y:CE0000002LX=L00P2LX=L00`000000TigJ
09>MfP0209>MfP050000002:V]d0RY[M08ZJg@0000000`20V>401@000000MYKT07JFi01fU^@00000
00@0K9KX00D0000006BFk01TU^`0I9K/0000000205fIk`060000001IWO80FIgb05VMlP000000FZ?f
3P00000207C1oP8000000`23boh4000000<0TmKn0P0000020:3Lo080000000@0Wmkm000000000000
ZN7j0`0000020:KRn`<000001@2YhoX3000000<0ZN3j100000030:CJnPD000000`2MdOX010000000
W/[f09k:mP2Nb_H500000080UKkd00<0000009JgkP2F]nh00P2F]nh00`000000V;7X09Raj00209Ra
j0040000002IZn<0VJ_S09V[h`T000000`2GXM/00`000000VYoH09ZOf00209ZOf0030000002MW]@0
WIkD0080WIkD00<000000:2NdP2PW]800P2PW]8400000080Y:3B00<000000:RRd@2XX]400`2XX]40
0`000000[:G@0:bUd0020:bUd08000001@2c[=4010000000^;3A0000000000030;bcdP0300000030
^=80`;SB00<0000000@0a;_B00000038`=80b<3B0P0000020<c3dP040000003?a]80clKB000000@0
d/WA00D000000=K=d@3FcM40e/gA000000020=S>d0030000003oool0oooo01<0oooo00<000000?oo
o`3oool08@3oool004@0oooo00D0oono07@000000000001d0;ooo`0:0?ooo`050;md000000000000
00000018W=l01@3oool2000000D0oooo00L000000=73c@3A`ld0dL?=0000003@`lh0000000D0cl;>
00<000000<^oc`3;_ll0103;_ll00`000000bKg?000000040<Bjd00300000032^=00000000D0_KCB
00<000000;B^dP2d[]800`2d[]800`000000[J[D0:fZe0020:fZe0<000000`2TY=D00`000000W:3G
09bPe`0209bPe`050000002CWMX0TigJ09>MfP0000000`2:V]dD0000000305fIk`000000FIgb0080
FIgb00@0000005ZSmP1JXoH0FZ?f0P00000205j/n@8000000`1W]_`010000000M<7n07C1oP1d`Oh5
00000080Q/oo0P00000309?FoP@00000102Og_d3000000<0Y^;k0P0000000`2Sh_`0000000000002
0:WSnP8000000P2Th_`4000000<0X]kl1000000309_Go0040000002MdOX0WM7j09gAnPD000000`2C
aOP00`000000UKkd09Fnm00209Fnm0050000002F]nh0U[O^09JgkP0000000`2H/NP500000080TJSU
00<0000009BTh02DY>000P2DY>000`000000Uj7K09NQf`0209NQf`030000002JWmP0VYoH0080VYoH
00<0000009fNe02MW]@00`2MW]@01P000000X9kB000000000000WioD000000D0Y:3B00<000000:RR
d@2XX]400`2XX]401P000000[:G@000000000000[ZSA000000H0/jcA00<000000;bcdP2l/m800`2l
/m800`000000`;SB0000000200000080a;_B00<000000<S0dP000000103<`m800`000000clKB0000
00040=;9d@030000003Dbm4000000080e/gA00D000000=S>d03Hc]00f<k@0000000D0?ooo`030000
003oool0oooo0240oooo001M0?ooo`030000003oool0oooo00<0oooo00<000000=73c@3A`ld00`3A
`ld00`000000cl;>0<o2cP030<o2cP030000003>`<d0000000<0bko?0P0000030<Vmc`0400000034
^]00000000000080`[S@00<000000;fddP2m]=800P2m]=800`000000^[7A000000050;B^dP060000
002]Z]@000000000002/Ym<000001@2TY=D01@000000W:3G09bPe`2LX=L0000000<0TigJ10000002
08jJfP0500000027VMh0QiWN08NIgP0000000`1oVN801@000000MiWV07NIiP1gVNH0000000@0LI_[
5`00000307[5oP8000000`26col4000000@0U=Wn1000000309kOo@<00000102Sh_`3000000<0Y>;l
0P0000000`2Oh?d00000000000020:;No080000000<0V]_m0000000000000`2Keo`4000000<0T/cl
00<0000009?5n02CaOP00P2CaOP010000000UKkd09Fnm02E__@900000080SZg[00<00000096Xi@2A
Z>D00`2AZ>D01@000000U:CP09BTh02DY>00000000D0Uj7K00<0000009ZOf02JWmP00P2JWmP00`00
0000WIkD09fNe004000000@0WioD00<000000:BPdP2TX=800P2TX=800`000000Z:;A0:RRd@020:RR
d@<00000102^Z=4010000000/jcA0;>/d@2c[=43000000H0_;?B00D000000<BkdP34^m80a;_B0000
00020<BkdP030000003<`m80c<?B00@0c<?B00<000000=;9d@3BbM400P3BbM401P000000e<_A0=C;
d@000000e/gA000000<0f<k@00<000000=S?d0000000503oool00`000000oooo0?ooo`0P0?ooo`00
G@3oool00`000000oooo0?ooo`020?ooo`030000003A`ld0dL?=00<0dL?=00<000000=73c0000000
0`3?`/h01`000000c/3=0<k0c@3>`<d000000<^oc`0000001P39_Ll00`000000`[S@0<:hd0030<:h
d0040000002m]=80000000000080^[7A00<000000;B^dP2d[]800`2d[]82000000@0[:OC00<00000
0:BTe@2TY=D00P2TY=D010000000W:3G09bPe`2LX=L400000080UYcG00D0000008jJfP2>V]X0SY[J
0000000308NIgP050000001oVN80OiWR07nIhP000000101gVNH01@000000LI_[076Kj`1aVn/00000
00<0Jio_00D0000006RTl`1XY?<0J:Cc0000000406N[m`040000001Z/oX0J[?j06ZcnP8000000P1`
_?d2000000<0N/Gn1@00000208WBo`8000000P2DfOh30000000409GKoP000000W]om09kOo@800000
0`2MhOd4000000<0X>;l0`00000309oPo@@000000`2Jfod500000080TM;n0P00000309;<o0T00000
0`29^OD01@000000Rk;`08^bl02;/_00000000@0SZg[00<00000096Xi@2AZ>D00P2AZ>D00`000000
U:CP09BTh00209BTh0030000002GXM/0Uj7K0080Uj7K00@0000009ZOf02JWmP0VYoH1000000309^N
eP030000002OWm@0WioD0080WioD00<000000:BPdP2TX=800`2TX=801P000000Z:;A000000000000
ZJGB000000D0[ZSA00@000000;>/d@00000000000`2g[m800`000000_;?B0;bcdP020;bcdP800000
0`34^m8010000000a;_B0000000000050<c3dP030000003?a]4000000080d/WA00<000000=C;d@3D
bm400P3Dbm401@000000f<k@0=S>d03Hc]0000000080f<o@00<000000?ooo`3oool04`3oool00`00
0000oooo0?ooo`0O0?ooo`00G@3oool200000080oooo00<000000=73c@3A`ld00`3A`ld01`000000
dL?<0=73c03A`l`000000<o2cP0000001@3>`<d00`000000bKg?0<Vmc`040<Vmc`0300000037^ll0
000000D0`[S@00<000000;Zad@2j/M400`2j/M400`000000]:kB0;B^dP02000000030;:[d@000000
[:OC00<0[:OC00@000000:BTe@2TY=D0Y:CE1@0000000`2MW]D0000009JLe`0309JLe`030000002>
V]X0SY[J0080SY[J00D0000008NIgP27VMh0QiWN0000000307nIhP050000001gVNH0MiWV07NIiP00
0000101aVn/01@000000Jio_06^Ok`1[Wnl0000000<0J:Cc00D0000006N[m`1WZoL0Ij_g00000003
06ZcnPT000000`21b_l010000000RM;o08WBo`29d_l4000000@0UM_n0`00000309gQo@8000000P2L
hOd00`000000X>;l0:3Ro00200000080WN7m1000000309WNoP@000000`2Beoh010000000TM;n097B
oP2Ad_h500000080Q/Om00<0000008O0nP27`?X00P27`?X2000000<0RKWe00<0000008^bl02;/_00
0P2;/_001@000000SZg[08j]j`2>[N/0000000D0TJSU00<0000009BTh02DY>002P00000209JNfP03
0000002KW]H0VikF0080VikF00<0000009nOe02OWm@00`2OWm@00`000000Y:3B0:BPdP04000000@0
ZJGB00<000000:jXd@2^Z=400`2^Z=400`000000]joB0;N_dP030;N_dP030000002l/m80_;?B0080
00000P2o]m800`000000a;_B0<BkdP0200000080akoB00<000000<c3dP3<`m800P0000020<o6d@03
0000003BbM40000000@0e<_A00<000000=K=d@0000000P3Hc]001@000000f<o@0=S?d03Hcm000000
01@0oooo00<000000?ooo`3oool07`3oool005d0oooo00H000000?ooo`3oool000000=;4c0000003
0=73c@030000003A`l`0dL?<00<0dL?<00<000000<k0c@3>`<d00`3>`<d00`000000c;o>00000003
0<Vmc`8000000P37^ll00`000000`[S@0<:hd0020<:hd00300000030]M00000000D0^[7A0P000003
0;:[d@030000002/Ym<0[:OC00<0[:OC0`0000020:BQd`030000002MW]D0WIkE0080WIkE00<00000
09JLe`2FW=L00P2FW=L010000000SY[J08jJfP2>V]XQ000000<0M[Wk0P00000207[2o@8000000`21
b_l500000080S]Ko1@00000209OMoP@00000102LhOd3000000<0WN7m0P0000000`2Jh?d000000000
000209WNoP80000000<0U=_n0000000000000`2Beoh4000000<0Qlgn00<0000008K7o@26aod00P26
aod010000000Ql3j08O0nP27`?XG00000080SJ7R00<00000096Og@2AWmd00P2AWmd00`000000UYkJ
09JNfP0209JNfP030000002KW]H0VikF00<0VikF00<0000009nOe02OWm@00P2OWm@3000000<0Y:?D
00<000000:VUdP2YYM800P2YYM8010000000[ZSA0:jXd@2^Z=42000000030;6/d`000000]joB00@0
]joB0P0000040;ngdP0300000034^m8000000080akoB00D000000<NodP000000c<?B000000050<o6
d@030000003Dbm40e<_A0080e<_A00H000000=K=d@3FcM4000000=S>d00000030=S?d0030000003I
d=0000000180oooo00<000000?ooo`000000803oool005d0oooo00<000000?ooo`0000000`3Ba<`0
0`000000dL?=000000050=73c0030000003@`/d0000000<0c/3=00L000000<bocP3<_lh0c;o>0000
0039_Ll0000000D0ak_?00@000000<:hd000000000000P30]M000`000000^[7A0;Zad@020;Zad@03
0000002h[m00000000@0/Z_A00<000000:bWd`2/Ym<00P0000000`2ZYM4000000:BQd`030:BQd`05
0000002MW]D0WIkE09fNe@0000000`2FW=L50000000309>Jf0000000SI[K00<0SI[K00D0000008NK
g`27Vml0Qi_O0000000308:Mi0030000001mX>T0OJ3Y0080OJ3Y00D0000007VUkP1iYNh0NJG^0000
000307J[l`8000000P1e/OP2000000<0M[Wk2`0000000`28col0000000000002000000<0TmWn1000
000409[OoP@000000P2LhOd4000000<0V^3m1000000309CKoPD000000P2:dol2000000<0Qlgn2@00
000307^inP050000001l/oH0O;?f07bcmP000000101o[O401@000000Q:S/08BXk024Z>`0000000@0
R:CW00<0000008fQhP2=XN800P2=XN800`000000TIoM096Og@02096Og@030000002FW]X0UYkJ0080
UYkJ00<0000009^NeP2KW]H00P2KW]H00`000000WioD09nOe0020000000309jQeP000000Y:?D00<0
Y:?D00<000000:VUdP2YYM800`2YYM83000000<0/JcC00<000000;N_dP2g[m800`2g[m800`000000
_kOB0;ngdP040;ngdP0300000037_m80akoB0080akoB00<000000<NodP0000001@3?a]400`000000
dLWB000000020=C;d@030000003FcM40e/gA0080e/gA00D000000=S?d03Hcm00f<o@000000020=W@
d0030000003oool0oooo01<0oooo00<000000?ooo`3oool07@3oool005`0oooo00<000000?ooo`00
00001@3Ba<`00`000000dL?<0=73c0020=73c08000000`3@`/d00`000000c/3=000000050<bocP03
00000037^ll0ak_?00D0ak_?00<000000<2ed030]M000`30]M0010000000^[7A0000000000020;R_
d0030000002bZm40/Z_A00<0/Z_A0P0000030:ZUd@030000002TXM<0Y:7C0080Y:7C00@0000009fN
e@2MW]D0WIkE1000000209RKe@030000002CV]P0Ti[H0080Ti[H00D0000008fJf`2=V]/0SI[K0000
000308NKg`0300000022WN@0PYgT0080PYgT00D0000007fPj@1mX>T0OJ3Y0000000407VUkP040000
001fZo<0MZ_c07J[l`P000000`23_o/00`000000Q<Om0000000600000080T=?n0P00000209?IoP80
000000<0V=_n0000000000000P2Jgoh200000080W>3m1000000309cQo@@000000`2Hg_h4000000<0
SmSo2@00000207k7oP8000000`1k`?d00`000000NkWj07^inP0207^inP030000001l/oH0O;?f0080
O;?f00D0000007n]l@1o[O40Ojga0000000408BXk00300000028Y>L0R:CW0080R:CW00D0000008fQ
hP2=XN80SJ7R00000004096Og@040000002FW]X0UYkJ09JNfPT000000`2NXMH00`000000Y:?D0:BS
e0030:BSe0030000002YYM80ZJGB0080ZJGB00<000000;6/d`2a[=<0102a[=<01`000000]joB0;N_
dP2g[m8000000;Zcd`000000102o]m800`000000`[_C000000030<NodP80000000@0bl?B0000003?
a]40clKA0P0000020=79dP030000003Dbm40000000@0e/gA00<000000=O?d@0000000P3Hcm001@00
0000fM3@0=W@d03Id=00000001@0oooo00<000000?ooo`3oool07@3oool005`0oooo0P0000050=;4
c0060000003Ba<`000000000003A`l`000001P3@`/d00`000000c;o>0<bocP030<bocP030000003;
_/h0000000D0ak_?00<000000<Bic`0000001@30]M000`000000^:o@0;R_d0030;R_d0030000002b
Zm40/Z_A0080000000<0/:S@0000002ZYM400`2ZYM400`000000Y:7C0:BQd`020:BQd`<000000P2N
WM<00`000000V9_E09RKe@0209RKe@050000002CV]P0Ti[H09>Jf0000000102=V]/01@000000Qi_O
08NKg`27Vml0000000<0PYgT3@00000208>`lP0400000022^?L0P[Sg08:hm`8000000`23_o/40000
00<0Slcm0P000003093CoP<00000102Hfoh3000000<0W>3m0P0000000`2Ph?`000000000000209cQ
o@8000000P2Mh?d4000000<0U]cn1000000208_Do`8000000`23c_l^00000080TYkL00<0000009RO
f@2HWmT00P2HWmT00`000000WZ7F09jQeP0309jQeP030000002TXm@0Y:?D0080Y:?D00<000000:VU
dP2YYM800P0000000`2[Z=@000000;6/d`040;6/d`040000002g[m80000000000080^[?C00D00000
0;ngdP2o]m80_kOB000000020<:kd`0400000037_m80akoB0<NodP80000000@0bl?B0000003?a]40
00001@3AbM800`000000e/gA0=K=d@020=K=d@060000003Gcm40eloA0000003Hcm0000000`3Id=00
0`000000fM7A0000000D0?ooo`030000003oool0oooo01`0oooo001>0?ooo`050?oOW01800000000
0000M02oool02@3oool00`000000d/C<0=;4c0020=;4c0800000103Ba<`00`000000d<;=0=32c@04
0=32c@030000003?`/d0000000<0c;o>00@000000<^ncP3;_/h000000`37^ll200000080a;W?00<0
00000<2ed030]M000P30]M000`000000_K?@000000050;R_d08000000`2`Z=000`000000ZZGA0:ZU
d@020:ZUd@030000002TXM<0Y:7C0080000000<0XioA0000002NWM<00P2NWM<00`000000V9_E09RK
e@0209RKe@030000002CV]P0Ti[H0080Ti[H00<0000008fJf`2=V]/02@00000208bQhP0500000028
YNL0R:GW08RUi`0000001025Z^d00`000000Pk3b08>`lP0208>`lP0400000022^?L0P[Sg08:hm`@0
00000P2?aOX2000000<0Slcm1@00000309SFo@@000000`2NgOd3000000@0X>3l0`00000309gPo@80
000000<0W]om0000000000000P2Fg?h20000000309GJoP000000RmCo0080RmCo1@00000207g9o`80
00000`1e`Oh010000000L;Wl072io01`^O`2000000<0K[;i00<0000006n/mP1_[?H00P1_[?H01@00
0000LjOb07>WlP1cYo80000000<0N:?^00<0000007jQj@1nXNT00P1nXNT00`000000QIoT08FOi002
08FOi0050000002<W^00S9kP08bNh00000001@2BW]`00`000000V9oI09ROf@0209ROf@030000002N
XMH0WZ7F0080WZ7F00@000000:BSe02TXm@0Y:?D100000030:^Xe0030000002a[=<0/JcC00<0/JcC
0P0000050;Zcd`030000002o]m80000000@0`[_C00<000000<NodP0000000P3;`m800`000000bl?B
000000050=79dP030000003Cc=8000000080e/gA00<000000=O?d@3Gcm400P3Gcm401@000000fM3@
0=W@d03Id=0000000080fM7A00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0K0?ooo`00
C@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0203oool3000000030=;4c0000000d/C<
00@0d/C<00<000000=;4c@0000000`3@`/d2000000<0cl;=00<000000<bocP000000103;_/h00`00
0000ak_?000000050<Bic`0400000030]M00000000000080_K?@00<000000;R_d02h[m000P2h[m00
0`000000]Jc@000000040;2Xd0040000002ZYM40ZZGA0:ZUd@<000000`2SWm400`000000WYgC09jM
d`0209jMd`030000002HVmD0V9_E0080V9_E00<0000009>Jf02CV]P01000000209>Lf@030000002?
WMd0SigM0080SigM00D0000008bQhP2<XN80S:7R0000000408RUi`0500000025Z^d0QJ[]08FZk@00
00000`23/?860000000308nmmP000000000000<0SlGj1000000309W@n`030000002He_d0V=Km0080
00000P2PfO/00`000000W]gm09kMo@02000000<0Y=kk100000020:?Po0@000000`2Ngod4000000<0
UM[n1000000308S@o`040000001mbOl0OLWo07g9o`d0000000<0K[;i0000001_[?H00`1_[?H00`00
0000LjOb07>WlP0207>WlP050000001hXnh0N:?^07RSkP000000101nXNT01@000000QIoT08FOi025
Wn@0000000@0S9kP00<0000009:Ng02BW]`00P2BW]`00`000000V9oI09ROf@0209ROf@040000002N
XMH0WZ7F09jQePD000000P2TYMH00`000000ZjSD0:^Xe0030:^Xe0030000002a[=<0/JcC00800000
00<0/joD0000002j/m<01@2j/m<00`000000`[_C0<:kd`040<:kd`040000003;`m80bl?B0<_3dP80
000000<0cLKC0000003AbM800P3AbM801P000000dlcB0=?<dP000000e/gA000000@0eloA00<00000
0=SAdP0000000P3Id=001@000000fM7A0=WAd@3IdM40000001@0oooo00<000000?ooo`3oool06`3o
ool004d0oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`040000003oool0
oooo000000D0d/C<00L000000=;4c@3Ba<d0d/C=0000003@`/d0000000H0cl;=00<000000<^ncP3;
_/h0103;_/h00`000000a;W?0<Bic`050<Bic`030000002m/m00_K?@00<0_K?@00@000000;R_d000
000000000P2e[=000`000000/:S@0;2Xd0030;2Xd0<000000P2XX]000`000000XioA0:>Od@020:>O
d@050000002NWM<0WYgC09jMd`0000000`2HVmD400000080Ui_F00<0000009>Lf@2CW=T00P2CW=T0
1@000000SigM08nMg@2?WMd0000000@0S:7R00@0000008RUi`28YNL0R:GW3000000308nmmPD00000
0P2JbOP2000000<0VM3k0`0000040:3In`<000000`2Tg_/200000080ZMoj100000030:OOn`@00000
0`2PgO`4000000<0UMOn2000000307S4oP040000001]^_d0KK[m06fjo@8000000`1V/_/?00000003
07RSkP000000OZ7Y00<0OZ7Y00<0000008FOi025Wn@00P25Wn@01@000000S9kP08bNh02<W^000000
00D0TYkL00<0000009ROf@2HWmT01@00000209fSf@030000002TYMH0Y:GF00<0Y:GF00<000000:^X
e02[Z=@00P2[Z=@3000000@0/joD00@000000;Zcd`2j/m<0^[?C0P0000000`2l]mD000000<:kd`02
0<:kd`80000000@0aKoD0000003;`m80bl?B0P0000020<g6d`030000003AbM80000000@0dlcB00<0
00000=O?d@3Gcm400P3Gcm401P000000f=7B0=SAdP000000fM3@000000<0fM7A00<000000=SBd`00
0000503oool00`000000oooo0?ooo`0J0?ooo`00C03oool00`3ookl0M00004RLg`030?ooo`030=nL
B00007@0_ooo00L0oooo0P0000020?ooo`030000003Ba<`0d/C<0080d/C<00<000000=;4c@3Ba<d0
0P3Ba<d00`000000cl;=0<o2c@040<o2c@030000003=`Lh0000000@0bkk>00<000000<Rlc`000000
1034^Ll3000000H0_K?@00<000000;F/d02e[=000`2e[=001`000000/:S@0;2Xd02`Z=0000000:jV
d00000000`2XX]000`000000XioA0:>Od@020:>Od@040000002NWM<0WYgC09jMd`@000000P2KVm<0
1@000000Ui_F09NKeP2GVmH0000000@0TicI00<0000008nMg@2?WMd00P2?WMd010000000S:7R08bQ
hP2<XN8400000080TjWU00<00000096_j`2A[n/00`000003092fl0@000000`2K`_<2000000<0V/Wh
100000030:;Cn@@000000`2Xf_T3000000<0ZMoj0P0000000`2^g_P00000000000020:OOn`800000
0P2Zg_T4000000<0X]_k1000000209GCo@0400000026c?h0Q/cn08K<oPd0000000<0I[;k0000001R
ZoP00P1RZoP2000000<0HjGe00H0000006NQlP1WXO80Ij7b0000001]W^hH00000080UZ3K00<00000
09fSf@2MXmT00`2MXmT00`000000Y:GF0:BUeP020:BUeP030000002[Z=@0ZjSD0080000000<0[:cG
0000002c[m@0102c[m@010000000^[?C0000000000030;bge@0400000032^m<00000000000<0aKoD
00<000000<_3dP0000000P3=a]<010000000cLKC0<g6d`0000040=?<dP030000003Ecm<000000080
eloA00<000000=SAdP3HdM800P3HdM801@000000fM7A0=WAd@3IdM4000000080f=;C00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`0I0?ooo`005`3oool01@3OW4P00000000000000000M;oo
01h0oooo00D0gia8000000000000000007Boo`0=0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol01`3oool00`000000oooo0?ooo`020?ooo`80000000<0d/C<0000003Ba<d00`3Ba<d0
0`000000dLC>000000030<o2c@8000000`3=`Lh010000000bkk>0000000000030<Rlc`0400000034
^Ll00000000000<0`[O@00@000000;fcd02m/m00_K?@0`0000050;F/d0040000002`Z=0000000000
0080[ZK@00<000000:RRd02XX]000P2XX]0010000000XioA0:>Od@2SWm4400000080WicA00D00000
09^Kd`2KVm<0Vi_C0000000409NKeP040000002CW=T0TicI09>Lf@T000000P2DY=l010000000TjWU
09>Yi@2CZND300000080TJo[1000000209^jkP030000002K`_<0Vl;c00<0Vl;c100000020:C<m@80
00000P2RdoT2000000030:_EmP00000000000080Z=[i0P0000020:kKm`<00000102^g_P300000080
Z]ki0P0000020:gLn0030000002Rfo/0X]_k008000000P2Sf?X010000000UM?m09GCo@2Edod40000
00<0QLSm00@0000007Bno@1d__d0M;km0P00000306Ndn`X0000000<0HjGe0000001WXO800`1WXO80
1@000000KIk^06fNkP1]W^h0000000<0MIgZ00<0000007fMiP1mWNH00P1mWNH00`000000QYgS08JM
h`0208JMh`030000002>W]l0SYkO0080SYkO00<0000009JPf`2FX=/00P2FX=/00`000000WJ?I09fS
f@0209fSf@030000002TYMH0Y:GF0080Y:GF0`0000040:b/e`030000002c[m@0/joD0080/joD0P00
00060;bge@0300000035_m@0aKoD00@0aKoD00<000000<g6d`3=a]<00P3=a]<2000000050=3:e000
0000dlcB0=?<dP0000000P3Ecm<00`000000eloA000000040=SAdP050000003Idm<000000=WAd@00
0000103Hd]<00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01T0oooo000G0?ooo`030=nL
B00007@0_ooo00l0oooo0100_g@004RLg`3oool0oooo0;md00000000B9c409a800000000M;oO09a8
00000000M;oo0=nLB000000007Bo0`3oool00`3ookl0M00007Boo`0=0?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol01P3oool00`000000oooo0?ooo`050?ooo`030000003Ba<d0d/C=
0080d/C=00L000000=74cP3Aa<h0dLC>0000003?`/d0000000H0cL7>00<000000<Rlc`38_<l01038
_<l00`000000`[O@0<:gd0030<:gd0040000002m/m000000000000<0^[3@00<000000;F/d02e[=00
0P2e[=02000000D0[ZK@00<000000:RRd02XX]000P2XX]0300000080Xik@00D0000009nLd@2OW=40
WicA0000000409^Kd`040000002GVmH0Ui_F09NKeP@000000P2IWML01@000000Uj3K09NPf`2GX=/0
000000D0U:CO0`0000000`2CZND000000000000200000080W;?X00<0000009^jkP2K^^h00P2K^^h0
0`000000Vl;c09_2l`0200000080YLG`00<000000:C<m@2Tc?D00P2Tc?D3000000@0ZmGf0`000002
0:kKm`8000000P2dfo@00`000000[]kh0:kNn00200000080/]gf0`0000050:gLn0800000102Sf?X4
00000080ULol0P00000308G8o@l000000`1HXoH60000000306NQlP000000KIk^00<0KIk^00D00000
07FMjP1eWNX0MIgZ0000000407fMiP0500000026WN<0QYgS08JMh`000000102>W]l00`000000UZ3K
09JPf`0209JPf`030000002MXmT0WJ?I00<0WJ?I00H000000:BUeP00000000000:FXf@0000050:b/
e`030000002c[m@0/joD0080000000<0][CG0000002l]mD0102l]mD2000000D0aKoD0P0000040<g6
d`060000003@b]@0d<[D0000003Cc=800000103Ecm<010000000f=7B0=SAdP3HdM82000000<0fM?C
00<000000=SBd`3Hd]<00P3Hd]<00`000000em?D0000000D0?ooo`030000003oool0oooo00d0oooo
00D0omnL04P000000000000007Boo`060?ooo`005`3oool00`3OW4P004RL0=ooo`0?0?ooo`0@0?oo
_`1dB9`0gooo0?oOW01807@0_ooo0?ooo`3ogi`0B01d0;ooo`3ogi`0B01d0;ooo`3ogi`0B01d0;oo
o`<0oooo00<0oono07@0001d_ol03@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00H0oooo00<000000?ooo`3oool01P3oool01@000000d/C=0=;4c@3Ba<d0000000@0dLC>00<00000
0<g1cP3=`Lh0103=`Lh00`000000bl3@000000040<Rlc`0300000035^m00000000D0`[O@00<00000
0;Z`d02j/=000`2j/=000`000000]Jc@0;F/d002000000030;:Zd0000000[ZK@00@0[ZK@00<00000
0:RRd02XX]000`0000040:>Nd0030000002OW=40WicA0080WicA00@0000009^Kd`2KVm<0Vi_C1000
000209^Le0050000002IWML0VIgG09VMe`000000102GX=/00`000000U:CO0000000600000080W:cR
00<0000009bcj02L/nP00P2L/nP010000000Vk[^09^jkP2K^^h3000000@0YLG`00<000000:C<m@2T
c?D00P0000020:g>l@030000002[eOH0ZmGf008000000P2be_<3000000D0]=_d0P0000030;;MmP80
000000<0]]_c0000000000000P2]g?P200000080[m[f100000020:GDn0040000002Eco`0ULol09G?
o0D000000P24`o`2000000<0L[Wl00H0000006>_nP1S[oX0Hjoj0000001GYOL40000000305RSmP00
0000F9kc0080F9kc00<0000005^Kl@000000200000000`1eWNX0000007fMiP0307fMiP0300000026
WN<0QYgS0080QYgS00<0000008jNg`2>W]l00P2>W]l00`000000UZ3K09JPf`0209JPf`030000002M
XmT0WJ?I00@00000102UZ=T00`000000[:cG0:b/e`020:b/e`<00000102f]=L00`000000_;OE0;bg
e@0200000080_kcG00<000000<Foe035_m@00P0000020<S4eP060000003=a]<0cLKC0000003@b]@0
00000P3@b]@00`000000eLoC0=G?d`020=G?d`050000003Fd]@000000=SAdP000000103Idm<00`00
0000f=GE000000020=SBd`050000003Gdm@0em?D0=OCe00000004P3oool00`000000oooo0000000=
0?ooo`070?oOW0180000B9cO0?ooo`3ookl0M00007Boo`050?ooo`005`3oool00`3OW4P004RL0=oo
o`0@0?ooo`040?noM01d_ol0okmd001d_`80oooo00T0omnL04P0M02oool0omnL04P0M02oool0omnL
04P0M02oool00`3oool00`3ookl0M00007Boo`0=0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol01P3oool2000000P0oooo00<000000=;4c@000000103Aa<h00`000000d<C?00000004
0<g1cP040000003;`=00bl3@000000<0b;c?00@000000<Fkd035^m0000001032]m000`000000_[GA
000000050;Z`d0800000102bZ]000`000000[ZK@0:jVd0020:jVd0<000000`2WXLl00`000000Xik@
0:>Nd0020:>Nd0030000002OW=40WicA0080WicA0`00000209fKd@050000002KW=@0VicD09^Le000
0000102IWML010000000Uj3K09NPf`2GX=/20000000509BTg`000000000009fWg@000000102L[>80
0`000000W;?X09bcj00309bcj0<000000P2U_N/2000000<0YLG`0`0000050:g>l@800000102be_<0
10000000^=K_0000000000020;CKm08000000P2if_43000000@0]]_c0`0000020:oJmP8000000P2a
eo<010000000YMCh0:GDn02Ue?P4000000<0U/_j00@0000008C3o024`o`0Q<?l200000000`1S[oX0
000005NUm`0305NUm`H0000000<0F9kc0000001KVo400P1KVo401P000000HYW^06:IkP1RVNh00000
06^IjaP000000`2LYM`00`000000YJSI0:FXf@020:FXf@030000002/[=L0[:cG0080000000<0[[3I
0000002f]=L0102f]=L00`000000_;OE000000040;nle`0300000035_m@0000000@0b<CF00<00000
0<g6d`0000000`3@b]@2000000050=;>eP000000eLoC0=G?d`0000000`3Fd]@00`000000fM?C0=WC
d`020=WCd`060000003HeMD0f=GE0000003Hd]<000000`3Gdm@00`000000eM?G0000000C0?ooo`03
0000003oool0oooo00X0oooo00<0oooO09a80000M;l00`3oool00`3o_g@004RL0=ooo`040?ooo`00
0P3oool00`3ooml0W4P000000002000000050000B02Lgol0oooo0?ooo`3o_g@00P000000100007@0
_ooo0?ooo`2oM002000000060000M02oool0oooo0=nLB000B9`0gooo1@3oool00`3ooml0W4P00000
0002000000030000B02Lgol0oooo00@0oooo00<0_gAd09bLB000M;l00P3oool02@3ogi`0B01d0;oo
o`3ogi`0B01d0;ooo`3ogi`0B01d0;ooo`030?ooo`030?oo_`1d0000M;oo00d0oooo00<0oono07@0
0018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`030000003oool0oooo00P0oooo00<000000=74
cP3Aa<h00P3Aa<h010000000d<C?0=34c`3@a<l2000000030<g1cP000000bl3@00<0bl3@00<00000
0<Rlc`0000001035^m0010000000`[O@0000000000030;jed@030000002j/=00^[3@0080000000<0
]jkA0000002bZ]000`2bZ]000`000000[ZK@0:jVd002000000030:^Tc`000000Yj7?00<0Yj7?00<0
00000:>Nd02SW]000P2SW]000`000000WicA09nLd@03000000@0WI_A00<0000009^Le02KW=@00P2K
W=@010000000VIgG09VMe`2IWML7000000@0WJOM00<0000009b/hP2L[>800P2L[>800`000000W;?X
09bcj002000000030:Ffi@000000YKg[00@0YKg[0`0000020:k7k@8000000P2]c_42000000040;C@
kP00000000000;;Fl`800000102he^l2000000@0^M[a00<000000;gIkP0000000P2ffo<200000080
^MW`0`0000050;7Gl`<000000P2Vd?H2000000<0U/_j1000000308BonP040000001a]OX0LKGj076e
nPX0000000<0Cigd0000001=Uo80100000000`1KVo40000006:IkP0306:IkP050000001[VN/0JiW[
06^Ij`0000000`1eVnP00`000000OigU07nMi@0207nMi@0300000029Wn80RIoR0080RIoR00<00000
09>Rg`2CX]l00P2CX]l00`000000W:GL09bUg00209bUg0030000002UZ=T0YJSI00<0YJSI0P000004
0:j`f@030000002f]=L0][CG0080][CG0P0000060;nle`0300000038a=H0b<CF00@0b<CF00<00000
0=3:e03@b]@00P3@b]@2000000040=;>eP000000eLoC000000<0e];D00<000000=OEeP0000000P3I
dm<00`000000f=GE0=SEe@020=SEe@050000003Gdm@0em?D0=OCe00000000P3EdmL00`000000oooo
0?ooo`0B0?ooo`030000003oool0oooo00T0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`040?ooo`00103oool00`3ogi`0B00004RLg`030?ooo`040?oo_`1d0000001809cOo`80oooo
00X0_g@000000000M;l0oooo0?oOW01807@0_ooo0=nLB000B9`0gooo1`3oool00`3ogi`0B00004RL
g`070?ooo`0>0?oo_`1d0000B9cO0?ooo`3ookl0M00000000018W=l0omnL04P00018W=l0omnL04P0
M02oool30?ooo`030?oo_`1d0000M;oo00h0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo
0?ooo`060?ooo`030000003oool0oooo00T0oooo00D000000=74cP3Aa<h0dLC>000000050=34c`03
0000003;`=00bl3@00@0bl3@00<000000<Fkd035^m001035^m000`000000_[GA0;jed@040;jed@80
00000`2g[]400`000000/Z[@0;:Zd0030;:Zd0800000102[Y<l00`000000Yj7?0:NQc`020:NQc`03
0000002SW]00Xik@00@000000`2PW=000`000000WI_A09fKd@0209fKd@060000002KW=@0VicD09^L
e0000000VIgG1P00000209fSf@030000002MYmd0WJOM0080WJOM00@0000009b/hP2L[>80W:cR1000
00040:Ffi@030000002U_N/0YKg[0080000000<0[/3W0000002^and0102^and2000000D0]=3^0P00
00030;SFk`80000000@0_MK/000000000000^M[a0P0000030;gIkP800000102ifO0010000000^mK]
0000000000020;7Gl`80000000<0/m?a0000002Vd?H00P2Vd?H500000080UlOh0P00000308BonPX0
00000`1BWoD60000000304fGlP000000000000@0000000<0HYW^0000001[VN/00`1[VN/01@000000
MI_X07FKj01eVnP0000000@0OigU00<0000008VOhP29Wn800P29Wn801@000000Tj;O09>Rg`2CX]l0
000000D0W:GL00<000000:FXf@2UZ=T00P0000000`2U[=`000000:j`f@030:j`f@030000002f]=L0
][CG008000000P2h^MT010000000_kcG0;nle`2o_=L2000000030<;1f@000000b<CF0080b<CF0P00
00001`3;b=P000000=3:e03@b]@000000=;>eP0000000P3Bc]H01@000000e];D0=KBe03Fd]@00000
0080emGF00<000000=WCd`000000103HeMD200000080em?D00<000000=GCe`3EdmL00P3EdmL00`00
0000oooo0?ooo`0B0?ooo`030000003oool0oooo00P0oooo00<0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`040?ooo`00103oool00`3ookl0M00004RLg`020?ooo`040?oog`2LB00004RL0=oo
o`80oooo00<0omnL04P00018W=l00`3oool01@3ogi`0M9cO0=nLB000B9`0gooo00L0oooo00<0oono
07@00018W=l0203oool0302oM000B9cO0?ooo`3oool0okmd0018W02oW4P0000004RLg`3OW4P00000
04RLg`@0oooo00<0oono07@0001d_ol03P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0
1`3oool00`000000oooo0?ooo`0:0?ooo`030000003Aa<h0000000D0d<C?00<000000<k4d@000000
103;`=000`000000bKoA000000040<Fkd00300000032^]80000000@0_[GA00<000000;^cdP000000
102g[]400`000000/Z[@0;:Zd002000000030:nWd0000000ZjC?00<0ZjC?00<000000:NQc`2WXLl0
0P2WXLl300000080Xig?00<000000:2Ld02PW=000P2PW=000`000000WI_A09fKd@0209fKd@030000
002KW=@0VicD00D000000P2NWmD00`000000WJ?I09fSf@0209fSf@040000002MYmd0WJOM09fWg@@0
00000`2T[ml00`000000YKKU0:Ffi@020:Ffi@<00000102^`>L00`000000[/O]0:k7k@0200000003
0;G9j@00000000000080]=3^0P0000020;_@jP<00000102me^`2000000<0_MW^0P0000001030en/0
00000000002ifO02000000<0^mK]0`0000040;?Cl@<000000`2Wc?<010000000UlOh09O7n02GaoP5
00000080QK[h00H0000007:`n01b/?P0L[3h0000001OY_H40000000305:Om@000000BIOb0080BIOb
00<0000004FBl00000000`0000000`1@UO00000005NDkP090000000307FKj0000000OigU00<0OigU
00<0000008VOhP29Wn800P29Wn800`000000Tj;O09>Rg`0209>Rg`040000002LYM`0W:GL09bUg0@0
00000`2U[=`00`000000[[3I0:j`f@030:j`f@800000102h^MT010000000_kcG0000000000030<;1
f@0400000038a=H00000000000<0blSH00<000000=3:e00000000`3Bc]H2000000050=?Bf0000000
e];D0=KBe00000000`3GeMH010000000f=GE0=SEe@3HeMD200000080f=OG0P0000040=GCe`030000
003Ce=X0000001<0oooo00<000000?ooo`3oool0203oool00`3ogi`0B00007Boo`030?ooo`030;md
000004P0W=oo00@0oooo00040?ooo`030?oo_`1d0000B9cO0080oooo00<0gia80000B02Lgol00`3o
ool00`3o_g@0001809cOo`050?ooo`030=nLB000B9`0gooo00L0oooo00<0oono07@00018W=l01`3o
ool00`3ooml0W4P004RLg`0>0?ooo`030?oo_`1d0000M;oo00l0oooo00D0oono07@000000000001d
0;ooo`070?ooo`030000003oool0000000`0oooo00<000000=34c`3@a<l00P3@a<l2000000<0c/CA
00@000000<_0d000000000000`39_m4010000000aK_@0000000000030<:jdP040000002n]M400000
00000080^k?B00<000000;N^d@2g[]400`2g[]42000000<0[jO@00<000000:^Tc`2[Y<l00`2[Y<l0
1P000000Yj7?000000000000YYo?000000@0Xig?00<000000:2Ld02PW=000P2PW=000`000000WI_A
09fKd@050000000309nMdP000000WYoE00<0WYoE00<0000009fSf@2MXmT00`2MXmT300000080Y:[K
00<000000:B_g`2T[ml00P2T[ml00`000000YKKU0:Ffi@0200000080[KWR00<000000:k0i`2^`>L0
0P2^`>L3000000@0]LWY0`0000030;_@jP80000000@0`=3V0000002me^`0_MK/0P0000020<;Ej0<0
00001@30en/010000000^mK]0;_Fk@2ke^d200000080_M?[00<000000;?Cl@2cdo400P0000000`2d
cnh00000000000030:O<l`@000000`2H`_D01@000000QK[h08Fjn025^_P0000000@0L[3h00@00000
05nVmP1OY_H0GjKf1P0000000`19Uo80000004FBl00204FBl0@000000`1GU>h01P000000HIK/066F
k01QU^`0000006bIjQ@000000P2KZ=l00`000000YJcL0:F/g0030:F/g0030000002^/=T0[[3I0080
000000<0/;GL0000002h^MT0102h^MT00`000000`/7I0<;1f@040<;1f@030000003;b=P0blSH00@0
blSH00@000000=;>eP3Bc]H0d/kF0`000000103Cd]P000000=KBe00000030=OEeP050000003Gf=P0
00000=SEe@000000103HemL010000000eM?G0=GCe`3EdmL200000080dmCJ00<000000?ooo`3oool0
4P3oool00`000000oooo0?ooo`070?ooo`080?oo_`1d0000000004RLg`3oool0okmd00000000M;l5
0?ooo`00103oool01`3ookl0M00004RLg`3oool0okmd0000001d_ol0103oool00`3o_g@0001d0;oo
o`050?ooo`030=nLB000B9`0gooo00L0oooo00<0oono07@00018W=l01`3oool00`3ogi`0B01809cO
o`0>0?ooo`030?oo_`1d0000M;oo01/0oooo0P00000>0?ooo`80000000<0d<C?0000003>a=401@3>
a=400`000000bKoA0<Vod@040<Vod@0300000032^]80`[[B00@0`[[B00<000000;^cdP2k/m800`2k
/m800`000000]jkA0;N^d@02000000030;>/d@000000[jO@00<0[jO@00<000000:^Tc`2[Y<l00P2[
Y<l2000000<0YYo?00<000000:>Mc`2SWLl00P2SWLl00`000000X9c@0:2Ld0020:2Ld0@0000000<0
X9c@0000002OWM800`2OWM800`000000WYoE09jOe@0209jOe@030000002MXmT0WJ?I0080000000<0
XjGG0000002TZ]/00P2TZ]/00`000000Y:oO0:B_g`030:B_g`800000102]^N800`000000[/3W0:k0
i`0200000080]L;T00<000000;G9j@2ebNT00P0000020;c:i@040000002kd>X00000000000@0`=3V
0P0000040<;Ej00400000035dnD0000000000080`=O[0P0000001033e>L000000000000000050;gC
j`8000001@2dcnh300000080YlS`0P00000309S2mA00000000<0Ci_c00000014To00100000001015
T_00000004N?kP17Snh30000000305NDkP000000HIK/00<0HIK/00D0000006bIjP1/VNX0K9WZ0000
000307VLj00300000025X>D0QJ3U0080QJ3U00<00000096ThP2AY>800P2AY>800`000000VjSO09^X
g`0309^Xg`030000002U[=`0YJcL0080YJcL0`0000030;2eg0030000002h^MT0^;WI0080^;WI00<0
00000;Zng00000001032`MT00`000000a<KK000000030<_8f080000000@0cLgJ0000003Bc]H00000
0`3Cd]P00`000000dm;H000000030=OEeP050000003Gf=P0emSH0=OHf0000000103HemL01@000000
e]WJ0000003EdmL0000000D0dmCJ00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`070?oo
o`030;md0000B9`0_ia80080000000<004RL0=ooo`3oool0103oool000@0oooo00<0oono07@00000
00000P0000000`0007@0_ooo0?ooo`030?ooo`030?oOW0180000000000@0000000@0B9cO0=nLB000
B9`0gooo1`3oool0102oM00000000018W03Oool50?ooo`030;md00000000B9cO00l0oooo00<0oono
07@0001d_ol06`3oool00`000000oooo0?ooo`0?0?ooo`030000003>a=40c/CA00@0c/CA00<00000
0<c4d`0000001039_m400`000000a[oC000000040<:jdP030000002o^=@0000000D0^k?B0P000003
0;>/d@030000002_Ym00[jO@0080[jO@00<000000:^Tc`2[Y<l00P0000000`2YX/l000000:JOc`03
0:JOc`030000002SWLl0Xig?0080Xig?00<000000:2Ld00000000`0000040:2Ld0050000002OWM80
WigB09nMdP0000000`2NWmD4000000<0XjGG00<000000:BZf`2TZ]/00P2TZ]/00`000000Y:oO0:B_
g`02000000030:bbg@000000[KWR00@0[KWR0P0000050;G2i08000001@2lb^D00`000000`=3V0<3@
iP030<3@iP0500000034d><0000000000032eNP00P0000040<GCi@8000000`33e>L010000000a=;U
0000000000020;gCj`80000000<0_]3X0000000000000P2dcnh2000000030;C;j`000000YlS`0080
YlS`1@00000209VnlP0500000026]_D0Q[Ke08Jfm@000000101c[OH010000000H:;d062Rm01PX_@5
00000080A9?`1@00000204N?kPH0000000<0HIK/0000001/VNX00P1/VNX00`000000NIcX07VLj002
07VLj00300000025X>D0QJ3U0080QJ3U00<00000096ThP2AY>800P2AY>800`000000VjSO09^Xg`02
09^Xg`030000002U[=`0YJcL0080000000<0Y[3P0000002`]M`0102`]M`010000000^;WI00000000
00020;Zng00500000032`MT0`/7I0<;1f@0000000P34a]/010000000blSH0000000000040<g=fP03
0000003Cd]P0dm;H0080dm;H0P000000103De]X000000=OEeP0000030=OHf0030000003Gfm/00000
0080f=OG00D000000=KIfP3FfMX0e]WJ000000050=?DfP030000003?e=d0000001<0oooo00<00000
0?ooo`3oool01`3oool00`3ogi`0B00007Boo`090?ooo`00103oool00`3ookl0M00004RLg`020?oo
o`030;md0000000007Bo00@0oooo00X0_g@004RLg`3oool0oooo0=nLB00004P0W=oo0=nLB000B9`0
gooo1P3oool01@3o_g@00000001d_`3OW4P0B9cO01L0oooo00<0oono07@0001d_ol06`3oool00`00
0000oooo0?ooo`0@0?ooo`040000003>a=40c/CA0<k4d@8000000`3<a=<010000000bKoA00000000
00030<Jod`0400000032^]80000000000080_kSD00<000000;^cdP2k/m800P2k/m800`000000]k7C
000000040;>/d@040000002_Ym00[jO@0:nWd0<000000`2YX/l00`000000YYo?0:JOc`020:JOc`04
0000002SWLl0Xig?0:>Mc`<000000`2QW<l00`000000X9c@0:2Ld0020:2Ld0030000002OWM80WigB
0080WigB0`0000020:>Rd`030000002SYML0XjGG00<0XjGG00@000000:BZf`2TZ]/0Y:[K0`000004
0:bbg@030000002]^N80[KWR0080000000<0];_O0000002e`^@00`2e`^@00`000000^l?P00000004
0;c:i@0500000031b^80000000000030d>H00P0000040<C@h`0300000035dnD0aM?U0080aM?U0P00
00000`37dN8000000<?Di`02000000@0a=;U0P0000050;k@j0800000102dbn/3000000<0Z<?]00D0
000009VnlP2I__80VKkb0000000308Jfm@P0000000<0H:;d0000001?V?800P1?V?84000000<0?hg^
1000000304f@k@030000001HTnd0000000@0000000<0K9WZ0000001iW>P00P1iW>P00`000000QJ3U
08FPi@0208FPi@030000002AY>80TJCR0080TJCR00@0000009^Xg`2KZ=l0VjSO100000040:J`h003
0000002`]M`0/;GL0080/;GL0P0000050;Zng00300000032`MT0000000@0a<KK00<000000<g=fP3=
cMX00`3=cMX00`000000cm;L000000020=?Bf0<000000P3De]X01@000000emSH0=OHf03Gf=P00000
0080em_K00<000000=SGe`0000000`3FfMX00`000000dm[M000000020=?DfP8000000P3?e=d00`00
0000oooo0?ooo`0B0?ooo`030000003oool0oooo00L0oooo00<0gia80000B02Lgol0203oool000@0
oooo00<0oono07@00018W=l00`3oool00`2oM000000007Boo`040?ooo`090=nLB000000000000000
B02Lgol0oooo0=nLB000B9`0gooo00D0oooo00L0oooO09a8000007@0_ooo0?oo_`1d07@0_ooo01H0
oooo00<0oono07@0001d_ol06`3oool200000180oooo00<000000<k4d@0000001P3<a=<00`000000
a[oC0<Jod`040<Jod`030000002o^=@0_kSD00<0_kSD00@000000;^cdP00000000000P2g/M<00`00
0000/jcA0;>/d@030;>/d@<000000P2/Y]400`000000ZJ;?0:VRc`020:VRc`030000002VWll0YYo?
0080YYo?0`0000020:>Lc`030000002QW<l0XIc?0080XIc?00<000000:2Ld02PW=000P2PW=001P00
0000WigB000000000000XioA000000@0Xj;C00<000000:>Ue`2SYML00P2SYML400000080ZjgI00<0
00000:bbg@2//]d00`2//]d2000000@0];_O00@000000;G2i000000000000`2k`n0010000000_<[U
0000000000040<7:hP0300000034d><0a=3S0080a=3S0P0000000`37cn0000000<GCi@02000000@0
am7R00<000000<CBi@34d^D00`34d^D00`000000aLoR000000030;k@j080000000@0_/cU0000002d
bn/0]<_[0P0000000`2danP000000:S3k@040:S3k@L000000P28/_800`000000MJ[c07FZl`0207FZ
l`H0000000<0CiSb00000013T>l0100000000`0oSNh0000003n;k@040000000304f@k@000000F9?]
0080F9?]00<0000006BGk00000004P00000309Z/h`030000002V/>00Y[3P0080Y[3P00<000000;2e
g02`]M`00P0000020;6jh0030000002j_]`0^[kL0080^[kL0P0000040<C6f`0300000036c=h00000
0080cLgJ0P0000020<oBg0030000003Cd]P0000000<0e=KJ0P0000001@3Df]`000000=OHf03Gf=P0
000000<0em_K00D000000=KIfP3FfMX0e]WJ000000020=?Jg@030000003Ce=X0000000D0cmCM00<0
00000?ooo`3oool04@3oool2000000T0oooo00<0okmd00000000M;l01`3oool000@0oooo00<0oono
07@00018W=l00`3oool00`3o_g@00000001d_`0:0?ooo`030=nLB000B9`0gooo00D0oooo00<0gia8
00000018W=l00P3oool00`3o_g@007Bo0?ooo`0E0?ooo`030?oo_`1d0000M;oo01/0oooo00<00000
0?ooo`3oool04P3oool00`000000c<CC0<c4d`040<c4d`0300000039a=H0000000<0a[oC0P000000
0`32_]H000000;nhe0040;nhe0030000002g/M<0]k7C00<0]k7C00L000000;>/d@2c[=40/jcA0000
002_Z]<0000000<0[:KA00<000000:VRc`2YX/l00P2YX/l00`000000YYo?0:JOc`04000000<0Xic?
00<000000:6Lc`2QW<l00P2QW<l010000000X9c@0:2Ld02PW=03000000<0XioA00<000000:>Rd`2S
X]<00P2SX]<00`000000XjGG0:>Ue`02000000030:VXe@000000ZjgI00@0ZjgI00<000000:bbg@2/
/]d00P0000000`2b]=X000000;Bkg`040;Bkg`030000002k`n00^l?P00<0^l?P0P0000040<7:hP80
000000<0aL[N00000034d><00P0000040<O?h00300000037dN80am7R00<0am7R00D000000<W?g`00
000000000<CBi@0200000080aLoR00@000000;k@j00000000000102nc>D2000000@0]<OX00<00000
0:S3k@2X`nd00P0000000`2X_nX0000009VjkP0209VjkP0300000028/_80R;;b0080R;;b00D00000
07FZl`1eZ_<0MJ[c0000000306:PlP030000001AU_00000000<0000000<0@i3_0000000kR^d01000
00000`0oRnd0000004F<k@040000000305RCk@000000I9O/00<0I9O/00D0000007:Lj`1bW>/0LYc[
00000003082Rj@030000002>YnH0SZOV00<0SZOV00<0000009Z/h`2J[><00P2J[><010000000Y[3P
0:J`h02V/>04000000@0/K[P00<000000;Zng02j_]`00P0000020;c3g`0400000034a]/0a<KK0000
0080a/cN00<000000<g=fP0000001@3?d]`010000000e=KJ0=CFfP3De]X200000080e=[L00<00000
0=OHf00000000`3Gfm/01@000000eMkN0000003FfMX0000000@0dm[M00<000000<oDg@3?e=d00P3?
e=d300000140oooo0P0000000`3oool000000?ooo`090?ooo`050?oog`2LB00000000000M02oool0
103oool000@0oooo00<0oono07@00018W=l00`3oool00`3OW4P0000004RLg`0:0?ooo`030=nLB000
B9`0gooo00@0oooo00@0oono07@00000B9`0gooo0`3oool00`2oM000M;oo0?ooo`0D0?ooo`030?oo
_`1d0000M;oo01/0oooo00<000000?ooo`3oool04`3oool010000000c<CC0<c4d`3<a=<200000080
bLCF00@000000<Jod`00000000000`32_]H00`000000_kSD0;nhe0020;nhe0030000002j]mH00000
00@0]k7C00@000000;>/d@00000000000P2_Z]<00`000000[:KA0:bVd@020:bVd@040000002YX/l0
ZJ;?0:VRc`<000000`2UW/l00`000000Xic?0:>Lc`020:>Lc`040000002QW<l0XIc?0:6Lc`@00000
0P2RWLl00`000000XioA0:>Od@020:>Od@030000002SX]<0Xj;C0080Xj;C0P0000040:VXe@030000
002[[MT0ZjgI0080ZjgI0`0000040;:dfP030000002d^ml0];_O0080000000@0^[gL0000002k`n00
^l?P0P0000020<34g@0400000031b^800000000000@0aL[N00<000000<O?h037cn000`37cn001@00
0000b/kM000000000000am7R008000001039cml00`000000aLoR0<G?hP030<G?hP030000002nc>D0
_/cU00<0_/cU00<000000;k8hP000000102danP3000000<0Z;oZ00@0000009VjkP2I^^h0VK[^3@00
00000`1RX?800000056Fl002056Fl0D000000P0kR^d500000080AHc]1P0000000`1TUn`0000007:L
j`0207:Lj`0300000020X^T0P:;Y0080P:;Y00<0000008jWiP2>YnH00P2>YnH010000000VZcS09Z/
h`2J[><5000000<0Y[KT00<000000;6jh02a^^000P2a^^03000000@0_<?O00<000000<C6f`000000
1036c=h00`000000cm;L0<oBg0030<oBg0050000003@eml000000=CFfP0000000P3Df]`010000000
e=[L0=CJg00000020=OKf`8000000`3Eg]h00`000000dm[M0=?Jg@020=?Jg@040000003@fn400000
0<oDg@8000000`3;e>400`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0100oooo00040?oo
o`030?oOW0180000B9cO0080oooo00@0okmd0000000007@0_ooo2P3oool00`3OW4P004RL0=ooo`03
0?ooo`040?oo_`1d0000001809cOo`@0oooo00<0omnL04P0001d_ol0503oool00`3ookl0M00007Bo
o`0J0?ooo`030000003oool0oooo01D0oooo00<000000<c4d`0000001@39a=H00`000000`[kF0<:n
eP040<:neP040000002o^=@0000000000080^[OF00<000000;Nad`2g/M<00`2g/M<00`000000[j[C
0:nZd`030:nZd`030000002/Y]40[:KA0080[:KA0`0000020:JQd0030000002UW/l0YIk?0080YIk?
00<000000:>Lc`2SW<l00P2SW<l300000080Xic?00D000000::Mc`2RWLl0XYg?000000040:>Od@03
0000002SX]<0Xj;C0080000000<0YjCB0000002YZ=D00`2YZ=D00`000000ZjgI0:^]f@0200000003
0;2_e`000000/[CJ00D0/[CJ0P0000040;Zmg08000001@30a=d00`000000aL[N0<G:gP030<G:gP03
00000038b]/0000000<0aloP00<000000<[>g@3:c]d00P3:c]d00`000000bLoO0<W?g`030<W?g`03
0000003:cM`000000080aLoR0P0000001035c=l000000000002nc>D200000080_/SR00<000000;C7
j02danP00P0000020;C3i@040000002X_nX0Z;oZ0:RojP@000000P2J]^/00`000000RZo^08Z_kP02
08Z_kP060000001hYo00N:O`07RWl0000000IYk`100000000`1AU_00000004>>kP0204>>kP@00000
0`0jR>`4000000<0Ci3]00<0000005bFkP000000100000000`1bW>/00000082Rj@02082Rj@040000
002>YnH0SZOV08jWiPD000000P2I/>L00`000000Y[KT0:Jfi0030:Jfi0030000002a^^00/K[P0080
00000P2b_n@00`000000_<?O0;c3g`020;c3g`8000001036c=h00`000000b=7Q000000020<oBg080
00000`3@eml00`000000e=[L0=CJg0020=CJg0<0000000<0em_K0000003Eg]h00`3Eg]h010000000
dn3R0000003Cf]d2000000<0d=_Q00<000000<_Dh@3;e>400P3;e>400`000000a]?T0000000D0?oo
o`030000003oool0oooo00l0oooo00020?ooo`030?oog`2LB000000000D0000000<0001d0;ooo`3o
ool02P3oool00`3OW4P0001d0;ooo`020?ooo`030=nLB000000000000080000000@0B9cO0?ooo`3o
ool0okmd0`0000000`0007@0_ooo0?ooo`0A0?ooo`030?oo_`1d0000M;oo01X0oooo00<000000?oo
o`0000005P3oool00`000000bLCF0<W4eP030<W4eP0300000035a=T0000000D0`[kF0P0000050;Zg
eP030000002g/M<0]k7C0080000000<0/ZoE0000002_Z]<00`2_Z]<00`000000[:KA000000030000
00@0YZ7@00<000000:FNc`2UW/l00P2UW/l00`000000Xic?0:>Lc`03000000@0Xic?00<000000::M
c`2RWLl00P2RWLl010000000XioA0:>Od@2SWm43000000<0YjCB00<000000:VXe@2YZ=D00`2YZ=D2
000000@0/:oG00@000000;:dfP2b]=X0/[CJ0P0000000`2h]mP000000;Zmg0020;Zmg0030000002o
_]T0000000@0`<CM00<000000<G5fP0000000`35b]h01`000000b<[K0<S:f`38b]/000000<O?h000
0000103:c]d00`000000c<gJ000000020<W?g`8000000`3:cM`2000000D0aLcO00<000000;k8hP2n
b>800`2nb>82000000D0]<?U0`0000020:Rki`030000002J]^/0V[K[0080V[K[00D0000008Z_kP2:
[nh0RZo^0000000307RWl0030000001VW_00IYk`0080IYk`1@0000001013S^h0000003Z9k00jRN`3
0000000403Z8k0000000?XW/03j9k0<0000000<0Ci3]0000001LU^h00`1LU^h00`000000K9c]0000
0008000000<0Rj[Z00<0000009V`i`2I/>L00P2I/>L00`000000Y[KT0:Jfi0020:Jfi0<00000102b
_n@00`000000_<?O0;c3g`0200000080_/WS00@000000<K<gP00000000000P38dN400`000000cm;L
000000040=3Gg`040000003@g>8000000=CJg080000000@0e=kO0000003Dg]l000000`3Eg]h20000
00<0dn3R00<000000=3Kh@3@fn400P3@fn4010000000c=cU0000003;e>4200000080a]?T00<00000
0?ooo`3oool04P3oool00`000000oooo0?ooo`0?0?ooo`005`3oool01@3OW4P00000000000000000
M;oo01h0oooo00D0gia8000000000000000007Boo`0J0?ooo`800000603oool2000000030<W4eP00
0000000000<0aLCI00<000000<:neP32_]H00P0000020;fmf@030000002j]mH0^[OF0080^[OF0`00
00030;:_e@040000002_Z]<0[j[C0:nZd`@000000`2XY=800`000000YZ7@0:JQd0020:JQd0040000
002UW/l0YIk?0:FNc`<000000`2SW<h00`000000Xic?0:>Lc`020:>Lc`030000002RWLl0XYg?0080
XYg?0`0000020:JQd0030000002WY=80YjCB0080YjCB00<000000:VXe@2YZ=D00P0000000`2^Z]@0
00000;2_e`030;2_e`040000002b]=X00000000000<0^;OH00@000000;Zmg000000000000`2o_]T0
10000000`<CM0000000000030<G5fP0300000035b]h0000000D0b<[K00@000000<[>g@3:c]d0b/kM
0P0000020<c=fP0300000039cml0000000@0b/gL00<000000<[;fP0000000`35c=l2000000040<G9
g@000000_/SR0;k8hP80000000<0_LGP0000000000000P2d`nD2000000030;>nhP000000Z;_W0080
Z;_W00@0000009Zfj`2J]^/0V[K[4@0000000`1DUNl00000000000040000000303Z9k00000000000
00<000000P0nRN`60000000305bFkP000000K9c]00<0K9c]00D0000007^Sk01kXn`0Nj?/00000004
08^ZjP030000002I/>L0VK3W0080VK3W00<000000:Jfi02V]^@00P0000020:Jkj0030000002b_n@0
/[oT0080/[oT0`0000050;k9h`0300000038dN40b=7Q00<0b=7Q00@000000=3Gg`3@eml0d=OO0P00
00030=3LhP040000003Dg]l0e=kO0=CNg`80000000@0d^;S0000003Eg]h00000103Ch>8010000000
cn;V0000003@fn42000000<0c=cU00<000000<KCi036dn@00`36dn@00`000000oooo0?ooo`0A0?oo
o`800000403oool003d0oooo00D0omnL04P000000000001d0;ooo`030?ooo`060?noM0000000M;oo
0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`080?ooo`030000003oool0oooo01T0oooo00<0
00000<G4f@35a=T01035a=T2000000@0_KgI00<000000;ZgeP2j]mH00P0000020;Fef@030000002b
[mD0/ZoE0080/ZoE1P0000040:RTdP030000002VXM00YZ7@0080YZ7@100000000`2SWLl000000:>L
cP030:>LcP030000002SW<l0Xic?0080Xic?00<000000::Mc`2RWLl00`0000040:JQd0030000002W
Y=80YjCB00<0YjCB0P0000030:jZe0030000002`[mL0/:oG00<0/:oG00<000000;Rgf02h]mP0102h
]mP00`000000_kkI0;nnf@040;nnf@0300000035aMX0aLGJ00@0aLGJ00<000000<S:f`38b]/00`38
b]/01@000000bl[I0000003:c]d0000000D0c<gJ00<000000<[=g03:cM`00P3:cM`01@000000b/_J
0<[;fP000000aLcO008000001035bMd2000000D0_LGP0P0000030;>nhP040000002X^nL0Z;_W0:Rk
i`@000000`2K/^P01@000000S:c[08b/j`2<[>/0000000<0O:G]00<0000006ZMkP1ZWNh00P1ZWNh<
0000000303N7j`0000000000008000000`17S^d00`000000EIG_000000040000000306bLk@000000
Nj?/00<0Nj?/00<0000008^ZjP2;Z^X00P2;Z^X010000000VK3W09V`i`2I/>L3000000@0Y[_X00<0
00000;:oi02b_n@00P0000020;?6j0040000002nbN<0_/WS0;k9h`80000000<0_m3W00000038dN40
0P38dN401@000000bMOU0000003@eml0000000@0d=cR00@000000<oQi@000000e=kO0`0000020=;R
h`040000003Ch>80dn3R0=?PhP8000000`3?h^H010000000c=cU0<cLi@3<g>D2000000040<KLj@00
0000a]?T0<KCi080000000<0_m;X0000003oool03P3oool3000000030?ooo`000000oooo00l0oooo
000l0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo0?oo
o`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00P0oooo00<000000?ooo`3oool06P3oool0
0`000000aLCI0<G4f@020<G4f@0300000030a=d0000000D0_KgI0P0000040;Fef@030000002b[mD0
/ZoE00D000000`2[Z=@00`000000Z:CB0:RTdP020:RTdP060000002VXM0000000000002TWm400000
102SWLl00`000000Xic>0:>LcP020:>LcP030000002SW<l0Xic?00@000000`2UW/l00`000000YZ7@
0:JQd0020:JQd0030000002WY=80YjCB0080000000<0[:KA0000002^Z]@00`2^Z]@00`000000/:oG
0;2_e`02000000030;Jae@000000^;OH00<0^;OH00<000000;fieP000000102o_]T00`000000`l3G
000000040<G5fP0500000038a]P0000000000038b]/00P0000030<_:f@030000003<cMX0c<gJ00<0
c<gJ00H000000<g;f000000000000<[=g00000040<[;fP0300000035bMd0aLWM00<0aLWM00<00000
0<G6f`0000000`2maN02000000030;g1g@000000/kkR00<0/kkR0`0000020:Rgi0030000002K/^P0
Vk;X0080Vk;X00@0000008b/j`2<[>/0S:c[100000000`1lYNd0000006ZMkP0206ZMkP060000001I
U^h0FIK^05VFkP000000Bho]0`00000203^9k0@0000000<0>HS/0000000000000P0000000`17S^d0
000005FEk`0205FEk`030000001UWO00000000h000000P2H]N`00`000000Y[_X0:Jkj0020:Jkj0<0
0000102ca^P010000000_/WS0000000000030;o@i`0400000038dN400000000000<0bMOU00@00000
0=3LhP3@g>80d=cR0P0000030<oQi@040000003Bh^<0d^;S0=;Rh`80000000@0d>KV0000003Ch>80
00000`3?h^H3000000030<cLi@000000000000@0a]cY0P0000030;oBj0030000003oool0oooo00T0
oooo100000040?ooo`030000003oool0oooo00h0oooo000l0?ooo`030;md0000B9`0gooo0080oooo
00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd0;ooo`080?ooo`800000703oool010000000aLCI
0000000000030<34g@030000002m_MT0_KgI0080000000<0^;cL0000002e]MT00`2e]MT400000080
[JgG00<000000:^Xe02[Z=@00P2[Z=@010000000Z:CB0:RTdP2XY=83000000@0Y9oA00D000000:>M
c`2SWLl0Xig?000000040:>LcP<000000P2TWLh00`000000YIk?0:FNc`020:FNc`030000002VXM00
YZ7@0080YZ7@0`0000030:bVd@030000002^Z]@0[Z[D00<0[Z[D0P0000040;Jae@040000002h]mP0
0000000000<0_KWF00@000000;nnf@00000000000`33`=L010000000aLGJ0000000000040<S6f003
0000003;b]T0bl[I0080bl[I00D000000<g9e`00000000000<c=fP02000000@0cL_H00<000000<[;
fP3:bmX00P3:bmX01@000000b/WI000000000000aLWM008000000P35a]/010000000_LGP00000000
00030;g1g@030000002c_^80/kkR0080000000<0/k_O0000002X]n@00`2X]n@010000000Vk;X09^b
j02K/^P400000080SZWX00<0000007nSjP0000001`0000000`1IU^h0000004^?k@0304^?k@<00000
0P0fQn/30000000303V8k0000000@Xg^00@0000000<0EIG_0000001UWO000`1UWO000`000000MZK`
07JVl00207JVl00300000028[^l0R:k_0080R:k_00D0000009Rek02H]N`0V;G/000000020:Jkj080
00000P2V`Nd00`000000/lKX0;?6j0020;?6j08000001@2od>L2000000@0bMOU00D000000<WMj000
0000d=cR000000040<oQi@040000003=i^P000000=;Rh`<000000P3@i^H00`000000cn;V0<oRiP02
000000<0b^CZ00<000000<KLj@36g>T00P36g>T3000000@0_m;X0P0000070?ooo`<00000203oool0
0`000000oooo0?ooo`0>0?ooo`00>`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo
00/0oooo00<0_g@007Boo`3oool01P3oool2000001h0oooo00<000000<34g@30a=d01030a=d20000
00@0^;cL1@0000000`2_/m/000000:f]e`030:f]e`030000002[Z=@0ZjSD0080ZjSD0`0000020:FR
d`030000002TWm40Y9oA0080Y9oA00<000000:>Mc`2SWLl00P2SWLl01P000000Xic>000000000000
Xig?000000@0Y9g>00<000000:FNc`2UW/l00P2UW/l01P000000YZ7@000000000000ZZ?@000000D0
[:KA00<000000:jZe02^Z]@00`0000060;Jae@030000002m^MH0_KWF00@0_KWF00<000000<?0e`33
`=L01033`=L00`000000b<KH0<S6f0030<S6f0040000003;a]H000000<_:f@800000103=bML00`00
0000cL_H0<g;f0020<g;f080000000<0c/[F0000003:bmX00P0000040<[9f@0300000035a]/0aLKK
00<0aLKK00<000000;g1g@2m`Md0102m`Md2000000<0/k_O00@000000:Rgi02X]n@0Z;OT1@000002
09b_i0050000002>ZNP0SZWX08jYj0000000101oXnX010000000L9g/072Mk01`WN`60000000304^?
k@000000?h[/00@0000000<0=XO[0000000000000`00000204:=kPH0000000<0IIg`0000001fY_00
0`1fY_000`000000R:k_08R^k`0208R^k`040000002H]N`0V;G/09Rek0<00000102V`Nd00`000000
/lKX0;?6j00200000080/lc]00<000000;o@i`2od>L00P0000020;oFj`0400000039enD000000000
00<0bMgX00@000000<oQi@3?hND0cn7U0P0000030<gVj0070000003@i^H0d>KV0000003@i^H00000
0<cXjP02000000@0b^CZ00@000000<CUkP000000a]cY0P0000030;oKkP040000002od^P000000000
0080]m3/00@000000?ooo`3oool0oooo0`00000<0?ooo`030000003oool0oooo00d0oooo000k0?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3ooml0W4P004RLg`070?oo
o`030000003oool0oooo01h0oooo00<000000<34g@30a=d00P30a=d00`000000^/?Q000000020;Rl
g0D000000`2_/m/00`000000[JgG0:f]e`020:f]e`030000002[Z=@0ZjSD0080000000<0YZKF0000
002UX]<00P2UX]<00`000000Y9oA0:BOd@030:BOd@030000002SWLl0Xig?00<000000`2SWLl00`00
0000Y9g>0:BMcP020:BMcP030000002UW/l0YIk?0080YIk?0P0000040:ZSd0030000002/Y]40[:KA
0080[:KA0`0000030;>]dP040000002f/MD0][7E0;Jae@80000000<0^[CC0000002m^MH00P2m^MH2
000000030<6ke0000000`l3G0080`l3G0P0000001037`MD000000<S6f038a]P2000000<0blKF00<0
00000<g9e`3=bML00`3=bML010000000clWE0000003=bmP2000000@0c/[F00<000000<[9f@3:bMT0
0`3:bMT00`000000b/OG000000030<G6f`0400000034`mT0000000000080_L7M0P0000000`2k_]/0
00000;>kg`030;>kg`<000000`2X/n401@000000W:oT09b_i02L[n@0000000@0SZWX00D0000007nS
jP1oXnX0Oj?Z00000003072Mk0060000001PU^d0H9K]062Fk@000000DI7]100000000`0oR^`00000
03R8k00300000080=hS/1000000304jEl0030000001OW_80000000h000000P2F^_400`000000Y/7]
0:K1k@030:K1k@8000001@2cc>d2000000D0_mK[00<000000<WMj039gNP00P39gNP01@000000b>?[
0000003?hND0000000<0cNKX0P0000000`3:j^`000000=3ViP02000000060<cXjP00000000000<K[
kP000000b^CZ0P0000030<CUkP030000002ofnh0_m_^00<0_m_^0P0000040;O@k0@000003`3oool0
0`000000oooo0?ooo`0=0?ooo`00>`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00X0oooo00<0gia80000001d_ol01`3oool00`000000oooo0?ooo`0O0?ooo`0400000030a=d00000
00000080^/?Q0`0000001@2h_=`000000000002a^^00000000@0[k?K00@000000:f]e`2][ML0[JgG
0`0000030:JVeP030000002UX]<0YJ;C0080YJ;C00@000000:BOd@2TWm40Y9oA100000020::Md003
0000002SWLl0Xig?0080Xig?00<000000:BMcP2TWLh00P2TWLh01P000000YIk?000000000000Yj3?
000000@0ZZ?@00<000000:bVd@2/Y]400P0000000`2`ZM4000000;>]dP030;>]dP040000002f/MD0
0000000000<0^[CC00@000000;fieP00000000000`31^m@010000000`l3G0000000000030<O1e@03
00000038a]P0000000@0blKF00<000000<g6d`0000000`3=bML01@000000clWE0<o9e@3?bMD00000
00D0c/[F00<000000<k8e@0000000`3:bMT01@000000b/OG0<[7e`000000aLKK008000001034`mT2
000000<0^kkK00<000000;>kg`2c^ml00P0000000`2b]md000000:Rch@030:Rch@040000002L[n@0
W:oT09b_i140000000<0H9K]0000001ATNd00P1ATNd500000080>8S/0`00000203f=kP@0000000<0
CYG`0000001OW_800P1OW_800`000000LJSc076Xl`02076Xl`0300000024/_<0Q;;c0080Q;;c00<0
000009Jjl@2F^_400P2F^_400`000000Y/7]0:K1k@02000000030:G7l@000000/lc]0080/lc]0P00
00000`2cdo4000000;oFj`020;oFj`80000000<0_mg^00000039gNP00P0000030<SSj`030000003=
i^P0cNKX00800000103:j^`010000000c>SZ0000000000030<K[kP8000000`34iNh2000000040;cT
l`000000_m_^0;oKkP80000000@0]][b0000002gd>`0]m3/0P0000020:k=l0030000003oool0oooo
00h0oooo00<000000?ooo`3oool03@3oool003/0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`060?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`070?ooo`030000003oool0
oooo0200oooo00<000000;[3h@0000000P0000020;[3h@8000000`2a^^000`000000[k?K0:ncf`03
0:ncf`<000000P2WZmX00`000000YZKF0:JVeP020:JVeP030000002UX]<0YJ;C0080YJ;C0`000002
0::NdP050000002RWM00XYg@0::Md0000000102SWLl010000000Y9g>0:BMcP2TWLh3000000<0Yj3?
00<000000:ZSd02ZXm000`2ZXm02000000@0/:WA00<000000;>]dP2c[M800P2c[M800`000000^[CC
0;Zdd`040;Zdd`0300000031^m@0`K_D00@0`K_D00<000000<O1e@37`MD01037`MD00`000000blKF
0<_6eP020<_6eP070000003=a]<0cLKC0<g6d`000000cLWG000000030<o9e@030000003?b=<00000
00<0c/[F00L000000<k8e@3>b=D0c/SE0000003:bMT0000000@0b/OG00<000000<C3f@34`mT00`34
`mT00`000000`/7H000000040;^nf`800000102b]md010000000Z;?Q0:Rch@2X/n4400000080WJcQ
00<00000096Wi02AYn@00P2AYn@01@000000Q:;W08BRi`24X^L0000000<0MIgZ00<0000006NHk000
00000`0000000`1ATNd0000004F=k@0204F=k@@0000000<0=h[]0000000000000P0000000`0mSNh0
000004VFl@0900000003076Xl`000000Q;;c0080Q;;c2@0000030:G7l@040000002cc>d000000000
00<0/m?a00@000000;oFj`00000000000`2ogNh2000000@0b>?[00@000000<GXk`0000000000103:
j^`2000000050<G]k`000000a^_^00000036jnh00P0000020;o/lP<000000`2li?<3000000@0]][b
0P0000040:k=l0030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool003/0oooo00<0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`060=nLB000000000000000000004P0
W=oo203oool200000200oooo100000020;[3h@80000000<0/l;U0000002a^^000`2a^^000`000000
[k?K0:ncf`02000000030:RagP000000Yj_J0080Yj_J00<000000:JVeP2VY]H00P2VY]H00`000000
YJ;C0:FRd`02000000030:6Qe0000000XYkB0080XYkB00<000000::Md02RWM000P2RWM000`000000
Xig?0:>Mc`020:>Mc`<000000P2UW/l00`000000Yj3?0:NPc`020:NPc`030000002ZXm00ZZ?@0080
000000<0[JG?0000002`ZM400`2`ZM400`000000/jgB0;>]dP02000000030;N_d@000000^[CC00<0
^[CC00<000000;jfdP0000001031^m@00`000000aKgB000000040<O1e@060000003:`]<000000000
003;a]H000001@3=a]<01@000000clWE0<o9e@3?bMD0000000<0clSC00<000000<k:eP0000001@3>
b=D00`000000b/OG0<[7e`020<[7e`0500000039aMH0000000000034`mT00P0000020<;1f0030000
002k_]/0^kkK0080000000<0^[_J0000002b]md0102b]md300000080Yk3N00D0000009f/h@2M[>40
WJcQ00000004096Wi00500000024X^L0Q:;W08BRi`0000000`1eWNX00`000000IiS/06NHk00206NH
k0H0000000<0AHg]0000000mRnd02`0000000`19U_40000005VPm00205VPm0030000001/[?L00000
00@000000`21]oL00`000000U<3e09C0m@0209C0m@040000002Uao40YLOa0:G7l@8000001P2cdo40
10000000_mg^0;oMkP2ogNh200000080_N?b00@000000<SSj`00000000000P35j>l010000000`Nca
0000003:j^`200000080aNg_00<000000;o_lP0000000P0000030;o/lP<000000P2li?<200000080
/^?g00<000000;KJlP2ff_800P0000000`2/e_H000000:k=l0020:k=l0<000003`3oool00`000000
oooo0?ooo`0=0?ooo`00>`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo
00@0oono07@000000000B9cO2P3oool00`000000oooo0?ooo`0L0?ooo`<00000103oool2000000@0
/l;U00@000000;6jh02a^^00/K[P0`0000030:RagP030000002WZmX0Yj_J0080Yj_J00@000000:JV
eP2VY]H0YZKF0`0000030:6Qe0030000002RW]80XYkB0080XYkB00D000000::Md02RWM00XYg@0000
00020:>Mc`80000000<0Xig?0000002UW/l00P2UW/l00`000000Yj3?0:NPc`030:NPc`8000000`2]
YLl00`000000/:WA0;2Yd@030;2Yd@800000102g[m4010000000^[CC0000000000020;jfdP050000
0031^m@0`K_D0<6ke00000000`35_M8010000000al7E0000000000040<[2d`030000003=a]<0cLKC
00<0cLKC00D000000<o7dP000000clWE000000050<o8d`030000003>b=D0c/SE0080c/SE0P000000
0`3>am@000000<[7e`02000000@0bLGF00<000000<;1f032`MP00`32`MP2000000@0^[_J00<00000
0;:gg@2b]md00P0000000`2a]=/000000:N`gP020:N`gP040000002M[>40WJcQ09f/hA40000000<0
IiS/0000001IU>d00P1IU>d00`000000CI7^000000030000000303f;k@000000>Xg^00@0000000<0
>ho_00000016V?801@0000000`1IX?@0000006b/m`0206b/m`0300000021]oL0PKOg0080PKOg00@0
000009C0m@2D`?D0U<3e1@0000020:?>mP8000000P2cdo42000000040;;Jm@0000000000000000@0
_N?b0P0000020<GXk`8000000`31k?400`000000aNg_0<G]k`0200000080_nob00<000000;K`m@00
00000`0000030;K[mP8000000`2bhoL400000080[=Kf1@0000020:?8l`030000003oool0oooo00d0
oooo00<000000?ooo`3oool03@3oool003`0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo
0?ooo`070?ooo`030;md001d_ol0oooo00X0oooo00<000000?ooo`3oool06P3oool2000000P0oooo
00<000000;?2i@2c`^D00`2c`^D300000080ZKST00<000000:RagP2X/Mh00P2X/Mh010000000Yj_J
0:N[fP2WZmX400000080XJCH00<000000:6Qe02QXM@00P2QXM@01@000000XYkB0::NdP2RW]800000
00<0XYg@0`0000030:>Mc`030000002UW/l0YIk?0080YIk?00<000000:NPc`2WX<l00P0000000`2Z
X/l000000:fUc`030:fUc`070000002`ZM40/:WA0;2Yd@000000]:_@000000050;N_d@030000002n
]]80_[KB00<0_[KB00<000000<6ke00000001@35_M800`000000b/;C0<[2d`030<[2d`030000003<
`m4000000080cLKC0P0000030<o7dP030000003?b=<0clSC00<0clSC00@000000=38dP000000c/SE
0P0000040<k7e00300000039aMH0bLGF00<0bLGF00<000000<S3e@0000000`32`MP00`000000`KkG
000000040;ZkfP<000000`2a]=/010000000Yk3N0:N`gP2W/=h400000080WZWN00<0000009>UhP2C
YN800P2CYN801@000000R:7U08RQi@28XND0000000<0NigX00<0000006nJj`000000200000000`1=
TNh00000000000040000000303Z=kP000000000000@0000000<0AYSb0000001EXoH03P0000030976
n@040000002Sc_H0Xlkf0:?>mP@000000P2bf_D2000000040:oPn0000000_N?b0;gSlP8000000P2j
j?@2000000<0`Nca0P0000000`2kko@00000000000020;o_lP8000000P2fl?D300000080]^_f0P00
00000`2[jOT0000000000002000000<0Ymok0`0000030:3Bn@040000002Sb?<0XlSc0:?8l`800000
3P3oool00`000000oooo0?ooo`0=0?ooo`00?03oool01`3ookl0M00007Boo`3oool0okmd0000M02o
ool0203oool01P3ogi`0B000000000000000001809cOo`H0oooo00<000000?ooo`3oool0603oool3
000000/0oooo00<000000;?2i@2c`^D00P0000000`2Z`NX000000:Vhi0030:Vhi0040000002X/Mh0
Z;7N0:RagP@000000P2QZM`01@000000XJCH0:6Tf02QY=P0000000@0XJ7D00@000000::NdP2RW]80
XYkB100000020:6Md@030000002SWLl0Xig?0080Xig?00<000000:FNc`2UW/l00P2UW/l2000000<0
ZZ;?00<000000:fUc`2]YLl00P2]YLl010000000/:WA0000000000020;B[d0030000002g[m40]joA
0080]joA00<000000;^bd@0000001@2n]]800`000000aKgB0<FmdP030<FmdP0300000038_]400000
0080b/;C0P0000020<c3d@030000003=a]<0000000@0clOB00<000000=37d@0000000`3?b=<01@00
0000d<SB0=38dP3@b=80000000D0c/OD00<000000<g6d`0000000P39aMH200000080b<?E00@00000
0<;1f000000000000P31_]L00`000000^[_J0;ZkfP0200000080^;SH00<000000;6df`2a]=/00P2a
]=/300000080YjgL00D0000009jYgP2NZMh0WZWN0000000409>UhP0500000028XND0R:7U08RQi@00
0000101kWNP01@000000Ki[[06nJj`1_V^/0000000<0HYO]00<0000005JFk`0000000P00000204B@
k`030000000oTO000000008000000P0lT_000`000000A9[d000000040000000305FSmP000000J:oi
0080J:oi00<0000007bknP1l^oX00P1l^oX010000000TLKi0976n@2Aa_T5000000<0XMCj0`000003
0:oPn0<000000P2jj?@200000080]Ncg0`0000000`2kko@00000000000020;?am@8000000P2fl?D2
00000080[>og0`0000000`2[jOT000000000000209oUo0030000002Wgo/0Ymok0080000000<0V][m
0000002Pd_T00P2Pd_T500000080V<?e00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0=
0?ooo`00?@3oool01@3ookl0M0000000000007@0_ooo00X0oooo00D0_g@0000000000000000004RL
g`060?ooo`8000005P3oool3000000l0oooo0P0000040:[1jP040000002Y^>@0ZKST0:Vhi0@00000
0P2P[n801@000000XJWL0:6Yg02QZM`0000000@0XJCH00@000000:6Qe02QXM@0XJ7D100000000`2O
W]<000000:6Md@030:6Md@030000002SWLl0Xig?0080Xig?00<000000:FNc`2UW/l00`0000050:ZR
c`030000002]YLl0[JG?0080[JG?00<000000;B[d02dZm000`2dZm0010000000]joA000000000002
0;^bd@030000002n]]80_[KB0080_[KB00<000000<:id@0000000`35_M801P000000b;kA0<Rnd@00
0000b/;C000000D0c<?A00<000000<o7dP3?am800P3?am801`000000d<OA0=37d@3@am4000000<o8
d`0000000`3@b=800`000000d<OA000000030<k7e0060000003=a]<0cLKC00000039aMH000001@38
`mD00`000000`KkG0<6ne`030<6ne`800000102h^=P00`000000/KCK0;6df`03000000@0YjgL00<0
000009jYgP2NZMh00P2NZMh010000000TjGR09>UhP2CYN8@0000000306:Gk@000000EYK_00<0EYK_
0`00000403nAl0<000001014V_@00`000000DZKh0000000=000000<0SLcm00<000000:7DnP2Qe?X0
0P000000102Mf_`000000:oPn02_h?P200000080ZnGj0`0000020;G/m`8000000P2^koL200000080
/o7e0P0000020:[amP8000000P2/koL2000000040:7/n@0000000000000000<0WnGl1000000209[J
o@D000000P2Cc?/00`000000V<?e09S3m@0209S3m@030000003oool0oooo00`0oooo00<000000?oo
o`3oool0303oool005L0oooo00<000000?ooo`3oool04P3oool3000001<0oooo0P0000030:[1jP@0
00000P2P]^P01@000000X:oR0:2_hP2P[n80000000@0XJWL00@000000:6Tf02QY=P0XJCH10000000
0`2MX=H0000009nNd`0309nNd`030000002QWM40XIgA0080XIgA00@000000:>Mc`2SWLl0Xig?0`00
00030:NPc`030000002ZX/l0ZZ;?0080ZZ;?00<000000:fUc`2]YLl00`0000060;B[d0030000002k
/]40^k;A00<0^k;A00@000000;jfdP00000000000`32^M400`000000aKgB000000040<Rnd@030000
003<`m40c<?A00<0c<?A00H000000<k4d000000000000<o7dP0000050=37d@050000003@b=80d<SB
0=38dP0000000`3@am400`000000c/OD000000040<g6d`0300000038`mD0b<?E00<0b<?E00<00000
0<K1e00000000`31_]L00`000000_k_F000000050;Rhf08000000`2_/MT00`000000YjgL0:N]g002
0:N]g0P000000P2FY=l01@000000S:7R08bQhP2<XN80000000@0PYkU00@0000007NLj@1gW>T0MicY
1P0000000`1FU^l0000004fEl@040000000303nAl0000000?iOb00T0000000<0DZKh0000001T/o/0
0P1T/o/00`000000N<3m07S0o@0207S0o@040000002=c?d0SLcm08g<o@@000000`2Mf_`300000080
ZnGj0P0000020:GYn`<000000P2^koL2000000040:G_m`000000000000000080Z_7f0P0000000`2O
koH00000000000040000000309CXnP000000000000<000000P2Bh?h4000000<0S=?o00@0000009?<
n`2Cc?/0Tlck1`00000=0?ooo`030000003oool0oooo00`0oooo001G0?ooo`030000003oool0oooo
00l0oooo0`00000H0?ooo`<000000P2O_^h00`000000X;KX0:2fj0020:2fj0050000002P[n80X:oR
0:2_hP0000000`2QZM`40000000309^SfP000000WJ3F00<0WJ3F00D0000009nNd`2OW]<0WikC0000
00030:6Md@D0000000<0Xio@0000002WX<l00`2WX<l00`000000ZZ;?0:ZRc`020:ZRc`<000000`2`
Z=0010000000]:_@0;B[d02dZm03000000H0^k;A00<000000<:id@32^M401032^M400`000000b;kA
0<Rnd@020<Rnd@030000003;`=00000000<0c<?A00<000000<k4d03>a=000P3>a=000`000000d<OA
0=37d@030=37d@050000003Aam0000000=38dP0000001@3@am400`000000cLKC0<g6d`020<g6d`05
0000003<a=<0000000000038`mD00P0000020<K1e00400000031_]L00000000000<0_k_F00<00000
0;Rhf02h^=P00`0000040:naf@030000002W[M`0YjgL0080000000<0Yj_J0000002OYm`00`2OYm`0
1@000000UZCO09JTg`2FY=l0000000<0S:7R00<0000008:Ni@22W^D00P22W^D01@000000MicY07NL
j@1gW>T0000000<0K9_/00<00000066Kk`0000000`0000000`1=UO40000004JHlP090000000304FO
m@000000D:[i00h000000P29dOl500000080V=kn0`0000020:GYn`8000000P2MjoX300000080YNog
0P0000000`2Kk_D00000000000060000000308SWm@000000000000T0000000<0Q=Wo000000000000
2000000208G4o0030000002;_OL0Rkgg0080Rkgg00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`0<0?ooo`00E`3oool00`000000oooo0?ooo`0<0?ooo`<00000703oool00`000000Wkk^09nn
kP0209nnkP040000002P]^P0X;KX0:2fj0P0000000<0VJOO0000002KXmX00`2KXmX01@000000WJ3F
09fPeP2MX=H0000000<0WikC100000020:2NdP030000002SWm00Xio@0080Xio@00<000000:NPc`2W
X<l00P2WX<l01P000000ZZ;?000000000000[:G@000000@0/:S@00@000000;B[d000000000000`2h
[]0010000000^k;A0;^bd@2k/]42000000030;ned0000000`[WA0080`[WA0P0000001@35^m000000
0<Rnd@38_]40000000<0bl3@00<000000<c3d@000000103>a=000`000000d<G@000000020=37d@80
00000`3Aam000`000000d<OA0=37d@030=37d@040000003@am4000000<g6d`800000103<a=<00`00
0000a/7D0<K1e0030<K1e0030000002o^mH0_k_F00@0_k_F0P0000030;Nee`030000002_/MT0[k7I
0080[k7I0P0000030:N[fP030000002OYm`0WjOL0080WjOL00@0000009JTg`2FY=l0UZCO4@000000
0`1/Vn`00000066Kk`02066Kk`H0000000<0AYSb00000015W?@00`15W?@600000003052Zn@000000
H;Ol0080H;Ol00@0000007C4o`1da?l0M<Co1@00000308CEo`030000002Hg_h0V=kn0080000000@0
T^;m00000000000000000P2MjoX200000080UN_g0`0000000`2Kk_D00000000000020000000308OX
l@000000000000@000000P1mh_D400000080Nmck1000000307OAo`050000001nc?l0O/co07k<o`00
00000`25a?`8000000/0oooo00<000000?ooo`3oool0303oool005L0oooo0P00000:0?ooo`<00000
803oool010000000Wkk^09nnkP2O_^h40000000309Bbk0000000UjcU00<0UjcU00D0000009VWg`2I
Yml0VJOO0000000309^SfP040000002MX=H0WJ3F09fPeP@000000P2LW]D00`000000X9kB0:2NdP02
0:2NdP050000002SWm00Xio@0:>Od00000000`2WX<l3000000@0[:G@00<000000;2Xd02`Z=000P2`
Z=000`000000^:k@0;R^d0030;R^d0040000002k/]400000000000<0_kG@00@000000<:id@000000
00000`35^m000`000000b;kA000000050<_0d0030000003>a=00c/C@0080c/C@00H000000=35d03@
aM0000000=37d@0000040=77d0030000003Bam0000000080d<OA0P0000030=37d@030000003<a=<0
c<CC00<0c<CC00<000000<_3d`0000000`36`M@00`000000aKkD000000030;nkeP<00000102g]ML0
0`000000[k7I0:naf@02000000030:j^f0000000Yj_J00<0Yj_J00@0000009nWg02OYm`0WjOL1000
00000`2HXm`00000092Qg`03092Qg`0300000027Wn<0QioS0080QioS00H0000007jOi`1nWnL0OYoW
0000001eWn`D0000000304RTm`000000D:oj00T000000P1`b?l900000080T^;m0P000000102<ho/0
000000000000000209G[m`80000000<0S>Wd000000000000200000000`1fh>l00000000000030000
000307CLm@000000000000<0000000<0L=Ck0000000000003000000307Nlo0030000001n]OL0O[Gg
0080O[Gg00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0<0?ooo`00EP3oool2000000T0
oooo0P00000T0?ooo`<000000P2A^O801@000000U;;/09Bbk02D/^`0000000<0UjcU00D0000009VW
g`2IYml0VJOO0000000309^SfP@000000P2IWmP01@000000W9kE09bNe@2LW]D0000000@0X9kB00<0
00000:>Od02SWm000P2SWm0300000080Z:;@00<000000:bUd02/YM000P2/YM000`000000/:S@0;2X
d002000000030;>[d0000000^:k@00@0^:k@00<000000;ned02o]M00102o]M000`000000aK_@0<Fk
d0040<Fkd0030000003;`=00bl3@00<0bl3@00<000000<g2d00000000P3>a=000`000000d<G@0=35
d0020=35d0030000003Aam00dLO@0080dLO@00H000000=;7d03Bam0000000=37d@0000040=37d@03
0000003?a]4000000080c<CC0P0000020<_3d`0400000036`M@0000000000080aKkD00@000000;nk
eP00000000000`2m^MD00`000000]kGG0;Nee`020;Nee`8000000`2^[]P010000000Yj_J0:N[fP2W
ZmX50000000309nUfP000000V:?L00<0V:?L00<00000092Qg`2@XMl00P2@XMl01@000000QioS08NO
h`27Wn<0000000<0OYoW00<0000007FOk01eWn`00P1eWn`01P000000K:3`06bPl01/X?00000006>S
m0@0000000<0D9ke0000001=Y?L00`1=Y?L?000000<0KLcn00@0000007oHoP1of?h0OmSn1`000000
0`24i?P00000000000050000000508?Wl@00000000000000001fhN`01P000002073Ml1X0000000<0
Mkcl0000001n]OL00`1n]OL00`000000oooo0?ooo`080?ooo`030000003oool0oooo00`0oooo001F
0?ooo`030000003oool0oooo00D0oooo0`00000W0?ooo`8000000`2A^O8010000000U;;/09Bbk000
000309N/i@P000000P2EXMd01@000000VIoH09VOf02IWmP0000000@0W9kE00<000000:2NdP2PW]80
0P2PW]801P000000Xio@000000000000Y:7B000000<0Z:;@00<000000:bUd02/YM000`2/YM020000
00<0/j_@00<000000;R^d02h[]000P2h[]000`000000^k7@000000040;ned00300000032^=000000
00@0aK_@00<000000<Rmd00000000`3;`=001P000000cL;@0<g2d0000000c/C@000000@0d<G@00H0
00000=76c`00000000000=77d00000040=;7d0030000003@am40d<OA0080d<OA00H000000<o6d@3?
a]4000000<c4d`0000050<_3d`0300000035_]@0aKkD00<0aKkD00<000000;fie@2m^MD00`2m^MD0
0`000000]kGG0;Nee`02000000030;FbeP000000[ZkH00<0[ZkH0`0000020:JXf0030000002OYMX0
WjGJ0080WjGJ00D0000009RSg02HXm`0V:?L00000003092Qg`d0000000<0MIo/0000001/X?000P1/
X?000`000000Hj?d06>Sm00206>Sm1l000000P1efoP600000080N>7`2P0000000`1ag^d000000000
000400000080K=Of1@0000000`1TcOX000000000000700000003062oo@000000I;Wm00<0I;Wm00<0
000006Zcn`0000001`0000090?ooo`030000003oool0oooo00`0oooo001F0?ooo`030000003oool0
oooo0080oooo0`00000/0?ooo`X000000P2<Z>P01@000000T:CR092ThP2@Y>80000000<0UJ7M00<0
000009VOf02IWmP00P2IWmP01@000000W9kE09bNe@2LW]D0000000<0X9kB0`0000030:BQdP030000
002XX]00Z:;@0080Z:;@00<000000:bUd02/YM000P0000000`2_Z=0000000;>[d0030;>[d0040000
002h[]00000000000080^k7@00D000000;ned02o]M00_kG@000000020<:hd00500000035^m00aK_@
0<Fkd00000000`38_M000`000000bl3@000000040<g2d0030000003@aM00d<G@0080d<G@00<00000
0=76c`3Aa/l00P3Aa/l00`000000d/O@0=;7d0020=;7d0060000003Bam0000000000003@am400000
103?a]400`000000bl?C0<_3d`030<_3d`0300000039`]<0000000<0aKkD00<000000<>le0000000
1@2m^MD2000000<0]K;F00<000000:j^f02^[]P00P0000000`2][=H000000:JXf0030:JXf0050000
002OYMX0WjGJ09nUfP0000000`2HXm`400000080U:7M00D0000008jQh@2>XN40SZ7Q0000000408JR
i@030000001oXnX000000140000000<0Ejci0000001E/_/00P1E/_/00`000000F;_l05Rko00205Rk
o0040000001OaO`0GlGl05o5o0@000000P1Zd_X50000000407;Lm00000000000077Ml@L0000000@0
Ln3/000000000000Mn7^1P00000207GMm@/0000000<0HlSl0000001K__`00`1K__`00`000000F;Ol
000000080000000306Zcn`000000LJkg00<0LJkg00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`0<0?ooo`00EP3oool00`000000oooo0000000200000300oooo00D00000086cm@21/oD0PK?e
0000000208J]k`050000002<Z>P0S:SX08bXj00000000`2@Y>800`000000UJ7M09FQg@0209FQg@05
0000002IWmP0VIoH09VOf00000000`2LW]D400000080WioD00<000000:BQdP2TXM800P2TXM800`00
0000Z:;@0:RRd0020:RRd08000000`2_Z=000`000000/j_@0;>[d0030;>[d0030000002k/M00^k7@
00<0^k7@00<000000;ned00000001032^=000`000000aK_@000000050<Rmd0030000003=`]00cL;@
0080cL;@00H000000<o4c`00000000000=35d00000040=76c`030000003Bam0000000080d/O@00<0
00000=;7d03Bam000P3Bam000`000000clKA0<o6d@020<o6d@030000003>aM80000000<0bl?C00D0
00000<W2d`39`]<000000<Fne002000000<0`kcD00<000000;fie@2m^MD00P0000000`2k]]D00000
0;FbeP030;FbeP8000000`2][=H00`000000YZSH0:JXf0020:JXf0040000002OYMX0WjGJ09nUfP@0
00000P2KX]X00`000000U:7M09BQg@0209BQg@050000002>XN40SZ7Q08jQh@0000000`26X^D00`00
0000Oj?Z07nSjP0207nSjP050000001hY^l0N:K_07RVk`000000101`ZO@01P000000JJki06V^n@1Y
[_T0000006>do1@000000P1[eOL80000000407COk`000000000007WRl0P0000000<0P^Cf00000000
000010000002083KoAH0000000<0FK3k0000001NZ_P0200000090?ooo`030000003oool0oooo00/0
oooo00180?ooo`060;md00000000000000000000B9`0gooo203oool2000003<0oooo00D00000086c
m@21/oD0PK?e0000000308J]k`050000002<Z>P0S:SX08bXj00000000`2@Y>8010000000UJ7M09FQ
g@2EXMd800000080VYkG00D0000009nOe02OWm@0WioD000000040:BQdP030000002XX]00Z:;@0080
000000<0ZZGB0000002_Z=000`2_Z=000`000000/j_@0;>[d002000000030;J^d@000000^k7@00@0
^k7@00<000000<:hd032^=001032^=000`000000b;g@0<Rmd0030<Rmd0030000003;`=0000000080
cL;@00<000000<o4c`3?a<l00P3?a<l00`000000dLK?0=76c`020=76c`060000003Bam00d/O@0000
003Bam000000103Bam000`000000dLSA000000020<o6d@070000003>aM80c/GB0<k5dP000000bl?C
000000040<W2d`0300000033_=@0`kcD00@0`kcD0P0000030;^fe@030000002e/]H0]K;F00800000
00<0/k3E0000002][=H00`2][=H010000000YZSH0:JXf02VZ=P400000080XJCH00D0000009^RfP2K
X]X0Vj;J0000000409BQg@050000002>XN40SZ7Q08jQh@0000000`26X^DO00000080J<Wm2`000000
101cg?@000000000001jh?@600000004083UlP000000000000000080T^Wh0`000000102:i?X00000
0000002Bho`400000080Q]Wo4@0000000`1B[oX00000056Wn0050000000305jZn0000000IJKe00<0
IJKe00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0;0?ooo`00B@3oool00`3ogi`0B000
07Boo`0;0?ooo`030000003oool0oooo0380oooo4`00000208jOg`050000002DW]/0U9kK09BNf`00
00000`2JW]L00`000000WioD09nOe00209nOe0040000002TXM80Y:7B0:BQdP<000000`2ZYM800`00
0000[jS@0:nXd0030:nXd08000000`2f[]400`000000^k7@0;^ad0020;^ad0030000002n]=000000
00@0`[S@00<000000<Fkd00000000P38_M0200000080bl3@00<000000<g2d0000000103?a<l00`00
0000d<G?000000020=76c`030000003Bam00d/O@0080d/O@00<000000=;7d03Bam000P3Bam001P00
0000dLSA0=78d@000000clKA000000D0c/GB00<000000<W2d`39`]<00P39`]<00`000000al3D0000
00040<>le00300000030^]@0000000@0^kKE0P0000030;>`e@030000002][=H0[JcF00<0[JcF0`00
00020:JWeP050000002QY=P0XJCH0:6Tf0000000102KX]X010000000U:7M09BQg@2DXMd800000080
SZCT00<0000008VWj@29YnT00P29YnT01@000000Pj__08>[k`23Znl0000000@0OK3d00D0000007Nf
n@1g]_T0MkKi00000003076lo@T000000P1cdOh50000000307SKnP000000000000L0000000<0SNSg
0000000000000`00000209;Yn080000000<0VnWj000000000000100000000`2IhOd0000000000003
00000080S=Co00@0000007S:o`1hb_l0N<[o200000000`1H/_/0000004nXn00500000003056Wn000
0000Dj;f00D0000000<0IJKe0000003oool01`3oool00`000000oooo0?ooo`0;0?ooo`00B@3oool0
0`3ogi`0B00007Boo`0<0?ooo`800000<`3oool200000080KZ_f00D0000007NWl@1gYo40MjOa0000
000307nTj`0500000027XND0Qj7U08NQi@0000000`2>Wml01@000000U9kK09BNf`2DW]/0000000@0
VYkG00@0000009nOe02OWm@0WioD100000000`2TXm@000000:ZUdP040:ZUdP030000002_Z=00[jS@
00<000001@2f[]4010000000^k7@0000000000030;jdd00400000032^=00000000000080aK_@00<0
00000<Rmd00000001@3;`=000`000000clC?0<o4c`020<o4c`060000003@aLl0d<G?0000003Aa/l0
0000103Bam001P000000d/S@000000000000d/O@000000@0dLSA00<000000<k5dP3>aM800`3>aM80
10000000cLGB00000039`]<2000000<0al3D00@000000<>le000000000000P30^]@00`000000^kKE
0;^fe@02000000030;Rde@000000/k3E00<0/k3E00<000000:f/eP2][=H00`0000040:JWeP040000
002QY=P0XJCH0:6Tf0T000000P2HX]/01@000000Tj;O09>Rg`2CX]l0000000@0SZCT00D0000008VW
j@29YnT0RJOY0000000308>[k`040000001m/?@0OK3d07f`m0T000000`1na?d010000000NL_o07W;
o`1ibol400000080OmSn100000001023h?X000000000002>iO/60000000409O[n000000000000:7/
n@P000000P2YiOX3000000<0W]cl2000000207g5oP060000001Z^Od0J[Wm06Zio@000000FZgj1@00
00000`1?Z?P0000004^PmP050000000305>RmP000000FYkc00<000001`3oool00`000000oooo0?oo
o`0;0?ooo`00B@3oool00`3ogi`0B00007Boo`0>0?ooo`030000003oool0oooo0380oooo0P000000
0`1^ZoH0000007NWl@0207NWl@050000001oY>/0OjC[07nTj`0000000`27XND01@000000SYoO08jO
g`2>Wml0000000<0U9kK00@0000009ZNe`2JW]L0VYkG100000000`2NXMH000000:BSe0030:BSe003
0000002ZYM80ZZGB0080ZZGB0`0000030;6[dP030000002f[]40]ZkA0080]ZkA0P0000060;jdd003
00000035^m00aK_@00<0aK_@00<000000<_0d03;`=000`3;`=001P000000cL;@000000000000clC?
000000@0d<G?00<000000=;7d03Bam000P3Bam000`000000d/S@0=;8d0020=;8d0030000003Ab=40
dLSA0080dLSA00D000000=78dP00000000000<k5dP02000000<0cLGB00<000000<O0e037`=@01037
`=@00`000000`;[D0<2je0030<2je08000000`2h]=D010000000/k3E0;>`e@2c/=D4000000<0[:[E
00<000000:JWeP2VYmH00P2VYmH3000000030::SeP000000WJ;I00@0WJ;I00D0000009RRf`2HX]/0
V:;K0000000309>Rg`040000002>Y>@0SZCT08jTi0d000000`29]_<01@000000Q;gi08Bmn@24_OT0
000000<0O/Cm1@00000208KBo`@000000P2<g_h60000000409WYnP00000000000:C[nPP000000P2a
j_P200000080ZNGj0P000000102_h?P0000009kLo02Ng?`2000000030:?FnP000000T<om0080T<om
3@0000000`1J[OX0000004jRmP050000000304^PmP000000Bi[c00D0000000<0FYkc0000003oool0
1@3oool00`000000oooo0?ooo`0;0?ooo`00B@3oool00`3ogi`0B00007Boo`0?0?ooo`030000003o
ool0oooo0380oooo6P0000000`2GWmX0000009jQeP0309jQeP050000002TXm@0Y:?D0:BSe0000000
0P2ZYM82000000030:bXd`000000/J_B00<0/J_B00<000000;J^d@2f[]400P0000020;Vad@040000
002n]=00_[C@0;jdd080000000<0`KSA00000035^m000P35^m000`000000b;g@000000030<_0d003
0000003=`]00cL;@0080cL;@00<000000=35c`3@aLl00P3@aLl01P000000d/O@000000000000d/O@
000000@0d/S@00<000000=?9d@0000000P3Ab=400`000000dLSB0=78dP020=78dP030000003=aM80
cLGB0080cLGB00<000000<_4e00000000`37`=@2000000030<Foe0000000`;[D0080`;[D0P000005
0;Rde@<000000`2a[]D01@000000[:[E0:bZe@2/Z]D000000080YZOF0`0000040::SeP030000002M
X]T0WJ;I0080WJ;I00@0000009RRf`2HX]/0V:;K2000000309FWhP050000002BZnL0TZ_W09:[i`00
00000`2>/>d010000000RKKc08Vfl`29]_<900000080S<_m1@0000000`2Cf?h00000000000030000
000409SSo@00000000000:CVn`<000000P2TjoX4000000030;KZmP00000000000080/N[h0P000000
102gi_H000000000000000030:oPn0<000000`2Se_X400000080U<Wj00@00000086on`21_o/0PKok
0P00000306jdnP030000001MZ?P0000000<0000000@0CZ;f00000017V_<0Ai[c10000000101;V_<0
0000052Gl@1@Uo43000000D0oooo00<000000?ooo`3oool02`3oool004T0oooo00<0omnL04P0001d
_ol0403oool00`000000oooo0?ooo`0b0?ooo`030000001HXOD0F:7e008000000P1UX?8010000000
LIk]076Nk@00000307bNj00500000026W^<0QYkS08JNh`0000000`2?W]h00`000000UioJ09NOfP02
09NOfP050000002NXMH0WZ7F09jQeP000000102TXm@2000000@0[:SC00<000000;6[dP2aZm800P2a
Zm82000000@0^K7A00@000000;jdd000000000000`31^=4010000000aK_@0000000000030<Rmd003
0000003;`=00000000@0cL;@00<000000<o5d00000000P3@aLl00`000000d/O@0=;7d0020=;7d004
0000003Bb=00d/S@0=;8d08000000P3CbM400`000000dLSA000000040=78dP030000003@b=<00000
0080cLGB00D000000<_4e03;a=@000000<O0e002000000<0aKoD00@000000<2je000000000000P2m
^=D01`000000^;CE0;Rde@2h]=D000000;Jbe@000000102a[]D010000000[:[E0:bZe@2/Z]D30000
00<0YZGE00D000000::SeP2RXmH0XZ?F0000000309fRf@D0000000<0W:?J0000002IY=d00`2IY=d0
0`000000UJOR09FWhP0209FWhP050000002BZnL0TZ_W09:[i`0000000`2>/>d4000000<0UK_a00@0
00000973n02A`oP0TL?h1000000309[An`<000000`2PgO`3000000030:CVn`00000000000080[^Oi
1@0000020;KZmP030000002lj?<00000008000000P2gi_H2000000030;_QlP000000000000800000
0P2bf_D400000080Y]3f600000000`17V_<0000004FDl@0400000080D9Oa00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0;0?ooo`00B@3oool00`3ogi`0B00007Boo`0A0?ooo`800000<`3o
ool60000000506FPlP000000LIk]076Nk@0000000`1lW^P01@000000QYkS08JNh`26W^<0000000<0
SikN00D0000009NOfP2GWmX0UioJ0000000409jQeP060000002TXm@000000000002UYMH00000102/
Z=<00`000000/J_B0;6[dP02000000030;B^d`000000^K7A00@0^K7A00<000000<6hd@31^=401031
^=400`000000b;g@0<Rmd0040<Rmd0040000003=`]00cL;@0<g2d08000000P3?aM000`000000d<G?
000000040=;7d0050000003BbM4000000=;8d00000001@3CbM4010000000dLSB0=78dP3Ab=820000
0080d<SC00<000000<g5dP000000103;a=@00`000000aKoD0<Foe0040<Foe0030000002m^=D0_KSE
0080_KSE00@000000;Rde@00000000000P2f/]D00`000000/JkE0;6^e@030;6^e@<000000P2[Z=@0
1@000000YZGE0:JUe@2VYMD0000000<0XZ?F1000000209nRe`030000002LXmX0W:?J0080W:?J00D0
000009VTg@2IY=d0VJCM0000000309FWhPT000000P2H]>/010000000UK_a09Fkl@2E^o4500000080
WLSf00<0000009[An`2JdO/00P000000102UeOP000000:3Mo02PgO`200000080[>3i0P0000020:kW
n@8000000P2gioH00`000000]^[f0;KZmP0200000080_>Sc00<000000<7Tl00000000P0000020;_Q
lP<000000P2bf_D2000000040;CDl0000000Y]3f0:K@mP80000000<0YlWa0000002H`_D00`2H`_D0
10000000QK[g08Fjm`25^_L30000000307>_m`000000HJCe00<0HJCe100000000`16UO4000000000
00030000000604FDl@00000000000000001@Uo4000001@3oool00`000000oooo0?ooo`0:0?ooo`00
B@3oool00`3ogi`0B00007Boo`0C0?ooo`030000003oool0oooo0340oooo0P00000204:FlPH00000
0`1lW^P01@000000QYkS08JNh`26W^<0000000<0SikN00@0000009NOfP2GWmX0UioJ1`0000030:FU
eP030000002/Z=<0[:SC00<0[:SC0P0000030;B^d`030000002i/M40^K7A0080^K7A00<000000;be
d`0000001031^=400`000000a;_B000000040<Rmd0050000003:`M4000000<g2d00000001@3?aM00
0`000000d/O@0=;7d0020=;7d0050000003BbM40d/WA0=;9d@0000001@3CbM401@000000d/_B0000
003Ab=80000000D0d<SC00<000000<_4e03;a=@00P3;a=@00`000000bLCE000000040<Foe0030000
0032_MH0000000@0_KSE00<000000;Jbe@2f/]D00`2f/]D00`000000/JkE0;6^e@03000000@0ZjSD
00@000000:JUe@2VYMD0YZGE100000020:>Se@030000002OX]L0Wj;G0080Wj;G00D0000009bSfP2L
XmX0W:?J0000000309VTg@@000000`2LZN001@000000VZkU09Z^i@2J[^D0000000<0V;C[10000003
0:2ok`040000002Mb?H0WLSf09g8mP<000000`2UeOP300000080[>3i0P000000102ehOH000000000
000000020;OWmP80000000<0_^Kb0000000000000P2lj?<200000080`NC`0P0000000`2khO800000
000000030;kKk`800000102de?02000000@0YlWa200000000`29]?8000000000000500000004066T
m@000000DY[b05:JlP@0000000<0AYGa00000012Snh01P0000050?ooo`030000003oool0oooo00X0
oooo00190?ooo`030?oOW0180000M;oo01@0oooo00<000000?ooo`3oool0<P3oool20000000404:F
lP000000DIOa056GlA4000000P2DXMd01@000000WJ?I09fSf@2MXmT0000000@0YJGF00<000000:bX
d`2/Z=<00P0000000`2][=H000000;B^d`030;B^d`040000002i/M400000000000<0_;GC00@00000
0<6hd@00000000000`34^m8010000000b;g@0000000000030<[1d@030000003?aM00clG@00<0clG@
00H000000=37d@00000000000=;7d00000030=;9d@030000003Cbm80000000<0dlWA00D000000=;;
dP3Bbm80d/_B000000050=38d`040000003>b=@000000<_4e08000000`39a=D010000000aKoD0000
000000020<:meP030000002m^=D0_KSE0080000000<0^[OF0000002f/]D00`2f/]D3000000<0[jcD
00<000000:^Xe02[Z=@00P2[Z=@3000000030:JTe0000000Xj?E00<0Xj?E00<0000009nRe`2OX]L0
0P2OX]L010000000W:?J09bSfP2LXmX400000080WZKK00<0000009bYh02LZN000P2LZN0010000000
VZkU09Z^i@2J[^D400000080XKSX00@000000:2ok`2P_nl0X;o_1@0000020:S<lP@000000P2`eo@3
00000080]N7f0P000000102mhO4000000000000000020;kVlP80000000<0a>C^0000000000000P31
i?0200000080aMo/00@000000;kKk`2nfnl0_]__0P000000102oeN/000000;CDl02de?0200000004
0;G=k0000000YlWa0:O9l@80000000<0Z<;/0000002J_?000P2J_?000`000000RKCb08VdlP0208Vd
lP060000001gZo<0Mj_c07N[l`000000IZ7b1@0000000`1BV_80000004NAk`0400000080@Xo^1000
00040?ooo`030000003oool0oooo00X0oooo00190?ooo`030?oOW0180000M;oo01D0oooo00<00000
0?ooo`3oool0<P3oool600000080HIS^00D00000072JjP1`V^X0L9[Z0000000307jLiP050000002:
W^40RYkQ08ZNh@0000000`2DXMd01@000000WJ?I09fSf@2MXmT0000000D0YJGF0P0000030:f/eP03
0000002d[]<0]:kC00<0]:kC00<000000;bed`2l]M<0102l]M<00`000000a;_B0<BkdP040<BkdP03
0000003:`M40b/7A0080b/7A00<000000<c4dP0000000`3?aM000`000000d<OA0=37d@020=37d@05
0000003BbM40d/WA0=;9d@0000000`3Cbm800`000000dlWA000000030=;;dP050000003Ac=@00000
0000003@b=<00P0000030<k8e00300000039a=D0bLCE00@0bLCE00<000000<:meP32_MH00`32_MH2
000000<0^[OF00<000000;Jbe@2f/]D00P0000000`2c/=D000000:n/e0030:n/e0030000002[Z=@0
ZjSD00<00000102VY=@00`000000Xj?E0:>Se@020:>Se@040000002OX]L0Wj;G09nRe`@000000P2P
Y=P00`000000WZKK09jVf`0209jVf`040000002LZN00W:WP09bYh0@000000P2R/N<00`000000XKSX
0:6hj0020:6hj0<000000`2Y`n`00`000000Z<cb0:S<lP02000000040;;>k`000000/=Od0;3Gm080
000000@0^MW`00000000000000000P2mhO42000000040<?Pk@000000000000000080a>C^0P000000
0`38h>X00000000000020<GOk080000000@0a]_X0000000000000000102oeN/2000000<0]Lg/0`00
00030:S2k0040000002J_?00V[c`09Zll0d0000000<0IZ7b0000001FUo00100000000`17TNl00000
046<k@0300000080@hc]0P0000030?ooo`030000003oool0oooo00X0oooo00170?ooo`050?oog`2L
B00000000000001d_ol05P3oool2000003@0oooo00<0000002b9k00000000`0000001@1QV>h00000
072JjP1`V^X0000000<0OYcV00D0000008ZNh@2:W^40RYkQ0000000409BQg@050000002MXmT0WJ?I
09fSf@0000000P2UYMH2000000030:FXf@000000[JcF00<0[JcF00<000000;B^d`2d[]<00P000000
0`2f/]D000000;bed`030;bed`8000001@34^m800`000000a[oC000000020<[1d@070000003<a=80
c<CB0<c4dP000000clG@000000040=37d@030000003Bb]8000000080d/WA00<000000=?;dP3Cbm80
0P3Cbm801@000000d/_B0=;;dP3Bbm80000000@0dLcD00<000000<k8e03>b=@00P3>b=@00`000000
c<SG000000040<W4e@0300000036`mL0000000<0`[gF00<000000;jle`000000102j]mH2000000@0
/k3E00<000000:n/e02_[=@00P2_[=@2000000<0ZJOD00D000000:JTe02VY=@0YZCD000000030:>S
e@D0000000<0XZ?F0000002PY=P00`2PY=P010000000WZKK09jVf`2NY]/500000080XjcN00D00000
0::ah`2R/N<0X[7S000000020:6hj080000000<0ZK_V0000002Y`n`00`2Y`n`3000000@0//k_0P00
00020;WIl080000000@0`=W/00000000000000000`33h>d00`000000b=kY0000000200000080b>3Z
0P0000000`3;g>H00000000000030<KKj00300000037eN@000000080_mG[0P0000000`2ocnH00000
0000000200000080]LKW00@000000:S2k02X`^`0Z<;/100000060?ooo`030000003oool0oooo00@0
oooo100000000`1FUo00000004Z?kP0400000003046<k@00000000000080000000D0@hc]0000003o
ool0oooo0000000<0?ooo`00B@3oool00`3o_g@0000007Boo`0H0?ooo`030000003oool0oooo0380
oooo0P00000203f=kPD0000000<0L9[Z0000001nW>H00P1nW>H01@000000RYkQ08ZNh@2:W^400000
00<0U:7M1`0000030:FXf@030000002][=H0[JcF00<0[JcF0P0000040;Jbe@040000002l]M<00000
00000080_kWD00<000000<BkdP34^m800P0000020<Jod`030000003:`M40000000D0c<CB00@00000
0=37d@3@am40d<OA0P0000020=;:dP030000003BbM40000000@0dl_B00D000000=?=e0000000d/_B
000000040=7<e0030000003@cMH000000080c/SD00H000000<c8e`3<b=L0c<SG00000039a=D20000
0080a/?G00@000000<:meP00000000000P2n_=L00`000000^[OF0;ZgeP02000000030;Nee`000000
/k3E00<0/k3E00<000000:n/e02_[=@00`0000040:VWe0040000002VY=@0YZCD0:JTe0@000000P2T
Xm@00`000000XZ?F0::SeP020::SeP030000002PY=P0X:CH0080X:CH0`0000020:>XfP030000002S
[=h0XjcN0080XjcN00@000000::ah`2R/N<0X[7S0`0000030:VkiP030000002Y`n`0ZL?/00800000
0P2ba^T00`000000//k_0;;>k`02000000040;[@j`0000000000000000<0`=W/00@000000<GIj000
0000`n3]0P0000020<SNj@80000000<0b>3Z0000000000000P3;g>H010000000c=SS00000036fnP2
000000<0amGT0P0000030;o?iP800000102ea^L4000000/0oooo00<000000?ooo`3oool01`3oool5
0000000304Z?kP0000000000008000000P10RN/3000000030?ooo`000000oooo00/0oooo001U0?oo
o`030000003oool0oooo0340oooo00H0000001QliP000000000003f=kP000002052BkPh0000000<0
TJ;P0000002LY]`00`2LY]`00`000000YJSI0:FXf@020:FXf@030000002][=H0[JcF00<000001P2f
/]D00`000000_kWD0;nie0020;nie00300000034^m80000000D0a[oC00<000000<c4dP3<a=800`3<
a=801@000000c/OC0000003@am40000000D0d/[B00<000000=?;dP3Cbm800P3Cbm801@000000dlgD
0=?=e03CcM@0000000@0dLcD00H000000=3=eP3@cMH000000<k8e00000050<c8e`0300000036`mL0
a/?G00<0a/?G00<000000;jle`2n_=L00`2n_=L2000000@0]kGG00@000000;>`e@2c/=D0/k3E0`00
00030:fZe0030000002YYm@0ZJOD0080ZJOD0`0000020:JTd`030000002TXm@0Y:?D0080Y:?D00D0
00000::SeP2RXmH0XZ?F000000020:2Tf0<000001@2SZ=X00`000000XjcN0:>/gP020:>/gP<00000
0P2Y]>4010000000ZK_V0:VkiP2Y^nH3000000@0//KY0`0000030;[@j`0400000030dNL000000<3I
k08000000`35fNP00`000000b=kY0<SNj@0200000080c=_U00<000000<_LiP3;g>H00P0000030<cH
h`0300000037eN@0amGT0080amGT00@000000<O@h@000000_loV0P0000020;o9hP040000002ea^L0
000000000100oooo00<000000?ooo`3oool02`3oool700000080@8W[00<000000?ooo`000000303o
ool006H0oooo00<000000?ooo`3oool0<@3oool700000005052BkP000000HiO/06>Gk00000000`1e
VnP010000000Q9oU08BOi@000004096Rh0050000002LY]`0W:KL09bVg00000000`2UZ=T4000000<0
[joH00@000000;Jbe@2f/]D0][;E0P0000000`2h]mL000000;nie0030;nie00300000036_m<0a[oC
00<0a[oC00<000000<S3e@0000000`3<a=801@000000c/OC0<k7d`3>am<0000000D0d/[B00H00000
0=;=e000000000000=?;dP0000030=?=e0060000003BcmH000000000003Ac=@00000103@cMH00`00
0000c<SG0<c8e`020<c8e`80000000@0bLSI00000036`mL0a/?G0P0000000`32`]T000000;jle`02
0;jle`030000002k^]T0000000D0]kGG0`0000020;2^eP050000002]Z]@0[J[D0:fZe00000000P2Y
Ym@3000000@0YZCC00<000000:BSe02TXm@00P2TXm@010000000XZ?F0::SeP2RXmH3000000<0Y:GG
00<000000:>XfP2SZ=X00P2SZ=X00`000000XjcN0:>/gP02000000030:R^g0000000ZKCQ00<0ZKCQ
0`0000020;6mh`030000002ba^T0//KY008000000P2ianD010000000^]3[0000000000030<3Ai`03
00000035fNP0aMWX0080aMWX00@000000<[Gi000000000000`3<fnD010000000c]SQ000000000004
0<cHh`040000003=e=l000000<OEi08000000`37d>400`000000_lWR0;o9hP020;o9hP8000004P3o
ool00`000000oooo0?ooo`0?0?ooo`D0000000<0@8W[000000000000303oool006L0oooo00<00000
0?ooo`3oool0<@3oool40000000302V3j@000000000000<0000000D0HiO/0000001eVnP0MI_X0000
000308BOi@050000002AX^00TJ;P096Rh00000000`2LY]`400000080YZcL00<000000:n_f02_[mP0
0P2_[mP010000000][;E0000000000030;Rge`030000002o^M@0_kWD0080000000@0`KgF00000000
0000a[oC0P0000030<S3e@030000003<a=80000000<0c/OC00<000000<o;e@0000000`3Bb]800`00
0000d/gD0=;=e0020=;=e0050000003CcM@0dlgD0=?=e0000000103BcmH00`000000d<gF0=3=eP02
0=3=eP040000003>c]T000000<c8e`8000001039b=T2000000<0`/;I00@000000;jle`0000000000
0`2k^]T00`000000]kGG0;Nee`03000000@0/:kF00@000000:fZe02]Z]@0[J[D0`0000030:RVe003
0000002VY=<0YZCC0080YZCC00@000000:BSe02TXm@0Y:?D100000020:BTe@050000002TYML0Y:GG
0:BUe`0000000`2SZ=X4000000<0Z:kL00<000000:Vdh@2Y]>400P0000000`2`]mh000000;6mh`03
0;6mh`800000102ianD2000000@0`=7W00@000000<KAh`000000aMWX0P0000001@3:en@000000000
00000000c=_U008000000P3>f>4010000000d=GM0000003<f><2000000<0cMCO00<000000<O@h@37
d>400P37d>4200000080_lWR0P00000E0?ooo`030000003oool0oooo0100oooo1@0000000`3oool0
00000?ooo`0:0?ooo`00J03oool2000003<0oooo00<0000000IagP0000000P00000203n:k0P00000
00<0Q9oU0000002AX^000P2AX^0400000080VjWP00<000000:J/g02V[=`00P2V[=`00`000000[joH
0:n_f0020:n_f0030000002h]mL0^;OG00@0^;OG0P0000050<6meP0300000038`mD0b<?E00@0b<?E
00D000000<k7d`3>am<0c/OC000000030<o;e@030000003Bb]80000000@0d/gD00D000000=;@eP00
0000dlgD000000040=;?eP040000003AdMT000000=3=eP8000000`3>c]T00`000000bLSI0<W8f@03
0<W8f@0300000036b=`0000000@0`/;I00<000000;^jf@2k^]T00`2k^]T3000000<0/k?H00<00000
0;2^eP2`[]H00P2`[]H300000080ZZSE00D000000:RVe02XY]@0Z:KD000000030:JTd`D0000000<0
YJ?D0000002TY=D00`2TY=D010000000Y:GG0:BUe`2TYML4000000<0Z:[H00<000000:R^g02X[]`0
0P2X[]`2000000<0/;ON00<000000;6mh`2a_N<00P000000102h`>0000000;W7i@2ianD200000080
_lWR00@000000<3Ai`00000000000`36dN<010000000b]OT0000000000030<gFh0050000003>f>40
c]SQ0<kHh@0000000`3@eMd00`000000cMCO0<gDg`020<gDg`040000003=d=`000000<O@h@800000
0P36bmd2000001P0oooo00<000000?ooo`3oool04P3oool4000000/0oooo001Z0?ooo`030000003o
ool0oooo0340oooo1`00000305JBk@030000001[V^`0Ji[/00H000000P2>YN@01@000000VjWP09^Y
h02KZN00000000@0YZcL00<000000:n_f02_[mP00P0000000`2`]=/000000;Rge`030;Rge`030000
002k^mX0000000@0`KgF00<000000<?2f@0000000`38`mD2000000040<_7e`000000c/OC000000D0
cl_E00@000000=;=e03BcM@0d/gD0P0000030=;@eP040000003BcmH0d/oF0=;?eP8000000`3AdMT0
0`000000c/kI0<k>f@020<k>f@050000003<c]`0000000000039b=T00P0000020<K8g00300000032
`]T0`/;I0080000000@0_/7L0000002k^]T0^k[I0P0000020;Nif`050000002c/mP0/k?H0;>cf000
00000P2`[]H3000000@0ZZSE00@000000:RVe02XY]@0Z:KD100000020:JSd`030000002UXm@0YJ?D
0080YJ?D00<000000:BTe@2TY=D00P2TY=D300000080YjOF00<000000:RZf02XZ]P00P2XZ]P00`00
0000Z:kL0:R^g002000000030:jafP000000/;ON00<0/;ON0P0000040;S0h0800000102obN820000
00<0a]7S0P0000000`3:dMl000000<gFh0030<gFh0050000003@e=`000000<kHh@0000000`3@eMd0
10000000d=;K0000003=e=l2000000<0cM3L00D000000<K;g@36bmd0a/_M0000000J0?ooo`030000
003oool0oooo01@0oooo0P00000;0?ooo`00J`3oool00`000000oooo0?ooo`0b0?ooo`L0000000D0
EY;]0000001[V^`0Ji[/0000000307jPj0050000002>YN@0SZGT08jUi0000000102KZN0010000000
YZcL0:J/g02V[=`3000000@0/;CK00@000000;Rge`00000000000`2k^mX010000000`KgF00000000
00020<?2f@0400000038`mD00000000000@0blOG00<000000<o;e@3?bmD00`3?bmD01@000000d<oH
0000003BcM@0000000@0d]3F00D000000=;Bf@000000d/oF000000040=7Af@040000003@dm`00000
0<k>f@800000103<c]`00`000000a/SL0<K8g0030<K8g0800000102n`M`2000000@0]kWK00@00000
0;>cf02c/mP0/k?H0`0000030:f/e`030000002ZZ=D0ZZSE0080ZZSE0`0000020:NUe0030000002V
Xm<0YZ?C0080YZ?C00D000000:FSe02UXm@0YJ?D000000020:BTe@<00000102WYmH010000000Z:[H
0:RZf02XZ]P4000000<0[[7J00<000000;2ggP2`]mh00P0000000`2e^M`000000;S0h0020;S0h003
0000002m`]d000000080_lWR0P0000000`34b]h000000<KAh`0200000080b]7O00@000000<g@g000
0000cMKP0P0000030=3Dg0030000003@eMd0d=GM008000000`3@d]/00`000000cM3L0<g@g0020<g@
g080000000<0a/_M000000000000703oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo
00T0oooo001/0?ooo`030000003oool0oooo0380oooo100000000`0^PnT000000000000300000003
06^Jk0000000OZ3X0080OZ3X00D0000008jUi02>YN@0SZGT0000000309^Yh0@000000P2V/Ml00`00
0000/;CK0;2df`030;2df`030000002k^mX0^k_J00@0^k_J00<000000<?2f@33`]T00`33`]T00`00
0000blOG0<_7e`030<_7e`050000003<c=T000000000003?bmD00P0000030=3?f0030000003Bd=H0
d]3F0080d]3F00D000000=;Bf@3Bd]T0d];I000000040=7Af@050000003@dm`0d=?L0=3Cg0000000
1@3<c]`00`000000b<oO000000030<K8g00300000031b=l0000000<0_/7L00<000000;VogP000000
1@2g^M/300000080[k7I00D000000:f/e`2][=L0[JcG000000020:ZXe@<00000102WYM@00`000000
YZ?C0:JSd`020:JSd`040000002UXm@0YJ?D0:FSe0<000000`2VYM@00`000000YjOF0:NWeP020:NW
eP<000000`2/[=L00`000000[[7J0:jafP020:jafP8000000`2e^M`010000000^<3P000000000003
0;g2g@8000001034b]h01@000000b]7O0<[Ag`3:dMl0000000<0cM3L00<000000=3Dg03@e=`00P3@
e=`010000000dM;J0000000000040=3Bf`040000003@cmP000000<g@g08000000P3<bmT00`000000
oooo0?ooo`0M0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool02@3oool006d0oooo
0P00000b0?ooo`0400000000HM<0000000Q`gP<000000P18S^d>00000080VZgT00<000000:Jag`2V
/Ml00P2V/Ml00`000000/;CK0;2df`02000000030;:igP000000^k_J0080^k_J0P0000000`2l`Md0
00000<?2f@020<?2f@0500000035am/000000000003;amL00P0000040<c<f@030000003@cmP0d<oH
0080d<oH00H000000=7BfP00000000000=;@eP0000030=;Bf@060000003Ae]`000000000003AdMT0
00000`3@dm`01@000000cMGO000000000000c<kL008000000P38cml010000000a/SL000000000002
0<78g`040000002n`M`00000000000<0^KoN00<000000;Nif`2g^M/00`0000040:naf@040000002]
[=L0[JcG0:f/e`<000000`2XYmH00`000000YjGD0:NUe0020:NUe0040000002VXm<0YZ?C0:JSd`@0
00000P2VY=<01@000000YZGD0:JUe02VYM@000000080YjOF0P0000000`2[ZMD000000:b/e`030:b/
e`030000002^/MX0[[7J0080000000<0/k?I0000002e^M`00`2e^M`00`000000_L;M0;g2g@020;g2
g@0300000032`m/0000000<0a<[N00D000000<S;f`000000b]7O000000030<g@g0040000003@cmT0
00000=3Dg08000000P3Ad]X010000000d]3G0000003@d]/2000000<0d<oH00D000000<c;f@3<bmT0
c<_I0000000Q0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool02@3oool006l0oooo
00<000000?ooo`3oool0<03oool00`000000067C000000050000000504R>k@000000HIS^066HkP00
00000`1gXN`01@000000RZSX08ZXj02:Z>P0000000@0VZgT00@000000:Jag`2V/Ml0Y[7O0`000003
0;:igP040000002k^mX00000000000<0_<7M00@000000<?2f@00000000001035am/00`000000c<cI
0<c<f@030<c<f@040000003=d=`000000=3?f0800000103Ad]X00`000000d];I0=;Bf@02000000@0
dMKL00<000000=3Cg03@dm`00P0000040<gEg`0300000038cml0b<oO00<0b<oO00<000000<78g`31
b=l00`31b=l00`000000^KoN0;VogP040;VogP8000000`2b]md00`000000[k7I0:naf@020:naf@<0
00000P2YZmP01@000000Z:OF0:RWeP2XYmH0000000<0YjGD1@0000000`2VY=<000000:JTd`030:JT
d`040000002VYM@0YZGD0:JUe0<000000`2[ZMD00`000000[:cG0:b/e`030:b/e`8000000`2c/mT0
0`000000]KWL0;Fig002000000030;ZkfP000000_L;M008000000P32`m/010000000a<[N00000000
00030<S;f`050000003=d=`0cM3L0<g@g00000000`3@cmT01@000000dM;J0=7BfP3Ad]X0000000<0
d]3G00<000000=3?f03@cmP00P3@cmP2000000030<c;f@00000000000280oooo00<000000?ooo`3o
ool04P3oool00`000000oooo0?ooo`090?ooo`00L03oool00`000000oooo0?ooo`0`0?ooo`/00000
00<0HIS^0000001gXN`00P1gXN`01@000000RZSX08ZXj02:Z>P0000000<0VZgT100000020:Nfh`03
0000002b^Mh0/[WN0080/[WN00<000000;c1g@2l`Md0102l`Md00`000000aLOK0<G7f`030<G7f`03
00000037c=h000000080c<cI0P0000030<g@g0030000003Ad]X0dM;J00<0dM;J00@000000=7Fg@00
000000001@3Ae]`010000000d=SO0000000000040<gEg`80000000@0bMKS00000038cml0b<oO0P00
00001033cn<000000<78g`31b=l2000000030;c7h`00000000000080^KoN0P0000000`2d_^400000
0;:gg@020;:gg@030000002_/MT0[k7I00<00000102YZmP010000000Z:OF0:RWeP2XYmH400000080
YZCD00<000000:JTd`2VY=<00P2VY=<00`000000YZCC0:JTd`020:JTd`<0000000<0ZJKD0000002[
ZMD0102[ZMD00`000000[:cG0:b/e`02000000030;2_eP000000/k?I00<0/k?I0P0000040;ZkfP03
00000032`m/0`/?K00<0`/?K00<000000<S;f`38bm/00P38bm/01@000000c<_I0000003=d=`00000
00<0d<oI00D000000=7>e`000000dM;J000000030=;@e`040000003Bc]D000000=3?f08000000P3@
c=H00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0203o
ool00740oooo00<000000?ooo`3oool0<@3oool70000000303Z:k@000000000000`000000P2I/NP0
1@000000YkKS0:Nfh`2W]^<000000080/[WN0P0000000`2c_^8000000;c1g@030;c1g@050000002n
a^00000000000035am/00P0000020<O<gP030000003<c=T0000000@0cM3L00D000000<kEg`000000
00000=7BfP0200000080dMKM00D000000=3Jh000000000000=7Fg00200000080d=SO00@000000<gK
i0000000cMGO0P0000040<WFh`8000001033cn<2000000@0_<OS0`0000030;Bnh@040000002b]md0
/[OM0;:gg@<000000`2[[m/00`000000ZJ_H0:V[f0020:V[f0<000000P2VY]H00`000000YZCD0:JT
e0020:JTe0050000002VY=<0YZCC0:JTd`0000000P2VY=<3000000@0ZJKD00<000000:^Ye@2[ZMD0
0P2[ZMD3000000<0/:oF00<000000;>cf@2c/mT00P0000000`2g]]L000000;ZkfP020;ZkfP030000
002o_MP0000000<0`/?K00@000000<K5f@000000b<_K0P0000030<c;f@050000003@cmT0d<oI0=3?
f@0000000`3Ac]L01@000000d]3G0=;@e`3Bd=L0000000<0d/kE00D000000=3<eP3@c=H0d<cF0000
000W0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0203oool00780oooo0P00000b
0?ooo`D0000000<04WKR0000000000000`1FUo001@000000L:;`072Rl01`X_00000000<0QZ_]00D0
000009Vaj02I/NP0VK7X000000030:Nfh`<00000102c_^8010000000_<7M0000000000040;k6h003
00000037c=h0alcN00<0alcN00<000000<g@g03=d=`00P3=d=`00`000000c]GO0<kEg`020<kEg`05
0000003Ae]d0dMKM0=7Fg@000000103@f^000`000000d=SO0=3Hg`02000000<0cM_T00<000000<WF
h`39e^<00`39e^<00`000000a=SX000000020<??h`80000000@0_/oW0000002lan<0_<OS0P000002
0;K6iP030000002d_^40];kQ0080];kQ0`0000020:beg`050000002[[m/0ZjoK0:^_f`0000000P2Y
ZmP3000000@0YZKF00<000000:JTe02VY=@00P2VY=@010000000YZCC0:JTd`2VY=<3000000<0YjGC
00D000000:VVe02YY]@0ZJKD000000020:^Ye@8000000P2^Zm@00`000000/:oF0;2_eP020;2_eP80
00000`2g]]L010000000^[_J0000000000020;nmf00400000032`m/00000000000<0a/GI00<00000
0<c;f@3<bmT00P3<bmT01@000000c/_G0000003@cmT0000000<0dLkG00D000000=;=e0000000d]3G
000000030=;>e@80000000<0d<cF000000000000:@3oool00`000000oooo0?ooo`0@0?ooo`030000
003oool0oooo00P0oooo001d0?ooo`030000003oool0oooo0340oooo0P0000001000IMH000000000
000]Q^/20000000305JGl0000000L:;`0080L:;`00D0000008J[k@26Znd0QZ_]0000000309Vaj0@0
00000P2W^nP2000000<0/kkR0P0000050;k6h0030000002oc>@0000000<0alcN00H000000<SAh@00
000000000<g@g00000030<kEg`80000000@0c][R0000003Ae]d00000103@f^0010000000c]kT0000
000000040<gKi00500000039gNP0000000000039e^<00P0000030<CHj0800000102ncnL2000000@0
]/KV00<000000;Bnh@2d_^400P0000000`2]_>@000000:beg`020:beg`040000002[[m/0ZjoK0:^_
f`<000000`2VZMT00`000000YZKF0:JVeP020:JVeP040000002VY=@0YZCD0:JTe0@000000P2VY=@0
1@000000YjGC0:NUd`2WYM<0000000<0ZJKD0`0000040:j[e0030000002`[mH0/:oF0080000000<0
];7E0000002g]]L00`2g]]L00`000000_kgH0;nmf0030;nmf00300000036aMT0a/GI0080a/GI00@0
00000<W6e`000000c<_I0P0000030<k;e`050000003Ac]L0dLkG0=7>e`0000000`3BcM@01@000000
d/kE0=;>e@3Bc]D000000080d/_C00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0@0?oo
o`030000003oool0oooo00P0oooo001e0?ooo`030000003oool0oooo0340oooo0`0000000`07L]l0
00000000000>00000080UkK]00<000000:Nkj02W^nP00`2W^nP300000080]<CV0P0000000`2na^00
0000000000020;o<i00400000037c=h00000000000@0b=7Q00<000000<kEg`3>eMl00P0000040<kJ
hP8000000P3@f^0200000080c]kT00@000000<_Qj0000000cM_T0P0000040<WMj00400000034f>P0
a=SX0<CHj080000000@0_]S]0000002ncnL0_/oW0P0000000`2gcn`00000000000030;K6iP<00000
0`2]_>@010000000[;GO0:beg`2/]Ml400000080YZgM00D000000:JYf@2VZMT0YZWI000000030:JV
ePD0000000<0Y:CE0000002VY=@00`2VY=@010000000YjGC0:NUd`2WYM<400000080ZjSD00<00000
0:j[e02^Zm@00P2^Zm@3000000<0];7E00<000000;Nfe`2g]]L00P0000000`2l^=H000000;nmf002
0;nmf00400000033_mL000000<K5f@8000000`39a]L00`000000c/_G0<k;e`020<k;e`050000003@
bm@000000=7>e`0000000`3BcM@01@000000dlcC0000003Bc]D000000080d/_C00<000000?ooo`3o
ool0:`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00P0oooo001f0?ooo`030000
003oool0oooo0340oooo2`00000206VTl`0400000022[o40PZoa08:_l@8000000`2G]^d00`000000
Yk_X0:Nkj002000000030:O1k@000000]<CV00<0]<CV0P0000050;o<i08000000`38dN4200000003
0<SGi@000000000000D0c][R00<000000<gNiP3=g^H00P0000020<kNi08000000`3;hNP2000000<0
bMgX0P0000001034gnd000000000000000040;kHk@8000001@2gcn`300000080[/CZ00@000000:fl
i02]_>@0[KcT100000020:JchP050000002V[Md0YZgM0:J]g@0000000`2VZMT400000080XjKG00<0
00000:BTe@2TY=D00P2TY=D00`000000YZCD0:JTe0020:JTe0<0000000<0Z:KD0000002[Z=@00`2[
Z=@00`000000[Z_D0:j[e00200000080/JgD00<000000;Bae@2d/MD00P2d/MD2000000<0_;SF00@0
00000;nmf000000000000`33_mL00`000000bLKG0<W6e`020<W6e`040000003<amD000000<k;e`80
00000`3@bm@01@000000d/gD0=;=e03BcM@0000000<0dlcC00@000000=;;d`3Bbm<00000;`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00P0oooo001g0?ooo`800000<P3oool60000
000302:3j@000000000000<0000000<0JJCc00000022[o400P22[o47000000<0Yl7]00<000000;C4
iP2da>H00P0000020;G;j`030000002oc>@0_lcT008000000P30d^P3000000<0b=OU00D000000<SL
j000000000000<kJhP0200000080cMkV0P0000000`3:hnT00000000000030<_Qj08000000P36i>d3
00000080a=o]0P000000102lhO4000000;kHk@2nf>d2000000030;KHl@00000000000080]lo/0P00
00000`2^cO0000000:k4jP020:k4jP@000000P2V^^L01@000000Y[?R0:JchP2V/n80000000<0YZgM
100000020::Xf`030000002SY]L0XjKG0080XjKG00D000000:BTe@2TY=D0Y:CE000000020:JTe0<0
0000102XY]@00`000000ZjSD0:^Xe0020:^Xe0800000102a[M@00`000000];7E0;Bae@0200000003
0;Rce@000000_;SF00<0_;SF00<000000<>oe`33_mL00P33_mL00`000000a/7E000000020<W6e`05
0000003<amD0c<OE0<c7e@000000103@bm@01@000000d/_C0000003BcM@0000000<0dlcC0P000000
0`3Bbm<000000?ooo`0`0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0203oool0
07T0oooo00<000000?ooo`3oool0<03oool800000080@IKb1P00000207fcmP050000002E_?80UKcb
09FllP0000000`2W`Nd3000000D0]L_[0P0000040<3Bj00400000030f>`000000<SGi@8000001038
g>P00`000000cMkV0<gNiP02000000<0b^?Y00<000000<KWk@0000000P0000030<KTk@80000000@0
`>Oa00000000000000000`2lhO43000000@0]]Sa0`0000030:k=l0@000000P2U`Nh01@000000Y[[W
0:Jji`2V^^L0000000<0Y[?R100000020:6/g`050000002RZ=/0XZSK0::Xf`000000102SY]L01000
0000Y:CE0:BTe@2TY=D3000000<0YZGD00D000000:RVe02XY]@0Z:KD000000020:^Xe080000000<0
[J[D0000002a[M@0102a[M@2000000<0^;?E00<000000;bheP2l^=H00P0000000`2o^]D000000<>o
e`0200000080a/7E00<000000<W6e`0000000`3<amD010000000clSC0000003@bm@2000000<0d/_C
00D000000=?<d`3Cc=<0dlcC000000020=?;dP030000003oool0oooo0300oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0?ooo`080?ooo`00NP3oool00`000000oooo0?ooo`0`0?ooo`h00000
0`1m/oH010000000UKcb09FllP2E_?8400000080Y/Sa0P0000020;G;j`80000000@0]M7_00000000
0000`=;X0P0000030<3Hk08000000`38g>P2000000040<KRk0000000000000000080b^?Y0P000002
0<KWk@0300000031j_40`N[a00<000000P30io42000000030;SXm@0000000000008000000`2ch?H4
00000080[=Kf100000020:?:m0050000002U`Nh0YL7^0:G1kP0000000`2V^^L400000080Wk3T00<0
00000:6/g`2Q[=l00P2Q[=l01@000000XZSK0::Xf`2RZ=/0000000<0XjKG100000020:>UeP050000
002VYM@0YZGD0:JUe00000000`2XY]@3000000<0[J[D00<000000;6]e02a[M@00P2a[M@2000000D0
^;?E0P0000040;nje@0300000036`MD0a/7E00<0a/7E00D000000<c7e@3<amD0c<OE000000030<o8
d`030000003Bbm<0d/_C0080d/_C00D000000=?;dP000000dlcC000000020=?;dP030000003oool0
oooo0380oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`070?ooo`00N`3oool00`00
0000oooo0?ooo`0`0?ooo`P000000`0iU_8900000080T/;g0P0000030:K8l@<000001@2edNl01000
0000`=S/0<3Hk030f>`200000080_mk_0`0000020<KRk08000000P33inl200000080a^O]0P000003
0<7Zl@030000002jk?D00000008000000P2hj?D200000080[NOi100000020:ONnP@000000P2OdoX0
10000000Xl[d0:?:m02Sb_@800000080WKOZ00D0000009n`i02O/>@0Wk3T000000030:6/g`T00000
0P2QYMT01@000000XjGF0:>UeP2SYMH0000000<0YZGD100000020:ZWe0030000002]Z]@0[J[D0080
[J[D00@000000;6]e000000000000P2d/=@01`000000^;?E0;Rce@2h/mD000000;^fe00000000`2o
^]D00`000000`kgD000000030<K1e@0500000039`m@000000<c7e@0000000`3?b=<00`000000d<SB
000000020=;;d`050000003Cbm80dl_B0=?;dP0000000P3Cbm800`000000oooo0?ooo`0d0?ooo`03
0000003oool0oooo00`0oooo00<000000?ooo`3oool01`3oool007`0oooo0P00000a0?ooo`l00000
0P1h]oX2000000<0T/;g100000020:G>mP8000000P2edNl2000000040;CHl`0000000000000000<0
_mk_0P000000102li?8000000000000000030<?Wk`80000000@0_^_b00000000000000000P31j_42
000000040;[/m@00000000000;3]n0<000000`2]ioT2000000040:3Uo0000000Ymkj0:ONnP800000
00<0VM_m0000002OdoX00P2OdoX400000080U/Kg00D0000009Znl@2J__40V[ka0000000309fgjP04
0000002O/>@0Wk3T09n`i0@000000P2KZ^400`000000WZOL09jWg00209jWg0050000002QYMT0XJGI
0:6Uf@0000000`2SYMH4000000030:JVeP000000ZZOD00<0ZZOD00@000000:fZe02]Z]@0[J[D0`00
00040;B`e0040000002h/mD0000000000080^kKD00@000000;nje@00000000000P33_M@010000000
a/7E0000000000030<W3e0050000003?b=<0clSC0<o8d`0000000P3@b=800`000000d/_C00000003
0=?;dP80000000<0dl_B0000003oool0=P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo00L0oooo001n0?ooo`030000003oool0oooo02l0oooo5@00000208o7n`8000000`2Uc_H30000
0080]=Sc0P0000020;;NmP<000000P2li?8200000080^>We0`0000020;k[lP8000000P2hk_D50000
0080/>gh0P0000020:C[nP<000000`2PiO`3000000<0VM_m10000002097>o@040000002Fa_L0U/Kg
09K6m``000000P2G[^H00`000000Vj[Q09^Zh@0209^Zh@050000002NYm`0WZOL09jWg00000000`2Q
YMT4000000030::Uf0000000YZKF00<0YZKF00<000000:ZWe02ZYm@00P2ZYm@3000000<0/:cD00<0
00000;B`e02d/=@00P2d/=@00`000000^kKD0;^fe0030;^fe00300000033_M@0`kgD00<0`kgD00<0
00000<W3e039`m@00P39`m@01@000000c<GC0000003?b=<0000000@0d<SB00D000000=?;dP3Cbm80
dl_B000000020=?;d@030000003oool0oooo03L0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`070?ooo`00O`3oool00`000000oooo0?ooo`0`0?ooo`l000000`1c_?`800000080X]Ci
0`0000030;;NmP80000000@0[^Ci00000000000000000`2hjOD2000000040;;/m`00000000000000
00<0^>ke100000000`2Tk_P000000000000500000080U^Sk2P00000208WEo`P000000P28`?X01@00
0000S[Wc08jil`2>^O<0000000<0Tk?]00<0000009N^iP2G[^H00P2G[^H01@000000Vj[Q09^Zh@2K
Z^40000000<0WZOL100000000`2NYM/000000::Uf0030::Uf0050000002VY]H0YZKF0:JVeP000000
0P2ZYm@3000000D0/:cD00<000000;B`e02d/=@00P0000000`2g/]@000000;^fe0020;^fe0030000
002n^M@0000000<0`kgD00@000000<Joe0000000bL?D0P0000030<c5d`030000003@b=80d<SB0080
d<SB00D000000=;9dP000000dl_B000000020=?;d@030000003oool0oooo03T0oooo00<000000?oo
o`3oool02P3oool00`000000oooo0?ooo`070?ooo`00P03oool00`000000oooo0?ooo`0`0?ooo`X0
000000<0;IW`0000000000001000000206g0oP040000002;cOh0Rlgn08_=oP@000000`2Of_`30000
0080[^Ci0P0000020:WXnP<000000P2bk?L200000080Znkh2000000209S/m`80000000<0RnOf0000
000000001000000207OKnPP000000P1hcOl01@000000P<On0837oP20aoh0000000<0R<3j00@00000
08jil`2>^O<0S[Wc4@0000000`2JY]h0000009jUf`0309jUf`050000002RYMP0XZGH0::Uf0000000
0`2VY]H3000000<0ZjWE00<000000;2/e02`[=@00P2`[=@3000000<0]k;D00@000000;^fe0000000
00000P2n^M@010000000`kgD0000000000030<Joe0030000003<aM<0c<GC0080c<GC00<000000<k6
d`0000000P3@b=801@000000d/WB0=;9dP3BbM8000000080dl_A00<000000?ooo`3oool0>`3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00L0oooo00210?ooo`030000003oool0oooo
0300oooo5P00000308KBo`@000000P2Jgod3000000<0ZNSj0P0000000`2Rj_X00000000000050000
00030:7^m`0000000000008000000P2;jO<8000000<0L=_c6@000002086dm@0500000028[nl0R:o_
08R_k`000000102?ZnT01@000000U:SS09BXh`2DZ><0000000<0VZKN00D0000009jUf`2NYM/0WZGK
000000030::Uf0@000000P2VYmL01@000000ZjWE0:^Ye@2[ZMD000000080/:cD0P0000020;:_e@03
0000002g/]@0]k;D0080]k;D00<000000;jie02n^M@00`2n^M@00`000000a[oD0<Joe0020<Joe003
00000038`]@000000080c<GC00H000000<k6d`3>a]<000000=38dP0000030=;9dP80000000<0dl_A
0000003oool0?@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00L0oooo00220?oo
o`800000<@3ooolH00000080PMKo1000000309CRo@H000000P2IjoT200000003093ZmP0000000000
02/000000`21]?D01@000000R:o_08R_k`28[nl0000000<0Sj_Y00D0000009BXh`2DZ><0U:SS0000
000309ZVgP040000002NYM/0WZGK09jUf`@000000P2QY]T01@000000YZOG0:JWe`2VYmL0000000<0
ZjWE0`0000040;:_e@030000002g/]@0]k;D0080000000<0^KGE0000002n^M@00P2n^M@00`000000
`K_D000000020<Joe00600000038`]@0b<;D0000003<aM<00000103>a]<01@000000d/WB0=;9dP3B
bM8000000080dl[A00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo00L0oooo00240?ooo`030000003oool0oooo0300oooo6000000307_Ho@@0000000<0SNCk
0000000000000P26i?PB00000080Kmc_5`0000000`1D]o/0000005fdo00205fdo1X000000P2KYMd0
1@000000XJKI0:6Vf@2QY]T0000000<0YZOG100000020:f/e`030000002b[mD0/ZoE00<0/ZoE0P00
00030;Vee@040000002n^M@0000000000080`K_D00<000000<Joe00000001038`]@00`000000c/KC
0<k6d`020<k6d`050000003@b=8000000=;9dP0000000P3Cb]400`000000oooo0?ooo`100?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool01P3oool008D0oooo00<000000?ooo`3oool0
<03oool40000000303RIlP000000000002X000000P1kh_88000000<0N][l600000000`1X/?X00000
07:]mP0207:]mP050000001mZ_40OJ[a07fZl@0000000`26Z>/01@000000SZKV08jViP2>Y^H00000
00<0UJGQ00D0000009^Ug@2KYMd0VjGM000000030:6Vf@@0000000<0YjWI0000002][=L0102][=L0
0`000000/ZoE0;:_e@02000000030;BbeP000000^KGE00<0^KGE00<000000<6ke031^m@00`31^m@0
0`000000b<;D0<S2e0020<S2e0040000003;a=@000000<k6d`8000000`3@b=8010000000dl[A0=?:
d@0000140?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool008H0oooo00<0
00000?ooo`3oool0;`3oool00`000000F:7e05RQm@0E000000<0Gloe2`000000101cgnh000000000
001ih_040000000308[WmP000000000000L000000P2@hod400000080P=Sn3P0000000`0iZ?@00000
000000030000000304bYn0000000000000L0000000<0OJ[a00000026Z>/00P26Z>/01@000000SZKV
08jViP2>Y^H0000000<0UJGQ00@0000009^Ug@2KYMd0VjGM100000000`2PZ=`000000:NYf@030:NY
f@030000002][=L0[JcG0080[JcG0`0000030;BbeP030000002i]MD0^KGE0080000000<0_;SF0000
0031^m@00P31^m@010000000a;kE00000038`]@2000000<0blCD00<000000=38dP3@b=800P3@b=82
000000030=?:d@000000oooo04@0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`06
0?ooo`00Q`3oool200000300oooo7`000000101^fo8000000000001hh?<>0000000409cYn`000000
00000:CWn`@000000P2Gh?hF0000000303RQm0000000000000<0000000<0EjOh0000001TY_D00P1T
Y_DB0000000309VVg`000000X:SL00<0X:SL00D000000:NYf@2WZMT0YjWI000000020:f/e`800000
0P2^[mP00`000000];;F0;BbeP020;BbeP8000000`2l^=H010000000`K_D0000000000030<Bne@03
0000003;a=@0blCD0080blCD00@000000<g6e0000000d<SB0P0000020=79d`030000003oool0oooo
04D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`00R@3oool00`000000
oooo0?ooo`0^0?ooo`8000000P1^ZoH01@000000IK7k06Fan`1U/O/0000000<0GKSl3P00000207;I
n0/0000000@0V^_i000000000000Y>_j0P0000000`2]j_T000000000000400000080ZnCj10000000
1@2Mg?d000000000002Qe_X0000000<0S<on4P0000000`0mW_@00000000000030000000306BVm@00
0000L:Ga0080L:Ga00D0000007bUk@1lYNd0O:G]0000000308NUj0050000002@YN<0T:GS092Uh`00
00000`2IY]l01@000000X:SL0:2Xg02PZ=`0000000<0YjWI0`0000040:j_f0030000002d/]H0];;F
0080000000<0][GH0000002l^=H00`2l^=H00`000000a;kE0<Bne@020<Bne@0300000036`]H00000
0080blCD00D000000<g6e03=a]@0cLKD000000030=79d`030000003oool0oooo04L0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`00RP3oool00`000000oooo0?ooo`0_0?oo
o`X000000P1c`ohC0000000309_Xn`00000000000080Yn[j1@0000020:gZn@@0000000<0^N?d0000
000000000P0000020:oNm`<000000`2Qe_X400000080T<Wk1`0000000`1M/_/00000000000020000
0003036Jl@000000000000D0000000<0B9ge0000000000001`0000000`1lYNd0000008NUj00208NU
j0050000002@YN<0T:GS092Uh`0000000`2IY]l010000000X:SL0:2Xg02PZ=`400000080YjcK00<0
00000:j_f02^[mP00P2^[mP3000000<0][GH00<000000;bheP2l^=H00P0000000`2n^mL000000<Bn
e@0200000080a/;F00<000000<_4e00000000`3=a]@01@000000c/WE0000003AbM<0000004X0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`00R`3oool00`000000oooo0?oo
o`0_0?ooo`0400000021/oD0PK?e086cm@L000000`27boh400000080T]Sn100000000`2Ihod00000
000000020:KUn`H0000000@0/NWh000000000000^NOe1@0000020;WSm08000000P2lgO03000000<0
/]Sd0`0000020:C?mP@000000P2D`oL010000000OK[j07fjnP1m^_X:0000000302RAkP0000000000
00D0000000<0EIkd0000001TWo800P1TWo8B00000080WjWN00D000000:N/f`2W[=/0YjcK00000002
0:j_f08000000P2_/]X00`000000][GH0;Jef0020;Jef0800000102n^mL00`000000a/;F0<K2eP03
0<K2eP050000003=a]@0cLKD0<g6e00000000P3>bMD00`000000dLWC0000001;0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool01P3oool008`0oooo0P00000`0?ooo`<000000P2A^O80
10000000SL;i08g2n@2=`_T400000080V=3k100000020:7Lo0<000000`2ViO/4000000040;[Tm000
0000000000000080^NOe100000020<?Ok@8000000`2lgO02000000030;kHk0000000000000800000
0P2dd^l300000080Y/Sa00@0000009C3m`2D`oL0U<?g300000000`0bU_4000000000000500000003
03ZDl@000000000000<0000000<0I9ob0000001bXNh00P1bXNh01@000000Oj?Z07nSjP1oXnX00000
00<0RjGV00D0000009JWhP2FYn80UZOR0000000309nYgP040000002W[=/0YjcK0:N/f`<00000102_
/]X00`000000][GH0;Jef002000000030;Rhf@000000_[_G0080_[_G00<000000<6of00000000`36
`]H01@000000b<GG0000003=a]@0000000@0c/WE00<000000?ooo`3oool0B`3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo00D0oooo002>0?ooo`030000003oool0oooo02h0oooo00@0
0000096ilP2A^O80TKWb1000000209c7mP030000002Hd?/0V=3k008000000P2Ve?L3000000<0[]kh
1@0000020;[Tm080000000<0`N;_0000000000000`0000030<?Ok@80000000@0aM[Y000000000000
00000P2nf>`2000000040;oBj0000000]=;_0;CBk`80000000@0]L_[0000002Vb?40Y/Sa10000002
09NmlP0500000022]OD0P[Ge08:em@0000000`1[Z_H60000000302N<k@000000000000D0000000<0
B9Kb0000000000001`0000000`1oXnX0000008^UiP0208^UiP050000002FYn80UZOR09JWhP000000
0`2OZMh400000080YjoM00<000000:nbfP2_/]X00P2_/]X3000000<0^;SI00@000000;jke`000000
00000P31_mP010000000a/;F0000000000030<S5e`030000003>bMD0c/WE0080c/WE00<000000?oo
o`3oool0C@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo002?0?ooo`03
0000003oool0oooo02h0oooo0`00000209nnkP8000000P2LaoH3000000<0Y]Cg0P0000000`2beO80
000000000002000000040;SNl`00000000000;oMkP<000000P31h^l200000080a]k[00<000000<WJ
iP0000000P0000030<GJj@0300000037eND0000000800000102od^P2000000<0]L_[0`0000020:O2
k08000000P2G_O8;0000000305:Mm0000000000000@0000000<08hK[0000000000001@0000000`1H
V_40000006RMk`0206RMk`h000000P2N[>401@000000YjoM0:N_g@2W[md000000080[k;J0P000002
0;2eg0030000002h^=T0^;SI0080^;SI00<000000<6of031_mP00`31_mP00`000000b<GG0<S5e`02
0<S5e`040000003:b=P000000<k9e@800000D03oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00D0oooo002@0?ooo`030000003oool0oooo02h0oooo0P00000309nnkP<000000`2Yb_43
00000080/]Gb0P000000102keNd000000000000000020;oMkP8000000P35f^T200000080a]k[0P00
00020<WJiP040000003;e^8000000<GJj@8000000P37eND00`000000am3Q000000020;oBj0800000
00<0_lcT0000000000000P0000020;G5iP040000002W`^`0Yl;/0:O2k0<000002`3oool=00000003
03b?k`000000000000<0000000<0J9g_0000001hXN`00P1hXN`01@000000QZGX08JUj026YNP00000
00<0TjSU00D0000009j/h@2N[>40WZcQ000000030:N_g@<00000102`]M`00`000000^;SI0;Rhf@02
000000040;Zlf`000000`KoH0<6of080000000<0`l?J00000038aML00P0000030<[8f0030000003o
ool0oooo0540oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`050?ooo`00T@3oool2
00000300oooo0`0000020:[1jP030000002Yb_40ZL[a0080000000@0/lc/00000000000000000`2k
eNd00`000000`]CX0<;Dj002000000@0aM[Y00@000000<WHi000000000000P39f^H2000000<0bmKR
00<000000<OEi@37eND00P0000030<O@h@8000000P2oc>@2000000030;k6h0000000]LGV0080]LGV
1000000B0?ooo`D0000000<08X;Y0000000000001000000204bDl0D0000000<0N:7/00000026YNP0
0P26YNP01@000000TjSU09>Xi@2CZ>D0000000<0WZcQ100000020:Nbh0030000002`]M`0/;GL0080
/;GL0`0000040;Zlf`8000001033`mX00`000000b/SH0<[8f0020<[8f0030000003oool0oooo05<0
oooo00D000000?ooo`3oool0oooo000000070?ooo`00T`3oool00`000000oooo0?ooo`0^0?ooo`04
0000002Z`NX0Z/7Z0:[1jP<000000P2cc>`2000000030;_<i`000000^mG]008000000`32e>P01000
0000am?S00000035f^T200000080bMST00@000000<cDh000000000000`3;e^82000000030<cBgP00
0000000000<0am3Q0P0000001036bmd000000000000000040;k6h0<00000603oool90000000403:9
k00000000000000000<0GY[`3P00000209b^i0050000002W/^00Yk;P0:Nbh00000000P2`]M`20000
0080/KWO00L000000;Zlf`2j_=/0^[cK0000002k`Mh0000000<0`l?J00@000000<C7g0000000b/SH
0P00001G0?ooo`040000003oool0oooo000000L0oooo002D0?ooo`030000003oool0oooo02h0oooo
0`0000020;?2i@<00000102kc>L200000080`]CX0P0000030<OCh`0300000039f>@0bMST00800000
0P3<e>000`000000c]7L00000002000000<0c=;N00<000000<c>f`0000000P0000030<K;g@030000
0036a]X000000080_/KP0P00000M0?ooo`<0000000<08h7X0000000000001`0000000`1NV_000000
072PkP02072PkP0500000021YN/0PJG[086Uj`0000000`2@Z^P01@000000W:kT09b^i02L[^@00000
00<0Yk;P0`0000040;6ig`040000002j_=/0000000000080^l7N00@000000<?3fP00000000000`34
am`00`000000oooo0?ooo`1G0?ooo`040000003oool0oooo000000L0oooo002E0?ooo`030000003o
ool0oooo02h0oooo0P000000102c`^D000000000002j`n42000000030;_<i`00000000000080`/cR
0P0000030<OCh`80000000<0b]7O0000000000000P3<e>0200000080c]7L00H000000<o>f@000000
00000<cBgP0000020<c>f`060000003<b]P0000000000036bmd000000`36a]X200000280oooo1@00
000202^4jP/0000000<0PJG[0000002@Z^P00P2@Z^P800000080YkKS00@000000;6ig`2a^Ml0/KWO
0`0000050;_1gP0400000034am`0a<OL0<C7g0800000F`3oool00`000000oooo000000070?ooo`00
UP3oool200000300oooo00<000000;[3h@2j`n400`2j`n400`000000`/cR0<;<hP020<;<hP030000
0036bmh00000008000000`3:dMl010000000cLoK0000000000030<kAg0030000003?c]T0clkI0080
clkI00@000000<c>f`3<c]/00000103<b]P01@000000a/KJ0<K6fP36a]X0000002L0oooo1@000000
1P0jRnd00000000000000000EIO`000000<0J9o`1P00000208b/j`050000002K/NL0Vk7W09^ai`00
00000`2W]^<4000000<0/[gR00<000000;_1gP2k`Mh00`0000000`34am`000000000001N0?ooo`03
0000003oool0000000H0oooo002H0?ooo`030000003oool0oooo02h0oooo00<000000;[3h@2j`n40
0P0000000`30a=d000000<;<hP0200000080a/_N00@000000<[;fP000000b]7O0P0000020<g?f`06
0000003?cMP000000000003>dM`000000`3?c]T010000000cl_F0000000000040<c:f0050000003;
a]H000000<K6fP000000:P3oool50000000303Z;k@000000000000<0000000<0J9o`0000001kY^h0
0P1kY^h01@000000S:c[08b/j`2<[>/0000000<0Vk7W100000020:Nji`030000002b_N80/[gR0080
/[gR0`0000030;g5h0030000003oool0oooo05l0oooo0P0000060?ooo`00V@3oool00`000000oooo
0?ooo`0^0?ooo`8000001030a=d00`000000a/_N0<K;gP02000000<0b/_J0P0000020<g?f`030000
003?cMP0clgH0080clgH00<000000<o>f@3?c]T00P0000020<o;eP040000003?b=@000000<c:f080
00000`3;a]H00`000000oooo0?ooo`0[0?ooo`L0000000<0C9C`0000000000003000000209Vej`05
0000002W^^L0Yk[W0:Nji`0000000P2b_N82000000040;?2i@000000_LGP0;g5h0800000HP3oool2
000000H0oooo002J0?ooo`030000003oool0oooo02h0oooo00<000000<34g@30a=d00P0000000`35
a=T00000000000040<[;fP030000003=bML0cLWG00800000103?cMP010000000d<_E000000000003
0<o;eP050000003?b=@0clSD0<o8e0000000103;a]H00`000000oooo0?ooo`0^0?ooo`<0000000<0
=8S/0000000000000`0000000`1QW_40000007JWl@0207JWl@0500000029[nl0RJo_08V_k`000000
0`2I]N/010000000Yk[W0:Nji`2W^^L3000000@0/l;U0P00001T0?ooo`8000001P3oool009/0oooo
00<000000?ooo`3oool0;P3oool2000000<0aLCI00H000000<W4eP00000000000<[;fP0000030<g9
e`040000003?b=@000000<o=f08000000P3@bmD01@000000dLWB0000003?bmH0000000<0clSD00@0
00000<k5dP000000blKF0P00000d0?ooo`@0000000<0AI;`000000000000400000020:NnjP030000
002c`^D0/l;U00800000HP3oool4000000P0oooo002L0?ooo`800000;`3oool010000000aLCI0000
000000040<W4eP050000003=bML0cLWG0<g9e`0000000`3?b=@01@000000d<_E0=3;e@3@bmD00000
00<0dLWB00<000000<o8e03?b=@00P0000030<k5dP030000003oool0oooo03P0oooo100000000`1J
WO<00000072Xl`02072Xl`0500000025/_80QK;b08FblP0000000`2H^Nl010000000YkkZ0:NnjP2W
_^X300000600oooo1000000<0?ooo`00WP3oool00`000000oooo0?ooo`0]0?ooo`0300000039a=H0
bLCF00<0bLCF00D000000<c4d`000000cLWG000000030<o8e0050000003@am4000000=3;e@000000
0`3AbM8010000000dLO@0000000000040<k5dP030000003oool0oooo03d0oooo100000000`1`Z?<0
000008FblP0208FblP040000002H^Nl0V;W_09Rik`@00000G`3oool400000100oooo002O0?ooo`03
0000003oool0oooo02d0oooo0P0000000`39a=H00000000000030<c4d`030000003?b=@0clSD0080
00000`3@am400`000000dLWB0=79dP0200000080dLO@00@000000=35c`000000c/GB0P0000140?oo
o`T00000H03oool3000001@0oooo002P0?ooo`030000003oool0oooo02h0oooo00<000000<c4d`3<
a=<00P3<a=<010000000c/CA0000000000040=37d@040000003Aa]000000000000<0dLO@00D00000
0=35c`3@aLl0d<G?0000002[0?ooo`@000005`3oool00:40oooo0P00000_0?ooo`040000003<a=<0
000000000080c/CA00@000000=34c`000000d<OA0P0000020=76d0050000003BaLh000000=77d000
00000`3@aLl00`000000oooo0?ooo`2V0?ooo`@000006`3oool00:<0oooo00<000000?ooo`3oool0
;@3oool01@000000c/CA0<k4d@3>a=40000000<0d<C?00D000000=76d03Aa]00dLK@000000030=;5
cP050000003@aLl0d<G?0=35c`000000Y@3oool4000001l0oooo002T0?ooo`030000003oool0oooo
02d0oooo00<000000<k4d@0000000`3@a<l01@000000dLC>0000003Aa]00000000<0d/G>00H00000
0=;4c@00000000000=35c`00002Q0?ooo`@000008`3oool00:D0oooo00<000000?ooo`3oool0;@3o
ool01@000000d<C?0=34c`3@a<l0000000<0dLC>00D000000=;5cP3BaLh0d/G>000000040=;4c@03
0000003oool0oooo09d0oooo0`00000W0?ooo`00YP3oool2000002h0oooo00<000000=34c`000000
0`3Aa<h01P000000d/C=000000000000d/G>000000<0d/C=00<000000?ooo`3oool0VP3oool40000
02X0oooo002X0?ooo`030000003oool0oooo02`0oooo00D000000=74cP3Aa<h0dLC>000000040=;4
c@050000003Ba<d0d/C=0=;4c@000000V@3oool4000002h0oooo002Y0?ooo`030000003oool0oooo
02`0oooo0P0000000`3Aa<h000000=;4c@020=;4c@050000003Ba<`000000=;4c@000000UP3oool4
00000380oooo002Z0?ooo`030000003oool0oooo02d0oooo00D000000=;4c@3Ba<d0d/C=00000003
0=;4c0030000003oool0oooo0980oooo0`00000f0?ooo`00Z`3oool2000002h0oooo00<000000=;4
c@0000000`3Ba<`00`000000oooo0?ooo`2?0?ooo`@00000>@3oool00:d0oooo00<000000?ooo`3o
ool0;03oool01@000000d/C<0=;4c03Ba<`0000008h0oooo1000000m0?ooo`00[P3oool00`000000
oooo0?ooo`0/0?ooo`030000003Ba<`0000008/0oooo100000110?ooo`00[`3oool00`000000oooo
0?ooo`0/0?ooo`030000003oool0oooo08H0oooo100000150?ooo`00/03oool2000002`0oooo00<0
00000?ooo`3oool0P`3oool3000004T0oooo002b0?ooo`030000003oool0oooo02T0oooo00<00000
0?ooo`3oool0O`3oool4000004`0oooo002c0?ooo`030000003oool0oooo02P0oooo00<000000?oo
o`3oool0N`3oool400000500oooo002d0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3o
ool0M`3oool4000005@0oooo002e0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0
M03oool3000005P0oooo002f0?ooo`8000009P3oool00`000000oooo0?ooo`1`0?ooo`@00000F`3o
ool00;P0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`1/0?ooo`@00000G`3oool0
0;T0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`1X0?ooo`@00000H`3oool00;X0
oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`1T0?ooo`@00000I`3oool00;/0oooo
0P00000Q0?ooo`030000003oool0oooo0640oooo0`00001[0?ooo`00_@3oool00`000000oooo0?oo
o`0N0?ooo`030000003oool0oooo05d0oooo1000001^0?ooo`00_P3oool00`000000oooo0?ooo`0M
0?ooo`030000003oool0oooo05T0oooo1000001b0?ooo`00_`3oool00`000000oooo0?ooo`0L0?oo
o`030000003oool0oooo05D0oooo1000001f0?ooo`00`03oool2000001`0oooo00<000000?ooo`3o
ool0DP3oool3000007X0oooo00320?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0
CP3oool4000007d0oooo00330?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0BP3o
ool400000840oooo00340?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0AP3oool4
000008D0oooo00350?ooo`8000005`3oool00`000000oooo0?ooo`120?ooo`@00000R@3oool00<L0
oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`100?ooo`<00000S@3oool00<P0oooo
00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0l0?ooo`@00000T03oool00<T0oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`0h0?ooo`@00000U03oool00<X0oooo0P00000A
0?ooo`030000003oool0oooo03@0oooo1000002H0?ooo`00c03oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo0340oooo0`00002L0?ooo`00c@3oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo02d0oooo1000002O0?ooo`00cP3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo02T0oooo1000002S0?ooo`00c`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo02D0oooo1000002W0?ooo`00d03oool2000000/0oooo00<000000?ooo`3oool08@3oool4
00000:/0oooo003B0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool07P3oool30000
0:l0oooo003C0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool06P3oool400000;80
oooo003D0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool05P3oool400000;H0oooo
003E0?ooo`8000001P3oool00`000000oooo0?ooo`0B0?ooo`@00000^P3oool00=L0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0?0?ooo`<00000_P3oool00=P0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0;0?ooo`@00000`@3oool00=T0oooo00D000000?ooo`3o
ool0oooo000000090?ooo`@00000a@3oool00=X0oooo0P0000000`3oool000000?ooo`040?ooo`@0
0000b@3oool00=`0oooo0P0000000`3oool000000000000200000<d0oooo003M0?ooo`800000d@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {349.313, 0}} -> {-0.241953, -0.0711471, \
0.00420975, 0.00420975}}]
}, Closed]],

Cell[TextData[{
  "In addition to the options available to ",
  StyleBox["Plot", "InlineInput",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Plot3D", "InlineInput",
    FontWeight->"Bold"],
  " comes with options to control the shading scheme and the viewpoint.  An \
example using the ",
  StyleBox["ViewPoint", "InlineInput",
    FontWeight->"Bold"],
  " option is given below.  Note that these values were obtain by using the \
",
  StyleBox["3D ViewPoint Selector",
    FontWeight->"Bold"],
  " from the ",
  StyleBox["Input",
    FontWeight->"Bold"],
  " menu.  Alternatively, there is a shortcut key for those who remember such \
things easily."
}], "Text"],

Cell["To change the viewpoint follow these steps:", "Text"],

Cell[TextData[{
  "if no ",
  StyleBox["ViewPoint", "InlineInput"],
  " data are present in the plot command, click at an appropriate insertion \
point near the end of the argument sequence to indicate where the ",
  StyleBox["ViewPoint", "InlineInput"],
  " data should be inserted"
}], "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell["\<\
if there are pre-existing viewpoint data , highlight those data with the \
cursor to replace old data with new data\
\>", "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell["\<\
activate the selector using Input \[RightTeeArrow] 3DViewPoint Selector or \
the corresponding shortcut\
\>", "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell["use the control bars to vary to taste", "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
  "click ",
  StyleBox["Paste.  ",
    FontSlant->"Italic"],
  "This should insert or replace the viewpoint data.  "
}], "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell[TextData[{
  "replot to see the result.  Note that values that are out of range are \
clipped.  Of course, you can vary the ",
  StyleBox["PlotRange", "InlineInput",
    FontWeight->"Bold"],
  " also."
}], "Text",
  CellDingbat->"\[EmptySquare]",
  CellMargins->{{30, Inherited}, {Inherited, Inherited}}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot3D[
        Evaluate[Re[Gamma[x\  + \ I\ y]]], {x, \(-5\), 5}, {y, \(-5\), 5}, 
        PlotPoints \[Rule] 50, 
        ViewPoint -> {\(-5.009\), \ \(-6.671\), \ 1.524}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .46147 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% SurfaceGraphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0116045 1.0116 -0.257307 1.0116 [
[.47918 -0.00253 -9.15205 -15.6875 ]
[.47918 -0.00253 9.16045 0 ]
[.59984 .01724 -9.18576 -15.6875 ]
[.59984 .01724 9.12674 0 ]
[.71699 .03644 -5.03397 -15.6875 ]
[.71699 .03644 4.96603 0 ]
[.83079 .0551 -5.05128 -15.6875 ]
[.83079 .0551 4.94872 0 ]
[.94136 .07322 -5.06809 -15.6875 ]
[.94136 .07322 4.93191 0 ]
[.95247 .39944 -9.02971 0 ]
[.95247 .39944 9.28279 15.6875 ]
[.85966 .4171 -9.05539 0 ]
[.85966 .4171 9.25711 15.6875 ]
[.77019 .43412 -4.95843 0 ]
[.77019 .43412 5.04157 15.6875 ]
[.68388 .45054 -4.97144 0 ]
[.68388 .45054 5.02856 15.6875 ]
[.60057 .4664 -4.98399 0 ]
[.60057 .4664 5.01601 15.6875 ]
[1.00867 .1294 0 -6.59202 ]
[1.00867 .1294 27.3125 9.09548 ]
[1.00999 .21438 0 -6.67896 ]
[1.00999 .21438 10 9.00854 ]
[1.01131 .2998 0 -6.76638 ]
[1.01131 .2998 19 8.92112 ]
[ 0 0 0 0 ]
[ 1 .46147 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.41745 0 m
.99568 .09473 L
s
.47917 .01011 m
.47917 .01643 L
s
gsave
.47918 -0.00253 -70.1521 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.59988 .02989 m
.5999 .03621 L
s
gsave
.59984 .01724 -70.1858 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.71708 .04909 m
.71712 .05541 L
s
gsave
.71699 .03644 -66.034 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.83092 .06774 m
.83098 .07406 L
s
gsave
.83079 .0551 -66.0513 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.94154 .08586 m
.94162 .09219 L
s
gsave
.94136 .07322 -66.0681 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.50969 .01511 m
.50969 .0189 L
s
.53998 .02007 m
.53998 .02387 L
s
.57004 .025 m
.57005 .02879 L
s
.6295 .03474 m
.62952 .03853 L
s
.65891 .03956 m
.65893 .04335 L
s
.6881 .04434 m
.68812 .04813 L
s
.74585 .0538 m
.74588 .0576 L
s
.77441 .05848 m
.77444 .06228 L
s
.80276 .06313 m
.8028 .06692 L
s
.85887 .07232 m
.85891 .07611 L
s
.88662 .07687 m
.88666 .08066 L
s
.91417 .08138 m
.91422 .08517 L
s
.44843 .00507 m
.44843 .00887 L
s
.9687 .09031 m
.96876 .09411 L
s
.25 Mabswid
1 .37771 m
.55997 .46147 L
s
.9523 .38679 m
.95221 .38047 L
s
gsave
.95247 .39944 -70.0297 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.85952 .40445 m
.85945 .39813 L
s
gsave
.85966 .4171 -70.0554 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.77008 .42148 m
.77003 .41515 L
s
gsave
.77019 .43412 -65.9584 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.68381 .4379 m
.68377 .43158 L
s
gsave
.68388 .45054 -65.9714 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.60053 .45375 m
.60051 .44743 L
s
gsave
.60057 .4664 -65.984 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.92878 .39127 m
.92873 .38747 L
s
.90548 .3957 m
.90543 .39191 L
s
.8824 .4001 m
.88235 .3963 L
s
.83686 .40877 m
.83682 .40497 L
s
.8144 .41304 m
.81436 .40925 L
s
.79214 .41728 m
.79211 .41348 L
s
.74823 .42564 m
.7482 .42184 L
s
.72656 .42976 m
.72654 .42597 L
s
.70509 .43385 m
.70507 .43005 L
s
.66272 .44191 m
.6627 .43812 L
s
.64181 .44589 m
.64179 .4421 L
s
.62108 .44984 m
.62107 .44605 L
s
.97604 .38227 m
.97598 .37848 L
s
.58016 .45763 m
.58015 .45383 L
s
.25 Mabswid
.99568 .09473 m
1 .37771 L
s
.99618 .1274 m
.98994 .12641 L
s
gsave
1.00867 .1294 -61 -10.592 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.99748 .21252 m
.99122 .21159 L
s
gsave
1.00999 .21438 -61 -10.679 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.99878 .29808 m
.99252 .29722 L
s
gsave
1.01131 .2998 -61 -10.7664 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.125 Mabswid
.99644 .14439 m
.99269 .1438 L
s
.9967 .1614 m
.99295 .16082 L
s
.99696 .17842 m
.99321 .17785 L
s
.99722 .19546 m
.99347 .1949 L
s
.99774 .2296 m
.99399 .22905 L
s
.998 .24669 m
.99425 .24615 L
s
.99826 .2638 m
.99451 .26327 L
s
.99852 .28093 m
.99477 .28041 L
s
.99592 .11043 m
.99218 .10983 L
s
.99905 .31524 m
.99529 .31474 L
s
.99931 .33243 m
.99555 .33193 L
s
.99957 .34963 m
.99581 .34913 L
s
.99983 .36684 m
.99607 .36636 L
s
.25 Mabswid
.41745 0 m
.41674 .30365 L
s
.41674 .30365 m
0 .40002 L
s
0 .40002 m
.00412 .1233 L
s
.00412 .1233 m
.41745 0 L
s
.99568 .09473 m
.55946 .20202 L
s
.55946 .20202 m
.55997 .46147 L
s
.55997 .46147 m
1 .37771 L
s
1 .37771 m
.99568 .09473 L
s
.41745 0 m
.41674 .30365 L
s
.41674 .30365 m
1 .37771 L
s
1 .37771 m
.99568 .09473 L
s
.99568 .09473 m
.41745 0 L
s
.00412 .1233 m
.55946 .20202 L
s
.55946 .20202 m
.55997 .46147 L
s
.55997 .46147 m
0 .40002 L
s
0 .40002 m
.00412 .1233 L
s
0 0 m
1 0 L
1 .46147 L
0 .46147 L
closepath
clip
newpath
.5 Mabswid
.715 .564 .683 r
.55725 .33613 .54891 .24347 .55946 .20202 .56779 .28122 Metetra
.732 .574 .676 r
.54659 .3587 .53831 .29099 .54891 .24347 .55725 .33613 Metetra
.753 .582 .664 r
.53586 .36312 .52763 .31505 .53831 .29099 .54659 .3587 Metetra
.78 .588 .641 r
.52509 .3581 .5169 .325 .52763 .31505 .53586 .36312 Metetra
.819 .587 .587 r
.51429 .34886 .50613 .32685 .5169 .325 .52509 .3581 Metetra
.874 .542 .421 r
.50347 .33836 .49532 .32439 .50613 .32685 .51429 .34886 Metetra
.642 .12 0 r
.49262 .32818 .48449 .31992 .49532 .32439 .50347 .33836 Metetra
.229 .502 .904 r
.48175 .31909 .47363 .31477 .48449 .31992 .49262 .32818 Metetra
.477 .617 .91 r
.47086 .31134 .46275 .30969 .47363 .31477 .48175 .31909 Metetra
.56 .657 .901 r
.45994 .30494 .45184 .30503 .46275 .30969 .47086 .31134 Metetra
.599 .684 .901 r
.44899 .29976 .4409 .30095 .45184 .30503 .45994 .30494 Metetra
.62 .706 .907 r
.43802 .29561 .42993 .29744 .4409 .30095 .44899 .29976 Metetra
.631 .726 .916 r
.42701 .29228 .41893 .29447 .42993 .29744 .43802 .29561 Metetra
.635 .744 .927 r
.41598 .2896 .40791 .29194 .41893 .29447 .42701 .29228 Metetra
.635 .76 .938 r
.40491 .28739 .39685 .28977 .40791 .29194 .41598 .2896 Metetra
.631 .773 .948 r
.39382 .28553 .38577 .28788 .39685 .28977 .40491 .28739 Metetra
.625 .783 .957 r
.38269 .2839 .37465 .28618 .38577 .28788 .39382 .28553 Metetra
.619 .79 .964 r
.37153 .28242 .3635 .28462 .37465 .28618 .38269 .2839 Metetra
.612 .795 .969 r
.36035 .28102 .35233 .28315 .3635 .28462 .37153 .28242 Metetra
.607 .798 .973 r
.34913 .27967 .34112 .28173 .35233 .28315 .36035 .28102 Metetra
.602 .799 .975 r
.33788 .27832 .32988 .28034 .34112 .28173 .34913 .27967 Metetra
.599 .8 .976 r
.3266 .27697 .31861 .27894 .32988 .28034 .33788 .27832 Metetra
.597 .799 .977 r
.3153 .27559 .30731 .27755 .31861 .27894 .3266 .27697 Metetra
.596 .799 .977 r
.30396 .2742 .29598 .27613 .30731 .27755 .3153 .27559 Metetra
.596 .798 .977 r
.29259 .27277 .28462 .27471 .29598 .27613 .30396 .2742 Metetra
.596 .798 .977 r
.28118 .27133 .27323 .27327 .28462 .27471 .29259 .27277 Metetra
.596 .797 .976 r
.26975 .26987 .26181 .27181 .27323 .27327 .28118 .27133 Metetra
.597 .797 .976 r
.25829 .2684 .25036 .27035 .26181 .27181 .26975 .26987 Metetra
.597 .797 .976 r
.24679 .26691 .23887 .26887 .25036 .27035 .25829 .2684 Metetra
.597 .796 .976 r
.23527 .26542 .22736 .26739 .23887 .26887 .24679 .26691 Metetra
.598 .796 .975 r
.22371 .26393 .21581 .26591 .22736 .26739 .23527 .26542 Metetra
.598 .796 .975 r
.21212 .26243 .20423 .26442 .21581 .26591 .22371 .26393 Metetra
.598 .796 .975 r
.2005 .26093 .19262 .26292 .20423 .26442 .21212 .26243 Metetra
.598 .796 .975 r
.18885 .25943 .18098 .26143 .19262 .26292 .2005 .26093 Metetra
.598 .796 .975 r
.17716 .25792 .1693 .25993 .18098 .26143 .18885 .25943 Metetra
.598 .796 .975 r
.16544 .25641 .15759 .25842 .1693 .25993 .17716 .25792 Metetra
.598 .796 .975 r
.15369 .2549 .14585 .25692 .15759 .25842 .16544 .25641 Metetra
.598 .797 .975 r
.14191 .25338 .13408 .2554 .14585 .25692 .15369 .2549 Metetra
.598 .797 .976 r
.13009 .25186 .12228 .25389 .13408 .2554 .14191 .25338 Metetra
.598 .797 .976 r
.11825 .25034 .11044 .25237 .12228 .25389 .13009 .25186 Metetra
.598 .796 .976 r
.10636 .24881 .09857 .25084 .11044 .25237 .11825 .25034 Metetra
.598 .796 .976 r
.09445 .24727 .08667 .24931 .09857 .25084 .10636 .24881 Metetra
.598 .796 .976 r
.0825 .24573 .07473 .24778 .08667 .24931 .09445 .24727 Metetra
.598 .796 .975 r
.07052 .24419 .06276 .24624 .07473 .24778 .0825 .24573 Metetra
.598 .796 .975 r
.05851 .24264 .05076 .2447 .06276 .24624 .07052 .24419 Metetra
.598 .796 .975 r
.04646 .24109 .03872 .24316 .05076 .2447 .05851 .24264 Metetra
.598 .796 .975 r
.03438 .23954 .02665 .24161 .03872 .24316 .04646 .24109 Metetra
.598 .796 .975 r
.02226 .23798 .01455 .24005 .02665 .24161 .03438 .23954 Metetra
.598 .796 .975 r
.01011 .23641 .00241 .23849 .01455 .24005 .02226 .23798 Metetra
.765 .604 .679 r
.56572 .45715 .55725 .33613 .56779 .28122 .57644 .45834 Metetra
.782 .612 .67 r
.55496 .44436 .54659 .3587 .55725 .33613 .56572 .45715 Metetra
.826 .635 .645 r
.54415 .42039 .53586 .36312 .54659 .3587 .55496 .44436 Metetra
.886 .655 .583 r
.53333 .39525 .52509 .3581 .53586 .36312 .54415 .42039 Metetra
.958 .626 .334 r
.52249 .37199 .51429 .34886 .52509 .3581 .53333 .39525 Metetra
0 .013 .587 r
.51165 .3519 .50347 .33836 .51429 .34886 .52249 .37199 Metetra
.332 .44 .821 r
.50079 .3353 .49262 .32818 .50347 .33836 .51165 .3519 Metetra
.502 .547 .834 r
.4899 .32201 .48175 .31909 .49262 .32818 .50079 .3353 Metetra
.572 .598 .84 r
.479 .31161 .47086 .31134 .48175 .31909 .4899 .32201 Metetra
.61 .633 .849 r
.46807 .30361 .45994 .30494 .47086 .31134 .479 .31161 Metetra
.632 .663 .862 r
.45711 .29753 .44899 .29976 .45994 .30494 .46807 .30361 Metetra
.645 .69 .877 r
.44613 .29293 .43802 .29561 .44899 .29976 .45711 .29753 Metetra
.652 .715 .895 r
.43512 .28944 .42701 .29228 .43802 .29561 .44613 .29293 Metetra
.652 .738 .912 r
.42407 .28676 .41598 .2896 .42701 .29228 .43512 .28944 Metetra
.649 .758 .929 r
.413 .28466 .40491 .28739 .41598 .2896 .42407 .28676 Metetra
.642 .774 .944 r
.40189 .28294 .39382 .28553 .40491 .28739 .413 .28466 Metetra
.632 .786 .955 r
.39076 .28146 .38269 .2839 .39382 .28553 .40189 .28294 Metetra
.622 .794 .964 r
.37959 .28011 .37153 .28242 .38269 .2839 .39076 .28146 Metetra
.613 .799 .971 r
.3684 .27884 .36035 .28102 .37153 .28242 .37959 .28011 Metetra
.606 .802 .975 r
.35717 .27758 .34913 .27967 .36035 .28102 .3684 .27884 Metetra
.6 .802 .977 r
.34591 .27631 .33788 .27832 .34913 .27967 .35717 .27758 Metetra
.597 .802 .978 r
.33463 .275 .3266 .27697 .33788 .27832 .34591 .27631 Metetra
.595 .801 .979 r
.32331 .27365 .3153 .27559 .3266 .27697 .33463 .275 Metetra
.594 .8 .978 r
.31196 .27227 .30396 .2742 .3153 .27559 .32331 .27365 Metetra
.594 .799 .978 r
.30058 .27085 .29259 .27277 .30396 .2742 .31196 .27227 Metetra
.595 .798 .977 r
.28917 .2694 .28118 .27133 .29259 .27277 .30058 .27085 Metetra
.596 .797 .977 r
.27772 .26793 .26975 .26987 .28118 .27133 .28917 .2694 Metetra
.596 .797 .976 r
.26625 .26644 .25829 .2684 .26975 .26987 .27772 .26793 Metetra
.597 .796 .976 r
.25475 .26495 .24679 .26691 .25829 .2684 .26625 .26644 Metetra
.598 .796 .975 r
.24321 .26345 .23527 .26542 .24679 .26691 .25475 .26495 Metetra
.598 .796 .975 r
.23164 .26194 .22371 .26393 .23527 .26542 .24321 .26345 Metetra
.598 .796 .975 r
.22004 .26044 .21212 .26243 .22371 .26393 .23164 .26194 Metetra
.598 .796 .975 r
.20841 .25893 .2005 .26093 .21212 .26243 .22004 .26044 Metetra
.598 .796 .975 r
.19675 .25742 .18885 .25943 .2005 .26093 .20841 .25893 Metetra
.598 .796 .975 r
.18505 .25591 .17716 .25792 .18885 .25943 .19675 .25742 Metetra
.598 .796 .975 r
.17332 .2544 .16544 .25641 .17716 .25792 .18505 .25591 Metetra
.598 .797 .975 r
.16156 .25288 .15369 .2549 .16544 .25641 .17332 .2544 Metetra
.598 .797 .976 r
.14977 .25136 .14191 .25338 .15369 .2549 .16156 .25288 Metetra
.598 .797 .976 r
.13794 .24983 .13009 .25186 .14191 .25338 .14977 .25136 Metetra
.598 .797 .976 r
.12608 .2483 .11825 .25034 .13009 .25186 .13794 .24983 Metetra
.598 .797 .976 r
.11419 .24676 .10636 .24881 .11825 .25034 .12608 .2483 Metetra
.598 .796 .976 r
.10226 .24522 .09445 .24727 .10636 .24881 .11419 .24676 Metetra
.598 .796 .976 r
.0903 .24368 .0825 .24573 .09445 .24727 .10226 .24522 Metetra
.598 .796 .975 r
.07831 .24213 .07052 .24419 .0825 .24573 .0903 .24368 Metetra
.598 .796 .975 r
.06629 .24058 .05851 .24264 .07052 .24419 .07831 .24213 Metetra
.598 .796 .975 r
.05423 .23902 .04646 .24109 .05851 .24264 .06629 .24058 Metetra
.598 .796 .975 r
.04213 .23746 .03438 .23954 .04646 .24109 .05423 .23902 Metetra
.598 .796 .975 r
.03 .23589 .02226 .23798 .03438 .23954 .04213 .23746 Metetra
.598 .796 .975 r
.01784 .23432 .01011 .23641 .02226 .23798 .03 .23589 Metetra
.568 .808 .989 r
.574 .45557 .56572 .45715 .57644 .45834 .58472 .45676 Metetra
.759 .491 .528 r
.56324 .45438 .55496 .44436 .56572 .45715 .574 .45557 Metetra
.766 .579 .646 r
.55246 .45319 .54415 .42039 .55496 .44436 .56324 .45438 Metetra
.878 .647 .585 r
.54161 .43043 .53333 .39525 .54415 .42039 .55246 .45319 Metetra
.212 .261 .697 r
.53074 .39187 .52249 .37199 .53333 .39525 .54161 .43043 Metetra
.445 .434 .753 r
.51986 .36189 .51165 .3519 .52249 .37199 .53074 .39187 Metetra
.539 .509 .771 r
.50898 .33908 .50079 .3353 .51165 .3519 .51986 .36189 Metetra
.59 .555 .783 r
.49809 .32202 .4899 .32201 .50079 .3353 .50898 .33908 Metetra
.622 .59 .797 r
.48717 .30947 .479 .31161 .4899 .32201 .49809 .32202 Metetra
.643 .622 .814 r
.47623 .30035 .46807 .30361 .479 .31161 .48717 .30947 Metetra
.658 .652 .833 r
.46527 .2938 .45711 .29753 .46807 .30361 .47623 .30035 Metetra
.667 .682 .855 r
.45427 .28913 .44613 .29293 .45711 .29753 .46527 .2938 Metetra
.671 .711 .878 r
.44325 .28578 .43512 .28944 .44613 .29293 .45427 .28913 Metetra
.669 .738 .901 r
.4322 .28335 .42407 .28676 .43512 .28944 .44325 .28578 Metetra
.662 .761 .923 r
.42112 .28153 .413 .28466 .42407 .28676 .4322 .28335 Metetra
.651 .78 .942 r
.41 .28009 .40189 .28294 .413 .28466 .42112 .28153 Metetra
.637 .793 .957 r
.39886 .27887 .39076 .28146 .40189 .28294 .41 .28009 Metetra
.624 .802 .968 r
.38768 .27774 .37959 .28011 .39076 .28146 .39886 .27887 Metetra
.612 .806 .974 r
.37648 .27663 .3684 .27884 .37959 .28011 .38768 .27774 Metetra
.602 .807 .978 r
.36524 .2755 .35717 .27758 .3684 .27884 .37648 .27663 Metetra
.596 .806 .98 r
.35397 .27431 .34591 .27631 .35717 .27758 .36524 .2755 Metetra
.593 .805 .981 r
.34267 .27306 .33463 .275 .34591 .27631 .35397 .27431 Metetra
.591 .803 .981 r
.33135 .27174 .32331 .27365 .33463 .275 .34267 .27306 Metetra
.591 .801 .98 r
.31999 .27037 .31196 .27227 .32331 .27365 .33135 .27174 Metetra
.592 .799 .979 r
.3086 .26894 .30058 .27085 .31196 .27227 .31999 .27037 Metetra
.594 .798 .978 r
.29718 .26748 .28917 .2694 .30058 .27085 .3086 .26894 Metetra
.595 .797 .977 r
.28573 .26599 .27772 .26793 .28917 .2694 .29718 .26748 Metetra
.596 .796 .976 r
.27424 .26449 .26625 .26644 .27772 .26793 .28573 .26599 Metetra
.597 .796 .975 r
.26273 .26298 .25475 .26495 .26625 .26644 .27424 .26449 Metetra
.598 .796 .975 r
.25118 .26146 .24321 .26345 .25475 .26495 .26273 .26298 Metetra
.598 .796 .975 r
.2396 .25995 .23164 .26194 .24321 .26345 .25118 .26146 Metetra
.598 .796 .975 r
.22799 .25843 .22004 .26044 .23164 .26194 .2396 .25995 Metetra
.598 .796 .975 r
.21635 .25692 .20841 .25893 .22004 .26044 .22799 .25843 Metetra
.598 .796 .975 r
.20468 .2554 .19675 .25742 .20841 .25893 .21635 .25692 Metetra
.598 .796 .975 r
.19297 .25389 .18505 .25591 .19675 .25742 .20468 .2554 Metetra
.598 .797 .975 r
.18123 .25237 .17332 .2544 .18505 .25591 .19297 .25389 Metetra
.598 .797 .976 r
.16946 .25085 .16156 .25288 .17332 .2544 .18123 .25237 Metetra
.598 .797 .976 r
.15765 .24932 .14977 .25136 .16156 .25288 .16946 .25085 Metetra
.598 .797 .976 r
.14582 .24779 .13794 .24983 .14977 .25136 .15765 .24932 Metetra
.598 .797 .976 r
.13395 .24625 .12608 .2483 .13794 .24983 .14582 .24779 Metetra
.598 .797 .976 r
.12204 .24471 .11419 .24676 .12608 .2483 .13395 .24625 Metetra
.598 .796 .976 r
.11011 .24316 .10226 .24522 .11419 .24676 .12204 .24471 Metetra
.598 .796 .975 r
.09814 .24161 .0903 .24368 .10226 .24522 .11011 .24316 Metetra
.598 .796 .975 r
.08613 .24006 .07831 .24213 .0903 .24368 .09814 .24161 Metetra
.598 .796 .975 r
.07409 .2385 .06629 .24058 .07831 .24213 .08613 .24006 Metetra
.598 .796 .975 r
.06202 .23694 .05423 .23902 .06629 .24058 .07409 .2385 Metetra
.598 .796 .975 r
.04992 .23537 .04213 .23746 .05423 .23902 .06202 .23694 Metetra
.598 .796 .975 r
.03778 .2338 .03 .23589 .04213 .23746 .04992 .23537 Metetra
.598 .796 .975 r
.0256 .23222 .01784 .23432 .03 .23589 .03778 .2338 Metetra
.568 .808 .989 r
.5823 .45399 .574 .45557 .58472 .45676 .59303 .45518 Metetra
.568 .808 .989 r
.57153 .45279 .56324 .45438 .574 .45557 .5823 .45399 Metetra
.568 .808 .989 r
.56074 .4516 .55246 .45319 .56324 .45438 .57153 .45279 Metetra
.766 .556 .614 r
.54992 .45039 .54161 .43043 .55246 .45319 .56074 .4516 Metetra
.51 .449 .724 r
.53901 .40318 .53074 .39187 .54161 .43043 .54992 .45039 Metetra
.575 .5 .733 r
.5281 .36473 .51986 .36189 .53074 .39187 .53901 .40318 Metetra
.61 .534 .745 r
.51721 .33709 .50898 .33908 .51986 .36189 .5281 .36473 Metetra
.635 .564 .758 r
.5063 .31752 .49809 .32202 .50898 .33908 .51721 .33709 Metetra
.655 .593 .773 r
.49537 .30388 .48717 .30947 .49809 .32202 .5063 .31752 Metetra
.67 .621 .792 r
.48442 .29452 .47623 .30035 .48717 .30947 .49537 .30388 Metetra
.681 .651 .814 r
.47345 .2882 .46527 .2938 .47623 .30035 .48442 .29452 Metetra
.688 .682 .839 r
.46245 .28399 .45427 .28913 .46527 .2938 .47345 .2882 Metetra
.69 .714 .867 r
.45142 .2812 .44325 .28578 .45427 .28913 .46245 .28399 Metetra
.686 .745 .896 r
.44035 .27933 .4322 .28335 .44325 .28578 .45142 .2812 Metetra
.675 .772 .923 r
.42926 .27803 .42112 .28153 .4322 .28335 .44035 .27933 Metetra
.659 .793 .946 r
.41814 .27703 .41 .28009 .42112 .28153 .42926 .27803 Metetra
.639 .806 .963 r
.40698 .27617 .39886 .27887 .41 .28009 .41814 .27703 Metetra
.621 .813 .974 r
.3958 .27533 .38768 .27774 .39886 .27887 .40698 .27617 Metetra
.605 .815 .981 r
.38458 .27445 .37648 .27663 .38768 .27774 .3958 .27533 Metetra
.594 .814 .984 r
.37334 .27346 .36524 .2755 .37648 .27663 .38458 .27445 Metetra
.588 .811 .985 r
.36206 .27238 .35397 .27431 .36524 .2755 .37334 .27346 Metetra
.586 .808 .984 r
.35075 .27118 .34267 .27306 .35397 .27431 .36206 .27238 Metetra
.586 .805 .983 r
.33942 .26988 .33135 .27174 .34267 .27306 .35075 .27118 Metetra
.588 .802 .981 r
.32805 .2685 .31999 .27037 .33135 .27174 .33942 .26988 Metetra
.59 .799 .979 r
.31665 .26706 .3086 .26894 .31999 .27037 .32805 .2685 Metetra
.592 .797 .978 r
.30522 .26557 .29718 .26748 .3086 .26894 .31665 .26706 Metetra
.595 .796 .976 r
.29376 .26406 .28573 .26599 .29718 .26748 .30522 .26557 Metetra
.596 .795 .976 r
.28226 .26253 .27424 .26449 .28573 .26599 .29376 .26406 Metetra
.598 .795 .975 r
.27074 .261 .26273 .26298 .27424 .26449 .28226 .26253 Metetra
.598 .795 .975 r
.25918 .25946 .25118 .26146 .26273 .26298 .27074 .261 Metetra
.599 .796 .975 r
.24759 .25794 .2396 .25995 .25118 .26146 .25918 .25946 Metetra
.599 .796 .975 r
.23597 .25641 .22799 .25843 .2396 .25995 .24759 .25794 Metetra
.598 .796 .975 r
.22432 .25489 .21635 .25692 .22799 .25843 .23597 .25641 Metetra
.598 .796 .975 r
.21264 .25338 .20468 .2554 .21635 .25692 .22432 .25489 Metetra
.598 .797 .975 r
.20092 .25186 .19297 .25389 .20468 .2554 .21264 .25338 Metetra
.598 .797 .976 r
.18917 .25034 .18123 .25237 .19297 .25389 .20092 .25186 Metetra
.598 .797 .976 r
.17739 .24881 .16946 .25085 .18123 .25237 .18917 .25034 Metetra
.598 .797 .976 r
.16557 .24728 .15765 .24932 .16946 .25085 .17739 .24881 Metetra
.598 .797 .976 r
.15372 .24574 .14582 .24779 .15765 .24932 .16557 .24728 Metetra
.598 .797 .976 r
.14184 .2442 .13395 .24625 .14582 .24779 .15372 .24574 Metetra
.598 .797 .976 r
.12993 .24265 .12204 .24471 .13395 .24625 .14184 .2442 Metetra
.598 .796 .976 r
.11798 .2411 .11011 .24316 .12204 .24471 .12993 .24265 Metetra
.598 .796 .975 r
.106 .23954 .09814 .24161 .11011 .24316 .11798 .2411 Metetra
.598 .796 .975 r
.09398 .23798 .08613 .24006 .09814 .24161 .106 .23954 Metetra
.598 .796 .975 r
.08193 .23642 .07409 .2385 .08613 .24006 .09398 .23798 Metetra
.598 .796 .975 r
.06985 .23485 .06202 .23694 .07409 .2385 .08193 .23642 Metetra
.598 .796 .975 r
.05773 .23328 .04992 .23537 .06202 .23694 .06985 .23485 Metetra
.598 .796 .975 r
.04558 .2317 .03778 .2338 .04992 .23537 .05773 .23328 Metetra
.598 .796 .975 r
.03339 .23011 .0256 .23222 .03778 .2338 .04558 .2317 Metetra
.568 .808 .989 r
.59063 .4524 .5823 .45399 .59303 .45518 .60137 .45359 Metetra
.568 .808 .989 r
.57985 .4512 .57153 .45279 .5823 .45399 .59063 .4524 Metetra
.568 .808 .989 r
.56905 .45 .56074 .4516 .57153 .45279 .57985 .4512 Metetra
.568 .808 .989 r
.55822 .44879 .54992 .45039 .56074 .4516 .56905 .45 Metetra
.609 .513 .721 r
.54729 .40037 .53901 .40318 .54992 .45039 .55822 .44879 Metetra
.63 .53 .722 r
.53636 .35683 .5281 .36473 .53901 .40318 .54729 .40037 Metetra
.65 .553 .732 r
.52545 .32708 .51721 .33709 .5281 .36473 .53636 .35683 Metetra
.667 .577 .744 r
.51453 .30713 .5063 .31752 .51721 .33709 .52545 .32708 Metetra
.682 .601 .759 r
.5036 .29403 .49537 .30388 .5063 .31752 .51453 .30713 Metetra
.694 .628 .779 r
.49264 .28566 .48442 .29452 .49537 .30388 .5036 .29403 Metetra
.704 .659 .803 r
.48166 .28049 .47345 .2882 .48442 .29452 .49264 .28566 Metetra
.71 .692 .831 r
.47065 .27743 .46245 .28399 .47345 .2882 .48166 .28049 Metetra
.71 .728 .864 r
.45961 .27569 .45142 .2812 .46245 .28399 .47065 .27743 Metetra
.703 .763 .899 r
.44854 .27474 .44035 .27933 .45142 .2812 .45961 .27569 Metetra
.686 .793 .931 r
.43744 .27421 .42926 .27803 .44035 .27933 .44854 .27474 Metetra
.662 .815 .957 r
.4263 .27383 .41814 .27703 .42926 .27803 .43744 .27421 Metetra
.634 .826 .974 r
.41514 .27345 .40698 .27617 .41814 .27703 .4263 .27383 Metetra
.609 .829 .984 r
.40395 .27298 .3958 .27533 .40698 .27617 .41514 .27345 Metetra
.59 .827 .989 r
.39272 .27234 .38458 .27445 .3958 .27533 .40395 .27298 Metetra
.579 .822 .99 r
.38146 .27153 .37334 .27346 .38458 .27445 .39272 .27234 Metetra
.575 .816 .99 r
.37018 .27054 .36206 .27238 .37334 .27346 .38146 .27153 Metetra
.575 .811 .988 r
.35886 .26939 .35075 .27118 .36206 .27238 .37018 .27054 Metetra
.579 .806 .986 r
.34751 .2681 .33942 .26988 .35075 .27118 .35886 .26939 Metetra
.583 .802 .983 r
.33614 .26669 .32805 .2685 .33942 .26988 .34751 .2681 Metetra
.587 .798 .98 r
.32473 .26522 .31665 .26706 .32805 .2685 .33614 .26669 Metetra
.591 .796 .978 r
.31329 .26368 .30522 .26557 .31665 .26706 .32473 .26522 Metetra
.595 .795 .976 r
.30182 .26213 .29376 .26406 .30522 .26557 .31329 .26368 Metetra
.597 .794 .975 r
.29031 .26056 .28226 .26253 .29376 .26406 .30182 .26213 Metetra
.598 .794 .974 r
.27878 .259 .27074 .261 .28226 .26253 .29031 .26056 Metetra
.599 .795 .974 r
.26721 .25745 .25918 .25946 .27074 .261 .27878 .259 Metetra
.599 .795 .974 r
.25561 .25591 .24759 .25794 .25918 .25946 .26721 .25745 Metetra
.599 .796 .975 r
.24398 .25438 .23597 .25641 .24759 .25794 .25561 .25591 Metetra
.599 .796 .975 r
.23232 .25286 .22432 .25489 .23597 .25641 .24398 .25438 Metetra
.598 .797 .975 r
.22062 .25134 .21264 .25338 .22432 .25489 .23232 .25286 Metetra
.598 .797 .975 r
.2089 .24982 .20092 .25186 .21264 .25338 .22062 .25134 Metetra
.598 .797 .976 r
.19714 .2483 .18917 .25034 .20092 .25186 .2089 .24982 Metetra
.598 .797 .976 r
.18534 .24677 .17739 .24881 .18917 .25034 .19714 .2483 Metetra
.598 .797 .976 r
.17352 .24523 .16557 .24728 .17739 .24881 .18534 .24677 Metetra
.598 .797 .976 r
.16166 .24369 .15372 .24574 .16557 .24728 .17352 .24523 Metetra
.598 .797 .976 r
.14976 .24214 .14184 .2442 .15372 .24574 .16166 .24369 Metetra
.598 .796 .976 r
.13784 .24059 .12993 .24265 .14184 .2442 .14976 .24214 Metetra
.598 .796 .975 r
.12588 .23903 .11798 .2411 .12993 .24265 .13784 .24059 Metetra
.598 .796 .975 r
.11389 .23746 .106 .23954 .11798 .2411 .12588 .23903 Metetra
.598 .796 .975 r
.10186 .2359 .09398 .23798 .106 .23954 .11389 .23746 Metetra
.598 .796 .975 r
.0898 .23433 .08193 .23642 .09398 .23798 .10186 .2359 Metetra
.598 .796 .975 r
.0777 .23275 .06985 .23485 .08193 .23642 .0898 .23433 Metetra
.598 .796 .975 r
.06557 .23117 .05773 .23328 .06985 .23485 .0777 .23275 Metetra
.598 .796 .975 r
.05341 .22959 .04558 .2317 .05773 .23328 .06557 .23117 Metetra
.598 .796 .975 r
.04121 .228 .03339 .23011 .04558 .2317 .05341 .22959 Metetra
.568 .808 .989 r
.59899 .4508 .59063 .4524 .60137 .45359 .60974 .452 Metetra
.568 .808 .989 r
.58821 .4496 .57985 .4512 .59063 .4524 .59899 .4508 Metetra
.568 .808 .989 r
.5774 .44839 .56905 .45 .57985 .4512 .58821 .4496 Metetra
.681 .689 .851 r
.56655 .44434 .55822 .44879 .56905 .45 .5774 .44839 Metetra
.65 .535 .71 r
.55557 .37863 .54729 .40037 .55822 .44879 .56655 .44434 Metetra
.666 .553 .717 r
.54463 .33532 .53636 .35683 .54729 .40037 .55557 .37863 Metetra
.681 .571 .726 r
.53371 .3074 .52545 .32708 .53636 .35683 .54463 .33532 Metetra
.694 .592 .737 r
.52279 .28994 .51453 .30713 .52545 .32708 .53371 .3074 Metetra
.707 .616 .752 r
.51185 .27948 .5036 .29403 .51453 .30713 .52279 .28994 Metetra
.719 .643 .772 r
.50089 .27362 .49264 .28566 .5036 .29403 .51185 .27948 Metetra
.729 .676 .799 r
.4899 .2707 .48166 .28049 .49264 .28566 .50089 .27362 Metetra
.735 .714 .832 r
.47888 .26955 .47065 .27743 .48166 .28049 .4899 .2707 Metetra
.733 .756 .871 r
.46783 .26941 .45961 .27569 .47065 .27743 .47888 .26955 Metetra
.72 .797 .912 r
.45675 .26974 .44854 .27474 .45961 .27569 .46783 .26941 Metetra
.692 .83 .949 r
.44564 .27022 .43744 .27421 .44854 .27474 .45675 .26974 Metetra
.654 .849 .975 r
.4345 .27063 .4263 .27383 .43744 .27421 .44564 .27022 Metetra
.613 .853 .989 r
.42333 .27083 .41514 .27345 .4263 .27383 .4345 .27063 Metetra
.58 .848 .995 r
.41212 .27077 .40395 .27298 .41514 .27345 .42333 .27083 Metetra
.561 .839 .997 r
.40089 .27041 .39272 .27234 .40395 .27298 .41212 .27077 Metetra
.553 .829 .997 r
.38962 .26977 .38146 .27153 .39272 .27234 .40089 .27041 Metetra
.555 .82 .995 r
.37832 .26886 .37018 .27054 .38146 .27153 .38962 .26977 Metetra
.561 .813 .992 r
.367 .26772 .35886 .26939 .37018 .27054 .37832 .26886 Metetra
.569 .806 .988 r
.35564 .26641 .34751 .2681 .35886 .26939 .367 .26772 Metetra
.578 .801 .984 r
.34425 .26495 .33614 .26669 .34751 .2681 .35564 .26641 Metetra
.585 .797 .98 r
.33284 .26341 .32473 .26522 .33614 .26669 .34425 .26495 Metetra
.591 .794 .977 r
.32139 .26181 .31329 .26368 .32473 .26522 .33284 .26341 Metetra
.595 .793 .975 r
.3099 .26019 .30182 .26213 .31329 .26368 .32139 .26181 Metetra
.598 .793 .973 r
.29839 .25858 .29031 .26056 .30182 .26213 .3099 .26019 Metetra
.6 .793 .973 r
.28685 .25698 .27878 .259 .29031 .26056 .29839 .25858 Metetra
.601 .794 .973 r
.27527 .25541 .26721 .25745 .27878 .259 .28685 .25698 Metetra
.601 .795 .974 r
.26366 .25386 .25561 .25591 .26721 .25745 .27527 .25541 Metetra
.6 .796 .974 r
.25202 .25233 .24398 .25438 .25561 .25591 .26366 .25386 Metetra
.599 .796 .975 r
.24035 .25081 .23232 .25286 .24398 .25438 .25202 .25233 Metetra
.598 .797 .975 r
.22864 .24929 .22062 .25134 .23232 .25286 .24035 .25081 Metetra
.598 .797 .976 r
.2169 .24777 .2089 .24982 .22062 .25134 .22864 .24929 Metetra
.598 .797 .976 r
.20513 .24625 .19714 .2483 .2089 .24982 .2169 .24777 Metetra
.597 .797 .976 r
.19333 .24472 .18534 .24677 .19714 .2483 .20513 .24625 Metetra
.597 .797 .976 r
.18149 .24318 .17352 .24523 .18534 .24677 .19333 .24472 Metetra
.597 .797 .976 r
.16962 .24163 .16166 .24369 .17352 .24523 .18149 .24318 Metetra
.598 .797 .976 r
.15772 .24007 .14976 .24214 .16166 .24369 .16962 .24163 Metetra
.598 .796 .976 r
.14578 .23851 .13784 .24059 .14976 .24214 .15772 .24007 Metetra
.598 .796 .975 r
.13381 .23695 .12588 .23903 .13784 .24059 .14578 .23851 Metetra
.598 .796 .975 r
.1218 .23538 .11389 .23746 .12588 .23903 .13381 .23695 Metetra
.598 .796 .975 r
.10977 .2338 .10186 .2359 .11389 .23746 .1218 .23538 Metetra
.598 .796 .975 r
.09769 .23223 .0898 .23433 .10186 .2359 .10977 .2338 Metetra
.598 .796 .975 r
.08559 .23065 .0777 .23275 .0898 .23433 .09769 .23223 Metetra
.598 .796 .975 r
.07345 .22906 .06557 .23117 .0777 .23275 .08559 .23065 Metetra
.598 .796 .975 r
.06127 .22747 .05341 .22959 .06557 .23117 .07345 .22906 Metetra
.598 .796 .976 r
.04906 .22588 .04121 .228 .05341 .22959 .06127 .22747 Metetra
.568 .808 .989 r
.60738 .4492 .59899 .4508 .60974 .452 .61813 .4504 Metetra
.568 .808 .989 r
.59659 .44799 .58821 .4496 .59899 .4508 .60738 .4492 Metetra
.568 .808 .989 r
.58577 .44678 .5774 .44839 .58821 .4496 .59659 .44799 Metetra
.694 .567 .708 r
.5748 .39397 .56655 .44434 .5774 .44839 .58577 .44678 Metetra
.683 .559 .708 r
.56382 .33493 .55557 .37863 .56655 .44434 .5748 .39397 Metetra
.696 .574 .714 r
.55289 .29867 .54463 .33532 .55557 .37863 .56382 .33493 Metetra
.709 .592 .722 r
.54198 .2774 .53371 .3074 .54463 .33532 .55289 .29867 Metetra
.722 .612 .734 r
.53106 .26583 .52279 .28994 .53371 .3074 .54198 .2774 Metetra
.736 .637 .75 r
.52012 .2604 .51185 .27948 .52279 .28994 .53106 .26583 Metetra
.749 .669 .772 r
.50915 .2587 .50089 .27362 .51185 .27948 .52012 .2604 Metetra
.76 .708 .803 r
.49816 .25914 .4899 .2707 .50089 .27362 .50915 .2587 Metetra
.767 .756 .844 r
.48714 .26068 .47888 .26955 .4899 .2707 .49816 .25914 Metetra
.761 .81 .892 r
.47608 .26265 .46783 .26941 .47888 .26955 .48714 .26068 Metetra
.733 .859 .94 r
.46499 .26461 .45675 .26974 .46783 .26941 .47608 .26265 Metetra
.678 .888 .975 r
.45387 .26632 .44564 .27022 .45675 .26974 .46499 .26461 Metetra
.609 .89 .989 r
.44272 .26763 .4345 .27063 .44564 .27022 .45387 .26632 Metetra
.548 .876 .992 r
.43154 .26848 .42333 .27083 .4345 .27063 .44272 .26763 Metetra
.513 .858 .993 r
.42033 .26885 .41212 .27077 .42333 .27083 .43154 .26848 Metetra
.502 .843 .997 r
.40908 .26877 .40089 .27041 .41212 .27077 .42033 .26885 Metetra
.509 .831 .999 r
.39781 .26826 .38962 .26977 .40089 .27041 .40908 .26877 Metetra
.524 .821 .999 r
.3865 .2674 .37832 .26886 .38962 .26977 .39781 .26826 Metetra
.542 .812 .996 r
.37516 .26624 .367 .26772 .37832 .26886 .3865 .2674 Metetra
.558 .804 .991 r
.3638 .26485 .35564 .26641 .367 .26772 .37516 .26624 Metetra
.572 .798 .985 r
.3524 .26329 .34425 .26495 .35564 .26641 .3638 .26485 Metetra
.583 .793 .979 r
.34097 .26164 .33284 .26341 .34425 .26495 .3524 .26329 Metetra
.592 .791 .975 r
.32951 .25994 .32139 .26181 .33284 .26341 .34097 .26164 Metetra
.597 .79 .973 r
.31802 .25824 .3099 .26019 .32139 .26181 .32951 .25994 Metetra
.601 .79 .971 r
.3065 .25656 .29839 .25858 .3099 .26019 .31802 .25824 Metetra
.603 .791 .971 r
.29495 .25493 .28685 .25698 .29839 .25858 .3065 .25656 Metetra
.603 .793 .972 r
.28336 .25334 .27527 .25541 .28685 .25698 .29495 .25493 Metetra
.602 .795 .973 r
.27174 .25178 .26366 .25386 .27527 .25541 .28336 .25334 Metetra
.601 .796 .974 r
.26009 .25025 .25202 .25233 .26366 .25386 .27174 .25178 Metetra
.6 .797 .975 r
.24841 .24874 .24035 .25081 .25202 .25233 .26009 .25025 Metetra
.598 .797 .976 r
.23669 .24723 .22864 .24929 .24035 .25081 .24841 .24874 Metetra
.598 .797 .976 r
.22494 .24572 .2169 .24777 .22864 .24929 .23669 .24723 Metetra
.597 .797 .976 r
.21316 .2442 .20513 .24625 .2169 .24777 .22494 .24572 Metetra
.597 .797 .976 r
.20134 .24266 .19333 .24472 .20513 .24625 .21316 .2442 Metetra
.597 .797 .976 r
.1895 .24112 .18149 .24318 .19333 .24472 .20134 .24266 Metetra
.597 .797 .976 r
.17761 .23956 .16962 .24163 .18149 .24318 .1895 .24112 Metetra
.598 .796 .976 r
.1657 .238 .15772 .24007 .16962 .24163 .17761 .23956 Metetra
.598 .796 .975 r
.15375 .23643 .14578 .23851 .15772 .24007 .1657 .238 Metetra
.598 .796 .975 r
.14177 .23486 .13381 .23695 .14578 .23851 .15375 .23643 Metetra
.598 .796 .975 r
.12975 .23328 .1218 .23538 .13381 .23695 .14177 .23486 Metetra
.598 .796 .975 r
.1177 .2317 .10977 .2338 .1218 .23538 .12975 .23328 Metetra
.598 .796 .975 r
.10562 .23012 .09769 .23223 .10977 .2338 .1177 .2317 Metetra
.598 .796 .975 r
.0935 .22853 .08559 .23065 .09769 .23223 .10562 .23012 Metetra
.598 .797 .975 r
.08135 .22694 .07345 .22906 .08559 .23065 .0935 .22853 Metetra
.598 .797 .976 r
.06916 .22535 .06127 .22747 .07345 .22906 .08135 .22694 Metetra
.598 .797 .976 r
.05694 .22375 .04906 .22588 .06127 .22747 .06916 .22535 Metetra
.568 .808 .989 r
.61579 .44759 .60738 .4492 .61813 .4504 .62656 .4488 Metetra
.568 .808 .989 r
.605 .44638 .59659 .44799 .60738 .4492 .61579 .44759 Metetra
.696 .567 .705 r
.59399 .38274 .58577 .44678 .59659 .44799 .605 .44638 Metetra
.702 .569 .703 r
.58296 .31033 .5748 .39397 .58577 .44678 .59399 .38274 Metetra
.714 .583 .707 r
.57203 .26902 .56382 .33493 .5748 .39397 .58296 .31033 Metetra
.727 .598 .713 r
.56113 .24737 .55289 .29867 .56382 .33493 .57203 .26902 Metetra
.741 .617 .721 r
.55024 .2379 .54198 .2774 .55289 .29867 .56113 .24737 Metetra
.756 .641 .734 r
.53933 .23572 .53106 .26583 .54198 .2774 .55024 .2379 Metetra
.773 .672 .752 r
.5284 .23765 .52012 .2604 .53106 .26583 .53933 .23572 Metetra
.791 .714 .779 r
.51744 .24167 .50915 .2587 .52012 .2604 .5284 .23765 Metetra
.807 .77 .818 r
.50645 .24652 .49816 .25914 .50915 .2587 .51744 .24167 Metetra
.812 .84 .871 r
.49542 .25141 .48714 .26068 .49816 .25914 .50645 .24652 Metetra
.785 .913 .927 r
.48436 .2559 .47608 .26265 .48714 .26068 .49542 .25141 Metetra
.695 .95 .956 r
.47326 .25974 .46499 .26461 .47608 .26265 .48436 .2559 Metetra
.554 .923 .938 r
.46214 .26281 .45387 .26632 .46499 .26461 .47326 .25974 Metetra
.429 .867 .908 r
.45098 .26509 .44272 .26763 .45387 .26632 .46214 .26281 Metetra
.364 .829 .905 r
.43978 .2666 .43154 .26848 .44272 .26763 .45098 .26509 Metetra
.358 .817 .93 r
.42856 .26739 .42033 .26885 .43154 .26848 .43978 .2666 Metetra
.39 .818 .964 r
.4173 .26753 .40908 .26877 .42033 .26885 .42856 .26739 Metetra
.436 .819 .989 r
.40602 .26711 .39781 .26826 .40908 .26877 .4173 .26753 Metetra
.481 .816 .999 r
.39471 .26623 .3865 .2674 .39781 .26826 .40602 .26711 Metetra
.518 .808 .998 r
.38336 .26497 .37516 .26624 .3865 .2674 .39471 .26623 Metetra
.547 .8 .992 r
.37198 .26343 .3638 .26485 .37516 .26624 .38336 .26497 Metetra
.568 .793 .984 r
.36058 .26172 .3524 .26329 .3638 .26485 .37198 .26343 Metetra
.584 .788 .977 r
.34914 .25991 .34097 .26164 .3524 .26329 .36058 .26172 Metetra
.594 .786 .972 r
.33767 .25807 .32951 .25994 .34097 .26164 .34914 .25991 Metetra
.601 .786 .969 r
.32617 .25625 .31802 .25824 .32951 .25994 .33767 .25807 Metetra
.605 .787 .968 r
.31464 .2545 .3065 .25656 .31802 .25824 .32617 .25625 Metetra
.607 .789 .969 r
.30308 .25282 .29495 .25493 .3065 .25656 .31464 .2545 Metetra
.606 .792 .97 r
.29148 .25121 .28336 .25334 .29495 .25493 .30308 .25282 Metetra
.604 .795 .972 r
.27985 .24966 .27174 .25178 .28336 .25334 .29148 .25121 Metetra
.602 .796 .974 r
.26819 .24815 .26009 .25025 .27174 .25178 .27985 .24966 Metetra
.6 .798 .975 r
.25649 .24666 .24841 .24874 .26009 .25025 .26819 .24815 Metetra
.598 .798 .976 r
.24477 .24517 .23669 .24723 .24841 .24874 .25649 .24666 Metetra
.597 .798 .977 r
.23301 .24367 .22494 .24572 .23669 .24723 .24477 .24517 Metetra
.596 .798 .977 r
.22121 .24215 .21316 .2442 .22494 .24572 .23301 .24367 Metetra
.596 .797 .976 r
.20939 .24061 .20134 .24266 .21316 .2442 .22121 .24215 Metetra
.597 .797 .976 r
.19753 .23906 .1895 .24112 .20134 .24266 .20939 .24061 Metetra
.597 .796 .976 r
.18564 .23749 .17761 .23956 .1895 .24112 .19753 .23906 Metetra
.598 .796 .975 r
.17371 .23592 .1657 .238 .17761 .23956 .18564 .23749 Metetra
.598 .796 .975 r
.16175 .23434 .15375 .23643 .1657 .238 .17371 .23592 Metetra
.598 .796 .975 r
.14976 .23276 .14177 .23486 .15375 .23643 .16175 .23434 Metetra
.598 .796 .975 r
.13773 .23118 .12975 .23328 .14177 .23486 .14976 .23276 Metetra
.598 .796 .975 r
.12567 .22959 .1177 .2317 .12975 .23328 .13773 .23118 Metetra
.598 .796 .975 r
.11358 .228 .10562 .23012 .1177 .2317 .12567 .22959 Metetra
.598 .797 .975 r
.10145 .22641 .0935 .22853 .10562 .23012 .11358 .228 Metetra
.598 .797 .976 r
.08928 .22482 .08135 .22694 .0935 .22853 .10145 .22641 Metetra
.598 .797 .976 r
.07708 .22322 .06916 .22535 .08135 .22694 .08928 .22482 Metetra
.598 .797 .976 r
.06485 .22161 .05694 .22375 .06916 .22535 .07708 .22322 Metetra
.694 .625 .775 r
.62421 .4384 .61579 .44759 .62656 .4488 .63502 .44719 Metetra
.698 .563 .698 r
.61294 .30446 .605 .44638 .61579 .44759 .62421 .4384 Metetra
.723 .584 .699 r
.60192 .2319 .59399 .38274 .605 .44638 .61294 .30446 Metetra
.735 .597 .702 r
.59103 .19681 .58296 .31033 .59399 .38274 .60192 .2319 Metetra
.749 .612 .706 r
.58018 .18412 .57203 .26902 .58296 .31033 .59103 .19681 Metetra
.765 .631 .712 r
.56935 .18428 .56113 .24737 .57203 .26902 .58018 .18412 Metetra
.784 .656 .721 r
.5585 .19134 .55024 .2379 .56113 .24737 .56935 .18428 Metetra
.807 .69 .735 r
.54762 .20161 .53933 .23572 .55024 .2379 .5585 .19134 Metetra
.834 .738 .758 r
.53671 .21289 .5284 .23765 .53933 .23572 .54762 .20161 Metetra
.864 .808 .793 r
.52575 .22387 .51744 .24167 .5284 .23765 .53671 .21289 Metetra
.883 .906 .839 r
.51476 .23388 .50645 .24652 .51744 .24167 .52575 .22387 Metetra
.825 .994 .86 r
.50373 .24257 .49542 .25141 .50645 .24652 .51476 .23388 Metetra
.571 .925 .747 r
.49266 .24982 .48436 .2559 .49542 .25141 .50373 .24257 Metetra
.48156 .25564 .47326 .25974 .48436 .2559 .49266 .24982 Metetra
.47043 .2601 .46214 .26281 .47326 .25974 .48156 .25564 Metetra
.036 0 0 r
.45926 .26333 .45098 .26509 .46214 .26281 .47043 .2601 Metetra
.021 0 0 r
.44805 .26544 .43978 .2666 .45098 .26509 .45926 .26333 Metetra
.43682 .26657 .42856 .26739 .43978 .2666 .44805 .26544 Metetra
.193 .719 .849 r
.42556 .26686 .4173 .26753 .42856 .26739 .43682 .26657 Metetra
.321 .778 .95 r
.41426 .26643 .40602 .26711 .4173 .26753 .42556 .26686 Metetra
.422 .8 .992 r
.40294 .26542 .39471 .26623 .40602 .26711 .41426 .26643 Metetra
.491 .8 .999 r
.39158 .26397 .38336 .26497 .39471 .26623 .40294 .26542 Metetra
.537 .792 .992 r
.3802 .26219 .37198 .26343 .38336 .26497 .39158 .26397 Metetra
.567 .785 .981 r
.36878 .26023 .36058 .26172 .37198 .26343 .3802 .26219 Metetra
.587 .78 .972 r
.35734 .25818 .34914 .25991 .36058 .26172 .36878 .26023 Metetra
.6 .778 .966 r
.34586 .25615 .33767 .25807 .34914 .25991 .35734 .25818 Metetra
.608 .779 .963 r
.33435 .25418 .32617 .25625 .33767 .25807 .34586 .25615 Metetra
.612 .783 .963 r
.32281 .25234 .31464 .2545 .32617 .25625 .33435 .25418 Metetra
.612 .787 .965 r
.31124 .25062 .30308 .25282 .31464 .2545 .32281 .25234 Metetra
.61 .791 .968 r
.29963 .24901 .29148 .25121 .30308 .25282 .31124 .25062 Metetra
.607 .795 .971 r
.28799 .24749 .27985 .24966 .29148 .25121 .29963 .24901 Metetra
.603 .798 .974 r
.27632 .24602 .26819 .24815 .27985 .24966 .28799 .24749 Metetra
.599 .799 .976 r
.26461 .24457 .25649 .24666 .26819 .24815 .27632 .24602 Metetra
.597 .8 .977 r
.25287 .24311 .24477 .24517 .25649 .24666 .26461 .24457 Metetra
.595 .799 .978 r
.2411 .24162 .23301 .24367 .24477 .24517 .25287 .24311 Metetra
.595 .798 .977 r
.2293 .24011 .22121 .24215 .23301 .24367 .2411 .24162 Metetra
.596 .797 .977 r
.21746 .23856 .20939 .24061 .22121 .24215 .2293 .24011 Metetra
.596 .797 .976 r
.20559 .237 .19753 .23906 .20939 .24061 .21746 .23856 Metetra
.597 .796 .976 r
.19369 .23542 .18564 .23749 .19753 .23906 .20559 .237 Metetra
.598 .796 .975 r
.18175 .23383 .17371 .23592 .18564 .23749 .19369 .23542 Metetra
.598 .796 .975 r
.16978 .23224 .16175 .23434 .17371 .23592 .18175 .23383 Metetra
.598 .796 .975 r
.15778 .23065 .14976 .23276 .16175 .23434 .16978 .23224 Metetra
.598 .796 .975 r
.14574 .22906 .13773 .23118 .14976 .23276 .15778 .23065 Metetra
.598 .796 .975 r
.13367 .22747 .12567 .22959 .13773 .23118 .14574 .22906 Metetra
.598 .797 .975 r
.12156 .22588 .11358 .228 .12567 .22959 .13367 .22747 Metetra
.598 .797 .976 r
.10942 .22428 .10145 .22641 .11358 .228 .12156 .22588 Metetra
.598 .797 .976 r
.09725 .22268 .08928 .22482 .10145 .22641 .10942 .22428 Metetra
.598 .797 .976 r
.08504 .22107 .07708 .22322 .08928 .22482 .09725 .22268 Metetra
.598 .797 .976 r
.07279 .21946 .06485 .22161 .07708 .22322 .08504 .22107 Metetra
.761 .614 .695 r
.63162 .18008 .62421 .4384 .63502 .44719 .64235 .18869 Metetra
.764 .621 .701 r
.6209 .17852 .61294 .30446 .62421 .4384 .63162 .18008 Metetra
.762 .63 .715 r
.61014 .17695 .60192 .2319 .61294 .30446 .6209 .17852 Metetra
.755 .655 .75 r
.59936 .17538 .59103 .19681 .60192 .2319 .61014 .17695 Metetra
.741 .69 .802 r
.58854 .1738 .58018 .18412 .59103 .19681 .59936 .17538 Metetra
.745 .681 .789 r
.5777 .17222 .56935 .18428 .58018 .18412 .58854 .1738 Metetra
.755 .656 .752 r
.56683 .17064 .5585 .19134 .56935 .18428 .5777 .17222 Metetra
.759 .642 .731 r
.55593 .16905 .54762 .20161 .5585 .19134 .56683 .17064 Metetra
.923 .85 .753 r
.54503 .18849 .53671 .21289 .54762 .20161 .55593 .16905 Metetra
.905 .969 .661 r
.53409 .20711 .52575 .22387 .53671 .21289 .54503 .18849 Metetra
.5231 .22259 .51476 .23388 .52575 .22387 .53409 .20711 Metetra
.009 0 .019 r
.51207 .23518 .50373 .24257 .51476 .23388 .5231 .22259 Metetra
.245 0 .153 r
.501 .24518 .49266 .24982 .50373 .24257 .51207 .23518 Metetra
.354 0 .183 r
.48989 .2529 .48156 .25564 .49266 .24982 .501 .24518 Metetra
.403 0 .153 r
.47874 .25864 .47043 .2601 .48156 .25564 .48989 .2529 Metetra
.411 0 .07 r
.46757 .26268 .45926 .26333 .47043 .2601 .47874 .25864 Metetra
.374 0 0 r
.45636 .26527 .44805 .26544 .45926 .26333 .46757 .26268 Metetra
.27 0 0 r
.44511 .2666 .43682 .26657 .44805 .26544 .45636 .26527 Metetra
.077 0 0 r
.43384 .26689 .42556 .26686 .43682 .26657 .44511 .2666 Metetra
.162 .69 .864 r
.42254 .26632 .41426 .26643 .42556 .26686 .43384 .26689 Metetra
.35 .77 .977 r
.4112 .26505 .40294 .26542 .41426 .26643 .42254 .26632 Metetra
.465 .786 .999 r
.39984 .26327 .39158 .26397 .40294 .26542 .4112 .26505 Metetra
.53 .781 .989 r
.38845 .26114 .3802 .26219 .39158 .26397 .39984 .26327 Metetra
.569 .773 .975 r
.37702 .25881 .36878 .26023 .3802 .26219 .38845 .26114 Metetra
.593 .769 .963 r
.36557 .25644 .35734 .25818 .36878 .26023 .37702 .25881 Metetra
.608 .768 .956 r
.35408 .25413 .34586 .25615 .35734 .25818 .36557 .25644 Metetra
.617 .771 .954 r
.34256 .25199 .33435 .25418 .34586 .25615 .35408 .25413 Metetra
.62 .777 .956 r
.33101 .25004 .32281 .25234 .33435 .25418 .34256 .25199 Metetra
.62 .784 .96 r
.31943 .24829 .31124 .25062 .32281 .25234 .33101 .25004 Metetra
.615 .791 .966 r
.30781 .24671 .29963 .24901 .31124 .25062 .31943 .24829 Metetra
.609 .797 .971 r
.29616 .24526 .28799 .24749 .29963 .24901 .30781 .24671 Metetra
.603 .801 .975 r
.28448 .24386 .27632 .24602 .28799 .24749 .29616 .24526 Metetra
.597 .802 .978 r
.27276 .24248 .26461 .24457 .27632 .24602 .28448 .24386 Metetra
.594 .802 .979 r
.26101 .24106 .25287 .24311 .26461 .24457 .27276 .24248 Metetra
.593 .8 .979 r
.24923 .2396 .2411 .24162 .25287 .24311 .26101 .24106 Metetra
.593 .799 .978 r
.23742 .23808 .2293 .24011 .2411 .24162 .24923 .2396 Metetra
.594 .797 .977 r
.22557 .23652 .21746 .23856 .2293 .24011 .23742 .23808 Metetra
.596 .796 .976 r
.21369 .23493 .20559 .237 .21746 .23856 .22557 .23652 Metetra
.597 .796 .975 r
.20178 .23333 .19369 .23542 .20559 .237 .21369 .23493 Metetra
.598 .796 .975 r
.18983 .23173 .18175 .23383 .19369 .23542 .20178 .23333 Metetra
.598 .796 .975 r
.17785 .23013 .16978 .23224 .18175 .23383 .18983 .23173 Metetra
.598 .796 .975 r
.16583 .22853 .15778 .23065 .16978 .23224 .17785 .23013 Metetra
.598 .796 .975 r
.15378 .22694 .14574 .22906 .15778 .23065 .16583 .22853 Metetra
.598 .797 .975 r
.1417 .22534 .13367 .22747 .14574 .22906 .15378 .22694 Metetra
.598 .797 .976 r
.12958 .22375 .12156 .22588 .13367 .22747 .1417 .22534 Metetra
.598 .797 .976 r
.11743 .22215 .10942 .22428 .12156 .22588 .12958 .22375 Metetra
.598 .797 .976 r
.10524 .22054 .09725 .22268 .10942 .22428 .11743 .22215 Metetra
.598 .797 .976 r
.09302 .21893 .08504 .22107 .09725 .22268 .10524 .22054 Metetra
.598 .797 .976 r
.08076 .21731 .07279 .21946 .08504 .22107 .09302 .21893 Metetra
.614 .788 .965 r
.64006 .178 .63162 .18008 .64235 .18869 .65077 .17957 Metetra
.614 .788 .965 r
.62933 .17643 .6209 .17852 .63162 .18008 .64006 .178 Metetra
.614 .788 .965 r
.61856 .17486 .61014 .17695 .6209 .17852 .62933 .17643 Metetra
.614 .788 .965 r
.60777 .17328 .59936 .17538 .61014 .17695 .61856 .17486 Metetra
.614 .788 .965 r
.59695 .1717 .58854 .1738 .59936 .17538 .60777 .17328 Metetra
.614 .788 .965 r
.5861 .17011 .5777 .17222 .58854 .1738 .59695 .1717 Metetra
.614 .788 .965 r
.57522 .16852 .56683 .17064 .5777 .17222 .5861 .17011 Metetra
.614 .788 .965 r
.56431 .16693 .55593 .16905 .56683 .17064 .57522 .16852 Metetra
.766 .668 .755 r
.55338 .16739 .54503 .18849 .55593 .16905 .56431 .16693 Metetra
.196 .006 .381 r
.54245 .19352 .53409 .20711 .54503 .18849 .55338 .16739 Metetra
.365 .149 .447 r
.53147 .21422 .5231 .22259 .53409 .20711 .54245 .19352 Metetra
.454 .218 .465 r
.52044 .2304 .51207 .23518 .5231 .22259 .53147 .21422 Metetra
.507 .249 .459 r
.50936 .24283 .501 .24518 .51207 .23518 .52044 .2304 Metetra
.54 .257 .435 r
.49825 .25215 .48989 .2529 .501 .24518 .50936 .24283 Metetra
.558 .244 .391 r
.48709 .25889 .47874 .25864 .48989 .2529 .49825 .25215 Metetra
.563 .206 .321 r
.47591 .26351 .46757 .26268 .47874 .25864 .48709 .25889 Metetra
.547 .131 .204 r
.46469 .26634 .45636 .26527 .46757 .26268 .47591 .26351 Metetra
.487 0 .005 r
.45344 .26768 .44511 .2666 .45636 .26527 .46469 .26634 Metetra
.328 0 0 r
.44215 .2678 .43384 .26689 .44511 .2666 .45344 .26768 Metetra
0 .554 .733 r
.43084 .2669 .42254 .26632 .43384 .26689 .44215 .2678 Metetra
.273 .728 .955 r
.4195 .2652 .4112 .26505 .42254 .26632 .43084 .2669 Metetra
.444 .766 .997 r
.40812 .26292 .39984 .26327 .4112 .26505 .4195 .2652 Metetra
.53 .764 .983 r
.39672 .26026 .38845 .26114 .39984 .26327 .40812 .26292 Metetra
.577 .756 .963 r
.38529 .25742 .37702 .25881 .38845 .26114 .39672 .26026 Metetra
.604 .753 .949 r
.37383 .2546 .36557 .25644 .37702 .25881 .38529 .25742 Metetra
.62 .755 .942 r
.36233 .25195 .35408 .25413 .36557 .25644 .37383 .2546 Metetra
.629 .761 .942 r
.3508 .24957 .34256 .25199 .35408 .25413 .36233 .25195 Metetra
.632 .77 .947 r
.33924 .24751 .33101 .25004 .34256 .25199 .3508 .24957 Metetra
.629 .782 .955 r
.32765 .24577 .31943 .24829 .33101 .25004 .33924 .24751 Metetra
.622 .793 .964 r
.31602 .24428 .30781 .24671 .31943 .24829 .32765 .24577 Metetra
.612 .801 .972 r
.30436 .24295 .29616 .24526 .30781 .24671 .31602 .24428 Metetra
.601 .805 .978 r
.29267 .24169 .28448 .24386 .29616 .24526 .30436 .24295 Metetra
.593 .806 .981 r
.28094 .24041 .27276 .24248 .28448 .24386 .29267 .24169 Metetra
.589 .805 .982 r
.26918 .23906 .26101 .24106 .27276 .24248 .28094 .24041 Metetra
.588 .802 .981 r
.25739 .23761 .24923 .2396 .26101 .24106 .26918 .23906 Metetra
.59 .799 .979 r
.24556 .23609 .23742 .23808 .24923 .2396 .25739 .23761 Metetra
.593 .797 .977 r
.23371 .2345 .22557 .23652 .23742 .23808 .24556 .23609 Metetra
.596 .795 .976 r
.22181 .23287 .21369 .23493 .22557 .23652 .23371 .2345 Metetra
.598 .795 .975 r
.20989 .23124 .20178 .23333 .21369 .23493 .22181 .23287 Metetra
.599 .795 .974 r
.19793 .22961 .18983 .23173 .20178 .23333 .20989 .23124 Metetra
.599 .795 .974 r
.18594 .228 .17785 .23013 .18983 .23173 .19793 .22961 Metetra
.599 .796 .975 r
.17391 .2264 .16583 .22853 .17785 .23013 .18594 .228 Metetra
.599 .796 .975 r
.16185 .2248 .15378 .22694 .16583 .22853 .17391 .2264 Metetra
.598 .797 .975 r
.14976 .22321 .1417 .22534 .15378 .22694 .16185 .2248 Metetra
.598 .797 .976 r
.13763 .22161 .12958 .22375 .1417 .22534 .14976 .22321 Metetra
.598 .797 .976 r
.12546 .22 .11743 .22215 .12958 .22375 .13763 .22161 Metetra
.598 .797 .976 r
.11326 .21839 .10524 .22054 .11743 .22215 .12546 .22 Metetra
.598 .797 .976 r
.10103 .21677 .09302 .21893 .10524 .22054 .11326 .21839 Metetra
.598 .796 .976 r
.08876 .21515 .08076 .21731 .09302 .21893 .10103 .21677 Metetra
.614 .788 .965 r
.64853 .17591 .64006 .178 .65077 .17957 .65925 .17748 Metetra
.614 .788 .965 r
.63779 .17433 .62933 .17643 .64006 .178 .64853 .17591 Metetra
.614 .788 .965 r
.62702 .17276 .61856 .17486 .62933 .17643 .63779 .17433 Metetra
.614 .788 .965 r
.61622 .17117 .60777 .17328 .61856 .17486 .62702 .17276 Metetra
.614 .788 .965 r
.60539 .16958 .59695 .1717 .60777 .17328 .61622 .17117 Metetra
.614 .788 .965 r
.59453 .16799 .5861 .17011 .59695 .1717 .60539 .16958 Metetra
.614 .788 .965 r
.58364 .1664 .57522 .16852 .5861 .17011 .59453 .16799 Metetra
.614 .788 .965 r
.57273 .1648 .56431 .16693 .57522 .16852 .58364 .1664 Metetra
.69 .752 .899 r
.56178 .16319 .55338 .16739 .56431 .16693 .57273 .1648 Metetra
.41 .206 .489 r
.55084 .18541 .54245 .19352 .55338 .16739 .56178 .16319 Metetra
.54 .357 .581 r
.53987 .21042 .53147 .21422 .54245 .19352 .55084 .18541 Metetra
.571 .371 .571 r
.52883 .2294 .52044 .2304 .53147 .21422 .53987 .21042 Metetra
.594 .376 .554 r
.51776 .2436 .50936 .24283 .52044 .2304 .52883 .2294 Metetra
.612 .371 .529 r
.50663 .25398 .49825 .25215 .50936 .24283 .51776 .2436 Metetra
.625 .357 .492 r
.49547 .26129 .48709 .25889 .49825 .25215 .50663 .25398 Metetra
.632 .328 .437 r
.48428 .26611 .47591 .26351 .48709 .25889 .49547 .26129 Metetra
.629 .274 .35 r
.47305 .2689 .46469 .26634 .47591 .26351 .48428 .26611 Metetra
.602 .172 .197 r
.46179 .27 .45344 .26768 .46469 .26634 .47305 .2689 Metetra
.501 0 0 r
.4505 .26971 .44215 .2678 .45344 .26768 .46179 .27 Metetra
.205 0 0 r
.43917 .26827 .43084 .2669 .44215 .2678 .4505 .26971 Metetra
.203 .679 .935 r
.42782 .26593 .4195 .2652 .43084 .2669 .43917 .26827 Metetra
.435 .742 .992 r
.41644 .26294 .40812 .26292 .4195 .2652 .42782 .26593 Metetra
.537 .741 .969 r
.40503 .25954 .39672 .26026 .40812 .26292 .41644 .26294 Metetra
.588 .734 .944 r
.39359 .256 .38529 .25742 .39672 .26026 .40503 .25954 Metetra
.617 .732 .929 r
.38212 .25257 .37383 .2546 .38529 .25742 .39359 .256 Metetra
.634 .737 .923 r
.37061 .24946 .36233 .25195 .37383 .2546 .38212 .25257 Metetra
.643 .748 .925 r
.35908 .2468 .3508 .24957 .36233 .25195 .37061 .24946 Metetra
.646 .763 .934 r
.34751 .24466 .33924 .24751 .3508 .24957 .35908 .2468 Metetra
.641 .78 .948 r
.3359 .24298 .32765 .24577 .33924 .24751 .34751 .24466 Metetra
.63 .797 .962 r
.32427 .24167 .31602 .24428 .32765 .24577 .3359 .24298 Metetra
.613 .808 .975 r
.31259 .24057 .30436 .24295 .31602 .24428 .32427 .24167 Metetra
.595 .814 .983 r
.30089 .23952 .29267 .24169 .30436 .24295 .31259 .24057 Metetra
.583 .813 .987 r
.28915 .2384 .28094 .24041 .29267 .24169 .30089 .23952 Metetra
.579 .809 .987 r
.27738 .23713 .26918 .23906 .28094 .24041 .28915 .2384 Metetra
.581 .804 .984 r
.26558 .23571 .25739 .23761 .26918 .23906 .27738 .23713 Metetra
.586 .799 .981 r
.25374 .23415 .24556 .23609 .25739 .23761 .26558 .23571 Metetra
.592 .795 .977 r
.24187 .2325 .23371 .2345 .24556 .23609 .25374 .23415 Metetra
.597 .793 .974 r
.22997 .23081 .22181 .23287 .23371 .2345 .24187 .2325 Metetra
.599 .793 .973 r
.21803 .22913 .20989 .23124 .22181 .23287 .22997 .23081 Metetra
.601 .794 .973 r
.20607 .22747 .19793 .22961 .20989 .23124 .21803 .22913 Metetra
.601 .795 .974 r
.19406 .22584 .18594 .228 .19793 .22961 .20607 .22747 Metetra
.6 .796 .975 r
.18202 .22424 .17391 .2264 .18594 .228 .19406 .22584 Metetra
.599 .797 .975 r
.16995 .22265 .16185 .2248 .17391 .2264 .18202 .22424 Metetra
.598 .797 .976 r
.15784 .22106 .14976 .22321 .16185 .2248 .16995 .22265 Metetra
.597 .797 .976 r
.1457 .21946 .13763 .22161 .14976 .22321 .15784 .22106 Metetra
.597 .797 .976 r
.13353 .21786 .12546 .22 .13763 .22161 .1457 .21946 Metetra
.597 .797 .976 r
.12132 .21624 .11326 .21839 .12546 .22 .13353 .21786 Metetra
.597 .797 .976 r
.10907 .21461 .10103 .21677 .11326 .21839 .12132 .21624 Metetra
.598 .796 .976 r
.09679 .21298 .08876 .21515 .10103 .21677 .10907 .21461 Metetra
.614 .788 .965 r
.65703 .17381 .64853 .17591 .65925 .17748 .66776 .17539 Metetra
.614 .788 .965 r
.64628 .17223 .63779 .17433 .64853 .17591 .65703 .17381 Metetra
.614 .788 .965 r
.6355 .17065 .62702 .17276 .63779 .17433 .64628 .17223 Metetra
.614 .788 .965 r
.6247 .16906 .61622 .17117 .62702 .17276 .6355 .17065 Metetra
.614 .788 .965 r
.61386 .16746 .60539 .16958 .61622 .17117 .6247 .16906 Metetra
.614 .788 .965 r
.60299 .16587 .59453 .16799 .60539 .16958 .61386 .16746 Metetra
.614 .788 .965 r
.5921 .16426 .58364 .1664 .59453 .16799 .60299 .16587 Metetra
.614 .788 .965 r
.58117 .16266 .57273 .1648 .58364 .1664 .5921 .16426 Metetra
.614 .788 .965 r
.57022 .16105 .56178 .16319 .57273 .1648 .58117 .16266 Metetra
.584 .402 .601 r
.55928 .18494 .55084 .18541 .56178 .16319 .57022 .16105 Metetra
.608 .436 .623 r
.54831 .21266 .53987 .21042 .55084 .18541 .55928 .18494 Metetra
.624 .439 .611 r
.53727 .2332 .52883 .2294 .53987 .21042 .54831 .21266 Metetra
.639 .438 .595 r
.52619 .2482 .51776 .2436 .52883 .2294 .53727 .2332 Metetra
.652 .432 .573 r
.51506 .25889 .50663 .25398 .51776 .2436 .52619 .2482 Metetra
.663 .42 .542 r
.50389 .26618 .49547 .26129 .50663 .25398 .51506 .25889 Metetra
.672 .397 .497 r
.49268 .27077 .48428 .26611 .49547 .26129 .50389 .26618 Metetra
.676 .358 .428 r
.48144 .27316 .47305 .2689 .48428 .26611 .49268 .27077 Metetra
.668 .283 .307 r
.47017 .27372 .46179 .27 .47305 .2689 .48144 .27316 Metetra
.61 .118 .06 r
.45887 .27277 .4505 .26971 .46179 .27 .47017 .27372 Metetra
.353 0 0 r
.44754 .27055 .43917 .26827 .4505 .26971 .45887 .27277 Metetra
.151 .63 .923 r
.43618 .26732 .42782 .26593 .43917 .26827 .44754 .27055 Metetra
.437 .713 .982 r
.42479 .26335 .41644 .26294 .42782 .26593 .43618 .26732 Metetra
.549 .712 .947 r
.41337 .25895 .40503 .25954 .41644 .26294 .42479 .26335 Metetra
.601 .707 .918 r
.40192 .25445 .39359 .256 .40503 .25954 .41337 .25895 Metetra
.631 .708 .902 r
.39044 .25019 .38212 .25257 .39359 .256 .40192 .25445 Metetra
.649 .716 .898 r
.37893 .24647 .37061 .24946 .38212 .25257 .39044 .25019 Metetra
.659 .732 .903 r
.36738 .24349 .35908 .2468 .37061 .24946 .37893 .24647 Metetra
.662 .754 .918 r
.3558 .2413 .34751 .24466 .35908 .2468 .36738 .24349 Metetra
.656 .78 .94 r
.34419 .23981 .3359 .24298 .34751 .24466 .3558 .2413 Metetra
.638 .806 .963 r
.33254 .23884 .32427 .24167 .3359 .24298 .34419 .23981 Metetra
.609 .823 .982 r
.32086 .23813 .31259 .24057 .32427 .24167 .33254 .23884 Metetra
.58 .827 .992 r
.30914 .23742 .30089 .23952 .31259 .24057 .32086 .23813 Metetra
.562 .823 .995 r
.29739 .23653 .28915 .2384 .30089 .23952 .30914 .23742 Metetra
.56 .814 .993 r
.28561 .23537 .27738 .23713 .28915 .2384 .29739 .23653 Metetra
.569 .805 .988 r
.27379 .23393 .26558 .23571 .27738 .23713 .28561 .23537 Metetra
.581 .797 .981 r
.26195 .23228 .25374 .23415 .26558 .23571 .27379 .23393 Metetra
.592 .792 .976 r
.25007 .23052 .24187 .2325 .25374 .23415 .26195 .23228 Metetra
.599 .79 .972 r
.23816 .22873 .22997 .23081 .24187 .2325 .25007 .23052 Metetra
.603 .79 .971 r
.22621 .22697 .21803 .22913 .22997 .23081 .23816 .22873 Metetra
.604 .792 .971 r
.21423 .22528 .20607 .22747 .21803 .22913 .22621 .22697 Metetra
.603 .795 .973 r
.20222 .22365 .19406 .22584 .20607 .22747 .21423 .22528 Metetra
.601 .796 .974 r
.19017 .22206 .18202 .22424 .19406 .22584 .20222 .22365 Metetra
.599 .797 .976 r
.17808 .22048 .16995 .22265 .18202 .22424 .19017 .22206 Metetra
.597 .798 .976 r
.16596 .21891 .15784 .22106 .16995 .22265 .17808 .22048 Metetra
.597 .798 .976 r
.15381 .21731 .1457 .21946 .15784 .22106 .16596 .21891 Metetra
.597 .797 .976 r
.14162 .2157 .13353 .21786 .1457 .21946 .15381 .21731 Metetra
.597 .797 .976 r
.1294 .21408 .12132 .21624 .13353 .21786 .14162 .2157 Metetra
.597 .796 .976 r
.11714 .21244 .10907 .21461 .12132 .21624 .1294 .21408 Metetra
.598 .796 .975 r
.10485 .2108 .09679 .21298 .10907 .21461 .11714 .21244 Metetra
.614 .788 .965 r
.66557 .1717 .65703 .17381 .66776 .17539 .67629 .17329 Metetra
.614 .788 .965 r
.65481 .17012 .64628 .17223 .65703 .17381 .66557 .1717 Metetra
.614 .788 .965 r
.64402 .16853 .6355 .17065 .64628 .17223 .65481 .17012 Metetra
.614 .788 .965 r
.6332 .16693 .6247 .16906 .6355 .17065 .64402 .16853 Metetra
.614 .788 .965 r
.62236 .16533 .61386 .16746 .6247 .16906 .6332 .16693 Metetra
.614 .788 .965 r
.61148 .16373 .60299 .16587 .61386 .16746 .62236 .16533 Metetra
.614 .788 .965 r
.60058 .16212 .5921 .16426 .60299 .16587 .61148 .16373 Metetra
.614 .788 .965 r
.58965 .16051 .58117 .16266 .5921 .16426 .60058 .16212 Metetra
.614 .788 .965 r
.57868 .1589 .57022 .16105 .58117 .16266 .58965 .16051 Metetra
.638 .477 .648 r
.56777 .19387 .55928 .18494 .57022 .16105 .57868 .1589 Metetra
.647 .479 .642 r
.55679 .22212 .54831 .21266 .55928 .18494 .56777 .19387 Metetra
.658 .479 .631 r
.54575 .24258 .53727 .2332 .54831 .21266 .55679 .22212 Metetra
.669 .477 .616 r
.53466 .25714 .52619 .2482 .53727 .2332 .54575 .24258 Metetra
.68 .471 .596 r
.52352 .26721 .51506 .25889 .52619 .2482 .53466 .25714 Metetra
.69 .46 .569 r
.51234 .2738 .50389 .26618 .51506 .25889 .52352 .26721 Metetra
.7 .442 .531 r
.50112 .27764 .49268 .27077 .50389 .26618 .51234 .2738 Metetra
.708 .412 .473 r
.48987 .27925 .48144 .27316 .49268 .27077 .50112 .27764 Metetra
.711 .355 .372 r
.47859 .27898 .47017 .27372 .48144 .27316 .48987 .27925 Metetra
.682 .225 .161 r
.46728 .27711 .45887 .27277 .47017 .27372 .47859 .27898 Metetra
.47 0 0 r
.45593 .27386 .44754 .27055 .45887 .27277 .46728 .27711 Metetra
.121 .584 .919 r
.44456 .26946 .43618 .26732 .44754 .27055 .45593 .27386 Metetra
.449 .679 .962 r
.43316 .2642 .42479 .26335 .43618 .26732 .44456 .26946 Metetra
.563 .679 .917 r
.42174 .25844 .41337 .25895 .42479 .26335 .43316 .2642 Metetra
.614 .676 .886 r
.41028 .25263 .40192 .25445 .41337 .25895 .42174 .25844 Metetra
.644 .681 .87 r
.39879 .24724 .39044 .25019 .40192 .25445 .41028 .25263 Metetra
.663 .693 .868 r
.38727 .24272 .37893 .24647 .39044 .25019 .39879 .24724 Metetra
.675 .714 .877 r
.37572 .23934 .36738 .24349 .37893 .24647 .38727 .24272 Metetra
.681 .745 .899 r
.36413 .23719 .3558 .2413 .36738 .24349 .37572 .23934 Metetra
.675 .784 .932 r
.35251 .2361 .34419 .23981 .3558 .2413 .36413 .23719 Metetra
.647 .824 .968 r
.34085 .23575 .33254 .23884 .34419 .23981 .35251 .2361 Metetra
.592 .847 .993 r
.32915 .23568 .32086 .23813 .33254 .23884 .34085 .23575 Metetra
.534 .845 .999 r
.31742 .23551 .30914 .23742 .32086 .23813 .32915 .23568 Metetra
.51 .831 1 r
.30566 .23495 .29739 .23653 .30914 .23742 .31742 .23551 Metetra
.523 .817 .999 r
.29386 .23388 .28561 .23537 .29739 .23653 .30566 .23495 Metetra
.551 .803 .992 r
.28204 .23236 .27379 .23393 .28561 .23537 .29386 .23388 Metetra
.576 .791 .981 r
.27018 .23052 .26195 .23228 .27379 .23393 .28204 .23236 Metetra
.594 .785 .971 r
.2583 .22855 .25007 .23052 .26195 .23228 .27018 .23052 Metetra
.604 .783 .967 r
.24637 .22659 .23816 .22873 .25007 .23052 .2583 .22855 Metetra
.608 .786 .966 r
.23442 .22474 .22621 .22697 .23816 .22873 .24637 .22659 Metetra
.608 .79 .969 r
.22243 .22302 .21423 .22528 .22621 .22697 .23442 .22474 Metetra
.605 .795 .972 r
.2104 .2214 .20222 .22365 .21423 .22528 .22243 .22302 Metetra
.601 .798 .975 r
.19834 .21985 .19017 .22206 .20222 .22365 .2104 .2214 Metetra
.598 .799 .976 r
.18625 .21831 .17808 .22048 .19017 .22206 .19834 .21985 Metetra
.596 .799 .977 r
.17412 .21676 .16596 .21891 .17808 .22048 .18625 .21831 Metetra
.596 .798 .977 r
.16195 .21517 .15381 .21731 .16596 .21891 .17412 .21676 Metetra
.596 .797 .976 r
.14975 .21356 .14162 .2157 .15381 .21731 .16195 .21517 Metetra
.597 .797 .976 r
.13752 .21192 .1294 .21408 .14162 .2157 .14975 .21356 Metetra
.597 .796 .975 r
.12525 .21027 .11714 .21244 .1294 .21408 .13752 .21192 Metetra
.598 .796 .975 r
.11294 .20861 .10485 .2108 .11714 .21244 .12525 .21027 Metetra
.614 .788 .965 r
.67413 .16959 .66557 .1717 .67629 .17329 .68486 .17118 Metetra
.614 .788 .965 r
.66336 .168 .65481 .17012 .66557 .1717 .67413 .16959 Metetra
.614 .788 .965 r
.65257 .1664 .64402 .16853 .65481 .17012 .66336 .168 Metetra
.614 .788 .965 r
.64174 .1648 .6332 .16693 .64402 .16853 .65257 .1664 Metetra
.614 .788 .965 r
.63089 .1632 .62236 .16533 .6332 .16693 .64174 .1648 Metetra
.614 .788 .965 r
.62001 .16159 .61148 .16373 .62236 .16533 .63089 .1632 Metetra
.614 .788 .965 r
.60909 .15998 .60058 .16212 .61148 .16373 .62001 .16159 Metetra
.614 .788 .965 r
.59815 .15836 .58965 .16051 .60058 .16212 .60909 .15998 Metetra
.692 .516 .645 r
.58724 .17658 .57868 .1589 .58965 .16051 .59815 .15836 Metetra
.667 .508 .66 r
.57632 .21327 .56777 .19387 .57868 .1589 .58724 .17658 Metetra
.676 .509 .652 r
.56533 .23944 .55679 .22212 .56777 .19387 .57632 .21327 Metetra
.685 .507 .641 r
.55428 .25788 .54575 .24258 .55679 .22212 .56533 .23944 Metetra
.694 .504 .626 r
.54317 .2706 .53466 .25714 .54575 .24258 .55428 .25788 Metetra
.703 .498 .608 r
.53202 .27904 .52352 .26721 .53466 .25714 .54317 .2706 Metetra
.713 .489 .583 r
.52083 .28419 .51234 .2738 .52352 .26721 .53202 .27904 Metetra
.723 .474 .549 r
.5096 .28676 .50112 .27764 .51234 .2738 .52083 .28419 Metetra
.734 .45 .499 r
.49834 .28722 .48987 .27925 .50112 .27764 .5096 .28676 Metetra
.743 .407 .413 r
.48704 .28585 .47859 .27898 .48987 .27925 .49834 .28722 Metetra
.735 .305 .229 r
.47572 .28283 .46728 .27711 .47859 .27898 .48704 .28585 Metetra
.566 0 0 r
.46436 .27832 .45593 .27386 .46728 .27711 .47572 .28283 Metetra
.102 .534 .911 r
.45298 .27247 .44456 .26946 .45593 .27386 .46436 .27832 Metetra
.461 .64 .934 r
.44157 .26554 .43316 .2642 .44456 .26946 .45298 .27247 Metetra
.574 .644 .882 r
.43014 .25797 .42174 .25844 .43316 .2642 .44157 .26554 Metetra
.625 .646 .851 r
.41867 .25035 .41028 .25263 .42174 .25844 .43014 .25797 Metetra
.654 .653 .837 r
.40718 .24339 .39879 .24724 .41028 .25263 .41867 .25035 Metetra
.675 .669 .836 r
.39566 .23775 .38727 .24272 .39879 .24724 .40718 .24339 Metetra
.691 .695 .848 r
.38409 .2339 .37572 .23934 .38727 .24272 .39566 .23775 Metetra
.702 .735 .877 r
.3725 .23196 .36413 .23719 .37572 .23934 .38409 .2339 Metetra
.699 .795 .925 r
.36086 .23164 .35251 .2361 .36413 .23719 .3725 .23196 Metetra
.651 .861 .979 r
.34918 .23235 .34085 .23575 .35251 .2361 .36086 .23164 Metetra
.518 .875 .985 r
.33747 .23336 .32915 .23568 .34085 .23575 .34918 .23235 Metetra
.379 .823 .945 r
.32573 .23401 .31742 .23551 .32915 .23568 .33747 .23336 Metetra
.367 .804 .961 r
.31395 .23388 .30566 .23495 .31742 .23551 .32573 .23401 Metetra
.448 .807 .996 r
.30215 .23286 .29386 .23388 .30566 .23495 .31395 .23388 Metetra
.526 .795 .995 r
.29032 .23109 .28204 .23236 .29386 .23388 .30215 .23286 Metetra
.574 .78 .976 r
.27845 .22888 .27018 .23052 .28204 .23236 .29032 .23109 Metetra
.601 .772 .962 r
.26656 .22654 .2583 .22855 .27018 .23052 .27845 .22888 Metetra
.614 .773 .957 r
.25462 .22432 .24637 .22659 .2583 .22855 .26656 .22654 Metetra
.618 .78 .959 r
.24266 .22235 .23442 .22474 .24637 .22659 .25462 .22432 Metetra
.615 .789 .965 r
.23066 .22063 .22243 .22302 .23442 .22474 .24266 .22235 Metetra
.609 .796 .971 r
.21862 .21908 .2104 .2214 .22243 .22302 .23066 .22063 Metetra
.601 .801 .976 r
.20655 .21761 .19834 .21985 .2104 .2214 .21862 .21908 Metetra
.596 .802 .979 r
.19444 .21614 .18625 .21831 .19834 .21985 .20655 .21761 Metetra
.593 .801 .979 r
.1823 .21463 .17412 .21676 .18625 .21831 .19444 .21614 Metetra
.593 .799 .978 r
.17012 .21305 .16195 .21517 .17412 .21676 .1823 .21463 Metetra
.595 .797 .977 r
.15791 .21142 .14975 .21356 .16195 .21517 .17012 .21305 Metetra
.596 .796 .976 r
.14566 .20976 .13752 .21192 .14975 .21356 .15791 .21142 Metetra
.598 .796 .975 r
.13338 .20809 .12525 .21027 .13752 .21192 .14566 .20976 Metetra
.598 .796 .975 r
.12107 .20641 .11294 .20861 .12525 .21027 .13338 .20809 Metetra
.614 .788 .965 r
.68272 .16747 .67413 .16959 .68486 .17118 .69346 .16906 Metetra
.614 .788 .965 r
.67195 .16587 .66336 .168 .67413 .16959 .68272 .16747 Metetra
.614 .788 .965 r
.66114 .16427 .65257 .1664 .66336 .168 .67195 .16587 Metetra
.614 .788 .965 r
.65031 .16266 .64174 .1648 .65257 .1664 .66114 .16427 Metetra
.614 .788 .965 r
.63945 .16105 .63089 .1632 .64174 .1648 .65031 .16266 Metetra
.614 .788 .965 r
.62856 .15944 .62001 .16159 .63089 .1632 .63945 .16105 Metetra
.614 .788 .965 r
.61764 .15782 .60909 .15998 .62001 .16159 .62856 .15944 Metetra
.687 .485 .608 r
.60673 .16831 .59815 .15836 .60909 .15998 .61764 .15782 Metetra
.686 .531 .671 r
.59589 .21246 .58724 .17658 .59815 .15836 .60673 .16831 Metetra
.693 .532 .664 r
.58495 .24328 .57632 .21327 .58724 .17658 .59589 .21246 Metetra
.701 .532 .656 r
.57393 .26456 .56533 .23944 .57632 .21327 .58495 .24328 Metetra
.708 .53 .645 r
.56286 .27896 .55428 .25788 .56533 .23944 .57393 .26456 Metetra
.716 .526 .631 r
.55173 .28839 .54317 .2706 .55428 .25788 .56286 .27896 Metetra
.725 .52 .613 r
.54056 .29417 .53202 .27904 .54317 .2706 .55173 .28839 Metetra
.734 .511 .589 r
.52936 .29719 .52083 .28419 .53202 .27904 .54056 .29417 Metetra
.744 .498 .558 r
.51811 .29803 .5096 .28676 .52083 .28419 .52936 .29719 Metetra
.756 .479 .513 r
.50684 .29704 .49834 .28722 .5096 .28676 .51811 .29803 Metetra
.77 .446 .439 r
.49553 .29438 .48704 .28585 .49834 .28722 .50684 .29704 Metetra
.776 .371 .283 r
.48419 .29009 .47572 .28283 .48704 .28585 .49553 .29438 Metetra
.659 .105 0 r
.47283 .28414 .46436 .27832 .47572 .28283 .48419 .29009 Metetra
.066 .471 .888 r
.46143 .27657 .45298 .27247 .46436 .27832 .47283 .28414 Metetra
.465 .599 .901 r
.45001 .26755 .44157 .26554 .45298 .27247 .46143 .27657 Metetra
.581 .611 .847 r
.43857 .25755 .43014 .25797 .44157 .26554 .45001 .26755 Metetra
.632 .617 .817 r
.4271 .24739 .41867 .25035 .43014 .25797 .43857 .25755 Metetra
.662 .627 .804 r
.4156 .23813 .40718 .24339 .41867 .25035 .4271 .24739 Metetra
.685 .645 .804 r
.40407 .23089 .39566 .23775 .40718 .24339 .4156 .23813 Metetra
.705 .675 .818 r
.3925 .22645 .38409 .2339 .39566 .23775 .40407 .23089 Metetra
.724 .727 .852 r
.3809 .22502 .3725 .23196 .38409 .2339 .3925 .22645 Metetra
.731 .818 .919 r
.36924 .2261 .36086 .23164 .3725 .23196 .3809 .22502 Metetra
.617 .922 .975 r
.35755 .22866 .34918 .23235 .36086 .23164 .36924 .2261 Metetra
.34582 .23143 .33747 .23336 .34918 .23235 .35755 .22866 Metetra
.126 0 0 r
.33406 .23334 .32573 .23401 .33747 .23336 .34582 .23143 Metetra
.035 0 0 r
.32228 .23375 .31395 .23388 .32573 .23401 .33406 .23334 Metetra
.287 .748 .95 r
.31046 .2326 .30215 .23286 .31395 .23388 .32228 .23375 Metetra
.498 .776 .994 r
.29862 .23026 .29032 .23109 .30215 .23286 .31046 .2326 Metetra
.579 .757 .963 r
.28675 .22733 .27845 .22888 .29032 .23109 .29862 .23026 Metetra
.614 .751 .943 r
.27485 .22437 .26656 .22654 .27845 .22888 .28675 .22733 Metetra
.629 .758 .94 r
.26291 .22178 .25462 .22432 .26656 .22654 .27485 .22437 Metetra
.632 .773 .948 r
.25093 .21969 .24266 .22235 .25462 .22432 .26291 .22178 Metetra
.625 .789 .961 r
.23892 .21805 .23066 .22063 .24266 .22235 .25093 .21969 Metetra
.612 .801 .972 r
.22687 .21667 .21862 .21908 .23066 .22063 .23892 .21805 Metetra
.598 .807 .98 r
.21478 .21537 .20655 .21761 .21862 .21908 .22687 .21667 Metetra
.59 .807 .983 r
.20266 .21402 .19444 .21614 .20655 .21761 .21478 .21537 Metetra
.588 .803 .982 r
.19051 .21255 .1823 .21463 .19444 .21614 .20266 .21402 Metetra
.59 .799 .98 r
.17832 .21097 .17012 .21305 .1823 .21463 .19051 .21255 Metetra
.593 .796 .977 r
.1661 .2093 .15791 .21142 .17012 .21305 .17832 .21097 Metetra
.597 .795 .975 r
.15384 .2076 .14566 .20976 .15791 .21142 .1661 .2093 Metetra
.598 .795 .974 r
.14155 .20589 .13338 .20809 .14566 .20976 .15384 .2076 Metetra
.599 .795 .974 r
.12922 .2042 .12107 .20641 .13338 .20809 .14155 .20589 Metetra
.614 .788 .965 r
.69134 .16534 .68272 .16747 .69346 .16906 .7021 .16694 Metetra
.614 .788 .965 r
.68056 .16374 .67195 .16587 .68272 .16747 .69134 .16534 Metetra
.614 .788 .965 r
.66975 .16213 .66114 .16427 .67195 .16587 .68056 .16374 Metetra
.614 .788 .965 r
.65891 .16052 .65031 .16266 .66114 .16427 .66975 .16213 Metetra
.614 .788 .965 r
.64804 .1589 .63945 .16105 .65031 .16266 .65891 .16052 Metetra
.614 .788 .965 r
.63714 .15728 .62856 .15944 .63945 .16105 .64804 .1589 Metetra
.693 .528 .659 r
.62632 .18288 .61764 .15782 .62856 .15944 .63714 .15728 Metetra
.706 .552 .677 r
.61554 .23006 .60673 .16831 .61764 .15782 .62632 .18288 Metetra
.712 .554 .672 r
.60463 .26187 .59589 .21246 .60673 .16831 .61554 .23006 Metetra
.719 .554 .665 r
.59365 .28298 .58495 .24328 .59589 .21246 .60463 .26187 Metetra
.725 .553 .656 r
.5826 .29663 .57393 .26456 .58495 .24328 .59365 .28298 Metetra
.732 .55 .645 r
.57149 .30509 .56286 .27896 .57393 .26456 .5826 .29663 Metetra
.739 .545 .63 r
.56034 .30989 .55173 .28839 .56286 .27896 .57149 .30509 Metetra
.747 .538 .612 r
.54915 .3121 .54056 .29417 .55173 .28839 .56034 .30989 Metetra
.755 .528 .588 r
.53793 .31239 .52936 .29719 .54056 .29417 .54915 .3121 Metetra
.764 .516 .559 r
.52667 .31115 .51811 .29803 .52936 .29719 .53793 .31239 Metetra
.776 .501 .519 r
.51538 .30855 .50684 .29704 .51811 .29803 .52667 .31115 Metetra
.791 .477 .459 r
.50406 .30456 .49553 .29438 .50684 .29704 .51538 .30855 Metetra
.807 .428 .337 r
.4927 .29902 .48419 .29009 .49553 .29438 .50406 .30456 Metetra
.756 .244 0 r
.48132 .29164 .47283 .28414 .48419 .29009 .4927 .29902 Metetra
0 .366 .83 r
.46992 .28218 .46143 .27657 .47283 .28414 .48132 .29164 Metetra
.454 .557 .87 r
.45849 .2706 .45001 .26755 .46143 .27657 .46992 .28218 Metetra
.582 .58 .816 r
.44703 .25733 .43857 .25755 .45001 .26755 .45849 .2706 Metetra
.635 .591 .787 r
.43556 .24349 .4271 .24739 .43857 .25755 .44703 .25733 Metetra
.667 .604 .776 r
.42406 .23077 .4156 .23813 .4271 .24739 .43556 .24349 Metetra
.692 .624 .776 r
.41252 .22108 .40407 .23089 .4156 .23813 .42406 .23077 Metetra
.717 .657 .789 r
.40095 .21584 .3925 .22645 .40407 .23089 .41252 .22108 Metetra
.748 .719 .825 r
.38933 .21545 .3809 .22502 .3925 .22645 .40095 .21584 Metetra
.778 .867 .915 r
.37767 .21904 .36924 .2261 .3809 .22502 .38933 .21545 Metetra
.36595 .22478 .35755 .22866 .36924 .2261 .37767 .21904 Metetra
.428 .023 .166 r
.35421 .23046 .34582 .23143 .35755 .22866 .36595 .22478 Metetra
.542 .145 .238 r
.34243 .23426 .33406 .23334 .34582 .23143 .35421 .23046 Metetra
.511 .015 .015 r
.33062 .23527 .32228 .23375 .33406 .23334 .34243 .23426 Metetra
0 .555 .773 r
.3188 .23357 .31046 .2326 .32228 .23375 .33062 .23527 Metetra
.481 .737 .984 r
.30695 .23001 .29862 .23026 .31046 .2326 .3188 .23357 Metetra
.594 .718 .931 r
.29508 .22575 .28675 .22733 .29862 .23026 .30695 .23001 Metetra
.632 .719 .91 r
.28317 .22178 .27485 .22437 .28675 .22733 .29508 .22575 Metetra
.649 .737 .914 r
.27123 .21868 .26291 .22178 .27485 .22437 .28317 .22178 Metetra
.651 .765 .933 r
.25924 .21656 .25093 .21969 .26291 .22178 .27123 .21868 Metetra
.637 .794 .958 r
.24721 .21518 .23892 .21805 .25093 .21969 .25924 .21656 Metetra
.612 .814 .978 r
.23515 .21417 .22687 .21667 .23892 .21805 .24721 .21518 Metetra
.587 .819 .988 r
.22305 .21319 .21478 .21537 .22687 .21667 .23515 .21417 Metetra
.575 .814 .989 r
.21092 .21201 .20266 .21402 .21478 .21537 .22305 .21319 Metetra
.577 .806 .986 r
.19875 .21059 .19051 .21255 .20266 .21402 .21092 .21201 Metetra
.585 .799 .981 r
.18655 .20895 .17832 .21097 .19051 .21255 .19875 .21059 Metetra
.592 .794 .976 r
.17432 .2072 .1661 .2093 .17832 .21097 .18655 .20895 Metetra
.598 .792 .974 r
.16205 .20542 .15384 .2076 .1661 .2093 .17432 .2072 Metetra
.6 .793 .973 r
.14975 .20366 .14155 .20589 .15384 .2076 .16205 .20542 Metetra
.601 .794 .973 r
.13741 .20194 .12922 .2042 .14155 .20589 .14975 .20366 Metetra
.614 .788 .965 r
.7 .16321 .69134 .16534 .7021 .16694 .71076 .16481 Metetra
.614 .788 .965 r
.68921 .1616 .68056 .16374 .69134 .16534 .7 .16321 Metetra
.614 .788 .965 r
.67839 .15998 .66975 .16213 .68056 .16374 .68921 .1616 Metetra
.614 .788 .965 r
.66754 .15837 .65891 .16052 .66975 .16213 .67839 .15998 Metetra
.693 .525 .656 r
.65679 .18159 .64804 .1589 .65891 .16052 .66754 .15837 Metetra
.708 .558 .681 r
.64617 .24338 .63714 .15728 .64804 .1589 .65679 .18159 Metetra
.729 .574 .681 r
.63538 .28296 .62632 .18288 .63714 .15728 .64617 .24338 Metetra
.735 .576 .676 r
.62447 .3076 .61554 .23006 .62632 .18288 .63538 .28296 Metetra
.741 .576 .67 r
.61348 .32226 .60463 .26187 .61554 .23006 .62447 .3076 Metetra
.747 .576 .663 r
.60243 .33031 .59365 .28298 .60463 .26187 .61348 .32226 Metetra
.754 .574 .653 r
.59132 .33403 .5826 .29663 .59365 .28298 .60243 .33031 Metetra
.76 .57 .64 r
.58018 .33495 .57149 .30509 .5826 .29663 .59132 .33403 Metetra
.766 .563 .623 r
.569 .33406 .56034 .30989 .57149 .30509 .58018 .33495 Metetra
.773 .554 .603 r
.55779 .33201 .54915 .3121 .56034 .30989 .569 .33406 Metetra
.779 .543 .579 r
.54654 .32916 .53793 .31239 .54915 .3121 .55779 .33201 Metetra
.785 .531 .552 r
.53526 .32566 .52667 .31115 .53793 .31239 .54654 .32916 Metetra
.793 .517 .52 r
.52396 .32146 .51538 .30855 .52667 .31115 .53526 .32566 Metetra
.805 .502 .476 r
.51262 .31633 .50406 .30456 .51538 .30855 .52396 .32146 Metetra
.825 .477 .399 r
.50125 .30981 .4927 .29902 .50406 .30456 .51262 .31633 Metetra
.835 .392 .176 r
.48986 .30127 .48132 .29164 .4927 .29902 .50125 .30981 Metetra
0 .132 .643 r
.47844 .28998 .46992 .28218 .48132 .29164 .48986 .30127 Metetra
.409 .505 .845 r
.46699 .27543 .45849 .2706 .46992 .28218 .47844 .28998 Metetra
.575 .551 .79 r
.45553 .25777 .44703 .25733 .45849 .2706 .46699 .27543 Metetra
.635 .569 .763 r
.44405 .23845 .43556 .24349 .44703 .25733 .45553 .25777 Metetra
.669 .585 .752 r
.43255 .22024 .42406 .23077 .43556 .24349 .44405 .23845 Metetra
.696 .606 .751 r
.42101 .20657 .41252 .22108 .42406 .23077 .43255 .22024 Metetra
.726 .64 .762 r
.40944 .20012 .40095 .21584 .41252 .22108 .42101 .20657 Metetra
.773 .714 .797 r
.39781 .20168 .38933 .21545 .40095 .21584 .40944 .20012 Metetra
.842 .957 .886 r
.38612 .20977 .37767 .21904 .38933 .21545 .39781 .20168 Metetra
.371 .055 .294 r
.37439 .22104 .36595 .22478 .37767 .21904 .38612 .20977 Metetra
.588 .331 .496 r
.36261 .23154 .35421 .23046 .36595 .22478 .37439 .22104 Metetra
.648 .38 .5 r
.35081 .23818 .34243 .23426 .35421 .23046 .36261 .23154 Metetra
.678 .333 .381 r
.33899 .23963 .33062 .23527 .34243 .23426 .35081 .23818 Metetra
.381 0 0 r
.32716 .23644 .3188 .23357 .33062 .23527 .33899 .23963 Metetra
.495 .673 .944 r
.31532 .23043 .30695 .23001 .3188 .23357 .32716 .23644 Metetra
.613 .666 .877 r
.30344 .22378 .29508 .22575 .30695 .23001 .31532 .23043 Metetra
.652 .68 .864 r
.29154 .21821 .28317 .22178 .29508 .22575 .30344 .22378 Metetra
.671 .712 .879 r
.27958 .21452 .27123 .21868 .28317 .22178 .29154 .21821 Metetra
.675 .76 .915 r
.26758 .21263 .25924 .21656 .27123 .21868 .27958 .21452 Metetra
.651 .812 .961 r
.25554 .21194 .24721 .21518 .25924 .21656 .26758 .21263 Metetra
.597 .841 .991 r
.24346 .21166 .23515 .21417 .24721 .21518 .25554 .21194 Metetra
.548 .838 .998 r
.23135 .2112 .22305 .21319 .23515 .21417 .24346 .21166 Metetra
.54 .823 .998 r
.2192 .21026 .21092 .21201 .22305 .21319 .23135 .2112 Metetra
.558 .807 .992 r
.20702 .20883 .19875 .21059 .21092 .21201 .2192 .21026 Metetra
.579 .795 .981 r
.19481 .20705 .18655 .20895 .19875 .21059 .20702 .20883 Metetra
.594 .789 .973 r
.18257 .20512 .17432 .2072 .18655 .20895 .19481 .20705 Metetra
.602 .788 .97 r
.17029 .2032 .16205 .20542 .17432 .2072 .18257 .20512 Metetra
.605 .79 .97 r
.15798 .20137 .14975 .20366 .16205 .20542 .17029 .2032 Metetra
.604 .793 .972 r
.14563 .19964 .13741 .20194 .14975 .20366 .15798 .20137 Metetra
.614 .788 .965 r
.70868 .16106 .7 .16321 .71076 .16481 .71945 .16267 Metetra
.695 .544 .679 r
.69827 .22154 .68921 .1616 .7 .16321 .70868 .16106 Metetra
.716 .567 .686 r
.68795 .30948 .67839 .15998 .68921 .1616 .69827 .22154 Metetra
.741 .588 .685 r
.67733 .35957 .66754 .15837 .67839 .15998 .68795 .30948 Metetra
.752 .597 .684 r
.66652 .38565 .65679 .18159 .66754 .15837 .67733 .35957 Metetra
.758 .599 .681 r
.65559 .39678 .64617 .24338 .65679 .18159 .66652 .38565 Metetra
.764 .602 .677 r
.64458 .39885 .63538 .28296 .64617 .24338 .65559 .39678 Metetra
.77 .603 .671 r
.63352 .39565 .62447 .3076 .63538 .28296 .64458 .39885 Metetra
.777 .603 .664 r
.62242 .3896 .61348 .32226 .62447 .3076 .63352 .39565 Metetra
.784 .602 .654 r
.61128 .38221 .60243 .33031 .61348 .32226 .62242 .3896 Metetra
.79 .599 .642 r
.60012 .3744 .59132 .33403 .60243 .33031 .61128 .38221 Metetra
.797 .592 .625 r
.58892 .36672 .58018 .33495 .59132 .33403 .60012 .3744 Metetra
.802 .583 .606 r
.57771 .35947 .569 .33406 .58018 .33495 .58892 .36672 Metetra
.805 .57 .582 r
.56647 .35277 .55779 .33201 .569 .33406 .57771 .35947 Metetra
.807 .555 .558 r
.5552 .34662 .54654 .32916 .55779 .33201 .56647 .35277 Metetra
.807 .54 .535 r
.5439 .34089 .53526 .32566 .54654 .32916 .5552 .34662 Metetra
.808 .528 .515 r
.53258 .33533 .52396 .32146 .53526 .32566 .5439 .34089 Metetra
.812 .52 .495 r
.52122 .32948 .51262 .31633 .52396 .32146 .53258 .33533 Metetra
.826 .516 .464 r
.50984 .32264 .50125 .30981 .51262 .31633 .52122 .32948 Metetra
.856 .501 .38 r
.49843 .31369 .48986 .30127 .50125 .30981 .50984 .32264 Metetra
.813 .335 0 r
.48699 .30119 .47844 .28998 .48986 .30127 .49843 .31369 Metetra
.255 .399 .816 r
.47553 .28358 .46699 .27543 .47844 .28998 .48699 .30119 Metetra
.553 .518 .771 r
.46406 .26011 .45553 .25777 .46699 .27543 .47553 .28358 Metetra
.63 .548 .743 r
.45257 .23236 .44405 .23845 .45553 .25777 .46406 .26011 Metetra
.668 .568 .733 r
.44107 .20499 .43255 .22024 .44405 .23845 .45257 .23236 Metetra
.698 .59 .732 r
.42954 .18444 .42101 .20657 .43255 .22024 .44107 .20499 Metetra
.732 .626 .74 r
.41797 .17591 .40944 .20012 .42101 .20657 .42954 .18444 Metetra
.796 .709 .769 r
.40633 .18103 .39781 .20168 .40944 .20012 .41797 .17591 Metetra
.723 .913 .554 r
.39462 .1972 .38612 .20977 .39781 .20168 .40633 .18103 Metetra
.529 .325 .547 r
.38284 .21825 .37439 .22104 .38612 .20977 .39462 .1972 Metetra
.633 .44 .603 r
.37103 .23686 .36261 .23154 .37439 .22104 .38284 .21825 Metetra
.682 .473 .597 r
.3592 .24765 .35081 .23818 .36261 .23154 .37103 .23686 Metetra
.733 .468 .528 r
.34737 .24889 .33899 .23963 .35081 .23818 .3592 .24765 Metetra
.603 .034 0 r
.33555 .2422 .32716 .23644 .33899 .23963 .34737 .24889 Metetra
.531 .595 .863 r
.32371 .23137 .31532 .23043 .32716 .23644 .33555 .2422 Metetra
.63 .614 .816 r
.31185 .22055 .30344 .22378 .31532 .23043 .32371 .23137 Metetra
.669 .643 .815 r
.29994 .21259 .29154 .21821 .30344 .22378 .31185 .22055 Metetra
.695 .69 .841 r
.28798 .20843 .27958 .21452 .29154 .21821 .29994 .21259 Metetra
.706 .766 .899 r
.27597 .20747 .26758 .21263 .27958 .21452 .28798 .20843 Metetra
.66 .862 .976 r
.2639 .2083 .25554 .21194 .26758 .21263 .27597 .20747 Metetra
.497 .871 .979 r
.2518 .20938 .24346 .21166 .25554 .21194 .2639 .2083 Metetra
.398 .826 .962 r
.23967 .20974 .23135 .2112 .24346 .21166 .2518 .20938 Metetra
.45 .817 .994 r
.22751 .20903 .2192 .21026 .23135 .2112 .23967 .20974 Metetra
.528 .802 .996 r
.21533 .20741 .20702 .20883 .2192 .21026 .22751 .20903 Metetra
.575 .784 .978 r
.20311 .20526 .19481 .20705 .20702 .20883 .21533 .20741 Metetra
.6 .777 .965 r
.19086 .203 .18257 .20512 .19481 .20705 .20311 .20526 Metetra
.61 .78 .962 r
.17857 .20087 .17029 .2032 .18257 .20512 .19086 .203 Metetra
.612 .786 .965 r
.16624 .19897 .15798 .20137 .17029 .2032 .17857 .20087 Metetra
.608 .793 .97 r
.15388 .19726 .14563 .19964 .15798 .20137 .16624 .19897 Metetra
.767 .61 .684 r
.71923 .42783 .70868 .16106 .71945 .16267 .73009 .42909 Metetra
.767 .608 .682 r
.70833 .42656 .69827 .22154 .70868 .16106 .71923 .42783 Metetra
.768 .604 .676 r
.69741 .42529 .68795 .30948 .69827 .22154 .70833 .42656 Metetra
.768 .596 .665 r
.68646 .42402 .67733 .35957 .68795 .30948 .69741 .42529 Metetra
.769 .583 .646 r
.67548 .42275 .66652 .38565 .67733 .35957 .68646 .42402 Metetra
.769 .565 .623 r
.66447 .42147 .65559 .39678 .66652 .38565 .67548 .42275 Metetra
.769 .557 .612 r
.65343 .42018 .64458 .39885 .65559 .39678 .66447 .42147 Metetra
.769 .562 .619 r
.64236 .4189 .63352 .39565 .64458 .39885 .65343 .42018 Metetra
.769 .571 .631 r
.63126 .41761 .62242 .3896 .63352 .39565 .64236 .4189 Metetra
.769 .579 .642 r
.62012 .41632 .61128 .38221 .62242 .3896 .63126 .41761 Metetra
.769 .585 .65 r
.60896 .41483 .60012 .3744 .61128 .38221 .62012 .41632 Metetra
.853 .622 .588 r
.59772 .39821 .58892 .36672 .60012 .3744 .60896 .41483 Metetra
.855 .606 .561 r
.58646 .38443 .57771 .35947 .58892 .36672 .59772 .39821 Metetra
.852 .584 .533 r
.57519 .37305 .56647 .35277 .57771 .35947 .58646 .38443 Metetra
.843 .561 .511 r
.5639 .3637 .5552 .34662 .56647 .35277 .57519 .37305 Metetra
.831 .542 .5 r
.55258 .35598 .5439 .34089 .5552 .34662 .5639 .3637 Metetra
.818 .531 .503 r
.54124 .34949 .53258 .33533 .5439 .34089 .55258 .35598 Metetra
.811 .53 .514 r
.52987 .34363 .52122 .32948 .53258 .33533 .54124 .34949 Metetra
.812 .537 .524 r
.51847 .33753 .50984 .32264 .52122 .32948 .52987 .34363 Metetra
.828 .55 .517 r
.50704 .32969 .49843 .31369 .50984 .32264 .51847 .33753 Metetra
.88 .56 .441 r
.49559 .31767 .48699 .30119 .49843 .31369 .50704 .32969 Metetra
0 0 .408 r
.48411 .29798 .47553 .28358 .48699 .30119 .49559 .31767 Metetra
.493 .466 .756 r
.47262 .26732 .46406 .26011 .47553 .28358 .48411 .29798 Metetra
.619 .527 .727 r
.46113 .22618 .45257 .23236 .46406 .26011 .47262 .26732 Metetra
.665 .553 .718 r
.44963 .18257 .44107 .20499 .45257 .23236 .46113 .22618 Metetra
.697 .578 .717 r
.43812 .14957 .42954 .18444 .44107 .20499 .44963 .18257 Metetra
.736 .614 .723 r
.42655 .13726 .41797 .17591 .42954 .18444 .43812 .14957 Metetra
.817 .705 .742 r
.4149 .14867 .40633 .18103 .41797 .17591 .42655 .13726 Metetra
0 0 .245 r
.40315 .17956 .39462 .1972 .40633 .18103 .4149 .14867 Metetra
.582 .42 .627 r
.39133 .21838 .38284 .21825 .39462 .1972 .40315 .17956 Metetra
.654 .489 .647 r
.37946 .25076 .37103 .23686 .38284 .21825 .39133 .21838 Metetra
.7 .52 .641 r
.36759 .26735 .3592 .24765 .37103 .23686 .37946 .25076 Metetra
.764 .54 .593 r
.35576 .2665 .34737 .24889 .3592 .24765 .36759 .26735 Metetra
0 0 .492 r
.34395 .25219 .33555 .2422 .34737 .24889 .35576 .2665 Metetra
.566 .539 .784 r
.33214 .23203 .32371 .23137 .33555 .2422 .34395 .25219 Metetra
.645 .579 .765 r
.3203 .21406 .31185 .22055 .32371 .23137 .33214 .23203 Metetra
.685 .617 .774 r
.3084 .20285 .29994 .21259 .31185 .22055 .3203 .21406 Metetra
.72 .677 .807 r
.29643 .19896 .28798 .20843 .29994 .21259 .3084 .20285 Metetra
.753 .803 .893 r
.28439 .20052 .27597 .20747 .28798 .20843 .29643 .19896 Metetra
.524 .918 .899 r
.2723 .20448 .2639 .2083 .27597 .20747 .28439 .20052 Metetra
.125 0 0 r
.26017 .20795 .2518 .20938 .2639 .2083 .2723 .20448 Metetra
.151 0 0 r
.24802 .20943 .23967 .20974 .2518 .20938 .26017 .20795 Metetra
.216 .722 .901 r
.23585 .20874 .22751 .20903 .23967 .20974 .24802 .20943 Metetra
.492 .78 .996 r
.22366 .20649 .21533 .20741 .22751 .20903 .23585 .20874 Metetra
.582 .76 .963 r
.21143 .20356 .20311 .20526 .21533 .20741 .22366 .20649 Metetra
.614 .757 .947 r
.19917 .20069 .19086 .203 .20311 .20526 .21143 .20356 Metetra
.625 .767 .948 r
.18688 .19828 .17857 .20087 .19086 .203 .19917 .20069 Metetra
.623 .783 .958 r
.17454 .19635 .16624 .19897 .17857 .20087 .18688 .19828 Metetra
.614 .796 .969 r
.16216 .19475 .15388 .19726 .16624 .19897 .17454 .19635 Metetra
.568 .808 .989 r
.72805 .42614 .71923 .42783 .73009 .42909 .73892 .42741 Metetra
.568 .808 .989 r
.71715 .42487 .70833 .42656 .71923 .42783 .72805 .42614 Metetra
.568 .808 .989 r
.70622 .4236 .69741 .42529 .70833 .42656 .71715 .42487 Metetra
.568 .808 .989 r
.69526 .42232 .68646 .42402 .69741 .42529 .70622 .4236 Metetra
.568 .808 .989 r
.68428 .42104 .67548 .42275 .68646 .42402 .69526 .42232 Metetra
.568 .808 .989 r
.67326 .41976 .66447 .42147 .67548 .42275 .68428 .42104 Metetra
.568 .808 .989 r
.66221 .41847 .65343 .42018 .66447 .42147 .67326 .41976 Metetra
.568 .808 .989 r
.65113 .41718 .64236 .4189 .65343 .42018 .66221 .41847 Metetra
.568 .808 .989 r
.64002 .41589 .63126 .41761 .64236 .4189 .65113 .41718 Metetra
.568 .808 .989 r
.62888 .41459 .62012 .41632 .63126 .41761 .64002 .41589 Metetra
.549 .808 .994 r
.6177 .41329 .60896 .41483 .62012 .41632 .62888 .41459 Metetra
.766 .52 .562 r
.6065 .41198 .59772 .39821 .60896 .41483 .6177 .41329 Metetra
.79 .57 .602 r
.59525 .40708 .58646 .38443 .59772 .39821 .6065 .41198 Metetra
.918 .576 .375 r
.58395 .39139 .57519 .37305 .58646 .38443 .59525 .40708 Metetra
.89 .544 .384 r
.57263 .37919 .5639 .3637 .57519 .37305 .58395 .39139 Metetra
.858 .526 .424 r
.5613 .36989 .55258 .35598 .5639 .3637 .57263 .37919 Metetra
.825 .523 .478 r
.54994 .363 .54124 .34949 .55258 .35598 .5613 .36989 Metetra
.801 .53 .529 r
.53855 .35798 .52987 .34363 .54124 .34949 .54994 .363 Metetra
.788 .543 .567 r
.52714 .35407 .51847 .33753 .52987 .34363 .53855 .35798 Metetra
.788 .56 .591 r
.5157 .34974 .50704 .32969 .51847 .33753 .52714 .35407 Metetra
.808 .582 .595 r
.50423 .34182 .49559 .31767 .50704 .32969 .5157 .34974 Metetra
.887 .622 .533 r
.49272 .32401 .48411 .29798 .49559 .31767 .50423 .34182 Metetra
.168 .24 .695 r
.48121 .28655 .47262 .26732 .48411 .29798 .49272 .32401 Metetra
.597 .5 .715 r
.46971 .22351 .46113 .22618 .47262 .26732 .48121 .28655 Metetra
.659 .54 .707 r
.45823 .14878 .44963 .18257 .46113 .22618 .46971 .22351 Metetra
.696 .576 .717 r
.44671 .11471 .43812 .14957 .44963 .18257 .45823 .14878 Metetra
.758 .654 .747 r
.43512 .11296 .42655 .13726 .43812 .14957 .44671 .11471 Metetra
.761 .641 .728 r
.4235 .11122 .4149 .14867 .42655 .13726 .43512 .11296 Metetra
0 0 .197 r
.41173 .15348 .40315 .17956 .4149 .14867 .4235 .11122 Metetra
.609 .464 .66 r
.39982 .2263 .39133 .21838 .40315 .17956 .41173 .15348 Metetra
.668 .515 .669 r
.38787 .2821 .37946 .25076 .39133 .21838 .39982 .2263 Metetra
.714 .548 .663 r
.37596 .30554 .36759 .26735 .37946 .25076 .38787 .2821 Metetra
.793 .59 .626 r
.36413 .29836 .35576 .2665 .36759 .26735 .37596 .30554 Metetra
.051 .193 .692 r
.35236 .26816 .34395 .25219 .35576 .2665 .36413 .29836 Metetra
.596 .516 .734 r
.34062 .22975 .33214 .23203 .34395 .25219 .35236 .26816 Metetra
.659 .561 .733 r
.32882 .20002 .3203 .21406 .33214 .23203 .34062 .22975 Metetra
.701 .605 .746 r
.31693 .18537 .3084 .20285 .3203 .21406 .32882 .20002 Metetra
.751 .677 .78 r
.30494 .18379 .29643 .19896 .3084 .20285 .31693 .18537 Metetra
.836 .924 .891 r
.29286 .19115 .28439 .20052 .29643 .19896 .30494 .18379 Metetra
.374 .038 .261 r
.28072 .20142 .2723 .20448 .28439 .20052 .29286 .19115 Metetra
.565 .253 .398 r
.26856 .20878 .26017 .20795 .2723 .20448 .28072 .20142 Metetra
.583 .177 .238 r
.25639 .21138 .24802 .20943 .26017 .20795 .26856 .20878 Metetra
.216 0 0 r
.24421 .21006 .23585 .20874 .24802 .20943 .25639 .21138 Metetra
.479 .733 .982 r
.23201 .20624 .22366 .20649 .23585 .20874 .24421 .21006 Metetra
.602 .714 .923 r
.21979 .20168 .21143 .20356 .22366 .20649 .23201 .20624 Metetra
.635 .725 .913 r
.20753 .19786 .19917 .20069 .21143 .20356 .21979 .20168 Metetra
.645 .753 .927 r
.19522 .19517 .18688 .19828 .19917 .20069 .20753 .19786 Metetra
.639 .782 .95 r
.18287 .19336 .17454 .19635 .18688 .19828 .19522 .19517 Metetra
.619 .805 .971 r
.17048 .19209 .16216 .19475 .17454 .19635 .18287 .19336 Metetra
.568 .808 .989 r
.73691 .42445 .72805 .42614 .73892 .42741 .74778 .42572 Metetra
.568 .808 .989 r
.726 .42318 .71715 .42487 .72805 .42614 .73691 .42445 Metetra
.568 .808 .989 r
.71506 .4219 .70622 .4236 .71715 .42487 .726 .42318 Metetra
.568 .808 .989 r
.7041 .42062 .69526 .42232 .70622 .4236 .71506 .4219 Metetra
.568 .808 .989 r
.6931 .41933 .68428 .42104 .69526 .42232 .7041 .42062 Metetra
.568 .808 .989 r
.68207 .41804 .67326 .41976 .68428 .42104 .6931 .41933 Metetra
.568 .808 .989 r
.67102 .41675 .66221 .41847 .67326 .41976 .68207 .41804 Metetra
.568 .808 .989 r
.65993 .41545 .65113 .41718 .66221 .41847 .67102 .41675 Metetra
.568 .808 .989 r
.64881 .41416 .64002 .41589 .65113 .41718 .65993 .41545 Metetra
.568 .808 .989 r
.63766 .41285 .62888 .41459 .64002 .41589 .64881 .41416 Metetra
.568 .808 .989 r
.62648 .41155 .6177 .41329 .62888 .41459 .63766 .41285 Metetra
.568 .808 .989 r
.61527 .41024 .6065 .41198 .6177 .41329 .62648 .41155 Metetra
.348 0 0 r
.60402 .40892 .59525 .40708 .6065 .41198 .61527 .41024 Metetra
.777 .53 .562 r
.59274 .40633 .58395 .39139 .59525 .40708 .60402 .40892 Metetra
.787 .319 0 r
.5814 .39184 .57263 .37919 .58395 .39139 .59274 .40633 Metetra
.865 .437 .196 r
.57005 .38142 .5613 .36989 .57263 .37919 .5814 .39184 Metetra
.829 .491 .416 r
.55868 .37463 .54994 .363 .5613 .36989 .57005 .38142 Metetra
.786 .517 .53 r
.54728 .37118 .53855 .35798 .54994 .363 .55868 .37463 Metetra
.761 .536 .591 r
.53586 .37083 .52714 .35407 .53855 .35798 .54728 .37118 Metetra
.753 .553 .626 r
.5244 .37295 .5157 .34974 .52714 .35407 .53586 .37083 Metetra
.757 .571 .645 r
.51292 .37526 .50423 .34182 .5157 .34974 .5244 .37295 Metetra
.78 .595 .65 r
.50139 .37001 .49272 .32401 .50423 .34182 .51292 .37526 Metetra
.903 .675 .584 r
.48984 .3354 .48121 .28655 .49272 .32401 .50139 .37001 Metetra
.543 .451 .701 r
.47831 .23684 .46971 .22351 .48121 .28655 .48984 .3354 Metetra
.64 .52 .701 r
.46684 .11413 .45823 .14878 .46971 .22351 .47831 .23684 Metetra
.614 .788 .965 r
.45527 .11238 .44671 .11471 .45823 .14878 .46684 .11413 Metetra
.614 .788 .965 r
.44367 .11063 .43512 .11296 .44671 .11471 .45527 .11238 Metetra
.614 .788 .965 r
.43204 .10888 .4235 .11122 .43512 .11296 .44367 .11063 Metetra
.77 .646 .724 r
.42037 .11004 .41173 .15348 .4235 .11122 .43204 .10888 Metetra
.627 .49 .676 r
.40829 .2562 .39982 .2263 .41173 .15348 .42037 .11004 Metetra
.68 .533 .679 r
.39623 .34944 .38787 .2821 .39982 .2263 .40829 .2562 Metetra
.731 .57 .673 r
.38428 .37503 .37596 .30554 .38787 .2821 .39623 .34944 Metetra
.817 .626 .644 r
.37246 .35502 .36413 .29836 .37596 .30554 .38428 .37503 Metetra
.312 .302 .687 r
.36078 .29327 .35236 .26816 .36413 .29836 .37246 .35502 Metetra
.621 .512 .709 r
.34917 .21617 .34062 .22975 .35236 .26816 .36078 .29327 Metetra
.675 .558 .715 r
.33745 .16938 .32882 .20002 .34062 .22975 .34917 .21617 Metetra
.724 .609 .729 r
.32555 .15469 .31693 .18537 .32882 .20002 .33745 .16938 Metetra
.789 .69 .756 r
.31353 .15921 .30494 .18379 .31693 .18537 .32555 .15469 Metetra
.30138 .17879 .29286 .19115 .30494 .18379 .31353 .15921 Metetra
.582 .39 .587 r
.28915 .20224 .28072 .20142 .29286 .19115 .30138 .17879 Metetra
.651 .434 .577 r
.27694 .21484 .26856 .20878 .28072 .20142 .28915 .20224 Metetra
.696 .396 .464 r
.26476 .21727 .25639 .21138 .26856 .20878 .27694 .21484 Metetra
.589 .033 0 r
.25258 .21398 .24421 .21006 .25639 .21138 .26476 .21727 Metetra
.506 .651 .923 r
.2404 .20669 .23201 .20624 .24421 .21006 .25258 .21398 Metetra
.628 .65 .852 r
.22819 .1988 .21979 .20168 .23201 .20624 .2404 .20669 Metetra
.661 .691 .868 r
.21593 .19378 .20753 .19786 .21979 .20168 .22819 .1988 Metetra
.671 .743 .905 r
.20361 .19115 .19522 .19517 .20753 .19786 .21593 .19378 Metetra
.661 .788 .942 r
.19124 .18982 .18287 .19336 .19522 .19517 .20361 .19115 Metetra
.618 .83 .982 r
.17883 .1893 .17048 .19209 .18287 .19336 .19124 .18982 Metetra
.568 .808 .989 r
.74579 .42275 .73691 .42445 .74778 .42572 .75668 .42403 Metetra
.568 .808 .989 r
.73488 .42147 .726 .42318 .73691 .42445 .74579 .42275 Metetra
.568 .808 .989 r
.72394 .42019 .71506 .4219 .726 .42318 .73488 .42147 Metetra
.568 .808 .989 r
.71296 .4189 .7041 .42062 .71506 .4219 .72394 .42019 Metetra
.568 .808 .989 r
.70196 .41761 .6931 .41933 .7041 .42062 .71296 .4189 Metetra
.568 .808 .989 r
.69092 .41632 .68207 .41804 .6931 .41933 .70196 .41761 Metetra
.568 .808 .989 r
.67986 .41502 .67102 .41675 .68207 .41804 .69092 .41632 Metetra
.568 .808 .989 r
.66876 .41372 .65993 .41545 .67102 .41675 .67986 .41502 Metetra
.568 .808 .989 r
.65763 .41242 .64881 .41416 .65993 .41545 .66876 .41372 Metetra
.568 .808 .989 r
.64648 .41111 .63766 .41285 .64881 .41416 .65763 .41242 Metetra
.568 .808 .989 r
.63529 .4098 .62648 .41155 .63766 .41285 .64648 .41111 Metetra
.568 .808 .989 r
.62406 .40848 .61527 .41024 .62648 .41155 .63529 .4098 Metetra
.568 .808 .989 r
.61281 .40717 .60402 .40892 .61527 .41024 .62406 .40848 Metetra
.383 .78 .986 r
.60153 .40584 .59274 .40633 .60402 .40892 .61281 .40717 Metetra
.801 .406 .304 r
.5902 .40047 .5814 .39184 .59274 .40633 .60153 .40584 Metetra
0 .133 .641 r
.57883 .38939 .57005 .38142 .5814 .39184 .5902 .40047 Metetra
.802 .359 .192 r
.56744 .38296 .55868 .37463 .57005 .38142 .57883 .38939 Metetra
.764 .477 .499 r
.55604 .38129 .54728 .37118 .55868 .37463 .56744 .38296 Metetra
.733 .514 .595 r
.54461 .38498 .53586 .37083 .54728 .37118 .55604 .38129 Metetra
.722 .536 .639 r
.53315 .39535 .5244 .37295 .53586 .37083 .54461 .38498 Metetra
.756 .549 .617 r
.52165 .39648 .51292 .37526 .5244 .37295 .53315 .39535 Metetra
.769 .565 .622 r
.51011 .39513 .50139 .37001 .51292 .37526 .52165 .39648 Metetra
.769 .594 .661 r
.49853 .39377 .48984 .3354 .50139 .37001 .51011 .39513 Metetra
.943 .725 .579 r
.48694 .32275 .47831 .23684 .48984 .3354 .49853 .39377 Metetra
.589 .477 .693 r
.47544 .1118 .46684 .11413 .47831 .23684 .48694 .32275 Metetra
.614 .788 .965 r
.46387 .11005 .45527 .11238 .46684 .11413 .47544 .1118 Metetra
.614 .788 .965 r
.45226 .1083 .44367 .11063 .45527 .11238 .46387 .11005 Metetra
.614 .788 .965 r
.44062 .10654 .43204 .10888 .44367 .11063 .45226 .1083 Metetra
.702 .743 .883 r
.42895 .10477 .42037 .11004 .43204 .10888 .44062 .10654 Metetra
.655 .519 .685 r
.41662 .36864 .40829 .2562 .42037 .11004 .42895 .10477 Metetra
.726 .552 .654 r
.40474 .38278 .39623 .34944 .40829 .2562 .41662 .36864 Metetra
.733 .372 .367 r
.39287 .38139 .38428 .37503 .39623 .34944 .40474 .38278 Metetra
.769 .564 .62 r
.38096 .37999 .37246 .35502 .38428 .37503 .39287 .38139 Metetra
.906 .679 .583 r
.36915 .34312 .36078 .29327 .37246 .35502 .38096 .37999 Metetra
.642 .518 .697 r
.35794 .15821 .34917 .21617 .36078 .29327 .36915 .34312 Metetra
.703 .574 .707 r
.34624 .10608 .33745 .16938 .34917 .21617 .35794 .15821 Metetra
.77 .644 .722 r
.33427 .10736 .32555 .15469 .33745 .16938 .34624 .10608 Metetra
.801 .679 .729 r
.32222 .11787 .31353 .15921 .32555 .15469 .33427 .10736 Metetra
.263 .111 .479 r
.30995 .16061 .30138 .17879 .31353 .15921 .32222 .11787 Metetra
.635 .484 .66 r
.29752 .21859 .28915 .20224 .30138 .17879 .30995 .16061 Metetra
.702 .51 .625 r
.2853 .23108 .27694 .21484 .28915 .20224 .29752 .21859 Metetra
.777 .484 .492 r
.27314 .22813 .26476 .21727 .27694 .21484 .2853 .23108 Metetra
.799 .4 .298 r
.26096 .22244 .25258 .21398 .26476 .21727 .27314 .22813 Metetra
.533 .562 .83 r
.24881 .20793 .2404 .20669 .25258 .21398 .26096 .22244 Metetra
.651 .59 .773 r
.23665 .19194 .22819 .1988 .2404 .20669 .24881 .20793 Metetra
.694 .685 .837 r
.22437 .18734 .21593 .19378 .22819 .1988 .23665 .19194 Metetra
.701 .761 .899 r
.21203 .18601 .20361 .19115 .21593 .19378 .22437 .18734 Metetra
.691 .79 .926 r
.19965 .1853 .19124 .18982 .20361 .19115 .21203 .18601 Metetra
.57 .884 .991 r
.18721 .18647 .17883 .1893 .19124 .18982 .19965 .1853 Metetra
.568 .808 .989 r
.75472 .42105 .74579 .42275 .75668 .42403 .76561 .42233 Metetra
.568 .808 .989 r
.74379 .41976 .73488 .42147 .74579 .42275 .75472 .42105 Metetra
.568 .808 .989 r
.73284 .41848 .72394 .42019 .73488 .42147 .74379 .41976 Metetra
.568 .808 .989 r
.72186 .41719 .71296 .4189 .72394 .42019 .73284 .41848 Metetra
.568 .808 .989 r
.71085 .41589 .70196 .41761 .71296 .4189 .72186 .41719 Metetra
.568 .808 .989 r
.69981 .41459 .69092 .41632 .70196 .41761 .71085 .41589 Metetra
.568 .808 .989 r
.68873 .41329 .67986 .41502 .69092 .41632 .69981 .41459 Metetra
.568 .808 .989 r
.67763 .41199 .66876 .41372 .67986 .41502 .68873 .41329 Metetra
.568 .808 .989 r
.66649 .41068 .65763 .41242 .66876 .41372 .67763 .41199 Metetra
.568 .808 .989 r
.65532 .40936 .64648 .41111 .65763 .41242 .66649 .41068 Metetra
.568 .808 .989 r
.64413 .40805 .63529 .4098 .64648 .41111 .65532 .40936 Metetra
.568 .808 .989 r
.6329 .40673 .62406 .40848 .63529 .4098 .64413 .40805 Metetra
.568 .808 .989 r
.62164 .4054 .61281 .40717 .62406 .40848 .6329 .40673 Metetra
.568 .808 .989 r
.61034 .40408 .60153 .40584 .61281 .40717 .62164 .4054 Metetra
.448 0 0 r
.59902 .40275 .5902 .40047 .60153 .40584 .61034 .40408 Metetra
.4 .563 .899 r
.58764 .39277 .57883 .38939 .5902 .40047 .59902 .40275 Metetra
0 .468 .863 r
.57624 .38665 .56744 .38296 .57883 .38939 .58764 .39277 Metetra
.699 .314 .315 r
.56483 .38617 .55604 .38129 .56744 .38296 .57624 .38665 Metetra
.69 .454 .561 r
.55339 .39267 .54461 .38498 .55604 .38129 .56483 .38617 Metetra
.52 .099 .183 r
.54192 .39603 .53315 .39535 .54461 .38498 .55339 .39267 Metetra
.568 .808 .989 r
.5304 .39468 .52165 .39648 .53315 .39535 .54192 .39603 Metetra
.568 .808 .989 r
.51885 .39332 .51011 .39513 .52165 .39648 .5304 .39468 Metetra
.568 .808 .989 r
.50726 .39196 .49853 .39377 .51011 .39513 .51885 .39332 Metetra
.769 .597 .665 r
.49565 .39059 .48694 .32275 .49853 .39377 .50726 .39196 Metetra
.589 .474 .691 r
.48408 .10947 .47544 .1118 .48694 .32275 .49565 .39059 Metetra
.614 .788 .965 r
.4725 .10771 .46387 .11005 .47544 .1118 .48408 .10947 Metetra
.614 .788 .965 r
.46088 .10595 .45226 .1083 .46387 .11005 .4725 .10771 Metetra
.614 .788 .965 r
.44923 .10418 .44062 .10654 .45226 .1083 .46088 .10595 Metetra
.614 .788 .965 r
.43755 .10241 .42895 .10477 .44062 .10654 .44923 .10418 Metetra
.596 .468 .677 r
.42525 .38232 .41662 .36864 .42895 .10477 .43755 .10241 Metetra
.568 .808 .989 r
.4134 .38092 .40474 .38278 .41662 .36864 .42525 .38232 Metetra
.568 .808 .989 r
.40151 .37953 .39287 .38139 .40474 .38278 .4134 .38092 Metetra
.568 .808 .989 r
.38959 .37813 .38096 .37999 .39287 .38139 .40151 .37953 Metetra
.77 .577 .639 r
.37763 .37672 .36915 .34312 .38096 .37999 .38959 .37813 Metetra
.631 .507 .693 r
.36675 .09166 .35794 .15821 .36915 .34312 .37763 .37672 Metetra
.751 .673 .774 r
.35483 .08985 .34624 .10608 .35794 .15821 .36675 .09166 Metetra
.754 .665 .762 r
.34287 .08804 .33427 .10736 .34624 .10608 .35483 .08985 Metetra
.76 .647 .736 r
.33088 .08622 .32222 .11787 .33427 .10736 .34287 .08804 Metetra
.765 .628 .709 r
.31886 .08439 .30995 .16061 .32222 .11787 .33088 .08622 Metetra
.649 .512 .682 r
.30555 .30185 .29752 .21859 .30995 .16061 .31886 .08439 Metetra
.025 .121 .628 r
.29365 .25537 .2853 .23108 .29752 .21859 .30555 .30185 Metetra
0 0 .464 r
.28155 .23918 .27314 .22813 .2853 .23108 .29365 .25537 Metetra
.759 .532 .589 r
.26932 .23918 .26096 .22244 .27314 .22813 .28155 .23918 Metetra
0 .115 .667 r
.25719 .21941 .24881 .20793 .26096 .22244 .26932 .23918 Metetra
.655 .545 .718 r
.24525 .1708 .23665 .19194 .24881 .20793 .25719 .21941 Metetra
.787 .994 .876 r
.23287 .17927 .22437 .18734 .23665 .19194 .24525 .1708 Metetra
.72 .849 .943 r
.22049 .1808 .21203 .18601 .22437 .18734 .23287 .17927 Metetra
.713 .727 .861 r
.20811 .17874 .19965 .1853 .21203 .18601 .22049 .1808 Metetra
.19563 .18382 .18721 .18647 .19965 .1853 .20811 .17874 Metetra
.568 .808 .989 r
.76367 .41934 .75472 .42105 .76561 .42233 .77457 .42062 Metetra
.568 .808 .989 r
.75274 .41805 .74379 .41976 .75472 .42105 .76367 .41934 Metetra
.568 .808 .989 r
.74178 .41676 .73284 .41848 .74379 .41976 .75274 .41805 Metetra
.568 .808 .989 r
.73079 .41546 .72186 .41719 .73284 .41848 .74178 .41676 Metetra
.568 .808 .989 r
.71977 .41416 .71085 .41589 .72186 .41719 .73079 .41546 Metetra
.568 .808 .989 r
.70872 .41286 .69981 .41459 .71085 .41589 .71977 .41416 Metetra
.568 .808 .989 r
.69764 .41155 .68873 .41329 .69981 .41459 .70872 .41286 Metetra
.568 .808 .989 r
.68653 .41024 .67763 .41199 .68873 .41329 .69764 .41155 Metetra
.568 .808 .989 r
.67538 .40893 .66649 .41068 .67763 .41199 .68653 .41024 Metetra
.568 .808 .989 r
.66421 .40761 .65532 .40936 .66649 .41068 .67538 .40893 Metetra
.568 .808 .989 r
.653 .40629 .64413 .40805 .65532 .40936 .66421 .40761 Metetra
.568 .808 .989 r
.64176 .40496 .6329 .40673 .64413 .40805 .653 .40629 Metetra
.568 .808 .989 r
.63049 .40364 .62164 .4054 .6329 .40673 .64176 .40496 Metetra
.568 .808 .989 r
.61919 .4023 .61034 .40408 .62164 .4054 .63049 .40364 Metetra
.568 .808 .989 r
.60786 .40097 .59902 .40275 .61034 .40408 .61919 .4023 Metetra
.614 .617 .83 r
.59647 .39097 .58764 .39277 .59902 .40275 .60786 .40097 Metetra
.615 .671 .881 r
.58506 .38484 .57624 .38665 .58764 .39277 .59647 .39097 Metetra
.533 .836 .999 r
.57364 .38435 .56483 .38617 .57624 .38665 .58506 .38484 Metetra
.397 .002 .169 r
.56219 .39086 .55339 .39267 .56483 .38617 .57364 .38435 Metetra
.062 0 0 r
.55071 .39423 .54192 .39603 .55339 .39267 .56219 .39086 Metetra
.568 .808 .989 r
.53918 .39287 .5304 .39468 .54192 .39603 .55071 .39423 Metetra
.568 .808 .989 r
.52762 .39151 .51885 .39332 .5304 .39468 .53918 .39287 Metetra
.568 .808 .989 r
.51603 .39014 .50726 .39196 .51885 .39332 .52762 .39151 Metetra
.568 .808 .989 r
.5044 .38877 .49565 .39059 .50726 .39196 .51603 .39014 Metetra
.589 .474 .691 r
.49276 .10713 .48408 .10947 .49565 .39059 .5044 .38877 Metetra
.614 .788 .965 r
.48116 .10536 .4725 .10771 .48408 .10947 .49276 .10713 Metetra
.614 .788 .965 r
.46954 .10359 .46088 .10595 .4725 .10771 .48116 .10536 Metetra
.614 .788 .965 r
.45788 .10182 .44923 .10418 .46088 .10595 .46954 .10359 Metetra
.614 .788 .965 r
.44618 .10004 .43755 .10241 .44923 .10418 .45788 .10182 Metetra
.582 .456 .674 r
.43394 .38046 .42525 .38232 .43755 .10241 .44618 .10004 Metetra
.568 .808 .989 r
.42208 .37906 .4134 .38092 .42525 .38232 .43394 .38046 Metetra
.568 .808 .989 r
.41019 .37766 .40151 .37953 .4134 .38092 .42208 .37906 Metetra
.568 .808 .989 r
.39825 .37625 .38959 .37813 .40151 .37953 .41019 .37766 Metetra
.568 .808 .989 r
.38629 .37484 .37763 .37672 .38959 .37813 .39825 .37625 Metetra
.583 .47 .69 r
.37532 .08925 .36675 .09166 .37763 .37672 .38629 .37484 Metetra
.614 .788 .965 r
.36339 .08744 .35483 .08985 .36675 .09166 .37532 .08925 Metetra
.614 .788 .965 r
.35142 .08561 .34287 .08804 .35483 .08985 .36339 .08744 Metetra
.614 .788 .965 r
.33943 .08379 .33088 .08622 .34287 .08804 .35142 .08561 Metetra
.614 .788 .965 r
.32739 .08196 .31886 .08439 .33088 .08622 .33943 .08379 Metetra
.575 .448 .67 r
.31413 .29981 .30555 .30185 .31886 .08439 .32739 .08196 Metetra
.59 .505 .727 r
.3022 .25324 .29365 .25537 .30555 .30185 .31413 .29981 Metetra
.604 .574 .793 r
.29009 .23702 .28155 .23918 .29365 .25537 .3022 .25324 Metetra
.541 .851 .998 r
.27784 .23702 .26932 .23918 .28155 .23918 .29009 .23702 Metetra
.601 .557 .778 r
.2657 .21721 .25719 .21941 .26932 .23918 .27784 .23702 Metetra
.588 .503 .725 r
.25374 .1685 .24525 .1708 .25719 .21941 .2657 .21721 Metetra
.37 0 .193 r
.24134 .17698 .23287 .17927 .24525 .1708 .25374 .1685 Metetra
.448 .864 .951 r
.22896 .17851 .22049 .1808 .23287 .17927 .24134 .17698 Metetra
.603 .785 .968 r
.21657 .17645 .20811 .17874 .22049 .1808 .22896 .17851 Metetra
.089 0 0 r
.20407 .18154 .19563 .18382 .20811 .17874 .21657 .17645 Metetra
.568 .808 .989 r
.77266 .41762 .76367 .41934 .77457 .42062 .78356 .41891 Metetra
.568 .808 .989 r
.76172 .41633 .75274 .41805 .76367 .41934 .77266 .41762 Metetra
.568 .808 .989 r
.75075 .41503 .74178 .41676 .75274 .41805 .76172 .41633 Metetra
.568 .808 .989 r
.73976 .41373 .73079 .41546 .74178 .41676 .75075 .41503 Metetra
.568 .808 .989 r
.72873 .41243 .71977 .41416 .73079 .41546 .73976 .41373 Metetra
.568 .808 .989 r
.71767 .41112 .70872 .41286 .71977 .41416 .72873 .41243 Metetra
.568 .808 .989 r
.70658 .40981 .69764 .41155 .70872 .41286 .71767 .41112 Metetra
.568 .808 .989 r
.69546 .40849 .68653 .41024 .69764 .41155 .70658 .40981 Metetra
.568 .808 .989 r
.6843 .40717 .67538 .40893 .68653 .41024 .69546 .40849 Metetra
.568 .808 .989 r
.67312 .40585 .66421 .40761 .67538 .40893 .6843 .40717 Metetra
.568 .808 .989 r
.66191 .40452 .653 .40629 .66421 .40761 .67312 .40585 Metetra
.568 .808 .989 r
.65066 .40319 .64176 .40496 .653 .40629 .66191 .40452 Metetra
.568 .808 .989 r
.63938 .40186 .63049 .40364 .64176 .40496 .65066 .40319 Metetra
.568 .808 .989 r
.62807 .40052 .61919 .4023 .63049 .40364 .63938 .40186 Metetra
.686 .67 .829 r
.61671 .39511 .60786 .40097 .61919 .4023 .62807 .40052 Metetra
.67 .614 .784 r
.60531 .38396 .59647 .39097 .60786 .40097 .61671 .39511 Metetra
.692 .653 .807 r
.59389 .37749 .58506 .38484 .59647 .39097 .60531 .38396 Metetra
.732 .711 .83 r
.58246 .3758 .57364 .38435 .58506 .38484 .59389 .37749 Metetra
.797 .78 .838 r
.57101 .37951 .56219 .39086 .57364 .38435 .58246 .3758 Metetra
.313 0 .219 r
.55953 .38992 .55071 .39423 .56219 .39086 .57101 .37951 Metetra
.371 .827 .924 r
.548 .39105 .53918 .39287 .55071 .39423 .55953 .38992 Metetra
.568 .808 .989 r
.53643 .38969 .52762 .39151 .53918 .39287 .548 .39105 Metetra
.568 .808 .989 r
.52483 .38831 .51603 .39014 .52762 .39151 .53643 .38969 Metetra
.695 .565 .704 r
.51315 .31689 .5044 .38877 .51603 .39014 .52483 .38831 Metetra
.589 .477 .693 r
.50146 .10478 .49276 .10713 .5044 .38877 .51315 .31689 Metetra
.614 .788 .965 r
.48986 .10301 .48116 .10536 .49276 .10713 .50146 .10478 Metetra
.614 .788 .965 r
.47822 .10123 .46954 .10359 .48116 .10536 .48986 .10301 Metetra
.614 .788 .965 r
.46655 .09945 .45788 .10182 .46954 .10359 .47822 .10123 Metetra
.614 .788 .965 r
.45485 .09766 .44618 .10004 .45788 .10182 .46655 .09945 Metetra
.564 .441 .671 r
.4427 .36298 .43394 .38046 .44618 .10004 .45485 .09766 Metetra
.52 .272 .479 r
.4308 .37719 .42208 .37906 .43394 .38046 .4427 .36298 Metetra
.568 .808 .989 r
.41889 .37579 .41019 .37766 .42208 .37906 .4308 .37719 Metetra
.568 .808 .989 r
.40695 .37437 .39825 .37625 .41019 .37766 .41889 .37579 Metetra
.695 .575 .716 r
.39508 .33729 .38629 .37484 .39825 .37625 .40695 .37437 Metetra
.401 .331 .664 r
.38371 .15133 .37532 .08925 .38629 .37484 .39508 .33729 Metetra
.489 .433 .722 r
.37193 .09891 .36339 .08744 .37532 .08925 .38371 .15133 Metetra
.748 .517 .581 r
.35994 .10018 .35142 .08561 .36339 .08744 .37193 .09891 Metetra
.728 .545 .643 r
.34787 .11074 .33943 .08379 .35142 .08561 .35994 .10018 Metetra
.716 .559 .675 r
.3356 .1537 .32739 .08196 .33943 .08379 .34787 .11074 Metetra
.89 .742 .691 r
.32318 .21201 .31413 .29981 .32739 .08196 .3356 .1537 Metetra
.832 .731 .753 r
.31093 .22455 .3022 .25324 .31413 .29981 .32318 .21201 Metetra
.744 .668 .777 r
.29874 .22158 .29009 .23702 .3022 .25324 .31093 .22455 Metetra
.739 .642 .752 r
.28652 .21584 .27784 .23702 .29009 .23702 .29874 .22158 Metetra
.697 .606 .751 r
.27433 .20125 .2657 .21721 .27784 .23702 .28652 .21584 Metetra
.916 .552 .325 r
.26213 .18515 .25374 .1685 .2657 .21721 .27433 .20125 Metetra
.352 0 0 r
.24981 .18051 .24134 .17698 .25374 .1685 .26213 .18515 Metetra
.013 .581 .78 r
.23744 .17917 .22896 .17851 .24134 .17698 .24981 .18051 Metetra
.455 0 0 r
.22502 .17845 .21657 .17645 .22896 .17851 .23744 .17917 Metetra
.405 .841 .943 r
.21255 .17961 .20407 .18154 .21657 .17645 .22502 .17845 Metetra
.568 .808 .989 r
.78168 .4159 .77266 .41762 .78356 .41891 .79259 .41719 Metetra
.568 .808 .989 r
.77073 .4146 .76172 .41633 .77266 .41762 .78168 .4159 Metetra
.568 .808 .989 r
.75976 .4133 .75075 .41503 .76172 .41633 .77073 .4146 Metetra
.568 .808 .989 r
.74875 .41199 .73976 .41373 .75075 .41503 .75976 .4133 Metetra
.568 .808 .989 r
.73772 .41068 .72873 .41243 .73976 .41373 .74875 .41199 Metetra
.568 .808 .989 r
.72665 .40937 .71767 .41112 .72873 .41243 .73772 .41068 Metetra
.568 .808 .989 r
.71555 .40805 .70658 .40981 .71767 .41112 .72665 .40937 Metetra
.568 .808 .989 r
.70442 .40673 .69546 .40849 .70658 .40981 .71555 .40805 Metetra
.568 .808 .989 r
.69326 .40541 .6843 .40717 .69546 .40849 .70442 .40673 Metetra
.568 .808 .989 r
.68207 .40408 .67312 .40585 .6843 .40717 .69326 .40541 Metetra
.568 .808 .989 r
.67085 .40275 .66191 .40452 .67312 .40585 .68207 .40408 Metetra
.568 .808 .989 r
.65959 .40142 .65066 .40319 .66191 .40452 .67085 .40275 Metetra
.568 .808 .989 r
.6483 .40008 .63938 .40186 .65066 .40319 .65959 .40142 Metetra
.651 .747 .92 r
.63698 .39745 .62807 .40052 .63938 .40186 .6483 .40008 Metetra
.686 .6 .755 r
.62558 .38281 .61671 .39511 .62807 .40052 .63698 .39745 Metetra
.697 .62 .766 r
.61417 .37229 .60531 .38396 .61671 .39511 .62558 .38281 Metetra
.714 .643 .776 r
.60274 .36542 .59389 .37749 .60531 .38396 .61417 .37229 Metetra
.737 .665 .779 r
.5913 .36193 .58246 .3758 .59389 .37749 .60274 .36542 Metetra
.762 .678 .77 r
.57983 .36156 .57101 .37951 .58246 .3758 .5913 .36193 Metetra
.777 .673 .749 r
.56834 .36369 .55953 .38992 .57101 .37951 .57983 .36156 Metetra
.778 .677 .752 r
.55681 .36601 .548 .39105 .55953 .38992 .56834 .36369 Metetra
.748 .636 .737 r
.54523 .36069 .53643 .38969 .548 .39105 .55681 .36601 Metetra
.719 .588 .709 r
.53359 .32574 .52483 .38831 .53643 .38969 .54523 .36069 Metetra
.691 .559 .701 r
.52188 .22626 .51315 .31689 .52483 .38831 .53359 .32574 Metetra
.591 .484 .7 r
.5102 .10242 .50146 .10478 .51315 .31689 .52188 .22626 Metetra
.614 .788 .965 r
.49859 .10064 .48986 .10301 .50146 .10478 .5102 .10242 Metetra
.614 .788 .965 r
.48694 .09885 .47822 .10123 .48986 .10301 .49859 .10064 Metetra
.614 .788 .965 r
.47526 .09707 .46655 .09945 .47822 .10123 .48694 .09885 Metetra
.353 0 0 r
.46355 .09821 .45485 .09766 .46655 .09945 .47526 .09707 Metetra
.859 .742 .226 r
.45161 .24571 .4427 .36298 .45485 .09766 .46355 .09821 Metetra
.346 .234 .581 r
.43961 .3398 .4308 .37719 .4427 .36298 .45161 .24571 Metetra
.336 .134 .451 r
.42765 .36562 .41889 .37579 .4308 .37719 .43961 .3398 Metetra
.71 .596 .727 r
.41575 .34541 .40695 .37437 .41889 .37579 .42765 .36562 Metetra
.688 .562 .707 r
.40394 .28307 .39508 .33729 .40695 .37437 .41575 .34541 Metetra
0 0 .212 r
.39219 .20523 .38371 .15133 .39508 .33729 .40394 .28307 Metetra
.927 .7 .574 r
.38035 .15799 .37193 .09891 .38371 .15133 .39219 .20523 Metetra
.816 .615 .631 r
.36839 .14314 .35994 .10018 .37193 .09891 .38035 .15799 Metetra
.752 .57 .649 r
.35632 .14769 .34787 .11074 .35994 .10018 .36839 .14314 Metetra
.674 .491 .63 r
.34414 .16743 .3356 .1537 .34787 .11074 .35632 .14769 Metetra
.943 .958 .67 r
.33189 .19109 .32318 .21201 .3356 .1537 .34414 .16743 Metetra
.865 .799 .784 r
.31964 .20379 .31093 .22455 .32318 .21201 .33189 .19109 Metetra
.781 .719 .795 r
.30741 .20622 .29874 .22158 .31093 .22455 .31964 .20379 Metetra
.736 .673 .789 r
.29517 .20289 .28652 .21584 .29874 .22158 .30741 .20622 Metetra
.671 .653 .824 r
.28293 .19551 .27433 .20125 .28652 .21584 .29517 .20289 Metetra
.348 .598 .94 r
.27065 .18752 .26213 .18515 .27433 .20125 .28293 .19551 Metetra
.197 .613 .947 r
.25833 .18244 .24981 .18051 .26213 .18515 .27065 .18752 Metetra
.232 .691 .949 r
.24595 .17977 .23744 .17917 .24981 .18051 .25833 .18244 Metetra
.234 .72 .928 r
.23352 .17841 .22502 .17845 .23744 .17917 .24595 .17977 Metetra
.508 .833 .999 r
.22105 .17786 .21255 .17961 .22502 .17845 .23352 .17841 Metetra
.568 .808 .989 r
.79073 .41417 .78168 .4159 .79259 .41719 .80165 .41547 Metetra
.568 .808 .989 r
.77978 .41286 .77073 .4146 .78168 .4159 .79073 .41417 Metetra
.568 .808 .989 r
.7688 .41156 .75976 .4133 .77073 .4146 .77978 .41286 Metetra
.568 .808 .989 r
.75778 .41025 .74875 .41199 .75976 .4133 .7688 .41156 Metetra
.568 .808 .989 r
.74674 .40893 .73772 .41068 .74875 .41199 .75778 .41025 Metetra
.568 .808 .989 r
.73566 .40762 .72665 .40937 .73772 .41068 .74674 .40893 Metetra
.568 .808 .989 r
.72456 .40629 .71555 .40805 .72665 .40937 .73566 .40762 Metetra
.568 .808 .989 r
.71342 .40497 .70442 .40673 .71555 .40805 .72456 .40629 Metetra
.568 .808 .989 r
.70225 .40364 .69326 .40541 .70442 .40673 .71342 .40497 Metetra
.568 .808 .989 r
.69105 .40231 .68207 .40408 .69326 .40541 .70225 .40364 Metetra
.568 .808 .989 r
.67982 .40097 .67085 .40275 .68207 .40408 .69105 .40231 Metetra
.568 .808 .989 r
.66856 .39963 .65959 .40142 .67085 .40275 .67982 .40097 Metetra
.684 .678 .838 r
.65724 .39465 .6483 .40008 .65959 .40142 .66856 .39963 Metetra
.701 .601 .741 r
.64585 .37875 .63698 .39745 .6483 .40008 .65724 .39465 Metetra
.706 .612 .75 r
.63445 .36637 .62558 .38281 .63698 .39745 .64585 .37875 Metetra
.713 .625 .758 r
.62304 .35693 .61417 .37229 .62558 .38281 .63445 .36637 Metetra
.723 .638 .763 r
.61161 .34993 .60274 .36542 .61417 .37229 .62304 .35693 Metetra
.735 .648 .762 r
.60015 .34483 .5913 .36193 .60274 .36542 .61161 .34993 Metetra
.746 .649 .754 r
.58867 .34085 .57983 .36156 .5913 .36193 .60015 .34483 Metetra
.75 .641 .74 r
.57716 .33645 .56834 .36369 .57983 .36156 .58867 .34085 Metetra
.745 .625 .726 r
.56561 .32841 .55681 .36601 .56834 .36369 .57716 .33645 Metetra
.733 .606 .716 r
.55401 .31034 .54523 .36069 .55681 .36601 .56561 .32841 Metetra
.709 .582 .711 r
.54236 .27238 .53359 .32574 .54523 .36069 .55401 .31034 Metetra
.639 .529 .712 r
.53068 .2085 .52188 .22626 .53359 .32574 .54236 .27238 Metetra
.358 .324 .684 r
.519 .1328 .5102 .10242 .52188 .22626 .53068 .2085 Metetra
.602 .523 .738 r
.50735 .09826 .49859 .10064 .5102 .10242 .519 .1328 Metetra
.614 .788 .965 r
.49569 .09647 .48694 .09885 .49859 .10064 .50735 .09826 Metetra
.614 .788 .965 r
.484 .09468 .47526 .09707 .48694 .09885 .49569 .09647 Metetra
.69 .533 .669 r
.47225 .13746 .46355 .09821 .47526 .09707 .484 .09468 Metetra
.245 .135 .524 r
.4604 .2112 .45161 .24571 .46355 .09821 .47225 .13746 Metetra
.932 .789 .679 r
.44849 .2677 .43961 .3398 .45161 .24571 .4604 .2112 Metetra
.815 .676 .709 r
.43655 .29142 .42765 .36562 .43961 .3398 .44849 .2677 Metetra
.755 .623 .714 r
.42463 .28413 .41575 .34541 .42765 .36562 .43655 .29142 Metetra
.693 .578 .722 r
.41274 .25351 .40394 .28307 .41575 .34541 .42463 .28413 Metetra
.477 .442 .74 r
.40086 .21458 .39219 .20523 .40394 .28307 .41274 .25351 Metetra
.949 .607 .209 r
.38894 .18444 .38035 .15799 .39219 .20523 .40086 .21458 Metetra
.855 .608 .564 r
.37695 .16957 .36839 .14314 .38035 .15799 .38894 .18444 Metetra
.771 .55 .599 r
.36489 .16795 .35632 .14769 .36839 .14314 .37695 .16957 Metetra
.685 .45 .561 r
.35275 .17538 .34414 .16743 .35632 .14769 .36489 .16795 Metetra
.064 0 0 r
.34055 .18577 .33189 .19109 .34414 .16743 .35275 .17538 Metetra
.842 .893 .876 r
.32833 .1932 .31964 .20379 .33189 .19109 .34055 .18577 Metetra
.78 .772 .845 r
.31608 .19581 .30741 .20622 .31964 .20379 .32833 .1932 Metetra
.734 .723 .841 r
.30383 .19445 .29517 .20289 .30741 .20622 .31608 .19581 Metetra
.68 .708 .869 r
.29155 .19055 .28293 .19551 .29517 .20289 .30383 .19445 Metetra
.589 .715 .931 r
.27924 .18591 .27065 .18752 .28293 .19551 .29155 .19055 Metetra
.495 .735 .979 r
.26689 .18201 .25833 .18244 .27065 .18752 .27924 .18591 Metetra
.456 .763 .995 r
.2545 .17926 .24595 .17977 .25833 .18244 .26689 .18201 Metetra
.474 .792 .999 r
.24207 .17741 .23352 .17841 .24595 .17977 .2545 .17926 Metetra
.543 .814 .996 r
.22959 .17609 .22105 .17786 .23352 .17841 .24207 .17741 Metetra
.568 .808 .989 r
.79982 .41243 .79073 .41417 .80165 .41547 .81075 .41374 Metetra
.568 .808 .989 r
.78886 .41112 .77978 .41286 .79073 .41417 .79982 .41243 Metetra
.568 .808 .989 r
.77787 .40981 .7688 .41156 .77978 .41286 .78886 .41112 Metetra
.568 .808 .989 r
.76685 .4085 .75778 .41025 .7688 .41156 .77787 .40981 Metetra
.568 .808 .989 r
.7558 .40718 .74674 .40893 .75778 .41025 .76685 .4085 Metetra
.568 .808 .989 r
.74471 .40586 .73566 .40762 .74674 .40893 .7558 .40718 Metetra
.568 .808 .989 r
.7336 .40453 .72456 .40629 .73566 .40762 .74471 .40586 Metetra
.568 .808 .989 r
.72245 .4032 .71342 .40497 .72456 .40629 .7336 .40453 Metetra
.568 .808 .989 r
.71128 .40187 .70225 .40364 .71342 .40497 .72245 .4032 Metetra
.568 .808 .989 r
.70007 .40053 .69105 .40231 .70225 .40364 .71128 .40187 Metetra
.589 .799 .98 r
.68883 .399 .67982 .40097 .69105 .40231 .70007 .40053 Metetra
.696 .597 .741 r
.67747 .38207 .66856 .39963 .67982 .40097 .68883 .399 Metetra
.72 .609 .732 r
.6661 .36804 .65724 .39465 .66856 .39963 .67747 .38207 Metetra
.72 .616 .74 r
.65473 .35646 .64585 .37875 .65724 .39465 .6661 .36804 Metetra
.722 .624 .748 r
.64334 .34693 .63445 .36637 .64585 .37875 .65473 .35646 Metetra
.725 .632 .754 r
.63192 .33906 .62304 .35693 .63445 .36637 .64334 .34693 Metetra
.729 .639 .758 r
.62049 .33243 .61161 .34993 .62304 .35693 .63192 .33906 Metetra
.734 .642 .757 r
.60903 .32646 .60015 .34483 .61161 .34993 .62049 .33243 Metetra
.736 .638 .751 r
.59754 .32023 .58867 .34085 .60015 .34483 .60903 .32646 Metetra
.734 .629 .742 r
.58601 .31223 .57716 .33645 .58867 .34085 .59754 .32023 Metetra
.727 .615 .732 r
.57445 .29998 .56561 .32841 .57716 .33645 .58601 .31223 Metetra
.712 .597 .726 r
.56284 .27994 .55401 .31034 .56561 .32841 .57445 .29998 Metetra
.682 .571 .724 r
.55119 .24875 .54236 .27238 .55401 .31034 .56284 .27994 Metetra
.596 .512 .73 r
.53952 .20689 .53068 .2085 .54236 .27238 .55119 .24875 Metetra
0 0 .461 r
.52784 .16254 .519 .1328 .53068 .2085 .53952 .20689 Metetra
.958 .633 .236 r
.51617 .12896 .50735 .09826 .519 .1328 .52784 .16254 Metetra
.867 .583 .504 r
.50448 .11641 .49569 .09647 .50735 .09826 .51617 .12896 Metetra
.732 .555 .652 r
.49278 .12798 .484 .09468 .49569 .09647 .50448 .11641 Metetra
.677 .512 .655 r
.48102 .15935 .47225 .13746 .484 .09468 .49278 .12798 Metetra
.43 .268 .559 r
.4692 .1988 .4604 .2112 .47225 .13746 .48102 .15935 Metetra
.965 .871 .694 r
.45732 .23168 .44849 .2677 .4604 .2112 .4692 .1988 Metetra
.829 .706 .729 r
.4454 .24852 .43655 .29142 .44849 .2677 .45732 .23168 Metetra
.765 .647 .731 r
.43348 .24764 .42463 .28413 .43655 .29142 .4454 .24852 Metetra
.708 .607 .742 r
.42154 .23305 .41274 .25351 .42463 .28413 .43348 .24764 Metetra
.602 .554 .773 r
.4096 .21253 .40086 .21458 .41274 .25351 .42154 .23305 Metetra
.049 .284 .775 r
.39762 .19423 .38894 .18444 .40086 .21458 .4096 .21253 Metetra
.879 .494 .304 r
.38561 .1828 .37695 .16957 .38894 .18444 .39762 .19423 Metetra
.789 .479 .463 r
.37353 .17881 .36489 .16795 .37695 .16957 .38561 .1828 Metetra
.686 .349 .398 r
.36141 .18036 .35275 .17538 .36489 .16795 .37353 .17881 Metetra
.211 0 0 r
.34923 .18436 .34055 .18577 .35275 .17538 .36141 .18036 Metetra
.715 .927 .962 r
.33701 .18787 .32833 .1932 .34055 .18577 .34923 .18436 Metetra
.744 .818 .91 r
.32477 .18933 .31608 .19581 .32833 .1932 .33701 .18787 Metetra
.715 .768 .894 r
.3125 .1886 .30383 .19445 .31608 .19581 .32477 .18933 Metetra
.676 .75 .907 r
.3002 .18629 .29155 .19055 .30383 .19445 .3125 .1886 Metetra
.625 .751 .937 r
.28787 .18327 .27924 .18591 .29155 .19055 .3002 .18629 Metetra
.575 .763 .967 r
.2755 .18033 .26689 .18201 .27924 .18591 .28787 .18327 Metetra
.545 .778 .985 r
.2631 .17784 .2545 .17926 .26689 .18201 .2755 .18033 Metetra
.544 .793 .99 r
.25065 .17584 .24207 .17741 .2545 .17926 .2631 .17784 Metetra
.565 .803 .988 r
.23817 .17419 .22959 .17609 .24207 .17741 .25065 .17584 Metetra
.768 .619 .694 r
.80634 .13696 .79982 .41243 .81075 .41374 .81719 .13863 Metetra
.815 .66 .691 r
.79604 .19858 .78886 .41112 .79982 .41243 .80634 .13696 Metetra
.942 .78 .652 r
.78592 .28823 .77787 .40981 .78886 .41112 .79604 .19858 Metetra
.939 .796 .675 r
.77537 .33928 .76685 .4085 .77787 .40981 .78592 .28823 Metetra
.903 .785 .721 r
.76456 .36585 .7558 .40718 .76685 .4085 .77537 .33928 Metetra
.832 .73 .752 r
.75358 .37717 .74471 .40586 .7558 .40718 .76456 .36585 Metetra
.778 .675 .751 r
.7425 .37925 .7336 .40453 .74471 .40586 .75358 .37717 Metetra
.754 .645 .741 r
.73134 .37596 .72245 .4032 .7336 .40453 .7425 .37925 Metetra
.747 .631 .732 r
.72013 .36976 .71128 .40187 .72245 .4032 .73134 .37596 Metetra
.747 .626 .725 r
.70888 .36219 .70007 .40053 .71128 .40187 .72013 .36976 Metetra
.749 .623 .721 r
.6976 .3542 .68883 .399 .70007 .40053 .70888 .36219 Metetra
.745 .626 .727 r
.6863 .34634 .67747 .38207 .68883 .399 .6976 .3542 Metetra
.741 .629 .735 r
.67497 .33891 .6661 .36804 .67747 .38207 .6863 .34634 Metetra
.738 .633 .742 r
.66362 .33204 .65473 .35646 .6661 .36804 .67497 .33891 Metetra
.737 .637 .749 r
.65224 .32574 .64334 .34693 .65473 .35646 .66362 .33204 Metetra
.735 .641 .755 r
.64083 .31986 .63192 .33906 .64334 .34693 .65224 .32574 Metetra
.735 .643 .758 r
.6294 .31416 .62049 .33243 .63192 .33906 .64083 .31986 Metetra
.733 .642 .757 r
.61793 .30817 .60903 .32646 .62049 .33243 .6294 .31416 Metetra
.731 .636 .753 r
.60643 .30115 .59754 .32023 .60903 .32646 .61793 .30817 Metetra
.725 .626 .748 r
.5949 .29199 .58601 .31223 .59754 .32023 .60643 .30115 Metetra
.714 .612 .742 r
.58332 .27921 .57445 .29998 .58601 .31223 .5949 .29199 Metetra
.696 .595 .739 r
.57172 .2612 .56284 .27994 .57445 .29998 .58332 .27921 Metetra
.661 .57 .742 r
.56007 .23722 .55119 .24875 .56284 .27994 .57172 .2612 Metetra
.565 .513 .756 r
.54841 .20887 .53952 .20689 .55119 .24875 .56007 .23722 Metetra
0 0 .574 r
.53672 .18091 .52784 .16254 .53952 .20689 .54841 .20887 Metetra
.893 .637 .545 r
.52503 .15992 .51617 .12896 .52784 .16254 .53672 .18091 Metetra
.801 .596 .625 r
.51333 .15118 .50448 .11641 .51617 .12896 .52503 .15992 Metetra
.746 .557 .639 r
.5016 .15636 .49278 .12798 .50448 .11641 .51333 .15118 Metetra
.682 .494 .626 r
.48983 .17282 .48102 .15935 .49278 .12798 .5016 .15636 Metetra
.492 .283 .523 r
.47802 .19427 .4692 .1988 .48102 .15935 .48983 .17282 Metetra
.939 .964 .732 r
.46615 .21322 .45732 .23168 .4692 .1988 .47802 .19427 Metetra
.837 .749 .767 r
.45424 .2242 .4454 .24852 .45732 .23168 .46615 .21322 Metetra
.773 .679 .76 r
.44231 .22543 .43348 .24764 .4454 .24852 .45424 .2242 Metetra
.721 .643 .77 r
.43035 .21857 .42154 .23305 .43348 .24764 .44231 .22543 Metetra
.651 .617 .802 r
.41838 .20748 .4096 .21253 .42154 .23305 .43035 .21857 Metetra
.476 .579 .879 r
.40637 .1964 .39762 .19423 .4096 .21253 .41838 .20748 Metetra
0 .261 .704 r
.39433 .18823 .38561 .1828 .39762 .19423 .40637 .1964 Metetra
.663 .13 0 r
.38224 .18395 .37353 .17881 .38561 .1828 .39433 .18823 Metetra
.524 0 0 r
.37011 .18294 .36141 .18036 .37353 .17881 .38224 .18395 Metetra
.062 .626 .72 r
.35794 .18374 .34923 .18436 .36141 .18036 .37011 .18294 Metetra
.592 .883 .992 r
.34572 .18481 .33701 .18787 .34923 .18436 .35794 .18374 Metetra
.681 .835 .957 r
.33347 .18514 .32477 .18933 .33701 .18787 .34572 .18481 Metetra
.681 .797 .936 r
.32119 .18438 .3125 .1886 .32477 .18933 .33347 .18514 Metetra
.66 .778 .936 r
.30888 .18268 .3002 .18629 .3125 .1886 .32119 .18438 Metetra
.63 .774 .949 r
.29653 .18046 .28787 .18327 .3002 .18629 .30888 .18268 Metetra
.6 .777 .965 r
.28415 .17811 .2755 .18033 .28787 .18327 .29653 .18046 Metetra
.58 .785 .977 r
.27173 .1759 .2631 .17784 .2755 .18033 .28415 .17811 Metetra
.574 .792 .982 r
.25927 .17391 .25065 .17584 .2631 .17784 .27173 .1759 Metetra
.579 .798 .982 r
.24678 .17213 .23817 .17419 .25065 .17584 .25927 .17391 Metetra
.614 .788 .965 r
.81542 .13472 .80634 .13696 .81719 .13863 .82628 .1364 Metetra
.766 .631 .712 r
.80453 .13303 .79604 .19858 .80634 .13696 .81542 .13472 Metetra
.767 .622 .698 r
.79361 .13134 .78592 .28823 .79604 .19858 .80453 .13303 Metetra
.768 .62 .696 r
.78266 .12964 .77537 .33928 .78592 .28823 .79361 .13134 Metetra
.784 .634 .695 r
.7719 .15336 .76456 .36585 .77537 .33928 .78266 .12964 Metetra
.836 .681 .69 r
.76141 .21656 .75358 .37717 .76456 .36585 .7719 .15336 Metetra
.822 .673 .697 r
.75066 .25703 .7425 .37925 .75358 .37717 .76141 .21656 Metetra
.81 .666 .703 r
.73975 .28221 .73134 .37596 .7425 .37925 .75066 .25703 Metetra
.799 .66 .71 r
.72871 .29718 .72013 .36976 .73134 .37596 .73975 .28221 Metetra
.789 .656 .717 r
.71758 .30539 .70888 .36219 .72013 .36976 .72871 .29718 Metetra
.78 .654 .724 r
.70639 .30916 .6976 .3542 .70888 .36219 .71758 .30539 Metetra
.771 .653 .731 r
.69514 .31006 .6863 .34634 .6976 .3542 .70639 .30916 Metetra
.764 .653 .739 r
.68385 .30911 .67497 .33891 .6863 .34634 .69514 .31006 Metetra
.757 .653 .747 r
.67252 .30697 .66362 .33204 .67497 .33891 .68385 .30911 Metetra
.752 .654 .753 r
.66116 .30402 .65224 .32574 .66362 .33204 .67252 .30697 Metetra
.746 .654 .758 r
.64976 .3004 .64083 .31986 .65224 .32574 .66116 .30402 Metetra
.741 .652 .761 r
.63833 .29606 .6294 .31416 .64083 .31986 .64976 .3004 Metetra
.735 .647 .761 r
.62686 .29077 .61793 .30817 .6294 .31416 .63833 .29606 Metetra
.728 .639 .76 r
.61536 .28406 .60643 .30115 .61793 .30817 .62686 .29077 Metetra
.719 .629 .757 r
.60382 .27528 .5949 .29199 .60643 .30115 .61536 .28406 Metetra
.706 .616 .754 r
.59224 .26368 .58332 .27921 .5949 .29199 .60382 .27528 Metetra
.685 .6 .755 r
.58063 .24874 .57172 .2612 .58332 .27921 .59224 .26368 Metetra
.647 .579 .764 r
.56899 .23062 .56007 .23722 .57172 .2612 .58063 .24874 Metetra
.553 .534 .788 r
.55733 .21079 .54841 .20887 .56007 .23722 .56899 .23062 Metetra
0 .223 .74 r
.54564 .19211 .53672 .18091 .54841 .20887 .55733 .21079 Metetra
.898 .57 .419 r
.53394 .17807 .52503 .15992 .53672 .18091 .54564 .19211 Metetra
.808 .564 .57 r
.52222 .17141 .51333 .15118 .52503 .15992 .53394 .17807 Metetra
.75 .528 .594 r
.51047 .17297 .5016 .15636 .51333 .15118 .52222 .17141 Metetra
.685 .459 .573 r
.49869 .18121 .48983 .17282 .5016 .15636 .51047 .17297 Metetra
.511 .234 .431 r
.48687 .19271 .47802 .19427 .48983 .17282 .49869 .18121 Metetra
.784 .996 .787 r
.475 .20342 .46615 .21322 .47802 .19427 .48687 .19271 Metetra
.83 .804 .826 r
.4631 .21018 .45424 .2242 .46615 .21322 .475 .20342 Metetra
.773 .72 .803 r
.45116 .21163 .44231 .22543 .45424 .2242 .4631 .21018 Metetra
.727 .684 .808 r
.43919 .20832 .43035 .21857 .44231 .22543 .45116 .21163 Metetra
.674 .667 .836 r
.42719 .20212 .41838 .20748 .43035 .21857 .43919 .20832 Metetra
.584 .666 .895 r
.41516 .19526 .40637 .1964 .41838 .20748 .42719 .20212 Metetra
.393 .663 .967 r
.4031 .18951 .39433 .18823 .40637 .1964 .41516 .19526 Metetra
.113 .601 .909 r
.391 .18568 .38224 .18395 .39433 .18823 .4031 .18951 Metetra
.102 .633 .868 r
.37886 .18371 .37011 .18294 .38224 .18395 .391 .18568 Metetra
.353 .786 .968 r
.36668 .18295 .35794 .18374 .37011 .18294 .37886 .18371 Metetra
.553 .843 .998 r
.35446 .18263 .34572 .18481 .35794 .18374 .36668 .18295 Metetra
.628 .829 .978 r
.3422 .18211 .33347 .18514 .34572 .18481 .35446 .18263 Metetra
.645 .807 .961 r
.32991 .1811 .32119 .18438 .33347 .18514 .3422 .18211 Metetra
.639 .793 .956 r
.31759 .17959 .30888 .18268 .32119 .18438 .32991 .1811 Metetra
.624 .787 .96 r
.30523 .17772 .29653 .18046 .30888 .18268 .31759 .17959 Metetra
.607 .786 .967 r
.29283 .1757 .28415 .17811 .29653 .18046 .30523 .17772 Metetra
.594 .789 .974 r
.2804 .17369 .27173 .1759 .28415 .17811 .29283 .1757 Metetra
.588 .793 .977 r
.26793 .17177 .25927 .17391 .27173 .1759 .2804 .17369 Metetra
.588 .796 .979 r
.25543 .16995 .24678 .17213 .25927 .17391 .26793 .17177 Metetra
.614 .788 .965 r
.82453 .13247 .81542 .13472 .82628 .1364 .83539 .13416 Metetra
.614 .788 .965 r
.81363 .13077 .80453 .13303 .81542 .13472 .82453 .13247 Metetra
.614 .788 .965 r
.80271 .12908 .79361 .13134 .80453 .13303 .81363 .13077 Metetra
.614 .788 .965 r
.79175 .12737 .78266 .12964 .79361 .13134 .80271 .12908 Metetra
.76 .651 .741 r
.78076 .12567 .7719 .15336 .78266 .12964 .79175 .12737 Metetra
.767 .627 .705 r
.76974 .12395 .76141 .21656 .7719 .15336 .78076 .12567 Metetra
.806 .66 .701 r
.75892 .1502 .75066 .25703 .76141 .21656 .76974 .12395 Metetra
.885 .738 .693 r
.74822 .1986 .73975 .28221 .75066 .25703 .75892 .1502 Metetra
.861 .721 .707 r
.73733 .23123 .72871 .29718 .73975 .28221 .74822 .1986 Metetra
.839 .708 .719 r
.72632 .25287 .71758 .30539 .72871 .29718 .73733 .23123 Metetra
.821 .698 .729 r
.7152 .26685 .70639 .30916 .71758 .30539 .72632 .25287 Metetra
.805 .69 .738 r
.70401 .27549 .69514 .31006 .70639 .30916 .7152 .26685 Metetra
.791 .685 .747 r
.69276 .28039 .68385 .30911 .69514 .31006 .70401 .27549 Metetra
.779 .68 .754 r
.68146 .28262 .67252 .30697 .68385 .30911 .69276 .28039 Metetra
.768 .675 .761 r
.67011 .28287 .66116 .30402 .67252 .30697 .68146 .28262 Metetra
.758 .67 .765 r
.65872 .28155 .64976 .3004 .66116 .30402 .67011 .28287 Metetra
.748 .664 .768 r
.64729 .27883 .63833 .29606 .64976 .3004 .65872 .28155 Metetra
.739 .657 .769 r
.63582 .27469 .62686 .29077 .63833 .29606 .64729 .27883 Metetra
.728 .648 .769 r
.62432 .26895 .61536 .28406 .62686 .29077 .63582 .27469 Metetra
.716 .637 .768 r
.61278 .26132 .60382 .27528 .61536 .28406 .62432 .26895 Metetra
.701 .625 .769 r
.6012 .25156 .59224 .26368 .60382 .27528 .61278 .26132 Metetra
.679 .612 .774 r
.58959 .23961 .58063 .24874 .59224 .26368 .6012 .25156 Metetra
.642 .596 .787 r
.57795 .22593 .56899 .23062 .58063 .24874 .58959 .23961 Metetra
.559 .569 .82 r
.56628 .21165 .55733 .21079 .56899 .23062 .57795 .22593 Metetra
.211 .436 .861 r
.55459 .19853 .54564 .19211 .55733 .21079 .56628 .21165 Metetra
.832 .365 .084 r
.54288 .18853 .53394 .17807 .54564 .19211 .55459 .19853 Metetra
.809 .488 .446 r
.53114 .18309 .52222 .17141 .53394 .17807 .54288 .18853 Metetra
.749 .465 .5 r
.51938 .18264 .51047 .17297 .52222 .17141 .53114 .18309 Metetra
.677 .381 .465 r
.50759 .18629 .49869 .18121 .51047 .17297 .51938 .18264 Metetra
.474 .088 .228 r
.49576 .19213 .48687 .19271 .49869 .18121 .50759 .18629 Metetra
.6 .952 .859 r
.48388 .19792 .475 .20342 .48687 .19271 .49576 .19213 Metetra
.79 .851 .896 r
.47197 .20178 .4631 .21018 .475 .20342 .48388 .19792 Metetra
.759 .763 .856 r
.46003 .20277 .45116 .21163 .4631 .21018 .47197 .20178 Metetra
.723 .725 .852 r
.44805 .20098 .43919 .20832 .45116 .21163 .46003 .20277 Metetra
.68 .711 .872 r
.43604 .19727 .42719 .20212 .43919 .20832 .44805 .20098 Metetra
.623 .714 .911 r
.42399 .19284 .41516 .19526 .42719 .20212 .43604 .19727 Metetra
.541 .729 .961 r
.41191 .1887 .4031 .18951 .41516 .19526 .42399 .19284 Metetra
.452 .746 .992 r
.3998 .18546 .391 .18568 .4031 .18951 .41191 .1887 Metetra
.427 .771 .997 r
.38765 .18323 .37886 .18371 .391 .18568 .3998 .18546 Metetra
.483 .804 1 r
.37546 .18176 .36668 .18295 .37886 .18371 .38765 .18323 Metetra
.555 .82 .995 r
.36323 .18068 .35446 .18263 .36668 .18295 .37546 .18176 Metetra
.601 .817 .983 r
.35097 .17961 .3422 .18211 .35446 .18263 .36323 .18068 Metetra
.619 .807 .972 r
.33867 .17835 .32991 .1811 .3422 .18211 .35097 .17961 Metetra
.621 .798 .967 r
.32633 .17684 .31759 .17959 .32991 .1811 .33867 .17835 Metetra
.615 .793 .967 r
.31396 .1751 .30523 .17772 .31759 .17959 .32633 .17684 Metetra
.607 .792 .97 r
.30155 .17325 .29283 .1757 .30523 .17772 .31396 .1751 Metetra
.599 .792 .973 r
.28911 .17137 .2804 .17369 .29283 .1757 .30155 .17325 Metetra
.594 .794 .976 r
.27663 .1695 .26793 .17177 .2804 .17369 .28911 .17137 Metetra
.593 .796 .977 r
.26411 .16768 .25543 .16995 .26793 .17177 .27663 .1695 Metetra
.614 .788 .965 r
.83367 .13021 .82453 .13247 .83539 .13416 .84454 .13191 Metetra
.614 .788 .965 r
.82277 .12851 .81363 .13077 .82453 .13247 .83367 .13021 Metetra
.614 .788 .965 r
.81183 .12681 .80271 .12908 .81363 .13077 .82277 .12851 Metetra
.614 .788 .965 r
.80087 .1251 .79175 .12737 .80271 .12908 .81183 .12681 Metetra
.614 .788 .965 r
.78987 .12338 .78076 .12567 .79175 .12737 .80087 .1251 Metetra
.614 .788 .965 r
.77885 .12166 .76974 .12395 .78076 .12567 .78987 .12338 Metetra
.761 .649 .737 r
.76779 .11994 .75892 .1502 .76974 .12395 .77885 .12166 Metetra
.792 .655 .712 r
.7568 .1307 .74822 .1986 .75892 .1502 .76779 .11994 Metetra
.967 .838 .663 r
.74604 .17615 .73733 .23123 .74822 .1986 .7568 .1307 Metetra
.924 .803 .708 r
.73512 .20787 .72632 .25287 .73733 .23123 .74604 .17615 Metetra
.885 .772 .733 r
.72407 .22975 .7152 .26685 .72632 .25287 .73512 .20787 Metetra
.853 .749 .748 r
.71292 .24455 .70401 .27549 .7152 .26685 .72407 .22975 Metetra
.826 .73 .759 r
.70171 .25423 .69276 .28039 .70401 .27549 .71292 .24455 Metetra
.805 .716 .767 r
.69042 .26013 .68146 .28262 .69276 .28039 .70171 .25423 Metetra
.786 .703 .773 r
.67909 .26319 .67011 .28287 .68146 .28262 .69042 .26013 Metetra
.771 .692 .777 r
.66771 .26401 .65872 .28155 .67011 .28287 .67909 .26319 Metetra
.756 .682 .779 r
.65628 .26294 .64729 .27883 .65872 .28155 .66771 .26401 Metetra
.743 .671 .78 r
.64482 .26015 .63582 .27469 .64729 .27883 .65628 .26294 Metetra
.73 .661 .781 r
.63331 .25567 .62432 .26895 .63582 .27469 .64482 .26015 Metetra
.717 .65 .782 r
.62177 .24949 .61278 .26132 .62432 .26895 .63331 .25567 Metetra
.7 .64 .785 r
.61019 .24163 .6012 .25156 .61278 .26132 .62177 .24949 Metetra
.678 .629 .794 r
.59858 .23227 .58959 .23961 .6012 .25156 .61019 .24163 Metetra
.644 .62 .811 r
.58694 .22191 .57795 .22593 .58959 .23961 .59858 .23227 Metetra
.577 .608 .847 r
.57527 .21136 .56628 .21165 .57795 .22593 .58694 .22191 Metetra
.384 .57 .91 r
.56357 .20176 .55459 .19853 .56628 .21165 .57527 .21136 Metetra
0 .185 .608 r
.55185 .19424 .54288 .18853 .55459 .19853 .56357 .20176 Metetra
.74 .264 .114 r
.5401 .18961 .53114 .18309 .54288 .18853 .55185 .19424 Metetra
.711 .303 .273 r
.52833 .18807 .51938 .18264 .53114 .18309 .5401 .18961 Metetra
.615 .189 .21 r
.51652 .18913 .50759 .18629 .51938 .18264 .52833 .18807 Metetra
.296 0 0 r
.50468 .19172 .49576 .19213 .50759 .18629 .51652 .18913 Metetra
.515 .906 .936 r
.4928 .19452 .48388 .19792 .49576 .19213 .50468 .19172 Metetra
.722 .868 .95 r
.48088 .19643 .47197 .20178 .48388 .19792 .4928 .19452 Metetra
.727 .798 .908 r
.46892 .1968 .46003 .20277 .47197 .20178 .48088 .19643 Metetra
.705 .762 .896 r
.45694 .19556 .44805 .20098 .46003 .20277 .46892 .1968 Metetra
.674 .746 .905 r
.44491 .19309 .43604 .19727 .44805 .20098 .45694 .19556 Metetra
.635 .746 .929 r
.43286 .19 .42399 .19284 .43604 .19727 .44491 .19309 Metetra
.588 .755 .958 r
.42076 .1869 .41191 .1887 .42399 .19284 .43286 .19 Metetra
.546 .77 .981 r
.40864 .18416 .3998 .18546 .41191 .1887 .42076 .1869 Metetra
.527 .785 .992 r
.39647 .18195 .38765 .18323 .3998 .18546 .40864 .18416 Metetra
.54 .8 .993 r
.38427 .1802 .37546 .18176 .38765 .18323 .39647 .18195 Metetra
.567 .808 .989 r
.37204 .17872 .36323 .18068 .37546 .18176 .38427 .1802 Metetra
.591 .808 .983 r
.35976 .17732 .35097 .17961 .36323 .18068 .37204 .17872 Metetra
.605 .804 .976 r
.34745 .17587 .33867 .17835 .35097 .17961 .35976 .17732 Metetra
.609 .8 .972 r
.33511 .17429 .32633 .17684 .33867 .17835 .34745 .17587 Metetra
.608 .796 .971 r
.32272 .1726 .31396 .1751 .32633 .17684 .33511 .17429 Metetra
.604 .795 .972 r
.3103 .17082 .30155 .17325 .31396 .1751 .32272 .1726 Metetra
.6 .794 .974 r
.29785 .169 .28911 .17137 .30155 .17325 .3103 .17082 Metetra
.597 .795 .975 r
.28535 .16717 .27663 .1695 .28911 .17137 .29785 .169 Metetra
.596 .796 .976 r
.27282 .16536 .26411 .16768 .27663 .1695 .28535 .16717 Metetra
.614 .788 .965 r
.84285 .12795 .83367 .13021 .84454 .13191 .85373 .12965 Metetra
.614 .788 .965 r
.83193 .12624 .82277 .12851 .83367 .13021 .84285 .12795 Metetra
.614 .788 .965 r
.82099 .12453 .81183 .12681 .82277 .12851 .83193 .12624 Metetra
.614 .788 .965 r
.81002 .12281 .80087 .1251 .81183 .12681 .82099 .12453 Metetra
.614 .788 .965 r
.79902 .12109 .78987 .12338 .80087 .1251 .81002 .12281 Metetra
.614 .788 .965 r
.78799 .11937 .77885 .12166 .78987 .12338 .79902 .12109 Metetra
.614 .788 .965 r
.77692 .11764 .76779 .11994 .77885 .12166 .78799 .11937 Metetra
.751 .678 .781 r
.76583 .1159 .7568 .1307 .76779 .11994 .77692 .11764 Metetra
.85 .729 .729 r
.75486 .13469 .74604 .17615 .7568 .1307 .76583 .1159 Metetra
.931 .89 .481 r
.744 .17258 .73512 .20787 .74604 .17615 .75486 .13469 Metetra
.972 .906 .69 r
.733 .1996 .72407 .22975 .73512 .20787 .744 .17258 Metetra
.923 .851 .754 r
.72189 .21862 .71292 .24455 .72407 .22975 .733 .1996 Metetra
.875 .804 .777 r
.71069 .23173 .70171 .25423 .71292 .24455 .72189 .21862 Metetra
.837 .769 .786 r
.69943 .2404 .69042 .26013 .70171 .25423 .71069 .23173 Metetra
.808 .742 .791 r
.68811 .24567 .67909 .26319 .69042 .26013 .69943 .2404 Metetra
.785 .722 .793 r
.67673 .24828 .66771 .26401 .67909 .26319 .68811 .24567 Metetra
.766 .705 .794 r
.66531 .24869 .65628 .26294 .66771 .26401 .67673 .24828 Metetra
.749 .69 .795 r
.65385 .24722 .64482 .26015 .65628 .26294 .66531 .24869 Metetra
.734 .678 .796 r
.64234 .24404 .63331 .25567 .64482 .26015 .65385 .24722 Metetra
.718 .667 .799 r
.6308 .23931 .62177 .24949 .63331 .25567 .64234 .24404 Metetra
.701 .658 .804 r
.61922 .23319 .61019 .24163 .62177 .24949 .6308 .23931 Metetra
.679 .65 .814 r
.60761 .22597 .59858 .23227 .61019 .24163 .61922 .23319 Metetra
.648 .646 .834 r
.59596 .21807 .58694 .22191 .59858 .23227 .60761 .22597 Metetra
.597 .645 .869 r
.58429 .21012 .57527 .21136 .58694 .22191 .59596 .21807 Metetra
.487 .643 .926 r
.57258 .20285 .56357 .20176 .57527 .21136 .58429 .21012 Metetra
.195 .584 .942 r
.56086 .19695 .55185 .19424 .56357 .20176 .57258 .20285 Metetra
.293 0 0 r
.5491 .1929 .5401 .18961 .55185 .19424 .56086 .19695 Metetra
.451 0 0 r
.53731 .19083 .52833 .18807 .5401 .18961 .5491 .1929 Metetra
.329 0 0 r
.52549 .19044 .51652 .18913 .52833 .18807 .53731 .19083 Metetra
.071 .634 .734 r
.51364 .19111 .50468 .19172 .51652 .18913 .52549 .19044 Metetra
.506 .872 .982 r
.50175 .1921 .4928 .19452 .50468 .19172 .51364 .19111 Metetra
.656 .859 .977 r
.48982 .19271 .48088 .19643 .4928 .19452 .50175 .1921 Metetra
.685 .815 .945 r
.47785 .19251 .46892 .1968 .48088 .19643 .48982 .19271 Metetra
.679 .786 .931 r
.46586 .19139 .45694 .19556 .46892 .1968 .47785 .19251 Metetra
.66 .771 .932 r
.45382 .1895 .44491 .19309 .45694 .19556 .46586 .19139 Metetra
.634 .768 .944 r
.44175 .18714 .43286 .19 .44491 .19309 .45382 .1895 Metetra
.605 .771 .96 r
.42965 .18465 .42076 .1869 .43286 .19 .44175 .18714 Metetra
.58 .779 .974 r
.41751 .1823 .40864 .18416 .42076 .1869 .42965 .18465 Metetra
.567 .788 .982 r
.40534 .18019 .39647 .18195 .40864 .18416 .41751 .1823 Metetra
.568 .796 .985 r
.39313 .17834 .38427 .1802 .39647 .18195 .40534 .18019 Metetra
.578 .801 .984 r
.38088 .17667 .37204 .17872 .38427 .1802 .39313 .17834 Metetra
.59 .803 .981 r
.36859 .17508 .35976 .17732 .37204 .17872 .38088 .17667 Metetra
.598 .801 .977 r
.35627 .1735 .34745 .17587 .35976 .17732 .36859 .17508 Metetra
.603 .799 .975 r
.34392 .17186 .33511 .17429 .34745 .17587 .35627 .1735 Metetra
.603 .797 .974 r
.33152 .17016 .32272 .1726 .33511 .17429 .34392 .17186 Metetra
.602 .796 .974 r
.31909 .1684 .3103 .17082 .32272 .1726 .33152 .17016 Metetra
.6 .795 .974 r
.30662 .16661 .29785 .169 .3103 .17082 .31909 .1684 Metetra
.598 .796 .975 r
.29412 .16481 .28535 .16717 .29785 .169 .30662 .16661 Metetra
.597 .796 .975 r
.28157 .16301 .27282 .16536 .28535 .16717 .29412 .16481 Metetra
.614 .788 .965 r
.85206 .12567 .84285 .12795 .85373 .12965 .86294 .12738 Metetra
.614 .788 .965 r
.84114 .12396 .83193 .12624 .84285 .12795 .85206 .12567 Metetra
.614 .788 .965 r
.83019 .12224 .82099 .12453 .83193 .12624 .84114 .12396 Metetra
.614 .788 .965 r
.81921 .12052 .81002 .12281 .82099 .12453 .83019 .12224 Metetra
.614 .788 .965 r
.8082 .11879 .79902 .12109 .81002 .12281 .81921 .12052 Metetra
.614 .788 .965 r
.79716 .11706 .78799 .11937 .79902 .12109 .8082 .11879 Metetra
.614 .788 .965 r
.78609 .11533 .77692 .11764 .78799 .11937 .79716 .11706 Metetra
.614 .788 .965 r
.77498 .11358 .76583 .1159 .77692 .11764 .78609 .11533 Metetra
.758 .659 .752 r
.76385 .11184 .75486 .13469 .76583 .1159 .77498 .11358 Metetra
0 0 .239 r
.75298 .14807 .744 .17258 .75486 .13469 .76385 .11184 Metetra
.742 .17734 .733 .1996 .744 .17258 .75298 .14807 Metetra
.904 .955 .615 r
.73091 .19852 .72189 .21862 .733 .1996 .742 .17734 Metetra
.928 .926 .786 r
.71974 .21358 .71069 .23173 .72189 .21862 .73091 .19852 Metetra
.876 .851 .815 r
.70848 .22397 .69943 .2404 .71069 .23173 .71974 .21358 Metetra
.832 .798 .819 r
.69717 .23075 .68811 .24567 .69943 .2404 .70848 .22397 Metetra
.8 .761 .818 r
.6858 .23468 .67673 .24828 .68811 .24567 .69717 .23075 Metetra
.775 .735 .816 r
.67438 .23628 .66531 .24869 .67673 .24828 .6858 .23468 Metetra
.755 .715 .815 r
.66292 .23593 .65385 .24722 .66531 .24869 .67438 .23628 Metetra
.737 .699 .815 r
.65141 .23393 .64234 .24404 .65385 .24722 .66292 .23593 Metetra
.72 .687 .818 r
.63987 .23049 .6308 .23931 .64234 .24404 .65141 .23393 Metetra
.702 .679 .824 r
.62829 .22585 .61922 .23319 .6308 .23931 .63987 .23049 Metetra
.681 .673 .835 r
.61667 .22032 .60761 .22597 .61922 .23319 .62829 .22585 Metetra
.654 .672 .855 r
.60502 .21427 .59596 .21807 .60761 .22597 .61667 .22032 Metetra
.614 .676 .886 r
.59334 .20817 .58429 .21012 .59596 .21807 .60502 .21427 Metetra
.546 .687 .931 r
.58163 .20251 .57258 .20285 .58429 .21012 .59334 .20817 Metetra
.424 .693 .976 r
.56989 .19773 .56086 .19695 .57258 .20285 .58163 .20251 Metetra
.238 .667 .959 r
.55813 .19416 .5491 .1929 .56086 .19695 .56989 .19773 Metetra
.104 .628 .88 r
.54633 .19185 .53731 .19083 .5491 .1929 .55813 .19416 Metetra
.154 .677 .879 r
.5345 .19066 .52549 .19044 .53731 .19083 .54633 .19185 Metetra
.344 .789 .958 r
.52263 .19022 .51364 .19111 .52549 .19044 .5345 .19066 Metetra
.521 .847 .998 r
.51073 .19008 .50175 .1921 .51364 .19111 .52263 .19022 Metetra
.613 .841 .986 r
.49879 .18983 .48982 .19271 .50175 .1921 .51073 .19008 Metetra
.647 .818 .966 r
.48682 .18918 .47785 .19251 .48982 .19271 .49879 .18983 Metetra
.652 .799 .953 r
.47481 .188 .46586 .19139 .47785 .19251 .48682 .18918 Metetra
.643 .787 .951 r
.46277 .18635 .45382 .1895 .46586 .19139 .47481 .188 Metetra
.627 .781 .955 r
.45069 .18437 .44175 .18714 .45382 .1895 .46277 .18635 Metetra
.61 .782 .963 r
.43857 .18225 .42965 .18465 .44175 .18714 .45069 .18437 Metetra
.594 .785 .972 r
.42642 .18014 .41751 .1823 .42965 .18465 .43857 .18225 Metetra
.585 .79 .977 r
.41424 .17814 .40534 .18019 .41751 .1823 .42642 .18014 Metetra
.582 .795 .98 r
.40201 .17628 .39313 .17834 .40534 .18019 .41424 .17814 Metetra
.586 .798 .98 r
.38976 .17453 .38088 .17667 .39313 .17834 .40201 .17628 Metetra
.591 .799 .979 r
.37746 .17284 .36859 .17508 .38088 .17667 .38976 .17453 Metetra
.596 .799 .977 r
.36513 .17117 .35627 .1735 .36859 .17508 .37746 .17284 Metetra
.599 .798 .976 r
.35276 .16948 .34392 .17186 .35627 .1735 .36513 .17117 Metetra
.6 .797 .975 r
.34035 .16776 .33152 .17016 .34392 .17186 .35276 .16948 Metetra
.6 .797 .975 r
.32791 .16601 .31909 .1684 .33152 .17016 .34035 .16776 Metetra
.599 .796 .975 r
.31543 .16423 .30662 .16661 .31909 .1684 .32791 .16601 Metetra
.598 .796 .975 r
.30291 .16244 .29412 .16481 .30662 .16661 .31543 .16423 Metetra
.598 .796 .975 r
.29036 .16064 .28157 .16301 .29412 .16481 .30291 .16244 Metetra
.614 .788 .965 r
.8613 .12339 .85206 .12567 .86294 .12738 .87219 .12511 Metetra
.614 .788 .965 r
.85037 .12167 .84114 .12396 .85206 .12567 .8613 .12339 Metetra
.614 .788 .965 r
.83942 .11995 .83019 .12224 .84114 .12396 .85037 .12167 Metetra
.614 .788 .965 r
.82843 .11822 .81921 .12052 .83019 .12224 .83942 .11995 Metetra
.614 .788 .965 r
.81741 .11649 .8082 .11879 .81921 .12052 .82843 .11822 Metetra
.614 .788 .965 r
.80637 .11475 .79716 .11706 .8082 .11879 .81741 .11649 Metetra
.614 .788 .965 r
.79529 .113 .78609 .11533 .79716 .11706 .80637 .11475 Metetra
.614 .788 .965 r
.78417 .11126 .77498 .11358 .78609 .11533 .79529 .113 Metetra
.614 .788 .965 r
.77303 .1095 .76385 .11184 .77498 .11358 .78417 .11126 Metetra
.198 0 .369 r
.76207 .13432 .75298 .14807 .76385 .11184 .77303 .1095 Metetra
.294 .113 .456 r
.7511 .16313 .742 .17734 .75298 .14807 .76207 .13432 Metetra
0 0 .137 r
.74001 .18446 .73091 .19852 .742 .17734 .7511 .16313 Metetra
.722 .91 .548 r
.72884 .20002 .71974 .21358 .73091 .19852 .74001 .18446 Metetra
.889 .969 .832 r
.71759 .21109 .70848 .22397 .71974 .21358 .72884 .20002 Metetra
.852 .878 .858 r
.70628 .21862 .69717 .23075 .70848 .22397 .71759 .21109 Metetra
.812 .814 .852 r
.69491 .22333 .6858 .23468 .69717 .23075 .70628 .21862 Metetra
.782 .773 .845 r
.68349 .22575 .67438 .23628 .6858 .23468 .69491 .22333 Metetra
.758 .745 .84 r
.67202 .22627 .66292 .23593 .67438 .23628 .68349 .22575 Metetra
.739 .725 .839 r
.66052 .22521 .65141 .23393 .66292 .23593 .67202 .22627 Metetra
.721 .711 .84 r
.64897 .22282 .63987 .23049 .65141 .23393 .66052 .22521 Metetra
.703 .702 .846 r
.63739 .21939 .62829 .22585 .63987 .23049 .64897 .22282 Metetra
.683 .697 .857 r
.62577 .21517 .61667 .22032 .62829 .22585 .63739 .21939 Metetra
.658 .697 .875 r
.61411 .21051 .60502 .21427 .61667 .22032 .62577 .21517 Metetra
.626 .703 .9 r
.60243 .20574 .59334 .20817 .60502 .21427 .61411 .21051 Metetra
.581 .715 .935 r
.59071 .20123 .58163 .20251 .59334 .20817 .60243 .20574 Metetra
.518 .73 .97 r
.57896 .19728 .56989 .19773 .58163 .20251 .59071 .20123 Metetra
.443 .744 .992 r
.56719 .19409 .55813 .19416 .56989 .19773 .57896 .19728 Metetra
.391 .756 .992 r
.55538 .19173 .54633 .19185 .55813 .19416 .56719 .19409 Metetra
.4 .78 .992 r
.54354 .19011 .5345 .19066 .54633 .19185 .55538 .19173 Metetra
.465 .811 .998 r
.53166 .18902 .52263 .19022 .5345 .19066 .54354 .19011 Metetra
.54 .828 .998 r
.51975 .1882 .51073 .19008 .52263 .19022 .53166 .18902 Metetra
.593 .826 .988 r
.5078 .18738 .49879 .18983 .51073 .19008 .51975 .1882 Metetra
.62 .815 .975 r
.49582 .18638 .48682 .18918 .49879 .18983 .5078 .18738 Metetra
.629 .803 .966 r
.4838 .18509 .47481 .188 .48682 .18918 .49582 .18638 Metetra
.627 .794 .962 r
.47174 .18351 .46277 .18635 .47481 .188 .4838 .18509 Metetra
.62 .79 .963 r
.45965 .18171 .45069 .18437 .46277 .18635 .47174 .18351 Metetra
.609 .788 .967 r
.44753 .1798 .43857 .18225 .45069 .18437 .45965 .18171 Metetra
.6 .789 .971 r
.43537 .17785 .42642 .18014 .43857 .18225 .44753 .1798 Metetra
.593 .792 .975 r
.42317 .17594 .41424 .17814 .42642 .18014 .43537 .17785 Metetra
.59 .794 .977 r
.41094 .1741 .40201 .17628 .41424 .17814 .42317 .17594 Metetra
.591 .797 .978 r
.39867 .17231 .38976 .17453 .40201 .17628 .41094 .1741 Metetra
.593 .798 .978 r
.38636 .17057 .37746 .17284 .38976 .17453 .39867 .17231 Metetra
.596 .798 .977 r
.37402 .16885 .36513 .17117 .37746 .17284 .38636 .17057 Metetra
.598 .798 .976 r
.36164 .16712 .35276 .16948 .36513 .17117 .37402 .16885 Metetra
.599 .797 .975 r
.34922 .16538 .34035 .16776 .35276 .16948 .36164 .16712 Metetra
.599 .797 .975 r
.33677 .16362 .32791 .16601 .34035 .16776 .34922 .16538 Metetra
.599 .796 .975 r
.32428 .16184 .31543 .16423 .32791 .16601 .33677 .16362 Metetra
.598 .796 .975 r
.31175 .16005 .30291 .16244 .31543 .16423 .32428 .16184 Metetra
.598 .796 .975 r
.29918 .15825 .29036 .16064 .30291 .16244 .31175 .16005 Metetra
.614 .788 .965 r
.87058 .1211 .8613 .12339 .87219 .12511 .88148 .12282 Metetra
.614 .788 .965 r
.85965 .11938 .85037 .12167 .8613 .12339 .87058 .1211 Metetra
.614 .788 .965 r
.84868 .11765 .83942 .11995 .85037 .12167 .85965 .11938 Metetra
.614 .788 .965 r
.83769 .11591 .82843 .11822 .83942 .11995 .84868 .11765 Metetra
.614 .788 .965 r
.82666 .11417 .81741 .11649 .82843 .11822 .83769 .11591 Metetra
.614 .788 .965 r
.81561 .11243 .80637 .11475 .81741 .11649 .82666 .11417 Metetra
.614 .788 .965 r
.80452 .11068 .79529 .113 .80637 .11475 .81561 .11243 Metetra
.614 .788 .965 r
.7934 .10892 .78417 .11126 .79529 .113 .80452 .11068 Metetra
.614 .788 .965 r
.78225 .10716 .77303 .1095 .78417 .11126 .7934 .10892 Metetra
.535 .343 .567 r
.77128 .13031 .76207 .13432 .77303 .1095 .78225 .10716 Metetra
.497 .31 .558 r
.76028 .15637 .7511 .16313 .76207 .13432 .77128 .13031 Metetra
.372 .147 .438 r
.74918 .17615 .74001 .18446 .7511 .16313 .76028 .15637 Metetra
.043 0 .1 r
.738 .19092 .72884 .20002 .74001 .18446 .74918 .17615 Metetra
.623 .912 .628 r
.72675 .2017 .71759 .21109 .72884 .20002 .738 .19092 Metetra
.828 .971 .89 r
.71543 .20927 .70628 .21862 .71759 .21109 .72675 .2017 Metetra
.811 .881 .896 r
.70406 .21424 .69491 .22333 .70628 .21862 .71543 .20927 Metetra
.781 .819 .882 r
.69264 .21708 .68349 .22575 .69491 .22333 .70406 .21424 Metetra
.757 .779 .871 r
.68117 .21816 .67202 .22627 .68349 .22575 .69264 .21708 Metetra
.737 .753 .866 r
.66966 .21777 .66052 .22521 .67202 .22627 .68117 .21816 Metetra
.719 .736 .865 r
.65811 .21619 .64897 .22282 .66052 .22521 .66966 .21777 Metetra
.701 .726 .869 r
.64652 .21367 .63739 .21939 .64897 .22282 .65811 .21619 Metetra
.682 .72 .878 r
.6349 .21045 .62577 .21517 .63739 .21939 .64652 .21367 Metetra
.66 .72 .893 r
.62324 .20681 .61411 .21051 .62577 .21517 .6349 .21045 Metetra
.634 .725 .914 r
.61155 .20303 .60243 .20574 .61411 .21051 .62324 .20681 Metetra
.602 .735 .939 r
.59983 .19936 .59071 .20123 .60243 .20574 .61155 .20303 Metetra
.563 .749 .964 r
.58807 .19602 .57896 .19728 .59071 .20123 .59983 .19936 Metetra
.524 .763 .984 r
.57628 .19316 .56719 .19409 .57896 .19728 .58807 .19602 Metetra
.498 .778 .995 r
.56446 .19083 .55538 .19173 .56719 .19409 .57628 .19316 Metetra
.498 .794 .998 r
.55261 .189 .54354 .19011 .55538 .19173 .56446 .19083 Metetra
.522 .807 .998 r
.54072 .18755 .53166 .18902 .54354 .19011 .55261 .189 Metetra
.556 .815 .994 r
.5288 .18631 .51975 .1882 .53166 .18902 .54072 .18755 Metetra
.586 .815 .987 r
.51684 .18513 .5078 .18738 .51975 .1882 .5288 .18631 Metetra
.605 .81 .979 r
.50485 .18387 .49582 .18638 .5078 .18738 .51684 .18513 Metetra
.614 .803 .972 r
.49282 .18246 .4838 .18509 .49582 .18638 .50485 .18387 Metetra
.616 .798 .969 r
.48076 .18089 .47174 .18351 .4838 .18509 .49282 .18246 Metetra
.612 .794 .969 r
.46866 .17917 .45965 .18171 .47174 .18351 .48076 .18089 Metetra
.607 .792 .97 r
.45652 .17735 .44753 .1798 .45965 .18171 .46866 .17917 Metetra
.601 .792 .972 r
.44435 .1755 .43537 .17785 .44753 .1798 .45652 .17735 Metetra
.597 .793 .974 r
.43214 .17365 .42317 .17594 .43537 .17785 .44435 .1755 Metetra
.595 .795 .976 r
.4199 .17183 .41094 .1741 .42317 .17594 .43214 .17365 Metetra
.594 .796 .977 r
.40762 .17004 .39867 .17231 .41094 .1741 .4199 .17183 Metetra
.595 .797 .977 r
.3953 .16827 .38636 .17057 .39867 .17231 .40762 .17004 Metetra
.596 .797 .976 r
.38295 .16652 .37402 .16885 .38636 .17057 .3953 .16827 Metetra
.597 .797 .976 r
.37055 .16477 .36164 .16712 .37402 .16885 .38295 .16652 Metetra
.598 .797 .976 r
.35813 .16301 .34922 .16538 .36164 .16712 .37055 .16477 Metetra
.598 .797 .975 r
.34566 .16123 .33677 .16362 .34922 .16538 .35813 .16301 Metetra
.598 .797 .975 r
.33316 .15945 .32428 .16184 .33677 .16362 .34566 .16123 Metetra
.598 .796 .975 r
.32061 .15766 .31175 .16005 .32428 .16184 .33316 .15945 Metetra
.598 .796 .975 r
.30804 .15586 .29918 .15825 .31175 .16005 .32061 .15766 Metetra
.614 .788 .965 r
.87989 .1188 .87058 .1211 .88148 .12282 .8908 .12053 Metetra
.614 .788 .965 r
.86895 .11707 .85965 .11938 .87058 .1211 .87989 .1188 Metetra
.614 .788 .965 r
.85798 .11533 .84868 .11765 .85965 .11938 .86895 .11707 Metetra
.614 .788 .965 r
.84698 .11359 .83769 .11591 .84868 .11765 .85798 .11533 Metetra
.614 .788 .965 r
.83595 .11185 .82666 .11417 .83769 .11591 .84698 .11359 Metetra
.614 .788 .965 r
.82488 .11009 .81561 .11243 .82666 .11417 .83595 .11185 Metetra
.614 .788 .965 r
.81379 .10834 .80452 .11068 .81561 .11243 .82488 .11009 Metetra
.614 .788 .965 r
.80266 .10658 .7934 .10892 .80452 .11068 .81379 .10834 Metetra
.022 0 0 r
.79152 .10697 .78225 .10716 .7934 .10892 .80266 .10658 Metetra
.619 .445 .625 r
.78058 .1343 .77128 .13031 .78225 .10716 .79152 .10697 Metetra
.582 .39 .587 r
.76955 .15594 .76028 .15637 .77128 .13031 .78058 .1343 Metetra
.518 .294 .514 r
.75842 .17284 .74918 .17615 .76028 .15637 .76955 .15594 Metetra
.378 .092 .346 r
.74722 .18579 .738 .19092 .74918 .17615 .75842 .17284 Metetra
.73596 .19549 .72675 .2017 .738 .19092 .74722 .18579 Metetra
.628 .958 .808 r
.72464 .20249 .71543 .20927 .72675 .2017 .73596 .19549 Metetra
.768 .944 .937 r
.71326 .20725 .70406 .21424 .71543 .20927 .72464 .20249 Metetra
.764 .867 .924 r
.70183 .21014 .69264 .21708 .70406 .21424 .71326 .20725 Metetra
.746 .816 .907 r
.69036 .21147 .68117 .21816 .69264 .21708 .70183 .21014 Metetra
.729 .783 .896 r
.67884 .2115 .66966 .21777 .68117 .21816 .69036 .21147 Metetra
.712 .762 .892 r
.66729 .21047 .65811 .21619 .66966 .21777 .67884 .2115 Metetra
.696 .749 .893 r
.6557 .20861 .64652 .21367 .65811 .21619 .66729 .21047 Metetra
.678 .743 .899 r
.64407 .20613 .6349 .21045 .64652 .21367 .6557 .20861 Metetra
.659 .741 .91 r
.6324 .20324 .62324 .20681 .6349 .21045 .64407 .20613 Metetra
.638 .744 .925 r
.6207 .20018 .61155 .20303 .62324 .20681 .6324 .20324 Metetra
.613 .751 .943 r
.60897 .19713 .59983 .19936 .61155 .20303 .6207 .20018 Metetra
.587 .761 .961 r
.59721 .19425 .58807 .19602 .59983 .19936 .60897 .19713 Metetra
.563 .772 .977 r
.58541 .19167 .57628 .19316 .58807 .19602 .59721 .19425 Metetra
.546 .783 .986 r
.57358 .18942 .56446 .19083 .57628 .19316 .58541 .19167 Metetra
.543 .794 .991 r
.56172 .1875 .55261 .189 .56446 .19083 .57358 .18942 Metetra
.552 .802 .991 r
.54982 .18585 .54072 .18755 .55261 .189 .56172 .1875 Metetra
.568 .807 .989 r
.53789 .18436 .5288 .18631 .54072 .18755 .54982 .18585 Metetra
.585 .808 .984 r
.52592 .18294 .51684 .18513 .5288 .18631 .53789 .18436 Metetra
.598 .806 .98 r
.51392 .18151 .50485 .18387 .51684 .18513 .52592 .18294 Metetra
.605 .802 .975 r
.50188 .18 .49282 .18246 .50485 .18387 .51392 .18151 Metetra
.608 .799 .973 r
.48981 .17839 .48076 .18089 .49282 .18246 .50188 .18 Metetra
.607 .796 .972 r
.4777 .1767 .46866 .17917 .48076 .18089 .48981 .17839 Metetra
.604 .795 .972 r
.46555 .17493 .45652 .17735 .46866 .17917 .4777 .1767 Metetra
.601 .794 .973 r
.45337 .17313 .44435 .1755 .45652 .17735 .46555 .17493 Metetra
.598 .795 .974 r
.44115 .17132 .43214 .17365 .44435 .1755 .45337 .17313 Metetra
.597 .795 .975 r
.42889 .16952 .4199 .17183 .43214 .17365 .44115 .17132 Metetra
.596 .796 .976 r
.4166 .16773 .40762 .17004 .4199 .17183 .42889 .16952 Metetra
.596 .796 .976 r
.40427 .16595 .3953 .16827 .40762 .17004 .4166 .16773 Metetra
.597 .797 .976 r
.39191 .16418 .38295 .16652 .3953 .16827 .40427 .16595 Metetra
.597 .797 .976 r
.3795 .16241 .37055 .16477 .38295 .16652 .39191 .16418 Metetra
.598 .797 .976 r
.36706 .16063 .35813 .16301 .37055 .16477 .3795 .16241 Metetra
.598 .797 .976 r
.35459 .15885 .34566 .16123 .35813 .16301 .36706 .16063 Metetra
.598 .797 .975 r
.34207 .15706 .33316 .15945 .34566 .16123 .35459 .15885 Metetra
.598 .796 .975 r
.32952 .15526 .32061 .15766 .33316 .15945 .34207 .15706 Metetra
.598 .796 .975 r
.31693 .15345 .30804 .15586 .32061 .15766 .32952 .15526 Metetra
.627 .625 .829 r
.88924 .11649 .87989 .1188 .8908 .12053 .90025 .12564 Metetra
.614 .788 .965 r
.87829 .11476 .86895 .11707 .87989 .1188 .88924 .11649 Metetra
.614 .788 .965 r
.86732 .11301 .85798 .11533 .86895 .11707 .87829 .11476 Metetra
.614 .788 .965 r
.85631 .11126 .84698 .11359 .85798 .11533 .86732 .11301 Metetra
.614 .788 .965 r
.84527 .10951 .83595 .11185 .84698 .11359 .85631 .11126 Metetra
.614 .788 .965 r
.83419 .10775 .82488 .11009 .83595 .11185 .84527 .10951 Metetra
.614 .788 .965 r
.82309 .10599 .81379 .10834 .82488 .11009 .83419 .10775 Metetra
.614 .788 .965 r
.81196 .10422 .80266 .10658 .81379 .10834 .82309 .10599 Metetra
.686 .51 .643 r
.801 .12459 .79152 .10697 .80266 .10658 .81196 .10422 Metetra
.661 .477 .625 r
.78998 .14411 .78058 .1343 .79152 .10697 .801 .12459 Metetra
.633 .432 .593 r
.77889 .16032 .76955 .15594 .78058 .1343 .78998 .14411 Metetra
.59 .359 .534 r
.76773 .17349 .75842 .17284 .76955 .15594 .77889 .16032 Metetra
.507 .222 .416 r
.75651 .18393 .74722 .18579 .75842 .17284 .76773 .17349 Metetra
.299 0 .124 r
.74523 .19197 .73596 .19549 .74722 .18579 .75651 .18393 Metetra
.73389 .19794 .72464 .20249 .73596 .19549 .74523 .19197 Metetra
.644 .956 .936 r
.7225 .2021 .71326 .20725 .72464 .20249 .73389 .19794 Metetra
.718 .904 .961 r
.71107 .20473 .70183 .21014 .71326 .20725 .7225 .2021 Metetra
.721 .848 .942 r
.69959 .20605 .69036 .21147 .70183 .21014 .71107 .20473 Metetra
.712 .811 .926 r
.68807 .20627 .67884 .2115 .69036 .21147 .69959 .20605 Metetra
.7 .786 .918 r
.67651 .20557 .66729 .21047 .67884 .2115 .68807 .20627 Metetra
.686 .771 .916 r
.66491 .20415 .6557 .20861 .66729 .21047 .67651 .20557 Metetra
.671 .762 .919 r
.65327 .20218 .64407 .20613 .6557 .20861 .66491 .20415 Metetra
.655 .759 .926 r
.6416 .19983 .6324 .20324 .64407 .20613 .65327 .20218 Metetra
.638 .759 .936 r
.6299 .19729 .6207 .20018 .6324 .20324 .6416 .19983 Metetra
.619 .763 .949 r
.61816 .19469 .60897 .19713 .6207 .20018 .6299 .19729 Metetra
.6 .77 .961 r
.60638 .19217 .59721 .19425 .60897 .19713 .61816 .19469 Metetra
.583 .778 .972 r
.59458 .18981 .58541 .19167 .59721 .19425 .60638 .19217 Metetra
.571 .785 .98 r
.58274 .18766 .57358 .18942 .58541 .19167 .59458 .18981 Metetra
.567 .793 .984 r
.57087 .18572 .56172 .1875 .57358 .18942 .58274 .18766 Metetra
.57 .798 .985 r
.55896 .18397 .54982 .18585 .56172 .1875 .57087 .18572 Metetra
.578 .802 .984 r
.54702 .18233 .53789 .18436 .54982 .18585 .55896 .18397 Metetra
.587 .803 .982 r
.53504 .18077 .52592 .18294 .53789 .18436 .54702 .18233 Metetra
.595 .802 .979 r
.52302 .17921 .51392 .18151 .52592 .18294 .53504 .18077 Metetra
.6 .801 .977 r
.51098 .17762 .50188 .18 .51392 .18151 .52302 .17921 Metetra
.603 .799 .975 r
.49889 .17598 .48981 .17839 .50188 .18 .51098 .17762 Metetra
.603 .797 .974 r
.48677 .17428 .4777 .1767 .48981 .17839 .49889 .17598 Metetra
.602 .796 .974 r
.47461 .17253 .46555 .17493 .4777 .1767 .48677 .17428 Metetra
.6 .795 .974 r
.46242 .17076 .45337 .17313 .46555 .17493 .47461 .17253 Metetra
.599 .795 .975 r
.45019 .16897 .44115 .17132 .45337 .17313 .46242 .17076 Metetra
.598 .796 .975 r
.43792 .16718 .42889 .16952 .44115 .17132 .45019 .16897 Metetra
.597 .796 .976 r
.42562 .16539 .4166 .16773 .42889 .16952 .43792 .16718 Metetra
.597 .796 .976 r
.41328 .1636 .40427 .16595 .4166 .16773 .42562 .16539 Metetra
.597 .797 .976 r
.4009 .16182 .39191 .16418 .40427 .16595 .41328 .1636 Metetra
.597 .797 .976 r
.38849 .16003 .3795 .16241 .39191 .16418 .4009 .16182 Metetra
.598 .797 .976 r
.37604 .15825 .36706 .16063 .3795 .16241 .38849 .16003 Metetra
.598 .797 .976 r
.36355 .15645 .35459 .15885 .36706 .16063 .37604 .15825 Metetra
.598 .797 .975 r
.35102 .15465 .34207 .15706 .35459 .15885 .36355 .15645 Metetra
.598 .797 .975 r
.33846 .15285 .32952 .15526 .34207 .15706 .35102 .15465 Metetra
.598 .796 .975 r
.32586 .15103 .31693 .15345 .32952 .15526 .33846 .15285 Metetra
.775 .615 .681 r
.90196 .3849 .88924 .11649 .90025 .12564 .91307 .39426 Metetra
.858 .648 .112 r
.88924 .24327 .87829 .11476 .88924 .11649 .90196 .3849 Metetra
0 0 .526 r
.87733 .16652 .86732 .11301 .87829 .11476 .88924 .24327 Metetra
.336 .353 .73 r
.8659 .12936 .85631 .11126 .86732 .11301 .87733 .16652 Metetra
.333 .481 .858 r
.85471 .11585 .84527 .10951 .85631 .11126 .8659 .12936 Metetra
.737 .437 .473 r
.84365 .11592 .83419 .10775 .84527 .10951 .85471 .11585 Metetra
.723 .522 .618 r
.83263 .12328 .82309 .10599 .83419 .10775 .84365 .11592 Metetra
.731 .55 .647 r
.82161 .13403 .81196 .10422 .82309 .10599 .83263 .12328 Metetra
.716 .53 .637 r
.81055 .14584 .801 .12459 .81196 .10422 .82161 .13403 Metetra
.696 .498 .616 r
.79945 .15736 .78998 .14411 .801 .12459 .81055 .14584 Metetra
.671 .455 .584 r
.7883 .16784 .77889 .16032 .78998 .14411 .79945 .15736 Metetra
.636 .389 .528 r
.7771 .17692 .76773 .17349 .77889 .16032 .7883 .16784 Metetra
.573 .274 .423 r
.76585 .18449 .75651 .18393 .76773 .17349 .7771 .17692 Metetra
.43 .036 .189 r
.75454 .19055 .74523 .19197 .75651 .18393 .76585 .18449 Metetra
.048 0 0 r
.74319 .19517 .73389 .19794 .74523 .19197 .75454 .19055 Metetra
.458 .887 .878 r
.73179 .19849 .7225 .2021 .73389 .19794 .74319 .19517 Metetra
.641 .914 .98 r
.72034 .20062 .71107 .20473 .7225 .2021 .73179 .19849 Metetra
.68 .868 .97 r
.70886 .20172 .69959 .20605 .71107 .20473 .72034 .20062 Metetra
.686 .831 .953 r
.69733 .20193 .68807 .20627 .69959 .20605 .70886 .20172 Metetra
.681 .805 .942 r
.68576 .20138 .67651 .20557 .68807 .20627 .69733 .20193 Metetra
.672 .789 .936 r
.67415 .20021 .66491 .20415 .67651 .20557 .68576 .20138 Metetra
.661 .779 .936 r
.66251 .19857 .65327 .20218 .66491 .20415 .67415 .20021 Metetra
.649 .773 .94 r
.65083 .1966 .6416 .19983 .65327 .20218 .66251 .19857 Metetra
.635 .772 .946 r
.63912 .19442 .6299 .19729 .6416 .19983 .65083 .1966 Metetra
.62 .774 .954 r
.62738 .19216 .61816 .19469 .6299 .19729 .63912 .19442 Metetra
.606 .777 .962 r
.61559 .1899 .60638 .19217 .61816 .19469 .62738 .19216 Metetra
.594 .782 .97 r
.60378 .18773 .59458 .18981 .60638 .19217 .61559 .1899 Metetra
.585 .787 .976 r
.59193 .18567 .58274 .18766 .59458 .18981 .60378 .18773 Metetra
.581 .792 .98 r
.58005 .18375 .57087 .18572 .58274 .18766 .59193 .18567 Metetra
.581 .796 .981 r
.56813 .18194 .55896 .18397 .57087 .18572 .58005 .18375 Metetra
.584 .799 .981 r
.55618 .18023 .54702 .18233 .55896 .18397 .56813 .18194 Metetra
.589 .8 .98 r
.54419 .17857 .53504 .18077 .54702 .18233 .55618 .18023 Metetra
.594 .8 .978 r
.53217 .17693 .52302 .17921 .53504 .18077 .54419 .17857 Metetra
.598 .799 .977 r
.52011 .17528 .51098 .17762 .52302 .17921 .53217 .17693 Metetra
.6 .798 .976 r
.50801 .1736 .49889 .17598 .51098 .17762 .52011 .17528 Metetra
.601 .797 .975 r
.49588 .17189 .48677 .17428 .49889 .17598 .50801 .1736 Metetra
.6 .797 .975 r
.48371 .17015 .47461 .17253 .48677 .17428 .49588 .17189 Metetra
.6 .796 .975 r
.47151 .16839 .46242 .17076 .47461 .17253 .48371 .17015 Metetra
.599 .796 .975 r
.45927 .16661 .45019 .16897 .46242 .17076 .47151 .16839 Metetra
.598 .796 .975 r
.44699 .16482 .43792 .16718 .45019 .16897 .45927 .16661 Metetra
.598 .796 .975 r
.43468 .16303 .42562 .16539 .43792 .16718 .44699 .16482 Metetra
.597 .796 .976 r
.42233 .16124 .41328 .1636 .42562 .16539 .43468 .16303 Metetra
.597 .796 .976 r
.40994 .15945 .4009 .16182 .41328 .1636 .42233 .16124 Metetra
.597 .797 .976 r
.39751 .15765 .38849 .16003 .4009 .16182 .40994 .15945 Metetra
.598 .797 .976 r
.38505 .15585 .37604 .15825 .38849 .16003 .39751 .15765 Metetra
.598 .797 .976 r
.37255 .15405 .36355 .15645 .37604 .15825 .38505 .15585 Metetra
.598 .797 .976 r
.36001 .15224 .35102 .15465 .36355 .15645 .37255 .15405 Metetra
.598 .797 .975 r
.34743 .15043 .33846 .15285 .35102 .15465 .36001 .15224 Metetra
.598 .796 .975 r
.33482 .14861 .32586 .15103 .33846 .15285 .34743 .15043 Metetra
.727 .345 .327 r
.91156 .39108 .90196 .3849 .91307 .39426 .92258 .39245 Metetra
.771 .608 .676 r
.90051 .38972 .88924 .24327 .90196 .3849 .91156 .39108 Metetra
.85 .667 .654 r
.88864 .32213 .87733 .16652 .88924 .24327 .90051 .38972 Metetra
.92 .713 .606 r
.87667 .24524 .8659 .12936 .87733 .16652 .88864 .32213 Metetra
.871 .671 .63 r
.86509 .20133 .85471 .11585 .8659 .12936 .87667 .24524 Metetra
.833 .636 .637 r
.85374 .17827 .84365 .11592 .85471 .11585 .86509 .20133 Metetra
.803 .606 .636 r
.84251 .16813 .83263 .12328 .84365 .11592 .85374 .17827 Metetra
.777 .577 .629 r
.83133 .16571 .82161 .13403 .83263 .12328 .84251 .16813 Metetra
.754 .547 .615 r
.82016 .16765 .81055 .14584 .82161 .13403 .83133 .16571 Metetra
.731 .511 .593 r
.80898 .17182 .79945 .15736 .81055 .14584 .82016 .16765 Metetra
.705 .464 .557 r
.79777 .17685 .7883 .16784 .79945 .15736 .80898 .17182 Metetra
.67 .396 .497 r
.78653 .18194 .7771 .17692 .7883 .16784 .79777 .17685 Metetra
.611 .28 .385 r
.77524 .1866 .76585 .18449 .7771 .17692 .78653 .18194 Metetra
.482 .057 .156 r
.76391 .19057 .75454 .19055 .76585 .18449 .77524 .1866 Metetra
.163 0 0 r
.75254 .19373 .74319 .19517 .75454 .19055 .76391 .19057 Metetra
.311 .803 .831 r
.74112 .19605 .73179 .19849 .74319 .19517 .75254 .19373 Metetra
.554 .895 .982 r
.72966 .19754 .72034 .20062 .73179 .19849 .74112 .19605 Metetra
.63 .872 .987 r
.71817 .19828 .70886 .20172 .72034 .20062 .72966 .19754 Metetra
.653 .841 .973 r
.70663 .19833 .69733 .20193 .70886 .20172 .71817 .19828 Metetra
.658 .818 .96 r
.69505 .19778 .68576 .20138 .69733 .20193 .70663 .19833 Metetra
.655 .802 .953 r
.68344 .19673 .67415 .20021 .68576 .20138 .69505 .19778 Metetra
.649 .791 .95 r
.67179 .19529 .66251 .19857 .67415 .20021 .68344 .19673 Metetra
.64 .785 .951 r
.66011 .19356 .65083 .1966 .66251 .19857 .67179 .19529 Metetra
.63 .782 .954 r
.64838 .19163 .63912 .19442 .65083 .1966 .66011 .19356 Metetra
.619 .781 .959 r
.63663 .1896 .62738 .19216 .63912 .19442 .64838 .19163 Metetra
.609 .783 .965 r
.62484 .18753 .61559 .1899 .62738 .19216 .63663 .1896 Metetra
.6 .786 .97 r
.61302 .1855 .60378 .18773 .61559 .1899 .62484 .18753 Metetra
.593 .789 .974 r
.60116 .18353 .59193 .18567 .60378 .18773 .61302 .1855 Metetra
.589 .792 .977 r
.58926 .18163 .58005 .18375 .59193 .18567 .60116 .18353 Metetra
.588 .795 .978 r
.57734 .17982 .56813 .18194 .58005 .18375 .58926 .18163 Metetra
.589 .797 .979 r
.56537 .17806 .55618 .18023 .56813 .18194 .57734 .17982 Metetra
.591 .798 .978 r
.55338 .17635 .54419 .17857 .55618 .18023 .56537 .17806 Metetra
.594 .799 .978 r
.54134 .17465 .53217 .17693 .54419 .17857 .55338 .17635 Metetra
.597 .798 .977 r
.52927 .17296 .52011 .17528 .53217 .17693 .54134 .17465 Metetra
.598 .798 .976 r
.51717 .17125 .50801 .1736 .52011 .17528 .52927 .17296 Metetra
.599 .797 .975 r
.50502 .16952 .49588 .17189 .50801 .1736 .51717 .17125 Metetra
.599 .797 .975 r
.49285 .16778 .48371 .17015 .49588 .17189 .50502 .16952 Metetra
.599 .796 .975 r
.48063 .16601 .47151 .16839 .48371 .17015 .49285 .16778 Metetra
.599 .796 .975 r
.46838 .16423 .45927 .16661 .47151 .16839 .48063 .16601 Metetra
.598 .796 .975 r
.45609 .16245 .44699 .16482 .45927 .16661 .46838 .16423 Metetra
.598 .796 .975 r
.44377 .16065 .43468 .16303 .44699 .16482 .45609 .16245 Metetra
.598 .796 .975 r
.43141 .15886 .42233 .16124 .43468 .16303 .44377 .16065 Metetra
.598 .796 .976 r
.41901 .15706 .40994 .15945 .42233 .16124 .43141 .15886 Metetra
.598 .796 .976 r
.40657 .15526 .39751 .15765 .40994 .15945 .41901 .15706 Metetra
.598 .797 .976 r
.3941 .15345 .38505 .15585 .39751 .15765 .40657 .15526 Metetra
.598 .797 .976 r
.38158 .15164 .37255 .15405 .38505 .15585 .3941 .15345 Metetra
.598 .797 .976 r
.36904 .14982 .36001 .15224 .37255 .15405 .38158 .15164 Metetra
.598 .797 .975 r
.35645 .148 .34743 .15043 .36001 .15224 .36904 .14982 Metetra
.598 .796 .975 r
.34382 .14617 .33482 .14861 .34743 .15043 .35645 .148 Metetra
.568 .808 .989 r
.9211 .38926 .91156 .39108 .92258 .39245 .93213 .39063 Metetra
.568 .808 .989 r
.91004 .38789 .90051 .38972 .91156 .39108 .9211 .38926 Metetra
.772 .597 .661 r
.89895 .38651 .88864 .32213 .90051 .38972 .91004 .38789 Metetra
.92 .704 .593 r
.88718 .33019 .87667 .24524 .88864 .32213 .89895 .38651 Metetra
.987 .723 .355 r
.87534 .26724 .86509 .20133 .87667 .24524 .88718 .33019 Metetra
.952 .697 .514 r
.86377 .22855 .85374 .17827 .86509 .20133 .87534 .26724 Metetra
.892 .647 .561 r
.85237 .2058 .84251 .16813 .85374 .17827 .86377 .22855 Metetra
.845 .602 .571 r
.84107 .19338 .83133 .16571 .84251 .16813 .85237 .2058 Metetra
.807 .559 .563 r
.8298 .18749 .82016 .16765 .83133 .16571 .84107 .19338 Metetra
.774 .512 .539 r
.81855 .18557 .80898 .17182 .82016 .16765 .8298 .18749 Metetra
.739 .454 .496 r
.80728 .18594 .79777 .17685 .80898 .17182 .81855 .18557 Metetra
.696 .371 .422 r
.79599 .18747 .78653 .18194 .79777 .17685 .80728 .18594 Metetra
.625 .236 .286 r
.78467 .18945 .77524 .1866 .78653 .18194 .79599 .18747 Metetra
.476 0 .023 r
.77332 .19143 .76391 .19057 .77524 .1866 .78467 .18945 Metetra
.154 0 0 r
.76193 .19314 .75254 .19373 .76391 .19057 .77332 .19143 Metetra
.258 .767 .846 r
.7505 .19443 .74112 .19605 .75254 .19373 .76193 .19314 Metetra
.492 .866 .981 r
.73903 .19522 .72966 .19754 .74112 .19605 .7505 .19443 Metetra
.584 .863 .995 r
.72752 .19551 .71817 .19828 .72966 .19754 .73903 .19522 Metetra
.621 .842 .984 r
.71597 .19531 .70663 .19833 .71817 .19828 .72752 .19551 Metetra
.634 .823 .973 r
.70439 .19466 .69505 .19778 .70663 .19833 .71597 .19531 Metetra
.638 .809 .965 r
.69276 .19363 .68344 .19673 .69505 .19778 .70439 .19466 Metetra
.636 .799 .961 r
.68111 .19228 .67179 .19529 .68344 .19673 .69276 .19363 Metetra
.631 .792 .96 r
.66941 .19068 .66011 .19356 .67179 .19529 .68111 .19228 Metetra
.624 .789 .961 r
.65768 .18892 .64838 .19163 .66011 .19356 .66941 .19068 Metetra
.616 .787 .964 r
.64592 .18704 .63663 .1896 .64838 .19163 .65768 .18892 Metetra
.609 .788 .967 r
.63412 .18512 .62484 .18753 .63663 .1896 .64592 .18704 Metetra
.602 .789 .97 r
.62229 .18319 .61302 .1855 .62484 .18753 .63412 .18512 Metetra
.597 .791 .973 r
.61042 .18129 .60116 .18353 .61302 .1855 .62229 .18319 Metetra
.594 .793 .975 r
.59852 .17943 .58926 .18163 .60116 .18353 .61042 .18129 Metetra
.592 .795 .977 r
.58658 .17761 .57734 .17982 .58926 .18163 .59852 .17943 Metetra
.592 .796 .977 r
.57461 .17584 .56537 .17806 .57734 .17982 .58658 .17761 Metetra
.593 .797 .977 r
.5626 .17409 .55338 .17635 .56537 .17806 .57461 .17584 Metetra
.595 .798 .977 r
.55056 .17236 .54134 .17465 .55338 .17635 .5626 .17409 Metetra
.596 .798 .976 r
.53848 .17064 .52927 .17296 .54134 .17465 .55056 .17236 Metetra
.598 .797 .976 r
.52636 .16891 .51717 .17125 .52927 .17296 .53848 .17064 Metetra
.598 .797 .976 r
.51421 .16716 .50502 .16952 .51717 .17125 .52636 .16891 Metetra
.599 .797 .975 r
.50202 .1654 .49285 .16778 .50502 .16952 .51421 .16716 Metetra
.598 .797 .975 r
.48979 .16363 .48063 .16601 .49285 .16778 .50202 .1654 Metetra
.598 .796 .975 r
.47753 .16185 .46838 .16423 .48063 .16601 .48979 .16363 Metetra
.598 .796 .975 r
.46523 .16006 .45609 .16245 .46838 .16423 .47753 .16185 Metetra
.598 .796 .975 r
.4529 .15827 .44377 .16065 .45609 .16245 .46523 .16006 Metetra
.598 .796 .975 r
.44052 .15647 .43141 .15886 .44377 .16065 .4529 .15827 Metetra
.598 .796 .976 r
.42811 .15466 .41901 .15706 .43141 .15886 .44052 .15647 Metetra
.598 .796 .976 r
.41567 .15285 .40657 .15526 .41901 .15706 .42811 .15466 Metetra
.598 .796 .976 r
.40318 .15104 .3941 .15345 .40657 .15526 .41567 .15285 Metetra
.598 .797 .976 r
.39066 .14922 .38158 .15164 .3941 .15345 .40318 .15104 Metetra
.598 .797 .976 r
.3781 .1474 .36904 .14982 .38158 .15164 .39066 .14922 Metetra
.598 .797 .976 r
.3655 .14557 .35645 .148 .36904 .14982 .3781 .1474 Metetra
.598 .796 .975 r
.35286 .14373 .34382 .14617 .35645 .148 .3655 .14557 Metetra
.568 .808 .989 r
.93068 .38743 .9211 .38926 .93213 .39063 .94171 .38881 Metetra
.568 .808 .989 r
.91961 .38605 .91004 .38789 .9211 .38926 .93068 .38743 Metetra
.568 .808 .989 r
.90852 .38467 .89895 .38651 .91004 .38789 .91961 .38605 Metetra
.781 .596 .649 r
.89735 .38025 .88718 .33019 .89895 .38651 .90852 .38467 Metetra
.146 .174 .633 r
.88538 .30993 .87534 .26724 .88718 .33019 .89735 .38025 Metetra
0 0 .463 r
.87369 .26355 .86377 .22855 .87534 .26724 .88538 .30993 Metetra
.853 .503 0 r
.86219 .23361 .85237 .2058 .86377 .22855 .87369 .26355 Metetra
.938 .574 .287 r
.85079 .21485 .84107 .19338 .85237 .2058 .86219 .23361 Metetra
.884 .532 .376 r
.83945 .20357 .8298 .18749 .84107 .19338 .85079 .21485 Metetra
.826 .469 .381 r
.82814 .1972 .81855 .18557 .8298 .18749 .83945 .20357 Metetra
.768 .389 .336 r
.81683 .19396 .80728 .18594 .81855 .18557 .82814 .1972 Metetra
.698 .274 .237 r
.8055 .19262 .79599 .18747 .80728 .18594 .81683 .19396 Metetra
.585 .093 .053 r
.79415 .19235 .78467 .18945 .79599 .18747 .8055 .19262 Metetra
.373 0 0 r
.78277 .19259 .77332 .19143 .78467 .18945 .79415 .19235 Metetra
.034 0 0 r
.77136 .19299 .76193 .19314 .77332 .19143 .78277 .19259 Metetra
.286 .774 .906 r
.75991 .19331 .7505 .19443 .76193 .19314 .77136 .19299 Metetra
.468 .843 .988 r
.74843 .19341 .73903 .19522 .7505 .19443 .75991 .19331 Metetra
.554 .848 .998 r
.73691 .19322 .72752 .19551 .73903 .19522 .74843 .19341 Metetra
.594 .837 .99 r
.72535 .19273 .71597 .19531 .72752 .19551 .73691 .19322 Metetra
.613 .823 .981 r
.71376 .19192 .70439 .19466 .71597 .19531 .72535 .19273 Metetra
.621 .812 .974 r
.70213 .19083 .69276 .19363 .70439 .19466 .71376 .19192 Metetra
.623 .803 .969 r
.69046 .1895 .68111 .19228 .69276 .19363 .70213 .19083 Metetra
.621 .797 .967 r
.67876 .18797 .66941 .19068 .68111 .19228 .69046 .1895 Metetra
.618 .794 .966 r
.66702 .18629 .65768 .18892 .66941 .19068 .67876 .18797 Metetra
.613 .792 .967 r
.65525 .18452 .64592 .18704 .65768 .18892 .66702 .18629 Metetra
.608 .791 .969 r
.64344 .18269 .63412 .18512 .64592 .18704 .65525 .18452 Metetra
.603 .792 .971 r
.6316 .18084 .62229 .18319 .63412 .18512 .64344 .18269 Metetra
.599 .792 .973 r
.61972 .17899 .61042 .18129 .62229 .18319 .6316 .18084 Metetra
.597 .794 .975 r
.60781 .17716 .59852 .17943 .61042 .18129 .61972 .17899 Metetra
.595 .795 .976 r
.59586 .17535 .58658 .17761 .59852 .17943 .60781 .17716 Metetra
.595 .796 .976 r
.58388 .17357 .57461 .17584 .58658 .17761 .59586 .17535 Metetra
.595 .797 .977 r
.57186 .17181 .5626 .17409 .57461 .17584 .58388 .17357 Metetra
.596 .797 .976 r
.55981 .17006 .55056 .17236 .5626 .17409 .57186 .17181 Metetra
.597 .797 .976 r
.54772 .16831 .53848 .17064 .55056 .17236 .55981 .17006 Metetra
.597 .797 .976 r
.53559 .16656 .52636 .16891 .53848 .17064 .54772 .16831 Metetra
.598 .797 .976 r
.52343 .1648 .51421 .16716 .52636 .16891 .53559 .16656 Metetra
.598 .797 .975 r
.51123 .16303 .50202 .1654 .51421 .16716 .52343 .1648 Metetra
.598 .797 .975 r
.49899 .16125 .48979 .16363 .50202 .1654 .51123 .16303 Metetra
.598 .796 .975 r
.48672 .15947 .47753 .16185 .48979 .16363 .49899 .16125 Metetra
.598 .796 .975 r
.47441 .15767 .46523 .16006 .47753 .16185 .48672 .15947 Metetra
.598 .796 .975 r
.46206 .15587 .4529 .15827 .46523 .16006 .47441 .15767 Metetra
.598 .796 .975 r
.44968 .15406 .44052 .15647 .4529 .15827 .46206 .15587 Metetra
.598 .796 .975 r
.43726 .15225 .42811 .15466 .44052 .15647 .44968 .15406 Metetra
.598 .796 .976 r
.4248 .15044 .41567 .15285 .42811 .15466 .43726 .15225 Metetra
.598 .796 .976 r
.4123 .14862 .40318 .15104 .41567 .15285 .4248 .15044 Metetra
.598 .796 .976 r
.39977 .14679 .39066 .14922 .40318 .15104 .4123 .14862 Metetra
.598 .796 .976 r
.38719 .14496 .3781 .1474 .39066 .14922 .39977 .14679 Metetra
.598 .796 .976 r
.37458 .14312 .3655 .14557 .3781 .1474 .38719 .14496 Metetra
.598 .796 .975 r
.36193 .14128 .35286 .14373 .3655 .14557 .37458 .14312 Metetra
.568 .808 .989 r
.94029 .3856 .93068 .38743 .94171 .38881 .95134 .38697 Metetra
.568 .808 .989 r
.92922 .38421 .91961 .38605 .93068 .38743 .94029 .3856 Metetra
.568 .808 .989 r
.91812 .38282 .90852 .38467 .91961 .38605 .92922 .38421 Metetra
0 .508 .709 r
.90698 .38143 .89735 .38025 .90852 .38467 .91812 .38282 Metetra
.444 .404 .719 r
.8952 .3294 .88538 .30993 .89735 .38025 .90698 .38143 Metetra
.417 .391 .723 r
.88348 .2826 .87369 .26355 .88538 .30993 .8952 .3294 Metetra
.274 .328 .737 r
.87193 .25059 .86219 .23361 .87369 .26355 .88348 .2826 Metetra
0 .211 .724 r
.86048 .22908 .85079 .21485 .86219 .23361 .87193 .25059 Metetra
0 .02 .564 r
.8491 .21491 .83945 .20357 .85079 .21485 .86048 .22908 Metetra
.638 .096 0 r
.83774 .20582 .82814 .1972 .83945 .20357 .8491 .21491 Metetra
.631 .068 0 r
.82639 .20016 .81683 .19396 .82814 .1972 .83774 .20582 Metetra
.527 0 0 r
.81504 .19676 .8055 .19262 .81683 .19396 .82639 .20016 Metetra
.349 0 0 r
.80366 .19477 .79415 .19235 .8055 .19262 .81504 .19676 Metetra
.1 0 0 r
.79226 .19363 .78277 .19259 .79415 .19235 .80366 .19477 Metetra
.16 .688 .866 r
.78083 .19293 .77136 .19299 .78277 .19259 .79226 .19363 Metetra
.355 .791 .964 r
.76937 .19241 .75991 .19331 .77136 .19299 .78083 .19293 Metetra
.474 .828 .996 r
.75787 .19188 .74843 .19341 .75991 .19331 .76937 .19241 Metetra
.541 .834 .999 r
.74634 .19125 .73691 .19322 .74843 .19341 .75787 .19188 Metetra
.578 .828 .992 r
.73477 .19044 .72535 .19273 .73691 .19322 .74634 .19125 Metetra
.598 .82 .985 r
.72316 .18945 .71376 .19192 .72535 .19273 .73477 .19044 Metetra
.608 .812 .978 r
.71152 .18826 .70213 .19083 .71376 .19192 .72316 .18945 Metetra
.612 .805 .974 r
.69985 .1869 .69046 .1895 .70213 .19083 .71152 .18826 Metetra
.613 .8 .971 r
.68814 .18539 .67876 .18797 .69046 .1895 .69985 .1869 Metetra
.612 .797 .97 r
.67639 .18376 .66702 .18629 .67876 .18797 .68814 .18539 Metetra
.609 .795 .97 r
.66461 .18204 .65525 .18452 .66702 .18629 .67639 .18376 Metetra
.606 .794 .971 r
.65279 .18027 .64344 .18269 .65525 .18452 .66461 .18204 Metetra
.603 .793 .972 r
.64094 .17847 .6316 .18084 .64344 .18269 .65279 .18027 Metetra
.6 .794 .973 r
.62906 .17666 .61972 .17899 .6316 .18084 .64094 .17847 Metetra
.598 .794 .974 r
.61713 .17485 .60781 .17716 .61972 .17899 .62906 .17666 Metetra
.597 .795 .975 r
.60518 .17306 .59586 .17535 .60781 .17716 .61713 .17485 Metetra
.596 .796 .976 r
.59318 .17127 .58388 .17357 .59586 .17535 .60518 .17306 Metetra
.596 .796 .976 r
.58116 .1695 .57186 .17181 .58388 .17357 .59318 .17127 Metetra
.596 .797 .976 r
.56909 .16773 .55981 .17006 .57186 .17181 .58116 .1695 Metetra
.597 .797 .976 r
.55699 .16597 .54772 .16831 .55981 .17006 .56909 .16773 Metetra
.597 .797 .976 r
.54485 .1642 .53559 .16656 .54772 .16831 .55699 .16597 Metetra
.598 .797 .976 r
.53268 .16243 .52343 .1648 .53559 .16656 .54485 .1642 Metetra
.598 .797 .976 r
.52047 .16065 .51123 .16303 .52343 .1648 .53268 .16243 Metetra
.598 .797 .975 r
.50823 .15887 .49899 .16125 .51123 .16303 .52047 .16065 Metetra
.598 .797 .975 r
.49594 .15707 .48672 .15947 .49899 .16125 .50823 .15887 Metetra
.598 .796 .975 r
.48362 .15527 .47441 .15767 .48672 .15947 .49594 .15707 Metetra
.598 .796 .975 r
.47126 .15347 .46206 .15587 .47441 .15767 .48362 .15527 Metetra
.598 .796 .975 r
.45887 .15165 .44968 .15406 .46206 .15587 .47126 .15347 Metetra
.598 .796 .975 r
.44644 .14983 .43726 .15225 .44968 .15406 .45887 .15165 Metetra
.598 .796 .975 r
.43397 .14801 .4248 .15044 .43726 .15225 .44644 .14983 Metetra
.598 .796 .976 r
.42146 .14618 .4123 .14862 .4248 .15044 .43397 .14801 Metetra
.598 .796 .976 r
.40891 .14435 .39977 .14679 .4123 .14862 .42146 .14618 Metetra
.598 .796 .976 r
.39633 .14251 .38719 .14496 .39977 .14679 .40891 .14435 Metetra
.598 .796 .976 r
.3837 .14066 .37458 .14312 .38719 .14496 .39633 .14251 Metetra
.598 .796 .975 r
.37104 .13881 .36193 .14128 .37458 .14312 .3837 .14066 Metetra
.568 .808 .989 r
.94995 .38375 .94029 .3856 .95134 .38697 .96099 .38514 Metetra
.568 .808 .989 r
.93887 .38236 .92922 .38421 .94029 .3856 .94995 .38375 Metetra
.568 .808 .989 r
.92775 .38097 .91812 .38282 .92922 .38421 .93887 .38236 Metetra
.568 .808 .989 r
.91661 .37957 .90698 .38143 .91812 .38282 .92775 .38097 Metetra
.599 .508 .722 r
.90482 .32864 .8952 .3294 .90698 .38143 .91661 .37957 Metetra
.567 .494 .733 r
.89314 .28715 .88348 .2826 .8952 .3294 .90482 .32864 Metetra
.523 .486 .757 r
.8816 .25729 .87193 .25059 .88348 .2826 .89314 .28715 Metetra
.461 .482 .794 r
.87014 .23611 .86048 .22908 .87193 .25059 .8816 .25729 Metetra
.371 .486 .847 r
.85873 .2213 .8491 .21491 .86048 .22908 .87014 .23611 Metetra
.246 .499 .898 r
.84735 .21111 .83774 .20582 .8491 .21491 .85873 .2213 Metetra
.123 .518 .913 r
.83598 .20418 .82639 .20016 .83774 .20582 .84735 .21111 Metetra
.076 .554 .899 r
.8246 .19952 .81504 .19676 .82639 .20016 .83598 .20418 Metetra
.122 .618 .907 r
.8132 .19639 .80366 .19477 .81504 .19676 .8246 .19952 Metetra
.223 .695 .94 r
.80178 .19425 .79226 .19363 .80366 .19477 .8132 .19639 Metetra
.334 .759 .974 r
.79034 .19272 .78083 .19293 .79226 .19363 .80178 .19425 Metetra
.428 .799 .994 r
.77886 .19152 .76937 .19241 .78083 .19293 .79034 .19272 Metetra
.495 .817 1 r
.76735 .19047 .75787 .19188 .76937 .19241 .77886 .19152 Metetra
.541 .821 .998 r
.7558 .18943 .74634 .19125 .75787 .19188 .76735 .19047 Metetra
.57 .82 .992 r
.74422 .18835 .73477 .19044 .74634 .19125 .7558 .18943 Metetra
.588 .815 .986 r
.73261 .18716 .72316 .18945 .73477 .19044 .74422 .18835 Metetra
.599 .81 .981 r
.72096 .18586 .71152 .18826 .72316 .18945 .73261 .18716 Metetra
.604 .805 .977 r
.70928 .18443 .69985 .1869 .71152 .18826 .72096 .18586 Metetra
.607 .801 .974 r
.69756 .18291 .68814 .18539 .69985 .1869 .70928 .18443 Metetra
.607 .798 .973 r
.6858 .18129 .67639 .18376 .68814 .18539 .69756 .18291 Metetra
.606 .796 .972 r
.67401 .1796 .66461 .18204 .67639 .18376 .6858 .18129 Metetra
.604 .795 .973 r
.66219 .17786 .65279 .18027 .66461 .18204 .67401 .1796 Metetra
.602 .795 .973 r
.65032 .17609 .64094 .17847 .65279 .18027 .66219 .17786 Metetra
.6 .795 .974 r
.63843 .17431 .62906 .17666 .64094 .17847 .65032 .17609 Metetra
.599 .795 .974 r
.6265 .17252 .61713 .17485 .62906 .17666 .63843 .17431 Metetra
.598 .795 .975 r
.61453 .17073 .60518 .17306 .61713 .17485 .6265 .17252 Metetra
.597 .796 .975 r
.60253 .16895 .59318 .17127 .60518 .17306 .61453 .17073 Metetra
.597 .796 .976 r
.59049 .16717 .58116 .1695 .59318 .17127 .60253 .16895 Metetra
.597 .796 .976 r
.57842 .16539 .56909 .16773 .58116 .1695 .59049 .16717 Metetra
.597 .797 .976 r
.5663 .16361 .55699 .16597 .56909 .16773 .57842 .16539 Metetra
.597 .797 .976 r
.55416 .16184 .54485 .1642 .55699 .16597 .5663 .16361 Metetra
.598 .797 .976 r
.54197 .16005 .53268 .16243 .54485 .1642 .55416 .16184 Metetra
.598 .797 .976 r
.52975 .15827 .52047 .16065 .53268 .16243 .54197 .16005 Metetra
.598 .797 .976 r
.5175 .15647 .50823 .15887 .52047 .16065 .52975 .15827 Metetra
.598 .797 .975 r
.5052 .15467 .49594 .15707 .50823 .15887 .5175 .15647 Metetra
.598 .796 .975 r
.49287 .15286 .48362 .15527 .49594 .15707 .5052 .15467 Metetra
.598 .796 .975 r
.4805 .15105 .47126 .15347 .48362 .15527 .49287 .15286 Metetra
.598 .796 .975 r
.4681 .14923 .45887 .15165 .47126 .15347 .4805 .15105 Metetra
.598 .796 .975 r
.45565 .14741 .44644 .14983 .45887 .15165 .4681 .14923 Metetra
.598 .796 .975 r
.44317 .14557 .43397 .14801 .44644 .14983 .45565 .14741 Metetra
.598 .796 .975 r
.43065 .14374 .42146 .14618 .43397 .14801 .44317 .14557 Metetra
.598 .796 .976 r
.41809 .1419 .40891 .14435 .42146 .14618 .43065 .14374 Metetra
.598 .796 .976 r
.4055 .14005 .39633 .14251 .40891 .14435 .41809 .1419 Metetra
.598 .796 .975 r
.39286 .1382 .3837 .14066 .39633 .14251 .4055 .14005 Metetra
.598 .796 .975 r
.38019 .13634 .37104 .13881 .3837 .14066 .39286 .1382 Metetra
.568 .808 .989 r
.95963 .3819 .94995 .38375 .96099 .38514 .97069 .38329 Metetra
.568 .808 .989 r
.94855 .38051 .93887 .38236 .94995 .38375 .95963 .3819 Metetra
.568 .808 .989 r
.93743 .37911 .92775 .38097 .93887 .38236 .94855 .38051 Metetra
.695 .585 .727 r
.92597 .35441 .91661 .37957 .92775 .38097 .93743 .37911 Metetra
.654 .548 .722 r
.91428 .31264 .90482 .32864 .91661 .37957 .92597 .35441 Metetra
.633 .544 .736 r
.90271 .28015 .89314 .28715 .90482 .32864 .91428 .31264 Metetra
.609 .544 .756 r
.89121 .25538 .8816 .25729 .89314 .28715 .90271 .28015 Metetra
.58 .551 .786 r
.87977 .23684 .87014 .23611 .8816 .25729 .89121 .25538 Metetra
.545 .567 .828 r
.86836 .22315 .85873 .2213 .87014 .23611 .87977 .23684 Metetra
.502 .596 .88 r
.85697 .21317 .84735 .21111 .85873 .2213 .86836 .22315 Metetra
.455 .635 .933 r
.84558 .20597 .83598 .20418 .84735 .21111 .85697 .21317 Metetra
.415 .677 .97 r
.83419 .20078 .8246 .19952 .83598 .20418 .84558 .20597 Metetra
.4 .715 .987 r
.82277 .19704 .8132 .19639 .8246 .19952 .83419 .20078 Metetra
.414 .749 .994 r
.81134 .19428 .80178 .19425 .8132 .19639 .82277 .19704 Metetra
.447 .776 .998 r
.79988 .19218 .79034 .19272 .80178 .19425 .81134 .19428 Metetra
.485 .796 .999 r
.78839 .19049 .77886 .19152 .79034 .19272 .79988 .19218 Metetra
.521 .807 .998 r
.77686 .18903 .76735 .19047 .77886 .19152 .78839 .19049 Metetra
.549 .811 .995 r
.76531 .18767 .7558 .18943 .76735 .19047 .77686 .18903 Metetra
.569 .812 .99 r
.75372 .18634 .74422 .18835 .7558 .18943 .76531 .18767 Metetra
.584 .81 .986 r
.7421 .18498 .73261 .18716 .74422 .18835 .75372 .18634 Metetra
.593 .807 .982 r
.73044 .18356 .72096 .18586 .73261 .18716 .7421 .18498 Metetra
.599 .804 .978 r
.71874 .18206 .70928 .18443 .72096 .18586 .73044 .18356 Metetra
.602 .801 .976 r
.70701 .1805 .69756 .18291 .70928 .18443 .71874 .18206 Metetra
.603 .799 .975 r
.69525 .17887 .6858 .18129 .69756 .18291 .70701 .1805 Metetra
.603 .797 .974 r
.68345 .17719 .67401 .1796 .6858 .18129 .69525 .17887 Metetra
.602 .796 .974 r
.67161 .17547 .66219 .17786 .67401 .1796 .68345 .17719 Metetra
.601 .796 .974 r
.65974 .17372 .65032 .17609 .66219 .17786 .67161 .17547 Metetra
.6 .796 .974 r
.64784 .17195 .63843 .17431 .65032 .17609 .65974 .17372 Metetra
.599 .796 .975 r
.6359 .17017 .6265 .17252 .63843 .17431 .64784 .17195 Metetra
.598 .796 .975 r
.62392 .16838 .61453 .17073 .6265 .17252 .6359 .17017 Metetra
.598 .796 .975 r
.61191 .1666 .60253 .16895 .61453 .17073 .62392 .16838 Metetra
.597 .796 .976 r
.59986 .16482 .59049 .16717 .60253 .16895 .61191 .1666 Metetra
.597 .796 .976 r
.58778 .16303 .57842 .16539 .59049 .16717 .59986 .16482 Metetra
.597 .796 .976 r
.57566 .16125 .5663 .16361 .57842 .16539 .58778 .16303 Metetra
.597 .797 .976 r
.5635 .15946 .55416 .16184 .5663 .16361 .57566 .16125 Metetra
.598 .797 .976 r
.5513 .15767 .54197 .16005 .55416 .16184 .5635 .15946 Metetra
.598 .797 .976 r
.53907 .15587 .52975 .15827 .54197 .16005 .5513 .15767 Metetra
.598 .797 .976 r
.52681 .15407 .5175 .15647 .52975 .15827 .53907 .15587 Metetra
.598 .797 .975 r
.5145 .15226 .5052 .15467 .5175 .15647 .52681 .15407 Metetra
.598 .797 .975 r
.50216 .15045 .49287 .15286 .5052 .15467 .5145 .15226 Metetra
.598 .796 .975 r
.48978 .14863 .4805 .15105 .49287 .15286 .50216 .15045 Metetra
.598 .796 .975 r
.47736 .1468 .4681 .14923 .4805 .15105 .48978 .14863 Metetra
.598 .796 .975 r
.46491 .14497 .45565 .14741 .4681 .14923 .47736 .1468 Metetra
.598 .796 .975 r
.45242 .14313 .44317 .14557 .45565 .14741 .46491 .14497 Metetra
.598 .796 .975 r
.43988 .14129 .43065 .14374 .44317 .14557 .45242 .14313 Metetra
.598 .796 .975 r
.42732 .13944 .41809 .1419 .43065 .14374 .43988 .14129 Metetra
.598 .796 .975 r
.41471 .13758 .4055 .14005 .41809 .1419 .42732 .13944 Metetra
.598 .796 .975 r
.40206 .13572 .39286 .1382 .4055 .14005 .41471 .13758 Metetra
.598 .796 .975 r
.38938 .13386 .38019 .13634 .39286 .1382 .40206 .13572 Metetra
.568 .808 .989 r
.96936 .38004 .95963 .3819 .97069 .38329 .98042 .38144 Metetra
.695 .606 .753 r
.95809 .36606 .94855 .38051 .95963 .3819 .96936 .38004 Metetra
.712 .591 .719 r
.94663 .33997 .93743 .37911 .94855 .38051 .95809 .36606 Metetra
.713 .59 .717 r
.93514 .3126 .92597 .35441 .93743 .37911 .94663 .33997 Metetra
.695 .584 .727 r
.92366 .28726 .91428 .31264 .92597 .35441 .93514 .3126 Metetra
.677 .581 .74 r
.9122 .26536 .90271 .28015 .91428 .31264 .92366 .28726 Metetra
.658 .582 .759 r
.90078 .24724 .89121 .25538 .90271 .28015 .9122 .26536 Metetra
.637 .59 .785 r
.88937 .2327 .87977 .23684 .89121 .25538 .90078 .24724 Metetra
.615 .606 .819 r
.87798 .22129 .86836 .22315 .87977 .23684 .88937 .2327 Metetra
.589 .63 .86 r
.8666 .21249 .85697 .21317 .86836 .22315 .87798 .22129 Metetra
.562 .663 .905 r
.8552 .20575 .84558 .20597 .85697 .21317 .8666 .21249 Metetra
.535 .697 .944 r
.8438 .20063 .83419 .20078 .84558 .20597 .8552 .20575 Metetra
.516 .73 .97 r
.83237 .1967 .82277 .19704 .83419 .20078 .8438 .20063 Metetra
.509 .756 .985 r
.82093 .19366 .81134 .19428 .82277 .19704 .83237 .1967 Metetra
.514 .776 .991 r
.80945 .19124 .79988 .19218 .81134 .19428 .82093 .19366 Metetra
.527 .79 .993 r
.79795 .18923 .78839 .19049 .79988 .19218 .80945 .19124 Metetra
.543 .799 .992 r
.78642 .18748 .77686 .18903 .78839 .19049 .79795 .18923 Metetra
.559 .804 .99 r
.77485 .18588 .76531 .18767 .77686 .18903 .78642 .18748 Metetra
.572 .805 .987 r
.76325 .18435 .75372 .18634 .76531 .18767 .77485 .18588 Metetra
.583 .805 .984 r
.75162 .18284 .7421 .18498 .75372 .18634 .76325 .18435 Metetra
.59 .804 .981 r
.73995 .18132 .73044 .18356 .7421 .18498 .75162 .18284 Metetra
.595 .802 .979 r
.72825 .17975 .71874 .18206 .73044 .18356 .73995 .18132 Metetra
.598 .8 .977 r
.71651 .17815 .70701 .1805 .71874 .18206 .72825 .17975 Metetra
.6 .799 .976 r
.70473 .1765 .69525 .17887 .70701 .1805 .71651 .17815 Metetra
.601 .798 .975 r
.69292 .17481 .68345 .17719 .69525 .17887 .70473 .1765 Metetra
.601 .797 .975 r
.68108 .17309 .67161 .17547 .68345 .17719 .69292 .17481 Metetra
.6 .796 .975 r
.6692 .17134 .65974 .17372 .67161 .17547 .68108 .17309 Metetra
.599 .796 .975 r
.65729 .16958 .64784 .17195 .65974 .17372 .6692 .17134 Metetra
.599 .796 .975 r
.64534 .16781 .6359 .17017 .64784 .17195 .65729 .16958 Metetra
.598 .796 .975 r
.63335 .16603 .62392 .16838 .6359 .17017 .64534 .16781 Metetra
.598 .796 .975 r
.62133 .16424 .61191 .1666 .62392 .16838 .63335 .16603 Metetra
.598 .796 .975 r
.60927 .16245 .59986 .16482 .61191 .1666 .62133 .16424 Metetra
.598 .796 .976 r
.59717 .16066 .58778 .16303 .59986 .16482 .60927 .16245 Metetra
.598 .796 .976 r
.58504 .15887 .57566 .16125 .58778 .16303 .59717 .16066 Metetra
.598 .796 .976 r
.57288 .15707 .5635 .15946 .57566 .16125 .58504 .15887 Metetra
.598 .797 .976 r
.56067 .15527 .5513 .15767 .5635 .15946 .57288 .15707 Metetra
.598 .797 .976 r
.54843 .15347 .53907 .15587 .5513 .15767 .56067 .15527 Metetra
.598 .797 .976 r
.53615 .15166 .52681 .15407 .53907 .15587 .54843 .15347 Metetra
.598 .797 .976 r
.52384 .14984 .5145 .15226 .52681 .15407 .53615 .15166 Metetra
.598 .797 .975 r
.51149 .14802 .50216 .15045 .5145 .15226 .52384 .14984 Metetra
.598 .796 .975 r
.49909 .14619 .48978 .14863 .50216 .15045 .51149 .14802 Metetra
.598 .796 .975 r
.48667 .14436 .47736 .1468 .48978 .14863 .49909 .14619 Metetra
.598 .796 .975 r
.4742 .14252 .46491 .14497 .47736 .1468 .48667 .14436 Metetra
.598 .796 .975 r
.4617 .14067 .45242 .14313 .46491 .14497 .4742 .14252 Metetra
.598 .796 .975 r
.44915 .13882 .43988 .14129 .45242 .14313 .4617 .14067 Metetra
.598 .796 .975 r
.43657 .13697 .42732 .13944 .43988 .14129 .44915 .13882 Metetra
.598 .796 .975 r
.42395 .1351 .41471 .13758 .42732 .13944 .43657 .13697 Metetra
.598 .796 .975 r
.4113 .13324 .40206 .13572 .41471 .13758 .42395 .1351 Metetra
.598 .796 .975 r
.3986 .13136 .38938 .13386 .40206 .13572 .4113 .13324 Metetra
.859 .704 .688 r
.97717 .24487 .96936 .38004 .98042 .38144 .98729 .18534 Metetra
.814 .669 .703 r
.96648 .26928 .95809 .36606 .96936 .38004 .97717 .24487 Metetra
.782 .647 .713 r
.95547 .27394 .94663 .33997 .95809 .36606 .96648 .26928 Metetra
.756 .632 .723 r
.94428 .26835 .93514 .3126 .94663 .33997 .95547 .27394 Metetra
.734 .622 .734 r
.933 .25817 .92366 .28726 .93514 .3126 .94428 .26835 Metetra
.714 .618 .749 r
.92168 .24661 .9122 .26536 .92366 .28726 .933 .25817 Metetra
.694 .618 .767 r
.91033 .23541 .90078 .24724 .9122 .26536 .92168 .24661 Metetra
.675 .625 .791 r
.89898 .22538 .88937 .2327 .90078 .24724 .91033 .23541 Metetra
.655 .638 .82 r
.88761 .21681 .87798 .22129 .88937 .2327 .89898 .22538 Metetra
.634 .657 .855 r
.87623 .2097 .8666 .21249 .87798 .22129 .88761 .21681 Metetra
.612 .682 .892 r
.86484 .20392 .8552 .20575 .8666 .21249 .87623 .2097 Metetra
.591 .709 .926 r
.85343 .19926 .8438 .20063 .8552 .20575 .86484 .20392 Metetra
.572 .735 .952 r
.842 .19549 .83237 .1967 .8438 .20063 .85343 .19926 Metetra
.559 .756 .97 r
.83055 .19242 .82093 .19366 .83237 .1967 .842 .19549 Metetra
.555 .773 .98 r
.81906 .18988 .80945 .19124 .82093 .19366 .83055 .19242 Metetra
.556 .785 .984 r
.80755 .1877 .79795 .18923 .80945 .19124 .81906 .18988 Metetra
.562 .793 .986 r
.79601 .18578 .78642 .18748 .79795 .18923 .80755 .1877 Metetra
.57 .798 .985 r
.78443 .18401 .77485 .18588 .78642 .18748 .79601 .18578 Metetra
.578 .801 .984 r
.77282 .18234 .76325 .18435 .77485 .18588 .78443 .18401 Metetra
.584 .801 .982 r
.76118 .18072 .75162 .18284 .76325 .18435 .77282 .18234 Metetra
.59 .801 .98 r
.7495 .1791 .73995 .18132 .75162 .18284 .76118 .18072 Metetra
.594 .801 .979 r
.73779 .17747 .72825 .17975 .73995 .18132 .7495 .1791 Metetra
.597 .8 .977 r
.72604 .17583 .71651 .17815 .72825 .17975 .73779 .17747 Metetra
.598 .799 .976 r
.71426 .17415 .70473 .1765 .71651 .17815 .72604 .17583 Metetra
.599 .798 .976 r
.70244 .17245 .69292 .17481 .70473 .1765 .71426 .17415 Metetra
.599 .797 .975 r
.69058 .17072 .68108 .17309 .69292 .17481 .70244 .17245 Metetra
.599 .797 .975 r
.6787 .16897 .6692 .17134 .68108 .17309 .69058 .17072 Metetra
.599 .796 .975 r
.66677 .16721 .65729 .16958 .6692 .17134 .6787 .16897 Metetra
.599 .796 .975 r
.65481 .16544 .64534 .16781 .65729 .16958 .66677 .16721 Metetra
.598 .796 .975 r
.64282 .16366 .63335 .16603 .64534 .16781 .65481 .16544 Metetra
.598 .796 .975 r
.63078 .16187 .62133 .16424 .63335 .16603 .64282 .16366 Metetra
.598 .796 .975 r
.61872 .16008 .60927 .16245 .62133 .16424 .63078 .16187 Metetra
.598 .796 .975 r
.60661 .15828 .59717 .16066 .60927 .16245 .61872 .16008 Metetra
.598 .796 .976 r
.59447 .15648 .58504 .15887 .59717 .16066 .60661 .15828 Metetra
.598 .796 .976 r
.58229 .15468 .57288 .15707 .58504 .15887 .59447 .15648 Metetra
.598 .796 .976 r
.57008 .15287 .56067 .15527 .57288 .15707 .58229 .15468 Metetra
.598 .797 .976 r
.55783 .15106 .54843 .15347 .56067 .15527 .57008 .15287 Metetra
.598 .797 .976 r
.54554 .14924 .53615 .15166 .54843 .15347 .55783 .15106 Metetra
.598 .797 .976 r
.53321 .14741 .52384 .14984 .53615 .15166 .54554 .14924 Metetra
.598 .797 .976 r
.52085 .14558 .51149 .14802 .52384 .14984 .53321 .14741 Metetra
.598 .796 .975 r
.50845 .14375 .49909 .14619 .51149 .14802 .52085 .14558 Metetra
.598 .796 .975 r
.49601 .14191 .48667 .14436 .49909 .14619 .50845 .14375 Metetra
.598 .796 .975 r
.48353 .14006 .4742 .14252 .48667 .14436 .49601 .14191 Metetra
.598 .796 .975 r
.47102 .13821 .4617 .14067 .4742 .14252 .48353 .14006 Metetra
.598 .796 .975 r
.45846 .13635 .44915 .13882 .4617 .14067 .47102 .13821 Metetra
.598 .796 .975 r
.44587 .13448 .43657 .13697 .44915 .13882 .45846 .13635 Metetra
.598 .796 .975 r
.43324 .13261 .42395 .1351 .43657 .13697 .44587 .13448 Metetra
.598 .796 .975 r
.42057 .13074 .4113 .13324 .42395 .1351 .43324 .13261 Metetra
.598 .796 .975 r
.40786 .12886 .3986 .13136 .4113 .13324 .42057 .13074 Metetra
.856 .705 .694 r
.98539 .13976 .97717 .24487 .98729 .18534 .99568 .09473 Metetra
.934 .787 .675 r
.97514 .19143 .96648 .26928 .97717 .24487 .98539 .13976 Metetra
.864 .731 .714 r
.96445 .21753 .95547 .27394 .96648 .26928 .97514 .19143 Metetra
.816 .697 .733 r
.9535 .22823 .94428 .26835 .95547 .27394 .96445 .21753 Metetra
.78 .675 .748 r
.94238 .23011 .933 .25817 .94428 .26835 .9535 .22823 Metetra
.752 .664 .764 r
.93117 .22728 .92168 .24661 .933 .25817 .94238 .23011 Metetra
.727 .659 .782 r
.91991 .22225 .91033 .23541 .92168 .24661 .93117 .22728 Metetra
.705 .662 .805 r
.9086 .21649 .89898 .22538 .91033 .23541 .91991 .22225 Metetra
.684 .671 .831 r
.89726 .21079 .88761 .21681 .89898 .22538 .9086 .21649 Metetra
.663 .685 .861 r
.88589 .20556 .87623 .2097 .88761 .21681 .89726 .21079 Metetra
.642 .703 .891 r
.87451 .20095 .86484 .20392 .87623 .2097 .88589 .20556 Metetra
.621 .723 .919 r
.8631 .19697 .85343 .19926 .86484 .20392 .87451 .20095 Metetra
.603 .742 .943 r
.85166 .19357 .842 .19549 .85343 .19926 .8631 .19697 Metetra
.589 .759 .96 r
.8402 .19066 .83055 .19242 .842 .19549 .85166 .19357 Metetra
.58 .772 .97 r
.82871 .18813 .81906 .18988 .83055 .19242 .8402 .19066 Metetra
.576 .782 .977 r
.81719 .18591 .80755 .1877 .81906 .18988 .82871 .18813 Metetra
.576 .79 .98 r
.80563 .1839 .79601 .18578 .80755 .1877 .81719 .18591 Metetra
.579 .794 .981 r
.79405 .18204 .78443 .18401 .79601 .18578 .80563 .1839 Metetra
.583 .797 .981 r
.78243 .18028 .77282 .18234 .78443 .18401 .79405 .18204 Metetra
.587 .799 .98 r
.77077 .17857 .76118 .18072 .77282 .18234 .78243 .18028 Metetra
.591 .799 .979 r
.75909 .17689 .7495 .1791 .76118 .18072 .77077 .17857 Metetra
.594 .799 .978 r
.74737 .17521 .73779 .17747 .7495 .1791 .75909 .17689 Metetra
.596 .799 .977 r
.73561 .17352 .72604 .17583 .73779 .17747 .74737 .17521 Metetra
.597 .798 .976 r
.72382 .17181 .71426 .17415 .72604 .17583 .73561 .17352 Metetra
.598 .798 .976 r
.71199 .17009 .70244 .17245 .71426 .17415 .72382 .17181 Metetra
.599 .797 .976 r
.70013 .16836 .69058 .17072 .70244 .17245 .71199 .17009 Metetra
.599 .797 .975 r
.68823 .1666 .6787 .16897 .69058 .17072 .70013 .16836 Metetra
.599 .797 .975 r
.67629 .16484 .66677 .16721 .6787 .16897 .68823 .1666 Metetra
.598 .796 .975 r
.66432 .16306 .65481 .16544 .66677 .16721 .67629 .16484 Metetra
.598 .796 .975 r
.65232 .16128 .64282 .16366 .65481 .16544 .66432 .16306 Metetra
.598 .796 .975 r
.64028 .15948 .63078 .16187 .64282 .16366 .65232 .16128 Metetra
.598 .796 .975 r
.6282 .15769 .61872 .16008 .63078 .16187 .64028 .15948 Metetra
.598 .796 .975 r
.61608 .15589 .60661 .15828 .61872 .16008 .6282 .15769 Metetra
.598 .796 .975 r
.60393 .15408 .59447 .15648 .60661 .15828 .61608 .15589 Metetra
.598 .796 .976 r
.59175 .15227 .58229 .15468 .59447 .15648 .60393 .15408 Metetra
.598 .796 .976 r
.57952 .15045 .57008 .15287 .58229 .15468 .59175 .15227 Metetra
.598 .796 .976 r
.56726 .14863 .55783 .15106 .57008 .15287 .57952 .15045 Metetra
.598 .797 .976 r
.55496 .14681 .54554 .14924 .55783 .15106 .56726 .14863 Metetra
.598 .797 .976 r
.54262 .14497 .53321 .14741 .54554 .14924 .55496 .14681 Metetra
.598 .796 .976 r
.53025 .14314 .52085 .14558 .53321 .14741 .54262 .14497 Metetra
.598 .796 .975 r
.51784 .14129 .50845 .14375 .52085 .14558 .53025 .14314 Metetra
.598 .796 .975 r
.50539 .13944 .49601 .14191 .50845 .14375 .51784 .14129 Metetra
.598 .796 .975 r
.4929 .13759 .48353 .14006 .49601 .14191 .50539 .13944 Metetra
.598 .796 .975 r
.48037 .13573 .47102 .13821 .48353 .14006 .4929 .13759 Metetra
.598 .796 .975 r
.46781 .13386 .45846 .13635 .47102 .13821 .48037 .13573 Metetra
.598 .796 .975 r
.4552 .13199 .44587 .13448 .45846 .13635 .46781 .13386 Metetra
.598 .796 .975 r
.44256 .13012 .43324 .13261 .44587 .13448 .4552 .13199 Metetra
.598 .796 .975 r
.42988 .12823 .42057 .13074 .43324 .13261 .44256 .13012 Metetra
.598 .796 .975 r
.41716 .12634 .40786 .12886 .42057 .13074 .42988 .12823 Metetra
0 g
.25 Mabswid
.41745 0 m
.41674 .30365 L
s
.41674 .30365 m
0 .40002 L
s
0 .40002 m
.00412 .1233 L
s
.00412 .1233 m
.41745 0 L
s
.41745 0 m
.41674 .30365 L
s
.41674 .30365 m
1 .37771 L
s
1 .37771 m
.99568 .09473 L
s
.99568 .09473 m
.41745 0 L
s
.41745 0 m
.99568 .09473 L
s
.47917 .01011 m
.47917 .01643 L
s
gsave
.47918 -0.00253 -70.1521 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.59988 .02989 m
.5999 .03621 L
s
gsave
.59984 .01724 -70.1858 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.71708 .04909 m
.71712 .05541 L
s
gsave
.71699 .03644 -66.034 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.83092 .06774 m
.83098 .07406 L
s
gsave
.83079 .0551 -66.0513 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.94154 .08586 m
.94162 .09219 L
s
gsave
.94136 .07322 -66.0681 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.50969 .01511 m
.50969 .0189 L
s
.53998 .02007 m
.53998 .02387 L
s
.57004 .025 m
.57005 .02879 L
s
.6295 .03474 m
.62952 .03853 L
s
.65891 .03956 m
.65893 .04335 L
s
.6881 .04434 m
.68812 .04813 L
s
.74585 .0538 m
.74588 .0576 L
s
.77441 .05848 m
.77444 .06228 L
s
.80276 .06313 m
.8028 .06692 L
s
.85887 .07232 m
.85891 .07611 L
s
.88662 .07687 m
.88666 .08066 L
s
.91417 .08138 m
.91422 .08517 L
s
.44843 .00507 m
.44843 .00887 L
s
.9687 .09031 m
.96876 .09411 L
s
.25 Mabswid
1 .37771 m
.55997 .46147 L
s
.9523 .38679 m
.95221 .38047 L
s
gsave
.95247 .39944 -70.0297 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.85952 .40445 m
.85945 .39813 L
s
gsave
.85966 .4171 -70.0554 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.77008 .42148 m
.77003 .41515 L
s
gsave
.77019 .43412 -65.9584 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.68381 .4379 m
.68377 .43158 L
s
gsave
.68388 .45054 -65.9714 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.60053 .45375 m
.60051 .44743 L
s
gsave
.60057 .4664 -65.984 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.92878 .39127 m
.92873 .38747 L
s
.90548 .3957 m
.90543 .39191 L
s
.8824 .4001 m
.88235 .3963 L
s
.83686 .40877 m
.83682 .40497 L
s
.8144 .41304 m
.81436 .40925 L
s
.79214 .41728 m
.79211 .41348 L
s
.74823 .42564 m
.7482 .42184 L
s
.72656 .42976 m
.72654 .42597 L
s
.70509 .43385 m
.70507 .43005 L
s
.66272 .44191 m
.6627 .43812 L
s
.64181 .44589 m
.64179 .4421 L
s
.62108 .44984 m
.62107 .44605 L
s
.97604 .38227 m
.97598 .37848 L
s
.58016 .45763 m
.58015 .45383 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{325.438, 150.125},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40007X0000hB000`400?l00000o`00003oo`3ooooY
0?ooo`00o`3ooooY0?ooo`00o`3ooooY0?ooo`00o`3ooooY0?ooo`00o`3ooooY0?ooo`00o`3ooooY
0?ooo`00o`3ooooY0?ooo`00f`3oool0103ooml0W4P00018W03Ooooo0?ooo`X0oooo003K0?ooo`04
0?oog`2LB00004RL0=ooool0oooo2P3oool00=/0oooo00@0oooO09a80000B9`0goooo`3oool:0?oo
o`00e`3oool00`2oM0000000000000050000000307Boo`3oool0oooo0?l0oooo1`3oool00<X0oooo
00<0oooO09a8000000001`0000001@18W=l0oooo0?ooo`3ogi`0M9cO0080oooo00@0oooO09a80000
B9`0goooo`3oool:0?ooo`00f03oool01`3OW4P0M;oo0?ooo`3ooml0W4P00018W03Oool0o`3oool:
0?ooo`00f03oool01`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool0o`3oool:0?ooo`00f@3oool0
1P3ogi`0B7Bo0?oog`2LB00004RL0=ooool0oooo2P3oool00=X0oooo00D0gia807Bog`2LB00004RL
0=ooo`0U0?ooo`030;md00000000000000@0000000<0B9cO0?ooo`3oool0fP3oool00=X0oooo00D0
oooO09a8B018B00004RL0=ooo`0V0?ooo`030=nLB0000000000000<0000000<004RL0=ooo`3oool0
fP3oool00=/0oooo00@0omnL04P00000B9`0gooo9`3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nL
M02Lgol0oooo0=X0oooo003L0?ooo`030=nLB000B9`0gooo02P0oooo00<0gia807Boo`3oool0gP3o
ool00?P0oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`030?ooo`030;mdB02Lgol0
oooo0=d0oooo003o0?ooo`T0oooo00<0oooO09a8M02oool0g@3oool00?l0oooo2P3oool00`3o_g@0
07Bo0?ooo`3L0?ooo`00o`3oool:0?ooo`030?oo_`1d04P0W=oo0200oooo00D0omnL04P000000000
001d0;ooo`2g0?ooo`00o`3oool:0?ooo`030?oo_`1d0000B9cO01l0oooo00L0oono07@0001d_ol0
oooo0?noM00007@0_ooo0;H0oooo003o0?ooo`D0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0
B00004RLg`0O0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool0]@3oool00;d0oooo
1@0000120?ooo`080?ooa01O51@0GlCo0?ooo`3oohl0=@000000=@2?oolN0?ooo`030?ooS`0e0000
55o400<0oooo00<0a5lD0000=@2?ool0]@3oool00;T0oooo100000000`3oool000000?ooo`020?oo
o`H00000?@3oool01P3ool@0Ga@0000000000000000e08oooal0oooo00<0olAO01@0000eSol00`3o
ool00`3oScD0000D05o4o`2e0?ooo`00]P3oool3000000D0oooo00<000000?ooo`3oool01`3oool6
000005`0oooo00<0olAO01@0000eSol00`3oool00`3oScD0000D05o4o`2e0?ooo`00/`3oool30000
00P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`020?ooo`H00000EP3oool00`3o
a5l0500003F?o`030?ooo`030?n?=@0001@0GlCo0;D0oooo002_0?ooo`@000002`3oool00`000000
oooo0?ooo`0C0?ooo`H00000D03oool00`3oa5l0500003F?o`030?ooo`030?n?=@0001@0GlCo0280
oooo00<0ScD0000000000000100000000`0DGl@0oooo0?ooo`290?ooo`00[03oool3000000l0oooo
00<000000?ooo`3oool0503oool00`000000oooo0?ooo`020?ooo`H00000BP3oool00`3oohl0=@00
01AOa0030?ooo`030<AO500003D0Sooo02<0oooo00<0a5lD0000000000000`0000000`0055l0a?oo
0?ooo`290?ooo`00Z03oool400000180oooo00<000000?ooo`3oool0503oool00`000000oooo0?oo
o`080?ooo`H00000A@3oool00`34Ga@001AO0<Coo`020?ooo`0308le0000=Hl0oooo02@0oooo00<0
a5lD01AOa03oool00P3oool00`34GcD0GlCo0?ooo`290?ooo`00Y@3oool3000001H0oooo00<00000
0?ooo`3oool0803oool00`000000oooo0?ooo`020?ooo`L00000?P3oool01`3oohl0=@0003F?o`3o
ool0ohle0000=@2?ool09P3oool00`34Ga@0=Hoo0?ooo`2=0?ooo`00XP3oool3000001T0oooo00<0
00000?ooo`3oool0:`3oool6000003X0oooo00D0oon?03D000000000000e08ooo`0X0?ooo`0308le
501Oa?l0oooo08`0oooo002N0?ooo`@00000703oool00`000000oooo0?ooo`0[0?ooo`030000003o
ool0oooo00<0oooo1P00001Q0?ooo`030?ooa01O53D0Sooo08`0oooo002K0?ooo`<00000803oool0
0`000000oooo0?ooo`0f0?ooo`L00000G03oool00`3oScD003F?0?ooo`2;0?ooo`00V03oool30000
02<0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`040?ooo`H00000EP3oool00`3o
ohl0=@0D05o4o`0V0?ooo`040?ooa01O500001AO0<Coof40oooo002D0?ooo`@000009P3oool00`00
0000oooo0?ooo`120?ooo`L00000D03oool00`3oohl0=@0001AOa00V0?ooo`040?ooa01O500001AO
0<Coof40oooo002A0?ooo`<00000:P3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo
00@0oooo1P0000150?ooo`030?n?G`1Oa?l0oooo0080oooo00<0olAO01@0000DGl@09P3oool0103o
ol@0Ga@0000DG`34oomQ0?ooo`00S@3oool4000002d0oooo00<000000?ooo`3oool0C@3oool00`00
0000oooo00000005000003l0oooo00P0ooo405lD501Oa?l0oooo0?ooS`0e0000000e08ooob80oooo
00<0ScD00000000000001@0000000`0eSol0oooo0?ooo`1N0?ooo`00RP3oool300000340oooo00<0
00000?ooo`3oool0E@3oool6000003X0oooo00H0ooo405lD0000000000000000=@2?oolS0?ooo`08
0?o4G`0eGl@0oooo0?ooo`3ool@0Ga@0000DG`34oomQ0?ooo`00Q`3oool3000003@0oooo00<00000
0?ooo`3oool0F03oool00`000000oooo0?ooo`06000005h0oooo00L0a5lD03F?o`3oool0ooo405lD
000055l0a?oo0640oooo00230?ooo`@00000=`3oool00`000000oooo0?ooo`1Q0?ooo`H00000F03o
ool01`3ool@0Ga@e08ooo`3ool@0Ga@0000DG`34ool0H@3oool00800oooo0`00000k0?ooo`030000
003oool0oooo06<0oooo00@000000?ooo`3oool0oooo2000001A0?ooo`060?o4G`0D=Hl0ooo405lD
000055l0a?ooH@3oool007`0oooo1000000n0?ooo`030000003oool0oooo06h0oooo1P00001=0?oo
o`050<AO500eSl@0Ga@0000DG`34ool0H@3oool007T0oooo0`0000120?ooo`030000003oool0oooo
06h0oooo00<000000?ooo`3oool00`3oool6000004L0oooo00D0ooo405lD500D500001AO0<Coo`1Q
0?ooo`00MP3oool3000004D0oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`080?oo
o`L00000@P3oool0103oa5l05000000DG`34oomQ0?ooo`00LP3oool4000004P0oooo00<000000?oo
o`3oool0N@3oool00`000000oooo0?ooo`040?ooo`H00000?@3oool00`34Ga@001AO0<Coo`1Q0?oo
o`00K`3oool3000004`0oooo00<000000?ooo`3oool0Q03oool00`000000oooo00000005000009/0
oooo001[0?ooo`@00000C`3oool00`000000oooo0?ooo`2<0?ooo`H00000U@3oool006P0oooo0`00
001C0?ooo`030000003oool0oooo08h0oooo00@000000?ooo`3oool0oooo1P0000000`3oool00000
0?ooo`2<0?ooo`00I@3oool3000005H0oooo00<000000?ooo`3oool0V03oool6000008T0oooo001Q
0?ooo`@00000F@3oool00`000000oooo0?ooo`2I0?ooo`030000003oool0oooo0080oooo1P000023
0?ooo`00GP3oool3000005d0oooo00<000000?ooo`3oool0V@3oool00`000000oooo0?ooo`070?oo
o`L00000O@3oool005X0oooo1000001P0?ooo`030000003oool0oooo0:@0oooo00<000000?ooo`3o
ool00`3oool6000007L0oooo001G0?ooo`<00000I03oool00`000000oooo0?ooo`2^0?ooo`030000
003oool0000000D00000L@3oool005@0oooo0`0000140?ooo`P000006`3oool00`000000oooo0?oo
o`2f0?ooo`H00000J`3oool00500oooo100000150?ooo`80000000<0Tg:[0000002gSj`00P2gSj`0
0`000000^X^T0;Z;Y004000001L0oooo00<000000?ooo`3oool0^03oool010000000oooo0?ooo`3o
ool6000000030?ooo`000000oooo0680oooo001=0?ooo`<00000B@3oool010000000Tg:[09=bZ`00
00040;N?[0060000002jRj@0^X^T0;Z;Y0000000_hBD1@00000B0?ooo`030000003oool0oooo0<80
oooo1P00001O0?ooo`00BP3oool3000004`0oooo00@0000009=bZ`2CLZ/00000102gSj`00`000000
^X^T0;Z;Y0020;Z;Y0060000002oQ9@0_hBD0;n4U0000000OFjh0`00000?0?ooo`030000003oool0
oooo0<<0oooo00<000000?ooo`3oool00P3oool6000005T0oooo00160?ooo`@00000C`3oool01000
0000Tg:[09=bZ`0000040;N?[0030000002jRj@0^X^T0080^X^T00<000000;n4U02oQ9@00P2oQ9@0
10000000OFjh07e^^000000?0?ooo`030000003oool0oooo0<<0oooo00<000000?ooo`3oool01`3o
ool7000005<0oooo00130?ooo`<00000DP3oool200000080Tg:[0P0000030;N?[0030000002jRj@0
^X^T00<0^X^T00<000000;n4U02oQ9@00P2oQ9@00`000000OFjh0000000?0?ooo`030000003oool0
oooo0140oooo3000002g0?ooo`@00000C`3oool003l0oooo1000001E0?ooo`8000000P2CLZ/20000
00@0]hn/00<000000;Z;Y02jRj@00P2jRj@010000000_hBD0;n4U02oQ9@4000000l0oooo00<00000
0?ooo`3oool02P3oool7000000D0/8R[00D000000;^>YP2kSZH0^hjV000000020=fEP@<00000/@3o
ool400000080oooo00<000000?ooo`3oool0C@3oool003`0oooo0`00001I0?ooo`8000000P2CLZ/2
000000@0]hn/00<000000;Z;Y02jRj@00`2jRj@300000080d9fQ00<0000007e^^00000003P3oool0
0`000000oooo0?ooo`030?ooo`L0000000D0T72[0000003K_CT0fkdi000000060;28Z`800000102k
SZH010000000gIF10=fEP@3MUH4200000:`0oooo100000060?ooo`030000003oool0oooo04d0oooo
000i0?ooo`<00000G03oool010000000Tg:[09=bZ`2CLZ/2000000@0]hn/00<000000;Z;Y02jRj@0
0P2jRj@2000000@0d9fQ00<0000007e^^00000003P3oool010000000oooo0?ooo`3oool200000003
09Qg[@000000U7B/0080U7B/00@00000091`Z`2@L:/000000P3K_CT00`000000/8R[0;28Z`040;28
Z`030000002]PjL0000000<0^hjV00@000000=fEP@3MUH40gIF10P00001`0?ooo`<00000=@3oool4
000000T0oooo0P00001?0?ooo`00=@3oool4000005l0oooo00H0000009=bZ`2CLZ/000000>>m/000
00040;N?[0030000002jRj@0^X^T0080000000<0`96V0000003@WJ400`3@WJ43000000h0oooo00<0
00000?ooo`0000000P00000209Qg[@050000002DM:`0U7B/09Ad[00000000P2@L:/00`000000fkdi
0=^m>@02000000D0/8R[00<000000:f3Y`000000102kSZH01@000000gIF10=fEP@3MUH40000006D0
oooo3@0000000`2kS:D0000000000006000002T0oooo1000000=0?ooo`800000C`3oool00380oooo
0`00001S0?ooo`060000002CLZ/0Tg:[0000003S_K0000001@2gSj`2000000@0`96V00<000000=2M
X@3@WJ400P3@WJ42000000h0oooo0P0000020:N4[`040000002HMjd0V7N]000000<0U7B/00@00000
091`Z`2@L:/000000P3K_CT2000000@0/8R[00@000000:f3Y`2]PjL00000102kSZH010000000gIF1
0=fEP@3MUH4200000600oooo1@00000309iaW`8000001@2_PZ@00`000000]hNS000000030;^<Y@04
0000002iQIh0^HFN0;V5WPH00000803oool400000140oooo0P00001?0?ooo`00;P3oool4000006H0
oooo00H0000009=bZ`2CLZ/000000>>m/00000050;N?[00300000030TJH0`96V00<0`96V00<00000
0=2MX@3@WJ400P3@WJ42000000h0oooo0P0000020:N4[`040000002HMjd0V7N]000000<0U7B/00@0
0000091`Z`2@L:/000000P3K_CT2000000@0/8R[00@000000:f3Y`2]PjL000001@2kSZH00`000000
gIF10=fEP@02000005h0oooo0P0000001@0b05h0000008QGT@28Ei4000000080WW6O00<000000:Uj
W`000000102_PZ@010000000]hNS0;N7X`0000030;^<Y@030000002iQIh0^HFN0080^HFN00D00000
0;a_N@2lKgT0000005Ekf`08000001@0oooo1000000E0?ooo`800000C`3oool002/0oooo0`00001Y
0?ooo`8000000P2CLZ/00`000000hkf`000000050;N?[00300000030TJH0`96V00<0`96V00<00000
0=2MX@3@WJ400P3@WJ42000000h0oooo0P0000020:N4[`040000002HMjd0V7N]000000<0U7B/00@0
0000091`Z`2@L:/000000P3K_CT2000000@0/8R[00D000000:f3Y`2]PjL0[H>W000000040;^>YP03
0000003MUH40gIF100800000GP3oool00`000000<P1N0000000208QGT@050000002NLIl0WW6O09ia
W`0000000P2YNYl01@000000[h:T0:n2Y02_PZ@000000080]hNS00<000000;^<Y@2kS:D00P2kS:D0
0`000000^HFN0;V5WP020;V5WP030000002lKgT0_6mi0080000000<0EG_K0000001FF[X00P1FF[X0
0`000000002600000003000000`0oooo1000000H0?ooo`<00000C`3oool002P0oooo0`00001/0?oo
o`8000000P2CLZ/00`000000hkf`000000040;N?[0030000002/OJ40000000@0`96V00<000000=2M
X@3@WJ400P3@WJ42000000h0oooo0P0000020:N4[`040000002HMjd0V7N]000000<0U7B/00P00000
091`Z`2@L:/000000=^m>@000000?R:6000000<0/8R[00<000000:f3Y`2]PjL00P2]PjL00`000000
^hjV0;^>YP030;^>YP040000003MUH40gIF1000005d0oooo00@000000380GP0b05h000000P28Ei40
1@000000WW6O09iaW`2NLIl000000080ZGZO00@000000:n2Y02_PZ@000000`2gQj<00`000000^hbU
0;^<Y@030;^<Y@030000002iQIh0^HFN0080^HFN1@0000001@1ENm/0000005IJ^P1FF[X0000000<0
002600<000000=^U700000000P0000040?ooo`D00000703oool3000004l0oooo000T0?ooo`@00000
K`3oool200000080Tg:[00@000000>>m/03S_K0000000`2gSj`00`000000[7fQ000000040<2AYP03
0000003@WJ40d9fQ0080d9fQ0P00000>0?ooo`8000000P2WQ:l010000000V7N]09Qg[@00000309Ad
[0060000002@L:/0T72[0000003K_CT000000P0n8XH010000000/8R[0;28Z`0000040:f3Y`030000
002kSZH0^hjV0080^hjV0`0000000`3MUH4000000000001L0?ooo`030000000b05h000000080R5NA
00D0000009iaW`2NLIl0WW6O000000040:UjW`030000002_PZ@0000000@0]hNS00<000000;^<Y@2k
S:D00P2kS:D00`000000^HFN0;V5WP03000000<0eJ:S0`00000205IJ^P050000000008H000260000
QP0000000`3KYA`500000240oooo0`00001?0?ooo`008P3oool400000740oooo0P00000209=bZ`04
0000003S_K00hkf`000000<0]hn/00<000000:amX@0000001030TJH00`000000d9fQ0=2MX@020=2M
X@<000003@3oool200000080YhB_00@0000009Qg[@2HMjd000000`2DM:`01P000000T72[091`Z`00
0000fkdi00000080?R:600@000000;28Z`2`R:/00000102]PjL00`000000^hjV0;^>YP0200000003
0;j>X`000000cIRO00<00000F`3oool2000000050380GP000000R5NA08QGT@0000000`2NLIl00`00
0000ZGZO0:UjW`020:UjW`8000001@2gQj<00`000000^hbU0;^<Y@030;^<Y@8000000P3=Vj801P00
0000eJ:S0=FRX`3EXZ<000000=j[X@8000000P1FF[X01000000000260000QP0000030=^U70040000
0036WJh0000000000280oooo0`00000F0?ooo`050?oOW018000000000000M02oool00`3oool01P3o
_g@0000007Boo`3oool0oooO09a800<0000000<0M;oo0?ooo`3oool09@3oool00280oooo00@00000
0?ooo`3oool0oooo1`00001Z0?ooo`8000000P2CLZ/010000000hkf`0>>m/00000030;N?[0030000
002/OJ40000000D0`96V00@000000=2MX@3@WJ40d9fQ0`00000=0?ooo`050000002WQ:l0YhB_0:N4
[`0000000P2HMjd01`000000U7B/09Ad[02DM:`00000091`Z`0000000P3K_CT010000000?R:603hR
QP0000020;28Z`030000002]PjL0[H>W00<0[H>W0P0000040;j>X`@00000F`3oool01P000000<P1N
0380GP000000R5NA000000<0WW6O00<000000:UjW`2YNYl0102YNYl00`000000]hNS0;N7X`030;N7
X`030000002kS:D0^hbU0080^hbU0P0000040<fKXP050000003EXZ<0eJ:S0000003NZj400P000002
05IJ^P040000000008H00026000000<0fjDL00D000000<JM[P36WJh0aYf^0000000P0?ooo`040000
003K];400000000001D0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00L0okmd
0000001d_ol0oooo0?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool09@3oool00280oooo00<0
00000?ooo`3oool0203oool7000006<0oooo0P00000209=bZ`040000003S_K00hkf`00000080]hn/
00D000000:amX@2/OJ40[7fQ000000040<2AYP040000003@WJ40d9fQ0=2MX@<00000203oool=0000
00<0U7B/00<00000091`Z`0000000P3K_CT01P000000?R:603hRQP000000/8R[000000H0[H>W00<0
00000;j>X`2nSZ<00`2nSZ<010000000cIRO00000000001J0?ooo`80000000D0<P1N00000028Ei40
R5NA0000000309iaW`030000002YNYl0ZGZO00<0ZGZO0P0000050;N7X`030000002kS:D0^hbU0080
000000<0aY>Q0000003=Vj800`3=Vj8010000000eJ:S0=FRX`0000020=j[X@80000000D0EUZj0000
000008H00026000000030=^U700500000036WJh0aYf^0<JM[P000000803oool010000000fkBa0000
0000000E0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd
0;ooo`0U0?ooo`008P3oool00`000000oooo0?ooo`0?0?ooo`L00000G03oool200000080Tg:[00@0
00000>>m/03S_K0000000P2gSj`01@000000[7fQ0:amX@2/OJ40000000@0`96V00<000000=2MX@3@
WJ400P3@WJ42000000D0oooo0`00000409W;n@D000000`2IboT500000005091`Z`000000fkdi0=^m
>@0000000P0n8XH00`000000/8R[000000060:f3Y`030000002nSZ<0_XjS00<0_XjS00<000000<fH
W`3=V9l00P00001I0?ooo`060000000b05h0<P1N00000028Ei400000102NLIl00`000000ZGZO0:Uj
W`030:UjW`030000002bOid0000000D0]hNS0P0000040<JCX@050000003=Vj80cI^R0<fKXP000000
0`3EXZ<01P000000gZ^Q0=j[X@000000EUZj00000080002600D000000=^U703KYA`0fjDL00000003
0<JM[P030000003oool0oooo01h0oooo00@000000=^d/@0000000000503oool00`3ookl0M00004RL
g`030?ooo`030=nLB00007@0_ooo00/0oooo00<0_g@007Boo`3oool0903oool00280oooo00<00000
0?ooo`3oool05P3oool7000005@0oooo00<000000:J3[P0000000P2CLZ/010000000hkf`0>>m/000
00020;N?[0050000002/OJ40[7fQ0:amX@0000001030TJH00`000000d9fQ0=2MX@020=2MX@800000
00<0oooo0000000000000P00000309W;n@H000000`2IboT4000000@0VL_i2P0000050:f3Y`030000
002^OZ00000000@0_XjS00<000000<fHW`3=V9l00P00001H0?ooo`80000000<0<P1N000000000000
100000000`2NLIl000000:UjW`040:UjW`040000002bOid0/WnM000000D0]hNS00<000000<JCX@36
Tj400`36Tj401@000000cI^R0<fKXP3=Vj80000000<0eJ:S00H000000=j[X@3NZj40000005IJ^P00
00030000QP040000003KYA`0fjDL000000<0aYf^00<000000?ooo`3oool07P3oool010000000fkBa
0000000000070?ooo`030?oog`2LB000000000L0000000H0B9cO0?ooo`3oool0omnL04P0001d_ol3
0?ooo`030?noM00004P0W=oo00X0oooo00<0oooO09a80018W=l09@3oool00280oooo00<000000?oo
o`3oool07@3oool7000004d0oooo00<000000:J3[P0000000P2CLZ/010000000hkf`0>>m/0000002
0;N?[0030000002/OJ40[7fQ0080[7fQ00D000000<2AYP30TJH0`96V000000040=2MX@<00000102I
boT5000000<0VL_i1@00000309W;n@X00000102IboT4000000040:f3Y`000000[WjP000000D0_XjS
00@000000<fHW`3=V9l00000F03oool2000000040380GP000000Odn>000000<0U6>F0P0000040:Uj
W`050000002bOid0/WnM0;9oW@000000102gQj<2000000D0aY>Q00D000000<fKXP3=Vj80cI^R0000
00030=FRX`030000003NZj40gZ^Q00<000000`0008H010000000fjDL0=^U700000030<JM[P030000
003oool0oooo01d0oooo00<000000=^d/@3K];400P00000D0?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol02P3oool00`3OW4P0000007Boo`0U0?ooo`008P3oool00`000000oooo0?oo
o`0T0?ooo`L00000AP3oool00`000000YX>^0000000209=bZ`060000003S_K00hkf`0000002gSj`0
00001@2/OJ400`000000`96V0<2AYP020<2AYP@000000`2IboT5000000@0VL_i4@00000409W;n@D0
00000`2IboT:00000080cIRO0P00001G0?ooo`060000000b05h000000000001oChh000000`2DHiH0
0`000000XVjG000000020:UjW`030000002bOid0/WnM00<0/WnM00@000000;N7X`2gQj<000000P31
S9d00`000000aY>Q0<JCX@020<JCX@050000003=Vj80cI^R0<fKXP0000000`3EXZ<00`000000gZ^Q
0=j[X@03000000<0002600@000000=^U703KYA`000000`36WJh00`000000oooo0?ooo`0M0?ooo`03
0000003K];40fkBa00800000503oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0
oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo02D0oooo000R0?ooo`030000003oool0oooo
02/0oooo1`00000o0?ooo`8000000`2CLZ/01P000000hkf`0>>m/0000000]hn/000000D0[7fQ00<0
00000<2AYP0000000`00000309W;n@D000000`2IboT:000000<0VL_i1@00000309W;n@D00000102I
boT9000000@0VL_i1000001F0?ooo`@0000000<0Odn>0000002DHiH00`2DHiH01P000000XVjG0:9^
U`000000ZGZO000000D0/WnM00@000000;N7X`2gQj<000000P31S9d00`000000aY>Q0<JCX@020<JC
X@030000003=Vj80cI^R0080cI^R00@000000=FRX`3EXZ<000000P3NZj420000000505IJ^P000000
00260000QP0000000P3KYA`01@000000aYf^0<JM[P36WJh0000001l0oooo00D000000=^d/@000000
oooo0000000D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01P3OW4P0
0000000000000000001809cOobH0oooo000R0?ooo`030000003oool0oooo0380oooo1`00000h0?oo
o`8000000`2CLZ/010000000hkf`0>>m/03S_K02000000<0[7fQ1000000409W;n@@00000102IboT9
0000000409W;n@000000VL_i09W;n@@00000102IboT5000000<0VL_i2P00000309W;n@D000000`2I
boT500000540oooo100000000`1oChh0000009ASUP0209ASUP030000002RKYL0XVjG0080XVjG00<0
00000;9oW@2bOid01@2bOid2000000<0`HbM00<000000<JCX@36Tj400P36Tj400`000000cI^R0<fK
XP020<fKXP040000003EXZ<0eJ:S00000080gZ^Q0P0000001@1FF[X000000000QP0008H000000080
fjDL00D000000<JM[P36WJh0aYf^0000000O0?ooo`030000003K];40000000800000503oool00`3o
okl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono07@000000000B9cO:03oool0
0280oooo00<000000?ooo`3oool0>@3oool700000340oooo0P00000309=bZ`040000003S_K00hkf`
0>>m/0H000000`2IboTB000000<0VL_i1@00000309W;nA800000102IboT4000000@0VL_i4@000018
0?ooo`<000000P1oChh01@000000U6>F09ASUP2DHiH0000000<0XVjG00<000000:]dU@0000001P2b
Oid00`000000`HbM0<6<W@030<6<W@0300000036Tj40aY>Q0080aY>Q00D000000<fKXP3=Vj80cI^R
000000020=FRX`040000003NZj40gZ^Q0=j[X@<000000P0008H01@000000fjDL0=^U703KYA`00000
0080aYf^00<000000?ooo`3oool07@3oool01@000000fkBa0000003oool0000001D0oooo00<0gia8
0018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030;md001d_ol0oooo02P0oooo000R0?oo
o`030000003oool0oooo0400oooo1`00000Z0?ooo`8000000P2CLZ/010000000hkf`0>>m/03S_K03
000000@0VL_i2@00000409W<n@@00000102IboT5000000<0VL_i2@00000409W;n@D000000`2IboT5
000000<0VL_i2P00000309W<n@D000000`2Ic?T5000004<0oooo00D0000004/LM0000000Odn>0000
000309ASUP030000002RKYL0XVjG0080XVjG00@000000:]dU@2[M9D00000102bOid2000000D0`HbM
00<000000<JCX@36Tj400P36Tj401@000000cI^R0<fKXP3=Vj8000000080eJ:S00@000000=j[X@3N
Zj40gZ^Q0`0000020000QP050000003KYA`0fjDL0=^U700000000P36WJh00`000000oooo0?ooo`0L
0?ooo`80000000@0fkBa0000003oool000005@3oool01`3ookl0M00007Boo`3oool0okmd0000M02o
ool0203oool01P3ogi`0B000000000000000001809cOobD0oooo000R0?ooo`030000003oool0oooo
04L0oooo1`00000R0?ooo`030000002VPjh000000080Tg:[1000000309W;n@X000000`2Ic?T50000
00<0VLci1@00000209W;n@030000002IboT0000000P000000`2IboT5000000@0VL_i1000000409W;
n@T00000102Ic?T4000000@0VLci1000000409W;n@T00000>@3oool20000000404/LM0000000Odn>
000000<0U6>F00D000000:9^U`2RKYL0XVjG000000030:]dU@050000002bOid0/WnM0;9oW@000000
0P2kPYL00`000000`HbM0<6<W@020<6<W@0300000036Tj40aY>Q0080aY>Q00D000000<fKXP3=Vj80
cI^R000000030=FRX`030000003NZj40gZ^Q00<000000P0008H01@000000fjDL0=^U703KYA`00000
0080aYf^00<000000?ooo`3oool0703oool2000000040=^d/@000000oooo000001H0oooo00D0oono
07@000000000001d0;ooo`0:0?ooo`050;md00000000000000000018W=l09@3oool00280oooo00<0
00000?ooo`3oool0CP3oool7000001/0oooo00<000000:J3[P0000003`00000409W<n@D000000`2I
c?TA000000@0VL_i1@00000309W;nA8000000`2Ic?T5000000<0VLci4P00000309W;n@D00000=03o
ool01@000000Baad0000001oChh0000000<0U6>F00D000000:9^U`2RKYL0XVjG000000050:]dU@03
0000002bOid0000000<0^h:G00<000000<6<W@31S9d00`31S9d01@000000aY>Q0<JCX@36Tj400000
00<0cI^R00D000000=FRX`3EXZ<0eJ:S000000020=j[X@030000003[]Y/000000080002600D00000
0=^U703KYA`0fjDL000000030<JM[P030000003oool0oooo01/0oooo0P000000103K];4000000?oo
o`00001?0?ooo`008P3oool00`000000oooo0?ooo`1E0?ooo`L000004P3oool9000000<0VL_i1@00
000309W<n@D00000102Ic?T9000000<0VL_i1@00000309W;n@D00000102IboT9000000<0VL_i1@00
000409W<n@@00000102Ic?T9000000<0VLci1@00000409W;n@@00000102IboT<000002D0oooo1000
00000`1;77@0000007m?SP0600000080XVjG00<000000:]dU@2[M9D0102[M9D00`000000^h:G0;^2
U`030;^2U`0300000031S9d0`HbM0080`HbM00D000000<JCX@36Tj40aY>Q000000040<fKXP040000
003EXZ<0eJ:S00000080gZ^Q00<000000>^fV`0000000P0008H01@000000fjDL0=^U703KYA`00000
00<0aYf^00<000000?ooo`3oool06`3oool2000000040=^d/@000000oooo000004l0oooo000R0?oo
o`030000003oool0oooo05`0oooo1`0000050?ooo``000000`2IboT5000000<0VLci4P00000309S<
n@D000000`2IboTA000000@0VL_i1@00000309W<nA400000102Ic?T5000000<0VL_i4@00000409W;
nA8000004@3oool500000080oooo1@00000308A;P`030000002GFhP0Ue^8008000001P2[M9D00`00
0000]7J>000000040;^2U`0300000031S9d0`HbM0080`HbM00D000000<JCX@36Tj40aY>Q00000004
0<fKXP040000003EXZ<0eJ:S00000080gZ^Q00H000000>^fV`00000000000000QP0000030=^U7005
00000036WJh0aYf^0<JM[P0000007@3oool01P000000fkBa0=^d/@000000oooo000004l0oooo000R
0?ooo`030000003oool0oooo06<0oooo1@00000309W;n@D00000102IboT4000000@0VLci2@000003
09W<n@D00000102Hc?T4000000<0V<_i2P00000309W;n@D000000`2IboT5000000@0VL_i2@000003
09W<n@D000000`2Ic?T5000000@0V<ci2@00000309W;n@D000000`2IboT5000000<0VL_i2P000002
0;fF[`030000002gTJl00000010000001`3oool01@000000Baad0000001O9G00000000<0Q4^300H0
000009MKR02GFhP000000:9SQ`0000040:]dU@040000002dMXh0]7J>000000D0^h:G00<000000<6<
W@31S9d00P31S9d01@000000aY>Q0<JCX@36Tj40000000<0cI^R00@000000=FRX`3EXZ<000000P3N
Zj401P000000jkJK0000000000000026000000<0fjDL00D000000<JM[P36WJh0aYf^0000000L0?oo
o`8000000P3K];44000004h0oooo000R0?ooo`030000003oool0oooo0640oooo1000000309S;n@D0
00000`2IboT5000000<0VL_i2P00000309W<n@D000000`2Hc?TA000000@0VL_i1@00000309W;n@D0
00000`2IboT9000000@0VLci1000000409W<nA400000102HboT4000000@0VL_i1000000409W;n@T0
0000102IboT400000080]i6_00<000000;6;[@2aRjd01@2aRjd01@000000a9b_0<BL[`34W:l00000
00T0oooo0P0000000`1O9G00000008A;P`0208A;P`040000002GFhP0Ue^8000000<0XV>70P000000
0`2[M9D000000;AfSP030;AfSP030000002kPYL0^h:G0080^h:G00@000000<6<W@31S9d0`HbM0`00
00020<JCX@050000003=Vj80cI^R0<fKXP0000000P3EXZ<010000000gZ^Q0=j[X@0000020>^fV`04
0000000008H0002600000080fjDL00D000000<JM[P36WJh0aYf^0000000L0?ooo`8000000P3K];40
0`000000oooo0000001?0?ooo`008P3oool00`000000oooo0?ooo`1J0?ooo`8000000P3oool40000
00<0V<_i1@00000309W;nA400000102Ic?T4000000@0V<ci2@00000309S;n@D000000`2HboT50000
00@0VL_i4@00000309W<n@D000000`2Ic?T:000000<0V<ci1@00000309S;n@D000000`2HboTA0000
00@0VL_i1@00000309W<n@T0000000<0/H^]00000034W:l00P34W:l00`000000oooo0?ooo`060?oo
o`050000001;77@0000005lUL00000000P24Bh<01@000000Ue^809MKR02GFhP0000000D0XV>700<0
00000;AfSP2dMXh0102dMXh00`000000^h:G0;^2U`020;^2U`<0000000<0cXn@0000003HVY800`00
00030<fKXP040000003EXZ<0eJ:S00000080gZ^Q00@000000>^fV`3[]Y/000000P0008H010000000
fjDL0=^U700000030<JM[P030000003oool0oooo01X0oooo0P0000020=^d/@040000003oool0oooo
000004h0oooo000R0?ooo`030000003oool0oooo05P0oooo1P00000409O;nP@00000102HboT90000
00<0V/_i1@00000309W<nA8000000`2Hc?X600000080V<_i1@00000309W;n@T00000102IboT40000
00@0VLci4@00000409S<n@@00000102Hc?TA000000<0VL_i1@00000309W;n@D000000`2IboT:0000
00<0VLci100000000`34W:l000000?ooo`070?ooo`80000000<0GbE`00000024Bh<00P24Bh<01000
0000Ue^809MKR00000050:9SQ`030000002[IGl0000000D0]7J>00<000000;^2U`2kPYL00`000004
0<j?T0030000003HVY80f9ZB008000000P3=Vj8010000000eJ:S0=FRX`0000020=j[X@040000003[
]Y/0jkJK00000080002600@000000=^U703KYA`000000`36WJh00`000000oooo0?ooo`0J0?ooo`80
00000P3K];4010000000oooo0?ooo`00001>0?ooo`008P3oool00`000000oooo0?ooo`1G0?oooaT0
00000`2JboP5000000<0V/ci2@00000409S<nP@00000102Gc?X4000000@0V<_j4@00000309W;n@D0
00000`2Ic?T:000000<0VLci1@00000309S<n@D000000`2Hc?T9000000@0V<_i1000000409W;n@@0
0000102IboTA000000<0VLci1@00000309W<n@T0000000H0oooo000000000000GbE`00000024Bh<5
000000D0XV>700@000000:]UO`2[IGl000001@2dMXh3000000<0aX>900D000000<j?T03>Si00cXn@
000000030=RJTP<0000000L0cI^R0000003EXZ<0eJ:S0000003NZj4000000080jkJK00@000000000
QP0008H000000P3KYA`01@000000aYf^0<JM[P36WJh0000001/0oooo00D000000>o9[0000000fkBa
000000030?ooo`030000003oool0oooo04`0oooo000R0?ooo`030000003oool0oooo04`0oooo1000
00000`3oool0000000000009000000@0T/[k1000000409S9n@@00000102Kb_P4000000@0Vl_h2@00
000309W=nPD000000`2Gc?XB000000<0V<_i1@00000309W;n@D000000`2JboT9000000@0VLci1000
000409S<nQ4000000`2HboT5000000<0V<_i1@00000309W;n@T00000102IboT4000000@0VLci4@00
000309S<nPL000000P21>6X010000000TTI/0000000000020:9SQ`030000002[IGl0ZfEo0080ZfEo
00<000000;AfSP2dMXh00P0000020;edOP0500000036PhT0aX>90<J3R@0000000`3>Si001@000000
f9ZB0=RJTP3HVY8000000080i:F?0P0000020=FRX`050000003NZj40gZ^Q0000003[]Y/00P000000
1@0008H000000=^U703KYA`0000000<0aYf^00<000000?ooo`3oool06@3oool01@000000klV/0000
003K];40000000<0oooo00<000000?ooo`3oool0C03oool00280oooo00<000000?ooo`3oool0B@3o
ool300000080ImOa2`00000308_;o@D000000`2Db?X5000000<0VlWg1@00000309g;m`X000000`2J
cOTA000000@0U<_j1000000409O:n@@00000102Ib_PA000000<0V/ci4@00000409O<nP@00000102G
boX4000000@0V<_i4@00000309[;n@D000000`2JboT9000000@0VLci1000000409S<nQ8000001P2[
IGl2000000@0_GAn00<000000<J3R@36PhT00P36PhT01@000000cXn@0<j?T03>Si0000000080f9ZB
00D000000>BUS`3TYHl0i:F?000000020=FRX`050000003NZj40gZ^Q0000003[]Y/00P0000001@00
08H000000=^U703KYA`0000000<0aYf^00<000000?ooo`3oool06@3oool01P000000klV/0000003K
];4000000?ooo`<00000CP3oool00280oooo00<000000?ooo`3oool0A`3oool>000000030<Wnh000
0000000000H000000`2;ao`5000000<0VLKg1@00000309o9m@T00000102Nc_PA000000<0TL_l1@00
00000`2Eb_X0000009G:nP05000000<0VLWh2@00000409c;n0@00000102Kc?T9000000<0V<gj1@00
000309K<nPD000000`2Fc?XA000000@0VL[i1000000409[;n0@00000102JboT9000000<0V/ci1@00
000309W<nPT00000102GcOX4000000@0U/cj100000001@2[IGl000000000002bGV`0000000D0_GAn
00D000000<J3R@36PhT0aX>9000000030<j?T0040000003HVY80f9ZB000000<0i:F?00@000000=FR
X`3EXZ<000000P3NZj401P000000jkJK000000000000002600000080fjDL00D000000<JM[P36WJh0
aYf^0000000K0?ooo`050000003_bJ`000000=^d/@0000000`3oool00`000000oooo0?ooo`1<0?oo
o`008P3oool00`000000oooo0?ooo`140?ooo`D000000`2f^]`2000000@0^=Wa00@000000<Wnh039
o^0000000`2WRkL00`000000UX2i00000007000000<0Tl?g100000040:76lPT000000`2Uc_HQ0000
00@0V<Sh2@00000309k:mPD000000`2Mc?P9000000@0VLkj6@00000309K:nPD000000`2Hb_T50000
00<0V/[h4@00000409_<nA4000000`2GcO/I000000030;edOP000000aX>900<0aX>900D000000<j?
T03>Si00cXn@000000020=RJTP030000003TYHl0i:F?0080000000D0eJ:S0000003NZj40gZ^Q0000
00020>^fV`030000000008H000000080fjDL00D000000<JM[P36WJh0aYf^0000000K0?ooo`050000
003_bJ`000000=^d/@0000000`3oool00`000000oooo0?ooo`1<0?ooo`008P3oool00`000000oooo
0?ooo`110?ooo`d000000`2hfO401@000000bOkP0<Wnh039o^00000000<0Yh^g00<0000009J0^@2F
P;T01000000307jlnP@000000`2O`>l5000000<0Z<O_0P0000030:k<k`/0000000<0Y8Bb0000002Q
PK400P2QPK4010000000UGR`09Eh/000000206IDZ@P000000`2;aO/<000000@0X<Od100000040:3;
mPT000000`2JcoX9000000@0TLkm10000004097<o0@00000102Db_X4000000@0UlWi1000000409[9
n0T000000`2Lb_L5000000<0W<ch2@00000309[>nQ400000102Bc_`4000000@0TLcl1000000409?:
nPD0000000<0aX>90000003>Si000P3>Si0010000000f9ZB0=RJTP0000020>BUS`80000000D0eJ:S
0000003NZj40gZ^Q000000020>^fV`030000000008H000000080fjDL00D000000<JM[P36WJh0aYf^
0000000J0?ooo`060000003_bJ`0klV/0000003K];4000000`3oool00`000000oooo0?ooo`1<0?oo
o`008P3oool00`000000oooo0?ooo`0n0?ooo`P000000`2`b^d2000000<0/l;V100000020<Wnh003
0000002WRkL0Yh^g0080Yh^g00H0000009J0^@2FP;T0UX2i0000003ZSE<500000080U[O^0P000002
09o0k`8000000`2]_nP?00000080/i:d00@000000:B4/P2TQ;8000000`2QPK4010000000UGR`09Eh
/000000206IDZ@`000000`2F`OD200000080V/Ge0P0000020:;4l@D000000`2TbO<9000000@0W]3i
:@00000309W9n0T00000102MbOH4000000<0W/_g2@00000409c>n@T000000`2Fd?`M000000<0TlSj
100000020<j?T0040000003HVY80f9ZB000000<0i:F?0`0000020=j[X@060000003[]Y/0jkJK0000
000008H000000P3KYA`01@000000aYf^0<JM[P36WJh0000001X0oooo00H000000>o9[03_bJ`00000
0=^d/@0000030?ooo`030000003oool0oooo04`0oooo000R0?ooo`030000003oool0oooo03X0oooo
1000000309kDn`8000000`2YbO45000000<0/l;V0P0000020;6_eP<000001@2WRkL010000000UX2i
09J0^@0000020>Z=D`040000001IVO000000000000@0U[O^100000030:foj0800000102ga>@00`00
0000_]7X0;kAj002000000<0[8jg00@000000;>B]02cT[@000000P2TQ;801@000000XH6a0:61/@2Q
PK4000000080UGR`00<0000006IDZ@1VE:T01P0000000`1TZOL000000000000208ZjmP800000102F
`OD3000000<0X/Ca0P0000030:S5kPD000000`2Vc?<200000080YM7g7P00000308_8o0D000000`2E
aoP=000000<0W/Oe1@00000309o9mPT000000`2OcOP9000000@0VM7l;@0000030>BUS`<000000P3N
Zj400`000000jkJK0>^fV`0300000080fjDL00D000000<JM[P36WJh0aYf^0000000J0?ooo`060000
003_bJ`0klV/0000003K];4000000`3oool00`000000oooo0?ooo`1<0?ooo`008@3oool00`000000
oooo0?ooo`0h0?ooo`@000000`2Nc_P5000000<0ZLWa0P0000030:^ni`<000001@2a[mH00`000000
YYO5000000040:N;]`040000002FP;T0UX2i00000080jXeC00@0000005VIl01IVO00FIW`0P000000
0`2F]nh00000000000030:jegP@000000`2ga>@6000000@0[8jg00@000000;>B]02cT[@000000P2T
Q;801@000000XH6a0:61/@2QPK4000000080UGR`00<0000006IDZ@1VE:T00`0000000`0<B<H00000
000000020000000506BYm`00000000000000002:^_H0100000030::nk@D000000`2XaNh2000000<0
[LW^2P00000209Ij/@030000002GNk<0Ug^c0080Ug^c00D0000005]B[`1KDZl0Fe:_000000020000
MP03000000000980000000h00000102@aOX4000000@0VLGf1000000309k6m0D000000`2Qb?@50000
00<0XL[e0P0000030:;<mPT000000`2Ld_/Y00000080Sl7h0P00000309K2m@D0000000D0lk:30000
00000000gZ^Q000000020>^fV`<000000P3KYA`01@000000aYf^0<JM[P36WJh0000001X0oooo00@0
00000>o9[0000000fkBa0P000000103oool000000?ooo`00001>0?ooo`008@3oool00`000000oooo
0?ooo`0e0?ooo`@000000`2MboP5000000<0XlSc100000040:^ni`80000000<0ZK3N000000000000
0P2a[mH01@000000YYO50:JGa@2VUlD0000000<0Yh^g00@0000009J0^@2FP;T000000`3ZSE<00`00
0000FIW`05VIl00205VIl08000001@2^]Mh00`000000^kWG0;^ie`05000000030<O5f00000000000
00D0[8jg00@000000;>B]02cT[@000000P2TQ;801@000000XH6a0:61/@2QPK4000000080UGR`00<0
000006IDZ@1VE:T00P0000000`3bVcD0000000a8aP08000000@0WkKY100000040:bni`l000000P2F
NK0010000000UWZa09Ij/@00000409Mk/`050000001KDZl0Fe:_05]B[`0000000P0007H010000000
002B00000000>KdB000000<0U/;f1@00000309k3lP@00000102RaO84000000@0XlSc2@0000030:??
m`8000000`2Rd_T?00000080h[ba00H000000<jX/`3>Z;<000000=J^/00000030<RR/@0700000034
W[80a9jb0<BN/P000000a9nb000000020=2X/14000000`2J_O44000000040=j[X@000000jkJK0>^f
V`80000000@000260000003KYA`000000`36WJh00`000000oooo0?ooo`0G0?ooo`80000000<0klV/
0000003K];400P0000020?ooo`800000CP3oool00240oooo00<000000?ooo`3oool0<P3oool40000
00<0Vl_h1@00000309o8m@@000000`2U`>d5000000@0ZK3N0P0000040:JGa@050000002WRkL0Yh^g
0:N;]`0000000`2FP;T01P000000jXeC0>Z=D`3ZSE<0000005VIl08000000P2[Ym8200000080[[GN
00<000000;^ie`2k^ML00P2k^ML010000000alGH0<O5f037aMP3000000H0[8jg00<000000;>B]000
00000P2TQ;800`000000XH6a0:61/@020:61/@070000002EN;00UGR`0000001VE:T000000000=P02
0000000300a8aP000000NICQ0080NICQ00<0000009FZi@0000000P00000309nfj@80000000<0[[KO
0000000000000`2/_^L200000080]<?U0P0000030;[<j0D000000`2FNK0010000000UWZa09Ij/@00
000409Mk/`050000001KDZl0Fe:_05]B[`0000000P0007H01`000000002B0000TP00000003Vm0000
000fKm`01@00000206ban@80000000<0Q;[h0000000000000`2@_oH4000000@0W<3a100000030:?2
k`D000000`2VaO09000000@0Ylgc4`000000103/cKD000000?OFZ@0000020>:l/@060000003>Z;<0
cZRc0000003F[[0000000`38X[401`000000a9jb0<BN/P34W[8000000<BO/P0000000P3@Z;001P00
0000a9ja0<BN/@34W[4000000?ooo`@000000P24^_P200000080T[cc1@00000209jij`<000000P3[
]Y/2000000050000QP000000fjDL0=^U700000000P36WJh00`000000oooo0?ooo`0G0?ooo`800000
00<0klV/0000003K];400P0000030?ooo`030000003oool0oooo04`0oooo000Q0?ooo`030000003o
ool0oooo02l0oooo1000000309[;n0@00000102LbOL<000000<0X[WY100000020:2VfP8000000`2V
UlD01@000000Yh^g0:N;]`2WRkL000000080UX2i00<000000>Z=D`3ZSE<00P3ZSE<00`000000ZjOB
0:^WdP040:^WdP800000102k^ML00`000000alGH0<O5f0020<O5f0<000001P2/S[L00`000000/i:d
000000020:B4/P030000002QPK40XH6a0080XH6a00L0000009Eh/02EN;00000006IDZ@000000000f
008000000P0<B<H010000000NICQ07VDh@1iU>4200000080UJ[U0`0000050:jfg`l0000000<0UWV`
0000002FNK000P2FNK0010000000UWZa09Ij/@00000409Mk/`050000001KDZl0Fe:_05]B[`000000
0P0007H010000000002B0000TP000002000i_@030000000fKm`0000000h000000`2J_?0500000080
Xkk/0P0000020:?2k`8000000`2W`^d4000000@0ZLO_0P0000020:c9kaL0000000@0k<fe0000003g
eZT000000P3R_;401P000000cZRc0<jX/`000000eZj`000000<0b::a00L000000<BN/P34W[80a9jb
00000034Wk8000000080d:R`00D000000<BN/@34W[40a9ja000000040?ooo`P000000`2G]Nd20000
00<0W[W[1@0000001@0008H000000=^U703KYA`000000080aYf^00<000000?ooo`3oool05P3oool0
1P000000gK^f0000003_bJ`000000=^d/@8000000`3oool00`000000oooo0?ooo`1<0?ooo`008@3o
ool00`000000oooo0?ooo`0/0?ooo`@000000`2IboTA000000<0VLKg2000000209Zfk08000000P2P
Y]X200000080YYO500@000000:N;]`2WRkL000000P2FP;T01`000000jXeC0>Z=D`3ZSE<000000;:K
`00000001@2[Ym8010000000_:c90000000000020;^ie`0300000037aMP000000080000000@0gLc8
0000002XSk/000001@2/S[L00`000000/i:d000000020:B4/P030000002QPK40XH6a0080XH6a00X0
000009Eh/02EN;00000006IDZ@000000000f0000003bVcD000000P0<B<H01P000000NICQ07VDh@1i
U>40000009FZi@8000000P2/Z]H200000080[[KO0P0000040;VifP0600000032`m/0`/?K0<;3f`00
0000b]WU0P000000102FN[40000009Ii/000000309Ii/0040000002FN[40UWZa000000@0Ug^c00<0
000005]B[`1KDZl00P1KDZl01P000000001f0000MP000000002B000000<003Vm00D0000003I_g000
0000HY7X0000000207bTk@8000000P2;[nh200000080U;O_0P00000209Zll080000000<0X[WY0000
00000000102S_^`3000000@0Yl;]0P0000020:_3j`T000000`2^c_42000000<0[mCc2`0000001@3h
ik0000000>c=]@000000mmJY0080000000@0h[ba0000003>Z;<0cZRc0P0000000`3F[[0000000<RR
/@020<RR/@0700000034W[80a9jb0<BN/P000000a9nb000000020=2X/00500000034W[40a9ja0<BN
/@0000001P3oool300000080R;;a0P00000409Nek@<00000102T/n@300000080fjDL00@000000<JM
[P36WJh00000603oool200000080klV/00@000000=^d/@00000000000`3oool00`000000oooo0?oo
o`1<0?ooo`008@3oool00`000000oooo0?ooo`0Y0?oooa0000000`2Ib_T5000000<0V<Wi10000003
09?8nQ00000000D0Yh^g0000002FP;T0UX2i000000030>Z=D`040000002bVl00/Y_0000000<0ZjOB
00<000000;b/b@2l[<T00P2l[<T400000080akS;00D000000=g<b0000000Z8nk000000050:b>]`80
00000`2TQ;800`000000XH6a0:61/@020:61/@070000002EN;00UGR`0000001VE:T000000000=P03
000000<034S600<0000007VDh@1iU>400P0000060:bZeP8000001@2i^MX010000000`/?K0<;3f`32
`m/3000000060:>5/`000000UWZa0000002FNK0000000`2FNK0010000000UWZa09Ij/@00000409Mk
/`030000001KDZl0Fe:_0080Fe:_00<000000000MP0000000P0000001`0009800000000i_@00>Kd0
000003I_g00000000P1RTNP010000000O:C]07bTk@1lY>d200000080Rjo^1`0000040::ij@@00000
0`2X_NP5000000<0Zl?[0P0000030:o5jP8000000P2cbN/C000000050?SW/0000000k<fe0000003g
eZT00P000000103R_;4000000<jX/`3>Z;<2000000030=J^/0000000b::a0080b::a00L000000<BN
/P34W[80a9jb00000034Wk8000000080d:R`00D000000<BN/@34W[40a9ja000000080?ooo`T00000
0P2L[^@2000000<0Y;?T0P000000102Y[m`000000000000000030<JM[P030000003oool0oooo01D0
oooo0P0000020>o9[0040000003K];400000000000<0oooo00<000000?ooo`3oool01`3oool01@3o
gi`0B0000000000007@0_ooo0400oooo000Q0?ooo`030000003oool0oooo02<0oooo2`00000309W;
n@D000000`2HboTP00000080R8oD0P00000209J0^@040000003ZSE<0jXeC000000<0/Y_000D00000
0:^WdP2[Ym80ZjOB000000040;b/b@0300000037^</0akS;0080akS;00<000000=g<b00000000`2X
Sk/00`000000[8jg0:b>]`030:b>]`050000002TQ;80Y8Bb0:B4/P000000102QPK401`000000UGR`
09Eh/0000000IUBY0000000003H00`00000300a8aP060000001iU>4000000000002VWLd000001@2/
Z]H010000000^Zo>0000000000020;VifP0300000032`m/0000000@000000P2SQK<01@000000UWZa
0000002FNK00000000<0UWV`00@0000009Ij/@2FN[400000102GNk<00`000000Fe:_05]B[`0205]B
[`80000000<0V8>j0000002@Pl402000000206:Aj0040000001lY>d0O:C]07bTk@<000000P2M[N82
000000<0X;?V100000040:Rhi0@00000102]_^H8000000@0/lW[0P0000020;K?k@L0000000<0a?7_
0000003Cn><01@0000001@3hik0000000>c=]@000000mmJY0080000000D0h[ba0000003>Z;<0cZRc
000000020=J^/00500000038X[40b::a0<RR/@0000000P34W[8010000000a9nb0<BO/P0000020=2X
/00500000034W[40a9ja0<BN/@0000002@3oool20000000407BRkP000000000008nYi`800000102L
[^@3000000D0ZJoL0P0000020<JM[P030000003oool0oooo01@0oooo00<000000=fk]P0000000P3_
bJ`2000000030=^d/0000000oooo0080oooo00<000000?ooo`3oool01P3oool01`3ookl0M00007Bo
o`3oool0okmd0000M02oool0?`3oool00240oooo00<000000?ooo`3oool0803oool3000000<0VL_i
1@00000309S;nA4000000`2Gc?X5000000<0Tlkl4`00000309J0^@8000001@2bVl000`000000ZjOB
000000050;b/b@0300000037^</0akS;0080akS;00<000000=g<b00000000`2XSk/00`000000[8jg
0:b>]`030:b>]`050000002TQ;80Y8Bb0:B4/P000000102QPK402@000000UGR`09Eh/0000000IUBY
0000000003H0000007Ya_@020000008034S60P000000102VWLd000000:JMc@2VWLd200000080[:[F
00<000000;Z_cP2j[lh00P2j[lh4000000030<Fhc@000000e<gC008000000P2SQK<01@000000UWZa
0000002FNK00000000<0UWV`00D0000009Ij/@2FN[40UWZa0000000409Mk/`030000001KDZl0Fe:_
008000000P2HPkX010000000T8?10923`@00000308f8b@030000002?TM40000000H000000P2HYMh0
0`000000WJgR09f]hP0209f]hP80000000@0X;?V000000000000Z;;O0P0000030:Rhi080000000<0
[[SP0000000000000`2]_^H200000080/[oT0P0000020;K3i0/000000P33gN`3000000060=?hh`00
000000000>g]b@000000k=W10`0000001`3/cKD000000000003L^;@000000>:l/@0000000P3>Z;<0
10000000eZj`0=J^/00000030<RR/@0500000034W[80a9jb00000034Wk800P0000020=2X/0050000
0034W[40a9ja0<BN/@0000002P3oool;00000080XZSJ00<000000:V_g02Y[m`00`2Y[m`01@000000
[j_D00000036WJh0000001H0oooo00<000000=fk]P0000000P3_bJ`2000000030=^d/00000000000
00<00000203oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo03h0oooo000Q0?ooo`03
0000003oool0oooo01d0oooo0`00000309W;nA4000000`2Hc?T5000000<0Ulcj4@00000309KAo0/0
00000P2FP;T2000000H0/Y_00P0000040;b/b@0300000037^</0akS;0080akS;00@000000=g<b000
000000000`2XSk/00`000000[8jg0:b>]`030:b>]`030000002TQ;80Y8Bb0080Y8Bb00D000000:61
/@2QPK40XH6a0000000209Eh/0030000001VE:T000000080000f00<0000007Ya_@0000000P000000
102JSLD000000:JMc@0000040:JMc@030000002/Z]H0000000@0^Zo>00<000000<Fhc@35^<d00P35
^<d400000080XhFc00<0000009Ij/@0000000P2FNK0010000000UWV`09Ii/000000309Ij/@030000
002GNk<0Ug^c0080Ug^c00<0000005]B[`1KDZl00P00000309R3^P040000002@Pl40T8?1000000<0
SHS900@0000008nAd@2?TM4000000P2CVmP00`000000V:GN09RUgP0209RUgP8000000`2M[N820000
00D0Z;;O0`0000050:jhh0P000000`2f`n@2000000<0^/SU0P0000020;k@j00300000033gN`00000
00X0000000@0n>N`0000003/cKD0k<fe0`0000020>:l/@040000003>Z;<0cZRc00000080eZj`00D0
00000<RR/@38X[40b::a000000020<BN/P0400000034Wk80000000000080d:R`00D000000<BN/@34
W[40a9ja0000000<0?ooo`0500000020V>40P9SQ082Hh@0000000P2FXM`2000000<0XZSJ0P000000
0`2Y[m`00000000000030:n[e08000005@3oool010000000gK^f0=fk]P0000020>o9[080000000<0
fkB`0000003oool00P3oool00`000000oooo0?ooo`050?ooo`030?oo_`1d0000B9cO00<0oooo00<0
gia80000M02oool0?P3oool00240oooo00<000000?ooo`3oool06P3oool?000000@0V<ci10000003
09S<nA4000000`2HcOX5000000<0VLkj2@0000030:?;m@80000000<0UX2i0000000000001`2bVl00
10000000_JC00000000000020;b/b@0300000037^</0akS;0080akS;0P0000000`3D^l0000000:R?
^`030:R?^`040000002/S[L0[8jg0:b>]`800000102TQ;801@000000XH6a0:61/@2QPK4000000080
UGR`00<0000006IDZ@0000000P0003H00`000000NW6m0000000409Z=a@8000001@2VWLd00`000000
^Zo>0;Z_cP030;Z_cP0400000035^<d0aKS=0<Fhc@<000000P2XR[@01@000000XhFc0000002FN[40
00000080UWV`00@0000009Ii/02FNK0000000`2FN[400`000000Ug^c09Mk/`0209Mk/`050000001K
DZl000000:>7]P0000000`2HPkX010000000T8?10923`@00000308f8b@050000002?TM40Si7A08nA
d@0000000P2CVmP2000000@0V:GN0`0000020:N/fP8000000P2X/]l200000080[[;K0P0000020:jh
h0800000102c^Mh00`000000^;cM0;Rlg@0?000000030<oQi@000000000000<0000000<0kNg90000
003/fL400`0000020>c=]@<000000P3R_;4010000000cZRc0<jX/`0000020=J^/00500000038X[40
b::a0<RR/@0000000P34W[8010000000a9nb0000000000020=2X/00500000034W[40a9ja0<BN/@00
00003@3oool200000080P9SQ00@0000009JQg02FXM`0UZ7L0P0000020::XfP8000001`2_Zm@20000
01<0oooo00@000000=fk]P3M^kH000000P3_bJ`2000000030=^d/0000000oooo0080oooo00<00000
0?ooo`3oool01@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo03h0oooo000Q0?oo
o`030000003oool0oooo01L0oooo2000000309W<n@@000000`2Ic?T5000000<0V<ci2@00000309S<
n@D000000`2Ic?TA000000<0WlWf1P00000209V>a`030000002bVl00/Y_000<0/Y_000<000000;fT
`02mY<000P2mY<03000000@0akS;0P0000000`3D^l000000000000030:R?^`040000002/S[L0[8jg
0:b>]`800000102TQ;801@000000XH6a0:61/@2QPK4000000080UGR`00<0000006IDZ@0000000P00
03H010000000NW6m07Ya_@00000309Z=a@800000102VWLd00`000000^:C5000000030;Z_cP030000
0035^<d0aKS=0080aKS=00<000000=Joa00000000P2XR[@01@000000XhFc0000002FN[4000000080
UWV`00@0000009Ii/02FNK0000000`2FN[400`000000Ug^c09Mk/`0209Mk/`8000000`2SQkH01000
0000V8>j09R3^P0000030923`@050000002=R<T0SHS908f8b@0000000P2?TM400`000000Ti_H09>K
f00209>Kf0050000002HYMh000000000002UYMD00P0000040:N/fP8000001P2^/]/3000000<0/kWN
0P0000040;Rlg@050000002l`=d0_<3M0;c0g@0000001031amh00`000000am7Q0<OAh@0300000080
f^3K00<000000>3Id00000000P000000103/fL4000000000000000020>c=]@030000003L^;@00000
0080h[ba00@000000<jX/`3>Z;<000000P3F[[001@000000b::a0<RR/@38X[4000000080a9jb00@0
00000<BO/P00000000000P3@Z;001@000000a9ja0<BN/@34W[40000000d0oooo100000000`20V>40
000009JQg00209JQg0<000000P2WXm42000000@0[j_D00<000000;BYcP2dZLh00P00000@0?ooo`80
00000P3M^kH00`000000klV/0>o9[003000000@0oooo00<000000?ooo`3oool01@3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo03h0oooo000Q0?ooo`030000003oool0oooo0140oooo
2`00000309W<n@@000000`2Ic?TA000000<0VLci4@00000309[;n0D000000`2MbOL4000000<0XLGb
0P00000209V>a`030000002bVl00/Y_000<0/Y_000<000000;fT`02mY<000P2mY<000`000000_Z_6
0;j[aP02000000030<Nhb`00000000000080e;_00P0000040:R?^`050000002/S[L0000009j3]@00
0000102TQ;801@000000XH6a0:61/@2QPK4000000080UGR`00H0000006IDZ@1VE:T000000000=P00
000207Ya_@030000002JSLD0VXg50080VXg500<000000:JMc@2VWLd00P2VWLd00`000000^:C50;RT
a@02000000030;Z_cP000000aKS=00<0000000@0e[o40000002XR[@0Z8Zd0P000000102SQK<00000
09Ij/@00000209Ii/0040000002FNK00UWV`000000<0UWZa00<0000009Mk/`2GNk<00P2GNk<20000
00<0XhNf00@0000009R3^P2HPkX000000`2@Pl401@000000SHS908f8b@2=R<T0000000<0Si7A00@0
000009>Kf02CVmP0Ti_H0P0000050:FUe@80000000@0YjcJ000000000000[ZcE0P0000030:jbf`80
00000P2c/mP5000000<0^;cM00D000000;c0g@2l`=d0_<3M000000030<77gPL000000P3Jh=/00`00
0000h=W@0000000200000080k=W100@000000>;5^`000000k<fe0P0000001P3L^;@000000>:l/@3R
_;4000000<jX/`8000000P3F[[001@000000b::a0<RR/@38X[4000000080a9jb00@000000<BO/P00
000000000P3@Z;001@000000a9ja0<BN/@34W[40000000h0oooo00<0000005ilf00000001@000002
09fKd@030000002WXm40Yj?A0080Yj?A0P0000000`2_Zm@000000;BYcP040;BYcP8000003@3oool0
0`000000d;:k000000020=fk]P060000003_bJ`0klV/0000003K];000000103oool00`000000oooo
0?ooo`050?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0?P3oool00240oooo00<0
00000?ooo`3oool03P3oool3000000<0VLci1@00000309W<nA4000000`2Ic?T4000000<0VLci2@00
000309W;n@D000000`2JboT5000000<0Vl[h3000000309V>a`050000002bVl00/Y_00;:K`0000000
1@2mY<000`000000_Z_60;j[aP020;j[aP8000000`3D^l000`000000YIC3000000030:R?^`030000
002/S[L000000080WX>e00D000000:B4/P2TQ;80Y8Bb000000030:61/@040000002EN;00UGR`0000
0080IUBY00<000000000=P0000000P1jLKd00`000000VXg509Z=a@0209Z=a@050000002VWLd0YYg=
0:JMc@0000001@2hY<D300000080aJg200<000000=Joa00000000P000000102HOkL000000000002S
QK<200000080UWV`00@0000009Ii/02FNK0000000`2FN[400`000000Ug^c09Mk/`0209Mk/`030000
002SQkH0XhNf0080XhNf00D0000009R3^P2HPkX0V8>j000000030923`@050000002=R<T0SHS908f8
b@0000000`2?TM4300000080Y9k?00<000000:FUe@2UYMD0102UYMD00`000000[ZcE0:j/e@030:j/
e@<000001@2c/mP010000000^;KG0;Rfe`2h]]L=000000040<S5f0000000cm3J0<o@fP<000000P3P
fM03000000040>cI`@000000000000000080k<fe0P0000001P3L^;@000000>:l/@3R_;4000000<jX
/`8000000P3F[[001@000000b::a0<RR/@38X[4000000080a9jb00@000000<BO/P00000000000P3@
Z;001@000000a9ja0<BN/@34W[40000000l0oooo00<0000005ilf00000000P2;TM<00`000000WI_A
09fKd@0209fKd@030000002WXm40Yj?A00<0Yj?A00<000000;BYcP2dZLh0102dZLh010000000^ZS8
00000000000:0?ooo`040000003@/[/0d;:k00000080gK^f00H000000>o9[03_bJ`000000=^d/000
00040?ooo`030000003oool0oooo00D0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`0n0?ooo`008@3oool00`000000oooo0?ooo`0;0?ooo`<000000`2IboTA000000<0VLci1@000003
09W<nA0000000`2IboT5000000<0VL_i4@00000309[8m`@000000`2IS/L010000000/Y_00;:K`000
00040;fT`0030000002nZlH0_Z_600@0_Z_600@000000=Bk`03D^l0000000`2UU<<01@000000Z8nk
0:R?^`2XSk/0000000<0WX>e00D000000:B4/P2TQ;80Y8Bb000000030:61/@040000002EN;00UGR`
00000080IUBY00<000000000=P0000000`1jLKd010000000VXg509Z=a@2JSLD2000000030:JMc@00
0000^:C500D0^:C500H000000<F]`P35[L80aJg20000003F_l@2000000<0V7ng0P0000000`2SQK<0
00000000000209Ii/0040000002FNK00UWV`000000<0UWZa00<0000009Mk/`2GNk<00P2GNk<00`00
0000XhNf0:>7]P020:>7]P050000002HPkX0V8>j09R3^P0000000`2@Pl401P000000SHS908f8b@2=
R<T0000008nAd@80000000<0Y9S90000002TW/l00`2TW/l00`000000YJGE0:FUe@020:FUe@030000
002]Y]00000000@0[ZcE00<000000;>]dP2c[M800P0000020;>cf0030000002h]]L0^;KG00<0^;KG
00D000000;fieP2m^MH0_KWF000000030<:ne`0500000038aMP0b<GH0<S5f00000000P3?d=X20000
00030=C<d@000000h=W@00<000000P3/fL400`000000h/Fk000000020>c=]@040000003L^;@0g;Rd
00000080h[ba00@000000<jX/`00000000000P3F[[001@000000b::a0<RR/@38X[4000000080a9jb
00@000000<BO/P00000000000P3@Z;001@000000a9ja0<BN/@34W[4000000100oooo0P00000308^A
d`050000002MVm40WI_A09fKd@000000102WXm400`000000[9o:000000040;BYcP030000002jZ<P0
^ZS80080^ZS80`0000050?ooo`8000000P3@/[/01P000000gK^f0=fk]P3M^kH000000>o9[0800000
00<0fkB`0000003oool00P3oool2000000L0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d
0;ooo`0n0?ooo`008@3oool00`000000oooo0?ooo`080?ooo`l000000`2IboT5000000<0VL_i4@00
000309W;n@@000000`2IboTA000000<0VL_i1000000409W:n0@000000`2Hb?P01`000000VHk709V>
a`2IS/L000000;:K`0000000102mY<000`000000_Z_60;j[aP030;j[aP040000003D^l00e;_00=Bk
`0T000000`2NPkD01@000000Y8Bb0:B4/P2TQ;80000000<0XH6a00@0000009Eh/02EN;0000000P1V
E:T010000000000f0000=P00000207Ya_@040000002JSLD0VXg509Z=a@80000000<0YYg=0000002h
Y<D0102hY<D00`000000aJg20<F]`P020<F]`P050000003F_l@0000009j6^P0000000P2HOkL00`00
0000Rg>c0000000200000080UWV`00@0000009Ii/02FNK0000000`2FN[400`000000Ug^c09Mk/`02
09Mk/`030000002SQkH0XhNf0080XhNf00D0000009R3^P2HPkX0V8>j000000040923`@050000002=
R<T0SHS908f8b@000000102TV<T00`000000Y9k?0:BNc`020:BNc`040000002UYMD00000000000<0
[JK@0P0000000`2^[=D000000;>]dP040;>]dP<000000`2h]]L2000000<0_KWF00D000000<:ne`32
_]L0`[kG000000030<S5f0040000003?d=X0000000000080e<cA00H000000>3Id03PfM0000000=o=
aP0000020>cI`@030000003RaK/000000080k<fe00@000000=bh]03L^;@000000P3R_;4010000000
cZRc0000000000020=J^/00500000038X[40b::a0<RR/@0000000P34W[8010000000a9nb00000000
00020=2X/00500000034W[40a9ja0<BN/@000000403oool300000080Ri7C00<0000009fKd@2MVm40
0P2MVm4010000000Yj?A0000000000030:bObP050000002dZLh0]:W>0;BYcP000000102jZ<P00`00
0000`:W30<2Y``05000000050<N/_`000000d;:k0=2b^`0000000P3M^kH00`000000klV/0>o9[002
000000030=^d/0000000oooo00<0oooo00<000000?ooo`3oool01P3oool00`3OW4P004RL0=ooo`02
0?ooo`030;md0000M;l0oooo03h0oooo000Q0?ooo`030000003oool0oooo0080oooo2`00000309W;
n@@000000`2IboTA000000<0VL_i1@00000309W;nA0000000`2IboT5000000<0VL_i5000000209V>
a`800000102mY<000`000000_Z_60;j[aP040;j[aP040000003D^l00e;_00=Bk`0<000000P2@RLP0
10000000V8Bk09R4^`00000309j3]@050000002TQ;80Y8Bb0:B4/P0000000`2QPK4010000000UGR`
09Eh/000000206IDZ@040000000003H0000f00000080NW6m00L0000009Z=a@2JSLD0VXg50000002J
SLD0000000H0^:C500H000000<F]`P35[L80aJg20000003F_l@20000000309j6^P000000V7ng0080
V7ng1000000209Ii/0040000002FNK00UWV`000000<0UWZa00<0000009Mk/`2GNk<00P2GNk<00`00
0000XhNf0:>7]P020:>7]P050000002HPkX0V8>j09R3^P000000102@Pl401P000000SHS900000000
0000YIC3000000@0Y9S900<000000:BNc`2TW/l00P2TW/l00`000000[JK@0:fVd0050:fVd0030000
002c[M80/jgB00@0/jgB00<000000;R_d@2h[m40100000030;fieP0300000032_]L0`[kG0080`[kG
00<000000<S5f0000000100000030=C<d@070000003PfM0000000000003OcLH000000>cI`@020000
00030>;5^`000000k<fe008000000P3L^;@00`000000h[ba0>:l/@02000000050=://P000000eZj`
0=J^/00000000`38X[400`000000a9jb0<BN/P02000000050?77YP000000d:R`0=2X/00000000`34
W[400`000000oooo0?ooo`0?0?ooo`<000000P2;TM<010000000WI_A09fKd@2MVm42000000H0[9o:
00<000000;BYcP0000001@2jZ<P00`000000`:W30<2Y``020<2Y``0500000037[;l0ajbo0<N/_`00
00000P3@/[/010000000gK^f0=fk]P0000020>o9[080000000<0fkB`0000003oool00`3oool00`00
0000oooo0?ooo`060?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0o0?ooo`008@3oool0
0`000000oooo00000002000000<0VL_i6@00000309W;nAP000000`2IboT4000000<0VL_i4@000003
09O<nP@000000`2Fc?/400000080VHk71`0000050;j[aP030000003D^l00e;_00080e;_01@000000
1@2@RLP0000009R4^`2HQ;/000000080WX>e00D000000:B4/P2TQ;80Y8Bb000000030:61/@040000
002EN;00UGR`00000080IUBY00@000000000=P0003H000000`1jLKd01@000000VXg509Z=a@000000
VXg500800000102hY<D00`000000aJg20<F]`P020<F]`P030000003F_l@000000080WXJj00H00000
09Qo]`2HOkL0V7ng0000002;Lk<200000080UWV`00@0000009Ii/02FNK0000000`2FN[400`000000
Ug^c09Mk/`0209Mk/`800000102SQkH01@000000V8>j09R3^P2HPkX0000000@0T8?100D000000:FD
``2UU<<0YIC3000000050:BHb@050000002TW/l000000000002]XL/00P0000040:fVd0040000002c
Z<d0000000000080/jgB0P0000040;R_d@030000002l/]00_;;@00/000000`3<`]400`000000e<cA
0=C<d@04000000030=o=aP000000k=W10080000000@0h/Fk0000003/cKD0k<fe0P0000020=bh]003
0000003R_;40h[ba0080000000D0dZbb0000003F[[00eZj`000000020<RR/@0400000034W[80a9jb
0<BN/P<000000`3@Z;001@000000a9ja0<BN/@34W[4000000140oooo00<0000007Ikb`1fNl/01P00
00020:>Gb0030000002/WlX0[9o:00@0[9o:00<000000;ZXb02jZ<P00`2jZ<P00`000000`:W30<2Y
``020<2Y``0500000037[;l0ajbo0<N/_`0000000`3@/[/010000000gK^f0=fk]P0000020>o9[080
000000<0fkB`0000003oool00`3oool00`000000oooo0?ooo`070?ooo`050?oo_`1d000000000000
M02oool0@03oool00240oooo3`00000309W;n@@000000`2IboTA000000<0VL_i1000000309W;nAP0
00000`2Hc?TM00000080VY;;00<000000;j[aP2nZlH00P2nZlH00`000000e;_00=Bk`0020=Bk`0L0
00000P2HQ;/010000000WX>e09j3]@0000030:B4/P050000002QPK40XH6a0:61/@0000000P2EN;00
10000000IUBY06IDZ@0000020000=P050000001jLKd0NW6m07Ya_@0000000P2JSLD200000080]I^m
00D000000;RTa@2hY<D0^:C5000000040<F]`P<000000`2NQ[X01@000000V7ng09Qo]`000000Rg>c
008000000P2FNK0010000000UWV`09Ii/000000309Ij/@030000002GNk<0Ug^c0080Ug^c0P000004
0:>7]P050000002HPkX0V8>j09R3^P0000000`2@Pl42000000@0YIC300<000000:BHb@2TV<T00`2T
V<T00`000000[J7;0:fQb`030:fQb`80000000<0[JK@0000002cZ<d00`2cZ<d3000000@0^:oA00<0
00000;bbd02l/]000P2l/]000`000000`KO@0<6gd0020<6gd00500000036_=00a[c@0<Jld0000000
0P3<`]401@000000e<cA0=C<d@3Dc=4000000080e/C900`000000=o=aP00000000000=Zo_`000000
h/Fk0000003/cKD000000=Je^00000020=bh]0030000003R_;40h[ba0080000000D0dZbb0000003F
[[00eZj`000000020<RR/@0400000034W[80a9jb0<BN/P<000000`3@Z;001@000000a9ja0<BN/@34
W[4000000180oooo00<0000007Ikb`0000000`2DSLT00`000000XiO80:>Gb0020:>Gb0030000002/
WlX0[9o:0080[9o:00<000000;6Na0000000102jZ<P00`000000`:W30<2Y``020<2Y``0500000037
[;l0ajbo0<N/_`0000000`3@/[/010000000gK^f0=fk]P0000020>o9[0040000003K];00fkB`0000
00@0oooo00<000000?ooo`3oool0C03oool00240oooo00<000000?ooo`3oool0103oool5000000<0
VL_i4000000309W;n@D000000`2IboT@000000<0VLci1@00000309S<nA0000000`2Hc?X5000000<0
Ulcj2`00000309ZBb`030000002nZlH0_Z_60080_Z_600<000000=Bk`03D^l000P3D^l0400000080
Xlkf0P0000001@2HQ;/0000009j3]@2NPkD0000000<0Y8Bb00D000000:61/@2QPK40XH6a00000002
09Eh/0040000001VE:T0IUBY00000080000f00D0000007Ya_@1jLKd0NW6m0000000209Z=a@800000
0`2eVkd010000000^:C50;RTa@0000030<F]`P80000000<0e;Bj0000002NQ[X00P2NQ[X01@000000
V7ng09Qo]`000000Rg>c008000000P2FNK0010000000UWV`09Ii/000000309Ij/@070000002GNk<0
Ug^c09Mk/`000000/8nc000000040:>7]P050000002HPkX0V8>j09R3^P0000000P2@Pl4010000000
ZI6m0:VA_@0000030:FD``030000002TV<T0Y9S90080Y9S900<000000:fLaP0000001P2]XL/00`00
0000/jS=0;>Xc@030;>Xc@030000002gZ/`0]j[<008000000P2h[m401@000000_;;@0;bbd02l/]00
000000@0`KO@00D000000<Jld036_=00a[c@000000030<c2d@040000003Dc=40e<cA00000080e/C9
00@000000=o=aP00000000000P3J_kl01P000000h/Fk000000000000e[Fh00000080g;Rd00P00000
0>:l/@3R_;4000000<jX/`000000dZbb00000080eZj`00@000000<RR/@38X[4000000`34W[830000
00<0d:R`00D000000<BN/@34W[40a9ja0000000B0?ooo`040000001fNl/0MW_;00000080U8g900<0
00000:>Gb02SUlP00P2SUlP01@000000[9o:0:bObP2/WlX0000000<0/Ik400@000000;ZXb02jZ<P0
00001030ZL<00`000000ajbo0<N/_`020<N/_`050000003@/[/0d;:k0=2b^`0000000P3M^kH01000
0000klV/0>o9[00000020=^d/0030000003oool0oooo0080oooo00<000000?ooo`3oool0C03oool0
0240oooo00<000000?ooo`3oool02@3oool9000000<0VL_i1000000309W;nA4000000`2Ic?T40000
00<0VLci4@00000309S<n@@000000`2Hc?X@000000@0V<gj0`00000309ZBb`040000002nZlH0_Z_6
000000D0e;_000@000000:7<mP2Qc?H0XLcf100000001@2HQ;/0000009j3]@2NPkD0000000<0Y8Bb
00D000000:61/@2QPK40XH6a0000000209Eh/0040000001VE:T0IUBY00000080000f00<0000007Ya
_@1jLKd00P1jLKd00`000000VXg5000000050;FK_@8000001035[L800`000000e;Bj0=Bd^P020000
0080WXJj00D0000009Qo]`2HOkL0000008]c/`0200000080UWV`00@0000009Ii/02FNK0000000`2F
N[401`000000Ug^c09Mk/`2GNk<000000;2?/`000000102SQkH01@000000V8>j09R3^P2HPkX00000
0080T8?100@000000:VA_@2YTKd00000102UU<<010000000Y9S90:BHb@0000030:fLaP030000002]
XL/0[J7;0080[J7;00<000000;>Sb00000000`2cZ<d00`000000]j[<0;NZc0030;NZc0<000000`2l
/]000`000000`KO@0<6gd0020<6gd00500000036_=00a[c@0<Jld00000000P3<`]400`000000e<cA
0000000200000080e/C900<000000=o=aP0000000`3J_kl01P000000h/Fk000000000000e[Fh0000
0080g;Rd00<000000>:l/@3R_;400P0000000`3B[;8000000=J^/0020=J^/00400000038X[40b::a
000000<0a9jb0`0000030=2X/00500000034W[40a9ja0<BN/@0000004P3oool20000000307Ikb`00
0000U8g90080U8g900<000000:>Gb02SUlP00P2SUlP00`000000[9o:000000050;6Na0030000002j
Z<P0000000@0`:W300<000000<N/_`37[;l00P37[;l01@000000d;:k0=2b^`3@/[/000000080gK^f
00@000000>o9[03_bJ`000000P3K];000`000000oooo0?ooo`020?ooo`030000003oool0oooo04`0
oooo000Q0?ooo`030000003oool0oooo0180oooo4000000309W;n@D000000`2Ic?T@000000<0VLci
1@00000309S<nA0000000`2Ic?T4000000<0VLci1`00000409ZBb`030000002nZlH0000000D0e;_0
2@0000001@2HQ;/0000009j3]@2NPkD000000080Y8Bb00D000000:61/@2QPK40XH6a0000000209Eh
/0040000001VE:T0IUBY000000<0000f00L0000007Ya_@1jLKd0NW6m0000002JSLD0000000H0]I^m
0P0000020<F]`P80000000@0e;Bj0000002QS;h000000P2NQ[X02P000000V7ng09Qo]`000000Rg>c
0000002FN[40000009Ii/000000209Ii/0050000002FN[40UWZa09Ij/@0000000`2GNk<00`000000
/8nc000000040:>7]P030000002HPkX0V8>j0080V8>j0P0000040:VA_@070000002UU<<0YIC30:FD
``000000Y9S9000000040:fLaP050000002]XL/0[J7;0:fQb`0000000`2cXlP2000000030;>Xc@00
0000]j[<00@0]j[<00<000000;^]b`2k[L/0100000030<6gd00500000036_=00a[c@0<Jld0000000
0P3<`]43000000060<jmbP000000e/C90=K4b@000000glg60P0000020=Zo_`040000003RaK/00000
00000080e[Fh00@000000=bh]03L^;@000000P3R_;42000000030=://P000000eZj`0080eZj`00@0
00000<RR/@38X[4000000`34W[83000000<0d:R`00D000000<BN/@34W[40a9ja0000000C0?ooo`04
0000001fNl/0MW_;000000<0U8g900<000000:>Gb02SUlP00P2SUlP00`000000/Ik40;6Na0050;6N
a00300000030ZL<0`:W300<0`:W300<000000<N/_`37[;l00P37[;l010000000d;:k0=2b^`000003
0=fk]P040000003_bJ`0klV/00000080fkB`00@000000?ooo`3oool0oooo0P00001>0?ooo`008@3o
ool00`000000oooo0?ooo`0K0?ooo`@000000`2IboTA000000<0VLci1000000309W<nA0000000`2I
c?T5000000<0VLci400000000`2KboP0000009ZBb`0209ZBb`8000001P3D^l0600000080XLGb00<0
000009R4^`0000000P2NPkD010000000Y8Bb0:B4/P0000030:61/@040000002EN;00UGR`00000080
IUBY00D000000000=P0003H0000f0000000307Ya_@030000002JSLD0000000H0]I^m00<000000<>U
^`0000000P0000020=Bd^P030000002QS;h0000000<0WXJj00<0000009Qo]`2HOkL00P000000102F
N[40000009Ii/000000309Ii/0040000002FN[40UWZa000000<0Ug^c00<000000;2?/`000000102S
QkH00`000000V8>j09R3^P0209R3^P030000002YTKd0ZI6m00<0ZI6m00<000000:FD``2UU<<00P2U
U<<00`000000[Ic60:fLaP040:fLaP030000002]XL/0000000H0/j?800<000000;NZc02gZ/`00`2g
Z/`00`000000^jg;0;^]b`020;^]b`030000002o/</0_k3;00/000000`3>_LX00`000000e/C90=K4
b@03000000<0f[no00@000000>;5^`00000000000P3F]KP010000000g;Rd0=bh]00000020>:l/@80
000000<0dZbb0000003F[[000P3F[[0010000000b::a0<RR/@0000030<BN/P<000000`3@Z;001@00
0000a9ja0<BN/@34W[40000001<0oooo0P0000000`1fNl/0000009B=b@0309B=b@060000002SUlP0
XiO80000002XU<8000001P2aW/@2000000<0`:W300<000000<N/_`37[;l00P37[;l01@000000d;:k
0=2b^`3@/[/0000000<0gK^f00@000000>o9[000000000000P3K];000`000000oooo0?ooo`020?oo
o`030000003oool0oooo04`0oooo000Q0?ooo`030000003oool0oooo01l0oooo2@00000309W<n@D0
00000`2Ic?T@000000<0VLci1000000309W<nA4000000`2IboT:00000080VY;;0P0000060=Bk`0<0
00000`2Na_@500000080WX>e00D000000:B4/P2TQ;80Y8Bb000000020:61/@040000002EN;00UGR`
00000080IUBY00D000000000=P0003H0000f0000000407Ya_@8000001@2eVkd00`000000`jFk0<>U
^`020<>U^`040000003D];X0e;Bj00000080XHbn00@0000009j6^P2NQ[X000000P2HOkL01P000000
Rg>c000000000000UWV`000000<0UWV`00@0000009Ij/@2FN[4000000`2GNk<00`000000/8nc0000
00050:>7]P050000002HPkX0V8>j09R3^P0000001P2YTKd01P000000YIC30:FD``000000[iW10000
00H0[Ic600<000000;>Sb02cXlP0102cXlP010000000]jK80000000000020;NZc0030000002k[L/0
^jg;00<0^jg;00D000000;n`b`2o/</0_k3;000000040<Bdb`0500000038^</0b;S;0<Rhb`000000
0P3>_LX00`000000e/C90=K4b@02000000030=>j`P000000f[no0080f[no00<000000>;5^`000000
0`3F]KP010000000g;Rd0=bh]00000020>:l/@80000000<0dZbb0000003F[[000P3F[[0010000000
b::a0<RR/@0000030<BN/P<000000`3@Z;001@000000a9ja0<BN/@34W[40000001@0oooo0P000000
0`1fNl/0000009B=b@0209B=b@030000002SUlP000000080Z9C200<000000;6Na02aW/@00`2aW/@0
10000000]Yjo0;JN_`0000020<2Y``0300000037[;l0ajbo0080ajbo00D000000=2b^`3@/[/0d;:k
000000030=fk]P040000003_bJ`0000000000080fkB`00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`1<0?ooo`008@3oool00`000000oooo0?ooo`0X0?oooa4000000`2Ic?TG000000<0VL_i
4000000309[;n0@0000000<0Uh6j0000003D^l001@3D^l000`000000W<Sf09c8mP050000000609g5
m0000000V8Bk0000002NPkD000000`2TQ;8010000000XH6a0:61/@00000209Eh/0040000001VE:T0
IUBY000000<0000f00<0000007Ya_@1jLKd00P1jLKd2000000D0]I^m00<000000<>U^`33YK/00P33
YK/010000000e;Bj0000000000020:6<_P040000002NQ[X0WXJj00000080V7ng00H0000008]c/`00
0000000009Ii/000000309Ii/0050000002FN[40UWZa09Ij/@0000000P2GNk<00`000000/8nc0000
00050:>7]P060000002HPkX0V8>j0000002^T[T000001@2YTKd00`000000YIC3000000030:nI`@03
0000002]W<H0[Ic60080[Ic600<000000;>Oa00000001@2cXlP010000000]jK80;NVb02gY/P30000
00@0^jg;00<000000;n`b`2o/</00P2o/</00`000000a;C;0<Bdb`020<Bdb`0400000038^</0b;S;
000000<0c[g:00@000000=K4b@3Fa<T000000P3C^/801@000000f[no0=Zo_`000000h/Fk00800000
0`3F]KP010000000g;Rd0=bh]00000020>:l/@040000003B[;80dZbb000000<0eZj`00@000000<RR
/@38X[4000000`34W[83000000<0d:R`00D000000<BN/@34W[40a9ja0000000D0?ooo`@000000`2D
SLT2000000@0Z9C200<000000;6Na02aW/@00P2aW/@010000000]Yjo0;JN_`2fW[l2000000D0ajbo
00D000000=2b^`3@/[/0d;:k000000030=fk]P80000000D0d:^c0000003K];00fkB`000000040?oo
o`030000003oool0oooo04`0oooo000Q0?ooo`030000003oool0oooo0340oooo4000000309W;n@D0
00000`2IboT@000000<0VL_i1000000309W;n@H000000`2GPKX00`000000e;_00=Bk`0020=Bk`0d0
000000<0WX>e0000002TQ;800P2TQ;8010000000XH6a0:61/@00000209Eh/0040000001VE:T0IUBY
000000<0000f00<0000007Ya_@1jLKd00P1jLKd2000000D0]I^m00D000000<>U^`33YK/0`jFk0000
00020=Bd^P@0000000P0XHbn0000002NQ[X0WXJj0000002HOkL0000008]c/`80000000<0UWV`0000
002FNK000P2FNK001@000000UWZa09Ij/@2FN[4000000080Ug^c00@000000;2?/`2`Sk<00000102S
QkH01P000000V8>j09R3^P000000[Y:i000000H0ZI6m0P0000040:nI`@040000002]W<H0[Ic60000
00<0/io40P0000020;>Sb0030000002gY/P0]jK80080]jK800<000000;ZYa`2jZLL00P0000020;^]
b`050000002o/</0_k3;0;n`b`0000001034]</01@000000b;S;0<Rhb`38^</000000080c[g:00<0
00000=K4b@0000000P0000020=>j`P050000003J_kl0f[no0000003RaK/00P0000020=Je^0050000
003L^;@0g;Rd0=bh]00000000P3R_;4010000000dZbb0=://P0000030=J^/00400000038X[40b::a
000000<0a9jb00<000000?77YP0000000`3@Z;001@000000a9ja0<BN/@34W[40000001@0oooo0P00
000208El`@<0000000<0Vh_10000002XU<80102XU<8010000000/Ik40;6Na00000050;JN_`030000
0037[;l0ajbo00<0ajbo00D000000=2b^`3@/[/0d;:k000000020=fk]P050000003_bJ`000000=2[
/`0000000P3K];000`000000oooo0?ooo`020?ooo`030000003oool0oooo04`0oooo000Q0?ooo`03
0000003oool0oooo03X0oooo1000000309W;nA4000000`2IboT4000000<0VL_i4000000309N1^P03
0000003D^l00e;_00080e;_000<0000009S:n@2Hb_T01000000309S9n0<0000000<0WX>e0000002T
Q;800P2TQ;8010000000XH6a0:61/@00000209Eh/0040000001VE:T0IUBY000000<0000f00L00000
07Ya_@1jLKd0NW6m0000002aTkP0000000D0]I^m00D000000<>U^`33YK/0`jFk000000020=Bd^P80
000000@0Thg:00000000000000000P2NQ[X010000000V7ng0000002;Lk<20000000309Ii/0000000
UWV`0080UWV`00D0000009Ij/@2FN[40UWZa0000000209Mk/`040000002`Sk<0/8nc000000@0XhNf
00@0000009R3^P2HPkX000000P2^T[T00`000000ZI6m0:VA_@030:VA_@030000002_VL40[iW100@0
[iW100<000000:fLaP0000001@2cWl@00`000000/j?8000000040;NVb0030000002jZLL0^ZW70080
^ZW70`0000030;n`b`0500000034]</0a;C;0<Bdb`0000000`38^</00`000000c[g:0<jmbP030000
00050<jga0000000dk[20=>j`P0000000`3J_kl2000000050=:b^P000000e[Fh0=Je^00000000`3L
^;@010000000h[ba0>:l/@0000020=://P040000003F[[00eZj`000000<0b::a00L000000<BN/P34
W[80a9jb0000003aajH0000000<0d:R`00D000000<BN/@34W[40a9ja0000000E0?ooo`80000000<0
QGc10000002KRl400`2KRl400`000000Z9C20:RD`P030:RD`P030000002aW/@0000000D0]Yjo0P00
00040<N/_`050000003@/[/0d;:k0=2b^`0000000P3M^kH01@000000klV/0000003@Zk<000000080
fkB`00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1<0?ooo`008@3oool00`000000oooo
0?ooo`0n0?ooo`T000000`2IboTH000000<0VL_i1000000309S;n@040000002GPKX0Uh6j000000@0
e;_03@0000000`2NPkD000000:B4/P020:B4/P040000002QPK40XH6a00000080UGR`00@0000006ID
Z@1VE:T00000100003H01P000000NW6m07Ya_@000000/I>h000000@0]I^m00<000000<>U^`33YK/0
0P33YK/00`000000e;Bj0=Bd^P020000000:09>=bP000000SHC508f4a@000000WXJj0000002HOkL0
000008]c/`80000000<0UWV`0000002FNK000P2FNK001@000000UWZa09Ij/@2FN[4000000080Ug^c
00@000000;2?/`2`Sk<00000102SQkH00`000000V8>j000000030:jB^@030000002YTKd0ZI6m00<0
ZI6m0P0000060:nI`@030000002cWl@0/io400D0/io400<000000;NVb02gY/P00P2gY/P00`000000
^ZW70;ZYa`030;ZYa`030000002nZlL0_Z_700@000000`34]</00`000000b;S;0000000400000005
0<Vca@000000c[O40<jga00000000P3C^/801`000000f[no0=Zo_`3J_kl000000=:b^P0000000`3F
]KP01P000000g;Rd0=bh]03L^;@000000>:l/@8000000P3B[;8010000000eZj`0=J^/00000030<RR
/@0300000034W[80a9jb0080000000<0lLNV0000003@Z;000P3@Z;001@000000a9ja0<BN/@34W[40
000001D0oooo0P00000208El`@050000002KRl40Vh_109^;`@0000001@2XU<82000000D0]Yjo00<0
00000;^O^`2kWk/00P0000000`37[;l000000=2b^`030=2b^`030000003M^kH0gK^f008000000P3@
Zk<010000000fkB`0=^d/00000030?ooo`800000CP3oool00240oooo00<000000?ooo`3oool0A`3o
ool9000000<0VL_i1000000309W;nA<000000P2GPKX00`000000e;_00=Bk`0020=Bk`0D000000`2E
c?X60000000509j3]@000000Y8Bb0:B4/P0000000P2QPK4010000000UGR`09Eh/000000206IDZ@03
0000000003H0000f0080000f00H0000007Ya_@1jLKd000000;6C^00000040;FK_@0300000033YK/0
`jFk0080`jFk00<000000=Bd^P0000000P00000209>=bP030000002=Q<D0000000<0000000H0V7ng
000000000000000009Ii/000000309Ii/0050000002FN[40UWZa09Ij/@0000000P2GNk<010000000
/8nc0;2?/`0000040:>7]P030000002HPkX0000000<0[Y:i00<000000:VA_@2YTKd00P2YTKd00`00
0000/YRm000000050:nI`@030000002dWL40000000D0/io400@000000;NSa000000000000P2gY/P0
0`000000^ZW70;ZYa`030;ZYa`050000002nZlL0_Z_70;j[a`0000000P31[/L5000000030<FaaP00
0000bK?50080bK?500D000000<jga03>]l@0c[O4000000020=>j`P030000003J_kl0f[no00800000
00<0d[:j0000003F]KP00P3F]KP01P000000g;Rd0=bh]03L^;@000000>:l/@8000000P3B[;801000
0000eZj`0=J^/00000030<RR/@0300000034W[80a9jb0080000000<0lLNV0000003@Z;000P3@Z;00
1@000000a9ja0<BN/@34W[40000001D0oooo0P00000208El`@030000002KRl40Vh_10080Vh_100<0
00000:RD`P2XU<800`2XU<800`000000]Yjo0;JN_`030;JN_`030000002kWk/0^ink0080^ink0P00
00030=2b^`040000003M^kH0gK^f0=fk]P8000000P3@Zk<010000000fkB`0=^d/00000040?ooo`03
0000003oool0oooo04`0oooo000Q0?ooo`030000003oool0oooo0500oooo4000000309S<n@@00000
0`2Hc?X3000000<0Uh6j00@000000=Bk`03D^l00e;_03P0000001@2NPkD000000:B4/P2TQ;800000
0080XH6a00D0000009Eh/02EN;00UGR`0000000206IDZ@050000000003H0000f0000=P0000000P1j
LKd010000000/I>h0000000000030;FK_@0300000033YK/0`jFk0080`jFk00<000000=Bd^P000000
0`000000202CSLX0000008f4a@2=Q<D0000007ig`@000000V7ng0`0000000`2FNK00000009Ii/002
09Ii/0050000002FN[40UWZa09Ij/@0000000P2GNk<010000000/8nc0;2?/`0000050:>7]P800000
102^T[T01@000000ZI6m0:VA_@2YTKd000000080/YRm00<000000:nI`@2_VL400P2_VL4010000000
]9g10;BM`@0000030;>Oa0030000002gXl@0]j?40080]j?40P0000050;ZYa`030000002nZlL0_Z_7
0080_Z_700D000000<6^a`31[/L0`Jk7000000040<FaaP0500000039/lD0bK?50<Vca@0000000P3>
]l@01@000000dk[20=>j`P3C^/8000000080f[no00@000000=:b^P3B/[X000000`3F]KP020000000
g;Rd0=bh]0000000h[ba0000003?Z[<000000P3B[;8010000000eZj`0=J^/00000030<RR/@030000
0034W[80a9jb0080000000<0lLNV0000003@Z;000P3@Z;001@000000a9ja0<BN/@34W[40000001H0
oooo0P0000000`25O<40000009^;`@0309^;`@030000002XU<80Z9C20080Z9C20P0000050;JN_`03
0000002kWk/0^ink00<0^ink00D000000=2b^`3@/[/0d;:k000000030=fk]P040000003@Zk<0d:^c
000000<0fkB`00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1<0?ooo`008@3oool00`00
0000oooo0?ooo`1I0?oooa@000000`2GPKX010000000e;_00=Bk`03D^l0>0000000509j3]@000000
Y8Bb0:B4/P0000000`2QPK4010000000UGR`09Eh/000000206IDZ@070000000003H0000f0000=P00
0000NW6m000000020;6C^08000000P2eVkd00`000000`jFk0<>U^`020<>U^`030000003D];X0e;Bj
00D0000000L0Thg:0000002=Q<D0000007ig`@000000V7ng00<0000000<0UWV`0000002FNK000P2F
NK001`000000UWZa09Ij/@2FN[40000009Mk/`0000000`2`Sk<00`000000XhNf0:>7]P030:>7]P03
0000002^T[T0[Y:i00<0[Y:i00@000000:VA_@2YTKd000000`2bV;d01@000000[iW10:nI`@2_VL40
000000<0]9g100D000000;>Oa02cWl@0/io4000000040;NSa0030000002jYL@0000000<0^ZW700<0
00000;j[a`2nZlL00P2nZlL00`000000`Jk70<6^a`020<6^a`0500000035/LH0aK760<FaaP000000
0`39/lD01@000000c[O40<jga03>]l@000000080dk[200<000000=Zo_`3J_kl00P0000020=:b^P05
0000003F]KP0e[Fh0=Je^00000000P3L^;@01@000000h[ba0000003?Z[<000000080dZbb00@00000
0=J^/03F[[0000000`38X[400`000000a9jb0<BN/P02000000030?77YP000000d:R`0080d:R`00D0
00000<BN/@34W[40a9ja0000000F0?ooo`8000000P25O<401@000000Vh_109^;`@2KRl40000000@0
Z9C200<000000:fD_@000000102fW[l00`000000^ink0;^O^`030;^O^`<0000000@0d;:k0000003M
^kH0gK^f0P0000020=2[/`050000003K];00fkB`0=^d/0000000103oool00`000000oooo0?ooo`1<
0?ooo`008@3oool00`000000oooo0?ooo`1Q0?ooo`D000000`2Hc?X500000080Uh6j00@000000=Bk
`03D^l00e;_01000000409O=n`P000000P2TQ;801@000000XH6a0:61/@2QPK4000000080UGR`00@0
000006IDZ@1VE:T000000`0003H00`000000NW6m000000020;6C^0050000002aTkP000000;FK_@00
00001033YK/00`000000e;Bj0000000700000080SHC500<0000007ig`@0000000`0000000`2FNK00
000009Ii/00209Ii/0070000002FN[40UWZa09Ij/@000000Ug^c000000030;2?/`030000002SQkH0
XhNf00<0XhNf00<000000:jB^@2^T[T0102^T[T00`000000ZI6m000000040;:H_@030000002_VL40
000000D0]9g100<000000;>Oa00000001@2gXl@00`000000^ZG40;ZUa002000000030;ZYa`000000
_Z_700<0_Z_700D000000<6^a`31[/L0`Jk7000000040<FaaP0500000039/lD0bK?50<Vca@000000
0`3>]l@00`000000dk[20=>j`P03000000050<j`_0000000d[:j0=:b^P0000000P3F]KP010000000
g;Rd0=bh]03L^;@200000080cjZc00@000000=://P3B[;8000000P3F[[001@000000b::a0<RR/@38
X[4000000080a9jb0P0000000`3aajH000000=2X/0020=2X/00500000034W[40a9ja0<BN/@000000
5P3oool200000080QGc100<0000009^;`@2KRl400P2KRl4010000000Z9C20:RD`P0000020:fD_@05
0000002fW[l0]Yjo0;JN_`0000001P2kWk/00`000000`J6i0<6Q^@04000000050<RU]P000000d:^c
0=2[/`0000000`3K];000`000000oooo0?ooo`020?ooo`030000003oool0oooo04`0oooo000Q0?oo
o`030000003oool0oooo06H0oooo2000000209N1^P040000003D^l00e;_00=Bk`0d0000000<0V<kj
0000000000000P2TQ;801@000000XH6a0:61/@2QPK4000000080UGR`00@0000006IDZ@1VE:T00000
0`0003H00`000000NW6m000000020;6C^0050000002aTkP000000;FK_@0000001033YK/00`000000
e;Bj000000080000000308f4a@000000OWO100@0000000<0UWV`0000002FNK000P2FNK001`000000
UWZa09Ij/@2FN[40000009Mk/`0000000`2`Sk<00`000000XhNf0:>7]P030:>7]P030000002^T[T0
[Y:i00@0[Y:i00<000000:VA_@000000102bV;d00`000000[iW1000000060;BM`@030000002gXl@0
]j?400<0]j?400<000000;ZUa02jYL@00`2jYL@300000080_Z_700<000000<6^a`31[/L00P31[/L0
1@000000aK760<FaaP35/LH0000000<0bK?500<000000<jga03>]l@00P3>]l@010000000dk[20=>j
`P0000020<j`_0050000003B/[X0d[:j0=:b^P0000000P3F]KP010000000g;Rd0=bh]03L^;@20000
0080cjZc00@000000=://P3B[;8000000P3F[[001@000000b::a0<RR/@38X[4000000080a9jb0P00
00001@3aajH000000=2X/03@Z;00000000<0a9ja00<000000?ooo`3oool05@3oool300000080QGc1
00L0000009^;`@2KRl40Vh_10000002XU<80000000@0[IBm00@000000;JN_`2fW[l000001@2kWk/0
0`000000`J6i0<6Q^@020<6Q^@0500000038YKH0b:Ff0<RU]P0000000P3@Zk<01@000000fkB`0=^d
/03K];00000000@0oooo00<000000?ooo`3oool0C03oool00240oooo00<000000?ooo`3oool0K@3o
ool01@000000Uh6j09N1^P2GPKX000000080e;_00`00000309W<n@X000000P2TQ;801@000000XH6a
0:61/@2QPK4000000080UGR`00@0000006IDZ@1VE:T00000100003H200000080/I>h00D000000;6C
^0000000]I^m000000040<>U^`80000000@0d:fe000000000000/l?V1P0000000`2=Q<D000000000
00030000000309Ii/0000000UWV`0080UWV`00L0000009Ij/@2FN[40UWZa0000002GNk<0000000@0
/8nc00<000000:>7]P2SQkH00P2SQkH2000000H0[Y:i00<000000;:H_@2bV;d0102bV;d00`000000
]9g10;BM`@040;BM`@040000002hXL400000000000<0]j?400<000000;ZUa02jYL@00P2jYL@01000
0000_JS40;fXa02mZ<@400000080`Jk700<000000<FaaP35/LH00P35/LH01@000000bK?50<Vca@39
/lD0000000<0c[O400<000000=>j`P0000000P0000020<j`_0040000003B/[X0d[:j000000<0e[Fh
00D000000=bh]03L^;@0g;Rd000000020<nZ/`050000003B[;80dZbb0=://P0000000P3F[[001@00
0000b::a0<RR/@38X[4000000080a9jb00@000000?77YP3aajH000000P3@Z;001@000000a9ja0<BN
/@34W[40000001P0oooo0P00000208El`@030000002KRl40Vh_10080Vh_10P0000040:fD_@040000
002fW[l0]Yjo000000D0^ink00<000000<6Q^@31XKT00P31XKT01@000000b:Ff0<RU]P38YKH00000
0080d:^c00D000000=^d/03K];00fkB`000000040?ooo`030000003oool0oooo04`0oooo000Q0?oo
o`030000003oool0oooo06d0oooo00D0000009N1^P2GPKX0Uh6j000000020=Bk`0030000003oool0
oooo00P000000`2Kc?T200000080Y8Bb00D000000:61/@2QPK40XH6a0000000209Eh/0040000001V
E:T0IUBY000000@0000f00D000000;6C^02aTkP0/I>h000000020;6C^00300000033YK/0`jFk00<0
`jFk1`0000030;Fjg@@0000000H0OWO1000000000000000009Ii/000000309Ii/0070000002FN[40
UWZa09Ij/@000000Ug^c000000040;2?/`030000002SQkH0XhNf0080XhNf0P0000060:jB^@030000
002bV;d0/YRm00@0/YRm0P0000050;BM`@040000002hXL40^:710;RQ`@<000001@2jYL@00`000000
_JS40;fXa0020;fXa0030000002oZL@0_jW400@000000P35/LH01@000000bK?50<Vca@39/lD00000
0080c[O4100000020<Z__`040000003>/;`0c[2l00000080d[:j00D000000=Je^03F]KP0e[Fh0000
00030=bh]0040000003?Z[<0cjZc000000<0dZbb00@000000=J^/03F[[0000000`38X[4010000000
a9jb0<BN/P0000020?77YP040000003@Z;00d:R`000000<0a9ja00<000000?ooo`3oool05P3oool3
0000000308El`@000000Vh_100<0Vh_100<000000:fD_@2]U;d0102]U;d00`000000]Yjo00000005
0;^O^`0300000031XKT0`J6i0080`J6i00D000000<RU]P38YKH0b:Ff000000020=2[/`050000003K
];00fkB`0=^d/00000000`3oool2000004h0oooo000Q0?ooo`030000003oool0oooo06d0oooo00D0
000009N1^P2GPKX0Uh6j000000020=Bk`0030000003oool0oooo00P0oooo1@0000020:B4/P050000
002QPK40XH6a0:61/@0000000P2EN;0010000000IUBY06IDZ@0000040000=P050000002aTkP0/I>h
0;6C^00000000P2aTkP00`000000`jFk0<>U^`030<>U^`030000003@[KD00000008000000P2`]]d3
00000080]K7D1P0000000`2FNK00000009Ii/00209Ii/0070000002FN[40UWZa09Ij/@000000Ug^c
000000040;2?/`070000002SQkH0XhNf0:>7]P000000]IBf000000060:jB^@800000102bV;d01000
0000]Ybm0;JL_@0000030;BM`@030000002hXL40^:7100@0^:7100<000000;ZUa02jYL@00P2jYL@0
0`000000_JS40;fXa0030;fXa0030000002oZL@0_jW40080_jW400<000000<:[``00000020000002
0<N^`@040000003:[kl0bZno000000<0c[2l00@000000=:b^P3B/[X000000`3F]KP00`000000g;Rd
0=bh]00200000080cjZc00D000000=://P3B[;80dZbb000000020=J^/00500000038X[40b::a0<RR
/@0000000P34W[8010000000lLNV0?77YP0000020=2X/00500000034W[40a9ja0<BN/@000000603o
ool01`000000JVBi000000000000QGc10000002KRl400`0000060:fD_@8000001P2kWk/00`000000
`J6i0<6Q^@020<6Q^@0500000038YKH0b:Ff0<RU]P0000000P3@Zk<010000000fkB`0=^d/0000005
0?ooo`030000003oool0oooo04`0oooo000Q0?ooo`030000003oool0oooo06h0oooo00@0000009N1
^P2GPKX0Uh6j0P00000?0?ooo`80000000<0Y8Bb0000002QPK400P2QPK4010000000UGR`09Eh/000
000206IDZ@030000000003H0000f0080000f00D000000;6C^02aTkP0/I>h000000020;6C^0<00000
0P33YK/2000000040=2]]@000000000000000080/;KM0P0000000`2`[]H00000000000020;Fae0@0
00000P2FNK0010000000UWV`09Ii/000000309Ij/@030000002GNk<0000000@0/8nc00L000000:>7
]P2SQkH0XhNf0000002eU;H0000000D0[Y:i00<000000;JH^@000000102bV;d01@000000]Ybm0;JL
_@2fW;d000000080]9g100<000000;RQ`@2hXL400`2hXL400`000000^Z?2000000030;ZUa0030000
002mZ<@0_JS40080_JS400<000000;nYa02oZL@00P2oZL@00`000000`Z_30<:[``020<:[``050000
0034[<80a:c20<B/`P0000000`37[/401@000000bZno0<Z__`3:[kl0000000<0c[2l00@000000=:b
^P3B/[X000000`3F]KP01@000000g;Rd0000003<Z;D000000080cjZc00D000000=://P3B[;80dZbb
000000020=J^/00500000038X[40b::a0<RR/@0000000P34W[8010000000lLNV0?77YP0000020=2X
/00500000034W[40a9ja0<BN/@000000603oool010000000JVBi0000002AO[/3000000<0XX^l00<0
00000:fD_@2]U;d0102]U;d00`000000^ink0;^O^`030;^O^`0300000031XKT0`J6i0080`J6i00D0
00000<RU]P38YKH0b:Ff000000030=2[/`040000003K];00fkB`000000D0oooo00<000000?ooo`3o
ool0C03oool00240oooo00<000000?ooo`3oool0KP3oool010000000Uh6j09N1^P2GPKX2000000l0
oooo0P0000020:B4/P040000002QPK40XH6a00000080UGR`00@0000006IDZ@1VE:T00000100003H0
1@000000/I>h0;6C^02aTkP000000080/I>h00<000000<6O]P31WkH00`0000020=2]]@D000001@2`
[]H2000000040;BXc@000000000000000080UWV`00@0000009Ii/02FNK0000000`2FN[400`000000
Ug^c000000040;2?/`050000002SQkH0XhNf0:>7]P0000000P2eU;H00`000000[Y:i0:jB^@020:jB
^@040000002fV;T0]YRi00000080/YRm00<000000;JL_@2fW;d00`2fW;d2000000H0^:7100<00000
0;ZS`P2jXl800`0000050;fXa0030000002oZL@0_jW40080_jW400D000000<:[``32Zl<0`Z_30000
00030<B/`P0300000037[/40ajk10080ajk100D000000<Z__`3:[kl0bZno000000020<j`_0050000
003B/[X0d[:j0=:b^P0000000P3F]KP00`000000g;Rd000000020<bX]@040000003?Z[<0cjZc0000
00<0dZbb00@000000=J^/03F[[0000000`38X[4010000000a9jb0<BN/P0000020?77YP040000003@
Z;00d:R`000000<0a9ja00<000000?ooo`3oool05P3oool300000080TGjk00<000000::;_02RRk`0
0P2RRk`00`000000[IBm0:fD_@040:fD_@030000002kWk/0^ink00<0^ink00<000000<6Q^@31XKT0
0P31XKT01@000000b:Ff0<RU]P38YKH0000000<0d:^c00@000000=^d/03K];0000001@3oool00`00
0000oooo0?ooo`1<0?ooo`008@3oool00`000000oooo0?ooo`1^0?ooo`040000002GPKX0Uh6j09N1
^P800000403oool010000000Y8Bb0:B4/P0000020:61/@040000002EN;00UGR`00000080IUBY00L0
00000000=P0003H0000f0000002`Sk@000000080/I>h00@000000;6C^02aTkP000001031WkH00`00
0000d:fe0=2]]@0300000080ZZkJ0P0000020;2^eP80000000<0[ZK@0000000000000P00000209Ii
/0040000002FNK00UWV`000000<0UWZa00<0000009Mk/`000000102`Sk<01@000000XhNf0:>7]P2S
QkH000000080]IBf00<000000:jB^@2^T[T00P2^T[T010000000]YRi0;JH^@0000020;:H_@030000
002fW;d0]Ybm00@0]Ybm00<000000;RQ`@2hXL400`2hXL400`000000^Z?20;ZS`P030;ZS`P030000
002mZ<@0_JS40080_JS400<000000;nYa02oZL@00P2oZL@00`000000`Z_30<:[``020<:[``050000
0034[<80a:c20<B/`P0000001037[/401@000000bZno0<Z__`3:[kl000000080c[2l00D000000=:b
^P3B/[X0d[:j000000020=Je^0030000003L^;@000000080c:Re00@000000<nZ/`3?Z[<000000`3B
[;8010000000eZj`0=J^/00000030<RR/@0400000034W[80a9jb00000080lLNV00@000000=2X/03@
Z;0000000`34W[400`000000oooo0?ooo`0F0?ooo`80000000D0JVBi0000002AO[/0TGjk00000003
0::;_0030000002]U;d0[IBm00<0[IBm00<000000;6E^P000000102kWk/00`000000`J6i0<6Q^@02
0<6Q^@0500000038YKH0b:Ff0<RU]P0000000`3@Zk<010000000fkB`0=^d/00000050?ooo`030000
003oool0oooo00L0oooo00D0omnL04P000000000001d0;ooo`030?ooo`060?noM0000000M;oo0?oo
o`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`0a0?ooo`008@3oool00`000000oooo0?ooo`1^0?oo
o`040000002GPKX0Uh6j09N1^P800000403oool010000000Y8Bb0:B4/P0000020:61/@040000002E
N;00UGR`00000080IUBY00L000000000=P0003H0000f0000002`Sk@000000080/I>h00<000000;6C
^00000001@31WkH00`000000d:fe0=2]]@04000000<0ZZkJ0P0000050:jVd08000000P2FNK001000
0000UWV`09Ii/000000309Ij/@030000002GNk<0000000@0/8nc00D000000:>7]P2SQkH0XhNf0000
00020;FD]P050000002^T[T0[Y:i0:jB^@000000102fV;T00`000000/YRm000000050;JL_@030000
002iX;l0000000@0^:7100<000000;ZS`P2jXl800P2jXl8010000000_:G20000000000020;fXa003
0000002oZL@0_jW40080_jW400D000000<:[``32Zl<0`Z_3000000040<B/`P0500000037[/40ajk1
0<N^`@0000000`3:[kl01@000000c[2l0<j`_03>/;`0000000<0d[:j00<000000=Je^03F]KP00P00
00020<bX]@050000003?Z[<0cjZc0<nZ/`0000000`3B[;8010000000eZj`0=J^/00000030<RR/@04
00000034W[80a9jb00000080lLNV00@000000=2X/03@Z;0000000`34W[400`000000oooo0?ooo`0G
0?ooo`030000001ZI;T000000080TGjk00<000000::;_02RRk`00P2RRk`00`000000[IBm0:fD_@02
0:fD_@030000002aUKX0000000@0^ink00<000000<6Q^@31XKT00P31XKT01@000000b:Ff0<RU]P38
YKH0000000<0d:^c00@000000=^d/03K];0000001@3oool00`000000oooo0?ooo`060?ooo`070?oo
_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo0?ooo`3ooml0W4Qd0;oo
o`020?ooo`030;md001d_ol0oooo0340oooo000Q0?ooo`030000003oool0oooo06l0oooo00D00000
09N1^P2GPKX0Uh6j0000000@0?ooo`040000002TQ;80Y8Bb00000080XH6a00@0000009Eh/02EN;00
00000P1VE:T01`000000000f0000=P0003H000000;2?]00000000P2aTkP00`000000/I>h00000005
0<6O]P030000003@[KD0d:fe008000000P2U/>0300000080Z:KE0P0000020:jVd0<000000P2FNK00
10000000UWV`09Ii/000000409Ij/@8000001@2`Sk<010000000XhNf0:>7]P0000030;FD]P040000
002^T[T0[Y:i000000@0]YRi0P0000060;JL_@040000002iX;l0^J2o00000080^:7100<000000;ZS
`P2jXl800`2jXl8010000000_:G20;bU`P2lYL8400000080_jW400<000000<:[``32Zl<00P32Zl<0
1@000000a:c20<B/`P34[<80000000@0ajk100D000000<Z__`3:[kl0bZno000000030<j`_0040000
003B/[X0d[:j000000<0e[Fh00D000000<bX]@3<Z;D0c:Re000000030<nZ/`040000003B[;80dZbb
000000<0eZj`00H000000<RR/@38X[40b::a00000034W[8200000080lLNV00@000000=2X/03@Z;00
00000`34W[400`000000oooo0?ooo`0G0?ooo`<000000P2AO[/00`000000XX^l0::;_0020::;_005
0000002]U;d0[IBm0:fD_@0000000`2aUKX010000000^ink0;^O^`0000050<6Q^@0500000038YKH0
b:Ff0<RU]P0000000`3@Zk<010000000fkB`0=^d/00000030?ooo`<00000203oool00`2oM00004RL
0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO09a8M02oool0<@3oool00240oooo00<0
00000?ooo`3oool0K`3oool00`000000Uh6j09N1^P0200000100oooo00@000000:B4/P2TQ;800000
0P2QPK4010000000UGR`09Eh/000000206IDZ@070000000003H0000f0000=P000000/8nd00000002
0;6C^0030000002aTkP0000000D0`Inf00<000000=2]]@3@[KD0100000020:6Yg08000000`2XY]D2
00000080Zik:0P0000001@2FNK00000009Ii/02FNK00000000@0UWZa0P0000050;2?/`030000002S
QkH0000000@0]IBf00<000000:jB^@0000001P2fV;T00`000000]Ybm0;JL_@030;JL_@030000002i
X;l0^J2o0080^J2o00<000000;RQ`@000000102jXl800`000000_:G20;bU`P020;bU`P040000002m
Y/80_JK20;fV`P<000001032Zl<01@000000a:c20<B/`P34[<80000000<0ajk100<000000<Z__`3:
[kl00P3:[kl01@000000c[2l0<j`_03>/;`000000080d[:j00<000000=Je^03F]KP00P0000030<bX
]@050000003?Z[<0cjZc0<nZ/`0000000P3B[;801@000000eZj`0=J^/03F[[00000000<0b::a00@0
00000<BN/P00000000000P3aajH010000000d:R`0=2X/00000030<BN/@030000003oool0oooo01L0
oooo0`000003095n^`050000002RRk`0XX^l0::;_00000000`2]U;d01@000000/IFj0;6E^P2aUKX0
00000080^ink00<000000<6Q^@31XKT00`31XKT01@000000b:Ff0<RU]P38YKH0000000<0d:^c00@0
00000=^d/03K];0000001@3oool00`000000oooo0?ooo`050?ooo`030?oo_`1d0000B9cO00<0oooo
00<0gia80000M02oool02`3oool00`2oM000M;oo0?ooo`0`0?ooo`008@3oool00`000000oooo0?oo
o`1_0?ooo`<000004`3oool00`000000Y8Bb000000020:61/@040000002EN;00UGR`00000080IUBY
00H000000000=P0003H0000f0000002`Sk@7000000@0`Inf00<000000=2]]@3@[KD00P0000000`2I
[nH00000000000030:6Yg0<000000P2TWm001P000000Zik:000000000000UWV`00000080UWV`00<0
000009Ij/@2FN[400P2FN[400`000000/8nc0;2?/`040;2?/`030000002SQkH0000000@0]IBf00<0
00000:jB^@0000001P2fV;T2000000@0]Ybm00<000000;VP_`2iX;l00`2iX;l00`000000^Z?20;ZS
`P020;ZS`P030000002lYL80_:G200<0_:G200<000000;fV`P2mY/800P2mY/801P000000_ZO20000
00000000`Z_3000000@0a:c200D000000<N^`@37[/40ajk1000000040<Z__`050000003>/;`0c[2l
0<j`_00000000P3B/[X3000000030<VW]`000000c:Re0080c:Re00D000000<nZ/`3?Z[<0cjZc0000
00020=://P050000003F[[00eZj`0=J^/00000000`38X[4010000000a9jb0000000000020?77YP04
0000003@Z;00d:R`000000<0a9ja00<000000?ooo`3oool05`3oool20000000506YT^@000000TGjk
095n^`000000102RRk`010000000[IBm0:fD_@0000030;6E^P040000002kWk/0^ink000000D0`J6i
00D000000<RU]P38YKH0b:Ff000000030=2[/`040000003K];00fkB`000000D0oooo00<000000?oo
o`3oool01@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0oooO09a8
0018W=l0<@3oool00240oooo00<000000?ooo`3oool0Q@3oool00`000000Y8Bb000000020:61/@04
0000002EN;00UGR`00000080IUBY00<000000000=P0003H0100000040;6C^00300000031WkH00000
00h000000P2LX]T010000000Y9o@0:BOd02TWm030000000309Ii/0000000UWV`0080UWV`00D00000
09Ij/@2FN[40UWZa000000060;2?/`030000002SQkH0000000@0]IBf00<000000:jB^@0000001@2f
V;T010000000^Yfk0;ZM^`0000020;JL_@030000002iX;l0^J2o00@0^J2o0`0000020;ZS`P030000
002lYL80_:G20080_:G200<000000;fV`P2mY/800`2mY/8010000000_ZO20;jW`P2nYl8300000080
a:c200<000000<N^`@37[/400P37[/401@000000bZno0<Z__`3:[kl0000000<0c[2l00D000000=:b
^P3B/[X0d[:j000000030<VW]`050000003<Z;D0c:Re0<bX]@0000000`3?Z[<010000000dZbb0=:/
/P0000030=J^/00400000038X[40b::a00000080a9jb0P0000000`3aajH000000=2X/0020=2X/005
00000034W[40a9ja0<BN/@0000006@3oool20000000506YT^@000000TGjk095n^`000000102RRk`0
0`000000[IBm000000050;6E^P030000002kWk/0000000D0`J6i00D000000<RU]P38YKH0b:Ff0000
00030=2[/`040000003K];00fkB`000000D0oooo00<000000?ooo`3oool01@3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0gia80000001d_ol0<@3oool00240oooo00<0
00000?ooo`3oool0Q@3oool00`000000Y8Bb000000020:61/@050000002EN;00UGR`0000001VE:T0
1@0000030;2?]0050000002aTkP0/I>h0;6C^00000001031WkH00`000000d:fe0=2]]@0:00000080
W:;I0P0000020:BOd080000000<0UWV`0000002FNK000P2FNK001@000000UWZa09Ij/@2FN[400000
00H0/8nc00<000000:>7]P0000001@2eU;H00`000000]YRi0;JH^@040;JH^@040000002jWK/0^Yfk
00000080]Ybm00<000000;VP_`2iX;l00`2iX;l010000000^j;00;^R`02kX/03000000@0_:G200<0
00000;fV`P2mY/800P2mY/800`000000_ZO20;jW`P020;jW`P0300000030Yl00`:O000@000000P37
[/400`000000bZno0<Z__`020<Z__`060000003>/;`0c[2l0<j`_0000000d[:j0`0000030<VW]`05
0000003<Z;D0c:Re0<bX]@0000000`3?Z[<010000000dZbb0=://P0000030=J^/00400000038X[40
b::a00000080a9jb0P0000000`3aajH000000=2X/0020=2X/00500000034W[40a9ja0<BN/@000000
6@3oool0100000009BbR000000000002095n^`030000002RRk`0XX^l0080XX^l00<000000:fD_@00
00001@2aUKX2000000D0`J6i00<000000<RU]P38YKH00P38YKH01@000000d:^c0=2[/`3@Zk<00000
0080fkB`00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol01P3oool01`3Ogml0oooo0?oo_`1d000000000000M02oool0<@3o
ool00240oooo00<000000?ooo`3oool0Q@3oool00`000000Y8Bb000000020:61/@D0000000D0IUBY
0000000003H0000f000000030;2?]0050000002aTkP0/I>h0;6C^00000001031WkH00`000000d:fe
0=2]]@03000000@0oooo2`0000000`2FNK00000009Ii/00209Ii/0050000002FN[40UWZa09Ij/@00
00001P2`Sk<00`000000XhNf000000050;FD]P030000002fV;T0]YRi00@0]YRi00@000000;ZM^`2j
WK/0^Yfk0P0000060;VP_`030000002kX/00^j;00080^j;000H000000;^T`@00000000000;bU`P00
00050;fV`P030000002nYl80_ZO20080_ZO200<000000<2W`030Yl000P30Yl000`000000`JNo0000
000<00000080ajNi00@000000<VW]`39YkL00000103<Z;D01@000000cjZc0<nZ/`3?Z[<000000080
dZbb00D000000=J^/03F[[00eZj`000000020<RR/@0400000034W[80000000000080lLNV00D00000
0=2X/03@Z;00d:R`000000030<BN/@030000003oool0oooo01P0oooo0`000003095n^`030000002R
Rk`0XX^l0080XX^l00<000000;6E^P2aUKX01@2aUKX00`000000`J6i0<6Q^@030<6Q^@0300000038
YKH0b:Ff0080b:Ff00D000000=2[/`3@Zk<0d:^c000000020=^d/0030000003oool0oooo00@0oooo
00<000000?ooo`3oool0103oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo
00H0gia800000000000000000000B02Lgolb0?ooo`008@3oool00`000000oooo0?ooo`260?ooo`D0
00000P2EN;0010000000IUBY06IDZ@0000020000=P050000002`Sk@0/8nd0;2?]00000000`2aTkP0
0`000000`Inf0<6O]P020<6O]P060000003@[KD0d:fe0000001H>i@000001P3oool20000000607ZM
j0000000TYWG09:Ie`000000WiO<2@0000000`2FN[4000000;2?/`050;2?/`030000002SQkH00000
00D0]IBf0P0000050;JH^@030000002jWK/0^Yfk0080^Yfk00<000000;VP_`2iX;l00`2iX;l00`00
0000^j;00;^R`0020;^R`0030000002kY<40^jC10080^jC10P0000040;fV`P030000002nYl80_ZO2
0080_ZO200<000000<2W`030Yl000P30Yl000`000000`JNo0<6W_`020<6W_`0300000033Ykd0`jNm
0080`jNm00D000000<FW^`35Yk/0aJNk000000030<NW^@0400000039YkL0bJNg000000@0c:Re00D0
00000<nZ/`3?Z[<0cjZc000000020=://P050000003F[[00eZj`0=J^/00000000P38X[4010000000
a9jb0000000000020?77YP050000003@Z;00d:R`0=2X/00000000`34W[400`000000oooo0?ooo`0H
0?ooo`80000000D0JVBi0000002AO[/0TGjk000000040::;_0030000002aUKX0/IFj00D0/IFj00<0
00000<6Q^@31XKT00`31XKT00`000000b:Ff0<RU]P020<RU]P040000003@Zk<0d:^c000000<0fkB`
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030?oo_`1d0000B9cO00<0oooo
00<0gia80000M02oool01P3oool0103ookl0M00000000018W=ld0?ooo`008@3oool00`000000oooo
0?ooo`220?ooo`H000000P2QPK4010000000UGR`09Eh/000000306IDZ@030000000003H0000000<0
/8nd00@000000;6C^02aTkP000001@31WkH01P000000d:fe0=2]]@000000F3^D000000P0oooo1000
00001@2BVML000000000002FNK00000000<0UWV`00<0000009Ij/@2FN[402@0000000`2SQkH00000
0;FD]P040;FD]P800000102fV;T00`000000^Yfk0;ZM^`030;ZM^`<000000`2iX;l00`000000^j;0
0;^R`0020;^R`0030000002kY<40^jC10080^jC100<000000;bT`P0000000`2mY/800`000000_ZO2
0;jW`P020;jW`P0300000030Yl00`:O00080`:O000D000000<6W_`31Ykl0`JNo000000040<>W_@03
00000035Yk/0aJNk0080aJNk00@000000<NW^@37YkT000000`39YkL00`000000c:Re0<bX]@020<bX
]@050000003?Z[<0cjZc0<nZ/`0000000P3B[;801@000000eZj`0=J^/03F[[0000000080b::a00@0
00000<BN/P00000000000P3aajH01@000000d:R`0=2X/03@Z;00000000<0a9ja00<000000?ooo`3o
ool0603oool400000080TGjk00@000000::;_02RRk`0XX^l0P0000070;6E^P0300000031XKT0`J6i
00<0`J6i00<000000<RU]P38YKH00P38YKH010000000d:^c0=2[/`0000030=^d/0030000003oool0
oooo0080oooo0`0000070?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool0
0`2oM000M;oo0?ooo`0d0?ooo`008@3oool00`000000oooo0?ooo`1n0?ooo`@00000103oool20000
0080XH6a00@0000009Eh/02EN;0000000`1VE:T2000000@0/8nd00@000000;6C^02aTkP000001@31
WkH01P000000d:fe0=2]]@000000F3^D000000X0oooo1@0000000`2FNK00000009Ii/00209Ii/005
0000002FN[40UWZa09Ij/@0000001P2`Sk<9000000030;^K]`000000]YRi0080]YRi00<000000;ZM
^`2jWK/00P2jWK/010000000^j6m0;^Q_@2kXKd3000000@0^j;000<000000;^T`@2kY<400`2kY<40
0`000000_:C20;bT`P03000000@0_ZO200<000000<2W`030Yl000`30Yl001@000000`JNo0<6W_`31
Ykl0000000@0`jNm00D000000<FW^`35Yk/0aJNk000000030<NW^@0500000039YkL0bJNg0<VW]`00
00000`3<Z;D01@000000cjZc0<nZ/`3?Z[<0000000<0dZbb00D000000=J^/03F[[00eZj`00000002
0<RR/@0400000034W[80000000000080lLNV00D000000=2X/03@Z;00d:R`000000030<BN/@030000
003oool0oooo01P0oooo10000003095n^`060000002RRk`0XX^l0000002WS;P000001P2aUKX20000
00@0`J6i00D000000<RU]P38YKH0b:Ff000000030=2[/`050000003K];00fkB`0=^d/00000001P3o
ool00`000000oooo0?ooo`050?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`06
0?oOW018000000000000000004P0W=oo<@3oool00240oooo00<000000?ooo`3oool0NP3oool40000
00P0oooo0P0000020:61/@040000002EN;00UGR`000000<0IUBY0P0000040;2?]0040000002aTkP0
/I>h000000D0`Inf00<000000=2]]@3@[KD00P00000=0?ooo`<0000000<0UWV`0000002FNK000P2F
NK001@000000UWZa09Ij/@2FN[40000000L0/8nc00<000000;FD]P2eU;H00`2eU;H9000000<0^Yfk
00<000000;^Q_@2kXKd00`2kXKd00`000000^j;00;^R`0020;^R`0030000002kY<40^jC10080^jC1
00<000000;bT`P2lY<800`2lY<800`000000_ZO20;jW`P020;jW`P0300000030Yl00`:O00080`:O0
00<000000<6W_`31Ykl00P31Ykl00`000000`jNm0<>W_@020<>W_@0500000035Yk/0aJNk0<FW^`00
00000`37YkT01@000000bJNg0<VW]`39YkL0000000<0c:Re00D000000<nZ/`3?Z[<0cjZc00000003
0=://P050000003F[[00eZj`0=J^/00000000P38X[42000000050?3;[0000000lLNV0?77YP000000
0`3@Z;001@000000a9ja0<BN/@34W[40000001X0oooo00@0000002D/XP00000000000`2AO[/00`00
0000XX^l000000020:N<^0030000002aUKX0/IFj00<0/IFj00<000000;JG]`0000001031XKT01@00
0000b:Ff0<RU]P38YKH0000000<0d:^c00D000000=^d/03K];00fkB`000000060?ooo`030000003o
ool0oooo00H0oooo00D0oono07@000000000001d0;ooo`0:0?ooo`050;md00000000000000000018
W=l0<@3oool00240oooo00<000000?ooo`3oool0M@3oool5000000`0oooo0P0000020:61/@040000
002EN;00UGR`000000<0IUBY0P0000050;2?]0030000002aTkP0000000D0`Inf00<000000=2]]@3@
[KD00P00000?0?ooo`030000002FNK00000000<0UWV`00D0000009Ij/@2FN[40UWZa000000070;2?
/`030000002eU;H0]IBf00<0]IBf00<000000;^K]`2kVkL00P2kVkL010000000]YRi0000002jWK/9
000000050;bS`000000000000;^R`00000001@2kY<400`000000_:C20;bT`P030;bT`P<0000000<0
_ZO200000030Yl001030Yl000`000000`JNo0<6W_`020<6W_`0300000033Ykd0`jNm0080`jNm00D0
00000<FW^`35Yk/0aJNk000000030<NW^@0500000039YkL0bJNg0<VW]`0000000`3<Z;D01@000000
cjZc0<nZ/`3?Z[<0000000<0dZbb00D000000=J^/03F[[00eZj`000000020<RR/@80000000D0l<^/
0000003aajH0lLNV000000030=2X/00500000034W[40a9ja0<BN/@0000006P3oool00`0000009BbR
0000000200000080TGjk00<000000::;_00000000P2WS;P00`000000/IFj0;6E^P030;6E^P040000
002fUkL0]YNg000000<0`J6i00D000000<RU]P38YKH0b:Ff000000030=2[/`050000003K];00fkB`
0=^d/00000001P3oool00`000000oooo0?ooo`1;0?ooo`008@3oool00`000000oooo0?ooo`1a0?oo
o`@000004P3oool2000000050:61/@000000UGR`09Eh/00000000`1VE:T2000000D0/8nd00<00000
0;6C^00000001@31WkH00`000000d:fe0=2]]@02000000l0oooo00<0000009Ii/00000000`2FNK00
10000000UWZa09Ij/@2FN[42000000H0/8nc00<000000;FD]P2eU;H00`2eU;H00`000000^i^g0;^K
]`020;^K]`800000102jWK/00`000000^j6m0;^Q_@030;^Q_@L000000`2kY<400`000000_:C20;bT
`P030;bT`P040000002lY<40_:C10;bT`@<000001030Yl001@000000`JNo0<6W_`31Ykl0000000@0
`jNm00<000000<FW^`35Yk/00P35Yk/01@000000ajNi0<NW^@37YkT0000000<0bJNg00D000000<bX
]@3<Z;D0c:Re000000030<nZ/`050000003B[;80dZbb0=://P0000000`3F[[000`000000b::a0<RR
/@02000000050?3;[0000000lLNV0?77YP0000000`3@Z;001@000000a9ja0<BN/@34W[40000001/0
oooo0P0000000`1aIkP00000095n^`02095n^`030000002WS;P0Yhbh0080Yhbh00<000000;6E^P2a
UKX00P2aUKX010000000]YNg0;JG]`0000030<6Q^@0500000038YKH0b:Ff0<RU]P0000000`3@Zk<0
1@000000fkB`0=^d/03K];00000000H0oooo00<000000?ooo`3oool0B`3oool00240oooo00<00000
0?ooo`3oool0K@3oool4000001H0oooo0P0000001@2QPK40000009Eh/02EN;00000000<0IUBY0P00
00050;2?]0030000002aTkP0000000@0`Inf00<000000=2]]@3@[KD00`00000?0?ooo`030000002F
NK00000000<0UWV`00H0000009Ij/@2FN[4000000;6@]00000060;2?/`030000002eU;H0]IBf00<0
]IBf00<000000;^K]`2kVkL00`2kVkL00`000000^Yfk0;ZM^`020;ZM^`030000002kXKd0^j6m0080
^j6m00<000000;bS`02lXl000`2lXl0:000000030;bT`P000000_:C100<0_:C100H000000;bS_`00
000000000<2W`00000040<6W_`0300000033Ykd0`jNm0080`jNm00<000000<FW^`35Yk/00P35Yk/0
1@000000ajNi0<NW^@37YkT0000000<0bJNg00D000000<bX]@3<Z;D0c:Re000000030<nZ/`050000
003B[;80dZbb0=://P0000000`3F[[000`000000b::a0<RR/@02000000050?3;[0000000lLNV0?77
YP0000000`3@Z;001@000000a9ja0<BN/@34W[40000001/0oooo0P0000000`1aIkP0000000000003
000000@0Yhbh00<000000;6E^P2aUKX00P2aUKX01`000000]YNg0;JG]`2fUkL000000<6Q^@000000
1038YKH01@000000d:^c0=2[/`3@Zk<0000000<0fkB`00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`1;0?ooo`00803oool00`000000oooo0?ooo`1Z0?ooo`@000006P3oool2000000050:61
/@000000UGR`09Eh/00000000`1VE:T00`000000/8nd0;2?]0040;2?]0030000002aTkP0000000@0
`Inf00H000000=2]]@3@[KD0000005PkU000000?0?ooo`030000002FNK00000000<0UWV`00H00000
09Ij/@2FN[4000000;6@]00000060;2?/`800000102eU;H00`000000^i^g0;^K]`030;^K]`800000
0`2jWK/00`000000^j6m0;^Q_@020;^Q_@030000002lXl00_:?00080_:?000@000000;^T`@2kY<40
^jC10P0000050;bT`PP000000P2lXkl2000000@0`JNo00<000000<>W_@33Ykd00P33Ykd00`000000
aJNk0<FW^`020<FW^`0500000037YkT0ajNi0<NW^@0000000`39YkL01@000000c:Re0<bX]@3<Z;D0
000000<0cjZc00D000000=://P3B[;80dZbb000000020=J^/00400000038X[40b::a0<RR/@800000
00D0l<^/0000003aajH0lLNV000000030=2X/00500000034W[40a9ja0<BN/@0000006`3oool20000
0003075W^0000000VH:h0080VH:h00<000000:N<^02WS;P00P2WS;P00`000000/IFj0;6E^P020;6E
^P070000002fUkL0]YNg0;JG]`000000`J6i000000040<RU]P050000003@Zk<0d:^c0=2[/`000000
0`3K];000`000000oooo0?ooo`020?ooo`<00000C@3oool00200oooo00<000000?ooo`3oool0I@3o
ool5000001h0oooo0P0000001@2QPK40000009Eh/02EN;00000000<0IUBY00<000000;2?]02`Sk@0
102`Sk@00`000000/I>h000000040<6O]P060000003@[KD0d:fe0000001H>i@000003`3oool00`00
0000UWV`0000000309Ii/0060000002FN[40UWZa0000002aT;@000001@2`Sk<00`000000^9Je0000
00040;FD]P030000002kVkL0^i^g0080^i^g00<000000;jO^@2nWkT00`0000040;^Q_@030000002l
Xl00_:?000<0_:?000<000000;^T`@2kY<400P2kY<43000000<0_:C200<000000;bT`@2lY<400P2l
Y<400`000000_:>o0;bS_`09000000<0`jNm00<000000<FW^`35Yk/00P35Yk/01@000000ajNi0<NW
^@37YkT0000000<0bJNg00<000000<bX]@3<Z;D00P3<Z;D01@000000cjZc0<nZ/`3?Z[<0000000<0
dZbb00@000000=J^/03F[[0000000`38X[4010000000l<^/0?3;[00000020?77YP050000003@Z;00
d:R`0=2X/00000000`34W[400`000000oooo0?ooo`0I0?ooo`80000000<0LFNh0000002IP[P00P2I
P[P00`000000Yhbh0:N<^0020:N<^0050000002aUKX0/IFj0;6E^P0000001@2fUkL2000000@0b:Ff
00D000000=2[/`3@Zk<0d:^c000000030=^d/0030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0B`3oool00200oooo00<000000?ooo`3oool0H@3oool4000002<0oooo0P0000001@2QPK400000
09Eh/02EN;00000000<0IUBY00<000000;2?]02`Sk@0102`Sk@2000000D0`Inf00H000000=2]]@3@
[KD0000005PkU000000@0?ooo`8000000`2FNK001P000000UWZa09Ij/@000000/I2d000000D0/8nc
00<000000;RF]@0000000`2eU;H00`000000^i^g0;^K]`030;^K]`030000002nWkT0_Yni0080_Yni
00<000000;^Q_@2kXKd00P2kXKd00`000000_:?00;bS`0020;bS`0030000002kY<40^jC10080^jC1
00@000000;ZS`P2jXl80^Z?20`0000040;bT`@030000002lXkl0_:>o00<0_:>o00<000000;bR_@2l
X[d02`0000000`35Yk/000000<NW^@020<NW^@0500000039YkL0bJNg0<VW]`000000103<Z;D01@00
0000cjZc0<nZ/`3?Z[<0000000<0dZbb00@000000=J^/03F[[0000000`38X[4010000000l<^/0?3;
[00000020?77YP050000003@Z;00d:R`0=2X/00000000`34W[400`000000oooo0?ooo`0I0?ooo`05
0000000U;:800000075W^00000000`2IP[P00`000000Yhbh0:N<^0020:N<^0040000002aUKX0/IFj
000000D0]YNg0P0000040<RU]P050000003@Zk<0d:^c0=2[/`0000000`3K];000`000000oooo0?oo
o`040?ooo`030000003oool0oooo04/0oooo000P0?ooo`030000003oool0oooo05d0oooo1000000X
0?ooo`030000002QPK40000000<0UGR`00@0000006IDZ@1VE:T000001`2`Sk@00`000000`Inf0<6O
]P030<6O]P060000003@[KD0d:fe0000001H>i@00000403oool2000000<0UWV`00H0000009Ij/@2F
N[4000000;6@]00000050;2?/`040000002hU[D0^9Je00000080]IBf00<000000;^K]`2kVkL00`2k
VkL00`000000_Yni0;jO^@020;jO^@<0000000<0^j6m0000002lXl00102lXl000`000000^jC10;^T
`@020;^T`@030000002jXl80^Z?200<0^Z?200<000000;bT`@2lY<400P2lY<400`000000_:>o0;bS
_`030;bS_`030000002lX[d0_::m0080_::m00<000000<>W_@33Ykd00P33Ykd010000000aJNk0<FW
^`35Yk/8000000030<VW]`000000c:Re00<0c:Re00D000000<nZ/`3?Z[<0cjZc000000030=://P04
0000003F[[00eZj`000000<0b::a00@000000?3;[03`bj`000000P3aajH01@000000d:R`0=2X/03@
Z;00000000<0a9ja00<000000?ooo`3oool06P3oool200000003075W^0000000VH:h0080VH:h00<0
00000:N<^02WS;P00P2WS;P010000000/IFj0;6E^P0000060;JG]`0300000038YKH0b:Ff0080b:Ff
00D000000=2[/`3@Zk<0d:^c000000030=^d/0030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0B`3oool00200oooo00<000000?ooo`3oool0F@3oool4000002`0oooo00@000000:61/@2QPK40
00000P2EN;000`000000IUBY06IDZ@02000000H0/8nd00<000000<6O]P31WkH00`31WkH01P000000
d:fe0=2]]@000000F3^D00000100oooo0P00000309Ii/0060000002FN[40UWZa0000002aT;@00000
1@2`Sk<010000000^9Je0;RF]@0000020;FD]P030000002kVkL0^i^g00<0^i^g00D000000;jO^@2n
WkT0_Yni000000030;nT_@@000000P2lXl000`000000^jC10;^T`@020;^T`@030000002jXl80^Z?2
00<0^Z?200<000000;VQ`@0000000`2lY<400`000000_:>o0;bS_`020;bS_`030000002lX[d0_::m
00<0_::m0P0000030<>W_@0300000035Yk/0aJNk0080aJNk00D000000<NW^@37YkT0ajNi00000002
0<VW]`P000000P3?Z[<01@000000dZbb0=://P3B[;8000000080eZj`00D000000<RR/@38X[40b::a
000000020?3;[0040000003aajH0lLNV000000<0d:R`00D000000<BN/@34W[40a9ja0000000L0?oo
o`80000000<0LFNh0000002IP[P00P2IP[P00`000000Yhbh0:N<^0020:N<^0040000002aUKX0/IFj
000000H0]YNg00<000000<RU]P38YKH00P38YKH01@000000d:^c0=2[/`3@Zk<0000000<0fkB`00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`1;0?ooo`00803oool00`000000oooo0?ooo`1D
0?ooo`D00000<03oool010000000XH6a0:61/@00000209Eh/0050000001VE:T000000;6C]`000000
1P2`Sk@00`000000`Inf0<6O]P030<6O]P060000003@[KD000000000001H>i@00000403oool20000
00<0UWV`00@0000009Ij/@2FN[4000000P2aT;@00`000000/8nc0;2?/`020;2?/`040000002hU[D0
^9Je00000080]IBf00<000000;^K]`2kVkL00P2kVkL00`000000_Yni0;jO^@020;jO^@030000002o
Y;d0_jBm0080_jBm00<000000;jV`@2nY/400`0000040;^T`@030000002jXl80^Z?200<0^Z?200<0
00000;VQ`@2iXL400`0000050;bS_`030000002lX[d0_::m0080_::m00<000000;fQ_02mXK`00`00
00050<FW^`0500000037YkT0ajNi0<NW^@0000000`39YkL01@000000c:Re0<bX]@3<Z;D000000080
cjZc20000000103F[[0000000<RR/@38X[4200000080l<^/00@000000?77YP3aajH000000`3@Z;00
1@000000a9ja0<BN/@34W[40000001`0oooo0P0000000`1aIkP0000009V2^00209V2^0030000002W
S;P0Yhbh00<0Yhbh00<000000;6E^P0000001@2fUkL00`000000]YNh000000030<RU]P050000003@
Zk<0d:^c0=2[/`0000000`3K];000`000000oooo0?ooo`040?ooo`030000003oool0oooo04/0oooo
000P0?ooo`030000003oool0oooo0500oooo1000000e0?ooo`040000002QPK40XH6a00000080UGR`
00D0000006IDZ@000000/I>g000000060;2?]08000001031WkH00`000000d:fe0000000205PkU003
0000003oool0oooo00h0oooo0P00000309Ii/0040000002FN[40UWZa00000080/I2d00<000000;2?
/`2`Sk<00P2`Sk<010000000^9Je0;RF]@0000020;FD]P030000002kVkL0^i^g0080^i^g00<00000
0;jO^@2nWkT00P2nWkT01@000000_jBm0;nT_@2oY;d0000000@0_ZK100H000000;bU``0000000000
0;^T`@0000050;ZS`P030000002iXL40^J7100<0^J7100<000000;bS_`2lXkl00P2lXkl00`000000
_::m0;bR_@030;bR_@030000002mXK`0_J6l0080_J6l00<000000<FW^`35Yk/00`35Yk/01@000000
ajNi0<NW^@37YkT0000000<0bJNg00D000000<bX]@3<Z;D0c:Re000000040<nZ/`060000003B[;80
dZbb0=://P000000eZj`2@0000020?77YP050000003@Z;00d:R`0=2X/00000000`34W[400`000000
oooo0?ooo`0J0?ooo`8000000P1aIkP010000000VH:h09V2^00000050:N<^0030000002aUKX00000
00D0]YNg00<000000;JG^00000000`38YKH01@000000d:^c0=2[/`3@Zk<0000000<0fkB`00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`1;0?ooo`00803oool00`000000oooo0?ooo`1<0?oo
o`@00000>P3oool00`000000XH6a0000000209Eh/0030000001VE:T000000080/I>g00<000000;2?
]02`Sk@00P2`Sk@00`000000]IRj000000040<6O]P<0000000<0F3^D0000003oool0403oool20000
00<0UWV`00@0000009Ij/@2FN[4000000P2aT;@01@000000/8nc0;2?/`2`Sk<0000000@0^9Je00<0
00000;FD]P000000102kVkL01@000000_Yni0;jO^@2nWkT0000000@0_jBm00<000000;jV`@2nY/40
0P2nY/4010000000_:G30;bU``2lYL<2000000D0^Z?200<000000;VQ`@2iXL400`2iXL42000000<0
_:>o00<000000;bR_@2lX[d00`2lX[d00`000000_J6l0;fQ_0020;fQ_0<000000P35Yk/00`000000
ajNi0<NW^@020<NW^@0500000039YkL0bJNg0<VW]`0000000`3<Z;D00`000000cjZc0<nZ/`020<nZ
/`040000003B[;80dZbb000000<0eZj`00H000000<RR/@000000ilRh0000003`bj`9000000<0a9ja
00<000000?ooo`3oool06`3oool010000000LFNh075W^000000309V2^0030000002WS;P0Yhbh0080
Yhbh0P0000060;JG]`060000002fUkP0]YNh00000038YKH00000103@Zk<010000000fkB`0=^d/000
00050?ooo`<00000C@3oool00200oooo00<000000?ooo`3oool0B03oool4000003h0oooo00<00000
0:61/@0000000P2EN;000`000000IUBY000000020;6C]`030000002`Sk@0/8nd0080/8nd00@00000
0;FH^P2eV;X000000`31WkH30000000305PkU0000000oooo0100oooo0P00000309Ii/0030000002F
N[40000000<0/I2d00D000000;2?/`2`Sk<0/8nc000000040;RF]@800000102kVkL00`000000_Yni
0;jO^@020;jO^@050000002oY;d0_jBm0;nT_@000000102nY/400`000000_:G30;bU``020;bU``03
0000002iXl<0000000<0^Z?200<000000;VQ`@2iXL400`2iXL400`000000^9no0;RO_`03000000H0
_::m00<000000;fQ_02mXK`00P2mXK`00`000000_Z2j0;jP^P03000000@0ajNi00D000000<VW]`39
YkL0bJNg000000030<bX]@030000003?Z[<0cjZc0080cjZc00@000000=://P3B[;8000000`3F[[00
10000000b::a0000000000020?3;[0050000003aajH0lLNV0?77YP0000000`3@Z;08000001X0oooo
00@00000075W^01aIkP000000`2IP[P00`000000Yhbh0:N<^0030:N<^0030000002fUkL0]YNg00@0
]YNg00H000000;JG^02fUkP000000<RU]P0000040=2[/`040000003K];00fkB`000000L0oooo00<0
00000?ooo`3oool0B`3oool00200oooo00<000000?ooo`3oool0A03oool400000480oooo00<00000
0:61/@0000000P2EN;02000000<0/I>g00<000000;2?]02`Sk@00P2`Sk@010000000]IRj0;FH^P00
00020<6O]P060000003@[KD000000000001H>i@000004@3oool00`000000UWV`0000000209Ii/003
0000002FN[40000000<0/I2d00D000000;2?/`2`Sk<0/8nc000000040;RF]@800000102kVkL00`00
0000_Yni0;jO^@020;jO^@050000002oY;d0_jBm0;nT_@000000102nY/400`000000_:G30;bU``02
0;bU``030000002iXl<0^J?30080000000<0^Z?20000002iXL40102iXL400`000000^9no0;RO_`02
0;RO_`030000002lX[d0_::m00<0_::m00<000000;fQ_02mXK`00P2mXK`00`000000_Z2j0;jP^P03
0;jP^P0500000037YkT0ajNi0<NW^@0000001039YkL01@000000c:Re0<bX]@3<Z;D0000000@0cjZc
00@000000=://P3B[;8000000`3F[[0010000000b::a0000000000020?3;[0050000003aajH0lLNV
0?77YP0000000`3@Z;001@000000a9ja0<BN/@34W[40000000<0oooo2000000B0?ooo`80000000<0
LFNh0000002IP[P00P2IP[P00`000000Yhbh0:N<^0030:N<^0030000002fUkL0]YNg00<0]YNg00L0
00000;JG^02fUkP0]YNh00000038YKH0000000@0d:^c00@000000=^d/03K];0000001`3oool00`00
0000oooo0?ooo`1;0?ooo`00803oool00`000000oooo0?ooo`0o0?ooo`D00000AP3oool00`000000
XH6a0000000209Eh/0800000102aTkL01@000000/8nd0;2?]02`Sk@0000000<0]IRj00<000000<6O
]P0000000P00000205PkU0030000003oool0oooo0100oooo0P00000209Ii/0030000002FN[400000
00<0/I2d00D000000;2?/`2`Sk<0/8nc000000050;RF]@030000002kVkL0^i^g0080^i^g00<00000
0;jO^@2nWkT00P2nWkT010000000_jBm0;nT_@0000040;jV`@030000002lYL<0_:G30080_:G300<0
00000;VS``2iXl<00`2iXl<00`000000^J710;VQ`@030;VQ`@030000002hWkl0^9no00<0^9no0P00
00040;bR_@030000002mXK`0_J6l0080_J6l00<000000;jP^P2nX;X00`2nX;X200000080ajNi00D0
00000<VW]`39YkL0bJNg000000040<bX]@030000003?Z[<0cjZc0080cjZc00@000000=://P3B[;80
00000`3F[[02000000050>O8^0000000l<^/0?3;[00000000`3aajH01@000000d:R`0=2X/03@Z;00
000000<0a9ja00<000000?ooo`3oool02@3oool8000000X0oooo0P000002075W^0040000002IP[P0
VH:h000000D0Yhbh00<000000;JG]`2fUkL00`2fUkL00`000000]YNh0;JG^0020;JG^08000000`3@
Zk<01@000000fkB`0=^d/03K];00000000L0oooo00<000000?ooo`3oool0B`3oool00200oooo00<0
00000?ooo`3oool0>`3oool4000004`0oooo0P00000209Eh/0800000102aTkL010000000/8nd0;2?
]00000050;FH^P@000000P1H>i@00`000000oooo0?ooo`0@0?ooo`8000000P2FNK000`000000UWZa
000000040;6@]0040000002`Sk<0/8nc000000D0^9Je00<000000;^K]`2kVkL00P2kVkL01@000000
_Yni0;jO^@2nWkT000000080_jBm0`0000030;jV`@030000002lYL<0_:G30080_:G300<000000;VS
``2iXl<00P2iXl<010000000]Yo20000000000030;VQ`@030000002hWkl0^9no0080^9no00@00000
0;RM_@2hWKd000000`2lX[d00`000000_J6l0;fQ_0020;fQ_0030000002nX;X0_Z2j00<0_Z2j00<0
00000;nO^00000000P0000030<VW]`030000003<Z;D0c:Re0080c:Re00D000000<nZ/`3?Z[<0cjZc
000000030=://P040000003F[[00eZj`0=J^/080000000D0ilRh0000003`bj`0l<^/000000030?77
YP050000003@Z;00d:R`0=2X/00000000`34W[400`000000oooo0?ooo`0A0?ooo`P000000P3oool2
00000080LFNh00D0000009V2^02IP[P0VH:h000000030:N<^0030000002aUKX0000000@0]YNg00<0
00000;JG^02fUkP00P2fUkP2000000<0d:^c00D000000=^d/03K];00fkB`000000070?ooo`030000
003oool0oooo04/0oooo000P0?ooo`030000003oool0oooo03L0oooo1000001@0?ooo`8000000P2E
N;02000000D0/I>g00<000000;2?]00000001@2eV;X20000000505HRL`000000F3^D05PkU0000000
4P3oool2000000<0UWV`0P0000040;6@]0040000002`Sk<0/8nc000000D0^9Je0`000000102kVkL0
00000;jO^@2nWkT5000000030<J/_`000000`jg500D000000P2lYL<00`000000^J?30;VS``030;VS
``040000002fWl80]Yo20;JO`P<000001@2hWkl01@000000^9fm0;RM_@2hWKd000000080_::m00<0
00000;fQ_02mXK`00P2mXK`00`000000_Z2j0;jP^P020;jP^P030000002oWkP0_inh0080_inh00D0
00000<VW]`39YkL0bJNg000000040<bX]@050000003?Z[<0cjZc0<nZ/`0000000`3B[;801@000000
eZj`0=J^/03F[[0000000080ilRh00@000000?3;[03`bj`000000`3aajH01@000000d:R`0=2X/03@
Z;00000000<0a9ja00<000000?ooo`3oool06@3oool800000080VH:h00L000000:N<^02WS;P0Yhbh
0000002aUKX0000000@0]YNg00<000000;JG^02fUkP00`2fUkP01@000000d:^c0=2[/`3@Zk<00000
00<0fkB`00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`1;0?ooo`00803oool00`000000
oooo0?ooo`0c0?ooo`@00000E03oool200000080UGR`0P0000050;6C]`030000002`Sk@0000000H0
]IRj00@0000005HRL`000000F3^D0P00000B0?ooo`8000000`2FNK02000000@0/I2d00<000000;2?
/`0000001@2hU[D010000000_j:l0;nR_02oX[`400000080ajg000@000000<J/_`36[;l000000`33
[LD010000000_:[70;bZa`2lZ/L3000000@0^J?300<000000;JO`P2fWl80102fWl800`000000^9no
0;RO_`030;RO_`030000002hWKd0^9fm0080^9fm0P0000040;fQ_0030000002nX;X0_Z2j00<0_Z2j
00<000000;nO^02oWkP00P2oWkP3000000030<VW]`000000c:Re0080c:Re00<000000<nZ/`3?Z[<0
0P3?Z[<01@000000dZbb0=://P3B[;80000000<0eZj`00@000000>O8^03Wb;P000000P3`bj`01@00
0000lLNV0?77YP3aajH0000000<0d:R`00D000000<BN/@34W[40a9ja0000000N0?ooo`050000001a
IkP0LFNh0000002IP[P0200000001@2aUKX000000;JG]`2fUkL0000000H0]YNh00D000000=2[/`3@
Zk<0d:^c000000030=^d/0030000003oool0oooo00<0oooo0`00001=0?ooo`00803oool00`000000
oooo0?ooo`0^0?ooo`D00000F03oool200000080UGR`00<000000;6C]`2aTkL0102aTkL00`000000
/8nd000000050;FH^P030000001F8W<0ER9c00<000004`3oool2000000<0UWV`0P0000040;6@]003
0000002`Sk<0000000D0^9Je00<000000;nR_02oX[`00P2oX[`01@000000ajg00<N]`037[L000000
00<0aZbo00D000000<>]a@33[LD0`jg5000000040;bZa`040000002fY<H0000000000080^J?300<0
00000;JO`P2fWl800`2fWl800`000000]9bo000000030;RO_`030000002hWKd0^9fm00@0^9fm00<0
00000;fQ_02mXK`00P2mXK`00`000000_Z2j0;jP^P030;jP^P030000002oWkP0_inh0080_inh00<0
00000;nP^@2oX;T00P0000030<bX]@030000003?Z[<0cjZc0080cjZc00D000000=://P3B[;80dZbb
000000020=J^/08000000P3Wb;P010000000l<^/0?3;[00000030?77YP040000003@Z;00d:R`0000
00<0a9ja00<000000?ooo`3oool07@3oool01@000000LFNh075W^01aIkP000000080VH:h00@00000
0:N<^02WS;P000000P2aUKX800000080]YNh00<000000;6K`00000000P3@Zk<01@000000fkB`0=^d
/03K];00000000L0oooo00<000000?ooo`3oool0B`3oool00200oooo00<000000?ooo`3oool0:P3o
ool4000005h0oooo00@0000009Eh/02EN;0000001`2aTkL00`000000]IRj0;FH^P030;FH^P050000
001F8W<0ER9c0000001H>i@00P00000C0?ooo`8000000`2FNK02000000@0/I2d00<000000;2?/`00
00001@2hU[D01@000000_j:l0;nR_02oX[`0000000@0ajg000@000000<J/_`36[;l000000`33[LD0
0`000000_:[70;bZa`020;bZa`030000002fY<H0]ZC60080]ZC60P0000060;JO`P040000002dW;l0
]9bo00000080^9no00<000000;RM_@2hWKd0102hWKd2000000<0_J6l00<000000;jP^P2nX;X00`2n
X;X01@000000_inh0;nO^02oWkP0000000@0_j2i0P0000020<bX]@030000003?Z[<0cjZc0080cjZc
00T000000=://P3B[;80dZbb0000003F[[0000000=Bj`00000000P3Wb;P010000000l<^/0?3;[000
00030?77YP040000003@Z;00d:R`000000<0a9ja00<000000?ooo`3oool07@3oool200000080LFNh
00H0000009V2^02IP[P000000:N<^00000050;6E^P030000002fUkL0000000<0]YNh200000030=^d
/0030000003oool0oooo00D0oooo00<000000?ooo`3oool0B`3oool00200oooo00<000000?ooo`3o
ool09P3oool400000680oooo00<0000009Eh/00000000`0000040;6C]`<000000`2eV;X010000000
ER9c05HRL`1F8W<3000001@0oooo0P00000309Ii/0800000102aT;@00`000000/8nc000000050;RF
]@050000002oX[`0_j:l0;nR_00000000`37[L001@000000aZbo0<J/_`36[;l0000000<0`jg500D0
00000;bZa`2lZ/L0_:[7000000050;JTaP030000002bW/@0000000@0]Yo200D000000;BL_`2dW;l0
]9bo000000020;RO_`030000002hWKd0^9fm00<0^9fm00@000000;RK^`2hVk/000000P2mXK`00`00
0000_Z2j0;jP^P020;jP^P030000002oWkP0_inh0080_inh00<000000;nP^@2oX;T00P2oX;T00`00
0000_j6k00000002000000@0cjZc00@000000=://P3B[;80dZbb0P0000001@3D^/0000000>O8^03W
b;P0000000<0l<^/00D000000?77YP3aajH0lLNV000000020=2X/00500000034W[40a9ja0<BN/@00
0000803oool010000000LFNh075W^000000309V2^08000001@2aUKX00`000000]YNg000000040;JG
^0060000002aVl00/I_00;6K`0000000d:^c2@0000030?ooo`030000003oool0oooo04/0oooo000P
0?ooo`030000003oool0oooo0280oooo1000001V0?ooo`800000102Ac_d?0000000308E5NP000000
oooo01<0oooo0P00000309Ii/0800000102aT;@00`000000/8nc000000040;RF]@030000002oX[`0
_j:l0080_j:l00D000000<N]`037[L00ajg0000000020<J/_`0300000033[LD0`jg500H000001P2f
Y<H010000000/Yk40;:Na00000020;JO`P030000002dW;l0]9bo00<0]9bo0P0000060;RM_@060000
002hVk/0^9^k0000002mXK`000001@2nX;X00`000000_inh0;nO^0020;nO^0050000002oX;T0_j2i
0;nP^@000000102oXK/300000080cjZc00<000000=://P3B[;800P0000020=Bj`0040000003Wb;P0
ilRh000000<0l<^/00@000000?77YP3aajH000000`3@Z;001@000000a9ja0<BN/@34W[4000000200
oooo00@00000075W^01aIkP000000`2IP[P00`000000/IFj0;6E^P050;6E^P030000002fUkP0]YNh
0080]YNh00<000000;6K`02aVl000`2aVl02000000<0fkB`00<000000?ooo`3oool00P3oool40000
04d0oooo000P0?ooo`030000003oool0oooo01d0oooo1@00001[0?ooo`X00000102Ac_d400000080
QDEj00<000000?ooo`3oool0503oool01@000000UWV`09Ii/02FNK00000000H0/I2d0P0000040;RF
]@050000002oX[`0_j:l0;nR_00000001037[L000`000000aZbo0<J/_`04000000<0^kKD00@00000
0;6WcP00000000000`2fY<H00`000000/Yk40;:Na0020;:Na0030000002fWl80000000H0]9bo00<0
00000;RM_@2hWKd00`2hWKd00`000000^9^k0;RK^`020;RK^`8000001@2nX;X00`000000_inh0;nO
^0020;nO^0050000002oX;T0_j2i0;nP^@000000102oXK/00`000000`JFm0<6U_@05000000050<N/
`0000000e;[00=Bj`00000000P3Wb;P01@000000l<^/0?3;[03`bj`000000080lLNV00D000000=2X
/03@Z;00d:R`000000030<BN/@030000003oool0oooo01h0oooo0P000002075W^0030000002IP[P0
000000<00000102aUKX3000000<0]YNh00<000000;6K`02aVl000`2aVl03000000030=^d/0000000
000000D00000D03oool00200oooo00<000000?ooo`3oool06@3oool400000780oooo4`00000F0?oo
o`050000002FNK00UWV`09Ii/00000001P2aT;@00`000000^9Je0;RF]@030;RF]@050000002oX[`0
_j:l0;nR_00000000`37[L0010000000aZbo0<J/_`0000020<c7eP050000002k]]@0^kKD0;^fe000
0000102aYlh2000000030;JTaP000000/Yk400@0/Yk400<000000;BL_`2dW;l01@2dW;l2000000@0
^9fm00<000000;RK^`2hVk/00`2hVk/00`000000_Z2j0;jP^P030;jP^P050000002oWkP0_inh0;nO
^0000000102oX;T01@000000_j6k0;nQ^`2oXK/0000000D0`JFm00@000000<N/`037[<0000000`3D
^/0010000000ilRh0>O8^00000030?3;[0040000003aajH0lLNV000000<0d:R`00D000000<BN/@34
W[40a9ja0000000Q0?oooaT0000000<0TLkm0000000000000`00001E0?ooo`00803oool00`000000
oooo0?ooo`0E0?ooo`@00000W`3oool01@000000UWV`09Ii/02FNK00000000H0/I2d00<000000;RF
]@2hU[D00`2hU[D01@000000_j:l0;nR_02oX[`0000000<0ajg00`0000000`3<amH0000000000004
000000L0/JO>00<000000;:Na02bW/@00`2bW/@00`000000[iW1000000050;BL_`030000002cVKd0
000000@0^9fm00<000000;RK^`2hVk/00`2hVk/2000000<0_Z2j00<000000;nO^02oWkP00P2oWkP0
0`000000_j2i0;nP^@020;nP^@050000002oXK/0_j6k0;nQ^`0000001031YKd01@000000ajc00<N/
`037[<00000000<0e;[000@000000>O8^03Wb;P000000`3`bj`010000000lLNV0?77YP0000030=2X
/00500000034W[40a9ja0<BN/@0000008@3oool00`000000aiRV00000003000000<0TLkm5000001G
0?ooo`00803oool00`000000oooo0?ooo`0A0?ooo`@00000X`3oool01@000000UWV`09Ii/02FNK00
000000D0/I2d100000020;RF]@030000002oX[`0_j:l0080_j:l00<000000<N]`037[L002@000002
09f[h@8000000`2aYlh2000000030:^Mb0000000/Yk400<0/Yk400@000000:nI`@2_VL400000102d
W;l010000000/iVm0;>I_@0000020;RM_@030000002hVk/0^9^k00<0^9^k00<000000;:H_@000000
0`2nX;X00`000000_inh0;nO^0020;nO^0050000002oX;T0_j2i0;nP^@000000102oXK/00`000000
`JFm0<6U_@020<6U_@0500000037[<00ajc00<N/`00000000P3D^/001@000000ilRh0>O8^03Wb;P0
000000<0l<^/00@000000?77YP3aajH000000`3@Z;001@000000a9ja0<BN/@34W[4000000240oooo
3P0000000`2Ac_d0000000000005000005h0oooo000P0?ooo`030000003oool0oooo00`0oooo1@00
002W0?ooo`X00000102Ac_d400000080_j:l00<000000<N]`037[L00100000030:j?]@040000002c
TK<00000000000<0WJ_Q0`0000040:^Mb0040000002bW/@0/Yk4000000<0[iW100D000000;BL_`2d
W;l0]9bo000000040;>I_@030000002hWKd0000000D0^9^k00@000000;:H_@2bV;d000000P2nX;X0
0`000000_inh0;nO^0020;nO^0050000002oX;T0_j2i0;nP^@0000000`2oXK/00`000000`JFm0<6U
_@020<6U_@0500000037[<00ajc00<N/`00000000`3D^/001@000000ilRh0>O8^03Wb;P0000000<0
l<^/00@000000?77YP3aajH000000`3@Z;001@000000a9ja0<BN/@34W[4000000240oooo0`000003
097>o@`00000HP3oool00200oooo00<000000?ooo`3oool0203oool400000:d0oooo00<000000<BH
ZP0000005@0000050:j?]@030000002cTK<0/i6c00@000000P2MWM@00`000000Zig80:^Mb0030:^M
b0030000002bW/@0000000@0[iW100@000000;BL_`2dW;l00000102cVKd2000000D0^9^k00D00000
0;:H_@2bV;d0/YRm000000020;jP^P030000002oWkP0_inh0080_inh00D000000;nP^@2oX;T0_j2i
000000030;nQ^`0300000031YKd0`JFm0080`JFm00D000000<N/`037[<00ajc0000000030=Bj`005
0000003Wb;P0ilRh0>O8^00000000P3`bj`01@000000lLNV0?77YP3aajH0000000<0d:R`00D00000
0<BN/@34W[40a9ja0000000S0?ooo`X00000J03oool00200oooo00<000000?ooo`3oool0103oool4
00000;40oooo1`000003097>o@`0000000<0YXRe0000002^SkD0102^SkD01`000000/i6c0;>A/`2c
TK<0000006J@iP0000000`2MWM@2000000@0Zig800<000000:nI`@2_VL40102_VL42000000H0/iVm
0P0000040;RK^`030000002bV;d0/YRm0080/YRm0P0000040;nO^0030000002oX;T0_j2i0080_j2i
00D000000;nQ^`2oXK/0_j6k000000040<6U_@0400000037[<00ajc0000000@0e;[000D000000>O8
^03Wb;P0ilRh000000020?3;[0050000003aajH0lLNV0?77YP0000000`3@Z;001@000000a9ja0<BN
/@34W[40000001X0oooo00<000000?ooo`3oool0103oool8000006`0oooo000P0?ooo`030000003o
ool0oooo00@00000_03oool500000080oooo0P0000030:67^0050000002VR;D0YXRe0:J8]@000000
1@2^SkD00`000000/i6c0;>A/`020;>A/`030000001VT>H0000000@0WIgD00D000000:^Mb0000000
00000:n[e002000000@0[iW10`0000030;>I_@8000000P2_[MH2000000030;RK^`000000/YRm00@0
/YRm0P0000040;nO^0050000002oX;T0_j2i0;nP^@000000102oXK/01@000000`JFm0<6U_@31YKd0
000000<0ajc000<000000=Bj`03D^/000P3D^/0010000000ilRh0>O8^00000030?3;[0050000003a
ajH0lLNV0?77YP0000000`3@Z;001@000000a9ja0<BN/@34W[4000000180oooo00<000000?ooo`3o
ool01@3oool7000007@0oooo000P0?ooo`D00000a@3oool3000000030:67^0000000000000<0YXRe
00<000000:j?]@2^SkD00`2^SkD00`000000/i6c0;>A/`020;>A/`@000000`2MWM@3000000@0[j_D
100000030:JojaP000000P2oX;T01@000000_j6k0;nQ^`2oXK/0000000@0`JFm00D000000<N/`037
[<00ajc0000000040=Bj`0040000003Wb;P0ilRh000000<0l<^/00D000000?77YP3aajH0lLNV0000
00030=2X/00500000034W[40a9ja0<BN/@0000004P3oool010000000oooo0?ooo`3oool5000007X0
oooo000U0?ooo`T00000_03oool01P000000PW:i000000000000Vh>h000000<0YXRe00<000000:j?
]@2^SkD00`2^SkD2000000@0/i6c6@000003097>oAP000000P31YKd01@000000ajc00<N/`037[<00
000000@0e;[000D000000>O8^03Wb;P0ilRh000000030?3;[0050000003aajH0lLNV0?77YP000000
0`3@Z;001@000000a9ja0<BN/@34W[4000000140oooo1@00001o0?ooo`00;P3oool:00000;80oooo
00<00000089b^@0000000P2KPkP00`000000YXRe0:J8]@020:J8]@070000002^SkD0[Xne0:j?]@00
0000/I6e000000040;>A/`L000000`2Ac_dH000000<0TLkm600000020=Bj`0050000003Wb;P0ilRh
0>O8^00000000`3`bj`01@000000lLNV0?77YP3aajH0000000<0d:R`00D000000<BN/@34W[40a9ja
000000090?ooo`030000003oool0oooo00D0000000<0oooo0000003oool0P@3oool003P0oooo2@00
002Z0?ooo`8000000`2KPkP01@000000YXRe0:J8]@2VR;D0000000<0[Xne00<000000;6A]@000000
102cTK<00`000000/Y6d0000000<000000<0TLkm60000003097>oAP000000P3`bj`01@000000lLNV
0?77YP3aajH0000000<0d:R`00L000000<BN/@34W[40a9ja0000003oool0000000D0oooo1@000029
0?ooo`00@@3oool900000:40oooo00<00000089b^@0000000P2KPkP01@000000YXRe0:J8]@2VR;D0
000000<0[Xne00<000000;6A]@000000102cTK<00`000000/Y6d0;:A]00E000000<0TLkm60000003
097>oAL000000P3@Z;001P000000a9ja0<BN/@34W[4000000?ooo`H00000SP3oool004X0oooo2@00
002H0?ooo`0300000022L[T000000080Vh>h00D000000:J8]@2VR;D0YXRe000000030:j?]@030000
002aTKD0000000@0/i6c00@000000;:A]02bTK@000000`2iUK8I000000<0TLkm60000003097>oAP0
0000T`3oool005<0oooo2@00002@0?ooo`8000000P2KPkP01@000000YXRe0:J8]@2VR;D0000000<0
[Xne00@000000;6A]@2aTKD000000P2cTK<00`000000/Y6d0;:A]0020;:A]0050000002iUK80^IFb
0000002bT;801@000003097>oAL00000102Ac_dG00000003097>o@000000TLkm00`00000?@3oool0
103ool@0Ga@0000DG`34oomF0?ooo`00G03oool:000008H0oooo0P00000309^3^0040000002VR;D0
YXRe000000<0[Xne00@000000;6A]@2aTKD000000P2cTK<00`000000/Y6d0;:A]0020;:A]0040000
002iUK80^IFb000000@0/Y2b2P000003097>oAP000000`2Ac_dG000004@0oooo00@0ooo405lD0000
55l0a?ooEP3oool006H0oooo2@00001m0?ooo`0300000022L[T000000080Vh>h00L000000:J8]@2V
R;D0YXRe0000002^SkD0000000<0/I6e00@000000;>A/`2cTK<00000102bTK@010000000^IFb0;VE
/P0000050;:@/P0500000032WK40`Yfa0<:M/@000000103ooolb000004L0oooo00@0ooo405lD0000
55l0a?ooEP3oool006l0oooo2@00001d0?ooo`0300000022L[T000000080Vh>h00L000000:J8]@2V
R;D0YXRe0000002^SkD0000000<0/I6e00@000000;>A/`2cTK<00000102bTK@00`000000^IFb0000
00060;:@/P0500000032WK40`Yfa0<:M/@000000403ooolH00000080TLkm1@00001:0?ooo`0308le
00000000000000D0000000<0=Hoo0?ooo`3oool0D`3oool007P0oooo2@00001/0?ooo`8000000P2K
PkP01`000000YXRe0:J8]@2VR;D000000:j?]@0000000`2aTKD010000000/i6c0;>A/`0000050;:A
]08000001P2bT;801@000000`Yfa0<:M/@32WK40000001T0oooo4`0000100?ooo`030?ooa01O5000
000000L0000000D055o40?ooo`3oool0olAO03EOa0020?ooo`040?ooa01O500001AO0<CooeH0oooo
00210?ooo`T00000H`3oool200000080Vh>h00L000000:J8]@2VR;D0YXRe0000002^SkD0000000@0
/I6e00<000000;>A/`0000001@2bTK@2000000H0/Y2b00D000000<:M/@32WK40`Yfa0000000I0?oo
o`030000003oool0oooo00<0oooo1P00001E0?ooo`070<AO500eSol0oooo0?ooa01O500001AO0<Co
o`1F0?ooo`00RP3oool:000005T0oooo0P00000309^3^0060000002VR;D0YXRe0000002^SkD00000
102aTKD00`000000/i6c000000050;:A]08000001P2bT;801@000000`Yfa0<:M/@32WK4000000140
oooo00<000000?ooo`3oool01@3oool6000005/0oooo00L0ooo405lD=@2?ool0ooo405lD000055l0
a?oo05H0oooo002D0?ooo`T00000D@3oool200000080Vh>h00H000000:J8]@2VR;D000000:j?]@00
00040;6A]@030000002cTK<0000000D0/Y6d0P0000060;:@/P0500000032WK40`Yfa0<:M/@000000
4@3oool010000000oooo0?ooo`3oool5000003`0oooo00<0ScD0000000000000100000000`0DGl@0
oooo0?ooo`0K0?ooo`060?o4G`0D=Hl0ooo405lD000055l0a?ooEP3oool009d0oooo2@0000180?oo
o`8000000P2KPkP01@000000YXRe0:J8]@2VR;D0000000D0/I6e0P0000060;:A]08000001P2bT;80
1@000000`Yfa0<:M/@32WK40000000T0oooo00<000000?ooo`3oool0103oool500000480oooo00<0
a5lD0000000000000`0000000`0055l0a?oo0?ooo`0L0?ooo`050<AO500eSl@0Ga@0000DG`34ool0
EP3oool00:H0oooo2@00000o0?ooo`8000000P2KPkP01@000000YXRe0:J8]@2VR;D0000000D0/I6e
0P0000070;:A]0030000002bT;80/Y2b00<0/Y2b00<000000;6OaP0000000P32WK400`000000oooo
0?ooo`070?ooo`L00000B03oool00`34Ga@055o40?ooo`020?ooo`030<AO=@1Oa?l0oooo01`0oooo
00D0ooo405lD500D500001AO0<Coo`1F0?ooo`00[`3oool:000003H0oooo00D0000009^3^02KPkP0
Vh>h000000020:J8]@030000002aTKD0/I6e00@0/I6e00<000000;:A]02bTK@01@2bTK@00`000000
/Y2b0;:@/P020;:@/P090000002aWlH0/Io60;6OaP000000`Yfa0000003oool0000000<0oooo1@00
001?0?ooo`030<AO500eSol0oooo0240oooo00@0olAO01@0000055l0a?ooEP3oool00;T0oooo2@00
000]0?ooo`8000000P2KPkP00`000000YXRe00000002000000D0/I6e00<000000;:A]02bTK@0102b
TK@3000000<0/Y2b0P0000030;6OaPL00000A03oool00`3ool@0Ga@0000000070000000301AOa03o
ool0oooo00@0oooo00<0ScDD05o4o`3oool08@3oool00`34Ga@001AO0<Coo`1F0?ooo`00`P3oool9
000002D0oooo00<0000009^3^00000000P0000030:j`fAT00000FP3oool00`3ool@0Ga@e08ooo`1j
0?ooo`00b`3oool9000001`0oooo5@000002097>o@D00000H03oool00`3oScD003F?0?ooo`1i0?oo
o`00e03oool9000001@0oooo6000000h0?ooo`050?o4G`0D000000000000=@2?ool09P3oool00`3o
ohl0=@0D05o4o`1i0?ooo`00g@3oool:000000d0oooo00<000000?ooo`0000002`00000n0?ooo`07
0?ooS`0e0000=Hoo0?ooo`3oScD0000e08ooo`0U0?ooo`030?ooS`0e000055o407T0oooo003W0?oo
o`T00000103oool00`000000oooo0?ooo`05000004@0oooo00<0ScD0000DG`34ool00P3oool00`2?
=@0003F?0?ooo`0O0?ooo`030?n?G`1Oa?l0oooo0080oooo00<0olAO01@0000DGl@0N@3oool00?00
oooo1`0000180?ooo`030?ooS`0e000055o400<0oooo00<0a5lD0000=@2?ool07`3oool0203ool@0
Ga@D05o4o`3oool0oon?03D0000003D0SoooN@3oool00?l0oooo@03oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo0200oooo00H0oooO09a80000000000000000M02ooomj0?ooo`00o`3o
oom00?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0X03oool00?l0oooo7`3oool0
0`2oM0000000000000040000000304RLg`3oool0oooo01L0oooo00<0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`2P0?ooo`00o`3ooolP0?ooo`030=nLB0000000000000<0000000<004RL0=oo
o`3oool05`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo0:00oooo003o0?ooob40
oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`0G0?ooo`030?oo_`1d0000B9cO00<0
oooo00<0gia80000M02oool0X03oool00?l0oooo8P3oool00`3OW4P0M;oo0?ooo`0L0?ooo`030=nL
B000B9`0gooo0080oooo00<0_g@0001d_`3oool0X03oool00?l0oooo8`3oool00`2oM4P0W=oo0?oo
o`0K0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`2Q0?ooo`00o`3ooolS0?ooo`030?oo
g`2LB7@0_ooo01`0oooo00D0oono07@000000000001d0;ooo`2R0?ooo`00o`3oool30?ooo`040?oo
g`2LB00004RL0=oooad0oooo00<0okmd001d_`3oool0`P3oool00?l0oooo0`3oool0103ooml0W4P0
0018W03OoolM0?ooo`030?oo_`1d04P0W=oo0<80oooo003o0?ooo`<0oooo00@0oooO09a80000B9`0
gooo7@3oool00`3ookl0M00004RLg`320?ooo`00oP3oool00`2oM0000000000000050000000307Bo
o`3oool0oooo01D0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`320?ooo`00oP3o
ool0203ogi`0M9cO0?ooo`3oool0oooO09a80000B9`0gooo603oool0203ooml0W4Q809cOo`3oool0
oono07@0000007@0_ooo`P3oool00?l0oooo00L0gia807Boo`3oool0oooO09a80000B9`0gooo01T0
oooo00H0oooO09a80000000000000000M02oooo30?ooo`00o`3oool01`3ooml0W4Qd0;ooo`3ooml0
W4P00018W03Oool0hP3oool00?l0oooo0@3oool01P3ogi`0B7Bo0?oog`2LB00004RL0=ooon80oooo
003o0?ooo`80oooo00D0gia807Bog`2LB00004RL0=ooo`3R0?ooo`00o`3oool20?ooo`050?oog`2L
B4P0B4P00018W03Oool0hP3oool00?l0oooo0`3oool0103ogi`0B0000018W03OoooR0?ooo`00o`3o
ool40?ooo`030=nLB000B9`0gooo0>80oooo003o0?ooonT0oooo003o0?ooonT0oooo003o0?ooonT0
oooo003o0?ooonT0oooo003o0?ooonT0oooo0000\
\>"],
  ImageRangeCache->{{{0, 487.125}, {224.188, 0}} -> {-0.073799, 0.190269, \
0.00392568, 0.00392568}}]
}, Closed]],

Cell["\<\
Copy the input cell above and practice varying the ViewPoint by modifying \
your copy.\
\>", "Exercise"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Parametric plots", "Section"],

Cell[CellGroupData[{

Cell["Two-dimensional parametric functions", "Subsection"],

Cell[TextData[{
  "Many functions are expressed parametrically.  Two variables connected by a \
single parameter can be plotted against each other using ",
  ButtonBox["ParametricPlot",
    ButtonStyle->"RefGuideLink"],
  ".  The basic syntax is ",
  StyleBox["ParametricPlot[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["{x[t],y[t]},{t,tmin,tmax}", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " with the usual options.  In the following example we define a plotting \
function with options, create a table of parametric plots, and display a \
graphics array.  Of course, you will probably apply similar techniques to \
problems more interesting than Lissajous figures!"
}], "Text"],

Cell[BoxData[
    \(lissajous[n_, \[Phi]_]\  := \ {Cos[t], 
        Sin[n\ t\  + \ \[Phi]]}\)], "Input"],

Cell[BoxData[
    \(plotLissajous[n_, \[Phi]_, opts___]\  := \ 
      ParametricPlot[Evaluate[lissajous[n, \[Phi]]], {t, 0, 2  \[Pi]}, 
        opts]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogetherArray[
        Evaluate[
          Table[plotLissajous[n, \[Phi], AspectRatio \[Rule] 1, 
              Axes \[Rule] False, DisplayFunction \[Rule] Identity], {n, 1, 
              4}, {\[Phi], 0, \[Pi], \[Pi]\/4}]]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .7963 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.190476 0.0189594 0.190476 [
[ 0 0 0 0 ]
[ 1 .7963 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .7963 L
0 .7963 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0189594 0.200176 0.195326 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.499999 0.476191 0.5 0.476191 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .5 m
.97618 .51477 L
.97614 .5282 L
.97606 .54367 L
.97597 .55825 L
.97572 .58404 L
.97536 .61175 L
.97483 .6418 L
.97414 .67299 L
.97253 .72676 L
.97039 .77857 L
.96795 .82295 L
.96478 .86624 L
.96081 .90566 L
.95851 .92294 L
.95624 .93693 L
.9536 .94992 L
.95218 .95559 L
.9506 .96097 L
.94905 .96535 L
.94758 .96874 L
.94614 .97139 L
.94539 .97252 L
.94458 .97356 L
.94315 .97496 L
.94179 .97579 L
.94094 .97607 L
.94013 .97619 L
.93859 .97598 L
.93717 .97534 L
.93556 .97411 L
.93386 .97226 L
.93222 .96999 L
.92924 .96467 L
.92591 .95713 L
.91795 .93318 L
.91017 .90328 L
.87685 .73277 L
.83371 .48308 L
.7871 .24986 L
.77275 .19248 L
.75886 .14454 L
.74621 .10772 L
.73202 .07434 L
.72534 .06158 L
.71811 .04986 L
.71181 .04145 L
.70489 .03411 L
.70082 .03072 L
.69703 .02818 L
Mistroke
.69339 .02629 L
.69153 .02552 L
.68951 .02486 L
.68751 .02436 L
.68567 .02404 L
.68205 .02381 L
.67875 .02405 L
.67525 .02476 L
.67188 .0259 L
.6688 .02731 L
.66186 .0318 L
.6554 .03758 L
.64925 .04446 L
.63539 .06474 L
.62018 .09409 L
.59132 .16787 L
.5591 .27266 L
.49321 .52714 L
.46107 .65311 L
.42748 .7734 L
.39669 .86543 L
.38184 .90147 L
.36848 .92848 L
.35462 .95053 L
.34831 .95842 L
.34168 .96522 L
.33831 .96806 L
.33462 .9707 L
.33123 .97267 L
.32816 .97409 L
.32453 .9753 L
.32254 .97574 L
.32066 .97603 L
.31732 .97619 L
.3137 .97586 L
.3101 .97502 L
.30673 .97377 L
.30304 .97187 L
.29909 .96921 L
.29108 .96187 L
.2838 .95287 L
.27066 .931 L
.25705 .90069 L
.2318 .82406 L
.20598 .71974 L
.16287 .49769 L
.12177 .25939 L
.10269 .15636 L
.09305 .1106 L
.08455 .07581 L
Mistroke
.07686 .05034 L
.07365 .0418 L
.07026 .03434 L
.06839 .03099 L
.06669 .02845 L
.06509 .02653 L
.06342 .02505 L
.06251 .02448 L
.06153 .02405 L
.06062 .02384 L
.05979 .02382 L
.05824 .02419 L
.05748 .0246 L
.05665 .02521 L
.05519 .02672 L
.05384 .02865 L
.05138 .03358 L
.04885 .04079 L
.04623 .05081 L
.04407 .06138 L
.04188 .07442 L
.03827 .10235 L
.03475 .13941 L
.03195 .17866 L
.02935 .22691 L
.02715 .28264 L
.02638 .30746 L
.02568 .33467 L
.02516 .3587 L
.02469 .38519 L
.02445 .4016 L
.02427 .41683 L
.02412 .43119 L
.02399 .44704 L
.02391 .46132 L
.02385 .4747 L
.02382 .48885 L
.02381 .50405 L
.02382 .51272 L
.02384 .52086 L
.02389 .53606 L
.02399 .55238 L
.02413 .56994 L
.02449 .60136 L
.02494 .62982 L
.02557 .66053 L
.02641 .69332 L
.0286 .75604 L
.03109 .80691 L
.03433 .85534 L
Mistroke
.03792 .8945 L
.04165 .92402 L
.04366 .93637 L
.04599 .94812 L
.04857 .95826 L
.0511 .96573 L
.05236 .96861 L
.05358 .97092 L
.05471 .97265 L
.05593 .97413 L
.05728 .97528 L
.05856 .97593 L
.06012 .97619 L
.06155 .97594 L
.06316 .97513 L
.06399 .97451 L
.0649 .97366 L
.0681 .96948 L
.06996 .96625 L
.07198 .96209 L
.07575 .95278 L
.08353 .92795 L
.09106 .89808 L
.16848 .47084 L
.21184 .25443 L
.23843 .15331 L
.2519 .1128 L
.26451 .08173 L
.27727 .0571 L
.28384 .04708 L
.29092 .03831 L
.29468 .0345 L
.29811 .03154 L
.30161 .02902 L
.30485 .02713 L
.30865 .02546 L
.31065 .02481 L
.31278 .0243 L
.31637 .02385 L
.31841 .02382 L
.32033 .02394 L
.32411 .0246 L
.32626 .02521 L
.32825 .02595 L
.33191 .02769 L
.33582 .03011 L
.34333 .03632 L
.35016 .04375 L
.36698 .06883 L
.38296 .10104 L
Mistroke
.41298 .18064 L
.44505 .28748 L
.50299 .51196 L
.56528 .74893 L
.59218 .83463 L
.62106 .9078 L
.63684 .93769 L
.64372 .94823 L
.65128 .95795 L
.65787 .96481 L
.66493 .97041 L
.66881 .97269 L
.67235 .97429 L
.67572 .97536 L
.67926 .97602 L
.68308 .97618 L
.68653 .97583 L
.68844 .97543 L
.6905 .97483 L
.69424 .97332 L
.69773 .9714 L
.7014 .96884 L
.70795 .96289 L
.71552 .95379 L
.72254 .94322 L
.73826 .912 L
.75171 .87707 L
.76604 .83162 L
.79123 .73221 L
.83503 .5095 L
.87539 .27558 L
.89439 .1712 L
.90366 .12558 L
.91286 .08574 L
.92051 .0583 L
.92437 .0469 L
.92835 .03718 L
.93019 .03348 L
.93189 .03054 L
.93352 .02818 L
.93499 .02645 L
.93637 .02521 L
.93781 .02432 L
.93922 .02387 L
.94075 .02388 L
.94224 .02443 L
.94363 .02545 L
.94485 .02677 L
.94616 .02864 L
.94754 .03119 L
Mistroke
.94899 .03449 L
.95149 .04193 L
.95391 .05144 L
.95607 .06214 L
.95983 .08659 L
.96342 .11868 L
.96685 .16049 L
.96986 .21071 L
.97215 .26278 L
.97383 .31528 L
.97497 .36559 L
.97544 .39417 L
.97565 .40942 L
.97583 .42572 L
.97595 .44 L
.97606 .45583 L
.97613 .47026 L
.97617 .48384 L
.97619 .5 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.217813 0.0189594 0.39418 0.195326 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.499999 0.476191 0.499998 0.476196 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .83672 m
.97618 .847 L
.97614 .85607 L
.97606 .86618 L
.97597 .87538 L
.97585 .88357 L
.97572 .89086 L
.97536 .90633 L
.97488 .92048 L
.97423 .9343 L
.97342 .94672 L
.97253 .95644 L
.97204 .96058 L
.97147 .96455 L
.97093 .96758 L
.97033 .9703 L
.96971 .9725 L
.96911 .97405 L
.96849 .97518 L
.96788 .97587 L
.96713 .97619 L
.96631 .97597 L
.96558 .97532 L
.96475 .97413 L
.96382 .97227 L
.96278 .96964 L
.96081 .9632 L
.95622 .94271 L
.95056 .91018 L
.93855 .82634 L
.91017 .60611 L
.87597 .36527 L
.85526 .24763 L
.84372 .19276 L
.83111 .14177 L
.81953 .10329 L
.80647 .06926 L
.80059 .05715 L
.79423 .04622 L
.7882 .0379 L
.78256 .03191 L
.7796 .02943 L
.77631 .02722 L
.77299 .02555 L
.77116 .02487 L
.76945 .02439 L
.76628 .02387 L
.76337 .02384 L
.76 .0243 L
.75678 .02525 L
Mistroke
.7539 .02651 L
.75081 .02829 L
.74522 .03258 L
.73852 .03953 L
.73239 .04752 L
.71886 .07038 L
.70434 .10224 L
.67777 .17716 L
.61557 .40765 L
.55409 .65432 L
.52691 .75264 L
.49728 .84433 L
.48088 .88624 L
.46575 .91848 L
.45132 .94297 L
.44335 .95373 L
.43606 .96178 L
.42837 .96838 L
.42434 .97105 L
.4201 .97326 L
.41641 .97469 L
.41298 .97559 L
.40923 .97612 L
.40523 .97613 L
.40295 .97589 L
.40086 .97551 L
.39856 .97492 L
.39613 .97408 L
.39206 .97222 L
.38759 .9695 L
.37937 .96265 L
.37172 .95413 L
.35668 .93135 L
.34048 .89807 L
.31156 .81712 L
.25958 .61285 L
.20918 .37222 L
.18428 .25387 L
.15904 .14665 L
.14797 .10675 L
.13675 .07253 L
.12753 .05018 L
.12241 .04036 L
.11784 .03338 L
.11565 .03066 L
.11331 .02822 L
.1113 .02654 L
.10912 .02515 L
.10683 .02421 L
.10558 .02393 L
Mistroke
.10443 .02381 L
.10238 .02396 L
.10021 .02463 L
.09822 .02572 L
.09608 .02744 L
.09408 .02955 L
.09224 .03194 L
.08884 .03757 L
.08531 .04516 L
.08134 .05594 L
.07729 .06961 L
.07038 .09969 L
.06345 .13962 L
.05647 .19152 L
.0456 .3018 L
.03739 .41836 L
.03391 .48095 L
.0307 .54934 L
.02838 .60883 L
.02643 .67007 L
.02514 .72251 L
.0246 .75095 L
.02422 .77656 L
.02409 .78761 L
.02398 .79898 L
.0239 .80847 L
.02385 .81861 L
.02381 .82928 L
.02381 .84034 L
.02383 .85044 L
.02388 .85933 L
.02396 .869 L
.02406 .87771 L
.0242 .8869 L
.02439 .89631 L
.02462 .90564 L
.02488 .914 L
.02546 .92835 L
.02611 .94012 L
.02681 .94959 L
.02769 .95837 L
.02874 .96582 L
.0294 .96918 L
.03004 .97169 L
.03078 .97381 L
.03117 .97462 L
.0316 .97532 L
.032 .97577 L
.03237 .97604 L
.03281 .97618 L
Mistroke
.03323 .97616 L
.03401 .97573 L
.03446 .97527 L
.03488 .97471 L
.03569 .97331 L
.03647 .97158 L
.03835 .96616 L
.0403 .95891 L
.04253 .94896 L
.05234 .89131 L
.06533 .79701 L
.09483 .56805 L
.11399 .43067 L
.13389 .3063 L
.15489 .19891 L
.1646 .15807 L
.17557 .11858 L
.18585 .08807 L
.19716 .06155 L
.20353 .0498 L
.20942 .04094 L
.21541 .03387 L
.22116 .02888 L
.22413 .02697 L
.22744 .02538 L
.22911 .02479 L
.23091 .02431 L
.23419 .02384 L
.2373 .02389 L
.24058 .02444 L
.24341 .02532 L
.24651 .02672 L
.25022 .02897 L
.2536 .03157 L
.26127 .03928 L
.26844 .04872 L
.27528 .05963 L
.30263 .11997 L
.33248 .21086 L
.38784 .42148 L
.44904 .66622 L
.47614 .76291 L
.50577 .85263 L
.52201 .8929 L
.53705 .92373 L
.55139 .94686 L
.55933 .95689 L
.56662 .9643 L
.57505 .97068 L
.57947 .97307 L
Mistroke
.58172 .97402 L
.58413 .97486 L
.58818 .97581 L
.59196 .97618 L
.59426 .97616 L
.59644 .97598 L
.59839 .97567 L
.60051 .97518 L
.60416 .97397 L
.60818 .97209 L
.61237 .96953 L
.61629 .96657 L
.62327 .95994 L
.63076 .95092 L
.6457 .92702 L
.65945 .89822 L
.69007 .8118 L
.71903 .70472 L
.74884 .57436 L
.79995 .32805 L
.82482 .21317 L
.84942 .11563 L
.85947 .08327 L
.86984 .05595 L
.87561 .04392 L
.88074 .03536 L
.88317 .03207 L
.88578 .0291 L
.88822 .02691 L
.89045 .02539 L
.89241 .02447 L
.8945 .02391 L
.89646 .02382 L
.89823 .02409 L
.90017 .0248 L
.90227 .02607 L
.90444 .02793 L
.90643 .03017 L
.91047 .03631 L
.91405 .04366 L
.91777 .05333 L
.92174 .06611 L
.92858 .0946 L
.93535 .13198 L
.94625 .21536 L
.95166 .26994 L
.95689 .33338 L
.96478 .45617 L
.96968 .55841 L
.97172 .61163 L
Mistroke
.97347 .66675 L
.97475 .7174 L
.97556 .76078 L
.97583 .78073 L
.97595 .79093 L
.97604 .80148 L
.97611 .81052 L
.97616 .82023 L
.97618 .82885 L
.97619 .83672 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.411817 0.0189594 0.588183 0.195326 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.499999 0.476191 0.499994 0.476197 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .97619 m
.97253 .91873 L
.96081 .7494 L
.94087 .50979 L
.92592 .36667 L
.91017 .24677 L
.90226 .19761 L
.89322 .14982 L
.88503 .11386 L
.8756 .08064 L
.86517 .05335 L
.85928 .04206 L
.85363 .0339 L
.85089 .03083 L
.84826 .02843 L
.84588 .0267 L
.84327 .02527 L
.84041 .02426 L
.8388 .02394 L
.83733 .02381 L
.83445 .02397 L
.83172 .02462 L
.8291 .0257 L
.82626 .02736 L
.82108 .03164 L
.81483 .03887 L
.80905 .04745 L
.7962 .07252 L
.7837 .10397 L
.7351 .27396 L
.68375 .49343 L
.62402 .73533 L
.59493 .83082 L
.56315 .91036 L
.55457 .92682 L
.54551 .94171 L
.53711 .9532 L
.52928 .96184 L
.52191 .96814 L
.51813 .97067 L
.51403 .97288 L
.51169 .97389 L
.50954 .97466 L
.50747 .97525 L
.50529 .97572 L
.5015 .97615 L
.49935 .97618 L
.49734 .97607 L
.49341 .97546 L
.49121 .97489 L
Mistroke
.48922 .97424 L
.48535 .97259 L
.4817 .97057 L
.47351 .96444 L
.46468 .95535 L
.44857 .93227 L
.43382 .90404 L
.40589 .83322 L
.37679 .73823 L
.26646 .2803 L
.2395 .17729 L
.21503 .10048 L
.20199 .06852 L
.19526 .05497 L
.18823 .04319 L
.18241 .03541 L
.17936 .03208 L
.1761 .02911 L
.17314 .02697 L
.17045 .02549 L
.1679 .0245 L
.16524 .02392 L
.16248 .02382 L
.16093 .02398 L
.15951 .02428 L
.15692 .02518 L
.15421 .02664 L
.1518 .02838 L
.14956 .03038 L
.14453 .03627 L
.1392 .0447 L
.13443 .05422 L
.12401 .08183 L
.11547 .11189 L
.10656 .15087 L
.09143 .23628 L
.07708 .34152 L
.06311 .46873 L
.04113 .72346 L
.02797 .91097 L
.02504 .95666 L
.02419 .97017 L
.02393 .97431 L
.02385 .97549 L
.02382 .97607 L
.02381 .97617 L
.02383 .97576 L
.024 .97318 L
.02463 .96311 L
.02715 .92359 L
Mistroke
.03855 .75815 L
.05776 .52435 L
.08192 .30321 L
.09779 .1973 L
.11448 .11583 L
.12426 .08107 L
.13372 .05579 L
.13934 .04445 L
.14453 .03628 L
.14746 .03261 L
.15065 .02936 L
.1537 .02697 L
.15652 .02536 L
.15929 .02434 L
.16193 .02386 L
.16431 .02383 L
.16693 .02424 L
.16979 .02519 L
.17124 .02588 L
.17285 .02679 L
.17842 .03116 L
.18369 .03696 L
.18945 .04505 L
.20006 .06442 L
.21306 .09521 L
.22711 .13598 L
.25312 .22746 L
.36224 .68406 L
.39289 .79321 L
.42225 .87734 L
.43863 .91397 L
.44729 .93008 L
.45646 .94461 L
.4642 .95478 L
.47272 .96373 L
.47646 .96691 L
.48048 .9698 L
.48422 .97201 L
.48774 .97367 L
.48976 .97443 L
.49194 .9751 L
.49575 .97589 L
.49791 .97612 L
.50019 .97619 L
.50427 .97588 L
.50805 .9751 L
.51158 .97394 L
.51545 .97218 L
.51959 .96976 L
.52892 .96219 L
Mistroke
.53773 .95243 L
.55419 .9275 L
.5698 .8961 L
.58414 .86095 L
.61628 .76256 L
.67692 .5229 L
.73089 .29111 L
.7561 .19303 L
.77916 .11688 L
.79147 .08365 L
.80456 .0553 L
.8102 .0456 L
.81613 .03718 L
.82165 .03109 L
.82659 .02714 L
.82932 .02559 L
.83186 .02458 L
.83462 .02394 L
.83755 .02382 L
.83902 .02398 L
.8406 .02431 L
.84346 .02536 L
.84623 .02692 L
.84876 .02885 L
.85327 .03346 L
.85817 .04026 L
.86319 .04924 L
.86783 .05939 L
.87803 .08839 L
.88699 .12188 L
.90577 .21853 L
.92102 .32609 L
.93363 .43674 L
.95638 .6911 L
.97595 .97231 L
.97619 .97619 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.60582 0.0189594 0.782187 0.195326 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.5 0.476191 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .83672 m
.97618 .82611 L
.97614 .81619 L
.97606 .80442 L
.97597 .793 L
.97572 .77201 L
.97536 .7483 L
.97414 .69139 L
.97253 .63574 L
.96784 .51671 L
.96081 .38951 L
.95154 .26864 L
.94094 .17081 L
.93406 .12407 L
.92725 .08836 L
.92328 .07178 L
.91879 .05636 L
.91439 .04446 L
.91017 .03578 L
.9081 .03245 L
.90589 .02951 L
.90397 .02749 L
.90184 .02577 L
.90063 .02504 L
.89951 .02452 L
.89822 .0241 L
.89698 .02387 L
.89476 .02388 L
.8935 .02413 L
.89232 .02451 L
.8902 .02554 L
.88819 .02694 L
.88596 .02893 L
.88353 .03163 L
.87849 .03886 L
.87376 .04751 L
.86518 .06738 L
.85546 .09555 L
.83482 .17094 L
.78422 .40442 L
.73337 .64412 L
.70653 .75367 L
.67649 .85379 L
.66242 .89117 L
.64735 .92391 L
.63435 .94589 L
.62688 .95584 L
.61993 .96333 L
.61233 .96956 L
.6083 .97203 L
Mistroke
.60624 .97307 L
.60401 .97403 L
.60187 .97478 L
.5999 .97533 L
.59603 .97602 L
.59251 .97619 L
.58878 .9759 L
.58536 .9752 L
.5817 .97401 L
.57823 .97246 L
.57508 .97069 L
.56732 .96492 L
.55998 .95761 L
.54372 .93527 L
.52866 .90732 L
.4945 .82081 L
.43594 .61581 L
.37255 .36001 L
.31507 .15511 L
.29981 .11263 L
.28404 .07617 L
.2708 .05228 L
.2634 .04187 L
.25659 .03428 L
.25314 .03119 L
.24951 .0285 L
.24642 .02668 L
.24305 .0252 L
.24116 .0246 L
.23945 .0242 L
.23747 .02391 L
.2356 .02381 L
.23229 .02405 L
.23043 .02442 L
.22872 .02492 L
.22568 .02616 L
.22283 .02775 L
.21973 .02996 L
.21641 .03288 L
.20971 .04057 L
.20364 .04963 L
.19313 .07017 L
.18191 .09905 L
.16024 .17571 L
.13846 .28077 L
.12008 .3904 L
.08416 .64997 L
.05636 .86357 L
.04675 .92642 L
.04232 .94995 L
Mistroke
.04007 .95981 L
.03807 .96706 L
.0363 .97201 L
.03534 .97397 L
.03449 .97524 L
.03373 .97594 L
.03306 .97619 L
.03233 .97601 L
.03166 .97539 L
.03105 .97438 L
.03043 .97289 L
.02992 .97127 L
.02939 .96912 L
.02884 .96637 L
.02828 .96293 L
.02736 .95544 L
.02646 .94508 L
.02603 .9388 L
.02562 .93154 L
.02529 .9247 L
.02497 .91663 L
.02471 .90873 L
.0245 .90108 L
.02433 .89333 L
.02416 .88447 L
.02403 .87516 L
.02392 .86493 L
.02386 .85553 L
.02382 .84514 L
.02381 .835 L
.02383 .82553 L
.02387 .81441 L
.02394 .80373 L
.02405 .79171 L
.02419 .77846 L
.02461 .75041 L
.02512 .7237 L
.02644 .66968 L
.02813 .61598 L
.03335 .49206 L
.04133 .35794 L
.04612 .29555 L
.05113 .24056 L
.06181 .1507 L
.06855 .10927 L
.07519 .07783 L
.08298 .05122 L
.08709 .04113 L
.08949 .03639 L
.09174 .03268 L
Mistroke
.0937 .03002 L
.09588 .02763 L
.0981 .02581 L
.09934 .02506 L
.1005 .02451 L
.10265 .02392 L
.10465 .02383 L
.10698 .02426 L
.10922 .02522 L
.11124 .0265 L
.11341 .02833 L
.11739 .03279 L
.12221 .04004 L
.12668 .04842 L
.137 .0732 L
.14863 .10896 L
.17094 .19488 L
.21791 .4148 L
.27485 .67939 L
.30257 .78673 L
.3182 .83803 L
.33322 .88027 L
.34735 .91328 L
.3629 .94173 L
.3698 .95166 L
.37717 .96041 L
.38349 .96638 L
.39049 .97134 L
.3942 .97327 L
.39767 .97463 L
.40148 .97564 L
.40362 .97598 L
.40556 .97615 L
.4095 .97609 L
.41172 .97582 L
.41377 .97542 L
.41754 .9743 L
.42154 .97257 L
.42495 .97068 L
.4286 .96821 L
.43522 .96259 L
.45034 .9444 L
.4587 .93123 L
.46783 .91444 L
.48448 .87764 L
.51576 .78932 L
.58241 .54263 L
.64141 .30555 L
.6702 .20184 L
.70052 .11179 L
Mistroke
.71413 .07997 L
.72827 .05375 L
.74013 .03769 L
.74678 .03125 L
.7497 .02903 L
.75292 .02704 L
.75628 .02545 L
.75811 .02481 L
.75982 .02435 L
.76287 .02388 L
.76613 .02387 L
.76786 .02408 L
.76974 .02447 L
.77169 .02506 L
.7735 .02578 L
.77674 .02749 L
.78017 .02988 L
.78691 .03639 L
.7932 .04465 L
.80473 .06547 L
.81578 .09251 L
.8257 .12283 L
.84715 .20824 L
.86731 .31319 L
.88479 .42247 L
.91933 .67726 L
.94531 .87531 L
.95039 .90907 L
.9553 .93789 L
.9594 .95763 L
.96124 .96477 L
.96285 .96982 L
.96422 .97314 L
.96489 .97436 L
.9656 .97534 L
.96628 .97595 L
.96699 .97619 L
.96764 .97603 L
.96821 .97555 L
.96886 .97457 L
.96944 .97327 L
.97004 .97141 L
.97065 .96894 L
.9716 .96367 L
.97211 .96004 L
.97255 .95623 L
.97343 .94659 L
.97415 .93575 L
.97471 .92465 L
.97521 .9111 L
Mistroke
.97564 .89479 L
.97583 .88496 L
.97596 .87561 L
.97606 .8665 L
.97614 .85626 L
.97618 .84624 L
.97619 .83672 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.799824 0.0189594 0.97619 0.195326 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.499999 0.476191 0.5 0.476191 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .5 m
.97618 .48523 L
.97614 .4718 L
.97606 .45633 L
.97597 .44175 L
.97572 .41596 L
.97536 .38825 L
.97483 .3582 L
.97414 .32701 L
.97253 .27324 L
.97039 .22143 L
.96795 .17705 L
.96478 .13376 L
.96081 .09434 L
.95851 .07706 L
.95624 .06307 L
.9536 .05008 L
.95218 .04441 L
.9506 .03903 L
.94905 .03465 L
.94758 .03126 L
.94614 .02861 L
.94539 .02748 L
.94458 .02644 L
.94315 .02504 L
.94179 .02421 L
.94094 .02393 L
.94013 .02381 L
.93859 .02402 L
.93717 .02466 L
.93556 .02589 L
.93386 .02774 L
.93222 .03001 L
.92924 .03533 L
.92591 .04287 L
.91795 .06682 L
.91017 .09672 L
.87685 .26723 L
.83371 .51692 L
.7871 .75014 L
.77275 .80752 L
.75886 .85546 L
.74621 .89228 L
.73202 .92566 L
.72534 .93842 L
.71811 .95014 L
.71181 .95855 L
.70489 .96589 L
.70082 .96928 L
.69703 .97182 L
Mistroke
.69339 .97371 L
.69153 .97448 L
.68951 .97514 L
.68751 .97564 L
.68567 .97596 L
.68205 .97619 L
.67875 .97595 L
.67525 .97524 L
.67188 .9741 L
.6688 .97269 L
.66186 .9682 L
.6554 .96242 L
.64925 .95554 L
.63539 .93526 L
.62018 .90591 L
.59132 .83213 L
.5591 .72734 L
.49321 .47286 L
.46107 .34689 L
.42748 .2266 L
.39669 .13457 L
.38184 .09853 L
.36848 .07152 L
.35462 .04947 L
.34831 .04158 L
.34168 .03478 L
.33831 .03194 L
.33462 .0293 L
.33123 .02733 L
.32816 .02591 L
.32453 .0247 L
.32254 .02426 L
.32066 .02397 L
.31732 .02381 L
.3137 .02414 L
.3101 .02498 L
.30673 .02623 L
.30304 .02813 L
.29909 .03079 L
.29108 .03813 L
.2838 .04713 L
.27066 .069 L
.25705 .09931 L
.2318 .17594 L
.20598 .28026 L
.16287 .50231 L
.12177 .74061 L
.10269 .84364 L
.09305 .8894 L
.08455 .92419 L
Mistroke
.07686 .94966 L
.07365 .9582 L
.07026 .96566 L
.06839 .96901 L
.06669 .97155 L
.06509 .97347 L
.06342 .97495 L
.06251 .97552 L
.06153 .97595 L
.06062 .97616 L
.05979 .97618 L
.05824 .97581 L
.05748 .9754 L
.05665 .97479 L
.05519 .97328 L
.05384 .97135 L
.05138 .96642 L
.04885 .95921 L
.04623 .94919 L
.04407 .93862 L
.04188 .92558 L
.03827 .89765 L
.03475 .86059 L
.03195 .82134 L
.02935 .77309 L
.02715 .71736 L
.02638 .69254 L
.02568 .66533 L
.02516 .6413 L
.02469 .61481 L
.02445 .5984 L
.02427 .58317 L
.02412 .56881 L
.02399 .55296 L
.02391 .53868 L
.02385 .5253 L
.02382 .51115 L
.02381 .49595 L
.02382 .48728 L
.02384 .47914 L
.02389 .46394 L
.02399 .44762 L
.02413 .43006 L
.02449 .39864 L
.02494 .37018 L
.02557 .33947 L
.02641 .30668 L
.0286 .24396 L
.03109 .19309 L
.03433 .14466 L
Mistroke
.03792 .1055 L
.04165 .07598 L
.04366 .06363 L
.04599 .05188 L
.04857 .04174 L
.0511 .03427 L
.05236 .03139 L
.05358 .02908 L
.05471 .02735 L
.05593 .02587 L
.05728 .02472 L
.05856 .02407 L
.06012 .02381 L
.06155 .02406 L
.06316 .02487 L
.06399 .02549 L
.0649 .02634 L
.0681 .03052 L
.06996 .03375 L
.07198 .03791 L
.07575 .04722 L
.08353 .07205 L
.09106 .10192 L
.16848 .52916 L
.21184 .74557 L
.23843 .84669 L
.2519 .8872 L
.26451 .91827 L
.27727 .9429 L
.28384 .95292 L
.29092 .96169 L
.29468 .9655 L
.29811 .96846 L
.30161 .97098 L
.30485 .97287 L
.30865 .97454 L
.31065 .97519 L
.31278 .9757 L
.31637 .97615 L
.31841 .97618 L
.32033 .97606 L
.32411 .9754 L
.32626 .97479 L
.32825 .97405 L
.33191 .97231 L
.33582 .96989 L
.34333 .96368 L
.35016 .95625 L
.36698 .93117 L
.38296 .89896 L
Mistroke
.41298 .81936 L
.44505 .71252 L
.50299 .48804 L
.56528 .25107 L
.59218 .16537 L
.62106 .0922 L
.63684 .06231 L
.64372 .05177 L
.65128 .04205 L
.65787 .03519 L
.66493 .02959 L
.66881 .02731 L
.67235 .02571 L
.67572 .02464 L
.67926 .02398 L
.68308 .02382 L
.68653 .02417 L
.68844 .02457 L
.6905 .02517 L
.69424 .02668 L
.69773 .0286 L
.7014 .03116 L
.70795 .03711 L
.71552 .04621 L
.72254 .05678 L
.73826 .088 L
.75171 .12293 L
.76604 .16838 L
.79123 .26779 L
.83503 .4905 L
.87539 .72442 L
.89439 .8288 L
.90366 .87442 L
.91286 .91426 L
.92051 .9417 L
.92437 .9531 L
.92835 .96282 L
.93019 .96652 L
.93189 .96946 L
.93352 .97182 L
.93499 .97355 L
.93637 .97479 L
.93781 .97568 L
.93922 .97613 L
.94075 .97612 L
.94224 .97557 L
.94363 .97455 L
.94485 .97323 L
.94616 .97136 L
.94754 .96881 L
Mistroke
.94899 .96551 L
.95149 .95807 L
.95391 .94856 L
.95607 .93786 L
.95983 .91341 L
.96342 .88132 L
.96685 .83951 L
.96986 .78929 L
.97215 .73722 L
.97383 .68472 L
.97497 .63441 L
.97544 .60583 L
.97565 .59058 L
.97583 .57428 L
.97595 .56 L
.97606 .54417 L
.97613 .52974 L
.97617 .51616 L
.97619 .5 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.0238095 0.212963 0.200176 0.38933 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.499999 0.476191 0.5 0.476191 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .5 m
.97618 .51108 L
.97614 .52115 L
.97606 .53277 L
.97597 .54374 L
.97572 .56317 L
.97536 .58415 L
.97483 .60706 L
.97414 .63106 L
.97253 .67319 L
.97039 .71509 L
.96795 .75254 L
.96081 .82967 L
.95621 .86465 L
.95129 .894 L
.94643 .91685 L
.94065 .93779 L
.93383 .9555 L
.92978 .96308 L
.92595 .96848 L
.92223 .97225 L
.92008 .97382 L
.91905 .97443 L
.91807 .97492 L
.91628 .9756 L
.91429 .97605 L
.91231 .97619 L
.9112 .97614 L
.91017 .97601 L
.90827 .97557 L
.90648 .97492 L
.90448 .97395 L
.90229 .9726 L
.89775 .96886 L
.89348 .96429 L
.88541 .95314 L
.87654 .93757 L
.85668 .89254 L
.83322 .82611 L
.73141 .47695 L
.67833 .30616 L
.61775 .15144 L
.60126 .11885 L
.58366 .08909 L
.56725 .06619 L
.55182 .04906 L
.53735 .03696 L
.52921 .03186 L
.52177 .02828 L
.51755 .02672 L
Mistroke
.51363 .02557 L
.50934 .02463 L
.50694 .02426 L
.50476 .02402 L
.50076 .02382 L
.49851 .02383 L
.49644 .02393 L
.49263 .02432 L
.48861 .02503 L
.48477 .026 L
.48059 .02737 L
.473 .03069 L
.46562 .03495 L
.45881 .03979 L
.44338 .05392 L
.42708 .07356 L
.40013 .11632 L
.3708 .17663 L
.3134 .331 L
.25503 .52391 L
.20421 .70277 L
.17931 .78659 L
.15441 .86259 L
.132 .91974 L
.12205 .94024 L
.11327 .95517 L
.1048 .96624 L
.1005 .97044 L
.09835 .97214 L
.09607 .97364 L
.09409 .97467 L
.09231 .97538 L
.09054 .97587 L
.08889 .97613 L
.08699 .97618 L
.08525 .97598 L
.08337 .97549 L
.0814 .97467 L
.07972 .97369 L
.0779 .97236 L
.07432 .96879 L
.0711 .96446 L
.0682 .95955 L
.06196 .94538 L
.0558 .92565 L
.05046 .90288 L
.04586 .87785 L
.03839 .82259 L
.03504 .78926 L
.03185 .74979 L
Mistroke
.02909 .70587 L
.02702 .66268 L
.02621 .64142 L
.02555 .62093 L
.02504 .6021 L
.0246 .58185 L
.02439 .57028 L
.02423 .55975 L
.02409 .54902 L
.02399 .53914 L
.0239 .52753 L
.02386 .52143 L
.02383 .51497 L
.02381 .50411 L
.02381 .49219 L
.02385 .47988 L
.02392 .46852 L
.02397 .46259 L
.02403 .45606 L
.02419 .44287 L
.02458 .41897 L
.02507 .39674 L
.0263 .35597 L
.02809 .31332 L
.03062 .26832 L
.0339 .22375 L
.03721 .18835 L
.04135 .15285 L
.0461 .12068 L
.05109 .09412 L
.05599 .07365 L
.06173 .05526 L
.0685 .03989 L
.07212 .03405 L
.07426 .03128 L
.07625 .02913 L
.07798 .02758 L
.07989 .02619 L
.08096 .02557 L
.08195 .02507 L
.08394 .02433 L
.08584 .02393 L
.08758 .02381 L
.08962 .02396 L
.09157 .02439 L
.09347 .02506 L
.09553 .02606 L
.09932 .02859 L
.10393 .03283 L
.10824 .03782 L
Mistroke
.11797 .05241 L
.12759 .07075 L
.16647 .17294 L
.26231 .5014 L
.31703 .68001 L
.34735 .76566 L
.38076 .8454 L
.41328 .90611 L
.42887 .92882 L
.44341 .94611 L
.45798 .95956 L
.4657 .9651 L
.47394 .96978 L
.47771 .9715 L
.4818 .97306 L
.4854 .97418 L
.48921 .97509 L
.49265 .97568 L
.49642 .97607 L
.49986 .97619 L
.50307 .9761 L
.50702 .97572 L
.50902 .97542 L
.51123 .975 L
.51512 .97403 L
.51878 .97286 L
.52699 .96931 L
.53585 .96408 L
.552 .95077 L
.56675 .93443 L
.59473 .8928 L
.62382 .83545 L
.68246 .68153 L
.73429 .51316 L
.83214 .17719 L
.85441 .11335 L
.8655 .08595 L
.87673 .06207 L
.88611 .04576 L
.89092 .0389 L
.89596 .03294 L
.90045 .02877 L
.90445 .02606 L
.90634 .02513 L
.90838 .0244 L
.91027 .02398 L
.91201 .02382 L
.91403 .02391 L
.91503 .02408 L
.91612 .02435 L
Mistroke
.91804 .02506 L
.91979 .026 L
.92184 .02743 L
.92367 .02905 L
.92767 .03376 L
.931 .03901 L
.9345 .04594 L
.9405 .06177 L
.94589 .08094 L
.95114 .10523 L
.95934 .15828 L
.96316 .192 L
.96623 .22544 L
.96909 .26385 L
.97165 .30827 L
.97343 .34887 L
.97469 .38775 L
.97523 .40964 L
.97561 .42968 L
.97578 .44085 L
.97593 .45298 L
.97604 .46446 L
.97612 .47505 L
.97617 .48693 L
.97619 .49369 L
.97619 .5 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.217813 0.212963 0.39418 0.38933 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.499999 0.476191 0.5 0.476193 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .83672 m
.97618 .84446 L
.97614 .85135 L
.97606 .85909 L
.97597 .86622 L
.97585 .87264 L
.97572 .87841 L
.97536 .89093 L
.97488 .90275 L
.97423 .91483 L
.97342 .92636 L
.97253 .93613 L
.97147 .94524 L
.97033 .95285 L
.96921 .95878 L
.96788 .96429 L
.96631 .96908 L
.96537 .97123 L
.96448 .97285 L
.96362 .97407 L
.96315 .97461 L
.96265 .97509 L
.9617 .97574 L
.96081 .97609 L
.96029 .97618 L
.95972 .97618 L
.95866 .97596 L
.95759 .97545 L
.95658 .97473 L
.95546 .97367 L
.95419 .97217 L
.9515 .96808 L
.94626 .95713 L
.9409 .94278 L
.91017 .82723 L
.83581 .49614 L
.78935 .31589 L
.76276 .23008 L
.733 .15107 L
.71769 .11775 L
.70297 .09049 L
.68841 .06815 L
.67479 .05141 L
.66171 .0391 L
.65507 .03424 L
.64771 .02995 L
.6403 .02678 L
.63613 .0255 L
.6323 .02464 L
.62874 .02411 L
Mistroke
.62485 .02383 L
.62117 .02385 L
.61775 .02411 L
.61381 .0247 L
.61183 .02512 L
.60965 .02566 L
.60218 .02824 L
.59796 .03018 L
.59343 .03265 L
.58524 .03809 L
.56906 .05248 L
.55427 .06973 L
.52432 .11583 L
.49589 .17211 L
.37673 .50007 L
.31896 .67865 L
.28944 .7644 L
.25876 .84436 L
.23068 .90527 L
.21879 .92647 L
.20611 .94553 L
.19438 .95949 L
.18375 .96873 L
.17799 .97228 L
.17529 .97356 L
.17283 .97452 L
.17 .97536 L
.16852 .97568 L
.16696 .97594 L
.16436 .97617 L
.16152 .97613 L
.15866 .97578 L
.15603 .97518 L
.15317 .9742 L
.15017 .97281 L
.14519 .96966 L
.13979 .96498 L
.13012 .95316 L
.12137 .93834 L
.11161 .91671 L
.10306 .8929 L
.08526 .82654 L
.07103 .75372 L
.05938 .67731 L
.0478 .58102 L
.0389 .48667 L
.03218 .39517 L
.02943 .34834 L
.0271 .30046 L
.02562 .26198 L
Mistroke
.02502 .24234 L
.0245 .2219 L
.02431 .21255 L
.02415 .20382 L
.02403 .19548 L
.02394 .18787 L
.02388 .18083 L
.02383 .17346 L
.02381 .16545 L
.02381 .15836 L
.02385 .15055 L
.02391 .14344 L
.024 .13585 L
.02414 .12795 L
.02432 .11995 L
.02453 .11263 L
.02501 .09962 L
.02558 .08832 L
.0262 .07863 L
.02699 .06885 L
.02795 .05946 L
.02903 .05119 L
.03033 .04347 L
.03169 .03732 L
.03305 .03269 L
.03468 .02868 L
.03566 .02693 L
.03661 .02566 L
.03715 .0251 L
.03766 .02467 L
.03865 .02409 L
.03921 .0239 L
.03975 .02382 L
.04034 .02382 L
.04098 .02393 L
.04201 .02433 L
.04316 .02507 L
.04429 .02607 L
.04535 .02726 L
.04744 .03017 L
.04977 .03424 L
.05549 .04725 L
.06114 .06336 L
.08662 .15913 L
.1629 .49839 L
.20619 .66837 L
.2327 .75627 L
.25856 .82842 L
.28519 .88795 L
.29919 .91312 L
Mistroke
.31464 .93595 L
.32321 .94639 L
.33118 .95467 L
.33871 .96124 L
.34705 .96711 L
.35455 .97114 L
.35817 .97266 L
.36163 .97386 L
.36554 .97491 L
.36913 .9756 L
.373 .97605 L
.3752 .97617 L
.37725 .97619 L
.38115 .976 L
.38312 .97578 L
.38528 .97546 L
.38912 .97466 L
.39271 .97364 L
.40083 .97041 L
.40957 .96551 L
.42568 .95273 L
.44044 .93688 L
.47032 .89348 L
.49873 .83939 L
.61787 .51666 L
.67573 .33739 L
.73624 .16785 L
.7623 .10863 L
.77556 .08321 L
.78919 .06107 L
.80087 .04571 L
.80692 .0392 L
.8133 .03349 L
.8191 .02938 L
.8243 .02662 L
.82678 .02562 L
.82945 .02479 L
.83195 .02423 L
.83424 .02392 L
.83703 .02381 L
.83966 .02397 L
.842 .02434 L
.84453 .02499 L
.84726 .02597 L
.85013 .02735 L
.85522 .03065 L
.86031 .03511 L
.865 .0403 L
.87355 .05255 L
.8834 .07153 L
Mistroke
.89206 .09293 L
.91017 .15407 L
.92488 .2233 L
.93702 .29718 L
.94852 .3857 L
.95868 .48534 L
.96621 .58064 L
.96934 .62981 L
.97203 .6804 L
.97383 .72269 L
.97458 .7443 L
.97523 .76682 L
.97568 .78671 L
.97583 .79518 L
.97597 .80427 L
.97606 .81272 L
.97613 .82042 L
.97617 .8284 L
.97619 .83672 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.411817 0.212963 0.588183 0.38933 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.47619 0.5 0.47619 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .97619 m
.96081 .84362 L
.94087 .689 L
.91017 .48676 L
.87383 .30007 L
.85535 .22549 L
.83383 .15477 L
.82139 .12142 L
.80953 .09455 L
.79812 .07301 L
.7852 .05361 L
.77953 .0467 L
.77336 .04025 L
.76797 .03552 L
.76203 .03127 L
.75854 .02922 L
.75528 .02762 L
.75188 .02625 L
.74879 .02527 L
.74574 .02455 L
.74284 .0241 L
.73978 .02385 L
.73645 .02384 L
.73306 .02413 L
.72995 .02464 L
.72635 .02552 L
.72292 .02665 L
.71679 .02936 L
.71008 .03331 L
.69467 .04613 L
.68019 .06263 L
.6501 .10935 L
.61739 .17636 L
.50002 .49994 L
.43434 .69199 L
.4059 .7676 L
.37519 .84013 L
.34781 .89439 L
.31917 .93819 L
.30363 .95553 L
.28973 .96681 L
.28311 .9707 L
.27964 .97233 L
.27598 .97374 L
.2728 .97472 L
.26931 .97551 L
.26615 .97596 L
.26321 .97617 L
.25981 .97614 L
.25792 .97599 L
Mistroke
.25621 .97578 L
.25288 .97515 L
.24977 .9743 L
.24615 .97299 L
.24284 .97149 L
.23543 .96703 L
.22947 .96233 L
.223 .95609 L
.2102 .94007 L
.19879 .92156 L
.18688 .89783 L
.16627 .84552 L
.14419 .7728 L
.12202 .68136 L
.08699 .49631 L
.0388 .15316 L
.02711 .05321 L
.02446 .02964 L
.02394 .025 L
.02383 .02396 L
.02381 .02381 L
.02382 .02394 L
.02394 .02501 L
.02416 .02692 L
.02486 .0332 L
.0277 .05842 L
.03947 .15865 L
.08471 .48243 L
.11881 .66652 L
.13878 .75226 L
.16201 .83286 L
.18546 .89467 L
.20811 .93699 L
.21942 .9521 L
.22549 .95863 L
.23201 .96446 L
.23775 .96859 L
.24083 .97043 L
.24416 .97212 L
.24722 .97342 L
.25005 .97439 L
.25276 .97512 L
.25562 .97569 L
.25895 .97608 L
.26083 .97618 L
.26258 .97618 L
.26579 .976 L
.26921 .97552 L
.27228 .97485 L
.27517 .97401 L
Mistroke
.28178 .97135 L
.28897 .96731 L
.2956 .96256 L
.3107 .94824 L
.32376 .93215 L
.33819 .9107 L
.36488 .86184 L
.42516 .71711 L
.49076 .52772 L
.55344 .34237 L
.5861 .25295 L
.62014 .17017 L
.65013 .1093 L
.66576 .08306 L
.68228 .05999 L
.69726 .04362 L
.7107 .0329 L
.71711 .0292 L
.72046 .02763 L
.72401 .02626 L
.72709 .02531 L
.73048 .02453 L
.73354 .02407 L
.73641 .02385 L
.73961 .02384 L
.74302 .02412 L
.74622 .02465 L
.74911 .02536 L
.7525 .02648 L
.75616 .02802 L
.76273 .03172 L
.76955 .03683 L
.77668 .04358 L
.78923 .05912 L
.80047 .07711 L
.81073 .09705 L
.83316 .15283 L
.85574 .22691 L
.87561 .30795 L
.90852 .47708 L
.95669 .8101 L
.97611 .97545 L
.97619 .97619 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.60582 0.212963 0.782187 0.38933 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.5 0.476191 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .83672 m
.97618 .82879 L
.97614 .82143 L
.97606 .81275 L
.97597 .80437 L
.97572 .78907 L
.97536 .77191 L
.97414 .73104 L
.97253 .69119 L
.96784 .60516 L
.96081 .50987 L
.95154 .4128 L
.94094 .32509 L
.92725 .23638 L
.91017 .15405 L
.89314 .09594 L
.88273 .07006 L
.87257 .05096 L
.86737 .04333 L
.86238 .03727 L
.85776 .03273 L
.85279 .02893 L
.85017 .02737 L
.84726 .02598 L
.8446 .02501 L
.84209 .02436 L
.84056 .02409 L
.83911 .02392 L
.83636 .02381 L
.83482 .02388 L
.83338 .02402 L
.8318 .02426 L
.83011 .02462 L
.82752 .02536 L
.82468 .02645 L
.81887 .02952 L
.81342 .0334 L
.80827 .03789 L
.79642 .05114 L
.78342 .06992 L
.75869 .11614 L
.73533 .17012 L
.61925 .51239 L
.56281 .68041 L
.53027 .76675 L
.49969 .83737 L
.46709 .89887 L
.43743 .94042 L
.42126 .9567 L
.41386 .96258 L
Mistroke
.406 .96769 L
.40163 .97002 L
.39768 .97181 L
.39368 .97332 L
.39 .97443 L
.3857 .97538 L
.38345 .97574 L
.38107 .976 L
.37708 .97619 L
.37481 .97615 L
.37271 .97603 L
.36888 .97556 L
.36695 .97521 L
.36482 .97474 L
.36108 .97368 L
.35758 .97243 L
.34972 .96868 L
.34132 .96324 L
.32608 .94953 L
.31231 .93284 L
.28653 .89054 L
.26024 .83261 L
.20874 .67741 L
.16531 .50857 L
.08997 .17339 L
.07452 .11016 L
.06717 .0832 L
.06 .05987 L
.05423 .04406 L
.05137 .03748 L
.04845 .03184 L
.04591 .02797 L
.04371 .02553 L
.04269 .02473 L
.04162 .02415 L
.04063 .02386 L
.03973 .02382 L
.0388 .02403 L
.03785 .02453 L
.03706 .02518 L
.03622 .02614 L
.03535 .02744 L
.03445 .02916 L
.03292 .03308 L
.03149 .03811 L
.03026 .04386 L
.02823 .05713 L
.02723 .06631 L
.02631 .07718 L
.02563 .08738 L
Mistroke
.02502 .09956 L
.02449 .11372 L
.02426 .12214 L
.0241 .13013 L
.02397 .13779 L
.02392 .14223 L
.02388 .14641 L
.02383 .15423 L
.02381 .16278 L
.02382 .17116 L
.02385 .176 L
.02388 .18058 L
.02392 .18548 L
.02396 .18999 L
.0241 .2003 L
.02444 .21923 L
.02489 .23779 L
.02631 .28113 L
.02853 .331 L
.03427 .42632 L
.04157 .51738 L
.06088 .68815 L
.07487 .77537 L
.0899 .84623 L
.10647 .90296 L
.11543 .92584 L
.12569 .94617 L
.13487 .95955 L
.14016 .96535 L
.14505 .96955 L
.1497 .97256 L
.15236 .97386 L
.15482 .9748 L
.157 .97543 L
.15941 .9759 L
.16195 .97616 L
.16433 .97617 L
.16714 .97591 L
.16872 .97564 L
.17015 .97532 L
.17296 .97447 L
.17562 .97342 L
.18168 .97013 L
.18832 .96517 L
.20074 .95239 L
.21245 .9365 L
.23554 .89574 L
.26091 .83915 L
.37068 .51889 L
.43584 .3234 L
Mistroke
.49626 .17131 L
.52642 .11212 L
.54228 .08644 L
.55905 .06374 L
.57363 .04793 L
.58134 .04112 L
.58961 .03503 L
.59716 .03059 L
.60405 .0275 L
.60735 .02634 L
.61095 .02533 L
.61435 .02461 L
.6175 .02414 L
.6214 .02384 L
.62336 .02381 L
.62552 .02386 L
.62932 .02418 L
.63287 .02475 L
.63501 .02522 L
.63704 .02575 L
.64084 .02697 L
.64934 .0308 L
.6567 .03534 L
.66348 .04054 L
.67868 .05578 L
.69109 .07192 L
.70461 .0933 L
.73138 .14731 L
.78245 .29232 L
.83236 .48163 L
.87363 .66333 L
.9116 .83334 L
.92638 .89327 L
.94015 .94056 L
.94606 .95665 L
.94878 .96283 L
.95117 .9675 L
.95322 .97083 L
.95537 .97357 L
.95645 .97461 L
.95743 .97535 L
.95844 .97588 L
.95949 .97616 L
.96055 .97614 L
.96148 .97585 L
.96198 .97558 L
.96252 .9752 L
.96346 .97426 L
.96431 .97311 L
.96518 .97159 L
Mistroke
.96666 .96814 L
.96825 .96291 L
.9696 .95688 L
.97094 .94899 L
.97221 .9391 L
.97317 .92934 L
.97406 .91752 L
.97476 .90522 L
.97529 .89291 L
.97565 .88107 L
.97582 .87403 L
.97595 .86722 L
.97606 .85985 L
.97613 .85293 L
.97617 .84513 L
.97619 .83672 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.799824 0.212963 0.97619 0.38933 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.499999 0.476191 0.5 0.476191 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .5 m
.97618 .48892 L
.97614 .47885 L
.97606 .46723 L
.97597 .45626 L
.97572 .43683 L
.97536 .41585 L
.97483 .39294 L
.97414 .36894 L
.97253 .32681 L
.97039 .28491 L
.96795 .24746 L
.96081 .17033 L
.95621 .13535 L
.95129 .106 L
.94643 .08315 L
.94065 .06221 L
.93383 .0445 L
.92978 .03692 L
.92595 .03152 L
.92223 .02775 L
.92008 .02618 L
.91905 .02557 L
.91807 .02508 L
.91628 .0244 L
.91429 .02395 L
.91231 .02381 L
.9112 .02386 L
.91017 .02399 L
.90827 .02443 L
.90648 .02508 L
.90448 .02605 L
.90229 .0274 L
.89775 .03114 L
.89348 .03571 L
.88541 .04686 L
.87654 .06243 L
.85668 .10746 L
.83322 .17389 L
.73141 .52305 L
.67833 .69384 L
.61775 .84856 L
.60126 .88115 L
.58366 .91091 L
.56725 .93381 L
.55182 .95094 L
.53735 .96304 L
.52921 .96814 L
.52177 .97172 L
.51755 .97328 L
Mistroke
.51363 .97443 L
.50934 .97537 L
.50694 .97574 L
.50476 .97598 L
.50076 .97618 L
.49851 .97617 L
.49644 .97607 L
.49263 .97568 L
.48861 .97497 L
.48477 .974 L
.48059 .97263 L
.473 .96931 L
.46562 .96505 L
.45881 .96021 L
.44338 .94608 L
.42708 .92644 L
.40013 .88368 L
.3708 .82337 L
.3134 .669 L
.25503 .47609 L
.20421 .29723 L
.17931 .21341 L
.15441 .13741 L
.132 .08026 L
.12205 .05976 L
.11327 .04483 L
.1048 .03376 L
.1005 .02956 L
.09835 .02786 L
.09607 .02636 L
.09409 .02533 L
.09231 .02462 L
.09054 .02413 L
.08889 .02387 L
.08699 .02382 L
.08525 .02402 L
.08337 .02451 L
.0814 .02533 L
.07972 .02631 L
.0779 .02764 L
.07432 .03121 L
.0711 .03554 L
.0682 .04045 L
.06196 .05462 L
.0558 .07435 L
.05046 .09712 L
.04586 .12215 L
.03839 .17741 L
.03504 .21074 L
.03185 .25021 L
Mistroke
.02909 .29413 L
.02702 .33732 L
.02621 .35858 L
.02555 .37907 L
.02504 .3979 L
.0246 .41815 L
.02439 .42972 L
.02423 .44025 L
.02409 .45098 L
.02399 .46086 L
.0239 .47247 L
.02386 .47857 L
.02383 .48503 L
.02381 .49589 L
.02381 .50781 L
.02385 .52012 L
.02392 .53148 L
.02397 .53741 L
.02403 .54394 L
.02419 .55713 L
.02458 .58103 L
.02507 .60326 L
.0263 .64403 L
.02809 .68668 L
.03062 .73168 L
.0339 .77625 L
.03721 .81165 L
.04135 .84715 L
.0461 .87932 L
.05109 .90588 L
.05599 .92635 L
.06173 .94474 L
.0685 .96011 L
.07212 .96595 L
.07426 .96872 L
.07625 .97087 L
.07798 .97242 L
.07989 .97381 L
.08096 .97443 L
.08195 .97493 L
.08394 .97567 L
.08584 .97607 L
.08758 .97619 L
.08962 .97604 L
.09157 .97561 L
.09347 .97494 L
.09553 .97394 L
.09932 .97141 L
.10393 .96717 L
.10824 .96218 L
Mistroke
.11797 .94759 L
.12759 .92925 L
.16647 .82706 L
.26231 .4986 L
.31703 .31999 L
.34735 .23434 L
.38076 .1546 L
.41328 .09389 L
.42887 .07118 L
.44341 .05389 L
.45798 .04044 L
.4657 .0349 L
.47394 .03022 L
.47771 .0285 L
.4818 .02694 L
.4854 .02582 L
.48921 .02491 L
.49265 .02432 L
.49642 .02393 L
.49986 .02381 L
.50307 .0239 L
.50702 .02428 L
.50902 .02458 L
.51123 .025 L
.51512 .02597 L
.51878 .02714 L
.52699 .03069 L
.53585 .03592 L
.552 .04923 L
.56675 .06557 L
.59473 .1072 L
.62382 .16455 L
.68246 .31847 L
.73429 .48684 L
.83214 .82281 L
.85441 .88665 L
.8655 .91405 L
.87673 .93793 L
.88611 .95424 L
.89092 .9611 L
.89596 .96706 L
.90045 .97123 L
.90445 .97394 L
.90634 .97487 L
.90838 .9756 L
.91027 .97602 L
.91201 .97618 L
.91403 .97609 L
.91503 .97592 L
.91612 .97565 L
Mistroke
.91804 .97494 L
.91979 .974 L
.92184 .97257 L
.92367 .97095 L
.92767 .96624 L
.931 .96099 L
.9345 .95406 L
.9405 .93823 L
.94589 .91906 L
.95114 .89477 L
.95934 .84172 L
.96316 .808 L
.96623 .77456 L
.96909 .73615 L
.97165 .69173 L
.97343 .65113 L
.97469 .61225 L
.97523 .59036 L
.97561 .57032 L
.97578 .55915 L
.97593 .54702 L
.97604 .53554 L
.97612 .52495 L
.97617 .51307 L
.97619 .50631 L
.97619 .5 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.0238095 0.406966 0.200176 0.583333 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.499999 0.476192 0.5 0.476191 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .5 m
.97618 .50739 L
.97614 .51411 L
.97606 .52186 L
.97597 .52918 L
.97572 .54219 L
.97536 .55627 L
.97483 .57172 L
.97414 .58801 L
.97253 .61696 L
.97039 .64637 L
.96795 .67338 L
.96081 .73237 L
.95129 .78805 L
.94065 .83409 L
.92613 .88039 L
.91801 .90044 L
.91017 .91674 L
.89406 .94247 L
.88454 .95362 L
.87515 .96212 L
.86608 .96824 L
.85743 .97234 L
.8528 .97391 L
.8501 .97462 L
.84762 .97516 L
.84498 .9756 L
.84213 .97594 L
.83945 .97613 L
.8369 .97619 L
.83437 .97614 L
.83161 .97597 L
.82899 .9757 L
.82656 .97535 L
.82386 .97485 L
.82088 .97418 L
.8148 .9724 L
.8087 .9701 L
.80209 .96704 L
.78993 .95999 L
.7653 .94063 L
.74199 .91683 L
.68829 .84589 L
.6267 .74426 L
.49969 .49938 L
.3773 .26289 L
.31842 .16428 L
.2673 .09396 L
.24225 .06655 L
.21605 .04411 L
Mistroke
.20222 .03525 L
.19564 .03185 L
.18976 .02927 L
.18415 .02726 L
.17818 .0256 L
.17493 .02491 L
.17197 .02443 L
.16916 .02409 L
.16622 .02388 L
.16349 .02381 L
.16196 .02382 L
.16054 .02387 L
.15797 .02405 L
.15526 .02436 L
.15286 .02475 L
.15063 .02521 L
.1456 .02658 L
.14028 .02858 L
.13551 .03089 L
.12506 .03771 L
.11694 .04488 L
.10841 .05437 L
.09236 .07857 L
.084 .0951 L
.07659 .1125 L
.06396 .1495 L
.05222 .19527 L
.04174 .25085 L
.03454 .30348 L
.03126 .3348 L
.02873 .36482 L
.02663 .39712 L
.02521 .42705 L
.02464 .44372 L
.0244 .45254 L
.02419 .46195 L
.02404 .47015 L
.02398 .47451 L
.02392 .47922 L
.02385 .48747 L
.02381 .49523 L
.02381 .5031 L
.02383 .51026 L
.0239 .51851 L
.02399 .52631 L
.0241 .53352 L
.02423 .54016 L
.02461 .55515 L
.02511 .57021 L
.0258 .5867 L
Mistroke
.02765 .6197 L
.03002 .65131 L
.0332 .68448 L
.04058 .74171 L
.04995 .79416 L
.06247 .84538 L
.07641 .88706 L
.09115 .91921 L
.09956 .93341 L
.10917 .94656 L
.11871 .95683 L
.12787 .96438 L
.13649 .96963 L
.1415 .97193 L
.14627 .97363 L
.14889 .97437 L
.15139 .97496 L
.15369 .97539 L
.15614 .97575 L
.15883 .97602 L
.16133 .97616 L
.16405 .97619 L
.1656 .97615 L
.16707 .97607 L
.16972 .97585 L
.17258 .97548 L
.1753 .97502 L
.17779 .9745 L
.18395 .97281 L
.18986 .97068 L
.20187 .96494 L
.21525 .95646 L
.24007 .93558 L
.26386 .91011 L
.31677 .83824 L
.37815 .73559 L
.50357 .49285 L
.62415 .26028 L
.68287 .1623 L
.71268 .11943 L
.74328 .08173 L
.7719 .05356 L
.78493 .04341 L
.79662 .03591 L
.80812 .03015 L
.81405 .02786 L
.82021 .02599 L
.82277 .02538 L
.8255 .02483 L
.82783 .02446 L
Mistroke
.83036 .02414 L
.8333 .02391 L
.83605 .02381 L
.83872 .02384 L
.84116 .02398 L
.84387 .02425 L
.84644 .02463 L
.84872 .02506 L
.85119 .02565 L
.85665 .02736 L
.86163 .02945 L
.87043 .03445 L
.87948 .04151 L
.88978 .05219 L
.89902 .06446 L
.91499 .09294 L
.92915 .12804 L
.94219 .1718 L
.9517 .214 L
.96039 .26476 L
.9642 .29298 L
.96734 .32059 L
.96978 .34631 L
.97194 .37419 L
.97379 .40499 L
.97504 .43387 L
.97553 .45001 L
.97587 .46522 L
.97601 .47394 L
.9761 .48194 L
.97614 .48614 L
.97617 .49073 L
.97619 .49553 L
.97619 .5 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.217813 0.406966 0.39418 0.583333 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.5 0.476191 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .83672 m
.97618 .8419 L
.97614 .84654 L
.97606 .85182 L
.97597 .85672 L
.97585 .86117 L
.97572 .86522 L
.97536 .87415 L
.97488 .8828 L
.97423 .89194 L
.97341 .90113 L
.97253 .90911 L
.97036 .92409 L
.96747 .9382 L
.96429 .94936 L
.96081 .95822 L
.95863 .96247 L
.95654 .96584 L
.95417 .96894 L
.95144 .97168 L
.95016 .9727 L
.94878 .97364 L
.94746 .97438 L
.94621 .97496 L
.94501 .9754 L
.9437 .97576 L
.94221 .97604 L
.94083 .97617 L
.94006 .97619 L
.93919 .97617 L
.93828 .97611 L
.93742 .97601 L
.93585 .97573 L
.93495 .97552 L
.93413 .97529 L
.93035 .97384 L
.92839 .97286 L
.92621 .97159 L
.91816 .9655 L
.91017 .9576 L
.8924 .93496 L
.87389 .9059 L
.83501 .83329 L
.73931 .62597 L
.62085 .37198 L
.55809 .25485 L
.4984 .16103 L
.44117 .091 L
.40937 .06185 L
.38066 .04219 L
Mistroke
.36618 .03484 L
.3583 .0316 L
.35092 .02905 L
.34404 .02712 L
.34045 .02629 L
.33666 .02553 L
.33348 .02501 L
.32999 .02454 L
.32683 .02421 L
.3239 .02399 L
.32013 .02384 L
.31802 .02381 L
.31606 .02382 L
.31249 .02395 L
.3087 .02422 L
.30534 .02458 L
.30219 .02502 L
.29507 .0264 L
.28746 .02846 L
.28055 .03088 L
.2652 .03813 L
.25232 .0463 L
.23885 .057 L
.21535 .0812 L
.18875 .11786 L
.16599 .15793 L
.12575 .252 L
.08897 .37152 L
.06101 .49504 L
.04964 .55871 L
.03969 .62583 L
.0325 .68663 L
.0278 .73917 L
.02618 .7632 L
.0255 .77552 L
.02489 .78831 L
.02448 .79919 L
.0243 .80487 L
.02414 .81091 L
.02401 .81636 L
.02393 .8213 L
.02387 .82596 L
.02383 .83081 L
.02381 .83633 L
.02381 .83941 L
.02383 .84222 L
.02387 .84732 L
.02395 .85263 L
.02405 .85731 L
.02416 .86161 L
Mistroke
.02451 .87115 L
.02501 .88103 L
.02559 .88969 L
.02641 .89932 L
.02731 .90779 L
.02923 .92167 L
.0318 .93511 L
.03448 .94545 L
.03774 .95487 L
.03978 .95945 L
.04208 .96369 L
.04419 .96687 L
.04657 .96976 L
.04928 .97228 L
.05194 .97407 L
.05363 .97489 L
.05448 .97523 L
.05527 .97549 L
.05599 .97569 L
.05678 .97587 L
.05759 .97601 L
.05847 .97612 L
.05998 .97619 L
.06075 .97618 L
.06161 .97612 L
.06314 .97593 L
.0646 .97563 L
.06636 .97514 L
.06799 .97456 L
.07174 .97278 L
.07511 .97074 L
.07888 .96798 L
.08604 .96156 L
.10155 .94335 L
.12039 .91536 L
.16384 .8356 L
.26444 .6176 L
.37703 .3762 L
.43788 .26184 L
.50535 .1558 L
.53674 .11548 L
.56996 .07995 L
.601 .05399 L
.62871 .03723 L
.64345 .03095 L
.65081 .02853 L
.65883 .02645 L
.66566 .02514 L
.66946 .0246 L
.67291 .02423 L
Mistroke
.67612 .02399 L
.6796 .02384 L
.68287 .02381 L
.68588 .02388 L
.68934 .02406 L
.69304 .02439 L
.69653 .02483 L
.69971 .02534 L
.70742 .02701 L
.7147 .02916 L
.72895 .03505 L
.74421 .04391 L
.75838 .05461 L
.77144 .0667 L
.79505 .0943 L
.81921 .13091 L
.84377 .17784 L
.884 .28007 L
.91931 .40437 L
.94494 .52703 L
.96216 .64012 L
.96867 .69824 L
.97113 .72566 L
.97299 .75021 L
.97428 .77126 L
.97532 .79362 L
.9757 .80482 L
.97596 .81509 L
.97606 .82064 L
.97613 .82565 L
.97617 .83109 L
.97619 .83672 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.411817 0.406966 0.588183 0.583333 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.499999 0.476192 0.5 0.476191 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .97619 m
.96081 .91564 L
.91017 .73041 L
.83383 .49187 L
.739 .26371 L
.68397 .16596 L
.65198 .12082 L
.62174 .08606 L
.59046 .05818 L
.57381 .04669 L
.55605 .037 L
.5486 .03373 L
.54058 .03073 L
.53364 .02856 L
.52606 .02666 L
.52146 .02574 L
.51719 .02505 L
.5131 .02453 L
.51101 .02432 L
.50875 .02413 L
.50652 .02399 L
.50447 .02389 L
.50044 .02381 L
.49678 .02385 L
.49291 .02402 L
.48908 .02431 L
.48691 .02453 L
.48491 .02477 L
.47734 .02597 L
.47047 .02747 L
.46317 .02951 L
.44806 .03514 L
.43145 .04354 L
.4022 .06398 L
.37509 .08934 L
.31519 .16725 L
.26312 .25947 L
.16441 .49682 L
.09091 .72668 L
.04102 .90856 L
.02888 .956 L
.02537 .96995 L
.0243 .97422 L
.02384 .97605 L
.02381 .97618 L
.02381 .97617 L
.02385 .976 L
.02393 .9757 L
.0247 .97264 L
.02774 .96051 L
Mistroke
.04001 .91247 L
.08533 .74601 L
.16229 .50281 L
.25933 .26708 L
.31969 .16035 L
.34861 .12007 L
.3798 .08449 L
.4092 .05844 L
.42491 .04749 L
.44183 .03802 L
.44986 .03437 L
.45748 .0314 L
.4649 .02898 L
.47176 .02716 L
.47827 .02579 L
.48517 .02473 L
.48875 .02434 L
.49211 .02407 L
.4958 .02388 L
.49976 .02381 L
.50373 .02387 L
.50732 .02403 L
.51147 .02436 L
.51539 .0248 L
.52235 .02591 L
.52989 .02756 L
.5382 .02994 L
.54698 .03308 L
.56266 .0403 L
.57882 .0499 L
.59383 .06079 L
.62759 .09219 L
.65885 .12979 L
.68779 .17192 L
.7398 .26533 L
.83503 .49524 L
.91288 .73978 L
.95985 .91196 L
.97614 .97601 L
.97619 .97619 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.60582 0.406966 0.782187 0.583333 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.499999 0.476191 0.5 0.476191 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .83672 m
.97618 .83145 L
.97614 .8266 L
.97606 .82091 L
.97597 .81545 L
.97572 .80556 L
.97536 .79457 L
.97414 .76869 L
.97253 .7437 L
.96784 .68992 L
.96081 .62959 L
.94094 .50527 L
.91017 .36825 L
.87592 .25663 L
.83222 .15447 L
.80876 .11398 L
.78158 .07762 L
.76866 .06394 L
.75455 .05148 L
.74194 .04242 L
.72837 .03477 L
.72129 .03161 L
.71346 .02876 L
.70627 .02672 L
.69951 .0253 L
.69538 .02467 L
.69335 .02442 L
.69146 .02423 L
.68793 .02397 L
.68606 .02388 L
.68407 .02383 L
.6821 .02381 L
.67997 .02383 L
.67796 .0239 L
.67611 .02399 L
.67406 .02414 L
.67181 .02434 L
.66726 .0249 L
.66325 .02555 L
.65945 .02631 L
.65231 .02809 L
.63615 .03383 L
.62205 .04068 L
.60663 .05008 L
.57746 .07302 L
.54503 .10591 L
.479 .19427 L
.35462 .42184 L
.24448 .66212 L
.19444 .772 L
Mistroke
.14582 .87064 L
.12562 .90673 L
.1059 .93738 L
.08908 .95842 L
.08236 .96504 L
.07539 .97055 L
.07185 .97272 L
.06866 .97428 L
.06697 .97493 L
.0654 .97543 L
.06455 .97565 L
.06363 .97584 L
.06204 .97608 L
.06069 .97618 L
.05922 .97617 L
.05772 .97603 L
.05634 .97578 L
.05508 .97543 L
.05395 .97503 L
.05147 .9738 L
.0502 .97297 L
.04888 .97195 L
.04658 .96977 L
.04399 .9666 L
.04177 .96318 L
.03958 .95905 L
.03738 .95395 L
.03379 .94306 L
.03089 .93085 L
.02863 .9178 L
.02665 .90173 L
.02577 .89194 L
.0251 .88253 L
.02455 .87223 L
.02418 .86228 L
.02404 .85694 L
.02392 .85092 L
.02388 .84784 L
.02384 .84449 L
.02382 .84129 L
.02381 .83833 L
.02382 .83283 L
.02385 .82763 L
.02392 .82184 L
.02397 .81856 L
.02403 .81555 L
.02436 .80263 L
.02482 .79007 L
.02599 .76653 L
.02778 .73947 L
Mistroke
.03303 .68159 L
.0398 .62497 L
.0478 .57015 L
.07172 .44323 L
.1003 .33048 L
.1358 .22538 L
.15873 .17258 L
.18102 .13051 L
.2037 .09599 L
.22951 .06575 L
.24424 .05245 L
.25811 .04239 L
.27162 .03477 L
.27906 .03147 L
.28616 .02888 L
.29286 .02694 L
.29641 .0261 L
.3002 .02535 L
.3034 .02484 L
.30693 .02439 L
.31016 .0241 L
.31319 .02391 L
.31679 .02381 L
.3206 .02385 L
.32385 .02399 L
.32743 .02426 L
.3313 .0247 L
.33545 .02532 L
.34297 .02686 L
.35002 .02878 L
.35771 .03138 L
.37177 .03747 L
.39837 .05354 L
.42763 .07768 L
.46012 .1118 L
.52618 .20229 L
.65037 .43221 L
.75999 .67206 L
.80965 .78073 L
.85775 .87729 L
.8777 .91225 L
.89711 .94155 L
.90586 .95269 L
.91364 .96124 L
.92077 .9677 L
.92704 .97209 L
.93013 .97374 L
.93333 .97504 L
.93511 .97556 L
.93676 .97591 L
Mistroke
.93821 .97611 L
.93976 .97619 L
.94124 .97614 L
.94207 .97606 L
.94284 .97594 L
.94363 .97578 L
.94436 .9756 L
.94596 .97506 L
.94728 .97448 L
.94869 .9737 L
.95118 .9719 L
.95357 .96961 L
.95563 .96712 L
.95774 .96399 L
.95991 .96008 L
.96343 .95182 L
.96636 .94252 L
.96857 .93342 L
.97038 .92399 L
.97209 .91262 L
.97359 .8993 L
.97421 .89217 L
.9748 .8841 L
.97523 .87671 L
.9756 .86858 L
.97585 .86105 L
.97597 .85666 L
.97605 .85247 L
.97611 .84859 L
.97615 .84499 L
.97618 .84099 L
.97619 .83672 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.799824 0.406966 0.97619 0.583333 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.499999 0.476192 0.5 0.476191 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .5 m
.97618 .49261 L
.97614 .48589 L
.97606 .47814 L
.97597 .47082 L
.97572 .45781 L
.97536 .44373 L
.97483 .42828 L
.97414 .41199 L
.97253 .38304 L
.97039 .35363 L
.96795 .32662 L
.96081 .26763 L
.95129 .21195 L
.94065 .16591 L
.92613 .11961 L
.91801 .09956 L
.91017 .08326 L
.89406 .05753 L
.88454 .04638 L
.87515 .03788 L
.86608 .03176 L
.85743 .02766 L
.8528 .02609 L
.8501 .02538 L
.84762 .02484 L
.84498 .0244 L
.84213 .02406 L
.83945 .02387 L
.8369 .02381 L
.83437 .02386 L
.83161 .02403 L
.82899 .0243 L
.82656 .02465 L
.82386 .02515 L
.82088 .02582 L
.8148 .0276 L
.8087 .0299 L
.80209 .03296 L
.78993 .04001 L
.7653 .05937 L
.74199 .08317 L
.68829 .15411 L
.6267 .25574 L
.49969 .50062 L
.3773 .73711 L
.31842 .83572 L
.2673 .90604 L
.24225 .93345 L
.21605 .95589 L
Mistroke
.20222 .96475 L
.19564 .96815 L
.18976 .97073 L
.18415 .97274 L
.17818 .9744 L
.17493 .97509 L
.17197 .97557 L
.16916 .97591 L
.16622 .97612 L
.16349 .97619 L
.16196 .97618 L
.16054 .97613 L
.15797 .97595 L
.15526 .97564 L
.15286 .97525 L
.15063 .97479 L
.1456 .97342 L
.14028 .97142 L
.13551 .96911 L
.12506 .96229 L
.11694 .95512 L
.10841 .94563 L
.09236 .92143 L
.084 .9049 L
.07659 .8875 L
.06396 .8505 L
.05222 .80473 L
.04174 .74915 L
.03454 .69652 L
.03126 .6652 L
.02873 .63518 L
.02663 .60288 L
.02521 .57295 L
.02464 .55628 L
.0244 .54746 L
.02419 .53805 L
.02404 .52985 L
.02398 .52549 L
.02392 .52078 L
.02385 .51253 L
.02381 .50477 L
.02381 .4969 L
.02383 .48974 L
.0239 .48149 L
.02399 .47369 L
.0241 .46648 L
.02423 .45984 L
.02461 .44485 L
.02511 .42979 L
.0258 .4133 L
Mistroke
.02765 .3803 L
.03002 .34869 L
.0332 .31552 L
.04058 .25829 L
.04995 .20584 L
.06247 .15462 L
.07641 .11294 L
.09115 .08079 L
.09956 .06659 L
.10917 .05344 L
.11871 .04317 L
.12787 .03562 L
.13649 .03037 L
.1415 .02807 L
.14627 .02637 L
.14889 .02563 L
.15139 .02504 L
.15369 .02461 L
.15614 .02425 L
.15883 .02398 L
.16133 .02384 L
.16405 .02381 L
.1656 .02385 L
.16707 .02393 L
.16972 .02415 L
.17258 .02452 L
.1753 .02498 L
.17779 .0255 L
.18395 .02719 L
.18986 .02932 L
.20187 .03506 L
.21525 .04354 L
.24007 .06442 L
.26386 .08989 L
.31677 .16176 L
.37815 .26441 L
.50357 .50715 L
.62415 .73972 L
.68287 .8377 L
.71268 .88057 L
.74328 .91827 L
.7719 .94644 L
.78493 .95659 L
.79662 .96409 L
.80812 .96985 L
.81405 .97214 L
.82021 .97401 L
.82277 .97462 L
.8255 .97517 L
.82783 .97554 L
Mistroke
.83036 .97586 L
.8333 .97609 L
.83605 .97619 L
.83872 .97616 L
.84116 .97602 L
.84387 .97575 L
.84644 .97537 L
.84872 .97494 L
.85119 .97435 L
.85665 .97264 L
.86163 .97055 L
.87043 .96555 L
.87948 .95849 L
.88978 .94781 L
.89902 .93554 L
.91499 .90706 L
.92915 .87196 L
.94219 .8282 L
.9517 .786 L
.96039 .73524 L
.9642 .70702 L
.96734 .67941 L
.96978 .65369 L
.97194 .62581 L
.97379 .59501 L
.97504 .56613 L
.97553 .54999 L
.97587 .53478 L
.97601 .52606 L
.9761 .51806 L
.97614 .51386 L
.97617 .50927 L
.97619 .50447 L
.97619 .5 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.0238095 0.60097 0.200176 0.777337 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.499998 0.476192 0.5 0.476191 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .5 m
.97618 .50369 L
.97614 .50705 L
.97606 .51093 L
.97597 .5146 L
.97572 .52111 L
.97536 .52818 L
.97483 .53596 L
.97414 .5442 L
.97253 .55893 L
.97039 .57409 L
.96795 .58822 L
.96081 .62007 L
.95129 .65197 L
.94065 .68052 L
.91017 .74191 L
.87673 .79125 L
.83352 .83988 L
.78673 .88019 L
.73155 .9161 L
.7028 .93084 L
.67066 .94456 L
.64214 .95448 L
.61017 .96327 L
.5949 .96664 L
.57842 .96969 L
.5641 .97186 L
.54845 .97372 L
.53926 .97457 L
.53074 .9752 L
.52259 .97565 L
.51847 .97583 L
.51392 .97599 L
.50983 .97609 L
.50609 .97615 L
.50197 .97619 L
.4999 .97619 L
.49764 .97618 L
.49381 .97615 L
.48973 .97608 L
.48587 .97598 L
.48236 .97586 L
.47512 .97554 L
.46715 .97506 L
.45109 .97367 L
.43626 .9719 L
.42244 .96983 L
.39155 .96368 L
.3589 .95481 L
.32884 .94437 L
Mistroke
.29695 .93073 L
.24178 .9001 L
.19207 .86323 L
.14489 .81726 L
.10369 .764 L
.07215 .70905 L
.05955 .681 L
.04825 .65059 L
.03956 .62147 L
.03328 .59452 L
.02828 .56509 L
.02631 .54878 L
.0256 .54128 L
.02497 .5333 L
.02454 .52649 L
.02436 .52287 L
.02418 .51897 L
.02406 .5154 L
.02396 .51212 L
.02389 .509 L
.02384 .50571 L
.02381 .50211 L
.02381 .49827 L
.02384 .49465 L
.02388 .49136 L
.02397 .48749 L
.0241 .4833 L
.02443 .4757 L
.02486 .46828 L
.02546 .46033 L
.02688 .446 L
.02913 .42902 L
.03171 .41361 L
.03905 .38047 L
.04818 .3496 L
.07174 .29179 L
.10012 .24143 L
.13486 .19433 L
.17963 .14769 L
.22702 .10982 L
.28309 .07608 L
.3408 .05121 L
.37094 .04163 L
.38817 .03713 L
.40459 .03347 L
.41978 .03062 L
.43635 .02808 L
.45149 .02629 L
.4655 .02506 L
.47334 .02456 L
Mistroke
.48068 .0242 L
.48489 .02405 L
.48874 .02394 L
.49087 .0239 L
.49319 .02386 L
.49538 .02383 L
.49741 .02382 L
.49954 .02381 L
.50184 .02381 L
.50401 .02383 L
.50601 .02385 L
.5098 .02391 L
.5138 .02401 L
.5181 .02415 L
.52212 .02432 L
.53124 .02484 L
.5394 .02544 L
.54693 .02613 L
.56393 .02812 L
.57886 .03038 L
.59461 .0333 L
.62161 .0396 L
.65091 .04836 L
.70722 .07126 L
.76408 .10374 L
.81405 .14205 L
.86069 .18909 L
.89804 .23861 L
.92672 .28865 L
.94037 .31879 L
.95129 .34804 L
.95954 .37519 L
.9665 .4044 L
.9707 .42788 L
.97256 .44131 L
.97395 .45391 L
.97494 .4655 L
.97559 .47611 L
.97585 .48197 L
.97596 .48507 L
.97605 .48836 L
.97612 .49153 L
.97616 .4944 L
.97618 .49729 L
.97619 .5 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.217813 0.60097 0.39418 0.777337 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.5 0.476191 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .83672 m
.97618 .83932 L
.97614 .84167 L
.97606 .84436 L
.97597 .84688 L
.97572 .85132 L
.97536 .85606 L
.97483 .86118 L
.97414 .86652 L
.97253 .8758 L
.97039 .88501 L
.96795 .89327 L
.96081 .91074 L
.95621 .91911 L
.95129 .92657 L
.94065 .93923 L
.92718 .95085 L
.91865 .95648 L
.91017 .96109 L
.90184 .96481 L
.89232 .96826 L
.88286 .97094 L
.87376 .97293 L
.86521 .97432 L
.86023 .97494 L
.85549 .97541 L
.85289 .97561 L
.8504 .97578 L
.84813 .97591 L
.84569 .97602 L
.84302 .97611 L
.84053 .97616 L
.83783 .97619 L
.83628 .97619 L
.83482 .97618 L
.83218 .97615 L
.82934 .97608 L
.82664 .97598 L
.82417 .97587 L
.81804 .9755 L
.81217 .97501 L
.80022 .97365 L
.7869 .97161 L
.7622 .96648 L
.73848 .96008 L
.68577 .9414 L
.62453 .91306 L
.49926 .8362 L
.37865 .73979 L
.25927 .6203 L
Mistroke
.16096 .49671 L
.08822 .37793 L
.06068 .31927 L
.03953 .2602 L
.03235 .23279 L
.02944 .21887 L
.02702 .20457 L
.02545 .19237 L
.02479 .1856 L
.02433 .17939 L
.02416 .17648 L
.02403 .17375 L
.02394 .17132 L
.02387 .1687 L
.02382 .16587 L
.02381 .16445 L
.02381 .16289 L
.02383 .16015 L
.02388 .1576 L
.02396 .15503 L
.02407 .15232 L
.02436 .14748 L
.02459 .14457 L
.02484 .14184 L
.02541 .13682 L
.02695 .12691 L
.02919 .11662 L
.03162 .10808 L
.03467 .09945 L
.04139 .08506 L
.05028 .0713 L
.06177 .05838 L
.07441 .04801 L
.08744 .04012 L
.09517 .03644 L
.10383 .03304 L
.11168 .03052 L
.12052 .02826 L
.12561 .02719 L
.13038 .02634 L
.14002 .02503 L
.14528 .02453 L
.14834 .0243 L
.15115 .02413 L
.15415 .02399 L
.15582 .02393 L
.1574 .02388 L
.16023 .02383 L
.16184 .02381 L
.16333 .02381 L
Mistroke
.16612 .02383 L
.16879 .02387 L
.1712 .02394 L
.17385 .02404 L
.17675 .02418 L
.17985 .02436 L
.18551 .02478 L
.19687 .02597 L
.20743 .02745 L
.23447 .03273 L
.26112 .0398 L
.31766 .06001 L
.37251 .08542 L
.49505 .1598 L
.61331 .25307 L
.73125 .36917 L
.8294 .48976 L
.87144 .55194 L
.90992 .61851 L
.93694 .6751 L
.9493 .70615 L
.95883 .73434 L
.96643 .76201 L
.97151 .78631 L
.97343 .79853 L
.9749 .81104 L
.97545 .81741 L
.97582 .82324 L
.97598 .82656 L
.97608 .82954 L
.97613 .83118 L
.97616 .83295 L
.97618 .83462 L
.97619 .83614 L
.97619 .83672 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.411817 0.60097 0.588183 0.777337 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.499999 0.476192 0.499999 0.476192 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .97619 m
.96081 .96081 L
.91017 .91017 L
.83383 .83383 L
.739 .739 L
.62175 .62175 L
.50061 .50061 L
.372 .372 L
.25698 .25698 L
.16207 .16207 L
.08491 .08491 L
.03892 .03892 L
.02736 .02736 L
.02482 .02482 L
.0241 .0241 L
.02389 .02389 L
.02383 .02383 L
.02381 .02381 L
.02381 .02381 L
.02386 .02386 L
.02404 .02404 L
.02474 .02474 L
.02753 .02753 L
.04016 .04016 L
.08754 .08754 L
.15942 .15942 L
.25917 .25917 L
.37135 .37135 L
.49968 .49968 L
.62343 .62343 L
.73457 .73457 L
.83629 .83629 L
.91143 .91143 L
.95783 .95783 L
.97615 .97615 L
.97619 .97619 L
s
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.60582 0.60097 0.782187 0.777337 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.47619 0.5 0.476191 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .83672 m
.97618 .8341 L
.97614 .83169 L
.97606 .8289 L
.97597 .82624 L
.97572 .82146 L
.97536 .8162 L
.97483 .81033 L
.97414 .80401 L
.97253 .79246 L
.96795 .76851 L
.96081 .74094 L
.93889 .67971 L
.91017 .61898 L
.83562 .49845 L
.73253 .37058 L
.61698 .25632 L
.4982 .16201 L
.43147 .11833 L
.3721 .08522 L
.31177 .05761 L
.25924 .03925 L
.23434 .0327 L
.20915 .02773 L
.19655 .02593 L
.19063 .02527 L
.1851 .02475 L
.18201 .02451 L
.17867 .02429 L
.17592 .02413 L
.1729 .024 L
.17134 .02394 L
.16967 .02389 L
.16792 .02385 L
.16628 .02383 L
.16474 .02381 L
.16312 .02381 L
.16026 .02383 L
.15755 .02388 L
.15604 .02392 L
.15465 .02397 L
.15211 .02408 L
.14944 .02423 L
.14487 .02456 L
.13994 .02504 L
.13477 .02568 L
.12933 .02652 L
.1198 .02842 L
.11069 .03081 L
.10242 .03355 L
Mistroke
.08793 .03987 L
.07364 .04856 L
.0603 .05981 L
.04956 .07227 L
.0414 .08505 L
.0375 .0928 L
.03391 .10144 L
.03124 .10928 L
.02884 .11804 L
.02677 .12791 L
.02533 .1375 L
.02468 .14353 L
.02443 .14655 L
.02423 .14937 L
.02409 .15189 L
.02397 .1547 L
.02388 .15756 L
.02385 .15913 L
.02382 .16061 L
.02381 .16322 L
.02383 .16603 L
.02387 .16871 L
.02394 .17115 L
.02405 .17405 L
.0242 .17722 L
.02458 .18303 L
.02513 .18923 L
.02585 .19588 L
.02755 .20806 L
.03182 .23044 L
.03833 .25608 L
.0571 .31052 L
.08536 .37238 L
.16374 .50064 L
.26282 .62426 L
.37357 .73523 L
.50005 .83676 L
.56341 .87856 L
.62194 .91171 L
.67668 .93762 L
.73155 .95796 L
.75748 .96532 L
.77079 .96845 L
.78468 .97122 L
.79715 .97323 L
.80821 .97461 L
.81345 .97513 L
.81905 .97557 L
.8216 .97573 L
.82427 .97588 L
Mistroke
.82675 .97599 L
.82903 .97607 L
.83176 .97614 L
.83471 .97618 L
.8373 .97619 L
.84003 .97617 L
.84269 .97611 L
.84518 .97604 L
.84739 .97594 L
.84977 .97582 L
.85519 .97543 L
.86081 .97487 L
.87057 .9735 L
.87931 .97178 L
.88717 .9698 L
.9039 .96396 L
.91202 .96016 L
.92033 .95546 L
.93405 .94541 L
.94562 .9338 L
.95472 .9215 L
.95884 .91452 L
.96279 .90657 L
.96865 .89107 L
.97129 .88144 L
.97249 .876 L
.97345 .87083 L
.97481 .86137 L
.97534 .85623 L
.97577 .85063 L
.97595 .84725 L
.97603 .84533 L
.97609 .84355 L
.97613 .84198 L
.97616 .84025 L
.97618 .83843 L
.97619 .83672 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.799824 0.60097 0.97619 0.777337 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.499998 0.476192 0.5 0.476191 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .5 m
.97618 .49631 L
.97614 .49295 L
.97606 .48907 L
.97597 .4854 L
.97572 .47889 L
.97536 .47182 L
.97483 .46404 L
.97414 .4558 L
.97253 .44107 L
.97039 .42591 L
.96795 .41178 L
.96081 .37993 L
.95129 .34803 L
.94065 .31948 L
.91017 .25809 L
.87673 .20875 L
.83352 .16012 L
.78673 .11981 L
.73155 .0839 L
.7028 .06916 L
.67066 .05544 L
.64214 .04552 L
.61017 .03673 L
.5949 .03336 L
.57842 .03031 L
.5641 .02814 L
.54845 .02628 L
.53926 .02543 L
.53074 .0248 L
.52259 .02435 L
.51847 .02417 L
.51392 .02401 L
.50983 .02391 L
.50609 .02385 L
.50197 .02381 L
.4999 .02381 L
.49764 .02382 L
.49381 .02385 L
.48973 .02392 L
.48587 .02402 L
.48236 .02414 L
.47512 .02446 L
.46715 .02494 L
.45109 .02633 L
.43626 .0281 L
.42244 .03017 L
.39155 .03632 L
.3589 .04519 L
.32884 .05563 L
Mistroke
.29695 .06927 L
.24178 .0999 L
.19207 .13677 L
.14489 .18274 L
.10369 .236 L
.07215 .29095 L
.05955 .319 L
.04825 .34941 L
.03956 .37853 L
.03328 .40548 L
.02828 .43491 L
.02631 .45122 L
.0256 .45872 L
.02497 .4667 L
.02454 .47351 L
.02436 .47713 L
.02418 .48103 L
.02406 .4846 L
.02396 .48788 L
.02389 .491 L
.02384 .49429 L
.02381 .49789 L
.02381 .50173 L
.02384 .50535 L
.02388 .50864 L
.02397 .51251 L
.0241 .5167 L
.02443 .5243 L
.02486 .53172 L
.02546 .53967 L
.02688 .554 L
.02913 .57098 L
.03171 .58639 L
.03905 .61953 L
.04818 .6504 L
.07174 .70821 L
.10012 .75857 L
.13486 .80567 L
.17963 .85231 L
.22702 .89018 L
.28309 .92392 L
.3408 .94879 L
.37094 .95837 L
.38817 .96287 L
.40459 .96653 L
.41978 .96938 L
.43635 .97192 L
.45149 .97371 L
.4655 .97494 L
.47334 .97544 L
Mistroke
.48068 .9758 L
.48489 .97595 L
.48874 .97606 L
.49087 .9761 L
.49319 .97614 L
.49538 .97617 L
.49741 .97618 L
.49954 .97619 L
.50184 .97619 L
.50401 .97617 L
.50601 .97615 L
.5098 .97609 L
.5138 .97599 L
.5181 .97585 L
.52212 .97568 L
.53124 .97516 L
.5394 .97456 L
.54693 .97387 L
.56393 .97188 L
.57886 .96962 L
.59461 .9667 L
.62161 .9604 L
.65091 .95164 L
.70722 .92874 L
.76408 .89626 L
.81405 .85795 L
.86069 .81091 L
.89804 .76139 L
.92672 .71135 L
.94037 .68121 L
.95129 .65196 L
.95954 .62481 L
.9665 .5956 L
.9707 .57212 L
.97256 .55869 L
.97395 .54609 L
.97494 .5345 L
.97559 .52389 L
.97585 .51803 L
.97596 .51493 L
.97605 .51164 L
.97612 .50847 L
.97616 .5056 L
.97618 .50271 L
.97619 .5 L
Mfstroke
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 229.313},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001Eb000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`003P3oool2
00000100oooo1000000H0?ooo`@000004@3oool200000140oooo0`0000070?ooo`@00000903oool4
000000H0oooo0`00000I0?ooo`@00000;`3oool4000001T0oooo0P0000070?ooo`@00000903oool4
000000L0oooo0`00000A0?ooo`8000004@3oool4000001L0oooo1000000A0?ooo`8000003`3oool0
00h0oooo00<000000?ooo`0000003P3oool2000000<0oooo00<000000?ooo`3oool0503oool20000
00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0000000A0?ooo`040000003oool0oooo
000000D0oooo00@000000?ooo`3oool0oooo0P00000R0?ooo`8000000P3oool2000000D0oooo00@0
00000?ooo`3oool00000603oool010000000oooo0?ooo`00000^0?ooo`8000000P3oool2000001L0
oooo00@000000?ooo`3oool000001@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
0240oooo00<000000?ooo`3oool00P0000050?ooo`040000003oool0oooo00000140oooo00<00000
0?ooo`0000003`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01@0oooo00D00000
0?ooo`3oool0oooo0000000?0?ooo`030000003oool0000000l0oooo000=0?ooo`040000003oool0
oooo000000h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0B0?ooo`030000003o
ool0oooo00<0oooo0P00000>0?ooo`040000003oool0000000000100oooo00@000000?ooo`3oool0
0000103oool2000000@0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo00D000000?ooo`3oool0oooo0000000G0?ooo`040000003oool0oooo
0?ooo`800000;03oool2000000@0oooo00<000000?ooo`3oool05@3oool010000000oooo0?ooo`00
00050?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0803oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo0000000A0?ooo`030000
003oool0000000h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0B0?ooo`800000
1@3oool00`000000oooo0?ooo`0<0?ooo`040000003oool0oooo000000h0oooo000=0?ooo`040000
003oool0oooo000000d0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0A0?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool02`3oool01@000000oooo0?ooo`3oool00000
00l0oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000000D0oooo00<00000
0?ooo`3oool07`3oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000000`3o
ool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0:P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool00P3o
ool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool07`3oool00`000000oooo
0?ooo`040?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`0=0?ooo`04
0000003oool0oooo000000h0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool0303oool010000000oooo0?ooo`00000>
0?ooo`003@3oool010000000oooo0?ooo`00000=0?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool0403oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00X0oooo00D00000
0?ooo`3oool0oooo0000000?0?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo
000000060?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0103oool01@000000oooo
0?ooo`3oool0000000<0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo02X0oooo00<000000?ooo`3oool00`3oool2000001D0oooo00<000000?ooo`3o
ool00P3oool010000000oooo0?ooo`0000070?ooo`030000003oool0oooo01h0oooo00<000000?oo
o`3oool0103oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool03@3oool0
10000000oooo0?ooo`00000=0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00X0oooo00D000000?ooo`3oool0oooo
0000000>0?ooo`003@3oool010000000oooo0?ooo`00000<0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00X0oooo
00D000000?ooo`3oool0oooo0000000?0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3o
ool000001P3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo00H0oooo00@000000?oo
o`3oool00000103oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool0:@3oool00`000000oooo0?ooo`030?ooo`8000005@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`06
0?ooo`040000003oool0oooo000000@0oooo00<000000?ooo`3oool03@3oool01@000000oooo0?oo
o`3oool0000000`0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0?0?ooo`030000
003oool0oooo00H0oooo00<000000?ooo`3oool02P3oool01@000000oooo0?ooo`3oool0000000h0
oooo000=0?ooo`050000003oool0oooo0?ooo`0000002`3oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`090?ooo`05
0000003oool0oooo0?ooo`0000003`3oool00`000000oooo0?ooo`030?ooo`030000003oool00000
00P0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`060?ooo`030000003oool00000
00D0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
02P0oooo0P0000060?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo000000080?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool01P3oool01000
0000oooo0?ooo`0000040?ooo`030000003oool0oooo00d0oooo00D000000?ooo`3oool0oooo0000
000<0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo00T0oooo00D000000?ooo`3oool0oooo0000000>0?ooo`003@3o
ool01@000000oooo0?ooo`3oool0000000/0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`0=0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool01@000000oooo
0?ooo`3oool0000000h0oooo00<000000?ooo`3oool0103oool00`000000oooo000000080?ooo`03
0000003oool0oooo01`0oooo00<000000?ooo`3oool01P3oool00`000000oooo000000050?ooo`03
0000003oool0oooo01<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0W0?ooo`80
00001P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`00
0000203oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`00
00001@3oool00`000000oooo0?ooo`0=0?ooo`050000003oool0oooo0?ooo`0000002`3oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`090?ooo`050000003oool0oooo0?ooo`0000003P3oool000d0oooo00D000000?oo
o`3oool0oooo0000000:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00D000000?ooo`3oool0oooo0000
000>0?ooo`030000003oool0oooo00D0oooo0P0000080?ooo`030000003oool0oooo01/0oooo00<0
00000?ooo`3oool0203oool2000000D0oooo00<000000?ooo`3oool04P3oool2000000L0oooo00<0
00000?ooo`3oool09`3oool00`000000oooo0?ooo`050?ooo`8000004`3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?oo
o`070?ooo`030000003oool0000000D0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool01@000000oooo0?oo
o`3oool0000000h0oooo000=0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool04P3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo02H0oooo0P0000080?ooo`030000003o
ool0oooo0140oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool0203oool2000000D0oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00`0oooo000<0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo02D0oooo0P0000080?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00H0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo
01P0oooo00<000000?ooo`3oool02@3oool2000000D0oooo00<000000?ooo`3oool04P3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo02@0oooo0P0000090?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool01P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0103oool2000000/0
oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`090?ooo`8000001P3oool00`000000
oooo0?ooo`0@0?ooo`8000002@3oool00`000000oooo0?ooo`0T0?ooo`8000002@3oool200000180
oooo00<000000?ooo`3oool01@3oool2000000X0oooo00<000000?ooo`3oool05`3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`090?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0303oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`040?ooo`800000303oool00`000000oooo
0?ooo`0F0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`0000001P3oool00`000000oooo
0?ooo`0@0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0903oool2000000X0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`050?ooo`030000003oool0000000X0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0:0?ooo`8000001P3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`060?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02`3oool000`0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`0000002`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`0000001P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool08`3oool2000000X0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`050?ooo`030000003oool0000000X0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`090?ooo`030000003oool0000000H0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`040?ooo`030000003oool0000000/0oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`090?ooo`030000003oool0000000H0oooo00<00000
0?ooo`3oool0403oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo02<0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`0000002`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`0000001`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`0000002`3oool00`000000oooo0?ooo`0F0?oo
o`030000003oool0oooo00T0oooo00<000000?ooo`0000001P3oool00`000000oooo0?ooo`0@0?oo
o`030000003oool0oooo00T0oooo00<000000?ooo`3oool08P3oool2000000/0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`050?ooo`040000003oool0oooo000000X0oooo00<00000
0?ooo`3oool05@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0000000L0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00/0oooo000<0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`040?oo
o`040000003oool0oooo000000`0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`09
0?ooo`040000003oool0oooo000000L0oooo00<000000?ooo`3oool03P3oool2000000/0oooo0P00
000S0?ooo`8000002`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00@0oooo00D0
00000?ooo`3oool0oooo0000000:0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo000000070?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00l0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`00000<0?ooo`030000
003oool0oooo01@0oooo00<000000?ooo`3oool02@3oool010000000oooo0?ooo`0000070?ooo`03
0000003oool0oooo00h0oooo0P00000<0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3o
ool0oooo0000000;0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool02@3oool01000
0000oooo0?ooo`0000070?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`00000<0?ooo`030000003oool0
oooo01@0oooo00<000000?ooo`3oool02@3oool010000000oooo0?ooo`0000070?ooo`030000003o
ool0oooo00h0oooo0P00000<0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool02P3o
ool200000100oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000/0oooo
00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0:0?ooo`050000003oool0oooo0?ooo`00
00001P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool03`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool02`3oool000`0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0100oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0A0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0:0?ooo`030000
003oool0oooo00@0oooo00D000000?ooo`3oool0oooo0000000;0?ooo`030000003oool0oooo01<0
oooo00<000000?ooo`3oool02P3oool01@000000oooo0?ooo`3oool0000000H0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0240oooo00<000000?oo
o`3oool02P3oool200000100oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0
000000/0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0:0?ooo`050000003oool0
oooo0?ooo`0000001P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0:
0?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo0000000<0?ooo`030000003o
ool0oooo0180oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0<0?ooo`030000003o
ool0oooo01l0oooo0P00000<0?ooo`800000403oool00`000000oooo0?ooo`040?ooo`050000003o
ool0oooo0?ooo`000000303oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00X0oooo
00D000000?ooo`3oool0oooo000000060?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo0140oooo00D000000?ooo`3oool0oooo000000080?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000`0oooo00<000000?oo
o`3oool04P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool07`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool04P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool02`3oool000`0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo0140oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0A0?ooo`05
0000003oool0oooo0?ooo`000000203oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
00@0oooo00D000000?ooo`3oool0oooo0000000<0?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01l0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool04@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0140oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00/0oooo000<0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool04P3oool01@000000oooo0?ooo`3oool0
000000P0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0
oooo00X0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0O0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0
oooo00X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo0180oooo00D000000?ooo`3oool0oooo00000008
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0:
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool07@3oool2000000h0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`3oool01P3oool01@000000oooo0?ooo`3oool0000001<0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0B0?ooo`050000003oool0oooo0?ooo`0000001`3oool00`000000oooo
0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`050?ooo`050000003oool0oooo0?ooo`000000
503oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01<0oooo00@000000?ooo`3oool0
0000203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool02@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool03@3oool2000000l0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00H0oooo00@000000?ooo`3oool000005@3oool01@000000oooo0?ooo`3oool0
000001D0oooo00@000000?ooo`3oool00000203oool00`000000oooo0?ooo`0;0?ooo`00303oool0
0`000000oooo0?ooo`060?ooo`040000003oool0oooo000001D0oooo00D000000?ooo`3oool0oooo
0000000E0?ooo`040000003oool0oooo000000P0oooo00<000000?ooo`3oool02@3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`0@0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo0P00000?0?ooo`030000003oool0oooo01d0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`060?ooo`040000003oool0oooo0000
01D0oooo00D000000?ooo`3oool0oooo0000000E0?ooo`040000003oool0oooo000000P0oooo00<0
00000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool01P3oool010000000oooo0?ooo`00
000E0?ooo`040000003oool0oooo000001H0oooo00<000000?ooo`0000002@3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0303oool2000000l0oooo
00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0=0?ooo`8000003`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool01P3oool010000000oooo
0?ooo`00000E0?ooo`050000003oool0oooo0?ooo`0000005@3oool010000000oooo0?ooo`000008
0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00H0oooo00@000000?oo
o`3oool000005@3oool010000000oooo0?ooo`00000F0?ooo`030000003oool0000000T0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00X0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00h0oooo0P00
000?0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`060?ooo`040000003oool0oooo000001D0oooo00D000000?ooo`3oool0oooo0000000F0?ooo`03
0000003oool0000000P0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool0
1P3oool010000000oooo0?ooo`00000E0?ooo`040000003oool0oooo000001H0oooo00<000000?oo
o`0000002@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool0303oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01/0oooo00<000000?oo
o`3oool03P3oool2000000h0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`090?oo
o`030000003oool0oooo00L0oooo0P00000G0?ooo`030000003oool0000001L0oooo0P0000090?oo
o`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00H0oooo00@000000?ooo`3o
ool000005P3oool00`000000oooo0000000F0?ooo`030000003oool0000000T0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool02P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00h0oooo0P00000>0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`070?oo
o`8000005`3oool00`000000oooo0000000G0?ooo`8000002@3oool00`000000oooo0?ooo`0;0?oo
o`00303oool00`000000oooo0?ooo`060?ooo`030000003oool0000001L0oooo0P00000H0?ooo`80
00002@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool0303oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01X0oooo0P00000A0?ooo`03
0000003oool0oooo00`0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00L0oooo0P00000G0?ooo`030000003oool0000001L0oooo0P0000090?ooo`03
0000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00L0oooo0P00000G0?ooo`800000
603oool2000000T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00H0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo00X0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo0100oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00H0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo00X0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00T0oooo00<000000?ooo`3oool01`3oool2000001P0oooo0P00000H0?ooo`030000
003oool0oooo00L0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool01`3o
ool2000001L0oooo0P00000H0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool01P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00/0oooo0P00000B0?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0<0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00/0oooo000<0?ooo`030000003oool0oooo00L0oooo0P00000H0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool02`3oool200000180oooo00<000000?ooo`3oool06@3oool0
0`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0303oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`0F0?ooo`8000002@3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`00
0000oooo0?ooo`070?ooo`8000005`3oool2000001P0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool02`3oool20000
0180oooo00<000000?ooo`3oool0603oool200000180oooo0P00000=0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`070?ooo`800000603oool20000
01L0oooo0P0000090?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool05P3oool2000001P0oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`080?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool02`3oool2000001<0
oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0A0?ooo`8000003@3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool01`3oool2000001P0
oooo0P00000G0?ooo`8000002@3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo01H0oooo0P00000H0?ooo`8000002@3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0B0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool04@3oool2000000d0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00L0oooo
0P00000G0?ooo`030000003oool0000001L0oooo0P0000090?ooo`030000003oool0oooo00/0oooo
000<0?ooo`030000003oool0oooo00L0oooo0P00000G0?ooo`030000003oool0000001H0oooo00<0
00000?ooo`0000002@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool04@3oool2000000d0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00H0oooo00<000000?ooo`0000005`3oool00`000000oooo0000
000G0?ooo`030000003oool0000000P0oooo00<000000?ooo`3oool02`3oool000`0oooo00<00000
0?ooo`3oool01`3oool2000001L0oooo00<000000?ooo`0000005P3oool00`000000oooo00000009
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool05P3oool2000001<0oooo0P00000=
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`06
0?ooo`030000003oool0000001L0oooo00@000000?ooo`3oool000005@3oool010000000oooo0?oo
o`0000080?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00L0oooo0P00
000F0?ooo`040000003oool0oooo000001H0oooo00<000000?ooo`0000002@3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`0:0?ooo`050000003oool0oooo0?ooo`0000003@3oool00`000000oooo0?ooo`0C0?ooo`03
0000003oool0oooo01D0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`060?ooo`04
0000003oool0oooo000001H0oooo00@000000?ooo`3oool000005@3oool010000000oooo0?ooo`00
00080?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`0000005P3oool010000000oooo0?ooo`00000F0?ooo`030000003oool0000000T0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00X0oooo00<0
00000?ooo`3oool02P3oool01@000000oooo0?ooo`3oool0000000d0oooo00<000000?ooo`3oool0
4`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
1P3oool010000000oooo0?ooo`00000E0?ooo`050000003oool0oooo0?ooo`0000005@3oool01000
0000oooo0?ooo`0000080?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo
00H0oooo00@000000?ooo`3oool000005@3oool01@000000oooo0?ooo`3oool0000001D0oooo00<0
00000?ooo`0000002@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`050000003oool0oooo0?ooo`000000
3@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0
4`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`060?ooo`040000003oool0oooo000001D0oooo00D000000?ooo`3o
ool0oooo0000000E0?ooo`040000003oool0oooo000000P0oooo00<000000?ooo`3oool02`3oool0
00`0oooo00<000000?ooo`3oool01P3oool010000000oooo0?ooo`00000E0?ooo`050000003oool0
oooo0?ooo`0000005@3oool010000000oooo0?ooo`0000080?ooo`030000003oool0oooo00T0oooo
00D000000?ooo`3oool0oooo0000000<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00/0oooo00D000000?ooo`3o
ool0oooo0000000=0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`0C0?ooo`030000003oool0oooo00/0oooo00D000000?ooo`3oool0oooo0000000<
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0:
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00H0oooo00@000000?ooo`3oool000005@3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo0180oooo00@000000?ooo`3oool00000203oool00`000000oooo
0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`060?ooo`040000003oool0oooo000001@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0B0?ooo`050000003oool0oooo0?ooo`00
0000203oool00`000000oooo0?ooo`090?ooo`050000003oool0oooo0?ooo`000000303oool00`00
0000oooo0?ooo`0;0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`0;0?ooo`050000003oool0oooo0?ooo`0000003@3oool00`000000oooo0?ooo`0D
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0;
0?ooo`050000003oool0oooo0?ooo`000000303oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?ooo`050000003o
ool0oooo0?ooo`0000002`3oool00`000000oooo0?ooo`050?ooo`050000003oool0oooo0?ooo`00
00005@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0180oooo00D000000?ooo`3o
ool0oooo000000070?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00H0
oooo00@000000?ooo`3oool00000503oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
0180oooo00D000000?ooo`3oool0oooo000000080?ooo`030000003oool0oooo00T0oooo00D00000
0?ooo`3oool0oooo0000000<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool01P3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00D000000?ooo`3oool0oooo
0000000=0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool04`3oool00`000000oooo
0?ooo`0D0?ooo`800000303oool01@000000oooo0?ooo`3oool0000000`0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool02`3oool01@000000oooo0?ooo`3oool0000000/0oooo00<000000?ooo`3oool01@3oool01@00
0000oooo0?ooo`3oool0000001@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0;
0?ooo`00303oool00`000000oooo0?ooo`060?ooo`050000003oool0oooo0?ooo`0000004`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0140oooo00D000000?ooo`3oool0oooo0000
00080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00/0oooo00D000000?ooo`3oool0oooo0000000<0?ooo`800000
5P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01@0oooo0P00000<0?ooo`050000
003oool0oooo0?ooo`0000003@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?ooo`050000003oool0oooo0?oo
o`0000002`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00/0oooo000<0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0A0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool04@3oool01@000000oooo0?ooo`3oool00000
00P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool02`3oool01@000000oooo0?ooo`3oool0000000`0oooo00<00000
0?ooo`3oool05@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01@0oooo0P00000<
0?ooo`050000003oool0oooo0?ooo`0000003@3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?ooo`050000003o
ool0oooo0?ooo`0000002`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00/0oooo
000<0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool04@3oool01@000000oooo0?oo
o`3oool0000000P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00X0oooo00<000000?ooo`3oool03@3oool010000000oooo0?ooo`00000<0?ooo`03
0000003oool0oooo01H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0E0?ooo`80
0000303oool01@000000oooo0?ooo`3oool0000000d0oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0303oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00P0
oooo00D000000?ooo`3oool0oooo0000000>0?ooo`030000003oool0oooo00/0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00d0oooo00@000000?oo
o`3oool00000303oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0140oooo00<00000
0?ooo`3oool05@3oool2000000`0oooo00D000000?ooo`3oool0oooo0000000=0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`080?ooo`050000003oool0oooo0?ooo`0000003`3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`0=0?ooo`040000003oool0oooo000000`0oooo00<000000?ooo`3oool05P3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool02@3oool01@000000oooo
0?ooo`3oool0000000l0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0100oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0;0?ooo`00
303oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00P0oooo00D000000?ooo`3oool0
oooo0000000?0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo00d0oooo00@000000?ooo`3oool00000303oool00`00
0000oooo0?ooo`0F0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool05P3oool00`00
0000oooo0?ooo`090?ooo`050000003oool0oooo0?ooo`0000003`3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`06
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool0203oool01@000000oooo0?ooo`3oool0000000l0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool03P3oool0
10000000oooo0?ooo`00000<0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool03`3o
ool2000001P0oooo00<000000?ooo`3oool02@3oool01@000000oooo0?ooo`3oool0000000l0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool03P3oool01@000000oooo0?ooo`3oool0000000X0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`070?ooo`030000
003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo00P0oooo00D000000?ooo`3oool0oooo0000000?0?ooo`030000003oool0oooo00/0
oooo00D000000?ooo`3oool0oooo0000000=0?ooo`030000003oool0oooo00h0oooo00@000000?oo
o`3oool00000303oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`090?ooo`050000003oool0oooo0?ooo`000000403o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`0>0?ooo`050000003oool0oooo0?ooo`0000002P3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool03P3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool0203oool01@000000oooo0?ooo`3oool000000100oooo00<000000?ooo`3oool0
2`3oool010000000oooo0?ooo`00000=0?ooo`030000003oool0oooo00h0oooo00D000000?ooo`3o
ool0oooo0000000;0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool03`3oool00`00
0000oooo0?ooo`0G0?ooo`030000003oool0oooo00T0oooo00D000000?ooo`3oool0oooo0000000@
0?ooo`030000003oool0oooo00/0oooo00D000000?ooo`3oool0oooo0000000<0?ooo`030000003o
ool0oooo00h0oooo00D000000?ooo`3oool0oooo0000000:0?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`080?ooo`040000003oool0oooo00000140oooo00<000000?ooo`3oool02`3oool010000000oooo
0?ooo`00000=0?ooo`030000003oool0oooo00h0oooo00D000000?ooo`3oool0oooo0000000;0?oo
o`030000003oool0oooo01L0oooo0P00000?0?ooo`8000006@3oool00`000000oooo0?ooo`090?oo
o`050000003oool0oooo0?ooo`0000004@3oool00`000000oooo0?ooo`0:0?ooo`050000003oool0
oooo0?ooo`000000303oool00`000000oooo0?ooo`0>0?ooo`050000003oool0oooo0?ooo`000000
2P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0<0?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool0203oool010000000oooo0?ooo`00000A0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`0000003@3oool00`000000oooo0?ooo`0@0?ooo`040000
003oool0oooo000000/0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0=0?ooo`80
00006@3oool00`000000oooo0?ooo`090?ooo`050000003oool0oooo0?ooo`0000004@3oool00`00
0000oooo0?ooo`0:0?ooo`040000003oool0oooo000000`0oooo00<000000?ooo`3oool03`3oool0
1@000000oooo0?ooo`3oool0000000X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`0;0?ooo`00303oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00P0oooo00@00000
0?ooo`3oool000004@3oool00`000000oooo0?ooo`0<0?ooo`8000003@3oool00`000000oooo0?oo
o`0@0?ooo`040000003oool0oooo000000/0oooo00<000000?ooo`3oool0603oool00`000000oooo
0?ooo`0=0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool02@3oool01@000000oooo
0?ooo`3oool000000140oooo00<000000?ooo`3oool02P3oool010000000oooo0?ooo`00000<0?oo
o`030000003oool0oooo00l0oooo00D000000?ooo`3oool0oooo0000000:0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`060?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`0;0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`090?ooo`040000003oool0oooo00000180oooo00<000000?ooo`3oool02`3oool2
000000d0oooo00<000000?ooo`3oool0403oool010000000oooo0?ooo`00000;0?ooo`030000003o
ool0oooo01P0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0H0?ooo`030000003o
ool0oooo00T0oooo00@000000?ooo`3oool000004P3oool00`000000oooo0?ooo`0;0?ooo`800000
3@3oool00`000000oooo0?ooo`0?0?ooo`050000003oool0oooo0?ooo`0000002`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool01P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool02@3oool010000000oooo0?ooo`00000B0?ooo`030000003oool0oooo00/0
oooo0P00000=0?ooo`030000003oool0oooo0100oooo00@000000?ooo`3oool000002P3oool20000
01/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
00T0oooo00@000000?ooo`3oool000004`3oool00`000000oooo0?ooo`0:0?ooo`8000003@3oool0
0`000000oooo0?ooo`0?0?ooo`050000003oool0oooo0?ooo`0000002`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool01`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool02@3oool010000000oooo0?ooo`00000B0?ooo`030000003oool0oooo00/0oooo0P00
000<0?ooo`030000003oool0oooo0140oooo00@000000?ooo`3oool000002P3oool2000001/0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00T0oooo
00@000000?ooo`3oool000004`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool04@3oool010000000oooo0?ooo`00000;0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool02`3oool000d0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`070?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`090?ooo`030000003oool0000001@0oooo00<000000?ooo`3oool02P3oool2000000`0
oooo00<000000?ooo`3oool04@3oool010000000oooo0?ooo`00000:0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool02P3oool2000001`0oooo00<000000?ooo`3oool0203oool01000
0000oooo0?ooo`00000C0?ooo`030000003oool0oooo00X0oooo0P00000<0?ooo`030000003oool0
oooo0140oooo00@000000?ooo`3oool000002`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0<0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00/0oooo000=0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0000000D0?ooo`030000003oool0oooo00X0oooo0P00000<0?ooo`030000003oool0
oooo0140oooo00@000000?ooo`3oool000002P3oool00`000000oooo0?ooo`0K0?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`080?ooo`040000003o
ool0oooo000001<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0000000<0?ooo`030000
003oool0oooo0140oooo00@000000?ooo`3oool000002`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool01@000000oooo0?ooo`3oool00000
00h0oooo000=0?ooo`050000003oool0oooo0?ooo`0000002`3oool00`000000oooo0?ooo`080?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?oo
o`050000003oool0oooo0?ooo`0000002`3oool00`000000oooo0000000D0?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`000000303oool00`000000oooo0?ooo`0B0?ooo`030000003oool0
000000X0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo01/0oooo00<000000?ooo`3oool0203oool010000000oooo0?ooo`00000D0?ooo`030000003o
ool0oooo00P0oooo00@000000?ooo`3oool000002P3oool00`000000oooo0?ooo`0B0?ooo`040000
003oool0oooo000000/0oooo00D000000?ooo`3oool0oooo0000000;0?ooo`030000003oool0oooo
00P0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo
00P0oooo00D000000?ooo`3oool0oooo0000000>0?ooo`003@3oool01@000000oooo0?ooo`3oool0
000000/0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool02@3oool01@000000oooo0?ooo`3oool0000000/0oooo00<0
00000?ooo`0000005@3oool00`000000oooo0?ooo`080?ooo`040000003oool0oooo000000X0oooo
00<000000?ooo`3oool04`3oool00`000000oooo0000000:0?ooo`030000003oool0oooo01/0oooo
0P0000090?ooo`8000007@3oool00`000000oooo0?ooo`080?ooo`040000003oool0oooo000001@0
oooo00<000000?ooo`3oool0203oool010000000oooo0?ooo`00000:0?ooo`030000003oool0oooo
0180oooo00@000000?ooo`3oool000002`3oool01@000000oooo0?ooo`3oool0000000/0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool02@3oool01@000000oooo0?ooo`3oool0000000h0oooo000=0?ooo`050000003o
ool0oooo0?ooo`0000002`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0:0?ooo`050000003oool0oooo0?ooo`00
00002`3oool00`000000oooo0000000E0?ooo`030000003oool0oooo00L0oooo00D000000?ooo`3o
ool0oooo0000000:0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`0000002P3oool00`00
0000oooo0?ooo`0L0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0703oool00`00
0000oooo0?ooo`080?ooo`030000003oool0000001D0oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool04`3oool00`00
0000oooo0000000;0?ooo`050000003oool0oooo0?ooo`0000002`3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`09
0?ooo`050000003oool0oooo0?ooo`0000003P3oool000d0oooo00@000000?ooo`3oool00000303o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool01P3o
ool00`000000oooo0?ooo`0:0?ooo`050000003oool0oooo0?ooo`0000002`3oool00`000000oooo
0000000F0?ooo`030000003oool0oooo00H0oooo00D000000?ooo`3oool0oooo000000090?ooo`03
0000003oool0oooo01@0oooo00<000000?ooo`0000002P3oool00`000000oooo0?ooo`0L0?ooo`03
0000003oool0oooo00H0oooo0P00000N0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`00
00005P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0000000/0oooo00D000000?ooo`3o
ool0oooo0000000<0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0403oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo00X0oooo00D000000?ooo`3oool0oooo0000000>
0?ooo`003@3oool010000000oooo0?ooo`00000=0?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool0403oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00`0oooo00@00000
0?ooo`3oool000002`3oool00`000000oooo0000000G0?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01D0oooo0P00
000;0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`0M0?ooo`030000003oool0oooo00P0oooo0`00000F0?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo01@0oooo00<0
00000?ooo`0000002`3oool01@000000oooo0?ooo`3oool0000000`0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
2P3oool01@000000oooo0?ooo`3oool0000000h0oooo000=0?ooo`040000003oool0oooo000000d0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool0303oool010000000oooo0?ooo`00000;0?ooo`030000003oool00000
01L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool05P3oool2000000/0oooo00<000000?ooo`3oool07@3oool20000
00D0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`090?ooo`8000005`3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0000000/0oooo00@000000?ooo`3oool000003P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`0<0?ooo`040000003oool0oooo000000h0oooo000=0?ooo`040000003oool0
oooo000000h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0C0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool03@3oool00`000000oooo0000000<0?ooo`030000003o
ool0000001P0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`050?ooo`800000103o
ool00`000000oooo0?ooo`0F0?ooo`8000002`3oool00`000000oooo0?ooo`0N0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`090?ooo`800000603o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool00P3o
ool2000001L0oooo0P00000=0?ooo`030000003oool0000000l0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0303o
ool010000000oooo0?ooo`00000>0?ooo`003P3oool200000100oooo0P0000000`3oool000000000
000G0?ooo`030000003oool0oooo008000003`3oool3000000d0oooo0P00000H0?ooo`8000000P3o
ool2000000P0oooo0P0000020?ooo`800000603oool2000000/0oooo00<000000?ooo`3oool07`3o
ool2000000030?ooo`00000000000240oooo00<000000?ooo`3oool02@3oool2000001P0oooo0P00
00020?ooo`8000002@3oool00`000000oooo0?ooo`02000001T0oooo00<000000?ooo`3oool02`3o
ool3000000l0oooo0P0000020?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool00P00
000@0?ooo`8000003`3oool000h0oooo0P00000A0?ooo`<000006@3oool300000140oooo0P00000=
0?ooo`030000003oool0oooo01P0oooo0`00000;0?ooo`@000006P3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo0200oooo0`00000R0?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`3oool0603oool4000000X0oooo1000000J0?ooo`030000003oool0oooo00/0oooo0P00000A
0?ooo`<000006@3oool300000140oooo0P00000?0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00
o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00
o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00403oool2000001`0oooo1@00000K0?oo
o`<000003`3oool2000002T0oooo1000000=0?ooo`<00000503oool00`000000oooo0?ooo`0b0?oo
o`<00000703oool2000002T0oooo1000000=0?ooo`<00000603oool3000001/0oooo1000000M0?oo
o`8000004@3oool000l0oooo0P0000000`3oool000000?ooo`0I0?ooo`8000000`3oool2000001X0
oooo00@000000?ooo`00000000003P3oool2000002P0oooo0P0000030?ooo`8000002P3oool20000
00030?ooo`000000000001<0oooo00<000000?ooo`3oool0<@3oool010000000oooo0?ooo`3oool2
000001X0oooo0P00000X0?ooo`8000000`3oool2000000X0oooo0P0000000`3oool000000000000F
0?ooo`80000000<0oooo0000003oool06@3oool2000000<0oooo0P00000J0?ooo`040000003oool0
000000000100oooo000?0?ooo`050000003oool0oooo0?ooo`000000603oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo01H0oooo00D000000?ooo`3oool0oooo0000000>0?ooo`030000
003oool0000002H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0`0?ooo`030000
003oool0oooo00<0oooo0P00000I0?ooo`030000003oool0000002H0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0080oooo0P00000E0?ooo`050000
003oool0oooo0?ooo`000000603oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01H0
oooo00D000000?ooo`3oool0oooo0000000@0?ooo`003`3oool01@000000oooo0?ooo`3oool00000
01L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0303oool00`000000oooo0000000U0?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
01L0oooo00<000000?ooo`0000009@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
01D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0=0?ooo`003P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`0D0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02P3oool010000000
oooo0?ooo`00000U0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0;P3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo01H0oooo00<000000?ooo`000000903oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0=0?ooo`003P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0180
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool02P3oool01@000000oooo0?ooo`3oool0000002<0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0]0?ooo`8000002@3oool00`000000oooo0?ooo`0F0?oo
o`040000003oool0oooo000002<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0180oooo0P00000;0?ooo`030000003oool0oooo01<0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`003@3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool02P3oool01@000000oooo0?ooo`3oool0
00000280oooo00<000000?ooo`3oool02P3oool01@000000oooo0?ooo`3oool0000000T0oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo00P0oooo0P00
000F0?ooo`050000003oool0oooo0?ooo`0000008P3oool00`000000oooo0?ooo`0:0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0A0?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo0140oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0=0?ooo`003@3o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool02P3o
ool01@000000oooo0?ooo`3oool000000240oooo00<000000?ooo`3oool0303oool010000000oooo
0?ooo`00000:0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0:`3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo01@0oooo00D000000?ooo`3oool0oooo0000000R0?oo
o`030000003oool0oooo00/0oooo00@000000?ooo`3oool000002@3oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0=
0?ooo`003@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00l0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool02P3oool01@000000oooo0?ooo`3oool000000200oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0000000;0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0:P3oool2
000000`0oooo00<000000?ooo`3oool0503oool01@000000oooo0?ooo`3oool000000240oooo00<0
00000?ooo`3oool0303oool00`000000oooo0000000;0?ooo`030000003oool0oooo0100oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00`0oooo000=
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0>
0?ooo`800000303oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo02X0oooo00<00000
0?ooo`3oool02`3oool2000001D0oooo00D000000?ooo`3oool0oooo0000000P0?ooo`030000003o
ool0oooo00h0oooo0P00000;0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0Y0?ooo`8000003P3oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool07@3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0303oool0
00d0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
02T0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0?0?ooo`800000303oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0303oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`0<0?ooo`003@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00`0
oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01`0
oooo00<000000?ooo`3oool03`3oool2000000`0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0Y0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool04P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool03`3oool2000000`0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00`0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`0000003@3oool00`000000oooo
0?ooo`0=0?ooo`030000003oool0oooo02P0oooo0P00000@0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`000000303oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0<0?ooo`00303oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool03`3oool00`000000oooo0000
000=0?ooo`030000003oool0oooo00d0oooo0P00000Y0?ooo`030000003oool0oooo00l0oooo0P00
000C0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0?0?ooo`040000003oool0oooo000000`0oooo00<000000?ooo`3oool03@3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool04P3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0303oool000`0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00l0oooo00@0
00000?ooo`3oool00000303oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo02L0oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0?0?ooo`040000003oool0oooo000000`0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo00l0oooo00D000000?ooo`3oool0oooo0000000<0?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool09P3oool200000180oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3o
ool03`3oool01@000000oooo0?ooo`3oool0000000`0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02`3oool000`0
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo01@0
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00l0
oooo00D000000?ooo`3oool0oooo0000000<0?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool09P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00l0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo01D0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00/0oooo000<0?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0E0?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo02H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0?0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0F0?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?ooo`00
303oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
5P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo0180oooo0P00000C0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0@0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo00X0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0G0?ooo`030000
003oool0oooo00L0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?ooo`00303o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool05P3o
ool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool03@3o
ool2000002H0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0A0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?ooo`00303oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool03@3oool2
000002D0oooo0P00000E0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo
00H0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0=0?ooo`8000009@3oool00`00
0000oooo0?ooo`0D0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`090?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0280
oooo0P00000F0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`0E0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool01P3oool00`000000
oooo0?ooo`080?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`050?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0280oooo
00<000000?ooo`3oool05@3oool200000180oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`0E0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool0503oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0280oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool0503oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0Q
0?ooo`800000603oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool0503oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0<0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo00P0oooo00<000000?ooo`3oool03P3oool200000280oooo00<000000?oo
o`3oool05`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00P0oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0303oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0=0?ooo`05
0000003oool0oooo0?ooo`0000007P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
0180oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool03P3oool200000280oooo00<000000?ooo`3oool05`3oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool04P3oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00d0oooo00D000000?ooo`3oool0oooo0000000O
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;
0?ooo`00303oool00`000000oooo0?ooo`0=0?ooo`050000003oool0oooo0?ooo`0000007P3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool03P3oool2
00000240oooo0P00000J0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool03@3oool0
1@000000oooo0?ooo`3oool0000001l0oooo00D000000?ooo`3oool0oooo0000000?0?ooo`030000
003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00d0oooo00D000000?ooo`3oool0oooo
0000000N0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03@3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0403oool00`000000oooo
0?ooo`0@0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`0>0?ooo`8000008@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool03@3oool01@000000oooo0?ooo`3oool0000001l0oooo00D000000?ooo`3o
ool0oooo0000000?0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`000000803oool01@000000oooo0?ooo`3oool0000000l0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0N0?ooo`8000006`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo00d0oooo00D000000?ooo`3oool0oooo0000000P0?ooo`040000003oool0
oooo000000l0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0000000P0?ooo`040000003oool0oooo00000100oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool000000240oooo00@000000?ooo`3oool00000
3`3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool000000200oooo00@000000?ooo`3oool00000403oool00`000000oooo0?ooo`090?ooo`030000
003oool0oooo00H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0@0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0>0?ooo`800000
7`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0000000Q0?ooo`030000003oool000000100oooo00<000000?ooo`3oool0
2`3oool000`0oooo00<000000?ooo`3oool03P3oool200000280oooo00<000000?ooo`000000403o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`0>0?ooo`8000007P3oool2000001d0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool000000280oooo0P00000@0?ooo`030000003o
ool0oooo00/0oooo000<0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0803oool0
0`000000oooo0000000@0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00h0oooo0P00000N0?ooo`030000003oool0
oooo01`0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`060?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0>0?ooo`8000008`3oool2
00000100oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool03`3oool00`00
0000oooo0?ooo`0P0?ooo`8000004@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0>0?ooo`8000007P3oool00`00
0000oooo0?ooo`0L0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool03`3oool00`00
0000oooo0?ooo`0Q0?ooo`800000403oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000
oooo0?ooo`0>0?ooo`8000008`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00h0
oooo0P00000M0?ooo`8000007P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool08@3oool200000100oooo00<000000?ooo`3oool02`3oool000`0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0000000R0?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01d0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0>0?ooo`8000008`3oool200000100oooo
00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool03P3oool00`000000oooo0000
000Q0?ooo`8000004@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01/0oooo00<0
00000?ooo`3oool07@3oool200000100oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool03P3oool00`000000oooo0000
000R0?ooo`800000403oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0=
0?ooo`040000003oool0oooo00000240oooo00<000000?ooo`000000403oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`0>0?ooo`8000006`3oool200000200oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool000000240oooo00<000000?ooo`000000403oool00`000000oooo0?oo
o`0;0?ooo`00303oool00`000000oooo0?ooo`0=0?ooo`040000003oool0oooo00000240oooo00<0
00000?ooo`000000403oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0100oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`0>0?ooo`8000006`3oool00`000000oooo0?oo
o`0O0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool03P3oool00`000000oooo0000
000Q0?ooo`040000003oool0oooo000000l0oooo00<000000?ooo`3oool02`3oool000`0oooo00<0
00000?ooo`3oool03@3oool01@000000oooo0?ooo`3oool0000001l0oooo00@000000?ooo`3oool0
0000403oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`0>0?ooo`8000006`3oool00`000000oooo0?ooo`0O0?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0:0?ooo`03
0000003oool0oooo0100oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool03@3oool01@000000oooo0?ooo`3oool00000
0200oooo00@000000?ooo`3oool000003`3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`00
0000oooo0?ooo`0=0?ooo`050000003oool0oooo0?ooo`0000007`3oool010000000oooo0?ooo`00
000@0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0603oool200000240oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00X0oooo00<0
00000?ooo`3oool0403oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`0=0?ooo`050000003oool0oooo0?ooo`000000
7`3oool01@000000oooo0?ooo`3oool0000000l0oooo00<000000?ooo`3oool02`3oool000`0oooo
00<000000?ooo`3oool0303oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01d0oooo
00D000000?ooo`3oool0oooo0000000?0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool04P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool04P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool03@3oool01@000000oooo0?ooo`3oool0000001l0oooo00D000000?ooo`3oool0oooo0000000?
0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0L0?ooo`050000003oool0oooo0?ooo`0000003`3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0803oool2
00000100oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo0100oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo01`0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo00/0oooo000<0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0K0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo01L0oooo0P00000R0?ooo`800000403oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0403oool00`00
0000oooo0?ooo`0B0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool02@3oool00`00
0000oooo0?ooo`0<0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0703oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool000`0
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01/0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0Q0?ooo`800000403oool00`000000
oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0D0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool02@3oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool06`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool000`0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01/0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool0503oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0303oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo
000<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`0I0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo01D0oooo0P00000T0?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool05@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01X0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo000<
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0I
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;
0?ooo`00303oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00/0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool0403oool2000001H0oooo00<000000?ooo`3oool08`3oool2000000l0oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool05@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo000<0?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0G0?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0?0?ooo`03
0000003oool0oooo01L0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0A0?ooo`03
0000003oool0oooo01<0oooo0P00000U0?ooo`8000003`3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0G0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
603oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
4`3oool00`000000oooo0?ooo`0T0?ooo`8000003`3oool00`000000oooo0?ooo`090?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0G0?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0:0?ooo`030000
003oool0oooo00L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`070?ooo`030000
003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0603o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0140oooo0P00000D0?ooo`030000003o
ool0oooo02D0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool02@3oool0
1@000000oooo0?ooo`3oool000000180oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo0140oooo0P00000C0?ooo`8000009`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`0E0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0;0?ooo`00303oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`3oool0503oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`090?ooo`050000003oool0oooo0?ooo`0000004@3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool04P3o
ool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`090?ooo`050000003oool0oooo0?ooo`0000004P3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool05@3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0303oool000d0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01@0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool02@3oool01@000000oooo0?ooo`3oool000000140oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0
4@3oool00`000000oooo0?ooo`0V0?ooo`8000003`3oool00`000000oooo0?ooo`090?ooo`050000
003oool0oooo0?ooo`0000004P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0000000A0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool09`3oool2000000l0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0000000C0?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0<0?ooo`003@3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`090?ooo`030000003oool000000140oooo00<000000?ooo`3oool0703oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`0W0?ooo`8000003`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool000000180oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00`0oooo000=
0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00X0oooo0P00000A0?ooo`030000003oool0oooo01`0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool0:@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`0000004P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool04@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0<0?oo
o`030000003oool0oooo0100oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo01d0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0B0?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0L0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0<0?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`0B0?ooo`8000003P3oool2000002/0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool0703oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000
003oool0oooo00H0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo00l0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo00X0oooo0P00000?0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0B0?ooo`8000003@3oool2000002`0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`090?ooo`8000004@3oool00`000000oooo0?ooo`0M0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`060?ooo`030000
003oool0oooo00h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo00H0oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool01P3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0000000?0?ooo`030000003oool0oooo01l0oooo00D000000?ooo`3oool0oooo
0000000E0?ooo`030000003oool0oooo00/0oooo0P00000/0?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0000000@0?ooo`030000003oool0oooo01d0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0=0?ooo`003@3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`080?ooo`040000003oool0oooo000000d0oooo00<000000?ooo`3oool0803oool01@000000oooo
0?ooo`3oool0000001D0oooo0P00000;0?ooo`800000;@3oool00`000000oooo0?ooo`0?0?ooo`03
0000003oool0oooo00L0oooo00@000000?ooo`3oool000003`3oool00`000000oooo0?ooo`0O0?oo
o`050000003oool0oooo0?ooo`000000303oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo0100oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool03@3oool000d0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool01`3oool01@00
0000oooo0?ooo`3oool0000000d0oooo00<000000?ooo`3oool0803oool01@000000oooo0?ooo`3o
ool0000001H0oooo00<000000?ooo`3oool02@3oool2000002d0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`070?ooo`050000003oool0oooo0?ooo`0000003P3oool00`000000oooo
0?ooo`0O0?ooo`050000003oool0oooo0?ooo`000000303oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0140oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0A0?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool03@3oool000h0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
4P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0
1`3oool01@000000oooo0?ooo`3oool0000000`0oooo00<000000?ooo`3oool08@3oool010000000
oooo0?ooo`00000G0?ooo`8000002@3oool2000002h0oooo0P00000A0?ooo`030000003oool0oooo
00H0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
0200oooo00D000000?ooo`3oool0oooo0000000<0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool04@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`0=0?ooo`003P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0B0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0R0?oo
o`030000003oool0000001P0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0^0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00T0oooo00<000000?ooo`3oool08@3oool01@000000oooo0?ooo`3oool0
000000`0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00d0oooo000>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0B0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo02<0oooo00<000000?ooo`000000603oool20000
00L0oooo0P00000`0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool08P3oool00`00
0000oooo0000000=0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0=0?ooo`003`3oool01@000000oooo0?ooo`3oool0000001L0oooo0P0000070?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0C0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo02@0oooo00<000000?ooo`0000006@3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo02l0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0S0?oo
o`030000003oool0000000h0oooo00D000000?ooo`3oool0oooo0000000H0?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00d0oooo000?0?ooo`050000003oool0oooo0?ooo`0000006@3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo01H0oooo00D000000?ooo`3oool0oooo0000000E0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo02@0oooo00<000000?ooo`0000006P3oool00`000000oooo0?ooo`020?ooo`800000<P3o
ool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`030?ooo`8000009P3oool3000000h0oooo00@000000?ooo`3oool00000
6P3oool00`000000oooo0?ooo`030?ooo`800000603oool200000080oooo00<000000?ooo`3oool0
3P3oool000l0oooo1000000J0?ooo`L000006P3oool3000001L0oooo1000000<0?ooo`H00000:03o
ool2000001/0oooo1@00000c0?ooo`030000003oool0oooo0140oooo1@00000;0?ooo`D00000:@3o
ool2000000l0oooo0`00000J0?ooo`H000006`3oool400000100oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo000D0?oo
o`H00000;@3oool6000002H0oooo1`00000D0?ooo`L00000?P3oool9000003h0oooo1`00000D0?oo
o`L000009P3oool6000002d0oooo1P00000E0?ooo`004`3oool00`000000oooo0?ooo`030?ooo`80
0000:`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0240oooo0P0000060?ooo`80
00004P3oool00`000000oooo0?ooo`050?ooo`800000>P3oool3000000L0oooo0`00000j0?ooo`80
00001`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00D0oooo0P00000S0?ooo`80
0000103oool2000002/0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0B0?ooo`00
4P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo00P0oooo0P00000>0?ooo`800000
2P3oool00`000000oooo0?ooo`0f0?ooo`8000002`3oool2000003P0oooo00<000000?ooo`3oool0
203oool2000000h0oooo0P00000:0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`0V0?ooo`800000203oool00`000000oooo0?ooo`0A0?ooo`004@3o
ool00`000000oooo0?ooo`080?ooo`030000003oool0oooo02@0oooo0P0000090?ooo`030000003o
ool0oooo01h0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool0=03oool2000000h0oooo00<000000?ooo`3oool0=03o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00T0oooo0P00000=0?ooo`030000003o
ool0oooo01h0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0T0?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool04@3oool00140oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`0R0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool07@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0b0?ooo`800000403oool00`000000oooo0?ooo`0b0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo01`0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0
oooo00X0oooo00<000000?ooo`3oool0403oool00100oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`0P0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool06`3oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool03`3oool00`00
0000oooo0?ooo`0`0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0<03oool00`00
0000oooo0?ooo`0?0?ooo`8000001`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool03`3oool00100oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0O0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool06`3oool00`000000
oooo0?ooo`0@0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool04@3oool00`000000
oooo0?ooo`0^0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0;P3oool00`000000
oooo0?ooo`0B0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`0J0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool07`3oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo00l0oooo000@0?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool0;@3oool2000001D0oooo0P00000_0?ooo`030000003oool0oooo01<0oooo
00D000000?ooo`3oool0oooo0000000D0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3o
ool03`3oool000l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0L0?ooo`030000
003oool0oooo00h0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0C0?ooo`050000
003oool0oooo0?ooo`0000005P3oool00`000000oooo0?ooo`0[0?ooo`8000005`3oool2000002d0
oooo00<000000?ooo`3oool0503oool01@000000oooo0?ooo`3oool0000001D0oooo00<000000?oo
o`3oool05`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01`0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0>0?ooo`003`3oool00`000000oooo0?ooo`0?0?ooo`03
0000003oool0oooo01/0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0H0?ooo`03
0000003oool0oooo01@0oooo00@000000?ooo`3oool000005`3oool00`000000oooo0?ooo`0Z0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0000001H0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo01/0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0>0?oo
o`003`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3o
ool0403oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01H0oooo0P00000H0?ooo`03
0000003oool0oooo02T0oooo0P00000I0?ooo`800000:`3oool00`000000oooo0?ooo`0G0?ooo`03
0000003oool0oooo01H0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0=0?ooo`00
3P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
4@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01H0oooo0P00000I0?ooo`030000
003oool0oooo02L0oooo0P00000K0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0
5P3oool00`000000oooo0000000G0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0
4@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0
3@3oool000h0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo0140oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0G0?ooo`8000006@3o
ool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool09`3o
ool00`000000oooo0?ooo`0G0?ooo`030000003oool0000001P0oooo00<000000?ooo`3oool0503o
ool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool04P3o
ool00`000000oooo0?ooo`0=0?ooo`003P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0
oooo01D0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0
oooo01H0oooo00@000000?ooo`3oool000006@3oool00`000000oooo0?ooo`0U0?ooo`8000007@3o
ool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo01H0oooo00D000000?ooo`3oool0oooo
0000000G0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool04`3oool00`000000oooo
0?ooo`0E0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool03@3oool000h0oooo00<0
00000?ooo`3oool04`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo01@0oooo00<0
00000?ooo`3oool04`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo01d0oooo00<0
00000?ooo`3oool09@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool05@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00d0oooo000=
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0D
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0903oool2000001l0oooo00<00000
0?ooo`3oool09@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool05@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01@0oooo00<00000
0?ooo`3oool04`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00d0oooo000=0?oo
o`030000003oool0oooo01H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo01H0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0O0?oo
o`030000003oool0oooo02@0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo01H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0F0?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0=0?oo
o`003@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3o
ool05P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0280oooo0P00000Q0?ooo`03
0000003oool0oooo02<0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo01D0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0F0?ooo`03
0000003oool0oooo0140oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0<0?ooo`00
3@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
5`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0
8@3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
5`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
303oool000d0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo01P0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0E0?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0Q0?ooo`030000003o
ool0oooo0240oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0E0?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo00`0oooo000=0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0803oool0
0`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0603oool0
0`000000oooo0?ooo`0<0?ooo`003@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
01@0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo
0200oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo
01@0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool06P3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0D0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0503oool00`000000
oooo0?ooo`0N0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool07P3oool00`000000
oooo0?ooo`0E0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool05@3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`0J0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01@0oooo
00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo01h0oooo
00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01@0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?oo
o`0K0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`0C0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?oo
o`0N0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0D0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool02`3oool000`0oooo00<00000
0?ooo`3oool06`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo01`0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool04`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo02L0oooo00<00000
0?ooo`3oool07@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool0503oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01/0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00/0oooo000<0?oo
o`030000003oool0oooo01`0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0L0?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0@0?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0X0?oo
o`030000003oool0oooo01`0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo01@0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0L0?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0;0?oo
o`00303oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool07@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3o
ool0403oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3o
ool0:@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3o
ool0403oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3o
ool0703oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3o
ool02`3oool000`0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo01d0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0B0?ooo`030000
003oool0oooo0180oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0K0?ooo`030000
003oool0oooo02T0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0B0?ooo`030000
003oool0oooo0140oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0=0?ooo`030000
003oool0oooo01d0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0M0?ooo`030000
003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool04`3o
ool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool04P3o
ool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool07@3o
ool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0N0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo0180oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0A0?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo0180oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo01h0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool07P3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0303oool00`00
0000oooo0?ooo`0B0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`0I0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool06@3oool00`00
0000oooo0?ooo`0B0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool04P3oool00`00
0000oooo0?ooo`0<0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0N0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo01l0
oooo00D000000?ooo`3oool0oooo0000000Q0?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0180oooo00<000000?oo
o`3oool0303oool00`000000oooo0?ooo`0O0?ooo`050000003oool0oooo0?ooo`0000008@3oool0
0`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0O0?ooo`040000003oool0oooo
00000280oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
oooo02h0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0A0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo01l0oooo00D000000?ooo`3oool0oooo0000000Q0?ooo`030000003oool0oooo00/0oooo000<
0?ooo`030000003oool0oooo0200oooo00<000000?ooo`0000008P3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0B
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0803oool00`000000oooo0000000R
0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo0200oooo0P00000S0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo0100oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0`0?oo
o`030000003oool0oooo01H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0P0?oo
o`030000003oool000000280oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3o
ool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool05@3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3o
ool0403oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3o
ool02`3oool000`0oooo00<000000?ooo`3oool0803oool2000002<0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0
5@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0200oooo0P00000S0?ooo`030000
003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo0200oooo00<000000?ooo`0000008P3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool06`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0<P3o
ool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool06`3o
ool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0803o
ool00`000000oooo0000000R0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0
oooo01l0oooo00@000000?ooo`3oool000008P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo0140oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo01@0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0D0?ooo`030000003o
ool0oooo0100oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool07`3oool010000000oooo0?ooo`00000R0?ooo`030000
003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo01l0oooo00D000000?ooo`3oool0oooo
0000000Q0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0403oool00`000000oooo
0?ooo`0M0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0503oool00`000000oooo
0?ooo`0c0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0403oool00`000000oooo
0?ooo`0L0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`0O0?ooo`050000003oool0oooo0?ooo`0000008@3oool00`000000oooo0?ooo`0;0?ooo`00
303oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
7`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0
7@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0180oooo0P00000e0?ooo`030000
003oool0oooo01<0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0M0?ooo`030000
003oool0oooo0100oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0N0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0;0?ooo`00303o
ool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool07P3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool07`3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0=03o
ool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool07P3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool07P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool02`3o
ool000`0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo01h0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0
oooo01l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0
oooo03D0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo01l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo01d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0N0?ooo`030000003oool0
oooo00/0oooo000<0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`0M0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool03`3oool00`00
0000oooo0?ooo`0Q0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool04@3oool20000
03L0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
0200oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo
01d0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo
00/0oooo000<0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool01P3oool00`000000
oooo0?ooo`0M0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0Q0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`0f0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0Q0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool02@3oool00`000000
oooo0?ooo`0L0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool07@3oool00`000000
oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool0703oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool0=P3oool2000001<0oooo00<000000?ooo`3oool03@3oool00`000000oooo
0?ooo`0R0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`0L0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0703oool00`000000oooo
0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0703oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool08`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0100oooo00<0
00000?ooo`3oool0=`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool08`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool06`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo01`0oooo00<0
00000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0K
0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool02P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0h0?ooo`8000004@3oool00`000000oooo0?ooo`0<0?oo
o`030000003oool0oooo02D0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`090?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0K0?oo
o`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool0303oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3o
ool03`3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool0303oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0:0?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo00h0oooo0P00000k0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`0<0?ooo`00303oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool0>P3oool200000100oooo00<000000?ooo`3oool02`3o
ool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0603o
ool00`000000oooo0?ooo`0<0?ooo`003@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00X0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo00X0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo01P0oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool05`3oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`0=0?ooo`800000?@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
01L0oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool05P3oool00`000000
oooo0?ooo`0A0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool02@3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo03`0oooo0P00000?0?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo01H0
oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool05P3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`0=0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool03@3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`0E0?ooo`030000003oool0oooo00d0oooo000>0?ooo`030000003oool0oooo01@0oooo00<0
00000?ooo`3oool04`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00X0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool0303oool2000003l0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?oo
o`0D0?ooo`030000003oool0oooo00d0oooo000>0?ooo`030000003oool0oooo01<0oooo00<00000
0?ooo`3oool0503oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`3oool02`3oool2000003l0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0:
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0C
0?ooo`030000003oool0oooo00d0oooo000>0?ooo`030000003oool0oooo01<0oooo00<000000?oo
o`3oool05@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo00/0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool02P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool04P3oool00`000000oooo0?ooo`0=0?ooo`003P3oool00`000000oooo0?ooo`0B0?ooo`03
0000003oool0oooo01L0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo00/0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0G0?ooo`03
0000003oool0oooo0180oooo00<000000?ooo`3oool03@3oool000l0oooo00<000000?ooo`3oool0
4@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`0:0?ooo`800000@@3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo00P0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`080?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo0140oooo00<000000?ooo`3oool03@3oool000l0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool02`3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0400oooo0P00000<0?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo0140oooo00<000000?ooo`3oool03@3oool000l0oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0303oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool02P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool06`3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0oooo000@0?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo
00H0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
01`0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0>0?ooo`00403oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0=03oool00`000000
oooo0?ooo`050?ooo`030000003oool0oooo00X0oooo0P0000130?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01h0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`00403oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0=@3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0@`3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0=@3oool00`000000oooo
0?ooo`040?ooo`8000003P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01l0oooo
00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0?0?ooo`004@3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`0Q0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool03`3oool00140oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool03`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo03L0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo04<0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo03P0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`3oool08`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0100oooo000A0?oo
o`030000003oool0oooo00P0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0h0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`140?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0i0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`080?oo
o`030000003oool0oooo02D0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0A0?oo
o`004P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0>P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool0A@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0>P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo0P0000070?ooo`03
0000003oool0oooo02L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0A0?ooo`00
4`3oool2000000@0oooo0P00000[0?ooo`800000103oool00`000000oooo0?ooo`0A0?ooo`800000
00<0oooo000000000000?P3oool010000000oooo0?ooo`00000<0?ooo`030000003oool0oooo04D0
oooo00<000000?ooo`3oool02P3oool010000000oooo0?ooo`00000n0?ooo`040000003oool0oooo
000001@0oooo00<000000?ooo`3oool00P3oool2000002/0oooo0P0000040?ooo`800000503oool0
01@0oooo1@00000^0?ooo`D000005@3oool300000400oooo0`00000<0?ooo`030000003oool0oooo
04D0oooo00<000000?ooo`3oool02P3oool300000400oooo0`00000E0?ooo`D00000;@3oool60000
01D0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo000/0?ooo`P00000<`3oool6000004H0oooo00<000000?ooo`3oool0F03oool6000006H0
oooo2@00000]0?ooo`009`3oool5000000P0oooo1P00000Z0?ooo`@000001@3oool600000400oooo
0P00001F0?ooo`@000001@3oool5000005`0oooo1@0000090?ooo`D00000:03oool002<0oooo1000
000C0?ooo`<000009P3oool00`000000oooo0?ooo`0=0?ooo`800000@03oool00`000000oooo0?oo
o`1B0?ooo`030000003oool0oooo00`0oooo0P00001H0?ooo`8000004`3oool2000002H0oooo000Q
0?ooo`8000006P3oool2000002<0oooo00<000000?ooo`3oool0403oool2000003h0oooo0P00001B
0?ooo`030000003oool0oooo00l0oooo0`00001C0?ooo`8000005`3oool2000002@0oooo000O0?oo
o`8000007P3oool200000200oooo00<000000?ooo`3oool04`3oool2000003d0oooo0P00001@0?oo
o`030000003oool0oooo01<0oooo0P00001>0?ooo`<000006`3oool300000240oooo000N0?ooo`03
0000003oool0oooo0200oooo0P00000N0?ooo`030000003oool0oooo01D0oooo0P00000l0?ooo`80
0000CP3oool2000001L0oooo0P00001:0?ooo`8000008@3oool2000001l0oooo000L0?ooo`800000
9@3oool2000001/0oooo00<000000?ooo`3oool0603oool2000003/0oooo0P00001=0?ooo`030000
003oool0oooo01P0oooo0P0000160?ooo`8000009@3oool00`000000oooo0?ooo`0L0?ooo`006`3o
ool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0i0?ooo`800000C03oool00`000000oooo0?ooo`0J0?ooo`800000@`3o
ool00`000000oooo0?ooo`0V0?ooo`800000703oool001X0oooo00<000000?ooo`3oool0:@3oool2
000001L0oooo00<000000?ooo`3oool0703oool2000003X0oooo00<000000?ooo`3oool0BP3oool0
0`000000oooo0?ooo`0L0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0:@3oool0
0`000000oooo0?ooo`0I0?ooo`006@3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo
01@0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo
04P0oooo00<000000?ooo`3oool07P3oool2000003h0oooo00<000000?ooo`3oool0:`3oool00`00
0000oooo0?ooo`0H0?ooo`00603oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo01<0
oooo00<000000?ooo`3oool07`3oool2000003T0oooo00<000000?ooo`3oool0A`3oool00`000000
oooo0?ooo`0P0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0;@3oool00`000000
oooo0?ooo`0G0?ooo`005`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01<0oooo
00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo04H0oooo
00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0g0?ooo`800000<@3oool00`000000oooo
0?ooo`0F0?ooo`005@3oool2000003<0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`0R0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?oo
o`0R0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?oo
o`0E0?ooo`00503oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo0140oooo00<00000
0?ooo`3oool08`3oool2000003P0oooo00<000000?ooo`3oool0A03oool2000002@0oooo0P00000d
0?ooo`030000003oool0oooo03@0oooo0P00000E0?ooo`004`3oool00`000000oooo0?ooo`0f0?oo
o`030000003oool0oooo0100oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0f0?oo
o`030000003oool0oooo04@0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`0`0?oo
o`030000003oool0oooo03L0oooo00<000000?ooo`3oool04P3oool001<0oooo00<000000?ooo`3o
ool0=P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3o
ool0=P3oool2000004@0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0_0?ooo`03
0000003oool0oooo03L0oooo00<000000?ooo`3oool04P3oool00180oooo00<000000?ooo`3oool0
>03oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo02H0oooo0P00000h0?ooo`030000
003oool0oooo0440oooo00<000000?ooo`3oool09P3oool2000002h0oooo00<000000?ooo`3oool0
>@3oool00`000000oooo0?ooo`0A0?ooo`004P3oool00`000000oooo0?ooo`0i0?ooo`030000003o
ool0oooo00l0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0f0?ooo`030000003o
ool0oooo0400oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0[0?ooo`030000003o
ool0oooo03T0oooo00<000000?ooo`3oool04@3oool00140oooo00<000000?ooo`3oool0>`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0=P3oool0
0`000000oooo0?ooo`0o0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0:@3oool0
0`000000oooo0?ooo`0k0?ooo`030000003oool0oooo0100oooo000@0?ooo`030000003oool0oooo
03`0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo
03H0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo
02P0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`0?0?ooo`00403oool00`000000
oooo0?ooo`0m0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0:P3oool00`000000
oooo0?ooo`0f0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool0:P3oool00`000000
oooo0?ooo`0V0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool03`3oool000l0oooo
00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo02/0oooo
00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo02/0oooo
00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo00h0oooo
000?0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`0[0?ooo`800000=`3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo02/0oooo00<0
00000?ooo`3oool08`3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo00h0oooo000>
0?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0]
0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`0/
0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`0=
0?ooo`003P3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool0;P3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo03d0oooo00<000000?oo
o`3oool0;03oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0440oooo00<000000?oo
o`3oool03@3oool000h0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo02h0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0l0?ooo`03
0000003oool0oooo02d0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`110?ooo`03
0000003oool0oooo00d0oooo000=0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0
?03oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0
@P3oool00`000000oooo0?ooo`0=0?ooo`003@3oool00`000000oooo0?ooo`130?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0d0?ooo`030000003o
ool0oooo03/0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0L0?ooo`030000003o
ool0oooo04<0oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool0@`3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0=@3oool0
0`000000oooo0?ooo`0k0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool06`3oool0
0`000000oooo0?ooo`130?ooo`030000003oool0oooo00`0oooo000<0?ooo`030000003oool0oooo
04@0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo
03D0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000
oooo0?ooo`140?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0;`3oool00`000000
oooo0?ooo`0e0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0;`3oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool02`3oool000`0oooo
00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo02h0oooo
00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo02l0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo00/0oooo
000<0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`0_0?ooo`030000003oool0oooo03H0oooo0P00000j0?ooo`030000003oool0oooo02l0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo00/0oooo000<
0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0_
0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`0_
0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0;
0?ooo`00303oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo0180oooo00<000000?oo
o`3oool0;`3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo03P0oooo00<000000?oo
o`3oool0;`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo04D0oooo00<000000?oo
o`3oool02`3oool000`0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0B0?ooo`03
0000003oool0oooo02l0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`0g0?ooo`03
0000003oool0oooo0300oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`150?ooo`03
0000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0
4`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool0
=`3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0
A@3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`150?ooo`030000003o
ool0oooo01@0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0h0?ooo`030000003o
ool0oooo03L0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0B0?ooo`030000003o
ool0oooo04D0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool0A@3oool0
0`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0>03oool0
0`000000oooo0?ooo`0g0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool04P3oool0
0`000000oooo0?ooo`150?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo
04D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo
03P0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000
oooo0?ooo`150?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0<03oool00`000000
oooo0?ooo`0h0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0<03oool00`000000
oooo0?ooo`0@0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool02`3oool000`0oooo
00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0340oooo
00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo02l0oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo00/0oooo
000<0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`0`0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?oo
o`0_0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?oo
o`0;0?ooo`00303oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo01T0oooo00<00000
0?ooo`3oool0<03oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo03L0oooo00<00000
0?ooo`3oool0;`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo04<0oooo00<00000
0?ooo`3oool0303oool000`0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`0J0?oo
o`030000003oool0oooo02l0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0f0?oo
o`030000003oool0oooo02l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`130?oo
o`030000003oool0oooo00`0oooo000<0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3o
ool0=P3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3o
ool0@`3oool00`000000oooo0?ooo`0<0?ooo`003@3oool00`000000oooo0?ooo`130?ooo`030000
003oool0oooo01/0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0j0?ooo`030000
003oool0oooo03H0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo04<0oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool0@P3o
ool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0>P3o
ool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`110?ooo`030000003oool0oooo00d0oooo000>0?ooo`030000003oool0
oooo0440oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0
oooo03/0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`0=0?ooo`003P3oool00`00
0000oooo0?ooo`100?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0;P3oool00`00
0000oooo0?ooo`0k0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0;@3oool00`00
0000oooo0?ooo`0=0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool03@3oool000h0
oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo02d0
oooo00<000000?ooo`3oool0?03oool2000003L0oooo00<000000?ooo`3oool0;03oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool03@3oool000h0oooo
00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo02`0oooo
00<000000?ooo`3oool0?@3oool2000003L0oooo00<000000?ooo`3oool0:`3oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool03P3oool000l0oooo00<0
00000?ooo`3oool0?P3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo02`0oooo00<0
00000?ooo`3oool0?@3oool2000003L0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool03`3oool00100oooo00<00000
0?ooo`3oool0?@3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo02/0oooo00<00000
0?ooo`3oool0?P3oool2000003L0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0403oool00100oooo00<000000?oo
o`3oool0?03oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo02X0oooo00<000000?oo
o`3oool0?`3oool2000003L0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo03/0oooo00<000000?ooo`3oool0403oool00140oooo00<000000?ooo`3o
ool0>`3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3o
ool0@03oool2000003L0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo03X0oooo00<000000?ooo`3oool04@3oool00180oooo00<000000?ooo`3oool0
>@3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0
@03oool2000003L0oooo0P00000Y0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
>@3oool00`000000oooo0?ooo`0A0?ooo`004`3oool00`000000oooo0?ooo`0g0?ooo`030000003o
ool0oooo02`0oooo0P00000Y0?ooo`030000003oool0oooo0440oooo0P00000h0?ooo`030000003o
ool0oooo02H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0h0?ooo`030000003o
ool0oooo0180oooo000C0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0;P3oool0
0`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0480oooo0P00000h0?ooo`030000003oool0
oooo02D0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0f0?ooo`030000003oool0
oooo01<0oooo000D0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0<03oool00`00
0000oooo0?ooo`0U0?ooo`030000003oool0oooo04<0oooo0P00000h0?ooo`030000003oool0oooo
02@0oooo0P00000A0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool04`3oool001D0
oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0b0?ooo`8000009@3oool00`000000
oooo0?ooo`140?ooo`800000>03oool2000002D0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`0d0?ooo`030000003oool0oooo01@0oooo000F0?ooo`030000003oool0oooo0340oooo
00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo04@0oooo
0P00000i0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`0b0?ooo`030000003oool0oooo01D0oooo000F0?ooo`030000003oool0oooo0300oooo00<0
00000?ooo`3oool0=`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo04D0oooo0P00
000i0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool04P3oool200000340oooo00<0
00000?ooo`3oool05P3oool001L0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0i
0?ooo`8000008P3oool00`000000oooo0?ooo`160?ooo`800000>@3oool00`000000oooo0?ooo`0P
0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0F
0?ooo`00603oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo03`0oooo0P00000O0?oo
o`030000003oool0oooo04P0oooo0P00000i0?ooo`800000803oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo02`0oooo00<000000?ooo`3oool05`3oool001T0oooo00<000000?ooo`3o
ool0:P3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3o
ool0B@3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3o
ool05`3oool00`000000oooo0?ooo`0Y0?ooo`8000006P3oool001X0oooo0P00000X0?ooo`800000
@`3oool2000001`0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0i0?ooo`800000
703oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo02H0oooo0P00000L0?ooo`00703o
ool00`000000oooo0?ooo`0S0?ooo`800000A`3oool2000001T0oooo00<000000?ooo`3oool0C03o
ool00`000000oooo0?ooo`0j0?ooo`8000006P3oool00`000000oooo0?ooo`0I0?ooo`8000009@3o
ool00`000000oooo0?ooo`0L0?ooo`007@3oool200000280oooo00<000000?ooo`3oool0B@3oool2
000001L0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool0703oool200000240oooo0P00000O0?ooo`007`3oool30000
01d0oooo0P00001>0?ooo`8000005@3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo
03/0oooo0P00000E0?ooo`030000003oool0oooo01l0oooo0P00000N0?ooo`030000003oool0oooo
01l0oooo000R0?ooo`8000006@3oool200000580oooo0`00000A0?ooo`030000003oool0oooo0500
oooo0P00000m0?ooo`<000004@3oool00`000000oooo0?ooo`0R0?ooo`<000006@3oool200000280
oooo000T0?ooo`<000004`3oool3000005L0oooo0`00000<0?ooo`800000E@3oool00`000000oooo
0?ooo`0m0?ooo`<000003@3oool00`000000oooo0?ooo`0V0?ooo`8000004P3oool5000002@0oooo
000W0?ooo`L00000103oool8000005d0oooo1@0000030?ooo`@00000E`3oool00`000000oooo0?oo
o`100?ooo`H000000`3oool4000002/0oooo1`0000040?ooo`L00000:@3oool002h0oooo1@00001Y
0?ooo`<00000G03oool00`000000oooo0?ooo`150?ooo`@00000=@3oool5000002l0oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {342.938, 0}} -> {-0.127076, -0.0995579, \
0.0182859, 0.0182859}, {{10.375, 86.375}, {334.75, 258.813}} -> {-1.33779, \
-1.27643, 0.0414819, 0.0414819}, {{93.9375, 169.875}, {334.75, 258.813}} -> \
{-3.64781, -1.27641, 0.0414819, 0.0414814}, {{177.5, 253.438}, {334.75, \
258.813}} -> {-5.95869, -1.2764, 0.0414819, 0.0414813}, {{261.063, 337}, \
{334.75, 258.813}} -> {-8.26958, -1.27643, 0.0414819, 0.0414819}, {{344.563, \
420.563}, {334.75, 258.813}} -> {-10.5796, -1.27643, 0.0414819, 0.0414819}, \
{{10.375, 86.375}, {251.188, 175.25}} -> {-1.33779, -3.58732, 0.0414819, \
0.0414819}, {{93.9375, 169.875}, {251.188, 175.25}} -> {-3.64781, -3.5873, \
0.0414819, 0.0414817}, {{177.5, 253.438}, {251.188, 175.25}} -> {-5.95871, \
-3.58732, 0.0414819, 0.0414819}, {{261.063, 337}, {251.188, 175.25}} -> \
{-8.26958, -3.58732, 0.0414819, 0.0414819}, {{344.563, 420.563}, {251.188, \
175.25}} -> {-10.5796, -3.58732, 0.0414819, 0.0414819}, {{10.375, 86.375}, \
{167.625, 91.6875}} -> {-1.33778, -5.8982, 0.0414818, 0.0414819}, {{93.9375, \
169.875}, {167.625, 91.6875}} -> {-3.64781, -5.8982, 0.0414819, 0.0414819}, \
{{177.5, 253.438}, {167.625, 91.6875}} -> {-5.95868, -5.8982, 0.0414818, \
0.0414819}, {{261.063, 337}, {167.625, 91.6875}} -> {-8.26958, -5.8982, \
0.0414819, 0.0414819}, {{344.563, 420.563}, {167.625, 91.6875}} -> {-10.5796, \
-5.8982, 0.0414818, 0.0414819}, {{10.375, 86.375}, {84.0625, 8.125}} -> \
{-1.33778, -8.20909, 0.0414818, 0.0414819}, {{93.9375, 169.875}, {84.0625, \
8.125}} -> {-3.64781, -8.20909, 0.0414819, 0.0414819}, {{177.5, 253.438}, \
{84.0625, 8.125}} -> {-5.95868, -8.20907, 0.0414818, 0.0414818}, {{261.063, \
337}, {84.0625, 8.125}} -> {-8.2696, -8.20909, 0.0414819, 0.0414819}, \
{{344.563, 420.563}, {84.0625, 8.125}} -> {-10.5796, -8.20909, 0.0414818, \
0.0414819}}]
}, Closed]],

Cell[TextData[{
  "Plot the curve described by the parametric equations ",
  Cell[BoxData[
      \(TraditionalForm\`{x = r\ \((1 - sin(t))\), \ 
        y = r\ \((1 - cos(t))\)}\)]],
  " ",
  "for the interval {t,-2\[Pi],2\[Pi]}."
}], "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["ImplicitPlot", "Subsection"],

Cell[TextData[{
  "The function ",
  ButtonBox["ImplicitPlot",
    ButtonStyle->"RefGuideLink"],
  ", found in a package by the same name, plots the solution to an equation \
and is closely related to ",
  StyleBox["ParametricPlot", "InlineInput",
    FontWeight->"Bold"],
  ".  Thus, ",
  StyleBox["ImplicitPlot[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["f[x,y]==g[x,y],{x,xmin,xmax}", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " uses ",
  StyleBox["Solve", "InlineInput",
    FontWeight->"Bold"],
  " to obtain a solution of the form ",
  StyleBox["y[x]", "InlineInput"],
  " and plots the result over the specified range of ",
  StyleBox["x", "InlineInput"],
  ".  Implicit contour plots and methods for avoiding singularities are also \
available; further detail can be obtained from ",
  StyleBox["Help",
    FontSlant->"Italic"],
  " when needed."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ImplicitPlot[y\^2 - \ x\^2\  == \ 1, \ {x, 5, \(-5\)}, 
        PlotPoints \[Rule] 50, \ ImageSize \[Rule] 144];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.0198 
%%ImageSize: 144 146.852 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.0952381 0.509902 0.0952381 [
[.11905 .4974 -9.15625 -15.6875 ]
[.11905 .4974 9.15625 0 ]
[.30952 .4974 -9.15625 -15.6875 ]
[.30952 .4974 9.15625 0 ]
[.69048 .4974 -5 -15.6875 ]
[.69048 .4974 5 0 ]
[.88095 .4974 -5 -15.6875 ]
[.88095 .4974 5 0 ]
[.4875 .12895 -18.3125 -7.84375 ]
[.4875 .12895 0 7.84375 ]
[.4875 .31943 -18.3125 -7.84375 ]
[.4875 .31943 0 7.84375 ]
[.4875 .70038 -10 -7.84375 ]
[.4875 .70038 0 7.84375 ]
[.4875 .89085 -10 -7.84375 ]
[.4875 .89085 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 1.0198 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.11905 .5099 m
.11905 .51615 L
s
gsave
.11905 .4974 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.30952 .5099 m
.30952 .51615 L
s
gsave
.30952 .4974 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.69048 .5099 m
.69048 .51615 L
s
gsave
.69048 .4974 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.88095 .5099 m
.88095 .51615 L
s
gsave
.88095 .4974 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.16667 .5099 m
.16667 .51365 L
s
.21429 .5099 m
.21429 .51365 L
s
.2619 .5099 m
.2619 .51365 L
s
.35714 .5099 m
.35714 .51365 L
s
.40476 .5099 m
.40476 .51365 L
s
.45238 .5099 m
.45238 .51365 L
s
.54762 .5099 m
.54762 .51365 L
s
.59524 .5099 m
.59524 .51365 L
s
.64286 .5099 m
.64286 .51365 L
s
.7381 .5099 m
.7381 .51365 L
s
.78571 .5099 m
.78571 .51365 L
s
.83333 .5099 m
.83333 .51365 L
s
.07143 .5099 m
.07143 .51365 L
s
.02381 .5099 m
.02381 .51365 L
s
.92857 .5099 m
.92857 .51365 L
s
.97619 .5099 m
.97619 .51365 L
s
.25 Mabswid
0 .5099 m
1 .5099 L
s
.5 .12895 m
.50625 .12895 L
s
gsave
.4875 .12895 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.5 .31943 m
.50625 .31943 L
s
gsave
.4875 .31943 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 .70038 m
.50625 .70038 L
s
gsave
.4875 .70038 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 .89085 m
.50625 .89085 L
s
gsave
.4875 .89085 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .17657 m
.50375 .17657 L
s
.5 .22419 m
.50375 .22419 L
s
.5 .27181 m
.50375 .27181 L
s
.5 .36704 m
.50375 .36704 L
s
.5 .41466 m
.50375 .41466 L
s
.5 .46228 m
.50375 .46228 L
s
.5 .55752 m
.50375 .55752 L
s
.5 .60514 m
.50375 .60514 L
s
.5 .65276 m
.50375 .65276 L
s
.5 .748 m
.50375 .748 L
s
.5 .79562 m
.50375 .79562 L
s
.5 .84324 m
.50375 .84324 L
s
.5 .08133 m
.50375 .08133 L
s
.5 .03371 m
.50375 .03371 L
s
.5 .93847 m
.50375 .93847 L
s
.5 .98609 m
.50375 .98609 L
s
.25 Mabswid
.5 0 m
.5 1.0198 L
s
0 0 m
1 0 L
1 1.0198 L
0 1.0198 L
closepath
clip
newpath
.5 Mabswid
.02381 .02428 m
.02479 .02524 L
.02772 .02811 L
.03259 .03289 L
.03938 .03954 L
.04807 .04804 L
.05861 .05836 L
.07097 .07043 L
.08509 .0842 L
.10091 .09961 L
.11837 .11657 L
.13741 .13501 L
.15793 .15482 L
.17986 .17589 L
.2031 .1981 L
.22756 .2213 L
.25315 .24531 L
.27974 .26994 L
.30725 .2949 L
.33554 .31986 L
.36451 .34429 L
.39404 .36743 L
.424 .38806 L
.45427 .40426 L
.48474 .41345 L
.51526 .41345 L
.54573 .40426 L
.576 .38806 L
.60596 .36743 L
.63549 .34429 L
.66446 .31986 L
.69275 .2949 L
.72026 .26994 L
.74685 .24531 L
.77244 .2213 L
.7969 .1981 L
.82014 .17589 L
.84207 .15482 L
.86259 .13501 L
.88163 .11657 L
.89909 .09961 L
.91491 .0842 L
.92903 .07043 L
.94139 .05836 L
.95193 .04804 L
.96062 .03954 L
.96741 .03289 L
.97228 .02811 L
.97521 .02524 L
.97619 .02428 L
Mistroke
Mfstroke
.02381 .99552 m
.02479 .99456 L
.02772 .99169 L
.03259 .98692 L
.03938 .98026 L
.04807 .97176 L
.05861 .96145 L
.07097 .94938 L
.08509 .93561 L
.10091 .9202 L
.11837 .90323 L
.13741 .88479 L
.15793 .86498 L
.17986 .84391 L
.2031 .8217 L
.22756 .79851 L
.25315 .77449 L
.27974 .74987 L
.30725 .7249 L
.33554 .69995 L
.36451 .67551 L
.39404 .65237 L
.424 .63175 L
.45427 .61555 L
.48474 .60636 L
.51526 .60636 L
.54573 .61555 L
.576 .63175 L
.60596 .65237 L
.63549 .67551 L
.66446 .69995 L
.69275 .7249 L
.72026 .74987 L
.74685 .77449 L
.77244 .79851 L
.7969 .8217 L
.82014 .84391 L
.84207 .86498 L
.86259 .88479 L
.88163 .90323 L
.89909 .9202 L
.91491 .93561 L
.92903 .94938 L
.94139 .96145 L
.95193 .97176 L
.96062 .98026 L
.96741 .98692 L
.97228 .99169 L
.97521 .99456 L
.97619 .99552 L
Mistroke
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{144, 146.813},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40003H0000g2000`400?l00000o`00003oJ`3oool0
0`000000oooo0?ooo`1Z0?ooo`00J`3oool00`000000oooo0?ooo`1Z0?ooo`00J`3oool00`000000
oooo0?ooo`1Z0?ooo`00J`3oool00`000000oooo0?ooo`1Z0?ooo`00J`3oool00`000000oooo0?oo
o`1Z0?ooo`00J`3oool00`000000oooo0?ooo`1Z0?ooo`001@3oool00`000000oooo0?ooo`1S0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool0103oool000H0oooo0P00001S0?ooo`03
0000003oool0oooo0640oooo0P0000070?ooo`001`3oool00`000000oooo0?ooo`1Q0?ooo`800000
H@3oool00`000000oooo0?ooo`070?ooo`00203oool00`000000oooo0?ooo`1P0?ooo`030000003o
ool0oooo0600oooo00<000000?ooo`3oool01`3oool000T0oooo00<000000?ooo`3oool0G`3oool0
0`000000oooo0?ooo`1O0?ooo`030000003oool0oooo00P0oooo000:0?ooo`030000003oool0oooo
05h0oooo00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`090?ooo`002`3oool00`000000
oooo0?ooo`1M0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool02P3oool000`0oooo
00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo00/0oooo
000=0?ooo`800000G03oool00`000000oooo0?ooo`1J0?ooo`8000003P3oool000l0oooo00<00000
0?ooo`3oool0F@3oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo00h0oooo000@0?oo
o`030000003oool0oooo05P0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`0?0?oo
o`004@3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3o
ool0403oool00180oooo00<000000?ooo`3oool0EP3oool2000005L0oooo00<000000?ooo`3oool0
4@3oool001<0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`1E0?ooo`030000003o
ool0oooo0180oooo000C0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0E03oool0
0`000000oooo0?ooo`0C0?ooo`00503oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo
05@0oooo00<000000?ooo`3oool04`3oool001D0oooo00<000000?ooo`3oool0D`3oool00`000000
oooo0?ooo`1C0?ooo`030000003oool0oooo01@0oooo000F0?ooo`030000003oool0oooo04L0oooo
00@0oooO09a80000B9`0gooo1`3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo01D0
oooo000G0?ooo`800000A`3oool0103ooml0W4P00018W03Oool70?ooo`030000003oool0oooo0500
oooo0P00000H0?ooo`006@3oool00`000000oooo0?ooo`140?ooo`040?oog`2LB00004RL0=ooo`L0
oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`0H0?ooo`006P3oool00`000000oooo
0?ooo`0o0?ooo`030;md00000000000000D0000000<0M;oo0?ooo`3oool0103oool00`000000oooo
0?ooo`1>0?ooo`030000003oool0oooo01T0oooo000K0?ooo`030000003oool0oooo0380oooo00<0
oooO09a8000000001`0000001018W=l0oooo0?oOW01dW=l20?ooo`040?oog`2LB00004RL0=ooo`L0
oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`0J0?ooo`00703oool00`000000oooo
0?ooo`0n0?ooo`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`070?ooo`800000C@3oool00`00
0000oooo0?ooo`0K0?ooo`007@3oool00`000000oooo0?ooo`0m0?ooo`070?oog`2LB7@0_ooo0?oo
g`2LB00004RL0=ooo`070?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0703oool0
01h0oooo00<000000?ooo`3oool0?@3oool01P3ogi`0B7Bo0?oog`2LB00004RL0=ooo`L0oooo00<0
00000?ooo`3oool0BP3oool00`000000oooo0?ooo`0M0?ooo`007`3oool00`000000oooo0?ooo`0m
0?ooo`050=nLB01d_ml0W4P00018W03Oool01`3oool00`000000oooo0?ooo`190?ooo`030000003o
ool0oooo01h0oooo000O0?ooo`030000003oool0oooo03d0oooo00D0oooO09a8B018B00004RL0=oo
o`070?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07P3oool00200oooo00<00000
0?ooo`3oool0?@3oool0103ogi`0B0000018W03Oool70?ooo`030000003oool0oooo04P0oooo00<0
00000?ooo`3oool07`3oool00240oooo00<000000?ooo`3oool0?@3oool00`3OW4P004RL0=ooo`07
0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0803oool00280oooo00<000000?oo
o`3oool0AP3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo0240oooo000S0?ooo`80
0000AP3oool00`000000oooo0?ooo`140?ooo`800000903oool002D0oooo00<000000?ooo`3oool0
@`3oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo02@0oooo000V0?ooo`030000003o
ool0oooo0480oooo0P0000130?ooo`030000003oool0oooo02D0oooo000W0?ooo`030000003oool0
oooo0440oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`0V0?ooo`00:03oool00`00
0000oooo0?ooo`100?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool09`3oool002T0
oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo02P0
oooo000Z0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool0?P3oool00`000000oooo
0?ooo`0Y0?ooo`00:`3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo03d0oooo00<0
00000?ooo`3oool0:P3oool002`0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`0l
0?ooo`030000003oool0oooo02/0oooo000]0?ooo`030000003oool0oooo03/0oooo00<000000?oo
o`3oool0>`3oool00`000000oooo0?ooo`0/0?ooo`00;P3oool00`000000oooo0?ooo`0j0?ooo`03
0000003oool0oooo03X0oooo00<000000?ooo`3oool0;@3oool002l0oooo00<000000?ooo`3oool0
>@3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo02h0oooo000`0?ooo`030000003o
ool0oooo03P0oooo0P00000i0?ooo`030000003oool0oooo02l0oooo000a0?ooo`030000003oool0
oooo03L0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0`0?ooo`00<P3oool00`00
0000oooo0?ooo`0f0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0<@3oool003<0
oooo0P00000f0?ooo`030000003oool0oooo03@0oooo0P00000d0?ooo`00=@3oool00`000000oooo
0?ooo`0c0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0=03oool003H0oooo00<0
00000?ooo`3oool0<P3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo03D0oooo000g
0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0f
0?ooo`00>03oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo0300oooo00<000000?oo
o`3oool0=`3oool003T0oooo0P00000`0?ooo`030000003oool0oooo02h0oooo0P00000j0?ooo`00
>`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0
>P3oool003`0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0/0?ooo`030000003o
ool0oooo03/0oooo000m0?ooo`030000003oool0oooo02/0oooo0P00000/0?ooo`030000003oool0
oooo03`0oooo000n0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0:P3oool00`00
0000oooo0?ooo`0m0?ooo`00?`3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo02T0
oooo00<000000?ooo`3oool0?P3oool00400oooo00<000000?ooo`3oool0:03oool00`000000oooo
0?ooo`0X0?ooo`030000003oool0oooo03l0oooo00110?ooo`030000003oool0oooo02L0oooo00<0
00000?ooo`3oool09`3oool00`000000oooo0?ooo`100?ooo`00@P3oool00`000000oooo0?ooo`0G
0?ooo`030;md00000000000000@0000000<0B9cO0?ooo`3oool01@3oool00`000000oooo0?ooo`0V
0?ooo`030000003oool0oooo0440oooo00130?ooo`030000003oool0oooo01L0oooo00<0gia80000
000000000`0000000`00B9`0gooo0?ooo`050?ooo`030000003oool0oooo02D0oooo00<000000?oo
o`3oool0@P3oool004@0oooo00<000000?ooo`3oool05`3oool00`3OW4P0B9cO0?ooo`020?ooo`03
0=nLM02Lgol0oooo00D0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`130?ooo`00
A@3oool2000001P0oooo00<0gia807Boo`3oool02@3oool00`000000oooo0?ooo`0R0?ooo`800000
AP3oool004L0oooo00<000000?ooo`3oool01P3oool00`3ooml0W4P0000000070000000304RLg`3o
ool0oooo00<0oooo00<0_gA809cOo`3oool0203oool00`000000oooo0?ooo`0Q0?ooo`030000003o
ool0oooo04H0oooo00180?ooo`030000003oool0oooo01D0oooo00<0oooO09a8M02oool0203oool2
00000240oooo00<000000?ooo`3oool0A`3oool004T0oooo00<000000?ooo`3oool05@3oool00`3o
_g@007Bo0?ooo`070?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0B03oool004X0
oooo00<000000?ooo`3oool0503oool00`3ookl0M01809cOo`070?ooo`030000003oool0oooo01h0
oooo00<000000?ooo`3oool0B@3oool004/0oooo0P00000D0?ooo`030?oo_`1d0000B9cO00L0oooo
00<000000?ooo`3oool0703oool2000004`0oooo001=0?ooo`030000003oool0oooo00`0oooo00<0
oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`070?ooo`030000003oool0oooo01/0oooo00<0
00000?ooo`3oool0C03oool004h0oooo00<000000?ooo`3oool02`3oool0203ooml0W4Q809cOo`3o
ool0oono07@0000007@0_ooo1`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo04d0
oooo001?0?ooo`030000003oool0oooo00/0oooo00H0oooO09a80000000000000000M02oool80?oo
o`030000003oool0oooo01T0oooo00<000000?ooo`3oool0CP3oool00500oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo04l0oooo001A0?ooo`800000
603oool00`000000oooo0?ooo`0F0?ooo`800000DP3oool005<0oooo00<000000?ooo`3oool05@3o
ool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0580oooo001D0?ooo`030000003oool0
oooo01@0oooo0P00000E0?ooo`030000003oool0oooo05<0oooo001E0?ooo`800000503oool00`00
0000oooo0?ooo`0B0?ooo`800000EP3oool005L0oooo00<000000?ooo`3oool04@3oool00`000000
oooo0?ooo`0A0?ooo`030000003oool0oooo05H0oooo001H0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1G0?ooo`00F@3oool200000100oooo00<0
00000?ooo`3oool03P3oool2000005X0oooo001K0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`1J0?ooo`00G03oool2000000d0oooo00<000000?oo
o`3oool02`3oool2000005d0oooo001N0?ooo`8000002`3oool00`000000oooo0?ooo`090?ooo`80
0000G`3oool00600oooo0`0000080?ooo`030000003oool0oooo00H0oooo0`00001Q0?ooo`00H`3o
ool4000000@0oooo00<000000?ooo`3oool00P3oool4000006@0oooo001W0?ooo`T00000J03oool0
01l0oooo00@0oooO09a80000B9`0gooo8@3oool00`2oM0000000000000040000000304RLg`3oool0
oooo01d0oooo00<000000?ooo`3oool08@3oool00`2oM0000000000000040000000304RLg`3oool0
oooo02<0oooo00@0oooO09a80000B9`0gooo603oool001l0oooo00@0oooO09a80000B9`0gooo8P3o
ool00`3OW4P0000000000003000000030018W03Oool0oooo01d0oooo00<000000?ooo`3oool08P3o
ool00`3OW4P0000000000003000000030018W03Oool0oooo02<0oooo00@0oooO09a80000B9`0gooo
603oool001l0oooo00@0oooO09a80000B9`0gooo8`3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nL
M02Lgol0oooo01d0oooo00<000000?ooo`3oool08`3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nL
M02Lgol0oooo02<0oooo00@0oooO09a80000B9`0gooo603oool001/0oooo00<0_g@0000000000000
1@0000000`1d_ol0oooo0?ooo`0Q0?ooo`030=nLB01d_ol0oooo0240oooo00<000000?ooo`3oool0
903oool00`3OW4P0M;oo0?ooo`0S0?ooo`030;md00000000000000D0000000<0M;oo0?ooo`3oool0
5@3oool000h0oooo00<0oooO09a8000000001`0000001@18W=l0oooo0?ooo`3ogi`0M9cO0080oooo
00@0oooO09a80000B9`0gooo503oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00@0
oooo00<0_gA809cOo`3oool0803oool00`000000oooo0?ooo`0U0?ooo`030;mdB02Lgol0oooo0280
oooo00P0omnL07BLg`3oool0oooo0?oog`2LB00004RL0=oooaP0oooo000L0?ooo`070=nLB01d_ol0
oooo0?oog`2LB00004RL0=ooo`0U0?ooo`030?oog`2LB7@0_ooo0200oooo00<000000?ooo`3oool0
9@3oool00`3ooml0W4Qd0;ooo`0S0?ooo`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`0H0?oo
o`00703oool01`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool09P3oool00`3o_g@007Bo0?ooo`0O
0?ooo`030000003oool0oooo02H0oooo00<0okmd001d_`3oool08P3oool01`3ooml0W4Qd0;ooo`3o
oml0W4P00018W03Oool0603oool001d0oooo00H0omnL04Qd_`3ooml0W4P00018W03OoolV0?ooo`03
0?oo_`1d04P0W=oo01l0oooo00<000000?ooo`3oool09P3oool00`3ookl0M01809cOo`0S0?ooo`06
0?oOW018M;l0oooO09a80000B9`0gooo603oool001h0oooo00D0gia807Bog`2LB00004RL0=ooo`0V
0?ooo`030?oo_`1d0000B9cO01l0oooo00<000000?ooo`3oool09P3oool00`3ookl0M00004RLg`0T
0?ooo`050=nLB01d_ml0W4P00018W03Oool0603oool001h0oooo00D0oooO09a8B018B00004RL0=oo
o`0Q0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l07`3oool00`000000oooo0?oo
o`0Q0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l0903oool01@3ooml0W4Q804Q8
0000B9`0gooo01P0oooo000O0?ooo`040?oOW018000004RL0=ooob40oooo00P0oooO09a8B02Lgol0
oooo0?oo_`1d0000001d0;oooal0oooo0P00000R0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M000
0000M02ooolU0?ooo`040?oOW018000004RL0=oooaP0oooo000P0?ooo`030=nLB000B9`0gooo0280
oooo00H0oooO09a80000000000000000M02ooolP0?ooo`030000003oool0oooo0280oooo00H0oooO
09a80000000000000000M02ooolW0?ooo`030=nLB000B9`0gooo01P0oooo001[0?ooo`030000003o
ool0oooo06X0oooo001[0?ooo`030000003oool0oooo06X0oooo001[0?ooo`030000003oool0oooo
06X0oooo001[0?ooo`030000003oool0oooo06X0oooo001[0?ooo`030000003oool0oooo06X0oooo
001[0?ooo`030000003oool0oooo06X0oooo001[0?ooo`030000003oool0oooo06X0oooo001[0?oo
o`030000003oool0oooo06X0oooo003G00000040oooo00050?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`040?ooo`00J`3oool00`000000oooo0?oo
o`1Z0?ooo`00J`3oool00`000000oooo0?ooo`1Z0?ooo`00J`3oool00`000000oooo0?ooo`1Z0?oo
o`00J`3oool00`000000oooo0?ooo`1Z0?ooo`00J`3oool00`000000oooo0?ooo`1Z0?ooo`00J`3o
ool00`000000oooo0?ooo`1Z0?ooo`00J`3oool00`000000oooo0?ooo`1Z0?ooo`00J`3oool00`00
0000oooo0?ooo`1Z0?ooo`00J`3oool2000006/0oooo001[0?ooo`030000003oool0oooo06X0oooo
001[0?ooo`030000003oool0oooo06X0oooo001[0?ooo`030000003oool0oooo06X0oooo001[0?oo
o`030000003oool0oooo06X0oooo001[0?ooo`030000003oool0oooo06X0oooo001[0?ooo`030000
003oool0oooo06X0oooo001[0?ooo`030000003oool0oooo06X0oooo001[0?ooo`030000003oool0
oooo06X0oooo001[0?ooo`030000003oool0oooo06X0oooo001[0?ooo`800000J`3oool006L0oooo
2@00001X0?ooo`00H`3oool4000000@0oooo00<000000?ooo`3oool00P3oool4000006@0oooo001Q
0?ooo`800000203oool00`000000oooo0?ooo`060?ooo`800000HP3oool005l0oooo0P00000:0?oo
o`030000003oool0oooo00P0oooo0P00001P0?ooo`00G@3oool2000000`0oooo00<000000?ooo`3o
ool02P3oool2000005h0oooo001K0?ooo`8000003P3oool00`000000oooo0?ooo`0<0?ooo`800000
G03oool005T0oooo0P00000@0?ooo`030000003oool0oooo00h0oooo0P00001J0?ooo`00F03oool0
0`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0E`3oool0
05L0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
05H0oooo001E0?ooo`800000503oool00`000000oooo0?ooo`0B0?ooo`800000EP3oool005@0oooo
00<000000?ooo`3oool0503oool2000001D0oooo00<000000?ooo`3oool0D`3oool00580oooo0P00
000G0?ooo`030000003oool0oooo01D0oooo0P00001C0?ooo`00D@3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0D03oool004l0oooo0P00000J0?oo
o`030000003oool0oooo01P0oooo0P00001@0?ooo`00CP3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo01X0oooo00<000000?ooo`3oool0C@3oool004d0oooo00<000000?ooo`3oool0
303oool00`2oM0000000000000040000000304RLg`3oool0oooo00D0oooo00<000000?ooo`3oool0
6`3oool00`000000oooo0?ooo`1<0?ooo`00C03oool00`000000oooo0?ooo`0>0?ooo`030=nLB000
0000000000<0000000<004RL0=ooo`3oool01@3oool00`000000oooo0?ooo`0L0?ooo`030000003o
ool0oooo04/0oooo001;0?ooo`030000003oool0oooo0100oooo00<0gia804RLg`3oool00P3oool0
0`3OW7@0W=oo0?ooo`050?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0BP3oool0
04X0oooo00<000000?ooo`3oool04P3oool00`3OW4P0M;oo0?ooo`090?ooo`030000003oool0oooo
01h0oooo00<000000?ooo`3oool0B@3oool004T0oooo00<000000?ooo`3oool0503oool00`2oM4P0
W=oo0?ooo`080?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0B03oool004P0oooo
00<000000?ooo`3oool05@3oool00`3ooml0W4Qd0;ooo`080?ooo`8000008@3oool00`000000oooo
0?ooo`170?ooo`00A`3oool00`000000oooo0?ooo`0G0?ooo`030?noM000M;l0oooo00L0oooo00<0
00000?ooo`3oool08@3oool00`000000oooo0?ooo`160?ooo`00AP3oool00`000000oooo0?ooo`0H
0?ooo`030?oo_`1d04P0W=oo00L0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`15
0?ooo`00A03oool2000001/0oooo00<0oono07@00018W=l01`3oool00`000000oooo0?ooo`0S0?oo
o`800000A@3oool004<0oooo00<000000?ooo`3oool05P3oool00`3o_i`0W=oo0?ooo`020?ooo`03
0?oOW0180000B9cO00L0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`120?ooo`00
@P3oool00`000000oooo0?ooo`0G0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool7
0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0@@3oool00440oooo00<000000?oo
o`3oool06@3oool01P3ooml0W4P0000000000000001d0;ooo`P0oooo00<000000?ooo`3oool09`3o
ool00`000000oooo0?ooo`100?ooo`00@03oool00`000000oooo0?ooo`0X0?ooo`030000003oool0
oooo02P0oooo00<000000?ooo`3oool0?`3oool003h0oooo0P00000[0?ooo`030000003oool0oooo
02T0oooo0P00000o0?ooo`00?@3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo02/0
oooo00<000000?ooo`3oool0?03oool003`0oooo00<000000?ooo`3oool0;03oool2000002d0oooo
00<000000?ooo`3oool0>`3oool003/0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?oo
o`0]0?ooo`030000003oool0oooo03X0oooo000j0?ooo`030000003oool0oooo02h0oooo00<00000
0?ooo`3oool0;P3oool00`000000oooo0?ooo`0i0?ooo`00>@3oool00`000000oooo0?ooo`0_0?oo
o`030000003oool0oooo02l0oooo00<000000?ooo`3oool0>03oool003P0oooo00<000000?ooo`3o
ool0<03oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo03L0oooo000g0?ooo`030000
003oool0oooo0340oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0f0?ooo`00=P3o
ool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0=@3o
ool003D0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0
oooo03@0oooo000d0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0=03oool00`00
0000oooo0?ooo`0c0?ooo`00<P3oool2000003L0oooo00<000000?ooo`3oool0=@3oool2000003<0
oooo000a0?ooo`030000003oool0oooo03L0oooo0P00000h0?ooo`030000003oool0oooo0300oooo
000`0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?oo
o`0_0?ooo`00;`3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo03T0oooo00<00000
0?ooo`3oool0;P3oool002h0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0j0?oo
o`030000003oool0oooo02d0oooo000]0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3o
ool0>`3oool00`000000oooo0?ooo`0/0?ooo`00;03oool00`000000oooo0?ooo`0l0?ooo`030000
003oool0oooo03`0oooo00<000000?ooo`3oool0:`3oool002/0oooo00<000000?ooo`3oool0?@3o
ool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo02X0oooo000Z0?ooo`030000003oool0
oooo03h0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`0Y0?ooo`00:@3oool00`00
0000oooo0?ooo`0o0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0:03oool002P0
oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo02L0
oooo000W0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool0@@3oool00`000000oooo
0?ooo`0V0?ooo`009P3oool00`000000oooo0?ooo`120?ooo`800000@`3oool00`000000oooo0?oo
o`0U0?ooo`009@3oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo04<0oooo00<00000
0?ooo`3oool0903oool002<0oooo0P0000160?ooo`030000003oool0oooo04@0oooo0P00000T0?oo
o`008P3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3o
ool08@3oool00240oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`170?ooo`030000
003oool0oooo0200oooo000P0?ooo`030000003oool0oooo03d0oooo00@0oooO09a80000B9`0gooo
1`3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo01l0oooo000O0?ooo`030000003o
ool0oooo03h0oooo00@0oooO09a80000B9`0gooo1`3oool00`000000oooo0?ooo`190?ooo`030000
003oool0oooo01h0oooo000N0?ooo`030000003oool0oooo03l0oooo00@0oooO09a80000B9`0gooo
1`3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo01d0oooo000M0?ooo`030000003o
ool0oooo03`0oooo00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`040?ooo`030000003o
ool0oooo04/0oooo00<000000?ooo`3oool0703oool001`0oooo00<000000?ooo`3oool0?@3oool0
203ogi`0M9cO0?ooo`3oool0oooO09a80000B9`0gooo1`3oool00`000000oooo0?ooo`1<0?ooo`03
0000003oool0oooo01/0oooo000K0?ooo`030000003oool0oooo03l0oooo00L0gia807Boo`3oool0
oooO09a80000B9`0gooo00L0oooo0P00001>0?ooo`030000003oool0oooo01X0oooo000J0?ooo`03
0000003oool0oooo0400oooo00L0oooO09a8M02oool0oooO09a80000B9`0gooo00L0oooo00<00000
0?ooo`3oool0CP3oool00`000000oooo0?ooo`0I0?ooo`006@3oool00`000000oooo0?ooo`120?oo
o`060?oOW018M;l0oooO09a80000B9`0gooo1`3oool00`000000oooo0?ooo`1?0?ooo`030000003o
ool0oooo01P0oooo000H0?ooo`030000003oool0oooo04@0oooo00D0gia807Bog`2LB00004RL0=oo
o`070?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool05`3oool001L0oooo00<00000
0?ooo`3oool0A@3oool01@3ooml0W4Q804Q80000B9`0gooo00L0oooo00<000000?ooo`3oool0D@3o
ool00`000000oooo0?ooo`0F0?ooo`005P3oool00`000000oooo0?ooo`170?ooo`040?oOW0180000
04RL0=ooo`L0oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`0E0?ooo`005@3oool0
0`000000oooo0?ooo`190?ooo`030=nLB000B9`0gooo00L0oooo00<000000?ooo`3oool0D`3oool0
0`000000oooo0?ooo`0D0?ooo`00503oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo
05@0oooo00<000000?ooo`3oool04`3oool001<0oooo00<000000?ooo`3oool0E@3oool00`000000
oooo0?ooo`1E0?ooo`030000003oool0oooo0180oooo000B0?ooo`030000003oool0oooo05H0oooo
00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`0A0?ooo`004@3oool00`000000oooo0?oo
o`1G0?ooo`800000F03oool00`000000oooo0?ooo`0@0?ooo`00403oool00`000000oooo0?ooo`1H
0?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool03`3oool000l0oooo00<000000?oo
o`3oool0F@3oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo00h0oooo000>0?ooo`03
0000003oool0oooo05X0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`0=0?ooo`00
3@3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0
303oool000`0oooo00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`1L0?ooo`030000003o
ool0oooo00/0oooo000;0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool0G@3oool0
0`000000oooo0?ooo`0:0?ooo`002P3oool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo
05h0oooo00<000000?ooo`3oool02@3oool000T0oooo00<000000?ooo`3oool0G`3oool00`000000
oooo0?ooo`1O0?ooo`030000003oool0oooo00P0oooo00080?ooo`030000003oool0oooo0600oooo
00<000000?ooo`3oool0H03oool00`000000oooo0?ooo`070?ooo`001`3oool00`000000oooo0?oo
o`1Q0?ooo`800000HP3oool00`000000oooo0?ooo`060?ooo`001@3oool2000006@0oooo00<00000
0?ooo`3oool0HP3oool2000000H0oooo00050?ooo`030000003oool0oooo06<0oooo00<000000?oo
o`3oool0H`3oool00`000000oooo0?ooo`040?ooo`00J`3oool00`000000oooo0?ooo`1Z0?ooo`00
J`3oool00`000000oooo0?ooo`1Z0?ooo`00J`3oool00`000000oooo0?ooo`1Z0?ooo`00J`3oool0
0`000000oooo0?ooo`1Z0?ooo`00J`3oool00`000000oooo0?ooo`1Z0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 215}, {219.188, 0}} -> {-5.25172, -5.35402, \
0.0732798, 0.0732798}}]
}, Closed]],

Cell[TextData[{
  "Notice that both branches are obtained automatically.  ",
  StyleBox["ImplicitPlot", "InlineInput",
    FontWeight->"Bold"],
  " will also accept a list of equations."
}], "Text"],

Cell[TextData[{
  "Plot the real solutions to the lemniscate equation, ",
  Cell[BoxData[
      FormBox[
        RowBox[{
          SuperscriptBox[
            RowBox[{"(", 
              FormBox[\((x\^2 + y\^2)\),
                "TraditionalForm"], ")"}], \(2\ \)], "=", " ", 
          \(x\^2 - y\^2\)}], TraditionalForm]]],
  ". "
}], "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Three-dimensional parametric functions", "Subsection"],

Cell[TextData[{
  "A three-dimensional trajectory expressed in terms of three coordinate \
functions ",
  StyleBox["{x[t],y[t],z[t]}", "InlineInput"],
  StyleBox[" ",
    FontSlant->"Italic"],
  "connected by a single parameter ",
  StyleBox["t", "InlineInput"],
  " can be displayed using ",
  ButtonBox["ParametricPlot3D",
    ButtonStyle->"RefGuideLink"],
  StyleBox[" ",
    FontWeight->"Bold"],
  "with basic syntax \n",
  StyleBox["ParametricPlot3D[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["{", "InlineInput"],
  StyleBox["x[t],y[t],z[t]}", "InlineInput"],
  StyleBox[", {", "InlineInput"],
  StyleBox["t,tmin,tmax", "InlineInput"],
  StyleBox["}", "InlineInput"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  ". "
}], "Text"],

Cell[TextData[{
  "For example, a charged particle in a uniform magnetic field follows a \
helical trajectory which can be parametrized and displayed as follows.  For \
simplicity we have chosen the magnetic field along the ",
  StyleBox["z",
    FontSlant->"Italic"],
  "-axis and have chosen an arbitrary ratio between angular and linear \
velocities.  In order to display just the shape, the box and axes have been \
suppressed. "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(path = {Cos[5  t], Sin[5  t], t};\)\), "\n", 
    \(\(ParametricPlot3D[Evaluate[path], {t, 0, 10}, Boxed \[Rule] False, 
        Axes \[Rule] False, AspectRatio \[Rule] Automatic];\)\)}], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 3.26553 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-1.21738 3.44527 -0.179738 3.44527 [
[ 0 0 0 0 ]
[ 1 3.26553 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 3.26553 L
0 3.26553 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.51699 .63392 m
.33101 .62665 L
s
.69733 .55696 m
.51699 .63392 L
s
.79612 .41422 m
.69733 .55696 L
s
.33101 .62665 m
.21767 .55021 L
s
.76823 .25263 m
.79612 .41422 L
s
.57663 .90313 m
.37993 .92232 L
s
.74286 .80271 m
.57663 .90313 L
s
.21767 .55021 m
.22865 .4489 L
s
.37993 .92232 m
.23485 .86531 L
s
.80769 .64862 m
.74286 .80271 L
s
.23485 .86531 m
.20526 .76826 L
s
.22865 .4489 m
.3641 .381 L
s
.73953 .49303 m
.80769 .64862 L
s
.63577 1.17557 m
.43564 1.22111 L
s
.7807 1.05478 m
.63577 1.17557 L
s
.43564 1.22111 m
.26365 1.1866 L
s
.56533 .39269 m
.73953 .49303 L
s
.8072 .89444 m
.7807 1.05478 L
s
.3641 .381 m
.56533 .39269 L
s
.26365 1.1866 m
.1936 1.09917 L
s
.20526 .76826 m
.30911 .68695 L
s
.49624 1.52345 m
.30323 1.51346 L
s
.69213 1.45267 m
.49624 1.52345 L
s
.69962 .75001 m
.8072 .89444 L
s
.80895 1.31536 m
.69213 1.45267 L
s
.30911 .68695 m
.50353 .67364 L
s
.1936 1.09917 m
.26049 1.00992 L
s
.30323 1.51346 m
.19447 1.4404 L
s
.50353 .67364 m
.69962 .75001 L
s
.79404 1.15409 m
.80895 1.31536 L
s
.5596 1.83014 m
.3524 1.84566 L
s
.74341 1.73621 m
.5596 1.83014 L
s
.19447 1.4404 m
.2205 1.34896 L
s
.26049 1.00992 m
.43932 .97442 L
s
.3524 1.84566 m
.20817 1.79094 L
s
.64963 1.02542 m
.79404 1.15409 L
s
.82601 1.58685 m
.74341 1.73621 L
s
.43932 .97442 m
.64963 1.02542 L
s
.76808 1.42997 m
.82601 1.58685 L
s
.20817 1.79094 m
.19111 1.70297 L
s
.2205 1.34896 m
.37539 1.2952 L
s
.62343 2.14242 m
.40957 2.18337 L
s
.78736 2.0283 m
.62343 2.14242 L
s
.40957 2.18337 m
.2345 2.15 L
s
.59125 1.32081 m
.76808 1.42997 L
s
.83056 1.87179 m
.78736 2.0283 L
s
.37539 1.2952 m
.59125 1.32081 L
s
.2345 2.15 m
.17382 2.07081 L
s
.19111 1.70297 m
.31451 1.63575 L
s
.72971 1.72433 m
.83056 1.87179 L
s
.47284 2.5272 m
.27279 2.51717 L
s
.68533 2.46191 m
.47284 2.5272 L
s
.82188 2.33132 m
.68533 2.46191 L
s
.31451 1.63575 m
.52659 1.63736 L
s
.27279 2.51717 m
.16971 2.45143 L
s
.17382 2.07081 m
.25944 1.99554 L
s
.52659 1.63736 m
.72971 1.72433 L
s
.82171 2.17279 m
.82188 2.33132 L
s
.54006 2.87819 m
.32194 2.8924 L
s
.74282 2.79064 m
.54006 2.87819 L
s
.67992 2.03921 m
.82171 2.17279 L
s
.16971 2.45143 m
.21271 2.37378 L
s
.84509 2.64788 m
.74282 2.79064 L
s
.25944 1.99554 m
.45821 1.97586 L
s
.32194 2.8924 m
.17931 2.84398 L
s
.45821 1.97586 m
.67992 2.03921 L
s
.79906 2.49243 m
.84509 2.64788 L
s
.17931 2.84398 m
.17659 2.76952 L
s
.21271 2.37378 m
.38891 2.33665 L
s
.62024 2.37637 m
.79906 2.49243 L
s
.38891 2.33665 m
.62024 2.37637 L
s
.17659 2.76952 m
.32169 2.71971 L
s
.32169 2.71971 m
.55278 2.73719 L
s
.55278 2.73719 m
.76276 2.83319 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{88.1875, 287.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000240001[b000`400?l00000o`00003oQ03oool0
08@0oooo00240?ooo`00Q03oool008@0oooo00240?ooo`00Q03oool008@0oooo00240?ooo`00Q03o
ool008@0oooo00240?ooo`00Q03oool008@0oooo00240?ooo`00Q03oool008@0oooo00240?ooo`00
Q03oool008@0oooo00240?ooo`00Q03oool008@0oooo00240?ooo`00Q03oool008@0oooo00240?oo
o`00Q03oool008@0oooo00240?ooo`00Q03oool008@0oooo00240?ooo`00Q03oool006@0oooo00<0
00000?ooo`3oool07@3oool006@0oooo00<000000?ooo`3oool07@3oool006@0oooo00<000000?oo
o`3oool07@3oool006D0oooo00<000000?ooo`3oool0703oool006D0oooo00<000000?ooo`3oool0
703oool006D0oooo00<000000?ooo`3oool0703oool006D0oooo00<000000?ooo`3oool0703oool0
06D0oooo00<000000?ooo`3oool0703oool006H0oooo00<000000?ooo`3oool06`3oool006H0oooo
00<000000?ooo`3oool06`3oool006H0oooo00<000000?ooo`3oool06`3oool006H0oooo00<00000
0?ooo`3oool06`3oool006H0oooo00<000000?ooo`3oool06`3oool006H0oooo00<000000?ooo`3o
ool06`3oool006L0oooo00<000000?ooo`3oool06P3oool006L0oooo00<000000?ooo`3oool06P3o
ool006L0oooo00<000000?ooo`3oool06P3oool002l0oooo3P00000Z0?ooo`030000003oool0oooo
01X0oooo000]0?ooo`8000003P3oool>000001`0oooo00<000000?ooo`3oool06P3oool002/0oooo
0P00000N0?ooo`8000006`3oool00`000000oooo0?ooo`0I0?ooo`00:@3oool200000280oooo0P00
000I0?ooo`030000003oool0oooo01T0oooo000W0?ooo`8000009P3oool2000001L0oooo00<00000
0?ooo`3oool06@3oool002D0oooo0P00000Z0?ooo`030000003oool0oooo01<0oooo00<000000?oo
o`3oool06P3oool002<0oooo0P00000]0?ooo`8000004`3oool00`000000oooo0?ooo`0J0?ooo`00
8@3oool200000340oooo0P00000@0?ooo`030000003oool0oooo01/0oooo000O0?ooo`800000=@3o
ool2000000d0oooo00<000000?ooo`3oool0703oool001h0oooo00<000000?ooo`3oool0=`3oool2
000000/0oooo00<000000?ooo`3oool0703oool001h0oooo00<000000?ooo`3oool0>@3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo01d0oooo000N0?ooo`030000003oool0oooo03X0
oooo0P0000060?ooo`030000003oool0oooo01h0oooo000N0?ooo`030000003oool0oooo03`0oooo
0P0000040?ooo`030000003oool0oooo01h0oooo000M0?ooo`030000003oool0oooo03l0oooo0P00
00000`3oool000000?ooo`0P0?ooo`007@3oool00`000000oooo0?ooo`110?ooo`030000003oool0
oooo0200oooo000M0?ooo`030000003oool0oooo0400oooo0P00000R0?ooo`007@3oool00`000000
oooo0?ooo`100?ooo`030000003oool000000240oooo000M0?ooo`030000003oool0oooo03l0oooo
00@000000?ooo`3oool000008@3oool001d0oooo00<000000?ooo`3oool0?P3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo01h0oooo000L0?ooo`030000003oool0oooo03l0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0N0?ooo`00703oool00`000000oooo0?ooo`0n
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool07@3oool001`0oooo00<000000?oo
o`3oool0?@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01d0oooo000L0?ooo`03
0000003oool0oooo03d0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0M0?ooo`00
7@3oool2000003/0oooo0P0000090?ooo`030000003oool0oooo01`0oooo000O0?ooo`030000003o
ool0oooo03H0oooo0P00000;0?ooo`030000003oool0oooo01`0oooo000P0?ooo`800000=03oool2
000000h0oooo00<000000?ooo`3oool06`3oool00280oooo00<000000?ooo`3oool0;P3oool30000
0100oooo00<000000?ooo`3oool06`3oool002<0oooo0P00000/0?ooo`800000503oool00`000000
oooo0?ooo`0J0?ooo`009@3oool00`000000oooo0?ooo`0V0?ooo`<000005P3oool00`000000oooo
0?ooo`0J0?ooo`009P3oool2000002@0oooo0P00000I0?ooo`030000003oool0oooo01X0oooo000X
0?ooo`030000003oool0oooo01l0oooo0P00000L0?ooo`030000003oool0oooo01T0oooo000Y0?oo
o`800000703oool3000001h0oooo00<000000?ooo`3oool06@3oool002/0oooo3000000>0?ooo`80
00008P3oool00`000000oooo0?ooo`0H0?ooo`00=`3oool>000002@0oooo00<000000?ooo`3oool0
603oool006X0oooo00<000000?ooo`3oool05`3oool006X0oooo00<000000?ooo`3oool05`3oool0
06X0oooo00<000000?ooo`3oool05`3oool006T0oooo00<000000?ooo`3oool0603oool003`0oooo
2000000U0?ooo`030000003oool0oooo01P0oooo000_0?ooo`d00000203oool200000280oooo00<0
00000?ooo`3oool06@3oool002P0oooo1`00000G0?ooo`<000007`3oool00`000000oooo0?ooo`0I
0?ooo`009`3oool00`000000oooo0?ooo`0O0?ooo`800000703oool00`000000oooo0?ooo`0J0?oo
o`009@3oool2000002@0oooo0`00000I0?ooo`030000003oool0oooo01X0oooo000T0?ooo`030000
003oool0oooo02L0oooo0P00000F0?ooo`030000003oool0oooo01/0oooo000S0?ooo`030000003o
ool0oooo02X0oooo0`00000C0?ooo`030000003oool0oooo01/0oooo000R0?ooo`030000003oool0
oooo02h0oooo0P00000@0?ooo`030000003oool0oooo01`0oooo000P0?ooo`800000<`3oool30000
00d0oooo00<000000?ooo`3oool0703oool001l0oooo00<000000?ooo`3oool0=P3oool2000000/0
oooo00<000000?ooo`3oool0703oool001h0oooo00<000000?ooo`3oool0>@3oool2000000P0oooo
00<000000?ooo`3oool07@3oool001`0oooo0P00000n0?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool07@3oool001/0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo01h0oooo000J0?ooo`030000003oool0oooo0400oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0N0?ooo`006P3oool00`000000oooo0?ooo`110?ooo`05
0000003oool0oooo0?ooo`0000008@3oool001/0oooo00<000000?ooo`3oool0@@3oool010000000
oooo0?ooo`00000Q0?ooo`006`3oool00`000000oooo0?ooo`110?ooo`030000003oool000000280
oooo000K0?ooo`030000003oool0oooo0480oooo0P00000R0?ooo`00703oool00`000000oooo0?oo
o`100?ooo`<000008P3oool001`0oooo00<000000?ooo`3oool0?P3oool2000000<0oooo00<00000
0?ooo`3oool07`3oool001`0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo01l0oooo000M0?ooo`030000003oool0oooo03X0oooo0P0000070?ooo`03
0000003oool0oooo01h0oooo000M0?ooo`030000003oool0oooo03P0oooo0P00000:0?ooo`030000
003oool0oooo01d0oooo000M0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool02`3o
ool00`000000oooo0?ooo`0L0?ooo`007P3oool00`000000oooo0?ooo`0d0?ooo`8000003P3oool0
0`000000oooo0?ooo`0L0?ooo`007P3oool2000003<0oooo0P00000A0?ooo`030000003oool0oooo
01/0oooo000P0?ooo`800000;`3oool2000001@0oooo00<000000?ooo`3oool06P3oool00280oooo
0P00000/0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool06@3oool002@0oooo0`00
000W0?ooo`800000603oool00`000000oooo0?ooo`0I0?ooo`009`3oool2000002<0oooo0P00000K
0?ooo`030000003oool0oooo01P0oooo000Y0?ooo`<000006`3oool5000001d0oooo00<000000?oo
o`3oool0603oool002`0oooo0P00000A0?ooo`P000008P3oool00`000000oooo0?ooo`0H0?ooo`00
;P3oool2000000H0oooo2@00000Z0?ooo`030000003oool0oooo01P0oooo000`0?ooo`H00000<P3o
ool00`000000oooo0?ooo`0I0?ooo`00J03oool00`000000oooo0?ooo`0I0?ooo`00J03oool00`00
0000oooo0?ooo`0I0?ooo`00J03oool00`000000oooo0?ooo`0I0?ooo`00J03oool00`000000oooo
0?ooo`0I0?ooo`00J03oool00`000000oooo0?ooo`0I0?ooo`00J03oool00`000000oooo0?ooo`0I
0?ooo`00=`3oool5000002/0oooo00<000000?ooo`3oool06P3oool00340oooo1P0000050?ooo`@0
00009`3oool00`000000oooo0?ooo`0J0?ooo`00:`3oool6000000l0oooo1@00000R0?ooo`030000
003oool0oooo01X0oooo000U0?ooo`H000006P3oool5000001d0oooo00<000000?ooo`3oool06P3o
ool00280oooo0`00000U0?ooo`@000006@3oool00`000000oooo0?ooo`0J0?ooo`008@3oool00`00
0000oooo0?ooo`0Z0?ooo`D00000503oool00`000000oooo0?ooo`0J0?ooo`00803oool00`000000
oooo0?ooo`0`0?ooo`<000004@3oool00`000000oooo0?ooo`0J0?ooo`00803oool00`000000oooo
0?ooo`0c0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool06`3oool001l0oooo00<0
00000?ooo`3oool0=@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01/0oooo000N
0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0K
0?ooo`007@3oool00`000000oooo0?ooo`0i0?ooo`8000002`3oool00`000000oooo0?ooo`0K0?oo
o`007@3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool0703oool001`0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo01d0oooo000K0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool00P3o
ool200000200oooo000J0?ooo`030000003oool0oooo0440oooo00@000000?ooo`3oool000008P3o
ool001X0oooo00<000000?ooo`3oool0@P3oool2000002<0oooo000I0?ooo`030000003oool0oooo
04<0oooo0P00000S0?ooo`006P3oool00`000000oooo0?ooo`110?ooo`040000003oool0oooo0000
0280oooo000J0?ooo`030000003oool0oooo03l0oooo0P0000040?ooo`030000003oool0oooo01l0
oooo000K0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool01@3oool2000001l0oooo
000L0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`0L0?ooo`007@3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool06`3oool001d0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`0<0?oo
o`030000003oool0oooo01X0oooo000N0?ooo`030000003oool0oooo03D0oooo0P00000@0?ooo`03
0000003oool0oooo01T0oooo000O0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`0I0?ooo`00803oool00`000000oooo0?ooo`0a0?ooo`030000003o
ool0oooo0140oooo00<000000?ooo`3oool06@3oool00200oooo00<000000?ooo`3oool0;P3oool3
000001@0oooo00<000000?ooo`3oool06@3oool00240oooo00<000000?ooo`3oool0:@3oool40000
01L0oooo00<000000?ooo`3oool06@3oool00280oooo0`00000T0?ooo`@000006`3oool00`000000
oooo0?ooo`0I0?ooo`009@3oool6000001T0oooo1@00000P0?ooo`030000003oool0oooo01P0oooo
000[0?ooo`H000003`3oool4000002D0oooo00<000000?ooo`3oool0603oool00340oooo1P000005
0?ooo`@00000:@3oool00`000000oooo0?ooo`0H0?ooo`00=`3oool5000002d0oooo00<000000?oo
o`3oool0603oool006T0oooo00<000000?ooo`3oool0603oool006T0oooo00<000000?ooo`3oool0
603oool006T0oooo00<000000?ooo`3oool0603oool006T0oooo00<000000?ooo`3oool0603oool0
06T0oooo00<000000?ooo`3oool0603oool006T0oooo00<000000?ooo`3oool0603oool006X0oooo
00<000000?ooo`3oool05`3oool006X0oooo00<000000?ooo`3oool05`3oool006X0oooo00<00000
0?ooo`3oool05`3oool00300oooo1P00000d0?ooo`030000003oool0oooo01L0oooo000]0?ooo`<0
00001P3oool9000002/0oooo00<000000?ooo`3oool05`3oool002X0oooo0`00000B0?ooo`X00000
8@3oool00`000000oooo0?ooo`0G0?ooo`009`3oool3000001l0oooo1@00000K0?ooo`030000003o
ool0oooo01P0oooo000T0?ooo`<000009`3oool2000001P0oooo00<000000?ooo`3oool06@3oool0
0240oooo0`00000/0?ooo`8000005@3oool00`000000oooo0?ooo`0J0?ooo`007P3oool300000340
oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0K0?ooo`00703oool2000003D0oooo
0P00000A0?ooo`030000003oool0oooo01/0oooo000L0?ooo`030000003oool0oooo03H0oooo0P00
000>0?ooo`030000003oool0oooo01`0oooo000K0?ooo`030000003oool0oooo03T0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`0M0?ooo`006`3oool00`000000oooo0?ooo`0j0?oo
o`8000002@3oool00`000000oooo0?ooo`0N0?ooo`006`3oool00`000000oooo0?ooo`0l0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool07`3oool001/0oooo00<000000?ooo`3oool0
?@3oool2000000@0oooo00<000000?ooo`3oool0803oool001X0oooo00<000000?ooo`3oool0@03o
ool2000000030?ooo`000000oooo0280oooo000J0?ooo`030000003oool0oooo0480oooo00<00000
0?ooo`3oool08P3oool001X0oooo00<000000?ooo`3oool0@@3oool010000000oooo00000000000R
0?ooo`006P3oool00`000000oooo0?ooo`110?ooo`040000003oool0oooo0?ooo`800000803oool0
01T0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
01d0oooo000I0?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`0M0?ooo`006@3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool0703oool001X0oooo0P00000m0?ooo`8000002P3oool00`000000oooo0?oo
o`0L0?ooo`00703oool00`000000oooo0?ooo`0h0?ooo`8000003@3oool00`000000oooo0?ooo`0K
0?ooo`007@3oool2000003D0oooo0`00000?0?ooo`030000003oool0oooo01/0oooo000O0?ooo`03
0000003oool0oooo0300oooo0P00000B0?ooo`030000003oool0oooo01/0oooo000P0?ooo`800000
;@3oool3000001D0oooo00<000000?ooo`3oool06P3oool00280oooo0P00000Y0?ooo`800000603o
ool00`000000oooo0?ooo`0J0?ooo`00903oool00`000000oooo0?ooo`0S0?ooo`<000006P3oool0
0`000000oooo0?ooo`0J0?ooo`009@3oool200000240oooo0P00000N0?ooo`030000003oool0oooo
01T0oooo000W0?ooo`L000005`3oool300000200oooo00<000000?ooo`3oool06@3oool002h0oooo
3@0000080?ooo`800000903oool00`000000oooo0?ooo`0H0?ooo`00>`3oool8000002H0oooo00<0
00000?ooo`3oool0603oool006T0oooo00<000000?ooo`3oool0603oool006X0oooo00<000000?oo
o`3oool05`3oool006X0oooo00<000000?ooo`3oool05`3oool006X0oooo00<000000?ooo`3oool0
5`3oool006/0oooo00<000000?ooo`3oool05P3oool006/0oooo00<000000?ooo`3oool05P3oool0
06`0oooo00<000000?ooo`3oool05@3oool006`0oooo00<000000?ooo`3oool05@3oool006/0oooo
00<000000?ooo`3oool05P3oool006/0oooo00<000000?ooo`3oool05P3oool006X0oooo00<00000
0?ooo`3oool05`3oool006X0oooo00<000000?ooo`3oool05`3oool006T0oooo00<000000?ooo`3o
ool0603oool002T0oooo3P00000b0?ooo`030000003oool0oooo01P0oooo000W0?ooo`8000003P3o
ool@00000240oooo00<000000?ooo`3oool06@3oool002D0oooo0P00000P0?ooo`8000007`3oool0
0`000000oooo0?ooo`0I0?ooo`008`3oool2000002@0oooo0P00000L0?ooo`030000003oool0oooo
01X0oooo000R0?ooo`030000003oool0oooo02H0oooo0`00000H0?ooo`030000003oool0oooo01/0
oooo000P0?ooo`800000;03oool2000001H0oooo00<000000?ooo`3oool06`3oool001h0oooo0P00
000`0?ooo`8000004`3oool00`000000oooo0?ooo`0L0?ooo`00703oool2000003@0oooo0`00000@
0?ooo`030000003oool0oooo01`0oooo000J0?ooo`800000>@3oool2000000d0oooo00<000000?oo
o`3oool07@3oool001T0oooo00<000000?ooo`3oool0>`3oool3000000X0oooo00<000000?ooo`3o
ool07@3oool001T0oooo00<000000?ooo`3oool0?P3oool2000000L0oooo00<000000?ooo`3oool0
7P3oool001T0oooo00<000000?ooo`3oool0@03oool2000000D0oooo00<000000?ooo`3oool07P3o
ool001T0oooo00<000000?ooo`3oool0@P3oool200000080oooo00<000000?ooo`3oool07`3oool0
01X0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0000000Q0?ooo`006P3oool00`000000
oooo0?ooo`130?ooo`8000008P3oool001X0oooo00<000000?ooo`3oool0@P3oool300000280oooo
000J0?ooo`030000003oool0oooo0400oooo0P0000030?ooo`030000003oool0oooo01l0oooo000J
0?ooo`030000003oool0oooo03h0oooo0P0000060?ooo`030000003oool0oooo01h0oooo000J0?oo
o`030000003oool0oooo03`0oooo0P0000080?ooo`030000003oool0oooo01h0oooo000K0?ooo`03
0000003oool0oooo03T0oooo0P00000;0?ooo`030000003oool0oooo01d0oooo000K0?ooo`030000
003oool0oooo03L0oooo0P00000>0?ooo`030000003oool0oooo01`0oooo000K0?ooo`800000=P3o
ool200000140oooo00<000000?ooo`3oool06`3oool001d0oooo0`00000a0?ooo`8000004`3oool0
0`000000oooo0?ooo`0K0?ooo`00803oool2000002d0oooo0P00000F0?ooo`030000003oool0oooo
01X0oooo000R0?ooo`<00000:03oool2000001T0oooo00<000000?ooo`3oool06@3oool002D0oooo
0`00000S0?ooo`800000703oool00`000000oooo0?ooo`0H0?ooo`00:03oool2000001T0oooo2000
000N0?ooo`030000003oool0oooo01P0oooo000Z0?ooo`<00000203oool>000002L0oooo00<00000
0?ooo`3oool05`3oool002d0oooo2000000f0?ooo`030000003oool0oooo01H0oooo001/0?ooo`03
0000003oool0oooo01D0oooo001/0?ooo`030000003oool0oooo01D0oooo001]0?ooo`030000003o
ool0oooo01@0oooo001]0?ooo`030000003oool0oooo01@0oooo001/0?ooo`030000003oool0oooo
01D0oooo001/0?ooo`030000003oool0oooo01D0oooo001/0?ooo`030000003oool0oooo01D0oooo
001/0?ooo`030000003oool0oooo01D0oooo001[0?ooo`030000003oool0oooo01H0oooo001[0?oo
o`030000003oool0oooo01H0oooo001[0?ooo`030000003oool0oooo01H0oooo001Z0?ooo`030000
003oool0oooo01L0oooo001Z0?ooo`030000003oool0oooo01L0oooo001Z0?ooo`030000003oool0
oooo01L0oooo001Z0?ooo`030000003oool0oooo01L0oooo001Y0?ooo`030000003oool0oooo01P0
oooo000h0?ooo`H00000:`3oool00`000000oooo0?ooo`0H0?ooo`00<03oool8000000H0oooo1000
000W0?ooo`030000003oool0oooo01P0oooo000W0?ooo`T000004P3oool400000280oooo00<00000
0?ooo`3oool06@3oool00280oooo1@00000O0?ooo`<000007`3oool00`000000oooo0?ooo`0I0?oo
o`008@3oool00`000000oooo0?ooo`0U0?ooo`@000006`3oool00`000000oooo0?ooo`0I0?ooo`00
803oool00`000000oooo0?ooo`0Z0?ooo`<00000603oool00`000000oooo0?ooo`0I0?ooo`007P3o
ool200000300oooo1000000C0?ooo`030000003oool0oooo01X0oooo000M0?ooo`030000003oool0
oooo03@0oooo1000000?0?ooo`030000003oool0oooo01X0oooo000L0?ooo`030000003oool0oooo
03T0oooo0P00000;0?ooo`8000007@3oool001/0oooo00<000000?ooo`3oool0?03oool00`000000
oooo0?ooo`070?ooo`030000003oool0oooo01d0oooo000J0?ooo`030000003oool0oooo03h0oooo
00<000000?ooo`3oool0103oool200000200oooo000H0?ooo`800000@P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo0200oooo000G0?ooo`030000003oool0oooo04<0oooo00@00000
0?ooo`00000000008`3oool001H0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0S
0?ooo`005`3oool00`000000oooo0?ooo`130?ooo`030000003oool0000002@0oooo000H0?ooo`03
0000003oool0oooo0400oooo0P0000030?ooo`030000003oool0oooo0240oooo000H0?ooo`030000
003oool0oooo03l0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0P0?ooo`006@3o
ool00`000000oooo0?ooo`0l0?ooo`800000203oool00`000000oooo0?ooo`0O0?ooo`006P3oool0
0`000000oooo0?ooo`0j0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool07P3oool0
01/0oooo00<000000?ooo`3oool0=`3oool2000000d0oooo00<000000?ooo`3oool07@3oool001`0
oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01`0
oooo000L0?ooo`030000003oool0oooo03<0oooo0P00000B0?ooo`030000003oool0oooo01/0oooo
000M0?ooo`030000003oool0oooo02l0oooo0`00000E0?ooo`030000003oool0oooo01X0oooo000N
0?ooo`<00000:03oool6000001T0oooo00<000000?ooo`3oool06@3oool00240oooo1P00000M0?oo
o`D00000803oool00`000000oooo0?ooo`0H0?ooo`009`3oool600000140oooo1P00000V0?ooo`03
0000003oool0oooo01L0oooo000]0?ooo`H000001@3oool6000002d0oooo00<000000?ooo`3oool0
5P3oool003<0oooo1@00000c0?ooo`030000003oool0oooo01H0oooo001[0?ooo`030000003oool0
oooo01H0oooo001[0?ooo`030000003oool0oooo01H0oooo001[0?ooo`030000003oool0oooo01H0
oooo001[0?ooo`030000003oool0oooo01H0oooo001[0?ooo`030000003oool0oooo01H0oooo001[
0?ooo`030000003oool0oooo01H0oooo001[0?ooo`030000003oool0oooo01H0oooo001[0?ooo`03
0000003oool0oooo01H0oooo001[0?ooo`030000003oool0oooo01H0oooo001[0?ooo`030000003o
ool0oooo01H0oooo001[0?ooo`030000003oool0oooo01H0oooo001[0?ooo`030000003oool0oooo
01H0oooo001[0?ooo`030000003oool0oooo01H0oooo001[0?ooo`030000003oool0oooo01H0oooo
001[0?ooo`030000003oool0oooo01H0oooo001[0?ooo`030000003oool0oooo01H0oooo001[0?oo
o`030000003oool0oooo01H0oooo001[0?ooo`030000003oool0oooo01H0oooo001[0?ooo`030000
003oool0oooo01H0oooo000a0?ooo`D00000=@3oool00`000000oooo0?ooo`0F0?ooo`00;03oool5
000000D0oooo1@00000_0?ooo`030000003oool0oooo01L0oooo000X0?ooo`@000003`3oool50000
02T0oooo00<000000?ooo`3oool0603oool002<0oooo1@00000H0?ooo`D000008`3oool00`000000
oooo0?ooo`0I0?ooo`007P3oool500000280oooo1@00000M0?ooo`030000003oool0oooo01X0oooo
000K0?ooo`<00000;03oool5000001L0oooo00<000000?ooo`3oool06`3oool001/0oooo00<00000
0?ooo`3oool0<@3oool3000001<0oooo00<000000?ooo`3oool0703oool001X0oooo00<000000?oo
o`3oool0=@3oool200000100oooo00<000000?ooo`3oool07@3oool001X0oooo00<000000?ooo`3o
ool0=`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01h0oooo000I0?ooo`030000
003oool0oooo03T0oooo0P00000:0?ooo`8000008@3oool001T0oooo00<000000?ooo`3oool0>`3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0240oooo000H0?ooo`030000003oool0
oooo03d0oooo0P0000050?ooo`030000003oool0oooo0280oooo000H0?ooo`030000003oool0oooo
03l0oooo00D000000?ooo`3oool0oooo0000000U0?ooo`005`3oool00`000000oooo0?ooo`110?oo
o`<000009P3oool001L0oooo00<000000?ooo`3oool0@P3oool3000002D0oooo000F0?ooo`030000
003oool0oooo0480oooo00D000000?ooo`3oool0oooo0000000T0?ooo`005P3oool00`000000oooo
0?ooo`110?ooo`030000003oool0oooo00<0oooo0P00000R0?ooo`005`3oool2000003l0oooo0P00
00080?ooo`030000003oool0oooo01l0oooo000I0?ooo`800000>@3oool4000000/0oooo0P00000O
0?ooo`006`3oool00`000000oooo0?ooo`0c0?ooo`<000004@3oool00`000000oooo0?ooo`0L0?oo
o`00703oool200000300oooo0`00000E0?ooo`800000703oool001h0oooo00<000000?ooo`3oool0
:@3oool4000001X0oooo00<000000?ooo`3oool06@3oool001l0oooo0P00000V0?ooo`<000007P3o
ool00`000000oooo0?ooo`0I0?ooo`008@3oool200000240oooo0`00000R0?ooo`030000003oool0
oooo01P0oooo000S0?ooo`h000003`3oool4000002D0oooo00<000000?ooo`3oool0603oool00340
oooo3`00000Y0?ooo`030000003oool0oooo01P0oooo001Y0?ooo`030000003oool0oooo01P0oooo
001Z0?ooo`030000003oool0oooo01L0oooo001Z0?ooo`030000003oool0oooo01L0oooo001Z0?oo
o`030000003oool0oooo01L0oooo001[0?ooo`030000003oool0oooo01H0oooo001[0?ooo`030000
003oool0oooo01H0oooo001[0?ooo`030000003oool0oooo01H0oooo001/0?ooo`030000003oool0
oooo01D0oooo001/0?ooo`030000003oool0oooo01D0oooo001/0?ooo`030000003oool0oooo01D0
oooo001/0?ooo`030000003oool0oooo01D0oooo001]0?ooo`030000003oool0oooo01@0oooo001]
0?ooo`030000003oool0oooo01@0oooo001]0?ooo`030000003oool0oooo01@0oooo001^0?ooo`03
0000003oool0oooo01<0oooo001^0?ooo`030000003oool0oooo01<0oooo001]0?ooo`030000003o
ool0oooo01@0oooo001]0?ooo`030000003oool0oooo01@0oooo001/0?ooo`030000003oool0oooo
01D0oooo001[0?ooo`030000003oool0oooo01H0oooo001[0?ooo`030000003oool0oooo01H0oooo
001Z0?ooo`030000003oool0oooo01L0oooo001Y0?ooo`030000003oool0oooo01P0oooo001Y0?oo
o`030000003oool0oooo01P0oooo001X0?ooo`030000003oool0oooo01T0oooo000Y0?ooo`T00000
=@3oool00`000000oooo0?ooo`0J0?ooo`009P3oool3000000T0oooo3`00000U0?ooo`030000003o
ool0oooo01/0oooo000S0?ooo`<000006`3oool9000001`0oooo00<000000?ooo`3oool06`3oool0
01l0oooo1000000W0?ooo`8000006@3oool00`000000oooo0?ooo`0L0?ooo`00703oool3000002d0
oooo0P00000F0?ooo`030000003oool0oooo01d0oooo000I0?ooo`<00000<P3oool2000001@0oooo
00<000000?ooo`3oool07@3oool001L0oooo0P00000g0?ooo`8000004@3oool00`000000oooo0?oo
o`0N0?ooo`005`3oool00`000000oooo0?ooo`0h0?ooo`8000003P3oool00`000000oooo0?ooo`0O
0?ooo`005`3oool00`000000oooo0?ooo`0j0?ooo`800000303oool00`000000oooo0?ooo`0O0?oo
o`005`3oool00`000000oooo0?ooo`0l0?ooo`<000001`3oool200000280oooo000G0?ooo`030000
003oool0oooo03l0oooo0P0000030?ooo`800000903oool001L0oooo00<000000?ooo`3oool0@@3o
ool3000002H0oooo000G0?ooo`030000003oool0oooo0400oooo0P0000000`3oool000000000000U
0?ooo`005`3oool00`000000oooo0?ooo`0m0?ooo`<000001@3oool2000002<0oooo000G0?ooo`03
0000003oool0oooo03/0oooo0P00000:0?ooo`8000008@3oool001L0oooo00<000000?ooo`3oool0
>@3oool2000000h0oooo0P00000O0?ooo`005`3oool2000003P0oooo0P00000a0?ooo`006@3oool3
00000380oooo0`00000c0?ooo`00703oool3000002d0oooo0P00000f0?ooo`007`3oool4000002L0
oooo0P00000h0?ooo`008`3oool300000280oooo0P00000j0?ooo`009P3oool300000100oooo3`00
000l0?ooo`00:@3oool@000004/0oooo00240?ooo`00Q03oool008@0oooo00240?ooo`00Q03oool0
08@0oooo00240?ooo`00Q03oool008@0oooo00240?ooo`00Q03oool008@0oooo00240?ooo`00Q03o
ool008@0oooo00240?ooo`00Q03oool008@0oooo00240?ooo`00Q03oool008@0oooo00240?ooo`00
Q03oool008@0oooo00240?ooo`00Q03oool008@0oooo00240?ooo`00Q03oool008@0oooo00240?oo
o`00Q03oool008@0oooo00240?ooo`00Q03oool008@0oooo00240?ooo`00Q03oool008@0oooo0024
0?ooo`00Q03oool008@0oooo00240?ooo`00Q03oool008@0oooo00240?ooo`00Q03oool008@0oooo
00240?ooo`00Q03oool00001\
\>"],
  ImageRangeCache->{{{0, 131.25}, {430.875, 0}} -> {0.353347, 0.0496501, \
0.00331721, 0.00331721}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["SphericalPlot3D", "Subsection"],

Cell[TextData[{
  "The package ",
  StyleBox["Graphics`ParametricPlot3D`", "InlineInput",
    FontWeight->"Bold"],
  " provides extensions to the built-in version of ",
  StyleBox["ParametricPlot3D", "InlineInput",
    FontWeight->"Bold"],
  ".  Among these is a function ",
  StyleBox["SphericalPlot3D", "InlineInput",
    FontWeight->"Bold"],
  " which displays ",
  Cell[BoxData[
      \(TraditionalForm\`f[\[Theta], \[Phi]]\)]],
  " as a parametric function of polar angle, \[Theta], and azimuthal angle, \
\[Phi], by interpreting the function as the radial coordinate.  Because the \
radius must be positive, one often wraps the target function within ",
  StyleBox["Abs", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[TextData[{
  "For example, below we create functions which plot individual spherical \
harmonics and then assemble a graphics array for ",
  StyleBox["{m,0,\[Lambda]}", "InlineInput"],
  "."
}], "Text"],

Cell[BoxData[
    \(plotYlm[\[Lambda]_, m_]\  := 
      SphericalPlot3D[\ 
        Evaluate[
          Abs[SphericalHarmonicY[\[Lambda], 
              m, \[Theta], \[Phi]]]], {\[Theta], 0, \[Pi]}, {\[Phi], 0, 
          2  \[Pi]}, Boxed \[Rule] False, Axes \[Rule] False, 
        DisplayFunction \[Rule] Identity]\)], "Input"],

Cell[BoxData[
    \(tablePlotYlm[\[Lambda]_]\  := \ 
      Table[plotYlm[\[Lambda], m], {m, 0, \[Lambda]}]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[GraphicsArray[tablePlotYlm[2]]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .5121 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.31746 0.0121928 0.31746 [
[ 0 0 0 0 ]
[ 1 .5121 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .5121 L
0 .5121 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0121928 0.321429 0.499904 MathSubStart
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.312503 1.65106 -0.01235 1.65106 [
[ 0 0 0 0 ]
[ 1 1.63871 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.63871 L
0 1.63871 L
closepath
clip
newpath
.5 Mabswid
[ ] 0 setdash
.728 .644 .765 r
.49266 .38098 m .47549 .44627 L .50054 .44972 L p F P
0 g
s
.728 .644 .765 r
.50054 .44972 m .50601 .38286 L .49266 .38098 L p F P
0 g
s
.74 .598 .699 r
.4964 .52457 m .50054 .44972 L .47549 .44627 L p F P
0 g
s
.704 .623 .764 r
.50601 .38286 m .50054 .44972 L .5261 .44957 L p F P
0 g
s
.706 .568 .698 r
.50054 .44972 m .4964 .52457 L .53099 .52438 L p F P
0 g
s
.706 .568 .698 r
.53099 .52438 m .5261 .44957 L .50054 .44972 L p F P
0 g
s
.704 .623 .764 r
.5261 .44957 m .51965 .38277 L .50601 .38286 L p F P
0 g
s
.678 .604 .766 r
.51965 .38277 m .5261 .44957 L .55106 .44583 L p F P
0 g
s
.67 .541 .699 r
.5261 .44957 m .53099 .52438 L .56477 .51941 L p F P
0 g
s
.67 .541 .699 r
.56477 .51941 m .55106 .44583 L .5261 .44957 L p F P
0 g
s
.74 .598 .699 r
.47549 .44627 m .46249 .51999 L .4964 .52457 L p F P
0 g
s
.744 .55 .631 r
.49403 .60107 m .4964 .52457 L .46249 .51999 L p F P
0 g
s
.702 .514 .631 r
.4964 .52457 m .49403 .60107 L .53377 .60084 L p F P
0 g
s
.702 .514 .631 r
.53377 .60084 m .53099 .52438 L .4964 .52457 L p F P
0 g
s
.657 .478 .631 r
.53099 .52438 m .53377 .60084 L .57261 .59525 L p F P
0 g
s
.657 .478 .631 r
.57261 .59525 m .56477 .51941 L .53099 .52438 L p F P
0 g
s
.751 .668 .77 r
.4802 .37723 m .45206 .43935 L .47549 .44627 L p F P
0 g
s
.751 .668 .77 r
.47549 .44627 m .49266 .38098 L .4802 .37723 L p F P
0 g
s
.773 .63 .703 r
.46249 .51999 m .47549 .44627 L .45206 .43935 L p F P
0 g
s
.678 .604 .766 r
.55106 .44583 m .53295 .38074 L .51965 .38277 L p F P
0 g
s
.651 .587 .77 r
.53295 .38074 m .55106 .44583 L .57431 .43864 L p F P
0 g
s
.631 .515 .703 r
.55106 .44583 m .56477 .51941 L .59629 .50986 L p F P
0 g
s
.631 .515 .703 r
.59629 .50986 m .57431 .43864 L .55106 .44583 L p F P
0 g
s
.773 .63 .703 r
.45206 .43935 m .43073 .51081 L .46249 .51999 L p F P
0 g
s
.786 .588 .632 r
.45506 .59591 m .46249 .51999 L .43073 .51081 L p F P
0 g
s
.744 .55 .631 r
.46249 .51999 m .45506 .59591 L .49403 .60107 L p F P
0 g
s
.607 .442 .633 r
.56477 .51941 m .57261 .59525 L .60887 .5845 L p F P
0 g
s
.607 .442 .633 r
.60887 .5845 m .59629 .50986 L .56477 .51941 L p F P
0 g
s
.773 .696 .778 r
.46919 .37178 m .4313 .42926 L .45206 .43935 L p F P
0 g
s
.773 .696 .778 r
.45206 .43935 m .4802 .37723 L .46919 .37178 L p F P
0 g
s
.806 .667 .709 r
.43073 .51081 m .45206 .43935 L .4313 .42926 L p F P
0 g
s
.651 .587 .77 r
.57431 .43864 m .54531 .37685 L .53295 .38074 L p F P
0 g
s
.622 .573 .779 r
.54531 .37685 m .57431 .43864 L .5948 .4283 L p F P
0 g
s
.587 .489 .71 r
.57431 .43864 m .59629 .50986 L .62416 .49611 L p F P
0 g
s
.587 .489 .71 r
.62416 .49611 m .5948 .4283 L .57431 .43864 L p F P
0 g
s
.701 .697 .843 r
.50601 .38286 m .51226 .33017 L .49266 .38098 L closepath p F P
0 g
s
.69 .687 .842 r
.51965 .38277 m .51226 .33017 L .50601 .38286 L closepath p F P
0 g
s
.806 .667 .709 r
.4313 .42926 m .4025 .49739 L .43073 .51081 L p F P
0 g
s
.829 .631 .634 r
.41851 .58557 m .43073 .51081 L .4025 .49739 L p F P
0 g
s
.786 .588 .632 r
.43073 .51081 m .41851 .58557 L .45506 .59591 L p F P
0 g
s
.679 .68 .843 r
.53295 .38074 m .51226 .33017 L .51965 .38277 L closepath p F P
0 g
s
.547 .401 .634 r
.59629 .50986 m .60887 .5845 L .64099 .569 L p F P
0 g
s
.547 .401 .634 r
.64099 .569 m .62416 .49611 L .59629 .50986 L p F P
0 g
s
.71 .708 .846 r
.49266 .38098 m .51226 .33017 L .4802 .37723 L closepath p F P
0 g
s
.668 .673 .846 r
.54531 .37685 m .51226 .33017 L .53295 .38074 L closepath p F P
0 g
s
.794 .726 .789 r
.46015 .36485 m .41415 .41642 L .4313 .42926 L p F P
0 g
s
.794 .726 .789 r
.4313 .42926 m .46919 .37178 L .46015 .36485 L p F P
0 g
s
.84 .71 .719 r
.4025 .49739 m .4313 .42926 L .41415 .41642 L p F P
0 g
s
.622 .573 .779 r
.5948 .4283 m .55617 .37126 L .54531 .37685 L p F P
0 g
s
.59 .562 .79 r
.55617 .37126 m .5948 .4283 L .61161 .41526 L p F P
0 g
s
.536 .463 .72 r
.5948 .4283 m .62416 .49611 L .64711 .47873 L p F P
0 g
s
.536 .463 .72 r
.64711 .47873 m .61161 .41526 L .5948 .4283 L p F P
0 g
s
.718 .719 .85 r
.4802 .37723 m .51226 .33017 L .46919 .37178 L closepath p F P
0 g
s
.656 .669 .851 r
.55617 .37126 m .51226 .33017 L .54531 .37685 L closepath p F P
0 g
s
.742 .492 .552 r
.49374 .67298 m .49403 .60107 L .45506 .59591 L p F P
0 g
s
.691 .45 .553 r
.49403 .60107 m .49374 .67298 L .5341 .67276 L p F P
0 g
s
.691 .45 .553 r
.5341 .67276 m .53377 .60084 L .49403 .60107 L p F P
0 g
s
.636 .406 .552 r
.53377 .60084 m .5341 .67276 L .57353 .66719 L p F P
0 g
s
.636 .406 .552 r
.57353 .66719 m .57261 .59525 L .53377 .60084 L p F P
0 g
s
.724 .732 .857 r
.46919 .37178 m .51226 .33017 L .46015 .36485 L closepath p F P
0 g
s
.644 .667 .857 r
.56502 .36423 m .51226 .33017 L .55617 .37126 L closepath p F P
0 g
s
.59 .562 .79 r
.61161 .41526 m .56502 .36423 L .55617 .37126 L p F P
0 g
s
.84 .71 .719 r
.41415 .41642 m .37909 .48028 L .4025 .49739 L p F P
0 g
s
.876 .681 .636 r
.38599 .57045 m .4025 .49739 L .37909 .48028 L p F P
0 g
s
.829 .631 .634 r
.4025 .49739 m .38599 .57045 L .41851 .58557 L p F P
0 g
s
.472 .353 .636 r
.62416 .49611 m .64099 .569 L .6675 .54938 L p F P
0 g
s
.472 .353 .636 r
.6675 .54938 m .64711 .47873 L .62416 .49611 L p F P
0 g
s
.792 .534 .548 r
.45416 .66785 m .45506 .59591 L .41851 .58557 L p F P
0 g
s
.742 .492 .552 r
.45506 .59591 m .45416 .66785 L .49374 .67298 L p F P
0 g
s
.573 .356 .547 r
.57261 .59525 m .57353 .66719 L .61036 .65649 L p F P
0 g
s
.573 .356 .547 r
.61036 .65649 m .60887 .5845 L .57261 .59525 L p F P
0 g
s
.812 .761 .805 r
.4535 .35677 m .40142 .40138 L .41415 .41642 L p F P
0 g
s
.812 .761 .805 r
.41415 .41642 m .46015 .36485 L .4535 .35677 L p F P
0 g
s
.875 .76 .733 r
.37909 .48028 m .41415 .41642 L .40142 .40138 L p F P
0 g
s
.555 .553 .806 r
.56502 .36423 m .61161 .41526 L .62394 .40007 L p F P
0 g
s
.472 .434 .734 r
.61161 .41526 m .64711 .47873 L .66406 .45842 L p F P
0 g
s
.472 .434 .734 r
.66406 .45842 m .62394 .40007 L .61161 .41526 L p F P
0 g
s
.729 .745 .865 r
.46015 .36485 m .51226 .33017 L .4535 .35677 L closepath p F P
0 g
s
.632 .667 .865 r
.57146 .35606 m .51226 .33017 L .56502 .36423 L closepath p F P
0 g
s
.555 .553 .806 r
.62394 .40007 m .57146 .35606 L .56502 .36423 L p F P
0 g
s
.731 .758 .874 r
.4535 .35677 m .51226 .33017 L .44956 .3479 L closepath p F P
0 g
s
.827 .8 .825 r
.40142 .40138 m .4535 .35677 L .44956 .3479 L p F P
0 g
s
.844 .58 .539 r
.41705 .65756 m .41851 .58557 L .38599 .57045 L p F P
0 g
s
.792 .534 .548 r
.41851 .58557 m .41705 .65756 L .45416 .66785 L p F P
0 g
s
.622 .67 .875 r
.57516 .34714 m .51226 .33017 L .57146 .35606 L closepath p F P
0 g
s
.516 .547 .827 r
.57146 .35606 m .62394 .40007 L .63116 .3834 L p F P
0 g
s
.516 .547 .827 r
.63116 .3834 m .57516 .34714 L .57146 .35606 L p F P
0 g
s
.494 .295 .538 r
.60887 .5845 m .61036 .65649 L .64298 .64107 L p F P
0 g
s
.494 .295 .538 r
.64298 .64107 m .64099 .569 L .60887 .5845 L p F P
0 g
s
.875 .76 .733 r
.40142 .40138 m .36161 .46019 L .37909 .48028 L p F P
0 g
s
.925 .742 .635 r
.35895 .55114 m .37909 .48028 L .36161 .46019 L p F P
0 g
s
.876 .681 .636 r
.37909 .48028 m .35895 .55114 L .38599 .57045 L p F P
0 g
s
.827 .8 .825 r
.44956 .3479 m .39375 .38481 L .40142 .40138 L p F P
0 g
s
.908 .821 .75 r
.36161 .46019 m .40142 .40138 L .39375 .38481 L p F P
0 g
s
.388 .401 .751 r
.62394 .40007 m .66406 .45842 L .67414 .43608 L p F P
0 g
s
.388 .401 .751 r
.67414 .43608 m .63116 .3834 L .62394 .40007 L p F P
0 g
s
.369 .29 .635 r
.64711 .47873 m .6675 .54938 L .68716 .52644 L p F P
0 g
s
.369 .29 .635 r
.68716 .52644 m .66406 .45842 L .64711 .47873 L p F P
0 g
s
.73 .771 .885 r
.44956 .3479 m .51226 .33017 L .44854 .33864 L closepath p F P
0 g
s
.836 .842 .85 r
.39375 .38481 m .44956 .3479 L .44854 .33864 L p F P
0 g
s
.596 .706 .921 r
.56083 .31228 m .51226 .33017 L .56861 .32002 L closepath p F P
0 g
s
.597 .721 .931 r
.55072 .30585 m .51226 .33017 L .56083 .31228 L closepath p F P
0 g
s
.603 .737 .939 r
.53877 .30104 m .51226 .33017 L .55072 .30585 L closepath p F P
0 g
s
.612 .753 .945 r
.52555 .2981 m .51226 .33017 L .53877 .30104 L closepath p F P
0 g
s
.677 .797 .939 r
.47287 .30633 m .51226 .33017 L .48469 .30137 L closepath p F P
0 g
s
.694 .799 .93 r
.46295 .31288 m .51226 .33017 L .47287 .30633 L closepath p F P
0 g
s
.708 .797 .92 r
.45537 .32071 m .51226 .33017 L .46295 .31288 L closepath p F P
0 g
s
.719 .791 .908 r
.4505 .32943 m .51226 .33017 L .45537 .32071 L closepath p F P
0 g
s
.726 .782 .896 r
.44854 .33864 m .51226 .33017 L .4505 .32943 L closepath p F P
0 g
s
.612 .675 .886 r
.57593 .33788 m .51226 .33017 L .57516 .34714 L closepath p F P
0 g
s
.604 .683 .897 r
.57372 .32869 m .51226 .33017 L .57593 .33788 L closepath p F P
0 g
s
.599 .694 .909 r
.56861 .32002 m .51226 .33017 L .57372 .32869 L closepath p F P
0 g
s
.625 .768 .948 r
.51168 .29715 m .51226 .33017 L .52555 .2981 L closepath p F P
0 g
s
.642 .781 .948 r
.49784 .29826 m .51226 .33017 L .51168 .29715 L closepath p F P
0 g
s
.659 .791 .945 r
.48469 .30137 m .51226 .33017 L .49784 .29826 L closepath p F P
0 g
s
.47 .546 .852 r
.57516 .34714 m .63116 .3834 L .63288 .366 L p F P
0 g
s
.47 .546 .852 r
.63288 .366 m .57593 .33788 L .57516 .34714 L p F P
0 g
s
.9 .632 .522 r
.38402 .64251 m .38599 .57045 L .35895 .55114 L p F P
0 g
s
.844 .58 .539 r
.38599 .57045 m .38402 .64251 L .41705 .65756 L p F P
0 g
s
.386 .213 .52 r
.64099 .569 m .64298 .64107 L .66993 .62154 L p F P
0 g
s
.386 .213 .52 r
.66993 .62154 m .6675 .54938 L .64099 .569 L p F P
0 g
s
.834 .885 .878 r
.39156 .36745 m .44854 .33864 L .4505 .32943 L p F P
0 g
s
.836 .842 .85 r
.44854 .33864 m .39156 .36745 L .39375 .38481 L p F P
0 g
s
.419 .55 .881 r
.57593 .33788 m .63288 .366 L .62894 .34869 L p F P
0 g
s
.419 .55 .881 r
.62894 .34869 m .57372 .32869 L .57593 .33788 L p F P
0 g
s
.932 .892 .768 r
.35093 .43798 m .39375 .38481 L .39156 .36745 L p F P
0 g
s
.908 .821 .75 r
.39375 .38481 m .35093 .43798 L .36161 .46019 L p F P
0 g
s
.272 .358 .769 r
.63116 .3834 m .67414 .43608 L .67675 .41268 L p F P
0 g
s
.272 .358 .769 r
.67675 .41268 m .63288 .366 L .63116 .3834 L p F P
0 g
s
.973 .815 .626 r
.33868 .52843 m .36161 .46019 L .35093 .43798 L p F P
0 g
s
.925 .742 .635 r
.36161 .46019 m .33868 .52843 L .35895 .55114 L p F P
0 g
s
.814 .925 .909 r
.39502 .35009 m .4505 .32943 L .45537 .32071 L p F P
0 g
s
.834 .885 .878 r
.4505 .32943 m .39502 .35009 L .39156 .36745 L p F P
0 g
s
.361 .56 .911 r
.57372 .32869 m .62894 .34869 L .61946 .33228 L p F P
0 g
s
.361 .56 .911 r
.61946 .33228 m .56861 .32002 L .57372 .32869 L p F P
0 g
s
.213 .198 .624 r
.66406 .45842 m .68716 .52644 L .69894 .50113 L p F P
0 g
s
.213 .198 .624 r
.69894 .50113 m .67414 .43608 L .66406 .45842 L p F P
0 g
s
.728 .411 .443 r
.49569 .73451 m .49374 .67298 L .45416 .66785 L p F P
0 g
s
.668 .363 .447 r
.49374 .67298 m .49569 .73451 L .53177 .73432 L p F P
0 g
s
.668 .363 .447 r
.53177 .73432 m .5341 .67276 L .49374 .67298 L p F P
0 g
s
.602 .308 .442 r
.5341 .67276 m .53177 .73432 L .56701 .72943 L p F P
0 g
s
.602 .308 .442 r
.56701 .72943 m .57353 .66719 L .5341 .67276 L p F P
0 g
s
.786 .455 .429 r
.46032 .73001 m .45416 .66785 L .41705 .65756 L p F P
0 g
s
.728 .411 .443 r
.45416 .66785 m .46032 .73001 L .49569 .73451 L p F P
0 g
s
.521 .239 .427 r
.57353 .66719 m .56701 .72943 L .59991 .72004 L p F P
0 g
s
.521 .239 .427 r
.59991 .72004 m .61036 .65649 L .57353 .66719 L p F P
0 g
s
.767 .952 .935 r
.40405 .33358 m .45537 .32071 L .46295 .31288 L p F P
0 g
s
.814 .925 .909 r
.45537 .32071 m .40405 .33358 L .39502 .35009 L p F P
0 g
s
.3 .576 .937 r
.56861 .32002 m .61946 .33228 L .60482 .31758 L p F P
0 g
s
.3 .576 .937 r
.60482 .31758 m .56083 .31228 L .56861 .32002 L p F P
0 g
s
.958 .69 .486 r
.35654 .62329 m .35895 .55114 L .33868 .52843 L p F P
0 g
s
.9 .632 .522 r
.35895 .55114 m .35654 .62329 L .38402 .64251 L p F P
0 g
s
.222 .091 .481 r
.6675 .54938 m .66993 .62154 L .68991 .59869 L p F P
0 g
s
.222 .091 .481 r
.68991 .59869 m .68716 .52644 L .6675 .54938 L p F P
0 g
s
.926 .963 .776 r
.34767 .41463 m .39156 .36745 L .39502 .35009 L p F P
0 g
s
.932 .892 .768 r
.39156 .36745 m .34767 .41463 L .35093 .43798 L p F P
0 g
s
.103 .299 .776 r
.63288 .366 m .67675 .41268 L .67164 .38932 L p F P
0 g
s
.103 .299 .776 r
.67164 .38932 m .62894 .34869 L .63288 .366 L p F P
0 g
s
.685 .955 .951 r
.41828 .31871 m .46295 .31288 L .47287 .30633 L p F P
0 g
s
.767 .952 .935 r
.46295 .31288 m .41828 .31871 L .40405 .33358 L p F P
0 g
s
.243 .6 .951 r
.56083 .31228 m .60482 .31758 L .58567 .30532 L p F P
0 g
s
.243 .6 .951 r
.58567 .30532 m .55072 .30585 L .56083 .31228 L p F P
0 g
s
.844 .498 .4 r
.42717 .72097 m .41705 .65756 L .38402 .64251 L p F P
0 g
s
.786 .455 .429 r
.41705 .65756 m .42717 .72097 L .46032 .73001 L p F P
0 g
s
.413 .146 .396 r
.61036 .65649 m .59991 .72004 L .629 .70651 L p F P
0 g
s
.413 .146 .396 r
.629 .70651 m .64298 .64107 L .61036 .65649 L p F P
0 g
s
.991 .887 .585 r
.32621 .50328 m .35093 .43798 L .34767 .41463 L p F P
0 g
s
.973 .815 .626 r
.35093 .43798 m .32621 .50328 L .33868 .52843 L p F P
0 g
s
.571 .925 .951 r
.43709 .30623 m .47287 .30633 L .48469 .30137 L p F P
0 g
s
.685 .955 .951 r
.47287 .30633 m .43709 .30623 L .41828 .31871 L p F P
0 g
s
.204 .63 .95 r
.55072 .30585 m .58567 .30532 L .56292 .29613 L p F P
0 g
s
.204 .63 .95 r
.56292 .29613 m .53877 .30104 L .55072 .30585 L p F P
0 g
s
0 .053 .578 r
.67414 .43608 m .69894 .50113 L .70214 .4746 L p F P
0 g
s
0 .053 .578 r
.70214 .4746 m .67675 .41268 L .67414 .43608 L p F P
0 g
s
.442 .866 .939 r
.45958 .29676 m .48469 .30137 L .49784 .29826 L p F P
0 g
s
.571 .925 .951 r
.48469 .30137 m .45958 .29676 L .43709 .30623 L p F P
0 g
s
.2 .671 .938 r
.53877 .30104 m .56292 .29613 L .53766 .29049 L p F P
0 g
s
.2 .671 .938 r
.53766 .29049 m .52555 .2981 L .53877 .30104 L p F P
0 g
s
.836 .993 .741 r
.35213 .39122 m .39502 .35009 L .40405 .33358 L p F P
0 g
s
.926 .963 .776 r
.39502 .35009 m .35213 .39122 L .34767 .41463 L p F P
0 g
s
.324 .794 .928 r
.48467 .29081 m .49784 .29826 L .51168 .29715 L p F P
0 g
s
.442 .866 .939 r
.49784 .29826 m .48467 .29081 L .45958 .29676 L p F P
0 g
s
.24 .726 .927 r
.52555 .2981 m .53766 .29049 L .51114 .28868 L p F P
0 g
s
.24 .726 .927 r
.51114 .28868 m .51168 .29715 L .52555 .2981 L p F P
0 g
s
0 .209 .734 r
.62894 .34869 m .67164 .38932 L .65892 .36711 L p F P
0 g
s
0 .209 .734 r
.65892 .36711 m .61946 .33228 L .62894 .34869 L p F P
0 g
s
.324 .794 .928 r
.51168 .29715 m .51114 .28868 L .48467 .29081 L p F P
0 g
s
.993 .743 .402 r
.33594 .60068 m .33868 .52843 L .32621 .50328 L p F P
0 g
s
.958 .69 .486 r
.33868 .52843 m .33594 .60068 L .35654 .62329 L p F P
0 g
s
.901 .538 .344 r
.3977 .70777 m .38402 .64251 L .35654 .62329 L p F P
0 g
s
.844 .498 .4 r
.38402 .64251 m .3977 .70777 L .42717 .72097 L p F P
0 g
s
0 0 .391 r
.68716 .52644 m .68991 .59869 L .70189 .5735 L p F P
0 g
s
0 0 .391 r
.70189 .5735 m .69894 .50113 L .68716 .52644 L p F P
0 g
s
.253 .009 .337 r
.64298 .64107 m .629 .70651 L .65298 .6894 L p F P
0 g
s
.253 .009 .337 r
.65298 .6894 m .66993 .62154 L .64298 .64107 L p F P
0 g
s
.876 .882 .448 r
.32228 .4768 m .34767 .41463 L .35213 .39122 L p F P
0 g
s
.991 .887 .585 r
.34767 .41463 m .32228 .4768 L .32621 .50328 L p F P
0 g
s
.595 .896 .611 r
.36423 .36887 m .40405 .33358 L .41828 .31871 L p F P
0 g
s
.836 .993 .741 r
.40405 .33358 m .36423 .36887 L .35213 .39122 L p F P
0 g
s
0 0 .429 r
.67675 .41268 m .70214 .4746 L .69642 .44805 L p F P
0 g
s
0 0 .429 r
.69642 .44805 m .67164 .38932 L .67675 .41268 L p F P
0 g
s
0 .091 .595 r
.61946 .33228 m .65892 .36711 L .63907 .34715 L p F P
0 g
s
0 .091 .595 r
.63907 .34715 m .60482 .31758 L .61946 .33228 L p F P
0 g
s
.937 .564 .231 r
.37324 .69093 m .35654 .62329 L .33594 .60068 L p F P
0 g
s
.901 .538 .344 r
.35654 .62329 m .37324 .69093 L .3977 .70777 L p F P
0 g
s
0 0 .217 r
.66993 .62154 m .65298 .6894 L .67071 .6694 L p F P
0 g
s
0 0 .217 r
.67071 .6694 m .68991 .59869 L .66993 .62154 L p F P
0 g
s
.896 .714 .205 r
.32325 .57564 m .32621 .50328 L .32228 .4768 L p F P
0 g
s
.993 .743 .402 r
.32621 .50328 m .32325 .57564 L .33594 .60068 L p F P
0 g
s
.595 .896 .611 r
.41828 .31871 m .38352 .34869 L .36423 .36887 L p F P
0 g
s
0 0 0 r
.38352 .34869 m .41828 .31871 L .43709 .30623 L p F P
0 g
s
0 0 .181 r
.69894 .50113 m .70189 .5735 L .70516 .54707 L p F P
0 g
s
0 0 .181 r
.70516 .54707 m .70214 .4746 L .69894 .50113 L p F P
0 g
s
.57 .008 0 r
.60482 .31758 m .63907 .34715 L .61295 .33047 L p F P
0 g
s
.57 .008 0 r
.61295 .33047 m .58567 .30532 L .60482 .31758 L p F P
0 g
s
.686 .279 .269 r
.49986 .78059 m .49569 .73451 L .46032 .73001 L p F P
0 g
s
.617 .225 .279 r
.49569 .73451 m .49986 .78059 L .52682 .78045 L p F P
0 g
s
.617 .225 .279 r
.52682 .78045 m .53177 .73432 L .49569 .73451 L p F P
0 g
s
.533 .154 .268 r
.53177 .73432 m .52682 .78045 L .55314 .77684 L p F P
0 g
s
.533 .154 .268 r
.55314 .77684 m .56701 .72943 L .53177 .73432 L p F P
0 g
s
.748 .32 .235 r
.47344 .77727 m .46032 .73001 L .42717 .72097 L p F P
0 g
s
.686 .279 .269 r
.46032 .73001 m .47344 .77727 L .49986 .78059 L p F P
0 g
s
.424 .055 .231 r
.56701 .72943 m .55314 .77684 L .57767 .76993 L p F P
0 g
s
.424 .055 .231 r
.57767 .76993 m .59991 .72004 L .56701 .72943 L p F P
0 g
s
.876 .882 .448 r
.35213 .39122 m .32724 .45021 L .32228 .4768 L p F P
0 g
s
0 0 0 r
.32724 .45021 m .35213 .39122 L .36423 .36887 L p F P
0 g
s
.776 .387 0 r
.67164 .38932 m .69642 .44805 L .68187 .42276 L p F P
0 g
s
.776 .387 0 r
.68187 .42276 m .65892 .36711 L .67164 .38932 L p F P
0 g
s
.738 .615 .721 r
.42534 .93353 m .41769 .97103 L .48213 .97865 L p F P
0 g
s
.738 .615 .721 r
.48213 .97865 m .48456 .94062 L .42534 .93353 L p F P
0 g
s
.741 .493 .554 r
.482 1.00601 m .48213 .97865 L .41769 .97103 L p F P
0 g
s
.706 .587 .72 r
.48456 .94062 m .48213 .97865 L .54776 .97832 L p F P
0 g
s
.691 .452 .555 r
.48213 .97865 m .482 1.00601 L .5479 1.00569 L p F P
0 g
s
.691 .452 .555 r
.5479 1.00569 m .54776 .97832 L .48213 .97865 L p F P
0 g
s
.706 .587 .72 r
.54776 .97832 m .54489 .94031 L .48456 .94062 L p F P
0 g
s
.672 .561 .721 r
.54489 .94031 m .54776 .97832 L .61198 .97005 L p F P
0 g
s
.638 .409 .554 r
.54776 .97832 m .5479 1.00569 L .6124 .99745 L p F P
0 g
s
.638 .409 .554 r
.6124 .99745 m .61198 .97005 L .54776 .97832 L p F P
0 g
s
.056 0 0 r
.40915 .3317 m .43709 .30623 L .45958 .29676 L p F P
0 g
s
0 0 0 r
.43709 .30623 m .40915 .3317 L .38352 .34869 L p F P
0 g
s
.714 .685 .821 r
.43953 .89245 m .42534 .93353 L .48456 .94062 L p F P
0 g
s
.714 .685 .821 r
.48456 .94062 m .48907 .89846 L .43953 .89245 L p F P
0 g
s
.696 .668 .819 r
.48907 .89846 m .48456 .94062 L .54489 .94031 L p F P
0 g
s
.696 .668 .819 r
.54489 .94031 m .53956 .8982 L .48907 .89846 L p F P
0 g
s
.676 .654 .821 r
.53956 .8982 m .54489 .94031 L .60391 .93262 L p F P
0 g
s
.672 .561 .721 r
.61198 .97005 m .60391 .93262 L .54489 .94031 L p F P
0 g
s
.618 .052 0 r
.58567 .30532 m .61295 .33047 L .5818 .31794 L p F P
0 g
s
.618 .052 0 r
.5818 .31794 m .56292 .29613 L .58567 .30532 L p F P
0 g
s
.799 .346 .166 r
.44872 .77062 m .42717 .72097 L .3977 .70777 L p F P
0 g
s
.748 .32 .235 r
.42717 .72097 m .44872 .77062 L .47344 .77727 L p F P
0 g
s
.267 0 .157 r
.59991 .72004 m .57767 .76993 L .5993 .75999 L p F P
0 g
s
.267 0 .157 r
.5993 .75999 m .629 .70651 L .59991 .72004 L p F P
0 g
s
.864 .518 .01 r
.35495 .67114 m .33594 .60068 L .32325 .57564 L p F P
0 g
s
.937 .564 .231 r
.33594 .60068 m .35495 .67114 L .37324 .69093 L p F P
0 g
s
.769 .646 .726 r
.36962 .91928 m .357 .95569 L .41769 .97103 L p F P
0 g
s
.769 .646 .726 r
.41769 .97103 m .42534 .93353 L .36962 .91928 L p F P
0 g
s
.79 .534 .55 r
.41728 .99842 m .41769 .97103 L .357 .95569 L p F P
0 g
s
.741 .493 .554 r
.41769 .97103 m .41728 .99842 L .482 1.00601 L p F P
0 g
s
0 0 0 r
.68991 .59869 m .67071 .6694 L .68127 .64739 L p F P
0 g
s
0 0 0 r
.68127 .64739 m .70189 .5735 L .68991 .59869 L p F P
0 g
s
.635 .537 .726 r
.60391 .93262 m .61198 .97005 L .67224 .9541 L p F P
0 g
s
.577 .36 .55 r
.61198 .97005 m .6124 .99745 L .67292 .98156 L p F P
0 g
s
.577 .36 .55 r
.67292 .98156 m .67224 .9541 L .61198 .97005 L p F P
0 g
s
.731 .704 .825 r
.393 .88039 m .36962 .91928 L .42534 .93353 L p F P
0 g
s
.731 .704 .825 r
.42534 .93353 m .43953 .89245 L .393 .88039 L p F P
0 g
s
.272 0 0 r
.43993 .3188 m .45958 .29676 L .48467 .29081 L p F P
0 g
s
.056 0 0 r
.45958 .29676 m .43993 .3188 L .40915 .3317 L p F P
0 g
s
.676 .654 .821 r
.60391 .93262 m .58892 .89168 L .53956 .8982 L p F P
0 g
s
.655 .643 .826 r
.58892 .89168 m .60391 .93262 L .65923 .91779 L p F P
0 g
s
.635 .537 .726 r
.67224 .9541 m .65923 .91779 L .60391 .93262 L p F P
0 g
s
.59 .045 0 r
.56292 .29613 m .5818 .31794 L .54714 .31023 L p F P
0 g
s
.59 .045 0 r
.54714 .31023 m .53766 .29049 L .56292 .29613 L p F P
0 g
s
.896 .714 .205 r
.32228 .4768 m .31923 .54926 L .32325 .57564 L p F P
0 g
s
0 0 .108 r
.31923 .54926 m .32228 .4768 L .32724 .45021 L p F P
0 g
s
.576 .063 0 r
.48422 1.01679 m .482 1.00601 L .41728 .99842 L p F P
0 g
s
.503 .007 .013 r
.482 1.00601 m .48422 1.01679 L .54527 1.01649 L p F P
0 g
s
.503 .007 .013 r
.54527 1.01649 m .5479 1.00569 L .482 1.00601 L p F P
0 g
s
.402 0 0 r
.5479 1.00569 m .54527 1.01649 L .60499 1.00889 L p F P
0 g
s
.402 0 0 r
.60499 1.00889 m .6124 .99745 L .5479 1.00569 L p F P
0 g
s
.419 0 0 r
.47436 .31067 m .48467 .29081 L .51114 .28868 L p F P
0 g
s
.272 0 0 r
.48467 .29081 m .47436 .31067 L .43993 .3188 L p F P
0 g
s
.521 0 0 r
.53766 .29049 m .54714 .31023 L .51071 .30776 L p F P
0 g
s
.521 0 0 r
.51071 .30776 m .51114 .28868 L .53766 .29049 L p F P
0 g
s
.879 .658 .133 r
.70214 .4746 m .70516 .54707 L .69937 .52061 L p F P
0 g
s
.879 .658 .133 r
.69937 .52061 m .69642 .44805 L .70214 .4746 L p F P
0 g
s
.419 0 0 r
.51114 .28868 m .51071 .30776 L .47436 .31067 L p F P
0 g
s
0 0 .09 r
.34107 .42476 m .36423 .36887 L .38352 .34869 L p F P
0 g
s
0 0 0 r
.36423 .36887 m .34107 .42476 L .32724 .45021 L p F P
0 g
s
.823 .345 .035 r
.4268 .76091 m .3977 .70777 L .37324 .69093 L p F P
0 g
s
.799 .346 .166 r
.3977 .70777 m .4268 .76091 L .44872 .77062 L p F P
0 g
s
.891 .467 .107 r
.65892 .36711 m .68187 .42276 L .65903 .39999 L p F P
0 g
s
.891 .467 .107 r
.65903 .39999 m .63907 .34715 L .65892 .36711 L p F P
0 g
s
.023 0 .02 r
.629 .70651 m .5993 .75999 L .61705 .74744 L p F P
0 g
s
.023 0 .02 r
.61705 .74744 m .65298 .6894 L .629 .70651 L p F P
0 g
s
.667 .742 .906 r
.4591 .85453 m .43953 .89245 L .48907 .89846 L p F P
0 g
s
.667 .742 .906 r
.48907 .89846 m .49529 .85897 L .4591 .85453 L p F P
0 g
s
.665 .74 .906 r
.49529 .85897 m .48907 .89846 L .53956 .8982 L p F P
0 g
s
.665 .74 .906 r
.53956 .8982 m .5322 .85878 L .49529 .85897 L p F P
0 g
s
.662 .738 .906 r
.5322 .85878 m .53956 .8982 L .58892 .89168 L p F P
0 g
s
.79 .534 .55 r
.357 .95569 m .35633 .98315 L .41728 .99842 L p F P
0 g
s
.625 .088 0 r
.42429 1.00979 m .41728 .99842 L .35633 .98315 L p F P
0 g
s
.576 .063 0 r
.41728 .99842 m .42429 1.00979 L .48422 1.01679 L p F P
0 g
s
.261 0 0 r
.6124 .99745 m .60499 1.00889 L .66097 .99424 L p F P
0 g
s
.261 0 0 r
.66097 .99424 m .67292 .98156 L .6124 .99745 L p F P
0 g
s
.746 .725 .832 r
.35145 .86271 m .31972 .89835 L .36962 .91928 L p F P
0 g
s
.746 .725 .832 r
.36962 .91928 m .393 .88039 L .35145 .86271 L p F P
0 g
s
.801 .682 .733 r
.357 .95569 m .36962 .91928 L .31972 .89835 L p F P
0 g
s
.801 .682 .733 r
.31972 .89835 m .30257 .93316 L .357 .95569 L p F P
0 g
s
.841 .579 .543 r
.35633 .98315 m .357 .95569 L .30257 .93316 L p F P
0 g
s
.655 .643 .826 r
.65923 .91779 m .63511 .87913 L .58892 .89168 L p F P
0 g
s
.633 .634 .833 r
.63511 .87913 m .65923 .91779 L .70853 .89635 L p F P
0 g
s
.594 .514 .734 r
.65923 .91779 m .67224 .9541 L .72603 .931 L p F P
0 g
s
.594 .514 .734 r
.72603 .931 m .70853 .89635 L .65923 .91779 L p F P
0 g
s
.501 .302 .542 r
.67224 .9541 m .67292 .98156 L .72695 .95856 L p F P
0 g
s
.501 .302 .542 r
.72695 .95856 m .72603 .931 L .67224 .9541 L p F P
0 g
s
.669 .744 .907 r
.42518 .84562 m .393 .88039 L .43953 .89245 L p F P
0 g
s
.669 .744 .907 r
.43953 .89245 m .4591 .85453 L .42518 .84562 L p F P
0 g
s
.662 .738 .906 r
.58892 .89168 m .56826 .85396 L .5322 .85878 L p F P
0 g
s
.659 .736 .907 r
.56826 .85396 m .58892 .89168 L .63511 .87913 L p F P
0 g
s
.864 .518 .01 r
.32325 .57564 m .34376 .64926 L .35495 .67114 L p F P
0 g
s
0 0 .298 r
.34376 .64926 m .32325 .57564 L .31923 .54926 L p F P
0 g
s
.845 .809 .322 r
.70189 .5735 m .68127 .64739 L .68405 .62433 L p F P
0 g
s
.845 .809 .322 r
.68405 .62433 m .70516 .54707 L .70189 .5735 L p F P
0 g
s
.753 .281 0 r
.40869 .74856 m .37324 .69093 L .35495 .67114 L p F P
0 g
s
.823 .345 .035 r
.37324 .69093 m .40869 .74856 L .4268 .76091 L p F P
0 g
s
.221 0 .24 r
.36325 .40174 m .38352 .34869 L .40915 .3317 L p F P
0 g
s
0 0 .09 r
.38352 .34869 m .36325 .40174 L .34107 .42476 L p F P
0 g
s
0 0 0 r
.65298 .6894 m .61705 .74744 L .63008 .73281 L p F P
0 g
s
0 0 0 r
.63008 .73281 m .67071 .6694 L .65298 .6894 L p F P
0 g
s
.87 .461 .249 r
.63907 .34715 m .65903 .39999 L .62888 .38093 L p F P
0 g
s
.87 .461 .249 r
.62888 .38093 m .61295 .33047 L .63907 .34715 L p F P
0 g
s
0 0 .338 r
.32426 .52276 m .32724 .45021 L .34107 .42476 L p F P
0 g
s
0 0 .108 r
.32724 .45021 m .32426 .52276 L .31923 .54926 L p F P
0 g
s
.841 .579 .543 r
.30257 .93316 m .30167 .96071 L .35633 .98315 L p F P
0 g
s
.639 .079 0 r
.3679 .9957 m .35633 .98315 L .30167 .96071 L p F P
0 g
s
.625 .088 0 r
.35633 .98315 m .3679 .9957 L .42429 1.00979 L p F P
0 g
s
.984 .7 .351 r
.69642 .44805 m .69937 .52061 L .68461 .49541 L p F P
0 g
s
.984 .7 .351 r
.68461 .49541 m .68187 .42276 L .69642 .44805 L p F P
0 g
s
.671 .747 .908 r
.39502 .83259 m .35145 .86271 L .393 .88039 L p F P
0 g
s
.671 .747 .908 r
.393 .88039 m .42518 .84562 L .39502 .83259 L p F P
0 g
s
.055 0 0 r
.67292 .98156 m .66097 .99424 L .71088 .97304 L p F P
0 g
s
.055 0 0 r
.71088 .97304 m .72695 .95856 L .67292 .98156 L p F P
0 g
s
.659 .736 .907 r
.63511 .87913 m .60192 .84469 L .56826 .85396 L p F P
0 g
s
.656 .736 .908 r
.60192 .84469 m .63511 .87913 L .67615 .86102 L p F P
0 g
s
.633 .634 .833 r
.70853 .89635 m .67615 .86102 L .63511 .87913 L p F P
0 g
s
.387 .772 .99 r
.48866 1.00626 m .48422 1.01679 L .42429 1.00979 L p F P
0 g
s
.425 .81 .99 r
.48422 1.01679 m .48866 1.00626 L .54001 1.006 L p F P
0 g
s
.425 .81 .99 r
.54001 1.006 m .54527 1.01649 L .48422 1.01679 L p F P
0 g
s
.482 .848 .99 r
.54527 1.01649 m .54001 1.006 L .59021 .99959 L p F P
0 g
s
.482 .848 .99 r
.59021 .99959 m .60499 1.00889 L .54527 1.01649 L p F P
0 g
s
.759 .748 .843 r
.31674 .84007 m .27784 .8715 L .31972 .89835 L p F P
0 g
s
.759 .748 .843 r
.31972 .89835 m .35145 .86271 L .31674 .84007 L p F P
0 g
s
.832 .723 .745 r
.30257 .93316 m .31972 .89835 L .27784 .8715 L p F P
0 g
s
.832 .723 .745 r
.27784 .8715 m .25677 .90421 L .30257 .93316 L p F P
0 g
s
.896 .63 .527 r
.30167 .96071 m .30257 .93316 L .25677 .90421 L p F P
0 g
s
.61 .628 .844 r
.67615 .86102 m .70853 .89635 L .74965 .86906 L p F P
0 g
s
.546 .491 .746 r
.70853 .89635 m .72603 .931 L .771 .90156 L p F P
0 g
s
.546 .491 .746 r
.771 .90156 m .74965 .86906 L .70853 .89635 L p F P
0 g
s
.371 .738 .99 r
.43829 1.00035 m .42429 1.00979 L .3679 .9957 L p F P
0 g
s
.387 .772 .99 r
.42429 1.00979 m .43829 1.00035 L .48866 1.00626 L p F P
0 g
s
.398 .223 .525 r
.72603 .931 m .72695 .95856 L .77212 .92925 L p F P
0 g
s
.398 .223 .525 r
.77212 .92925 m .771 .90156 L .72603 .931 L p F P
0 g
s
.549 .881 .99 r
.60499 1.00889 m .59021 .99959 L .63719 .98724 L p F P
0 g
s
.549 .881 .99 r
.63719 .98724 m .66097 .99424 L .60499 1.00889 L p F P
0 g
s
.4 .088 .317 r
.39283 .38234 m .40915 .3317 L .43993 .3188 L p F P
0 g
s
.221 0 .24 r
.40915 .3317 m .39283 .38234 L .36325 .40174 L p F P
0 g
s
.813 .426 .322 r
.61295 .33047 m .62888 .38093 L .59284 .3666 L p F P
0 g
s
.813 .426 .322 r
.59284 .3666 m .5818 .31794 L .61295 .33047 L p F P
0 g
s
.753 .281 0 r
.35495 .67114 m .39524 .73408 L .40869 .74856 L p F P
0 g
s
0 0 .464 r
.39524 .73408 m .35495 .67114 L .34376 .64926 L p F P
0 g
s
0 0 .5 r
.34032 .62624 m .31923 .54926 L .32426 .52276 L p F P
0 g
s
0 0 .298 r
.31923 .54926 m .34032 .62624 L .34376 .64926 L p F P
0 g
s
.745 .887 .485 r
.67071 .6694 m .63008 .73281 L .63774 .71676 L p F P
0 g
s
.745 .887 .485 r
.63774 .71676 m .68127 .64739 L .67071 .6694 L p F P
0 g
s
.995 .852 .51 r
.70516 .54707 m .68405 .62433 L .67876 .60128 L p F P
0 g
s
.995 .852 .51 r
.67876 .60128 m .69937 .52061 L .70516 .54707 L p F P
0 g
s
.672 .75 .909 r
.36998 .81597 m .31674 .84007 L .35145 .86271 L p F P
0 g
s
.672 .75 .909 r
.35145 .86271 m .39502 .83259 L .36998 .81597 L p F P
0 g
s
.559 .031 0 r
.50601 .80738 m .49986 .78059 L .47344 .77727 L p F P
0 g
s
.481 0 0 r
.49986 .78059 m .50601 .80738 L .51953 .8073 L p F P
0 g
s
.481 0 0 r
.51953 .8073 m .52682 .78045 L .49986 .78059 L p F P
0 g
s
.372 0 0 r
.52682 .78045 m .51953 .8073 L .53271 .80551 L p F P
0 g
s
.372 0 0 r
.53271 .80551 m .55314 .77684 L .52682 .78045 L p F P
0 g
s
.564 .785 .982 r
.48245 .82629 m .4591 .85453 L .49529 .85897 L p F P
0 g
s
.564 .785 .982 r
.49529 .85897 m .50272 .8288 L .48245 .82629 L p F P
0 g
s
.583 .804 .984 r
.50272 .8288 m .49529 .85897 L .5322 .85878 L p F P
0 g
s
.583 .804 .984 r
.5322 .85878 m .52342 .82869 L .50272 .8288 L p F P
0 g
s
.607 .82 .982 r
.52342 .82869 m .5322 .85878 L .56826 .85396 L p F P
0 g
s
.656 .736 .908 r
.67615 .86102 m .63171 .83135 L .60192 .84469 L p F P
0 g
s
.653 .735 .91 r
.63171 .83135 m .67615 .86102 L .71021 .83801 L p F P
0 g
s
.61 .628 .844 r
.74965 .86906 m .71021 .83801 L .67615 .86102 L p F P
0 g
s
.225 .065 .443 r
.33828 .49741 m .34107 .42476 L .36325 .40174 L p F P
0 g
s
0 0 .338 r
.34107 .42476 m .33828 .49741 L .32426 .52276 L p F P
0 g
s
.607 .057 0 r
.49277 .80572 m .47344 .77727 L .44872 .77062 L p F P
0 g
s
.559 .031 0 r
.47344 .77727 m .49277 .80572 L .50601 .80738 L p F P
0 g
s
.52 .195 .356 r
.42844 .36758 m .43993 .3188 L .47436 .31067 L p F P
0 g
s
.4 .088 .317 r
.43993 .3188 m .42844 .36758 L .39283 .38234 L p F P
0 g
s
.216 0 0 r
.55314 .77684 m .53271 .80551 L .54497 .80208 L p F P
0 g
s
.216 0 0 r
.54497 .80208 m .57767 .76993 L .55314 .77684 L p F P
0 g
s
.896 .63 .527 r
.25677 .90421 m .25566 .93188 L .30167 .96071 L p F P
0 g
s
.582 .021 0 r
.3174 .97502 m .30167 .96071 L .25566 .93188 L p F P
0 g
s
.639 .079 0 r
.30167 .96071 m .3174 .97502 L .3679 .9957 L p F P
0 g
s
.948 .654 .449 r
.68187 .42276 m .68461 .49541 L .66141 .47272 L p F P
0 g
s
.948 .654 .449 r
.66141 .47272 m .65903 .39999 L .68187 .42276 L p F P
0 g
s
.75 .383 .358 r
.5818 .31794 m .59284 .3666 L .55269 .35777 L p F P
0 g
s
.75 .383 .358 r
.55269 .35777 m .54714 .31023 L .5818 .31794 L p F P
0 g
s
.551 .765 .977 r
.4635 .82126 m .42518 .84562 L .4591 .85453 L p F P
0 g
s
.551 .765 .977 r
.4591 .85453 m .48245 .82629 L .4635 .82126 L p F P
0 g
s
.607 .82 .982 r
.56826 .85396 m .54361 .82596 L .52342 .82869 L p F P
0 g
s
.634 .832 .977 r
.54361 .82596 m .56826 .85396 L .60192 .84469 L p F P
0 g
s
.375 .708 .986 r
.39096 .98848 m .3679 .9957 L .3174 .97502 L p F P
0 g
s
.371 .738 .99 r
.3679 .9957 m .39096 .98848 L .43829 1.00035 L p F P
0 g
s
0 0 0 r
.72695 .95856 m .71088 .97304 L .75251 .94606 L p F P
0 g
s
0 0 0 r
.75251 .94606 m .77212 .92925 L .72695 .95856 L p F P
0 g
s
.618 .903 .985 r
.66097 .99424 m .63719 .98724 L .67894 .96942 L p F P
0 g
s
.618 .903 .985 r
.67894 .96942 m .71088 .97304 L .66097 .99424 L p F P
0 g
s
.611 .042 0 r
.48042 .80242 m .44872 .77062 L .4268 .76091 L p F P
0 g
s
.607 .057 0 r
.44872 .77062 m .48042 .80242 L .49277 .80572 L p F P
0 g
s
0 0 0 r
.57767 .76993 m .54497 .80208 L .55573 .79716 L p F P
0 g
s
0 0 0 r
.55573 .79716 m .5993 .75999 L .57767 .76993 L p F P
0 g
s
.61 .272 .373 r
.46832 .35828 m .47436 .31067 L .51071 .30776 L p F P
0 g
s
.52 .195 .356 r
.47436 .31067 m .46832 .35828 L .42844 .36758 L p F P
0 g
s
.684 .332 .373 r
.54714 .31023 m .55269 .35777 L .51046 .35494 L p F P
0 g
s
.684 .332 .373 r
.51046 .35494 m .51071 .30776 L .54714 .31023 L p F P
0 g
s
.61 .272 .373 r
.51071 .30776 m .51046 .35494 L .46832 .35828 L p F P
0 g
s
.544 .745 .969 r
.44673 .81393 m .39502 .83259 L .42518 .84562 L p F P
0 g
s
.544 .745 .969 r
.42518 .84562 m .4635 .82126 L .44673 .81393 L p F P
0 g
s
.769 .773 .856 r
.29051 .81337 m .24597 .83974 L .27784 .8715 L p F P
0 g
s
.769 .773 .856 r
.27784 .8715 m .31674 .84007 L .29051 .81337 L p F P
0 g
s
.862 .772 .76 r
.25677 .90421 m .27784 .8715 L .24597 .83974 L p F P
0 g
s
.634 .832 .977 r
.60192 .84469 m .56241 .82074 L .54361 .82596 L p F P
0 g
s
.66 .838 .968 r
.56241 .82074 m .60192 .84469 L .63171 .83135 L p F P
0 g
s
0 .086 .636 r
.38711 .71812 m .34376 .64926 L .34032 .62624 L p F P
0 g
s
0 0 .464 r
.34376 .64926 m .38711 .71812 L .39524 .73408 L p F P
0 g
s
.586 .625 .857 r
.71021 .83801 m .74965 .86906 L .7806 .83693 L p F P
0 g
s
.487 .468 .762 r
.74965 .86906 m .771 .90156 L .80499 .86687 L p F P
0 g
s
.487 .468 .762 r
.80499 .86687 m .7806 .83693 L .74965 .86906 L p F P
0 g
s
.945 .952 .644 r
.68127 .64739 m .63774 .71676 L .63959 .7 L p F P
0 g
s
.945 .952 .644 r
.63959 .7 m .68405 .62433 L .68127 .64739 L p F P
0 g
s
.862 .772 .76 r
.24597 .83974 m .22179 .8699 L .25677 .90421 L p F P
0 g
s
.953 .687 .495 r
.25566 .93188 m .25677 .90421 L .22179 .8699 L p F P
0 g
s
.532 0 0 r
.46951 .79762 m .4268 .76091 L .40869 .74856 L p F P
0 g
s
.611 .042 0 r
.4268 .76091 m .46951 .79762 L .48042 .80242 L p F P
0 g
s
.642 .761 .935 r
.49498 .97146 m .48866 1.00626 L .43829 1.00035 L p F P
0 g
s
.645 .764 .935 r
.48866 1.00626 m .49498 .97146 L .53254 .97127 L p F P
0 g
s
.645 .764 .935 r
.53254 .97127 m .54001 1.006 L .48866 1.00626 L p F P
0 g
s
.648 .766 .935 r
.54001 1.006 m .53254 .97127 L .56925 .96653 L p F P
0 g
s
.648 .766 .935 r
.56925 .96653 m .59021 .99959 L .54001 1.006 L p F P
0 g
s
0 0 0 r
.5993 .75999 m .55573 .79716 L .56451 .79096 L p F P
0 g
s
0 0 0 r
.56451 .79096 m .61705 .74744 L .5993 .75999 L p F P
0 g
s
.188 .145 .575 r
.34493 .60316 m .32426 .52276 L .33828 .49741 L p F P
0 g
s
0 0 .5 r
.32426 .52276 m .34493 .60316 L .34032 .62624 L p F P
0 g
s
.243 .109 .491 r
.771 .90156 m .77212 .92925 L .80627 .89469 L p F P
0 g
s
.243 .109 .491 r
.80627 .89469 m .80499 .86687 L .771 .90156 L p F P
0 g
s
.672 .753 .911 r
.35124 .79644 m .29051 .81337 L .31674 .84007 L p F P
0 g
s
.672 .753 .911 r
.31674 .84007 m .36998 .81597 L .35124 .79644 L p F P
0 g
s
.978 .788 .578 r
.69937 .52061 m .67876 .60128 L .66551 .57937 L p F P
0 g
s
.978 .788 .578 r
.66551 .57937 m .68461 .49541 L .69937 .52061 L p F P
0 g
s
.396 .197 .49 r
.36081 .47447 m .36325 .40174 L .39283 .38234 L p F P
0 g
s
.225 .065 .443 r
.36325 .40174 m .36081 .47447 L .33828 .49741 L p F P
0 g
s
.887 .598 .492 r
.65903 .39999 m .66141 .47272 L .63078 .45372 L p F P
0 g
s
.887 .598 .492 r
.63078 .45372 m .62888 .38093 L .65903 .39999 L p F P
0 g
s
.653 .735 .91 r
.71021 .83801 m .65626 .81446 L .63171 .83135 L p F P
0 g
s
.651 .736 .911 r
.65626 .81446 m .71021 .83801 L .73567 .81102 L p F P
0 g
s
.586 .625 .857 r
.7806 .83693 m .73567 .81102 L .71021 .83801 L p F P
0 g
s
.64 .758 .934 r
.45814 .96709 m .43829 1.00035 L .39096 .98848 L p F P
0 g
s
.642 .761 .935 r
.43829 1.00035 m .45814 .96709 L .49498 .97146 L p F P
0 g
s
.651 .767 .934 r
.59021 .99959 m .56925 .96653 L .60351 .95742 L p F P
0 g
s
.651 .767 .934 r
.60351 .95742 m .63719 .98724 L .59021 .99959 L p F P
0 g
s
.544 .726 .958 r
.43291 .80461 m .36998 .81597 L .39502 .83259 L p F P
0 g
s
.544 .726 .958 r
.39502 .83259 m .44673 .81393 L .43291 .80461 L p F P
0 g
s
.582 .021 0 r
.25566 .93188 m .27499 .94848 L .3174 .97502 L p F P
0 g
s
.394 .683 .976 r
.34869 .97108 m .3174 .97502 L .27499 .94848 L p F P
0 g
s
.375 .708 .986 r
.3174 .97502 m .34869 .97108 L .39096 .98848 L p F P
0 g
s
.66 .838 .968 r
.63171 .83135 m .57896 .81323 L .56241 .82074 L p F P
0 g
s
.684 .839 .957 r
.57896 .81323 m .63171 .83135 L .65626 .81446 L p F P
0 g
s
.678 .911 .975 r
.71088 .97304 m .67894 .96942 L .71361 .94678 L p F P
0 g
s
.678 .911 .975 r
.71361 .94678 m .75251 .94606 L .71088 .97304 L p F P
0 g
s
.532 0 0 r
.40869 .74856 m .46055 .79152 L .46951 .79762 L p F P
0 g
s
0 .138 .647 r
.46055 .79152 m .40869 .74856 L .39524 .73408 L p F P
0 g
s
.648 .931 .661 r
.61705 .74744 m .56451 .79096 L .57089 .78378 L p F P
0 g
s
.648 .931 .661 r
.57089 .78378 m .63008 .73281 L .61705 .74744 L p F P
0 g
s
.953 .687 .495 r
.22179 .8699 m .22051 .89771 L .25566 .93188 L p F P
0 g
s
0 .092 .589 r
.27499 .94848 m .25566 .93188 L .22051 .89771 L p F P
0 g
s
.573 .886 .606 r
.77212 .92925 m .75251 .94606 L .78388 .9143 L p F P
0 g
s
.573 .886 .606 r
.78388 .9143 m .80627 .89469 L .77212 .92925 L p F P
0 g
s
.151 .227 .688 r
.38476 .70139 m .34032 .62624 L .34493 .60316 L p F P
0 g
s
0 .086 .636 r
.34032 .62624 m .38476 .70139 L .38711 .71812 L p F P
0 g
s
.638 .755 .933 r
.42362 .95833 m .39096 .98848 L .34869 .97108 L p F P
0 g
s
.64 .758 .934 r
.39096 .98848 m .42362 .95833 L .45814 .96709 L p F P
0 g
s
.971 .891 .69 r
.68405 .62433 m .63959 .7 L .63546 .68331 L p F P
0 g
s
.971 .891 .69 r
.63546 .68331 m .67876 .60128 L .68405 .62433 L p F P
0 g
s
.655 .768 .933 r
.63719 .98724 m .60351 .95742 L .63383 .94429 L p F P
0 g
s
.655 .768 .933 r
.63383 .94429 m .67894 .96942 L .63719 .98724 L p F P
0 g
s
.508 .286 .512 r
.39087 .45512 m .39283 .38234 L .42844 .36758 L p F P
0 g
s
.396 .197 .49 r
.39283 .38234 m .39087 .45512 L .36081 .47447 L p F P
0 g
s
.827 .546 .513 r
.62888 .38093 m .63078 .45372 L .59416 .43943 L p F P
0 g
s
.827 .546 .513 r
.59416 .43943 m .59284 .3666 L .62888 .38093 L p F P
0 g
s
.55 .71 .946 r
.4227 .79372 m .35124 .79644 L .36998 .81597 L p F P
0 g
s
.55 .71 .946 r
.36998 .81597 m .43291 .80461 L .4227 .79372 L p F P
0 g
s
.362 .258 .6 r
.35753 .5811 m .33828 .49741 L .36081 .47447 L p F P
0 g
s
.188 .145 .575 r
.33828 .49741 m .35753 .5811 L .34493 .60316 L p F P
0 g
s
.684 .839 .957 r
.65626 .81446 m .5925 .80377 L .57896 .81323 L p F P
0 g
s
.704 .834 .945 r
.5925 .80377 m .65626 .81446 L .67443 .79473 L p F P
0 g
s
.651 .736 .911 r
.73567 .81102 m .67443 .79473 L .65626 .81446 L p F P
0 g
s
0 .254 .766 r
.45395 .7844 m .39524 .73408 L .38711 .71812 L p F P
0 g
s
0 .138 .647 r
.39524 .73408 m .45395 .7844 L .46055 .79152 L p F P
0 g
s
.925 .716 .601 r
.68461 .49541 m .66551 .57937 L .64479 .55966 L p F P
0 g
s
.925 .716 .601 r
.64479 .55966 m .66141 .47272 L .68461 .49541 L p F P
0 g
s
.852 .997 .771 r
.63008 .73281 m .57089 .78378 L .57455 .77592 L p F P
0 g
s
.852 .997 .771 r
.57455 .77592 m .63774 .71676 L .63008 .73281 L p F P
0 g
s
.775 .8 .873 r
.27413 .78369 m .22579 .80431 L .24597 .83974 L p F P
0 g
s
.775 .8 .873 r
.24597 .83974 m .29051 .81337 L .27413 .78369 L p F P
0 g
s
.89 .829 .78 r
.22179 .8699 m .24597 .83974 L .22579 .80431 L p F P
0 g
s
.672 .755 .913 r
.33976 .77484 m .27413 .78369 L .29051 .81337 L p F P
0 g
s
.672 .755 .913 r
.29051 .81337 m .35124 .79644 L .33976 .77484 L p F P
0 g
s
.56 .626 .874 r
.73567 .81102 m .7806 .83693 L .79973 .80127 L p F P
0 g
s
.412 .443 .782 r
.7806 .83693 m .80499 .86687 L .82617 .82827 L p F P
0 g
s
.412 .443 .782 r
.82617 .82827 m .79973 .80127 L .7806 .83693 L p F P
0 g
s
.648 .736 .914 r
.67443 .79473 m .73567 .81102 L .75117 .78115 L p F P
0 g
s
.56 .626 .874 r
.79973 .80127 m .75117 .78115 L .73567 .81102 L p F P
0 g
s
.59 .351 .523 r
.42705 .44041 m .42844 .36758 L .46832 .35828 L p F P
0 g
s
.508 .286 .512 r
.42844 .36758 m .42705 .44041 L .39087 .45512 L p F P
0 g
s
.3 .741 .965 r
.50761 .81366 m .48245 .82629 L .50272 .8288 L p F P
0 g
s
.3 .741 .965 r
.50272 .8288 m .51074 .81405 L .50761 .81366 L p F P
0 g
s
.348 .788 .962 r
.51074 .81405 m .50272 .8288 L .52342 .82869 L p F P
0 g
s
.348 .788 .962 r
.52342 .82869 m .51393 .81404 L .51074 .81405 L p F P
0 g
s
.423 .839 .965 r
.51393 .81404 m .52342 .82869 L .54361 .82596 L p F P
0 g
s
.771 .499 .524 r
.59284 .3666 m .59416 .43943 L .55335 .43063 L p F P
0 g
s
.771 .499 .524 r
.55335 .43063 m .55269 .35777 L .59284 .3666 L p F P
0 g
s
.89 .829 .78 r
.22579 .80431 m .19947 .83156 L .22179 .8699 L p F P
0 g
s
.993 .743 .42 r
.22051 .89771 m .22179 .8699 L .19947 .83156 L p F P
0 g
s
0 .092 .589 r
.22051 .89771 m .2427 .91707 L .27499 .94848 L p F P
0 g
s
.423 .662 .96 r
.31336 .9488 m .27499 .94848 L .2427 .91707 L p F P
0 g
s
.394 .683 .976 r
.27499 .94848 m .31336 .9488 L .34869 .97108 L p F P
0 g
s
.284 .702 .971 r
.5047 .81289 m .4635 .82126 L .48245 .82629 L p F P
0 g
s
.284 .702 .971 r
.48245 .82629 m .50761 .81366 L .5047 .81289 L p F P
0 g
s
.423 .839 .965 r
.54361 .82596 m .51704 .81361 L .51393 .81404 L p F P
0 g
s
.514 .887 .971 r
.51704 .81361 m .54361 .82596 L .56241 .82074 L p F P
0 g
s
.638 .752 .931 r
.39291 .94551 m .34869 .97108 L .31336 .9488 L p F P
0 g
s
.638 .755 .933 r
.34869 .97108 m .39291 .94551 L .42362 .95833 L p F P
0 g
s
0 0 .41 r
.80499 .86687 m .80627 .89469 L .82756 .85625 L p F P
0 g
s
0 0 .41 r
.82756 .85625 m .82617 .82827 L .80499 .86687 L p F P
0 g
s
.724 .905 .959 r
.75251 .94606 m .71361 .94678 L .73953 .9202 L p F P
0 g
s
.724 .905 .959 r
.73953 .9202 m .78388 .9143 L .75251 .94606 L p F P
0 g
s
.658 .769 .931 r
.67894 .96942 m .63383 .94429 L .65884 .92768 L p F P
0 g
s
.658 .769 .931 r
.65884 .92768 m .71361 .94678 L .67894 .96942 L p F P
0 g
s
.698 .713 .86 r
.50258 .91194 m .49498 .97146 L .45814 .96709 L p F P
0 g
s
.685 .701 .859 r
.49498 .97146 m .50258 .91194 L .52355 .91183 L p F P
0 g
s
.685 .701 .859 r
.52355 .91183 m .53254 .97127 L .49498 .97146 L p F P
0 g
s
.326 .315 .694 r
.38839 .68466 m .34493 .60316 L .35753 .5811 L p F P
0 g
s
.151 .227 .688 r
.34493 .60316 m .38839 .68466 L .38476 .70139 L p F P
0 g
s
.671 .691 .86 r
.53254 .97127 m .52355 .91183 L .54402 .90914 L p F P
0 g
s
.671 .691 .86 r
.54402 .90914 m .56925 .96653 L .53254 .97127 L p F P
0 g
s
.326 0 0 r
.50761 .81366 m .49277 .80572 L .48042 .80242 L p F P
0 g
s
.326 0 0 r
.48042 .80242 m .5047 .81289 L .50761 .81366 L p F P
0 g
s
0 .302 .654 r
.5047 .81289 m .48042 .80242 L .46951 .79762 L p F P
0 g
s
0 .302 .654 r
.46951 .79762 m .50213 .81176 L .5047 .81289 L p F P
0 g
s
.296 .671 .974 r
.50213 .81176 m .44673 .81393 L .4635 .82126 L p F P
0 g
s
.296 .671 .974 r
.4635 .82126 m .5047 .81289 L .50213 .81176 L p F P
0 g
s
.153 .337 .795 r
.45005 .77658 m .38711 .71812 L .38476 .70139 L p F P
0 g
s
0 .254 .766 r
.38711 .71812 m .45005 .77658 L .45395 .7844 L p F P
0 g
s
.656 .405 .528 r
.46759 .43113 m .46832 .35828 L .51046 .35494 L p F P
0 g
s
.59 .351 .523 r
.46832 .35828 m .46759 .43113 L .42705 .44041 L p F P
0 g
s
.938 .811 .694 r
.67876 .60128 m .63546 .68331 L .62547 .66748 L p F P
0 g
s
.938 .811 .694 r
.62547 .66748 m .66551 .57937 L .67876 .60128 L p F P
0 g
s
.514 .887 .971 r
.56241 .82074 m .51993 .81281 L .51704 .81361 L p F P
0 g
s
.608 .921 .973 r
.51993 .81281 m .56241 .82074 L .57896 .81323 L p F P
0 g
s
.715 .453 .528 r
.55269 .35777 m .55335 .43063 L .51042 .42781 L p F P
0 g
s
.715 .453 .528 r
.51042 .42781 m .51046 .35494 L .55269 .35777 L p F P
0 g
s
.92 .956 .795 r
.63774 .71676 m .57455 .77592 L .57532 .76776 L p F P
0 g
s
.92 .956 .795 r
.57532 .76776 m .63959 .7 L .63774 .71676 L p F P
0 g
s
.559 .696 .932 r
.41658 .78173 m .33976 .77484 L .35124 .79644 L p F P
0 g
s
.559 .696 .932 r
.35124 .79644 m .4227 .79372 L .41658 .78173 L p F P
0 g
s
.656 .405 .528 r
.51046 .35494 m .51042 .42781 L .46759 .43113 L p F P
0 g
s
.709 .726 .864 r
.48204 .90946 m .45814 .96709 L .42362 .95833 L p F P
0 g
s
.698 .713 .86 r
.45814 .96709 m .48204 .90946 L .50258 .91194 L p F P
0 g
s
.476 .333 .608 r
.37765 .56118 m .36081 .47447 L .39087 .45512 L p F P
0 g
s
.362 .258 .6 r
.36081 .47447 m .37765 .56118 L .35753 .5811 L p F P
0 g
s
.704 .834 .945 r
.67443 .79473 m .60239 .79276 L .5925 .80377 L p F P
0 g
s
.719 .824 .931 r
.60239 .79276 m .67443 .79473 L .68526 .77299 L p F P
0 g
s
.648 .736 .914 r
.75117 .78115 m .68526 .77299 L .67443 .79473 L p F P
0 g
s
.656 .683 .864 r
.56925 .96653 m .54402 .90914 L .56306 .90397 L p F P
0 g
s
.656 .683 .864 r
.56306 .90397 m .60351 .95742 L .56925 .96653 L p F P
0 g
s
.87 .654 .608 r
.66141 .47272 m .64479 .55966 L .61751 .54318 L p F P
0 g
s
.87 .654 .608 r
.61751 .54318 m .63078 .45372 L .66141 .47272 L p F P
0 g
s
0 .351 .777 r
.50213 .81176 m .46951 .79762 L .46055 .79152 L p F P
0 g
s
0 .351 .777 r
.46055 .79152 m .50004 .81033 L .50213 .81176 L p F P
0 g
s
.329 .646 .967 r
.50004 .81033 m .43291 .80461 L .44673 .81393 L p F P
0 g
s
.329 .646 .967 r
.44673 .81393 m .50213 .81176 L .50004 .81033 L p F P
0 g
s
.993 .743 .42 r
.19947 .83156 m .19808 .85953 L .22051 .89771 L p F P
0 g
s
0 .222 .742 r
.2427 .91707 m .22051 .89771 L .19808 .85953 L p F P
0 g
s
0 0 0 r
.51993 .81281 m .54497 .80208 L .53271 .80551 L p F P
0 g
s
.312 .781 .666 r
.54497 .80208 m .51993 .81281 L .52246 .81165 L p F P
0 g
s
.608 .921 .973 r
.57896 .81323 m .52246 .81165 L .51993 .81281 L p F P
0 g
s
.688 .934 .966 r
.52246 .81165 m .57896 .81323 L .5925 .80377 L p F P
0 g
s
.816 .993 .75 r
.80627 .89469 m .78388 .9143 L .80328 .87902 L p F P
0 g
s
.816 .993 .75 r
.80328 .87902 m .82756 .85625 L .80627 .89469 L p F P
0 g
s
.307 0 0 r
.51074 .81405 m .50601 .80738 L .49277 .80572 L p F P
0 g
s
.233 0 0 r
.50601 .80738 m .51074 .81405 L .51393 .81404 L p F P
0 g
s
.233 0 0 r
.51393 .81404 m .51953 .8073 L .50601 .80738 L p F P
0 g
s
.107 0 0 r
.51953 .8073 m .51393 .81404 L .51704 .81361 L p F P
0 g
s
.107 0 0 r
.51704 .81361 m .53271 .80551 L .51953 .8073 L p F P
0 g
s
.719 .741 .869 r
.46285 .90449 m .42362 .95833 L .39291 .94551 L p F P
0 g
s
.709 .726 .864 r
.42362 .95833 m .46285 .90449 L .48204 .90946 L p F P
0 g
s
.307 0 0 r
.49277 .80572 m .50761 .81366 L .51074 .81405 L p F P
0 g
s
0 0 0 r
.53271 .80551 m .51704 .81361 L .51993 .81281 L p F P
0 g
s
.64 .678 .87 r
.60351 .95742 m .56306 .90397 L .57983 .89656 L p F P
0 g
s
.64 .678 .87 r
.57983 .89656 m .63383 .94429 L .60351 .95742 L p F P
0 g
s
.309 .391 .784 r
.44904 .76843 m .38476 .70139 L .38839 .68466 L p F P
0 g
s
.153 .337 .795 r
.38476 .70139 m .44904 .76843 L .45005 .77658 L p F P
0 g
s
.312 .781 .666 r
.52246 .81165 m .55573 .79716 L .54497 .80208 L p F P
0 g
s
.009 .401 .852 r
.50004 .81033 m .46055 .79152 L .45395 .7844 L p F P
0 g
s
.009 .401 .852 r
.45395 .7844 m .49851 .80866 L .50004 .81033 L p F P
0 g
s
.372 .626 .951 r
.49851 .80866 m .4227 .79372 L .43291 .80461 L p F P
0 g
s
.372 .626 .951 r
.43291 .80461 m .50004 .81033 L .49851 .80866 L p F P
0 g
s
.558 .927 .786 r
.55573 .79716 m .52246 .81165 L .52451 .8102 L p F P
0 g
s
.688 .934 .966 r
.5925 .80377 m .52451 .8102 L .52246 .81165 L p F P
0 g
s
.746 .926 .95 r
.52451 .8102 m .5925 .80377 L .60239 .79276 L p F P
0 g
s
.919 .884 .783 r
.63959 .7 m .57532 .76776 L .57313 .75967 L p F P
0 g
s
.919 .884 .783 r
.57313 .75967 m .63546 .68331 L .63959 .7 L p F P
0 g
s
.671 .758 .916 r
.3362 .75211 m .2686 .75231 L .27413 .78369 L p F P
0 g
s
.671 .758 .916 r
.27413 .78369 m .33976 .77484 L .3362 .75211 L p F P
0 g
s
.775 .826 .892 r
.22579 .80431 m .27413 .78369 L .2686 .75231 L p F P
0 g
s
.423 .662 .96 r
.2427 .91707 m .28665 .92252 L .31336 .9488 L p F P
0 g
s
.638 .749 .929 r
.3674 .92917 m .31336 .9488 L .28665 .92252 L p F P
0 g
s
.638 .752 .931 r
.31336 .9488 m .3674 .92917 L .39291 .94551 L p F P
0 g
s
.441 .376 .688 r
.3979 .66874 m .35753 .5811 L .37765 .56118 L p F P
0 g
s
.326 .315 .694 r
.35753 .5811 m .3979 .66874 L .38839 .68466 L p F P
0 g
s
.646 .738 .916 r
.68526 .77299 m .75117 .78115 L .75575 .74969 L p F P
0 g
s
.533 .631 .894 r
.75117 .78115 m .79973 .80127 L .8058 .76358 L p F P
0 g
s
.533 .631 .894 r
.8058 .76358 m .75575 .74969 L .75117 .78115 L p F P
0 g
s
.893 .742 .687 r
.66551 .57937 m .62547 .66748 L .61002 .6533 L p F P
0 g
s
.893 .742 .687 r
.61002 .6533 m .64479 .55966 L .66551 .57937 L p F P
0 g
s
.661 .768 .929 r
.71361 .94678 m .65884 .92768 L .67735 .90826 L p F P
0 g
s
.661 .768 .929 r
.67735 .90826 m .73953 .9202 L .71361 .94678 L p F P
0 g
s
.557 .39 .61 r
.40443 .54439 m .39087 .45512 L .42705 .44041 L p F P
0 g
s
.476 .333 .608 r
.39087 .45512 m .40443 .54439 L .37765 .56118 L p F P
0 g
s
.775 .826 .892 r
.2686 .75231 m .21859 .76672 L .22579 .80431 L p F P
0 g
s
.909 .895 .803 r
.19947 .83156 m .22579 .80431 L .21859 .76672 L p F P
0 g
s
.558 .927 .786 r
.52451 .8102 m .56451 .79096 L .55573 .79716 L p F P
0 g
s
.819 .603 .61 r
.63078 .45372 m .61751 .54318 L .58495 .53079 L p F P
0 g
s
.819 .603 .61 r
.58495 .53079 m .59416 .43943 L .63078 .45372 L p F P
0 g
s
.572 .686 .917 r
.4149 .76919 m .3362 .75211 L .33976 .77484 L p F P
0 g
s
.572 .686 .917 r
.33976 .77484 m .41658 .78173 L .4149 .76919 L p F P
0 g
s
.312 .413 .805 r
.79973 .80127 m .82617 .82827 L .83312 .7874 L p F P
0 g
s
.312 .413 .805 r
.83312 .7874 m .8058 .76358 L .79973 .80127 L p F P
0 g
s
.719 .824 .931 r
.68526 .77299 m .60814 .7807 L .60239 .79276 L p F P
0 g
s
.728 .811 .916 r
.60814 .7807 m .68526 .77299 L .68813 .75022 L p F P
0 g
s
.646 .738 .916 r
.75575 .74969 m .68813 .75022 L .68526 .77299 L p F P
0 g
s
.727 .757 .877 r
.44585 .89725 m .39291 .94551 L .3674 .92917 L p F P
0 g
s
.719 .741 .869 r
.39291 .94551 m .44585 .89725 L .46285 .90449 L p F P
0 g
s
0 .222 .742 r
.19808 .85953 m .22224 .88203 L .2427 .91707 L p F P
0 g
s
.456 .646 .94 r
.28665 .92252 m .2427 .91707 L .22224 .88203 L p F P
0 g
s
.624 .675 .878 r
.63383 .94429 m .57983 .89656 L .59355 .8872 L p F P
0 g
s
.624 .675 .878 r
.59355 .8872 m .65884 .92768 L .63383 .94429 L p F P
0 g
s
.17 .438 .87 r
.49851 .80866 m .45395 .7844 L .45005 .77658 L p F P
0 g
s
.17 .438 .87 r
.45005 .77658 m .49762 .80684 L .49851 .80866 L p F P
0 g
s
.419 .611 .928 r
.49762 .80684 m .41658 .78173 L .4227 .79372 L p F P
0 g
s
.419 .611 .928 r
.4227 .79372 m .49851 .80866 L .49762 .80684 L p F P
0 g
s
.745 .993 .856 r
.56451 .79096 m .52451 .8102 L .52599 .80852 L p F P
0 g
s
.745 .993 .856 r
.52599 .80852 m .57089 .78378 L .56451 .79096 L p F P
0 g
s
.746 .926 .95 r
.60239 .79276 m .52599 .80852 L .52451 .8102 L p F P
0 g
s
.782 .903 .926 r
.52599 .80852 m .60239 .79276 L .60814 .7807 L p F P
0 g
s
.755 .887 .938 r
.78388 .9143 m .73953 .9202 L .75534 .89079 L p F P
0 g
s
.755 .887 .938 r
.75534 .89079 m .80328 .87902 L .78388 .9143 L p F P
0 g
s
.909 .895 .803 r
.21859 .76672 m .19129 .79081 L .19947 .83156 L p F P
0 g
s
.921 .732 .242 r
.19808 .85953 m .19947 .83156 L .19129 .79081 L p F P
0 g
s
.417 .43 .766 r
.45098 .76032 m .38839 .68466 L .3979 .66874 L p F P
0 g
s
.309 .391 .784 r
.38839 .68466 m .45098 .76032 L .44904 .76843 L p F P
0 g
s
.893 .815 .764 r
.63546 .68331 m .57313 .75967 L .56806 .75203 L p F P
0 g
s
.893 .815 .764 r
.56806 .75203 m .62547 .66748 L .63546 .68331 L p F P
0 g
s
.62 .436 .611 r
.4366 .53164 m .42705 .44041 L .46759 .43113 L p F P
0 g
s
.557 .39 .61 r
.42705 .44041 m .4366 .53164 L .40443 .54439 L p F P
0 g
s
.844 .98 .87 r
.57089 .78378 m .52599 .80852 L .52683 .80669 L p F P
0 g
s
.844 .98 .87 r
.52683 .80669 m .57455 .77592 L .57089 .78378 L p F P
0 g
s
0 0 .22 r
.82617 .82827 m .82756 .85625 L .83455 .81553 L p F P
0 g
s
0 0 .22 r
.83455 .81553 m .83312 .7874 L .82617 .82827 L p F P
0 g
s
.771 .558 .611 r
.59416 .43943 m .58495 .53079 L .54871 .52317 L p F P
0 g
s
.771 .558 .611 r
.54871 .52317 m .55335 .43063 L .59416 .43943 L p F P
0 g
s
.72 .676 .806 r
.51073 .83025 m .50258 .91194 L .48204 .90946 L p F P
0 g
s
.698 .656 .804 r
.50258 .91194 m .51073 .83025 L .51393 .83023 L p F P
0 g
s
.698 .656 .804 r
.51393 .83023 m .52355 .91183 L .50258 .91194 L p F P
0 g
s
.675 .639 .806 r
.52355 .91183 m .51393 .83023 L .51705 .82981 L p F P
0 g
s
.675 .639 .806 r
.51705 .82981 m .54402 .90914 L .52355 .91183 L p F P
0 g
s
.523 .421 .681 r
.41291 .65439 m .37765 .56118 L .40443 .54439 L p F P
0 g
s
.441 .376 .688 r
.37765 .56118 m .41291 .65439 L .3979 .66874 L p F P
0 g
s
.741 .698 .81 r
.5076 .82986 m .48204 .90946 L .46285 .90449 L p F P
0 g
s
.72 .676 .806 r
.48204 .90946 m .5076 .82986 L .51073 .83025 L p F P
0 g
s
.849 .686 .68 r
.64479 .55966 m .61002 .6533 L .5898 .64147 L p F P
0 g
s
.849 .686 .68 r
.5898 .64147 m .61751 .54318 L .64479 .55966 L p F P
0 g
s
.299 .463 .855 r
.49762 .80684 m .45005 .77658 L .44904 .76843 L p F P
0 g
s
.299 .463 .855 r
.44904 .76843 m .49741 .80495 L .49762 .80684 L p F P
0 g
s
.464 .601 .903 r
.49741 .80495 m .4149 .76919 L .41658 .78173 L p F P
0 g
s
.464 .601 .903 r
.41658 .78173 m .49762 .80684 L .49741 .80495 L p F P
0 g
s
.65 .624 .811 r
.54402 .90914 m .51705 .82981 L .51995 .82901 L p F P
0 g
s
.65 .624 .811 r
.51995 .82901 m .56306 .90397 L .54402 .90914 L p F P
0 g
s
.782 .903 .926 r
.60814 .7807 m .52683 .80669 L .52599 .80852 L p F P
0 g
s
.799 .871 .901 r
.52683 .80669 m .60814 .7807 L .60944 .76814 L p F P
0 g
s
.728 .811 .916 r
.68813 .75022 m .60944 .76814 L .60814 .7807 L p F P
0 g
s
.877 .929 .853 r
.57455 .77592 m .52683 .80669 L .52698 .80479 L p F P
0 g
s
.877 .929 .853 r
.52698 .80479 m .57532 .76776 L .57455 .77592 L p F P
0 g
s
.638 .749 .929 r
.28665 .92252 m .34831 .90995 L .3674 .92917 L p F P
0 g
s
.731 .774 .887 r
.43185 .88804 m .3674 .92917 L .34831 .90995 L p F P
0 g
s
.727 .757 .877 r
.3674 .92917 m .43185 .88804 L .44585 .89725 L p F P
0 g
s
.674 .477 .611 r
.4726 .52361 m .46759 .43113 L .51042 .42781 L p F P
0 g
s
.62 .436 .611 r
.46759 .43113 m .4726 .52361 L .4366 .53164 L p F P
0 g
s
.723 .517 .611 r
.55335 .43063 m .54871 .52317 L .51062 .52073 L p F P
0 g
s
.723 .517 .611 r
.51062 .52073 m .51042 .42781 L .55335 .43063 L p F P
0 g
s
.608 .674 .888 r
.65884 .92768 m .59355 .8872 L .60358 .87633 L p F P
0 g
s
.608 .674 .888 r
.60358 .87633 m .67735 .90826 L .65884 .92768 L p F P
0 g
s
.76 .724 .819 r
.50468 .82909 m .46285 .90449 L .44585 .89725 L p F P
0 g
s
.741 .698 .81 r
.46285 .90449 m .50468 .82909 L .5076 .82986 L p F P
0 g
s
.623 .612 .82 r
.56306 .90397 m .51995 .82901 L .52249 .82785 L p F P
0 g
s
.623 .612 .82 r
.52249 .82785 m .57983 .89656 L .56306 .90397 L p F P
0 g
s
.674 .477 .611 r
.51042 .42781 m .51062 .52073 L .4726 .52361 L p F P
0 g
s
.587 .678 .904 r
.41777 .75668 m .3409 .72931 L .3362 .75211 L p F P
0 g
s
.587 .678 .904 r
.3362 .75211 m .4149 .76919 L .41777 .75668 L p F P
0 g
s
.669 .759 .918 r
.2686 .75231 m .3362 .75211 L .3409 .72931 L p F P
0 g
s
.456 .646 .94 r
.22224 .88203 m .26995 .8933 L .28665 .92252 L p F P
0 g
s
.638 .747 .927 r
.34831 .90995 m .28665 .92252 L .26995 .8933 L p F P
0 g
s
.921 .732 .242 r
.19129 .79081 m .18985 .81893 L .19808 .85953 L p F P
0 g
s
.105 .32 .793 r
.22224 .88203 m .19808 .85953 L .18985 .81893 L p F P
0 g
s
.733 .796 .903 r
.60944 .76814 m .68813 .75022 L .68275 .72746 L p F P
0 g
s
.645 .74 .918 r
.68813 .75022 m .75575 .74969 L .7489 .7181 L p F P
0 g
s
.645 .74 .918 r
.7489 .7181 m .68275 .72746 L .68813 .75022 L p F P
0 g
s
.496 .461 .748 r
.45582 .75263 m .3979 .66874 L .41291 .65439 L p F P
0 g
s
.417 .43 .766 r
.3979 .66874 m .45582 .75263 L .45098 .76032 L p F P
0 g
s
.663 .767 .927 r
.73953 .9202 m .67735 .90826 L .6884 .88687 L p F P
0 g
s
.663 .767 .927 r
.6884 .88687 m .75534 .89079 L .73953 .9202 L p F P
0 g
s
.669 .759 .918 r
.3409 .72931 m .27449 .72067 L .2686 .75231 L p F P
0 g
s
.768 .852 .914 r
.21859 .76672 m .2686 .75231 L .27449 .72067 L p F P
0 g
s
.86 .756 .747 r
.62547 .66748 m .56806 .75203 L .56035 .74522 L p F P
0 g
s
.86 .756 .747 r
.56035 .74522 m .61002 .6533 L .62547 .66748 L p F P
0 g
s
.396 .482 .83 r
.49741 .80495 m .44904 .76843 L .45098 .76032 L p F P
0 g
s
.912 .971 .794 r
.82756 .85625 m .80328 .87902 L .80946 .84174 L p F P
0 g
s
.912 .971 .794 r
.80946 .84174 m .83455 .81553 L .82756 .85625 L p F P
0 g
s
.875 .869 .828 r
.57532 .76776 m .52698 .80479 L .52644 .80292 L p F P
0 g
s
.875 .869 .828 r
.52644 .80292 m .57313 .75967 L .57532 .76776 L p F P
0 g
s
.506 .641 .915 r
.75575 .74969 m .8058 .76358 L .79807 .72557 L p F P
0 g
s
.506 .641 .915 r
.79807 .72557 m .7489 .7181 L .75575 .74969 L p F P
0 g
s
.777 .752 .831 r
.50211 .82796 m .44585 .89725 L .43185 .88804 L p F P
0 g
s
.76 .724 .819 r
.44585 .89725 m .50211 .82796 L .50468 .82909 L p F P
0 g
s
.396 .482 .83 r
.45098 .76032 m .49789 .80307 L .49741 .80495 L p F P
0 g
s
.506 .597 .879 r
.49789 .80307 m .41777 .75668 L .4149 .76919 L p F P
0 g
s
.506 .597 .879 r
.4149 .76919 m .49741 .80495 L .49789 .80307 L p F P
0 g
s
.593 .604 .832 r
.57983 .89656 m .52249 .82785 L .52454 .8264 L p F P
0 g
s
.593 .604 .832 r
.52454 .8264 m .59355 .8872 L .57983 .89656 L p F P
0 g
s
.799 .871 .901 r
.60944 .76814 m .52698 .80479 L .52683 .80669 L p F P
0 g
s
.802 .835 .877 r
.52698 .80479 m .60944 .76814 L .60619 .75567 L p F P
0 g
s
.733 .796 .903 r
.68275 .72746 m .60619 .75567 L .60944 .76814 L p F P
0 g
s
.586 .459 .675 r
.43277 .64234 m .40443 .54439 L .4366 .53164 L p F P
0 g
s
.523 .421 .681 r
.40443 .54439 m .43277 .64234 L .41291 .65439 L p F P
0 g
s
.806 .639 .674 r
.61751 .54318 m .5898 .64147 L .56576 .63259 L p F P
0 g
s
.806 .639 .674 r
.56576 .63259 m .58495 .53079 L .61751 .54318 L p F P
0 g
s
.74 .613 .716 r
.49249 .8925 m .47205 .996 L .4994 .99922 L p F P
0 g
s
.74 .613 .716 r
.4994 .99922 m .50591 .89413 L .49249 .8925 L p F P
0 g
s
.744 .579 .67 r
.49465 1.10504 m .4994 .99922 L .47205 .996 L p F P
0 g
s
.706 .583 .715 r
.50591 .89413 m .4994 .99922 L .5273 .99908 L p F P
0 g
s
.704 .545 .669 r
.4994 .99922 m .49465 1.10504 L .53289 1.10485 L p F P
0 g
s
.704 .545 .669 r
.53289 1.10485 m .5273 .99908 L .4994 .99922 L p F P
0 g
s
.706 .583 .715 r
.5273 .99908 m .51962 .89406 L .50591 .89413 L p F P
0 g
s
.669 .555 .717 r
.51962 .89406 m .5273 .99908 L .55454 .99559 L p F P
0 g
s
.662 .512 .67 r
.5273 .99908 m .53289 1.10485 L .57024 1.10023 L p F P
0 g
s
.662 .512 .67 r
.57024 1.10023 m .55454 .99559 L .5273 .99908 L p F P
0 g
s
.735 .637 .75 r
.5076 .82986 m .49249 .8925 L .50591 .89413 L p F P
0 g
s
.735 .637 .75 r
.50591 .89413 m .51073 .83025 L .5076 .82986 L p F P
0 g
s
.705 .61 .748 r
.51073 .83025 m .50591 .89413 L .51962 .89406 L p F P
0 g
s
.705 .61 .748 r
.51962 .89406 m .51393 .83023 L .51073 .83025 L p F P
0 g
s
.673 .586 .75 r
.51393 .83023 m .51962 .89406 L .53298 .89229 L p F P
0 g
s
.669 .555 .717 r
.55454 .99559 m .53298 .89229 L .51962 .89406 L p F P
0 g
s
.471 .499 .806 r
.49789 .80307 m .45098 .76032 L .45582 .75263 L p F P
0 g
s
.774 .646 .721 r
.47997 .88924 m .44647 .98956 L .47205 .996 L p F P
0 g
s
.774 .646 .721 r
.47205 .996 m .49249 .8925 L .47997 .88924 L p F P
0 g
s
.784 .617 .673 r
.45717 1.10077 m .47205 .996 L .44647 .98956 L p F P
0 g
s
.744 .579 .67 r
.47205 .996 m .45717 1.10077 L .49465 1.10504 L p F P
0 g
s
.629 .528 .721 r
.53298 .89229 m .55454 .99559 L .57993 .98889 L p F P
0 g
s
.615 .48 .673 r
.55454 .99559 m .57024 1.10023 L .60511 1.09134 L p F P
0 g
s
.615 .48 .673 r
.60511 1.09134 m .57993 .98889 L .55454 .99559 L p F P
0 g
s
.768 .852 .914 r
.27449 .72067 m .22515 .72867 L .21859 .76672 L p F P
0 g
s
.901 .962 .821 r
.19129 .79081 m .21859 .76672 L .22515 .72867 L p F P
0 g
s
.858 .812 .804 r
.57313 .75967 m .52644 .80292 L .52523 .80116 L p F P
0 g
s
.858 .812 .804 r
.52523 .80116 m .56806 .75203 L .57313 .75967 L p F P
0 g
s
.764 .667 .755 r
.50468 .82909 m .47997 .88924 L .49249 .8925 L p F P
0 g
s
.764 .667 .755 r
.49249 .8925 m .5076 .82986 L .50468 .82909 L p F P
0 g
s
.673 .586 .75 r
.53298 .89229 m .51705 .82981 L .51393 .83023 L p F P
0 g
s
.638 .564 .756 r
.51705 .82981 m .53298 .89229 L .54541 .8889 L p F P
0 g
s
.629 .528 .721 r
.57993 .98889 m .54541 .8889 L .53298 .89229 L p F P
0 g
s
.731 .774 .887 r
.34831 .90995 m .42149 .87727 L .43185 .88804 L p F P
0 g
s
.791 .784 .846 r
.5 .82653 m .43185 .88804 L .42149 .87727 L p F P
0 g
s
.777 .752 .831 r
.43185 .88804 m .5 .82653 L .50211 .82796 L p F P
0 g
s
.557 .489 .735 r
.46332 .74574 m .41291 .65439 L .43277 .64234 L p F P
0 g
s
.496 .461 .748 r
.41291 .65439 m .46332 .74574 L .45582 .75263 L p F P
0 g
s
.173 .374 .821 r
.8058 .76358 m .83312 .7874 L .82499 .7461 L p F P
0 g
s
.173 .374 .821 r
.82499 .7461 m .79807 .72557 L .8058 .76358 L p F P
0 g
s
.561 .598 .848 r
.59355 .8872 m .52454 .8264 L .52602 .82473 L p F P
0 g
s
.561 .598 .848 r
.52602 .82473 m .60358 .87633 L .59355 .8872 L p F P
0 g
s
.826 .707 .734 r
.61002 .6533 m .56035 .74522 L .55033 .73955 L p F P
0 g
s
.826 .707 .734 r
.55033 .73955 m .5898 .64147 L .61002 .6533 L p F P
0 g
s
.638 .747 .927 r
.26995 .8933 m .3366 .88869 L .34831 .90995 L p F P
0 g
s
.733 .791 .899 r
.42149 .87727 m .34831 .90995 L .3366 .88869 L p F P
0 g
s
.792 .701 .764 r
.50211 .82796 m .46891 .88449 L .47997 .88924 L p F P
0 g
s
.792 .701 .764 r
.47997 .88924 m .50468 .82909 L .50211 .82796 L p F P
0 g
s
.808 .685 .728 r
.44647 .98956 m .47997 .88924 L .46891 .88449 L p F P
0 g
s
.105 .32 .793 r
.18985 .81893 m .21491 .84485 L .22224 .88203 L p F P
0 g
s
.491 .635 .918 r
.26995 .8933 m .22224 .88203 L .21491 .84485 L p F P
0 g
s
.638 .564 .756 r
.54541 .8889 m .51995 .82901 L .51705 .82981 L p F P
0 g
s
.599 .544 .764 r
.51995 .82901 m .54541 .8889 L .55632 .88404 L p F P
0 g
s
.584 .503 .729 r
.54541 .8889 m .57993 .98889 L .60233 .97926 L p F P
0 g
s
.584 .503 .729 r
.60233 .97926 m .55632 .88404 L .54541 .8889 L p F P
0 g
s
.471 .499 .806 r
.45582 .75263 m .49904 .8013 L .49789 .80307 L p F P
0 g
s
.544 .597 .857 r
.49904 .8013 m .42512 .74479 L .41777 .75668 L p F P
0 g
s
.544 .597 .857 r
.41777 .75668 m .49789 .80307 L .49904 .8013 L p F P
0 g
s
.602 .674 .891 r
.3409 .72931 m .41777 .75668 L .42512 .74479 L p F P
0 g
s
.591 .677 .9 r
.67735 .90826 m .60358 .87633 L .60941 .86441 L p F P
0 g
s
.591 .677 .9 r
.60941 .86441 m .6884 .88687 L .67735 .90826 L p F P
0 g
s
.808 .685 .728 r
.46891 .88449 m .42377 .98015 L .44647 .98956 L p F P
0 g
s
.824 .659 .678 r
.42203 1.09222 m .44647 .98956 L .42377 .98015 L p F P
0 g
s
.784 .617 .673 r
.44647 .98956 m .42203 1.09222 L .45717 1.10077 L p F P
0 g
s
.802 .835 .877 r
.60619 .75567 m .52644 .80292 L .52698 .80479 L p F P
0 g
s
.796 .8 .855 r
.52644 .80292 m .60619 .75567 L .59848 .74386 L p F P
0 g
s
.733 .78 .89 r
.60619 .75567 m .68275 .72746 L .6692 .7058 L p F P
0 g
s
.733 .78 .89 r
.6692 .7058 m .59848 .74386 L .60619 .75567 L p F P
0 g
s
.638 .494 .671 r
.45653 .6332 m .4366 .53164 L .4726 .52361 L p F P
0 g
s
.586 .459 .675 r
.4366 .53164 m .45653 .6332 L .43277 .64234 L p F P
0 g
s
.53 .517 .786 r
.49904 .8013 m .45582 .75263 L .46332 .74574 L p F P
0 g
s
.561 .445 .679 r
.57993 .98889 m .60511 1.09134 L .63598 1.07853 L p F P
0 g
s
.561 .445 .679 r
.63598 1.07853 m .60233 .97926 L .57993 .98889 L p F P
0 g
s
.772 .861 .916 r
.80328 .87902 m .75534 .89079 L .76004 .85981 L p F P
0 g
s
.772 .861 .916 r
.76004 .85981 m .80946 .84174 L .80328 .87902 L p F P
0 g
s
.766 .598 .67 r
.58495 .53079 m .56576 .63259 L .53907 .62714 L p F P
0 g
s
.766 .598 .67 r
.53907 .62714 m .54871 .52317 L .58495 .53079 L p F P
0 g
s
.833 .763 .784 r
.56806 .75203 m .52523 .80116 L .52342 .7996 L p F P
0 g
s
.833 .763 .784 r
.52342 .7996 m .56035 .74522 L .56806 .75203 L p F P
0 g
s
.82 .74 .776 r
.5 .82653 m .45982 .87847 L .46891 .88449 L p F P
0 g
s
.82 .74 .776 r
.46891 .88449 m .50211 .82796 L .5 .82653 L p F P
0 g
s
.842 .729 .74 r
.42377 .98015 m .46891 .88449 L .45982 .87847 L p F P
0 g
s
.599 .544 .764 r
.55632 .88404 m .52249 .82785 L .51995 .82901 L p F P
0 g
s
.555 .526 .778 r
.52249 .82785 m .55632 .88404 L .56523 .87792 L p F P
0 g
s
.53 .476 .741 r
.55632 .88404 m .60233 .97926 L .62072 .96709 L p F P
0 g
s
.53 .476 .741 r
.62072 .96709 m .56523 .87792 L .55632 .88404 L p F P
0 g
s
.602 .674 .891 r
.42512 .74479 m .35378 .70751 L .3409 .72931 L p F P
0 g
s
.667 .761 .92 r
.27449 .72067 m .3409 .72931 L .35378 .70751 L p F P
0 g
s
.745 .54 .617 r
.49207 1.20452 m .49465 1.10504 L .45717 1.10077 L p F P
0 g
s
.7 .502 .617 r
.49465 1.10504 m .49207 1.20452 L .53592 1.20432 L p F P
0 g
s
.7 .502 .617 r
.53592 1.20432 m .53289 1.10485 L .49465 1.10504 L p F P
0 g
s
.652 .464 .617 r
.53289 1.10485 m .53592 1.20432 L .57876 1.19918 L p F P
0 g
s
.652 .464 .617 r
.57876 1.19918 m .57024 1.10023 L .53289 1.10485 L p F P
0 g
s
.555 .526 .778 r
.56523 .87792 m .52454 .8264 L .52249 .82785 L p F P
0 g
s
.796 .8 .855 r
.59848 .74386 m .52523 .80116 L .52644 .80292 L p F P
0 g
s
.644 .742 .92 r
.68275 .72746 m .7489 .7181 L .73062 .68788 L p F P
0 g
s
.644 .742 .92 r
.73062 .68788 m .6692 .7058 L .68275 .72746 L p F P
0 g
s
.791 .784 .846 r
.42149 .87727 m .49847 .82487 L .5 .82653 L p F P
0 g
s
.846 .786 .794 r
.45982 .87847 m .5 .82653 L .49847 .82487 L p F P
0 g
s
.502 .508 .795 r
.52454 .8264 m .56523 .87792 L .57169 .87082 L p F P
0 g
s
.502 .508 .795 r
.57169 .87082 m .52602 .82473 L .52454 .8264 L p F P
0 g
s
.53 .517 .786 r
.46332 .74574 m .50081 .79972 L .49904 .8013 L p F P
0 g
s
.578 .6 .839 r
.42512 .74479 m .49904 .8013 L .50081 .79972 L p F P
0 g
s
.784 .767 .838 r
.52523 .80116 m .59848 .74386 L .58664 .7333 L p F P
0 g
s
.784 .767 .838 r
.58664 .7333 m .52342 .7996 L .52523 .80116 L p F P
0 g
s
.733 .791 .899 r
.3366 .88869 m .41529 .86542 L .42149 .87727 L p F P
0 g
s
.8 .818 .866 r
.49847 .82487 m .42149 .87727 L .41529 .86542 L p F P
0 g
s
.557 .489 .735 r
.43277 .64234 m .47316 .73997 L .46332 .74574 L p F P
0 g
s
.578 .535 .771 r
.50081 .79972 m .46332 .74574 L .47316 .73997 L p F P
0 g
s
.8 .818 .866 r
.41529 .86542 m .49758 .82306 L .49847 .82487 L p F P
0 g
s
.868 .839 .816 r
.45313 .87144 m .49847 .82487 L .49758 .82306 L p F P
0 g
s
.846 .786 .794 r
.49847 .82487 m .45313 .87144 L .45982 .87847 L p F P
0 g
s
.901 .962 .821 r
.22515 .72867 m .19814 .74947 L .19129 .79081 L p F P
0 g
s
0 0 .065 r
.18985 .81893 m .19129 .79081 L .19814 .74947 L p F P
0 g
s
.683 .527 .669 r
.48305 .62745 m .4726 .52361 L .51062 .52073 L p F P
0 g
s
.638 .494 .671 r
.4726 .52361 m .48305 .62745 L .45653 .6332 L p F P
0 g
s
.525 .596 .868 r
.60358 .87633 m .52602 .82473 L .52686 .8229 L p F P
0 g
s
.436 .491 .818 r
.52602 .82473 m .57169 .87082 L .57541 .86306 L p F P
0 g
s
.436 .491 .818 r
.57541 .86306 m .52686 .8229 L .52602 .82473 L p F P
0 g
s
.807 .72 .769 r
.56035 .74522 m .52342 .7996 L .52107 .7983 L p F P
0 g
s
.807 .72 .769 r
.52107 .7983 m .55033 .73955 L .56035 .74522 L p F P
0 g
s
.578 .535 .771 r
.47316 .73997 m .50311 .7984 L .50081 .79972 L p F P
0 g
s
.608 .608 .825 r
.43663 .73411 m .50081 .79972 L .50311 .7984 L p F P
0 g
s
.578 .6 .839 r
.50081 .79972 m .43663 .73411 L .42512 .74479 L p F P
0 g
s
.606 .515 .725 r
.47316 .73997 m .43277 .64234 L .45653 .6332 L p F P
0 g
s
.769 .737 .824 r
.52342 .7996 m .58664 .7333 L .5712 .7245 L p F P
0 g
s
.769 .737 .824 r
.5712 .7245 m .52107 .7983 L .52342 .7996 L p F P
0 g
s
.79 .581 .617 r
.44907 1.19979 m .45717 1.10077 L .42203 1.09222 L p F P
0 g
s
.745 .54 .617 r
.45717 1.10077 m .44907 1.19979 L .49207 1.20452 L p F P
0 g
s
.725 .562 .669 r
.54871 .52317 m .53907 .62714 L .51103 .6254 L p F P
0 g
s
.725 .562 .669 r
.51103 .6254 m .51062 .52073 L .54871 .52317 L p F P
0 g
s
.525 .596 .868 r
.52686 .8229 m .60941 .86441 L .60358 .87633 L p F P
0 g
s
.597 .424 .617 r
.57024 1.10023 m .57876 1.19918 L .61881 1.18931 L p F P
0 g
s
.597 .424 .617 r
.61881 1.18931 m .60511 1.09134 L .57024 1.10023 L p F P
0 g
s
.876 .783 .755 r
.40501 .96818 m .45982 .87847 L .45313 .87144 L p F P
0 g
s
.842 .729 .74 r
.45982 .87847 m .40501 .96818 L .42377 .98015 L p F P
0 g
s
.792 .665 .724 r
.5898 .64147 m .55033 .73955 L .53848 .73532 L p F P
0 g
s
.792 .665 .724 r
.53848 .73532 m .56576 .63259 L .5898 .64147 L p F P
0 g
s
.619 .554 .76 r
.48488 .73561 m .50585 .7974 L .50311 .7984 L p F P
0 g
s
.619 .554 .76 r
.50311 .7984 m .47316 .73997 L .48488 .73561 L p F P
0 g
s
.636 .618 .815 r
.45179 .72515 m .50311 .7984 L .50585 .7974 L p F P
0 g
s
.608 .608 .825 r
.50311 .7984 m .45179 .72515 L .43663 .73411 L p F P
0 g
s
.462 .448 .756 r
.56523 .87792 m .62072 .96709 L .63423 .95292 L p F P
0 g
s
.462 .448 .756 r
.63423 .95292 m .57169 .87082 L .56523 .87792 L p F P
0 g
s
.778 .683 .759 r
.55033 .73955 m .52107 .7983 L .51831 .79733 L p F P
0 g
s
.778 .683 .759 r
.51831 .79733 m .53848 .73532 L .55033 .73955 L p F P
0 g
s
.751 .71 .814 r
.52107 .7983 m .5712 .7245 L .5529 .71792 L p F P
0 g
s
.751 .71 .814 r
.5529 .71792 m .51831 .79733 L .52107 .7983 L p F P
0 g
s
.802 .853 .889 r
.41358 .85302 m .49737 .82117 L .49758 .82306 L p F P
0 g
s
.802 .853 .889 r
.49758 .82306 m .41529 .86542 L .41358 .85302 L p F P
0 g
s
.879 .899 .841 r
.44918 .86372 m .49758 .82306 L .49737 .82117 L p F P
0 g
s
.868 .839 .816 r
.49758 .82306 m .44918 .86372 L .45313 .87144 L p F P
0 g
s
.485 .598 .891 r
.60941 .86441 m .52686 .8229 L .52702 .82101 L p F P
0 g
s
.352 .473 .843 r
.52686 .8229 m .57541 .86306 L .57619 .85499 L p F P
0 g
s
.352 .473 .843 r
.57619 .85499 m .52702 .82101 L .52686 .8229 L p F P
0 g
s
.866 .709 .685 r
.39077 1.07972 m .42377 .98015 L .40501 .96818 L p F P
0 g
s
.824 .659 .678 r
.42377 .98015 m .39077 1.07972 L .42203 1.09222 L p F P
0 g
s
.656 .575 .753 r
.49791 .73287 m .50888 .79677 L .50585 .7974 L p F P
0 g
s
.656 .575 .753 r
.50585 .7974 m .48488 .73561 L .49791 .73287 L p F P
0 g
s
.606 .515 .725 r
.45653 .6332 m .48488 .73561 L .47316 .73997 L p F P
0 g
s
.663 .631 .808 r
.46989 .71837 m .50585 .7974 L .50888 .79677 L p F P
0 g
s
.636 .618 .815 r
.50585 .7974 m .46989 .71837 L .45179 .72515 L p F P
0 g
s
.75 .651 .752 r
.53848 .73532 m .51831 .79733 L .51526 .79674 L p F P
0 g
s
.75 .651 .752 r
.51526 .79674 m .52537 .73272 L .53848 .73532 L p F P
0 g
s
.758 .629 .718 r
.56576 .63259 m .53848 .73532 L .52537 .73272 L p F P
0 g
s
.758 .629 .718 r
.52537 .73272 m .53907 .62714 L .56576 .63259 L p F P
0 g
s
.731 .686 .808 r
.51831 .79733 m .5529 .71792 L .5326 .71388 L p F P
0 g
s
.731 .686 .808 r
.5326 .71388 m .51526 .79674 L .51831 .79733 L p F P
0 g
s
.683 .527 .669 r
.51062 .52073 m .51103 .6254 L .48305 .62745 L p F P
0 g
s
.493 .406 .686 r
.60233 .97926 m .63598 1.07853 L .66144 1.06232 L p F P
0 g
s
.493 .406 .686 r
.66144 1.06232 m .62072 .96709 L .60233 .97926 L p F P
0 g
s
.618 .673 .88 r
.35378 .70751 m .42512 .74479 L .43663 .73411 L p F P
0 g
s
.689 .598 .749 r
.51162 .73189 m .51207 .79655 L .50888 .79677 L p F P
0 g
s
.689 .598 .749 r
.50888 .79677 m .49791 .73287 L .51162 .73189 L p F P
0 g
s
.649 .541 .719 r
.48305 .62745 m .49791 .73287 L .48488 .73561 L p F P
0 g
s
.687 .647 .805 r
.49005 .71411 m .50888 .79677 L .51207 .79655 L p F P
0 g
s
.649 .541 .719 r
.48488 .73561 m .45653 .6332 L .48305 .62745 L p F P
0 g
s
.663 .631 .808 r
.50888 .79677 m .49005 .71411 L .46989 .71837 L p F P
0 g
s
.72 .623 .749 r
.52537 .73272 m .51526 .79674 L .51207 .79655 L p F P
0 g
s
.72 .623 .749 r
.51207 .79655 m .51162 .73189 L .52537 .73272 L p F P
0 g
s
.687 .568 .715 r
.51103 .6254 m .51162 .73189 L .49791 .73287 L p F P
0 g
s
.723 .597 .715 r
.53907 .62714 m .52537 .73272 L .51162 .73189 L p F P
0 g
s
.723 .597 .715 r
.51162 .73189 m .51103 .6254 L .53907 .62714 L p F P
0 g
s
.71 .665 .804 r
.51526 .79674 m .5326 .71388 L .51131 .71258 L p F P
0 g
s
.71 .665 .804 r
.51131 .71258 m .51207 .79655 L .51526 .79674 L p F P
0 g
s
.687 .568 .715 r
.49791 .73287 m .48305 .62745 L .51103 .6254 L p F P
0 g
s
.687 .647 .805 r
.51207 .79655 m .51131 .71258 L .49005 .71411 L p F P
0 g
s
.846 .634 .091 r
.83312 .7874 m .83455 .81553 L .82641 .77439 L p F P
0 g
s
.846 .634 .091 r
.82641 .77439 m .82499 .7461 L .83312 .7874 L p F P
0 g
s
.792 .888 .914 r
.41649 .84066 m .49785 .8193 L .49737 .82117 L p F P
0 g
s
.792 .888 .914 r
.49737 .82117 m .41358 .85302 L .41649 .84066 L p F P
0 g
s
.866 .957 .864 r
.44815 .85566 m .49737 .82117 L .49785 .8193 L p F P
0 g
s
.879 .899 .841 r
.49737 .82117 m .44815 .85566 L .44918 .86372 L p F P
0 g
s
.491 .635 .918 r
.21491 .84485 m .26429 .86239 L .26995 .8933 L p F P
0 g
s
.64 .745 .925 r
.3366 .88869 m .26995 .8933 L .26429 .86239 L p F P
0 g
s
.485 .598 .891 r
.52702 .82101 m .61074 .85199 L .60941 .86441 L p F P
0 g
s
.441 .605 .916 r
.61074 .85199 m .52702 .82101 L .52648 .81915 L p F P
0 g
s
.24 .452 .865 r
.52702 .82101 m .57619 .85499 L .57397 .847 L p F P
0 g
s
.24 .452 .865 r
.57397 .847 m .52648 .81915 L .52702 .82101 L p F P
0 g
s
.73 .763 .88 r
.59848 .74386 m .6692 .7058 L .64799 .68632 L p F P
0 g
s
.73 .763 .88 r
.64799 .68632 m .58664 .7333 L .59848 .74386 L p F P
0 g
s
.766 .916 .939 r
.42393 .82891 m .499 .81753 L .49785 .8193 L p F P
0 g
s
.766 .916 .939 r
.49785 .8193 m .41649 .84066 L .42393 .82891 L p F P
0 g
s
.805 .994 .869 r
.45012 .84764 m .49785 .8193 L .499 .81753 L p F P
0 g
s
.866 .957 .864 r
.49785 .8193 m .45012 .84764 L .44815 .85566 L p F P
0 g
s
.441 .605 .916 r
.52648 .81915 m .60744 .83966 L .61074 .85199 L p F P
0 g
s
.395 .618 .941 r
.60744 .83966 m .52648 .81915 L .52527 .81739 L p F P
0 g
s
.094 .422 .868 r
.52648 .81915 m .57397 .847 L .56884 .83945 L p F P
0 g
s
.094 .422 .868 r
.56884 .83945 m .52527 .81739 L .52648 .81915 L p F P
0 g
s
.665 .765 .925 r
.75534 .89079 m .6884 .88687 L .69134 .86445 L p F P
0 g
s
.665 .765 .925 r
.69134 .86445 m .76004 .85981 L .75534 .89079 L p F P
0 g
s
.667 .761 .92 r
.35378 .70751 m .29183 .69028 L .27449 .72067 L p F P
0 g
s
.751 .873 .936 r
.22515 .72867 m .27449 .72067 L .29183 .69028 L p F P
0 g
s
.72 .933 .959 r
.4356 .81834 m .50078 .81595 L .499 .81753 L p F P
0 g
s
.72 .933 .959 r
.499 .81753 m .42393 .82891 L .4356 .81834 L p F P
0 g
s
.663 .972 .83 r
.45501 .84005 m .499 .81753 L .50078 .81595 L p F P
0 g
s
.805 .994 .869 r
.499 .81753 m .45501 .84005 L .45012 .84764 L p F P
0 g
s
.395 .618 .941 r
.52527 .81739 m .59963 .82798 L .60744 .83966 L p F P
0 g
s
.349 .635 .96 r
.59963 .82798 m .52527 .81739 L .52344 .81583 L p F P
0 g
s
0 .378 .824 r
.52527 .81739 m .56884 .83945 L .56101 .83272 L p F P
0 g
s
0 .378 .824 r
.56101 .83272 m .52344 .81583 L .52527 .81739 L p F P
0 g
s
.906 .846 .774 r
.39106 .95415 m .45313 .87144 L .44918 .86372 L p F P
0 g
s
.876 .783 .755 r
.45313 .87144 m .39106 .95415 L .40501 .96818 L p F P
0 g
s
.372 .414 .775 r
.57169 .87082 m .63423 .95292 L .64218 .93735 L p F P
0 g
s
.372 .414 .775 r
.64218 .93735 m .57541 .86306 L .57169 .87082 L p F P
0 g
s
.65 .931 .971 r
.45096 .80948 m .50309 .81463 L .50078 .81595 L p F P
0 g
s
.65 .931 .971 r
.50078 .81595 m .4356 .81834 L .45096 .80948 L p F P
0 g
s
.44 .863 .731 r
.46262 .83323 m .50078 .81595 L .50309 .81463 L p F P
0 g
s
.663 .972 .83 r
.50078 .81595 m .46262 .83323 L .45501 .84005 L p F P
0 g
s
.349 .635 .96 r
.52344 .81583 m .58764 .81754 L .59963 .82798 L p F P
0 g
s
.309 .657 .972 r
.58764 .81754 m .52344 .81583 L .52109 .81453 L p F P
0 g
s
0 .327 .721 r
.52344 .81583 m .56101 .83272 L .55085 .82712 L p F P
0 g
s
0 .327 .721 r
.55085 .82712 m .52109 .81453 L .52344 .81583 L p F P
0 g
s
.481 .656 .938 r
.7489 .7181 m .79807 .72557 L .77647 .68909 L p F P
0 g
s
.481 .656 .938 r
.77647 .68909 m .73062 .68788 L .7489 .7181 L p F P
0 g
s
.837 .626 .617 r
.40872 1.19029 m .42203 1.09222 L .39077 1.07972 L p F P
0 g
s
.79 .581 .617 r
.42203 1.09222 m .40872 1.19029 L .44907 1.19979 L p F P
0 g
s
.561 .906 .973 r
.4693 .80278 m .50583 .81363 L .50309 .81463 L p F P
0 g
s
.561 .906 .973 r
.50309 .81463 m .45096 .80948 L .4693 .80278 L p F P
0 g
s
.44 .863 .731 r
.50309 .81463 m .4726 .82753 L .46262 .83323 L p F P
0 g
s
0 0 0 r
.4726 .82753 m .50309 .81463 L .50583 .81363 L p F P
0 g
s
.309 .657 .972 r
.52109 .81453 m .572 .80884 L .58764 .81754 L p F P
0 g
s
.285 .685 .973 r
.572 .80884 m .52109 .81453 L .51832 .81357 L p F P
0 g
s
.308 0 0 r
.52109 .81453 m .55085 .82712 L .53884 .82294 L p F P
0 g
s
.308 0 0 r
.53884 .82294 m .51832 .81357 L .52109 .81453 L p F P
0 g
s
.532 .377 .617 r
.60511 1.09134 m .61881 1.18931 L .65432 1.17507 L p F P
0 g
s
.532 .377 .617 r
.65432 1.17507 m .63598 1.07853 L .60511 1.09134 L p F P
0 g
s
.466 .864 .968 r
.48974 .79856 m .50887 .81301 L .50583 .81363 L p F P
0 g
s
.466 .864 .968 r
.50583 .81363 m .4693 .80278 L .48974 .79856 L p F P
0 g
s
.02 0 0 r
.48448 .82322 m .50583 .81363 L .50887 .81301 L p F P
0 g
s
0 0 0 r
.50583 .81363 m .48448 .82322 L .4726 .82753 L p F P
0 g
s
.285 .685 .973 r
.51832 .81357 m .55344 .80233 L .572 .80884 L p F P
0 g
s
.287 .72 .967 r
.55344 .80233 m .51832 .81357 L .51527 .81298 L p F P
0 g
s
.321 0 0 r
.51832 .81357 m .53884 .82294 L .52555 .82037 L p F P
0 g
s
.321 0 0 r
.52555 .82037 m .51527 .81298 L .51832 .81357 L p F P
0 g
s
.731 .807 .912 r
.33296 .86631 m .41358 .85302 L .41529 .86542 L p F P
0 g
s
.731 .807 .912 r
.41529 .86542 m .3366 .88869 L .33296 .86631 L p F P
0 g
s
.64 .745 .925 r
.26429 .86239 m .33296 .86631 L .3366 .88869 L p F P
0 g
s
.381 .813 .963 r
.51128 .79705 m .51206 .81279 L .50887 .81301 L p F P
0 g
s
.381 .813 .963 r
.50887 .81301 m .48974 .79856 L .51128 .79705 L p F P
0 g
s
.174 0 0 r
.49769 .82052 m .50887 .81301 L .51206 .81279 L p F P
0 g
s
.02 0 0 r
.50887 .81301 m .49769 .82052 L .48448 .82322 L p F P
0 g
s
.287 .72 .967 r
.51527 .81298 m .53287 .79833 L .55344 .80233 L p F P
0 g
s
.318 .763 .962 r
.53287 .79833 m .51527 .81298 L .51206 .81279 L p F P
0 g
s
.274 0 0 r
.51527 .81298 m .52555 .82037 L .5116 .81955 L p F P
0 g
s
.274 0 0 r
.5116 .81955 m .51206 .81279 L .51527 .81298 L p F P
0 g
s
.318 .763 .962 r
.51206 .81279 m .51128 .79705 L .53287 .79833 L p F P
0 g
s
.174 0 0 r
.51206 .81279 m .5116 .81955 L .49769 .82052 L p F P
0 g
s
.576 .683 .913 r
.6884 .88687 m .60941 .86441 L .61074 .85199 L p F P
0 g
s
.576 .683 .913 r
.61074 .85199 m .69134 .86445 L .6884 .88687 L p F P
0 g
s
.634 .675 .871 r
.37437 .68782 m .43663 .73411 L .45179 .72515 L p F P
0 g
s
.618 .673 .88 r
.43663 .73411 m .37437 .68782 L .35378 .70751 L p F P
0 g
s
.723 .747 .871 r
.58664 .7333 m .64799 .68632 L .62006 .67002 L p F P
0 g
s
.723 .747 .871 r
.62006 .67002 m .5712 .7245 L .58664 .7333 L p F P
0 g
s
.91 .77 .693 r
.3648 1.06377 m .40501 .96818 L .39106 .95415 L p F P
0 g
s
.866 .709 .685 r
.40501 .96818 m .3648 1.06377 L .39077 1.07972 L p F P
0 g
s
0 0 .065 r
.19814 .74947 m .19671 .77775 L .18985 .81893 L p F P
0 g
s
.279 .383 .791 r
.21491 .84485 m .18985 .81893 L .19671 .77775 L p F P
0 g
s
.402 .357 .694 r
.62072 .96709 m .66144 1.06232 L .6803 1.04337 L p F P
0 g
s
.402 .357 .694 r
.6803 1.04337 m .63423 .95292 L .62072 .96709 L p F P
0 g
s
.923 .918 .791 r
.38263 .93867 m .44918 .86372 L .44815 .85566 L p F P
0 g
s
.906 .846 .774 r
.44918 .86372 m .38263 .93867 L .39106 .95415 L p F P
0 g
s
.246 .37 .793 r
.57541 .86306 m .64218 .93735 L .64411 .92109 L p F P
0 g
s
.246 .37 .793 r
.64411 .92109 m .57619 .85499 L .57541 .86306 L p F P
0 g
s
.665 .762 .922 r
.29183 .69028 m .35378 .70751 L .37437 .68782 L p F P
0 g
s
.644 .745 .922 r
.6692 .7058 m .73062 .68788 L .70154 .66058 L p F P
0 g
s
.644 .745 .922 r
.70154 .66058 m .64799 .68632 L .6692 .7058 L p F P
0 g
s
.921 .902 .79 r
.83455 .81553 m .80946 .84174 L .80168 .80412 L p F P
0 g
s
.921 .902 .79 r
.80168 .80412 m .82641 .77439 L .83455 .81553 L p F P
0 g
s
.649 .679 .865 r
.40178 .67122 m .45179 .72515 L .46989 .71837 L p F P
0 g
s
.634 .675 .871 r
.45179 .72515 m .40178 .67122 L .37437 .68782 L p F P
0 g
s
.715 .732 .864 r
.5712 .7245 m .62006 .67002 L .58672 .65778 L p F P
0 g
s
.715 .732 .864 r
.58672 .65778 m .5529 .71792 L .5712 .7245 L p F P
0 g
s
.751 .873 .936 r
.29183 .69028 m .2456 .69199 L .22515 .72867 L p F P
0 g
s
.829 1 .81 r
.19814 .74947 m .22515 .72867 L .2456 .69199 L p F P
0 g
s
.887 .679 .615 r
.37277 1.1764 m .39077 1.07972 L .3648 1.06377 L p F P
0 g
s
.837 .626 .617 r
.39077 1.07972 m .37277 1.1764 L .40872 1.19029 L p F P
0 g
s
.9 .982 .793 r
.38018 .92244 m .44815 .85566 L .45012 .84764 L p F P
0 g
s
.923 .918 .791 r
.44815 .85566 m .38018 .92244 L .38263 .93867 L p F P
0 g
s
.723 .822 .926 r
.33772 .84386 m .41649 .84066 L .41358 .85302 L p F P
0 g
s
.723 .822 .926 r
.41358 .85302 m .33296 .86631 L .33772 .84386 L p F P
0 g
s
.447 .32 .615 r
.63598 1.07853 m .65432 1.17507 L .68369 1.15702 L p F P
0 g
s
.447 .32 .615 r
.68369 1.15702 m .66144 1.06232 L .63598 1.07853 L p F P
0 g
s
.063 .306 .791 r
.57619 .85499 m .64411 .92109 L .63986 .9049 L p F P
0 g
s
.063 .306 .791 r
.63986 .9049 m .57397 .847 L .57619 .85499 L p F P
0 g
s
.561 .691 .927 r
.69134 .86445 m .61074 .85199 L .60744 .83966 L p F P
0 g
s
.741 .488 .547 r
.49187 1.2906 m .49207 1.20452 L .44907 1.19979 L p F P
0 g
s
.69 .446 .548 r
.49207 1.20452 m .49187 1.2906 L .53613 1.2904 L p F P
0 g
s
.69 .446 .548 r
.53613 1.2904 m .53592 1.20432 L .49207 1.20452 L p F P
0 g
s
.635 .401 .546 r
.53592 1.20432 m .53613 1.2904 L .57937 1.28536 L p F P
0 g
s
.635 .401 .546 r
.57937 1.28536 m .57876 1.19918 L .53592 1.20432 L p F P
0 g
s
0 .317 .807 r
.79807 .72557 m .82499 .7461 L .80165 .70638 L p F P
0 g
s
0 .317 .807 r
.80165 .70638 m .77647 .68909 L .79807 .72557 L p F P
0 g
s
.279 .383 .791 r
.19671 .77775 m .22151 .80719 L .21491 .84485 L p F P
0 g
s
.525 .629 .896 r
.26429 .86239 m .21491 .84485 L .22151 .80719 L p F P
0 g
s
.665 .686 .86 r
.43473 .65861 m .46989 .71837 L .49005 .71411 L p F P
0 g
s
.649 .679 .865 r
.46989 .71837 m .43473 .65861 L .40178 .67122 L p F P
0 g
s
.704 .718 .86 r
.5529 .71792 m .58672 .65778 L .54959 .65024 L p F P
0 g
s
.704 .718 .86 r
.54959 .65024 m .5326 .71388 L .5529 .71792 L p F P
0 g
s
.952 .843 .699 r
.34535 1.04502 m .39106 .95415 L .38263 .93867 L p F P
0 g
s
.91 .77 .693 r
.39106 .95415 m .34535 1.04502 L .3648 1.06377 L p F P
0 g
s
.271 .291 .699 r
.63423 .95292 m .6803 1.04337 L .69157 1.02249 L p F P
0 g
s
.271 .291 .699 r
.69157 1.02249 m .64218 .93735 L .63423 .95292 L p F P
0 g
s
.792 .53 .542 r
.44847 1.28596 m .44907 1.19979 L .40872 1.19029 L p F P
0 g
s
.741 .488 .547 r
.44907 1.19979 m .44847 1.28596 L .49187 1.2906 L p F P
0 g
s
.778 .833 .895 r
.80946 .84174 m .76004 .85981 L .75311 .82868 L p F P
0 g
s
.778 .833 .895 r
.75311 .82868 m .80168 .80412 L .80946 .84174 L p F P
0 g
s
.571 .35 .541 r
.57876 1.19918 m .57937 1.28536 L .61979 1.27568 L p F P
0 g
s
.571 .35 .541 r
.61979 1.27568 m .61881 1.18931 L .57876 1.19918 L p F P
0 g
s
.778 .986 .738 r
.38391 .90621 m .45012 .84764 L .45501 .84005 L p F P
0 g
s
.9 .982 .793 r
.45012 .84764 m .38391 .90621 L .38018 .92244 L p F P
0 g
s
0 .21 .729 r
.57397 .847 m .63986 .9049 L .62953 .88954 L p F P
0 g
s
0 .21 .729 r
.62953 .88954 m .56884 .83945 L .57397 .847 L p F P
0 g
s
.679 .694 .858 r
.47161 .65067 m .49005 .71411 L .51131 .71258 L p F P
0 g
s
.665 .686 .86 r
.49005 .71411 m .47161 .65067 L .43473 .65861 L p F P
0 g
s
.692 .705 .858 r
.5326 .71388 m .54959 .65024 L .51056 .64782 L p F P
0 g
s
.692 .705 .858 r
.51056 .64782 m .51131 .71258 L .5326 .71388 L p F P
0 g
s
.642 .743 .922 r
.27024 .83122 m .33772 .84386 L .33296 .86631 L p F P
0 g
s
.642 .743 .922 r
.33296 .86631 m .26429 .86239 L .27024 .83122 L p F P
0 g
s
.679 .694 .858 r
.51131 .71258 m .51056 .64782 L .47161 .65067 L p F P
0 g
s
.525 .629 .896 r
.22151 .80719 m .27024 .83122 L .26429 .86239 L p F P
0 g
s
.662 .762 .924 r
.32005 .66269 m .37437 .68782 L .40178 .67122 L p F P
0 g
s
.665 .762 .922 r
.37437 .68782 m .32005 .66269 L .29183 .69028 L p F P
0 g
s
.667 .763 .922 r
.76004 .85981 m .69134 .86445 L .68588 .84204 L p F P
0 g
s
.561 .691 .927 r
.60744 .83966 m .68588 .84204 L .69134 .86445 L p F P
0 g
s
.721 .888 .957 r
.2456 .69199 m .29183 .69028 L .32005 .66269 L p F P
0 g
s
.711 .833 .94 r
.35082 .8224 m .42393 .82891 L .41649 .84066 L p F P
0 g
s
.711 .833 .94 r
.41649 .84066 m .33772 .84386 L .35082 .8224 L p F P
0 g
s
.644 .748 .924 r
.64799 .68632 m .70154 .66058 L .66287 .63765 L p F P
0 g
s
.644 .748 .924 r
.66287 .63765 m .62006 .67002 L .64799 .68632 L p F P
0 g
s
.55 .704 .942 r
.68588 .84204 m .60744 .83966 L .59963 .82798 L p F P
0 g
s
.667 .763 .922 r
.68588 .84204 m .75311 .82868 L .76004 .85981 L p F P
0 g
s
.459 .676 .958 r
.73062 .68788 m .77647 .68909 L .74166 .65602 L p F P
0 g
s
.459 .676 .958 r
.74166 .65602 m .70154 .66058 L .73062 .68788 L p F P
0 g
s
.51 .852 .584 r
.39374 .89075 m .45501 .84005 L .46262 .83323 L p F P
0 g
s
.778 .986 .738 r
.45501 .84005 m .39374 .89075 L .38391 .90621 L p F P
0 g
s
0 .09 .566 r
.56884 .83945 m .62953 .88954 L .61354 .87576 L p F P
0 g
s
0 .09 .566 r
.61354 .87576 m .56101 .83272 L .56884 .83945 L p F P
0 g
s
.94 .742 .608 r
.34282 1.15863 m .3648 1.06377 L .34535 1.04502 L p F P
0 g
s
.887 .679 .615 r
.3648 1.06377 m .34282 1.15863 L .37277 1.1764 L p F P
0 g
s
.829 1 .81 r
.2456 .69199 m .22022 .70954 L .19814 .74947 L p F P
0 g
s
0 0 .313 r
.19671 .77775 m .19814 .74947 L .22022 .70954 L p F P
0 g
s
.844 .576 .533 r
.40775 1.27665 m .40872 1.19029 L .37277 1.1764 L p F P
0 g
s
.792 .53 .542 r
.40872 1.19029 m .40775 1.27665 L .44847 1.28596 L p F P
0 g
s
.325 .242 .607 r
.66144 1.06232 m .68369 1.15702 L .70554 1.13588 L p F P
0 g
s
.325 .242 .607 r
.70554 1.13588 m .6803 1.04337 L .66144 1.06232 L p F P
0 g
s
.491 .288 .532 r
.61881 1.18931 m .61979 1.27568 L .65564 1.26173 L p F P
0 g
s
.491 .288 .532 r
.65564 1.26173 m .65432 1.17507 L .61881 1.18931 L p F P
0 g
s
.981 .697 .327 r
.82499 .7461 m .82641 .77439 L .80298 .73482 L p F P
0 g
s
.981 .697 .327 r
.80298 .73482 m .80165 .70638 L .82499 .7461 L p F P
0 g
s
.971 .922 .688 r
.33341 1.02426 m .38263 .93867 L .38018 .92244 L p F P
0 g
s
.952 .843 .699 r
.38263 .93867 m .33341 1.02426 L .34535 1.04502 L p F P
0 g
s
.51 .852 .584 r
.46262 .83323 m .40928 .87682 L .39374 .89075 L p F P
0 g
s
0 0 0 r
.40928 .87682 m .46262 .83323 L .4726 .82753 L p F P
0 g
s
.068 .191 .685 r
.64218 .93735 m .69157 1.02249 L .69459 1.00059 L p F P
0 g
s
.068 .191 .685 r
.69459 1.00059 m .64411 .92109 L .64218 .93735 L p F P
0 g
s
.574 .004 0 r
.56101 .83272 m .61354 .87576 L .5926 .86427 L p F P
0 g
s
.574 .004 0 r
.5926 .86427 m .55085 .82712 L .56101 .83272 L p F P
0 g
s
.692 .84 .954 r
.37178 .80299 m .4356 .81834 L .42393 .82891 L p F P
0 g
s
.692 .84 .954 r
.42393 .82891 m .35082 .8224 L .37178 .80299 L p F P
0 g
s
.55 .704 .942 r
.59963 .82798 m .6721 .8207 L .68588 .84204 L p F P
0 g
s
.541 .719 .955 r
.6721 .8207 m .59963 .82798 L .58764 .81754 L p F P
0 g
s
.659 .761 .926 r
.35797 .63935 m .40178 .67122 L .43473 .65861 L p F P
0 g
s
.662 .762 .924 r
.40178 .67122 m .35797 .63935 L .32005 .66269 L p F P
0 g
s
.099 0 0 r
.42985 .86512 m .4726 .82753 L .48448 .82322 L p F P
0 g
s
0 0 0 r
.4726 .82753 m .42985 .86512 L .40928 .87682 L p F P
0 g
s
.646 .751 .926 r
.62006 .67002 m .66287 .63765 L .61643 .62036 L p F P
0 g
s
.646 .751 .926 r
.61643 .62036 m .58672 .65778 L .62006 .67002 L p F P
0 g
s
.609 .044 0 r
.55085 .82712 m .5926 .86427 L .56769 .85565 L p F P
0 g
s
.609 .044 0 r
.56769 .85565 m .53884 .82294 L .55085 .82712 L p F P
0 g
s
0 0 .313 r
.22022 .70954 m .21887 .73797 L .19671 .77775 L p F P
0 g
s
.399 .427 .774 r
.22151 .80719 m .19671 .77775 L .21887 .73797 L p F P
0 g
s
.291 0 0 r
.45447 .85624 m .48448 .82322 L .49769 .82052 L p F P
0 g
s
.099 0 0 r
.48448 .82322 m .45447 .85624 L .42985 .86512 L p F P
0 g
s
.581 .038 0 r
.53884 .82294 m .56769 .85565 L .54002 .85035 L p F P
0 g
s
.581 .038 0 r
.54002 .85035 m .52555 .82037 L .53884 .82294 L p F P
0 g
s
.901 .628 .514 r
.37145 1.26303 m .37277 1.1764 L .34282 1.15863 L p F P
0 g
s
.844 .576 .533 r
.37277 1.1764 m .37145 1.26303 L .40775 1.27665 L p F P
0 g
s
.644 .742 .92 r
.28785 .80127 m .35082 .8224 L .33772 .84386 L p F P
0 g
s
.644 .742 .92 r
.33772 .84386 m .27024 .83122 L .28785 .80127 L p F P
0 g
s
.381 .203 .512 r
.65432 1.17507 m .65564 1.26173 L .6853 1.24403 L p F P
0 g
s
.381 .203 .512 r
.6853 1.24403 m .68369 1.15702 L .65432 1.17507 L p F P
0 g
s
.721 .888 .957 r
.32005 .66269 m .27935 .65857 L .2456 .69199 L p F P
0 g
s
.643 .949 .731 r
.22022 .70954 m .2456 .69199 L .27935 .65857 L p F P
0 g
s
.669 .842 .966 r
.39969 .78665 m .45096 .80948 L .4356 .81834 L p F P
0 g
s
.669 .842 .966 r
.4356 .81834 m .37178 .80299 L .39969 .78665 L p F P
0 g
s
.668 .761 .92 r
.75311 .82868 m .68588 .84204 L .6721 .8207 L p F P
0 g
s
.897 .831 .772 r
.82641 .77439 m .80168 .80412 L .77984 .76801 L p F P
0 g
s
.897 .831 .772 r
.77984 .76801 m .80298 .73482 L .82641 .77439 L p F P
0 g
s
.424 0 0 r
.48196 .85066 m .49769 .82052 L .5116 .81955 L p F P
0 g
s
.291 0 0 r
.49769 .82052 m .48196 .85066 L .45447 .85624 L p F P
0 g
s
.517 0 0 r
.52555 .82037 m .54002 .85035 L .51096 .84866 L p F P
0 g
s
.517 0 0 r
.51096 .84866 m .5116 .81955 L .52555 .82037 L p F P
0 g
s
.541 .719 .955 r
.58764 .81754 m .65051 .80152 L .6721 .8207 L p F P
0 g
s
.538 .737 .967 r
.65051 .80152 m .58764 .81754 L .572 .80884 L p F P
0 g
s
.424 0 0 r
.5116 .81955 m .51096 .84866 L .48196 .85066 L p F P
0 g
s
.988 .816 .583 r
.3203 1.13772 m .34535 1.04502 L .33341 1.02426 L p F P
0 g
s
.94 .742 .608 r
.34535 1.04502 m .3203 1.13772 L .34282 1.15863 L p F P
0 g
s
0 .237 .72 r
.77647 .68909 m .80165 .70638 L .76374 .6703 L p F P
0 g
s
0 .237 .72 r
.76374 .6703 m .74166 .65602 L .77647 .68909 L p F P
0 g
s
.897 .956 .615 r
.32968 1.00241 m .38018 .92244 L .38391 .90621 L p F P
0 g
s
.971 .922 .688 r
.38018 .92244 m .32968 1.00241 L .33341 1.02426 L p F P
0 g
s
.134 .123 .579 r
.6803 1.04337 m .70554 1.13588 L .71871 1.11254 L p F P
0 g
s
.134 .123 .579 r
.71871 1.11254 m .69157 1.02249 L .6803 1.04337 L p F P
0 g
s
.68 .892 .975 r
.27935 .65857 m .32005 .66269 L .35797 .63935 L p F P
0 g
s
0 .033 .603 r
.64411 .92109 m .69459 1.00059 L .68905 .9787 L p F P
0 g
s
0 .033 .603 r
.68905 .9787 m .63986 .9049 L .64411 .92109 L p F P
0 g
s
.655 .76 .927 r
.40384 .62155 m .43473 .65861 L .47161 .65067 L p F P
0 g
s
.659 .761 .926 r
.43473 .65861 m .40384 .62155 L .35797 .63935 L p F P
0 g
s
.647 .753 .927 r
.58672 .65778 m .61643 .62036 L .56455 .6097 L p F P
0 g
s
.647 .753 .927 r
.56455 .6097 m .54959 .65024 L .58672 .65778 L p F P
0 g
s
.443 .702 .976 r
.70154 .66058 m .74166 .65602 L .69505 .62816 L p F P
0 g
s
.443 .702 .976 r
.69505 .62816 m .66287 .63765 L .70154 .66058 L p F P
0 g
s
.399 .427 .774 r
.21887 .73797 m .24218 .77089 L .22151 .80719 L p F P
0 g
s
.556 .626 .877 r
.27024 .83122 m .22151 .80719 L .24218 .77089 L p F P
0 g
s
.556 .626 .877 r
.24218 .77089 m .28785 .80127 L .27024 .83122 L p F P
0 g
s
.643 .839 .976 r
.43326 .77422 m .4693 .80278 L .45096 .80948 L p F P
0 g
s
.643 .839 .976 r
.45096 .80948 m .39969 .78665 L .43326 .77422 L p F P
0 g
s
.538 .737 .967 r
.572 .80884 m .62207 .78546 L .65051 .80152 L p F P
0 g
s
.54 .756 .976 r
.62207 .78546 m .572 .80884 L .55344 .80233 L p F P
0 g
s
.775 .803 .875 r
.80168 .80412 m .75311 .82868 L .73454 .7989 L p F P
0 g
s
.775 .803 .875 r
.73454 .7989 m .77984 .76801 L .80168 .80412 L p F P
0 g
s
.668 .761 .92 r
.6721 .8207 m .73454 .7989 L .75311 .82868 L p F P
0 g
s
.725 .405 .437 r
.4941 1.35653 m .49187 1.2906 L .44847 1.28596 L p F P
0 g
s
.667 .358 .441 r
.49187 1.2906 m .4941 1.35653 L .53348 1.35636 L p F P
0 g
s
.667 .358 .441 r
.53348 1.35636 m .53613 1.2904 L .49187 1.2906 L p F P
0 g
s
.602 .304 .436 r
.53613 1.2904 m .53348 1.35636 L .57195 1.35198 L p F P
0 g
s
.602 .304 .436 r
.57195 1.35198 m .57937 1.28536 L .53613 1.2904 L p F P
0 g
s
.652 .758 .927 r
.45537 .61031 m .47161 .65067 L .51056 .64782 L p F P
0 g
s
.655 .76 .927 r
.47161 .65067 m .45537 .61031 L .40384 .62155 L p F P
0 g
s
.65 .756 .927 r
.54959 .65024 m .56455 .6097 L .5099 .60628 L p F P
0 g
s
.65 .756 .927 r
.5099 .60628 m .51056 .64782 L .54959 .65024 L p F P
0 g
s
.652 .758 .927 r
.51056 .64782 m .5099 .60628 L .45537 .61031 L p F P
0 g
s
.615 .829 .982 r
.47083 .76639 m .48974 .79856 L .4693 .80278 L p F P
0 g
s
.615 .829 .982 r
.4693 .80278 m .43326 .77422 L .47083 .76639 L p F P
0 g
s
.781 .448 .423 r
.45548 1.3525 m .44847 1.28596 L .40775 1.27665 L p F P
0 g
s
.725 .405 .437 r
.44847 1.28596 m .45548 1.3525 L .4941 1.35653 L p F P
0 g
s
.54 .756 .976 r
.55344 .80233 m .58811 .77339 L .62207 .78546 L p F P
0 g
s
.55 .777 .983 r
.58811 .77339 m .55344 .80233 L .53287 .79833 L p F P
0 g
s
.524 .237 .421 r
.57937 1.28536 m .57195 1.35198 L .60788 1.34357 L p F P
0 g
s
.524 .237 .421 r
.60788 1.34357 m .61979 1.27568 L .57937 1.28536 L p F P
0 g
s
.959 .685 .475 r
.34121 1.24561 m .34282 1.15863 L .3203 1.13772 L p F P
0 g
s
.901 .628 .514 r
.34282 1.15863 m .34121 1.24561 L .37145 1.26303 L p F P
0 g
s
.589 .815 .986 r
.51051 .76359 m .51128 .79705 L .48974 .79856 L p F P
0 g
s
.589 .815 .986 r
.48974 .79856 m .47083 .76639 L .51051 .76359 L p F P
0 g
s
.647 .741 .918 r
.31655 .77406 m .37178 .80299 L .35082 .8224 L p F P
0 g
s
.647 .741 .918 r
.35082 .8224 m .28785 .80127 L .31655 .77406 L p F P
0 g
s
.212 .077 .47 r
.68369 1.15702 m .6853 1.24403 L .70736 1.2233 L p F P
0 g
s
.212 .077 .47 r
.70736 1.2233 m .70554 1.13588 L .68369 1.15702 L p F P
0 g
s
.616 .819 .411 r
.3345 .98048 m .38391 .90621 L .39374 .89075 L p F P
0 g
s
.897 .956 .615 r
.38391 .90621 m .3345 .98048 L .32968 1.00241 L p F P
0 g
s
.55 .777 .983 r
.53287 .79833 m .55028 .76596 L .58811 .77339 L p F P
0 g
s
.566 .797 .986 r
.55028 .76596 m .53287 .79833 L .51128 .79705 L p F P
0 g
s
.643 .949 .731 r
.27935 .65857 m .25695 .67309 L .22022 .70954 L p F P
0 g
s
.201 .043 .431 r
.21887 .73797 m .22022 .70954 L .25695 .67309 L p F P
0 g
s
.566 .797 .986 r
.51128 .79705 m .51051 .76359 L .55028 .76596 L p F P
0 g
s
.593 .163 0 r
.63986 .9049 m .68905 .9787 L .67503 .95786 L p F P
0 g
s
.593 .163 0 r
.67503 .95786 m .62953 .88954 L .63986 .9049 L p F P
0 g
s
.668 .758 .918 r
.73454 .7989 m .6721 .8207 L .65051 .80152 L p F P
0 g
s
.952 .655 .437 r
.80165 .70638 m .80298 .73482 L .76491 .69887 L p F P
0 g
s
.952 .655 .437 r
.76491 .69887 m .76374 .6703 L .80165 .70638 L p F P
0 g
s
.981 .872 .503 r
.30636 1.11453 m .33341 1.02426 L .32968 1.00241 L p F P
0 g
s
.988 .816 .583 r
.33341 1.02426 m .30636 1.11453 L .3203 1.13772 L p F P
0 g
s
0 0 .491 r
.69157 1.02249 m .71871 1.11254 L .72241 1.08803 L p F P
0 g
s
0 0 .491 r
.72241 1.08803 m .69459 1.00059 L .69157 1.02249 L p F P
0 g
s
.628 .884 .988 r
.32505 .63022 m .35797 .63935 L .40384 .62155 L p F P
0 g
s
.68 .892 .975 r
.35797 .63935 m .32505 .63022 L .27935 .65857 L p F P
0 g
s
.837 .489 .395 r
.41928 1.34441 m .40775 1.27665 L .37145 1.26303 L p F P
0 g
s
.781 .448 .423 r
.40775 1.27665 m .41928 1.34441 L .45548 1.3525 L p F P
0 g
s
.42 .147 .391 r
.61979 1.27568 m .60788 1.34357 L .63969 1.33145 L p F P
0 g
s
.42 .147 .391 r
.63969 1.33145 m .65564 1.26173 L .61979 1.27568 L p F P
0 g
s
.439 .732 .989 r
.66287 .63765 m .69505 .62816 L .63882 .60709 L p F P
0 g
s
.439 .732 .989 r
.63882 .60709 m .61643 .62036 L .66287 .63765 L p F P
0 g
s
0 0 0 r
.25695 .67309 m .27935 .65857 L .32505 .63022 L p F P
0 g
s
.412 0 0 r
.74166 .65602 m .76374 .6703 L .7128 .63986 L p F P
0 g
s
.412 0 0 r
.7128 .63986 m .69505 .62816 L .74166 .65602 L p F P
0 g
s
.616 .819 .411 r
.39374 .89075 m .34782 .95952 L .3345 .98048 L p F P
0 g
s
0 0 0 r
.34782 .95952 m .39374 .89075 L .40928 .87682 L p F P
0 g
s
.201 .043 .431 r
.25695 .67309 m .25576 .70164 L .21887 .73797 L p F P
0 g
s
.485 .461 .756 r
.24218 .77089 m .21887 .73797 L .25576 .70164 L p F P
0 g
s
.782 .292 0 r
.62953 .88954 m .67503 .95786 L .65307 .93911 L p F P
0 g
s
.782 .292 0 r
.65307 .93911 m .61354 .87576 L .62953 .88954 L p F P
0 g
s
.586 .628 .86 r
.27634 .7378 m .31655 .77406 L .28785 .80127 L p F P
0 g
s
.586 .628 .86 r
.28785 .80127 m .24218 .77089 L .27634 .7378 L p F P
0 g
s
.65 .741 .916 r
.35514 .75105 m .39969 .78665 L .37178 .80299 L p F P
0 g
s
.65 .741 .916 r
.37178 .80299 m .31655 .77406 L .35514 .75105 L p F P
0 g
s
.864 .769 .755 r
.80298 .73482 m .77984 .76801 L .74461 .73527 L p F P
0 g
s
.864 .769 .755 r
.74461 .73527 m .76491 .69887 L .80298 .73482 L p F P
0 g
s
.767 .775 .859 r
.77984 .76801 m .73454 .7989 L .70496 .77199 L p F P
0 g
s
.668 .758 .918 r
.65051 .80152 m .70496 .77199 L .73454 .7989 L p F P
0 g
s
.667 .755 .916 r
.70496 .77199 m .65051 .80152 L .62207 .78546 L p F P
0 g
s
.991 .735 .385 r
.31847 1.22511 m .3203 1.13772 L .30636 1.11453 L p F P
0 g
s
.959 .685 .475 r
.3203 1.13772 m .31847 1.22511 L .34121 1.24561 L p F P
0 g
s
.893 .527 .34 r
.38706 1.33258 m .37145 1.26303 L .34121 1.24561 L p F P
0 g
s
.837 .489 .395 r
.37145 1.26303 m .38706 1.33258 L .41928 1.34441 L p F P
0 g
s
.485 .461 .756 r
.25576 .70164 m .27634 .7378 L .24218 .77089 L p F P
0 g
s
.573 .864 .996 r
.38057 .60854 m .40384 .62155 L .45537 .61031 L p F P
0 g
s
.628 .884 .988 r
.40384 .62155 m .38057 .60854 L .32505 .63022 L p F P
0 g
s
0 0 .373 r
.70554 1.13588 m .70736 1.2233 L .72067 1.20041 L p F P
0 g
s
0 0 .373 r
.72067 1.20041 m .71871 1.11254 L .70554 1.13588 L p F P
0 g
s
.267 .015 .334 r
.65564 1.26173 m .63969 1.33145 L .66595 1.3161 L p F P
0 g
s
.267 .015 .334 r
.66595 1.3161 m .6853 1.24403 L .65564 1.26173 L p F P
0 g
s
.45 .765 .996 r
.61643 .62036 m .63882 .60709 L .57583 .59408 L p F P
0 g
s
.45 .765 .996 r
.57583 .59408 m .56455 .6097 L .61643 .62036 L p F P
0 g
s
.767 .775 .859 r
.70496 .77199 m .74461 .73527 L .77984 .76801 L p F P
0 g
s
.772 .787 .291 r
.30185 1.09007 m .32968 1.00241 L .3345 .98048 L p F P
0 g
s
.981 .872 .503 r
.32968 1.00241 m .30185 1.09007 L .30636 1.11453 L p F P
0 g
s
0 0 .266 r
.69459 1.00059 m .72241 1.08803 L .71623 1.06346 L p F P
0 g
s
0 0 .266 r
.71623 1.06346 m .68905 .9787 L .69459 1.00059 L p F P
0 g
s
.121 0 .058 r
.36915 .94056 m .40928 .87682 L .42985 .86512 L p F P
0 g
s
0 0 0 r
.40928 .87682 m .36915 .94056 L .34782 .95952 L p F P
0 g
s
.808 .329 .071 r
.61354 .87576 m .65307 .93911 L .62411 .92342 L p F P
0 g
s
.808 .329 .071 r
.62411 .92342 m .5926 .86427 L .61354 .87576 L p F P
0 g
s
.52 .835 .999 r
.44312 .59483 m .45537 .61031 L .5099 .60628 L p F P
0 g
s
.573 .864 .996 r
.45537 .61031 m .44312 .59483 L .38057 .60854 L p F P
0 g
s
.477 .8 1 r
.56455 .6097 m .57583 .59408 L .5094 .58991 L p F P
0 g
s
.477 .8 1 r
.5094 .58991 m .5099 .60628 L .56455 .6097 L p F P
0 g
s
.653 .742 .915 r
.40184 .73349 m .43326 .77422 L .39969 .78665 L p F P
0 g
s
.653 .742 .915 r
.39969 .78665 m .35514 .75105 L .40184 .73349 L p F P
0 g
s
0 0 0 r
.32505 .63022 m .30688 .64211 L .25695 .67309 L p F P
0 g
s
.383 .184 .482 r
.25576 .70164 m .25695 .67309 L .30688 .64211 L p F P
0 g
s
.667 .755 .916 r
.62207 .78546 m .66561 .74938 L .70496 .77199 L p F P
0 g
s
.666 .752 .915 r
.66561 .74938 m .62207 .78546 L .58811 .77339 L p F P
0 g
s
.52 .835 .999 r
.5099 .60628 m .5094 .58991 L .44312 .59483 L p F P
0 g
s
.891 .599 .485 r
.76374 .6703 m .76491 .69887 L .71373 .66853 L p F P
0 g
s
.891 .599 .485 r
.71373 .66853 m .7128 .63986 L .76374 .6703 L p F P
0 g
s
0 0 0 r
.30688 .64211 m .32505 .63022 L .38057 .60854 L p F P
0 g
s
.488 0 0 r
.69505 .62816 m .7128 .63986 L .6512 .61682 L p F P
0 g
s
.488 0 0 r
.6512 .61682 m .63882 .60709 L .69505 .62816 L p F P
0 g
s
.33 0 .17 r
.39757 .92458 m .42985 .86512 L .45447 .85624 L p F P
0 g
s
.121 0 .058 r
.42985 .86512 m .39757 .92458 L .36915 .94056 L p F P
0 g
s
.931 .551 .233 r
.36028 1.31747 m .34121 1.24561 L .31847 1.22511 L p F P
0 g
s
.893 .527 .34 r
.34121 1.24561 m .36028 1.31747 L .38706 1.33258 L p F P
0 g
s
.77 .319 .177 r
.5926 .86427 m .62411 .92342 L .58951 .91163 L p F P
0 g
s
.77 .319 .177 r
.58951 .91163 m .56769 .85565 L .5926 .86427 L p F P
0 g
s
.021 0 .22 r
.6853 1.24403 m .66595 1.3161 L .68542 1.29815 L p F P
0 g
s
.021 0 .22 r
.68542 1.29815 m .70736 1.2233 L .6853 1.24403 L p F P
0 g
s
.663 .231 .213 r
.49855 1.39648 m .4941 1.35653 L .45548 1.3525 L p F P
0 g
s
.596 .18 .224 r
.4941 1.35653 m .49855 1.39648 L .52819 1.39635 L p F P
0 g
s
.596 .18 .224 r
.52819 1.39635 m .53348 1.35636 L .4941 1.35653 L p F P
0 g
s
.613 .633 .847 r
.3226 .70972 m .35514 .75105 L .31655 .77406 L p F P
0 g
s
.613 .633 .847 r
.31655 .77406 m .27634 .7378 L .3226 .70972 L p F P
0 g
s
.514 .11 .211 r
.53348 1.35636 m .52819 1.39635 L .55714 1.39311 L p F P
0 g
s
.514 .11 .211 r
.55714 1.39311 m .57195 1.35198 L .53348 1.35636 L p F P
0 g
s
.755 .748 .846 r
.74461 .73527 m .70496 .77199 L .66561 .74938 L p F P
0 g
s
.883 .697 .178 r
.30438 1.20236 m .30636 1.11453 L .30185 1.09007 L p F P
0 g
s
.991 .735 .385 r
.30636 1.11453 m .30438 1.20236 L .31847 1.22511 L p F P
0 g
s
.656 .743 .914 r
.45431 .7224 m .47083 .76639 L .43326 .77422 L p F P
0 g
s
.656 .743 .914 r
.43326 .77422 m .40184 .73349 L .45431 .7224 L p F P
0 g
s
.666 .752 .915 r
.58811 .77339 m .61833 .73232 L .66561 .74938 L p F P
0 g
s
.664 .749 .914 r
.61833 .73232 m .58811 .77339 L .55028 .76596 L p F P
0 g
s
0 0 .153 r
.71871 1.11254 m .72067 1.20041 L .72442 1.17637 L p F P
0 g
s
0 0 .153 r
.72442 1.17637 m .72241 1.08803 L .71871 1.11254 L p F P
0 g
s
.719 .267 .176 r
.46949 1.39349 m .45548 1.3525 L .41928 1.34441 L p F P
0 g
s
.663 .231 .213 r
.45548 1.3525 m .46949 1.39349 L .49855 1.39648 L p F P
0 g
s
.407 .011 .171 r
.57195 1.35198 m .55714 1.39311 L .58412 1.38687 L p F P
0 g
s
.407 .011 .171 r
.58412 1.38687 m .60788 1.34357 L .57195 1.35198 L p F P
0 g
s
.772 .787 .291 r
.3345 .98048 m .30719 1.06546 L .30185 1.09007 L p F P
0 g
s
0 0 .013 r
.30719 1.06546 m .3345 .98048 L .34782 .95952 L p F P
0 g
s
.468 .085 .229 r
.43176 .91243 m .45447 .85624 L .48196 .85066 L p F P
0 g
s
.33 0 .17 r
.45447 .85624 m .43176 .91243 L .39757 .92458 L p F P
0 g
s
.383 .184 .482 r
.30688 .64211 m .30592 .67077 L .25576 .70164 L p F P
0 g
s
.55 .49 .742 r
.27634 .7378 m .25576 .70164 L .30592 .67077 L p F P
0 g
s
.713 .286 .232 r
.56769 .85565 m .58951 .91163 L .55098 .90436 L p F P
0 g
s
.713 .286 .232 r
.55098 .90436 m .54002 .85035 L .56769 .85565 L p F P
0 g
s
.883 .507 .036 r
.68905 .9787 m .71623 1.06346 L .70025 1.04003 L p F P
0 g
s
.883 .507 .036 r
.70025 1.04003 m .67503 .95786 L .68905 .9787 L p F P
0 g
s
.829 .717 .741 r
.76491 .69887 m .74461 .73527 L .69742 .70768 L p F P
0 g
s
.829 .717 .741 r
.69742 .70768 m .71373 .66853 L .76491 .69887 L p F P
0 g
s
.55 .49 .742 r
.30592 .67077 m .3226 .70972 L .27634 .7378 L p F P
0 g
s
.659 .745 .913 r
.50984 .71843 m .51051 .76359 L .47083 .76639 L p F P
0 g
s
.659 .745 .913 r
.47083 .76639 m .45431 .7224 L .50984 .71843 L p F P
0 g
s
.664 .749 .914 r
.55028 .76596 m .56549 .7218 L .61833 .73232 L p F P
0 g
s
.662 .747 .913 r
.56549 .7218 m .55028 .76596 L .51051 .76359 L p F P
0 g
s
.567 .173 .254 r
.47003 .90478 m .48196 .85066 L .51096 .84866 L p F P
0 g
s
.468 .085 .229 r
.48196 .85066 m .47003 .90478 L .43176 .91243 L p F P
0 g
s
.646 .238 .255 r
.54002 .85035 m .55098 .90436 L .51046 .90204 L p F P
0 g
s
.646 .238 .255 r
.51046 .90204 m .51096 .84866 L .54002 .85035 L p F P
0 g
s
.755 .748 .846 r
.66561 .74938 m .69742 .70768 L .74461 .73527 L p F P
0 g
s
.662 .747 .913 r
.51051 .76359 m .50984 .71843 L .56549 .7218 L p F P
0 g
s
.567 .173 .254 r
.51096 .84866 m .51046 .90204 L .47003 .90478 L p F P
0 g
s
.763 .284 .101 r
.4423 1.3875 m .41928 1.34441 L .38706 1.33258 L p F P
0 g
s
.719 .267 .176 r
.41928 1.34441 m .4423 1.3875 L .46949 1.39349 L p F P
0 g
s
.163 0 0 r
.36771 .6184 m .38057 .60854 L .44312 .59483 L p F P
0 g
s
0 0 0 r
.38057 .60854 m .36771 .6184 L .30688 .64211 L p F P
0 g
s
.254 0 .092 r
.60788 1.34357 m .58412 1.38687 L .60794 1.37791 L p F P
0 g
s
.254 0 .092 r
.60794 1.37791 m .63969 1.33145 L .60788 1.34357 L p F P
0 g
s
.483 0 0 r
.63882 .60709 m .6512 .61682 L .58207 .60257 L p F P
0 g
s
.483 0 0 r
.58207 .60257 m .57583 .59408 L .63882 .60709 L p F P
0 g
s
.501 .278 .507 r
.30592 .67077 m .30688 .64211 L .36771 .6184 L p F P
0 g
s
.874 .513 .022 r
.34021 1.29971 m .31847 1.22511 L .30438 1.20236 L p F P
0 g
s
.931 .551 .233 r
.31847 1.22511 m .34021 1.29971 L .36028 1.31747 L p F P
0 g
s
.638 .642 .838 r
.37883 .68825 m .40184 .73349 L .35514 .75105 L p F P
0 g
s
.638 .642 .838 r
.35514 .75105 m .3226 .70972 L .37883 .68825 L p F P
0 g
s
.83 .546 .509 r
.7128 .63986 m .71373 .66853 L .65185 .64556 L p F P
0 g
s
.83 .546 .509 r
.65185 .64556 m .6512 .61682 L .7128 .63986 L p F P
0 g
s
0 0 0 r
.70736 1.2233 m .68542 1.29815 L .69707 1.27837 L p F P
0 g
s
0 0 0 r
.69707 1.27837 m .72067 1.20041 L .70736 1.2233 L p F P
0 g
s
.74 .724 .837 r
.69742 .70768 m .66561 .74938 L .61833 .73232 L p F P
0 g
s
.06 0 .223 r
.32237 1.04189 m .34782 .95952 L .36915 .94056 L p F P
0 g
s
0 0 .013 r
.34782 .95952 m .32237 1.04189 L .30719 1.06546 L p F P
0 g
s
.319 0 0 r
.43634 .60338 m .44312 .59483 L .5094 .58991 L p F P
0 g
s
.163 0 0 r
.44312 .59483 m .43634 .60338 L .36771 .6184 L p F P
0 g
s
.777 .271 0 r
.41817 1.37874 m .38706 1.33258 L .36028 1.31747 L p F P
0 g
s
.763 .284 .101 r
.38706 1.33258 m .41817 1.37874 L .4423 1.3875 L p F P
0 g
s
.423 0 0 r
.57583 .59408 m .58207 .60257 L .50912 .59799 L p F P
0 g
s
.423 0 0 r
.50912 .59799 m .5094 .58991 L .57583 .59408 L p F P
0 g
s
.883 .697 .178 r
.30185 1.09007 m .29981 1.17837 L .30438 1.20236 L p F P
0 g
s
0 0 .137 r
.29981 1.17837 m .30185 1.09007 L .30719 1.06546 L p F P
0 g
s
.921 .534 .235 r
.67503 .95786 m .70025 1.04003 L .67502 1.01891 L p F P
0 g
s
.921 .534 .235 r
.67502 1.01891 m .65307 .93911 L .67503 .95786 L p F P
0 g
s
.021 0 0 r
.63969 1.33145 m .60794 1.37791 L .62752 1.36658 L p F P
0 g
s
.021 0 0 r
.62752 1.36658 m .66595 1.3161 L .63969 1.33145 L p F P
0 g
s
.893 .676 .161 r
.72241 1.08803 m .72442 1.17637 L .7182 1.15227 L p F P
0 g
s
.893 .676 .161 r
.7182 1.15227 m .71623 1.06346 L .72241 1.08803 L p F P
0 g
s
.319 0 0 r
.5094 .58991 m .50912 .59799 L .43634 .60338 L p F P
0 g
s
.501 .278 .507 r
.36771 .6184 m .36702 .64714 L .30592 .67077 L p F P
0 g
s
.602 .518 .732 r
.3226 .70972 m .30592 .67077 L .36702 .64714 L p F P
0 g
s
.602 .518 .732 r
.36702 .64714 m .37883 .68825 L .3226 .70972 L p F P
0 g
s
.794 .674 .731 r
.71373 .66853 m .69742 .70768 L .64047 .68682 L p F P
0 g
s
.794 .674 .731 r
.64047 .68682 m .65185 .64556 L .71373 .66853 L p F P
0 g
s
.661 .653 .831 r
.4422 .67467 m .45431 .7224 L .40184 .73349 L p F P
0 g
s
.661 .653 .831 r
.40184 .73349 m .37883 .68825 L .4422 .67467 L p F P
0 g
s
.74 .724 .837 r
.61833 .73232 m .64047 .68682 L .69742 .70768 L p F P
0 g
s
.723 .702 .831 r
.64047 .68682 m .61833 .73232 L .56549 .7218 L p F P
0 g
s
.586 .346 .519 r
.36702 .64714 m .36771 .6184 L .43634 .60338 L p F P
0 g
s
.772 .498 .52 r
.6512 .61682 m .65185 .64556 L .5824 .63136 L p F P
0 g
s
.772 .498 .52 r
.5824 .63136 m .58207 .60257 L .6512 .61682 L p F P
0 g
s
.702 .198 0 r
.3982 1.36759 m .36028 1.31747 L .34021 1.29971 L p F P
0 g
s
.777 .271 0 r
.36028 1.31747 m .3982 1.36759 L .41817 1.37874 L p F P
0 g
s
.874 .513 .022 r
.30438 1.20236 m .32786 1.28005 L .34021 1.29971 L p F P
0 g
s
0 0 .285 r
.32786 1.28005 m .30438 1.20236 L .29981 1.17837 L p F P
0 g
s
.298 .031 .33 r
.34685 1.02054 m .36915 .94056 L .39757 .92458 L p F P
0 g
s
.06 0 .223 r
.36915 .94056 m .34685 1.02054 L .32237 1.04189 L p F P
0 g
s
0 0 0 r
.66595 1.3161 m .62752 1.36658 L .64192 1.35338 L p F P
0 g
s
0 0 0 r
.64192 1.35338 m .68542 1.29815 L .66595 1.3161 L p F P
0 g
s
.877 .505 .336 r
.65307 .93911 m .67502 1.01891 L .64164 1.0012 L p F P
0 g
s
.877 .505 .336 r
.64164 1.0012 m .62411 .92342 L .65307 .93911 L p F P
0 g
s
.81 .801 .309 r
.72067 1.20041 m .69707 1.27837 L .70022 1.25762 L p F P
0 g
s
.81 .801 .309 r
.70022 1.25762 m .72442 1.17637 L .72067 1.20041 L p F P
0 g
s
.683 .667 .828 r
.50935 .66979 m .50984 .71843 L .45431 .7224 L p F P
0 g
s
.683 .667 .828 r
.45431 .7224 m .4422 .67467 L .50935 .66979 L p F P
0 g
s
.723 .702 .831 r
.56549 .7218 m .57666 .67393 L .64047 .68682 L p F P
0 g
s
.704 .683 .828 r
.57666 .67393 m .56549 .7218 L .50984 .71843 L p F P
0 g
s
.704 .683 .828 r
.50984 .71843 m .50935 .66979 L .57666 .67393 L p F P
0 g
s
0 0 .357 r
.30519 1.15423 m .30719 1.06546 L .32237 1.04189 L p F P
0 g
s
0 0 .137 r
.30719 1.06546 m .30519 1.15423 L .29981 1.17837 L p F P
0 g
s
.655 .402 .524 r
.43597 .63218 m .43634 .60338 L .50912 .59799 L p F P
0 g
s
.586 .346 .519 r
.43634 .60338 m .43597 .63218 L .36702 .64714 L p F P
0 g
s
.986 .709 .369 r
.71623 1.06346 m .7182 1.15227 L .70208 1.12928 L p F P
0 g
s
.986 .709 .369 r
.70208 1.12928 m .70025 1.04003 L .71623 1.06346 L p F P
0 g
s
.715 .451 .525 r
.58207 .60257 m .5824 .63136 L .5091 .6268 L p F P
0 g
s
.715 .451 .525 r
.5091 .6268 m .50912 .59799 L .58207 .60257 L p F P
0 g
s
.647 .545 .725 r
.43597 .63218 m .4422 .67467 L .37883 .68825 L p F P
0 g
s
.647 .545 .725 r
.37883 .68825 m .36702 .64714 L .43597 .63218 L p F P
0 g
s
.319 0 0 r
.50479 1.4063 m .49855 1.39648 L .46949 1.39349 L p F P
0 g
s
.255 0 0 r
.49855 1.39648 m .50479 1.4063 L .5208 1.40623 L p F P
0 g
s
.255 0 0 r
.5208 1.40623 m .52819 1.39635 L .49855 1.39648 L p F P
0 g
s
.147 0 0 r
.52819 1.39635 m .5208 1.40623 L .53643 1.40448 L p F P
0 g
s
.147 0 0 r
.53643 1.40448 m .55714 1.39311 L .52819 1.39635 L p F P
0 g
s
.759 .636 .725 r
.65185 .64556 m .64047 .68682 L .57666 .67393 L p F P
0 g
s
.759 .636 .725 r
.57666 .67393 m .5824 .63136 L .65185 .64556 L p F P
0 g
s
.655 .402 .524 r
.50912 .59799 m .5091 .6268 L .43597 .63218 L p F P
0 g
s
.338 0 0 r
.4891 1.40469 m .46949 1.39349 L .4423 1.3875 L p F P
0 g
s
.319 0 0 r
.46949 1.39349 m .4891 1.40469 L .50479 1.4063 L p F P
0 g
s
0 0 0 r
.55714 1.39311 m .53643 1.40448 L .55095 1.40113 L p F P
0 g
s
0 0 0 r
.55095 1.40113 m .58412 1.38687 L .55714 1.39311 L p F P
0 g
s
.446 .16 .384 r
.3796 1.00251 m .39757 .92458 L .43176 .91243 L p F P
0 g
s
.298 .031 .33 r
.39757 .92458 m .3796 1.00251 L .34685 1.02054 L p F P
0 g
s
.817 .464 .388 r
.62411 .92342 m .64164 1.0012 L .60167 .98787 L p F P
0 g
s
.817 .464 .388 r
.60167 .98787 m .58951 .91163 L .62411 .92342 L p F P
0 g
s
.702 .198 0 r
.34021 1.29971 m .38334 1.35452 L .3982 1.36759 L p F P
0 g
s
0 0 .518 r
.38334 1.35452 m .34021 1.29971 L .32786 1.28005 L p F P
0 g
s
.702 .901 .537 r
.68542 1.29815 m .64192 1.35338 L .65041 1.33886 L p F P
0 g
s
.702 .901 .537 r
.65041 1.33886 m .69707 1.27837 L .68542 1.29815 L p F P
0 g
s
.302 0 0 r
.47446 1.40147 m .4423 1.3875 L .41817 1.37874 L p F P
0 g
s
.338 0 0 r
.4423 1.3875 m .47446 1.40147 L .4891 1.40469 L p F P
0 g
s
0 0 0 r
.58412 1.38687 m .55095 1.40113 L .56373 1.39633 L p F P
0 g
s
0 0 0 r
.56373 1.39633 m .60794 1.37791 L .58412 1.38687 L p F P
0 g
s
.687 .573 .722 r
.5091 .6268 m .50935 .66979 L .4422 .67467 L p F P
0 g
s
.687 .573 .722 r
.4422 .67467 m .43597 .63218 L .5091 .6268 L p F P
0 g
s
.724 .603 .722 r
.5824 .63136 m .57666 .67393 L .50935 .66979 L p F P
0 g
s
0 0 .501 r
.32398 1.25934 m .29981 1.17837 L .30519 1.15423 L p F P
0 g
s
0 0 .285 r
.29981 1.17837 m .32398 1.25934 L .32786 1.28005 L p F P
0 g
s
.724 .603 .722 r
.50935 .66979 m .5091 .6268 L .5824 .63136 L p F P
0 g
s
.989 .865 .512 r
.72442 1.17637 m .70022 1.25762 L .69455 1.23686 L p F P
0 g
s
.989 .865 .512 r
.69455 1.23686 m .7182 1.15227 L .72442 1.17637 L p F P
0 g
s
.549 .249 .412 r
.41909 .98879 m .43176 .91243 L .47003 .90478 L p F P
0 g
s
.446 .16 .384 r
.43176 .91243 m .41909 .98879 L .3796 1.00251 L p F P
0 g
s
.757 .419 .413 r
.58951 .91163 m .60167 .98787 L .55709 .97966 L p F P
0 g
s
.757 .419 .413 r
.55709 .97966 m .55098 .90436 L .58951 .91163 L p F P
0 g
s
.302 0 0 r
.41817 1.37874 m .46151 1.39678 L .47446 1.40147 L p F P
0 g
s
0 .345 .738 r
.46151 1.39678 m .41817 1.37874 L .3982 1.36759 L p F P
0 g
s
.235 .079 .455 r
.3205 1.13111 m .32237 1.04189 L .34685 1.02054 L p F P
0 g
s
0 0 .357 r
.32237 1.04189 m .3205 1.13111 L .30519 1.15423 L p F P
0 g
s
.449 .871 .748 r
.60794 1.37791 m .56373 1.39633 L .57415 1.39028 L p F P
0 g
s
.449 .871 .748 r
.57415 1.39028 m .62752 1.36658 L .60794 1.37791 L p F P
0 g
s
.947 .66 .46 r
.70025 1.04003 m .70208 1.12928 L .67662 1.10855 L p F P
0 g
s
.947 .66 .46 r
.67662 1.10855 m .67502 1.01891 L .70025 1.04003 L p F P
0 g
s
.628 .316 .424 r
.46337 .98013 m .47003 .90478 L .51046 .90204 L p F P
0 g
s
.549 .249 .412 r
.47003 .90478 m .46337 .98013 L .41909 .98879 L p F P
0 g
s
.695 .371 .424 r
.55098 .90436 m .55709 .97966 L .51018 .97703 L p F P
0 g
s
.695 .371 .424 r
.51018 .97703 m .51046 .90204 L .55098 .90436 L p F P
0 g
s
.628 .316 .424 r
.51046 .90204 m .51018 .97703 L .46337 .98013 L p F P
0 g
s
0 .141 .683 r
.37432 1.34009 m .32786 1.28005 L .32398 1.25934 L p F P
0 g
s
0 0 .518 r
.32786 1.28005 m .37432 1.34009 L .38334 1.35452 L p F P
0 g
s
.913 .975 .691 r
.69707 1.27837 m .65041 1.33886 L .65251 1.3237 L p F P
0 g
s
.913 .975 .691 r
.65251 1.3237 m .70022 1.25762 L .69707 1.27837 L p F P
0 g
s
0 .406 .839 r
.45087 1.39082 m .3982 1.36759 L .38334 1.35452 L p F P
0 g
s
0 .345 .738 r
.3982 1.36759 m .45087 1.39082 L .46151 1.39678 L p F P
0 g
s
.654 .97 .846 r
.62752 1.36658 m .57415 1.39028 L .58174 1.38326 L p F P
0 g
s
.654 .97 .846 r
.58174 1.38326 m .64192 1.35338 L .62752 1.36658 L p F P
0 g
s
.168 .142 .583 r
.32891 1.23855 m .30519 1.15423 L .3205 1.13111 L p F P
0 g
s
0 0 .501 r
.30519 1.15423 m .32891 1.23855 L .32398 1.25934 L p F P
0 g
s
.982 .803 .587 r
.7182 1.15227 m .69455 1.23686 L .68013 1.21709 L p F P
0 g
s
.982 .803 .587 r
.68013 1.21709 m .70208 1.12928 L .7182 1.15227 L p F P
0 g
s
.401 .207 .498 r
.34521 1.11015 m .34685 1.02054 L .3796 1.00251 L p F P
0 g
s
.235 .079 .455 r
.34685 1.02054 m .34521 1.11015 L .3205 1.13111 L p F P
0 g
s
.887 .603 .501 r
.67502 1.01891 m .67662 1.10855 L .64292 1.09118 L p F P
0 g
s
.887 .603 .501 r
.64292 1.09118 m .64164 1.0012 L .67502 1.01891 L p F P
0 g
s
.119 .453 .885 r
.44303 1.38387 m .38334 1.35452 L .37432 1.34009 L p F P
0 g
s
0 .406 .839 r
.38334 1.35452 m .44303 1.38387 L .45087 1.39082 L p F P
0 g
s
.786 .99 .886 r
.64192 1.35338 m .58174 1.38326 L .58612 1.37557 L p F P
0 g
s
.786 .99 .886 r
.58612 1.37557 m .65041 1.33886 L .64192 1.35338 L p F P
0 g
s
.15 .269 .732 r
.37167 1.32496 m .32398 1.25934 L .32891 1.23855 L p F P
0 g
s
0 .141 .683 r
.32398 1.25934 m .37167 1.32496 L .37432 1.34009 L p F P
0 g
s
.955 .921 .734 r
.70022 1.25762 m .65251 1.3237 L .64803 1.30858 L p F P
0 g
s
.955 .921 .734 r
.64803 1.30858 m .69455 1.23686 L .70022 1.25762 L p F P
0 g
s
.475 .796 .999 r
.51212 1.38439 m .50479 1.4063 L .4891 1.40469 L closepath p F P
0 g
s
.497 .817 1 r
.51212 1.38439 m .5208 1.40623 L .50479 1.4063 L closepath p F P
0 g
s
.527 .837 .999 r
.51212 1.38439 m .53643 1.40448 L .5208 1.40623 L closepath p F P
0 g
s
.464 .775 .997 r
.51212 1.38439 m .4891 1.40469 L .47446 1.40147 L closepath p F P
0 g
s
.562 .855 .997 r
.51212 1.38439 m .55095 1.40113 L .53643 1.40448 L closepath p F P
0 g
s
.511 .293 .519 r
.37827 1.09246 m .3796 1.00251 L .41909 .98879 L p F P
0 g
s
.401 .207 .498 r
.3796 1.00251 m .37827 1.09246 L .34521 1.11015 L p F P
0 g
s
.463 .756 .993 r
.51212 1.38439 m .47446 1.40147 L .46151 1.39678 L closepath p F P
0 g
s
.6 .867 .993 r
.51212 1.38439 m .56373 1.39633 L .55095 1.40113 L closepath p F P
0 g
s
.827 .551 .52 r
.64164 1.0012 m .64292 1.09118 L .60255 1.07809 L p F P
0 g
s
.827 .551 .52 r
.60255 1.07809 m .60167 .98787 L .64164 1.0012 L p F P
0 g
s
.352 .259 .609 r
.34261 1.21866 m .3205 1.13111 L .34521 1.11015 L p F P
0 g
s
.168 .142 .583 r
.3205 1.13111 m .34261 1.21866 L .32891 1.23855 L p F P
0 g
s
.26 .485 .885 r
.43837 1.37622 m .37432 1.34009 L .37167 1.32496 L p F P
0 g
s
.119 .453 .885 r
.37432 1.34009 m .43837 1.37622 L .44303 1.38387 L p F P
0 g
s
.845 .956 .884 r
.65041 1.33886 m .58612 1.37557 L .58706 1.36758 L p F P
0 g
s
.845 .956 .884 r
.58706 1.36758 m .65251 1.3237 L .65041 1.33886 L p F P
0 g
s
.93 .729 .61 r
.70208 1.12928 m .68013 1.21709 L .6575 1.1993 L p F P
0 g
s
.93 .729 .61 r
.6575 1.1993 m .67662 1.10855 L .70208 1.12928 L p F P
0 g
s
.47 .739 .987 r
.51212 1.38439 m .46151 1.39678 L .45087 1.39082 L closepath p F P
0 g
s
.636 .872 .986 r
.51212 1.38439 m .57415 1.39028 L .56373 1.39633 L closepath p F P
0 g
s
.484 .724 .977 r
.51212 1.38439 m .45087 1.39082 L .44303 1.38387 L closepath p F P
0 g
s
.666 .871 .976 r
.51212 1.38439 m .58174 1.38326 L .57415 1.39028 L closepath p F P
0 g
s
.592 .357 .529 r
.41815 1.07899 m .41909 .98879 L .46337 .98013 L p F P
0 g
s
.511 .293 .519 r
.41909 .98879 m .41815 1.07899 L .37827 1.09246 L p F P
0 g
s
.325 .35 .733 r
.3756 1.30981 m .32891 1.23855 L .34261 1.21866 L p F P
0 g
s
.15 .269 .732 r
.32891 1.23855 m .3756 1.30981 L .37167 1.32496 L p F P
0 g
s
.771 .503 .53 r
.60167 .98787 m .60255 1.07809 L .55753 1.07003 L p F P
0 g
s
.771 .503 .53 r
.55753 1.07003 m .55709 .97966 L .60167 .98787 L p F P
0 g
s
.931 .841 .733 r
.69455 1.23686 m .64803 1.30858 L .63708 1.29423 L p F P
0 g
s
.931 .841 .733 r
.63708 1.29423 m .68013 1.21709 L .69455 1.23686 L p F P
0 g
s
.369 .507 .866 r
.43713 1.36824 m .37167 1.32496 L .3756 1.30981 L p F P
0 g
s
.26 .485 .885 r
.37167 1.32496 m .43713 1.36824 L .43837 1.37622 L p F P
0 g
s
.503 .712 .965 r
.51212 1.38439 m .44303 1.38387 L .43837 1.37622 L closepath p F P
0 g
s
.69 .863 .964 r
.51212 1.38439 m .58612 1.37557 L .58174 1.38326 L closepath p F P
0 g
s
.858 .902 .864 r
.65251 1.3237 m .58706 1.36758 L .58449 1.35965 L p F P
0 g
s
.858 .902 .864 r
.58449 1.35965 m .64803 1.30858 L .65251 1.3237 L p F P
0 g
s
.658 .41 .534 r
.46287 1.07049 m .46337 .98013 L .51018 .97703 L p F P
0 g
s
.592 .357 .529 r
.46337 .98013 m .46287 1.07049 L .41815 1.07899 L p F P
0 g
s
.716 .458 .534 r
.55709 .97966 m .55753 1.07003 L .51015 1.06745 L p F P
0 g
s
.716 .458 .534 r
.51015 1.06745 m .51018 .97703 L .55709 .97966 L p F P
0 g
s
.47 .336 .616 r
.36459 1.20067 m .34521 1.11015 L .37827 1.09246 L p F P
0 g
s
.352 .259 .609 r
.34521 1.11015 m .36459 1.20067 L .34261 1.21866 L p F P
0 g
s
.658 .41 .534 r
.51018 .97703 m .51015 1.06745 L .46287 1.07049 L p F P
0 g
s
.874 .665 .617 r
.67662 1.10855 m .6575 1.1993 L .62763 1.1844 L p F P
0 g
s
.874 .665 .617 r
.62763 1.1844 m .64292 1.09118 L .67662 1.10855 L p F P
0 g
s
.524 .703 .952 r
.51212 1.38439 m .43837 1.37622 L .43713 1.36824 L closepath p F P
0 g
s
.707 .85 .951 r
.51212 1.38439 m .58706 1.36758 L .58612 1.37557 L closepath p F P
0 g
s
.441 .405 .723 r
.38603 1.29537 m .34261 1.21866 L .36459 1.20067 L p F P
0 g
s
.325 .35 .733 r
.34261 1.21866 m .38603 1.29537 L .3756 1.30981 L p F P
0 g
s
.891 .77 .722 r
.68013 1.21709 m .63708 1.29423 L .62008 1.28136 L p F P
0 g
s
.891 .77 .722 r
.62008 1.28136 m .6575 1.1993 L .68013 1.21709 L p F P
0 g
s
.452 .525 .842 r
.43941 1.3603 m .3756 1.30981 L .38603 1.29537 L p F P
0 g
s
.369 .507 .866 r
.3756 1.30981 m .43941 1.3603 L .43713 1.36824 L p F P
0 g
s
.546 .697 .938 r
.51212 1.38439 m .43713 1.36824 L .43941 1.3603 L closepath p F P
0 g
s
.849 .846 .84 r
.64803 1.30858 m .58449 1.35965 L .5785 1.35217 L p F P
0 g
s
.849 .846 .84 r
.5785 1.35217 m .63708 1.29423 L .64803 1.30858 L p F P
0 g
s
.717 .834 .937 r
.51212 1.38439 m .58449 1.35965 L .58706 1.36758 L closepath p F P
0 g
s
.719 .797 .913 r
.51212 1.38439 m .56935 1.34548 L .5785 1.35217 L closepath p F P
0 g
s
.714 .779 .903 r
.51212 1.38439 m .55746 1.33992 L .56935 1.34548 L closepath p F P
0 g
s
.705 .761 .896 r
.51212 1.38439 m .54338 1.33576 L .55746 1.33992 L closepath p F P
0 g
s
.694 .745 .891 r
.51212 1.38439 m .52779 1.33321 L .54338 1.33576 L closepath p F P
0 g
s
.681 .731 .888 r
.51212 1.38439 m .51143 1.33239 L .52779 1.33321 L closepath p F P
0 g
s
.665 .719 .889 r
.51212 1.38439 m .49511 1.33335 L .51143 1.33239 L closepath p F P
0 g
s
.649 .709 .891 r
.51212 1.38439 m .47962 1.33604 L .49511 1.33335 L closepath p F P
0 g
s
.63 .701 .896 r
.51212 1.38439 m .4657 1.34033 L .47962 1.33604 L closepath p F P
0 g
s
.611 .696 .904 r
.51212 1.38439 m .45402 1.34599 L .4657 1.34033 L closepath p F P
0 g
s
.59 .693 .914 r
.51212 1.38439 m .44512 1.35276 L .45402 1.34599 L closepath p F P
0 g
s
.568 .693 .925 r
.51212 1.38439 m .43941 1.3603 L .44512 1.35276 L closepath p F P
0 g
s
.721 .816 .924 r
.51212 1.38439 m .5785 1.35217 L .58449 1.35965 L closepath p F P
0 g
s
.554 .394 .618 r
.39389 1.1855 m .37827 1.09246 L .41815 1.07899 L p F P
0 g
s
.47 .336 .616 r
.37827 1.09246 m .39389 1.1855 L .36459 1.20067 L p F P
0 g
s
.822 .612 .618 r
.64292 1.09118 m .62763 1.1844 L .59195 1.1732 L p F P
0 g
s
.822 .612 .618 r
.59195 1.1732 m .60255 1.07809 L .64292 1.09118 L p F P
0 g
s
.452 .525 .842 r
.38603 1.29537 m .44512 1.35276 L .43941 1.3603 L p F P
0 g
s
.441 .405 .723 r
.36459 1.20067 m .40254 1.28235 L .38603 1.29537 L p F P
0 g
s
.517 .542 .821 r
.44512 1.35276 m .38603 1.29537 L .40254 1.28235 L p F P
0 g
s
.829 .794 .819 r
.63708 1.29423 m .5785 1.35217 L .56935 1.34548 L p F P
0 g
s
.829 .794 .819 r
.56935 1.34548 m .62008 1.28136 L .63708 1.29423 L p F P
0 g
s
.523 .447 .712 r
.40254 1.28235 m .36459 1.20067 L .39389 1.1855 L p F P
0 g
s
.517 .542 .821 r
.40254 1.28235 m .45402 1.34599 L .44512 1.35276 L p F P
0 g
s
.848 .711 .712 r
.6575 1.1993 m .62008 1.28136 L .59779 1.27062 L p F P
0 g
s
.848 .711 .712 r
.59779 1.27062 m .62763 1.1844 L .6575 1.1993 L p F P
0 g
s
.619 .441 .619 r
.42916 1.17397 m .41815 1.07899 L .46287 1.07049 L p F P
0 g
s
.554 .394 .618 r
.41815 1.07899 m .42916 1.17397 L .39389 1.1855 L p F P
0 g
s
.773 .566 .619 r
.60255 1.07809 m .59195 1.1732 L .55219 1.1663 L p F P
0 g
s
.773 .566 .619 r
.55219 1.1663 m .55753 1.07003 L .60255 1.07809 L p F P
0 g
s
.569 .56 .804 r
.42442 1.27141 m .4657 1.34033 L .45402 1.34599 L p F P
0 g
s
.569 .56 .804 r
.45402 1.34599 m .40254 1.28235 L .42442 1.27141 L p F P
0 g
s
.804 .75 .803 r
.62008 1.28136 m .56935 1.34548 L .55746 1.33992 L p F P
0 g
s
.675 .483 .619 r
.46865 1.1667 m .46287 1.07049 L .51015 1.06745 L p F P
0 g
s
.619 .441 .619 r
.46287 1.07049 m .46865 1.1667 L .42916 1.17397 L p F P
0 g
s
.725 .524 .619 r
.55753 1.07003 m .55219 1.1663 L .51037 1.16409 L p F P
0 g
s
.725 .524 .619 r
.51037 1.16409 m .51015 1.06745 L .55753 1.07003 L p F P
0 g
s
.523 .447 .712 r
.39389 1.1855 m .42442 1.27141 L .40254 1.28235 L p F P
0 g
s
.804 .75 .803 r
.55746 1.33992 m .59779 1.27062 L .62008 1.28136 L p F P
0 g
s
.675 .483 .619 r
.51015 1.06745 m .51037 1.16409 L .46865 1.1667 L p F P
0 g
s
.613 .579 .792 r
.45064 1.26311 m .47962 1.33604 L .4657 1.34033 L p F P
0 g
s
.613 .579 .792 r
.4657 1.34033 m .42442 1.27141 L .45064 1.26311 L p F P
0 g
s
.776 .711 .791 r
.59779 1.27062 m .55746 1.33992 L .54338 1.33576 L p F P
0 g
s
.586 .484 .704 r
.42442 1.27141 m .39389 1.1855 L .42916 1.17397 L p F P
0 g
s
.807 .663 .704 r
.62763 1.1844 m .59779 1.27062 L .57127 1.26255 L p F P
0 g
s
.807 .663 .704 r
.57127 1.26255 m .59195 1.1732 L .62763 1.1844 L p F P
0 g
s
.651 .6 .784 r
.47992 1.25788 m .49511 1.33335 L .47962 1.33604 L p F P
0 g
s
.651 .6 .784 r
.47962 1.33604 m .45064 1.26311 L .47992 1.25788 L p F P
0 g
s
.776 .711 .791 r
.54338 1.33576 m .57127 1.26255 L .59779 1.27062 L p F P
0 g
s
.747 .678 .783 r
.57127 1.26255 m .54338 1.33576 L .52779 1.33321 L p F P
0 g
s
.685 .623 .78 r
.51082 1.25601 m .51143 1.33239 L .49511 1.33335 L p F P
0 g
s
.685 .623 .78 r
.49511 1.33335 m .47992 1.25788 L .51082 1.25601 L p F P
0 g
s
.747 .678 .783 r
.52779 1.33321 m .54179 1.2576 L .57127 1.26255 L p F P
0 g
s
.717 .648 .78 r
.54179 1.2576 m .52779 1.33321 L .51143 1.33239 L p F P
0 g
s
.717 .648 .78 r
.51143 1.33239 m .51082 1.25601 L .54179 1.2576 L p F P
0 g
s
.586 .484 .704 r
.42916 1.17397 m .45064 1.26311 L .42442 1.27141 L p F P
0 g
s
.638 .517 .699 r
.45064 1.26311 m .42916 1.17397 L .46865 1.1667 L p F P
0 g
s
.766 .621 .698 r
.59195 1.1732 m .57127 1.26255 L .54179 1.2576 L p F P
0 g
s
.766 .621 .698 r
.54179 1.2576 m .55219 1.1663 L .59195 1.1732 L p F P
0 g
s
.638 .517 .699 r
.46865 1.1667 m .47992 1.25788 L .45064 1.26311 L p F P
0 g
s
.684 .55 .696 r
.47992 1.25788 m .46865 1.1667 L .51037 1.16409 L p F P
0 g
s
.726 .584 .696 r
.55219 1.1663 m .54179 1.2576 L .51082 1.25601 L p F P
0 g
s
.726 .584 .696 r
.51082 1.25601 m .51037 1.16409 L .55219 1.1663 L p F P
0 g
s
.684 .55 .696 r
.51037 1.16409 m .51082 1.25601 L .47992 1.25788 L p F P
0 g
s
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.35119 0.0121928 0.64881 0.499904 MathSubStart
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0568615 1.16644 2.13046e-017 1.16644 [
[ 0 0 0 0 ]
[ 1 1.08584 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.08584 L
0 1.08584 L
closepath
clip
newpath
.5 Mabswid
[ ] 0 setdash
.736 .605 .711 r
.44995 .51286 m .44273 .55053 L .49929 .55767 L p F P
0 g
s
.736 .605 .711 r
.49929 .55767 m .50158 .51953 L .44995 .51286 L p F P
0 g
s
.741 .529 .606 r
.49846 .59474 m .49929 .55767 L .44273 .55053 L p F P
0 g
s
.706 .578 .71 r
.50158 .51953 m .49929 .55767 L .55675 .55736 L p F P
0 g
s
.699 .494 .606 r
.49929 .55767 m .49846 .59474 L .55772 .59443 L p F P
0 g
s
.699 .494 .606 r
.55772 .59443 m .55675 .55736 L .49929 .55767 L p F P
0 g
s
.706 .578 .71 r
.55675 .55736 m .55405 .51924 L .50158 .51953 L p F P
0 g
s
.674 .555 .711 r
.55405 .51924 m .55675 .55736 L .61315 .54961 L p F P
0 g
s
.654 .458 .606 r
.55675 .55736 m .55772 .59443 L .6159 .58661 L p F P
0 g
s
.654 .458 .606 r
.6159 .58661 m .61315 .54961 L .55675 .55736 L p F P
0 g
s
.715 .671 .806 r
.46081 .47928 m .44995 .51286 L .50158 .51953 L p F P
0 g
s
.715 .671 .806 r
.50158 .51953 m .50502 .48511 L .46081 .47928 L p F P
0 g
s
.698 .657 .805 r
.50502 .48511 m .50158 .51953 L .55405 .51924 L p F P
0 g
s
.698 .657 .805 r
.55405 .51924 m .54998 .48486 L .50502 .48511 L p F P
0 g
s
.681 .644 .806 r
.54998 .48486 m .55405 .51924 L .60553 .51201 L p F P
0 g
s
.674 .555 .711 r
.61315 .54961 m .60553 .51201 L .55405 .51924 L p F P
0 g
s
.73 .687 .81 r
.419 .46754 m .40099 .4994 L .44995 .51286 L p F P
0 g
s
.73 .687 .81 r
.44995 .51286 m .46081 .47928 L .419 .46754 L p F P
0 g
s
.765 .633 .715 r
.44273 .55053 m .44995 .51286 L .40099 .4994 L p F P
0 g
s
.765 .633 .715 r
.40099 .4994 m .38899 .53609 L .44273 .55053 L p F P
0 g
s
.783 .566 .606 r
.44011 .58754 m .44273 .55053 L .38899 .53609 L p F P
0 g
s
.741 .529 .606 r
.44273 .55053 m .44011 .58754 L .49846 .59474 L p F P
0 g
s
.681 .644 .806 r
.60553 .51201 m .59405 .47854 L .54998 .48486 L p F P
0 g
s
.664 .634 .81 r
.59405 .47854 m .60553 .51201 L .65418 .498 L p F P
0 g
s
.641 .532 .715 r
.60553 .51201 m .61315 .54961 L .66655 .53458 L p F P
0 g
s
.641 .532 .715 r
.66655 .53458 m .65418 .498 L .60553 .51201 L p F P
0 g
s
.605 .421 .606 r
.61315 .54961 m .6159 .58661 L .67102 .57145 L p F P
0 g
s
.605 .421 .606 r
.67102 .57145 m .66655 .53458 L .61315 .54961 L p F P
0 g
s
.726 .417 .455 r
.49933 .626 m .49846 .59474 L .44011 .58754 L p F P
0 g
s
.671 .372 .458 r
.49846 .59474 m .49933 .626 L .55667 .62571 L p F P
0 g
s
.671 .372 .458 r
.55667 .62571 m .55772 .59443 L .49846 .59474 L p F P
0 g
s
.61 .321 .454 r
.55772 .59443 m .55667 .62571 L .61295 .61828 L p F P
0 g
s
.61 .321 .454 r
.61295 .61828 m .6159 .58661 L .55772 .59443 L p F P
0 g
s
.656 .741 .913 r
.47415 .45443 m .46081 .47928 L .50502 .48511 L p F P
0 g
s
.656 .741 .913 r
.50502 .48511 m .50924 .45912 L .47415 .45443 L p F P
0 g
s
.661 .746 .913 r
.50924 .45912 m .50502 .48511 L .54998 .48486 L p F P
0 g
s
.661 .746 .913 r
.54998 .48486 m .54497 .45892 L .50924 .45912 L p F P
0 g
s
.666 .749 .912 r
.54497 .45892 m .54998 .48486 L .59405 .47854 L p F P
0 g
s
.743 .706 .816 r
.3812 .45021 m .35653 .47947 L .40099 .4994 L p F P
0 g
s
.743 .706 .816 r
.40099 .4994 m .419 .46754 L .3812 .45021 L p F P
0 g
s
.794 .666 .722 r
.38899 .53609 m .40099 .4994 L .35653 .47947 L p F P
0 g
s
.664 .634 .81 r
.65418 .498 m .63559 .46632 L .59405 .47854 L p F P
0 g
s
.645 .626 .817 r
.63559 .46632 m .65418 .498 L .69816 .47755 L p F P
0 g
s
.603 .511 .723 r
.65418 .498 m .66655 .53458 L .71498 .51261 L p F P
0 g
s
.603 .511 .723 r
.71498 .51261 m .69816 .47755 L .65418 .498 L p F P
0 g
s
.652 .736 .911 r
.44106 .44499 m .419 .46754 L .46081 .47928 L p F P
0 g
s
.652 .736 .911 r
.46081 .47928 m .47415 .45443 L .44106 .44499 L p F P
0 g
s
.794 .666 .722 r
.35653 .47947 m .34006 .51467 L .38899 .53609 L p F P
0 g
s
.826 .607 .606 r
.38465 .57297 m .38899 .53609 L .34006 .51467 L p F P
0 g
s
.783 .566 .606 r
.38899 .53609 m .38465 .57297 L .44011 .58754 L p F P
0 g
s
.666 .749 .912 r
.59405 .47854 m .57995 .45383 L .54497 .45892 L p F P
0 g
s
.672 .752 .911 r
.57995 .45383 m .59405 .47854 L .63559 .46632 L p F P
0 g
s
.78 .458 .443 r
.44288 .61916 m .44011 .58754 L .38465 .57297 L p F P
0 g
s
.726 .417 .455 r
.44011 .58754 m .44288 .61916 L .49933 .626 L p F P
0 g
s
.546 .379 .606 r
.66655 .53458 m .67102 .57145 L .72106 .54926 L p F P
0 g
s
.546 .379 .606 r
.72106 .54926 m .71498 .51261 L .66655 .53458 L p F P
0 g
s
.537 .26 .442 r
.6159 .58661 m .61295 .61828 L .66623 .60389 L p F P
0 g
s
.537 .26 .442 r
.66623 .60389 m .67102 .57145 L .6159 .58661 L p F P
0 g
s
.649 .731 .909 r
.41132 .4311 m .3812 .45021 L .419 .46754 L p F P
0 g
s
.649 .731 .909 r
.419 .46754 m .44106 .44499 L .41132 .4311 L p F P
0 g
s
.672 .752 .911 r
.63559 .46632 m .61281 .44401 L .57995 .45383 L p F P
0 g
s
.677 .753 .909 r
.61281 .44401 m .63559 .46632 L .67296 .44855 L p F P
0 g
s
.645 .626 .817 r
.69816 .47755 m .67296 .44855 L .63559 .46632 L p F P
0 g
s
.826 .607 .606 r
.34006 .51467 m .33409 .55134 L .38465 .57297 L p F P
0 g
s
.834 .499 .42 r
.38927 .60534 m .38465 .57297 L .33409 .55134 L p F P
0 g
s
.78 .458 .443 r
.38465 .57297 m .38927 .60534 L .44288 .61916 L p F P
0 g
s
.443 .179 .417 r
.67102 .57145 m .66623 .60389 L .71455 .58286 L p F P
0 g
s
.443 .179 .417 r
.71455 .58286 m .72106 .54926 L .67102 .57145 L p F P
0 g
s
.755 .726 .825 r
.34903 .4278 m .31842 .45361 L .35653 .47947 L p F P
0 g
s
.755 .726 .825 r
.35653 .47947 m .3812 .45021 L .34903 .4278 L p F P
0 g
s
.823 .703 .732 r
.34006 .51467 m .35653 .47947 L .31842 .45361 L p F P
0 g
s
.626 .621 .826 r
.67296 .44855 m .69816 .47755 L .73564 .45123 L p F P
0 g
s
.561 .491 .733 r
.69816 .47755 m .71498 .51261 L .75646 .48425 L p F P
0 g
s
.561 .491 .733 r
.75646 .48425 m .73564 .45123 L .69816 .47755 L p F P
0 g
s
.823 .703 .732 r
.31842 .45361 m .29791 .48682 L .34006 .51467 L p F P
0 g
s
.873 .654 .604 r
.33409 .55134 m .34006 .51467 L .29791 .48682 L p F P
0 g
s
.645 .192 .16 r
.50206 .64678 m .49933 .626 L .44288 .61916 L p F P
0 g
s
.576 .138 .173 r
.49933 .626 m .50206 .64678 L .5534 .64652 L p F P
0 g
s
.576 .138 .173 r
.5534 .64652 m .55667 .62571 L .49933 .626 L p F P
0 g
s
.488 .062 .157 r
.55667 .62571 m .5534 .64652 L .60376 .63997 L p F P
0 g
s
.488 .062 .157 r
.60376 .63997 m .61295 .61828 L .55667 .62571 L p F P
0 g
s
.471 .328 .604 r
.71498 .51261 m .72106 .54926 L .764 .52059 L p F P
0 g
s
.471 .328 .604 r
.764 .52059 m .75646 .48425 L .71498 .51261 L p F P
0 g
s
.44 .776 .998 r
.4886 .44271 m .47415 .45443 L .50924 .45912 L p F P
0 g
s
.44 .776 .998 r
.50924 .45912 m .51383 .44611 L .4886 .44271 L p F P
0 g
s
.482 .818 .999 r
.51383 .44611 m .50924 .45912 L .54497 .45892 L p F P
0 g
s
.539 .856 .997 r
.56467 .44228 m .53953 .44596 L p .54719 .44745 L F P
0 g
s
.539 .856 .997 r
.57995 .45383 m .56467 .44228 L p .54719 .44745 L F P
0 g
s
.605 .885 .991 r
.58819 .43518 m .56467 .44228 L p .57376 .4426 L F P
0 g
s
.605 .885 .991 r
.61281 .44401 m .58819 .43518 L p .57376 .4426 L F P
0 g
s
.404 .848 .746 r
.58819 .43518 m .6091 .42493 L p .60417 .42696 L F P
0 g
s
0 0 0 r
.56467 .44228 m .58819 .43518 L p .58398 .43656 L F P
0 g
s
.605 .885 .991 r
.56467 .44228 m .57995 .45383 L .61281 .44401 L p F P
0 g
s
0 0 0 r
.54984 .44773 m .56467 .44228 L p .58398 .43656 L F P
0 g
s
0 0 0 r
.56438 .44334 m .54984 .44773 L p .5556 .44584 L F P
0 g
s
0 0 0 r
.58819 .43518 m .56438 .44334 L p .5556 .44584 L F P
0 g
s
.404 .848 .746 r
.56438 .44334 m .58819 .43518 L p .60417 .42696 L F P
0 g
s
.404 .848 .746 r
.57723 .43703 m .56438 .44334 L p .57109 .44058 L F P
0 g
s
.404 .848 .746 r
.6091 .42493 m .57723 .43703 L p .57109 .44058 L F P
0 g
s
.627 .961 .836 r
.5878 .42906 m .57723 .43703 L p .58459 .43328 L F P
0 g
s
.627 .961 .836 r
.62647 .41192 m .5878 .42906 L p .58459 .43328 L F P
0 g
s
.59 .861 .524 r
.5878 .42906 m .5956 .41975 L p .5912 .42459 L F P
0 g
s
0 0 0 r
.57723 .43703 m .5878 .42906 L p .58373 .43328 L F P
0 g
s
.539 .856 .997 r
.53953 .44596 m .54497 .45892 L .57995 .45383 L p F P
0 g
s
.627 .961 .836 r
.57723 .43703 m .6091 .42493 L .62647 .41192 L p F P
0 g
s
0 0 0 r
.55035 .46786 m .57723 .43703 L p .58373 .43328 L F P
0 g
s
0 0 0 r
.55536 .46411 m .55035 .46786 L p .55349 .46461 L F P
0 g
s
0 0 0 r
.5878 .42906 m .55536 .46411 L p .55349 .46461 L F P
0 g
s
.59 .861 .524 r
.55536 .46411 m .5878 .42906 L p .5912 .42459 L F P
0 g
s
.59 .861 .524 r
.559 .45975 m .55536 .46411 L p .55872 .4604 L F P
0 g
s
.59 .861 .524 r
.5956 .41975 m .559 .45975 L p .55872 .4604 L F P
0 g
s
.095 0 0 r
.57723 .43703 m .55035 .46786 L .54422 .47084 L p F P
0 g
s
.649 .763 .288 r
.559 .45975 m .5611 .45498 L p .55928 .45908 L F P
0 g
s
0 0 .001 r
.55536 .46411 m .559 .45975 L p .55726 .4636 L F P
0 g
s
0 0 .001 r
.53437 .51536 m .53342 .51648 L p .53403 .51513 L F P
0 g
s
0 0 .001 r
.559 .45975 m .53437 .51536 L p .53403 .51513 L F P
0 g
s
.649 .763 .288 r
.53437 .51536 m .559 .45975 L p .55928 .45908 L F P
0 g
s
.649 .763 .288 r
.5349 .51415 m .53437 .51536 L p .535 .51393 L F P
0 g
s
.649 .763 .288 r
.5611 .45498 m .5349 .51415 L p .535 .51393 L F P
0 g
s
0 0 .001 r
.53342 .51648 m .55536 .46411 L p .55726 .4636 L F P
0 g
s
.765 .69 .146 r
.5349 .51415 m .535 .51289 L .5348 .51437 L p F P
0 g
s
0 0 .159 r
.53437 .51536 m .5349 .51415 L .53471 .5156 L p F P
0 g
s
0 0 .159 r
.53471 .5156 m .52553 .58349 L .53437 .51536 L p F P
0 g
s
.765 .69 .146 r
.5348 .51437 m .52553 .58349 L .5349 .51415 L p F P
0 g
s
.095 0 0 r
.56438 .44334 m .57723 .43703 L p .57365 .44071 L F P
0 g
s
.095 0 0 r
.54422 .47084 m .56438 .44334 L p .57365 .44071 L F P
0 g
s
.021 0 0 r
.53953 .44596 m .56467 .44228 L p .56133 .44312 L F P
0 g
s
.482 .818 .999 r
.54497 .45892 m .53953 .44596 L p .51974 .44854 L F P
0 g
s
.021 0 0 r
.53424 .45001 m .53953 .44596 L p .56133 .44312 L F P
0 g
s
.482 .818 .999 r
.53953 .44596 m .51383 .44611 L p .51974 .44854 L F P
0 g
s
.647 .725 .906 r
.38622 .41322 m .34903 .4278 L .3812 .45021 L p F P
0 g
s
.647 .725 .906 r
.3812 .45021 m .41132 .4311 L .38622 .41322 L p F P
0 g
s
.419 .736 .991 r
.4649 .43589 m .44106 .44499 L .47415 .45443 L p F P
0 g
s
.419 .736 .991 r
.47415 .45443 m .4886 .44271 L .4649 .43589 L p F P
0 g
s
.677 .753 .909 r
.67296 .44855 m .6422 .42977 L .61281 .44401 L p F P
0 g
s
.682 .754 .905 r
.6422 .42977 m .67296 .44855 L .70458 .42575 L p F P
0 g
s
.626 .621 .826 r
.73564 .45123 m .70458 .42575 L .67296 .44855 L p F P
0 g
s
.702 .225 .116 r
.45155 .64074 m .44288 .61916 L .38927 .60534 L p F P
0 g
s
.645 .192 .16 r
.44288 .61916 m .45155 .64074 L .50206 .64678 L p F P
0 g
s
.37 0 .11 r
.61295 .61828 m .60376 .63997 L .65133 .62727 L p F P
0 g
s
.37 0 .11 r
.65133 .62727 m .66623 .60389 L .61295 .61828 L p F P
0 g
s
.417 .699 .98 r
.44373 .42589 m .41132 .4311 L .44106 .44499 L p F P
0 g
s
.417 .699 .98 r
.44106 .44499 m .4649 .43589 L .44373 .42589 L p F P
0 g
s
.873 .654 .604 r
.29791 .48682 m .29047 .52318 L .33409 .55134 L p F P
0 g
s
.889 .538 .375 r
.34044 .58483 m .33409 .55134 L .29047 .52318 L p F P
0 g
s
.834 .499 .42 r
.33409 .55134 m .34044 .58483 L .38927 .60534 L p F P
0 g
s
.668 .901 .978 r
.58819 .43518 m .61281 .44401 L .6422 .42977 L p F P
0 g
s
.307 .062 .37 r
.72106 .54926 m .71455 .58286 L .75593 .55571 L p F P
0 g
s
.307 .062 .37 r
.75593 .55571 m .764 .52059 L .72106 .54926 L p F P
0 g
s
.765 .747 .837 r
.32399 .40104 m .28844 .42257 L .31842 .45361 L p F P
0 g
s
.765 .747 .837 r
.31842 .45361 m .34903 .4278 L .32399 .40104 L p F P
0 g
s
.852 .747 .747 r
.29791 .48682 m .31842 .45361 L .28844 .42257 L p F P
0 g
s
.74 .236 .028 r
.40367 .62854 m .38927 .60534 L .34044 .58483 L p F P
0 g
s
.702 .225 .116 r
.38927 .60534 m .40367 .62854 L .45155 .64074 L p F P
0 g
s
.607 .619 .838 r
.70458 .42575 m .73564 .45123 L .76485 .41981 L p F P
0 g
s
.511 .471 .748 r
.73564 .45123 m .75646 .48425 L .78902 .45029 L p F P
0 g
s
.511 .471 .748 r
.78902 .45029 m .76485 .41981 L .73564 .45123 L p F P
0 g
s
.199 0 .017 r
.66623 .60389 m .65133 .62727 L .69431 .60875 L p F P
0 g
s
.199 0 .017 r
.69431 .60875 m .71455 .58286 L .66623 .60389 L p F P
0 g
s
.852 .747 .747 r
.28844 .42257 m .26451 .45328 L .29791 .48682 L p F P
0 g
s
.923 .711 .599 r
.29047 .52318 m .29791 .48682 L .26451 .45328 L p F P
0 g
s
.647 .72 .902 r
.36694 .39199 m .32399 .40104 L .34903 .4278 L p F P
0 g
s
.647 .72 .902 r
.34903 .4278 m .38622 .41322 L .36694 .39199 L p F P
0 g
s
.43 .668 .962 r
.42603 .41308 m .38622 .41322 L .41132 .4311 L p F P
0 g
s
.43 .668 .962 r
.41132 .4311 m .44373 .42589 L .42603 .41308 L p F P
0 g
s
.682 .754 .905 r
.70458 .42575 m .66685 .41159 L .6422 .42977 L p F P
0 g
s
.687 .752 .902 r
.66685 .41159 m .70458 .42575 L .72895 .39867 L p F P
0 g
s
.607 .619 .838 r
.76485 .41981 m .72895 .39867 L .70458 .42575 L p F P
0 g
s
.369 .261 .598 r
.75646 .48425 m .764 .52059 L .79779 .48622 L p F P
0 g
s
.369 .261 .598 r
.79779 .48622 m .78902 .45029 L .75646 .48425 L p F P
0 g
s
.668 .901 .978 r
.6422 .42977 m .6091 .42493 L .58819 .43518 L p F P
0 g
s
.72 .902 .96 r
.6091 .42493 m .6422 .42977 L .66685 .41159 L p F P
0 g
s
.232 0 0 r
.50262 .448 m .4886 .44271 L .51383 .44611 L p F P
0 g
s
.232 0 0 r
.51383 .44611 m .51827 .4501 L .50262 .448 L p F P
0 g
s
.156 0 0 r
.51827 .4501 m .51383 .44611 L .53953 .44596 L p F P
0 g
s
.156 0 0 r
.53953 .44596 m .53424 .45001 L .51827 .4501 L p F P
0 g
s
.285 0 0 r
.50672 .65276 m .50206 .64678 L .45155 .64074 L p F P
0 g
s
.215 0 0 r
.50206 .64678 m .50672 .65276 L .54787 .65256 L p F P
0 g
s
.215 0 0 r
.54787 .65256 m .5534 .64652 L .50206 .64678 L p F P
0 g
s
.094 0 0 r
.5534 .64652 m .54787 .65256 L .58818 .64732 L p F P
0 g
s
.094 0 0 r
.58818 .64732 m .60376 .63997 L .5534 .64652 L p F P
0 g
s
.889 .538 .375 r
.29047 .52318 m .29839 .55816 L .34044 .58483 L p F P
0 g
s
.735 .209 0 r
.36023 .61049 m .34044 .58483 L .29839 .55816 L p F P
0 g
s
.74 .236 .028 r
.34044 .58483 m .36023 .61049 L .40367 .62854 L p F P
0 g
s
.923 .711 .599 r
.26451 .45328 m .25582 .48925 L .29047 .52318 L p F P
0 g
s
.935 .57 .288 r
.29839 .55816 m .29047 .52318 L .25582 .48925 L p F P
0 g
s
.453 .643 .939 r
.41263 .39797 m .36694 .39199 L .38622 .41322 L p F P
0 g
s
.453 .643 .939 r
.38622 .41322 m .42603 .41308 L .41263 .39797 L p F P
0 g
s
.246 0 0 r
.48796 .44378 m .4649 .43589 L .4886 .44271 L p F P
0 g
s
.246 0 0 r
.4886 .44271 m .50262 .448 L .48796 .44378 L p F P
0 g
s
0 0 0 r
.71455 .58286 m .69431 .60875 L .73091 .58492 L p F P
0 g
s
0 0 0 r
.73091 .58492 m .75593 .55571 L .71455 .58286 L p F P
0 g
s
.021 0 0 r
.56467 .44228 m .54984 .44773 L .53424 .45001 L p F P
0 g
s
.092 0 .277 r
.764 .52059 m .75593 .55571 L .78841 .52319 L p F P
0 g
s
.092 0 .277 r
.78841 .52319 m .79779 .48622 L .764 .52059 L p F P
0 g
s
.72 .902 .96 r
.66685 .41159 m .62647 .41192 L .6091 .42493 L p F P
0 g
s
.758 .889 .937 r
.62647 .41192 m .66685 .41159 L .68558 .39011 L p F P
0 g
s
.687 .752 .902 r
.72895 .39867 m .68558 .39011 L .66685 .41159 L p F P
0 g
s
.302 0 0 r
.46627 .64794 m .45155 .64074 L .40367 .62854 L p F P
0 g
s
.285 0 0 r
.45155 .64074 m .46627 .64794 L .50672 .65276 L p F P
0 g
s
.771 .77 .851 r
.30747 .37087 m .26827 .3874 L .28844 .42257 L p F P
0 g
s
.771 .77 .851 r
.28844 .42257 m .32399 .40104 L .30747 .37087 L p F P
0 g
s
.879 .798 .766 r
.26451 .45328 m .28844 .42257 L .26827 .3874 L p F P
0 g
s
0 0 0 r
.60376 .63997 m .58818 .64732 L .62612 .63722 L p F P
0 g
s
0 0 0 r
.62612 .63722 m .65133 .62727 L .60376 .63997 L p F P
0 g
s
.647 .716 .898 r
.35454 .36822 m .30747 .37087 L .32399 .40104 L p F P
0 g
s
.647 .716 .898 r
.32399 .40104 m .36694 .39199 L .35454 .36822 L p F P
0 g
s
0 .367 .737 r
.47494 .43762 m .44373 .42589 L .4649 .43589 L p F P
0 g
s
0 .367 .737 r
.4649 .43589 m .48337 .44221 L p .47494 .43762 L F P
0 g
s
0 .401 .83 r
.44373 .42589 m .46874 .43529 L p .46416 .42977 L F P
0 g
s
0 .401 .83 r
.46874 .43529 m .47494 .43762 L .46416 .42977 L p F P
0 g
s
.673 .137 0 r
.49616 .46444 m .46782 .43375 L p .47494 .43762 L F P
0 g
s
.673 .137 0 r
.46782 .43375 m .46416 .42977 L .47494 .43762 L p F P
0 g
s
0 0 .504 r
.49239 .46013 m .46028 .42511 L p .46416 .42977 L F P
0 g
s
0 0 .504 r
.46028 .42511 m .45611 .42056 L .46416 .42977 L p F P
0 g
s
0 .367 .737 r
.48337 .44221 m .48796 .44378 L .47494 .43762 L p F P
0 g
s
.587 .621 .853 r
.72895 .39867 m .76485 .41981 L .78412 .38435 L p F P
0 g
s
.45 .451 .768 r
.76485 .41981 m .78902 .45029 L .81079 .41182 L p F P
0 g
s
.45 .451 .768 r
.81079 .41182 m .78412 .38435 L .76485 .41981 L p F P
0 g
s
.69 .75 .898 r
.68558 .39011 m .72895 .39867 L .74469 .36827 L p F P
0 g
s
.587 .621 .853 r
.78412 .38435 m .74469 .36827 L .72895 .39867 L p F P
0 g
s
.879 .798 .766 r
.26827 .3874 m .24175 .41514 L .26451 .45328 L p F P
0 g
s
.973 .779 .583 r
.25582 .48925 m .26451 .45328 L .24175 .41514 L p F P
0 g
s
0 .401 .83 r
.46416 .42977 m .42603 .41308 L .44373 .42589 L p F P
0 g
s
.214 .163 .58 r
.78902 .45029 m .79779 .48622 L .8205 .44724 L p F P
0 g
s
.214 .163 .58 r
.8205 .44724 m .81079 .41182 L .78902 .45029 L p F P
0 g
s
.302 0 0 r
.40367 .62854 m .42807 .63823 L .46627 .64794 L p F P
0 g
s
0 .325 .668 r
.42807 .63823 m .40367 .62854 L .36023 .61049 L p F P
0 g
s
.304 .779 .678 r
.65133 .62727 m .62612 .63722 L .66018 .62254 L p F P
0 g
s
.304 .779 .678 r
.66018 .62254 m .69431 .60875 L .65133 .62727 L p F P
0 g
s
.481 .624 .914 r
.40424 .38116 m .35454 .36822 L .36694 .39199 L p F P
0 g
s
.481 .624 .914 r
.36694 .39199 m .41263 .39797 L .40424 .38116 L p F P
0 g
s
.758 .889 .937 r
.68558 .39011 m .63947 .39664 L .62647 .41192 L p F P
0 g
s
.781 .866 .912 r
.63947 .39664 m .68558 .39011 L .69736 .36617 L p F P
0 g
s
.69 .75 .898 r
.74469 .36827 m .69736 .36617 L .68558 .39011 L p F P
0 g
s
.935 .57 .288 r
.25582 .48925 m .26508 .52605 L .29839 .55816 L p F P
0 g
s
.623 .116 0 r
.32302 .58707 m .29839 .55816 L .26508 .52605 L p F P
0 g
s
.735 .209 0 r
.29839 .55816 m .32302 .58707 L .36023 .61049 L p F P
0 g
s
0 0 0 r
.75593 .55571 m .73091 .58492 L .75937 .5565 L p F P
0 g
s
0 0 0 r
.75937 .5565 m .78841 .52319 L .75593 .55571 L p F P
0 g
s
.083 .432 .873 r
.45611 .42056 m .41263 .39797 L .42603 .41308 L p F P
0 g
s
.083 .432 .873 r
.42603 .41308 m .46416 .42977 L .45611 .42056 L p F P
0 g
s
.781 .994 .875 r
.5878 .42906 m .62647 .41192 L .63947 .39664 L p F P
0 g
s
.973 .779 .583 r
.24175 .41514 m .23209 .45059 L .25582 .48925 L p F P
0 g
s
.917 .56 .113 r
.26508 .52605 m .25582 .48925 L .23209 .45059 L p F P
0 g
s
.649 .712 .894 r
.34982 .34288 m .30063 .33849 L .30747 .37087 L p F P
0 g
s
.649 .712 .894 r
.30747 .37087 m .35454 .36822 L .34982 .34288 L p F P
0 g
s
.773 .792 .869 r
.26827 .3874 m .30747 .37087 L .30063 .33849 L p F P
0 g
s
0 0 .091 r
.79779 .48622 m .78841 .52319 L .81012 .48637 L p F P
0 g
s
0 0 .091 r
.81012 .48637 m .8205 .44724 L .79779 .48622 L p F P
0 g
s
0 .371 .785 r
.39362 .62391 m .36023 .61049 L .32302 .58707 L p F P
0 g
s
0 .325 .668 r
.36023 .61049 m .39362 .62391 L .42807 .63823 L p F P
0 g
s
.693 .747 .893 r
.69736 .36617 m .74469 .36827 L .75066 .33576 L p F P
0 g
s
.568 .627 .87 r
.74469 .36827 m .78412 .38435 L .792 .3462 L p F P
0 g
s
.568 .627 .87 r
.792 .3462 m .75066 .33576 L .74469 .36827 L p F P
0 g
s
.773 .792 .869 r
.30063 .33849 m .25937 .34941 L .26827 .3874 L p F P
0 g
s
.901 .858 .79 r
.24175 .41514 m .26827 .3874 L .25937 .34941 L p F P
0 g
s
.539 .921 .794 r
.69431 .60875 m .66018 .62254 L .68889 .60374 L p F P
0 g
s
.539 .921 .794 r
.68889 .60374 m .73091 .58492 L .69431 .60875 L p F P
0 g
s
.371 .43 .792 r
.78412 .38435 m .81079 .41182 L .82011 .37027 L p F P
0 g
s
.371 .43 .792 r
.82011 .37027 m .792 .3462 L .78412 .38435 L p F P
0 g
s
.511 .612 .889 r
.40139 .36337 m .34982 .34288 L .35454 .36822 L p F P
0 g
s
.511 .612 .889 r
.35454 .36822 m .40424 .38116 L .40139 .36337 L p F P
0 g
s
.222 .455 .872 r
.45121 .41038 m .40424 .38116 L .41263 .39797 L p F P
0 g
s
.222 .455 .872 r
.41263 .39797 m .45611 .42056 L .45121 .41038 L p F P
0 g
s
.781 .866 .912 r
.69736 .36617 m .6474 .37972 L .63947 .39664 L p F P
0 g
s
.791 .838 .887 r
.6474 .37972 m .69736 .36617 L .70139 .34075 L p F P
0 g
s
.693 .747 .893 r
.75066 .33576 m .70139 .34075 L .69736 .36617 L p F P
0 g
s
.781 .994 .875 r
.63947 .39664 m .5956 .41975 L .5878 .42906 L p F P
0 g
s
.856 .966 .871 r
.5956 .41975 m .63947 .39664 L .6474 .37972 L p F P
0 g
s
.627 .147 .095 r
.51452 .47305 m .50262 .448 L .51827 .4501 L p F P
0 g
s
.627 .147 .095 r
.51827 .4501 m .52205 .47405 L .51452 .47305 L p F P
0 g
s
.549 .088 .112 r
.52205 .47405 m .51827 .4501 L .53424 .45001 L p F P
0 g
s
.549 .088 .112 r
.53424 .45001 m .52974 .474 L .52205 .47405 L p F P
0 g
s
.447 0 .092 r
.52974 .474 m .53424 .45001 L .54984 .44773 L p F P
0 g
s
.685 .181 .039 r
.50748 .47105 m .48796 .44378 L .50262 .448 L p F P
0 g
s
.685 .181 .039 r
.50262 .448 m .51452 .47305 L .50748 .47105 L p F P
0 g
s
.447 0 .092 r
.54984 .44773 m .53724 .47292 L .52974 .474 L p F P
0 g
s
.305 0 .032 r
.53724 .47292 m .54984 .44773 L .56438 .44334 L p F P
0 g
s
.901 .858 .79 r
.25937 .34941 m .23132 .37377 L .24175 .41514 L p F P
0 g
s
.998 .847 .533 r
.23209 .45059 m .24175 .41514 L .23132 .37377 L p F P
0 g
s
.313 .747 .969 r
.51319 .64039 m .50672 .65276 L .46627 .64794 L p F P
0 g
s
.359 .792 .967 r
.50672 .65276 m .51319 .64039 L .54018 .64025 L p F P
0 g
s
.359 .792 .967 r
.54018 .64025 m .54787 .65256 L .50672 .65276 L p F P
0 g
s
.43 .84 .97 r
.54787 .65256 m .54018 .64025 L .56658 .63679 L p F P
0 g
s
.43 .84 .97 r
.56658 .63679 m .58818 .64732 L .54787 .65256 L p F P
0 g
s
.714 .184 0 r
.50127 .46814 m .47494 .43762 L .48796 .44378 L p F P
0 g
s
.714 .184 0 r
.48796 .44378 m .50748 .47105 L .50127 .46814 L p F P
0 g
s
.305 0 .032 r
.56438 .44334 m .54422 .47084 L .53724 .47292 L p F P
0 g
s
0 .006 .525 r
.81079 .41182 m .8205 .44724 L .83037 .40506 L p F P
0 g
s
0 .006 .525 r
.83037 .40506 m .82011 .37027 L .81079 .41182 L p F P
0 g
s
.297 .709 .974 r
.48669 .6372 m .46627 .64794 L .42807 .63823 L p F P
0 g
s
.313 .747 .969 r
.46627 .64794 m .48669 .6372 L .51319 .64039 L p F P
0 g
s
.516 .885 .975 r
.58818 .64732 m .56658 .63679 L .5913 .63014 L p F P
0 g
s
.516 .885 .975 r
.5913 .63014 m .62612 .63722 L .58818 .64732 L p F P
0 g
s
.673 .137 0 r
.47494 .43762 m .50127 .46814 L .49616 .46444 L p F P
0 g
s
.333 .472 .85 r
.44974 .39969 m .40139 .36337 L .40424 .38116 L p F P
0 g
s
.333 .472 .85 r
.40424 .38116 m .45121 .41038 L .44974 .39969 L p F P
0 g
s
.917 .56 .113 r
.23209 .45059 m .24238 .48954 L .26508 .52605 L p F P
0 g
s
.623 .116 0 r
.26508 .52605 m .2938 .559 L .32302 .58707 L p F P
0 g
s
0 .046 .6 r
.2938 .559 m .26508 .52605 L .24238 .48954 L p F P
0 g
s
.856 .966 .871 r
.6474 .37972 m .60021 .40951 L .5956 .41975 L p F P
0 g
s
.878 .911 .847 r
.60021 .40951 m .6474 .37972 L .64977 .36189 L p F P
0 g
s
.791 .838 .887 r
.70139 .34075 m .64977 .36189 L .6474 .37972 L p F P
0 g
s
.016 .419 .86 r
.36439 .60543 m .32302 .58707 L .2938 .559 L p F P
0 g
s
0 .371 .785 r
.32302 .58707 m .36439 .60543 L .39362 .62391 L p F P
0 g
s
.76 .946 .617 r
.78841 .52319 m .75937 .5565 L .77809 .52445 L p F P
0 g
s
.76 .946 .617 r
.77809 .52445 m .81012 .48637 L .78841 .52319 L p F P
0 g
s
.723 .989 .864 r
.73091 .58492 m .68889 .60374 L .71089 .58147 L p F P
0 g
s
.723 .989 .864 r
.71089 .58147 m .75937 .5565 L .73091 .58492 L p F P
0 g
s
.652 .709 .89 r
.35334 .31711 m .30428 .3053 L .30063 .33849 L p F P
0 g
s
.652 .709 .89 r
.30063 .33849 m .34982 .34288 L .35334 .31711 L p F P
0 g
s
.77 .814 .888 r
.25937 .34941 m .30063 .33849 L .30428 .3053 L p F P
0 g
s
.54 .605 .867 r
.40435 .34543 m .35334 .31711 L .34982 .34288 L p F P
0 g
s
.54 .605 .867 r
.34982 .34288 m .40139 .36337 L .40435 .34543 L p F P
0 g
s
0 0 .504 r
.46416 .42977 m .4929 .46091 L p .49239 .46013 L F P
0 g
s
0 .135 .682 r
.45611 .42056 m .4887 .45611 L p .49016 .45539 L F P
0 g
s
0 .135 .682 r
.4887 .45611 m .49239 .46013 L .49016 .45539 L p F P
0 g
s
.674 .311 0 r
.51623 .51425 m .49196 .45945 L p .49239 .46013 L F P
0 g
s
.674 .311 0 r
.49196 .45945 m .49016 .45539 L .49239 .46013 L p F P
0 g
s
0 0 .504 r
.4929 .46091 m .49616 .46444 L .49239 .46013 L p F P
0 g
s
.869 .442 .022 r
.5168 .51546 m .49239 .46013 L .49616 .46444 L p F P
0 g
s
.869 .442 .022 r
.49616 .46444 m .51778 .51657 L .5168 .51546 L p F P
0 g
s
.674 .311 0 r
.49239 .46013 m .51616 .51401 L p .51623 .51425 L F P
0 g
s
.674 .311 0 r
.51616 .51401 m .5168 .51546 L .51623 .51425 L p F P
0 g
s
.939 .638 .18 r
.5168 .51546 m .52553 .58349 L .51623 .51425 L closepath p F P
0 g
s
0 0 .118 r
.51623 .51425 m .52553 .58349 L .51629 .51447 L p F P
0 g
s
0 0 .118 r
.51629 .51447 m .51609 .51299 L .51623 .51425 L p F P
0 g
s
0 0 .51 r
.51609 .51299 m .4914 .45477 L p .49016 .45539 L F P
0 g
s
0 0 .51 r
.4914 .45477 m .48956 .45044 L .49016 .45539 L p F P
0 g
s
.308 .679 .976 r
.46179 .63081 m .42807 .63823 L .39362 .62391 L p F P
0 g
s
.297 .709 .974 r
.42807 .63823 m .46179 .63081 L .48669 .6372 L p F P
0 g
s
.792 .808 .865 r
.64977 .36189 m .70139 .34075 L .69716 .315 L p F P
0 g
s
.694 .743 .889 r
.70139 .34075 m .75066 .33576 L .74609 .30258 L p F P
0 g
s
.694 .743 .889 r
.74609 .30258 m .69716 .315 L .70139 .34075 L p F P
0 g
s
.55 .637 .89 r
.75066 .33576 m .792 .3462 L .78742 .30701 L p F P
0 g
s
.55 .637 .89 r
.78742 .30701 m .74609 .30258 L .75066 .33576 L p F P
0 g
s
.605 .917 .976 r
.62612 .63722 m .5913 .63014 L .61329 .62053 L p F P
0 g
s
.605 .917 .976 r
.61329 .62053 m .66018 .62254 L .62612 .63722 L p F P
0 g
s
.998 .847 .533 r
.23132 .37377 m .22107 .40862 L .23209 .45059 L p F P
0 g
s
0 0 .182 r
.24238 .48954 m .23209 .45059 L .22107 .40862 L p F P
0 g
s
.77 .814 .888 r
.30428 .3053 m .26286 .31024 L .25937 .34941 L p F P
0 g
s
.907 .923 .816 r
.23132 .37377 m .25937 .34941 L .26286 .31024 L p F P
0 g
s
.722 .731 .21 r
.8205 .44724 m .81012 .48637 L .8194 .44659 L p F P
0 g
s
.722 .731 .21 r
.8194 .44659 m .83037 .40506 L .8205 .44724 L p F P
0 g
s
0 .135 .682 r
.49016 .45539 m .45121 .41038 L .45611 .42056 L p F P
0 g
s
.419 .486 .823 r
.45182 .38899 m .40435 .34543 L .40139 .36337 L p F P
0 g
s
.419 .486 .823 r
.40139 .36337 m .44974 .39969 L .45182 .38899 L p F P
0 g
s
.854 .982 .692 r
.559 .45975 m .5956 .41975 L .60021 .40951 L p F P
0 g
s
.267 .406 .818 r
.792 .3462 m .82011 .37027 L .81568 .3274 L p F P
0 g
s
.267 .406 .818 r
.81568 .3274 m .78742 .30701 L .792 .3462 L p F P
0 g
s
.878 .911 .847 r
.64977 .36189 m .60138 .3988 L .60021 .40951 L p F P
0 g
s
.872 .852 .821 r
.60138 .3988 m .64977 .36189 L .64631 .34397 L p F P
0 g
s
.792 .808 .865 r
.69716 .315 m .64631 .34397 L .64977 .36189 L p F P
0 g
s
.338 .654 .97 r
.43952 .62143 m .39362 .62391 L .36439 .60543 L p F P
0 g
s
.308 .679 .976 r
.39362 .62391 m .43952 .62143 L .46179 .63081 L p F P
0 g
s
.682 .93 .969 r
.66018 .62254 m .61329 .62053 L .63158 .60831 L p F P
0 g
s
.682 .93 .969 r
.63158 .60831 m .68889 .60374 L .66018 .62254 L p F P
0 g
s
.08 .256 .744 r
.48956 .45044 m .44974 .39969 L .45121 .41038 L p F P
0 g
s
.08 .256 .744 r
.45121 .41038 m .49016 .45539 L .48956 .45044 L p F P
0 g
s
.854 .982 .692 r
.60021 .40951 m .5611 .45498 L .559 .45975 L p F P
0 g
s
.943 .953 .746 r
.5611 .45498 m .60021 .40951 L .60138 .3988 L p F P
0 g
s
.732 .641 .758 r
.46457 .73769 m .44767 .80486 L .50081 .81056 L p F P
0 g
s
.732 .641 .758 r
.50081 .81056 m .50616 .74237 L .46457 .73769 L p F P
0 g
s
.74 .601 .703 r
.4972 .87401 m .50081 .81056 L .44767 .80486 L p F P
0 g
s
.704 .616 .756 r
.50616 .74237 m .50081 .81056 L .5548 .81032 L p F P
0 g
s
.706 .572 .702 r
.50081 .81056 m .4972 .87401 L .55905 .87374 L p F P
0 g
s
.706 .572 .702 r
.55905 .87374 m .5548 .81032 L .50081 .81056 L p F P
0 g
s
.704 .616 .756 r
.5548 .81032 m .54847 .74217 L .50616 .74237 L p F P
0 g
s
.675 .594 .758 r
.54847 .74217 m .5548 .81032 L .60778 .80412 L p F P
0 g
s
.67 .544 .703 r
.5548 .81032 m .55905 .87374 L .61981 .86695 L p F P
0 g
s
.67 .544 .703 r
.61981 .86695 m .60778 .80412 L .5548 .81032 L p F P
0 g
s
.74 .601 .703 r
.44767 .80486 m .43627 .86776 L .4972 .87401 L p F P
0 g
s
.743 .55 .634 r
.49548 .92772 m .4972 .87401 L .43627 .86776 L p F P
0 g
s
.702 .516 .634 r
.4972 .87401 m .49548 .92772 L .56107 .92744 L p F P
0 g
s
.702 .516 .634 r
.56107 .92744 m .55905 .87374 L .4972 .87401 L p F P
0 g
s
.659 .482 .634 r
.55905 .87374 m .56107 .92744 L .62552 .92052 L p F P
0 g
s
.659 .482 .634 r
.62552 .92052 m .61981 .86695 L .55905 .87374 L p F P
0 g
s
.569 .603 .847 r
.41315 .32819 m .36523 .29215 L .35334 .31711 L p F P
0 g
s
.569 .603 .847 r
.35334 .31711 m .40435 .34543 L .41315 .32819 L p F P
0 g
s
.655 .707 .886 r
.30428 .3053 m .35334 .31711 L .36523 .29215 L p F P
0 g
s
.907 .923 .816 r
.26286 .31024 m .23455 .33093 L .23132 .37377 L p F P
0 g
s
.9 .844 .387 r
.22107 .40862 m .23132 .37377 L .23455 .33093 L p F P
0 g
s
.786 .778 .846 r
.64631 .34397 m .69716 .315 L .68456 .29017 L p F P
0 g
s
.694 .739 .885 r
.69716 .315 m .74609 .30258 L .73064 .27034 L p F P
0 g
s
.694 .739 .885 r
.73064 .27034 m .68456 .29017 L .69716 .315 L p F P
0 g
s
0 .046 .6 r
.24238 .48954 m .27419 .5272 L .2938 .559 L p F P
0 g
s
.17 .454 .881 r
.34178 .58342 m .2938 .559 L .27419 .5272 L p F P
0 g
s
.016 .419 .86 r
.2938 .559 m .34178 .58342 L .36439 .60543 L p F P
0 g
s
.485 .501 .799 r
.45742 .37877 m .41315 .32819 L .40435 .34543 L p F P
0 g
s
.485 .501 .799 r
.40435 .34543 m .45182 .38899 L .45742 .37877 L p F P
0 g
s
.759 .669 .762 r
.42527 .72826 m .39725 .79333 L .44767 .80486 L p F P
0 g
s
.759 .669 .762 r
.44767 .80486 m .46457 .73769 L .42527 .72826 L p F P
0 g
s
.773 .634 .707 r
.43627 .86776 m .44767 .80486 L .39725 .79333 L p F P
0 g
s
0 0 .367 r
.82011 .37027 m .83037 .40506 L .82602 .36148 L p F P
0 g
s
0 0 .367 r
.82602 .36148 m .81568 .3274 L .82011 .37027 L p F P
0 g
s
.872 .852 .821 r
.64631 .34397 m .599 .38812 L .60138 .3988 L p F P
0 g
s
.853 .799 .797 r
.599 .38812 m .64631 .34397 L .63701 .32682 L p F P
0 g
s
.786 .778 .846 r
.68456 .29017 m .63701 .32682 L .64631 .34397 L p F P
0 g
s
.826 .982 .881 r
.75937 .5565 m .71089 .58147 L .72494 .55653 L p F P
0 g
s
.826 .982 .881 r
.72494 .55653 m .77809 .52445 L .75937 .5565 L p F P
0 g
s
.675 .594 .758 r
.60778 .80412 m .58993 .73709 L .54847 .74217 L p F P
0 g
s
.643 .575 .763 r
.58993 .73709 m .60778 .80412 L .65788 .79212 L p F P
0 g
s
.63 .518 .707 r
.60778 .80412 m .61981 .86695 L .67742 .85378 L p F P
0 g
s
.63 .518 .707 r
.67742 .85378 m .65788 .79212 L .60778 .80412 L p F P
0 g
s
.268 .334 .747 r
.49066 .44551 m .45182 .38899 L .44974 .39969 L p F P
0 g
s
.268 .334 .747 r
.44974 .39969 m .48956 .45044 L .49066 .44551 L p F P
0 g
s
.655 .707 .886 r
.36523 .29215 m .31881 .27292 L .30428 .3053 L p F P
0 g
s
.759 .833 .909 r
.26286 .31024 m .30428 .3053 L .31881 .27292 L p F P
0 g
s
.773 .634 .707 r
.39725 .79333 m .37831 .8551 L .43627 .86776 L p F P
0 g
s
.783 .587 .635 r
.43085 .92134 m .43627 .86776 L .37831 .8551 L p F P
0 g
s
.743 .55 .634 r
.43627 .86776 m .43085 .92134 L .49548 .92772 L p F P
0 g
s
0 0 .182 r
.22107 .40862 m .23199 .44995 L .24238 .48954 L p F P
0 g
s
0 .202 .723 r
.27419 .5272 m .24238 .48954 L .23199 .44995 L p F P
0 g
s
.943 .953 .746 r
.60138 .3988 m .56155 .45003 L .5611 .45498 L p F P
0 g
s
.941 .879 .746 r
.56155 .45003 m .60138 .3988 L .599 .38812 L p F P
0 g
s
.611 .447 .636 r
.61981 .86695 m .62552 .92052 L .68672 .90709 L p F P
0 g
s
.611 .447 .636 r
.68672 .90709 m .67742 .85378 L .61981 .86695 L p F P
0 g
s
.535 .652 .91 r
.74609 .30258 m .78742 .30701 L .76984 .2687 L p F P
0 g
s
.535 .652 .91 r
.76984 .2687 m .73064 .27034 L .74609 .30258 L p F P
0 g
s
.922 .976 .728 r
.81012 .48637 m .77809 .52445 L .78566 .49 L p F P
0 g
s
.922 .976 .728 r
.78566 .49 m .8194 .44659 L .81012 .48637 L p F P
0 g
s
.719 .677 .809 r
.48627 .67122 m .46457 .73769 L .50616 .74237 L p F P
0 g
s
.719 .677 .809 r
.50616 .74237 m .51304 .67437 L .48627 .67122 L p F P
0 g
s
.698 .659 .808 r
.51304 .67437 m .50616 .74237 L .54847 .74217 L p F P
0 g
s
.698 .659 .808 r
.54847 .74217 m .54033 .67424 L .51304 .67437 L p F P
0 g
s
.676 .642 .809 r
.54033 .67424 m .54847 .74217 L .58993 .73709 L p F P
0 g
s
.713 .354 .365 r
.52262 .51878 m .51452 .47305 L .52205 .47405 L p F P
0 g
s
.713 .354 .365 r
.52205 .47405 m .52462 .51904 L .52262 .51878 L p F P
0 g
s
.647 .301 .372 r
.52462 .51904 m .52205 .47405 L .52974 .474 L p F P
0 g
s
.647 .301 .372 r
.52974 .474 m .52666 .51903 L .52462 .51904 L p F P
0 g
s
.571 .237 .364 r
.52666 .51903 m .52974 .474 L .53724 .47292 L p F P
0 g
s
.775 .398 .342 r
.52076 .51827 m .50748 .47105 L .51452 .47305 L p F P
0 g
s
.775 .398 .342 r
.51452 .47305 m .52262 .51878 L .52076 .51827 L p F P
0 g
s
.571 .237 .364 r
.53724 .47292 m .52865 .51875 L .52666 .51903 L p F P
0 g
s
.475 .152 .339 r
.52865 .51875 m .53724 .47292 L .54422 .47084 L p F P
0 g
s
.379 .635 .955 r
.42088 .60941 m .36439 .60543 L .34178 .58342 L p F P
0 g
s
.338 .654 .97 r
.36439 .60543 m .42088 .60941 L .43952 .62143 L p F P
0 g
s
.735 .467 .523 r
.4957 .96673 m .49548 .92772 L .43085 .92134 L p F P
0 g
s
.686 .426 .524 r
.49548 .92772 m .4957 .96673 L .56079 .96647 L p F P
0 g
s
.686 .426 .524 r
.56079 .96647 m .56107 .92744 L .49548 .92772 L p F P
0 g
s
.633 .383 .522 r
.56107 .92744 m .56079 .96647 L .62475 .95981 L p F P
0 g
s
.633 .383 .522 r
.62475 .95981 m .62552 .92052 L .56107 .92744 L p F P
0 g
s
.834 .436 .294 r
.51912 .51752 m .50127 .46814 L .50748 .47105 L p F P
0 g
s
.834 .436 .294 r
.50748 .47105 m .52076 .51827 L .51912 .51752 L p F P
0 g
s
.475 .152 .339 r
.54422 .47084 m .53049 .51821 L .52865 .51875 L p F P
0 g
s
.34 .032 .288 r
.53049 .51821 m .54422 .47084 L .55035 .46786 L p F P
0 g
s
.738 .923 .953 r
.68889 .60374 m .63158 .60831 L .64529 .59396 L p F P
0 g
s
.738 .923 .953 r
.64529 .59396 m .71089 .58147 L .68889 .60374 L p F P
0 g
s
.738 .699 .814 r
.46109 .6649 m .42527 .72826 L .46457 .73769 L p F P
0 g
s
.738 .699 .814 r
.46457 .73769 m .48627 .67122 L .46109 .6649 L p F P
0 g
s
.393 .387 .734 r
.4934 .44083 m .45742 .37877 L .45182 .38899 L p F P
0 g
s
.393 .387 .734 r
.45182 .38899 m .49066 .44551 L .4934 .44083 L p F P
0 g
s
.676 .642 .809 r
.58993 .73709 m .56701 .67082 L .54033 .67424 L p F P
0 g
s
.652 .629 .814 r
.56701 .67082 m .58993 .73709 L .62897 .72728 L p F P
0 g
s
.643 .575 .763 r
.65788 .79212 m .62897 .72728 L .58993 .73709 L p F P
0 g
s
.941 .879 .746 r
.599 .38812 m .56031 .44511 L .56155 .45003 L p F P
0 g
s
.909 .805 .733 r
.56031 .44511 m .599 .38812 L .59311 .37797 L p F P
0 g
s
.853 .799 .797 r
.63701 .32682 m .59311 .37797 L .599 .38812 L p F P
0 g
s
.88 .46 .2 r
.51778 .51657 m .49616 .46444 L .50127 .46814 L p F P
0 g
s
.88 .46 .2 r
.50127 .46814 m .51912 .51752 L .51778 .51657 L p F P
0 g
s
.34 .032 .288 r
.55035 .46786 m .53211 .51744 L .53049 .51821 L p F P
0 g
s
.132 0 .189 r
.53211 .51744 m .55035 .46786 L .55536 .46411 L p F P
0 g
s
.785 .7 .771 r
.3898 .71435 m .35143 .77625 L .39725 .79333 L p F P
0 g
s
.785 .7 .771 r
.39725 .79333 m .42527 .72826 L .3898 .71435 L p F P
0 g
s
.807 .671 .714 r
.37831 .8551 m .39725 .79333 L .35143 .77625 L p F P
0 g
s
.539 .517 .779 r
.46632 .36956 m .42752 .31252 L .41315 .32819 L p F P
0 g
s
.539 .517 .779 r
.41315 .32819 m .45742 .37877 L .46632 .36956 L p F P
0 g
s
.596 .605 .831 r
.36523 .29215 m .41315 .32819 L .42752 .31252 L p F P
0 g
s
.608 .557 .772 r
.62897 .72728 m .65788 .79212 L .70322 .77461 L p F P
0 g
s
.586 .492 .714 r
.65788 .79212 m .67742 .85378 L .72981 .83449 L p F P
0 g
s
.586 .492 .714 r
.72981 .83449 m .70322 .77461 L .65788 .79212 L p F P
0 g
s
.83 .752 .778 r
.59311 .37797 m .63701 .32682 L .6222 .31132 L p F P
0 g
s
.776 .75 .83 r
.63701 .32682 m .68456 .29017 L .6639 .26755 L p F P
0 g
s
.776 .75 .83 r
.6639 .26755 m .6222 .31132 L .63701 .32682 L p F P
0 g
s
.759 .833 .909 r
.31881 .27292 m .27934 .27177 L .26286 .31024 L p F P
0 g
s
.877 .981 .834 r
.23455 .33093 m .26286 .31024 L .27934 .27177 L p F P
0 g
s
.568 .785 .981 r
.52115 .60758 m .51319 .64039 L .48669 .6372 L p F P
0 g
s
.568 .785 .981 r
.51178 .60643 m .52115 .60758 L p .52031 .60831 L F P
0 g
s
.568 .785 .981 r
.48669 .6372 m .51178 .60643 L p .52031 .60831 L F P
0 g
s
.555 .766 .976 r
.50302 .60413 m .51178 .60643 L p .51055 .60702 L F P
0 g
s
.555 .766 .976 r
.46179 .63081 m .50302 .60413 L p .51055 .60702 L F P
0 g
s
.549 .747 .969 r
.49527 .60077 m .50302 .60413 L p .50147 .60455 L F P
0 g
s
.549 .747 .969 r
.43952 .62143 m .49527 .60077 L p .50147 .60455 L F P
0 g
s
.549 .73 .959 r
.48888 .59651 m .49527 .60077 L p .49346 .60098 L F P
0 g
s
.549 .73 .959 r
.42088 .60941 m .48888 .59651 L p .49346 .60098 L F P
0 g
s
.554 .714 .947 r
.48417 .59153 m .48888 .59651 L p .4869 .59648 L F P
0 g
s
.554 .714 .947 r
.40675 .59521 m .48417 .59153 L p .4869 .59648 L F P
0 g
s
.563 .701 .933 r
.48135 .58604 m .48417 .59153 L p .4821 .59124 L F P
0 g
s
.563 .701 .933 r
.39787 .57941 m .48135 .58604 L p .4821 .59124 L F P
0 g
s
.575 .691 .92 r
.48058 .58031 m .48135 .58604 L p .47929 .58549 L F P
0 g
s
.575 .691 .92 r
.3948 .56268 m .48058 .58031 L p .47929 .58549 L F P
0 g
s
0 0 .174 r
.48058 .58031 m .48192 .5746 L p .48179 .57544 L F P
0 g
s
.852 .675 .14 r
.48135 .58604 m .48058 .58031 L p .48064 .58116 L F P
0 g
s
.575 .691 .92 r
.48135 .58604 m .39787 .57941 L .3948 .56268 L p F P
0 g
s
.852 .675 .14 r
.48129 .58998 m .48135 .58604 L p .48064 .58116 L F P
0 g
s
.852 .675 .14 r
.48052 .58425 m .48129 .58998 L p .48123 .58914 L F P
0 g
s
.852 .675 .14 r
.48058 .58031 m .48052 .58425 L p .48123 .58914 L F P
0 g
s
0 0 .174 r
.48052 .58425 m .48058 .58031 L p .48179 .57544 L F P
0 g
s
.589 .684 .907 r
.48058 .58031 m .3948 .56268 L .39786 .54578 L p F P
0 g
s
.589 .684 .907 r
.39786 .54578 m .48192 .5746 L .48058 .58031 L p F P
0 g
s
0 0 .174 r
.48186 .57854 m .48052 .58425 L p .48064 .58341 L F P
0 g
s
0 0 .174 r
.48192 .5746 m .48186 .57854 L p .48064 .58341 L F P
0 g
s
0 0 .376 r
.48186 .57854 m .48192 .5746 L .48532 .56918 L p F P
0 g
s
0 0 .376 r
.48526 .57313 m .48186 .57854 L p .48216 .57773 L F P
0 g
s
0 0 .376 r
.48532 .56918 m .48526 .57313 L p .48216 .57773 L F P
0 g
s
.255 .097 .465 r
.49059 .56826 m .48526 .57313 L p .48573 .57236 L F P
0 g
s
.255 .097 .465 r
.49064 .5643 m .49059 .56826 L p .48573 .57236 L F P
0 g
s
.412 .218 .504 r
.4976 .56417 m .49059 .56826 L p .4912 .56756 L F P
0 g
s
.412 .218 .504 r
.49764 .56021 m .4976 .56417 L p .4912 .56756 L F P
0 g
s
.517 .3 .523 r
.50597 .56108 m .4976 .56417 L p .49833 .56356 L F P
0 g
s
.517 .3 .523 r
.506 .55712 m .50597 .56108 L p .49833 .56356 L F P
0 g
s
.595 .362 .533 r
.51529 .55914 m .50597 .56108 L p .50678 .56057 L F P
0 g
s
.595 .362 .533 r
.5153 .55518 m .51529 .55914 L p .50678 .56057 L F P
0 g
s
.737 .452 .496 r
.52462 .51904 m .52553 .58349 L .52262 .51878 L closepath p F P
0 g
s
.793 .497 .487 r
.52262 .51878 m .52553 .58349 L .52076 .51827 L closepath p F P
0 g
s
.851 .543 .467 r
.52076 .51827 m .52553 .58349 L .51912 .51752 L closepath p F P
0 g
s
.911 .592 .429 r
.51912 .51752 m .52553 .58349 L .51778 .51657 L closepath p F P
0 g
s
.962 .638 .351 r
.51778 .51657 m .52553 .58349 L .5168 .51546 L closepath p F P
0 g
s
0 0 .379 r
.51609 .51299 m .52553 .58349 L .5164 .51175 L closepath p F P
0 g
s
.125 .054 .498 r
.5164 .51175 m .52553 .58349 L .51713 .51057 L closepath p F P
0 g
s
.327 .196 .544 r
.51713 .51057 m .52553 .58349 L .51826 .50951 L closepath p F P
0 g
s
.454 .287 .563 r
.51826 .50951 m .52553 .58349 L .51974 .50864 L closepath p F P
0 g
s
.543 .352 .572 r
.51974 .50864 m .52553 .58349 L .52149 .50797 L closepath p F P
0 g
s
.611 .404 .576 r
.52149 .50797 m .52553 .58349 L .52342 .50756 L closepath p F P
0 g
s
.68 .405 .498 r
.52666 .51903 m .52553 .58349 L .52462 .51904 L closepath p F P
0 g
s
.669 .449 .578 r
.52342 .50756 m .52553 .58349 L .52546 .50741 L closepath p F P
0 g
s
.659 .413 .537 r
.52511 .55845 m .51529 .55914 L p .51614 .55874 L F P
0 g
s
.659 .413 .537 r
.52511 .55448 m .52511 .55845 L p .51614 .55874 L F P
0 g
s
.618 .354 .495 r
.52865 .51875 m .52553 .58349 L .52666 .51903 L closepath p F P
0 g
s
.543 .292 .485 r
.53049 .51821 m .52553 .58349 L .52865 .51875 L closepath p F P
0 g
s
.445 .212 .465 r
.53211 .51744 m .52553 .58349 L .53049 .51821 L closepath p F P
0 g
s
.304 .097 .424 r
.53342 .51648 m .52553 .58349 L .53211 .51744 L closepath p F P
0 g
s
.076 0 .34 r
.53437 .51536 m .52553 .58349 L .53342 .51648 L closepath p F P
0 g
s
.132 0 .189 r
.55536 .46411 m .53342 .51648 L .53211 .51744 L p F P
0 g
s
.984 .797 .395 r
.53466 .51165 m .52553 .58349 L .535 .51289 L closepath p F P
0 g
s
.985 .754 .504 r
.53389 .51048 m .52553 .58349 L .53466 .51165 L closepath p F P
0 g
s
.932 .688 .547 r
.53273 .50943 m .52553 .58349 L .53389 .51048 L closepath p F P
0 g
s
.874 .629 .564 r
.53123 .50857 m .52553 .58349 L .53273 .50943 L closepath p F P
0 g
s
.82 .578 .572 r
.52946 .50793 m .52553 .58349 L .53123 .50857 L closepath p F P
0 g
s
.77 .533 .576 r
.52751 .50753 m .52553 .58349 L .52946 .50793 L closepath p F P
0 g
s
.72 .491 .578 r
.52546 .50741 m .52553 .58349 L .52751 .50753 L closepath p F P
0 g
s
.716 .459 .537 r
.53494 .55507 m .53495 .55904 L .52511 .55845 L p F P
0 g
s
.579 .793 .98 r
.52511 .55845 m .53495 .55904 L p .53471 .55814 L F P
0 g
s
.575 .427 .642 r
.52149 .50797 m .50996 .43047 L .50328 .43313 L p F P
0 g
s
.575 .427 .642 r
.50328 .43313 m .51974 .50864 L .52149 .50797 L p F P
0 g
s
.631 .466 .64 r
.52342 .50756 m .5174 .4288 L .50996 .43047 L p F P
0 g
s
.631 .466 .64 r
.50996 .43047 m .52149 .50797 L .52342 .50756 L p F P
0 g
s
.68 .502 .639 r
.52546 .50741 m .52523 .42821 L .5174 .4288 L p F P
0 g
s
.68 .502 .639 r
.5174 .4288 m .52342 .50756 L .52546 .50741 L p F P
0 g
s
.677 .699 .863 r
.52427 .48445 m .52511 .55448 L .5153 .55518 L p F P
0 g
s
.659 .413 .537 r
.51529 .55914 m .5153 .55518 L .52511 .55448 L p F P
0 g
s
.69 .709 .863 r
.55371 .48629 m .53494 .55507 L .52511 .55448 L p F P
0 g
s
.716 .459 .537 r
.52511 .55845 m .52511 .55448 L .53494 .55507 L p F P
0 g
s
.937 .91 .524 r
.5349 .51415 m .5611 .45498 L .56155 .45003 L p F P
0 g
s
.937 .91 .524 r
.56155 .45003 m .535 .51289 L .5349 .51415 L p F P
0 g
s
.99 .873 .618 r
.535 .51289 m .56155 .45003 L .56031 .44511 L p F P
0 g
s
.99 .873 .618 r
.56031 .44511 m .53466 .51165 L .535 .51289 L p F P
0 g
s
.956 .795 .642 r
.53466 .51165 m .56031 .44511 L .55743 .44047 L p F P
0 g
s
.906 .725 .646 r
.53389 .51048 m .55743 .44047 L .55304 .43632 L p F P
0 g
s
.956 .795 .642 r
.55743 .44047 m .53389 .51048 L .53466 .51165 L p F P
0 g
s
.906 .725 .646 r
.55304 .43632 m .53273 .50943 L .53389 .51048 L p F P
0 g
s
.857 .667 .644 r
.53273 .50943 m .55304 .43632 L .54732 .43287 L p F P
0 g
s
.857 .667 .644 r
.54732 .43287 m .53123 .50857 L .53273 .50943 L p F P
0 g
s
.811 .618 .642 r
.53123 .50857 m .54732 .43287 L .54057 .43029 L p F P
0 g
s
.811 .618 .642 r
.54057 .43029 m .52946 .50793 L .53123 .50857 L p F P
0 g
s
.767 .577 .64 r
.52946 .50793 m .54057 .43029 L .53308 .42871 L p F P
0 g
s
.767 .577 .64 r
.53308 .42871 m .52751 .50753 L .52946 .50793 L p F P
0 g
s
.724 .539 .639 r
.52751 .50753 m .53308 .42871 L .52523 .42821 L p F P
0 g
s
.724 .539 .639 r
.52523 .42821 m .52546 .50741 L .52751 .50753 L p F P
0 g
s
.69 .709 .863 r
.52511 .55448 m .52427 .48445 L .55371 .48629 L p F P
0 g
s
.579 .793 .98 r
.52425 .51993 m .52511 .55845 L p .53471 .55814 L F P
0 g
s
.599 .809 .98 r
.51529 .55914 m .52511 .55845 L p .52441 .55761 L F P
0 g
s
.664 .69 .865 r
.49491 .48662 m .5153 .55518 L .506 .55712 L p F P
0 g
s
.664 .69 .865 r
.506 .55712 m .46717 .49266 L .49491 .48662 L p F P
0 g
s
.595 .362 .533 r
.50597 .56108 m .506 .55712 L .5153 .55518 L p F P
0 g
s
0 0 .51 r
.49016 .45539 m .51623 .51425 L .51609 .51299 L p F P
0 g
s
.064 .123 .613 r
.5164 .51175 m .49066 .44551 L .48956 .45044 L p F P
0 g
s
.064 .123 .613 r
.48956 .45044 m .51609 .51299 L .5164 .51175 L p F P
0 g
s
.279 .246 .641 r
.51713 .51057 m .4934 .44083 L .49066 .44551 L p F P
0 g
s
.279 .246 .641 r
.49066 .44551 m .5164 .51175 L .51713 .51057 L p F P
0 g
s
.414 .325 .646 r
.51826 .50951 m .49767 .43664 L .4934 .44083 L p F P
0 g
s
.414 .325 .646 r
.4934 .44083 m .51713 .51057 L .51826 .50951 L p F P
0 g
s
.506 .381 .645 r
.51974 .50864 m .50328 .43313 L .49767 .43664 L p F P
0 g
s
.506 .381 .645 r
.49767 .43664 m .51826 .50951 L .51974 .50864 L p F P
0 g
s
.677 .699 .863 r
.5153 .55518 m .49491 .48662 L .52427 .48445 L p F P
0 g
s
.599 .809 .98 r
.49453 .52208 m .51529 .55914 L p .52441 .55761 L F P
0 g
s
.623 .822 .977 r
.50597 .56108 m .51529 .55914 L p .51416 .55843 L F P
0 g
s
.634 .68 .876 r
.44251 .50222 m .49764 .56021 L .49064 .5643 L p F P
0 g
s
.649 .684 .869 r
.46717 .49266 m .506 .55712 L .49764 .56021 L p F P
0 g
s
.649 .684 .869 r
.49764 .56021 m .44251 .50222 L .46717 .49266 L p F P
0 g
s
.517 .3 .523 r
.4976 .56417 m .49764 .56021 L .506 .55712 L p F P
0 g
s
.623 .822 .977 r
.46647 .52807 m .50597 .56108 L p .51416 .55843 L F P
0 g
s
.648 .831 .971 r
.4976 .56417 m .50597 .56108 L p .50448 .56054 L F P
0 g
s
.619 .679 .885 r
.42216 .51474 m .49064 .5643 L .48532 .56918 L p F P
0 g
s
.619 .679 .885 r
.48532 .56918 m .40707 .52952 L .42216 .51474 L p F P
0 g
s
.634 .68 .876 r
.49064 .5643 m .42216 .51474 L .44251 .50222 L p F P
0 g
s
.412 .218 .504 r
.49059 .56826 m .49064 .5643 L .49764 .56021 L p F P
0 g
s
.648 .831 .971 r
.44153 .53754 m .4976 .56417 L p .50448 .56054 L F P
0 g
s
.671 .834 .962 r
.49059 .56826 m .4976 .56417 L p .49582 .56384 L F P
0 g
s
.604 .68 .895 r
.40707 .52952 m .48532 .56918 L .48192 .5746 L p F P
0 g
s
.604 .68 .895 r
.48192 .5746 m .39786 .54578 L .40707 .52952 L p F P
0 g
s
.255 .097 .465 r
.48526 .57313 m .48532 .56918 L .49064 .5643 L p F P
0 g
s
.671 .834 .962 r
.42095 .54995 m .49059 .56826 L p .49582 .56384 L F P
0 g
s
.692 .832 .95 r
.48526 .57313 m .49059 .56826 L p .4886 .56817 L F P
0 g
s
.692 .832 .95 r
.4057 .5646 m .48526 .57313 L p .4886 .56817 L F P
0 g
s
.708 .825 .937 r
.48186 .57854 m .48526 .57313 L p .48317 .5733 L F P
0 g
s
.424 .62 .933 r
.327 .55867 m .39787 .57941 L .40675 .59521 L p F P
0 g
s
.563 .701 .933 r
.48417 .59153 m .40675 .59521 L .39787 .57941 L p F P
0 g
s
.708 .825 .937 r
.39639 .58071 m .48186 .57854 L p .48317 .5733 L F P
0 g
s
.72 .815 .924 r
.48052 .58425 m .48186 .57854 L p .47976 .57899 L F P
0 g
s
.424 .62 .933 r
.40675 .59521 m .34178 .58342 L .327 .55867 L p F P
0 g
s
.379 .635 .955 r
.34178 .58342 m .40675 .59521 L .42088 .60941 L p F P
0 g
s
.554 .714 .947 r
.48888 .59651 m .42088 .60941 L .40675 .59521 L p F P
0 g
s
.72 .815 .924 r
.39332 .59745 m .48052 .58425 L p .47976 .57899 L F P
0 g
s
.989 .733 .365 r
.48417 .59153 m .48135 .58604 L p .48159 .58686 L F P
0 g
s
.989 .733 .365 r
.48411 .59545 m .48417 .59153 L p .48159 .58686 L F P
0 g
s
.963 .688 .465 r
.48888 .59651 m .48417 .59153 L p .48457 .5923 L F P
0 g
s
.963 .688 .465 r
.48883 .60043 m .48888 .59651 L p .48457 .5923 L F P
0 g
s
.905 .63 .509 r
.49527 .60077 m .48888 .59651 L p .48943 .59722 L F P
0 g
s
.549 .73 .959 r
.49527 .60077 m .43952 .62143 L .42088 .60941 L p F P
0 g
s
.905 .63 .509 r
.49523 .60469 m .49527 .60077 L p .48943 .59722 L F P
0 g
s
.847 .577 .529 r
.50302 .60413 m .49527 .60077 L p .49594 .6014 L F P
0 g
s
.549 .747 .969 r
.50302 .60413 m .46179 .63081 L .43952 .62143 L p F P
0 g
s
.847 .577 .529 r
.50299 .60804 m .50302 .60413 L p .49594 .6014 L F P
0 g
s
.794 .53 .539 r
.51178 .60643 m .50302 .60413 L p .50378 .60467 L F P
0 g
s
.555 .766 .976 r
.51178 .60643 m .48669 .6372 L .46179 .63081 L p F P
0 g
s
.794 .53 .539 r
.51176 .61034 m .51178 .60643 L p .50378 .60467 L F P
0 g
s
.742 .487 .544 r
.52115 .60758 m .51178 .60643 L p .51259 .60687 L F P
0 g
s
.586 .803 .982 r
.51319 .64039 m .52115 .60758 L p .53029 .60834 L F P
0 g
s
.742 .487 .544 r
.52115 .61149 m .52115 .60758 L p .51259 .60687 L F P
0 g
s
.586 .803 .982 r
.52115 .60758 m .53073 .60753 L p .53029 .60834 L F P
0 g
s
.586 .803 .982 r
.53073 .60753 m .54018 .64025 L .51319 .64039 L p F P
0 g
s
.608 .818 .98 r
.54018 .64025 m .53073 .60753 L .54007 .60628 L p F P
0 g
s
.608 .818 .98 r
.54007 .60628 m .56658 .63679 L .54018 .64025 L p F P
0 g
s
.783 .587 .635 r
.37831 .8551 m .36929 .90844 L .43085 .92134 L p F P
0 g
s
.783 .506 .517 r
.43156 .9606 m .43085 .92134 L .36929 .90844 L p F P
0 g
s
.735 .467 .523 r
.43085 .92134 m .43156 .9606 L .4957 .96673 L p F P
0 g
s
.9 .844 .387 r
.23455 .33093 m .22417 .36508 L .22107 .40862 L p F P
0 g
s
0 0 .444 r
.23199 .44995 m .22107 .40862 L .22417 .36508 L p F P
0 g
s
.807 .671 .714 r
.35143 .77625 m .32539 .83631 L .37831 .8551 L p F P
0 g
s
.825 .629 .638 r
.36929 .90844 m .37831 .8551 L .32539 .83631 L p F P
0 g
s
.596 .605 .831 r
.42752 .31252 m .38524 .2693 L .36523 .29215 L p F P
0 g
s
.659 .706 .882 r
.31881 .27292 m .36523 .29215 L .38524 .2693 L p F P
0 g
s
.572 .334 .516 r
.62552 .92052 m .62475 .95981 L .68547 .94687 L p F P
0 g
s
.572 .334 .516 r
.68547 .94687 m .68672 .90709 L .62552 .92052 L p F P
0 g
s
.127 .377 .834 r
.78742 .30701 m .81568 .3274 L .79677 .2853 L p F P
0 g
s
.127 .377 .834 r
.79677 .2853 m .76984 .2687 L .78742 .30701 L p F P
0 g
s
.555 .409 .638 r
.67742 .85378 m .68672 .90709 L .74251 .88738 L p F P
0 g
s
.555 .409 .638 r
.74251 .88738 m .72981 .83449 L .67742 .85378 L p F P
0 g
s
.694 .734 .882 r
.68456 .29017 m .73064 .27034 L .70459 .24078 L p F P
0 g
s
.694 .734 .882 r
.70459 .24078 m .6639 .26755 L .68456 .29017 L p F P
0 g
s
.633 .829 .976 r
.56658 .63679 m .54007 .60628 L .54876 .60389 L p F P
0 g
s
.633 .829 .976 r
.54876 .60389 m .5913 .63014 L .56658 .63679 L p F P
0 g
s
.973 .856 .459 r
.83037 .40506 m .8194 .44659 L .81496 .40556 L p F P
0 g
s
.973 .856 .459 r
.81496 .40556 m .82602 .36148 L .83037 .40506 L p F P
0 g
s
.481 .427 .72 r
.49767 .43664 m .46632 .36956 L .45742 .37877 L p F P
0 g
s
.481 .427 .72 r
.45742 .37877 m .4934 .44083 L .49767 .43664 L p F P
0 g
s
.871 .744 .719 r
.55743 .44047 m .59311 .37797 L .58395 .36885 L p F P
0 g
s
.83 .752 .778 r
.6222 .31132 m .58395 .36885 L .59311 .37797 L p F P
0 g
s
.909 .805 .733 r
.59311 .37797 m .55743 .44047 L .56031 .44511 L p F P
0 g
s
.756 .723 .822 r
.43858 .65561 m .3898 .71435 L .42527 .72826 L p F P
0 g
s
.756 .723 .822 r
.42527 .72826 m .46109 .6649 L .43858 .65561 L p F P
0 g
s
.652 .629 .814 r
.62897 .72728 m .592 .66424 L .56701 .67082 L p F P
0 g
s
.626 .618 .822 r
.592 .66424 m .62897 .72728 L .66403 .71302 L p F P
0 g
s
.608 .557 .772 r
.70322 .77461 m .66403 .71302 L .62897 .72728 L p F P
0 g
s
.658 .835 .968 r
.5913 .63014 m .54876 .60389 L .5564 .60045 L p F P
0 g
s
.658 .835 .968 r
.5564 .60045 m .61329 .62053 L .5913 .63014 L p F P
0 g
s
.584 .534 .764 r
.47813 .3618 m .44685 .29926 L .42752 .31252 L p F P
0 g
s
.584 .534 .764 r
.42752 .31252 m .46632 .36956 L .47813 .3618 L p F P
0 g
s
.621 .61 .819 r
.38524 .2693 m .42752 .31252 L .44685 .29926 L p F P
0 g
s
.803 .711 .763 r
.58395 .36885 m .6222 .31132 L .60248 .2983 L p F P
0 g
s
.762 .724 .818 r
.6222 .31132 m .6639 .26755 L .63599 .24843 L p F P
0 g
s
.762 .724 .818 r
.63599 .24843 m .60248 .2983 L .6222 .31132 L p F P
0 g
s
.547 .46 .709 r
.50328 .43313 m .47813 .3618 L .46632 .36956 L p F P
0 g
s
.547 .46 .709 r
.46632 .36956 m .49767 .43664 L .50328 .43313 L p F P
0 g
s
0 .202 .723 r
.23199 .44995 m .26562 .4929 L .27419 .5272 L p F P
0 g
s
.297 .478 .868 r
.327 .55867 m .27419 .5272 L .26562 .4929 L p F P
0 g
s
.17 .454 .881 r
.27419 .5272 m .327 .55867 L .34178 .58342 L p F P
0 g
s
.871 .744 .719 r
.58395 .36885 m .55304 .43632 L .55743 .44047 L p F P
0 g
s
.833 .693 .708 r
.55304 .43632 m .58395 .36885 L .57191 .36124 L p F P
0 g
s
.803 .711 .763 r
.60248 .2983 m .57191 .36124 L .58395 .36885 L p F P
0 g
s
.659 .706 .882 r
.38524 .2693 m .34399 .24307 L .31881 .27292 L p F P
0 g
s
.74 .847 .93 r
.27934 .27177 m .31881 .27292 L .34399 .24307 L p F P
0 g
s
.524 .673 .931 r
.73064 .27034 m .76984 .2687 L .73942 .23335 L p F P
0 g
s
.524 .673 .931 r
.73942 .23335 m .70459 .24078 L .73064 .27034 L p F P
0 g
s
.865 .935 .866 r
.77809 .52445 m .72494 .55653 L .73006 .52995 L p F P
0 g
s
.865 .935 .866 r
.73006 .52995 m .78566 .49 L .77809 .52445 L p F P
0 g
s
.809 .736 .783 r
.35968 .6964 m .31209 .75407 L .35143 .77625 L p F P
0 g
s
.809 .736 .783 r
.35143 .77625 m .3898 .71435 L .35968 .6964 L p F P
0 g
s
.841 .714 .724 r
.32539 .83631 m .35143 .77625 L .31209 .75407 L p F P
0 g
s
.774 .901 .931 r
.71089 .58147 m .64529 .59396 L .65369 .57806 L p F P
0 g
s
.774 .901 .931 r
.65369 .57806 m .72494 .55653 L .71089 .58147 L p F P
0 g
s
.825 .629 .638 r
.32539 .83631 m .31295 .88924 L .36929 .90844 L p F P
0 g
s
.833 .548 .506 r
.37047 .94817 m .36929 .90844 L .31295 .88924 L p F P
0 g
s
.783 .506 .517 r
.36929 .90844 m .37047 .94817 L .43156 .9606 L p F P
0 g
s
.663 .24 .229 r
.49778 .98646 m .4957 .96673 L .43156 .9606 L p F P
0 g
s
.602 .192 .238 r
.4957 .96673 m .49778 .98646 L .55831 .98621 L p F P
0 g
s
.602 .192 .238 r
.55831 .98621 m .56079 .96647 L .4957 .96673 L p F P
0 g
s
.528 .128 .227 r
.56079 .96647 m .55831 .98621 L .61775 .98012 L p F P
0 g
s
.528 .128 .227 r
.61775 .98012 m .62475 .95981 L .56079 .96647 L p F P
0 g
s
.696 .714 .862 r
.51176 .61034 m .48627 .67122 L .51304 .67437 L p F P
0 g
s
.696 .714 .862 r
.51304 .67437 m .52115 .61149 L .51176 .61034 L p F P
0 g
s
.684 .703 .861 r
.52115 .61149 m .51304 .67437 L .54033 .67424 L p F P
0 g
s
.684 .703 .861 r
.54033 .67424 m .53073 .61144 L .52115 .61149 L p F P
0 g
s
.671 .694 .862 r
.53073 .61144 m .54033 .67424 L .56701 .67082 L p F P
0 g
s
.877 .981 .834 r
.27934 .27177 m .25221 .28869 L .23455 .33093 L p F P
0 g
s
0 0 0 r
.22417 .36508 m .23455 .33093 L .25221 .28869 L p F P
0 g
s
.569 .542 .785 r
.66403 .71302 m .70322 .77461 L .74191 .75204 L p F P
0 g
s
.534 .465 .725 r
.70322 .77461 m .72981 .83449 L .77487 .80954 L p F P
0 g
s
.534 .465 .725 r
.77487 .80954 m .74191 .75204 L .70322 .77461 L p F P
0 g
s
.496 .273 .505 r
.68672 .90709 m .68547 .94687 L .74081 .92789 L p F P
0 g
s
.496 .273 .505 r
.74081 .92789 m .74251 .88738 L .68672 .90709 L p F P
0 g
s
.601 .491 .7 r
.50996 .43047 m .4923 .35591 L .47813 .3618 L p F P
0 g
s
.601 .491 .7 r
.47813 .3618 m .50328 .43313 L .50996 .43047 L p F P
0 g
s
.623 .552 .754 r
.44685 .29926 m .47813 .3618 L .4923 .35591 L p F P
0 g
s
.833 .693 .708 r
.57191 .36124 m .54732 .43287 L .55304 .43632 L p F P
0 g
s
.797 .65 .7 r
.54732 .43287 m .57191 .36124 L .55757 .35552 L p F P
0 g
s
.776 .676 .753 r
.57191 .36124 m .60248 .2983 L .57879 .28847 L p F P
0 g
s
.776 .676 .753 r
.57879 .28847 m .55757 .35552 L .57191 .36124 L p F P
0 g
s
.681 .836 .958 r
.61329 .62053 m .5564 .60045 L .56266 .59612 L p F P
0 g
s
.681 .836 .958 r
.56266 .59612 m .63158 .60831 L .61329 .62053 L p F P
0 g
s
.621 .61 .819 r
.44685 .29926 m .41257 .24985 L .38524 .2693 L p F P
0 g
s
.663 .706 .879 r
.34399 .24307 m .38524 .2693 L .41257 .24985 L p F P
0 g
s
.772 .75 .833 r
.41973 .64372 m .35968 .6964 L .3898 .71435 L p F P
0 g
s
.772 .75 .833 r
.3898 .71435 m .43858 .65561 L .41973 .64372 L p F P
0 g
s
.707 .727 .865 r
.50299 .60804 m .46109 .6649 L .48627 .67122 L p F P
0 g
s
.707 .727 .865 r
.48627 .67122 m .51176 .61034 L .50299 .60804 L p F P
0 g
s
.808 .467 0 r
.81568 .3274 m .82602 .36148 L .80666 .31862 L p F P
0 g
s
.808 .467 0 r
.80666 .31862 m .79677 .2853 L .81568 .3274 L p F P
0 g
s
.671 .694 .862 r
.56701 .67082 m .54008 .61019 L .53073 .61144 L p F P
0 g
s
.657 .686 .866 r
.54008 .61019 m .56701 .67082 L .592 .66424 L p F P
0 g
s
.692 .729 .879 r
.6639 .26755 m .70459 .24078 L .66888 .21562 L p F P
0 g
s
.692 .729 .879 r
.66888 .21562 m .63599 .24843 L .6639 .26755 L p F P
0 g
s
.626 .618 .822 r
.66403 .71302 m .61424 .65473 L .592 .66424 L p F P
0 g
s
.599 .61 .834 r
.61424 .65473 m .66403 .71302 L .69363 .69475 L p F P
0 g
s
.569 .542 .785 r
.74191 .75204 m .69363 .69475 L .66403 .71302 L p F P
0 g
s
.623 .552 .754 r
.4923 .35591 m .47025 .28915 L .44685 .29926 L p F P
0 g
s
.645 .619 .809 r
.41257 .24985 m .44685 .29926 L .47025 .28915 L p F P
0 g
s
.841 .714 .724 r
.31209 .75407 m .27961 .8118 L .32539 .83631 L p F P
0 g
s
.869 .677 .64 r
.31295 .88924 m .32539 .83631 L .27961 .8118 L p F P
0 g
s
0 0 .444 r
.22417 .36508 m .23522 .40895 L .23199 .44995 L p F P
0 g
s
.198 .306 .75 r
.26562 .4929 m .23199 .44995 L .23522 .40895 L p F P
0 g
s
.746 .701 .809 r
.60248 .2983 m .63599 .24843 L .60214 .23393 L p F P
0 g
s
.746 .701 .809 r
.60214 .23393 m .57879 .28847 L .60248 .2983 L p F P
0 g
s
.716 .273 .197 r
.43816 .98084 m .43156 .9606 L .37047 .94817 L p F P
0 g
s
.663 .24 .229 r
.43156 .9606 m .43816 .98084 L .49778 .98646 L p F P
0 g
s
.646 .52 .695 r
.5174 .4288 m .50814 .35221 L .4923 .35591 L p F P
0 g
s
.646 .52 .695 r
.4923 .35591 m .50996 .43047 L .5174 .4288 L p F P
0 g
s
.658 .572 .747 r
.47025 .28915 m .4923 .35591 L .50814 .35221 L p F P
0 g
s
.797 .65 .7 r
.55757 .35552 m .54057 .43029 L .54732 .43287 L p F P
0 g
s
.76 .613 .695 r
.54057 .43029 m .55757 .35552 L .54162 .35201 L p F P
0 g
s
.748 .645 .747 r
.55757 .35552 m .57879 .28847 L .5523 .28241 L p F P
0 g
s
.748 .645 .747 r
.5523 .28241 m .54162 .35201 L .55757 .35552 L p F P
0 g
s
.433 .04 .193 r
.62475 .95981 m .61775 .98012 L .67409 .96828 L p F P
0 g
s
.433 .04 .193 r
.67409 .96828 m .68547 .94687 L .62475 .95981 L p F P
0 g
s
.485 .365 .64 r
.72981 .83449 m .74251 .88738 L .79066 .86184 L p F P
0 g
s
.485 .365 .64 r
.79066 .86184 m .77487 .80954 L .72981 .83449 L p F P
0 g
s
.947 .911 .75 r
.8194 .44659 m .78566 .49 L .78106 .45465 L p F P
0 g
s
.947 .911 .75 r
.78106 .45465 m .81496 .40556 L .8194 .44659 L p F P
0 g
s
.716 .741 .871 r
.49523 .60469 m .43858 .65561 L .46109 .6649 L p F P
0 g
s
.716 .741 .871 r
.46109 .6649 m .50299 .60804 L .49523 .60469 L p F P
0 g
s
.686 .549 .692 r
.52523 .42821 m .52486 .35088 L .50814 .35221 L p F P
0 g
s
.686 .549 .692 r
.50814 .35221 m .5174 .4288 L .52523 .42821 L p F P
0 g
s
.69 .594 .744 r
.49656 .28276 m .50814 .35221 L .52486 .35088 L p F P
0 g
s
.658 .572 .747 r
.50814 .35221 m .49656 .28276 L .47025 .28915 L p F P
0 g
s
.76 .613 .695 r
.54162 .35201 m .53308 .42871 L .54057 .43029 L p F P
0 g
s
.724 .579 .692 r
.53308 .42871 m .54162 .35201 L .52486 .35088 L p F P
0 g
s
.72 .618 .744 r
.54162 .35201 m .5523 .28241 L .5244 .28047 L p F P
0 g
s
.72 .618 .744 r
.5244 .28047 m .52486 .35088 L .54162 .35201 L p F P
0 g
s
.657 .686 .866 r
.592 .66424 m .54878 .6078 L .54008 .61019 L p F P
0 g
s
.642 .681 .871 r
.54878 .6078 m .592 .66424 L .61424 .65473 L p F P
0 g
s
.724 .579 .692 r
.52486 .35088 m .52523 .42821 L .53308 .42871 L p F P
0 g
s
.69 .594 .744 r
.52486 .35088 m .5244 .28047 L .49656 .28276 L p F P
0 g
s
.668 .63 .803 r
.44598 .23493 m .47025 .28915 L .49656 .28276 L p F P
0 g
s
.645 .619 .809 r
.47025 .28915 m .44598 .23493 L .41257 .24985 L p F P
0 g
s
.728 .679 .803 r
.57879 .28847 m .60214 .23393 L .56408 .22495 L p F P
0 g
s
.728 .679 .803 r
.56408 .22495 m .5523 .28241 L .57879 .28847 L p F P
0 g
s
.74 .847 .93 r
.34399 .24307 m .3087 .2361 L .27934 .27177 L p F P
0 g
s
.769 .996 .819 r
.25221 .28869 m .27934 .27177 L .3087 .2361 L p F P
0 g
s
.7 .831 .946 r
.63158 .60831 m .56266 .59612 L .56722 .59109 L p F P
0 g
s
.7 .831 .946 r
.56722 .59109 m .64529 .59396 L .63158 .60831 L p F P
0 g
s
0 .337 .815 r
.76984 .2687 m .79677 .2853 L .76346 .24632 L p F P
0 g
s
0 .337 .815 r
.76346 .24632 m .73942 .23335 L .76984 .2687 L p F P
0 g
s
.689 .644 .8 r
.48377 .22547 m .49656 .28276 L .5244 .28047 L p F P
0 g
s
.668 .63 .803 r
.49656 .28276 m .48377 .22547 L .44598 .23493 L p F P
0 g
s
.668 .707 .877 r
.37894 .21749 m .41257 .24985 L .44598 .23493 L p F P
0 g
s
.663 .706 .879 r
.41257 .24985 m .37894 .21749 L .34399 .24307 L p F P
0 g
s
.709 .661 .8 r
.5523 .28241 m .56408 .22495 L .52387 .22207 L p F P
0 g
s
.709 .661 .8 r
.52387 .22207 m .5244 .28047 L .5523 .28241 L p F P
0 g
s
.723 .756 .878 r
.48883 .60043 m .41973 .64372 L .43858 .65561 L p F P
0 g
s
.723 .756 .878 r
.43858 .65561 m .49523 .60469 L .48883 .60043 L p F P
0 g
s
.689 .724 .876 r
.63599 .24843 m .66888 .21562 L .62516 .19646 L p F P
0 g
s
.689 .724 .876 r
.62516 .19646 m .60214 .23393 L .63599 .24843 L p F P
0 g
s
.642 .681 .871 r
.61424 .65473 m .55644 .60437 L .54878 .6078 L p F P
0 g
s
.627 .678 .879 r
.55644 .60437 m .61424 .65473 L .63274 .64264 L p F P
0 g
s
.599 .61 .834 r
.69363 .69475 m .63274 .64264 L .61424 .65473 L p F P
0 g
s
.713 .854 .949 r
.3087 .2361 m .34399 .24307 L .37894 .21749 L p F P
0 g
s
.689 .644 .8 r
.5244 .28047 m .52387 .22207 L .48377 .22547 L p F P
0 g
s
.297 .478 .868 r
.26562 .4929 m .32109 .53218 L .327 .55867 L p F P
0 g
s
.468 .61 .909 r
.39787 .57941 m .327 .55867 L .32109 .53218 L p F P
0 g
s
0 0 0 r
.25221 .28869 m .24224 .32207 L .22417 .36508 L p F P
0 g
s
.109 .094 .557 r
.23522 .40895 m .22417 .36508 L .24224 .32207 L p F P
0 g
s
.784 .779 .848 r
.40543 .62967 m .33633 .67498 L .35968 .6964 L p F P
0 g
s
.784 .779 .848 r
.35968 .6964 m .41973 .64372 L .40543 .62967 L p F P
0 g
s
.832 .778 .8 r
.31209 .75407 m .35968 .6964 L .33633 .67498 L p F P
0 g
s
.869 .677 .64 r
.27961 .8118 m .26405 .86415 L .31295 .88924 L p F P
0 g
s
.886 .593 .485 r
.31459 .92968 m .31295 .88924 L .26405 .86415 L p F P
0 g
s
.833 .548 .506 r
.31295 .88924 m .31459 .92968 L .37047 .94817 L p F P
0 g
s
.52 .699 .951 r
.70459 .24078 m .73942 .23335 L .69718 .20312 L p F P
0 g
s
.52 .699 .951 r
.69718 .20312 m .66888 .21562 L .70459 .24078 L p F P
0 g
s
.759 .29 .133 r
.38147 .96947 m .37047 .94817 L .31459 .92968 L p F P
0 g
s
.716 .273 .197 r
.37047 .94817 m .38147 .96947 L .43816 .98084 L p F P
0 g
s
.792 .871 .906 r
.72494 .55653 m .65369 .57806 L .65625 .56128 L p F P
0 g
s
.792 .871 .906 r
.65625 .56128 m .73006 .52995 L .72494 .55653 L p F P
0 g
s
.832 .778 .8 r
.33633 .67498 m .28107 .72744 L .31209 .75407 L p F P
0 g
s
.875 .766 .738 r
.27961 .8118 m .31209 .75407 L .28107 .72744 L p F P
0 g
s
.569 .605 .849 r
.63274 .64264 m .69363 .69475 L .71634 .67309 L p F P
0 g
s
.524 .528 .802 r
.69363 .69475 m .74191 .75204 L .77213 .72507 L p F P
0 g
s
.524 .528 .802 r
.77213 .72507 m .71634 .67309 L .69363 .69475 L p F P
0 g
s
.3 0 .125 r
.68547 .94687 m .67409 .96828 L .72529 .95097 L p F P
0 g
s
.3 0 .125 r
.72529 .95097 m .74081 .92789 L .68547 .94687 L p F P
0 g
s
.394 .191 .483 r
.74251 .88738 m .74081 .92789 L .78855 .9033 L p F P
0 g
s
.394 .191 .483 r
.78855 .9033 m .79066 .86184 L .74251 .88738 L p F P
0 g
s
.991 .81 .563 r
.82602 .36148 m .81496 .40556 L .79608 .36528 L p F P
0 g
s
.991 .81 .563 r
.79608 .36528 m .80666 .31862 L .82602 .36148 L p F P
0 g
s
.469 .436 .739 r
.74191 .75204 m .77487 .80954 L .81048 .77957 L p F P
0 g
s
.469 .436 .739 r
.81048 .77957 m .77213 .72507 L .74191 .75204 L p F P
0 g
s
.198 .306 .75 r
.23522 .40895 m .26915 .45756 L .26562 .4929 L p F P
0 g
s
.395 .496 .844 r
.32109 .53218 m .26562 .4929 L .26915 .45756 L p F P
0 g
s
.13 .661 .865 r
.50148 .98323 m .49778 .98646 L .43816 .98084 L p F P
0 g
s
.172 .701 .856 r
.49778 .98646 m .50148 .98323 L .5539 .98302 L p F P
0 g
s
.172 .701 .856 r
.5539 .98302 m .55831 .98621 L .49778 .98646 L p F P
0 g
s
.714 .822 .932 r
.64529 .59396 m .56722 .59109 L .56987 .58558 L p F P
0 g
s
.714 .822 .932 r
.56987 .58558 m .65369 .57806 L .64529 .59396 L p F P
0 g
s
.247 .756 .867 r
.55831 .98621 m .5539 .98302 L .60532 .97773 L p F P
0 g
s
.247 .756 .867 r
.60532 .97773 m .61775 .98012 L .55831 .98621 L p F P
0 g
s
.868 .877 .842 r
.78566 .49 m .73006 .52995 L .72559 .5029 L p F P
0 g
s
.868 .877 .842 r
.72559 .5029 m .78106 .45465 L .78566 .49 L p F P
0 g
s
.673 .709 .875 r
.42206 .19778 m .44598 .23493 L .48377 .22547 L p F P
0 g
s
.668 .707 .877 r
.44598 .23493 m .42206 .19778 L .37894 .21749 L p F P
0 g
s
.728 .771 .887 r
.48411 .59545 m .40543 .62967 L .41973 .64372 L p F P
0 g
s
.728 .771 .887 r
.41973 .64372 m .48883 .60043 L .48411 .59545 L p F P
0 g
s
.686 .719 .875 r
.60214 .23393 m .62516 .19646 L .57572 .18455 L p F P
0 g
s
.686 .719 .875 r
.57572 .18455 m .56408 .22495 L .60214 .23393 L p F P
0 g
s
.627 .678 .879 r
.63274 .64264 m .5627 .60004 L .55644 .60437 L p F P
0 g
s
.612 .678 .888 r
.5627 .60004 m .63274 .64264 L .64661 .62843 L p F P
0 g
s
.569 .605 .849 r
.71634 .67309 m .64661 .62843 L .63274 .64264 L p F P
0 g
s
.875 .766 .738 r
.28107 .72744 m .24311 .78221 L .27961 .8118 L p F P
0 g
s
.916 .735 .642 r
.26405 .86415 m .27961 .8118 L .24311 .78221 L p F P
0 g
s
.769 .996 .819 r
.3087 .2361 m .28433 .24936 L .25221 .28869 L p F P
0 g
s
0 0 .159 r
.24224 .32207 m .25221 .28869 L .28433 .24936 L p F P
0 g
s
.122 .638 .89 r
.44991 .97835 m .43816 .98084 L .38147 .96947 L p F P
0 g
s
.13 .661 .865 r
.43816 .98084 m .44991 .97835 L .50148 .98323 L p F P
0 g
s
.391 .308 .642 r
.77487 .80954 m .79066 .86184 L .82893 .83109 L p F P
0 g
s
.391 .308 .642 r
.82893 .83109 m .81048 .77957 L .77487 .80954 L p F P
0 g
s
.348 .821 .893 r
.61775 .98012 m .60532 .97773 L .65391 .96748 L p F P
0 g
s
.348 .821 .893 r
.65391 .96748 m .67409 .96828 L .61775 .98012 L p F P
0 g
s
.69 .444 .546 r
.52115 .60758 m .52115 .61149 L .53073 .61144 L p F P
0 g
s
.69 .444 .546 r
.53073 .61144 m .53073 .60753 L .52115 .60758 L p F P
0 g
s
.633 .398 .544 r
.53073 .60753 m .53073 .61144 L .54008 .61019 L p F P
0 g
s
.633 .398 .544 r
.54008 .61019 m .54007 .60628 L .53073 .60753 L p F P
0 g
s
.678 .712 .874 r
.47112 .18523 m .48377 .22547 L .52387 .22207 L p F P
0 g
s
.673 .709 .875 r
.48377 .22547 m .47112 .18523 L .42206 .19778 L p F P
0 g
s
.926 .539 .176 r
.79677 .2853 m .80666 .31862 L .77231 .27887 L p F P
0 g
s
.926 .539 .176 r
.77231 .27887 m .76346 .24632 L .79677 .2853 L p F P
0 g
s
.682 .715 .874 r
.56408 .22495 m .57572 .18455 L .52335 .18071 L p F P
0 g
s
.682 .715 .874 r
.52335 .18071 m .52387 .22207 L .56408 .22495 L p F P
0 g
s
.742 .487 .544 r
.51178 .60643 m .51176 .61034 L .52115 .61149 L p F P
0 g
s
.567 .346 .539 r
.54007 .60628 m .54008 .61019 L .54878 .6078 L p F P
0 g
s
.567 .346 .539 r
.54878 .6078 m .54876 .60389 L .54007 .60628 L p F P
0 g
s
.678 .851 .965 r
.35 .20537 m .37894 .21749 L .42206 .19778 L p F P
0 g
s
.713 .854 .949 r
.37894 .21749 m .35 .20537 L .3087 .2361 L p F P
0 g
s
.678 .712 .874 r
.52387 .22207 m .52335 .18071 L .47112 .18523 L p F P
0 g
s
.794 .53 .539 r
.50302 .60413 m .50299 .60804 L .51176 .61034 L p F P
0 g
s
.484 .281 .528 r
.54876 .60389 m .54878 .6078 L .55644 .60437 L p F P
0 g
s
.484 .281 .528 r
.55644 .60437 m .5564 .60045 L .54876 .60389 L p F P
0 g
s
.109 .094 .557 r
.24224 .32207 m .25286 .36852 L .23522 .40895 L p F P
0 g
s
.352 .374 .743 r
.26915 .45756 m .23522 .40895 L .25286 .36852 L p F P
0 g
s
.525 .728 .967 r
.66888 .21562 m .69718 .20312 L .64504 .17998 L p F P
0 g
s
.525 .728 .967 r
.64504 .17998 m .62516 .19646 L .66888 .21562 L p F P
0 g
s
.468 .61 .909 r
.32109 .53218 m .3948 .56268 L .39787 .57941 L p F P
0 g
s
.553 .917 .741 r
.28433 .24936 m .3087 .2361 L .35 .20537 L p F P
0 g
s
.847 .577 .529 r
.49527 .60077 m .49523 .60469 L .50299 .60804 L p F P
0 g
s
.793 .811 .867 r
.39644 .61402 m .32103 .65089 L .33633 .67498 L p F P
0 g
s
.793 .811 .867 r
.33633 .67498 m .40543 .62967 L .39644 .61402 L p F P
0 g
s
.851 .825 .822 r
.28107 .72744 m .33633 .67498 L .32103 .65089 L p F P
0 g
s
.723 .81 .919 r
.65369 .57806 m .56987 .58558 L .57046 .57984 L p F P
0 g
s
.723 .81 .919 r
.57046 .57984 m .65625 .56128 L .65369 .57806 L p F P
0 g
s
.886 .593 .485 r
.26405 .86415 m .2661 .90553 L .31459 .92968 L p F P
0 g
s
.781 .284 .018 r
.32975 .9526 m .31459 .92968 L .2661 .90553 L p F P
0 g
s
.759 .29 .133 r
.31459 .92968 m .32975 .9526 L .38147 .96947 L p F P
0 g
s
.921 .834 .742 r
.81496 .40556 m .78106 .45465 L .76379 .42012 L p F P
0 g
s
.921 .834 .742 r
.76379 .42012 m .79608 .36528 L .81496 .40556 L p F P
0 g
s
.369 .192 .506 r
.5564 .60045 m .55644 .60437 L .5627 .60004 L p F P
0 g
s
.369 .192 .506 r
.5627 .60004 m .56266 .59612 L .5564 .60045 L p F P
0 g
s
0 .288 .73 r
.73942 .23335 m .76346 .24632 L .71679 .21284 L p F P
0 g
s
0 .288 .73 r
.71679 .21284 m .69718 .20312 L .73942 .23335 L p F P
0 g
s
.729 .787 .899 r
.48129 .58998 m .39644 .61402 L .40543 .62967 L p F P
0 g
s
.729 .787 .899 r
.40543 .62967 m .48411 .59545 L .48129 .58998 L p F P
0 g
s
.537 .605 .868 r
.64661 .62843 m .71634 .67309 L .7309 .64881 L p F P
0 g
s
.469 .517 .824 r
.71634 .67309 m .77213 .72507 L .79215 .69459 L p F P
0 g
s
.469 .517 .824 r
.79215 .69459 m .7309 .64881 L .71634 .67309 L p F P
0 g
s
.395 .496 .844 r
.26915 .45756 m .32473 .50511 L .32109 .53218 L p F P
0 g
s
.508 .605 .885 r
.3948 .56268 m .32109 .53218 L .32473 .50511 L p F P
0 g
s
.612 .678 .888 r
.64661 .62843 m .56727 .59501 L .5627 .60004 L p F P
0 g
s
.597 .681 .9 r
.56727 .59501 m .64661 .62843 L .65512 .61268 L p F P
0 g
s
.537 .605 .868 r
.7309 .64881 m .65512 .61268 L .64661 .62843 L p F P
0 g
s
.905 .63 .509 r
.48888 .59651 m .48883 .60043 L .49523 .60469 L p F P
0 g
s
.102 0 .005 r
.74081 .92789 m .72529 .95097 L .76926 .92858 L p F P
0 g
s
.102 0 .005 r
.76926 .92858 m .78855 .9033 L .74081 .92789 L p F P
0 g
s
.796 .837 .883 r
.73006 .52995 m .65625 .56128 L .65267 .5444 L p F P
0 g
s
.796 .837 .883 r
.65267 .5444 m .72559 .5029 L .73006 .52995 L p F P
0 g
s
.19 .058 .46 r
.56266 .59612 m .5627 .60004 L .56727 .59501 L p F P
0 g
s
.19 .058 .46 r
.56727 .59501 m .56722 .59109 L .56266 .59612 L p F P
0 g
s
.147 .628 .922 r
.401 .96851 m .38147 .96947 L .32975 .9526 L p F P
0 g
s
.122 .638 .89 r
.38147 .96947 m .401 .96851 L .44991 .97835 L p F P
0 g
s
.963 .688 .465 r
.48417 .59153 m .48411 .59545 L .48883 .60043 L p F P
0 g
s
.851 .825 .822 r
.32103 .65089 m .26012 .69721 L .28107 .72744 L p F P
0 g
s
.908 .828 .755 r
.24311 .78221 m .28107 .72744 L .26012 .69721 L p F P
0 g
s
.916 .735 .642 r
.24311 .78221 m .22484 .8338 L .26405 .86415 L p F P
0 g
s
.942 .643 .445 r
.2661 .90553 m .26405 .86415 L .22484 .8338 L p F P
0 g
s
.461 .881 .924 r
.67409 .96828 m .65391 .96748 L .69784 .95254 L p F P
0 g
s
.461 .881 .924 r
.69784 .95254 m .72529 .95097 L .67409 .96828 L p F P
0 g
s
.383 .403 .757 r
.77213 .72507 m .81048 .77957 L .83457 .74551 L p F P
0 g
s
.383 .403 .757 r
.83457 .74551 m .79215 .69459 L .77213 .72507 L p F P
0 g
s
.24 .071 .439 r
.79066 .86184 m .78855 .9033 L .82646 .87372 L p F P
0 g
s
.24 .071 .439 r
.82646 .87372 m .82893 .83109 L .79066 .86184 L p F P
0 g
s
0 0 .351 r
.56722 .59109 m .56727 .59501 L .56992 .58951 L p F P
0 g
s
0 0 .351 r
.56992 .58951 m .56987 .58558 L .56722 .59109 L p F P
0 g
s
.989 .733 .365 r
.48135 .58604 m .48129 .58998 L .48411 .59545 L p F P
0 g
s
.639 .839 .977 r
.40139 .18158 m .42206 .19778 L .47112 .18523 L p F P
0 g
s
.678 .851 .965 r
.42206 .19778 m .40139 .18158 L .35 .20537 L p F P
0 g
s
.597 .681 .9 r
.65512 .61268 m .56992 .58951 L .56727 .59501 L p F P
0 g
s
.54 .759 .978 r
.62516 .19646 m .64504 .17998 L .58579 .16556 L p F P
0 g
s
.54 .759 .978 r
.58579 .16556 m .57572 .18455 L .62516 .19646 L p F P
0 g
s
0 0 .159 r
.28433 .24936 m .27534 .28197 L .24224 .32207 L p F P
0 g
s
.321 .231 .595 r
.25286 .36852 m .24224 .32207 L .27534 .28197 L p F P
0 g
s
.508 .605 .885 r
.32473 .50511 m .39786 .54578 L .3948 .56268 L p F P
0 g
s
.352 .374 .743 r
.25286 .36852 m .28535 .42289 L .26915 .45756 L p F P
0 g
s
.469 .512 .819 r
.32473 .50511 m .26915 .45756 L .28535 .42289 L p F P
0 g
s
0 0 .113 r
.56987 .58558 m .56992 .58951 L .57051 .58378 L p F P
0 g
s
0 0 .113 r
.57051 .58378 m .57046 .57984 L .56987 .58558 L p F P
0 g
s
.655 .764 .929 r
.5064 .95523 m .50148 .98323 L .44991 .97835 L p F P
0 g
s
.649 .758 .929 r
.50148 .98323 m .5064 .95523 L .54807 .95506 L p F P
0 g
s
.649 .758 .929 r
.54807 .95506 m .5539 .98302 L .50148 .98323 L p F P
0 g
s
.728 .796 .906 r
.65625 .56128 m .57046 .57984 L .56894 .57414 L p F P
0 g
s
.728 .796 .906 r
.56894 .57414 m .65267 .5444 L .65625 .56128 L p F P
0 g
s
.643 .754 .929 r
.5539 .98302 m .54807 .95506 L .58891 .95076 L p F P
0 g
s
.643 .754 .929 r
.58891 .95076 m .60532 .97773 L .5539 .98302 L p F P
0 g
s
.727 .802 .911 r
.48052 .58425 m .39332 .59745 L .39644 .61402 L p F P
0 g
s
.727 .802 .911 r
.39644 .61402 m .48129 .58998 L .48052 .58425 L p F P
0 g
s
.794 .844 .889 r
.32103 .65089 m .39644 .61402 L .39332 .59745 L p F P
0 g
s
.941 .735 .597 r
.80666 .31862 m .79608 .36528 L .76285 .32797 L p F P
0 g
s
.941 .735 .597 r
.76285 .32797 m .77231 .27887 L .80666 .31862 L p F P
0 g
s
.583 .686 .912 r
.56992 .58951 m .65512 .61268 L .65772 .59607 L p F P
0 g
s
.583 .686 .912 r
.65772 .59607 m .57051 .58378 L .56992 .58951 L p F P
0 g
s
.502 .608 .891 r
.65512 .61268 m .7309 .64881 L .73627 .62291 L p F P
0 g
s
.502 .608 .891 r
.73627 .62291 m .65772 .59607 L .65512 .61268 L p F P
0 g
s
.853 .822 .818 r
.78106 .45465 m .72559 .5029 L .71129 .4767 L p F P
0 g
s
.853 .822 .818 r
.71129 .4767 m .76379 .42012 L .78106 .45465 L p F P
0 g
s
.397 .245 .556 r
.52553 .58349 m .53376 .66014 L .53238 .66106 L closepath p F P
0 g
s
.239 .133 .526 r
.52553 .58349 m .53475 .65906 L .53376 .66014 L closepath p F P
0 g
s
0 0 .453 r
.52553 .58349 m .53531 .6579 L .53475 .65906 L closepath p F P
0 g
s
.939 .617 .397 r
.52553 .58349 m .53303 .65337 L .53425 .65437 L closepath p F P
0 g
s
.502 .322 .568 r
.52553 .58349 m .53238 .66106 L .53069 .6618 L closepath p F P
0 g
s
.881 .567 .452 r
.52553 .58349 m .53146 .65254 L .53303 .65337 L closepath p F P
0 g
s
.579 .379 .574 r
.52553 .58349 m .53069 .6618 L .52876 .66232 L closepath p F P
0 g
s
.822 .52 .479 r
.52553 .58349 m .52961 .65192 L .53146 .65254 L closepath p F P
0 g
s
.641 .427 .577 r
.52553 .58349 m .52876 .66232 L .52667 .66258 L closepath p F P
0 g
s
.765 .474 .492 r
.52553 .58349 m .52756 .65154 L .52961 .65192 L closepath p F P
0 g
s
.695 .47 .578 r
.52553 .58349 m .52667 .66258 L .52454 .6626 L closepath p F P
0 g
s
.709 .429 .498 r
.52553 .58349 m .52542 .65142 L .52756 .65154 L closepath p F P
0 g
s
.745 .512 .577 r
.52553 .58349 m .52454 .6626 L .52244 .66235 L closepath p F P
0 g
s
.65 .38 .498 r
.52553 .58349 m .52328 .65156 L .52542 .65142 L closepath p F P
0 g
s
.795 .555 .575 r
.52553 .58349 m .52244 .66235 L .5205 .66185 L closepath p F P
0 g
s
.582 .325 .491 r
.52553 .58349 m .52125 .65196 L .52328 .65156 L closepath p F P
0 g
s
.903 .657 .557 r
.52553 .58349 m .51878 .66113 L .51737 .66022 L closepath p F P
0 g
s
.847 .602 .569 r
.52553 .58349 m .5205 .66185 L .51878 .66113 L closepath p F P
0 g
s
.498 .255 .477 r
.52553 .58349 m .51942 .6526 L .52125 .65196 L closepath p F P
0 g
s
.96 .721 .53 r
.52553 .58349 m .51737 .66022 L .51635 .65916 L closepath p F P
0 g
s
.382 .16 .448 r
.52553 .58349 m .51788 .65344 L .51942 .6526 L closepath p F P
0 g
s
.998 .783 .463 r
.52553 .58349 m .51635 .65916 L .51575 .658 L closepath p F P
0 g
s
.205 .017 .39 r
.52553 .58349 m .5167 .65446 L .51788 .65344 L closepath p F P
0 g
s
.915 .774 .29 r
.51561 .65679 m .52553 .58349 L .51596 .65637 L p F P
0 g
s
.915 .774 .29 r
.51596 .65637 m .51575 .658 L .51561 .65679 L p F P
0 g
s
.8 .848 .391 r
.48733 .72225 m .51521 .65725 L p .51561 .65679 L F P
0 g
s
0 0 .016 r
.51593 .65559 m .52553 .58349 L .51582 .65518 L p F P
0 g
s
0 0 .016 r
.51582 .65518 m .51561 .65679 L .51593 .65559 L p F P
0 g
s
.8 .848 .391 r
.51521 .65725 m .51593 .65559 L .51561 .65679 L p F P
0 g
s
.979 .903 .569 r
.48614 .72681 m .51561 .65679 L .51575 .658 L p F P
0 g
s
.979 .903 .569 r
.51575 .658 m .4868 .73138 L .48614 .72681 L p F P
0 g
s
.8 .848 .391 r
.51561 .65679 m .48826 .72176 L p .48733 .72225 L F P
0 g
s
.8 .848 .391 r
.48826 .72176 m .48614 .72681 L .48733 .72225 L p F P
0 g
s
.651 .881 .524 r
.44816 .7683 m .48622 .72282 L p .48733 .72225 L F P
0 g
s
0 0 0 r
.51593 .65559 m .48939 .71743 L p .49032 .71792 L F P
0 g
s
0 0 0 r
.48939 .71743 m .48733 .72225 L .49032 .71792 L p F P
0 g
s
.651 .881 .524 r
.48622 .72282 m .49032 .71792 L .48733 .72225 L p F P
0 g
s
.89 .979 .687 r
.44187 .77747 m .48733 .72225 L .48614 .72681 L p F P
0 g
s
.89 .979 .687 r
.48614 .72681 m .43958 .78706 L .44187 .77747 L p F P
0 g
s
.651 .881 .524 r
.48733 .72225 m .44731 .77086 L p .44816 .7683 L F P
0 g
s
.651 .881 .524 r
.44731 .77086 m .44187 .77747 L .44816 .7683 L p F P
0 g
s
.565 .92 .733 r
.41097 .78698 m .4492 .76515 L p .44816 .7683 L F P
0 g
s
0 0 0 r
.49032 .71792 m .45321 .76226 L p .45821 .76001 L F P
0 g
s
0 0 0 r
.45321 .76226 m .44816 .7683 L .45821 .76001 L p F P
0 g
s
.565 .92 .733 r
.4492 .76515 m .45821 .76001 L .44816 .7683 L p F P
0 g
s
.953 .935 .735 r
.43958 .78706 m .48614 .72681 L .4868 .73138 L p F P
0 g
s
0 0 0 r
.43343 .77528 m .46433 .75726 L p .45821 .76001 L F P
0 g
s
.953 .935 .735 r
.4868 .73138 m .44131 .79662 L .43958 .78706 L p F P
0 g
s
0 0 0 r
.46433 .75726 m .47158 .75304 L .45821 .76001 L p F P
0 g
s
0 0 0 r
.45821 .76001 m .49142 .71849 L p .49032 .71792 L F P
0 g
s
.981 .838 .628 r
.4868 .73138 m .51575 .658 L .51635 .65916 L p F P
0 g
s
.981 .838 .628 r
.51635 .65916 m .48925 .73577 L .4868 .73138 L p F P
0 g
s
0 0 0 r
.49142 .71849 m .49498 .71403 L .49032 .71792 L p F P
0 g
s
0 0 0 r
.49032 .71792 m .51604 .65604 L p .51593 .65559 L F P
0 g
s
0 0 0 r
.51604 .65604 m .5167 .65446 L .51593 .65559 L p F P
0 g
s
0 0 .266 r
.52553 .58349 m .51593 .65559 L .5167 .65446 L closepath p F P
0 g
s
0 0 .267 r
.53509 .65627 m .52553 .58349 L .53541 .65669 L p F P
0 g
s
.845 .585 .043 r
.53505 .65549 m .53541 .65669 L .5352 .65508 L p F P
0 g
s
.845 .585 .043 r
.5352 .65508 m .52553 .58349 L .53505 .65549 L p F P
0 g
s
.829 .42 0 r
.53505 .65549 m .53425 .65437 L p .53491 .6559 L F P
0 g
s
0 0 .367 r
.53541 .65669 m .53505 .65549 L p .53573 .65711 L F P
0 g
s
0 0 .367 r
.56482 .72643 m .53541 .65669 L p .53573 .65711 L F P
0 g
s
0 0 .267 r
.53541 .65669 m .53531 .6579 L .53509 .65627 L p F P
0 g
s
0 .008 .56 r
.56432 .73101 m .53531 .6579 L .53541 .65669 L p F P
0 g
s
0 .008 .56 r
.53541 .65669 m .56482 .72643 L .56432 .73101 L p F P
0 g
s
0 0 .367 r
.56348 .72188 m .56482 .72643 L p .56271 .72141 L F P
0 g
s
0 0 .367 r
.53505 .65549 m .56348 .72188 L p .56271 .72141 L F P
0 g
s
.829 .42 0 r
.56348 .72188 m .53505 .65549 L p .53491 .6559 L F P
0 g
s
.829 .42 0 r
.56034 .71758 m .56348 .72188 L p .56141 .71709 L F P
0 g
s
.829 .42 0 r
.53425 .65437 m .56034 .71758 L p .56141 .71709 L F P
0 g
s
.164 .176 .625 r
.53531 .6579 m .56432 .73101 L .56202 .73542 L p F P
0 g
s
.697 .179 0 r
.56034 .71758 m .55554 .71374 L p .55928 .71806 L F P
0 g
s
0 0 .504 r
.56348 .72188 m .56034 .71758 L p .56423 .72234 L F P
0 g
s
0 0 .504 r
.6087 .77669 m .56348 .72188 L p .56423 .72234 L F P
0 g
s
0 .13 .678 r
.61134 .78626 m .56482 .72643 L .56348 .72188 L p F P
0 g
s
0 .13 .678 r
.56348 .72188 m .6087 .77669 L .61134 .78626 L p F P
0 g
s
0 0 .504 r
.60208 .76758 m .6087 .77669 L p .60323 .77 L F P
0 g
s
0 0 .504 r
.56034 .71758 m .60208 .76758 L p .60323 .77 L F P
0 g
s
.697 .179 0 r
.60208 .76758 m .56034 .71758 L p .55928 .71806 L F P
0 g
s
.697 .179 0 r
.59173 .75938 m .60208 .76758 L p .5968 .76147 L F P
0 g
s
.697 .179 0 r
.55554 .71374 m .59173 .75938 L p .5968 .76147 L F P
0 g
s
.296 .338 .736 r
.56432 .73101 m .60995 .79585 L .60467 .80499 L p F P
0 g
s
.114 .257 .734 r
.60995 .79585 m .56432 .73101 L .56482 .72643 L p F P
0 g
s
.114 .257 .734 r
.56482 .72643 m .61134 .78626 L .60995 .79585 L p F P
0 g
s
.39 0 0 r
.59173 .75938 m .57809 .75253 L p .58679 .75735 L F P
0 g
s
0 .269 .721 r
.60208 .76758 m .59173 .75938 L p .60096 .76521 L F P
0 g
s
0 .269 .721 r
.65539 .79957 m .60208 .76758 L p .60096 .76521 L F P
0 g
s
0 .269 .721 r
.63826 .78592 m .65539 .79957 L p .64813 .79499 L F P
0 g
s
0 .269 .721 r
.59173 .75938 m .63826 .78592 L p .64813 .79499 L F P
0 g
s
.39 0 0 r
.63826 .78592 m .59173 .75938 L p .58679 .75735 L F P
0 g
s
.39 0 0 r
.61534 .77443 m .63826 .78592 L p .63137 .7821 L F P
0 g
s
.39 0 0 r
.57809 .75253 m .61534 .77443 L p .63137 .7821 L F P
0 g
s
0 .348 .822 r
.66602 .81463 m .6087 .77669 L .60208 .76758 L p F P
0 g
s
0 .348 .822 r
.60208 .76758 m .65539 .79957 L .66602 .81463 L p F P
0 g
s
.502 .434 .713 r
.55803 .73944 m .59581 .8133 L .58383 .82041 L p F P
0 g
s
.417 .393 .725 r
.59581 .8133 m .55803 .73944 L .56202 .73542 L p F P
0 g
s
.417 .393 .725 r
.56202 .73542 m .60467 .80499 L .59581 .8133 L p F P
0 g
s
.296 .338 .736 r
.60467 .80499 m .56202 .73542 L .56432 .73101 L p F P
0 g
s
.454 0 0 r
.58769 .76574 m .61534 .77443 L p .60919 .77132 L F P
0 g
s
.454 0 0 r
.56182 .74738 m .58769 .76574 L p .60919 .77132 L F P
0 g
s
.567 .469 .703 r
.55256 .74291 m .58383 .82041 L .56929 .82603 L p F P
0 g
s
.502 .434 .713 r
.58383 .82041 m .55256 .74291 L .55803 .73944 L p F P
0 g
s
.602 .444 .64 r
.53069 .6618 m .54586 .74566 L .53824 .74758 L p F P
0 g
s
.539 .401 .642 r
.54586 .74566 m .53069 .6618 L .53238 .66106 L p F P
0 g
s
.539 .401 .642 r
.53238 .66106 m .55256 .74291 L .54586 .74566 L p F P
0 g
s
.619 .501 .697 r
.55286 .82993 m .53824 .74758 L .54586 .74566 L p F P
0 g
s
.619 .501 .697 r
.54586 .74566 m .56929 .82603 L .55286 .82993 L p F P
0 g
s
.567 .469 .703 r
.56929 .82603 m .54586 .74566 L .55256 .74291 L p F P
0 g
s
.45 0 0 r
.55672 .76037 m .58769 .76574 L p .58262 .76325 L F P
0 g
s
.45 0 0 r
.5437 .74421 m .55672 .76037 L p .58262 .76325 L F P
0 g
s
.705 .563 .691 r
.51723 .83204 m .52165 .74862 L .53005 .74858 L p F P
0 g
s
.705 .563 .691 r
.53005 .74858 m .53526 .83196 L .51723 .83204 L p F P
0 g
s
.664 .532 .693 r
.53526 .83196 m .53005 .74858 L .53824 .74758 L p F P
0 g
s
.664 .532 .693 r
.53824 .74758 m .55286 .82993 L .53526 .83196 L p F P
0 g
s
.702 .519 .638 r
.52667 .66258 m .53005 .74858 L .52165 .74862 L p F P
0 g
s
.655 .482 .638 r
.53005 .74858 m .52667 .66258 L .52876 .66232 L p F P
0 g
s
.655 .482 .638 r
.52876 .66232 m .53824 .74758 L .53005 .74858 L p F P
0 g
s
.397 0 0 r
.52466 .7432 m .52406 .75864 L .55672 .76037 L p F P
0 g
s
.185 .701 .888 r
.55672 .76037 m .52406 .75864 L p .53265 .75768 L F P
0 g
s
.45 0 0 r
.58769 .76574 m .56182 .74738 L .5437 .74421 L p F P
0 g
s
.602 .444 .64 r
.53824 .74758 m .52876 .66232 L .53069 .6618 L p F P
0 g
s
.397 0 0 r
.55672 .76037 m .5437 .74421 L .52466 .7432 L p F P
0 g
s
.185 .701 .888 r
.57134 .75336 m .55672 .76037 L p .53265 .75768 L F P
0 g
s
.156 .663 .904 r
.58769 .76574 m .55672 .76037 L p .56765 .76053 L F P
0 g
s
.454 0 0 r
.61534 .77443 m .57809 .75253 L .56182 .74738 L p F P
0 g
s
.156 .663 .904 r
.61665 .76127 m .58769 .76574 L p .56765 .76053 L F P
0 g
s
.164 .64 .928 r
.61534 .77443 m .58769 .76574 L p .60035 .76725 L F P
0 g
s
.164 .64 .928 r
.6568 .774 m .61534 .77443 L p .60035 .76725 L F P
0 g
s
.203 .626 .95 r
.63826 .78592 m .61534 .77443 L p .62902 .77743 L F P
0 g
s
.203 .626 .95 r
.68971 .79075 m .63826 .78592 L p .62902 .77743 L F P
0 g
s
.968 .649 .282 r
.52553 .58349 m .53425 .65437 L .53505 .65549 L closepath p F P
0 g
s
.917 .694 .198 r
.57046 .57984 m .57051 .58378 L .56899 .57808 L p F P
0 g
s
.917 .694 .198 r
.56899 .57808 m .56894 .57414 L .57046 .57984 L p F P
0 g
s
.6 .818 .983 r
.46016 .16639 m .47112 .18523 L .52335 .18071 L p F P
0 g
s
.639 .839 .977 r
.47112 .18523 m .46016 .16639 L .40139 .18158 L p F P
0 g
s
.908 .828 .755 r
.26012 .69721 m .21796 .74845 L .24311 .78221 L p F P
0 g
s
.963 .805 .638 r
.22484 .8338 m .24311 .78221 L .21796 .74845 L p F P
0 g
s
.66 .77 .931 r
.46543 .95127 m .44991 .97835 L .401 .96851 L p F P
0 g
s
.655 .764 .929 r
.44991 .97835 m .46543 .95127 L .5064 .95523 L p F P
0 g
s
.565 .79 .984 r
.57572 .18455 m .58579 .16556 L .5229 .16091 L p F P
0 g
s
.565 .79 .984 r
.5229 .16091 m .52335 .18071 L .57572 .18455 L p F P
0 g
s
.571 .695 .925 r
.57051 .58378 m .65772 .59607 L .65411 .57935 L p F P
0 g
s
.571 .695 .925 r
.65411 .57935 m .56899 .57808 L .57051 .58378 L p F P
0 g
s
.635 .75 .931 r
.60532 .97773 m .58891 .95076 L .62734 .94246 L p F P
0 g
s
.635 .75 .931 r
.62734 .94246 m .65391 .96748 L .60532 .97773 L p F P
0 g
s
.553 .917 .741 r
.35 .20537 m .32993 .21534 L .28433 .24936 L p F P
0 g
s
0 0 0 r
.32993 .21534 m .35 .20537 L .40139 .18158 L p F P
0 g
s
.226 0 .301 r
.27534 .28197 m .28433 .24936 L .32993 .21534 L p F P
0 g
s
.794 .844 .889 r
.39332 .59745 m .31484 .62509 L .32103 .65089 L p F P
0 g
s
.861 .877 .848 r
.26012 .69721 m .32103 .65089 L .31484 .62509 L p F P
0 g
s
.336 0 0 r
.69718 .20312 m .71679 .21284 L .65886 .18715 L p F P
0 g
s
.336 0 0 r
.65886 .18715 m .64504 .17998 L .69718 .20312 L p F P
0 g
s
.253 .228 .637 r
.81048 .77957 m .82893 .83109 L .85507 .79604 L p F P
0 g
s
.253 .228 .637 r
.85507 .79604 m .83457 .74551 L .81048 .77957 L p F P
0 g
s
.6 .818 .983 r
.52335 .18071 m .5229 .16091 L .46016 .16639 L p F P
0 g
s
.896 .518 .31 r
.76346 .24632 m .77231 .27887 L .72403 .24469 L p F P
0 g
s
.896 .518 .31 r
.72403 .24469 m .71679 .21284 L .76346 .24632 L p F P
0 g
s
.402 .507 .851 r
.7309 .64881 m .79215 .69459 L .80046 .66177 L p F P
0 g
s
.402 .507 .851 r
.80046 .66177 m .73627 .62291 L .7309 .64881 L p F P
0 g
s
.729 .781 .894 r
.65267 .5444 m .56894 .57414 L .56537 .56875 L p F P
0 g
s
.986 .71 .388 r
.56894 .57414 m .56899 .57808 L .56542 .5727 L p F P
0 g
s
.986 .71 .388 r
.56542 .5727 m .56537 .56875 L .56894 .57414 L p F P
0 g
s
.469 .512 .819 r
.28535 .42289 m .3382 .47879 L .32473 .50511 L p F P
0 g
s
.545 .604 .864 r
.39786 .54578 m .32473 .50511 L .3382 .47879 L p F P
0 g
s
.781 .284 .018 r
.2661 .90553 m .28507 .9306 L .32975 .9526 L p F P
0 g
s
.198 .625 .948 r
.35659 .95394 m .32975 .9526 L .28507 .9306 L p F P
0 g
s
.147 .628 .922 r
.32975 .9526 m .35659 .95394 L .401 .96851 L p F P
0 g
s
.72 .815 .924 r
.48186 .57854 m .39639 .58071 L .39332 .59745 L p F P
0 g
s
.561 .707 .939 r
.56899 .57808 m .65411 .57935 L .64428 .56333 L p F P
0 g
s
.561 .707 .939 r
.64428 .56333 m .56542 .5727 L .56899 .57808 L p F P
0 g
s
.791 .803 .862 r
.72559 .5029 m .65267 .5444 L .64294 .52824 L p F P
0 g
s
.791 .803 .862 r
.64294 .52824 m .71129 .4767 L .72559 .5029 L p F P
0 g
s
.571 .925 .95 r
.72529 .95097 m .69784 .95254 L .73528 .93329 L p F P
0 g
s
.571 .925 .95 r
.73528 .93329 m .76926 .92858 L .72529 .95097 L p F P
0 g
s
.942 .643 .445 r
.22484 .8338 m .22726 .87633 L .2661 .90553 L p F P
0 g
s
.74 .228 0 r
.28507 .9306 m .2661 .90553 L .22726 .87633 L p F P
0 g
s
.545 .604 .864 r
.3382 .47879 m .40707 .52952 L .39786 .54578 L p F P
0 g
s
.729 .781 .894 r
.56537 .56875 m .64294 .52824 L .65267 .5444 L p F P
0 g
s
.726 .765 .884 r
.64294 .52824 m .56537 .56875 L .55989 .56393 L p F P
0 g
s
.943 .658 .47 r
.56537 .56875 m .56542 .5727 L .55994 .56789 L p F P
0 g
s
.943 .658 .47 r
.55994 .56789 m .55989 .56393 L .56537 .56875 L p F P
0 g
s
0 0 0 r
.78855 .9033 m .76926 .92858 L .80394 .90172 L p F P
0 g
s
0 0 0 r
.80394 .90172 m .82646 .87372 L .78855 .9033 L p F P
0 g
s
.786 .875 .913 r
.31484 .62509 m .39332 .59745 L .39639 .58071 L p F P
0 g
s
.321 .231 .595 r
.27534 .28197 m .28491 .33088 L .25286 .36852 L p F P
0 g
s
.457 .421 .73 r
.28535 .42289 m .25286 .36852 L .28491 .33088 L p F P
0 g
s
.708 .825 .937 r
.48526 .57313 m .4057 .5646 L .39639 .58071 L p F P
0 g
s
.554 .721 .951 r
.56542 .5727 m .64428 .56333 L .62855 .54883 L p F P
0 g
s
.554 .721 .951 r
.62855 .54883 m .55994 .56789 L .56542 .5727 L p F P
0 g
s
.465 .617 .915 r
.65772 .59607 m .73627 .62291 L .73176 .59653 L p F P
0 g
s
.465 .617 .915 r
.73176 .59653 m .65411 .57935 L .65772 .59607 L p F P
0 g
s
.663 .777 .933 r
.42673 .94329 m .401 .96851 L .35659 .95394 L p F P
0 g
s
.66 .77 .931 r
.401 .96851 m .42673 .94329 L .46543 .95127 L p F P
0 g
s
.882 .766 .729 r
.79608 .36528 m .76379 .42012 L .73402 .38829 L p F P
0 g
s
.882 .766 .729 r
.73402 .38829 m .76285 .32797 L .79608 .36528 L p F P
0 g
s
.627 .748 .934 r
.65391 .96748 m .62734 .94246 L .66186 .93041 L p F P
0 g
s
.627 .748 .934 r
.66186 .93041 m .69784 .95254 L .65391 .96748 L p F P
0 g
s
.726 .765 .884 r
.55989 .56393 m .62737 .51361 L .64294 .52824 L p F P
0 g
s
.72 .749 .875 r
.62737 .51361 m .55989 .56393 L .55276 .55992 L p F P
0 g
s
.882 .601 .506 r
.55989 .56393 m .55994 .56789 L .5528 .56388 L p F P
0 g
s
.882 .601 .506 r
.5528 .56388 m .55276 .55992 L .55989 .56393 L p F P
0 g
s
.692 .832 .95 r
.49059 .56826 m .42095 .54995 L .4057 .5646 L p F P
0 g
s
.551 .738 .962 r
.55994 .56789 m .62855 .54883 L .60756 .53664 L p F P
0 g
s
.551 .738 .962 r
.60756 .53664 m .5528 .56388 L .55994 .56789 L p F P
0 g
s
.861 .877 .848 r
.31484 .62509 m .25077 .66453 L .26012 .69721 L p F P
0 g
s
.931 .899 .774 r
.21796 .74845 m .26012 .69721 L .25077 .66453 L p F P
0 g
s
.72 .749 .875 r
.55276 .55992 m .60662 .50131 L .62737 .51361 L p F P
0 g
s
.712 .734 .869 r
.60662 .50131 m .55276 .55992 L .54431 .55692 L p F P
0 g
s
.824 .55 .524 r
.55276 .55992 m .5528 .56388 L .54433 .56088 L p F P
0 g
s
.824 .55 .524 r
.54433 .56088 m .54431 .55692 L .55276 .55992 L p F P
0 g
s
.264 .36 .775 r
.79215 .69459 m .83457 .74551 L .84527 .70858 L p F P
0 g
s
.264 .36 .775 r
.84527 .70858 m .80046 .66177 L .79215 .69459 L p F P
0 g
s
0 0 0 r
.38701 .18892 m .40139 .18158 L .46016 .16639 L p F P
0 g
s
0 0 0 r
.40139 .18158 m .38701 .18892 L .32993 .21534 L p F P
0 g
s
.671 .834 .962 r
.4976 .56417 m .44153 .53754 L .42095 .54995 L p F P
0 g
s
.555 .756 .971 r
.5528 .56388 m .60756 .53664 L .5823 .52744 L p F P
0 g
s
.555 .756 .971 r
.5823 .52744 m .54433 .56088 L .5528 .56388 L p F P
0 g
s
.963 .805 .638 r
.21796 .74845 m .19758 .79907 L .22484 .8338 L p F P
0 g
s
.98 .687 .358 r
.22726 .87633 m .22484 .8338 L .19758 .79907 L p F P
0 g
s
.712 .734 .869 r
.54431 .55692 m .58165 .49202 L .60662 .50131 L p F P
0 g
s
.702 .721 .865 r
.58165 .49202 m .54431 .55692 L .53494 .55507 L p F P
0 g
s
.77 .504 .533 r
.54431 .55692 m .54433 .56088 L .53495 .55904 L p F P
0 g
s
.77 .504 .533 r
.53495 .55904 m .53494 .55507 L .54431 .55692 L p F P
0 g
s
.356 0 0 r
.64504 .17998 m .65886 .18715 L .59281 .1711 L p F P
0 g
s
.356 0 0 r
.59281 .1711 m .58579 .16556 L .64504 .17998 L p F P
0 g
s
.702 .721 .865 r
.53494 .55507 m .55371 .48629 L .58165 .49202 L p F P
0 g
s
.719 .688 .819 r
.51189 .9034 m .5064 .95523 L .46543 .95127 L p F P
0 g
s
.696 .667 .818 r
.5064 .95523 m .51189 .9034 L .54156 .90327 L p F P
0 g
s
.696 .667 .818 r
.54156 .90327 m .54807 .95506 L .5064 .95523 L p F P
0 g
s
.648 .831 .971 r
.50597 .56108 m .46647 .52807 L .44153 .53754 L p F P
0 g
s
.67 .648 .82 r
.54807 .95506 m .54156 .90327 L .57059 .90009 L p F P
0 g
s
.67 .648 .82 r
.57059 .90009 m .58891 .95076 L .54807 .95506 L p F P
0 g
s
.564 .775 .977 r
.54433 .56088 m .5823 .52744 L .55403 .52175 L p F P
0 g
s
.564 .775 .977 r
.55403 .52175 m .53495 .55904 L .54433 .56088 L p F P
0 g
s
0 0 .347 r
.82893 .83109 m .82646 .87372 L .85233 .84 L p F P
0 g
s
0 0 .347 r
.85233 .84 m .85507 .79604 L .82893 .83109 L p F P
0 g
s
.578 .608 .846 r
.36124 .4546 m .42216 .51474 L .40707 .52952 L p F P
0 g
s
.578 .608 .846 r
.40707 .52952 m .3382 .47879 L .36124 .4546 L p F P
0 g
s
.457 .421 .73 r
.28491 .33088 m .31408 .39077 L .28535 .42289 L p F P
0 g
s
.529 .528 .799 r
.3382 .47879 m .28535 .42289 L .31408 .39077 L p F P
0 g
s
.623 .822 .977 r
.51529 .55914 m .49453 .52208 L .46647 .52807 L p F P
0 g
s
.579 .793 .98 r
.53495 .55904 m .55403 .52175 L .52425 .51993 L p F P
0 g
s
.226 0 .301 r
.32993 .21534 m .32252 .24724 L .27534 .28197 L p F P
0 g
s
.453 .319 .608 r
.28491 .33088 m .27534 .28197 L .32252 .24724 L p F P
0 g
s
.78 .771 .845 r
.71129 .4767 m .64294 .52824 L .62737 .51361 L p F P
0 g
s
.599 .809 .98 r
.52511 .55845 m .52425 .51993 L .49453 .52208 L p F P
0 g
s
.83 .772 .797 r
.76379 .42012 m .71129 .4767 L .68746 .45275 L p F P
0 g
s
.83 .772 .797 r
.68746 .45275 m .73402 .38829 L .76379 .42012 L p F P
0 g
s
.764 .902 .937 r
.31849 .5987 m .39639 .58071 L .4057 .5646 L p F P
0 g
s
.786 .875 .913 r
.39639 .58071 m .31849 .5987 L .31484 .62509 L p F P
0 g
s
.882 .668 .608 r
.77231 .27887 m .76285 .32797 L .71628 .29594 L p F P
0 g
s
.882 .668 .608 r
.71628 .29594 m .72403 .24469 L .77231 .27887 L p F P
0 g
s
.741 .713 .825 r
.48277 .90047 m .46543 .95127 L .42673 .94329 L p F P
0 g
s
.719 .688 .819 r
.46543 .95127 m .48277 .90047 L .51189 .9034 L p F P
0 g
s
.855 .931 .876 r
.25077 .66453 m .31484 .62509 L .31849 .5987 L p F P
0 g
s
.643 .633 .825 r
.58891 .95076 m .57059 .90009 L .59779 .89397 L p F P
0 g
s
.643 .633 .825 r
.59779 .89397 m .62734 .94246 L .58891 .95076 L p F P
0 g
s
.426 .63 .94 r
.65411 .57935 m .73176 .59653 L .7171 .57096 L p F P
0 g
s
.426 .63 .94 r
.7171 .57096 m .64428 .56333 L .65411 .57935 L p F P
0 g
s
.41 .127 .371 r
.32252 .24724 m .32993 .21534 L .38701 .18892 L p F P
0 g
s
.529 .528 .799 r
.31408 .39077 m .36124 .4546 L .3382 .47879 L p F P
0 g
s
.318 .499 .878 r
.73627 .62291 m .80046 .66177 L .7959 .62801 L p F P
0 g
s
.318 .499 .878 r
.7959 .62801 m .73176 .59653 L .73627 .62291 L p F P
0 g
s
.746 .557 .638 r
.51343 .7477 m .52244 .66235 L .52454 .6626 L p F P
0 g
s
.746 .557 .638 r
.52454 .6626 m .52165 .74862 L .51343 .7477 L p F P
0 g
s
.702 .519 .638 r
.52165 .74862 m .52454 .6626 L .52667 .66258 L p F P
0 g
s
.79 .597 .64 r
.50575 .74586 m .5205 .66185 L .52244 .66235 L p F P
0 g
s
.79 .597 .64 r
.52244 .66235 m .51343 .7477 L .50575 .74586 L p F P
0 g
s
.833 .474 .375 r
.71679 .21284 m .72403 .24469 L .66397 .21843 L p F P
0 g
s
.833 .474 .375 r
.66397 .21843 m .65886 .18715 L .71679 .21284 L p F P
0 g
s
.198 .625 .948 r
.28507 .9306 m .31852 .93503 L .35659 .95394 L p F P
0 g
s
.665 .783 .937 r
.39181 .93153 m .35659 .95394 L .31852 .93503 L p F P
0 g
s
.663 .777 .933 r
.35659 .95394 m .39181 .93153 L .42673 .94329 L p F P
0 g
s
.78 .771 .845 r
.62737 .51361 m .68746 .45275 L .71129 .4767 L p F P
0 g
s
.836 .643 .642 r
.49896 .74317 m .51878 .66113 L .5205 .66185 L p F P
0 g
s
.836 .643 .642 r
.5205 .66185 m .50575 .74586 L .49896 .74317 L p F P
0 g
s
.177 0 0 r
.45252 .17202 m .46016 .16639 L .5229 .16091 L p F P
0 g
s
0 0 0 r
.46016 .16639 m .45252 .17202 L .38701 .18892 L p F P
0 g
s
.884 .696 .643 r
.49337 .73975 m .51737 .66022 L .51878 .66113 L p F P
0 g
s
.884 .696 .643 r
.51878 .66113 m .49896 .74317 L .49337 .73975 L p F P
0 g
s
.457 .349 .643 r
.55256 .74291 m .53238 .66106 L .53376 .66014 L p F P
0 g
s
.299 0 0 r
.58579 .16556 m .59281 .1711 L .52257 .16592 L p F P
0 g
s
.299 0 0 r
.52257 .16592 m .5229 .16091 L .58579 .16556 L p F P
0 g
s
.74 .228 0 r
.22726 .87633 m .24951 .90408 L .28507 .9306 L p F P
0 g
s
.263 .624 .96 r
.31852 .93503 m .28507 .9306 L .24951 .90408 L p F P
0 g
s
.619 .746 .937 r
.69784 .95254 m .66186 .93041 L .69097 .91496 L p F P
0 g
s
.619 .746 .937 r
.69097 .91496 m .73528 .93329 L .69784 .95254 L p F P
0 g
s
.935 .762 .642 r
.48925 .73577 m .51635 .65916 L .51737 .66022 L p F P
0 g
s
.935 .762 .642 r
.51737 .66022 m .49337 .73975 L .48925 .73577 L p F P
0 g
s
.457 .349 .643 r
.53376 .66014 m .55803 .73944 L .55256 .74291 L p F P
0 g
s
.342 .28 .641 r
.55803 .73944 m .53376 .66014 L .53475 .65906 L p F P
0 g
s
.342 .28 .641 r
.53475 .65906 m .56202 .73542 L .55803 .73944 L p F P
0 g
s
.164 .176 .625 r
.56202 .73542 m .53475 .65906 L .53531 .6579 L p F P
0 g
s
.661 .944 .96 r
.76926 .92858 m .73528 .93329 L .76445 .91032 L p F P
0 g
s
.661 .944 .96 r
.76445 .91032 m .80394 .90172 L .76926 .92858 L p F P
0 g
s
.177 0 0 r
.5229 .16091 m .52257 .16592 L .45252 .17202 L p F P
0 g
s
.931 .899 .774 r
.25077 .66453 m .20605 .7117 L .21796 .74845 L p F P
0 g
s
.991 .881 .611 r
.19758 .79907 m .21796 .74845 L .20605 .7117 L p F P
0 g
s
.761 .74 .834 r
.45537 .89458 m .42673 .94329 L .39181 .93153 L p F P
0 g
s
.741 .713 .825 r
.42673 .94329 m .45537 .89458 L .48277 .90047 L p F P
0 g
s
.608 .615 .832 r
.39302 .43392 m .44251 .50222 L .42216 .51474 L p F P
0 g
s
.608 .615 .832 r
.42216 .51474 m .36124 .4546 L .39302 .43392 L p F P
0 g
s
.612 .62 .835 r
.62734 .94246 m .59779 .89397 L .62203 .88512 L p F P
0 g
s
.612 .62 .835 r
.62203 .88512 m .66186 .93041 L .62734 .94246 L p F P
0 g
s
.016 0 .138 r
.49498 .71403 m .51788 .65344 L .5167 .65446 L p F P
0 g
s
.016 0 .138 r
.5167 .65446 m .49032 .71792 L .49498 .71403 L p F P
0 g
s
.894 .472 .153 r
.56034 .71758 m .53425 .65437 L .53303 .65337 L p F P
0 g
s
.766 .742 .831 r
.68746 .45275 m .62737 .51361 L .60662 .50131 L p F P
0 g
s
.268 0 .265 r
.50111 .71078 m .51942 .6526 L .51788 .65344 L p F P
0 g
s
.268 0 .265 r
.51788 .65344 m .49498 .71403 L .50111 .71078 L p F P
0 g
s
.894 .472 .153 r
.53303 .65337 m .55554 .71374 L .56034 .71758 L p F P
0 g
s
.861 .458 .273 r
.55554 .71374 m .53303 .65337 L .53146 .65254 L p F P
0 g
s
.032 .104 .607 r
.83457 .74551 m .85507 .79604 L .86704 .75792 L p F P
0 g
s
.032 .104 .607 r
.86704 .75792 m .84527 .70858 L .83457 .74551 L p F P
0 g
s
.425 .113 .331 r
.50842 .70832 m .52125 .65196 L .51942 .6526 L p F P
0 g
s
.425 .113 .331 r
.51942 .6526 m .50111 .71078 L .50842 .70832 L p F P
0 g
s
.861 .458 .273 r
.53146 .65254 m .5493 .71054 L .55554 .71374 L p F P
0 g
s
.805 .425 .335 r
.5493 .71054 m .53146 .65254 L .52961 .65192 L p F P
0 g
s
.533 .209 .365 r
.51656 .70677 m .52328 .65156 L .52125 .65196 L p F P
0 g
s
.533 .209 .365 r
.52125 .65196 m .50842 .70832 L .51656 .70677 L p F P
0 g
s
.805 .425 .335 r
.52961 .65192 m .5419 .70815 L .5493 .71054 L p F P
0 g
s
.746 .384 .367 r
.5419 .70815 m .52961 .65192 L .52756 .65154 L p F P
0 g
s
.615 .28 .38 r
.52513 .70622 m .52542 .65142 L .52328 .65156 L p F P
0 g
s
.615 .28 .38 r
.52328 .65156 m .51656 .70677 L .52513 .70622 L p F P
0 g
s
.746 .384 .367 r
.52756 .65154 m .53371 .70669 L .5419 .70815 L p F P
0 g
s
.683 .336 .38 r
.53371 .70669 m .52756 .65154 L .52542 .65142 L p F P
0 g
s
.683 .336 .38 r
.52542 .65142 m .52513 .70622 L .53371 .70669 L p F P
0 g
s
.724 .919 .959 r
.33229 .573 m .4057 .5646 L .42095 .54995 L p F P
0 g
s
.764 .902 .937 r
.4057 .5646 m .33229 .573 L .31849 .5987 L p F P
0 g
s
.98 .687 .358 r
.19758 .79907 m .20028 .84291 L .22726 .87633 L p F P
0 g
s
.555 .091 0 r
.24951 .90408 m .22726 .87633 L .20028 .84291 L p F P
0 g
s
.453 .319 .608 r
.32252 .24724 m .33041 .29833 L .28491 .33088 L p F P
0 g
s
.533 .459 .719 r
.31408 .39077 m .28491 .33088 L .33041 .29833 L p F P
0 g
s
.389 .65 .961 r
.64428 .56333 m .7171 .57096 L .6926 .54756 L p F P
0 g
s
.389 .65 .961 r
.6926 .54756 m .62855 .54883 L .64428 .56333 L p F P
0 g
s
.737 .637 .749 r
.51723 .83204 m .51189 .9034 L .48277 .90047 L p F P
0 g
s
.705 .609 .747 r
.51189 .9034 m .51723 .83204 L .53526 .83196 L p F P
0 g
s
.705 .609 .747 r
.53526 .83196 m .54156 .90327 L .51189 .9034 L p F P
0 g
s
.67 .583 .749 r
.54156 .90327 m .53526 .83196 L .55286 .82993 L p F P
0 g
s
.67 .583 .749 r
.55286 .82993 m .57059 .90009 L .54156 .90327 L p F P
0 g
s
.842 .71 .718 r
.76285 .32797 m .73402 .38829 L .69278 .36109 L p F P
0 g
s
.842 .71 .718 r
.69278 .36109 m .71628 .29594 L .76285 .32797 L p F P
0 g
s
.579 .83 .462 r
.82646 .87372 m .80394 .90172 L .82732 .87122 L p F P
0 g
s
.579 .83 .462 r
.82732 .87122 m .85233 .84 L .82646 .87372 L p F P
0 g
s
.768 .67 .754 r
.49956 .83017 m .48277 .90047 L .45537 .89458 L p F P
0 g
s
.737 .637 .749 r
.48277 .90047 m .49956 .83017 L .51723 .83204 L p F P
0 g
s
.633 .559 .754 r
.57059 .90009 m .55286 .82993 L .56929 .82603 L p F P
0 g
s
.633 .559 .754 r
.56929 .82603 m .59779 .89397 L .57059 .90009 L p F P
0 g
s
.744 .596 .692 r
.49956 .83017 m .51343 .7477 L .52165 .74862 L p F P
0 g
s
.744 .596 .692 r
.52165 .74862 m .51723 .83204 L .49956 .83017 L p F P
0 g
s
.636 .624 .822 r
.43207 .41802 m .46717 .49266 L .44251 .50222 L p F P
0 g
s
.636 .624 .822 r
.44251 .50222 m .39302 .43392 L .43207 .41802 L p F P
0 g
s
.531 .232 .405 r
.38169 .22024 m .38701 .18892 L .45252 .17202 L p F P
0 g
s
.41 .127 .371 r
.38701 .18892 m .38169 .22024 L .32252 .24724 L p F P
0 g
s
.782 .633 .696 r
.483 .82642 m .50575 .74586 L .51343 .7477 L p F P
0 g
s
.782 .633 .696 r
.51343 .7477 m .49956 .83017 L .483 .82642 L p F P
0 g
s
.766 .742 .831 r
.60662 .50131 m .655 .43242 L .68746 .45275 L p F P
0 g
s
.748 .715 .821 r
.655 .43242 m .60662 .50131 L .58165 .49202 L p F P
0 g
s
.855 .931 .876 r
.31849 .5987 m .25419 .63079 L .25077 .66453 L p F P
0 g
s
.92 .969 .78 r
.20605 .7117 m .25077 .66453 L .25419 .63079 L p F P
0 g
s
.577 .546 .783 r
.35441 .36311 m .39302 .43392 L .36124 .4546 L p F P
0 g
s
.577 .546 .783 r
.36124 .4546 m .31408 .39077 L .35441 .36311 L p F P
0 g
s
.766 .424 .406 r
.65886 .18715 m .66397 .21843 L .5954 .20201 L p F P
0 g
s
.766 .424 .406 r
.5954 .20201 m .59281 .1711 L .65886 .18715 L p F P
0 g
s
.665 .783 .937 r
.31852 .93503 m .36218 .91635 L .39181 .93153 L p F P
0 g
s
.779 .772 .846 r
.43083 .88595 m .39181 .93153 L .36218 .91635 L p F P
0 g
s
.761 .74 .834 r
.39181 .93153 m .43083 .88595 L .45537 .89458 L p F P
0 g
s
.533 .459 .719 r
.33041 .29833 m .35441 .36311 L .31408 .39077 L p F P
0 g
s
.821 .675 .703 r
.46828 .82093 m .49896 .74317 L .50575 .74586 L p F P
0 g
s
.821 .675 .703 r
.50575 .74586 m .483 .82642 L .46828 .82093 L p F P
0 g
s
.804 .73 .782 r
.73402 .38829 m .68746 .45275 L .655 .43242 L p F P
0 g
s
.544 .382 .612 r
.33041 .29833 m .32252 .24724 L .38169 .22024 L p F P
0 g
s
.089 .298 .779 r
.80046 .66177 m .84527 .70858 L .84107 .67034 L p F P
0 g
s
.089 .298 .779 r
.84107 .67034 m .7959 .62801 L .80046 .66177 L p F P
0 g
s
.579 .61 .848 r
.66186 .93041 m .62203 .88512 L .64224 .87385 L p F P
0 g
s
.579 .61 .848 r
.64224 .87385 m .69097 .91496 L .66186 .93041 L p F P
0 g
s
.815 .974 .897 r
.25419 .63079 m .31849 .5987 L .33229 .573 L p F P
0 g
s
.263 .624 .96 r
.24951 .90408 m .28858 .91233 L .31852 .93503 L p F P
0 g
s
.665 .79 .941 r
.36218 .91635 m .31852 .93503 L .28858 .91233 L p F P
0 g
s
.799 .707 .763 r
.483 .82642 m .45537 .89458 L .43083 .88595 L p F P
0 g
s
.768 .67 .754 r
.45537 .89458 m .483 .82642 L .49956 .83017 L p F P
0 g
s
.804 .73 .782 r
.655 .43242 m .69278 .36109 L .73402 .38829 L p F P
0 g
s
.59 .537 .764 r
.59779 .89397 m .56929 .82603 L .58383 .82041 L p F P
0 g
s
.59 .537 .764 r
.58383 .82041 m .62203 .88512 L .59779 .89397 L p F P
0 g
s
.828 .613 .612 r
.72403 .24469 m .71628 .29594 L .6585 .27136 L p F P
0 g
s
.828 .613 .612 r
.6585 .27136 m .66397 .21843 L .72403 .24469 L p F P
0 g
s
.86 .725 .712 r
.45606 .81394 m .49337 .73975 L .49896 .74317 L p F P
0 g
s
.86 .725 .712 r
.49896 .74317 m .46828 .82093 L .45606 .81394 L p F P
0 g
s
.214 .49 .898 r
.73176 .59653 m .7959 .62801 L .77791 .59496 L p F P
0 g
s
.214 .49 .898 r
.77791 .59496 m .7171 .57096 L .73176 .59653 L p F P
0 g
s
.611 .747 .941 r
.73528 .93329 m .69097 .91496 L .71329 .89666 L p F P
0 g
s
.611 .747 .941 r
.71329 .89666 m .76445 .91032 L .73528 .93329 L p F P
0 g
s
.665 .92 .975 r
.35598 .54937 m .42095 .54995 L .44153 .53754 L p F P
0 g
s
.724 .919 .959 r
.42095 .54995 m .35598 .54937 L .33229 .573 L p F P
0 g
s
.36 .675 .976 r
.62855 .54883 m .6926 .54756 L .65915 .52768 L p F P
0 g
s
.36 .675 .976 r
.65915 .52768 m .60756 .53664 L .62855 .54883 L p F P
0 g
s
.662 .637 .815 r
.47639 .40792 m .49491 .48662 L .46717 .49266 L p F P
0 g
s
.662 .637 .815 r
.46717 .49266 m .43207 .41802 L .47639 .40792 L p F P
0 g
s
.901 .785 .724 r
.44691 .80572 m .48925 .73577 L .49337 .73975 L p F P
0 g
s
.901 .785 .724 r
.49337 .73975 m .45606 .81394 L .44691 .80572 L p F P
0 g
s
.748 .715 .821 r
.58165 .49202 m .61541 .41696 L .655 .43242 L p F P
0 g
s
.729 .691 .814 r
.61541 .41696 m .58165 .49202 L .55371 .48629 L p F P
0 g
s
.991 .881 .611 r
.20605 .7117 m .18432 .76115 L .19758 .79907 L p F P
0 g
s
.914 .666 .167 r
.20028 .84291 m .19758 .79907 L .18432 .76115 L p F P
0 g
s
.799 .707 .763 r
.43083 .88595 m .46828 .82093 L .483 .82642 L p F P
0 g
s
0 0 .143 r
.85507 .79604 m .85233 .84 L .86412 .80335 L p F P
0 g
s
0 0 .143 r
.86412 .80335 m .86704 .75792 L .85507 .79604 L p F P
0 g
s
.622 .309 .419 r
.44969 .20295 m .45252 .17202 L .52257 .16592 L p F P
0 g
s
.531 .232 .405 r
.45252 .17202 m .44969 .20295 L .38169 .22024 L p F P
0 g
s
.938 .857 .735 r
.44131 .79662 m .4868 .73138 L .48925 .73577 L p F P
0 g
s
.938 .857 .735 r
.48925 .73577 m .44691 .80572 L .44131 .79662 L p F P
0 g
s
.697 .371 .419 r
.59281 .1711 m .5954 .20201 L .52245 .19671 L p F P
0 g
s
.697 .371 .419 r
.52245 .19671 m .52257 .16592 L .59281 .1711 L p F P
0 g
s
.686 .653 .811 r
.52352 .40428 m .52427 .48445 L .49491 .48662 L p F P
0 g
s
.686 .653 .811 r
.49491 .48662 m .47639 .40792 L .52352 .40428 L p F P
0 g
s
.729 .691 .814 r
.55371 .48629 m .57076 .40737 L .61541 .41696 L p F P
0 g
s
.708 .671 .811 r
.57076 .40737 m .55371 .48629 L .52427 .48445 L p F P
0 g
s
.555 .091 0 r
.20028 .84291 m .22509 .87384 L .24951 .90408 L p F P
0 g
s
.333 .623 .957 r
.28858 .91233 m .24951 .90408 L .22509 .87384 L p F P
0 g
s
.779 .772 .846 r
.36218 .91635 m .41023 .87486 L .43083 .88595 L p F P
0 g
s
.829 .749 .776 r
.46828 .82093 m .43083 .88595 L .41023 .87486 L p F P
0 g
s
.54 .516 .777 r
.62203 .88512 m .58383 .82041 L .59581 .8133 L p F P
0 g
s
.54 .516 .777 r
.59581 .8133 m .64224 .87385 L .62203 .88512 L p F P
0 g
s
.708 .671 .811 r
.52427 .48445 m .52352 .40428 L .57076 .40737 L p F P
0 g
s
.829 .749 .776 r
.41023 .87486 m .45606 .81394 L .46828 .82093 L p F P
0 g
s
.725 .938 .956 r
.80394 .90172 m .76445 .91032 L .78369 .88439 L p F P
0 g
s
.725 .938 .956 r
.78369 .88439 m .82732 .87122 L .80394 .90172 L p F P
0 g
s
.622 .309 .419 r
.52257 .16592 m .52245 .19671 L .44969 .20295 L p F P
0 g
s
.591 .904 .984 r
.38872 .52915 m .44153 .53754 L .46647 .52807 L p F P
0 g
s
.665 .92 .975 r
.44153 .53754 m .38872 .52915 L .35598 .54937 L p F P
0 g
s
.344 .705 .984 r
.60756 .53664 m .65915 .52768 L .61831 .51255 L p F P
0 g
s
.344 .705 .984 r
.61831 .51255 m .5823 .52744 L .60756 .53664 L p F P
0 g
s
.618 .564 .772 r
.40454 .34172 m .43207 .41802 L .39302 .43392 L p F P
0 g
s
.618 .564 .772 r
.39302 .43392 m .35441 .36311 L .40454 .34172 L p F P
0 g
s
.544 .382 .612 r
.38169 .22024 m .38735 .27306 L .33041 .29833 L p F P
0 g
s
.592 .492 .71 r
.35441 .36311 m .33041 .29833 L .38735 .27306 L p F P
0 g
s
.665 .79 .941 r
.28858 .91233 m .33923 .89826 L .36218 .91635 L p F P
0 g
s
.792 .806 .863 r
.41023 .87486 m .36218 .91635 L .33923 .89826 L p F P
0 g
s
.777 .693 .771 r
.69278 .36109 m .655 .43242 L .61541 .41696 L p F P
0 g
s
.803 .663 .709 r
.71628 .29594 m .69278 .36109 L .64193 .34029 L p F P
0 g
s
.803 .663 .709 r
.64193 .34029 m .6585 .27136 L .71628 .29594 L p F P
0 g
s
.541 .603 .865 r
.69097 .91496 m .64224 .87385 L .65745 .86059 L p F P
0 g
s
.541 .603 .865 r
.65745 .86059 m .71329 .89666 L .69097 .91496 L p F P
0 g
s
.857 .8 .793 r
.39456 .86175 m .44691 .80572 L .45606 .81394 L p F P
0 g
s
.857 .8 .793 r
.45606 .81394 m .41023 .87486 L .39456 .86175 L p F P
0 g
s
.614 .433 .613 r
.38735 .27306 m .38169 .22024 L .44969 .20295 L p F P
0 g
s
.479 .494 .795 r
.64224 .87385 m .59581 .8133 L .60467 .80499 L p F P
0 g
s
.92 .969 .78 r
.25419 .63079 m .20893 .6735 L .20605 .7117 L p F P
0 g
s
.921 .909 .512 r
.18432 .76115 m .20605 .7117 L .20893 .6735 L p F P
0 g
s
.592 .492 .71 r
.38735 .27306 m .40454 .34172 L .35441 .36311 L p F P
0 g
s
.776 .564 .613 r
.66397 .21843 m .6585 .27136 L .59262 .25601 L p F P
0 g
s
.776 .564 .613 r
.59262 .25601 m .5954 .20201 L .66397 .21843 L p F P
0 g
s
.51 .871 .985 r
.42901 .51359 m .46647 .52807 L .49453 .52208 L p F P
0 g
s
.591 .904 .984 r
.46647 .52807 m .42901 .51359 L .38872 .52915 L p F P
0 g
s
.792 .806 .863 r
.33923 .89826 m .39456 .86175 L .41023 .87486 L p F P
0 g
s
.35 .741 .985 r
.5823 .52744 m .61831 .51255 L .57221 .50316 L p F P
0 g
s
.35 .741 .985 r
.57221 .50316 m .55403 .52175 L .5823 .52744 L p F P
0 g
s
.777 .693 .771 r
.61541 .41696 m .64193 .34029 L .69278 .36109 L p F P
0 g
s
.479 .494 .795 r
.60467 .80499 m .65745 .86059 L .64224 .87385 L p F P
0 g
s
.722 .984 .897 r
.27104 .59761 m .33229 .573 L .35598 .54937 L p F P
0 g
s
.815 .974 .897 r
.33229 .573 m .27104 .59761 L .25419 .63079 L p F P
0 g
s
0 0 .497 r
.84527 .70858 m .86704 .75792 L .86312 .71832 L p F P
0 g
s
0 0 .497 r
.86312 .71832 m .84107 .67034 L .84527 .70858 L p F P
0 g
s
.094 .479 .895 r
.7171 .57096 m .77791 .59496 L .7466 .56443 L p F P
0 g
s
.094 .479 .895 r
.7466 .56443 m .6926 .54756 L .7171 .57096 L p F P
0 g
s
.914 .666 .167 r
.18432 .76115 m .18721 .80645 L .20028 .84291 L p F P
0 g
s
0 .098 .654 r
.22509 .87384 m .20028 .84291 L .18721 .80645 L p F P
0 g
s
.333 .623 .957 r
.22509 .87384 m .26843 .88662 L .28858 .91233 L p F P
0 g
s
.664 .797 .946 r
.33923 .89826 m .28858 .91233 L .26843 .88662 L p F P
0 g
s
.437 .828 .984 r
.47476 .5037 m .49453 .52208 L .52425 .51993 L p F P
0 g
s
.51 .871 .985 r
.49453 .52208 m .47476 .5037 L .42901 .51359 L p F P
0 g
s
.381 .783 .984 r
.55403 .52175 m .57221 .50316 L .52342 .50015 L p F P
0 g
s
.381 .783 .984 r
.52342 .50015 m .52425 .51993 L .55403 .52175 L p F P
0 g
s
.88 .858 .815 r
.38463 .84712 m .44131 .79662 L .44691 .80572 L p F P
0 g
s
.88 .858 .815 r
.44691 .80572 m .39456 .86175 L .38463 .84712 L p F P
0 g
s
.655 .585 .764 r
.46181 .32807 m .47639 .40792 L .43207 .41802 L p F P
0 g
s
.655 .585 .764 r
.43207 .41802 m .40454 .34172 L .46181 .32807 L p F P
0 g
s
.401 .471 .817 r
.65745 .86059 m .60467 .80499 L .60995 .79585 L p F P
0 g
s
.603 .748 .946 r
.76445 .91032 m .71329 .89666 L .72757 .87616 L p F P
0 g
s
.603 .748 .946 r
.72757 .87616 m .78369 .88439 L .76445 .91032 L p F P
0 g
s
.748 .661 .764 r
.64193 .34029 m .61541 .41696 L .57076 .40737 L p F P
0 g
s
.821 .991 .739 r
.20893 .6735 m .25419 .63079 L .27104 .59761 L p F P
0 g
s
.437 .828 .984 r
.52425 .51993 m .52342 .50015 L .47476 .5037 L p F P
0 g
s
.858 .975 .666 r
.85233 .84 m .82732 .87122 L .83759 .83818 L p F P
0 g
s
.858 .975 .666 r
.83759 .83818 m .86412 .80335 L .85233 .84 L p F P
0 g
s
.06 0 0 r
.47158 .75304 m .50111 .71078 L .49498 .71403 L p F P
0 g
s
.06 0 0 r
.49498 .71403 m .45821 .76001 L .47158 .75304 L p F P
0 g
s
.753 .238 0 r
.59173 .75938 m .55554 .71374 L .5493 .71054 L p F P
0 g
s
0 .205 .731 r
.7959 .62801 m .84107 .67034 L .82107 .63265 L p F P
0 g
s
0 .205 .731 r
.82107 .63265 m .77791 .59496 L .7959 .62801 L p F P
0 g
s
.673 .478 .613 r
.45269 .25689 m .44969 .20295 L .52245 .19671 L p F P
0 g
s
.614 .433 .613 r
.44969 .20295 m .45269 .25689 L .38735 .27306 L p F P
0 g
s
.725 .52 .613 r
.5954 .20201 m .59262 .25601 L .52256 .25105 L p F P
0 g
s
.725 .52 .613 r
.52256 .25105 m .52245 .19671 L .5954 .20201 L p F P
0 g
s
.283 0 .073 r
.48766 .74773 m .50842 .70832 L .50111 .71078 L p F P
0 g
s
.283 0 .073 r
.50111 .71078 m .47158 .75304 L .48766 .74773 L p F P
0 g
s
.753 .238 0 r
.5493 .71054 m .57809 .75253 L .59173 .75938 L p F P
0 g
s
.732 .243 .081 r
.57809 .75253 m .5493 .71054 L .5419 .70815 L p F P
0 g
s
.641 .523 .704 r
.40454 .34172 m .38735 .27306 L .45269 .25689 L p F P
0 g
s
.43 .012 .143 r
.50566 .74439 m .51656 .70677 L .50842 .70832 L p F P
0 g
s
.43 .012 .143 r
.50842 .70832 m .48766 .74773 L .50566 .74439 L p F P
0 g
s
.764 .623 .704 r
.6585 .27136 m .64193 .34029 L .5842 .32733 L p F P
0 g
s
.764 .623 .704 r
.5842 .32733 m .59262 .25601 L .6585 .27136 L p F P
0 g
s
.664 .797 .946 r
.26843 .88662 m .32422 .87792 L .33923 .89826 L p F P
0 g
s
.8 .843 .884 r
.39456 .86175 m .33923 .89826 L .32422 .87792 L p F P
0 g
s
.732 .243 .081 r
.5419 .70815 m .56182 .74738 L .57809 .75253 L p F P
0 g
s
.681 .218 .147 r
.56182 .74738 m .5419 .70815 L .53371 .70669 L p F P
0 g
s
.688 .607 .76 r
.52292 .32315 m .52352 .40428 L .47639 .40792 L p F P
0 g
s
.688 .607 .76 r
.47639 .40792 m .46181 .32807 L .52292 .32315 L p F P
0 g
s
.748 .661 .764 r
.57076 .40737 m .5842 .32733 L .64193 .34029 L p F P
0 g
s
.719 .632 .76 r
.5842 .32733 m .57076 .40737 L .52352 .40428 L p F P
0 g
s
.8 .843 .884 r
.32422 .87792 m .38463 .84712 L .39456 .86175 L p F P
0 g
s
.673 .478 .613 r
.52245 .19671 m .52256 .25105 L .45269 .25689 L p F P
0 g
s
.889 .922 .838 r
.3811 .83161 m .43958 .78706 L .44131 .79662 L p F P
0 g
s
.889 .922 .838 r
.44131 .79662 m .38463 .84712 L .3811 .83161 L p F P
0 g
s
.535 .107 .174 r
.52466 .7432 m .52513 .70622 L .51656 .70677 L p F P
0 g
s
.535 .107 .174 r
.51656 .70677 m .50566 .74439 L .52466 .7432 L p F P
0 g
s
.681 .218 .147 r
.53371 .70669 m .5437 .74421 L .56182 .74738 L p F P
0 g
s
.616 .173 .175 r
.5437 .74421 m .53371 .70669 L .52513 .70622 L p F P
0 g
s
.498 .599 .886 r
.71329 .89666 m .65745 .86059 L .66685 .84587 L p F P
0 g
s
.498 .599 .886 r
.66685 .84587 m .72757 .87616 L .71329 .89666 L p F P
0 g
s
.401 .471 .817 r
.60995 .79585 m .66685 .84587 L .65745 .86059 L p F P
0 g
s
.296 .443 .84 r
.66685 .84587 m .60995 .79585 L .61134 .78626 L p F P
0 g
s
.616 .173 .175 r
.52513 .70622 m .52466 .7432 L .5437 .74421 L p F P
0 g
s
.641 .523 .704 r
.45269 .25689 m .46181 .32807 L .40454 .34172 L p F P
0 g
s
.719 .632 .76 r
.52352 .40428 m .52292 .32315 L .5842 .32733 L p F P
0 g
s
0 .098 .654 r
.18721 .80645 m .21364 .84097 L .22509 .87384 L p F P
0 g
s
.398 .622 .943 r
.26843 .88662 m .22509 .87384 L .21364 .84097 L p F P
0 g
s
.921 .909 .512 r
.20893 .6735 m .18683 .7216 L .18432 .76115 L p F P
0 g
s
0 0 .144 r
.18721 .80645 m .18432 .76115 L .18683 .7216 L p F P
0 g
s
.564 .937 .859 r
.30124 .56681 m .35598 .54937 L .38872 .52915 L p F P
0 g
s
.722 .984 .897 r
.35598 .54937 m .30124 .56681 L .27104 .59761 L p F P
0 g
s
0 .464 .854 r
.6926 .54756 m .7466 .56443 L .70301 .53829 L p F P
0 g
s
0 .464 .854 r
.70301 .53829 m .65915 .52768 L .6926 .54756 L p F P
0 g
s
.762 .915 .941 r
.82732 .87122 m .78369 .88439 L .79156 .8565 L p F P
0 g
s
.762 .915 .941 r
.79156 .8565 m .83759 .83818 L .82732 .87122 L p F P
0 g
s
.685 .555 .701 r
.46181 .32807 m .45269 .25689 L .52256 .25105 L p F P
0 g
s
.864 .98 .851 r
.38432 .81591 m .44187 .77747 L .43958 .78706 L p F P
0 g
s
.864 .98 .851 r
.43958 .78706 m .3811 .83161 L .38432 .81591 L p F P
0 g
s
.85 .702 .171 r
.86704 .75792 m .86412 .80335 L .86016 .76529 L p F P
0 g
s
.85 .702 .171 r
.86016 .76529 m .86312 .71832 L .86704 .75792 L p F P
0 g
s
.725 .587 .701 r
.59262 .25601 m .5842 .32733 L .52292 .32315 L p F P
0 g
s
.725 .587 .701 r
.52292 .32315 m .52256 .25105 L .59262 .25601 L p F P
0 g
s
.296 .443 .84 r
.61134 .78626 m .66982 .83031 L .66685 .84587 L p F P
0 g
s
.151 .405 .851 r
.66982 .83031 m .61134 .78626 L .6087 .77669 L p F P
0 g
s
.685 .555 .701 r
.52256 .25105 m .52292 .32315 L .46181 .32807 L p F P
0 g
s
.798 .881 .907 r
.31818 .85613 m .3811 .83161 L .38463 .84712 L p F P
0 g
s
.798 .881 .907 r
.38463 .84712 m .32422 .87792 L .31818 .85613 L p F P
0 g
s
.448 .599 .909 r
.72757 .87616 m .66685 .84587 L .66982 .83031 L p F P
0 g
s
.398 .622 .943 r
.21364 .84097 m .25954 .85885 L .26843 .88662 L p F P
0 g
s
.66 .802 .951 r
.32422 .87792 m .26843 .88662 L .25954 .85885 L p F P
0 g
s
.767 .997 .827 r
.39437 .80077 m .44816 .7683 L .44187 .77747 L p F P
0 g
s
.767 .997 .827 r
.44187 .77747 m .38432 .81591 L .39437 .80077 L p F P
0 g
s
.596 .751 .951 r
.78369 .88439 m .72757 .87616 L .7328 .8543 L p F P
0 g
s
.596 .751 .951 r
.7328 .8543 m .79156 .8565 L .78369 .88439 L p F P
0 g
s
.821 .991 .739 r
.27104 .59761 m .22757 .63568 L .20893 .6735 L p F P
0 g
s
0 0 0 r
.18683 .7216 m .20893 .6735 L .22757 .63568 L p F P
0 g
s
.151 .405 .851 r
.6087 .77669 m .66602 .81463 L .66982 .83031 L p F P
0 g
s
.66 .802 .951 r
.25954 .85885 m .31818 .85613 L .32422 .87792 L p F P
0 g
s
.57 .882 .6 r
.22757 .63568 m .27104 .59761 L .30124 .56681 L p F P
0 g
s
.448 .599 .909 r
.66982 .83031 m .7328 .8543 L .72757 .87616 L p F P
0 g
s
.696 .317 0 r
.84107 .67034 m .86312 .71832 L .84225 .67915 L p F P
0 g
s
.696 .317 0 r
.84225 .67915 m .82107 .63265 L .84107 .67034 L p F P
0 g
s
0 .086 .583 r
.77791 .59496 m .82107 .63265 L .78523 .5976 L p F P
0 g
s
0 .086 .583 r
.78523 .5976 m .7466 .56443 L .77791 .59496 L p F P
0 g
s
.364 .832 .787 r
.34385 .54023 m .38872 .52915 L .42901 .51359 L p F P
0 g
s
.564 .937 .859 r
.38872 .52915 m .34385 .54023 L .30124 .56681 L p F P
0 g
s
0 .453 .78 r
.65915 .52768 m .70301 .53829 L .64912 .51826 L p F P
0 g
s
0 .453 .78 r
.64912 .51826 m .61831 .51255 L .65915 .52768 L p F P
0 g
s
0 0 .144 r
.18683 .7216 m .18977 .76844 L .18721 .80645 L p F P
0 g
s
.077 .249 .738 r
.21364 .84097 m .18721 .80645 L .18977 .76844 L p F P
0 g
s
.565 .92 .733 r
.44816 .7683 m .39437 .80077 L .41097 .78698 L p F P
0 g
s
.946 .951 .741 r
.86412 .80335 m .83759 .83818 L .83331 .80402 L p F P
0 g
s
.946 .951 .741 r
.83331 .80402 m .86016 .76529 L .86412 .80335 L p F P
0 g
s
.781 .917 .93 r
.32183 .83387 m .38432 .81591 L .3811 .83161 L p F P
0 g
s
.781 .917 .93 r
.3811 .83161 m .31818 .85613 L .32183 .83387 L p F P
0 g
s
.391 .602 .932 r
.7328 .8543 m .66982 .83031 L .66602 .81463 L p F P
0 g
s
0 0 0 r
.45821 .76001 m .41679 .78365 L p .43343 .77528 L F P
0 g
s
0 0 0 r
.47158 .75304 m .43343 .77528 L .46073 .76633 L p F P
0 g
s
.464 .881 .931 r
.43061 .76214 m .45326 .76529 L p .43343 .77528 L F P
0 g
s
.464 .881 .931 r
.45326 .76529 m .46073 .76633 L .43343 .77528 L p F P
0 g
s
0 0 0 r
.41679 .78365 m .41097 .78698 L .43343 .77528 L p F P
0 g
s
.168 .702 .712 r
.39697 .51965 m .42901 .51359 L .47476 .5037 L p F P
0 g
s
.364 .832 .787 r
.42901 .51359 m .39697 .51965 L .34385 .54023 L p F P
0 g
s
0 .463 .707 r
.61831 .51255 m .64912 .51826 L .58782 .50578 L p F P
0 g
s
0 .463 .707 r
.58782 .50578 m .57221 .50316 L .61831 .51255 L p F P
0 g
s
.077 .249 .738 r
.18977 .76844 m .21663 .80684 L .21364 .84097 L p F P
0 g
s
.455 .622 .923 r
.25954 .85885 m .21364 .84097 L .21663 .80684 L p F P
0 g
s
0 0 0 r
.46073 .76633 m .48766 .74773 L .47158 .75304 L p F P
0 g
s
.778 .882 .921 r
.83759 .83818 m .79156 .8565 L .78697 .82785 L p F P
0 g
s
.778 .882 .921 r
.78697 .82785 m .83331 .80402 L .83759 .83818 L p F P
0 g
s
.655 .807 .956 r
.26303 .83021 m .32183 .83387 L .31818 .85613 L p F P
0 g
s
.655 .807 .956 r
.31818 .85613 m .25954 .85885 L .26303 .83021 L p F P
0 g
s
0 0 0 r
.22757 .63568 m .20624 .68231 L .18683 .7216 L p F P
0 g
s
0 0 .383 r
.18977 .76844 m .18683 .7216 L .20624 .68231 L p F P
0 g
s
.59 .755 .957 r
.79156 .8565 m .7328 .8543 L .72831 .83204 L p F P
0 g
s
.391 .602 .932 r
.66602 .81463 m .72831 .83204 L .7328 .8543 L p F P
0 g
s
.168 .702 .712 r
.47476 .5037 m .45777 .50649 L .39697 .51965 L p F P
0 g
s
0 0 0 r
.45777 .50649 m .47476 .5037 L .52342 .50015 L p F P
0 g
s
.086 0 0 r
.57221 .50316 m .58782 .50578 L .52271 .50175 L p F P
0 g
s
.086 0 0 r
.52271 .50175 m .52342 .50015 L .57221 .50316 L p F P
0 g
s
.455 .622 .923 r
.21663 .80684 m .26303 .83021 L .25954 .85885 L p F P
0 g
s
.742 .942 .949 r
.33544 .8122 m .39437 .80077 L .38432 .81591 L p F P
0 g
s
.742 .942 .949 r
.38432 .81591 m .32183 .83387 L .33544 .8122 L p F P
0 g
s
.993 .76 .397 r
.86312 .71832 m .86016 .76529 L .8394 .72766 L p F P
0 g
s
.993 .76 .397 r
.8394 .72766 m .84225 .67915 L .86312 .71832 L p F P
0 g
s
.166 0 0 r
.49148 .76067 m .50566 .74439 L .48766 .74773 L p F P
0 g
s
.166 0 0 r
.48766 .74773 m .46073 .76633 L .49148 .76067 L p F P
0 g
s
.327 .608 .95 r
.72831 .83204 m .66602 .81463 L .65539 .79957 L p F P
0 g
s
.57 .882 .6 r
.30124 .56681 m .26209 .60034 L .22757 .63568 L p F P
0 g
s
0 0 0 r
.26209 .60034 m .30124 .56681 L .34385 .54023 L p F P
0 g
s
0 0 0 r
.52342 .50015 m .52271 .50175 L .45777 .50649 L p F P
0 g
s
.59 .755 .957 r
.72831 .83204 m .78697 .82785 L .79156 .8565 L p F P
0 g
s
.576 .012 0 r
.7466 .56443 m .78523 .5976 L .73459 .56741 L p F P
0 g
s
.576 .012 0 r
.73459 .56741 m .70301 .53829 L .7466 .56443 L p F P
0 g
s
.304 0 0 r
.52406 .75864 m .52466 .7432 L .50566 .74439 L p F P
0 g
s
.304 0 0 r
.50566 .74439 m .49148 .76067 L .52406 .75864 L p F P
0 g
s
0 0 .009 r
.20624 .68231 m .22757 .63568 L .26209 .60034 L p F P
0 g
s
.867 .43 .029 r
.82107 .63265 m .84225 .67915 L .80427 .64261 L p F P
0 g
s
.867 .43 .029 r
.80427 .64261 m .78523 .5976 L .82107 .63265 L p F P
0 g
s
.674 .948 .958 r
.35878 .79228 m .41097 .78698 L .39437 .80077 L p F P
0 g
s
.674 .948 .958 r
.39437 .80077 m .33544 .8122 L .35878 .79228 L p F P
0 g
s
.262 .616 .958 r
.71385 .81048 m .65539 .79957 L .63826 .78592 L p F P
0 g
s
.327 .608 .95 r
.65539 .79957 m .71385 .81048 L .72831 .83204 L p F P
0 g
s
0 0 .383 r
.20624 .68231 m .20908 .7307 L .18977 .76844 L p F P
0 g
s
.283 .344 .75 r
.21663 .80684 m .18977 .76844 L .20908 .7307 L p F P
0 g
s
.648 .81 .961 r
.27949 .80209 m .33544 .8122 L .32183 .83387 L p F P
0 g
s
.648 .81 .961 r
.32183 .83387 m .26303 .83021 L .27949 .80209 L p F P
0 g
s
.936 .873 .75 r
.86016 .76529 m .83331 .80402 L .81363 .77037 L p F P
0 g
s
.936 .873 .75 r
.81363 .77037 m .8394 .72766 L .86016 .76529 L p F P
0 g
s
.586 .76 .962 r
.78697 .82785 m .72831 .83204 L .71385 .81048 L p F P
0 g
s
.072 0 0 r
.31155 .56967 m .34385 .54023 L .39697 .51965 L p F P
0 g
s
0 0 0 r
.34385 .54023 m .31155 .56967 L .26209 .60034 L p F P
0 g
s
.619 .055 0 r
.70301 .53829 m .73459 .56741 L .67141 .54418 L p F P
0 g
s
.619 .055 0 r
.67141 .54418 m .64912 .51826 L .70301 .53829 L p F P
0 g
s
.283 .344 .75 r
.20908 .7307 m .23499 .77309 L .21663 .80684 L p F P
0 g
s
.505 .625 .903 r
.26303 .83021 m .21663 .80684 L .23499 .77309 L p F P
0 g
s
.578 .928 .951 r
.39099 .77525 m .43343 .77528 L .41097 .78698 L p F P
0 g
s
.578 .928 .951 r
.41097 .78698 m .35878 .79228 L .39099 .77525 L p F P
0 g
s
.505 .625 .903 r
.23499 .77309 m .27949 .80209 L .26303 .83021 L p F P
0 g
s
.262 .616 .958 r
.63826 .78592 m .68971 .79075 L .71385 .81048 L p F P
0 g
s
.781 .847 .901 r
.83331 .80402 m .78697 .82785 L .7694 .79984 L p F P
0 g
s
.781 .847 .901 r
.7694 .79984 m .81363 .77037 L .83331 .80402 L p F P
0 g
s
.586 .76 .962 r
.71385 .81048 m .7694 .79984 L .78697 .82785 L p F P
0 g
s
0 0 .009 r
.26209 .60034 m .24278 .64547 L .20624 .68231 L p F P
0 g
s
.203 .086 .488 r
.20908 .7307 m .20624 .68231 L .24278 .64547 L p F P
0 g
s
.965 .706 .494 r
.84225 .67915 m .8394 .72766 L .80171 .69258 L p F P
0 g
s
.965 .706 .494 r
.80171 .69258 m .80427 .64261 L .84225 .67915 L p F P
0 g
s
.166 0 .187 r
.24278 .64547 m .26209 .60034 L .31155 .56967 L p F P
0 g
s
.464 .881 .931 r
.43343 .77528 m .39099 .77525 L .43061 .76214 L p F P
0 g
s
.866 .439 .198 r
.78523 .5976 m .80427 .64261 L .75022 .61105 L p F P
0 g
s
.866 .439 .198 r
.75022 .61105 m .73459 .56741 L .78523 .5976 L p F P
0 g
s
.203 .626 .95 r
.61534 .77443 m .6568 .774 L .68971 .79075 L p F P
0 g
s
.28 0 0 r
.37378 .54578 m .39697 .51965 L .45777 .50649 L p F P
0 g
s
.072 0 0 r
.39697 .51965 m .37378 .54578 L .31155 .56967 L p F P
0 g
s
.639 .812 .966 r
.30882 .77602 m .35878 .79228 L .33544 .8122 L p F P
0 g
s
.639 .812 .966 r
.33544 .8122 m .27949 .80209 L .30882 .77602 L p F P
0 g
s
.59 .048 0 r
.64912 .51826 m .67141 .54418 L .59915 .52964 L p F P
0 g
s
.59 .048 0 r
.59915 .52964 m .58782 .50578 L .64912 .51826 L p F P
0 g
s
.583 .766 .966 r
.7694 .79984 m .71385 .81048 L .68971 .79075 L p F P
0 g
s
.349 .818 .906 r
.47557 .75381 m .49148 .76067 L .46073 .76633 L p F P
0 g
s
.349 .818 .906 r
.46073 .76633 m .43061 .76214 L .47557 .75381 L p F P
0 g
s
.164 .64 .928 r
.58769 .76574 m .61665 .76127 L .6568 .774 L p F P
0 g
s
.423 0 0 r
.44543 .53047 m .45777 .50649 L .52271 .50175 L p F P
0 g
s
.28 0 0 r
.45777 .50649 m .44543 .53047 L .37378 .54578 L p F P
0 g
s
.523 .002 0 r
.58782 .50578 m .59915 .52964 L .52218 .52494 L p F P
0 g
s
.523 .002 0 r
.52218 .52494 m .52271 .50175 L .58782 .50578 L p F P
0 g
s
.203 .086 .488 r
.24278 .64547 m .24536 .69532 L .20908 .7307 L p F P
0 g
s
.416 .407 .741 r
.23499 .77309 m .20908 .7307 L .24536 .69532 L p F P
0 g
s
.252 .754 .889 r
.52339 .75082 m .52406 .75864 L .49148 .76067 L p F P
0 g
s
.252 .754 .889 r
.49148 .76067 m .47557 .75381 L .52339 .75082 L p F P
0 g
s
.156 .663 .904 r
.55672 .76037 m .57134 .75336 L .61665 .76127 L p F P
0 g
s
.423 0 0 r
.52271 .50175 m .52218 .52494 L .44543 .53047 L p F P
0 g
s
.898 .798 .74 r
.8394 .72766 m .81363 .77037 L .77855 .73914 L p F P
0 g
s
.898 .798 .74 r
.77855 .73914 m .80171 .69258 L .8394 .72766 L p F P
0 g
s
.185 .701 .888 r
.52406 .75864 m .52339 .75082 L .57134 .75336 L p F P
0 g
s
.548 .63 .884 r
.26878 .74157 m .30882 .77602 L .27949 .80209 L p F P
0 g
s
.548 .63 .884 r
.27949 .80209 m .23499 .77309 L .26878 .74157 L p F P
0 g
s
.774 .811 .883 r
.81363 .77037 m .7694 .79984 L .73901 .77401 L p F P
0 g
s
.583 .766 .966 r
.68971 .79075 m .73901 .77401 L .7694 .79984 L p F P
0 g
s
.629 .811 .97 r
.35007 .75356 m .39099 .77525 L .35878 .79228 L p F P
0 g
s
.629 .811 .97 r
.35878 .79228 m .30882 .77602 L .35007 .75356 L p F P
0 g
s
.416 .407 .741 r
.24536 .69532 m .26878 .74157 L .23499 .77309 L p F P
0 g
s
.582 .773 .97 r
.73901 .77401 m .68971 .79075 L .6568 .774 L p F P
0 g
s
.774 .811 .883 r
.73901 .77401 m .77855 .73914 L .81363 .77037 L p F P
0 g
s
.369 .045 .28 r
.29555 .61341 m .31155 .56967 L .37378 .54578 L p F P
0 g
s
.166 0 .187 r
.31155 .56967 m .29555 .61341 L .24278 .64547 L p F P
0 g
s
.812 .409 .286 r
.73459 .56741 m .75022 .61105 L .68248 .5867 L p F P
0 g
s
.812 .409 .286 r
.68248 .5867 m .67141 .54418 L .73459 .56741 L p F P
0 g
s
.387 .221 .531 r
.24536 .69532 m .24278 .64547 L .29555 .61341 L p F P
0 g
s
.902 .64 .533 r
.80427 .64261 m .80171 .69258 L .7481 .66228 L p F P
0 g
s
.902 .64 .533 r
.7481 .66228 m .75022 .61105 L .80427 .64261 L p F P
0 g
s
.619 .809 .973 r
.4014 .73617 m .43061 .76214 L .39099 .77525 L p F P
0 g
s
.619 .809 .973 r
.39099 .77525 m .35007 .75356 L .4014 .73617 L p F P
0 g
s
.582 .773 .97 r
.6568 .774 m .69681 .75191 L .73901 .77401 L p F P
0 g
s
.583 .78 .973 r
.69681 .75191 m .6568 .774 L .61665 .76127 L p F P
0 g
s
.586 .638 .869 r
.31707 .71426 m .35007 .75356 L .30882 .77602 L p F P
0 g
s
.586 .638 .869 r
.30882 .77602 m .26878 .74157 L .31707 .71426 L p F P
0 g
s
.76 .779 .868 r
.77855 .73914 m .73901 .77401 L .69681 .75191 L p F P
0 g
s
.502 .164 .327 r
.36226 .58838 m .37378 .54578 L .44543 .53047 L p F P
0 g
s
.369 .045 .28 r
.37378 .54578 m .36226 .58838 L .29555 .61341 L p F P
0 g
s
.387 .221 .531 r
.29555 .61341 m .29768 .66455 L .24536 .69532 L p F P
0 g
s
.509 .453 .73 r
.26878 .74157 m .24536 .69532 L .29768 .66455 L p F P
0 g
s
.748 .365 .329 r
.67141 .54418 m .68248 .5867 L .60477 .57143 L p F P
0 g
s
.748 .365 .329 r
.60477 .57143 m .59915 .52964 L .67141 .54418 L p F P
0 g
s
.855 .734 .729 r
.80171 .69258 m .77855 .73914 L .72911 .71225 L p F P
0 g
s
.855 .734 .729 r
.72911 .71225 m .7481 .66228 L .80171 .69258 L p F P
0 g
s
.609 .806 .975 r
.4601 .72506 m .47557 .75381 L .43061 .76214 L p F P
0 g
s
.609 .806 .975 r
.43061 .76214 m .4014 .73617 L .4601 .72506 L p F P
0 g
s
.583 .78 .973 r
.61665 .76127 m .64474 .735 L .69681 .75191 L p F P
0 g
s
.587 .787 .975 r
.64474 .735 m .61665 .76127 L .57134 .75336 L p F P
0 g
s
.509 .453 .73 r
.29768 .66455 m .31707 .71426 L .26878 .74157 L p F P
0 g
s
.76 .779 .868 r
.69681 .75191 m .72911 .71225 L .77855 .73914 L p F P
0 g
s
.506 .311 .549 r
.29768 .66455 m .29555 .61341 L .36226 .58838 L p F P
0 g
s
.599 .248 .346 r
.43929 .57231 m .44543 .53047 L .52218 .52494 L p F P
0 g
s
.502 .164 .327 r
.44543 .53047 m .43929 .57231 L .36226 .58838 L p F P
0 g
s
.839 .582 .55 r
.75022 .61105 m .7481 .66228 L .68097 .63892 L p F P
0 g
s
.839 .582 .55 r
.68097 .63892 m .68248 .5867 L .75022 .61105 L p F P
0 g
s
.6 .801 .976 r
.52275 .72106 m .52339 .75082 L .47557 .75381 L p F P
0 g
s
.6 .801 .976 r
.47557 .75381 m .4601 .72506 L .52275 .72106 L p F P
0 g
s
.678 .313 .347 r
.59915 .52964 m .60477 .57143 L .52191 .5665 L p F P
0 g
s
.678 .313 .347 r
.52191 .5665 m .52218 .52494 L .59915 .52964 L p F P
0 g
s
.587 .787 .975 r
.57134 .75336 m .58557 .72446 L .64474 .735 L p F P
0 g
s
.593 .794 .976 r
.58557 .72446 m .57134 .75336 L .52339 .75082 L p F P
0 g
s
.593 .794 .976 r
.52339 .75082 m .52275 .72106 L .58557 .72446 L p F P
0 g
s
.599 .248 .346 r
.52218 .52494 m .52191 .5665 L .43929 .57231 L p F P
0 g
s
.619 .649 .858 r
.37774 .69301 m .4014 .73617 L .35007 .75356 L p F P
0 g
s
.619 .649 .858 r
.35007 .75356 m .31707 .71426 L .37774 .69301 L p F P
0 g
s
.743 .749 .857 r
.72911 .71225 m .69681 .75191 L .64474 .735 L p F P
0 g
s
.506 .311 .549 r
.36226 .58838 m .36379 .64054 L .29768 .66455 L p F P
0 g
s
.578 .492 .721 r
.31707 .71426 m .29768 .66455 L .36379 .64054 L p F P
0 g
s
.812 .681 .72 r
.7481 .66228 m .72911 .71225 L .66752 .69158 L p F P
0 g
s
.812 .681 .72 r
.66752 .69158 m .68097 .63892 L .7481 .66228 L p F P
0 g
s
.578 .492 .721 r
.36379 .64054 m .37774 .69301 L .31707 .71426 L p F P
0 g
s
.592 .377 .558 r
.36379 .64054 m .36226 .58838 L .43929 .57231 L p F P
0 g
s
.743 .749 .857 r
.64474 .735 m .66752 .69158 L .72911 .71225 L p F P
0 g
s
.78 .53 .558 r
.68248 .5867 m .68097 .63892 L .604 .62427 L p F P
0 g
s
.78 .53 .558 r
.604 .62427 m .60477 .57143 L .68248 .5867 L p F P
0 g
s
.649 .663 .85 r
.44752 .67939 m .4601 .72506 L .4014 .73617 L p F P
0 g
s
.649 .663 .85 r
.4014 .73617 m .37774 .69301 L .44752 .67939 L p F P
0 g
s
.723 .723 .85 r
.66752 .69158 m .64474 .735 L .58557 .72446 L p F P
0 g
s
.676 .68 .846 r
.52223 .67447 m .52275 .72106 L .4601 .72506 L p F P
0 g
s
.676 .68 .846 r
.4601 .72506 m .44752 .67939 L .52223 .67447 L p F P
0 g
s
.661 .432 .561 r
.4401 .62512 m .43929 .57231 L .52191 .5665 L p F P
0 g
s
.592 .377 .558 r
.43929 .57231 m .4401 .62512 L .36379 .64054 L p F P
0 g
s
.723 .723 .85 r
.58557 .72446 m .59715 .67865 L .66752 .69158 L p F P
0 g
s
.701 .7 .846 r
.59715 .67865 m .58557 .72446 L .52275 .72106 L p F P
0 g
s
.722 .481 .562 r
.60477 .57143 m .604 .62427 L .52193 .61954 L p F P
0 g
s
.722 .481 .562 r
.52193 .61954 m .52191 .5665 L .60477 .57143 L p F P
0 g
s
.701 .7 .846 r
.52275 .72106 m .52223 .67447 L .59715 .67865 L p F P
0 g
s
.634 .527 .714 r
.37774 .69301 m .36379 .64054 L .4401 .62512 L p F P
0 g
s
.661 .432 .561 r
.52191 .5665 m .52193 .61954 L .4401 .62512 L p F P
0 g
s
.634 .527 .714 r
.4401 .62512 m .44752 .67939 L .37774 .69301 L p F P
0 g
s
.769 .636 .714 r
.68097 .63892 m .66752 .69158 L .59715 .67865 L p F P
0 g
s
.769 .636 .714 r
.59715 .67865 m .604 .62427 L .68097 .63892 L p F P
0 g
s
.683 .561 .711 r
.52193 .61954 m .52223 .67447 L .44752 .67939 L p F P
0 g
s
.683 .561 .711 r
.44752 .67939 m .4401 .62512 L .52193 .61954 L p F P
0 g
s
.727 .597 .711 r
.604 .62427 m .59715 .67865 L .52223 .67447 L p F P
0 g
s
.727 .597 .711 r
.52223 .67447 m .52193 .61954 L .604 .62427 L p F P
0 g
s
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.678571 0.0121928 0.97619 0.499904 MathSubStart
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-5.56738e-017 1.0498 -0.0709627 1.0498 [
[ 0 0 0 0 ]
[ 1 .80506 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .80506 L
0 .80506 L
closepath
clip
newpath
.5 Mabswid
[ ] 0 setdash
.729 .642 .761 r
.44785 .51317 m .43934 .54633 L .49716 .55312 L p F P
0 g
s
.729 .642 .761 r
.49716 .55312 m .49986 .5194 L .44785 .51317 L p F P
0 g
s
.738 .603 .708 r
.49516 .58724 m .49716 .55312 L .43934 .54633 L p F P
0 g
s
.704 .619 .76 r
.49986 .5194 m .49716 .55312 L .55591 .55283 L p F P
0 g
s
.706 .576 .706 r
.49716 .55312 m .49516 .58724 L .55827 .58693 L p F P
0 g
s
.706 .576 .706 r
.55827 .58693 m .55591 .55283 L .49716 .55312 L p F P
0 g
s
.738 .603 .708 r
.43934 .54633 m .43301 .58009 L .49516 .58724 L p F P
0 g
s
.742 .559 .647 r
.49399 .61949 m .49516 .58724 L .43301 .58009 L p F P
0 g
s
.703 .526 .646 r
.49516 .58724 m .49399 .61949 L .55965 .61917 L p F P
0 g
s
.703 .526 .646 r
.55965 .61917 m .55827 .58693 L .49516 .58724 L p F P
0 g
s
.704 .619 .76 r
.55591 .55283 m .55272 .51913 L .49986 .5194 L p F P
0 g
s
.677 .6 .761 r
.55272 .51913 m .55591 .55283 L .61356 .54546 L p F P
0 g
s
.672 .55 .708 r
.55591 .55283 m .55827 .58693 L .62024 .57917 L p F P
0 g
s
.672 .55 .708 r
.62024 .57917 m .61356 .54546 L .55591 .55283 L p F P
0 g
s
.662 .494 .647 r
.55827 .58693 m .55965 .61917 L .62414 .61126 L p F P
0 g
s
.662 .494 .647 r
.62414 .61126 m .62024 .57917 L .55827 .58693 L p F P
0 g
s
.753 .667 .766 r
.39858 .5006 m .38444 .53262 L .43934 .54633 L p F P
0 g
s
.753 .667 .766 r
.43934 .54633 m .44785 .51317 L .39858 .5006 L p F P
0 g
s
.769 .634 .712 r
.43301 .58009 m .43934 .54633 L .38444 .53262 L p F P
0 g
s
.715 .678 .813 r
.45801 .48271 m .44785 .51317 L .49986 .5194 L p F P
0 g
s
.715 .678 .813 r
.49986 .5194 m .50308 .48821 L .45801 .48271 L p F P
0 g
s
.697 .662 .812 r
.50308 .48821 m .49986 .5194 L .55272 .51913 L p F P
0 g
s
.697 .662 .812 r
.55272 .51913 m .54891 .48797 L .50308 .48821 L p F P
0 g
s
.678 .648 .813 r
.54891 .48797 m .55272 .51913 L .60457 .51237 L p F P
0 g
s
.677 .6 .761 r
.61356 .54546 m .60457 .51237 L .55272 .51913 L p F P
0 g
s
.742 .559 .647 r
.43301 .58009 m .42931 .6122 L .49399 .61949 L p F P
0 g
s
.74 .5 .566 r
.49374 .64761 m .49399 .61949 L .42931 .6122 L p F P
0 g
s
.693 .461 .567 r
.49399 .61949 m .49374 .64761 L .55993 .6473 L p F P
0 g
s
.693 .461 .567 r
.55993 .6473 m .55965 .61917 L .49399 .61949 L p F P
0 g
s
.769 .634 .712 r
.38444 .53262 m .37393 .56564 L .43301 .58009 L p F P
0 g
s
.78 .594 .649 r
.42931 .6122 m .43301 .58009 L .37393 .56564 L p F P
0 g
s
.644 .422 .566 r
.55965 .61917 m .55993 .6473 L .62495 .63946 L p F P
0 g
s
.644 .422 .566 r
.62495 .63946 m .62414 .61126 L .55965 .61917 L p F P
0 g
s
.649 .582 .766 r
.60457 .51237 m .61356 .54546 L .66811 .53119 L p F P
0 g
s
.636 .526 .712 r
.61356 .54546 m .62024 .57917 L .67896 .56413 L p F P
0 g
s
.636 .526 .712 r
.67896 .56413 m .66811 .53119 L .61356 .54546 L p F P
0 g
s
.617 .462 .649 r
.62024 .57917 m .62414 .61126 L .6853 .59591 L p F P
0 g
s
.617 .462 .649 r
.6853 .59591 m .67896 .56413 L .62024 .57917 L p F P
0 g
s
.731 .696 .817 r
.41542 .47163 m .39858 .5006 L .44785 .51317 L p F P
0 g
s
.731 .696 .817 r
.44785 .51317 m .45801 .48271 L .41542 .47163 L p F P
0 g
s
.678 .648 .813 r
.60457 .51237 m .59383 .48201 L .54891 .48797 L p F P
0 g
s
.658 .637 .818 r
.59383 .48201 m .60457 .51237 L .65352 .49929 L p F P
0 g
s
.649 .582 .766 r
.66811 .53119 m .65352 .49929 L .60457 .51237 L p F P
0 g
s
.78 .594 .649 r
.37393 .56564 m .36777 .59745 L .42931 .6122 L p F P
0 g
s
.785 .539 .564 r
.42853 .64039 m .42931 .6122 L .36777 .59745 L p F P
0 g
s
.74 .5 .566 r
.42931 .6122 m .42853 .64039 L .49374 .64761 L p F P
0 g
s
.588 .378 .564 r
.62414 .61126 m .62495 .63946 L .68663 .62426 L p F P
0 g
s
.588 .378 .564 r
.68663 .62426 m .6853 .59591 L .62414 .61126 L p F P
0 g
s
.691 .715 .868 r
.46918 .45682 m .45801 .48271 L .50308 .48821 L p F P
0 g
s
.691 .715 .868 r
.50308 .48821 m .50661 .46146 L .46918 .45682 L p F P
0 g
s
.682 .708 .867 r
.50661 .46146 m .50308 .48821 L .54891 .48797 L p F P
0 g
s
.682 .708 .867 r
.54891 .48797 m .54472 .46126 L .50661 .46146 L p F P
0 g
s
.673 .701 .868 r
.54472 .46126 m .54891 .48797 L .59383 .48201 L p F P
0 g
s
.777 .695 .774 r
.35392 .48201 m .33452 .51229 L .38444 .53262 L p F P
0 g
s
.777 .695 .774 r
.38444 .53262 m .39858 .5006 L .35392 .48201 L p F P
0 g
s
.8 .669 .719 r
.37393 .56564 m .38444 .53262 L .33452 .51229 L p F P
0 g
s
.723 .405 .441 r
.49445 .66948 m .49374 .64761 L .42853 .64039 L p F P
0 g
s
.668 .361 .444 r
.49374 .64761 m .49445 .66948 L .55907 .66918 L p F P
0 g
s
.668 .361 .444 r
.55907 .66918 m .55993 .6473 L .49374 .64761 L p F P
0 g
s
.607 .31 .44 r
.55993 .6473 m .55907 .66918 L .62255 .66164 L p F P
0 g
s
.607 .31 .44 r
.62255 .66164 m .62495 .63946 L .55993 .6473 L p F P
0 g
s
.618 .567 .774 r
.65352 .49929 m .66811 .53119 L .71752 .51034 L p F P
0 g
s
.595 .503 .719 r
.66811 .53119 m .67896 .56413 L .73228 .54213 L p F P
0 g
s
.595 .503 .719 r
.73228 .54213 m .71752 .51034 L .66811 .53119 L p F P
0 g
s
.8 .669 .719 r
.33452 .51229 m .32006 .5442 L .37393 .56564 L p F P
0 g
s
.819 .634 .652 r
.36777 .59745 m .37393 .56564 L .32006 .5442 L p F P
0 g
s
.565 .428 .652 r
.67896 .56413 m .6853 .59591 L .74093 .57343 L p F P
0 g
s
.565 .428 .652 r
.74093 .57343 m .73228 .54213 L .67896 .56413 L p F P
0 g
s
.746 .716 .824 r
.37696 .45529 m .35392 .48201 L .39858 .5006 L p F P
0 g
s
.746 .716 .824 r
.39858 .5006 m .41542 .47163 L .37696 .45529 L p F P
0 g
s
.698 .724 .87 r
.43387 .44749 m .41542 .47163 L .45801 .48271 L p F P
0 g
s
.698 .724 .87 r
.45801 .48271 m .46918 .45682 L .43387 .44749 L p F P
0 g
s
.673 .701 .868 r
.59383 .48201 m .58204 .45623 L .54472 .46126 L p F P
0 g
s
.664 .697 .87 r
.58204 .45623 m .59383 .48201 L .63614 .47048 L p F P
0 g
s
.658 .637 .818 r
.65352 .49929 m .63614 .47048 L .59383 .48201 L p F P
0 g
s
.638 .629 .825 r
.63614 .47048 m .65352 .49929 L .6977 .48022 L p F P
0 g
s
.618 .567 .774 r
.71752 .51034 m .6977 .48022 L .65352 .49929 L p F P
0 g
s
.785 .539 .564 r
.36777 .59745 m .36648 .62579 L .42853 .64039 L p F P
0 g
s
.776 .445 .428 r
.43079 .66254 m .42853 .64039 L .36648 .62579 L p F P
0 g
s
.723 .405 .441 r
.42853 .64039 m .43079 .66254 L .49445 .66948 L p F P
0 g
s
.534 .248 .426 r
.62495 .63946 m .62255 .66164 L .68272 .64701 L p F P
0 g
s
.534 .248 .426 r
.68272 .64701 m .68663 .62426 L .62495 .63946 L p F P
0 g
s
.819 .634 .652 r
.32006 .5442 m .31158 .57555 L .36777 .59745 L p F P
0 g
s
.832 .581 .559 r
.36648 .62579 m .36777 .59745 L .31158 .57555 L p F P
0 g
s
.521 .327 .558 r
.6853 .59591 m .68663 .62426 L .74274 .60199 L p F P
0 g
s
.521 .327 .558 r
.74274 .60199 m .74093 .57343 L .6853 .59591 L p F P
0 g
s
.704 .733 .874 r
.40214 .43375 m .37696 .45529 L .41542 .47163 L p F P
0 g
s
.704 .733 .874 r
.41542 .47163 m .43387 .44749 L .40214 .43375 L p F P
0 g
s
.664 .697 .87 r
.63614 .47048 m .61709 .44652 L .58204 .45623 L p F P
0 g
s
.654 .693 .874 r
.61709 .44652 m .63614 .47048 L .67417 .45372 L p F P
0 g
s
.638 .629 .825 r
.6977 .48022 m .67417 .45372 L .63614 .47048 L p F P
0 g
s
.646 .754 .928 r
.48064 .43695 m .46918 .45682 L .50661 .46146 L p F P
0 g
s
.646 .754 .928 r
.50661 .46146 m .51025 .44066 L .48064 .43695 L p F P
0 g
s
.65 .758 .928 r
.51025 .44066 m .50661 .46146 L .54472 .46126 L p F P
0 g
s
.65 .758 .928 r
.54472 .46126 m .54041 .4405 L .51025 .44066 L p F P
0 g
s
.655 .761 .928 r
.54041 .4405 m .54472 .46126 L .58204 .45623 L p F P
0 g
s
.653 .213 .191 r
.4961 .68329 m .49445 .66948 L .43079 .66254 L p F P
0 g
s
.588 .162 .202 r
.49445 .66948 m .4961 .68329 L .5571 .68301 L p F P
0 g
s
.588 .162 .202 r
.5571 .68301 m .55907 .66918 L .49445 .66948 L p F P
0 g
s
.507 .092 .189 r
.55907 .66918 m .5571 .68301 L .61699 .67596 L p F P
0 g
s
.507 .092 .189 r
.61699 .67596 m .62255 .66164 L .55907 .66918 L p F P
0 g
s
.799 .727 .785 r
.31571 .45791 m .29159 .48588 L .33452 .51229 L p F P
0 g
s
.799 .727 .785 r
.33452 .51229 m .35392 .48201 L .31571 .45791 L p F P
0 g
s
.832 .709 .729 r
.32006 .5442 m .33452 .51229 L .29159 .48588 L p F P
0 g
s
.585 .554 .787 r
.6977 .48022 m .71752 .51034 L .75975 .48345 L p F P
0 g
s
.548 .479 .73 r
.71752 .51034 m .73228 .54213 L .77805 .51369 L p F P
0 g
s
.548 .479 .73 r
.77805 .51369 m .75975 .48345 L .71752 .51034 L p F P
0 g
s
.759 .739 .834 r
.34426 .43417 m .31571 .45791 L .35392 .48201 L p F P
0 g
s
.759 .739 .834 r
.35392 .48201 m .37696 .45529 L .34426 .43417 L p F P
0 g
s
.642 .749 .926 r
.45279 .42949 m .43387 .44749 L .46918 .45682 L p F P
0 g
s
.642 .749 .926 r
.46918 .45682 m .48064 .43695 L .45279 .42949 L p F P
0 g
s
.832 .709 .729 r
.29159 .48588 m .27356 .51627 L .32006 .5442 L p F P
0 g
s
.861 .68 .657 r
.31158 .57555 m .32006 .5442 L .27356 .51627 L p F P
0 g
s
.832 .581 .559 r
.31158 .57555 m .3098 .60408 L .36648 .62579 L p F P
0 g
s
.829 .484 .403 r
.37025 .64848 m .36648 .62579 L .3098 .60408 L p F P
0 g
s
.776 .445 .428 r
.36648 .62579 m .37025 .64848 L .43079 .66254 L p F P
0 g
s
.655 .761 .928 r
.58204 .45623 m .56992 .43647 L .54041 .4405 L p F P
0 g
s
.661 .764 .926 r
.56992 .43647 m .58204 .45623 L .61709 .44652 L p F P
0 g
s
.616 .623 .835 r
.67417 .45372 m .6977 .48022 L .73527 .4557 L p F P
0 g
s
.585 .554 .787 r
.75975 .48345 m .73527 .4557 L .6977 .48022 L p F P
0 g
s
.439 .165 .4 r
.68663 .62426 m .68272 .64701 L .73741 .62559 L p F P
0 g
s
.439 .165 .4 r
.73741 .62559 m .74274 .60199 L .68663 .62426 L p F P
0 g
s
.502 .388 .657 r
.73228 .54213 m .74093 .57343 L .78878 .54435 L p F P
0 g
s
.502 .388 .657 r
.78878 .54435 m .77805 .51369 L .73228 .54213 L p F P
0 g
s
.708 .246 .153 r
.43603 .6768 m .43079 .66254 L .37025 .64848 L p F P
0 g
s
.653 .213 .191 r
.43079 .66254 m .43603 .6768 L .4961 .68329 L p F P
0 g
s
.4 0 .148 r
.62255 .66164 m .61699 .67596 L .67369 .66229 L p F P
0 g
s
.4 0 .148 r
.67369 .66229 m .68272 .64701 L .62255 .66164 L p F P
0 g
s
.708 .743 .879 r
.37535 .41608 m .34426 .43417 L .37696 .45529 L p F P
0 g
s
.708 .743 .879 r
.37696 .45529 m .40214 .43375 L .37535 .41608 L p F P
0 g
s
.64 .744 .924 r
.42788 .41855 m .40214 .43375 L .43387 .44749 L p F P
0 g
s
.64 .744 .924 r
.43387 .44749 m .45279 .42949 L .42788 .41855 L p F P
0 g
s
.861 .68 .657 r
.27356 .51627 m .26298 .54698 L .31158 .57555 L p F P
0 g
s
.883 .629 .549 r
.3098 .60408 m .31158 .57555 L .26298 .54698 L p F P
0 g
s
.661 .764 .926 r
.61709 .44652 m .59756 .42872 L .56992 .43647 L p F P
0 g
s
.666 .765 .924 r
.59756 .42872 m .61709 .44652 L .64845 .43244 L p F P
0 g
s
.654 .693 .874 r
.67417 .45372 m .64845 .43244 L .61709 .44652 L p F P
0 g
s
.644 .692 .879 r
.64845 .43244 m .67417 .45372 L .7063 .43224 L p F P
0 g
s
.616 .623 .835 r
.73527 .4557 m .7063 .43224 L .67417 .45372 L p F P
0 g
s
.432 .261 .547 r
.74093 .57343 m .74274 .60199 L .79104 .57316 L p F P
0 g
s
.432 .261 .547 r
.79104 .57316 m .78878 .54435 L .74093 .57343 L p F P
0 g
s
.553 .784 .985 r
.4917 .42401 m .48064 .43695 L .51025 .44066 L p F P
0 g
s
.553 .784 .985 r
.51025 .44066 m .51376 .4268 L .4917 .42401 L p F P
0 g
s
.576 .806 .987 r
.51376 .4268 m .51025 .44066 L .54041 .4405 L p F P
0 g
s
.576 .806 .987 r
.54041 .4405 m .53625 .42668 L .51376 .4268 L p F P
0 g
s
.604 .825 .985 r
.53625 .42668 m .54041 .4405 L .56992 .43647 L p F P
0 g
s
.829 .484 .403 r
.3098 .60408 m .315 .62761 L .37025 .64848 L p F P
0 g
s
.749 .26 .076 r
.37897 .66366 m .37025 .64848 L .315 .62761 L p F P
0 g
s
.708 .246 .153 r
.37025 .64848 m .37897 .66366 L .43603 .6768 L p F P
0 g
s
.344 0 0 r
.4986 .68771 m .4961 .68329 L .43603 .6768 L p F P
0 g
s
.278 0 0 r
.4961 .68329 m .4986 .68771 L .55412 .68745 L p F P
0 g
s
.278 0 0 r
.55412 .68745 m .5571 .68301 L .4961 .68329 L p F P
0 g
s
.168 0 0 r
.5571 .68301 m .55412 .68745 L .60859 .68105 L p F P
0 g
s
.168 0 0 r
.60859 .68105 m .61699 .67596 L .5571 .68301 L p F P
0 g
s
.248 0 .067 r
.68272 .64701 m .67369 .66229 L .72512 .6423 L p F P
0 g
s
.248 0 .067 r
.72512 .6423 m .73741 .62559 L .68272 .64701 L p F P
0 g
s
.638 .739 .921 r
.40699 .40451 m .37535 .41608 L .40214 .43375 L p F P
0 g
s
.638 .739 .921 r
.40214 .43375 m .42788 .41855 L .40699 .40451 L p F P
0 g
s
.538 .761 .979 r
.471 .41842 m .45279 .42949 L .48064 .43695 L p F P
0 g
s
.538 .761 .979 r
.48064 .43695 m .4917 .42401 L .471 .41842 L p F P
0 g
s
.604 .825 .985 r
.56992 .43647 m .55823 .42365 L .53625 .42668 L p F P
0 g
s
.636 .839 .978 r
.55823 .42365 m .56992 .43647 L .59756 .42872 L p F P
0 g
s
.819 .764 .801 r
.28577 .42904 m .25767 .45412 L .29159 .48588 L p F P
0 g
s
.819 .764 .801 r
.29159 .48588 m .31571 .45791 L .28577 .42904 L p F P
0 g
s
.863 .756 .743 r
.27356 .51627 m .29159 .48588 L .25767 .45412 L p F P
0 g
s
.769 .763 .847 r
.31888 .40898 m .28577 .42904 L .31571 .45791 L p F P
0 g
s
.769 .763 .847 r
.31571 .45791 m .34426 .43417 L .31888 .40898 L p F P
0 g
s
.666 .765 .924 r
.64845 .43244 m .62218 .4175 L .59756 .42872 L p F P
0 g
s
.672 .766 .921 r
.62218 .4175 m .64845 .43244 L .67476 .41446 L p F P
0 g
s
.644 .692 .879 r
.7063 .43224 m .67476 .41446 L .64845 .43244 L p F P
0 g
s
.883 .629 .549 r
.26298 .54698 m .26075 .57577 L .3098 .60408 L p F P
0 g
s
.883 .521 .355 r
.315 .62761 m .3098 .60408 L .26075 .57577 L p F P
0 g
s
.593 .62 .848 r
.7063 .43224 m .73527 .4557 L .76443 .42648 L p F P
0 g
s
.547 .544 .803 r
.73527 .4557 m .75975 .48345 L .79281 .45129 L p F P
0 g
s
.547 .544 .803 r
.79281 .45129 m .76443 .42648 L .73527 .4557 L p F P
0 g
s
.492 .456 .745 r
.75975 .48345 m .77805 .51369 L .81409 .47959 L p F P
0 g
s
.492 .456 .745 r
.81409 .47959 m .79281 .45129 L .75975 .48345 L p F P
0 g
s
.303 .047 .35 r
.74274 .60199 m .73741 .62559 L .78442 .59789 L p F P
0 g
s
.303 .047 .35 r
.78442 .59789 m .79104 .57316 L .74274 .60199 L p F P
0 g
s
.365 0 0 r
.44397 .68181 m .43603 .6768 L .37897 .66366 L p F P
0 g
s
.344 0 0 r
.43603 .6768 m .44397 .68181 L .4986 .68771 L p F P
0 g
s
.008 0 0 r
.61699 .67596 m .60859 .68105 L .66006 .66867 L p F P
0 g
s
.008 0 0 r
.66006 .66867 m .67369 .66229 L .61699 .67596 L p F P
0 g
s
.863 .756 .743 r
.25767 .45412 m .23661 .4826 L .27356 .51627 L p F P
0 g
s
.904 .735 .662 r
.26298 .54698 m .27356 .51627 L .23661 .4826 L p F P
0 g
s
.711 .754 .885 r
.35477 .39508 m .31888 .40898 L .34426 .43417 L p F P
0 g
s
.711 .754 .885 r
.34426 .43417 m .37535 .41608 L .35477 .39508 L p F P
0 g
s
.531 .738 .97 r
.45256 .41023 m .42788 .41855 L .45279 .42949 L p F P
0 g
s
.531 .738 .97 r
.45279 .42949 m .471 .41842 L .45256 .41023 L p F P
0 g
s
.636 .839 .978 r
.59756 .42872 m .57877 .41784 L .55823 .42365 L p F P
0 g
s
.666 .846 .969 r
.57877 .41784 m .59756 .42872 L .62218 .4175 L p F P
0 g
s
.635 .692 .886 r
.67476 .41446 m .7063 .43224 L .73099 .40675 L p F P
0 g
s
.593 .62 .848 r
.76443 .42648 m .73099 .40675 L .7063 .43224 L p F P
0 g
s
.418 .341 .662 r
.77805 .51369 m .78878 .54435 L .82661 .50942 L p F P
0 g
s
.418 .341 .662 r
.82661 .50942 m .81409 .47959 L .77805 .51369 L p F P
0 g
s
.904 .735 .662 r
.23661 .4826 m .22421 .51249 L .26298 .54698 L p F P
0 g
s
.938 .685 .528 r
.26075 .57577 m .26298 .54698 L .22421 .51249 L p F P
0 g
s
.638 .734 .918 r
.39111 .38792 m .35477 .39508 L .37535 .41608 L p F P
0 g
s
.638 .734 .918 r
.37535 .41608 m .40699 .40451 L .39111 .38792 L p F P
0 g
s
.532 .716 .957 r
.4372 .39977 m .40699 .40451 L .42788 .41855 L p F P
0 g
s
.532 .716 .957 r
.42788 .41855 m .45256 .41023 L .4372 .39977 L p F P
0 g
s
.672 .766 .921 r
.67476 .41446 m .64267 .40323 L .62218 .4175 L p F P
0 g
s
.676 .765 .918 r
.64267 .40323 m .67476 .41446 L .69475 .39322 L p F P
0 g
s
.635 .692 .886 r
.73099 .40675 m .69475 .39322 L .67476 .41446 L p F P
0 g
s
.666 .846 .969 r
.62218 .4175 m .59698 .40945 L .57877 .41784 L p F P
0 g
s
.694 .847 .956 r
.59698 .40945 m .62218 .4175 L .64267 .40323 L p F P
0 g
s
.303 .168 .525 r
.78878 .54435 m .79104 .57316 L .82924 .53853 L p F P
0 g
s
.303 .168 .525 r
.82924 .53853 m .82661 .50942 L .78878 .54435 L p F P
0 g
s
.749 .26 .076 r
.315 .62761 m .32701 .64417 L .37897 .66366 L p F P
0 g
s
.33 0 0 r
.39216 .66991 m .37897 .66366 L .32701 .64417 L p F P
0 g
s
.365 0 0 r
.37897 .66366 m .39216 .66991 L .44397 .68181 L p F P
0 g
s
.883 .521 .355 r
.26075 .57577 m .26725 .6004 L .315 .62761 L p F P
0 g
s
.758 .243 0 r
.32701 .64417 m .315 .62761 L .26725 .6004 L p F P
0 g
s
0 0 0 r
.67369 .66229 m .66006 .66867 L .70659 .65059 L p F P
0 g
s
0 0 0 r
.70659 .65059 m .72512 .6423 L .67369 .66229 L p F P
0 g
s
.019 0 0 r
.73741 .62559 m .72512 .6423 L .76918 .61649 L p F P
0 g
s
.019 0 0 r
.76918 .61649 m .78442 .59789 L .73741 .62559 L p F P
0 g
s
.34 .754 .979 r
.50167 .41823 m .4917 .42401 L .51376 .4268 L p F P
0 g
s
.34 .754 .979 r
.51376 .4268 m .51692 .42016 L .50167 .41823 L p F P
0 g
s
.386 .8 .978 r
.51692 .42016 m .51376 .4268 L .53625 .42668 L p F P
0 g
s
.119 .669 .723 r
.53249 .42008 m .54768 .41798 L p .54416 .4184 L F P
0 g
s
.386 .8 .978 r
.53249 .42008 m .51692 .42016 L p .5214 .42167 L F P
0 g
s
.003 0 0 r
.51958 .42024 m .51692 .42016 L .53249 .42008 L p F P
0 g
s
.003 0 0 r
.52934 .42019 m .51958 .42024 L p .52222 .42021 L F P
0 g
s
.003 0 0 r
.53249 .42008 m .52934 .42019 L p .52222 .42021 L F P
0 g
s
.119 .669 .723 r
.52934 .42019 m .53249 .42008 L p .54416 .4184 L F P
0 g
s
.386 .8 .978 r
.53625 .42668 m .53249 .42008 L p .5214 .42167 L F P
0 g
s
.457 .848 .979 r
.53249 .42008 m .53625 .42668 L .55823 .42365 L p F P
0 g
s
.457 .848 .979 r
.55823 .42365 m .54768 .41798 L .53249 .42008 L p F P
0 g
s
.119 .669 .723 r
.53886 .41887 m .52934 .42019 L p .53309 .41973 L F P
0 g
s
.119 .669 .723 r
.54768 .41798 m .53886 .41887 L p .53309 .41973 L F P
0 g
s
.286 .785 .78 r
.53886 .41887 m .54768 .41798 L .56185 .41395 L p F P
0 g
s
.286 .785 .78 r
.54772 .41635 m .53886 .41887 L p .54356 .41786 L F P
0 g
s
.286 .785 .78 r
.56185 .41395 m .54772 .41635 L p .54356 .41786 L F P
0 g
s
.475 .896 .848 r
.55552 .41272 m .54772 .41635 L p .55315 .41467 L F P
0 g
s
.475 .896 .848 r
.57436 .40815 m .55552 .41272 L p .55315 .41467 L F P
0 g
s
.468 .856 .652 r
.55552 .41272 m .56189 .40816 L p .55793 .41073 L F P
0 g
s
0 0 0 r
.54772 .41635 m .55552 .41272 L p .55197 .4148 L F P
0 g
s
.475 .896 .848 r
.54772 .41635 m .56185 .41395 L .57436 .40815 L p F P
0 g
s
0 0 0 r
.53698 .42356 m .54772 .41635 L p .55197 .4148 L F P
0 g
s
0 0 0 r
.54122 .42159 m .53698 .42356 L p .5399 .42185 L F P
0 g
s
0 0 0 r
.55552 .41272 m .54122 .42159 L p .5399 .42185 L F P
0 g
s
.468 .856 .652 r
.54122 .42159 m .55552 .41272 L p .55793 .41073 L F P
0 g
s
.468 .856 .652 r
.54468 .41911 m .54122 .42159 L p .54448 .41948 L F P
0 g
s
.468 .856 .652 r
.56189 .40816 m .54468 .41911 L p .54448 .41948 L F P
0 g
s
.074 0 0 r
.53886 .41887 m .54772 .41635 L p .54474 .41799 L F P
0 g
s
.074 0 0 r
.54772 .41635 m .53698 .42356 L .53214 .42493 L p F P
0 g
s
.561 .843 .507 r
.54468 .41911 m .54718 .41624 L p .54488 .41875 L F P
0 g
s
0 0 0 r
.54122 .42159 m .54468 .41911 L p .54254 .42133 L F P
0 g
s
0 0 0 r
.53166 .4326 m .54122 .42159 L p .54254 .42133 L F P
0 g
s
0 0 0 r
.53318 .43151 m .53166 .4326 L p .53297 .43124 L F P
0 g
s
0 0 0 r
.54468 .41911 m .53318 .43151 L p .53297 .43124 L F P
0 g
s
.561 .843 .507 r
.53318 .43151 m .54468 .41911 L p .54488 .41875 L F P
0 g
s
.561 .843 .507 r
.53428 .43025 m .53318 .43151 L p .5346 .42997 L F P
0 g
s
.561 .843 .507 r
.54718 .41624 m .53428 .43025 L p .5346 .42997 L F P
0 g
s
.729 .86 .434 r
.53428 .43025 m .5349 .42887 L p .53397 .43053 L F P
0 g
s
0 0 0 r
.53318 .43151 m .53428 .43025 L p .53338 .43178 L F P
0 g
s
0 0 0 r
.52684 .44297 m .53318 .43151 L p .53338 .43178 L F P
0 g
s
0 0 0 r
.52717 .44258 m .52684 .44297 L p .52719 .44236 L F P
0 g
s
0 0 0 r
.53428 .43025 m .52717 .44258 L p .52719 .44236 L F P
0 g
s
0 0 0 r
.52717 .44258 m .52736 .44216 L p .52715 .4428 L F P
0 g
s
0 0 0 r
.52443 .45112 m .52717 .44258 L p .52715 .4428 L F P
0 g
s
.729 .86 .434 r
.52717 .44258 m .53428 .43025 L p .53397 .43053 L F P
0 g
s
.074 0 0 r
.53214 .42493 m .53886 .41887 L p .54474 .41799 L F P
0 g
s
.252 0 0 r
.52934 .42019 m .53886 .41887 L p .53658 .42017 L F P
0 g
s
.252 0 0 r
.52693 .42565 m .52934 .42019 L p .53658 .42017 L F P
0 g
s
.292 .742 .959 r
.5018 .6821 m .4986 .68771 L .44397 .68181 L p F P
0 g
s
.333 .782 .956 r
.4986 .68771 m .5018 .6821 L .55032 .68187 L p F P
0 g
s
.333 .782 .956 r
.55032 .68187 m .55412 .68745 L .4986 .68771 L p F P
0 g
s
.399 .828 .959 r
.55412 .68745 m .55032 .68187 L .59787 .67627 L p F P
0 g
s
.399 .828 .959 r
.59787 .67627 m .60859 .68105 L .55412 .68745 L p F P
0 g
s
.322 .714 .981 r
.48739 .41435 m .471 .41842 L .4917 .42401 L p F P
0 g
s
.322 .714 .981 r
.4917 .42401 m .50167 .41823 L .48739 .41435 L p F P
0 g
s
.542 .891 .981 r
.54768 .41798 m .55823 .42365 L .57877 .41784 L p F P
0 g
s
.776 .788 .863 r
.30221 .38062 m .26575 .39638 L .28577 .42904 L p F P
0 g
s
.776 .788 .863 r
.28577 .42904 m .31888 .40898 L .30221 .38062 L p F P
0 g
s
.835 .805 .822 r
.25767 .45412 m .28577 .42904 L .26575 .39638 L p F P
0 g
s
.33 .681 .978 r
.47472 .40869 m .45256 .41023 L .471 .41842 L p F P
0 g
s
.33 .681 .978 r
.471 .41842 m .48739 .41435 L .47472 .40869 L p F P
0 g
s
.712 .764 .893 r
.34152 .37156 m .30221 .38062 L .31888 .40898 L p F P
0 g
s
.712 .764 .893 r
.31888 .40898 m .35477 .39508 L .34152 .37156 L p F P
0 g
s
.542 .891 .981 r
.57877 .41784 m .56185 .41395 L .54768 .41798 L p F P
0 g
s
.627 .919 .978 r
.56185 .41395 m .57877 .41784 L .59698 .40945 L p F P
0 g
s
.278 .709 .967 r
.45409 .67693 m .44397 .68181 L .39216 .66991 L p F P
0 g
s
.292 .742 .959 r
.44397 .68181 m .45409 .67693 L .5018 .6821 L p F P
0 g
s
.539 .698 .942 r
.42566 .38746 m .39111 .38792 L .40699 .40451 L p F P
0 g
s
.539 .698 .942 r
.40699 .40451 m .4372 .39977 L .42566 .38746 L p F P
0 g
s
.938 .685 .528 r
.22421 .51249 m .22159 .54158 L .26075 .57577 L p F P
0 g
s
.926 .548 .264 r
.26725 .6004 m .26075 .57577 L .22159 .54158 L p F P
0 g
s
.569 .621 .865 r
.73099 .40675 m .76443 .42648 L .78353 .39356 L p F P
0 g
s
.503 .537 .824 r
.76443 .42648 m .79281 .45129 L .8148 .41492 L p F P
0 g
s
.503 .537 .824 r
.8148 .41492 m .78353 .39356 L .76443 .42648 L p F P
0 g
s
.482 .873 .968 r
.60859 .68105 m .59787 .67627 L .64271 .66542 L p F P
0 g
s
.482 .873 .968 r
.64271 .66542 m .66006 .66867 L .60859 .68105 L p F P
0 g
s
.694 .847 .956 r
.64267 .40323 m .61204 .39882 L .59698 .40945 L p F P
0 g
s
.716 .84 .941 r
.61204 .39882 m .64267 .40323 L .65808 .38645 L p F P
0 g
s
.676 .765 .918 r
.69475 .39322 m .65808 .38645 L .64267 .40323 L p F P
0 g
s
.835 .805 .822 r
.26575 .39638 m .23467 .41804 L .25767 .45412 L p F P
0 g
s
.893 .813 .762 r
.23661 .4826 m .25767 .45412 L .23467 .41804 L p F P
0 g
s
.626 .695 .894 r
.69475 .39322 m .73099 .40675 L .74686 .37817 L p F P
0 g
s
.569 .621 .865 r
.78353 .39356 m .74686 .37817 L .73099 .40675 L p F P
0 g
s
.088 0 .253 r
.79104 .57316 m .78442 .59789 L .82151 .56465 L p F P
0 g
s
.088 0 .253 r
.82151 .56465 m .82924 .53853 L .79104 .57316 L p F P
0 g
s
.42 .43 .764 r
.79281 .45129 m .81409 .47959 L .83833 .44091 L p F P
0 g
s
.42 .43 .764 r
.83833 .44091 m .8148 .41492 L .79281 .45129 L p F P
0 g
s
.639 .73 .914 r
.3811 .36943 m .34152 .37156 L .35477 .39508 L p F P
0 g
s
.639 .73 .914 r
.35477 .39508 m .39111 .38792 L .3811 .36943 L p F P
0 g
s
.357 .653 .968 r
.46423 .40148 m .4372 .39977 L .45256 .41023 L p F P
0 g
s
.357 .653 .968 r
.45256 .41023 m .47472 .40869 L .46423 .40148 L p F P
0 g
s
.68 .763 .914 r
.65808 .38645 m .69475 .39322 L .70734 .36954 L p F P
0 g
s
.626 .695 .894 r
.74686 .37817 m .70734 .36954 L .69475 .39322 L p F P
0 g
s
.627 .919 .978 r
.59698 .40945 m .57436 .40815 L .56185 .41395 L p F P
0 g
s
.699 .927 .967 r
.57436 .40815 m .59698 .40945 L .61204 .39882 L p F P
0 g
s
.893 .813 .762 r
.23467 .41804 m .21128 .44424 L .23661 .4826 L p F P
0 g
s
.949 .803 .665 r
.22421 .51249 m .23661 .4826 L .21128 .44424 L p F P
0 g
s
.758 .243 0 r
.26725 .6004 m .28224 .61882 L .32701 .64417 L p F P
0 g
s
.33 0 0 r
.32701 .64417 m .34513 .65229 L .39216 .66991 L p F P
0 g
s
0 .321 .719 r
.34513 .65229 m .32701 .64417 L .28224 .61882 L p F P
0 g
s
.3 .276 .665 r
.81409 .47959 m .82661 .50942 L .85221 .46971 L p F P
0 g
s
.3 .276 .665 r
.85221 .46971 m .83833 .44091 L .81409 .47959 L p F P
0 g
s
.447 .866 .731 r
.72512 .6423 m .70659 .65059 L .74625 .62732 L p F P
0 g
s
.447 .866 .731 r
.74625 .62732 m .76918 .61649 L .72512 .6423 L p F P
0 g
s
.291 .685 .974 r
.40896 .66651 m .39216 .66991 L .34513 .65229 L p F P
0 g
s
.278 .709 .967 r
.39216 .66991 m .40896 .66651 L .45409 .67693 L p F P
0 g
s
.569 .908 .974 r
.66006 .66867 m .64271 .66542 L .68307 .64965 L p F P
0 g
s
.569 .908 .974 r
.68307 .64965 m .70659 .65059 L .66006 .66867 L p F P
0 g
s
.55 .683 .926 r
.41853 .37381 m .3811 .36943 L .39111 .38792 L p F P
0 g
s
.55 .683 .926 r
.39111 .38792 m .42566 .38746 L .41853 .37381 L p F P
0 g
s
.395 .631 .949 r
.45643 .39302 m .42566 .38746 L .4372 .39977 L p F P
0 g
s
.395 .631 .949 r
.4372 .39977 m .46423 .40148 L .45643 .39302 L p F P
0 g
s
.716 .84 .941 r
.65808 .38645 m .62322 .38638 L .61204 .39882 L p F P
0 g
s
.732 .829 .925 r
.62322 .38638 m .65808 .38645 L .66756 .36784 L p F P
0 g
s
.68 .763 .914 r
.70734 .36954 m .66756 .36784 L .65808 .38645 L p F P
0 g
s
0 .516 .72 r
.51001 .41902 m .50167 .41823 L .51692 .42016 L p F P
0 g
s
0 .516 .72 r
.51692 .42016 m .51958 .42024 L .51001 .41902 L p F P
0 g
s
.699 .927 .967 r
.61204 .39882 m .58463 .40082 L .57436 .40815 L p F P
0 g
s
.751 .917 .947 r
.58463 .40082 m .61204 .39882 L .62322 .38638 L p F P
0 g
s
.926 .548 .264 r
.22159 .54158 m .22922 .56757 L .26725 .6004 L p F P
0 g
s
.681 .167 0 r
.28224 .61882 m .26725 .6004 L .22922 .56757 L p F P
0 g
s
0 .49 .774 r
.50108 .4166 m .48739 .41435 L .50167 .41823 L p F P
0 g
s
0 .49 .774 r
.50167 .41823 m .50748 .41878 L p .50108 .4166 L F P
0 g
s
0 .489 .843 r
.48739 .41435 m .49692 .41591 L p .49318 .41306 L F P
0 g
s
0 .489 .843 r
.49692 .41591 m .50108 .4166 L .49318 .41306 L p F P
0 g
s
.458 0 0 r
.50718 .42177 m .49617 .41492 L p .50108 .4166 L F P
0 g
s
.458 0 0 r
.49617 .41492 m .49318 .41306 L .50108 .4166 L p F P
0 g
s
0 .201 .637 r
.50365 .41933 m .4903 .41087 L p .49318 .41306 L F P
0 g
s
0 .201 .637 r
.4903 .41087 m .48667 .40857 L .49318 .41306 L p F P
0 g
s
0 .49 .774 r
.50748 .41878 m .51001 .41902 L .50108 .4166 L p F P
0 g
s
.577 .788 .979 r
.50547 .66672 m .5018 .6821 L .45409 .67693 L p F P
0 g
s
.59 .801 .98 r
.5018 .6821 m .50547 .66672 L .54596 .66653 L p F P
0 g
s
.59 .801 .98 r
.54596 .66653 m .55032 .68187 L .5018 .6821 L p F P
0 g
s
.606 .812 .979 r
.55032 .68187 m .54596 .66653 L .58561 .66181 L p F P
0 g
s
.606 .812 .979 r
.58561 .66181 m .59787 .67627 L .55032 .68187 L p F P
0 g
s
0 0 0 r
.78442 .59789 m .76918 .61649 L .80377 .58558 L p F P
0 g
s
0 0 0 r
.80377 .58558 m .82151 .56465 L .78442 .59789 L p F P
0 g
s
.949 .803 .665 r
.21128 .44424 m .19748 .47313 L .22421 .51249 L p F P
0 g
s
.987 .746 .48 r
.22159 .54158 m .22421 .51249 L .19748 .47313 L p F P
0 g
s
0 .489 .843 r
.49318 .41306 m .47472 .40869 L .48739 .41435 L p F P
0 g
s
.099 .023 .474 r
.82661 .50942 m .82924 .53853 L .85513 .49916 L p F P
0 g
s
.099 .023 .474 r
.85513 .49916 m .85221 .46971 L .82661 .50942 L p F P
0 g
s
.568 .775 .976 r
.46569 .66237 m .45409 .67693 L .40896 .66651 L p F P
0 g
s
.577 .788 .979 r
.45409 .67693 m .46569 .66237 L .50547 .66672 L p F P
0 g
s
.71 .773 .901 r
.33647 .34649 m .2954 .35021 L .30221 .38062 L p F P
0 g
s
.71 .773 .901 r
.30221 .38062 m .34152 .37156 L .33647 .34649 L p F P
0 g
s
.777 .814 .882 r
.26575 .39638 m .30221 .38062 L .2954 .35021 L p F P
0 g
s
.624 .82 .976 r
.59787 .67627 m .58561 .66181 L .62289 .65269 L p F P
0 g
s
.624 .82 .976 r
.62289 .65269 m .64271 .66542 L .59787 .67627 L p F P
0 g
s
.641 .726 .91 r
.37758 .34983 m .33647 .34649 L .34152 .37156 L p F P
0 g
s
.641 .726 .91 r
.34152 .37156 m .3811 .36943 L .37758 .34983 L p F P
0 g
s
.065 .498 .894 r
.48667 .40857 m .46423 .40148 L .47472 .40869 L p F P
0 g
s
.065 .498 .894 r
.47472 .40869 m .49318 .41306 L .48667 .40857 L p F P
0 g
s
.436 .615 .925 r
.45169 .38369 m .41853 .37381 L .42566 .38746 L p F P
0 g
s
.436 .615 .925 r
.42566 .38746 m .45643 .39302 L .45169 .38369 L p F P
0 g
s
.619 .699 .902 r
.70734 .36954 m .74686 .37817 L .75278 .34765 L p F P
0 g
s
.545 .627 .884 r
.74686 .37817 m .78353 .39356 L .79116 .35822 L p F P
0 g
s
.545 .627 .884 r
.79116 .35822 m .75278 .34765 L .74686 .37817 L p F P
0 g
s
.646 .966 .897 r
.55552 .41272 m .57436 .40815 L .58463 .40082 L p F P
0 g
s
.777 .814 .882 r
.2954 .35021 m .2571 .36119 L .26575 .39638 L p F P
0 g
s
.845 .85 .847 r
.23467 .41804 m .26575 .39638 L .2571 .36119 L p F P
0 g
s
.751 .917 .947 r
.62322 .38638 m .59219 .39228 L .58463 .40082 L p F P
0 g
s
.782 .893 .923 r
.59219 .39228 m .62322 .38638 L .62996 .37264 L p F P
0 g
s
.732 .829 .925 r
.66756 .36784 m .62996 .37264 L .62322 .38638 L p F P
0 g
s
.683 .76 .91 r
.66756 .36784 m .70734 .36954 L .71166 .34439 L p F P
0 g
s
.619 .699 .902 r
.75278 .34765 m .71166 .34439 L .70734 .36954 L p F P
0 g
s
.453 .534 .85 r
.78353 .39356 m .8148 .41492 L .82404 .37569 L p F P
0 g
s
.453 .534 .85 r
.82404 .37569 m .79116 .35822 L .78353 .39356 L p F P
0 g
s
0 .321 .719 r
.28224 .61882 m .30481 .62942 L .34513 .65229 L p F P
0 g
s
.322 .665 .974 r
.36814 .65112 m .34513 .65229 L .30481 .62942 L p F P
0 g
s
.291 .685 .974 r
.34513 .65229 m .36814 .65112 L .40896 .66651 L p F P
0 g
s
.565 .671 .91 r
.41622 .35941 m .37758 .34983 L .3811 .36943 L p F P
0 g
s
.565 .671 .91 r
.3811 .36943 m .41853 .37381 L .41622 .35941 L p F P
0 g
s
.741 .813 .909 r
.62996 .37264 m .66756 .36784 L .67051 .34819 L p F P
0 g
s
.683 .76 .91 r
.71166 .34439 m .67051 .34819 L .66756 .36784 L p F P
0 g
s
.648 .926 .974 r
.70659 .65059 m .68307 .64965 L .71725 .6294 L p F P
0 g
s
.648 .926 .974 r
.71725 .6294 m .74625 .62732 L .70659 .65059 L p F P
0 g
s
.176 .507 .911 r
.48187 .40331 m .45643 .39302 L .46423 .40148 L p F P
0 g
s
.176 .507 .911 r
.46423 .40148 m .48667 .40857 L .48187 .40331 L p F P
0 g
s
.987 .746 .48 r
.19748 .47313 m .19457 .50255 L .22159 .54158 L p F P
0 g
s
.905 .531 .084 r
.22922 .56757 m .22159 .54158 L .19457 .50255 L p F P
0 g
s
.646 .966 .897 r
.58463 .40082 m .56189 .40816 L .55552 .41272 L p F P
0 g
s
.765 .981 .911 r
.56189 .40816 m .58463 .40082 L .59219 .39228 L p F P
0 g
s
.845 .85 .847 r
.2571 .36119 m .22428 .37899 L .23467 .41804 L p F P
0 g
s
.915 .878 .784 r
.21128 .44424 m .23467 .41804 L .22428 .37899 L p F P
0 g
s
.681 .167 0 r
.22922 .56757 m .24676 .5883 L .28224 .61882 L p F P
0 g
s
0 .386 .828 r
.30481 .62942 m .28224 .61882 L .24676 .5883 L p F P
0 g
s
.563 .761 .971 r
.42815 .6536 m .40896 .66651 L .36814 .65112 L p F P
0 g
s
.568 .775 .976 r
.40896 .66651 m .42815 .6536 L .46569 .66237 L p F P
0 g
s
0 0 .062 r
.82924 .53853 m .82151 .56465 L .84656 .52688 L p F P
0 g
s
0 0 .062 r
.84656 .52688 m .85513 .49916 L .82924 .53853 L p F P
0 g
s
.642 .825 .97 r
.64271 .66542 m .62289 .65269 L .65629 .63946 L p F P
0 g
s
.642 .825 .97 r
.65629 .63946 m .68307 .64965 L .64271 .66542 L p F P
0 g
s
.323 .4 .786 r
.8148 .41492 m .83833 .44091 L .84889 .39904 L p F P
0 g
s
.323 .4 .786 r
.84889 .39904 m .82404 .37569 L .8148 .41492 L p F P
0 g
s
.661 .972 .835 r
.76918 .61649 m .74625 .62732 L .77715 .59954 L p F P
0 g
s
.661 .972 .835 r
.77715 .59954 m .80377 .58558 L .76918 .61649 L p F P
0 g
s
.477 .604 .9 r
.45029 .3739 m .41622 .35941 L .41853 .37381 L p F P
0 g
s
.477 .604 .9 r
.41853 .37381 m .45169 .38369 L .45029 .3739 L p F P
0 g
s
.452 0 0 r
.51637 .42502 m .51001 .41902 L .51958 .42024 L p F P
0 g
s
.452 0 0 r
.51958 .42024 m .52159 .42568 L .51637 .42502 L p F P
0 g
s
.374 0 0 r
.52159 .42568 m .51958 .42024 L .52934 .42019 L p F P
0 g
s
.374 0 0 r
.52934 .42019 m .52693 .42565 L .52159 .42568 L p F P
0 g
s
.782 .893 .923 r
.62996 .37264 m .59665 .3829 L .59219 .39228 L p F P
0 g
s
.797 .862 .897 r
.59665 .3829 m .62996 .37264 L .63184 .35821 L p F P
0 g
s
.741 .813 .909 r
.67051 .34819 m .63184 .35821 L .62996 .37264 L p F P
0 g
s
.282 .513 .9 r
.479 .39754 m .45169 .38369 L .45643 .39302 L p F P
0 g
s
.282 .513 .9 r
.45643 .39302 m .48187 .40331 L .479 .39754 L p F P
0 g
s
.486 0 0 r
.51149 .4237 m .50108 .4166 L .51001 .41902 L p F P
0 g
s
.486 0 0 r
.51001 .41902 m .51637 .42502 L .51149 .4237 L p F P
0 g
s
.252 0 0 r
.53886 .41887 m .53214 .42493 L .52693 .42565 L p F P
0 g
s
.765 .981 .911 r
.59219 .39228 m .56654 .40285 L .56189 .40816 L p F P
0 g
s
.828 .953 .898 r
.56654 .40285 m .59219 .39228 L .59665 .3829 L p F P
0 g
s
.458 0 0 r
.50108 .4166 m .51149 .4237 L .50718 .42177 L p F P
0 g
s
.665 .749 .912 r
.50934 .6428 m .50547 .66672 L .46569 .66237 L p F P
0 g
s
.661 .744 .912 r
.50547 .66672 m .50934 .6428 L .54137 .64265 L p F P
0 g
s
.661 .744 .912 r
.54137 .64265 m .54596 .66653 L .50547 .66672 L p F P
0 g
s
.656 .741 .912 r
.54596 .66653 m .54137 .64265 L .57271 .63886 L p F P
0 g
s
.656 .741 .912 r
.57271 .63886 m .58561 .66181 L .54596 .66653 L p F P
0 g
s
.915 .878 .784 r
.22428 .37899 m .19949 .40257 L .21128 .44424 L p F P
0 g
s
.981 .881 .656 r
.19748 .47313 m .21128 .44424 L .19949 .40257 L p F P
0 g
s
0 .201 .637 r
.50408 .41984 m .50718 .42177 L .50365 .41933 L p F P
0 g
s
.681 .143 0 r
.515 .43161 m .50365 .41933 L .50718 .42177 L p F P
0 g
s
0 0 .486 r
.51386 .43036 m .50333 .41895 L p .50365 .41933 L F P
0 g
s
0 0 .486 r
.50333 .41895 m .50107 .41649 L .50365 .41933 L p F P
0 g
s
0 .201 .637 r
.49318 .41306 m .50408 .41984 L p .50365 .41933 L F P
0 g
s
.117 .181 .654 r
.83833 .44091 m .85221 .46971 L .86359 .42666 L p F P
0 g
s
.117 .181 .654 r
.86359 .42666 m .84889 .39904 L .83833 .44091 L p F P
0 g
s
.669 .754 .914 r
.4779 .63931 m .46569 .66237 L .42815 .6536 L p F P
0 g
s
.665 .749 .912 r
.46569 .66237 m .4779 .63931 L .50934 .6428 L p F P
0 g
s
.644 .723 .906 r
.38091 .33001 m .3402 .32099 L .33647 .34649 L p F P
0 g
s
.644 .723 .906 r
.33647 .34649 m .37758 .34983 L .38091 .33001 L p F P
0 g
s
.706 .781 .91 r
.2954 .35021 m .33647 .34649 L .3402 .32099 L p F P
0 g
s
.581 .663 .895 r
.41898 .34494 m .38091 .33001 L .37758 .34983 L p F P
0 g
s
.581 .663 .895 r
.37758 .34983 m .41622 .35941 L .41898 .34494 L p F P
0 g
s
.65 .738 .914 r
.58561 .66181 m .57271 .63886 L .60208 .63156 L p F P
0 g
s
.65 .738 .914 r
.60208 .63156 m .62289 .65269 L .58561 .66181 L p F P
0 g
s
.371 .519 .875 r
.47822 .3915 m .45029 .3739 L .45169 .38369 L p F P
0 g
s
.371 .519 .875 r
.45169 .38369 m .479 .39754 L .47822 .3915 L p F P
0 g
s
.905 .531 .084 r
.19457 .50255 m .20306 .53014 L .22922 .56757 L p F P
0 g
s
0 0 .537 r
.24676 .5883 m .22922 .56757 L .20306 .53014 L p F P
0 g
s
.745 .796 .893 r
.63184 .35821 m .67051 .34819 L .66658 .3284 L p F P
0 g
s
.684 .756 .906 r
.67051 .34819 m .71166 .34439 L .70717 .31891 L p F P
0 g
s
.684 .756 .906 r
.70717 .31891 m .66658 .3284 L .67051 .34819 L p F P
0 g
s
.828 .953 .898 r
.59665 .3829 m .56923 .39704 L .56654 .40285 L p F P
0 g
s
.851 .906 .872 r
.56923 .39704 m .59665 .3829 L .59776 .37309 L p F P
0 g
s
.797 .862 .897 r
.63184 .35821 m .59776 .37309 L .59665 .3829 L p F P
0 g
s
.613 .706 .911 r
.71166 .34439 m .75278 .34765 L .74798 .31654 L p F P
0 g
s
.613 .706 .911 r
.74798 .31654 m .70717 .31891 L .71166 .34439 L p F P
0 g
s
0 .287 .773 r
.50107 .41649 m .48187 .40331 L .48667 .40857 L p F P
0 g
s
0 .287 .773 r
.48667 .40857 m .50365 .41933 L .50107 .41649 L p F P
0 g
s
.72 .982 .782 r
.54468 .41911 m .56189 .40816 L .56654 .40285 L p F P
0 g
s
.322 .665 .974 r
.30481 .62942 m .33337 .63122 L .36814 .65112 L p F P
0 g
s
.563 .748 .964 r
.39436 .6407 m .36814 .65112 L .33337 .63122 L p F P
0 g
s
.563 .761 .971 r
.36814 .65112 m .39436 .6407 L .42815 .6536 L p F P
0 g
s
.515 .599 .875 r
.45235 .3641 m .41898 .34494 L .41622 .35941 L p F P
0 g
s
.515 .599 .875 r
.41622 .35941 m .45029 .3739 L .45235 .3641 L p F P
0 g
s
.706 .781 .91 r
.3402 .32099 m .29928 .3191 L .2954 .35021 L p F P
0 g
s
.772 .839 .903 r
.2571 .36119 m .2954 .35021 L .29928 .3191 L p F P
0 g
s
.686 .904 .556 r
.82151 .56465 m .80377 .58558 L .82691 .55057 L p F P
0 g
s
.686 .904 .556 r
.82691 .55057 m .84656 .52688 L .82151 .56465 L p F P
0 g
s
.8 .828 .873 r
.59776 .37309 m .63184 .35821 L .62865 .34376 L p F P
0 g
s
.745 .796 .893 r
.66658 .3284 m .62865 .34376 L .63184 .35821 L p F P
0 g
s
.66 .826 .963 r
.68307 .64965 m .65629 .63946 L .68436 .62255 L p F P
0 g
s
.66 .826 .963 r
.68436 .62255 m .71725 .6294 L .68307 .64965 L p F P
0 g
s
.521 .637 .905 r
.75278 .34765 m .79116 .35822 L .78628 .32199 L p F P
0 g
s
.521 .637 .905 r
.78628 .32199 m .74798 .31654 L .75278 .34765 L p F P
0 g
s
0 .386 .828 r
.24676 .5883 m .27309 .60199 L .30481 .62942 L p F P
0 g
s
.365 .649 .965 r
.33337 .63122 m .30481 .62942 L .27309 .60199 L p F P
0 g
s
.069 .358 .829 r
.49954 .41337 m .479 .39754 L .48187 .40331 L p F P
0 g
s
.069 .358 .829 r
.48187 .40331 m .50107 .41649 L .49954 .41337 L p F P
0 g
s
.72 .982 .782 r
.56654 .40285 m .54718 .41624 L .54468 .41911 L p F P
0 g
s
.857 .993 .83 r
.54718 .41624 m .56654 .40285 L .56923 .39704 L p F P
0 g
s
.709 .926 .964 r
.74625 .62732 m .71725 .6294 L .74362 .60533 L p F P
0 g
s
.709 .926 .964 r
.74362 .60533 m .77715 .59954 L .74625 .62732 L p F P
0 g
s
.981 .881 .656 r
.19949 .40257 m .18482 .43029 L .19748 .47313 L p F P
0 g
s
.978 .78 .363 r
.19457 .50255 m .19748 .47313 L .18482 .43029 L p F P
0 g
s
.672 .759 .916 r
.44831 .63229 m .42815 .6536 L .39436 .6407 L p F P
0 g
s
.669 .754 .914 r
.42815 .6536 m .44831 .63229 L .4779 .63931 L p F P
0 g
s
.442 .525 .847 r
.47959 .38548 m .45235 .3641 L .45029 .3739 L p F P
0 g
s
.442 .525 .847 r
.45029 .3739 m .47822 .3915 L .47959 .38548 L p F P
0 g
s
.645 .737 .916 r
.62289 .65269 m .60208 .63156 L .62827 .62099 L p F P
0 g
s
.645 .737 .916 r
.62827 .62099 m .65629 .63946 L .62289 .65269 L p F P
0 g
s
.851 .906 .872 r
.59776 .37309 m .56983 .391 L .56923 .39704 L p F P
0 g
s
.85 .854 .845 r
.56983 .391 m .59776 .37309 L .5954 .36331 L p F P
0 g
s
.8 .828 .873 r
.62865 .34376 m .5954 .36331 L .59776 .37309 L p F P
0 g
s
.772 .839 .903 r
.29928 .3191 m .26089 .32497 L .2571 .36119 L p F P
0 g
s
.843 .897 .877 r
.22428 .37899 m .2571 .36119 L .26089 .32497 L p F P
0 g
s
0 0 .346 r
.85221 .46971 m .85513 .49916 L .86669 .45644 L p F P
0 g
s
0 0 .346 r
.86669 .45644 m .86359 .42666 L .85221 .46971 L p F P
0 g
s
.628 .151 .102 r
.52063 .43412 m .51637 .42502 L .52159 .42568 L p F P
0 g
s
.628 .151 .102 r
.52159 .42568 m .52295 .43441 L .52063 .43412 L p F P
0 g
s
.552 .093 .117 r
.52295 .43441 m .52159 .42568 L .52693 .42565 L p F P
0 g
s
.552 .093 .117 r
.52693 .42565 m .52532 .4344 L .52295 .43441 L p F P
0 g
s
.453 .007 .099 r
.52532 .4344 m .52693 .42565 L .53214 .42493 L p F P
0 g
s
.237 .406 .828 r
.49915 .41012 m .47822 .3915 L .479 .39754 L p F P
0 g
s
.237 .406 .828 r
.479 .39754 m .49954 .41337 L .49915 .41012 L p F P
0 g
s
.857 .993 .83 r
.56923 .39704 m .54861 .41311 L .54718 .41624 L p F P
0 g
s
.901 .941 .827 r
.54861 .41311 m .56923 .39704 L .56983 .391 L p F P
0 g
s
.685 .184 .047 r
.51846 .43353 m .51149 .4237 L .51637 .42502 L p F P
0 g
s
.685 .184 .047 r
.51637 .42502 m .52063 .43412 L .51846 .43353 L p F P
0 g
s
.453 .007 .099 r
.53214 .42493 m .52763 .43408 L .52532 .4344 L p F P
0 g
s
.315 0 .04 r
.52763 .43408 m .53214 .42493 L .53698 .42356 L p F P
0 g
s
.393 .535 .879 r
.79116 .35822 m .82404 .37569 L .81927 .33529 L p F P
0 g
s
.393 .535 .879 r
.81927 .33529 m .78628 .32199 L .79116 .35822 L p F P
0 g
s
.715 .188 0 r
.51656 .43269 m .50718 .42177 L .51149 .4237 L p F P
0 g
s
.715 .188 0 r
.51149 .4237 m .51846 .43353 L .51656 .43269 L p F P
0 g
s
.315 0 .04 r
.53698 .42356 m .52978 .43347 L .52763 .43408 L p F P
0 g
s
.113 0 0 r
.52978 .43347 m .53698 .42356 L .54122 .42159 L p F P
0 g
s
.681 .143 0 r
.50718 .42177 m .51656 .43269 L .515 .43161 L p F P
0 g
s
.113 0 0 r
.54122 .42159 m .53166 .4326 L .52978 .43347 L p F P
0 g
s
.55 .598 .854 r
.45782 .35476 m .42677 .33108 L .41898 .34494 L p F P
0 g
s
.55 .598 .854 r
.41898 .34494 m .45235 .3641 L .45782 .35476 L p F P
0 g
s
.598 .659 .881 r
.38091 .33001 m .41898 .34494 L .42677 .33108 L p F P
0 g
s
.702 .71 .855 r
.51312 .61251 m .50934 .6428 L .4779 .63931 L p F P
0 g
s
.687 .697 .854 r
.50934 .6428 m .51312 .61251 L .5369 .61239 L p F P
0 g
s
.687 .697 .854 r
.5369 .61239 m .54137 .64265 L .50934 .6428 L p F P
0 g
s
.67 .685 .855 r
.54137 .64265 m .5369 .61239 L .56013 .60953 L p F P
0 g
s
.67 .685 .855 r
.56013 .60953 m .57271 .63886 L .54137 .64265 L p F P
0 g
s
.794 .794 .853 r
.5954 .36331 m .62865 .34376 L .62044 .32998 L p F P
0 g
s
.744 .777 .88 r
.62865 .34376 m .66658 .3284 L .65575 .30941 L p F P
0 g
s
.744 .777 .88 r
.65575 .30941 m .62044 .32998 L .62865 .34376 L p F P
0 g
s
.598 .659 .881 r
.42677 .33108 m .39117 .31092 L .38091 .33001 L p F P
0 g
s
.647 .721 .903 r
.3402 .32099 m .38091 .33001 L .39117 .31092 L p F P
0 g
s
.359 .439 .808 r
.49993 .40689 m .47959 .38548 L .47822 .3915 L p F P
0 g
s
.359 .439 .808 r
.47822 .3915 m .49915 .41012 L .49993 .40689 L p F P
0 g
s
0 0 .486 r
.51351 .43 m .515 .43161 L .51386 .43036 L p F P
0 g
s
.791 .332 0 r
.52091 .44261 m .51386 .43036 L .515 .43161 L p F P
0 g
s
.791 .332 0 r
.515 .43161 m .5209 .44235 L p .52091 .44261 L F P
0 g
s
.791 .332 0 r
.5209 .44235 m .52126 .443 L .52091 .44261 L p F P
0 g
s
.803 .453 0 r
.5236 .45107 m .52091 .44283 L p .52091 .44261 L F P
0 g
s
.803 .453 0 r
.52091 .44283 m .52071 .4422 L .52091 .44261 L p F P
0 g
s
0 0 .411 r
.52071 .4422 m .51413 .43064 L p .51386 .43036 L F P
0 g
s
0 0 .411 r
.51413 .43064 m .51319 .42899 L .51386 .43036 L p F P
0 g
s
0 0 .486 r
.50365 .41933 m .51351 .43 L p .51386 .43036 L F P
0 g
s
.901 .941 .827 r
.56983 .391 m .5489 .40985 L .54861 .41311 L p F P
0 g
s
.897 .874 .806 r
.5489 .40985 m .56983 .391 L .56828 .38499 L p F P
0 g
s
.85 .854 .845 r
.5954 .36331 m .56828 .38499 L .56983 .391 L p F P
0 g
s
0 0 .537 r
.20306 .53014 m .22257 .55358 L .24676 .5883 L p F P
0 g
s
.107 .438 .876 r
.27309 .60199 m .24676 .5883 L .22257 .55358 L p F P
0 g
s
.501 .534 .823 r
.48306 .37977 m .45782 .35476 L .45235 .3641 L p F P
0 g
s
.501 .534 .823 r
.45235 .3641 m .47959 .38548 L .48306 .37977 L p F P
0 g
s
.685 .752 .902 r
.66658 .3284 m .70717 .31891 L .69374 .29433 L p F P
0 g
s
.685 .752 .902 r
.69374 .29433 m .65575 .30941 L .66658 .3284 L p F P
0 g
s
.836 .806 .821 r
.56828 .38499 m .5954 .36331 L .58965 .35403 L p F P
0 g
s
.794 .794 .853 r
.62044 .32998 m .58965 .35403 L .5954 .36331 L p F P
0 g
s
0 .127 .675 r
.51319 .42899 m .49954 .41337 L .50107 .41649 L p F P
0 g
s
0 .127 .675 r
.50107 .41649 m .51386 .43036 L .51319 .42899 L p F P
0 g
s
.715 .726 .859 r
.4898 .60986 m .4779 .63931 L .44831 .63229 L p F P
0 g
s
.702 .71 .855 r
.4779 .63931 m .4898 .60986 L .51312 .61251 L p F P
0 g
s
.839 .98 .685 r
.53428 .43025 m .54718 .41624 L .54861 .41311 L p F P
0 g
s
.652 .676 .859 r
.57271 .63886 m .56013 .60953 L .58186 .60401 L p F P
0 g
s
.652 .676 .859 r
.58186 .60401 m .60208 .63156 L .57271 .63886 L p F P
0 g
s
.796 .992 .878 r
.80377 .58558 m .77715 .59954 L .79753 .5682 L p F P
0 g
s
.796 .992 .878 r
.79753 .5682 m .82691 .55057 L .80377 .58558 L p F P
0 g
s
.843 .897 .877 r
.26089 .32497 m .22783 .33862 L .22428 .37899 L p F P
0 g
s
.916 .948 .804 r
.19949 .40257 m .22428 .37899 L .22783 .33862 L p F P
0 g
s
.647 .721 .903 r
.39117 .31092 m .35288 .29629 L .3402 .32099 L p F P
0 g
s
.699 .787 .919 r
.29928 .3191 m .3402 .32099 L .35288 .29629 L p F P
0 g
s
.563 .748 .964 r
.33337 .63122 m .36578 .62407 L .39436 .6407 L p F P
0 g
s
.674 .764 .919 r
.42181 .62197 m .39436 .6407 L .36578 .62407 L p F P
0 g
s
.672 .759 .916 r
.39436 .6407 m .42181 .62197 L .44831 .63229 L p F P
0 g
s
.447 .465 .785 r
.50185 .40383 m .48306 .37977 L .47959 .38548 L p F P
0 g
s
.447 .465 .785 r
.47959 .38548 m .49993 .40689 L .50185 .40383 L p F P
0 g
s
.978 .78 .363 r
.18482 .43029 m .18172 .46005 L .19457 .50255 L p F P
0 g
s
0 0 .212 r
.20306 .53014 m .19457 .50255 L .18172 .46005 L p F P
0 g
s
.063 .252 .745 r
.51304 .42757 m .49915 .41012 L .49954 .41337 L p F P
0 g
s
.063 .252 .745 r
.49954 .41337 m .51319 .42899 L .51304 .42757 L p F P
0 g
s
.897 .874 .806 r
.56828 .38499 m .54802 .40663 L .5489 .40985 L p F P
0 g
s
.875 .812 .783 r
.54802 .40663 m .56828 .38499 L .56463 .37932 L p F P
0 g
s
.836 .806 .821 r
.58965 .35403 m .56463 .37932 L .56828 .38499 L p F P
0 g
s
.839 .98 .685 r
.54861 .41311 m .5349 .42887 L .53428 .43025 L p F P
0 g
s
.939 .959 .747 r
.5349 .42887 m .54861 .41311 L .5489 .40985 L p F P
0 g
s
.639 .736 .919 r
.65629 .63946 m .62827 .62099 L .65011 .60753 L p F P
0 g
s
.639 .736 .919 r
.65011 .60753 m .68436 .62255 L .65629 .63946 L p F P
0 g
s
.189 .361 .805 r
.82404 .37569 m .84889 .39904 L .8443 .35576 L p F P
0 g
s
.189 .361 .805 r
.8443 .35576 m .81927 .33529 L .82404 .37569 L p F P
0 g
s
.608 .714 .92 r
.70717 .31891 m .74798 .31654 L .73215 .28636 L p F P
0 g
s
.608 .714 .92 r
.73215 .28636 m .69374 .29433 L .70717 .31891 L p F P
0 g
s
.365 .649 .965 r
.27309 .60199 m .30627 .60744 L .33337 .63122 L p F P
0 g
s
.567 .736 .955 r
.36578 .62407 m .33337 .63122 L .30627 .60744 L p F P
0 g
s
.675 .823 .955 r
.71725 .6294 m .68436 .62255 L .70578 .60255 L p F P
0 g
s
.675 .823 .955 r
.70578 .60255 m .74362 .60533 L .71725 .6294 L p F P
0 g
s
.258 .333 .751 r
.51339 .42615 m .49993 .40689 L .49915 .41012 L p F P
0 g
s
.258 .333 .751 r
.49915 .41012 m .51304 .42757 L .51339 .42615 L p F P
0 g
s
.717 .749 .24 r
.85513 .49916 m .84656 .52688 L .85761 .48597 L p F P
0 g
s
.717 .749 .24 r
.85761 .48597 m .86669 .45644 L .85513 .49916 L p F P
0 g
s
.939 .959 .747 r
.5489 .40985 m .53502 .42745 L .5349 .42887 L p F P
0 g
s
.941 .886 .75 r
.53502 .42745 m .5489 .40985 L .54802 .40663 L p F P
0 g
s
.549 .546 .803 r
.48849 .37463 m .46649 .34634 L .45782 .35476 L p F P
0 g
s
.549 .546 .803 r
.45782 .35476 m .48306 .37977 L .48849 .37463 L p F P
0 g
s
.582 .601 .837 r
.42677 .33108 m .45782 .35476 L .46649 .34634 L p F P
0 g
s
.817 .762 .801 r
.56463 .37932 m .58965 .35403 L .58072 .3457 L p F P
0 g
s
.782 .763 .835 r
.58965 .35403 m .62044 .32998 L .6075 .31756 L p F P
0 g
s
.782 .763 .835 r
.6075 .31756 m .58072 .3457 L .58965 .35403 L p F P
0 g
s
.727 .744 .865 r
.46791 .60456 m .44831 .63229 L .42181 .62197 L p F P
0 g
s
.715 .726 .859 r
.44831 .63229 m .46791 .60456 L .4898 .60986 L p F P
0 g
s
.515 .489 .766 r
.50481 .40109 m .48849 .37463 L .48306 .37977 L p F P
0 g
s
.515 .489 .766 r
.48306 .37977 m .50185 .40383 L .50481 .40109 L p F P
0 g
s
.696 .304 .3 r
.52298 .44376 m .52063 .43412 L .52295 .43441 L p F P
0 g
s
.696 .304 .3 r
.52295 .43441 m .5237 .44385 L .52298 .44376 L p F P
0 g
s
.627 .25 .309 r
.5237 .44385 m .52295 .43441 L .52532 .4344 L p F P
0 g
s
.627 .25 .309 r
.52532 .4344 m .52442 .44385 L .5237 .44385 L p F P
0 g
s
.545 .18 .299 r
.52442 .44385 m .52532 .4344 L .52763 .43408 L p F P
0 g
s
.582 .601 .837 r
.46649 .34634 m .43932 .31852 L .42677 .33108 L p F P
0 g
s
.616 .659 .869 r
.39117 .31092 m .42677 .33108 L .43932 .31852 L p F P
0 g
s
.634 .669 .866 r
.60208 .63156 m .58186 .60401 L .60113 .59606 L p F P
0 g
s
.634 .669 .866 r
.60113 .59606 m .62827 .62099 L .60208 .63156 L p F P
0 g
s
.758 .346 .27 r
.52232 .44358 m .51846 .43353 L .52063 .43412 L p F P
0 g
s
.758 .346 .27 r
.52063 .43412 m .52298 .44376 L .52232 .44358 L p F P
0 g
s
.875 .812 .783 r
.56463 .37932 m .54601 .40359 L .54802 .40663 L p F P
0 g
s
.847 .758 .765 r
.54601 .40359 m .56463 .37932 L .55905 .37425 L p F P
0 g
s
.817 .762 .801 r
.58072 .3457 m .55905 .37425 L .56463 .37932 L p F P
0 g
s
.545 .18 .299 r
.52763 .43408 m .52513 .44375 L .52442 .44385 L p F P
0 g
s
.439 .085 .266 r
.52513 .44375 m .52763 .43408 L .52978 .43347 L p F P
0 g
s
.813 .376 .206 r
.52173 .44332 m .51656 .43269 L .51846 .43353 L p F P
0 g
s
.813 .376 .206 r
.51846 .43353 m .52232 .44358 L .52173 .44332 L p F P
0 g
s
.439 .085 .266 r
.52978 .43347 m .52579 .44356 L .52513 .44375 L p F P
0 g
s
.288 0 .199 r
.52579 .44356 m .52978 .43347 L .53166 .4326 L p F P
0 g
s
.387 .388 .738 r
.51424 .42482 m .50185 .40383 L .49993 .40689 L p F P
0 g
s
.387 .388 .738 r
.49993 .40689 m .51339 .42615 L .51424 .42482 L p F P
0 g
s
.941 .886 .75 r
.54802 .40663 m .53462 .42604 L .53502 .42745 L p F P
0 g
s
.911 .812 .737 r
.53462 .42604 m .54802 .40663 L .54601 .40359 L p F P
0 g
s
.74 .758 .868 r
.62044 .32998 m .65575 .30941 L .63834 .29222 L p F P
0 g
s
.74 .758 .868 r
.63834 .29222 m .6075 .31756 L .62044 .32998 L p F P
0 g
s
.844 .383 .086 r
.52126 .443 m .515 .43161 L .51656 .43269 L p F P
0 g
s
.844 .383 .086 r
.51656 .43269 m .52173 .44332 L .52126 .443 L p F P
0 g
s
.288 0 .199 r
.53166 .4326 m .52637 .4433 L .52579 .44356 L p F P
0 g
s
.052 0 .072 r
.52637 .4433 m .53166 .4326 L .53318 .43151 L p F P
0 g
s
.699 .787 .919 r
.35288 .29629 m .31417 .28878 L .29928 .3191 L p F P
0 g
s
.757 .86 .925 r
.26089 .32497 m .29928 .3191 L .31417 .28878 L p F P
0 g
s
.052 0 .072 r
.53318 .43151 m .52684 .44297 L .52637 .4433 L p F P
0 g
s
.477 .428 .724 r
.51556 .42363 m .50481 .40109 L .50185 .40383 L p F P
0 g
s
.477 .428 .724 r
.50185 .40383 m .51424 .42482 L .51556 .42363 L p F P
0 g
s
.803 .453 0 r
.52091 .44261 m .52363 .45112 L .5236 .45107 L p F P
0 g
s
0 0 .383 r
.5236 .45101 m .52087 .44243 L p .52071 .4422 L F P
0 g
s
0 0 .383 r
.52087 .44243 m .52066 .44176 L .52071 .4422 L p F P
0 g
s
0 0 .411 r
.51386 .43036 m .52051 .44192 L p .52071 .4422 L F P
0 g
s
0 0 .411 r
.52051 .44192 m .52091 .44261 L .52071 .4422 L p F P
0 g
s
.911 .812 .737 r
.54601 .40359 m .53373 .42472 L .53462 .42604 L p F P
0 g
s
.873 .749 .723 r
.53373 .42472 m .54601 .40359 L .54297 .40089 L p F P
0 g
s
.847 .758 .765 r
.55905 .37425 m .54297 .40089 L .54601 .40359 L p F P
0 g
s
.568 .511 .751 r
.50869 .3988 m .49562 .37033 L .48849 .37463 L p F P
0 g
s
.568 .511 .751 r
.48849 .37463 m .50481 .40109 L .50869 .3988 L p F P
0 g
s
.59 .559 .788 r
.46649 .34634 m .48849 .37463 L .49562 .37033 L p F P
0 g
s
.499 .653 .927 r
.74798 .31654 m .78628 .32199 L .7684 .28665 L p F P
0 g
s
.499 .653 .927 r
.7684 .28665 m .73215 .28636 L .74798 .31654 L p F P
0 g
s
.816 .712 .75 r
.54297 .40089 m .55905 .37425 L .55178 .37002 L p F P
0 g
s
.794 .724 .787 r
.55905 .37425 m .58072 .3457 L .56902 .33875 L p F P
0 g
s
.794 .724 .787 r
.56902 .33875 m .55178 .37002 L .55905 .37425 L p F P
0 g
s
0 .043 .603 r
.52066 .44176 m .51304 .42757 L .51319 .42899 L p F P
0 g
s
0 .043 .603 r
.51319 .42899 m .52071 .4422 L .52066 .44176 L p F P
0 g
s
.896 .855 .404 r
.52736 .44216 m .5274 .44173 L p .52719 .44239 L F P
0 g
s
0 0 .148 r
.52717 .44258 m .52443 .45112 L .52438 .45117 L p F P
0 g
s
0 0 0 r
.52736 .44216 m .52445 .45106 L .52443 .45112 L p F P
0 g
s
.896 .855 .404 r
.52445 .45106 m .52736 .44216 L p .52719 .44239 L F P
0 g
s
.729 .86 .434 r
.52736 .44216 m .52717 .44258 L p .52754 .44193 L F P
0 g
s
.729 .86 .434 r
.5349 .42887 m .52736 .44216 L p .52754 .44193 L F P
0 g
s
.948 .941 .613 r
.52736 .44216 m .5349 .42887 L .53502 .42745 L p F P
0 g
s
.72 .678 .809 r
.51654 .57871 m .51312 .61251 L .4898 .60986 L p F P
0 g
s
.698 .658 .807 r
.51312 .61251 m .51654 .57871 L .53286 .57863 L p F P
0 g
s
.698 .658 .807 r
.53286 .57863 m .5369 .61239 L .51312 .61251 L p F P
0 g
s
.916 .948 .804 r
.22783 .33862 m .20274 .35934 L .19949 .40257 L p F P
0 g
s
.949 .937 .602 r
.18482 .43029 m .19949 .40257 L .20274 .35934 L p F P
0 g
s
.616 .659 .869 r
.43932 .31852 m .40804 .29355 L .39117 .31092 L p F P
0 g
s
.651 .721 .899 r
.35288 .29629 m .39117 .31092 L .40804 .29355 L p F P
0 g
s
.59 .559 .788 r
.49562 .37033 m .47798 .33926 L .46649 .34634 L p F P
0 g
s
.612 .608 .823 r
.43932 .31852 m .46649 .34634 L .47798 .33926 L p F P
0 g
s
.674 .641 .809 r
.5369 .61239 m .53286 .57863 L .5488 .57662 L p F P
0 g
s
.674 .641 .809 r
.5488 .57662 m .56013 .60953 L .5369 .61239 L p F P
0 g
s
.545 .462 .712 r
.51727 .42263 m .50869 .3988 L .50481 .40109 L p F P
0 g
s
.545 .462 .712 r
.50481 .40109 m .51556 .42363 L .51727 .42263 L p F P
0 g
s
.107 .438 .876 r
.22257 .55358 m .25174 .57089 L .27309 .60199 L p F P
0 g
s
.412 .637 .948 r
.30627 .60744 m .27309 .60199 L .25174 .57089 L p F P
0 g
s
.131 .199 .667 r
.52077 .44133 m .51339 .42615 L .51304 .42757 L p F P
0 g
s
.131 .199 .667 r
.51304 .42757 m .52066 .44176 L .52077 .44133 L p F P
0 g
s
.873 .749 .723 r
.54297 .40089 m .53238 .42354 L .53373 .42472 L p F P
0 g
s
.835 .697 .712 r
.53238 .42354 m .54297 .40089 L .53902 .39864 L p F P
0 g
s
.816 .712 .75 r
.55178 .37002 m .53902 .39864 L .54297 .40089 L p F P
0 g
s
.948 .941 .613 r
.53502 .42745 m .5274 .44173 L .52736 .44216 L p F P
0 g
s
.977 .884 .67 r
.5274 .44173 m .53502 .42745 L .53462 .42604 L p F P
0 g
s
.767 .734 .822 r
.58072 .3457 m .6075 .31756 L .59039 .30716 L p F P
0 g
s
.767 .734 .822 r
.59039 .30716 m .56902 .33875 L .58072 .3457 L p F P
0 g
s
.685 .747 .899 r
.65575 .30941 m .69374 .29433 L .6717 .27194 L p F P
0 g
s
.685 .747 .899 r
.6717 .27194 m .63834 .29222 L .65575 .30941 L p F P
0 g
s
.316 .296 .678 r
.52103 .44093 m .51424 .42482 L .51339 .42615 L p F P
0 g
s
.316 .296 .678 r
.51339 .42615 m .52077 .44133 L .52103 .44093 L p F P
0 g
s
.977 .884 .67 r
.53462 .42604 m .52727 .4413 L .5274 .44173 L p F P
0 g
s
.943 .805 .678 r
.52727 .4413 m .53462 .42604 L .53373 .42472 L p F P
0 g
s
.599 .493 .704 r
.5193 .42188 m .51329 .39707 L .50869 .3988 L p F P
0 g
s
.599 .493 .704 r
.50869 .3988 m .51727 .42263 L .5193 .42188 L p F P
0 g
s
.614 .534 .741 r
.49562 .37033 m .50869 .3988 L .51329 .39707 L p F P
0 g
s
.614 .534 .741 r
.51329 .39707 m .50414 .36708 L .49562 .37033 L p F P
0 g
s
.627 .575 .777 r
.47798 .33926 m .49562 .37033 L .50414 .36708 L p F P
0 g
s
.835 .697 .712 r
.53902 .39864 m .53063 .42256 L .53238 .42354 L p F P
0 g
s
.797 .653 .703 r
.53063 .42256 m .53902 .39864 L .53436 .39696 L p F P
0 g
s
.785 .673 .74 r
.53902 .39864 m .55178 .37002 L .54317 .36686 L p F P
0 g
s
.785 .673 .74 r
.54317 .36686 m .53436 .39696 L .53902 .39864 L p F P
0 g
s
.741 .701 .814 r
.50054 .57686 m .4898 .60986 L .46791 .60456 L p F P
0 g
s
.72 .678 .809 r
.4898 .60986 m .50054 .57686 L .51654 .57871 L p F P
0 g
s
.77 .691 .776 r
.55178 .37002 m .56902 .33875 L .55509 .33353 L p F P
0 g
s
.77 .691 .776 r
.55509 .33353 m .54317 .36686 L .55178 .37002 L p F P
0 g
s
.75 .909 .946 r
.77715 .59954 m .74362 .60533 L .7607 .5783 L p F P
0 g
s
.75 .909 .946 r
.7607 .5783 m .79753 .5682 L .77715 .59954 L p F P
0 g
s
.435 .361 .675 r
.52144 .44057 m .51556 .42363 L .51424 .42482 L p F P
0 g
s
.435 .361 .675 r
.51424 .42482 m .52103 .44093 L .52144 .44057 L p F P
0 g
s
.648 .626 .814 r
.56013 .60953 m .5488 .57662 L .56366 .57276 L p F P
0 g
s
.648 .626 .814 r
.56366 .57276 m .58186 .60401 L .56013 .60953 L p F P
0 g
s
.943 .805 .678 r
.53373 .42472 m .527 .4409 L .52727 .4413 L p F P
0 g
s
.896 .736 .674 r
.527 .4409 m .53373 .42472 L .53238 .42354 L p F P
0 g
s
0 .029 .592 r
.84889 .39904 m .86359 .42666 L .85916 .38206 L p F P
0 g
s
0 .029 .592 r
.85916 .38206 m .8443 .35576 L .84889 .39904 L p F P
0 g
s
.645 .522 .698 r
.52156 .42141 m .51841 .39599 L .51329 .39707 L p F P
0 g
s
.645 .522 .698 r
.51329 .39707 m .5193 .42188 L .52156 .42141 L p F P
0 g
s
.653 .557 .734 r
.50414 .36708 m .51329 .39707 L .51841 .39599 L p F P
0 g
s
.519 .409 .669 r
.52196 .44027 m .51727 .42263 L .51556 .42363 L p F P
0 g
s
.519 .409 .669 r
.51556 .42363 m .52144 .44057 L .52196 .44027 L p F P
0 g
s
.797 .653 .703 r
.53436 .39696 m .52858 .42183 L .53063 .42256 L p F P
0 g
s
.761 .616 .698 r
.52858 .42183 m .53436 .39696 L .52921 .39593 L p F P
0 g
s
.754 .639 .734 r
.53436 .39696 m .54317 .36686 L .53363 .36492 L p F P
0 g
s
.754 .639 .734 r
.53363 .36492 m .52921 .39593 L .53436 .39696 L p F P
0 g
s
.896 .736 .674 r
.53238 .42354 m .52658 .44054 L .527 .4409 L p F P
0 g
s
.851 .679 .669 r
.52658 .44054 m .53238 .42354 L .53063 .42256 L p F P
0 g
s
.674 .764 .919 r
.36578 .62407 m .39955 .60874 L .42181 .62197 L p F P
0 g
s
.736 .763 .874 r
.4484 .5968 m .42181 .62197 L .39955 .60874 L p F P
0 g
s
.727 .744 .865 r
.42181 .62197 m .4484 .5968 L .46791 .60456 L p F P
0 g
s
0 0 .212 r
.18172 .46005 m .19075 .48942 L .20306 .53014 L p F P
0 g
s
0 .162 .7 r
.22257 .55358 m .20306 .53014 L .19075 .48942 L p F P
0 g
s
.686 .551 .696 r
.52393 .42124 m .52381 .3956 L .51841 .39599 L p F P
0 g
s
.686 .551 .696 r
.51841 .39599 m .52156 .42141 L .52393 .42124 L p F P
0 g
s
.688 .582 .731 r
.51362 .36503 m .51841 .39599 L .52381 .3956 L p F P
0 g
s
.653 .557 .734 r
.51841 .39599 m .51362 .36503 L .50414 .36708 L p F P
0 g
s
.583 .449 .664 r
.52258 .44004 m .5193 .42188 L .51727 .42263 L p F P
0 g
s
.583 .449 .664 r
.51727 .42263 m .52196 .44027 L .52258 .44004 L p F P
0 g
s
.761 .616 .698 r
.52921 .39593 m .52631 .42138 L .52858 .42183 L p F P
0 g
s
.724 .582 .696 r
.52631 .42138 m .52921 .39593 L .52381 .3956 L p F P
0 g
s
.722 .609 .731 r
.52921 .39593 m .53363 .36492 L .52361 .3643 L p F P
0 g
s
.722 .609 .731 r
.52361 .3643 m .52381 .3956 L .52921 .39593 L p F P
0 g
s
.851 .679 .669 r
.53063 .42256 m .52605 .44024 L .52658 .44054 L p F P
0 g
s
.807 .632 .664 r
.52605 .44024 m .53063 .42256 L .52858 .42183 L p F P
0 g
s
.612 .608 .823 r
.47798 .33926 m .45611 .30792 L .43932 .31852 L p F P
0 g
s
.633 .661 .859 r
.40804 .29355 m .43932 .31852 L .45611 .30792 L p F P
0 g
s
.724 .582 .696 r
.52381 .3956 m .52393 .42124 L .52631 .42138 L p F P
0 g
s
.688 .582 .731 r
.52381 .3956 m .52361 .3643 L .51362 .36503 L p F P
0 g
s
.614 .665 .875 r
.62827 .62099 m .60113 .59606 L .61708 .58597 L p F P
0 g
s
.614 .665 .875 r
.61708 .58597 m .65011 .60753 L .62827 .62099 L p F P
0 g
s
.659 .593 .77 r
.49175 .33389 m .50414 .36708 L .51362 .36503 L p F P
0 g
s
.627 .575 .777 r
.50414 .36708 m .49175 .33389 L .47798 .33926 L p F P
0 g
s
.725 .395 .421 r
.52389 .45127 m .52298 .44376 L .5237 .44385 L p F P
0 g
s
.725 .395 .421 r
.5237 .44385 m .52398 .45128 L .52389 .45127 L p F P
0 g
s
.663 .345 .425 r
.52398 .45128 m .5237 .44385 L .52442 .44385 L p F P
0 g
s
.663 .345 .425 r
.52442 .44385 m .52408 .45128 L .52398 .45128 L p F P
0 g
s
.592 .286 .42 r
.52408 .45128 m .52442 .44385 L .52513 .44375 L p F P
0 g
s
.785 .44 .404 r
.52381 .45124 m .52232 .44358 L .52298 .44376 L p F P
0 g
s
.785 .44 .404 r
.52298 .44376 m .52389 .45127 L .52381 .45124 L p F P
0 g
s
.592 .286 .42 r
.52513 .44375 m .52417 .45127 L .52408 .45128 L p F P
0 g
s
.506 .212 .401 r
.52417 .45127 m .52513 .44375 L .52579 .44356 L p F P
0 g
s
.744 .662 .769 r
.54317 .36686 m .55509 .33353 L .53961 .33032 L p F P
0 g
s
.744 .662 .769 r
.53961 .33032 m .53363 .36492 L .54317 .36686 L p F P
0 g
s
.844 .482 .369 r
.52373 .45121 m .52173 .44332 L .52232 .44358 L p F P
0 g
s
.844 .482 .369 r
.52232 .44358 m .52381 .45124 L .52373 .45121 L p F P
0 g
s
.506 .212 .401 r
.52579 .44356 m .52425 .45124 L .52417 .45127 L p F P
0 g
s
.388 .109 .364 r
.52425 .45124 m .52579 .44356 L .52637 .4433 L p F P
0 g
s
.636 .485 .661 r
.52327 .4399 m .52156 .42141 L .5193 .42188 L p F P
0 g
s
.636 .485 .661 r
.5193 .42188 m .52258 .44004 L .52327 .4399 L p F P
0 g
s
.807 .632 .664 r
.52858 .42183 m .52542 .44002 L .52605 .44024 L p F P
0 g
s
.766 .591 .661 r
.52542 .44002 m .52858 .42183 L .52631 .42138 L p F P
0 g
s
.567 .736 .955 r
.30627 .60744 m .34374 .6043 L .36578 .62407 L p F P
0 g
s
.674 .769 .922 r
.39955 .60874 m .36578 .62407 L .34374 .6043 L p F P
0 g
s
.899 .519 .3 r
.52367 .45117 m .52126 .443 L .52173 .44332 L p F P
0 g
s
.899 .519 .3 r
.52173 .44332 m .52373 .45121 L .52367 .45117 L p F P
0 g
s
.388 .109 .364 r
.52637 .4433 m .52433 .45121 L .52425 .45124 L p F P
0 g
s
.21 0 .292 r
.52433 .45121 m .52637 .4433 L .52684 .44297 L p F P
0 g
s
.682 .519 .659 r
.524 .43985 m .52393 .42124 L .52156 .42141 L p F P
0 g
s
.682 .519 .659 r
.52156 .42141 m .52327 .4399 L .524 .43985 L p F P
0 g
s
.766 .591 .661 r
.52631 .42138 m .52472 .43989 L .52542 .44002 L p F P
0 g
s
.725 .554 .659 r
.52472 .43989 m .52631 .42138 L .52393 .42124 L p F P
0 g
s
.732 .74 .858 r
.6075 .31756 m .63834 .29222 L .61509 .27775 L p F P
0 g
s
.732 .74 .858 r
.61509 .27775 m .59039 .30716 L .6075 .31756 L p F P
0 g
s
.922 .532 .165 r
.52363 .45112 m .52091 .44261 L .52126 .443 L p F P
0 g
s
.922 .532 .165 r
.52126 .443 m .52367 .45117 L .52363 .45112 L p F P
0 g
s
.21 0 .292 r
.52684 .44297 m .52438 .45117 L .52433 .45121 L p F P
0 g
s
0 0 .148 r
.52438 .45117 m .52684 .44297 L .52717 .44258 L p F P
0 g
s
.639 .618 .813 r
.45611 .30792 m .47798 .33926 L .49175 .33389 L p F P
0 g
s
.725 .554 .659 r
.52393 .42124 m .524 .43985 L .52472 .43989 L p F P
0 g
s
.75 .708 .812 r
.56902 .33875 m .59039 .30716 L .5699 .29932 L p F P
0 g
s
.75 .708 .812 r
.5699 .29932 m .55509 .33353 L .56902 .33875 L p F P
0 g
s
0 0 .383 r
.52071 .4422 m .5236 .45107 L .5236 .45101 L p F P
0 g
s
.689 .613 .766 r
.50712 .33051 m .51362 .36503 L .52361 .3643 L p F P
0 g
s
.659 .593 .77 r
.51362 .36503 m .50712 .33051 L .49175 .33389 L p F P
0 g
s
.717 .636 .766 r
.53363 .36492 m .53961 .33032 L .52335 .3293 L p F P
0 g
s
.717 .636 .766 r
.52335 .3293 m .52361 .3643 L .53363 .36492 L p F P
0 g
s
0 .026 .544 r
.52361 .45096 m .52077 .44133 L .52066 .44176 L p F P
0 g
s
0 .026 .544 r
.52066 .44176 m .5236 .45101 L .52361 .45096 L p F P
0 g
s
.896 .855 .404 r
.5274 .44173 m .52446 .45101 L .52445 .45106 L p F P
0 g
s
.997 .858 .552 r
.52446 .45101 m .5274 .44173 L .52727 .4413 L p F P
0 g
s
.633 .736 .922 r
.68436 .62255 m .65011 .60753 L .66657 .59169 L p F P
0 g
s
.633 .736 .922 r
.66657 .59169 m .70578 .60255 L .68436 .62255 L p F P
0 g
s
.899 .982 .708 r
.84656 .52688 m .82691 .55057 L .83682 .51274 L p F P
0 g
s
.899 .982 .708 r
.83682 .51274 m .85761 .48597 L .84656 .52688 L p F P
0 g
s
.228 .184 .598 r
.52364 .45091 m .52103 .44093 L .52077 .44133 L p F P
0 g
s
.228 .184 .598 r
.52077 .44133 m .52361 .45096 L .52364 .45091 L p F P
0 g
s
.997 .858 .552 r
.52727 .4413 m .52444 .45095 L .52446 .45101 L p F P
0 g
s
.97 .787 .601 r
.52444 .45095 m .52727 .4413 L .527 .4409 L p F P
0 g
s
.689 .613 .766 r
.52361 .3643 m .52335 .3293 L .50712 .33051 L p F P
0 g
s
.761 .727 .822 r
.48556 .57315 m .46791 .60456 L .4484 .5968 L p F P
0 g
s
.741 .701 .814 r
.46791 .60456 m .48556 .57315 L .50054 .57686 L p F P
0 g
s
.385 .283 .614 r
.5237 .45086 m .52144 .44057 L .52103 .44093 L p F P
0 g
s
.385 .283 .614 r
.52103 .44093 m .52364 .45091 L .5237 .45086 L p F P
0 g
s
.917 .715 .615 r
.5244 .4509 m .527 .4409 L .52658 .44054 L p F P
0 g
s
.97 .787 .601 r
.527 .4409 m .5244 .4509 L .52444 .45095 L p F P
0 g
s
.489 .35 .619 r
.52376 .45082 m .52196 .44027 L .52144 .44057 L p F P
0 g
s
.489 .35 .619 r
.52144 .44057 m .5237 .45086 L .52376 .45082 L p F P
0 g
s
.917 .715 .615 r
.52658 .44054 m .52435 .45086 L .5244 .4509 L p F P
0 g
s
.864 .656 .619 r
.52435 .45086 m .52658 .44054 L .52605 .44024 L p F P
0 g
s
.757 .86 .925 r
.31417 .28878 m .27765 .28949 L .26089 .32497 L p F P
0 g
s
.82 .939 .906 r
.22783 .33862 m .26089 .32497 L .27765 .28949 L p F P
0 g
s
.62 .614 .823 r
.58186 .60401 m .56366 .57276 L .57678 .56722 L p F P
0 g
s
.62 .614 .823 r
.57678 .56722 m .60113 .59606 L .58186 .60401 L p F P
0 g
s
.564 .402 .62 r
.52384 .45079 m .52258 .44004 L .52196 .44027 L p F P
0 g
s
.564 .402 .62 r
.52196 .44027 m .52376 .45082 L .52384 .45079 L p F P
0 g
s
.864 .656 .619 r
.52605 .44024 m .52428 .45082 L .52435 .45086 L p F P
0 g
s
.815 .606 .62 r
.52428 .45082 m .52605 .44024 L .52542 .44002 L p F P
0 g
s
.625 .445 .619 r
.52393 .45078 m .52327 .4399 L .52258 .44004 L p F P
0 g
s
.625 .445 .619 r
.52258 .44004 m .52384 .45079 L .52393 .45078 L p F P
0 g
s
.815 .606 .62 r
.52542 .44002 m .5242 .45079 L .52428 .45082 L p F P
0 g
s
.769 .563 .619 r
.5242 .45079 m .52542 .44002 L .52472 .43989 L p F P
0 g
s
.676 .485 .619 r
.52402 .45077 m .524 .43985 L .52327 .4399 L p F P
0 g
s
.676 .485 .619 r
.52327 .4399 m .52393 .45078 L .52402 .45077 L p F P
0 g
s
.769 .563 .619 r
.52472 .43989 m .52412 .45078 L .5242 .45079 L p F P
0 g
s
.723 .523 .619 r
.52412 .45078 m .52472 .43989 L .524 .43985 L p F P
0 g
s
.723 .523 .619 r
.524 .43985 m .52402 .45077 L .52412 .45078 L p F P
0 g
s
.651 .721 .899 r
.40804 .29355 m .37421 .27367 L .35288 .29629 L p F P
0 g
s
.69 .791 .927 r
.31417 .28878 m .35288 .29629 L .37421 .27367 L p F P
0 g
s
.664 .631 .806 r
.47635 .29986 m .49175 .33389 L .50712 .33051 L p F P
0 g
s
.639 .618 .813 r
.49175 .33389 m .47635 .29986 L .45611 .30792 L p F P
0 g
s
.324 .541 .908 r
.78628 .32199 m .81927 .33529 L .79979 .29568 L p F P
0 g
s
.324 .541 .908 r
.79979 .29568 m .7684 .28665 L .78628 .32199 L p F P
0 g
s
.73 .684 .806 r
.55509 .33353 m .5699 .29932 L .54706 .2945 L p F P
0 g
s
.73 .684 .806 r
.54706 .2945 m .53961 .33032 L .55509 .33353 L p F P
0 g
s
.607 .724 .928 r
.69374 .29433 m .73215 .28636 L .7056 .25872 L p F P
0 g
s
.607 .724 .928 r
.7056 .25872 m .6717 .27194 L .69374 .29433 L p F P
0 g
s
.737 .452 .496 r
.52398 .45128 m .52403 .45422 L .52389 .45127 L closepath p F P
0 g
s
.68 .405 .498 r
.52408 .45128 m .52403 .45422 L .52398 .45128 L closepath p F P
0 g
s
.618 .354 .495 r
.52417 .45127 m .52403 .45422 L .52408 .45128 L closepath p F P
0 g
s
.793 .497 .487 r
.52389 .45127 m .52403 .45422 L .52381 .45124 L closepath p F P
0 g
s
.543 .292 .485 r
.52425 .45124 m .52403 .45422 L .52417 .45127 L closepath p F P
0 g
s
.851 .543 .467 r
.52381 .45124 m .52403 .45422 L .52373 .45121 L closepath p F P
0 g
s
.445 .212 .465 r
.52433 .45121 m .52403 .45422 L .52425 .45124 L closepath p F P
0 g
s
.911 .592 .429 r
.52373 .45121 m .52403 .45422 L .52367 .45117 L closepath p F P
0 g
s
.304 .097 .424 r
.52438 .45117 m .52403 .45422 L .52433 .45121 L closepath p F P
0 g
s
.962 .638 .351 r
.52367 .45117 m .52403 .45422 L .52363 .45112 L closepath p F P
0 g
s
.076 0 .34 r
.52443 .45112 m .52403 .45422 L .52438 .45117 L closepath p F P
0 g
s
.939 .638 .18 r
.52363 .45112 m .52403 .45422 L .5236 .45107 L closepath p F P
0 g
s
0 0 .159 r
.52445 .45106 m .52403 .45422 L .52443 .45112 L closepath p F P
0 g
s
0 0 .118 r
.5236 .45107 m .52403 .45422 L .5236 .45101 L closepath p F P
0 g
s
.765 .69 .146 r
.52446 .45101 m .52403 .45422 L .52445 .45106 L closepath p F P
0 g
s
0 0 .379 r
.5236 .45101 m .52403 .45422 L .52361 .45096 L closepath p F P
0 g
s
.984 .797 .395 r
.52444 .45095 m .52403 .45422 L .52446 .45101 L closepath p F P
0 g
s
.125 .054 .498 r
.52361 .45096 m .52403 .45422 L .52364 .45091 L closepath p F P
0 g
s
.985 .754 .504 r
.5244 .4509 m .52403 .45422 L .52444 .45095 L closepath p F P
0 g
s
.327 .196 .544 r
.52364 .45091 m .52403 .45422 L .5237 .45086 L closepath p F P
0 g
s
.932 .688 .547 r
.52435 .45086 m .52403 .45422 L .5244 .4509 L closepath p F P
0 g
s
.454 .287 .563 r
.5237 .45086 m .52403 .45422 L .52376 .45082 L closepath p F P
0 g
s
.874 .629 .564 r
.52428 .45082 m .52403 .45422 L .52435 .45086 L closepath p F P
0 g
s
.543 .352 .572 r
.52376 .45082 m .52403 .45422 L .52384 .45079 L closepath p F P
0 g
s
.82 .578 .572 r
.5242 .45079 m .52403 .45422 L .52428 .45082 L closepath p F P
0 g
s
.611 .404 .576 r
.52384 .45079 m .52403 .45422 L .52393 .45078 L closepath p F P
0 g
s
.77 .533 .576 r
.52412 .45078 m .52403 .45422 L .5242 .45079 L closepath p F P
0 g
s
.669 .449 .578 r
.52393 .45078 m .52403 .45422 L .52402 .45077 L closepath p F P
0 g
s
.72 .491 .578 r
.52402 .45077 m .52403 .45422 L .52412 .45078 L closepath p F P
0 g
s
.731 .65 .769 r
.51937 .54462 m .51654 .57871 L .50054 .57686 L p F P
0 g
s
.703 .625 .767 r
.51654 .57871 m .51937 .54462 L .52951 .54457 L p F P
0 g
s
.703 .625 .767 r
.52951 .54457 m .53286 .57863 L .51654 .57871 L p F P
0 g
s
.673 .603 .769 r
.53286 .57863 m .52951 .54457 L .5394 .5433 L p F P
0 g
s
.673 .603 .769 r
.5394 .5433 m .5488 .57662 L .53286 .57863 L p F P
0 g
s
.239 .133 .526 r
.52403 .45422 m .52443 .45751 L .52438 .45756 L closepath p F P
0 g
s
.939 .617 .397 r
.52403 .45422 m .52435 .45725 L .5244 .45729 L closepath p F P
0 g
s
.397 .245 .556 r
.52403 .45422 m .52438 .45756 L .52432 .4576 L closepath p F P
0 g
s
.881 .567 .452 r
.52403 .45422 m .52428 .45721 L .52435 .45725 L closepath p F P
0 g
s
.502 .322 .568 r
.52403 .45422 m .52432 .4576 L .52425 .45763 L closepath p F P
0 g
s
.822 .52 .479 r
.52403 .45422 m .5242 .45718 L .52428 .45721 L closepath p F P
0 g
s
.579 .379 .574 r
.52403 .45422 m .52425 .45763 L .52417 .45766 L closepath p F P
0 g
s
.765 .474 .492 r
.52403 .45422 m .52411 .45717 L .5242 .45718 L closepath p F P
0 g
s
.695 .47 .578 r
.52403 .45422 m .52408 .45767 L .52398 .45767 L closepath p F P
0 g
s
.641 .427 .577 r
.52403 .45422 m .52417 .45766 L .52408 .45767 L closepath p F P
0 g
s
.709 .429 .498 r
.52403 .45422 m .52402 .45716 L .52411 .45717 L closepath p F P
0 g
s
.745 .512 .577 r
.52403 .45422 m .52398 .45767 L .52389 .45766 L closepath p F P
0 g
s
.65 .38 .498 r
.52403 .45422 m .52393 .45717 L .52402 .45716 L closepath p F P
0 g
s
.795 .555 .575 r
.52403 .45422 m .52389 .45766 L .52381 .45763 L closepath p F P
0 g
s
.582 .325 .491 r
.52403 .45422 m .52384 .45719 L .52393 .45717 L closepath p F P
0 g
s
.847 .602 .569 r
.52403 .45422 m .52381 .45763 L .52373 .4576 L closepath p F P
0 g
s
.498 .255 .477 r
.52403 .45422 m .52376 .45721 L .52384 .45719 L closepath p F P
0 g
s
.903 .657 .557 r
.52403 .45422 m .52373 .4576 L .52367 .45756 L closepath p F P
0 g
s
.382 .16 .448 r
.52403 .45422 m .52369 .45725 L .52376 .45721 L closepath p F P
0 g
s
.96 .721 .53 r
.52403 .45422 m .52367 .45756 L .52363 .45751 L closepath p F P
0 g
s
.205 .017 .39 r
.52403 .45422 m .52364 .4573 L .52369 .45725 L closepath p F P
0 g
s
.998 .783 .463 r
.52403 .45422 m .52363 .45751 L .5236 .45746 L closepath p F P
0 g
s
0 0 .266 r
.52403 .45422 m .52361 .45735 L .52364 .4573 L closepath p F P
0 g
s
.915 .774 .29 r
.52403 .45422 m .5236 .45746 L .52359 .4574 L closepath p F P
0 g
s
0 0 .016 r
.52403 .45422 m .52359 .4574 L .52361 .45735 L closepath p F P
0 g
s
0 0 .453 r
.52403 .45422 m .52445 .45745 L .52443 .45751 L closepath p F P
0 g
s
0 0 .267 r
.52403 .45422 m .52446 .4574 L .52445 .45745 L closepath p F P
0 g
s
.845 .585 .043 r
.52403 .45422 m .52444 .45735 L .52446 .4574 L closepath p F P
0 g
s
.968 .649 .282 r
.52403 .45422 m .5244 .45729 L .52444 .45735 L closepath p F P
0 g
s
.687 .646 .803 r
.49904 .29477 m .50712 .33051 L .52335 .3293 L p F P
0 g
s
.664 .631 .806 r
.50712 .33051 m .49904 .29477 L .47635 .29986 L p F P
0 g
s
.758 .679 .774 r
.50944 .54345 m .50054 .57686 L .48556 .57315 L p F P
0 g
s
.731 .65 .769 r
.50054 .57686 m .50944 .54345 L .51937 .54462 L p F P
0 g
s
.71 .664 .803 r
.53961 .33032 m .54706 .2945 L .52302 .29295 L p F P
0 g
s
.71 .664 .803 r
.52302 .29295 m .52335 .3293 L .53961 .33032 L p F P
0 g
s
.641 .584 .774 r
.5488 .57662 m .5394 .5433 L .5486 .54086 L p F P
0 g
s
.641 .584 .774 r
.5486 .54086 m .56366 .57276 L .5488 .57662 L p F P
0 g
s
.949 .937 .602 r
.20274 .35934 m .18786 .38575 L .18482 .43029 L p F P
0 g
s
.744 .66 .105 r
.18172 .46005 m .18482 .43029 L .18786 .38575 L p F P
0 g
s
.65 .666 .852 r
.43083 .27882 m .45611 .30792 L .47635 .29986 L p F P
0 g
s
.633 .661 .859 r
.45611 .30792 m .43083 .27882 L .40804 .29355 L p F P
0 g
s
.412 .637 .948 r
.25174 .57089 m .28835 .58062 L .30627 .60744 L p F P
0 g
s
.574 .726 .946 r
.34374 .6043 m .30627 .60744 L .28835 .58062 L p F P
0 g
s
.746 .543 .619 r
.52297 .46857 m .52389 .45766 L .52398 .45767 L p F P
0 g
s
.746 .543 .619 r
.52398 .45767 m .52369 .46866 L .52297 .46857 L p F P
0 g
s
.7 .504 .619 r
.52369 .46866 m .52398 .45767 L .52408 .45767 L p F P
0 g
s
.7 .504 .619 r
.52408 .45767 m .52442 .46865 L .52369 .46866 L p F P
0 g
s
.651 .465 .619 r
.52442 .46865 m .52408 .45767 L .52417 .45766 L p F P
0 g
s
.792 .584 .62 r
.5223 .46839 m .52381 .45763 L .52389 .45766 L p F P
0 g
s
.792 .584 .62 r
.52389 .45766 m .52297 .46857 L .5223 .46839 L p F P
0 g
s
.651 .465 .619 r
.52417 .45766 m .52514 .46856 L .52442 .46865 L p F P
0 g
s
.596 .424 .62 r
.52514 .46856 m .52417 .45766 L .52425 .45763 L p F P
0 g
s
.839 .63 .62 r
.52171 .46813 m .52373 .4576 L .52381 .45763 L p F P
0 g
s
.839 .63 .62 r
.52381 .45763 m .5223 .46839 L .52171 .46813 L p F P
0 g
s
.596 .424 .62 r
.52425 .45763 m .5258 .46837 L .52514 .46856 L p F P
0 g
s
.529 .377 .619 r
.5258 .46837 m .52425 .45763 L .52432 .4576 L p F P
0 g
s
.89 .684 .618 r
.52123 .46781 m .52367 .45756 L .52373 .4576 L p F P
0 g
s
.89 .684 .618 r
.52373 .4576 m .52171 .46813 L .52123 .46781 L p F P
0 g
s
.529 .377 .619 r
.52432 .4576 m .52638 .46811 L .5258 .46837 L p F P
0 g
s
.441 .319 .617 r
.52638 .46811 m .52432 .4576 L .52438 .45756 L p F P
0 g
s
.944 .75 .61 r
.52088 .46743 m .52363 .45751 L .52367 .45756 L p F P
0 g
s
.944 .75 .61 r
.52367 .45756 m .52123 .46781 L .52088 .46743 L p F P
0 g
s
.441 .319 .617 r
.52438 .45756 m .52685 .46778 L .52638 .46811 L p F P
0 g
s
.315 .238 .609 r
.52685 .46778 m .52438 .45756 L .52443 .45751 L p F P
0 g
s
.687 .646 .803 r
.52335 .3293 m .52302 .29295 L .49904 .29477 L p F P
0 g
s
.99 .825 .583 r
.52067 .46701 m .5236 .45746 L .52363 .45751 L p F P
0 g
s
.99 .825 .583 r
.52363 .45751 m .52088 .46743 L .52067 .46701 L p F P
0 g
s
.656 .721 .896 r
.37421 .27367 m .40804 .29355 L .43083 .27882 L p F P
0 g
s
.315 .238 .609 r
.52443 .45751 m .52719 .4674 L .52685 .46778 L p F P
0 g
s
.115 .114 .579 r
.52719 .4674 m .52443 .45751 L .52445 .45745 L p F P
0 g
s
.974 .875 .495 r
.52063 .46658 m .52359 .4574 L .5236 .45746 L p F P
0 g
s
.974 .875 .495 r
.5236 .45746 m .52067 .46701 L .52063 .46658 L p F P
0 g
s
.115 .114 .579 r
.52445 .45745 m .52738 .46698 L .52719 .4674 L p F P
0 g
s
0 0 .482 r
.52738 .46698 m .52445 .45745 L .52446 .4574 L p F P
0 g
s
.74 .77 .269 r
.52074 .46616 m .52361 .45735 L .52359 .4574 L p F P
0 g
s
.868 .927 .539 r
.5132 .47931 m .52074 .46616 L .52063 .46658 L p F P
0 g
s
.868 .927 .539 r
.52063 .46658 m .51382 .47892 L p .5132 .47931 L F P
0 g
s
.868 .927 .539 r
.51382 .47892 m .51284 .4807 L .5132 .47931 L p F P
0 g
s
.71 .928 .598 r
.50122 .492 m .51241 .47981 L p .5132 .47931 L F P
0 g
s
0 0 0 r
.52074 .46616 m .51415 .47765 L p .51407 .478 L F P
0 g
s
0 0 0 r
.51415 .47765 m .5132 .47931 L .51407 .478 L p F P
0 g
s
.71 .928 .598 r
.51241 .47981 m .51407 .478 L .5132 .47931 L p F P
0 g
s
.71 .928 .598 r
.5132 .47931 m .50175 .49218 L p .50122 .492 L F P
0 g
s
.71 .928 .598 r
.50175 .49218 m .49925 .49499 L .50122 .492 L p F P
0 g
s
.577 .916 .699 r
.48706 .5005 m .50051 .49176 L p .50122 .492 L F P
0 g
s
0 0 0 r
.51407 .478 m .50352 .48949 L p .50426 .48932 L F P
0 g
s
0 0 0 r
.50352 .48949 m .50122 .492 L .50426 .48932 L p F P
0 g
s
.577 .916 .699 r
.50051 .49176 m .50426 .48932 L .50122 .492 L p F P
0 g
s
.904 .983 .723 r
.49925 .49499 m .5132 .47931 L .51284 .4807 L p F P
0 g
s
.904 .983 .723 r
.51284 .4807 m .49846 .49814 L .49925 .49499 L p F P
0 g
s
.577 .916 .699 r
.50122 .492 m .48572 .50255 L p .48706 .5005 L F P
0 g
s
.577 .916 .699 r
.48572 .50255 m .48142 .50547 L .48706 .5005 L p F P
0 g
s
.521 .918 .843 r
.47558 .50117 m .48884 .49778 L p .48706 .5005 L F P
0 g
s
0 0 0 r
.50426 .48932 m .49079 .49808 L p .49447 .49634 L F P
0 g
s
0 0 0 r
.49079 .49808 m .48706 .5005 L .49447 .49634 L p F P
0 g
s
0 0 0 r
.49447 .49634 m .50524 .4891 L p .50426 .48932 L F P
0 g
s
.791 .998 .803 r
.48142 .50547 m .50122 .492 L .49925 .49499 L p F P
0 g
s
.947 .927 .751 r
.49846 .49814 m .51284 .4807 L .513 .48211 L p F P
0 g
s
.947 .927 .751 r
.513 .48211 m .49885 .50132 L .49846 .49814 L p F P
0 g
s
0 0 0 r
.50824 .48708 m .49447 .49634 L .50332 .49319 L p F P
0 g
s
.32 .804 .762 r
.49005 .49601 m .49937 .49403 L p .49447 .49634 L F P
0 g
s
.32 .804 .762 r
.49937 .49403 m .50332 .49319 L .49447 .49634 L p F P
0 g
s
.521 .918 .843 r
.48884 .49778 m .49447 .49634 L .48706 .5005 L p F P
0 g
s
0 0 0 r
.50524 .4891 m .50824 .48708 L .50426 .48932 L p F P
0 g
s
0 0 0 r
.50426 .48932 m .51396 .47844 L p .51407 .478 L F P
0 g
s
.977 .921 .65 r
.51284 .4807 m .52063 .46658 L .52067 .46701 L p F P
0 g
s
.977 .921 .65 r
.52067 .46701 m .513 .48211 L .51284 .4807 L p F P
0 g
s
0 0 0 r
.51396 .47844 m .51541 .47682 L .51407 .478 L p F P
0 g
s
0 0 0 r
.51407 .478 m .52061 .46644 L p .52074 .46616 L F P
0 g
s
.74 .77 .269 r
.52084 .46592 m .52063 .46658 L .52074 .46616 L p F P
0 g
s
0 0 0 r
.52061 .46644 m .521 .46575 L .52074 .46616 L p F P
0 g
s
.74 .77 .269 r
.52359 .4574 m .52084 .46592 L p .52074 .46616 L F P
0 g
s
0 0 .482 r
.52446 .4574 m .52741 .46655 L .52738 .46698 L p F P
0 g
s
0 0 .243 r
.52741 .46655 m .52446 .4574 L .52444 .45735 L p F P
0 g
s
0 0 .032 r
.521 .46575 m .52364 .4573 L .52361 .45735 L p F P
0 g
s
0 0 .032 r
.52361 .45735 m .52074 .46616 L .521 .46575 L p F P
0 g
s
0 0 .243 r
.52729 .46612 m .52741 .46655 L p .5272 .46589 L F P
0 g
s
.7 .27 0 r
.52729 .46612 m .52701 .46572 L p .52738 .46636 L F P
0 g
s
0 0 .523 r
.53518 .48059 m .52741 .46655 L .52729 .46612 L p F P
0 g
s
0 0 .523 r
.53478 .4792 m .53518 .48059 L p .53423 .47883 L F P
0 g
s
0 0 .523 r
.52729 .46612 m .53478 .4792 L p .53423 .47883 L F P
0 g
s
.7 .27 0 r
.53478 .4792 m .52729 .46612 L p .52738 .46636 L F P
0 g
s
.7 .27 0 r
.53387 .47789 m .53478 .4792 L p .53384 .47756 L F P
0 g
s
.7 .27 0 r
.52701 .46572 m .53387 .47789 L p .53384 .47756 L F P
0 g
s
0 .127 .645 r
.52741 .46655 m .53518 .48059 L .53507 .48199 L p F P
0 g
s
.631 .108 0 r
.53387 .47789 m .5325 .47673 L p .53391 .47823 L F P
0 g
s
0 .035 .581 r
.53478 .4792 m .53387 .47789 L p .53535 .47958 L F P
0 g
s
0 .035 .581 r
.54865 .49473 m .53478 .4792 L p .53535 .47958 L F P
0 g
s
0 .035 .581 r
.54659 .49177 m .54865 .49473 L p .5462 .49194 L F P
0 g
s
0 .035 .581 r
.53387 .47789 m .54659 .49177 L p .5462 .49194 L F P
0 g
s
.631 .108 0 r
.54659 .49177 m .53387 .47789 L p .53391 .47823 L F P
0 g
s
0 .186 .716 r
.53478 .4792 m .54865 .49473 L .54955 .49788 L p F P
0 g
s
0 .186 .716 r
.54955 .49788 m .53518 .48059 L .53478 .4792 L p F P
0 g
s
.631 .108 0 r
.54346 .48912 m .54659 .49177 L p .54425 .48927 L F P
0 g
s
.631 .108 0 r
.5325 .47673 m .54346 .48912 L p .54425 .48927 L F P
0 g
s
.161 .292 .75 r
.54926 .50106 m .53507 .48199 L .53518 .48059 L p F P
0 g
s
.161 .292 .75 r
.53518 .48059 m .54955 .49788 L .54926 .50106 L p F P
0 g
s
.446 0 0 r
.54346 .48912 m .53941 .48692 L p .54266 .48897 L F P
0 g
s
0 .217 .686 r
.54659 .49177 m .54346 .48912 L p .54699 .49159 L F P
0 g
s
0 .217 .686 r
.56619 .50504 m .54659 .49177 L p .54699 .49159 L F P
0 g
s
0 .217 .686 r
.56039 .50013 m .56619 .50504 L p .56178 .50195 L F P
0 g
s
0 .217 .686 r
.54346 .48912 m .56039 .50013 L p .56178 .50195 L F P
0 g
s
.446 0 0 r
.56039 .50013 m .54346 .48912 L p .54266 .48897 L F P
0 g
s
0 .307 .798 r
.54659 .49177 m .56619 .50504 L .56997 .51053 L p F P
0 g
s
0 .307 .798 r
.56997 .51053 m .54865 .49473 L .54659 .49177 L p F P
0 g
s
.446 0 0 r
.55284 .49604 m .56039 .50013 L p .55631 .49756 L F P
0 g
s
.446 0 0 r
.53941 .48692 m .55284 .49604 L p .55631 .49756 L F P
0 g
s
.433 .407 .731 r
.54521 .50693 m .53332 .4846 L .53443 .48335 L p F P
0 g
s
.433 .407 .731 r
.53443 .48335 m .54778 .50412 L .54521 .50693 L p F P
0 g
s
.323 .359 .745 r
.54778 .50412 m .53443 .48335 L .53507 .48199 L p F P
0 g
s
.323 .359 .745 r
.53507 .48199 m .54926 .50106 L .54778 .50412 L p F P
0 g
s
.508 0 0 r
.54389 .49298 m .55284 .49604 L p .54929 .49394 L F P
0 g
s
.508 0 0 r
.53462 .48528 m .54389 .49298 L p .54929 .49394 L F P
0 g
s
.572 .477 .708 r
.53176 .48567 m .54166 .50935 L .5373 .51127 L p F P
0 g
s
.511 .444 .718 r
.54166 .50935 m .53176 .48567 L .53332 .4846 L p F P
0 g
s
.511 .444 .718 r
.53332 .4846 m .54521 .50693 L .54166 .50935 L p F P
0 g
s
.149 0 0 r
.54389 .49298 m .53397 .4911 L p .53851 .49201 L F P
0 g
s
0 .44 .755 r
.55284 .49604 m .54389 .49298 L p .54938 .49452 L F P
0 g
s
.508 0 0 r
.55284 .49604 m .53941 .48692 L .53462 .48528 L p F P
0 g
s
0 .44 .755 r
.55664 .49567 m .57127 .50068 L p .56591 .49917 L F P
0 g
s
0 .44 .755 r
.54389 .49298 m .55664 .49567 L p .56591 .49917 L F P
0 g
s
.149 0 0 r
.55664 .49567 m .54389 .49298 L p .53851 .49201 L F P
0 g
s
.61 .467 .662 r
.5258 .46837 m .52986 .48653 L .52767 .48713 L p F P
0 g
s
.552 .43 .666 r
.52986 .48653 m .5258 .46837 L .52638 .46811 L p F P
0 g
s
.552 .43 .666 r
.52638 .46811 m .53176 .48567 L .52986 .48653 L p F P
0 g
s
.479 .386 .672 r
.53176 .48567 m .52638 .46811 L .52685 .46778 L p F P
0 g
s
.479 .386 .672 r
.52685 .46778 m .53332 .4846 L .53176 .48567 L p F P
0 g
s
.572 .477 .708 r
.5373 .51127 m .52986 .48653 L .53176 .48567 L p F P
0 g
s
.497 0 0 r
.54389 .49298 m .53462 .48528 L .52933 .48427 L p F P
0 g
s
.497 0 0 r
.52933 .48427 m .53397 .4911 L .54389 .49298 L p F P
0 g
s
.149 0 0 r
.54037 .49259 m .55664 .49567 L p .55219 .49477 L F P
0 g
s
.149 0 0 r
.53397 .4911 m .54037 .49259 L p .55219 .49477 L F P
0 g
s
.705 .566 .695 r
.5215 .51335 m .52292 .48745 L .52533 .48744 L p F P
0 g
s
.705 .566 .695 r
.52533 .48744 m .52699 .51332 L .5215 .51335 L p F P
0 g
s
.666 .536 .697 r
.52699 .51332 m .52533 .48744 L .52767 .48713 L p F P
0 g
s
.666 .536 .697 r
.52767 .48713 m .53234 .51262 L .52699 .51332 L p F P
0 g
s
.622 .507 .701 r
.53234 .51262 m .52767 .48713 L .52986 .48653 L p F P
0 g
s
.622 .507 .701 r
.52986 .48653 m .5373 .51127 L .53234 .51262 L p F P
0 g
s
.704 .536 .659 r
.52442 .46865 m .52533 .48744 L .52292 .48745 L p F P
0 g
s
.66 .502 .66 r
.52533 .48744 m .52442 .46865 L .52514 .46856 L p F P
0 g
s
.66 .502 .66 r
.52514 .46856 m .52767 .48713 L .52533 .48744 L p F P
0 g
s
.61 .467 .662 r
.52767 .48713 m .52514 .46856 L .5258 .46837 L p F P
0 g
s
.439 0 0 r
.52378 .48395 m .52356 .4905 L .53397 .4911 L p F P
0 g
s
.11 0 0 r
.54037 .49259 m .53397 .4911 L .52356 .4905 L p F P
0 g
s
.11 0 0 r
.52356 .4905 m .52327 .4916 L .54037 .49259 L p F P
0 g
s
.286 .75 .947 r
.54037 .49259 m .52327 .4916 L p .52895 .49048 L F P
0 g
s
.38 .33 .677 r
.52719 .4674 m .53443 .48335 L .53332 .4846 L p F P
0 g
s
.232 .251 .675 r
.53443 .48335 m .52719 .4674 L .52738 .46698 L p F P
0 g
s
.232 .251 .675 r
.52738 .46698 m .53507 .48199 L .53443 .48335 L p F P
0 g
s
0 .127 .645 r
.53507 .48199 m .52738 .46698 L .52741 .46655 L p F P
0 g
s
.717 .184 0 r
.54346 .48912 m .5325 .47673 L .53073 .47577 L p F P
0 g
s
.717 .184 0 r
.53073 .47577 m .53941 .48692 L .54346 .48912 L p F P
0 g
s
.439 0 0 r
.53397 .4911 m .52933 .48427 L .52378 .48395 L p F P
0 g
s
.286 .75 .947 r
.54842 .48663 m .54037 .49259 L p .52895 .49048 L F P
0 g
s
.254 .708 .955 r
.55664 .49567 m .54037 .49259 L p .54767 .49228 L F P
0 g
s
.254 .708 .955 r
.57266 .49124 m .55664 .49567 L p .54767 .49228 L F P
0 g
s
0 .44 .755 r
.57127 .50068 m .55284 .49604 L p .54938 .49452 L F P
0 g
s
0 .457 .837 r
.56039 .50013 m .55284 .49604 L p .55901 .49831 L F P
0 g
s
0 .457 .837 r
.58356 .50735 m .56039 .50013 L p .55901 .49831 L F P
0 g
s
0 0 .243 r
.52444 .45735 m .52729 .46612 L p .5272 .46589 L F P
0 g
s
.892 .511 .055 r
.52729 .46612 m .52444 .45735 L .5244 .45729 L p F P
0 g
s
.722 .724 .851 r
.59039 .30716 m .61509 .27775 L .58707 .26681 L p F P
0 g
s
.722 .724 .851 r
.58707 .26681 m .5699 .29932 L .59039 .30716 L p F P
0 g
s
.083 0 .233 r
.52141 .4654 m .52369 .45725 L .52364 .4573 L p F P
0 g
s
.083 0 .233 r
.52364 .4573 m .521 .46575 L .52141 .4654 L p F P
0 g
s
.919 .531 .245 r
.52701 .46572 m .5244 .45729 L .52435 .45725 L p F P
0 g
s
.892 .511 .055 r
.5244 .45729 m .52701 .46572 L .52729 .46612 L p F P
0 g
s
.31 .041 .335 r
.52194 .4651 m .52376 .45721 L .52369 .45725 L p F P
0 g
s
.31 .041 .335 r
.52369 .45725 m .52141 .4654 L .52194 .4651 L p F P
0 g
s
.919 .531 .245 r
.52435 .45725 m .52659 .46537 L .52701 .46572 L p F P
0 g
s
.873 .502 .341 r
.52659 .46537 m .52435 .45725 L .52428 .45721 L p F P
0 g
s
.452 .166 .386 r
.52257 .46487 m .52384 .45719 L .52376 .45721 L p F P
0 g
s
.452 .166 .386 r
.52376 .45721 m .52194 .4651 L .52257 .46487 L p F P
0 g
s
.873 .502 .341 r
.52428 .45721 m .52606 .46507 L .52659 .46537 L p F P
0 g
s
.814 .462 .389 r
.52606 .46507 m .52428 .45721 L .5242 .45718 L p F P
0 g
s
.552 .252 .413 r
.52326 .46473 m .52393 .45717 L .52384 .45719 L p F P
0 g
s
.552 .252 .413 r
.52384 .45719 m .52257 .46487 L .52326 .46473 L p F P
0 g
s
.814 .462 .389 r
.5242 .45718 m .52542 .46485 L .52606 .46507 L p F P
0 g
s
.755 .418 .414 r
.52542 .46485 m .5242 .45718 L .52411 .45717 L p F P
0 g
s
.629 .317 .424 r
.52399 .46468 m .52402 .45716 L .52393 .45717 L p F P
0 g
s
.629 .317 .424 r
.52393 .45717 m .52326 .46473 L .52399 .46468 L p F P
0 g
s
.755 .418 .414 r
.52411 .45717 m .52472 .46472 L .52542 .46485 L p F P
0 g
s
.694 .371 .425 r
.52472 .46472 m .52411 .45717 L .52402 .45716 L p F P
0 g
s
.694 .371 .425 r
.52402 .45716 m .52399 .46468 L .52472 .46472 L p F P
0 g
s
.736 .763 .874 r
.39955 .60874 m .43213 .58688 L .4484 .5968 L p F P
0 g
s
.778 .757 .834 r
.47226 .56773 m .4484 .5968 L .43213 .58688 L p F P
0 g
s
.761 .727 .822 r
.4484 .5968 m .47226 .56773 L .48556 .57315 L p F P
0 g
s
.683 .743 .896 r
.63834 .29222 m .6717 .27194 L .64191 .253 L p F P
0 g
s
.683 .743 .896 r
.64191 .253 m .61509 .27775 L .63834 .29222 L p F P
0 g
s
.59 .605 .835 r
.60113 .59606 m .57678 .56722 L .58758 .5602 L p F P
0 g
s
.59 .605 .835 r
.58758 .5602 m .61708 .58597 L .60113 .59606 L p F P
0 g
s
.687 .818 .945 r
.74362 .60533 m .70578 .60255 L .71934 .5802 L p F P
0 g
s
.687 .818 .945 r
.71934 .5802 m .7607 .5783 L .74362 .60533 L p F P
0 g
s
.784 .711 .783 r
.50017 .5411 m .48556 .57315 L .47226 .56773 L p F P
0 g
s
.758 .679 .774 r
.48556 .57315 m .50017 .5411 L .50944 .54345 L p F P
0 g
s
.746 .572 .66 r
.52056 .48716 m .52297 .46857 L .52369 .46866 L p F P
0 g
s
.746 .572 .66 r
.52369 .46866 m .52292 .48745 L .52056 .48716 L p F P
0 g
s
.704 .536 .659 r
.52292 .48745 m .52369 .46866 L .52442 .46865 L p F P
0 g
s
0 0 .077 r
.86359 .42666 m .86669 .45644 L .86229 .41218 L p F P
0 g
s
0 0 .077 r
.86229 .41218 m .85916 .38206 L .86359 .42666 L p F P
0 g
s
.787 .611 .662 r
.51836 .48659 m .5223 .46839 L .52297 .46857 L p F P
0 g
s
.787 .611 .662 r
.52297 .46857 m .52056 .48716 L .51836 .48659 L p F P
0 g
s
.674 .769 .922 r
.34374 .6043 m .38259 .59308 L .39955 .60874 L p F P
0 g
s
.742 .783 .886 r
.43213 .58688 m .39955 .60874 L .38259 .59308 L p F P
0 g
s
.829 .654 .666 r
.51642 .48575 m .52171 .46813 L .5223 .46839 L p F P
0 g
s
.829 .654 .666 r
.5223 .46839 m .51836 .48659 L .51642 .48575 L p F P
0 g
s
.605 .566 .784 r
.56366 .57276 m .5486 .54086 L .55669 .53735 L p F P
0 g
s
.605 .566 .784 r
.55669 .53735 m .57678 .56722 L .56366 .57276 L p F P
0 g
s
.874 .706 .671 r
.51484 .48469 m .52123 .46781 L .52171 .46813 L p F P
0 g
s
.874 .706 .671 r
.52171 .46813 m .51642 .48575 L .51484 .48469 L p F P
0 g
s
.92 .769 .676 r
.51368 .48346 m .52088 .46743 L .52123 .46781 L p F P
0 g
s
.92 .769 .676 r
.52123 .46781 m .51484 .48469 L .51368 .48346 L p F P
0 g
s
.38 .33 .677 r
.53332 .4846 m .52685 .46778 L .52719 .4674 L p F P
0 g
s
.964 .845 .676 r
.513 .48211 m .52067 .46701 L .52088 .46743 L p F P
0 g
s
.964 .845 .676 r
.52088 .46743 m .51368 .48346 L .513 .48211 L p F P
0 g
s
0 .162 .7 r
.19075 .48942 m .21147 .51593 L .22257 .55358 L p F P
0 g
s
.254 .472 .877 r
.25174 .57089 m .22257 .55358 L .21147 .51593 L p F P
0 g
s
.594 .663 .887 r
.65011 .60753 m .61708 .58597 L .62896 .57416 L p F P
0 g
s
.594 .663 .887 r
.62896 .57416 m .66657 .59169 L .65011 .60753 L p F P
0 g
s
.738 .624 .732 r
.5215 .51335 m .51937 .54462 L .50944 .54345 L p F P
0 g
s
.705 .596 .731 r
.51937 .54462 m .5215 .51335 L .52699 .51332 L p F P
0 g
s
.705 .596 .731 r
.52699 .51332 m .52951 .54457 L .51937 .54462 L p F P
0 g
s
.67 .569 .732 r
.52951 .54457 m .52699 .51332 L .53234 .51262 L p F P
0 g
s
.67 .569 .732 r
.53234 .51262 m .5394 .5433 L .52951 .54457 L p F P
0 g
s
.77 .657 .737 r
.51614 .5127 m .50944 .54345 L .50017 .5411 L p F P
0 g
s
.738 .624 .732 r
.50944 .54345 m .51614 .5127 L .5215 .51335 L p F P
0 g
s
.632 .545 .737 r
.5394 .5433 m .53234 .51262 L .5373 .51127 L p F P
0 g
s
.632 .545 .737 r
.5373 .51127 m .5486 .54086 L .5394 .5433 L p F P
0 g
s
0 0 0 r
.51541 .47682 m .52141 .4654 L .521 .46575 L p F P
0 g
s
0 0 0 r
.521 .46575 m .51407 .478 L .51541 .47682 L p F P
0 g
s
.838 .371 0 r
.53387 .47789 m .52701 .46572 L .52659 .46537 L p F P
0 g
s
.743 .599 .697 r
.51614 .5127 m .52056 .48716 L .52292 .48745 L p F P
0 g
s
.743 .599 .697 r
.52292 .48745 m .5215 .51335 L .51614 .5127 L p F P
0 g
s
.188 0 .149 r
.51715 .47584 m .52194 .4651 L .52141 .4654 L p F P
0 g
s
.188 0 .149 r
.52141 .4654 m .51541 .47682 L .51715 .47584 L p F P
0 g
s
.838 .371 0 r
.52659 .46537 m .5325 .47673 L .53387 .47789 L p F P
0 g
s
.834 .386 .159 r
.5325 .47673 m .52659 .46537 L .52606 .46507 L p F P
0 g
s
.78 .634 .7 r
.51113 .51141 m .51836 .48659 L .52056 .48716 L p F P
0 g
s
.78 .634 .7 r
.52056 .48716 m .51614 .5127 L .51113 .51141 L p F P
0 g
s
.373 .025 .24 r
.51921 .4751 m .52257 .46487 L .52194 .4651 L p F P
0 g
s
.373 .025 .24 r
.52194 .4651 m .51715 .47584 L .51921 .4751 L p F P
0 g
s
.834 .386 .159 r
.52606 .46507 m .53073 .47577 L .5325 .47673 L p F P
0 g
s
.787 .365 .246 r
.53073 .47577 m .52606 .46507 L .52542 .46485 L p F P
0 g
s
.497 .139 .288 r
.5215 .47463 m .52326 .46473 L .52257 .46487 L p F P
0 g
s
.497 .139 .288 r
.52257 .46487 m .51921 .4751 L .5215 .47463 L p F P
0 g
s
.787 .365 .246 r
.52542 .46485 m .52864 .47505 L .53073 .47577 L p F P
0 g
s
.728 .328 .29 r
.52864 .47505 m .52542 .46485 L .52472 .46472 L p F P
0 g
s
.667 .673 .847 r
.4585 .26757 m .47635 .29986 L .49904 .29477 L p F P
0 g
s
.65 .666 .852 r
.47635 .29986 m .4585 .26757 L .43083 .27882 L p F P
0 g
s
.817 .675 .707 r
.50671 .50953 m .51642 .48575 L .51836 .48659 L p F P
0 g
s
.817 .675 .707 r
.51836 .48659 m .51113 .51141 L .50671 .50953 L p F P
0 g
s
.589 .219 .308 r
.52391 .47446 m .52399 .46468 L .52326 .46473 L p F P
0 g
s
.589 .219 .308 r
.52326 .46473 m .5215 .47463 L .52391 .47446 L p F P
0 g
s
.728 .328 .29 r
.52472 .46472 m .52633 .47461 L .52864 .47505 L p F P
0 g
s
.663 .28 .309 r
.52633 .47461 m .52472 .46472 L .52399 .46468 L p F P
0 g
s
.663 .28 .309 r
.52399 .46468 m .52391 .47446 L .52633 .47461 L p F P
0 g
s
.71 .708 .846 r
.5699 .29932 m .58707 .26681 L .5557 .26006 L p F P
0 g
s
.71 .708 .846 r
.5557 .26006 m .54706 .2945 L .5699 .29932 L p F P
0 g
s
.854 .956 .876 r
.82691 .55057 m .79753 .5682 L .80586 .53447 L p F P
0 g
s
.854 .956 .876 r
.80586 .53447 m .83682 .51274 L .82691 .55057 L p F P
0 g
s
.77 .657 .737 r
.50017 .5411 m .51113 .51141 L .51614 .5127 L p F P
0 g
s
.855 .723 .717 r
.50308 .50715 m .51484 .48469 L .51642 .48575 L p F P
0 g
s
.855 .723 .717 r
.51642 .48575 m .50671 .50953 L .50308 .50715 L p F P
0 g
s
.784 .711 .783 r
.47226 .56773 m .49196 .53768 L .50017 .5411 L p F P
0 g
s
.802 .694 .745 r
.51113 .51141 m .50017 .5411 L .49196 .53768 L p F P
0 g
s
.82 .939 .906 r
.27765 .28949 m .24605 .29887 L .22783 .33862 L p F P
0 g
s
.86 .996 .801 r
.20274 .35934 m .22783 .33862 L .24605 .29887 L p F P
0 g
s
.69 .791 .927 r
.37421 .27367 m .33984 .26086 L .31417 .28878 L p F P
0 g
s
.732 .875 .947 r
.27765 .28949 m .31417 .28878 L .33984 .26086 L p F P
0 g
s
.59 .521 .746 r
.5486 .54086 m .5373 .51127 L .54166 .50935 L p F P
0 g
s
.59 .521 .746 r
.54166 .50935 m .55669 .53735 L .5486 .54086 L p F P
0 g
s
.778 .757 .834 r
.43213 .58688 m .46124 .56083 L .47226 .56773 L p F P
0 g
s
.808 .748 .796 r
.49196 .53768 m .47226 .56773 L .46124 .56083 L p F P
0 g
s
.893 .781 .73 r
.50042 .50437 m .51368 .48346 L .51484 .48469 L p F P
0 g
s
.893 .781 .73 r
.51484 .48469 m .50308 .50715 L .50042 .50437 L p F P
0 g
s
.802 .694 .745 r
.49196 .53768 m .50671 .50953 L .51113 .51141 L p F P
0 g
s
.565 .551 .797 r
.57678 .56722 m .55669 .53735 L .56332 .53294 L p F P
0 g
s
.565 .551 .797 r
.56332 .53294 m .58758 .5602 L .57678 .56722 L p F P
0 g
s
.929 .85 .744 r
.49885 .50132 m .513 .48211 L .51368 .48346 L p F P
0 g
s
.929 .85 .744 r
.51368 .48346 m .50042 .50437 L .49885 .50132 L p F P
0 g
s
.482 .674 .949 r
.73215 .28636 m .7684 .28665 L .73776 .25412 L p F P
0 g
s
.482 .674 .949 r
.73776 .25412 m .7056 .25872 L .73215 .28636 L p F P
0 g
s
.833 .737 .757 r
.4852 .53334 m .50308 .50715 L .50671 .50953 L p F P
0 g
s
.833 .737 .757 r
.50671 .50953 m .49196 .53768 L .4852 .53334 L p F P
0 g
s
.539 .498 .759 r
.55669 .53735 m .54166 .50935 L .54521 .50693 L p F P
0 g
s
.003 .305 .798 r
.81927 .33529 m .8443 .35576 L .82371 .3132 L p F P
0 g
s
.003 .305 .798 r
.82371 .3132 m .79979 .29568 L .81927 .33529 L p F P
0 g
s
.682 .683 .844 r
.48964 .26045 m .49904 .29477 L .52302 .29295 L p F P
0 g
s
.667 .673 .847 r
.49904 .29477 m .48964 .26045 L .4585 .26757 L p F P
0 g
s
.808 .748 .796 r
.46124 .56083 m .4852 .53334 L .49196 .53768 L p F P
0 g
s
.697 .694 .844 r
.54706 .2945 m .5557 .26006 L .52263 .2579 L p F P
0 g
s
.697 .694 .844 r
.52263 .2579 m .52302 .29295 L .54706 .2945 L p F P
0 g
s
.742 .783 .886 r
.38259 .59308 m .41988 .57519 L .43213 .58688 L p F P
0 g
s
.791 .789 .85 r
.46124 .56083 m .43213 .58688 L .41988 .57519 L p F P
0 g
s
.574 .726 .946 r
.28835 .58062 m .32947 .58211 L .34374 .6043 L p F P
0 g
s
.674 .774 .926 r
.38259 .59308 m .34374 .6043 L .32947 .58211 L p F P
0 g
s
.539 .498 .759 r
.54521 .50693 m .56332 .53294 L .55669 .53735 L p F P
0 g
s
.556 .599 .852 r
.61708 .58597 m .58758 .5602 L .59553 .55203 L p F P
0 g
s
.556 .599 .852 r
.59553 .55203 m .62896 .57416 L .61708 .58597 L p F P
0 g
s
.682 .683 .844 r
.52302 .29295 m .52263 .2579 L .48964 .26045 L p F P
0 g
s
.661 .722 .894 r
.40339 .25441 m .43083 .27882 L .4585 .26757 L p F P
0 g
s
.656 .721 .896 r
.43083 .27882 m .40339 .25441 L .37421 .27367 L p F P
0 g
s
.864 .787 .774 r
.48021 .52826 m .50042 .50437 L .50308 .50715 L p F P
0 g
s
.864 .787 .774 r
.50308 .50715 m .4852 .53334 L .48021 .52826 L p F P
0 g
s
.478 .475 .776 r
.56332 .53294 m .54521 .50693 L .54778 .50412 L p F P
0 g
s
.628 .738 .926 r
.70578 .60255 m .66657 .59169 L .67675 .5741 L p F P
0 g
s
.628 .738 .926 r
.67675 .5741 m .71934 .5802 L .70578 .60255 L p F P
0 g
s
.681 .738 .894 r
.61509 .27775 m .64191 .253 L .60577 .23865 L p F P
0 g
s
.681 .738 .894 r
.60577 .23865 m .58707 .26681 L .61509 .27775 L p F P
0 g
s
.679 .793 .935 r
.33984 .26086 m .37421 .27367 L .40339 .25441 L p F P
0 g
s
.744 .66 .105 r
.18786 .38575 m .18471 .41584 L .18172 .46005 L p F P
0 g
s
0 0 .469 r
.19075 .48942 m .18172 .46005 L .18471 .41584 L p F P
0 g
s
.791 .789 .85 r
.41988 .57519 m .45304 .55274 L .46124 .56083 L p F P
0 g
s
.83 .79 .813 r
.4852 .53334 m .46124 .56083 L .45304 .55274 L p F P
0 g
s
.83 .79 .813 r
.45304 .55274 m .48021 .52826 L .4852 .53334 L p F P
0 g
s
.607 .735 .935 r
.6717 .27194 m .7056 .25872 L .6693 .23523 L p F P
0 g
s
.607 .735 .935 r
.6693 .23523 m .64191 .253 L .6717 .27194 L p F P
0 g
s
.89 .847 .795 r
.47722 .52267 m .49885 .50132 L .50042 .50437 L p F P
0 g
s
.89 .847 .795 r
.50042 .50437 m .48021 .52826 L .47722 .52267 L p F P
0 g
s
.478 .475 .776 r
.54778 .50412 m .56815 .52781 L .56332 .53294 L p F P
0 g
s
.398 .449 .797 r
.56815 .52781 m .54778 .50412 L .54926 .50106 L p F P
0 g
s
.518 .537 .815 r
.58758 .5602 m .56332 .53294 L .56815 .52781 L p F P
0 g
s
.518 .537 .815 r
.56815 .52781 m .59553 .55203 L .58758 .5602 L p F P
0 g
s
0 0 0 r
.50824 .48708 m .51715 .47584 L .51541 .47682 L p F P
0 g
s
0 0 0 r
.51541 .47682 m .50426 .48932 L .50824 .48708 L p F P
0 g
s
.235 0 .005 r
.51298 .48539 m .51921 .4751 L .51715 .47584 L p F P
0 g
s
.235 0 .005 r
.51715 .47584 m .50824 .48708 L .51298 .48539 L p F P
0 g
s
.709 .199 .015 r
.53941 .48692 m .53073 .47577 L .52864 .47505 L p F P
0 g
s
.254 .472 .877 r
.21147 .51593 m .24233 .53731 L .25174 .57089 L p F P
0 g
s
.457 .628 .927 r
.28835 .58062 m .25174 .57089 L .24233 .53731 L p F P
0 g
s
.397 0 .086 r
.51824 .48433 m .5215 .47463 L .51921 .4751 L p F P
0 g
s
.397 0 .086 r
.51921 .4751 m .51298 .48539 L .51824 .48433 L p F P
0 g
s
.709 .199 .015 r
.52864 .47505 m .53462 .48528 L .53941 .48692 L p F P
0 g
s
.662 .178 .091 r
.53462 .48528 m .52864 .47505 L .52633 .47461 L p F P
0 g
s
.97 .873 .484 r
.86669 .45644 m .85761 .48597 L .85311 .44363 L p F P
0 g
s
.97 .873 .484 r
.85311 .44363 m .86229 .41218 L .86669 .45644 L p F P
0 g
s
.511 .063 .122 r
.52378 .48395 m .52391 .47446 L .5215 .47463 L p F P
0 g
s
.511 .063 .122 r
.5215 .47463 m .51824 .48433 L .52378 .48395 L p F P
0 g
s
.662 .178 .091 r
.52633 .47461 m .52933 .48427 L .53462 .48528 L p F P
0 g
s
.596 .133 .124 r
.52933 .48427 m .52633 .47461 L .52391 .47446 L p F P
0 g
s
.596 .133 .124 r
.52391 .47446 m .52378 .48395 L .52933 .48427 L p F P
0 g
s
.904 .914 .817 r
.47641 .51683 m .49846 .49814 L .49885 .50132 L p F P
0 g
s
.904 .914 .817 r
.49885 .50132 m .47722 .52267 L .47641 .51683 L p F P
0 g
s
.398 .449 .797 r
.54926 .50106 m .57095 .52219 L .56815 .52781 L p F P
0 g
s
.289 .417 .818 r
.57095 .52219 m .54926 .50106 L .54955 .49788 L p F P
0 g
s
.674 .774 .926 r
.32947 .58211 m .37184 .5756 L .38259 .59308 L p F P
0 g
s
.745 .804 .9 r
.41988 .57519 m .38259 .59308 L .37184 .5756 L p F P
0 g
s
.773 .883 .925 r
.79753 .5682 m .7607 .5783 L .76724 .54937 L p F P
0 g
s
.773 .883 .925 r
.76724 .54937 m .80586 .53447 L .79753 .5682 L p F P
0 g
s
.848 .838 .836 r
.44805 .5438 m .47722 .52267 L .48021 .52826 L p F P
0 g
s
.848 .838 .836 r
.48021 .52826 m .45304 .55274 L .44805 .5438 L p F P
0 g
s
.573 .666 .901 r
.66657 .59169 m .62896 .57416 L .63614 .56111 L p F P
0 g
s
.573 .666 .901 r
.63614 .56111 m .67675 .5741 L .66657 .59169 L p F P
0 g
s
.461 .526 .838 r
.59553 .55203 m .56815 .52781 L .57095 .52219 L p F P
0 g
s
.886 .977 .829 r
.47782 .511 m .49925 .49499 L .49846 .49814 L p F P
0 g
s
.886 .977 .829 r
.49846 .49814 m .47641 .51683 L .47782 .511 L p F P
0 g
s
.289 .417 .818 r
.54955 .49788 m .57157 .51634 L .57095 .52219 L p F P
0 g
s
.137 .373 .829 r
.57157 .51634 m .54955 .49788 L .54865 .49473 L p F P
0 g
s
.665 .723 .892 r
.43905 .23963 m .4585 .26757 L .48964 .26045 L p F P
0 g
s
.661 .722 .894 r
.4585 .26757 m .43905 .23963 L .40339 .25441 L p F P
0 g
s
.678 .733 .892 r
.58707 .26681 m .60577 .23865 L .56514 .22975 L p F P
0 g
s
.678 .733 .892 r
.56514 .22975 m .5557 .26006 L .58707 .26681 L p F P
0 g
s
.745 .804 .9 r
.37184 .5756 m .41228 .56222 L .41988 .57519 L p F P
0 g
s
.8 .824 .87 r
.45304 .55274 m .41988 .57519 L .41228 .56222 L p F P
0 g
s
.8 .824 .87 r
.41228 .56222 m .44805 .5438 L .45304 .55274 L p F P
0 g
s
.86 .996 .801 r
.24605 .29887 m .22196 .31662 L .20274 .35934 L p F P
0 g
s
.728 .856 .428 r
.18786 .38575 m .20274 .35934 L .22196 .31662 L p F P
0 g
s
.518 .596 .872 r
.62896 .57416 m .59553 .55203 L .60022 .54303 L p F P
0 g
s
.461 .526 .838 r
.57095 .52219 m .60022 .54303 L .59553 .55203 L p F P
0 g
s
.518 .596 .872 r
.60022 .54303 m .63614 .56111 L .62896 .57416 L p F P
0 g
s
.732 .875 .947 r
.33984 .26086 m .30725 .25665 L .27765 .28949 L p F P
0 g
s
.763 .966 .929 r
.24605 .29887 m .27765 .28949 L .30725 .25665 L p F P
0 g
s
.791 .998 .803 r
.49925 .49499 m .47782 .511 L .48142 .50547 L p F P
0 g
s
.137 .373 .829 r
.54865 .49473 m .56997 .51053 L .57157 .51634 L p F P
0 g
s
.855 .89 .862 r
.44656 .53441 m .47641 .51683 L .47722 .52267 L p F P
0 g
s
.855 .89 .862 r
.47722 .52267 m .44805 .5438 L .44656 .53441 L p F P
0 g
s
.391 .516 .864 r
.60022 .54303 m .57095 .52219 L .57157 .51634 L p F P
0 g
s
.248 .552 .93 r
.7684 .28665 m .79979 .29568 L .76573 .25906 L p F P
0 g
s
.248 .552 .93 r
.76573 .25906 m .73776 .25412 L .7684 .28665 L p F P
0 g
s
0 0 .405 r
.8443 .35576 m .85916 .38206 L .83794 .3381 L p F P
0 g
s
0 0 .405 r
.83794 .3381 m .82371 .3132 L .8443 .35576 L p F P
0 g
s
.457 .628 .927 r
.24233 .53731 m .28087 .5518 L .28835 .58062 L p F P
0 g
s
.583 .718 .936 r
.32947 .58211 m .28835 .58062 L .28087 .5518 L p F P
0 g
s
.67 .726 .892 r
.47939 .23026 m .48964 .26045 L .52263 .2579 L p F P
0 g
s
.665 .723 .892 r
.48964 .26045 m .47939 .23026 L .43905 .23963 L p F P
0 g
s
.674 .73 .892 r
.5557 .26006 m .56514 .22975 L .52221 .2269 L p F P
0 g
s
.674 .73 .892 r
.52221 .2269 m .52263 .2579 L .5557 .26006 L p F P
0 g
s
0 0 .469 r
.18471 .41584 m .19388 .44712 L .19075 .48942 L p F P
0 g
s
.149 .285 .749 r
.21147 .51593 m .19075 .48942 L .19388 .44712 L p F P
0 g
s
.696 .81 .935 r
.7607 .5783 m .71934 .5802 L .72413 .55643 L p F P
0 g
s
.696 .81 .935 r
.72413 .55643 m .76724 .54937 L .7607 .5783 L p F P
0 g
s
.666 .791 .941 r
.37537 .23697 m .40339 .25441 L .43905 .23963 L p F P
0 g
s
.679 .793 .935 r
.40339 .25441 m .37537 .23697 L .33984 .26086 L p F P
0 g
s
.67 .726 .892 r
.52263 .2579 m .52221 .2269 L .47939 .23026 L p F P
0 g
s
.611 .747 .941 r
.64191 .253 m .6693 .23523 L .62493 .21734 L p F P
0 g
s
.611 .747 .941 r
.62493 .21734 m .60577 .23865 L .64191 .253 L p F P
0 g
s
.843 .942 .887 r
.4487 .52502 m .47782 .511 L .47641 .51683 L p F P
0 g
s
.843 .942 .887 r
.47641 .51683 m .44656 .53441 L .4487 .52502 L p F P
0 g
s
.696 .881 .967 r
.30725 .25665 m .33984 .26086 L .37537 .23697 L p F P
0 g
s
.391 .516 .864 r
.57157 .51634 m .60141 .53363 L .60022 .54303 L p F P
0 g
s
.302 .506 .889 r
.60141 .53363 m .57157 .51634 L .56997 .51053 L p F P
0 g
s
0 0 0 r
.50332 .49319 m .51298 .48539 L .50824 .48708 L p F P
0 g
s
.801 .86 .893 r
.40979 .54853 m .44656 .53441 L .44805 .5438 L p F P
0 g
s
.801 .86 .893 r
.44805 .5438 m .41228 .56222 L .40979 .54853 L p F P
0 g
s
.948 .931 .75 r
.85761 .48597 m .83682 .51274 L .83213 .47371 L p F P
0 g
s
.948 .931 .75 r
.83213 .47371 m .85311 .44363 L .85761 .48597 L p F P
0 g
s
.476 .598 .895 r
.63614 .56111 m .60022 .54303 L .60141 .53363 L p F P
0 g
s
.471 .7 .968 r
.7056 .25872 m .73776 .25412 L .69539 .22634 L p F P
0 g
s
.471 .7 .968 r
.69539 .22634 m .6693 .23523 L .7056 .25872 L p F P
0 g
s
.2 0 0 r
.51318 .49121 m .51824 .48433 L .51298 .48539 L p F P
0 g
s
.2 0 0 r
.51298 .48539 m .50332 .49319 L .51318 .49121 L p F P
0 g
s
.583 .718 .936 r
.28087 .5518 m .32391 .55842 L .32947 .58211 L p F P
0 g
s
.672 .779 .93 r
.37184 .5756 m .32947 .58211 L .32391 .55842 L p F P
0 g
s
.343 0 0 r
.52356 .4905 m .52378 .48395 L .51824 .48433 L p F P
0 g
s
.343 0 0 r
.51824 .48433 m .51318 .49121 L .52356 .4905 L p F P
0 g
s
.624 .74 .931 r
.71934 .5802 m .67675 .5741 L .68 .5555 L p F P
0 g
s
.624 .74 .931 r
.68 .5555 m .72413 .55643 L .71934 .5802 L p F P
0 g
s
.742 .824 .915 r
.36798 .55706 m .40979 .54853 L .41228 .56222 L p F P
0 g
s
.742 .824 .915 r
.41228 .56222 m .37184 .5756 L .36798 .55706 L p F P
0 g
s
.672 .779 .93 r
.32391 .55842 m .36798 .55706 L .37184 .5756 L p F P
0 g
s
.552 .671 .917 r
.67675 .5741 m .63614 .56111 L .63818 .54739 L p F P
0 g
s
.476 .598 .895 r
.60141 .53363 m .63818 .54739 L .63614 .56111 L p F P
0 g
s
.794 .979 .904 r
.45443 .51605 m .48142 .50547 L .47782 .511 L p F P
0 g
s
.794 .979 .904 r
.47782 .511 m .4487 .52502 L .45443 .51605 L p F P
0 g
s
.302 .506 .889 r
.56997 .51053 m .59895 .52425 L .60141 .53363 L p F P
0 g
s
.192 .495 .904 r
.59895 .52425 m .56997 .51053 L .56619 .50504 L p F P
0 g
s
.552 .671 .917 r
.63818 .54739 m .68 .5555 L .67675 .5741 L p F P
0 g
s
.728 .856 .428 r
.22196 .31662 m .20763 .34164 L .18786 .38575 L p F P
0 g
s
0 0 .195 r
.18471 .41584 m .18786 .38575 L .20763 .34164 L p F P
0 g
s
.149 .285 .749 r
.19388 .44712 m .21487 .47693 L .21147 .51593 L p F P
0 g
s
.366 .496 .858 r
.24233 .53731 m .21147 .51593 L .21487 .47693 L p F P
0 g
s
.79 .895 .918 r
.41265 .53475 m .4487 .52502 L .44656 .53441 L p F P
0 g
s
.79 .895 .918 r
.44656 .53441 m .40979 .54853 L .41265 .53475 L p F P
0 g
s
.429 .604 .92 r
.63818 .54739 m .60141 .53363 L .59895 .52425 L p F P
0 g
s
.763 .966 .929 r
.30725 .25665 m .2789 .26192 L .24605 .29887 L p F P
0 g
s
.691 .965 .735 r
.22196 .31662 m .24605 .29887 L .2789 .26192 L p F P
0 g
s
.689 .978 .893 r
.46352 .50797 m .48706 .5005 L .48142 .50547 L p F P
0 g
s
.689 .978 .893 r
.48142 .50547 m .45443 .51605 L .46352 .50797 L p F P
0 g
s
.653 .786 .945 r
.41913 .21857 m .43905 .23963 L .47939 .23026 L p F P
0 g
s
.666 .791 .941 r
.43905 .23963 m .41913 .21857 L .37537 .23697 L p F P
0 g
s
.067 .478 .89 r
.59293 .51535 m .56619 .50504 L .56039 .50013 L p F P
0 g
s
.192 .495 .904 r
.56619 .50504 m .59293 .51535 L .59895 .52425 L p F P
0 g
s
.618 .759 .946 r
.60577 .23865 m .62493 .21734 L .57482 .20624 L p F P
0 g
s
.618 .759 .946 r
.57482 .20624 m .56514 .22975 L .60577 .23865 L p F P
0 g
s
.951 .66 .215 r
.85916 .38206 m .86229 .41218 L .84094 .36852 L p F P
0 g
s
.951 .66 .215 r
.84094 .36852 m .83794 .3381 L .85916 .38206 L p F P
0 g
s
.865 .899 .856 r
.83682 .51274 m .80586 .53447 L .80101 .49983 L p F P
0 g
s
.865 .899 .856 r
.80101 .49983 m .83213 .47371 L .83682 .51274 L p F P
0 g
s
0 .227 .725 r
.79979 .29568 m .82371 .3132 L .78713 .2737 L p F P
0 g
s
0 .227 .725 r
.78713 .2737 m .76573 .25906 L .79979 .29568 L p F P
0 g
s
.521 .918 .843 r
.48706 .5005 m .46352 .50797 L .47558 .50117 L p F P
0 g
s
.067 .478 .89 r
.56039 .50013 m .58356 .50735 L .59293 .51535 L p F P
0 g
s
.733 .842 .932 r
.37143 .53827 m .41265 .53475 L .40979 .54853 L p F P
0 g
s
.733 .842 .932 r
.40979 .54853 m .36798 .55706 L .37143 .53827 L p F P
0 g
s
.366 .496 .858 r
.21487 .47693 m .24603 .50268 L .24233 .53731 L p F P
0 g
s
.499 .624 .905 r
.28087 .5518 m .24233 .53731 L .24603 .50268 L p F P
0 g
s
.64 .779 .948 r
.46886 .20687 m .47939 .23026 L .52221 .2269 L p F P
0 g
s
.653 .786 .945 r
.47939 .23026 m .46886 .20687 L .41913 .21857 L p F P
0 g
s
.762 .924 .942 r
.42086 .52154 m .45443 .51605 L .4487 .52502 L p F P
0 g
s
.762 .924 .942 r
.4487 .52502 m .41265 .53475 L .42086 .52154 L p F P
0 g
s
.532 .681 .933 r
.68 .5555 m .63818 .54739 L .63485 .53363 L p F P
0 g
s
.429 .604 .92 r
.59895 .52425 m .63485 .53363 L .63818 .54739 L p F P
0 g
s
.628 .77 .948 r
.56514 .22975 m .57482 .20624 L .52177 .20266 L p F P
0 g
s
.628 .77 .948 r
.52177 .20266 m .52221 .2269 L .56514 .22975 L p F P
0 g
s
.379 .615 .944 r
.63485 .53363 m .59895 .52425 L .59293 .51535 L p F P
0 g
s
.32 .804 .762 r
.49447 .49634 m .47558 .50117 L .49005 .49601 L p F P
0 g
s
0 .457 .837 r
.55284 .49604 m .57127 .50068 L .58356 .50735 L p F P
0 g
s
.65 .876 .982 r
.34868 .2284 m .37537 .23697 L .41913 .21857 L p F P
0 g
s
.696 .881 .967 r
.37537 .23697 m .34868 .2284 L .30725 .25665 L p F P
0 g
s
.664 .961 .936 r
.2789 .26192 m .30725 .25665 L .34868 .2284 L p F P
0 g
s
.782 .851 .903 r
.80586 .53447 m .76724 .54937 L .76238 .51983 L p F P
0 g
s
.782 .851 .903 r
.76238 .51983 m .80101 .49983 L .80586 .53447 L p F P
0 g
s
.47 .731 .983 r
.6693 .23523 m .69539 .22634 L .64325 .20511 L p F P
0 g
s
.47 .731 .983 r
.64325 .20511 m .62493 .21734 L .6693 .23523 L p F P
0 g
s
.64 .779 .948 r
.52221 .2269 m .52177 .20266 L .46886 .20687 L p F P
0 g
s
.175 .568 .935 r
.73776 .25412 m .76573 .25906 L .71818 .22767 L p F P
0 g
s
.175 .568 .935 r
.71818 .22767 m .69539 .22634 L .73776 .25412 L p F P
0 g
s
0 0 0 r
.50621 .49276 m .51318 .49121 L .50332 .49319 L p F P
0 g
s
0 0 0 r
.50332 .49319 m .49005 .49601 L .50621 .49276 L p F P
0 g
s
.668 .783 .934 r
.3277 .53427 m .37143 .53827 L .36798 .55706 L p F P
0 g
s
.668 .783 .934 r
.36798 .55706 m .32391 .55842 L .3277 .53427 L p F P
0 g
s
0 0 .195 r
.20763 .34164 m .20459 .37204 L .18471 .41584 L p F P
0 g
s
.112 .112 .578 r
.19388 .44712 m .18471 .41584 L .20459 .37204 L p F P
0 g
s
.499 .624 .905 r
.24603 .50268 m .28473 .52225 L .28087 .5518 L p F P
0 g
s
.594 .713 .927 r
.32391 .55842 m .28087 .5518 L .28473 .52225 L p F P
0 g
s
.016 0 0 r
.52327 .4916 m .52356 .4905 L .51318 .49121 L p F P
0 g
s
.016 0 0 r
.51318 .49121 m .50621 .49276 L .52327 .4916 L p F P
0 g
s
.62 .744 .935 r
.72413 .55643 m .68 .5555 L .67592 .53674 L p F P
0 g
s
.532 .681 .933 r
.63485 .53363 m .67592 .53674 L .68 .5555 L p F P
0 g
s
.594 .713 .927 r
.28473 .52225 m .3277 .53427 L .32391 .55842 L p F P
0 g
s
.701 .799 .926 r
.76724 .54937 m .72413 .55643 L .71951 .53229 L p F P
0 g
s
.701 .799 .926 r
.71951 .53229 m .76238 .51983 L .76724 .54937 L p F P
0 g
s
.991 .825 .583 r
.86229 .41218 m .85311 .44363 L .83214 .40192 L p F P
0 g
s
.991 .825 .583 r
.83214 .40192 m .84094 .36852 L .86229 .41218 L p F P
0 g
s
.71 .938 .96 r
.43413 .50957 m .46352 .50797 L .45443 .51605 L p F P
0 g
s
.71 .938 .96 r
.45443 .51605 m .42086 .52154 L .43413 .50957 L p F P
0 g
s
.62 .744 .935 r
.67592 .53674 m .71951 .53229 L .72413 .55643 L p F P
0 g
s
.379 .615 .944 r
.59293 .51535 m .6262 .5205 L .63485 .53363 L p F P
0 g
s
.328 .631 .961 r
.6262 .5205 m .59293 .51535 L .58356 .50735 L p F P
0 g
s
.691 .965 .735 r
.2789 .26192 m .25721 .27678 L .22196 .31662 L p F P
0 g
s
0 0 0 r
.20763 .34164 m .22196 .31662 L .25721 .27678 L p F P
0 g
s
.716 .856 .949 r
.38227 .52016 m .42086 .52154 L .41265 .53475 L p F P
0 g
s
.716 .856 .949 r
.41265 .53475 m .37143 .53827 L .38227 .52016 L p F P
0 g
s
.514 .694 .95 r
.67592 .53674 m .63485 .53363 L .6262 .5205 L p F P
0 g
s
.791 .328 0 r
.82371 .3132 m .83794 .3381 L .79989 .29722 L p F P
0 g
s
.791 .328 0 r
.79989 .29722 m .78713 .2737 L .82371 .3132 L p F P
0 g
s
.6 .859 .993 r
.40009 .20658 m .41913 .21857 L .46886 .20687 L p F P
0 g
s
.65 .876 .982 r
.41913 .21857 m .40009 .20658 L .34868 .2284 L p F P
0 g
s
.634 .932 .968 r
.45192 .49945 m .47558 .50117 L .46352 .50797 L p F P
0 g
s
.634 .932 .968 r
.46352 .50797 m .43413 .50957 L .45192 .49945 L p F P
0 g
s
.328 .631 .961 r
.58356 .50735 m .61251 .50865 L .6262 .5205 L p F P
0 g
s
.283 .65 .968 r
.61251 .50865 m .58356 .50735 L .57127 .50068 L p F P
0 g
s
.483 .765 .993 r
.62493 .21734 m .64325 .20511 L .5841 .19189 L p F P
0 g
s
.483 .765 .993 r
.5841 .19189 m .57482 .20624 L .62493 .21734 L p F P
0 g
s
.112 .112 .578 r
.20459 .37204 m .21343 .40528 L .19388 .44712 L p F P
0 g
s
.324 .363 .748 r
.21487 .47693 m .19388 .44712 L .21343 .40528 L p F P
0 g
s
.928 .852 .748 r
.85311 .44363 m .83213 .47371 L .81207 .43538 L p F P
0 g
s
.928 .852 .748 r
.81207 .43538 m .83214 .40192 L .85311 .44363 L p F P
0 g
s
.539 .903 .965 r
.47337 .49175 m .49005 .49601 L .47558 .50117 L p F P
0 g
s
.539 .903 .965 r
.47558 .50117 m .45192 .49945 L .47337 .49175 L p F P
0 g
s
.664 .785 .939 r
.34108 .51084 m .38227 .52016 L .37143 .53827 L p F P
0 g
s
.664 .785 .939 r
.37143 .53827 m .3277 .53427 L .34108 .51084 L p F P
0 g
s
.664 .961 .936 r
.34868 .2284 m .32537 .23001 L .2789 .26192 L p F P
0 g
s
.401 .804 .584 r
.25721 .27678 m .2789 .26192 L .32537 .23001 L p F P
0 g
s
.283 .65 .968 r
.57127 .50068 m .59438 .49872 L .61251 .50865 L p F P
0 g
s
.256 .675 .965 r
.59438 .49872 m .57127 .50068 L .55664 .49567 L p F P
0 g
s
.618 .748 .939 r
.71951 .53229 m .67592 .53674 L .66446 .51873 L p F P
0 g
s
.514 .694 .95 r
.6262 .5205 m .66446 .51873 L .67592 .53674 L p F P
0 g
s
.527 .916 .921 r
.32537 .23001 m .34868 .2284 L .40009 .20658 L p F P
0 g
s
.692 .865 .964 r
.40018 .50365 m .43413 .50957 L .42086 .52154 L p F P
0 g
s
.692 .865 .964 r
.42086 .52154 m .38227 .52016 L .40018 .50365 L p F P
0 g
s
.412 0 0 r
.76573 .25906 m .78713 .2737 L .73567 .23973 L p F P
0 g
s
.412 0 0 r
.73567 .23973 m .71818 .22767 L .76573 .25906 L p F P
0 g
s
.551 .832 .998 r
.45877 .19265 m .46886 .20687 L .52177 .20266 L p F P
0 g
s
.6 .859 .993 r
.46886 .20687 m .45877 .19265 L .40009 .20658 L p F P
0 g
s
.5 .711 .965 r
.66446 .51873 m .6262 .5205 L .61251 .50865 L p F P
0 g
s
.51 .8 .998 r
.57482 .20624 m .5841 .19189 L .52135 .18764 L p F P
0 g
s
.51 .8 .998 r
.52135 .18764 m .52177 .20266 L .57482 .20624 L p F P
0 g
s
.123 .59 .919 r
.69539 .22634 m .71818 .22767 L .65929 .2036 L p F P
0 g
s
.123 .59 .919 r
.65929 .2036 m .64325 .20511 L .69539 .22634 L p F P
0 g
s
.324 .363 .748 r
.21343 .40528 m .23362 .43847 L .21487 .47693 L p F P
0 g
s
.45 .516 .834 r
.24603 .50268 m .21487 .47693 L .23362 .43847 L p F P
0 g
s
.438 .856 .956 r
.49745 .48689 m .50621 .49276 L .49005 .49601 L p F P
0 g
s
.438 .856 .956 r
.49005 .49601 m .47337 .49175 L .49745 .48689 L p F P
0 g
s
.256 .675 .965 r
.55664 .49567 m .57266 .49124 L .59438 .49872 L p F P
0 g
s
0 0 0 r
.25721 .27678 m .24427 .30042 L .20763 .34164 L p F P
0 g
s
.105 0 .366 r
.20459 .37204 m .20763 .34164 L .24427 .30042 L p F P
0 g
s
.551 .832 .998 r
.52177 .20266 m .52135 .18764 L .45877 .19265 L p F P
0 g
s
.606 .709 .917 r
.30036 .4934 m .34108 .51084 L .3277 .53427 L p F P
0 g
s
.606 .709 .917 r
.3277 .53427 m .28473 .52225 L .30036 .4934 L p F P
0 g
s
.854 .842 .832 r
.83213 .47371 m .80101 .49983 L .7824 .46596 L p F P
0 g
s
.854 .842 .832 r
.7824 .46596 m .81207 .43538 L .83213 .47371 L p F P
0 g
s
.349 .801 .947 r
.5229 .48515 m .52327 .4916 L .50621 .49276 L p F P
0 g
s
.349 .801 .947 r
.50621 .49276 m .49745 .48689 L .5229 .48515 L p F P
0 g
s
.254 .708 .955 r
.54037 .49259 m .54842 .48663 L .57266 .49124 L p F P
0 g
s
.962 .647 .376 r
.83794 .3381 m .84094 .36852 L .80259 .32791 L p F P
0 g
s
.962 .647 .376 r
.80259 .32791 m .79989 .29722 L .83794 .3381 L p F P
0 g
s
.286 .75 .947 r
.52327 .4916 m .5229 .48515 L .54842 .48663 L p F P
0 g
s
.704 .788 .917 r
.76238 .51983 m .71951 .53229 L .70533 .50898 L p F P
0 g
s
.618 .748 .939 r
.66446 .51873 m .70533 .50898 L .71951 .53229 L p F P
0 g
s
.45 .516 .834 r
.23362 .43847 m .26346 .46868 L .24603 .50268 L p F P
0 g
s
.538 .623 .885 r
.28473 .52225 m .24603 .50268 L .26346 .46868 L p F P
0 g
s
.538 .623 .885 r
.26346 .46868 m .30036 .4934 L .28473 .52225 L p F P
0 g
s
.78 .819 .883 r
.80101 .49983 m .76238 .51983 L .74575 .4911 L p F P
0 g
s
.78 .819 .883 r
.74575 .4911 m .7824 .46596 L .80101 .49983 L p F P
0 g
s
.704 .788 .917 r
.70533 .50898 m .74575 .4911 L .76238 .51983 L p F P
0 g
s
.66 .867 .977 r
.42446 .48964 m .45192 .49945 L .43413 .50957 L p F P
0 g
s
.66 .867 .977 r
.43413 .50957 m .40018 .50365 L .42446 .48964 L p F P
0 g
s
.5 .711 .965 r
.61251 .50865 m .64597 .50239 L .66446 .51873 L p F P
0 g
s
.491 .731 .978 r
.64597 .50239 m .61251 .50865 L .59438 .49872 L p F P
0 g
s
.658 .787 .943 r
.36375 .48935 m .40018 .50365 L .38227 .52016 L p F P
0 g
s
.658 .787 .943 r
.38227 .52016 m .34108 .51084 L .36375 .48935 L p F P
0 g
s
.617 .753 .943 r
.70533 .50898 m .66446 .51873 L .64597 .50239 L p F P
0 g
s
.376 .838 .894 r
.38341 .20526 m .40009 .20658 L .45877 .19265 L p F P
0 g
s
.527 .916 .921 r
.40009 .20658 m .38341 .20526 L .32537 .23001 L p F P
0 g
s
.111 .625 .891 r
.64325 .20511 m .65929 .2036 L .59224 .18858 L p F P
0 g
s
.111 .625 .891 r
.59224 .18858 m .5841 .19189 L .64325 .20511 L p F P
0 g
s
.105 0 .366 r
.24427 .30042 m .24153 .33108 L .20459 .37204 L p F P
0 g
s
.322 .245 .612 r
.21343 .40528 m .20459 .37204 L .24153 .33108 L p F P
0 g
s
.401 .804 .584 r
.32537 .23001 m .30749 .24226 L .25721 .27678 L p F P
0 g
s
.068 0 .069 r
.24427 .30042 m .25721 .27678 L .30749 .24226 L p F P
0 g
s
.624 .861 .987 r
.45397 .47893 m .47337 .49175 L .45192 .49945 L p F P
0 g
s
.624 .861 .987 r
.45192 .49945 m .42446 .48964 L .45397 .47893 L p F P
0 g
s
.491 .731 .978 r
.59438 .49872 m .6212 .48862 L .64597 .50239 L p F P
0 g
s
.491 .754 .988 r
.6212 .48862 m .59438 .49872 L .57266 .49124 L p F P
0 g
s
.836 .371 .083 r
.78713 .2737 m .79989 .29722 L .74612 .26201 L p F P
0 g
s
.836 .371 .083 r
.74612 .26201 m .73567 .23973 L .78713 .2737 L p F P
0 g
s
.941 .748 .613 r
.84094 .36852 m .83214 .40192 L .79468 .36318 L p F P
0 g
s
.941 .748 .613 r
.79468 .36318 m .80259 .32791 L .84094 .36852 L p F P
0 g
s
0 0 0 r
.30749 .24226 m .32537 .23001 L .38341 .20526 L p F P
0 g
s
.499 0 0 r
.71818 .22767 m .73567 .23973 L .67163 .21362 L p F P
0 g
s
.499 0 0 r
.67163 .21362 m .65929 .2036 L .71818 .22767 L p F P
0 g
s
.618 .708 .909 r
.32752 .46679 m .36375 .48935 L .34108 .51084 L p F P
0 g
s
.618 .708 .909 r
.34108 .51084 m .30036 .4934 L .32752 .46679 L p F P
0 g
s
.585 .847 .994 r
.48723 .47215 m .49745 .48689 L .47337 .49175 L p F P
0 g
s
.585 .847 .994 r
.47337 .49175 m .45397 .47893 L .48723 .47215 L p F P
0 g
s
.243 .752 .872 r
.44991 .18944 m .45877 .19265 L .52135 .18764 L p F P
0 g
s
.376 .838 .894 r
.45877 .19265 m .44991 .18944 L .38341 .20526 L p F P
0 g
s
.491 .754 .988 r
.57266 .49124 m .59131 .47821 L .6212 .48862 L p F P
0 g
s
.5 .779 .994 r
.59131 .47821 m .57266 .49124 L .54842 .48663 L p F P
0 g
s
.151 .678 .871 r
.5841 .19189 m .59224 .18858 L .52098 .18374 L p F P
0 g
s
.151 .678 .871 r
.52098 .18374 m .52135 .18764 L .5841 .19189 L p F P
0 g
s
.703 .776 .909 r
.74575 .4911 m .70533 .50898 L .6819 .48771 L p F P
0 g
s
.617 .753 .943 r
.64597 .50239 m .6819 .48771 L .70533 .50898 L p F P
0 g
s
.322 .245 .612 r
.24153 .33108 m .24951 .3662 L .21343 .40528 L p F P
0 g
s
.44 .417 .737 r
.23362 .43847 m .21343 .40528 L .24951 .3662 L p F P
0 g
s
.652 .787 .946 r
.39486 .47102 m .42446 .48964 L .40018 .50365 L p F P
0 g
s
.652 .787 .946 r
.40018 .50365 m .36375 .48935 L .39486 .47102 L p F P
0 g
s
.617 .759 .946 r
.6819 .48771 m .64597 .50239 L .6212 .48862 L p F P
0 g
s
.243 .752 .872 r
.52135 .18764 m .52098 .18374 L .44991 .18944 L p F P
0 g
s
.55 .827 .997 r
.52248 .46972 m .5229 .48515 L .49745 .48689 L p F P
0 g
s
.55 .827 .997 r
.49745 .48689 m .48723 .47215 L .52248 .46972 L p F P
0 g
s
.5 .779 .994 r
.54842 .48663 m .55781 .47178 L .59131 .47821 L p F P
0 g
s
.52 .804 .997 r
.55781 .47178 m .54842 .48663 L .5229 .48515 L p F P
0 g
s
.889 .781 .736 r
.83214 .40192 m .81207 .43538 L .7767 .39988 L p F P
0 g
s
.889 .781 .736 r
.7767 .39988 m .79468 .36318 L .83214 .40192 L p F P
0 g
s
.52 .804 .997 r
.5229 .48515 m .52248 .46972 L .55781 .47178 L p F P
0 g
s
.572 .626 .867 r
.29453 .43715 m .32752 .46679 L .30036 .4934 L p F P
0 g
s
.572 .626 .867 r
.30036 .4934 m .26346 .46868 L .29453 .43715 L p F P
0 g
s
.44 .417 .737 r
.24951 .3662 m .26771 .40264 L .23362 .43847 L p F P
0 g
s
.516 .534 .813 r
.26346 .46868 m .23362 .43847 L .26771 .40264 L p F P
0 g
s
.772 .787 .866 r
.7824 .46596 m .74575 .4911 L .71764 .46474 L p F P
0 g
s
.703 .776 .909 r
.6819 .48771 m .71764 .46474 L .74575 .4911 L p F P
0 g
s
.068 0 .069 r
.30749 .24226 m .29679 .26464 L .24427 .30042 L p F P
0 g
s
.331 .126 .443 r
.24153 .33108 m .24427 .30042 L .29679 .26464 L p F P
0 g
s
.516 .534 .813 r
.26771 .40264 m .29453 .43715 L .26346 .46868 L p F P
0 g
s
.832 .79 .811 r
.81207 .43538 m .7824 .46596 L .75021 .43472 L p F P
0 g
s
.832 .79 .811 r
.75021 .43472 m .7767 .39988 L .81207 .43538 L p F P
0 g
s
.904 .593 .448 r
.79989 .29722 m .80259 .32791 L .74833 .29291 L p F P
0 g
s
.904 .593 .448 r
.74833 .29291 m .74612 .26201 L .79989 .29722 L p F P
0 g
s
.772 .787 .866 r
.71764 .46474 m .75021 .43472 L .7824 .46596 L p F P
0 g
s
.645 .786 .949 r
.43298 .45695 m .45397 .47893 L .42446 .48964 L p F P
0 g
s
.645 .786 .949 r
.42446 .48964 m .39486 .47102 L .43298 .45695 L p F P
0 g
s
.151 0 0 r
.37057 .21542 m .38341 .20526 L .44991 .18944 L p F P
0 g
s
0 0 0 r
.38341 .20526 m .37057 .21542 L .30749 .24226 L p F P
0 g
s
.617 .759 .946 r
.6212 .48862 m .65013 .46969 L .6819 .48771 L p F P
0 g
s
.619 .764 .949 r
.65013 .46969 m .6212 .48862 L .59131 .47821 L p F P
0 g
s
.494 0 0 r
.65929 .2036 m .67163 .21362 L .59851 .1973 L p F P
0 g
s
.494 0 0 r
.59851 .1973 m .59224 .18858 L .65929 .2036 L p F P
0 g
s
.308 0 .192 r
.29679 .26464 m .30749 .24226 L .37057 .21542 L p F P
0 g
s
.631 .709 .903 r
.36529 .44397 m .39486 .47102 L .36375 .48935 L p F P
0 g
s
.631 .709 .903 r
.36375 .48935 m .32752 .46679 L .36529 .44397 L p F P
0 g
s
.699 .764 .902 r
.71764 .46474 m .6819 .48771 L .65013 .46969 L p F P
0 g
s
.796 .355 .199 r
.73567 .23973 m .74612 .26201 L .67902 .2349 L p F P
0 g
s
.796 .355 .199 r
.67902 .2349 m .67163 .21362 L .73567 .23973 L p F P
0 g
s
.639 .783 .951 r
.47615 .44802 m .48723 .47215 L .45397 .47893 L p F P
0 g
s
.639 .783 .951 r
.45397 .47893 m .43298 .45695 L .47615 .44802 L p F P
0 g
s
.331 .126 .443 r
.29679 .26464 m .29451 .29552 L .24153 .33108 L p F P
0 g
s
.454 .331 .622 r
.24951 .3662 m .24153 .33108 L .29451 .29552 L p F P
0 g
s
.619 .764 .949 r
.59131 .47821 m .6115 .45601 L .65013 .46969 L p F P
0 g
s
.622 .77 .951 r
.6115 .45601 m .59131 .47821 L .55781 .47178 L p F P
0 g
s
.318 0 0 r
.44309 .19824 m .44991 .18944 L .52098 .18374 L p F P
0 g
s
.151 0 0 r
.44991 .18944 m .44309 .19824 L .37057 .21542 L p F P
0 g
s
.883 .68 .622 r
.80259 .32791 m .79468 .36318 L .74185 .32982 L p F P
0 g
s
.883 .68 .622 r
.74185 .32982 m .74833 .29291 L .80259 .32791 L p F P
0 g
s
.43 0 0 r
.59224 .18858 m .59851 .1973 L .52069 .19203 L p F P
0 g
s
.43 0 0 r
.52069 .19203 m .52098 .18374 L .59224 .18858 L p F P
0 g
s
.632 .78 .952 r
.52202 .44481 m .52248 .46972 L .48723 .47215 L p F P
0 g
s
.632 .78 .952 r
.48723 .47215 m .47615 .44802 L .52202 .44481 L p F P
0 g
s
.622 .77 .951 r
.55781 .47178 m .568 .44753 L .6115 .45601 L p F P
0 g
s
.627 .775 .952 r
.568 .44753 m .55781 .47178 L .52248 .46972 L p F P
0 g
s
.603 .632 .853 r
.33823 .40999 m .36529 .44397 L .32752 .46679 L p F P
0 g
s
.603 .632 .853 r
.32752 .46679 m .29453 .43715 L .33823 .40999 L p F P
0 g
s
.318 0 0 r
.52098 .18374 m .52069 .19203 L .44309 .19824 L p F P
0 g
s
.627 .775 .952 r
.52248 .46972 m .52202 .44481 L .568 .44753 L p F P
0 g
s
.759 .758 .852 r
.75021 .43472 m .71764 .46474 L .67905 .4423 L p F P
0 g
s
.699 .764 .902 r
.65013 .46969 m .67905 .4423 L .71764 .46474 L p F P
0 g
s
.643 .712 .897 r
.41193 .42637 m .43298 .45695 L .39486 .47102 L p F P
0 g
s
.643 .712 .897 r
.39486 .47102 m .36529 .44397 L .41193 .42637 L p F P
0 g
s
.454 .331 .622 r
.29451 .29552 m .30112 .33231 L .24951 .3662 L p F P
0 g
s
.523 .458 .725 r
.26771 .40264 m .24951 .3662 L .30112 .33231 L p F P
0 g
s
.693 .752 .897 r
.67905 .4423 m .65013 .46969 L .6115 .45601 L p F P
0 g
s
.847 .721 .724 r
.79468 .36318 m .7767 .39988 L .72713 .36939 L p F P
0 g
s
.847 .721 .724 r
.72713 .36939 m .74185 .32982 L .79468 .36318 L p F P
0 g
s
.568 .552 .796 r
.31615 .37166 m .33823 .40999 L .29453 .43715 L p F P
0 g
s
.568 .552 .796 r
.29453 .43715 m .26771 .40264 L .31615 .37166 L p F P
0 g
s
.308 0 .192 r
.37057 .21542 m .36288 .23678 L .29679 .26464 L p F P
0 g
s
.472 .239 .479 r
.29451 .29552 m .29679 .26464 L .36288 .23678 L p F P
0 g
s
.463 .095 .254 r
.36288 .23678 m .37057 .21542 L .44309 .19824 L p F P
0 g
s
.523 .458 .725 r
.30112 .33231 m .31615 .37166 L .26771 .40264 L p F P
0 g
s
.839 .539 .481 r
.74612 .26201 m .74833 .29291 L .68058 .26596 L p F P
0 g
s
.839 .539 .481 r
.68058 .26596 m .67902 .2349 L .74612 .26201 L p F P
0 g
s
.806 .744 .795 r
.7767 .39988 m .75021 .43472 L .70552 .408 L p F P
0 g
s
.759 .758 .852 r
.67905 .4423 m .70552 .408 L .75021 .43472 L p F P
0 g
s
.733 .316 .258 r
.67163 .21362 m .67902 .2349 L .60227 .21794 L p F P
0 g
s
.733 .316 .258 r
.60227 .21794 m .59851 .1973 L .67163 .21362 L p F P
0 g
s
.806 .744 .795 r
.70552 .408 m .72713 .36939 L .7767 .39988 L p F P
0 g
s
.655 .717 .894 r
.46502 .41517 m .47615 .44802 L .43298 .45695 L p F P
0 g
s
.655 .717 .894 r
.43298 .45695 m .41193 .42637 L .46502 .41517 L p F P
0 g
s
.693 .752 .897 r
.6115 .45601 m .63175 .42519 L .67905 .4423 L p F P
0 g
s
.686 .741 .893 r
.63175 .42519 m .6115 .45601 L .568 .44753 L p F P
0 g
s
.631 .642 .843 r
.3926 .38896 m .41193 .42637 L .36529 .44397 L p F P
0 g
s
.631 .642 .843 r
.36529 .44397 m .33823 .40999 L .3926 .38896 L p F P
0 g
s
.742 .732 .842 r
.70552 .408 m .67905 .4423 L .63175 .42519 L p F P
0 g
s
.573 .191 .281 r
.43899 .21891 m .44309 .19824 L .52069 .19203 L p F P
0 g
s
.463 .095 .254 r
.44309 .19824 m .43899 .21891 L .36288 .23678 L p F P
0 g
s
.659 .262 .282 r
.59851 .1973 m .60227 .21794 L .52052 .21246 L p F P
0 g
s
.659 .262 .282 r
.52052 .21246 m .52069 .19203 L .59851 .1973 L p F P
0 g
s
.472 .239 .479 r
.36288 .23678 m .36124 .26783 L .29451 .29552 L p F P
0 g
s
.545 .392 .624 r
.30112 .33231 m .29451 .29552 L .36124 .26783 L p F P
0 g
s
.667 .724 .892 r
.52156 .41113 m .52202 .44481 L .47615 .44802 L p F P
0 g
s
.667 .724 .892 r
.47615 .44802 m .46502 .41517 L .52156 .41113 L p F P
0 g
s
.686 .741 .893 r
.568 .44753 m .57825 .41455 L .63175 .42519 L p F P
0 g
s
.677 .732 .892 r
.57825 .41455 m .568 .44753 L .52202 .44481 L p F P
0 g
s
.829 .623 .625 r
.74833 .29291 m .74185 .32982 L .676 .30416 L p F P
0 g
s
.829 .623 .625 r
.676 .30416 m .68058 .26596 L .74833 .29291 L p F P
0 g
s
.573 .191 .281 r
.52069 .19203 m .52052 .21246 L .43899 .21891 L p F P
0 g
s
.677 .732 .892 r
.52202 .44481 m .52156 .41113 L .57825 .41455 L p F P
0 g
s
.571 .32 .497 r
.36124 .26783 m .36288 .23678 L .43899 .21891 L p F P
0 g
s
.776 .487 .498 r
.67902 .2349 m .68058 .26596 L .60306 .24909 L p F P
0 g
s
.776 .487 .498 r
.60306 .24909 m .60227 .21794 L .67902 .2349 L p F P
0 g
s
.613 .572 .784 r
.37678 .3476 m .3926 .38896 L .33823 .40999 L p F P
0 g
s
.613 .572 .784 r
.33823 .40999 m .31615 .37166 L .37678 .3476 L p F P
0 g
s
.545 .392 .624 r
.36124 .26783 m .36599 .30593 L .30112 .33231 L p F P
0 g
s
.586 .493 .716 r
.31615 .37166 m .30112 .33231 L .36599 .30593 L p F P
0 g
s
.778 .706 .783 r
.72713 .36939 m .70552 .408 L .65036 .38755 L p F P
0 g
s
.742 .732 .842 r
.63175 .42519 m .65036 .38755 L .70552 .408 L p F P
0 g
s
.806 .672 .715 r
.74185 .32982 m .72713 .36939 L .6656 .34598 L p F P
0 g
s
.806 .672 .715 r
.6656 .34598 m .676 .30416 L .74185 .32982 L p F P
0 g
s
.586 .493 .716 r
.36599 .30593 m .37678 .3476 L .31615 .37166 L p F P
0 g
s
.657 .654 .836 r
.45477 .37553 m .46502 .41517 L .41193 .42637 L p F P
0 g
s
.657 .654 .836 r
.41193 .42637 m .3926 .38896 L .45477 .37553 L p F P
0 g
s
.724 .708 .835 r
.65036 .38755 m .63175 .42519 L .57825 .41455 L p F P
0 g
s
.778 .706 .783 r
.65036 .38755 m .6656 .34598 L .72713 .36939 L p F P
0 g
s
.647 .382 .504 r
.43811 .25006 m .43899 .21891 L .52052 .21246 L p F P
0 g
s
.571 .32 .497 r
.43899 .21891 m .43811 .25006 L .36124 .26783 L p F P
0 g
s
.714 .436 .504 r
.60227 .21794 m .60306 .24909 L .52047 .24364 L p F P
0 g
s
.714 .436 .504 r
.52047 .24364 m .52052 .21246 L .60227 .21794 L p F P
0 g
s
.615 .443 .625 r
.36599 .30593 m .36124 .26783 L .43811 .25006 L p F P
0 g
s
.681 .67 .832 r
.52114 .37069 m .52156 .41113 L .46502 .41517 L p F P
0 g
s
.681 .67 .832 r
.46502 .41517 m .45477 .37553 L .52114 .37069 L p F P
0 g
s
.724 .708 .835 r
.57825 .41455 m .58768 .3748 L .65036 .38755 L p F P
0 g
s
.703 .687 .832 r
.58768 .3748 m .57825 .41455 L .52156 .41113 L p F P
0 g
s
.777 .574 .625 r
.68058 .26596 m .676 .30416 L .60072 .28811 L p F P
0 g
s
.777 .574 .625 r
.60072 .28811 m .60306 .24909 L .68058 .26596 L p F P
0 g
s
.647 .382 .504 r
.52052 .21246 m .52047 .24364 L .43811 .25006 L p F P
0 g
s
.703 .687 .832 r
.52156 .41113 m .52114 .37069 L .58768 .3748 L p F P
0 g
s
.651 .593 .776 r
.44637 .33221 m .45477 .37553 L .3926 .38896 L p F P
0 g
s
.651 .593 .776 r
.3926 .38896 m .37678 .3476 L .44637 .33221 L p F P
0 g
s
.749 .672 .776 r
.6656 .34598 m .65036 .38755 L .58768 .3748 L p F P
0 g
s
.615 .443 .625 r
.43811 .25006 m .44064 .28903 L .36599 .30593 L p F P
0 g
s
.638 .526 .71 r
.37678 .3476 m .36599 .30593 L .44064 .28903 L p F P
0 g
s
.638 .526 .71 r
.44064 .28903 m .44637 .33221 L .37678 .3476 L p F P
0 g
s
.766 .63 .709 r
.676 .30416 m .6656 .34598 L .59542 .33136 L p F P
0 g
s
.766 .63 .709 r
.59542 .33136 m .60072 .28811 L .676 .30416 L p F P
0 g
s
.749 .672 .776 r
.58768 .3748 m .59542 .33136 L .6656 .34598 L p F P
0 g
s
.674 .487 .625 r
.44064 .28903 m .43811 .25006 L .52047 .24364 L p F P
0 g
s
.726 .53 .625 r
.60306 .24909 m .60072 .28811 L .52057 .28293 L p F P
0 g
s
.726 .53 .625 r
.52057 .28293 m .52047 .24364 L .60306 .24909 L p F P
0 g
s
.686 .616 .772 r
.5208 .32665 m .52114 .37069 L .45477 .37553 L p F P
0 g
s
.686 .616 .772 r
.45477 .37553 m .44637 .33221 L .5208 .32665 L p F P
0 g
s
.718 .642 .772 r
.59542 .33136 m .58768 .3748 L .52114 .37069 L p F P
0 g
s
.674 .487 .625 r
.52047 .24364 m .52057 .28293 L .44064 .28903 L p F P
0 g
s
.718 .642 .772 r
.52114 .37069 m .5208 .32665 L .59542 .33136 L p F P
0 g
s
.684 .558 .707 r
.44637 .33221 m .44064 .28903 L .52057 .28293 L p F P
0 g
s
.684 .558 .707 r
.52057 .28293 m .5208 .32665 L .44637 .33221 L p F P
0 g
s
.726 .593 .706 r
.60072 .28811 m .59542 .33136 L .5208 .32665 L p F P
0 g
s
.726 .593 .706 r
.5208 .32665 m .52057 .28293 L .60072 .28811 L p F P
0 g
s
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{263.375, 134.813},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006;0000bR000`400?l00000o`00003oo`3ooon<
0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<
0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<
0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<
0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<
0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<
0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<
0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<
0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00@@3oool900000?l0oooo@P3o
ool003l0oooo0`00000306/00005000000250000Q@000000002G2`000`00003o0?ooocl0oooo000l
0?ooo`<0000000<0A@000000000000002@000000102G2`00000000000000003o0?ooocd0oooo000j
0?ooo`80000000H03P00000000000000000008DaF`00000309a5G`060000002_E5l0[eAO0:mDG`00
0000_f9K0`0000000`2N3@0000000000003o0?oooc/0oooo000i0?ooo`D000000`25<E/01@000000
W4EO09a5G`2LAEl0000000<0[eAO00@000000;mRF`2oHU/0_f9K1@00003o0?ooocX0oooo000h0?oo
o`<000000P1V5U4010000000QC5K08DaF`00000409a5G`050000002_E5l0[eAO0:mDG`000000102o
HU/00`000000d6eB0000000200000?l0oooo>@3oool003L0oooo0`00000206HFD@0500000025<E/0
QC5K08DaF`000000102LAEl00`000000[eAO0:mDG`020:mDG`050000002oHU/0_f9K0;mRF`000000
0P3@KE8300000?l0oooo>03oool003H0oooo00<0000003P0?@0000000`1V5U401@000000QC5K08Da
F`25<E/0000000@0W4EO00<000000:mDG`2_E5l00P2_E5l01@000000_f9K0;mRF`2oHU/0000000<0
d6eB00<000000=if?`000000o`3ooolg0?ooo`00=@3oool20000000303P0?@000000IQIA0080IQIA
00<0000008DaF`25<E/03P0000020;mRF`050000003@KE80d6eB0000003NMSl00P00003o0?ooocH0
oooo000d0?ooo`8000000P0h03d00`000000IQIA06HFD@040000000309MIQ@000000YfN700<0YfN7
00<000000;McQ`2gLhL00P2gLhL00`000000aGn60<EoQP0300000080d6eB00<000000=if?`3NMSl0
0P00003o0?ooocD0oooo000d0?ooo`80000000<0>00m0000001V5U400P0000000`22BH<0000009MI
Q@0209MIQ@030000002WIhL0YfN70080YfN700<000000;McQ`2gLhL00P2gLhL00`000000aGn60<Eo
QP020<EoQP<0000000H0d6eB0000003NMSl000000>Ag6`00003o0?oooc@0oooo000c0?ooo`040000
000001L0000003P0?@8000000`22BH<01@000000UeV509MIQ@2GFHD0000000@0YfN700<000000;Mc
Q`2gLhL00P2gLhL00`000000aGn60<EoQP020<EoQP030000003CRh<0dh^3008000000P3NMSl20000
0?l0oooo=03oool003<0oooo1@0000040899P`050000002GFHD0UeV509MIQ@000000102WIhL00`00
0000]g>70;McQ`020;McQ`0300000035OhH0aGn60080aGn600@000000=>;P`3CRh<0dh^31P00003o
0?oooc<0oooo000b0?ooo`030000000001L000000080IC9m00<000000899P`22BH<00P22BH<01@00
0000UeV509MIQ@2GFHD0000000@0YfN700<000000;McQ`2gLhL00P2gLhL00`000000aGn60<EoQP02
0<EoQP050000003CRh<0dh^30=>;P`0000000P3SVGd300000?l0oooo<`3oool00380oooo0`000002
06DbO@0500000022BH<0PTV30899P`000000102GFHD00`000000YfN70:MWQ`020:MWQ`030000002g
LhL0]g>70080]g>700<000000<EoQP35OhH00P35OhH01@000000dh^30=>;P`3CRh<0000000<0hiUm
0`00003o0?oooc80oooo000a0?ooo`80000000D0>A1a0000001U<Wd0IC9m000000030899P`030000
002GFHD0UeV50080UeV500<000000:MWQ`2WIhL00P2WIhL00`000000]g>70;McQ`020;McQ`030000
0035OhH0aGn60080aGn600D000000=>;P`3CRh<0dh^3000000030>>IO@<00000o`3ooolb0?ooo`00
<@3oool010000000>A1a03T@L@00000206DbO@0500000022BH<0PTV30899P`000000102GFHD00`00
0000YfN70:MWQ`09000000<0aGn600D000000=>;P`3CRh<0dh^3000000030>>IO@040000003bYg80
000000000?l0oooo<@3oool00340oooo00@0000003T@L@0i474000000P1U<Wd01@000000PTV30899
P`22BH<000000080UeV51@0000020:ajW0030000002iQ9`0^HBL0080^HBL00H000000<F>W0000000
00000<EoQP0000030=>;P`060000003SVGd0hiUm0>>IO@000000lZMb0P00001Y0?ooo`L00000`03o
ool00300oooo0P00000203T@L@040000001U<Wd0IC9m00000080PTV31000000209i_W0030000002/
NY`0[7ZL0080[7ZL00<000000;V4W02iQ9`00P2iQ9`010000000aHjL0<F>W035SY`300000080dh^3
00H000000>>IO@3SVGd0hiUm0000003bYg83000006<0oooo100000000`0]000000000000000;0000
0;X0oooo000`0?ooo`8000000P0i47400`000000IC9m06DbO@030000000308iSW0000000WVnL00<0
WVnL00<000000:ajW02/NY`00P2/NY`00`000000^HBL0;V4W0020;V4W00300000035SY`0aHjL0080
aHjL00<000000=6JW00000000P0000030>>IO@050000003bYg8000000?^cF@000000G`3oool90000
00<0Wdm[00<000000;9NJ`2bGV/01@2bGV/00`000000a6aW0000000300000;H0oooo000`0?ooo`80
00000P0i474010000000IC9m00000000000308iSW0030000002NKi`0WVnL0080WVnL00<000000:aj
W02/NY`00P2/NY`00`000000^HBL0;V4W0020;V4W00300000035SY`0aHjL0080aHjL00@000000=6J
W03AVY`0dIZL0`0000001P3SVGd000000?:WLP000000nk=I000005`0oooo1P0000000`27>fL00000
09m?J`0609m?J`030000002bGV/0/Ui[00D0/Ui[00<000000<A/I`34K6L00`34K6L400000;80oooo
000_0?ooo`<000000P0i47420000000307UEV`000000SV>L0080SV>L00<0000009i_W02NKi`00P2N
Ki`00`000000[7ZL0:ajW0020:ajW0030000002iQ9`0^HBL0080^HBL00<000000<F>W035SY`00P35
SY`01@000000dIZL0=6JW03AVY`000000080gZNK0P0000000`3bYg8000000?^cF@02000005P0oooo
1@00000508LkI`030000002OCf/0Wdm[0080Wdm[200000040;9NJ`0300000034K6L0a6aW00@0a6aW
00@000000=EiG`0000000000/03oool002l0oooo00<000000000EP0000000P0i47401@000000NEFK
07UEV`1iEI/000000080SV>L00<0000009i_W02NKi`00P2NKi`00`000000[7ZL0:ajW0020:ajW003
0000002iQ9`0^HBL0080^HBL00<000000<F>W035SY`00P35SY`01@000000dIZL0=6JW03AVY`00000
00<0gZNK0`0000000`3k/eT000000000001F0?ooo`<0000000<0J21N00000027>fL01@27>fL60000
00<0[7ZL00<000000;V5W02iQI`00P2iQI`6000000D0a6aW00@000000=EiG`3ENEl0eGUO0P00002^
0?ooo`00;`3oool00`000000001F00000003000000<0NEFK00@0000008iSW02>Hi`000001@2NKi`0
1@000000[7ZL0:ajW02/NY`0000000@0^HBL00D000000<F>W035SY`0aHjL000000040=6JW0040000
003NYi/0gZNK0=jWV`<0000000<0nk=I000000000000E@3oool300000080J21N00<0000008LkI`27
>fL0100000000`2MKY`000000:ajW0060:ajW0030000002iQI`0^HFL00H0^HFL00<000000<J@W000
00000`0000020<A/I`030000003ENEl0eGUO0080eGUO0P00002/0?ooo`00;`3oool00`000000001F
0000000205a2V@050000001iEI/0NEFK07UEV`0000000`2>Hi`00`000000WVnL09i_W00209i_W005
0000002/NY`0[7ZL0:ajW0000000102iQ9`01@000000aHjL0<F>W035SY`0000000@0dIZL00@00000
0=jWV`3NYi/000000P3/]iT010000000nk=I00000000001C0?ooo`<000000`1X85h4000000D0WFjL
00<000000:ajW02/NY`01@2/NY`00`000000^HFL0;V5W0060;V5W00300000036T9`0aY2L00<0aY2L
100000040=EiG`800000Z`3oool002l0oooo0`00000305a2V@040000001iEI/0NEFK000000<0SV>L
00<0000009i_W02NKi`00P2NKi`01@000000[7ZL0:ajW02/NY`0000000@0^HBL00D000000<F>W035
SY`0aHjL000000030=6JW0050000003NYi/0gZNK0=jWV`0000000P3/]iT400000580oooo00<00000
03T0C@0000000P1X85h20000000308]QW0000000WFjL00H0WFjL00<000000:ajW02/NY`01@2/NY`0
0`000000^HFL0;V5W0050;V5W00300000036T9`0aY2L00H0aY2L00@000000=>LW000000000000P3E
NEl010000000iHA?00000000002Y0?ooo`00;`3oool3000000<0G4:I00@0000007UEV`1iEI/00000
0`2>Hi`00`000000WVnL09i_W00209i_W0050000002/NY`0[7ZL0:ajW0000000102iQ9`01@000000
aHjL0<F>W035SY`0000000<0dIZL00D000000=jWV`3NYi/0gZNK000000020>bgV@@00000D@3oool0
10000000>@1=0000001X85h2000000<0Rf6L00<0000009e^W02MKY`01@2MKY`00`000000[7ZL0:aj
W0050:ajW0030000002iQI`0^HFL00D0^HFL00<000000<J@W036T9`01P36T9`010000000dibL0=>L
W03CW9`2000000050=EiG`000000iHA?0>F4C`000000Z03oool002l0oooo00@00000030UT`0`9I<0
00000P1L@YT010000000NEFK07UEV`00000308iSW0030000002NKi`0WVnL00d000000P35SY`01@00
0000dIZL0=6JW03AVY`0000000<0gZNK00<000000>bgV@0000000P3jbI<2000004l0oooo0P000002
03T0C@8000001@2;HI`00`000000WFjL09e^W00509e^W0030000002/NY`0[7ZL00D0[7ZL00<00000
0;V5W02iQI`01@2iQI`00`000000aY2L0<J@W0050<J@W0030000003CW9`0dibL00@0dibL0P000003
0>F4C`030000003oool0oooo0:D0oooo000_0?ooo`040000000`9I<0<2FC00000080G4:I00D00000
07UEV`1iEI/0NEFK0000000208iSW0<000000P2SOZ/01@000000[XJ[0:j6Z`2^QZ/0000000@0^Hn[
00<000000<BIZ`0000000P0000030=6JW0070000003NYi/0gZNK0=jWV`000000k;NI000000020?[9
T`800000CP3oool00`000000>@1=03T0C@03000000H0Rf6L00<0000009e^W02MKY`01@2MKY`00`00
0000[7ZL0:ajW0020:ajW0d000001P36T9`00`000000dibL0=>LW0050=>LW08000000P3UQ4l20000
0:H0oooo000_0?ooo`80000000D0<2FC0000001L@YT0G4:I0000000307UEV`@00000102SOZ/01@00
0000[XJ[0:j6Z`2^QZ/0000000@0^Hn[00@000000<BIZ`34VJ/0a9V[0`0000000`3AVY`000000=jW
V`020=jWV`030000003/]iT000000080n/VC0P00001=0?ooo`80000000D0>@1=0000001cDI/0Le6K
0000000608]QW0030000002MKY`0WFjL0080WFjL200000030:n>/`030000002iU[<0^IJc00D0^IJc
00<000000<>O]00000000`0000020<J@W0030000003CW9`0dibL00D0dibL00<000000>6[V`000000
0P0000000`3]RBd000000?ooo`2T0?ooo`00;`3oool200000003030UT`000000G4:I0080G4:I00@0
000007UEV`1iEI/000000P2FMJ`00`000000Xgj[0:=nZ`020:=nZ`050000002^QZ/0[XJ[0:j6Z`00
00000`2iSj/00`000000a9V[0<BIZ`020<BIZ`030000003>Xj`0cZ>/008000000P3NYi/020000000
k;NI0>bgV@000000n/VC0000003nfX800000C03oool4000000<0Le6K00<0000008]QW02;HI`0102;
HI`5000000<0Y8Fd00<000000:n>/`2_S[<01@2_S[<00`000000^IJc0;VF/`050;VF/`0300000033
Wk@0`ind00<0`ind100000050=>LW0030000003QZi/0hJ^K0080hJ^K0`00002U0?ooo`00;`3oool2
00000003030UT`000000G4:I0080G4:I1000000309Ie[0050000002SOZ/0Xgj[0:=nZ`0000000`2^
QZ/01@000000^Hn[0;V?Z`2iSj/0000000<0a9V[00<000000<jS[03>Xj`00P3>Xj`400000080k;NI
00D000000?[9T`000000o]Z20000001<0?ooo`8000001P1cDI/010000000Rf6L08]QW02;HI`30000
00L0Y8Fd00<000000:n>/`2_S[<01@2_S[<00`000000^IJc0;VF/`050;VF/`0300000033Wk@0`ind
00D0`ind00@000000<fY]@00000000000`3CW9`00`000000hJ^K0>6[V`030>6[V`030000003]RBd0
00000:@0oooo000_0?ooo`8000000P0`9I<010000000G4:I05a2V@00000308E[[P040000002FMJ`0
UWF/000000<0Xgj[00D000000:j6Z`2^QZ/0[XJ[000000030;V?Z`0500000034VJ/0a9V[0<BIZ`00
00000`3>Xj`01`000000fJn^0=V_[P3I[jh000000>bgV@0000000P3jbI<3000004/0oooo0`000006
07=AV`040000002;HI`0000000000080Ugfe00<000000:B5]02TQK@01@2TQK@00`000000[hjc0:n>
/`050:n>/`030000002iU[<0^IJc00D0^IJc00<000000<>O]033Wk@01033Wk@00`000000cJVe0<fY
]@020<fY]@80000000<0dibL0000003QZi/0103QZi/2000000030>f9;@000000oooo0:80oooo000_
0?ooo`040000000007l00000030UT`@000000`25Jjh010000000UWF/09Ie[00000030:=nZ`050000
002^QZ/0[XJ[0:j6Z`0000000`2iSj/01@000000a9V[0<BIZ`34VJ/0000000<0cZ>/00@000000=V_
[P3I[jh0fJn^0`0000020?[9T`<00000BP3oool200000003058kV0000000Le6K00D0Le6K0P000005
09Mm]@030000002TQK@0Y8Fd00@0Y8Fd00<000000:n>/`2_S[<01@2_S[<00`000000^IJc0;VF/`04
0;VF/`0300000033Wk@0`ind00D0`ind00<000000<fY]@3=ZKD0103=ZKD2000000D0hJ^K00@00000
0?2lV00000000000X`3oool002l0oooo0`0000000`0`9I<00000071P/002071P/00400000025Jjh0
QF^^00000080UWF/00<000000:=nZ`2SOZ/00P2SOZ/010000000[XJ[0:j6Z`0000030;V?Z`050000
0034VJ/0a9V[0<BIZ`0000000`3>Xj`010000000fJn^0=V_[P0000020>Bm[`040000003jbI<00000
0?kJPP800000BP3oool200000003058kV0000000Le6K00<0Le6K0`00000609Mm]@030000002TQK@0
Y8Fd00D0Y8Fd00<000000:n>/`2_S[<0102_S[<00`000000^IJc0;VF/`040;VF/`0300000033Wk@0
`ind00D0`ind00<000000<fY]@3=ZKD01@3=ZKD2000000<0hJ^K00<000000?2lV03`_9P00`00002R
0?ooo`00<03oool200000003030UT`000000L62`0080L62`00@0000008E[[P25Jjh000000`2FMJ`0
10000000Xgj[0:=nZ`2SOZ/8000000<0a9V[00@000000<jS[03>Xj`000000`3I[jh00`000000i;f_
0>Bm[`05000004X0oooo0P000003058kV0040000001cDI/0Le6K00000080R7Fh00<0000009Mm]@2G
OKD0102GOKD00`000000Y8Fd0:B5]0050:B5]0030000002_S[<0[hjc00@0[hjc00<000000;VF/`2i
U[<0102iU[<00`000000`ind0<>O]0050<>O]0030000003=ZKD0cJVe00@0cJVe00<000000=Ne]`3G
]KL00P0000000`3QZi/000000?2lV0020?2lV0800000XP3oool00300oooo00D000000000O`000000
De2a00000002071P/00300000025Jjh0QF^^01L000000`3I[jh00`000000i;f_000000020?3?/@<0
0000BP3oool00`000000DS^H058kV002058kV0<000001028MKP00`000000Ugfe09Mm]@0309Mm]@03
0000002TQK@0Y8Fd00D0Y8Fd00@000000:n>/`2_S[<0[hjc1`0000030;VF/`0300000033Wk@0`ind
00@0`ind00<000000<fY]@3=ZKD01@3=ZKD00`000000ekFg0=Ne]`020=Ne]`800000103`_9P20000
05X0oooo2000000o0?ooo`00<03oool010000000001o0000001CD;4=000000@0YfJ600<000000;Mc
QP2gLhH00P2gLhH=00000080l<na0`0000190?ooo`800000101B>iP00`000000R7Fh08Qe^00408Qe
^0030000002GOKD0Ugfe00@0Ugfe00@000000:B5]02TQK@0Y8Fd1`0000030;2K`P040000002hXL80
^:720;RQ`PL000000`33Wk@00`000000cJVe0<fY]@040<fY]@030000003G]KL0ekFg00@0ekFg00<0
00000?2lV03`_9P00P3`_9P2000005@0oooo1@0000000`1A000000000000000<000003T0oooo000a
0?ooo`030000000V>[00000000D000000`2FF8@00`000000YfJ60:MVQP060:MVQP030000002gLhH0
]g>600H0]g>600@000000<EoQ@35OhD0aGn52@00001:0?ooo`030000000K68h000000080DS^H00<0
000007A[^P0000001@28MKP00`000000Ugfe09Mm]@0409Mm]@@000000`2WUL<00`000000/9_20;2K
`P040;2K`P030000002hXL80^:7200@0^:7200@000000;nY``2oZL<0_jW3100000060<fY]@030000
003G]KL0ekFg00@0ekFg0P0000030?2lV0800000D03oool20000000302H000000000000000D00000
102B<4L00`000000Z4=80:Q3B0020:Q3B0T00000=@3oool00340oooo0P0000000`20Ah40000009IH
Q00609IHQ0030000002WIXH0YfJ60080YfJ62@0000040;McQP0300000035OhD0aGn500D0aGn500<0
00000=B;PP0000000P0000190?ooo`80000000L06aR>0000001B>iP0DS^H0000001dJkX0000000D0
R7Fh00<0000009Mm]@2GOKD00P2GOKD3000000H0YiG300<000000;2K`P2`Vl80102`Vl800`000000
^:720;RQ`P040;RQ`P030000002oZL<0_jW300@0_jW30`0000040<fY]@030000003G]KL0ekFg00<0
ekFg00H000000>74^P3Qa;X000000?2lV00000020?g?T0030000003oool0oooo04X0oooo1P000000
0`1f64400000098`A`07098`A`030000002X@dP0Z4=800L0Z4=800<000000;]@@P0000001P00000a
0?ooo`00;`3oool2000000<0P4N100<0000009IHQ02FF8@0102FF8@6000000@0[i:h00<000000;VJ
^02iV[P00P2iV[P6000000H0aGn500@000000=B;PP3DRh80e8^20P0000180?ooo`060000000K68h0
6aR>0000001B>iP000000`1dJkX00`000000R7Fh08Qe^00308Qe^0040000002GOKD00000000000<0
WY3500<000000:NE``2WUL<0102WUL<00`000000/9_20;2K`P030;2K`P030000002hXL80^:7200@0
^:7200<000000;nY``2oZL<00`2oZL<010000000a[750<Jaa@36/LD2000000030<fY]@000000ekFg
00D0ekFg00@000000>74^P3Qa;X0hLBj0P0000020?g?T0030000003oool0oooo04P0oooo10000004
07HH@AL000000`2kD487000002d0oooo000]0?ooo`8000001@20Ah400`000000UUR409IHQ0040000
00030:F;^@000000[i:h00L0[i:h00<000000;VJ^02iV[P01P2iV[P00`000000`Z:i000000030000
0080aGn500<000000=B;PP3DRh800`3DRh82000004H0oooo00@0000001/HSP0K68h06aR>0P000003
07A[^P0300000028MKP0R7Fh00<0R7Fh0P00000509j@a@030000002WUL<0YiG300@0YiG300<00000
0;2K`P2`Vl800`2`Vl800`000000^:720;RQ`P040;RQ`P030000002oZL<0_jW300<0_jW300<00000
0<Jaa@36/LD00`36/LD2000000D0ekFg00<000000>74^P3Qa;X00`3Qa;X01@000000oLn@0?g?T03m
ci00000004L0oooo0P0000000`1>0340000007HH@@0207HH@@<0000000<0TU5o0000002UHH401`2U
HH400`000000]Vn10;I_P@070;I_P@0300000036O7l0aWao00@0000000<0^e120000003;FS800P00
00000`1o000000000000000[0?ooo`00:`3oool3000000D0P4N1100000050:F;^@030000002_T[P0
[i:h00H0[i:h00<000000;VJ^02iV[P01P2iV[P00`000000`Z:i0<:R^@030<:R^@@000001@3DRh83
000004@0oooo00D0000001/HSP0K68h06aR>0000000507A[^P0700000028MKP0R7Fh08Qe^0000000
Th_80000000609j@a@030000002WUL<0YiG300<0YiG300<000000;2K`P2`Vl800`2`Vl800`000000
^:720;RQ`P030;RQ`P030000002oZL<0_jW300@0_jW300<000000<Jaa@36/LD00`36/LD010000000
cK[80000000000030=Ne]`030000003Qa;X0hLBj0080hLBj0P0000030?g?T0030000003oool0oooo
04<0oooo0P00000204h0<@030000001f6440000000800000102BDGl00`000000YF610:EQP@060:EQ
P@030000002fKh40]Vn100L0]Vn100<000000<IlO`36O7l01036O7l300000080beXb1000000Z0?oo
o`00:P3oool010000000HRmk068_N`0000030817P@80000000<0VXBk0000002URkT01`2URkT00`00
0000[i:h0:nB^0050:nB^0030000002iV[P0^IZh00D0^IZh00<000000<:R^@32X[T01@32X[T00`00
0000bjbk0<^/^`04000000050=B;PP000000i9Ul0>BIO0000000@`3oool20000000301/HSP000000
000000D0M6^j00@0000008Qe^000000000000P2CRlP00`000000WY3509j@a@0409j@a@030000002W
UL<0YiG300<0YiG300<000000;2K`P2`Vl800`2`Vl800`000000^:720;RQ`P030;RQ`P030000002o
ZL<0_jW300<0_jW300<000000<Jaa@36/LD01036/LD010000000cK[80<fjb03=^/P3000000D0hLBj
00@000000>_E_@000000oLn@0P0000130?ooo`8000000`1>0343000000L0TU5o00<000000:EQP@2U
HH400P2UHH49000000D0]Vn100<000000<IlO`36O7l01P36O7l3000000<0beXb0`00000X0?ooo`00
:03oool2000000<0HRmk00@000000817P@0000000000102JQ;/00`000000YH^i0:F;^@050:F;^@03
0000002_T[P0[i:h0080[i:h1`0000040;VJ^00300000032X[T0`Z:i00D0`Z:i00<000000<^/^`3;
[;/0103;[;/2000000<0i9Ul0P0000110?ooo`80000000@06aR>0000001JGkh000001@1dJkX00`00
0000Th_809>;b00309>;b0030000002NT<D0WY3500@0WY3500<000000:NE``2WUL<00`2WUL<00`00
0000/9_20;2K`P020;2K`P030000002hXL80^:7200<0^:7200<000000;nY``2oZL<00`2oZL<00`00
0000a[750<Jaa@030<Jaa@030000003=^/P0cK[800@0cK[800<000000>74^P3Qa;X00P3Qa;X01@00
0000jmFm0>_E_@000000oLn@00800000@@3oool2000000<0CP0a0P0000000`1h?GX00000099AO`05
099AO`L00000102/O9l00`000000^HNO0;V7W`020;V7W`P000001P36O7l00`000000eXUk0=J9N`02
00000080beXb0`00000W0?ooo`009`3oool00`000000HRmk068_N`03068_N`8000001P2JQ;/00`00
0000YH^i0:F;^@040:F;^@H000000`2^Z]<010000000]:kC0;B^d`2d[]<8000000<0`Z:i00<00000
0<^/^`3;[;/01P3;[;/3000000<0i9Ul00<000000?ooo`3oool0?P3oool00`000000001a00000002
05YO_P030000001dJkX0M6^j0080M6^j00<0000008N7c0000000102CRlP00`000000WY3509j@a@04
09j@a@030000002WUL<0YiG300<0YiG300<000000;2K`P2`Vl800P2`Vl800`000000^:720;RQ`P03
0;RQ`P030000002oZL<0_jW300<0_jW300<000000<Jaa@36/LD00`36/LD00`000000cK[80<fjb003
0<fjb0030000003DaL`0000000<0hLBj00L000000>_E_@3[eKd0jmFm0000003ifgD0000003l0oooo
0P0000000`0A0140000000000002000000<0N3ej00<00000099AO`2BDGl01000000209eaW`030000
002/O9l0[7bO00H0[7bO00<000000;V7W`2iQil01`2iQil00`000000aY:O0<JBW`0400000080aWao
00<000000=J9N`3FRG/00P3FRG/2000000040<]J<P000000eUhE000002H0oooo000V0?ooo`800000
0`1R;g/3000000P0VXBk00<000000:F;^@0000000`0000000`2YYm@000000:jZd`060:jZd`030000
002d[]<0]:kC00D0]:kC00@000000;Vce02i/m@0^K?D1@0000060<^/^`050000003D]kd0e;Nm0=Bg
_@0000000P3TVG`2000003l0oooo00<000000000L@0000000`1JGkh010000000M6^j07A[^P000002
08N7c0030000002CRlP0Th_800<0Th_800<0000009j@a@2NT<D0102NT<D00`000000YiG30:NE``0@
00000080_jW300<000000<Jaa@36/LD00`36/LD00`000000cK[80<fjb0030<fjb0040000003DaL`0
e<G<00000080hLBj00<000000>_E_@3[eKd00P3[eKd00`000000nM]e0000000n0?ooo`80000000<0
4@0A0000000000001@1h?GX4000000H0WG6O00<000000:alW`2/O9l01`2/O9l00`000000^HNO0;V7
W`060;V7W`0300000036TYl0aY:O00@0aY:O100000050=J9N`8000000P3FGQD2000002@0oooo000U
0?ooo`040000000c2fh00000068_N`8000000`2<OKd00`000000VXBk09Z4^`0409Z4^`<000001P2Y
Ym@00`000000[Z[C0:jZd`040:jZd`030000002d[]<0]:kC00D0]:kC00<000000;Vce02i/m@0102i
/m@010000000_KWF0000000000040<^/^`030000003D]kd0e;Nm0080e;Nm0`0000000`3cYfl00000
0?ooo`0m0?ooo`04000000000740000000000080FUnn00<0000007A[^P0000001027Ql`00`000000
Th_809>;b00309>;b0030000002NT<D0WY350080WY35100000020:VSd0030000002_Yll0[jO?0080
[jO?00<000000;F[c`2eZll00`2eZll00`000000^[3@0;Z`d004000000@0a[7500<000000<fjb03=
^/P00P3=^/P00`000000e<G<0=C5c0020=C5c0030000003Qa;X0000000<0jmFm0P0000000`3ifgD0
00000?ooo`0l0?ooo`030000000A0140000000800000101h?GX20000000308]TW`000000WG6O00L0
WG6O00<000000:alW`2/O9l01`2/O9l00`000000^HNO0;V7W`060;V7W`0300000036TYl0aY:O00H0
aY:O00@000000=BOW`0000000000103FRG/300000080eUhE00<000000?ooo`3oool08@3oool002@0
oooo00<0000003<;KP0c2fh00P00000508am_@030000002JQ;/0VXBk00@0000000<0XjCF0000002Y
Ym@01P2YYm@00`000000[Z[C0:jZd`040:jZd`030000002d[]<0]:kC00@0]:kC00<000000;Vce02i
/m@0102i/m@00`000000_KWF0;fieP020;fieP80000000<0bjbk0000003D]kd01P3D]kd010000000
ljM_0?>WK`00000m0?ooo`040000000007400000000000<0FUnn00<0000008N7c027Ql`00`27Ql`0
0`000000Th_809>;b00309>;b0030000002NT<D0WY350080000000<0XYoB0000002YXm000`2YXm00
0`000000[jO?0:nWc`020:nWc`030000002eZll0]J_?00<0]J_?00<000000;Z`d02j/=000P2j/=00
1P000000_kOB000000000000a[75000000D0cK[800<000000=C5c03DaL`00`3DaL`00`000000jmFm
0>_E_@020>_E_@80000000<0nM]e0000003oool0>`3oool00`0000004@0A0000000205@PL@030000
001h?GX00000008000000`2;I9l00`000000WG6O09eaW`0609eaW`030000002/O9l0[7bO00<0[7bO
2@0000040;V7W`0300000036TYl0aY:O00H0aY:O00@000000=BOW`3DWil0e9nO0`0000001P3FRG/0
00000>NGLP3WUg8000000=IN5@8000008P3oool002<0oooo00H0000003<;KP0c2fh0000007af`@00
000608am_@8000001P2SY=H00`000000ZJOD0:VWe0040:VWe0030000002^Z]<0[Z[C00@0[Z[C00<0
00000;B^d`2d[]<0102d[]<00`000000^K?D0;Vce0040;Vce0030000002m^MH0_KWF00@0_KWF0P00
00060=Bg_@030000003Ma<4000000080ljM_00<000000?ooo`3oool0>`3oool010000000<Tjo039>
_`00000205YO_P8000001@27Ql`00`000000Th_809>;b00309>;b0800000102RWm800`000000ZJ?@
0:VSd0020:VSd0050000002_Yll0[jO?0:nWc`000000102eZll00`000000^[3@0;Z`d0020;Z`d003
0000002o]m80_kOB0080_kOB0P0000040<fjb0030000003DaL`0e<G<00<0e<G<0P0000030>_E_@04
0000003bjKl0l^Vo000003`0oooo0`00000305@PL@8000001P2;I9l00`000000WG6O09eaW`0409ea
W`P00000102_S[@00`000000^INd0;VG]0020;VG]0T000001036TYl00`000000e9nO0=BOW`040=BO
W`8000000`3WUg8400000240oooo000S0?ooo`030000000c2fh000000080O7K100<0000008am_@2<
OKd00P2<OKd20000000309bRf0000000XjCF00D0XjCF00<000000:VWe02YYm@01@2YYm@010000000
[Z[C0:jZd`2^Z]<6000000<0]:kC00<000000;Vce02i/m@0102i/m@00`000000_KWF0;fieP040;fi
eP0400000031_mP00000000000@0e;Nm00D000000=g4`@3Ma<4000000?>WK`02000003`0oooo00@0
0000039>_`0bC[l0<Tjo0P00000207R3d@0300000027Ql`0QhO<00<0QhO<00L0000009>;b02CRlP0
Th_80000002KWM@0000000@0XYoB00<000000:VSd02YXm000P2YXm001@000000[jO?0:nWc`2_Yll0
000000@0]J_?00<000000;Z`d02j/=000P2j/=000`000000_kOB0;ngdP020;ngdP0600000034_M@0
00000000003=^/P000001@3DaL`01P000000f];A0=[Bd@000000jmFm000000<0l^Vo00<000000?oo
o`3oool0>P3oool00`000000E21a05@PL@0205@PL@8000001`2;I9l00`000000WG6O09eaW`040000
0080XhJe00<000000:n>]02_S[@01`2_S[@00`000000^INd0;VG]0060;VG]00300000034XKD0a:6e
0080a:6e1P0000060=BOW`<000000`3WUg800`000000bU<00000000P0?ooo`008P3oool3000000@0
O7K100@0000008am_@0000000000102LX]P00`000000XjCF0:>TeP040:>TeP040000002YYm@0ZJOD
0:VWe0L000000P2X_^T010000000ZKoY0:Voj@2Y_nT6000000<0^K?D00<000000;fieP2m^MH0102m
^MH00`000000`KoH0<6of0020<6of080000000<0e;Nm0000003Ma<400`3Ma<43000003`0oooo00@0
0000039>_`0bC[l0<Tjo0P00000307R3d@8000000`27Ql`010000000Th_800000000000209^Me003
0000002RWm80XYoB00<0XYoB00D000000:VSd02YXm00ZJ?@000000030:nWc`030000002eZll0]J_?
0080]J_?00D000000;Z`d02j/=00^[3@000000040;ngdP0300000034_M@0a;gD0080a;gD00<00000
0=C5c03DaL`00`3DaL`01`000000f];A0=[Bd@3Jd]4000000>_E_@0000000`3bjKl00`000000oooo
0?ooo`0i0?ooo`8000000`1D87400`000000M5BO0000000608]TW`@000001P2SQ[D00`000000[hjd
0:n>]0070:n>]0030000002iUk@0^INd00H0^INd00<000000<BQ]@34XKD01P34XKD010000000cZbg
0000000000030=BOW`050000003R[Yl0hZjO0>:^W`0000000P3WUg82000000030<YC00000000oooo
01h0oooo000R0?ooo`030000D0000000O7K100D0O7K100<0000009bRf02LX]P0102LX]P00`000000
XjCF0:>TeP030:>TePD000000`2W_^T00`000000Z;kY0:Rnj@030:Rnj@030000002Y_nT0ZKoY00@0
ZKoY00<000000:Voj@2Y_nT01@0000050;fieP0300000031_mP0`KoH00@0`KoH0P0000050=g4`@03
0000003k/U<0000003/0oooo0P000003039>_`030000001hPm40N8?A00<0N8?A00D0000008N7c027
Ql`0QhO<0000000509^Me0030000002RWm80XYoB0080XYoB00D000000:VSd02YXm00ZJ?@00000003
0:nWc`030000002eZll0]J_?0080]J_?00D000000;Z`d02j/=00^[3@000000040;ngdP0700000034
_M@0a;gD0<Bme0000000alGH000000030=C5c0030000003Jd]40f];A00<0f];A00@000000?;Y_`3b
jKl0l^Vo0P00000j0?ooo`040000000J05d0000005@PL@8000000`1dE9l010000000RfBO08]TW`2;
I9l20000000309In]`000000XhJe00P0XhJe00<000000:n>]02_S[@01P2_S[@00`000000^INd0;VG
]0050;VG]00300000034XKD0a:6e00H0a:6e00<000000<j/]`3>[;L00P3>[;L2000000030=BOW`00
0000hZjO00<0hZjO0`000000103fYF00000000000000000N0?ooo`008@3oool20000000306I]aP00
0000O7K100<0O7K100<0000009JPg00000001P2LX]P00`000000XjCF0:>TeP0200000080YkgZ00<0
00000:Nnj@2W_^T00`2W_^T00`000000Z;kY0:Rnj@030:Rnj@030000002Y_nT0ZKoY00<0ZKoY00<0
00000:Voj@2Y_nT00`2Y_nT00`000000Z/3Z0:[0jP0200000080_KWF00<000000<6of031_mP01@31
_mP00`000000a<KK000000030=g4`@050000003Ue<D000000?^bD`000000>P3oool00`00000003>i
00000002039>_`8000001@1hPm401@000000QhO<0000002CVmP0000000D0VigD00<000000::OdP2R
Wm800P2RWm801@000000ZJ?@0:VSd02YXm0000000080[jO?00D000000;F[c`2eZll0]J_?00000003
0;Z`d0030000002o]m80_kOB0080_kOB00D000000<Bme034_M@0a;gD000000030<O5f0030000003D
aL`0000000D0f];A0P0000020?;Y_`800000>P3oool20000000301X0G@000000000000D0M5BO00@0
000008]TW`0000000000102FO[L00`000000XhJe0:>6]@060:>6]@030000002_S[@0[hjd0080[hjd
200000040;VG]00300000034XKD0a:6e00H0a:6e00<000000<j/]`3>[;L0103>[;L2000000H0hZjO
00D000000?JUH03fYF00mZEP0000000N0?ooo`008@3oool20000000306I]aP000000O7K10080O7K1
00D0000009JPg02FX=`0UZ3L0000000609bRf08000001@2W_NX00`000000YkkY0:Nnj@030:Nnj@03
0000002X_^T0Z;kY0080Z;kY00<000000:Voj@2Y_nT00`2Y_nT00`000000ZKoY0:Voj@020:Voj@03
0000002Z`>X0Z/3Z00<0Z/3Z0`0000050<6of08000000`34a]/020000000gLC10=g4`@000000iMC5
0000003k/U<00000>`3oool200000005039>_`000000IGkH06Enf00000001@1hPm401@000000Ti_H
09>Kf02CVmP0000000D0VigD00D000000::OdP2RWm80XYoB000000030:VSd0040000002_Yll0[jO?
000000<0]J_?00D000000;Z`d02j/=00^[3@000000030;ngdP0500000034_M@0a;gD0<Bme0000000
1@37aMP00`000000f];A0=[Bd@030=[Bd@030000003Nh=L000000080l^Vo0P00000j0?ooo`040000
000J05d06P1M000000H0M5BO0P00000609In]`030000002SQ[D0XhJe00D0XhJe1P0000040:nMa@04
0000002gY<D0]jC50;NTa@P000001034XKD00`000000cZbg0<j/]`060<j/]`<00000103R[Yl20000
00<0mZEP00<000000?ooo`3oool06`3oool00240oooo00L0000006I]aP1VKLH0IVg60000001lM/40
000000@0UZ3L00<0000009bRf02LX]P00P2LX]P2000000030:JmjP000000YkgZ00@0YkgZ00<00000
0:Nnj@2W_^T00`2W_^T00`000000Z;kY0:Rnj@0500000080ZKoY00<000000:Voj@2Y_nT00`2Y_nT0
0`000000Z/3Z0:[0jP030:[0jP030000002Z`NX0Z/7Z008000000P31_mP00`000000a<KK0<C6f`04
0<C6f`8000000`3Ue<D2000003/0oooo00@00000000c^@00000000000`1UO]P01`000000N8?A07R3
d@1hPm40000008^Jg@0000000`2CVmP00`000000VigD09^Me00209^Me0030000002RWm80XYoB0080
XYoB00<000000:VSd00000002@0000000`2j/=0000000;ngdP020;ngdP0500000034_M@0a;gD0<Bm
e00000001@37aMP00`000000b/gL000000020=[Bd@8000000`3Nh=L2000000030>ci^P000000oooo
03P0oooo00<0000001X0G@0J05d00P00000407ADW`<00000202FO[L00`000000XhJe0:>6]@040000
00030:JGaP000000[ig500L0[ig500<000000;NTa@2gY<D01@2gY<D010000000_jc60;n/aP2o[<H6
000000L0cZbg00<000000=Rh^@3H^;T00P0000020>:^W`030000003`_i`000000080mZEP00<00000
0?ooo`3oool06`3oool00200oooo0P00000406I]aP030000002FX=`0UZ3L00@0UZ3L00@0000009bR
f00000000000102V_NX2000000@0YkgZ00@000000:Nnj@2W_^T0YkkY1000000209K@o0030000002@
c?`0T<cl00@000000`2Y_nT00`000000Z/3Z0:[0jP020:[0jP8000001@2Z`NX2000000P0a<KK00@0
00000>GDa@3Ue<D0iMC50`00000j0?ooo`0400000000<kT003>i000000@0IGkH00@0000007R3d@1h
Pm4000000P2;V]d200000080Ti_H00<0000009^Me02KWM@00P2KWM@00`000000XYoB0::OdP030000
00050:V`g@000000[K;L0:fbg00000000`2`]M`00`000000/kSM0000000200000080_kOB00D00000
0<Bme034_M@0a;gD000000030<O5f08000000P3:cM`00`000000f];A000000050=kPe`030000003/
nKX0000003X0oooo00D0000001X0G@000000DSjL0000000307ADW`0500000025MKT0QGFi08Ee^@00
00001P2FO[L4000000D0YYO600<000000:nMa@2_WLD01P2_WLD00`000000]jC50;NTa@050;NTa@03
0000002o[<H0_jc600D0_jc600@000000<Ndb00000000000103>[;L00`000000f;Ri0=Rh^@030=Rh
^@<000000P3`_i`010000000mZEP00000000000L0?ooo`00803oool00`000000IVg606I]aP0206I]
aP030000002>X>00000000H0UZ3L0P0000070:JmjP040000002W_NX0YkgZ0:NmjP@000000P2Me?/0
1@000000U]3l09K@o02Fd?`0000000<0T<cl00<0000008c6n`2<a_/0100000030:[0jP030000002Z
`NX0Z/7Z00@0Z/7Z0P0000000`2[`^/000000<C6f`040<C6f`<00000103Ue<D2000003/0oooo00<0
0000000c^@0000000P00000306Enf08000001@2;V]d010000000Ti_H09>Kf000000409^Me0800000
00<0YZoN0000000000003`0000030<Bme00500000037aMP0alGH0<O5f00000001@3:cM`00`000000
g^3G0=kPe`020=kPe`D00000>@3oool3000000<0DSjL0`00000408Ee^@030000002FO[L0UWjg00@0
000000<0W9;80000002VUlH01`2VUlH00`000000[ig50:nMa@050:nMa@030000002gY<D0]jC500@0
]jC500<000000;n/aP2o[<H01P2o[<H010000000akC80<Ndb037]<P3000000030<j/]`000000f;Ri
00H0f;Ri00@000000?2oW03`_i`0l;nL0`00000L0?ooo`00803oool2000000<0IVg600<0000008jP
h02>X>000P00000309JPg0030000002U_N/0YKg[008000001P2V_NX300000080Y=Ki00D0000009gD
n`2Me?/0WMCk0000000309K@o0050000002@c?`0T<cl093<o00000000`2<a_/00`000000R/7i08[1
n@03000000H0Z/7Z00D000000:_2j`2[`^/0Zl;[000000040<C6f`0500000036cMl0a/gO0<K=g`00
00000P3Ue<D3000003/0oooo0`00000204ajgP040000001UO]P0IGkH06Enf08000001@2;V]d01@00
0000Ti_H09>Kf0000000VigD00P000000`1/000010000000Q@0008D00025000:000000<0alGH00<0
00000<[=g03:cM`00P3:cM`300000080g^3G0P000000103Mkmd000000>ci^P00000j0?ooo`030000
<@000000DSjL00<0DSjL00<0000008Ee^@25MKT01@25MKT2000000H0W9;800<000000:JGaP2VUlH0
1@2VUlH00`000000[ig50:nMa@020:nMa@L000000`2gY<D00`000000_jc60;n/aP050;n/aP030000
0037]<P0akC800D0akC80P0000060=Rh^@030000003Sak`0000000<0l;nL0P00000L0?ooo`00803o
ool200000080IVg600<0000008jPh02>X>000`2>X>000`000000UZ3L000000050:Fmj`8000000P2V
_NX2000000030:_Gm`000000000000<0Y=Ki0P00000209gDn`030000002Fd?`0U]3l00@000000P2<
a_/2000000<0R/7i0P0000000`29_?L00000000000020:[1jP8000001@2[`^/010000000a<KK0000
000000040<K=g`040000003Ue<D000000>_VbP800000?03oool300000080C7[N00D0000006Enf01U
O]P0000008:JhP02000000@0Ri[M1@00000204L000030000001/0000K00000D0K00000<0000008D0
002500001@25000010000000U``009L<002G3005000000<0b/gL0P0000020<_FhP<000000`3Mkmd3
000003T0oooo0`000003058nW08000001@25MKT3000000L0W9;800<000000:JGaP2VUlH00P2VUlH8
000000<0[Z_D00@000000;>_e02c[m@0/joD1`0000040;n/aP0300000037]<P0akC800D0akC800<0
00000<jnb`0000001@3H^;T010000000hlNl0>?7_00000020?2oW0030000003cZ3L0000001/0oooo
000P0?ooo`0500000017H/X0000006I]aP0000001P2>X>02000000L0YKg[0`0000040:_Gm`800000
0`2Te_T6000000030573mP000000BKSg008000000`2:`OT2000000D0RKcg0P0000080:_2j`030000
0036cMl0a/gO00D0a/gO00D000000>_VbP3[i/X0jnK:0000000m0?ooo`03000;V@000000C7[N0080
C7[N0`00000308:JhP@000001@17000F000000@0U``0100000040<_FhP030000003Mkmd0gNoM00<0
gNoM0P00000j0?ooo`<000000P1B?Y`010000000L6Zl071Z_000000308Ee^@050000002ASL/0THg;
096=b`0000001P2LT/P5000000<0Z:OF00<000000:j[e02^Zm@01@2^Zm@00`000000/joD0;>_e004
0;>_e0040000002i]MD0^KGE0;Vee@D000001`37]<P00`000000c[k;0<jnb`0200000080f;Ri00P0
00000>?7_03Sak`0hlNl0000003`_i`000000?gCU@8000006`3oool00200oooo00<0000004MRbP17
H/X00P00000508jPh0030000002T_N/0Y;g[00@000000P2U_N/200000080/MOd100000020:_Gm``0
00000P18[oT3000000<0RKcg1000000208Zhm0030000002[`^/0Zl;[00@000000P2[`^/00`000000
a/gO0<K=g`020<K=g`<000000`3[i/X00`000000oooo0?ooo`0l0?ooo`<000000`1<N]h200000080
PY[R00D000000180000B0000000004L0000400000003080YD`000000VCmH00L0VCmH00<000000:e@
F02]D5P01P2]D5P00`000000_eeD0;mME0040000000409L<00000000WPh009h>00D000000`3Mkmd3
000003/0oooo00<0000001`LT`0L79<00P000003071Z_00400000025MKT00000000000@0THg;00<0
000009bBb02LT/P00P2LT/P2000000030:6Te`000000Z:OF00D0Z:OF00<000000:j[e02^Zm@01@2^
Zm@00`000000/joD0;>_e0030;>_e0030000002i]MD0^KGE00H0^KGE0`0000040<Ndb0030000003>
_//0c[k;00<0c[k;00<000000=Rh^@000000103Sak`2000000030?gCU@000000000001/0oooo000P
0?ooo`0500000017H/X0Af;:00000026XND00`00000208jPh0030000002T_N/0Y;g[00@0Y;g[1P00
00040;7Gm1H000001`2:^?@2000000L0Zl;[00@000000<K=g`00000000000P37eND00`000000jnK:
0>_VbP02000003h0oooo0P0000001`0[M>40000004ajgP1<N]h0000007NLj00000000P0B00040000
00D0P2UC00<0000009ToF02I?eP01P2I?eP00`000000[E1H0:e@F0060:e@F0030000002oGE@0_eeD
00@0_eeD1000000209h>00<000000P3Mkmd00`000000do_Q0000000k0?ooo`8000000P0L79<00`00
0000L6Zl071Z_003071Z_0030000002ASL/0THg;00D0THg;00@0000009bBb00000000000102QY=L0
0`000000Z:OF0:RWeP040:RWeP030000002^Zm@0[Z_D00@0[Z_D00<000000;>_e02c[m@00`2c[m@0
0`000000^KGE0;Vee@050;Vee@040000002m^mL0_K_G0;fke`8000000P37]<P00`000000c[k;0<jn
b`040<jnb`030000003Sak`0hlNl00@0hlNl00@000000?gCU@3mdiD000006`3oool00200oooo0P00
00000`17H/X0000008JQi@0308JQi@<000001`2T_N/00`000000]]G`0;KEl0020;KEl0/0000000L0
BP000000001/0000000008@000000000U0T000`00000102<]?400`000000Zl;[0:_2j`050:_2j`80
00001@37eND2000000030>Whb`000000oooo03h0oooo00<0000002]dh@0[M>40100000001@0B0000
00000000001N2dL0000000H0P2UC00<0000009ToF02I?eP01`2I?eP00`000000[E1H0:e@F0070:e@
F0030000002oGE@0_eeD00D0_eeD00D000000<mXB@000000000009h>0004000000030=?kh@000000
000003/0oooo00<0000001`LT`0L79<00P000004071Z_00300000024R=00000000H0THg;0P000006
0:6Te`030000002XYmH0Z:OF00<0Z:OF400000060;Vee@030000002m^mL0_K_G00<0_K_G0P000006
0<jnb`030000003Db/l0000000@0hlNl0P0000020?gCU@8000006P3oool00200oooo00@0000001YA
b`00000000001@26XND00`000000Y;k/0000000300000080Y;g[00<000000;KEl02feO000P2feO07
000000040<[Sj@000000000000000080BP0000<0000006`0000000000P2400000`000000U0T00000
000200000003072JjP000000000000D00000102<]?4010000000Zl;[0:_2j`2[`^/4000000030:[3
k0000000amGU00@0amGU0`0000110?ooo`030000000[M>40000000@000000P1N2dL00`000000P2UC
080YD`05080YD`030000002I?eP0VCmH00L0VCmH00<000000:e@F02]D5P01`2]D5P00`000000_eeD
0;mME0050;mME0040000003?J4T0cfQ90<mXB@<0000000@0T`<00000003Cnn400000?03oool01P00
000071bC01`LT`000000DUbo00000080L6Zl00@0000008B8d024R=000000102ASL/20000000309ZQ
fP000000XJCG00H0XJCG00<000000:RWeP0000000P0000000`2W]n@000000:Zii0050:Zii0030000
002]^n@0[K_T00<0[K_T00<000000:nmi00000000`0000000`2i]MD000000;fke`050;fke`040000
0032`]X00000000000<0c[k;00@000000=C:c`3Db/l0e<[?0P0000020>?7_0040000003]f[l00000
0?gCU@8000006P3oool00240oooo00@0000001YAb`00000000000`26XND00`000000Y;k/0:Bnk004
0:Bnk0L00000102id^d00`000000b^?Y000000020000000501T000000000BP000000001/00000`00
000010240000000009@9002D2@0400000003072JjP000000Sk3]00<0Sk3]200000050:[3k0030000
0037eND0amGU0080amGU0P0000000`3Yn</000000?ooo`110?ooo`<000001@1N2dL00`000000P2UC
080YD`03080YDaT000001P2oGE@00`000000cfQ90<mXB@020<mXB@80000000<0T`<00000003oool0
?03oool00`00000071bC00000002059L_`030000001`J[`0000000@0Q8S@00@00000096=b`000000
00000`2JXMX00`000000XJCG0:6Te`040:6Te`<00000102W]n@00`000000Z[WT0:Zii0040:Zii003
0000002]^n@0[K_T0080[K_T00<000000:nmi02__N@0102__N@3000000@0_K_G00<000000<;2fP32
`]X00P32`]X2000000030<jnb`000000e<[?00@0e<[?0P0000020>gJ_`@000006P3oool00240oooo
00@0000001YAb`000000OJ;[0P0000000`26XND000000:Bnk0060:Bnk0030000002id^d0^M;]00<0
^M;]3P0000000`2A;44000000:Dl@@080000000307bIi0000000000000@00000102?/>d00`000000
Z/?/0:[3k0050:[3k00300000037eND000000080000000@0aMcZ0000003Yn</00000@@3oool30000
00D0GP]700<00000080YD`20:E<01000000209MPSP030000002YKXl0ZFj?00L0ZFj?00<000000;Qk
S`2hNhl01`2hNhl00`000000ahN>0<N7SP0400000080_eeD00<000000<mXB@3?J4T00`3?J4T50000
03/0oooo0`000003059L_`0300000024R=00Q8S@00@0Q8S@00<0000009ZQfP2JXMX0102JXMX00`00
0000XJCG0:6Te`0300000080Y;KU00<000000:Ngi02W]n@00`2W]n@00`000000Z[WT0:Zii0030:Zi
i0030000002]^n@0[K_T0080[K_T00<000000:nmi02__N@0102__N@00`000000/L3U0;70i@020000
0080_K_G00<000000<;2fP32`]X01032`]X00`000000e<[?0=C:c`050=C:c`070000003]f[l0kMZo
0>gJ_`000000n=mk0000000J0?ooo`008@3oool00`0000006U7;0000000307fRj`l000000`2kc^T5
0000000505@0:`000000MaDj07LE>P0000000P2A;4401@000000YCa10:Dl@@2U?44000000080]TTk
00<000000<EA;@0000001000000309>^j@h000000`35g>X010000000jOS;0000000000100?ooo`80
000000<0:P0_0000001N2dL00`1N2dL4000000H0Uf2>00<000000:U^S`2YKXl01`2YKXl00`000000
^7^?0;QkS`070;QkS`0300000037Qhh0ahN>00@0ahN>100000050<mXB@030000003ML38000000080
0000>P3oool2000000@0DUbo0P00000408B8d0030000002BX=d0000000H0VZ7J0P0000050:Bfi@04
0000002W]n@0YkOT0:Ngi10000000`2__N@00`000000/L3U0;70i@030;70i@<000001@32`]X00`00
0000aLWM000000050=C:c`8000000`3]f[l00`000000n=mk0000000J0?ooo`008@3oool3000000@0
OJ;[00<000000:>nk02S_^`01@2S_^`00`000000^lkY0;_>j@030;_>j@<0000000<0^l[V00000000
00000P1D02/010000000MaDj07LE>P000003094/@@050000002U?440YCa10:Dl@@0000000`2fBC/0
0`000000aE4]000000020000000409N]iP0000000000000000@0TjkY00<000000:[3k02Z`n`01@2Z
`n`00`000000aMcZ0<GLjP020<GLjP<00000@03oool010000000:P0_02X0;`00000305h;A`800000
00<0PDn<0000002GH8h01`2GH8h00`000000ZFj?0:U^S`070:U^S`030000002hNhl0^7^?00L0^7^?
00<000000<N7SP37Qhh01P37Qhh010000000eYB<0000000000030<mXB@040000003ML380gG0b0=e`
<P800000>@3oool300000080DUbo00<0000007>4e@0000000`24R=0010000000TZ3M09:Pg@2BX=d2
00000080VZ7J0P0000000`2Q]NL000000:Bfi@050:Bfi@<000000`2Sb?<00`000000XLOc0:77l`02
0:77l`030000002Pa_<0X<Kc0080X<Kc00<0000009o5l`2OaO<0100000050;70i@030000002b`nH0
//?V008000000P32`]X010000000aLWM0<G9g@35bMd2000000030=C:c`00000000000080f]OD00H0
00000>gJ_`00000000000?SON`00000J0?ooo`008P3oool3000000<0OJ;[00<000000:>nk02S_^`0
1P2S_^`00`000000^lkY0000000200000080^l[V0P0000001@0N00h0000005@0:`1D02/000000080
MaDj00D00000094/@@2A;440TBa1000000030:Dl@@050000002fBC/0]TTk0;I9>`0000000P35DBd3
000000<0UjgV0P00000209>^j@030000002Z`n`0Z/?/00D0Z/?/00<000000<GLjP35g>X00`000000
0`3@o[l000000?ooo`0n0?ooo`060000000Z02l0:P0_02X0;`000000GP]70P000003085?S0030000
002GH8h0Uf2>00L0Uf2>00<000000:U^S`2YKXl01P2YKXl00`000000^7^?0;QkS`070;QkS`030000
0037Qhh0ahN>00H0ahN>00@000000=JDS03FU8`0eYB<0`0000000`3?J4T000000=e`<P020=e`<P80
0000>03oool300000080DUbo00H0000007>4e@1cQ=D0000008B8d000000609:Pg@800000102Q]NL2
00000080Y;KU0P0000000`2UbO8000000:?8l`030:?8l`030000002Qao<0XLOc0080XLOc00D00000
0:36l`2Pa_<0X<Kc0000000409o5l`030000002N`o80W/?b008000000P2a`>D00`000000//?V0;;3
iP030;;3iP8000001P35bMd00`000000f]OD0=[Ge0020=[Ge0060000003]f[l000000000003hgg/0
00006P3oool00280oooo00<00000000h_@0000000`0000000`1mX^/000000:>nk0030:>nk0D00000
102kb^H20000000501h03P000000E00[05@0:`0000000`1g5CX01@000000TBa1094/@@2A;4400000
00@0YCa100@000000;I9>`2fBC/000000P35DBd00`000000cU@B00000002000000@0UjgV30000000
1031ho0000000=3n_`00000o0?ooo`030000000Z02l0:P0_0080:P0_0P000005085?S0030000002G
H8h0Uf2>00L0Uf2>00<000000:U^S`2YKXl01P2YKXl00`000000^7^?0;QkS`070;QkS`0300000037
Qhh0ahN>00H0ahN>00<000000=JDS03FU8`0103FU8`2000000@0gG0b0P00000g0?ooo`80000000<0
9TRo000000000000101cQ=D00`000000TZ3M09:Pg@0409:Pg@030000002N]NP0000000H0XKGW0P00
00040:G9lP040000002Sb?<0XlSc0:?8l`/000000`2OaO<00`000000W/?b09k3lP0209k3lP800000
1@2b`nH2000000030;?6j0000000aLWM00<0aLWM0P0000050=[Ge0030000003bk[l0l^jo00800000
6P3oool002<0oooo0P00000307BUl0H000000`2S_nd00`000000^l[V0;_:iP020;_:iP<0000000D0
7P0>0000001D02/0E00[0000000207LE>P030000002A;440TBa10080TBa100<000000:Dl@@2U?440
0P2U?44010000000]TTk0;I9>`0000030<EA;@80000000<0b4X00000000000000P0000000`2G[NH0
00000:W4k@060:W4k@0300000031ho00`N?`0080`N?`00<000000=3n_`000000?P3oool200000080
:P0_0`000007085?S0030000002GH8h0Uf2>00H0Uf2>00<000000:U^S`2YKXl00P2YKXlA000000H0
ahN>00<000000=JDS03FU8`01@3FU8`300000080gG0b00<000000=e^1`000000=P3oool200000080
9TRo00H0000007>4e@1cQ=D0LhCE00000029Wn82000000<0TZ3M00@0000009jej02N]NP0W[GX0P00
00020:6ei`80000000<0Y/Wb0000000000000`2UbO840000000309GHoP000000S=?o0080S=?o00@0
000008G=o`25cOl000000P1oaoh4000000<0W/?b0P000000102M`_8000000000002b`nH2000000@0
/lKX00@000000<G9g@00000000000P37dN8010000000f]OD0=[Ge03Jem@200000080l^jo00<00000
0?ooo`3oool06@3oool002@0oooo0P00000407BUl0030000002S_nd0Xko]00D0Xko]00<000000;_:
iP0000000`0000001@0N00h0000005@0:`1D02/000000080MaDj3@0000020;I9>`0500000035DBd0
aE4]0000003>E1800P00000309^/h`<000001P2Ya>d010000000`N?`0<7Sl031ho03000003h0oooo
0P0000000`0Z02l000000000000206<hQ`0300000021Ch`0PDn<00D0PDn<00<0000009MPSP2GH8h0
0P2GH8h9000000@0[Xnf00<000000;ZH]P2jV;H01`2jV;H00`000000a::f0<BR]P0400000080ahN>
00<000000=JDS03FU8`01@3FU8`01P000000iZ>80>JSR0000000gG0b00000080gFh700<000000?oo
o`3oool0=03oool00`0000009TRo02I8_`0200000080LhCE00@0000008VOhP29Wn80RIoR0P000000
0`2BX=d0000009jej00409jej08000001@2VbO8300000080Wmcl3`00000207g1o0<000001@2M`_80
0`000000/lKX0;?6j0050;?6j00300000037dN80am7R0080am7R00H000000=[Ge000000000000=gV
f`0000020?;^_`030000003oool0oooo01T0oooo000U0?ooo`@0000000<0M:G`0000002S_nd00`2S
_nd5000000050;_6h`00000000000000000N00h00P1D02/40000000308`oJ@000000X51/00<0X51/
00<000000;5NK02aGV`00P2aGV`00`000000`F]Y0000000300000080aE4]00@000000<iD4P000000
00000P2K[><00`000000ZLC]0:W4k@04000000040:W4k@000000`N?`0<7Sl0@00000?P3oool01000
0000000200000000000406<hQ`0300000021Ch`0PDn<00@0PDn<1P0000040::6]P030000002^SkH0
[Xnf00H0[Xnf00<000000;ZH]P2jV;H01`2jV;H00`000000a::f0<BR]P040<BR]P@000001@3FU8`0
0`000000iZ>80>JSR0020>JSR0<0000000<0gFh70000003oool0=@3oool300000080GGkK0P000006
08VOhP800000102N]NP010000000Z<Wa0000000000020:K9lP8000000P2XgOX@00000004046/m`00
0000000000000080OK_j0P00000209g2lP80000000<0WL3a0000000000000P2ca^P2000000030;C9
jP000000am7R00@0am7R00H000000=gVf`3Mi]/0gNKK0000003bk[l2000001/0oooo000V0?ooo`80
000000<0K:We000000000000100000040:>ok@@0000000H07P0>0000000000000000078XHP000003
08`oJ@030000002PD6`0X51/0080X51/00<000000;5NK02aGV`00P2aGV`01P000000`F]Y0<5[J@31
JfT000000=5fH`<000000P3>E1800`000000VjcS000000070:W4k@@0000000<0^NOe000000000000
?P3oool00`00000000020000000606<hQ`0300000021Ch`0PDn<0080PDn<0P0000000`2COKP00000
0::6]P070::6]P030000002^SkH0[Xnf00H0[Xnf00<000000;ZH]P2jV;H01`2jV;H00`000000a::f
0<BR]P060<BR]P040000003?[[P00000000000<0eYB<00<000000>JSR03VXhP0103VXhP010000000
gFh70=e^1`00000e0?ooo`<000000`1MO]/2000000@0RIoR00<0000009^ejP2K]NX00P00000209je
j0040000002XbO40Z<Wa0:S9lB0000001@2M`?42000000@0]<WZ00@000000<OAhP37dN80am7R0P00
00030=gVf`@000006`3oool002L0oooo0P00000306bYm@030000002S_nd0Xko]00@0Xko]1P000003
078XHP050000002<?fT0S3mY08`oJ@000000102PD6`00`000000/Ei/0;5NK0020;5NK00500000031
JfT0`F]Y0<5[J@0000000`3AMV<6000000L0ZLC]00@000000;WWm@2iioD0^NOe0P00000n0?ooo`@0
00001P1S>8L00`000000PDn<00000002000000<0Tgfh00<000000::6]P2RQ[H01P2RQ[H00`000000
[Xnf0:j?]P060:j?]P030000002jV;H0^YRf00H0^YRf00<000000<BR]P34X[H01P34X[H00`000000
cjjh0<n^^0020<n^^080000000<0eYB<0000003VXhP01@3VXhP2000000030=e^1`000000oooo03@0
oooo0`00000305enf`040000001oWnL0000000000080RIoR00<0000009^ejP2K]NX00P2K]NX30000
00<0Z<Wa00D000000;7MmP2agOH0/Mgf000000020;G`mA<0000000<0Dj7f0000001o]_L00P1o]_L0
0`000000WL3a09g0l@0200000080W[o`00<000000;C9jP2dbNX00P2dbNX01P000000am7R00000000
0000b=WW000000<0gNKK00<000000>Kk]@000000703oool002P0oooo0P00000306bYm@030000002S
_nd0Xko]00H000000P1<1e@010000000LRQR078XHP00000408`oJ@030000002PD6`0X51/0080X51/
00<000000;5NK02aGV`00P2aGV`00`000000`F]Y0<5[J@020<5[J@040000003AMV<0dGIS00000080
h85F1`0000020:W4k@040000002iioD0^NOe0;WWm@800000?`3oool20000000303<FO0000000HcR7
00D0HcR70P00000609=m^0030000002RQ[H0XXJf00H0XXJf00<000000:j?]P2^SkH01P2^SkH00`00
0000^YRf0;ZH]P060;ZH]P0300000034X[H0a::f00H0a::f00<000000<n^^03?[[P0103?[[P30000
00D0iZ>800<000000?NdOP0000000P00000e0?ooo`<000000P1MO]/010000000OioW07nOi`1oWnL4
000000<0VkGZ00<000000:W8l02Yb?003P0000000`3Sf</00000000000050000000308Uo`P000000
000000`00000102N_o0010000000]<WZ0;C9jP2dbNX3000000@0b=WW100000000`3VnkD000000?oo
o`0K0?ooo`00:P3oool6000000<0Xl3^0`0000000`1<1e@00000078XHP02078XHP030000002<?fT0
S3mY0080S3mY00<000000:1@K02PD6`00P2PD6`00`000000/Ei/0;5NK0030;5NK00500000031JfT0
`F]Y0<5[J@0000000`3AMV<00`000000h85F0>21EP02000000@0Z<C^1P0000100?ooo`8000000P0c
5W`00`000000HcR706<hQ`0206<hQ`<00000202COKP00`000000XXJf0::6]P050::6]P030000002^
SkH0[Xnf0080[Xnf2@0000040;ZH]P0300000034X[H0a::f00H0a::f00<000000<n^^03?[[P01@3?
[[P010000000f[^j0000000000030>JSR0050000003g]7h0mkAn0?NdOP000000=@3oool00`000000
02Vc00000003000000@0OioW00<0000009Rfk@2H]^d01@0000020:W8l0030000002gfo80]m_b00D0
000000D0b^G[00000000000000000=WFeP06000000040:^A^`000000000008Uo`PD0000000<0KI[[
0000000000000P00000208>al`040000002N_o00W[o`09jol0D000000P2cc>d010000000b=WW0<SI
i`38fNL2000000040=[dh0000000i_^e000001d0oooo000[0?ooo`8000000P1T[_T01P000000Xl3^
0:?0kP0000003`0i00000080C0MD00D00000078XHP1b:680LRQR0000000408`oJ@030000002PD6`0
X51/0080X51/00<000000;5NK02aGV`00`2aGV`01@000000`F]Y0<5[J@31JfT0000000<0dGIS00H0
00000>21EP3PPEH000000>^8?00000020:S4kP8000000`2]jOT00`000000oooo0?ooo`0o0?ooo`03
0000000c5W`0<aIl0080<aIl00@0000006<hQ`1S>8L000000`22LkX00`000000Tgfh09=m^00509=m
^0030000002RQ[H0XXJf0080XXJf200000040:f^f0030000002c/mP0/k?H0080/k?H200000040<BR
]P030000003?[[P0cjjh00D0cjjh00<000000=Zk^P3J^kX00P3J^kX2000000030>JSR0000000mkAn
00<0mkAn00<000000?ooo`3oool0<`3oool200000080@GSP0P00000207nOi`030000002H]^d0V;K]
00<0V;K]00<000000:_8k`2[b>l01@0000020;_GkP030000003:iN/00000008000000P3Ie]H00`00
0000c[o;00000003000000030;BH^`2[TK/000000080000000<0T8g<0000001fQ]H010000000101]
V^/0000008R^kP28[^h500000080W[k_00<000000;?<k@2cc>d00P2cc>d01`000000b=WW00000000
0000aN;/0000003Jm>000`00000M0?ooo`00;@3oool20000000606B^n@00000000000000000?03T0
00000P1<1e@01@000000LRQR078XHP1b:680000000@0S3mY00<000000:1@K02PD6`00P2PD6`00`00
0000/Ei/0;5NK0030;5NK00500000031JfT0`F]Y0<5[J@0000000`3AMV<010000000h85F0>21EP3P
PEH4000000<0[NWi0P0000110?ooo`800000100c5W`3000000@0PW>j00<0000009=m^02COKP0102C
OKP6000000<0YZWI00<000000:f^f02][]P01P2][]P00`000000/k?H0;>cf0060;>cf0040000002i
^MT0^KWI0;Vif@H000001P3?[[P00`000000f[^j0=Zk^P040=Zk^P800000103g]7h00`000000oooo
0?ooo`0c0?ooo`<0000000<0@GSP0000001dX>d01@00000309Rfk@030000002[b>l0ZlS_0080ZlS_
00<000000;_GkP2kenh010000000103=g>@000000000003Ie]H4000000030<:]aP000000000000@0
000000H0VY38000000000000Q8W@00000024V=l600000080R:k^00<0000009jnk`2N_^l00P2N_^l0
0`000000/lc]0;?<k@0400000080aN;/00<000000=[dh03Jm>000P00000N0?ooo`00;`3oool30000
00803`0i00@0000004`7E01<1e@000000P1b:6800`000000S3mY08`oJ@0308`oJ@030000002PD6`0
X51/00D000000`2aGV`00`000000`F]Y0<5[J@020<5[J@040000003AMV<0dGIS000000<0h85F00@0
00000>^8?0000000W^Ol0`0000130?ooo`800000100c5W`00`000000PW>j089c^P04089c^P030000
002COKP0Tgfh0080Tgfh0P0000000`2NY]/000000:JYf@070:JYf@030000002][]P0[JkH00D0[JkH
00<000000;>cf02c/mP01P2c/mP00`000000^KWI0;Vif@040;Vif@030000002n_m/0_[oK00800000
103?[[P00`000000f[^j0=Zk^P040=Zk^P030000003Uc;d0000000<0mkAn0P00000e0?ooo`800000
00<0@GSP0000001dX>d00P1dX>d00`000000UKO_09Fgk`0;00000080_M;Z0`0000020<cBgP800000
00L0c[o;00000000000000000;^Va0000000/io40080000000L0Y9G6000000000000T8g<0000002>
VMX000000080Q9SO0`0000000`2=ZnX000000000000800000080//o_00@000000<GRk035h^`0aN;/
0`0000000`2_ihh000000?ooo`0M0?ooo`00<03oool20000000500l0>@000000C0MD04`7E0000000
0`1b:68010000000S3mY08`oJ@2<?fT500000080Z6R800<000000;MeR02gMHP01P0000020<5[J@05
0000003AMV<0dGIS0=5fH`0000000P3PPEH00`000000jhPl00000002000004D0oooo0P00000303<F
O08000001`22LkX010000000Tgfh00000000000409jVf`030000002VZMT0YZWI00D0YZWI00<00000
0:f^f02][]P01@2][]P00`000000/k?H0;>cf0050;>cf0030000002i^MT0^KWI00D0^KWI00<00000
0;jof`2n_m/00P2n_m/2000000030<n^^0000000f[^j00H0f[^j00P000000>G<_@3Uc;d000000?Nd
OP000000o/9U000003H0oooo1000000207BPk@030000002E]nl0UKO_00<0UKO_00<000000:c7kP2/
anh00P2/anh00`000000_M;Z0;gBjP0300000080c=;N0`0000004@37`MD0000000000032[LH00000
0;^Va0000000/io40000002/V/@0Y9G60000002JT<P0000009NJe@000000SYWJ00<000000P2BZ^H2
0000000308f[jP000000Wkg^00<0Wkg^00<000000;;?k`2bcnl00P2bcnl00`000000aN;/0<GRk002
000000030<_mh0000000000001l0oooo000`0?ooo`80000000D03`0i0000001<1e@0C0MD00000002
078XHPD000000P2GFhL00`000000Z6R80:QXR0020:QXR0030000002gMHP0]gF800<0]gF800<00000
0<F0Q`35P8L0100000020=5fH`070000003PPEH0h85F0000003[R3`000000>:12@150?ooo`060000
00000680001R0000000c5W`000000P1ZJ;d00`000000PW>j089c^P04089c^P8000001P2NY]/00`00
0000YZWI0:JYf@060:JYf@030000002][]P0[JkH00@0[JkH00<000000;>cf02c/mP01@2c/mP00`00
0000^KWI0;Vif@040;Vif@030000002n_m/0_[oK00D0_[oK0P0000060=Zk^P070000003Uc;d0iLbm
0>G<_@000000mkAn000000020?k2I@030000003oool0oooo03@0oooo00@0000001]`h00000000000
0P1dX>d00`000000UKO_09Fgk`0209Fgk`030000002/anh0[<O^00D000000`2ncNH4000000030<O1
e@000000`[[B00<0000000D0^jK40000002cWl@000000:bJa0030000000309jMdP000000Ui[E0080
000000<0SYWJ0000000000000`2BZ^H400000080Wkg^00<000000;;?k`2bcnl01P0000000`3;oN00
00000000000P0?ooo`00<03oool0100000003`0i00l0>@00000204`7E0<0000000<0PT^40000002G
FhL0102GFhL00`000000Z6R80:QXR0020:QXR0030000002gMHP0]gF800<0]gF800<000000<F0Q`35
P8L00P35P8L00`000000dhf500000002000000<0h85F0P0000000`3RP@T000000?ooo`130?ooo`03
000000000680001R008000000`1ZJ;d00`000000PW>j089c^P02089c^P80000000<0UZ?N0000002N
Y]/01P2NY]/00`000000YZWI0:JYf@040:JYfA<00000102i^MT00`000000_[oK0;jof`050;jof`04
00000032amh00000000000@0f[^j00<000000>G<_@3Uc;d00P3Uc;d200000080o/9U00<000000?oo
o`3oool0=03oool00`0000006g3P01]`h00;00000080[<K]00<000000;k=iP2ncNH00P0000020;g8
hP0600000037`MD0al7E00000032^]8000000P2m/m0700000080YJ3@0P00000309NJe@030000002H
ZN<0V:WS0080000000<0TZ[V0000002P_>d00P2P_>d5000000030:oAl@000000_nSb0080_nSb00<0
00000<_mh00000008@3oool00300oooo00@0000000l0>@000000C0MD0`000003089;Q0030000002G
FhL0Ue^700<0Ue^700<000000:QXR02XJ8P00P2XJ8P00`000000]gF80;MeR0030;MeR00300000035
P8L0aH270080aH2700@000000=>=Q@3CSHD0dhf50`0000000`3PPEH000000>^8?002000004@0oooo
00@000000000HP00068000001@1ZJ;d010000000PW>j00000000000409JSgP030000002NY]/0WZKK
00@0WZKK00<000000:JYf@2VZMT0100000000`2Kc_T0000009W=n@0509W=n@030000002GboT0Ul_i
00@0Ul_i00@0000009K9n@2FbOT0U/Wi1@0000060;jof`0300000032amh0`/ON0080`/ON0P000000
0`3J^kX000000>G<_@050>G<_@040000003n`VD0o/9U000003L0oooo0P0000001@0KL>00000006VS
lP1YXo80000000@0T[Wb00<000000:c6k@2/a^d00P2/a^d2000000<0_LSR00T000000;c3g`000000
al7E00000032^]8000000;fcd00000000P2h[Ll01@000000/ZS>0000002/Y<l000000080YJ3@00L0
000009jMdP000000Ui[E0000002MZ]l0000000<0V:WS0`0000030:2lk@030000002_dO40[m7a0080
[m7a00<000000;oXlP2oj?800`00000R0?ooo`00;`3oool20000000300l0>@00000000000080ISAo
00D00000089;Q022Bh@0PT^40000000509MKQ`030000002XJ8P0Z6R80080Z6R800<000000;MeR02g
MHP00`2gMHP00`000000aH270<F0Q`020<F0Q`050000003CSHD0dhf50=>=Q@0000000P3SVX020000
00040>^8?00000000000000004<0oooo00D000000000HP000000B5Ro0000000406YX_@8000001P2F
Xmh00`000000WZKK09jVf`0409jVf`<000001P2Kc_T00`000000VLgi09W=n@0309W=n@030000002G
boT0Ul_i00@0Ul_i00<0000009K9n@2FbOT00`2FbOT00`000000ULOi09G7n@0400000080_[oK00<0
00000<;7gP32amh01032amh2000000D0iLbm00D000000>oO_`000000o/9U0000000h0?ooo`H00000
102B^O800`000000[<K]0:c6k@02000000030:c4k0000000_LSR008000000`2l`ml00`000000`[[B
000000020;fcd0060000002h[Ll0^:g?0000002bZ<h000000P2/Y<l00`000000YJ3@000000030000
00<0WJ[O0P000000102HZN<000000::lk02R_>`500000080[m7a00<000000;oXlP2oj?800`00000S
0?ooo`00;`3oool0100000<03`0i00000000000306HdO`0500000022Bh@0PT^4089;Q00000001@2G
FhL00`000000Z6R80:QXR0020:QXR0030000002gMHP0]gF800<0]gF800<000000<F0Q`35P8L00P35
P8L01@000000dhf50=>=Q@3CSHD0000000<0hiZ00`0000000`3RP@T000000?ooo`120?ooo`800000
0P18F;l01`000000JVRm06YX_@1ZJ;d0000008bQhP0000001`2FXmh00`000000WZKK09jVf`030000
0080W/oi00<0000009_>n@2Kc_T0102Kc_T010000000VLgi09W=n@2IcOT5000000<0Ul_i00<00000
09K9n@2FbOT00`2FbOT00`000000ULOi09G7n@0509G7n@8000001`32amh00`000000a/oR00000004
0>G<_@050000003_gkl000000?k2I@000000>P3oool00`00000006;C00000008000000@0[<C/00<0
00000;c3g`2l`ml00P2l`ml00`000000^[kM0000000=000000040:J]f`000000XZ_M0::[g@800000
0P2MZ]l2000000@0X[c/00@000000:cBm02/d_@0[=;d1`00000T0?ooo`00;`3oool20000000303`D
M0000000ISAo0080ISAo00D00000089;Q022Bh@0PT^40000000509MKQ`030000002XJ8P0Z6R80080
Z6R800<000000;MeR02gMHP00`2gMHP00`000000aH270<F0Q`020<F0Q`050000003CSHD0dhf50=>=
Q@0000000`3SVX000`000000lZQe00000002000004<0oooo0P00000304QH_`030000001ZJ;d00000
00<0S:7R0P00000509JSgP<000001P2NcoT00`000000Vlki09_>n@0209_>n@P000000P0_/n<:0000
00H0ULOi0P0000000`2DaOP000000<;7gP030<;7gP8000000`36cn8010000000iLbm0>G<_@000003
0>oO_`800000?03oool20000000305fVm`000000TKcd00<0TKcd00<000000:c4k02/a>`00P000000
1@2/`n/000000000002l`ml000000080^[kM00D000000;Nif`2g^M/0]kWK000000020;>efP030000
002_/]X000000080ZjoJ00D000000:J]f`2V[M/0YZgK000000030::[g@8000000P2S_>/200000080
X[c/00@000000:cBm02/d_@0[=;d0P000000102ekOH0000000000000000U0?ooo`00;P3oool20000
0080?1Ad00D0000006HdO`1V=7l0ISAo00000003089;Q0030000002GFhL0Ue^700<0Ue^700<00000
0:QXR02XJ8P00P2XJ8P00`000000]gF80;MeR0030;MeR00300000035P8L0aH270080aH2700D00000
0=>=Q@3CSHD0dhf5000000030>>JP0030000003bZ7D0lZQe00800000@`3oool00`000000B5Ro04QH
_`0204QH_`030000001ZJ;d0000000D0S:7R00<0000009JSgP2FXmh00P0000030:7?n0800000102N
coT9000000030=://`000000000000<0000000<0WX2b0000000000001000000302VSk@8000000P2E
aoT2000000@0U<Gh0P0000000`32amh000000<K?hP050<K?hP030000003Uc;d0000000@0kmno00<0
00000?ooo`3oool0>`3oool500000080TKcd0`0000050:c3j`8000000P2j_]d010000000]kWK0;Ni
f`0000020;>efP040000002_/]X0[k;J000000<0ZjoJ00D000000:J]f`2V[M/000000::[g@020000
00030:FljP000000Xkc[00<0Xkc[2`00000V0?ooo`00;P3oool200000080?1Ad00D0000006HdO`1V
=7l0ISAo00000003089;Q0030000002GFhL0Ue^700<0Ue^700<000000:QXR02XJ8P00P2XJ8P00`00
0000]gF80;MeR0030;MeR00300000035P8L0aH270080aH2700<000000=>=Q@3CSHD00P3CSHD01000
0000hiZ00>>JP00000020?:XM@<00000@P3oool2000000@0B5Ro00<0000008bQhP2<XN801@2<XN82
000000L0XLoh200000000`3ckk`000000>K8^@04000000040;jH/@00000000000:V8/@@0000000<0
JVBi0000000M@K/01`00000809C5n00300000036cn80a/oR00H0a/oR00<000000>oO_`3_gkl00P3_
gkl2000003l0oooo1@000002092omP8000000P2/`n/200000080[<;Z0`0000020;Nif`040000002c
]MX0/kGJ000000<0[k;J00@000000:^_fP2[[mX000000P2V[M/2000000@0YKcZ00<000000:>lj`2S
_>/00P0000030:SCmP030000002UkOT0000000800000:03oool002d0oooo0`00000203`DM0050000
001V=7l0ISAo06HdO`0000000`22Bh@00`000000Ue^709MKQ`0309MKQ`030000002XJ8P0Z6R80080
Z6R800<000000;MeR02gMHP00`2gMHP00`000000aH270<F0Q`020<F0Q`030000003CSHD0dhf50080
dhf500@000000>>JP03SVX0000000P3bZ7D00`000000o;EN000000120?ooo`8000000`18F;l01000
0000PIoW00000000000408bQhP030000002ScoL0Xlog00<000000`2QcoP8000000030??__0000000
00000080000000@0dZbc000000000000_YRa0P000000102YR;4000000000002AN;<40000000301e1
^`000000000000<0000000<0<j3c0000000000000`2DaOP200000080ULCg00<000000<K?hP36cn80
0`36cn82000000070<OHiP000000kmno0>oO_`000000l_>m000000110?ooo`@000000`2@_oH30000
00<0[<;Z00<000000:_1j@2[`NT03`0000020:Jlj@040000002U_>X0YKcZ0:FljP<000000`2XdoH5
000002X0oooo000]0?ooo`030000000005/000000080?1Ad00D0000006HdO`1V=7l0ISAo00000003
089;Q0040000002GFhL0Ue^709MKQa0000000P35P8L00`000000dhf50=>=Q@020=>=Q@090000003S
VX00hiZ00>>JP0000000lZQe0000003l]Eh000000480oooo00D0000001<o_0000000B5Ro00000004
086Oi`80000000<0S:7R0000002ScoL01@2ScoL5000000030:cbm@000000000000<0000000<0lnnl
0000003Vb;T00P000000103B[;<000000000002nV;42000000030:V8/@000000WX2b00<0000000D0
JVBi000000000000Bg7G02IWf@030000000304>Mm@000000000000<000001@2Ea?L2000000030<K?
hP000000000000<0amSV00<000000>oO_`0000000P3blkd00`000000oooo0?ooo`100?ooo`H00000
0`2@`_P4000000<0Zl7Y00D000000:Zoj@2Z_nT0Z[oY000000030:Vnj@050000002W_NT0YkgY0:Nm
j@0000000`2V_>T4000000<0Y=?h1P00000/0?ooo`00;@3oool00`000000001K0000000203`DM005
0000001V=7l0ISAo06HdO`0000000`22Bh@400000080WW2N00<000000:akWP2/Nih00P2/Nih00`00
0000^HJN0;V6WP030;V6WP0300000035T9h0aI2N00@000000`3CSHD020000000hiZ00>>JP03SVX00
00000?:XM@000000o;EN0P0000110?ooo`050000000C?k`0000004QH_`0000001P21WnL3000000@0
Xlog0P0000000`2/l_D0000000000002000000030<Cod`00000000000080hncF0`0000000`3L^KH0
00000=://`02000000060;jH/@000000]92`0:V8/@000000WX2b0P0000000`20KkH0000000000002
0000000404]ae`0VImT00000001IdP<000000P13WOD00`000000ULCg09G4m`0209G4m`@000001P37
f>H01@000000l_>m0?;c_@3blkd0000004D0oooo100000020932n08000000P2Aa_T500000080Z[oY
00D000000:Vnj@2Y_^T0ZKkY000000020:Nmj@D000000P2OdOT200000080Y=?h1000000_0?ooo`00
;@3oool00`000000001K0000000203`DM0060000001V=7l0ISAo06HdO`000000PT^40`00000509i`
WP030000002/Nih0[7^N0080[7^N00<000000;V6WP2iQYh00`2iQYh00`000000aI2N0<F@WP020<F@
WP060000003BW9h000000000003CSHD000000`3SVX0010000000lZQe0000003l]Eh200000440oooo
00@0000001<o_00C?k`000001P21WnL010000000YLof0:G?mP2UcoH700000080_O7b100000001@3S
k=H00000000000000000i/Ri0080000000L0dZbc00000038X[8000000;jH/@000000]92`0080ZHRa
0P0000002@2AN;<000000000001ZI;T0000006Ihd@000000Bg7G02IWf@0200000080Di_c1`000004
09K2mP0400000037f>H0amSV0<OHiP<000000P3blkd2000004L0oooo1P0000030976n@040000002C
bOX0TlWj09?9nPT000000`2KcoX010000000Wm7i09oAn@2OdOT600000340oooo000]0?ooo`030000
000005/000000080?1Ad00@0000006HdO`1V=7l0ISAo0P00000208eTWP030000002NL9h0WW2N00<0
WW2N00<000000:akWP2/Nih00P2/Nih00`000000^HJN0;V6WP030;V6WP0300000035T9h0aI2N0080
aI2N00@000000=:LWP3BW9h0dYbN0P0000030>>JP0030000003bZ7D0lZQe00<00000@@3oool20000
000401<o_00000000000000000@0PIoW00<000000:G?mP2UcoH0102UcoH00`000000_O7b0;galP02
0;galP80000000L0gO[I0000003Sk=H000000>7Kd0000000flc;00<0000000<0dZbc000000000000
0P2nV;401P000000]92`0:V8/@2YR;40000009j0/P@0000000<0NWk;0000001VN=400P00000202IW
f@800000101CVo<00`000000U/;f09K2mP0409K2mP0400000037f>H0000000000080an7[00D00000
0?;c_@000000foVZ0000001:0?ooo`T000000P2CbOX00`000000Ulgj09O=nP0209O=nP030000002K
coX0Vloj0080Vloj2000000d0?ooo`00;@3oool00`000000001K0000000203`DM0030000001V=7l0
000000800000102=I9h00`000000WW2N09i`WP0209i`WP030000002/Nih0[7^N0080[7^N00<00000
0;V6WP2iQYh00`2iQYh00`000000aI2N0<F@WP020<F@WP030000003BW9h0dYbN0080dYbN0`000000
103SVX0000000?:XM@3bZ7D300000440oooo0P000000100C?k`0000007BOk01dWn`200000003086O
i`000000YLof00D0YLof00<000000;galP0000000`0000003037j^h000000000003Sk=H000000>7K
d0000000flc;0000003D_lH000000<bd``<000000P2nV;4010000000]92`0:V8/@2YR;4300000006
09N9``000000RXC60000001jO//000000P1VN=40100000009VOI0000001TV^h30000000305>Kl`00
0000U/;f00<0U/;f100000050<OQj`<00000C`3oool2000000<0B[ge4@00000g0?ooo`00;@3oool0
0`000000001K0000000203`DM08000000P1hEYd00`000000SFBN08eTWP0208eTWP030000002NL9h0
WW2N0080WW2N00<000000:akWP2/Nih00P2/Nih00`000000^HJN0;V6WP030;V6WP0300000035T9h0
aI2N0080aI2N00<000000=:LWP3BW9h00P3BW9h00`000000gjZM0=nZW@0200000080lZQe0`000012
0?ooo`<00000101dWn`:000000@0an[^100000002P3Qfm0000000=_<b`3Kc</000000=BoaP000000
c;C300000034Zl48000000060::?`@000000UhW30000002:Q<H000000P1jO//00`000000IWSA06Ih
d@03000000@0I9[^1@00000309O1m00300000037hN/0an7[00<0an7[00<000000=_iZP000000W03o
ool002d0oooo00<000000000F`0000000`00000407QFW@050000002=I9h0SFBN08eTWP000000102N
L9h00`000000[7^N0:akWP020:akWP030000002iQYh0^HJN0080^HJN00<000000<F@WP35T9h00`35
T9h01@000000dYbN0=:LWP3BW9h0000000@0gjZM1P0000120?ooo`04000000006JL00000000000@0
M9o/00<000000:O>m02Wc_@0102Wc_@010000000an[^0<OZkP37j^h2000000050<cQj0000000hM_@
0>7Kd00000000P3Kc</00`000000e;o6000000020<bd``0300000034Zl4000000080_:>o00D00000
0;BK_`000000ZiFo000000020::?`@050000002GRL<0000008Z4aP0000000P1jO//01@000000IWSA
06Ihd@000000LYWX00@0000000<0I9[^0000002G`O@01@2G`O@010000000an7[0<OQj`37hN/30000
00030=_iZP000000oooo09/0oooo000]0?ooo`030000000005/000000080FT:K00<0000007QFW@1h
EYd00P1hEYd01@000000SFBN08eTWP2=I9h0000000@0WW2N00<000000:akWP2/Nih00P2/Nih00`00
0000^HJN0;V6WP020;V6WP0300000035T9h0aI2N0080aI2N00<000000=:LWP3BW9h00P3BW9h01@00
0000gjZM0=nZW@3OZYd0000000<0k[ZL0`0000130?ooo`80000000D0IIoa000000000000M9o/0000
00060:O>m00300000037j^h000000080000000H0c>7X00000000000000000>7Kd00000020=_<b`03
0000003D_lH000000080c;C300<000000<B[`@0000000`2lXkl00`000000]9^o000000020:^E_`05
0000002RSl40000009N9``0000000P2:Q<H00`000000NWk;07Ynb`020000000307nIhP0000000000
00<0LYWX1P00000309O1m0<0000000@0`n[`0000003KnJX00000W@3oool002`0oooo1000000305Y2
V`050000001hEYd0N5JM07QFW@0000000`2=I9h00`000000WW2N09i`WP0209i`WP030000002/Nih0
[7^N0080[7^N00<000000;V6WP2iQYh00P2iQYh00`000000aI2N0<F@WP020<F@WP030000003BW9h0
dYbN0080dYbN00D000000=nZW@3OZYd0gjZM000000030>jjW0<00000@`3oool2000000<0IIoa2P00
00020<cQj08000000`3<en8010000000flc;0=_<b`0000020=BoaP060000003<]<<0c;C300000034
Zl4000000`2lXkl00`000000]9^o000000020:^E_`030000002RSl4000000080UhW300<0000008Z4
aP2:Q<H00`1jO//010000000OiWR07nIhP1oVN82000000<0LYWX00@0000009S0l`2H`?<0V<3c1000
00030<?Zl0800000WP3oool002`0oooo0`0000001@0[99D0000005Y2V`1J@Y/0000000<0N5JM00D0
000008eTWP2=I9h0SFBN0000000409i`WP030000002/Nih0[7^N0080[7^N00<000000;V6WP2iQYh0
0P2iQYh00`000000aI2N0<F@WP020<F@WP030000003BW9h0dYbN0080dYbN00D000000=nZW@3OZYd0
gjZM000000020>jjW0040000003kcIH00000000004@0oooo0P00000306FOl@030000002XcO<0Z<gc
00@0Z<gc00@000000<cQj000000000000`3<en82000000060<[>g0000000flc;0000003D_lH00000
0P3<]<<010000000a:_10<B[`@0000020;bS_`040000002dVkl0]9^o00000080ZiFo00@000000::?
`@2RSl4000000P2GRL<01P000000RXC60000001jO//0000008ZJg@8000000`1oVN800`000000LYWX
0000000709S0l`0300000033j_00`n[`00<00000WP3oool002`0oooo0P00000202/TU@040000001J
@Y/0FT:K000000<0N5JM00<0000008eTWP2=I9h00P2=I9h01@000000WW2N09i`WP2NL9h0000000@0
[7^N00<000000;V6WP2iQYh00P2iQYh00`000000aI2N0<F@WP020<F@WP030000003BW9h0dYbN0080
dYbN00@000000=nZW@3OZYd000000`3^^Y`010000000nlfF0000000000150?ooo`<000000P1UWo40
10000000Z<gc0:S=l`2XcO<4000000@0c=OR00<000000<[>g03:c]`00P3:c]`010000000e;o60=Bo
aP0000020<bd``0400000034Zl40a:_100000080_:>o00@000000;BK_`2dVkl000000P2[UKl01000
0000XXo10::?`@00000209N9``040000002:Q<H0RXC6000000@0RY[M00<0000007nIhP1oVN800P1o
VN84000000<0V<3c00<000000<?Zl00000000`00002O0?ooo`00;03oool200000080:bBE00@00000
05Y2V`1J@Y/000000`1hEYd00`000000SFBN08eTWP0208eTWP050000002NL9h0WW2N09i`WP000000
0P2/Nih500000080^HJN00<000000<F@WP35T9h00P35T9h00`000000dYbN0=:LWP020=:LWP040000
003OZYd0gjZM000000<0k[ZL00@000000?_=UP0000000000AP3oool00`000000EIod000000050000
00@0ZLcb00@000000<cGhP0000000000103:c]`00`000000alGH00000002000000050<bd``000000
a:_10<B[`@0000000`2lXkl010000000]9^o0;BK_`0000030:^E_`<0XXo10P2GRL<30000000309BL
f@000000000000<0RY[M0P0000000`1oVN80000009ZolP0309ZolPD0000000<0^O3d000000000000
X03oool002`0oooo00D000000000P0000000:bBE0000000205Y2V`030000001hEYd0N5JM0080N5JM
00@0000008eTWP2=I9h0SFBN1`0000020:n</P030000002iUK80^IFb00P000000`3BW9h01@000000
gjZM0=nZW@3OZYd000000080k[ZL00<000000?_=UP3kcIH00P0000170?ooo`050000001EWo@0EIod
05FOm00000001P2Yc?82000000D0b/kL00@000000<O5f037aMP000000P32_MD3000000030<B[`@00
0000_:>o0080_:>o00@000000;BK_`2dVkl000000`2[UKl20::?`@<000000P2LW]D01@000000U9cI
09BLf@2DW=T0000000@0RY[M0P00000609ZolP050000002il?@0^O3d0;W`m0000000X@3oool002`0
oooo00<000000000P00000000P0[99D00`000000FT:K0000000407QFW@030000002=I9h000000080
00000`2SQ;800`000000[hbb0:n</P020:n</P030000002iUK80^IFb0080^IFb00<000000<BN/P34
W[800P34W[83000000030=:LWP000000gjZM0080gjZM00P000000>jjW03^^Y`000000?_=UP000000
oMf3000004P0oooo0`000000101EWo@000000:W<lP2Yc?8300000080Z/[`00<000000<[>g03:c]`0
0P0000020<O5f00500000032_MD0`[gE0<:me@0000000P2m]]<>000000030:BRd`000000W9kE00<0
W9kE00@0000009BLf@2DW=T0U9cI0P000000102:V]d0000009bol02L_o0300000080V[ob00<00000
0;W`m00000000P00002R0?ooo`00;03oool00`00000000200000000202/TU@040000001J@Y/0FT:K
000000<0N5JM0P00000209Ik]0050000002SQ;80XhBb0:>4/P000000102_S;800`000000^IFb0;VE
/P020;VE/P0500000034W[80a9jb0<BN/P0000000`3>ZK@3000000030=nZW@000000k[ZL0080k[ZL
00D000000?_=UP000000oMf3000000190?ooo`0300000013WoD0000000<000001@2Zb_000`000000
b/kL000000040<O5f00500000032_MD0`[gE0<:me@0000000`2m]]<01@000000^;3A0;R`d@2h/=40
00000080/Z[A00D000000:^Ud@2[YM40ZjGA000000030:BRd`050000002LW]D0W9kE09bNe@000000
1@2DW=T00`000000W;o`09bol00409bol0@0000000<0ZO7e0000003oool0XP3oool002d0oooo0P00
000202/TU@050000001J@Y/0FT:K0000001hEYd00`00000409Ik]0050000002SQ;80XhBb0:>4/P00
00000`2_S;800`000000^IFb0;VE/P020;VE/P0500000034W[80a9jb0<BN/P000000103>ZK@01000
0000fKJf0000000000030>jjW0050000003kcIH000000?gMP`000000BP3oool200000080@ioe00<0
00000:[:l02Zb_000`2Zb_03000000<0alGH00D000000<:me@32_MD0`[gE000000030;ffd`030000
002h/=40^;3A0080^;3A00@000000;:Zd@2bZ]400000102[YM4010000000Y:;C0:BRd`00000409bN
e@030000002DW=T0U9cI0080000000<0W[k_000000000000102L_o000`000000ZO7e0:Wam@020000
0:@0oooo000]0?ooo`05000000000800000002/TU@0000000P1J@Y/2000000<0QG:f00D0000009Ik
]02FNk@0UW^d000000030:>4/P050000002_S;80[hbb0:n</P000000102iUK801@000000a9jb0<BN
/P34W[80000000<0cZVd00@000000=Vf]P3I][H0fKJf100000020?_=UP<00000C03oool200000004
04>Om@000000Z/[`0:[:l08000000`2Zb>l3000000@0`[gE00D000000;ffd`2m]]<0_KKC00000003
0;R`d@050000002bZ]40/Z[A0;:Zd@000000102[YM401@000000Y:;C0:BRd`2TX]<0000000@0W9kE
0P00000509jnk`80000000<0W;o`0000002YlOD00P00002V0?ooo`00;@3oool00`00000000200000
000202/TU@0405Y2V`000000L6Ni000000<0QG:f00D0000009Ik]02FNk@0UW^d000000030:>4/P05
0000002_S;80[hbb0:n</P0000000`2iUK800`000000a9jb0<BN/P020<BN/P050000003>ZK@0cZVd
0<jY]00000000P3I][H01@000000i<Fh0>C5^03TaKP000000080nlfF0P00001>0?ooo`D000001P2Z
b>l01P000000ZLK^000000000000`[gE000000@0_KKC00D000000;R`d@2h/=40^;3A000000040;:Z
d@050000002[YM40ZjGA0:^Ud@0000000`2TX]<00`000000W9kE09bNe@02000000030:2ok`000000
000000H0W[k_1000002W0?ooo`00;P3oool200000080:bBE00@00000071W^@1`IkT000000`25L[H0
1@000000UW^d09Ik]02FNk@0000000<0XhBb00D000000:n</P2_S;80[hbb000000030;VE/P050000
0034W[80a9jb0<BN/P0000000`3>ZK@01@000000fKJf0=Vf]P3I][H0000000<0i<Fh00<0k]Nk0000
000000000P00001@0?ooo`80000000<0<Yob0000000000000P2Zb>l2000000@0ZLK^0`0000020;ff
d`030000002h/=40^;3A0080^;3A00<000000;:Zd@2bZ]400P2bZ]400`000000ZjGA0:^Ud@020:^U
d@030000002TX]<0Y:;C00<000001@2P_nl200000080W[k_1000002Y0?ooo`00;P3oool01P000000
002002/TU@000000DeVk00000080L6Ni00@0000008Eb]P25L[H000000`2FNk@01@000000XhBb0:>4
/P2SQ;80000000<0[hbb00D000000;VE/P2iUK80^IFb000000030<BN/P050000003>ZK@0cZVd0<jY
]00000000P3I][H01P000000i<Fh0>C5^03TaKP000000>kG^`@00000DP3oool6000000D0ZLK^00<0
00000:S5kP2XaNh0500000020::okP030000002P_nl0X;o_00@0X;o_1@00002[0?ooo`00;P3oool0
0`00000000200000000205=I^`040000001`IkT0L6Ni000000<0QG:f00@0000009Ik]02FNk@00000
102SQ;8010000000[hbb0:n</P0000030;VE/P0500000034W[80a9jb0<BN/P0000000`3>ZK@01000
0000fKJf0=Vf]P0000020>C5^0040000003^ek/0k]Nk0>kG^`800000EP3oool20000000302FPk000
000000000080ZLK^00<000000:S5kP2XaNh00`2XaNh00`000000Yl?]0:O3k@020:O3k@030000002V
`^d0Y/;]0080Y/;]00<000000:C1k@2T`Nd00`2T`Nd00`000000X[o^0::okP020::okP030000002P
_nl0X;o_00H00000[@3oool002l0oooo00<0000002I4^`0000000P1CFK/010000000L6Ni071W^@00
000208Eb]P050000002FNk@0UW^d09Ik]00000000`2SQ;8010000000[hbb0:n</P0000030;VE/P05
00000034W[80a9jb0<BN/P0000000P3>ZK@01@000000fKJf0=Vf]P3I][H000000080i<Fh00<00000
0>kG^`3^ek/00`00001I0?ooo`H000000`2XaNh00`000000Yl?]0:O3k@030:O3k@030000002V`^d0
Y/;]00<0Y/;]00<000000:C1k@2T`Nd00P2T`Nd010000000X[o^0::okP2R_nh700000;00oooo000_
0?ooo`030000000VA;/000000080DeVk00@00000071W^@1`IkT000000P25L[H01@000000UW^d09Ik
]02FNk@000000080XhBb2P0000000`34W[8000000<jY]0020<jY]0040000003I][H0fKJf00000080
i<Fh00D000000>kG^`3^ek/000000?C[^`02000005d0oooo7000002e0?ooo`00;`3oool200000005
02I4^`000000DeVk05=I^`0000000P1`IkT01@000000QG:f08Eb]P000000UW^d00@0000000D0YYW8
0000002_WlL0[io7000000030;NUa`030000002o[LP000000080000000<0cZVd0000003I][H00P3I
][H010000000i<Fh0>C5^00000020>kG^`@00000I03oool>00000;`0oooo000`0?ooo`030000000V
A;/000000080DeVk00@00000071W^@1`IkT000000P25L[H20000000309bDbP000000YYW80080YYW8
00@000000:nOa`2_WlL000000`2gYLL01@000000_jg80;n]b0000000a[K:00<0000000D0fKJf0000
003TaKP0i<Fh000000020>kG^`040000003djk/0000000000?l0oooo<03oool00300oooo0P000000
1@0VA;/0000005=I^`1CFK/000000080L6Ni0P0000001@2ASld0000009bDbP2LU<X000000080YYW8
00@000000:nOa`2_WlL000000P2gYLL010000000_jg80;n]b00000030<JfbP80cKo=0P0000020>C5
^0070000003^ek/000000?C[^`000000jOV`0000003o0?oooc00oooo000`0?ooo`<0000000X09TBk
0000001CFK/00000071W^@000000Q8[B0000002ASld000000P2LU<X010000000YYW80:JIb0000002
0:nOa`040000002gYLL0]jG700000080_jg800@000000<JfbP36]/X000000P3=_ld02P000000e<_A
000000000000k]Nk0000003djk/000000>Wi/000003o0?oooc40oooo000a0?ooo`04000000009:h0
000000000080DeVk00@0000008B:dP24R]8000000P2ASld010000000W9C:09bDbP0000020:JIb003
0000002_WlL000000080]jG700<000000;n]b00000000P36]/X010000000cKo=0<foc@0000020=C;
d@80000000<0k]Nk0000003djk/00`00003o0?oooc40oooo000b0?ooo`80000000D09TBk00000000
0000LhKG0000000208B:dP030000002ASld000000080W9C:00H000000:JIb02VVLP000000:nOa`00
00020;NUa`030000002o[LP000000080a[K:00<000000<foc@0000000P3Dbm401@000000fMSG0=WH
e`000000m>^k00<00000o`3ooolb0?ooo`00<P3oool500000080LhKG00H0000008B:dP24R]800000
096?c@00000209bDbP070000002VVLP000000:nOa`000000]jG7000000020;n]b00300000036]/X0
00000080cKo=00<000000=C;d@0000000P3If=L2000000030?C[^`00000000000?l0oooo<`3oool0
03<0oooo00@00000000T[P000000GX7M0P0000004@1cQ]L0000008B:dP24R]800000096?c@000000
W9C:0000002VVLP000000:nOa`000000]jG70000002o[LP000000080a[K:00D000000<foc@000000
e<_A000000020=WHe`030000003Ki]d0fnKM00800000o`3ooold0?ooo`00=03oool40000000805j1
g@000000LhKG00000024R]800000096?c@00000209bDbPP0000000L0a[K:0000003=_ld000000=C;
d@000000fMSG0080000000<0fnKM0000000000000P00003o0?oooc@0oooo000e0?ooo`0;00000012
O>80@WcR0000001NPMd0000007>6e`000000Q8[B0000002ASld0100000000`2Z^><000000:jkh`02
000000080;C2i@00000000000000003Dbm4000000=WHe`0000020=_Vg@040000003Hm>8000000000
0?l0oooo=@3oool003D0oooo0`0000000`12O>800000000000050000000309bbi`000000XK?U00@0
000000<0/KkT0000002d`^D00P0000001@2hc>T000000=WHe`000000fnKM00D00000o`3ooolf0?oo
o`00=P3oool400000003049lhP000000GX7M00/0000000<0]<;U0000002hc>T02P00003o0?ooocL0
oooo000g0?ooo`D0000000<0@WcR0000000000005@00003o0?ooocP0oooo000i0?oooaX00000o`3o
ooli0?ooo`00>`3ooolF00000?l0oooo>`3oool003d0oooo4P00003o0?ooocd0oooo00100?ooo`D0
000000<0Om7o00000026e_l01000003o0?oood00oooo00140?ooo`@00000o`3ooom40?ooo`00o`3o
oon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3o
oon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3o
oon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3o
oon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3o
oon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3o
oon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3ooon<0?ooo`00o`3o
oon<0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 394.063}, {201.188, 0}} -> {-0.0798061, -0.0384154, \
0.0120271, 0.0120271}, {{9.9375, 126.875}, {196.375, 4.75}} -> {0.1378, \
-0.0174463, 0.00776924, 0.00776924}, {{138.563, 255.438}, {196.375, 4.75}} -> \
{-0.96766, -0.272668, 0.011003, 0.011003}, {{267.125, 384.063}, {196.375, \
4.75}} -> {-2.176, -0.368659, 0.012219, 0.012219}}]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Contour plots", "Section"],

Cell["\<\
Contour or density plots often prove very useful in locating extrema, ridges, \
or other important features of functions of two or more dimensions.  These \
plots can also help in finding starting points for numerical solution of \
equations.\
\>", "Text"],

Cell[TextData[{
  "The basic syntax is\n\t",
  ButtonBox["DensityPlot",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["f[x,y],{x,xmin,xmax},{y,ymin,ymax}", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\t",
    FontWeight->"Bold"],
  "or\n\t",
  ButtonBox["ContourPlot",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["f[x,y],{x,xmin,xmax},{y,ymin,ymax}", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\t",
    FontWeight->"Bold"]
}], "Text"],

Cell[TextData[{
  "Common options include ",
  StyleBox["PlotPoints", "InlineInput",
    FontWeight->"Bold"],
  " to increase the resolution of the figure or ",
  StyleBox["ColorFunction", "InlineInput",
    FontWeight->"Bold"],
  " to apply a coloration function based upon the values of the function.  \
For ",
  StyleBox["ContourPlot", "InlineInput",
    FontWeight->"Bold"],
  " the option ",
  StyleBox["Contours\[Rule]", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["n", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[" ",
    FontSlant->"Italic"],
  "specifies that ",
  StyleBox["n", "InlineInput"],
  " equally spaced contours or ",
  StyleBox["Contours\[Rule]", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["list", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[" ",
    FontWeight->"Bold"],
  "specifies particular contours in the form of a list."
}], "Text"],

Cell["\<\
Evaluate the following cells to compare density and contour plots of a simple \
function.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DensityPlot[
        Exp[\(-\((x\^2 + y\^2)\)\)/20] Cos[x\ y], {x, \(-4\), 4}, {y, \(-4\), 
          4}, PlotPoints \[Rule] 100, 
        ColorFunction \[Rule] Hue];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% DensityGraphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.120192 0.5 0.120192 [
[.01923 -0.0125 -9.15625 -15.6875 ]
[.01923 -0.0125 9.15625 0 ]
[.25962 -0.0125 -9.15625 -15.6875 ]
[.25962 -0.0125 9.15625 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.74038 -0.0125 -5 -15.6875 ]
[.74038 -0.0125 5 0 ]
[.98077 -0.0125 -5 -15.6875 ]
[.98077 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .01923 -18.3125 -7.84375 ]
[-0.0125 .01923 0 7.84375 ]
[-0.0125 .25962 -18.3125 -7.84375 ]
[-0.0125 .25962 0 7.84375 ]
[-0.0125 .5 -10 -7.84375 ]
[-0.0125 .5 0 7.84375 ]
[-0.0125 .74038 -10 -7.84375 ]
[-0.0125 .74038 0 7.84375 ]
[-0.0125 .98077 -10 -7.84375 ]
[-0.0125 .98077 0 7.84375 ]
[ 0 0 -0.125 0 ]
[ 0 1 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.01923 0 m
.01923 .00625 L
s
gsave
.01923 -0.0125 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.25962 0 m
.25962 .00625 L
s
gsave
.25962 -0.0125 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.74038 0 m
.74038 .00625 L
s
gsave
.74038 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.98077 0 m
.98077 .00625 L
s
gsave
.98077 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.07933 0 m
.07933 .00375 L
s
.13942 0 m
.13942 .00375 L
s
.19952 0 m
.19952 .00375 L
s
.31971 0 m
.31971 .00375 L
s
.37981 0 m
.37981 .00375 L
s
.4399 0 m
.4399 .00375 L
s
.5601 0 m
.5601 .00375 L
s
.62019 0 m
.62019 .00375 L
s
.68029 0 m
.68029 .00375 L
s
.80048 0 m
.80048 .00375 L
s
.86058 0 m
.86058 .00375 L
s
.92067 0 m
.92067 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .01923 m
.00625 .01923 L
s
gsave
-0.0125 .01923 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
0 .25962 m
.00625 .25962 L
s
gsave
-0.0125 .25962 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .5 m
.00625 .5 L
s
gsave
-0.0125 .5 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .74038 m
.00625 .74038 L
s
gsave
-0.0125 .74038 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .98077 m
.00625 .98077 L
s
gsave
-0.0125 .98077 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .07933 m
.00375 .07933 L
s
0 .13942 m
.00375 .13942 L
s
0 .19952 m
.00375 .19952 L
s
0 .31971 m
.00375 .31971 L
s
0 .37981 m
.00375 .37981 L
s
0 .4399 m
.00375 .4399 L
s
0 .5601 m
.00375 .5601 L
s
0 .62019 m
.00375 .62019 L
s
0 .68029 m
.00375 .68029 L
s
0 .80048 m
.00375 .80048 L
s
0 .86058 m
.00375 .86058 L
s
0 .92067 m
.00375 .92067 L
s
.25 Mabswid
0 0 m
0 1 L
s
.01923 .99375 m
.01923 1 L
s
.25962 .99375 m
.25962 1 L
s
.5 .99375 m
.5 1 L
s
.74038 .99375 m
.74038 1 L
s
.98077 .99375 m
.98077 1 L
s
.125 Mabswid
.07933 .99625 m
.07933 1 L
s
.13942 .99625 m
.13942 1 L
s
.19952 .99625 m
.19952 1 L
s
.31971 .99625 m
.31971 1 L
s
.37981 .99625 m
.37981 1 L
s
.4399 .99625 m
.4399 1 L
s
.5601 .99625 m
.5601 1 L
s
.62019 .99625 m
.62019 1 L
s
.68029 .99625 m
.68029 1 L
s
.80048 .99625 m
.80048 1 L
s
.86058 .99625 m
.86058 1 L
s
.92067 .99625 m
.92067 1 L
s
.25 Mabswid
0 1 m
1 1 L
s
.99375 .01923 m
1 .01923 L
s
.99375 .25962 m
1 .25962 L
s
.99375 .5 m
1 .5 L
s
.99375 .74038 m
1 .74038 L
s
.99375 .98077 m
1 .98077 L
s
.125 Mabswid
.99625 .07933 m
1 .07933 L
s
.99625 .13942 m
1 .13942 L
s
.99625 .19952 m
1 .19952 L
s
.99625 .31971 m
1 .31971 L
s
.99625 .37981 m
1 .37981 L
s
.99625 .4399 m
1 .4399 L
s
.99625 .5601 m
1 .5601 L
s
.99625 .62019 m
1 .62019 L
s
.99625 .68029 m
1 .68029 L
s
.99625 .80048 m
1 .80048 L
s
.99625 .86058 m
1 .86058 L
s
.99625 .92067 m
1 .92067 L
s
.25 Mabswid
1 0 m
1 1 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.01923 .01923 translate
.96154 .96154 scale
300 string
100 100 8 [100 0 0 100 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
21FF002EFF0028FF000EFF0000FF1C00FF5700FF9A00FFE000DFFF00ACFF008BFF0082FF
0092FF00BCFF00FCFF00FFB400FF5C00FF0547FF0080FF009FFF00A0FF0081FF0044FF00
00FF1000FF7400FFDE00BCFF0068FF002EFF0016FF0022FF0052FF00A1FF00FFF600FF80
00FF0961FF00B5FF00E9FF00F7FF00DCFF009CFF003CFF0000FF3900FFB800CDFF0063FF
0015FF1300FF1300FF0015FF0063FF00CDFF00FFB800FF393CFF009CFF00DCFF00F7FF00
E9FF00B5FF0061FF0000FF0900FF8000FFF600A1FF0052FF0022FF0016FF002EFF0068FF
00BCFF00FFDE00FF7400FF1044FF0081FF00A0FF009FFF0080FF0047FF0000FF0500FF5C
00FFB400FCFF00BCFF0092FF0082FF008BFF00ACFF00DFFF00FFE000FF9A00FF5700FF1C
0EFF0028FF002EFF0021FF00
2EFF0029FF0010FF0000FF1900FF5100FF9300FFD800E6FF00B1FF008DFF007FFF0089FF
00ADFF00E7FF00FFCC00FF7600FF1E31FF0073FF009CFF00AAFF0098FF0068FF001DFF00
00FF4100FFAA00ECFF008DFF0044FF0017FF000CFF0026FF0063FF00BDFF00FFD400FF5C
19FF0081FF00CFFF00FBFF00FFFE00DFFF0098FF0033FF0000FF4400FFC500BFFF0055FF
0008FF2000FF2000FF0008FF0055FF00BFFF00FFC500FF4433FF0098FF00DFFF00FFFE00
FBFF00CFFF0081FF0019FF0000FF5C00FFD400BDFF0063FF0026FF000CFF0017FF0044FF
008DFF00ECFF00FFAA00FF411DFF0068FF0098FF00AAFF009CFF0073FF0031FF0000FF1E
00FF7600FFCC00E7FF00ADFF0089FF007FFF008DFF00B1FF00E6FF00FFD800FF9300FF51
00FF1910FF0029FF002EFF00
28FF0010FF0000FF1800FF4F00FF8F00FFD300ECFF00B5FF008FFF007DFF0083FF00A1FF
00D6FF00FFE200FF8E00FF361CFF0063FF0096FF00AEFF00A9FF0085FF0045FF0000FF11
00FF7700FFE200B8FF0061FF0023FF0005FF0009FF0031FF007AFF00DCFF00FFAF00FF35
3EFF00A0FF00E7FF00FFF300FFF600E0FF0093FF002AFF0000FF5100FFD300B0FF0047FF
0400FF2D00FF2D00FF0400FF0047FF00B0FF00FFD300FF512AFF0093FF00E0FF00FFF600
FFF300E7FF00A0FF003EFF0000FF3500FFAF00DCFF007AFF0031FF0009FF0005FF0023FF
0061FF00B8FF00FFE200FF7700FF1145FF0085FF00A9FF00AEFF0096FF0063FF001CFF00
00FF3600FF8E00FFE200D6FF00A1FF0083FF007DFF008FFF00B5FF00ECFF00FFD300FF8F
00FF4F00FF1810FF0028FF00
0EFF0000FF1900FF4F00FF8E00FFD000EFFF00B9FF0091FF007CFF007FFF0098FF00C9FF
00FFF300FFA200FF4B09FF0054FF008DFF00AFFF00B4FF009BFF0066FF0018FF0000FF47
00FFB100E4FF0085FF0039FF0008FF0600FF000CFF0042FF0095FF00FFFF00FF8700FF0D
62FF00BFFF00FFFF00FFE400FFEF00DFFF008CFF001EFF0000FF5F00FFE200A1FF0039FF
1200FF3A00FF3A00FF1200FF0039FF00A1FF00FFE200FF5F1EFF008CFF00DFFF00FFEF00
FFE400FFFF00BFFF0062FF0000FF0D00FF8700FFFF0095FF0042FF000CFF0600FF0008FF
0039FF0085FF00E4FF00FFB100FF4718FF0066FF009BFF00B4FF00AFFF008DFF0054FF00
09FF0000FF4B00FFA200FFF300C9FF0098FF007FFF007CFF0091FF00B9FF00EFFF00FFD0
00FF8E00FF4F00FF190EFF00
00FF1C00FF5100FF8F00FFD000F0FF00BAFF0092FF007CFF007CFF0092FF00BFFF00FEFF
00FFB300FF5E00FF0845FF0083FF00ACFF00BBFF00ACFF0081FF003CFF0000FF1C00FF83
00FFEE00ACFF0055FF0016FF0B00FF0B00FF0016FF0059FF00B6FF00FFD800FF5D1BFF00
87FF00DDFF00FFEA00FFD700FFEA00DDFF0084FF0012FF0000FF6D00FFF10092FF002AFF
2000FF4700FF4700FF2000FF002AFF0092FF00FFF100FF6D12FF0084FF00DDFF00FFEA00
FFD700FFEA00DDFF0087FF001BFF0000FF5D00FFD800B6FF0059FF0016FF0B00FF0B00FF
0016FF0055FF00ACFF00FFEE00FF8300FF1C3CFF0081FF00ACFF00BBFF00ACFF0083FF00
45FF0000FF0800FF5E00FFB300FEFF00BFFF0092FF007CFF007CFF0092FF00BAFF00F0FF
00FFD000FF8F00FF5100FF1C
00FF5700FF9300FFD300EFFF00BAFF0092FF007CFF007AFF008EFF00B8FF00F4FF00FFC0
00FF6D00FF1837FF0079FF00A8FF00BEFF00B9FF0097FF005BFF0009FF0000FF5800FFC2
00D4FF0076FF002AFF0600FF1900FF0900FF0027FF0076FF00DCFF00FFAD00FF3145FF00
ACFF00FAFF00FFD600FFCC00FFE700D8FF007AFF0004FF0000FF7D00FDFF0082FF001BFF
2E00FF5500FF5500FF2E00FF001BFF0082FF00FDFF00FF7D04FF007AFF00D8FF00FFE700
FFCC00FFD600FAFF00ACFF0045FF0000FF3100FFAD00DCFF0076FF0027FF0900FF1900FF
0600FF002AFF0076FF00D4FF00FFC200FF5809FF005BFF0097FF00B9FF00BEFF00A8FF00
79FF0037FF0000FF1800FF6D00FFC000F4FF00B8FF008EFF007AFF007CFF0092FF00BAFF
00EFFF00FFD300FF9300FF57
00FF9A00FFD800ECFF00B9FF0092FF007CFF0079FF008CFF00B3FF00EDFF00FFCA00FF78
00FF242BFF0070FF00A3FF00BFFF00C1FF00A8FF0075FF002AFF0000FF3200FF9A00FBFF
0098FF0043FF0005FF1C00FF1E00FF0000FF003FFF0099FF00FFF800FF7F00FF036FFF00
D0FF00FFE900FFC400FFC300FFE700D2FF006EFF0000FF0900FF8E00ECFF0072FF000CFF
3C00FF6200FF6200FF3C00FF000CFF0072FF00ECFF00FF8E00FF096EFF00D2FF00FFE700
FFC300FFC400FFE900D0FF006FFF0000FF0300FF7F00FFF80099FF003FFF0000FF1E00FF
1C00FF0005FF0043FF0098FF00FBFF00FF9A00FF322AFF0075FF00A8FF00C1FF00BFFF00
A3FF0070FF002BFF0000FF2400FF7800FFCA00EDFF00B3FF008CFF0079FF007CFF0092FF
00B9FF00ECFF00FFD800FF9A
00FFE000E6FF00B5FF0091FF007CFF007AFF008CFF00B2FF00EAFF00FFCF00FF8000FF2D
22FF0068FF009EFF00BFFF00C7FF00B5FF0089FF0045FF0000FF1100FF7500FFDE00BAFF
005FFF0016FF1800FF2A00FF1B00FF0012FF005EFF00C1FF00FFCA00FF4E2BFF0099FF00
F2FF00FFD000FFB400FFBC00FFE800CAFF0061FF0000FF1A00FF9F00DBFF0061FF0300FF
4B00FF7000FF7000FF4B00FF0300FF0061FF00DBFF00FF9F00FF1A61FF00CAFF00FFE800
FFBC00FFB400FFD000F2FF0099FF002BFF0000FF4E00FFCA00C1FF005EFF0012FF1B00FF
2A00FF1800FF0016FF005FFF00BAFF00FFDE00FF7500FF1145FF0089FF00B5FF00C7FF00
BFFF009EFF0068FF0022FF0000FF2D00FF8000FFCF00EAFF00B2FF008CFF007AFF007CFF
0091FF00B5FF00E6FF00FFE0
00DFFF00B1FF008FFF007CFF007CFF008EFF00B3FF00EAFF00FFD100FF8400FF331BFF00
61FF0099FF00BEFF00CBFF00BFFF0099FF005CFF000AFF0000FF5500FFBC00DBFF007BFF
002BFF0F00FF2E00FF2E00FF0F00FF002DFF0084FF00EFFF00FF9700FF1B5AFF00C2FF00
FFEC00FFB800FFA600FFB800FFEC00BFFF0052FF0000FF2B00FFB200C9FF0050FF1200FF
5900FF7E00FF7E00FF5900FF1200FF0050FF00C9FF00FFB200FF2B52FF00BFFF00FFEC00
FFB800FFA600FFB800FFEC00C2FF005AFF0000FF1B00FF9700EFFF0084FF002DFF0F00FF
2E00FF2E00FF0F00FF002BFF007BFF00DBFF00FFBC00FF550AFF005CFF0099FF00BFFF00
CBFF00BEFF0099FF0061FF001BFF0000FF3300FF8400FFD100EAFF00B3FF008EFF007CFF
007CFF008FFF00B1FF00DFFF
00ACFF008DFF007DFF007FFF0092FF00B8FF00EDFF00FFCF00FF8400FF3418FF005EFF00
96FF00BCFF00CDFF00C5FF00A5FF006DFF0021FF0000FF3900FF9E00F9FF0097FF0041FF
0100FF2C00FF3A00FF2900FF0005FF004FFF00B0FF00FFDD00FF6217FF0089FF00EAFF00
FFCD00FFA300FF9B00FFB600FFF300B3FF0042FF0000FF3E00FFC500B6FF003FFF2200FF
6800FF8C00FF8C00FF6800FF2200FF003FFF00B6FF00FFC500FF3E42FF00B3FF00FFF300
FFB600FF9B00FFA300FFCD00EAFF0089FF0017FF0000FF6200FFDD00B0FF004FFF0005FF
2900FF3A00FF2C00FF0100FF0041FF0097FF00F9FF00FF9E00FF3921FF006DFF00A5FF00
C5FF00CDFF00BCFF0096FF005EFF0018FF0000FF3400FF8400FFCF00EDFF00B8FF0092FF
007FFF007DFF008DFF00ACFF
008BFF007FFF0083FF0098FF00BFFF00F4FF00FFCA00FF8000FF3318FF005CFF0094FF00
BBFF00CEFF00C9FF00ADFF007BFF0034FF0000FF2300FF8500FFEB00B1FF0057FF000DFF
2600FF3F00FF3B00FF1A00FF0023FF0079FF00E3FF00FFA600FF2B4CFF00B8FF00FFEF00
FFB100FF9200FF9400FFB800FFFB00A4FF002FFF0000FF5200FFDA00A2FF002DFF3200FF
7600FF9A00FF9A00FF7600FF3200FF002DFF00A2FF00FFDA00FF522FFF00A4FF00FFFB00
FFB800FF9400FF9200FFB100FFEF00B8FF004CFF0000FF2B00FFA600E3FF0079FF0023FF
1A00FF3B00FF3F00FF2600FF000DFF0057FF00B1FF00FFEB00FF8500FF2334FF007BFF00
ADFF00C9FF00CEFF00BBFF0094FF005CFF0018FF0000FF3300FF8000FFCA00F4FF00BFFF
0098FF0083FF007FFF008BFF
0082FF0089FF00A1FF00C9FF00FEFF00FFC000FF7800FF2D1BFF005EFF0094FF00BBFF00
CEFF00CCFF00B3FF0084FF0041FF0000FF1100FF7000FFD400C7FF006BFF001DFF1D00FF
4000FF4700FF3200FF0100FF0049FF00AAFF00FFE500FF6C0DFF0081FF00E5FF00FFCB00
FF9700FF8300FF8F00FFBC00F8FF0093FF001BFF0000FF6800FFEF008FFF001BFF4200FF
8500FFA700FFA700FF8500FF4200FF001BFF008FFF00FFEF00FF681BFF0093FF00F8FF00
FFBC00FF8F00FF8300FF9700FFCB00E5FF0081FF000DFF0000FF6C00FFE500AAFF0049FF
0100FF3200FF4700FF4000FF1D00FF001DFF006BFF00C7FF00FFD400FF7000FF1141FF00
84FF00B3FF00CCFF00CEFF00BBFF0094FF005EFF001BFF0000FF2D00FF7800FFC000FEFF
00C9FF00A1FF0089FF0082FF
0092FF00ADFF00D6FF00FFF300FFB300FF6D00FF2422FF0061FF0096FF00BBFF00CEFF00
CDFF00B6FF0089FF004AFF0000FF0500FF6000FFC200DAFF007DFF002CFF1300FF3D00FF
4E00FF4300FF1D00FF0022FF0079FF00E2FF00FFA900FF2F47FF00B4FF00FFEF00FFAA00
FF8100FF7700FF8E00FFC300EAFF0081FF0006FF0000FF7E00FAFF007AFF0008FF5300FF
9400FFB500FFB500FF9400FF5300FF0008FF007AFF00FAFF00FF7E06FF0081FF00EAFF00
FFC300FF8E00FF7700FF8100FFAA00FFEF00B4FF0047FF0000FF2F00FFA900E2FF0079FF
0022FF1D00FF4300FF4E00FF3D00FF1300FF002CFF007DFF00DAFF00FFC200FF6000FF05
4AFF0089FF00B6FF00CDFF00CEFF00BBFF0096FF0061FF0022FF0000FF2400FF6D00FFB3
00FFF300D6FF00ADFF0092FF
00BCFF00E7FF00FFE200FFA200FF5E00FF182BFF0068FF0099FF00BCFF00CEFF00CCFF00
B6FF008BFF004EFF0002FF0000FF5600FFB500E9FF008CFF0039FF0A00FF3900FF5100FF
5000FF3300FF0002FF004FFF00B0FF00FFE000FF690EFF0081FF00E6FF00FFC600FF8C00
FF6F00FF7000FF9000FFCE00D9FF006CFF0000FF1100FF9600E3FF0065FF0900FF6300FF
A200FFC300FFC300FFA200FF6300FF0900FF0065FF00E3FF00FF9600FF116CFF00D9FF00
FFCE00FF9000FF7000FF6F00FF8C00FFC600E6FF0081FF000EFF0000FF6900FFE000B0FF
004FFF0002FF3300FF5000FF5100FF3900FF0A00FF0039FF008CFF00E9FF00FFB500FF56
02FF004EFF008BFF00B6FF00CCFF00CEFF00BCFF0099FF0068FF002BFF0000FF1800FF5E
00FFA200FFE200E7FF00BCFF
00FCFF00FFCC00FF8E00FF4B00FF0837FF0070FF009EFF00BEFF00CDFF00C9FF00B3FF00
89FF004EFF0004FF0000FF5100FFAE00F2FF0097FF0043FF0100FF3500FF5200FF5800FF
4400FF1700FF002CFF0085FF00EEFF00FF9E00FF274DFF00B9FF00FFE900FFA100FF7300
FF6100FF6C00FF9600FFDB00C6FF0054FF0000FF2900FFAF00CCFF0050FF1C00FF7400FF
B100FFD100FFD100FFB100FF7400FF1C00FF0050FF00CCFF00FFAF00FF2954FF00C6FF00
FFDB00FF9600FF6C00FF6100FF7300FFA100FFE900B9FF004DFF0000FF2700FF9E00EEFF
0085FF002CFF1700FF4400FF5800FF5200FF3500FF0100FF0043FF0097FF00F2FF00FFAE
00FF5104FF004EFF0089FF00B3FF00C9FF00CDFF00BEFF009EFF0070FF0037FF0000FF08
00FF4B00FF8E00FFCC00FCFF
00FFB400FF7600FF3609FF0045FF0079FF00A3FF00BFFF00CBFF00C5FF00ADFF0084FF00
4AFF0002FF0000FF5100FFAB00F7FF009DFF004BFF0004FF3000FF5200FF5D00FF5000FF
2B00FF0011FF0062FF00C4FF00FFCD00FF581CFF008CFF00F0FF00FFBD00FF8000FF5D00
FF5700FF6D00FF9F00FFEC00B0FF003BFF0000FF4400FFC900B4FF003AFF2F00FF8500FF
C000FFDF00FFDF00FFC000FF8500FF2F00FF003AFF00B4FF00FFC900FF443BFF00B0FF00
FFEC00FF9F00FF6D00FF5700FF5D00FF8000FFBD00F0FF008CFF001CFF0000FF5800FFCD
00C4FF0062FF0011FF2B00FF5000FF5D00FF5200FF3000FF0004FF004BFF009DFF00F7FF
00FFAB00FF5102FF004AFF0084FF00ADFF00C5FF00CBFF00BFFF00A3FF0079FF0045FF00
09FF0000FF3600FF7600FFB4
00FF5C00FF1E1CFF0054FF0083FF00A8FF00BFFF00C7FF00BFFF00A5FF007BFF0041FF00
00FF0500FF5600FFAE00F7FF00A0FF004FFF0009FF2D00FF5100FF6000FF5800FF3900FF
0400FF0045FF00A0FF00FFF600FF8500FF1160FF00C8FF00FFDC00FF9400FF6300FF4D00
FF5100FF7200FFAD00FFFF0097FF0020FF0000FF6000FFE4009BFF0024FF4300FF9500FF
CF00FFEC00FFEC00FFCF00FF9500FF4300FF0024FF009BFF00FFE400FF6020FF0097FF00
FFFF00FFAD00FF7200FF5100FF4D00FF6300FF9400FFDC00C8FF0060FF0000FF1100FF85
00FFF600A0FF0045FF0400FF3900FF5800FF6000FF5100FF2D00FF0009FF004FFF00A0FF
00F7FF00FFAE00FF5600FF0541FF007BFF00A5FF00BFFF00C7FF00BFFF00A8FF0083FF00
54FF001CFF0000FF1E00FF5C
00FF0531FF0063FF008DFF00ACFF00BEFF00C1FF00B5FF0099FF006DFF0034FF0000FF11
00FF6000FFB500F2FF009DFF004FFF000AFF2B00FF5100FF6200FF5E00FF4400FF1400FF
002FFF0084FF00E7FF00FFAB00FF3937FF00A2FF00FFFD00FFAC00FF7000FF4C00FF4100
FF5100FF7B00FFBE00E7FF007CFF0003FF0000FF7D00FFFF0081FF000DFF5600FFA600FF
DD00FFFA00FFFA00FFDD00FFA600FF5600FF000DFF0081FF00FFFF00FF7D03FF007CFF00
E7FF00FFBE00FF7B00FF5100FF4100FF4C00FF7000FFAC00FFFD00A2FF0037FF0000FF39
00FFAB00E7FF0084FF002FFF1400FF4400FF5E00FF6200FF5100FF2B00FF000AFF004FFF
009DFF00F2FF00FFB500FF6000FF1134FF006DFF0099FF00B5FF00C1FF00BEFF00ACFF00
8DFF0063FF0031FF0000FF05
47FF0073FF0096FF00AFFF00BBFF00B9FF00A8FF0089FF005CFF0021FF0000FF2300FF70
00FFC200E9FF0097FF004BFF0009FF2B00FF5000FF6300FF6100FF4A00FF1F00FF001EFF
006EFF00CCFF00FFCA00FF5C13FF007FFF00E3FF00FFC600FF8200FF5200FF3A00FF3B00
FF5600FF8900FFD400CDFF005EFF0000FF1B00FF9B00E3FF0067FF0800FF6A00FFB600FF
EC00FFFF00F8FF00F8EC00FFB600FF6A00FF0800FF0067FF00E3FF00FF9B00FF1B5EFF00
CDFF00FFD400FF8900FF5600FF3B00FF3A00FF5200FF8200FFC600E3FF007FFF0013FF00
00FF5C00FFCA00CCFF006EFF001EFF1F00FF4A00FF6100FF6300FF5000FF2B00FF0009FF
004BFF0097FF00E9FF00FFC200FF7000FF2321FF005CFF0089FF00A8FF00B9FF00BBFF00
AFFF0096FF0073FF0047FF00
80FF009CFF00AEFF00B4FF00ACFF0097FF0075FF0045FF000AFF0000FF3900FF8500FFD4
00DAFF008CFF0043FF0004FF2D00FF5100FF6300FF6200FF4D00FF2600FF0014FF005FFF
00B8FF00FFE300FF7800FF0B60FF00C5FF00FFE000FF9500FF5D00FF3A00FF2E00FF3B00
FF6000FF9C00FFED00AFFF003EFF0000FF3C00FFBB00C5FF004DFF1F00FF7D00FFC700FF
FA00FFFF00EBFF00EBFA00FFC700FF7D00FF1F00FF004DFF00C5FF00FFBB00FF3C3EFF00
AFFF00FFED00FF9C00FF6000FF3B00FF2E00FF3A00FF5D00FF9500FFE000C5FF0060FF00
00FF0B00FF7800FFE300B8FF005FFF0014FF2600FF4D00FF6200FF6300FF5100FF2D00FF
0004FF0043FF008CFF00DAFF00FFD400FF8500FF390AFF0045FF0075FF0097FF00ACFF00
B4FF00AEFF009CFF0080FF00
9FFF00AAFF00A9FF009BFF0081FF005BFF002AFF0000FF1100FF5500FF9E00FFEB00C7FF
007DFF0039FF0100FF3000FF5100FF6200FF6100FF4D00FF2800FF000EFF0056FF00ABFF
00FFF500FF8E00FF2346FF00ABFF00FFF800FFA900FF6A00FF3E00FF2800FF2900FF4000
FF6F00FFB300F5FF008EFF001BFF0000FF5E00FFDC00A7FF0032FF3600FF9100FFD700FF
FF00F7FF00DEFF00DEFF00F7D700FF9100FF3600FF0032FF00A7FF00FFDC00FF5E1BFF00
8EFF00F5FF00FFB300FF6F00FF4000FF2900FF2800FF3E00FF6A00FFA900FFF800ABFF00
46FF0000FF2300FF8E00FFF500ABFF0056FF000EFF2800FF4D00FF6100FF6200FF5100FF
3000FF0100FF0039FF007DFF00C7FF00FFEB00FF9E00FF5500FF112AFF005BFF0081FF00
9BFF00A9FF00AAFF009FFF00
A0FF0098FF0085FF0066FF003CFF0009FF0000FF3200FF7500FFBC00F9FF00B1FF006BFF
002CFF0A00FF3500FF5200FF6000FF5E00FF4A00FF2600FF000EFF0053FF00A4FF00FFFF
00FF9C00FF3532FF0096FF00F3FF00FFBB00FF7700FF4500FF2700FF1D00FF2A00FF4C00
FF8400FFCF00D4FF006AFF0000FF0800FF8200FFFE0088FF0017FF4E00FFA400FFE800FF
FF00E9FF00D1FF00D1FF00E9E800FFA400FF4E00FF0017FF0088FF00FFFE00FF8200FF08
6AFF00D4FF00FFCF00FF8400FF4C00FF2A00FF1D00FF2700FF4500FF7700FFBB00F3FF00
96FF0032FF0000FF3500FF9C00FFFF00A4FF0053FF000EFF2600FF4A00FF5E00FF6000FF
5200FF3500FF0A00FF002CFF006BFF00B1FF00F9FF00FFBC00FF7500FF3209FF003CFF00
66FF0085FF0098FF00A0FF00
81FF0068FF0045FF0018FF0000FF1C00FF5800FF9A00FFDE00DBFF0097FF0057FF001DFF
1300FF3900FF5200FF5D00FF5800FF4400FF1F00FF0014FF0056FF00A4FF00FCFF00FFA3
00FF3F24FF0087FF00E3FF00FFCA00FF8400FF4E00FF2900FF1700FF1A00FF3200FF5F00
FF9E00FFEF00B0FF0044FF0000FF2F00FFA800DEFF0069FF0300FF6500FFB700FFF800FF
FF00DBFF00C4FF00C4FF00DBF800FFB700FF6500FF0300FF0069FF00DEFF00FFA800FF2F
44FF00B0FF00FFEF00FF9E00FF5F00FF3200FF1A00FF1700FF2900FF4E00FF8400FFCA00
E3FF0087FF0024FF0000FF3F00FFA300FCFF00A4FF0056FF0014FF1F00FF4400FF5800FF
5D00FF5200FF3900FF1300FF001DFF0057FF0097FF00DBFF00FFDE00FF9A00FF5800FF1C
18FF0045FF0068FF0081FF00
44FF001DFF0000FF1100FF4700FF8300FFC200FBFF00BAFF007BFF0041FF000DFF1D00FF
3D00FF5100FF5800FF5000FF3900FF1400FF001EFF005FFF00ABFF00FFFF00FFA300FF43
1EFF007DFF00D7FF00FFD600FF8F00FF5500FF2C00FF1500FF1000FF1F00FF4200FF7700
FFBE00EBFF0088FF001BFF0000FF5800FFCE00BBFF004AFF1E00FF7C00FFCB00FFFF00F7
FF00CDFF00B8FF00B8FF00CDFF00F7CB00FF7C00FF1E00FF004AFF00BBFF00FFCE00FF58
1BFF0088FF00EBFF00FFBE00FF7700FF4200FF1F00FF1000FF1500FF2C00FF5500FF8F00
FFD600D7FF007DFF001EFF0000FF4300FFA300FFFF00ABFF005FFF001EFF1400FF3900FF
5000FF5800FF5100FF3D00FF1D00FF000DFF0041FF007BFF00BAFF00FBFF00FFC200FF83
00FF4700FF111DFF0044FF00
00FF1000FF4100FF7700FFB100FFEE00D4FF0098FF005FFF002BFF0100FF2600FF4000FF
4E00FF5000FF4400FF2B00FF0400FF002FFF006EFF00B8FF00FFF500FF9C00FF3F1EFF00
7AFF00D2FF00FFDD00FF9600FF5C00FF3000FF1400FF0A00FF1200FF2C00FF5900FF9600
FFE300C2FF005DFF0000FF1000FF8200FFF60097FF002AFF3A00FF9300FFDE00FFFF00E7
FF00C0FF00ABFF00ABFF00C0FF00E7DE00FF9300FF3A00FF002AFF0097FF00FFF600FF82
00FF105DFF00C2FF00FFE300FF9600FF5900FF2C00FF1200FF0A00FF1400FF3000FF5C00
FF9600FFDD00D2FF007AFF001EFF0000FF3F00FF9C00FFF500B8FF006EFF002FFF0400FF
2B00FF4400FF5000FF4E00FF4000FF2600FF0100FF002BFF005FFF0098FF00D4FF00FFEE
00FFB100FF7700FF4100FF10
00FF7400FFAA00FFE200E4FF00ACFF0076FF0043FF0016FF0F00FF2C00FF3F00FF4700FF
4300FF3300FF1700FF0011FF0045FF0084FF00CCFF00FFE300FF8E00FF3524FF007DFF00
D2FF00FFDF00FF9A00FF6000FF3300FF1500FF0700FF0900FF1D00FF4100FF7600FFBB00
F3FF0095FF002FFF0000FF3D00FFAE00E0FF0072FF000AFF5500FFAA00FFF000FFFF00D8
FF00B2FF009FFF009FFF00B2FF00D8F000FFAA00FF5500FF000AFF0072FF00E0FF00FFAE
00FF3D2FFF0095FF00F3FF00FFBB00FF7600FF4100FF1D00FF0900FF0700FF1500FF3300
FF6000FF9A00FFDF00D2FF007DFF0024FF0000FF3500FF8E00FFE300CCFF0084FF0045FF
0011FF1700FF3300FF4300FF4700FF3F00FF2C00FF0F00FF0016FF0043FF0076FF00ACFF
00E4FF00FFE200FFAA00FF74
00FFDE00ECFF00B8FF0085FF0055FF002AFF0005FF1800FF2E00FF3A00FF3B00FF3200FF
1D00FF0002FF002CFF0062FF00A0FF00E7FF00FFCA00FF7800FF2332FF0087FF00D7FF00
FFDD00FF9A00FF6100FF3500FF1500FF0500FF0400FF1200FF3100FF5F00FF9B00FFE500
C4FF0065FF0000FF0100FF6D00FFDB00B7FF004DFF1500FF7000FFC000FFFF00FCFF00C9
FF00A6FF0094FF0094FF00A6FF00C9FF00FCC000FF7000FF1500FF004DFF00B7FF00FFDB
00FF6D00FF0165FF00C4FF00FFE500FF9B00FF5F00FF3100FF1200FF0400FF0500FF1500
FF3500FF6100FF9A00FFDD00D7FF0087FF0032FF0000FF2300FF7800FFCA00E7FF00A0FF
0062FF002CFF0002FF1D00FF3200FF3B00FF3A00FF2E00FF1800FF0005FF002AFF0055FF
0085FF00B8FF00ECFF00FFDE
00BCFF008DFF0061FF0039FF0016FF0600FF1C00FF2A00FF2E00FF2900FF1A00FF0100FF
0022FF004FFF0085FF00C4FF00FFF600FFAB00FF5C00FF0B46FF0096FF00E3FF00FFD600
FF9600FF6000FF3500FF1600FF0400FF0000FF0C00FF2500FF4E00FF8400FFC700EAFF00
91FF0031FF0000FF3400FF9E00F6FF008DFF0028FF3600FF8B00FFD700FFFF00EAFF00BA
FF0099FF0088FF0088FF0099FF00BAFF00EAD700FF8B00FF3600FF0028FF008DFF00F6FF
00FF9E00FF3431FF0091FF00EAFF00FFC700FF8400FF4E00FF2500FF0C00FF0000FF0400
FF1600FF3500FF6000FF9600FFD600E3FF0096FF0046FF0000FF0B00FF5C00FFAB00FFF6
00C4FF0085FF004FFF0022FF0100FF1A00FF2900FF2E00FF2A00FF1C00FF0600FF0016FF
0039FF0061FF008DFF00BCFF
0068FF0044FF0023FF0008FF0B00FF1900FF1E00FF1B00FF0F00FF0005FF0023FF0049FF
0079FF00B0FF00EEFF00FFCD00FF8500FF3913FF0060FF00ABFF00F3FF00FFCA00FF8F00
FF5C00FF3300FF1500FF0400FF0000FF0800FF1F00FF4300FF7400FFB100FFF900B4FF00
5AFF0000FF0500FF6A00FFD100C8FF0063FF0003FF5500FFA600FFEC00FFFF00D8FF00AB
FF008DFF007DFF007DFF008DFF00ABFF00D8EC00FFA600FF5500FF0003FF0063FF00C8FF
00FFD100FF6A00FF055AFF00B4FF00FFF900FFB100FF7400FF4300FF1F00FF0800FF0000
FF0400FF1500FF3300FF5C00FF8F00FFCA00F3FF00ABFF0060FF0013FF0000FF3900FF85
00FFCD00EEFF00B0FF0079FF0049FF0023FF0005FF0F00FF1B00FF1E00FF1900FF0B00FF
0008FF0023FF0044FF0068FF
002EFF0017FF0005FF0600FF0B00FF0900FF0000FF0012FF002DFF004FFF0079FF00AAFF
00E2FF00FFE000FF9E00FF5800FF1137FF007FFF00C5FF00FFF800FFBB00FF8400FF5500
FF3000FF1500FF0500FF0000FF0800FF1D00FF3D00FF6A00FFA200FFE500CEFF007AFF00
20FF0000FF3F00FFA100FBFF0098FF0039FF2100FF7500FFC000FFFF00FDFF00C7FF009D
FF0081FF0073FF0073FF0081FF009DFF00C7FF00FDC000FF7500FF2100FF0039FF0098FF
00FBFF00FFA100FF3F20FF007AFF00CEFF00FFE500FFA200FF6A00FF3D00FF1D00FF0800
FF0000FF0500FF1500FF3000FF5500FF8400FFBB00FFF800C5FF007FFF0037FF0000FF11
00FF5800FF9E00FFE000E2FF00AAFF0079FF004FFF002DFF0012FF0000FF0900FF0B00FF
0600FF0005FF0017FF002EFF
0016FF000CFF0009FF000CFF0016FF0027FF003FFF005EFF0084FF00B0FF00E3FF00FFE5
00FFA900FF6900FF271CFF0060FF00A2FF00E3FF00FFE000FFA900FF7700FF4E00FF2C00
FF1400FF0700FF0400FF0C00FF1F00FF3D00FF6700FF9B00FFD900DFFF0091FF003CFF00
00FF1D00FF7A00FFD900C6FF0069FF000FFF4500FF9400FFDA00FFFF00E8FF00B6FF0090
FF0076FF0068FF0068FF0076FF0090FF00B6FF00E8DA00FF9400FF4500FF000FFF0069FF
00C6FF00FFD900FF7A00FF1D3CFF0091FF00DFFF00FFD900FF9B00FF6700FF3D00FF1F00
FF0C00FF0400FF0700FF1400FF2C00FF4E00FF7700FFA900FFE000E3FF00A2FF0060FF00
1CFF0000FF2700FF6900FFA900FFE500E3FF00B0FF0084FF005EFF003FFF0027FF0016FF
000CFF0009FF000CFF0016FF
0022FF0026FF0031FF0042FF0059FF0076FF0099FF00C1FF00EFFF00FFDD00FFA600FF6C
00FF2F0EFF004DFF008CFF00C8FF00FFFD00FFC600FF9500FF6A00FF4500FF2900FF1500
FF0A00FF0900FF1200FF2500FF4300FF6A00FF9B00FFD500E8FF009FFF004FFF0000FF05
00FF5D00FFB700ECFF0092FF0039FF1A00FF6A00FFB200FFF300FFFF00D4FF00A6FF0082
FF006BFF005FFF005FFF006BFF0082FF00A6FF00D4F300FFB200FF6A00FF1A00FF0039FF
0092FF00ECFF00FFB700FF5D00FF054FFF009FFF00E8FF00FFD500FF9B00FF6A00FF4300
FF2500FF1200FF0900FF0A00FF1500FF2900FF4500FF6A00FF9500FFC600FFFD00C8FF00
8CFF004DFF000EFF0000FF2F00FF6C00FFA600FFDD00EFFF00C1FF0099FF0076FF0059FF
0042FF0031FF0026FF0022FF
0052FF0063FF007AFF0095FF00B6FF00DCFF00FFF800FFCA00FF9700FF6200FF2B0DFF00
47FF0081FF00B9FF00F0FF00FFDC00FFAC00FF8200FF5D00FF3E00FF2700FF1700FF1000
FF1200FF1D00FF3100FF4E00FF7400FFA200FFD900E8FF00A3FF0058FF0009FF0000FF49
00FF9F00FFF600B2FF005DFF000AFF4300FF8D00FFD000FFFF00F3FF00C0FF0096FF0076
FF0060FF0055FF0055FF0060FF0076FF0096FF00C0FF00F3D000FF8D00FF4300FF000AFF
005DFF00B2FF00FFF600FF9F00FF4909FF0058FF00A3FF00E8FF00FFD900FFA200FF7400
FF4E00FF3100FF1D00FF1200FF1000FF1700FF2700FF3E00FF5D00FF8200FFAC00FFDC00
F0FF00B9FF0081FF0047FF000DFF0000FF2B00FF6200FF9700FFCA00FFF800DCFF00B6FF
0095FF007AFF0063FF0052FF
00A1FF00BDFF00DCFF00FFFF00FFD800FFAD00FF7F00FF4E00FF1B17FF004CFF0081FF00
B4FF00E6FF00FFE900FFBD00FF9400FF7000FF5200FF3A00FF2800FF1D00FF1A00FF1F00
FF2C00FF4100FF5F00FF8400FFB100FFE500DFFF009FFF0058FF000EFF0000FF3F00FF90
00FFE200CAFF0078FF0029FF2300FF6C00FFB000FFED00FFFF00DCFF00ADFF0087FF006A
FF0056FF004CFF004CFF0056FF006AFF0087FF00ADFF00DCED00FFB000FF6C00FF2300FF
0029FF0078FF00CAFF00FFE200FF9000FF3F0EFF0058FF009FFF00DFFF00FFE500FFB100
FF8400FF5F00FF4100FF2C00FF1F00FF1A00FF1D00FF2800FF3A00FF5200FF7000FF9400
FFBD00FFE900E6FF00B4FF0081FF004CFF0017FF0000FF1B00FF4E00FF7F00FFAD00FFD8
00FFFF00DCFF00BDFF00A1FF
00FFF600FFD400FFAF00FF8700FF5D00FF3100FF032BFF005AFF0089FF00B8FF00E5FF00
FFEF00FFC600FFA100FF8000FF6300FF4C00FF3A00FF2E00FF2900FF2A00FF3200FF4200
FF5900FF7600FF9B00FFC700FFF900CEFF0091FF004FFF0009FF0000FF3F00FF8B00FFD8
00D9FF008BFF003FFF0A00FF5100FF9300FFD100FFFF00F6FF00C5FF009BFF0078FF005E
FF004CFF0043FF0043FF004CFF005EFF0078FF009BFF00C5FF00F6D100FF9300FF5100FF
0A00FF003FFF008BFF00D9FF00FFD800FF8B00FF3F09FF004FFF0091FF00CEFF00FFF900
FFC700FF9B00FF7600FF5900FF4200FF3200FF2A00FF2900FF2E00FF3A00FF4C00FF6300
FF8000FFA100FFC600FFEF00E5FF00B8FF0089FF005AFF002BFF0000FF0300FF3100FF5D
00FF8700FFAF00FFD400FFF6
00FF8000FF5C00FF3500FF0D1BFF0045FF006FFF0099FF00C2FF00EAFF00FFEF00FFCB00
FFAA00FF8C00FF7300FF5D00FF4D00FF4100FF3B00FF3B00FF4000FF4C00FF5F00FF7700
FF9600FFBB00FFE500EAFF00B4FF007AFF003CFF0000FF0500FF4900FF9000FFD800DDFF
0095FF004CFF0006FF3D00FF7D00FFBA00FFF200FFFF00DBFF00AFFF0089FF006AFF0053
FF0043FF003BFF003BFF0043FF0053FF006AFF0089FF00AFFF00DBF200FFBA00FF7D00FF
3D00FF0006FF004CFF0095FF00DDFF00FFD800FF9000FF4900FF053CFF007AFF00B4FF00
EAFF00FFE500FFBB00FF9600FF7700FF5F00FF4C00FF4000FF3B00FF3B00FF4100FF4D00
FF5D00FF7300FF8C00FFAA00FFCB00FFEF00EAFF00C2FF0099FF006FFF0045FF001BFF00
00FF0D00FF3500FF5C00FF80
00FF0919FF003EFF0062FF0087FF00ACFF00D0FF00F2FF00FFEC00FFCD00FFB100FF9700
FF8100FF6F00FF6100FF5700FF5100FF5100FF5600FF6000FF6F00FF8400FF9E00FFBE00
FFE300F3FF00C4FF0091FF005AFF0020FF0000FF1D00FF5D00FF9F00FFE200D9FF0095FF
0051FF000FFF3000FF6E00FFA800FFDF00FFFF00EEFF00C1FF009AFF0078FF005DFF0048
FF003AFF0033FF0033FF003AFF0048FF005DFF0078FF009AFF00C1FF00EEDF00FFA800FF
6E00FF3000FF000FFF0051FF0095FF00D9FF00FFE200FF9F00FF5D00FF1D20FF005AFF00
91FF00C4FF00F3FF00FFE300FFBE00FF9E00FF8400FF6F00FF6000FF5600FF5100FF5100
FF5700FF6100FF6F00FF8100FF9700FFB100FFCD00FFEC00F2FF00D0FF00ACFF0087FF00
62FF003EFF0019FF0000FF09
61FF0081FF00A0FF00BFFF00DDFF00FAFF00FFE900FFD000FFB800FFA300FF9200FF8300
FF7700FF7000FF6C00FF6D00FF7200FF7B00FF8900FF9C00FFB300FFCF00FFEF00EBFF00
C2FF0095FF0065FF0031FF0000FF0500FF3F00FF7A00FFB700FFF600CAFF008BFF004CFF
000FFF2C00FF6600FF9D00FFD100FFFF00FDFF00D1FF00A9FF0086FF0069FF0051FF003F
FF0032FF002CFF002CFF0032FF003FFF0051FF0069FF0086FF00A9FF00D1FF00FDD100FF
9D00FF6600FF2C00FF000FFF004CFF008BFF00CAFF00FFF600FFB700FF7A00FF3F00FF05
31FF0065FF0095FF00C2FF00EBFF00FFEF00FFCF00FFB300FF9C00FF8900FF7B00FF7200
FF6D00FF6C00FF7000FF7700FF8300FF9200FFA300FFB800FFD000FFE900FAFF00DDFF00
BFFF00A0FF0081FF0061FF00
B5FF00CFFF00E7FF00FFFF00FFEA00FFD600FFC400FFB400FFA600FF9B00FF9400FF8F00
FF8E00FF9000FF9600FF9F00FFAD00FFBE00FFD400FFED00F5FF00D4FF00B0FF0088FF00
5DFF002FFF0000FF0100FF3400FF6A00FFA100FFD900ECFF00B2FF0078FF003FFF0006FF
3000FF6600FF9A00FFCB00FFF900FFFF00DCFF00B5FF0092FF0074FF005AFF0045FF0035
FF002BFF0025FF0025FF002BFF0035FF0045FF005AFF0074FF0092FF00B5FF00DCF900FF
CB00FF9A00FF6600FF3000FF0006FF003FFF0078FF00B2FF00ECFF00FFD900FFA100FF6A
00FF3400FF012FFF005DFF0088FF00B0FF00D4FF00F5FF00FFED00FFD400FFBE00FFAD00
FF9F00FF9600FF9000FF8E00FF8F00FF9400FF9B00FFA600FFB400FFC400FFD600FFEA00
FFFF00E7FF00CFFF00B5FF00
E9FF00FBFF00FFF300FFE400FFD700FFCC00FFC300FFBC00FFB800FFB600FFB800FFBC00
FFC300FFCE00FFDB00FFEC00FFFF00E7FF00CDFF00AFFF008EFF006AFF0044FF001BFF00
00FF1000FF3D00FF6D00FF9E00FFD100FBFF00C6FF0092FF005DFF0029FF0A00FF3D00FF
6E00FF9D00FFCB00FFF600FFFF00E1FF00BCFF009BFF007DFF0063FF004CFF003AFF002D
FF0024FF001FFF001FFF0024FF002DFF003AFF004CFF0063FF007DFF009BFF00BCFF00E1
F600FFCB00FF9D00FF6E00FF3D00FF0A00FF0029FF005DFF0092FF00C6FF00FBFF00FFD1
00FF9E00FF6D00FF3D00FF101BFF0044FF006AFF008EFF00AFFF00CDFF00E7FF00FFFF00
FFEC00FFDB00FFCE00FFC300FFBC00FFB800FFB600FFB800FFBC00FFC300FFCC00FFD700
FFE400FFF300FBFF00E9FF00
F7FF00FFFE00FFF600FFEF00FFEA00FFE700FFE700FFE800FFEC00FFF300FFFB00F8FF00
EAFF00D9FF00C6FF00B0FF0097FF007CFF005EFF003EFF001BFF0000FF0800FF2F00FF58
00FF8200FFAE00FFDB00F6FF00C8FF0098FF0069FF0039FF000AFF2300FF5100FF7D00FF
A800FFD100FFF900FFFF00E1FF00BFFF009FFF0083FF0069FF0053FF0040FF0031FF0025
FF001DFF001AFF001AFF001DFF0025FF0031FF0040FF0053FF0069FF0083FF009FFF00BF
FF00E1F900FFD100FFA800FF7D00FF5100FF2300FF000AFF0039FF0069FF0098FF00C8FF
00F6FF00FFDB00FFAE00FF8200FF5800FF2F00FF081BFF003EFF005EFF007CFF0097FF00
B0FF00C6FF00D9FF00EAFF00F8FF00FFFB00FFF300FFEC00FFE800FFE700FFE700FFEA00
FFEF00FFF600FFFE00F7FF00
DCFF00DFFF00E0FF00DFFF00DDFF00D8FF00D2FF00CAFF00BFFF00B3FF00A4FF0093FF00
81FF006CFF0054FF003BFF0020FF0003FF0000FF1B00FF3C00FF5E00FF8200FFA800FFCE
00FFF600E0FF00B7FF008DFF0063FF0039FF000FFF1A00FF4300FF6C00FF9300FFBA00FF
DF00FFFF00FDFF00DCFF00BCFF009FFF0085FF006CFF0057FF0044FF0034FF0028FF001E
FF0018FF0014FF0014FF0018FF001EFF0028FF0034FF0044FF0057FF006CFF0085FF009F
FF00BCFF00DCFF00FDDF00FFBA00FF9300FF6C00FF4300FF1A00FF000FFF0039FF0063FF
008DFF00B7FF00E0FF00FFF600FFCE00FFA800FF8200FF5E00FF3C00FF1B03FF0020FF00
3BFF0054FF006CFF0081FF0093FF00A4FF00B3FF00BFFF00CAFF00D2FF00D8FF00DDFF00
DFFF00E0FF00DFFF00DCFF00
9CFF0098FF0093FF008CFF0084FF007AFF006EFF0061FF0052FF0042FF002FFF001BFF00
06FF0000FF1100FF2900FF4400FF6000FF7D00FF9B00FFBB00FFDC00FFFE00DEFF00BBFF
0097FF0072FF004DFF0028FF0003FF2100FF4500FF6A00FF8D00FFB000FFD100FFF200FF
FF00EEFF00D1FF00B5FF009BFF0083FF006CFF0058FF0046FF0037FF002AFF0020FF0018
FF0012FF0010FF0010FF0012FF0018FF0020FF002AFF0037FF0046FF0058FF006CFF0083
FF009BFF00B5FF00D1FF00EEF200FFD100FFB000FF8D00FF6A00FF4500FF2100FF0003FF
0028FF004DFF0072FF0097FF00BBFF00DEFF00FFFE00FFDC00FFBB00FF9B00FF7D00FF60
00FF4400FF2900FF1106FF001BFF002FFF0042FF0052FF0061FF006EFF007AFF0084FF00
8CFF0093FF0098FF009CFF00
3CFF0033FF002AFF001EFF0012FF0004FF0000FF0900FF1A00FF2B00FF3E00FF5200FF68
00FF7E00FF9600FFAF00FFC900FFE400FFFF00E3FF00C5FF00A7FF0088FF0069FF004AFF
002AFF000AFF1500FF3600FF5500FF7500FF9400FFB200FFD000FFED00FFFF00F6FF00DB
FF00C1FF00A9FF0092FF007DFF0069FF0057FF0046FF0038FF002CFF0021FF0019FF0012
FF000EFF000CFF000CFF000EFF0012FF0019FF0021FF002CFF0038FF0046FF0057FF0069
FF007DFF0092FF00A9FF00C1FF00DBFF00F6ED00FFD000FFB200FF9400FF7500FF5500FF
3600FF1500FF000AFF002AFF004AFF0069FF0088FF00A7FF00C5FF00E3FF00FFFF00FFE4
00FFC900FFAF00FF9600FF7E00FF6800FF5200FF3E00FF2B00FF1A00FF0904FF0012FF00
1EFF002AFF0033FF003CFF00
00FF3900FF4400FF5100FF5F00FF6D00FF7D00FF8E00FF9F00FFB200FFC500FFDA00FFEF
00FAFF00E3FF00CCFF00B4FF009BFF0081FF0067FF004DFF0032FF0017FF0300FF1E00FF
3A00FF5500FF7000FF8B00FFA600FFC000FFDA00FFF300FFFF00F3FF00DCFF00C5FF00AF
FF009AFF0086FF0074FF0063FF0053FF0044FF0037FF002CFF0022FF0019FF0012FF000D
FF000AFF0008FF0008FF000AFF000DFF0012FF0019FF0022FF002CFF0037FF0044FF0053
FF0063FF0074FF0086FF009AFF00AFFF00C5FF00DCFF00F3F300FFDA00FFC000FFA600FF
8B00FF7000FF5500FF3A00FF1E00FF0300FF0017FF0032FF004DFF0067FF0081FF009BFF
00B4FF00CCFF00E3FF00FAFF00FFEF00FFDA00FFC500FFB200FF9F00FF8E00FF7D00FF6D
00FF5F00FF5100FF4400FF39
00FFB800FFC500FFD300FFE200FFF100FDFF00ECFF00DBFF00C9FF00B6FF00A2FF008FFF
007AFF0065FF0050FF003AFF0024FF000DFF0800FF1F00FF3600FF4E00FF6500FF7C00FF
9300FFAA00FFC000FFD700FFEC00FFFF00FDFF00E8FF00D4FF00C0FF00ADFF009BFF0089
FF0078FF0069FF005AFF004CFF0040FF0034FF002AFF0021FF0019FF0013FF000DFF0009
FF0007FF0005FF0005FF0007FF0009FF000DFF0013FF0019FF0021FF002AFF0034FF0040
FF004CFF005AFF0069FF0078FF0089FF009BFF00ADFF00C0FF00D4FF00E8FF00FDEC00FF
D700FFC000FFAA00FF9300FF7C00FF6500FF4E00FF3600FF1F00FF0800FF000DFF0024FF
003AFF0050FF0065FF007AFF008FFF00A2FF00B6FF00C9FF00DBFF00ECFF00FDFF00FFF1
00FFE200FFD300FFC500FFB8
00CDFF00BFFF00B0FF00A1FF0092FF0082FF0072FF0061FF0050FF003FFF002DFF001BFF
0008FF0900FF1C00FF2F00FF4300FF5600FF6A00FF7D00FF9100FFA400FFB700FFCB00FF
DE00FFF000FFFF00FCFF00EAFF00D8FF00C7FF00B6FF00A6FF0096FF0087FF0078FF006A
FF005DFF0051FF0045FF003AFF0031FF0028FF0020FF0019FF0012FF000DFF0009FF0006
FF0004FF0003FF0003FF0004FF0006FF0009FF000DFF0012FF0019FF0020FF0028FF0031
FF003AFF0045FF0051FF005DFF006AFF0078FF0087FF0096FF00A6FF00B6FF00C7FF00D8
FF00EAFF00FCF000FFDE00FFCB00FFB700FFA400FF9100FF7D00FF6A00FF5600FF4300FF
2F00FF1C00FF0900FF0008FF001BFF002DFF003FFF0050FF0061FF0072FF0082FF0092FF
00A1FF00B0FF00BFFF00CDFF
0063FF0055FF0047FF0039FF002AFF001BFF000CFF0300FF1200FF2200FF3200FF4200FF
5300FF6300FF7400FF8500FF9500FFA600FFB600FFC700FFD700FFE800FFF800FFFF00F7
FF00E7FF00D8FF00C9FF00BAFF00ABFF009DFF0090FF0082FF0076FF006AFF005EFF0053
FF0048FF003FFF0035FF002DFF0025FF001EFF0018FF0012FF000DFF0009FF0006FF0004
FF0002FF0001FF0001FF0002FF0004FF0006FF0009FF000DFF0012FF0018FF001EFF0025
FF002DFF0035FF003FFF0048FF0053FF005EFF006AFF0076FF0082FF0090FF009DFF00AB
FF00BAFF00C9FF00D8FF00E7FF00F7F800FFE800FFD700FFC700FFB600FFA600FF9500FF
8500FF7400FF6300FF5300FF4200FF3200FF2200FF1200FF0300FF000CFF001BFF002AFF
0039FF0047FF0055FF0063FF
0015FF0008FF0400FF1200FF2000FF2E00FF3C00FF4B00FF5900FF6800FF7600FF8500FF
9400FFA200FFB100FFC000FFCF00FFDD00FFEC00FFFA00FFFF00F7FF00E9FF00DBFF00CD
FF00C0FF00B2FF00A6FF0099FF008DFF0081FF0076FF006BFF0060FF0056FF004CFF0043
FF003AFF0032FF002BFF0024FF001DFF0018FF0012FF000EFF000AFF0007FF0004FF0002
FF0001FF0000FF0000FF0001FF0002FF0004FF0007FF000AFF000EFF0012FF0018FF001D
FF0024FF002BFF0032FF003AFF0043FF004CFF0056FF0060FF006BFF0076FF0081FF008D
FF0099FF00A6FF00B2FF00C0FF00CDFF00DBFF00E9FF00F7FA00FFEC00FFDD00FFCF00FF
C000FFB100FFA200FF9400FF8500FF7600FF6800FF5900FF4B00FF3C00FF2E00FF2000FF
1200FF0400FF0008FF0015FF
1300FF2000FF2D00FF3A00FF4700FF5500FF6200FF7000FF7E00FF8C00FF9A00FFA700FF
B500FFC300FFD100FFDF00FFEC00FFFA00FFFF00F8FF00EBFF00DEFF00D1FF00C4FF00B8
FF00ABFF009FFF0094FF0088FF007DFF0073FF0068FF005FFF0055FF004CFF0043FF003B
FF0033FF002CFF0025FF001FFF001AFF0014FF0010FF000CFF0008FF0005FF0003FF0001
FF0000FF0000FF0000FF0000FF0001FF0003FF0005FF0008FF000CFF0010FF0014FF001A
FF001FFF0025FF002CFF0033FF003BFF0043FF004CFF0055FF005FFF0068FF0073FF007D
FF0088FF0094FF009FFF00ABFF00B8FF00C4FF00D1FF00DEFF00EBFF00F8FA00FFEC00FF
DF00FFD100FFC300FFB500FFA700FF9A00FF8C00FF7E00FF7000FF6200FF5500FF4700FF
3A00FF2D00FF2000FF1300FF
1300FF2000FF2D00FF3A00FF4700FF5500FF6200FF7000FF7E00FF8C00FF9A00FFA700FF
B500FFC300FFD100FFDF00FFEC00FFFA00FFFF00F8FF00EBFF00DEFF00D1FF00C4FF00B8
FF00ABFF009FFF0094FF0088FF007DFF0073FF0068FF005FFF0055FF004CFF0043FF003B
FF0033FF002CFF0025FF001FFF001AFF0014FF0010FF000CFF0008FF0005FF0003FF0001
FF0000FF0000FF0000FF0000FF0001FF0003FF0005FF0008FF000CFF0010FF0014FF001A
FF001FFF0025FF002CFF0033FF003BFF0043FF004CFF0055FF005FFF0068FF0073FF007D
FF0088FF0094FF009FFF00ABFF00B8FF00C4FF00D1FF00DEFF00EBFF00F8FA00FFEC00FF
DF00FFD100FFC300FFB500FFA700FF9A00FF8C00FF7E00FF7000FF6200FF5500FF4700FF
3A00FF2D00FF2000FF1300FF
0015FF0008FF0400FF1200FF2000FF2E00FF3C00FF4B00FF5900FF6800FF7600FF8500FF
9400FFA200FFB100FFC000FFCF00FFDD00FFEC00FFFA00FFFF00F7FF00E9FF00DBFF00CD
FF00C0FF00B2FF00A6FF0099FF008DFF0081FF0076FF006BFF0060FF0056FF004CFF0043
FF003AFF0032FF002BFF0024FF001DFF0018FF0012FF000EFF000AFF0007FF0004FF0002
FF0001FF0000FF0000FF0001FF0002FF0004FF0007FF000AFF000EFF0012FF0018FF001D
FF0024FF002BFF0032FF003AFF0043FF004CFF0056FF0060FF006BFF0076FF0081FF008D
FF0099FF00A6FF00B2FF00C0FF00CDFF00DBFF00E9FF00F7FA00FFEC00FFDD00FFCF00FF
C000FFB100FFA200FF9400FF8500FF7600FF6800FF5900FF4B00FF3C00FF2E00FF2000FF
1200FF0400FF0008FF0015FF
0063FF0055FF0047FF0039FF002AFF001BFF000CFF0300FF1200FF2200FF3200FF4200FF
5300FF6300FF7400FF8500FF9500FFA600FFB600FFC700FFD700FFE800FFF800FFFF00F7
FF00E7FF00D8FF00C9FF00BAFF00ABFF009DFF0090FF0082FF0076FF006AFF005EFF0053
FF0048FF003FFF0035FF002DFF0025FF001EFF0018FF0012FF000DFF0009FF0006FF0004
FF0002FF0001FF0001FF0002FF0004FF0006FF0009FF000DFF0012FF0018FF001EFF0025
FF002DFF0035FF003FFF0048FF0053FF005EFF006AFF0076FF0082FF0090FF009DFF00AB
FF00BAFF00C9FF00D8FF00E7FF00F7F800FFE800FFD700FFC700FFB600FFA600FF9500FF
8500FF7400FF6300FF5300FF4200FF3200FF2200FF1200FF0300FF000CFF001BFF002AFF
0039FF0047FF0055FF0063FF
00CDFF00BFFF00B0FF00A1FF0092FF0082FF0072FF0061FF0050FF003FFF002DFF001BFF
0008FF0900FF1C00FF2F00FF4300FF5600FF6A00FF7D00FF9100FFA400FFB700FFCB00FF
DE00FFF000FFFF00FCFF00EAFF00D8FF00C7FF00B6FF00A6FF0096FF0087FF0078FF006A
FF005DFF0051FF0045FF003AFF0031FF0028FF0020FF0019FF0012FF000DFF0009FF0006
FF0004FF0003FF0003FF0004FF0006FF0009FF000DFF0012FF0019FF0020FF0028FF0031
FF003AFF0045FF0051FF005DFF006AFF0078FF0087FF0096FF00A6FF00B6FF00C7FF00D8
FF00EAFF00FCF000FFDE00FFCB00FFB700FFA400FF9100FF7D00FF6A00FF5600FF4300FF
2F00FF1C00FF0900FF0008FF001BFF002DFF003FFF0050FF0061FF0072FF0082FF0092FF
00A1FF00B0FF00BFFF00CDFF
00FFB800FFC500FFD300FFE200FFF100FDFF00ECFF00DBFF00C9FF00B6FF00A2FF008FFF
007AFF0065FF0050FF003AFF0024FF000DFF0800FF1F00FF3600FF4E00FF6500FF7C00FF
9300FFAA00FFC000FFD700FFEC00FFFF00FDFF00E8FF00D4FF00C0FF00ADFF009BFF0089
FF0078FF0069FF005AFF004CFF0040FF0034FF002AFF0021FF0019FF0013FF000DFF0009
FF0007FF0005FF0005FF0007FF0009FF000DFF0013FF0019FF0021FF002AFF0034FF0040
FF004CFF005AFF0069FF0078FF0089FF009BFF00ADFF00C0FF00D4FF00E8FF00FDEC00FF
D700FFC000FFAA00FF9300FF7C00FF6500FF4E00FF3600FF1F00FF0800FF000DFF0024FF
003AFF0050FF0065FF007AFF008FFF00A2FF00B6FF00C9FF00DBFF00ECFF00FDFF00FFF1
00FFE200FFD300FFC500FFB8
00FF3900FF4400FF5100FF5F00FF6D00FF7D00FF8E00FF9F00FFB200FFC500FFDA00FFEF
00FAFF00E3FF00CCFF00B4FF009BFF0081FF0067FF004DFF0032FF0017FF0300FF1E00FF
3A00FF5500FF7000FF8B00FFA600FFC000FFDA00FFF300FFFF00F3FF00DCFF00C5FF00AF
FF009AFF0086FF0074FF0063FF0053FF0044FF0037FF002CFF0022FF0019FF0012FF000D
FF000AFF0008FF0008FF000AFF000DFF0012FF0019FF0022FF002CFF0037FF0044FF0053
FF0063FF0074FF0086FF009AFF00AFFF00C5FF00DCFF00F3F300FFDA00FFC000FFA600FF
8B00FF7000FF5500FF3A00FF1E00FF0300FF0017FF0032FF004DFF0067FF0081FF009BFF
00B4FF00CCFF00E3FF00FAFF00FFEF00FFDA00FFC500FFB200FF9F00FF8E00FF7D00FF6D
00FF5F00FF5100FF4400FF39
3CFF0033FF002AFF001EFF0012FF0004FF0000FF0900FF1A00FF2B00FF3E00FF5200FF68
00FF7E00FF9600FFAF00FFC900FFE400FFFF00E3FF00C5FF00A7FF0088FF0069FF004AFF
002AFF000AFF1500FF3600FF5500FF7500FF9400FFB200FFD000FFED00FFFF00F6FF00DB
FF00C1FF00A9FF0092FF007DFF0069FF0057FF0046FF0038FF002CFF0021FF0019FF0012
FF000EFF000CFF000CFF000EFF0012FF0019FF0021FF002CFF0038FF0046FF0057FF0069
FF007DFF0092FF00A9FF00C1FF00DBFF00F6ED00FFD000FFB200FF9400FF7500FF5500FF
3600FF1500FF000AFF002AFF004AFF0069FF0088FF00A7FF00C5FF00E3FF00FFFF00FFE4
00FFC900FFAF00FF9600FF7E00FF6800FF5200FF3E00FF2B00FF1A00FF0904FF0012FF00
1EFF002AFF0033FF003CFF00
9CFF0098FF0093FF008CFF0084FF007AFF006EFF0061FF0052FF0042FF002FFF001BFF00
06FF0000FF1100FF2900FF4400FF6000FF7D00FF9B00FFBB00FFDC00FFFE00DEFF00BBFF
0097FF0072FF004DFF0028FF0003FF2100FF4500FF6A00FF8D00FFB000FFD100FFF200FF
FF00EEFF00D1FF00B5FF009BFF0083FF006CFF0058FF0046FF0037FF002AFF0020FF0018
FF0012FF0010FF0010FF0012FF0018FF0020FF002AFF0037FF0046FF0058FF006CFF0083
FF009BFF00B5FF00D1FF00EEF200FFD100FFB000FF8D00FF6A00FF4500FF2100FF0003FF
0028FF004DFF0072FF0097FF00BBFF00DEFF00FFFE00FFDC00FFBB00FF9B00FF7D00FF60
00FF4400FF2900FF1106FF001BFF002FFF0042FF0052FF0061FF006EFF007AFF0084FF00
8CFF0093FF0098FF009CFF00
DCFF00DFFF00E0FF00DFFF00DDFF00D8FF00D2FF00CAFF00BFFF00B3FF00A4FF0093FF00
81FF006CFF0054FF003BFF0020FF0003FF0000FF1B00FF3C00FF5E00FF8200FFA800FFCE
00FFF600E0FF00B7FF008DFF0063FF0039FF000FFF1A00FF4300FF6C00FF9300FFBA00FF
DF00FFFF00FDFF00DCFF00BCFF009FFF0085FF006CFF0057FF0044FF0034FF0028FF001E
FF0018FF0014FF0014FF0018FF001EFF0028FF0034FF0044FF0057FF006CFF0085FF009F
FF00BCFF00DCFF00FDDF00FFBA00FF9300FF6C00FF4300FF1A00FF000FFF0039FF0063FF
008DFF00B7FF00E0FF00FFF600FFCE00FFA800FF8200FF5E00FF3C00FF1B03FF0020FF00
3BFF0054FF006CFF0081FF0093FF00A4FF00B3FF00BFFF00CAFF00D2FF00D8FF00DDFF00
DFFF00E0FF00DFFF00DCFF00
F7FF00FFFE00FFF600FFEF00FFEA00FFE700FFE700FFE800FFEC00FFF300FFFB00F8FF00
EAFF00D9FF00C6FF00B0FF0097FF007CFF005EFF003EFF001BFF0000FF0800FF2F00FF58
00FF8200FFAE00FFDB00F6FF00C8FF0098FF0069FF0039FF000AFF2300FF5100FF7D00FF
A800FFD100FFF900FFFF00E1FF00BFFF009FFF0083FF0069FF0053FF0040FF0031FF0025
FF001DFF001AFF001AFF001DFF0025FF0031FF0040FF0053FF0069FF0083FF009FFF00BF
FF00E1F900FFD100FFA800FF7D00FF5100FF2300FF000AFF0039FF0069FF0098FF00C8FF
00F6FF00FFDB00FFAE00FF8200FF5800FF2F00FF081BFF003EFF005EFF007CFF0097FF00
B0FF00C6FF00D9FF00EAFF00F8FF00FFFB00FFF300FFEC00FFE800FFE700FFE700FFEA00
FFEF00FFF600FFFE00F7FF00
E9FF00FBFF00FFF300FFE400FFD700FFCC00FFC300FFBC00FFB800FFB600FFB800FFBC00
FFC300FFCE00FFDB00FFEC00FFFF00E7FF00CDFF00AFFF008EFF006AFF0044FF001BFF00
00FF1000FF3D00FF6D00FF9E00FFD100FBFF00C6FF0092FF005DFF0029FF0A00FF3D00FF
6E00FF9D00FFCB00FFF600FFFF00E1FF00BCFF009BFF007DFF0063FF004CFF003AFF002D
FF0024FF001FFF001FFF0024FF002DFF003AFF004CFF0063FF007DFF009BFF00BCFF00E1
F600FFCB00FF9D00FF6E00FF3D00FF0A00FF0029FF005DFF0092FF00C6FF00FBFF00FFD1
00FF9E00FF6D00FF3D00FF101BFF0044FF006AFF008EFF00AFFF00CDFF00E7FF00FFFF00
FFEC00FFDB00FFCE00FFC300FFBC00FFB800FFB600FFB800FFBC00FFC300FFCC00FFD700
FFE400FFF300FBFF00E9FF00
B5FF00CFFF00E7FF00FFFF00FFEA00FFD600FFC400FFB400FFA600FF9B00FF9400FF8F00
FF8E00FF9000FF9600FF9F00FFAD00FFBE00FFD400FFED00F5FF00D4FF00B0FF0088FF00
5DFF002FFF0000FF0100FF3400FF6A00FFA100FFD900ECFF00B2FF0078FF003FFF0006FF
3000FF6600FF9A00FFCB00FFF900FFFF00DCFF00B5FF0092FF0074FF005AFF0045FF0035
FF002BFF0025FF0025FF002BFF0035FF0045FF005AFF0074FF0092FF00B5FF00DCF900FF
CB00FF9A00FF6600FF3000FF0006FF003FFF0078FF00B2FF00ECFF00FFD900FFA100FF6A
00FF3400FF012FFF005DFF0088FF00B0FF00D4FF00F5FF00FFED00FFD400FFBE00FFAD00
FF9F00FF9600FF9000FF8E00FF8F00FF9400FF9B00FFA600FFB400FFC400FFD600FFEA00
FFFF00E7FF00CFFF00B5FF00
61FF0081FF00A0FF00BFFF00DDFF00FAFF00FFE900FFD000FFB800FFA300FF9200FF8300
FF7700FF7000FF6C00FF6D00FF7200FF7B00FF8900FF9C00FFB300FFCF00FFEF00EBFF00
C2FF0095FF0065FF0031FF0000FF0500FF3F00FF7A00FFB700FFF600CAFF008BFF004CFF
000FFF2C00FF6600FF9D00FFD100FFFF00FDFF00D1FF00A9FF0086FF0069FF0051FF003F
FF0032FF002CFF002CFF0032FF003FFF0051FF0069FF0086FF00A9FF00D1FF00FDD100FF
9D00FF6600FF2C00FF000FFF004CFF008BFF00CAFF00FFF600FFB700FF7A00FF3F00FF05
31FF0065FF0095FF00C2FF00EBFF00FFEF00FFCF00FFB300FF9C00FF8900FF7B00FF7200
FF6D00FF6C00FF7000FF7700FF8300FF9200FFA300FFB800FFD000FFE900FAFF00DDFF00
BFFF00A0FF0081FF0061FF00
00FF0919FF003EFF0062FF0087FF00ACFF00D0FF00F2FF00FFEC00FFCD00FFB100FF9700
FF8100FF6F00FF6100FF5700FF5100FF5100FF5600FF6000FF6F00FF8400FF9E00FFBE00
FFE300F3FF00C4FF0091FF005AFF0020FF0000FF1D00FF5D00FF9F00FFE200D9FF0095FF
0051FF000FFF3000FF6E00FFA800FFDF00FFFF00EEFF00C1FF009AFF0078FF005DFF0048
FF003AFF0033FF0033FF003AFF0048FF005DFF0078FF009AFF00C1FF00EEDF00FFA800FF
6E00FF3000FF000FFF0051FF0095FF00D9FF00FFE200FF9F00FF5D00FF1D20FF005AFF00
91FF00C4FF00F3FF00FFE300FFBE00FF9E00FF8400FF6F00FF6000FF5600FF5100FF5100
FF5700FF6100FF6F00FF8100FF9700FFB100FFCD00FFEC00F2FF00D0FF00ACFF0087FF00
62FF003EFF0019FF0000FF09
00FF8000FF5C00FF3500FF0D1BFF0045FF006FFF0099FF00C2FF00EAFF00FFEF00FFCB00
FFAA00FF8C00FF7300FF5D00FF4D00FF4100FF3B00FF3B00FF4000FF4C00FF5F00FF7700
FF9600FFBB00FFE500EAFF00B4FF007AFF003CFF0000FF0500FF4900FF9000FFD800DDFF
0095FF004CFF0006FF3D00FF7D00FFBA00FFF200FFFF00DBFF00AFFF0089FF006AFF0053
FF0043FF003BFF003BFF0043FF0053FF006AFF0089FF00AFFF00DBF200FFBA00FF7D00FF
3D00FF0006FF004CFF0095FF00DDFF00FFD800FF9000FF4900FF053CFF007AFF00B4FF00
EAFF00FFE500FFBB00FF9600FF7700FF5F00FF4C00FF4000FF3B00FF3B00FF4100FF4D00
FF5D00FF7300FF8C00FFAA00FFCB00FFEF00EAFF00C2FF0099FF006FFF0045FF001BFF00
00FF0D00FF3500FF5C00FF80
00FFF600FFD400FFAF00FF8700FF5D00FF3100FF032BFF005AFF0089FF00B8FF00E5FF00
FFEF00FFC600FFA100FF8000FF6300FF4C00FF3A00FF2E00FF2900FF2A00FF3200FF4200
FF5900FF7600FF9B00FFC700FFF900CEFF0091FF004FFF0009FF0000FF3F00FF8B00FFD8
00D9FF008BFF003FFF0A00FF5100FF9300FFD100FFFF00F6FF00C5FF009BFF0078FF005E
FF004CFF0043FF0043FF004CFF005EFF0078FF009BFF00C5FF00F6D100FF9300FF5100FF
0A00FF003FFF008BFF00D9FF00FFD800FF8B00FF3F09FF004FFF0091FF00CEFF00FFF900
FFC700FF9B00FF7600FF5900FF4200FF3200FF2A00FF2900FF2E00FF3A00FF4C00FF6300
FF8000FFA100FFC600FFEF00E5FF00B8FF0089FF005AFF002BFF0000FF0300FF3100FF5D
00FF8700FFAF00FFD400FFF6
00A1FF00BDFF00DCFF00FFFF00FFD800FFAD00FF7F00FF4E00FF1B17FF004CFF0081FF00
B4FF00E6FF00FFE900FFBD00FF9400FF7000FF5200FF3A00FF2800FF1D00FF1A00FF1F00
FF2C00FF4100FF5F00FF8400FFB100FFE500DFFF009FFF0058FF000EFF0000FF3F00FF90
00FFE200CAFF0078FF0029FF2300FF6C00FFB000FFED00FFFF00DCFF00ADFF0087FF006A
FF0056FF004CFF004CFF0056FF006AFF0087FF00ADFF00DCED00FFB000FF6C00FF2300FF
0029FF0078FF00CAFF00FFE200FF9000FF3F0EFF0058FF009FFF00DFFF00FFE500FFB100
FF8400FF5F00FF4100FF2C00FF1F00FF1A00FF1D00FF2800FF3A00FF5200FF7000FF9400
FFBD00FFE900E6FF00B4FF0081FF004CFF0017FF0000FF1B00FF4E00FF7F00FFAD00FFD8
00FFFF00DCFF00BDFF00A1FF
0052FF0063FF007AFF0095FF00B6FF00DCFF00FFF800FFCA00FF9700FF6200FF2B0DFF00
47FF0081FF00B9FF00F0FF00FFDC00FFAC00FF8200FF5D00FF3E00FF2700FF1700FF1000
FF1200FF1D00FF3100FF4E00FF7400FFA200FFD900E8FF00A3FF0058FF0009FF0000FF49
00FF9F00FFF600B2FF005DFF000AFF4300FF8D00FFD000FFFF00F3FF00C0FF0096FF0076
FF0060FF0055FF0055FF0060FF0076FF0096FF00C0FF00F3D000FF8D00FF4300FF000AFF
005DFF00B2FF00FFF600FF9F00FF4909FF0058FF00A3FF00E8FF00FFD900FFA200FF7400
FF4E00FF3100FF1D00FF1200FF1000FF1700FF2700FF3E00FF5D00FF8200FFAC00FFDC00
F0FF00B9FF0081FF0047FF000DFF0000FF2B00FF6200FF9700FFCA00FFF800DCFF00B6FF
0095FF007AFF0063FF0052FF
0022FF0026FF0031FF0042FF0059FF0076FF0099FF00C1FF00EFFF00FFDD00FFA600FF6C
00FF2F0EFF004DFF008CFF00C8FF00FFFD00FFC600FF9500FF6A00FF4500FF2900FF1500
FF0A00FF0900FF1200FF2500FF4300FF6A00FF9B00FFD500E8FF009FFF004FFF0000FF05
00FF5D00FFB700ECFF0092FF0039FF1A00FF6A00FFB200FFF300FFFF00D4FF00A6FF0082
FF006BFF005FFF005FFF006BFF0082FF00A6FF00D4F300FFB200FF6A00FF1A00FF0039FF
0092FF00ECFF00FFB700FF5D00FF054FFF009FFF00E8FF00FFD500FF9B00FF6A00FF4300
FF2500FF1200FF0900FF0A00FF1500FF2900FF4500FF6A00FF9500FFC600FFFD00C8FF00
8CFF004DFF000EFF0000FF2F00FF6C00FFA600FFDD00EFFF00C1FF0099FF0076FF0059FF
0042FF0031FF0026FF0022FF
0016FF000CFF0009FF000CFF0016FF0027FF003FFF005EFF0084FF00B0FF00E3FF00FFE5
00FFA900FF6900FF271CFF0060FF00A2FF00E3FF00FFE000FFA900FF7700FF4E00FF2C00
FF1400FF0700FF0400FF0C00FF1F00FF3D00FF6700FF9B00FFD900DFFF0091FF003CFF00
00FF1D00FF7A00FFD900C6FF0069FF000FFF4500FF9400FFDA00FFFF00E8FF00B6FF0090
FF0076FF0068FF0068FF0076FF0090FF00B6FF00E8DA00FF9400FF4500FF000FFF0069FF
00C6FF00FFD900FF7A00FF1D3CFF0091FF00DFFF00FFD900FF9B00FF6700FF3D00FF1F00
FF0C00FF0400FF0700FF1400FF2C00FF4E00FF7700FFA900FFE000E3FF00A2FF0060FF00
1CFF0000FF2700FF6900FFA900FFE500E3FF00B0FF0084FF005EFF003FFF0027FF0016FF
000CFF0009FF000CFF0016FF
002EFF0017FF0005FF0600FF0B00FF0900FF0000FF0012FF002DFF004FFF0079FF00AAFF
00E2FF00FFE000FF9E00FF5800FF1137FF007FFF00C5FF00FFF800FFBB00FF8400FF5500
FF3000FF1500FF0500FF0000FF0800FF1D00FF3D00FF6A00FFA200FFE500CEFF007AFF00
20FF0000FF3F00FFA100FBFF0098FF0039FF2100FF7500FFC000FFFF00FDFF00C7FF009D
FF0081FF0073FF0073FF0081FF009DFF00C7FF00FDC000FF7500FF2100FF0039FF0098FF
00FBFF00FFA100FF3F20FF007AFF00CEFF00FFE500FFA200FF6A00FF3D00FF1D00FF0800
FF0000FF0500FF1500FF3000FF5500FF8400FFBB00FFF800C5FF007FFF0037FF0000FF11
00FF5800FF9E00FFE000E2FF00AAFF0079FF004FFF002DFF0012FF0000FF0900FF0B00FF
0600FF0005FF0017FF002EFF
0068FF0044FF0023FF0008FF0B00FF1900FF1E00FF1B00FF0F00FF0005FF0023FF0049FF
0079FF00B0FF00EEFF00FFCD00FF8500FF3913FF0060FF00ABFF00F3FF00FFCA00FF8F00
FF5C00FF3300FF1500FF0400FF0000FF0800FF1F00FF4300FF7400FFB100FFF900B4FF00
5AFF0000FF0500FF6A00FFD100C8FF0063FF0003FF5500FFA600FFEC00FFFF00D8FF00AB
FF008DFF007DFF007DFF008DFF00ABFF00D8EC00FFA600FF5500FF0003FF0063FF00C8FF
00FFD100FF6A00FF055AFF00B4FF00FFF900FFB100FF7400FF4300FF1F00FF0800FF0000
FF0400FF1500FF3300FF5C00FF8F00FFCA00F3FF00ABFF0060FF0013FF0000FF3900FF85
00FFCD00EEFF00B0FF0079FF0049FF0023FF0005FF0F00FF1B00FF1E00FF1900FF0B00FF
0008FF0023FF0044FF0068FF
00BCFF008DFF0061FF0039FF0016FF0600FF1C00FF2A00FF2E00FF2900FF1A00FF0100FF
0022FF004FFF0085FF00C4FF00FFF600FFAB00FF5C00FF0B46FF0096FF00E3FF00FFD600
FF9600FF6000FF3500FF1600FF0400FF0000FF0C00FF2500FF4E00FF8400FFC700EAFF00
91FF0031FF0000FF3400FF9E00F6FF008DFF0028FF3600FF8B00FFD700FFFF00EAFF00BA
FF0099FF0088FF0088FF0099FF00BAFF00EAD700FF8B00FF3600FF0028FF008DFF00F6FF
00FF9E00FF3431FF0091FF00EAFF00FFC700FF8400FF4E00FF2500FF0C00FF0000FF0400
FF1600FF3500FF6000FF9600FFD600E3FF0096FF0046FF0000FF0B00FF5C00FFAB00FFF6
00C4FF0085FF004FFF0022FF0100FF1A00FF2900FF2E00FF2A00FF1C00FF0600FF0016FF
0039FF0061FF008DFF00BCFF
00FFDE00ECFF00B8FF0085FF0055FF002AFF0005FF1800FF2E00FF3A00FF3B00FF3200FF
1D00FF0002FF002CFF0062FF00A0FF00E7FF00FFCA00FF7800FF2332FF0087FF00D7FF00
FFDD00FF9A00FF6100FF3500FF1500FF0500FF0400FF1200FF3100FF5F00FF9B00FFE500
C4FF0065FF0000FF0100FF6D00FFDB00B7FF004DFF1500FF7000FFC000FFFF00FCFF00C9
FF00A6FF0094FF0094FF00A6FF00C9FF00FCC000FF7000FF1500FF004DFF00B7FF00FFDB
00FF6D00FF0165FF00C4FF00FFE500FF9B00FF5F00FF3100FF1200FF0400FF0500FF1500
FF3500FF6100FF9A00FFDD00D7FF0087FF0032FF0000FF2300FF7800FFCA00E7FF00A0FF
0062FF002CFF0002FF1D00FF3200FF3B00FF3A00FF2E00FF1800FF0005FF002AFF0055FF
0085FF00B8FF00ECFF00FFDE
00FF7400FFAA00FFE200E4FF00ACFF0076FF0043FF0016FF0F00FF2C00FF3F00FF4700FF
4300FF3300FF1700FF0011FF0045FF0084FF00CCFF00FFE300FF8E00FF3524FF007DFF00
D2FF00FFDF00FF9A00FF6000FF3300FF1500FF0700FF0900FF1D00FF4100FF7600FFBB00
F3FF0095FF002FFF0000FF3D00FFAE00E0FF0072FF000AFF5500FFAA00FFF000FFFF00D8
FF00B2FF009FFF009FFF00B2FF00D8F000FFAA00FF5500FF000AFF0072FF00E0FF00FFAE
00FF3D2FFF0095FF00F3FF00FFBB00FF7600FF4100FF1D00FF0900FF0700FF1500FF3300
FF6000FF9A00FFDF00D2FF007DFF0024FF0000FF3500FF8E00FFE300CCFF0084FF0045FF
0011FF1700FF3300FF4300FF4700FF3F00FF2C00FF0F00FF0016FF0043FF0076FF00ACFF
00E4FF00FFE200FFAA00FF74
00FF1000FF4100FF7700FFB100FFEE00D4FF0098FF005FFF002BFF0100FF2600FF4000FF
4E00FF5000FF4400FF2B00FF0400FF002FFF006EFF00B8FF00FFF500FF9C00FF3F1EFF00
7AFF00D2FF00FFDD00FF9600FF5C00FF3000FF1400FF0A00FF1200FF2C00FF5900FF9600
FFE300C2FF005DFF0000FF1000FF8200FFF60097FF002AFF3A00FF9300FFDE00FFFF00E7
FF00C0FF00ABFF00ABFF00C0FF00E7DE00FF9300FF3A00FF002AFF0097FF00FFF600FF82
00FF105DFF00C2FF00FFE300FF9600FF5900FF2C00FF1200FF0A00FF1400FF3000FF5C00
FF9600FFDD00D2FF007AFF001EFF0000FF3F00FF9C00FFF500B8FF006EFF002FFF0400FF
2B00FF4400FF5000FF4E00FF4000FF2600FF0100FF002BFF005FFF0098FF00D4FF00FFEE
00FFB100FF7700FF4100FF10
44FF001DFF0000FF1100FF4700FF8300FFC200FBFF00BAFF007BFF0041FF000DFF1D00FF
3D00FF5100FF5800FF5000FF3900FF1400FF001EFF005FFF00ABFF00FFFF00FFA300FF43
1EFF007DFF00D7FF00FFD600FF8F00FF5500FF2C00FF1500FF1000FF1F00FF4200FF7700
FFBE00EBFF0088FF001BFF0000FF5800FFCE00BBFF004AFF1E00FF7C00FFCB00FFFF00F7
FF00CDFF00B8FF00B8FF00CDFF00F7CB00FF7C00FF1E00FF004AFF00BBFF00FFCE00FF58
1BFF0088FF00EBFF00FFBE00FF7700FF4200FF1F00FF1000FF1500FF2C00FF5500FF8F00
FFD600D7FF007DFF001EFF0000FF4300FFA300FFFF00ABFF005FFF001EFF1400FF3900FF
5000FF5800FF5100FF3D00FF1D00FF000DFF0041FF007BFF00BAFF00FBFF00FFC200FF83
00FF4700FF111DFF0044FF00
81FF0068FF0045FF0018FF0000FF1C00FF5800FF9A00FFDE00DBFF0097FF0057FF001DFF
1300FF3900FF5200FF5D00FF5800FF4400FF1F00FF0014FF0056FF00A4FF00FCFF00FFA3
00FF3F24FF0087FF00E3FF00FFCA00FF8400FF4E00FF2900FF1700FF1A00FF3200FF5F00
FF9E00FFEF00B0FF0044FF0000FF2F00FFA800DEFF0069FF0300FF6500FFB700FFF800FF
FF00DBFF00C4FF00C4FF00DBF800FFB700FF6500FF0300FF0069FF00DEFF00FFA800FF2F
44FF00B0FF00FFEF00FF9E00FF5F00FF3200FF1A00FF1700FF2900FF4E00FF8400FFCA00
E3FF0087FF0024FF0000FF3F00FFA300FCFF00A4FF0056FF0014FF1F00FF4400FF5800FF
5D00FF5200FF3900FF1300FF001DFF0057FF0097FF00DBFF00FFDE00FF9A00FF5800FF1C
18FF0045FF0068FF0081FF00
A0FF0098FF0085FF0066FF003CFF0009FF0000FF3200FF7500FFBC00F9FF00B1FF006BFF
002CFF0A00FF3500FF5200FF6000FF5E00FF4A00FF2600FF000EFF0053FF00A4FF00FFFF
00FF9C00FF3532FF0096FF00F3FF00FFBB00FF7700FF4500FF2700FF1D00FF2A00FF4C00
FF8400FFCF00D4FF006AFF0000FF0800FF8200FFFE0088FF0017FF4E00FFA400FFE800FF
FF00E9FF00D1FF00D1FF00E9E800FFA400FF4E00FF0017FF0088FF00FFFE00FF8200FF08
6AFF00D4FF00FFCF00FF8400FF4C00FF2A00FF1D00FF2700FF4500FF7700FFBB00F3FF00
96FF0032FF0000FF3500FF9C00FFFF00A4FF0053FF000EFF2600FF4A00FF5E00FF6000FF
5200FF3500FF0A00FF002CFF006BFF00B1FF00F9FF00FFBC00FF7500FF3209FF003CFF00
66FF0085FF0098FF00A0FF00
9FFF00AAFF00A9FF009BFF0081FF005BFF002AFF0000FF1100FF5500FF9E00FFEB00C7FF
007DFF0039FF0100FF3000FF5100FF6200FF6100FF4D00FF2800FF000EFF0056FF00ABFF
00FFF500FF8E00FF2346FF00ABFF00FFF800FFA900FF6A00FF3E00FF2800FF2900FF4000
FF6F00FFB300F5FF008EFF001BFF0000FF5E00FFDC00A7FF0032FF3600FF9100FFD700FF
FF00F7FF00DEFF00DEFF00F7D700FF9100FF3600FF0032FF00A7FF00FFDC00FF5E1BFF00
8EFF00F5FF00FFB300FF6F00FF4000FF2900FF2800FF3E00FF6A00FFA900FFF800ABFF00
46FF0000FF2300FF8E00FFF500ABFF0056FF000EFF2800FF4D00FF6100FF6200FF5100FF
3000FF0100FF0039FF007DFF00C7FF00FFEB00FF9E00FF5500FF112AFF005BFF0081FF00
9BFF00A9FF00AAFF009FFF00
80FF009CFF00AEFF00B4FF00ACFF0097FF0075FF0045FF000AFF0000FF3900FF8500FFD4
00DAFF008CFF0043FF0004FF2D00FF5100FF6300FF6200FF4D00FF2600FF0014FF005FFF
00B8FF00FFE300FF7800FF0B60FF00C5FF00FFE000FF9500FF5D00FF3A00FF2E00FF3B00
FF6000FF9C00FFED00AFFF003EFF0000FF3C00FFBB00C5FF004DFF1F00FF7D00FFC700FF
FA00FFFF00EBFF00EBFA00FFC700FF7D00FF1F00FF004DFF00C5FF00FFBB00FF3C3EFF00
AFFF00FFED00FF9C00FF6000FF3B00FF2E00FF3A00FF5D00FF9500FFE000C5FF0060FF00
00FF0B00FF7800FFE300B8FF005FFF0014FF2600FF4D00FF6200FF6300FF5100FF2D00FF
0004FF0043FF008CFF00DAFF00FFD400FF8500FF390AFF0045FF0075FF0097FF00ACFF00
B4FF00AEFF009CFF0080FF00
47FF0073FF0096FF00AFFF00BBFF00B9FF00A8FF0089FF005CFF0021FF0000FF2300FF70
00FFC200E9FF0097FF004BFF0009FF2B00FF5000FF6300FF6100FF4A00FF1F00FF001EFF
006EFF00CCFF00FFCA00FF5C13FF007FFF00E3FF00FFC600FF8200FF5200FF3A00FF3B00
FF5600FF8900FFD400CDFF005EFF0000FF1B00FF9B00E3FF0067FF0800FF6A00FFB600FF
EC00FFFF00F8FF00F8EC00FFB600FF6A00FF0800FF0067FF00E3FF00FF9B00FF1B5EFF00
CDFF00FFD400FF8900FF5600FF3B00FF3A00FF5200FF8200FFC600E3FF007FFF0013FF00
00FF5C00FFCA00CCFF006EFF001EFF1F00FF4A00FF6100FF6300FF5000FF2B00FF0009FF
004BFF0097FF00E9FF00FFC200FF7000FF2321FF005CFF0089FF00A8FF00B9FF00BBFF00
AFFF0096FF0073FF0047FF00
00FF0531FF0063FF008DFF00ACFF00BEFF00C1FF00B5FF0099FF006DFF0034FF0000FF11
00FF6000FFB500F2FF009DFF004FFF000AFF2B00FF5100FF6200FF5E00FF4400FF1400FF
002FFF0084FF00E7FF00FFAB00FF3937FF00A2FF00FFFD00FFAC00FF7000FF4C00FF4100
FF5100FF7B00FFBE00E7FF007CFF0003FF0000FF7D00FFFF0081FF000DFF5600FFA600FF
DD00FFFA00FFFA00FFDD00FFA600FF5600FF000DFF0081FF00FFFF00FF7D03FF007CFF00
E7FF00FFBE00FF7B00FF5100FF4100FF4C00FF7000FFAC00FFFD00A2FF0037FF0000FF39
00FFAB00E7FF0084FF002FFF1400FF4400FF5E00FF6200FF5100FF2B00FF000AFF004FFF
009DFF00F2FF00FFB500FF6000FF1134FF006DFF0099FF00B5FF00C1FF00BEFF00ACFF00
8DFF0063FF0031FF0000FF05
00FF5C00FF1E1CFF0054FF0083FF00A8FF00BFFF00C7FF00BFFF00A5FF007BFF0041FF00
00FF0500FF5600FFAE00F7FF00A0FF004FFF0009FF2D00FF5100FF6000FF5800FF3900FF
0400FF0045FF00A0FF00FFF600FF8500FF1160FF00C8FF00FFDC00FF9400FF6300FF4D00
FF5100FF7200FFAD00FFFF0097FF0020FF0000FF6000FFE4009BFF0024FF4300FF9500FF
CF00FFEC00FFEC00FFCF00FF9500FF4300FF0024FF009BFF00FFE400FF6020FF0097FF00
FFFF00FFAD00FF7200FF5100FF4D00FF6300FF9400FFDC00C8FF0060FF0000FF1100FF85
00FFF600A0FF0045FF0400FF3900FF5800FF6000FF5100FF2D00FF0009FF004FFF00A0FF
00F7FF00FFAE00FF5600FF0541FF007BFF00A5FF00BFFF00C7FF00BFFF00A8FF0083FF00
54FF001CFF0000FF1E00FF5C
00FFB400FF7600FF3609FF0045FF0079FF00A3FF00BFFF00CBFF00C5FF00ADFF0084FF00
4AFF0002FF0000FF5100FFAB00F7FF009DFF004BFF0004FF3000FF5200FF5D00FF5000FF
2B00FF0011FF0062FF00C4FF00FFCD00FF581CFF008CFF00F0FF00FFBD00FF8000FF5D00
FF5700FF6D00FF9F00FFEC00B0FF003BFF0000FF4400FFC900B4FF003AFF2F00FF8500FF
C000FFDF00FFDF00FFC000FF8500FF2F00FF003AFF00B4FF00FFC900FF443BFF00B0FF00
FFEC00FF9F00FF6D00FF5700FF5D00FF8000FFBD00F0FF008CFF001CFF0000FF5800FFCD
00C4FF0062FF0011FF2B00FF5000FF5D00FF5200FF3000FF0004FF004BFF009DFF00F7FF
00FFAB00FF5102FF004AFF0084FF00ADFF00C5FF00CBFF00BFFF00A3FF0079FF0045FF00
09FF0000FF3600FF7600FFB4
00FCFF00FFCC00FF8E00FF4B00FF0837FF0070FF009EFF00BEFF00CDFF00C9FF00B3FF00
89FF004EFF0004FF0000FF5100FFAE00F2FF0097FF0043FF0100FF3500FF5200FF5800FF
4400FF1700FF002CFF0085FF00EEFF00FF9E00FF274DFF00B9FF00FFE900FFA100FF7300
FF6100FF6C00FF9600FFDB00C6FF0054FF0000FF2900FFAF00CCFF0050FF1C00FF7400FF
B100FFD100FFD100FFB100FF7400FF1C00FF0050FF00CCFF00FFAF00FF2954FF00C6FF00
FFDB00FF9600FF6C00FF6100FF7300FFA100FFE900B9FF004DFF0000FF2700FF9E00EEFF
0085FF002CFF1700FF4400FF5800FF5200FF3500FF0100FF0043FF0097FF00F2FF00FFAE
00FF5104FF004EFF0089FF00B3FF00C9FF00CDFF00BEFF009EFF0070FF0037FF0000FF08
00FF4B00FF8E00FFCC00FCFF
00BCFF00E7FF00FFE200FFA200FF5E00FF182BFF0068FF0099FF00BCFF00CEFF00CCFF00
B6FF008BFF004EFF0002FF0000FF5600FFB500E9FF008CFF0039FF0A00FF3900FF5100FF
5000FF3300FF0002FF004FFF00B0FF00FFE000FF690EFF0081FF00E6FF00FFC600FF8C00
FF6F00FF7000FF9000FFCE00D9FF006CFF0000FF1100FF9600E3FF0065FF0900FF6300FF
A200FFC300FFC300FFA200FF6300FF0900FF0065FF00E3FF00FF9600FF116CFF00D9FF00
FFCE00FF9000FF7000FF6F00FF8C00FFC600E6FF0081FF000EFF0000FF6900FFE000B0FF
004FFF0002FF3300FF5000FF5100FF3900FF0A00FF0039FF008CFF00E9FF00FFB500FF56
02FF004EFF008BFF00B6FF00CCFF00CEFF00BCFF0099FF0068FF002BFF0000FF1800FF5E
00FFA200FFE200E7FF00BCFF
0092FF00ADFF00D6FF00FFF300FFB300FF6D00FF2422FF0061FF0096FF00BBFF00CEFF00
CDFF00B6FF0089FF004AFF0000FF0500FF6000FFC200DAFF007DFF002CFF1300FF3D00FF
4E00FF4300FF1D00FF0022FF0079FF00E2FF00FFA900FF2F47FF00B4FF00FFEF00FFAA00
FF8100FF7700FF8E00FFC300EAFF0081FF0006FF0000FF7E00FAFF007AFF0008FF5300FF
9400FFB500FFB500FF9400FF5300FF0008FF007AFF00FAFF00FF7E06FF0081FF00EAFF00
FFC300FF8E00FF7700FF8100FFAA00FFEF00B4FF0047FF0000FF2F00FFA900E2FF0079FF
0022FF1D00FF4300FF4E00FF3D00FF1300FF002CFF007DFF00DAFF00FFC200FF6000FF05
4AFF0089FF00B6FF00CDFF00CEFF00BBFF0096FF0061FF0022FF0000FF2400FF6D00FFB3
00FFF300D6FF00ADFF0092FF
0082FF0089FF00A1FF00C9FF00FEFF00FFC000FF7800FF2D1BFF005EFF0094FF00BBFF00
CEFF00CCFF00B3FF0084FF0041FF0000FF1100FF7000FFD400C7FF006BFF001DFF1D00FF
4000FF4700FF3200FF0100FF0049FF00AAFF00FFE500FF6C0DFF0081FF00E5FF00FFCB00
FF9700FF8300FF8F00FFBC00F8FF0093FF001BFF0000FF6800FFEF008FFF001BFF4200FF
8500FFA700FFA700FF8500FF4200FF001BFF008FFF00FFEF00FF681BFF0093FF00F8FF00
FFBC00FF8F00FF8300FF9700FFCB00E5FF0081FF000DFF0000FF6C00FFE500AAFF0049FF
0100FF3200FF4700FF4000FF1D00FF001DFF006BFF00C7FF00FFD400FF7000FF1141FF00
84FF00B3FF00CCFF00CEFF00BBFF0094FF005EFF001BFF0000FF2D00FF7800FFC000FEFF
00C9FF00A1FF0089FF0082FF
008BFF007FFF0083FF0098FF00BFFF00F4FF00FFCA00FF8000FF3318FF005CFF0094FF00
BBFF00CEFF00C9FF00ADFF007BFF0034FF0000FF2300FF8500FFEB00B1FF0057FF000DFF
2600FF3F00FF3B00FF1A00FF0023FF0079FF00E3FF00FFA600FF2B4CFF00B8FF00FFEF00
FFB100FF9200FF9400FFB800FFFB00A4FF002FFF0000FF5200FFDA00A2FF002DFF3200FF
7600FF9A00FF9A00FF7600FF3200FF002DFF00A2FF00FFDA00FF522FFF00A4FF00FFFB00
FFB800FF9400FF9200FFB100FFEF00B8FF004CFF0000FF2B00FFA600E3FF0079FF0023FF
1A00FF3B00FF3F00FF2600FF000DFF0057FF00B1FF00FFEB00FF8500FF2334FF007BFF00
ADFF00C9FF00CEFF00BBFF0094FF005CFF0018FF0000FF3300FF8000FFCA00F4FF00BFFF
0098FF0083FF007FFF008BFF
00ACFF008DFF007DFF007FFF0092FF00B8FF00EDFF00FFCF00FF8400FF3418FF005EFF00
96FF00BCFF00CDFF00C5FF00A5FF006DFF0021FF0000FF3900FF9E00F9FF0097FF0041FF
0100FF2C00FF3A00FF2900FF0005FF004FFF00B0FF00FFDD00FF6217FF0089FF00EAFF00
FFCD00FFA300FF9B00FFB600FFF300B3FF0042FF0000FF3E00FFC500B6FF003FFF2200FF
6800FF8C00FF8C00FF6800FF2200FF003FFF00B6FF00FFC500FF3E42FF00B3FF00FFF300
FFB600FF9B00FFA300FFCD00EAFF0089FF0017FF0000FF6200FFDD00B0FF004FFF0005FF
2900FF3A00FF2C00FF0100FF0041FF0097FF00F9FF00FF9E00FF3921FF006DFF00A5FF00
C5FF00CDFF00BCFF0096FF005EFF0018FF0000FF3400FF8400FFCF00EDFF00B8FF0092FF
007FFF007DFF008DFF00ACFF
00DFFF00B1FF008FFF007CFF007CFF008EFF00B3FF00EAFF00FFD100FF8400FF331BFF00
61FF0099FF00BEFF00CBFF00BFFF0099FF005CFF000AFF0000FF5500FFBC00DBFF007BFF
002BFF0F00FF2E00FF2E00FF0F00FF002DFF0084FF00EFFF00FF9700FF1B5AFF00C2FF00
FFEC00FFB800FFA600FFB800FFEC00BFFF0052FF0000FF2B00FFB200C9FF0050FF1200FF
5900FF7E00FF7E00FF5900FF1200FF0050FF00C9FF00FFB200FF2B52FF00BFFF00FFEC00
FFB800FFA600FFB800FFEC00C2FF005AFF0000FF1B00FF9700EFFF0084FF002DFF0F00FF
2E00FF2E00FF0F00FF002BFF007BFF00DBFF00FFBC00FF550AFF005CFF0099FF00BFFF00
CBFF00BEFF0099FF0061FF001BFF0000FF3300FF8400FFD100EAFF00B3FF008EFF007CFF
007CFF008FFF00B1FF00DFFF
00FFE000E6FF00B5FF0091FF007CFF007AFF008CFF00B2FF00EAFF00FFCF00FF8000FF2D
22FF0068FF009EFF00BFFF00C7FF00B5FF0089FF0045FF0000FF1100FF7500FFDE00BAFF
005FFF0016FF1800FF2A00FF1B00FF0012FF005EFF00C1FF00FFCA00FF4E2BFF0099FF00
F2FF00FFD000FFB400FFBC00FFE800CAFF0061FF0000FF1A00FF9F00DBFF0061FF0300FF
4B00FF7000FF7000FF4B00FF0300FF0061FF00DBFF00FF9F00FF1A61FF00CAFF00FFE800
FFBC00FFB400FFD000F2FF0099FF002BFF0000FF4E00FFCA00C1FF005EFF0012FF1B00FF
2A00FF1800FF0016FF005FFF00BAFF00FFDE00FF7500FF1145FF0089FF00B5FF00C7FF00
BFFF009EFF0068FF0022FF0000FF2D00FF8000FFCF00EAFF00B2FF008CFF007AFF007CFF
0091FF00B5FF00E6FF00FFE0
00FF9A00FFD800ECFF00B9FF0092FF007CFF0079FF008CFF00B3FF00EDFF00FFCA00FF78
00FF242BFF0070FF00A3FF00BFFF00C1FF00A8FF0075FF002AFF0000FF3200FF9A00FBFF
0098FF0043FF0005FF1C00FF1E00FF0000FF003FFF0099FF00FFF800FF7F00FF036FFF00
D0FF00FFE900FFC400FFC300FFE700D2FF006EFF0000FF0900FF8E00ECFF0072FF000CFF
3C00FF6200FF6200FF3C00FF000CFF0072FF00ECFF00FF8E00FF096EFF00D2FF00FFE700
FFC300FFC400FFE900D0FF006FFF0000FF0300FF7F00FFF80099FF003FFF0000FF1E00FF
1C00FF0005FF0043FF0098FF00FBFF00FF9A00FF322AFF0075FF00A8FF00C1FF00BFFF00
A3FF0070FF002BFF0000FF2400FF7800FFCA00EDFF00B3FF008CFF0079FF007CFF0092FF
00B9FF00ECFF00FFD800FF9A
00FF5700FF9300FFD300EFFF00BAFF0092FF007CFF007AFF008EFF00B8FF00F4FF00FFC0
00FF6D00FF1837FF0079FF00A8FF00BEFF00B9FF0097FF005BFF0009FF0000FF5800FFC2
00D4FF0076FF002AFF0600FF1900FF0900FF0027FF0076FF00DCFF00FFAD00FF3145FF00
ACFF00FAFF00FFD600FFCC00FFE700D8FF007AFF0004FF0000FF7D00FDFF0082FF001BFF
2E00FF5500FF5500FF2E00FF001BFF0082FF00FDFF00FF7D04FF007AFF00D8FF00FFE700
FFCC00FFD600FAFF00ACFF0045FF0000FF3100FFAD00DCFF0076FF0027FF0900FF1900FF
0600FF002AFF0076FF00D4FF00FFC200FF5809FF005BFF0097FF00B9FF00BEFF00A8FF00
79FF0037FF0000FF1800FF6D00FFC000F4FF00B8FF008EFF007AFF007CFF0092FF00BAFF
00EFFF00FFD300FF9300FF57
00FF1C00FF5100FF8F00FFD000F0FF00BAFF0092FF007CFF007CFF0092FF00BFFF00FEFF
00FFB300FF5E00FF0845FF0083FF00ACFF00BBFF00ACFF0081FF003CFF0000FF1C00FF83
00FFEE00ACFF0055FF0016FF0B00FF0B00FF0016FF0059FF00B6FF00FFD800FF5D1BFF00
87FF00DDFF00FFEA00FFD700FFEA00DDFF0084FF0012FF0000FF6D00FFF10092FF002AFF
2000FF4700FF4700FF2000FF002AFF0092FF00FFF100FF6D12FF0084FF00DDFF00FFEA00
FFD700FFEA00DDFF0087FF001BFF0000FF5D00FFD800B6FF0059FF0016FF0B00FF0B00FF
0016FF0055FF00ACFF00FFEE00FF8300FF1C3CFF0081FF00ACFF00BBFF00ACFF0083FF00
45FF0000FF0800FF5E00FFB300FEFF00BFFF0092FF007CFF007CFF0092FF00BAFF00F0FF
00FFD000FF8F00FF5100FF1C
0EFF0000FF1900FF4F00FF8E00FFD000EFFF00B9FF0091FF007CFF007FFF0098FF00C9FF
00FFF300FFA200FF4B09FF0054FF008DFF00AFFF00B4FF009BFF0066FF0018FF0000FF47
00FFB100E4FF0085FF0039FF0008FF0600FF000CFF0042FF0095FF00FFFF00FF8700FF0D
62FF00BFFF00FFFF00FFE400FFEF00DFFF008CFF001EFF0000FF5F00FFE200A1FF0039FF
1200FF3A00FF3A00FF1200FF0039FF00A1FF00FFE200FF5F1EFF008CFF00DFFF00FFEF00
FFE400FFFF00BFFF0062FF0000FF0D00FF8700FFFF0095FF0042FF000CFF0600FF0008FF
0039FF0085FF00E4FF00FFB100FF4718FF0066FF009BFF00B4FF00AFFF008DFF0054FF00
09FF0000FF4B00FFA200FFF300C9FF0098FF007FFF007CFF0091FF00B9FF00EFFF00FFD0
00FF8E00FF4F00FF190EFF00
28FF0010FF0000FF1800FF4F00FF8F00FFD300ECFF00B5FF008FFF007DFF0083FF00A1FF
00D6FF00FFE200FF8E00FF361CFF0063FF0096FF00AEFF00A9FF0085FF0045FF0000FF11
00FF7700FFE200B8FF0061FF0023FF0005FF0009FF0031FF007AFF00DCFF00FFAF00FF35
3EFF00A0FF00E7FF00FFF300FFF600E0FF0093FF002AFF0000FF5100FFD300B0FF0047FF
0400FF2D00FF2D00FF0400FF0047FF00B0FF00FFD300FF512AFF0093FF00E0FF00FFF600
FFF300E7FF00A0FF003EFF0000FF3500FFAF00DCFF007AFF0031FF0009FF0005FF0023FF
0061FF00B8FF00FFE200FF7700FF1145FF0085FF00A9FF00AEFF0096FF0063FF001CFF00
00FF3600FF8E00FFE200D6FF00A1FF0083FF007DFF008FFF00B5FF00ECFF00FFD300FF8F
00FF4F00FF1810FF0028FF00
2EFF0029FF0010FF0000FF1900FF5100FF9300FFD800E6FF00B1FF008DFF007FFF0089FF
00ADFF00E7FF00FFCC00FF7600FF1E31FF0073FF009CFF00AAFF0098FF0068FF001DFF00
00FF4100FFAA00ECFF008DFF0044FF0017FF000CFF0026FF0063FF00BDFF00FFD400FF5C
19FF0081FF00CFFF00FBFF00FFFE00DFFF0098FF0033FF0000FF4400FFC500BFFF0055FF
0008FF2000FF2000FF0008FF0055FF00BFFF00FFC500FF4433FF0098FF00DFFF00FFFE00
FBFF00CFFF0081FF0019FF0000FF5C00FFD400BDFF0063FF0026FF000CFF0017FF0044FF
008DFF00ECFF00FFAA00FF411DFF0068FF0098FF00AAFF009CFF0073FF0031FF0000FF1E
00FF7600FFCC00E7FF00ADFF0089FF007FFF008DFF00B1FF00E6FF00FFD800FF9300FF51
00FF1910FF0029FF002EFF00
21FF002EFF0028FF000EFF0000FF1C00FF5700FF9A00FFE000DFFF00ACFF008BFF0082FF
0092FF00BCFF00FCFF00FFB400FF5C00FF0547FF0080FF009FFF00A0FF0081FF0044FF00
00FF1000FF7400FFDE00BCFF0068FF002EFF0016FF0022FF0052FF00A1FF00FFF600FF80
00FF0961FF00B5FF00E9FF00F7FF00DCFF009CFF003CFF0000FF3900FFB800CDFF0063FF
0015FF1300FF1300FF0015FF0063FF00CDFF00FFB800FF393CFF009CFF00DCFF00F7FF00
E9FF00B5FF0061FF0000FF0900FF8000FFF600A1FF0052FF0022FF0016FF002EFF0068FF
00BCFF00FFDE00FF7400FF1044FF0081FF00A0FF009FFF0080FF0047FF0000FF0500FF5C
00FFB400FCFF00BCFF0092FF0082FF008BFF00ACFF00DFFF00FFE000FF9A00FF5700FF1C
0EFF0028FF002EFF0021FF00
pop
P
% End of image
.01923 .01923 m
.01923 .98077 L
s
.02885 .01923 m
.02885 .98077 L
s
.03846 .01923 m
.03846 .98077 L
s
.04808 .01923 m
.04808 .98077 L
s
.05769 .01923 m
.05769 .98077 L
s
.06731 .01923 m
.06731 .98077 L
s
.07692 .01923 m
.07692 .98077 L
s
.08654 .01923 m
.08654 .98077 L
s
.09615 .01923 m
.09615 .98077 L
s
.10577 .01923 m
.10577 .98077 L
s
.11538 .01923 m
.11538 .98077 L
s
.125 .01923 m
.125 .98077 L
s
.13462 .01923 m
.13462 .98077 L
s
.14423 .01923 m
.14423 .98077 L
s
.15385 .01923 m
.15385 .98077 L
s
.16346 .01923 m
.16346 .98077 L
s
.17308 .01923 m
.17308 .98077 L
s
.18269 .01923 m
.18269 .98077 L
s
.19231 .01923 m
.19231 .98077 L
s
.20192 .01923 m
.20192 .98077 L
s
.21154 .01923 m
.21154 .98077 L
s
.22115 .01923 m
.22115 .98077 L
s
.23077 .01923 m
.23077 .98077 L
s
.24038 .01923 m
.24038 .98077 L
s
.25 .01923 m
.25 .98077 L
s
.25962 .01923 m
.25962 .98077 L
s
.26923 .01923 m
.26923 .98077 L
s
.27885 .01923 m
.27885 .98077 L
s
.28846 .01923 m
.28846 .98077 L
s
.29808 .01923 m
.29808 .98077 L
s
.30769 .01923 m
.30769 .98077 L
s
.31731 .01923 m
.31731 .98077 L
s
.32692 .01923 m
.32692 .98077 L
s
.33654 .01923 m
.33654 .98077 L
s
.34615 .01923 m
.34615 .98077 L
s
.35577 .01923 m
.35577 .98077 L
s
.36538 .01923 m
.36538 .98077 L
s
.375 .01923 m
.375 .98077 L
s
.38462 .01923 m
.38462 .98077 L
s
.39423 .01923 m
.39423 .98077 L
s
.40385 .01923 m
.40385 .98077 L
s
.41346 .01923 m
.41346 .98077 L
s
.42308 .01923 m
.42308 .98077 L
s
.43269 .01923 m
.43269 .98077 L
s
.44231 .01923 m
.44231 .98077 L
s
.45192 .01923 m
.45192 .98077 L
s
.46154 .01923 m
.46154 .98077 L
s
.47115 .01923 m
.47115 .98077 L
s
.48077 .01923 m
.48077 .98077 L
s
.49038 .01923 m
.49038 .98077 L
s
.5 .01923 m
.5 .98077 L
s
.50962 .01923 m
.50962 .98077 L
s
.51923 .01923 m
.51923 .98077 L
s
.52885 .01923 m
.52885 .98077 L
s
.53846 .01923 m
.53846 .98077 L
s
.54808 .01923 m
.54808 .98077 L
s
.55769 .01923 m
.55769 .98077 L
s
.56731 .01923 m
.56731 .98077 L
s
.57692 .01923 m
.57692 .98077 L
s
.58654 .01923 m
.58654 .98077 L
s
.59615 .01923 m
.59615 .98077 L
s
.60577 .01923 m
.60577 .98077 L
s
.61538 .01923 m
.61538 .98077 L
s
.625 .01923 m
.625 .98077 L
s
.63462 .01923 m
.63462 .98077 L
s
.64423 .01923 m
.64423 .98077 L
s
.65385 .01923 m
.65385 .98077 L
s
.66346 .01923 m
.66346 .98077 L
s
.67308 .01923 m
.67308 .98077 L
s
.68269 .01923 m
.68269 .98077 L
s
.69231 .01923 m
.69231 .98077 L
s
.70192 .01923 m
.70192 .98077 L
s
.71154 .01923 m
.71154 .98077 L
s
.72115 .01923 m
.72115 .98077 L
s
.73077 .01923 m
.73077 .98077 L
s
.74038 .01923 m
.74038 .98077 L
s
.75 .01923 m
.75 .98077 L
s
.75962 .01923 m
.75962 .98077 L
s
.76923 .01923 m
.76923 .98077 L
s
.77885 .01923 m
.77885 .98077 L
s
.78846 .01923 m
.78846 .98077 L
s
.79808 .01923 m
.79808 .98077 L
s
.80769 .01923 m
.80769 .98077 L
s
.81731 .01923 m
.81731 .98077 L
s
.82692 .01923 m
.82692 .98077 L
s
.83654 .01923 m
.83654 .98077 L
s
.84615 .01923 m
.84615 .98077 L
s
.85577 .01923 m
.85577 .98077 L
s
.86538 .01923 m
.86538 .98077 L
s
.875 .01923 m
.875 .98077 L
s
.88462 .01923 m
.88462 .98077 L
s
.89423 .01923 m
.89423 .98077 L
s
.90385 .01923 m
.90385 .98077 L
s
.91346 .01923 m
.91346 .98077 L
s
.92308 .01923 m
.92308 .98077 L
s
.93269 .01923 m
.93269 .98077 L
s
.94231 .01923 m
.94231 .98077 L
s
.95192 .01923 m
.95192 .98077 L
s
.96154 .01923 m
.96154 .98077 L
s
.97115 .01923 m
.97115 .98077 L
s
.98077 .01923 m
.98077 .98077 L
s
.01923 .01923 m
.98077 .01923 L
s
.01923 .02885 m
.98077 .02885 L
s
.01923 .03846 m
.98077 .03846 L
s
.01923 .04808 m
.98077 .04808 L
s
.01923 .05769 m
.98077 .05769 L
s
.01923 .06731 m
.98077 .06731 L
s
.01923 .07692 m
.98077 .07692 L
s
.01923 .08654 m
.98077 .08654 L
s
.01923 .09615 m
.98077 .09615 L
s
.01923 .10577 m
.98077 .10577 L
s
.01923 .11538 m
.98077 .11538 L
s
.01923 .125 m
.98077 .125 L
s
.01923 .13462 m
.98077 .13462 L
s
.01923 .14423 m
.98077 .14423 L
s
.01923 .15385 m
.98077 .15385 L
s
.01923 .16346 m
.98077 .16346 L
s
.01923 .17308 m
.98077 .17308 L
s
.01923 .18269 m
.98077 .18269 L
s
.01923 .19231 m
.98077 .19231 L
s
.01923 .20192 m
.98077 .20192 L
s
.01923 .21154 m
.98077 .21154 L
s
.01923 .22115 m
.98077 .22115 L
s
.01923 .23077 m
.98077 .23077 L
s
.01923 .24038 m
.98077 .24038 L
s
.01923 .25 m
.98077 .25 L
s
.01923 .25962 m
.98077 .25962 L
s
.01923 .26923 m
.98077 .26923 L
s
.01923 .27885 m
.98077 .27885 L
s
.01923 .28846 m
.98077 .28846 L
s
.01923 .29808 m
.98077 .29808 L
s
.01923 .30769 m
.98077 .30769 L
s
.01923 .31731 m
.98077 .31731 L
s
.01923 .32692 m
.98077 .32692 L
s
.01923 .33654 m
.98077 .33654 L
s
.01923 .34615 m
.98077 .34615 L
s
.01923 .35577 m
.98077 .35577 L
s
.01923 .36538 m
.98077 .36538 L
s
.01923 .375 m
.98077 .375 L
s
.01923 .38462 m
.98077 .38462 L
s
.01923 .39423 m
.98077 .39423 L
s
.01923 .40385 m
.98077 .40385 L
s
.01923 .41346 m
.98077 .41346 L
s
.01923 .42308 m
.98077 .42308 L
s
.01923 .43269 m
.98077 .43269 L
s
.01923 .44231 m
.98077 .44231 L
s
.01923 .45192 m
.98077 .45192 L
s
.01923 .46154 m
.98077 .46154 L
s
.01923 .47115 m
.98077 .47115 L
s
.01923 .48077 m
.98077 .48077 L
s
.01923 .49038 m
.98077 .49038 L
s
.01923 .5 m
.98077 .5 L
s
.01923 .50962 m
.98077 .50962 L
s
.01923 .51923 m
.98077 .51923 L
s
.01923 .52885 m
.98077 .52885 L
s
.01923 .53846 m
.98077 .53846 L
s
.01923 .54808 m
.98077 .54808 L
s
.01923 .55769 m
.98077 .55769 L
s
.01923 .56731 m
.98077 .56731 L
s
.01923 .57692 m
.98077 .57692 L
s
.01923 .58654 m
.98077 .58654 L
s
.01923 .59615 m
.98077 .59615 L
s
.01923 .60577 m
.98077 .60577 L
s
.01923 .61538 m
.98077 .61538 L
s
.01923 .625 m
.98077 .625 L
s
.01923 .63462 m
.98077 .63462 L
s
.01923 .64423 m
.98077 .64423 L
s
.01923 .65385 m
.98077 .65385 L
s
.01923 .66346 m
.98077 .66346 L
s
.01923 .67308 m
.98077 .67308 L
s
.01923 .68269 m
.98077 .68269 L
s
.01923 .69231 m
.98077 .69231 L
s
.01923 .70192 m
.98077 .70192 L
s
.01923 .71154 m
.98077 .71154 L
s
.01923 .72115 m
.98077 .72115 L
s
.01923 .73077 m
.98077 .73077 L
s
.01923 .74038 m
.98077 .74038 L
s
.01923 .75 m
.98077 .75 L
s
.01923 .75962 m
.98077 .75962 L
s
.01923 .76923 m
.98077 .76923 L
s
.01923 .77885 m
.98077 .77885 L
s
.01923 .78846 m
.98077 .78846 L
s
.01923 .79808 m
.98077 .79808 L
s
.01923 .80769 m
.98077 .80769 L
s
.01923 .81731 m
.98077 .81731 L
s
.01923 .82692 m
.98077 .82692 L
s
.01923 .83654 m
.98077 .83654 L
s
.01923 .84615 m
.98077 .84615 L
s
.01923 .85577 m
.98077 .85577 L
s
.01923 .86538 m
.98077 .86538 L
s
.01923 .875 m
.98077 .875 L
s
.01923 .88462 m
.98077 .88462 L
s
.01923 .89423 m
.98077 .89423 L
s
.01923 .90385 m
.98077 .90385 L
s
.01923 .91346 m
.98077 .91346 L
s
.01923 .92308 m
.98077 .92308 L
s
.01923 .93269 m
.98077 .93269 L
s
.01923 .94231 m
.98077 .94231 L
s
.01923 .95192 m
.98077 .95192 L
s
.01923 .96154 m
.98077 .96154 L
s
.01923 .97115 m
.98077 .97115 L
s
.01923 .98077 m
.98077 .98077 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00;@3oool0103ooml0W4P00018W03OoomH0?ooo`030;md000000000000
00@0000000<0B9cO0?ooo`3oool0DP3oool01@3ogi`0B0000000000007@0_ooo05P0oooo00<0_g@0
000000000000100000000`18W=l0oooo0?ooo`1J0?ooo`040?oog`2LB00004RL0=ooo`H0oooo000]
0?ooo`040?oog`2LB00004RL0=oooeT0oooo00<0gia80000000000000`0000000`00B9`0gooo0?oo
o`1A0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`1H0?ooo`030=nLB0000000000000<0
000000<004RL0=ooo`3oool0FP3oool0103ooml0W4P00018W03Oool60?ooo`00;@3oool0103ooml0
W4P00018W03OoomJ0?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0D@3oool00`2o
M00004RL0=ooo`020?ooo`030;md0000M;l0oooo05P0oooo00<0gia804RLg`3oool00P3oool00`3O
W7@0W=oo0?ooo`1J0?ooo`040?oog`2LB00004RL0=ooo`H0oooo000Y0?ooo`030;md000000000000
00D0000000<0M;oo0?ooo`3oool0F03oool00`3OW4P0M;oo0?ooo`1D0?ooo`030?oo_`1d0000B9cO
00<0oooo00<0gia80000M02oool0F@3oool00`3OW4P0M;oo0?ooo`1J0?ooo`030;md000000000000
00D0000000<0M;oo0?ooo`3oool00`3oool001d0oooo00<0oooO09a8000000001`0000001018W=l0
oooo0?oOW01dW=l20?ooo`040?oog`2LB00004RL0=oood`0oooo00<0oooO09a8000000001`000000
0`18W=l0oooo0?ooo`030?ooo`030;mdB02Lgol0oooo05<0oooo00<0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`1J0?ooo`030;mdB02Lgol0oooo05T0oooo00P0omnL07BLg`3oool0oooo0?oo
g`2LB00004RL0=ooo`H0oooo000Z0?ooo`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`1L0?oo
o`030?oog`2LB7@0_ooo05<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`1J0?oo
o`030?oog`2LB7@0_ooo05X0oooo00L0gia807Boo`3oool0oooO09a80000B9`0gooo00H0oooo000Z
0?ooo`070?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`1M0?ooo`030?noM000M;l0oooo0580oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`1K0?ooo`030?noM000M;l0oooo05T0oooo
00L0oooO09a8M02oool0oooO09a80000B9`0gooo00H0oooo000[0?ooo`060?oOW018M;l0oooO09a8
0000B9`0goooG@3oool00`3ookl0M01809cOo`1B0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol0F`3oool00`3ookl0M01809cOo`1J0?ooo`060?oOW018M;l0oooO09a80000B9`0gooo
1P3oool002`0oooo00D0gia807Bog`2LB00004RL0=ooo`1M0?ooo`030?oo_`1d0000B9cO0580oooo
00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`1K0?ooo`030?oo_`1d0000B9cO05/0oooo
00D0gia807Bog`2LB00004RL0=ooo`060?ooo`00;03oool01@3ooml0W4Q804Q80000B9`0gooo05P0
oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`1C0?ooo`030=nLB000B9`0gooo0080
oooo00<0_g@0001d_`3oool0EP3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO05/0
oooo00D0oooO09a8B018B00004RL0=ooo`060?ooo`00;@3oool0103ogi`0B0000018W03OoomH0?oo
o`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02ooomC0?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`1G0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02ooomL0?ooo`040?oO
W018000004RL0=ooo`H0oooo000^0?ooo`030=nLB000B9`0gooo05T0oooo00H0oooO09a800000000
00000000M02ooomE0?ooo`050?oo_`1d000000000000M02oool0F@3oool01P3ooml0W4P000000000
0000001d0;oooeh0oooo00<0gia80018W03Oool01P3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00200ooooo`00002@00000040oooo000P
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0D
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`04
0?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo01@0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo00@0oooo0@0000010?ooo`004`3oool0103ooml0W4P00018W03Oool90?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`1M0?oo
o`030000003oool0oooo05d0oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`040?oo
o`4000000@3oool001<0oooo00@0oooO09a80000B9`0gooo2@3oool00`000000oooo0?ooo`3o0?oo
oh`0oooo0@0000010?ooo`004`3oool0103ooml0W4P00018W03Oool90?ooo`030000003oool0oooo
0?l0ooooS03oool100000040oooo000?0?ooo`030;md00000000000000D0000000<0M;oo0?ooo`3o
ool01P3oool00`000000oooo0?ooo`3o0?oooh`0oooo0@0000010?ooo`000`3oool00`3ooml0W4P0
000000070000000404RLg`3oool0omnL07BLg`80oooo00@0oooO09a80000B9`0gooo2@3oool00`00
0000oooo0?ooo`3o0?oooh`0oooo0@0000010?ooo`00403oool01`3OW4P0M;oo0?ooo`3ooml0W4P0
0018W03Oool02@3oool3000000D0ooooo`000021000000@0oooo0`0000010?ooo`00403oool01`3o
oml0W4Qd0;ooo`3ooml0W4P00018W03Oool02@3oool00`000000oooo0?ooo`050?ooo`040000000Q
o`008Ol0000000<0;_l000D0000002So000Xo`00:?l00000000300ko000500000000oa`00?lL003o
700000000`00oeL0100000000?nJ003oVP000003003oh00500000000gol00=oo003Oo`0000000`00
[?l01@00000008_o002;o`00Rol0000000<008;o00@00000002Bo`00T_l000000`00_?l01@000000
0?co003lo`00o?l0000000<00?nd00D00000003oG000oe`00?mL00000003003o1@0400000017o`00
Aol0000000<0P?l000D0000009oo002Oo`00Wol0000000030:3o000500000021o`00POl0087o0000
00000`14o`00100000000?l@003o40000003003oM00500000000omh00?oN003ogP0000000`00_?l0
1@00000006So001Xo`00J?l0000000<002ko00@00000000Fo`005_l000000`008_l01@00000005;o
001Bo`00D_l0000000<00:7o00D00000003omP00ooH00?of00000003003oP00400000000o`T00?l9
000000<0HOl000D000000;Go002eo`00]Ol0000000030>Wo00050000003go`00mol00?Oo00000000
0P3Lo`001@000000W?l009co002Lo`00000000<0??l000D00000003o>@00ocT00?li00000003003o
^00500000000cOl00<go003=o`0000000P00Hol01@00000001Go000Eo`005Ol0000000<04`3o00D0
000001<0o`0C0?l04`3o00000003000Eo`0500000000Hol006?o001So`0000000P00cOl01@000000
0?nh003o^000okP0000000<00?li00D0000003co000lo`00??l00000000309co00050000003Lo`00
g?l00=co000000000P3go`001@000000jOl00>Wo003Yo`00000000<0]Ol000D00000067o001Qo`00
HOl000000003003o2@0500000000oh000?n0003oP00000000P00ooH01@0000000:7o002Qo`00XOl0
000000<005;o00D00000000Ro`008_l002;o00000003000Fo`0500000000;_l002ko000^o`000000
0P00J?l01@0000000;co002lo`00_?l0000000<00?oN00D00000003oM000og@00?md00000003003o
400500000014o`00A?l004Co000000000P21o`001@000000X?l00:3o002Po`00000000<0Wol000D0
0000083o0020o`00P?l00000000304Oo000500000000o`D00?l5003o1@0000000P00oe`01@000000
0?nd003o]000ok@0000000<00?co00D00000002lo`00_?l00;co00000003002Bo`0500000000P_l0
08;o0022o`0000000P00Rol01@0000000:co002/o`00[?l0000000<00=oo00D00000003oh000on00
0?oP00000003003oVP0400000000oeL00?mG000000<00?lL00D0000000ko000>o`003_l000000003
02So00050000000^o`00;_l002ko000000000`0Qo`000`000000oooo0?ooo`040?ooo`4000000@3o
ool00140oooo00H0omnL04Qd_`3ooml0W4P00018W03Oool90?ooo`030000003oool0oooo00D0oooo
00@00000027o000Qo`0000000`0^o`001@000000:?l002So000Xo`00000000<03_l000D00000003o
7000oa`00?lL00000003003oE`0400000000oiX00?nJ000000<00?oP00D00000003Oo`00gol00=oo
00000003002/o`0500000000Rol008_o002;o`0000000`00P_l01000000009;o002Bo`000003002l
o`0500000000o?l00?co003lo`0000000`00ok@01@0000000?mL003oG000oe`0000000<00?l500@0
000004Oo0017o`0000000`20o`001@000000Wol009oo002Oo`00000000<0X?l000D00000087o0021
o`00POl00000000304Co000400000000oa000?l@000000<00?md00D00000003ogP00omh00?oN0000
0003002lo`0500000000J?l006So001Xo`0000000`00;_l01000000001Ko000Fo`000003000Ro`05
00000000D_l005;o001Bo`0000000`00XOl01@0000000?of003omP00ooH0000000<00?n000@00000
003o2@00o`T000000`1Qo`001@000000]Ol00;Go002eo`00000000<0jOl000D000000?Oo003go`00
mol0000000020=co00050000002Lo`00W?l009co000000000`0lo`001@0000000?li003o>@00ocT0
000000<00?nh00D00000003=o`00cOl00<go00000002001So`05000000005Ol001Go000Eo`000000
0`0C0?l01@0000004`3o01<0o`0C0?l0000000<001Go00D00000001So`00Hol006?o00000002003=
o`0500000000okP00?nh003o^00000000`00ocT01@000000??l003co000lo`00000000<0W?l000D0
00000=co003Lo`00g?l0000000020?Oo00050000003Yo`00jOl00>Wo000000000`2eo`001@000000
HOl0067o001Qo`00000000<00?l900D00000003oP000oh000?n000000002003omP0500000000XOl0
0:7o002Qo`0000000`00D_l01@00000002;o000Ro`008_l0000000<001Ko00D00000000^o`00;_l0
02ko00000002001Xo`0500000000_?l00;co002lo`0000000`00omh01@0000000?md003oM000og@0
000000<00?l@00D0000004Co0014o`00A?l000000002087o00050000002Po`00X?l00:3o00000000
0`2Oo`001@000000P?l0083o0020o`00000000<0Aol000D00000003o1@00o`D00?l500000002003o
G00500000000ok@00?nd003o]00000000`00o?l01@0000000;co002lo`00_?l0000000<009;o00D0
00000022o`00P_l008;o00000002002;o`0500000000[?l00:co002/o`0000000`00gol01@000000
0?oP003oh000on00000000<00?nJ00@00000003oE`00oeL000000`00oa`01@0000003_l000ko000>
o`00000000<0:?l000D0000002ko000^o`00;_l000000003027o00030000003oool0oooo00@0oooo
0@0000010?ooo`004P3oool01@3OW4P0M;oO09a80000B9`0gooo00T0oooo00<000000?ooo`3oool0
1@3oool0100000008Ol0027o0000000302ko00050000000Xo`00:?l002So000000000`0>o`001@00
00000?lL003o7000oa`0000000<00?mG00@00000003oVP00oiX000000`00on001@0000000=oo003O
o`00gol0000000<00:co00D00000002;o`00Rol008_o000000030022o`0400000000T_l009;o0000
00<00;co00D00000003lo`00o?l00?co00000003003o]00500000000oe`00?mL003oG00000000`00
o`D010000000Aol004Oo00000003083o00050000002Oo`00Wol009oo000000000`2Po`001@000000
POl0087o0021o`00000000<0A?l000@00000003o4000oa0000000`00og@01@0000000?oN003ogP00
omh0000000<00;co00D00000001Xo`00J?l006So00000003000^o`04000000005_l001Ko000000<0
02;o00D00000001Bo`00D_l005;o00000003002Qo`0500000000ooH00?of003omP0000000`00oh00
100000000?l9003o2@000003067o00050000002eo`00]Ol00;Go000000000`3Yo`001@000000mol0
0?Oo003go`0000000080g?l000D0000009co002Lo`00W?l00000000303co000500000000ocT00?li
003o>@0000000`00okP01@0000000<go003=o`00cOl00000008006?o00D00000000Eo`005Ol001Go
0000000301<0o`050000000C0?l04`3o01<0o`0000000`005Ol01@00000006?o001So`00Hol00000
00800<go00D00000003o^000okP00?nh00000003003o>@050000000lo`00??l003co000000000`2L
o`001@000000g?l00=co003Lo`0000000080mol000D000000>Wo003Yo`00jOl0000000030;Go0005
0000001Qo`00HOl0067o000000000`00o`T01@0000000?n0003oP000oh00000000800?of00D00000
002Qo`00XOl00:7o00000003001Bo`05000000008_l002;o000Ro`0000000`005_l01@00000002ko
000^o`00;_l00000008006So00D00000002lo`00_?l00;co00000003003ogP0500000000og@00?md
003oM00000000`00oa001@000000A?l004Co0014o`0000000080POl000D000000:3o002Po`00X?l0
0000000309oo000500000020o`00P?l0083o000000000`17o`001@0000000?l5003o1@00o`D00000
00800?mL00D00000003o]000ok@00?nd00000003003lo`0500000000_?l00;co002lo`0000000`00
T_l01@00000008;o0022o`00P_l00000008008_o00D00000002/o`00[?l00:co00000003003Oo`05
00000000on000?oP003oh00000000`00oiX0100000000?mG003oE`000003003o70050000000>o`00
3_l000ko000000000`0Xo`001@000000;_l002ko000^o`00000000<08Ol000<000000?ooo`3oool0
103oool100000040oooo000B0?ooo`050?oog`2LB4P0B4P00018W03Oool02@3oool00`000000oooo
0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool001<0oooo00@0omnL04P00000B9`0gooo
2@3oool00`000000oooo0?ooo`050?ooo`040000000^o`00;_l0000000<0:Ol000D00000013o000@
o`004?l000000003003o6@0500000000oe400?mA003oD@0000000`00oi<0100000000?oH003of000
0003003Vo`0500000000/Ol00;7o002ao`0000000`00SOl01@00000007oo001oo`00Ool0000000<0
08Wo00@00000002]o`00[Ol000000`00iol01@0000000?o<003oc000ol`0000000<00?mf00D00000
003o7P00oah00?lN00000003037o00040000001co`00Lol0000000<0W?l000D000000:[o002Zo`00
Z_l00000000309So00050000001Xo`00J?l006So000000000`0Mo`00100000000?m1003o@@000003
003oZP0500000000k?l00>co003/o`0000000`00SOl01@00000004Co0014o`00A?l0000000<001Oo
00@00000000<o`003?l000000`009_l01@00000006?o001So`00Hol0000000<00;go00D00000003o
e000om@00?oD00000003003oG0040000000Io`006Ol0000000<0POl000D000000<oo003?o`00col0
000000030?_o00050000003ooP00ooh00?on000000000P3Oo`001@000000V?l009So002Ho`000000
00<0<ol000D00000003oA000od@00?m400000003003oa@0500000000_ol00;oo002oo`0000000P00
EOl01@00000000So0008o`002?l0000000<0803o00D000000200o`0P0?l0803o000000030008o`05
00000000EOl005Go001Eo`0000000P00_ol01@0000000?o5003oa@00olD0000000<00?m400D00000
03?o000co`00<ol00000000309So00050000003Oo`00gol00=oo000000000P3ooP001@000000nol0
0?_o003ko`00000000<0col000D00000087o0021o`00POl00000000301Wo000500000000oe`00?mL
003oG00000000P00om@01@0000000;go002mo`00_Ol0000000<006?o00D00000000Vo`009_l002Ko
00000003000<o`05000000005ol001Oo000Go`0000000P00A?l01@00000008go002=o`00SOl00000
00<00>co00D00000003oZP00ojX00?nZ00000003003o@@050000000Mo`007Ol001go000000000P1X
o`001@000000V?l009So002Ho`00000000<0Z_l000D0000009co002Lo`00W?l00000000307?o0005
0000000ao`00<Ol0037o000000000P00oah01@0000000?mf003oMP00ogH0000000<00?o<00D00000
003Wo`00iol00>Oo00000003002]o`0500000000ROl008Wo0029o`0000000P00Ool01@00000008go
002=o`00SOl0000000<00;7o00D00000003Vo`00i_l00>Ko00000003003of00400000000oi<00?nC
000000<00?mA00D00000003o6@00oaT00?lI00000003013o00050000000Yo`00:Ol002Wo00000000
0`0^o`000`000000oooo0?ooo`040?ooo`4000000@3oool001@0oooo00<0gia80018W03Oool02@3o
ool00`000000oooo0?ooo`050?ooo`040000000^o`00;_l0000000<0:Ol000D00000013o000@o`00
4?l000000003003o6@0500000000oe400?mA003oD@0000000`00oi<0100000000?oH003of0000003
003Vo`0500000000/Ol00;7o002ao`0000000`00SOl01@00000007oo001oo`00Ool0000000<008Wo
00@00000002]o`00[Ol000000`00iol01@0000000?o<003oc000ol`0000000<00?mf00D00000003o
7P00oah00?lN00000003037o00040000001co`00Lol0000000<0W?l000D000000:[o002Zo`00Z_l0
0000000309So00050000001Xo`00J?l006So000000000`0Mo`00100000000?m1003o@@000003003o
ZP0500000000k?l00>co003/o`0000000`00SOl01@00000004Co0014o`00A?l0000000<001Oo00@0
0000000<o`003?l000000`009_l01@00000006?o001So`00Hol0000000<00;go00D00000003oe000
om@00?oD00000003003oG0040000000Io`006Ol0000000<0POl000D000000<oo003?o`00col00000
00030?_o00050000003ooP00ooh00?on000000000P3Oo`001@000000V?l009So002Ho`00000000<0
<ol000D00000003oA000od@00?m400000003003oa@0500000000_ol00;oo002oo`0000000P00EOl0
1@00000000So0008o`002?l0000000<0803o00D000000200o`0P0?l0803o000000030008o`050000
0000EOl005Go001Eo`0000000P00_ol01@0000000?o5003oa@00olD0000000<00?m400D0000003?o
000co`00<ol00000000309So00050000003Oo`00gol00=oo000000000P3ooP001@000000nol00?_o
003ko`00000000<0col000D00000087o0021o`00POl00000000301Wo000500000000oe`00?mL003o
G00000000P00om@01@0000000;go002mo`00_Ol0000000<006?o00D00000000Vo`009_l002Ko0000
0003000<o`05000000005ol001Oo000Go`0000000P00A?l01@00000008go002=o`00SOl0000000<0
0>co00D00000003oZP00ojX00?nZ00000003003o@@050000000Mo`007Ol001go000000000P1Xo`00
1@000000V?l009So002Ho`00000000<0Z_l000D0000009co002Lo`00W?l00000000307?o00050000
000ao`00<Ol0037o000000000P00oah01@0000000?mf003oMP00ogH0000000<00?o<00D00000003W
o`00iol00>Oo00000003002]o`0500000000ROl008Wo0029o`0000000P00Ool01@00000008go002=
o`00SOl0000000<00;7o00D00000003Vo`00i_l00>Ko00000003003of00400000000oi<00?nC0000
00<00?mA00D00000003o6@00oaT00?lI00000003013o00050000000Yo`00:Ol002Wo000000000`0^
o`000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo
0000084000001P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@0000002So
000Xo`0000000`0@o`001@0000000?lH003o6000oaP0000000<00?m?00D00000003oS`00ohl00?n?
00000003003od`0400000000k?l00>co000000<00;Go00D00000002?o`00Sol008oo00000003001m
o`0500000000Pol008?o0023o`0000000`00XOl0100000000=Ko003Fo`000003003ohP0500000000
ohh00?n>003oSP0000000`00ocH01@0000007?l001co000Lo`00000000<0Hol000@0000009Ko002F
o`0000000`2^o`001@000000ZOl00:Wo002Yo`00000000<0QOl000D0000004Go0015o`00AOl00000
0003003o4@0400000000ogL00?mg000000<00?oR00D00000002ho`00^?l00;So00000003001Qo`05
000000008ol002?o000So`0000000`001Ol01000000000Wo0009o`000003000ao`0500000000N_l0
07[o001jo`0000000`00g?l01@0000000?n_003o[`00ojl0000000<00?le00@0000003ko000no`00
00000`2Po`001@000000iol00>Oo003Wo`00000000<0oo<000D000000?of003omP00ooH000000002
0>3o00050000002Co`00Tol009?o000000000`0Zo`001@0000000?mA003oD@00oe40000000<00?oC
00D00000002`o`00/?l00;3o000000020017o`05000000040?l0103o00@0o`0000000`0]0?l01@00
0000;@3o02d0o`0]0?l0000000<0103o00D000000017o`00Aol004Oo00000002002`o`0500000000
om<00?oC003od`0000000`00oe401@000000:_l002[o000Zo`00000000<0Tol000D000000>3o003P
o`00h?l0000000020?of00050000003ol`00oo<00?oc000000000`3Wo`001@000000X?l00:3o002P
o`00000000<0?_l000D00000003o=@00ocD00?le00000002003o[`0500000000g?l00=co003Lo`00
00000`00N_l01@000000037o000ao`00<Ol0000000<000Wo00D000000005o`001Ol000Go00000002
000So`0500000000HOl0067o001Qo`0000000`00^?l01@0000000?oR003ohP00on80000000<00?mg
00D00000003o4@00oa400?lA0000000204Go000500000025o`00QOl008Go000000000`2Yo`001@00
0000[_l00:ko002^o`00000000<0U_l000D0000006?o001So`00Hol00000000201co000500000000
ocH00?lf003o=P0000000`00ohh01@0000000?oR003ohP00on80000000<00=Ko00D00000002Qo`00
XOl00:7o000000020023o`0500000000OOl007go001mo`0000000`00Sol01@0000000;Go002eo`00
]Ol0000000<00>co00@00000003od`00om<000000`00ohl01@0000000?m?003oC`00odl0000000<0
0?lH00D00000013o000@o`004?l00000000302So00030000003oool0oooo00@0oooo0@0000010?oo
o`00803oool00`000000oooo0?ooo`050?ooo`040000000Xo`00:?l0000000<04?l000D00000003o
6000oaP00?lH00000003003oC`0500000000ohl00?n?003oS`0000000`00om<0100000000>co003/
o`000003002eo`0500000000Sol008oo002?o`0000000`00OOl01@00000008?o0023o`00Pol00000
00<00:7o00@00000003Fo`00e_l000000`00on801@0000000?n>003oSP00ohh0000000<00?lf00D0
000001co000Lo`007?l00000000306?o00040000002Fo`00U_l0000000<0[_l000D000000:Wo002Y
o`00ZOl00000000308Go000500000015o`00AOl004Go000000000`00oa40100000000?mg003oM`00
0003003ohP0500000000^?l00;So002ho`0000000`00HOl01@00000002?o000So`008ol0000000<0
00Go00@000000009o`002Ol000000`00<Ol01@00000007[o001jo`00N_l0000000<00=co00D00000
003o[`00ojl00?n_00000003003o=@040000000no`00?_l0000000<0X?l000D000000>Oo003Wo`00
iol0000000030?oc00050000003omP00ooH00?of000000000P3Po`001@000000Tol009?o002Co`00
000000<0:_l000D00000003oD@00oe400?mA00000003003od`0500000000/?l00;3o002`o`000000
0P00Aol01@000000103o00@0o`040?l0000000<0;@3o00D0000002d0o`0]0?l0;@3o0000000300@0
o`0500000000Aol004Oo0017o`0000000P00/?l01@0000000?oC003od`00om<0000000<00?mA00D0
000002[o000Zo`00:_l00000000309?o00050000003Po`00h?l00>3o000000000P3omP001@000000
oo<00?oc003ol`00000000<0iol000D000000:3o002Po`00X?l00000000303ko000500000000ocD0
0?le003o=@0000000P00ojl01@0000000=co003Lo`00g?l0000000<007[o00D00000000ao`00<Ol0
037o000000030009o`05000000001Ol000Go0005o`0000000P008ol01@000000067o001Qo`00HOl0
000000<00;So00D00000003ohP00on800?oR00000003003oM`0500000000oa400?lA003o4@000000
0P15o`001@000000QOl008Go0025o`00000000<0ZOl000D000000:ko002^o`00[_l00000000309Ko
00050000001So`00Hol006?o000000000P0Lo`001@0000000?lf003o=P00ocH0000000<00?n>00D0
0000003ohP00on800?oR00000003003Fo`0500000000XOl00:7o002Qo`0000000P00Pol01@000000
07go001mo`00OOl0000000<008oo00D00000002eo`00]Ol00;Go00000003003/o`0400000000om<0
0?oC000000<00?n?00D00000003oC`00odl00?m?00000003003o60050000000@o`004?l0013o0000
00000`0Xo`000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool0
1@3oool010000000:?l002So00000003013o000500000000oaP00?lH003o600000000`00odl01@00
00000?n?003oS`00ohl0000000<00?oC00@00000003/o`00k?l000000`00]Ol01@00000008oo002?
o`00Sol0000000<007go00D000000023o`00Pol008?o00000003002Qo`0400000000e_l00=Ko0000
00<00?oR00D00000003oSP00ohh00?n>00000003003o=P050000000Lo`007?l001co000000000`1S
o`0010000000U_l009Ko000000030:ko00050000002Yo`00ZOl00:Wo000000000`25o`001@000000
AOl004Go0015o`00000000<00?lA00@00000003oM`00ogL000000`00on801@0000000;So002ho`00
^?l0000000<0067o00D00000000So`008ol002?o000000030005o`04000000002Ol000Wo000000<0
037o00D00000001jo`00N_l007[o00000003003Lo`0500000000ojl00?n_003o[`0000000`00ocD0
10000000?_l003ko000000030:3o00050000003Wo`00iol00>Oo000000000`3ol`001@000000ooH0
0?of003omP0000000080h?l000D0000009?o002Co`00Tol00000000302[o000500000000oe400?mA
003oD@0000000`00om<01@0000000;3o002`o`00/?l00000008004Oo00D0000000@0o`040?l0103o
0000000302d0o`050000000]0?l0;@3o02d0o`0000000`040?l01@00000004Oo0017o`00Aol00000
00800;3o00D00000003od`00om<00?oC00000003003oD@050000000Zo`00:_l002[o000000000`2C
o`001@000000h?l00>3o003Po`0000000080ooH000D000000?oc003ol`00oo<0000000030>Oo0005
0000002Po`00X?l00:3o000000000`0no`001@0000000?le003o=@00ocD0000000800?n_00D00000
003Lo`00g?l00=co00000003001jo`0500000000<Ol0037o000ao`0000000`002Ol01@00000000Go
0005o`001Ol00000008002?o00D00000001Qo`00HOl0067o00000003002ho`0500000000on800?oR
003ohP0000000`00ogL01@0000000?lA003o4@00oa4000000080AOl000D0000008Go0025o`00QOl0
000000030:Wo00050000002^o`00[_l00:ko000000000`2Fo`001@000000Hol006?o001So`000000
00807?l000D00000003o=P00ocH00?lf00000003003oSP0500000000on800?oR003ohP0000000`00
e_l01@0000000:7o002Qo`00XOl00000008008?o00D00000001mo`00OOl007go00000003002?o`05
00000000]Ol00;Go002eo`0000000`00k?l0100000000?oC003od`000003003oS`0500000000odl0
0?m?003oC`0000000`00oaP01@0000004?l0013o000@o`00000000<0:?l000<000000?ooo`3oool0
103oool100000040oooo000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@00
00010?ooo`00803oool00`000000oooo0?ooo`050?ooo`040000000>o`003_l0000000<00?lI00D0
0000003oC`00odl00?m?00000003003oSP0500000000om000?o@003od00000000`00kol010000000
0;Wo002io`000003002Ao`0500000000O?l007co001lo`0000000`00Ool01@00000009So002Ho`00
V?l0000000<00<Wo00@00000003ol`00oo<000000`00oj801@0000000?m;003oB`00od/0000000<0
2Ol000D0000005Co001Do`00E?l00000000308go00040000002_o`00[ol0000000<0]?l000D00000
09_o002Ko`00Vol00000000306Ko00050000000Ho`006?l001So000000000`00odL0100000000?na
003o/@000003003To`0500000000QOl008Go0025o`0000000`00>Ol01@00000000So0008o`002?l0
000000<01P3o00@00000000<o`003?l000000`00@_l01@00000009Go002Eo`00UOl0000000<00?oo
00D00000003oQ`00ohL00?n700000003003o3@040000001Ro`00H_l0000000<0_ol000D000000?oo
003oo`00ool0000000030?oT00050000003ok`00onl00?o_000000000P3Oo`001@000000S?l008co
002<o`00000000<07_l000D00000003oG`00oel00?mO00000003003ohP0500000000XOl00:7o002Q
o`0000000P00>Ol01@0000004P3o0180o`0B0?l0000000<0>P3o00D0000003X0o`0j0?l0>P3o0000
00030180o`0500000000>Ol003Wo000io`0000000P00XOl01@0000000?oR003ohP00on80000000<0
0?mO00D0000001ko000No`007_l00000000308co00050000003Oo`00gol00=oo000000000P3ok`00
1@000000on@00?oT003oi000000000<0ool000D000000;oo002oo`00_ol00000000306;o00050000
0000o`d00?l=003o3@0000000P00ohL01@0000000?oo003oo`00ool0000000<009Go00D000000012
o`00@_l004;o00000003000<o`05000000060?l01P3o00H0o`0000000P002?l01@00000003Wo000i
o`00>Ol0000000<008Go00D00000003To`00i?l00>Co00000003003o/@0500000000odL00?m7003o
A`0000000P0Ho`001@000000I_l006Ko001Vo`00000000<0Vol000D000000;Co002do`00]?l00000
00030:oo00050000002=o`00SOl008go000000000P1Do`001@0000002Ol000Wo0009o`00000000<0
0?m;00D00000003oXP00oj800?nR00000003003ol`0500000000bOl00<Wo0039o`0000000P00V?l0
1@00000007oo001oo`00Ool0000000<007co00D00000002Ao`00TOl0097o00000003002io`040000
0000kol00>oo000000<00?o@00D00000003oSP00ohh00?n>00000003003oC`0500000000oaT00?lI
003o6@0000000`0>o`000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?oo
o`3oool01@3oool0100000003_l000ko00000003003o6@0500000000odl00?m?003oC`0000000`00
ohh01@0000000?o@003od000om00000000<00>oo00@00000002io`00^Ol000000`00TOl01@000000
07co001lo`00O?l0000000<007oo00D00000002Ho`00V?l009So000000030039o`0400000000oo<0
0?oc000000<00?nR00D00000003oB`00od/00?m;0000000300Wo00050000001Do`00E?l005Co0000
00000`2=o`0010000000[ol00:oo000000030;Co00050000002Ko`00Vol009_o000000000`1Vo`00
1@0000006?l001So000Ho`00000000<00?m700@00000003o/@00ok4000000`00i?l01@00000008Go
0025o`00QOl0000000<003Wo00D000000008o`002?l000So0000000300H0o`04000000003?l000co
000000<004;o00D00000002Eo`00UOl009Go00000003003oo`0500000000ohL00?n7003oQ`000000
0`00o`d010000000H_l006;o000000030;oo00050000003oo`00ool00?oo000000000`3oi0001@00
0000onl00?o_003ok`0000000080gol000D0000008co002<o`00S?l00000000301ko000500000000
oel00?mO003oG`0000000`00on801@0000000:7o002Qo`00XOl00000008003Wo00D000000180o`0B
0?l04P3o0000000303X0o`050000000j0?l0>P3o03X0o`0000000`0B0?l01@00000003Wo000io`00
>Ol0000000800:7o00D00000003ohP00on800?oR00000003003oG`050000000No`007_l001ko0000
00000`2<o`001@000000gol00=oo003Oo`0000000080onl000D000000?oT003oi000on@000000003
0?oo00050000002oo`00_ol00;oo000000000`1Ro`001@0000000?l=003o3@00o`d0000000800?n7
00D00000003oo`00ool00?oo00000003002Eo`0500000000@_l004;o0012o`0000000`003?l01@00
00001P3o00H0o`060?l00000008000So00D00000000io`00>Ol003Wo000000030025o`0500000000
i?l00>Co003To`0000000`00ok401@0000000?m7003oA`00odL0000000806?l000D0000006Ko001V
o`00I_l00000000309_o00050000002do`00]?l00;Co000000000`2_o`001@000000SOl008go002=
o`0000000080E?l000D0000000Wo0009o`002Ol000000003003oB`0500000000oj800?nR003oXP00
00000`00oo<01@0000000<Wo0039o`00bOl00000008009So00D00000001oo`00Ool007oo00000003
001lo`0500000000TOl0097o002Ao`0000000`00^Ol0100000000>oo003_o`000003003od0050000
0000ohh00?n>003oSP0000000`00odl01@0000000?lI003o6@00oaT0000000<03_l000<000000?oo
o`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@0000000ko000>
o`0000000`00oaT01@0000000?m?003oC`00odl0000000<00?n>00D00000003od000om000?o@0000
0003003_o`0400000000^Ol00;Wo000000<0097o00D00000001lo`00O?l007co00000003001oo`05
00000000V?l009So002Ho`0000000`00bOl0100000000?oc003ol`000003003oXP0500000000od/0
0?m;003oB`0000000`09o`001@000000E?l005Co001Do`00000000<0SOl000@000000:oo002_o`00
00000`2do`001@000000Vol009_o002Ko`00000000<0I_l000D0000001So000Ho`006?l000000003
003oA`0400000000ok400?na000000<00>Co00D000000025o`00QOl008Go00000003000io`050000
00002?l000So0008o`0000000`060?l01000000000co000<o`0000030012o`0500000000UOl009Go
002Eo`0000000`00ool01@0000000?n7003oQ`00ohL0000000<00?l=00@0000006;o001Ro`000000
0`2oo`001@000000ool00?oo003oo`00000000<0on@000D000000?o_003ok`00onl0000000020=oo
00050000002<o`00S?l008co000000000`0No`001@0000000?mO003oG`00oel0000000<00?oR00D0
0000002Qo`00XOl00:7o00000002000io`050000000B0?l04P3o0180o`0000000`0j0?l01@000000
>P3o03X0o`0j0?l0000000<04P3o00D00000000io`00>Ol003Wo00000002002Qo`0500000000on80
0?oR003ohP0000000`00oel01@0000007_l001ko000No`00000000<0S?l000D000000=oo003Oo`00
gol0000000020?o_00050000003oi000on@00?oT000000000`3oo`001@000000_ol00;oo002oo`00
000000<0H_l000D00000003o3@00o`d00?l=00000002003oQ`0500000000ool00?oo003oo`000000
0`00UOl01@00000004;o0012o`00@_l0000000<000co00D0000000H0o`060?l01P3o000000020008
o`0500000000>Ol003Wo000io`0000000`00QOl01@0000000>Co003To`00i?l0000000<00?na00D0
0000003oA`00odL00?m70000000201So00050000001Vo`00I_l006Ko000000000`2Ko`001@000000
]?l00;Co002do`00000000<0[ol000D0000008go002=o`00SOl00000000205Co000500000009o`00
2Ol000Wo000000000`00od/01@0000000?nR003oXP00oj80000000<00?oc00D000000039o`00bOl0
0<Wo00000002002Ho`0500000000Ool007oo001oo`0000000`00O?l01@000000097o002Ao`00TOl0
000000<00;Wo00@00000003_o`00kol000000`00om001@0000000?n>003oSP00ohh0000000<00?m?
00D00000003o6@00oaT00?lI0000000300ko00030000003oool0oooo00@0oooo0@0000010?ooo`00
803oool00`000000oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<0
00000?ooo`3oool01@3oool0100000000?lL003o70000003003oD@0500000000ohl00?n?003oS`00
00000`00om001@0000000?3o003`o`00l?l0000000<00;[o00@00000002Bo`00T_l000000`00O?l0
1@00000007co001lo`00O?l0000000<009;o00D00000002oo`00_ol00;oo00000003003no`040000
0000ok<00?nc000000<00?mN00D00000003o2000o`P00?l80000000304Go000500000023o`00Pol0
08?o000000000`2/o`0010000000^ol00;_o000000030:co000500000021o`00POl0087o00000000
0`0lo`001@0000000?lL003o7000oa`0000000<00?n300@00000003okP00onh000000`00[?l01@00
000005Go001Eo`00EOl0000000<001Ko00D0000000/0o`0;0?l02`3o0000000300/0o`0400000000
5_l001Ko000000<005Wo00D00000002fo`00]_l00;Ko00000003003of00500000000oed00?mM003o
G@0000000`0Ko`0010000000Qol008Oo000000030=go00050000003ojP00onX00?oZ000000000`3o
e`001@000000onX00?oZ003ojP0000000080gOl000D0000008Co0024o`00Q?l00000000301;o0005
00000000ofd00?m]003oK@0000000`00oo401@00000009;o002Bo`00T_l00000008002[o00D00000
0200o`0P0?l0803o0000000304L0o`05000000170?l0A`3o04L0o`0000000`0P0?l01@00000002[o
000Zo`00:_l00000008009;o00D00000003ol@00oo400?oa00000003003oK@050000000Bo`004_l0
01;o000000000`24o`001@000000gOl00=go003Mo`0000000080onX000D000000?oG003oe`00omL0
000000030?oZ00050000003Mo`00gOl00=go000000000`27o`001@0000006ol001_o000Ko`000000
00800?mM00D00000003of000omP00?oH00000003002fo`0500000000FOl005Wo001Io`0000000`00
5_l01@0000002`3o00/0o`0;0?l0000000802`3o00D00000000Fo`005_l001Ko00000003001Eo`05
00000000[?l00:co002/o`0000000`00onh01@0000000?n3003oP`00oh<0000000800?lL00D00000
03co000lo`00??l000000003087o00050000002/o`00[?l00:co000000000`2ko`001@000000[?l0
0:co002/o`0000000080Pol000D0000004Go0015o`00AOl000000003003o200500000000oeh00?mN
003oGP0000000`00ok<01@0000000?ko003no`00o_l0000000800;oo00D00000002Bo`00T_l009;o
00000003001lo`0500000000O?l007co001lo`0000000`00T_l0100000000;[o002jo`000003003`
o`0500000000om000?o@003od00000000`00ohl01@0000000?mA003oD@00oe40000000<00?lL00<0
00000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000
003o7000oa`000000`00oe401@0000000?n?003oS`00ohl0000000<00?o@00D00000003`o`00l?l0
0?3o00000003002jo`0400000000T_l009;o000000<007co00D00000001lo`00O?l007co00000003
002Bo`0500000000_ol00;oo002oo`0000000`00o_l0100000000?nc003o/`000003003oGP050000
0000o`P00?l8003o200000000`15o`001@000000Pol008?o0023o`00000000<0[?l000@000000;_o
002ko`0000000`2/o`001@000000POl0087o0021o`00000000<0??l000D00000003o7000oa`00?lL
00000003003oP`0400000000onh00?o^000000<00:co00D00000001Eo`00EOl005Go00000003000F
o`050000000;0?l02`3o00/0o`0000000`0;0?l01000000001Ko000Fo`000003001Io`0500000000
]_l00;Ko002fo`0000000`00omP01@0000000?mM003oG@00oed0000000<06ol000@0000008Oo0027
o`0000000`3Mo`001@000000onX00?oZ003ojP00000000<0omL000D000000?oZ003ojP00onX00000
00020=go000500000024o`00Q?l008Co000000000`0Bo`001@0000000?m]003oK@00ofd0000000<0
0?oa00D00000002Bo`00T_l009;o00000002000Zo`050000000P0?l0803o0200o`0000000`170?l0
1@000000A`3o04L0o`170?l0000000<0803o00D00000000Zo`00:_l002[o00000002002Bo`050000
0000oo400?oa003ol@0000000`00ofd01@0000004_l001;o000Bo`00000000<0Q?l000D000000=go
003Mo`00gOl0000000020?oZ00050000003oe`00omL00?oG000000000`3ojP001@000000gOl00=go
003Mo`00000000<0Qol000D0000001_o000Ko`006ol000000002003oG@0500000000omP00?oH003o
f00000000`00]_l01@00000005Wo001Io`00FOl0000000<001Ko00D0000000/0o`0;0?l02`3o0000
000200/0o`05000000005_l001Ko000Fo`0000000`00EOl01@0000000:co002/o`00[?l0000000<0
0?o^00D00000003oP`00oh<00?n300000002003o70050000000lo`00??l003co000000000`21o`00
1@000000[?l00:co002/o`00000000<0^ol000D000000:co002/o`00[?l00000000208?o00050000
0015o`00AOl004Go000000000`00o`P01@0000000?mN003oGP00oeh0000000<00?nc00D00000003n
o`00o_l00?ko00000002002oo`0500000000T_l009;o002Bo`0000000`00O?l01@00000007co001l
o`00O?l0000000<009;o00@00000002jo`00^_l000000`00l?l01@0000000?o@003od000om000000
00<00?n?00D00000003oD@00oe400?mA00000003003o70030000003oool0oooo00@0oooo0@000001
0?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000oa`00?lL000000<00?mA00D00000
003oS`00ohl00?n?00000003003od00500000000l?l00?3o003`o`0000000`00^_l01000000009;o
002Bo`000003001lo`0500000000O?l007co001lo`0000000`00T_l01@0000000;oo002oo`00_ol0
000000<00?ko00@00000003o/`00ok<000000`00oeh01@0000000?l8003o2000o`P0000000<0AOl0
00D0000008?o0023o`00Pol0000000030:co00040000002ko`00^ol0000000<0[?l000D00000087o
0021o`00POl00000000303co000500000000oa`00?lL003o700000000`00oh<0100000000?o^003o
kP000003002/o`0500000000EOl005Go001Eo`0000000`005_l01@0000002`3o00/0o`0;0?l00000
00<02`3o00@00000000Fo`005_l000000`00FOl01@0000000;Ko002fo`00]_l0000000<00?oH00D0
0000003oG@00oed00?mM0000000301_o000400000027o`00Qol0000000<0gOl000D000000?oZ003o
jP00onX0000000030?oG00050000003ojP00onX00?oZ000000000P3Mo`001@000000Q?l008Co0024
o`00000000<04_l000D00000003oK@00ofd00?m]00000003003ol@0500000000T_l009;o002Bo`00
00000P00:_l01@000000803o0200o`0P0?l0000000<0A`3o00D0000004L0o`170?l0A`3o00000003
0200o`0500000000:_l002[o000Zo`0000000P00T_l01@0000000?oa003ol@00oo40000000<00?m]
00D0000001;o000Bo`004_l00000000308Co00050000003Mo`00gOl00=go000000000P3ojP001@00
0000omL00?oG003oe`00000000<0onX000D000000=go003Mo`00gOl00000000308Oo00050000000K
o`006ol001_o000000000P00oed01@0000000?oH003of000omP0000000<00;Ko00D00000001Io`00
FOl005Wo00000003000Fo`050000000;0?l02`3o00/0o`0000000P0;0?l01@00000001Ko000Fo`00
5_l0000000<005Go00D00000002/o`00[?l00:co00000003003okP0500000000oh<00?n3003oP`00
00000P00oa`01@000000??l003co000lo`00000000<0POl000D000000:co002/o`00[?l000000003
0;_o00050000002/o`00[?l00:co000000000P23o`001@000000AOl004Go0015o`00000000<00?l8
00D00000003oGP00oeh00?mN00000003003o/`0500000000o_l00?ko003no`0000000P00_ol01@00
000009;o002Bo`00T_l0000000<007co00D00000001lo`00O?l007co00000003002Bo`0400000000
^_l00;[o000000<00?3o00D00000003od000om000?o@00000003003oS`0500000000oe400?mA003o
D@0000000`00oa`00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3o
ool01@3ooooo0000084000001P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo
00@00000003oE`00oeL000000`00oi<01@0000000?oC003od`00om<0000000<00>oo00D00000002j
o`00^_l00;[o00000003002Bo`0400000000O?l007co000000<007[o00D00000002>o`00S_l008ko
00000003002ho`0500000000m?l00?Co003do`0000000`00ol00100000000?m]003oK@000003003o
60050000000go`00=ol003Oo000000000`1io`001@000000Z?l00:So002Xo`00000000<0__l000@0
00000;Wo002io`0000000`2Go`001@000000Fol005_o001Ko`00000000<02Ol000D00000003oF000
oeP00?mH00000003003o`P0400000000e?l00=Co000000<007Ko00D00000000Zo`00:_l002[o0000
000300H0o`050000000I0?l06@3o01T0o`0000000`090?l01000000002Oo000Wo`000003001fo`05
00000000g?l00=co003Lo`0000000`00ojd01@0000000?la003o<@00oc40000000<0AOl000@00000
0:co002/o`0000000`3jo`001@000000omH00?oF003oeP00000000<0ol`000D000000?oW003oi`00
onL0000000020=So00050000001jo`00N_l007[o000000000`04o`001@0000000?mm003oO@00ogd0
000000<00?go00D000000022o`00P_l008;o00000002000Ko`050000000^0?l0;P3o02h0o`000000
0`1E0?l01@000000E@3o05D0o`1E0?l0000000<0;P3o00D00000000Ko`006ol001_o000000020022
o`0500000000oOl00?go003mo`0000000`00ogd01@0000001?l000Co0004o`00000000<0N_l000D0
00000=So003Ho`00f?l0000000020?oW00050000003oc000ol`00?o<000000000`3oeP001@000000
n_l00?[o003jo`00000000<0[?l000D0000004Go0015o`00AOl000000002003o<@0500000000ojd0
0?n]003o[@0000000`00g?l01@00000007Ko001fo`00M_l0000000<002Oo00D0000000T0o`090?l0
2@3o0000000201T0o`05000000060?l01P3o00H0o`0000000`00:_l01@00000007Ko001fo`00M_l0
000000<00=Co00D00000003o`P00ol800?o200000002003oF00500000009o`002Ol000Wo00000000
0`1Ko`001@000000Uol009Oo002Go`00000000<0^Ol000D000000;ko002no`00__l0000000020:So
00050000001io`00NOl007Wo000000000`0go`001@0000000?lH003o6000oaP0000000<00?m]00D0
0000003o`000ol000?o000000002003do`0500000000^?l00;So002ho`0000000`00S_l01@000000
07[o001jo`00N_l0000000<007co00@00000002Bo`00T_l000000`00^_l01@0000000>oo003_o`00
kol0000000<00?oC00D00000003oT`00oi<00?nC00000003003oE`030000003oool0oooo00@0oooo
0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000oeL00?mG000000<00?nC
00D00000003od`00om<00?oC00000003003_o`0500000000^_l00;[o002jo`0000000`00T_l01000
000007co001lo`000003001jo`0500000000S_l008ko002>o`0000000`00^?l01@0000000?Co003d
o`00m?l0000000<00?o000@00000003oK@00ofd000000`00oaP01@000000=ol003Oo000go`000000
00<0NOl000D000000:So002Xo`00Z?l0000000030;ko00040000002io`00^Ol0000000<0Uol000D0
000005_o001Ko`00Fol00000000300Wo000500000000oeP00?mH003oF00000000`00ol8010000000
0=Co003Do`000003001fo`0500000000:_l002[o000Zo`0000000`060?l01@0000006@3o01T0o`0I
0?l0000000<02@3o00@00000000Wo`009ol000000`00M_l01@0000000=co003Lo`00g?l0000000<0
0?n]00D00000003o<@00oc400?la0000000304Go00040000002/o`00[?l0000000<0n_l000D00000
0?oF003oeP00omH0000000030?o<00050000003oi`00onL00?oW000000000P3Ho`001@000000N_l0
07[o001jo`00000000<01?l000D00000003oO@00ogd00?mm00000003003mo`0500000000P_l008;o
0022o`0000000P006ol01@000000;P3o02h0o`0^0?l0000000<0E@3o00D0000005D0o`1E0?l0E@3o
0000000302h0o`05000000006ol001_o000Ko`0000000P00P_l01@0000000?go003mo`00oOl00000
00<00?mm00D0000000Co0004o`001?l00000000307[o00050000003Ho`00f?l00=So000000000P3o
i`001@000000ol`00?o<003oc000000000<0omH000D000000?[o003jo`00n_l0000000030:co0005
00000015o`00AOl004Go000000000P00oc401@0000000?n]003o[@00ojd0000000<00=co00D00000
001fo`00M_l007Ko00000003000Wo`05000000090?l02@3o00T0o`0000000P0I0?l01@0000001P3o
00H0o`060?l0000000<002[o00D00000001fo`00M_l007Ko00000003003Do`0500000000ol800?o2
003o`P0000000P00oeP01@0000002Ol000Wo0009o`00000000<0Fol000D0000009Oo002Go`00Uol0
000000030;Wo00050000002no`00__l00;ko000000000P2Xo`001@000000NOl007Wo001io`000000
00<0=ol000D00000003o6000oaP00?lH00000003003oK@0500000000ol000?o0003o`00000000P00
m?l01@0000000;So002ho`00^?l0000000<008ko00D00000001jo`00N_l007[o00000003001lo`04
00000000T_l009;o000000<00;[o00D00000003_o`00kol00>oo00000003003od`0500000000oi<0
0?nC003oT`0000000`00oeL00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<00000
0?ooo`3oool01@3oool0100000000?mG003oE`000003003oT`0500000000om<00?oC003od`000000
0`00kol01@0000000;[o002jo`00^_l0000000<009;o00@00000001lo`00O?l000000`00N_l01@00
000008ko002>o`00S_l0000000<00;So00D00000003do`00m?l00?Co00000003003o`00400000000
ofd00?m]000000<00?lH00D0000003Oo000go`00=ol00000000307Wo00050000002Xo`00Z?l00:So
000000000`2no`0010000000^Ol00;Wo0000000309Oo00050000001Ko`00Fol005_o000000000`09
o`001@0000000?mH003oF000oeP0000000<00?o200@00000003Do`00e?l000000`00M_l01@000000
02[o000Zo`00:_l0000000<01P3o00D0000001T0o`0I0?l06@3o0000000300T0o`04000000009ol0
02Oo000000<007Ko00D00000003Lo`00g?l00=co00000003003o[@0500000000oc400?la003o<@00
00000`15o`0010000000[?l00:co000000030?[o00050000003oeP00omH00?oF000000000`3oc000
1@000000onL00?oW003oi`0000000080f?l000D0000007[o001jo`00N_l00000000300Co00050000
0000ogd00?mm003oO@0000000`00oOl01@00000008;o0022o`00P_l00000008001_o00D0000002h0
o`0^0?l0;P3o0000000305D0o`050000001E0?l0E@3o05D0o`0000000`0^0?l01@00000001_o000K
o`006ol00000008008;o00D00000003mo`00oOl00?go00000003003oO@0500000004o`001?l000Co
000000000`1jo`001@000000f?l00=So003Ho`0000000080onL000D000000?o<003oc000ol`00000
00030?oF00050000003jo`00n_l00?[o000000000`2/o`001@000000AOl004Go0015o`0000000080
0?la00D00000003o[@00ojd00?n]00000003003Lo`0500000000M_l007Ko001fo`0000000`009ol0
1@0000002@3o00T0o`090?l0000000806@3o00D0000000H0o`060?l01P3o00000003000Zo`050000
0000M_l007Ko001fo`0000000`00e?l01@0000000?o2003o`P00ol80000000800?mH00D0000000Wo
0009o`002Ol00000000305_o00050000002Go`00Uol009Oo000000000`2io`001@000000__l00;ko
002no`0000000080Z?l000D0000007Wo001io`00NOl00000000303Oo000500000000oaP00?lH003o
600000000`00ofd01@0000000?o0003o`000ol00000000800?Co00D00000002ho`00^?l00;So0000
0003002>o`0500000000N_l007[o001jo`0000000`00O?l01000000009;o002Bo`000003002jo`05
00000000kol00>oo003_o`0000000`00om<01@0000000?nC003oT`00oi<0000000<00?mG00<00000
0?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0ooooo`0000210000
00H0oooo0@0000010?ooo`00803oool2000000H0oooo00@00000003oVP00oiX000000`00omP01@00
00000>co003/o`00k?l0000000<00;Wo00D00000002Bo`00T_l009;o00000003001lo`0400000000
NOl007Wo000000<008co00D00000002co`00/ol00;?o00000003003]o`0500000000olX00?o:003o
bP0000000`00ogP0100000000?lT003o9000000302_o00050000001`o`00L?l0073o000000000`2S
o`001@000000_ol00;oo002oo`00000000<0`Ol000@000000:So002Xo`0000000`1eo`001@000000
:_l002[o000Zo`00000000<00?lb00D00000003oVP00oiX00?nJ00000003003ko`0400000000V?l0
09So000000<004?o00D000000005o`001Ol000Go0000000301`0o`050000000N0?l07P3o01h0o`00
00000`000?l01000000003oo000oo`000003002Io`0500000000ooP00?oh003on00000000`00ogl0
1@0000000?l3003o0`00o`<0000000<0Kol000@000000=3o003@o`0000000`3oj@001@000000ol@0
0?o4003oa000000000<0ol<000D000000?oW003oi`00onL0000000020=;o00050000001^o`00K_l0
06ko000000000`00o`T01@0000000?n>003oSP00ohh0000000<00>co00D00000001bo`00L_l007;o
00000002000<o`050000000l0?l0?03o03`0o`0000000`1R0?l01@000000HP3o0680o`1R0?l00000
00<0?03o00D00000000<o`003?l000co00000002001bo`0500000000k?l00>co003/o`0000000`00
ohh01@0000000?l9003o2@00o`T0000000<0K_l000D000000=;o003Bo`00d_l0000000020?oW0005
0000003o``00ol<00?o3000000000`3oa0001@000000onT00?oY003oj@00000000<0d?l000D00000
06oo001_o`00Kol000000002003o0`0500000000ogl00?mo003oO`0000000`00ooP01@00000009Wo
002Io`00VOl0000000<003oo00D000000000o`000?l0003o0000000201h0o`050000000L0?l0703o
01`0o`0000000`001Ol01@00000004?o0013o`00@ol0000000<009So00D00000003ko`00nol00?_o
00000002003oVP0500000000oc800?lb003o<P0000000`0Zo`001@000000MOl007Go001eo`000000
00<0Z?l000D000000<7o0031o`00`Ol0000000020;oo00050000002So`00Xol00:?o000000000`1`
o`001@000000:ol002_o000[o`00000000<00?lT00D00000003oN000ogP00?mh00000002003obP05
00000000kOl00>go003]o`0000000`00/ol01@00000008co002<o`00S?l0000000<007Wo00@00000
001lo`00O?l000000`00T_l01@0000000;Wo002io`00^Ol0000000<00>co00D00000003of000omP0
0?oH00000003003oVP030000003oool0oooo00<0oooo0P0000010?ooo`00803oool00`000000oooo
0?ooo`050?ooo`0400000000oiX00?nJ000000<00?oH00D00000003/o`00k?l00>co00000003002i
o`0500000000T_l009;o002Bo`0000000`00O?l01000000007Wo001io`000003002<o`0500000000
/ol00;?o002co`0000000`00kOl01@0000000?o:003obP00olX0000000<00?mh00@00000003o9000
ob@000000`0[o`001@000000L?l0073o001`o`00000000<0Xol000D000000;oo002oo`00_ol00000
00030<7o00040000002Xo`00Z?l0000000<0MOl000D0000002[o000Zo`00:_l000000003003o<P05
00000000oiX00?nJ003oVP0000000`00nol01000000009So002Ho`0000030013o`05000000001Ol0
00Go0005o`0000000`0L0?l01@0000007P3o01h0o`0N0?l0000000<0003o00@00000000oo`00?ol0
00000`00VOl01@0000000?oh003on000ooP0000000<00?mo00D00000003o0`00o`<00?l300000003
06oo00040000003@o`00d?l0000000<0onT000D000000?o4003oa000ol@0000000030?o300050000
003oi`00onL00?oW000000000P3Bo`001@000000K_l006ko001^o`00000000<00?l900D00000003o
SP00ohh00?n>00000003003/o`0500000000L_l007;o001bo`0000000P003?l01@000000?03o03`0
o`0l0?l0000000<0HP3o00D000000680o`1R0?l0HP3o0000000303`0o`05000000003?l000co000<
o`0000000P00L_l01@0000000>co003/o`00k?l0000000<00?n>00D00000003o2@00o`T00?l90000
000306ko00050000003Bo`00d_l00=;o000000000P3oi`001@000000ol<00?o3003o``00000000<0
ol@000D000000?oY003oj@00onT0000000030=3o00050000001_o`00Kol006oo000000000P00o`<0
1@0000000?mo003oO`00ogl0000000<00?oh00D00000002Io`00VOl009Wo00000003000oo`050000
00000?l0003o0000o`0000000P0N0?l01@000000703o01`0o`0L0?l0000000<000Go00D000000013
o`00@ol004?o00000003002Ho`0500000000nol00?_o003ko`0000000P00oiX01@0000000?lb003o
<P00oc80000000<0:_l000D0000007Go001eo`00MOl0000000030:So000500000031o`00`Ol00<7o
000000000P2oo`001@000000Xol00:?o002So`00000000<0L?l000D0000002_o000[o`00:ol00000
0003003o900500000000ogP00?mh003oN00000000P00olX01@0000000>go003]o`00kOl0000000<0
0;?o00D00000002<o`00S?l008co00000003001io`0400000000O?l007co000000<009;o00D00000
002io`00^Ol00;Wo00000003003/o`0500000000omP00?oH003of00000000`00oiX00`000000oooo
0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo0000084000001P3o
ool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003oh000on0000000`00
i_l01@0000000;Go002eo`00]Ol0000000<0097o00D00000001lo`00O?l007co00000003001jo`04
00000000S?l008co000000<00;;o00D00000003Zo`00j_l00>[o00000003003oc`0500000000oh00
0?n0003oP00000000`00obd0100000008_l002;o0000000306So00050000002No`00W_l009ko0000
00000`2oo`001@000000aol00<Oo0037o`00000000<0]Ol000@0000008Wo0029o`0000000`15o`00
1@0000000?lA003o4@00oa40000000<00?me00D00000003ogP00omh00?oN00000003002jo`040000
0000Gol005oo000000<001Ko00D0000001P0o`0H0?l0603o0000000302X0o`050000000K0?l06`3o
01/0o`0000000`004_l01000000005ko001No`0000030031o`0500000000olX00?o:003obP000000
0`00odh01@000000:ol002_o000[o`00000000<0VOl000@000000?;o003bo`0000000`3od0001@00
0000ok@00?nd003o]000000000<0ok`000D000000?oX003oj000onP0000000020<[o00050000001Q
o`00HOl0067o000000000`00oaX01@0000000?nO003oW`00oil0000000<00=_o00D00000001Qo`00
HOl0067o0000000200<0o`050000001;0?l0B`3o04/0o`0000000`1`0?l01@000000L03o0700o`1`
0?l0000000<0B`3o00D0000000<0o`030?l00`3o00000002001Qo`0500000000fol00=_o003Ko`00
00000`00oil01@0000000?lJ003o6P00oaX0000000<0HOl000D000000<[o003:o`00b_l000000002
0?oX00050000003o_000ok`00?nl000000000`3o]0001@000000om000?o@003od000000000<0l_l0
00D0000009Wo002Io`00VOl00000000202_o000500000000odh00?m>003oCP0000000`00olX01@00
00000<7o0031o`00`Ol0000000<005ko00D00000000Bo`004_l001;o0000000201/0o`050000000Z
0?l0:P3o02X0o`0000000`0H0?l01@00000001Ko000Fo`005_l0000000<005oo00D00000002jo`00
^_l00;[o00000002003ogP0500000000ogD00?me003oM@0000000`00oa401@000000AOl004Go0015
o`00000000<0ROl000D000000;Go002eo`00]Ol0000000020<Oo00050000002oo`00_ol00;oo0000
00000`2No`001@000000J?l006So001Xo`00000000<08_l000D00000003o;@00obd00?l]00000002
003oP00500000000oll00?o?003oc`0000000`00j_l01@0000000;;o002bo`00/_l0000000<008co
00@00000001jo`00N_l000000`00O?l01@000000097o002Ao`00TOl0000000<00;Go00D00000003V
o`00i_l00>Ko00000003003oh0030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`00
0000oooo0?ooo`050?ooo`0400000000on000?oP000000<00>Ko00D00000002eo`00]Ol00;Go0000
0003002Ao`0500000000O?l007co001lo`0000000`00N_l01000000008co002<o`000003002bo`05
00000000j_l00>[o003Zo`0000000`00oll01@0000000?n0003oP000oh00000000<00?l]00@00000
02;o000Ro`0000000`1Xo`001@000000W_l009ko002No`00000000<0_ol000D000000<Oo0037o`00
aol0000000030;Go000400000029o`00ROl0000000<0AOl000D00000003o4@00oa400?lA00000003
003oM@0500000000omh00?oN003ogP0000000`00^_l01000000005oo001Oo`000003000Fo`050000
000H0?l0603o01P0o`0000000`0Z0?l01@0000006`3o01/0o`0K0?l0000000<001;o00@00000001N
o`00G_l000000`00`Ol01@0000000?o:003obP00olX0000000<00?m>00D0000002_o000[o`00:ol0
0000000309Wo00040000003bo`00l_l0000000<0om0000D000000?nd003o]000ok@0000000030?nl
00050000003oj000onP00?oX000000000P3:o`001@000000HOl0067o001Qo`00000000<00?lJ00D0
0000003oW`00oil00?nO00000003003Ko`0500000000HOl0067o001Qo`0000000P030?l01@000000
B`3o04/0o`1;0?l0000000<0L03o00D000000700o`1`0?l0L03o0000000304/0o`05000000030?l0
0`3o00<0o`0000000P00HOl01@0000000=_o003Ko`00fol0000000<00?nO00D00000003o6P00oaX0
0?lJ00000003067o00050000003:o`00b_l00<[o000000000P3oj0001@000000ok`00?nl003o_000
000000<0ok@000D000000?o@003od000om00000000030?;o00050000002Io`00VOl009Wo00000000
0P0[o`001@0000000?m>003oCP00odh0000000<00?o:00D000000031o`00`Ol00<7o00000003001N
o`05000000004_l001;o000Bo`0000000P0K0?l01@000000:P3o02X0o`0Z0?l0000000<0603o00D0
0000000Fo`005_l001Ko00000003001Oo`0500000000^_l00;[o002jo`0000000P00omh01@000000
0?me003oM@00ogD0000000<00?lA00D0000004Go0015o`00AOl00000000308Wo00050000002eo`00
]Ol00;Go000000000P37o`001@000000_ol00;oo002oo`00000000<0W_l000D0000006So001Xo`00
J?l00000000302;o000500000000obd00?l]003o;@0000000P00oh001@0000000?o?003oc`00oll0
000000<00>[o00D00000002bo`00/_l00;;o00000003002<o`0400000000N_l007[o000000<007co
00D00000002Ao`00TOl0097o00000003002eo`0500000000i_l00>Ko003Vo`0000000`00on000`00
0000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool010000000
0?oP003oh0000003003Vo`0500000000]Ol00;Go002eo`0000000`00TOl01@00000007co001lo`00
O?l0000000<007[o00@00000002<o`00S?l000000`00/_l01@0000000>[o003Zo`00j_l0000000<0
0?o?00D00000003oP000oh000?n000000003003o;@040000000Ro`008_l0000000<0J?l000D00000
09ko002No`00W_l0000000030;oo000500000037o`00aol00<Oo000000000`2eo`0010000000ROl0
08Wo0000000304Go000500000000oa400?lA003o4@0000000`00ogD01@0000000?oN003ogP00omh0
000000<00;[o00@00000001Oo`00Gol000000`005_l01@000000603o01P0o`0H0?l0000000<0:P3o
00D0000001/0o`0K0?l06`3o00000003000Bo`0400000000G_l005ko000000<00<7o00D00000003o
bP00olX00?o:00000003003oCP050000000[o`00:ol002_o000000000`2Io`0010000000l_l00?;o
000000030?o@00050000003o]000ok@00?nd000000000`3o_0001@000000onP00?oX003oj0000000
0080b_l000D00000067o001Qo`00HOl000000003003o6P0500000000oil00?nO003oW`0000000`00
fol01@000000067o001Qo`00HOl0000000800`3o00D0000004/0o`1;0?l0B`3o000000030700o`05
0000001`0?l0L03o0700o`0000000`1;0?l01@0000000`3o00<0o`030?l000000080067o00D00000
003Ko`00fol00=_o00000003003oW`0500000000oaX00?lJ003o6P0000000`1Qo`001@000000b_l0
0<[o003:o`0000000080onP000D000000?nl003o_000ok`0000000030?nd00050000003od000om00
0?o@000000000`3bo`001@000000VOl009Wo002Io`0000000080:ol000D00000003oCP00odh00?m>
00000003003obP0500000000`Ol00<7o0031o`0000000`00G_l01@00000001;o000Bo`004_l00000
00806`3o00D0000002X0o`0Z0?l0:P3o0000000301P0o`05000000005_l001Ko000Fo`0000000`00
Gol01@0000000;[o002jo`00^_l0000000800?oN00D00000003oM@00ogD00?me00000003003o4@05
00000015o`00AOl004Go000000000`29o`001@000000]Ol00;Go002eo`0000000080aol000D00000
0;oo002oo`00_ol00000000309ko00050000001Xo`00J?l006So000000000`0Ro`001@0000000?l]
003o;@00obd0000000800?n000D00000003oc`00oll00?o?00000003003Zo`0500000000/_l00;;o
002bo`0000000`00S?l01000000007[o001jo`000003001lo`0500000000TOl0097o002Ao`000000
0`00]Ol01@0000000>Ko003Vo`00i_l0000000<00?oP00<000000?ooo`3oool0103oool100000040
oooo000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooo`0400000000gol00=oo000000<00;7o00D00000002?o`00Sol0
08oo00000003001lo`0500000000O?l007co001lo`0000000`00S_l0100000000;?o002co`000003
003Zo`0500000000om400?oA003od@0000000`00oh@01@0000000?lc003o<`00oc<0000000<06ol0
00@00000067o001Qo`0000000`2Io`001@000000__l00;ko002no`00000000<0bol000D000000;oo
002oo`00_ol00000000309Wo00040000001Lo`00G?l0000000<02_l000D00000003oE@00oeD00?mE
00000003003o_00500000000fol00=_o003Ko`0000000`00Nol01000000002_o000[o`00000300l0
o`050000000^0?l0;P3o02h0o`0000000`0^0?l01@0000003`3o00l0o`0?0?l0000000<002go00@0
00000024o`00Q?l000000`00kol01@0000000?nG003oU`00oiL0000000<00?lK00D0000005[o001J
o`00F_l0000000030<;o00040000003ok000on`0000000<0okP000D000000?nV003oYP00ojH00000
00030?nh00050000003ok000on`00?o/000000000P2oo`001@000000D_l005;o001Bo`00000000<0
0?l[00D00000003o/P00ok800?nb000000030039o`0500000000D?l0053o001@o`0000000P0B0?l0
1@000000F@3o05T0o`1I0?l0000000<0OP3o00D0000007h0o`1n0?l0OP3o0000000305T0o`050000
000B0?l04P3o0180o`0000000P00D?l01@0000000<Wo0039o`00bOl0000000<00?nb00D00000003o
:`00ob/00?l[0000000305;o00050000002oo`00_ol00;oo000000000P3ok0001@000000okP00?nh
003o^000000000<0ojH000D000000?nh003o^000okP0000000030?o/000500000032o`00`_l00<;o
000000000P1Jo`001@0000000?lK003o6`00oa/0000000<00?nG00D00000003_o`00kol00>oo0000
00030024o`0500000000;Ol002go000]o`0000000P0?0?l01@000000;P3o02h0o`0^0?l0000000<0
;P3o00D0000000l0o`0?0?l03`3o00000003000[o`0500000000Nol007_o001ko`0000000P00fol0
1@0000000?nl003o_000ok`0000000<00?mE00D0000000[o000:o`002_l00000000305co00050000
002Io`00VOl009Wo000000000P2oo`001@000000bol00<_o003;o`00000000<0__l000D0000009Wo
002Io`00VOl000000003067o00050000000Ko`006ol001_o000000000P00oc<01@0000000?n4003o
Q000oh@0000000<00?oA00D00000003Zo`00j_l00>[o00000003002co`0400000000S_l008ko0000
00<007co00D00000001lo`00O?l007co00000003002?o`0500000000/Ol00;7o002ao`0000000`00
gol00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool0
100000000=oo003Oo`000003002ao`0500000000Sol008oo002?o`0000000`00O?l01@00000007co
001lo`00O?l0000000<008ko00@00000002co`00/ol000000`00j_l01@0000000?oA003od@00om40
000000<00?n400D00000003o<`00oc<00?lc0000000301_o00040000001Qo`00HOl0000000<0VOl0
00D000000;ko002no`00__l0000000030<_o00050000002oo`00_ol00;oo000000000`2Io`001000
0000G?l005co0000000300[o000500000000oeD00?mE003oE@0000000`00ok`01@0000000=_o003K
o`00fol0000000<007_o00@00000000[o`00:ol000000`0?0?l01@000000;P3o02h0o`0^0?l00000
00<0;P3o00D0000000l0o`0?0?l03`3o00000003000]o`0400000000Q?l008Co000000<00>oo00D0
0000003oU`00oiL00?nG00000003003o6`050000001Jo`00F_l005[o000000000`32o`0010000000
on`00?o/000000030?nh00050000003oYP00ojH00?nV000000000`3o^0001@000000on`00?o/003o
k00000000080_ol000D0000005;o001Bo`00D_l000000003003o:`0500000000ok800?nb003o/P00
00000`00bOl01@000000053o001@o`00D?l0000000804P3o00D0000005T0o`1I0?l0F@3o00000003
07h0o`050000001n0?l0OP3o07h0o`0000000`1I0?l01@0000004P3o0180o`0B0?l000000080053o
00D000000039o`00bOl00<Wo00000003003o/P0500000000ob/00?l[003o:`0000000`1Bo`001@00
0000_ol00;oo002oo`0000000080on`000D000000?nh003o^000okP0000000030?nV00050000003o
^000okP00?nh000000000`3ok0001@000000`_l00<;o0032o`0000000080F_l000D00000003o6`00
oa/00?lK00000003003oU`0500000000kol00>oo003_o`0000000`00Q?l01@00000002go000]o`00
;Ol0000000803`3o00D0000002h0o`0^0?l0;P3o0000000302h0o`050000000?0?l03`3o00l0o`00
00000`00:ol01@00000007_o001ko`00Nol0000000800=_o00D00000003o_000ok`00?nl00000003
003oE@050000000:o`002_l000[o000000000`1Lo`001@000000VOl009Wo002Io`0000000080_ol0
00D000000<_o003;o`00bol0000000030;ko00050000002Io`00VOl009Wo000000000`1Qo`001@00
00006ol001_o000Ko`00000000800?lc00D00000003oQ000oh@00?n400000003003od@0500000000
j_l00>[o003Zo`0000000`00/ol01000000008ko002>o`000003001lo`0500000000O?l007co001l
o`0000000`00Sol01@0000000;7o002ao`00/Ol0000000<00=oo00<000000?ooo`3oool0103oool1
00000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003Oo`00gol000000`00/Ol0
1@00000008oo002?o`00Sol0000000<007co00D00000001lo`00O?l007co00000003002>o`040000
0000/ol00;?o000000<00>[o00D00000003od@00om400?oA00000003003oQ00500000000oc<00?lc
003o<`0000000`0Ko`0010000000HOl0067o0000000309Wo00050000002no`00__l00;ko00000000
0`3;o`001@000000_ol00;oo002oo`00000000<0VOl000@0000005co001Lo`0000000`0:o`001@00
00000?mE003oE@00oeD0000000<00?nl00D00000003Ko`00fol00=_o00000003001ko`0400000000
:ol002_o000000<03`3o00D0000002h0o`0^0?l0;P3o0000000302h0o`050000000?0?l03`3o00l0
o`0000000`00;Ol01000000008Co0024o`000003003_o`0500000000oiL00?nG003oU`0000000`00
oa/01@000000F_l005[o001Jo`00000000<0`_l000@000000?o/003ok00000000`3o^0001@000000
ojH00?nV003oYP00000000<0okP000D000000?o/003ok000on`0000000020;oo00050000001Bo`00
D_l005;o000000000`00ob/01@0000000?nb003o/P00ok80000000<00<Wo00D00000001@o`00D?l0
053o000000020180o`050000001I0?l0F@3o05T0o`0000000`1n0?l01@000000OP3o07h0o`1n0?l0
000000<0F@3o00D000000180o`0B0?l04P3o00000002001@o`0500000000bOl00<Wo0039o`000000
0`00ok801@0000000?l[003o:`00ob/0000000<0D_l000D000000;oo002oo`00_ol0000000020?o/
00050000003o^000okP00?nh000000000`3oYP001@000000okP00?nh003o^000000000<0on`000D0
00000<;o0032o`00`_l00000000205[o000500000000oa/00?lK003o6`0000000`00oiL01@000000
0>oo003_o`00kol0000000<008Co00D00000000]o`00;Ol002go0000000200l0o`050000000^0?l0
;P3o02h0o`0000000`0^0?l01@0000003`3o00l0o`0?0?l0000000<002_o00D00000001ko`00Nol0
07_o00000002003Ko`0500000000ok`00?nl003o_00000000`00oeD01@0000002_l000[o000:o`00
000000<0G?l000D0000009Wo002Io`00VOl0000000020;oo00050000003;o`00bol00<_o00000000
0`2no`001@000000VOl009Wo002Io`00000000<0HOl000D0000001_o000Ko`006ol000000002003o
<`0500000000oh@00?n4003oQ00000000`00om401@0000000>[o003Zo`00j_l0000000<00;?o00@0
0000002>o`00S_l000000`00O?l01@00000007co001lo`00O?l0000000<008oo00D00000002ao`00
/Ol00;7o00000003003Oo`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000
oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool0
1@3oool0100000000:co002/o`000003002=o`0500000000OOl007go001mo`0000000`00Ool01@00
000009;o002Bo`00T_l0000000<00;So00@00000003]o`00kOl000000`00oll01@0000000?n4003o
Q000oh@0000000<00?ld00D0000001So000Ho`006?l00000000305ko00040000002Fo`00U_l00000
00<0_?l000D000000<go003=o`00cOl0000000030<Go00050000002Uo`00YOl00:Go000000000`1]
o`00100000008Ol0027o00000003003o>@0500000000oih00?nN003oWP0000000`00nOl01@000000
09Oo002Go`00Uol0000000<0047o00@000000040o`010?l000000`0/0?l01@000000>P3o03X0o`0j
0?l0000000<0:@3o00D000000005o`001Ol000Go00000003001?o`0400000000/?l00;3o000000<0
0?oM00D00000003oHP00of800?mR0000000301Oo000500000029o`00ROl008Wo000000000`3Zo`00
10000000old00?o=000000030?nS00050000003oV`00oi/00?nK000000000`3o]P001@000000oo<0
0?oc003ol`0000000080/ol000D0000004;o0012o`00@_l000000003003o?P0500000000olD00?o5
003oa@0000000`00]_l01@00000003oo000oo`00?ol0000000808P3o00D0000006P0o`1X0?l0J03o
0000000308`0o`050000002<0?l0S03o08`0o`0000000`1X0?l01@0000008P3o0280o`0R0?l00000
008003oo00D00000002fo`00]_l00;Ko00000003003oa@0500000000och00?ln003o?P0000000`12
o`001@000000/ol00;?o002co`0000000080oo<000D000000?nf003o]P00okH0000000030?nK0005
0000003oX`00oj<00?nS000000000`3oc@001@000000j_l00>[o003Zo`0000000080ROl000D00000
01Oo000Go`005ol000000003003oHP0500000000omd00?oM003og@0000000`00/?l01@00000004oo
001?o`00Col00000008000Go00D0000002T0o`0Y0?l0:@3o0000000303X0o`050000000/0?l0;03o
02`0o`0000000`010?l01@000000047o0011o`00@Ol00000008009Oo00D00000003io`00nOl00?Wo
00000003003oWP0500000000ocT00?li003o>@0000000`0Qo`001@000000KOl006go001]o`000000
0080YOl000D000000<Go0035o`00aOl0000000030<go00050000002lo`00_?l00;co000000000`2F
o`001@000000G_l005ko001No`00000000806?l000D00000003o=000oc@00?ld00000003003oQ005
00000000oll00?o?003oc`0000000`00kOl0100000000;So002ho`000003002Bo`0500000000Ool0
07oo001oo`0000000`00OOl01@00000008go002=o`00SOl0000000<00:co00<000000?ooo`3oool0
103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000002/o`00[?l00000
0`00SOl01@00000007go001mo`00OOl0000000<007oo00D00000002Bo`00T_l009;o00000003002h
o`0400000000kOl00>go000000<00?o?00D00000003oQ000oh@00?n400000003003o=0050000000H
o`006?l001So000000000`1No`0010000000U_l009Ko000000030;co00050000003=o`00cOl00<go
000000000`35o`001@000000YOl00:Go002Uo`00000000<0KOl000@00000027o000Qo`0000000`00
ocT01@0000000?nN003oWP00oih0000000<00?Wo00D00000002Go`00Uol009Oo000000030011o`04
000000010?l00@3o000000<0;03o00D0000003X0o`0j0?l0>P3o0000000302T0o`05000000001Ol0
00Go0005o`0000000`00Col0100000000;3o002`o`000003003og@0500000000of800?mR003oHP00
00000`0Go`001@000000ROl008Wo0029o`00000000<0j_l000@000000?o=003oc@0000000`3oX`00
1@000000oi/00?nK003oV`00000000<0okH000D000000?oc003ol`00oo<0000000020;?o00050000
0012o`00@_l004;o000000000`00och01@0000000?o5003oa@00olD0000000<00;Ko00D00000000o
o`00?ol003oo000000020280o`050000001X0?l0J03o06P0o`0000000`2<0?l01@000000S03o08`0
o`2<0?l0000000<0J03o00D000000280o`0R0?l08P3o00000002000oo`0500000000]_l00;Ko002f
o`0000000`00olD01@0000000?ln003o?P00och0000000<0@_l000D000000;?o002co`00/ol00000
00020?oc00050000003o]P00okH00?nf000000000`3oV`001@000000oj<00?nS003oX`00000000<0
old000D000000>[o003Zo`00j_l00000000208Wo00050000000Go`005ol001Oo000000000`00of80
1@0000000?oM003og@00omd0000000<00;3o00D00000001?o`00Col004oo000000020005o`050000
000Y0?l0:@3o02T0o`0000000`0j0?l01@000000;03o02`0o`0/0?l0000000<00@3o00D000000011
o`00@Ol0047o00000002002Go`0500000000nOl00?Wo003io`0000000`00oih01@0000000?li003o
>@00ocT0000000<08Ol000D0000006go001]o`00KOl0000000020:Go000500000035o`00aOl00<Go
000000000`3=o`001@000000_?l00;co002lo`00000000<0U_l000D0000005ko001No`00G_l00000
000201So000500000000oc@00?ld003o=00000000`00oh@01@0000000?o?003oc`00oll0000000<0
0>go00@00000002ho`00^?l000000`00T_l01@00000007oo001oo`00Ool0000000<007go00D00000
002=o`00SOl008go00000003002/o`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool0
0`000000oooo0?ooo`050?ooo`0400000000[?l00:co000000<008go00D00000001mo`00OOl007go
00000003001oo`0500000000T_l009;o002Bo`0000000`00^?l0100000000>go003]o`000003003o
c`0500000000oh@00?n4003oQ00000000`00oc@01@0000006?l001So000Ho`00000000<0G_l000@0
000009Ko002Fo`0000000`2lo`001@000000cOl00<go003=o`00000000<0aOl000D000000:Go002U
o`00YOl00000000306go00040000000Qo`008Ol0000000<00?li00D00000003oWP00oih00?nN0000
0003003io`0500000000Uol009Oo002Go`0000000`00@Ol0100000000@3o0040o`00000302`0o`05
0000000j0?l0>P3o03X0o`0000000`0Y0?l01@00000000Go0005o`001Ol0000000<004oo00@00000
002`o`00/?l000000`00omd01@0000000?mR003oHP00of80000000<05ol000D0000008Wo0029o`00
ROl0000000030>[o00040000003oc@00old0000000<0oj<000D000000?nK003oV`00oi/000000003
0?nf00050000003ol`00oo<00?oc000000000P2co`001@000000@_l004;o0012o`00000000<00?ln
00D00000003oa@00olD00?o500000003002fo`0500000000?ol003oo000oo`0000000P0R0?l01@00
0000J03o06P0o`1X0?l0000000<0S03o00D0000008`0o`2<0?l0S03o0000000306P0o`050000000R
0?l08P3o0280o`0000000P00?ol01@0000000;Ko002fo`00]_l0000000<00?o500D00000003o?P00
och00?ln0000000304;o00050000002co`00/ol00;?o000000000P3ol`001@000000okH00?nf003o
]P00000000<0oi/000D000000?nS003oX`00oj<0000000030?o=00050000003Zo`00j_l00>[o0000
00000P29o`001@0000005ol001Oo000Go`00000000<00?mR00D00000003og@00omd00?oM00000003
002`o`0500000000Col004oo001?o`0000000P001Ol01@000000:@3o02T0o`0Y0?l0000000<0>P3o
00D0000002`0o`0/0?l0;03o000000030040o`0500000000@Ol0047o0011o`0000000P00Uol01@00
00000?Wo003io`00nOl0000000<00?nN00D00000003o>@00ocT00?li00000003027o00050000001]
o`00KOl006go000000000P2Uo`001@000000aOl00<Go0035o`00000000<0cOl000D000000;co002l
o`00_?l00000000309Ko00050000001No`00G_l005ko000000000P0Ho`001@0000000?ld003o=000
oc@0000000<00?n400D00000003oc`00oll00?o?00000003003]o`0400000000^?l00;So000000<0
09;o00D00000001oo`00Ool007oo00000003001mo`0500000000SOl008go002=o`0000000`00[?l0
0`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo0000
084000001P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000002;o`00
Rol000000`00Ool01@00000008?o0023o`00Pol0000000<009So00D00000002oo`00_ol00;oo0000
0003003do`0400000000olX00?o:000000<00?n000D00000003o<`00oc<00?lc0000000301So0005
0000001Lo`00G?l005co000000000`2Do`0010000000^ol00;_o000000030<ko000500000039o`00
bOl00<Wo000000000`2]o`001@000000Nol007_o001ko`00000000<0=?l000@00000003o8`00ob<0
00000`00ohD01@0000000?o[003oj`00on/0000000<00;7o00D00000001Go`00Eol005Oo00000003
000=o`040000000V0?l09P3o000000<0?`3o00D0000003/0o`0k0?l0>`3o0000000301X0o`050000
00008ol002?o000So`0000000`00NOl0100000000>?o003So`000003003oYP0500000000ob/00?l[
003o:`0000000`1<o`001@000000^?l00;So002ho`00000000<0onl000@000000?na003o/@000000
0`3oTP001@000000oi@00?nD003oU000000000<0okP000D000000?ok003on`00oo/0000000020:Co
00050000000_o`00;ol002oo000000000`00oe801@0000000?oJ003ofP00omX0000000<00:;o00D0
0000000]o`00;Ol002go000000020380o`050000001f0?l0MP3o07H0o`0000000`2J0?l01@000000
VP3o09X0o`2J0?l0000000<0MP3o00D000000380o`0b0?l0<P3o00000002000]o`0500000000X_l0
0:;o002Ro`0000000`00omX01@0000000?mB003oDP00oe80000000<0;ol000D000000:Co002To`00
Y?l0000000020?ok00050000003o^000okP00?nh000000000`3oU0001@000000oi800?nB003oTP00
000000<0ok4000D000000?o_003ok`00onl0000000020;So00050000001<o`00C?l004co00000000
0`00ob/01@0000000?nV003oYP00ojH0000000<00>?o00D00000001io`00NOl007Wo00000002000S
o`050000000J0?l06P3o01X0o`0000000`0k0?l01@000000?`3o03l0o`0o0?l0000000<09P3o00D0
0000000=o`003Ol000go00000002001Go`0500000000/Ol00;7o002ao`0000000`00on/01@000000
0?n5003oQ@00ohD0000000<00?lS00D0000003Co000do`00=?l00000000207_o00050000002]o`00
[Ol00:go000000000`39o`001@000000c_l00<ko003>o`00000000<0^ol000D0000009Co002Do`00
U?l00000000205co00050000000Ho`006?l001So000000000`00oc<01@0000000?n0003oP000oh00
000000<00?o:00@00000003do`00m?l000000`00_ol01@00000009So002Ho`00V?l0000000<008?o
00D00000001oo`00Ool007oo00000003002;o`030000003oool0oooo00@0oooo0@0000010?ooo`00
803oool00`000000oooo0?ooo`050?ooo`0400000000Rol008_o000000<007oo00D000000023o`00
Pol008?o00000003002Ho`0500000000_ol00;oo002oo`0000000`00m?l0100000000?o:003obP00
0003003oP00500000000oc<00?lc003o<`0000000`0Ho`001@000000G?l005co001Lo`00000000<0
U?l000@000000;_o002ko`0000000`3>o`001@000000bOl00<Wo0039o`00000000<0[Ol000D00000
07_o001ko`00Nol00000000303Co000400000000ob<00?lS000000<00?n500D00000003oj`00on/0
0?o[00000003002ao`0500000000Eol005Oo001Go`0000000`003Ol0100000009P3o02H0o`000003
03l0o`050000000k0?l0>`3o03/0o`0000000`0J0?l01@00000002?o000So`008ol0000000<007Wo
00@00000003So`00hol000000`00ojH01@0000000?l[003o:`00ob/0000000<0C?l000D000000;So
002ho`00^?l0000000030?o_00040000003o/@00ok40000000<0oi8000D000000?nD003oU000oi@0
000000030?nh00050000003on`00oo/00?ok000000000P2To`001@000000;ol002oo000_o`000000
00<00?mB00D00000003ofP00omX00?oJ00000003002Ro`0500000000;Ol002go000]o`0000000P0b
0?l01@000000MP3o07H0o`1f0?l0000000<0VP3o00D0000009X0o`2J0?l0VP3o0000000307H0o`05
0000000b0?l0<P3o0380o`0000000P00;Ol01@0000000:;o002Ro`00X_l0000000<00?oJ00D00000
003oDP00oe800?mB0000000302oo00050000002To`00Y?l00:Co000000000P3on`001@000000okP0
0?nh003o^000000000<0oi@000D000000?nB003oTP00oi80000000030?na00050000003ok`00onl0
0?o_000000000P2ho`001@000000C?l004co001<o`00000000<00?l[00D00000003oYP00ojH00?nV
00000003003So`0500000000NOl007Wo001io`0000000P008ol01@0000006P3o01X0o`0J0?l00000
00<0>`3o00D0000003l0o`0o0?l0?`3o0000000302H0o`05000000003Ol000go000=o`0000000P00
Eol01@0000000;7o002ao`00/Ol0000000<00?o[00D00000003oQ@00ohD00?n500000003003o8`05
0000000do`00=?l003Co000000000P1ko`001@000000[Ol00:go002]o`00000000<0bOl000D00000
0<ko003>o`00c_l0000000030;_o00050000002Do`00U?l009Co000000000P1Lo`001@0000006?l0
01So000Ho`00000000<00?lc00D00000003oP000oh000?n000000003003obP0400000000m?l00?Co
000000<00;oo00D00000002Ho`00V?l009So000000030023o`0500000000Ool007oo001oo`000000
0`00Rol00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3o
ool01000000008_o002;o`000003001oo`0500000000Pol008?o0023o`0000000`00V?l01@000000
0;oo002oo`00_ol0000000<00?Co00@00000003obP00olX000000`00oh001@0000000?lc003o<`00
oc<0000000<06?l000D0000005co001Lo`00G?l00000000309Co00040000002ko`00^ol0000000<0
c_l000D000000<Wo0039o`00bOl0000000030:go00050000001ko`00Nol007_o000000000`0do`00
100000000?lS003o8`000003003oQ@0500000000on/00?o[003oj`0000000`00/Ol01@00000005Oo
001Go`00Eol0000000<000go00@0000002H0o`0V0?l000000`0o0?l01@000000>`3o03/0o`0k0?l0
000000<06P3o00D00000000So`008ol002?o00000003001io`0400000000hol00>?o000000<00?nV
00D00000003o:`00ob/00?l[0000000304co00050000002ho`00^?l00;So000000000`3ok`001000
0000ok400?na000000030?nB00050000003oU000oi@00?nD000000000`3o^0001@000000oo/00?ok
003on`0000000080Y?l000D0000002oo000_o`00;ol000000003003oDP0500000000omX00?oJ003o
fP0000000`00X_l01@00000002go000]o`00;Ol000000080<P3o00D0000007H0o`1f0?l0MP3o0000
000309X0o`050000002J0?l0VP3o09X0o`0000000`1f0?l01@000000<P3o0380o`0b0?l000000080
02go00D00000002Ro`00X_l00:;o00000003003ofP0500000000oe800?mB003oDP0000000`0_o`00
1@000000Y?l00:Co002To`0000000080oo/000D000000?nh003o^000okP0000000030?nD00050000
003oTP00oi800?nB000000000`3o/@001@000000onl00?o_003ok`0000000080^?l000D0000004co
001<o`00C?l000000003003o:`0500000000ojH00?nV003oYP0000000`00hol01@00000007Wo001i
o`00NOl00000008002?o00D0000001X0o`0J0?l06P3o0000000303/0o`050000000o0?l0?`3o03l0
o`0000000`0V0?l01@00000000go000=o`003Ol00000008005Oo00D00000002ao`00/Ol00;7o0000
0003003oj`0500000000ohD00?n5003oQ@0000000`00ob<01@000000=?l003Co000do`0000000080
Nol000D000000:go002]o`00[Ol0000000030<Wo00050000003>o`00c_l00<ko000000000`2ko`00
1@000000U?l009Co002Do`0000000080G?l000D0000001So000Ho`006?l000000003003o<`050000
0000oh000?n0003oP00000000`00olX0100000000?Co003do`000003002oo`0500000000V?l009So
002Ho`0000000`00Pol01@00000007oo001oo`00Ool0000000<008_o00<000000?ooo`3oool0103o
ool100000040oooo000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@000001
0?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000P_l008;o000000<008Wo00D00000
002Qo`00XOl00:7o000000030039o`0500000000o_l00?ko003no`0000000`00ol00100000000?mh
003oN0000003003o;@050000000Ko`006ol001_o000000000`1No`001@000000U?l009Co002Do`00
000000<0^ol000@000000<ko003>o`0000000`3<o`001@000000/ol00;?o002co`00000000<0Q?l0
00D00000047o0011o`00@Ol000000003003o4@0400000000og000?m`000000<00?oD00D000000037
o`00aol00<Oo00000003001[o`05000000007Ol001go000Mo`0000000`0M0?l010000000@03o0400
o`00000304L0o`050000000b0?l0<P3o0380o`0000000`010?l01@00000004Wo0019o`00BOl00000
00<00:[o00@00000003oi@00onD000000`00of`01@0000003Ol000go000=o`00000000<0POl000D0
00000>Go003Uo`00iOl0000000030?o;00040000003oU`00oiL0000000<0oh<000D000000?n?003o
S`00ohl0000000030?nl00050000003ho`00n?l00?So000000000P2Co`001@0000006ol001_o000K
o`00000000<00?mX00D00000003ok`00onl00?o_00000003002?o`05000000006ol001_o000Ko`00
00000P120?l01@000000Q@3o08D0o`250?l0000000<0Y`3o00D000000:L0o`2W0?l0Y`3o00000003
08D0o`05000000120?l0@P3o0480o`0000000P006ol01@00000008oo002?o`00Sol0000000<00?o_
00D00000003oJ000ofP00?mX0000000301_o00050000002Co`00Tol009?o000000000P3ho`001@00
0000ok`00?nl003o_000000000<0ohl000D000000?n3003oP`00oh<0000000030?nG00050000003o
b`00ol/00?o;000000000P3Uo`001@000000POl0087o0021o`00000000<03Ol000D00000003oK000
of`00?m/00000003003oi@0500000000Z_l00:[o002Zo`0000000P00BOl01@0000000@3o0040o`01
0?l0000000<0<P3o00D0000004L0o`170?l0A`3o000000030400o`050000000M0?l07@3o01d0o`00
00000P007Ol01@00000006_o001[o`00Jol0000000<00<Oo00D00000003oe000om@00?oD00000003
003oL00500000000oa400?lA003o4@0000000P11o`001@000000Q?l008Co0024o`00000000<0/ol0
00D000000<co003<o`00c?l0000000030<ko00050000002ko`00^ol00;_o000000000P2Do`001@00
0000G_l005ko001No`00000000<06ol000D00000003o;@00obd00?l]00000003003oN00400000000
ol000?o0000000<00?ko00D000000039o`00bOl00<Wo00000003002Qo`0500000000ROl008Wo0029
o`0000000`00P_l00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3o
ool01@3oool01000000008;o0022o`0000030029o`0500000000XOl00:7o002Qo`0000000`00bOl0
1@0000000?ko003no`00o_l0000000<00?o000@00000003oN000ogP000000`00obd01@0000006ol0
01_o000Ko`00000000<0G_l000D0000009Co002Do`00U?l0000000030;_o00040000003>o`00c_l0
000000<0c?l000D000000;?o002co`00/ol00000000308Co000500000011o`00@Ol0047o00000000
0`00oa40100000000?m`003oL0000003003oe00500000000aol00<Oo0037o`0000000`00Jol01@00
000001go000Mo`007Ol0000000<07@3o00@000000400o`100?l000000`170?l01@000000<P3o0380
o`0b0?l0000000<00@3o00D000000019o`00BOl004Wo00000003002Zo`0400000000onD00?oU0000
00<00?m/00D0000000go000=o`003Ol000000003087o00050000003Uo`00iOl00>Go000000000`3o
b`0010000000oiL00?nG000000030?n300050000003oS`00ohl00?n?000000000`3o_0001@000000
n?l00?So003ho`0000000080Tol000D0000001_o000Ko`006ol000000003003oJ00500000000onl0
0?o_003ok`0000000`00Sol01@00000001_o000Ko`006ol000000080@P3o00D0000008D0o`250?l0
Q@3o000000030:L0o`050000002W0?l0Y`3o0:L0o`0000000`250?l01@000000@P3o0480o`120?l0
0000008001_o00D00000002?o`00Sol008oo00000003003ok`0500000000ofP00?mX003oJ0000000
0`0Ko`001@000000Tol009?o002Co`0000000080n?l000D000000?nl003o_000ok`0000000030?n?
00050000003oP`00oh<00?n3000000000`3oU`001@000000ol/00?o;003ob`0000000080iOl000D0
0000087o0021o`00POl00000000300go000500000000of`00?m/003oK00000000`00onD01@000000
0:[o002Zo`00Z_l00000008004Wo00D000000040o`010?l00@3o000000030380o`05000000170?l0
A`3o04L0o`0000000`100?l01@0000007@3o01d0o`0M0?l00000008001go00D00000001[o`00Jol0
06_o000000030037o`0500000000om@00?oD003oe00000000`00og001@0000000?lA003o4@00oa40
00000080@Ol000D0000008Co0024o`00Q?l0000000030;?o00050000003<o`00c?l00<co00000000
0`3>o`001@000000^ol00;_o002ko`0000000080U?l000D0000005ko001No`00G_l00000000301_o
000500000000obd00?l]003o;@0000000`00ogP0100000000?o0003o`0000003003no`0500000000
bOl00<Wo0039o`0000000`00XOl01@00000008Wo0029o`00ROl0000000<008;o00<000000?ooo`3o
ool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@000000022o`00P_l0
00000`00ROl01@0000000:7o002Qo`00XOl0000000<00<Wo00D00000003no`00o_l00?ko00000003
003o`00400000000ogP00?mh000000<00?l]00D0000001_o000Ko`006ol00000000305ko00050000
002Do`00U?l009Co000000000`2ko`0010000000c_l00<ko000000030<co00050000002co`00/ol0
0;?o000000000`24o`001@000000@Ol0047o0011o`00000000<00?lA00@00000003oL000og000000
0`00om@01@0000000<Oo0037o`00aol0000000<006_o00D00000000Mo`007Ol001go0000000301d0
o`04000000100?l0@03o000000<0A`3o00D000000380o`0b0?l0<P3o000000030040o`0500000000
BOl004Wo0019o`0000000`00Z_l0100000000?oU003oi@000003003oK0050000000=o`003Ol000go
000000000`21o`001@000000iOl00>Go003Uo`00000000<0ol/000@000000?nG003oU`0000000`3o
P`001@000000ohl00?n?003oS`00000000<0ok`000D000000?So003ho`00n?l00000000209?o0005
0000000Ko`006ol001_o000000000`00ofP01@0000000?o_003ok`00onl0000000<008oo00D00000
000Ko`006ol001_o000000020480o`05000000250?l0Q@3o08D0o`0000000`2W0?l01@000000Y`3o
0:L0o`2W0?l0000000<0Q@3o00D000000480o`120?l0@P3o00000002000Ko`0500000000Sol008oo
002?o`0000000`00onl01@0000000?mX003oJ000ofP0000000<06ol000D0000009?o002Co`00Tol0
000000020?So00050000003o_000ok`00?nl000000000`3oS`001@000000oh<00?n3003oP`000000
00<0oiL000D000000?o;003ob`00ol/0000000020>Go000500000021o`00POl0087o000000000`0=
o`001@0000000?m/003oK000of`0000000<00?oU00D00000002Zo`00Z_l00:[o000000020019o`05
000000010?l00@3o0040o`0000000`0b0?l01@000000A`3o04L0o`170?l0000000<0@03o00D00000
01d0o`0M0?l07@3o00000002000Mo`0500000000Jol006_o001[o`0000000`00aol01@0000000?oD
003oe000om@0000000<00?m`00D00000003o4@00oa400?lA00000002047o000500000024o`00Q?l0
08Co000000000`2co`001@000000c?l00<co003<o`00000000<0c_l000D000000;_o002ko`00^ol0
0000000209Co00050000001No`00G_l005ko000000000`0Ko`001@0000000?l]003o;@00obd00000
00<00?mh00@00000003o`000ol0000000`00o_l01@0000000<Wo0039o`00bOl0000000<00:7o00D0
00000029o`00ROl008Wo000000030022o`030000003oool0oooo00@0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<00000
0?ooo`3oool01@3oool01000000009;o002Bo`000003002]o`0500000000e_l00=Ko003Fo`000000
0`00oo<01@0000000?nc003o/`00ok<0000000<00?m]00@00000003o9000ob@000000`0Ro`001@00
0000HOl0067o001Qo`00000000<0U_l000D000000;_o002ko`00^ol0000000030<ko00040000003=
o`00cOl0000000<0]_l000D0000008Wo0029o`00ROl00000000304[o000500000000o`D00?l5003o
1@0000000`00of00100000000?o2003o`P000003003Jo`0500000000OOl007go001mo`0000000`00
;?l01@0000004`3o01<0o`0C0?l0000000<0?@3o00@0000004h0o`1>0?l000000`130?l01@000000
7@3o01d0o`0M0?l0000000<002;o00D00000001io`00NOl007Wo00000003003Ro`0400000000ojT0
0?nY000000<00?l_00D0000004Oo0017o`00Aol0000000030;Co00050000003ok`00onl00?o_0000
00000`3oZP0010000000oh400?n1000000030?mg00050000003oSP00ohh00?n>000000000`3o``00
1@000000j_l00>[o003Zo`0000000080POl000D0000000Ko0006o`001_l000000003003oOP050000
0000n_l00?[o003jo`0000000`00N_l01@00000000So0008o`002?l000000080D`3o00D0000009@0
o`2D0?l0U03o000000030;D0o`050000002e0?l0]@3o0;D0o`0000000`2D0?l01@000000D`3o05<0
o`1C0?l00000008000So00D00000001jo`00N_l007[o00000003003jo`0500000000ogh00?mn003o
OP0000000`06o`001@000000POl0087o0021o`0000000080j_l000D000000?o3003o``00ol<00000
00030?n>00050000003oM`00ogL00?mg000000000`3oP@001@000000ojX00?nZ003oZP0000000080
onl000D000000;Co002do`00]?l00000000304Oo000500000000obl00?l_003o;`0000000`00ojT0
1@0000000>;o003Ro`00h_l00000008007Wo00D00000000Ro`008_l002;o0000000301d0o`050000
00130?l0@`3o04<0o`0000000`1>0?l01@000000?@3o03d0o`0m0?l0000000804`3o00D00000000/
o`00;?l002co00000003001mo`0500000000f_l00=[o003Jo`0000000`00ol801@0000000?mP003o
H000of00000000800?l500D0000004[o001:o`00B_l00000000308Wo00050000002fo`00]_l00;Ko
000000000`3=o`001@000000c_l00<ko003>o`0000000080^ol000D0000009Ko002Fo`00U_l00000
0003067o00050000000Ro`008_l002;o000000000`00ob@0100000000?m]003oK@000003003o/`05
00000000oo<00?oc003ol`0000000`00e_l01@0000000:go002]o`00[Ol0000000<009;o00<00000
0?ooo`3oool0103oool100000040oooo000P0?ooo`8000001P3oool01000000009;o002Bo`000003
002]o`0500000000e_l00=Ko003Fo`0000000`00oo<01@0000000?nc003o/`00ok<0000000<00?m]
00@00000003o9000ob@000000`0Ro`001@000000HOl0067o001Qo`00000000<0U_l000D000000;_o
002ko`00^ol0000000030<ko00040000003=o`00cOl0000000<0]_l000D0000008Wo0029o`00ROl0
0000000304[o000500000000o`D00?l5003o1@0000000`00of00100000000?o2003o`P000003003J
o`0500000000OOl007go001mo`0000000`00;?l01@0000004`3o01<0o`0C0?l0000000<0?@3o00@0
000004h0o`1>0?l000000`130?l01@0000007@3o01d0o`0M0?l0000000<002;o00D00000001io`00
NOl007Wo00000003003Ro`0400000000ojT00?nY000000<00?l_00D0000004Oo0017o`00Aol00000
00030;Co00050000003ok`00onl00?o_000000000`3oZP0010000000oh400?n1000000030?mg0005
0000003oSP00ohh00?n>000000000`3o``001@000000j_l00>[o003Zo`0000000080POl000D00000
00Ko0006o`001_l000000003003oOP0500000000n_l00?[o003jo`0000000`00N_l01@00000000So
0008o`002?l000000080D`3o00D0000009@0o`2D0?l0U03o000000030;D0o`050000002e0?l0]@3o
0;D0o`0000000`2D0?l01@000000D`3o05<0o`1C0?l00000008000So00D00000001jo`00N_l007[o
00000003003jo`0500000000ogh00?mn003oOP0000000`06o`001@000000POl0087o0021o`000000
0080j_l000D000000?o3003o``00ol<0000000030?n>00050000003oM`00ogL00?mg000000000`3o
P@001@000000ojX00?nZ003oZP0000000080onl000D000000;Co002do`00]?l00000000304Oo0005
00000000obl00?l_003o;`0000000`00ojT01@0000000>;o003Ro`00h_l00000008007Wo00D00000
000Ro`008_l002;o0000000301d0o`05000000130?l0@`3o04<0o`0000000`1>0?l01@000000?@3o
03d0o`0m0?l0000000804`3o00D00000000/o`00;?l002co00000003001mo`0500000000f_l00=[o
003Jo`0000000`00ol801@0000000?mP003oH000of00000000800?l500D0000004[o001:o`00B_l0
0000000308Wo00050000002fo`00]_l00;Ko000000000`3=o`001@000000c_l00<ko003>o`000000
0080^ol000D0000009Ko002Fo`00U_l000000003067o00050000000Ro`008_l002;o000000000`00
ob@0100000000?m]003oK@000003003o/`0500000000oo<00?oc003ol`0000000`00e_l01@000000
0:go002]o`00[Ol0000000<009;o00<000000?ooo`3oool00`3oool200000040oooo000P0?ooo`03
0000003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?ooo`00803oool00`000000oooo
0?ooo`050?ooo`0400000000_?l00;co000000<00>Oo00D00000003ohP00on800?oR00000003003o
XP0500000000oeh00?mN003oGP0000000`00oaP010000000:ol002_o0000000306So00050000002I
o`00VOl009Wo000000000`2lo`001@000000c_l00<ko003>o`00000000<0c?l000@000000;Ko002f
o`0000000`2;o`001@000000C_l004ko001>o`00000000<00_l000D00000003oEP00oeH00?mF0000
0003003o]@0400000000jOl00>Wo000000<008co00D00000000io`00>Ol003Wo0000000300X0o`05
0000000i0?l0>@3o03T0o`0000000`1A0?l010000000D03o0500o`00000303<0o`05000000000_l0
00;o0002o`0000000`00Col01@0000000;3o002`o`00/?l0000000<00?oP00@00000003oJ@00ofT0
00000`0>o`001@000000POl0087o0021o`00000000<0i_l000D000000?o6003oaP00olH000000003
0?n<00040000003oK`00ofl0000000<0og0000D000000?n@003oT000oi00000000030?o>00050000
003Io`00fOl00=Wo000000000P1/o`001@0000000?lA003o4@00oa40000000<00?nF00D00000003S
o`00hol00>?o00000003001Uo`05000000090?l02@3o00T0o`0000000P1S0?l01@000000XP3o0:80
o`2R0?l0000000<0``3o00D000000<<0o`330?l0``3o000000030:80o`050000001S0?l0H`3o06<0
o`0000000P090?l01@00000006Go001Uo`00IOl0000000<00>?o00D00000003oUP00oiH00?nF0000
0003003o4@050000001/o`00K?l006co000000000P3Io`001@000000olh00?o>003ocP00000000<0
oi0000D000000?m`003oL000og00000000030?m_00050000003oS000oh`00?n<000000000P3oaP00
1@000000i_l00>Ko003Vo`00000000<0POl000D0000000ko000>o`003_l000000003003oJ@050000
0000on000?oP003oh00000000P00/?l01@00000004oo001?o`00Col0000000<000;o00D0000003<0
o`0c0?l0<`3o000000030500o`050000001A0?l0D@3o0540o`0000000P0i0?l01@0000002P3o00X0
o`0:0?l0000000<003Wo00D00000002<o`00S?l008co00000003003Yo`0500000000okD00?ne003o
]@0000000P00oeH01@0000000_l000;o0002o`00000000<0C_l000D0000008_o002;o`00Rol00000
00030;Ko00050000003<o`00c?l00<co000000000P3>o`001@000000_?l00;co002lo`00000000<0
VOl000D0000006So001Xo`00J?l00000000302_o000400000000oaP00?lH000000<00?mN00D00000
003oXP00oj800?nR00000003003ohP0500000000iol00>Oo003Wo`0000000`00_?l00`000000oooo
0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool0100000000;co002l
o`000003003Wo`0500000000on800?oR003ohP0000000`00oj801@0000000?mN003oGP00oeh00000
00<00?lH00@0000002_o000[o`0000000`1Xo`001@000000VOl009Wo002Io`00000000<0_?l000D0
00000<ko003>o`00c_l0000000030<co00040000002fo`00]_l0000000<0Rol000D0000004ko001>
o`00C_l00000000300;o000500000000oeH00?mF003oEP0000000`00okD0100000000>Wo003Yo`00
0003002<o`0500000000>Ol003Wo000io`0000000`0:0?l01@000000>@3o03T0o`0i0?l0000000<0
D@3o00@000000500o`1@0?l000000`0c0?l01@00000000;o0002o`000_l0000000<004oo00D00000
002`o`00/?l00;3o00000003003oh00400000000ofT00?mY000000<03_l000D00000087o0021o`00
POl0000000030>Ko00050000003oaP00olH00?o6000000000`3oS00010000000ofl00?m_00000003
0?m`00050000003oT000oi000?n@000000000`3ocP001@000000fOl00=Wo003Io`0000000080K?l0
00D00000003o4@00oa400?lA00000003003oUP0500000000hol00>?o003So`0000000`00IOl01@00
00002@3o00T0o`090?l000000080H`3o00D000000:80o`2R0?l0XP3o000000030<<0o`0500000033
0?l0``3o0<<0o`0000000`2R0?l01@000000H`3o06<0o`1S0?l0000000802@3o00D00000001Uo`00
IOl006Go00000003003So`0500000000oiH00?nF003oUP0000000`00oa401@000000K?l006co001/
o`0000000080fOl000D000000?o>003ocP00olh0000000030?n@00050000003oL000og000?m`0000
00000`3oK`001@000000oh`00?n<003oS00000000080olH000D000000>Ko003Vo`00i_l000000003
087o00050000000>o`003_l000ko000000000`00ofT01@0000000?oP003oh000on00000000800;3o
00D00000001?o`00Col004oo000000030002o`050000000c0?l0<`3o03<0o`0000000`1@0?l01@00
0000D@3o0540o`1A0?l000000080>@3o00D0000000X0o`0:0?l02P3o00000003000io`0500000000
S?l008co002<o`0000000`00jOl01@0000000?ne003o]@00okD0000000800?mF00D0000000;o0002
o`000_l00000000304ko00050000002;o`00Rol008_o000000000`2fo`001@000000c?l00<co003<
o`0000000080c_l000D000000;co002lo`00_?l00000000309Wo00050000001Xo`00J?l006So0000
00000`0[o`00100000000?lH003o60000003003oGP0500000000oj800?nR003oXP0000000`00on80
1@0000000>Oo003Wo`00iol0000000<00;co00<000000?ooo`3oool0103oool100000040oooo000P
0?ooo`030000003oool0oooo00D0oooo00@00000002lo`00_?l000000`00iol01@0000000?oR003o
hP00on80000000<00?nR00D00000003oGP00oeh00?mN00000003003o60040000000[o`00:ol00000
00<0J?l000D0000009Wo002Io`00VOl0000000030;co00050000003>o`00c_l00<ko000000000`3<
o`0010000000]_l00;Ko0000000308_o00050000001>o`00C_l004ko000000000`02o`001@000000
0?mF003oEP00oeH0000000<00?ne00@00000003Yo`00jOl000000`00S?l01@00000003Wo000io`00
>Ol0000000<02P3o00D0000003T0o`0i0?l0>@3o000000030540o`040000001@0?l0D03o000000<0
<`3o00D000000002o`000_l000;o00000003001?o`0500000000/?l00;3o002`o`0000000`00on00
100000000?mY003oJ@00000300ko000500000021o`00POl0087o000000000`3Vo`001@000000olH0
0?o6003oaP00000000<0oh`000@000000?m_003oK`0000000`3oL0001@000000oi000?n@003oT000
000000<0olh000D000000=Wo003Io`00fOl00000000206co000500000000oa400?lA003o4@000000
0`00oiH01@0000000>?o003So`00hol0000000<006Go00D0000000T0o`090?l02@3o0000000206<0
o`050000002R0?l0XP3o0:80o`0000000`330?l01@000000``3o0<<0o`330?l0000000<0XP3o00D0
000006<0o`1S0?l0H`3o0000000200T0o`0500000000IOl006Go001Uo`0000000`00hol01@000000
0?nF003oUP00oiH0000000<00?lA00D0000006co001/o`00K?l0000000020=Wo00050000003ocP00
olh00?o>000000000`3oT0001@000000og000?m`003oL000000000<0ofl000D000000?n<003oS000
oh`0000000020?o600050000003Vo`00i_l00>Ko000000000`21o`001@0000003_l000ko000>o`00
000000<00?mY00D00000003oh000on000?oP00000002002`o`0500000000Col004oo001?o`000000
0`000_l01@000000<`3o03<0o`0c0?l0000000<0D03o00D000000540o`1A0?l0D@3o0000000203T0
o`050000000:0?l02P3o00X0o`0000000`00>Ol01@00000008co002<o`00S?l0000000<00>Wo00D0
0000003o]@00okD00?ne00000002003oEP0500000002o`000_l000;o000000000`1>o`001@000000
Rol008_o002;o`00000000<0]_l000D000000<co003<o`00c?l0000000020<ko00050000002lo`00
_?l00;co000000000`2Io`001@000000J?l006So001Xo`00000000<0:ol000@00000003o6000oaP0
00000`00oeh01@0000000?nR003oXP00oj80000000<00?oR00D00000003Wo`00iol00>Oo00000003
002lo`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oo
ool00000P@0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool010000000
0?co003lo`000003003oc00500000000ohh00?n>003oSP0000000`00od/01@0000000?l8003o2000
o`P0000000<0=ol000@00000073o001`o`0000000`2No`001@000000__l00;ko002no`00000000<0
cOl000D000000<Wo0039o`00bOl0000000030;?o000400000029o`00ROl0000000<0C_l000D00000
00Co0004o`001?l000000003003oD@0500000000ojh00?n^003o[P0000000`00l_l01000000009Oo
002Go`0000030013o`05000000010?l00@3o0040o`0000000`0e0?l01@000000DP3o0580o`1B0?l0
000000<0F03o00@0000004@0o`140?l000000`0G0?l01@00000002co000/o`00;?l0000000<008Go
00D00000003^o`00k_l00>ko00000003003oWP0400000000obL00?lW000000<0COl000D000000;Wo
002io`00^Ol0000000030?oY00050000003oX@00oj400?nQ000000000`3oL`0010000000of400?mQ
000000030?m/00050000003oUP00oiH00?nF000000000`3of`001@000000a_l00<Ko0036o`000000
0080E?l000D00000003o:@00obT00?lY00000003003o[`0500000000c?l00<co003<o`0000000`00
D?l01@000000703o01`0o`0L0?l000000080M03o00D000000;40o`2a0?l0/@3o000000030=40o`05
0000003A0?l0d@3o0=40o`0000000`2a0?l01@000000M03o07@0o`1d0?l000000080703o00D00000
001@o`00D?l0053o00000003003<o`0500000000ojl00?n_003o[`0000000`00obT01@000000E?l0
05Co001Do`0000000080a_l000D000000?oK003of`00om/0000000030?nF00050000003oK000of`0
0?m/000000000`3oH@001@000000og<00?mc003oL`0000000080oj4000D000000?oY003oj@00onT0
000000030;Wo00050000001=o`00COl004go000000000`00obL01@0000000?nN003oWP00oih00000
00800>ko00D000000025o`00QOl008Go00000003000/o`050000000G0?l05`3o01L0o`0000000`14
0?l01@000000F03o05P0o`1H0?l000000080DP3o00D0000003D0o`0e0?l0=@3o000000030040o`05
00000000@ol004?o0013o`0000000`00Uol01@0000000?;o003bo`00l_l0000000800?n^00D00000
003oD@00oe400?mA0000000300Co00050000001>o`00C_l004ko000000000`29o`001@000000/ol0
0;?o002co`0000000080bOl000D000000<go003=o`00cOl0000000030;ko00050000002No`00W_l0
09ko000000000`1`o`0010000000=ol003Oo00000003003o200500000000od/00?m;003oB`000000
0`00ohh01@0000000?o<003oc000ol`0000000<00?co00<000000?ooo`3oool0103oool100000040
oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003lo`00o?l000000`00ol`01@000000
0?n>003oSP00ohh0000000<00?m;00D00000003o2000o`P00?l80000000303Oo00040000001`o`00
L?l0000000<0W_l000D000000;ko002no`00__l0000000030<go000500000039o`00bOl00<Wo0000
00000`2co`0010000000ROl008Wo0000000304ko000500000004o`001?l000Co000000000`00oe40
1@0000000?n^003o[P00ojh0000000<00?;o00@00000002Go`00Uol000000`00@ol01@0000000@3o
0040o`010?l0000000<0=@3o00D000000580o`1B0?l0DP3o0000000305P0o`04000000140?l0A03o
000000<05`3o00D00000000/o`00;?l002co000000030025o`0500000000k_l00>ko003^o`000000
0`00oih0100000000?lW003o9`00000304go00050000002io`00^Ol00;Wo000000000`3oj@001@00
0000oj400?nQ003oX@00000000<0og<000@000000?mQ003oH@0000000`3oK0001@000000oiH00?nF
003oUP00000000<0om/000D000000<Ko0036o`00a_l00000000205Co000500000000obT00?lY003o
:@0000000`00ojl01@0000000<co003<o`00c?l0000000<0053o00D0000001`0o`0L0?l0703o0000
000207@0o`050000002a0?l0/@3o0;40o`0000000`3A0?l01@000000d@3o0=40o`3A0?l0000000<0
/@3o00D0000007@0o`1d0?l0M03o0000000201`0o`0500000000D?l0053o001@o`0000000`00c?l0
1@0000000?n_003o[`00ojl0000000<00?lY00D0000005Co001Do`00E?l0000000020<Ko00050000
003of`00om/00?oK000000000`3oUP001@000000of`00?m/003oK000000000<0of4000D000000?mc
003oL`00og<0000000020?nQ00050000003oj@00onT00?oY000000000`2io`001@000000COl004go
001=o`00000000<00?lW00D00000003oWP00oih00?nN00000002003^o`0500000000QOl008Go0025
o`0000000`00;?l01@0000005`3o01L0o`0G0?l0000000<0A03o00D0000005P0o`1H0?l0F03o0000
00020580o`050000000e0?l0=@3o03D0o`0000000`010?l01@00000004?o0013o`00@ol0000000<0
09Oo00D00000003bo`00l_l00?;o00000002003o[P0500000000oe400?mA003oD@0000000`04o`00
1@000000C_l004ko001>o`00000000<0ROl000D000000;?o002co`00/ol0000000020<Wo00050000
003=o`00cOl00<go000000000`2no`001@000000W_l009ko002No`00000000<0L?l000@0000003Oo
000go`0000000`00o`P01@0000000?m;003oB`00od/0000000<00?n>00D00000003oc000ol`00?o<
00000003003lo`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?oo
o`050?ooo`0400000000o?l00?co000000<00?o<00D00000003oSP00ohh00?n>00000003003oB`05
00000000o`P00?l8003o200000000`0go`0010000000L?l0073o0000000309ko00050000002no`00
__l00;ko000000000`3=o`001@000000bOl00<Wo0039o`00000000<0/ol000@0000008Wo0029o`00
00000`1>o`001@0000001?l000Co0004o`00000000<00?mA00D00000003o[P00ojh00?n^00000003
003bo`0400000000Uol009Oo000000<004?o00D000000040o`010?l00@3o0000000303D0o`050000
001B0?l0DP3o0580o`0000000`1H0?l010000000A03o04@0o`00000301L0o`0500000000;?l002co
000/o`0000000`00QOl01@0000000>ko003^o`00k_l0000000<00?nN00@00000003o9`00obL00000
0`1=o`001@000000^Ol00;Wo002io`00000000<0onT000D000000?nQ003oX@00oj40000000030?mc
00040000003oH@00of40000000<0of`000D000000?nF003oUP00oiH0000000030?oK000500000036
o`00a_l00<Ko000000000P1Do`001@0000000?lY003o:@00obT0000000<00?n_00D00000003<o`00
c?l00<co00000003001@o`050000000L0?l0703o01`0o`0000000P1d0?l01@000000/@3o0;40o`2a
0?l0000000<0d@3o00D000000=40o`3A0?l0d@3o000000030;40o`050000001d0?l0M03o07@0o`00
00000P0L0?l01@000000053o001@o`00D?l0000000<00<co00D00000003o[`00ojl00?n_00000003
003o:@050000001Do`00E?l005Co000000000P36o`001@000000om/00?oK003of`00000000<0oiH0
00D000000?m/003oK000of`0000000030?mQ00050000003oL`00og<00?mc000000000P3oX@001@00
0000onT00?oY003oj@00000000<0^Ol000D0000004go001=o`00COl000000003003o9`0500000000
oih00?nN003oWP0000000P00k_l01@00000008Go0025o`00QOl0000000<002co00D0000001L0o`0G
0?l05`3o0000000304@0o`050000001H0?l0F03o05P0o`0000000P1B0?l01@000000=@3o03D0o`0e
0?l0000000<00@3o00D000000013o`00@ol004?o00000003002Go`0500000000l_l00?;o003bo`00
00000P00ojh01@0000000?mA003oD@00oe40000000<01?l000D0000004ko001>o`00C_l000000003
08Wo00050000002co`00/ol00;?o000000000P39o`001@000000cOl00<go003=o`00000000<0__l0
00D0000009ko002No`00W_l000000003073o00040000000go`00=ol0000000<00?l800D00000003o
B`00od/00?m;00000003003oSP0500000000ol`00?o<003oc00000000`00o?l00`000000oooo0?oo
o`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo0000084000001P3oool1
00000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003o]000ok@000000`00ogH0
1@0000000?lf003o=P00ocH0000000<02Ol000D0000004Go0015o`00AOl00000000307Wo00040000
002So`00Xol0000000<0_ol000D000000<_o003;o`00bol0000000030<Go00050000002]o`00[Ol0
0:go000000000`24o`0010000000B_l004[o0000000300;o000500000000oe400?mA003oD@000000
0`00oj/01@0000000?Oo003go`00mol0000000<009go00@00000001;o`00Bol000000`001?l01@00
0000<03o0300o`0`0?l0000000<0DP3o00D0000005d0o`1M0?l0G@3o000000030500o`040000000[
0?l0:`3o000000<0017o00D00000001Ro`00H_l006;o000000030034o`0500000000old00?o=003o
c@0000000`00oeP0100000007?l001co0000000308co00050000003`o`00l?l00?3o000000000`3o
_@001@000000oh000?n0003oP000000000<0oed000@000000?mG003oE`0000000`3oK@001@000000
oil00?nO003oW`00000000<0on`000D000000;3o002`o`00/?l00000000203_o000500000000od@0
0?m4003oA00000000`00olT01@0000000;Co002do`00]?l0000000<003[o00D0000002l0o`0_0?l0
;`3o0000000208D0o`05000000300?l0`03o0<00o`0000000`3O0?l01@000000g`3o0=l0o`3O0?l0
000000<0`03o00D0000008D0o`250?l0Q@3o0000000202l0o`0500000000>_l003[o000jo`000000
0`00]?l01@0000000?o9003ob@00olT0000000<00?m400D0000003_o000ko`00>ol0000000020;3o
00050000003ok000on`00?o/000000000`3oW`001@000000ofd00?m]003oK@00000000<0oeL000D0
00000?mM003oG@00oed0000000020?n000050000003o_@00okd00?nm000000000`3`o`001@000000
S?l008co002<o`00000000<07?l000D00000003oF000oeP00?mH00000002003oc@0500000000a?l0
0<Co0034o`0000000`00H_l01@000000017o000Ao`004Ol0000000<0:`3o00D000000500o`1@0?l0
D03o0000000205d0o`050000001B0?l0DP3o0580o`0000000`0`0?l01@00000000Co0004o`001?l0
000000<004_o00D00000002Mo`00WOl009go00000002003go`0500000000oj/00?n[003oZ`000000
0`00oe401@0000000_l000;o0002o`00000000<0B_l000D0000008Co0024o`00Q?l0000000020:go
000500000035o`00aOl00<Go000000000`3;o`001@000000_ol00;oo002oo`00000000<0Xol000@0
000007Wo001io`0000000`15o`001@0000002Ol000Wo0009o`00000000<00?lf00D00000003oMP00
ogH00?mf00000003003o]0030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000
oooo0?ooo`050?ooo`0400000000ok@00?nd000000<00?mf00D00000003o=P00ocH00?lf00000003
00Wo000500000015o`00AOl004Go000000000`1io`0010000000Xol00:?o000000030;oo00050000
003;o`00bol00<_o000000000`35o`001@000000[Ol00:go002]o`00000000<0Q?l000@0000004[o
001:o`0000000`02o`001@0000000?mA003oD@00oe40000000<00?n[00D00000003go`00mol00?Oo
00000003002Mo`0400000000Bol004_o000000<000Co00D000000300o`0`0?l0<03o000000030580
o`050000001M0?l0G@3o05d0o`0000000`1@0?l010000000:`3o02/0o`000003000Ao`0500000000
H_l006;o001Ro`0000000`00a?l01@0000000?o=003oc@00old0000000<00?mH00@0000001co000L
o`0000000`2<o`001@000000l?l00?3o003`o`00000000<0okd000D000000?n0003oP000oh000000
00030?mM00040000003oE`00oeL0000000<0ofd000D000000?nO003oW`00oil0000000030?o/0005
0000002`o`00/?l00;3o000000000P0ko`001@0000000?m4003oA000od@0000000<00?o900D00000
002do`00]?l00;Co00000003000jo`050000000_0?l0;`3o02l0o`0000000P250?l01@000000`03o
0<00o`300?l0000000<0g`3o00D000000=l0o`3O0?l0g`3o000000030<00o`05000000250?l0Q@3o
08D0o`0000000P0_0?l01@00000003[o000jo`00>_l0000000<00;Co00D00000003ob@00olT00?o9
00000003003oA0050000000ko`00>ol003_o000000000P2`o`001@000000on`00?o/003ok0000000
00<0oil000D000000?m]003oK@00ofd0000000030?mG00050000003oG@00oed00?mM000000000P3o
P0001@000000okd00?nm003o_@00000000<0l?l000D0000008co002<o`00S?l00000000301co0005
00000000oeP00?mH003oF00000000P00old01@0000000<Co0034o`00a?l0000000<006;o00D00000
000Ao`004Ol0017o0000000302/0o`050000001@0?l0D03o0500o`0000000P1M0?l01@000000DP3o
0580o`1B0?l0000000<0<03o00D000000004o`001?l000Co00000003001;o`0500000000WOl009go
002Mo`0000000P00mol01@0000000?n[003oZ`00oj/0000000<00?mA00D0000000;o0002o`000_l0
0000000304[o000500000024o`00Q?l008Co000000000P2]o`001@000000aOl00<Go0035o`000000
00<0bol000D000000;oo002oo`00_ol0000000030:?o00040000001io`00NOl0000000<0AOl000D0
000000Wo0009o`002Ol000000003003o=P0500000000ogH00?mf003oMP0000000`00ok@00`000000
oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool0100000000?nd
003o]0000003003oMP0500000000ocH00?lf003o=P0000000`09o`001@000000AOl004Go0015o`00
000000<0NOl000@000000:?o002So`0000000`2oo`001@000000bol00<_o003;o`00000000<0aOl0
00D000000:go002]o`00[Ol00000000308Co00040000001:o`00B_l0000000<00_l000D00000003o
D@00oe400?mA00000003003oZ`0500000000mol00?Oo003go`0000000`00WOl01000000004_o001;
o`0000030004o`050000000`0?l0<03o0300o`0000000`1B0?l01@000000G@3o05d0o`1M0?l00000
00<0D03o00@0000002/0o`0[0?l000000`004Ol01@00000006;o001Ro`00H_l0000000<00<Co00D0
0000003oc@00old00?o=00000003003oF0040000000Lo`007?l0000000<0S?l000D000000?3o003`
o`00l?l0000000030?nm00050000003oP000oh000?n0000000000`3oG@0010000000oeL00?mG0000
00030?m]00050000003oW`00oil00?nO000000000`3ok0001@000000/?l00;3o002`o`0000000080
>ol000D00000003oA000od@00?m400000003003ob@0500000000]?l00;Co002do`0000000`00>_l0
1@000000;`3o02l0o`0_0?l000000080Q@3o00D000000<00o`300?l0`03o000000030=l0o`050000
003O0?l0g`3o0=l0o`0000000`300?l01@000000Q@3o08D0o`250?l000000080;`3o00D00000000j
o`00>_l003[o00000003002do`0500000000olT00?o9003ob@0000000`00od@01@000000>ol003_o
000ko`0000000080/?l000D000000?o/003ok000on`0000000030?nO00050000003oK@00ofd00?m]
000000000`3oE`001@000000oed00?mM003oG@0000000080oh0000D000000?nm003o_@00okd00000
00030?3o00050000002<o`00S?l008co000000000`0Lo`001@0000000?mH003oF000oeP000000080
0?o=00D000000034o`00a?l00<Co00000003001Ro`05000000004Ol0017o000Ao`0000000`0[0?l0
1@000000D03o0500o`1@0?l000000080G@3o00D000000580o`1B0?l0DP3o000000030300o`050000
00001?l000Co0004o`0000000`00Bol01@00000009go002Mo`00WOl0000000800?Oo00D00000003o
Z`00oj/00?n[00000003003oD@0500000002o`000_l000;o000000000`1:o`001@000000Q?l008Co
0024o`0000000080[Ol000D000000<Go0035o`00aOl0000000030<_o00050000002oo`00_ol00;oo
000000000`2So`0010000000NOl007Wo0000000304Go000500000009o`002Ol000Wo000000000`00
ocH01@0000000?mf003oMP00ogH0000000<00?nd00<000000?ooo`3oool0103oool100000040oooo
000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?ooo`00803oool0
0`000000oooo0?ooo`050?ooo`0400000000oe`00?mL000000<00?lN00D0000001co000Lo`007?l0
0000000305Co000500000023o`00Pol008?o000000000`2Xo`0010000000_ol00;oo000000030<Oo
00050000002oo`00_ol00;oo000000000`2Uo`001@000000Nol007_o001ko`00000000<0@Ol000@0
0000003o1@00o`D000000`00oeH01@0000000?n^003o[P00ojh0000000<00?Oo00D00000002Po`00
X?l00:3o00000003001?o`04000000002Ol000Wo000000<0;@3o00D000000540o`1A0?l0D@3o0000
00030600o`050000001H0?l0F03o05P0o`0000000`0i0?l010000000103o00@0o`0000030015o`05
00000000X?l00:3o002Po`0000000`00ooH01@0000000?n5003oQ@00ohD0000000<00?lA00@00000
063o001Po`0000000`38o`001@000000om`00?oL003og000000000<0oi@000D000000?mS003oH`00
of<0000000030?m=00040000003oD@00oe40000000<0og8000D000000?n]003o[@00ojd000000003
0?oo00050000002Go`00Uol009Oo000000000P0Po`001@0000000?mP003oH000of00000000<00?oT
00D00000002Ko`00Vol009_o00000003000To`05000000130?l0@`3o04<0o`0000000P2E0?l01@00
0000c`3o0<l0o`3?0?l0000000<0k03o00D000000>`0o`3/0?l0k03o000000030<l0o`050000002E
0?l0U@3o09D0o`0000000P130?l01@00000002Co000To`009?l0000000<009_o00D00000003oi000
on@00?oT00000003003oH0050000000Po`008?l0023o000000000P2Go`001@000000ool00?oo003o
o`00000000<0ojd000D000000?mb003oLP00og80000000030?mA00050000003oC@00odd00?m=0000
00000P3oH`001@000000oi@00?nD003oU000000000<0om`000D000000<So0038o`00b?l000000003
063o000500000000oa400?lA003o4@0000000P00ohD01@0000000?of003omP00ooH0000000<00:3o
00D000000015o`00AOl004Go0000000300@0o`050000000i0?l0>@3o03T0o`0000000P1H0?l01@00
0000H03o0600o`1P0?l0000000<0D@3o00D0000002d0o`0]0?l0;@3o000000030009o`0500000000
Col004oo001?o`0000000P00X?l01@0000000?Oo003go`00mol0000000<00?n^00D00000003oEP00
oeH00?mF00000003003o1@0500000011o`00@Ol0047o000000000P1ko`001@000000YOl00:Go002U
o`00000000<0_ol000D000000<Oo0037o`00aol0000000030;oo00040000002Xo`00Z?l0000000<0
Pol000D0000005Co001Do`00E?l00000000301co000500000000oah00?lN003o7P0000000`00oe`0
0`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool01000
00000?mL003oG0000003003o7P050000000Lo`007?l001co000000000`1Do`001@000000Pol008?o
0023o`00000000<0Z?l000@000000;oo002oo`0000000`37o`001@000000_ol00;oo002oo`000000
00<0YOl000D0000007_o001ko`00Nol000000003047o000400000000o`D00?l5000000<00?mF00D0
0000003o[P00ojh00?n^00000003003go`0500000000X?l00:3o002Po`0000000`00Col010000000
00Wo0009o`00000302d0o`050000001A0?l0D@3o0540o`0000000`1P0?l01@000000F03o05P0o`1H
0?l0000000<0>@3o00@0000000@0o`040?l000000`00AOl01@0000000:3o002Po`00X?l0000000<0
0?of00D00000003oQ@00ohD00?n500000003003o4@040000001Po`00H?l0000000<0b?l000D00000
0?oL003og000om`0000000030?nD00050000003oH`00of<00?mS000000000`3oC@0010000000oe40
0?mA000000030?mb00050000003o[@00ojd00?n]000000000`3oo`001@000000Uol009Oo002Go`00
000000808?l000D00000003oH000of000?mP00000003003oi00500000000Vol009_o002Ko`000000
0`009?l01@000000@`3o04<0o`130?l000000080U@3o00D000000<l0o`3?0?l0c`3o000000030>`0
o`050000003/0?l0k03o0>`0o`0000000`3?0?l01@000000U@3o09D0o`2E0?l000000080@`3o00D0
0000000To`009?l002Co00000003002Ko`0500000000on@00?oT003oi00000000`00of001@000000
8?l0023o000Po`0000000080Uol000D000000?oo003oo`00ool0000000030?n]00050000003oLP00
og800?mb000000000`3oD@001@000000odd00?m=003oC@0000000080of<000D000000?nD003oU000
oi@0000000030?oL000500000038o`00b?l00<So000000000`1Po`001@0000000?lA003o4@00oa40
000000800?n500D00000003omP00ooH00?of00000003002Po`0500000000AOl004Go0015o`000000
0`040?l01@000000>@3o03T0o`0i0?l000000080F03o00D000000600o`1P0?l0H03o000000030540
o`050000000]0?l0;@3o02d0o`0000000`002Ol01@00000004oo001?o`00Col0000000800:3o00D0
0000003go`00mol00?Oo00000003003o[P0500000000oeH00?mF003oEP0000000`00o`D01@000000
@Ol0047o0011o`0000000080Nol000D000000:Go002Uo`00YOl0000000030;oo000500000037o`00
aol00<Oo000000000`2oo`0010000000Z?l00:So0000000308?o00050000001Do`00E?l005Co0000
00000`0Lo`001@0000000?lN003o7P00oah0000000<00?mL00<000000?ooo`3oool0103oool10000
0040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003oG000oe`000000`00oah01@00
00007?l001co000Lo`00000000<0E?l000D0000008?o0023o`00Pol0000000030:So00040000002o
o`00_ol0000000<0aol000D000000;oo002oo`00_ol0000000030:Go00050000001ko`00Nol007_o
000000000`11o`00100000000?l5003o1@000003003oEP0500000000ojh00?n^003o[P0000000`00
mol01@0000000:3o002Po`00X?l0000000<004oo00@000000009o`002Ol000000`0]0?l01@000000
D@3o0540o`1A0?l0000000<0H03o00D0000005P0o`1H0?l0F03o0000000303T0o`04000000040?l0
103o000000<004Go00D00000002Po`00X?l00:3o00000003003omP0500000000ohD00?n5003oQ@00
00000`00oa4010000000H?l0063o000000030<So00050000003og000om`00?oL000000000`3oU000
1@000000of<00?mS003oH`00000000<0odd000@000000?mA003oD@0000000`3oLP001@000000ojd0
0?n]003o[@00000000<0ool000D0000009Oo002Go`00Uol000000002023o000500000000of000?mP
003oH00000000`00on@01@00000009_o002Ko`00Vol0000000<002Co00D0000004<0o`130?l0@`3o
0000000209D0o`050000003?0?l0c`3o0<l0o`0000000`3/0?l01@000000k03o0>`0o`3/0?l00000
00<0c`3o00D0000009D0o`2E0?l0U@3o0000000204<0o`05000000009?l002Co000To`0000000`00
Vol01@0000000?oT003oi000on@0000000<00?mP00D00000023o000Po`008?l00000000209Oo0005
0000003oo`00ool00?oo000000000`3o[@001@000000og800?mb003oLP00000000<0oe4000D00000
0?m=003oC@00odd0000000020?mS00050000003oU000oi@00?nD000000000`3og0001@000000b?l0
0<So0038o`00000000<0H?l000D00000003o4@00oa400?lA00000002003oQ@0500000000ooH00?of
003omP0000000`00X?l01@00000004Go0015o`00AOl0000000<0103o00D0000003T0o`0i0?l0>@3o
0000000205P0o`050000001P0?l0H03o0600o`0000000`1A0?l01@000000;@3o02d0o`0]0?l00000
00<000Wo00D00000001?o`00Col004oo00000002002Po`0500000000mol00?Oo003go`0000000`00
ojh01@0000000?mF003oEP00oeH0000000<00?l500D00000047o0011o`00@Ol00000000207_o0005
0000002Uo`00YOl00:Go000000000`2oo`001@000000aol00<Oo0037o`00000000<0_ol000@00000
0:So002Xo`0000000`23o`001@000000E?l005Co001Do`00000000<07?l000D00000003o7P00oah0
0?lN00000003003oG0030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo
0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3o
ool0100000000?l5003o1@000003037o00050000001So`00Hol006?o000000000`2=o`001@000000
[?l00:co002/o`00000000<0__l000@000000<7o0031o`0000000`2eo`001@000000VOl009Wo002I
o`00000000<0KOl000D0000003Co000do`00=?l000000003003o4@0400000000of000?mP000000<0
0?ne00D00000003bo`00l_l00?;o00000003002Mo`0500000000Col004oo001?o`0000000`002_l0
10000000:`3o02/0o`0000030540o`050000001R0?l0HP3o0680o`0000000`1N0?l01@000000A03o
04@0o`140?l0000000<0503o00@00000000_o`00;ol000000`00Q?l01@0000000>Oo003Wo`00iol0
000000<00?n[00D00000003o>@00ocT00?li0000000303Oo00040000002Ro`00X_l0000000<0ood0
00D000000?n/003o[000oj`0000000030?m`00050000003oC000od`00?m<000000000`3o@@001000
0000oe400?mA000000030?mk00050000003o_P00okh00?nn000000000`3Wo`001@000000O?l007co
001lo`00000000800ol000D00000003oO@00ogd00?mm00000003003oo`0500000000POl0087o0021
o`0000000`003Ol01@000000EP3o05H0o`1F0?l000000080YP3o00D000000=d0o`3M0?l0g@3o0000
00030?X0o`050000003j0?l0nP3o0?X0o`0000000`3M0?l01@000000YP3o0:H0o`2V0?l000000080
EP3o00D00000000=o`003Ol000go000000030021o`0500000000ool00?oo003oo`0000000`00ogd0
1@0000000ol000?o0003o`0000000080O?l000D000000>Oo003Wo`00iol0000000030?nn00050000
003oN`00og/00?mk000000000`3oD@001@000000od400?m1003o@@0000000080od`000D000000?m`
003oL000og00000000030?n/00050000003oo@00ood00?om000000000`2Ro`001@000000=ol003Oo
000go`00000000800?li00D00000003oZ`00oj/00?n[00000003003Wo`0500000000Q?l008Co0024
o`0000000`00;ol01@000000503o01@0o`0D0?l000000080A03o00D0000005h0o`1N0?l0GP3o0000
00030680o`050000001A0?l0D@3o0540o`0000000`0[0?l01@00000000[o000:o`002_l000000080
04oo00D00000002Mo`00WOl009go00000003003bo`0500000000okD00?ne003o]@0000000`00of00
1@0000000?lA003o4@00oa4000000080=?l000D0000006go001]o`00KOl00000000309Wo00050000
002eo`00]Ol00;Go000000000`31o`0010000000__l00;ko000000030:co00050000002=o`00SOl0
08go000000000`1So`001@000000<Ol0037o000ao`00000000<00?l500<000000?ooo`3oool0103o
ool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003o1@00o`D000000`0a
o`001@000000Hol006?o001So`00000000<0SOl000D000000:co002/o`00[?l0000000030;ko0004
00000031o`00`Ol0000000<0]Ol000D0000009Wo002Io`00VOl00000000306go00050000000do`00
=?l003Co000000000`00oa40100000000?mP003oH0000003003o]@0500000000l_l00?;o003bo`00
00000`00WOl01@00000004oo001?o`00Col0000000<000[o00@0000002/0o`0[0?l000000`1A0?l0
1@000000HP3o0680o`1R0?l0000000<0GP3o00D0000004@0o`140?l0A03o0000000301@0o`040000
0000;ol002oo000000<008Co00D00000003Wo`00iol00>Oo00000003003oZ`0500000000ocT00?li
003o>@0000000`0go`0010000000X_l00:;o000000030?om00050000003o[000oj`00?n/00000000
0`3oL0001@000000od`00?m<003oC000000000<0od4000@000000?mA003oD@0000000`3oN`001@00
0000okh00?nn003o_P00000000<0iol000D0000007co001lo`00O?l00000000200?o000500000000
ogd00?mm003oO@0000000`00ool01@000000087o0021o`00POl0000000<000go00D0000005H0o`1F
0?l0EP3o000000020:H0o`050000003M0?l0g@3o0=d0o`0000000`3j0?l01@000000nP3o0?X0o`3j
0?l0000000<0g@3o00D000000:H0o`2V0?l0YP3o0000000205H0o`05000000003Ol000go000=o`00
00000`00POl01@0000000?oo003oo`00ool0000000<00?mm00D0000000?o0003o`000ol000000002
07co00050000003Wo`00iol00>Oo000000000`3o_P001@000000og/00?mk003oN`00000000<0oe40
00D000000?m1003o@@00od40000000020?m<00050000003oL000og000?m`000000000`3o[0001@00
0000ood00?om003oo@00000000<0X_l000D0000003Oo000go`00=ol000000002003o>@0500000000
oj/00?n[003oZ`0000000`00iol01@00000008Co0024o`00Q?l0000000<002oo00D0000001@0o`0D
0?l0503o0000000204@0o`050000001N0?l0GP3o05h0o`0000000`1R0?l01@000000D@3o0540o`1A
0?l0000000<0:`3o00D00000000:o`002_l000[o00000002001?o`0500000000WOl009go002Mo`00
00000`00l_l01@0000000?ne003o]@00okD0000000<00?mP00D00000003o4@00oa400?lA00000002
03Co00050000001]o`00KOl006go000000000`2Io`001@000000]Ol00;Go002eo`00000000<0`Ol0
00@000000;ko002no`0000000`2/o`001@000000SOl008go002=o`00000000<0Hol000D00000037o
000ao`00<Ol000000003003o1@030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`00
0000oooo0?ooo`050?ooo`0400000000o`D00?l5000000<0<Ol000D0000006?o001So`00Hol00000
000308go00050000002/o`00[?l00:co000000000`2no`0010000000`Ol00<7o000000030;Go0005
0000002Io`00VOl009Wo000000000`1]o`001@000000=?l003Co000do`00000000<00?lA00@00000
003oH000of0000000`00okD01@0000000?;o003bo`00l_l0000000<009go00D00000001?o`00Col0
04oo00000003000:o`040000000[0?l0:`3o000000<0D@3o00D000000680o`1R0?l0HP3o00000003
05h0o`05000000140?l0A03o04@0o`0000000`0D0?l01000000002oo000_o`0000030024o`050000
0000iol00>Oo003Wo`0000000`00oj/01@0000000?li003o>@00ocT0000000<0=ol000@000000:;o
002Ro`0000000`3oo@001@000000oj`00?n/003o[000000000<0og0000D000000?m<003oC000od`0
000000030?m100040000003oD@00oe40000000<0og/000D000000?nn003o_P00okh0000000030>Oo
00050000001lo`00O?l007co000000000P03o`001@0000000?mm003oO@00ogd0000000<00?oo00D0
00000021o`00POl0087o00000003000=o`050000001F0?l0EP3o05H0o`0000000P2V0?l01@000000
g@3o0=d0o`3M0?l0000000<0nP3o00D000000?X0o`3j0?l0nP3o000000030=d0o`050000002V0?l0
YP3o0:H0o`0000000P1F0?l01@00000000go000=o`003Ol0000000<0087o00D00000003oo`00ool0
0?oo00000003003oO@0500000003o`000ol000?o000000000P1lo`001@000000iol00>Oo003Wo`00
000000<0okh000D000000?mk003oN`00og/0000000030?mA00050000003o@@00od400?m100000000
0P3oC0001@000000og000?m`003oL000000000<0oj`000D000000?om003oo@00ood0000000030:;o
00050000000go`00=ol003Oo000000000P00ocT01@0000000?n[003oZ`00oj/0000000<00>Oo00D0
00000024o`00Q?l008Co00000003000_o`050000000D0?l0503o01@0o`0000000P140?l01@000000
GP3o05h0o`1N0?l0000000<0HP3o00D000000540o`1A0?l0D@3o0000000302/0o`05000000002_l0
00[o000:o`0000000P00Col01@00000009go002Mo`00WOl0000000<00?;o00D00000003o]@00okD0
0?ne00000003003oH00500000000oa400?lA003o4@0000000P0do`001@000000KOl006go001]o`00
000000<0VOl000D000000;Go002eo`00]Ol0000000030<7o00040000002no`00__l0000000<0[?l0
00D0000008go002=o`00SOl00000000306?o00050000000ao`00<Ol0037o000000000`00o`D00`00
0000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo00000840
00001P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@0000004Oo0017o`00
00000`1co`001@000000U_l009Ko002Fo`00000000<0[ol000D000000;_o002ko`00^ol000000003
0;Wo00040000002Xo`00Z?l0000000<0ROl000D0000005co001Lo`00G?l000000003027o00050000
0000ob<00?lS003o8`0000000`00og00100000000?o2003o`P000003003Yo`0500000000Uol009Oo
002Go`0000000`00Bol01@00000000Wo0009o`002Ol0000000<0:`3o00@000000500o`1@0?l00000
0`1S0?l01@000000H@3o0640o`1Q0?l0000000<0BP3o00D0000001l0o`0O0?l07`3o00000003000N
o`0400000000K_l006ko000000<00<co00D00000003obP00olX00?o:00000003003oG0050000000C
o`004ol001?o000000000`1oo`0010000000hol00>?o000000030?o600050000003oPP00oh800?n2
000000000`3oDP001@000000ocX00?lj003o>P00000000<0oc/000@000000?mF003oEP0000000`3o
R@001@000000om@00?oD003oe000000000<0cOl000D0000005ko001No`00G_l000000002003o6`05
00000000oi/00?nK003oV`0000000`00hol01@00000006Oo001Wo`00Iol0000000<0203o00D00000
06X0o`1Z0?l0JP3o000000020;H0o`050000003/0?l0k03o0>`0o`0000000`3o0?P01@000000o`3h
0?l0n03o0?P0000000<0k03o00D000000;H0o`2f0?l0]P3o0000000206X0o`05000000080?l0203o
00P0o`0000000`00Iol01@0000000>?o003So`00hol0000000<00?nK00D00000003o6`00oa/00?lK
0000000205ko00050000003=o`00cOl00<go000000000`3oe0001@000000ohT00?n9003oR@000000
00<0oeH000D000000?lk003o>`00oc/0000000020?lj00050000003oDP00oe800?mB000000000`3o
PP001@000000olH00?o6003oaP00000000<0hol000D0000007oo001oo`00Ool00000000201?o0005
00000000oe`00?mL003oG00000000`00olX01@0000000<co003<o`00c?l0000000<006ko00D00000
000No`007_l001ko0000000201l0o`050000001:0?l0BP3o04X0o`0000000`1Q0?l01@000000H`3o
06<0o`1S0?l0000000<0D03o00D0000002/0o`0[0?l0:`3o000000020009o`0500000000Bol004_o
001;o`0000000`00Uol01@0000000>Wo003Yo`00jOl0000000<00?o200D00000003oL000og000?m`
00000002003o8`050000000Qo`008Ol0027o000000000`1Lo`001@000000ROl008Wo0029o`000000
00<0Z?l000@000000;Wo002io`0000000`2ko`001@000000[ol00:oo002_o`00000000<0U_l000D0
000007?o001co`00Lol00000000304Oo00030000003oool0oooo00@0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooo`0400000017o`00Aol0000000<0Lol000D0000009Ko002Fo`00
U_l0000000030:oo00050000002ko`00^ol00;_o000000000`2io`0010000000Z?l00:So00000003
08Wo00050000001Lo`00G?l005co000000000`0Qo`001@0000000?lS003o8`00ob<0000000<00?m`
00@00000003o`P00ol8000000`00jOl01@00000009Oo002Go`00Uol0000000<004_o00D000000009
o`002Ol000Wo0000000302/0o`040000001@0?l0D03o000000<0H`3o00D000000640o`1Q0?l0H@3o
0000000304X0o`050000000O0?l07`3o01l0o`0000000`007_l01000000006ko001^o`000003003<
o`0500000000olX00?o:003obP0000000`00oe`01@0000004ol001?o000Co`00000000<0Ool000@0
00000>?o003So`0000000`3oaP001@000000oh800?n2003oPP00000000<0oe8000D000000?lj003o
>P00ocX0000000030?lk00040000003oEP00oeH0000000<0ohT000D000000?oD003oe000om@00000
00030<go00050000001No`00G_l005ko000000000P00oa/01@0000000?nK003oV`00oi/0000000<0
0>?o00D00000001Wo`00Iol006Oo0000000300P0o`050000001Z0?l0JP3o06X0o`0000000P2f0?l0
1@000000k03o0>`0o`3/0?l0000000<0o`3h00D000000?l0n03o0?P0o`3h000000030>`0o`050000
002f0?l0]P3o0;H0o`0000000P1Z0?l01@000000203o00P0o`080?l0000000<006Oo00D00000003S
o`00hol00>?o00000003003oV`0500000000oa/00?lK003o6`0000000P1No`001@000000cOl00<go
003=o`00000000<0om@000D000000?n9003oR@00ohT0000000030?mF00050000003o>`00oc/00?lk
000000000P3o>P001@000000oe800?mB003oDP00000000<0oh8000D000000?o6003oaP00olH00000
00030>?o00050000001oo`00Ool007oo000000000P0Co`001@0000000?mL003oG000oe`0000000<0
0?o:00D00000003<o`00c?l00<co00000003001^o`05000000007_l001ko000No`0000000P0O0?l0
1@000000BP3o04X0o`1:0?l0000000<0H@3o00D0000006<0o`1S0?l0H`3o000000030500o`050000
000[0?l0:`3o02/0o`0000000P002Ol01@00000004_o001;o`00Bol0000000<009Oo00D00000003Y
o`00jOl00>Wo00000003003o`P0500000000og000?m`003oL00000000P00ob<01@0000008Ol0027o
000Qo`00000000<0G?l000D0000008Wo0029o`00ROl0000000030:So00040000002io`00^Ol00000
00<0^ol000D000000:oo002_o`00[ol00000000309Ko00050000001co`00Lol007?o000000000`17
o`000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo0P0000060?ooool00000P@000005
0?ooo`8000000@3oool00200oooo00<000000?ooo`3oool01@3oool010000000P?l0083o00000003
09co00050000002^o`00[_l00:ko000000000`2do`001@000000[?l00:co002/o`00000000<0Uol0
00@0000007Go001eo`0000000`15o`001@0000002_l000[o000:o`00000000<00?li00D00000003o
Q@00ohD00?n500000003003oe00400000000f_l00=[o000000<008co00D000000013o`00@ol004?o
000000030004o`050000000]0?l0;@3o02d0o`0000000`1A0?l010000000H`3o06<0o`0000030680
o`050000001=0?l0C@3o04d0o`0000000`0V0?l01@00000001Co000Do`005?l0000000<005oo00@0
0000002ho`00^?l000000`00on<01@0000000?mh003oN000ogP0000000<00?l;00D00000063o001P
o`00H?l0000000030<Go00040000003oh000on00000000<0oiD000D000000?mM003oG@00oed00000
00030?lj00050000003o;P00obh00?l^000000000`3o>`0010000000of000?mP000000030?nL0005
0000003ok@00ond00?o]000000000`2_o`001@000000?_l003ko000no`00000000800?ll00D00000
003o^`00ok/00?nk000000030035o`0500000000COl004go001=o`0000000`0O0?l01@000000O@3o
07d0o`1m0?l000000080a`3o00D000000?X0o`3j0?l0nP3o000000030?l0j`050000003o0>/0o`3[
0?l0j`0000000`3j0?l01@000000a`3o0<L0o`370?l000000080O@3o00D0000001l0o`0O0?l07`3o
00000003001=o`0500000000aOl00<Go0035o`0000000`00ok/01@0000000?ll003o?000oc`00000
0080?_l000D000000:oo002_o`00[ol0000000030?o]00050000003oW000oi`00?nL000000000`3o
H0001@000000oc/00?lk003o>`0000000080obh000D000000?lj003o>P00ocX0000000030?mM0005
0000003oU@00oiD00?nE000000000`3oh0001@000000aOl00<Go0035o`0000000080H?l000D00000
003o2`00o`/00?l;00000003003oN00500000000on<00?oS003oh`0000000`00^?l01@00000005oo
001Oo`00Gol00000008001Co00D0000002H0o`0V0?l09P3o0000000304d0o`050000001R0?l0HP3o
0680o`0000000`1S0?l01@000000D@3o0540o`1A0?l000000080;@3o00D000000004o`001?l000Co
000000030013o`0500000000S?l008co002<o`0000000`00f_l01@0000000?oD003oe000om@00000
00800?n500D00000003o>@00ocT00?li0000000300[o000500000015o`00AOl004Go000000000`1e
o`0010000000Uol009Oo000000030:co00050000002do`00]?l00;Co000000000`2^o`001@000000
W?l009co002Lo`00000000<0P?l000<000000?ooo`3oool0103oool100000040oooo000P0?ooo`03
0000003oool0oooo00D0oooo00@00000083o0020o`0000000`2Lo`001@000000[_l00:ko002^o`00
000000<0]?l000D000000:co002/o`00[?l00000000309Oo00040000001eo`00MOl0000000<0AOl0
00D0000000[o000:o`002_l000000003003o>@0500000000ohD00?n5003oQ@0000000`00om@01000
00000=[o003Jo`000003002<o`0500000000@ol004?o0013o`0000000`001?l01@000000;@3o02d0
o`0]0?l0000000<0D@3o00@0000006<0o`1S0?l000000`1R0?l01@000000C@3o04d0o`1=0?l00000
00<09P3o00D00000000Do`005?l001Co00000003001Oo`0400000000^?l00;So000000<00?oS00D0
0000003oN000ogP00?mh00000003003o2`050000001Po`00H?l0063o000000000`35o`0010000000
on000?oP000000030?nE00050000003oG@00oed00?mM000000000`3o>P001@000000obh00?l^003o
;P00000000<0oc/000@000000?mP003oH00000000`3oW0001@000000ond00?o]003ok@00000000<0
[ol000D0000003ko000no`00?_l000000002003o?00500000000ok/00?nk003o^`0000000`00aOl0
1@00000004go001=o`00COl0000000<07`3o00D0000007d0o`1m0?l0O@3o000000020<L0o`050000
003j0?l0nP3o0?X0o`0000000`3o0>/01@000000o`3[0?l0j`3o0>/0000000<0nP3o00D000000<L0
o`370?l0a`3o0000000207d0o`050000000O0?l07`3o01l0o`0000000`00COl01@0000000<Go0035
o`00aOl0000000<00?nk00D00000003o?000oc`00?ll0000000203ko00050000002_o`00[ol00:oo
000000000`3ok@001@000000oi`00?nL003oW000000000<0of0000D000000?lk003o>`00oc/00000
00020?l^00050000003o>P00ocX00?lj000000000`3oG@001@000000oiD00?nE003oU@00000000<0
on0000D000000<Go0035o`00aOl000000002063o000500000000o`/00?l;003o2`0000000`00ogP0
1@0000000?oS003oh`00on<0000000<00;So00D00000001Oo`00Gol005oo00000002000Do`050000
000V0?l09P3o02H0o`0000000`1=0?l01@000000HP3o0680o`1R0?l0000000<0H`3o00D000000540
o`1A0?l0D@3o0000000202d0o`05000000001?l000Co0004o`0000000`00@ol01@00000008co002<
o`00S?l0000000<00=[o00D00000003oe000om@00?oD00000002003oQ@0500000000ocT00?li003o
>@0000000`0:o`001@000000AOl004Go0015o`00000000<0MOl000@0000009Oo002Go`0000000`2/
o`001@000000]?l00;Co002do`00000000<0[_l000D0000009co002Lo`00W?l000000003083o0003
0000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`040000
0020o`00P?l0000000<0W?l000D000000:ko002^o`00[_l0000000030;Co00050000002/o`00[?l0
0:co000000000`2Go`0010000000MOl007Go0000000304Go00050000000:o`002_l000[o00000000
0`00ocT01@0000000?n5003oQ@00ohD0000000<00?oD00@00000003Jo`00f_l000000`00S?l01@00
000004?o0013o`00@ol0000000<000Co00D0000002d0o`0]0?l0;@3o000000030540o`040000001S
0?l0H`3o000000<0HP3o00D0000004d0o`1=0?l0C@3o0000000302H0o`05000000005?l001Co000D
o`0000000`00Gol0100000000;So002ho`000003003oh`0500000000ogP00?mh003oN00000000`00
o`/01@000000H?l0063o001Po`00000000<0aOl000@000000?oP003oh00000000`3oU@001@000000
oed00?mM003oG@00000000<0ocX000D000000?l^003o;P00obh0000000030?lk00040000003oH000
of00000000<0oi`000D000000?o]003ok@00ond0000000030:oo00050000000no`00?_l003ko0000
00000P00oc`01@0000000?nk003o^`00ok/0000000<00<Go00D00000001=o`00COl004go00000003
01l0o`050000001m0?l0O@3o07d0o`0000000P370?l01@000000nP3o0?X0o`3j0?l0000000<0o`3[
00D000000?l0j`3o0>/0o`3[000000030?X0o`05000000370?l0a`3o0<L0o`0000000P1m0?l01@00
00007`3o01l0o`0O0?l0000000<004go00D000000035o`00aOl00<Go00000003003o^`0500000000
oc`00?ll003o?00000000P0no`001@000000[ol00:oo002_o`00000000<0ond000D000000?nL003o
W000oi`0000000030?mP00050000003o>`00oc/00?lk000000000P3o;P001@000000ocX00?lj003o
>P00000000<0oed000D000000?nE003oU@00oiD0000000030?oP000500000035o`00aOl00<Go0000
00000P1Po`001@0000000?l;003o2`00o`/0000000<00?mh00D00000003oh`00on<00?oS00000003
002ho`0500000000Gol005oo001Oo`0000000P005?l01@0000009P3o02H0o`0V0?l0000000<0C@3o
00D000000680o`1R0?l0HP3o0000000306<0o`050000001A0?l0D@3o0540o`0000000P0]0?l01@00
000000Co0004o`001?l0000000<004?o00D00000002<o`00S?l008co00000003003Jo`0500000000
om@00?oD003oe00000000P00ohD01@0000000?li003o>@00ocT0000000<02_l000D0000004Go0015
o`00AOl00000000307Go00040000002Go`00Uol0000000<0[?l000D000000;Co002do`00]?l00000
00030:ko00050000002Lo`00W?l009co000000000`20o`000`000000oooo0?ooo`040?ooo`400000
0@3oool00200oooo00<000000?ooo`3oool01@3ooooo0000084000001P3oool100000040oooo000P
0?ooo`030000003oool0oooo00D0oooo00@0000009oo002Oo`0000000`2Zo`001@000000ZOl00:Wo
002Yo`00000000<0Vol000D00000087o0021o`00POl00000000305_o00040000000Zo`00:_l00000
00<00?lA00D00000003oE@00oeD00?mE00000003003oWP0500000000on/00?o[003oj`0000000`00
aol01000000007go001mo`000003000io`05000000010?l00@3o0040o`0000000`0`0?l01@000000
D@3o0540o`1A0?l0000000<0HP3o00@000000640o`1Q0?l000000`1=0?l01@000000:03o02P0o`0X
0?l0000000<000ko00D00000001Fo`00E_l005Ko00000003002[o`0400000000ooD00?oe000000<0
0?n>00D00000003o8`00ob<00?lS0000000304Ko00050000002[o`00Zol00:_o000000000`3on000
10000000ojT00?nY000000030?mZ00050000003o?P00och00?ln000000000`3o:0001@000000obT0
0?lY003o:@00000000<0od0000@000000?m_003oK`0000000`3o/`001@000000mOl00?Go003eo`00
000000<0S_l000D0000001_o000Ko`006ol000000002003oGP0500000000om`00?oL003og0000000
0`00Yol01@00000003;o000bo`00<_l0000000<0=P3o00D000000940o`2A0?l0T@3o000000020=L0
o`050000003o0?L0o`3g0?l0m`0000000`3o0=h01@000000o`3N0?l0gP3o0=h0000000<0o`3g00D0
00000=L0o`3G0?l0e`3o000000020940o`050000000f0?l0=P3o03H0o`0000000`00<_l01@000000
0:Oo002Wo`00Yol0000000<00?oL00D00000003oGP00oeh00?mN0000000201_o00050000002>o`00
S_l008ko000000000`3eo`001@000000ok<00?nc003o/`00000000<0ofl000D000000?m0003o@000
od00000000020?lY00050000003o:000obP00?lX000000000`3o?P001@000000ofX00?mZ003oJP00
000000<0ojT000D000000?oh003on000ooP0000000020:_o000500000016o`00A_l004Ko00000000
0`00ob<01@0000000?n>003oSP00ohh0000000<00?oe00D00000002[o`00Zol00:_o00000002001F
o`05000000003_l000ko000>o`0000000`0X0?l01@000000C@3o04d0o`1=0?l0000000<0H@3o00D0
00000680o`1R0?l0HP3o000000020540o`050000000`0?l0<03o0300o`0000000`010?l01@000000
03Wo000io`00>Ol0000000<007go00D000000037o`00aol00<Oo00000002003oj`0500000000oih0
0?nN003oWP0000000`00oeD01@0000000?lA003o4@00oa40000000<0:_l000@0000005_o001Ko`00
00000`21o`001@000000Vol009_o002Ko`00000000<0ZOl000D000000:[o002Zo`00Z_l000000003
09oo00030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oo
o`040000002Oo`00Wol0000000<0Z_l000D000000:Wo002Yo`00ZOl00000000309_o000500000021
o`00POl0087o000000000`1Ko`0010000000:_l002[o00000003003o4@0500000000oeD00?mE003o
E@0000000`00oih01@0000000?o[003oj`00on/0000000<00<Oo00@00000001mo`00OOl000000`00
>Ol01@0000000@3o0040o`010?l0000000<0<03o00D000000540o`1A0?l0D@3o000000030680o`04
0000001Q0?l0H@3o000000<0C@3o00D0000002P0o`0X0?l0:03o00000003000>o`0500000000E_l0
05Ko001Fo`0000000`00Zol0100000000?oe003om@000003003oSP0500000000ob<00?lS003o8`00
00000`16o`001@000000Zol00:_o002[o`00000000<0ooP000@000000?nY003oZ@0000000`3oJP00
1@000000och00?ln003o?P00000000<0obP000D000000?lY003o:@00obT0000000030?m000040000
003oK`00ofl0000000<0ok<000D000000?Go003eo`00mOl00000000308ko00050000000Ko`006ol0
01_o000000000P00oeh01@0000000?oL003og000om`0000000<00:Oo00D00000000bo`00<_l003;o
0000000303H0o`050000002A0?l0T@3o0940o`0000000P3G0?l01@000000o`3g0?l0m`3o0?L00000
00<0o`3N00D000000?l0gP3o0=h0o`3N000000030?l0m`050000003G0?l0e`3o0=L0o`0000000P2A
0?l01@000000=P3o03H0o`0f0?l0000000<003;o00D00000002Wo`00Yol00:Oo00000003003og005
00000000oeh00?mN003oGP0000000P0Ko`001@000000S_l008ko002>o`00000000<0mOl000D00000
0?nc003o/`00ok<0000000030?m_00050000003o@000od000?m0000000000P3o:@001@000000obP0
0?lX003o:000000000<0och000D000000?mZ003oJP00ofX0000000030?nY00050000003on000ooP0
0?oh000000000P2[o`001@000000A_l004Ko0016o`00000000<00?lS00D00000003oSP00ohh00?n>
00000003003om@0500000000Zol00:_o002[o`0000000P00E_l01@00000000ko000>o`003_l00000
00<0:03o00D0000004d0o`1=0?l0C@3o000000030640o`050000001R0?l0HP3o0680o`0000000P1A
0?l01@000000<03o0300o`0`0?l0000000<00@3o00D00000000io`00>Ol003Wo00000003001mo`05
00000000aol00<Oo0037o`0000000P00on/01@0000000?nN003oWP00oih0000000<00?mE00D00000
003o4@00oa400?lA0000000302[o00040000001Ko`00Fol0000000<0POl000D0000009_o002Ko`00
Vol0000000030:Wo00050000002Zo`00Z_l00:[o000000000`2Oo`000`000000oooo0?ooo`040?oo
o`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool010000000Wol009oo000000030:[o
00050000002Yo`00ZOl00:Wo000000000`2Ko`001@000000POl0087o0021o`00000000<0Fol000@0
000002[o000Zo`0000000`00oa401@0000000?mE003oE@00oeD0000000<00?nN00D00000003oj`00
on/00?o[000000030037o`0400000000OOl007go000000<003Wo00D000000040o`010?l00@3o0000
00030300o`050000001A0?l0D@3o0540o`0000000`1R0?l010000000H@3o0640o`00000304d0o`05
0000000X0?l0:03o02P0o`0000000`003_l01@00000005Ko001Fo`00E_l0000000<00:_o00@00000
003om@00ooD000000`00ohh01@0000000?lS003o8`00ob<0000000<0A_l000D000000:_o002[o`00
Zol0000000030?oh00040000003oZ@00ojT0000000<0ofX000D000000?ln003o?P00och000000003
0?lX00050000003o:@00obT00?lY000000000`3o@00010000000ofl00?m_000000030?nc00050000
003eo`00mOl00?Go000000000`2>o`001@0000006ol001_o000Ko`00000000800?mN00D00000003o
g000om`00?oL00000003002Wo`0500000000<_l003;o000bo`0000000`0f0?l01@000000T@3o0940
o`2A0?l000000080e`3o00D000000?l0m`3o0?L0o`3g000000030?l0gP050000003o0=h0o`3N0?l0
gP0000000`3o0?L01@000000e`3o0=L0o`3G0?l000000080T@3o00D0000003H0o`0f0?l0=P3o0000
0003000bo`0500000000Yol00:Oo002Wo`0000000`00om`01@0000000?mN003oGP00oeh000000080
6ol000D0000008ko002>o`00S_l0000000030?Go00050000003o/`00ok<00?nc000000000`3oK`00
1@000000od000?m0003o@00000000080obT000D000000?lX003o:000obP0000000030?ln00050000
003oJP00ofX00?mZ000000000`3oZ@001@000000ooP00?oh003on00000000080Zol000D0000004Ko
0016o`00A_l000000003003o8`0500000000ohh00?n>003oSP0000000`00ooD01@0000000:_o002[
o`00Zol00000008005Ko00D00000000>o`003_l000ko0000000302P0o`050000001=0?l0C@3o04d0
o`0000000`1Q0?l01@000000HP3o0680o`1R0?l000000080D@3o00D000000300o`0`0?l0<03o0000
00030040o`0500000000>Ol003Wo000io`0000000`00OOl01@0000000<Oo0037o`00aol000000080
0?o[00D00000003oWP00oih00?nN00000003003oE@0500000000oa400?lA003o4@0000000`0Zo`00
10000000Fol005_o00000003087o00050000002Ko`00Vol009_o000000000`2Yo`001@000000Z_l0
0:[o002Zo`00000000<0Wol000<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000
003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?ooo`00803oool00`000000oooo0?oo
o`050?ooo`040000002Po`00X?l0000000<0V?l000D0000008Go0025o`00QOl00000000306Ko0005
0000000lo`00??l003co000000000`09o`00100000000?lb003o<P000003003oM@0500000000ok`0
0?nl003o_00000000`00nOl01@0000000;7o002ao`00/Ol0000000<006_o00@00000000/o`00;?l0
00000`0:0?l01@000000=@3o03D0o`0e0?l0000000<0DP3o00D000000600o`1P0?l0H03o00000003
05h0o`040000001:0?l0BP3o000000<09P3o00D00000000>o`003_l000ko00000003001Co`050000
0000Y?l00:Co002To`0000000`00ool0100000000?nL003oW0000003003o=@050000000bo`00<_l0
03;o000000000`2Fo`001@000000lol00??o003co`00000000<0ok/000@000000?mg003oM`000000
0`3oA@001@000000obL00?lW003o9`00000000<0oad000D000000?lZ003o:P00obX0000000030?m<
00040000003oQ000oh@0000000<0oll000D000000=Co003Do`00e?l00000000306[o000500000000
o`P00?l8003o200000000P00oh801@0000000?on003ooP00ooh0000000<008So00D00000000Go`00
5ol001Oo0000000304h0o`050000002T0?l0Y03o0:@0o`0000000P3X0?l01@000000o`3Y0?l0j@3o
0>T0000000<0o`3A00D000000?l0d@3o0=40o`3A000000030?l0j@050000003X0?l0j03o0>P0o`00
00000P2T0?l01@000000CP3o04h0o`1>0?l0000000<001Oo00D000000028o`00R?l008So00000003
003ooP0500000000oh800?n2003oPP0000000P00o`P01@000000J_l006[o001Zo`00000000<0e?l0
00D000000?o?003oc`00oll0000000030?n400050000003oC000od`00?m<000000000P3o:P001@00
0000oad00?lM003o7@00000000<0obL000D000000?m5003oA@00odD0000000030?mg00050000003o
^`00ok/00?nk000000000P3co`001@000000U_l009Ko002Fo`00000000<0<_l000D00000003o=@00
ocD00?le00000003003oW00500000000ool00?oo003oo`0000000P00Y?l01@00000005?o001Co`00
Dol0000000<000ko00D0000002H0o`0V0?l09P3o0000000304X0o`050000001N0?l0GP3o05h0o`00
00000P1P0?l01@000000DP3o0580o`1B0?l0000000<0=@3o00D0000000X0o`0:0?l02P3o00000003
000/o`0500000000Jol006_o001[o`0000000P00/Ol01@0000000?Wo003io`00nOl0000000<00?nl
00D00000003oM@00ogD00?me00000003003o<P0400000009o`002Ol0000000<0??l000D0000006Ko
001Vo`00I_l00000000308Go00050000002Ho`00V?l009So000000000`2Po`000`000000oooo0?oo
o`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool010000000X?l00:3o0000
000309So000500000025o`00QOl008Go000000000`1Vo`001@000000??l003co000lo`00000000<0
2Ol000@00000003o<P00oc8000000`00ogD01@0000000?nl003o_000ok`0000000<00?Wo00D00000
002ao`00/Ol00;7o00000003001[o`0400000000;?l002co000000<02P3o00D0000003D0o`0e0?l0
=@3o000000030580o`050000001P0?l0H03o0600o`0000000`1N0?l010000000BP3o04X0o`000003
02H0o`05000000003_l000ko000>o`0000000`00Dol01@0000000:Co002To`00Y?l0000000<00?oo
00@00000003oW000oi`000000`00ocD01@000000<_l003;o000bo`00000000<0U_l000D000000??o
003co`00lol0000000030?nk00040000003oM`00ogL0000000<0odD000D000000?lW003o9`00obL0
000000030?lM00050000003o:P00obX00?lZ000000000`3oC00010000000oh@00?n4000000030?o?
00050000003Do`00e?l00=Co000000000`1Zo`001@0000000?l8003o2000o`P0000000800?n200D0
0000003ooP00ooh00?on000000030028o`05000000005ol001Oo000Go`0000000`1>0?l01@000000
Y03o0:@0o`2T0?l000000080j03o00D000000?l0j@3o0>T0o`3Y000000030?l0d@050000003o0=40
o`3A0?l0d@0000000`3o0>T01@000000j03o0>P0o`3X0?l000000080Y03o00D0000004h0o`1>0?l0
CP3o00000003000Go`0500000000R?l008So0028o`0000000`00ooh01@0000000?n2003oPP00oh80
000000800?l800D0000006[o001Zo`00J_l0000000030=Co00050000003oc`00oll00?o?00000000
0`3oQ0001@000000od`00?m<003oC00000000080obX000D000000?lM003o7@00oad0000000030?lW
00050000003oA@00odD00?m5000000000`3oM`001@000000ok/00?nk003o^`0000000080lol000D0
000009Ko002Fo`00U_l00000000303;o000500000000ocD00?le003o=@0000000`00oi`01@000000
0?oo003oo`00ool0000000800:Co00D00000001Co`00Dol005?o00000003000>o`050000000V0?l0
9P3o02H0o`0000000`1:0?l01@000000GP3o05h0o`1N0?l000000080H03o00D000000580o`1B0?l0
DP3o0000000303D0o`050000000:0?l02P3o00X0o`0000000`00;?l01@00000006_o001[o`00Jol0
000000800;7o00D00000003io`00nOl00?Wo00000003003o_00500000000ogD00?me003oM@000000
0`00oc80100000002Ol000Wo0000000303co00050000001Vo`00I_l006Ko000000000`25o`001@00
0000V?l009So002Ho`00000000<0X?l000<000000?ooo`3oool0103oool100000040oooo000P0?oo
o`030000003oool0oooo00D0oooo00@000000:3o002Po`0000000`2Ho`001@000000QOl008Go0025
o`00000000<0I_l000D0000003co000lo`00??l00000000300Wo000400000000oc800?lb000000<0
0?me00D00000003o_000ok`00?nl00000003003io`0500000000/Ol00;7o002ao`0000000`00Jol0
1000000002co000/o`00000300X0o`050000000e0?l0=@3o03D0o`0000000`1B0?l01@000000H03o
0600o`1P0?l0000000<0GP3o00@0000004X0o`1:0?l000000`0V0?l01@00000000ko000>o`003_l0
000000<005?o00D00000002To`00Y?l00:Co00000003003oo`0400000000oi`00?nL000000<00?le
00D0000003;o000bo`00<_l00000000309Ko00050000003co`00lol00??o000000000`3o^`001000
0000ogL00?mg000000030?m500050000003o9`00obL00?lW000000000`3o7@001@000000obX00?lZ
003o:P00000000<0od`000@000000?n4003oQ00000000`3oc`001@000000e?l00=Co003Do`000000
00<0J_l000D00000003o2000o`P00?l800000002003oPP0500000000ooh00?on003ooP0000000`00
R?l01@00000001Oo000Go`005ol0000000<0CP3o00D000000:@0o`2T0?l0Y03o000000020>P0o`05
0000003o0>T0o`3Y0?l0j@0000000`3o0=401@000000o`3A0?l0d@3o0=40000000<0o`3Y00D00000
0>P0o`3X0?l0j03o000000020:@0o`050000001>0?l0CP3o04h0o`0000000`005ol01@00000008So
0028o`00R?l0000000<00?on00D00000003oPP00oh800?n200000002003o20050000001Zo`00J_l0
06[o000000000`3Do`001@000000oll00?o?003oc`00000000<0oh@000D000000?m<003oC000od`0
000000020?lZ00050000003o7@00oad00?lM000000000`3o9`001@000000odD00?m5003oA@000000
00<0ogL000D000000?nk003o^`00ok/0000000020??o00050000002Fo`00U_l009Ko000000000`0b
o`001@0000000?le003o=@00ocD0000000<00?nL00D00000003oo`00ool00?oo00000002002To`05
00000000Dol005?o001Co`0000000`003_l01@0000009P3o02H0o`0V0?l0000000<0BP3o00D00000
05h0o`1N0?l0GP3o000000020600o`050000001B0?l0DP3o0580o`0000000`0e0?l01@0000002P3o
00X0o`0:0?l0000000<002co00D00000001[o`00Jol006_o00000002002ao`0500000000nOl00?Wo
003io`0000000`00ok`01@0000000?me003oM@00ogD0000000<00?lb00@0000000Wo0009o`000000
0`0lo`001@000000I_l006Ko001Vo`00000000<0QOl000D0000009So002Ho`00V?l0000000030:3o
00030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooool0
0000P@0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool010000000POl0
087o0000000306So000500000015o`00AOl004Go000000000`0Ho`001@0000000?lL003o7000oa`0
000000<00?mH00@00000003oVP00oiX000000`00omh01@0000000=_o003Ko`00fol0000000<009Oo
00D00000001Go`00Eol005Oo00000003000Mo`040000000C0?l04`3o000000<0>@3o00D000000580
o`1B0?l0DP3o0000000305d0o`050000001H0?l0F03o05P0o`0000000`140?l0100000007`3o01l0
o`000003000Do`0500000000E_l005Ko001Fo`0000000`00Y?l01@0000000?co003lo`00o?l00000
00<00?nS00@00000003o?`00ocl000000`0To`001@000000Qol008Oo0027o`00000000<0hol000D0
00000?o:003obP00olX0000000030?n400040000003oCP00odh0000000<0obT000D000000?lG003o
5`00oaL0000000030?lJ00050000003o<P00oc800?lb000000000`3oG`0010000000oih00?nN0000
00030?o_00050000002`o`00/?l00;3o000000000`14o`001@0000000?l_003o;`00obl000000080
0?nX00D00000003No`00g_l00=ko00000003001Yo`05000000030?l00`3o00<0o`0000000`1U0?l0
1@000000]`3o0;L0o`2g0?l000000080n03o00D000000?l0f`3o0=/0o`3K000000030?l0a0050000
003o0<@0o`340?l0a00000000`3o0=/01@000000n03o0?P0o`3h0?l000000080]`3o00D0000006D0
o`1U0?l0I@3o0000000300<0o`0500000000JOl006Wo001Yo`0000000`00g_l01@0000000?nX003o
Z000ojP0000000800?l_00D0000004Co0014o`00A?l0000000030;3o00050000003ok`00onl00?o_
000000000`3oWP001@000000oel00?mO003oG`0000000080oc8000D000000?lJ003o6P00oaX00000
00030?lG00050000003o:@00obT00?lY000000000`3oCP001@000000oh@00?n4003oQ00000000080
olX000D000000>?o003So`00hol00000000308Oo00050000000To`009?l002Co000000000`00ocl0
1@0000000?nS003oX`00oj<0000000800?co00D00000002To`00Y?l00:Co00000003001Fo`050000
00005?l001Co000Do`0000000`0O0?l01@000000A03o04@0o`140?l000000080F03o00D0000005d0
o`1M0?l0G@3o000000030580o`050000000i0?l0>@3o03T0o`0000000`0C0?l01@00000001go000M
o`007Ol00000008005Oo00D00000002Go`00Uol009Oo00000003003Ko`0500000000omh00?oN003o
gP0000000`00oiX0100000000?mH003oF0000003003o70050000000Ho`006?l001So000000000`15
o`001@000000J?l006So001Xo`00000000<0POl000<000000?ooo`3oool0103oool100000040oooo
000P0?ooo`030000003oool0oooo00D0oooo00@00000087o0021o`0000000`1Xo`001@000000AOl0
04Go0015o`00000000<06?l000D00000003o7000oa`00?lL00000003003oF00400000000oiX00?nJ
000000<00?oN00D00000003Ko`00fol00=_o00000003002Go`0500000000Eol005Oo001Go`000000
0`007Ol0100000004`3o01<0o`00000303T0o`050000001B0?l0DP3o0580o`0000000`1M0?l01@00
0000F03o05P0o`1H0?l0000000<0A03o00@0000001l0o`0O0?l000000`005?l01@00000005Ko001F
o`00E_l0000000<00:Co00D00000003lo`00o?l00?co00000003003oX`0400000000ocl00?lo0000
00<09?l000D0000008Oo0027o`00Qol0000000030>?o00050000003obP00olX00?o:000000000`3o
Q00010000000odh00?m>000000030?lY00050000003o5`00oaL00?lG000000000`3o6P001@000000
oc800?lb003o<P00000000<0oel000@000000?nN003oWP0000000`3ok`001@000000/?l00;3o002`
o`00000000<0A?l000D00000003o;`00obl00?l_00000002003oZ00500000000g_l00=ko003No`00
00000`00JOl01@0000000`3o00<0o`030?l0000000<0I@3o00D000000;L0o`2g0?l0]`3o00000002
0?P0o`050000003o0=/0o`3K0?l0f`0000000`3o0<@01@000000o`340?l0a03o0<@0000000<0o`3K
00D000000?P0o`3h0?l0n03o000000020;L0o`050000001U0?l0I@3o06D0o`0000000`030?l01@00
000006Wo001Yo`00JOl0000000<00=ko00D00000003oZ000ojP00?nX00000002003o;`0500000014
o`00A?l004Co000000000`2`o`001@000000onl00?o_003ok`00000000<0oih000D000000?mO003o
G`00oel0000000020?lb00050000003o6P00oaX00?lJ000000000`3o5`001@000000obT00?lY003o
:@00000000<0odh000D000000?n4003oQ000oh@0000000020?o:00050000003So`00hol00>?o0000
00000`27o`001@0000009?l002Co000To`00000000<00?lo00D00000003oX`00oj<00?nS00000002
003lo`0500000000Y?l00:Co002To`0000000`00E_l01@00000001Co000Do`005?l0000000<07`3o
00D0000004@0o`140?l0A03o0000000205P0o`050000001M0?l0G@3o05d0o`0000000`1B0?l01@00
0000>@3o03T0o`0i0?l0000000<04`3o00D00000000Mo`007Ol001go00000002001Go`0500000000
Uol009Oo002Go`0000000`00fol01@0000000?oN003ogP00omh0000000<00?nJ00@00000003oF000
oeP000000`00oa`01@0000006?l001So000Ho`00000000<0AOl000D0000006So001Xo`00J?l00000
0003087o00030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`05
0?ooo`0400000021o`00POl0000000<0J?l000D0000004Go0015o`00AOl00000000301So00050000
0000oa`00?lL003o700000000`00oeP0100000000?nJ003oVP000003003ogP0500000000fol00=_o
003Ko`0000000`00Uol01@00000005Oo001Go`00Eol0000000<001go00@0000001<0o`0C0?l00000
0`0i0?l01@000000DP3o0580o`1B0?l0000000<0G@3o00D0000005P0o`1H0?l0F03o0000000304@0
o`040000000O0?l07`3o000000<001Co00D00000001Fo`00E_l005Ko00000003002To`0500000000
o?l00?co003lo`0000000`00oj<0100000000?lo003o?`00000302Co000500000027o`00Qol008Oo
000000000`3So`001@000000olX00?o:003obP00000000<0oh@000@000000?m>003oCP0000000`3o
:@001@000000oaL00?lG003o5`00000000<0oaX000D000000?lb003o<P00oc80000000030?mO0004
0000003oWP00oih0000000<0onl000D000000;3o002`o`00/?l00000000304Co000500000000obl0
0?l_003o;`0000000P00ojP01@0000000=ko003No`00g_l0000000<006Wo00D0000000<0o`030?l0
0`3o0000000306D0o`050000002g0?l0]`3o0;L0o`0000000P3h0?l01@000000o`3K0?l0f`3o0=/0
000000<0o`3400D000000?l0a03o0<@0o`34000000030?l0f`050000003h0?l0n03o0?P0o`000000
0P2g0?l01@000000I@3o06D0o`1U0?l0000000<00`3o00D00000001Yo`00JOl006Wo00000003003N
o`0500000000ojP00?nX003oZ00000000P00obl01@000000A?l004Co0014o`00000000<0/?l000D0
00000?o_003ok`00onl0000000030?nN00050000003oG`00oel00?mO000000000P3o<P001@000000
oaX00?lJ003o6P00000000<0oaL000D000000?lY003o:@00obT0000000030?m>00050000003oQ000
oh@00?n4000000000P3obP001@000000hol00>?o003So`00000000<0Qol000D0000002Co000To`00
9?l000000003003o?`0500000000oj<00?nS003oX`0000000P00o?l01@0000000:Co002To`00Y?l0
000000<005Ko00D00000000Do`005?l001Co0000000301l0o`05000000140?l0A03o04@0o`000000
0P1H0?l01@000000G@3o05d0o`1M0?l0000000<0DP3o00D0000003T0o`0i0?l0>@3o0000000301<0
o`05000000007Ol001go000Mo`0000000P00Eol01@00000009Oo002Go`00Uol0000000<00=_o00D0
0000003ogP00omh00?oN00000003003oVP0400000000oeP00?mH000000<00?lL00D0000001So000H
o`006?l00000000304Go00050000001Xo`00J?l006So000000000`21o`000`000000oooo0?ooo`04
0?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo0000084000001P3oool10000
0040oooo000P0?ooo`030000003oool0oooo00D0oooo00@0000004Co0014o`0000000`0Mo`001@00
00000?lA003o4@00oa40000000<00?m700D00000003oP`00oh<00?n300000003003o`P0400000000
nol00?_o000000<00;[o00D00000001ko`00Nol007_o000000030011o`05000000003Ol000go000=
o`0000000`0M0?l010000000?@3o03d0o`0000030540o`050000001H0?l0F03o05P0o`0000000`1@
0?l01@000000>@3o03T0o`0i0?l0000000<0503o00@00000000No`007_l000000`00Gol01@000000
0:_o002[o`00Zol0000000<00?oo00D00000003oX`00oj<00?nS00000003003o@`040000000No`00
7_l0000000<0OOl000D000000=Oo003Go`00eol0000000030?oF00050000003oS`00ohl00?n?0000
00000`3oE@0010000000ob`00?l/000000030?lE00050000003o4000oa000?l@000000000`3o7`00
1@000000od800?m2003o@P00000000<0ogL000@000000?nn003o_P0000000`3[o`001@000000R?l0
08So0028o`00000000<06ol000D00000003oF000oeP00?mH00000002003ocP0500000000^ol00;_o
002ko`0000000`00B_l01@0000007P3o01h0o`0N0?l0000000<0O03o00D000000</0o`3;0?l0b`3o
000000020?l0m`050000003o0<d0o`3=0?l0c@0000000`3o0;P01@000000o`2h0?l0^03o0;P00000
00<0o`3=00D000000?l0m`3o0?L0o`3g000000020</0o`050000001l0?l0O03o07`0o`0000000`0N
0?l01@00000004[o001:o`00B_l0000000<00;_o00D00000003ocP00olh00?o>00000002003oF005
0000000Ko`006ol001_o000000000`28o`001@000000jol00>_o003[o`00000000<0okh000D00000
0?mg003oM`00ogL0000000020?m200050000003o7`00oal00?lO000000000`3o40001@000000oaD0
0?lE003o5@00000000<0ob`000D000000?mE003oE@00oeD0000000020?n?00050000003oeP00omH0
0?oF000000000`3Go`001@000000OOl007go001mo`00000000<07_l000D00000003o@`00od<00?m3
00000002003oX`0500000000ool00?oo003oo`0000000`00Zol01@00000005oo001Oo`00Gol00000
00<001ko00D0000001@0o`0D0?l0503o0000000203T0o`050000001@0?l0D03o0500o`0000000`1H
0?l01@000000D@3o0540o`1A0?l0000000<0?@3o00D0000001d0o`0M0?l07@3o00000002000=o`05
00000000@Ol0047o0011o`0000000`00Nol01@0000000;[o002jo`00^_l0000000<00?_o00@00000
003o`P00ol8000000`00oh<01@0000000?m7003oA`00odL0000000<00?lA00D0000001go000Mo`00
7Ol00000000304Co00030000003oool0oooo00@0oooo0@0000010?ooo`003`3oool00`2oM0000000
000000040000000304RLg`3oool0oooo00L0oooo00<000000?ooo`3oool01@3oool010000000A?l0
04Co0000000301go000500000000oa400?lA003o4@0000000`00odL01@0000000?n3003oP`00oh<0
000000<00?o200@00000003ko`00nol000000`00^_l01@00000007_o001ko`00Nol0000000<0047o
00D00000000=o`003Ol000go0000000301d0o`040000000m0?l0?@3o000000<0D@3o00D0000005P0
o`1H0?l0F03o000000030500o`050000000i0?l0>@3o03T0o`0000000`0D0?l01000000001ko000N
o`000003001Oo`0500000000Zol00:_o002[o`0000000`00ool01@0000000?nS003oX`00oj<00000
00<00?m300@0000001ko000No`0000000`1mo`001@000000eol00=Oo003Go`00000000<0omH000D0
00000?n?003oS`00ohl0000000030?mE00040000003o;000ob`0000000<0oaD000D000000?l@003o
4000oa00000000030?lO00050000003o@P00od800?m2000000000`3oM`0010000000okh00?nn0000
00030>_o000500000028o`00R?l008So000000000`0Ko`001@0000000?mH003oF000oeP000000080
0?o>00D00000002ko`00^ol00;_o00000003001:o`050000000N0?l07P3o01h0o`0000000`1l0?l0
1@000000b`3o0</0o`3;0?l000000080o`3g00D000000?l0c@3o0<d0o`3=000000030?l0^0050000
003o0;P0o`2h0?l0^00000000`3o0<d01@000000o`3g0?l0m`3o0?L000000080b`3o00D0000007`0
o`1l0?l0O03o0000000301h0o`0500000000B_l004[o001:o`0000000`00^ol01@0000000?o>003o
cP00olh0000000800?mH00D0000001_o000Ko`006ol00000000308So00050000003[o`00jol00>_o
000000000`3o_P001@000000ogL00?mg003oM`0000000080od8000D000000?lO003o7`00oal00000
00030?l@00050000003o5@00oaD00?lE000000000`3o;0001@000000oeD00?mE003oE@0000000080
ohl000D000000?oF003oeP00omH0000000030=Oo00050000001mo`00OOl007go000000000`0No`00
1@0000000?m3003o@`00od<0000000800?nS00D00000003oo`00ool00?oo00000003002[o`050000
0000Gol005oo001Oo`0000000`007_l01@000000503o01@0o`0D0?l000000080>@3o00D000000500
o`1@0?l0D03o0000000305P0o`050000001A0?l0D@3o0540o`0000000`0m0?l01@0000007@3o01d0
o`0M0?l00000008000go00D000000011o`00@Ol0047o00000003001ko`0500000000^_l00;[o002j
o`0000000`00nol0100000000?o2003o`P000003003oP`0500000000odL00?m7003oA`0000000`00
oa401@0000007Ol001go000Mo`00000000<0A?l000<000000?ooo`3oool0103oool100000040oooo
000@0?ooo`030=nLB0000000000000<0000000<004RL0=ooo`3oool01`3oool00`000000oooo0?oo
o`050?ooo`0400000014o`00A?l0000000<07Ol000D00000003o4@00oa400?lA00000003003oA`05
00000000oh<00?n3003oP`0000000`00ol80100000000?_o003ko`000003002jo`0500000000Nol0
07_o001ko`0000000`00@Ol01@00000000go000=o`003Ol0000000<07@3o00@0000003d0o`0m0?l0
00000`1A0?l01@000000F03o05P0o`1H0?l0000000<0D03o00D0000003T0o`0i0?l0>@3o00000003
01@0o`04000000007_l001ko000000<005oo00D00000002[o`00Zol00:_o00000003003oo`050000
0000oj<00?nS003oX`0000000`00od<0100000007_l001ko0000000307go00050000003Go`00eol0
0=Oo000000000`3oeP001@000000ohl00?n?003oS`00000000<0oeD000@000000?l/003o;0000000
0`3o5@001@000000oa000?l@003o4000000000<0oal000D000000?m2003o@P00od80000000030?mg
00040000003o_P00okh0000000<0jol000D0000008So0028o`00R?l00000000301_o000500000000
oeP00?mH003oF00000000P00olh01@0000000;_o002ko`00^ol0000000<004[o00D0000001h0o`0N
0?l07P3o0000000307`0o`050000003;0?l0b`3o0</0o`0000000P3o0?L01@000000o`3=0?l0c@3o
0<d0000000<0o`2h00D000000?l0^03o0;P0o`2h000000030?l0c@050000003o0?L0o`3g0?l0m`00
00000P3;0?l01@000000O03o07`0o`1l0?l0000000<07P3o00D00000001:o`00B_l004[o00000003
002ko`0500000000olh00?o>003ocP0000000P00oeP01@0000006ol001_o000Ko`00000000<0R?l0
00D000000>_o003[o`00jol0000000030?nn00050000003oM`00ogL00?mg000000000P3o@P001@00
0000oal00?lO003o7`00000000<0oa0000D000000?lE003o5@00oaD0000000030?l/00050000003o
E@00oeD00?mE000000000P3oS`001@000000omH00?oF003oeP00000000<0eol000D0000007go001m
o`00OOl00000000301ko000500000000od<00?m3003o@`0000000P00oj<01@0000000?oo003oo`00
ool0000000<00:_o00D00000001Oo`00Gol005oo00000003000No`050000000D0?l0503o01@0o`00
00000P0i0?l01@000000D03o0500o`1@0?l0000000<0F03o00D000000540o`1A0?l0D@3o00000003
03d0o`050000000M0?l07@3o01d0o`0000000P003Ol01@000000047o0011o`00@Ol0000000<007_o
00D00000002jo`00^_l00;[o00000003003ko`0400000000ol800?o2000000<00?n300D00000003o
A`00odL00?m700000003003o4@050000000Mo`007Ol001go000000000`14o`000`000000oooo0?oo
o`040?ooo`4000000@3oool00140oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`07
0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?ooo`004P3oool00`3O
W4P0M;oo0?ooo`0;0?ooo`030000003oool0oooo00D0oooo00@00000003o4000oa0000000`00od40
1@0000000?mg003oM`00ogL0000000<00?na00D00000003okP00onh00?o^00000003003Do`040000
0000V?l009So000000<005oo00D00000000[o`00:ol002_o000000030040o`050000000V0?l09P3o
02H0o`0000000`100?l010000000CP3o04h0o`0000030500o`05000000140?l0A03o04@0o`000000
0`0[0?l01@000000103o00@0o`040?l0000000<002oo00@00000001^o`00K_l000000`00^?l01@00
00000?oe003om@00ooD0000000<00?nL00D00000003o?`00ocl00?lo0000000301ko00040000001j
o`00N_l0000000<0d_l000D000000?oM003og@00omd0000000030?nF00050000003oG000oe`00?mL
000000000`3o<00010000000oa@00?lD000000030?l:00050000003o4P00oa800?lB000000000`3o
;0001@000000oeT00?mI003oF@00000000<0oiH000@000000?oS003oh`0000000`32o`001@000000
GOl005go001Mo`00000000<00?l@00D00000003oPP00oh800?n200000002003omP0500000000Uol0
09Oo002Go`0000000`00:_l01@000000>P3o03X0o`0j0?l0000000<0T`3o00D000000=h0o`3N0?l0
gP3o000000020?l0i`050000003o0<00o`300?l0`00000000`3o0:/01@000000o`2[0?l0Z`3o0:/0
000000<0o`3000D000000?l0i`3o0>L0o`3W000000020=h0o`050000002C0?l0T`3o09<0o`000000
0`0j0?l01@00000002[o000Zo`00:_l0000000<009Oo00D00000003omP00ooH00?of00000002003o
PP0500000000oa000?l@003o400000000`1Mo`001@000000`_l00<;o0032o`00000000<0on<000D0
00000?nF003oUP00oiH0000000020?mI00050000003o;000ob`00?l/000000000`3o4P001@000000
o`X00?l:003o2P00000000<0oa@000D000000?l`003o<000oc00000000020?mL00050000003oUP00
oiH00?nF000000000`3og@001@000000d_l00=;o003Bo`00000000<0N_l000D0000001ko000No`00
7_l000000002003o?`0500000000oi`00?nL003oW00000000`00ooD01@0000000;So002ho`00^?l0
000000<006ko00D00000000_o`00;ol002oo0000000200@0o`050000000[0?l0:`3o02/0o`000000
0`140?l01@000000D03o0500o`1@0?l0000000<0CP3o00D000000400o`100?l0@03o0000000202H0
o`05000000010?l00@3o0040o`0000000`00:ol01@00000005oo001Oo`00Gol0000000<009So00@0
0000003Do`00e?l000000`00onh01@0000000?na003o/@00ok40000000<00?mg00D00000003o@@00
od400?m100000003003o40030000003oool0oooo00@0oooo0@0000010?ooo`000`3oool00`3ooml0
W4P0000000070000000304RLg`3oool0oooo00<0oooo00<0_gA809cOo`3oool02P3oool00`000000
oooo0?ooo`050?ooo`0400000000oa000?l@000000<00?m100D00000003oM`00ogL00?mg00000003
003o/@0500000000onh00?o^003okP0000000`00e?l01000000009So002Ho`000003001Oo`050000
0000:ol002_o000[o`0000000`010?l01@0000009P3o02H0o`0V0?l0000000<0@03o00@0000004h0
o`1>0?l000000`1@0?l01@000000A03o04@0o`140?l0000000<0:`3o00D0000000@0o`040?l0103o
00000003000_o`0400000000K_l006ko000000<00;So00D00000003om@00ooD00?oe00000003003o
W00500000000ocl00?lo003o?`0000000`0No`0010000000N_l007[o000000030=;o00050000003o
g@00omd00?oM000000000`3oUP001@000000oe`00?mL003oG000000000<0oc0000@000000?lD003o
500000000`3o2P001@000000oa800?lB003o4P00000000<0ob`000D000000?mI003oF@00oeT00000
00030?nF00040000003oh`00on<0000000<0`_l000D0000005go001Mo`00GOl000000003003o4005
00000000oh800?n2003oPP0000000P00ooH01@00000009Oo002Go`00Uol0000000<002[o00D00000
03X0o`0j0?l0>P3o0000000309<0o`050000003N0?l0gP3o0=h0o`0000000P3o0>L01@000000o`30
0?l0`03o0<00000000<0o`2[00D000000?l0Z`3o0:/0o`2[000000030?l0`0050000003o0>L0o`3W
0?l0i`0000000P3N0?l01@000000T`3o09<0o`2C0?l0000000<0>P3o00D00000000Zo`00:_l002[o
00000003002Go`0500000000ooH00?of003omP0000000P00oh801@0000000?l@003o4000oa000000
00<0GOl000D000000<;o0032o`00`_l0000000030?oS00050000003oUP00oiH00?nF000000000P3o
F@001@000000ob`00?l/003o;000000000<0oa8000D000000?l:003o2P00o`X0000000030?lD0005
0000003o<000oc000?l`000000000P3oG0001@000000oiH00?nF003oUP00000000<0omd000D00000
0=;o003Bo`00d_l00000000307[o00050000000No`007_l001ko000000000P00ocl01@0000000?nL
003oW000oi`0000000<00?oe00D00000002ho`00^?l00;So00000003001^o`0500000000;ol002oo
000_o`0000000P040?l01@000000:`3o02/0o`0[0?l0000000<0A03o00D000000500o`1@0?l0D03o
0000000304h0o`05000000100?l0@03o0400o`0000000P0V0?l01@0000000@3o0040o`010?l00000
00<002_o00D00000001Oo`00Gol005oo00000003002Ho`0400000000e?l00=Co000000<00?o^00D0
0000003o/@00ok400?na00000003003oM`0500000000od400?m1003o@@0000000`00oa000`000000
oooo0?ooo`040?ooo`4000000@3oool001<0oooo00<0oooO09a8M02oool02P3oool3000000D0oooo
o`000021000000@0oooo0`0000010?ooo`00503oool00`3o_g@007Bo0?ooo`090?ooo`030000003o
ool0oooo00D0oooo00@00000003oM000og@000000`00ojX01@0000000?oR003ohP00on80000000<0
0>Co00D00000002/o`00[?l00:co00000003001fo`0400000000@ol004?o000000<001Ko00D00000
00l0o`0?0?l03`3o0000000302`0o`050000000o0?l0?`3o03l0o`0000000`170?l010000000@`3o
04<0o`00000303<0o`050000000G0?l05`3o01L0o`0000000`004Ol01@00000004Go0015o`00AOl0
000000<008Co00@00000003<o`00c?l000000`00on<01@0000000?n>003oSP00ohh0000000<00?le
00D0000002Co000To`009?l00000000307go00040000003Bo`00d_l0000000<0oml000D000000?nJ
003oVP00oiX0000000030?mP00050000003o<`00oc<00?lc000000000`3o5@0010000000o`L00?l7
000000030?l900050000003o7@00oad00?lM000000000`3o@@001@000000ogH00?mf003oMP000000
00<0ok/000@000000??o003co`0000000`2Eo`001@000000;ol002oo000_o`00000000<00?lm00D0
0000003o[P00ojh00?n^00000002003Po`0500000000L_l007;o001bo`0000000`002_l01@000000
E@3o05D0o`1E0?l0000000<0ZP3o00D000000?00o`3`0?l0l03o000000020?l0f0050000003o0;80
o`2b0?l0/P0000000`3o09l01@000000o`2O0?l0W`3o09l0000000<0o`2b00D000000?l0f03o0=P0
o`3H000000020?00o`050000002Z0?l0ZP3o0:X0o`0000000`1E0?l01@00000000[o000:o`002_l0
000000<007;o00D00000003Po`00h?l00>3o00000002003o[P0500000000ocd00?lm003o?@000000
0`0_o`001@000000UOl009Go002Eo`00000000<0lol000D000000?nk003o^`00ok/0000000020?mf
00050000003o@@00od400?m1000000000`3o7@001@000000o`T00?l9003o2@00000000<0o`L000D0
00000?lE003o5@00oaD0000000020?lc00050000003oH000of000?mP000000000`3oVP001@000000
oml00?oO003og`00000000<0d_l000D0000007go001mo`00OOl00000000202Co000500000000ocD0
0?le003o=@0000000`00ohh01@0000000?oS003oh`00on<0000000<00<co00D000000024o`00Q?l0
08Co000000020015o`05000000004Ol0017o000Ao`0000000`0G0?l01@000000<`3o03<0o`0c0?l0
000000<0@`3o00D0000004L0o`170?l0A`3o0000000203l0o`050000000/0?l0;03o02`0o`000000
0`0?0?l01@00000001Ko000Fo`005_l0000000<004?o00@00000001fo`00M_l000000`00[?l01@00
00000>Co003To`00i?l0000000<00?oR00D00000003oZP00ojX00?nZ00000003003oM0030000003o
ool0oooo00@0oooo0@0000010?ooo`00503oool00`3ookl0M01809cOo`090?ooo`030000003oool0
oooo00D0oooo00@00000003oM000og@000000`00ojX01@0000000?oR003ohP00on80000000<00>Co
00D00000002/o`00[?l00:co00000003001fo`0400000000@ol004?o000000<001Ko00D0000000l0
o`0?0?l03`3o0000000302`0o`050000000o0?l0?`3o03l0o`0000000`170?l010000000@`3o04<0
o`00000303<0o`050000000G0?l05`3o01L0o`0000000`004Ol01@00000004Go0015o`00AOl00000
00<008Co00@00000003<o`00c?l000000`00on<01@0000000?n>003oSP00ohh0000000<00?le00D0
000002Co000To`009?l00000000307go00040000003Bo`00d_l0000000<0oml000D000000?nJ003o
VP00oiX0000000030?mP00050000003o<`00oc<00?lc000000000`3o5@0010000000o`L00?l70000
00030?l900050000003o7@00oad00?lM000000000`3o@@001@000000ogH00?mf003oMP00000000<0
ok/000@000000??o003co`0000000`2Eo`001@000000;ol002oo000_o`00000000<00?lm00D00000
003o[P00ojh00?n^00000002003Po`0500000000L_l007;o001bo`0000000`002_l01@000000E@3o
05D0o`1E0?l0000000<0ZP3o00D000000?00o`3`0?l0l03o000000020?l0f0050000003o0;80o`2b
0?l0/P0000000`3o09l01@000000o`2O0?l0W`3o09l0000000<0o`2b00D000000?l0f03o0=P0o`3H
000000020?00o`050000002Z0?l0ZP3o0:X0o`0000000`1E0?l01@00000000[o000:o`002_l00000
00<007;o00D00000003Po`00h?l00>3o00000002003o[P0500000000ocd00?lm003o?@0000000`0_
o`001@000000UOl009Go002Eo`00000000<0lol000D000000?nk003o^`00ok/0000000020?mf0005
0000003o@@00od400?m1000000000`3o7@001@000000o`T00?l9003o2@00000000<0o`L000D00000
0?lE003o5@00oaD0000000020?lc00050000003oH000of000?mP000000000`3oVP001@000000oml0
0?oO003og`00000000<0d_l000D0000007go001mo`00OOl00000000202Co000500000000ocD00?le
003o=@0000000`00ohh01@0000000?oS003oh`00on<0000000<00<co00D000000024o`00Q?l008Co
000000020015o`05000000004Ol0017o000Ao`0000000`0G0?l01@000000<`3o03<0o`0c0?l00000
00<0@`3o00D0000004L0o`170?l0A`3o0000000203l0o`050000000/0?l0;03o02`0o`0000000`0?
0?l01@00000001Ko000Fo`005_l0000000<004?o00@00000001fo`00M_l000000`00[?l01@000000
0>Co003To`00i?l0000000<00?oR00D00000003oZP00ojX00?nZ00000003003oM0030000003oool0
oooo00@0oooo0@0000010?ooo`00503oool00`3ookl0M00004RLg`090?ooo`030000003oool0oooo
00D0oooo00@00000003oM000og@000000`00ojX01@0000000?oR003ohP00on80000000<00>Co00D0
0000002/o`00[?l00:co00000003001fo`0400000000@ol004?o000000<001Ko00D0000000l0o`0?
0?l03`3o0000000302`0o`050000000o0?l0?`3o03l0o`0000000`170?l010000000@`3o04<0o`00
000303<0o`050000000G0?l05`3o01L0o`0000000`004Ol01@00000004Go0015o`00AOl0000000<0
08Co00@00000003<o`00c?l000000`00on<01@0000000?n>003oSP00ohh0000000<00?le00D00000
02Co000To`009?l00000000307go00040000003Bo`00d_l0000000<0oml000D000000?nJ003oVP00
oiX0000000030?mP00050000003o<`00oc<00?lc000000000`3o5@0010000000o`L00?l700000003
0?l900050000003o7@00oad00?lM000000000`3o@@001@000000ogH00?mf003oMP00000000<0ok/0
00@000000??o003co`0000000`2Eo`001@000000;ol002oo000_o`00000000<00?lm00D00000003o
[P00ojh00?n^00000002003Po`0500000000L_l007;o001bo`0000000`002_l01@000000E@3o05D0
o`1E0?l0000000<0ZP3o00D000000?00o`3`0?l0l03o000000020?l0f0050000003o0;80o`2b0?l0
/P0000000`3o09l01@000000o`2O0?l0W`3o09l0000000<0o`2b00D000000?l0f03o0=P0o`3H0000
00020?00o`050000002Z0?l0ZP3o0:X0o`0000000`1E0?l01@00000000[o000:o`002_l0000000<0
07;o00D00000003Po`00h?l00>3o00000002003o[P0500000000ocd00?lm003o?@0000000`0_o`00
1@000000UOl009Go002Eo`00000000<0lol000D000000?nk003o^`00ok/0000000020?mf00050000
003o@@00od400?m1000000000`3o7@001@000000o`T00?l9003o2@00000000<0o`L000D000000?lE
003o5@00oaD0000000020?lc00050000003oH000of000?mP000000000`3oVP001@000000oml00?oO
003og`00000000<0d_l000D0000007go001mo`00OOl00000000202Co000500000000ocD00?le003o
=@0000000`00ohh01@0000000?oS003oh`00on<0000000<00<co00D000000024o`00Q?l008Co0000
00020015o`05000000004Ol0017o000Ao`0000000`0G0?l01@000000<`3o03<0o`0c0?l0000000<0
@`3o00D0000004L0o`170?l0A`3o0000000203l0o`050000000/0?l0;03o02`0o`0000000`0?0?l0
1@00000001Ko000Fo`005_l0000000<004?o00@00000001fo`00M_l000000`00[?l01@0000000>Co
003To`00i?l0000000<00?oR00D00000003oZP00ojX00?nZ00000003003oM0030000003oool0oooo
00@0oooo0@0000010?ooo`003`3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO00T0
oooo00<000000?ooo`3oool01@3ooooo0000084000001P3oool100000040oooo000?0?ooo`080?oo
g`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool90?ooo`030000003oool0oooo00D0oooo00@00000
003ogP00omh000000`00k?l01@0000000;So002ho`00^?l0000000<008Go00D00000001Eo`00EOl0
05Go00000003000Zo`04000000001Ol000Go000000<0603o00D0000002h0o`0^0?l0;P3o00000003
03X0o`050000000k0?l0>`3o03/0o`0000000`0b0?l0100000007@3o01d0o`0000030002o`050000
0000;?l002co000/o`0000000`00H_l01@0000000:3o002Po`00X?l0000000<00>Oo00@00000003o
bP00olX000000`00ogP01@0000000?lS003o8`00ob<0000000<0<_l000D0000008Oo0027o`00Qol0
000000030=Oo00040000003og@00omd0000000<0oiX000D000000?mQ003oH@00of40000000030?le
00050000003o5@00oaD00?lE000000000`3o1@0010000000o`@00?l4000000030?lB00050000003o
<@00oc400?la000000000`3oG`001@000000oi/00?nK003oV`00000000<0onD000@000000<Co0034
o`0000000`1Uo`001@0000000?l1003o0@00o`40000000<00?m]00D00000003of`00om/00?oK0000
0002002go`0500000000COl004go001=o`0000000`0E0?l01@000000L03o0700o`1`0?l0000000<0
`03o00D000000?l0o03o0?`0o`3l000000020?l0b@050000003o0:H0o`2V0?l0YP0000000`3o09@0
1@000000o`2D0?l0U03o09@0000000<0o`2V00D000000?l0b@3o0<T0o`39000000020?l0o0050000
00300?l0`03o0<00o`0000000`1`0?l01@0000005@3o01D0o`0E0?l0000000<004go00D00000002g
o`00]ol00;Oo00000002003of`0500000000ofd00?m]003oK@0000000`00o`401@000000IOl006Go
001Uo`00000000<0a?l000D000000?oU003oi@00onD0000000020?nK00050000003oG`00oel00?mO
000000000`3o<@001@000000oa800?lB003o4P00000000<0o`@000D000000?l5003o1@00o`D00000
00020?lE00050000003o=@00ocD00?le000000000`3oH@001@000000oiX00?nJ003oVP00000000<0
omd000D000000=Oo003Go`00eol00000000208Oo00050000000bo`00<_l003;o000000000`00ob<0
1@0000000?mh003oN000ogP0000000<00?o:00D00000003Wo`00iol00>Oo00000002002Po`050000
0000H_l006;o001Ro`0000000`00;?l01@00000000;o0002o`000_l0000000<07@3o00D000000380
o`0b0?l0<P3o0000000203/0o`050000000j0?l0>P3o03X0o`0000000`0^0?l01@000000603o01P0
o`0H0?l0000000<000Go00@00000000Zo`00:_l000000`00EOl01@00000008Go0025o`00QOl00000
00<00;So00D00000003/o`00k?l00>co00000003003ogP030000003oool0oooo00@0oooo0@000001
0?ooo`00403oool01P3ooml0W4P0000000000000001d0;ooo`X0oooo00<000000?ooo`3oool01@3o
ool0100000000?oN003ogP000003003/o`0500000000^?l00;So002ho`0000000`00QOl01@000000
05Go001Eo`00EOl0000000<002[o00@000000005o`001Ol000000`0H0?l01@000000;P3o02h0o`0^
0?l0000000<0>P3o00D0000003/0o`0k0?l0>`3o000000030380o`040000000M0?l07@3o000000<0
00;o00D00000000/o`00;?l002co00000003001Ro`0500000000X?l00:3o002Po`0000000`00iol0
100000000?o:003obP000003003oN00500000000ob<00?lS003o8`0000000`0bo`001@000000Qol0
08Oo0027o`00000000<0eol000@000000?oM003og@0000000`3oVP001@000000of400?mQ003oH@00
000000<0ocD000D000000?lE003o5@00oaD0000000030?l500040000003o1000o`@0000000<0oa80
00D000000?la003o<@00oc40000000030?mO00050000003oV`00oi/00?nK000000000`3oi@001000
0000a?l00<Co0000000306Go000500000000o`400?l1003o0@0000000`00ofd01@0000000?oK003o
f`00om/0000000800;Oo00D00000001=o`00COl004go0000000301D0o`050000001`0?l0L03o0700
o`0000000`300?l01@000000o`3l0?l0o03o0?`000000080o`3900D000000?l0YP3o0:H0o`2V0000
00030?l0U0050000003o09@0o`2D0?l0U00000000`3o0:H01@000000o`390?l0b@3o0<T000000080
o`3l00D000000<00o`300?l0`03o000000030700o`050000000E0?l05@3o01D0o`0000000`00COl0
1@0000000;Oo002go`00]ol0000000800?oK00D00000003oK@00ofd00?m]00000003003o0@050000
001Uo`00IOl006Go000000000`34o`001@000000onD00?oU003oi@0000000080oi/000D000000?mO
003oG`00oel0000000030?la00050000003o4P00oa800?lB000000000`3o10001@000000o`D00?l5
003o1@0000000080oaD000D000000?le003o=@00ocD0000000030?mQ00050000003oVP00oiX00?nJ
000000000`3og@001@000000eol00=Oo003Go`0000000080Qol000D0000003;o000bo`00<_l00000
0003003o8`0500000000ogP00?mh003oN00000000`00olX01@0000000>Oo003Wo`00iol000000080
0:3o00D00000001Ro`00H_l006;o00000003000/o`05000000000_l000;o0002o`0000000`0M0?l0
1@000000<P3o0380o`0b0?l000000080>`3o00D0000003X0o`0j0?l0>P3o0000000302h0o`050000
000H0?l0603o01P0o`0000000`001Ol01000000002[o000Zo`000003001Eo`0500000000QOl008Go
0025o`0000000`00^?l01@0000000>co003/o`00k?l0000000<00?oN00<000000?ooo`3oool0103o
ool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003ogP00omh000000`00
k?l01@0000000;So002ho`00^?l0000000<008Go00D00000001Eo`00EOl005Go00000003000Zo`04
000000001Ol000Go000000<0603o00D0000002h0o`0^0?l0;P3o0000000303X0o`050000000k0?l0
>`3o03/0o`0000000`0b0?l0100000007@3o01d0o`0000030002o`0500000000;?l002co000/o`00
00000`00H_l01@0000000:3o002Po`00X?l0000000<00>Oo00@00000003obP00olX000000`00ogP0
1@0000000?lS003o8`00ob<0000000<0<_l000D0000008Oo0027o`00Qol0000000030=Oo00040000
003og@00omd0000000<0oiX000D000000?mQ003oH@00of40000000030?le00050000003o5@00oaD0
0?lE000000000`3o1@0010000000o`@00?l4000000030?lB00050000003o<@00oc400?la00000000
0`3oG`001@000000oi/00?nK003oV`00000000<0onD000@000000<Co0034o`0000000`1Uo`001@00
00000?l1003o0@00o`40000000<00?m]00D00000003of`00om/00?oK00000002002go`0500000000
COl004go001=o`0000000`0E0?l01@000000L03o0700o`1`0?l0000000<0`03o00D000000?l0o03o
0?`0o`3l000000020?l0b@050000003o0:H0o`2V0?l0YP0000000`3o09@01@000000o`2D0?l0U03o
09@0000000<0o`2V00D000000?l0b@3o0<T0o`39000000020?l0o005000000300?l0`03o0<00o`00
00000`1`0?l01@0000005@3o01D0o`0E0?l0000000<004go00D00000002go`00]ol00;Oo00000002
003of`0500000000ofd00?m]003oK@0000000`00o`401@000000IOl006Go001Uo`00000000<0a?l0
00D000000?oU003oi@00onD0000000020?nK00050000003oG`00oel00?mO000000000`3o<@001@00
0000oa800?lB003o4P00000000<0o`@000D000000?l5003o1@00o`D0000000020?lE00050000003o
=@00ocD00?le000000000`3oH@001@000000oiX00?nJ003oVP00000000<0omd000D000000=Oo003G
o`00eol00000000208Oo00050000000bo`00<_l003;o000000000`00ob<01@0000000?mh003oN000
ogP0000000<00?o:00D00000003Wo`00iol00>Oo00000002002Po`0500000000H_l006;o001Ro`00
00000`00;?l01@00000000;o0002o`000_l0000000<07@3o00D000000380o`0b0?l0<P3o00000002
03/0o`050000000j0?l0>P3o03X0o`0000000`0^0?l01@000000603o01P0o`0H0?l0000000<000Go
00@00000000Zo`00:_l000000`00EOl01@00000008Go0025o`00QOl0000000<00;So00D00000003/
o`00k?l00>co00000003003ogP030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`00
0000oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3o
ool01@3oool0100000000;co002lo`000003002=o`0500000000HOl0067o001Qo`0000000`00>Ol0
1@00000001Ko000Fo`005_l0000000<01P3o00@0000001`0o`0L0?l000000`0Z0?l01@000000;P3o
02h0o`0^0?l0000000<0:@3o00D0000001X0o`0J0?l06P3o000000030040o`04000000008_l002;o
000000<004oo00D000000025o`00QOl008Go000000030034o`0500000000ooH00?of003omP000000
0`00oj/0100000000?mL003oG0000003003o2`0500000016o`00A_l004Ko000000000`2Fo`001@00
0000hol00>?o003So`00000000<0omH000@000000?nF003oUP0000000`3oH0001@000000ocD00?le
003o=@00000000<0oaH000D000000?l4003o1000o`@0000000030?l000040000003o3000o``00000
00<0obD000D000000?m>003oCP00odh0000000030?n400050000003oa`00olL00?o7000000000`3Z
o`0010000000TOl0097o00000003037o000500000000oc@00?ld003o=00000000`00oih01@000000
0?Ko003fo`00m_l00000008008go00D00000000Xo`00:?l002So0000000303H0o`050000002;0?l0
R`3o08/0o`0000000`3G0?l01@000000o`3Z0?l0jP3o0>X000000080o`2j00D000000?l0V@3o09T0
o`2I000000030?l0R0050000003o08P0o`280?l0R00000000`3o09T01@000000o`2j0?l0^P3o0;X0
00000080o`3Z00D000000=L0o`3G0?l0e`3o0000000308/0o`050000000f0?l0=P3o03H0o`000000
0`00:?l01@00000008go002=o`00SOl0000000800?Ko00D00000003oWP00oih00?nN00000003003o
=0050000000ao`00<Ol0037o000000000`2Ao`001@000000j_l00>[o003Zo`0000000080olL000D0
00000?n4003oQ000oh@0000000030?m>00050000003o9@00obD00?lU000000000`3o30001@000000
o`000?l0003o000000000080o`@000D000000?lF003o5P00oaH0000000030?le00050000003oH000
of000?mP000000000`3oUP001@000000omH00?oF003oeP0000000080hol000D0000009Ko002Fo`00
U_l00000000304Ko000500000000o`/00?l;003o2`0000000`00oe`01@0000000?n[003oZ`00oj/0
000000800?of00D000000034o`00a?l00<Co000000030025o`0500000000Col004oo001?o`000000
0`008_l01@0000000@3o0040o`010?l0000000806P3o00D0000002T0o`0Y0?l0:@3o0000000302h0
o`050000000Z0?l0:P3o02X0o`0000000`0L0?l0100000001P3o00H0o`000003000Fo`0500000000
>Ol003Wo000io`0000000`00HOl01@00000008go002=o`00SOl0000000<00;co00<000000?ooo`3o
ool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000002lo`00_?l0
00000`00SOl01@000000067o001Qo`00HOl0000000<003Wo00D00000000Fo`005_l001Ko00000003
00H0o`040000000L0?l0703o000000<0:P3o00D0000002h0o`0^0?l0;P3o0000000302T0o`050000
000J0?l06P3o01X0o`0000000`010?l01000000002;o000Ro`000003001?o`0500000000QOl008Go
0025o`0000000`00a?l01@0000000?of003omP00ooH0000000<00?n[00@00000003oG000oe`00000
0`00o`/01@000000A_l004Ko0016o`00000000<0U_l000D000000>?o003So`00hol0000000030?oF
00040000003oUP00oiH0000000<0of0000D000000?le003o=@00ocD0000000030?lF00050000003o
1000o`@00?l4000000000`3o000010000000o``00?l<000000030?lU00050000003oCP00odh00?m>
000000000`3oQ0001@000000olL00?o7003oa`00000000<0j_l000@00000097o002Ao`0000000`0a
o`001@0000000?ld003o=000oc@0000000<00?nN00D00000003fo`00m_l00?Ko00000002002=o`05
00000000:?l002So000Xo`0000000`0f0?l01@000000R`3o08/0o`2;0?l0000000<0e`3o00D00000
0?l0jP3o0>X0o`3Z000000020?l0^P050000003o09T0o`2I0?l0V@0000000`3o08P01@000000o`28
0?l0R03o08P0000000<0o`2I00D000000?l0^P3o0;X0o`2j000000020?l0jP050000003G0?l0e`3o
0=L0o`0000000`2;0?l01@000000=P3o03H0o`0f0?l0000000<002So00D00000002=o`00SOl008go
00000002003fo`0500000000oih00?nN003oWP0000000`00oc@01@000000<Ol0037o000ao`000000
00<0TOl000D000000>[o003Zo`00j_l0000000020?o700050000003oQ000oh@00?n4000000000`3o
CP001@000000obD00?lU003o9@00000000<0o``000D000000?l0003o0000o`00000000020?l40005
0000003o5P00oaH00?lF000000000`3o=@001@000000of000?mP003oH000000000<0oiH000D00000
0?oF003oeP00omH0000000020>?o00050000002Fo`00U_l009Ko000000000`16o`001@0000000?l;
003o2`00o`/0000000<00?mL00D00000003oZ`00oj/00?n[00000002003omP0500000000a?l00<Co
0034o`0000000`00QOl01@00000004oo001?o`00Col0000000<002;o00D000000040o`010?l00@3o
0000000201X0o`050000000Y0?l0:@3o02T0o`0000000`0^0?l01@000000:P3o02X0o`0Z0?l00000
00<0703o00@0000000H0o`060?l000000`005_l01@00000003Wo000io`00>Ol0000000<0067o00D0
0000002=o`00SOl008go00000003002lo`030000003oool0oooo00@0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooo`0400000000_?l00;co000000<008go00D00000001Qo`00HOl0
067o00000003000io`05000000005_l001Ko000Fo`0000000`060?l010000000703o01`0o`000003
02X0o`050000000^0?l0;P3o02h0o`0000000`0Y0?l01@0000006P3o01X0o`0J0?l0000000<00@3o
00@00000000Ro`008_l000000`00Col01@00000008Go0025o`00QOl0000000<00<Co00D00000003o
mP00ooH00?of00000003003oZ`0400000000oe`00?mL000000<00?l;00D0000004Ko0016o`00A_l0
0000000309Ko00050000003So`00hol00>?o000000000`3oeP0010000000oiH00?nF000000030?mP
00050000003o=@00ocD00?le000000000`3o5P001@000000o`@00?l4003o1000000000<0o`0000@0
00000?l<003o300000000`3o9@001@000000odh00?m>003oCP00000000<0oh@000D000000?o7003o
a`00olL0000000030>[o00040000002Ao`00TOl0000000<0<Ol000D00000003o=000oc@00?ld0000
0003003oWP0500000000m_l00?Ko003fo`0000000P00SOl01@00000002So000Xo`00:?l0000000<0
=P3o00D0000008/0o`2;0?l0R`3o000000030=L0o`050000003o0>X0o`3Z0?l0jP0000000P3o0;X0
1@000000o`2I0?l0V@3o09T0000000<0o`2800D000000?l0R03o08P0o`28000000030?l0V@050000
003o0;X0o`2j0?l0^P0000000P3o0>X01@000000e`3o0=L0o`3G0?l0000000<0R`3o00D0000003H0
o`0f0?l0=P3o00000003000Xo`0500000000SOl008go002=o`0000000P00m_l01@0000000?nN003o
WP00oih0000000<00?ld00D00000037o000ao`00<Ol000000003097o00050000003Zo`00j_l00>[o
000000000P3oa`001@000000oh@00?n4003oQ000000000<0odh000D000000?lU003o9@00obD00000
00030?l<00050000003o0000o`000?l0000000000P3o10001@000000oaH00?lF003o5P00000000<0
ocD000D000000?mP003oH000of00000000030?nF00050000003oeP00omH00?oF000000000P3So`00
1@000000U_l009Ko002Fo`00000000<0A_l000D00000003o2`00o`/00?l;00000003003oG0050000
0000oj/00?n[003oZ`0000000P00ooH01@0000000<Co0034o`00a?l0000000<008Go00D00000001?
o`00Col004oo00000003000Ro`05000000010?l00@3o0040o`0000000P0J0?l01@000000:@3o02T0
o`0Y0?l0000000<0;P3o00D0000002X0o`0Z0?l0:P3o0000000301`0o`04000000060?l01P3o0000
00<001Ko00D00000000io`00>Ol003Wo00000003001Qo`0500000000SOl008go002=o`0000000`00
_?l00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo
0000084000001P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000001X
o`00J?l000000`00A?l01@00000002?o000So`008ol0000000<000So00D0000000/0o`0;0?l02`3o
0000000301T0o`040000000N0?l07P3o000000<06`3o00D0000000l0o`0?0?l03`3o000000030005
o`05000000008ol002?o000So`0000000`00BOl01000000007Wo001io`000003002`o`0500000000
k_l00>ko003^o`0000000`00old01@0000000?n5003oQ@00ohD0000000<00?li00@0000001?o000C
o`0000000`1Po`001@000000Zol00:_o002[o`00000000<0lol000D000000?o:003obP00olX00000
00030?n?00040000003oG000oe`0000000<0oc<000D000000?lE003o5@00oaD0000000030?l40005
0000003o0000o`000?l0000000000`3o200010000000oal00?lO000000030?m300050000003oM000
og@00?md000000000`3o/@001@000000ooT00?oi003on@00000000<0]?l000@0000005[o001Jo`00
00000`00o`D01@0000000?mZ003oJP00ofX0000000<00?oA00D000000038o`00b?l00<So00000002
001So`05000000000ol000?o0003o`0000000`1E0?l01@000000YP3o0:H0o`2V0?l0000000<0k03o
00D000000?l0f03o0=P0o`3H000000020?l0Z`050000003o08d0o`2=0?l0S@0000000`3o07d01@00
0000o`1m0?l0O@3o07d0000000<0o`2=00D000000?l0Z`3o0:/0o`2[000000020?l0f0050000003/
0?l0k03o0>`0o`0000000`2V0?l01@000000E@3o05D0o`1E0?l0000000<000?o00D00000001So`00
Hol006?o000000020038o`0500000000om400?oA003od@0000000`00ofX01@0000000?l5003o1@00
o`D0000000<0F_l000D000000;Co002do`00]?l0000000020?oi00050000003o/@00ok400?na0000
00000`3oM0001@000000od<00?m3003o@`00000000<0oal000D000000?l8003o2000o`P000000002
0?l000050000003o1000o`@00?l4000000000`3o5@001@000000oc<00?lc003o<`00000000<0oe`0
00D000000?n?003oS`00ohl0000000020?o:00050000003co`00lol00??o000000000`2[o`001@00
0000H?l0063o001Po`00000000<04ol000D00000003o>@00ocT00?li00000002003oQ@0500000000
old00?o=003oc@0000000`00k_l01@0000000;3o002`o`00/?l0000000<007Wo00D000000019o`00
BOl004Wo00000002000So`05000000001Ol000Go0005o`0000000`0?0?l01@0000006`3o01/0o`0K
0?l0000000<07P3o00@0000001T0o`0I0?l000000`0;0?l01@00000000So0008o`002?l0000000<0
02?o00D000000014o`00A?l004Co00000003001Xo`030000003oool0oooo00@0oooo0@0000010?oo
o`00803oool00`000000oooo0?ooo`050?ooo`0400000000J?l006So000000<004Co00D00000000S
o`008ol002?o000000030008o`050000000;0?l02`3o00/0o`0000000`0I0?l0100000007P3o01h0
o`00000301/0o`050000000?0?l03`3o00l0o`0000000`001Ol01@00000002?o000So`008ol00000
00<004Wo00@00000001io`00NOl000000`00/?l01@0000000>ko003^o`00k_l0000000<00?o=00D0
0000003oQ@00ohD00?n500000003003o>@040000000Co`004ol0000000<0H?l000D000000:_o002[
o`00Zol0000000030??o00050000003obP00olX00?o:000000000`3oS`0010000000oe`00?mL0000
00030?lc00050000003o5@00oaD00?lE000000000`3o10001@000000o`000?l0003o0000000000<0
o`P000@000000?lO003o7`0000000`3o@`001@000000og@00?md003oM000000000<0ok4000D00000
0?oi003on@00ooT0000000030;Co00040000001Jo`00F_l0000000<00?l500D00000003oJP00ofX0
0?mZ00000003003od@0500000000b?l00<So0038o`0000000P00Hol01@00000000?o0003o`000ol0
000000<0E@3o00D000000:H0o`2V0?l0YP3o000000030>`0o`050000003o0=P0o`3H0?l0f0000000
0P3o0:/01@000000o`2=0?l0S@3o08d0000000<0o`1m00D000000?l0O@3o07d0o`1m000000030?l0
S@050000003o0:/0o`2[0?l0Z`0000000P3o0=P01@000000k03o0>`0o`3/0?l0000000<0YP3o00D0
000005D0o`1E0?l0E@3o000000030003o`0500000000Hol006?o001So`0000000P00b?l01@000000
0?oA003od@00om40000000<00?mZ00D00000003o1@00o`D00?l50000000305[o00050000002do`00
]?l00;Co000000000P3on@001@000000ok400?na003o/@00000000<0og@000D000000?m3003o@`00
od<0000000030?lO00050000003o2000o`P00?l8000000000P3o00001@000000o`@00?l4003o1000
000000<0oaD000D000000?lc003o<`00oc<0000000030?mL00050000003oS`00ohl00?n?00000000
0P3obP001@000000lol00??o003co`00000000<0Zol000D00000063o001Po`00H?l00000000301?o
000500000000ocT00?li003o>@0000000P00ohD01@0000000?o=003oc@00old0000000<00>ko00D0
0000002`o`00/?l00;3o00000003001io`0500000000BOl004Wo0019o`0000000P008ol01@000000
00Go0005o`001Ol0000000<03`3o00D0000001/0o`0K0?l06`3o0000000301h0o`040000000I0?l0
6@3o000000<02`3o00D000000008o`002?l000So00000003000So`0500000000A?l004Co0014o`00
00000`00J?l00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool0
1@3oool01000000006So001Xo`0000030014o`05000000008ol002?o000So`0000000`002?l01@00
00002`3o00/0o`0;0?l0000000<06@3o00@0000001h0o`0N0?l000000`0K0?l01@0000003`3o00l0
o`0?0?l0000000<000Go00D00000000So`008ol002?o000000030019o`0400000000NOl007Wo0000
00<00;3o00D00000003^o`00k_l00>ko00000003003oc@0500000000ohD00?n5003oQ@0000000`00
ocT0100000004ol001?o00000003063o00050000002[o`00Zol00:_o000000000`3co`001@000000
olX00?o:003obP00000000<0ohl000@000000?mL003oG00000000`3o<`001@000000oaD00?lE003o
5@00000000<0o`@000D000000?l0003o0000o`00000000030?l800040000003o7`00oal0000000<0
od<000D000000?md003oM000og@0000000030?na00050000003on@00ooT00?oi000000000`2do`00
10000000F_l005[o00000003003o1@0500000000ofX00?mZ003oJP0000000`00om401@0000000<So
0038o`00b?l00000008006?o00D000000003o`000ol000?o0000000305D0o`050000002V0?l0YP3o
0:H0o`0000000`3/0?l01@000000o`3H0?l0f03o0=P000000080o`2[00D000000?l0S@3o08d0o`2=
000000030?l0O@050000003o07d0o`1m0?l0O@0000000`3o08d01@000000o`2[0?l0Z`3o0:/00000
0080o`3H00D000000>`0o`3/0?l0k03o000000030:H0o`050000001E0?l0E@3o05D0o`0000000`00
0ol01@00000006?o001So`00Hol0000000800<So00D00000003od@00om400?oA00000003003oJP05
00000000o`D00?l5003o1@0000000`1Jo`001@000000]?l00;Co002do`0000000080ooT000D00000
0?na003o/@00ok40000000030?md00050000003o@`00od<00?m3000000000`3o7`001@000000o`P0
0?l8003o200000000080o`0000D000000?l4003o1000o`@0000000030?lE00050000003o<`00oc<0
0?lc000000000`3oG0001@000000ohl00?n?003oS`0000000080olX000D000000??o003co`00lol0
000000030:_o00050000001Po`00H?l0063o000000000`0Co`001@0000000?li003o>@00ocT00000
00800?n500D00000003oc@00old00?o=00000003003^o`0500000000/?l00;3o002`o`0000000`00
NOl01@00000004Wo0019o`00BOl00000008002?o00D000000005o`001Ol000Go0000000300l0o`05
0000000K0?l06`3o01/0o`0000000`0N0?l0100000006@3o01T0o`00000300/0o`05000000002?l0
00So0008o`0000000`008ol01@00000004Co0014o`00A?l0000000<006So00<000000?ooo`3oool0
103oool100000040oooo000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@00
00010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000;_l002ko000000<001Oo00D0
00000005o`001Ol000Go0000000300H0o`050000000;0?l02`3o00/0o`0000000`090?l010000000
003o0000o`000003000Bo`0500000000;Ol002go000]o`0000000`00Col01@00000007Wo001io`00
NOl0000000<00:[o00@00000003Ro`00h_l000000`00on001@0000000?nN003oWP00oih0000000<0
0?mH00D00000003o4@00oa400?lA0000000303Oo00040000001oo`00Ool0000000<0aOl000D00000
0?oh003on000ooP0000000030?nk00050000003oQ000oh@00?n4000000000`3oE@0010000000oc00
0?l`000000030?lE00050000003o1@00o`D00?l5000000000`3o00001@000000o`P00?l8003o2000
000000<0oad000@000000?lm003o?@0000000`3oJP001@000000oj800?nR003oXP00000000<0onD0
00D000000<ko003>o`00c_l00000000307[o00040000000Po`008?l0000000<00?lo00D00000003o
X@00oj400?nQ00000003003ko`0500000000V?l009So002Ho`0000000P00>Ol01@0000008@3o0240
o`0Q0?l0000000<0M@3o00D000000<00o`300?l0`03o000000030?l0o@050000003o0<L0o`370?l0
a`0000000P3o09d01@000000o`210?l0P@3o0840000000<0o`1c00D000000?l0L`3o07<0o`1c0000
00030?l0P@050000003o09d0o`2M0?l0W@0000000P3o0<L01@000000o`3m0?l0o@3o0?d0000000<0
`03o00D0000007D0o`1e0?l0M@3o000000030240o`0500000000>Ol003Wo000io`0000000P00V?l0
1@0000000?_o003ko`00nol0000000<00?nQ00D00000003o?`00ocl00?lo00000003023o00050000
001jo`00N_l007[o000000000P3>o`001@000000onD00?oU003oi@00000000<0oj8000D000000?mZ
003oJP00ofX0000000030?lm00050000003o7@00oad00?lM000000000P3o20001@000000o`000?l0
003o0000000000<0o`D000D000000?lE003o5@00oaD0000000030?l`00050000003oE@00oeD00?mE
000000000P3oQ0001@000000ok/00?nk003o^`00000000<0ooP000D000000<Go0035o`00aOl00000
000307oo00050000000go`00=ol003Oo000000000P00oa401@0000000?mH003oF000oeP0000000<0
0?nN00D00000003oh000on000?oP00000003003Ro`0500000000Z_l00:[o002Zo`0000000P00NOl0
1@00000004oo001?o`00Col0000000<002go00D00000000Bo`004_l001;o000000030000o`040000
00090?l02@3o000000<02`3o00D0000000H0o`060?l01P3o000000030005o`05000000005ol001Oo
000Go`0000000`00;_l00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?oo
o`3oool01@3oool01000000002ko000^o`000003000Go`05000000001Ol000Go0005o`0000000`06
0?l01@0000002`3o00/0o`0;0?l0000000<02@3o00@000000000o`000?l000000`004_l01@000000
02go000]o`00;Ol0000000<004oo00D00000001io`00NOl007Wo00000003002Zo`0400000000h_l0
0>;o000000<00?oP00D00000003oWP00oih00?nN00000003003oF00500000000oa400?lA003o4@00
00000`0go`0010000000Ool007oo000000030<Go00050000003on000ooP00?oh000000000`3o^`00
1@000000oh@00?n4003oQ000000000<0oeD000@000000?l`003o<00000000`3o5@001@000000o`D0
0?l5003o1@00000000<0o`0000D000000?l8003o2000o`P0000000030?lM00040000003o?@00ocd0
000000<0ofX000D000000?nR003oXP00oj80000000030?oU00050000003>o`00c_l00<ko00000000
0`1jo`00100000008?l0023o00000003003o?`0500000000oj400?nQ003oX@0000000`00nol01@00
000009So002Ho`00V?l00000008003Wo00D000000240o`0Q0?l08@3o0000000307D0o`0500000030
0?l0`03o0<00o`0000000`3o0?d01@000000o`370?l0a`3o0<L000000080o`2M00D000000?l0P@3o
0840o`21000000030?l0L`050000003o07<0o`1c0?l0L`0000000`3o08401@000000o`2M0?l0W@3o
09d000000080o`3700D000000?l0o@3o0?d0o`3m000000030<00o`050000001e0?l0M@3o07D0o`00
00000`0Q0?l01@00000003Wo000io`00>Ol00000008009So00D00000003ko`00nol00?_o00000003
003oX@0500000000ocl00?lo003o?`0000000`0Po`001@000000N_l007[o001jo`0000000080c_l0
00D000000?oU003oi@00onD0000000030?nR00050000003oJP00ofX00?mZ000000000`3o?@001@00
0000oad00?lM003o7@0000000080o`P000D000000?l0003o0000o`00000000030?l500050000003o
5@00oaD00?lE000000000`3o<0001@000000oeD00?mE003oE@0000000080oh@000D000000?nk003o
^`00ok/0000000030?oh000500000035o`00aOl00<Go000000000`1oo`001@000000=ol003Oo000g
o`00000000800?lA00D00000003oF000oeP00?mH00000003003oWP0500000000on000?oP003oh000
00000`00h_l01@0000000:[o002Zo`00Z_l00000008007Wo00D00000001?o`00Col004oo00000003
000]o`05000000004_l001;o000Bo`0000000`000?l0100000002@3o00T0o`00000300/0o`050000
00060?l01P3o00H0o`0000000`001Ol01@00000001Oo000Go`005ol0000000<002ko00<000000?oo
o`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000000^o`00
;_l000000`005ol01@00000000Go0005o`001Ol0000000<01P3o00D0000000/0o`0;0?l02`3o0000
000300T0o`04000000000?l0003o000000<001;o00D00000000]o`00;Ol002go00000003001?o`05
00000000NOl007Wo001io`0000000`00Z_l0100000000>;o003Ro`000003003oh00500000000oih0
0?nN003oWP0000000`00oeP01@0000000?lA003o4@00oa40000000<0=ol000@0000007oo001oo`00
00000`35o`001@000000ooP00?oh003on000000000<0ok/000D000000?n4003oQ000oh@000000003
0?mE00040000003o<000oc00000000<0oaD000D000000?l5003o1@00o`D0000000030?l000050000
003o2000o`P00?l8000000000`3o7@0010000000ocd00?lm000000030?mZ00050000003oXP00oj80
0?nR000000000`3oi@001@000000c_l00<ko003>o`00000000<0N_l000@00000023o000Po`000000
0`00ocl01@0000000?nQ003oX@00oj40000000<00?_o00D00000002Ho`00V?l009So00000002000i
o`050000000Q0?l08@3o0240o`0000000`1e0?l01@000000`03o0<00o`300?l0000000<0o`3m00D0
00000?l0a`3o0<L0o`37000000020?l0W@050000003o0840o`210?l0P@0000000`3o07<01@000000
o`1c0?l0L`3o07<0000000<0o`2100D000000?l0W@3o09d0o`2M000000020?l0a`050000003o0?d0
o`3m0?l0o@0000000`300?l01@000000M@3o07D0o`1e0?l0000000<08@3o00D00000000io`00>Ol0
03Wo00000002002Ho`0500000000nol00?_o003ko`0000000`00oj401@0000000?lo003o?`00ocl0
000000<08?l000D0000007[o001jo`00N_l0000000020<ko00050000003oi@00onD00?oU00000000
0`3oXP001@000000ofX00?mZ003oJP00000000<0ocd000D000000?lM003o7@00oad0000000020?l8
00050000003o0000o`000?l0000000000`3o1@001@000000oaD00?lE003o5@00000000<0oc0000D0
00000?mE003oE@00oeD0000000020?n400050000003o^`00ok/00?nk000000000`3on0001@000000
aOl00<Go0035o`00000000<0Ool000D0000003Oo000go`00=ol000000002003o4@0500000000oeP0
0?mH003oF00000000`00oih01@0000000?oP003oh000on00000000<00>;o00D00000002Zo`00Z_l0
0:[o00000002001io`0500000000Col004oo001?o`0000000`00;Ol01@00000001;o000Bo`004_l0
000000<0003o00@0000000T0o`090?l000000`0;0?l01@0000001P3o00H0o`060?l0000000<000Go
00D00000000Go`005ol001Oo00000003000^o`030000003oool0oooo00@0oooo0@0000010?ooo`00
803oool00`000000oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<0
00000?ooo`3oool01@3oool01000000001Ko000Fo`000003000<o`05000000002Ol000Wo0009o`00
00000`003?l01@00000001Ko000Fo`005_l0000000<002Oo00@00000000oo`00?ol000000`00G_l0
1@00000008Co0024o`00Q?l0000000<00;3o00D00000003So`00hol00>?o00000003003oi@040000
0000ojT00?nY000000<00?mY00D00000003o9`00obL00?lW0000000301co00050000001Po`00H?l0
063o000000000`2Ro`0010000000hol00>?o000000030?oP00050000003oZ@00ojT00?nY00000000
0`3oM`001@000000odh00?m>003oCP00000000<0ob`000@000000?lD003o500000000`3o1`001@00
0000o`@00?l4003o1000000000<0o``000D000000?lO003o7`00oal0000000030?lm00040000003o
I`00ofL0000000<0oi/000D000000?oI003of@00omT0000000030=oo00050000002Ao`00TOl0097o
000000000`0lo`00100000000?lM003o7@000003003oNP0500000000omT00?oI003of@0000000`00
a_l01@00000006Wo001Yo`00JOl00000008000oo00D0000004D0o`150?l0A@3o0000000309@0o`05
0000003J0?l0fP3o0=X0o`0000000`3o0>P01@000000o`2f0?l0]P3o0;H000000080o`2@00D00000
0?l0MP3o07H0o`1f000000030?l0J0050000003o06P0o`1X0?l0J00000000`3o07H01@000000o`2@
0?l0T03o090000000080o`2f00D000000?l0j03o0>P0o`3X000000030=X0o`050000002D0?l0U03o
09@0o`0000000`150?l01@00000000oo000?o`003ol00000008006Wo00D000000036o`00a_l00<Ko
00000003003of@0500000000ogX00?mj003oNP0000000`00oad01@000000??l003co000lo`000000
0080TOl000D000000=oo003Oo`00gol0000000030?oI00050000003oV`00oi/00?nK000000000`3o
I`001@000000ocd00?lm003o?@0000000080oal000D000000?l<003o3000o``0000000030?l40005
0000003o1`00o`L00?l7000000000`3o50001@000000ob`00?l/003o;00000000080odh000D00000
0?mg003oM`00ogL0000000030?nY00050000003oh000on000?oP000000000`3So`001@000000X_l0
0:;o002Ro`0000000080H?l000D0000001co000Lo`007?l000000003003o9`0500000000ofT00?mY
003oJ@0000000`00ojT01@0000000?oU003oi@00onD0000000800>?o00D00000002`o`00/?l00;3o
000000030024o`0500000000G_l005ko001No`0000000`00?ol01000000002Oo000Wo`000003000F
o`05000000003?l000co000<o`0000000`002Ol01@00000000co000<o`003?l0000000<001Ko00<0
00000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000
000Fo`005_l000000`003?l01@00000000Wo0009o`002Ol0000000<000co00D00000000Fo`005_l0
01Ko00000003000Wo`0400000000?ol003oo000000<005ko00D000000024o`00Q?l008Co00000003
002`o`0500000000hol00>?o003So`0000000`00onD0100000000?nY003oZ@000003003oJ@050000
0000obL00?lW003o9`0000000`0Lo`001@000000H?l0063o001Po`00000000<0X_l000@000000>?o
003So`0000000`3oh0001@000000ojT00?nY003oZ@00000000<0ogL000D000000?m>003oCP00odh0
000000030?l/00040000003o5000oa@0000000<0o`L000D000000?l4003o1000o`@0000000030?l<
00050000003o7`00oal00?lO000000000`3o?@0010000000ofL00?mW000000030?nK00050000003o
f@00omT00?oI000000000`3Oo`001@000000TOl0097o002Ao`00000000<0??l000@00000003o7@00
oad000000`00ogX01@0000000?oI003of@00omT0000000<00<Ko00D00000001Yo`00JOl006Wo0000
0002000?o`05000000150?l0A@3o04D0o`0000000`2D0?l01@000000fP3o0=X0o`3J0?l0000000<0
o`3X00D000000?l0]P3o0;H0o`2f000000020?l0T0050000003o07H0o`1f0?l0MP0000000`3o06P0
1@000000o`1X0?l0J03o06P0000000<0o`1f00D000000?l0T03o0900o`2@000000020?l0]P050000
003o0>P0o`3X0?l0j00000000`3J0?l01@000000U03o09@0o`2D0?l0000000<0A@3o00D00000000?
o`003ol000oo00000002001Yo`0500000000a_l00<Ko0036o`0000000`00omT01@0000000?mj003o
NP00ogX0000000<00?lM00D0000003co000lo`00??l000000002097o00050000003Oo`00gol00=oo
000000000`3of@001@000000oi/00?nK003oV`00000000<0ofL000D000000?lm003o?@00ocd00000
00020?lO00050000003o3000o``00?l<000000000`3o10001@000000o`L00?l7003o1`00000000<0
oa@000D000000?l/003o;000ob`0000000020?m>00050000003oM`00ogL00?mg000000000`3oZ@00
1@000000on000?oP003oh000000000<0hol000D000000:;o002Ro`00X_l000000002063o00050000
000Lo`007?l001co000000000`00obL01@0000000?mY003oJ@00ofT0000000<00?nY00D00000003o
i@00onD00?oU00000002003So`0500000000/?l00;3o002`o`0000000`00Q?l01@00000005ko001N
o`00G_l0000000<003oo00@00000000Wo`009ol000000`005_l01@00000000co000<o`003?l00000
00<000Wo00D00000000<o`003?l000co00000003000Fo`030000003oool0oooo00@0oooo0@000001
0?ooo`00803oool00`000000oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200
oooo0P0000060?ooo`04000000008_l002;o000000<002Ko00D00000000ao`00<Ol0037o00000003
0012o`0500000000FOl005Wo001Io`0000000`00M_l01000000009Wo002Io`0000030031o`050000
0000kol00>oo003_o`0000000`00omd01@0000000?nV003oYP00ojH0000000<00?m/00@00000003o
;`00obl000000`0>o`001@000000COl004go001=o`00000000<0S?l000D000000<So0038o`00b?l0
000000030?om00040000003oaP00olH0000000<0oiD000D000000?mZ003oJP00ofX0000000030?m5
00050000003o:@00obT00?lY000000000`3o5@0010000000o`X00?l:000000030?l900050000003o
4P00oa800?lB000000000`3o9@001@000000od<00?m3003o@`00000000<0ofX000@000000?nK003o
V`0000000`3oe@001@000000j?l00>So003Xo`00000000<0Wol000D0000004oo001?o`00Col00000
0003003o1@0400000000oed00?mM000000<00?ng00D00000003/o`00k?l00>co00000003002Bo`05
00000000>Ol003Wo000io`0000000P0J0?l01@000000JP3o06X0o`1Z0?l0000000<0/P3o00D00000
0?<0o`3c0?l0l`3o000000030?l0e0050000003o0:H0o`2V0?l0YP0000000P3o08801@000000o`1[
0?l0J`3o06/0000000<0o`1O00D000000?l0G`3o05l0o`1O000000030?l0J`050000003o0880o`22
0?l0PP0000000P3o0:H01@000000o`3D0?l0e03o0=@0000000<0l`3o00D000000;80o`2b0?l0/P3o
0000000306X0o`050000000J0?l06P3o01X0o`0000000P00>Ol01@00000009;o002Bo`00T_l00000
00<00>co00D00000003o]`00okL00?ng00000003003oG@0500000000o`D00?l5003o1@0000000P1?
o`001@000000Wol009oo002Oo`00000000<0j?l000D000000?oE003oe@00omD0000000030?nK0005
0000003oJP00ofX00?mZ000000000P3o@`001@000000obD00?lU003o9@00000000<0oa8000D00000
0?l9003o2@00o`T0000000030?l:00050000003o5@00oaD00?lE000000000P3o:@001@000000odD0
0?m5003oA@00000000<0ofX000D000000?nE003oU@00oiD0000000030?o600050000003oo@00ood0
0?om000000000P38o`001@000000S?l008co002<o`00000000<0COl000D0000000ko000>o`003_l0
00000003003o;`0500000000of`00?m/003oK00000000P00ojH01@0000000?oM003og@00omd00000
00<00>oo00D000000031o`00`Ol00<7o00000003002Io`0400000000M_l007Ko000000<005Wo00D0
00000012o`00@_l004;o00000003000ao`05000000009_l002Ko000Vo`0000000`008_l00`000000
oooo0?ooo`030?ooo`8000000@3oool00200oooo00<000000?ooo`3oool01@3oool01000000002;o
000Ro`000003000Vo`0500000000<Ol0037o000ao`0000000`00@_l01@00000005Wo001Io`00FOl0
000000<007Ko00@00000002Io`00VOl000000`00`Ol01@0000000>oo003_o`00kol0000000<00?oM
00D00000003oYP00ojH00?nV00000003003oK00400000000obl00?l_000000<03_l000D0000004go
001=o`00COl00000000308co000500000038o`00b?l00<So000000000`3oo@0010000000olH00?o6
000000030?nE00050000003oJP00ofX00?mZ000000000`3oA@001@000000obT00?lY003o:@000000
00<0oaD000@000000?l:003o2P0000000`3o2@001@000000oa800?lB003o4P00000000<0obD000D0
00000?m3003o@`00od<0000000030?mZ00040000003oV`00oi/0000000<0omD000D000000>So003X
o`00j?l00000000309oo00050000001?o`00Col004oo000000000`00o`D0100000000?mM003oG@00
0003003o]`0500000000k?l00>co003/o`0000000`00T_l01@00000003Wo000io`00>Ol000000080
6P3o00D0000006X0o`1Z0?l0JP3o000000030;80o`050000003c0?l0l`3o0?<0o`0000000`3o0=@0
1@000000o`2V0?l0YP3o0:H000000080o`2200D000000?l0J`3o06/0o`1[000000030?l0G`050000
003o05l0o`1O0?l0G`0000000`3o06/01@000000o`220?l0PP3o088000000080o`2V00D000000?l0
e03o0=@0o`3D000000030?<0o`050000002b0?l0/P3o0;80o`0000000`1Z0?l01@0000006P3o01X0
o`0J0?l00000008003Wo00D00000002Bo`00T_l009;o00000003003/o`0500000000okL00?ng003o
]`0000000`00oed01@0000000?l5003o1@00o`D000000080Col000D0000009oo002Oo`00Wol00000
00030>So00050000003oe@00omD00?oE000000000`3oV`001@000000ofX00?mZ003oJP0000000080
od<000D000000?lU003o9@00obD0000000030?lB00050000003o2@00o`T00?l9000000000`3o2P00
1@000000oaD00?lE003o5@0000000080obT000D000000?m5003oA@00odD0000000030?mZ00050000
003oU@00oiD00?nE000000000`3oaP001@000000ood00?om003oo@0000000080b?l000D0000008co
002<o`00S?l00000000304go00050000000>o`003_l000ko000000000`00obl01@0000000?m/003o
K000of`0000000800?nV00D00000003og@00omd00?oM00000003003_o`0500000000`Ol00<7o0031
o`0000000`00VOl01000000007Ko001fo`000003001Io`0500000000@_l004;o0012o`0000000`00
<Ol01@00000002Ko000Vo`009_l0000000<002;o00<000000?ooo`3oool0103oool100000040oooo
000P0?ooo`030000003oool0oooo00D0oooo00@00000000Ro`008_l000000`009_l01@000000037o
000ao`00<Ol0000000<004;o00D00000001Io`00FOl005Wo00000003001fo`0400000000VOl009Wo
000000<00<7o00D00000003_o`00kol00>oo00000003003og@0500000000ojH00?nV003oYP000000
0`00of`0100000000?l_003o;`00000300ko00050000001=o`00COl004go000000000`2<o`001@00
0000b?l00<So0038o`00000000<0ood000@000000?o6003oaP0000000`3oU@001@000000ofX00?mZ
003oJP00000000<0odD000D000000?lY003o:@00obT0000000030?lE00040000003o2P00o`X00000
00<0o`T000D000000?lB003o4P00oa80000000030?lU00050000003o@`00od<00?m3000000000`3o
JP0010000000oi/00?nK000000030?oE00050000003Xo`00j?l00>So000000000`2Oo`001@000000
Col004oo001?o`00000000<00?l500@00000003oG@00oed000000`00okL01@0000000>co003/o`00
k?l0000000<009;o00D00000000io`00>Ol003Wo0000000201X0o`050000001Z0?l0JP3o06X0o`00
00000`2b0?l01@000000l`3o0?<0o`3c0?l0000000<0o`3D00D000000?l0YP3o0:H0o`2V00000002
0?l0PP050000003o06/0o`1[0?l0J`0000000`3o05l01@000000o`1O0?l0G`3o05l0000000<0o`1[
00D000000?l0PP3o0880o`22000000020?l0YP050000003o0=@0o`3D0?l0e00000000`3c0?l01@00
0000/P3o0;80o`2b0?l0000000<0JP3o00D0000001X0o`0J0?l06P3o00000002000io`0500000000
T_l009;o002Bo`0000000`00k?l01@0000000?ng003o]`00okL0000000<00?mM00D00000003o1@00
o`D00?l50000000204oo00050000002Oo`00Wol009oo000000000`3Xo`001@000000omD00?oE003o
e@00000000<0oi/000D000000?mZ003oJP00ofX0000000020?m300050000003o9@00obD00?lU0000
00000`3o4P001@000000o`T00?l9003o2@00000000<0o`X000D000000?lE003o5@00oaD000000002
0?lY00050000003oA@00odD00?m5000000000`3oJP001@000000oiD00?nE003oU@00000000<0olH0
00D000000?om003oo@00ood0000000020<So00050000002<o`00S?l008co000000000`1=o`001@00
00003_l000ko000>o`00000000<00?l_00D00000003oK000of`00?m/00000002003oYP0500000000
omd00?oM003og@0000000`00kol01@0000000<7o0031o`00`Ol0000000<009Wo00@00000001fo`00
M_l000000`00FOl01@00000004;o0012o`00@_l0000000<0037o00D00000000Vo`009_l002Ko0000
0003000Ro`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`05
0?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool01000
000005;o001Bo`000003001So`0500000000N_l007[o001jo`0000000`00UOl01@0000000;Ko002f
o`00]_l0000000<00=co00@00000003on000ooP000000`00olX01@0000000?nG003oU`00oiL00000
00<00?mR00D00000003o:`00ob/00?l[0000000300go000400000017o`00Aol0000000<0POl000D0
00000;Wo002io`00^Ol0000000030?3o00050000003og000om`00?oL000000000`3o[00010000000
oh800?n2000000030?mM00050000003o?P00och00?ln000000000`3o9`001@000000oaL00?lG003o
5`00000000<0oa0000@000000?lB003o4P0000000`3o7@001@000000oc400?la003o<@00000000<0
odh000D000000?md003oM000og@0000000030?nR00040000003of@00omT0000000<0j?l000D00000
0:?o002So`00Xol00000000305So000500000009o`002Ol000Wo000000000`00odT0100000000?nO
003oW`000003003omP0500000000/_l00;;o002bo`0000000`00GOl01@00000000[o000:o`002_l0
00000080@`3o00D0000008d0o`2=0?l0S@3o000000030=00o`050000003o0?<0o`3c0?l0l`000000
0`3o0<001@000000o`2F0?l0UP3o09H000000080o`1f00D000000?l0H03o0600o`1P000000030?l0
E@050000003o05D0o`1E0?l0E@0000000`3o06001@000000o`1f0?l0MP3o07H000000080o`2F00D0
00000?l0`03o0<00o`30000000030?l0l`050000003@0?l0d03o0=00o`0000000`2=0?l01@000000
@`3o04<0o`130?l00000008000[o00D00000001Mo`00GOl005go00000003002bo`0500000000ooH0
0?of003omP0000000`00oil01@0000000?m9003oB@00odT0000000802Ol000D0000005So001Ho`00
F?l0000000030:?o00050000003Xo`00j?l00>So000000000`3of@001@000000oj800?nR003oXP00
00000080og@000D000000?m>003oCP00odh0000000030?la00050000003o7@00oad00?lM00000000
0`3o4P001@000000oa000?l@003o400000000080oaL000D000000?lW003o9`00obL0000000030?ln
00050000003oG@00oed00?mM000000000`3oPP001@000000oj`00?n/003o[00000000080om`000D0
00000?3o003`o`00l?l0000000030;Wo000500000021o`00POl0087o000000000`17o`001@000000
3Ol000go000=o`00000000800?l[00D00000003oHP00of800?mR00000003003oU`0500000000olX0
0?o:003obP0000000`00ooP0100000000=co003Lo`000003002fo`0500000000UOl009Go002Eo`00
00000`00N_l01@00000006?o001So`00Hol0000000<005;o00<000000?ooo`3oool0103oool10000
0040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000001Bo`00D_l000000`00Hol01@00
000007[o001jo`00N_l0000000<009Go00D00000002fo`00]_l00;Ko00000003003Lo`0400000000
ooP00?oh000000<00?o:00D00000003oU`00oiL00?nG00000003003oHP0500000000ob/00?l[003o
:`0000000`0=o`0010000000Aol004Oo00000003087o00050000002io`00^Ol00;Wo000000000`3`
o`001@000000om`00?oL003og000000000<0oj`000@000000?n2003oPP0000000`3oG@001@000000
och00?ln003o?P00000000<0obL000D000000?lG003o5`00oaL0000000030?l@00040000003o4P00
oa80000000<0oad000D000000?la003o<@00oc40000000030?m>00050000003oM000og@00?md0000
00000`3oXP0010000000omT00?oI000000030>So00050000002So`00Xol00:?o000000000`1Ho`00
1@0000002Ol000Wo0009o`00000000<00?m900@00000003oW`00oil000000`00ooH01@0000000;;o
002bo`00/_l0000000<005go00D00000000:o`002_l000[o0000000204<0o`050000002=0?l0S@3o
08d0o`0000000`3@0?l01@000000o`3c0?l0l`3o0?<0000000<0o`3000D000000?l0UP3o09H0o`2F
000000020?l0MP050000003o0600o`1P0?l0H00000000`3o05D01@000000o`1E0?l0E@3o05D00000
00<0o`1P00D000000?l0MP3o07H0o`1f000000020?l0UP050000003o0<00o`300?l0`00000000`3o
0?<01@000000d03o0=00o`3@0?l0000000<0S@3o00D0000004<0o`130?l0@`3o00000002000:o`05
00000000GOl005go001Mo`0000000`00/_l01@0000000?of003omP00ooH0000000<00?nO00D00000
003oB@00odT00?m90000000200Wo00050000001Ho`00F?l005So000000000`2So`001@000000j?l0
0>So003Xo`00000000<0omT000D000000?nR003oXP00oj80000000020?md00050000003oCP00odh0
0?m>000000000`3o<@001@000000oad00?lM003o7@00000000<0oa8000D000000?l@003o4000oa00
000000020?lG00050000003o9`00obL00?lW000000000`3o?P001@000000oed00?mM003oG@000000
00<0oh8000D000000?n/003o[000oj`0000000020?oL00050000003`o`00l?l00?3o000000000`2i
o`001@000000POl0087o0021o`00000000<0Aol000D0000000go000=o`003Ol000000002003o:`05
00000000of800?mR003oHP0000000`00oiL01@0000000?o:003obP00olX0000000<00?oh00@00000
003Lo`00g?l000000`00]_l01@00000009Go002Eo`00UOl0000000<007[o00D00000001So`00Hol0
06?o00000003001Bo`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo
0?ooo`050?ooo`0400000000D_l005;o000000<006?o00D00000001jo`00N_l007[o00000003002E
o`0500000000]_l00;Ko002fo`0000000`00g?l0100000000?oh003on0000003003obP0500000000
oiL00?nG003oU`0000000`00of801@0000000?l[003o:`00ob/0000000<03Ol000@0000004Oo0017
o`0000000`21o`001@000000^Ol00;Wo002io`00000000<0l?l000D000000?oL003og000om`00000
00030?n/00040000003oPP00oh80000000<0oed000D000000?ln003o?P00och0000000030?lW0005
0000003o5`00oaL00?lG000000000`3o400010000000oa800?lB000000030?lM00050000003o<@00
oc400?la000000000`3oCP001@000000og@00?md003oM000000000<0oj8000@000000?oI003of@00
00000`3Xo`001@000000Xol00:?o002So`00000000<0F?l000D0000000Wo0009o`002Ol000000003
003oB@0400000000oil00?nO000000<00?of00D00000002bo`00/_l00;;o00000003001Mo`050000
00002_l000[o000:o`0000000P130?l01@000000S@3o08d0o`2=0?l0000000<0d03o00D000000?l0
l`3o0?<0o`3c000000030?l0`0050000003o09H0o`2F0?l0UP0000000P3o07H01@000000o`1P0?l0
H03o0600000000<0o`1E00D000000?l0E@3o05D0o`1E000000030?l0H0050000003o07H0o`1f0?l0
MP0000000P3o09H01@000000o`300?l0`03o0<00000000<0o`3c00D000000=00o`3@0?l0d03o0000
000308d0o`05000000130?l0@`3o04<0o`0000000P002_l01@00000005go001Mo`00GOl0000000<0
0;;o00D00000003omP00ooH00?of00000003003oW`0500000000odT00?m9003oB@0000000P09o`00
1@000000F?l005So001Ho`00000000<0Xol000D000000>So003Xo`00j?l0000000030?oI00050000
003oXP00oj800?nR000000000P3oM0001@000000odh00?m>003oCP00000000<0oc4000D000000?lM
003o7@00oad0000000030?lB00050000003o4000oa000?l@000000000P3o5`001@000000obL00?lW
003o9`00000000<0och000D000000?mM003oG@00oed0000000030?n200050000003o[000oj`00?n/
000000000P3og0001@000000l?l00?3o003`o`00000000<0^Ol000D00000087o0021o`00POl00000
000304Oo00050000000=o`003Ol000go000000000P00ob/01@0000000?mR003oHP00of80000000<0
0?nG00D00000003obP00olX00?o:00000003003on00400000000g?l00=co000000<00;Ko00D00000
002Eo`00UOl009Go00000003001jo`0500000000Hol006?o001So`0000000`00D_l00`000000oooo
0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo0000084000001P3o
ool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000002Qo`00XOl000000`00
_Ol01@0000000=co003Lo`00g?l0000000<00?oo00D00000003of000omP00?oH00000003003o[@04
00000000ogl00?mo000000<00?m>00D00000003o6`00oa/00?lK0000000301Oo00050000001<o`00
C?l004co000000000`21o`0010000000]?l00;Co000000030>Ko00050000003oj@00onT00?oY0000
00000`3o_@001@000000oi@00?nD003oU000000000<0og0000@000000?mB003oDP0000000`3o>P00
1@000000obP00?lX003o:000000000<0oad000D000000?lJ003o6P00oaX0000000030?lO00040000
003o;000ob`0000000<0od4000D000000?mO003oG`00oel0000000030?n400050000003o/@00ok40
0?na000000000`3oi@0010000000gol00=oo0000000309oo00050000001Ho`00F?l005So00000000
0`0>o`001@0000000?lo003o?`00ocl0000000<00?n@00@00000003ohP00on8000000`00b_l01@00
000007So001ho`00N?l0000000<002Wo00D0000002<0o`0S0?l08`3o0000000206`0o`050000002`
0?l0/03o0;00o`0000000`3]0?l01@000000o`3L0?l0g03o0=`0000000<0o`2]00D000000?l0Q`3o
08L0o`27000000020?l0JP050000003o05H0o`1F0?l0EP0000000`3o04`01@000000o`1<0?l0C03o
04`0000000<0o`1F00D000000?l0JP3o06X0o`1Z000000020?l0Q`050000003o0:d0o`2]0?l0[@00
00000`3o0=`01@000000k@3o0>d0o`3]0?l0000000<0/03o00D0000006`0o`1/0?l0K03o00000002
02<0o`0500000000:Ol002Wo000Yo`0000000`00N?l01@0000000<[o003:o`00b_l0000000<00?oR
00D00000003oT000oi000?n@00000002003o?`050000000>o`003_l000ko000000000`1Ho`001@00
0000Wol009oo002Oo`00000000<0gol000D000000?oU003oi@00onD0000000020?na00050000003o
Q000oh@00?n4000000000`3oG`001@000000od400?m1003o@@00000000<0ob`000D000000?lO003o
7`00oal0000000020?lJ00050000003o7@00oad00?lM000000000`3o:0001@000000ocX00?lj003o
>P00000000<0oe8000D000000?m`003oL000og00000000020?nD00050000003o_@00okd00?nm0000
00000`3oj@001@000000i_l00>Ko003Vo`00000000<0]?l000D00000087o0021o`00POl000000002
04co00050000000Go`005ol001Oo000000000`00oa/01@0000000?m>003oCP00odh0000000<00?mo
00@00000003o[@00ojd000000`00omP01@0000000?oo003oo`00ool0000000<00=co00D00000002m
o`00_Ol00;go00000003002Qo`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`00
0000oooo0?ooo`050?ooo`0400000000XOl00:7o000000<00;go00D00000003Lo`00g?l00=co0000
0003003oo`0500000000omP00?oH003of00000000`00ojd0100000000?mo003oO`000003003oCP05
00000000oa/00?lK003o6`0000000`0Go`001@000000C?l004co001<o`00000000<0POl000@00000
0;Co002do`0000000`3Vo`001@000000onT00?oY003oj@00000000<0okd000D000000?nD003oU000
oi@0000000030?m`00040000003oDP00oe80000000<0ocX000D000000?lX003o:000obP000000003
0?lM00050000003o6P00oaX00?lJ000000000`3o7`0010000000ob`00?l/000000030?m100050000
003oG`00oel00?mO000000000`3oQ0001@000000ok400?na003o/@00000000<0onD000@000000=oo
003Oo`0000000`2Oo`001@000000F?l005So001Ho`00000000<03_l000D00000003o?`00ocl00?lo
00000003003oT00400000000on800?oR000000<00<[o00D00000001ho`00N?l007So00000003000Y
o`050000000S0?l08`3o02<0o`0000000P1/0?l01@000000/03o0;00o`2`0?l0000000<0k@3o00D0
00000?l0g03o0=`0o`3L000000030?l0[@050000003o08L0o`270?l0Q`0000000P3o06X01@000000
o`1F0?l0EP3o05H0000000<0o`1<00D000000?l0C03o04`0o`1<000000030?l0EP050000003o06X0
o`1Z0?l0JP0000000P3o08L01@000000o`2]0?l0[@3o0:d0000000<0o`3L00D000000>d0o`3]0?l0
k@3o000000030;00o`050000001/0?l0K03o06`0o`0000000P0S0?l01@00000002Wo000Yo`00:Ol0
000000<007So00D00000003:o`00b_l00<[o00000003003ohP0500000000oi000?n@003oT0000000
0P00ocl01@0000003_l000ko000>o`00000000<0F?l000D0000009oo002Oo`00Wol0000000030=oo
00050000003oi@00onD00?oU000000000P3o/@001@000000oh@00?n4003oQ000000000<0oel000D0
00000?m1003o@@00od40000000030?l/00050000003o7`00oal00?lO000000000P3o6P001@000000
oad00?lM003o7@00000000<0obP000D000000?lj003o>P00ocX0000000030?mB00050000003oL000
og000?m`000000000P3oU0001@000000okd00?nm003o_@00000000<0onT000D000000>Ko003Vo`00
i_l0000000030;Co000500000021o`00POl0087o000000000P1<o`001@0000005ol001Oo000Go`00
000000<00?lK00D00000003oCP00odh00?m>00000003003oO`0400000000ojd00?n]000000<00?oH
00D00000003oo`00ool00?oo00000003003Lo`0500000000_Ol00;go002mo`0000000`00XOl00`00
0000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool010000000
0:7o002Qo`000003002mo`0500000000g?l00=co003Lo`0000000`00ool01@0000000?oH003of000
omP0000000<00?n]00@00000003oO`00ogl000000`00odh01@0000000?lK003o6`00oa/0000000<0
5ol000D0000004co001<o`00C?l000000003087o00040000002do`00]?l0000000<0i_l000D00000
0?oY003oj@00onT0000000030?nm00050000003oU000oi@00?nD000000000`3oL00010000000oe80
0?mB000000030?lj00050000003o:000obP00?lX000000000`3o7@001@000000oaX00?lJ003o6P00
000000<0oal000@000000?l/003o;00000000`3o@@001@000000oel00?mO003oG`00000000<0oh@0
00D000000?na003o/@00ok40000000030?oU00040000003Oo`00gol0000000<0Wol000D0000005So
001Ho`00F?l00000000300ko000500000000ocl00?lo003o?`0000000`00oi00100000000?oR003o
hP000003003:o`0500000000N?l007So001ho`0000000`00:Ol01@0000008`3o02<0o`0S0?l00000
0080K03o00D000000;00o`2`0?l0/03o000000030>d0o`050000003o0=`0o`3L0?l0g00000000`3o
0:d01@000000o`270?l0Q`3o08L000000080o`1Z00D000000?l0EP3o05H0o`1F000000030?l0C005
0000003o04`0o`1<0?l0C00000000`3o05H01@000000o`1Z0?l0JP3o06X000000080o`2700D00000
0?l0[@3o0:d0o`2]000000030?l0g0050000003]0?l0k@3o0>d0o`0000000`2`0?l01@000000K03o
06`0o`1/0?l0000000808`3o00D00000000Yo`00:Ol002Wo00000003001ho`0500000000b_l00<[o
003:o`0000000`00on801@0000000?n@003oT000oi00000000800?lo00D0000000ko000>o`003_l0
0000000305So00050000002Oo`00Wol009oo000000000`3Oo`001@000000onD00?oU003oi@000000
0080ok4000D000000?n4003oQ000oh@0000000030?mO00050000003o@@00od400?m1000000000`3o
;0001@000000oal00?lO003o7`0000000080oaX000D000000?lM003o7@00oad0000000030?lX0005
0000003o>P00ocX00?lj000000000`3oDP001@000000og000?m`003oL00000000080oi@000D00000
0?nm003o_@00okd0000000030?oY00050000003Vo`00i_l00>Ko000000000`2do`001@000000POl0
087o0021o`0000000080C?l000D0000001Oo000Go`005ol000000003003o6`0500000000odh00?m>
003oCP0000000`00ogl0100000000?n]003o[@000003003of00500000000ool00?oo003oo`000000
0`00g?l01@0000000;go002mo`00_Ol0000000<00:7o00<000000?ooo`3oool0103oool100000040
oooo000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooo`0400000000ooH00?of000000<00?oD00D00000003o[`00ojl0
0?n_00000003003oQ`0500000000oed00?mM003oG@0000000`00oc40100000000?l3003o0`000003
02_o00050000001Jo`00F_l005[o000000000`29o`001@000000^?l00;So002ho`00000000<0iOl0
00@000000?o_003ok`0000000`3oaP001@000000oj400?nQ003oX@00000000<0oh0000D000000?mS
003oH`00of<0000000030?m<00040000003o>P00ocX0000000<0obh000D000000?lY003o:@00obT0
000000030?lZ00050000003o<P00oc800?lb000000000`3o@P0010000000oeT00?mI000000030?mf
00050000003oV`00oi/00?nK000000000`3oa`001@000000ooT00?oi003on@00000000<0c_l000@0
0000097o002Ao`0000000`1?o`001@0000002Ol000Wo0009o`00000000<00?lo00D00000003oR`00
oh/00?n;00000003003of00400000000fOl00=Wo000000<008_o00D00000000oo`00?ol003oo0000
000300X0o`050000001A0?l0D@3o0540o`0000000P2C0?l01@000000d@3o0=40o`3A0?l0000000<0
o`3f00D000000?l0a@3o0<D0o`35000000030?l0V`050000003o07P0o`1h0?l0N00000000P3o05h0
1@000000o`1<0?l0C03o04`0000000<0o`1300D000000?l0@`3o04<0o`13000000030?l0C0050000
003o05h0o`1N0?l0GP0000000P3o07P01@000000o`2K0?l0V`3o09/0000000<0o`3500D000000?l0
mP3o0?H0o`3f000000030=40o`050000002C0?l0T`3o09<0o`0000000P1A0?l01@0000002P3o00X0
o`0:0?l0000000<003oo00D00000002;o`00Rol008_o00000003003Io`0500000000omP00?oH003o
f00000000P00oh/01@0000000?lo003o?`00ocl0000000<02Ol000D0000004oo001?o`00Col00000
0003097o00050000003>o`00c_l00<ko000000000P3on@001@000000olL00?o7003oa`00000000<0
oi/000D000000?mf003oMP00ogH0000000030?mI00050000003o@P00od800?m2000000000P3o<P00
1@000000obX00?lZ003o:P00000000<0obT000D000000?l^003o;P00obh0000000030?lj00050000
003oC000od`00?m<000000000P3oH`001@000000oh000?n0003oP000000000<0oj4000D000000?o6
003oaP00olH0000000030?o_00050000003Uo`00iOl00>Go000000000P2ho`001@000000ROl008Wo
0029o`00000000<0F_l000D0000002_o000[o`00:ol000000003003o0`0400000000oc400?la0000
00<00?mM00D00000003oQ`00ohL00?n700000003003o[`0500000000om@00?oD003oe00000000`00
ooH00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool0
100000000?of003omP000003003oe00500000000ojl00?n_003o[`0000000`00ohL01@0000000?mM
003oG@00oed0000000<00?la00@00000003o0`00o`<000000`0[o`001@000000F_l005[o001Jo`00
000000<0ROl000D000000;So002ho`00^?l0000000030>Go00040000003ok`00onl0000000<0olH0
00D000000?nQ003oX@00oj40000000030?n000050000003oH`00of<00?mS000000000`3oC0001000
0000ocX00?lj000000030?l^00050000003o:@00obT00?lY000000000`3o:P001@000000oc800?lb
003o<P00000000<0od8000@000000?mI003oF@0000000`3oMP001@000000oi/00?nK003oV`000000
00<0olL000D000000?oi003on@00ooT0000000030<ko00040000002Ao`00TOl0000000<0Col000D0
000000Wo0009o`002Ol000000003003o?`0500000000oh/00?n;003oR`0000000`00omP010000000
0=Wo003Io`000003002;o`0500000000?ol003oo000oo`0000000`0:0?l01@000000D@3o0540o`1A
0?l000000080T`3o00D000000=40o`3A0?l0d@3o000000030?l0mP050000003o0<D0o`350?l0a@00
00000`3o09/01@000000o`1h0?l0N03o07P000000080o`1N00D000000?l0C03o04`0o`1<00000003
0?l0@`050000003o04<0o`130?l0@`0000000`3o04`01@000000o`1N0?l0GP3o05h000000080o`1h
00D000000?l0V`3o09/0o`2K000000030?l0a@050000003o0?H0o`3f0?l0mP0000000`3A0?l01@00
0000T`3o09<0o`2C0?l000000080D@3o00D0000000X0o`0:0?l02P3o00000003000oo`0500000000
Rol008_o002;o`0000000`00fOl01@0000000?oH003of000omP0000000800?n;00D00000003o?`00
ocl00?lo0000000300Wo00050000001?o`00Col004oo000000000`2Ao`001@000000c_l00<ko003>
o`0000000080ooT000D000000?o7003oa`00olL0000000030?nK00050000003oMP00ogH00?mf0000
00000`3oF@001@000000od800?m2003o@P0000000080oc8000D000000?lZ003o:P00obX000000003
0?lY00050000003o;P00obh00?l^000000000`3o>P001@000000od`00?m<003oC00000000080of<0
00D000000?n0003oP000oh00000000030?nQ00050000003oaP00olH00?o6000000000`3ok`001@00
0000iOl00>Go003Uo`0000000080^?l000D0000008Wo0029o`00ROl00000000305[o00050000000[
o`00:ol002_o000000000`00o`<0100000000?la003o<@000003003oG@0500000000ohL00?n7003o
Q`0000000`00ojl01@0000000?oD003oe000om@0000000<00?of00<000000?ooo`3oool0103oool1
00000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003omP00ooH000000`00om@0
1@0000000?n_003o[`00ojl0000000<00?n700D00000003oG@00oed00?mM00000003003o<@040000
0000o`<00?l3000000<0:ol000D0000005[o001Jo`00F_l00000000308Wo00050000002ho`00^?l0
0;So000000000`3Uo`0010000000onl00?o_000000030?o600050000003oX@00oj400?nQ00000000
0`3oP0001@000000of<00?mS003oH`00000000<0od`000@000000?lj003o>P0000000`3o;P001@00
0000obT00?lY003o:@00000000<0obX000D000000?lb003o<P00oc80000000030?m200040000003o
F@00oeT0000000<0ogH000D000000?nK003oV`00oi/0000000030?o700050000003on@00ooT00?oi
000000000`3>o`0010000000TOl0097o0000000304oo000500000009o`002Ol000Wo000000000`00
ocl01@0000000?n;003oR`00oh/0000000<00?oH00@00000003Io`00fOl000000`00Rol01@000000
03oo000oo`00?ol0000000<02P3o00D000000540o`1A0?l0D@3o0000000209<0o`050000003A0?l0
d@3o0=40o`0000000`3o0?H01@000000o`350?l0a@3o0<D0000000<0o`2K00D000000?l0N03o07P0
o`1h000000020?l0GP050000003o04`0o`1<0?l0C00000000`3o04<01@000000o`130?l0@`3o04<0
000000<0o`1<00D000000?l0GP3o05h0o`1N000000020?l0N0050000003o09/0o`2K0?l0V`000000
0`3o0<D01@000000o`3f0?l0mP3o0?H0000000<0d@3o00D0000009<0o`2C0?l0T`3o000000020540
o`050000000:0?l02P3o00X0o`0000000`00?ol01@00000008_o002;o`00Rol0000000<00=Wo00D0
0000003of000omP00?oH00000002003oR`0500000000ocl00?lo003o?`0000000`09o`001@000000
Col004oo001?o`00000000<0TOl000D000000<ko003>o`00c_l0000000020?oi00050000003oa`00
olL00?o7000000000`3oV`001@000000ogH00?mf003oMP00000000<0oeT000D000000?m2003o@P00
od80000000020?lb00050000003o:P00obX00?lZ000000000`3o:@001@000000obh00?l^003o;P00
000000<0ocX000D000000?m<003oC000od`0000000020?mS00050000003oP000oh000?n000000000
0`3oX@001@000000olH00?o6003oaP00000000<0onl000D000000>Go003Uo`00iOl0000000020;So
000500000029o`00ROl008Wo000000000`1Jo`001@000000:ol002_o000[o`00000000<00?l300@0
0000003o<@00oc4000000`00oed01@0000000?n7003oQ`00ohL0000000<00?n_00D00000003oe000
om@00?oD00000003003omP030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000
oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool0
1@3oool0100000000?n0003oP0000003003oG00500000000ocD00?le003o=@0000000`00o`d01@00
00006ol001_o000Ko`00000000<0AOl000@0000006oo001_o`0000000`2Io`001@000000`_l00<;o
0032o`00000000<0j_l000D000000?o_003ok`00onl0000000030?o;00040000003oZP00ojX00000
00<0oh`000D000000?mc003oL`00og<0000000030?mM00050000003oC@00odd00?m=000000000`3o
@@0010000000oc/00?lk000000030?lk00050000003o@000od000?m0000000000`3oC0001@000000
oel00?mO003oG`00000000<0ogL000@000000?nF003oUP0000000`3o^`001@000000onD00?oU003o
i@00000000<0j_l000D000000;Co002do`00]?l00000000307[o00040000000lo`00??l0000000<0
0?l500D00000003oB@00odT00?m900000003003oT00500000000omP00?oH003of00000000`00gOl0
1000000009Go002Eo`000003001<o`05000000001_l000Ko0006o`0000000`0m0?l01@000000O@3o
07d0o`1m0?l000000080^P3o00D000000?80o`3b0?l0lP3o000000030?l0f`050000003o0:l0o`2_
0?l0[`0000000`3o08T01@000000o`1Z0?l0JP3o06X000000080o`1C00D000000?l0@`3o04<0o`13
000000030?l0>`050000003o03/0o`0k0?l0>`0000000`3o04<01@000000o`1C0?l0D`3o05<00000
0080o`1Z00D000000?l0R@3o08T0o`29000000030?l0[`050000003o0=/0o`3K0?l0f`0000000`3b
0?l01@000000^P3o0;X0o`2j0?l000000080O@3o00D0000003d0o`0m0?l0?@3o000000030006o`05
00000000C?l004co001<o`0000000`00UOl01@0000000=go003Mo`00gOl0000000800?oH00D00000
003oT000oi000?n@00000003003oB@0500000000o`D00?l5003o1@0000000`0lo`001@000000N_l0
07[o001jo`0000000080]?l000D000000>[o003Zo`00j_l0000000030?oU00050000003o^`00ok/0
0?nk000000000`3oUP001@000000ogL00?mg003oM`0000000080oel000D000000?m<003oC000od`0
000000030?m000050000003o>`00oc/00?lk000000000`3o>`001@000000od400?m1003o@@000000
0080odd000D000000?mM003oG@00oed0000000030?mc00050000003oS000oh`00?n<000000000`3o
ZP001@000000ol/00?o;003ob`0000000080onl000D000000>[o003Zo`00j_l0000000030<;o0005
0000002Io`00VOl009Wo000000000`1_o`0010000000AOl004Go0000000301_o000500000000o`d0
0?l=003o3@0000000`00ocD01@0000000?mL003oG000oe`0000000<00?n000<000000?ooo`3oool0
103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003oP000oh000000
0`00oe`01@0000000?le003o=@00ocD0000000<00?l=00D0000001_o000Ko`006ol00000000304Go
00040000001_o`00Kol0000000<0VOl000D000000<;o0032o`00`_l0000000030>[o00050000003o
k`00onl00?o_000000000`3ob`0010000000ojX00?nZ000000030?n<00050000003oL`00og<00?mc
000000000`3oG@001@000000odd00?m=003oC@00000000<0od4000@000000?lk003o>`0000000`3o
>`001@000000od000?m0003o@000000000<0od`000D000000?mO003oG`00oel0000000030?mg0004
0000003oUP00oiH0000000<0ok/000D000000?oU003oi@00onD0000000030>[o00050000002do`00
]?l00;Co000000000`1jo`0010000000??l003co00000003003o1@0500000000odT00?m9003oB@00
00000`00oi001@0000000?oH003of000omP0000000<00=go00@00000002Eo`00UOl000000`00C?l0
1@00000000Ko0006o`001_l0000000<0?@3o00D0000007d0o`1m0?l0O@3o000000020;X0o`050000
003b0?l0lP3o0?80o`0000000`3o0=/01@000000o`2_0?l0[`3o0:l0000000<0o`2900D000000?l0
JP3o06X0o`1Z000000020?l0D`050000003o04<0o`130?l0@`0000000`3o03/01@000000o`0k0?l0
>`3o03/0000000<0o`1300D000000?l0D`3o05<0o`1C000000020?l0JP050000003o08T0o`290?l0
R@0000000`3o0:l01@000000o`3K0?l0f`3o0=/0000000<0lP3o00D000000;X0o`2j0?l0^P3o0000
000207d0o`050000000m0?l0?@3o03d0o`0000000`001_l01@00000004co001<o`00C?l0000000<0
09Go00D00000003Mo`00gOl00=go00000002003of00500000000oi000?n@003oT00000000`00odT0
1@0000000?l5003o1@00o`D0000000<0??l000D0000007[o001jo`00N_l0000000020;Co00050000
003Zo`00j_l00>[o000000000`3oi@001@000000ok/00?nk003o^`00000000<0oiH000D000000?mg
003oM`00ogL0000000020?mO00050000003oC000od`00?m<000000000`3o@0001@000000oc/00?lk
003o>`00000000<0oc/000D000000?m1003o@@00od40000000020?m=00050000003oG@00oed00?mM
000000000`3oL`001@000000oh`00?n<003oS000000000<0ojX000D000000?o;003ob`00ol/00000
00020?o_00050000003Zo`00j_l00>[o000000000`32o`001@000000VOl009Wo002Io`00000000<0
Kol000@0000004Go0015o`0000000`0Ko`001@0000000?l=003o3@00o`d0000000<00?le00D00000
003oG000oe`00?mL00000003003oP0030000003oool0oooo00@0oooo0@0000010?ooo`00803oool0
0`000000oooo0?ooo`050?ooo`0400000000oh000?n0000000<00?mL00D00000003o=@00ocD00?le
00000003003o3@050000000Ko`006ol001_o000000000`15o`0010000000Kol006oo0000000309Wo
000500000032o`00`_l00<;o000000000`3Zo`001@000000onl00?o_003ok`00000000<0ol/000@0
00000?nZ003oZP0000000`3oS0001@000000og<00?mc003oL`00000000<0oed000D000000?m=003o
C@00odd0000000030?m100040000003o>`00oc/0000000<0oc/000D000000?m0003o@000od000000
00030?m<00050000003oG`00oel00?mO000000000`3oM`0010000000oiH00?nF000000030?nk0005
0000003oi@00onD00?oU000000000`3Zo`001@000000]?l00;Co002do`00000000<0N_l000@00000
03co000lo`0000000`00o`D01@0000000?m9003oB@00odT0000000<00?n@00D00000003of000omP0
0?oH00000003003Mo`0400000000UOl009Go000000<004co00D000000006o`001_l000Ko00000003
03d0o`050000001m0?l0O@3o07d0o`0000000P2j0?l01@000000lP3o0?80o`3b0?l0000000<0o`3K
00D000000?l0[`3o0:l0o`2_000000030?l0R@050000003o06X0o`1Z0?l0JP0000000P3o05<01@00
0000o`130?l0@`3o04<0000000<0o`0k00D000000?l0>`3o03/0o`0k000000030?l0@`050000003o
05<0o`1C0?l0D`0000000P3o06X01@000000o`290?l0R@3o08T0000000<0o`2_00D000000?l0f`3o
0=/0o`3K000000030?80o`050000002j0?l0^P3o0;X0o`0000000P1m0?l01@000000?@3o03d0o`0m
0?l0000000<000Ko00D00000001<o`00C?l004co00000003002Eo`0500000000gOl00=go003Mo`00
00000P00omP01@0000000?n@003oT000oi00000000<00?m900D00000003o1@00o`D00?l500000003
03co00050000001jo`00N_l007[o000000000P2do`001@000000j_l00>[o003Zo`00000000<0onD0
00D000000?nk003o^`00ok/0000000030?nF00050000003oM`00ogL00?mg000000000P3oG`001@00
0000od`00?m<003oC000000000<0od0000D000000?lk003o>`00oc/0000000030?lk00050000003o
@@00od400?m1000000000P3oC@001@000000oed00?mM003oG@00000000<0og<000D000000?n<003o
S000oh`0000000030?nZ00050000003ob`00ol/00?o;000000000P3ok`001@000000j_l00>[o003Z
o`00000000<0`_l000D0000009Wo002Io`00VOl00000000306oo000400000015o`00AOl0000000<0
6ol000D00000003o3@00o`d00?l=00000003003o=@0500000000oe`00?mL003oG00000000`00oh00
0`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo0000
084000001P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003o2@00
o`T000000`0Io`001@000000?_l003ko000no`00000000<0H_l000D0000008Oo0027o`00Qol00000
00030:co00040000003@o`00d?l0000000<0l_l000D000000?o/003ok000on`0000000030?o=0005
0000003o/@00ok400?na000000000`3oU`0010000000oh400?n1000000030?m_00050000003oH@00
of400?mQ000000000`3oE`001@000000oe400?mA003oD@00000000<0oe4000@000000?mF003oEP00
00000`3oH0001@000000ofl00?m_003oK`00000000<0oh@000D000000?nN003oWP00oih000000003
0?nn00040000003oh`00on<0000000<0lol000D000000<Co0034o`00a?l000000003097o00050000
001Jo`00F_l005[o000000000`0Po`00100000000?lM003o7@000003003oG@0500000000oil00?nO
003oW`0000000`00on801@0000000=Wo003Io`00fOl0000000<009Go00@00000001Ao`00DOl00000
0`003ol01@000000<03o0300o`0`0?l0000000<0KP3o00D000000:P0o`2X0?l0Z03o000000020=l0
o`050000003o0>h0o`3^0?l0kP0000000`3o0<401@000000o`2J0?l0VP3o09X0000000<0o`1h00D0
00000?l0G@3o05d0o`1M000000020?l0B0050000003o03X0o`0j0?l0>P0000000`3o03<01@000000
o`0c0?l0<`3o03<0000000<0o`0j00D000000?l0B03o04P0o`18000000020?l0G@050000003o07P0
o`1h0?l0N00000000`3o09X01@000000o`310?l0`@3o0<40000000<0o`3^00D000000=l0o`3O0?l0
g`3o000000020:P0o`050000001^0?l0KP3o06h0o`0000000`0`0?l01@00000000oo000?o`003ol0
000000<0057o00D00000002Eo`00UOl009Go00000002003Io`0500000000on800?oR003ohP000000
0`00oil01@0000000?mM003oG@00oed0000000<00?lM00D00000023o000Po`008?l00000000205[o
00050000002Ao`00TOl0097o000000000`34o`001@000000lol00??o003co`00000000<0on<000D0
00000?nn003o_P00okh0000000020?nN00050000003oQ000oh@00?n4000000000`3oK`001@000000
of000?mP003oH000000000<0oeH000D000000?mA003oD@00oe40000000020?mA00050000003oE`00
oeL00?mG000000000`3oH@001@000000ofl00?m_003oK`00000000<0oh4000D000000?nG003oU`00
oiL0000000020?na00050000003oc@00old00?o=000000000`3ok0001@000000l_l00?;o003bo`00
000000<0d?l000@000000:co002/o`0000000`27o`001@000000H_l006;o001Ro`00000000<0?_l0
00D0000001Wo000Io`006Ol000000003003o2@030000003oool0oooo00@0oooo0@0000010?ooo`00
803oool00`000000oooo0?ooo`050?ooo`0400000000o`T00?l9000000<06Ol000D0000003ko000n
o`00?_l00000000306;o000500000027o`00Qol008Oo000000000`2/o`0010000000d?l00=3o0000
00030?;o00050000003ok000on`00?o/000000000`3oc@001@000000ok400?na003o/@00000000<0
oiL000@000000?n1003oP@0000000`3oK`001@000000of400?mQ003oH@00000000<0oeL000D00000
0?mA003oD@00oe40000000030?mA00040000003oEP00oeH0000000<0of0000D000000?m_003oK`00
ofl0000000030?n400050000003oWP00oih00?nN000000000`3o_P0010000000on<00?oS00000003
0??o000500000034o`00a?l00<Co000000000`2Ao`001@000000F_l005[o001Jo`00000000<08?l0
00@00000003o7@00oad000000`00oed01@0000000?nO003oW`00oil0000000<00?oR00D00000003I
o`00fOl00=Wo00000003002Eo`0400000000DOl0057o000000<000oo00D000000300o`0`0?l0<03o
0000000306h0o`050000002X0?l0Z03o0:P0o`0000000P3O0?l01@000000o`3^0?l0kP3o0>h00000
00<0o`3100D000000?l0VP3o09X0o`2J000000030?l0N0050000003o05d0o`1M0?l0G@0000000P3o
04P01@000000o`0j0?l0>P3o03X0000000<0o`0c00D000000?l0<`3o03<0o`0c000000030?l0>P05
0000003o04P0o`180?l0B00000000P3o05d01@000000o`1h0?l0N03o07P0000000<0o`2J00D00000
0?l0`@3o0<40o`31000000030?l0kP050000003O0?l0g`3o0=l0o`0000000P2X0?l01@000000KP3o
06h0o`1^0?l0000000<0<03o00D00000000?o`003ol000oo00000003001Ao`0500000000UOl009Go
002Eo`0000000P00fOl01@0000000?oR003ohP00on80000000<00?nO00D00000003oG@00oed00?mM
00000003003o7@050000000Po`008?l0023o000000000P1Jo`001@000000TOl0097o002Ao`000000
00<0a?l000D000000??o003co`00lol0000000030?oS00050000003o_P00okh00?nn000000000P3o
WP001@000000oh@00?n4003oQ000000000<0ofl000D000000?mP003oH000of00000000030?mF0005
0000003oD@00oe400?mA000000000P3oD@001@000000oeL00?mG003oE`00000000<0of4000D00000
0?m_003oK`00ofl0000000030?n100050000003oU`00oiL00?nG000000000P3o/@001@000000old0
0?o=003oc@00000000<0on`000D000000?;o003bo`00l_l0000000030=3o00040000002/o`00[?l0
000000<0Qol000D0000006;o001Ro`00H_l00000000303ko00050000000Io`006Ol001Wo00000000
0`00o`T00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3o
oooo0000084000001P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000
067o001Qo`0000000`21o`001@000000X?l00:3o002Po`00000000<0_ol000D000000=go003Mo`00
gOl0000000030?[o00040000003oj@00onT0000000<0om0000D000000?nh003o^000okP000000003
0?nS00050000003oTP00oi800?nB000000000`3oP`0010000000ogL00?mg000000030?m`00050000
003oK000of`00?m/000000000`3oK@001@000000og800?mb003oLP00000000<0og/000@000000?n9
003oR@0000000`3oW0001@000000ok<00?nc003o/`00000000<0oll000D000000?o_003ok`00onl0
000000030>_o000400000032o`00`_l0000000<0UOl000D0000006Go001Uo`00IOl000000003037o
000500000000o`D00?l5003o1@0000000`00ocl0100000000?mj003oNP000003003o]`0500000000
ooH00?of003omP0000000`00b_l01@00000008_o002;o`00Rol0000000<004co00@00000000?o`00
3ol000000`0/0?l01@000000IP3o06H0o`1V0?l0000000<0W@3o00D000000=40o`3A0?l0d@3o0000
00020?l0o@050000003o0=40o`3A0?l0d@0000000`3o0:T01@000000o`260?l0QP3o08H0000000<0
o`1Y00D000000?l0D@3o0540o`1A000000020?l0?`050000003o0380o`0b0?l0<P0000000`3o02`0
1@000000o`0/0?l0;03o02`0000000<0o`0b00D000000?l0?`3o03l0o`0o000000020?l0D@050000
003o06T0o`1Y0?l0J@0000000`3o08H01@000000o`2Y0?l0Z@3o0:T0000000<0o`3A00D000000?l0
o@3o0?d0o`3m000000020=40o`050000002M0?l0W@3o09d0o`0000000`1V0?l01@000000;03o02`0
o`0/0?l0000000<000oo00D00000001<o`00C?l004co00000002002;o`0500000000b_l00<[o003:
o`0000000`00ooH01@0000000?ng003o]`00okL0000000<00?mj00D00000003o?`00ocl00?lo0000
0002003o1@050000000ao`00<Ol0037o000000000`1Uo`001@000000UOl009Go002Eo`00000000<0
`_l000D000000>_o003[o`00jol0000000020?o_00050000003oc`00oll00?o?000000000`3o/`00
1@000000oi`00?nL003oW000000000<0ohT000D000000?mk003oN`00og/0000000020?mb00050000
003oK@00ofd00?m]000000000`3oK0001@000000og000?m`003oL000000000<0ogL000D000000?n3
003oP`00oh<0000000020?nB00050000003oX`00oj<00?nS000000000`3o^0001@000000om000?o@
003od000000000<0onT000@000000?[o003jo`0000000`3Mo`001@000000_ol00;oo002oo`000000
00<0X?l000D00000087o0021o`00POl000000003067o00030000003oool0oooo00@0oooo0@000001
0?ooo`00803oool2000000H0oooo00@00000067o001Qo`0000000`21o`001@000000X?l00:3o002P
o`00000000<0_ol000D000000=go003Mo`00gOl0000000030?[o00040000003oj@00onT0000000<0
om0000D000000?nh003o^000okP0000000030?nS00050000003oTP00oi800?nB000000000`3oP`00
10000000ogL00?mg000000030?m`00050000003oK000of`00?m/000000000`3oK@001@000000og80
0?mb003oLP00000000<0og/000@000000?n9003oR@0000000`3oW0001@000000ok<00?nc003o/`00
000000<0oll000D000000?o_003ok`00onl0000000030>_o000400000032o`00`_l0000000<0UOl0
00D0000006Go001Uo`00IOl000000003037o000500000000o`D00?l5003o1@0000000`00ocl01000
00000?mj003oNP000003003o]`0500000000ooH00?of003omP0000000`00b_l01@00000008_o002;
o`00Rol0000000<004co00@00000000?o`003ol000000`0/0?l01@000000IP3o06H0o`1V0?l00000
00<0W@3o00D000000=40o`3A0?l0d@3o000000020?l0o@050000003o0=40o`3A0?l0d@0000000`3o
0:T01@000000o`260?l0QP3o08H0000000<0o`1Y00D000000?l0D@3o0540o`1A000000020?l0?`05
0000003o0380o`0b0?l0<P0000000`3o02`01@000000o`0/0?l0;03o02`0000000<0o`0b00D00000
0?l0?`3o03l0o`0o000000020?l0D@050000003o06T0o`1Y0?l0J@0000000`3o08H01@000000o`2Y
0?l0Z@3o0:T0000000<0o`3A00D000000?l0o@3o0?d0o`3m000000020=40o`050000002M0?l0W@3o
09d0o`0000000`1V0?l01@000000;03o02`0o`0/0?l0000000<000oo00D00000001<o`00C?l004co
00000002002;o`0500000000b_l00<[o003:o`0000000`00ooH01@0000000?ng003o]`00okL00000
00<00?mj00D00000003o?`00ocl00?lo00000002003o1@050000000ao`00<Ol0037o000000000`1U
o`001@000000UOl009Go002Eo`00000000<0`_l000D000000>_o003[o`00jol0000000020?o_0005
0000003oc`00oll00?o?000000000`3o/`001@000000oi`00?nL003oW000000000<0ohT000D00000
0?mk003oN`00og/0000000020?mb00050000003oK@00ofd00?m]000000000`3oK0001@000000og00
0?m`003oL000000000<0ogL000D000000?n3003oP`00oh<0000000020?nB00050000003oX`00oj<0
0?nS000000000`3o^0001@000000om000?o@003od000000000<0onT000@000000?[o003jo`000000
0`3Mo`001@000000_ol00;oo002oo`00000000<0X?l000D00000087o0021o`00POl000000003067o
00030000003oool0oooo00<0oooo0P0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`04
0000001Qo`00HOl0000000<0POl000D000000:3o002Po`00X?l0000000030;oo00050000003Mo`00
gOl00=go000000000`3jo`0010000000onT00?oY000000030?o@00050000003o^000okP00?nh0000
00000`3oX`001@000000oi800?nB003oTP00000000<0oh<000@000000?mg003oM`0000000`3oL000
1@000000of`00?m/003oK000000000<0ofd000D000000?mb003oLP00og80000000030?mk00040000
003oR@00ohT0000000<0oi`000D000000?nc003o/`00ok<0000000030?o?00050000003ok`00onl0
0?o_000000000`3[o`0010000000`_l00<;o0000000309Go00050000001Uo`00IOl006Go00000000
0`0ao`001@0000000?l5003o1@00o`D0000000<00?lo00@00000003oNP00ogX000000`00okL01@00
00000?of003omP00ooH0000000<00<[o00D00000002;o`00Rol008_o00000003001<o`0400000000
3ol000oo000000<0;03o00D0000006H0o`1V0?l0IP3o0000000309d0o`050000003A0?l0d@3o0=40
o`0000000P3o0?d01@000000o`3A0?l0d@3o0=40000000<0o`2Y00D000000?l0QP3o08H0o`260000
00030?l0J@050000003o0540o`1A0?l0D@0000000P3o03l01@000000o`0b0?l0<P3o0380000000<0
o`0/00D000000?l0;03o02`0o`0/000000030?l0<P050000003o03l0o`0o0?l0?`0000000P3o0540
1@000000o`1Y0?l0J@3o06T0000000<0o`2600D000000?l0Z@3o0:T0o`2Y000000030?l0d@050000
003o0?d0o`3m0?l0o@0000000P3A0?l01@000000W@3o09d0o`2M0?l0000000<0IP3o00D0000002`0
o`0/0?l0;03o00000003000?o`0500000000C?l004co001<o`0000000P00Rol01@0000000<[o003:
o`00b_l0000000<00?of00D00000003o]`00okL00?ng00000003003oNP0500000000ocl00?lo003o
?`0000000P00o`D01@000000<Ol0037o000ao`00000000<0IOl000D0000009Go002Eo`00UOl00000
00030<;o00050000003[o`00jol00>_o000000000P3ok`001@000000oll00?o?003oc`00000000<0
ok<000D000000?nL003oW000oi`0000000030?n900050000003oN`00og/00?mk000000000P3oLP00
1@000000ofd00?m]003oK@00000000<0of`000D000000?m`003oL000og00000000030?mg00050000
003oP`00oh<00?n3000000000P3oTP001@000000oj<00?nS003oX`00000000<0okP000D000000?o@
003od000om00000000030?oY00040000003jo`00n_l0000000<0gOl000D000000;oo002oo`00_ol0
000000030:3o000500000021o`00POl0087o000000000`1Qo`000`000000oooo0?ooo`040?ooo`40
00000@3oool00200oooo00<000000?ooo`3oool01@3ooooo0000084000001P3oool100000040oooo
000P0?ooo`030000003oool0oooo00D0oooo00@000000;Go002eo`0000000`3?o`001@000000iol0
0>Oo003Wo`00000000<0ool000D000000?oZ003ojP00onX0000000030?oF00040000003oa000ol@0
000000<0ok@000D000000?nV003oYP00ojH0000000030?nK00050000003oU000oi@00?nD00000000
0`3oS`0010000000ohh00?n>000000030?n@00050000003oUP00oiH00?nF000000000`3oW`001@00
0000ojd00?n]003o[@00000000<0okh000@000000?oD003oe00000000`3ok@001@000000mOl00?Go
003eo`00000000<0e?l000D000000;3o002`o`00/?l00000000308So00040000001Mo`00GOl00000
00<0;ol000D00000003o0@00o`400?l100000003003o=00500000000ofX00?mZ003oJP0000000`00
oj40100000000?oI003of@000003003/o`0500000000/_l00;;o002bo`0000000`00N?l01@000000
03oo000oo`00?ol0000000<000Ko00@000000300o`0`0?l000000`1V0?l01@000000VP3o09X0o`2J
0?l0000000<0b`3o00D000000?T0o`3i0?l0n@3o000000020?l0g0050000003o0;D0o`2e0?l0]@00
00000`3o09801@000000o`1d0?l0M03o07@0000000<0o`1J00D000000?l0A@3o04D0o`1500000002
0?l0=@050000003o02/0o`0[0?l0:`0000000`3o02D01@000000o`0U0?l09@3o02D0000000<0o`0[
00D000000?l0=@3o03D0o`0e000000020?l0A@050000003o05X0o`1J0?l0FP0000000`3o07@01@00
0000o`2B0?l0TP3o0980000000<0o`2e00D000000?l0g03o0=`0o`3L000000020?T0o`050000003;
0?l0b`3o0</0o`0000000`2J0?l01@000000IP3o06H0o`1V0?l0000000<0<03o00D000000006o`00
1_l000Ko00000002000oo`0500000000N?l007So001ho`0000000`00/_l01@0000000>co003/o`00
k?l0000000<00?oI00D00000003oX@00oj400?nQ00000002003oJP0500000000oc@00?ld003o=000
00000`00o`401@000000;ol002oo000_o`00000000<0GOl000D0000008So0028o`00R?l000000002
0;3o00050000003Do`00e?l00=Co000000000`3eo`001@000000ond00?o]003ok@00000000<0om@0
00D000000?nn003o_P00okh0000000020?n]00050000003oW`00oil00?nO000000000`3oUP001@00
0000oi000?n@003oT000000000<0ohh000D000000?n?003oS`00ohl0000000020?nD00050000003o
V`00oi/00?nK000000000`3oYP001@000000ok@00?nd003o]000000000<0ol@000@000000?oF003o
eP0000000`3ojP001@000000ool00?oo003oo`00000000<0iol000D000000<oo003?o`00col00000
00030;Go00030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`05
0?ooo`040000002eo`00]Ol0000000<0col000D000000>Oo003Wo`00iol0000000030?oo00050000
003ojP00onX00?oZ000000000`3oeP0010000000ol@00?o4000000030?nd00050000003oYP00ojH0
0?nV000000000`3oV`001@000000oi@00?nD003oU000000000<0ohl000@000000?n>003oSP000000
0`3oT0001@000000oiH00?nF003oUP00000000<0oil000D000000?n]003o[@00ojd0000000030?nn
00040000003oe000om@0000000<0ond000D000000?Go003eo`00mOl0000000030=Co00050000002`
o`00/?l00;3o000000000`28o`0010000000GOl005go0000000302oo000500000000o`400?l1003o
0@0000000`00oc@01@0000000?mZ003oJP00ofX0000000<00?nQ00@00000003of@00omT000000`00
k?l01@0000000;;o002bo`00/_l0000000<007So00D00000000oo`00?ol003oo000000030006o`04
0000000`0?l0<03o000000<0IP3o00D0000009X0o`2J0?l0VP3o000000030</0o`050000003i0?l0
n@3o0?T0o`0000000P3o0=`01@000000o`2e0?l0]@3o0;D0000000<0o`2B00D000000?l0M03o07@0
o`1d000000030?l0FP050000003o04D0o`150?l0A@0000000P3o03D01@000000o`0[0?l0:`3o02/0
000000<0o`0U00D000000?l09@3o02D0o`0U000000030?l0:`050000003o03D0o`0e0?l0=@000000
0P3o04D01@000000o`1J0?l0FP3o05X0000000<0o`1d00D000000?l0TP3o0980o`2B000000030?l0
]@050000003o0=`0o`3L0?l0g00000000P3i0?l01@000000b`3o0</0o`3;0?l0000000<0VP3o00D0
000006H0o`1V0?l0IP3o000000030300o`05000000001_l000Ko0006o`0000000P00?ol01@000000
07So001ho`00N?l0000000<00;;o00D00000003/o`00k?l00>co00000003003of@0500000000oj40
0?nQ003oX@0000000P00ofX01@0000000?ld003o=000oc@0000000<00?l100D0000002oo000_o`00
;ol00000000305go000500000028o`00R?l008So000000000P2`o`001@000000e?l00=Co003Do`00
000000<0mOl000D000000?o]003ok@00ond0000000030?oD00050000003o_P00okh00?nn00000000
0P3o[@001@000000oil00?nO003oW`00000000<0oiH000D000000?n@003oT000oi00000000030?n>
00050000003oS`00ohl00?n?000000000P3oU0001@000000oi/00?nK003oV`00000000<0ojH000D0
00000?nd003o]000ok@0000000030?o400040000003oeP00omH0000000<0onX000D000000?oo003o
o`00ool0000000030>Oo00050000003?o`00col00<oo000000000`2eo`000`000000oooo0?ooo`04
0?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool010000000]Ol00;Go00000003
0<oo00050000003Wo`00iol00>Oo000000000`3oo`001@000000onX00?oZ003ojP00000000<0omH0
00@000000?o4003oa00000000`3o]0001@000000ojH00?nV003oYP00000000<0oi/000D000000?nD
003oU000oi@0000000030?n?00040000003oSP00ohh0000000<0oi0000D000000?nF003oUP00oiH0
000000030?nO00050000003o[@00ojd00?n]000000000`3o_P0010000000om@00?oD000000030?o]
00050000003eo`00mOl00?Go000000000`3Do`001@000000/?l00;3o002`o`00000000<0R?l000@0
000005go001Mo`0000000`0_o`001@0000000?l1003o0@00o`40000000<00?ld00D00000003oJP00
ofX00?mZ00000003003oX@0400000000omT00?oI000000<00>co00D00000002bo`00/_l00;;o0000
0003001ho`0500000000?ol003oo000oo`0000000`001_l010000000<03o0300o`00000306H0o`05
0000002J0?l0VP3o09X0o`0000000`3;0?l01@000000n@3o0?T0o`3i0?l000000080o`3L00D00000
0?l0]@3o0;D0o`2e000000030?l0TP050000003o07@0o`1d0?l0M00000000`3o05X01@000000o`15
0?l0A@3o04D000000080o`0e00D000000?l0:`3o02/0o`0[000000030?l09@050000003o02D0o`0U
0?l09@0000000`3o02/01@000000o`0e0?l0=@3o03D000000080o`1500D000000?l0FP3o05X0o`1J
000000030?l0M0050000003o0980o`2B0?l0TP0000000`3o0;D01@000000o`3L0?l0g03o0=`00000
0080n@3o00D000000</0o`3;0?l0b`3o0000000309X0o`050000001V0?l0IP3o06H0o`0000000`0`
0?l01@00000000Ko0006o`001_l00000008003oo00D00000001ho`00N?l007So00000003002bo`05
00000000k?l00>co003/o`0000000`00omT01@0000000?nQ003oX@00oj40000000800?mZ00D00000
003o=000oc@00?ld00000003003o0@050000000_o`00;ol002oo000000000`1Mo`001@000000R?l0
08So0028o`0000000080/?l000D000000=Co003Do`00e?l0000000030?Go00050000003ok@00ond0
0?o]000000000`3oe0001@000000okh00?nn003o_P0000000080ojd000D000000?nO003oW`00oil0
000000030?nF00050000003oT000oi000?n@000000000`3oSP001@000000ohl00?n?003oS`000000
0080oi@000D000000?nK003oV`00oi/0000000030?nV00050000003o]000ok@00?nd000000000`3o
a00010000000omH00?oF000000030?oZ00050000003oo`00ool00?oo000000000`3Wo`001@000000
col00<oo003?o`00000000<0]Ol000<000000?ooo`3oool0103oool100000040oooo000P0?ooo`03
0000003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?ooo`00803oool00`000000oooo
0?ooo`050?ooo`040000003Yo`00jOl0000000<0nol000D000000?oc003ol`00oo<0000000030?oT
00050000003oe`00omL00?oG000000000`3oc00010000000ol<00?o3000000030?nl00050000003o
^000okP00?nh000000000`3o]P001@000000okP00?nh003o^000000000<0ok`000@000000?o3003o
``0000000`3ocP001@000000om/00?oK003of`00000000<0on`000D000000?oo003oo`00ool00000
00030>Oo00040000003=o`00cOl0000000<0[ol000D0000008ko002>o`00S_l00000000306[o0005
00000014o`00A?l004Co000000000`0Ko`00100000000?l@003o40000003003o?@0500000000ofd0
0?m]003oK@0000000`00oih01@0000000?oA003od@00om40000000<00?_o00@000000036o`00a_l0
00000`00T_l01@00000005go001Mo`00GOl0000000<002Wo00D0000000X0o`0:0?l02P3o00000003
03d0o`040000001^0?l0KP3o000000<0W@3o00D000000</0o`3;0?l0b`3o000000030?H0o`050000
003o0>40o`3Q0?l0h@0000000P3o0;`01@000000o`2K0?l0V`3o09/0000000<0o`1m00D000000?l0
H`3o06<0o`1S000000030?l0C0050000003o03X0o`0j0?l0>P0000000P3o02d01@000000o`0T0?l0
903o02@0000000<0o`0O00D000000?l07`3o01l0o`0O000000030?l090050000003o02d0o`0]0?l0
;@0000000P3o03X01@000000o`1<0?l0C03o04`0000000<0o`1S00D000000?l0O@3o07d0o`1m0000
00030?l0V`050000003o0;`0o`2l0?l0_00000000P3o0>401@000000mP3o0?H0o`3f0?l0000000<0
b`3o00D0000009d0o`2M0?l0W@3o0000000306h0o`050000000m0?l0?@3o03d0o`0000000P0:0?l0
1@00000002Wo000Yo`00:Ol0000000<005go00D00000002Bo`00T_l009;o000000030036o`050000
0000nol00?_o003ko`0000000P00om401@0000000?nN003oWP00oih0000000<00?m]00D00000003o
?@00ocd00?lm00000003003o40050000000Ko`006ol001_o000000000P14o`001@000000J_l006[o
001Zo`00000000<0S_l000D000000:oo002_o`00[ol0000000030<go00050000003Wo`00iol00>Oo
000000000P3oo`001@000000on`00?o/003ok000000000<0om/000D000000?o>003ocP00olh00000
00030?o300050000003o_000ok`00?nl000000000P3o^0001@000000okH00?nf003o]P00000000<0
okP000D000000?nl003o_000ok`0000000030?o300040000003oc000ol`0000000<0omL000D00000
0?oT003oi000on@0000000030?oc00050000003ko`00nol00?_o000000000`3Yo`000`000000oooo
0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool010000000jOl00>Wo
000000030?_o00050000003ol`00oo<00?oc000000000`3oi0001@000000omL00?oG003oe`000000
00<0ol`000@000000?o3003o``0000000`3o_0001@000000okP00?nh003o^000000000<0okH000D0
00000?nh003o^000okP0000000030?nl00040000003o``00ol<0000000<0olh000D000000?oK003o
f`00om/0000000030?o/00050000003oo`00ool00?oo000000000`3Wo`0010000000cOl00<go0000
00030:oo00050000002>o`00S_l008ko000000000`1Zo`001@000000A?l004Co0014o`00000000<0
6ol000@00000003o4000oa0000000`00ocd01@0000000?m]003oK@00ofd0000000<00?nN00D00000
003od@00om400?oA00000003003ko`0400000000a_l00<Ko000000<009;o00D00000001Mo`00GOl0
05go00000003000Yo`050000000:0?l02P3o00X0o`0000000`0m0?l010000000KP3o06h0o`000003
09d0o`050000003;0?l0b`3o0</0o`0000000`3f0?l01@000000o`3Q0?l0h@3o0>4000000080o`2l
00D000000?l0V`3o09/0o`2K000000030?l0O@050000003o06<0o`1S0?l0H`0000000`3o04`01@00
0000o`0j0?l0>P3o03X000000080o`0]00D000000?l0903o02@0o`0T000000030?l07`050000003o
01l0o`0O0?l07`0000000`3o02@01@000000o`0]0?l0;@3o02d000000080o`0j00D000000?l0C03o
04`0o`1<000000030?l0H`050000003o07d0o`1m0?l0O@0000000`3o09/01@000000o`2l0?l0_03o
0;`000000080o`3Q00D000000?H0o`3f0?l0mP3o000000030</0o`050000002M0?l0W@3o09d0o`00
00000`1^0?l01@000000?@3o03d0o`0m0?l0000000802P3o00D00000000Yo`00:Ol002Wo00000003
001Mo`0500000000T_l009;o002Bo`0000000`00a_l01@0000000?_o003ko`00nol0000000800?oA
00D00000003oWP00oih00?nN00000003003oK@0500000000ocd00?lm003o?@0000000`00oa001@00
00006ol001_o000Ko`0000000080A?l000D0000006[o001Zo`00J_l00000000308ko00050000002_
o`00[ol00:oo000000000`3=o`001@000000iol00>Oo003Wo`0000000080ool000D000000?o/003o
k000on`0000000030?oK00050000003ocP00olh00?o>000000000`3o``001@000000ok`00?nl003o
_00000000080okP000D000000?nf003o]P00okH0000000030?nh00050000003o_000ok`00?nl0000
00000`3o``0010000000ol`00?o<000000030?oG00050000003oi000on@00?oT000000000`3ol`00
1@000000nol00?_o003ko`00000000<0jOl000<000000?ooo`3oool0103oool100000040oooo000P
0?ooo`030000003oool0oooo00D0oooo00@000000>Wo003Yo`0000000`3ko`001@000000oo<00?oc
003ol`00000000<0on@000D000000?oG003oe`00omL0000000030?o<00040000003o``00ol<00000
00<0ok`000D000000?nh003o^000okP0000000030?nf00050000003o^000okP00?nh000000000`3o
_00010000000ol<00?o3000000030?o>00050000003of`00om/00?oK000000000`3ok0001@000000
ool00?oo003oo`00000000<0iol000@000000<go003=o`0000000`2_o`001@000000S_l008ko002>
o`00000000<0J_l000D0000004Co0014o`00A?l00000000301_o000400000000oa000?l@000000<0
0?lm00D00000003oK@00ofd00?m]00000003003oWP0500000000om400?oA003od@0000000`00nol0
100000000<Ko0036o`000003002Bo`0500000000GOl005go001Mo`0000000`00:Ol01@0000002P3o
00X0o`0:0?l0000000<0?@3o00@0000006h0o`1^0?l000000`2M0?l01@000000b`3o0</0o`3;0?l0
000000<0mP3o00D000000?l0h@3o0>40o`3Q000000020?l0_0050000003o09/0o`2K0?l0V`000000
0`3o07d01@000000o`1S0?l0H`3o06<0000000<0o`1<00D000000?l0>P3o03X0o`0j000000020?l0
;@050000003o02@0o`0T0?l0900000000`3o01l01@000000o`0O0?l07`3o01l0000000<0o`0T00D0
00000?l0;@3o02d0o`0]000000020?l0>P050000003o04`0o`1<0?l0C00000000`3o06<01@000000
o`1m0?l0O@3o07d0000000<0o`2K00D000000?l0_03o0;`0o`2l000000020?l0h@050000003f0?l0
mP3o0?H0o`0000000`3;0?l01@000000W@3o09d0o`2M0?l0000000<0KP3o00D0000003d0o`0m0?l0
?@3o0000000200X0o`0500000000:Ol002Wo000Yo`0000000`00GOl01@00000009;o002Bo`00T_l0
000000<00<Ko00D00000003ko`00nol00?_o00000002003od@0500000000oih00?nN003oWP000000
0`00ofd01@0000000?lm003o?@00ocd0000000<00?l@00D0000001_o000Ko`006ol00000000204Co
00050000001Zo`00J_l006[o000000000`2>o`001@000000[ol00:oo002_o`00000000<0cOl000D0
00000>Oo003Wo`00iol0000000020?oo00050000003ok000on`00?o/000000000`3of`001@000000
olh00?o>003ocP00000000<0ol<000D000000?nl003o_000ok`0000000020?nh00050000003o]P00
okH00?nf000000000`3o^0001@000000ok`00?nl003o_000000000<0ol<000@000000?o<003oc000
00000`3oe`001@000000on@00?oT003oi000000000<0oo<000D000000?_o003ko`00nol000000003
0>Wo00030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oo
ool00000P@0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool010000000
mol00?Oo000000030?on00050000003omP00ooH00?of000000000`3ok`001@000000onX00?oZ003o
jP00000000<0onL000@000000?oW003oi`0000000`3oj0001@000000on`00?o/003ok000000000<0
oo<000D000000?ok003on`00oo/0000000030?So00040000003Zo`00j_l0000000<0fOl000D00000
0<Ko0036o`00a_l0000000030;3o00050000002Go`00Uol009Oo000000000`1lo`0010000000G_l0
05ko0000000303ko00050000000Ko`006ol001_o000000000`00o`P01@0000000?l_003o;`00obl0
000000<00?mH00@00000003oPP00oh8000000`00ojh01@0000000?oK003of`00om/0000000<00?Ko
00D000000038o`00b?l00<So00000003002Ho`0400000000JOl006Wo000000<003Wo00D00000000:
o`002_l000[o0000000302<0o`050000001A0?l0D@3o0540o`0000000`1m0?l010000000Z03o0:P0
o`0000030=40o`050000003i0?l0n@3o0?T0o`0000000`3o0>401@000000o`2o0?l0_`3o0;l00000
0080o`2O00D000000?l0P`3o08<0o`23000000030?l0J@050000003o05<0o`1C0?l0D`0000000`3o
04001@000000o`0a0?l0<@3o034000000080o`0U00D000000?l07@3o01d0o`0M000000030?l06P05
0000003o01X0o`0J0?l06P0000000`3o01d01@000000o`0U0?l09@3o02D000000080o`0a00D00000
0?l0@03o0400o`10000000030?l0D`050000003o06T0o`1Y0?l0J@0000000`3o08<01@000000o`2O
0?l0W`3o09l000000080o`2o00D000000?l0h@3o0>40o`3Q000000030?T0o`050000003A0?l0d@3o
0=40o`0000000`2X0?l01@000000O@3o07d0o`1m0?l000000080D@3o00D0000002<0o`0S0?l08`3o
00000003000:o`0500000000>Ol003Wo000io`0000000`00JOl01@00000009So002Ho`00V?l00000
00800<So00D00000003fo`00m_l00?Ko00000003003of`0500000000ojh00?n^003o[P0000000`00
oh801@0000000?mH003oF000oeP0000000800?l_00D00000003o2000o`P00?l80000000301_o0005
0000000no`00?_l003ko000000000`1No`001@000000O?l007co001lo`0000000080Uol000D00000
0;3o002`o`00/?l0000000030<Ko00050000003Io`00fOl00=Wo000000000`3Zo`001@000000n?l0
0?So003ho`0000000080oo/000D000000?oc003ol`00oo<0000000030?o/00050000003oj000onP0
0?oX000000000`3oi`0010000000onL00?oW000000030?oZ00050000003ok`00onl00?o_00000000
0`3omP001@000000ooh00?on003ooP00000000<0mol000<000000?ooo`3oool0103oool100000040
oooo000P0?ooo`030000003oool0oooo00D0oooo00@000000?Oo003go`0000000`3ooP001@000000
ooH00?of003omP00000000<0onl000D000000?oZ003ojP00onX0000000030?oW00040000003oi`00
onL0000000<0onP000D000000?o/003ok000on`0000000030?oc00050000003on`00oo/00?ok0000
00000`3ho`0010000000j_l00>[o000000030=Wo000500000036o`00a_l00<Ko000000000`2`o`00
1@000000Uol009Oo002Go`00000000<0O?l000@0000005ko001No`0000000`0no`001@0000006ol0
01_o000Ko`00000000<00?l800D00000003o;`00obl00?l_00000003003oF00400000000oh800?n2
000000<00?n^00D00000003of`00om/00?oK00000003003fo`0500000000b?l00<So0038o`000000
0`00V?l01000000006Wo001Yo`000003000io`05000000002_l000[o000:o`0000000`0S0?l01@00
0000D@3o0540o`1A0?l0000000<0O@3o00@000000:P0o`2X0?l000000`3A0?l01@000000n@3o0?T0
o`3i0?l0000000<0o`3Q00D000000?l0_`3o0;l0o`2o000000020?l0W`050000003o08<0o`230?l0
P`0000000`3o06T01@000000o`1C0?l0D`3o05<0000000<0o`1000D000000?l0<@3o0340o`0a0000
00020?l09@050000003o01d0o`0M0?l07@0000000`3o01X01@000000o`0J0?l06P3o01X0000000<0
o`0M00D000000?l09@3o02D0o`0U000000020?l0<@050000003o0400o`100?l0@00000000`3o05<0
1@000000o`1Y0?l0J@3o06T0000000<0o`2300D000000?l0W`3o09l0o`2O000000020?l0_`050000
003o0>40o`3Q0?l0h@0000000`3i0?l01@000000d@3o0=40o`3A0?l0000000<0Z03o00D0000007d0
o`1m0?l0O@3o000000020540o`050000000S0?l08`3o02<0o`0000000`002_l01@00000003Wo000i
o`00>Ol0000000<006Wo00D00000002Ho`00V?l009So000000020038o`0500000000m_l00?Ko003f
o`0000000`00om/01@0000000?n^003o[P00ojh0000000<00?n200D00000003oF000oeP00?mH0000
0002003o;`0500000000o`P00?l8003o200000000`0Ko`001@000000?_l003ko000no`00000000<0
G_l000D0000007co001lo`00O?l00000000209Oo00050000002`o`00/?l00;3o000000000`36o`00
1@000000fOl00=Wo003Io`00000000<0j_l000D000000?So003ho`00n?l0000000020?ok00050000
003ol`00oo<00?oc000000000`3ok0001@000000onP00?oX003oj000000000<0onL000@000000?oW
003oi`0000000`3ojP001@000000onl00?o_003ok`00000000<0ooH000D000000?on003ooP00ooh0
000000030?Oo00030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?oo
o`050?ooo`040000003go`00mol0000000<0ooh000D000000?of003omP00ooH0000000030?o_0005
0000003ojP00onX00?oZ000000000`3oi`0010000000onL00?oW000000030?oX00050000003ok000
on`00?o/000000000`3ol`001@000000oo/00?ok003on`00000000<0n?l000@000000>[o003Zo`00
00000`3Io`001@000000a_l00<Ko0036o`00000000<0/?l000D0000009Oo002Go`00Uol000000003
07co00040000001No`00G_l0000000<0?_l000D0000001_o000Ko`006ol000000003003o20050000
0000obl00?l_003o;`0000000`00oeP0100000000?n2003oPP000003003o[P0500000000om/00?oK
003of`0000000`00m_l01@0000000<So0038o`00b?l0000000<009So00@00000001Yo`00JOl00000
0`00>Ol01@00000000[o000:o`002_l0000000<08`3o00D000000540o`1A0?l0D@3o0000000307d0
o`040000002X0?l0Z03o000000<0d@3o00D000000?T0o`3i0?l0n@3o000000030?l0h@050000003o
0;l0o`2o0?l0_`0000000P3o09l01@000000o`230?l0P`3o08<0000000<0o`1Y00D000000?l0D`3o
05<0o`1C000000030?l0@0050000003o0340o`0a0?l0<@0000000P3o02D01@000000o`0M0?l07@3o
01d0000000<0o`0J00D000000?l06P3o01X0o`0J000000030?l07@050000003o02D0o`0U0?l09@00
00000P3o03401@000000o`100?l0@03o0400000000<0o`1C00D000000?l0J@3o06T0o`1Y00000003
0?l0P`050000003o09l0o`2O0?l0W`0000000P3o0;l01@000000o`3Q0?l0h@3o0>40000000<0n@3o
00D000000=40o`3A0?l0d@3o000000030:P0o`050000001m0?l0O@3o07d0o`0000000P1A0?l01@00
00008`3o02<0o`0S0?l0000000<000[o00D00000000io`00>Ol003Wo00000003001Yo`0500000000
V?l009So002Ho`0000000P00b?l01@0000000?Ko003fo`00m_l0000000<00?oK00D00000003o[P00
ojh00?n^00000003003oPP0500000000oeP00?mH003oF00000000P00obl01@0000000?l8003o2000
o`P0000000<06ol000D0000003ko000no`00?_l00000000305ko00050000001lo`00O?l007co0000
00000P2Go`001@000000/?l00;3o002`o`00000000<0a_l000D000000=Wo003Io`00fOl000000003
0>[o00050000003ho`00n?l00?So000000000P3on`001@000000oo<00?oc003ol`00000000<0on`0
00D000000?oX003oj000onP0000000030?oW00040000003oi`00onL0000000<0onX000D000000?o_
003ok`00onl0000000030?of00050000003ooP00ooh00?on000000000`3go`000`000000oooo0?oo
o`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo0000084000001P3oool1
00000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@000000=co003Lo`0000000`3Oo`00
1@000000h?l00>3o003Po`00000000<0gol000D000000=go003Mo`00gOl0000000030=So00040000
003Bo`00d_l0000000<0b_l000D000000;oo002oo`00_ol0000000030;?o00050000002To`00Y?l0
0:Co000000000`2Co`0010000000POl0087o0000000306co00050000001Do`00E?l005Co00000000
0`0ko`001@0000008?l0023o000Po`00000000<00ol000@00000003o6`00oa/000000`00oc`01@00
00000?mN003oGP00oeh0000000<00?n200D00000003oZ000ojP00?nX00000003003ocP0400000000
ooH00?of000000<00>3o00D00000002go`00]ol00;Oo00000003002=o`0500000000Hol006?o001S
o`0000000`00>Ol01000000000oo000?o`00000301X0o`05000000130?l0@`3o04<0o`0000000`1/
0?l01@000000T`3o09<0o`2C0?l0000000<0^P3o00@000000=l0o`3O0?l000000`3o0?d01@000000
o`3L0?l0g03o0=`0000000<0o`2l00D000000?l0W`3o09l0o`2O000000020?l0Q@050000003o06`0
o`1/0?l0K00000000`3o05L01@000000o`140?l0A03o04@0000000<0o`0d00D000000?l0:03o02P0
o`0X000000020?l07P050000003o01P0o`0H0?l0600000000`3o01@01@000000o`0D0?l0503o01@0
000000<0o`0H00D000000?l07P3o01h0o`0N000000020?l0:0050000003o03@0o`0d0?l0=0000000
0`3o04@01@000000o`1G0?l0E`3o05L0000000<0o`1/00D000000?l0Q@3o08D0o`25000000020?l0
W`050000003o0;`0o`2l0?l0_00000000`3o0=`01@000000o`3m0?l0o@3o0?d0000000<0g`3o00D0
00000;X0o`2j0?l0^P3o0000000209<0o`050000001/0?l0K03o06`0o`0000000`130?l01@000000
6P3o01X0o`0J0?l0000000<000oo00D00000000io`00>Ol003Wo00000002001So`0500000000SOl0
08go002=o`0000000`00]ol01@0000000>3o003Po`00h?l0000000<00?of00D00000003ocP00olh0
0?o>00000002003oZ00500000000oh800?n2003oPP0000000`00oeh01@0000000?ll003o?000oc`0
000000<00?lK00D0000000?o0003o`000ol000000002023o00050000000ko`00>ol003_o00000000
0`1Do`001@000000K?l006co001/o`00000000<0POl000D0000009?o002Co`00Tol0000000020:Co
00050000002co`00/ol00;?o000000000`2oo`001@000000b_l00<[o003:o`00000000<0d_l000@0
00000=So003Ho`0000000`3Mo`001@000000gol00=oo003Oo`00000000<0h?l000D000000=oo003O
o`00gol0000000030=co00030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000
oooo0?ooo`050?ooo`040000003Lo`00g?l0000000<0gol000D000000>3o003Po`00h?l000000003
0=oo00050000003Mo`00gOl00=go000000000`3Ho`0010000000d_l00=;o000000030<[o00050000
002oo`00_ol00;oo000000000`2co`001@000000Y?l00:Co002To`00000000<0Tol000@00000087o
0021o`0000000`1/o`001@000000E?l005Co001Do`00000000<0>ol000D00000023o000Po`008?l0
0000000300?o000400000000oa/00?lK000000<00?ll00D00000003oGP00oeh00?mN00000003003o
PP0500000000ojP00?nX003oZ00000000`00olh0100000000?of003omP000003003Po`0500000000
]ol00;Oo002go`0000000`00SOl01@00000006?o001So`00Hol0000000<003Wo00@00000000?o`00
3ol000000`0J0?l01@000000@`3o04<0o`130?l0000000<0K03o00D0000009<0o`2C0?l0T`3o0000
00030;X0o`040000003O0?l0g`3o000000<0o`3m00D000000?l0g03o0=`0o`3L000000030?l0_005
0000003o09l0o`2O0?l0W`0000000P3o08D01@000000o`1/0?l0K03o06`0000000<0o`1G00D00000
0?l0A03o04@0o`14000000030?l0=0050000003o02P0o`0X0?l0:00000000P3o01h01@000000o`0H
0?l0603o01P0000000<0o`0D00D000000?l0503o01@0o`0D000000030?l060050000003o01h0o`0N
0?l07P0000000P3o02P01@000000o`0d0?l0=03o03@0000000<0o`1400D000000?l0E`3o05L0o`1G
000000030?l0K0050000003o08D0o`250?l0Q@0000000P3o09l01@000000o`2l0?l0_03o0;`00000
00<0o`3L00D000000?l0o@3o0?d0o`3m000000030=l0o`050000002j0?l0^P3o0;X0o`0000000P2C
0?l01@000000K03o06`0o`1/0?l0000000<0@`3o00D0000001X0o`0J0?l06P3o00000003000?o`05
00000000>Ol003Wo000io`0000000P00Hol01@00000008go002=o`00SOl0000000<00;Oo00D00000
003Po`00h?l00>3o00000003003omP0500000000olh00?o>003ocP0000000P00ojP01@0000000?n2
003oPP00oh80000000<00?mN00D00000003o?000oc`00?ll00000003003o6`0500000003o`000ol0
00?o000000000P0Po`001@000000>ol003_o000ko`00000000<0E?l000D0000006co001/o`00K?l0
00000003087o00050000002Co`00Tol009?o000000000P2To`001@000000/ol00;?o002co`000000
00<0_ol000D000000<[o003:o`00b_l0000000030=;o00040000003Ho`00f?l0000000<0gOl000D0
00000=oo003Oo`00gol0000000030>3o00050000003Oo`00gol00=oo000000000`3Lo`000`000000
oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool010000000g?l0
0=co000000030=oo00050000003Po`00h?l00>3o000000000`3Oo`001@000000gOl00=go003Mo`00
000000<0f?l000@000000=;o003Bo`0000000`3:o`001@000000_ol00;oo002oo`00000000<0/ol0
00D000000:Co002To`00Y?l00000000309?o000400000021o`00POl0000000<0K?l000D0000005Co
001Do`00E?l00000000303_o00050000000Po`008?l0023o000000000`03o`00100000000?lK003o
6`000003003o?00500000000oeh00?mN003oGP0000000`00oh801@0000000?nX003oZ000ojP00000
00<00?o>00@00000003omP00ooH000000`00h?l01@0000000;Oo002go`00]ol0000000<008go00D0
0000001So`00Hol006?o00000003000io`04000000003ol000oo000000<06P3o00D0000004<0o`13
0?l0@`3o0000000306`0o`050000002C0?l0T`3o09<0o`0000000`2j0?l010000000g`3o0=l0o`00
00030?l0o@050000003o0=`0o`3L0?l0g00000000`3o0;`01@000000o`2O0?l0W`3o09l000000080
o`2500D000000?l0K03o06`0o`1/000000030?l0E`050000003o04@0o`140?l0A00000000`3o03@0
1@000000o`0X0?l0:03o02P000000080o`0N00D000000?l0603o01P0o`0H000000030?l050050000
003o01@0o`0D0?l0500000000`3o01P01@000000o`0N0?l07P3o01h000000080o`0X00D000000?l0
=03o03@0o`0d000000030?l0A0050000003o05L0o`1G0?l0E`0000000`3o06`01@000000o`250?l0
Q@3o08D000000080o`2O00D000000?l0_03o0;`0o`2l000000030?l0g0050000003o0?d0o`3m0?l0
o@0000000`3O0?l01@000000^P3o0;X0o`2j0?l000000080T`3o00D0000006`0o`1/0?l0K03o0000
000304<0o`050000000J0?l06P3o01X0o`0000000`003ol01@00000003Wo000io`00>Ol000000080
06?o00D00000002=o`00SOl008go00000003002go`0500000000h?l00>3o003Po`0000000`00ooH0
1@0000000?o>003ocP00olh0000000800?nX00D00000003oPP00oh800?n200000003003oGP050000
0000oc`00?ll003o?00000000`00oa/01@0000000ol000?o0003o`00000000808?l000D0000003_o
000ko`00>ol00000000305Co00050000001/o`00K?l006co000000000`21o`001@000000Tol009?o
002Co`0000000080Y?l000D000000;?o002co`00/ol0000000030;oo00050000003:o`00b_l00<[o
000000000`3Bo`0010000000f?l00=So000000030=go00050000003Oo`00gol00=oo000000000`3P
o`001@000000gol00=oo003Oo`00000000<0g?l000<000000?ooo`3oool0103oool100000040oooo
000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?ooo`00803oool0
0`000000oooo0?ooo`050?ooo`040000002Lo`00W?l0000000<0V?l000D0000009?o002Co`00Tol0
0000000308co000500000024o`00Q?l008Co000000000`1jo`0010000000K_l006ko00000003067o
00050000001Bo`00D_l005;o000000000`12o`001@000000;ol002oo000_o`00000000<06ol000@0
000000Ko0006o`0000000`00oa401@0000000?lY003o:@00obT0000000<00?m400D00000003oH000
of000?mP00000003003oO@0400000000oi/00?nK000000<00?nk00D00000003og000om`00?oL0000
0003003ooP0500000000g_l00=ko003No`0000000`00^ol01000000009Oo002Go`000003001bo`05
00000000COl004go001=o`0000000`00:?l01@00000000?o0003o`000ol0000000<08@3o00@00000
04D0o`150?l000000`1Z0?l01@000000S@3o08d0o`2=0?l0000000<0/03o00D000000=40o`3A0?l0
d@3o000000030?80o`040000003o0>h0o`3^000000<0o`3A00D000000?l0]@3o0;D0o`2e00000003
0?l0V`050000003o08<0o`230?l0P`0000000P3o06`01@000000o`1H0?l0F03o05P0000000<0o`16
00D000000?l0=`3o03L0o`0g000000030?l0:P050000003o0200o`0P0?l0800000000P3o01P01@00
0000o`0B0?l04P3o0180000000<0o`0@00D000000?l0403o0100o`0@000000030?l04P050000003o
01P0o`0H0?l0600000000P3o02001@000000o`0Z0?l0:P3o02X0000000<0o`0g00D000000?l0AP3o
04H0o`16000000030?l0F0050000003o06`0o`1/0?l0K00000000P3o08<01@000000o`2K0?l0V`3o
09/0000000<0o`2e00D000000?l0d@3o0=40o`3A000000030?l0kP050000003b0?l0lP3o0?80o`00
00000P3A0?l01@000000/03o0;00o`2`0?l0000000<0S@3o00D0000006X0o`1Z0?l0JP3o00000003
04D0o`050000000Q0?l08@3o0240o`0000000P000ol01@00000002So000Xo`00:?l0000000<004go
00D00000001bo`00L_l007;o00000003002Go`0500000000^ol00;_o002ko`0000000P00g_l01@00
00000?on003ooP00ooh0000000<00?oL00D00000003o^`00ok/00?nk00000003003oV`0500000000
ogd00?mm003oO@0000000P00of001@0000000?m4003oA000od@0000000<00?lY00D00000003o4@00
oa400?lA0000000300Ko00050000000Ko`006ol001_o000000000P0_o`001@000000@_l004;o0012
o`00000000<0D_l000D00000067o001Qo`00HOl00000000306ko00040000001jo`00N_l0000000<0
Q?l000D0000008co002<o`00S?l00000000309?o00050000002Ho`00V?l009So000000000`2Lo`00
0`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool01000
0000W?l009co0000000309So00050000002Co`00Tol009?o000000000`2<o`001@000000Q?l008Co
0024o`00000000<0N_l000@0000006ko001^o`0000000`1Qo`001@000000D_l005;o001Bo`000000
00<0@_l000D0000002oo000_o`00;ol00000000301_o000400000006o`001_l0000000<00?lA00D0
0000003o:@00obT00?lY00000003003oA00500000000of000?mP003oH00000000`00ogd010000000
0?nK003oV`000003003o^`0500000000om`00?oL003og00000000`00ooh01@0000000=ko003No`00
g_l0000000<00;_o00@00000002Go`00Uol000000`00L_l01@00000004go001=o`00COl0000000<0
02So00D000000003o`000ol000?o000000030240o`04000000150?l0A@3o000000<0JP3o00D00000
08d0o`2=0?l0S@3o000000030;00o`050000003A0?l0d@3o0=40o`0000000`3b0?l010000000o`3^
0?l0kP0000030?l0d@050000003o0;D0o`2e0?l0]@0000000`3o09/01@000000o`230?l0P`3o08<0
00000080o`1/00D000000?l0F03o05P0o`1H000000030?l0AP050000003o03L0o`0g0?l0=`000000
0`3o02X01@000000o`0P0?l0803o020000000080o`0H00D000000?l04P3o0180o`0B000000030?l0
40050000003o0100o`0@0?l0400000000`3o01801@000000o`0H0?l0603o01P000000080o`0P00D0
00000?l0:P3o02X0o`0Z000000030?l0=`050000003o04H0o`160?l0AP0000000`3o05P01@000000
o`1/0?l0K03o06`000000080o`2300D000000?l0V`3o09/0o`2K000000030?l0]@050000003o0=40
o`3A0?l0d@0000000`3o0>h01@000000lP3o0?80o`3b0?l000000080d@3o00D000000;00o`2`0?l0
/03o0000000308d0o`050000001Z0?l0JP3o06X0o`0000000`150?l01@0000008@3o0240o`0Q0?l0
0000008000?o00D00000000Xo`00:?l002So00000003001=o`0500000000L_l007;o001bo`000000
0`00Uol01@0000000;_o002ko`00^ol0000000800=ko00D00000003ooP00ooh00?on00000003003o
g00500000000ok/00?nk003o^`0000000`00oi/01@0000000?mm003oO@00ogd0000000800?mP00D0
0000003oA000od@00?m400000003003o:@0500000000oa400?lA003o4@0000000`06o`001@000000
6ol001_o000Ko`0000000080;ol000D0000004;o0012o`00@_l00000000305;o00050000001Qo`00
HOl0067o000000000`1^o`0010000000N_l007[o0000000308Co00050000002<o`00S?l008co0000
00000`2Co`001@000000V?l009So002Ho`00000000<0W?l000<000000?ooo`3oool0103oool10000
0040oooo000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?ooo`00
803oool00`000000oooo0?ooo`050?ooo`040000000lo`00??l0000000<0<ol000D0000002[o000Z
o`00:_l00000000301ko00050000000Bo`004_l001;o000000000`04o`00100000000?l9003o2@00
0003003o6P0500000000ob/00?l[003o:`0000000`00och01@0000000?mB003oDP00oe80000000<0
0?mX00@00000003oOP00ogh000000`00oiH01@0000000?n_003o[`00ojl0000000<00?o900D00000
003oi000on@00?oT00000003003oo`0400000000hol00>?o000000<00<Go00D00000002Wo`00Yol0
0:Oo000000030028o`0500000000JOl006Wo001Yo`0000000`00B_l01000000002[o000Zo`000003
000:o`050000000E0?l05@3o01D0o`0000000`0f0?l01@000000E@3o05D0o`1E0?l0000000<0M@3o
00@0000009@0o`2D0?l000000`2b0?l01@000000d03o0=00o`3@0?l0000000<0k@3o00D000000?l0
mP3o0?H0o`3f000000030?l0f`040000003o0<40o`31000000<0o`2Y00D000000?l0TP3o0980o`2B
000000030?l0O@050000003o06T0o`1Y0?l0J@0000000P3o05L01@000000o`160?l0AP3o04H00000
00<0o`0h00D000000?l0;03o02`0o`0/000000030?l08@050000003o01T0o`0I0?l06@0000000P3o
01801@000000o`0>0?l03P3o00h0000000<0o`0<00D000000?l0303o00`0o`0<000000030?l03P05
0000003o0180o`0B0?l04P0000000P3o01T01@000000o`0Q0?l08@3o0240000000<0o`0/00D00000
0?l0>03o03P0o`0h000000030?l0AP050000003o05L0o`1G0?l0E`0000000P3o06T01@000000o`1m
0?l0O@3o07d0000000<0o`2B00D000000?l0Z@3o0:T0o`2Y000000030?l0`@050000003o0=/0o`3K
0?l0f`0000000P3o0?H01@000000k@3o0>d0o`3]0?l0000000<0d03o00D000000;80o`2b0?l0/P3o
0000000309@0o`050000001e0?l0M@3o07D0o`0000000P1E0?l01@000000=P3o03H0o`0f0?l00000
00<05@3o00D00000000:o`002_l000[o00000003000Zo`0500000000B_l004[o001:o`0000000P00
JOl01@00000008So0028o`00R?l0000000<00:Oo00D000000035o`00aOl00<Go00000003003So`05
00000000ool00?oo003oo`0000000P00on@01@0000000?o9003ob@00olT0000000<00?n_00D00000
003oUP00oiH00?nF00000003003oOP0500000000ofP00?mX003oJ00000000P00oe801@0000000?ln
003o?P00och0000000<00?l[00D00000003o6P00oaX00?lJ00000003003o2@0400000004o`001?l0
000000<04_l000D0000001ko000No`007_l00000000302[o00050000000co`00<ol003?o00000000
0`0lo`000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3o
ool010000000??l003co0000000303?o00050000000Zo`00:_l002[o000000000`0No`001@000000
4_l001;o000Bo`00000000<01?l000@00000003o2@00o`T000000`00oaX01@0000000?l[003o:`00
ob/0000000<00?ln00D00000003oDP00oe800?mB00000003003oJ00400000000ogh00?mn000000<0
0?nF00D00000003o[`00ojl00?n_00000003003ob@0500000000on@00?oT003oi00000000`00ool0
100000000>?o003So`0000030035o`0500000000Yol00:Oo002Wo`0000000`00R?l01@00000006Wo
001Yo`00JOl0000000<004[o00@00000000Zo`00:_l000000`002_l01@0000005@3o01D0o`0E0?l0
000000<0=P3o00D0000005D0o`1E0?l0E@3o0000000307D0o`040000002D0?l0U03o000000<0/P3o
00D000000=00o`3@0?l0d03o000000030>d0o`050000003o0?H0o`3f0?l0mP0000000`3o0=/01000
0000o`310?l0`@0000030?l0Z@050000003o0980o`2B0?l0TP0000000`3o07d01@000000o`1Y0?l0
J@3o06T000000080o`1G00D000000?l0AP3o04H0o`16000000030?l0>0050000003o02`0o`0/0?l0
;00000000`3o02401@000000o`0I0?l06@3o01T000000080o`0B00D000000?l03P3o00h0o`0>0000
00030?l030050000003o00`0o`0<0?l0300000000`3o00h01@000000o`0B0?l04P3o018000000080
o`0I00D000000?l08@3o0240o`0Q000000030?l0;0050000003o03P0o`0h0?l0>00000000`3o04H0
1@000000o`1G0?l0E`3o05L000000080o`1Y00D000000?l0O@3o07d0o`1m000000030?l0TP050000
003o0:T0o`2Y0?l0Z@0000000`3o0<401@000000o`3K0?l0f`3o0=/000000080o`3f00D000000>d0
o`3]0?l0k@3o000000030=00o`050000002b0?l0/P3o0;80o`0000000`2D0?l01@000000M@3o07D0
o`1e0?l000000080E@3o00D0000003H0o`0f0?l0=P3o0000000301D0o`05000000002_l000[o000:
o`0000000`00:_l01@00000004[o001:o`00B_l00000008006Wo00D000000028o`00R?l008So0000
0003002Wo`0500000000aOl00<Go0035o`0000000`00hol01@0000000?oo003oo`00ool000000080
0?oT00D00000003ob@00olT00?o900000003003o[`0500000000oiH00?nF003oUP0000000`00ogh0
1@0000000?mX003oJ000ofP0000000800?mB00D00000003o?P00och00?ln00000003003o:`050000
0000oaX00?lJ003o6P0000000`00o`T0100000001?l000Co0000000301;o00050000000No`007_l0
01ko000000000`0Zo`001@000000<ol003?o000co`00000000<0??l000<000000?ooo`3oool0103o
ool100000040oooo000P0?ooo`8000001P3oool010000000??l003co0000000303?o00050000000Z
o`00:_l002[o000000000`0No`001@0000004_l001;o000Bo`00000000<01?l000@00000003o2@00
o`T000000`00oaX01@0000000?l[003o:`00ob/0000000<00?ln00D00000003oDP00oe800?mB0000
0003003oJ00400000000ogh00?mn000000<00?nF00D00000003o[`00ojl00?n_00000003003ob@05
00000000on@00?oT003oi00000000`00ool0100000000>?o003So`0000030035o`0500000000Yol0
0:Oo002Wo`0000000`00R?l01@00000006Wo001Yo`00JOl0000000<004[o00@00000000Zo`00:_l0
00000`002_l01@0000005@3o01D0o`0E0?l0000000<0=P3o00D0000005D0o`1E0?l0E@3o00000003
07D0o`040000002D0?l0U03o000000<0/P3o00D000000=00o`3@0?l0d03o000000030>d0o`050000
003o0?H0o`3f0?l0mP0000000`3o0=/010000000o`310?l0`@0000030?l0Z@050000003o0980o`2B
0?l0TP0000000`3o07d01@000000o`1Y0?l0J@3o06T000000080o`1G00D000000?l0AP3o04H0o`16
000000030?l0>0050000003o02`0o`0/0?l0;00000000`3o02401@000000o`0I0?l06@3o01T00000
0080o`0B00D000000?l03P3o00h0o`0>000000030?l030050000003o00`0o`0<0?l0300000000`3o
00h01@000000o`0B0?l04P3o018000000080o`0I00D000000?l08@3o0240o`0Q000000030?l0;005
0000003o03P0o`0h0?l0>00000000`3o04H01@000000o`1G0?l0E`3o05L000000080o`1Y00D00000
0?l0O@3o07d0o`1m000000030?l0TP050000003o0:T0o`2Y0?l0Z@0000000`3o0<401@000000o`3K
0?l0f`3o0=/000000080o`3f00D000000>d0o`3]0?l0k@3o000000030=00o`050000002b0?l0/P3o
0;80o`0000000`2D0?l01@000000M@3o07D0o`1e0?l000000080E@3o00D0000003H0o`0f0?l0=P3o
0000000301D0o`05000000002_l000[o000:o`0000000`00:_l01@00000004[o001:o`00B_l00000
008006Wo00D000000028o`00R?l008So00000003002Wo`0500000000aOl00<Go0035o`0000000`00
hol01@0000000?oo003oo`00ool0000000800?oT00D00000003ob@00olT00?o900000003003o[`05
00000000oiH00?nF003oUP0000000`00ogh01@0000000?mX003oJ000ofP0000000800?mB00D00000
003o?P00och00?ln00000003003o:`0500000000oaX00?lJ003o6P0000000`00o`T0100000001?l0
00Co0000000301;o00050000000No`007_l001ko000000000`0Zo`001@000000<ol003?o000co`00
000000<0??l000<000000?ooo`3oool00`3oool200000040oooo000P0?ooo`030000003oool0oooo
00D0ooooo`000021000000H0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`04
00000000ocT00?li000000<00?m400D00000003oD@00oe400?mA00000003003oG`0500000000ofd0
0?m]003oK@0000000`00ogd0100000000?n>003oSP000003003oW`0500000000ok800?nb003o/P00
00000`00olD01@0000000?oJ003ofP00omX0000000<00?o_00@00000003jo`00n_l000000`00hol0
1@0000000<co003<o`00c?l0000000<00;Co00D00000002Ko`00Vol009_o000000030021o`040000
0000Iol006Oo000000<004go00D00000000bo`00<_l003;o00000003000Go`05000000030?l00`3o
00<0o`0000000`0N0?l010000000>P3o03X0o`00000305D0o`050000001`0?l0L03o0700o`000000
0`2;0?l01@000000YP3o0:H0o`2V0?l0000000<0`03o00@000000=X0o`3J0?l000000`3c0?l01@00
0000o`3c0?l0l`3o0?<0000000<0o`3L00D000000?l0a@3o0<D0o`35000000030?l0[`040000003o
09X0o`2J000000<0o`2600D000000?l0M03o07@0o`1d000000030?l0H`050000003o05<0o`1C0?l0
D`0000000P3o04@01@000000o`0g0?l0=`3o03L0000000<0o`0/00D000000?l08P3o0280o`0R0000
00030?l06@050000003o0180o`0B0?l04P0000000P3o00d01@000000o`0:0?l02P3o00X0000000<0
o`0800D000000?l0203o00P0o`08000000030?l02P050000003o00d0o`0=0?l03@0000000P3o0180
1@000000o`0I0?l06@3o01T0000000<0o`0R00D000000?l0;03o02`0o`0/000000030?l0=`050000
003o04@0o`140?l0A00000000P3o05<01@000000o`1S0?l0H`3o06<0000000<0o`1d00D000000?l0
QP3o08H0o`26000000030?l0VP050000003o0:l0o`2_0?l0[`0000000P3o0<D01@000000o`3L0?l0
g03o0=`0000000<0o`3c00D000000?<0o`3c0?l0l`3o000000030=X0o`05000000300?l0`03o0<00
o`0000000P2V0?l01@000000R`3o08/0o`2;0?l0000000<0L03o00D0000005D0o`1E0?l0E@3o0000
000303X0o`050000000N0?l07P3o01h0o`0000000P030?l01@00000001Oo000Go`005ol0000000<0
03;o00D00000001=o`00COl004go00000003001Wo`0500000000POl0087o0021o`0000000P00Vol0
1@0000000;Co002do`00]?l0000000<00<co00D00000003So`00hol00>?o00000003003jo`050000
0000onl00?o_003ok`0000000P00omX01@0000000?o5003oa@00olD0000000<00?nb00D00000003o
W`00oil00?nO00000003003oSP0400000000ogd00?mm000000<00?m]00D00000003oG`00oel00?mO
00000003003oD@0500000000od@00?m4003oA00000000`00ocT00`000000oooo0?ooo`040?ooo`40
00000@3oool00200oooo00<000000?ooo`3oool01@3oool0100000000?li003o>@000003003oA005
00000000oe400?mA003oD@0000000`00oel01@0000000?m]003oK@00ofd0000000<00?mm00@00000
003oSP00ohh000000`00oil01@0000000?nb003o/P00ok80000000<00?o500D00000003ofP00omX0
0?oJ00000003003ok`0400000000n_l00?[o000000<00>?o00D00000003<o`00c?l00<co00000003
002do`0500000000Vol009_o002Ko`0000000`00POl01000000006Oo001Wo`000003001=o`050000
0000<_l003;o000bo`0000000`005ol01@0000000`3o00<0o`030?l0000000<07P3o00@0000003X0
o`0j0?l000000`1E0?l01@000000L03o0700o`1`0?l0000000<0R`3o00D000000:H0o`2V0?l0YP3o
000000030<00o`040000003J0?l0fP3o000000<0l`3o00D000000?l0l`3o0?<0o`3c000000030?l0
g0050000003o0<D0o`350?l0a@0000000`3o0:l010000000o`2J0?l0VP0000030?l0QP050000003o
07@0o`1d0?l0M00000000`3o06<01@000000o`1C0?l0D`3o05<000000080o`1400D000000?l0=`3o
03L0o`0g000000030?l0;0050000003o0280o`0R0?l08P0000000`3o01T01@000000o`0B0?l04P3o
018000000080o`0=00D000000?l02P3o00X0o`0:000000030?l020050000003o00P0o`080?l02000
00000`3o00X01@000000o`0=0?l03@3o00d000000080o`0B00D000000?l06@3o01T0o`0I00000003
0?l08P050000003o02`0o`0/0?l0;00000000`3o03L01@000000o`140?l0A03o04@000000080o`1C
00D000000?l0H`3o06<0o`1S000000030?l0M0050000003o08H0o`260?l0QP0000000`3o09X01@00
0000o`2_0?l0[`3o0:l000000080o`3500D000000?l0g03o0=`0o`3L000000030?l0l`050000003c
0?l0l`3o0?<0o`0000000`3J0?l01@000000`03o0<00o`300?l000000080YP3o00D0000008/0o`2;
0?l0R`3o000000030700o`050000001E0?l0E@3o05D0o`0000000`0j0?l01@0000007P3o01h0o`0N
0?l0000000800`3o00D00000000Go`005ol001Oo00000003000bo`0500000000COl004go001=o`00
00000`00Iol01@000000087o0021o`00POl00000008009_o00D00000002do`00]?l00;Co00000003
003<o`0500000000hol00>?o003So`0000000`00n_l01@0000000?o_003ok`00onl0000000800?oJ
00D00000003oa@00olD00?o500000003003o/P0500000000oil00?nO003oW`0000000`00ohh01000
00000?mm003oO@000003003oK@0500000000oel00?mO003oG`0000000`00oe401@0000000?m4003o
A000od@0000000<00?li00<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003o
ool0oooo00D0oooo00@00000003o>@00ocT000000`00od@01@0000000?mA003oD@00oe40000000<0
0?mO00D00000003oK@00ofd00?m]00000003003oO@0400000000ohh00?n>000000<00?nO00D00000
003o/P00ok800?nb00000003003oa@0500000000omX00?oJ003ofP0000000`00onl0100000000?[o
003jo`000003003So`0500000000c?l00<co003<o`0000000`00]?l01@00000009_o002Ko`00Vol0
000000<0087o00@00000001Wo`00Iol000000`00COl01@00000003;o000bo`00<_l0000000<001Oo
00D0000000<0o`030?l00`3o0000000301h0o`040000000j0?l0>P3o000000<0E@3o00D000000700
o`1`0?l0L03o0000000308/0o`050000002V0?l0YP3o0:H0o`0000000`300?l010000000fP3o0=X0
o`0000030?<0o`050000003o0?<0o`3c0?l0l`0000000`3o0=`01@000000o`350?l0a@3o0<D00000
00<0o`2_00@000000?l0VP3o09X000000`3o08H01@000000o`1d0?l0M03o07@0000000<0o`1S00D0
00000?l0D`3o05<0o`1C000000020?l0A0050000003o03L0o`0g0?l0=`0000000`3o02`01@000000
o`0R0?l08P3o0280000000<0o`0I00D000000?l04P3o0180o`0B000000020?l03@050000003o00X0
o`0:0?l02P0000000`3o00P01@000000o`080?l0203o00P0000000<0o`0:00D000000?l03@3o00d0
o`0=000000020?l04P050000003o01T0o`0I0?l06@0000000`3o02801@000000o`0/0?l0;03o02`0
000000<0o`0g00D000000?l0A03o04@0o`14000000020?l0D`050000003o06<0o`1S0?l0H`000000
0`3o07@01@000000o`260?l0QP3o08H0000000<0o`2J00D000000?l0[`3o0:l0o`2_000000020?l0
a@050000003o0=`0o`3L0?l0g00000000`3o0?<01@000000l`3o0?<0o`3c0?l0000000<0fP3o00D0
00000<00o`300?l0`03o000000020:H0o`050000002;0?l0R`3o08/0o`0000000`1`0?l01@000000
E@3o05D0o`1E0?l0000000<0>P3o00D0000001h0o`0N0?l07P3o0000000200<0o`05000000005ol0
01Oo000Go`0000000`00<_l01@00000004go001=o`00COl0000000<006Oo00D000000021o`00POl0
087o00000002002Ko`0500000000]?l00;Co002do`0000000`00c?l01@0000000>?o003So`00hol0
000000<00?[o00D00000003ok`00onl00?o_00000002003ofP0500000000olD00?o5003oa@000000
0`00ok801@0000000?nO003oW`00oil0000000<00?n>00@00000003oO@00ogd000000`00ofd01@00
00000?mO003oG`00oel0000000<00?mA00D00000003oA000od@00?m400000003003o>@030000003o
ool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooool00000P@000006
0?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool0100000000?nh003o^0000003
003oa@0500000000om<00?oC003od`0000000`00on801@0000000?oa003ol@00oo40000000<00?go
00@00000003/o`00k?l000000`00fol01@0000000<Wo0039o`00bOl0000000<00;Ko00D00000002R
o`00X_l00:;o00000003002?o`0400000000N_l007[o000000<006Go00D00000001@o`00D?l0053o
00000003000jo`05000000009?l002Co000To`0000000`003Ol010000000203o00P0o`00000301l0
o`050000000f0?l0=P3o03H0o`0000000`1>0?l01@000000I@3o06D0o`1U0?l0000000<0O03o00@0
000009<0o`2C0?l000000`2Z0?l01@000000`03o0<00o`300?l0000000<0e`3o00D000000>`0o`3/
0?l0k03o000000030?l0o@040000003o0>P0o`3X000000<0o`3D00D000000?l0`03o0<00o`300000
00030?l0[@050000003o09/0o`2K0?l0V`0000000`3o08T010000000o`1h0?l0N00000030?l0J@05
0000003o05X0o`1J0?l0FP0000000`3o04`01@000000o`100?l0@03o040000000080o`0d00D00000
0?l0:P3o02X0o`0Z000000030?l08@050000003o01T0o`0I0?l06@0000000`3o01<01@000000o`0=
0?l03@3o00d000000080o`0900D000000?l01`3o00L0o`07000000030?l01@050000003o00D0o`05
0?l01@0000000`3o00L01@000000o`090?l02@3o00T000000080o`0=00D000000?l04`3o01<0o`0C
000000030?l06@050000003o0240o`0Q0?l08@0000000`3o02X01@000000o`0d0?l0=03o03@00000
0080o`1000D000000?l0C03o04`0o`1<000000030?l0FP050000003o06T0o`1Y0?l0J@0000000`3o
07P01@000000o`290?l0R@3o08T000000080o`2K00D000000?l0[@3o0:d0o`2]000000030?l0`005
0000003o0=@0o`3D0?l0e00000000`3o0>P01@000000o`3m0?l0o@3o0?d000000080k03o00D00000
0=L0o`3G0?l0e`3o000000030<00o`050000002Z0?l0ZP3o0:X0o`0000000`2C0?l01@000000O03o
07`0o`1l0?l000000080I@3o00D0000004h0o`1>0?l0CP3o0000000303H0o`050000000O0?l07`3o
01l0o`0000000`080?l01@00000000go000=o`003Ol00000008002Co00D00000000jo`00>_l003[o
00000003001@o`0500000000IOl006Go001Uo`0000000`00N_l01@00000008oo002?o`00Sol00000
00800:;o00D00000002fo`00]_l00;Ko000000030039o`0500000000fol00=_o003Ko`0000000`00
k?l0100000000?go003mo`000003003ol@0500000000on800?oR003ohP0000000`00om<01@000000
0?o5003oa@00olD0000000<00?nh00<000000?ooo`3oool0103oool100000040oooo000P0?ooo`03
0000003oool0oooo00D0oooo00@00000003o^000okP000000`00olD01@0000000?oC003od`00om<0
000000<00?oR00D00000003ol@00oo400?oa00000003003mo`0400000000k?l00>co000000<00=_o
00D000000039o`00bOl00<Wo00000003002fo`0500000000X_l00:;o002Ro`0000000`00Sol01000
000007[o001jo`000003001Uo`0500000000D?l0053o001@o`0000000`00>_l01@00000002Co000T
o`009?l0000000<000go00@0000000P0o`080?l000000`0O0?l01@000000=P3o03H0o`0f0?l00000
00<0CP3o00D0000006D0o`1U0?l0I@3o0000000307`0o`040000002C0?l0T`3o000000<0ZP3o00D0
00000<00o`300?l0`03o000000030=L0o`050000003/0?l0k03o0>`0o`0000000`3o0?d010000000
o`3X0?l0j00000030?l0e0050000003o0<00o`300?l0`00000000`3o0:d01@000000o`2K0?l0V`3o
09/0000000<0o`2900@000000?l0N03o07P000000`3o06T01@000000o`1J0?l0FP3o05X0000000<0
o`1<00D000000?l0@03o0400o`10000000020?l0=0050000003o02X0o`0Z0?l0:P0000000`3o0240
1@000000o`0I0?l06@3o01T0000000<0o`0C00D000000?l03@3o00d0o`0=000000020?l02@050000
003o00L0o`070?l01`0000000`3o00D01@000000o`050?l01@3o00D0000000<0o`0700D000000?l0
2@3o00T0o`09000000020?l03@050000003o01<0o`0C0?l04`0000000`3o01T01@000000o`0Q0?l0
8@3o0240000000<0o`0Z00D000000?l0=03o03@0o`0d000000020?l0@0050000003o04`0o`1<0?l0
C00000000`3o05X01@000000o`1Y0?l0J@3o06T0000000<0o`1h00D000000?l0R@3o08T0o`290000
00020?l0V`050000003o0:d0o`2]0?l0[@0000000`3o0<001@000000o`3D0?l0e03o0=@0000000<0
o`3X00D000000?l0o@3o0?d0o`3m000000020>`0o`050000003G0?l0e`3o0=L0o`0000000`300?l0
1@000000ZP3o0:X0o`2Z0?l0000000<0T`3o00D0000007`0o`1l0?l0O03o0000000206D0o`050000
001>0?l0CP3o04h0o`0000000`0f0?l01@0000007`3o01l0o`0O0?l0000000<0203o00D00000000=
o`003Ol000go00000002000To`0500000000>_l003[o000jo`0000000`00D?l01@00000006Go001U
o`00IOl0000000<007[o00D00000002?o`00Sol008oo00000002002Ro`0500000000]_l00;Ko002f
o`0000000`00bOl01@0000000=_o003Ko`00fol0000000<00>co00@00000003mo`00oOl000000`00
oo401@0000000?oR003ohP00on80000000<00?oC00D00000003oa@00olD00?o500000003003o^003
0000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`040000
0000okP00?nh000000<00?o500D00000003od`00om<00?oC00000003003ohP0500000000oo400?oa
003ol@0000000`00oOl0100000000>co003/o`000003003Ko`0500000000bOl00<Wo0039o`000000
0`00]_l01@0000000:;o002Ro`00X_l0000000<008oo00@00000001jo`00N_l000000`00IOl01@00
0000053o001@o`00D?l0000000<003[o00D00000000To`009?l002Co00000003000=o`0400000008
0?l0203o000000<07`3o00D0000003H0o`0f0?l0=P3o0000000304h0o`050000001U0?l0I@3o06D0
o`0000000`1l0?l010000000T`3o09<0o`0000030:X0o`05000000300?l0`03o0<00o`0000000`3G
0?l01@000000k03o0>`0o`3/0?l0000000<0o`3m00@000000?l0j03o0>P000000`3o0=@01@000000
o`300?l0`03o0<00000000<0o`2]00D000000?l0V`3o09/0o`2K000000030?l0R@040000003o07P0
o`1h000000<0o`1Y00D000000?l0FP3o05X0o`1J000000030?l0C0050000003o0400o`100?l0@000
00000P3o03@01@000000o`0Z0?l0:P3o02X0000000<0o`0Q00D000000?l06@3o01T0o`0I00000003
0?l04`050000003o00d0o`0=0?l03@0000000P3o00T01@000000o`070?l01`3o00L0000000<0o`05
00D000000?l01@3o00D0o`05000000030?l01`050000003o00T0o`090?l02@0000000P3o00d01@00
0000o`0C0?l04`3o01<0000000<0o`0I00D000000?l08@3o0240o`0Q000000030?l0:P050000003o
03@0o`0d0?l0=00000000P3o04001@000000o`1<0?l0C03o04`0000000<0o`1J00D000000?l0J@3o
06T0o`1Y000000030?l0N0050000003o08T0o`290?l0R@0000000P3o09/01@000000o`2]0?l0[@3o
0:d0000000<0o`3000D000000?l0e03o0=@0o`3D000000030?l0j0050000003o0?d0o`3m0?l0o@00
00000P3/0?l01@000000e`3o0=L0o`3G0?l0000000<0`03o00D000000:X0o`2Z0?l0ZP3o00000003
09<0o`050000001l0?l0O03o07`0o`0000000P1U0?l01@000000CP3o04h0o`1>0?l0000000<0=P3o
00D0000001l0o`0O0?l07`3o0000000300P0o`05000000003Ol000go000=o`0000000P009?l01@00
000003[o000jo`00>_l0000000<0053o00D00000001Uo`00IOl006Go00000003001jo`0500000000
Sol008oo002?o`0000000P00X_l01@0000000;Ko002fo`00]_l0000000<00<Wo00D00000003Ko`00
fol00=_o00000003003/o`0400000000oOl00?go000000<00?oa00D00000003ohP00on800?oR0000
0003003od`0500000000olD00?o5003oa@0000000`00okP00`000000oooo0?ooo`040?ooo`400000
0@3oool00200oooo00<000000?ooo`3oool01@3ooooo0000084000001P3oool100000040oooo000P
0?ooo`030000003oool0oooo00D0oooo00@00000003=o`00cOl000000`00_ol01@0000000;3o002`
o`00/?l0000000<00:7o00D00000002Bo`00T_l009;o000000030022o`0400000000L_l007;o0000
00<0067o00D00000001@o`00D?l0053o00000003000oo`0500000000;Ol002go000]o`0000000`00
6ol01000000000So0008o`00000300T0o`050000000L0?l0703o01`0o`0000000`0_0?l01@000000
@`3o04<0o`130?l0000000<0EP3o00@0000006X0o`1Z0?l000000`1m0?l01@000000T@3o0940o`2A
0?l0000000<0Y03o00D000000;L0o`2g0?l0]`3o000000030</0o`040000003N0?l0gP3o000000<0
l03o00D000000?l0o03o0?`0o`3l000000030?l0jP050000003o0=P0o`3H0?l0f00000000`3o0<L0
10000000o`2f0?l0]P0000030?l0YP050000003o09H0o`2F0?l0UP0000000`3o08L01@000000o`1h
0?l0N03o07P0000000<0o`1Z00@000000?l0G@3o05d000000`3o05401@000000o`150?l0A@3o04D0
000000<0o`0j00D000000?l0<@3o0340o`0a000000020?l0:0050000003o0200o`0P0?l080000000
0`3o01T01@000000o`0B0?l04P3o0180000000<0o`0=00D000000?l02@3o00T0o`09000000020?l0
1P050000003o00@0o`040?l0100000000`3o00<01@000000o`030?l00`3o00<0000000<0o`0400D0
00000?l01P3o00H0o`06000000020?l02@050000003o00d0o`0=0?l03@0000000`3o01801@000000
o`0I0?l06@3o01T0000000<0o`0P00D000000?l0:03o02P0o`0X000000020?l0<@050000003o03X0
o`0j0?l0>P0000000`3o04D01@000000o`1A0?l0D@3o0540000000<0o`1M00D000000?l0JP3o06X0
o`1Z000000020?l0N0050000003o08L0o`270?l0Q`0000000`3o09H01@000000o`2V0?l0YP3o0:H0
000000<0o`2f00D000000?l0a`3o0<L0o`37000000020?l0f0050000003o0>X0o`3Z0?l0jP000000
0`3o0?`01@000000l03o0?00o`3`0?l0000000<0gP3o00D000000</0o`3;0?l0b`3o000000020;L0
o`050000002T0?l0Y03o0:@0o`0000000`2A0?l01@000000O@3o07d0o`1m0?l0000000<0JP3o00D0
000005H0o`1F0?l0EP3o0000000204<0o`050000000_0?l0;`3o02l0o`0000000`0L0?l01@000000
2@3o00T0o`090?l0000000<000So00D00000000Ko`006ol001_o00000002000]o`0500000000?ol0
03oo000oo`0000000`00D?l01@000000067o001Qo`00HOl0000000<007;o00@000000022o`00P_l0
00000`00T_l01@0000000:7o002Qo`00XOl0000000<00;3o00D00000002oo`00_ol00;oo00000003
003=o`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oo
o`0400000000cOl00<go000000<00;oo00D00000002`o`00/?l00;3o00000003002Qo`0500000000
T_l009;o002Bo`0000000`00P_l01000000007;o001bo`000003001Qo`0500000000D?l0053o001@
o`0000000`00?ol01@00000002go000]o`00;Ol0000000<001_o00@000000008o`002?l000000`09
0?l01@000000703o01`0o`0L0?l0000000<0;`3o00D0000004<0o`130?l0@`3o0000000305H0o`04
0000001Z0?l0JP3o000000<0O@3o00D000000940o`2A0?l0T@3o000000030:@0o`050000002g0?l0
]`3o0;L0o`0000000`3;0?l010000000gP3o0=h0o`0000030?00o`050000003o0?`0o`3l0?l0o000
00000`3o0>X01@000000o`3H0?l0f03o0=P0000000<0o`3700@000000?l0]P3o0;H000000`3o0:H0
1@000000o`2F0?l0UP3o09H0000000<0o`2700D000000?l0N03o07P0o`1h000000030?l0JP040000
003o05d0o`1M000000<0o`1A00D000000?l0A@3o04D0o`15000000030?l0>P050000003o0340o`0a
0?l0<@0000000P3o02P01@000000o`0P0?l0803o0200000000<0o`0I00D000000?l04P3o0180o`0B
000000030?l03@050000003o00T0o`090?l02@0000000P3o00H01@000000o`040?l0103o00@00000
00<0o`0300D000000?l00`3o00<0o`03000000030?l010050000003o00H0o`060?l01P0000000P3o
00T01@000000o`0=0?l03@3o00d0000000<0o`0B00D000000?l06@3o01T0o`0I000000030?l08005
0000003o02P0o`0X0?l0:00000000P3o03401@000000o`0j0?l0>P3o03X0000000<0o`1500D00000
0?l0D@3o0540o`1A000000030?l0G@050000003o06X0o`1Z0?l0JP0000000P3o07P01@000000o`27
0?l0Q`3o08L0000000<0o`2F00D000000?l0YP3o0:H0o`2V000000030?l0]P050000003o0<L0o`37
0?l0a`0000000P3o0=P01@000000o`3Z0?l0jP3o0>X0000000<0o`3l00D000000?00o`3`0?l0l03o
000000030=h0o`050000003;0?l0b`3o0</0o`0000000P2g0?l01@000000Y03o0:@0o`2T0?l00000
00<0T@3o00D0000007d0o`1m0?l0O@3o0000000306X0o`050000001F0?l0EP3o05H0o`0000000P13
0?l01@000000;`3o02l0o`0_0?l0000000<0703o00D0000000T0o`090?l02@3o000000030008o`05
000000006ol001_o000Ko`0000000P00;Ol01@00000003oo000oo`00?ol0000000<0053o00D00000
001Qo`00HOl0067o00000003001bo`0400000000P_l008;o000000<009;o00D00000002Qo`00XOl0
0:7o00000003002`o`0500000000_ol00;oo002oo`0000000`00cOl00`000000oooo0?ooo`040?oo
o`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool0100000000<go003=o`000003002o
o`0500000000/?l00;3o002`o`0000000`00XOl01@00000009;o002Bo`00T_l0000000<008;o00@0
0000001bo`00L_l000000`00HOl01@000000053o001@o`00D?l0000000<003oo00D00000000]o`00
;Ol002go00000003000Ko`04000000002?l000So000000<02@3o00D0000001`0o`0L0?l0703o0000
000302l0o`05000000130?l0@`3o04<0o`0000000`1F0?l010000000JP3o06X0o`00000307d0o`05
0000002A0?l0T@3o0940o`0000000`2T0?l01@000000]`3o0;L0o`2g0?l0000000<0b`3o00@00000
0=h0o`3N0?l000000`3`0?l01@000000o`3l0?l0o03o0?`0000000<0o`3Z00D000000?l0f03o0=P0
o`3H000000030?l0a`040000003o0;H0o`2f000000<0o`2V00D000000?l0UP3o09H0o`2F00000003
0?l0Q`050000003o07P0o`1h0?l0N00000000`3o06X010000000o`1M0?l0G@0000030?l0D@050000
003o04D0o`150?l0A@0000000`3o03X01@000000o`0a0?l0<@3o034000000080o`0X00D000000?l0
803o0200o`0P000000030?l06@050000003o0180o`0B0?l04P0000000`3o00d01@000000o`090?l0
2@3o00T000000080o`0600D000000?l0103o00@0o`04000000030?l00`050000003o00<0o`030?l0
0`0000000`3o00@01@000000o`060?l01P3o00H000000080o`0900D000000?l03@3o00d0o`0=0000
00030?l04P050000003o01T0o`0I0?l06@0000000`3o02001@000000o`0X0?l0:03o02P000000080
o`0a00D000000?l0>P3o03X0o`0j000000030?l0A@050000003o0540o`1A0?l0D@0000000`3o05d0
1@000000o`1Z0?l0JP3o06X000000080o`1h00D000000?l0Q`3o08L0o`27000000030?l0UP050000
003o0:H0o`2V0?l0YP0000000`3o0;H01@000000o`370?l0a`3o0<L000000080o`3H00D000000?l0
jP3o0>X0o`3Z000000030?l0o0050000003`0?l0l03o0?00o`0000000`3N0?l01@000000b`3o0</0
o`3;0?l000000080]`3o00D000000:@0o`2T0?l0Y03o000000030940o`050000001m0?l0O@3o07d0
o`0000000`1Z0?l01@000000EP3o05H0o`1F0?l000000080@`3o00D0000002l0o`0_0?l0;`3o0000
000301`0o`05000000090?l02@3o00T0o`0000000`002?l01@00000001_o000Ko`006ol000000080
02go00D00000000oo`00?ol003oo00000003001@o`0500000000HOl0067o001Qo`0000000`00L_l0
1000000008;o0022o`000003002Bo`0500000000XOl00:7o002Qo`0000000`00/?l01@0000000;oo
002oo`00_ol0000000<00<go00<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000
003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?ooo`00803oool00`000000oooo0?oo
o`050?ooo`0400000000Hol006?o000000<005Go00D000000017o`00Aol004Oo00000003000io`05
00000000:_l002[o000Zo`0000000`006ol01000000000co000<o`00000300<0o`050000000B0?l0
4P3o0180o`0000000`0R0?l01@000000<P3o0380o`0b0?l0000000<0@P3o00@0000005<0o`1C0?l0
00000`1S0?l01@000000M03o07@0o`1d0?l0000000<0Q@3o00D0000009D0o`2E0?l0U@3o00000003
0:H0o`040000002f0?l0]P3o000000<0a`3o00D000000=L0o`3G0?l0e`3o000000030>P0o`050000
003h0?l0n03o0?P0o`0000000`3o0?L010000000o`3W0?l0i`0000030?l0f0050000003o0<T0o`39
0?l0b@0000000`3o0;X01@000000o`2[0?l0Z`3o0:/0000000<0o`2M00@000000?l0T03o09000000
0`3o08801@000000o`1f0?l0MP3o07H0000000<0o`1Z00D000000?l0GP3o05h0o`1N000000030?l0
D`040000003o04P0o`18000000<0o`0o00D000000?l0=@3o03D0o`0e000000030?l0;@050000003o
02D0o`0U0?l09@0000000P3o01h01@000000o`0H0?l0603o01P0000000<0o`0B00D000000?l03@3o
00d0o`0=000000030?l02@050000003o00H0o`060?l01P0000000P3o00@01@000000o`020?l00P3o
0080000000<0o`0100D000000?l00@3o0040o`01000000030?l00P050000003o00@0o`040?l01000
00000P3o00H01@000000o`090?l02@3o00T0000000<0o`0=00D000000?l04P3o0180o`0B00000003
0?l060050000003o01h0o`0N0?l07P0000000P3o02D01@000000o`0]0?l0;@3o02d0000000<0o`0e
00D000000?l0?`3o03l0o`0o000000030?l0B0050000003o05<0o`1C0?l0D`0000000P3o05h01@00
0000o`1Z0?l0JP3o06X0000000<0o`1f00D000000?l0PP3o0880o`22000000030?l0T0050000003o
09d0o`2M0?l0W@0000000P3o0:/01@000000o`2j0?l0^P3o0;X0000000<0o`3900D000000?l0f03o
0=P0o`3H000000030?l0i`050000003o0?L0o`3g0?l0m`0000000P3h0?l01@000000j03o0>P0o`3X
0?l0000000<0e`3o00D000000<L0o`370?l0a`3o000000030;H0o`050000002V0?l0YP3o0:H0o`00
00000P2E0?l01@000000Q@3o08D0o`250?l0000000<0M03o00D0000006<0o`1S0?l0H`3o00000003
05<0o`05000000120?l0@P3o0480o`0000000P0b0?l01@0000008P3o0280o`0R0?l0000000<04P3o
00D0000000<0o`030?l00`3o00000003000<o`04000000006ol001_o000000<002[o00D00000000i
o`00>Ol003Wo000000030017o`0500000000EOl005Go001Eo`0000000`00Hol00`000000oooo0?oo
o`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool01000000006?o001So`00
0003001Eo`0500000000Aol004Oo0017o`0000000`00>Ol01@00000002[o000Zo`00:_l0000000<0
01_o00@00000000<o`003?l000000`030?l01@0000004P3o0180o`0B0?l0000000<08P3o00D00000
0380o`0b0?l0<P3o000000030480o`040000001C0?l0D`3o000000<0H`3o00D0000007@0o`1d0?l0
M03o0000000308D0o`050000002E0?l0U@3o09D0o`0000000`2V0?l010000000]P3o0;H0o`000003
0<L0o`050000003G0?l0e`3o0=L0o`0000000`3X0?l01@000000n03o0?P0o`3h0?l0000000<0o`3g
00@000000?l0i`3o0>L000000`3o0=P01@000000o`390?l0b@3o0<T0000000<0o`2j00D000000?l0
Z`3o0:/0o`2[000000030?l0W@040000003o0900o`2@000000<0o`2200D000000?l0MP3o07H0o`1f
000000030?l0JP050000003o05h0o`1N0?l0GP0000000`3o05<010000000o`180?l0B00000030?l0
?`050000003o03D0o`0e0?l0=@0000000`3o02d01@000000o`0U0?l09@3o02D000000080o`0N00D0
00000?l0603o01P0o`0H000000030?l04P050000003o00d0o`0=0?l03@0000000`3o00T01@000000
o`060?l01P3o00H000000080o`0400D000000?l00P3o0080o`02000000030?l00@050000003o0040
o`010?l00@0000000`3o00801@000000o`040?l0103o00@000000080o`0600D000000?l02@3o00T0
o`09000000030?l03@050000003o0180o`0B0?l04P0000000`3o01P01@000000o`0N0?l07P3o01h0
00000080o`0U00D000000?l0;@3o02d0o`0]000000030?l0=@050000003o03l0o`0o0?l0?`000000
0`3o04P01@000000o`1C0?l0D`3o05<000000080o`1N00D000000?l0JP3o06X0o`1Z000000030?l0
MP050000003o0880o`220?l0PP0000000`3o09001@000000o`2M0?l0W@3o09d000000080o`2[00D0
00000?l0^P3o0;X0o`2j000000030?l0b@050000003o0=P0o`3H0?l0f00000000`3o0>L01@000000
o`3g0?l0m`3o0?L000000080n03o00D000000>P0o`3X0?l0j03o000000030=L0o`05000000370?l0
a`3o0<L0o`0000000`2f0?l01@000000YP3o0:H0o`2V0?l000000080U@3o00D0000008D0o`250?l0
Q@3o0000000307@0o`050000001S0?l0H`3o06<0o`0000000`1C0?l01@000000@P3o0480o`120?l0
00000080<P3o00D000000280o`0R0?l08P3o000000030180o`05000000030?l00`3o00<0o`000000
0`003?l01000000001_o000Ko`000003000Zo`0500000000>Ol003Wo000io`0000000`00Aol01@00
000005Go001Eo`00EOl0000000<006?o00<000000?ooo`3oool0103oool100000040oooo000P0?oo
o`030000003oool0oooo00D0oooo00@00000001So`00Hol000000`00EOl01@00000004Oo0017o`00
Aol0000000<003Wo00D00000000Zo`00:_l002[o00000003000Ko`04000000003?l000co000000<0
0`3o00D000000180o`0B0?l04P3o000000030280o`050000000b0?l0<P3o0380o`0000000`120?l0
10000000D`3o05<0o`00000306<0o`050000001d0?l0M03o07@0o`0000000`250?l01@000000U@3o
09D0o`2E0?l0000000<0YP3o00@000000;H0o`2f0?l000000`370?l01@000000e`3o0=L0o`3G0?l0
000000<0j03o00D000000?P0o`3h0?l0n03o000000030?l0m`040000003o0>L0o`3W000000<0o`3H
00D000000?l0b@3o0<T0o`39000000030?l0^P050000003o0:/0o`2[0?l0Z`0000000`3o09d01000
0000o`2@0?l0T00000030?l0PP050000003o07H0o`1f0?l0MP0000000`3o06X01@000000o`1N0?l0
GP3o05h0000000<0o`1C00@000000?l0B03o04P000000`3o03l01@000000o`0e0?l0=@3o03D00000
00<0o`0]00D000000?l09@3o02D0o`0U000000020?l07P050000003o01P0o`0H0?l0600000000`3o
01801@000000o`0=0?l03@3o00d0000000<0o`0900D000000?l01P3o00H0o`06000000020?l01005
0000003o0080o`020?l00P0000000`3o00401@000000o`010?l00@3o0040000000<0o`0200D00000
0?l0103o00@0o`04000000020?l01P050000003o00T0o`090?l02@0000000`3o00d01@000000o`0B
0?l04P3o0180000000<0o`0H00D000000?l07P3o01h0o`0N000000020?l09@050000003o02d0o`0]
0?l0;@0000000`3o03D01@000000o`0o0?l0?`3o03l0000000<0o`1800D000000?l0D`3o05<0o`1C
000000020?l0GP050000003o06X0o`1Z0?l0JP0000000`3o07H01@000000o`220?l0PP3o08800000
00<0o`2@00D000000?l0W@3o09d0o`2M000000020?l0Z`050000003o0;X0o`2j0?l0^P0000000`3o
0<T01@000000o`3H0?l0f03o0=P0000000<0o`3W00D000000?l0m`3o0?L0o`3g000000020?P0o`05
0000003X0?l0j03o0>P0o`0000000`3G0?l01@000000a`3o0<L0o`370?l0000000<0]P3o00D00000
0:H0o`2V0?l0YP3o0000000209D0o`05000000250?l0Q@3o08D0o`0000000`1d0?l01@000000H`3o
06<0o`1S0?l0000000<0D`3o00D000000480o`120?l0@P3o000000020380o`050000000R0?l08P3o
0280o`0000000`0B0?l01@0000000`3o00<0o`030?l0000000<000co00@00000000Ko`006ol00000
0`00:_l01@00000003Wo000io`00>Ol0000000<004Oo00D00000001Eo`00EOl005Go00000003001S
o`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooool0
0000P@0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool01000000001Go
000Eo`0000030008o`05000000040?l0103o00@0o`0000000`0B0?l01@000000803o0200o`0P0?l0
000000<0;P3o00@0000003`0o`0l0?l000000`1;0?l01@000000F@3o05T0o`1I0?l0000000<0J03o
00D0000007H0o`1f0?l0MP3o0000000308D0o`040000002D0?l0U03o000000<0XP3o00D000000;40
o`2a0?l0/@3o000000030<00o`050000003?0?l0c`3o0<l0o`0000000`3M0?l010000000k03o0>`0
o`0000030?X0o`050000003o0?L0o`3g0?l0m`0000000`3o0>T01@000000o`3K0?l0f`3o0=/00000
00<0o`3=00@000000?l0`03o0<0000000`3o0;801@000000o`2V0?l0YP3o0:H0000000<0o`2I00D0
00000?l0S@3o08d0o`2=000000030?l0P@040000003o07H0o`1f000000<0o`1[00D000000?l0H03o
0600o`1P000000030?l0EP050000003o04`0o`1<0?l0C00000000`3o04<010000000o`0j0?l0>P00
00030?l0<P050000003o02/0o`0[0?l0:`0000000`3o02@01@000000o`0M0?l07@3o01d000000080
o`0H00D000000?l04P3o0180o`0B000000030?l03P050000003o00X0o`0:0?l02P0000000`3o00L0
1@000000o`040?l0103o00@000000080o`0200D000000?l00@3o0040o`01000000030?l000050000
003o0000o`000?l0000000000`3o00401@000000o`020?l00P3o008000000080o`0400D000000?l0
1`3o00L0o`07000000030?l02P050000003o00h0o`0>0?l03P0000000`3o01801@000000o`0H0?l0
603o01P000000080o`0M00D000000?l0903o02@0o`0T000000030?l0:`050000003o0380o`0b0?l0
<P0000000`3o03X01@000000o`130?l0@`3o04<000000080o`1<00D000000?l0EP3o05H0o`1F0000
00030?l0H0050000003o06/0o`1[0?l0J`0000000`3o07H01@000000o`210?l0P@3o084000000080
o`2=00D000000?l0V@3o09T0o`2I000000030?l0YP050000003o0;80o`2b0?l0/P0000000`3o0<00
1@000000o`3=0?l0c@3o0<d000000080o`3K00D000000?l0j@3o0>T0o`3Y000000030?l0m`050000
003j0?l0nP3o0?X0o`0000000`3/0?l01@000000g@3o0=d0o`3M0?l000000080c`3o00D000000<00
o`300?l0`03o000000030;40o`050000002R0?l0XP3o0:80o`0000000`2D0?l01@000000Q@3o08D0
o`250?l000000080MP3o00D0000006P0o`1X0?l0J03o0000000305T0o`050000001;0?l0B`3o04/0
o`0000000`0l0?l010000000;P3o02h0o`0000030200o`050000000B0?l04P3o0180o`0000000`04
0?l01@00000000So0008o`002?l0000000<001Go00<000000?ooo`3oool0103oool100000040oooo
000A0?ooo`050?oOW018000000000000M02oool02P3oool00`000000oooo0?ooo`050?ooo`040000
00005Ol001Go000000<000So00D0000000@0o`040?l0103o000000030180o`050000000P0?l0803o
0200o`0000000`0^0?l010000000?03o03`0o`00000304/0o`050000001I0?l0F@3o05T0o`000000
0`1X0?l01@000000MP3o07H0o`1f0?l0000000<0Q@3o00@0000009@0o`2D0?l000000`2R0?l01@00
0000/@3o0;40o`2a0?l0000000<0`03o00D000000<l0o`3?0?l0c`3o000000030=d0o`040000003/
0?l0k03o000000<0nP3o00D000000?l0m`3o0?L0o`3g000000030?l0j@050000003o0=/0o`3K0?l0
f`0000000`3o0<d010000000o`300?l0`00000030?l0/P050000003o0:H0o`2V0?l0YP0000000`3o
09T01@000000o`2=0?l0S@3o08d0000000<0o`2100@000000?l0MP3o07H000000`3o06/01@000000
o`1P0?l0H03o0600000000<0o`1F00D000000?l0C03o04`0o`1<000000030?l0@`040000003o03X0
o`0j000000<0o`0b00D000000?l0:`3o02/0o`0[000000030?l090050000003o01d0o`0M0?l07@00
00000P3o01P01@000000o`0B0?l04P3o0180000000<0o`0>00D000000?l02P3o00X0o`0:00000003
0?l01`050000003o00@0o`040?l0100000000P3o00801@000000o`010?l00@3o0040000000<0o`00
00D000000?l0003o0000o`00000000030?l00@050000003o0080o`020?l00P0000000P3o00@01@00
0000o`070?l01`3o00L0000000<0o`0:00D000000?l03P3o00h0o`0>000000030?l04P050000003o
01P0o`0H0?l0600000000P3o01d01@000000o`0T0?l0903o02@0000000<0o`0[00D000000?l0<P3o
0380o`0b000000030?l0>P050000003o04<0o`130?l0@`0000000P3o04`01@000000o`1F0?l0EP3o
05H0000000<0o`1P00D000000?l0J`3o06/0o`1[000000030?l0MP050000003o0840o`210?l0P@00
00000P3o08d01@000000o`2I0?l0V@3o09T0000000<0o`2V00D000000?l0/P3o0;80o`2b00000003
0?l0`0050000003o0<d0o`3=0?l0c@0000000P3o0=/01@000000o`3Y0?l0j@3o0>T0000000<0o`3g
00D000000?X0o`3j0?l0nP3o000000030>`0o`050000003M0?l0g@3o0=d0o`0000000P3?0?l01@00
0000`03o0<00o`300?l0000000<0/@3o00D000000:80o`2R0?l0XP3o0000000309@0o`0500000025
0?l0Q@3o08D0o`0000000P1f0?l01@000000J03o06P0o`1X0?l0000000<0F@3o00D0000004/0o`1;
0?l0B`3o0000000303`0o`040000000^0?l0;P3o000000<0803o00D000000180o`0B0?l04P3o0000
000300@0o`05000000002?l000So0008o`0000000`005Ol00`000000oooo0?ooo`040?ooo`400000
0@3oool00100oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00T0oooo00<000000?ooo`3o
ool01@3ooooo0000084000001P3oool100000040oooo000@0?ooo`030;md0000B9`0gooo0080oooo
00<0_g@0001d_`3oool0203oool00`000000oooo0?ooo`050?ooo`040000000C0?l04`3o000000<0
803o00D0000002d0o`0]0?l0;@3o0000000303X0o`05000000170?l0A`3o04L0o`0000000`1E0?l0
10000000HP3o0680o`0000030700o`050000001n0?l0OP3o07h0o`0000000`2<0?l01@000000VP3o
09X0o`2J0?l0000000<0Y`3o00@000000;D0o`2e0?l000000`330?l01@000000d@3o0=40o`3A0?l0
000000<0g`3o00D000000>`0o`3/0?l0k03o000000030?X0o`040000003o0?P0o`3h000000<0o`3[
00D000000?l0gP3o0=h0o`3N000000030?l0d@050000003o0<@0o`340?l0a00000000`3o0;P01000
0000o`2[0?l0Z`0000030?l0W`050000003o09@0o`2D0?l0U00000000`3o08P01@000000o`1m0?l0
O@3o07d0000000<0o`1c00@000000?l0J03o06P000000`3o05l01@000000o`1E0?l0E@3o05D00000
00<0o`1<00D000000?l0@`3o04<0o`13000000030?l0>`040000003o03<0o`0c000000<0o`0/00D0
00000?l09@3o02D0o`0U000000030?l07`050000003o01X0o`0J0?l06P0000000P3o01@01@000000
o`0@0?l0403o0100000000<0o`0<00D000000?l0203o00P0o`08000000030?l01@050000003o00<0
o`030?l00`0000000P3o00401@000000o`000?l0003o0000000000<0o`0000D000000?l0003o0000
o`00000000030?l000050000003o0040o`010?l00@0000000P3o00<01@000000o`050?l01@3o00D0
000000<0o`0800D000000?l0303o00`0o`0<000000030?l040050000003o01@0o`0D0?l050000000
0P3o01X01@000000o`0O0?l07`3o01l0000000<0o`0U00D000000?l0;03o02`0o`0/000000030?l0
<`050000003o03/0o`0k0?l0>`0000000P3o04<01@000000o`1<0?l0C03o04`0000000<0o`1E00D0
00000?l0G`3o05l0o`1O000000030?l0J0050000003o07<0o`1c0?l0L`0000000P3o07d01@000000
o`280?l0R03o08P0000000<0o`2D00D000000?l0W`3o09l0o`2O000000030?l0Z`050000003o0;P0
o`2h0?l0^00000000P3o0<@01@000000o`3A0?l0d@3o0=40000000<0o`3N00D000000?l0j`3o0>/0
o`3[000000030?l0n0050000003j0?l0nP3o0?X0o`0000000P3/0?l01@000000g`3o0=l0o`3O0?l0
000000<0d@3o00D000000<<0o`330?l0``3o000000030;D0o`050000002W0?l0Y`3o0:L0o`000000
0P2J0?l01@000000S03o08`0o`2<0?l0000000<0OP3o00D000000700o`1`0?l0L03o000000030680
o`040000001E0?l0E@3o000000<0A`3o00D0000003X0o`0j0?l0>P3o0000000302d0o`050000000P
0?l0803o0200o`0000000`0C0?l00`000000oooo0?ooo`040?ooo`4000000@3oool000l0oooo00<0
oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`080?ooo`030000003oool0oooo00D0oooo00@0
000001<0o`0C0?l000000`0P0?l01@000000;@3o02d0o`0]0?l0000000<0>P3o00D0000004L0o`17
0?l0A`3o0000000305D0o`040000001R0?l0HP3o000000<0L03o00D0000007h0o`1n0?l0OP3o0000
000308`0o`050000002J0?l0VP3o09X0o`0000000`2W0?l010000000]@3o0;D0o`0000030<<0o`05
0000003A0?l0d@3o0=40o`0000000`3O0?l01@000000k03o0>`0o`3/0?l0000000<0nP3o00@00000
0?l0n03o0?P000000`3o0>/01@000000o`3N0?l0gP3o0=h0000000<0o`3A00D000000?l0a03o0<@0
o`34000000030?l0^0040000003o0:/0o`2[000000<0o`2O00D000000?l0U03o09@0o`2D00000003
0?l0R0050000003o07d0o`1m0?l0O@0000000`3o07<010000000o`1X0?l0J00000030?l0G`050000
003o05D0o`1E0?l0E@0000000`3o04`01@000000o`130?l0@`3o04<0000000<0o`0k00@000000?l0
<`3o03<000000`3o02`01@000000o`0U0?l09@3o02D0000000<0o`0O00D000000?l06P3o01X0o`0J
000000020?l050050000003o0100o`0@0?l0400000000`3o00`01@000000o`080?l0203o00P00000
00<0o`0500D000000?l00`3o00<0o`03000000020?l00@050000003o0000o`000?l0000000000`3o
00001@000000o`000?l0003o0000000000<0o`0000D000000?l00@3o0040o`01000000020?l00`05
0000003o00D0o`050?l01@0000000`3o00P01@000000o`0<0?l0303o00`0000000<0o`0@00D00000
0?l0503o01@0o`0D000000020?l06P050000003o01l0o`0O0?l07`0000000`3o02D01@000000o`0/
0?l0;03o02`0000000<0o`0c00D000000?l0>`3o03/0o`0k000000020?l0@`050000003o04`0o`1<
0?l0C00000000`3o05D01@000000o`1O0?l0G`3o05l0000000<0o`1X00D000000?l0L`3o07<0o`1c
000000020?l0O@050000003o08P0o`280?l0R00000000`3o09@01@000000o`2O0?l0W`3o09l00000
00<0o`2[00D000000?l0^03o0;P0o`2h000000020?l0a0050000003o0=40o`3A0?l0d@0000000`3o
0=h01@000000o`3[0?l0j`3o0>/0000000<0o`3h00D000000?X0o`3j0?l0nP3o000000020>`0o`05
0000003O0?l0g`3o0=l0o`0000000`3A0?l01@000000``3o0<<0o`330?l0000000<0]@3o00D00000
0:L0o`2W0?l0Y`3o0000000209X0o`050000002<0?l0S03o08`0o`0000000`1n0?l01@000000L03o
0700o`1`0?l0000000<0HP3o00@0000005D0o`1E0?l000000`170?l01@000000>P3o03X0o`0j0?l0
000000<0;@3o00D000000200o`0P0?l0803o0000000301<0o`030000003oool0oooo00@0oooo0@00
00010?ooo`003`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00P0oooo00<00000
0?ooo`3oool01@3oool0100000004`3o01<0o`0000030200o`050000000]0?l0;@3o02d0o`000000
0`0j0?l01@000000A`3o04L0o`170?l0000000<0E@3o00@000000680o`1R0?l000000`1`0?l01@00
0000OP3o07h0o`1n0?l0000000<0S03o00D0000009X0o`2J0?l0VP3o000000030:L0o`040000002e
0?l0]@3o000000<0``3o00D000000=40o`3A0?l0d@3o000000030=l0o`050000003/0?l0k03o0>`0
o`0000000`3j0?l010000000o`3h0?l0n00000030?l0j`050000003o0=h0o`3N0?l0gP0000000`3o
0=401@000000o`340?l0a03o0<@0000000<0o`2h00@000000?l0Z`3o0:/000000`3o09l01@000000
o`2D0?l0U03o09@0000000<0o`2800D000000?l0O@3o07d0o`1m000000030?l0L`040000003o06P0
o`1X000000<0o`1O00D000000?l0E@3o05D0o`1E000000030?l0C0050000003o04<0o`130?l0@`00
00000`3o03/010000000o`0c0?l0<`0000030?l0;0050000003o02D0o`0U0?l09@0000000`3o01l0
1@000000o`0J0?l06P3o01X000000080o`0D00D000000?l0403o0100o`0@000000030?l030050000
003o00P0o`080?l0200000000`3o00D01@000000o`030?l00`3o00<000000080o`0100D000000?l0
003o0000o`00000000030?l000050000003o0000o`000?l0000000000`3o00001@000000o`010?l0
0@3o004000000080o`0300D000000?l01@3o00D0o`05000000030?l020050000003o00`0o`0<0?l0
300000000`3o01001@000000o`0D0?l0503o01@000000080o`0J00D000000?l07`3o01l0o`0O0000
00030?l09@050000003o02`0o`0/0?l0;00000000`3o03<01@000000o`0k0?l0>`3o03/000000080
o`1300D000000?l0C03o04`0o`1<000000030?l0E@050000003o05l0o`1O0?l0G`0000000`3o06P0
1@000000o`1c0?l0L`3o07<000000080o`1m00D000000?l0R03o08P0o`28000000030?l0U0050000
003o09l0o`2O0?l0W`0000000`3o0:/01@000000o`2h0?l0^03o0;P000000080o`3400D000000?l0
d@3o0=40o`3A000000030?l0gP050000003o0>/0o`3[0?l0j`0000000`3o0?P01@000000nP3o0?X0
o`3j0?l000000080k03o00D000000=l0o`3O0?l0g`3o000000030=40o`05000000330?l0``3o0<<0
o`0000000`2e0?l01@000000Y`3o0:L0o`2W0?l000000080VP3o00D0000008`0o`2<0?l0S03o0000
000307h0o`050000001`0?l0L03o0700o`0000000`1R0?l010000000E@3o05D0o`00000304L0o`05
0000000j0?l0>P3o03X0o`0000000`0]0?l01@000000803o0200o`0P0?l0000000<04`3o00<00000
0?ooo`3oool0103oool100000040oooo000?0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol0203oool3000000D0ooooo`000021000000@0oooo0`0000010?ooo`003`3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00P0oooo00<000000?ooo`3oool01@3oool010000000
4`3o01<0o`0000030200o`050000000]0?l0;@3o02d0o`0000000`0j0?l01@000000A`3o04L0o`17
0?l0000000<0E@3o00@000000680o`1R0?l000000`1`0?l01@000000OP3o07h0o`1n0?l0000000<0
S03o00D0000009X0o`2J0?l0VP3o000000030:L0o`040000002e0?l0]@3o000000<0``3o00D00000
0=40o`3A0?l0d@3o000000030=l0o`050000003/0?l0k03o0>`0o`0000000`3j0?l010000000o`3h
0?l0n00000030?l0j`050000003o0=h0o`3N0?l0gP0000000`3o0=401@000000o`340?l0a03o0<@0
000000<0o`2h00@000000?l0Z`3o0:/000000`3o09l01@000000o`2D0?l0U03o09@0000000<0o`28
00D000000?l0O@3o07d0o`1m000000030?l0L`040000003o06P0o`1X000000<0o`1O00D000000?l0
E@3o05D0o`1E000000030?l0C0050000003o04<0o`130?l0@`0000000`3o03/010000000o`0c0?l0
<`0000030?l0;0050000003o02D0o`0U0?l09@0000000`3o01l01@000000o`0J0?l06P3o01X00000
0080o`0D00D000000?l0403o0100o`0@000000030?l030050000003o00P0o`080?l0200000000`3o
00D01@000000o`030?l00`3o00<000000080o`0100D000000?l0003o0000o`00000000030?l00005
0000003o0000o`000?l0000000000`3o00001@000000o`010?l00@3o004000000080o`0300D00000
0?l01@3o00D0o`05000000030?l020050000003o00`0o`0<0?l0300000000`3o01001@000000o`0D
0?l0503o01@000000080o`0J00D000000?l07`3o01l0o`0O000000030?l09@050000003o02`0o`0/
0?l0;00000000`3o03<01@000000o`0k0?l0>`3o03/000000080o`1300D000000?l0C03o04`0o`1<
000000030?l0E@050000003o05l0o`1O0?l0G`0000000`3o06P01@000000o`1c0?l0L`3o07<00000
0080o`1m00D000000?l0R03o08P0o`28000000030?l0U0050000003o09l0o`2O0?l0W`0000000`3o
0:/01@000000o`2h0?l0^03o0;P000000080o`3400D000000?l0d@3o0=40o`3A000000030?l0gP05
0000003o0>/0o`3[0?l0j`0000000`3o0?P01@000000nP3o0?X0o`3j0?l000000080k03o00D00000
0=l0o`3O0?l0g`3o000000030=40o`05000000330?l0``3o0<<0o`0000000`2e0?l01@000000Y`3o
0:L0o`2W0?l000000080VP3o00D0000008`0o`2<0?l0S03o0000000307h0o`050000001`0?l0L03o
0700o`0000000`1R0?l010000000E@3o05D0o`00000304L0o`050000000j0?l0>P3o03X0o`000000
0`0]0?l01@000000803o0200o`0P0?l0000000<04`3o00<000000?ooo`3oool0103oool100000040
oooo000?0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0203oool00`000000oooo
0?ooo`050?ooo`040000000C0?l04`3o000000<0803o00D0000002d0o`0]0?l0;@3o0000000303X0
o`05000000170?l0A`3o04L0o`0000000`1E0?l010000000HP3o0680o`0000030700o`050000001n
0?l0OP3o07h0o`0000000`2<0?l01@000000VP3o09X0o`2J0?l0000000<0Y`3o00@000000;D0o`2e
0?l000000`330?l01@000000d@3o0=40o`3A0?l0000000<0g`3o00D000000>`0o`3/0?l0k03o0000
00030?X0o`040000003o0?P0o`3h000000<0o`3[00D000000?l0gP3o0=h0o`3N000000030?l0d@05
0000003o0<@0o`340?l0a00000000`3o0;P010000000o`2[0?l0Z`0000030?l0W`050000003o09@0
o`2D0?l0U00000000`3o08P01@000000o`1m0?l0O@3o07d0000000<0o`1c00@000000?l0J03o06P0
00000`3o05l01@000000o`1E0?l0E@3o05D0000000<0o`1<00D000000?l0@`3o04<0o`1300000003
0?l0>`040000003o03<0o`0c000000<0o`0/00D000000?l09@3o02D0o`0U000000030?l07`050000
003o01X0o`0J0?l06P0000000P3o01@01@000000o`0@0?l0403o0100000000<0o`0<00D000000?l0
203o00P0o`08000000030?l01@050000003o00<0o`030?l00`0000000P3o00401@000000o`000?l0
003o0000000000<0o`0000D000000?l0003o0000o`00000000030?l000050000003o0040o`010?l0
0@0000000P3o00<01@000000o`050?l01@3o00D0000000<0o`0800D000000?l0303o00`0o`0<0000
00030?l040050000003o01@0o`0D0?l0500000000P3o01X01@000000o`0O0?l07`3o01l0000000<0
o`0U00D000000?l0;03o02`0o`0/000000030?l0<`050000003o03/0o`0k0?l0>`0000000P3o04<0
1@000000o`1<0?l0C03o04`0000000<0o`1E00D000000?l0G`3o05l0o`1O000000030?l0J0050000
003o07<0o`1c0?l0L`0000000P3o07d01@000000o`280?l0R03o08P0000000<0o`2D00D000000?l0
W`3o09l0o`2O000000030?l0Z`050000003o0;P0o`2h0?l0^00000000P3o0<@01@000000o`3A0?l0
d@3o0=40000000<0o`3N00D000000?l0j`3o0>/0o`3[000000030?l0n0050000003j0?l0nP3o0?X0
o`0000000P3/0?l01@000000g`3o0=l0o`3O0?l0000000<0d@3o00D000000<<0o`330?l0``3o0000
00030;D0o`050000002W0?l0Y`3o0:L0o`0000000P2J0?l01@000000S03o08`0o`2<0?l0000000<0
OP3o00D000000700o`1`0?l0L03o000000030680o`040000001E0?l0E@3o000000<0A`3o00D00000
03X0o`0j0?l0>P3o0000000302d0o`050000000P0?l0803o0200o`0000000`0C0?l00`000000oooo
0?ooo`040?ooo`4000000@3oool000l0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;oo
o`080?ooo`030000003oool0oooo00D0oooo00@0000001<0o`0C0?l000000`0P0?l01@000000;@3o
02d0o`0]0?l0000000<0>P3o00D0000004L0o`170?l0A`3o0000000305D0o`040000001R0?l0HP3o
000000<0L03o00D0000007h0o`1n0?l0OP3o0000000308`0o`050000002J0?l0VP3o09X0o`000000
0`2W0?l010000000]@3o0;D0o`0000030<<0o`050000003A0?l0d@3o0=40o`0000000`3O0?l01@00
0000k03o0>`0o`3/0?l0000000<0nP3o00@000000?l0n03o0?P000000`3o0>/01@000000o`3N0?l0
gP3o0=h0000000<0o`3A00D000000?l0a03o0<@0o`34000000030?l0^0040000003o0:/0o`2[0000
00<0o`2O00D000000?l0U03o09@0o`2D000000030?l0R0050000003o07d0o`1m0?l0O@0000000`3o
07<010000000o`1X0?l0J00000030?l0G`050000003o05D0o`1E0?l0E@0000000`3o04`01@000000
o`130?l0@`3o04<0000000<0o`0k00@000000?l0<`3o03<000000`3o02`01@000000o`0U0?l09@3o
02D0000000<0o`0O00D000000?l06P3o01X0o`0J000000020?l050050000003o0100o`0@0?l04000
00000`3o00`01@000000o`080?l0203o00P0000000<0o`0500D000000?l00`3o00<0o`0300000002
0?l00@050000003o0000o`000?l0000000000`3o00001@000000o`000?l0003o0000000000<0o`00
00D000000?l00@3o0040o`01000000020?l00`050000003o00D0o`050?l01@0000000`3o00P01@00
0000o`0<0?l0303o00`0000000<0o`0@00D000000?l0503o01@0o`0D000000020?l06P050000003o
01l0o`0O0?l07`0000000`3o02D01@000000o`0/0?l0;03o02`0000000<0o`0c00D000000?l0>`3o
03/0o`0k000000020?l0@`050000003o04`0o`1<0?l0C00000000`3o05D01@000000o`1O0?l0G`3o
05l0000000<0o`1X00D000000?l0L`3o07<0o`1c000000020?l0O@050000003o08P0o`280?l0R000
00000`3o09@01@000000o`2O0?l0W`3o09l0000000<0o`2[00D000000?l0^03o0;P0o`2h00000002
0?l0a0050000003o0=40o`3A0?l0d@0000000`3o0=h01@000000o`3[0?l0j`3o0>/0000000<0o`3h
00D000000?X0o`3j0?l0nP3o000000020>`0o`050000003O0?l0g`3o0=l0o`0000000`3A0?l01@00
0000``3o0<<0o`330?l0000000<0]@3o00D000000:L0o`2W0?l0Y`3o0000000209X0o`050000002<
0?l0S03o08`0o`0000000`1n0?l01@000000L03o0700o`1`0?l0000000<0HP3o00@0000005D0o`1E
0?l000000`170?l01@000000>P3o03X0o`0j0?l0000000<0;@3o00D000000200o`0P0?l0803o0000
000301<0o`030000003oool0oooo00@0oooo0@0000010?ooo`00403oool00`3OW4P004RL0=ooo`02
0?ooo`030;md0000M;l0oooo00P0oooo00<000000?ooo`3oool01@3ooooo0000084000001P3oool1
00000040oooo000@0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`030000003o
ool0oooo00D0oooo00@00000000Eo`005Ol000000`002?l01@000000103o00@0o`040?l0000000<0
4P3o00D000000200o`0P0?l0803o0000000302h0o`040000000l0?l0?03o000000<0B`3o00D00000
05T0o`1I0?l0F@3o0000000306P0o`050000001f0?l0MP3o07H0o`0000000`250?l010000000U03o
09@0o`0000030:80o`050000002a0?l0/@3o0;40o`0000000`300?l01@000000c`3o0<l0o`3?0?l0
000000<0g@3o00@000000>`0o`3/0?l000000`3j0?l01@000000o`3g0?l0m`3o0?L0000000<0o`3Y
00D000000?l0f`3o0=/0o`3K000000030?l0c@040000003o0<00o`30000000<0o`2b00D000000?l0
YP3o0:H0o`2V000000030?l0V@050000003o08d0o`2=0?l0S@0000000`3o084010000000o`1f0?l0
MP0000030?l0J`050000003o0600o`1P0?l0H00000000`3o05H01@000000o`1<0?l0C03o04`00000
00<0o`1300@000000?l0>P3o03X000000`3o03801@000000o`0[0?l0:`3o02/0000000<0o`0T00D0
00000?l07@3o01d0o`0M000000020?l060050000003o0180o`0B0?l04P0000000`3o00h01@000000
o`0:0?l02P3o00X0000000<0o`0700D000000?l0103o00@0o`04000000020?l00P050000003o0040
o`010?l00@0000000`3o00001@000000o`000?l0003o0000000000<0o`0100D000000?l00P3o0080
o`02000000020?l010050000003o00L0o`070?l01`0000000`3o00X01@000000o`0>0?l03P3o00h0
000000<0o`0B00D000000?l0603o01P0o`0H000000020?l07@050000003o02@0o`0T0?l090000000
0`3o02/01@000000o`0b0?l0<P3o0380000000<0o`0j00D000000?l0@`3o04<0o`13000000020?l0
C0050000003o05H0o`1F0?l0EP0000000`3o06001@000000o`1[0?l0J`3o06/0000000<0o`1f00D0
00000?l0P@3o0840o`21000000020?l0S@050000003o09T0o`2I0?l0V@0000000`3o0:H01@000000
o`2b0?l0/P3o0;80000000<0o`3000D000000?l0c@3o0<d0o`3=000000020?l0f`050000003o0>T0
o`3Y0?l0j@0000000`3o0?L01@000000nP3o0?X0o`3j0?l0000000<0k03o00D000000=d0o`3M0?l0
g@3o000000020<l0o`05000000300?l0`03o0<00o`0000000`2a0?l01@000000XP3o0:80o`2R0?l0
000000<0U03o00D0000008D0o`250?l0Q@3o0000000207H0o`050000001X0?l0J03o06P0o`000000
0`1I0?l01@000000B`3o04/0o`1;0?l0000000<0?03o00@0000002h0o`0^0?l000000`0P0?l01@00
00004P3o0180o`0B0?l0000000<0103o00D000000008o`002?l000So00000003000Eo`030000003o
ool0oooo00@0oooo0@0000010?ooo`004@3oool01@3ookl0M0000000000007@0_ooo00X0oooo00<0
00000?ooo`3oool01@3oool01000000001Go000Eo`0000030008o`05000000040?l0103o00@0o`00
00000`0B0?l01@000000803o0200o`0P0?l0000000<0;P3o00@0000003`0o`0l0?l000000`1;0?l0
1@000000F@3o05T0o`1I0?l0000000<0J03o00D0000007H0o`1f0?l0MP3o0000000308D0o`040000
002D0?l0U03o000000<0XP3o00D000000;40o`2a0?l0/@3o000000030<00o`050000003?0?l0c`3o
0<l0o`0000000`3M0?l010000000k03o0>`0o`0000030?X0o`050000003o0?L0o`3g0?l0m`000000
0`3o0>T01@000000o`3K0?l0f`3o0=/0000000<0o`3=00@000000?l0`03o0<0000000`3o0;801@00
0000o`2V0?l0YP3o0:H0000000<0o`2I00D000000?l0S@3o08d0o`2=000000030?l0P@040000003o
07H0o`1f000000<0o`1[00D000000?l0H03o0600o`1P000000030?l0EP050000003o04`0o`1<0?l0
C00000000`3o04<010000000o`0j0?l0>P0000030?l0<P050000003o02/0o`0[0?l0:`0000000`3o
02@01@000000o`0M0?l07@3o01d000000080o`0H00D000000?l04P3o0180o`0B000000030?l03P05
0000003o00X0o`0:0?l02P0000000`3o00L01@000000o`040?l0103o00@000000080o`0200D00000
0?l00@3o0040o`01000000030?l000050000003o0000o`000?l0000000000`3o00401@000000o`02
0?l00P3o008000000080o`0400D000000?l01`3o00L0o`07000000030?l02P050000003o00h0o`0>
0?l03P0000000`3o01801@000000o`0H0?l0603o01P000000080o`0M00D000000?l0903o02@0o`0T
000000030?l0:`050000003o0380o`0b0?l0<P0000000`3o03X01@000000o`130?l0@`3o04<00000
0080o`1<00D000000?l0EP3o05H0o`1F000000030?l0H0050000003o06/0o`1[0?l0J`0000000`3o
07H01@000000o`210?l0P@3o084000000080o`2=00D000000?l0V@3o09T0o`2I000000030?l0YP05
0000003o0;80o`2b0?l0/P0000000`3o0<001@000000o`3=0?l0c@3o0<d000000080o`3K00D00000
0?l0j@3o0>T0o`3Y000000030?l0m`050000003j0?l0nP3o0?X0o`0000000`3/0?l01@000000g@3o
0=d0o`3M0?l000000080c`3o00D000000<00o`300?l0`03o000000030;40o`050000002R0?l0XP3o
0:80o`0000000`2D0?l01@000000Q@3o08D0o`250?l000000080MP3o00D0000006P0o`1X0?l0J03o
0000000305T0o`050000001;0?l0B`3o04/0o`0000000`0l0?l010000000;P3o02h0o`0000030200
o`050000000B0?l04P3o0180o`0000000`040?l01@00000000So0008o`002?l0000000<001Go00<0
00000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000
000Eo`005Ol000000`002?l01@000000103o00@0o`040?l0000000<04P3o00D000000200o`0P0?l0
803o0000000302h0o`040000000l0?l0?03o000000<0B`3o00D0000005T0o`1I0?l0F@3o00000003
06P0o`050000001f0?l0MP3o07H0o`0000000`250?l010000000U03o09@0o`0000030:80o`050000
002a0?l0/@3o0;40o`0000000`300?l01@000000c`3o0<l0o`3?0?l0000000<0g@3o00@000000>`0
o`3/0?l000000`3j0?l01@000000o`3g0?l0m`3o0?L0000000<0o`3Y00D000000?l0f`3o0=/0o`3K
000000030?l0c@040000003o0<00o`30000000<0o`2b00D000000?l0YP3o0:H0o`2V000000030?l0
V@050000003o08d0o`2=0?l0S@0000000`3o084010000000o`1f0?l0MP0000030?l0J`050000003o
0600o`1P0?l0H00000000`3o05H01@000000o`1<0?l0C03o04`0000000<0o`1300@000000?l0>P3o
03X000000`3o03801@000000o`0[0?l0:`3o02/0000000<0o`0T00D000000?l07@3o01d0o`0M0000
00020?l060050000003o0180o`0B0?l04P0000000`3o00h01@000000o`0:0?l02P3o00X0000000<0
o`0700D000000?l0103o00@0o`04000000020?l00P050000003o0040o`010?l00@0000000`3o0000
1@000000o`000?l0003o0000000000<0o`0100D000000?l00P3o0080o`02000000020?l010050000
003o00L0o`070?l01`0000000`3o00X01@000000o`0>0?l03P3o00h0000000<0o`0B00D000000?l0
603o01P0o`0H000000020?l07@050000003o02@0o`0T0?l0900000000`3o02/01@000000o`0b0?l0
<P3o0380000000<0o`0j00D000000?l0@`3o04<0o`13000000020?l0C0050000003o05H0o`1F0?l0
EP0000000`3o06001@000000o`1[0?l0J`3o06/0000000<0o`1f00D000000?l0P@3o0840o`210000
00020?l0S@050000003o09T0o`2I0?l0V@0000000`3o0:H01@000000o`2b0?l0/P3o0;80000000<0
o`3000D000000?l0c@3o0<d0o`3=000000020?l0f`050000003o0>T0o`3Y0?l0j@0000000`3o0?L0
1@000000nP3o0?X0o`3j0?l0000000<0k03o00D000000=d0o`3M0?l0g@3o000000020<l0o`050000
00300?l0`03o0<00o`0000000`2a0?l01@000000XP3o0:80o`2R0?l0000000<0U03o00D0000008D0
o`250?l0Q@3o0000000207H0o`050000001X0?l0J03o06P0o`0000000`1I0?l01@000000B`3o04/0
o`1;0?l0000000<0?03o00@0000002h0o`0^0?l000000`0P0?l01@0000004P3o0180o`0B0?l00000
00<0103o00D000000008o`002?l000So00000003000Eo`030000003oool0oooo00@0oooo0@000001
0?ooo`00803oool00`000000oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200
oooo00<000000?ooo`3oool01@3oool01000000006?o001So`000003001Eo`0500000000Aol004Oo
0017o`0000000`00>Ol01@00000002[o000Zo`00:_l0000000<001_o00@00000000<o`003?l00000
0`030?l01@0000004P3o0180o`0B0?l0000000<08P3o00D000000380o`0b0?l0<P3o000000030480
o`040000001C0?l0D`3o000000<0H`3o00D0000007@0o`1d0?l0M03o0000000308D0o`050000002E
0?l0U@3o09D0o`0000000`2V0?l010000000]P3o0;H0o`0000030<L0o`050000003G0?l0e`3o0=L0
o`0000000`3X0?l01@000000n03o0?P0o`3h0?l0000000<0o`3g00@000000?l0i`3o0>L000000`3o
0=P01@000000o`390?l0b@3o0<T0000000<0o`2j00D000000?l0Z`3o0:/0o`2[000000030?l0W@04
0000003o0900o`2@000000<0o`2200D000000?l0MP3o07H0o`1f000000030?l0JP050000003o05h0
o`1N0?l0GP0000000`3o05<010000000o`180?l0B00000030?l0?`050000003o03D0o`0e0?l0=@00
00000`3o02d01@000000o`0U0?l09@3o02D000000080o`0N00D000000?l0603o01P0o`0H00000003
0?l04P050000003o00d0o`0=0?l03@0000000`3o00T01@000000o`060?l01P3o00H000000080o`04
00D000000?l00P3o0080o`02000000030?l00@050000003o0040o`010?l00@0000000`3o00801@00
0000o`040?l0103o00@000000080o`0600D000000?l02@3o00T0o`09000000030?l03@050000003o
0180o`0B0?l04P0000000`3o01P01@000000o`0N0?l07P3o01h000000080o`0U00D000000?l0;@3o
02d0o`0]000000030?l0=@050000003o03l0o`0o0?l0?`0000000`3o04P01@000000o`1C0?l0D`3o
05<000000080o`1N00D000000?l0JP3o06X0o`1Z000000030?l0MP050000003o0880o`220?l0PP00
00000`3o09001@000000o`2M0?l0W@3o09d000000080o`2[00D000000?l0^P3o0;X0o`2j00000003
0?l0b@050000003o0=P0o`3H0?l0f00000000`3o0>L01@000000o`3g0?l0m`3o0?L000000080n03o
00D000000>P0o`3X0?l0j03o000000030=L0o`05000000370?l0a`3o0<L0o`0000000`2f0?l01@00
0000YP3o0:H0o`2V0?l000000080U@3o00D0000008D0o`250?l0Q@3o0000000307@0o`050000001S
0?l0H`3o06<0o`0000000`1C0?l01@000000@P3o0480o`120?l000000080<P3o00D000000280o`0R
0?l08P3o000000030180o`05000000030?l00`3o00<0o`0000000`003?l01000000001_o000Ko`00
0003000Zo`0500000000>Ol003Wo000io`0000000`00Aol01@00000005Go001Eo`00EOl0000000<0
06?o00<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo
00@00000001So`00Hol000000`00EOl01@00000004Oo0017o`00Aol0000000<003Wo00D00000000Z
o`00:_l002[o00000003000Ko`04000000003?l000co000000<00`3o00D000000180o`0B0?l04P3o
000000030280o`050000000b0?l0<P3o0380o`0000000`120?l010000000D`3o05<0o`00000306<0
o`050000001d0?l0M03o07@0o`0000000`250?l01@000000U@3o09D0o`2E0?l0000000<0YP3o00@0
00000;H0o`2f0?l000000`370?l01@000000e`3o0=L0o`3G0?l0000000<0j03o00D000000?P0o`3h
0?l0n03o000000030?l0m`040000003o0>L0o`3W000000<0o`3H00D000000?l0b@3o0<T0o`390000
00030?l0^P050000003o0:/0o`2[0?l0Z`0000000`3o09d010000000o`2@0?l0T00000030?l0PP05
0000003o07H0o`1f0?l0MP0000000`3o06X01@000000o`1N0?l0GP3o05h0000000<0o`1C00@00000
0?l0B03o04P000000`3o03l01@000000o`0e0?l0=@3o03D0000000<0o`0]00D000000?l09@3o02D0
o`0U000000020?l07P050000003o01P0o`0H0?l0600000000`3o01801@000000o`0=0?l03@3o00d0
000000<0o`0900D000000?l01P3o00H0o`06000000020?l010050000003o0080o`020?l00P000000
0`3o00401@000000o`010?l00@3o0040000000<0o`0200D000000?l0103o00@0o`04000000020?l0
1P050000003o00T0o`090?l02@0000000`3o00d01@000000o`0B0?l04P3o0180000000<0o`0H00D0
00000?l07P3o01h0o`0N000000020?l09@050000003o02d0o`0]0?l0;@0000000`3o03D01@000000
o`0o0?l0?`3o03l0000000<0o`1800D000000?l0D`3o05<0o`1C000000020?l0GP050000003o06X0
o`1Z0?l0JP0000000`3o07H01@000000o`220?l0PP3o0880000000<0o`2@00D000000?l0W@3o09d0
o`2M000000020?l0Z`050000003o0;X0o`2j0?l0^P0000000`3o0<T01@000000o`3H0?l0f03o0=P0
000000<0o`3W00D000000?l0m`3o0?L0o`3g000000020?P0o`050000003X0?l0j03o0>P0o`000000
0`3G0?l01@000000a`3o0<L0o`370?l0000000<0]P3o00D000000:H0o`2V0?l0YP3o0000000209D0
o`05000000250?l0Q@3o08D0o`0000000`1d0?l01@000000H`3o06<0o`1S0?l0000000<0D`3o00D0
00000480o`120?l0@P3o000000020380o`050000000R0?l08P3o0280o`0000000`0B0?l01@000000
0`3o00<0o`030?l0000000<000co00@00000000Ko`006ol000000`00:_l01@00000003Wo000io`00
>Ol0000000<004Oo00D00000001Eo`00EOl005Go00000003001So`030000003oool0oooo00@0oooo
0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000Hol006?o000000<005Go
00D000000017o`00Aol004Oo00000003000io`0500000000:_l002[o000Zo`0000000`006ol01000
000000co000<o`00000300<0o`050000000B0?l04P3o0180o`0000000`0R0?l01@000000<P3o0380
o`0b0?l0000000<0@P3o00@0000005<0o`1C0?l000000`1S0?l01@000000M03o07@0o`1d0?l00000
00<0Q@3o00D0000009D0o`2E0?l0U@3o000000030:H0o`040000002f0?l0]P3o000000<0a`3o00D0
00000=L0o`3G0?l0e`3o000000030>P0o`050000003h0?l0n03o0?P0o`0000000`3o0?L010000000
o`3W0?l0i`0000030?l0f0050000003o0<T0o`390?l0b@0000000`3o0;X01@000000o`2[0?l0Z`3o
0:/0000000<0o`2M00@000000?l0T03o090000000`3o08801@000000o`1f0?l0MP3o07H0000000<0
o`1Z00D000000?l0GP3o05h0o`1N000000030?l0D`040000003o04P0o`18000000<0o`0o00D00000
0?l0=@3o03D0o`0e000000030?l0;@050000003o02D0o`0U0?l09@0000000P3o01h01@000000o`0H
0?l0603o01P0000000<0o`0B00D000000?l03@3o00d0o`0=000000030?l02@050000003o00H0o`06
0?l01P0000000P3o00@01@000000o`020?l00P3o0080000000<0o`0100D000000?l00@3o0040o`01
000000030?l00P050000003o00@0o`040?l0100000000P3o00H01@000000o`090?l02@3o00T00000
00<0o`0=00D000000?l04P3o0180o`0B000000030?l060050000003o01h0o`0N0?l07P0000000P3o
02D01@000000o`0]0?l0;@3o02d0000000<0o`0e00D000000?l0?`3o03l0o`0o000000030?l0B005
0000003o05<0o`1C0?l0D`0000000P3o05h01@000000o`1Z0?l0JP3o06X0000000<0o`1f00D00000
0?l0PP3o0880o`22000000030?l0T0050000003o09d0o`2M0?l0W@0000000P3o0:/01@000000o`2j
0?l0^P3o0;X0000000<0o`3900D000000?l0f03o0=P0o`3H000000030?l0i`050000003o0?L0o`3g
0?l0m`0000000P3h0?l01@000000j03o0>P0o`3X0?l0000000<0e`3o00D000000<L0o`370?l0a`3o
000000030;H0o`050000002V0?l0YP3o0:H0o`0000000P2E0?l01@000000Q@3o08D0o`250?l00000
00<0M03o00D0000006<0o`1S0?l0H`3o0000000305<0o`05000000120?l0@P3o0480o`0000000P0b
0?l01@0000008P3o0280o`0R0?l0000000<04P3o00D0000000<0o`030?l00`3o00000003000<o`04
000000006ol001_o000000<002[o00D00000000io`00>Ol003Wo000000030017o`0500000000EOl0
05Go001Eo`0000000`00Hol00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<00000
0?ooo`3oool01@3ooooo0000084000001P3oool100000040oooo000P0?ooo`030000003oool0oooo
00D0oooo00@00000003=o`00cOl000000`00_ol01@0000000;3o002`o`00/?l0000000<00:7o00D0
0000002Bo`00T_l009;o000000030022o`0400000000L_l007;o000000<0067o00D00000001@o`00
D?l0053o00000003000oo`0500000000;Ol002go000]o`0000000`006ol01000000000So0008o`00
000300T0o`050000000L0?l0703o01`0o`0000000`0_0?l01@000000@`3o04<0o`130?l0000000<0
EP3o00@0000006X0o`1Z0?l000000`1m0?l01@000000T@3o0940o`2A0?l0000000<0Y03o00D00000
0;L0o`2g0?l0]`3o000000030</0o`040000003N0?l0gP3o000000<0l03o00D000000?l0o03o0?`0
o`3l000000030?l0jP050000003o0=P0o`3H0?l0f00000000`3o0<L010000000o`2f0?l0]P000003
0?l0YP050000003o09H0o`2F0?l0UP0000000`3o08L01@000000o`1h0?l0N03o07P0000000<0o`1Z
00@000000?l0G@3o05d000000`3o05401@000000o`150?l0A@3o04D0000000<0o`0j00D000000?l0
<@3o0340o`0a000000020?l0:0050000003o0200o`0P0?l0800000000`3o01T01@000000o`0B0?l0
4P3o0180000000<0o`0=00D000000?l02@3o00T0o`09000000020?l01P050000003o00@0o`040?l0
100000000`3o00<01@000000o`030?l00`3o00<0000000<0o`0400D000000?l01P3o00H0o`060000
00020?l02@050000003o00d0o`0=0?l03@0000000`3o01801@000000o`0I0?l06@3o01T0000000<0
o`0P00D000000?l0:03o02P0o`0X000000020?l0<@050000003o03X0o`0j0?l0>P0000000`3o04D0
1@000000o`1A0?l0D@3o0540000000<0o`1M00D000000?l0JP3o06X0o`1Z000000020?l0N0050000
003o08L0o`270?l0Q`0000000`3o09H01@000000o`2V0?l0YP3o0:H0000000<0o`2f00D000000?l0
a`3o0<L0o`37000000020?l0f0050000003o0>X0o`3Z0?l0jP0000000`3o0?`01@000000l03o0?00
o`3`0?l0000000<0gP3o00D000000</0o`3;0?l0b`3o000000020;L0o`050000002T0?l0Y03o0:@0
o`0000000`2A0?l01@000000O@3o07d0o`1m0?l0000000<0JP3o00D0000005H0o`1F0?l0EP3o0000
000204<0o`050000000_0?l0;`3o02l0o`0000000`0L0?l01@0000002@3o00T0o`090?l0000000<0
00So00D00000000Ko`006ol001_o00000002000]o`0500000000?ol003oo000oo`0000000`00D?l0
1@000000067o001Qo`00HOl0000000<007;o00@000000022o`00P_l000000`00T_l01@0000000:7o
002Qo`00XOl0000000<00;3o00D00000002oo`00_ol00;oo00000003003=o`030000003oool0oooo
00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000cOl00<go0000
00<00;oo00D00000002`o`00/?l00;3o00000003002Qo`0500000000T_l009;o002Bo`0000000`00
P_l01000000007;o001bo`000003001Qo`0500000000D?l0053o001@o`0000000`00?ol01@000000
02go000]o`00;Ol0000000<001_o00@000000008o`002?l000000`090?l01@000000703o01`0o`0L
0?l0000000<0;`3o00D0000004<0o`130?l0@`3o0000000305H0o`040000001Z0?l0JP3o000000<0
O@3o00D000000940o`2A0?l0T@3o000000030:@0o`050000002g0?l0]`3o0;L0o`0000000`3;0?l0
10000000gP3o0=h0o`0000030?00o`050000003o0?`0o`3l0?l0o00000000`3o0>X01@000000o`3H
0?l0f03o0=P0000000<0o`3700@000000?l0]P3o0;H000000`3o0:H01@000000o`2F0?l0UP3o09H0
000000<0o`2700D000000?l0N03o07P0o`1h000000030?l0JP040000003o05d0o`1M000000<0o`1A
00D000000?l0A@3o04D0o`15000000030?l0>P050000003o0340o`0a0?l0<@0000000P3o02P01@00
0000o`0P0?l0803o0200000000<0o`0I00D000000?l04P3o0180o`0B000000030?l03@050000003o
00T0o`090?l02@0000000P3o00H01@000000o`040?l0103o00@0000000<0o`0300D000000?l00`3o
00<0o`03000000030?l010050000003o00H0o`060?l01P0000000P3o00T01@000000o`0=0?l03@3o
00d0000000<0o`0B00D000000?l06@3o01T0o`0I000000030?l080050000003o02P0o`0X0?l0:000
00000P3o03401@000000o`0j0?l0>P3o03X0000000<0o`1500D000000?l0D@3o0540o`1A00000003
0?l0G@050000003o06X0o`1Z0?l0JP0000000P3o07P01@000000o`270?l0Q`3o08L0000000<0o`2F
00D000000?l0YP3o0:H0o`2V000000030?l0]P050000003o0<L0o`370?l0a`0000000P3o0=P01@00
0000o`3Z0?l0jP3o0>X0000000<0o`3l00D000000?00o`3`0?l0l03o000000030=h0o`050000003;
0?l0b`3o0</0o`0000000P2g0?l01@000000Y03o0:@0o`2T0?l0000000<0T@3o00D0000007d0o`1m
0?l0O@3o0000000306X0o`050000001F0?l0EP3o05H0o`0000000P130?l01@000000;`3o02l0o`0_
0?l0000000<0703o00D0000000T0o`090?l02@3o000000030008o`05000000006ol001_o000Ko`00
00000P00;Ol01@00000003oo000oo`00?ol0000000<0053o00D00000001Qo`00HOl0067o00000003
001bo`0400000000P_l008;o000000<009;o00D00000002Qo`00XOl00:7o00000003002`o`050000
0000_ol00;oo002oo`0000000`00cOl00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo
00<000000?ooo`3oool01@3oool0100000000<go003=o`000003002oo`0500000000/?l00;3o002`
o`0000000`00XOl01@00000009;o002Bo`00T_l0000000<008;o00@00000001bo`00L_l000000`00
HOl01@000000053o001@o`00D?l0000000<003oo00D00000000]o`00;Ol002go00000003000Ko`04
000000002?l000So000000<02@3o00D0000001`0o`0L0?l0703o0000000302l0o`05000000130?l0
@`3o04<0o`0000000`1F0?l010000000JP3o06X0o`00000307d0o`050000002A0?l0T@3o0940o`00
00000`2T0?l01@000000]`3o0;L0o`2g0?l0000000<0b`3o00@000000=h0o`3N0?l000000`3`0?l0
1@000000o`3l0?l0o03o0?`0000000<0o`3Z00D000000?l0f03o0=P0o`3H000000030?l0a`040000
003o0;H0o`2f000000<0o`2V00D000000?l0UP3o09H0o`2F000000030?l0Q`050000003o07P0o`1h
0?l0N00000000`3o06X010000000o`1M0?l0G@0000030?l0D@050000003o04D0o`150?l0A@000000
0`3o03X01@000000o`0a0?l0<@3o034000000080o`0X00D000000?l0803o0200o`0P000000030?l0
6@050000003o0180o`0B0?l04P0000000`3o00d01@000000o`090?l02@3o00T000000080o`0600D0
00000?l0103o00@0o`04000000030?l00`050000003o00<0o`030?l00`0000000`3o00@01@000000
o`060?l01P3o00H000000080o`0900D000000?l03@3o00d0o`0=000000030?l04P050000003o01T0
o`0I0?l06@0000000`3o02001@000000o`0X0?l0:03o02P000000080o`0a00D000000?l0>P3o03X0
o`0j000000030?l0A@050000003o0540o`1A0?l0D@0000000`3o05d01@000000o`1Z0?l0JP3o06X0
00000080o`1h00D000000?l0Q`3o08L0o`27000000030?l0UP050000003o0:H0o`2V0?l0YP000000
0`3o0;H01@000000o`370?l0a`3o0<L000000080o`3H00D000000?l0jP3o0>X0o`3Z000000030?l0
o0050000003`0?l0l03o0?00o`0000000`3N0?l01@000000b`3o0</0o`3;0?l000000080]`3o00D0
00000:@0o`2T0?l0Y03o000000030940o`050000001m0?l0O@3o07d0o`0000000`1Z0?l01@000000
EP3o05H0o`1F0?l000000080@`3o00D0000002l0o`0_0?l0;`3o0000000301`0o`05000000090?l0
2@3o00T0o`0000000`002?l01@00000001_o000Ko`006ol00000008002go00D00000000oo`00?ol0
03oo00000003001@o`0500000000HOl0067o001Qo`0000000`00L_l01000000008;o0022o`000003
002Bo`0500000000XOl00:7o002Qo`0000000`00/?l01@0000000;oo002oo`00_ol0000000<00<go
00<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0ooooo`00
0021000000H0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000okP0
0?nh000000<00?o500D00000003od`00om<00?oC00000003003ohP0500000000oo400?oa003ol@00
00000`00oOl0100000000>co003/o`000003003Ko`0500000000bOl00<Wo0039o`0000000`00]_l0
1@0000000:;o002Ro`00X_l0000000<008oo00@00000001jo`00N_l000000`00IOl01@000000053o
001@o`00D?l0000000<003[o00D00000000To`009?l002Co00000003000=o`04000000080?l0203o
000000<07`3o00D0000003H0o`0f0?l0=P3o0000000304h0o`050000001U0?l0I@3o06D0o`000000
0`1l0?l010000000T`3o09<0o`0000030:X0o`05000000300?l0`03o0<00o`0000000`3G0?l01@00
0000k03o0>`0o`3/0?l0000000<0o`3m00@000000?l0j03o0>P000000`3o0=@01@000000o`300?l0
`03o0<00000000<0o`2]00D000000?l0V`3o09/0o`2K000000030?l0R@040000003o07P0o`1h0000
00<0o`1Y00D000000?l0FP3o05X0o`1J000000030?l0C0050000003o0400o`100?l0@00000000P3o
03@01@000000o`0Z0?l0:P3o02X0000000<0o`0Q00D000000?l06@3o01T0o`0I000000030?l04`05
0000003o00d0o`0=0?l03@0000000P3o00T01@000000o`070?l01`3o00L0000000<0o`0500D00000
0?l01@3o00D0o`05000000030?l01`050000003o00T0o`090?l02@0000000P3o00d01@000000o`0C
0?l04`3o01<0000000<0o`0I00D000000?l08@3o0240o`0Q000000030?l0:P050000003o03@0o`0d
0?l0=00000000P3o04001@000000o`1<0?l0C03o04`0000000<0o`1J00D000000?l0J@3o06T0o`1Y
000000030?l0N0050000003o08T0o`290?l0R@0000000P3o09/01@000000o`2]0?l0[@3o0:d00000
00<0o`3000D000000?l0e03o0=@0o`3D000000030?l0j0050000003o0?d0o`3m0?l0o@0000000P3/
0?l01@000000e`3o0=L0o`3G0?l0000000<0`03o00D000000:X0o`2Z0?l0ZP3o0000000309<0o`05
0000001l0?l0O03o07`0o`0000000P1U0?l01@000000CP3o04h0o`1>0?l0000000<0=P3o00D00000
01l0o`0O0?l07`3o0000000300P0o`05000000003Ol000go000=o`0000000P009?l01@00000003[o
000jo`00>_l0000000<0053o00D00000001Uo`00IOl006Go00000003001jo`0500000000Sol008oo
002?o`0000000P00X_l01@0000000;Ko002fo`00]_l0000000<00<Wo00D00000003Ko`00fol00=_o
00000003003/o`0400000000oOl00?go000000<00?oa00D00000003ohP00on800?oR00000003003o
d`0500000000olD00?o5003oa@0000000`00okP00`000000oooo0?ooo`040?ooo`4000000@3oool0
0200oooo00<000000?ooo`3oool01@3oool0100000000?nh003o^0000003003oa@0500000000om<0
0?oC003od`0000000`00on801@0000000?oa003ol@00oo40000000<00?go00@00000003/o`00k?l0
00000`00fol01@0000000<Wo0039o`00bOl0000000<00;Ko00D00000002Ro`00X_l00:;o00000003
002?o`0400000000N_l007[o000000<006Go00D00000001@o`00D?l0053o00000003000jo`050000
00009?l002Co000To`0000000`003Ol010000000203o00P0o`00000301l0o`050000000f0?l0=P3o
03H0o`0000000`1>0?l01@000000I@3o06D0o`1U0?l0000000<0O03o00@0000009<0o`2C0?l00000
0`2Z0?l01@000000`03o0<00o`300?l0000000<0e`3o00D000000>`0o`3/0?l0k03o000000030?l0
o@040000003o0>P0o`3X000000<0o`3D00D000000?l0`03o0<00o`30000000030?l0[@050000003o
09/0o`2K0?l0V`0000000`3o08T010000000o`1h0?l0N00000030?l0J@050000003o05X0o`1J0?l0
FP0000000`3o04`01@000000o`100?l0@03o040000000080o`0d00D000000?l0:P3o02X0o`0Z0000
00030?l08@050000003o01T0o`0I0?l06@0000000`3o01<01@000000o`0=0?l03@3o00d000000080
o`0900D000000?l01`3o00L0o`07000000030?l01@050000003o00D0o`050?l01@0000000`3o00L0
1@000000o`090?l02@3o00T000000080o`0=00D000000?l04`3o01<0o`0C000000030?l06@050000
003o0240o`0Q0?l08@0000000`3o02X01@000000o`0d0?l0=03o03@000000080o`1000D000000?l0
C03o04`0o`1<000000030?l0FP050000003o06T0o`1Y0?l0J@0000000`3o07P01@000000o`290?l0
R@3o08T000000080o`2K00D000000?l0[@3o0:d0o`2]000000030?l0`0050000003o0=@0o`3D0?l0
e00000000`3o0>P01@000000o`3m0?l0o@3o0?d000000080k03o00D000000=L0o`3G0?l0e`3o0000
00030<00o`050000002Z0?l0ZP3o0:X0o`0000000`2C0?l01@000000O03o07`0o`1l0?l000000080
I@3o00D0000004h0o`1>0?l0CP3o0000000303H0o`050000000O0?l07`3o01l0o`0000000`080?l0
1@00000000go000=o`003Ol00000008002Co00D00000000jo`00>_l003[o00000003001@o`050000
0000IOl006Go001Uo`0000000`00N_l01@00000008oo002?o`00Sol0000000800:;o00D00000002f
o`00]_l00;Ko000000030039o`0500000000fol00=_o003Ko`0000000`00k?l0100000000?go003m
o`000003003ol@0500000000on800?oR003ohP0000000`00om<01@0000000?o5003oa@00olD00000
00<00?nh00<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0
ooooo`000021000000H0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`040000
0000ocT00?li000000<00?m400D00000003oD@00oe400?mA00000003003oG`0500000000ofd00?m]
003oK@0000000`00ogd0100000000?n>003oSP000003003oW`0500000000ok800?nb003o/P000000
0`00olD01@0000000?oJ003ofP00omX0000000<00?o_00@00000003jo`00n_l000000`00hol01@00
00000<co003<o`00c?l0000000<00;Co00D00000002Ko`00Vol009_o000000030021o`0400000000
Iol006Oo000000<004go00D00000000bo`00<_l003;o00000003000Go`05000000030?l00`3o00<0
o`0000000`0N0?l010000000>P3o03X0o`00000305D0o`050000001`0?l0L03o0700o`0000000`2;
0?l01@000000YP3o0:H0o`2V0?l0000000<0`03o00@000000=X0o`3J0?l000000`3c0?l01@000000
o`3c0?l0l`3o0?<0000000<0o`3L00D000000?l0a@3o0<D0o`35000000030?l0[`040000003o09X0
o`2J000000<0o`2600D000000?l0M03o07@0o`1d000000030?l0H`050000003o05<0o`1C0?l0D`00
00000P3o04@01@000000o`0g0?l0=`3o03L0000000<0o`0/00D000000?l08P3o0280o`0R00000003
0?l06@050000003o0180o`0B0?l04P0000000P3o00d01@000000o`0:0?l02P3o00X0000000<0o`08
00D000000?l0203o00P0o`08000000030?l02P050000003o00d0o`0=0?l03@0000000P3o01801@00
0000o`0I0?l06@3o01T0000000<0o`0R00D000000?l0;03o02`0o`0/000000030?l0=`050000003o
04@0o`140?l0A00000000P3o05<01@000000o`1S0?l0H`3o06<0000000<0o`1d00D000000?l0QP3o
08H0o`26000000030?l0VP050000003o0:l0o`2_0?l0[`0000000P3o0<D01@000000o`3L0?l0g03o
0=`0000000<0o`3c00D000000?<0o`3c0?l0l`3o000000030=X0o`05000000300?l0`03o0<00o`00
00000P2V0?l01@000000R`3o08/0o`2;0?l0000000<0L03o00D0000005D0o`1E0?l0E@3o00000003
03X0o`050000000N0?l07P3o01h0o`0000000P030?l01@00000001Oo000Go`005ol0000000<003;o
00D00000001=o`00COl004go00000003001Wo`0500000000POl0087o0021o`0000000P00Vol01@00
00000;Co002do`00]?l0000000<00<co00D00000003So`00hol00>?o00000003003jo`0500000000
onl00?o_003ok`0000000P00omX01@0000000?o5003oa@00olD0000000<00?nb00D00000003oW`00
oil00?nO00000003003oSP0400000000ogd00?mm000000<00?m]00D00000003oG`00oel00?mO0000
0003003oD@0500000000od@00?m4003oA00000000`00ocT00`000000oooo0?ooo`040?ooo`400000
0@3oool00200oooo00<000000?ooo`3oool01@3oool0100000000?li003o>@000003003oA0050000
0000oe400?mA003oD@0000000`00oel01@0000000?m]003oK@00ofd0000000<00?mm00@00000003o
SP00ohh000000`00oil01@0000000?nb003o/P00ok80000000<00?o500D00000003ofP00omX00?oJ
00000003003ok`0400000000n_l00?[o000000<00>?o00D00000003<o`00c?l00<co00000003002d
o`0500000000Vol009_o002Ko`0000000`00POl01000000006Oo001Wo`000003001=o`0500000000
<_l003;o000bo`0000000`005ol01@0000000`3o00<0o`030?l0000000<07P3o00@0000003X0o`0j
0?l000000`1E0?l01@000000L03o0700o`1`0?l0000000<0R`3o00D000000:H0o`2V0?l0YP3o0000
00030<00o`040000003J0?l0fP3o000000<0l`3o00D000000?l0l`3o0?<0o`3c000000030?l0g005
0000003o0<D0o`350?l0a@0000000`3o0:l010000000o`2J0?l0VP0000030?l0QP050000003o07@0
o`1d0?l0M00000000`3o06<01@000000o`1C0?l0D`3o05<000000080o`1400D000000?l0=`3o03L0
o`0g000000030?l0;0050000003o0280o`0R0?l08P0000000`3o01T01@000000o`0B0?l04P3o0180
00000080o`0=00D000000?l02P3o00X0o`0:000000030?l020050000003o00P0o`080?l020000000
0`3o00X01@000000o`0=0?l03@3o00d000000080o`0B00D000000?l06@3o01T0o`0I000000030?l0
8P050000003o02`0o`0/0?l0;00000000`3o03L01@000000o`140?l0A03o04@000000080o`1C00D0
00000?l0H`3o06<0o`1S000000030?l0M0050000003o08H0o`260?l0QP0000000`3o09X01@000000
o`2_0?l0[`3o0:l000000080o`3500D000000?l0g03o0=`0o`3L000000030?l0l`050000003c0?l0
l`3o0?<0o`0000000`3J0?l01@000000`03o0<00o`300?l000000080YP3o00D0000008/0o`2;0?l0
R`3o000000030700o`050000001E0?l0E@3o05D0o`0000000`0j0?l01@0000007P3o01h0o`0N0?l0
000000800`3o00D00000000Go`005ol001Oo00000003000bo`0500000000COl004go001=o`000000
0`00Iol01@000000087o0021o`00POl00000008009_o00D00000002do`00]?l00;Co00000003003<
o`0500000000hol00>?o003So`0000000`00n_l01@0000000?o_003ok`00onl0000000800?oJ00D0
0000003oa@00olD00?o500000003003o/P0500000000oil00?nO003oW`0000000`00ohh010000000
0?mm003oO@000003003oK@0500000000oel00?mO003oG`0000000`00oe401@0000000?m4003oA000
od@0000000<00?li00<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0
oooo00D0oooo00@00000003o>@00ocT000000`00od@01@0000000?mA003oD@00oe40000000<00?mO
00D00000003oK@00ofd00?m]00000003003oO@0400000000ohh00?n>000000<00?nO00D00000003o
/P00ok800?nb00000003003oa@0500000000omX00?oJ003ofP0000000`00onl0100000000?[o003j
o`000003003So`0500000000c?l00<co003<o`0000000`00]?l01@00000009_o002Ko`00Vol00000
00<0087o00@00000001Wo`00Iol000000`00COl01@00000003;o000bo`00<_l0000000<001Oo00D0
000000<0o`030?l00`3o0000000301h0o`040000000j0?l0>P3o000000<0E@3o00D000000700o`1`
0?l0L03o0000000308/0o`050000002V0?l0YP3o0:H0o`0000000`300?l010000000fP3o0=X0o`00
00030?<0o`050000003o0?<0o`3c0?l0l`0000000`3o0=`01@000000o`350?l0a@3o0<D0000000<0
o`2_00@000000?l0VP3o09X000000`3o08H01@000000o`1d0?l0M03o07@0000000<0o`1S00D00000
0?l0D`3o05<0o`1C000000020?l0A0050000003o03L0o`0g0?l0=`0000000`3o02`01@000000o`0R
0?l08P3o0280000000<0o`0I00D000000?l04P3o0180o`0B000000020?l03@050000003o00X0o`0:
0?l02P0000000`3o00P01@000000o`080?l0203o00P0000000<0o`0:00D000000?l03@3o00d0o`0=
000000020?l04P050000003o01T0o`0I0?l06@0000000`3o02801@000000o`0/0?l0;03o02`00000
00<0o`0g00D000000?l0A03o04@0o`14000000020?l0D`050000003o06<0o`1S0?l0H`0000000`3o
07@01@000000o`260?l0QP3o08H0000000<0o`2J00D000000?l0[`3o0:l0o`2_000000020?l0a@05
0000003o0=`0o`3L0?l0g00000000`3o0?<01@000000l`3o0?<0o`3c0?l0000000<0fP3o00D00000
0<00o`300?l0`03o000000020:H0o`050000002;0?l0R`3o08/0o`0000000`1`0?l01@000000E@3o
05D0o`1E0?l0000000<0>P3o00D0000001h0o`0N0?l07P3o0000000200<0o`05000000005ol001Oo
000Go`0000000`00<_l01@00000004go001=o`00COl0000000<006Oo00D000000021o`00POl0087o
00000002002Ko`0500000000]?l00;Co002do`0000000`00c?l01@0000000>?o003So`00hol00000
00<00?[o00D00000003ok`00onl00?o_00000002003ofP0500000000olD00?o5003oa@0000000`00
ok801@0000000?nO003oW`00oil0000000<00?n>00@00000003oO@00ogd000000`00ofd01@000000
0?mO003oG`00oel0000000<00?mA00D00000003oA000od@00?m400000003003o>@030000003oool0
oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooool00000P@0000060?oo
o`4000000@3oool00200oooo0P0000060?ooo`040000000lo`00??l0000000<0<ol000D0000002[o
000Zo`00:_l00000000301ko00050000000Bo`004_l001;o000000000`04o`00100000000?l9003o
2@000003003o6P0500000000ob/00?l[003o:`0000000`00och01@0000000?mB003oDP00oe800000
00<00?mX00@00000003oOP00ogh000000`00oiH01@0000000?n_003o[`00ojl0000000<00?o900D0
0000003oi000on@00?oT00000003003oo`0400000000hol00>?o000000<00<Go00D00000002Wo`00
Yol00:Oo000000030028o`0500000000JOl006Wo001Yo`0000000`00B_l01000000002[o000Zo`00
0003000:o`050000000E0?l05@3o01D0o`0000000`0f0?l01@000000E@3o05D0o`1E0?l0000000<0
M@3o00@0000009@0o`2D0?l000000`2b0?l01@000000d03o0=00o`3@0?l0000000<0k@3o00D00000
0?l0mP3o0?H0o`3f000000030?l0f`040000003o0<40o`31000000<0o`2Y00D000000?l0TP3o0980
o`2B000000030?l0O@050000003o06T0o`1Y0?l0J@0000000P3o05L01@000000o`160?l0AP3o04H0
000000<0o`0h00D000000?l0;03o02`0o`0/000000030?l08@050000003o01T0o`0I0?l06@000000
0P3o01801@000000o`0>0?l03P3o00h0000000<0o`0<00D000000?l0303o00`0o`0<000000030?l0
3P050000003o0180o`0B0?l04P0000000P3o01T01@000000o`0Q0?l08@3o0240000000<0o`0/00D0
00000?l0>03o03P0o`0h000000030?l0AP050000003o05L0o`1G0?l0E`0000000P3o06T01@000000
o`1m0?l0O@3o07d0000000<0o`2B00D000000?l0Z@3o0:T0o`2Y000000030?l0`@050000003o0=/0
o`3K0?l0f`0000000P3o0?H01@000000k@3o0>d0o`3]0?l0000000<0d03o00D000000;80o`2b0?l0
/P3o0000000309@0o`050000001e0?l0M@3o07D0o`0000000P1E0?l01@000000=P3o03H0o`0f0?l0
000000<05@3o00D00000000:o`002_l000[o00000003000Zo`0500000000B_l004[o001:o`000000
0P00JOl01@00000008So0028o`00R?l0000000<00:Oo00D000000035o`00aOl00<Go00000003003S
o`0500000000ool00?oo003oo`0000000P00on@01@0000000?o9003ob@00olT0000000<00?n_00D0
0000003oUP00oiH00?nF00000003003oOP0500000000ofP00?mX003oJ00000000P00oe801@000000
0?ln003o?P00och0000000<00?l[00D00000003o6P00oaX00?lJ00000003003o2@0400000004o`00
1?l0000000<04_l000D0000001ko000No`007_l00000000302[o00050000000co`00<ol003?o0000
00000`0lo`000`000000oooo0?ooo`030?ooo`8000000@3oool00200oooo00<000000?ooo`3oool0
1@3oool010000000??l003co0000000303?o00050000000Zo`00:_l002[o000000000`0No`001@00
00004_l001;o000Bo`00000000<01?l000@00000003o2@00o`T000000`00oaX01@0000000?l[003o
:`00ob/0000000<00?ln00D00000003oDP00oe800?mB00000003003oJ00400000000ogh00?mn0000
00<00?nF00D00000003o[`00ojl00?n_00000003003ob@0500000000on@00?oT003oi00000000`00
ool0100000000>?o003So`0000030035o`0500000000Yol00:Oo002Wo`0000000`00R?l01@000000
06Wo001Yo`00JOl0000000<004[o00@00000000Zo`00:_l000000`002_l01@0000005@3o01D0o`0E
0?l0000000<0=P3o00D0000005D0o`1E0?l0E@3o0000000307D0o`040000002D0?l0U03o000000<0
/P3o00D000000=00o`3@0?l0d03o000000030>d0o`050000003o0?H0o`3f0?l0mP0000000`3o0=/0
10000000o`310?l0`@0000030?l0Z@050000003o0980o`2B0?l0TP0000000`3o07d01@000000o`1Y
0?l0J@3o06T000000080o`1G00D000000?l0AP3o04H0o`16000000030?l0>0050000003o02`0o`0/
0?l0;00000000`3o02401@000000o`0I0?l06@3o01T000000080o`0B00D000000?l03P3o00h0o`0>
000000030?l030050000003o00`0o`0<0?l0300000000`3o00h01@000000o`0B0?l04P3o01800000
0080o`0I00D000000?l08@3o0240o`0Q000000030?l0;0050000003o03P0o`0h0?l0>00000000`3o
04H01@000000o`1G0?l0E`3o05L000000080o`1Y00D000000?l0O@3o07d0o`1m000000030?l0TP05
0000003o0:T0o`2Y0?l0Z@0000000`3o0<401@000000o`3K0?l0f`3o0=/000000080o`3f00D00000
0>d0o`3]0?l0k@3o000000030=00o`050000002b0?l0/P3o0;80o`0000000`2D0?l01@000000M@3o
07D0o`1e0?l000000080E@3o00D0000003H0o`0f0?l0=P3o0000000301D0o`05000000002_l000[o
000:o`0000000`00:_l01@00000004[o001:o`00B_l00000008006Wo00D000000028o`00R?l008So
00000003002Wo`0500000000aOl00<Go0035o`0000000`00hol01@0000000?oo003oo`00ool00000
00800?oT00D00000003ob@00olT00?o900000003003o[`0500000000oiH00?nF003oUP0000000`00
ogh01@0000000?mX003oJ000ofP0000000800?mB00D00000003o?P00och00?ln00000003003o:`05
00000000oaX00?lJ003o6P0000000`00o`T0100000001?l000Co0000000301;o00050000000No`00
7_l001ko000000000`0Zo`001@000000<ol003?o000co`00000000<0??l000<000000?ooo`3oool0
103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@0000003co000lo`000000
0`0co`001@000000:_l002[o000Zo`00000000<07_l000D0000001;o000Bo`004_l00000000300Co
000400000000o`T00?l9000000<00?lJ00D00000003o:`00ob/00?l[00000003003o?P0500000000
oe800?mB003oDP0000000`00ofP0100000000?mn003oOP000003003oUP0500000000ojl00?n_003o
[`0000000`00olT01@0000000?oT003oi000on@0000000<00?oo00@00000003So`00hol000000`00
aOl01@0000000:Oo002Wo`00Yol0000000<008So00D00000001Yo`00JOl006Wo00000003001:o`04
00000000:_l002[o000000<000[o00D0000001D0o`0E0?l05@3o0000000303H0o`050000001E0?l0
E@3o05D0o`0000000`1e0?l010000000U03o09@0o`0000030;80o`050000003@0?l0d03o0=00o`00
00000`3]0?l01@000000o`3f0?l0mP3o0?H0000000<0o`3K00@000000?l0`@3o0<4000000`3o0:T0
1@000000o`2B0?l0TP3o0980000000<0o`1m00D000000?l0J@3o06T0o`1Y000000020?l0E`050000
003o04H0o`160?l0AP0000000`3o03P01@000000o`0/0?l0;03o02`0000000<0o`0Q00D000000?l0
6@3o01T0o`0I000000020?l04P050000003o00h0o`0>0?l03P0000000`3o00`01@000000o`0<0?l0
303o00`0000000<0o`0>00D000000?l04P3o0180o`0B000000020?l06@050000003o0240o`0Q0?l0
8@0000000`3o02`01@000000o`0h0?l0>03o03P0000000<0o`1600D000000?l0E`3o05L0o`1G0000
00020?l0J@050000003o07d0o`1m0?l0O@0000000`3o09801@000000o`2Y0?l0Z@3o0:T0000000<0
o`3100D000000?l0f`3o0=/0o`3K000000020?l0mP050000003]0?l0k@3o0>d0o`0000000`3@0?l0
1@000000/P3o0;80o`2b0?l0000000<0U03o00D0000007D0o`1e0?l0M@3o0000000205D0o`050000
000f0?l0=P3o03H0o`0000000`0E0?l01@00000000[o000:o`002_l0000000<002[o00D00000001:
o`00B_l004[o00000002001Yo`0500000000R?l008So0028o`0000000`00Yol01@0000000<Go0035
o`00aOl0000000<00>?o00D00000003oo`00ool00?oo00000002003oi00500000000olT00?o9003o
b@0000000`00ojl01@0000000?nF003oUP00oiH0000000<00?mn00D00000003oJ000ofP00?mX0000
0002003oDP0500000000och00?ln003o?P0000000`00ob/01@0000000?lJ003o6P00oaX0000000<0
0?l900@0000000Co0004o`0000000`0Bo`001@0000007_l001ko000No`00000000<0:_l000D00000
03?o000co`00<ol00000000303co00030000003oool0oooo00@0oooo0@0000010?ooo`00803oool0
0`000000oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<000000?oo
o`3oool01@3oool010000000W?l009co0000000309So00050000002Co`00Tol009?o000000000`2<
o`001@000000Q?l008Co0024o`00000000<0N_l000@0000006ko001^o`0000000`1Qo`001@000000
D_l005;o001Bo`00000000<0@_l000D0000002oo000_o`00;ol00000000301_o000400000006o`00
1_l0000000<00?lA00D00000003o:@00obT00?lY00000003003oA00500000000of000?mP003oH000
00000`00ogd0100000000?nK003oV`000003003o^`0500000000om`00?oL003og00000000`00ooh0
1@0000000=ko003No`00g_l0000000<00;_o00@00000002Go`00Uol000000`00L_l01@00000004go
001=o`00COl0000000<002So00D000000003o`000ol000?o000000030240o`04000000150?l0A@3o
000000<0JP3o00D0000008d0o`2=0?l0S@3o000000030;00o`050000003A0?l0d@3o0=40o`000000
0`3b0?l010000000o`3^0?l0kP0000030?l0d@050000003o0;D0o`2e0?l0]@0000000`3o09/01@00
0000o`230?l0P`3o08<000000080o`1/00D000000?l0F03o05P0o`1H000000030?l0AP050000003o
03L0o`0g0?l0=`0000000`3o02X01@000000o`0P0?l0803o020000000080o`0H00D000000?l04P3o
0180o`0B000000030?l040050000003o0100o`0@0?l0400000000`3o01801@000000o`0H0?l0603o
01P000000080o`0P00D000000?l0:P3o02X0o`0Z000000030?l0=`050000003o04H0o`160?l0AP00
00000`3o05P01@000000o`1/0?l0K03o06`000000080o`2300D000000?l0V`3o09/0o`2K00000003
0?l0]@050000003o0=40o`3A0?l0d@0000000`3o0>h01@000000lP3o0?80o`3b0?l000000080d@3o
00D000000;00o`2`0?l0/03o0000000308d0o`050000001Z0?l0JP3o06X0o`0000000`150?l01@00
00008@3o0240o`0Q0?l00000008000?o00D00000000Xo`00:?l002So00000003001=o`0500000000
L_l007;o001bo`0000000`00Uol01@0000000;_o002ko`00^ol0000000800=ko00D00000003ooP00
ooh00?on00000003003og00500000000ok/00?nk003o^`0000000`00oi/01@0000000?mm003oO@00
ogd0000000800?mP00D00000003oA000od@00?m400000003003o:@0500000000oa400?lA003o4@00
00000`06o`001@0000006ol001_o000Ko`0000000080;ol000D0000004;o0012o`00@_l000000003
05;o00050000001Qo`00HOl0067o000000000`1^o`0010000000N_l007[o0000000308Co00050000
002<o`00S?l008co000000000`2Co`001@000000V?l009So002Ho`00000000<0W?l000<000000?oo
o`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@0000009co002L
o`0000000`2Ho`001@000000Tol009?o002Co`00000000<0S?l000D0000008Co0024o`00Q?l00000
000307[o00040000001^o`00K_l0000000<0HOl000D0000005;o001Bo`00D_l00000000304;o0005
0000000_o`00;ol002oo000000000`0Ko`00100000001_l000Ko00000003003o4@0500000000obT0
0?lY003o:@0000000`00od@01@0000000?mP003oH000of00000000<00?mm00@00000003oV`00oi/0
00000`00ok/01@0000000?oL003og000om`0000000<00?on00D00000003No`00g_l00=ko00000003
002ko`0400000000Uol009Oo000000<007;o00D00000001=o`00COl004go00000003000Xo`050000
00000ol000?o0003o`0000000`0Q0?l010000000A@3o04D0o`00000306X0o`050000002=0?l0S@3o
08d0o`0000000`2`0?l01@000000d@3o0=40o`3A0?l0000000<0lP3o00@000000?l0kP3o0>h00000
0`3o0=401@000000o`2e0?l0]@3o0;D0000000<0o`2K00D000000?l0P`3o08<0o`23000000020?l0
K0050000003o05P0o`1H0?l0F00000000`3o04H01@000000o`0g0?l0=`3o03L0000000<0o`0Z00D0
00000?l0803o0200o`0P000000020?l060050000003o0180o`0B0?l04P0000000`3o01001@000000
o`0@0?l0403o0100000000<0o`0B00D000000?l0603o01P0o`0H000000020?l080050000003o02X0
o`0Z0?l0:P0000000`3o03L01@000000o`160?l0AP3o04H0000000<0o`1H00D000000?l0K03o06`0
o`1/000000020?l0P`050000003o09/0o`2K0?l0V`0000000`3o0;D01@000000o`3A0?l0d@3o0=40
000000<0o`3^00D000000?80o`3b0?l0lP3o000000020=40o`050000002`0?l0/03o0;00o`000000
0`2=0?l01@000000JP3o06X0o`1Z0?l0000000<0A@3o00D000000240o`0Q0?l08@3o000000020003
o`0500000000:?l002So000Xo`0000000`00COl01@00000007;o001bo`00L_l0000000<009Oo00D0
0000002ko`00^ol00;_o00000002003No`0500000000ooh00?on003ooP0000000`00om`01@000000
0?nk003o^`00ok/0000000<00?nK00D00000003oO@00ogd00?mm00000002003oH00500000000od@0
0?m4003oA00000000`00obT01@0000000?lA003o4@00oa40000000<01_l000D0000001_o000Ko`00
6ol00000000202oo000500000012o`00@_l004;o000000000`1Bo`001@000000HOl0067o001Qo`00
000000<0K_l000@0000007[o001jo`0000000`24o`001@000000S?l008co002<o`00000000<0Tol0
00D0000009So002Ho`00V?l00000000309co00030000003oool0oooo00@0oooo0@0000010?ooo`00
803oool00`000000oooo0?ooo`050?ooo`040000002Lo`00W?l0000000<0V?l000D0000009?o002C
o`00Tol00000000308co000500000024o`00Q?l008Co000000000`1jo`0010000000K_l006ko0000
0003067o00050000001Bo`00D_l005;o000000000`12o`001@000000;ol002oo000_o`00000000<0
6ol000@0000000Ko0006o`0000000`00oa401@0000000?lY003o:@00obT0000000<00?m400D00000
003oH000of000?mP00000003003oO@0400000000oi/00?nK000000<00?nk00D00000003og000om`0
0?oL00000003003ooP0500000000g_l00=ko003No`0000000`00^ol01000000009Oo002Go`000003
001bo`0500000000COl004go001=o`0000000`00:?l01@00000000?o0003o`000ol0000000<08@3o
00@0000004D0o`150?l000000`1Z0?l01@000000S@3o08d0o`2=0?l0000000<0/03o00D000000=40
o`3A0?l0d@3o000000030?80o`040000003o0>h0o`3^000000<0o`3A00D000000?l0]@3o0;D0o`2e
000000030?l0V`050000003o08<0o`230?l0P`0000000P3o06`01@000000o`1H0?l0F03o05P00000
00<0o`1600D000000?l0=`3o03L0o`0g000000030?l0:P050000003o0200o`0P0?l0800000000P3o
01P01@000000o`0B0?l04P3o0180000000<0o`0@00D000000?l0403o0100o`0@000000030?l04P05
0000003o01P0o`0H0?l0600000000P3o02001@000000o`0Z0?l0:P3o02X0000000<0o`0g00D00000
0?l0AP3o04H0o`16000000030?l0F0050000003o06`0o`1/0?l0K00000000P3o08<01@000000o`2K
0?l0V`3o09/0000000<0o`2e00D000000?l0d@3o0=40o`3A000000030?l0kP050000003b0?l0lP3o
0?80o`0000000P3A0?l01@000000/03o0;00o`2`0?l0000000<0S@3o00D0000006X0o`1Z0?l0JP3o
0000000304D0o`050000000Q0?l08@3o0240o`0000000P000ol01@00000002So000Xo`00:?l00000
00<004go00D00000001bo`00L_l007;o00000003002Go`0500000000^ol00;_o002ko`0000000P00
g_l01@0000000?on003ooP00ooh0000000<00?oL00D00000003o^`00ok/00?nk00000003003oV`05
00000000ogd00?mm003oO@0000000P00of001@0000000?m4003oA000od@0000000<00?lY00D00000
003o4@00oa400?lA0000000300Ko00050000000Ko`006ol001_o000000000P0_o`001@000000@_l0
04;o0012o`00000000<0D_l000D00000067o001Qo`00HOl00000000306ko00040000001jo`00N_l0
000000<0Q?l000D0000008co002<o`00S?l00000000309?o00050000002Ho`00V?l009So00000000
0`2Lo`000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3o
oooo0000084000001P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000
0=co003Lo`0000000`3Oo`001@000000h?l00>3o003Po`00000000<0gol000D000000=go003Mo`00
gOl0000000030=So00040000003Bo`00d_l0000000<0b_l000D000000;oo002oo`00_ol000000003
0;?o00050000002To`00Y?l00:Co000000000`2Co`0010000000POl0087o0000000306co00050000
001Do`00E?l005Co000000000`0ko`001@0000008?l0023o000Po`00000000<00ol000@00000003o
6`00oa/000000`00oc`01@0000000?mN003oGP00oeh0000000<00?n200D00000003oZ000ojP00?nX
00000003003ocP0400000000ooH00?of000000<00>3o00D00000002go`00]ol00;Oo00000003002=
o`0500000000Hol006?o001So`0000000`00>Ol01000000000oo000?o`00000301X0o`0500000013
0?l0@`3o04<0o`0000000`1/0?l01@000000T`3o09<0o`2C0?l0000000<0^P3o00@000000=l0o`3O
0?l000000`3o0?d01@000000o`3L0?l0g03o0=`0000000<0o`2l00D000000?l0W`3o09l0o`2O0000
00020?l0Q@050000003o06`0o`1/0?l0K00000000`3o05L01@000000o`140?l0A03o04@0000000<0
o`0d00D000000?l0:03o02P0o`0X000000020?l07P050000003o01P0o`0H0?l0600000000`3o01@0
1@000000o`0D0?l0503o01@0000000<0o`0H00D000000?l07P3o01h0o`0N000000020?l0:0050000
003o03@0o`0d0?l0=00000000`3o04@01@000000o`1G0?l0E`3o05L0000000<0o`1/00D000000?l0
Q@3o08D0o`25000000020?l0W`050000003o0;`0o`2l0?l0_00000000`3o0=`01@000000o`3m0?l0
o@3o0?d0000000<0g`3o00D000000;X0o`2j0?l0^P3o0000000209<0o`050000001/0?l0K03o06`0
o`0000000`130?l01@0000006P3o01X0o`0J0?l0000000<000oo00D00000000io`00>Ol003Wo0000
0002001So`0500000000SOl008go002=o`0000000`00]ol01@0000000>3o003Po`00h?l0000000<0
0?of00D00000003ocP00olh00?o>00000002003oZ00500000000oh800?n2003oPP0000000`00oeh0
1@0000000?ll003o?000oc`0000000<00?lK00D0000000?o0003o`000ol000000002023o00050000
000ko`00>ol003_o000000000`1Do`001@000000K?l006co001/o`00000000<0POl000D0000009?o
002Co`00Tol0000000020:Co00050000002co`00/ol00;?o000000000`2oo`001@000000b_l00<[o
003:o`00000000<0d_l000@000000=So003Ho`0000000`3Mo`001@000000gol00=oo003Oo`000000
00<0h?l000D000000=oo003Oo`00gol0000000030=co00030000003oool0oooo00@0oooo0@000001
0?ooo`00803oool00`000000oooo0?ooo`050?ooo`040000003Lo`00g?l0000000<0gol000D00000
0>3o003Po`00h?l0000000030=oo00050000003Mo`00gOl00=go000000000`3Ho`0010000000d_l0
0=;o000000030<[o00050000002oo`00_ol00;oo000000000`2co`001@000000Y?l00:Co002To`00
000000<0Tol000@00000087o0021o`0000000`1/o`001@000000E?l005Co001Do`00000000<0>ol0
00D00000023o000Po`008?l00000000300?o000400000000oa/00?lK000000<00?ll00D00000003o
GP00oeh00?mN00000003003oPP0500000000ojP00?nX003oZ00000000`00olh0100000000?of003o
mP000003003Po`0500000000]ol00;Oo002go`0000000`00SOl01@00000006?o001So`00Hol00000
00<003Wo00@00000000?o`003ol000000`0J0?l01@000000@`3o04<0o`130?l0000000<0K03o00D0
000009<0o`2C0?l0T`3o000000030;X0o`040000003O0?l0g`3o000000<0o`3m00D000000?l0g03o
0=`0o`3L000000030?l0_0050000003o09l0o`2O0?l0W`0000000P3o08D01@000000o`1/0?l0K03o
06`0000000<0o`1G00D000000?l0A03o04@0o`14000000030?l0=0050000003o02P0o`0X0?l0:000
00000P3o01h01@000000o`0H0?l0603o01P0000000<0o`0D00D000000?l0503o01@0o`0D00000003
0?l060050000003o01h0o`0N0?l07P0000000P3o02P01@000000o`0d0?l0=03o03@0000000<0o`14
00D000000?l0E`3o05L0o`1G000000030?l0K0050000003o08D0o`250?l0Q@0000000P3o09l01@00
0000o`2l0?l0_03o0;`0000000<0o`3L00D000000?l0o@3o0?d0o`3m000000030=l0o`050000002j
0?l0^P3o0;X0o`0000000P2C0?l01@000000K03o06`0o`1/0?l0000000<0@`3o00D0000001X0o`0J
0?l06P3o00000003000?o`0500000000>Ol003Wo000io`0000000P00Hol01@00000008go002=o`00
SOl0000000<00;Oo00D00000003Po`00h?l00>3o00000003003omP0500000000olh00?o>003ocP00
00000P00ojP01@0000000?n2003oPP00oh80000000<00?mN00D00000003o?000oc`00?ll00000003
003o6`0500000003o`000ol000?o000000000P0Po`001@000000>ol003_o000ko`00000000<0E?l0
00D0000006co001/o`00K?l000000003087o00050000002Co`00Tol009?o000000000P2To`001@00
0000/ol00;?o002co`00000000<0_ol000D000000<[o003:o`00b_l0000000030=;o00040000003H
o`00f?l0000000<0gOl000D000000=oo003Oo`00gol0000000030>3o00050000003Oo`00gol00=oo
000000000`3Lo`000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3o
ool01@3oool010000000g?l00=co000000030=oo00050000003Po`00h?l00>3o000000000`3Oo`00
1@000000gOl00=go003Mo`00000000<0f?l000@000000=;o003Bo`0000000`3:o`001@000000_ol0
0;oo002oo`00000000<0/ol000D000000:Co002To`00Y?l00000000309?o000400000021o`00POl0
000000<0K?l000D0000005Co001Do`00E?l00000000303_o00050000000Po`008?l0023o00000000
0`03o`00100000000?lK003o6`000003003o?00500000000oeh00?mN003oGP0000000`00oh801@00
00000?nX003oZ000ojP0000000<00?o>00@00000003omP00ooH000000`00h?l01@0000000;Oo002g
o`00]ol0000000<008go00D00000001So`00Hol006?o00000003000io`04000000003ol000oo0000
00<06P3o00D0000004<0o`130?l0@`3o0000000306`0o`050000002C0?l0T`3o09<0o`0000000`2j
0?l010000000g`3o0=l0o`0000030?l0o@050000003o0=`0o`3L0?l0g00000000`3o0;`01@000000
o`2O0?l0W`3o09l000000080o`2500D000000?l0K03o06`0o`1/000000030?l0E`050000003o04@0
o`140?l0A00000000`3o03@01@000000o`0X0?l0:03o02P000000080o`0N00D000000?l0603o01P0
o`0H000000030?l050050000003o01@0o`0D0?l0500000000`3o01P01@000000o`0N0?l07P3o01h0
00000080o`0X00D000000?l0=03o03@0o`0d000000030?l0A0050000003o05L0o`1G0?l0E`000000
0`3o06`01@000000o`250?l0Q@3o08D000000080o`2O00D000000?l0_03o0;`0o`2l000000030?l0
g0050000003o0?d0o`3m0?l0o@0000000`3O0?l01@000000^P3o0;X0o`2j0?l000000080T`3o00D0
000006`0o`1/0?l0K03o0000000304<0o`050000000J0?l06P3o01X0o`0000000`003ol01@000000
03Wo000io`00>Ol00000008006?o00D00000002=o`00SOl008go00000003002go`0500000000h?l0
0>3o003Po`0000000`00ooH01@0000000?o>003ocP00olh0000000800?nX00D00000003oPP00oh80
0?n200000003003oGP0500000000oc`00?ll003o?00000000`00oa/01@0000000ol000?o0003o`00
000000808?l000D0000003_o000ko`00>ol00000000305Co00050000001/o`00K?l006co00000000
0`21o`001@000000Tol009?o002Co`0000000080Y?l000D000000;?o002co`00/ol0000000030;oo
00050000003:o`00b_l00<[o000000000`3Bo`0010000000f?l00=So000000030=go00050000003O
o`00gol00=oo000000000`3Po`001@000000gol00=oo003Oo`00000000<0g?l000<000000?ooo`3o
ool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo
0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`040000003go`00mol0000000<0ooh0
00D000000?of003omP00ooH0000000030?o_00050000003ojP00onX00?oZ000000000`3oi`001000
0000onL00?oW000000030?oX00050000003ok000on`00?o/000000000`3ol`001@000000oo/00?ok
003on`00000000<0n?l000@000000>[o003Zo`0000000`3Io`001@000000a_l00<Ko0036o`000000
00<0/?l000D0000009Oo002Go`00Uol00000000307co00040000001No`00G_l0000000<0?_l000D0
000001_o000Ko`006ol000000003003o200500000000obl00?l_003o;`0000000`00oeP010000000
0?n2003oPP000003003o[P0500000000om/00?oK003of`0000000`00m_l01@0000000<So0038o`00
b?l0000000<009So00@00000001Yo`00JOl000000`00>Ol01@00000000[o000:o`002_l0000000<0
8`3o00D000000540o`1A0?l0D@3o0000000307d0o`040000002X0?l0Z03o000000<0d@3o00D00000
0?T0o`3i0?l0n@3o000000030?l0h@050000003o0;l0o`2o0?l0_`0000000P3o09l01@000000o`23
0?l0P`3o08<0000000<0o`1Y00D000000?l0D`3o05<0o`1C000000030?l0@0050000003o0340o`0a
0?l0<@0000000P3o02D01@000000o`0M0?l07@3o01d0000000<0o`0J00D000000?l06P3o01X0o`0J
000000030?l07@050000003o02D0o`0U0?l09@0000000P3o03401@000000o`100?l0@03o04000000
00<0o`1C00D000000?l0J@3o06T0o`1Y000000030?l0P`050000003o09l0o`2O0?l0W`0000000P3o
0;l01@000000o`3Q0?l0h@3o0>40000000<0n@3o00D000000=40o`3A0?l0d@3o000000030:P0o`05
0000001m0?l0O@3o07d0o`0000000P1A0?l01@0000008`3o02<0o`0S0?l0000000<000[o00D00000
000io`00>Ol003Wo00000003001Yo`0500000000V?l009So002Ho`0000000P00b?l01@0000000?Ko
003fo`00m_l0000000<00?oK00D00000003o[P00ojh00?n^00000003003oPP0500000000oeP00?mH
003oF00000000P00obl01@0000000?l8003o2000o`P0000000<06ol000D0000003ko000no`00?_l0
0000000305ko00050000001lo`00O?l007co000000000P2Go`001@000000/?l00;3o002`o`000000
00<0a_l000D000000=Wo003Io`00fOl0000000030>[o00050000003ho`00n?l00?So000000000P3o
n`001@000000oo<00?oc003ol`00000000<0on`000D000000?oX003oj000onP0000000030?oW0004
0000003oi`00onL0000000<0onX000D000000?o_003ok`00onl0000000030?of00050000003ooP00
ooh00?on000000000`3go`000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<00000
0?ooo`3oool01@3oool010000000mol00?Oo000000030?on00050000003omP00ooH00?of00000000
0`3ok`001@000000onX00?oZ003ojP00000000<0onL000@000000?oW003oi`0000000`3oj0001@00
0000on`00?o/003ok000000000<0oo<000D000000?ok003on`00oo/0000000030?So00040000003Z
o`00j_l0000000<0fOl000D000000<Ko0036o`00a_l0000000030;3o00050000002Go`00Uol009Oo
000000000`1lo`0010000000G_l005ko0000000303ko00050000000Ko`006ol001_o000000000`00
o`P01@0000000?l_003o;`00obl0000000<00?mH00@00000003oPP00oh8000000`00ojh01@000000
0?oK003of`00om/0000000<00?Ko00D000000038o`00b?l00<So00000003002Ho`0400000000JOl0
06Wo000000<003Wo00D00000000:o`002_l000[o0000000302<0o`050000001A0?l0D@3o0540o`00
00000`1m0?l010000000Z03o0:P0o`0000030=40o`050000003i0?l0n@3o0?T0o`0000000`3o0>40
1@000000o`2o0?l0_`3o0;l000000080o`2O00D000000?l0P`3o08<0o`23000000030?l0J@050000
003o05<0o`1C0?l0D`0000000`3o04001@000000o`0a0?l0<@3o034000000080o`0U00D000000?l0
7@3o01d0o`0M000000030?l06P050000003o01X0o`0J0?l06P0000000`3o01d01@000000o`0U0?l0
9@3o02D000000080o`0a00D000000?l0@03o0400o`10000000030?l0D`050000003o06T0o`1Y0?l0
J@0000000`3o08<01@000000o`2O0?l0W`3o09l000000080o`2o00D000000?l0h@3o0>40o`3Q0000
00030?T0o`050000003A0?l0d@3o0=40o`0000000`2X0?l01@000000O@3o07d0o`1m0?l000000080
D@3o00D0000002<0o`0S0?l08`3o00000003000:o`0500000000>Ol003Wo000io`0000000`00JOl0
1@00000009So002Ho`00V?l0000000800<So00D00000003fo`00m_l00?Ko00000003003of`050000
0000ojh00?n^003o[P0000000`00oh801@0000000?mH003oF000oeP0000000800?l_00D00000003o
2000o`P00?l80000000301_o00050000000no`00?_l003ko000000000`1No`001@000000O?l007co
001lo`0000000080Uol000D000000;3o002`o`00/?l0000000030<Ko00050000003Io`00fOl00=Wo
000000000`3Zo`001@000000n?l00?So003ho`0000000080oo/000D000000?oc003ol`00oo<00000
00030?o/00050000003oj000onP00?oX000000000`3oi`0010000000onL00?oW000000030?oZ0005
0000003ok`00onl00?o_000000000`3omP001@000000ooh00?on003ooP00000000<0mol000<00000
0?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@000000?Oo
003go`0000000`3ooP001@000000ooH00?of003omP00000000<0onl000D000000?oZ003ojP00onX0
000000030?oW00040000003oi`00onL0000000<0onP000D000000?o/003ok000on`0000000030?oc
00050000003on`00oo/00?ok000000000`3ho`0010000000j_l00>[o000000030=Wo000500000036
o`00a_l00<Ko000000000`2`o`001@000000Uol009Oo002Go`00000000<0O?l000@0000005ko001N
o`0000000`0no`001@0000006ol001_o000Ko`00000000<00?l800D00000003o;`00obl00?l_0000
0003003oF00400000000oh800?n2000000<00?n^00D00000003of`00om/00?oK00000003003fo`05
00000000b?l00<So0038o`0000000`00V?l01000000006Wo001Yo`000003000io`05000000002_l0
00[o000:o`0000000`0S0?l01@000000D@3o0540o`1A0?l0000000<0O@3o00@000000:P0o`2X0?l0
00000`3A0?l01@000000n@3o0?T0o`3i0?l0000000<0o`3Q00D000000?l0_`3o0;l0o`2o00000002
0?l0W`050000003o08<0o`230?l0P`0000000`3o06T01@000000o`1C0?l0D`3o05<0000000<0o`10
00D000000?l0<@3o0340o`0a000000020?l09@050000003o01d0o`0M0?l07@0000000`3o01X01@00
0000o`0J0?l06P3o01X0000000<0o`0M00D000000?l09@3o02D0o`0U000000020?l0<@050000003o
0400o`100?l0@00000000`3o05<01@000000o`1Y0?l0J@3o06T0000000<0o`2300D000000?l0W`3o
09l0o`2O000000020?l0_`050000003o0>40o`3Q0?l0h@0000000`3i0?l01@000000d@3o0=40o`3A
0?l0000000<0Z03o00D0000007d0o`1m0?l0O@3o000000020540o`050000000S0?l08`3o02<0o`00
00000`002_l01@00000003Wo000io`00>Ol0000000<006Wo00D00000002Ho`00V?l009So00000002
0038o`0500000000m_l00?Ko003fo`0000000`00om/01@0000000?n^003o[P00ojh0000000<00?n2
00D00000003oF000oeP00?mH00000002003o;`0500000000o`P00?l8003o200000000`0Ko`001@00
0000?_l003ko000no`00000000<0G_l000D0000007co001lo`00O?l00000000209Oo00050000002`
o`00/?l00;3o000000000`36o`001@000000fOl00=Wo003Io`00000000<0j_l000D000000?So003h
o`00n?l0000000020?ok00050000003ol`00oo<00?oc000000000`3ok0001@000000onP00?oX003o
j000000000<0onL000@000000?oW003oi`0000000`3ojP001@000000onl00?o_003ok`00000000<0
ooH000D000000?on003ooP00ooh0000000030?Oo00030000003oool0oooo00@0oooo0@0000010?oo
o`00803oool00`000000oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo
00<000000?ooo`3oool01@3oool010000000jOl00>Wo000000030?_o00050000003ol`00oo<00?oc
000000000`3oi0001@000000omL00?oG003oe`00000000<0ol`000@000000?o3003o``0000000`3o
_0001@000000okP00?nh003o^000000000<0okH000D000000?nh003o^000okP0000000030?nl0004
0000003o``00ol<0000000<0olh000D000000?oK003of`00om/0000000030?o/00050000003oo`00
ool00?oo000000000`3Wo`0010000000cOl00<go000000030:oo00050000002>o`00S_l008ko0000
00000`1Zo`001@000000A?l004Co0014o`00000000<06ol000@00000003o4000oa0000000`00ocd0
1@0000000?m]003oK@00ofd0000000<00?nN00D00000003od@00om400?oA00000003003ko`040000
0000a_l00<Ko000000<009;o00D00000001Mo`00GOl005go00000003000Yo`050000000:0?l02P3o
00X0o`0000000`0m0?l010000000KP3o06h0o`00000309d0o`050000003;0?l0b`3o0</0o`000000
0`3f0?l01@000000o`3Q0?l0h@3o0>4000000080o`2l00D000000?l0V`3o09/0o`2K000000030?l0
O@050000003o06<0o`1S0?l0H`0000000`3o04`01@000000o`0j0?l0>P3o03X000000080o`0]00D0
00000?l0903o02@0o`0T000000030?l07`050000003o01l0o`0O0?l07`0000000`3o02@01@000000
o`0]0?l0;@3o02d000000080o`0j00D000000?l0C03o04`0o`1<000000030?l0H`050000003o07d0
o`1m0?l0O@0000000`3o09/01@000000o`2l0?l0_03o0;`000000080o`3Q00D000000?H0o`3f0?l0
mP3o000000030</0o`050000002M0?l0W@3o09d0o`0000000`1^0?l01@000000?@3o03d0o`0m0?l0
000000802P3o00D00000000Yo`00:Ol002Wo00000003001Mo`0500000000T_l009;o002Bo`000000
0`00a_l01@0000000?_o003ko`00nol0000000800?oA00D00000003oWP00oih00?nN00000003003o
K@0500000000ocd00?lm003o?@0000000`00oa001@0000006ol001_o000Ko`0000000080A?l000D0
000006[o001Zo`00J_l00000000308ko00050000002_o`00[ol00:oo000000000`3=o`001@000000
iol00>Oo003Wo`0000000080ool000D000000?o/003ok000on`0000000030?oK00050000003ocP00
olh00?o>000000000`3o``001@000000ok`00?nl003o_00000000080okP000D000000?nf003o]P00
okH0000000030?nh00050000003o_000ok`00?nl000000000`3o``0010000000ol`00?o<00000003
0?oG00050000003oi000on@00?oT000000000`3ol`001@000000nol00?_o003ko`00000000<0jOl0
00<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@0
00000>Wo003Yo`0000000`3ko`001@000000oo<00?oc003ol`00000000<0on@000D000000?oG003o
e`00omL0000000030?o<00040000003o``00ol<0000000<0ok`000D000000?nh003o^000okP00000
00030?nf00050000003o^000okP00?nh000000000`3o_00010000000ol<00?o3000000030?o>0005
0000003of`00om/00?oK000000000`3ok0001@000000ool00?oo003oo`00000000<0iol000@00000
0<go003=o`0000000`2_o`001@000000S_l008ko002>o`00000000<0J_l000D0000004Co0014o`00
A?l00000000301_o000400000000oa000?l@000000<00?lm00D00000003oK@00ofd00?m]00000003
003oWP0500000000om400?oA003od@0000000`00nol0100000000<Ko0036o`000003002Bo`050000
0000GOl005go001Mo`0000000`00:Ol01@0000002P3o00X0o`0:0?l0000000<0?@3o00@0000006h0
o`1^0?l000000`2M0?l01@000000b`3o0</0o`3;0?l0000000<0mP3o00D000000?l0h@3o0>40o`3Q
000000020?l0_0050000003o09/0o`2K0?l0V`0000000`3o07d01@000000o`1S0?l0H`3o06<00000
00<0o`1<00D000000?l0>P3o03X0o`0j000000020?l0;@050000003o02@0o`0T0?l0900000000`3o
01l01@000000o`0O0?l07`3o01l0000000<0o`0T00D000000?l0;@3o02d0o`0]000000020?l0>P05
0000003o04`0o`1<0?l0C00000000`3o06<01@000000o`1m0?l0O@3o07d0000000<0o`2K00D00000
0?l0_03o0;`0o`2l000000020?l0h@050000003f0?l0mP3o0?H0o`0000000`3;0?l01@000000W@3o
09d0o`2M0?l0000000<0KP3o00D0000003d0o`0m0?l0?@3o0000000200X0o`0500000000:Ol002Wo
000Yo`0000000`00GOl01@00000009;o002Bo`00T_l0000000<00<Ko00D00000003ko`00nol00?_o
00000002003od@0500000000oih00?nN003oWP0000000`00ofd01@0000000?lm003o?@00ocd00000
00<00?l@00D0000001_o000Ko`006ol00000000204Co00050000001Zo`00J_l006[o000000000`2>
o`001@000000[ol00:oo002_o`00000000<0cOl000D000000>Oo003Wo`00iol0000000020?oo0005
0000003ok000on`00?o/000000000`3of`001@000000olh00?o>003ocP00000000<0ol<000D00000
0?nl003o_000ok`0000000020?nh00050000003o]P00okH00?nf000000000`3o^0001@000000ok`0
0?nl003o_000000000<0ol<000@000000?o<003oc00000000`3oe`001@000000on@00?oT003oi000
000000<0oo<000D000000?_o003ko`00nol0000000030>Wo00030000003oool0oooo00@0oooo0@00
00010?ooo`00803oool00`000000oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool0
0200oooo00<000000?ooo`3oool01@3oool010000000]Ol00;Go000000030<oo00050000003Wo`00
iol00>Oo000000000`3oo`001@000000onX00?oZ003ojP00000000<0omH000@000000?o4003oa000
00000`3o]0001@000000ojH00?nV003oYP00000000<0oi/000D000000?nD003oU000oi@000000003
0?n?00040000003oSP00ohh0000000<0oi0000D000000?nF003oUP00oiH0000000030?nO00050000
003o[@00ojd00?n]000000000`3o_P0010000000om@00?oD000000030?o]00050000003eo`00mOl0
0?Go000000000`3Do`001@000000/?l00;3o002`o`00000000<0R?l000@0000005go001Mo`000000
0`0_o`001@0000000?l1003o0@00o`40000000<00?ld00D00000003oJP00ofX00?mZ00000003003o
X@0400000000omT00?oI000000<00>co00D00000002bo`00/_l00;;o00000003001ho`0500000000
?ol003oo000oo`0000000`001_l010000000<03o0300o`00000306H0o`050000002J0?l0VP3o09X0
o`0000000`3;0?l01@000000n@3o0?T0o`3i0?l000000080o`3L00D000000?l0]@3o0;D0o`2e0000
00030?l0TP050000003o07@0o`1d0?l0M00000000`3o05X01@000000o`150?l0A@3o04D000000080
o`0e00D000000?l0:`3o02/0o`0[000000030?l09@050000003o02D0o`0U0?l09@0000000`3o02/0
1@000000o`0e0?l0=@3o03D000000080o`1500D000000?l0FP3o05X0o`1J000000030?l0M0050000
003o0980o`2B0?l0TP0000000`3o0;D01@000000o`3L0?l0g03o0=`000000080n@3o00D000000</0
o`3;0?l0b`3o0000000309X0o`050000001V0?l0IP3o06H0o`0000000`0`0?l01@00000000Ko0006
o`001_l00000008003oo00D00000001ho`00N?l007So00000003002bo`0500000000k?l00>co003/
o`0000000`00omT01@0000000?nQ003oX@00oj40000000800?mZ00D00000003o=000oc@00?ld0000
0003003o0@050000000_o`00;ol002oo000000000`1Mo`001@000000R?l008So0028o`0000000080
/?l000D000000=Co003Do`00e?l0000000030?Go00050000003ok@00ond00?o]000000000`3oe000
1@000000okh00?nn003o_P0000000080ojd000D000000?nO003oW`00oil0000000030?nF00050000
003oT000oi000?n@000000000`3oSP001@000000ohl00?n?003oS`0000000080oi@000D000000?nK
003oV`00oi/0000000030?nV00050000003o]000ok@00?nd000000000`3oa00010000000omH00?oF
000000030?oZ00050000003oo`00ool00?oo000000000`3Wo`001@000000col00<oo003?o`000000
00<0]Ol000<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0
oooo00@000000;Go002eo`0000000`3?o`001@000000iol00>Oo003Wo`00000000<0ool000D00000
0?oZ003ojP00onX0000000030?oF00040000003oa000ol@0000000<0ok@000D000000?nV003oYP00
ojH0000000030?nK00050000003oU000oi@00?nD000000000`3oS`0010000000ohh00?n>00000003
0?n@00050000003oUP00oiH00?nF000000000`3oW`001@000000ojd00?n]003o[@00000000<0okh0
00@000000?oD003oe00000000`3ok@001@000000mOl00?Go003eo`00000000<0e?l000D000000;3o
002`o`00/?l00000000308So00040000001Mo`00GOl0000000<0;ol000D00000003o0@00o`400?l1
00000003003o=00500000000ofX00?mZ003oJP0000000`00oj40100000000?oI003of@000003003/
o`0500000000/_l00;;o002bo`0000000`00N?l01@00000003oo000oo`00?ol0000000<000Ko00@0
00000300o`0`0?l000000`1V0?l01@000000VP3o09X0o`2J0?l0000000<0b`3o00D000000?T0o`3i
0?l0n@3o000000020?l0g0050000003o0;D0o`2e0?l0]@0000000`3o09801@000000o`1d0?l0M03o
07@0000000<0o`1J00D000000?l0A@3o04D0o`15000000020?l0=@050000003o02/0o`0[0?l0:`00
00000`3o02D01@000000o`0U0?l09@3o02D0000000<0o`0[00D000000?l0=@3o03D0o`0e00000002
0?l0A@050000003o05X0o`1J0?l0FP0000000`3o07@01@000000o`2B0?l0TP3o0980000000<0o`2e
00D000000?l0g03o0=`0o`3L000000020?T0o`050000003;0?l0b`3o0</0o`0000000`2J0?l01@00
0000IP3o06H0o`1V0?l0000000<0<03o00D000000006o`001_l000Ko00000002000oo`0500000000
N?l007So001ho`0000000`00/_l01@0000000>co003/o`00k?l0000000<00?oI00D00000003oX@00
oj400?nQ00000002003oJP0500000000oc@00?ld003o=00000000`00o`401@000000;ol002oo000_
o`00000000<0GOl000D0000008So0028o`00R?l0000000020;3o00050000003Do`00e?l00=Co0000
00000`3eo`001@000000ond00?o]003ok@00000000<0om@000D000000?nn003o_P00okh000000002
0?n]00050000003oW`00oil00?nO000000000`3oUP001@000000oi000?n@003oT000000000<0ohh0
00D000000?n?003oS`00ohl0000000020?nD00050000003oV`00oi/00?nK000000000`3oYP001@00
0000ok@00?nd003o]000000000<0ol@000@000000?oF003oeP0000000`3ojP001@000000ool00?oo
003oo`00000000<0iol000D000000<oo003?o`00col0000000030;Go00030000003oool0oooo00@0
oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`040000002eo`00]Ol0000000<0
col000D000000>Oo003Wo`00iol0000000030?oo00050000003ojP00onX00?oZ000000000`3oeP00
10000000ol@00?o4000000030?nd00050000003oYP00ojH00?nV000000000`3oV`001@000000oi@0
0?nD003oU000000000<0ohl000@000000?n>003oSP0000000`3oT0001@000000oiH00?nF003oUP00
000000<0oil000D000000?n]003o[@00ojd0000000030?nn00040000003oe000om@0000000<0ond0
00D000000?Go003eo`00mOl0000000030=Co00050000002`o`00/?l00;3o000000000`28o`001000
0000GOl005go0000000302oo000500000000o`400?l1003o0@0000000`00oc@01@0000000?mZ003o
JP00ofX0000000<00?nQ00@00000003of@00omT000000`00k?l01@0000000;;o002bo`00/_l00000
00<007So00D00000000oo`00?ol003oo000000030006o`040000000`0?l0<03o000000<0IP3o00D0
000009X0o`2J0?l0VP3o000000030</0o`050000003i0?l0n@3o0?T0o`0000000P3o0=`01@000000
o`2e0?l0]@3o0;D0000000<0o`2B00D000000?l0M03o07@0o`1d000000030?l0FP050000003o04D0
o`150?l0A@0000000P3o03D01@000000o`0[0?l0:`3o02/0000000<0o`0U00D000000?l09@3o02D0
o`0U000000030?l0:`050000003o03D0o`0e0?l0=@0000000P3o04D01@000000o`1J0?l0FP3o05X0
000000<0o`1d00D000000?l0TP3o0980o`2B000000030?l0]@050000003o0=`0o`3L0?l0g0000000
0P3i0?l01@000000b`3o0</0o`3;0?l0000000<0VP3o00D0000006H0o`1V0?l0IP3o000000030300
o`05000000001_l000Ko0006o`0000000P00?ol01@00000007So001ho`00N?l0000000<00;;o00D0
0000003/o`00k?l00>co00000003003of@0500000000oj400?nQ003oX@0000000P00ofX01@000000
0?ld003o=000oc@0000000<00?l100D0000002oo000_o`00;ol00000000305go000500000028o`00
R?l008So000000000P2`o`001@000000e?l00=Co003Do`00000000<0mOl000D000000?o]003ok@00
ond0000000030?oD00050000003o_P00okh00?nn000000000P3o[@001@000000oil00?nO003oW`00
000000<0oiH000D000000?n@003oT000oi00000000030?n>00050000003oS`00ohl00?n?00000000
0P3oU0001@000000oi/00?nK003oV`00000000<0ojH000D000000?nd003o]000ok@0000000030?o4
00040000003oeP00omH0000000<0onX000D000000?oo003oo`00ool0000000030>Oo00050000003?
o`00col00<oo000000000`2eo`000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<0
00000?ooo`3oool01@3ooooo0000084000001P3oool100000040oooo000P0?ooo`030000003oool0
oooo00D0oooo00@00000067o001Qo`0000000`21o`001@000000X?l00:3o002Po`00000000<0_ol0
00D000000=go003Mo`00gOl0000000030?[o00040000003oj@00onT0000000<0om0000D000000?nh
003o^000okP0000000030?nS00050000003oTP00oi800?nB000000000`3oP`0010000000ogL00?mg
000000030?m`00050000003oK000of`00?m/000000000`3oK@001@000000og800?mb003oLP000000
00<0og/000@000000?n9003oR@0000000`3oW0001@000000ok<00?nc003o/`00000000<0oll000D0
00000?o_003ok`00onl0000000030>_o000400000032o`00`_l0000000<0UOl000D0000006Go001U
o`00IOl000000003037o000500000000o`D00?l5003o1@0000000`00ocl0100000000?mj003oNP00
0003003o]`0500000000ooH00?of003omP0000000`00b_l01@00000008_o002;o`00Rol0000000<0
04co00@00000000?o`003ol000000`0/0?l01@000000IP3o06H0o`1V0?l0000000<0W@3o00D00000
0=40o`3A0?l0d@3o000000020?l0o@050000003o0=40o`3A0?l0d@0000000`3o0:T01@000000o`26
0?l0QP3o08H0000000<0o`1Y00D000000?l0D@3o0540o`1A000000020?l0?`050000003o0380o`0b
0?l0<P0000000`3o02`01@000000o`0/0?l0;03o02`0000000<0o`0b00D000000?l0?`3o03l0o`0o
000000020?l0D@050000003o06T0o`1Y0?l0J@0000000`3o08H01@000000o`2Y0?l0Z@3o0:T00000
00<0o`3A00D000000?l0o@3o0?d0o`3m000000020=40o`050000002M0?l0W@3o09d0o`0000000`1V
0?l01@000000;03o02`0o`0/0?l0000000<000oo00D00000001<o`00C?l004co00000002002;o`05
00000000b_l00<[o003:o`0000000`00ooH01@0000000?ng003o]`00okL0000000<00?mj00D00000
003o?`00ocl00?lo00000002003o1@050000000ao`00<Ol0037o000000000`1Uo`001@000000UOl0
09Go002Eo`00000000<0`_l000D000000>_o003[o`00jol0000000020?o_00050000003oc`00oll0
0?o?000000000`3o/`001@000000oi`00?nL003oW000000000<0ohT000D000000?mk003oN`00og/0
000000020?mb00050000003oK@00ofd00?m]000000000`3oK0001@000000og000?m`003oL0000000
00<0ogL000D000000?n3003oP`00oh<0000000020?nB00050000003oX`00oj<00?nS000000000`3o
^0001@000000om000?o@003od000000000<0onT000@000000?[o003jo`0000000`3Mo`001@000000
_ol00;oo002oo`00000000<0X?l000D00000087o0021o`00POl000000003067o00030000003oool0
oooo00@0oooo0@0000010?ooo`00803oool2000000H0oooo00@00000067o001Qo`0000000`21o`00
1@000000X?l00:3o002Po`00000000<0_ol000D000000=go003Mo`00gOl0000000030?[o00040000
003oj@00onT0000000<0om0000D000000?nh003o^000okP0000000030?nS00050000003oTP00oi80
0?nB000000000`3oP`0010000000ogL00?mg000000030?m`00050000003oK000of`00?m/00000000
0`3oK@001@000000og800?mb003oLP00000000<0og/000@000000?n9003oR@0000000`3oW0001@00
0000ok<00?nc003o/`00000000<0oll000D000000?o_003ok`00onl0000000030>_o000400000032
o`00`_l0000000<0UOl000D0000006Go001Uo`00IOl000000003037o000500000000o`D00?l5003o
1@0000000`00ocl0100000000?mj003oNP000003003o]`0500000000ooH00?of003omP0000000`00
b_l01@00000008_o002;o`00Rol0000000<004co00@00000000?o`003ol000000`0/0?l01@000000
IP3o06H0o`1V0?l0000000<0W@3o00D000000=40o`3A0?l0d@3o000000020?l0o@050000003o0=40
o`3A0?l0d@0000000`3o0:T01@000000o`260?l0QP3o08H0000000<0o`1Y00D000000?l0D@3o0540
o`1A000000020?l0?`050000003o0380o`0b0?l0<P0000000`3o02`01@000000o`0/0?l0;03o02`0
000000<0o`0b00D000000?l0?`3o03l0o`0o000000020?l0D@050000003o06T0o`1Y0?l0J@000000
0`3o08H01@000000o`2Y0?l0Z@3o0:T0000000<0o`3A00D000000?l0o@3o0?d0o`3m000000020=40
o`050000002M0?l0W@3o09d0o`0000000`1V0?l01@000000;03o02`0o`0/0?l0000000<000oo00D0
0000001<o`00C?l004co00000002002;o`0500000000b_l00<[o003:o`0000000`00ooH01@000000
0?ng003o]`00okL0000000<00?mj00D00000003o?`00ocl00?lo00000002003o1@050000000ao`00
<Ol0037o000000000`1Uo`001@000000UOl009Go002Eo`00000000<0`_l000D000000>_o003[o`00
jol0000000020?o_00050000003oc`00oll00?o?000000000`3o/`001@000000oi`00?nL003oW000
000000<0ohT000D000000?mk003oN`00og/0000000020?mb00050000003oK@00ofd00?m]00000000
0`3oK0001@000000og000?m`003oL000000000<0ogL000D000000?n3003oP`00oh<0000000020?nB
00050000003oX`00oj<00?nS000000000`3o^0001@000000om000?o@003od000000000<0onT000@0
00000?[o003jo`0000000`3Mo`001@000000_ol00;oo002oo`00000000<0X?l000D00000087o0021
o`00POl000000003067o00030000003oool0oooo00<0oooo0P0000010?ooo`00803oool00`000000
oooo0?ooo`050?ooo`040000001Qo`00HOl0000000<0POl000D000000:3o002Po`00X?l000000003
0;oo00050000003Mo`00gOl00=go000000000`3jo`0010000000onT00?oY000000030?o@00050000
003o^000okP00?nh000000000`3oX`001@000000oi800?nB003oTP00000000<0oh<000@000000?mg
003oM`0000000`3oL0001@000000of`00?m/003oK000000000<0ofd000D000000?mb003oLP00og80
000000030?mk00040000003oR@00ohT0000000<0oi`000D000000?nc003o/`00ok<0000000030?o?
00050000003ok`00onl00?o_000000000`3[o`0010000000`_l00<;o0000000309Go00050000001U
o`00IOl006Go000000000`0ao`001@0000000?l5003o1@00o`D0000000<00?lo00@00000003oNP00
ogX000000`00okL01@0000000?of003omP00ooH0000000<00<[o00D00000002;o`00Rol008_o0000
0003001<o`04000000003ol000oo000000<0;03o00D0000006H0o`1V0?l0IP3o0000000309d0o`05
0000003A0?l0d@3o0=40o`0000000P3o0?d01@000000o`3A0?l0d@3o0=40000000<0o`2Y00D00000
0?l0QP3o08H0o`26000000030?l0J@050000003o0540o`1A0?l0D@0000000P3o03l01@000000o`0b
0?l0<P3o0380000000<0o`0/00D000000?l0;03o02`0o`0/000000030?l0<P050000003o03l0o`0o
0?l0?`0000000P3o05401@000000o`1Y0?l0J@3o06T0000000<0o`2600D000000?l0Z@3o0:T0o`2Y
000000030?l0d@050000003o0?d0o`3m0?l0o@0000000P3A0?l01@000000W@3o09d0o`2M0?l00000
00<0IP3o00D0000002`0o`0/0?l0;03o00000003000?o`0500000000C?l004co001<o`0000000P00
Rol01@0000000<[o003:o`00b_l0000000<00?of00D00000003o]`00okL00?ng00000003003oNP05
00000000ocl00?lo003o?`0000000P00o`D01@000000<Ol0037o000ao`00000000<0IOl000D00000
09Go002Eo`00UOl0000000030<;o00050000003[o`00jol00>_o000000000P3ok`001@000000oll0
0?o?003oc`00000000<0ok<000D000000?nL003oW000oi`0000000030?n900050000003oN`00og/0
0?mk000000000P3oLP001@000000ofd00?m]003oK@00000000<0of`000D000000?m`003oL000og00
000000030?mg00050000003oP`00oh<00?n3000000000P3oTP001@000000oj<00?nS003oX`000000
00<0okP000D000000?o@003od000om00000000030?oY00040000003jo`00n_l0000000<0gOl000D0
00000;oo002oo`00_ol0000000030:3o000500000021o`00POl0087o000000000`1Qo`000`000000
oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo000008400000
1P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003o2@00o`T00000
0`0Io`001@000000?_l003ko000no`00000000<0H_l000D0000008Oo0027o`00Qol0000000030:co
00040000003@o`00d?l0000000<0l_l000D000000?o/003ok000on`0000000030?o=00050000003o
/@00ok400?na000000000`3oU`0010000000oh400?n1000000030?m_00050000003oH@00of400?mQ
000000000`3oE`001@000000oe400?mA003oD@00000000<0oe4000@000000?mF003oEP0000000`3o
H0001@000000ofl00?m_003oK`00000000<0oh@000D000000?nN003oWP00oih0000000030?nn0004
0000003oh`00on<0000000<0lol000D000000<Co0034o`00a?l000000003097o00050000001Jo`00
F_l005[o000000000`0Po`00100000000?lM003o7@000003003oG@0500000000oil00?nO003oW`00
00000`00on801@0000000=Wo003Io`00fOl0000000<009Go00@00000001Ao`00DOl000000`003ol0
1@000000<03o0300o`0`0?l0000000<0KP3o00D000000:P0o`2X0?l0Z03o000000020=l0o`050000
003o0>h0o`3^0?l0kP0000000`3o0<401@000000o`2J0?l0VP3o09X0000000<0o`1h00D000000?l0
G@3o05d0o`1M000000020?l0B0050000003o03X0o`0j0?l0>P0000000`3o03<01@000000o`0c0?l0
<`3o03<0000000<0o`0j00D000000?l0B03o04P0o`18000000020?l0G@050000003o07P0o`1h0?l0
N00000000`3o09X01@000000o`310?l0`@3o0<40000000<0o`3^00D000000=l0o`3O0?l0g`3o0000
00020:P0o`050000001^0?l0KP3o06h0o`0000000`0`0?l01@00000000oo000?o`003ol0000000<0
057o00D00000002Eo`00UOl009Go00000002003Io`0500000000on800?oR003ohP0000000`00oil0
1@0000000?mM003oG@00oed0000000<00?lM00D00000023o000Po`008?l00000000205[o00050000
002Ao`00TOl0097o000000000`34o`001@000000lol00??o003co`00000000<0on<000D000000?nn
003o_P00okh0000000020?nN00050000003oQ000oh@00?n4000000000`3oK`001@000000of000?mP
003oH000000000<0oeH000D000000?mA003oD@00oe40000000020?mA00050000003oE`00oeL00?mG
000000000`3oH@001@000000ofl00?m_003oK`00000000<0oh4000D000000?nG003oU`00oiL00000
00020?na00050000003oc@00old00?o=000000000`3ok0001@000000l_l00?;o003bo`00000000<0
d?l000@000000:co002/o`0000000`27o`001@000000H_l006;o001Ro`00000000<0?_l000D00000
01Wo000Io`006Ol000000003003o2@030000003oool0oooo00@0oooo0@0000010?ooo`00803oool0
0`000000oooo0?ooo`050?ooo`0400000000o`T00?l9000000<06Ol000D0000003ko000no`00?_l0
0000000306;o000500000027o`00Qol008Oo000000000`2/o`0010000000d?l00=3o000000030?;o
00050000003ok000on`00?o/000000000`3oc@001@000000ok400?na003o/@00000000<0oiL000@0
00000?n1003oP@0000000`3oK`001@000000of400?mQ003oH@00000000<0oeL000D000000?mA003o
D@00oe40000000030?mA00040000003oEP00oeH0000000<0of0000D000000?m_003oK`00ofl00000
00030?n400050000003oWP00oih00?nN000000000`3o_P0010000000on<00?oS000000030??o0005
00000034o`00a?l00<Co000000000`2Ao`001@000000F_l005[o001Jo`00000000<08?l000@00000
003o7@00oad000000`00oed01@0000000?nO003oW`00oil0000000<00?oR00D00000003Io`00fOl0
0=Wo00000003002Eo`0400000000DOl0057o000000<000oo00D000000300o`0`0?l0<03o00000003
06h0o`050000002X0?l0Z03o0:P0o`0000000P3O0?l01@000000o`3^0?l0kP3o0>h0000000<0o`31
00D000000?l0VP3o09X0o`2J000000030?l0N0050000003o05d0o`1M0?l0G@0000000P3o04P01@00
0000o`0j0?l0>P3o03X0000000<0o`0c00D000000?l0<`3o03<0o`0c000000030?l0>P050000003o
04P0o`180?l0B00000000P3o05d01@000000o`1h0?l0N03o07P0000000<0o`2J00D000000?l0`@3o
0<40o`31000000030?l0kP050000003O0?l0g`3o0=l0o`0000000P2X0?l01@000000KP3o06h0o`1^
0?l0000000<0<03o00D00000000?o`003ol000oo00000003001Ao`0500000000UOl009Go002Eo`00
00000P00fOl01@0000000?oR003ohP00on80000000<00?nO00D00000003oG@00oed00?mM00000003
003o7@050000000Po`008?l0023o000000000P1Jo`001@000000TOl0097o002Ao`00000000<0a?l0
00D000000??o003co`00lol0000000030?oS00050000003o_P00okh00?nn000000000P3oWP001@00
0000oh@00?n4003oQ000000000<0ofl000D000000?mP003oH000of00000000030?mF00050000003o
D@00oe400?mA000000000P3oD@001@000000oeL00?mG003oE`00000000<0of4000D000000?m_003o
K`00ofl0000000030?n100050000003oU`00oiL00?nG000000000P3o/@001@000000old00?o=003o
c@00000000<0on`000D000000?;o003bo`00l_l0000000030=3o00040000002/o`00[?l0000000<0
Qol000D0000006;o001Ro`00H_l00000000303ko00050000000Io`006Ol001Wo000000000`00o`T0
0`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool01000
00000?l9003o2@00000301Wo00050000000no`00?_l003ko000000000`1Ro`001@000000Qol008Oo
0027o`00000000<0[?l000@000000=3o003@o`0000000`3bo`001@000000on`00?o/003ok0000000
00<0old000D000000?na003o/@00ok40000000030?nG00040000003oP@00oh40000000<0ofl000D0
00000?mQ003oH@00of40000000030?mG00050000003oD@00oe400?mA000000000`3oD@0010000000
oeH00?mF000000030?mP00050000003oK`00ofl00?m_000000000`3oQ0001@000000oih00?nN003o
WP00000000<0okh000@000000?oS003oh`0000000`3co`001@000000a?l00<Co0034o`00000000<0
TOl000D0000005[o001Jo`00F_l000000003023o000400000000oad00?lM000000<00?mM00D00000
003oW`00oil00?nO00000003003ohP0500000000fOl00=Wo003Io`0000000`00UOl010000000057o
001Ao`000003000?o`050000000`0?l0<03o0300o`0000000`1^0?l01@000000Z03o0:P0o`2X0?l0
00000080g`3o00D000000?l0kP3o0>h0o`3^000000030?l0`@050000003o09X0o`2J0?l0VP000000
0`3o07P01@000000o`1M0?l0G@3o05d000000080o`1800D000000?l0>P3o03X0o`0j000000030?l0
<`050000003o03<0o`0c0?l0<`0000000`3o03X01@000000o`180?l0B03o04P000000080o`1M00D0
00000?l0N03o07P0o`1h000000030?l0VP050000003o0<40o`310?l0`@0000000`3o0>h01@000000
g`3o0=l0o`3O0?l000000080Z03o00D0000006h0o`1^0?l0KP3o000000030300o`05000000003ol0
00oo000?o`0000000`00DOl01@00000009Go002Eo`00UOl0000000800=Wo00D00000003ohP00on80
0?oR00000003003oW`0500000000oed00?mM003oG@0000000`00oad01@0000008?l0023o000Po`00
00000080F_l000D00000097o002Ao`00TOl0000000030<Co00050000003co`00lol00??o00000000
0`3oh`001@000000okh00?nn003o_P0000000080oih000D000000?n4003oQ000oh@0000000030?m_
00050000003oH000of000?mP000000000`3oEP001@000000oe400?mA003oD@0000000080oe4000D0
00000?mG003oE`00oeL0000000030?mQ00050000003oK`00ofl00?m_000000000`3oP@001@000000
oiL00?nG003oU`0000000080ok4000D000000?o=003oc@00old0000000030?o/00050000003bo`00
l_l00?;o000000000`3@o`0010000000[?l00:co0000000308Oo00050000001Ro`00H_l006;o0000
00000`0no`001@0000006Ol001Wo000Io`00000000<00?l900<000000?ooo`3oool0103oool10000
0040oooo000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?ooo`00
803oool00`000000oooo0?ooo`050?ooo`0400000000oh000?n0000000<00?mL00D00000003o=@00
ocD00?le00000003003o3@050000000Ko`006ol001_o000000000`15o`0010000000Kol006oo0000
000309Wo000500000032o`00`_l00<;o000000000`3Zo`001@000000onl00?o_003ok`00000000<0
ol/000@000000?nZ003oZP0000000`3oS0001@000000og<00?mc003oL`00000000<0oed000D00000
0?m=003oC@00odd0000000030?m100040000003o>`00oc/0000000<0oc/000D000000?m0003o@000
od00000000030?m<00050000003oG`00oel00?mO000000000`3oM`0010000000oiH00?nF00000003
0?nk00050000003oi@00onD00?oU000000000`3Zo`001@000000]?l00;Co002do`00000000<0N_l0
00@0000003co000lo`0000000`00o`D01@0000000?m9003oB@00odT0000000<00?n@00D00000003o
f000omP00?oH00000003003Mo`0400000000UOl009Go000000<004co00D000000006o`001_l000Ko
0000000303d0o`050000001m0?l0O@3o07d0o`0000000P2j0?l01@000000lP3o0?80o`3b0?l00000
00<0o`3K00D000000?l0[`3o0:l0o`2_000000030?l0R@050000003o06X0o`1Z0?l0JP0000000P3o
05<01@000000o`130?l0@`3o04<0000000<0o`0k00D000000?l0>`3o03/0o`0k000000030?l0@`05
0000003o05<0o`1C0?l0D`0000000P3o06X01@000000o`290?l0R@3o08T0000000<0o`2_00D00000
0?l0f`3o0=/0o`3K000000030?80o`050000002j0?l0^P3o0;X0o`0000000P1m0?l01@000000?@3o
03d0o`0m0?l0000000<000Ko00D00000001<o`00C?l004co00000003002Eo`0500000000gOl00=go
003Mo`0000000P00omP01@0000000?n@003oT000oi00000000<00?m900D00000003o1@00o`D00?l5
0000000303co00050000001jo`00N_l007[o000000000P2do`001@000000j_l00>[o003Zo`000000
00<0onD000D000000?nk003o^`00ok/0000000030?nF00050000003oM`00ogL00?mg000000000P3o
G`001@000000od`00?m<003oC000000000<0od0000D000000?lk003o>`00oc/0000000030?lk0005
0000003o@@00od400?m1000000000P3oC@001@000000oed00?mM003oG@00000000<0og<000D00000
0?n<003oS000oh`0000000030?nZ00050000003ob`00ol/00?o;000000000P3ok`001@000000j_l0
0>[o003Zo`00000000<0`_l000D0000009Wo002Io`00VOl00000000306oo000400000015o`00AOl0
000000<06ol000D00000003o3@00o`d00?l=00000003003o=@0500000000oe`00?mL003oG0000000
0`00oh000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3o
ool0100000000?n0003oP0000003003oG00500000000ocD00?le003o=@0000000`00o`d01@000000
6ol001_o000Ko`00000000<0AOl000@0000006oo001_o`0000000`2Io`001@000000`_l00<;o0032
o`00000000<0j_l000D000000?o_003ok`00onl0000000030?o;00040000003oZP00ojX0000000<0
oh`000D000000?mc003oL`00og<0000000030?mM00050000003oC@00odd00?m=000000000`3o@@00
10000000oc/00?lk000000030?lk00050000003o@000od000?m0000000000`3oC0001@000000oel0
0?mO003oG`00000000<0ogL000@000000?nF003oUP0000000`3o^`001@000000onD00?oU003oi@00
000000<0j_l000D000000;Co002do`00]?l00000000307[o00040000000lo`00??l0000000<00?l5
00D00000003oB@00odT00?m900000003003oT00500000000omP00?oH003of00000000`00gOl01000
000009Go002Eo`000003001<o`05000000001_l000Ko0006o`0000000`0m0?l01@000000O@3o07d0
o`1m0?l000000080^P3o00D000000?80o`3b0?l0lP3o000000030?l0f`050000003o0:l0o`2_0?l0
[`0000000`3o08T01@000000o`1Z0?l0JP3o06X000000080o`1C00D000000?l0@`3o04<0o`130000
00030?l0>`050000003o03/0o`0k0?l0>`0000000`3o04<01@000000o`1C0?l0D`3o05<000000080
o`1Z00D000000?l0R@3o08T0o`29000000030?l0[`050000003o0=/0o`3K0?l0f`0000000`3b0?l0
1@000000^P3o0;X0o`2j0?l000000080O@3o00D0000003d0o`0m0?l0?@3o000000030006o`050000
0000C?l004co001<o`0000000`00UOl01@0000000=go003Mo`00gOl0000000800?oH00D00000003o
T000oi000?n@00000003003oB@0500000000o`D00?l5003o1@0000000`0lo`001@000000N_l007[o
001jo`0000000080]?l000D000000>[o003Zo`00j_l0000000030?oU00050000003o^`00ok/00?nk
000000000`3oUP001@000000ogL00?mg003oM`0000000080oel000D000000?m<003oC000od`00000
00030?m000050000003o>`00oc/00?lk000000000`3o>`001@000000od400?m1003o@@0000000080
odd000D000000?mM003oG@00oed0000000030?mc00050000003oS000oh`00?n<000000000`3oZP00
1@000000ol/00?o;003ob`0000000080onl000D000000>[o003Zo`00j_l0000000030<;o00050000
002Io`00VOl009Wo000000000`1_o`0010000000AOl004Go0000000301_o000500000000o`d00?l=
003o3@0000000`00ocD01@0000000?mL003oG000oe`0000000<00?n000<000000?ooo`3oool0103o
ool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003oP000oh0000000`00
oe`01@0000000?le003o=@00ocD0000000<00?l=00D0000001_o000Ko`006ol00000000304Go0004
0000001_o`00Kol0000000<0VOl000D000000<;o0032o`00`_l0000000030>[o00050000003ok`00
onl00?o_000000000`3ob`0010000000ojX00?nZ000000030?n<00050000003oL`00og<00?mc0000
00000`3oG@001@000000odd00?m=003oC@00000000<0od4000@000000?lk003o>`0000000`3o>`00
1@000000od000?m0003o@000000000<0od`000D000000?mO003oG`00oel0000000030?mg00040000
003oUP00oiH0000000<0ok/000D000000?oU003oi@00onD0000000030>[o00050000002do`00]?l0
0;Co000000000`1jo`0010000000??l003co00000003003o1@0500000000odT00?m9003oB@000000
0`00oi001@0000000?oH003of000omP0000000<00=go00@00000002Eo`00UOl000000`00C?l01@00
000000Ko0006o`001_l0000000<0?@3o00D0000007d0o`1m0?l0O@3o000000020;X0o`050000003b
0?l0lP3o0?80o`0000000`3o0=/01@000000o`2_0?l0[`3o0:l0000000<0o`2900D000000?l0JP3o
06X0o`1Z000000020?l0D`050000003o04<0o`130?l0@`0000000`3o03/01@000000o`0k0?l0>`3o
03/0000000<0o`1300D000000?l0D`3o05<0o`1C000000020?l0JP050000003o08T0o`290?l0R@00
00000`3o0:l01@000000o`3K0?l0f`3o0=/0000000<0lP3o00D000000;X0o`2j0?l0^P3o00000002
07d0o`050000000m0?l0?@3o03d0o`0000000`001_l01@00000004co001<o`00C?l0000000<009Go
00D00000003Mo`00gOl00=go00000002003of00500000000oi000?n@003oT00000000`00odT01@00
00000?l5003o1@00o`D0000000<0??l000D0000007[o001jo`00N_l0000000020;Co00050000003Z
o`00j_l00>[o000000000`3oi@001@000000ok/00?nk003o^`00000000<0oiH000D000000?mg003o
M`00ogL0000000020?mO00050000003oC000od`00?m<000000000`3o@0001@000000oc/00?lk003o
>`00000000<0oc/000D000000?m1003o@@00od40000000020?m=00050000003oG@00oed00?mM0000
00000`3oL`001@000000oh`00?n<003oS000000000<0ojX000D000000?o;003ob`00ol/000000002
0?o_00050000003Zo`00j_l00>[o000000000`32o`001@000000VOl009Wo002Io`00000000<0Kol0
00@0000004Go0015o`0000000`0Ko`001@0000000?l=003o3@00o`d0000000<00?le00D00000003o
G000oe`00?mL00000003003oP0030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`00
0000oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3o
ool01@3oool0100000000?of003omP000003003oe00500000000ojl00?n_003o[`0000000`00ohL0
1@0000000?mM003oG@00oed0000000<00?la00@00000003o0`00o`<000000`0[o`001@000000F_l0
05[o001Jo`00000000<0ROl000D000000;So002ho`00^?l0000000030>Go00040000003ok`00onl0
000000<0olH000D000000?nQ003oX@00oj40000000030?n000050000003oH`00of<00?mS00000000
0`3oC00010000000ocX00?lj000000030?l^00050000003o:@00obT00?lY000000000`3o:P001@00
0000oc800?lb003o<P00000000<0od8000@000000?mI003oF@0000000`3oMP001@000000oi/00?nK
003oV`00000000<0olL000D000000?oi003on@00ooT0000000030<ko00040000002Ao`00TOl00000
00<0Col000D0000000Wo0009o`002Ol000000003003o?`0500000000oh/00?n;003oR`0000000`00
omP0100000000=Wo003Io`000003002;o`0500000000?ol003oo000oo`0000000`0:0?l01@000000
D@3o0540o`1A0?l000000080T`3o00D000000=40o`3A0?l0d@3o000000030?l0mP050000003o0<D0
o`350?l0a@0000000`3o09/01@000000o`1h0?l0N03o07P000000080o`1N00D000000?l0C03o04`0
o`1<000000030?l0@`050000003o04<0o`130?l0@`0000000`3o04`01@000000o`1N0?l0GP3o05h0
00000080o`1h00D000000?l0V`3o09/0o`2K000000030?l0a@050000003o0?H0o`3f0?l0mP000000
0`3A0?l01@000000T`3o09<0o`2C0?l000000080D@3o00D0000000X0o`0:0?l02P3o00000003000o
o`0500000000Rol008_o002;o`0000000`00fOl01@0000000?oH003of000omP0000000800?n;00D0
0000003o?`00ocl00?lo0000000300Wo00050000001?o`00Col004oo000000000`2Ao`001@000000
c_l00<ko003>o`0000000080ooT000D000000?o7003oa`00olL0000000030?nK00050000003oMP00
ogH00?mf000000000`3oF@001@000000od800?m2003o@P0000000080oc8000D000000?lZ003o:P00
obX0000000030?lY00050000003o;P00obh00?l^000000000`3o>P001@000000od`00?m<003oC000
00000080of<000D000000?n0003oP000oh00000000030?nQ00050000003oaP00olH00?o600000000
0`3ok`001@000000iOl00>Go003Uo`0000000080^?l000D0000008Wo0029o`00ROl00000000305[o
00050000000[o`00:ol002_o000000000`00o`<0100000000?la003o<@000003003oG@0500000000
ohL00?n7003oQ`0000000`00ojl01@0000000?oD003oe000om@0000000<00?of00<000000?ooo`3o
ool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003omP00ooH0
00000`00om@01@0000000?n_003o[`00ojl0000000<00?n700D00000003oG@00oed00?mM00000003
003o<@0400000000o`<00?l3000000<0:ol000D0000005[o001Jo`00F_l00000000308Wo00050000
002ho`00^?l00;So000000000`3Uo`0010000000onl00?o_000000030?o600050000003oX@00oj40
0?nQ000000000`3oP0001@000000of<00?mS003oH`00000000<0od`000@000000?lj003o>P000000
0`3o;P001@000000obT00?lY003o:@00000000<0obX000D000000?lb003o<P00oc80000000030?m2
00040000003oF@00oeT0000000<0ogH000D000000?nK003oV`00oi/0000000030?o700050000003o
n@00ooT00?oi000000000`3>o`0010000000TOl0097o0000000304oo000500000009o`002Ol000Wo
000000000`00ocl01@0000000?n;003oR`00oh/0000000<00?oH00@00000003Io`00fOl000000`00
Rol01@00000003oo000oo`00?ol0000000<02P3o00D000000540o`1A0?l0D@3o0000000209<0o`05
0000003A0?l0d@3o0=40o`0000000`3o0?H01@000000o`350?l0a@3o0<D0000000<0o`2K00D00000
0?l0N03o07P0o`1h000000020?l0GP050000003o04`0o`1<0?l0C00000000`3o04<01@000000o`13
0?l0@`3o04<0000000<0o`1<00D000000?l0GP3o05h0o`1N000000020?l0N0050000003o09/0o`2K
0?l0V`0000000`3o0<D01@000000o`3f0?l0mP3o0?H0000000<0d@3o00D0000009<0o`2C0?l0T`3o
000000020540o`050000000:0?l02P3o00X0o`0000000`00?ol01@00000008_o002;o`00Rol00000
00<00=Wo00D00000003of000omP00?oH00000002003oR`0500000000ocl00?lo003o?`0000000`09
o`001@000000Col004oo001?o`00000000<0TOl000D000000<ko003>o`00c_l0000000020?oi0005
0000003oa`00olL00?o7000000000`3oV`001@000000ogH00?mf003oMP00000000<0oeT000D00000
0?m2003o@P00od80000000020?lb00050000003o:P00obX00?lZ000000000`3o:@001@000000obh0
0?l^003o;P00000000<0ocX000D000000?m<003oC000od`0000000020?mS00050000003oP000oh00
0?n0000000000`3oX@001@000000olH00?o6003oaP00000000<0onl000D000000>Go003Uo`00iOl0
000000020;So000500000029o`00ROl008Wo000000000`1Jo`001@000000:ol002_o000[o`000000
00<00?l300@00000003o<@00oc4000000`00oed01@0000000?n7003oQ`00ohL0000000<00?n_00D0
0000003oe000om@00?oD00000003003omP030000003oool0oooo00@0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<00000
0?ooo`3oool01@3oool0100000000:7o002Qo`000003002mo`0500000000g?l00=co003Lo`000000
0`00ool01@0000000?oH003of000omP0000000<00?n]00@00000003oO`00ogl000000`00odh01@00
00000?lK003o6`00oa/0000000<05ol000D0000004co001<o`00C?l000000003087o00040000002d
o`00]?l0000000<0i_l000D000000?oY003oj@00onT0000000030?nm00050000003oU000oi@00?nD
000000000`3oL00010000000oe800?mB000000030?lj00050000003o:000obP00?lX000000000`3o
7@001@000000oaX00?lJ003o6P00000000<0oal000@000000?l/003o;00000000`3o@@001@000000
oel00?mO003oG`00000000<0oh@000D000000?na003o/@00ok40000000030?oU00040000003Oo`00
gol0000000<0Wol000D0000005So001Ho`00F?l00000000300ko000500000000ocl00?lo003o?`00
00000`00oi00100000000?oR003ohP000003003:o`0500000000N?l007So001ho`0000000`00:Ol0
1@0000008`3o02<0o`0S0?l000000080K03o00D000000;00o`2`0?l0/03o000000030>d0o`050000
003o0=`0o`3L0?l0g00000000`3o0:d01@000000o`270?l0Q`3o08L000000080o`1Z00D000000?l0
EP3o05H0o`1F000000030?l0C0050000003o04`0o`1<0?l0C00000000`3o05H01@000000o`1Z0?l0
JP3o06X000000080o`2700D000000?l0[@3o0:d0o`2]000000030?l0g0050000003]0?l0k@3o0>d0
o`0000000`2`0?l01@000000K03o06`0o`1/0?l0000000808`3o00D00000000Yo`00:Ol002Wo0000
0003001ho`0500000000b_l00<[o003:o`0000000`00on801@0000000?n@003oT000oi0000000080
0?lo00D0000000ko000>o`003_l00000000305So00050000002Oo`00Wol009oo000000000`3Oo`00
1@000000onD00?oU003oi@0000000080ok4000D000000?n4003oQ000oh@0000000030?mO00050000
003o@@00od400?m1000000000`3o;0001@000000oal00?lO003o7`0000000080oaX000D000000?lM
003o7@00oad0000000030?lX00050000003o>P00ocX00?lj000000000`3oDP001@000000og000?m`
003oL00000000080oi@000D000000?nm003o_@00okd0000000030?oY00050000003Vo`00i_l00>Ko
000000000`2do`001@000000POl0087o0021o`0000000080C?l000D0000001Oo000Go`005ol00000
0003003o6`0500000000odh00?m>003oCP0000000`00ogl0100000000?n]003o[@000003003of005
00000000ool00?oo003oo`0000000`00g?l01@0000000;go002mo`00_Ol0000000<00:7o00<00000
0?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000002Q
o`00XOl000000`00_Ol01@0000000=co003Lo`00g?l0000000<00?oo00D00000003of000omP00?oH
00000003003o[@0400000000ogl00?mo000000<00?m>00D00000003o6`00oa/00?lK0000000301Oo
00050000001<o`00C?l004co000000000`21o`0010000000]?l00;Co000000030>Ko00050000003o
j@00onT00?oY000000000`3o_@001@000000oi@00?nD003oU000000000<0og0000@000000?mB003o
DP0000000`3o>P001@000000obP00?lX003o:000000000<0oad000D000000?lJ003o6P00oaX00000
00030?lO00040000003o;000ob`0000000<0od4000D000000?mO003oG`00oel0000000030?n40005
0000003o/@00ok400?na000000000`3oi@0010000000gol00=oo0000000309oo00050000001Ho`00
F?l005So000000000`0>o`001@0000000?lo003o?`00ocl0000000<00?n@00@00000003ohP00on80
00000`00b_l01@00000007So001ho`00N?l0000000<002Wo00D0000002<0o`0S0?l08`3o00000002
06`0o`050000002`0?l0/03o0;00o`0000000`3]0?l01@000000o`3L0?l0g03o0=`0000000<0o`2]
00D000000?l0Q`3o08L0o`27000000020?l0JP050000003o05H0o`1F0?l0EP0000000`3o04`01@00
0000o`1<0?l0C03o04`0000000<0o`1F00D000000?l0JP3o06X0o`1Z000000020?l0Q`050000003o
0:d0o`2]0?l0[@0000000`3o0=`01@000000k@3o0>d0o`3]0?l0000000<0/03o00D0000006`0o`1/
0?l0K03o0000000202<0o`0500000000:Ol002Wo000Yo`0000000`00N?l01@0000000<[o003:o`00
b_l0000000<00?oR00D00000003oT000oi000?n@00000002003o?`050000000>o`003_l000ko0000
00000`1Ho`001@000000Wol009oo002Oo`00000000<0gol000D000000?oU003oi@00onD000000002
0?na00050000003oQ000oh@00?n4000000000`3oG`001@000000od400?m1003o@@00000000<0ob`0
00D000000?lO003o7`00oal0000000020?lJ00050000003o7@00oad00?lM000000000`3o:0001@00
0000ocX00?lj003o>P00000000<0oe8000D000000?m`003oL000og00000000020?nD00050000003o
_@00okd00?nm000000000`3oj@001@000000i_l00>Ko003Vo`00000000<0]?l000D00000087o0021
o`00POl00000000204co00050000000Go`005ol001Oo000000000`00oa/01@0000000?m>003oCP00
odh0000000<00?mo00@00000003o[@00ojd000000`00omP01@0000000?oo003oo`00ool0000000<0
0=co00D00000002mo`00_Ol00;go00000003002Qo`030000003oool0oooo00@0oooo0@0000010?oo
o`00803oool00`000000oooo0?ooo`050?ooo`0400000000XOl00:7o000000<00;go00D00000003L
o`00g?l00=co00000003003oo`0500000000omP00?oH003of00000000`00ojd0100000000?mo003o
O`000003003oCP0500000000oa/00?lK003o6`0000000`0Go`001@000000C?l004co001<o`000000
00<0POl000@000000;Co002do`0000000`3Vo`001@000000onT00?oY003oj@00000000<0okd000D0
00000?nD003oU000oi@0000000030?m`00040000003oDP00oe80000000<0ocX000D000000?lX003o
:000obP0000000030?lM00050000003o6P00oaX00?lJ000000000`3o7`0010000000ob`00?l/0000
00030?m100050000003oG`00oel00?mO000000000`3oQ0001@000000ok400?na003o/@00000000<0
onD000@000000=oo003Oo`0000000`2Oo`001@000000F?l005So001Ho`00000000<03_l000D00000
003o?`00ocl00?lo00000003003oT00400000000on800?oR000000<00<[o00D00000001ho`00N?l0
07So00000003000Yo`050000000S0?l08`3o02<0o`0000000P1/0?l01@000000/03o0;00o`2`0?l0
000000<0k@3o00D000000?l0g03o0=`0o`3L000000030?l0[@050000003o08L0o`270?l0Q`000000
0P3o06X01@000000o`1F0?l0EP3o05H0000000<0o`1<00D000000?l0C03o04`0o`1<000000030?l0
EP050000003o06X0o`1Z0?l0JP0000000P3o08L01@000000o`2]0?l0[@3o0:d0000000<0o`3L00D0
00000>d0o`3]0?l0k@3o000000030;00o`050000001/0?l0K03o06`0o`0000000P0S0?l01@000000
02Wo000Yo`00:Ol0000000<007So00D00000003:o`00b_l00<[o00000003003ohP0500000000oi00
0?n@003oT00000000P00ocl01@0000003_l000ko000>o`00000000<0F?l000D0000009oo002Oo`00
Wol0000000030=oo00050000003oi@00onD00?oU000000000P3o/@001@000000oh@00?n4003oQ000
000000<0oel000D000000?m1003o@@00od40000000030?l/00050000003o7`00oal00?lO00000000
0P3o6P001@000000oad00?lM003o7@00000000<0obP000D000000?lj003o>P00ocX0000000030?mB
00050000003oL000og000?m`000000000P3oU0001@000000okd00?nm003o_@00000000<0onT000D0
00000>Ko003Vo`00i_l0000000030;Co000500000021o`00POl0087o000000000P1<o`001@000000
5ol001Oo000Go`00000000<00?lK00D00000003oCP00odh00?m>00000003003oO`0400000000ojd0
0?n]000000<00?oH00D00000003oo`00ool00?oo00000003003Lo`0500000000_Ol00;go002mo`00
00000`00XOl00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool0
1@3ooooo0000084000001P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@0
0000001Bo`00D_l000000`00Hol01@00000007[o001jo`00N_l0000000<009Go00D00000002fo`00
]_l00;Ko00000003003Lo`0400000000ooP00?oh000000<00?o:00D00000003oU`00oiL00?nG0000
0003003oHP0500000000ob/00?l[003o:`0000000`0=o`0010000000Aol004Oo00000003087o0005
0000002io`00^Ol00;Wo000000000`3`o`001@000000om`00?oL003og000000000<0oj`000@00000
0?n2003oPP0000000`3oG@001@000000och00?ln003o?P00000000<0obL000D000000?lG003o5`00
oaL0000000030?l@00040000003o4P00oa80000000<0oad000D000000?la003o<@00oc4000000003
0?m>00050000003oM000og@00?md000000000`3oXP0010000000omT00?oI000000030>So00050000
002So`00Xol00:?o000000000`1Ho`001@0000002Ol000Wo0009o`00000000<00?m900@00000003o
W`00oil000000`00ooH01@0000000;;o002bo`00/_l0000000<005go00D00000000:o`002_l000[o
0000000204<0o`050000002=0?l0S@3o08d0o`0000000`3@0?l01@000000o`3c0?l0l`3o0?<00000
00<0o`3000D000000?l0UP3o09H0o`2F000000020?l0MP050000003o0600o`1P0?l0H00000000`3o
05D01@000000o`1E0?l0E@3o05D0000000<0o`1P00D000000?l0MP3o07H0o`1f000000020?l0UP05
0000003o0<00o`300?l0`00000000`3o0?<01@000000d03o0=00o`3@0?l0000000<0S@3o00D00000
04<0o`130?l0@`3o00000002000:o`0500000000GOl005go001Mo`0000000`00/_l01@0000000?of
003omP00ooH0000000<00?nO00D00000003oB@00odT00?m90000000200Wo00050000001Ho`00F?l0
05So000000000`2So`001@000000j?l00>So003Xo`00000000<0omT000D000000?nR003oXP00oj80
000000020?md00050000003oCP00odh00?m>000000000`3o<@001@000000oad00?lM003o7@000000
00<0oa8000D000000?l@003o4000oa00000000020?lG00050000003o9`00obL00?lW000000000`3o
?P001@000000oed00?mM003oG@00000000<0oh8000D000000?n/003o[000oj`0000000020?oL0005
0000003`o`00l?l00?3o000000000`2io`001@000000POl0087o0021o`00000000<0Aol000D00000
00go000=o`003Ol000000002003o:`0500000000of800?mR003oHP0000000`00oiL01@0000000?o:
003obP00olX0000000<00?oh00@00000003Lo`00g?l000000`00]_l01@00000009Go002Eo`00UOl0
000000<007[o00D00000001So`00Hol006?o00000003001Bo`030000003oool0oooo00@0oooo0@00
00010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000D_l005;o000000<006?o00D0
0000001jo`00N_l007[o00000003002Eo`0500000000]_l00;Ko002fo`0000000`00g?l010000000
0?oh003on0000003003obP0500000000oiL00?nG003oU`0000000`00of801@0000000?l[003o:`00
ob/0000000<03Ol000@0000004Oo0017o`0000000`21o`001@000000^Ol00;Wo002io`00000000<0
l?l000D000000?oL003og000om`0000000030?n/00040000003oPP00oh80000000<0oed000D00000
0?ln003o?P00och0000000030?lW00050000003o5`00oaL00?lG000000000`3o400010000000oa80
0?lB000000030?lM00050000003o<@00oc400?la000000000`3oCP001@000000og@00?md003oM000
000000<0oj8000@000000?oI003of@0000000`3Xo`001@000000Xol00:?o002So`00000000<0F?l0
00D0000000Wo0009o`002Ol000000003003oB@0400000000oil00?nO000000<00?of00D00000002b
o`00/_l00;;o00000003001Mo`05000000002_l000[o000:o`0000000P130?l01@000000S@3o08d0
o`2=0?l0000000<0d03o00D000000?l0l`3o0?<0o`3c000000030?l0`0050000003o09H0o`2F0?l0
UP0000000P3o07H01@000000o`1P0?l0H03o0600000000<0o`1E00D000000?l0E@3o05D0o`1E0000
00030?l0H0050000003o07H0o`1f0?l0MP0000000P3o09H01@000000o`300?l0`03o0<00000000<0
o`3c00D000000=00o`3@0?l0d03o0000000308d0o`05000000130?l0@`3o04<0o`0000000P002_l0
1@00000005go001Mo`00GOl0000000<00;;o00D00000003omP00ooH00?of00000003003oW`050000
0000odT00?m9003oB@0000000P09o`001@000000F?l005So001Ho`00000000<0Xol000D000000>So
003Xo`00j?l0000000030?oI00050000003oXP00oj800?nR000000000P3oM0001@000000odh00?m>
003oCP00000000<0oc4000D000000?lM003o7@00oad0000000030?lB00050000003o4000oa000?l@
000000000P3o5`001@000000obL00?lW003o9`00000000<0och000D000000?mM003oG@00oed00000
00030?n200050000003o[000oj`00?n/000000000P3og0001@000000l?l00?3o003`o`00000000<0
^Ol000D00000087o0021o`00POl00000000304Oo00050000000=o`003Ol000go000000000P00ob/0
1@0000000?mR003oHP00of80000000<00?nG00D00000003obP00olX00?o:00000003003on0040000
0000g?l00=co000000<00;Ko00D00000002Eo`00UOl009Go00000003001jo`0500000000Hol006?o
001So`0000000`00D_l00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?oo
o`3oool01@3oool01000000005;o001Bo`000003001So`0500000000N_l007[o001jo`0000000`00
UOl01@0000000;Ko002fo`00]_l0000000<00=co00@00000003on000ooP000000`00olX01@000000
0?nG003oU`00oiL0000000<00?mR00D00000003o:`00ob/00?l[0000000300go000400000017o`00
Aol0000000<0POl000D000000;Wo002io`00^Ol0000000030?3o00050000003og000om`00?oL0000
00000`3o[00010000000oh800?n2000000030?mM00050000003o?P00och00?ln000000000`3o9`00
1@000000oaL00?lG003o5`00000000<0oa0000@000000?lB003o4P0000000`3o7@001@000000oc40
0?la003o<@00000000<0odh000D000000?md003oM000og@0000000030?nR00040000003of@00omT0
000000<0j?l000D000000:?o002So`00Xol00000000305So000500000009o`002Ol000Wo00000000
0`00odT0100000000?nO003oW`000003003omP0500000000/_l00;;o002bo`0000000`00GOl01@00
000000[o000:o`002_l000000080@`3o00D0000008d0o`2=0?l0S@3o000000030=00o`050000003o
0?<0o`3c0?l0l`0000000`3o0<001@000000o`2F0?l0UP3o09H000000080o`1f00D000000?l0H03o
0600o`1P000000030?l0E@050000003o05D0o`1E0?l0E@0000000`3o06001@000000o`1f0?l0MP3o
07H000000080o`2F00D000000?l0`03o0<00o`30000000030?l0l`050000003@0?l0d03o0=00o`00
00000`2=0?l01@000000@`3o04<0o`130?l00000008000[o00D00000001Mo`00GOl005go00000003
002bo`0500000000ooH00?of003omP0000000`00oil01@0000000?m9003oB@00odT0000000802Ol0
00D0000005So001Ho`00F?l0000000030:?o00050000003Xo`00j?l00>So000000000`3of@001@00
0000oj800?nR003oXP0000000080og@000D000000?m>003oCP00odh0000000030?la00050000003o
7@00oad00?lM000000000`3o4P001@000000oa000?l@003o400000000080oaL000D000000?lW003o
9`00obL0000000030?ln00050000003oG@00oed00?mM000000000`3oPP001@000000oj`00?n/003o
[00000000080om`000D000000?3o003`o`00l?l0000000030;Wo000500000021o`00POl0087o0000
00000`17o`001@0000003Ol000go000=o`00000000800?l[00D00000003oHP00of800?mR00000003
003oU`0500000000olX00?o:003obP0000000`00ooP0100000000=co003Lo`000003002fo`050000
0000UOl009Go002Eo`0000000`00N_l01@00000006?o001So`00Hol0000000<005;o00<000000?oo
o`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0
oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`04000000008_l002;o000000<0
02Ko00D00000000ao`00<Ol0037o000000030012o`0500000000FOl005Wo001Io`0000000`00M_l0
1000000009Wo002Io`0000030031o`0500000000kol00>oo003_o`0000000`00omd01@0000000?nV
003oYP00ojH0000000<00?m/00@00000003o;`00obl000000`0>o`001@000000COl004go001=o`00
000000<0S?l000D000000<So0038o`00b?l0000000030?om00040000003oaP00olH0000000<0oiD0
00D000000?mZ003oJP00ofX0000000030?m500050000003o:@00obT00?lY000000000`3o5@001000
0000o`X00?l:000000030?l900050000003o4P00oa800?lB000000000`3o9@001@000000od<00?m3
003o@`00000000<0ofX000@000000?nK003oV`0000000`3oe@001@000000j?l00>So003Xo`000000
00<0Wol000D0000004oo001?o`00Col000000003003o1@0400000000oed00?mM000000<00?ng00D0
0000003/o`00k?l00>co00000003002Bo`0500000000>Ol003Wo000io`0000000P0J0?l01@000000
JP3o06X0o`1Z0?l0000000<0/P3o00D000000?<0o`3c0?l0l`3o000000030?l0e0050000003o0:H0
o`2V0?l0YP0000000P3o08801@000000o`1[0?l0J`3o06/0000000<0o`1O00D000000?l0G`3o05l0
o`1O000000030?l0J`050000003o0880o`220?l0PP0000000P3o0:H01@000000o`3D0?l0e03o0=@0
000000<0l`3o00D000000;80o`2b0?l0/P3o0000000306X0o`050000000J0?l06P3o01X0o`000000
0P00>Ol01@00000009;o002Bo`00T_l0000000<00>co00D00000003o]`00okL00?ng00000003003o
G@0500000000o`D00?l5003o1@0000000P1?o`001@000000Wol009oo002Oo`00000000<0j?l000D0
00000?oE003oe@00omD0000000030?nK00050000003oJP00ofX00?mZ000000000P3o@`001@000000
obD00?lU003o9@00000000<0oa8000D000000?l9003o2@00o`T0000000030?l:00050000003o5@00
oaD00?lE000000000P3o:@001@000000odD00?m5003oA@00000000<0ofX000D000000?nE003oU@00
oiD0000000030?o600050000003oo@00ood00?om000000000P38o`001@000000S?l008co002<o`00
000000<0COl000D0000000ko000>o`003_l000000003003o;`0500000000of`00?m/003oK0000000
0P00ojH01@0000000?oM003og@00omd0000000<00>oo00D000000031o`00`Ol00<7o00000003002I
o`0400000000M_l007Ko000000<005Wo00D000000012o`00@_l004;o00000003000ao`0500000000
9_l002Ko000Vo`0000000`008_l00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<0
00000?ooo`3oool01@3oool01000000002;o000Ro`000003000Vo`0500000000<Ol0037o000ao`00
00000`00@_l01@00000005Wo001Io`00FOl0000000<007Ko00@00000002Io`00VOl000000`00`Ol0
1@0000000>oo003_o`00kol0000000<00?oM00D00000003oYP00ojH00?nV00000003003oK0040000
0000obl00?l_000000<03_l000D0000004go001=o`00COl00000000308co000500000038o`00b?l0
0<So000000000`3oo@0010000000olH00?o6000000030?nE00050000003oJP00ofX00?mZ00000000
0`3oA@001@000000obT00?lY003o:@00000000<0oaD000@000000?l:003o2P0000000`3o2@001@00
0000oa800?lB003o4P00000000<0obD000D000000?m3003o@`00od<0000000030?mZ00040000003o
V`00oi/0000000<0omD000D000000>So003Xo`00j?l00000000309oo00050000001?o`00Col004oo
000000000`00o`D0100000000?mM003oG@000003003o]`0500000000k?l00>co003/o`0000000`00
T_l01@00000003Wo000io`00>Ol0000000806P3o00D0000006X0o`1Z0?l0JP3o000000030;80o`05
0000003c0?l0l`3o0?<0o`0000000`3o0=@01@000000o`2V0?l0YP3o0:H000000080o`2200D00000
0?l0J`3o06/0o`1[000000030?l0G`050000003o05l0o`1O0?l0G`0000000`3o06/01@000000o`22
0?l0PP3o088000000080o`2V00D000000?l0e03o0=@0o`3D000000030?<0o`050000002b0?l0/P3o
0;80o`0000000`1Z0?l01@0000006P3o01X0o`0J0?l00000008003Wo00D00000002Bo`00T_l009;o
00000003003/o`0500000000okL00?ng003o]`0000000`00oed01@0000000?l5003o1@00o`D00000
0080Col000D0000009oo002Oo`00Wol0000000030>So00050000003oe@00omD00?oE000000000`3o
V`001@000000ofX00?mZ003oJP0000000080od<000D000000?lU003o9@00obD0000000030?lB0005
0000003o2@00o`T00?l9000000000`3o2P001@000000oaD00?lE003o5@0000000080obT000D00000
0?m5003oA@00odD0000000030?mZ00050000003oU@00oiD00?nE000000000`3oaP001@000000ood0
0?om003oo@0000000080b?l000D0000008co002<o`00S?l00000000304go00050000000>o`003_l0
00ko000000000`00obl01@0000000?m/003oK000of`0000000800?nV00D00000003og@00omd00?oM
00000003003_o`0500000000`Ol00<7o0031o`0000000`00VOl01000000007Ko001fo`000003001I
o`0500000000@_l004;o0012o`0000000`00<Ol01@00000002Ko000Vo`009_l0000000<002;o00<0
00000?ooo`3oool0103oool100000040oooo000P0?ooo`8000001P3oool01000000002;o000Ro`00
0003000Vo`0500000000<Ol0037o000ao`0000000`00@_l01@00000005Wo001Io`00FOl0000000<0
07Ko00@00000002Io`00VOl000000`00`Ol01@0000000>oo003_o`00kol0000000<00?oM00D00000
003oYP00ojH00?nV00000003003oK00400000000obl00?l_000000<03_l000D0000004go001=o`00
COl00000000308co000500000038o`00b?l00<So000000000`3oo@0010000000olH00?o600000003
0?nE00050000003oJP00ofX00?mZ000000000`3oA@001@000000obT00?lY003o:@00000000<0oaD0
00@000000?l:003o2P0000000`3o2@001@000000oa800?lB003o4P00000000<0obD000D000000?m3
003o@`00od<0000000030?mZ00040000003oV`00oi/0000000<0omD000D000000>So003Xo`00j?l0
0000000309oo00050000001?o`00Col004oo000000000`00o`D0100000000?mM003oG@000003003o
]`0500000000k?l00>co003/o`0000000`00T_l01@00000003Wo000io`00>Ol0000000806P3o00D0
000006X0o`1Z0?l0JP3o000000030;80o`050000003c0?l0l`3o0?<0o`0000000`3o0=@01@000000
o`2V0?l0YP3o0:H000000080o`2200D000000?l0J`3o06/0o`1[000000030?l0G`050000003o05l0
o`1O0?l0G`0000000`3o06/01@000000o`220?l0PP3o088000000080o`2V00D000000?l0e03o0=@0
o`3D000000030?<0o`050000002b0?l0/P3o0;80o`0000000`1Z0?l01@0000006P3o01X0o`0J0?l0
0000008003Wo00D00000002Bo`00T_l009;o00000003003/o`0500000000okL00?ng003o]`000000
0`00oed01@0000000?l5003o1@00o`D000000080Col000D0000009oo002Oo`00Wol0000000030>So
00050000003oe@00omD00?oE000000000`3oV`001@000000ofX00?mZ003oJP0000000080od<000D0
00000?lU003o9@00obD0000000030?lB00050000003o2@00o`T00?l9000000000`3o2P001@000000
oaD00?lE003o5@0000000080obT000D000000?m5003oA@00odD0000000030?mZ00050000003oU@00
oiD00?nE000000000`3oaP001@000000ood00?om003oo@0000000080b?l000D0000008co002<o`00
S?l00000000304go00050000000>o`003_l000ko000000000`00obl01@0000000?m/003oK000of`0
000000800?nV00D00000003og@00omd00?oM00000003003_o`0500000000`Ol00<7o0031o`000000
0`00VOl01000000007Ko001fo`000003001Io`0500000000@_l004;o0012o`0000000`00<Ol01@00
000002Ko000Vo`009_l0000000<002;o00<000000?ooo`3oool00`3oool200000040oooo000P0?oo
o`030000003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?ooo`00803oool00`000000
oooo0?ooo`050?ooo`04000000005_l001Ko000000<000co00D000000009o`002Ol000Wo00000003
000<o`05000000005_l001Ko000Fo`0000000`009ol01000000003oo000oo`000003001No`050000
0000Q?l008Co0024o`0000000`00/?l01@0000000>?o003So`00hol0000000<00?oU00@00000003o
Z@00ojT000000`00ofT01@0000000?lW003o9`00obL0000000<07?l000D00000063o001Po`00H?l0
000000030:;o00040000003So`00hol0000000<0on0000D000000?nY003oZ@00ojT0000000030?mg
00050000003oCP00odh00?m>000000000`3o;00010000000oa@00?lD000000030?l700050000003o
1000o`@00?l4000000000`3o30001@000000oal00?lO003o7`00000000<0ocd000@000000?mW003o
I`0000000`3oV`001@000000omT00?oI003of@00000000<0gol000D00000097o002Ao`00TOl00000
000303co000400000000oad00?lM000000<00?mj00D00000003of@00omT00?oI000000030036o`05
00000000JOl006Wo001Yo`0000000P003ol01@000000A@3o04D0o`150?l0000000<0U03o00D00000
0=X0o`3J0?l0fP3o000000030?l0j0050000003o0;H0o`2f0?l0]P0000000P3o09001@000000o`1f
0?l0MP3o07H0000000<0o`1X00D000000?l0J03o06P0o`1X000000030?l0MP050000003o0900o`2@
0?l0T00000000P3o0;H01@000000o`3X0?l0j03o0>P0000000<0fP3o00D0000009@0o`2D0?l0U03o
0000000304D0o`05000000003ol000oo000?o`0000000P00JOl01@0000000<Ko0036o`00a_l00000
00<00?oI00D00000003oNP00ogX00?mj00000003003o7@050000000lo`00??l003co000000000P2A
o`001@000000gol00=oo003Oo`00000000<0omT000D000000?nK003oV`00oi/0000000030?mW0005
0000003o?@00ocd00?lm000000000P3o7`001@000000o``00?l<003o3000000000<0o`@000D00000
0?l7003o1`00o`L0000000030?lD00050000003o;000ob`00?l/000000000P3oCP001@000000ogL0
0?mg003oM`00000000<0ojT000D000000?oP003oh000on00000000030>?o00050000002Ro`00X_l0
0:;o000000000P1Po`001@0000007?l001co000Lo`00000000<00?lW00D00000003oJ@00ofT00?mY
00000003003oZ@0500000000onD00?oU003oi@0000000P00hol01@0000000;3o002`o`00/?l00000
00<008Co00D00000001No`00G_l005ko00000003000oo`04000000009ol002Oo000000<001Ko00D0
0000000<o`003?l000co000000030009o`05000000003?l000co000<o`0000000`005_l00`000000
oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool01000000001Ko
000Fo`000003000<o`05000000002Ol000Wo0009o`0000000`003?l01@00000001Ko000Fo`005_l0
000000<002Oo00@00000000oo`00?ol000000`00G_l01@00000008Co0024o`00Q?l0000000<00;3o
00D00000003So`00hol00>?o00000003003oi@0400000000ojT00?nY000000<00?mY00D00000003o
9`00obL00?lW0000000301co00050000001Po`00H?l0063o000000000`2Ro`0010000000hol00>?o
000000030?oP00050000003oZ@00ojT00?nY000000000`3oM`001@000000odh00?m>003oCP000000
00<0ob`000@000000?lD003o500000000`3o1`001@000000o`@00?l4003o1000000000<0o``000D0
00000?lO003o7`00oal0000000030?lm00040000003oI`00ofL0000000<0oi/000D000000?oI003o
f@00omT0000000030=oo00050000002Ao`00TOl0097o000000000`0lo`00100000000?lM003o7@00
0003003oNP0500000000omT00?oI003of@0000000`00a_l01@00000006Wo001Yo`00JOl000000080
00oo00D0000004D0o`150?l0A@3o0000000309@0o`050000003J0?l0fP3o0=X0o`0000000`3o0>P0
1@000000o`2f0?l0]P3o0;H000000080o`2@00D000000?l0MP3o07H0o`1f000000030?l0J0050000
003o06P0o`1X0?l0J00000000`3o07H01@000000o`2@0?l0T03o090000000080o`2f00D000000?l0
j03o0>P0o`3X000000030=X0o`050000002D0?l0U03o09@0o`0000000`150?l01@00000000oo000?
o`003ol00000008006Wo00D000000036o`00a_l00<Ko00000003003of@0500000000ogX00?mj003o
NP0000000`00oad01@000000??l003co000lo`0000000080TOl000D000000=oo003Oo`00gol00000
00030?oI00050000003oV`00oi/00?nK000000000`3oI`001@000000ocd00?lm003o?@0000000080
oal000D000000?l<003o3000o``0000000030?l400050000003o1`00o`L00?l7000000000`3o5000
1@000000ob`00?l/003o;00000000080odh000D000000?mg003oM`00ogL0000000030?nY00050000
003oh000on000?oP000000000`3So`001@000000X_l00:;o002Ro`0000000080H?l000D0000001co
000Lo`007?l000000003003o9`0500000000ofT00?mY003oJ@0000000`00ojT01@0000000?oU003o
i@00onD0000000800>?o00D00000002`o`00/?l00;3o000000030024o`0500000000G_l005ko001N
o`0000000`00?ol01000000002Oo000Wo`000003000Fo`05000000003?l000co000<o`0000000`00
2Ol01@00000000co000<o`003?l0000000<001Ko00<000000?ooo`3oool0103oool100000040oooo
000P0?ooo`030000003oool0oooo00D0oooo00@00000000Fo`005_l000000`003?l01@00000000Wo
0009o`002Ol0000000<000co00D00000000Fo`005_l001Ko00000003000Wo`0400000000?ol003oo
000000<005ko00D000000024o`00Q?l008Co00000003002`o`0500000000hol00>?o003So`000000
0`00onD0100000000?nY003oZ@000003003oJ@0500000000obL00?lW003o9`0000000`0Lo`001@00
0000H?l0063o001Po`00000000<0X_l000@000000>?o003So`0000000`3oh0001@000000ojT00?nY
003oZ@00000000<0ogL000D000000?m>003oCP00odh0000000030?l/00040000003o5000oa@00000
00<0o`L000D000000?l4003o1000o`@0000000030?l<00050000003o7`00oal00?lO000000000`3o
?@0010000000ofL00?mW000000030?nK00050000003of@00omT00?oI000000000`3Oo`001@000000
TOl0097o002Ao`00000000<0??l000@00000003o7@00oad000000`00ogX01@0000000?oI003of@00
omT0000000<00<Ko00D00000001Yo`00JOl006Wo00000002000?o`05000000150?l0A@3o04D0o`00
00000`2D0?l01@000000fP3o0=X0o`3J0?l0000000<0o`3X00D000000?l0]P3o0;H0o`2f00000002
0?l0T0050000003o07H0o`1f0?l0MP0000000`3o06P01@000000o`1X0?l0J03o06P0000000<0o`1f
00D000000?l0T03o0900o`2@000000020?l0]P050000003o0>P0o`3X0?l0j00000000`3J0?l01@00
0000U03o09@0o`2D0?l0000000<0A@3o00D00000000?o`003ol000oo00000002001Yo`0500000000
a_l00<Ko0036o`0000000`00omT01@0000000?mj003oNP00ogX0000000<00?lM00D0000003co000l
o`00??l000000002097o00050000003Oo`00gol00=oo000000000`3of@001@000000oi/00?nK003o
V`00000000<0ofL000D000000?lm003o?@00ocd0000000020?lO00050000003o3000o``00?l<0000
00000`3o10001@000000o`L00?l7003o1`00000000<0oa@000D000000?l/003o;000ob`000000002
0?m>00050000003oM`00ogL00?mg000000000`3oZ@001@000000on000?oP003oh000000000<0hol0
00D000000:;o002Ro`00X_l000000002063o00050000000Lo`007?l001co000000000`00obL01@00
00000?mY003oJ@00ofT0000000<00?nY00D00000003oi@00onD00?oU00000002003So`0500000000
/?l00;3o002`o`0000000`00Q?l01@00000005ko001No`00G_l0000000<003oo00@00000000Wo`00
9ol000000`005_l01@00000000co000<o`003?l0000000<000Wo00D00000000<o`003?l000co0000
0003000Fo`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`05
0?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool01000
000002ko000^o`000003000Go`05000000001Ol000Go0005o`0000000`060?l01@0000002`3o00/0
o`0;0?l0000000<02@3o00@000000000o`000?l000000`004_l01@00000002go000]o`00;Ol00000
00<004oo00D00000001io`00NOl007Wo00000003002Zo`0400000000h_l00>;o000000<00?oP00D0
0000003oWP00oih00?nN00000003003oF00500000000oa400?lA003o4@0000000`0go`0010000000
Ool007oo000000030<Go00050000003on000ooP00?oh000000000`3o^`001@000000oh@00?n4003o
Q000000000<0oeD000@000000?l`003o<00000000`3o5@001@000000o`D00?l5003o1@00000000<0
o`0000D000000?l8003o2000o`P0000000030?lM00040000003o?@00ocd0000000<0ofX000D00000
0?nR003oXP00oj80000000030?oU00050000003>o`00c_l00<ko000000000`1jo`00100000008?l0
023o00000003003o?`0500000000oj400?nQ003oX@0000000`00nol01@00000009So002Ho`00V?l0
0000008003Wo00D000000240o`0Q0?l08@3o0000000307D0o`05000000300?l0`03o0<00o`000000
0`3o0?d01@000000o`370?l0a`3o0<L000000080o`2M00D000000?l0P@3o0840o`21000000030?l0
L`050000003o07<0o`1c0?l0L`0000000`3o08401@000000o`2M0?l0W@3o09d000000080o`3700D0
00000?l0o@3o0?d0o`3m000000030<00o`050000001e0?l0M@3o07D0o`0000000`0Q0?l01@000000
03Wo000io`00>Ol00000008009So00D00000003ko`00nol00?_o00000003003oX@0500000000ocl0
0?lo003o?`0000000`0Po`001@000000N_l007[o001jo`0000000080c_l000D000000?oU003oi@00
onD0000000030?nR00050000003oJP00ofX00?mZ000000000`3o?@001@000000oad00?lM003o7@00
00000080o`P000D000000?l0003o0000o`00000000030?l500050000003o5@00oaD00?lE00000000
0`3o<0001@000000oeD00?mE003oE@0000000080oh@000D000000?nk003o^`00ok/0000000030?oh
000500000035o`00aOl00<Go000000000`1oo`001@000000=ol003Oo000go`00000000800?lA00D0
0000003oF000oeP00?mH00000003003oWP0500000000on000?oP003oh00000000`00h_l01@000000
0:[o002Zo`00Z_l00000008007Wo00D00000001?o`00Col004oo00000003000]o`05000000004_l0
01;o000Bo`0000000`000?l0100000002@3o00T0o`00000300/0o`05000000060?l01P3o00H0o`00
00000`001Ol01@00000001Oo000Go`005ol0000000<002ko00<000000?ooo`3oool0103oool10000
0040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000000^o`00;_l000000`005ol01@00
000000Go0005o`001Ol0000000<01P3o00D0000000/0o`0;0?l02`3o0000000300T0o`0400000000
0?l0003o000000<001;o00D00000000]o`00;Ol002go00000003001?o`0500000000NOl007Wo001i
o`0000000`00Z_l0100000000>;o003Ro`000003003oh00500000000oih00?nN003oWP0000000`00
oeP01@0000000?lA003o4@00oa40000000<0=ol000@0000007oo001oo`0000000`35o`001@000000
ooP00?oh003on000000000<0ok/000D000000?n4003oQ000oh@0000000030?mE00040000003o<000
oc00000000<0oaD000D000000?l5003o1@00o`D0000000030?l000050000003o2000o`P00?l80000
00000`3o7@0010000000ocd00?lm000000030?mZ00050000003oXP00oj800?nR000000000`3oi@00
1@000000c_l00<ko003>o`00000000<0N_l000@00000023o000Po`0000000`00ocl01@0000000?nQ
003oX@00oj40000000<00?_o00D00000002Ho`00V?l009So00000002000io`050000000Q0?l08@3o
0240o`0000000`1e0?l01@000000`03o0<00o`300?l0000000<0o`3m00D000000?l0a`3o0<L0o`37
000000020?l0W@050000003o0840o`210?l0P@0000000`3o07<01@000000o`1c0?l0L`3o07<00000
00<0o`2100D000000?l0W@3o09d0o`2M000000020?l0a`050000003o0?d0o`3m0?l0o@0000000`30
0?l01@000000M@3o07D0o`1e0?l0000000<08@3o00D00000000io`00>Ol003Wo00000002002Ho`05
00000000nol00?_o003ko`0000000`00oj401@0000000?lo003o?`00ocl0000000<08?l000D00000
07[o001jo`00N_l0000000020<ko00050000003oi@00onD00?oU000000000`3oXP001@000000ofX0
0?mZ003oJP00000000<0ocd000D000000?lM003o7@00oad0000000020?l800050000003o0000o`00
0?l0000000000`3o1@001@000000oaD00?lE003o5@00000000<0oc0000D000000?mE003oE@00oeD0
000000020?n400050000003o^`00ok/00?nk000000000`3on0001@000000aOl00<Go0035o`000000
00<0Ool000D0000003Oo000go`00=ol000000002003o4@0500000000oeP00?mH003oF00000000`00
oih01@0000000?oP003oh000on00000000<00>;o00D00000002Zo`00Z_l00:[o00000002001io`05
00000000Col004oo001?o`0000000`00;Ol01@00000001;o000Bo`004_l0000000<0003o00@00000
00T0o`090?l000000`0;0?l01@0000001P3o00H0o`060?l0000000<000Go00D00000000Go`005ol0
01Oo00000003000^o`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo
0?ooo`050?ooo`0400000000;_l002ko000000<001Oo00D000000005o`001Ol000Go0000000300H0
o`050000000;0?l02`3o00/0o`0000000`090?l010000000003o0000o`000003000Bo`0500000000
;Ol002go000]o`0000000`00Col01@00000007Wo001io`00NOl0000000<00:[o00@00000003Ro`00
h_l000000`00on001@0000000?nN003oWP00oih0000000<00?mH00D00000003o4@00oa400?lA0000
000303Oo00040000001oo`00Ool0000000<0aOl000D000000?oh003on000ooP0000000030?nk0005
0000003oQ000oh@00?n4000000000`3oE@0010000000oc000?l`000000030?lE00050000003o1@00
o`D00?l5000000000`3o00001@000000o`P00?l8003o2000000000<0oad000@000000?lm003o?@00
00000`3oJP001@000000oj800?nR003oXP00000000<0onD000D000000<ko003>o`00c_l000000003
07[o00040000000Po`008?l0000000<00?lo00D00000003oX@00oj400?nQ00000003003ko`050000
0000V?l009So002Ho`0000000P00>Ol01@0000008@3o0240o`0Q0?l0000000<0M@3o00D000000<00
o`300?l0`03o000000030?l0o@050000003o0<L0o`370?l0a`0000000P3o09d01@000000o`210?l0
P@3o0840000000<0o`1c00D000000?l0L`3o07<0o`1c000000030?l0P@050000003o09d0o`2M0?l0
W@0000000P3o0<L01@000000o`3m0?l0o@3o0?d0000000<0`03o00D0000007D0o`1e0?l0M@3o0000
00030240o`0500000000>Ol003Wo000io`0000000P00V?l01@0000000?_o003ko`00nol0000000<0
0?nQ00D00000003o?`00ocl00?lo00000003023o00050000001jo`00N_l007[o000000000P3>o`00
1@000000onD00?oU003oi@00000000<0oj8000D000000?mZ003oJP00ofX0000000030?lm00050000
003o7@00oad00?lM000000000P3o20001@000000o`000?l0003o0000000000<0o`D000D000000?lE
003o5@00oaD0000000030?l`00050000003oE@00oeD00?mE000000000P3oQ0001@000000ok/00?nk
003o^`00000000<0ooP000D000000<Go0035o`00aOl00000000307oo00050000000go`00=ol003Oo
000000000P00oa401@0000000?mH003oF000oeP0000000<00?nN00D00000003oh000on000?oP0000
0003003Ro`0500000000Z_l00:[o002Zo`0000000P00NOl01@00000004oo001?o`00Col0000000<0
02go00D00000000Bo`004_l001;o000000030000o`04000000090?l02@3o000000<02`3o00D00000
00H0o`060?l01P3o000000030005o`05000000005ol001Oo000Go`0000000`00;_l00`000000oooo
0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo0000084000001P3o
ool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000001Xo`00J?l000000`00
A?l01@00000002?o000So`008ol0000000<000So00D0000000/0o`0;0?l02`3o0000000301T0o`04
0000000N0?l07P3o000000<06`3o00D0000000l0o`0?0?l03`3o000000030005o`05000000008ol0
02?o000So`0000000`00BOl01000000007Wo001io`000003002`o`0500000000k_l00>ko003^o`00
00000`00old01@0000000?n5003oQ@00ohD0000000<00?li00@0000001?o000Co`0000000`1Po`00
1@000000Zol00:_o002[o`00000000<0lol000D000000?o:003obP00olX0000000030?n?00040000
003oG000oe`0000000<0oc<000D000000?lE003o5@00oaD0000000030?l400050000003o0000o`00
0?l0000000000`3o200010000000oal00?lO000000030?m300050000003oM000og@00?md00000000
0`3o/@001@000000ooT00?oi003on@00000000<0]?l000@0000005[o001Jo`0000000`00o`D01@00
00000?mZ003oJP00ofX0000000<00?oA00D000000038o`00b?l00<So00000002001So`0500000000
0ol000?o0003o`0000000`1E0?l01@000000YP3o0:H0o`2V0?l0000000<0k03o00D000000?l0f03o
0=P0o`3H000000020?l0Z`050000003o08d0o`2=0?l0S@0000000`3o07d01@000000o`1m0?l0O@3o
07d0000000<0o`2=00D000000?l0Z`3o0:/0o`2[000000020?l0f0050000003/0?l0k03o0>`0o`00
00000`2V0?l01@000000E@3o05D0o`1E0?l0000000<000?o00D00000001So`00Hol006?o00000002
0038o`0500000000om400?oA003od@0000000`00ofX01@0000000?l5003o1@00o`D0000000<0F_l0
00D000000;Co002do`00]?l0000000020?oi00050000003o/@00ok400?na000000000`3oM0001@00
0000od<00?m3003o@`00000000<0oal000D000000?l8003o2000o`P0000000020?l000050000003o
1000o`@00?l4000000000`3o5@001@000000oc<00?lc003o<`00000000<0oe`000D000000?n?003o
S`00ohl0000000020?o:00050000003co`00lol00??o000000000`2[o`001@000000H?l0063o001P
o`00000000<04ol000D00000003o>@00ocT00?li00000002003oQ@0500000000old00?o=003oc@00
00000`00k_l01@0000000;3o002`o`00/?l0000000<007Wo00D000000019o`00BOl004Wo00000002
000So`05000000001Ol000Go0005o`0000000`0?0?l01@0000006`3o01/0o`0K0?l0000000<07P3o
00@0000001T0o`0I0?l000000`0;0?l01@00000000So0008o`002?l0000000<002?o00D000000014
o`00A?l004Co00000003001Xo`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`00
0000oooo0?ooo`050?ooo`0400000000J?l006So000000<004Co00D00000000So`008ol002?o0000
00030008o`050000000;0?l02`3o00/0o`0000000`0I0?l0100000007P3o01h0o`00000301/0o`05
0000000?0?l03`3o00l0o`0000000`001Ol01@00000002?o000So`008ol0000000<004Wo00@00000
001io`00NOl000000`00/?l01@0000000>ko003^o`00k_l0000000<00?o=00D00000003oQ@00ohD0
0?n500000003003o>@040000000Co`004ol0000000<0H?l000D000000:_o002[o`00Zol000000003
0??o00050000003obP00olX00?o:000000000`3oS`0010000000oe`00?mL000000030?lc00050000
003o5@00oaD00?lE000000000`3o10001@000000o`000?l0003o0000000000<0o`P000@000000?lO
003o7`0000000`3o@`001@000000og@00?md003oM000000000<0ok4000D000000?oi003on@00ooT0
000000030;Co00040000001Jo`00F_l0000000<00?l500D00000003oJP00ofX00?mZ00000003003o
d@0500000000b?l00<So0038o`0000000P00Hol01@00000000?o0003o`000ol0000000<0E@3o00D0
00000:H0o`2V0?l0YP3o000000030>`0o`050000003o0=P0o`3H0?l0f00000000P3o0:/01@000000
o`2=0?l0S@3o08d0000000<0o`1m00D000000?l0O@3o07d0o`1m000000030?l0S@050000003o0:/0
o`2[0?l0Z`0000000P3o0=P01@000000k03o0>`0o`3/0?l0000000<0YP3o00D0000005D0o`1E0?l0
E@3o000000030003o`0500000000Hol006?o001So`0000000P00b?l01@0000000?oA003od@00om40
000000<00?mZ00D00000003o1@00o`D00?l50000000305[o00050000002do`00]?l00;Co00000000
0P3on@001@000000ok400?na003o/@00000000<0og@000D000000?m3003o@`00od<0000000030?lO
00050000003o2000o`P00?l8000000000P3o00001@000000o`@00?l4003o1000000000<0oaD000D0
00000?lc003o<`00oc<0000000030?mL00050000003oS`00ohl00?n?000000000P3obP001@000000
lol00??o003co`00000000<0Zol000D00000063o001Po`00H?l00000000301?o000500000000ocT0
0?li003o>@0000000P00ohD01@0000000?o=003oc@00old0000000<00>ko00D00000002`o`00/?l0
0;3o00000003001io`0500000000BOl004Wo0019o`0000000P008ol01@00000000Go0005o`001Ol0
000000<03`3o00D0000001/0o`0K0?l06`3o0000000301h0o`040000000I0?l06@3o000000<02`3o
00D000000008o`002?l000So00000003000So`0500000000A?l004Co0014o`0000000`00J?l00`00
0000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo00000840
00001P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000002lo`00_?l0
00000`00SOl01@000000067o001Qo`00HOl0000000<003Wo00D00000000Fo`005_l001Ko00000003
00H0o`040000000L0?l0703o000000<0:P3o00D0000002h0o`0^0?l0;P3o0000000302T0o`050000
000J0?l06P3o01X0o`0000000`010?l01000000002;o000Ro`000003001?o`0500000000QOl008Go
0025o`0000000`00a?l01@0000000?of003omP00ooH0000000<00?n[00@00000003oG000oe`00000
0`00o`/01@000000A_l004Ko0016o`00000000<0U_l000D000000>?o003So`00hol0000000030?oF
00040000003oUP00oiH0000000<0of0000D000000?le003o=@00ocD0000000030?lF00050000003o
1000o`@00?l4000000000`3o000010000000o``00?l<000000030?lU00050000003oCP00odh00?m>
000000000`3oQ0001@000000olL00?o7003oa`00000000<0j_l000@00000097o002Ao`0000000`0a
o`001@0000000?ld003o=000oc@0000000<00?nN00D00000003fo`00m_l00?Ko00000002002=o`05
00000000:?l002So000Xo`0000000`0f0?l01@000000R`3o08/0o`2;0?l0000000<0e`3o00D00000
0?l0jP3o0>X0o`3Z000000020?l0^P050000003o09T0o`2I0?l0V@0000000`3o08P01@000000o`28
0?l0R03o08P0000000<0o`2I00D000000?l0^P3o0;X0o`2j000000020?l0jP050000003G0?l0e`3o
0=L0o`0000000`2;0?l01@000000=P3o03H0o`0f0?l0000000<002So00D00000002=o`00SOl008go
00000002003fo`0500000000oih00?nN003oWP0000000`00oc@01@000000<Ol0037o000ao`000000
00<0TOl000D000000>[o003Zo`00j_l0000000020?o700050000003oQ000oh@00?n4000000000`3o
CP001@000000obD00?lU003o9@00000000<0o``000D000000?l0003o0000o`00000000020?l40005
0000003o5P00oaH00?lF000000000`3o=@001@000000of000?mP003oH000000000<0oiH000D00000
0?oF003oeP00omH0000000020>?o00050000002Fo`00U_l009Ko000000000`16o`001@0000000?l;
003o2`00o`/0000000<00?mL00D00000003oZ`00oj/00?n[00000002003omP0500000000a?l00<Co
0034o`0000000`00QOl01@00000004oo001?o`00Col0000000<002;o00D000000040o`010?l00@3o
0000000201X0o`050000000Y0?l0:@3o02T0o`0000000`0^0?l01@000000:P3o02X0o`0Z0?l00000
00<0703o00@0000000H0o`060?l000000`005_l01@00000003Wo000io`00>Ol0000000<0067o00D0
0000002=o`00SOl008go00000003002lo`030000003oool0oooo00@0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooo`0400000000_?l00;co000000<008go00D00000001Qo`00HOl0
067o00000003000io`05000000005_l001Ko000Fo`0000000`060?l010000000703o01`0o`000003
02X0o`050000000^0?l0;P3o02h0o`0000000`0Y0?l01@0000006P3o01X0o`0J0?l0000000<00@3o
00@00000000Ro`008_l000000`00Col01@00000008Go0025o`00QOl0000000<00<Co00D00000003o
mP00ooH00?of00000003003oZ`0400000000oe`00?mL000000<00?l;00D0000004Ko0016o`00A_l0
0000000309Ko00050000003So`00hol00>?o000000000`3oeP0010000000oiH00?nF000000030?mP
00050000003o=@00ocD00?le000000000`3o5P001@000000o`@00?l4003o1000000000<0o`0000@0
00000?l<003o300000000`3o9@001@000000odh00?m>003oCP00000000<0oh@000D000000?o7003o
a`00olL0000000030>[o00040000002Ao`00TOl0000000<0<Ol000D00000003o=000oc@00?ld0000
0003003oWP0500000000m_l00?Ko003fo`0000000P00SOl01@00000002So000Xo`00:?l0000000<0
=P3o00D0000008/0o`2;0?l0R`3o000000030=L0o`050000003o0>X0o`3Z0?l0jP0000000P3o0;X0
1@000000o`2I0?l0V@3o09T0000000<0o`2800D000000?l0R03o08P0o`28000000030?l0V@050000
003o0;X0o`2j0?l0^P0000000P3o0>X01@000000e`3o0=L0o`3G0?l0000000<0R`3o00D0000003H0
o`0f0?l0=P3o00000003000Xo`0500000000SOl008go002=o`0000000P00m_l01@0000000?nN003o
WP00oih0000000<00?ld00D00000037o000ao`00<Ol000000003097o00050000003Zo`00j_l00>[o
000000000P3oa`001@000000oh@00?n4003oQ000000000<0odh000D000000?lU003o9@00obD00000
00030?l<00050000003o0000o`000?l0000000000P3o10001@000000oaH00?lF003o5P00000000<0
ocD000D000000?mP003oH000of00000000030?nF00050000003oeP00omH00?oF000000000P3So`00
1@000000U_l009Ko002Fo`00000000<0A_l000D00000003o2`00o`/00?l;00000003003oG0050000
0000oj/00?n[003oZ`0000000P00ooH01@0000000<Co0034o`00a?l0000000<008Go00D00000001?
o`00Col004oo00000003000Ro`05000000010?l00@3o0040o`0000000P0J0?l01@000000:@3o02T0
o`0Y0?l0000000<0;P3o00D0000002X0o`0Z0?l0:P3o0000000301`0o`04000000060?l01P3o0000
00<001Ko00D00000000io`00>Ol003Wo00000003001Qo`0500000000SOl008go002=o`0000000`00
_?l00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool0
100000000;co002lo`000003002=o`0500000000HOl0067o001Qo`0000000`00>Ol01@00000001Ko
000Fo`005_l0000000<01P3o00@0000001`0o`0L0?l000000`0Z0?l01@000000;P3o02h0o`0^0?l0
000000<0:@3o00D0000001X0o`0J0?l06P3o000000030040o`04000000008_l002;o000000<004oo
00D000000025o`00QOl008Go000000030034o`0500000000ooH00?of003omP0000000`00oj/01000
00000?mL003oG0000003003o2`0500000016o`00A_l004Ko000000000`2Fo`001@000000hol00>?o
003So`00000000<0omH000@000000?nF003oUP0000000`3oH0001@000000ocD00?le003o=@000000
00<0oaH000D000000?l4003o1000o`@0000000030?l000040000003o3000o``0000000<0obD000D0
00000?m>003oCP00odh0000000030?n400050000003oa`00olL00?o7000000000`3Zo`0010000000
TOl0097o00000003037o000500000000oc@00?ld003o=00000000`00oih01@0000000?Ko003fo`00
m_l00000008008go00D00000000Xo`00:?l002So0000000303H0o`050000002;0?l0R`3o08/0o`00
00000`3G0?l01@000000o`3Z0?l0jP3o0>X000000080o`2j00D000000?l0V@3o09T0o`2I00000003
0?l0R0050000003o08P0o`280?l0R00000000`3o09T01@000000o`2j0?l0^P3o0;X000000080o`3Z
00D000000=L0o`3G0?l0e`3o0000000308/0o`050000000f0?l0=P3o03H0o`0000000`00:?l01@00
000008go002=o`00SOl0000000800?Ko00D00000003oWP00oih00?nN00000003003o=0050000000a
o`00<Ol0037o000000000`2Ao`001@000000j_l00>[o003Zo`0000000080olL000D000000?n4003o
Q000oh@0000000030?m>00050000003o9@00obD00?lU000000000`3o30001@000000o`000?l0003o
000000000080o`@000D000000?lF003o5P00oaH0000000030?le00050000003oH000of000?mP0000
00000`3oUP001@000000omH00?oF003oeP0000000080hol000D0000009Ko002Fo`00U_l000000003
04Ko000500000000o`/00?l;003o2`0000000`00oe`01@0000000?n[003oZ`00oj/0000000800?of
00D000000034o`00a?l00<Co000000030025o`0500000000Col004oo001?o`0000000`008_l01@00
00000@3o0040o`010?l0000000806P3o00D0000002T0o`0Y0?l0:@3o0000000302h0o`050000000Z
0?l0:P3o02X0o`0000000`0L0?l0100000001P3o00H0o`000003000Fo`0500000000>Ol003Wo000i
o`0000000`00HOl01@00000008go002=o`00SOl0000000<00;co00<000000?ooo`3oool0103oool1
00000040oooo000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?oo
o`00803oool00`000000oooo0?ooo`050?ooo`0400000000omh00?oN000000<00>co00D00000002h
o`00^?l00;So000000030025o`0500000000EOl005Go001Eo`0000000`00:_l01000000000Go0005
o`00000301P0o`050000000^0?l0;P3o02h0o`0000000`0j0?l01@000000>`3o03/0o`0k0?l00000
00<0<P3o00@0000001d0o`0M0?l000000`000_l01@00000002co000/o`00;?l0000000<006;o00D0
0000002Po`00X?l00:3o00000003003Wo`0400000000olX00?o:000000<00?mh00D00000003o8`00
ob<00?lS0000000303;o000500000027o`00Qol008Oo000000000`3Go`0010000000omd00?oM0000
00030?nJ00050000003oH@00of400?mQ000000000`3o=@001@000000oaD00?lE003o5@00000000<0
o`D000@000000?l4003o100000000`3o4P001@000000oc400?la003o<@00000000<0oel000D00000
0?nK003oV`00oi/0000000030?oU000400000034o`00a?l0000000<0IOl000D00000003o0@00o`40
0?l100000003003oK@0500000000om/00?oK003of`0000000P00]ol01@00000004go001=o`00COl0
000000<05@3o00D000000700o`1`0?l0L03o000000030<00o`050000003o0?`0o`3l0?l0o0000000
0P3o0<T01@000000o`2V0?l0YP3o0:H0000000<0o`2D00D000000?l0U03o09@0o`2D000000030?l0
YP050000003o0<T0o`390?l0b@0000000P3o0?`01@000000`03o0<00o`300?l0000000<0L03o00D0
000001D0o`0E0?l05@3o00000003001=o`0500000000]ol00;Oo002go`0000000P00om/01@000000
0?m]003oK@00ofd0000000<00?l100D0000006Go001Uo`00IOl0000000030<Co00050000003oi@00
onD00?oU000000000P3oV`001@000000oel00?mO003oG`00000000<0oc4000D000000?lB003o4P00
oa80000000030?l400050000003o1@00o`D00?l5000000000P3o5@001@000000ocD00?le003o=@00
000000<0of4000D000000?nJ003oVP00oiX0000000030?oM00050000003Go`00eol00=Oo00000000
0P27o`001@000000<_l003;o000bo`00000000<00?lS00D00000003oN000ogP00?mh00000003003o
bP0500000000iol00>Oo003Wo`0000000P00X?l01@00000006;o001Ro`00H_l0000000<002co00D0
00000002o`000_l000;o0000000301d0o`050000000b0?l0<P3o0380o`0000000P0k0?l01@000000
>P3o03X0o`0j0?l0000000<0;P3o00D0000001P0o`0H0?l0603o000000030005o`0400000000:_l0
02[o000000<005Go00D000000025o`00QOl008Go00000003002ho`0500000000k?l00>co003/o`00
00000`00omh00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool0
1@3oool0100000000?oN003ogP000003003/o`0500000000^?l00;So002ho`0000000`00QOl01@00
000005Go001Eo`00EOl0000000<002[o00@000000005o`001Ol000000`0H0?l01@000000;P3o02h0
o`0^0?l0000000<0>P3o00D0000003/0o`0k0?l0>`3o000000030380o`040000000M0?l07@3o0000
00<000;o00D00000000/o`00;?l002co00000003001Ro`0500000000X?l00:3o002Po`0000000`00
iol0100000000?o:003obP000003003oN00500000000ob<00?lS003o8`0000000`0bo`001@000000
Qol008Oo0027o`00000000<0eol000@000000?oM003og@0000000`3oVP001@000000of400?mQ003o
H@00000000<0ocD000D000000?lE003o5@00oaD0000000030?l500040000003o1000o`@0000000<0
oa8000D000000?la003o<@00oc40000000030?mO00050000003oV`00oi/00?nK000000000`3oi@00
10000000a?l00<Co0000000306Go000500000000o`400?l1003o0@0000000`00ofd01@0000000?oK
003of`00om/0000000800;Oo00D00000001=o`00COl004go0000000301D0o`050000001`0?l0L03o
0700o`0000000`300?l01@000000o`3l0?l0o03o0?`000000080o`3900D000000?l0YP3o0:H0o`2V
000000030?l0U0050000003o09@0o`2D0?l0U00000000`3o0:H01@000000o`390?l0b@3o0<T00000
0080o`3l00D000000<00o`300?l0`03o000000030700o`050000000E0?l05@3o01D0o`0000000`00
COl01@0000000;Oo002go`00]ol0000000800?oK00D00000003oK@00ofd00?m]00000003003o0@05
0000001Uo`00IOl006Go000000000`34o`001@000000onD00?oU003oi@0000000080oi/000D00000
0?mO003oG`00oel0000000030?la00050000003o4P00oa800?lB000000000`3o10001@000000o`D0
0?l5003o1@0000000080oaD000D000000?le003o=@00ocD0000000030?mQ00050000003oVP00oiX0
0?nJ000000000`3og@001@000000eol00=Oo003Go`0000000080Qol000D0000003;o000bo`00<_l0
00000003003o8`0500000000ogP00?mh003oN00000000`00olX01@0000000>Oo003Wo`00iol00000
00800:3o00D00000001Ro`00H_l006;o00000003000/o`05000000000_l000;o0002o`0000000`0M
0?l01@000000<P3o0380o`0b0?l000000080>`3o00D0000003X0o`0j0?l0>P3o0000000302h0o`05
0000000H0?l0603o01P0o`0000000`001Ol01000000002[o000Zo`000003001Eo`0500000000QOl0
08Go0025o`0000000`00^?l01@0000000>co003/o`00k?l0000000<00?oN00<000000?ooo`3oool0
103oool100000040oooo000?0?ooo`030;md00000000000000@0000000<0B9cO0?ooo`3oool01`3o
ool00`000000oooo0?ooo`050?ooo`0400000000omh00?oN000000<00>co00D00000002ho`00^?l0
0;So000000030025o`0500000000EOl005Go001Eo`0000000`00:_l01000000000Go0005o`000003
01P0o`050000000^0?l0;P3o02h0o`0000000`0j0?l01@000000>`3o03/0o`0k0?l0000000<0<P3o
00@0000001d0o`0M0?l000000`000_l01@00000002co000/o`00;?l0000000<006;o00D00000002P
o`00X?l00:3o00000003003Wo`0400000000olX00?o:000000<00?mh00D00000003o8`00ob<00?lS
0000000303;o000500000027o`00Qol008Oo000000000`3Go`0010000000omd00?oM000000030?nJ
00050000003oH@00of400?mQ000000000`3o=@001@000000oaD00?lE003o5@00000000<0o`D000@0
00000?l4003o100000000`3o4P001@000000oc400?la003o<@00000000<0oel000D000000?nK003o
V`00oi/0000000030?oU000400000034o`00a?l0000000<0IOl000D00000003o0@00o`400?l10000
0003003oK@0500000000om/00?oK003of`0000000P00]ol01@00000004go001=o`00COl0000000<0
5@3o00D000000700o`1`0?l0L03o000000030<00o`050000003o0?`0o`3l0?l0o00000000P3o0<T0
1@000000o`2V0?l0YP3o0:H0000000<0o`2D00D000000?l0U03o09@0o`2D000000030?l0YP050000
003o0<T0o`390?l0b@0000000P3o0?`01@000000`03o0<00o`300?l0000000<0L03o00D0000001D0
o`0E0?l05@3o00000003001=o`0500000000]ol00;Oo002go`0000000P00om/01@0000000?m]003o
K@00ofd0000000<00?l100D0000006Go001Uo`00IOl0000000030<Co00050000003oi@00onD00?oU
000000000P3oV`001@000000oel00?mO003oG`00000000<0oc4000D000000?lB003o4P00oa800000
00030?l400050000003o1@00o`D00?l5000000000P3o5@001@000000ocD00?le003o=@00000000<0
of4000D000000?nJ003oVP00oiX0000000030?oM00050000003Go`00eol00=Oo000000000P27o`00
1@000000<_l003;o000bo`00000000<00?lS00D00000003oN000ogP00?mh00000003003obP050000
0000iol00>Oo003Wo`0000000P00X?l01@00000006;o001Ro`00H_l0000000<002co00D000000002
o`000_l000;o0000000301d0o`050000000b0?l0<P3o0380o`0000000P0k0?l01@000000>P3o03X0
o`0j0?l0000000<0;P3o00D0000001P0o`0H0?l0603o000000030005o`0400000000:_l002[o0000
00<005Go00D000000025o`00QOl008Go00000003002ho`0500000000k?l00>co003/o`0000000`00
omh00`000000oooo0?ooo`040?ooo`4000000@3oool00100oooo00<0gia80000000000000`000000
0`00B9`0gooo0?ooo`070?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@000001
0?ooo`004@3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo00L0oooo00<000000?oo
o`3oool01@3oool0100000000?md003oM0000003003oZP0500000000on800?oR003ohP0000000`00
i?l01@0000000:co002/o`00[?l0000000<007Ko00@000000013o`00@ol000000`005_l01@000000
3`3o00l0o`0?0?l0000000<0;03o00D0000003l0o`0o0?l0?`3o0000000304L0o`04000000130?l0
@`3o000000<0<`3o00D0000001L0o`0G0?l05`3o00000003000Ao`0500000000AOl004Go0015o`00
00000`00Q?l0100000000<co003<o`000003003oh`0500000000ohh00?n>003oSP0000000`00ocD0
1@0000009?l002Co000To`00000000<0OOl000@000000=;o003Bo`0000000`3og`001@000000oiX0
0?nJ003oVP00000000<0of0000D000000?lc003o<`00oc<0000000030?lE00040000003o1`00o`L0
000000<0o`T000D000000?lM003o7@00oad0000000030?m100050000003oMP00ogH00?mf00000000
0`3o^`0010000000lol00??o0000000309Go00050000000_o`00;ol002oo000000000`00ocd01@00
00000?n^003o[P00ojh0000000800>3o00D00000001bo`00L_l007;o00000003000:o`050000001E
0?l0E@3o05D0o`0000000`2Z0?l01@000000l03o0?00o`3`0?l000000080o`3H00D000000?l0/P3o
0;80o`2b000000030?l0W`050000003o09l0o`2O0?l0W`0000000`3o0;801@000000o`3H0?l0f03o
0=P000000080l03o00D000000:X0o`2Z0?l0ZP3o0000000305D0o`05000000002_l000[o000:o`00
00000`00L_l01@0000000>3o003Po`00h?l0000000800?n^00D00000003o?@00ocd00?lm00000003
02oo00050000002Eo`00UOl009Go000000000`3co`001@000000ok/00?nk003o^`0000000080ogH0
00D000000?m1003o@@00od40000000030?lM00050000003o2@00o`T00?l9000000000`3o1`001@00
0000oaD00?lE003o5@0000000080oc<000D000000?mP003oH000of00000000030?nJ00050000003o
g`00oml00?oO000000000`3Bo`001@000000OOl007go001mo`00000000809?l000D00000003o=@00
ocD00?le00000003003oSP0500000000on<00?oS003oh`0000000`00c?l01@00000008Co0024o`00
Q?l00000008004Go00D00000000Ao`004Ol0017o0000000301L0o`050000000c0?l0<`3o03<0o`00
00000`130?l01@000000A`3o04L0o`170?l000000080?`3o00D0000002`0o`0/0?l0;03o00000003
00l0o`05000000005_l001Ko000Fo`0000000`00@ol01000000007Ko001fo`000003002/o`050000
0000i?l00>Co003To`0000000`00on801@0000000?nZ003oZP00ojX0000000<00?md00<000000?oo
o`3oool0103oool100000040oooo000B0?ooo`030=nLB01d_ol0oooo00/0oooo00<000000?ooo`3o
ool01@3oool0100000000?md003oM0000003003oZP0500000000on800?oR003ohP0000000`00i?l0
1@0000000:co002/o`00[?l0000000<007Ko00@000000013o`00@ol000000`005_l01@0000003`3o
00l0o`0?0?l0000000<0;03o00D0000003l0o`0o0?l0?`3o0000000304L0o`04000000130?l0@`3o
000000<0<`3o00D0000001L0o`0G0?l05`3o00000003000Ao`0500000000AOl004Go0015o`000000
0`00Q?l0100000000<co003<o`000003003oh`0500000000ohh00?n>003oSP0000000`00ocD01@00
00009?l002Co000To`00000000<0OOl000@000000=;o003Bo`0000000`3og`001@000000oiX00?nJ
003oVP00000000<0of0000D000000?lc003o<`00oc<0000000030?lE00040000003o1`00o`L00000
00<0o`T000D000000?lM003o7@00oad0000000030?m100050000003oMP00ogH00?mf000000000`3o
^`0010000000lol00??o0000000309Go00050000000_o`00;ol002oo000000000`00ocd01@000000
0?n^003o[P00ojh0000000800>3o00D00000001bo`00L_l007;o00000003000:o`050000001E0?l0
E@3o05D0o`0000000`2Z0?l01@000000l03o0?00o`3`0?l000000080o`3H00D000000?l0/P3o0;80
o`2b000000030?l0W`050000003o09l0o`2O0?l0W`0000000`3o0;801@000000o`3H0?l0f03o0=P0
00000080l03o00D000000:X0o`2Z0?l0ZP3o0000000305D0o`05000000002_l000[o000:o`000000
0`00L_l01@0000000>3o003Po`00h?l0000000800?n^00D00000003o?@00ocd00?lm0000000302oo
00050000002Eo`00UOl009Go000000000`3co`001@000000ok/00?nk003o^`0000000080ogH000D0
00000?m1003o@@00od40000000030?lM00050000003o2@00o`T00?l9000000000`3o1`001@000000
oaD00?lE003o5@0000000080oc<000D000000?mP003oH000of00000000030?nJ00050000003og`00
oml00?oO000000000`3Bo`001@000000OOl007go001mo`00000000809?l000D00000003o=@00ocD0
0?le00000003003oSP0500000000on<00?oS003oh`0000000`00c?l01@00000008Co0024o`00Q?l0
0000008004Go00D00000000Ao`004Ol0017o0000000301L0o`050000000c0?l0<`3o03<0o`000000
0`130?l01@000000A`3o04L0o`170?l000000080?`3o00D0000002`0o`0/0?l0;03o0000000300l0
o`05000000005_l001Ko000Fo`0000000`00@ol01000000007Ko001fo`000003002/o`0500000000
i?l00>Co003To`0000000`00on801@0000000?nZ003oZP00ojX0000000<00?md00<000000?ooo`3o
ool0103oool100000040oooo000C0?ooo`030;mdB02Lgol0oooo00X0oooo00<000000?ooo`3oool0
1@3oool0100000000?md003oM0000003003oZP0500000000on800?oR003ohP0000000`00i?l01@00
00000:co002/o`00[?l0000000<007Ko00@000000013o`00@ol000000`005_l01@0000003`3o00l0
o`0?0?l0000000<0;03o00D0000003l0o`0o0?l0?`3o0000000304L0o`04000000130?l0@`3o0000
00<0<`3o00D0000001L0o`0G0?l05`3o00000003000Ao`0500000000AOl004Go0015o`0000000`00
Q?l0100000000<co003<o`000003003oh`0500000000ohh00?n>003oSP0000000`00ocD01@000000
9?l002Co000To`00000000<0OOl000@000000=;o003Bo`0000000`3og`001@000000oiX00?nJ003o
VP00000000<0of0000D000000?lc003o<`00oc<0000000030?lE00040000003o1`00o`L0000000<0
o`T000D000000?lM003o7@00oad0000000030?m100050000003oMP00ogH00?mf000000000`3o^`00
10000000lol00??o0000000309Go00050000000_o`00;ol002oo000000000`00ocd01@0000000?n^
003o[P00ojh0000000800>3o00D00000001bo`00L_l007;o00000003000:o`050000001E0?l0E@3o
05D0o`0000000`2Z0?l01@000000l03o0?00o`3`0?l000000080o`3H00D000000?l0/P3o0;80o`2b
000000030?l0W`050000003o09l0o`2O0?l0W`0000000`3o0;801@000000o`3H0?l0f03o0=P00000
0080l03o00D000000:X0o`2Z0?l0ZP3o0000000305D0o`05000000002_l000[o000:o`0000000`00
L_l01@0000000>3o003Po`00h?l0000000800?n^00D00000003o?@00ocd00?lm0000000302oo0005
0000002Eo`00UOl009Go000000000`3co`001@000000ok/00?nk003o^`0000000080ogH000D00000
0?m1003o@@00od40000000030?lM00050000003o2@00o`T00?l9000000000`3o1`001@000000oaD0
0?lE003o5@0000000080oc<000D000000?mP003oH000of00000000030?nJ00050000003og`00oml0
0?oO000000000`3Bo`001@000000OOl007go001mo`00000000809?l000D00000003o=@00ocD00?le
00000003003oSP0500000000on<00?oS003oh`0000000`00c?l01@00000008Co0024o`00Q?l00000
008004Go00D00000000Ao`004Ol0017o0000000301L0o`050000000c0?l0<`3o03<0o`0000000`13
0?l01@000000A`3o04L0o`170?l000000080?`3o00D0000002`0o`0/0?l0;03o0000000300l0o`05
000000005_l001Ko000Fo`0000000`00@ol01000000007Ko001fo`000003002/o`0500000000i?l0
0>Co003To`0000000`00on801@0000000?nZ003oZP00ojX0000000<00?md00<000000?ooo`3oool0
103oool100000040oooo000C0?ooo`030?oog`2LB7@0_ooo00X0oooo0`0000050?ooool00000P@00
00040?ooo`<000000@3oool001@0oooo00<0okmd001d_`3oool02@3oool00`000000oooo0?ooo`05
0?ooo`0400000000oa000?l@000000<00?m100D00000003oM`00ogL00?mg00000003003o/@050000
0000onh00?o^003okP0000000`00e?l01000000009So002Ho`000003001Oo`0500000000:ol002_o
000[o`0000000`010?l01@0000009P3o02H0o`0V0?l0000000<0@03o00@0000004h0o`1>0?l00000
0`1@0?l01@000000A03o04@0o`140?l0000000<0:`3o00D0000000@0o`040?l0103o00000003000_
o`0400000000K_l006ko000000<00;So00D00000003om@00ooD00?oe00000003003oW00500000000
ocl00?lo003o?`0000000`0No`0010000000N_l007[o000000030=;o00050000003og@00omd00?oM
000000000`3oUP001@000000oe`00?mL003oG000000000<0oc0000@000000?lD003o500000000`3o
2P001@000000oa800?lB003o4P00000000<0ob`000D000000?mI003oF@00oeT0000000030?nF0004
0000003oh`00on<0000000<0`_l000D0000005go001Mo`00GOl000000003003o400500000000oh80
0?n2003oPP0000000P00ooH01@00000009Oo002Go`00Uol0000000<002[o00D0000003X0o`0j0?l0
>P3o0000000309<0o`050000003N0?l0gP3o0=h0o`0000000P3o0>L01@000000o`300?l0`03o0<00
000000<0o`2[00D000000?l0Z`3o0:/0o`2[000000030?l0`0050000003o0>L0o`3W0?l0i`000000
0P3N0?l01@000000T`3o09<0o`2C0?l0000000<0>P3o00D00000000Zo`00:_l002[o00000003002G
o`0500000000ooH00?of003omP0000000P00oh801@0000000?l@003o4000oa00000000<0GOl000D0
00000<;o0032o`00`_l0000000030?oS00050000003oUP00oiH00?nF000000000P3oF@001@000000
ob`00?l/003o;000000000<0oa8000D000000?l:003o2P00o`X0000000030?lD00050000003o<000
oc000?l`000000000P3oG0001@000000oiH00?nF003oUP00000000<0omd000D000000=;o003Bo`00
d_l00000000307[o00050000000No`007_l001ko000000000P00ocl01@0000000?nL003oW000oi`0
000000<00?oe00D00000002ho`00^?l00;So00000003001^o`0500000000;ol002oo000_o`000000
0P040?l01@000000:`3o02/0o`0[0?l0000000<0A03o00D000000500o`1@0?l0D03o0000000304h0
o`05000000100?l0@03o0400o`0000000P0V0?l01@0000000@3o0040o`010?l0000000<002_o00D0
0000001Oo`00Gol005oo00000003002Ho`0400000000e?l00=Co000000<00?o^00D00000003o/@00
ok400?na00000003003oM`0500000000od400?m1003o@@0000000`00oa000`000000oooo0?ooo`04
0?ooo`4000000@3oool001@0oooo00<0oono07@0B02Lgol02@3oool00`000000oooo0?ooo`050?oo
o`0400000000oa000?l@000000<00?m100D00000003oM`00ogL00?mg00000003003o/@0500000000
onh00?o^003okP0000000`00e?l01000000009So002Ho`000003001Oo`0500000000:ol002_o000[
o`0000000`010?l01@0000009P3o02H0o`0V0?l0000000<0@03o00@0000004h0o`1>0?l000000`1@
0?l01@000000A03o04@0o`140?l0000000<0:`3o00D0000000@0o`040?l0103o00000003000_o`04
00000000K_l006ko000000<00;So00D00000003om@00ooD00?oe00000003003oW00500000000ocl0
0?lo003o?`0000000`0No`0010000000N_l007[o000000030=;o00050000003og@00omd00?oM0000
00000`3oUP001@000000oe`00?mL003oG000000000<0oc0000@000000?lD003o500000000`3o2P00
1@000000oa800?lB003o4P00000000<0ob`000D000000?mI003oF@00oeT0000000030?nF00040000
003oh`00on<0000000<0`_l000D0000005go001Mo`00GOl000000003003o400500000000oh800?n2
003oPP0000000P00ooH01@00000009Oo002Go`00Uol0000000<002[o00D0000003X0o`0j0?l0>P3o
0000000309<0o`050000003N0?l0gP3o0=h0o`0000000P3o0>L01@000000o`300?l0`03o0<000000
00<0o`2[00D000000?l0Z`3o0:/0o`2[000000030?l0`0050000003o0>L0o`3W0?l0i`0000000P3N
0?l01@000000T`3o09<0o`2C0?l0000000<0>P3o00D00000000Zo`00:_l002[o00000003002Go`05
00000000ooH00?of003omP0000000P00oh801@0000000?l@003o4000oa00000000<0GOl000D00000
0<;o0032o`00`_l0000000030?oS00050000003oUP00oiH00?nF000000000P3oF@001@000000ob`0
0?l/003o;000000000<0oa8000D000000?l:003o2P00o`X0000000030?lD00050000003o<000oc00
0?l`000000000P3oG0001@000000oiH00?nF003oUP00000000<0omd000D000000=;o003Bo`00d_l0
0000000307[o00050000000No`007_l001ko000000000P00ocl01@0000000?nL003oW000oi`00000
00<00?oe00D00000002ho`00^?l00;So00000003001^o`0500000000;ol002oo000_o`0000000P04
0?l01@000000:`3o02/0o`0[0?l0000000<0A03o00D000000500o`1@0?l0D03o0000000304h0o`05
000000100?l0@03o0400o`0000000P0V0?l01@0000000@3o0040o`010?l0000000<002_o00D00000
001Oo`00Gol005oo00000003002Ho`0400000000e?l00=Co000000<00?o^00D00000003o/@00ok40
0?na00000003003oM`0500000000od400?m1003o@@0000000`00oa000`000000oooo0?ooo`040?oo
o`4000000@3oool001@0oooo00<0oono07@00018W=l02@3oool00`000000oooo0?ooo`050?ooo`04
00000000oa000?l@000000<00?m100D00000003oM`00ogL00?mg00000003003o/@0500000000onh0
0?o^003okP0000000`00e?l01000000009So002Ho`000003001Oo`0500000000:ol002_o000[o`00
00000`010?l01@0000009P3o02H0o`0V0?l0000000<0@03o00@0000004h0o`1>0?l000000`1@0?l0
1@000000A03o04@0o`140?l0000000<0:`3o00D0000000@0o`040?l0103o00000003000_o`040000
0000K_l006ko000000<00;So00D00000003om@00ooD00?oe00000003003oW00500000000ocl00?lo
003o?`0000000`0No`0010000000N_l007[o000000030=;o00050000003og@00omd00?oM00000000
0`3oUP001@000000oe`00?mL003oG000000000<0oc0000@000000?lD003o500000000`3o2P001@00
0000oa800?lB003o4P00000000<0ob`000D000000?mI003oF@00oeT0000000030?nF00040000003o
h`00on<0000000<0`_l000D0000005go001Mo`00GOl000000003003o400500000000oh800?n2003o
PP0000000P00ooH01@00000009Oo002Go`00Uol0000000<002[o00D0000003X0o`0j0?l0>P3o0000
000309<0o`050000003N0?l0gP3o0=h0o`0000000P3o0>L01@000000o`300?l0`03o0<00000000<0
o`2[00D000000?l0Z`3o0:/0o`2[000000030?l0`0050000003o0>L0o`3W0?l0i`0000000P3N0?l0
1@000000T`3o09<0o`2C0?l0000000<0>P3o00D00000000Zo`00:_l002[o00000003002Go`050000
0000ooH00?of003omP0000000P00oh801@0000000?l@003o4000oa00000000<0GOl000D000000<;o
0032o`00`_l0000000030?oS00050000003oUP00oiH00?nF000000000P3oF@001@000000ob`00?l/
003o;000000000<0oa8000D000000?l:003o2P00o`X0000000030?lD00050000003o<000oc000?l`
000000000P3oG0001@000000oiH00?nF003oUP00000000<0omd000D000000=;o003Bo`00d_l00000
000307[o00050000000No`007_l001ko000000000P00ocl01@0000000?nL003oW000oi`0000000<0
0?oe00D00000002ho`00^?l00;So00000003001^o`0500000000;ol002oo000_o`0000000P040?l0
1@000000:`3o02/0o`0[0?l0000000<0A03o00D000000500o`1@0?l0D03o0000000304h0o`050000
00100?l0@03o0400o`0000000P0V0?l01@0000000@3o0040o`010?l0000000<002_o00D00000001O
o`00Gol005oo00000003002Ho`0400000000e?l00=Co000000<00?o^00D00000003o/@00ok400?na
00000003003oM`0500000000od400?m1003o@@0000000`00oa000`000000oooo0?ooo`040?ooo`40
00000@3oool000l0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`090?ooo`030000
003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?ooo`003`3oool0203ooml0W4Q809cO
o`3oool0oono07@0000007@0_ooo2@3oool00`000000oooo0?ooo`050?ooo`0400000014o`00A?l0
000000<07Ol000D00000003o4@00oa400?lA00000003003oA`0500000000oh<00?n3003oP`000000
0`00ol80100000000?_o003ko`000003002jo`0500000000Nol007_o001ko`0000000`00@Ol01@00
000000go000=o`003Ol0000000<07@3o00@0000003d0o`0m0?l000000`1A0?l01@000000F03o05P0
o`1H0?l0000000<0D03o00D0000003T0o`0i0?l0>@3o0000000301@0o`04000000007_l001ko0000
00<005oo00D00000002[o`00Zol00:_o00000003003oo`0500000000oj<00?nS003oX`0000000`00
od<0100000007_l001ko0000000307go00050000003Go`00eol00=Oo000000000`3oeP001@000000
ohl00?n?003oS`00000000<0oeD000@000000?l/003o;00000000`3o5@001@000000oa000?l@003o
4000000000<0oal000D000000?m2003o@P00od80000000030?mg00040000003o_P00okh0000000<0
jol000D0000008So0028o`00R?l00000000301_o000500000000oeP00?mH003oF00000000P00olh0
1@0000000;_o002ko`00^ol0000000<004[o00D0000001h0o`0N0?l07P3o0000000307`0o`050000
003;0?l0b`3o0</0o`0000000P3o0?L01@000000o`3=0?l0c@3o0<d0000000<0o`2h00D000000?l0
^03o0;P0o`2h000000030?l0c@050000003o0?L0o`3g0?l0m`0000000P3;0?l01@000000O03o07`0
o`1l0?l0000000<07P3o00D00000001:o`00B_l004[o00000003002ko`0500000000olh00?o>003o
cP0000000P00oeP01@0000006ol001_o000Ko`00000000<0R?l000D000000>_o003[o`00jol00000
00030?nn00050000003oM`00ogL00?mg000000000P3o@P001@000000oal00?lO003o7`00000000<0
oa0000D000000?lE003o5@00oaD0000000030?l/00050000003oE@00oeD00?mE000000000P3oS`00
1@000000omH00?oF003oeP00000000<0eol000D0000007go001mo`00OOl00000000301ko00050000
0000od<00?m3003o@`0000000P00oj<01@0000000?oo003oo`00ool0000000<00:_o00D00000001O
o`00Gol005oo00000003000No`050000000D0?l0503o01@0o`0000000P0i0?l01@000000D03o0500
o`1@0?l0000000<0F03o00D000000540o`1A0?l0D@3o0000000303d0o`050000000M0?l07@3o01d0
o`0000000P003Ol01@000000047o0011o`00@Ol0000000<007_o00D00000002jo`00^_l00;[o0000
0003003ko`0400000000ol800?o2000000<00?n300D00000003oA`00odL00?m700000003003o4@05
0000000Mo`007Ol001go000000000`14o`000`000000oooo0?ooo`040?ooo`4000000@3oool00100
oooo00H0oooO09a80000000000000000M02oool:0?ooo`030000003oool0oooo00D0oooo00@00000
04Co0014o`0000000`0Mo`001@0000000?lA003o4@00oa40000000<00?m700D00000003oP`00oh<0
0?n300000003003o`P0400000000nol00?_o000000<00;[o00D00000001ko`00Nol007_o00000003
0011o`05000000003Ol000go000=o`0000000`0M0?l010000000?@3o03d0o`0000030540o`050000
001H0?l0F03o05P0o`0000000`1@0?l01@000000>@3o03T0o`0i0?l0000000<0503o00@00000000N
o`007_l000000`00Gol01@0000000:_o002[o`00Zol0000000<00?oo00D00000003oX`00oj<00?nS
00000003003o@`040000000No`007_l0000000<0OOl000D000000=Oo003Go`00eol0000000030?oF
00050000003oS`00ohl00?n?000000000`3oE@0010000000ob`00?l/000000030?lE00050000003o
4000oa000?l@000000000`3o7`001@000000od800?m2003o@P00000000<0ogL000@000000?nn003o
_P0000000`3[o`001@000000R?l008So0028o`00000000<06ol000D00000003oF000oeP00?mH0000
0002003ocP0500000000^ol00;_o002ko`0000000`00B_l01@0000007P3o01h0o`0N0?l0000000<0
O03o00D000000</0o`3;0?l0b`3o000000020?l0m`050000003o0<d0o`3=0?l0c@0000000`3o0;P0
1@000000o`2h0?l0^03o0;P0000000<0o`3=00D000000?l0m`3o0?L0o`3g000000020</0o`050000
001l0?l0O03o07`0o`0000000`0N0?l01@00000004[o001:o`00B_l0000000<00;_o00D00000003o
cP00olh00?o>00000002003oF0050000000Ko`006ol001_o000000000`28o`001@000000jol00>_o
003[o`00000000<0okh000D000000?mg003oM`00ogL0000000020?m200050000003o7`00oal00?lO
000000000`3o40001@000000oaD00?lE003o5@00000000<0ob`000D000000?mE003oE@00oeD00000
00020?n?00050000003oeP00omH00?oF000000000`3Go`001@000000OOl007go001mo`00000000<0
7_l000D00000003o@`00od<00?m300000002003oX`0500000000ool00?oo003oo`0000000`00Zol0
1@00000005oo001Oo`00Gol0000000<001ko00D0000001@0o`0D0?l0503o0000000203T0o`050000
001@0?l0D03o0500o`0000000`1H0?l01@000000D@3o0540o`1A0?l0000000<0?@3o00D0000001d0
o`0M0?l07@3o00000002000=o`0500000000@Ol0047o0011o`0000000`00Nol01@0000000;[o002j
o`00^_l0000000<00?_o00@00000003o`P00ol8000000`00oh<01@0000000?m7003oA`00odL00000
00<00?lA00D0000001go000Mo`007Ol00000000304Co00030000003oool0oooo00@0oooo0@000001
0?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000014o`00A?l0000000<07Ol000D00000
003o4@00oa400?lA00000003003oA`0500000000oh<00?n3003oP`0000000`00ol80100000000?_o
003ko`000003002jo`0500000000Nol007_o001ko`0000000`00@Ol01@00000000go000=o`003Ol0
000000<07@3o00@0000003d0o`0m0?l000000`1A0?l01@000000F03o05P0o`1H0?l0000000<0D03o
00D0000003T0o`0i0?l0>@3o0000000301@0o`04000000007_l001ko000000<005oo00D00000002[
o`00Zol00:_o00000003003oo`0500000000oj<00?nS003oX`0000000`00od<0100000007_l001ko
0000000307go00050000003Go`00eol00=Oo000000000`3oeP001@000000ohl00?n?003oS`000000
00<0oeD000@000000?l/003o;00000000`3o5@001@000000oa000?l@003o4000000000<0oal000D0
00000?m2003o@P00od80000000030?mg00040000003o_P00okh0000000<0jol000D0000008So0028
o`00R?l00000000301_o000500000000oeP00?mH003oF00000000P00olh01@0000000;_o002ko`00
^ol0000000<004[o00D0000001h0o`0N0?l07P3o0000000307`0o`050000003;0?l0b`3o0</0o`00
00000P3o0?L01@000000o`3=0?l0c@3o0<d0000000<0o`2h00D000000?l0^03o0;P0o`2h00000003
0?l0c@050000003o0?L0o`3g0?l0m`0000000P3;0?l01@000000O03o07`0o`1l0?l0000000<07P3o
00D00000001:o`00B_l004[o00000003002ko`0500000000olh00?o>003ocP0000000P00oeP01@00
00006ol001_o000Ko`00000000<0R?l000D000000>_o003[o`00jol0000000030?nn00050000003o
M`00ogL00?mg000000000P3o@P001@000000oal00?lO003o7`00000000<0oa0000D000000?lE003o
5@00oaD0000000030?l/00050000003oE@00oeD00?mE000000000P3oS`001@000000omH00?oF003o
eP00000000<0eol000D0000007go001mo`00OOl00000000301ko000500000000od<00?m3003o@`00
00000P00oj<01@0000000?oo003oo`00ool0000000<00:_o00D00000001Oo`00Gol005oo00000003
000No`050000000D0?l0503o01@0o`0000000P0i0?l01@000000D03o0500o`1@0?l0000000<0F03o
00D000000540o`1A0?l0D@3o0000000303d0o`050000000M0?l07@3o01d0o`0000000P003Ol01@00
0000047o0011o`00@Ol0000000<007_o00D00000002jo`00^_l00;[o00000003003ko`0400000000
ol800?o2000000<00?n300D00000003oA`00odL00?m700000003003o4@050000000Mo`007Ol001go
000000000`14o`000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3o
ool01@3ooooo0000084000001P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo
00@00000087o0021o`0000000`1Xo`001@000000AOl004Go0015o`00000000<06?l000D00000003o
7000oa`00?lL00000003003oF00400000000oiX00?nJ000000<00?oN00D00000003Ko`00fol00=_o
00000003002Go`0500000000Eol005Oo001Go`0000000`007Ol0100000004`3o01<0o`00000303T0
o`050000001B0?l0DP3o0580o`0000000`1M0?l01@000000F03o05P0o`1H0?l0000000<0A03o00@0
000001l0o`0O0?l000000`005?l01@00000005Ko001Fo`00E_l0000000<00:Co00D00000003lo`00
o?l00?co00000003003oX`0400000000ocl00?lo000000<09?l000D0000008Oo0027o`00Qol00000
00030>?o00050000003obP00olX00?o:000000000`3oQ00010000000odh00?m>000000030?lY0005
0000003o5`00oaL00?lG000000000`3o6P001@000000oc800?lb003o<P00000000<0oel000@00000
0?nN003oWP0000000`3ok`001@000000/?l00;3o002`o`00000000<0A?l000D00000003o;`00obl0
0?l_00000002003oZ00500000000g_l00=ko003No`0000000`00JOl01@0000000`3o00<0o`030?l0
000000<0I@3o00D000000;L0o`2g0?l0]`3o000000020?P0o`050000003o0=/0o`3K0?l0f`000000
0`3o0<@01@000000o`340?l0a03o0<@0000000<0o`3K00D000000?P0o`3h0?l0n03o000000020;L0
o`050000001U0?l0I@3o06D0o`0000000`030?l01@00000006Wo001Yo`00JOl0000000<00=ko00D0
0000003oZ000ojP00?nX00000002003o;`0500000014o`00A?l004Co000000000`2`o`001@000000
onl00?o_003ok`00000000<0oih000D000000?mO003oG`00oel0000000020?lb00050000003o6P00
oaX00?lJ000000000`3o5`001@000000obT00?lY003o:@00000000<0odh000D000000?n4003oQ000
oh@0000000020?o:00050000003So`00hol00>?o000000000`27o`001@0000009?l002Co000To`00
000000<00?lo00D00000003oX`00oj<00?nS00000002003lo`0500000000Y?l00:Co002To`000000
0`00E_l01@00000001Co000Do`005?l0000000<07`3o00D0000004@0o`140?l0A03o0000000205P0
o`050000001M0?l0G@3o05d0o`0000000`1B0?l01@000000>@3o03T0o`0i0?l0000000<04`3o00D0
0000000Mo`007Ol001go00000002001Go`0500000000Uol009Oo002Go`0000000`00fol01@000000
0?oN003ogP00omh0000000<00?nJ00@00000003oF000oeP000000`00oa`01@0000006?l001So000H
o`00000000<0AOl000D0000006So001Xo`00J?l000000003087o00030000003oool0oooo00@0oooo
0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000021o`00POl0000000<0J?l0
00D0000004Go0015o`00AOl00000000301So000500000000oa`00?lL003o700000000`00oeP01000
00000?nJ003oVP000003003ogP0500000000fol00=_o003Ko`0000000`00Uol01@00000005Oo001G
o`00Eol0000000<001go00@0000001<0o`0C0?l000000`0i0?l01@000000DP3o0580o`1B0?l00000
00<0G@3o00D0000005P0o`1H0?l0F03o0000000304@0o`040000000O0?l07`3o000000<001Co00D0
0000001Fo`00E_l005Ko00000003002To`0500000000o?l00?co003lo`0000000`00oj<010000000
0?lo003o?`00000302Co000500000027o`00Qol008Oo000000000`3So`001@000000olX00?o:003o
bP00000000<0oh@000@000000?m>003oCP0000000`3o:@001@000000oaL00?lG003o5`00000000<0
oaX000D000000?lb003o<P00oc80000000030?mO00040000003oWP00oih0000000<0onl000D00000
0;3o002`o`00/?l00000000304Co000500000000obl00?l_003o;`0000000P00ojP01@0000000=ko
003No`00g_l0000000<006Wo00D0000000<0o`030?l00`3o0000000306D0o`050000002g0?l0]`3o
0;L0o`0000000P3h0?l01@000000o`3K0?l0f`3o0=/0000000<0o`3400D000000?l0a03o0<@0o`34
000000030?l0f`050000003h0?l0n03o0?P0o`0000000P2g0?l01@000000I@3o06D0o`1U0?l00000
00<00`3o00D00000001Yo`00JOl006Wo00000003003No`0500000000ojP00?nX003oZ00000000P00
obl01@000000A?l004Co0014o`00000000<0/?l000D000000?o_003ok`00onl0000000030?nN0005
0000003oG`00oel00?mO000000000P3o<P001@000000oaX00?lJ003o6P00000000<0oaL000D00000
0?lY003o:@00obT0000000030?m>00050000003oQ000oh@00?n4000000000P3obP001@000000hol0
0>?o003So`00000000<0Qol000D0000002Co000To`009?l000000003003o?`0500000000oj<00?nS
003oX`0000000P00o?l01@0000000:Co002To`00Y?l0000000<005Ko00D00000000Do`005?l001Co
0000000301l0o`05000000140?l0A03o04@0o`0000000P1H0?l01@000000G@3o05d0o`1M0?l00000
00<0DP3o00D0000003T0o`0i0?l0>@3o0000000301<0o`05000000007Ol001go000Mo`0000000P00
Eol01@00000009Oo002Go`00Uol0000000<00=_o00D00000003ogP00omh00?oN00000003003oVP04
00000000oeP00?mH000000<00?lL00D0000001So000Ho`006?l00000000304Go00050000001Xo`00
J?l006So000000000`21o`000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<00000
0?ooo`3oool01@3ooooo0000084000001P3oool100000040oooo000P0?ooo`030000003oool0oooo
00D0oooo00@000000:3o002Po`0000000`2Ho`001@000000QOl008Go0025o`00000000<0I_l000D0
000003co000lo`00??l00000000300Wo000400000000oc800?lb000000<00?me00D00000003o_000
ok`00?nl00000003003io`0500000000/Ol00;7o002ao`0000000`00Jol01000000002co000/o`00
000300X0o`050000000e0?l0=@3o03D0o`0000000`1B0?l01@000000H03o0600o`1P0?l0000000<0
GP3o00@0000004X0o`1:0?l000000`0V0?l01@00000000ko000>o`003_l0000000<005?o00D00000
002To`00Y?l00:Co00000003003oo`0400000000oi`00?nL000000<00?le00D0000003;o000bo`00
<_l00000000309Ko00050000003co`00lol00??o000000000`3o^`0010000000ogL00?mg00000003
0?m500050000003o9`00obL00?lW000000000`3o7@001@000000obX00?lZ003o:P00000000<0od`0
00@000000?n4003oQ00000000`3oc`001@000000e?l00=Co003Do`00000000<0J_l000D00000003o
2000o`P00?l800000002003oPP0500000000ooh00?on003ooP0000000`00R?l01@00000001Oo000G
o`005ol0000000<0CP3o00D000000:@0o`2T0?l0Y03o000000020>P0o`050000003o0>T0o`3Y0?l0
j@0000000`3o0=401@000000o`3A0?l0d@3o0=40000000<0o`3Y00D000000>P0o`3X0?l0j03o0000
00020:@0o`050000001>0?l0CP3o04h0o`0000000`005ol01@00000008So0028o`00R?l0000000<0
0?on00D00000003oPP00oh800?n200000002003o20050000001Zo`00J_l006[o000000000`3Do`00
1@000000oll00?o?003oc`00000000<0oh@000D000000?m<003oC000od`0000000020?lZ00050000
003o7@00oad00?lM000000000`3o9`001@000000odD00?m5003oA@00000000<0ogL000D000000?nk
003o^`00ok/0000000020??o00050000002Fo`00U_l009Ko000000000`0bo`001@0000000?le003o
=@00ocD0000000<00?nL00D00000003oo`00ool00?oo00000002002To`0500000000Dol005?o001C
o`0000000`003_l01@0000009P3o02H0o`0V0?l0000000<0BP3o00D0000005h0o`1N0?l0GP3o0000
00020600o`050000001B0?l0DP3o0580o`0000000`0e0?l01@0000002P3o00X0o`0:0?l0000000<0
02co00D00000001[o`00Jol006_o00000002002ao`0500000000nOl00?Wo003io`0000000`00ok`0
1@0000000?me003oM@00ogD0000000<00?lb00@0000000Wo0009o`0000000`0lo`001@000000I_l0
06Ko001Vo`00000000<0QOl000D0000009So002Ho`00V?l0000000030:3o00030000003oool0oooo
00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`040000002Po`00X?l00000
00<0V?l000D0000008Go0025o`00QOl00000000306Ko00050000000lo`00??l003co000000000`09
o`00100000000?lb003o<P000003003oM@0500000000ok`00?nl003o_00000000`00nOl01@000000
0;7o002ao`00/Ol0000000<006_o00@00000000/o`00;?l000000`0:0?l01@000000=@3o03D0o`0e
0?l0000000<0DP3o00D000000600o`1P0?l0H03o0000000305h0o`040000001:0?l0BP3o000000<0
9P3o00D00000000>o`003_l000ko00000003001Co`0500000000Y?l00:Co002To`0000000`00ool0
100000000?nL003oW0000003003o=@050000000bo`00<_l003;o000000000`2Fo`001@000000lol0
0??o003co`00000000<0ok/000@000000?mg003oM`0000000`3oA@001@000000obL00?lW003o9`00
000000<0oad000D000000?lZ003o:P00obX0000000030?m<00040000003oQ000oh@0000000<0oll0
00D000000=Co003Do`00e?l00000000306[o000500000000o`P00?l8003o200000000P00oh801@00
00000?on003ooP00ooh0000000<008So00D00000000Go`005ol001Oo0000000304h0o`050000002T
0?l0Y03o0:@0o`0000000P3X0?l01@000000o`3Y0?l0j@3o0>T0000000<0o`3A00D000000?l0d@3o
0=40o`3A000000030?l0j@050000003X0?l0j03o0>P0o`0000000P2T0?l01@000000CP3o04h0o`1>
0?l0000000<001Oo00D000000028o`00R?l008So00000003003ooP0500000000oh800?n2003oPP00
00000P00o`P01@000000J_l006[o001Zo`00000000<0e?l000D000000?o?003oc`00oll000000003
0?n400050000003oC000od`00?m<000000000P3o:P001@000000oad00?lM003o7@00000000<0obL0
00D000000?m5003oA@00odD0000000030?mg00050000003o^`00ok/00?nk000000000P3co`001@00
0000U_l009Ko002Fo`00000000<0<_l000D00000003o=@00ocD00?le00000003003oW00500000000
ool00?oo003oo`0000000P00Y?l01@00000005?o001Co`00Dol0000000<000ko00D0000002H0o`0V
0?l09P3o0000000304X0o`050000001N0?l0GP3o05h0o`0000000P1P0?l01@000000DP3o0580o`1B
0?l0000000<0=@3o00D0000000X0o`0:0?l02P3o00000003000/o`0500000000Jol006_o001[o`00
00000P00/Ol01@0000000?Wo003io`00nOl0000000<00?nl00D00000003oM@00ogD00?me00000003
003o<P0400000009o`002Ol0000000<0??l000D0000006Ko001Vo`00I_l00000000308Go00050000
002Ho`00V?l009So000000000`2Po`000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo
00<000000?ooo`3oool01@3oool010000000X?l00:3o0000000309So000500000025o`00QOl008Go
000000000`1Vo`001@000000??l003co000lo`00000000<02Ol000@00000003o<P00oc8000000`00
ogD01@0000000?nl003o_000ok`0000000<00?Wo00D00000002ao`00/Ol00;7o00000003001[o`04
00000000;?l002co000000<02P3o00D0000003D0o`0e0?l0=@3o000000030580o`050000001P0?l0
H03o0600o`0000000`1N0?l010000000BP3o04X0o`00000302H0o`05000000003_l000ko000>o`00
00000`00Dol01@0000000:Co002To`00Y?l0000000<00?oo00@00000003oW000oi`000000`00ocD0
1@000000<_l003;o000bo`00000000<0U_l000D000000??o003co`00lol0000000030?nk00040000
003oM`00ogL0000000<0odD000D000000?lW003o9`00obL0000000030?lM00050000003o:P00obX0
0?lZ000000000`3oC00010000000oh@00?n4000000030?o?00050000003Do`00e?l00=Co00000000
0`1Zo`001@0000000?l8003o2000o`P0000000800?n200D00000003ooP00ooh00?on000000030028
o`05000000005ol001Oo000Go`0000000`1>0?l01@000000Y03o0:@0o`2T0?l000000080j03o00D0
00000?l0j@3o0>T0o`3Y000000030?l0d@050000003o0=40o`3A0?l0d@0000000`3o0>T01@000000
j03o0>P0o`3X0?l000000080Y03o00D0000004h0o`1>0?l0CP3o00000003000Go`0500000000R?l0
08So0028o`0000000`00ooh01@0000000?n2003oPP00oh80000000800?l800D0000006[o001Zo`00
J_l0000000030=Co00050000003oc`00oll00?o?000000000`3oQ0001@000000od`00?m<003oC000
00000080obX000D000000?lM003o7@00oad0000000030?lW00050000003oA@00odD00?m500000000
0`3oM`001@000000ok/00?nk003o^`0000000080lol000D0000009Ko002Fo`00U_l00000000303;o
000500000000ocD00?le003o=@0000000`00oi`01@0000000?oo003oo`00ool0000000800:Co00D0
0000001Co`00Dol005?o00000003000>o`050000000V0?l09P3o02H0o`0000000`1:0?l01@000000
GP3o05h0o`1N0?l000000080H03o00D000000580o`1B0?l0DP3o0000000303D0o`050000000:0?l0
2P3o00X0o`0000000`00;?l01@00000006_o001[o`00Jol0000000800;7o00D00000003io`00nOl0
0?Wo00000003003o_00500000000ogD00?me003oM@0000000`00oc80100000002Ol000Wo00000003
03co00050000001Vo`00I_l006Ko000000000`25o`001@000000V?l009So002Ho`00000000<0X?l0
00<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0ooooo`00
0021000000H0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`040000002Oo`00
Wol0000000<0Z_l000D000000:Wo002Yo`00ZOl00000000309_o000500000021o`00POl0087o0000
00000`1Ko`0010000000:_l002[o00000003003o4@0500000000oeD00?mE003oE@0000000`00oih0
1@0000000?o[003oj`00on/0000000<00<Oo00@00000001mo`00OOl000000`00>Ol01@0000000@3o
0040o`010?l0000000<0<03o00D000000540o`1A0?l0D@3o000000030680o`040000001Q0?l0H@3o
000000<0C@3o00D0000002P0o`0X0?l0:03o00000003000>o`0500000000E_l005Ko001Fo`000000
0`00Zol0100000000?oe003om@000003003oSP0500000000ob<00?lS003o8`0000000`16o`001@00
0000Zol00:_o002[o`00000000<0ooP000@000000?nY003oZ@0000000`3oJP001@000000och00?ln
003o?P00000000<0obP000D000000?lY003o:@00obT0000000030?m000040000003oK`00ofl00000
00<0ok<000D000000?Go003eo`00mOl00000000308ko00050000000Ko`006ol001_o000000000P00
oeh01@0000000?oL003og000om`0000000<00:Oo00D00000000bo`00<_l003;o0000000303H0o`05
0000002A0?l0T@3o0940o`0000000P3G0?l01@000000o`3g0?l0m`3o0?L0000000<0o`3N00D00000
0?l0gP3o0=h0o`3N000000030?l0m`050000003G0?l0e`3o0=L0o`0000000P2A0?l01@000000=P3o
03H0o`0f0?l0000000<003;o00D00000002Wo`00Yol00:Oo00000003003og00500000000oeh00?mN
003oGP0000000P0Ko`001@000000S_l008ko002>o`00000000<0mOl000D000000?nc003o/`00ok<0
000000030?m_00050000003o@000od000?m0000000000P3o:@001@000000obP00?lX003o:0000000
00<0och000D000000?mZ003oJP00ofX0000000030?nY00050000003on000ooP00?oh000000000P2[
o`001@000000A_l004Ko0016o`00000000<00?lS00D00000003oSP00ohh00?n>00000003003om@05
00000000Zol00:_o002[o`0000000P00E_l01@00000000ko000>o`003_l0000000<0:03o00D00000
04d0o`1=0?l0C@3o000000030640o`050000001R0?l0HP3o0680o`0000000P1A0?l01@000000<03o
0300o`0`0?l0000000<00@3o00D00000000io`00>Ol003Wo00000003001mo`0500000000aol00<Oo
0037o`0000000P00on/01@0000000?nN003oWP00oih0000000<00?mE00D00000003o4@00oa400?lA
0000000302[o00040000001Ko`00Fol0000000<0POl000D0000009_o002Ko`00Vol0000000030:Wo
00050000002Zo`00Z_l00:[o000000000`2Oo`000`000000oooo0?ooo`040?ooo`4000000@3oool0
0200oooo00<000000?ooo`3oool01@3oool010000000Wol009oo000000030:[o00050000002Yo`00
ZOl00:Wo000000000`2Ko`001@000000POl0087o0021o`00000000<0Fol000@0000002[o000Zo`00
00000`00oa401@0000000?mE003oE@00oeD0000000<00?nN00D00000003oj`00on/00?o[00000003
0037o`0400000000OOl007go000000<003Wo00D000000040o`010?l00@3o000000030300o`050000
001A0?l0D@3o0540o`0000000`1R0?l010000000H@3o0640o`00000304d0o`050000000X0?l0:03o
02P0o`0000000`003_l01@00000005Ko001Fo`00E_l0000000<00:_o00@00000003om@00ooD00000
0`00ohh01@0000000?lS003o8`00ob<0000000<0A_l000D000000:_o002[o`00Zol0000000030?oh
00040000003oZ@00ojT0000000<0ofX000D000000?ln003o?P00och0000000030?lX00050000003o
:@00obT00?lY000000000`3o@00010000000ofl00?m_000000030?nc00050000003eo`00mOl00?Go
000000000`2>o`001@0000006ol001_o000Ko`00000000800?mN00D00000003og000om`00?oL0000
0003002Wo`0500000000<_l003;o000bo`0000000`0f0?l01@000000T@3o0940o`2A0?l000000080
e`3o00D000000?l0m`3o0?L0o`3g000000030?l0gP050000003o0=h0o`3N0?l0gP0000000`3o0?L0
1@000000e`3o0=L0o`3G0?l000000080T@3o00D0000003H0o`0f0?l0=P3o00000003000bo`050000
0000Yol00:Oo002Wo`0000000`00om`01@0000000?mN003oGP00oeh0000000806ol000D0000008ko
002>o`00S_l0000000030?Go00050000003o/`00ok<00?nc000000000`3oK`001@000000od000?m0
003o@00000000080obT000D000000?lX003o:000obP0000000030?ln00050000003oJP00ofX00?mZ
000000000`3oZ@001@000000ooP00?oh003on00000000080Zol000D0000004Ko0016o`00A_l00000
0003003o8`0500000000ohh00?n>003oSP0000000`00ooD01@0000000:_o002[o`00Zol000000080
05Ko00D00000000>o`003_l000ko0000000302P0o`050000001=0?l0C@3o04d0o`0000000`1Q0?l0
1@000000HP3o0680o`1R0?l000000080D@3o00D000000300o`0`0?l0<03o000000030040o`050000
0000>Ol003Wo000io`0000000`00OOl01@0000000<Oo0037o`00aol0000000800?o[00D00000003o
WP00oih00?nN00000003003oE@0500000000oa400?lA003o4@0000000`0Zo`0010000000Fol005_o
00000003087o00050000002Ko`00Vol009_o000000000`2Yo`001@000000Z_l00:[o002Zo`000000
00<0Wol000<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0
oooo00@0000009oo002Oo`0000000`2Zo`001@000000ZOl00:Wo002Yo`00000000<0Vol000D00000
087o0021o`00POl00000000305_o00040000000Zo`00:_l0000000<00?lA00D00000003oE@00oeD0
0?mE00000003003oWP0500000000on/00?o[003oj`0000000`00aol01000000007go001mo`000003
000io`05000000010?l00@3o0040o`0000000`0`0?l01@000000D@3o0540o`1A0?l0000000<0HP3o
00@000000640o`1Q0?l000000`1=0?l01@000000:03o02P0o`0X0?l0000000<000ko00D00000001F
o`00E_l005Ko00000003002[o`0400000000ooD00?oe000000<00?n>00D00000003o8`00ob<00?lS
0000000304Ko00050000002[o`00Zol00:_o000000000`3on00010000000ojT00?nY000000030?mZ
00050000003o?P00och00?ln000000000`3o:0001@000000obT00?lY003o:@00000000<0od0000@0
00000?m_003oK`0000000`3o/`001@000000mOl00?Go003eo`00000000<0S_l000D0000001_o000K
o`006ol000000002003oGP0500000000om`00?oL003og00000000`00Yol01@00000003;o000bo`00
<_l0000000<0=P3o00D000000940o`2A0?l0T@3o000000020=L0o`050000003o0?L0o`3g0?l0m`00
00000`3o0=h01@000000o`3N0?l0gP3o0=h0000000<0o`3g00D000000=L0o`3G0?l0e`3o00000002
0940o`050000000f0?l0=P3o03H0o`0000000`00<_l01@0000000:Oo002Wo`00Yol0000000<00?oL
00D00000003oGP00oeh00?mN0000000201_o00050000002>o`00S_l008ko000000000`3eo`001@00
0000ok<00?nc003o/`00000000<0ofl000D000000?m0003o@000od00000000020?lY00050000003o
:000obP00?lX000000000`3o?P001@000000ofX00?mZ003oJP00000000<0ojT000D000000?oh003o
n000ooP0000000020:_o000500000016o`00A_l004Ko000000000`00ob<01@0000000?n>003oSP00
ohh0000000<00?oe00D00000002[o`00Zol00:_o00000002001Fo`05000000003_l000ko000>o`00
00000`0X0?l01@000000C@3o04d0o`1=0?l0000000<0H@3o00D000000680o`1R0?l0HP3o00000002
0540o`050000000`0?l0<03o0300o`0000000`010?l01@00000003Wo000io`00>Ol0000000<007go
00D000000037o`00aol00<Oo00000002003oj`0500000000oih00?nN003oWP0000000`00oeD01@00
00000?lA003o4@00oa40000000<0:_l000@0000005_o001Ko`0000000`21o`001@000000Vol009_o
002Ko`00000000<0ZOl000D000000:[o002Zo`00Z_l00000000309oo00030000003oool0oooo00@0
oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooool00000P@0000060?ooo`400000
0@3oool00200oooo00<000000?ooo`3oool01@3oool010000000P?l0083o0000000309co00050000
002^o`00[_l00:ko000000000`2do`001@000000[?l00:co002/o`00000000<0Uol000@0000007Go
001eo`0000000`15o`001@0000002_l000[o000:o`00000000<00?li00D00000003oQ@00ohD00?n5
00000003003oe00400000000f_l00=[o000000<008co00D000000013o`00@ol004?o000000030004
o`050000000]0?l0;@3o02d0o`0000000`1A0?l010000000H`3o06<0o`0000030680o`050000001=
0?l0C@3o04d0o`0000000`0V0?l01@00000001Co000Do`005?l0000000<005oo00@00000002ho`00
^?l000000`00on<01@0000000?mh003oN000ogP0000000<00?l;00D00000063o001Po`00H?l00000
00030<Go00040000003oh000on00000000<0oiD000D000000?mM003oG@00oed0000000030?lj0005
0000003o;P00obh00?l^000000000`3o>`0010000000of000?mP000000030?nL00050000003ok@00
ond00?o]000000000`2_o`001@000000?_l003ko000no`00000000800?ll00D00000003o^`00ok/0
0?nk000000030035o`0500000000COl004go001=o`0000000`0O0?l01@000000O@3o07d0o`1m0?l0
00000080a`3o00D000000?X0o`3j0?l0nP3o000000030?l0j`050000003o0>/0o`3[0?l0j`000000
0`3j0?l01@000000a`3o0<L0o`370?l000000080O@3o00D0000001l0o`0O0?l07`3o00000003001=
o`0500000000aOl00<Go0035o`0000000`00ok/01@0000000?ll003o?000oc`000000080?_l000D0
00000:oo002_o`00[ol0000000030?o]00050000003oW000oi`00?nL000000000`3oH0001@000000
oc/00?lk003o>`0000000080obh000D000000?lj003o>P00ocX0000000030?mM00050000003oU@00
oiD00?nE000000000`3oh0001@000000aOl00<Go0035o`0000000080H?l000D00000003o2`00o`/0
0?l;00000003003oN00500000000on<00?oS003oh`0000000`00^?l01@00000005oo001Oo`00Gol0
0000008001Co00D0000002H0o`0V0?l09P3o0000000304d0o`050000001R0?l0HP3o0680o`000000
0`1S0?l01@000000D@3o0540o`1A0?l000000080;@3o00D000000004o`001?l000Co000000030013
o`0500000000S?l008co002<o`0000000`00f_l01@0000000?oD003oe000om@0000000800?n500D0
0000003o>@00ocT00?li0000000300[o000500000015o`00AOl004Go000000000`1eo`0010000000
Uol009Oo000000030:co00050000002do`00]?l00;Co000000000`2^o`001@000000W?l009co002L
o`00000000<0P?l000<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0
oooo00D0oooo00@00000083o0020o`0000000`2Lo`001@000000[_l00:ko002^o`00000000<0]?l0
00D000000:co002/o`00[?l00000000309Oo00040000001eo`00MOl0000000<0AOl000D0000000[o
000:o`002_l000000003003o>@0500000000ohD00?n5003oQ@0000000`00om@0100000000=[o003J
o`000003002<o`0500000000@ol004?o0013o`0000000`001?l01@000000;@3o02d0o`0]0?l00000
00<0D@3o00@0000006<0o`1S0?l000000`1R0?l01@000000C@3o04d0o`1=0?l0000000<09P3o00D0
0000000Do`005?l001Co00000003001Oo`0400000000^?l00;So000000<00?oS00D00000003oN000
ogP00?mh00000003003o2`050000001Po`00H?l0063o000000000`35o`0010000000on000?oP0000
00030?nE00050000003oG@00oed00?mM000000000`3o>P001@000000obh00?l^003o;P00000000<0
oc/000@000000?mP003oH00000000`3oW0001@000000ond00?o]003ok@00000000<0[ol000D00000
03ko000no`00?_l000000002003o?00500000000ok/00?nk003o^`0000000`00aOl01@00000004go
001=o`00COl0000000<07`3o00D0000007d0o`1m0?l0O@3o000000020<L0o`050000003j0?l0nP3o
0?X0o`0000000`3o0>/01@000000o`3[0?l0j`3o0>/0000000<0nP3o00D000000<L0o`370?l0a`3o
0000000207d0o`050000000O0?l07`3o01l0o`0000000`00COl01@0000000<Go0035o`00aOl00000
00<00?nk00D00000003o?000oc`00?ll0000000203ko00050000002_o`00[ol00:oo000000000`3o
k@001@000000oi`00?nL003oW000000000<0of0000D000000?lk003o>`00oc/0000000020?l^0005
0000003o>P00ocX00?lj000000000`3oG@001@000000oiD00?nE003oU@00000000<0on0000D00000
0<Go0035o`00aOl000000002063o000500000000o`/00?l;003o2`0000000`00ogP01@0000000?oS
003oh`00on<0000000<00;So00D00000001Oo`00Gol005oo00000002000Do`050000000V0?l09P3o
02H0o`0000000`1=0?l01@000000HP3o0680o`1R0?l0000000<0H`3o00D000000540o`1A0?l0D@3o
0000000202d0o`05000000001?l000Co0004o`0000000`00@ol01@00000008co002<o`00S?l00000
00<00=[o00D00000003oe000om@00?oD00000002003oQ@0500000000ocT00?li003o>@0000000`0:
o`001@000000AOl004Go0015o`00000000<0MOl000@0000009Oo002Go`0000000`2/o`001@000000
]?l00;Co002do`00000000<0[_l000D0000009co002Lo`00W?l000000003083o00030000003oool0
oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000020o`00P?l0
000000<0W?l000D000000:ko002^o`00[_l0000000030;Co00050000002/o`00[?l00:co00000000
0`2Go`0010000000MOl007Go0000000304Go00050000000:o`002_l000[o000000000`00ocT01@00
00000?n5003oQ@00ohD0000000<00?oD00@00000003Jo`00f_l000000`00S?l01@00000004?o0013
o`00@ol0000000<000Co00D0000002d0o`0]0?l0;@3o000000030540o`040000001S0?l0H`3o0000
00<0HP3o00D0000004d0o`1=0?l0C@3o0000000302H0o`05000000005?l001Co000Do`0000000`00
Gol0100000000;So002ho`000003003oh`0500000000ogP00?mh003oN00000000`00o`/01@000000
H?l0063o001Po`00000000<0aOl000@000000?oP003oh00000000`3oU@001@000000oed00?mM003o
G@00000000<0ocX000D000000?l^003o;P00obh0000000030?lk00040000003oH000of00000000<0
oi`000D000000?o]003ok@00ond0000000030:oo00050000000no`00?_l003ko000000000P00oc`0
1@0000000?nk003o^`00ok/0000000<00<Go00D00000001=o`00COl004go0000000301l0o`050000
001m0?l0O@3o07d0o`0000000P370?l01@000000nP3o0?X0o`3j0?l0000000<0o`3[00D000000?l0
j`3o0>/0o`3[000000030?X0o`05000000370?l0a`3o0<L0o`0000000P1m0?l01@0000007`3o01l0
o`0O0?l0000000<004go00D000000035o`00aOl00<Go00000003003o^`0500000000oc`00?ll003o
?00000000P0no`001@000000[ol00:oo002_o`00000000<0ond000D000000?nL003oW000oi`00000
00030?mP00050000003o>`00oc/00?lk000000000P3o;P001@000000ocX00?lj003o>P00000000<0
oed000D000000?nE003oU@00oiD0000000030?oP000500000035o`00aOl00<Go000000000P1Po`00
1@0000000?l;003o2`00o`/0000000<00?mh00D00000003oh`00on<00?oS00000003002ho`050000
0000Gol005oo001Oo`0000000P005?l01@0000009P3o02H0o`0V0?l0000000<0C@3o00D000000680
o`1R0?l0HP3o0000000306<0o`050000001A0?l0D@3o0540o`0000000P0]0?l01@00000000Co0004
o`001?l0000000<004?o00D00000002<o`00S?l008co00000003003Jo`0500000000om@00?oD003o
e00000000P00ohD01@0000000?li003o>@00ocT0000000<02_l000D0000004Go0015o`00AOl00000
000307Go00040000002Go`00Uol0000000<0[?l000D000000;Co002do`00]?l0000000030:ko0005
0000002Lo`00W?l009co000000000`20o`000`000000oooo0?ooo`040?ooo`4000000@3oool00200
oooo00<000000?ooo`3oool01@3ooooo0000084000001P3oool100000040oooo000P0?ooo`800000
1P3oool010000000Aol004Oo0000000307?o00050000002Fo`00U_l009Ko000000000`2_o`001@00
0000^ol00;_o002ko`00000000<0^Ol000@000000:So002Xo`0000000`29o`001@000000G?l005co
001Lo`00000000<08Ol000D00000003o8`00ob<00?lS00000003003oL00400000000ol800?o20000
00<00>Wo00D00000002Go`00Uol009Oo00000003001;o`05000000002Ol000Wo0009o`0000000`0[
0?l010000000D03o0500o`00000306<0o`050000001Q0?l0H@3o0640o`0000000`1:0?l01@000000
7`3o01l0o`0O0?l0000000<001ko00@00000001^o`00K_l000000`00c?l01@0000000?o:003obP00
olX0000000<00?mL00D0000001?o000Co`004ol00000000307oo00040000003So`00hol0000000<0
olH000D000000?n2003oPP00oh80000000030?mB00050000003o>P00ocX00?lj000000000`3o>`00
10000000oeH00?mF000000030?n900050000003oe000om@00?oD000000000`3=o`001@000000G_l0
05ko001No`00000000800?lK00D00000003oV`00oi/00?nK00000003003So`0500000000Iol006Oo
001Wo`0000000`080?l01@000000JP3o06X0o`1Z0?l000000080]P3o00D000000>`0o`3/0?l0k03o
000000030?l0n0050000003o0?P0o`3h0?l0n00000000`3/0?l01@000000]P3o0;H0o`2f0?l00000
0080JP3o00D0000000P0o`080?l0203o00000003001Wo`0500000000hol00>?o003So`0000000`00
oi/01@0000000?lK003o6`00oa/000000080G_l000D000000<go003=o`00cOl0000000030?oD0005
0000003oR@00ohT00?n9000000000`3oEP001@000000oc/00?lk003o>`0000000080ocX000D00000
0?mB003oDP00oe80000000030?n200050000003oaP00olH00?o6000000000`3So`001@000000Ool0
07oo001oo`00000000804ol000D00000003oG000oe`00?mL00000003003obP0500000000c?l00<co
003<o`0000000`00K_l01@00000001ko000No`007_l0000000807`3o00D0000004X0o`1:0?l0BP3o
000000030640o`050000001S0?l0H`3o06<0o`0000000`1@0?l01@000000:`3o02/0o`0[0?l00000
008000Wo00D00000001;o`00Bol004_o00000003002Go`0500000000jOl00>Wo003Yo`0000000`00
ol801@0000000?m`003oL000og00000000800?lS00D00000027o000Qo`008Ol00000000305co0005
00000029o`00ROl008Wo000000000`2Xo`0010000000^Ol00;Wo000000030;_o00050000002_o`00
[ol00:oo000000000`2Fo`001@000000Lol007?o001co`00000000<0Aol000<000000?ooo`3oool0
0`3oool200000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@0000004Oo0017o`000000
0`1co`001@000000U_l009Ko002Fo`00000000<0[ol000D000000;_o002ko`00^ol0000000030;Wo
00040000002Xo`00Z?l0000000<0ROl000D0000005co001Lo`00G?l000000003027o000500000000
ob<00?lS003o8`0000000`00og00100000000?o2003o`P000003003Yo`0500000000Uol009Oo002G
o`0000000`00Bol01@00000000Wo0009o`002Ol0000000<0:`3o00@000000500o`1@0?l000000`1S
0?l01@000000H@3o0640o`1Q0?l0000000<0BP3o00D0000001l0o`0O0?l07`3o00000003000No`04
00000000K_l006ko000000<00<co00D00000003obP00olX00?o:00000003003oG0050000000Co`00
4ol001?o000000000`1oo`0010000000hol00>?o000000030?o600050000003oPP00oh800?n20000
00000`3oDP001@000000ocX00?lj003o>P00000000<0oc/000@000000?mF003oEP0000000`3oR@00
1@000000om@00?oD003oe000000000<0cOl000D0000005ko001No`00G_l000000002003o6`050000
0000oi/00?nK003oV`0000000`00hol01@00000006Oo001Wo`00Iol0000000<0203o00D0000006X0
o`1Z0?l0JP3o000000020;H0o`050000003/0?l0k03o0>`0o`0000000`3o0?P01@000000o`3h0?l0
n03o0?P0000000<0k03o00D000000;H0o`2f0?l0]P3o0000000206X0o`05000000080?l0203o00P0
o`0000000`00Iol01@0000000>?o003So`00hol0000000<00?nK00D00000003o6`00oa/00?lK0000
000205ko00050000003=o`00cOl00<go000000000`3oe0001@000000ohT00?n9003oR@00000000<0
oeH000D000000?lk003o>`00oc/0000000020?lj00050000003oDP00oe800?mB000000000`3oPP00
1@000000olH00?o6003oaP00000000<0hol000D0000007oo001oo`00Ool00000000201?o00050000
0000oe`00?mL003oG00000000`00olX01@0000000<co003<o`00c?l0000000<006ko00D00000000N
o`007_l001ko0000000201l0o`050000001:0?l0BP3o04X0o`0000000`1Q0?l01@000000H`3o06<0
o`1S0?l0000000<0D03o00D0000002/0o`0[0?l0:`3o000000020009o`0500000000Bol004_o001;
o`0000000`00Uol01@0000000>Wo003Yo`00jOl0000000<00?o200D00000003oL000og000?m`0000
0002003o8`050000000Qo`008Ol0027o000000000`1Lo`001@000000ROl008Wo0029o`00000000<0
Z?l000@000000;Wo002io`0000000`2ko`001@000000[ol00:oo002_o`00000000<0U_l000D00000
07?o001co`00Lol00000000304Oo00030000003oool0oooo00@0oooo0@0000010?ooo`00803oool0
0`000000oooo0?ooo`050?ooo`0400000017o`00Aol0000000<0Lol000D0000009Ko002Fo`00U_l0
000000030:oo00050000002ko`00^ol00;_o000000000`2io`0010000000Z?l00:So0000000308Wo
00050000001Lo`00G?l005co000000000`0Qo`001@0000000?lS003o8`00ob<0000000<00?m`00@0
0000003o`P00ol8000000`00jOl01@00000009Oo002Go`00Uol0000000<004_o00D000000009o`00
2Ol000Wo0000000302/0o`040000001@0?l0D03o000000<0H`3o00D000000640o`1Q0?l0H@3o0000
000304X0o`050000000O0?l07`3o01l0o`0000000`007_l01000000006ko001^o`000003003<o`05
00000000olX00?o:003obP0000000`00oe`01@0000004ol001?o000Co`00000000<0Ool000@00000
0>?o003So`0000000`3oaP001@000000oh800?n2003oPP00000000<0oe8000D000000?lj003o>P00
ocX0000000030?lk00040000003oEP00oeH0000000<0ohT000D000000?oD003oe000om@000000003
0<go00050000001No`00G_l005ko000000000P00oa/01@0000000?nK003oV`00oi/0000000<00>?o
00D00000001Wo`00Iol006Oo0000000300P0o`050000001Z0?l0JP3o06X0o`0000000P2f0?l01@00
0000k03o0>`0o`3/0?l0000000<0o`3h00D000000?l0n03o0?P0o`3h000000030>`0o`050000002f
0?l0]P3o0;H0o`0000000P1Z0?l01@000000203o00P0o`080?l0000000<006Oo00D00000003So`00
hol00>?o00000003003oV`0500000000oa/00?lK003o6`0000000P1No`001@000000cOl00<go003=
o`00000000<0om@000D000000?n9003oR@00ohT0000000030?mF00050000003o>`00oc/00?lk0000
00000P3o>P001@000000oe800?mB003oDP00000000<0oh8000D000000?o6003oaP00olH000000003
0>?o00050000001oo`00Ool007oo000000000P0Co`001@0000000?mL003oG000oe`0000000<00?o:
00D00000003<o`00c?l00<co00000003001^o`05000000007_l001ko000No`0000000P0O0?l01@00
0000BP3o04X0o`1:0?l0000000<0H@3o00D0000006<0o`1S0?l0H`3o000000030500o`050000000[
0?l0:`3o02/0o`0000000P002Ol01@00000004_o001;o`00Bol0000000<009Oo00D00000003Yo`00
jOl00>Wo00000003003o`P0500000000og000?m`003oL00000000P00ob<01@0000008Ol0027o000Q
o`00000000<0G?l000D0000008Wo0029o`00ROl0000000030:So00040000002io`00^Ol0000000<0
^ol000D000000:oo002_o`00[ol00000000309Ko00050000001co`00Lol007?o000000000`17o`00
0`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo0000
084000001P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003o1@00
o`D000000`0ao`001@000000Hol006?o001So`00000000<0SOl000D000000:co002/o`00[?l00000
00030;ko000400000031o`00`Ol0000000<0]Ol000D0000009Wo002Io`00VOl00000000306go0005
0000000do`00=?l003Co000000000`00oa40100000000?mP003oH0000003003o]@0500000000l_l0
0?;o003bo`0000000`00WOl01@00000004oo001?o`00Col0000000<000[o00@0000002/0o`0[0?l0
00000`1A0?l01@000000HP3o0680o`1R0?l0000000<0GP3o00D0000004@0o`140?l0A03o00000003
01@0o`0400000000;ol002oo000000<008Co00D00000003Wo`00iol00>Oo00000003003oZ`050000
0000ocT00?li003o>@0000000`0go`0010000000X_l00:;o000000030?om00050000003o[000oj`0
0?n/000000000`3oL0001@000000od`00?m<003oC000000000<0od4000@000000?mA003oD@000000
0`3oN`001@000000okh00?nn003o_P00000000<0iol000D0000007co001lo`00O?l00000000200?o
000500000000ogd00?mm003oO@0000000`00ool01@000000087o0021o`00POl0000000<000go00D0
000005H0o`1F0?l0EP3o000000020:H0o`050000003M0?l0g@3o0=d0o`0000000`3j0?l01@000000
nP3o0?X0o`3j0?l0000000<0g@3o00D000000:H0o`2V0?l0YP3o0000000205H0o`05000000003Ol0
00go000=o`0000000`00POl01@0000000?oo003oo`00ool0000000<00?mm00D0000000?o0003o`00
0ol00000000207co00050000003Wo`00iol00>Oo000000000`3o_P001@000000og/00?mk003oN`00
000000<0oe4000D000000?m1003o@@00od40000000020?m<00050000003oL000og000?m`00000000
0`3o[0001@000000ood00?om003oo@00000000<0X_l000D0000003Oo000go`00=ol000000002003o
>@0500000000oj/00?n[003oZ`0000000`00iol01@00000008Co0024o`00Q?l0000000<002oo00D0
000001@0o`0D0?l0503o0000000204@0o`050000001N0?l0GP3o05h0o`0000000`1R0?l01@000000
D@3o0540o`1A0?l0000000<0:`3o00D00000000:o`002_l000[o00000002001?o`0500000000WOl0
09go002Mo`0000000`00l_l01@0000000?ne003o]@00okD0000000<00?mP00D00000003o4@00oa40
0?lA0000000203Co00050000001]o`00KOl006go000000000`2Io`001@000000]Ol00;Go002eo`00
000000<0`Ol000@000000;ko002no`0000000`2/o`001@000000SOl008go002=o`00000000<0Hol0
00D00000037o000ao`00<Ol000000003003o1@030000003oool0oooo00@0oooo0@0000010?ooo`00
803oool00`000000oooo0?ooo`050?ooo`0400000000o`D00?l5000000<0<Ol000D0000006?o001S
o`00Hol00000000308go00050000002/o`00[?l00:co000000000`2no`0010000000`Ol00<7o0000
00030;Go00050000002Io`00VOl009Wo000000000`1]o`001@000000=?l003Co000do`00000000<0
0?lA00@00000003oH000of0000000`00okD01@0000000?;o003bo`00l_l0000000<009go00D00000
001?o`00Col004oo00000003000:o`040000000[0?l0:`3o000000<0D@3o00D000000680o`1R0?l0
HP3o0000000305h0o`05000000140?l0A03o04@0o`0000000`0D0?l01000000002oo000_o`000003
0024o`0500000000iol00>Oo003Wo`0000000`00oj/01@0000000?li003o>@00ocT0000000<0=ol0
00@000000:;o002Ro`0000000`3oo@001@000000oj`00?n/003o[000000000<0og0000D000000?m<
003oC000od`0000000030?m100040000003oD@00oe40000000<0og/000D000000?nn003o_P00okh0
000000030>Oo00050000001lo`00O?l007co000000000P03o`001@0000000?mm003oO@00ogd00000
00<00?oo00D000000021o`00POl0087o00000003000=o`050000001F0?l0EP3o05H0o`0000000P2V
0?l01@000000g@3o0=d0o`3M0?l0000000<0nP3o00D000000?X0o`3j0?l0nP3o000000030=d0o`05
0000002V0?l0YP3o0:H0o`0000000P1F0?l01@00000000go000=o`003Ol0000000<0087o00D00000
003oo`00ool00?oo00000003003oO@0500000003o`000ol000?o000000000P1lo`001@000000iol0
0>Oo003Wo`00000000<0okh000D000000?mk003oN`00og/0000000030?mA00050000003o@@00od40
0?m1000000000P3oC0001@000000og000?m`003oL000000000<0oj`000D000000?om003oo@00ood0
000000030:;o00050000000go`00=ol003Oo000000000P00ocT01@0000000?n[003oZ`00oj/00000
00<00>Oo00D000000024o`00Q?l008Co00000003000_o`050000000D0?l0503o01@0o`0000000P14
0?l01@000000GP3o05h0o`1N0?l0000000<0HP3o00D000000540o`1A0?l0D@3o0000000302/0o`05
000000002_l000[o000:o`0000000P00Col01@00000009go002Mo`00WOl0000000<00?;o00D00000
003o]@00okD00?ne00000003003oH00500000000oa400?lA003o4@0000000P0do`001@000000KOl0
06go001]o`00000000<0VOl000D000000;Go002eo`00]Ol0000000030<7o00040000002no`00__l0
000000<0[?l000D0000008go002=o`00SOl00000000306?o00050000000ao`00<Ol0037o00000000
0`00o`D00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3o
ool0100000000?l5003o1@000003037o00050000001So`00Hol006?o000000000`2=o`001@000000
[?l00:co002/o`00000000<0__l000@000000<7o0031o`0000000`2eo`001@000000VOl009Wo002I
o`00000000<0KOl000D0000003Co000do`00=?l000000003003o4@0400000000of000?mP000000<0
0?ne00D00000003bo`00l_l00?;o00000003002Mo`0500000000Col004oo001?o`0000000`002_l0
10000000:`3o02/0o`0000030540o`050000001R0?l0HP3o0680o`0000000`1N0?l01@000000A03o
04@0o`140?l0000000<0503o00@00000000_o`00;ol000000`00Q?l01@0000000>Oo003Wo`00iol0
000000<00?n[00D00000003o>@00ocT00?li0000000303Oo00040000002Ro`00X_l0000000<0ood0
00D000000?n/003o[000oj`0000000030?m`00050000003oC000od`00?m<000000000`3o@@001000
0000oe400?mA000000030?mk00050000003o_P00okh00?nn000000000`3Wo`001@000000O?l007co
001lo`00000000800ol000D00000003oO@00ogd00?mm00000003003oo`0500000000POl0087o0021
o`0000000`003Ol01@000000EP3o05H0o`1F0?l000000080YP3o00D000000=d0o`3M0?l0g@3o0000
00030?X0o`050000003j0?l0nP3o0?X0o`0000000`3M0?l01@000000YP3o0:H0o`2V0?l000000080
EP3o00D00000000=o`003Ol000go000000030021o`0500000000ool00?oo003oo`0000000`00ogd0
1@0000000ol000?o0003o`0000000080O?l000D000000>Oo003Wo`00iol0000000030?nn00050000
003oN`00og/00?mk000000000`3oD@001@000000od400?m1003o@@0000000080od`000D000000?m`
003oL000og00000000030?n/00050000003oo@00ood00?om000000000`2Ro`001@000000=ol003Oo
000go`00000000800?li00D00000003oZ`00oj/00?n[00000003003Wo`0500000000Q?l008Co0024
o`0000000`00;ol01@000000503o01@0o`0D0?l000000080A03o00D0000005h0o`1N0?l0GP3o0000
00030680o`050000001A0?l0D@3o0540o`0000000`0[0?l01@00000000[o000:o`002_l000000080
04oo00D00000002Mo`00WOl009go00000003003bo`0500000000okD00?ne003o]@0000000`00of00
1@0000000?lA003o4@00oa4000000080=?l000D0000006go001]o`00KOl00000000309Wo00050000
002eo`00]Ol00;Go000000000`31o`0010000000__l00;ko000000030:co00050000002=o`00SOl0
08go000000000`1So`001@000000<Ol0037o000ao`00000000<00?l500<000000?ooo`3oool0103o
ool100000040oooo000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@000001
0?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000oe`00?mL000000<00?lN00D00000
01co000Lo`007?l00000000305Co000500000023o`00Pol008?o000000000`2Xo`0010000000_ol0
0;oo000000030<Oo00050000002oo`00_ol00;oo000000000`2Uo`001@000000Nol007_o001ko`00
000000<0@Ol000@00000003o1@00o`D000000`00oeH01@0000000?n^003o[P00ojh0000000<00?Oo
00D00000002Po`00X?l00:3o00000003001?o`04000000002Ol000Wo000000<0;@3o00D000000540
o`1A0?l0D@3o000000030600o`050000001H0?l0F03o05P0o`0000000`0i0?l010000000103o00@0
o`0000030015o`0500000000X?l00:3o002Po`0000000`00ooH01@0000000?n5003oQ@00ohD00000
00<00?lA00@00000063o001Po`0000000`38o`001@000000om`00?oL003og000000000<0oi@000D0
00000?mS003oH`00of<0000000030?m=00040000003oD@00oe40000000<0og8000D000000?n]003o
[@00ojd0000000030?oo00050000002Go`00Uol009Oo000000000P0Po`001@0000000?mP003oH000
of00000000<00?oT00D00000002Ko`00Vol009_o00000003000To`05000000130?l0@`3o04<0o`00
00000P2E0?l01@000000c`3o0<l0o`3?0?l0000000<0k03o00D000000>`0o`3/0?l0k03o00000003
0<l0o`050000002E0?l0U@3o09D0o`0000000P130?l01@00000002Co000To`009?l0000000<009_o
00D00000003oi000on@00?oT00000003003oH0050000000Po`008?l0023o000000000P2Go`001@00
0000ool00?oo003oo`00000000<0ojd000D000000?mb003oLP00og80000000030?mA00050000003o
C@00odd00?m=000000000P3oH`001@000000oi@00?nD003oU000000000<0om`000D000000<So0038
o`00b?l000000003063o000500000000oa400?lA003o4@0000000P00ohD01@0000000?of003omP00
ooH0000000<00:3o00D000000015o`00AOl004Go0000000300@0o`050000000i0?l0>@3o03T0o`00
00000P1H0?l01@000000H03o0600o`1P0?l0000000<0D@3o00D0000002d0o`0]0?l0;@3o00000003
0009o`0500000000Col004oo001?o`0000000P00X?l01@0000000?Oo003go`00mol0000000<00?n^
00D00000003oEP00oeH00?mF00000003003o1@0500000011o`00@Ol0047o000000000P1ko`001@00
0000YOl00:Go002Uo`00000000<0_ol000D000000<Oo0037o`00aol0000000030;oo00040000002X
o`00Z?l0000000<0Pol000D0000005Co001Do`00E?l00000000301co000500000000oah00?lN003o
7P0000000`00oe`00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3o
ool01@3oool0100000000?mL003oG0000003003o7P050000000Lo`007?l001co000000000`1Do`00
1@000000Pol008?o0023o`00000000<0Z?l000@000000;oo002oo`0000000`37o`001@000000_ol0
0;oo002oo`00000000<0YOl000D0000007_o001ko`00Nol000000003047o000400000000o`D00?l5
000000<00?mF00D00000003o[P00ojh00?n^00000003003go`0500000000X?l00:3o002Po`000000
0`00Col01000000000Wo0009o`00000302d0o`050000001A0?l0D@3o0540o`0000000`1P0?l01@00
0000F03o05P0o`1H0?l0000000<0>@3o00@0000000@0o`040?l000000`00AOl01@0000000:3o002P
o`00X?l0000000<00?of00D00000003oQ@00ohD00?n500000003003o4@040000001Po`00H?l00000
00<0b?l000D000000?oL003og000om`0000000030?nD00050000003oH`00of<00?mS000000000`3o
C@0010000000oe400?mA000000030?mb00050000003o[@00ojd00?n]000000000`3oo`001@000000
Uol009Oo002Go`00000000808?l000D00000003oH000of000?mP00000003003oi00500000000Vol0
09_o002Ko`0000000`009?l01@000000@`3o04<0o`130?l000000080U@3o00D000000<l0o`3?0?l0
c`3o000000030>`0o`050000003/0?l0k03o0>`0o`0000000`3?0?l01@000000U@3o09D0o`2E0?l0
00000080@`3o00D00000000To`009?l002Co00000003002Ko`0500000000on@00?oT003oi0000000
0`00of001@0000008?l0023o000Po`0000000080Uol000D000000?oo003oo`00ool0000000030?n]
00050000003oLP00og800?mb000000000`3oD@001@000000odd00?m=003oC@0000000080of<000D0
00000?nD003oU000oi@0000000030?oL000500000038o`00b?l00<So000000000`1Po`001@000000
0?lA003o4@00oa40000000800?n500D00000003omP00ooH00?of00000003002Po`0500000000AOl0
04Go0015o`0000000`040?l01@000000>@3o03T0o`0i0?l000000080F03o00D000000600o`1P0?l0
H03o000000030540o`050000000]0?l0;@3o02d0o`0000000`002Ol01@00000004oo001?o`00Col0
000000800:3o00D00000003go`00mol00?Oo00000003003o[P0500000000oeH00?mF003oEP000000
0`00o`D01@000000@Ol0047o0011o`0000000080Nol000D000000:Go002Uo`00YOl0000000030;oo
000500000037o`00aol00<Oo000000000`2oo`0010000000Z?l00:So0000000308?o00050000001D
o`00E?l005Co000000000`0Lo`001@0000000?lN003o7P00oah0000000<00?mL00<000000?ooo`3o
ool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo
0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000ok@00?nd000000<00?mf
00D00000003o=P00ocH00?lf0000000300Wo000500000015o`00AOl004Go000000000`1io`001000
0000Xol00:?o000000030;oo00050000003;o`00bol00<_o000000000`35o`001@000000[Ol00:go
002]o`00000000<0Q?l000@0000004[o001:o`0000000`02o`001@0000000?mA003oD@00oe400000
00<00?n[00D00000003go`00mol00?Oo00000003002Mo`0400000000Bol004_o000000<000Co00D0
00000300o`0`0?l0<03o000000030580o`050000001M0?l0G@3o05d0o`0000000`1@0?l010000000
:`3o02/0o`000003000Ao`0500000000H_l006;o001Ro`0000000`00a?l01@0000000?o=003oc@00
old0000000<00?mH00@0000001co000Lo`0000000`2<o`001@000000l?l00?3o003`o`00000000<0
okd000D000000?n0003oP000oh00000000030?mM00040000003oE`00oeL0000000<0ofd000D00000
0?nO003oW`00oil0000000030?o/00050000002`o`00/?l00;3o000000000P0ko`001@0000000?m4
003oA000od@0000000<00?o900D00000002do`00]?l00;Co00000003000jo`050000000_0?l0;`3o
02l0o`0000000P250?l01@000000`03o0<00o`300?l0000000<0g`3o00D000000=l0o`3O0?l0g`3o
000000030<00o`05000000250?l0Q@3o08D0o`0000000P0_0?l01@00000003[o000jo`00>_l00000
00<00;Co00D00000003ob@00olT00?o900000003003oA0050000000ko`00>ol003_o000000000P2`
o`001@000000on`00?o/003ok000000000<0oil000D000000?m]003oK@00ofd0000000030?mG0005
0000003oG@00oed00?mM000000000P3oP0001@000000okd00?nm003o_@00000000<0l?l000D00000
08co002<o`00S?l00000000301co000500000000oeP00?mH003oF00000000P00old01@0000000<Co
0034o`00a?l0000000<006;o00D00000000Ao`004Ol0017o0000000302/0o`050000001@0?l0D03o
0500o`0000000P1M0?l01@000000DP3o0580o`1B0?l0000000<0<03o00D000000004o`001?l000Co
00000003001;o`0500000000WOl009go002Mo`0000000P00mol01@0000000?n[003oZ`00oj/00000
00<00?mA00D0000000;o0002o`000_l00000000304[o000500000024o`00Q?l008Co000000000P2]
o`001@000000aOl00<Go0035o`00000000<0bol000D000000;oo002oo`00_ol0000000030:?o0004
0000001io`00NOl0000000<0AOl000D0000000Wo0009o`002Ol000000003003o=P0500000000ogH0
0?mf003oMP0000000`00ok@00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<00000
0?ooo`3oool01@3oool0100000000?nd003o]0000003003oMP0500000000ocH00?lf003o=P000000
0`09o`001@000000AOl004Go0015o`00000000<0NOl000@000000:?o002So`0000000`2oo`001@00
0000bol00<_o003;o`00000000<0aOl000D000000:go002]o`00[Ol00000000308Co00040000001:
o`00B_l0000000<00_l000D00000003oD@00oe400?mA00000003003oZ`0500000000mol00?Oo003g
o`0000000`00WOl01000000004_o001;o`0000030004o`050000000`0?l0<03o0300o`0000000`1B
0?l01@000000G@3o05d0o`1M0?l0000000<0D03o00@0000002/0o`0[0?l000000`004Ol01@000000
06;o001Ro`00H_l0000000<00<Co00D00000003oc@00old00?o=00000003003oF0040000000Lo`00
7?l0000000<0S?l000D000000?3o003`o`00l?l0000000030?nm00050000003oP000oh000?n00000
00000`3oG@0010000000oeL00?mG000000030?m]00050000003oW`00oil00?nO000000000`3ok000
1@000000/?l00;3o002`o`0000000080>ol000D00000003oA000od@00?m400000003003ob@050000
0000]?l00;Co002do`0000000`00>_l01@000000;`3o02l0o`0_0?l000000080Q@3o00D000000<00
o`300?l0`03o000000030=l0o`050000003O0?l0g`3o0=l0o`0000000`300?l01@000000Q@3o08D0
o`250?l000000080;`3o00D00000000jo`00>_l003[o00000003002do`0500000000olT00?o9003o
b@0000000`00od@01@000000>ol003_o000ko`0000000080/?l000D000000?o/003ok000on`00000
00030?nO00050000003oK@00ofd00?m]000000000`3oE`001@000000oed00?mM003oG@0000000080
oh0000D000000?nm003o_@00okd0000000030?3o00050000002<o`00S?l008co000000000`0Lo`00
1@0000000?mH003oF000oeP0000000800?o=00D000000034o`00a?l00<Co00000003001Ro`050000
00004Ol0017o000Ao`0000000`0[0?l01@000000D03o0500o`1@0?l000000080G@3o00D000000580
o`1B0?l0DP3o000000030300o`05000000001?l000Co0004o`0000000`00Bol01@00000009go002M
o`00WOl0000000800?Oo00D00000003oZ`00oj/00?n[00000003003oD@0500000002o`000_l000;o
000000000`1:o`001@000000Q?l008Co0024o`0000000080[Ol000D000000<Go0035o`00aOl00000
00030<_o00050000002oo`00_ol00;oo000000000`2So`0010000000NOl007Wo0000000304Go0005
00000009o`002Ol000Wo000000000`00ocH01@0000000?mf003oMP00ogH0000000<00?nd00<00000
0?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003o
]000ok@000000`00ogH01@0000000?lf003o=P00ocH0000000<02Ol000D0000004Go0015o`00AOl0
0000000307Wo00040000002So`00Xol0000000<0_ol000D000000<_o003;o`00bol0000000030<Go
00050000002]o`00[Ol00:go000000000`24o`0010000000B_l004[o0000000300;o000500000000
oe400?mA003oD@0000000`00oj/01@0000000?Oo003go`00mol0000000<009go00@00000001;o`00
Bol000000`001?l01@000000<03o0300o`0`0?l0000000<0DP3o00D0000005d0o`1M0?l0G@3o0000
00030500o`040000000[0?l0:`3o000000<0017o00D00000001Ro`00H_l006;o000000030034o`05
00000000old00?o=003oc@0000000`00oeP0100000007?l001co0000000308co00050000003`o`00
l?l00?3o000000000`3o_@001@000000oh000?n0003oP000000000<0oed000@000000?mG003oE`00
00000`3oK@001@000000oil00?nO003oW`00000000<0on`000D000000;3o002`o`00/?l000000002
03_o000500000000od@00?m4003oA00000000`00olT01@0000000;Co002do`00]?l0000000<003[o
00D0000002l0o`0_0?l0;`3o0000000208D0o`05000000300?l0`03o0<00o`0000000`3O0?l01@00
0000g`3o0=l0o`3O0?l0000000<0`03o00D0000008D0o`250?l0Q@3o0000000202l0o`0500000000
>_l003[o000jo`0000000`00]?l01@0000000?o9003ob@00olT0000000<00?m400D0000003_o000k
o`00>ol0000000020;3o00050000003ok000on`00?o/000000000`3oW`001@000000ofd00?m]003o
K@00000000<0oeL000D000000?mM003oG@00oed0000000020?n000050000003o_@00okd00?nm0000
00000`3`o`001@000000S?l008co002<o`00000000<07?l000D00000003oF000oeP00?mH00000002
003oc@0500000000a?l00<Co0034o`0000000`00H_l01@000000017o000Ao`004Ol0000000<0:`3o
00D000000500o`1@0?l0D03o0000000205d0o`050000001B0?l0DP3o0580o`0000000`0`0?l01@00
000000Co0004o`001?l0000000<004_o00D00000002Mo`00WOl009go00000002003go`0500000000
oj/00?n[003oZ`0000000`00oe401@0000000_l000;o0002o`00000000<0B_l000D0000008Co0024
o`00Q?l0000000020:go000500000035o`00aOl00<Go000000000`3;o`001@000000_ol00;oo002o
o`00000000<0Xol000@0000007Wo001io`0000000`15o`001@0000002Ol000Wo0009o`00000000<0
0?lf00D00000003oMP00ogH00?mf00000003003o]0030000003oool0oooo00@0oooo0@0000010?oo
o`00803oool00`000000oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo
00<000000?ooo`3oool01@3oool0100000000?co003lo`000003003oc00500000000ohh00?n>003o
SP0000000`00od/01@0000000?l8003o2000o`P0000000<0=ol000@00000073o001`o`0000000`2N
o`001@000000__l00;ko002no`00000000<0cOl000D000000<Wo0039o`00bOl0000000030;?o0004
00000029o`00ROl0000000<0C_l000D0000000Co0004o`001?l000000003003oD@0500000000ojh0
0?n^003o[P0000000`00l_l01000000009Oo002Go`0000030013o`05000000010?l00@3o0040o`00
00000`0e0?l01@000000DP3o0580o`1B0?l0000000<0F03o00@0000004@0o`140?l000000`0G0?l0
1@00000002co000/o`00;?l0000000<008Go00D00000003^o`00k_l00>ko00000003003oWP040000
0000obL00?lW000000<0COl000D000000;Wo002io`00^Ol0000000030?oY00050000003oX@00oj40
0?nQ000000000`3oL`0010000000of400?mQ000000030?m/00050000003oUP00oiH00?nF00000000
0`3of`001@000000a_l00<Ko0036o`0000000080E?l000D00000003o:@00obT00?lY00000003003o
[`0500000000c?l00<co003<o`0000000`00D?l01@000000703o01`0o`0L0?l000000080M03o00D0
00000;40o`2a0?l0/@3o000000030=40o`050000003A0?l0d@3o0=40o`0000000`2a0?l01@000000
M03o07@0o`1d0?l000000080703o00D00000001@o`00D?l0053o00000003003<o`0500000000ojl0
0?n_003o[`0000000`00obT01@000000E?l005Co001Do`0000000080a_l000D000000?oK003of`00
om/0000000030?nF00050000003oK000of`00?m/000000000`3oH@001@000000og<00?mc003oL`00
00000080oj4000D000000?oY003oj@00onT0000000030;Wo00050000001=o`00COl004go00000000
0`00obL01@0000000?nN003oWP00oih0000000800>ko00D000000025o`00QOl008Go00000003000/
o`050000000G0?l05`3o01L0o`0000000`140?l01@000000F03o05P0o`1H0?l000000080DP3o00D0
000003D0o`0e0?l0=@3o000000030040o`0500000000@ol004?o0013o`0000000`00Uol01@000000
0?;o003bo`00l_l0000000800?n^00D00000003oD@00oe400?mA0000000300Co00050000001>o`00
C_l004ko000000000`29o`001@000000/ol00;?o002co`0000000080bOl000D000000<go003=o`00
cOl0000000030;ko00050000002No`00W_l009ko000000000`1`o`0010000000=ol003Oo00000003
003o200500000000od/00?m;003oB`0000000`00ohh01@0000000?o<003oc000ol`0000000<00?co
00<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@0
0000003lo`00o?l000000`00ol`01@0000000?n>003oSP00ohh0000000<00?m;00D00000003o2000
o`P00?l80000000303Oo00040000001`o`00L?l0000000<0W_l000D000000;ko002no`00__l00000
00030<go000500000039o`00bOl00<Wo000000000`2co`0010000000ROl008Wo0000000304ko0005
00000004o`001?l000Co000000000`00oe401@0000000?n^003o[P00ojh0000000<00?;o00@00000
002Go`00Uol000000`00@ol01@0000000@3o0040o`010?l0000000<0=@3o00D000000580o`1B0?l0
DP3o0000000305P0o`04000000140?l0A03o000000<05`3o00D00000000/o`00;?l002co00000003
0025o`0500000000k_l00>ko003^o`0000000`00oih0100000000?lW003o9`00000304go00050000
002io`00^Ol00;Wo000000000`3oj@001@000000oj400?nQ003oX@00000000<0og<000@000000?mQ
003oH@0000000`3oK0001@000000oiH00?nF003oUP00000000<0om/000D000000<Ko0036o`00a_l0
0000000205Co000500000000obT00?lY003o:@0000000`00ojl01@0000000<co003<o`00c?l00000
00<0053o00D0000001`0o`0L0?l0703o0000000207@0o`050000002a0?l0/@3o0;40o`0000000`3A
0?l01@000000d@3o0=40o`3A0?l0000000<0/@3o00D0000007@0o`1d0?l0M03o0000000201`0o`05
00000000D?l0053o001@o`0000000`00c?l01@0000000?n_003o[`00ojl0000000<00?lY00D00000
05Co001Do`00E?l0000000020<Ko00050000003of`00om/00?oK000000000`3oUP001@000000of`0
0?m/003oK000000000<0of4000D000000?mc003oL`00og<0000000020?nQ00050000003oj@00onT0
0?oY000000000`2io`001@000000COl004go001=o`00000000<00?lW00D00000003oWP00oih00?nN
00000002003^o`0500000000QOl008Go0025o`0000000`00;?l01@0000005`3o01L0o`0G0?l00000
00<0A03o00D0000005P0o`1H0?l0F03o000000020580o`050000000e0?l0=@3o03D0o`0000000`01
0?l01@00000004?o0013o`00@ol0000000<009Oo00D00000003bo`00l_l00?;o00000002003o[P05
00000000oe400?mA003oD@0000000`04o`001@000000C_l004ko001>o`00000000<0ROl000D00000
0;?o002co`00/ol0000000020<Wo00050000003=o`00cOl00<go000000000`2no`001@000000W_l0
09ko002No`00000000<0L?l000@0000003Oo000go`0000000`00o`P01@0000000?m;003oB`00od/0
000000<00?n>00D00000003oc000ol`00?o<00000003003lo`030000003oool0oooo00@0oooo0@00
00010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000o?l00?co000000<00?o<00D0
0000003oSP00ohh00?n>00000003003oB`0500000000o`P00?l8003o200000000`0go`0010000000
L?l0073o0000000309ko00050000002no`00__l00;ko000000000`3=o`001@000000bOl00<Wo0039
o`00000000<0/ol000@0000008Wo0029o`0000000`1>o`001@0000001?l000Co0004o`00000000<0
0?mA00D00000003o[P00ojh00?n^00000003003bo`0400000000Uol009Oo000000<004?o00D00000
0040o`010?l00@3o0000000303D0o`050000001B0?l0DP3o0580o`0000000`1H0?l010000000A03o
04@0o`00000301L0o`0500000000;?l002co000/o`0000000`00QOl01@0000000>ko003^o`00k_l0
000000<00?nN00@00000003o9`00obL000000`1=o`001@000000^Ol00;Wo002io`00000000<0onT0
00D000000?nQ003oX@00oj40000000030?mc00040000003oH@00of40000000<0of`000D000000?nF
003oUP00oiH0000000030?oK000500000036o`00a_l00<Ko000000000P1Do`001@0000000?lY003o
:@00obT0000000<00?n_00D00000003<o`00c?l00<co00000003001@o`050000000L0?l0703o01`0
o`0000000P1d0?l01@000000/@3o0;40o`2a0?l0000000<0d@3o00D000000=40o`3A0?l0d@3o0000
00030;40o`050000001d0?l0M03o07@0o`0000000P0L0?l01@000000053o001@o`00D?l0000000<0
0<co00D00000003o[`00ojl00?n_00000003003o:@050000001Do`00E?l005Co000000000P36o`00
1@000000om/00?oK003of`00000000<0oiH000D000000?m/003oK000of`0000000030?mQ00050000
003oL`00og<00?mc000000000P3oX@001@000000onT00?oY003oj@00000000<0^Ol000D0000004go
001=o`00COl000000003003o9`0500000000oih00?nN003oWP0000000P00k_l01@00000008Go0025
o`00QOl0000000<002co00D0000001L0o`0G0?l05`3o0000000304@0o`050000001H0?l0F03o05P0
o`0000000P1B0?l01@000000=@3o03D0o`0e0?l0000000<00@3o00D000000013o`00@ol004?o0000
0003002Go`0500000000l_l00?;o003bo`0000000P00ojh01@0000000?mA003oD@00oe40000000<0
1?l000D0000004ko001>o`00C_l00000000308Wo00050000002co`00/ol00;?o000000000P39o`00
1@000000cOl00<go003=o`00000000<0__l000D0000009ko002No`00W_l000000003073o00040000
000go`00=ol0000000<00?l800D00000003oB`00od/00?m;00000003003oSP0500000000ol`00?o<
003oc00000000`00o?l00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?oo
o`3oool01@3ooooo0000084000001P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0
oooo00@00000002lo`00_?l000000`00iol01@0000000?oR003ohP00on80000000<00?nR00D00000
003oGP00oeh00?mN00000003003o60040000000[o`00:ol0000000<0J?l000D0000009Wo002Io`00
VOl0000000030;co00050000003>o`00c_l00<ko000000000`3<o`0010000000]_l00;Ko00000003
08_o00050000001>o`00C_l004ko000000000`02o`001@0000000?mF003oEP00oeH0000000<00?ne
00@00000003Yo`00jOl000000`00S?l01@00000003Wo000io`00>Ol0000000<02P3o00D0000003T0
o`0i0?l0>@3o000000030540o`040000001@0?l0D03o000000<0<`3o00D000000002o`000_l000;o
00000003001?o`0500000000/?l00;3o002`o`0000000`00on00100000000?mY003oJ@00000300ko
000500000021o`00POl0087o000000000`3Vo`001@000000olH00?o6003oaP00000000<0oh`000@0
00000?m_003oK`0000000`3oL0001@000000oi000?n@003oT000000000<0olh000D000000=Wo003I
o`00fOl00000000206co000500000000oa400?lA003o4@0000000`00oiH01@0000000>?o003So`00
hol0000000<006Go00D0000000T0o`090?l02@3o0000000206<0o`050000002R0?l0XP3o0:80o`00
00000`330?l01@000000``3o0<<0o`330?l0000000<0XP3o00D0000006<0o`1S0?l0H`3o00000002
00T0o`0500000000IOl006Go001Uo`0000000`00hol01@0000000?nF003oUP00oiH0000000<00?lA
00D0000006co001/o`00K?l0000000020=Wo00050000003ocP00olh00?o>000000000`3oT0001@00
0000og000?m`003oL000000000<0ofl000D000000?n<003oS000oh`0000000020?o600050000003V
o`00i_l00>Ko000000000`21o`001@0000003_l000ko000>o`00000000<00?mY00D00000003oh000
on000?oP00000002002`o`0500000000Col004oo001?o`0000000`000_l01@000000<`3o03<0o`0c
0?l0000000<0D03o00D000000540o`1A0?l0D@3o0000000203T0o`050000000:0?l02P3o00X0o`00
00000`00>Ol01@00000008co002<o`00S?l0000000<00>Wo00D00000003o]@00okD00?ne00000002
003oEP0500000002o`000_l000;o000000000`1>o`001@000000Rol008_o002;o`00000000<0]_l0
00D000000<co003<o`00c?l0000000020<ko00050000002lo`00_?l00;co000000000`2Io`001@00
0000J?l006So001Xo`00000000<0:ol000@00000003o6000oaP000000`00oeh01@0000000?nR003o
XP00oj80000000<00?oR00D00000003Wo`00iol00>Oo00000003002lo`030000003oool0oooo00@0
oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000_?l00;co000000<0
0>Oo00D00000003ohP00on800?oR00000003003oXP0500000000oeh00?mN003oGP0000000`00oaP0
10000000:ol002_o0000000306So00050000002Io`00VOl009Wo000000000`2lo`001@000000c_l0
0<ko003>o`00000000<0c?l000@000000;Ko002fo`0000000`2;o`001@000000C_l004ko001>o`00
000000<00_l000D00000003oEP00oeH00?mF00000003003o]@0400000000jOl00>Wo000000<008co
00D00000000io`00>Ol003Wo0000000300X0o`050000000i0?l0>@3o03T0o`0000000`1A0?l01000
0000D03o0500o`00000303<0o`05000000000_l000;o0002o`0000000`00Col01@0000000;3o002`
o`00/?l0000000<00?oP00@00000003oJ@00ofT000000`0>o`001@000000POl0087o0021o`000000
00<0i_l000D000000?o6003oaP00olH0000000030?n<00040000003oK`00ofl0000000<0og0000D0
00000?n@003oT000oi00000000030?o>00050000003Io`00fOl00=Wo000000000P1/o`001@000000
0?lA003o4@00oa40000000<00?nF00D00000003So`00hol00>?o00000003001Uo`05000000090?l0
2@3o00T0o`0000000P1S0?l01@000000XP3o0:80o`2R0?l0000000<0``3o00D000000<<0o`330?l0
``3o000000030:80o`050000001S0?l0H`3o06<0o`0000000P090?l01@00000006Go001Uo`00IOl0
000000<00>?o00D00000003oUP00oiH00?nF00000003003o4@050000001/o`00K?l006co00000000
0P3Io`001@000000olh00?o>003ocP00000000<0oi0000D000000?m`003oL000og00000000030?m_
00050000003oS000oh`00?n<000000000P3oaP001@000000i_l00>Ko003Vo`00000000<0POl000D0
000000ko000>o`003_l000000003003oJ@0500000000on000?oP003oh00000000P00/?l01@000000
04oo001?o`00Col0000000<000;o00D0000003<0o`0c0?l0<`3o000000030500o`050000001A0?l0
D@3o0540o`0000000P0i0?l01@0000002P3o00X0o`0:0?l0000000<003Wo00D00000002<o`00S?l0
08co00000003003Yo`0500000000okD00?ne003o]@0000000P00oeH01@0000000_l000;o0002o`00
000000<0C_l000D0000008_o002;o`00Rol0000000030;Ko00050000003<o`00c?l00<co00000000
0P3>o`001@000000_?l00;co002lo`00000000<0VOl000D0000006So001Xo`00J?l00000000302_o
000400000000oaP00?lH000000<00?mN00D00000003oXP00oj800?nR00000003003ohP0500000000
iol00>Oo003Wo`0000000`00_?l00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<0
00000?ooo`3oool01@3oool0100000000;co002lo`000003003Wo`0500000000on800?oR003ohP00
00000`00oj801@0000000?mN003oGP00oeh0000000<00?lH00@0000002_o000[o`0000000`1Xo`00
1@000000VOl009Wo002Io`00000000<0_?l000D000000<ko003>o`00c_l0000000030<co00040000
002fo`00]_l0000000<0Rol000D0000004ko001>o`00C_l00000000300;o000500000000oeH00?mF
003oEP0000000`00okD0100000000>Wo003Yo`000003002<o`0500000000>Ol003Wo000io`000000
0`0:0?l01@000000>@3o03T0o`0i0?l0000000<0D@3o00@000000500o`1@0?l000000`0c0?l01@00
000000;o0002o`000_l0000000<004oo00D00000002`o`00/?l00;3o00000003003oh00400000000
ofT00?mY000000<03_l000D00000087o0021o`00POl0000000030>Ko00050000003oaP00olH00?o6
000000000`3oS00010000000ofl00?m_000000030?m`00050000003oT000oi000?n@000000000`3o
cP001@000000fOl00=Wo003Io`0000000080K?l000D00000003o4@00oa400?lA00000003003oUP05
00000000hol00>?o003So`0000000`00IOl01@0000002@3o00T0o`090?l000000080H`3o00D00000
0:80o`2R0?l0XP3o000000030<<0o`05000000330?l0``3o0<<0o`0000000`2R0?l01@000000H`3o
06<0o`1S0?l0000000802@3o00D00000001Uo`00IOl006Go00000003003So`0500000000oiH00?nF
003oUP0000000`00oa401@000000K?l006co001/o`0000000080fOl000D000000?o>003ocP00olh0
000000030?n@00050000003oL000og000?m`000000000`3oK`001@000000oh`00?n<003oS0000000
0080olH000D000000>Ko003Vo`00i_l000000003087o00050000000>o`003_l000ko000000000`00
ofT01@0000000?oP003oh000on00000000800;3o00D00000001?o`00Col004oo000000030002o`05
0000000c0?l0<`3o03<0o`0000000`1@0?l01@000000D@3o0540o`1A0?l000000080>@3o00D00000
00X0o`0:0?l02P3o00000003000io`0500000000S?l008co002<o`0000000`00jOl01@0000000?ne
003o]@00okD0000000800?mF00D0000000;o0002o`000_l00000000304ko00050000002;o`00Rol0
08_o000000000`2fo`001@000000c?l00<co003<o`0000000080c_l000D000000;co002lo`00_?l0
0000000309Wo00050000001Xo`00J?l006So000000000`0[o`00100000000?lH003o60000003003o
GP0500000000oj800?nR003oXP0000000`00on801@0000000>Oo003Wo`00iol0000000<00;co00<0
00000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0ooooo`000021
000000H0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000T_l009;o
000000<00:go00D00000003Fo`00e_l00=Ko00000003003ol`0500000000ok<00?nc003o/`000000
0`00ofd0100000000?lT003o9000000302;o00050000001Qo`00HOl0067o000000000`2Fo`001@00
0000^ol00;_o002ko`00000000<0c_l000@000000<go003=o`0000000`2fo`001@000000ROl008Wo
0029o`00000000<0B_l000D00000003o1@00o`D00?l500000003003oH00400000000ol800?o20000
00<00=[o00D00000001mo`00OOl007go00000003000/o`050000000C0?l04`3o01<0o`0000000`0m
0?l010000000CP3o04h0o`00000304<0o`050000000M0?l07@3o01d0o`0000000`008_l01@000000
07Wo001io`00NOl0000000<00>;o00@00000003oZ@00ojT000000`00obl01@000000Aol004Oo0017
o`00000000<0]?l000D000000?o_003ok`00onl0000000030?nZ00040000003oP@00oh40000000<0
ogL000D000000?n>003oSP00ohh0000000030?o300050000003Zo`00j_l00>[o000000000P21o`00
1@0000001_l000Ko0006o`00000000<00?mn00D00000003jo`00n_l00?[o00000003001jo`050000
00002?l000So0008o`0000000P1C0?l01@000000U03o09@0o`2D0?l0000000<0]@3o00D000000;D0
o`2e0?l0]@3o0000000309@0o`050000001C0?l0D`3o05<0o`0000000P002?l01@00000007[o001j
o`00N_l0000000<00?[o00D00000003oOP00ogh00?mn0000000300Ko000500000021o`00POl0087o
000000000P3Zo`001@000000ol<00?o3003o``00000000<0ohh000D000000?mg003oM`00ogL00000
00030?n100050000003oZP00ojX00?nZ000000000P3ok`001@000000]?l00;Co002do`00000000<0
Aol000D00000003o;`00obl00?l_00000003003oZ@0500000000h_l00>;o003Ro`0000000P00NOl0
1@00000002;o000Ro`008_l0000000<07@3o00D0000004<0o`130?l0@`3o0000000304h0o`050000
000m0?l0?@3o03d0o`0000000P0C0?l01@00000002co000/o`00;?l0000000<007go00D00000003J
o`00f_l00=[o00000003003o`P0500000000of000?mP003oH00000000P00o`D01@000000B_l004[o
001:o`00000000<0ROl000D000000;Ko002fo`00]_l0000000030<go00050000003>o`00c_l00<ko
000000000P2ko`001@000000U_l009Ko002Fo`00000000<0HOl000D0000002;o000Ro`008_l00000
0003003o900400000000ofd00?m]000000<00?nc00D00000003ol`00oo<00?oc00000003003Fo`05
00000000[Ol00:go002]o`0000000`00T_l00`000000oooo0?ooo`040?ooo`4000000@3oool00200
oooo0P0000060?ooo`0400000000T_l009;o000000<00:go00D00000003Fo`00e_l00=Ko00000003
003ol`0500000000ok<00?nc003o/`0000000`00ofd0100000000?lT003o9000000302;o00050000
001Qo`00HOl0067o000000000`2Fo`001@000000^ol00;_o002ko`00000000<0c_l000@000000<go
003=o`0000000`2fo`001@000000ROl008Wo0029o`00000000<0B_l000D00000003o1@00o`D00?l5
00000003003oH00400000000ol800?o2000000<00=[o00D00000001mo`00OOl007go00000003000/
o`050000000C0?l04`3o01<0o`0000000`0m0?l010000000CP3o04h0o`00000304<0o`050000000M
0?l07@3o01d0o`0000000`008_l01@00000007Wo001io`00NOl0000000<00>;o00@00000003oZ@00
ojT000000`00obl01@000000Aol004Oo0017o`00000000<0]?l000D000000?o_003ok`00onl00000
00030?nZ00040000003oP@00oh40000000<0ogL000D000000?n>003oSP00ohh0000000030?o30005
0000003Zo`00j_l00>[o000000000P21o`001@0000001_l000Ko0006o`00000000<00?mn00D00000
003jo`00n_l00?[o00000003001jo`05000000002?l000So0008o`0000000P1C0?l01@000000U03o
09@0o`2D0?l0000000<0]@3o00D000000;D0o`2e0?l0]@3o0000000309@0o`050000001C0?l0D`3o
05<0o`0000000P002?l01@00000007[o001jo`00N_l0000000<00?[o00D00000003oOP00ogh00?mn
0000000300Ko000500000021o`00POl0087o000000000P3Zo`001@000000ol<00?o3003o``000000
00<0ohh000D000000?mg003oM`00ogL0000000030?n100050000003oZP00ojX00?nZ000000000P3o
k`001@000000]?l00;Co002do`00000000<0Aol000D00000003o;`00obl00?l_00000003003oZ@05
00000000h_l00>;o003Ro`0000000P00NOl01@00000002;o000Ro`008_l0000000<07@3o00D00000
04<0o`130?l0@`3o0000000304h0o`050000000m0?l0?@3o03d0o`0000000P0C0?l01@00000002co
000/o`00;?l0000000<007go00D00000003Jo`00f_l00=[o00000003003o`P0500000000of000?mP
003oH00000000P00o`D01@000000B_l004[o001:o`00000000<0ROl000D000000;Ko002fo`00]_l0
000000030<go00050000003>o`00c_l00<ko000000000P2ko`001@000000U_l009Ko002Fo`000000
00<0HOl000D0000002;o000Ro`008_l000000003003o900400000000ofd00?m]000000<00?nc00D0
0000003ol`00oo<00?oc00000003003Fo`0500000000[Ol00:go002]o`0000000`00T_l00`000000
oooo0?ooo`030?ooo`8000000@3oool00200oooo00<000000?ooo`3oool01@3oool01000000009;o
002Bo`000003002]o`0500000000e_l00=Ko003Fo`0000000`00oo<01@0000000?nc003o/`00ok<0
000000<00?m]00@00000003o9000ob@000000`0Ro`001@000000HOl0067o001Qo`00000000<0U_l0
00D000000;_o002ko`00^ol0000000030<ko00040000003=o`00cOl0000000<0]_l000D0000008Wo
0029o`00ROl00000000304[o000500000000o`D00?l5003o1@0000000`00of00100000000?o2003o
`P000003003Jo`0500000000OOl007go001mo`0000000`00;?l01@0000004`3o01<0o`0C0?l00000
00<0?@3o00@0000004h0o`1>0?l000000`130?l01@0000007@3o01d0o`0M0?l0000000<002;o00D0
0000001io`00NOl007Wo00000003003Ro`0400000000ojT00?nY000000<00?l_00D0000004Oo0017
o`00Aol0000000030;Co00050000003ok`00onl00?o_000000000`3oZP0010000000oh400?n10000
00030?mg00050000003oSP00ohh00?n>000000000`3o``001@000000j_l00>[o003Zo`0000000080
POl000D0000000Ko0006o`001_l000000003003oOP0500000000n_l00?[o003jo`0000000`00N_l0
1@00000000So0008o`002?l000000080D`3o00D0000009@0o`2D0?l0U03o000000030;D0o`050000
002e0?l0]@3o0;D0o`0000000`2D0?l01@000000D`3o05<0o`1C0?l00000008000So00D00000001j
o`00N_l007[o00000003003jo`0500000000ogh00?mn003oOP0000000`06o`001@000000POl0087o
0021o`0000000080j_l000D000000?o3003o``00ol<0000000030?n>00050000003oM`00ogL00?mg
000000000`3oP@001@000000ojX00?nZ003oZP0000000080onl000D000000;Co002do`00]?l00000
000304Oo000500000000obl00?l_003o;`0000000`00ojT01@0000000>;o003Ro`00h_l000000080
07Wo00D00000000Ro`008_l002;o0000000301d0o`05000000130?l0@`3o04<0o`0000000`1>0?l0
1@000000?@3o03d0o`0m0?l0000000804`3o00D00000000/o`00;?l002co00000003001mo`050000
0000f_l00=[o003Jo`0000000`00ol801@0000000?mP003oH000of00000000800?l500D0000004[o
001:o`00B_l00000000308Wo00050000002fo`00]_l00;Ko000000000`3=o`001@000000c_l00<ko
003>o`0000000080^ol000D0000009Ko002Fo`00U_l000000003067o00050000000Ro`008_l002;o
000000000`00ob@0100000000?m]003oK@000003003o/`0500000000oo<00?oc003ol`0000000`00
e_l01@0000000:go002]o`00[Ol0000000<009;o00<000000?ooo`3oool0103oool100000040oooo
000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?ooo`00803oool0
0`000000oooo0?ooo`050?ooo`0400000000P_l008;o000000<008Wo00D00000002Qo`00XOl00:7o
000000030039o`0500000000o_l00?ko003no`0000000`00ol00100000000?mh003oN0000003003o
;@050000000Ko`006ol001_o000000000`1No`001@000000U?l009Co002Do`00000000<0^ol000@0
00000<ko003>o`0000000`3<o`001@000000/ol00;?o002co`00000000<0Q?l000D00000047o0011
o`00@Ol000000003003o4@0400000000og000?m`000000<00?oD00D000000037o`00aol00<Oo0000
0003001[o`05000000007Ol001go000Mo`0000000`0M0?l010000000@03o0400o`00000304L0o`05
0000000b0?l0<P3o0380o`0000000`010?l01@00000004Wo0019o`00BOl0000000<00:[o00@00000
003oi@00onD000000`00of`01@0000003Ol000go000=o`00000000<0POl000D000000>Go003Uo`00
iOl0000000030?o;00040000003oU`00oiL0000000<0oh<000D000000?n?003oS`00ohl000000003
0?nl00050000003ho`00n?l00?So000000000P2Co`001@0000006ol001_o000Ko`00000000<00?mX
00D00000003ok`00onl00?o_00000003002?o`05000000006ol001_o000Ko`0000000P120?l01@00
0000Q@3o08D0o`250?l0000000<0Y`3o00D000000:L0o`2W0?l0Y`3o0000000308D0o`0500000012
0?l0@P3o0480o`0000000P006ol01@00000008oo002?o`00Sol0000000<00?o_00D00000003oJ000
ofP00?mX0000000301_o00050000002Co`00Tol009?o000000000P3ho`001@000000ok`00?nl003o
_000000000<0ohl000D000000?n3003oP`00oh<0000000030?nG00050000003ob`00ol/00?o;0000
00000P3Uo`001@000000POl0087o0021o`00000000<03Ol000D00000003oK000of`00?m/00000003
003oi@0500000000Z_l00:[o002Zo`0000000P00BOl01@0000000@3o0040o`010?l0000000<0<P3o
00D0000004L0o`170?l0A`3o000000030400o`050000000M0?l07@3o01d0o`0000000P007Ol01@00
000006_o001[o`00Jol0000000<00<Oo00D00000003oe000om@00?oD00000003003oL00500000000
oa400?lA003o4@0000000P11o`001@000000Q?l008Co0024o`00000000<0/ol000D000000<co003<
o`00c?l0000000030<ko00050000002ko`00^ol00;_o000000000P2Do`001@000000G_l005ko001N
o`00000000<06ol000D00000003o;@00obd00?l]00000003003oN00400000000ol000?o0000000<0
0?ko00D000000039o`00bOl00<Wo00000003002Qo`0500000000ROl008Wo0029o`0000000`00P_l0
0`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool01000
000008;o0022o`0000030029o`0500000000XOl00:7o002Qo`0000000`00bOl01@0000000?ko003n
o`00o_l0000000<00?o000@00000003oN000ogP000000`00obd01@0000006ol001_o000Ko`000000
00<0G_l000D0000009Co002Do`00U?l0000000030;_o00040000003>o`00c_l0000000<0c?l000D0
00000;?o002co`00/ol00000000308Co000500000011o`00@Ol0047o000000000`00oa4010000000
0?m`003oL0000003003oe00500000000aol00<Oo0037o`0000000`00Jol01@00000001go000Mo`00
7Ol0000000<07@3o00@000000400o`100?l000000`170?l01@000000<P3o0380o`0b0?l0000000<0
0@3o00D000000019o`00BOl004Wo00000003002Zo`0400000000onD00?oU000000<00?m/00D00000
00go000=o`003Ol000000003087o00050000003Uo`00iOl00>Go000000000`3ob`0010000000oiL0
0?nG000000030?n300050000003oS`00ohl00?n?000000000`3o_0001@000000n?l00?So003ho`00
00000080Tol000D0000001_o000Ko`006ol000000003003oJ00500000000onl00?o_003ok`000000
0`00Sol01@00000001_o000Ko`006ol000000080@P3o00D0000008D0o`250?l0Q@3o000000030:L0
o`050000002W0?l0Y`3o0:L0o`0000000`250?l01@000000@P3o0480o`120?l00000008001_o00D0
0000002?o`00Sol008oo00000003003ok`0500000000ofP00?mX003oJ00000000`0Ko`001@000000
Tol009?o002Co`0000000080n?l000D000000?nl003o_000ok`0000000030?n?00050000003oP`00
oh<00?n3000000000`3oU`001@000000ol/00?o;003ob`0000000080iOl000D00000087o0021o`00
POl00000000300go000500000000of`00?m/003oK00000000`00onD01@0000000:[o002Zo`00Z_l0
0000008004Wo00D000000040o`010?l00@3o000000030380o`05000000170?l0A`3o04L0o`000000
0`100?l01@0000007@3o01d0o`0M0?l00000008001go00D00000001[o`00Jol006_o000000030037
o`0500000000om@00?oD003oe00000000`00og001@0000000?lA003o4@00oa4000000080@Ol000D0
000008Co0024o`00Q?l0000000030;?o00050000003<o`00c?l00<co000000000`3>o`001@000000
^ol00;_o002ko`0000000080U?l000D0000005ko001No`00G_l00000000301_o000500000000obd0
0?l]003o;@0000000`00ogP0100000000?o0003o`0000003003no`0500000000bOl00<Wo0039o`00
00000`00XOl01@00000008Wo0029o`00ROl0000000<008;o00<000000?ooo`3oool0103oool10000
0040oooo000P0?ooo`030000003oool0oooo00D0oooo00@000000022o`00P_l000000`00ROl01@00
00000:7o002Qo`00XOl0000000<00<Wo00D00000003no`00o_l00?ko00000003003o`00400000000
ogP00?mh000000<00?l]00D0000001_o000Ko`006ol00000000305ko00050000002Do`00U?l009Co
000000000`2ko`0010000000c_l00<ko000000030<co00050000002co`00/ol00;?o000000000`24
o`001@000000@Ol0047o0011o`00000000<00?lA00@00000003oL000og0000000`00om@01@000000
0<Oo0037o`00aol0000000<006_o00D00000000Mo`007Ol001go0000000301d0o`04000000100?l0
@03o000000<0A`3o00D000000380o`0b0?l0<P3o000000030040o`0500000000BOl004Wo0019o`00
00000`00Z_l0100000000?oU003oi@000003003oK0050000000=o`003Ol000go000000000`21o`00
1@000000iOl00>Go003Uo`00000000<0ol/000@000000?nG003oU`0000000`3oP`001@000000ohl0
0?n?003oS`00000000<0ok`000D000000?So003ho`00n?l00000000209?o00050000000Ko`006ol0
01_o000000000`00ofP01@0000000?o_003ok`00onl0000000<008oo00D00000000Ko`006ol001_o
000000020480o`05000000250?l0Q@3o08D0o`0000000`2W0?l01@000000Y`3o0:L0o`2W0?l00000
00<0Q@3o00D000000480o`120?l0@P3o00000002000Ko`0500000000Sol008oo002?o`0000000`00
onl01@0000000?mX003oJ000ofP0000000<06ol000D0000009?o002Co`00Tol0000000020?So0005
0000003o_000ok`00?nl000000000`3oS`001@000000oh<00?n3003oP`00000000<0oiL000D00000
0?o;003ob`00ol/0000000020>Go000500000021o`00POl0087o000000000`0=o`001@0000000?m/
003oK000of`0000000<00?oU00D00000002Zo`00Z_l00:[o000000020019o`05000000010?l00@3o
0040o`0000000`0b0?l01@000000A`3o04L0o`170?l0000000<0@03o00D0000001d0o`0M0?l07@3o
00000002000Mo`0500000000Jol006_o001[o`0000000`00aol01@0000000?oD003oe000om@00000
00<00?m`00D00000003o4@00oa400?lA00000002047o000500000024o`00Q?l008Co000000000`2c
o`001@000000c?l00<co003<o`00000000<0c_l000D000000;_o002ko`00^ol00000000209Co0005
0000001No`00G_l005ko000000000`0Ko`001@0000000?l]003o;@00obd0000000<00?mh00@00000
003o`000ol0000000`00o_l01@0000000<Wo0039o`00bOl0000000<00:7o00D000000029o`00ROl0
08Wo000000030022o`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo
0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3o
ool01000000008_o002;o`000003001oo`0500000000Pol008?o0023o`0000000`00V?l01@000000
0;oo002oo`00_ol0000000<00?Co00@00000003obP00olX000000`00oh001@0000000?lc003o<`00
oc<0000000<06?l000D0000005co001Lo`00G?l00000000309Co00040000002ko`00^ol0000000<0
c_l000D000000<Wo0039o`00bOl0000000030:go00050000001ko`00Nol007_o000000000`0do`00
100000000?lS003o8`000003003oQ@0500000000on/00?o[003oj`0000000`00/Ol01@00000005Oo
001Go`00Eol0000000<000go00@0000002H0o`0V0?l000000`0o0?l01@000000>`3o03/0o`0k0?l0
000000<06P3o00D00000000So`008ol002?o00000003001io`0400000000hol00>?o000000<00?nV
00D00000003o:`00ob/00?l[0000000304co00050000002ho`00^?l00;So000000000`3ok`001000
0000ok400?na000000030?nB00050000003oU000oi@00?nD000000000`3o^0001@000000oo/00?ok
003on`0000000080Y?l000D0000002oo000_o`00;ol000000003003oDP0500000000omX00?oJ003o
fP0000000`00X_l01@00000002go000]o`00;Ol000000080<P3o00D0000007H0o`1f0?l0MP3o0000
000309X0o`050000002J0?l0VP3o09X0o`0000000`1f0?l01@000000<P3o0380o`0b0?l000000080
02go00D00000002Ro`00X_l00:;o00000003003ofP0500000000oe800?mB003oDP0000000`0_o`00
1@000000Y?l00:Co002To`0000000080oo/000D000000?nh003o^000okP0000000030?nD00050000
003oTP00oi800?nB000000000`3o/@001@000000onl00?o_003ok`0000000080^?l000D0000004co
001<o`00C?l000000003003o:`0500000000ojH00?nV003oYP0000000`00hol01@00000007Wo001i
o`00NOl00000008002?o00D0000001X0o`0J0?l06P3o0000000303/0o`050000000o0?l0?`3o03l0
o`0000000`0V0?l01@00000000go000=o`003Ol00000008005Oo00D00000002ao`00/Ol00;7o0000
0003003oj`0500000000ohD00?n5003oQ@0000000`00ob<01@000000=?l003Co000do`0000000080
Nol000D000000:go002]o`00[Ol0000000030<Wo00050000003>o`00c_l00<ko000000000`2ko`00
1@000000U?l009Co002Do`0000000080G?l000D0000001So000Ho`006?l000000003003o<`050000
0000oh000?n0003oP00000000`00olX0100000000?Co003do`000003002oo`0500000000V?l009So
002Ho`0000000`00Pol01@00000007oo001oo`00Ool0000000<008_o00<000000?ooo`3oool0103o
ool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000002;o`00Rol000000`00
Ool01@00000008?o0023o`00Pol0000000<009So00D00000002oo`00_ol00;oo00000003003do`04
00000000olX00?o:000000<00?n000D00000003o<`00oc<00?lc0000000301So00050000001Lo`00
G?l005co000000000`2Do`0010000000^ol00;_o000000030<ko000500000039o`00bOl00<Wo0000
00000`2]o`001@000000Nol007_o001ko`00000000<0=?l000@00000003o8`00ob<000000`00ohD0
1@0000000?o[003oj`00on/0000000<00;7o00D00000001Go`00Eol005Oo00000003000=o`040000
000V0?l09P3o000000<0?`3o00D0000003/0o`0k0?l0>`3o0000000301X0o`05000000008ol002?o
000So`0000000`00NOl0100000000>?o003So`000003003oYP0500000000ob/00?l[003o:`000000
0`1<o`001@000000^?l00;So002ho`00000000<0onl000@000000?na003o/@0000000`3oTP001@00
0000oi@00?nD003oU000000000<0okP000D000000?ok003on`00oo/0000000020:Co00050000000_
o`00;ol002oo000000000`00oe801@0000000?oJ003ofP00omX0000000<00:;o00D00000000]o`00
;Ol002go000000020380o`050000001f0?l0MP3o07H0o`0000000`2J0?l01@000000VP3o09X0o`2J
0?l0000000<0MP3o00D000000380o`0b0?l0<P3o00000002000]o`0500000000X_l00:;o002Ro`00
00000`00omX01@0000000?mB003oDP00oe80000000<0;ol000D000000:Co002To`00Y?l000000002
0?ok00050000003o^000okP00?nh000000000`3oU0001@000000oi800?nB003oTP00000000<0ok40
00D000000?o_003ok`00onl0000000020;So00050000001<o`00C?l004co000000000`00ob/01@00
00000?nV003oYP00ojH0000000<00>?o00D00000001io`00NOl007Wo00000002000So`050000000J
0?l06P3o01X0o`0000000`0k0?l01@000000?`3o03l0o`0o0?l0000000<09P3o00D00000000=o`00
3Ol000go00000002001Go`0500000000/Ol00;7o002ao`0000000`00on/01@0000000?n5003oQ@00
ohD0000000<00?lS00D0000003Co000do`00=?l00000000207_o00050000002]o`00[Ol00:go0000
00000`39o`001@000000c_l00<ko003>o`00000000<0^ol000D0000009Co002Do`00U?l000000002
05co00050000000Ho`006?l001So000000000`00oc<01@0000000?n0003oP000oh00000000<00?o:
00@00000003do`00m?l000000`00_ol01@00000009So002Ho`00V?l0000000<008?o00D00000001o
o`00Ool007oo00000003002;o`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`00
0000oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3o
ool01@3oool0100000000:co002/o`000003002=o`0500000000OOl007go001mo`0000000`00Ool0
1@00000009;o002Bo`00T_l0000000<00;So00@00000003]o`00kOl000000`00oll01@0000000?n4
003oQ000oh@0000000<00?ld00D0000001So000Ho`006?l00000000305ko00040000002Fo`00U_l0
000000<0_?l000D000000<go003=o`00cOl0000000030<Go00050000002Uo`00YOl00:Go00000000
0`1]o`00100000008Ol0027o00000003003o>@0500000000oih00?nN003oWP0000000`00nOl01@00
000009Oo002Go`00Uol0000000<0047o00@000000040o`010?l000000`0/0?l01@000000>P3o03X0
o`0j0?l0000000<0:@3o00D000000005o`001Ol000Go00000003001?o`0400000000/?l00;3o0000
00<00?oM00D00000003oHP00of800?mR0000000301Oo000500000029o`00ROl008Wo000000000`3Z
o`0010000000old00?o=000000030?nS00050000003oV`00oi/00?nK000000000`3o]P001@000000
oo<00?oc003ol`0000000080/ol000D0000004;o0012o`00@_l000000003003o?P0500000000olD0
0?o5003oa@0000000`00]_l01@00000003oo000oo`00?ol0000000808P3o00D0000006P0o`1X0?l0
J03o0000000308`0o`050000002<0?l0S03o08`0o`0000000`1X0?l01@0000008P3o0280o`0R0?l0
0000008003oo00D00000002fo`00]_l00;Ko00000003003oa@0500000000och00?ln003o?P000000
0`12o`001@000000/ol00;?o002co`0000000080oo<000D000000?nf003o]P00okH0000000030?nK
00050000003oX`00oj<00?nS000000000`3oc@001@000000j_l00>[o003Zo`0000000080ROl000D0
000001Oo000Go`005ol000000003003oHP0500000000omd00?oM003og@0000000`00/?l01@000000
04oo001?o`00Col00000008000Go00D0000002T0o`0Y0?l0:@3o0000000303X0o`050000000/0?l0
;03o02`0o`0000000`010?l01@000000047o0011o`00@Ol00000008009Oo00D00000003io`00nOl0
0?Wo00000003003oWP0500000000ocT00?li003o>@0000000`0Qo`001@000000KOl006go001]o`00
00000080YOl000D000000<Go0035o`00aOl0000000030<go00050000002lo`00_?l00;co00000000
0`2Fo`001@000000G_l005ko001No`00000000806?l000D00000003o=000oc@00?ld00000003003o
Q00500000000oll00?o?003oc`0000000`00kOl0100000000;So002ho`000003002Bo`0500000000
Ool007oo001oo`0000000`00OOl01@00000008go002=o`00SOl0000000<00:co00<000000?ooo`3o
ool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000002/o`00[?l0
00000`00SOl01@00000007go001mo`00OOl0000000<007oo00D00000002Bo`00T_l009;o00000003
002ho`0400000000kOl00>go000000<00?o?00D00000003oQ000oh@00?n400000003003o=0050000
000Ho`006?l001So000000000`1No`0010000000U_l009Ko000000030;co00050000003=o`00cOl0
0<go000000000`35o`001@000000YOl00:Go002Uo`00000000<0KOl000@00000027o000Qo`000000
0`00ocT01@0000000?nN003oWP00oih0000000<00?Wo00D00000002Go`00Uol009Oo000000030011
o`04000000010?l00@3o000000<0;03o00D0000003X0o`0j0?l0>P3o0000000302T0o`0500000000
1Ol000Go0005o`0000000`00Col0100000000;3o002`o`000003003og@0500000000of800?mR003o
HP0000000`0Go`001@000000ROl008Wo0029o`00000000<0j_l000@000000?o=003oc@0000000`3o
X`001@000000oi/00?nK003oV`00000000<0okH000D000000?oc003ol`00oo<0000000020;?o0005
00000012o`00@_l004;o000000000`00och01@0000000?o5003oa@00olD0000000<00;Ko00D00000
000oo`00?ol003oo000000020280o`050000001X0?l0J03o06P0o`0000000`2<0?l01@000000S03o
08`0o`2<0?l0000000<0J03o00D000000280o`0R0?l08P3o00000002000oo`0500000000]_l00;Ko
002fo`0000000`00olD01@0000000?ln003o?P00och0000000<0@_l000D000000;?o002co`00/ol0
000000020?oc00050000003o]P00okH00?nf000000000`3oV`001@000000oj<00?nS003oX`000000
00<0old000D000000>[o003Zo`00j_l00000000208Wo00050000000Go`005ol001Oo000000000`00
of801@0000000?oM003og@00omd0000000<00;3o00D00000001?o`00Col004oo000000020005o`05
0000000Y0?l0:@3o02T0o`0000000`0j0?l01@000000;03o02`0o`0/0?l0000000<00@3o00D00000
0011o`00@Ol0047o00000002002Go`0500000000nOl00?Wo003io`0000000`00oih01@0000000?li
003o>@00ocT0000000<08Ol000D0000006go001]o`00KOl0000000020:Go000500000035o`00aOl0
0<Go000000000`3=o`001@000000_?l00;co002lo`00000000<0U_l000D0000005ko001No`00G_l0
0000000201So000500000000oc@00?ld003o=00000000`00oh@01@0000000?o?003oc`00oll00000
00<00>go00@00000002ho`00^?l000000`00T_l01@00000007oo001oo`00Ool0000000<007go00D0
0000002=o`00SOl008go00000003002/o`030000003oool0oooo00@0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooo`0400000000[?l00:co000000<008go00D00000001mo`00OOl0
07go00000003001oo`0500000000T_l009;o002Bo`0000000`00^?l0100000000>go003]o`000003
003oc`0500000000oh@00?n4003oQ00000000`00oc@01@0000006?l001So000Ho`00000000<0G_l0
00@0000009Ko002Fo`0000000`2lo`001@000000cOl00<go003=o`00000000<0aOl000D000000:Go
002Uo`00YOl00000000306go00040000000Qo`008Ol0000000<00?li00D00000003oWP00oih00?nN
00000003003io`0500000000Uol009Oo002Go`0000000`00@Ol0100000000@3o0040o`00000302`0
o`050000000j0?l0>P3o03X0o`0000000`0Y0?l01@00000000Go0005o`001Ol0000000<004oo00@0
0000002`o`00/?l000000`00omd01@0000000?mR003oHP00of80000000<05ol000D0000008Wo0029
o`00ROl0000000030>[o00040000003oc@00old0000000<0oj<000D000000?nK003oV`00oi/00000
00030?nf00050000003ol`00oo<00?oc000000000P2co`001@000000@_l004;o0012o`00000000<0
0?ln00D00000003oa@00olD00?o500000003002fo`0500000000?ol003oo000oo`0000000P0R0?l0
1@000000J03o06P0o`1X0?l0000000<0S03o00D0000008`0o`2<0?l0S03o0000000306P0o`050000
000R0?l08P3o0280o`0000000P00?ol01@0000000;Ko002fo`00]_l0000000<00?o500D00000003o
?P00och00?ln0000000304;o00050000002co`00/ol00;?o000000000P3ol`001@000000okH00?nf
003o]P00000000<0oi/000D000000?nS003oX`00oj<0000000030?o=00050000003Zo`00j_l00>[o
000000000P29o`001@0000005ol001Oo000Go`00000000<00?mR00D00000003og@00omd00?oM0000
0003002`o`0500000000Col004oo001?o`0000000P001Ol01@000000:@3o02T0o`0Y0?l0000000<0
>P3o00D0000002`0o`0/0?l0;03o000000030040o`0500000000@Ol0047o0011o`0000000P00Uol0
1@0000000?Wo003io`00nOl0000000<00?nN00D00000003o>@00ocT00?li00000003027o00050000
001]o`00KOl006go000000000P2Uo`001@000000aOl00<Go0035o`00000000<0cOl000D000000;co
002lo`00_?l00000000309Ko00050000001No`00G_l005ko000000000P0Ho`001@0000000?ld003o
=000oc@0000000<00?n400D00000003oc`00oll00?o?00000003003]o`0400000000^?l00;So0000
00<009;o00D00000001oo`00Ool007oo00000003001mo`0500000000SOl008go002=o`0000000`00
[?l00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo
0000084000001P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003O
o`00gol000000`00/Ol01@00000008oo002?o`00Sol0000000<007co00D00000001lo`00O?l007co
00000003002>o`0400000000/ol00;?o000000<00>[o00D00000003od@00om400?oA00000003003o
Q00500000000oc<00?lc003o<`0000000`0Ko`0010000000HOl0067o0000000309Wo00050000002n
o`00__l00;ko000000000`3;o`001@000000_ol00;oo002oo`00000000<0VOl000@0000005co001L
o`0000000`0:o`001@0000000?mE003oE@00oeD0000000<00?nl00D00000003Ko`00fol00=_o0000
0003001ko`0400000000:ol002_o000000<03`3o00D0000002h0o`0^0?l0;P3o0000000302h0o`05
0000000?0?l03`3o00l0o`0000000`00;Ol01000000008Co0024o`000003003_o`0500000000oiL0
0?nG003oU`0000000`00oa/01@000000F_l005[o001Jo`00000000<0`_l000@000000?o/003ok000
00000`3o^0001@000000ojH00?nV003oYP00000000<0okP000D000000?o/003ok000on`000000002
0;oo00050000001Bo`00D_l005;o000000000`00ob/01@0000000?nb003o/P00ok80000000<00<Wo
00D00000001@o`00D?l0053o000000020180o`050000001I0?l0F@3o05T0o`0000000`1n0?l01@00
0000OP3o07h0o`1n0?l0000000<0F@3o00D000000180o`0B0?l04P3o00000002001@o`0500000000
bOl00<Wo0039o`0000000`00ok801@0000000?l[003o:`00ob/0000000<0D_l000D000000;oo002o
o`00_ol0000000020?o/00050000003o^000okP00?nh000000000`3oYP001@000000okP00?nh003o
^000000000<0on`000D000000<;o0032o`00`_l00000000205[o000500000000oa/00?lK003o6`00
00000`00oiL01@0000000>oo003_o`00kol0000000<008Co00D00000000]o`00;Ol002go00000002
00l0o`050000000^0?l0;P3o02h0o`0000000`0^0?l01@0000003`3o00l0o`0?0?l0000000<002_o
00D00000001ko`00Nol007_o00000002003Ko`0500000000ok`00?nl003o_00000000`00oeD01@00
00002_l000[o000:o`00000000<0G?l000D0000009Wo002Io`00VOl0000000020;oo00050000003;
o`00bol00<_o000000000`2no`001@000000VOl009Wo002Io`00000000<0HOl000D0000001_o000K
o`006ol000000002003o<`0500000000oh@00?n4003oQ00000000`00om401@0000000>[o003Zo`00
j_l0000000<00;?o00@00000002>o`00S_l000000`00O?l01@00000007co001lo`00O?l0000000<0
08oo00D00000002ao`00/Ol00;7o00000003003Oo`030000003oool0oooo00@0oooo0@0000010?oo
o`00803oool00`000000oooo0?ooo`050?ooo`0400000000gol00=oo000000<00;7o00D00000002?
o`00Sol008oo00000003001lo`0500000000O?l007co001lo`0000000`00S_l0100000000;?o002c
o`000003003Zo`0500000000om400?oA003od@0000000`00oh@01@0000000?lc003o<`00oc<00000
00<06ol000@00000067o001Qo`0000000`2Io`001@000000__l00;ko002no`00000000<0bol000D0
00000;oo002oo`00_ol00000000309Wo00040000001Lo`00G?l0000000<02_l000D00000003oE@00
oeD00?mE00000003003o_00500000000fol00=_o003Ko`0000000`00Nol01000000002_o000[o`00
000300l0o`050000000^0?l0;P3o02h0o`0000000`0^0?l01@0000003`3o00l0o`0?0?l0000000<0
02go00@000000024o`00Q?l000000`00kol01@0000000?nG003oU`00oiL0000000<00?lK00D00000
05[o001Jo`00F_l0000000030<;o00040000003ok000on`0000000<0okP000D000000?nV003oYP00
ojH0000000030?nh00050000003ok000on`00?o/000000000P2oo`001@000000D_l005;o001Bo`00
000000<00?l[00D00000003o/P00ok800?nb000000030039o`0500000000D?l0053o001@o`000000
0P0B0?l01@000000F@3o05T0o`1I0?l0000000<0OP3o00D0000007h0o`1n0?l0OP3o0000000305T0
o`050000000B0?l04P3o0180o`0000000P00D?l01@0000000<Wo0039o`00bOl0000000<00?nb00D0
0000003o:`00ob/00?l[0000000305;o00050000002oo`00_ol00;oo000000000P3ok0001@000000
okP00?nh003o^000000000<0ojH000D000000?nh003o^000okP0000000030?o/000500000032o`00
`_l00<;o000000000P1Jo`001@0000000?lK003o6`00oa/0000000<00?nG00D00000003_o`00kol0
0>oo000000030024o`0500000000;Ol002go000]o`0000000P0?0?l01@000000;P3o02h0o`0^0?l0
000000<0;P3o00D0000000l0o`0?0?l03`3o00000003000[o`0500000000Nol007_o001ko`000000
0P00fol01@0000000?nl003o_000ok`0000000<00?mE00D0000000[o000:o`002_l00000000305co
00050000002Io`00VOl009Wo000000000P2oo`001@000000bol00<_o003;o`00000000<0__l000D0
000009Wo002Io`00VOl000000003067o00050000000Ko`006ol001_o000000000P00oc<01@000000
0?n4003oQ000oh@0000000<00?oA00D00000003Zo`00j_l00>[o00000003002co`0400000000S_l0
08ko000000<007co00D00000001lo`00O?l007co00000003002?o`0500000000/Ol00;7o002ao`00
00000`00gol00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool0
1@3oool0100000000=oo003Oo`000003002ao`0500000000Sol008oo002?o`0000000`00O?l01@00
000007co001lo`00O?l0000000<008ko00@00000002co`00/ol000000`00j_l01@0000000?oA003o
d@00om40000000<00?n400D00000003o<`00oc<00?lc0000000301_o00040000001Qo`00HOl00000
00<0VOl000D000000;ko002no`00__l0000000030<_o00050000002oo`00_ol00;oo000000000`2I
o`0010000000G?l005co0000000300[o000500000000oeD00?mE003oE@0000000`00ok`01@000000
0=_o003Ko`00fol0000000<007_o00@00000000[o`00:ol000000`0?0?l01@000000;P3o02h0o`0^
0?l0000000<0;P3o00D0000000l0o`0?0?l03`3o00000003000]o`0400000000Q?l008Co000000<0
0>oo00D00000003oU`00oiL00?nG00000003003o6`050000001Jo`00F_l005[o000000000`32o`00
10000000on`00?o/000000030?nh00050000003oYP00ojH00?nV000000000`3o^0001@000000on`0
0?o/003ok00000000080_ol000D0000005;o001Bo`00D_l000000003003o:`0500000000ok800?nb
003o/P0000000`00bOl01@000000053o001@o`00D?l0000000804P3o00D0000005T0o`1I0?l0F@3o
0000000307h0o`050000001n0?l0OP3o07h0o`0000000`1I0?l01@0000004P3o0180o`0B0?l00000
0080053o00D000000039o`00bOl00<Wo00000003003o/P0500000000ob/00?l[003o:`0000000`1B
o`001@000000_ol00;oo002oo`0000000080on`000D000000?nh003o^000okP0000000030?nV0005
0000003o^000okP00?nh000000000`3ok0001@000000`_l00<;o0032o`0000000080F_l000D00000
003o6`00oa/00?lK00000003003oU`0500000000kol00>oo003_o`0000000`00Q?l01@00000002go
000]o`00;Ol0000000803`3o00D0000002h0o`0^0?l0;P3o0000000302h0o`050000000?0?l03`3o
00l0o`0000000`00:ol01@00000007_o001ko`00Nol0000000800=_o00D00000003o_000ok`00?nl
00000003003oE@050000000:o`002_l000[o000000000`1Lo`001@000000VOl009Wo002Io`000000
0080_ol000D000000<_o003;o`00bol0000000030;ko00050000002Io`00VOl009Wo000000000`1Q
o`001@0000006ol001_o000Ko`00000000800?lc00D00000003oQ000oh@00?n400000003003od@05
00000000j_l00>[o003Zo`0000000`00/ol01000000008ko002>o`000003001lo`0500000000O?l0
07co001lo`0000000`00Sol01@0000000;7o002ao`00/Ol0000000<00=oo00<000000?ooo`3oool0
103oool100000040oooo000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@00
00010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000on000?oP000000<00>Ko00D0
0000002eo`00]Ol00;Go00000003002Ao`0500000000O?l007co001lo`0000000`00N_l010000000
08co002<o`000003002bo`0500000000j_l00>[o003Zo`0000000`00oll01@0000000?n0003oP000
oh00000000<00?l]00@0000002;o000Ro`0000000`1Xo`001@000000W_l009ko002No`00000000<0
_ol000D000000<Oo0037o`00aol0000000030;Go000400000029o`00ROl0000000<0AOl000D00000
003o4@00oa400?lA00000003003oM@0500000000omh00?oN003ogP0000000`00^_l01000000005oo
001Oo`000003000Fo`050000000H0?l0603o01P0o`0000000`0Z0?l01@0000006`3o01/0o`0K0?l0
000000<001;o00@00000001No`00G_l000000`00`Ol01@0000000?o:003obP00olX0000000<00?m>
00D0000002_o000[o`00:ol00000000309Wo00040000003bo`00l_l0000000<0om0000D000000?nd
003o]000ok@0000000030?nl00050000003oj000onP00?oX000000000P3:o`001@000000HOl0067o
001Qo`00000000<00?lJ00D00000003oW`00oil00?nO00000003003Ko`0500000000HOl0067o001Q
o`0000000P030?l01@000000B`3o04/0o`1;0?l0000000<0L03o00D000000700o`1`0?l0L03o0000
000304/0o`05000000030?l00`3o00<0o`0000000P00HOl01@0000000=_o003Ko`00fol0000000<0
0?nO00D00000003o6P00oaX00?lJ00000003067o00050000003:o`00b_l00<[o000000000P3oj000
1@000000ok`00?nl003o_000000000<0ok@000D000000?o@003od000om00000000030?;o00050000
002Io`00VOl009Wo000000000P0[o`001@0000000?m>003oCP00odh0000000<00?o:00D000000031
o`00`Ol00<7o00000003001No`05000000004_l001;o000Bo`0000000P0K0?l01@000000:P3o02X0
o`0Z0?l0000000<0603o00D00000000Fo`005_l001Ko00000003001Oo`0500000000^_l00;[o002j
o`0000000P00omh01@0000000?me003oM@00ogD0000000<00?lA00D0000004Go0015o`00AOl00000
000308Wo00050000002eo`00]Ol00;Go000000000P37o`001@000000_ol00;oo002oo`00000000<0
W_l000D0000006So001Xo`00J?l00000000302;o000500000000obd00?l]003o;@0000000P00oh00
1@0000000?o?003oc`00oll0000000<00>[o00D00000002bo`00/_l00;;o00000003002<o`040000
0000N_l007[o000000<007co00D00000002Ao`00TOl0097o00000003002eo`0500000000i_l00>Ko
003Vo`0000000`00on000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?oo
o`3oool01@3oool0100000000?oP003oh0000003003Vo`0500000000]Ol00;Go002eo`0000000`00
TOl01@00000007co001lo`00O?l0000000<007[o00@00000002<o`00S?l000000`00/_l01@000000
0>[o003Zo`00j_l0000000<00?o?00D00000003oP000oh000?n000000003003o;@040000000Ro`00
8_l0000000<0J?l000D0000009ko002No`00W_l0000000030;oo000500000037o`00aol00<Oo0000
00000`2eo`0010000000ROl008Wo0000000304Go000500000000oa400?lA003o4@0000000`00ogD0
1@0000000?oN003ogP00omh0000000<00;[o00@00000001Oo`00Gol000000`005_l01@000000603o
01P0o`0H0?l0000000<0:P3o00D0000001/0o`0K0?l06`3o00000003000Bo`0400000000G_l005ko
000000<00<7o00D00000003obP00olX00?o:00000003003oCP050000000[o`00:ol002_o00000000
0`2Io`0010000000l_l00?;o000000030?o@00050000003o]000ok@00?nd000000000`3o_0001@00
0000onP00?oX003oj00000000080b_l000D00000067o001Qo`00HOl000000003003o6P0500000000
oil00?nO003oW`0000000`00fol01@000000067o001Qo`00HOl0000000800`3o00D0000004/0o`1;
0?l0B`3o000000030700o`050000001`0?l0L03o0700o`0000000`1;0?l01@0000000`3o00<0o`03
0?l000000080067o00D00000003Ko`00fol00=_o00000003003oW`0500000000oaX00?lJ003o6P00
00000`1Qo`001@000000b_l00<[o003:o`0000000080onP000D000000?nl003o_000ok`000000003
0?nd00050000003od000om000?o@000000000`3bo`001@000000VOl009Wo002Io`0000000080:ol0
00D00000003oCP00odh00?m>00000003003obP0500000000`Ol00<7o0031o`0000000`00G_l01@00
000001;o000Bo`004_l0000000806`3o00D0000002X0o`0Z0?l0:P3o0000000301P0o`0500000000
5_l001Ko000Fo`0000000`00Gol01@0000000;[o002jo`00^_l0000000800?oN00D00000003oM@00
ogD00?me00000003003o4@0500000015o`00AOl004Go000000000`29o`001@000000]Ol00;Go002e
o`0000000080aol000D000000;oo002oo`00_ol00000000309ko00050000001Xo`00J?l006So0000
00000`0Ro`001@0000000?l]003o;@00obd0000000800?n000D00000003oc`00oll00?o?00000003
003Zo`0500000000/_l00;;o002bo`0000000`00S?l01000000007[o001jo`000003001lo`050000
0000TOl0097o002Ao`0000000`00]Ol01@0000000>Ko003Vo`00i_l0000000<00?oP00<000000?oo
o`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003oh000
on0000000`00i_l01@0000000;Go002eo`00]Ol0000000<0097o00D00000001lo`00O?l007co0000
0003001jo`0400000000S?l008co000000<00;;o00D00000003Zo`00j_l00>[o00000003003oc`05
00000000oh000?n0003oP00000000`00obd0100000008_l002;o0000000306So00050000002No`00
W_l009ko000000000`2oo`001@000000aol00<Oo0037o`00000000<0]Ol000@0000008Wo0029o`00
00000`15o`001@0000000?lA003o4@00oa40000000<00?me00D00000003ogP00omh00?oN00000003
002jo`0400000000Gol005oo000000<001Ko00D0000001P0o`0H0?l0603o0000000302X0o`050000
000K0?l06`3o01/0o`0000000`004_l01000000005ko001No`0000030031o`0500000000olX00?o:
003obP0000000`00odh01@000000:ol002_o000[o`00000000<0VOl000@000000?;o003bo`000000
0`3od0001@000000ok@00?nd003o]000000000<0ok`000D000000?oX003oj000onP0000000020<[o
00050000001Qo`00HOl0067o000000000`00oaX01@0000000?nO003oW`00oil0000000<00=_o00D0
0000001Qo`00HOl0067o0000000200<0o`050000001;0?l0B`3o04/0o`0000000`1`0?l01@000000
L03o0700o`1`0?l0000000<0B`3o00D0000000<0o`030?l00`3o00000002001Qo`0500000000fol0
0=_o003Ko`0000000`00oil01@0000000?lJ003o6P00oaX0000000<0HOl000D000000<[o003:o`00
b_l0000000020?oX00050000003o_000ok`00?nl000000000`3o]0001@000000om000?o@003od000
000000<0l_l000D0000009Wo002Io`00VOl00000000202_o000500000000odh00?m>003oCP000000
0`00olX01@0000000<7o0031o`00`Ol0000000<005ko00D00000000Bo`004_l001;o0000000201/0
o`050000000Z0?l0:P3o02X0o`0000000`0H0?l01@00000001Ko000Fo`005_l0000000<005oo00D0
0000002jo`00^_l00;[o00000002003ogP0500000000ogD00?me003oM@0000000`00oa401@000000
AOl004Go0015o`00000000<0ROl000D000000;Go002eo`00]Ol0000000020<Oo00050000002oo`00
_ol00;oo000000000`2No`001@000000J?l006So001Xo`00000000<08_l000D00000003o;@00obd0
0?l]00000002003oP00500000000oll00?o?003oc`0000000`00j_l01@0000000;;o002bo`00/_l0
000000<008co00@00000001jo`00N_l000000`00O?l01@000000097o002Ao`00TOl0000000<00;Go
00D00000003Vo`00i_l00>Ko00000003003oh0030000003oool0oooo00@0oooo0@0000010?ooo`00
803oool00`000000oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<0
00000?ooo`3oool01@3oool0100000000?nJ003oVP000003003of00500000000k?l00>co003/o`00
00000`00^Ol01@00000009;o002Bo`00T_l0000000<007co00@00000001io`00NOl000000`00S?l0
1@0000000;?o002co`00/ol0000000<00>go00D00000003obP00olX00?o:00000003003oN0040000
0000ob@00?lT000000<0:ol000D00000073o001`o`00L?l0000000030:?o00050000002oo`00_ol0
0;oo000000000`31o`0010000000Z?l00:So0000000307Go00050000000Zo`00:_l002[o00000000
0`00oc801@0000000?nJ003oVP00oiX0000000<00?_o00@00000002Ho`00V?l000000`00@ol01@00
000000Go0005o`001Ol0000000<0703o00D0000001h0o`0N0?l07P3o000000030000o`0400000000
?ol003oo000000<009Wo00D00000003on000ooP00?oh00000003003oO`0500000000o`<00?l3003o
0`0000000`1_o`0010000000d?l00=3o000000030?oY00050000003oa000ol@00?o4000000000`3o
``001@000000onL00?oW003oi`0000000080d_l000D0000006ko001^o`00K_l000000003003o2@05
00000000ohh00?n>003oSP0000000`00k?l01@00000007;o001bo`00L_l00000008000co00D00000
03`0o`0l0?l0?03o000000030680o`050000001R0?l0HP3o0680o`0000000`0l0?l01@00000000co
000<o`003?l00000008007;o00D00000003/o`00k?l00>co00000003003oSP0500000000o`T00?l9
003o2@0000000`1^o`001@000000d_l00=;o003Bo`0000000080onL000D000000?o3003o``00ol<0
000000030?o400050000003oj@00onT00?oY000000000`3@o`001@000000Kol006oo001_o`000000
00800?l300D00000003oO`00ogl00?mo00000003003on00500000000VOl009Wo002Io`0000000`00
?ol01@000000003o0000o`000?l0000000807P3o00D0000001`0o`0L0?l0703o000000030005o`05
00000000@ol004?o0013o`0000000`00V?l01@0000000?_o003ko`00nol0000000800?nJ00D00000
003o<P00oc800?lb0000000302[o00050000001eo`00MOl007Go000000000`2Xo`001@000000`Ol0
0<7o0031o`0000000080_ol000D000000:?o002So`00Xol000000003073o00050000000[o`00:ol0
02_o000000000`00ob@01@0000000?mh003oN000ogP0000000800?o:00D00000003]o`00kOl00>go
00000003002co`0500000000S?l008co002<o`0000000`00NOl01000000007co001lo`000003002B
o`0500000000^Ol00;Wo002io`0000000`00k?l01@0000000?oH003of000omP0000000<00?nJ00<0
00000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000
003oVP00oiX000000`00omP01@0000000>co003/o`00k?l0000000<00;Wo00D00000002Bo`00T_l0
09;o00000003001lo`0400000000NOl007Wo000000<008co00D00000002co`00/ol00;?o00000003
003]o`0500000000olX00?o:003obP0000000`00ogP0100000000?lT003o9000000302_o00050000
001`o`00L?l0073o000000000`2So`001@000000_ol00;oo002oo`00000000<0`Ol000@000000:So
002Xo`0000000`1eo`001@000000:_l002[o000Zo`00000000<00?lb00D00000003oVP00oiX00?nJ
00000003003ko`0400000000V?l009So000000<004?o00D000000005o`001Ol000Go0000000301`0
o`050000000N0?l07P3o01h0o`0000000`000?l01000000003oo000oo`000003002Io`0500000000
ooP00?oh003on00000000`00ogl01@0000000?l3003o0`00o`<0000000<0Kol000@000000=3o003@
o`0000000`3oj@001@000000ol@00?o4003oa000000000<0ol<000D000000?oW003oi`00onL00000
00020=;o00050000001^o`00K_l006ko000000000`00o`T01@0000000?n>003oSP00ohh0000000<0
0>co00D00000001bo`00L_l007;o00000002000<o`050000000l0?l0?03o03`0o`0000000`1R0?l0
1@000000HP3o0680o`1R0?l0000000<0?03o00D00000000<o`003?l000co00000002001bo`050000
0000k?l00>co003/o`0000000`00ohh01@0000000?l9003o2@00o`T0000000<0K_l000D000000=;o
003Bo`00d_l0000000020?oW00050000003o``00ol<00?o3000000000`3oa0001@000000onT00?oY
003oj@00000000<0d?l000D0000006oo001_o`00Kol000000002003o0`0500000000ogl00?mo003o
O`0000000`00ooP01@00000009Wo002Io`00VOl0000000<003oo00D000000000o`000?l0003o0000
000201h0o`050000000L0?l0703o01`0o`0000000`001Ol01@00000004?o0013o`00@ol0000000<0
09So00D00000003ko`00nol00?_o00000002003oVP0500000000oc800?lb003o<P0000000`0Zo`00
1@000000MOl007Go001eo`00000000<0Z?l000D000000<7o0031o`00`Ol0000000020;oo00050000
002So`00Xol00:?o000000000`1`o`001@000000:ol002_o000[o`00000000<00?lT00D00000003o
N000ogP00?mh00000002003obP0500000000kOl00>go003]o`0000000`00/ol01@00000008co002<
o`00S?l0000000<007Wo00@00000001lo`00O?l000000`00T_l01@0000000;Wo002io`00^Ol00000
00<00>co00D00000003of000omP00?oH00000003003oVP030000003oool0oooo00@0oooo0@000001
0?ooo`00803oool2000000H0oooo00@00000003oVP00oiX000000`00omP01@0000000>co003/o`00
k?l0000000<00;Wo00D00000002Bo`00T_l009;o00000003001lo`0400000000NOl007Wo000000<0
08co00D00000002co`00/ol00;?o00000003003]o`0500000000olX00?o:003obP0000000`00ogP0
100000000?lT003o9000000302_o00050000001`o`00L?l0073o000000000`2So`001@000000_ol0
0;oo002oo`00000000<0`Ol000@000000:So002Xo`0000000`1eo`001@000000:_l002[o000Zo`00
000000<00?lb00D00000003oVP00oiX00?nJ00000003003ko`0400000000V?l009So000000<004?o
00D000000005o`001Ol000Go0000000301`0o`050000000N0?l07P3o01h0o`0000000`000?l01000
000003oo000oo`000003002Io`0500000000ooP00?oh003on00000000`00ogl01@0000000?l3003o
0`00o`<0000000<0Kol000@000000=3o003@o`0000000`3oj@001@000000ol@00?o4003oa0000000
00<0ol<000D000000?oW003oi`00onL0000000020=;o00050000001^o`00K_l006ko000000000`00
o`T01@0000000?n>003oSP00ohh0000000<00>co00D00000001bo`00L_l007;o00000002000<o`05
0000000l0?l0?03o03`0o`0000000`1R0?l01@000000HP3o0680o`1R0?l0000000<0?03o00D00000
000<o`003?l000co00000002001bo`0500000000k?l00>co003/o`0000000`00ohh01@0000000?l9
003o2@00o`T0000000<0K_l000D000000=;o003Bo`00d_l0000000020?oW00050000003o``00ol<0
0?o3000000000`3oa0001@000000onT00?oY003oj@00000000<0d?l000D0000006oo001_o`00Kol0
00000002003o0`0500000000ogl00?mo003oO`0000000`00ooP01@00000009Wo002Io`00VOl00000
00<003oo00D000000000o`000?l0003o0000000201h0o`050000000L0?l0703o01`0o`0000000`00
1Ol01@00000004?o0013o`00@ol0000000<009So00D00000003ko`00nol00?_o00000002003oVP05
00000000oc800?lb003o<P0000000`0Zo`001@000000MOl007Go001eo`00000000<0Z?l000D00000
0<7o0031o`00`Ol0000000020;oo00050000002So`00Xol00:?o000000000`1`o`001@000000:ol0
02_o000[o`00000000<00?lT00D00000003oN000ogP00?mh00000002003obP0500000000kOl00>go
003]o`0000000`00/ol01@00000008co002<o`00S?l0000000<007Wo00@00000001lo`00O?l00000
0`00T_l01@0000000;Wo002io`00^Ol0000000<00>co00D00000003of000omP00?oH00000003003o
VP030000003oool0oooo00<0oooo0P0000010?ooo`00803oool00`000000oooo0?ooo`050?ooool0
0000P@0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool0100000000?mG
003oE`000003003oT`0500000000om<00?oC003od`0000000`00kol01@0000000;[o002jo`00^_l0
000000<009;o00@00000001lo`00O?l000000`00N_l01@00000008ko002>o`00S_l0000000<00;So
00D00000003do`00m?l00?Co00000003003o`00400000000ofd00?m]000000<00?lH00D0000003Oo
000go`00=ol00000000307Wo00050000002Xo`00Z?l00:So000000000`2no`0010000000^Ol00;Wo
0000000309Oo00050000001Ko`00Fol005_o000000000`09o`001@0000000?mH003oF000oeP00000
00<00?o200@00000003Do`00e?l000000`00M_l01@00000002[o000Zo`00:_l0000000<01P3o00D0
000001T0o`0I0?l06@3o0000000300T0o`04000000009ol002Oo000000<007Ko00D00000003Lo`00
g?l00=co00000003003o[@0500000000oc400?la003o<@0000000`15o`0010000000[?l00:co0000
00030?[o00050000003oeP00omH00?oF000000000`3oc0001@000000onL00?oW003oi`0000000080
f?l000D0000007[o001jo`00N_l00000000300Co000500000000ogd00?mm003oO@0000000`00oOl0
1@00000008;o0022o`00P_l00000008001_o00D0000002h0o`0^0?l0;P3o0000000305D0o`050000
001E0?l0E@3o05D0o`0000000`0^0?l01@00000001_o000Ko`006ol00000008008;o00D00000003m
o`00oOl00?go00000003003oO@0500000004o`001?l000Co000000000`1jo`001@000000f?l00=So
003Ho`0000000080onL000D000000?o<003oc000ol`0000000030?oF00050000003jo`00n_l00?[o
000000000`2/o`001@000000AOl004Go0015o`00000000800?la00D00000003o[@00ojd00?n]0000
0003003Lo`0500000000M_l007Ko001fo`0000000`009ol01@0000002@3o00T0o`090?l000000080
6@3o00D0000000H0o`060?l01P3o00000003000Zo`0500000000M_l007Ko001fo`0000000`00e?l0
1@0000000?o2003o`P00ol80000000800?mH00D0000000Wo0009o`002Ol00000000305_o00050000
002Go`00Uol009Oo000000000`2io`001@000000__l00;ko002no`0000000080Z?l000D0000007Wo
001io`00NOl00000000303Oo000500000000oaP00?lH003o600000000`00ofd01@0000000?o0003o
`000ol00000000800?Co00D00000002ho`00^?l00;So00000003002>o`0500000000N_l007[o001j
o`0000000`00O?l01000000009;o002Bo`000003002jo`0500000000kol00>oo003_o`0000000`00
om<01@0000000?nC003oT`00oi<0000000<00?mG00<000000?ooo`3oool0103oool100000040oooo
000P0?ooo`030000003oool0oooo00D0oooo00@00000003oE`00oeL000000`00oi<01@0000000?oC
003od`00om<0000000<00>oo00D00000002jo`00^_l00;[o00000003002Bo`0400000000O?l007co
000000<007[o00D00000002>o`00S_l008ko00000003002ho`0500000000m?l00?Co003do`000000
0`00ol00100000000?m]003oK@000003003o60050000000go`00=ol003Oo000000000`1io`001@00
0000Z?l00:So002Xo`00000000<0__l000@000000;Wo002io`0000000`2Go`001@000000Fol005_o
001Ko`00000000<02Ol000D00000003oF000oeP00?mH00000003003o`P0400000000e?l00=Co0000
00<007Ko00D00000000Zo`00:_l002[o0000000300H0o`050000000I0?l06@3o01T0o`0000000`09
0?l01000000002Oo000Wo`000003001fo`0500000000g?l00=co003Lo`0000000`00ojd01@000000
0?la003o<@00oc40000000<0AOl000@000000:co002/o`0000000`3jo`001@000000omH00?oF003o
eP00000000<0ol`000D000000?oW003oi`00onL0000000020=So00050000001jo`00N_l007[o0000
00000`04o`001@0000000?mm003oO@00ogd0000000<00?go00D000000022o`00P_l008;o00000002
000Ko`050000000^0?l0;P3o02h0o`0000000`1E0?l01@000000E@3o05D0o`1E0?l0000000<0;P3o
00D00000000Ko`006ol001_o000000020022o`0500000000oOl00?go003mo`0000000`00ogd01@00
00001?l000Co0004o`00000000<0N_l000D000000=So003Ho`00f?l0000000020?oW00050000003o
c000ol`00?o<000000000`3oeP001@000000n_l00?[o003jo`00000000<0[?l000D0000004Go0015
o`00AOl000000002003o<@0500000000ojd00?n]003o[@0000000`00g?l01@00000007Ko001fo`00
M_l0000000<002Oo00D0000000T0o`090?l02@3o0000000201T0o`05000000060?l01P3o00H0o`00
00000`00:_l01@00000007Ko001fo`00M_l0000000<00=Co00D00000003o`P00ol800?o200000002
003oF00500000009o`002Ol000Wo000000000`1Ko`001@000000Uol009Oo002Go`00000000<0^Ol0
00D000000;ko002no`00__l0000000020:So00050000001io`00NOl007Wo000000000`0go`001@00
00000?lH003o6000oaP0000000<00?m]00D00000003o`000ol000?o000000002003do`0500000000
^?l00;So002ho`0000000`00S_l01@00000007[o001jo`00N_l0000000<007co00@00000002Bo`00
T_l000000`00^_l01@0000000>oo003_o`00kol0000000<00?oC00D00000003oT`00oi<00?nC0000
0003003oE`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`05
0?ooo`0400000000oeL00?mG000000<00?nC00D00000003od`00om<00?oC00000003003_o`050000
0000^_l00;[o002jo`0000000`00T_l01000000007co001lo`000003001jo`0500000000S_l008ko
002>o`0000000`00^?l01@0000000?Co003do`00m?l0000000<00?o000@00000003oK@00ofd00000
0`00oaP01@000000=ol003Oo000go`00000000<0NOl000D000000:So002Xo`00Z?l0000000030;ko
00040000002io`00^Ol0000000<0Uol000D0000005_o001Ko`00Fol00000000300Wo000500000000
oeP00?mH003oF00000000`00ol80100000000=Co003Do`000003001fo`0500000000:_l002[o000Z
o`0000000`060?l01@0000006@3o01T0o`0I0?l0000000<02@3o00@00000000Wo`009ol000000`00
M_l01@0000000=co003Lo`00g?l0000000<00?n]00D00000003o<@00oc400?la0000000304Go0004
0000002/o`00[?l0000000<0n_l000D000000?oF003oeP00omH0000000030?o<00050000003oi`00
onL00?oW000000000P3Ho`001@000000N_l007[o001jo`00000000<01?l000D00000003oO@00ogd0
0?mm00000003003mo`0500000000P_l008;o0022o`0000000P006ol01@000000;P3o02h0o`0^0?l0
000000<0E@3o00D0000005D0o`1E0?l0E@3o0000000302h0o`05000000006ol001_o000Ko`000000
0P00P_l01@0000000?go003mo`00oOl0000000<00?mm00D0000000Co0004o`001?l00000000307[o
00050000003Ho`00f?l00=So000000000P3oi`001@000000ol`00?o<003oc000000000<0omH000D0
00000?[o003jo`00n_l0000000030:co000500000015o`00AOl004Go000000000P00oc401@000000
0?n]003o[@00ojd0000000<00=co00D00000001fo`00M_l007Ko00000003000Wo`05000000090?l0
2@3o00T0o`0000000P0I0?l01@0000001P3o00H0o`060?l0000000<002[o00D00000001fo`00M_l0
07Ko00000003003Do`0500000000ol800?o2003o`P0000000P00oeP01@0000002Ol000Wo0009o`00
000000<0Fol000D0000009Oo002Go`00Uol0000000030;Wo00050000002no`00__l00;ko00000000
0P2Xo`001@000000NOl007Wo001io`00000000<0=ol000D00000003o6000oaP00?lH00000003003o
K@0500000000ol000?o0003o`00000000P00m?l01@0000000;So002ho`00^?l0000000<008ko00D0
0000001jo`00N_l007[o00000003001lo`0400000000T_l009;o000000<00;[o00D00000003_o`00
kol00>oo00000003003od`0500000000oi<00?nC003oT`0000000`00oeL00`000000oooo0?ooo`04
0?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo0000084000001P3oool10000
0040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003o7000oa`000000`00oe401@00
00000?n?003oS`00ohl0000000<00?o@00D00000003`o`00l?l00?3o00000003002jo`0400000000
T_l009;o000000<007co00D00000001lo`00O?l007co00000003002Bo`0500000000_ol00;oo002o
o`0000000`00o_l0100000000?nc003o/`000003003oGP0500000000o`P00?l8003o200000000`15
o`001@000000Pol008?o0023o`00000000<0[?l000@000000;_o002ko`0000000`2/o`001@000000
POl0087o0021o`00000000<0??l000D00000003o7000oa`00?lL00000003003oP`0400000000onh0
0?o^000000<00:co00D00000001Eo`00EOl005Go00000003000Fo`050000000;0?l02`3o00/0o`00
00000`0;0?l01000000001Ko000Fo`000003001Io`0500000000]_l00;Ko002fo`0000000`00omP0
1@0000000?mM003oG@00oed0000000<06ol000@0000008Oo0027o`0000000`3Mo`001@000000onX0
0?oZ003ojP00000000<0omL000D000000?oZ003ojP00onX0000000020=go000500000024o`00Q?l0
08Co000000000`0Bo`001@0000000?m]003oK@00ofd0000000<00?oa00D00000002Bo`00T_l009;o
00000002000Zo`050000000P0?l0803o0200o`0000000`170?l01@000000A`3o04L0o`170?l00000
00<0803o00D00000000Zo`00:_l002[o00000002002Bo`0500000000oo400?oa003ol@0000000`00
ofd01@0000004_l001;o000Bo`00000000<0Q?l000D000000=go003Mo`00gOl0000000020?oZ0005
0000003oe`00omL00?oG000000000`3ojP001@000000gOl00=go003Mo`00000000<0Qol000D00000
01_o000Ko`006ol000000002003oG@0500000000omP00?oH003of00000000`00]_l01@00000005Wo
001Io`00FOl0000000<001Ko00D0000000/0o`0;0?l02`3o0000000200/0o`05000000005_l001Ko
000Fo`0000000`00EOl01@0000000:co002/o`00[?l0000000<00?o^00D00000003oP`00oh<00?n3
00000002003o70050000000lo`00??l003co000000000`21o`001@000000[?l00:co002/o`000000
00<0^ol000D000000:co002/o`00[?l00000000208?o000500000015o`00AOl004Go000000000`00
o`P01@0000000?mN003oGP00oeh0000000<00?nc00D00000003no`00o_l00?ko00000002002oo`05
00000000T_l009;o002Bo`0000000`00O?l01@00000007co001lo`00O?l0000000<009;o00@00000
002jo`00^_l000000`00l?l01@0000000?o@003od000om00000000<00?n?00D00000003oD@00oe40
0?mA00000003003o70030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo
0?ooo`050?ooo`0400000000oa`00?lL000000<00?mA00D00000003oS`00ohl00?n?00000003003o
d00500000000l?l00?3o003`o`0000000`00^_l01000000009;o002Bo`000003001lo`0500000000
O?l007co001lo`0000000`00T_l01@0000000;oo002oo`00_ol0000000<00?ko00@00000003o/`00
ok<000000`00oeh01@0000000?l8003o2000o`P0000000<0AOl000D0000008?o0023o`00Pol00000
00030:co00040000002ko`00^ol0000000<0[?l000D00000087o0021o`00POl00000000303co0005
00000000oa`00?lL003o700000000`00oh<0100000000?o^003okP000003002/o`0500000000EOl0
05Go001Eo`0000000`005_l01@0000002`3o00/0o`0;0?l0000000<02`3o00@00000000Fo`005_l0
00000`00FOl01@0000000;Ko002fo`00]_l0000000<00?oH00D00000003oG@00oed00?mM00000003
01_o000400000027o`00Qol0000000<0gOl000D000000?oZ003ojP00onX0000000030?oG00050000
003ojP00onX00?oZ000000000P3Mo`001@000000Q?l008Co0024o`00000000<04_l000D00000003o
K@00ofd00?m]00000003003ol@0500000000T_l009;o002Bo`0000000P00:_l01@000000803o0200
o`0P0?l0000000<0A`3o00D0000004L0o`170?l0A`3o000000030200o`0500000000:_l002[o000Z
o`0000000P00T_l01@0000000?oa003ol@00oo40000000<00?m]00D0000001;o000Bo`004_l00000
000308Co00050000003Mo`00gOl00=go000000000P3ojP001@000000omL00?oG003oe`00000000<0
onX000D000000=go003Mo`00gOl00000000308Oo00050000000Ko`006ol001_o000000000P00oed0
1@0000000?oH003of000omP0000000<00;Ko00D00000001Io`00FOl005Wo00000003000Fo`050000
000;0?l02`3o00/0o`0000000P0;0?l01@00000001Ko000Fo`005_l0000000<005Go00D00000002/
o`00[?l00:co00000003003okP0500000000oh<00?n3003oP`0000000P00oa`01@000000??l003co
000lo`00000000<0POl000D000000:co002/o`00[?l0000000030;_o00050000002/o`00[?l00:co
000000000P23o`001@000000AOl004Go0015o`00000000<00?l800D00000003oGP00oeh00?mN0000
0003003o/`0500000000o_l00?ko003no`0000000P00_ol01@00000009;o002Bo`00T_l0000000<0
07co00D00000001lo`00O?l007co00000003002Bo`0400000000^_l00;[o000000<00?3o00D00000
003od000om000?o@00000003003oS`0500000000oe400?mA003oD@0000000`00oa`00`000000oooo
0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooooo0000084000001P3o
ool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@0000000ko000>o`0000000`00
oaT01@0000000?m?003oC`00odl0000000<00?n>00D00000003od000om000?o@00000003003_o`04
00000000^Ol00;Wo000000<0097o00D00000001lo`00O?l007co00000003001oo`0500000000V?l0
09So002Ho`0000000`00bOl0100000000?oc003ol`000003003oXP0500000000od/00?m;003oB`00
00000`09o`001@000000E?l005Co001Do`00000000<0SOl000@000000:oo002_o`0000000`2do`00
1@000000Vol009_o002Ko`00000000<0I_l000D0000001So000Ho`006?l000000003003oA`040000
0000ok400?na000000<00>Co00D000000025o`00QOl008Go00000003000io`05000000002?l000So
0008o`0000000`060?l01000000000co000<o`0000030012o`0500000000UOl009Go002Eo`000000
0`00ool01@0000000?n7003oQ`00ohL0000000<00?l=00@0000006;o001Ro`0000000`2oo`001@00
0000ool00?oo003oo`00000000<0on@000D000000?o_003ok`00onl0000000020=oo00050000002<
o`00S?l008co000000000`0No`001@0000000?mO003oG`00oel0000000<00?oR00D00000002Qo`00
XOl00:7o00000002000io`050000000B0?l04P3o0180o`0000000`0j0?l01@000000>P3o03X0o`0j
0?l0000000<04P3o00D00000000io`00>Ol003Wo00000002002Qo`0500000000on800?oR003ohP00
00000`00oel01@0000007_l001ko000No`00000000<0S?l000D000000=oo003Oo`00gol000000002
0?o_00050000003oi000on@00?oT000000000`3oo`001@000000_ol00;oo002oo`00000000<0H_l0
00D00000003o3@00o`d00?l=00000002003oQ`0500000000ool00?oo003oo`0000000`00UOl01@00
000004;o0012o`00@_l0000000<000co00D0000000H0o`060?l01P3o000000020008o`0500000000
>Ol003Wo000io`0000000`00QOl01@0000000>Co003To`00i?l0000000<00?na00D00000003oA`00
odL00?m70000000201So00050000001Vo`00I_l006Ko000000000`2Ko`001@000000]?l00;Co002d
o`00000000<0[ol000D0000008go002=o`00SOl00000000205Co000500000009o`002Ol000Wo0000
00000`00od/01@0000000?nR003oXP00oj80000000<00?oc00D000000039o`00bOl00<Wo00000002
002Ho`0500000000Ool007oo001oo`0000000`00O?l01@000000097o002Ao`00TOl0000000<00;Wo
00@00000003_o`00kol000000`00om001@0000000?n>003oSP00ohh0000000<00?m?00D00000003o
6@00oaT00?lI0000000300ko00030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`00
0000oooo0?ooo`050?ooo`040000000>o`003_l0000000<00?lI00D00000003oC`00odl00?m?0000
0003003oSP0500000000om000?o@003od00000000`00kol0100000000;Wo002io`000003002Ao`05
00000000O?l007co001lo`0000000`00Ool01@00000009So002Ho`00V?l0000000<00<Wo00@00000
003ol`00oo<000000`00oj801@0000000?m;003oB`00od/0000000<02Ol000D0000005Co001Do`00
E?l00000000308go00040000002_o`00[ol0000000<0]?l000D0000009_o002Ko`00Vol000000003
06Ko00050000000Ho`006?l001So000000000`00odL0100000000?na003o/@000003003To`050000
0000QOl008Go0025o`0000000`00>Ol01@00000000So0008o`002?l0000000<01P3o00@00000000<
o`003?l000000`00@_l01@00000009Go002Eo`00UOl0000000<00?oo00D00000003oQ`00ohL00?n7
00000003003o3@040000001Ro`00H_l0000000<0_ol000D000000?oo003oo`00ool0000000030?oT
00050000003ok`00onl00?o_000000000P3Oo`001@000000S?l008co002<o`00000000<07_l000D0
0000003oG`00oel00?mO00000003003ohP0500000000XOl00:7o002Qo`0000000P00>Ol01@000000
4P3o0180o`0B0?l0000000<0>P3o00D0000003X0o`0j0?l0>P3o000000030180o`0500000000>Ol0
03Wo000io`0000000P00XOl01@0000000?oR003ohP00on80000000<00?mO00D0000001ko000No`00
7_l00000000308co00050000003Oo`00gol00=oo000000000P3ok`001@000000on@00?oT003oi000
000000<0ool000D000000;oo002oo`00_ol00000000306;o000500000000o`d00?l=003o3@000000
0P00ohL01@0000000?oo003oo`00ool0000000<009Go00D000000012o`00@_l004;o00000003000<
o`05000000060?l01P3o00H0o`0000000P002?l01@00000003Wo000io`00>Ol0000000<008Go00D0
0000003To`00i?l00>Co00000003003o/@0500000000odL00?m7003oA`0000000P0Ho`001@000000
I_l006Ko001Vo`00000000<0Vol000D000000;Co002do`00]?l0000000030:oo00050000002=o`00
SOl008go000000000P1Do`001@0000002Ol000Wo0009o`00000000<00?m;00D00000003oXP00oj80
0?nR00000003003ol`0500000000bOl00<Wo0039o`0000000P00V?l01@00000007oo001oo`00Ool0
000000<007co00D00000002Ao`00TOl0097o00000003002io`0400000000kol00>oo000000<00?o@
00D00000003oSP00ohh00?n>00000003003oC`0500000000oaT00?lI003o6@0000000`0>o`000`00
0000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool010000000
3_l000ko00000003003o6@0500000000odl00?m?003oC`0000000`00ohh01@0000000?o@003od000
om00000000<00>oo00@00000002io`00^Ol000000`00TOl01@00000007co001lo`00O?l0000000<0
07oo00D00000002Ho`00V?l009So000000030039o`0400000000oo<00?oc000000<00?nR00D00000
003oB`00od/00?m;0000000300Wo00050000001Do`00E?l005Co000000000`2=o`0010000000[ol0
0:oo000000030;Co00050000002Ko`00Vol009_o000000000`1Vo`001@0000006?l001So000Ho`00
000000<00?m700@00000003o/@00ok4000000`00i?l01@00000008Go0025o`00QOl0000000<003Wo
00D000000008o`002?l000So0000000300H0o`04000000003?l000co000000<004;o00D00000002E
o`00UOl009Go00000003003oo`0500000000ohL00?n7003oQ`0000000`00o`d010000000H_l006;o
000000030;oo00050000003oo`00ool00?oo000000000`3oi0001@000000onl00?o_003ok`000000
0080gol000D0000008co002<o`00S?l00000000301ko000500000000oel00?mO003oG`0000000`00
on801@0000000:7o002Qo`00XOl00000008003Wo00D000000180o`0B0?l04P3o0000000303X0o`05
0000000j0?l0>P3o03X0o`0000000`0B0?l01@00000003Wo000io`00>Ol0000000800:7o00D00000
003ohP00on800?oR00000003003oG`050000000No`007_l001ko000000000`2<o`001@000000gol0
0=oo003Oo`0000000080onl000D000000?oT003oi000on@0000000030?oo00050000002oo`00_ol0
0;oo000000000`1Ro`001@0000000?l=003o3@00o`d0000000800?n700D00000003oo`00ool00?oo
00000003002Eo`0500000000@_l004;o0012o`0000000`003?l01@0000001P3o00H0o`060?l00000
008000So00D00000000io`00>Ol003Wo000000030025o`0500000000i?l00>Co003To`0000000`00
ok401@0000000?m7003oA`00odL0000000806?l000D0000006Ko001Vo`00I_l00000000309_o0005
0000002do`00]?l00;Co000000000`2_o`001@000000SOl008go002=o`0000000080E?l000D00000
00Wo0009o`002Ol000000003003oB`0500000000oj800?nR003oXP0000000`00oo<01@0000000<Wo
0039o`00bOl00000008009So00D00000001oo`00Ool007oo00000003001lo`0500000000TOl0097o
002Ao`0000000`00^Ol0100000000>oo003_o`000003003od00500000000ohh00?n>003oSP000000
0`00odl01@0000000?lI003o6@00oaT0000000<03_l000<000000?ooo`3oool0103oool100000040
oooo000P0?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooo`040000000Xo`00:?l0000000<04?l000D00000003o6000oaP0
0?lH00000003003oC`0500000000ohl00?n?003oS`0000000`00om<0100000000>co003/o`000003
002eo`0500000000Sol008oo002?o`0000000`00OOl01@00000008?o0023o`00Pol0000000<00:7o
00@00000003Fo`00e_l000000`00on801@0000000?n>003oSP00ohh0000000<00?lf00D0000001co
000Lo`007?l00000000306?o00040000002Fo`00U_l0000000<0[_l000D000000:Wo002Yo`00ZOl0
0000000308Go000500000015o`00AOl004Go000000000`00oa40100000000?mg003oM`000003003o
hP0500000000^?l00;So002ho`0000000`00HOl01@00000002?o000So`008ol0000000<000Go00@0
00000009o`002Ol000000`00<Ol01@00000007[o001jo`00N_l0000000<00=co00D00000003o[`00
ojl00?n_00000003003o=@040000000no`00?_l0000000<0X?l000D000000>Oo003Wo`00iol00000
00030?oc00050000003omP00ooH00?of000000000P3Po`001@000000Tol009?o002Co`00000000<0
:_l000D00000003oD@00oe400?mA00000003003od`0500000000/?l00;3o002`o`0000000P00Aol0
1@000000103o00@0o`040?l0000000<0;@3o00D0000002d0o`0]0?l0;@3o0000000300@0o`050000
0000Aol004Oo0017o`0000000P00/?l01@0000000?oC003od`00om<0000000<00?mA00D0000002[o
000Zo`00:_l00000000309?o00050000003Po`00h?l00>3o000000000P3omP001@000000oo<00?oc
003ol`00000000<0iol000D000000:3o002Po`00X?l00000000303ko000500000000ocD00?le003o
=@0000000P00ojl01@0000000=co003Lo`00g?l0000000<007[o00D00000000ao`00<Ol0037o0000
00030009o`05000000001Ol000Go0005o`0000000P008ol01@000000067o001Qo`00HOl0000000<0
0;So00D00000003ohP00on800?oR00000003003oM`0500000000oa400?lA003o4@0000000P15o`00
1@000000QOl008Go0025o`00000000<0ZOl000D000000:ko002^o`00[_l00000000309Ko00050000
001So`00Hol006?o000000000P0Lo`001@0000000?lf003o=P00ocH0000000<00?n>00D00000003o
hP00on800?oR00000003003Fo`0500000000XOl00:7o002Qo`0000000P00Pol01@00000007go001m
o`00OOl0000000<008oo00D00000002eo`00]Ol00;Go00000003003/o`0400000000om<00?oC0000
00<00?n?00D00000003oC`00odl00?m?00000003003o60050000000@o`004?l0013o000000000`0X
o`000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool0
10000000:?l002So00000003013o000500000000oaP00?lH003o600000000`00odl01@0000000?n?
003oS`00ohl0000000<00?oC00@00000003/o`00k?l000000`00]Ol01@00000008oo002?o`00Sol0
000000<007go00D000000023o`00Pol008?o00000003002Qo`0400000000e_l00=Ko000000<00?oR
00D00000003oSP00ohh00?n>00000003003o=P050000000Lo`007?l001co000000000`1So`001000
0000U_l009Ko000000030:ko00050000002Yo`00ZOl00:Wo000000000`25o`001@000000AOl004Go
0015o`00000000<00?lA00@00000003oM`00ogL000000`00on801@0000000;So002ho`00^?l00000
00<0067o00D00000000So`008ol002?o000000030005o`04000000002Ol000Wo000000<0037o00D0
0000001jo`00N_l007[o00000003003Lo`0500000000ojl00?n_003o[`0000000`00ocD010000000
?_l003ko000000030:3o00050000003Wo`00iol00>Oo000000000`3ol`001@000000ooH00?of003o
mP0000000080h?l000D0000009?o002Co`00Tol00000000302[o000500000000oe400?mA003oD@00
00000`00om<01@0000000;3o002`o`00/?l00000008004Oo00D0000000@0o`040?l0103o00000003
02d0o`050000000]0?l0;@3o02d0o`0000000`040?l01@00000004Oo0017o`00Aol0000000800;3o
00D00000003od`00om<00?oC00000003003oD@050000000Zo`00:_l002[o000000000`2Co`001@00
0000h?l00>3o003Po`0000000080ooH000D000000?oc003ol`00oo<0000000030>Oo00050000002P
o`00X?l00:3o000000000`0no`001@0000000?le003o=@00ocD0000000800?n_00D00000003Lo`00
g?l00=co00000003001jo`0500000000<Ol0037o000ao`0000000`002Ol01@00000000Go0005o`00
1Ol00000008002?o00D00000001Qo`00HOl0067o00000003002ho`0500000000on800?oR003ohP00
00000`00ogL01@0000000?lA003o4@00oa4000000080AOl000D0000008Go0025o`00QOl000000003
0:Wo00050000002^o`00[_l00:ko000000000`2Fo`001@000000Hol006?o001So`00000000807?l0
00D00000003o=P00ocH00?lf00000003003oSP0500000000on800?oR003ohP0000000`00e_l01@00
00000:7o002Qo`00XOl00000008008?o00D00000001mo`00OOl007go00000003002?o`0500000000
]Ol00;Go002eo`0000000`00k?l0100000000?oC003od`000003003oS`0500000000odl00?m?003o
C`0000000`00oaP01@0000004?l0013o000@o`00000000<0:?l000<000000?ooo`3oool0103oool1
00000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@0000002So000Xo`0000000`0@o`00
1@0000000?lH003o6000oaP0000000<00?m?00D00000003oS`00ohl00?n?00000003003od`040000
0000k?l00>co000000<00;Go00D00000002?o`00Sol008oo00000003001mo`0500000000Pol008?o
0023o`0000000`00XOl0100000000=Ko003Fo`000003003ohP0500000000ohh00?n>003oSP000000
0`00ocH01@0000007?l001co000Lo`00000000<0Hol000@0000009Ko002Fo`0000000`2^o`001@00
0000ZOl00:Wo002Yo`00000000<0QOl000D0000004Go0015o`00AOl000000003003o4@0400000000
ogL00?mg000000<00?oR00D00000002ho`00^?l00;So00000003001Qo`05000000008ol002?o000S
o`0000000`001Ol01000000000Wo0009o`000003000ao`0500000000N_l007[o001jo`0000000`00
g?l01@0000000?n_003o[`00ojl0000000<00?le00@0000003ko000no`0000000`2Po`001@000000
iol00>Oo003Wo`00000000<0oo<000D000000?of003omP00ooH0000000020>3o00050000002Co`00
Tol009?o000000000`0Zo`001@0000000?mA003oD@00oe40000000<00?oC00D00000002`o`00/?l0
0;3o000000020017o`05000000040?l0103o00@0o`0000000`0]0?l01@000000;@3o02d0o`0]0?l0
000000<0103o00D000000017o`00Aol004Oo00000002002`o`0500000000om<00?oC003od`000000
0`00oe401@000000:_l002[o000Zo`00000000<0Tol000D000000>3o003Po`00h?l0000000020?of
00050000003ol`00oo<00?oc000000000`3Wo`001@000000X?l00:3o002Po`00000000<0?_l000D0
0000003o=@00ocD00?le00000002003o[`0500000000g?l00=co003Lo`0000000`00N_l01@000000
037o000ao`00<Ol0000000<000Wo00D000000005o`001Ol000Go00000002000So`0500000000HOl0
067o001Qo`0000000`00^?l01@0000000?oR003ohP00on80000000<00?mg00D00000003o4@00oa40
0?lA0000000204Go000500000025o`00QOl008Go000000000`2Yo`001@000000[_l00:ko002^o`00
000000<0U_l000D0000006?o001So`00Hol00000000201co000500000000ocH00?lf003o=P000000
0`00ohh01@0000000?oR003ohP00on80000000<00=Ko00D00000002Qo`00XOl00:7o000000020023
o`0500000000OOl007go001mo`0000000`00Sol01@0000000;Go002eo`00]Ol0000000<00>co00@0
0000003od`00om<000000`00ohl01@0000000?m?003oC`00odl0000000<00?lH00D00000013o000@
o`004?l00000000302So00030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000
oooo0?ooo`050?ooool00000P@0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool0
1@3oool010000000;_l002ko0000000302Wo00050000000@o`004?l0013o000000000`00oaT01@00
00000?mA003oD@00oe40000000<00?nC00@00000003of000omP000000`00i_l01@0000000;7o002a
o`00/Ol0000000<008go00D00000001oo`00Ool007oo000000030029o`0400000000[Ol00:go0000
00<00>Oo00D00000003oc000ol`00?o<00000003003oMP0500000000oah00?lN003o7P0000000`0a
o`0010000000Lol007?o0000000309co00050000002Zo`00Z_l00:[o000000000`2Ho`001@000000
J?l006So001Xo`00000000<07Ol000@00000003o@@00od4000000`00ojX01@0000000>co003/o`00
k?l0000000<008go00D000000014o`00A?l004Co00000003000Go`04000000003?l000co000000<0
02Ko00D00000001So`00Hol006?o00000003002mo`0500000000om@00?oD003oe00000000`00oe`0
100000006Ol001Wo00000003087o00050000003?o`00col00<oo000000000`3ko`001@000000ooh0
0?on003ooP0000000080gol000D0000009So002Ho`00V?l00000000303?o000500000000od@00?m4
003oA00000000`00olD01@0000000;oo002oo`00_ol00000008005Go00D000000008o`002?l000So
000000030200o`050000000P0?l0803o0200o`0000000`002?l01@00000005Go001Eo`00EOl00000
00800;oo00D00000003oa@00olD00?o500000003003oA0050000000co`00<ol003?o000000000`2H
o`001@000000gol00=oo003Oo`0000000080ooh000D000000?_o003ko`00nol0000000030<oo0005
00000021o`00POl0087o000000000`0Io`001@0000000?mL003oG000oe`0000000800?oD00D00000
002mo`00_Ol00;go00000003001So`05000000009_l002Ko000Vo`0000000`003?l01@00000001Oo
000Go`005ol00000008004Co00D00000002=o`00SOl008go00000003003/o`0500000000ojX00?nZ
003oZP0000000`00od401@0000007Ol001go000Mo`0000000080J?l000D0000009So002Ho`00V?l0
000000030:[o00050000002Lo`00W?l009co000000000`1co`001@000000<Ol0037o000ao`000000
00800?lN00D00000003oMP00ogH00?mf00000003003oc00500000000iol00>Oo003Wo`0000000`00
[Ol01@00000008Wo0029o`00ROl00000008007oo00D00000002=o`00SOl008go00000003002ao`05
00000000i_l00>Ko003Vo`0000000`00omP0100000000?nC003oT`000003003oD@0500000000oaT0
0?lI003o6@0000000`0@o`001@000000:Ol002Wo000Yo`00000000<0;_l000<000000?ooo`3oool0
103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@0000002ko000^o`000000
0`0Yo`001@0000004?l0013o000@o`00000000<00?lI00D00000003oD@00oe400?mA00000003003o
T`0400000000omP00?oH000000<00>Ko00D00000002ao`00/Ol00;7o00000003002=o`0500000000
Ool007oo001oo`0000000`00ROl0100000000:go002]o`000003003Wo`0500000000ol`00?o<003o
c00000000`00ogH01@0000000?lN003o7P00oah0000000<0<Ol000@0000007?o001co`0000000`2L
o`001@000000Z_l00:[o002Zo`00000000<0V?l000D0000006So001Xo`00J?l00000000301go0004
00000000od400?m1000000<00?nZ00D00000003/o`00k?l00>co00000003002=o`0500000000A?l0
04Co0014o`0000000`005ol01000000000co000<o`000003000Vo`0500000000Hol006?o001So`00
00000`00_Ol01@0000000?oD003oe000om@0000000<00?mL00@0000001Wo000Io`0000000`21o`00
1@000000col00<oo003?o`00000000<0nol000D000000?on003ooP00ooh0000000020=oo00050000
002Ho`00V?l009So000000000`0co`001@0000000?m4003oA000od@0000000<00?o500D00000002o
o`00_ol00;oo00000002001Eo`05000000002?l000So0008o`0000000`0P0?l01@000000803o0200
o`0P0?l0000000<000So00D00000001Eo`00EOl005Go00000002002oo`0500000000olD00?o5003o
a@0000000`00od@01@000000<ol003?o000co`00000000<0V?l000D000000=oo003Oo`00gol00000
00020?on00050000003ko`00nol00?_o000000000`3?o`001@000000POl0087o0021o`00000000<0
6Ol000D00000003oG000oe`00?mL00000002003oe00500000000_Ol00;go002mo`0000000`00Hol0
1@00000002Ko000Vo`009_l0000000<000co00D00000000Go`005ol001Oo000000020014o`050000
0000SOl008go002=o`0000000`00k?l01@0000000?nZ003oZP00ojX0000000<00?m100D0000001go
000Mo`007Ol00000000206So00050000002Ho`00V?l009So000000000`2Zo`001@000000W?l009co
002Lo`00000000<0Lol000D00000037o000ao`00<Ol000000002003o7P0500000000ogH00?mf003o
MP0000000`00ol`01@0000000>Oo003Wo`00iol0000000<00:go00D000000029o`00ROl008Wo0000
0002001oo`0500000000SOl008go002=o`0000000`00/Ol01@0000000>Ko003Vo`00i_l0000000<0
0?oH00@00000003oT`00oi<000000`00oe401@0000000?lI003o6@00oaT0000000<04?l000D00000
02Wo000Yo`00:Ol00000000302ko00030000003oool0oooo00@0oooo0@0000010?ooo`004`3oool0
103ooml0W4P00018W03Oool90?ooo`030000003oool0oooo00D0oooo00@0000002ko000^o`000000
0`0Yo`001@0000004?l0013o000@o`00000000<00?lI00D00000003oD@00oe400?mA00000003003o
T`0400000000omP00?oH000000<00>Ko00D00000002ao`00/Ol00;7o00000003002=o`0500000000
Ool007oo001oo`0000000`00ROl0100000000:go002]o`000003003Wo`0500000000ol`00?o<003o
c00000000`00ogH01@0000000?lN003o7P00oah0000000<0<Ol000@0000007?o001co`0000000`2L
o`001@000000Z_l00:[o002Zo`00000000<0V?l000D0000006So001Xo`00J?l00000000301go0004
00000000od400?m1000000<00?nZ00D00000003/o`00k?l00>co00000003002=o`0500000000A?l0
04Co0014o`0000000`005ol01000000000co000<o`000003000Vo`0500000000Hol006?o001So`00
00000`00_Ol01@0000000?oD003oe000om@0000000<00?mL00@0000001Wo000Io`0000000`21o`00
1@000000col00<oo003?o`00000000<0nol000D000000?on003ooP00ooh0000000020=oo00050000
002Ho`00V?l009So000000000`0co`001@0000000?m4003oA000od@0000000<00?o500D00000002o
o`00_ol00;oo00000002001Eo`05000000002?l000So0008o`0000000`0P0?l01@000000803o0200
o`0P0?l0000000<000So00D00000001Eo`00EOl005Go00000002002oo`0500000000olD00?o5003o
a@0000000`00od@01@000000<ol003?o000co`00000000<0V?l000D000000=oo003Oo`00gol00000
00020?on00050000003ko`00nol00?_o000000000`3?o`001@000000POl0087o0021o`00000000<0
6Ol000D00000003oG000oe`00?mL00000002003oe00500000000_Ol00;go002mo`0000000`00Hol0
1@00000002Ko000Vo`009_l0000000<000co00D00000000Go`005ol001Oo000000020014o`050000
0000SOl008go002=o`0000000`00k?l01@0000000?nZ003oZP00ojX0000000<00?m100D0000001go
000Mo`007Ol00000000206So00050000002Ho`00V?l009So000000000`2Zo`001@000000W?l009co
002Lo`00000000<0Lol000D00000037o000ao`00<Ol000000002003o7P0500000000ogH00?mf003o
MP0000000`00ol`01@0000000>Oo003Wo`00iol0000000<00:go00D000000029o`00ROl008Wo0000
0002001oo`0500000000SOl008go002=o`0000000`00/Ol01@0000000>Ko003Vo`00i_l0000000<0
0?oH00@00000003oT`00oi<000000`00oe401@0000000?lI003o6@00oaT0000000<04?l000D00000
02Wo000Yo`00:Ol00000000302ko00030000003oool0oooo00@0oooo0@0000010?ooo`004`3oool0
103ooml0W4P00018W03Oool90?ooo`030000003oool0oooo00D0ooooo`000021000000H0oooo0@00
00010?ooo`004`3oool0103ooml0W4P00018W03Oool90?ooo`030000003oool0oooo00D0oooo00@0
0000027o000Qo`0000000`0^o`001@000000:?l002So000Xo`00000000<03_l000D00000003o7000
oa`00?lL00000003003oE`0400000000oiX00?nJ000000<00?oP00D00000003Oo`00gol00=oo0000
0003002/o`0500000000Rol008_o002;o`0000000`00P_l01000000009;o002Bo`000003002lo`05
00000000o?l00?co003lo`0000000`00ok@01@0000000?mL003oG000oe`0000000<00?l500@00000
04Oo0017o`0000000`20o`001@000000Wol009oo002Oo`00000000<0X?l000D00000087o0021o`00
POl00000000304Co000400000000oa000?l@000000<00?md00D00000003ogP00omh00?oN00000003
002lo`0500000000J?l006So001Xo`0000000`00;_l01000000001Ko000Fo`000003000Ro`050000
0000D_l005;o001Bo`0000000`00XOl01@0000000?of003omP00ooH0000000<00?n000@00000003o
2@00o`T000000`1Qo`001@000000]Ol00;Go002eo`00000000<0jOl000D000000?Oo003go`00mol0
000000020=co00050000002Lo`00W?l009co000000000`0lo`001@0000000?li003o>@00ocT00000
00<00?nh00D00000003=o`00cOl00<go00000002001So`05000000005Ol001Go000Eo`0000000`0C
0?l01@0000004`3o01<0o`0C0?l0000000<001Go00D00000001So`00Hol006?o00000002003=o`05
00000000okP00?nh003o^00000000`00ocT01@000000??l003co000lo`00000000<0W?l000D00000
0=co003Lo`00g?l0000000020?Oo00050000003Yo`00jOl00>Wo000000000`2eo`001@000000HOl0
067o001Qo`00000000<00?l900D00000003oP000oh000?n000000002003omP0500000000XOl00:7o
002Qo`0000000`00D_l01@00000002;o000Ro`008_l0000000<001Ko00D00000000^o`00;_l002ko
00000002001Xo`0500000000_?l00;co002lo`0000000`00omh01@0000000?md003oM000og@00000
00<00?l@00D0000004Co0014o`00A?l000000002087o00050000002Po`00X?l00:3o000000000`2O
o`001@000000P?l0083o0020o`00000000<0Aol000D00000003o1@00o`D00?l500000002003oG005
00000000ok@00?nd003o]00000000`00o?l01@0000000;co002lo`00_?l0000000<009;o00D00000
0022o`00P_l008;o00000002002;o`0500000000[?l00:co002/o`0000000`00gol01@0000000?oP
003oh000on00000000<00?nJ00@00000003oE`00oeL000000`00oa`01@0000003_l000ko000>o`00
000000<0:?l000D0000002ko000^o`00;_l000000003027o00030000003oool0oooo00@0oooo0@00
00010?ooo`003`3oool00`2oM0000000000000050000000307Boo`3oool0oooo00H0oooo00<00000
0?ooo`3oool01@3oool0100000008Ol0027o0000000302ko00050000000Xo`00:?l002So00000000
0`0>o`001@0000000?lL003o7000oa`0000000<00?mG00@00000003oVP00oiX000000`00on001@00
00000=oo003Oo`00gol0000000<00:co00D00000002;o`00Rol008_o000000030022o`0400000000
T_l009;o000000<00;co00D00000003lo`00o?l00?co00000003003o]00500000000oe`00?mL003o
G00000000`00o`D010000000Aol004Oo00000003083o00050000002Oo`00Wol009oo000000000`2P
o`001@000000POl0087o0021o`00000000<0A?l000@00000003o4000oa0000000`00og@01@000000
0?oN003ogP00omh0000000<00;co00D00000001Xo`00J?l006So00000003000^o`04000000005_l0
01Ko000000<002;o00D00000001Bo`00D_l005;o00000003002Qo`0500000000ooH00?of003omP00
00000`00oh00100000000?l9003o2@000003067o00050000002eo`00]Ol00;Go000000000`3Yo`00
1@000000mol00?Oo003go`0000000080g?l000D0000009co002Lo`00W?l00000000303co00050000
0000ocT00?li003o>@0000000`00okP01@0000000<go003=o`00cOl00000008006?o00D00000000E
o`005Ol001Go0000000301<0o`050000000C0?l04`3o01<0o`0000000`005Ol01@00000006?o001S
o`00Hol0000000800<go00D00000003o^000okP00?nh00000003003o>@050000000lo`00??l003co
000000000`2Lo`001@000000g?l00=co003Lo`0000000080mol000D000000>Wo003Yo`00jOl00000
00030;Go00050000001Qo`00HOl0067o000000000`00o`T01@0000000?n0003oP000oh0000000080
0?of00D00000002Qo`00XOl00:7o00000003001Bo`05000000008_l002;o000Ro`0000000`005_l0
1@00000002ko000^o`00;_l00000008006So00D00000002lo`00_?l00;co00000003003ogP050000
0000og@00?md003oM00000000`00oa001@000000A?l004Co0014o`0000000080POl000D000000:3o
002Po`00X?l00000000309oo000500000020o`00P?l0083o000000000`17o`001@0000000?l5003o
1@00o`D0000000800?mL00D00000003o]000ok@00?nd00000003003lo`0500000000_?l00;co002l
o`0000000`00T_l01@00000008;o0022o`00P_l00000008008_o00D00000002/o`00[?l00:co0000
0003003Oo`0500000000on000?oP003oh00000000`00oiX0100000000?mG003oE`000003003o7005
0000000>o`003_l000ko000000000`0Xo`001@000000;_l002ko000^o`00000000<08Ol000<00000
0?ooo`3oool0103oool100000040oooo000?0?ooo`080?oOW01dW=l0oooo0?ooo`3ooml0W4P00018
W03Oool90?ooo`030000003oool0oooo00D0oooo00@00000027o000Qo`0000000`0^o`001@000000
:?l002So000Xo`00000000<03_l000D00000003o7000oa`00?lL00000003003oE`0400000000oiX0
0?nJ000000<00?oP00D00000003Oo`00gol00=oo00000003002/o`0500000000Rol008_o002;o`00
00000`00P_l01000000009;o002Bo`000003002lo`0500000000o?l00?co003lo`0000000`00ok@0
1@0000000?mL003oG000oe`0000000<00?l500@0000004Oo0017o`0000000`20o`001@000000Wol0
09oo002Oo`00000000<0X?l000D00000087o0021o`00POl00000000304Co000400000000oa000?l@
000000<00?md00D00000003ogP00omh00?oN00000003002lo`0500000000J?l006So001Xo`000000
0`00;_l01000000001Ko000Fo`000003000Ro`0500000000D_l005;o001Bo`0000000`00XOl01@00
00000?of003omP00ooH0000000<00?n000@00000003o2@00o`T000000`1Qo`001@000000]Ol00;Go
002eo`00000000<0jOl000D000000?Oo003go`00mol0000000020=co00050000002Lo`00W?l009co
000000000`0lo`001@0000000?li003o>@00ocT0000000<00?nh00D00000003=o`00cOl00<go0000
0002001So`05000000005Ol001Go000Eo`0000000`0C0?l01@0000004`3o01<0o`0C0?l0000000<0
01Go00D00000001So`00Hol006?o00000002003=o`0500000000okP00?nh003o^00000000`00ocT0
1@000000??l003co000lo`00000000<0W?l000D000000=co003Lo`00g?l0000000020?Oo00050000
003Yo`00jOl00>Wo000000000`2eo`001@000000HOl0067o001Qo`00000000<00?l900D00000003o
P000oh000?n000000002003omP0500000000XOl00:7o002Qo`0000000`00D_l01@00000002;o000R
o`008_l0000000<001Ko00D00000000^o`00;_l002ko00000002001Xo`0500000000_?l00;co002l
o`0000000`00omh01@0000000?md003oM000og@0000000<00?l@00D0000004Co0014o`00A?l00000
0002087o00050000002Po`00X?l00:3o000000000`2Oo`001@000000P?l0083o0020o`00000000<0
Aol000D00000003o1@00o`D00?l500000002003oG00500000000ok@00?nd003o]00000000`00o?l0
1@0000000;co002lo`00_?l0000000<009;o00D000000022o`00P_l008;o00000002002;o`050000
0000[?l00:co002/o`0000000`00gol01@0000000?oP003oh000on00000000<00?nJ00@00000003o
E`00oeL000000`00oa`01@0000003_l000ko000>o`00000000<0:?l000D0000002ko000^o`00;_l0
00000003027o00030000003oool0oooo00@0oooo0@0000010?ooo`00403oool01`3OW4P0M;oo0?oo
o`3ooml0W4P00018W03Oool02@3oool3000000D0ooooo`000021000000@0oooo0`0000010?ooo`00
403oool01`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool02@3oool00`000000oooo0?ooo`3o0?oo
oh`0oooo0@0000010?ooo`004@3oool01P3ogi`0B7Bo0?oog`2LB00004RL0=ooo`T0oooo00<00000
0?ooo`3oool0o`3ooon<0?ooo`4000000@3oool00180oooo00D0gia807Bog`2LB00004RL0=ooo`09
0?ooo`030000003oool0oooo0?l0ooooS03oool100000040oooo000B0?ooo`050?oog`2LB4P0B4P0
0018W03Oool02@3oool00`000000oooo0?ooo`3o0?oooh`0oooo0@0000010?ooo`004`3oool0103o
gi`0B0000018W03Oool90?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0G03oool0
0`000000oooo0?ooo`1M0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool0G@3oool0
0`000000oooo0?ooo`040?ooo`4000000@3oool001@0oooo00<0gia80018W03Oool02@3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0103oool10000
0040oooo000P0?ooool00000T00000010?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-4.83776, -4.75573, 0.0313285, \
0.0313285}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ContourPlot[
        Exp[\(-\((x\^2 + y\^2)\)\)/20] Cos[x\ y], {x, \(-4\), 4}, {y, \(-4\), 
          4}, Contours \[Rule] 25, PlotPoints \[Rule] 100, 
        ColorFunction \[Rule] Hue];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% ContourGraphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.120192 0.5 0.120192 [
[.01923 -0.0125 -9.15625 -15.6875 ]
[.01923 -0.0125 9.15625 0 ]
[.25962 -0.0125 -9.15625 -15.6875 ]
[.25962 -0.0125 9.15625 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.74038 -0.0125 -5 -15.6875 ]
[.74038 -0.0125 5 0 ]
[.98077 -0.0125 -5 -15.6875 ]
[.98077 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .01923 -18.3125 -7.84375 ]
[-0.0125 .01923 0 7.84375 ]
[-0.0125 .25962 -18.3125 -7.84375 ]
[-0.0125 .25962 0 7.84375 ]
[-0.0125 .5 -10 -7.84375 ]
[-0.0125 .5 0 7.84375 ]
[-0.0125 .74038 -10 -7.84375 ]
[-0.0125 .74038 0 7.84375 ]
[-0.0125 .98077 -10 -7.84375 ]
[-0.0125 .98077 0 7.84375 ]
[ 0 0 -0.125 0 ]
[ 0 1 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.01923 0 m
.01923 .00625 L
s
gsave
.01923 -0.0125 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.25962 0 m
.25962 .00625 L
s
gsave
.25962 -0.0125 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.74038 0 m
.74038 .00625 L
s
gsave
.74038 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.98077 0 m
.98077 .00625 L
s
gsave
.98077 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.07933 0 m
.07933 .00375 L
s
.13942 0 m
.13942 .00375 L
s
.19952 0 m
.19952 .00375 L
s
.31971 0 m
.31971 .00375 L
s
.37981 0 m
.37981 .00375 L
s
.4399 0 m
.4399 .00375 L
s
.5601 0 m
.5601 .00375 L
s
.62019 0 m
.62019 .00375 L
s
.68029 0 m
.68029 .00375 L
s
.80048 0 m
.80048 .00375 L
s
.86058 0 m
.86058 .00375 L
s
.92067 0 m
.92067 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .01923 m
.00625 .01923 L
s
gsave
-0.0125 .01923 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
0 .25962 m
.00625 .25962 L
s
gsave
-0.0125 .25962 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .5 m
.00625 .5 L
s
gsave
-0.0125 .5 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .74038 m
.00625 .74038 L
s
gsave
-0.0125 .74038 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .98077 m
.00625 .98077 L
s
gsave
-0.0125 .98077 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .07933 m
.00375 .07933 L
s
0 .13942 m
.00375 .13942 L
s
0 .19952 m
.00375 .19952 L
s
0 .31971 m
.00375 .31971 L
s
0 .37981 m
.00375 .37981 L
s
0 .4399 m
.00375 .4399 L
s
0 .5601 m
.00375 .5601 L
s
0 .62019 m
.00375 .62019 L
s
0 .68029 m
.00375 .68029 L
s
0 .80048 m
.00375 .80048 L
s
0 .86058 m
.00375 .86058 L
s
0 .92067 m
.00375 .92067 L
s
.25 Mabswid
0 0 m
0 1 L
s
.01923 .99375 m
.01923 1 L
s
.25962 .99375 m
.25962 1 L
s
.5 .99375 m
.5 1 L
s
.74038 .99375 m
.74038 1 L
s
.98077 .99375 m
.98077 1 L
s
.125 Mabswid
.07933 .99625 m
.07933 1 L
s
.13942 .99625 m
.13942 1 L
s
.19952 .99625 m
.19952 1 L
s
.31971 .99625 m
.31971 1 L
s
.37981 .99625 m
.37981 1 L
s
.4399 .99625 m
.4399 1 L
s
.5601 .99625 m
.5601 1 L
s
.62019 .99625 m
.62019 1 L
s
.68029 .99625 m
.68029 1 L
s
.80048 .99625 m
.80048 1 L
s
.86058 .99625 m
.86058 1 L
s
.92067 .99625 m
.92067 1 L
s
.25 Mabswid
0 1 m
1 1 L
s
.99375 .01923 m
1 .01923 L
s
.99375 .25962 m
1 .25962 L
s
.99375 .5 m
1 .5 L
s
.99375 .74038 m
1 .74038 L
s
.99375 .98077 m
1 .98077 L
s
.125 Mabswid
.99625 .07933 m
1 .07933 L
s
.99625 .13942 m
1 .13942 L
s
.99625 .19952 m
1 .19952 L
s
.99625 .31971 m
1 .31971 L
s
.99625 .37981 m
1 .37981 L
s
.99625 .4399 m
1 .4399 L
s
.99625 .5601 m
1 .5601 L
s
.99625 .62019 m
1 .62019 L
s
.99625 .68029 m
1 .68029 L
s
.99625 .80048 m
1 .80048 L
s
.99625 .86058 m
1 .86058 L
s
.99625 .92067 m
1 .92067 L
s
.25 Mabswid
1 0 m
1 1 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
.08 0 1 r
.01923 .98077 m
.98077 .98077 L
.98077 .01923 L
.01923 .01923 L
F
0 g
.5 Mabswid
0 .16 1 r
.01923 .48789 m
.02894 .48555 L
.02945 .48543 L
.03866 .48316 L
.04837 .48104 L
.05808 .47919 L
.06779 .47756 L
.07751 .47611 L
.08035 .47572 L
.08722 .47449 L
.09693 .47277 L
.10664 .47107 L
.11636 .4694 L
.12607 .46775 L
.13578 .46609 L
.13627 .46601 L
.14549 .46432 L
.15521 .46251 L
.16492 .46066 L
.17463 .45876 L
.18434 .45681 L
.18686 .45629 L
.19406 .45475 L
.20377 .4526 L
.21348 .45036 L
.22319 .448 L
.22884 .44658 L
.23291 .44552 L
.24262 .44287 L
.25233 .44007 L
.26204 .43709 L
.26273 .43687 L
.27176 .43388 L
.28147 .43043 L
.29005 .42716 L
.29118 .42671 L
.30089 .42266 L
.31061 .41825 L
.31229 .41744 L
.32032 .41341 L
.33003 .40808 L
.33064 .40773 L
.33974 .40217 L
.34597 .39802 L
.34946 .39556 L
.35895 .38831 L
.35917 .38813 L
.36888 .37971 L
.37008 .37859 L
.37859 .37008 L
.37971 .36888 L
.38813 .35917 L
.38831 .35895 L
.39556 .34946 L
.39802 .34597 L
.40217 .33974 L
.40773 .33064 L
.40808 .33003 L
.41341 .32032 L
.41744 .31229 L
.41825 .31061 L
.42266 .30089 L
.42671 .29118 L
.42716 .29005 L
.43043 .28147 L
.43388 .27176 L
.43687 .26273 L
.43709 .26204 L
.44007 .25233 L
.44287 .24262 L
.44552 .23291 L
.44658 .22884 L
.448 .22319 L
.45036 .21348 L
.4526 .20377 L
.45475 .19406 L
.45629 .18686 L
.45681 .18434 L
.45876 .17463 L
.46066 .16492 L
.46251 .15521 L
.46432 .14549 L
.46601 .13627 L
.46609 .13578 L
.46775 .12607 L
.4694 .11636 L
.47107 .10664 L
.47277 .09693 L
.47449 .08722 L
.47572 .08035 L
.47611 .07751 L
.47756 .06779 L
.47919 .05808 L
.48104 .04837 L
.48316 .03866 L
.48543 .02945 L
.48555 .02894 L
.48789 .01923 L
.01923 .01923 L
F
0 g
.01923 .48789 m
.02894 .48555 L
.02945 .48543 L
.03866 .48316 L
.04837 .48104 L
.05808 .47919 L
.06779 .47756 L
.07751 .47611 L
.08035 .47572 L
.08722 .47449 L
.09693 .47277 L
.10664 .47107 L
.11636 .4694 L
.12607 .46775 L
.13578 .46609 L
.13627 .46601 L
.14549 .46432 L
.15521 .46251 L
.16492 .46066 L
.17463 .45876 L
.18434 .45681 L
.18686 .45629 L
.19406 .45475 L
.20377 .4526 L
.21348 .45036 L
.22319 .448 L
.22884 .44658 L
.23291 .44552 L
.24262 .44287 L
.25233 .44007 L
.26204 .43709 L
.26273 .43687 L
.27176 .43388 L
.28147 .43043 L
.29005 .42716 L
.29118 .42671 L
.30089 .42266 L
.31061 .41825 L
.31229 .41744 L
.32032 .41341 L
.33003 .40808 L
.33064 .40773 L
.33974 .40217 L
.34597 .39802 L
.34946 .39556 L
.35895 .38831 L
.35917 .38813 L
.36888 .37971 L
.37008 .37859 L
.37859 .37008 L
Mistroke
.37971 .36888 L
.38813 .35917 L
.38831 .35895 L
.39556 .34946 L
.39802 .34597 L
.40217 .33974 L
.40773 .33064 L
.40808 .33003 L
.41341 .32032 L
.41744 .31229 L
.41825 .31061 L
.42266 .30089 L
.42671 .29118 L
.42716 .29005 L
.43043 .28147 L
.43388 .27176 L
.43687 .26273 L
.43709 .26204 L
.44007 .25233 L
.44287 .24262 L
.44552 .23291 L
.44658 .22884 L
.448 .22319 L
.45036 .21348 L
.4526 .20377 L
.45475 .19406 L
.45629 .18686 L
.45681 .18434 L
.45876 .17463 L
.46066 .16492 L
.46251 .15521 L
.46432 .14549 L
.46601 .13627 L
.46609 .13578 L
.46775 .12607 L
.4694 .11636 L
.47107 .10664 L
.47277 .09693 L
.47449 .08722 L
.47572 .08035 L
.47611 .07751 L
.47756 .06779 L
.47919 .05808 L
.48104 .04837 L
.48316 .03866 L
.48543 .02945 L
.48555 .02894 L
.48789 .01923 L
Mfstroke
0 .16 1 r
.01923 .51211 m
.02894 .51445 L
.02945 .51457 L
.03866 .51684 L
.04837 .51896 L
.05808 .52081 L
.06779 .52244 L
.07751 .52389 L
.08035 .52428 L
.08722 .52551 L
.09693 .52723 L
.10664 .52893 L
.11636 .5306 L
.12607 .53225 L
.13578 .53391 L
.13627 .53399 L
.14549 .53568 L
.15521 .53749 L
.16492 .53934 L
.17463 .54124 L
.18434 .54319 L
.18686 .54371 L
.19406 .54525 L
.20377 .5474 L
.21348 .54964 L
.22319 .552 L
.22884 .55342 L
.23291 .55448 L
.24262 .55713 L
.25233 .55993 L
.26204 .56291 L
.26273 .56313 L
.27176 .56612 L
.28147 .56957 L
.29005 .57284 L
.29118 .57329 L
.30089 .57734 L
.31061 .58175 L
.31229 .58256 L
.32032 .58659 L
.33003 .59192 L
.33064 .59227 L
.33974 .59783 L
.34597 .60198 L
.34946 .60444 L
.35895 .61169 L
.35917 .61187 L
.36888 .62029 L
.37008 .62141 L
.37859 .62992 L
.37971 .63112 L
.38813 .64083 L
.38831 .64105 L
.39556 .65054 L
.39802 .65403 L
.40217 .66026 L
.40773 .66936 L
.40808 .66997 L
.41341 .67968 L
.41744 .68771 L
.41825 .68939 L
.42266 .69911 L
.42671 .70882 L
.42716 .70995 L
.43043 .71853 L
.43388 .72824 L
.43687 .73727 L
.43709 .73796 L
.44007 .74767 L
.44287 .75738 L
.44552 .76709 L
.44658 .77116 L
.448 .77681 L
.45036 .78652 L
.4526 .79623 L
.45475 .80594 L
.45629 .81314 L
.45681 .81566 L
.45876 .82537 L
.46066 .83508 L
.46251 .84479 L
.46432 .85451 L
.46601 .86373 L
.46609 .86422 L
.46775 .87393 L
.4694 .88364 L
.47107 .89336 L
.47277 .90307 L
.47449 .91278 L
.47572 .91965 L
.47611 .92249 L
.47756 .93221 L
.47919 .94192 L
.48104 .95163 L
.48316 .96134 L
.48543 .97055 L
.48555 .97106 L
.48789 .98077 L
.01923 .98077 L
F
0 g
.01923 .51211 m
.02894 .51445 L
.02945 .51457 L
.03866 .51684 L
.04837 .51896 L
.05808 .52081 L
.06779 .52244 L
.07751 .52389 L
.08035 .52428 L
.08722 .52551 L
.09693 .52723 L
.10664 .52893 L
.11636 .5306 L
.12607 .53225 L
.13578 .53391 L
.13627 .53399 L
.14549 .53568 L
.15521 .53749 L
.16492 .53934 L
.17463 .54124 L
.18434 .54319 L
.18686 .54371 L
.19406 .54525 L
.20377 .5474 L
.21348 .54964 L
.22319 .552 L
.22884 .55342 L
.23291 .55448 L
.24262 .55713 L
.25233 .55993 L
.26204 .56291 L
.26273 .56313 L
.27176 .56612 L
.28147 .56957 L
.29005 .57284 L
.29118 .57329 L
.30089 .57734 L
.31061 .58175 L
.31229 .58256 L
.32032 .58659 L
.33003 .59192 L
.33064 .59227 L
.33974 .59783 L
.34597 .60198 L
.34946 .60444 L
.35895 .61169 L
.35917 .61187 L
.36888 .62029 L
.37008 .62141 L
.37859 .62992 L
Mistroke
.37971 .63112 L
.38813 .64083 L
.38831 .64105 L
.39556 .65054 L
.39802 .65403 L
.40217 .66026 L
.40773 .66936 L
.40808 .66997 L
.41341 .67968 L
.41744 .68771 L
.41825 .68939 L
.42266 .69911 L
.42671 .70882 L
.42716 .70995 L
.43043 .71853 L
.43388 .72824 L
.43687 .73727 L
.43709 .73796 L
.44007 .74767 L
.44287 .75738 L
.44552 .76709 L
.44658 .77116 L
.448 .77681 L
.45036 .78652 L
.4526 .79623 L
.45475 .80594 L
.45629 .81314 L
.45681 .81566 L
.45876 .82537 L
.46066 .83508 L
.46251 .84479 L
.46432 .85451 L
.46601 .86373 L
.46609 .86422 L
.46775 .87393 L
.4694 .88364 L
.47107 .89336 L
.47277 .90307 L
.47449 .91278 L
.47572 .91965 L
.47611 .92249 L
.47756 .93221 L
.47919 .94192 L
.48104 .95163 L
.48316 .96134 L
.48543 .97055 L
.48555 .97106 L
.48789 .98077 L
Mfstroke
0 .4 1 r
.01923 .4785 m
.02894 .47711 L
.03866 .47586 L
.03979 .47572 L
.04837 .47436 L
.05808 .47284 L
.06779 .47134 L
.07751 .46986 L
.08722 .46839 L
.09693 .46692 L
.10294 .46601 L
.10664 .46541 L
.11636 .46383 L
.12607 .46221 L
.13578 .46056 L
.14549 .45887 L
.15521 .45713 L
.1598 .45629 L
.16492 .45532 L
.17463 .45342 L
.18434 .45144 L
.19406 .44938 L
.20377 .44721 L
.20653 .44658 L
.21348 .44493 L
.22319 .44251 L
.23291 .43995 L
.24262 .43724 L
.2439 .43687 L
.25233 .43434 L
.26204 .43123 L
.27176 .42789 L
.27381 .42716 L
.28147 .42429 L
.29118 .42038 L
.298 .41744 L
.30089 .41614 L
.31061 .41149 L
.31786 .40773 L
.32032 .40639 L
.33003 .40075 L
.3344 .39802 L
.33974 .39448 L
.34835 .38831 L
.34946 .38747 L
.35917 .37956 L
.36027 .37859 L
.36888 .37057 L
.37057 .36888 L
.37859 .36027 L
.37956 .35917 L
.38747 .34946 L
.38831 .34835 L
.39448 .33974 L
.39802 .3344 L
.40075 .33003 L
.40639 .32032 L
.40773 .31786 L
.41149 .31061 L
.41614 .30089 L
.41744 .298 L
.42038 .29118 L
.42429 .28147 L
.42716 .27381 L
.42789 .27176 L
.43123 .26204 L
.43434 .25233 L
.43687 .2439 L
.43724 .24262 L
.43995 .23291 L
.44251 .22319 L
.44493 .21348 L
.44658 .20653 L
.44721 .20377 L
.44938 .19406 L
.45144 .18434 L
.45342 .17463 L
.45532 .16492 L
.45629 .1598 L
.45713 .15521 L
.45887 .14549 L
.46056 .13578 L
.46221 .12607 L
.46383 .11636 L
.46541 .10664 L
.46601 .10294 L
.46692 .09693 L
.46839 .08722 L
.46986 .07751 L
.47134 .06779 L
.47284 .05808 L
.47436 .04837 L
.47572 .03979 L
.47586 .03866 L
.47711 .02894 L
.4785 .01923 L
.01923 .01923 L
F
0 g
.01923 .4785 m
.02894 .47711 L
.03866 .47586 L
.03979 .47572 L
.04837 .47436 L
.05808 .47284 L
.06779 .47134 L
.07751 .46986 L
.08722 .46839 L
.09693 .46692 L
.10294 .46601 L
.10664 .46541 L
.11636 .46383 L
.12607 .46221 L
.13578 .46056 L
.14549 .45887 L
.15521 .45713 L
.1598 .45629 L
.16492 .45532 L
.17463 .45342 L
.18434 .45144 L
.19406 .44938 L
.20377 .44721 L
.20653 .44658 L
.21348 .44493 L
.22319 .44251 L
.23291 .43995 L
.24262 .43724 L
.2439 .43687 L
.25233 .43434 L
.26204 .43123 L
.27176 .42789 L
.27381 .42716 L
.28147 .42429 L
.29118 .42038 L
.298 .41744 L
.30089 .41614 L
.31061 .41149 L
.31786 .40773 L
.32032 .40639 L
.33003 .40075 L
.3344 .39802 L
.33974 .39448 L
.34835 .38831 L
.34946 .38747 L
.35917 .37956 L
.36027 .37859 L
.36888 .37057 L
.37057 .36888 L
.37859 .36027 L
Mistroke
.37956 .35917 L
.38747 .34946 L
.38831 .34835 L
.39448 .33974 L
.39802 .3344 L
.40075 .33003 L
.40639 .32032 L
.40773 .31786 L
.41149 .31061 L
.41614 .30089 L
.41744 .298 L
.42038 .29118 L
.42429 .28147 L
.42716 .27381 L
.42789 .27176 L
.43123 .26204 L
.43434 .25233 L
.43687 .2439 L
.43724 .24262 L
.43995 .23291 L
.44251 .22319 L
.44493 .21348 L
.44658 .20653 L
.44721 .20377 L
.44938 .19406 L
.45144 .18434 L
.45342 .17463 L
.45532 .16492 L
.45629 .1598 L
.45713 .15521 L
.45887 .14549 L
.46056 .13578 L
.46221 .12607 L
.46383 .11636 L
.46541 .10664 L
.46601 .10294 L
.46692 .09693 L
.46839 .08722 L
.46986 .07751 L
.47134 .06779 L
.47284 .05808 L
.47436 .04837 L
.47572 .03979 L
.47586 .03866 L
.47711 .02894 L
.4785 .01923 L
Mfstroke
0 .4 1 r
.01923 .5215 m
.02894 .52289 L
.03866 .52414 L
.03979 .52428 L
.04837 .52564 L
.05808 .52716 L
.06779 .52866 L
.07751 .53014 L
.08722 .53161 L
.09693 .53308 L
.10294 .53399 L
.10664 .53459 L
.11636 .53617 L
.12607 .53779 L
.13578 .53944 L
.14549 .54113 L
.15521 .54287 L
.1598 .54371 L
.16492 .54468 L
.17463 .54658 L
.18434 .54856 L
.19406 .55062 L
.20377 .55279 L
.20653 .55342 L
.21348 .55507 L
.22319 .55749 L
.23291 .56005 L
.24262 .56276 L
.2439 .56313 L
.25233 .56566 L
.26204 .56877 L
.27176 .57211 L
.27381 .57284 L
.28147 .57571 L
.29118 .57962 L
.298 .58256 L
.30089 .58386 L
.31061 .58851 L
.31786 .59227 L
.32032 .59361 L
.33003 .59925 L
.3344 .60198 L
.33974 .60552 L
.34835 .61169 L
.34946 .61253 L
.35917 .62044 L
.36027 .62141 L
.36888 .62943 L
.37057 .63112 L
.37859 .63973 L
.37956 .64083 L
.38747 .65054 L
.38831 .65165 L
.39448 .66026 L
.39802 .6656 L
.40075 .66997 L
.40639 .67968 L
.40773 .68214 L
.41149 .68939 L
.41614 .69911 L
.41744 .702 L
.42038 .70882 L
.42429 .71853 L
.42716 .72619 L
.42789 .72824 L
.43123 .73796 L
.43434 .74767 L
.43687 .7561 L
.43724 .75738 L
.43995 .76709 L
.44251 .77681 L
.44493 .78652 L
.44658 .79347 L
.44721 .79623 L
.44938 .80594 L
.45144 .81566 L
.45342 .82537 L
.45532 .83508 L
.45629 .8402 L
.45713 .84479 L
.45887 .85451 L
.46056 .86422 L
.46221 .87393 L
.46383 .88364 L
.46541 .89336 L
.46601 .89706 L
.46692 .90307 L
.46839 .91278 L
.46986 .92249 L
.47134 .93221 L
.47284 .94192 L
.47436 .95163 L
.47572 .96021 L
.47586 .96134 L
.47711 .97106 L
.4785 .98077 L
.01923 .98077 L
F
0 g
.01923 .5215 m
.02894 .52289 L
.03866 .52414 L
.03979 .52428 L
.04837 .52564 L
.05808 .52716 L
.06779 .52866 L
.07751 .53014 L
.08722 .53161 L
.09693 .53308 L
.10294 .53399 L
.10664 .53459 L
.11636 .53617 L
.12607 .53779 L
.13578 .53944 L
.14549 .54113 L
.15521 .54287 L
.1598 .54371 L
.16492 .54468 L
.17463 .54658 L
.18434 .54856 L
.19406 .55062 L
.20377 .55279 L
.20653 .55342 L
.21348 .55507 L
.22319 .55749 L
.23291 .56005 L
.24262 .56276 L
.2439 .56313 L
.25233 .56566 L
.26204 .56877 L
.27176 .57211 L
.27381 .57284 L
.28147 .57571 L
.29118 .57962 L
.298 .58256 L
.30089 .58386 L
.31061 .58851 L
.31786 .59227 L
.32032 .59361 L
.33003 .59925 L
.3344 .60198 L
.33974 .60552 L
.34835 .61169 L
.34946 .61253 L
.35917 .62044 L
.36027 .62141 L
.36888 .62943 L
.37057 .63112 L
.37859 .63973 L
Mistroke
.37956 .64083 L
.38747 .65054 L
.38831 .65165 L
.39448 .66026 L
.39802 .6656 L
.40075 .66997 L
.40639 .67968 L
.40773 .68214 L
.41149 .68939 L
.41614 .69911 L
.41744 .702 L
.42038 .70882 L
.42429 .71853 L
.42716 .72619 L
.42789 .72824 L
.43123 .73796 L
.43434 .74767 L
.43687 .7561 L
.43724 .75738 L
.43995 .76709 L
.44251 .77681 L
.44493 .78652 L
.44658 .79347 L
.44721 .79623 L
.44938 .80594 L
.45144 .81566 L
.45342 .82537 L
.45532 .83508 L
.45629 .8402 L
.45713 .84479 L
.45887 .85451 L
.46056 .86422 L
.46221 .87393 L
.46383 .88364 L
.46541 .89336 L
.46601 .89706 L
.46692 .90307 L
.46839 .91278 L
.46986 .92249 L
.47134 .93221 L
.47284 .94192 L
.47436 .95163 L
.47572 .96021 L
.47586 .96134 L
.47711 .97106 L
.4785 .98077 L
Mfstroke
0 .64 1 r
.01923 .4724 m
.02894 .47109 L
.03866 .46979 L
.04837 .4685 L
.05808 .4672 L
.06706 .46601 L
.06779 .4659 L
.07751 .46452 L
.08722 .46311 L
.09693 .46167 L
.10664 .46021 L
.11636 .4587 L
.12607 .45715 L
.13135 .45629 L
.13578 .45555 L
.14549 .45386 L
.15521 .45211 L
.16492 .4503 L
.17463 .4484 L
.18354 .44658 L
.18434 .44641 L
.19406 .44431 L
.20377 .4421 L
.21348 .43976 L
.22319 .43729 L
.22478 .43687 L
.23291 .43465 L
.24262 .43184 L
.25233 .42883 L
.25747 .42716 L
.26204 .4256 L
.27176 .42212 L
.28147 .41836 L
.28373 .41744 L
.29118 .41428 L
.30089 .40982 L
.30517 .40773 L
.31061 .40494 L
.32032 .39957 L
.32295 .39802 L
.33003 .39362 L
.3379 .38831 L
.33974 .38699 L
.34946 .37956 L
.35064 .37859 L
.35917 .37117 L
.36162 .36888 L
.36888 .36162 L
.37117 .35917 L
.37859 .35064 L
.37956 .34946 L
.38699 .33974 L
.38831 .3379 L
.39362 .33003 L
.39802 .32295 L
.39957 .32032 L
.40494 .31061 L
.40773 .30517 L
.40982 .30089 L
.41428 .29118 L
.41744 .28373 L
.41836 .28147 L
.42212 .27176 L
.4256 .26204 L
.42716 .25747 L
.42883 .25233 L
.43184 .24262 L
.43465 .23291 L
.43687 .22478 L
.43729 .22319 L
.43976 .21348 L
.4421 .20377 L
.44431 .19406 L
.44641 .18434 L
.44658 .18354 L
.4484 .17463 L
.4503 .16492 L
.45211 .15521 L
.45386 .14549 L
.45555 .13578 L
.45629 .13135 L
.45715 .12607 L
.4587 .11636 L
.46021 .10664 L
.46167 .09693 L
.46311 .08722 L
.46452 .07751 L
.4659 .06779 L
.46601 .06706 L
.4672 .05808 L
.4685 .04837 L
.46979 .03866 L
.47109 .02894 L
.4724 .01923 L
.01923 .01923 L
F
0 g
.01923 .4724 m
.02894 .47109 L
.03866 .46979 L
.04837 .4685 L
.05808 .4672 L
.06706 .46601 L
.06779 .4659 L
.07751 .46452 L
.08722 .46311 L
.09693 .46167 L
.10664 .46021 L
.11636 .4587 L
.12607 .45715 L
.13135 .45629 L
.13578 .45555 L
.14549 .45386 L
.15521 .45211 L
.16492 .4503 L
.17463 .4484 L
.18354 .44658 L
.18434 .44641 L
.19406 .44431 L
.20377 .4421 L
.21348 .43976 L
.22319 .43729 L
.22478 .43687 L
.23291 .43465 L
.24262 .43184 L
.25233 .42883 L
.25747 .42716 L
.26204 .4256 L
.27176 .42212 L
.28147 .41836 L
.28373 .41744 L
.29118 .41428 L
.30089 .40982 L
.30517 .40773 L
.31061 .40494 L
.32032 .39957 L
.32295 .39802 L
.33003 .39362 L
.3379 .38831 L
.33974 .38699 L
.34946 .37956 L
.35064 .37859 L
.35917 .37117 L
.36162 .36888 L
.36888 .36162 L
.37117 .35917 L
.37859 .35064 L
Mistroke
.37956 .34946 L
.38699 .33974 L
.38831 .3379 L
.39362 .33003 L
.39802 .32295 L
.39957 .32032 L
.40494 .31061 L
.40773 .30517 L
.40982 .30089 L
.41428 .29118 L
.41744 .28373 L
.41836 .28147 L
.42212 .27176 L
.4256 .26204 L
.42716 .25747 L
.42883 .25233 L
.43184 .24262 L
.43465 .23291 L
.43687 .22478 L
.43729 .22319 L
.43976 .21348 L
.4421 .20377 L
.44431 .19406 L
.44641 .18434 L
.44658 .18354 L
.4484 .17463 L
.4503 .16492 L
.45211 .15521 L
.45386 .14549 L
.45555 .13578 L
.45629 .13135 L
.45715 .12607 L
.4587 .11636 L
.46021 .10664 L
.46167 .09693 L
.46311 .08722 L
.46452 .07751 L
.4659 .06779 L
.46601 .06706 L
.4672 .05808 L
.4685 .04837 L
.46979 .03866 L
.47109 .02894 L
.4724 .01923 L
Mfstroke
0 .64 1 r
.01923 .5276 m
.02894 .52891 L
.03866 .53021 L
.04837 .5315 L
.05808 .5328 L
.06706 .53399 L
.06779 .5341 L
.07751 .53548 L
.08722 .53689 L
.09693 .53833 L
.10664 .53979 L
.11636 .5413 L
.12607 .54285 L
.13135 .54371 L
.13578 .54445 L
.14549 .54614 L
.15521 .54789 L
.16492 .5497 L
.17463 .5516 L
.18354 .55342 L
.18434 .55359 L
.19406 .55569 L
.20377 .5579 L
.21348 .56024 L
.22319 .56271 L
.22478 .56313 L
.23291 .56535 L
.24262 .56816 L
.25233 .57117 L
.25747 .57284 L
.26204 .5744 L
.27176 .57788 L
.28147 .58164 L
.28373 .58256 L
.29118 .58572 L
.30089 .59018 L
.30517 .59227 L
.31061 .59506 L
.32032 .60043 L
.32295 .60198 L
.33003 .60638 L
.3379 .61169 L
.33974 .61301 L
.34946 .62044 L
.35064 .62141 L
.35917 .62883 L
.36162 .63112 L
.36888 .63838 L
.37117 .64083 L
.37859 .64936 L
.37956 .65054 L
.38699 .66026 L
.38831 .6621 L
.39362 .66997 L
.39802 .67705 L
.39957 .67968 L
.40494 .68939 L
.40773 .69483 L
.40982 .69911 L
.41428 .70882 L
.41744 .71627 L
.41836 .71853 L
.42212 .72824 L
.4256 .73796 L
.42716 .74253 L
.42883 .74767 L
.43184 .75738 L
.43465 .76709 L
.43687 .77522 L
.43729 .77681 L
.43976 .78652 L
.4421 .79623 L
.44431 .80594 L
.44641 .81566 L
.44658 .81646 L
.4484 .82537 L
.4503 .83508 L
.45211 .84479 L
.45386 .85451 L
.45555 .86422 L
.45629 .86865 L
.45715 .87393 L
.4587 .88364 L
.46021 .89336 L
.46167 .90307 L
.46311 .91278 L
.46452 .92249 L
.4659 .93221 L
.46601 .93294 L
.4672 .94192 L
.4685 .95163 L
.46979 .96134 L
.47109 .97106 L
.4724 .98077 L
.01923 .98077 L
F
0 g
.01923 .5276 m
.02894 .52891 L
.03866 .53021 L
.04837 .5315 L
.05808 .5328 L
.06706 .53399 L
.06779 .5341 L
.07751 .53548 L
.08722 .53689 L
.09693 .53833 L
.10664 .53979 L
.11636 .5413 L
.12607 .54285 L
.13135 .54371 L
.13578 .54445 L
.14549 .54614 L
.15521 .54789 L
.16492 .5497 L
.17463 .5516 L
.18354 .55342 L
.18434 .55359 L
.19406 .55569 L
.20377 .5579 L
.21348 .56024 L
.22319 .56271 L
.22478 .56313 L
.23291 .56535 L
.24262 .56816 L
.25233 .57117 L
.25747 .57284 L
.26204 .5744 L
.27176 .57788 L
.28147 .58164 L
.28373 .58256 L
.29118 .58572 L
.30089 .59018 L
.30517 .59227 L
.31061 .59506 L
.32032 .60043 L
.32295 .60198 L
.33003 .60638 L
.3379 .61169 L
.33974 .61301 L
.34946 .62044 L
.35064 .62141 L
.35917 .62883 L
.36162 .63112 L
.36888 .63838 L
.37117 .64083 L
.37859 .64936 L
Mistroke
.37956 .65054 L
.38699 .66026 L
.38831 .6621 L
.39362 .66997 L
.39802 .67705 L
.39957 .67968 L
.40494 .68939 L
.40773 .69483 L
.40982 .69911 L
.41428 .70882 L
.41744 .71627 L
.41836 .71853 L
.42212 .72824 L
.4256 .73796 L
.42716 .74253 L
.42883 .74767 L
.43184 .75738 L
.43465 .76709 L
.43687 .77522 L
.43729 .77681 L
.43976 .78652 L
.4421 .79623 L
.44431 .80594 L
.44641 .81566 L
.44658 .81646 L
.4484 .82537 L
.4503 .83508 L
.45211 .84479 L
.45386 .85451 L
.45555 .86422 L
.45629 .86865 L
.45715 .87393 L
.4587 .88364 L
.46021 .89336 L
.46167 .90307 L
.46311 .91278 L
.46452 .92249 L
.4659 .93221 L
.46601 .93294 L
.4672 .94192 L
.4685 .95163 L
.46979 .96134 L
.47109 .97106 L
.4724 .98077 L
Mfstroke
0 .88 1 r
.01923 .46691 m
.02705 .46601 L
.02894 .46577 L
.03866 .46458 L
.04837 .46336 L
.05808 .46211 L
.06779 .46084 L
.07751 .45954 L
.08722 .45821 L
.09693 .45684 L
.1007 .45629 L
.10664 .45541 L
.11636 .45392 L
.12607 .45238 L
.13578 .45077 L
.14549 .4491 L
.15521 .44736 L
.1594 .44658 L
.16492 .44553 L
.17463 .4436 L
.18434 .44157 L
.19406 .43943 L
.20377 .43718 L
.20505 .43687 L
.21348 .43478 L
.22319 .43223 L
.23291 .42951 L
.24083 .42716 L
.24262 .4266 L
.25233 .42349 L
.26204 .42013 L
.26933 .41744 L
.27176 .41651 L
.28147 .41258 L
.29118 .40832 L
.29245 .40773 L
.30089 .40365 L
.31061 .39853 L
.31153 .39802 L
.32032 .39288 L
.32751 .38831 L
.33003 .38662 L
.33974 .37963 L
.34109 .37859 L
.34946 .37178 L
.35276 .36888 L
.35917 .36289 L
.36289 .35917 L
.36888 .35276 L
.37178 .34946 L
.37859 .34109 L
.37963 .33974 L
.38662 .33003 L
.38831 .32751 L
.39288 .32032 L
.39802 .31153 L
.39853 .31061 L
.40365 .30089 L
.40773 .29245 L
.40832 .29118 L
.41258 .28147 L
.41651 .27176 L
.41744 .26933 L
.42013 .26204 L
.42349 .25233 L
.4266 .24262 L
.42716 .24083 L
.42951 .23291 L
.43223 .22319 L
.43478 .21348 L
.43687 .20505 L
.43718 .20377 L
.43943 .19406 L
.44157 .18434 L
.4436 .17463 L
.44553 .16492 L
.44658 .1594 L
.44736 .15521 L
.4491 .14549 L
.45077 .13578 L
.45238 .12607 L
.45392 .11636 L
.45541 .10664 L
.45629 .1007 L
.45684 .09693 L
.45821 .08722 L
.45954 .07751 L
.46084 .06779 L
.46211 .05808 L
.46336 .04837 L
.46458 .03866 L
.46577 .02894 L
.46601 .02705 L
.46691 .01923 L
.01923 .01923 L
F
0 g
.01923 .46691 m
.02705 .46601 L
.02894 .46577 L
.03866 .46458 L
.04837 .46336 L
.05808 .46211 L
.06779 .46084 L
.07751 .45954 L
.08722 .45821 L
.09693 .45684 L
.1007 .45629 L
.10664 .45541 L
.11636 .45392 L
.12607 .45238 L
.13578 .45077 L
.14549 .4491 L
.15521 .44736 L
.1594 .44658 L
.16492 .44553 L
.17463 .4436 L
.18434 .44157 L
.19406 .43943 L
.20377 .43718 L
.20505 .43687 L
.21348 .43478 L
.22319 .43223 L
.23291 .42951 L
.24083 .42716 L
.24262 .4266 L
.25233 .42349 L
.26204 .42013 L
.26933 .41744 L
.27176 .41651 L
.28147 .41258 L
.29118 .40832 L
.29245 .40773 L
.30089 .40365 L
.31061 .39853 L
.31153 .39802 L
.32032 .39288 L
.32751 .38831 L
.33003 .38662 L
.33974 .37963 L
.34109 .37859 L
.34946 .37178 L
.35276 .36888 L
.35917 .36289 L
.36289 .35917 L
.36888 .35276 L
.37178 .34946 L
Mistroke
.37859 .34109 L
.37963 .33974 L
.38662 .33003 L
.38831 .32751 L
.39288 .32032 L
.39802 .31153 L
.39853 .31061 L
.40365 .30089 L
.40773 .29245 L
.40832 .29118 L
.41258 .28147 L
.41651 .27176 L
.41744 .26933 L
.42013 .26204 L
.42349 .25233 L
.4266 .24262 L
.42716 .24083 L
.42951 .23291 L
.43223 .22319 L
.43478 .21348 L
.43687 .20505 L
.43718 .20377 L
.43943 .19406 L
.44157 .18434 L
.4436 .17463 L
.44553 .16492 L
.44658 .1594 L
.44736 .15521 L
.4491 .14549 L
.45077 .13578 L
.45238 .12607 L
.45392 .11636 L
.45541 .10664 L
.45629 .1007 L
.45684 .09693 L
.45821 .08722 L
.45954 .07751 L
.46084 .06779 L
.46211 .05808 L
.46336 .04837 L
.46458 .03866 L
.46577 .02894 L
.46601 .02705 L
.46691 .01923 L
Mfstroke
0 .88 1 r
.01923 .53309 m
.02705 .53399 L
.02894 .53423 L
.03866 .53542 L
.04837 .53664 L
.05808 .53789 L
.06779 .53916 L
.07751 .54046 L
.08722 .54179 L
.09693 .54316 L
.1007 .54371 L
.10664 .54459 L
.11636 .54608 L
.12607 .54762 L
.13578 .54923 L
.14549 .5509 L
.15521 .55264 L
.1594 .55342 L
.16492 .55447 L
.17463 .5564 L
.18434 .55843 L
.19406 .56057 L
.20377 .56282 L
.20505 .56313 L
.21348 .56522 L
.22319 .56777 L
.23291 .57049 L
.24083 .57284 L
.24262 .5734 L
.25233 .57651 L
.26204 .57987 L
.26933 .58256 L
.27176 .58349 L
.28147 .58742 L
.29118 .59168 L
.29245 .59227 L
.30089 .59635 L
.31061 .60147 L
.31153 .60198 L
.32032 .60712 L
.32751 .61169 L
.33003 .61338 L
.33974 .62037 L
.34109 .62141 L
.34946 .62822 L
.35276 .63112 L
.35917 .63711 L
.36289 .64083 L
.36888 .64724 L
.37178 .65054 L
.37859 .65891 L
.37963 .66026 L
.38662 .66997 L
.38831 .67249 L
.39288 .67968 L
.39802 .68847 L
.39853 .68939 L
.40365 .69911 L
.40773 .70755 L
.40832 .70882 L
.41258 .71853 L
.41651 .72824 L
.41744 .73067 L
.42013 .73796 L
.42349 .74767 L
.4266 .75738 L
.42716 .75917 L
.42951 .76709 L
.43223 .77681 L
.43478 .78652 L
.43687 .79495 L
.43718 .79623 L
.43943 .80594 L
.44157 .81566 L
.4436 .82537 L
.44553 .83508 L
.44658 .8406 L
.44736 .84479 L
.4491 .85451 L
.45077 .86422 L
.45238 .87393 L
.45392 .88364 L
.45541 .89336 L
.45629 .8993 L
.45684 .90307 L
.45821 .91278 L
.45954 .92249 L
.46084 .93221 L
.46211 .94192 L
.46336 .95163 L
.46458 .96134 L
.46577 .97106 L
.46601 .97295 L
.46691 .98077 L
.01923 .98077 L
F
0 g
.01923 .53309 m
.02705 .53399 L
.02894 .53423 L
.03866 .53542 L
.04837 .53664 L
.05808 .53789 L
.06779 .53916 L
.07751 .54046 L
.08722 .54179 L
.09693 .54316 L
.1007 .54371 L
.10664 .54459 L
.11636 .54608 L
.12607 .54762 L
.13578 .54923 L
.14549 .5509 L
.15521 .55264 L
.1594 .55342 L
.16492 .55447 L
.17463 .5564 L
.18434 .55843 L
.19406 .56057 L
.20377 .56282 L
.20505 .56313 L
.21348 .56522 L
.22319 .56777 L
.23291 .57049 L
.24083 .57284 L
.24262 .5734 L
.25233 .57651 L
.26204 .57987 L
.26933 .58256 L
.27176 .58349 L
.28147 .58742 L
.29118 .59168 L
.29245 .59227 L
.30089 .59635 L
.31061 .60147 L
.31153 .60198 L
.32032 .60712 L
.32751 .61169 L
.33003 .61338 L
.33974 .62037 L
.34109 .62141 L
.34946 .62822 L
.35276 .63112 L
.35917 .63711 L
.36289 .64083 L
.36888 .64724 L
.37178 .65054 L
Mistroke
.37859 .65891 L
.37963 .66026 L
.38662 .66997 L
.38831 .67249 L
.39288 .67968 L
.39802 .68847 L
.39853 .68939 L
.40365 .69911 L
.40773 .70755 L
.40832 .70882 L
.41258 .71853 L
.41651 .72824 L
.41744 .73067 L
.42013 .73796 L
.42349 .74767 L
.4266 .75738 L
.42716 .75917 L
.42951 .76709 L
.43223 .77681 L
.43478 .78652 L
.43687 .79495 L
.43718 .79623 L
.43943 .80594 L
.44157 .81566 L
.4436 .82537 L
.44553 .83508 L
.44658 .8406 L
.44736 .84479 L
.4491 .85451 L
.45077 .86422 L
.45238 .87393 L
.45392 .88364 L
.45541 .89336 L
.45629 .8993 L
.45684 .90307 L
.45821 .91278 L
.45954 .92249 L
.46084 .93221 L
.46211 .94192 L
.46336 .95163 L
.46458 .96134 L
.46577 .97106 L
.46601 .97295 L
.46691 .98077 L
Mfstroke
0 1 .88 r
.01923 .46185 m
.02894 .46077 L
.03866 .45966 L
.04837 .45852 L
.05808 .45734 L
.06655 .45629 L
.06779 .45613 L
.07751 .45487 L
.08722 .45356 L
.09693 .4522 L
.10664 .45079 L
.11636 .44932 L
.12607 .44779 L
.13348 .44658 L
.13578 .44619 L
.14549 .44451 L
.15521 .44275 L
.16492 .44089 L
.17463 .43894 L
.18434 .43688 L
.18438 .43687 L
.19406 .43469 L
.20377 .43237 L
.21348 .42991 L
.22319 .42729 L
.22366 .42716 L
.23291 .42448 L
.24262 .42147 L
.25233 .41824 L
.25462 .41744 L
.26204 .41476 L
.27176 .41099 L
.27955 .40773 L
.28147 .4069 L
.29118 .40244 L
.30001 .39802 L
.30089 .39756 L
.31061 .39219 L
.31708 .38831 L
.32032 .38626 L
.33003 .37968 L
.33154 .37859 L
.33974 .37232 L
.34392 .36888 L
.34946 .36404 L
.35465 .35917 L
.35917 .35465 L
.36404 .34946 L
.36888 .34392 L
.37232 .33974 L
.37859 .33154 L
.37968 .33003 L
.38626 .32032 L
.38831 .31708 L
.39219 .31061 L
.39756 .30089 L
.39802 .30001 L
.40244 .29118 L
.4069 .28147 L
.40773 .27955 L
.41099 .27176 L
.41476 .26204 L
.41744 .25462 L
.41824 .25233 L
.42147 .24262 L
.42448 .23291 L
.42716 .22366 L
.42729 .22319 L
.42991 .21348 L
.43237 .20377 L
.43469 .19406 L
.43687 .18438 L
.43688 .18434 L
.43894 .17463 L
.44089 .16492 L
.44275 .15521 L
.44451 .14549 L
.44619 .13578 L
.44658 .13348 L
.44779 .12607 L
.44932 .11636 L
.45079 .10664 L
.4522 .09693 L
.45356 .08722 L
.45487 .07751 L
.45613 .06779 L
.45629 .06655 L
.45734 .05808 L
.45852 .04837 L
.45966 .03866 L
.46077 .02894 L
.46185 .01923 L
.01923 .01923 L
F
0 g
.01923 .46185 m
.02894 .46077 L
.03866 .45966 L
.04837 .45852 L
.05808 .45734 L
.06655 .45629 L
.06779 .45613 L
.07751 .45487 L
.08722 .45356 L
.09693 .4522 L
.10664 .45079 L
.11636 .44932 L
.12607 .44779 L
.13348 .44658 L
.13578 .44619 L
.14549 .44451 L
.15521 .44275 L
.16492 .44089 L
.17463 .43894 L
.18434 .43688 L
.18438 .43687 L
.19406 .43469 L
.20377 .43237 L
.21348 .42991 L
.22319 .42729 L
.22366 .42716 L
.23291 .42448 L
.24262 .42147 L
.25233 .41824 L
.25462 .41744 L
.26204 .41476 L
.27176 .41099 L
.27955 .40773 L
.28147 .4069 L
.29118 .40244 L
.30001 .39802 L
.30089 .39756 L
.31061 .39219 L
.31708 .38831 L
.32032 .38626 L
.33003 .37968 L
.33154 .37859 L
.33974 .37232 L
.34392 .36888 L
.34946 .36404 L
.35465 .35917 L
.35917 .35465 L
.36404 .34946 L
.36888 .34392 L
.37232 .33974 L
Mistroke
.37859 .33154 L
.37968 .33003 L
.38626 .32032 L
.38831 .31708 L
.39219 .31061 L
.39756 .30089 L
.39802 .30001 L
.40244 .29118 L
.4069 .28147 L
.40773 .27955 L
.41099 .27176 L
.41476 .26204 L
.41744 .25462 L
.41824 .25233 L
.42147 .24262 L
.42448 .23291 L
.42716 .22366 L
.42729 .22319 L
.42991 .21348 L
.43237 .20377 L
.43469 .19406 L
.43687 .18438 L
.43688 .18434 L
.43894 .17463 L
.44089 .16492 L
.44275 .15521 L
.44451 .14549 L
.44619 .13578 L
.44658 .13348 L
.44779 .12607 L
.44932 .11636 L
.45079 .10664 L
.4522 .09693 L
.45356 .08722 L
.45487 .07751 L
.45613 .06779 L
.45629 .06655 L
.45734 .05808 L
.45852 .04837 L
.45966 .03866 L
.46077 .02894 L
.46185 .01923 L
Mfstroke
0 1 .88 r
.01923 .53815 m
.02894 .53923 L
.03866 .54034 L
.04837 .54148 L
.05808 .54266 L
.06655 .54371 L
.06779 .54387 L
.07751 .54513 L
.08722 .54644 L
.09693 .5478 L
.10664 .54921 L
.11636 .55068 L
.12607 .55221 L
.13348 .55342 L
.13578 .55381 L
.14549 .55549 L
.15521 .55725 L
.16492 .55911 L
.17463 .56106 L
.18434 .56312 L
.18438 .56313 L
.19406 .56531 L
.20377 .56763 L
.21348 .57009 L
.22319 .57271 L
.22366 .57284 L
.23291 .57552 L
.24262 .57853 L
.25233 .58176 L
.25462 .58256 L
.26204 .58524 L
.27176 .58901 L
.27955 .59227 L
.28147 .5931 L
.29118 .59756 L
.30001 .60198 L
.30089 .60244 L
.31061 .60781 L
.31708 .61169 L
.32032 .61374 L
.33003 .62032 L
.33154 .62141 L
.33974 .62768 L
.34392 .63112 L
.34946 .63596 L
.35465 .64083 L
.35917 .64535 L
.36404 .65054 L
.36888 .65608 L
.37232 .66026 L
.37859 .66846 L
.37968 .66997 L
.38626 .67968 L
.38831 .68292 L
.39219 .68939 L
.39756 .69911 L
.39802 .69999 L
.40244 .70882 L
.4069 .71853 L
.40773 .72045 L
.41099 .72824 L
.41476 .73796 L
.41744 .74538 L
.41824 .74767 L
.42147 .75738 L
.42448 .76709 L
.42716 .77634 L
.42729 .77681 L
.42991 .78652 L
.43237 .79623 L
.43469 .80594 L
.43687 .81562 L
.43688 .81566 L
.43894 .82537 L
.44089 .83508 L
.44275 .84479 L
.44451 .85451 L
.44619 .86422 L
.44658 .86652 L
.44779 .87393 L
.44932 .88364 L
.45079 .89336 L
.4522 .90307 L
.45356 .91278 L
.45487 .92249 L
.45613 .93221 L
.45629 .93345 L
.45734 .94192 L
.45852 .95163 L
.45966 .96134 L
.46077 .97106 L
.46185 .98077 L
.01923 .98077 L
F
0 g
.01923 .53815 m
.02894 .53923 L
.03866 .54034 L
.04837 .54148 L
.05808 .54266 L
.06655 .54371 L
.06779 .54387 L
.07751 .54513 L
.08722 .54644 L
.09693 .5478 L
.10664 .54921 L
.11636 .55068 L
.12607 .55221 L
.13348 .55342 L
.13578 .55381 L
.14549 .55549 L
.15521 .55725 L
.16492 .55911 L
.17463 .56106 L
.18434 .56312 L
.18438 .56313 L
.19406 .56531 L
.20377 .56763 L
.21348 .57009 L
.22319 .57271 L
.22366 .57284 L
.23291 .57552 L
.24262 .57853 L
.25233 .58176 L
.25462 .58256 L
.26204 .58524 L
.27176 .58901 L
.27955 .59227 L
.28147 .5931 L
.29118 .59756 L
.30001 .60198 L
.30089 .60244 L
.31061 .60781 L
.31708 .61169 L
.32032 .61374 L
.33003 .62032 L
.33154 .62141 L
.33974 .62768 L
.34392 .63112 L
.34946 .63596 L
.35465 .64083 L
.35917 .64535 L
.36404 .65054 L
.36888 .65608 L
.37232 .66026 L
Mistroke
.37859 .66846 L
.37968 .66997 L
.38626 .67968 L
.38831 .68292 L
.39219 .68939 L
.39756 .69911 L
.39802 .69999 L
.40244 .70882 L
.4069 .71853 L
.40773 .72045 L
.41099 .72824 L
.41476 .73796 L
.41744 .74538 L
.41824 .74767 L
.42147 .75738 L
.42448 .76709 L
.42716 .77634 L
.42729 .77681 L
.42991 .78652 L
.43237 .79623 L
.43469 .80594 L
.43687 .81562 L
.43688 .81566 L
.43894 .82537 L
.44089 .83508 L
.44275 .84479 L
.44451 .85451 L
.44619 .86422 L
.44658 .86652 L
.44779 .87393 L
.44932 .88364 L
.45079 .89336 L
.4522 .90307 L
.45356 .91278 L
.45487 .92249 L
.45613 .93221 L
.45629 .93345 L
.45734 .94192 L
.45852 .95163 L
.45966 .96134 L
.46077 .97106 L
.46185 .98077 L
Mfstroke
0 1 .64 r
.01923 .45705 m
.02648 .45629 L
.02894 .45603 L
.03866 .45497 L
.04837 .45388 L
.05808 .45274 L
.06779 .45155 L
.07751 .45032 L
.08722 .44904 L
.09693 .44771 L
.1048 .44658 L
.10664 .44631 L
.11636 .44485 L
.12607 .44331 L
.13578 .4417 L
.14549 .44001 L
.15521 .43823 L
.16227 .43687 L
.16492 .43635 L
.17463 .43436 L
.18434 .43225 L
.19406 .43002 L
.20377 .42764 L
.20567 .42716 L
.21348 .42511 L
.22319 .4224 L
.23291 .4195 L
.23942 .41744 L
.24262 .41639 L
.25233 .41305 L
.26204 .40943 L
.26635 .40773 L
.27176 .40551 L
.28147 .40125 L
.2883 .39802 L
.29118 .39659 L
.30089 .3915 L
.30653 .38831 L
.31061 .38588 L
.32032 .37967 L
.3219 .37859 L
.33003 .37276 L
.33503 .36888 L
.33974 .36502 L
.34639 .35917 L
.34946 .35629 L
.35629 .34946 L
.35917 .34639 L
.36502 .33974 L
.36888 .33503 L
.37276 .33003 L
.37859 .3219 L
.37967 .32032 L
.38588 .31061 L
.38831 .30653 L
.3915 .30089 L
.39659 .29118 L
.39802 .2883 L
.40125 .28147 L
.40551 .27176 L
.40773 .26635 L
.40943 .26204 L
.41305 .25233 L
.41639 .24262 L
.41744 .23942 L
.4195 .23291 L
.4224 .22319 L
.42511 .21348 L
.42716 .20567 L
.42764 .20377 L
.43002 .19406 L
.43225 .18434 L
.43436 .17463 L
.43635 .16492 L
.43687 .16227 L
.43823 .15521 L
.44001 .14549 L
.4417 .13578 L
.44331 .12607 L
.44485 .11636 L
.44631 .10664 L
.44658 .1048 L
.44771 .09693 L
.44904 .08722 L
.45032 .07751 L
.45155 .06779 L
.45274 .05808 L
.45388 .04837 L
.45497 .03866 L
.45603 .02894 L
.45629 .02648 L
.45705 .01923 L
.01923 .01923 L
F
0 g
.01923 .45705 m
.02648 .45629 L
.02894 .45603 L
.03866 .45497 L
.04837 .45388 L
.05808 .45274 L
.06779 .45155 L
.07751 .45032 L
.08722 .44904 L
.09693 .44771 L
.1048 .44658 L
.10664 .44631 L
.11636 .44485 L
.12607 .44331 L
.13578 .4417 L
.14549 .44001 L
.15521 .43823 L
.16227 .43687 L
.16492 .43635 L
.17463 .43436 L
.18434 .43225 L
.19406 .43002 L
.20377 .42764 L
.20567 .42716 L
.21348 .42511 L
.22319 .4224 L
.23291 .4195 L
.23942 .41744 L
.24262 .41639 L
.25233 .41305 L
.26204 .40943 L
.26635 .40773 L
.27176 .40551 L
.28147 .40125 L
.2883 .39802 L
.29118 .39659 L
.30089 .3915 L
.30653 .38831 L
.31061 .38588 L
.32032 .37967 L
.3219 .37859 L
.33003 .37276 L
.33503 .36888 L
.33974 .36502 L
.34639 .35917 L
.34946 .35629 L
.35629 .34946 L
.35917 .34639 L
.36502 .33974 L
.36888 .33503 L
Mistroke
.37276 .33003 L
.37859 .3219 L
.37967 .32032 L
.38588 .31061 L
.38831 .30653 L
.3915 .30089 L
.39659 .29118 L
.39802 .2883 L
.40125 .28147 L
.40551 .27176 L
.40773 .26635 L
.40943 .26204 L
.41305 .25233 L
.41639 .24262 L
.41744 .23942 L
.4195 .23291 L
.4224 .22319 L
.42511 .21348 L
.42716 .20567 L
.42764 .20377 L
.43002 .19406 L
.43225 .18434 L
.43436 .17463 L
.43635 .16492 L
.43687 .16227 L
.43823 .15521 L
.44001 .14549 L
.4417 .13578 L
.44331 .12607 L
.44485 .11636 L
.44631 .10664 L
.44658 .1048 L
.44771 .09693 L
.44904 .08722 L
.45032 .07751 L
.45155 .06779 L
.45274 .05808 L
.45388 .04837 L
.45497 .03866 L
.45603 .02894 L
.45629 .02648 L
.45705 .01923 L
Mfstroke
0 1 .64 r
.01923 .54295 m
.02648 .54371 L
.02894 .54397 L
.03866 .54503 L
.04837 .54612 L
.05808 .54726 L
.06779 .54845 L
.07751 .54968 L
.08722 .55096 L
.09693 .55229 L
.1048 .55342 L
.10664 .55369 L
.11636 .55515 L
.12607 .55669 L
.13578 .5583 L
.14549 .55999 L
.15521 .56177 L
.16227 .56313 L
.16492 .56365 L
.17463 .56564 L
.18434 .56775 L
.19406 .56998 L
.20377 .57236 L
.20567 .57284 L
.21348 .57489 L
.22319 .5776 L
.23291 .5805 L
.23942 .58256 L
.24262 .58361 L
.25233 .58695 L
.26204 .59057 L
.26635 .59227 L
.27176 .59449 L
.28147 .59875 L
.2883 .60198 L
.29118 .60341 L
.30089 .6085 L
.30653 .61169 L
.31061 .61412 L
.32032 .62033 L
.3219 .62141 L
.33003 .62724 L
.33503 .63112 L
.33974 .63498 L
.34639 .64083 L
.34946 .64371 L
.35629 .65054 L
.35917 .65361 L
.36502 .66026 L
.36888 .66497 L
.37276 .66997 L
.37859 .6781 L
.37967 .67968 L
.38588 .68939 L
.38831 .69347 L
.3915 .69911 L
.39659 .70882 L
.39802 .7117 L
.40125 .71853 L
.40551 .72824 L
.40773 .73365 L
.40943 .73796 L
.41305 .74767 L
.41639 .75738 L
.41744 .76058 L
.4195 .76709 L
.4224 .77681 L
.42511 .78652 L
.42716 .79433 L
.42764 .79623 L
.43002 .80594 L
.43225 .81566 L
.43436 .82537 L
.43635 .83508 L
.43687 .83773 L
.43823 .84479 L
.44001 .85451 L
.4417 .86422 L
.44331 .87393 L
.44485 .88364 L
.44631 .89336 L
.44658 .8952 L
.44771 .90307 L
.44904 .91278 L
.45032 .92249 L
.45155 .93221 L
.45274 .94192 L
.45388 .95163 L
.45497 .96134 L
.45603 .97106 L
.45629 .97352 L
.45705 .98077 L
.01923 .98077 L
F
0 g
.01923 .54295 m
.02648 .54371 L
.02894 .54397 L
.03866 .54503 L
.04837 .54612 L
.05808 .54726 L
.06779 .54845 L
.07751 .54968 L
.08722 .55096 L
.09693 .55229 L
.1048 .55342 L
.10664 .55369 L
.11636 .55515 L
.12607 .55669 L
.13578 .5583 L
.14549 .55999 L
.15521 .56177 L
.16227 .56313 L
.16492 .56365 L
.17463 .56564 L
.18434 .56775 L
.19406 .56998 L
.20377 .57236 L
.20567 .57284 L
.21348 .57489 L
.22319 .5776 L
.23291 .5805 L
.23942 .58256 L
.24262 .58361 L
.25233 .58695 L
.26204 .59057 L
.26635 .59227 L
.27176 .59449 L
.28147 .59875 L
.2883 .60198 L
.29118 .60341 L
.30089 .6085 L
.30653 .61169 L
.31061 .61412 L
.32032 .62033 L
.3219 .62141 L
.33003 .62724 L
.33503 .63112 L
.33974 .63498 L
.34639 .64083 L
.34946 .64371 L
.35629 .65054 L
.35917 .65361 L
.36502 .66026 L
.36888 .66497 L
Mistroke
.37276 .66997 L
.37859 .6781 L
.37967 .67968 L
.38588 .68939 L
.38831 .69347 L
.3915 .69911 L
.39659 .70882 L
.39802 .7117 L
.40125 .71853 L
.40551 .72824 L
.40773 .73365 L
.40943 .73796 L
.41305 .74767 L
.41639 .75738 L
.41744 .76058 L
.4195 .76709 L
.4224 .77681 L
.42511 .78652 L
.42716 .79433 L
.42764 .79623 L
.43002 .80594 L
.43225 .81566 L
.43436 .82537 L
.43635 .83508 L
.43687 .83773 L
.43823 .84479 L
.44001 .85451 L
.4417 .86422 L
.44331 .87393 L
.44485 .88364 L
.44631 .89336 L
.44658 .8952 L
.44771 .90307 L
.44904 .91278 L
.45032 .92249 L
.45155 .93221 L
.45274 .94192 L
.45388 .95163 L
.45497 .96134 L
.45603 .97106 L
.45629 .97352 L
.45705 .98077 L
Mfstroke
0 1 .4 r
.01923 .45233 m
.02894 .45136 L
.03866 .45035 L
.04837 .44929 L
.05808 .44819 L
.06779 .44704 L
.07158 .44658 L
.07751 .44584 L
.08722 .44458 L
.09693 .44326 L
.10664 .44187 L
.11636 .44041 L
.12607 .43887 L
.13578 .43725 L
.13799 .43687 L
.14549 .43554 L
.15521 .43374 L
.16492 .43183 L
.17463 .4298 L
.18434 .42765 L
.18649 .42716 L
.19406 .42536 L
.20377 .42292 L
.21348 .42032 L
.22319 .41753 L
.22349 .41744 L
.23291 .41454 L
.24262 .41133 L
.25233 .40785 L
.25266 .40773 L
.26204 .4041 L
.27176 .40003 L
.27626 .39802 L
.28147 .39559 L
.29118 .39074 L
.29574 .38831 L
.30089 .38542 L
.31061 .37955 L
.31209 .37859 L
.32032 .37304 L
.32602 .36888 L
.33003 .36579 L
.33802 .35917 L
.33974 .35766 L
.34848 .34946 L
.34946 .34848 L
.35766 .33974 L
.35917 .33802 L
.36579 .33003 L
.36888 .32602 L
.37304 .32032 L
.37859 .31209 L
.37955 .31061 L
.38542 .30089 L
.38831 .29574 L
.39074 .29118 L
.39559 .28147 L
.39802 .27626 L
.40003 .27176 L
.4041 .26204 L
.40773 .25266 L
.40785 .25233 L
.41133 .24262 L
.41454 .23291 L
.41744 .22349 L
.41753 .22319 L
.42032 .21348 L
.42292 .20377 L
.42536 .19406 L
.42716 .18649 L
.42765 .18434 L
.4298 .17463 L
.43183 .16492 L
.43374 .15521 L
.43554 .14549 L
.43687 .13799 L
.43725 .13578 L
.43887 .12607 L
.44041 .11636 L
.44187 .10664 L
.44326 .09693 L
.44458 .08722 L
.44584 .07751 L
.44658 .07158 L
.44704 .06779 L
.44819 .05808 L
.44929 .04837 L
.45035 .03866 L
.45136 .02894 L
.45233 .01923 L
.01923 .01923 L
F
0 g
.01923 .45233 m
.02894 .45136 L
.03866 .45035 L
.04837 .44929 L
.05808 .44819 L
.06779 .44704 L
.07158 .44658 L
.07751 .44584 L
.08722 .44458 L
.09693 .44326 L
.10664 .44187 L
.11636 .44041 L
.12607 .43887 L
.13578 .43725 L
.13799 .43687 L
.14549 .43554 L
.15521 .43374 L
.16492 .43183 L
.17463 .4298 L
.18434 .42765 L
.18649 .42716 L
.19406 .42536 L
.20377 .42292 L
.21348 .42032 L
.22319 .41753 L
.22349 .41744 L
.23291 .41454 L
.24262 .41133 L
.25233 .40785 L
.25266 .40773 L
.26204 .4041 L
.27176 .40003 L
.27626 .39802 L
.28147 .39559 L
.29118 .39074 L
.29574 .38831 L
.30089 .38542 L
.31061 .37955 L
.31209 .37859 L
.32032 .37304 L
.32602 .36888 L
.33003 .36579 L
.33802 .35917 L
.33974 .35766 L
.34848 .34946 L
.34946 .34848 L
.35766 .33974 L
.35917 .33802 L
.36579 .33003 L
.36888 .32602 L
Mistroke
.37304 .32032 L
.37859 .31209 L
.37955 .31061 L
.38542 .30089 L
.38831 .29574 L
.39074 .29118 L
.39559 .28147 L
.39802 .27626 L
.40003 .27176 L
.4041 .26204 L
.40773 .25266 L
.40785 .25233 L
.41133 .24262 L
.41454 .23291 L
.41744 .22349 L
.41753 .22319 L
.42032 .21348 L
.42292 .20377 L
.42536 .19406 L
.42716 .18649 L
.42765 .18434 L
.4298 .17463 L
.43183 .16492 L
.43374 .15521 L
.43554 .14549 L
.43687 .13799 L
.43725 .13578 L
.43887 .12607 L
.44041 .11636 L
.44187 .10664 L
.44326 .09693 L
.44458 .08722 L
.44584 .07751 L
.44658 .07158 L
.44704 .06779 L
.44819 .05808 L
.44929 .04837 L
.45035 .03866 L
.45136 .02894 L
.45233 .01923 L
Mfstroke
0 1 .4 r
.01923 .54767 m
.02894 .54864 L
.03866 .54965 L
.04837 .55071 L
.05808 .55181 L
.06779 .55296 L
.07158 .55342 L
.07751 .55416 L
.08722 .55542 L
.09693 .55674 L
.10664 .55813 L
.11636 .55959 L
.12607 .56113 L
.13578 .56275 L
.13799 .56313 L
.14549 .56446 L
.15521 .56626 L
.16492 .56817 L
.17463 .5702 L
.18434 .57235 L
.18649 .57284 L
.19406 .57464 L
.20377 .57708 L
.21348 .57968 L
.22319 .58247 L
.22349 .58256 L
.23291 .58546 L
.24262 .58867 L
.25233 .59215 L
.25266 .59227 L
.26204 .5959 L
.27176 .59997 L
.27626 .60198 L
.28147 .60441 L
.29118 .60926 L
.29574 .61169 L
.30089 .61458 L
.31061 .62045 L
.31209 .62141 L
.32032 .62696 L
.32602 .63112 L
.33003 .63421 L
.33802 .64083 L
.33974 .64234 L
.34848 .65054 L
.34946 .65152 L
.35766 .66026 L
.35917 .66198 L
.36579 .66997 L
.36888 .67398 L
.37304 .67968 L
.37859 .68791 L
.37955 .68939 L
.38542 .69911 L
.38831 .70426 L
.39074 .70882 L
.39559 .71853 L
.39802 .72374 L
.40003 .72824 L
.4041 .73796 L
.40773 .74734 L
.40785 .74767 L
.41133 .75738 L
.41454 .76709 L
.41744 .77651 L
.41753 .77681 L
.42032 .78652 L
.42292 .79623 L
.42536 .80594 L
.42716 .81351 L
.42765 .81566 L
.4298 .82537 L
.43183 .83508 L
.43374 .84479 L
.43554 .85451 L
.43687 .86201 L
.43725 .86422 L
.43887 .87393 L
.44041 .88364 L
.44187 .89336 L
.44326 .90307 L
.44458 .91278 L
.44584 .92249 L
.44658 .92842 L
.44704 .93221 L
.44819 .94192 L
.44929 .95163 L
.45035 .96134 L
.45136 .97106 L
.45233 .98077 L
.01923 .98077 L
F
0 g
.01923 .54767 m
.02894 .54864 L
.03866 .54965 L
.04837 .55071 L
.05808 .55181 L
.06779 .55296 L
.07158 .55342 L
.07751 .55416 L
.08722 .55542 L
.09693 .55674 L
.10664 .55813 L
.11636 .55959 L
.12607 .56113 L
.13578 .56275 L
.13799 .56313 L
.14549 .56446 L
.15521 .56626 L
.16492 .56817 L
.17463 .5702 L
.18434 .57235 L
.18649 .57284 L
.19406 .57464 L
.20377 .57708 L
.21348 .57968 L
.22319 .58247 L
.22349 .58256 L
.23291 .58546 L
.24262 .58867 L
.25233 .59215 L
.25266 .59227 L
.26204 .5959 L
.27176 .59997 L
.27626 .60198 L
.28147 .60441 L
.29118 .60926 L
.29574 .61169 L
.30089 .61458 L
.31061 .62045 L
.31209 .62141 L
.32032 .62696 L
.32602 .63112 L
.33003 .63421 L
.33802 .64083 L
.33974 .64234 L
.34848 .65054 L
.34946 .65152 L
.35766 .66026 L
.35917 .66198 L
.36579 .66997 L
.36888 .67398 L
Mistroke
.37304 .67968 L
.37859 .68791 L
.37955 .68939 L
.38542 .69911 L
.38831 .70426 L
.39074 .70882 L
.39559 .71853 L
.39802 .72374 L
.40003 .72824 L
.4041 .73796 L
.40773 .74734 L
.40785 .74767 L
.41133 .75738 L
.41454 .76709 L
.41744 .77651 L
.41753 .77681 L
.42032 .78652 L
.42292 .79623 L
.42536 .80594 L
.42716 .81351 L
.42765 .81566 L
.4298 .82537 L
.43183 .83508 L
.43374 .84479 L
.43554 .85451 L
.43687 .86201 L
.43725 .86422 L
.43887 .87393 L
.44041 .88364 L
.44187 .89336 L
.44326 .90307 L
.44458 .91278 L
.44584 .92249 L
.44658 .92842 L
.44704 .93221 L
.44819 .94192 L
.44929 .95163 L
.45035 .96134 L
.45136 .97106 L
.45233 .98077 L
Mfstroke
0 1 .16 r
.01923 .44757 m
.02894 .44666 L
.02975 .44658 L
.03866 .44571 L
.04837 .4447 L
.05808 .44364 L
.06779 .44252 L
.07751 .44134 L
.08722 .44009 L
.09693 .43878 L
.10664 .4374 L
.11023 .43687 L
.11636 .43594 L
.12607 .43441 L
.13578 .43278 L
.14549 .43105 L
.15521 .42922 L
.16492 .42728 L
.16551 .42716 L
.17463 .42522 L
.18434 .42302 L
.19406 .42068 L
.20377 .41818 L
.2065 .41744 L
.21348 .4155 L
.22319 .41263 L
.23291 .40954 L
.23829 .40773 L
.24262 .40622 L
.25233 .40263 L
.26204 .39873 L
.26374 .39802 L
.27176 .3945 L
.28147 .38989 L
.2846 .38831 L
.29118 .38483 L
.30089 .37928 L
.30202 .37859 L
.31061 .37314 L
.3168 .36888 L
.32032 .36633 L
.3295 .35917 L
.33003 .35873 L
.33974 .35019 L
.34052 .34946 L
.34946 .34052 L
.35019 .33974 L
.35873 .33003 L
.35917 .3295 L
.36633 .32032 L
.36888 .3168 L
.37314 .31061 L
.37859 .30202 L
.37928 .30089 L
.38483 .29118 L
.38831 .2846 L
.38989 .28147 L
.3945 .27176 L
.39802 .26374 L
.39873 .26204 L
.40263 .25233 L
.40622 .24262 L
.40773 .23829 L
.40954 .23291 L
.41263 .22319 L
.4155 .21348 L
.41744 .2065 L
.41818 .20377 L
.42068 .19406 L
.42302 .18434 L
.42522 .17463 L
.42716 .16551 L
.42728 .16492 L
.42922 .15521 L
.43105 .14549 L
.43278 .13578 L
.43441 .12607 L
.43594 .11636 L
.43687 .11023 L
.4374 .10664 L
.43878 .09693 L
.44009 .08722 L
.44134 .07751 L
.44252 .06779 L
.44364 .05808 L
.4447 .04837 L
.44571 .03866 L
.44658 .02975 L
.44666 .02894 L
.44757 .01923 L
.01923 .01923 L
F
0 g
.01923 .44757 m
.02894 .44666 L
.02975 .44658 L
.03866 .44571 L
.04837 .4447 L
.05808 .44364 L
.06779 .44252 L
.07751 .44134 L
.08722 .44009 L
.09693 .43878 L
.10664 .4374 L
.11023 .43687 L
.11636 .43594 L
.12607 .43441 L
.13578 .43278 L
.14549 .43105 L
.15521 .42922 L
.16492 .42728 L
.16551 .42716 L
.17463 .42522 L
.18434 .42302 L
.19406 .42068 L
.20377 .41818 L
.2065 .41744 L
.21348 .4155 L
.22319 .41263 L
.23291 .40954 L
.23829 .40773 L
.24262 .40622 L
.25233 .40263 L
.26204 .39873 L
.26374 .39802 L
.27176 .3945 L
.28147 .38989 L
.2846 .38831 L
.29118 .38483 L
.30089 .37928 L
.30202 .37859 L
.31061 .37314 L
.3168 .36888 L
.32032 .36633 L
.3295 .35917 L
.33003 .35873 L
.33974 .35019 L
.34052 .34946 L
.34946 .34052 L
.35019 .33974 L
.35873 .33003 L
.35917 .3295 L
.36633 .32032 L
Mistroke
.36888 .3168 L
.37314 .31061 L
.37859 .30202 L
.37928 .30089 L
.38483 .29118 L
.38831 .2846 L
.38989 .28147 L
.3945 .27176 L
.39802 .26374 L
.39873 .26204 L
.40263 .25233 L
.40622 .24262 L
.40773 .23829 L
.40954 .23291 L
.41263 .22319 L
.4155 .21348 L
.41744 .2065 L
.41818 .20377 L
.42068 .19406 L
.42302 .18434 L
.42522 .17463 L
.42716 .16551 L
.42728 .16492 L
.42922 .15521 L
.43105 .14549 L
.43278 .13578 L
.43441 .12607 L
.43594 .11636 L
.43687 .11023 L
.4374 .10664 L
.43878 .09693 L
.44009 .08722 L
.44134 .07751 L
.44252 .06779 L
.44364 .05808 L
.4447 .04837 L
.44571 .03866 L
.44658 .02975 L
.44666 .02894 L
.44757 .01923 L
Mfstroke
0 1 .16 r
.01923 .55243 m
.02894 .55334 L
.02975 .55342 L
.03866 .55429 L
.04837 .5553 L
.05808 .55636 L
.06779 .55748 L
.07751 .55866 L
.08722 .55991 L
.09693 .56122 L
.10664 .5626 L
.11023 .56313 L
.11636 .56406 L
.12607 .56559 L
.13578 .56722 L
.14549 .56895 L
.15521 .57078 L
.16492 .57272 L
.16551 .57284 L
.17463 .57478 L
.18434 .57698 L
.19406 .57932 L
.20377 .58182 L
.2065 .58256 L
.21348 .5845 L
.22319 .58737 L
.23291 .59046 L
.23829 .59227 L
.24262 .59378 L
.25233 .59737 L
.26204 .60127 L
.26374 .60198 L
.27176 .6055 L
.28147 .61011 L
.2846 .61169 L
.29118 .61517 L
.30089 .62072 L
.30202 .62141 L
.31061 .62686 L
.3168 .63112 L
.32032 .63367 L
.3295 .64083 L
.33003 .64127 L
.33974 .64981 L
.34052 .65054 L
.34946 .65948 L
.35019 .66026 L
.35873 .66997 L
.35917 .6705 L
.36633 .67968 L
.36888 .6832 L
.37314 .68939 L
.37859 .69798 L
.37928 .69911 L
.38483 .70882 L
.38831 .7154 L
.38989 .71853 L
.3945 .72824 L
.39802 .73626 L
.39873 .73796 L
.40263 .74767 L
.40622 .75738 L
.40773 .76171 L
.40954 .76709 L
.41263 .77681 L
.4155 .78652 L
.41744 .7935 L
.41818 .79623 L
.42068 .80594 L
.42302 .81566 L
.42522 .82537 L
.42716 .83449 L
.42728 .83508 L
.42922 .84479 L
.43105 .85451 L
.43278 .86422 L
.43441 .87393 L
.43594 .88364 L
.43687 .88977 L
.4374 .89336 L
.43878 .90307 L
.44009 .91278 L
.44134 .92249 L
.44252 .93221 L
.44364 .94192 L
.4447 .95163 L
.44571 .96134 L
.44658 .97025 L
.44666 .97106 L
.44757 .98077 L
.01923 .98077 L
F
0 g
.01923 .55243 m
.02894 .55334 L
.02975 .55342 L
.03866 .55429 L
.04837 .5553 L
.05808 .55636 L
.06779 .55748 L
.07751 .55866 L
.08722 .55991 L
.09693 .56122 L
.10664 .5626 L
.11023 .56313 L
.11636 .56406 L
.12607 .56559 L
.13578 .56722 L
.14549 .56895 L
.15521 .57078 L
.16492 .57272 L
.16551 .57284 L
.17463 .57478 L
.18434 .57698 L
.19406 .57932 L
.20377 .58182 L
.2065 .58256 L
.21348 .5845 L
.22319 .58737 L
.23291 .59046 L
.23829 .59227 L
.24262 .59378 L
.25233 .59737 L
.26204 .60127 L
.26374 .60198 L
.27176 .6055 L
.28147 .61011 L
.2846 .61169 L
.29118 .61517 L
.30089 .62072 L
.30202 .62141 L
.31061 .62686 L
.3168 .63112 L
.32032 .63367 L
.3295 .64083 L
.33003 .64127 L
.33974 .64981 L
.34052 .65054 L
.34946 .65948 L
.35019 .66026 L
.35873 .66997 L
.35917 .6705 L
.36633 .67968 L
Mistroke
.36888 .6832 L
.37314 .68939 L
.37859 .69798 L
.37928 .69911 L
.38483 .70882 L
.38831 .7154 L
.38989 .71853 L
.3945 .72824 L
.39802 .73626 L
.39873 .73796 L
.40263 .74767 L
.40622 .75738 L
.40773 .76171 L
.40954 .76709 L
.41263 .77681 L
.4155 .78652 L
.41744 .7935 L
.41818 .79623 L
.42068 .80594 L
.42302 .81566 L
.42522 .82537 L
.42716 .83449 L
.42728 .83508 L
.42922 .84479 L
.43105 .85451 L
.43278 .86422 L
.43441 .87393 L
.43594 .88364 L
.43687 .88977 L
.4374 .89336 L
.43878 .90307 L
.44009 .91278 L
.44134 .92249 L
.44252 .93221 L
.44364 .94192 L
.4447 .95163 L
.44571 .96134 L
.44658 .97025 L
.44666 .97106 L
.44757 .98077 L
Mfstroke
.08 1 0 r
.01923 .4427 m
.02894 .44185 L
.03866 .44094 L
.04837 .43997 L
.05808 .43895 L
.06779 .43786 L
.07618 .43687 L
.07751 .43671 L
.08722 .4355 L
.09693 .43422 L
.10664 .43285 L
.11636 .4314 L
.12607 .42985 L
.13578 .42821 L
.14173 .42716 L
.14549 .42647 L
.15521 .42463 L
.16492 .42266 L
.17463 .42056 L
.18434 .41831 L
.18795 .41744 L
.19406 .41592 L
.20377 .41336 L
.21348 .41061 L
.22293 .40773 L
.22319 .40765 L
.23291 .40447 L
.24262 .40103 L
.25054 .39802 L
.25233 .39731 L
.26204 .39328 L
.27176 .38888 L
.27296 .38831 L
.28147 .38408 L
.29118 .37882 L
.29157 .37859 L
.30089 .37302 L
.30728 .36888 L
.31061 .36661 L
.32032 .35948 L
.32072 .35917 L
.33003 .35151 L
.33236 .34946 L
.33974 .34253 L
.34253 .33974 L
.34946 .33236 L
.35151 .33003 L
.35917 .32072 L
.35948 .32032 L
.36661 .31061 L
.36888 .30728 L
.37302 .30089 L
.37859 .29157 L
.37882 .29118 L
.38408 .28147 L
.38831 .27296 L
.38888 .27176 L
.39328 .26204 L
.39731 .25233 L
.39802 .25054 L
.40103 .24262 L
.40447 .23291 L
.40765 .22319 L
.40773 .22293 L
.41061 .21348 L
.41336 .20377 L
.41592 .19406 L
.41744 .18795 L
.41831 .18434 L
.42056 .17463 L
.42266 .16492 L
.42463 .15521 L
.42647 .14549 L
.42716 .14173 L
.42821 .13578 L
.42985 .12607 L
.4314 .11636 L
.43285 .10664 L
.43422 .09693 L
.4355 .08722 L
.43671 .07751 L
.43687 .07618 L
.43786 .06779 L
.43895 .05808 L
.43997 .04837 L
.44094 .03866 L
.44185 .02894 L
.4427 .01923 L
.01923 .01923 L
F
0 g
.01923 .4427 m
.02894 .44185 L
.03866 .44094 L
.04837 .43997 L
.05808 .43895 L
.06779 .43786 L
.07618 .43687 L
.07751 .43671 L
.08722 .4355 L
.09693 .43422 L
.10664 .43285 L
.11636 .4314 L
.12607 .42985 L
.13578 .42821 L
.14173 .42716 L
.14549 .42647 L
.15521 .42463 L
.16492 .42266 L
.17463 .42056 L
.18434 .41831 L
.18795 .41744 L
.19406 .41592 L
.20377 .41336 L
.21348 .41061 L
.22293 .40773 L
.22319 .40765 L
.23291 .40447 L
.24262 .40103 L
.25054 .39802 L
.25233 .39731 L
.26204 .39328 L
.27176 .38888 L
.27296 .38831 L
.28147 .38408 L
.29118 .37882 L
.29157 .37859 L
.30089 .37302 L
.30728 .36888 L
.31061 .36661 L
.32032 .35948 L
.32072 .35917 L
.33003 .35151 L
.33236 .34946 L
.33974 .34253 L
.34253 .33974 L
.34946 .33236 L
.35151 .33003 L
.35917 .32072 L
.35948 .32032 L
.36661 .31061 L
Mistroke
.36888 .30728 L
.37302 .30089 L
.37859 .29157 L
.37882 .29118 L
.38408 .28147 L
.38831 .27296 L
.38888 .27176 L
.39328 .26204 L
.39731 .25233 L
.39802 .25054 L
.40103 .24262 L
.40447 .23291 L
.40765 .22319 L
.40773 .22293 L
.41061 .21348 L
.41336 .20377 L
.41592 .19406 L
.41744 .18795 L
.41831 .18434 L
.42056 .17463 L
.42266 .16492 L
.42463 .15521 L
.42647 .14549 L
.42716 .14173 L
.42821 .13578 L
.42985 .12607 L
.4314 .11636 L
.43285 .10664 L
.43422 .09693 L
.4355 .08722 L
.43671 .07751 L
.43687 .07618 L
.43786 .06779 L
.43895 .05808 L
.43997 .04837 L
.44094 .03866 L
.44185 .02894 L
.4427 .01923 L
Mfstroke
.08 1 0 r
.01923 .5573 m
.02894 .55815 L
.03866 .55906 L
.04837 .56003 L
.05808 .56105 L
.06779 .56214 L
.07618 .56313 L
.07751 .56329 L
.08722 .5645 L
.09693 .56578 L
.10664 .56715 L
.11636 .5686 L
.12607 .57015 L
.13578 .57179 L
.14173 .57284 L
.14549 .57353 L
.15521 .57537 L
.16492 .57734 L
.17463 .57944 L
.18434 .58169 L
.18795 .58256 L
.19406 .58408 L
.20377 .58664 L
.21348 .58939 L
.22293 .59227 L
.22319 .59235 L
.23291 .59553 L
.24262 .59897 L
.25054 .60198 L
.25233 .60269 L
.26204 .60672 L
.27176 .61112 L
.27296 .61169 L
.28147 .61592 L
.29118 .62118 L
.29157 .62141 L
.30089 .62698 L
.30728 .63112 L
.31061 .63339 L
.32032 .64052 L
.32072 .64083 L
.33003 .64849 L
.33236 .65054 L
.33974 .65747 L
.34253 .66026 L
.34946 .66764 L
.35151 .66997 L
.35917 .67928 L
.35948 .67968 L
.36661 .68939 L
.36888 .69272 L
.37302 .69911 L
.37859 .70843 L
.37882 .70882 L
.38408 .71853 L
.38831 .72704 L
.38888 .72824 L
.39328 .73796 L
.39731 .74767 L
.39802 .74946 L
.40103 .75738 L
.40447 .76709 L
.40765 .77681 L
.40773 .77707 L
.41061 .78652 L
.41336 .79623 L
.41592 .80594 L
.41744 .81205 L
.41831 .81566 L
.42056 .82537 L
.42266 .83508 L
.42463 .84479 L
.42647 .85451 L
.42716 .85827 L
.42821 .86422 L
.42985 .87393 L
.4314 .88364 L
.43285 .89336 L
.43422 .90307 L
.4355 .91278 L
.43671 .92249 L
.43687 .92382 L
.43786 .93221 L
.43895 .94192 L
.43997 .95163 L
.44094 .96134 L
.44185 .97106 L
.4427 .98077 L
.01923 .98077 L
F
0 g
.01923 .5573 m
.02894 .55815 L
.03866 .55906 L
.04837 .56003 L
.05808 .56105 L
.06779 .56214 L
.07618 .56313 L
.07751 .56329 L
.08722 .5645 L
.09693 .56578 L
.10664 .56715 L
.11636 .5686 L
.12607 .57015 L
.13578 .57179 L
.14173 .57284 L
.14549 .57353 L
.15521 .57537 L
.16492 .57734 L
.17463 .57944 L
.18434 .58169 L
.18795 .58256 L
.19406 .58408 L
.20377 .58664 L
.21348 .58939 L
.22293 .59227 L
.22319 .59235 L
.23291 .59553 L
.24262 .59897 L
.25054 .60198 L
.25233 .60269 L
.26204 .60672 L
.27176 .61112 L
.27296 .61169 L
.28147 .61592 L
.29118 .62118 L
.29157 .62141 L
.30089 .62698 L
.30728 .63112 L
.31061 .63339 L
.32032 .64052 L
.32072 .64083 L
.33003 .64849 L
.33236 .65054 L
.33974 .65747 L
.34253 .66026 L
.34946 .66764 L
.35151 .66997 L
.35917 .67928 L
.35948 .67968 L
.36661 .68939 L
Mistroke
.36888 .69272 L
.37302 .69911 L
.37859 .70843 L
.37882 .70882 L
.38408 .71853 L
.38831 .72704 L
.38888 .72824 L
.39328 .73796 L
.39731 .74767 L
.39802 .74946 L
.40103 .75738 L
.40447 .76709 L
.40765 .77681 L
.40773 .77707 L
.41061 .78652 L
.41336 .79623 L
.41592 .80594 L
.41744 .81205 L
.41831 .81566 L
.42056 .82537 L
.42266 .83508 L
.42463 .84479 L
.42647 .85451 L
.42716 .85827 L
.42821 .86422 L
.42985 .87393 L
.4314 .88364 L
.43285 .89336 L
.43422 .90307 L
.4355 .91278 L
.43671 .92249 L
.43687 .92382 L
.43786 .93221 L
.43895 .94192 L
.43997 .95163 L
.44094 .96134 L
.44185 .97106 L
.4427 .98077 L
Mfstroke
.32 1 0 r
.01923 .43743 m
.02643 .43687 L
.02894 .43668 L
.03866 .43588 L
.04837 .435 L
.05808 .43405 L
.06779 .43302 L
.07751 .43191 L
.08722 .43072 L
.09693 .42945 L
.10664 .4281 L
.11306 .42716 L
.11636 .42666 L
.12607 .42514 L
.13578 .42351 L
.14549 .42176 L
.15521 .41988 L
.16492 .41788 L
.16698 .41744 L
.17463 .41576 L
.18434 .41348 L
.19406 .41103 L
.20377 .4084 L
.20612 .40773 L
.21348 .40558 L
.22319 .40254 L
.23291 .39925 L
.23637 .39802 L
.24262 .39571 L
.25233 .39186 L
.26062 .38831 L
.26204 .38767 L
.27176 .38311 L
.28058 .37859 L
.28147 .37812 L
.29118 .37263 L
.29732 .36888 L
.30089 .36659 L
.31061 .35988 L
.31158 .35917 L
.32032 .35242 L
.32389 .34946 L
.33003 .34405 L
.33462 .33974 L
.33974 .33462 L
.34405 .33003 L
.34946 .32389 L
.35242 .32032 L
.35917 .31158 L
.35988 .31061 L
.36659 .30089 L
.36888 .29732 L
.37263 .29118 L
.37812 .28147 L
.37859 .28058 L
.38311 .27176 L
.38767 .26204 L
.38831 .26062 L
.39186 .25233 L
.39571 .24262 L
.39802 .23637 L
.39925 .23291 L
.40254 .22319 L
.40558 .21348 L
.40773 .20612 L
.4084 .20377 L
.41103 .19406 L
.41348 .18434 L
.41576 .17463 L
.41744 .16698 L
.41788 .16492 L
.41988 .15521 L
.42176 .14549 L
.42351 .13578 L
.42514 .12607 L
.42666 .11636 L
.42716 .11306 L
.4281 .10664 L
.42945 .09693 L
.43072 .08722 L
.43191 .07751 L
.43302 .06779 L
.43405 .05808 L
.435 .04837 L
.43588 .03866 L
.43668 .02894 L
.43687 .02643 L
.43743 .01923 L
.01923 .01923 L
F
0 g
.01923 .43743 m
.02643 .43687 L
.02894 .43668 L
.03866 .43588 L
.04837 .435 L
.05808 .43405 L
.06779 .43302 L
.07751 .43191 L
.08722 .43072 L
.09693 .42945 L
.10664 .4281 L
.11306 .42716 L
.11636 .42666 L
.12607 .42514 L
.13578 .42351 L
.14549 .42176 L
.15521 .41988 L
.16492 .41788 L
.16698 .41744 L
.17463 .41576 L
.18434 .41348 L
.19406 .41103 L
.20377 .4084 L
.20612 .40773 L
.21348 .40558 L
.22319 .40254 L
.23291 .39925 L
.23637 .39802 L
.24262 .39571 L
.25233 .39186 L
.26062 .38831 L
.26204 .38767 L
.27176 .38311 L
.28058 .37859 L
.28147 .37812 L
.29118 .37263 L
.29732 .36888 L
.30089 .36659 L
.31061 .35988 L
.31158 .35917 L
.32032 .35242 L
.32389 .34946 L
.33003 .34405 L
.33462 .33974 L
.33974 .33462 L
.34405 .33003 L
.34946 .32389 L
.35242 .32032 L
.35917 .31158 L
.35988 .31061 L
Mistroke
.36659 .30089 L
.36888 .29732 L
.37263 .29118 L
.37812 .28147 L
.37859 .28058 L
.38311 .27176 L
.38767 .26204 L
.38831 .26062 L
.39186 .25233 L
.39571 .24262 L
.39802 .23637 L
.39925 .23291 L
.40254 .22319 L
.40558 .21348 L
.40773 .20612 L
.4084 .20377 L
.41103 .19406 L
.41348 .18434 L
.41576 .17463 L
.41744 .16698 L
.41788 .16492 L
.41988 .15521 L
.42176 .14549 L
.42351 .13578 L
.42514 .12607 L
.42666 .11636 L
.42716 .11306 L
.4281 .10664 L
.42945 .09693 L
.43072 .08722 L
.43191 .07751 L
.43302 .06779 L
.43405 .05808 L
.435 .04837 L
.43588 .03866 L
.43668 .02894 L
.43687 .02643 L
.43743 .01923 L
Mfstroke
.32 1 0 r
.01923 .56257 m
.02643 .56313 L
.02894 .56332 L
.03866 .56412 L
.04837 .565 L
.05808 .56595 L
.06779 .56698 L
.07751 .56809 L
.08722 .56928 L
.09693 .57055 L
.10664 .5719 L
.11306 .57284 L
.11636 .57334 L
.12607 .57486 L
.13578 .57649 L
.14549 .57824 L
.15521 .58012 L
.16492 .58212 L
.16698 .58256 L
.17463 .58424 L
.18434 .58652 L
.19406 .58897 L
.20377 .5916 L
.20612 .59227 L
.21348 .59442 L
.22319 .59746 L
.23291 .60075 L
.23637 .60198 L
.24262 .60429 L
.25233 .60814 L
.26062 .61169 L
.26204 .61233 L
.27176 .61689 L
.28058 .62141 L
.28147 .62188 L
.29118 .62737 L
.29732 .63112 L
.30089 .63341 L
.31061 .64012 L
.31158 .64083 L
.32032 .64758 L
.32389 .65054 L
.33003 .65595 L
.33462 .66026 L
.33974 .66538 L
.34405 .66997 L
.34946 .67611 L
.35242 .67968 L
.35917 .68842 L
.35988 .68939 L
.36659 .69911 L
.36888 .70268 L
.37263 .70882 L
.37812 .71853 L
.37859 .71942 L
.38311 .72824 L
.38767 .73796 L
.38831 .73938 L
.39186 .74767 L
.39571 .75738 L
.39802 .76363 L
.39925 .76709 L
.40254 .77681 L
.40558 .78652 L
.40773 .79388 L
.4084 .79623 L
.41103 .80594 L
.41348 .81566 L
.41576 .82537 L
.41744 .83302 L
.41788 .83508 L
.41988 .84479 L
.42176 .85451 L
.42351 .86422 L
.42514 .87393 L
.42666 .88364 L
.42716 .88694 L
.4281 .89336 L
.42945 .90307 L
.43072 .91278 L
.43191 .92249 L
.43302 .93221 L
.43405 .94192 L
.435 .95163 L
.43588 .96134 L
.43668 .97106 L
.43687 .97357 L
.43743 .98077 L
.01923 .98077 L
F
0 g
.01923 .56257 m
.02643 .56313 L
.02894 .56332 L
.03866 .56412 L
.04837 .565 L
.05808 .56595 L
.06779 .56698 L
.07751 .56809 L
.08722 .56928 L
.09693 .57055 L
.10664 .5719 L
.11306 .57284 L
.11636 .57334 L
.12607 .57486 L
.13578 .57649 L
.14549 .57824 L
.15521 .58012 L
.16492 .58212 L
.16698 .58256 L
.17463 .58424 L
.18434 .58652 L
.19406 .58897 L
.20377 .5916 L
.20612 .59227 L
.21348 .59442 L
.22319 .59746 L
.23291 .60075 L
.23637 .60198 L
.24262 .60429 L
.25233 .60814 L
.26062 .61169 L
.26204 .61233 L
.27176 .61689 L
.28058 .62141 L
.28147 .62188 L
.29118 .62737 L
.29732 .63112 L
.30089 .63341 L
.31061 .64012 L
.31158 .64083 L
.32032 .64758 L
.32389 .65054 L
.33003 .65595 L
.33462 .66026 L
.33974 .66538 L
.34405 .66997 L
.34946 .67611 L
.35242 .67968 L
.35917 .68842 L
.35988 .68939 L
Mistroke
.36659 .69911 L
.36888 .70268 L
.37263 .70882 L
.37812 .71853 L
.37859 .71942 L
.38311 .72824 L
.38767 .73796 L
.38831 .73938 L
.39186 .74767 L
.39571 .75738 L
.39802 .76363 L
.39925 .76709 L
.40254 .77681 L
.40558 .78652 L
.40773 .79388 L
.4084 .79623 L
.41103 .80594 L
.41348 .81566 L
.41576 .82537 L
.41744 .83302 L
.41788 .83508 L
.41988 .84479 L
.42176 .85451 L
.42351 .86422 L
.42514 .87393 L
.42666 .88364 L
.42716 .88694 L
.4281 .89336 L
.42945 .90307 L
.43072 .91278 L
.43191 .92249 L
.43302 .93221 L
.43405 .94192 L
.435 .95163 L
.43588 .96134 L
.43668 .97106 L
.43687 .97357 L
.43743 .98077 L
Mfstroke
.56 1 0 r
.01923 .43173 m
.02894 .43108 L
.03866 .43036 L
.04837 .42956 L
.05808 .42868 L
.06779 .42772 L
.07314 .42716 L
.07751 .4267 L
.08722 .42561 L
.09693 .42441 L
.10664 .4231 L
.11636 .42169 L
.12607 .42016 L
.13578 .41852 L
.14184 .41744 L
.14549 .41678 L
.15521 .41492 L
.16492 .41291 L
.17463 .41074 L
.18434 .40842 L
.18707 .40773 L
.19406 .40593 L
.20377 .40325 L
.21348 .40035 L
.22077 .39802 L
.22319 .39722 L
.23291 .39384 L
.24262 .39017 L
.24727 .38831 L
.25233 .3862 L
.26204 .38186 L
.26882 .37859 L
.27176 .37712 L
.28147 .37193 L
.28676 .36888 L
.29118 .36621 L
.30089 .3599 L
.30195 .35917 L
.31061 .35289 L
.315 .34946 L
.32032 .34506 L
.32633 .33974 L
.33003 .33627 L
.33627 .33003 L
.33974 .32633 L
.34506 .32032 L
.34946 .315 L
.35289 .31061 L
.35917 .30195 L
.3599 .30089 L
.36621 .29118 L
.36888 .28676 L
.37193 .28147 L
.37712 .27176 L
.37859 .26882 L
.38186 .26204 L
.3862 .25233 L
.38831 .24727 L
.39017 .24262 L
.39384 .23291 L
.39722 .22319 L
.39802 .22077 L
.40035 .21348 L
.40325 .20377 L
.40593 .19406 L
.40773 .18707 L
.40842 .18434 L
.41074 .17463 L
.41291 .16492 L
.41492 .15521 L
.41678 .14549 L
.41744 .14184 L
.41852 .13578 L
.42016 .12607 L
.42169 .11636 L
.4231 .10664 L
.42441 .09693 L
.42561 .08722 L
.4267 .07751 L
.42716 .07314 L
.42772 .06779 L
.42868 .05808 L
.42956 .04837 L
.43036 .03866 L
.43108 .02894 L
.43173 .01923 L
.01923 .01923 L
F
0 g
.01923 .43173 m
.02894 .43108 L
.03866 .43036 L
.04837 .42956 L
.05808 .42868 L
.06779 .42772 L
.07314 .42716 L
.07751 .4267 L
.08722 .42561 L
.09693 .42441 L
.10664 .4231 L
.11636 .42169 L
.12607 .42016 L
.13578 .41852 L
.14184 .41744 L
.14549 .41678 L
.15521 .41492 L
.16492 .41291 L
.17463 .41074 L
.18434 .40842 L
.18707 .40773 L
.19406 .40593 L
.20377 .40325 L
.21348 .40035 L
.22077 .39802 L
.22319 .39722 L
.23291 .39384 L
.24262 .39017 L
.24727 .38831 L
.25233 .3862 L
.26204 .38186 L
.26882 .37859 L
.27176 .37712 L
.28147 .37193 L
.28676 .36888 L
.29118 .36621 L
.30089 .3599 L
.30195 .35917 L
.31061 .35289 L
.315 .34946 L
.32032 .34506 L
.32633 .33974 L
.33003 .33627 L
.33627 .33003 L
.33974 .32633 L
.34506 .32032 L
.34946 .315 L
.35289 .31061 L
.35917 .30195 L
.3599 .30089 L
Mistroke
.36621 .29118 L
.36888 .28676 L
.37193 .28147 L
.37712 .27176 L
.37859 .26882 L
.38186 .26204 L
.3862 .25233 L
.38831 .24727 L
.39017 .24262 L
.39384 .23291 L
.39722 .22319 L
.39802 .22077 L
.40035 .21348 L
.40325 .20377 L
.40593 .19406 L
.40773 .18707 L
.40842 .18434 L
.41074 .17463 L
.41291 .16492 L
.41492 .15521 L
.41678 .14549 L
.41744 .14184 L
.41852 .13578 L
.42016 .12607 L
.42169 .11636 L
.4231 .10664 L
.42441 .09693 L
.42561 .08722 L
.4267 .07751 L
.42716 .07314 L
.42772 .06779 L
.42868 .05808 L
.42956 .04837 L
.43036 .03866 L
.43108 .02894 L
.43173 .01923 L
Mfstroke
.56 1 0 r
.01923 .56827 m
.02894 .56892 L
.03866 .56964 L
.04837 .57044 L
.05808 .57132 L
.06779 .57228 L
.07314 .57284 L
.07751 .5733 L
.08722 .57439 L
.09693 .57559 L
.10664 .5769 L
.11636 .57831 L
.12607 .57984 L
.13578 .58148 L
.14184 .58256 L
.14549 .58322 L
.15521 .58508 L
.16492 .58709 L
.17463 .58926 L
.18434 .59158 L
.18707 .59227 L
.19406 .59407 L
.20377 .59675 L
.21348 .59965 L
.22077 .60198 L
.22319 .60278 L
.23291 .60616 L
.24262 .60983 L
.24727 .61169 L
.25233 .6138 L
.26204 .61814 L
.26882 .62141 L
.27176 .62288 L
.28147 .62807 L
.28676 .63112 L
.29118 .63379 L
.30089 .6401 L
.30195 .64083 L
.31061 .64711 L
.315 .65054 L
.32032 .65494 L
.32633 .66026 L
.33003 .66373 L
.33627 .66997 L
.33974 .67367 L
.34506 .67968 L
.34946 .685 L
.35289 .68939 L
.35917 .69805 L
.3599 .69911 L
.36621 .70882 L
.36888 .71324 L
.37193 .71853 L
.37712 .72824 L
.37859 .73118 L
.38186 .73796 L
.3862 .74767 L
.38831 .75273 L
.39017 .75738 L
.39384 .76709 L
.39722 .77681 L
.39802 .77923 L
.40035 .78652 L
.40325 .79623 L
.40593 .80594 L
.40773 .81293 L
.40842 .81566 L
.41074 .82537 L
.41291 .83508 L
.41492 .84479 L
.41678 .85451 L
.41744 .85816 L
.41852 .86422 L
.42016 .87393 L
.42169 .88364 L
.4231 .89336 L
.42441 .90307 L
.42561 .91278 L
.4267 .92249 L
.42716 .92686 L
.42772 .93221 L
.42868 .94192 L
.42956 .95163 L
.43036 .96134 L
.43108 .97106 L
.43173 .98077 L
.01923 .98077 L
F
0 g
.01923 .56827 m
.02894 .56892 L
.03866 .56964 L
.04837 .57044 L
.05808 .57132 L
.06779 .57228 L
.07314 .57284 L
.07751 .5733 L
.08722 .57439 L
.09693 .57559 L
.10664 .5769 L
.11636 .57831 L
.12607 .57984 L
.13578 .58148 L
.14184 .58256 L
.14549 .58322 L
.15521 .58508 L
.16492 .58709 L
.17463 .58926 L
.18434 .59158 L
.18707 .59227 L
.19406 .59407 L
.20377 .59675 L
.21348 .59965 L
.22077 .60198 L
.22319 .60278 L
.23291 .60616 L
.24262 .60983 L
.24727 .61169 L
.25233 .6138 L
.26204 .61814 L
.26882 .62141 L
.27176 .62288 L
.28147 .62807 L
.28676 .63112 L
.29118 .63379 L
.30089 .6401 L
.30195 .64083 L
.31061 .64711 L
.315 .65054 L
.32032 .65494 L
.32633 .66026 L
.33003 .66373 L
.33627 .66997 L
.33974 .67367 L
.34506 .67968 L
.34946 .685 L
.35289 .68939 L
.35917 .69805 L
.3599 .69911 L
Mistroke
.36621 .70882 L
.36888 .71324 L
.37193 .71853 L
.37712 .72824 L
.37859 .73118 L
.38186 .73796 L
.3862 .74767 L
.38831 .75273 L
.39017 .75738 L
.39384 .76709 L
.39722 .77681 L
.39802 .77923 L
.40035 .78652 L
.40325 .79623 L
.40593 .80594 L
.40773 .81293 L
.40842 .81566 L
.41074 .82537 L
.41291 .83508 L
.41492 .84479 L
.41678 .85451 L
.41744 .85816 L
.41852 .86422 L
.42016 .87393 L
.42169 .88364 L
.4231 .89336 L
.42441 .90307 L
.42561 .91278 L
.4267 .92249 L
.42716 .92686 L
.42772 .93221 L
.42868 .94192 L
.42956 .95163 L
.43036 .96134 L
.43108 .97106 L
.43173 .98077 L
Mfstroke
.8 1 0 r
.01923 .42536 m
.02894 .42475 L
.03866 .42415 L
.04837 .42349 L
.05808 .42275 L
.06779 .42191 L
.07751 .42097 L
.08722 .41993 L
.09693 .41879 L
.10664 .41754 L
.10734 .41744 L
.11636 .41624 L
.12607 .4148 L
.13578 .4132 L
.14549 .41147 L
.15521 .40958 L
.16411 .40773 L
.16492 .40756 L
.17463 .40541 L
.18434 .40307 L
.19406 .40052 L
.20292 .39802 L
.20377 .39777 L
.21348 .39483 L
.22319 .39161 L
.23242 .38831 L
.23291 .38813 L
.24262 .38436 L
.25233 .38023 L
.25597 .37859 L
.26204 .37575 L
.27176 .37082 L
.27535 .36888 L
.28147 .36542 L
.29118 .35945 L
.29162 .35917 L
.30089 .35286 L
.30552 .34946 L
.31061 .34551 L
.31754 .33974 L
.32032 .33729 L
.32804 .33003 L
.33003 .32804 L
.33729 .32032 L
.33974 .31754 L
.34551 .31061 L
.34946 .30552 L
.35286 .30089 L
.35917 .29162 L
.35945 .29118 L
.36542 .28147 L
.36888 .27535 L
.37082 .27176 L
.37575 .26204 L
.37859 .25597 L
.38023 .25233 L
.38436 .24262 L
.38813 .23291 L
.38831 .23242 L
.39161 .22319 L
.39483 .21348 L
.39777 .20377 L
.39802 .20292 L
.40052 .19406 L
.40307 .18434 L
.40541 .17463 L
.40756 .16492 L
.40773 .16411 L
.40958 .15521 L
.41147 .14549 L
.4132 .13578 L
.4148 .12607 L
.41624 .11636 L
.41744 .10734 L
.41754 .10664 L
.41879 .09693 L
.41993 .08722 L
.42097 .07751 L
.42191 .06779 L
.42275 .05808 L
.42349 .04837 L
.42415 .03866 L
.42475 .02894 L
.42536 .01923 L
.01923 .01923 L
F
0 g
.01923 .42536 m
.02894 .42475 L
.03866 .42415 L
.04837 .42349 L
.05808 .42275 L
.06779 .42191 L
.07751 .42097 L
.08722 .41993 L
.09693 .41879 L
.10664 .41754 L
.10734 .41744 L
.11636 .41624 L
.12607 .4148 L
.13578 .4132 L
.14549 .41147 L
.15521 .40958 L
.16411 .40773 L
.16492 .40756 L
.17463 .40541 L
.18434 .40307 L
.19406 .40052 L
.20292 .39802 L
.20377 .39777 L
.21348 .39483 L
.22319 .39161 L
.23242 .38831 L
.23291 .38813 L
.24262 .38436 L
.25233 .38023 L
.25597 .37859 L
.26204 .37575 L
.27176 .37082 L
.27535 .36888 L
.28147 .36542 L
.29118 .35945 L
.29162 .35917 L
.30089 .35286 L
.30552 .34946 L
.31061 .34551 L
.31754 .33974 L
.32032 .33729 L
.32804 .33003 L
.33003 .32804 L
.33729 .32032 L
.33974 .31754 L
.34551 .31061 L
.34946 .30552 L
.35286 .30089 L
.35917 .29162 L
.35945 .29118 L
Mistroke
.36542 .28147 L
.36888 .27535 L
.37082 .27176 L
.37575 .26204 L
.37859 .25597 L
.38023 .25233 L
.38436 .24262 L
.38813 .23291 L
.38831 .23242 L
.39161 .22319 L
.39483 .21348 L
.39777 .20377 L
.39802 .20292 L
.40052 .19406 L
.40307 .18434 L
.40541 .17463 L
.40756 .16492 L
.40773 .16411 L
.40958 .15521 L
.41147 .14549 L
.4132 .13578 L
.4148 .12607 L
.41624 .11636 L
.41744 .10734 L
.41754 .10664 L
.41879 .09693 L
.41993 .08722 L
.42097 .07751 L
.42191 .06779 L
.42275 .05808 L
.42349 .04837 L
.42415 .03866 L
.42475 .02894 L
.42536 .01923 L
Mfstroke
.8 1 0 r
.01923 .57464 m
.02894 .57525 L
.03866 .57585 L
.04837 .57651 L
.05808 .57725 L
.06779 .57809 L
.07751 .57903 L
.08722 .58007 L
.09693 .58121 L
.10664 .58246 L
.10734 .58256 L
.11636 .58376 L
.12607 .5852 L
.13578 .5868 L
.14549 .58853 L
.15521 .59042 L
.16411 .59227 L
.16492 .59244 L
.17463 .59459 L
.18434 .59693 L
.19406 .59948 L
.20292 .60198 L
.20377 .60223 L
.21348 .60517 L
.22319 .60839 L
.23242 .61169 L
.23291 .61187 L
.24262 .61564 L
.25233 .61977 L
.25597 .62141 L
.26204 .62425 L
.27176 .62918 L
.27535 .63112 L
.28147 .63458 L
.29118 .64055 L
.29162 .64083 L
.30089 .64714 L
.30552 .65054 L
.31061 .65449 L
.31754 .66026 L
.32032 .66271 L
.32804 .66997 L
.33003 .67196 L
.33729 .67968 L
.33974 .68246 L
.34551 .68939 L
.34946 .69448 L
.35286 .69911 L
.35917 .70838 L
.35945 .70882 L
.36542 .71853 L
.36888 .72465 L
.37082 .72824 L
.37575 .73796 L
.37859 .74403 L
.38023 .74767 L
.38436 .75738 L
.38813 .76709 L
.38831 .76758 L
.39161 .77681 L
.39483 .78652 L
.39777 .79623 L
.39802 .79708 L
.40052 .80594 L
.40307 .81566 L
.40541 .82537 L
.40756 .83508 L
.40773 .83589 L
.40958 .84479 L
.41147 .85451 L
.4132 .86422 L
.4148 .87393 L
.41624 .88364 L
.41744 .89266 L
.41754 .89336 L
.41879 .90307 L
.41993 .91278 L
.42097 .92249 L
.42191 .93221 L
.42275 .94192 L
.42349 .95163 L
.42415 .96134 L
.42475 .97106 L
.42536 .98077 L
.01923 .98077 L
F
0 g
.01923 .57464 m
.02894 .57525 L
.03866 .57585 L
.04837 .57651 L
.05808 .57725 L
.06779 .57809 L
.07751 .57903 L
.08722 .58007 L
.09693 .58121 L
.10664 .58246 L
.10734 .58256 L
.11636 .58376 L
.12607 .5852 L
.13578 .5868 L
.14549 .58853 L
.15521 .59042 L
.16411 .59227 L
.16492 .59244 L
.17463 .59459 L
.18434 .59693 L
.19406 .59948 L
.20292 .60198 L
.20377 .60223 L
.21348 .60517 L
.22319 .60839 L
.23242 .61169 L
.23291 .61187 L
.24262 .61564 L
.25233 .61977 L
.25597 .62141 L
.26204 .62425 L
.27176 .62918 L
.27535 .63112 L
.28147 .63458 L
.29118 .64055 L
.29162 .64083 L
.30089 .64714 L
.30552 .65054 L
.31061 .65449 L
.31754 .66026 L
.32032 .66271 L
.32804 .66997 L
.33003 .67196 L
.33729 .67968 L
.33974 .68246 L
.34551 .68939 L
.34946 .69448 L
.35286 .69911 L
.35917 .70838 L
.35945 .70882 L
Mistroke
.36542 .71853 L
.36888 .72465 L
.37082 .72824 L
.37575 .73796 L
.37859 .74403 L
.38023 .74767 L
.38436 .75738 L
.38813 .76709 L
.38831 .76758 L
.39161 .77681 L
.39483 .78652 L
.39777 .79623 L
.39802 .79708 L
.40052 .80594 L
.40307 .81566 L
.40541 .82537 L
.40756 .83508 L
.40773 .83589 L
.40958 .84479 L
.41147 .85451 L
.4132 .86422 L
.4148 .87393 L
.41624 .88364 L
.41744 .89266 L
.41754 .89336 L
.41879 .90307 L
.41993 .91278 L
.42097 .92249 L
.42191 .93221 L
.42275 .94192 L
.42349 .95163 L
.42415 .96134 L
.42475 .97106 L
.42536 .98077 L
Mfstroke
1 .96 0 r
.01923 .58788 m
.02894 .58611 L
.03866 .58539 L
.04837 .58516 L
.05808 .58368 L
.06779 .58474 L
.07751 .58562 L
.08722 .58654 L
.09693 .58755 L
.10664 .58869 L
.11636 .58995 L
.12607 .59134 L
.13206 .59227 L
.13578 .59282 L
.14549 .59443 L
.15521 .59625 L
.16492 .59826 L
.17463 .60045 L
.181 .60198 L
.18434 .60279 L
.19406 .60533 L
.20377 .60812 L
.21348 .61116 L
.21512 .61169 L
.22319 .61442 L
.23291 .61799 L
.24143 .62141 L
.24262 .6219 L
.25233 .62614 L
.26204 .63081 L
.26265 .63112 L
.27176 .63591 L
.28028 .64083 L
.28147 .64155 L
.29118 .64777 L
.2952 .65054 L
.30089 .65468 L
.30802 .66026 L
.31061 .6624 L
.31916 .66997 L
.32032 .67106 L
.32894 .67968 L
.33003 .68084 L
.3376 .68939 L
.33974 .69198 L
.34532 .69911 L
.34946 .7048 L
.35223 .70882 L
.35845 .71853 L
.35917 .71972 L
.36409 .72824 L
.36888 .73735 L
.36919 .73796 L
.37386 .74767 L
.3781 .75738 L
.37859 .75857 L
.38201 .76709 L
.38558 .77681 L
.38831 .78488 L
.38884 .78652 L
.39188 .79623 L
.39467 .80594 L
.39721 .81566 L
.39802 .819 L
.39955 .82537 L
.40174 .83508 L
.40375 .84479 L
.40557 .85451 L
.40718 .86422 L
.40773 .86794 L
.40866 .87393 L
.41005 .88364 L
.41131 .89336 L
.41245 .90307 L
.41346 .91278 L
.41438 .92249 L
.41526 .93221 L
.41632 .94192 L
.41484 .95163 L
.41461 .96134 L
.41389 .97106 L
.41212 .98077 L
.01923 .98077 L
F
0 g
.01923 .58788 m
.02894 .58611 L
.03866 .58539 L
.04837 .58516 L
.05808 .58368 L
.06779 .58474 L
.07751 .58562 L
.08722 .58654 L
.09693 .58755 L
.10664 .58869 L
.11636 .58995 L
.12607 .59134 L
.13206 .59227 L
.13578 .59282 L
.14549 .59443 L
.15521 .59625 L
.16492 .59826 L
.17463 .60045 L
.181 .60198 L
.18434 .60279 L
.19406 .60533 L
.20377 .60812 L
.21348 .61116 L
.21512 .61169 L
.22319 .61442 L
.23291 .61799 L
.24143 .62141 L
.24262 .6219 L
.25233 .62614 L
.26204 .63081 L
.26265 .63112 L
.27176 .63591 L
.28028 .64083 L
.28147 .64155 L
.29118 .64777 L
.2952 .65054 L
.30089 .65468 L
.30802 .66026 L
.31061 .6624 L
.31916 .66997 L
.32032 .67106 L
.32894 .67968 L
.33003 .68084 L
.3376 .68939 L
.33974 .69198 L
.34532 .69911 L
.34946 .7048 L
.35223 .70882 L
.35845 .71853 L
.35917 .71972 L
Mistroke
.36409 .72824 L
.36888 .73735 L
.36919 .73796 L
.37386 .74767 L
.3781 .75738 L
.37859 .75857 L
.38201 .76709 L
.38558 .77681 L
.38831 .78488 L
.38884 .78652 L
.39188 .79623 L
.39467 .80594 L
.39721 .81566 L
.39802 .819 L
.39955 .82537 L
.40174 .83508 L
.40375 .84479 L
.40557 .85451 L
.40718 .86422 L
.40773 .86794 L
.40866 .87393 L
.41005 .88364 L
.41131 .89336 L
.41245 .90307 L
.41346 .91278 L
.41438 .92249 L
.41526 .93221 L
.41632 .94192 L
.41484 .95163 L
.41461 .96134 L
.41389 .97106 L
.41212 .98077 L
Mfstroke
1 .96 0 r
.01923 .41212 m
.02894 .41389 L
.03866 .41461 L
.04837 .41484 L
.05808 .41632 L
.06779 .41526 L
.07751 .41438 L
.08722 .41346 L
.09693 .41245 L
.10664 .41131 L
.11636 .41005 L
.12607 .40866 L
.13206 .40773 L
.13578 .40718 L
.14549 .40557 L
.15521 .40375 L
.16492 .40174 L
.17463 .39955 L
.181 .39802 L
.18434 .39721 L
.19406 .39467 L
.20377 .39188 L
.21348 .38884 L
.21512 .38831 L
.22319 .38558 L
.23291 .38201 L
.24143 .37859 L
.24262 .3781 L
.25233 .37386 L
.26204 .36919 L
.26265 .36888 L
.27176 .36409 L
.28028 .35917 L
.28147 .35845 L
.29118 .35223 L
.2952 .34946 L
.30089 .34532 L
.30802 .33974 L
.31061 .3376 L
.31916 .33003 L
.32032 .32894 L
.32894 .32032 L
.33003 .31916 L
.3376 .31061 L
.33974 .30802 L
.34532 .30089 L
.34946 .2952 L
.35223 .29118 L
.35845 .28147 L
.35917 .28028 L
.36409 .27176 L
.36888 .26265 L
.36919 .26204 L
.37386 .25233 L
.3781 .24262 L
.37859 .24143 L
.38201 .23291 L
.38558 .22319 L
.38831 .21512 L
.38884 .21348 L
.39188 .20377 L
.39467 .19406 L
.39721 .18434 L
.39802 .181 L
.39955 .17463 L
.40174 .16492 L
.40375 .15521 L
.40557 .14549 L
.40718 .13578 L
.40773 .13206 L
.40866 .12607 L
.41005 .11636 L
.41131 .10664 L
.41245 .09693 L
.41346 .08722 L
.41438 .07751 L
.41526 .06779 L
.41632 .05808 L
.41484 .04837 L
.41461 .03866 L
.41389 .02894 L
.41212 .01923 L
.01923 .01923 L
F
0 g
.01923 .41212 m
.02894 .41389 L
.03866 .41461 L
.04837 .41484 L
.05808 .41632 L
.06779 .41526 L
.07751 .41438 L
.08722 .41346 L
.09693 .41245 L
.10664 .41131 L
.11636 .41005 L
.12607 .40866 L
.13206 .40773 L
.13578 .40718 L
.14549 .40557 L
.15521 .40375 L
.16492 .40174 L
.17463 .39955 L
.181 .39802 L
.18434 .39721 L
.19406 .39467 L
.20377 .39188 L
.21348 .38884 L
.21512 .38831 L
.22319 .38558 L
.23291 .38201 L
.24143 .37859 L
.24262 .3781 L
.25233 .37386 L
.26204 .36919 L
.26265 .36888 L
.27176 .36409 L
.28028 .35917 L
.28147 .35845 L
.29118 .35223 L
.2952 .34946 L
.30089 .34532 L
.30802 .33974 L
.31061 .3376 L
.31916 .33003 L
.32032 .32894 L
.32894 .32032 L
.33003 .31916 L
.3376 .31061 L
.33974 .30802 L
.34532 .30089 L
.34946 .2952 L
.35223 .29118 L
.35845 .28147 L
.35917 .28028 L
Mistroke
.36409 .27176 L
.36888 .26265 L
.36919 .26204 L
.37386 .25233 L
.3781 .24262 L
.37859 .24143 L
.38201 .23291 L
.38558 .22319 L
.38831 .21512 L
.38884 .21348 L
.39188 .20377 L
.39467 .19406 L
.39721 .18434 L
.39802 .181 L
.39955 .17463 L
.40174 .16492 L
.40375 .15521 L
.40557 .14549 L
.40718 .13578 L
.40773 .13206 L
.40866 .12607 L
.41005 .11636 L
.41131 .10664 L
.41245 .09693 L
.41346 .08722 L
.41438 .07751 L
.41526 .06779 L
.41632 .05808 L
.41484 .04837 L
.41461 .03866 L
.41389 .02894 L
.41212 .01923 L
Mfstroke
.8 1 0 r
.01923 .40023 m
.02271 .39802 L
.02894 .39462 L
.03866 .38989 L
.04208 .38831 L
.04837 .38478 L
.05808 .37977 L
.06419 .37859 L
.06779 .37689 L
.07751 .37236 L
.08656 .36888 L
.08722 .36858 L
.09693 .36406 L
.10664 .35977 L
.10812 .35917 L
.11636 .3552 L
.12607 .35055 L
.12845 .34946 L
.13578 .34571 L
.14549 .34071 L
.14739 .33974 L
.15521 .33545 L
.16492 .33004 L
.16493 .33003 L
.17463 .32427 L
.18114 .32032 L
.18434 .31829 L
.19406 .31197 L
.19612 .31061 L
.20377 .30531 L
.20996 .30089 L
.21348 .29828 L
.22277 .29118 L
.22319 .29085 L
.23291 .28293 L
.23464 .28147 L
.24262 .27451 L
.24568 .27176 L
.25233 .26554 L
.25595 .26204 L
.26204 .25595 L
.26554 .25233 L
.27176 .24568 L
.27451 .24262 L
.28147 .23464 L
.28293 .23291 L
.29085 .22319 L
.29118 .22277 L
.29828 .21348 L
.30089 .20996 L
.30531 .20377 L
.31061 .19612 L
.31197 .19406 L
.31829 .18434 L
.32032 .18114 L
.32427 .17463 L
.33003 .16493 L
.33004 .16492 L
.33545 .15521 L
.33974 .14739 L
.34071 .14549 L
.34571 .13578 L
.34946 .12845 L
.35055 .12607 L
.3552 .11636 L
.35917 .10812 L
.35977 .10664 L
.36406 .09693 L
.36858 .08722 L
.36888 .08656 L
.37236 .07751 L
.37689 .06779 L
.37859 .06419 L
.37977 .05808 L
.38478 .04837 L
.38831 .04208 L
.38989 .03866 L
.39462 .02894 L
.39802 .02271 L
.40023 .01923 L
.01923 .01923 L
F
0 g
.01923 .40023 m
.02271 .39802 L
.02894 .39462 L
.03866 .38989 L
.04208 .38831 L
.04837 .38478 L
.05808 .37977 L
.06419 .37859 L
.06779 .37689 L
.07751 .37236 L
.08656 .36888 L
.08722 .36858 L
.09693 .36406 L
.10664 .35977 L
.10812 .35917 L
.11636 .3552 L
.12607 .35055 L
.12845 .34946 L
.13578 .34571 L
.14549 .34071 L
.14739 .33974 L
.15521 .33545 L
.16492 .33004 L
.16493 .33003 L
.17463 .32427 L
.18114 .32032 L
.18434 .31829 L
.19406 .31197 L
.19612 .31061 L
.20377 .30531 L
.20996 .30089 L
.21348 .29828 L
.22277 .29118 L
.22319 .29085 L
.23291 .28293 L
.23464 .28147 L
.24262 .27451 L
.24568 .27176 L
.25233 .26554 L
.25595 .26204 L
.26204 .25595 L
.26554 .25233 L
.27176 .24568 L
.27451 .24262 L
.28147 .23464 L
.28293 .23291 L
.29085 .22319 L
.29118 .22277 L
.29828 .21348 L
.30089 .20996 L
Mistroke
.30531 .20377 L
.31061 .19612 L
.31197 .19406 L
.31829 .18434 L
.32032 .18114 L
.32427 .17463 L
.33003 .16493 L
.33004 .16492 L
.33545 .15521 L
.33974 .14739 L
.34071 .14549 L
.34571 .13578 L
.34946 .12845 L
.35055 .12607 L
.3552 .11636 L
.35917 .10812 L
.35977 .10664 L
.36406 .09693 L
.36858 .08722 L
.36888 .08656 L
.37236 .07751 L
.37689 .06779 L
.37859 .06419 L
.37977 .05808 L
.38478 .04837 L
.38831 .04208 L
.38989 .03866 L
.39462 .02894 L
.39802 .02271 L
.40023 .01923 L
Mfstroke
.8 1 0 r
.01923 .59977 m
.02271 .60198 L
.02894 .60538 L
.03866 .61011 L
.04208 .61169 L
.04837 .61522 L
.05808 .62023 L
.06419 .62141 L
.06779 .62311 L
.07751 .62764 L
.08656 .63112 L
.08722 .63142 L
.09693 .63594 L
.10664 .64023 L
.10812 .64083 L
.11636 .6448 L
.12607 .64945 L
.12845 .65054 L
.13578 .65429 L
.14549 .65929 L
.14739 .66026 L
.15521 .66455 L
.16492 .66996 L
.16493 .66997 L
.17463 .67573 L
.18114 .67968 L
.18434 .68171 L
.19406 .68803 L
.19612 .68939 L
.20377 .69469 L
.20996 .69911 L
.21348 .70172 L
.22277 .70882 L
.22319 .70915 L
.23291 .71707 L
.23464 .71853 L
.24262 .72549 L
.24568 .72824 L
.25233 .73446 L
.25595 .73796 L
.26204 .74405 L
.26554 .74767 L
.27176 .75432 L
.27451 .75738 L
.28147 .76536 L
.28293 .76709 L
.29085 .77681 L
.29118 .77723 L
.29828 .78652 L
.30089 .79004 L
.30531 .79623 L
.31061 .80388 L
.31197 .80594 L
.31829 .81566 L
.32032 .81886 L
.32427 .82537 L
.33003 .83507 L
.33004 .83508 L
.33545 .84479 L
.33974 .85261 L
.34071 .85451 L
.34571 .86422 L
.34946 .87155 L
.35055 .87393 L
.3552 .88364 L
.35917 .89188 L
.35977 .89336 L
.36406 .90307 L
.36858 .91278 L
.36888 .91344 L
.37236 .92249 L
.37689 .93221 L
.37859 .93581 L
.37977 .94192 L
.38478 .95163 L
.38831 .95792 L
.38989 .96134 L
.39462 .97106 L
.39802 .97729 L
.40023 .98077 L
.01923 .98077 L
F
0 g
.01923 .59977 m
.02271 .60198 L
.02894 .60538 L
.03866 .61011 L
.04208 .61169 L
.04837 .61522 L
.05808 .62023 L
.06419 .62141 L
.06779 .62311 L
.07751 .62764 L
.08656 .63112 L
.08722 .63142 L
.09693 .63594 L
.10664 .64023 L
.10812 .64083 L
.11636 .6448 L
.12607 .64945 L
.12845 .65054 L
.13578 .65429 L
.14549 .65929 L
.14739 .66026 L
.15521 .66455 L
.16492 .66996 L
.16493 .66997 L
.17463 .67573 L
.18114 .67968 L
.18434 .68171 L
.19406 .68803 L
.19612 .68939 L
.20377 .69469 L
.20996 .69911 L
.21348 .70172 L
.22277 .70882 L
.22319 .70915 L
.23291 .71707 L
.23464 .71853 L
.24262 .72549 L
.24568 .72824 L
.25233 .73446 L
.25595 .73796 L
.26204 .74405 L
.26554 .74767 L
.27176 .75432 L
.27451 .75738 L
.28147 .76536 L
.28293 .76709 L
.29085 .77681 L
.29118 .77723 L
.29828 .78652 L
.30089 .79004 L
Mistroke
.30531 .79623 L
.31061 .80388 L
.31197 .80594 L
.31829 .81566 L
.32032 .81886 L
.32427 .82537 L
.33003 .83507 L
.33004 .83508 L
.33545 .84479 L
.33974 .85261 L
.34071 .85451 L
.34571 .86422 L
.34946 .87155 L
.35055 .87393 L
.3552 .88364 L
.35917 .89188 L
.35977 .89336 L
.36406 .90307 L
.36858 .91278 L
.36888 .91344 L
.37236 .92249 L
.37689 .93221 L
.37859 .93581 L
.37977 .94192 L
.38478 .95163 L
.38831 .95792 L
.38989 .96134 L
.39462 .97106 L
.39802 .97729 L
.40023 .98077 L
Mfstroke
.56 1 0 r
.01923 .38779 m
.02894 .38388 L
.03866 .38026 L
.04413 .37859 L
.04837 .37692 L
.05808 .37309 L
.06779 .36944 L
.06938 .36888 L
.07751 .36554 L
.08722 .36153 L
.09309 .35917 L
.09693 .3575 L
.10664 .3532 L
.11507 .34946 L
.11636 .34886 L
.12607 .34423 L
.13531 .33974 L
.13578 .33951 L
.14549 .33449 L
.1539 .33003 L
.15521 .32931 L
.16492 .32383 L
.17096 .32032 L
.17463 .3181 L
.18434 .31206 L
.18663 .31061 L
.19406 .30569 L
.20105 .30089 L
.20377 .29897 L
.21348 .29184 L
.21435 .29118 L
.22319 .28426 L
.22663 .28147 L
.23291 .2762 L
.23802 .27176 L
.24262 .26761 L
.24859 .26204 L
.25233 .25843 L
.25843 .25233 L
.26204 .24859 L
.26761 .24262 L
.27176 .23802 L
.2762 .23291 L
.28147 .22663 L
.28426 .22319 L
.29118 .21435 L
.29184 .21348 L
.29897 .20377 L
.30089 .20105 L
.30569 .19406 L
.31061 .18663 L
.31206 .18434 L
.3181 .17463 L
.32032 .17096 L
.32383 .16492 L
.32931 .15521 L
.33003 .1539 L
.33449 .14549 L
.33951 .13578 L
.33974 .13531 L
.34423 .12607 L
.34886 .11636 L
.34946 .11507 L
.3532 .10664 L
.3575 .09693 L
.35917 .09309 L
.36153 .08722 L
.36554 .07751 L
.36888 .06938 L
.36944 .06779 L
.37309 .05808 L
.37692 .04837 L
.37859 .04413 L
.38026 .03866 L
.38388 .02894 L
.38779 .01923 L
.01923 .01923 L
F
0 g
.01923 .38779 m
.02894 .38388 L
.03866 .38026 L
.04413 .37859 L
.04837 .37692 L
.05808 .37309 L
.06779 .36944 L
.06938 .36888 L
.07751 .36554 L
.08722 .36153 L
.09309 .35917 L
.09693 .3575 L
.10664 .3532 L
.11507 .34946 L
.11636 .34886 L
.12607 .34423 L
.13531 .33974 L
.13578 .33951 L
.14549 .33449 L
.1539 .33003 L
.15521 .32931 L
.16492 .32383 L
.17096 .32032 L
.17463 .3181 L
.18434 .31206 L
.18663 .31061 L
.19406 .30569 L
.20105 .30089 L
.20377 .29897 L
.21348 .29184 L
.21435 .29118 L
.22319 .28426 L
.22663 .28147 L
.23291 .2762 L
.23802 .27176 L
.24262 .26761 L
.24859 .26204 L
.25233 .25843 L
.25843 .25233 L
.26204 .24859 L
.26761 .24262 L
.27176 .23802 L
.2762 .23291 L
.28147 .22663 L
.28426 .22319 L
.29118 .21435 L
.29184 .21348 L
.29897 .20377 L
.30089 .20105 L
.30569 .19406 L
Mistroke
.31061 .18663 L
.31206 .18434 L
.3181 .17463 L
.32032 .17096 L
.32383 .16492 L
.32931 .15521 L
.33003 .1539 L
.33449 .14549 L
.33951 .13578 L
.33974 .13531 L
.34423 .12607 L
.34886 .11636 L
.34946 .11507 L
.3532 .10664 L
.3575 .09693 L
.35917 .09309 L
.36153 .08722 L
.36554 .07751 L
.36888 .06938 L
.36944 .06779 L
.37309 .05808 L
.37692 .04837 L
.37859 .04413 L
.38026 .03866 L
.38388 .02894 L
.38779 .01923 L
Mfstroke
.56 1 0 r
.01923 .61221 m
.02894 .61612 L
.03866 .61974 L
.04413 .62141 L
.04837 .62308 L
.05808 .62691 L
.06779 .63056 L
.06938 .63112 L
.07751 .63446 L
.08722 .63847 L
.09309 .64083 L
.09693 .6425 L
.10664 .6468 L
.11507 .65054 L
.11636 .65114 L
.12607 .65577 L
.13531 .66026 L
.13578 .66049 L
.14549 .66551 L
.1539 .66997 L
.15521 .67069 L
.16492 .67617 L
.17096 .67968 L
.17463 .6819 L
.18434 .68794 L
.18663 .68939 L
.19406 .69431 L
.20105 .69911 L
.20377 .70103 L
.21348 .70816 L
.21435 .70882 L
.22319 .71574 L
.22663 .71853 L
.23291 .7238 L
.23802 .72824 L
.24262 .73239 L
.24859 .73796 L
.25233 .74157 L
.25843 .74767 L
.26204 .75141 L
.26761 .75738 L
.27176 .76198 L
.2762 .76709 L
.28147 .77337 L
.28426 .77681 L
.29118 .78565 L
.29184 .78652 L
.29897 .79623 L
.30089 .79895 L
.30569 .80594 L
.31061 .81337 L
.31206 .81566 L
.3181 .82537 L
.32032 .82904 L
.32383 .83508 L
.32931 .84479 L
.33003 .8461 L
.33449 .85451 L
.33951 .86422 L
.33974 .86469 L
.34423 .87393 L
.34886 .88364 L
.34946 .88493 L
.3532 .89336 L
.3575 .90307 L
.35917 .90691 L
.36153 .91278 L
.36554 .92249 L
.36888 .93062 L
.36944 .93221 L
.37309 .94192 L
.37692 .95163 L
.37859 .95587 L
.38026 .96134 L
.38388 .97106 L
.38779 .98077 L
.01923 .98077 L
F
0 g
.01923 .61221 m
.02894 .61612 L
.03866 .61974 L
.04413 .62141 L
.04837 .62308 L
.05808 .62691 L
.06779 .63056 L
.06938 .63112 L
.07751 .63446 L
.08722 .63847 L
.09309 .64083 L
.09693 .6425 L
.10664 .6468 L
.11507 .65054 L
.11636 .65114 L
.12607 .65577 L
.13531 .66026 L
.13578 .66049 L
.14549 .66551 L
.1539 .66997 L
.15521 .67069 L
.16492 .67617 L
.17096 .67968 L
.17463 .6819 L
.18434 .68794 L
.18663 .68939 L
.19406 .69431 L
.20105 .69911 L
.20377 .70103 L
.21348 .70816 L
.21435 .70882 L
.22319 .71574 L
.22663 .71853 L
.23291 .7238 L
.23802 .72824 L
.24262 .73239 L
.24859 .73796 L
.25233 .74157 L
.25843 .74767 L
.26204 .75141 L
.26761 .75738 L
.27176 .76198 L
.2762 .76709 L
.28147 .77337 L
.28426 .77681 L
.29118 .78565 L
.29184 .78652 L
.29897 .79623 L
.30089 .79895 L
.30569 .80594 L
Mistroke
.31061 .81337 L
.31206 .81566 L
.3181 .82537 L
.32032 .82904 L
.32383 .83508 L
.32931 .84479 L
.33003 .8461 L
.33449 .85451 L
.33951 .86422 L
.33974 .86469 L
.34423 .87393 L
.34886 .88364 L
.34946 .88493 L
.3532 .89336 L
.3575 .90307 L
.35917 .90691 L
.36153 .91278 L
.36554 .92249 L
.36888 .93062 L
.36944 .93221 L
.37309 .94192 L
.37692 .95163 L
.37859 .95587 L
.38026 .96134 L
.38388 .97106 L
.38779 .98077 L
Mfstroke
.32 1 0 r
.01923 .38035 m
.02499 .37859 L
.02894 .37723 L
.03866 .37382 L
.04837 .37043 L
.0529 .36888 L
.05808 .36696 L
.06779 .36329 L
.07751 .3596 L
.07864 .35917 L
.08722 .3557 L
.09693 .35167 L
.10219 .34946 L
.10664 .3475 L
.11636 .34313 L
.12369 .33974 L
.12607 .3386 L
.13578 .33383 L
.14328 .33003 L
.14549 .32887 L
.15521 .32364 L
.16117 .32032 L
.16492 .31816 L
.17463 .31238 L
.17752 .31061 L
.18434 .30628 L
.19251 .30089 L
.19406 .29984 L
.20377 .29301 L
.20629 .29118 L
.21348 .28577 L
.21898 .28147 L
.22319 .27806 L
.23071 .27176 L
.23291 .26985 L
.24158 .26204 L
.24262 .26107 L
.25167 .25233 L
.25233 .25167 L
.26107 .24262 L
.26204 .24158 L
.26985 .23291 L
.27176 .23071 L
.27806 .22319 L
.28147 .21898 L
.28577 .21348 L
.29118 .20629 L
.29301 .20377 L
.29984 .19406 L
.30089 .19251 L
.30628 .18434 L
.31061 .17752 L
.31238 .17463 L
.31816 .16492 L
.32032 .16117 L
.32364 .15521 L
.32887 .14549 L
.33003 .14328 L
.33383 .13578 L
.3386 .12607 L
.33974 .12369 L
.34313 .11636 L
.3475 .10664 L
.34946 .10219 L
.35167 .09693 L
.3557 .08722 L
.35917 .07864 L
.3596 .07751 L
.36329 .06779 L
.36696 .05808 L
.36888 .0529 L
.37043 .04837 L
.37382 .03866 L
.37723 .02894 L
.37859 .02499 L
.38035 .01923 L
.01923 .01923 L
F
0 g
.01923 .38035 m
.02499 .37859 L
.02894 .37723 L
.03866 .37382 L
.04837 .37043 L
.0529 .36888 L
.05808 .36696 L
.06779 .36329 L
.07751 .3596 L
.07864 .35917 L
.08722 .3557 L
.09693 .35167 L
.10219 .34946 L
.10664 .3475 L
.11636 .34313 L
.12369 .33974 L
.12607 .3386 L
.13578 .33383 L
.14328 .33003 L
.14549 .32887 L
.15521 .32364 L
.16117 .32032 L
.16492 .31816 L
.17463 .31238 L
.17752 .31061 L
.18434 .30628 L
.19251 .30089 L
.19406 .29984 L
.20377 .29301 L
.20629 .29118 L
.21348 .28577 L
.21898 .28147 L
.22319 .27806 L
.23071 .27176 L
.23291 .26985 L
.24158 .26204 L
.24262 .26107 L
.25167 .25233 L
.25233 .25167 L
.26107 .24262 L
.26204 .24158 L
.26985 .23291 L
.27176 .23071 L
.27806 .22319 L
.28147 .21898 L
.28577 .21348 L
.29118 .20629 L
.29301 .20377 L
.29984 .19406 L
.30089 .19251 L
Mistroke
.30628 .18434 L
.31061 .17752 L
.31238 .17463 L
.31816 .16492 L
.32032 .16117 L
.32364 .15521 L
.32887 .14549 L
.33003 .14328 L
.33383 .13578 L
.3386 .12607 L
.33974 .12369 L
.34313 .11636 L
.3475 .10664 L
.34946 .10219 L
.35167 .09693 L
.3557 .08722 L
.35917 .07864 L
.3596 .07751 L
.36329 .06779 L
.36696 .05808 L
.36888 .0529 L
.37043 .04837 L
.37382 .03866 L
.37723 .02894 L
.37859 .02499 L
.38035 .01923 L
Mfstroke
.32 1 0 r
.01923 .61965 m
.02499 .62141 L
.02894 .62277 L
.03866 .62618 L
.04837 .62957 L
.0529 .63112 L
.05808 .63304 L
.06779 .63671 L
.07751 .6404 L
.07864 .64083 L
.08722 .6443 L
.09693 .64833 L
.10219 .65054 L
.10664 .6525 L
.11636 .65687 L
.12369 .66026 L
.12607 .6614 L
.13578 .66617 L
.14328 .66997 L
.14549 .67113 L
.15521 .67636 L
.16117 .67968 L
.16492 .68184 L
.17463 .68762 L
.17752 .68939 L
.18434 .69372 L
.19251 .69911 L
.19406 .70016 L
.20377 .70699 L
.20629 .70882 L
.21348 .71423 L
.21898 .71853 L
.22319 .72194 L
.23071 .72824 L
.23291 .73015 L
.24158 .73796 L
.24262 .73893 L
.25167 .74767 L
.25233 .74833 L
.26107 .75738 L
.26204 .75842 L
.26985 .76709 L
.27176 .76929 L
.27806 .77681 L
.28147 .78102 L
.28577 .78652 L
.29118 .79371 L
.29301 .79623 L
.29984 .80594 L
.30089 .80749 L
.30628 .81566 L
.31061 .82248 L
.31238 .82537 L
.31816 .83508 L
.32032 .83883 L
.32364 .84479 L
.32887 .85451 L
.33003 .85672 L
.33383 .86422 L
.3386 .87393 L
.33974 .87631 L
.34313 .88364 L
.3475 .89336 L
.34946 .89781 L
.35167 .90307 L
.3557 .91278 L
.35917 .92136 L
.3596 .92249 L
.36329 .93221 L
.36696 .94192 L
.36888 .9471 L
.37043 .95163 L
.37382 .96134 L
.37723 .97106 L
.37859 .97501 L
.38035 .98077 L
.01923 .98077 L
F
0 g
.01923 .61965 m
.02499 .62141 L
.02894 .62277 L
.03866 .62618 L
.04837 .62957 L
.0529 .63112 L
.05808 .63304 L
.06779 .63671 L
.07751 .6404 L
.07864 .64083 L
.08722 .6443 L
.09693 .64833 L
.10219 .65054 L
.10664 .6525 L
.11636 .65687 L
.12369 .66026 L
.12607 .6614 L
.13578 .66617 L
.14328 .66997 L
.14549 .67113 L
.15521 .67636 L
.16117 .67968 L
.16492 .68184 L
.17463 .68762 L
.17752 .68939 L
.18434 .69372 L
.19251 .69911 L
.19406 .70016 L
.20377 .70699 L
.20629 .70882 L
.21348 .71423 L
.21898 .71853 L
.22319 .72194 L
.23071 .72824 L
.23291 .73015 L
.24158 .73796 L
.24262 .73893 L
.25167 .74767 L
.25233 .74833 L
.26107 .75738 L
.26204 .75842 L
.26985 .76709 L
.27176 .76929 L
.27806 .77681 L
.28147 .78102 L
.28577 .78652 L
.29118 .79371 L
.29301 .79623 L
.29984 .80594 L
.30089 .80749 L
Mistroke
.30628 .81566 L
.31061 .82248 L
.31238 .82537 L
.31816 .83508 L
.32032 .83883 L
.32364 .84479 L
.32887 .85451 L
.33003 .85672 L
.33383 .86422 L
.3386 .87393 L
.33974 .87631 L
.34313 .88364 L
.3475 .89336 L
.34946 .89781 L
.35167 .90307 L
.3557 .91278 L
.35917 .92136 L
.3596 .92249 L
.36329 .93221 L
.36696 .94192 L
.36888 .9471 L
.37043 .95163 L
.37382 .96134 L
.37723 .97106 L
.37859 .97501 L
.38035 .98077 L
Mfstroke
.08 1 0 r
.01923 .37443 m
.02894 .37132 L
.03654 .36888 L
.03866 .36817 L
.04837 .36482 L
.05808 .3614 L
.06432 .35917 L
.06779 .35787 L
.07751 .35417 L
.08722 .35035 L
.08947 .34946 L
.09693 .34636 L
.10664 .3422 L
.11223 .33974 L
.11636 .33787 L
.12607 .33331 L
.13284 .33003 L
.13578 .32856 L
.14549 .32356 L
.15156 .32032 L
.15521 .31831 L
.16492 .31278 L
.1686 .31061 L
.17463 .30694 L
.18417 .30089 L
.18434 .30078 L
.19406 .29424 L
.19842 .29118 L
.20377 .28731 L
.21153 .28147 L
.21348 .27995 L
.22319 .2721 L
.22361 .27176 L
.23291 .26372 L
.23477 .26204 L
.24262 .25475 L
.24512 .25233 L
.25233 .24512 L
.25475 .24262 L
.26204 .23477 L
.26372 .23291 L
.27176 .22361 L
.2721 .22319 L
.27995 .21348 L
.28147 .21153 L
.28731 .20377 L
.29118 .19842 L
.29424 .19406 L
.30078 .18434 L
.30089 .18417 L
.30694 .17463 L
.31061 .1686 L
.31278 .16492 L
.31831 .15521 L
.32032 .15156 L
.32356 .14549 L
.32856 .13578 L
.33003 .13284 L
.33331 .12607 L
.33787 .11636 L
.33974 .11223 L
.3422 .10664 L
.34636 .09693 L
.34946 .08947 L
.35035 .08722 L
.35417 .07751 L
.35787 .06779 L
.35917 .06432 L
.3614 .05808 L
.36482 .04837 L
.36817 .03866 L
.36888 .03654 L
.37132 .02894 L
.37443 .01923 L
.01923 .01923 L
F
0 g
.01923 .37443 m
.02894 .37132 L
.03654 .36888 L
.03866 .36817 L
.04837 .36482 L
.05808 .3614 L
.06432 .35917 L
.06779 .35787 L
.07751 .35417 L
.08722 .35035 L
.08947 .34946 L
.09693 .34636 L
.10664 .3422 L
.11223 .33974 L
.11636 .33787 L
.12607 .33331 L
.13284 .33003 L
.13578 .32856 L
.14549 .32356 L
.15156 .32032 L
.15521 .31831 L
.16492 .31278 L
.1686 .31061 L
.17463 .30694 L
.18417 .30089 L
.18434 .30078 L
.19406 .29424 L
.19842 .29118 L
.20377 .28731 L
.21153 .28147 L
.21348 .27995 L
.22319 .2721 L
.22361 .27176 L
.23291 .26372 L
.23477 .26204 L
.24262 .25475 L
.24512 .25233 L
.25233 .24512 L
.25475 .24262 L
.26204 .23477 L
.26372 .23291 L
.27176 .22361 L
.2721 .22319 L
.27995 .21348 L
.28147 .21153 L
.28731 .20377 L
.29118 .19842 L
.29424 .19406 L
.30078 .18434 L
.30089 .18417 L
Mistroke
.30694 .17463 L
.31061 .1686 L
.31278 .16492 L
.31831 .15521 L
.32032 .15156 L
.32356 .14549 L
.32856 .13578 L
.33003 .13284 L
.33331 .12607 L
.33787 .11636 L
.33974 .11223 L
.3422 .10664 L
.34636 .09693 L
.34946 .08947 L
.35035 .08722 L
.35417 .07751 L
.35787 .06779 L
.35917 .06432 L
.3614 .05808 L
.36482 .04837 L
.36817 .03866 L
.36888 .03654 L
.37132 .02894 L
.37443 .01923 L
Mfstroke
.08 1 0 r
.01923 .62557 m
.02894 .62868 L
.03654 .63112 L
.03866 .63183 L
.04837 .63518 L
.05808 .6386 L
.06432 .64083 L
.06779 .64213 L
.07751 .64583 L
.08722 .64965 L
.08947 .65054 L
.09693 .65364 L
.10664 .6578 L
.11223 .66026 L
.11636 .66213 L
.12607 .66669 L
.13284 .66997 L
.13578 .67144 L
.14549 .67644 L
.15156 .67968 L
.15521 .68169 L
.16492 .68722 L
.1686 .68939 L
.17463 .69306 L
.18417 .69911 L
.18434 .69922 L
.19406 .70576 L
.19842 .70882 L
.20377 .71269 L
.21153 .71853 L
.21348 .72005 L
.22319 .7279 L
.22361 .72824 L
.23291 .73628 L
.23477 .73796 L
.24262 .74525 L
.24512 .74767 L
.25233 .75488 L
.25475 .75738 L
.26204 .76523 L
.26372 .76709 L
.27176 .77639 L
.2721 .77681 L
.27995 .78652 L
.28147 .78847 L
.28731 .79623 L
.29118 .80158 L
.29424 .80594 L
.30078 .81566 L
.30089 .81583 L
.30694 .82537 L
.31061 .8314 L
.31278 .83508 L
.31831 .84479 L
.32032 .84844 L
.32356 .85451 L
.32856 .86422 L
.33003 .86716 L
.33331 .87393 L
.33787 .88364 L
.33974 .88777 L
.3422 .89336 L
.34636 .90307 L
.34946 .91053 L
.35035 .91278 L
.35417 .92249 L
.35787 .93221 L
.35917 .93568 L
.3614 .94192 L
.36482 .95163 L
.36817 .96134 L
.36888 .96346 L
.37132 .97106 L
.37443 .98077 L
.01923 .98077 L
F
0 g
.01923 .62557 m
.02894 .62868 L
.03654 .63112 L
.03866 .63183 L
.04837 .63518 L
.05808 .6386 L
.06432 .64083 L
.06779 .64213 L
.07751 .64583 L
.08722 .64965 L
.08947 .65054 L
.09693 .65364 L
.10664 .6578 L
.11223 .66026 L
.11636 .66213 L
.12607 .66669 L
.13284 .66997 L
.13578 .67144 L
.14549 .67644 L
.15156 .67968 L
.15521 .68169 L
.16492 .68722 L
.1686 .68939 L
.17463 .69306 L
.18417 .69911 L
.18434 .69922 L
.19406 .70576 L
.19842 .70882 L
.20377 .71269 L
.21153 .71853 L
.21348 .72005 L
.22319 .7279 L
.22361 .72824 L
.23291 .73628 L
.23477 .73796 L
.24262 .74525 L
.24512 .74767 L
.25233 .75488 L
.25475 .75738 L
.26204 .76523 L
.26372 .76709 L
.27176 .77639 L
.2721 .77681 L
.27995 .78652 L
.28147 .78847 L
.28731 .79623 L
.29118 .80158 L
.29424 .80594 L
.30078 .81566 L
.30089 .81583 L
Mistroke
.30694 .82537 L
.31061 .8314 L
.31278 .83508 L
.31831 .84479 L
.32032 .84844 L
.32356 .85451 L
.32856 .86422 L
.33003 .86716 L
.33331 .87393 L
.33787 .88364 L
.33974 .88777 L
.3422 .89336 L
.34636 .90307 L
.34946 .91053 L
.35035 .91278 L
.35417 .92249 L
.35787 .93221 L
.35917 .93568 L
.3614 .94192 L
.36482 .95163 L
.36817 .96134 L
.36888 .96346 L
.37132 .97106 L
.37443 .98077 L
Mfstroke
0 1 .16 r
.01923 .36906 m
.01981 .36888 L
.02894 .36602 L
.03866 .36287 L
.04837 .35964 L
.04978 .35917 L
.05808 .35626 L
.06779 .35275 L
.07662 .34946 L
.07751 .34911 L
.08722 .3453 L
.09693 .34133 L
.1007 .33974 L
.10664 .33718 L
.11636 .33283 L
.12238 .33003 L
.12607 .32827 L
.13578 .32349 L
.14197 .32032 L
.14549 .31846 L
.15521 .31316 L
.15972 .31061 L
.16492 .30758 L
.17463 .30167 L
.17588 .30089 L
.18434 .29542 L
.19063 .29118 L
.19406 .2888 L
.20377 .28176 L
.20416 .28147 L
.21348 .27426 L
.2166 .27176 L
.22319 .26626 L
.22807 .26204 L
.23291 .25771 L
.23869 .25233 L
.24262 .24854 L
.24854 .24262 L
.25233 .23869 L
.25771 .23291 L
.26204 .22807 L
.26626 .22319 L
.27176 .2166 L
.27426 .21348 L
.28147 .20416 L
.28176 .20377 L
.2888 .19406 L
.29118 .19063 L
.29542 .18434 L
.30089 .17588 L
.30167 .17463 L
.30758 .16492 L
.31061 .15972 L
.31316 .15521 L
.31846 .14549 L
.32032 .14197 L
.32349 .13578 L
.32827 .12607 L
.33003 .12238 L
.33283 .11636 L
.33718 .10664 L
.33974 .1007 L
.34133 .09693 L
.3453 .08722 L
.34911 .07751 L
.34946 .07662 L
.35275 .06779 L
.35626 .05808 L
.35917 .04978 L
.35964 .04837 L
.36287 .03866 L
.36602 .02894 L
.36888 .01981 L
.36906 .01923 L
.01923 .01923 L
F
0 g
.01923 .36906 m
.01981 .36888 L
.02894 .36602 L
.03866 .36287 L
.04837 .35964 L
.04978 .35917 L
.05808 .35626 L
.06779 .35275 L
.07662 .34946 L
.07751 .34911 L
.08722 .3453 L
.09693 .34133 L
.1007 .33974 L
.10664 .33718 L
.11636 .33283 L
.12238 .33003 L
.12607 .32827 L
.13578 .32349 L
.14197 .32032 L
.14549 .31846 L
.15521 .31316 L
.15972 .31061 L
.16492 .30758 L
.17463 .30167 L
.17588 .30089 L
.18434 .29542 L
.19063 .29118 L
.19406 .2888 L
.20377 .28176 L
.20416 .28147 L
.21348 .27426 L
.2166 .27176 L
.22319 .26626 L
.22807 .26204 L
.23291 .25771 L
.23869 .25233 L
.24262 .24854 L
.24854 .24262 L
.25233 .23869 L
.25771 .23291 L
.26204 .22807 L
.26626 .22319 L
.27176 .2166 L
.27426 .21348 L
.28147 .20416 L
.28176 .20377 L
.2888 .19406 L
.29118 .19063 L
.29542 .18434 L
.30089 .17588 L
Mistroke
.30167 .17463 L
.30758 .16492 L
.31061 .15972 L
.31316 .15521 L
.31846 .14549 L
.32032 .14197 L
.32349 .13578 L
.32827 .12607 L
.33003 .12238 L
.33283 .11636 L
.33718 .10664 L
.33974 .1007 L
.34133 .09693 L
.3453 .08722 L
.34911 .07751 L
.34946 .07662 L
.35275 .06779 L
.35626 .05808 L
.35917 .04978 L
.35964 .04837 L
.36287 .03866 L
.36602 .02894 L
.36888 .01981 L
.36906 .01923 L
Mfstroke
0 1 .16 r
.01923 .63094 m
.01981 .63112 L
.02894 .63398 L
.03866 .63713 L
.04837 .64036 L
.04978 .64083 L
.05808 .64374 L
.06779 .64725 L
.07662 .65054 L
.07751 .65089 L
.08722 .6547 L
.09693 .65867 L
.1007 .66026 L
.10664 .66282 L
.11636 .66717 L
.12238 .66997 L
.12607 .67173 L
.13578 .67651 L
.14197 .67968 L
.14549 .68154 L
.15521 .68684 L
.15972 .68939 L
.16492 .69242 L
.17463 .69833 L
.17588 .69911 L
.18434 .70458 L
.19063 .70882 L
.19406 .7112 L
.20377 .71824 L
.20416 .71853 L
.21348 .72574 L
.2166 .72824 L
.22319 .73374 L
.22807 .73796 L
.23291 .74229 L
.23869 .74767 L
.24262 .75146 L
.24854 .75738 L
.25233 .76131 L
.25771 .76709 L
.26204 .77193 L
.26626 .77681 L
.27176 .7834 L
.27426 .78652 L
.28147 .79584 L
.28176 .79623 L
.2888 .80594 L
.29118 .80937 L
.29542 .81566 L
.30089 .82412 L
.30167 .82537 L
.30758 .83508 L
.31061 .84028 L
.31316 .84479 L
.31846 .85451 L
.32032 .85803 L
.32349 .86422 L
.32827 .87393 L
.33003 .87762 L
.33283 .88364 L
.33718 .89336 L
.33974 .8993 L
.34133 .90307 L
.3453 .91278 L
.34911 .92249 L
.34946 .92338 L
.35275 .93221 L
.35626 .94192 L
.35917 .95022 L
.35964 .95163 L
.36287 .96134 L
.36602 .97106 L
.36888 .98019 L
.36906 .98077 L
.01923 .98077 L
F
0 g
.01923 .63094 m
.01981 .63112 L
.02894 .63398 L
.03866 .63713 L
.04837 .64036 L
.04978 .64083 L
.05808 .64374 L
.06779 .64725 L
.07662 .65054 L
.07751 .65089 L
.08722 .6547 L
.09693 .65867 L
.1007 .66026 L
.10664 .66282 L
.11636 .66717 L
.12238 .66997 L
.12607 .67173 L
.13578 .67651 L
.14197 .67968 L
.14549 .68154 L
.15521 .68684 L
.15972 .68939 L
.16492 .69242 L
.17463 .69833 L
.17588 .69911 L
.18434 .70458 L
.19063 .70882 L
.19406 .7112 L
.20377 .71824 L
.20416 .71853 L
.21348 .72574 L
.2166 .72824 L
.22319 .73374 L
.22807 .73796 L
.23291 .74229 L
.23869 .74767 L
.24262 .75146 L
.24854 .75738 L
.25233 .76131 L
.25771 .76709 L
.26204 .77193 L
.26626 .77681 L
.27176 .7834 L
.27426 .78652 L
.28147 .79584 L
.28176 .79623 L
.2888 .80594 L
.29118 .80937 L
.29542 .81566 L
.30089 .82412 L
Mistroke
.30167 .82537 L
.30758 .83508 L
.31061 .84028 L
.31316 .84479 L
.31846 .85451 L
.32032 .85803 L
.32349 .86422 L
.32827 .87393 L
.33003 .87762 L
.33283 .88364 L
.33718 .89336 L
.33974 .8993 L
.34133 .90307 L
.3453 .91278 L
.34911 .92249 L
.34946 .92338 L
.35275 .93221 L
.35626 .94192 L
.35917 .95022 L
.35964 .95163 L
.36287 .96134 L
.36602 .97106 L
.36888 .98019 L
.36906 .98077 L
Mfstroke
0 1 .4 r
.01923 .36392 m
.02894 .36095 L
.03462 .35917 L
.03866 .35787 L
.04837 .35466 L
.05808 .35132 L
.06334 .34946 L
.06779 .34784 L
.07751 .3442 L
.08722 .34041 L
.08888 .33974 L
.09693 .33644 L
.10664 .33228 L
.11172 .33003 L
.11636 .32792 L
.12607 .32334 L
.13223 .32032 L
.13578 .31853 L
.14549 .31346 L
.15074 .31061 L
.15521 .30811 L
.16492 .30246 L
.16752 .30089 L
.17463 .29648 L
.1828 .29118 L
.18434 .29015 L
.19406 .28342 L
.19677 .28147 L
.20377 .27626 L
.20958 .27176 L
.21348 .26863 L
.22138 .26204 L
.22319 .26048 L
.23227 .25233 L
.23291 .25174 L
.24237 .24262 L
.24262 .24237 L
.25174 .23291 L
.25233 .23227 L
.26048 .22319 L
.26204 .22138 L
.26863 .21348 L
.27176 .20958 L
.27626 .20377 L
.28147 .19677 L
.28342 .19406 L
.29015 .18434 L
.29118 .1828 L
.29648 .17463 L
.30089 .16752 L
.30246 .16492 L
.30811 .15521 L
.31061 .15074 L
.31346 .14549 L
.31853 .13578 L
.32032 .13223 L
.32334 .12607 L
.32792 .11636 L
.33003 .11172 L
.33228 .10664 L
.33644 .09693 L
.33974 .08888 L
.34041 .08722 L
.3442 .07751 L
.34784 .06779 L
.34946 .06334 L
.35132 .05808 L
.35466 .04837 L
.35787 .03866 L
.35917 .03462 L
.36095 .02894 L
.36392 .01923 L
.01923 .01923 L
F
0 g
.01923 .36392 m
.02894 .36095 L
.03462 .35917 L
.03866 .35787 L
.04837 .35466 L
.05808 .35132 L
.06334 .34946 L
.06779 .34784 L
.07751 .3442 L
.08722 .34041 L
.08888 .33974 L
.09693 .33644 L
.10664 .33228 L
.11172 .33003 L
.11636 .32792 L
.12607 .32334 L
.13223 .32032 L
.13578 .31853 L
.14549 .31346 L
.15074 .31061 L
.15521 .30811 L
.16492 .30246 L
.16752 .30089 L
.17463 .29648 L
.1828 .29118 L
.18434 .29015 L
.19406 .28342 L
.19677 .28147 L
.20377 .27626 L
.20958 .27176 L
.21348 .26863 L
.22138 .26204 L
.22319 .26048 L
.23227 .25233 L
.23291 .25174 L
.24237 .24262 L
.24262 .24237 L
.25174 .23291 L
.25233 .23227 L
.26048 .22319 L
.26204 .22138 L
.26863 .21348 L
.27176 .20958 L
.27626 .20377 L
.28147 .19677 L
.28342 .19406 L
.29015 .18434 L
.29118 .1828 L
.29648 .17463 L
.30089 .16752 L
Mistroke
.30246 .16492 L
.30811 .15521 L
.31061 .15074 L
.31346 .14549 L
.31853 .13578 L
.32032 .13223 L
.32334 .12607 L
.32792 .11636 L
.33003 .11172 L
.33228 .10664 L
.33644 .09693 L
.33974 .08888 L
.34041 .08722 L
.3442 .07751 L
.34784 .06779 L
.34946 .06334 L
.35132 .05808 L
.35466 .04837 L
.35787 .03866 L
.35917 .03462 L
.36095 .02894 L
.36392 .01923 L
Mfstroke
0 1 .4 r
.01923 .63608 m
.02894 .63905 L
.03462 .64083 L
.03866 .64213 L
.04837 .64534 L
.05808 .64868 L
.06334 .65054 L
.06779 .65216 L
.07751 .6558 L
.08722 .65959 L
.08888 .66026 L
.09693 .66356 L
.10664 .66772 L
.11172 .66997 L
.11636 .67208 L
.12607 .67666 L
.13223 .67968 L
.13578 .68147 L
.14549 .68654 L
.15074 .68939 L
.15521 .69189 L
.16492 .69754 L
.16752 .69911 L
.17463 .70352 L
.1828 .70882 L
.18434 .70985 L
.19406 .71658 L
.19677 .71853 L
.20377 .72374 L
.20958 .72824 L
.21348 .73137 L
.22138 .73796 L
.22319 .73952 L
.23227 .74767 L
.23291 .74826 L
.24237 .75738 L
.24262 .75763 L
.25174 .76709 L
.25233 .76773 L
.26048 .77681 L
.26204 .77862 L
.26863 .78652 L
.27176 .79042 L
.27626 .79623 L
.28147 .80323 L
.28342 .80594 L
.29015 .81566 L
.29118 .8172 L
.29648 .82537 L
.30089 .83248 L
.30246 .83508 L
.30811 .84479 L
.31061 .84926 L
.31346 .85451 L
.31853 .86422 L
.32032 .86777 L
.32334 .87393 L
.32792 .88364 L
.33003 .88828 L
.33228 .89336 L
.33644 .90307 L
.33974 .91112 L
.34041 .91278 L
.3442 .92249 L
.34784 .93221 L
.34946 .93666 L
.35132 .94192 L
.35466 .95163 L
.35787 .96134 L
.35917 .96538 L
.36095 .97106 L
.36392 .98077 L
.01923 .98077 L
F
0 g
.01923 .63608 m
.02894 .63905 L
.03462 .64083 L
.03866 .64213 L
.04837 .64534 L
.05808 .64868 L
.06334 .65054 L
.06779 .65216 L
.07751 .6558 L
.08722 .65959 L
.08888 .66026 L
.09693 .66356 L
.10664 .66772 L
.11172 .66997 L
.11636 .67208 L
.12607 .67666 L
.13223 .67968 L
.13578 .68147 L
.14549 .68654 L
.15074 .68939 L
.15521 .69189 L
.16492 .69754 L
.16752 .69911 L
.17463 .70352 L
.1828 .70882 L
.18434 .70985 L
.19406 .71658 L
.19677 .71853 L
.20377 .72374 L
.20958 .72824 L
.21348 .73137 L
.22138 .73796 L
.22319 .73952 L
.23227 .74767 L
.23291 .74826 L
.24237 .75738 L
.24262 .75763 L
.25174 .76709 L
.25233 .76773 L
.26048 .77681 L
.26204 .77862 L
.26863 .78652 L
.27176 .79042 L
.27626 .79623 L
.28147 .80323 L
.28342 .80594 L
.29015 .81566 L
.29118 .8172 L
.29648 .82537 L
.30089 .83248 L
Mistroke
.30246 .83508 L
.30811 .84479 L
.31061 .84926 L
.31346 .85451 L
.31853 .86422 L
.32032 .86777 L
.32334 .87393 L
.32792 .88364 L
.33003 .88828 L
.33228 .89336 L
.33644 .90307 L
.33974 .91112 L
.34041 .91278 L
.3442 .92249 L
.34784 .93221 L
.34946 .93666 L
.35132 .94192 L
.35466 .95163 L
.35787 .96134 L
.35917 .96538 L
.36095 .97106 L
.36392 .98077 L
Mfstroke
0 1 .64 r
.01923 .35891 m
.02894 .35599 L
.03866 .35294 L
.04837 .34976 L
.04928 .34946 L
.05808 .34645 L
.06779 .34298 L
.0765 .33974 L
.07751 .33936 L
.08722 .33557 L
.09693 .33159 L
.10063 .33003 L
.10664 .32743 L
.11636 .32305 L
.12217 .32032 L
.12607 .31844 L
.13578 .31359 L
.14151 .31061 L
.14549 .30848 L
.15521 .30307 L
.15897 .30089 L
.16492 .29736 L
.17463 .2913 L
.17481 .29118 L
.18434 .28487 L
.18925 .28147 L
.19406 .27804 L
.20247 .27176 L
.20377 .27075 L
.21348 .26298 L
.2146 .26204 L
.22319 .25466 L
.22579 .25233 L
.23291 .24573 L
.23614 .24262 L
.24262 .23614 L
.24573 .23291 L
.25233 .22579 L
.25466 .22319 L
.26204 .2146 L
.26298 .21348 L
.27075 .20377 L
.27176 .20247 L
.27804 .19406 L
.28147 .18925 L
.28487 .18434 L
.29118 .17481 L
.2913 .17463 L
.29736 .16492 L
.30089 .15897 L
.30307 .15521 L
.30848 .14549 L
.31061 .14151 L
.31359 .13578 L
.31844 .12607 L
.32032 .12217 L
.32305 .11636 L
.32743 .10664 L
.33003 .10063 L
.33159 .09693 L
.33557 .08722 L
.33936 .07751 L
.33974 .0765 L
.34298 .06779 L
.34645 .05808 L
.34946 .04928 L
.34976 .04837 L
.35294 .03866 L
.35599 .02894 L
.35891 .01923 L
.01923 .01923 L
F
0 g
.01923 .35891 m
.02894 .35599 L
.03866 .35294 L
.04837 .34976 L
.04928 .34946 L
.05808 .34645 L
.06779 .34298 L
.0765 .33974 L
.07751 .33936 L
.08722 .33557 L
.09693 .33159 L
.10063 .33003 L
.10664 .32743 L
.11636 .32305 L
.12217 .32032 L
.12607 .31844 L
.13578 .31359 L
.14151 .31061 L
.14549 .30848 L
.15521 .30307 L
.15897 .30089 L
.16492 .29736 L
.17463 .2913 L
.17481 .29118 L
.18434 .28487 L
.18925 .28147 L
.19406 .27804 L
.20247 .27176 L
.20377 .27075 L
.21348 .26298 L
.2146 .26204 L
.22319 .25466 L
.22579 .25233 L
.23291 .24573 L
.23614 .24262 L
.24262 .23614 L
.24573 .23291 L
.25233 .22579 L
.25466 .22319 L
.26204 .2146 L
.26298 .21348 L
.27075 .20377 L
.27176 .20247 L
.27804 .19406 L
.28147 .18925 L
.28487 .18434 L
.29118 .17481 L
.2913 .17463 L
.29736 .16492 L
.30089 .15897 L
Mistroke
.30307 .15521 L
.30848 .14549 L
.31061 .14151 L
.31359 .13578 L
.31844 .12607 L
.32032 .12217 L
.32305 .11636 L
.32743 .10664 L
.33003 .10063 L
.33159 .09693 L
.33557 .08722 L
.33936 .07751 L
.33974 .0765 L
.34298 .06779 L
.34645 .05808 L
.34946 .04928 L
.34976 .04837 L
.35294 .03866 L
.35599 .02894 L
.35891 .01923 L
Mfstroke
0 1 .64 r
.01923 .64109 m
.02894 .64401 L
.03866 .64706 L
.04837 .65024 L
.04928 .65054 L
.05808 .65355 L
.06779 .65702 L
.0765 .66026 L
.07751 .66064 L
.08722 .66443 L
.09693 .66841 L
.10063 .66997 L
.10664 .67257 L
.11636 .67695 L
.12217 .67968 L
.12607 .68156 L
.13578 .68641 L
.14151 .68939 L
.14549 .69152 L
.15521 .69693 L
.15897 .69911 L
.16492 .70264 L
.17463 .7087 L
.17481 .70882 L
.18434 .71513 L
.18925 .71853 L
.19406 .72196 L
.20247 .72824 L
.20377 .72925 L
.21348 .73702 L
.2146 .73796 L
.22319 .74534 L
.22579 .74767 L
.23291 .75427 L
.23614 .75738 L
.24262 .76386 L
.24573 .76709 L
.25233 .77421 L
.25466 .77681 L
.26204 .7854 L
.26298 .78652 L
.27075 .79623 L
.27176 .79753 L
.27804 .80594 L
.28147 .81075 L
.28487 .81566 L
.29118 .82519 L
.2913 .82537 L
.29736 .83508 L
.30089 .84103 L
.30307 .84479 L
.30848 .85451 L
.31061 .85849 L
.31359 .86422 L
.31844 .87393 L
.32032 .87783 L
.32305 .88364 L
.32743 .89336 L
.33003 .89937 L
.33159 .90307 L
.33557 .91278 L
.33936 .92249 L
.33974 .9235 L
.34298 .93221 L
.34645 .94192 L
.34946 .95072 L
.34976 .95163 L
.35294 .96134 L
.35599 .97106 L
.35891 .98077 L
.01923 .98077 L
F
0 g
.01923 .64109 m
.02894 .64401 L
.03866 .64706 L
.04837 .65024 L
.04928 .65054 L
.05808 .65355 L
.06779 .65702 L
.0765 .66026 L
.07751 .66064 L
.08722 .66443 L
.09693 .66841 L
.10063 .66997 L
.10664 .67257 L
.11636 .67695 L
.12217 .67968 L
.12607 .68156 L
.13578 .68641 L
.14151 .68939 L
.14549 .69152 L
.15521 .69693 L
.15897 .69911 L
.16492 .70264 L
.17463 .7087 L
.17481 .70882 L
.18434 .71513 L
.18925 .71853 L
.19406 .72196 L
.20247 .72824 L
.20377 .72925 L
.21348 .73702 L
.2146 .73796 L
.22319 .74534 L
.22579 .74767 L
.23291 .75427 L
.23614 .75738 L
.24262 .76386 L
.24573 .76709 L
.25233 .77421 L
.25466 .77681 L
.26204 .7854 L
.26298 .78652 L
.27075 .79623 L
.27176 .79753 L
.27804 .80594 L
.28147 .81075 L
.28487 .81566 L
.29118 .82519 L
.2913 .82537 L
.29736 .83508 L
.30089 .84103 L
Mistroke
.30307 .84479 L
.30848 .85451 L
.31061 .85849 L
.31359 .86422 L
.31844 .87393 L
.32032 .87783 L
.32305 .88364 L
.32743 .89336 L
.33003 .89937 L
.33159 .90307 L
.33557 .91278 L
.33936 .92249 L
.33974 .9235 L
.34298 .93221 L
.34645 .94192 L
.34946 .95072 L
.34976 .95163 L
.35294 .96134 L
.35599 .97106 L
.35891 .98077 L
Mfstroke
0 1 .88 r
.01923 .35387 m
.02894 .35097 L
.03389 .34946 L
.03866 .34797 L
.04837 .34483 L
.05808 .34153 L
.06316 .33974 L
.06779 .33809 L
.07751 .33448 L
.08722 .33068 L
.08883 .33003 L
.09693 .32671 L
.10664 .32252 L
.11156 .32032 L
.11636 .31813 L
.12607 .31349 L
.13185 .31061 L
.13578 .3086 L
.14549 .30343 L
.15007 .30089 L
.15521 .29797 L
.16492 .29218 L
.16654 .29118 L
.17463 .28604 L
.1815 .28147 L
.18434 .27952 L
.19406 .27256 L
.19514 .27176 L
.20377 .26515 L
.20765 .26204 L
.21348 .25722 L
.21916 .25233 L
.22319 .24873 L
.22977 .24262 L
.23291 .2396 L
.2396 .23291 L
.24262 .22977 L
.24873 .22319 L
.25233 .21916 L
.25722 .21348 L
.26204 .20765 L
.26515 .20377 L
.27176 .19514 L
.27256 .19406 L
.27952 .18434 L
.28147 .1815 L
.28604 .17463 L
.29118 .16654 L
.29218 .16492 L
.29797 .15521 L
.30089 .15007 L
.30343 .14549 L
.3086 .13578 L
.31061 .13185 L
.31349 .12607 L
.31813 .11636 L
.32032 .11156 L
.32252 .10664 L
.32671 .09693 L
.33003 .08883 L
.33068 .08722 L
.33448 .07751 L
.33809 .06779 L
.33974 .06316 L
.34153 .05808 L
.34483 .04837 L
.34797 .03866 L
.34946 .03389 L
.35097 .02894 L
.35387 .01923 L
.01923 .01923 L
F
0 g
.01923 .35387 m
.02894 .35097 L
.03389 .34946 L
.03866 .34797 L
.04837 .34483 L
.05808 .34153 L
.06316 .33974 L
.06779 .33809 L
.07751 .33448 L
.08722 .33068 L
.08883 .33003 L
.09693 .32671 L
.10664 .32252 L
.11156 .32032 L
.11636 .31813 L
.12607 .31349 L
.13185 .31061 L
.13578 .3086 L
.14549 .30343 L
.15007 .30089 L
.15521 .29797 L
.16492 .29218 L
.16654 .29118 L
.17463 .28604 L
.1815 .28147 L
.18434 .27952 L
.19406 .27256 L
.19514 .27176 L
.20377 .26515 L
.20765 .26204 L
.21348 .25722 L
.21916 .25233 L
.22319 .24873 L
.22977 .24262 L
.23291 .2396 L
.2396 .23291 L
.24262 .22977 L
.24873 .22319 L
.25233 .21916 L
.25722 .21348 L
.26204 .20765 L
.26515 .20377 L
.27176 .19514 L
.27256 .19406 L
.27952 .18434 L
.28147 .1815 L
.28604 .17463 L
.29118 .16654 L
.29218 .16492 L
.29797 .15521 L
Mistroke
.30089 .15007 L
.30343 .14549 L
.3086 .13578 L
.31061 .13185 L
.31349 .12607 L
.31813 .11636 L
.32032 .11156 L
.32252 .10664 L
.32671 .09693 L
.33003 .08883 L
.33068 .08722 L
.33448 .07751 L
.33809 .06779 L
.33974 .06316 L
.34153 .05808 L
.34483 .04837 L
.34797 .03866 L
.34946 .03389 L
.35097 .02894 L
.35387 .01923 L
Mfstroke
0 1 .88 r
.01923 .64613 m
.02894 .64903 L
.03389 .65054 L
.03866 .65203 L
.04837 .65517 L
.05808 .65847 L
.06316 .66026 L
.06779 .66191 L
.07751 .66552 L
.08722 .66932 L
.08883 .66997 L
.09693 .67329 L
.10664 .67748 L
.11156 .67968 L
.11636 .68187 L
.12607 .68651 L
.13185 .68939 L
.13578 .6914 L
.14549 .69657 L
.15007 .69911 L
.15521 .70203 L
.16492 .70782 L
.16654 .70882 L
.17463 .71396 L
.1815 .71853 L
.18434 .72048 L
.19406 .72744 L
.19514 .72824 L
.20377 .73485 L
.20765 .73796 L
.21348 .74278 L
.21916 .74767 L
.22319 .75127 L
.22977 .75738 L
.23291 .7604 L
.2396 .76709 L
.24262 .77023 L
.24873 .77681 L
.25233 .78084 L
.25722 .78652 L
.26204 .79235 L
.26515 .79623 L
.27176 .80486 L
.27256 .80594 L
.27952 .81566 L
.28147 .8185 L
.28604 .82537 L
.29118 .83346 L
.29218 .83508 L
.29797 .84479 L
.30089 .84993 L
.30343 .85451 L
.3086 .86422 L
.31061 .86815 L
.31349 .87393 L
.31813 .88364 L
.32032 .88844 L
.32252 .89336 L
.32671 .90307 L
.33003 .91117 L
.33068 .91278 L
.33448 .92249 L
.33809 .93221 L
.33974 .93684 L
.34153 .94192 L
.34483 .95163 L
.34797 .96134 L
.34946 .96611 L
.35097 .97106 L
.35387 .98077 L
.01923 .98077 L
F
0 g
.01923 .64613 m
.02894 .64903 L
.03389 .65054 L
.03866 .65203 L
.04837 .65517 L
.05808 .65847 L
.06316 .66026 L
.06779 .66191 L
.07751 .66552 L
.08722 .66932 L
.08883 .66997 L
.09693 .67329 L
.10664 .67748 L
.11156 .67968 L
.11636 .68187 L
.12607 .68651 L
.13185 .68939 L
.13578 .6914 L
.14549 .69657 L
.15007 .69911 L
.15521 .70203 L
.16492 .70782 L
.16654 .70882 L
.17463 .71396 L
.1815 .71853 L
.18434 .72048 L
.19406 .72744 L
.19514 .72824 L
.20377 .73485 L
.20765 .73796 L
.21348 .74278 L
.21916 .74767 L
.22319 .75127 L
.22977 .75738 L
.23291 .7604 L
.2396 .76709 L
.24262 .77023 L
.24873 .77681 L
.25233 .78084 L
.25722 .78652 L
.26204 .79235 L
.26515 .79623 L
.27176 .80486 L
.27256 .80594 L
.27952 .81566 L
.28147 .8185 L
.28604 .82537 L
.29118 .83346 L
.29218 .83508 L
.29797 .84479 L
Mistroke
.30089 .84993 L
.30343 .85451 L
.3086 .86422 L
.31061 .86815 L
.31349 .87393 L
.31813 .88364 L
.32032 .88844 L
.32252 .89336 L
.32671 .90307 L
.33003 .91117 L
.33068 .91278 L
.33448 .92249 L
.33809 .93221 L
.33974 .93684 L
.34153 .94192 L
.34483 .95163 L
.34797 .96134 L
.34946 .96611 L
.35097 .97106 L
.35387 .98077 L
Mfstroke
0 .88 1 r
.01923 .34862 m
.02894 .34581 L
.03866 .34283 L
.04825 .33974 L
.04837 .33971 L
.05808 .33647 L
.06779 .33304 L
.0759 .33003 L
.07751 .32943 L
.08722 .32566 L
.09693 .32167 L
.1001 .32032 L
.10664 .31748 L
.11636 .31306 L
.12151 .31061 L
.12607 .3084 L
.13578 .30346 L
.14062 .30089 L
.14549 .29825 L
.15521 .29272 L
.1578 .29118 L
.16492 .28686 L
.17335 .28147 L
.17463 .28063 L
.18434 .274 L
.18749 .27176 L
.19406 .26693 L
.20041 .26204 L
.20377 .25937 L
.21226 .25233 L
.21348 .25128 L
.22317 .24262 L
.22319 .24259 L
.23291 .23325 L
.23325 .23291 L
.24259 .22319 L
.24262 .22317 L
.25128 .21348 L
.25233 .21226 L
.25937 .20377 L
.26204 .20041 L
.26693 .19406 L
.27176 .18749 L
.274 .18434 L
.28063 .17463 L
.28147 .17335 L
.28686 .16492 L
.29118 .1578 L
.29272 .15521 L
.29825 .14549 L
.30089 .14062 L
.30346 .13578 L
.3084 .12607 L
.31061 .12151 L
.31306 .11636 L
.31748 .10664 L
.32032 .1001 L
.32167 .09693 L
.32566 .08722 L
.32943 .07751 L
.33003 .0759 L
.33304 .06779 L
.33647 .05808 L
.33971 .04837 L
.33974 .04825 L
.34283 .03866 L
.34581 .02894 L
.34862 .01923 L
.01923 .01923 L
F
0 g
.01923 .34862 m
.02894 .34581 L
.03866 .34283 L
.04825 .33974 L
.04837 .33971 L
.05808 .33647 L
.06779 .33304 L
.0759 .33003 L
.07751 .32943 L
.08722 .32566 L
.09693 .32167 L
.1001 .32032 L
.10664 .31748 L
.11636 .31306 L
.12151 .31061 L
.12607 .3084 L
.13578 .30346 L
.14062 .30089 L
.14549 .29825 L
.15521 .29272 L
.1578 .29118 L
.16492 .28686 L
.17335 .28147 L
.17463 .28063 L
.18434 .274 L
.18749 .27176 L
.19406 .26693 L
.20041 .26204 L
.20377 .25937 L
.21226 .25233 L
.21348 .25128 L
.22317 .24262 L
.22319 .24259 L
.23291 .23325 L
.23325 .23291 L
.24259 .22319 L
.24262 .22317 L
.25128 .21348 L
.25233 .21226 L
.25937 .20377 L
.26204 .20041 L
.26693 .19406 L
.27176 .18749 L
.274 .18434 L
.28063 .17463 L
.28147 .17335 L
.28686 .16492 L
.29118 .1578 L
.29272 .15521 L
.29825 .14549 L
Mistroke
.30089 .14062 L
.30346 .13578 L
.3084 .12607 L
.31061 .12151 L
.31306 .11636 L
.31748 .10664 L
.32032 .1001 L
.32167 .09693 L
.32566 .08722 L
.32943 .07751 L
.33003 .0759 L
.33304 .06779 L
.33647 .05808 L
.33971 .04837 L
.33974 .04825 L
.34283 .03866 L
.34581 .02894 L
.34862 .01923 L
Mfstroke
0 .88 1 r
.01923 .65138 m
.02894 .65419 L
.03866 .65717 L
.04825 .66026 L
.04837 .66029 L
.05808 .66353 L
.06779 .66696 L
.0759 .66997 L
.07751 .67057 L
.08722 .67434 L
.09693 .67833 L
.1001 .67968 L
.10664 .68252 L
.11636 .68694 L
.12151 .68939 L
.12607 .6916 L
.13578 .69654 L
.14062 .69911 L
.14549 .70175 L
.15521 .70728 L
.1578 .70882 L
.16492 .71314 L
.17335 .71853 L
.17463 .71937 L
.18434 .726 L
.18749 .72824 L
.19406 .73307 L
.20041 .73796 L
.20377 .74063 L
.21226 .74767 L
.21348 .74872 L
.22317 .75738 L
.22319 .75741 L
.23291 .76675 L
.23325 .76709 L
.24259 .77681 L
.24262 .77683 L
.25128 .78652 L
.25233 .78774 L
.25937 .79623 L
.26204 .79959 L
.26693 .80594 L
.27176 .81251 L
.274 .81566 L
.28063 .82537 L
.28147 .82665 L
.28686 .83508 L
.29118 .8422 L
.29272 .84479 L
.29825 .85451 L
.30089 .85938 L
.30346 .86422 L
.3084 .87393 L
.31061 .87849 L
.31306 .88364 L
.31748 .89336 L
.32032 .8999 L
.32167 .90307 L
.32566 .91278 L
.32943 .92249 L
.33003 .9241 L
.33304 .93221 L
.33647 .94192 L
.33971 .95163 L
.33974 .95175 L
.34283 .96134 L
.34581 .97106 L
.34862 .98077 L
.01923 .98077 L
F
0 g
.01923 .65138 m
.02894 .65419 L
.03866 .65717 L
.04825 .66026 L
.04837 .66029 L
.05808 .66353 L
.06779 .66696 L
.0759 .66997 L
.07751 .67057 L
.08722 .67434 L
.09693 .67833 L
.1001 .67968 L
.10664 .68252 L
.11636 .68694 L
.12151 .68939 L
.12607 .6916 L
.13578 .69654 L
.14062 .69911 L
.14549 .70175 L
.15521 .70728 L
.1578 .70882 L
.16492 .71314 L
.17335 .71853 L
.17463 .71937 L
.18434 .726 L
.18749 .72824 L
.19406 .73307 L
.20041 .73796 L
.20377 .74063 L
.21226 .74767 L
.21348 .74872 L
.22317 .75738 L
.22319 .75741 L
.23291 .76675 L
.23325 .76709 L
.24259 .77681 L
.24262 .77683 L
.25128 .78652 L
.25233 .78774 L
.25937 .79623 L
.26204 .79959 L
.26693 .80594 L
.27176 .81251 L
.274 .81566 L
.28063 .82537 L
.28147 .82665 L
.28686 .83508 L
.29118 .8422 L
.29272 .84479 L
.29825 .85451 L
Mistroke
.30089 .85938 L
.30346 .86422 L
.3084 .87393 L
.31061 .87849 L
.31306 .88364 L
.31748 .89336 L
.32032 .8999 L
.32167 .90307 L
.32566 .91278 L
.32943 .92249 L
.33003 .9241 L
.33304 .93221 L
.33647 .94192 L
.33971 .95163 L
.33974 .95175 L
.34283 .96134 L
.34581 .97106 L
.34862 .98077 L
Mfstroke
0 .64 1 r
.01923 .343 m
.02894 .3402 L
.03051 .33974 L
.03866 .33737 L
.04837 .33431 L
.05808 .33105 L
.06104 .33003 L
.06779 .3277 L
.07751 .32412 L
.08722 .32032 L
.08722 .32032 L
.09693 .31638 L
.10664 .31216 L
.11009 .31061 L
.11636 .30774 L
.12607 .30304 L
.13031 .30089 L
.13578 .29808 L
.14549 .2928 L
.14836 .29118 L
.15521 .28723 L
.1646 .28147 L
.16492 .28127 L
.17463 .27496 L
.17932 .27176 L
.18434 .26822 L
.1927 .26204 L
.19406 .26101 L
.20377 .2533 L
.20494 .25233 L
.21348 .24504 L
.21619 .24262 L
.22319 .23615 L
.22656 .23291 L
.23291 .22656 L
.23615 .22319 L
.24262 .21619 L
.24504 .21348 L
.25233 .20494 L
.2533 .20377 L
.26101 .19406 L
.26204 .1927 L
.26822 .18434 L
.27176 .17932 L
.27496 .17463 L
.28127 .16492 L
.28147 .1646 L
.28723 .15521 L
.29118 .14836 L
.2928 .14549 L
.29808 .13578 L
.30089 .13031 L
.30304 .12607 L
.30774 .11636 L
.31061 .11009 L
.31216 .10664 L
.31638 .09693 L
.32032 .08722 L
.32032 .08722 L
.32412 .07751 L
.3277 .06779 L
.33003 .06104 L
.33105 .05808 L
.33431 .04837 L
.33737 .03866 L
.33974 .03051 L
.3402 .02894 L
.343 .01923 L
.01923 .01923 L
F
0 g
.01923 .343 m
.02894 .3402 L
.03051 .33974 L
.03866 .33737 L
.04837 .33431 L
.05808 .33105 L
.06104 .33003 L
.06779 .3277 L
.07751 .32412 L
.08722 .32032 L
.08722 .32032 L
.09693 .31638 L
.10664 .31216 L
.11009 .31061 L
.11636 .30774 L
.12607 .30304 L
.13031 .30089 L
.13578 .29808 L
.14549 .2928 L
.14836 .29118 L
.15521 .28723 L
.1646 .28147 L
.16492 .28127 L
.17463 .27496 L
.17932 .27176 L
.18434 .26822 L
.1927 .26204 L
.19406 .26101 L
.20377 .2533 L
.20494 .25233 L
.21348 .24504 L
.21619 .24262 L
.22319 .23615 L
.22656 .23291 L
.23291 .22656 L
.23615 .22319 L
.24262 .21619 L
.24504 .21348 L
.25233 .20494 L
.2533 .20377 L
.26101 .19406 L
.26204 .1927 L
.26822 .18434 L
.27176 .17932 L
.27496 .17463 L
.28127 .16492 L
.28147 .1646 L
.28723 .15521 L
.29118 .14836 L
.2928 .14549 L
Mistroke
.29808 .13578 L
.30089 .13031 L
.30304 .12607 L
.30774 .11636 L
.31061 .11009 L
.31216 .10664 L
.31638 .09693 L
.32032 .08722 L
.32032 .08722 L
.32412 .07751 L
.3277 .06779 L
.33003 .06104 L
.33105 .05808 L
.33431 .04837 L
.33737 .03866 L
.33974 .03051 L
.3402 .02894 L
.343 .01923 L
Mfstroke
0 .64 1 r
.01923 .657 m
.02894 .6598 L
.03051 .66026 L
.03866 .66263 L
.04837 .66569 L
.05808 .66895 L
.06104 .66997 L
.06779 .6723 L
.07751 .67588 L
.08722 .67968 L
.08722 .67968 L
.09693 .68362 L
.10664 .68784 L
.11009 .68939 L
.11636 .69226 L
.12607 .69696 L
.13031 .69911 L
.13578 .70192 L
.14549 .7072 L
.14836 .70882 L
.15521 .71277 L
.1646 .71853 L
.16492 .71873 L
.17463 .72504 L
.17932 .72824 L
.18434 .73178 L
.1927 .73796 L
.19406 .73899 L
.20377 .7467 L
.20494 .74767 L
.21348 .75496 L
.21619 .75738 L
.22319 .76385 L
.22656 .76709 L
.23291 .77344 L
.23615 .77681 L
.24262 .78381 L
.24504 .78652 L
.25233 .79506 L
.2533 .79623 L
.26101 .80594 L
.26204 .8073 L
.26822 .81566 L
.27176 .82068 L
.27496 .82537 L
.28127 .83508 L
.28147 .8354 L
.28723 .84479 L
.29118 .85164 L
.2928 .85451 L
.29808 .86422 L
.30089 .86969 L
.30304 .87393 L
.30774 .88364 L
.31061 .88991 L
.31216 .89336 L
.31638 .90307 L
.32032 .91278 L
.32032 .91278 L
.32412 .92249 L
.3277 .93221 L
.33003 .93896 L
.33105 .94192 L
.33431 .95163 L
.33737 .96134 L
.33974 .96949 L
.3402 .97106 L
.343 .98077 L
.01923 .98077 L
F
0 g
.01923 .657 m
.02894 .6598 L
.03051 .66026 L
.03866 .66263 L
.04837 .66569 L
.05808 .66895 L
.06104 .66997 L
.06779 .6723 L
.07751 .67588 L
.08722 .67968 L
.08722 .67968 L
.09693 .68362 L
.10664 .68784 L
.11009 .68939 L
.11636 .69226 L
.12607 .69696 L
.13031 .69911 L
.13578 .70192 L
.14549 .7072 L
.14836 .70882 L
.15521 .71277 L
.1646 .71853 L
.16492 .71873 L
.17463 .72504 L
.17932 .72824 L
.18434 .73178 L
.1927 .73796 L
.19406 .73899 L
.20377 .7467 L
.20494 .74767 L
.21348 .75496 L
.21619 .75738 L
.22319 .76385 L
.22656 .76709 L
.23291 .77344 L
.23615 .77681 L
.24262 .78381 L
.24504 .78652 L
.25233 .79506 L
.2533 .79623 L
.26101 .80594 L
.26204 .8073 L
.26822 .81566 L
.27176 .82068 L
.27496 .82537 L
.28127 .83508 L
.28147 .8354 L
.28723 .84479 L
.29118 .85164 L
.2928 .85451 L
Mistroke
.29808 .86422 L
.30089 .86969 L
.30304 .87393 L
.30774 .88364 L
.31061 .88991 L
.31216 .89336 L
.31638 .90307 L
.32032 .91278 L
.32032 .91278 L
.32412 .92249 L
.3277 .93221 L
.33003 .93896 L
.33105 .94192 L
.33431 .95163 L
.33737 .96134 L
.33974 .96949 L
.3402 .97106 L
.343 .98077 L
Mfstroke
0 .4 1 r
.01923 .33674 m
.02894 .33403 L
.03866 .33115 L
.04234 .33003 L
.04837 .3283 L
.05808 .32516 L
.06779 .32177 L
.07181 .32032 L
.07751 .31831 L
.08722 .31457 L
.09681 .31061 L
.09693 .31056 L
.10664 .30644 L
.11636 .30195 L
.11858 .30089 L
.12607 .29729 L
.13578 .29225 L
.13778 .29118 L
.14549 .28698 L
.15491 .28147 L
.15521 .28129 L
.16492 .2753 L
.17033 .27176 L
.17463 .26888 L
.18428 .26204 L
.18434 .262 L
.19406 .25467 L
.19701 .25233 L
.20377 .2468 L
.20865 .24262 L
.21348 .23834 L
.21935 .23291 L
.22319 .22921 L
.22921 .22319 L
.23291 .21935 L
.23834 .21348 L
.24262 .20865 L
.2468 .20377 L
.25233 .19701 L
.25467 .19406 L
.262 .18434 L
.26204 .18428 L
.26888 .17463 L
.27176 .17033 L
.2753 .16492 L
.28129 .15521 L
.28147 .15491 L
.28698 .14549 L
.29118 .13778 L
.29225 .13578 L
.29729 .12607 L
.30089 .11858 L
.30195 .11636 L
.30644 .10664 L
.31056 .09693 L
.31061 .09681 L
.31457 .08722 L
.31831 .07751 L
.32032 .07181 L
.32177 .06779 L
.32516 .05808 L
.3283 .04837 L
.33003 .04234 L
.33115 .03866 L
.33403 .02894 L
.33674 .01923 L
.01923 .01923 L
F
0 g
.01923 .33674 m
.02894 .33403 L
.03866 .33115 L
.04234 .33003 L
.04837 .3283 L
.05808 .32516 L
.06779 .32177 L
.07181 .32032 L
.07751 .31831 L
.08722 .31457 L
.09681 .31061 L
.09693 .31056 L
.10664 .30644 L
.11636 .30195 L
.11858 .30089 L
.12607 .29729 L
.13578 .29225 L
.13778 .29118 L
.14549 .28698 L
.15491 .28147 L
.15521 .28129 L
.16492 .2753 L
.17033 .27176 L
.17463 .26888 L
.18428 .26204 L
.18434 .262 L
.19406 .25467 L
.19701 .25233 L
.20377 .2468 L
.20865 .24262 L
.21348 .23834 L
.21935 .23291 L
.22319 .22921 L
.22921 .22319 L
.23291 .21935 L
.23834 .21348 L
.24262 .20865 L
.2468 .20377 L
.25233 .19701 L
.25467 .19406 L
.262 .18434 L
.26204 .18428 L
.26888 .17463 L
.27176 .17033 L
.2753 .16492 L
.28129 .15521 L
.28147 .15491 L
.28698 .14549 L
.29118 .13778 L
.29225 .13578 L
Mistroke
.29729 .12607 L
.30089 .11858 L
.30195 .11636 L
.30644 .10664 L
.31056 .09693 L
.31061 .09681 L
.31457 .08722 L
.31831 .07751 L
.32032 .07181 L
.32177 .06779 L
.32516 .05808 L
.3283 .04837 L
.33003 .04234 L
.33115 .03866 L
.33403 .02894 L
.33674 .01923 L
Mfstroke
0 .4 1 r
.01923 .66326 m
.02894 .66597 L
.03866 .66885 L
.04234 .66997 L
.04837 .6717 L
.05808 .67484 L
.06779 .67823 L
.07181 .67968 L
.07751 .68169 L
.08722 .68543 L
.09681 .68939 L
.09693 .68944 L
.10664 .69356 L
.11636 .69805 L
.11858 .69911 L
.12607 .70271 L
.13578 .70775 L
.13778 .70882 L
.14549 .71302 L
.15491 .71853 L
.15521 .71871 L
.16492 .7247 L
.17033 .72824 L
.17463 .73112 L
.18428 .73796 L
.18434 .738 L
.19406 .74533 L
.19701 .74767 L
.20377 .7532 L
.20865 .75738 L
.21348 .76166 L
.21935 .76709 L
.22319 .77079 L
.22921 .77681 L
.23291 .78065 L
.23834 .78652 L
.24262 .79135 L
.2468 .79623 L
.25233 .80299 L
.25467 .80594 L
.262 .81566 L
.26204 .81572 L
.26888 .82537 L
.27176 .82967 L
.2753 .83508 L
.28129 .84479 L
.28147 .84509 L
.28698 .85451 L
.29118 .86222 L
.29225 .86422 L
.29729 .87393 L
.30089 .88142 L
.30195 .88364 L
.30644 .89336 L
.31056 .90307 L
.31061 .90319 L
.31457 .91278 L
.31831 .92249 L
.32032 .92819 L
.32177 .93221 L
.32516 .94192 L
.3283 .95163 L
.33003 .95766 L
.33115 .96134 L
.33403 .97106 L
.33674 .98077 L
.01923 .98077 L
F
0 g
.01923 .66326 m
.02894 .66597 L
.03866 .66885 L
.04234 .66997 L
.04837 .6717 L
.05808 .67484 L
.06779 .67823 L
.07181 .67968 L
.07751 .68169 L
.08722 .68543 L
.09681 .68939 L
.09693 .68944 L
.10664 .69356 L
.11636 .69805 L
.11858 .69911 L
.12607 .70271 L
.13578 .70775 L
.13778 .70882 L
.14549 .71302 L
.15491 .71853 L
.15521 .71871 L
.16492 .7247 L
.17033 .72824 L
.17463 .73112 L
.18428 .73796 L
.18434 .738 L
.19406 .74533 L
.19701 .74767 L
.20377 .7532 L
.20865 .75738 L
.21348 .76166 L
.21935 .76709 L
.22319 .77079 L
.22921 .77681 L
.23291 .78065 L
.23834 .78652 L
.24262 .79135 L
.2468 .79623 L
.25233 .80299 L
.25467 .80594 L
.262 .81566 L
.26204 .81572 L
.26888 .82537 L
.27176 .82967 L
.2753 .83508 L
.28129 .84479 L
.28147 .84509 L
.28698 .85451 L
.29118 .86222 L
.29225 .86422 L
Mistroke
.29729 .87393 L
.30089 .88142 L
.30195 .88364 L
.30644 .89336 L
.31056 .90307 L
.31061 .90319 L
.31457 .91278 L
.31831 .92249 L
.32032 .92819 L
.32177 .93221 L
.32516 .94192 L
.3283 .95163 L
.33003 .95766 L
.33115 .96134 L
.33403 .97106 L
.33674 .98077 L
Mfstroke
0 .16 1 r
.01923 .32811 m
.02894 .32694 L
.03866 .32385 L
.04837 .32086 L
.05015 .32032 L
.05808 .31833 L
.06779 .31501 L
.07751 .31144 L
.07974 .31061 L
.08722 .308 L
.09693 .30403 L
.10419 .30089 L
.10664 .2999 L
.11636 .29552 L
.12517 .29118 L
.12607 .29076 L
.13578 .28581 L
.14361 .28147 L
.14549 .28044 L
.15521 .27474 L
.16001 .27176 L
.16492 .26867 L
.17463 .2621 L
.17471 .26204 L
.18434 .25514 L
.18806 .25233 L
.19406 .24766 L
.20019 .24262 L
.20377 .23959 L
.21129 .23291 L
.21348 .23089 L
.22148 .22319 L
.22319 .22148 L
.23089 .21348 L
.23291 .21129 L
.23959 .20377 L
.24262 .20019 L
.24766 .19406 L
.25233 .18806 L
.25514 .18434 L
.26204 .17471 L
.2621 .17463 L
.26867 .16492 L
.27176 .16001 L
.27474 .15521 L
.28044 .14549 L
.28147 .14361 L
.28581 .13578 L
.29076 .12607 L
.29118 .12517 L
.29552 .11636 L
.2999 .10664 L
.30089 .10419 L
.30403 .09693 L
.308 .08722 L
.31061 .07974 L
.31144 .07751 L
.31501 .06779 L
.31833 .05808 L
.32032 .05015 L
.32086 .04837 L
.32385 .03866 L
.32694 .02894 L
.32811 .01923 L
.01923 .01923 L
F
0 g
.01923 .32811 m
.02894 .32694 L
.03866 .32385 L
.04837 .32086 L
.05015 .32032 L
.05808 .31833 L
.06779 .31501 L
.07751 .31144 L
.07974 .31061 L
.08722 .308 L
.09693 .30403 L
.10419 .30089 L
.10664 .2999 L
.11636 .29552 L
.12517 .29118 L
.12607 .29076 L
.13578 .28581 L
.14361 .28147 L
.14549 .28044 L
.15521 .27474 L
.16001 .27176 L
.16492 .26867 L
.17463 .2621 L
.17471 .26204 L
.18434 .25514 L
.18806 .25233 L
.19406 .24766 L
.20019 .24262 L
.20377 .23959 L
.21129 .23291 L
.21348 .23089 L
.22148 .22319 L
.22319 .22148 L
.23089 .21348 L
.23291 .21129 L
.23959 .20377 L
.24262 .20019 L
.24766 .19406 L
.25233 .18806 L
.25514 .18434 L
.26204 .17471 L
.2621 .17463 L
.26867 .16492 L
.27176 .16001 L
.27474 .15521 L
.28044 .14549 L
.28147 .14361 L
.28581 .13578 L
.29076 .12607 L
.29118 .12517 L
Mistroke
.29552 .11636 L
.2999 .10664 L
.30089 .10419 L
.30403 .09693 L
.308 .08722 L
.31061 .07974 L
.31144 .07751 L
.31501 .06779 L
.31833 .05808 L
.32032 .05015 L
.32086 .04837 L
.32385 .03866 L
.32694 .02894 L
.32811 .01923 L
Mfstroke
0 .16 1 r
.01923 .67189 m
.02894 .67306 L
.03866 .67615 L
.04837 .67914 L
.05015 .67968 L
.05808 .68167 L
.06779 .68499 L
.07751 .68856 L
.07974 .68939 L
.08722 .692 L
.09693 .69597 L
.10419 .69911 L
.10664 .7001 L
.11636 .70448 L
.12517 .70882 L
.12607 .70924 L
.13578 .71419 L
.14361 .71853 L
.14549 .71956 L
.15521 .72526 L
.16001 .72824 L
.16492 .73133 L
.17463 .7379 L
.17471 .73796 L
.18434 .74486 L
.18806 .74767 L
.19406 .75234 L
.20019 .75738 L
.20377 .76041 L
.21129 .76709 L
.21348 .76911 L
.22148 .77681 L
.22319 .77852 L
.23089 .78652 L
.23291 .78871 L
.23959 .79623 L
.24262 .79981 L
.24766 .80594 L
.25233 .81194 L
.25514 .81566 L
.26204 .82529 L
.2621 .82537 L
.26867 .83508 L
.27176 .83999 L
.27474 .84479 L
.28044 .85451 L
.28147 .85639 L
.28581 .86422 L
.29076 .87393 L
.29118 .87483 L
.29552 .88364 L
.2999 .89336 L
.30089 .89581 L
.30403 .90307 L
.308 .91278 L
.31061 .92026 L
.31144 .92249 L
.31501 .93221 L
.31833 .94192 L
.32032 .94985 L
.32086 .95163 L
.32385 .96134 L
.32694 .97106 L
.32811 .98077 L
.01923 .98077 L
F
0 g
.01923 .67189 m
.02894 .67306 L
.03866 .67615 L
.04837 .67914 L
.05015 .67968 L
.05808 .68167 L
.06779 .68499 L
.07751 .68856 L
.07974 .68939 L
.08722 .692 L
.09693 .69597 L
.10419 .69911 L
.10664 .7001 L
.11636 .70448 L
.12517 .70882 L
.12607 .70924 L
.13578 .71419 L
.14361 .71853 L
.14549 .71956 L
.15521 .72526 L
.16001 .72824 L
.16492 .73133 L
.17463 .7379 L
.17471 .73796 L
.18434 .74486 L
.18806 .74767 L
.19406 .75234 L
.20019 .75738 L
.20377 .76041 L
.21129 .76709 L
.21348 .76911 L
.22148 .77681 L
.22319 .77852 L
.23089 .78652 L
.23291 .78871 L
.23959 .79623 L
.24262 .79981 L
.24766 .80594 L
.25233 .81194 L
.25514 .81566 L
.26204 .82529 L
.2621 .82537 L
.26867 .83508 L
.27176 .83999 L
.27474 .84479 L
.28044 .85451 L
.28147 .85639 L
.28581 .86422 L
.29076 .87393 L
.29118 .87483 L
Mistroke
.29552 .88364 L
.2999 .89336 L
.30089 .89581 L
.30403 .90307 L
.308 .91278 L
.31061 .92026 L
.31144 .92249 L
.31501 .93221 L
.31833 .94192 L
.32032 .94985 L
.32086 .95163 L
.32385 .96134 L
.32694 .97106 L
.32811 .98077 L
Mfstroke
0 .4 1 r
.01923 .2954 m
.02845 .29118 L
.02894 .29087 L
.03866 .28471 L
.04526 .28147 L
.04837 .27952 L
.05808 .27351 L
.06149 .27176 L
.06779 .2677 L
.07703 .26204 L
.07751 .26173 L
.08722 .25523 L
.09179 .25233 L
.09693 .24875 L
.10582 .24262 L
.10664 .24202 L
.11636 .23488 L
.11909 .23291 L
.12607 .22753 L
.13168 .22319 L
.13578 .21988 L
.14361 .21348 L
.14549 .21189 L
.15493 .20377 L
.15521 .20352 L
.16492 .19472 L
.16564 .19406 L
.17463 .1855 L
.17582 .18434 L
.18434 .17582 L
.1855 .17463 L
.19406 .16564 L
.19472 .16492 L
.20352 .15521 L
.20377 .15493 L
.21189 .14549 L
.21348 .14361 L
.21988 .13578 L
.22319 .13168 L
.22753 .12607 L
.23291 .11909 L
.23488 .11636 L
.24202 .10664 L
.24262 .10582 L
.24875 .09693 L
.25233 .09179 L
.25523 .08722 L
.26173 .07751 L
.26204 .07703 L
.2677 .06779 L
.27176 .06149 L
.27351 .05808 L
.27952 .04837 L
.28147 .04526 L
.28471 .03866 L
.29087 .02894 L
.29118 .02845 L
.2954 .01923 L
.01923 .01923 L
F
0 g
.01923 .2954 m
.02845 .29118 L
.02894 .29087 L
.03866 .28471 L
.04526 .28147 L
.04837 .27952 L
.05808 .27351 L
.06149 .27176 L
.06779 .2677 L
.07703 .26204 L
.07751 .26173 L
.08722 .25523 L
.09179 .25233 L
.09693 .24875 L
.10582 .24262 L
.10664 .24202 L
.11636 .23488 L
.11909 .23291 L
.12607 .22753 L
.13168 .22319 L
.13578 .21988 L
.14361 .21348 L
.14549 .21189 L
.15493 .20377 L
.15521 .20352 L
.16492 .19472 L
.16564 .19406 L
.17463 .1855 L
.17582 .18434 L
.18434 .17582 L
.1855 .17463 L
.19406 .16564 L
.19472 .16492 L
.20352 .15521 L
.20377 .15493 L
.21189 .14549 L
.21348 .14361 L
.21988 .13578 L
.22319 .13168 L
.22753 .12607 L
.23291 .11909 L
.23488 .11636 L
.24202 .10664 L
.24262 .10582 L
.24875 .09693 L
.25233 .09179 L
.25523 .08722 L
.26173 .07751 L
.26204 .07703 L
.2677 .06779 L
Mistroke
.27176 .06149 L
.27351 .05808 L
.27952 .04837 L
.28147 .04526 L
.28471 .03866 L
.29087 .02894 L
.29118 .02845 L
.2954 .01923 L
Mfstroke
0 .4 1 r
.01923 .7046 m
.02845 .70882 L
.02894 .70913 L
.03866 .71529 L
.04526 .71853 L
.04837 .72048 L
.05808 .72649 L
.06149 .72824 L
.06779 .7323 L
.07703 .73796 L
.07751 .73827 L
.08722 .74477 L
.09179 .74767 L
.09693 .75125 L
.10582 .75738 L
.10664 .75798 L
.11636 .76512 L
.11909 .76709 L
.12607 .77247 L
.13168 .77681 L
.13578 .78012 L
.14361 .78652 L
.14549 .78811 L
.15493 .79623 L
.15521 .79648 L
.16492 .80528 L
.16564 .80594 L
.17463 .8145 L
.17582 .81566 L
.18434 .82418 L
.1855 .82537 L
.19406 .83436 L
.19472 .83508 L
.20352 .84479 L
.20377 .84507 L
.21189 .85451 L
.21348 .85639 L
.21988 .86422 L
.22319 .86832 L
.22753 .87393 L
.23291 .88091 L
.23488 .88364 L
.24202 .89336 L
.24262 .89418 L
.24875 .90307 L
.25233 .90821 L
.25523 .91278 L
.26173 .92249 L
.26204 .92297 L
.2677 .93221 L
.27176 .93851 L
.27351 .94192 L
.27952 .95163 L
.28147 .95474 L
.28471 .96134 L
.29087 .97106 L
.29118 .97155 L
.2954 .98077 L
.01923 .98077 L
F
0 g
.01923 .7046 m
.02845 .70882 L
.02894 .70913 L
.03866 .71529 L
.04526 .71853 L
.04837 .72048 L
.05808 .72649 L
.06149 .72824 L
.06779 .7323 L
.07703 .73796 L
.07751 .73827 L
.08722 .74477 L
.09179 .74767 L
.09693 .75125 L
.10582 .75738 L
.10664 .75798 L
.11636 .76512 L
.11909 .76709 L
.12607 .77247 L
.13168 .77681 L
.13578 .78012 L
.14361 .78652 L
.14549 .78811 L
.15493 .79623 L
.15521 .79648 L
.16492 .80528 L
.16564 .80594 L
.17463 .8145 L
.17582 .81566 L
.18434 .82418 L
.1855 .82537 L
.19406 .83436 L
.19472 .83508 L
.20352 .84479 L
.20377 .84507 L
.21189 .85451 L
.21348 .85639 L
.21988 .86422 L
.22319 .86832 L
.22753 .87393 L
.23291 .88091 L
.23488 .88364 L
.24202 .89336 L
.24262 .89418 L
.24875 .90307 L
.25233 .90821 L
.25523 .91278 L
.26173 .92249 L
.26204 .92297 L
.2677 .93221 L
Mistroke
.27176 .93851 L
.27351 .94192 L
.27952 .95163 L
.28147 .95474 L
.28471 .96134 L
.29087 .97106 L
.29118 .97155 L
.2954 .98077 L
Mfstroke
0 .64 1 r
.01923 .2876 m
.02894 .28251 L
.03109 .28147 L
.03866 .27725 L
.04837 .27191 L
.04867 .27176 L
.05808 .26622 L
.06526 .26204 L
.06779 .26048 L
.07751 .25443 L
.08089 .25233 L
.08722 .24819 L
.09562 .24262 L
.09693 .24172 L
.10664 .23492 L
.10949 .23291 L
.11636 .22785 L
.12257 .22319 L
.12607 .22048 L
.13491 .21348 L
.13578 .21277 L
.14549 .20467 L
.14656 .20377 L
.15521 .19618 L
.15757 .19406 L
.16492 .18726 L
.16799 .18434 L
.17463 .17788 L
.17788 .17463 L
.18434 .16799 L
.18726 .16492 L
.19406 .15757 L
.19618 .15521 L
.20377 .14656 L
.20467 .14549 L
.21277 .13578 L
.21348 .13491 L
.22048 .12607 L
.22319 .12257 L
.22785 .11636 L
.23291 .10949 L
.23492 .10664 L
.24172 .09693 L
.24262 .09562 L
.24819 .08722 L
.25233 .08089 L
.25443 .07751 L
.26048 .06779 L
.26204 .06526 L
.26622 .05808 L
.27176 .04867 L
.27191 .04837 L
.27725 .03866 L
.28147 .03109 L
.28251 .02894 L
.2876 .01923 L
.01923 .01923 L
F
0 g
.01923 .2876 m
.02894 .28251 L
.03109 .28147 L
.03866 .27725 L
.04837 .27191 L
.04867 .27176 L
.05808 .26622 L
.06526 .26204 L
.06779 .26048 L
.07751 .25443 L
.08089 .25233 L
.08722 .24819 L
.09562 .24262 L
.09693 .24172 L
.10664 .23492 L
.10949 .23291 L
.11636 .22785 L
.12257 .22319 L
.12607 .22048 L
.13491 .21348 L
.13578 .21277 L
.14549 .20467 L
.14656 .20377 L
.15521 .19618 L
.15757 .19406 L
.16492 .18726 L
.16799 .18434 L
.17463 .17788 L
.17788 .17463 L
.18434 .16799 L
.18726 .16492 L
.19406 .15757 L
.19618 .15521 L
.20377 .14656 L
.20467 .14549 L
.21277 .13578 L
.21348 .13491 L
.22048 .12607 L
.22319 .12257 L
.22785 .11636 L
.23291 .10949 L
.23492 .10664 L
.24172 .09693 L
.24262 .09562 L
.24819 .08722 L
.25233 .08089 L
.25443 .07751 L
.26048 .06779 L
.26204 .06526 L
.26622 .05808 L
Mistroke
.27176 .04867 L
.27191 .04837 L
.27725 .03866 L
.28147 .03109 L
.28251 .02894 L
.2876 .01923 L
Mfstroke
0 .64 1 r
.01923 .7124 m
.02894 .71749 L
.03109 .71853 L
.03866 .72275 L
.04837 .72809 L
.04867 .72824 L
.05808 .73378 L
.06526 .73796 L
.06779 .73952 L
.07751 .74557 L
.08089 .74767 L
.08722 .75181 L
.09562 .75738 L
.09693 .75828 L
.10664 .76508 L
.10949 .76709 L
.11636 .77215 L
.12257 .77681 L
.12607 .77952 L
.13491 .78652 L
.13578 .78723 L
.14549 .79533 L
.14656 .79623 L
.15521 .80382 L
.15757 .80594 L
.16492 .81274 L
.16799 .81566 L
.17463 .82212 L
.17788 .82537 L
.18434 .83201 L
.18726 .83508 L
.19406 .84243 L
.19618 .84479 L
.20377 .85344 L
.20467 .85451 L
.21277 .86422 L
.21348 .86509 L
.22048 .87393 L
.22319 .87743 L
.22785 .88364 L
.23291 .89051 L
.23492 .89336 L
.24172 .90307 L
.24262 .90438 L
.24819 .91278 L
.25233 .91911 L
.25443 .92249 L
.26048 .93221 L
.26204 .93474 L
.26622 .94192 L
.27176 .95133 L
.27191 .95163 L
.27725 .96134 L
.28147 .96891 L
.28251 .97106 L
.2876 .98077 L
.01923 .98077 L
F
0 g
.01923 .7124 m
.02894 .71749 L
.03109 .71853 L
.03866 .72275 L
.04837 .72809 L
.04867 .72824 L
.05808 .73378 L
.06526 .73796 L
.06779 .73952 L
.07751 .74557 L
.08089 .74767 L
.08722 .75181 L
.09562 .75738 L
.09693 .75828 L
.10664 .76508 L
.10949 .76709 L
.11636 .77215 L
.12257 .77681 L
.12607 .77952 L
.13491 .78652 L
.13578 .78723 L
.14549 .79533 L
.14656 .79623 L
.15521 .80382 L
.15757 .80594 L
.16492 .81274 L
.16799 .81566 L
.17463 .82212 L
.17788 .82537 L
.18434 .83201 L
.18726 .83508 L
.19406 .84243 L
.19618 .84479 L
.20377 .85344 L
.20467 .85451 L
.21277 .86422 L
.21348 .86509 L
.22048 .87393 L
.22319 .87743 L
.22785 .88364 L
.23291 .89051 L
.23492 .89336 L
.24172 .90307 L
.24262 .90438 L
.24819 .91278 L
.25233 .91911 L
.25443 .92249 L
.26048 .93221 L
.26204 .93474 L
.26622 .94192 L
Mistroke
.27176 .95133 L
.27191 .95163 L
.27725 .96134 L
.28147 .96891 L
.28251 .97106 L
.2876 .98077 L
Mfstroke
0 .88 1 r
.01923 .2809 m
.02894 .27587 L
.03683 .27176 L
.03866 .27077 L
.04837 .26539 L
.05435 .26204 L
.05808 .25987 L
.06779 .25411 L
.07075 .25233 L
.07751 .24812 L
.08612 .24262 L
.08722 .2419 L
.09693 .23537 L
.10053 .23291 L
.10664 .22858 L
.11407 .22319 L
.11636 .22149 L
.12607 .21405 L
.12679 .21348 L
.13578 .20624 L
.13877 .20377 L
.14549 .19805 L
.15007 .19406 L
.15521 .18944 L
.16074 .18434 L
.16492 .18038 L
.17082 .17463 L
.17463 .17082 L
.18038 .16492 L
.18434 .16074 L
.18944 .15521 L
.19406 .15007 L
.19805 .14549 L
.20377 .13877 L
.20624 .13578 L
.21348 .12679 L
.21405 .12607 L
.22149 .11636 L
.22319 .11407 L
.22858 .10664 L
.23291 .10053 L
.23537 .09693 L
.2419 .08722 L
.24262 .08612 L
.24812 .07751 L
.25233 .07075 L
.25411 .06779 L
.25987 .05808 L
.26204 .05435 L
.26539 .04837 L
.27077 .03866 L
.27176 .03683 L
.27587 .02894 L
.2809 .01923 L
.01923 .01923 L
F
0 g
.01923 .2809 m
.02894 .27587 L
.03683 .27176 L
.03866 .27077 L
.04837 .26539 L
.05435 .26204 L
.05808 .25987 L
.06779 .25411 L
.07075 .25233 L
.07751 .24812 L
.08612 .24262 L
.08722 .2419 L
.09693 .23537 L
.10053 .23291 L
.10664 .22858 L
.11407 .22319 L
.11636 .22149 L
.12607 .21405 L
.12679 .21348 L
.13578 .20624 L
.13877 .20377 L
.14549 .19805 L
.15007 .19406 L
.15521 .18944 L
.16074 .18434 L
.16492 .18038 L
.17082 .17463 L
.17463 .17082 L
.18038 .16492 L
.18434 .16074 L
.18944 .15521 L
.19406 .15007 L
.19805 .14549 L
.20377 .13877 L
.20624 .13578 L
.21348 .12679 L
.21405 .12607 L
.22149 .11636 L
.22319 .11407 L
.22858 .10664 L
.23291 .10053 L
.23537 .09693 L
.2419 .08722 L
.24262 .08612 L
.24812 .07751 L
.25233 .07075 L
.25411 .06779 L
.25987 .05808 L
.26204 .05435 L
.26539 .04837 L
Mistroke
.27077 .03866 L
.27176 .03683 L
.27587 .02894 L
.2809 .01923 L
Mfstroke
0 .88 1 r
.01923 .7191 m
.02894 .72413 L
.03683 .72824 L
.03866 .72923 L
.04837 .73461 L
.05435 .73796 L
.05808 .74013 L
.06779 .74589 L
.07075 .74767 L
.07751 .75188 L
.08612 .75738 L
.08722 .7581 L
.09693 .76463 L
.10053 .76709 L
.10664 .77142 L
.11407 .77681 L
.11636 .77851 L
.12607 .78595 L
.12679 .78652 L
.13578 .79376 L
.13877 .79623 L
.14549 .80195 L
.15007 .80594 L
.15521 .81056 L
.16074 .81566 L
.16492 .81962 L
.17082 .82537 L
.17463 .82918 L
.18038 .83508 L
.18434 .83926 L
.18944 .84479 L
.19406 .84993 L
.19805 .85451 L
.20377 .86123 L
.20624 .86422 L
.21348 .87321 L
.21405 .87393 L
.22149 .88364 L
.22319 .88593 L
.22858 .89336 L
.23291 .89947 L
.23537 .90307 L
.2419 .91278 L
.24262 .91388 L
.24812 .92249 L
.25233 .92925 L
.25411 .93221 L
.25987 .94192 L
.26204 .94565 L
.26539 .95163 L
.27077 .96134 L
.27176 .96317 L
.27587 .97106 L
.2809 .98077 L
.01923 .98077 L
F
0 g
.01923 .7191 m
.02894 .72413 L
.03683 .72824 L
.03866 .72923 L
.04837 .73461 L
.05435 .73796 L
.05808 .74013 L
.06779 .74589 L
.07075 .74767 L
.07751 .75188 L
.08612 .75738 L
.08722 .7581 L
.09693 .76463 L
.10053 .76709 L
.10664 .77142 L
.11407 .77681 L
.11636 .77851 L
.12607 .78595 L
.12679 .78652 L
.13578 .79376 L
.13877 .79623 L
.14549 .80195 L
.15007 .80594 L
.15521 .81056 L
.16074 .81566 L
.16492 .81962 L
.17082 .82537 L
.17463 .82918 L
.18038 .83508 L
.18434 .83926 L
.18944 .84479 L
.19406 .84993 L
.19805 .85451 L
.20377 .86123 L
.20624 .86422 L
.21348 .87321 L
.21405 .87393 L
.22149 .88364 L
.22319 .88593 L
.22858 .89336 L
.23291 .89947 L
.23537 .90307 L
.2419 .91278 L
.24262 .91388 L
.24812 .92249 L
.25233 .92925 L
.25411 .93221 L
.25987 .94192 L
.26204 .94565 L
.26539 .95163 L
Mistroke
.27077 .96134 L
.27176 .96317 L
.27587 .97106 L
.2809 .98077 L
Mfstroke
0 1 .88 r
.01923 .27481 m
.02535 .27176 L
.02894 .26991 L
.03866 .26479 L
.04377 .26204 L
.04837 .2595 L
.05808 .25397 L
.06092 .25233 L
.06779 .24823 L
.07691 .24262 L
.07751 .24224 L
.08722 .23598 L
.09186 .23291 L
.09693 .22945 L
.10584 .22319 L
.10664 .22261 L
.11636 .21544 L
.11895 .21348 L
.12607 .20793 L
.13126 .20377 L
.13578 .20004 L
.14284 .19406 L
.14549 .19174 L
.15375 .18434 L
.15521 .18301 L
.16405 .17463 L
.16492 .17379 L
.17379 .16492 L
.17463 .16405 L
.18301 .15521 L
.18434 .15375 L
.19174 .14549 L
.19406 .14284 L
.20004 .13578 L
.20377 .13126 L
.20793 .12607 L
.21348 .11895 L
.21544 .11636 L
.22261 .10664 L
.22319 .10584 L
.22945 .09693 L
.23291 .09186 L
.23598 .08722 L
.24224 .07751 L
.24262 .07691 L
.24823 .06779 L
.25233 .06092 L
.25397 .05808 L
.2595 .04837 L
.26204 .04377 L
.26479 .03866 L
.26991 .02894 L
.27176 .02535 L
.27481 .01923 L
.01923 .01923 L
F
0 g
.01923 .27481 m
.02535 .27176 L
.02894 .26991 L
.03866 .26479 L
.04377 .26204 L
.04837 .2595 L
.05808 .25397 L
.06092 .25233 L
.06779 .24823 L
.07691 .24262 L
.07751 .24224 L
.08722 .23598 L
.09186 .23291 L
.09693 .22945 L
.10584 .22319 L
.10664 .22261 L
.11636 .21544 L
.11895 .21348 L
.12607 .20793 L
.13126 .20377 L
.13578 .20004 L
.14284 .19406 L
.14549 .19174 L
.15375 .18434 L
.15521 .18301 L
.16405 .17463 L
.16492 .17379 L
.17379 .16492 L
.17463 .16405 L
.18301 .15521 L
.18434 .15375 L
.19174 .14549 L
.19406 .14284 L
.20004 .13578 L
.20377 .13126 L
.20793 .12607 L
.21348 .11895 L
.21544 .11636 L
.22261 .10664 L
.22319 .10584 L
.22945 .09693 L
.23291 .09186 L
.23598 .08722 L
.24224 .07751 L
.24262 .07691 L
.24823 .06779 L
.25233 .06092 L
.25397 .05808 L
.2595 .04837 L
.26204 .04377 L
Mistroke
.26479 .03866 L
.26991 .02894 L
.27176 .02535 L
.27481 .01923 L
Mfstroke
0 1 .88 r
.01923 .72519 m
.02535 .72824 L
.02894 .73009 L
.03866 .73521 L
.04377 .73796 L
.04837 .7405 L
.05808 .74603 L
.06092 .74767 L
.06779 .75177 L
.07691 .75738 L
.07751 .75776 L
.08722 .76402 L
.09186 .76709 L
.09693 .77055 L
.10584 .77681 L
.10664 .77739 L
.11636 .78456 L
.11895 .78652 L
.12607 .79207 L
.13126 .79623 L
.13578 .79996 L
.14284 .80594 L
.14549 .80826 L
.15375 .81566 L
.15521 .81699 L
.16405 .82537 L
.16492 .82621 L
.17379 .83508 L
.17463 .83595 L
.18301 .84479 L
.18434 .84625 L
.19174 .85451 L
.19406 .85716 L
.20004 .86422 L
.20377 .86874 L
.20793 .87393 L
.21348 .88105 L
.21544 .88364 L
.22261 .89336 L
.22319 .89416 L
.22945 .90307 L
.23291 .90814 L
.23598 .91278 L
.24224 .92249 L
.24262 .92309 L
.24823 .93221 L
.25233 .93908 L
.25397 .94192 L
.2595 .95163 L
.26204 .95623 L
.26479 .96134 L
.26991 .97106 L
.27176 .97465 L
.27481 .98077 L
.01923 .98077 L
F
0 g
.01923 .72519 m
.02535 .72824 L
.02894 .73009 L
.03866 .73521 L
.04377 .73796 L
.04837 .7405 L
.05808 .74603 L
.06092 .74767 L
.06779 .75177 L
.07691 .75738 L
.07751 .75776 L
.08722 .76402 L
.09186 .76709 L
.09693 .77055 L
.10584 .77681 L
.10664 .77739 L
.11636 .78456 L
.11895 .78652 L
.12607 .79207 L
.13126 .79623 L
.13578 .79996 L
.14284 .80594 L
.14549 .80826 L
.15375 .81566 L
.15521 .81699 L
.16405 .82537 L
.16492 .82621 L
.17379 .83508 L
.17463 .83595 L
.18301 .84479 L
.18434 .84625 L
.19174 .85451 L
.19406 .85716 L
.20004 .86422 L
.20377 .86874 L
.20793 .87393 L
.21348 .88105 L
.21544 .88364 L
.22261 .89336 L
.22319 .89416 L
.22945 .90307 L
.23291 .90814 L
.23598 .91278 L
.24224 .92249 L
.24262 .92309 L
.24823 .93221 L
.25233 .93908 L
.25397 .94192 L
.2595 .95163 L
.26204 .95623 L
Mistroke
.26479 .96134 L
.26991 .97106 L
.27176 .97465 L
.27481 .98077 L
Mfstroke
0 1 .64 r
.01923 .26909 m
.02894 .2642 L
.03313 .26204 L
.03866 .25912 L
.04837 .25383 L
.05105 .25233 L
.05808 .24831 L
.0677 .24262 L
.06779 .24256 L
.07751 .23655 L
.08319 .23291 L
.08722 .23027 L
.09693 .22369 L
.09764 .22319 L
.10664 .21679 L
.11115 .21348 L
.11636 .20956 L
.1238 .20377 L
.12607 .20196 L
.13568 .19406 L
.13578 .19397 L
.14549 .18555 L
.14684 .18434 L
.15521 .17667 L
.15736 .17463 L
.16492 .16729 L
.16729 .16492 L
.17463 .15736 L
.17667 .15521 L
.18434 .14684 L
.18555 .14549 L
.19397 .13578 L
.19406 .13568 L
.20196 .12607 L
.20377 .1238 L
.20956 .11636 L
.21348 .11115 L
.21679 .10664 L
.22319 .09764 L
.22369 .09693 L
.23027 .08722 L
.23291 .08319 L
.23655 .07751 L
.24256 .06779 L
.24262 .0677 L
.24831 .05808 L
.25233 .05105 L
.25383 .04837 L
.25912 .03866 L
.26204 .03313 L
.2642 .02894 L
.26909 .01923 L
.01923 .01923 L
F
0 g
.01923 .26909 m
.02894 .2642 L
.03313 .26204 L
.03866 .25912 L
.04837 .25383 L
.05105 .25233 L
.05808 .24831 L
.0677 .24262 L
.06779 .24256 L
.07751 .23655 L
.08319 .23291 L
.08722 .23027 L
.09693 .22369 L
.09764 .22319 L
.10664 .21679 L
.11115 .21348 L
.11636 .20956 L
.1238 .20377 L
.12607 .20196 L
.13568 .19406 L
.13578 .19397 L
.14549 .18555 L
.14684 .18434 L
.15521 .17667 L
.15736 .17463 L
.16492 .16729 L
.16729 .16492 L
.17463 .15736 L
.17667 .15521 L
.18434 .14684 L
.18555 .14549 L
.19397 .13578 L
.19406 .13568 L
.20196 .12607 L
.20377 .1238 L
.20956 .11636 L
.21348 .11115 L
.21679 .10664 L
.22319 .09764 L
.22369 .09693 L
.23027 .08722 L
.23291 .08319 L
.23655 .07751 L
.24256 .06779 L
.24262 .0677 L
.24831 .05808 L
.25233 .05105 L
.25383 .04837 L
.25912 .03866 L
.26204 .03313 L
Mistroke
.2642 .02894 L
.26909 .01923 L
Mfstroke
0 1 .64 r
.01923 .73091 m
.02894 .7358 L
.03313 .73796 L
.03866 .74088 L
.04837 .74617 L
.05105 .74767 L
.05808 .75169 L
.0677 .75738 L
.06779 .75744 L
.07751 .76345 L
.08319 .76709 L
.08722 .76973 L
.09693 .77631 L
.09764 .77681 L
.10664 .78321 L
.11115 .78652 L
.11636 .79044 L
.1238 .79623 L
.12607 .79804 L
.13568 .80594 L
.13578 .80603 L
.14549 .81445 L
.14684 .81566 L
.15521 .82333 L
.15736 .82537 L
.16492 .83271 L
.16729 .83508 L
.17463 .84264 L
.17667 .84479 L
.18434 .85316 L
.18555 .85451 L
.19397 .86422 L
.19406 .86432 L
.20196 .87393 L
.20377 .8762 L
.20956 .88364 L
.21348 .88885 L
.21679 .89336 L
.22319 .90236 L
.22369 .90307 L
.23027 .91278 L
.23291 .91681 L
.23655 .92249 L
.24256 .93221 L
.24262 .9323 L
.24831 .94192 L
.25233 .94895 L
.25383 .95163 L
.25912 .96134 L
.26204 .96687 L
.2642 .97106 L
.26909 .98077 L
.01923 .98077 L
F
0 g
.01923 .73091 m
.02894 .7358 L
.03313 .73796 L
.03866 .74088 L
.04837 .74617 L
.05105 .74767 L
.05808 .75169 L
.0677 .75738 L
.06779 .75744 L
.07751 .76345 L
.08319 .76709 L
.08722 .76973 L
.09693 .77631 L
.09764 .77681 L
.10664 .78321 L
.11115 .78652 L
.11636 .79044 L
.1238 .79623 L
.12607 .79804 L
.13568 .80594 L
.13578 .80603 L
.14549 .81445 L
.14684 .81566 L
.15521 .82333 L
.15736 .82537 L
.16492 .83271 L
.16729 .83508 L
.17463 .84264 L
.17667 .84479 L
.18434 .85316 L
.18555 .85451 L
.19397 .86422 L
.19406 .86432 L
.20196 .87393 L
.20377 .8762 L
.20956 .88364 L
.21348 .88885 L
.21679 .89336 L
.22319 .90236 L
.22369 .90307 L
.23027 .91278 L
.23291 .91681 L
.23655 .92249 L
.24256 .93221 L
.24262 .9323 L
.24831 .94192 L
.25233 .94895 L
.25383 .95163 L
.25912 .96134 L
.26204 .96687 L
Mistroke
.2642 .97106 L
.26909 .98077 L
Mfstroke
0 1 .4 r
.01923 .26343 m
.02204 .26204 L
.02894 .25858 L
.03866 .2535 L
.04084 .25233 L
.04837 .24822 L
.05808 .24269 L
.0582 .24262 L
.06779 .23692 L
.0743 .23291 L
.07751 .23089 L
.08722 .22456 L
.08925 .22319 L
.09693 .21793 L
.1032 .21348 L
.10664 .21098 L
.11621 .20377 L
.11636 .20366 L
.12607 .19597 L
.12841 .19406 L
.13578 .18787 L
.13985 .18434 L
.14549 .17932 L
.15061 .17463 L
.15521 .17029 L
.16073 .16492 L
.16492 .16073 L
.17029 .15521 L
.17463 .15061 L
.17932 .14549 L
.18434 .13985 L
.18787 .13578 L
.19406 .12841 L
.19597 .12607 L
.20366 .11636 L
.20377 .11621 L
.21098 .10664 L
.21348 .1032 L
.21793 .09693 L
.22319 .08925 L
.22456 .08722 L
.23089 .07751 L
.23291 .0743 L
.23692 .06779 L
.24262 .0582 L
.24269 .05808 L
.24822 .04837 L
.25233 .04084 L
.2535 .03866 L
.25858 .02894 L
.26204 .02204 L
.26343 .01923 L
.01923 .01923 L
F
0 g
.01923 .26343 m
.02204 .26204 L
.02894 .25858 L
.03866 .2535 L
.04084 .25233 L
.04837 .24822 L
.05808 .24269 L
.0582 .24262 L
.06779 .23692 L
.0743 .23291 L
.07751 .23089 L
.08722 .22456 L
.08925 .22319 L
.09693 .21793 L
.1032 .21348 L
.10664 .21098 L
.11621 .20377 L
.11636 .20366 L
.12607 .19597 L
.12841 .19406 L
.13578 .18787 L
.13985 .18434 L
.14549 .17932 L
.15061 .17463 L
.15521 .17029 L
.16073 .16492 L
.16492 .16073 L
.17029 .15521 L
.17463 .15061 L
.17932 .14549 L
.18434 .13985 L
.18787 .13578 L
.19406 .12841 L
.19597 .12607 L
.20366 .11636 L
.20377 .11621 L
.21098 .10664 L
.21348 .1032 L
.21793 .09693 L
.22319 .08925 L
.22456 .08722 L
.23089 .07751 L
.23291 .0743 L
.23692 .06779 L
.24262 .0582 L
.24269 .05808 L
.24822 .04837 L
.25233 .04084 L
.2535 .03866 L
.25858 .02894 L
Mistroke
.26204 .02204 L
.26343 .01923 L
Mfstroke
0 1 .4 r
.01923 .73657 m
.02204 .73796 L
.02894 .74142 L
.03866 .7465 L
.04084 .74767 L
.04837 .75178 L
.05808 .75731 L
.0582 .75738 L
.06779 .76308 L
.0743 .76709 L
.07751 .76911 L
.08722 .77544 L
.08925 .77681 L
.09693 .78207 L
.1032 .78652 L
.10664 .78902 L
.11621 .79623 L
.11636 .79634 L
.12607 .80403 L
.12841 .80594 L
.13578 .81213 L
.13985 .81566 L
.14549 .82068 L
.15061 .82537 L
.15521 .82971 L
.16073 .83508 L
.16492 .83927 L
.17029 .84479 L
.17463 .84939 L
.17932 .85451 L
.18434 .86015 L
.18787 .86422 L
.19406 .87159 L
.19597 .87393 L
.20366 .88364 L
.20377 .88379 L
.21098 .89336 L
.21348 .8968 L
.21793 .90307 L
.22319 .91075 L
.22456 .91278 L
.23089 .92249 L
.23291 .9257 L
.23692 .93221 L
.24262 .9418 L
.24269 .94192 L
.24822 .95163 L
.25233 .95916 L
.2535 .96134 L
.25858 .97106 L
.26204 .97796 L
.26343 .98077 L
.01923 .98077 L
F
0 g
.01923 .73657 m
.02204 .73796 L
.02894 .74142 L
.03866 .7465 L
.04084 .74767 L
.04837 .75178 L
.05808 .75731 L
.0582 .75738 L
.06779 .76308 L
.0743 .76709 L
.07751 .76911 L
.08722 .77544 L
.08925 .77681 L
.09693 .78207 L
.1032 .78652 L
.10664 .78902 L
.11621 .79623 L
.11636 .79634 L
.12607 .80403 L
.12841 .80594 L
.13578 .81213 L
.13985 .81566 L
.14549 .82068 L
.15061 .82537 L
.15521 .82971 L
.16073 .83508 L
.16492 .83927 L
.17029 .84479 L
.17463 .84939 L
.17932 .85451 L
.18434 .86015 L
.18787 .86422 L
.19406 .87159 L
.19597 .87393 L
.20366 .88364 L
.20377 .88379 L
.21098 .89336 L
.21348 .8968 L
.21793 .90307 L
.22319 .91075 L
.22456 .91278 L
.23089 .92249 L
.23291 .9257 L
.23692 .93221 L
.24262 .9418 L
.24269 .94192 L
.24822 .95163 L
.25233 .95916 L
.2535 .96134 L
.25858 .97106 L
Mistroke
.26204 .97796 L
.26343 .98077 L
Mfstroke
0 1 .16 r
.01923 .25767 m
.02894 .2528 L
.02986 .25233 L
.03866 .24777 L
.04808 .24262 L
.04837 .24246 L
.05808 .23694 L
.06488 .23291 L
.06779 .23115 L
.07751 .22507 L
.08042 .22319 L
.08722 .21872 L
.09485 .21348 L
.09693 .21203 L
.10664 .20499 L
.10829 .20377 L
.11636 .19761 L
.12084 .19406 L
.12607 .18982 L
.13258 .18434 L
.13578 .18159 L
.1436 .17463 L
.14549 .1729 L
.15395 .16492 L
.15521 .1637 L
.1637 .15521 L
.16492 .15395 L
.1729 .14549 L
.17463 .1436 L
.18159 .13578 L
.18434 .13258 L
.18982 .12607 L
.19406 .12084 L
.19761 .11636 L
.20377 .10829 L
.20499 .10664 L
.21203 .09693 L
.21348 .09485 L
.21872 .08722 L
.22319 .08042 L
.22507 .07751 L
.23115 .06779 L
.23291 .06488 L
.23694 .05808 L
.24246 .04837 L
.24262 .04808 L
.24777 .03866 L
.25233 .02986 L
.2528 .02894 L
.25767 .01923 L
.01923 .01923 L
F
0 g
.01923 .25767 m
.02894 .2528 L
.02986 .25233 L
.03866 .24777 L
.04808 .24262 L
.04837 .24246 L
.05808 .23694 L
.06488 .23291 L
.06779 .23115 L
.07751 .22507 L
.08042 .22319 L
.08722 .21872 L
.09485 .21348 L
.09693 .21203 L
.10664 .20499 L
.10829 .20377 L
.11636 .19761 L
.12084 .19406 L
.12607 .18982 L
.13258 .18434 L
.13578 .18159 L
.1436 .17463 L
.14549 .1729 L
.15395 .16492 L
.15521 .1637 L
.1637 .15521 L
.16492 .15395 L
.1729 .14549 L
.17463 .1436 L
.18159 .13578 L
.18434 .13258 L
.18982 .12607 L
.19406 .12084 L
.19761 .11636 L
.20377 .10829 L
.20499 .10664 L
.21203 .09693 L
.21348 .09485 L
.21872 .08722 L
.22319 .08042 L
.22507 .07751 L
.23115 .06779 L
.23291 .06488 L
.23694 .05808 L
.24246 .04837 L
.24262 .04808 L
.24777 .03866 L
.25233 .02986 L
.2528 .02894 L
.25767 .01923 L
Mistroke
Mfstroke
0 1 .16 r
.01923 .74233 m
.02894 .7472 L
.02986 .74767 L
.03866 .75223 L
.04808 .75738 L
.04837 .75754 L
.05808 .76306 L
.06488 .76709 L
.06779 .76885 L
.07751 .77493 L
.08042 .77681 L
.08722 .78128 L
.09485 .78652 L
.09693 .78797 L
.10664 .79501 L
.10829 .79623 L
.11636 .80239 L
.12084 .80594 L
.12607 .81018 L
.13258 .81566 L
.13578 .81841 L
.1436 .82537 L
.14549 .8271 L
.15395 .83508 L
.15521 .8363 L
.1637 .84479 L
.16492 .84605 L
.1729 .85451 L
.17463 .8564 L
.18159 .86422 L
.18434 .86742 L
.18982 .87393 L
.19406 .87916 L
.19761 .88364 L
.20377 .89171 L
.20499 .89336 L
.21203 .90307 L
.21348 .90515 L
.21872 .91278 L
.22319 .91958 L
.22507 .92249 L
.23115 .93221 L
.23291 .93512 L
.23694 .94192 L
.24246 .95163 L
.24262 .95192 L
.24777 .96134 L
.25233 .97014 L
.2528 .97106 L
.25767 .98077 L
.01923 .98077 L
F
0 g
.01923 .74233 m
.02894 .7472 L
.02986 .74767 L
.03866 .75223 L
.04808 .75738 L
.04837 .75754 L
.05808 .76306 L
.06488 .76709 L
.06779 .76885 L
.07751 .77493 L
.08042 .77681 L
.08722 .78128 L
.09485 .78652 L
.09693 .78797 L
.10664 .79501 L
.10829 .79623 L
.11636 .80239 L
.12084 .80594 L
.12607 .81018 L
.13258 .81566 L
.13578 .81841 L
.1436 .82537 L
.14549 .8271 L
.15395 .83508 L
.15521 .8363 L
.1637 .84479 L
.16492 .84605 L
.1729 .85451 L
.17463 .8564 L
.18159 .86422 L
.18434 .86742 L
.18982 .87393 L
.19406 .87916 L
.19761 .88364 L
.20377 .89171 L
.20499 .89336 L
.21203 .90307 L
.21348 .90515 L
.21872 .91278 L
.22319 .91958 L
.22507 .92249 L
.23115 .93221 L
.23291 .93512 L
.23694 .94192 L
.24246 .95163 L
.24262 .95192 L
.24777 .96134 L
.25233 .97014 L
.2528 .97106 L
.25767 .98077 L
Mistroke
Mfstroke
.08 1 0 r
.01923 .2515 m
.02894 .24671 L
.03678 .24262 L
.03866 .24165 L
.04837 .23638 L
.05448 .23291 L
.05808 .23085 L
.06779 .22502 L
.07075 .22319 L
.07751 .21895 L
.08578 .21348 L
.08722 .21252 L
.09693 .20578 L
.09972 .20377 L
.10664 .19869 L
.1127 .19406 L
.11636 .1912 L
.12479 .18434 L
.12607 .18328 L
.13578 .17491 L
.1361 .17463 L
.14549 .16607 L
.14671 .16492 L
.15521 .15669 L
.15669 .15521 L
.16492 .14671 L
.16607 .14549 L
.17463 .1361 L
.17491 .13578 L
.18328 .12607 L
.18434 .12479 L
.1912 .11636 L
.19406 .1127 L
.19869 .10664 L
.20377 .09972 L
.20578 .09693 L
.21252 .08722 L
.21348 .08578 L
.21895 .07751 L
.22319 .07075 L
.22502 .06779 L
.23085 .05808 L
.23291 .05448 L
.23638 .04837 L
.24165 .03866 L
.24262 .03678 L
.24671 .02894 L
.2515 .01923 L
.01923 .01923 L
F
0 g
.01923 .2515 m
.02894 .24671 L
.03678 .24262 L
.03866 .24165 L
.04837 .23638 L
.05448 .23291 L
.05808 .23085 L
.06779 .22502 L
.07075 .22319 L
.07751 .21895 L
.08578 .21348 L
.08722 .21252 L
.09693 .20578 L
.09972 .20377 L
.10664 .19869 L
.1127 .19406 L
.11636 .1912 L
.12479 .18434 L
.12607 .18328 L
.13578 .17491 L
.1361 .17463 L
.14549 .16607 L
.14671 .16492 L
.15521 .15669 L
.15669 .15521 L
.16492 .14671 L
.16607 .14549 L
.17463 .1361 L
.17491 .13578 L
.18328 .12607 L
.18434 .12479 L
.1912 .11636 L
.19406 .1127 L
.19869 .10664 L
.20377 .09972 L
.20578 .09693 L
.21252 .08722 L
.21348 .08578 L
.21895 .07751 L
.22319 .07075 L
.22502 .06779 L
.23085 .05808 L
.23291 .05448 L
.23638 .04837 L
.24165 .03866 L
.24262 .03678 L
.24671 .02894 L
.2515 .01923 L
s
.08 1 0 r
.01923 .7485 m
.02894 .75329 L
.03678 .75738 L
.03866 .75835 L
.04837 .76362 L
.05448 .76709 L
.05808 .76915 L
.06779 .77498 L
.07075 .77681 L
.07751 .78105 L
.08578 .78652 L
.08722 .78748 L
.09693 .79422 L
.09972 .79623 L
.10664 .80131 L
.1127 .80594 L
.11636 .8088 L
.12479 .81566 L
.12607 .81672 L
.13578 .82509 L
.1361 .82537 L
.14549 .83393 L
.14671 .83508 L
.15521 .84331 L
.15669 .84479 L
.16492 .85329 L
.16607 .85451 L
.17463 .8639 L
.17491 .86422 L
.18328 .87393 L
.18434 .87521 L
.1912 .88364 L
.19406 .8873 L
.19869 .89336 L
.20377 .90028 L
.20578 .90307 L
.21252 .91278 L
.21348 .91422 L
.21895 .92249 L
.22319 .92925 L
.22502 .93221 L
.23085 .94192 L
.23291 .94552 L
.23638 .95163 L
.24165 .96134 L
.24262 .96322 L
.24671 .97106 L
.2515 .98077 L
.01923 .98077 L
F
0 g
.01923 .7485 m
.02894 .75329 L
.03678 .75738 L
.03866 .75835 L
.04837 .76362 L
.05448 .76709 L
.05808 .76915 L
.06779 .77498 L
.07075 .77681 L
.07751 .78105 L
.08578 .78652 L
.08722 .78748 L
.09693 .79422 L
.09972 .79623 L
.10664 .80131 L
.1127 .80594 L
.11636 .8088 L
.12479 .81566 L
.12607 .81672 L
.13578 .82509 L
.1361 .82537 L
.14549 .83393 L
.14671 .83508 L
.15521 .84331 L
.15669 .84479 L
.16492 .85329 L
.16607 .85451 L
.17463 .8639 L
.17491 .86422 L
.18328 .87393 L
.18434 .87521 L
.1912 .88364 L
.19406 .8873 L
.19869 .89336 L
.20377 .90028 L
.20578 .90307 L
.21252 .91278 L
.21348 .91422 L
.21895 .92249 L
.22319 .92925 L
.22502 .93221 L
.23085 .94192 L
.23291 .94552 L
.23638 .95163 L
.24165 .96134 L
.24262 .96322 L
.24671 .97106 L
.2515 .98077 L
s
.32 1 0 r
.01923 .24452 m
.02302 .24262 L
.02894 .23989 L
.03866 .2348 L
.04217 .23291 L
.04837 .22964 L
.05808 .22402 L
.05948 .22319 L
.06779 .21829 L
.07538 .21348 L
.07751 .21215 L
.08722 .20568 L
.08998 .20377 L
.09693 .19892 L
.10352 .19406 L
.10664 .19174 L
.11605 .18434 L
.11636 .1841 L
.12607 .17608 L
.12775 .17463 L
.13578 .16758 L
.1387 .16492 L
.14549 .15855 L
.14894 .15521 L
.15521 .14894 L
.15855 .14549 L
.16492 .1387 L
.16758 .13578 L
.17463 .12775 L
.17608 .12607 L
.1841 .11636 L
.18434 .11605 L
.19174 .10664 L
.19406 .10352 L
.19892 .09693 L
.20377 .08998 L
.20568 .08722 L
.21215 .07751 L
.21348 .07538 L
.21829 .06779 L
.22319 .05948 L
.22402 .05808 L
.22964 .04837 L
.23291 .04217 L
.2348 .03866 L
.23989 .02894 L
.24262 .02302 L
.24452 .01923 L
.01923 .01923 L
F
0 g
.01923 .24452 m
.02302 .24262 L
.02894 .23989 L
.03866 .2348 L
.04217 .23291 L
.04837 .22964 L
.05808 .22402 L
.05948 .22319 L
.06779 .21829 L
.07538 .21348 L
.07751 .21215 L
.08722 .20568 L
.08998 .20377 L
.09693 .19892 L
.10352 .19406 L
.10664 .19174 L
.11605 .18434 L
.11636 .1841 L
.12607 .17608 L
.12775 .17463 L
.13578 .16758 L
.1387 .16492 L
.14549 .15855 L
.14894 .15521 L
.15521 .14894 L
.15855 .14549 L
.16492 .1387 L
.16758 .13578 L
.17463 .12775 L
.17608 .12607 L
.1841 .11636 L
.18434 .11605 L
.19174 .10664 L
.19406 .10352 L
.19892 .09693 L
.20377 .08998 L
.20568 .08722 L
.21215 .07751 L
.21348 .07538 L
.21829 .06779 L
.22319 .05948 L
.22402 .05808 L
.22964 .04837 L
.23291 .04217 L
.2348 .03866 L
.23989 .02894 L
.24262 .02302 L
.24452 .01923 L
s
.32 1 0 r
.01923 .75548 m
.02302 .75738 L
.02894 .76011 L
.03866 .7652 L
.04217 .76709 L
.04837 .77036 L
.05808 .77598 L
.05948 .77681 L
.06779 .78171 L
.07538 .78652 L
.07751 .78785 L
.08722 .79432 L
.08998 .79623 L
.09693 .80108 L
.10352 .80594 L
.10664 .80826 L
.11605 .81566 L
.11636 .8159 L
.12607 .82392 L
.12775 .82537 L
.13578 .83242 L
.1387 .83508 L
.14549 .84145 L
.14894 .84479 L
.15521 .85106 L
.15855 .85451 L
.16492 .8613 L
.16758 .86422 L
.17463 .87225 L
.17608 .87393 L
.1841 .88364 L
.18434 .88395 L
.19174 .89336 L
.19406 .89648 L
.19892 .90307 L
.20377 .91002 L
.20568 .91278 L
.21215 .92249 L
.21348 .92462 L
.21829 .93221 L
.22319 .94052 L
.22402 .94192 L
.22964 .95163 L
.23291 .95783 L
.2348 .96134 L
.23989 .97106 L
.24262 .97698 L
.24452 .98077 L
.01923 .98077 L
F
0 g
.01923 .75548 m
.02302 .75738 L
.02894 .76011 L
.03866 .7652 L
.04217 .76709 L
.04837 .77036 L
.05808 .77598 L
.05948 .77681 L
.06779 .78171 L
.07538 .78652 L
.07751 .78785 L
.08722 .79432 L
.08998 .79623 L
.09693 .80108 L
.10352 .80594 L
.10664 .80826 L
.11605 .81566 L
.11636 .8159 L
.12607 .82392 L
.12775 .82537 L
.13578 .83242 L
.1387 .83508 L
.14549 .84145 L
.14894 .84479 L
.15521 .85106 L
.15855 .85451 L
.16492 .8613 L
.16758 .86422 L
.17463 .87225 L
.17608 .87393 L
.1841 .88364 L
.18434 .88395 L
.19174 .89336 L
.19406 .89648 L
.19892 .90307 L
.20377 .91002 L
.20568 .91278 L
.21215 .92249 L
.21348 .92462 L
.21829 .93221 L
.22319 .94052 L
.22402 .94192 L
.22964 .95163 L
.23291 .95783 L
.2348 .96134 L
.23989 .97106 L
.24262 .97698 L
.24452 .98077 L
s
.56 1 0 r
.01923 .23582 m
.02466 .23291 L
.02894 .23198 L
.03866 .22649 L
.04464 .22319 L
.04837 .22166 L
.05808 .21586 L
.06204 .21348 L
.06779 .21034 L
.07751 .20391 L
.07772 .20377 L
.08722 .19763 L
.09226 .19406 L
.09693 .19087 L
.10549 .18434 L
.10664 .18351 L
.11636 .17574 L
.1177 .17463 L
.12607 .16762 L
.12913 .16492 L
.13578 .15894 L
.13976 .15521 L
.14549 .14968 L
.14968 .14549 L
.15521 .13976 L
.15894 .13578 L
.16492 .12913 L
.16762 .12607 L
.17463 .1177 L
.17574 .11636 L
.18351 .10664 L
.18434 .10549 L
.19087 .09693 L
.19406 .09226 L
.19763 .08722 L
.20377 .07772 L
.20391 .07751 L
.21034 .06779 L
.21348 .06204 L
.21586 .05808 L
.22166 .04837 L
.22319 .04464 L
.22649 .03866 L
.23198 .02894 L
.23291 .02466 L
.23582 .01923 L
.01923 .01923 L
F
0 g
.01923 .23582 m
.02466 .23291 L
.02894 .23198 L
.03866 .22649 L
.04464 .22319 L
.04837 .22166 L
.05808 .21586 L
.06204 .21348 L
.06779 .21034 L
.07751 .20391 L
.07772 .20377 L
.08722 .19763 L
.09226 .19406 L
.09693 .19087 L
.10549 .18434 L
.10664 .18351 L
.11636 .17574 L
.1177 .17463 L
.12607 .16762 L
.12913 .16492 L
.13578 .15894 L
.13976 .15521 L
.14549 .14968 L
.14968 .14549 L
.15521 .13976 L
.15894 .13578 L
.16492 .12913 L
.16762 .12607 L
.17463 .1177 L
.17574 .11636 L
.18351 .10664 L
.18434 .10549 L
.19087 .09693 L
.19406 .09226 L
.19763 .08722 L
.20377 .07772 L
.20391 .07751 L
.21034 .06779 L
.21348 .06204 L
.21586 .05808 L
.22166 .04837 L
.22319 .04464 L
.22649 .03866 L
.23198 .02894 L
.23291 .02466 L
.23582 .01923 L
s
.56 1 0 r
.01923 .76418 m
.02466 .76709 L
.02894 .76802 L
.03866 .77351 L
.04464 .77681 L
.04837 .77834 L
.05808 .78414 L
.06204 .78652 L
.06779 .78966 L
.07751 .79609 L
.07772 .79623 L
.08722 .80237 L
.09226 .80594 L
.09693 .80913 L
.10549 .81566 L
.10664 .81649 L
.11636 .82426 L
.1177 .82537 L
.12607 .83238 L
.12913 .83508 L
.13578 .84106 L
.13976 .84479 L
.14549 .85032 L
.14968 .85451 L
.15521 .86024 L
.15894 .86422 L
.16492 .87087 L
.16762 .87393 L
.17463 .8823 L
.17574 .88364 L
.18351 .89336 L
.18434 .89451 L
.19087 .90307 L
.19406 .90774 L
.19763 .91278 L
.20377 .92228 L
.20391 .92249 L
.21034 .93221 L
.21348 .93796 L
.21586 .94192 L
.22166 .95163 L
.22319 .95536 L
.22649 .96134 L
.23198 .97106 L
.23291 .97534 L
.23582 .98077 L
.01923 .98077 L
F
0 g
.01923 .76418 m
.02466 .76709 L
.02894 .76802 L
.03866 .77351 L
.04464 .77681 L
.04837 .77834 L
.05808 .78414 L
.06204 .78652 L
.06779 .78966 L
.07751 .79609 L
.07772 .79623 L
.08722 .80237 L
.09226 .80594 L
.09693 .80913 L
.10549 .81566 L
.10664 .81649 L
.11636 .82426 L
.1177 .82537 L
.12607 .83238 L
.12913 .83508 L
.13578 .84106 L
.13976 .84479 L
.14549 .85032 L
.14968 .85451 L
.15521 .86024 L
.15894 .86422 L
.16492 .87087 L
.16762 .87393 L
.17463 .8823 L
.17574 .88364 L
.18351 .89336 L
.18434 .89451 L
.19087 .90307 L
.19406 .90774 L
.19763 .91278 L
.20377 .92228 L
.20391 .92249 L
.21034 .93221 L
.21348 .93796 L
.21586 .94192 L
.22166 .95163 L
.22319 .95536 L
.22649 .96134 L
.23198 .97106 L
.23291 .97534 L
.23582 .98077 L
s
.32 1 0 r
.01923 .201 m
.02894 .19426 L
.0292 .19406 L
.03866 .18644 L
.04211 .18434 L
.04837 .17925 L
.05464 .17463 L
.05808 .17179 L
.06671 .16492 L
.06779 .164 L
.07751 .15577 L
.07825 .15521 L
.08722 .14725 L
.0893 .14549 L
.09693 .13849 L
.09994 .13578 L
.10664 .12941 L
.11016 .12607 L
.11636 .11998 L
.11998 .11636 L
.12607 .11016 L
.12941 .10664 L
.13578 .09994 L
.13849 .09693 L
.14549 .0893 L
.14725 .08722 L
.15521 .07825 L
.15577 .07751 L
.164 .06779 L
.16492 .06671 L
.17179 .05808 L
.17463 .05464 L
.17925 .04837 L
.18434 .04211 L
.18644 .03866 L
.19406 .0292 L
.19426 .02894 L
.201 .01923 L
.01923 .01923 L
F
0 g
.01923 .201 m
.02894 .19426 L
.0292 .19406 L
.03866 .18644 L
.04211 .18434 L
.04837 .17925 L
.05464 .17463 L
.05808 .17179 L
.06671 .16492 L
.06779 .164 L
.07751 .15577 L
.07825 .15521 L
.08722 .14725 L
.0893 .14549 L
.09693 .13849 L
.09994 .13578 L
.10664 .12941 L
.11016 .12607 L
.11636 .11998 L
.11998 .11636 L
.12607 .11016 L
.12941 .10664 L
.13578 .09994 L
.13849 .09693 L
.14549 .0893 L
.14725 .08722 L
.15521 .07825 L
.15577 .07751 L
.164 .06779 L
.16492 .06671 L
.17179 .05808 L
.17463 .05464 L
.17925 .04837 L
.18434 .04211 L
.18644 .03866 L
.19406 .0292 L
.19426 .02894 L
.201 .01923 L
s
.32 1 0 r
.01923 .799 m
.02894 .80574 L
.0292 .80594 L
.03866 .81356 L
.04211 .81566 L
.04837 .82075 L
.05464 .82537 L
.05808 .82821 L
.06671 .83508 L
.06779 .836 L
.07751 .84423 L
.07825 .84479 L
.08722 .85275 L
.0893 .85451 L
.09693 .86151 L
.09994 .86422 L
.10664 .87059 L
.11016 .87393 L
.11636 .88002 L
.11998 .88364 L
.12607 .88984 L
.12941 .89336 L
.13578 .90006 L
.13849 .90307 L
.14549 .9107 L
.14725 .91278 L
.15521 .92175 L
.15577 .92249 L
.164 .93221 L
.16492 .93329 L
.17179 .94192 L
.17463 .94536 L
.17925 .95163 L
.18434 .95789 L
.18644 .96134 L
.19406 .9708 L
.19426 .97106 L
.201 .98077 L
.01923 .98077 L
F
0 g
.01923 .799 m
.02894 .80574 L
.0292 .80594 L
.03866 .81356 L
.04211 .81566 L
.04837 .82075 L
.05464 .82537 L
.05808 .82821 L
.06671 .83508 L
.06779 .836 L
.07751 .84423 L
.07825 .84479 L
.08722 .85275 L
.0893 .85451 L
.09693 .86151 L
.09994 .86422 L
.10664 .87059 L
.11016 .87393 L
.11636 .88002 L
.11998 .88364 L
.12607 .88984 L
.12941 .89336 L
.13578 .90006 L
.13849 .90307 L
.14549 .9107 L
.14725 .91278 L
.15521 .92175 L
.15577 .92249 L
.164 .93221 L
.16492 .93329 L
.17179 .94192 L
.17463 .94536 L
.17925 .95163 L
.18434 .95789 L
.18644 .96134 L
.19406 .9708 L
.19426 .97106 L
.201 .98077 L
s
.08 1 0 r
.01923 .19166 m
.02894 .18485 L
.0297 .18434 L
.03866 .17772 L
.04292 .17463 L
.04837 .17045 L
.05556 .16492 L
.05808 .16291 L
.06763 .15521 L
.06779 .15507 L
.07751 .14686 L
.07912 .14549 L
.08722 .13835 L
.0901 .13578 L
.09693 .1295 L
.1006 .12607 L
.10664 .12027 L
.11065 .11636 L
.11636 .11065 L
.12027 .10664 L
.12607 .1006 L
.1295 .09693 L
.13578 .0901 L
.13835 .08722 L
.14549 .07912 L
.14686 .07751 L
.15507 .06779 L
.15521 .06763 L
.16291 .05808 L
.16492 .05556 L
.17045 .04837 L
.17463 .04292 L
.17772 .03866 L
.18434 .0297 L
.18485 .02894 L
.19166 .01923 L
.01923 .01923 L
F
0 g
.01923 .19166 m
.02894 .18485 L
.0297 .18434 L
.03866 .17772 L
.04292 .17463 L
.04837 .17045 L
.05556 .16492 L
.05808 .16291 L
.06763 .15521 L
.06779 .15507 L
.07751 .14686 L
.07912 .14549 L
.08722 .13835 L
.0901 .13578 L
.09693 .1295 L
.1006 .12607 L
.10664 .12027 L
.11065 .11636 L
.11636 .11065 L
.12027 .10664 L
.12607 .1006 L
.1295 .09693 L
.13578 .0901 L
.13835 .08722 L
.14549 .07912 L
.14686 .07751 L
.15507 .06779 L
.15521 .06763 L
.16291 .05808 L
.16492 .05556 L
.17045 .04837 L
.17463 .04292 L
.17772 .03866 L
.18434 .0297 L
.18485 .02894 L
.19166 .01923 L
s
.08 1 0 r
.01923 .80834 m
.02894 .81515 L
.0297 .81566 L
.03866 .82228 L
.04292 .82537 L
.04837 .82955 L
.05556 .83508 L
.05808 .83709 L
.06763 .84479 L
.06779 .84493 L
.07751 .85314 L
.07912 .85451 L
.08722 .86165 L
.0901 .86422 L
.09693 .8705 L
.1006 .87393 L
.10664 .87973 L
.11065 .88364 L
.11636 .88935 L
.12027 .89336 L
.12607 .8994 L
.1295 .90307 L
.13578 .9099 L
.13835 .91278 L
.14549 .92088 L
.14686 .92249 L
.15507 .93221 L
.15521 .93237 L
.16291 .94192 L
.16492 .94444 L
.17045 .95163 L
.17463 .95708 L
.17772 .96134 L
.18434 .9703 L
.18485 .97106 L
.19166 .98077 L
.01923 .98077 L
F
0 g
.01923 .80834 m
.02894 .81515 L
.0297 .81566 L
.03866 .82228 L
.04292 .82537 L
.04837 .82955 L
.05556 .83508 L
.05808 .83709 L
.06763 .84479 L
.06779 .84493 L
.07751 .85314 L
.07912 .85451 L
.08722 .86165 L
.0901 .86422 L
.09693 .8705 L
.1006 .87393 L
.10664 .87973 L
.11065 .88364 L
.11636 .88935 L
.12027 .89336 L
.12607 .8994 L
.1295 .90307 L
.13578 .9099 L
.13835 .91278 L
.14549 .92088 L
.14686 .92249 L
.15507 .93221 L
.15521 .93237 L
.16291 .94192 L
.16492 .94444 L
.17045 .95163 L
.17463 .95708 L
.17772 .96134 L
.18434 .9703 L
.18485 .97106 L
.19166 .98077 L
s
0 1 .16 r
.01923 .18405 m
.02894 .17723 L
.03261 .17463 L
.03866 .1702 L
.04574 .16492 L
.04837 .1629 L
.05808 .15531 L
.05821 .15521 L
.06779 .14738 L
.07007 .14549 L
.07751 .13914 L
.08136 .13578 L
.08722 .13055 L
.09213 .12607 L
.09693 .12158 L
.1024 .11636 L
.10664 .1122 L
.1122 .10664 L
.11636 .1024 L
.12158 .09693 L
.12607 .09213 L
.13055 .08722 L
.13578 .08136 L
.13914 .07751 L
.14549 .07007 L
.14738 .06779 L
.15521 .05821 L
.15531 .05808 L
.1629 .04837 L
.16492 .04574 L
.1702 .03866 L
.17463 .03261 L
.17723 .02894 L
.18405 .01923 L
.01923 .01923 L
F
0 g
.01923 .18405 m
.02894 .17723 L
.03261 .17463 L
.03866 .1702 L
.04574 .16492 L
.04837 .1629 L
.05808 .15531 L
.05821 .15521 L
.06779 .14738 L
.07007 .14549 L
.07751 .13914 L
.08136 .13578 L
.08722 .13055 L
.09213 .12607 L
.09693 .12158 L
.1024 .11636 L
.10664 .1122 L
.1122 .10664 L
.11636 .1024 L
.12158 .09693 L
.12607 .09213 L
.13055 .08722 L
.13578 .08136 L
.13914 .07751 L
.14549 .07007 L
.14738 .06779 L
.15521 .05821 L
.15531 .05808 L
.1629 .04837 L
.16492 .04574 L
.1702 .03866 L
.17463 .03261 L
.17723 .02894 L
.18405 .01923 L
s
0 1 .16 r
.01923 .81595 m
.02894 .82277 L
.03261 .82537 L
.03866 .8298 L
.04574 .83508 L
.04837 .8371 L
.05808 .84469 L
.05821 .84479 L
.06779 .85262 L
.07007 .85451 L
.07751 .86086 L
.08136 .86422 L
.08722 .86945 L
.09213 .87393 L
.09693 .87842 L
.1024 .88364 L
.10664 .8878 L
.1122 .89336 L
.11636 .8976 L
.12158 .90307 L
.12607 .90787 L
.13055 .91278 L
.13578 .91864 L
.13914 .92249 L
.14549 .92993 L
.14738 .93221 L
.15521 .94179 L
.15531 .94192 L
.1629 .95163 L
.16492 .95426 L
.1702 .96134 L
.17463 .96739 L
.17723 .97106 L
.18405 .98077 L
.01923 .98077 L
F
0 g
.01923 .81595 m
.02894 .82277 L
.03261 .82537 L
.03866 .8298 L
.04574 .83508 L
.04837 .8371 L
.05808 .84469 L
.05821 .84479 L
.06779 .85262 L
.07007 .85451 L
.07751 .86086 L
.08136 .86422 L
.08722 .86945 L
.09213 .87393 L
.09693 .87842 L
.1024 .88364 L
.10664 .8878 L
.1122 .89336 L
.11636 .8976 L
.12158 .90307 L
.12607 .90787 L
.13055 .91278 L
.13578 .91864 L
.13914 .92249 L
.14549 .92993 L
.14738 .93221 L
.15521 .94179 L
.15531 .94192 L
.1629 .95163 L
.16492 .95426 L
.1702 .96134 L
.17463 .96739 L
.17723 .97106 L
.18405 .98077 L
s
0 1 .4 r
.01923 .17706 m
.02278 .17463 L
.02894 .17028 L
.03637 .16492 L
.03866 .16323 L
.04837 .15588 L
.04924 .15521 L
.05808 .14822 L
.06145 .14549 L
.06779 .14024 L
.07305 .13578 L
.07751 .13191 L
.08407 .12607 L
.08722 .1232 L
.09456 .11636 L
.09693 .1141 L
.10456 .10664 L
.10664 .10456 L
.1141 .09693 L
.11636 .09456 L
.1232 .08722 L
.12607 .08407 L
.13191 .07751 L
.13578 .07305 L
.14024 .06779 L
.14549 .06145 L
.14822 .05808 L
.15521 .04924 L
.15588 .04837 L
.16323 .03866 L
.16492 .03637 L
.17028 .02894 L
.17463 .02278 L
.17706 .01923 L
.01923 .01923 L
F
0 g
.01923 .17706 m
.02278 .17463 L
.02894 .17028 L
.03637 .16492 L
.03866 .16323 L
.04837 .15588 L
.04924 .15521 L
.05808 .14822 L
.06145 .14549 L
.06779 .14024 L
.07305 .13578 L
.07751 .13191 L
.08407 .12607 L
.08722 .1232 L
.09456 .11636 L
.09693 .1141 L
.10456 .10664 L
.10664 .10456 L
.1141 .09693 L
.11636 .09456 L
.1232 .08722 L
.12607 .08407 L
.13191 .07751 L
.13578 .07305 L
.14024 .06779 L
.14549 .06145 L
.14822 .05808 L
.15521 .04924 L
.15588 .04837 L
.16323 .03866 L
.16492 .03637 L
.17028 .02894 L
.17463 .02278 L
.17706 .01923 L
s
0 1 .4 r
.01923 .82294 m
.02278 .82537 L
.02894 .82972 L
.03637 .83508 L
.03866 .83677 L
.04837 .84412 L
.04924 .84479 L
.05808 .85178 L
.06145 .85451 L
.06779 .85976 L
.07305 .86422 L
.07751 .86809 L
.08407 .87393 L
.08722 .8768 L
.09456 .88364 L
.09693 .8859 L
.10456 .89336 L
.10664 .89544 L
.1141 .90307 L
.11636 .90544 L
.1232 .91278 L
.12607 .91593 L
.13191 .92249 L
.13578 .92695 L
.14024 .93221 L
.14549 .93855 L
.14822 .94192 L
.15521 .95076 L
.15588 .95163 L
.16323 .96134 L
.16492 .96363 L
.17028 .97106 L
.17463 .97722 L
.17706 .98077 L
.01923 .98077 L
F
0 g
.01923 .82294 m
.02278 .82537 L
.02894 .82972 L
.03637 .83508 L
.03866 .83677 L
.04837 .84412 L
.04924 .84479 L
.05808 .85178 L
.06145 .85451 L
.06779 .85976 L
.07305 .86422 L
.07751 .86809 L
.08407 .87393 L
.08722 .8768 L
.09456 .88364 L
.09693 .8859 L
.10456 .89336 L
.10664 .89544 L
.1141 .90307 L
.11636 .90544 L
.1232 .91278 L
.12607 .91593 L
.13191 .92249 L
.13578 .92695 L
.14024 .93221 L
.14549 .93855 L
.14822 .94192 L
.15521 .95076 L
.15588 .95163 L
.16323 .96134 L
.16492 .96363 L
.17028 .97106 L
.17463 .97722 L
.17706 .98077 L
s
0 1 .64 r
.01923 .1703 m
.02692 .16492 L
.02894 .16349 L
.03866 .15639 L
.04023 .15521 L
.04837 .14899 L
.05282 .14549 L
.05808 .14127 L
.06474 .13578 L
.06779 .1332 L
.07603 .12607 L
.07751 .12476 L
.08676 .11636 L
.08722 .11593 L
.09693 .10667 L
.09696 .10664 L
.10664 .09696 L
.10667 .09693 L
.11593 .08722 L
.11636 .08676 L
.12476 .07751 L
.12607 .07603 L
.1332 .06779 L
.13578 .06474 L
.14127 .05808 L
.14549 .05282 L
.14899 .04837 L
.15521 .04023 L
.15639 .03866 L
.16349 .02894 L
.16492 .02692 L
.1703 .01923 L
.01923 .01923 L
F
0 g
.01923 .1703 m
.02692 .16492 L
.02894 .16349 L
.03866 .15639 L
.04023 .15521 L
.04837 .14899 L
.05282 .14549 L
.05808 .14127 L
.06474 .13578 L
.06779 .1332 L
.07603 .12607 L
.07751 .12476 L
.08676 .11636 L
.08722 .11593 L
.09693 .10667 L
.09696 .10664 L
.10664 .09696 L
.10667 .09693 L
.11593 .08722 L
.11636 .08676 L
.12476 .07751 L
.12607 .07603 L
.1332 .06779 L
.13578 .06474 L
.14127 .05808 L
.14549 .05282 L
.14899 .04837 L
.15521 .04023 L
.15639 .03866 L
.16349 .02894 L
.16492 .02692 L
.1703 .01923 L
s
0 1 .64 r
.01923 .8297 m
.02692 .83508 L
.02894 .83651 L
.03866 .84361 L
.04023 .84479 L
.04837 .85101 L
.05282 .85451 L
.05808 .85873 L
.06474 .86422 L
.06779 .8668 L
.07603 .87393 L
.07751 .87524 L
.08676 .88364 L
.08722 .88407 L
.09693 .89333 L
.09696 .89336 L
.10664 .90304 L
.10667 .90307 L
.11593 .91278 L
.11636 .91324 L
.12476 .92249 L
.12607 .92397 L
.1332 .93221 L
.13578 .93526 L
.14127 .94192 L
.14549 .94718 L
.14899 .95163 L
.15521 .95977 L
.15639 .96134 L
.16349 .97106 L
.16492 .97308 L
.1703 .98077 L
.01923 .98077 L
F
0 g
.01923 .8297 m
.02692 .83508 L
.02894 .83651 L
.03866 .84361 L
.04023 .84479 L
.04837 .85101 L
.05282 .85451 L
.05808 .85873 L
.06474 .86422 L
.06779 .8668 L
.07603 .87393 L
.07751 .87524 L
.08676 .88364 L
.08722 .88407 L
.09693 .89333 L
.09696 .89336 L
.10664 .90304 L
.10667 .90307 L
.11593 .91278 L
.11636 .91324 L
.12476 .92249 L
.12607 .92397 L
.1332 .93221 L
.13578 .93526 L
.14127 .94192 L
.14549 .94718 L
.14899 .95163 L
.15521 .95977 L
.15639 .96134 L
.16349 .97106 L
.16492 .97308 L
.1703 .98077 L
s
0 1 .88 r
.01923 .16335 m
.02894 .15649 L
.03072 .15521 L
.03866 .14937 L
.04374 .14549 L
.04837 .14191 L
.05602 .13578 L
.05808 .1341 L
.06762 .12607 L
.06779 .12592 L
.07751 .11737 L
.07862 .11636 L
.08722 .10839 L
.08906 .10664 L
.09693 .09897 L
.09897 .09693 L
.10664 .08906 L
.10839 .08722 L
.11636 .07862 L
.11737 .07751 L
.12592 .06779 L
.12607 .06762 L
.1341 .05808 L
.13578 .05602 L
.14191 .04837 L
.14549 .04374 L
.14937 .03866 L
.15521 .03072 L
.15649 .02894 L
.16335 .01923 L
.01923 .01923 L
F
0 g
.01923 .16335 m
.02894 .15649 L
.03072 .15521 L
.03866 .14937 L
.04374 .14549 L
.04837 .14191 L
.05602 .13578 L
.05808 .1341 L
.06762 .12607 L
.06779 .12592 L
.07751 .11737 L
.07862 .11636 L
.08722 .10839 L
.08906 .10664 L
.09693 .09897 L
.09897 .09693 L
.10664 .08906 L
.10839 .08722 L
.11636 .07862 L
.11737 .07751 L
.12592 .06779 L
.12607 .06762 L
.1341 .05808 L
.13578 .05602 L
.14191 .04837 L
.14549 .04374 L
.14937 .03866 L
.15521 .03072 L
.15649 .02894 L
.16335 .01923 L
s
0 1 .88 r
.01923 .83665 m
.02894 .84351 L
.03072 .84479 L
.03866 .85063 L
.04374 .85451 L
.04837 .85809 L
.05602 .86422 L
.05808 .8659 L
.06762 .87393 L
.06779 .87408 L
.07751 .88263 L
.07862 .88364 L
.08722 .89161 L
.08906 .89336 L
.09693 .90103 L
.09897 .90307 L
.10664 .91094 L
.10839 .91278 L
.11636 .92138 L
.11737 .92249 L
.12592 .93221 L
.12607 .93238 L
.1341 .94192 L
.13578 .94398 L
.14191 .95163 L
.14549 .95626 L
.14937 .96134 L
.15521 .96928 L
.15649 .97106 L
.16335 .98077 L
.01923 .98077 L
F
0 g
.01923 .83665 m
.02894 .84351 L
.03072 .84479 L
.03866 .85063 L
.04374 .85451 L
.04837 .85809 L
.05602 .86422 L
.05808 .8659 L
.06762 .87393 L
.06779 .87408 L
.07751 .88263 L
.07862 .88364 L
.08722 .89161 L
.08906 .89336 L
.09693 .90103 L
.09897 .90307 L
.10664 .91094 L
.10839 .91278 L
.11636 .92138 L
.11737 .92249 L
.12592 .93221 L
.12607 .93238 L
.1341 .94192 L
.13578 .94398 L
.14191 .95163 L
.14549 .95626 L
.14937 .96134 L
.15521 .96928 L
.15649 .97106 L
.16335 .98077 L
s
0 .88 1 r
.01923 .15572 m
.01995 .15521 L
.02894 .1489 L
.03357 .14549 L
.03866 .14174 L
.04632 .13578 L
.04837 .13419 L
.05808 .12625 L
.0583 .12607 L
.06779 .11799 L
.06965 .11636 L
.07751 .1093 L
.08038 .10664 L
.08722 .10017 L
.09054 .09693 L
.09693 .09054 L
.10017 .08722 L
.10664 .08038 L
.1093 .07751 L
.11636 .06965 L
.11799 .06779 L
.12607 .0583 L
.12625 .05808 L
.13419 .04837 L
.13578 .04632 L
.14174 .03866 L
.14549 .03357 L
.1489 .02894 L
.15521 .01995 L
.15572 .01923 L
.01923 .01923 L
F
0 g
.01923 .15572 m
.01995 .15521 L
.02894 .1489 L
.03357 .14549 L
.03866 .14174 L
.04632 .13578 L
.04837 .13419 L
.05808 .12625 L
.0583 .12607 L
.06779 .11799 L
.06965 .11636 L
.07751 .1093 L
.08038 .10664 L
.08722 .10017 L
.09054 .09693 L
.09693 .09054 L
.10017 .08722 L
.10664 .08038 L
.1093 .07751 L
.11636 .06965 L
.11799 .06779 L
.12607 .0583 L
.12625 .05808 L
.13419 .04837 L
.13578 .04632 L
.14174 .03866 L
.14549 .03357 L
.1489 .02894 L
.15521 .01995 L
.15572 .01923 L
s
0 .88 1 r
.01923 .84428 m
.01995 .84479 L
.02894 .8511 L
.03357 .85451 L
.03866 .85826 L
.04632 .86422 L
.04837 .86581 L
.05808 .87375 L
.0583 .87393 L
.06779 .88201 L
.06965 .88364 L
.07751 .8907 L
.08038 .89336 L
.08722 .89983 L
.09054 .90307 L
.09693 .90946 L
.10017 .91278 L
.10664 .91962 L
.1093 .92249 L
.11636 .93035 L
.11799 .93221 L
.12607 .9417 L
.12625 .94192 L
.13419 .95163 L
.13578 .95368 L
.14174 .96134 L
.14549 .96643 L
.1489 .97106 L
.15521 .98005 L
.15572 .98077 L
.01923 .98077 L
F
0 g
.01923 .84428 m
.01995 .84479 L
.02894 .8511 L
.03357 .85451 L
.03866 .85826 L
.04632 .86422 L
.04837 .86581 L
.05808 .87375 L
.0583 .87393 L
.06779 .88201 L
.06965 .88364 L
.07751 .8907 L
.08038 .89336 L
.08722 .89983 L
.09054 .90307 L
.09693 .90946 L
.10017 .91278 L
.10664 .91962 L
.1093 .92249 L
.11636 .93035 L
.11799 .93221 L
.12607 .9417 L
.12625 .94192 L
.13419 .95163 L
.13578 .95368 L
.14174 .96134 L
.14549 .96643 L
.1489 .97106 L
.15521 .98005 L
.15572 .98077 L
s
0 .64 1 r
.01923 .14652 m
.02058 .14549 L
.02894 .13985 L
.03435 .13578 L
.03866 .13273 L
.04696 .12607 L
.04837 .12504 L
.05808 .11687 L
.05869 .11636 L
.06779 .10852 L
.06988 .10664 L
.07751 .09968 L
.0804 .09693 L
.08722 .09032 L
.09032 .08722 L
.09693 .0804 L
.09968 .07751 L
.10664 .06988 L
.10852 .06779 L
.11636 .05869 L
.11687 .05808 L
.12504 .04837 L
.12607 .04696 L
.13273 .03866 L
.13578 .03435 L
.13985 .02894 L
.14549 .02058 L
.14652 .01923 L
.01923 .01923 L
F
0 g
.01923 .14652 m
.02058 .14549 L
.02894 .13985 L
.03435 .13578 L
.03866 .13273 L
.04696 .12607 L
.04837 .12504 L
.05808 .11687 L
.05869 .11636 L
.06779 .10852 L
.06988 .10664 L
.07751 .09968 L
.0804 .09693 L
.08722 .09032 L
.09032 .08722 L
.09693 .0804 L
.09968 .07751 L
.10664 .06988 L
.10852 .06779 L
.11636 .05869 L
.11687 .05808 L
.12504 .04837 L
.12607 .04696 L
.13273 .03866 L
.13578 .03435 L
.13985 .02894 L
.14549 .02058 L
.14652 .01923 L
s
0 .64 1 r
.01923 .85348 m
.02058 .85451 L
.02894 .86015 L
.03435 .86422 L
.03866 .86727 L
.04696 .87393 L
.04837 .87496 L
.05808 .88313 L
.05869 .88364 L
.06779 .89148 L
.06988 .89336 L
.07751 .90032 L
.0804 .90307 L
.08722 .90968 L
.09032 .91278 L
.09693 .9196 L
.09968 .92249 L
.10664 .93012 L
.10852 .93221 L
.11636 .94131 L
.11687 .94192 L
.12504 .95163 L
.12607 .95304 L
.13273 .96134 L
.13578 .96565 L
.13985 .97106 L
.14549 .97942 L
.14652 .98077 L
.01923 .98077 L
F
0 g
.01923 .85348 m
.02058 .85451 L
.02894 .86015 L
.03435 .86422 L
.03866 .86727 L
.04696 .87393 L
.04837 .87496 L
.05808 .88313 L
.05869 .88364 L
.06779 .89148 L
.06988 .89336 L
.07751 .90032 L
.0804 .90307 L
.08722 .90968 L
.09032 .91278 L
.09693 .9196 L
.09968 .92249 L
.10664 .93012 L
.10852 .93221 L
.11636 .94131 L
.11687 .94192 L
.12504 .95163 L
.12607 .95304 L
.13273 .96134 L
.13578 .96565 L
.13985 .97106 L
.14549 .97942 L
.14652 .98077 L
s
0 .4 1 r
.01923 .87159 m
.02394 .87393 L
.02894 .87696 L
.03866 .88341 L
.03898 .88364 L
.04837 .89064 L
.05182 .89336 L
.05808 .8985 L
.06342 .90307 L
.06779 .90697 L
.07406 .91278 L
.07751 .9161 L
.0839 .92249 L
.08722 .92594 L
.09303 .93221 L
.09693 .93658 L
.1015 .94192 L
.10664 .94818 L
.10936 .95163 L
.11636 .96102 L
.11659 .96134 L
.12304 .97106 L
.12607 .97606 L
.12841 .98077 L
.01923 .98077 L
F
0 g
.01923 .87159 m
.02394 .87393 L
.02894 .87696 L
.03866 .88341 L
.03898 .88364 L
.04837 .89064 L
.05182 .89336 L
.05808 .8985 L
.06342 .90307 L
.06779 .90697 L
.07406 .91278 L
.07751 .9161 L
.0839 .92249 L
.08722 .92594 L
.09303 .93221 L
.09693 .93658 L
.1015 .94192 L
.10664 .94818 L
.10936 .95163 L
.11636 .96102 L
.11659 .96134 L
.12304 .97106 L
.12607 .97606 L
.12841 .98077 L
s
0 .64 1 r
.01923 .87893 m
.02363 .88364 L
.02894 .88889 L
.03349 .89336 L
.03866 .89836 L
.04351 .90307 L
.04837 .90776 L
.05352 .91278 L
.05808 .91724 L
.06341 .92249 L
.06779 .92684 L
.07316 .93221 L
.07751 .93659 L
.08276 .94192 L
.08722 .94648 L
.09224 .95163 L
.09693 .95649 L
.10164 .96134 L
.10664 .96651 L
.11111 .97106 L
.11636 .97637 L
.12107 .98077 L
.01923 .98077 L
F
0 g
.01923 .87893 m
.02363 .88364 L
.02894 .88889 L
.03349 .89336 L
.03866 .89836 L
.04351 .90307 L
.04837 .90776 L
.05352 .91278 L
.05808 .91724 L
.06341 .92249 L
.06779 .92684 L
.07316 .93221 L
.07751 .93659 L
.08276 .94192 L
.08722 .94648 L
.09224 .95163 L
.09693 .95649 L
.10164 .96134 L
.10664 .96651 L
.11111 .97106 L
.11636 .97637 L
.12107 .98077 L
s
0 .88 1 r
.01923 .10192 m
.02592 .09693 L
.02894 .09349 L
.036 .08722 L
.03866 .08475 L
.04647 .07751 L
.04837 .07569 L
.05656 .06779 L
.05808 .0663 L
.0663 .05808 L
.06779 .05656 L
.07569 .04837 L
.07751 .04647 L
.08475 .03866 L
.08722 .036 L
.09349 .02894 L
.09693 .02592 L
.10192 .01923 L
.01923 .01923 L
F
0 g
.01923 .10192 m
.02592 .09693 L
.02894 .09349 L
.036 .08722 L
.03866 .08475 L
.04647 .07751 L
.04837 .07569 L
.05656 .06779 L
.05808 .0663 L
.0663 .05808 L
.06779 .05656 L
.07569 .04837 L
.07751 .04647 L
.08475 .03866 L
.08722 .036 L
.09349 .02894 L
.09693 .02592 L
.10192 .01923 L
s
0 .88 1 r
.01923 .89808 m
.02592 .90307 L
.02894 .90651 L
.036 .91278 L
.03866 .91525 L
.04647 .92249 L
.04837 .92431 L
.05656 .93221 L
.05808 .9337 L
.0663 .94192 L
.06779 .94344 L
.07569 .95163 L
.07751 .95353 L
.08475 .96134 L
.08722 .964 L
.09349 .97106 L
.09693 .97408 L
.10192 .98077 L
.01923 .98077 L
F
0 g
.01923 .89808 m
.02592 .90307 L
.02894 .90651 L
.036 .91278 L
.03866 .91525 L
.04647 .92249 L
.04837 .92431 L
.05656 .93221 L
.05808 .9337 L
.0663 .94192 L
.06779 .94344 L
.07569 .95163 L
.07751 .95353 L
.08475 .96134 L
.08722 .964 L
.09349 .97106 L
.09693 .97408 L
.10192 .98077 L
s
0 1 .88 r
.01923 .09183 m
.02469 .08722 L
.02894 .08328 L
.03532 .07751 L
.03866 .07441 L
.04568 .06779 L
.04837 .06521 L
.05564 .05808 L
.05808 .05564 L
.06521 .04837 L
.06779 .04568 L
.07441 .03866 L
.07751 .03532 L
.08328 .02894 L
.08722 .02469 L
.09183 .01923 L
.01923 .01923 L
F
0 g
.01923 .09183 m
.02469 .08722 L
.02894 .08328 L
.03532 .07751 L
.03866 .07441 L
.04568 .06779 L
.04837 .06521 L
.05564 .05808 L
.05808 .05564 L
.06521 .04837 L
.06779 .04568 L
.07441 .03866 L
.07751 .03532 L
.08328 .02894 L
.08722 .02469 L
.09183 .01923 L
s
0 1 .88 r
.01923 .90817 m
.02469 .91278 L
.02894 .91672 L
.03532 .92249 L
.03866 .92559 L
.04568 .93221 L
.04837 .93479 L
.05564 .94192 L
.05808 .94436 L
.06521 .95163 L
.06779 .95432 L
.07441 .96134 L
.07751 .96468 L
.08328 .97106 L
.08722 .97531 L
.09183 .98077 L
.01923 .98077 L
F
0 g
.01923 .90817 m
.02469 .91278 L
.02894 .91672 L
.03532 .92249 L
.03866 .92559 L
.04568 .93221 L
.04837 .93479 L
.05564 .94192 L
.05808 .94436 L
.06521 .95163 L
.06779 .95432 L
.07441 .96134 L
.07751 .96468 L
.08328 .97106 L
.08722 .97531 L
.09183 .98077 L
s
0 1 .64 r
.01923 .08297 m
.02544 .07751 L
.02894 .07432 L
.03603 .06779 L
.03866 .06533 L
.04621 .05808 L
.04837 .05596 L
.05596 .04837 L
.05808 .04621 L
.06533 .03866 L
.06779 .03603 L
.07432 .02894 L
.07751 .02544 L
.08297 .01923 L
.01923 .01923 L
F
0 g
.01923 .08297 m
.02544 .07751 L
.02894 .07432 L
.03603 .06779 L
.03866 .06533 L
.04621 .05808 L
.04837 .05596 L
.05596 .04837 L
.05808 .04621 L
.06533 .03866 L
.06779 .03603 L
.07432 .02894 L
.07751 .02544 L
.08297 .01923 L
s
0 1 .64 r
.01923 .91703 m
.02544 .92249 L
.02894 .92568 L
.03603 .93221 L
.03866 .93467 L
.04621 .94192 L
.04837 .94404 L
.05596 .95163 L
.05808 .95379 L
.06533 .96134 L
.06779 .96397 L
.07432 .97106 L
.07751 .97456 L
.08297 .98077 L
.01923 .98077 L
F
0 g
.01923 .91703 m
.02544 .92249 L
.02894 .92568 L
.03603 .93221 L
.03866 .93467 L
.04621 .94192 L
.04837 .94404 L
.05596 .95163 L
.05808 .95379 L
.06533 .96134 L
.06779 .96397 L
.07432 .97106 L
.07751 .97456 L
.08297 .98077 L
s
0 1 .4 r
.01923 .07434 m
.0265 .06779 L
.02894 .06557 L
.03692 .05808 L
.03866 .05642 L
.04688 .04837 L
.04837 .04688 L
.05642 .03866 L
.05808 .03692 L
.06557 .02894 L
.06779 .0265 L
.07434 .01923 L
.01923 .01923 L
F
0 g
.01923 .07434 m
.0265 .06779 L
.02894 .06557 L
.03692 .05808 L
.03866 .05642 L
.04688 .04837 L
.04837 .04688 L
.05642 .03866 L
.05808 .03692 L
.06557 .02894 L
.06779 .0265 L
.07434 .01923 L
s
0 1 .4 r
.01923 .92566 m
.0265 .93221 L
.02894 .93443 L
.03692 .94192 L
.03866 .94358 L
.04688 .95163 L
.04837 .95312 L
.05642 .96134 L
.05808 .96308 L
.06557 .97106 L
.06779 .9735 L
.07434 .98077 L
.01923 .98077 L
F
0 g
.01923 .92566 m
.0265 .93221 L
.02894 .93443 L
.03692 .94192 L
.03866 .94358 L
.04688 .95163 L
.04837 .95312 L
.05642 .96134 L
.05808 .96308 L
.06557 .97106 L
.06779 .9735 L
.07434 .98077 L
s
0 1 .16 r
.01923 .06516 m
.0269 .05808 L
.02894 .05623 L
.03716 .04837 L
.03866 .04691 L
.04691 .03866 L
.04837 .03716 L
.05623 .02894 L
.05808 .0269 L
.06516 .01923 L
.01923 .01923 L
F
0 g
.01923 .06516 m
.0269 .05808 L
.02894 .05623 L
.03716 .04837 L
.03866 .04691 L
.04691 .03866 L
.04837 .03716 L
.05623 .02894 L
.05808 .0269 L
.06516 .01923 L
s
0 1 .16 r
.01923 .93484 m
.0269 .94192 L
.02894 .94377 L
.03716 .95163 L
.03866 .95309 L
.04691 .96134 L
.04837 .96284 L
.05623 .97106 L
.05808 .9731 L
.06516 .98077 L
.01923 .98077 L
F
0 g
.01923 .93484 m
.0269 .94192 L
.02894 .94377 L
.03716 .95163 L
.03866 .95309 L
.04691 .96134 L
.04837 .96284 L
.05623 .97106 L
.05808 .9731 L
.06516 .98077 L
s
.08 1 0 r
.01923 .05419 m
.02506 .04837 L
.02894 .04509 L
.03551 .03866 L
.03866 .03551 L
.04509 .02894 L
.04837 .02506 L
.05419 .01923 L
.01923 .01923 L
F
0 g
.01923 .05419 m
.02506 .04837 L
.02894 .04509 L
.03551 .03866 L
.03866 .03551 L
.04509 .02894 L
.04837 .02506 L
.05419 .01923 L
s
.08 1 0 r
.01923 .94581 m
.02506 .95163 L
.02894 .95491 L
.03551 .96134 L
.03866 .96449 L
.04509 .97106 L
.04837 .97494 L
.05419 .98077 L
.01923 .98077 L
F
0 g
.01923 .94581 m
.02506 .95163 L
.02894 .95491 L
.03551 .96134 L
.03866 .96449 L
.04509 .97106 L
.04837 .97494 L
.05419 .98077 L
s
0 .4 1 r
.01923 .12841 m
.02394 .12607 L
.01923 .12107 L
F
0 g
.01923 .12841 m
.02394 .12607 L
.01923 .12107 L
s
0 .4 1 r
.11636 .02363 m
.12304 .02894 L
.11659 .03866 L
.11636 .03898 L
.10936 .04837 L
.10664 .05182 L
.1015 .05808 L
.09693 .06342 L
.09303 .06779 L
.08722 .07406 L
.0839 .07751 L
.07751 .0839 L
.07406 .08722 L
.06779 .09303 L
.06342 .09693 L
.05808 .1015 L
.05182 .10664 L
.04837 .10936 L
.03898 .11636 L
.03866 .11659 L
.02894 .12304 L
.02363 .11636 L
.02894 .11111 L
.03349 .10664 L
.03866 .10164 L
.04351 .09693 L
.04837 .09224 L
.05352 .08722 L
.05808 .08276 L
.06341 .07751 L
.06779 .07316 L
.07316 .06779 L
.07751 .06341 L
.08276 .05808 L
.08722 .05352 L
.09224 .04837 L
.09693 .04351 L
.10164 .03866 L
.10664 .03349 L
.11111 .02894 L
F
0 g
.11636 .02363 m
.12304 .02894 L
.11659 .03866 L
.11636 .03898 L
.10936 .04837 L
.10664 .05182 L
.1015 .05808 L
.09693 .06342 L
.09303 .06779 L
.08722 .07406 L
.0839 .07751 L
.07751 .0839 L
.07406 .08722 L
.06779 .09303 L
.06342 .09693 L
.05808 .1015 L
.05182 .10664 L
.04837 .10936 L
.03898 .11636 L
.03866 .11659 L
.02894 .12304 L
.02363 .11636 L
.02894 .11111 L
.03349 .10664 L
.03866 .10164 L
.04351 .09693 L
.04837 .09224 L
.05352 .08722 L
.05808 .08276 L
.06341 .07751 L
.06779 .07316 L
.07316 .06779 L
.07751 .06341 L
.08276 .05808 L
.08722 .05352 L
.09224 .04837 L
.09693 .04351 L
.10164 .03866 L
.10664 .03349 L
.11111 .02894 L
.11636 .02363 L
s
.08 0 1 r
.30089 .03687 m
.30984 .03866 L
.30927 .04837 L
.30749 .05808 L
.30511 .06779 L
.30255 .07751 L
.30089 .08222 L
.29907 .08722 L
.29607 .09693 L
.29166 .10664 L
.29118 .10773 L
.28782 .11636 L
.28289 .12607 L
.28147 .12887 L
.27817 .13578 L
.27262 .14549 L
.27176 .14698 L
.26705 .15521 L
.26204 .163 L
.26086 .16492 L
.2542 .17463 L
.25233 .17723 L
.24714 .18434 L
.24262 .19011 L
.23948 .19406 L
.23291 .20176 L
.23116 .20377 L
.22319 .21239 L
.22215 .21348 L
.21348 .22215 L
.21239 .22319 L
.20377 .23116 L
.20176 .23291 L
.19406 .23948 L
.19011 .24262 L
.18434 .24714 L
.17723 .25233 L
.17463 .2542 L
.16492 .26086 L
.163 .26204 L
.15521 .26705 L
.14698 .27176 L
.14549 .27262 L
.13578 .27817 L
.12887 .28147 L
.12607 .28289 L
.11636 .28782 L
.10773 .29118 L
.10664 .29166 L
.09693 .29607 L
.08722 .29907 L
.08222 .30089 L
.07751 .30255 L
.06779 .30511 L
.05808 .30749 L
.04837 .30927 L
.03866 .30984 L
.03687 .30089 L
.03866 .2993 L
.04837 .29152 L
.04879 .29118 L
.05808 .28461 L
.06252 .28147 L
.06779 .27653 L
.07675 .27176 L
.07751 .27117 L
.08722 .26332 L
.09072 .26204 L
.09693 .25722 L
.10436 .25233 L
.10664 .25057 L
.11636 .24327 L
.11747 .24262 L
.12607 .23576 L
.13001 .23291 L
.13578 .22815 L
.14202 .22319 L
.14549 .22024 L
.15348 .21348 L
.15521 .21196 L
.16441 .20377 L
.16492 .2033 L
.17463 .19423 L
.17482 .19406 L
.18434 .18474 L
.18474 .18434 L
.19406 .17482 L
.19423 .17463 L
.2033 .16492 L
.20377 .16441 L
.21196 .15521 L
.21348 .15348 L
.22024 .14549 L
.22319 .14202 L
.22815 .13578 L
.23291 .13001 L
.23576 .12607 L
.24262 .11747 L
.24327 .11636 L
.25057 .10664 L
.25233 .10436 L
.25722 .09693 L
.26204 .09072 L
.26332 .08722 L
.27117 .07751 L
.27176 .07675 L
.27653 .06779 L
.28147 .06252 L
.28461 .05808 L
.29118 .04879 L
.29152 .04837 L
.2993 .03866 L
F
0 g
.30089 .03687 m
.30984 .03866 L
.30927 .04837 L
.30749 .05808 L
.30511 .06779 L
.30255 .07751 L
.30089 .08222 L
.29907 .08722 L
.29607 .09693 L
.29166 .10664 L
.29118 .10773 L
.28782 .11636 L
.28289 .12607 L
.28147 .12887 L
.27817 .13578 L
.27262 .14549 L
.27176 .14698 L
.26705 .15521 L
.26204 .163 L
.26086 .16492 L
.2542 .17463 L
.25233 .17723 L
.24714 .18434 L
.24262 .19011 L
.23948 .19406 L
.23291 .20176 L
.23116 .20377 L
.22319 .21239 L
.22215 .21348 L
.21348 .22215 L
.21239 .22319 L
.20377 .23116 L
.20176 .23291 L
.19406 .23948 L
.19011 .24262 L
.18434 .24714 L
.17723 .25233 L
.17463 .2542 L
.16492 .26086 L
.163 .26204 L
.15521 .26705 L
.14698 .27176 L
.14549 .27262 L
.13578 .27817 L
.12887 .28147 L
.12607 .28289 L
.11636 .28782 L
.10773 .29118 L
.10664 .29166 L
.09693 .29607 L
Mistroke
.08722 .29907 L
.08222 .30089 L
.07751 .30255 L
.06779 .30511 L
.05808 .30749 L
.04837 .30927 L
.03866 .30984 L
.03687 .30089 L
.03866 .2993 L
.04837 .29152 L
.04879 .29118 L
.05808 .28461 L
.06252 .28147 L
.06779 .27653 L
.07675 .27176 L
.07751 .27117 L
.08722 .26332 L
.09072 .26204 L
.09693 .25722 L
.10436 .25233 L
.10664 .25057 L
.11636 .24327 L
.11747 .24262 L
.12607 .23576 L
.13001 .23291 L
.13578 .22815 L
.14202 .22319 L
.14549 .22024 L
.15348 .21348 L
.15521 .21196 L
.16441 .20377 L
.16492 .2033 L
.17463 .19423 L
.17482 .19406 L
.18434 .18474 L
.18474 .18434 L
.19406 .17482 L
.19423 .17463 L
.2033 .16492 L
.20377 .16441 L
.21196 .15521 L
.21348 .15348 L
.22024 .14549 L
.22319 .14202 L
.22815 .13578 L
.23291 .13001 L
.23576 .12607 L
.24262 .11747 L
.24327 .11636 L
.25057 .10664 L
Mistroke
.25233 .10436 L
.25722 .09693 L
.26204 .09072 L
.26332 .08722 L
.27117 .07751 L
.27176 .07675 L
.27653 .06779 L
.28147 .06252 L
.28461 .05808 L
.29118 .04879 L
.29152 .04837 L
.2993 .03866 L
.30089 .03687 L
Mfstroke
.08 0 1 r
.03866 .69016 m
.04837 .69073 L
.05808 .69251 L
.06779 .69489 L
.07751 .69745 L
.08222 .69911 L
.08722 .70093 L
.09693 .70393 L
.10664 .70834 L
.10773 .70882 L
.11636 .71218 L
.12607 .71711 L
.12887 .71853 L
.13578 .72183 L
.14549 .72738 L
.14698 .72824 L
.15521 .73295 L
.163 .73796 L
.16492 .73914 L
.17463 .7458 L
.17723 .74767 L
.18434 .75286 L
.19011 .75738 L
.19406 .76052 L
.20176 .76709 L
.20377 .76884 L
.21239 .77681 L
.21348 .77785 L
.22215 .78652 L
.22319 .78761 L
.23116 .79623 L
.23291 .79824 L
.23948 .80594 L
.24262 .80989 L
.24714 .81566 L
.25233 .82277 L
.2542 .82537 L
.26086 .83508 L
.26204 .837 L
.26705 .84479 L
.27176 .85302 L
.27262 .85451 L
.27817 .86422 L
.28147 .87113 L
.28289 .87393 L
.28782 .88364 L
.29118 .89227 L
.29166 .89336 L
.29607 .90307 L
.29907 .91278 L
.30089 .91778 L
.30255 .92249 L
.30511 .93221 L
.30749 .94192 L
.30927 .95163 L
.30984 .96134 L
.30089 .96313 L
.2993 .96134 L
.29152 .95163 L
.29118 .95121 L
.28461 .94192 L
.28147 .93748 L
.27653 .93221 L
.27176 .92325 L
.27117 .92249 L
.26332 .91278 L
.26204 .90928 L
.25722 .90307 L
.25233 .89564 L
.25057 .89336 L
.24327 .88364 L
.24262 .88253 L
.23576 .87393 L
.23291 .86999 L
.22815 .86422 L
.22319 .85798 L
.22024 .85451 L
.21348 .84652 L
.21196 .84479 L
.20377 .83559 L
.2033 .83508 L
.19423 .82537 L
.19406 .82518 L
.18474 .81566 L
.18434 .81526 L
.17482 .80594 L
.17463 .80577 L
.16492 .7967 L
.16441 .79623 L
.15521 .78804 L
.15348 .78652 L
.14549 .77976 L
.14202 .77681 L
.13578 .77185 L
.13001 .76709 L
.12607 .76424 L
.11747 .75738 L
.11636 .75673 L
.10664 .74943 L
.10436 .74767 L
.09693 .74278 L
.09072 .73796 L
.08722 .73668 L
.07751 .72883 L
.07675 .72824 L
.06779 .72347 L
.06252 .71853 L
.05808 .71539 L
.04879 .70882 L
.04837 .70848 L
.03866 .7007 L
.03687 .69911 L
F
0 g
.03866 .69016 m
.04837 .69073 L
.05808 .69251 L
.06779 .69489 L
.07751 .69745 L
.08222 .69911 L
.08722 .70093 L
.09693 .70393 L
.10664 .70834 L
.10773 .70882 L
.11636 .71218 L
.12607 .71711 L
.12887 .71853 L
.13578 .72183 L
.14549 .72738 L
.14698 .72824 L
.15521 .73295 L
.163 .73796 L
.16492 .73914 L
.17463 .7458 L
.17723 .74767 L
.18434 .75286 L
.19011 .75738 L
.19406 .76052 L
.20176 .76709 L
.20377 .76884 L
.21239 .77681 L
.21348 .77785 L
.22215 .78652 L
.22319 .78761 L
.23116 .79623 L
.23291 .79824 L
.23948 .80594 L
.24262 .80989 L
.24714 .81566 L
.25233 .82277 L
.2542 .82537 L
.26086 .83508 L
.26204 .837 L
.26705 .84479 L
.27176 .85302 L
.27262 .85451 L
.27817 .86422 L
.28147 .87113 L
.28289 .87393 L
.28782 .88364 L
.29118 .89227 L
.29166 .89336 L
.29607 .90307 L
.29907 .91278 L
Mistroke
.30089 .91778 L
.30255 .92249 L
.30511 .93221 L
.30749 .94192 L
.30927 .95163 L
.30984 .96134 L
.30089 .96313 L
.2993 .96134 L
.29152 .95163 L
.29118 .95121 L
.28461 .94192 L
.28147 .93748 L
.27653 .93221 L
.27176 .92325 L
.27117 .92249 L
.26332 .91278 L
.26204 .90928 L
.25722 .90307 L
.25233 .89564 L
.25057 .89336 L
.24327 .88364 L
.24262 .88253 L
.23576 .87393 L
.23291 .86999 L
.22815 .86422 L
.22319 .85798 L
.22024 .85451 L
.21348 .84652 L
.21196 .84479 L
.20377 .83559 L
.2033 .83508 L
.19423 .82537 L
.19406 .82518 L
.18474 .81566 L
.18434 .81526 L
.17482 .80594 L
.17463 .80577 L
.16492 .7967 L
.16441 .79623 L
.15521 .78804 L
.15348 .78652 L
.14549 .77976 L
.14202 .77681 L
.13578 .77185 L
.13001 .76709 L
.12607 .76424 L
.11747 .75738 L
.11636 .75673 L
.10664 .74943 L
.10436 .74767 L
Mistroke
.09693 .74278 L
.09072 .73796 L
.08722 .73668 L
.07751 .72883 L
.07675 .72824 L
.06779 .72347 L
.06252 .71853 L
.05808 .71539 L
.04879 .70882 L
.04837 .70848 L
.03866 .7007 L
.03687 .69911 L
.03866 .69016 L
Mfstroke
.8 1 0 r
.04837 .19587 m
.05323 .20377 L
.04837 .2054 L
.04225 .20377 L
F
0 g
.04837 .19587 m
.05323 .20377 L
.04837 .2054 L
.04225 .20377 L
.04837 .19587 L
s
.8 1 0 r
.04837 .7946 m
.05323 .79623 L
.05808 .79828 L
.06779 .80309 L
.07299 .80594 L
.07751 .80858 L
.08722 .8143 L
.08916 .81566 L
.09693 .82129 L
.10257 .82537 L
.10664 .82845 L
.11504 .83508 L
.11636 .83617 L
.12607 .84443 L
.12647 .84479 L
.13578 .85328 L
.13705 .85451 L
.14549 .86295 L
.14672 .86422 L
.15521 .87353 L
.15557 .87393 L
.16383 .88364 L
.16492 .88496 L
.17155 .89336 L
.17463 .89743 L
.17871 .90307 L
.18434 .91084 L
.1857 .91278 L
.19142 .92249 L
.19406 .92701 L
.19691 .93221 L
.20172 .94192 L
.20377 .94677 L
.2054 .95163 L
.20377 .95775 L
.19587 .95163 L
.19406 .94987 L
.18653 .94192 L
.18434 .93954 L
.17775 .93221 L
.17463 .93047 L
.16905 .92249 L
.16492 .91893 L
.16028 .91278 L
.15521 .90794 L
.14986 .90307 L
.14549 .89735 L
.14138 .89336 L
.13578 .88711 L
.13228 .88364 L
.12607 .87723 L
.12277 .87393 L
.11636 .86772 L
.11289 .86422 L
.10664 .85862 L
.10265 .85451 L
.09693 .85014 L
.09206 .84479 L
.08722 .83972 L
.08107 .83508 L
.07751 .83095 L
.06953 .82537 L
.06779 .82225 L
.06046 .81566 L
.05808 .81347 L
.05013 .80594 L
.04837 .80413 L
.04225 .79623 L
F
0 g
.04837 .7946 m
.05323 .79623 L
.05808 .79828 L
.06779 .80309 L
.07299 .80594 L
.07751 .80858 L
.08722 .8143 L
.08916 .81566 L
.09693 .82129 L
.10257 .82537 L
.10664 .82845 L
.11504 .83508 L
.11636 .83617 L
.12607 .84443 L
.12647 .84479 L
.13578 .85328 L
.13705 .85451 L
.14549 .86295 L
.14672 .86422 L
.15521 .87353 L
.15557 .87393 L
.16383 .88364 L
.16492 .88496 L
.17155 .89336 L
.17463 .89743 L
.17871 .90307 L
.18434 .91084 L
.1857 .91278 L
.19142 .92249 L
.19406 .92701 L
.19691 .93221 L
.20172 .94192 L
.20377 .94677 L
.2054 .95163 L
.20377 .95775 L
.19587 .95163 L
.19406 .94987 L
.18653 .94192 L
.18434 .93954 L
.17775 .93221 L
.17463 .93047 L
.16905 .92249 L
.16492 .91893 L
.16028 .91278 L
.15521 .90794 L
.14986 .90307 L
.14549 .89735 L
.14138 .89336 L
.13578 .88711 L
.13228 .88364 L
Mistroke
.12607 .87723 L
.12277 .87393 L
.11636 .86772 L
.11289 .86422 L
.10664 .85862 L
.10265 .85451 L
.09693 .85014 L
.09206 .84479 L
.08722 .83972 L
.08107 .83508 L
.07751 .83095 L
.06953 .82537 L
.06779 .82225 L
.06046 .81566 L
.05808 .81347 L
.05013 .80594 L
.04837 .80413 L
.04225 .79623 L
.04837 .7946 L
Mfstroke
.32 0 1 r
.49514 .04524 m
.50486 .04524 L
.50659 .04837 L
.51036 .05808 L
.51322 .06779 L
.51457 .07298 L
.51585 .07751 L
.51834 .08722 L
.52049 .09693 L
.52235 .10664 L
.52396 .11636 L
.52428 .1184 L
.52583 .12607 L
.52776 .13578 L
.52966 .14549 L
.53153 .15521 L
.53339 .16492 L
.53399 .16807 L
.53534 .17463 L
.53737 .18434 L
.53945 .19406 L
.54158 .20377 L
.54371 .21315 L
.54378 .21348 L
.54613 .22319 L
.54857 .23291 L
.55114 .24262 L
.55342 .25085 L
.55385 .25233 L
.55674 .26204 L
.55982 .27176 L
.56312 .28147 L
.56313 .28151 L
.56668 .29118 L
.57052 .30089 L
.57284 .3064 L
.5747 .31061 L
.57927 .32032 L
.58256 .32677 L
.5843 .33003 L
.58986 .33974 L
.59227 .34364 L
.59606 .34946 L
.60198 .3578 L
.60301 .35917 L
.61087 .36888 L
.61169 .36983 L
.61984 .37859 L
.62141 .38016 L
.63017 .38831 L
.63112 .38913 L
.64083 .39699 L
.6422 .39802 L
.65054 .40394 L
.65636 .40773 L
.66026 .41014 L
.66997 .4157 L
.67323 .41744 L
.67968 .42073 L
.68939 .4253 L
.6936 .42716 L
.69911 .42948 L
.70882 .43332 L
.71849 .43687 L
.71853 .43688 L
.72824 .44018 L
.73796 .44326 L
.74767 .44615 L
.74915 .44658 L
.75738 .44886 L
.76709 .45143 L
.77681 .45387 L
.78652 .45622 L
.78685 .45629 L
.79623 .45842 L
.80594 .46055 L
.81566 .46263 L
.82537 .46466 L
.83193 .46601 L
.83508 .46661 L
.84479 .46847 L
.85451 .47034 L
.86422 .47224 L
.87393 .47417 L
.8816 .47572 L
.88364 .47604 L
.89336 .47765 L
.90307 .47951 L
.91278 .48166 L
.92249 .48415 L
.92702 .48543 L
.93221 .48678 L
.94192 .48964 L
.95163 .49341 L
.95476 .49514 L
.95476 .50486 L
.95163 .50659 L
.94192 .51036 L
.93221 .51322 L
.92702 .51457 L
.92249 .51585 L
.91278 .51834 L
.90307 .52049 L
.89336 .52235 L
.88364 .52396 L
.8816 .52428 L
.87393 .52583 L
.86422 .52776 L
.85451 .52966 L
.84479 .53153 L
.83508 .53339 L
.83193 .53399 L
.82537 .53534 L
.81566 .53737 L
.80594 .53945 L
.79623 .54158 L
.78685 .54371 L
.78652 .54378 L
.77681 .54613 L
.76709 .54857 L
.75738 .55114 L
.74915 .55342 L
.74767 .55385 L
.73796 .55674 L
.72824 .55982 L
.71853 .56312 L
.71849 .56313 L
.70882 .56668 L
.69911 .57052 L
.6936 .57284 L
.68939 .5747 L
.67968 .57927 L
.67323 .58256 L
.66997 .5843 L
.66026 .58986 L
.65636 .59227 L
.65054 .59606 L
.6422 .60198 L
.64083 .60301 L
.63112 .61087 L
.63017 .61169 L
.62141 .61984 L
.61984 .62141 L
.61169 .63017 L
.61087 .63112 L
.60301 .64083 L
.60198 .6422 L
.59606 .65054 L
.59227 .65636 L
.58986 .66026 L
.5843 .66997 L
.58256 .67323 L
.57927 .67968 L
.5747 .68939 L
.57284 .6936 L
.57052 .69911 L
.56668 .70882 L
.56313 .71849 L
.56312 .71853 L
.55982 .72824 L
.55674 .73796 L
.55385 .74767 L
.55342 .74915 L
.55114 .75738 L
.54857 .76709 L
.54613 .77681 L
.54378 .78652 L
.54371 .78685 L
.54158 .79623 L
.53945 .80594 L
.53737 .81566 L
.53534 .82537 L
.53399 .83193 L
.53339 .83508 L
.53153 .84479 L
.52966 .85451 L
.52776 .86422 L
.52583 .87393 L
.52428 .8816 L
.52396 .88364 L
.52235 .89336 L
.52049 .90307 L
.51834 .91278 L
.51585 .92249 L
.51457 .92702 L
.51322 .93221 L
.51036 .94192 L
.50659 .95163 L
.50486 .95476 L
.49514 .95476 L
.49341 .95163 L
.48964 .94192 L
.48678 .93221 L
.48543 .92702 L
.48415 .92249 L
.48166 .91278 L
.47951 .90307 L
.47765 .89336 L
.47604 .88364 L
.47572 .8816 L
.47417 .87393 L
.47224 .86422 L
.47034 .85451 L
.46847 .84479 L
.46661 .83508 L
.46601 .83193 L
.46466 .82537 L
.46263 .81566 L
.46055 .80594 L
.45842 .79623 L
.45629 .78685 L
.45622 .78652 L
.45387 .77681 L
.45143 .76709 L
.44886 .75738 L
.44658 .74915 L
.44615 .74767 L
.44326 .73796 L
.44018 .72824 L
.43688 .71853 L
.43687 .71849 L
.43332 .70882 L
.42948 .69911 L
.42716 .6936 L
.4253 .68939 L
.42073 .67968 L
.41744 .67323 L
.4157 .66997 L
.41014 .66026 L
.40773 .65636 L
.40394 .65054 L
.39802 .6422 L
.39699 .64083 L
.38913 .63112 L
.38831 .63017 L
.38016 .62141 L
.37859 .61984 L
.36983 .61169 L
.36888 .61087 L
.35917 .60301 L
.3578 .60198 L
.34946 .59606 L
.34364 .59227 L
.33974 .58986 L
.33003 .5843 L
.32677 .58256 L
.32032 .57927 L
.31061 .5747 L
.3064 .57284 L
.30089 .57052 L
.29118 .56668 L
.28151 .56313 L
.28147 .56312 L
.27176 .55982 L
.26204 .55674 L
.25233 .55385 L
.25085 .55342 L
.24262 .55114 L
.23291 .54857 L
.22319 .54613 L
.21348 .54378 L
.21315 .54371 L
.20377 .54158 L
.19406 .53945 L
.18434 .53737 L
.17463 .53534 L
.16807 .53399 L
.16492 .53339 L
.15521 .53153 L
.14549 .52966 L
.13578 .52776 L
.12607 .52583 L
.1184 .52428 L
.11636 .52396 L
.10664 .52235 L
.09693 .52049 L
.08722 .51834 L
.07751 .51585 L
.07298 .51457 L
.06779 .51322 L
.05808 .51036 L
.04837 .50659 L
.04524 .50486 L
.04524 .49514 L
.04837 .49341 L
.05808 .48964 L
.06779 .48678 L
.07298 .48543 L
.07751 .48415 L
.08722 .48166 L
.09693 .47951 L
.10664 .47765 L
.11636 .47604 L
.1184 .47572 L
.12607 .47417 L
.13578 .47224 L
.14549 .47034 L
.15521 .46847 L
.16492 .46661 L
.16807 .46601 L
.17463 .46466 L
.18434 .46263 L
.19406 .46055 L
.20377 .45842 L
.21315 .45629 L
.21348 .45622 L
.22319 .45387 L
.23291 .45143 L
.24262 .44886 L
.25085 .44658 L
.25233 .44615 L
.26204 .44326 L
.27176 .44018 L
.28147 .43688 L
.28151 .43687 L
.29118 .43332 L
.30089 .42948 L
.3064 .42716 L
.31061 .4253 L
.32032 .42073 L
.32677 .41744 L
.33003 .4157 L
.33974 .41014 L
.34364 .40773 L
.34946 .40394 L
.3578 .39802 L
.35917 .39699 L
.36888 .38913 L
.36983 .38831 L
.37859 .38016 L
.38016 .37859 L
.38831 .36983 L
.38913 .36888 L
.39699 .35917 L
.39802 .3578 L
.40394 .34946 L
.40773 .34364 L
.41014 .33974 L
.4157 .33003 L
.41744 .32677 L
.42073 .32032 L
.4253 .31061 L
.42716 .3064 L
.42948 .30089 L
.43332 .29118 L
.43687 .28151 L
.43688 .28147 L
.44018 .27176 L
.44326 .26204 L
.44615 .25233 L
.44658 .25085 L
.44886 .24262 L
.45143 .23291 L
.45387 .22319 L
.45622 .21348 L
.45629 .21315 L
.45842 .20377 L
.46055 .19406 L
.46263 .18434 L
.46466 .17463 L
.46601 .16807 L
.46661 .16492 L
.46847 .15521 L
.47034 .14549 L
.47224 .13578 L
.47417 .12607 L
.47572 .1184 L
.47604 .11636 L
.47765 .10664 L
.47951 .09693 L
.48166 .08722 L
.48415 .07751 L
.48543 .07298 L
.48678 .06779 L
.48964 .05808 L
.49341 .04837 L
F
0 g
.49514 .04524 m
.50486 .04524 L
.50659 .04837 L
.51036 .05808 L
.51322 .06779 L
.51457 .07298 L
.51585 .07751 L
.51834 .08722 L
.52049 .09693 L
.52235 .10664 L
.52396 .11636 L
.52428 .1184 L
.52583 .12607 L
.52776 .13578 L
.52966 .14549 L
.53153 .15521 L
.53339 .16492 L
.53399 .16807 L
.53534 .17463 L
.53737 .18434 L
.53945 .19406 L
.54158 .20377 L
.54371 .21315 L
.54378 .21348 L
.54613 .22319 L
.54857 .23291 L
.55114 .24262 L
.55342 .25085 L
.55385 .25233 L
.55674 .26204 L
.55982 .27176 L
.56312 .28147 L
.56313 .28151 L
.56668 .29118 L
.57052 .30089 L
.57284 .3064 L
.5747 .31061 L
.57927 .32032 L
.58256 .32677 L
.5843 .33003 L
.58986 .33974 L
.59227 .34364 L
.59606 .34946 L
.60198 .3578 L
.60301 .35917 L
.61087 .36888 L
.61169 .36983 L
.61984 .37859 L
.62141 .38016 L
.63017 .38831 L
Mistroke
.63112 .38913 L
.64083 .39699 L
.6422 .39802 L
.65054 .40394 L
.65636 .40773 L
.66026 .41014 L
.66997 .4157 L
.67323 .41744 L
.67968 .42073 L
.68939 .4253 L
.6936 .42716 L
.69911 .42948 L
.70882 .43332 L
.71849 .43687 L
.71853 .43688 L
.72824 .44018 L
.73796 .44326 L
.74767 .44615 L
.74915 .44658 L
.75738 .44886 L
.76709 .45143 L
.77681 .45387 L
.78652 .45622 L
.78685 .45629 L
.79623 .45842 L
.80594 .46055 L
.81566 .46263 L
.82537 .46466 L
.83193 .46601 L
.83508 .46661 L
.84479 .46847 L
.85451 .47034 L
.86422 .47224 L
.87393 .47417 L
.8816 .47572 L
.88364 .47604 L
.89336 .47765 L
.90307 .47951 L
.91278 .48166 L
.92249 .48415 L
.92702 .48543 L
.93221 .48678 L
.94192 .48964 L
.95163 .49341 L
.95476 .49514 L
.95476 .50486 L
.95163 .50659 L
.94192 .51036 L
.93221 .51322 L
.92702 .51457 L
Mistroke
.92249 .51585 L
.91278 .51834 L
.90307 .52049 L
.89336 .52235 L
.88364 .52396 L
.8816 .52428 L
.87393 .52583 L
.86422 .52776 L
.85451 .52966 L
.84479 .53153 L
.83508 .53339 L
.83193 .53399 L
.82537 .53534 L
.81566 .53737 L
.80594 .53945 L
.79623 .54158 L
.78685 .54371 L
.78652 .54378 L
.77681 .54613 L
.76709 .54857 L
.75738 .55114 L
.74915 .55342 L
.74767 .55385 L
.73796 .55674 L
.72824 .55982 L
.71853 .56312 L
.71849 .56313 L
.70882 .56668 L
.69911 .57052 L
.6936 .57284 L
.68939 .5747 L
.67968 .57927 L
.67323 .58256 L
.66997 .5843 L
.66026 .58986 L
.65636 .59227 L
.65054 .59606 L
.6422 .60198 L
.64083 .60301 L
.63112 .61087 L
.63017 .61169 L
.62141 .61984 L
.61984 .62141 L
.61169 .63017 L
.61087 .63112 L
.60301 .64083 L
.60198 .6422 L
.59606 .65054 L
.59227 .65636 L
.58986 .66026 L
Mistroke
.5843 .66997 L
.58256 .67323 L
.57927 .67968 L
.5747 .68939 L
.57284 .6936 L
.57052 .69911 L
.56668 .70882 L
.56313 .71849 L
.56312 .71853 L
.55982 .72824 L
.55674 .73796 L
.55385 .74767 L
.55342 .74915 L
.55114 .75738 L
.54857 .76709 L
.54613 .77681 L
.54378 .78652 L
.54371 .78685 L
.54158 .79623 L
.53945 .80594 L
.53737 .81566 L
.53534 .82537 L
.53399 .83193 L
.53339 .83508 L
.53153 .84479 L
.52966 .85451 L
.52776 .86422 L
.52583 .87393 L
.52428 .8816 L
.52396 .88364 L
.52235 .89336 L
.52049 .90307 L
.51834 .91278 L
.51585 .92249 L
.51457 .92702 L
.51322 .93221 L
.51036 .94192 L
.50659 .95163 L
.50486 .95476 L
.49514 .95476 L
.49341 .95163 L
.48964 .94192 L
.48678 .93221 L
.48543 .92702 L
.48415 .92249 L
.48166 .91278 L
.47951 .90307 L
.47765 .89336 L
.47604 .88364 L
.47572 .8816 L
Mistroke
.47417 .87393 L
.47224 .86422 L
.47034 .85451 L
.46847 .84479 L
.46661 .83508 L
.46601 .83193 L
.46466 .82537 L
.46263 .81566 L
.46055 .80594 L
.45842 .79623 L
.45629 .78685 L
.45622 .78652 L
.45387 .77681 L
.45143 .76709 L
.44886 .75738 L
.44658 .74915 L
.44615 .74767 L
.44326 .73796 L
.44018 .72824 L
.43688 .71853 L
.43687 .71849 L
.43332 .70882 L
.42948 .69911 L
.42716 .6936 L
.4253 .68939 L
.42073 .67968 L
.41744 .67323 L
.4157 .66997 L
.41014 .66026 L
.40773 .65636 L
.40394 .65054 L
.39802 .6422 L
.39699 .64083 L
.38913 .63112 L
.38831 .63017 L
.38016 .62141 L
.37859 .61984 L
.36983 .61169 L
.36888 .61087 L
.35917 .60301 L
.3578 .60198 L
.34946 .59606 L
.34364 .59227 L
.33974 .58986 L
.33003 .5843 L
.32677 .58256 L
.32032 .57927 L
.31061 .5747 L
.3064 .57284 L
.30089 .57052 L
Mistroke
.29118 .56668 L
.28151 .56313 L
.28147 .56312 L
.27176 .55982 L
.26204 .55674 L
.25233 .55385 L
.25085 .55342 L
.24262 .55114 L
.23291 .54857 L
.22319 .54613 L
.21348 .54378 L
.21315 .54371 L
.20377 .54158 L
.19406 .53945 L
.18434 .53737 L
.17463 .53534 L
.16807 .53399 L
.16492 .53339 L
.15521 .53153 L
.14549 .52966 L
.13578 .52776 L
.12607 .52583 L
.1184 .52428 L
.11636 .52396 L
.10664 .52235 L
.09693 .52049 L
.08722 .51834 L
.07751 .51585 L
.07298 .51457 L
.06779 .51322 L
.05808 .51036 L
.04837 .50659 L
.04524 .50486 L
.04524 .49514 L
.04837 .49341 L
.05808 .48964 L
.06779 .48678 L
.07298 .48543 L
.07751 .48415 L
.08722 .48166 L
.09693 .47951 L
.10664 .47765 L
.11636 .47604 L
.1184 .47572 L
.12607 .47417 L
.13578 .47224 L
.14549 .47034 L
.15521 .46847 L
.16492 .46661 L
.16807 .46601 L
Mistroke
.17463 .46466 L
.18434 .46263 L
.19406 .46055 L
.20377 .45842 L
.21315 .45629 L
.21348 .45622 L
.22319 .45387 L
.23291 .45143 L
.24262 .44886 L
.25085 .44658 L
.25233 .44615 L
.26204 .44326 L
.27176 .44018 L
.28147 .43688 L
.28151 .43687 L
.29118 .43332 L
.30089 .42948 L
.3064 .42716 L
.31061 .4253 L
.32032 .42073 L
.32677 .41744 L
.33003 .4157 L
.33974 .41014 L
.34364 .40773 L
.34946 .40394 L
.3578 .39802 L
.35917 .39699 L
.36888 .38913 L
.36983 .38831 L
.37859 .38016 L
.38016 .37859 L
.38831 .36983 L
.38913 .36888 L
.39699 .35917 L
.39802 .3578 L
.40394 .34946 L
.40773 .34364 L
.41014 .33974 L
.4157 .33003 L
.41744 .32677 L
.42073 .32032 L
.4253 .31061 L
.42716 .3064 L
.42948 .30089 L
.43332 .29118 L
.43687 .28151 L
.43688 .28147 L
.44018 .27176 L
.44326 .26204 L
.44615 .25233 L
Mistroke
.44658 .25085 L
.44886 .24262 L
.45143 .23291 L
.45387 .22319 L
.45622 .21348 L
.45629 .21315 L
.45842 .20377 L
.46055 .19406 L
.46263 .18434 L
.46466 .17463 L
.46601 .16807 L
.46661 .16492 L
.46847 .15521 L
.47034 .14549 L
.47224 .13578 L
.47417 .12607 L
.47572 .1184 L
.47604 .11636 L
.47765 .10664 L
.47951 .09693 L
.48166 .08722 L
.48415 .07751 L
.48543 .07298 L
.48678 .06779 L
.48964 .05808 L
.49341 .04837 L
.49514 .04524 L
Mfstroke
.8 1 0 r
.19406 .05013 m
.20172 .05808 L
.19691 .06779 L
.19406 .07299 L
.19142 .07751 L
.1857 .08722 L
.18434 .08916 L
.17871 .09693 L
.17463 .10257 L
.17155 .10664 L
.16492 .11504 L
.16383 .11636 L
.15557 .12607 L
.15521 .12647 L
.14672 .13578 L
.14549 .13705 L
.13705 .14549 L
.13578 .14672 L
.12647 .15521 L
.12607 .15557 L
.11636 .16383 L
.11504 .16492 L
.10664 .17155 L
.10257 .17463 L
.09693 .17871 L
.08916 .18434 L
.08722 .1857 L
.07751 .19142 L
.07299 .19406 L
.06779 .19691 L
.05808 .20172 L
.05013 .19406 L
.05808 .18653 L
.06046 .18434 L
.06779 .17775 L
.06953 .17463 L
.07751 .16905 L
.08107 .16492 L
.08722 .16028 L
.09206 .15521 L
.09693 .14986 L
.10265 .14549 L
.10664 .14138 L
.11289 .13578 L
.11636 .13228 L
.12277 .12607 L
.12607 .12277 L
.13228 .11636 L
.13578 .11289 L
.14138 .10664 L
.14549 .10265 L
.14986 .09693 L
.15521 .09206 L
.16028 .08722 L
.16492 .08107 L
.16905 .07751 L
.17463 .06953 L
.17775 .06779 L
.18434 .06046 L
.18653 .05808 L
F
0 g
.19406 .05013 m
.20172 .05808 L
.19691 .06779 L
.19406 .07299 L
.19142 .07751 L
.1857 .08722 L
.18434 .08916 L
.17871 .09693 L
.17463 .10257 L
.17155 .10664 L
.16492 .11504 L
.16383 .11636 L
.15557 .12607 L
.15521 .12647 L
.14672 .13578 L
.14549 .13705 L
.13705 .14549 L
.13578 .14672 L
.12647 .15521 L
.12607 .15557 L
.11636 .16383 L
.11504 .16492 L
.10664 .17155 L
.10257 .17463 L
.09693 .17871 L
.08916 .18434 L
.08722 .1857 L
.07751 .19142 L
.07299 .19406 L
.06779 .19691 L
.05808 .20172 L
.05013 .19406 L
.05808 .18653 L
.06046 .18434 L
.06779 .17775 L
.06953 .17463 L
.07751 .16905 L
.08107 .16492 L
.08722 .16028 L
.09206 .15521 L
.09693 .14986 L
.10265 .14549 L
.10664 .14138 L
.11289 .13578 L
.11636 .13228 L
.12277 .12607 L
.12607 .12277 L
.13228 .11636 L
.13578 .11289 L
.14138 .10664 L
Mistroke
.14549 .10265 L
.14986 .09693 L
.15521 .09206 L
.16028 .08722 L
.16492 .08107 L
.16905 .07751 L
.17463 .06953 L
.17775 .06779 L
.18434 .06046 L
.18653 .05808 L
.19406 .05013 L
Mfstroke
1 .72 0 r
.06779 .59859 m
.07751 .59686 L
.08722 .5964 L
.09693 .59652 L
.10664 .59611 L
.11636 .59732 L
.12607 .59857 L
.13578 .59996 L
.14549 .6015 L
.14833 .60198 L
.15521 .60306 L
.16492 .60491 L
.17463 .60704 L
.18434 .6094 L
.19302 .61169 L
.19406 .61196 L
.20377 .61469 L
.21348 .61776 L
.22319 .62113 L
.22394 .62141 L
.23291 .62475 L
.24262 .62877 L
.2479 .63112 L
.25233 .63315 L
.26204 .63798 L
.26735 .64083 L
.27176 .64329 L
.28147 .64918 L
.28358 .65054 L
.29118 .65569 L
.2974 .66026 L
.30089 .66295 L
.30932 .66997 L
.31061 .6711 L
.31973 .67968 L
.32032 .68027 L
.3289 .68939 L
.33003 .69068 L
.33705 .69911 L
.33974 .7026 L
.34431 .70882 L
.34946 .71642 L
.35082 .71853 L
.35671 .72824 L
.35917 .73265 L
.36202 .73796 L
.36685 .74767 L
.36888 .7521 L
.37123 .75738 L
.37525 .76709 L
.37859 .77606 L
.37887 .77681 L
.38224 .78652 L
.38531 .79623 L
.38804 .80594 L
.38831 .80698 L
.3906 .81566 L
.39296 .82537 L
.39509 .83508 L
.39694 .84479 L
.39802 .85167 L
.3985 .85451 L
.40004 .86422 L
.40143 .87393 L
.40268 .88364 L
.40389 .89336 L
.40348 .90307 L
.4036 .91278 L
.40314 .92249 L
.40141 .93221 L
.39802 .93715 L
.39016 .93221 L
.38831 .92976 L
.3837 .92249 L
.37859 .9134 L
.37818 .91278 L
.37142 .90307 L
.36888 .89466 L
.36815 .89336 L
.3627 .88364 L
.35917 .87562 L
.35824 .87393 L
.35293 .86422 L
.34946 .85718 L
.34796 .85451 L
.34248 .84479 L
.33974 .83968 L
.33702 .83508 L
.33125 .82537 L
.33003 .82327 L
.32521 .81566 L
.32032 .80797 L
.31896 .80594 L
.31232 .79623 L
.31061 .79375 L
.30535 .78652 L
.30089 .78052 L
.29802 .77681 L
.29118 .76821 L
.29026 .76709 L
.28201 .75738 L
.28147 .75676 L
.27323 .74767 L
.27176 .74609 L
.26389 .73796 L
.26204 .73611 L
.25391 .72824 L
.25233 .72677 L
.24324 .71853 L
.24262 .71799 L
.23291 .70974 L
.23179 .70882 L
.22319 .70198 L
.21948 .69911 L
.21348 .69465 L
.20625 .68939 L
.20377 .68768 L
.19406 .68104 L
.19203 .67968 L
.18434 .67479 L
.17673 .66997 L
.17463 .66875 L
.16492 .66298 L
.16032 .66026 L
.15521 .65752 L
.14549 .65204 L
.14282 .65054 L
.13578 .64707 L
.12607 .64176 L
.12438 .64083 L
.11636 .6373 L
.10664 .63185 L
.10534 .63112 L
.09693 .62858 L
.08722 .62182 L
.0866 .62141 L
.07751 .6163 L
.07024 .61169 L
.06779 .60984 L
.06285 .60198 L
F
0 g
.06779 .59859 m
.07751 .59686 L
.08722 .5964 L
.09693 .59652 L
.10664 .59611 L
.11636 .59732 L
.12607 .59857 L
.13578 .59996 L
.14549 .6015 L
.14833 .60198 L
.15521 .60306 L
.16492 .60491 L
.17463 .60704 L
.18434 .6094 L
.19302 .61169 L
.19406 .61196 L
.20377 .61469 L
.21348 .61776 L
.22319 .62113 L
.22394 .62141 L
.23291 .62475 L
.24262 .62877 L
.2479 .63112 L
.25233 .63315 L
.26204 .63798 L
.26735 .64083 L
.27176 .64329 L
.28147 .64918 L
.28358 .65054 L
.29118 .65569 L
.2974 .66026 L
.30089 .66295 L
.30932 .66997 L
.31061 .6711 L
.31973 .67968 L
.32032 .68027 L
.3289 .68939 L
.33003 .69068 L
.33705 .69911 L
.33974 .7026 L
.34431 .70882 L
.34946 .71642 L
.35082 .71853 L
.35671 .72824 L
.35917 .73265 L
.36202 .73796 L
.36685 .74767 L
.36888 .7521 L
.37123 .75738 L
.37525 .76709 L
Mistroke
.37859 .77606 L
.37887 .77681 L
.38224 .78652 L
.38531 .79623 L
.38804 .80594 L
.38831 .80698 L
.3906 .81566 L
.39296 .82537 L
.39509 .83508 L
.39694 .84479 L
.39802 .85167 L
.3985 .85451 L
.40004 .86422 L
.40143 .87393 L
.40268 .88364 L
.40389 .89336 L
.40348 .90307 L
.4036 .91278 L
.40314 .92249 L
.40141 .93221 L
.39802 .93715 L
.39016 .93221 L
.38831 .92976 L
.3837 .92249 L
.37859 .9134 L
.37818 .91278 L
.37142 .90307 L
.36888 .89466 L
.36815 .89336 L
.3627 .88364 L
.35917 .87562 L
.35824 .87393 L
.35293 .86422 L
.34946 .85718 L
.34796 .85451 L
.34248 .84479 L
.33974 .83968 L
.33702 .83508 L
.33125 .82537 L
.33003 .82327 L
.32521 .81566 L
.32032 .80797 L
.31896 .80594 L
.31232 .79623 L
.31061 .79375 L
.30535 .78652 L
.30089 .78052 L
.29802 .77681 L
.29118 .76821 L
.29026 .76709 L
Mistroke
.28201 .75738 L
.28147 .75676 L
.27323 .74767 L
.27176 .74609 L
.26389 .73796 L
.26204 .73611 L
.25391 .72824 L
.25233 .72677 L
.24324 .71853 L
.24262 .71799 L
.23291 .70974 L
.23179 .70882 L
.22319 .70198 L
.21948 .69911 L
.21348 .69465 L
.20625 .68939 L
.20377 .68768 L
.19406 .68104 L
.19203 .67968 L
.18434 .67479 L
.17673 .66997 L
.17463 .66875 L
.16492 .66298 L
.16032 .66026 L
.15521 .65752 L
.14549 .65204 L
.14282 .65054 L
.13578 .64707 L
.12607 .64176 L
.12438 .64083 L
.11636 .6373 L
.10664 .63185 L
.10534 .63112 L
.09693 .62858 L
.08722 .62182 L
.0866 .62141 L
.07751 .6163 L
.07024 .61169 L
.06779 .60984 L
.06285 .60198 L
.06779 .59859 L
Mfstroke
1 .72 0 r
.39802 .06285 m
.40141 .06779 L
.40314 .07751 L
.4036 .08722 L
.40348 .09693 L
.40389 .10664 L
.40268 .11636 L
.40143 .12607 L
.40004 .13578 L
.3985 .14549 L
.39802 .14833 L
.39694 .15521 L
.39509 .16492 L
.39296 .17463 L
.3906 .18434 L
.38831 .19302 L
.38804 .19406 L
.38531 .20377 L
.38224 .21348 L
.37887 .22319 L
.37859 .22394 L
.37525 .23291 L
.37123 .24262 L
.36888 .2479 L
.36685 .25233 L
.36202 .26204 L
.35917 .26735 L
.35671 .27176 L
.35082 .28147 L
.34946 .28358 L
.34431 .29118 L
.33974 .2974 L
.33705 .30089 L
.33003 .30932 L
.3289 .31061 L
.32032 .31973 L
.31973 .32032 L
.31061 .3289 L
.30932 .33003 L
.30089 .33705 L
.2974 .33974 L
.29118 .34431 L
.28358 .34946 L
.28147 .35082 L
.27176 .35671 L
.26735 .35917 L
.26204 .36202 L
.25233 .36685 L
.2479 .36888 L
.24262 .37123 L
.23291 .37525 L
.22394 .37859 L
.22319 .37887 L
.21348 .38224 L
.20377 .38531 L
.19406 .38804 L
.19302 .38831 L
.18434 .3906 L
.17463 .39296 L
.16492 .39509 L
.15521 .39694 L
.14833 .39802 L
.14549 .3985 L
.13578 .40004 L
.12607 .40143 L
.11636 .40268 L
.10664 .40389 L
.09693 .40348 L
.08722 .4036 L
.07751 .40314 L
.06779 .40141 L
.06285 .39802 L
.06779 .39016 L
.07024 .38831 L
.07751 .3837 L
.0866 .37859 L
.08722 .37818 L
.09693 .37142 L
.10534 .36888 L
.10664 .36815 L
.11636 .3627 L
.12438 .35917 L
.12607 .35824 L
.13578 .35293 L
.14282 .34946 L
.14549 .34796 L
.15521 .34248 L
.16032 .33974 L
.16492 .33702 L
.17463 .33125 L
.17673 .33003 L
.18434 .32521 L
.19203 .32032 L
.19406 .31896 L
.20377 .31232 L
.20625 .31061 L
.21348 .30535 L
.21948 .30089 L
.22319 .29802 L
.23179 .29118 L
.23291 .29026 L
.24262 .28201 L
.24324 .28147 L
.25233 .27323 L
.25391 .27176 L
.26204 .26389 L
.26389 .26204 L
.27176 .25391 L
.27323 .25233 L
.28147 .24324 L
.28201 .24262 L
.29026 .23291 L
.29118 .23179 L
.29802 .22319 L
.30089 .21948 L
.30535 .21348 L
.31061 .20625 L
.31232 .20377 L
.31896 .19406 L
.32032 .19203 L
.32521 .18434 L
.33003 .17673 L
.33125 .17463 L
.33702 .16492 L
.33974 .16032 L
.34248 .15521 L
.34796 .14549 L
.34946 .14282 L
.35293 .13578 L
.35824 .12607 L
.35917 .12438 L
.3627 .11636 L
.36815 .10664 L
.36888 .10534 L
.37142 .09693 L
.37818 .08722 L
.37859 .0866 L
.3837 .07751 L
.38831 .07024 L
.39016 .06779 L
F
0 g
.39802 .06285 m
.40141 .06779 L
.40314 .07751 L
.4036 .08722 L
.40348 .09693 L
.40389 .10664 L
.40268 .11636 L
.40143 .12607 L
.40004 .13578 L
.3985 .14549 L
.39802 .14833 L
.39694 .15521 L
.39509 .16492 L
.39296 .17463 L
.3906 .18434 L
.38831 .19302 L
.38804 .19406 L
.38531 .20377 L
.38224 .21348 L
.37887 .22319 L
.37859 .22394 L
.37525 .23291 L
.37123 .24262 L
.36888 .2479 L
.36685 .25233 L
.36202 .26204 L
.35917 .26735 L
.35671 .27176 L
.35082 .28147 L
.34946 .28358 L
.34431 .29118 L
.33974 .2974 L
.33705 .30089 L
.33003 .30932 L
.3289 .31061 L
.32032 .31973 L
.31973 .32032 L
.31061 .3289 L
.30932 .33003 L
.30089 .33705 L
.2974 .33974 L
.29118 .34431 L
.28358 .34946 L
.28147 .35082 L
.27176 .35671 L
.26735 .35917 L
.26204 .36202 L
.25233 .36685 L
.2479 .36888 L
.24262 .37123 L
Mistroke
.23291 .37525 L
.22394 .37859 L
.22319 .37887 L
.21348 .38224 L
.20377 .38531 L
.19406 .38804 L
.19302 .38831 L
.18434 .3906 L
.17463 .39296 L
.16492 .39509 L
.15521 .39694 L
.14833 .39802 L
.14549 .3985 L
.13578 .40004 L
.12607 .40143 L
.11636 .40268 L
.10664 .40389 L
.09693 .40348 L
.08722 .4036 L
.07751 .40314 L
.06779 .40141 L
.06285 .39802 L
.06779 .39016 L
.07024 .38831 L
.07751 .3837 L
.0866 .37859 L
.08722 .37818 L
.09693 .37142 L
.10534 .36888 L
.10664 .36815 L
.11636 .3627 L
.12438 .35917 L
.12607 .35824 L
.13578 .35293 L
.14282 .34946 L
.14549 .34796 L
.15521 .34248 L
.16032 .33974 L
.16492 .33702 L
.17463 .33125 L
.17673 .33003 L
.18434 .32521 L
.19203 .32032 L
.19406 .31896 L
.20377 .31232 L
.20625 .31061 L
.21348 .30535 L
.21948 .30089 L
.22319 .29802 L
.23179 .29118 L
Mistroke
.23291 .29026 L
.24262 .28201 L
.24324 .28147 L
.25233 .27323 L
.25391 .27176 L
.26204 .26389 L
.26389 .26204 L
.27176 .25391 L
.27323 .25233 L
.28147 .24324 L
.28201 .24262 L
.29026 .23291 L
.29118 .23179 L
.29802 .22319 L
.30089 .21948 L
.30535 .21348 L
.31061 .20625 L
.31232 .20377 L
.31896 .19406 L
.32032 .19203 L
.32521 .18434 L
.33003 .17673 L
.33125 .17463 L
.33702 .16492 L
.33974 .16032 L
.34248 .15521 L
.34796 .14549 L
.34946 .14282 L
.35293 .13578 L
.35824 .12607 L
.35917 .12438 L
.3627 .11636 L
.36815 .10664 L
.36888 .10534 L
.37142 .09693 L
.37818 .08722 L
.37859 .0866 L
.3837 .07751 L
.38831 .07024 L
.39016 .06779 L
.39802 .06285 L
Mfstroke
.56 0 1 r
.49514 .08971 m
.50486 .08971 L
.50867 .09693 L
.51219 .10664 L
.51457 .11463 L
.51513 .11636 L
.518 .12607 L
.52045 .13578 L
.52253 .14549 L
.52428 .15507 L
.52431 .15521 L
.52652 .16492 L
.52868 .17463 L
.53081 .18434 L
.5329 .19406 L
.53399 .19911 L
.53507 .20377 L
.53736 .21348 L
.5397 .22319 L
.5421 .23291 L
.54371 .2392 L
.54462 .24262 L
.54729 .25233 L
.55009 .26204 L
.55305 .27176 L
.55342 .27292 L
.55623 .28147 L
.55962 .29118 L
.56313 .30054 L
.56327 .30089 L
.56724 .31061 L
.57155 .32032 L
.57284 .32308 L
.57627 .33003 L
.58148 .33974 L
.58256 .34163 L
.58727 .34946 L
.59227 .35705 L
.59375 .35917 L
.60105 .36888 L
.60198 .37004 L
.60935 .37859 L
.61169 .38111 L
.61889 .38831 L
.62141 .39065 L
.62996 .39802 L
.63112 .39895 L
.64083 .40625 L
.64295 .40773 L
.65054 .41273 L
.65837 .41744 L
.66026 .41852 L
.66997 .42373 L
.67692 .42716 L
.67968 .42845 L
.68939 .43276 L
.69911 .43673 L
.69946 .43687 L
.70882 .44038 L
.71853 .44377 L
.72708 .44658 L
.72824 .44695 L
.73796 .44991 L
.74767 .45271 L
.75738 .45538 L
.7608 .45629 L
.76709 .4579 L
.77681 .4603 L
.78652 .46264 L
.79623 .46493 L
.80089 .46601 L
.80594 .4671 L
.81566 .46919 L
.82537 .47132 L
.83508 .47348 L
.84479 .47569 L
.84493 .47572 L
.85451 .47747 L
.86422 .47955 L
.87393 .482 L
.88364 .48487 L
.88537 .48543 L
.89336 .48781 L
.90307 .49133 L
.91029 .49514 L
.91029 .50486 L
.90307 .50867 L
.89336 .51219 L
.88537 .51457 L
.88364 .51513 L
.87393 .518 L
.86422 .52045 L
.85451 .52253 L
.84493 .52428 L
.84479 .52431 L
.83508 .52652 L
.82537 .52868 L
.81566 .53081 L
.80594 .5329 L
.80089 .53399 L
.79623 .53507 L
.78652 .53736 L
.77681 .5397 L
.76709 .5421 L
.7608 .54371 L
.75738 .54462 L
.74767 .54729 L
.73796 .55009 L
.72824 .55305 L
.72708 .55342 L
.71853 .55623 L
.70882 .55962 L
.69946 .56313 L
.69911 .56327 L
.68939 .56724 L
.67968 .57155 L
.67692 .57284 L
.66997 .57627 L
.66026 .58148 L
.65837 .58256 L
.65054 .58727 L
.64295 .59227 L
.64083 .59375 L
.63112 .60105 L
.62996 .60198 L
.62141 .60935 L
.61889 .61169 L
.61169 .61889 L
.60935 .62141 L
.60198 .62996 L
.60105 .63112 L
.59375 .64083 L
.59227 .64295 L
.58727 .65054 L
.58256 .65837 L
.58148 .66026 L
.57627 .66997 L
.57284 .67692 L
.57155 .67968 L
.56724 .68939 L
.56327 .69911 L
.56313 .69946 L
.55962 .70882 L
.55623 .71853 L
.55342 .72708 L
.55305 .72824 L
.55009 .73796 L
.54729 .74767 L
.54462 .75738 L
.54371 .7608 L
.5421 .76709 L
.5397 .77681 L
.53736 .78652 L
.53507 .79623 L
.53399 .80089 L
.5329 .80594 L
.53081 .81566 L
.52868 .82537 L
.52652 .83508 L
.52431 .84479 L
.52428 .84493 L
.52253 .85451 L
.52045 .86422 L
.518 .87393 L
.51513 .88364 L
.51457 .88537 L
.51219 .89336 L
.50867 .90307 L
.50486 .91029 L
.49514 .91029 L
.49133 .90307 L
.48781 .89336 L
.48543 .88537 L
.48487 .88364 L
.482 .87393 L
.47955 .86422 L
.47747 .85451 L
.47572 .84493 L
.47569 .84479 L
.47348 .83508 L
.47132 .82537 L
.46919 .81566 L
.4671 .80594 L
.46601 .80089 L
.46493 .79623 L
.46264 .78652 L
.4603 .77681 L
.4579 .76709 L
.45629 .7608 L
.45538 .75738 L
.45271 .74767 L
.44991 .73796 L
.44695 .72824 L
.44658 .72708 L
.44377 .71853 L
.44038 .70882 L
.43687 .69946 L
.43673 .69911 L
.43276 .68939 L
.42845 .67968 L
.42716 .67692 L
.42373 .66997 L
.41852 .66026 L
.41744 .65837 L
.41273 .65054 L
.40773 .64295 L
.40625 .64083 L
.39895 .63112 L
.39802 .62996 L
.39065 .62141 L
.38831 .61889 L
.38111 .61169 L
.37859 .60935 L
.37004 .60198 L
.36888 .60105 L
.35917 .59375 L
.35705 .59227 L
.34946 .58727 L
.34163 .58256 L
.33974 .58148 L
.33003 .57627 L
.32308 .57284 L
.32032 .57155 L
.31061 .56724 L
.30089 .56327 L
.30054 .56313 L
.29118 .55962 L
.28147 .55623 L
.27292 .55342 L
.27176 .55305 L
.26204 .55009 L
.25233 .54729 L
.24262 .54462 L
.2392 .54371 L
.23291 .5421 L
.22319 .5397 L
.21348 .53736 L
.20377 .53507 L
.19911 .53399 L
.19406 .5329 L
.18434 .53081 L
.17463 .52868 L
.16492 .52652 L
.15521 .52431 L
.15507 .52428 L
.14549 .52253 L
.13578 .52045 L
.12607 .518 L
.11636 .51513 L
.11463 .51457 L
.10664 .51219 L
.09693 .50867 L
.08971 .50486 L
.08971 .49514 L
.09693 .49133 L
.10664 .48781 L
.11463 .48543 L
.11636 .48487 L
.12607 .482 L
.13578 .47955 L
.14549 .47747 L
.15507 .47572 L
.15521 .47569 L
.16492 .47348 L
.17463 .47132 L
.18434 .46919 L
.19406 .4671 L
.19911 .46601 L
.20377 .46493 L
.21348 .46264 L
.22319 .4603 L
.23291 .4579 L
.2392 .45629 L
.24262 .45538 L
.25233 .45271 L
.26204 .44991 L
.27176 .44695 L
.27292 .44658 L
.28147 .44377 L
.29118 .44038 L
.30054 .43687 L
.30089 .43673 L
.31061 .43276 L
.32032 .42845 L
.32308 .42716 L
.33003 .42373 L
.33974 .41852 L
.34163 .41744 L
.34946 .41273 L
.35705 .40773 L
.35917 .40625 L
.36888 .39895 L
.37004 .39802 L
.37859 .39065 L
.38111 .38831 L
.38831 .38111 L
.39065 .37859 L
.39802 .37004 L
.39895 .36888 L
.40625 .35917 L
.40773 .35705 L
.41273 .34946 L
.41744 .34163 L
.41852 .33974 L
.42373 .33003 L
.42716 .32308 L
.42845 .32032 L
.43276 .31061 L
.43673 .30089 L
.43687 .30054 L
.44038 .29118 L
.44377 .28147 L
.44658 .27292 L
.44695 .27176 L
.44991 .26204 L
.45271 .25233 L
.45538 .24262 L
.45629 .2392 L
.4579 .23291 L
.4603 .22319 L
.46264 .21348 L
.46493 .20377 L
.46601 .19911 L
.4671 .19406 L
.46919 .18434 L
.47132 .17463 L
.47348 .16492 L
.47569 .15521 L
.47572 .15507 L
.47747 .14549 L
.47955 .13578 L
.482 .12607 L
.48487 .11636 L
.48543 .11463 L
.48781 .10664 L
.49133 .09693 L
F
0 g
.49514 .08971 m
.50486 .08971 L
.50867 .09693 L
.51219 .10664 L
.51457 .11463 L
.51513 .11636 L
.518 .12607 L
.52045 .13578 L
.52253 .14549 L
.52428 .15507 L
.52431 .15521 L
.52652 .16492 L
.52868 .17463 L
.53081 .18434 L
.5329 .19406 L
.53399 .19911 L
.53507 .20377 L
.53736 .21348 L
.5397 .22319 L
.5421 .23291 L
.54371 .2392 L
.54462 .24262 L
.54729 .25233 L
.55009 .26204 L
.55305 .27176 L
.55342 .27292 L
.55623 .28147 L
.55962 .29118 L
.56313 .30054 L
.56327 .30089 L
.56724 .31061 L
.57155 .32032 L
.57284 .32308 L
.57627 .33003 L
.58148 .33974 L
.58256 .34163 L
.58727 .34946 L
.59227 .35705 L
.59375 .35917 L
.60105 .36888 L
.60198 .37004 L
.60935 .37859 L
.61169 .38111 L
.61889 .38831 L
.62141 .39065 L
.62996 .39802 L
.63112 .39895 L
.64083 .40625 L
.64295 .40773 L
.65054 .41273 L
Mistroke
.65837 .41744 L
.66026 .41852 L
.66997 .42373 L
.67692 .42716 L
.67968 .42845 L
.68939 .43276 L
.69911 .43673 L
.69946 .43687 L
.70882 .44038 L
.71853 .44377 L
.72708 .44658 L
.72824 .44695 L
.73796 .44991 L
.74767 .45271 L
.75738 .45538 L
.7608 .45629 L
.76709 .4579 L
.77681 .4603 L
.78652 .46264 L
.79623 .46493 L
.80089 .46601 L
.80594 .4671 L
.81566 .46919 L
.82537 .47132 L
.83508 .47348 L
.84479 .47569 L
.84493 .47572 L
.85451 .47747 L
.86422 .47955 L
.87393 .482 L
.88364 .48487 L
.88537 .48543 L
.89336 .48781 L
.90307 .49133 L
.91029 .49514 L
.91029 .50486 L
.90307 .50867 L
.89336 .51219 L
.88537 .51457 L
.88364 .51513 L
.87393 .518 L
.86422 .52045 L
.85451 .52253 L
.84493 .52428 L
.84479 .52431 L
.83508 .52652 L
.82537 .52868 L
.81566 .53081 L
.80594 .5329 L
.80089 .53399 L
Mistroke
.79623 .53507 L
.78652 .53736 L
.77681 .5397 L
.76709 .5421 L
.7608 .54371 L
.75738 .54462 L
.74767 .54729 L
.73796 .55009 L
.72824 .55305 L
.72708 .55342 L
.71853 .55623 L
.70882 .55962 L
.69946 .56313 L
.69911 .56327 L
.68939 .56724 L
.67968 .57155 L
.67692 .57284 L
.66997 .57627 L
.66026 .58148 L
.65837 .58256 L
.65054 .58727 L
.64295 .59227 L
.64083 .59375 L
.63112 .60105 L
.62996 .60198 L
.62141 .60935 L
.61889 .61169 L
.61169 .61889 L
.60935 .62141 L
.60198 .62996 L
.60105 .63112 L
.59375 .64083 L
.59227 .64295 L
.58727 .65054 L
.58256 .65837 L
.58148 .66026 L
.57627 .66997 L
.57284 .67692 L
.57155 .67968 L
.56724 .68939 L
.56327 .69911 L
.56313 .69946 L
.55962 .70882 L
.55623 .71853 L
.55342 .72708 L
.55305 .72824 L
.55009 .73796 L
.54729 .74767 L
.54462 .75738 L
.54371 .7608 L
Mistroke
.5421 .76709 L
.5397 .77681 L
.53736 .78652 L
.53507 .79623 L
.53399 .80089 L
.5329 .80594 L
.53081 .81566 L
.52868 .82537 L
.52652 .83508 L
.52431 .84479 L
.52428 .84493 L
.52253 .85451 L
.52045 .86422 L
.518 .87393 L
.51513 .88364 L
.51457 .88537 L
.51219 .89336 L
.50867 .90307 L
.50486 .91029 L
.49514 .91029 L
.49133 .90307 L
.48781 .89336 L
.48543 .88537 L
.48487 .88364 L
.482 .87393 L
.47955 .86422 L
.47747 .85451 L
.47572 .84493 L
.47569 .84479 L
.47348 .83508 L
.47132 .82537 L
.46919 .81566 L
.4671 .80594 L
.46601 .80089 L
.46493 .79623 L
.46264 .78652 L
.4603 .77681 L
.4579 .76709 L
.45629 .7608 L
.45538 .75738 L
.45271 .74767 L
.44991 .73796 L
.44695 .72824 L
.44658 .72708 L
.44377 .71853 L
.44038 .70882 L
.43687 .69946 L
.43673 .69911 L
.43276 .68939 L
.42845 .67968 L
Mistroke
.42716 .67692 L
.42373 .66997 L
.41852 .66026 L
.41744 .65837 L
.41273 .65054 L
.40773 .64295 L
.40625 .64083 L
.39895 .63112 L
.39802 .62996 L
.39065 .62141 L
.38831 .61889 L
.38111 .61169 L
.37859 .60935 L
.37004 .60198 L
.36888 .60105 L
.35917 .59375 L
.35705 .59227 L
.34946 .58727 L
.34163 .58256 L
.33974 .58148 L
.33003 .57627 L
.32308 .57284 L
.32032 .57155 L
.31061 .56724 L
.30089 .56327 L
.30054 .56313 L
.29118 .55962 L
.28147 .55623 L
.27292 .55342 L
.27176 .55305 L
.26204 .55009 L
.25233 .54729 L
.24262 .54462 L
.2392 .54371 L
.23291 .5421 L
.22319 .5397 L
.21348 .53736 L
.20377 .53507 L
.19911 .53399 L
.19406 .5329 L
.18434 .53081 L
.17463 .52868 L
.16492 .52652 L
.15521 .52431 L
.15507 .52428 L
.14549 .52253 L
.13578 .52045 L
.12607 .518 L
.11636 .51513 L
.11463 .51457 L
Mistroke
.10664 .51219 L
.09693 .50867 L
.08971 .50486 L
.08971 .49514 L
.09693 .49133 L
.10664 .48781 L
.11463 .48543 L
.11636 .48487 L
.12607 .482 L
.13578 .47955 L
.14549 .47747 L
.15507 .47572 L
.15521 .47569 L
.16492 .47348 L
.17463 .47132 L
.18434 .46919 L
.19406 .4671 L
.19911 .46601 L
.20377 .46493 L
.21348 .46264 L
.22319 .4603 L
.23291 .4579 L
.2392 .45629 L
.24262 .45538 L
.25233 .45271 L
.26204 .44991 L
.27176 .44695 L
.27292 .44658 L
.28147 .44377 L
.29118 .44038 L
.30054 .43687 L
.30089 .43673 L
.31061 .43276 L
.32032 .42845 L
.32308 .42716 L
.33003 .42373 L
.33974 .41852 L
.34163 .41744 L
.34946 .41273 L
.35705 .40773 L
.35917 .40625 L
.36888 .39895 L
.37004 .39802 L
.37859 .39065 L
.38111 .38831 L
.38831 .38111 L
.39065 .37859 L
.39802 .37004 L
.39895 .36888 L
.40625 .35917 L
Mistroke
.40773 .35705 L
.41273 .34946 L
.41744 .34163 L
.41852 .33974 L
.42373 .33003 L
.42716 .32308 L
.42845 .32032 L
.43276 .31061 L
.43673 .30089 L
.43687 .30054 L
.44038 .29118 L
.44377 .28147 L
.44658 .27292 L
.44695 .27176 L
.44991 .26204 L
.45271 .25233 L
.45538 .24262 L
.45629 .2392 L
.4579 .23291 L
.4603 .22319 L
.46264 .21348 L
.46493 .20377 L
.46601 .19911 L
.4671 .19406 L
.46919 .18434 L
.47132 .17463 L
.47348 .16492 L
.47569 .15521 L
.47572 .15507 L
.47747 .14549 L
.47955 .13578 L
.482 .12607 L
.48487 .11636 L
.48543 .11463 L
.48781 .10664 L
.49133 .09693 L
.49514 .08971 L
Mfstroke
.32 0 1 r
.27176 .10235 m
.27635 .10664 L
.2742 .11636 L
.27176 .12336 L
.27068 .12607 L
.26639 .13578 L
.26204 .14449 L
.26151 .14549 L
.25688 .15521 L
.25233 .16208 L
.25137 .16492 L
.24354 .17463 L
.24262 .1759 L
.23662 .18434 L
.23291 .18895 L
.22891 .19406 L
.22319 .2005 L
.22033 .20377 L
.21348 .21088 L
.21088 .21348 L
.20377 .22033 L
.2005 .22319 L
.19406 .22891 L
.18895 .23291 L
.18434 .23662 L
.1759 .24262 L
.17463 .24354 L
.16492 .25137 L
.16208 .25233 L
.15521 .25688 L
.14549 .26151 L
.14449 .26204 L
.13578 .26639 L
.12607 .27068 L
.12336 .27176 L
.11636 .2742 L
.10664 .27635 L
.10235 .27176 L
.10664 .26638 L
.10993 .26204 L
.11636 .25713 L
.12141 .25233 L
.12607 .24867 L
.13294 .24262 L
.13578 .23953 L
.14435 .23291 L
.14549 .23179 L
.15521 .22343 L
.1554 .22319 L
.16492 .21467 L
.16594 .21348 L
.17463 .20491 L
.17623 .20377 L
.18434 .19574 L
.18617 .19406 L
.19406 .18617 L
.19574 .18434 L
.20377 .17623 L
.20491 .17463 L
.21348 .16594 L
.21467 .16492 L
.22319 .1554 L
.22343 .15521 L
.23179 .14549 L
.23291 .14435 L
.23953 .13578 L
.24262 .13294 L
.24867 .12607 L
.25233 .12141 L
.25713 .11636 L
.26204 .10993 L
.26638 .10664 L
F
0 g
.27176 .10235 m
.27635 .10664 L
.2742 .11636 L
.27176 .12336 L
.27068 .12607 L
.26639 .13578 L
.26204 .14449 L
.26151 .14549 L
.25688 .15521 L
.25233 .16208 L
.25137 .16492 L
.24354 .17463 L
.24262 .1759 L
.23662 .18434 L
.23291 .18895 L
.22891 .19406 L
.22319 .2005 L
.22033 .20377 L
.21348 .21088 L
.21088 .21348 L
.20377 .22033 L
.2005 .22319 L
.19406 .22891 L
.18895 .23291 L
.18434 .23662 L
.1759 .24262 L
.17463 .24354 L
.16492 .25137 L
.16208 .25233 L
.15521 .25688 L
.14549 .26151 L
.14449 .26204 L
.13578 .26639 L
.12607 .27068 L
.12336 .27176 L
.11636 .2742 L
.10664 .27635 L
.10235 .27176 L
.10664 .26638 L
.10993 .26204 L
.11636 .25713 L
.12141 .25233 L
.12607 .24867 L
.13294 .24262 L
.13578 .23953 L
.14435 .23291 L
.14549 .23179 L
.15521 .22343 L
.1554 .22319 L
.16492 .21467 L
Mistroke
.16594 .21348 L
.17463 .20491 L
.17623 .20377 L
.18434 .19574 L
.18617 .19406 L
.19406 .18617 L
.19574 .18434 L
.20377 .17623 L
.20491 .17463 L
.21348 .16594 L
.21467 .16492 L
.22319 .1554 L
.22343 .15521 L
.23179 .14549 L
.23291 .14435 L
.23953 .13578 L
.24262 .13294 L
.24867 .12607 L
.25233 .12141 L
.25713 .11636 L
.26204 .10993 L
.26638 .10664 L
.27176 .10235 L
Mfstroke
.32 0 1 r
.10664 .72365 m
.11636 .7258 L
.12336 .72824 L
.12607 .72932 L
.13578 .73361 L
.14449 .73796 L
.14549 .73849 L
.15521 .74312 L
.16208 .74767 L
.16492 .74863 L
.17463 .75646 L
.1759 .75738 L
.18434 .76338 L
.18895 .76709 L
.19406 .77109 L
.2005 .77681 L
.20377 .77967 L
.21088 .78652 L
.21348 .78912 L
.22033 .79623 L
.22319 .7995 L
.22891 .80594 L
.23291 .81105 L
.23662 .81566 L
.24262 .8241 L
.24354 .82537 L
.25137 .83508 L
.25233 .83792 L
.25688 .84479 L
.26151 .85451 L
.26204 .85551 L
.26639 .86422 L
.27068 .87393 L
.27176 .87664 L
.2742 .88364 L
.27635 .89336 L
.27176 .89765 L
.26638 .89336 L
.26204 .89007 L
.25713 .88364 L
.25233 .87859 L
.24867 .87393 L
.24262 .86706 L
.23953 .86422 L
.23291 .85565 L
.23179 .85451 L
.22343 .84479 L
.22319 .8446 L
.21467 .83508 L
.21348 .83406 L
.20491 .82537 L
.20377 .82377 L
.19574 .81566 L
.19406 .81383 L
.18617 .80594 L
.18434 .80426 L
.17623 .79623 L
.17463 .79509 L
.16594 .78652 L
.16492 .78533 L
.1554 .77681 L
.15521 .77657 L
.14549 .76821 L
.14435 .76709 L
.13578 .76047 L
.13294 .75738 L
.12607 .75133 L
.12141 .74767 L
.11636 .74287 L
.10993 .73796 L
.10664 .73362 L
.10235 .72824 L
F
0 g
.10664 .72365 m
.11636 .7258 L
.12336 .72824 L
.12607 .72932 L
.13578 .73361 L
.14449 .73796 L
.14549 .73849 L
.15521 .74312 L
.16208 .74767 L
.16492 .74863 L
.17463 .75646 L
.1759 .75738 L
.18434 .76338 L
.18895 .76709 L
.19406 .77109 L
.2005 .77681 L
.20377 .77967 L
.21088 .78652 L
.21348 .78912 L
.22033 .79623 L
.22319 .7995 L
.22891 .80594 L
.23291 .81105 L
.23662 .81566 L
.24262 .8241 L
.24354 .82537 L
.25137 .83508 L
.25233 .83792 L
.25688 .84479 L
.26151 .85451 L
.26204 .85551 L
.26639 .86422 L
.27068 .87393 L
.27176 .87664 L
.2742 .88364 L
.27635 .89336 L
.27176 .89765 L
.26638 .89336 L
.26204 .89007 L
.25713 .88364 L
.25233 .87859 L
.24867 .87393 L
.24262 .86706 L
.23953 .86422 L
.23291 .85565 L
.23179 .85451 L
.22343 .84479 L
.22319 .8446 L
.21467 .83508 L
.21348 .83406 L
Mistroke
.20491 .82537 L
.20377 .82377 L
.19574 .81566 L
.19406 .81383 L
.18617 .80594 L
.18434 .80426 L
.17623 .79623 L
.17463 .79509 L
.16594 .78652 L
.16492 .78533 L
.1554 .77681 L
.15521 .77657 L
.14549 .76821 L
.14435 .76709 L
.13578 .76047 L
.13294 .75738 L
.12607 .75133 L
.12141 .74767 L
.11636 .74287 L
.10993 .73796 L
.10664 .73362 L
.10235 .72824 L
.10664 .72365 L
Mfstroke
1 .48 0 r
.37859 .11538 m
.38806 .11636 L
.38831 .11709 L
.38961 .12607 L
.39005 .13578 L
.38907 .14549 L
.38831 .15297 L
.38784 .15521 L
.38698 .16492 L
.38497 .17463 L
.38268 .18434 L
.38015 .19406 L
.37859 .19964 L
.3775 .20377 L
.37452 .21348 L
.37111 .22319 L
.36888 .22909 L
.36743 .23291 L
.36336 .24262 L
.35917 .25155 L
.3588 .25233 L
.35384 .26204 L
.34946 .26977 L
.34829 .27176 L
.34215 .28147 L
.33974 .28499 L
.33531 .29118 L
.33003 .29796 L
.32763 .30089 L
.32032 .30917 L
.31897 .31061 L
.31061 .31897 L
.30917 .32032 L
.30089 .32763 L
.29796 .33003 L
.29118 .33531 L
.28499 .33974 L
.28147 .34215 L
.27176 .34829 L
.26977 .34946 L
.26204 .35384 L
.25233 .3588 L
.25155 .35917 L
.24262 .36336 L
.23291 .36743 L
.22909 .36888 L
.22319 .37111 L
.21348 .37452 L
.20377 .3775 L
.19964 .37859 L
.19406 .38015 L
.18434 .38268 L
.17463 .38497 L
.16492 .38698 L
.15521 .38784 L
.15297 .38831 L
.14549 .38907 L
.13578 .39005 L
.12607 .38961 L
.11709 .38831 L
.11636 .38806 L
.11538 .37859 L
.11636 .37763 L
.12607 .37008 L
.12781 .36888 L
.13578 .36376 L
.14318 .35917 L
.14549 .35752 L
.15521 .35084 L
.15915 .34946 L
.16492 .34562 L
.17463 .33985 L
.17487 .33974 L
.18434 .33343 L
.1899 .33003 L
.19406 .32717 L
.20377 .32056 L
.20414 .32032 L
.21348 .3135 L
.21751 .31061 L
.22319 .30623 L
.23006 .30089 L
.23291 .29857 L
.24181 .29118 L
.24262 .29049 L
.25233 .2819 L
.25281 .28147 L
.26204 .27279 L
.26311 .27176 L
.27176 .26311 L
.27279 .26204 L
.28147 .25281 L
.2819 .25233 L
.29049 .24262 L
.29118 .24181 L
.29857 .23291 L
.30089 .23006 L
.30623 .22319 L
.31061 .21751 L
.3135 .21348 L
.32032 .20414 L
.32056 .20377 L
.32717 .19406 L
.33003 .1899 L
.33343 .18434 L
.33974 .17487 L
.33985 .17463 L
.34562 .16492 L
.34946 .15915 L
.35084 .15521 L
.35752 .14549 L
.35917 .14318 L
.36376 .13578 L
.36888 .12781 L
.37008 .12607 L
.37763 .11636 L
F
0 g
.37859 .11538 m
.38806 .11636 L
.38831 .11709 L
.38961 .12607 L
.39005 .13578 L
.38907 .14549 L
.38831 .15297 L
.38784 .15521 L
.38698 .16492 L
.38497 .17463 L
.38268 .18434 L
.38015 .19406 L
.37859 .19964 L
.3775 .20377 L
.37452 .21348 L
.37111 .22319 L
.36888 .22909 L
.36743 .23291 L
.36336 .24262 L
.35917 .25155 L
.3588 .25233 L
.35384 .26204 L
.34946 .26977 L
.34829 .27176 L
.34215 .28147 L
.33974 .28499 L
.33531 .29118 L
.33003 .29796 L
.32763 .30089 L
.32032 .30917 L
.31897 .31061 L
.31061 .31897 L
.30917 .32032 L
.30089 .32763 L
.29796 .33003 L
.29118 .33531 L
.28499 .33974 L
.28147 .34215 L
.27176 .34829 L
.26977 .34946 L
.26204 .35384 L
.25233 .3588 L
.25155 .35917 L
.24262 .36336 L
.23291 .36743 L
.22909 .36888 L
.22319 .37111 L
.21348 .37452 L
.20377 .3775 L
.19964 .37859 L
Mistroke
.19406 .38015 L
.18434 .38268 L
.17463 .38497 L
.16492 .38698 L
.15521 .38784 L
.15297 .38831 L
.14549 .38907 L
.13578 .39005 L
.12607 .38961 L
.11709 .38831 L
.11636 .38806 L
.11538 .37859 L
.11636 .37763 L
.12607 .37008 L
.12781 .36888 L
.13578 .36376 L
.14318 .35917 L
.14549 .35752 L
.15521 .35084 L
.15915 .34946 L
.16492 .34562 L
.17463 .33985 L
.17487 .33974 L
.18434 .33343 L
.1899 .33003 L
.19406 .32717 L
.20377 .32056 L
.20414 .32032 L
.21348 .3135 L
.21751 .31061 L
.22319 .30623 L
.23006 .30089 L
.23291 .29857 L
.24181 .29118 L
.24262 .29049 L
.25233 .2819 L
.25281 .28147 L
.26204 .27279 L
.26311 .27176 L
.27176 .26311 L
.27279 .26204 L
.28147 .25281 L
.2819 .25233 L
.29049 .24262 L
.29118 .24181 L
.29857 .23291 L
.30089 .23006 L
.30623 .22319 L
.31061 .21751 L
.3135 .21348 L
Mistroke
.32032 .20414 L
.32056 .20377 L
.32717 .19406 L
.33003 .1899 L
.33343 .18434 L
.33974 .17487 L
.33985 .17463 L
.34562 .16492 L
.34946 .15915 L
.35084 .15521 L
.35752 .14549 L
.35917 .14318 L
.36376 .13578 L
.36888 .12781 L
.37008 .12607 L
.37763 .11636 L
.37859 .11538 L
Mfstroke
1 .48 0 r
.12607 .61039 m
.13578 .60995 L
.14549 .61093 L
.15297 .61169 L
.15521 .61216 L
.16492 .61302 L
.17463 .61503 L
.18434 .61732 L
.19406 .61985 L
.19964 .62141 L
.20377 .6225 L
.21348 .62548 L
.22319 .62889 L
.22909 .63112 L
.23291 .63257 L
.24262 .63664 L
.25155 .64083 L
.25233 .6412 L
.26204 .64616 L
.26977 .65054 L
.27176 .65171 L
.28147 .65785 L
.28499 .66026 L
.29118 .66469 L
.29796 .66997 L
.30089 .67237 L
.30917 .67968 L
.31061 .68103 L
.31897 .68939 L
.32032 .69083 L
.32763 .69911 L
.33003 .70204 L
.33531 .70882 L
.33974 .71501 L
.34215 .71853 L
.34829 .72824 L
.34946 .73023 L
.35384 .73796 L
.3588 .74767 L
.35917 .74845 L
.36336 .75738 L
.36743 .76709 L
.36888 .77091 L
.37111 .77681 L
.37452 .78652 L
.3775 .79623 L
.37859 .80036 L
.38015 .80594 L
.38268 .81566 L
.38497 .82537 L
.38698 .83508 L
.38784 .84479 L
.38831 .84703 L
.38907 .85451 L
.39005 .86422 L
.38961 .87393 L
.38831 .88291 L
.38806 .88364 L
.37859 .88462 L
.37763 .88364 L
.37008 .87393 L
.36888 .87219 L
.36376 .86422 L
.35917 .85682 L
.35752 .85451 L
.35084 .84479 L
.34946 .84085 L
.34562 .83508 L
.33985 .82537 L
.33974 .82513 L
.33343 .81566 L
.33003 .8101 L
.32717 .80594 L
.32056 .79623 L
.32032 .79586 L
.3135 .78652 L
.31061 .78249 L
.30623 .77681 L
.30089 .76994 L
.29857 .76709 L
.29118 .75819 L
.29049 .75738 L
.2819 .74767 L
.28147 .74719 L
.27279 .73796 L
.27176 .73689 L
.26311 .72824 L
.26204 .72721 L
.25281 .71853 L
.25233 .7181 L
.24262 .70951 L
.24181 .70882 L
.23291 .70143 L
.23006 .69911 L
.22319 .69377 L
.21751 .68939 L
.21348 .6865 L
.20414 .67968 L
.20377 .67944 L
.19406 .67283 L
.1899 .66997 L
.18434 .66657 L
.17487 .66026 L
.17463 .66015 L
.16492 .65438 L
.15915 .65054 L
.15521 .64916 L
.14549 .64248 L
.14318 .64083 L
.13578 .63624 L
.12781 .63112 L
.12607 .62992 L
.11636 .62237 L
.11538 .62141 L
.11636 .61194 L
.11709 .61169 L
F
0 g
.12607 .61039 m
.13578 .60995 L
.14549 .61093 L
.15297 .61169 L
.15521 .61216 L
.16492 .61302 L
.17463 .61503 L
.18434 .61732 L
.19406 .61985 L
.19964 .62141 L
.20377 .6225 L
.21348 .62548 L
.22319 .62889 L
.22909 .63112 L
.23291 .63257 L
.24262 .63664 L
.25155 .64083 L
.25233 .6412 L
.26204 .64616 L
.26977 .65054 L
.27176 .65171 L
.28147 .65785 L
.28499 .66026 L
.29118 .66469 L
.29796 .66997 L
.30089 .67237 L
.30917 .67968 L
.31061 .68103 L
.31897 .68939 L
.32032 .69083 L
.32763 .69911 L
.33003 .70204 L
.33531 .70882 L
.33974 .71501 L
.34215 .71853 L
.34829 .72824 L
.34946 .73023 L
.35384 .73796 L
.3588 .74767 L
.35917 .74845 L
.36336 .75738 L
.36743 .76709 L
.36888 .77091 L
.37111 .77681 L
.37452 .78652 L
.3775 .79623 L
.37859 .80036 L
.38015 .80594 L
.38268 .81566 L
.38497 .82537 L
Mistroke
.38698 .83508 L
.38784 .84479 L
.38831 .84703 L
.38907 .85451 L
.39005 .86422 L
.38961 .87393 L
.38831 .88291 L
.38806 .88364 L
.37859 .88462 L
.37763 .88364 L
.37008 .87393 L
.36888 .87219 L
.36376 .86422 L
.35917 .85682 L
.35752 .85451 L
.35084 .84479 L
.34946 .84085 L
.34562 .83508 L
.33985 .82537 L
.33974 .82513 L
.33343 .81566 L
.33003 .8101 L
.32717 .80594 L
.32056 .79623 L
.32032 .79586 L
.3135 .78652 L
.31061 .78249 L
.30623 .77681 L
.30089 .76994 L
.29857 .76709 L
.29118 .75819 L
.29049 .75738 L
.2819 .74767 L
.28147 .74719 L
.27279 .73796 L
.27176 .73689 L
.26311 .72824 L
.26204 .72721 L
.25281 .71853 L
.25233 .7181 L
.24262 .70951 L
.24181 .70882 L
.23291 .70143 L
.23006 .69911 L
.22319 .69377 L
.21751 .68939 L
.21348 .6865 L
.20414 .67968 L
.20377 .67944 L
.19406 .67283 L
Mistroke
.1899 .66997 L
.18434 .66657 L
.17487 .66026 L
.17463 .66015 L
.16492 .65438 L
.15915 .65054 L
.15521 .64916 L
.14549 .64248 L
.14318 .64083 L
.13578 .63624 L
.12781 .63112 L
.12607 .62992 L
.11636 .62237 L
.11538 .62141 L
.11636 .61194 L
.11709 .61169 L
.12607 .61039 L
Mfstroke
0 .4 1 r
.12841 .01923 m
.12607 .02394 L
.12107 .01923 L
F
0 g
.12841 .01923 m
.12607 .02394 L
.12107 .01923 L
s
.8 0 1 r
.49514 .13332 m
.50486 .13332 L
.50656 .13578 L
.5111 .14549 L
.51447 .15521 L
.51457 .15552 L
.51778 .16492 L
.52056 .17463 L
.52287 .18434 L
.52428 .19128 L
.525 .19406 L
.52749 .20377 L
.52992 .21348 L
.53231 .22319 L
.53399 .23009 L
.53473 .23291 L
.53732 .24262 L
.53997 .25233 L
.54271 .26204 L
.54371 .2655 L
.5456 .27176 L
.54868 .28147 L
.55193 .29118 L
.55342 .29545 L
.55541 .30089 L
.55917 .31061 L
.56313 .32012 L
.56322 .32032 L
.56766 .33003 L
.57252 .33974 L
.57284 .34037 L
.5779 .34946 L
.58256 .3571 L
.58389 .35917 L
.59063 .36888 L
.59227 .37108 L
.59826 .37859 L
.60198 .3829 L
.607 .38831 L
.61169 .393 L
.6171 .39802 L
.62141 .40174 L
.62892 .40773 L
.63112 .40937 L
.64083 .41611 L
.6429 .41744 L
.65054 .4221 L
.65963 .42716 L
.66026 .42748 L
.66997 .43234 L
.67968 .43678 L
.67988 .43687 L
.68939 .44083 L
.69911 .44459 L
.70455 .44658 L
.70882 .44807 L
.71853 .45132 L
.72824 .4544 L
.7345 .45629 L
.73796 .45729 L
.74767 .46003 L
.75738 .46268 L
.76709 .46527 L
.76991 .46601 L
.77681 .46769 L
.78652 .47008 L
.79623 .47251 L
.80594 .475 L
.80872 .47572 L
.81566 .47713 L
.82537 .47944 L
.83508 .48222 L
.84448 .48543 L
.84479 .48553 L
.85451 .4889 L
.86422 .49344 L
.86668 .49514 L
.86668 .50486 L
.86422 .50656 L
.85451 .5111 L
.84479 .51447 L
.84448 .51457 L
.83508 .51778 L
.82537 .52056 L
.81566 .52287 L
.80872 .52428 L
.80594 .525 L
.79623 .52749 L
.78652 .52992 L
.77681 .53231 L
.76991 .53399 L
.76709 .53473 L
.75738 .53732 L
.74767 .53997 L
.73796 .54271 L
.7345 .54371 L
.72824 .5456 L
.71853 .54868 L
.70882 .55193 L
.70455 .55342 L
.69911 .55541 L
.68939 .55917 L
.67988 .56313 L
.67968 .56322 L
.66997 .56766 L
.66026 .57252 L
.65963 .57284 L
.65054 .5779 L
.6429 .58256 L
.64083 .58389 L
.63112 .59063 L
.62892 .59227 L
.62141 .59826 L
.6171 .60198 L
.61169 .607 L
.607 .61169 L
.60198 .6171 L
.59826 .62141 L
.59227 .62892 L
.59063 .63112 L
.58389 .64083 L
.58256 .6429 L
.5779 .65054 L
.57284 .65963 L
.57252 .66026 L
.56766 .66997 L
.56322 .67968 L
.56313 .67988 L
.55917 .68939 L
.55541 .69911 L
.55342 .70455 L
.55193 .70882 L
.54868 .71853 L
.5456 .72824 L
.54371 .7345 L
.54271 .73796 L
.53997 .74767 L
.53732 .75738 L
.53473 .76709 L
.53399 .76991 L
.53231 .77681 L
.52992 .78652 L
.52749 .79623 L
.525 .80594 L
.52428 .80872 L
.52287 .81566 L
.52056 .82537 L
.51778 .83508 L
.51457 .84448 L
.51447 .84479 L
.5111 .85451 L
.50656 .86422 L
.50486 .86668 L
.49514 .86668 L
.49344 .86422 L
.4889 .85451 L
.48553 .84479 L
.48543 .84448 L
.48222 .83508 L
.47944 .82537 L
.47713 .81566 L
.47572 .80872 L
.475 .80594 L
.47251 .79623 L
.47008 .78652 L
.46769 .77681 L
.46601 .76991 L
.46527 .76709 L
.46268 .75738 L
.46003 .74767 L
.45729 .73796 L
.45629 .7345 L
.4544 .72824 L
.45132 .71853 L
.44807 .70882 L
.44658 .70455 L
.44459 .69911 L
.44083 .68939 L
.43687 .67988 L
.43678 .67968 L
.43234 .66997 L
.42748 .66026 L
.42716 .65963 L
.4221 .65054 L
.41744 .6429 L
.41611 .64083 L
.40937 .63112 L
.40773 .62892 L
.40174 .62141 L
.39802 .6171 L
.393 .61169 L
.38831 .607 L
.3829 .60198 L
.37859 .59826 L
.37108 .59227 L
.36888 .59063 L
.35917 .58389 L
.3571 .58256 L
.34946 .5779 L
.34037 .57284 L
.33974 .57252 L
.33003 .56766 L
.32032 .56322 L
.32012 .56313 L
.31061 .55917 L
.30089 .55541 L
.29545 .55342 L
.29118 .55193 L
.28147 .54868 L
.27176 .5456 L
.2655 .54371 L
.26204 .54271 L
.25233 .53997 L
.24262 .53732 L
.23291 .53473 L
.23009 .53399 L
.22319 .53231 L
.21348 .52992 L
.20377 .52749 L
.19406 .525 L
.19128 .52428 L
.18434 .52287 L
.17463 .52056 L
.16492 .51778 L
.15552 .51457 L
.15521 .51447 L
.14549 .5111 L
.13578 .50656 L
.13332 .50486 L
.13332 .49514 L
.13578 .49344 L
.14549 .4889 L
.15521 .48553 L
.15552 .48543 L
.16492 .48222 L
.17463 .47944 L
.18434 .47713 L
.19128 .47572 L
.19406 .475 L
.20377 .47251 L
.21348 .47008 L
.22319 .46769 L
.23009 .46601 L
.23291 .46527 L
.24262 .46268 L
.25233 .46003 L
.26204 .45729 L
.2655 .45629 L
.27176 .4544 L
.28147 .45132 L
.29118 .44807 L
.29545 .44658 L
.30089 .44459 L
.31061 .44083 L
.32012 .43687 L
.32032 .43678 L
.33003 .43234 L
.33974 .42748 L
.34037 .42716 L
.34946 .4221 L
.3571 .41744 L
.35917 .41611 L
.36888 .40937 L
.37108 .40773 L
.37859 .40174 L
.3829 .39802 L
.38831 .393 L
.393 .38831 L
.39802 .3829 L
.40174 .37859 L
.40773 .37108 L
.40937 .36888 L
.41611 .35917 L
.41744 .3571 L
.4221 .34946 L
.42716 .34037 L
.42748 .33974 L
.43234 .33003 L
.43678 .32032 L
.43687 .32012 L
.44083 .31061 L
.44459 .30089 L
.44658 .29545 L
.44807 .29118 L
.45132 .28147 L
.4544 .27176 L
.45629 .2655 L
.45729 .26204 L
.46003 .25233 L
.46268 .24262 L
.46527 .23291 L
.46601 .23009 L
.46769 .22319 L
.47008 .21348 L
.47251 .20377 L
.475 .19406 L
.47572 .19128 L
.47713 .18434 L
.47944 .17463 L
.48222 .16492 L
.48543 .15552 L
.48553 .15521 L
.4889 .14549 L
.49344 .13578 L
F
0 g
.49514 .13332 m
.50486 .13332 L
.50656 .13578 L
.5111 .14549 L
.51447 .15521 L
.51457 .15552 L
.51778 .16492 L
.52056 .17463 L
.52287 .18434 L
.52428 .19128 L
.525 .19406 L
.52749 .20377 L
.52992 .21348 L
.53231 .22319 L
.53399 .23009 L
.53473 .23291 L
.53732 .24262 L
.53997 .25233 L
.54271 .26204 L
.54371 .2655 L
.5456 .27176 L
.54868 .28147 L
.55193 .29118 L
.55342 .29545 L
.55541 .30089 L
.55917 .31061 L
.56313 .32012 L
.56322 .32032 L
.56766 .33003 L
.57252 .33974 L
.57284 .34037 L
.5779 .34946 L
.58256 .3571 L
.58389 .35917 L
.59063 .36888 L
.59227 .37108 L
.59826 .37859 L
.60198 .3829 L
.607 .38831 L
.61169 .393 L
.6171 .39802 L
.62141 .40174 L
.62892 .40773 L
.63112 .40937 L
.64083 .41611 L
.6429 .41744 L
.65054 .4221 L
.65963 .42716 L
.66026 .42748 L
.66997 .43234 L
Mistroke
.67968 .43678 L
.67988 .43687 L
.68939 .44083 L
.69911 .44459 L
.70455 .44658 L
.70882 .44807 L
.71853 .45132 L
.72824 .4544 L
.7345 .45629 L
.73796 .45729 L
.74767 .46003 L
.75738 .46268 L
.76709 .46527 L
.76991 .46601 L
.77681 .46769 L
.78652 .47008 L
.79623 .47251 L
.80594 .475 L
.80872 .47572 L
.81566 .47713 L
.82537 .47944 L
.83508 .48222 L
.84448 .48543 L
.84479 .48553 L
.85451 .4889 L
.86422 .49344 L
.86668 .49514 L
.86668 .50486 L
.86422 .50656 L
.85451 .5111 L
.84479 .51447 L
.84448 .51457 L
.83508 .51778 L
.82537 .52056 L
.81566 .52287 L
.80872 .52428 L
.80594 .525 L
.79623 .52749 L
.78652 .52992 L
.77681 .53231 L
.76991 .53399 L
.76709 .53473 L
.75738 .53732 L
.74767 .53997 L
.73796 .54271 L
.7345 .54371 L
.72824 .5456 L
.71853 .54868 L
.70882 .55193 L
.70455 .55342 L
Mistroke
.69911 .55541 L
.68939 .55917 L
.67988 .56313 L
.67968 .56322 L
.66997 .56766 L
.66026 .57252 L
.65963 .57284 L
.65054 .5779 L
.6429 .58256 L
.64083 .58389 L
.63112 .59063 L
.62892 .59227 L
.62141 .59826 L
.6171 .60198 L
.61169 .607 L
.607 .61169 L
.60198 .6171 L
.59826 .62141 L
.59227 .62892 L
.59063 .63112 L
.58389 .64083 L
.58256 .6429 L
.5779 .65054 L
.57284 .65963 L
.57252 .66026 L
.56766 .66997 L
.56322 .67968 L
.56313 .67988 L
.55917 .68939 L
.55541 .69911 L
.55342 .70455 L
.55193 .70882 L
.54868 .71853 L
.5456 .72824 L
.54371 .7345 L
.54271 .73796 L
.53997 .74767 L
.53732 .75738 L
.53473 .76709 L
.53399 .76991 L
.53231 .77681 L
.52992 .78652 L
.52749 .79623 L
.525 .80594 L
.52428 .80872 L
.52287 .81566 L
.52056 .82537 L
.51778 .83508 L
.51457 .84448 L
.51447 .84479 L
Mistroke
.5111 .85451 L
.50656 .86422 L
.50486 .86668 L
.49514 .86668 L
.49344 .86422 L
.4889 .85451 L
.48553 .84479 L
.48543 .84448 L
.48222 .83508 L
.47944 .82537 L
.47713 .81566 L
.47572 .80872 L
.475 .80594 L
.47251 .79623 L
.47008 .78652 L
.46769 .77681 L
.46601 .76991 L
.46527 .76709 L
.46268 .75738 L
.46003 .74767 L
.45729 .73796 L
.45629 .7345 L
.4544 .72824 L
.45132 .71853 L
.44807 .70882 L
.44658 .70455 L
.44459 .69911 L
.44083 .68939 L
.43687 .67988 L
.43678 .67968 L
.43234 .66997 L
.42748 .66026 L
.42716 .65963 L
.4221 .65054 L
.41744 .6429 L
.41611 .64083 L
.40937 .63112 L
.40773 .62892 L
.40174 .62141 L
.39802 .6171 L
.393 .61169 L
.38831 .607 L
.3829 .60198 L
.37859 .59826 L
.37108 .59227 L
.36888 .59063 L
.35917 .58389 L
.3571 .58256 L
.34946 .5779 L
.34037 .57284 L
Mistroke
.33974 .57252 L
.33003 .56766 L
.32032 .56322 L
.32012 .56313 L
.31061 .55917 L
.30089 .55541 L
.29545 .55342 L
.29118 .55193 L
.28147 .54868 L
.27176 .5456 L
.2655 .54371 L
.26204 .54271 L
.25233 .53997 L
.24262 .53732 L
.23291 .53473 L
.23009 .53399 L
.22319 .53231 L
.21348 .52992 L
.20377 .52749 L
.19406 .525 L
.19128 .52428 L
.18434 .52287 L
.17463 .52056 L
.16492 .51778 L
.15552 .51457 L
.15521 .51447 L
.14549 .5111 L
.13578 .50656 L
.13332 .50486 L
.13332 .49514 L
.13578 .49344 L
.14549 .4889 L
.15521 .48553 L
.15552 .48543 L
.16492 .48222 L
.17463 .47944 L
.18434 .47713 L
.19128 .47572 L
.19406 .475 L
.20377 .47251 L
.21348 .47008 L
.22319 .46769 L
.23009 .46601 L
.23291 .46527 L
.24262 .46268 L
.25233 .46003 L
.26204 .45729 L
.2655 .45629 L
.27176 .4544 L
.28147 .45132 L
Mistroke
.29118 .44807 L
.29545 .44658 L
.30089 .44459 L
.31061 .44083 L
.32012 .43687 L
.32032 .43678 L
.33003 .43234 L
.33974 .42748 L
.34037 .42716 L
.34946 .4221 L
.3571 .41744 L
.35917 .41611 L
.36888 .40937 L
.37108 .40773 L
.37859 .40174 L
.3829 .39802 L
.38831 .393 L
.393 .38831 L
.39802 .3829 L
.40174 .37859 L
.40773 .37108 L
.40937 .36888 L
.41611 .35917 L
.41744 .3571 L
.4221 .34946 L
.42716 .34037 L
.42748 .33974 L
.43234 .33003 L
.43678 .32032 L
.43687 .32012 L
.44083 .31061 L
.44459 .30089 L
.44658 .29545 L
.44807 .29118 L
.45132 .28147 L
.4544 .27176 L
.45629 .2655 L
.45729 .26204 L
.46003 .25233 L
.46268 .24262 L
.46527 .23291 L
.46601 .23009 L
.46769 .22319 L
.47008 .21348 L
.47251 .20377 L
.475 .19406 L
.47572 .19128 L
.47713 .18434 L
.47944 .17463 L
.48222 .16492 L
Mistroke
.48543 .15552 L
.48553 .15521 L
.4889 .14549 L
.49344 .13578 L
.49514 .13332 L
Mfstroke
1 .24 0 r
.17463 .62933 m
.18434 .62921 L
.19345 .63112 L
.19406 .63123 L
.20377 .63197 L
.21348 .6354 L
.22319 .63891 L
.22841 .64083 L
.23291 .64232 L
.24262 .64647 L
.25108 .65054 L
.25233 .65112 L
.26204 .65619 L
.26893 .66026 L
.27176 .66194 L
.28147 .66845 L
.28359 .66997 L
.29118 .67566 L
.29607 .67968 L
.30089 .68386 L
.30679 .68939 L
.31061 .69321 L
.31614 .69911 L
.32032 .70393 L
.32434 .70882 L
.33003 .71641 L
.33155 .71853 L
.33806 .72824 L
.33974 .73107 L
.34381 .73796 L
.34888 .74767 L
.34946 .74892 L
.35353 .75738 L
.35768 .76709 L
.35917 .77159 L
.36109 .77681 L
.3646 .78652 L
.36803 .79623 L
.36877 .80594 L
.36888 .80655 L
.37079 .81566 L
.37067 .82537 L
.36888 .83325 L
.35917 .83088 L
.35408 .82537 L
.34946 .81984 L
.34624 .81566 L
.33974 .80708 L
.33875 .80594 L
.33167 .79623 L
.33003 .79409 L
.32378 .78652 L
.32032 .78133 L
.3166 .77681 L
.31061 .76903 L
.30898 .76709 L
.30094 .75738 L
.30089 .75732 L
.29237 .74767 L
.29118 .74628 L
.28345 .73796 L
.28147 .73582 L
.27408 .72824 L
.27176 .72592 L
.26418 .71853 L
.26204 .71655 L
.25372 .70882 L
.25233 .70763 L
.24268 .69911 L
.24262 .69906 L
.23291 .69102 L
.23097 .68939 L
.22319 .6834 L
.21867 .67968 L
.21348 .67622 L
.20591 .66997 L
.20377 .66833 L
.19406 .66125 L
.19292 .66026 L
.18434 .65376 L
.18016 .65054 L
.17463 .64592 L
.16912 .64083 L
.16675 .63112 L
F
0 g
.17463 .62933 m
.18434 .62921 L
.19345 .63112 L
.19406 .63123 L
.20377 .63197 L
.21348 .6354 L
.22319 .63891 L
.22841 .64083 L
.23291 .64232 L
.24262 .64647 L
.25108 .65054 L
.25233 .65112 L
.26204 .65619 L
.26893 .66026 L
.27176 .66194 L
.28147 .66845 L
.28359 .66997 L
.29118 .67566 L
.29607 .67968 L
.30089 .68386 L
.30679 .68939 L
.31061 .69321 L
.31614 .69911 L
.32032 .70393 L
.32434 .70882 L
.33003 .71641 L
.33155 .71853 L
.33806 .72824 L
.33974 .73107 L
.34381 .73796 L
.34888 .74767 L
.34946 .74892 L
.35353 .75738 L
.35768 .76709 L
.35917 .77159 L
.36109 .77681 L
.3646 .78652 L
.36803 .79623 L
.36877 .80594 L
.36888 .80655 L
.37079 .81566 L
.37067 .82537 L
.36888 .83325 L
.35917 .83088 L
.35408 .82537 L
.34946 .81984 L
.34624 .81566 L
.33974 .80708 L
.33875 .80594 L
.33167 .79623 L
Mistroke
.33003 .79409 L
.32378 .78652 L
.32032 .78133 L
.3166 .77681 L
.31061 .76903 L
.30898 .76709 L
.30094 .75738 L
.30089 .75732 L
.29237 .74767 L
.29118 .74628 L
.28345 .73796 L
.28147 .73582 L
.27408 .72824 L
.27176 .72592 L
.26418 .71853 L
.26204 .71655 L
.25372 .70882 L
.25233 .70763 L
.24268 .69911 L
.24262 .69906 L
.23291 .69102 L
.23097 .68939 L
.22319 .6834 L
.21867 .67968 L
.21348 .67622 L
.20591 .66997 L
.20377 .66833 L
.19406 .66125 L
.19292 .66026 L
.18434 .65376 L
.18016 .65054 L
.17463 .64592 L
.16912 .64083 L
.16675 .63112 L
.17463 .62933 L
Mfstroke
1 .24 0 r
.35917 .16912 m
.36888 .16675 L
.37067 .17463 L
.37079 .18434 L
.36888 .19345 L
.36877 .19406 L
.36803 .20377 L
.3646 .21348 L
.36109 .22319 L
.35917 .22841 L
.35768 .23291 L
.35353 .24262 L
.34946 .25108 L
.34888 .25233 L
.34381 .26204 L
.33974 .26893 L
.33806 .27176 L
.33155 .28147 L
.33003 .28359 L
.32434 .29118 L
.32032 .29607 L
.31614 .30089 L
.31061 .30679 L
.30679 .31061 L
.30089 .31614 L
.29607 .32032 L
.29118 .32434 L
.28359 .33003 L
.28147 .33155 L
.27176 .33806 L
.26893 .33974 L
.26204 .34381 L
.25233 .34888 L
.25108 .34946 L
.24262 .35353 L
.23291 .35768 L
.22841 .35917 L
.22319 .36109 L
.21348 .3646 L
.20377 .36803 L
.19406 .36877 L
.19345 .36888 L
.18434 .37079 L
.17463 .37067 L
.16675 .36888 L
.16912 .35917 L
.17463 .35408 L
.18016 .34946 L
.18434 .34624 L
.19292 .33974 L
.19406 .33875 L
.20377 .33167 L
.20591 .33003 L
.21348 .32378 L
.21867 .32032 L
.22319 .3166 L
.23097 .31061 L
.23291 .30898 L
.24262 .30094 L
.24268 .30089 L
.25233 .29237 L
.25372 .29118 L
.26204 .28345 L
.26418 .28147 L
.27176 .27408 L
.27408 .27176 L
.28147 .26418 L
.28345 .26204 L
.29118 .25372 L
.29237 .25233 L
.30089 .24268 L
.30094 .24262 L
.30898 .23291 L
.31061 .23097 L
.3166 .22319 L
.32032 .21867 L
.32378 .21348 L
.33003 .20591 L
.33167 .20377 L
.33875 .19406 L
.33974 .19292 L
.34624 .18434 L
.34946 .18016 L
.35408 .17463 L
F
0 g
.35917 .16912 m
.36888 .16675 L
.37067 .17463 L
.37079 .18434 L
.36888 .19345 L
.36877 .19406 L
.36803 .20377 L
.3646 .21348 L
.36109 .22319 L
.35917 .22841 L
.35768 .23291 L
.35353 .24262 L
.34946 .25108 L
.34888 .25233 L
.34381 .26204 L
.33974 .26893 L
.33806 .27176 L
.33155 .28147 L
.33003 .28359 L
.32434 .29118 L
.32032 .29607 L
.31614 .30089 L
.31061 .30679 L
.30679 .31061 L
.30089 .31614 L
.29607 .32032 L
.29118 .32434 L
.28359 .33003 L
.28147 .33155 L
.27176 .33806 L
.26893 .33974 L
.26204 .34381 L
.25233 .34888 L
.25108 .34946 L
.24262 .35353 L
.23291 .35768 L
.22841 .35917 L
.22319 .36109 L
.21348 .3646 L
.20377 .36803 L
.19406 .36877 L
.19345 .36888 L
.18434 .37079 L
.17463 .37067 L
.16675 .36888 L
.16912 .35917 L
.17463 .35408 L
.18016 .34946 L
.18434 .34624 L
.19292 .33974 L
Mistroke
.19406 .33875 L
.20377 .33167 L
.20591 .33003 L
.21348 .32378 L
.21867 .32032 L
.22319 .3166 L
.23097 .31061 L
.23291 .30898 L
.24262 .30094 L
.24268 .30089 L
.25233 .29237 L
.25372 .29118 L
.26204 .28345 L
.26418 .28147 L
.27176 .27408 L
.27408 .27176 L
.28147 .26418 L
.28345 .26204 L
.29118 .25372 L
.29237 .25233 L
.30089 .24268 L
.30094 .24262 L
.30898 .23291 L
.31061 .23097 L
.3166 .22319 L
.32032 .21867 L
.32378 .21348 L
.33003 .20591 L
.33167 .20377 L
.33875 .19406 L
.33974 .19292 L
.34624 .18434 L
.34946 .18016 L
.35408 .17463 L
.35917 .16912 L
Mfstroke
1 0 .96 r
.49514 .17722 m
.50486 .17722 L
.50942 .18434 L
.51358 .19406 L
.51457 .19673 L
.51735 .20377 L
.52058 .21348 L
.52318 .22319 L
.52428 .22796 L
.52575 .23291 L
.52859 .24262 L
.53135 .25233 L
.53399 .26176 L
.53408 .26204 L
.53703 .27176 L
.54007 .28147 L
.54321 .29118 L
.54371 .29267 L
.54658 .30089 L
.55016 .31061 L
.55342 .31893 L
.55399 .32032 L
.55815 .33003 L
.56267 .33974 L
.56313 .34068 L
.56766 .34946 L
.57284 .35863 L
.57316 .35917 L
.57933 .36888 L
.58256 .37354 L
.58629 .37859 L
.59227 .38605 L
.59421 .38831 L
.60198 .39667 L
.60333 .39802 L
.61169 .40579 L
.61395 .40773 L
.62141 .41371 L
.62646 .41744 L
.63112 .42067 L
.64083 .42684 L
.64137 .42716 L
.65054 .43234 L
.65932 .43687 L
.66026 .43733 L
.66997 .44185 L
.67968 .44601 L
.68107 .44658 L
.68939 .44984 L
.69911 .45342 L
.70733 .45629 L
.70882 .45679 L
.71853 .45993 L
.72824 .46297 L
.73796 .46592 L
.73824 .46601 L
.74767 .46865 L
.75738 .47141 L
.76709 .47425 L
.77204 .47572 L
.77681 .47682 L
.78652 .47942 L
.79623 .48265 L
.80327 .48543 L
.80594 .48642 L
.81566 .49058 L
.82278 .49514 L
.82278 .50486 L
.81566 .50942 L
.80594 .51358 L
.80327 .51457 L
.79623 .51735 L
.78652 .52058 L
.77681 .52318 L
.77204 .52428 L
.76709 .52575 L
.75738 .52859 L
.74767 .53135 L
.73824 .53399 L
.73796 .53408 L
.72824 .53703 L
.71853 .54007 L
.70882 .54321 L
.70733 .54371 L
.69911 .54658 L
.68939 .55016 L
.68107 .55342 L
.67968 .55399 L
.66997 .55815 L
.66026 .56267 L
.65932 .56313 L
.65054 .56766 L
.64137 .57284 L
.64083 .57316 L
.63112 .57933 L
.62646 .58256 L
.62141 .58629 L
.61395 .59227 L
.61169 .59421 L
.60333 .60198 L
.60198 .60333 L
.59421 .61169 L
.59227 .61395 L
.58629 .62141 L
.58256 .62646 L
.57933 .63112 L
.57316 .64083 L
.57284 .64137 L
.56766 .65054 L
.56313 .65932 L
.56267 .66026 L
.55815 .66997 L
.55399 .67968 L
.55342 .68107 L
.55016 .68939 L
.54658 .69911 L
.54371 .70733 L
.54321 .70882 L
.54007 .71853 L
.53703 .72824 L
.53408 .73796 L
.53399 .73824 L
.53135 .74767 L
.52859 .75738 L
.52575 .76709 L
.52428 .77204 L
.52318 .77681 L
.52058 .78652 L
.51735 .79623 L
.51457 .80327 L
.51358 .80594 L
.50942 .81566 L
.50486 .82278 L
.49514 .82278 L
.49058 .81566 L
.48642 .80594 L
.48543 .80327 L
.48265 .79623 L
.47942 .78652 L
.47682 .77681 L
.47572 .77204 L
.47425 .76709 L
.47141 .75738 L
.46865 .74767 L
.46601 .73824 L
.46592 .73796 L
.46297 .72824 L
.45993 .71853 L
.45679 .70882 L
.45629 .70733 L
.45342 .69911 L
.44984 .68939 L
.44658 .68107 L
.44601 .67968 L
.44185 .66997 L
.43733 .66026 L
.43687 .65932 L
.43234 .65054 L
.42716 .64137 L
.42684 .64083 L
.42067 .63112 L
.41744 .62646 L
.41371 .62141 L
.40773 .61395 L
.40579 .61169 L
.39802 .60333 L
.39667 .60198 L
.38831 .59421 L
.38605 .59227 L
.37859 .58629 L
.37354 .58256 L
.36888 .57933 L
.35917 .57316 L
.35863 .57284 L
.34946 .56766 L
.34068 .56313 L
.33974 .56267 L
.33003 .55815 L
.32032 .55399 L
.31893 .55342 L
.31061 .55016 L
.30089 .54658 L
.29267 .54371 L
.29118 .54321 L
.28147 .54007 L
.27176 .53703 L
.26204 .53408 L
.26176 .53399 L
.25233 .53135 L
.24262 .52859 L
.23291 .52575 L
.22796 .52428 L
.22319 .52318 L
.21348 .52058 L
.20377 .51735 L
.19673 .51457 L
.19406 .51358 L
.18434 .50942 L
.17722 .50486 L
.17722 .49514 L
.18434 .49058 L
.19406 .48642 L
.19673 .48543 L
.20377 .48265 L
.21348 .47942 L
.22319 .47682 L
.22796 .47572 L
.23291 .47425 L
.24262 .47141 L
.25233 .46865 L
.26176 .46601 L
.26204 .46592 L
.27176 .46297 L
.28147 .45993 L
.29118 .45679 L
.29267 .45629 L
.30089 .45342 L
.31061 .44984 L
.31893 .44658 L
.32032 .44601 L
.33003 .44185 L
.33974 .43733 L
.34068 .43687 L
.34946 .43234 L
.35863 .42716 L
.35917 .42684 L
.36888 .42067 L
.37354 .41744 L
.37859 .41371 L
.38605 .40773 L
.38831 .40579 L
.39667 .39802 L
.39802 .39667 L
.40579 .38831 L
.40773 .38605 L
.41371 .37859 L
.41744 .37354 L
.42067 .36888 L
.42684 .35917 L
.42716 .35863 L
.43234 .34946 L
.43687 .34068 L
.43733 .33974 L
.44185 .33003 L
.44601 .32032 L
.44658 .31893 L
.44984 .31061 L
.45342 .30089 L
.45629 .29267 L
.45679 .29118 L
.45993 .28147 L
.46297 .27176 L
.46592 .26204 L
.46601 .26176 L
.46865 .25233 L
.47141 .24262 L
.47425 .23291 L
.47572 .22796 L
.47682 .22319 L
.47942 .21348 L
.48265 .20377 L
.48543 .19673 L
.48642 .19406 L
.49058 .18434 L
F
0 g
.49514 .17722 m
.50486 .17722 L
.50942 .18434 L
.51358 .19406 L
.51457 .19673 L
.51735 .20377 L
.52058 .21348 L
.52318 .22319 L
.52428 .22796 L
.52575 .23291 L
.52859 .24262 L
.53135 .25233 L
.53399 .26176 L
.53408 .26204 L
.53703 .27176 L
.54007 .28147 L
.54321 .29118 L
.54371 .29267 L
.54658 .30089 L
.55016 .31061 L
.55342 .31893 L
.55399 .32032 L
.55815 .33003 L
.56267 .33974 L
.56313 .34068 L
.56766 .34946 L
.57284 .35863 L
.57316 .35917 L
.57933 .36888 L
.58256 .37354 L
.58629 .37859 L
.59227 .38605 L
.59421 .38831 L
.60198 .39667 L
.60333 .39802 L
.61169 .40579 L
.61395 .40773 L
.62141 .41371 L
.62646 .41744 L
.63112 .42067 L
.64083 .42684 L
.64137 .42716 L
.65054 .43234 L
.65932 .43687 L
.66026 .43733 L
.66997 .44185 L
.67968 .44601 L
.68107 .44658 L
.68939 .44984 L
.69911 .45342 L
Mistroke
.70733 .45629 L
.70882 .45679 L
.71853 .45993 L
.72824 .46297 L
.73796 .46592 L
.73824 .46601 L
.74767 .46865 L
.75738 .47141 L
.76709 .47425 L
.77204 .47572 L
.77681 .47682 L
.78652 .47942 L
.79623 .48265 L
.80327 .48543 L
.80594 .48642 L
.81566 .49058 L
.82278 .49514 L
.82278 .50486 L
.81566 .50942 L
.80594 .51358 L
.80327 .51457 L
.79623 .51735 L
.78652 .52058 L
.77681 .52318 L
.77204 .52428 L
.76709 .52575 L
.75738 .52859 L
.74767 .53135 L
.73824 .53399 L
.73796 .53408 L
.72824 .53703 L
.71853 .54007 L
.70882 .54321 L
.70733 .54371 L
.69911 .54658 L
.68939 .55016 L
.68107 .55342 L
.67968 .55399 L
.66997 .55815 L
.66026 .56267 L
.65932 .56313 L
.65054 .56766 L
.64137 .57284 L
.64083 .57316 L
.63112 .57933 L
.62646 .58256 L
.62141 .58629 L
.61395 .59227 L
.61169 .59421 L
.60333 .60198 L
Mistroke
.60198 .60333 L
.59421 .61169 L
.59227 .61395 L
.58629 .62141 L
.58256 .62646 L
.57933 .63112 L
.57316 .64083 L
.57284 .64137 L
.56766 .65054 L
.56313 .65932 L
.56267 .66026 L
.55815 .66997 L
.55399 .67968 L
.55342 .68107 L
.55016 .68939 L
.54658 .69911 L
.54371 .70733 L
.54321 .70882 L
.54007 .71853 L
.53703 .72824 L
.53408 .73796 L
.53399 .73824 L
.53135 .74767 L
.52859 .75738 L
.52575 .76709 L
.52428 .77204 L
.52318 .77681 L
.52058 .78652 L
.51735 .79623 L
.51457 .80327 L
.51358 .80594 L
.50942 .81566 L
.50486 .82278 L
.49514 .82278 L
.49058 .81566 L
.48642 .80594 L
.48543 .80327 L
.48265 .79623 L
.47942 .78652 L
.47682 .77681 L
.47572 .77204 L
.47425 .76709 L
.47141 .75738 L
.46865 .74767 L
.46601 .73824 L
.46592 .73796 L
.46297 .72824 L
.45993 .71853 L
.45679 .70882 L
.45629 .70733 L
Mistroke
.45342 .69911 L
.44984 .68939 L
.44658 .68107 L
.44601 .67968 L
.44185 .66997 L
.43733 .66026 L
.43687 .65932 L
.43234 .65054 L
.42716 .64137 L
.42684 .64083 L
.42067 .63112 L
.41744 .62646 L
.41371 .62141 L
.40773 .61395 L
.40579 .61169 L
.39802 .60333 L
.39667 .60198 L
.38831 .59421 L
.38605 .59227 L
.37859 .58629 L
.37354 .58256 L
.36888 .57933 L
.35917 .57316 L
.35863 .57284 L
.34946 .56766 L
.34068 .56313 L
.33974 .56267 L
.33003 .55815 L
.32032 .55399 L
.31893 .55342 L
.31061 .55016 L
.30089 .54658 L
.29267 .54371 L
.29118 .54321 L
.28147 .54007 L
.27176 .53703 L
.26204 .53408 L
.26176 .53399 L
.25233 .53135 L
.24262 .52859 L
.23291 .52575 L
.22796 .52428 L
.22319 .52318 L
.21348 .52058 L
.20377 .51735 L
.19673 .51457 L
.19406 .51358 L
.18434 .50942 L
.17722 .50486 L
.17722 .49514 L
Mistroke
.18434 .49058 L
.19406 .48642 L
.19673 .48543 L
.20377 .48265 L
.21348 .47942 L
.22319 .47682 L
.22796 .47572 L
.23291 .47425 L
.24262 .47141 L
.25233 .46865 L
.26176 .46601 L
.26204 .46592 L
.27176 .46297 L
.28147 .45993 L
.29118 .45679 L
.29267 .45629 L
.30089 .45342 L
.31061 .44984 L
.31893 .44658 L
.32032 .44601 L
.33003 .44185 L
.33974 .43733 L
.34068 .43687 L
.34946 .43234 L
.35863 .42716 L
.35917 .42684 L
.36888 .42067 L
.37354 .41744 L
.37859 .41371 L
.38605 .40773 L
.38831 .40579 L
.39667 .39802 L
.39802 .39667 L
.40579 .38831 L
.40773 .38605 L
.41371 .37859 L
.41744 .37354 L
.42067 .36888 L
.42684 .35917 L
.42716 .35863 L
.43234 .34946 L
.43687 .34068 L
.43733 .33974 L
.44185 .33003 L
.44601 .32032 L
.44658 .31893 L
.44984 .31061 L
.45342 .30089 L
.45629 .29267 L
.45679 .29118 L
Mistroke
.45993 .28147 L
.46297 .27176 L
.46592 .26204 L
.46601 .26176 L
.46865 .25233 L
.47141 .24262 L
.47425 .23291 L
.47572 .22796 L
.47682 .22319 L
.47942 .21348 L
.48265 .20377 L
.48543 .19673 L
.48642 .19406 L
.49058 .18434 L
.49514 .17722 L
Mfstroke
.8 1 0 r
.20377 .04225 m
.2054 .04837 L
.20377 .05323 L
.19587 .04837 L
F
0 g
.20377 .04225 m
.2054 .04837 L
.20377 .05323 L
.19587 .04837 L
.20377 .04225 L
s
1 0 .72 r
.49514 .22282 m
.50486 .22282 L
.50524 .22319 L
.51154 .23291 L
.51457 .23959 L
.51602 .24262 L
.52002 .25233 L
.52314 .26204 L
.52428 .26628 L
.52619 .27176 L
.52949 .28147 L
.53271 .29118 L
.53399 .2951 L
.53605 .30089 L
.53958 .31061 L
.54324 .32032 L
.54371 .32152 L
.54719 .33003 L
.55143 .33974 L
.55342 .34408 L
.55603 .34946 L
.56108 .35917 L
.56313 .36287 L
.56667 .36888 L
.57284 .37847 L
.57293 .37859 L
.58002 .38831 L
.58256 .39149 L
.58813 .39802 L
.59227 .40248 L
.59752 .40773 L
.60198 .41187 L
.60851 .41744 L
.61169 .41998 L
.62141 .42707 L
.62153 .42716 L
.63112 .43333 L
.63713 .43687 L
.64083 .43892 L
.65054 .44397 L
.65592 .44658 L
.66026 .44857 L
.66997 .45281 L
.67848 .45629 L
.67968 .45676 L
.68939 .46042 L
.69911 .46395 L
.7049 .46601 L
.70882 .46729 L
.71853 .47051 L
.72824 .47381 L
.73372 .47572 L
.73796 .47686 L
.74767 .47998 L
.75738 .48398 L
.76041 .48543 L
.76709 .48846 L
.77681 .49476 L
.77718 .49514 L
.77718 .50486 L
.77681 .50524 L
.76709 .51154 L
.76041 .51457 L
.75738 .51602 L
.74767 .52002 L
.73796 .52314 L
.73372 .52428 L
.72824 .52619 L
.71853 .52949 L
.70882 .53271 L
.7049 .53399 L
.69911 .53605 L
.68939 .53958 L
.67968 .54324 L
.67848 .54371 L
.66997 .54719 L
.66026 .55143 L
.65592 .55342 L
.65054 .55603 L
.64083 .56108 L
.63713 .56313 L
.63112 .56667 L
.62153 .57284 L
.62141 .57293 L
.61169 .58002 L
.60851 .58256 L
.60198 .58813 L
.59752 .59227 L
.59227 .59752 L
.58813 .60198 L
.58256 .60851 L
.58002 .61169 L
.57293 .62141 L
.57284 .62153 L
.56667 .63112 L
.56313 .63713 L
.56108 .64083 L
.55603 .65054 L
.55342 .65592 L
.55143 .66026 L
.54719 .66997 L
.54371 .67848 L
.54324 .67968 L
.53958 .68939 L
.53605 .69911 L
.53399 .7049 L
.53271 .70882 L
.52949 .71853 L
.52619 .72824 L
.52428 .73372 L
.52314 .73796 L
.52002 .74767 L
.51602 .75738 L
.51457 .76041 L
.51154 .76709 L
.50524 .77681 L
.50486 .77718 L
.49514 .77718 L
.49476 .77681 L
.48846 .76709 L
.48543 .76041 L
.48398 .75738 L
.47998 .74767 L
.47686 .73796 L
.47572 .73372 L
.47381 .72824 L
.47051 .71853 L
.46729 .70882 L
.46601 .7049 L
.46395 .69911 L
.46042 .68939 L
.45676 .67968 L
.45629 .67848 L
.45281 .66997 L
.44857 .66026 L
.44658 .65592 L
.44397 .65054 L
.43892 .64083 L
.43687 .63713 L
.43333 .63112 L
.42716 .62153 L
.42707 .62141 L
.41998 .61169 L
.41744 .60851 L
.41187 .60198 L
.40773 .59752 L
.40248 .59227 L
.39802 .58813 L
.39149 .58256 L
.38831 .58002 L
.37859 .57293 L
.37847 .57284 L
.36888 .56667 L
.36287 .56313 L
.35917 .56108 L
.34946 .55603 L
.34408 .55342 L
.33974 .55143 L
.33003 .54719 L
.32152 .54371 L
.32032 .54324 L
.31061 .53958 L
.30089 .53605 L
.2951 .53399 L
.29118 .53271 L
.28147 .52949 L
.27176 .52619 L
.26628 .52428 L
.26204 .52314 L
.25233 .52002 L
.24262 .51602 L
.23959 .51457 L
.23291 .51154 L
.22319 .50524 L
.22282 .50486 L
.22282 .49514 L
.22319 .49476 L
.23291 .48846 L
.23959 .48543 L
.24262 .48398 L
.25233 .47998 L
.26204 .47686 L
.26628 .47572 L
.27176 .47381 L
.28147 .47051 L
.29118 .46729 L
.2951 .46601 L
.30089 .46395 L
.31061 .46042 L
.32032 .45676 L
.32152 .45629 L
.33003 .45281 L
.33974 .44857 L
.34408 .44658 L
.34946 .44397 L
.35917 .43892 L
.36287 .43687 L
.36888 .43333 L
.37847 .42716 L
.37859 .42707 L
.38831 .41998 L
.39149 .41744 L
.39802 .41187 L
.40248 .40773 L
.40773 .40248 L
.41187 .39802 L
.41744 .39149 L
.41998 .38831 L
.42707 .37859 L
.42716 .37847 L
.43333 .36888 L
.43687 .36287 L
.43892 .35917 L
.44397 .34946 L
.44658 .34408 L
.44857 .33974 L
.45281 .33003 L
.45629 .32152 L
.45676 .32032 L
.46042 .31061 L
.46395 .30089 L
.46601 .2951 L
.46729 .29118 L
.47051 .28147 L
.47381 .27176 L
.47572 .26628 L
.47686 .26204 L
.47998 .25233 L
.48398 .24262 L
.48543 .23959 L
.48846 .23291 L
.49476 .22319 L
F
0 g
.49514 .22282 m
.50486 .22282 L
.50524 .22319 L
.51154 .23291 L
.51457 .23959 L
.51602 .24262 L
.52002 .25233 L
.52314 .26204 L
.52428 .26628 L
.52619 .27176 L
.52949 .28147 L
.53271 .29118 L
.53399 .2951 L
.53605 .30089 L
.53958 .31061 L
.54324 .32032 L
.54371 .32152 L
.54719 .33003 L
.55143 .33974 L
.55342 .34408 L
.55603 .34946 L
.56108 .35917 L
.56313 .36287 L
.56667 .36888 L
.57284 .37847 L
.57293 .37859 L
.58002 .38831 L
.58256 .39149 L
.58813 .39802 L
.59227 .40248 L
.59752 .40773 L
.60198 .41187 L
.60851 .41744 L
.61169 .41998 L
.62141 .42707 L
.62153 .42716 L
.63112 .43333 L
.63713 .43687 L
.64083 .43892 L
.65054 .44397 L
.65592 .44658 L
.66026 .44857 L
.66997 .45281 L
.67848 .45629 L
.67968 .45676 L
.68939 .46042 L
.69911 .46395 L
.7049 .46601 L
.70882 .46729 L
.71853 .47051 L
Mistroke
.72824 .47381 L
.73372 .47572 L
.73796 .47686 L
.74767 .47998 L
.75738 .48398 L
.76041 .48543 L
.76709 .48846 L
.77681 .49476 L
.77718 .49514 L
.77718 .50486 L
.77681 .50524 L
.76709 .51154 L
.76041 .51457 L
.75738 .51602 L
.74767 .52002 L
.73796 .52314 L
.73372 .52428 L
.72824 .52619 L
.71853 .52949 L
.70882 .53271 L
.7049 .53399 L
.69911 .53605 L
.68939 .53958 L
.67968 .54324 L
.67848 .54371 L
.66997 .54719 L
.66026 .55143 L
.65592 .55342 L
.65054 .55603 L
.64083 .56108 L
.63713 .56313 L
.63112 .56667 L
.62153 .57284 L
.62141 .57293 L
.61169 .58002 L
.60851 .58256 L
.60198 .58813 L
.59752 .59227 L
.59227 .59752 L
.58813 .60198 L
.58256 .60851 L
.58002 .61169 L
.57293 .62141 L
.57284 .62153 L
.56667 .63112 L
.56313 .63713 L
.56108 .64083 L
.55603 .65054 L
.55342 .65592 L
.55143 .66026 L
Mistroke
.54719 .66997 L
.54371 .67848 L
.54324 .67968 L
.53958 .68939 L
.53605 .69911 L
.53399 .7049 L
.53271 .70882 L
.52949 .71853 L
.52619 .72824 L
.52428 .73372 L
.52314 .73796 L
.52002 .74767 L
.51602 .75738 L
.51457 .76041 L
.51154 .76709 L
.50524 .77681 L
.50486 .77718 L
.49514 .77718 L
.49476 .77681 L
.48846 .76709 L
.48543 .76041 L
.48398 .75738 L
.47998 .74767 L
.47686 .73796 L
.47572 .73372 L
.47381 .72824 L
.47051 .71853 L
.46729 .70882 L
.46601 .7049 L
.46395 .69911 L
.46042 .68939 L
.45676 .67968 L
.45629 .67848 L
.45281 .66997 L
.44857 .66026 L
.44658 .65592 L
.44397 .65054 L
.43892 .64083 L
.43687 .63713 L
.43333 .63112 L
.42716 .62153 L
.42707 .62141 L
.41998 .61169 L
.41744 .60851 L
.41187 .60198 L
.40773 .59752 L
.40248 .59227 L
.39802 .58813 L
.39149 .58256 L
.38831 .58002 L
Mistroke
.37859 .57293 L
.37847 .57284 L
.36888 .56667 L
.36287 .56313 L
.35917 .56108 L
.34946 .55603 L
.34408 .55342 L
.33974 .55143 L
.33003 .54719 L
.32152 .54371 L
.32032 .54324 L
.31061 .53958 L
.30089 .53605 L
.2951 .53399 L
.29118 .53271 L
.28147 .52949 L
.27176 .52619 L
.26628 .52428 L
.26204 .52314 L
.25233 .52002 L
.24262 .51602 L
.23959 .51457 L
.23291 .51154 L
.22319 .50524 L
.22282 .50486 L
.22282 .49514 L
.22319 .49476 L
.23291 .48846 L
.23959 .48543 L
.24262 .48398 L
.25233 .47998 L
.26204 .47686 L
.26628 .47572 L
.27176 .47381 L
.28147 .47051 L
.29118 .46729 L
.2951 .46601 L
.30089 .46395 L
.31061 .46042 L
.32032 .45676 L
.32152 .45629 L
.33003 .45281 L
.33974 .44857 L
.34408 .44658 L
.34946 .44397 L
.35917 .43892 L
.36287 .43687 L
.36888 .43333 L
.37847 .42716 L
.37859 .42707 L
Mistroke
.38831 .41998 L
.39149 .41744 L
.39802 .41187 L
.40248 .40773 L
.40773 .40248 L
.41187 .39802 L
.41744 .39149 L
.41998 .38831 L
.42707 .37859 L
.42716 .37847 L
.43333 .36888 L
.43687 .36287 L
.43892 .35917 L
.44397 .34946 L
.44658 .34408 L
.44857 .33974 L
.45281 .33003 L
.45629 .32152 L
.45676 .32032 L
.46042 .31061 L
.46395 .30089 L
.46601 .2951 L
.46729 .29118 L
.47051 .28147 L
.47381 .27176 L
.47572 .26628 L
.47686 .26204 L
.47998 .25233 L
.48398 .24262 L
.48543 .23959 L
.48846 .23291 L
.49476 .22319 L
.49514 .22282 L
Mfstroke
1 0 0 r
.32032 .24104 m
.33003 .23314 L
.33591 .24262 L
.33254 .25233 L
.33003 .25791 L
.32794 .26204 L
.32269 .27176 L
.32032 .2754 L
.31662 .28147 L
.31061 .28891 L
.30891 .29118 L
.30089 .29967 L
.29967 .30089 L
.29118 .30891 L
.28891 .31061 L
.28147 .31662 L
.2754 .32032 L
.27176 .32269 L
.26204 .32794 L
.25791 .33003 L
.25233 .33254 L
.24262 .33591 L
.23314 .33003 L
.24104 .32032 L
.24262 .31888 L
.25081 .31061 L
.25233 .30868 L
.26093 .30089 L
.26204 .2997 L
.27099 .29118 L
.27176 .2904 L
.28083 .28147 L
.28147 .28083 L
.2904 .27176 L
.29118 .27099 L
.2997 .26204 L
.30089 .26093 L
.30868 .25233 L
.31061 .25081 L
.31888 .24262 L
F
0 g
.32032 .24104 m
.33003 .23314 L
.33591 .24262 L
.33254 .25233 L
.33003 .25791 L
.32794 .26204 L
.32269 .27176 L
.32032 .2754 L
.31662 .28147 L
.31061 .28891 L
.30891 .29118 L
.30089 .29967 L
.29967 .30089 L
.29118 .30891 L
.28891 .31061 L
.28147 .31662 L
.2754 .32032 L
.27176 .32269 L
.26204 .32794 L
.25791 .33003 L
.25233 .33254 L
.24262 .33591 L
.23314 .33003 L
.24104 .32032 L
.24262 .31888 L
.25081 .31061 L
.25233 .30868 L
.26093 .30089 L
.26204 .2997 L
.27099 .29118 L
.27176 .2904 L
.28083 .28147 L
.28147 .28083 L
.2904 .27176 L
.29118 .27099 L
.2997 .26204 L
.30089 .26093 L
.30868 .25233 L
.31061 .25081 L
.31888 .24262 L
.32032 .24104 L
s
1 0 0 r
.24262 .66409 m
.25233 .66746 L
.25791 .66997 L
.26204 .67206 L
.27176 .67731 L
.2754 .67968 L
.28147 .68338 L
.28891 .68939 L
.29118 .69109 L
.29967 .69911 L
.30089 .70033 L
.30891 .70882 L
.31061 .71109 L
.31662 .71853 L
.32032 .7246 L
.32269 .72824 L
.32794 .73796 L
.33003 .74209 L
.33254 .74767 L
.33591 .75738 L
.33003 .76686 L
.32032 .75896 L
.31888 .75738 L
.31061 .74919 L
.30868 .74767 L
.30089 .73907 L
.2997 .73796 L
.29118 .72901 L
.2904 .72824 L
.28147 .71917 L
.28083 .71853 L
.27176 .7096 L
.27099 .70882 L
.26204 .7003 L
.26093 .69911 L
.25233 .69132 L
.25081 .68939 L
.24262 .68112 L
.24104 .67968 L
.23314 .66997 L
F
0 g
.24262 .66409 m
.25233 .66746 L
.25791 .66997 L
.26204 .67206 L
.27176 .67731 L
.2754 .67968 L
.28147 .68338 L
.28891 .68939 L
.29118 .69109 L
.29967 .69911 L
.30089 .70033 L
.30891 .70882 L
.31061 .71109 L
.31662 .71853 L
.32032 .7246 L
.32269 .72824 L
.32794 .73796 L
.33003 .74209 L
.33254 .74767 L
.33591 .75738 L
.33003 .76686 L
.32032 .75896 L
.31888 .75738 L
.31061 .74919 L
.30868 .74767 L
.30089 .73907 L
.2997 .73796 L
.29118 .72901 L
.2904 .72824 L
.28147 .71917 L
.28083 .71853 L
.27176 .7096 L
.27099 .70882 L
.26204 .7003 L
.26093 .69911 L
.25233 .69132 L
.25081 .68939 L
.24262 .68112 L
.24104 .67968 L
.23314 .66997 L
.24262 .66409 L
s
1 0 .48 r
.49514 .27224 m
.50486 .27224 L
.51207 .28147 L
.51457 .28609 L
.51743 .29118 L
.52182 .30089 L
.52428 .30804 L
.52538 .31061 L
.52942 .32032 L
.53332 .33003 L
.53399 .33172 L
.53748 .33974 L
.54184 .34946 L
.54371 .35348 L
.54651 .35917 L
.5516 .36888 L
.55342 .37217 L
.5572 .37859 L
.56313 .38785 L
.56344 .38831 L
.57051 .39802 L
.57284 .40096 L
.57861 .40773 L
.58256 .412 L
.588 .41744 L
.59227 .42139 L
.59904 .42716 L
.60198 .42949 L
.61169 .43656 L
.61215 .43687 L
.62141 .4428 L
.62783 .44658 L
.63112 .4484 L
.64083 .45349 L
.64652 .45629 L
.65054 .45816 L
.66026 .46252 L
.66828 .46601 L
.66997 .46668 L
.67968 .47058 L
.68939 .47462 L
.69196 .47572 L
.69911 .47818 L
.70882 .48257 L
.71391 .48543 L
.71853 .48793 L
.72776 .49514 L
.72776 .50486 L
.71853 .51207 L
.71391 .51457 L
.70882 .51743 L
.69911 .52182 L
.69196 .52428 L
.68939 .52538 L
.67968 .52942 L
.66997 .53332 L
.66828 .53399 L
.66026 .53748 L
.65054 .54184 L
.64652 .54371 L
.64083 .54651 L
.63112 .5516 L
.62783 .55342 L
.62141 .5572 L
.61215 .56313 L
.61169 .56344 L
.60198 .57051 L
.59904 .57284 L
.59227 .57861 L
.588 .58256 L
.58256 .588 L
.57861 .59227 L
.57284 .59904 L
.57051 .60198 L
.56344 .61169 L
.56313 .61215 L
.5572 .62141 L
.55342 .62783 L
.5516 .63112 L
.54651 .64083 L
.54371 .64652 L
.54184 .65054 L
.53748 .66026 L
.53399 .66828 L
.53332 .66997 L
.52942 .67968 L
.52538 .68939 L
.52428 .69196 L
.52182 .69911 L
.51743 .70882 L
.51457 .71391 L
.51207 .71853 L
.50486 .72776 L
.49514 .72776 L
.48793 .71853 L
.48543 .71391 L
.48257 .70882 L
.47818 .69911 L
.47572 .69196 L
.47462 .68939 L
.47058 .67968 L
.46668 .66997 L
.46601 .66828 L
.46252 .66026 L
.45816 .65054 L
.45629 .64652 L
.45349 .64083 L
.4484 .63112 L
.44658 .62783 L
.4428 .62141 L
.43687 .61215 L
.43656 .61169 L
.42949 .60198 L
.42716 .59904 L
.42139 .59227 L
.41744 .588 L
.412 .58256 L
.40773 .57861 L
.40096 .57284 L
.39802 .57051 L
.38831 .56344 L
.38785 .56313 L
.37859 .5572 L
.37217 .55342 L
.36888 .5516 L
.35917 .54651 L
.35348 .54371 L
.34946 .54184 L
.33974 .53748 L
.33172 .53399 L
.33003 .53332 L
.32032 .52942 L
.31061 .52538 L
.30804 .52428 L
.30089 .52182 L
.29118 .51743 L
.28609 .51457 L
.28147 .51207 L
.27224 .50486 L
.27224 .49514 L
.28147 .48793 L
.28609 .48543 L
.29118 .48257 L
.30089 .47818 L
.30804 .47572 L
.31061 .47462 L
.32032 .47058 L
.33003 .46668 L
.33172 .46601 L
.33974 .46252 L
.34946 .45816 L
.35348 .45629 L
.35917 .45349 L
.36888 .4484 L
.37217 .44658 L
.37859 .4428 L
.38785 .43687 L
.38831 .43656 L
.39802 .42949 L
.40096 .42716 L
.40773 .42139 L
.412 .41744 L
.41744 .412 L
.42139 .40773 L
.42716 .40096 L
.42949 .39802 L
.43656 .38831 L
.43687 .38785 L
.4428 .37859 L
.44658 .37217 L
.4484 .36888 L
.45349 .35917 L
.45629 .35348 L
.45816 .34946 L
.46252 .33974 L
.46601 .33172 L
.46668 .33003 L
.47058 .32032 L
.47462 .31061 L
.47572 .30804 L
.47818 .30089 L
.48257 .29118 L
.48543 .28609 L
.48793 .28147 L
F
0 g
.49514 .27224 m
.50486 .27224 L
.51207 .28147 L
.51457 .28609 L
.51743 .29118 L
.52182 .30089 L
.52428 .30804 L
.52538 .31061 L
.52942 .32032 L
.53332 .33003 L
.53399 .33172 L
.53748 .33974 L
.54184 .34946 L
.54371 .35348 L
.54651 .35917 L
.5516 .36888 L
.55342 .37217 L
.5572 .37859 L
.56313 .38785 L
.56344 .38831 L
.57051 .39802 L
.57284 .40096 L
.57861 .40773 L
.58256 .412 L
.588 .41744 L
.59227 .42139 L
.59904 .42716 L
.60198 .42949 L
.61169 .43656 L
.61215 .43687 L
.62141 .4428 L
.62783 .44658 L
.63112 .4484 L
.64083 .45349 L
.64652 .45629 L
.65054 .45816 L
.66026 .46252 L
.66828 .46601 L
.66997 .46668 L
.67968 .47058 L
.68939 .47462 L
.69196 .47572 L
.69911 .47818 L
.70882 .48257 L
.71391 .48543 L
.71853 .48793 L
.72776 .49514 L
.72776 .50486 L
.71853 .51207 L
.71391 .51457 L
Mistroke
.70882 .51743 L
.69911 .52182 L
.69196 .52428 L
.68939 .52538 L
.67968 .52942 L
.66997 .53332 L
.66828 .53399 L
.66026 .53748 L
.65054 .54184 L
.64652 .54371 L
.64083 .54651 L
.63112 .5516 L
.62783 .55342 L
.62141 .5572 L
.61215 .56313 L
.61169 .56344 L
.60198 .57051 L
.59904 .57284 L
.59227 .57861 L
.588 .58256 L
.58256 .588 L
.57861 .59227 L
.57284 .59904 L
.57051 .60198 L
.56344 .61169 L
.56313 .61215 L
.5572 .62141 L
.55342 .62783 L
.5516 .63112 L
.54651 .64083 L
.54371 .64652 L
.54184 .65054 L
.53748 .66026 L
.53399 .66828 L
.53332 .66997 L
.52942 .67968 L
.52538 .68939 L
.52428 .69196 L
.52182 .69911 L
.51743 .70882 L
.51457 .71391 L
.51207 .71853 L
.50486 .72776 L
.49514 .72776 L
.48793 .71853 L
.48543 .71391 L
.48257 .70882 L
.47818 .69911 L
.47572 .69196 L
.47462 .68939 L
Mistroke
.47058 .67968 L
.46668 .66997 L
.46601 .66828 L
.46252 .66026 L
.45816 .65054 L
.45629 .64652 L
.45349 .64083 L
.4484 .63112 L
.44658 .62783 L
.4428 .62141 L
.43687 .61215 L
.43656 .61169 L
.42949 .60198 L
.42716 .59904 L
.42139 .59227 L
.41744 .588 L
.412 .58256 L
.40773 .57861 L
.40096 .57284 L
.39802 .57051 L
.38831 .56344 L
.38785 .56313 L
.37859 .5572 L
.37217 .55342 L
.36888 .5516 L
.35917 .54651 L
.35348 .54371 L
.34946 .54184 L
.33974 .53748 L
.33172 .53399 L
.33003 .53332 L
.32032 .52942 L
.31061 .52538 L
.30804 .52428 L
.30089 .52182 L
.29118 .51743 L
.28609 .51457 L
.28147 .51207 L
.27224 .50486 L
.27224 .49514 L
.28147 .48793 L
.28609 .48543 L
.29118 .48257 L
.30089 .47818 L
.30804 .47572 L
.31061 .47462 L
.32032 .47058 L
.33003 .46668 L
.33172 .46601 L
.33974 .46252 L
Mistroke
.34946 .45816 L
.35348 .45629 L
.35917 .45349 L
.36888 .4484 L
.37217 .44658 L
.37859 .4428 L
.38785 .43687 L
.38831 .43656 L
.39802 .42949 L
.40096 .42716 L
.40773 .42139 L
.412 .41744 L
.41744 .412 L
.42139 .40773 L
.42716 .40096 L
.42949 .39802 L
.43656 .38831 L
.43687 .38785 L
.4428 .37859 L
.44658 .37217 L
.4484 .36888 L
.45349 .35917 L
.45629 .35348 L
.45816 .34946 L
.46252 .33974 L
.46601 .33172 L
.46668 .33003 L
.47058 .32032 L
.47462 .31061 L
.47572 .30804 L
.47818 .30089 L
.48257 .29118 L
.48543 .28609 L
.48793 .28147 L
.49514 .27224 L
Mfstroke
1 0 .24 r
.49514 .32989 m
.50486 .32989 L
.50508 .33003 L
.5141 .33974 L
.51457 .34043 L
.52065 .34946 L
.52428 .35708 L
.52545 .35917 L
.53075 .36888 L
.53399 .37507 L
.53599 .37859 L
.54166 .38831 L
.54371 .3917 L
.5478 .39802 L
.55342 .40613 L
.55461 .40773 L
.56234 .41744 L
.56313 .41838 L
.57124 .42716 L
.57284 .42876 L
.58162 .43687 L
.58256 .43766 L
.59227 .44539 L
.59387 .44658 L
.60198 .4522 L
.6083 .45629 L
.61169 .45834 L
.62141 .46401 L
.62493 .46601 L
.63112 .46925 L
.64083 .47455 L
.64292 .47572 L
.65054 .47935 L
.65957 .48543 L
.66026 .4859 L
.66997 .49492 L
.67011 .49514 L
.67011 .50486 L
.66997 .50508 L
.66026 .5141 L
.65957 .51457 L
.65054 .52065 L
.64292 .52428 L
.64083 .52545 L
.63112 .53075 L
.62493 .53399 L
.62141 .53599 L
.61169 .54166 L
.6083 .54371 L
.60198 .5478 L
.59387 .55342 L
.59227 .55461 L
.58256 .56234 L
.58162 .56313 L
.57284 .57124 L
.57124 .57284 L
.56313 .58162 L
.56234 .58256 L
.55461 .59227 L
.55342 .59387 L
.5478 .60198 L
.54371 .6083 L
.54166 .61169 L
.53599 .62141 L
.53399 .62493 L
.53075 .63112 L
.52545 .64083 L
.52428 .64292 L
.52065 .65054 L
.51457 .65957 L
.5141 .66026 L
.50508 .66997 L
.50486 .67011 L
.49514 .67011 L
.49492 .66997 L
.4859 .66026 L
.48543 .65957 L
.47935 .65054 L
.47572 .64292 L
.47455 .64083 L
.46925 .63112 L
.46601 .62493 L
.46401 .62141 L
.45834 .61169 L
.45629 .6083 L
.4522 .60198 L
.44658 .59387 L
.44539 .59227 L
.43766 .58256 L
.43687 .58162 L
.42876 .57284 L
.42716 .57124 L
.41838 .56313 L
.41744 .56234 L
.40773 .55461 L
.40613 .55342 L
.39802 .5478 L
.3917 .54371 L
.38831 .54166 L
.37859 .53599 L
.37507 .53399 L
.36888 .53075 L
.35917 .52545 L
.35708 .52428 L
.34946 .52065 L
.34043 .51457 L
.33974 .5141 L
.33003 .50508 L
.32989 .50486 L
.32989 .49514 L
.33003 .49492 L
.33974 .4859 L
.34043 .48543 L
.34946 .47935 L
.35708 .47572 L
.35917 .47455 L
.36888 .46925 L
.37507 .46601 L
.37859 .46401 L
.38831 .45834 L
.3917 .45629 L
.39802 .4522 L
.40613 .44658 L
.40773 .44539 L
.41744 .43766 L
.41838 .43687 L
.42716 .42876 L
.42876 .42716 L
.43687 .41838 L
.43766 .41744 L
.44539 .40773 L
.44658 .40613 L
.4522 .39802 L
.45629 .3917 L
.45834 .38831 L
.46401 .37859 L
.46601 .37507 L
.46925 .36888 L
.47455 .35917 L
.47572 .35708 L
.47935 .34946 L
.48543 .34043 L
.4859 .33974 L
.49492 .33003 L
F
0 g
.49514 .32989 m
.50486 .32989 L
.50508 .33003 L
.5141 .33974 L
.51457 .34043 L
.52065 .34946 L
.52428 .35708 L
.52545 .35917 L
.53075 .36888 L
.53399 .37507 L
.53599 .37859 L
.54166 .38831 L
.54371 .3917 L
.5478 .39802 L
.55342 .40613 L
.55461 .40773 L
.56234 .41744 L
.56313 .41838 L
.57124 .42716 L
.57284 .42876 L
.58162 .43687 L
.58256 .43766 L
.59227 .44539 L
.59387 .44658 L
.60198 .4522 L
.6083 .45629 L
.61169 .45834 L
.62141 .46401 L
.62493 .46601 L
.63112 .46925 L
.64083 .47455 L
.64292 .47572 L
.65054 .47935 L
.65957 .48543 L
.66026 .4859 L
.66997 .49492 L
.67011 .49514 L
.67011 .50486 L
.66997 .50508 L
.66026 .5141 L
.65957 .51457 L
.65054 .52065 L
.64292 .52428 L
.64083 .52545 L
.63112 .53075 L
.62493 .53399 L
.62141 .53599 L
.61169 .54166 L
.6083 .54371 L
.60198 .5478 L
Mistroke
.59387 .55342 L
.59227 .55461 L
.58256 .56234 L
.58162 .56313 L
.57284 .57124 L
.57124 .57284 L
.56313 .58162 L
.56234 .58256 L
.55461 .59227 L
.55342 .59387 L
.5478 .60198 L
.54371 .6083 L
.54166 .61169 L
.53599 .62141 L
.53399 .62493 L
.53075 .63112 L
.52545 .64083 L
.52428 .64292 L
.52065 .65054 L
.51457 .65957 L
.5141 .66026 L
.50508 .66997 L
.50486 .67011 L
.49514 .67011 L
.49492 .66997 L
.4859 .66026 L
.48543 .65957 L
.47935 .65054 L
.47572 .64292 L
.47455 .64083 L
.46925 .63112 L
.46601 .62493 L
.46401 .62141 L
.45834 .61169 L
.45629 .6083 L
.4522 .60198 L
.44658 .59387 L
.44539 .59227 L
.43766 .58256 L
.43687 .58162 L
.42876 .57284 L
.42716 .57124 L
.41838 .56313 L
.41744 .56234 L
.40773 .55461 L
.40613 .55342 L
.39802 .5478 L
.3917 .54371 L
.38831 .54166 L
.37859 .53599 L
Mistroke
.37507 .53399 L
.36888 .53075 L
.35917 .52545 L
.35708 .52428 L
.34946 .52065 L
.34043 .51457 L
.33974 .5141 L
.33003 .50508 L
.32989 .50486 L
.32989 .49514 L
.33003 .49492 L
.33974 .4859 L
.34043 .48543 L
.34946 .47935 L
.35708 .47572 L
.35917 .47455 L
.36888 .46925 L
.37507 .46601 L
.37859 .46401 L
.38831 .45834 L
.3917 .45629 L
.39802 .4522 L
.40613 .44658 L
.40773 .44539 L
.41744 .43766 L
.41838 .43687 L
.42716 .42876 L
.42876 .42716 L
.43687 .41838 L
.43766 .41744 L
.44539 .40773 L
.44658 .40613 L
.4522 .39802 L
.45629 .3917 L
.45834 .38831 L
.46401 .37859 L
.46601 .37507 L
.46925 .36888 L
.47455 .35917 L
.47572 .35708 L
.47935 .34946 L
.48543 .34043 L
.4859 .33974 L
.49492 .33003 L
.49514 .32989 L
Mfstroke
1 0 0 r
.49514 .41219 m
.50486 .41219 L
.51351 .41744 L
.51457 .41835 L
.52356 .42716 L
.52428 .42817 L
.53216 .43687 L
.53399 .43901 L
.54101 .44658 L
.54371 .44942 L
.55058 .45629 L
.55342 .45899 L
.56099 .46601 L
.56313 .46784 L
.57183 .47572 L
.57284 .47644 L
.58165 .48543 L
.58256 .48649 L
.58781 .49514 L
.58781 .50486 L
.58256 .51351 L
.58165 .51457 L
.57284 .52356 L
.57183 .52428 L
.56313 .53216 L
.56099 .53399 L
.55342 .54101 L
.55058 .54371 L
.54371 .55058 L
.54101 .55342 L
.53399 .56099 L
.53216 .56313 L
.52428 .57183 L
.52356 .57284 L
.51457 .58165 L
.51351 .58256 L
.50486 .58781 L
.49514 .58781 L
.48649 .58256 L
.48543 .58165 L
.47644 .57284 L
.47572 .57183 L
.46784 .56313 L
.46601 .56099 L
.45899 .55342 L
.45629 .55058 L
.44942 .54371 L
.44658 .54101 L
.43901 .53399 L
.43687 .53216 L
.42817 .52428 L
.42716 .52356 L
.41835 .51457 L
.41744 .51351 L
.41219 .50486 L
.41219 .49514 L
.41744 .48649 L
.41835 .48543 L
.42716 .47644 L
.42817 .47572 L
.43687 .46784 L
.43901 .46601 L
.44658 .45899 L
.44942 .45629 L
.45629 .44942 L
.45899 .44658 L
.46601 .43901 L
.46784 .43687 L
.47572 .42817 L
.47644 .42716 L
.48543 .41835 L
.48649 .41744 L
F
0 g
.49514 .41219 m
.50486 .41219 L
.51351 .41744 L
.51457 .41835 L
.52356 .42716 L
.52428 .42817 L
.53216 .43687 L
.53399 .43901 L
.54101 .44658 L
.54371 .44942 L
.55058 .45629 L
.55342 .45899 L
.56099 .46601 L
.56313 .46784 L
.57183 .47572 L
.57284 .47644 L
.58165 .48543 L
.58256 .48649 L
.58781 .49514 L
.58781 .50486 L
.58256 .51351 L
.58165 .51457 L
.57284 .52356 L
.57183 .52428 L
.56313 .53216 L
.56099 .53399 L
.55342 .54101 L
.55058 .54371 L
.54371 .55058 L
.54101 .55342 L
.53399 .56099 L
.53216 .56313 L
.52428 .57183 L
.52356 .57284 L
.51457 .58165 L
.51351 .58256 L
.50486 .58781 L
.49514 .58781 L
.48649 .58256 L
.48543 .58165 L
.47644 .57284 L
.47572 .57183 L
.46784 .56313 L
.46601 .56099 L
.45899 .55342 L
.45629 .55058 L
.44942 .54371 L
.44658 .54101 L
.43901 .53399 L
.43687 .53216 L
Mistroke
.42817 .52428 L
.42716 .52356 L
.41835 .51457 L
.41744 .51351 L
.41219 .50486 L
.41219 .49514 L
.41744 .48649 L
.41835 .48543 L
.42716 .47644 L
.42817 .47572 L
.43687 .46784 L
.43901 .46601 L
.44658 .45899 L
.44942 .45629 L
.45629 .44942 L
.45899 .44658 L
.46601 .43901 L
.46784 .43687 L
.47572 .42817 L
.47644 .42716 L
.48543 .41835 L
.48649 .41744 L
.49514 .41219 L
Mfstroke
0 .16 1 r
.98077 .48789 m
.97106 .48555 L
.97055 .48543 L
.96134 .48316 L
.95163 .48104 L
.94192 .47919 L
.93221 .47756 L
.92249 .47611 L
.91965 .47572 L
.91278 .47449 L
.90307 .47277 L
.89336 .47107 L
.88364 .4694 L
.87393 .46775 L
.86422 .46609 L
.86373 .46601 L
.85451 .46432 L
.84479 .46251 L
.83508 .46066 L
.82537 .45876 L
.81566 .45681 L
.81314 .45629 L
.80594 .45475 L
.79623 .4526 L
.78652 .45036 L
.77681 .448 L
.77116 .44658 L
.76709 .44552 L
.75738 .44287 L
.74767 .44007 L
.73796 .43709 L
.73727 .43687 L
.72824 .43388 L
.71853 .43043 L
.70995 .42716 L
.70882 .42671 L
.69911 .42266 L
.68939 .41825 L
.68771 .41744 L
.67968 .41341 L
.66997 .40808 L
.66936 .40773 L
.66026 .40217 L
.65403 .39802 L
.65054 .39556 L
.64105 .38831 L
.64083 .38813 L
.63112 .37971 L
.62992 .37859 L
.62141 .37008 L
.62029 .36888 L
.61187 .35917 L
.61169 .35895 L
.60444 .34946 L
.60198 .34597 L
.59783 .33974 L
.59227 .33064 L
.59192 .33003 L
.58659 .32032 L
.58256 .31229 L
.58175 .31061 L
.57734 .30089 L
.57329 .29118 L
.57284 .29005 L
.56957 .28147 L
.56612 .27176 L
.56313 .26273 L
.56291 .26204 L
.55993 .25233 L
.55713 .24262 L
.55448 .23291 L
.55342 .22884 L
.552 .22319 L
.54964 .21348 L
.5474 .20377 L
.54525 .19406 L
.54371 .18686 L
.54319 .18434 L
.54124 .17463 L
.53934 .16492 L
.53749 .15521 L
.53568 .14549 L
.53399 .13627 L
.53391 .13578 L
.53225 .12607 L
.5306 .11636 L
.52893 .10664 L
.52723 .09693 L
.52551 .08722 L
.52428 .08035 L
.52389 .07751 L
.52244 .06779 L
.52081 .05808 L
.51896 .04837 L
.51684 .03866 L
.51457 .02945 L
.51445 .02894 L
.51211 .01923 L
.98077 .01923 L
F
0 g
.98077 .48789 m
.97106 .48555 L
.97055 .48543 L
.96134 .48316 L
.95163 .48104 L
.94192 .47919 L
.93221 .47756 L
.92249 .47611 L
.91965 .47572 L
.91278 .47449 L
.90307 .47277 L
.89336 .47107 L
.88364 .4694 L
.87393 .46775 L
.86422 .46609 L
.86373 .46601 L
.85451 .46432 L
.84479 .46251 L
.83508 .46066 L
.82537 .45876 L
.81566 .45681 L
.81314 .45629 L
.80594 .45475 L
.79623 .4526 L
.78652 .45036 L
.77681 .448 L
.77116 .44658 L
.76709 .44552 L
.75738 .44287 L
.74767 .44007 L
.73796 .43709 L
.73727 .43687 L
.72824 .43388 L
.71853 .43043 L
.70995 .42716 L
.70882 .42671 L
.69911 .42266 L
.68939 .41825 L
.68771 .41744 L
.67968 .41341 L
.66997 .40808 L
.66936 .40773 L
.66026 .40217 L
.65403 .39802 L
.65054 .39556 L
.64105 .38831 L
.64083 .38813 L
.63112 .37971 L
.62992 .37859 L
.62141 .37008 L
Mistroke
.62029 .36888 L
.61187 .35917 L
.61169 .35895 L
.60444 .34946 L
.60198 .34597 L
.59783 .33974 L
.59227 .33064 L
.59192 .33003 L
.58659 .32032 L
.58256 .31229 L
.58175 .31061 L
.57734 .30089 L
.57329 .29118 L
.57284 .29005 L
.56957 .28147 L
.56612 .27176 L
.56313 .26273 L
.56291 .26204 L
.55993 .25233 L
.55713 .24262 L
.55448 .23291 L
.55342 .22884 L
.552 .22319 L
.54964 .21348 L
.5474 .20377 L
.54525 .19406 L
.54371 .18686 L
.54319 .18434 L
.54124 .17463 L
.53934 .16492 L
.53749 .15521 L
.53568 .14549 L
.53399 .13627 L
.53391 .13578 L
.53225 .12607 L
.5306 .11636 L
.52893 .10664 L
.52723 .09693 L
.52551 .08722 L
.52428 .08035 L
.52389 .07751 L
.52244 .06779 L
.52081 .05808 L
.51896 .04837 L
.51684 .03866 L
.51457 .02945 L
.51445 .02894 L
.51211 .01923 L
Mfstroke
0 .16 1 r
.98077 .51211 m
.97106 .51445 L
.97055 .51457 L
.96134 .51684 L
.95163 .51896 L
.94192 .52081 L
.93221 .52244 L
.92249 .52389 L
.91965 .52428 L
.91278 .52551 L
.90307 .52723 L
.89336 .52893 L
.88364 .5306 L
.87393 .53225 L
.86422 .53391 L
.86373 .53399 L
.85451 .53568 L
.84479 .53749 L
.83508 .53934 L
.82537 .54124 L
.81566 .54319 L
.81314 .54371 L
.80594 .54525 L
.79623 .5474 L
.78652 .54964 L
.77681 .552 L
.77116 .55342 L
.76709 .55448 L
.75738 .55713 L
.74767 .55993 L
.73796 .56291 L
.73727 .56313 L
.72824 .56612 L
.71853 .56957 L
.70995 .57284 L
.70882 .57329 L
.69911 .57734 L
.68939 .58175 L
.68771 .58256 L
.67968 .58659 L
.66997 .59192 L
.66936 .59227 L
.66026 .59783 L
.65403 .60198 L
.65054 .60444 L
.64105 .61169 L
.64083 .61187 L
.63112 .62029 L
.62992 .62141 L
.62141 .62992 L
.62029 .63112 L
.61187 .64083 L
.61169 .64105 L
.60444 .65054 L
.60198 .65403 L
.59783 .66026 L
.59227 .66936 L
.59192 .66997 L
.58659 .67968 L
.58256 .68771 L
.58175 .68939 L
.57734 .69911 L
.57329 .70882 L
.57284 .70995 L
.56957 .71853 L
.56612 .72824 L
.56313 .73727 L
.56291 .73796 L
.55993 .74767 L
.55713 .75738 L
.55448 .76709 L
.55342 .77116 L
.552 .77681 L
.54964 .78652 L
.5474 .79623 L
.54525 .80594 L
.54371 .81314 L
.54319 .81566 L
.54124 .82537 L
.53934 .83508 L
.53749 .84479 L
.53568 .85451 L
.53399 .86373 L
.53391 .86422 L
.53225 .87393 L
.5306 .88364 L
.52893 .89336 L
.52723 .90307 L
.52551 .91278 L
.52428 .91965 L
.52389 .92249 L
.52244 .93221 L
.52081 .94192 L
.51896 .95163 L
.51684 .96134 L
.51457 .97055 L
.51445 .97106 L
.51211 .98077 L
.98077 .98077 L
F
0 g
.98077 .51211 m
.97106 .51445 L
.97055 .51457 L
.96134 .51684 L
.95163 .51896 L
.94192 .52081 L
.93221 .52244 L
.92249 .52389 L
.91965 .52428 L
.91278 .52551 L
.90307 .52723 L
.89336 .52893 L
.88364 .5306 L
.87393 .53225 L
.86422 .53391 L
.86373 .53399 L
.85451 .53568 L
.84479 .53749 L
.83508 .53934 L
.82537 .54124 L
.81566 .54319 L
.81314 .54371 L
.80594 .54525 L
.79623 .5474 L
.78652 .54964 L
.77681 .552 L
.77116 .55342 L
.76709 .55448 L
.75738 .55713 L
.74767 .55993 L
.73796 .56291 L
.73727 .56313 L
.72824 .56612 L
.71853 .56957 L
.70995 .57284 L
.70882 .57329 L
.69911 .57734 L
.68939 .58175 L
.68771 .58256 L
.67968 .58659 L
.66997 .59192 L
.66936 .59227 L
.66026 .59783 L
.65403 .60198 L
.65054 .60444 L
.64105 .61169 L
.64083 .61187 L
.63112 .62029 L
.62992 .62141 L
.62141 .62992 L
Mistroke
.62029 .63112 L
.61187 .64083 L
.61169 .64105 L
.60444 .65054 L
.60198 .65403 L
.59783 .66026 L
.59227 .66936 L
.59192 .66997 L
.58659 .67968 L
.58256 .68771 L
.58175 .68939 L
.57734 .69911 L
.57329 .70882 L
.57284 .70995 L
.56957 .71853 L
.56612 .72824 L
.56313 .73727 L
.56291 .73796 L
.55993 .74767 L
.55713 .75738 L
.55448 .76709 L
.55342 .77116 L
.552 .77681 L
.54964 .78652 L
.5474 .79623 L
.54525 .80594 L
.54371 .81314 L
.54319 .81566 L
.54124 .82537 L
.53934 .83508 L
.53749 .84479 L
.53568 .85451 L
.53399 .86373 L
.53391 .86422 L
.53225 .87393 L
.5306 .88364 L
.52893 .89336 L
.52723 .90307 L
.52551 .91278 L
.52428 .91965 L
.52389 .92249 L
.52244 .93221 L
.52081 .94192 L
.51896 .95163 L
.51684 .96134 L
.51457 .97055 L
.51445 .97106 L
.51211 .98077 L
Mfstroke
0 .4 1 r
.98077 .4785 m
.97106 .47711 L
.96134 .47586 L
.96021 .47572 L
.95163 .47436 L
.94192 .47284 L
.93221 .47134 L
.92249 .46986 L
.91278 .46839 L
.90307 .46692 L
.89706 .46601 L
.89336 .46541 L
.88364 .46383 L
.87393 .46221 L
.86422 .46056 L
.85451 .45887 L
.84479 .45713 L
.8402 .45629 L
.83508 .45532 L
.82537 .45342 L
.81566 .45144 L
.80594 .44938 L
.79623 .44721 L
.79347 .44658 L
.78652 .44493 L
.77681 .44251 L
.76709 .43995 L
.75738 .43724 L
.7561 .43687 L
.74767 .43434 L
.73796 .43123 L
.72824 .42789 L
.72619 .42716 L
.71853 .42429 L
.70882 .42038 L
.702 .41744 L
.69911 .41614 L
.68939 .41149 L
.68214 .40773 L
.67968 .40639 L
.66997 .40075 L
.6656 .39802 L
.66026 .39448 L
.65165 .38831 L
.65054 .38747 L
.64083 .37956 L
.63973 .37859 L
.63112 .37057 L
.62943 .36888 L
.62141 .36027 L
.62044 .35917 L
.61253 .34946 L
.61169 .34835 L
.60552 .33974 L
.60198 .3344 L
.59925 .33003 L
.59361 .32032 L
.59227 .31786 L
.58851 .31061 L
.58386 .30089 L
.58256 .298 L
.57962 .29118 L
.57571 .28147 L
.57284 .27381 L
.57211 .27176 L
.56877 .26204 L
.56566 .25233 L
.56313 .2439 L
.56276 .24262 L
.56005 .23291 L
.55749 .22319 L
.55507 .21348 L
.55342 .20653 L
.55279 .20377 L
.55062 .19406 L
.54856 .18434 L
.54658 .17463 L
.54468 .16492 L
.54371 .1598 L
.54287 .15521 L
.54113 .14549 L
.53944 .13578 L
.53779 .12607 L
.53617 .11636 L
.53459 .10664 L
.53399 .10294 L
.53308 .09693 L
.53161 .08722 L
.53014 .07751 L
.52866 .06779 L
.52716 .05808 L
.52564 .04837 L
.52428 .03979 L
.52414 .03866 L
.52289 .02894 L
.5215 .01923 L
.98077 .01923 L
F
0 g
.98077 .4785 m
.97106 .47711 L
.96134 .47586 L
.96021 .47572 L
.95163 .47436 L
.94192 .47284 L
.93221 .47134 L
.92249 .46986 L
.91278 .46839 L
.90307 .46692 L
.89706 .46601 L
.89336 .46541 L
.88364 .46383 L
.87393 .46221 L
.86422 .46056 L
.85451 .45887 L
.84479 .45713 L
.8402 .45629 L
.83508 .45532 L
.82537 .45342 L
.81566 .45144 L
.80594 .44938 L
.79623 .44721 L
.79347 .44658 L
.78652 .44493 L
.77681 .44251 L
.76709 .43995 L
.75738 .43724 L
.7561 .43687 L
.74767 .43434 L
.73796 .43123 L
.72824 .42789 L
.72619 .42716 L
.71853 .42429 L
.70882 .42038 L
.702 .41744 L
.69911 .41614 L
.68939 .41149 L
.68214 .40773 L
.67968 .40639 L
.66997 .40075 L
.6656 .39802 L
.66026 .39448 L
.65165 .38831 L
.65054 .38747 L
.64083 .37956 L
.63973 .37859 L
.63112 .37057 L
.62943 .36888 L
.62141 .36027 L
Mistroke
.62044 .35917 L
.61253 .34946 L
.61169 .34835 L
.60552 .33974 L
.60198 .3344 L
.59925 .33003 L
.59361 .32032 L
.59227 .31786 L
.58851 .31061 L
.58386 .30089 L
.58256 .298 L
.57962 .29118 L
.57571 .28147 L
.57284 .27381 L
.57211 .27176 L
.56877 .26204 L
.56566 .25233 L
.56313 .2439 L
.56276 .24262 L
.56005 .23291 L
.55749 .22319 L
.55507 .21348 L
.55342 .20653 L
.55279 .20377 L
.55062 .19406 L
.54856 .18434 L
.54658 .17463 L
.54468 .16492 L
.54371 .1598 L
.54287 .15521 L
.54113 .14549 L
.53944 .13578 L
.53779 .12607 L
.53617 .11636 L
.53459 .10664 L
.53399 .10294 L
.53308 .09693 L
.53161 .08722 L
.53014 .07751 L
.52866 .06779 L
.52716 .05808 L
.52564 .04837 L
.52428 .03979 L
.52414 .03866 L
.52289 .02894 L
.5215 .01923 L
Mfstroke
0 .4 1 r
.98077 .5215 m
.97106 .52289 L
.96134 .52414 L
.96021 .52428 L
.95163 .52564 L
.94192 .52716 L
.93221 .52866 L
.92249 .53014 L
.91278 .53161 L
.90307 .53308 L
.89706 .53399 L
.89336 .53459 L
.88364 .53617 L
.87393 .53779 L
.86422 .53944 L
.85451 .54113 L
.84479 .54287 L
.8402 .54371 L
.83508 .54468 L
.82537 .54658 L
.81566 .54856 L
.80594 .55062 L
.79623 .55279 L
.79347 .55342 L
.78652 .55507 L
.77681 .55749 L
.76709 .56005 L
.75738 .56276 L
.7561 .56313 L
.74767 .56566 L
.73796 .56877 L
.72824 .57211 L
.72619 .57284 L
.71853 .57571 L
.70882 .57962 L
.702 .58256 L
.69911 .58386 L
.68939 .58851 L
.68214 .59227 L
.67968 .59361 L
.66997 .59925 L
.6656 .60198 L
.66026 .60552 L
.65165 .61169 L
.65054 .61253 L
.64083 .62044 L
.63973 .62141 L
.63112 .62943 L
.62943 .63112 L
.62141 .63973 L
.62044 .64083 L
.61253 .65054 L
.61169 .65165 L
.60552 .66026 L
.60198 .6656 L
.59925 .66997 L
.59361 .67968 L
.59227 .68214 L
.58851 .68939 L
.58386 .69911 L
.58256 .702 L
.57962 .70882 L
.57571 .71853 L
.57284 .72619 L
.57211 .72824 L
.56877 .73796 L
.56566 .74767 L
.56313 .7561 L
.56276 .75738 L
.56005 .76709 L
.55749 .77681 L
.55507 .78652 L
.55342 .79347 L
.55279 .79623 L
.55062 .80594 L
.54856 .81566 L
.54658 .82537 L
.54468 .83508 L
.54371 .8402 L
.54287 .84479 L
.54113 .85451 L
.53944 .86422 L
.53779 .87393 L
.53617 .88364 L
.53459 .89336 L
.53399 .89706 L
.53308 .90307 L
.53161 .91278 L
.53014 .92249 L
.52866 .93221 L
.52716 .94192 L
.52564 .95163 L
.52428 .96021 L
.52414 .96134 L
.52289 .97106 L
.5215 .98077 L
.98077 .98077 L
F
0 g
.98077 .5215 m
.97106 .52289 L
.96134 .52414 L
.96021 .52428 L
.95163 .52564 L
.94192 .52716 L
.93221 .52866 L
.92249 .53014 L
.91278 .53161 L
.90307 .53308 L
.89706 .53399 L
.89336 .53459 L
.88364 .53617 L
.87393 .53779 L
.86422 .53944 L
.85451 .54113 L
.84479 .54287 L
.8402 .54371 L
.83508 .54468 L
.82537 .54658 L
.81566 .54856 L
.80594 .55062 L
.79623 .55279 L
.79347 .55342 L
.78652 .55507 L
.77681 .55749 L
.76709 .56005 L
.75738 .56276 L
.7561 .56313 L
.74767 .56566 L
.73796 .56877 L
.72824 .57211 L
.72619 .57284 L
.71853 .57571 L
.70882 .57962 L
.702 .58256 L
.69911 .58386 L
.68939 .58851 L
.68214 .59227 L
.67968 .59361 L
.66997 .59925 L
.6656 .60198 L
.66026 .60552 L
.65165 .61169 L
.65054 .61253 L
.64083 .62044 L
.63973 .62141 L
.63112 .62943 L
.62943 .63112 L
.62141 .63973 L
Mistroke
.62044 .64083 L
.61253 .65054 L
.61169 .65165 L
.60552 .66026 L
.60198 .6656 L
.59925 .66997 L
.59361 .67968 L
.59227 .68214 L
.58851 .68939 L
.58386 .69911 L
.58256 .702 L
.57962 .70882 L
.57571 .71853 L
.57284 .72619 L
.57211 .72824 L
.56877 .73796 L
.56566 .74767 L
.56313 .7561 L
.56276 .75738 L
.56005 .76709 L
.55749 .77681 L
.55507 .78652 L
.55342 .79347 L
.55279 .79623 L
.55062 .80594 L
.54856 .81566 L
.54658 .82537 L
.54468 .83508 L
.54371 .8402 L
.54287 .84479 L
.54113 .85451 L
.53944 .86422 L
.53779 .87393 L
.53617 .88364 L
.53459 .89336 L
.53399 .89706 L
.53308 .90307 L
.53161 .91278 L
.53014 .92249 L
.52866 .93221 L
.52716 .94192 L
.52564 .95163 L
.52428 .96021 L
.52414 .96134 L
.52289 .97106 L
.5215 .98077 L
Mfstroke
0 .64 1 r
.98077 .4724 m
.97106 .47109 L
.96134 .46979 L
.95163 .4685 L
.94192 .4672 L
.93294 .46601 L
.93221 .4659 L
.92249 .46452 L
.91278 .46311 L
.90307 .46167 L
.89336 .46021 L
.88364 .4587 L
.87393 .45715 L
.86865 .45629 L
.86422 .45555 L
.85451 .45386 L
.84479 .45211 L
.83508 .4503 L
.82537 .4484 L
.81646 .44658 L
.81566 .44641 L
.80594 .44431 L
.79623 .4421 L
.78652 .43976 L
.77681 .43729 L
.77522 .43687 L
.76709 .43465 L
.75738 .43184 L
.74767 .42883 L
.74253 .42716 L
.73796 .4256 L
.72824 .42212 L
.71853 .41836 L
.71627 .41744 L
.70882 .41428 L
.69911 .40982 L
.69483 .40773 L
.68939 .40494 L
.67968 .39957 L
.67705 .39802 L
.66997 .39362 L
.6621 .38831 L
.66026 .38699 L
.65054 .37956 L
.64936 .37859 L
.64083 .37117 L
.63838 .36888 L
.63112 .36162 L
.62883 .35917 L
.62141 .35064 L
.62044 .34946 L
.61301 .33974 L
.61169 .3379 L
.60638 .33003 L
.60198 .32295 L
.60043 .32032 L
.59506 .31061 L
.59227 .30517 L
.59018 .30089 L
.58572 .29118 L
.58256 .28373 L
.58164 .28147 L
.57788 .27176 L
.5744 .26204 L
.57284 .25747 L
.57117 .25233 L
.56816 .24262 L
.56535 .23291 L
.56313 .22478 L
.56271 .22319 L
.56024 .21348 L
.5579 .20377 L
.55569 .19406 L
.55359 .18434 L
.55342 .18354 L
.5516 .17463 L
.5497 .16492 L
.54789 .15521 L
.54614 .14549 L
.54445 .13578 L
.54371 .13135 L
.54285 .12607 L
.5413 .11636 L
.53979 .10664 L
.53833 .09693 L
.53689 .08722 L
.53548 .07751 L
.5341 .06779 L
.53399 .06706 L
.5328 .05808 L
.5315 .04837 L
.53021 .03866 L
.52891 .02894 L
.5276 .01923 L
.98077 .01923 L
F
0 g
.98077 .4724 m
.97106 .47109 L
.96134 .46979 L
.95163 .4685 L
.94192 .4672 L
.93294 .46601 L
.93221 .4659 L
.92249 .46452 L
.91278 .46311 L
.90307 .46167 L
.89336 .46021 L
.88364 .4587 L
.87393 .45715 L
.86865 .45629 L
.86422 .45555 L
.85451 .45386 L
.84479 .45211 L
.83508 .4503 L
.82537 .4484 L
.81646 .44658 L
.81566 .44641 L
.80594 .44431 L
.79623 .4421 L
.78652 .43976 L
.77681 .43729 L
.77522 .43687 L
.76709 .43465 L
.75738 .43184 L
.74767 .42883 L
.74253 .42716 L
.73796 .4256 L
.72824 .42212 L
.71853 .41836 L
.71627 .41744 L
.70882 .41428 L
.69911 .40982 L
.69483 .40773 L
.68939 .40494 L
.67968 .39957 L
.67705 .39802 L
.66997 .39362 L
.6621 .38831 L
.66026 .38699 L
.65054 .37956 L
.64936 .37859 L
.64083 .37117 L
.63838 .36888 L
.63112 .36162 L
.62883 .35917 L
.62141 .35064 L
Mistroke
.62044 .34946 L
.61301 .33974 L
.61169 .3379 L
.60638 .33003 L
.60198 .32295 L
.60043 .32032 L
.59506 .31061 L
.59227 .30517 L
.59018 .30089 L
.58572 .29118 L
.58256 .28373 L
.58164 .28147 L
.57788 .27176 L
.5744 .26204 L
.57284 .25747 L
.57117 .25233 L
.56816 .24262 L
.56535 .23291 L
.56313 .22478 L
.56271 .22319 L
.56024 .21348 L
.5579 .20377 L
.55569 .19406 L
.55359 .18434 L
.55342 .18354 L
.5516 .17463 L
.5497 .16492 L
.54789 .15521 L
.54614 .14549 L
.54445 .13578 L
.54371 .13135 L
.54285 .12607 L
.5413 .11636 L
.53979 .10664 L
.53833 .09693 L
.53689 .08722 L
.53548 .07751 L
.5341 .06779 L
.53399 .06706 L
.5328 .05808 L
.5315 .04837 L
.53021 .03866 L
.52891 .02894 L
.5276 .01923 L
Mfstroke
0 .64 1 r
.98077 .5276 m
.97106 .52891 L
.96134 .53021 L
.95163 .5315 L
.94192 .5328 L
.93294 .53399 L
.93221 .5341 L
.92249 .53548 L
.91278 .53689 L
.90307 .53833 L
.89336 .53979 L
.88364 .5413 L
.87393 .54285 L
.86865 .54371 L
.86422 .54445 L
.85451 .54614 L
.84479 .54789 L
.83508 .5497 L
.82537 .5516 L
.81646 .55342 L
.81566 .55359 L
.80594 .55569 L
.79623 .5579 L
.78652 .56024 L
.77681 .56271 L
.77522 .56313 L
.76709 .56535 L
.75738 .56816 L
.74767 .57117 L
.74253 .57284 L
.73796 .5744 L
.72824 .57788 L
.71853 .58164 L
.71627 .58256 L
.70882 .58572 L
.69911 .59018 L
.69483 .59227 L
.68939 .59506 L
.67968 .60043 L
.67705 .60198 L
.66997 .60638 L
.6621 .61169 L
.66026 .61301 L
.65054 .62044 L
.64936 .62141 L
.64083 .62883 L
.63838 .63112 L
.63112 .63838 L
.62883 .64083 L
.62141 .64936 L
.62044 .65054 L
.61301 .66026 L
.61169 .6621 L
.60638 .66997 L
.60198 .67705 L
.60043 .67968 L
.59506 .68939 L
.59227 .69483 L
.59018 .69911 L
.58572 .70882 L
.58256 .71627 L
.58164 .71853 L
.57788 .72824 L
.5744 .73796 L
.57284 .74253 L
.57117 .74767 L
.56816 .75738 L
.56535 .76709 L
.56313 .77522 L
.56271 .77681 L
.56024 .78652 L
.5579 .79623 L
.55569 .80594 L
.55359 .81566 L
.55342 .81646 L
.5516 .82537 L
.5497 .83508 L
.54789 .84479 L
.54614 .85451 L
.54445 .86422 L
.54371 .86865 L
.54285 .87393 L
.5413 .88364 L
.53979 .89336 L
.53833 .90307 L
.53689 .91278 L
.53548 .92249 L
.5341 .93221 L
.53399 .93294 L
.5328 .94192 L
.5315 .95163 L
.53021 .96134 L
.52891 .97106 L
.5276 .98077 L
.98077 .98077 L
F
0 g
.98077 .5276 m
.97106 .52891 L
.96134 .53021 L
.95163 .5315 L
.94192 .5328 L
.93294 .53399 L
.93221 .5341 L
.92249 .53548 L
.91278 .53689 L
.90307 .53833 L
.89336 .53979 L
.88364 .5413 L
.87393 .54285 L
.86865 .54371 L
.86422 .54445 L
.85451 .54614 L
.84479 .54789 L
.83508 .5497 L
.82537 .5516 L
.81646 .55342 L
.81566 .55359 L
.80594 .55569 L
.79623 .5579 L
.78652 .56024 L
.77681 .56271 L
.77522 .56313 L
.76709 .56535 L
.75738 .56816 L
.74767 .57117 L
.74253 .57284 L
.73796 .5744 L
.72824 .57788 L
.71853 .58164 L
.71627 .58256 L
.70882 .58572 L
.69911 .59018 L
.69483 .59227 L
.68939 .59506 L
.67968 .60043 L
.67705 .60198 L
.66997 .60638 L
.6621 .61169 L
.66026 .61301 L
.65054 .62044 L
.64936 .62141 L
.64083 .62883 L
.63838 .63112 L
.63112 .63838 L
.62883 .64083 L
.62141 .64936 L
Mistroke
.62044 .65054 L
.61301 .66026 L
.61169 .6621 L
.60638 .66997 L
.60198 .67705 L
.60043 .67968 L
.59506 .68939 L
.59227 .69483 L
.59018 .69911 L
.58572 .70882 L
.58256 .71627 L
.58164 .71853 L
.57788 .72824 L
.5744 .73796 L
.57284 .74253 L
.57117 .74767 L
.56816 .75738 L
.56535 .76709 L
.56313 .77522 L
.56271 .77681 L
.56024 .78652 L
.5579 .79623 L
.55569 .80594 L
.55359 .81566 L
.55342 .81646 L
.5516 .82537 L
.5497 .83508 L
.54789 .84479 L
.54614 .85451 L
.54445 .86422 L
.54371 .86865 L
.54285 .87393 L
.5413 .88364 L
.53979 .89336 L
.53833 .90307 L
.53689 .91278 L
.53548 .92249 L
.5341 .93221 L
.53399 .93294 L
.5328 .94192 L
.5315 .95163 L
.53021 .96134 L
.52891 .97106 L
.5276 .98077 L
Mfstroke
0 .88 1 r
.98077 .46691 m
.97295 .46601 L
.97106 .46577 L
.96134 .46458 L
.95163 .46336 L
.94192 .46211 L
.93221 .46084 L
.92249 .45954 L
.91278 .45821 L
.90307 .45684 L
.8993 .45629 L
.89336 .45541 L
.88364 .45392 L
.87393 .45238 L
.86422 .45077 L
.85451 .4491 L
.84479 .44736 L
.8406 .44658 L
.83508 .44553 L
.82537 .4436 L
.81566 .44157 L
.80594 .43943 L
.79623 .43718 L
.79495 .43687 L
.78652 .43478 L
.77681 .43223 L
.76709 .42951 L
.75917 .42716 L
.75738 .4266 L
.74767 .42349 L
.73796 .42013 L
.73067 .41744 L
.72824 .41651 L
.71853 .41258 L
.70882 .40832 L
.70755 .40773 L
.69911 .40365 L
.68939 .39853 L
.68847 .39802 L
.67968 .39288 L
.67249 .38831 L
.66997 .38662 L
.66026 .37963 L
.65891 .37859 L
.65054 .37178 L
.64724 .36888 L
.64083 .36289 L
.63711 .35917 L
.63112 .35276 L
.62822 .34946 L
.62141 .34109 L
.62037 .33974 L
.61338 .33003 L
.61169 .32751 L
.60712 .32032 L
.60198 .31153 L
.60147 .31061 L
.59635 .30089 L
.59227 .29245 L
.59168 .29118 L
.58742 .28147 L
.58349 .27176 L
.58256 .26933 L
.57987 .26204 L
.57651 .25233 L
.5734 .24262 L
.57284 .24083 L
.57049 .23291 L
.56777 .22319 L
.56522 .21348 L
.56313 .20505 L
.56282 .20377 L
.56057 .19406 L
.55843 .18434 L
.5564 .17463 L
.55447 .16492 L
.55342 .1594 L
.55264 .15521 L
.5509 .14549 L
.54923 .13578 L
.54762 .12607 L
.54608 .11636 L
.54459 .10664 L
.54371 .1007 L
.54316 .09693 L
.54179 .08722 L
.54046 .07751 L
.53916 .06779 L
.53789 .05808 L
.53664 .04837 L
.53542 .03866 L
.53423 .02894 L
.53399 .02705 L
.53309 .01923 L
.98077 .01923 L
F
0 g
.98077 .46691 m
.97295 .46601 L
.97106 .46577 L
.96134 .46458 L
.95163 .46336 L
.94192 .46211 L
.93221 .46084 L
.92249 .45954 L
.91278 .45821 L
.90307 .45684 L
.8993 .45629 L
.89336 .45541 L
.88364 .45392 L
.87393 .45238 L
.86422 .45077 L
.85451 .4491 L
.84479 .44736 L
.8406 .44658 L
.83508 .44553 L
.82537 .4436 L
.81566 .44157 L
.80594 .43943 L
.79623 .43718 L
.79495 .43687 L
.78652 .43478 L
.77681 .43223 L
.76709 .42951 L
.75917 .42716 L
.75738 .4266 L
.74767 .42349 L
.73796 .42013 L
.73067 .41744 L
.72824 .41651 L
.71853 .41258 L
.70882 .40832 L
.70755 .40773 L
.69911 .40365 L
.68939 .39853 L
.68847 .39802 L
.67968 .39288 L
.67249 .38831 L
.66997 .38662 L
.66026 .37963 L
.65891 .37859 L
.65054 .37178 L
.64724 .36888 L
.64083 .36289 L
.63711 .35917 L
.63112 .35276 L
.62822 .34946 L
Mistroke
.62141 .34109 L
.62037 .33974 L
.61338 .33003 L
.61169 .32751 L
.60712 .32032 L
.60198 .31153 L
.60147 .31061 L
.59635 .30089 L
.59227 .29245 L
.59168 .29118 L
.58742 .28147 L
.58349 .27176 L
.58256 .26933 L
.57987 .26204 L
.57651 .25233 L
.5734 .24262 L
.57284 .24083 L
.57049 .23291 L
.56777 .22319 L
.56522 .21348 L
.56313 .20505 L
.56282 .20377 L
.56057 .19406 L
.55843 .18434 L
.5564 .17463 L
.55447 .16492 L
.55342 .1594 L
.55264 .15521 L
.5509 .14549 L
.54923 .13578 L
.54762 .12607 L
.54608 .11636 L
.54459 .10664 L
.54371 .1007 L
.54316 .09693 L
.54179 .08722 L
.54046 .07751 L
.53916 .06779 L
.53789 .05808 L
.53664 .04837 L
.53542 .03866 L
.53423 .02894 L
.53399 .02705 L
.53309 .01923 L
Mfstroke
0 .88 1 r
.98077 .53309 m
.97295 .53399 L
.97106 .53423 L
.96134 .53542 L
.95163 .53664 L
.94192 .53789 L
.93221 .53916 L
.92249 .54046 L
.91278 .54179 L
.90307 .54316 L
.8993 .54371 L
.89336 .54459 L
.88364 .54608 L
.87393 .54762 L
.86422 .54923 L
.85451 .5509 L
.84479 .55264 L
.8406 .55342 L
.83508 .55447 L
.82537 .5564 L
.81566 .55843 L
.80594 .56057 L
.79623 .56282 L
.79495 .56313 L
.78652 .56522 L
.77681 .56777 L
.76709 .57049 L
.75917 .57284 L
.75738 .5734 L
.74767 .57651 L
.73796 .57987 L
.73067 .58256 L
.72824 .58349 L
.71853 .58742 L
.70882 .59168 L
.70755 .59227 L
.69911 .59635 L
.68939 .60147 L
.68847 .60198 L
.67968 .60712 L
.67249 .61169 L
.66997 .61338 L
.66026 .62037 L
.65891 .62141 L
.65054 .62822 L
.64724 .63112 L
.64083 .63711 L
.63711 .64083 L
.63112 .64724 L
.62822 .65054 L
.62141 .65891 L
.62037 .66026 L
.61338 .66997 L
.61169 .67249 L
.60712 .67968 L
.60198 .68847 L
.60147 .68939 L
.59635 .69911 L
.59227 .70755 L
.59168 .70882 L
.58742 .71853 L
.58349 .72824 L
.58256 .73067 L
.57987 .73796 L
.57651 .74767 L
.5734 .75738 L
.57284 .75917 L
.57049 .76709 L
.56777 .77681 L
.56522 .78652 L
.56313 .79495 L
.56282 .79623 L
.56057 .80594 L
.55843 .81566 L
.5564 .82537 L
.55447 .83508 L
.55342 .8406 L
.55264 .84479 L
.5509 .85451 L
.54923 .86422 L
.54762 .87393 L
.54608 .88364 L
.54459 .89336 L
.54371 .8993 L
.54316 .90307 L
.54179 .91278 L
.54046 .92249 L
.53916 .93221 L
.53789 .94192 L
.53664 .95163 L
.53542 .96134 L
.53423 .97106 L
.53399 .97295 L
.53309 .98077 L
.98077 .98077 L
F
0 g
.98077 .53309 m
.97295 .53399 L
.97106 .53423 L
.96134 .53542 L
.95163 .53664 L
.94192 .53789 L
.93221 .53916 L
.92249 .54046 L
.91278 .54179 L
.90307 .54316 L
.8993 .54371 L
.89336 .54459 L
.88364 .54608 L
.87393 .54762 L
.86422 .54923 L
.85451 .5509 L
.84479 .55264 L
.8406 .55342 L
.83508 .55447 L
.82537 .5564 L
.81566 .55843 L
.80594 .56057 L
.79623 .56282 L
.79495 .56313 L
.78652 .56522 L
.77681 .56777 L
.76709 .57049 L
.75917 .57284 L
.75738 .5734 L
.74767 .57651 L
.73796 .57987 L
.73067 .58256 L
.72824 .58349 L
.71853 .58742 L
.70882 .59168 L
.70755 .59227 L
.69911 .59635 L
.68939 .60147 L
.68847 .60198 L
.67968 .60712 L
.67249 .61169 L
.66997 .61338 L
.66026 .62037 L
.65891 .62141 L
.65054 .62822 L
.64724 .63112 L
.64083 .63711 L
.63711 .64083 L
.63112 .64724 L
.62822 .65054 L
Mistroke
.62141 .65891 L
.62037 .66026 L
.61338 .66997 L
.61169 .67249 L
.60712 .67968 L
.60198 .68847 L
.60147 .68939 L
.59635 .69911 L
.59227 .70755 L
.59168 .70882 L
.58742 .71853 L
.58349 .72824 L
.58256 .73067 L
.57987 .73796 L
.57651 .74767 L
.5734 .75738 L
.57284 .75917 L
.57049 .76709 L
.56777 .77681 L
.56522 .78652 L
.56313 .79495 L
.56282 .79623 L
.56057 .80594 L
.55843 .81566 L
.5564 .82537 L
.55447 .83508 L
.55342 .8406 L
.55264 .84479 L
.5509 .85451 L
.54923 .86422 L
.54762 .87393 L
.54608 .88364 L
.54459 .89336 L
.54371 .8993 L
.54316 .90307 L
.54179 .91278 L
.54046 .92249 L
.53916 .93221 L
.53789 .94192 L
.53664 .95163 L
.53542 .96134 L
.53423 .97106 L
.53399 .97295 L
.53309 .98077 L
Mfstroke
0 1 .88 r
.98077 .46185 m
.97106 .46077 L
.96134 .45966 L
.95163 .45852 L
.94192 .45734 L
.93345 .45629 L
.93221 .45613 L
.92249 .45487 L
.91278 .45356 L
.90307 .4522 L
.89336 .45079 L
.88364 .44932 L
.87393 .44779 L
.86652 .44658 L
.86422 .44619 L
.85451 .44451 L
.84479 .44275 L
.83508 .44089 L
.82537 .43894 L
.81566 .43688 L
.81562 .43687 L
.80594 .43469 L
.79623 .43237 L
.78652 .42991 L
.77681 .42729 L
.77634 .42716 L
.76709 .42448 L
.75738 .42147 L
.74767 .41824 L
.74538 .41744 L
.73796 .41476 L
.72824 .41099 L
.72045 .40773 L
.71853 .4069 L
.70882 .40244 L
.69999 .39802 L
.69911 .39756 L
.68939 .39219 L
.68292 .38831 L
.67968 .38626 L
.66997 .37968 L
.66846 .37859 L
.66026 .37232 L
.65608 .36888 L
.65054 .36404 L
.64535 .35917 L
.64083 .35465 L
.63596 .34946 L
.63112 .34392 L
.62768 .33974 L
.62141 .33154 L
.62032 .33003 L
.61374 .32032 L
.61169 .31708 L
.60781 .31061 L
.60244 .30089 L
.60198 .30001 L
.59756 .29118 L
.5931 .28147 L
.59227 .27955 L
.58901 .27176 L
.58524 .26204 L
.58256 .25462 L
.58176 .25233 L
.57853 .24262 L
.57552 .23291 L
.57284 .22366 L
.57271 .22319 L
.57009 .21348 L
.56763 .20377 L
.56531 .19406 L
.56313 .18438 L
.56312 .18434 L
.56106 .17463 L
.55911 .16492 L
.55725 .15521 L
.55549 .14549 L
.55381 .13578 L
.55342 .13348 L
.55221 .12607 L
.55068 .11636 L
.54921 .10664 L
.5478 .09693 L
.54644 .08722 L
.54513 .07751 L
.54387 .06779 L
.54371 .06655 L
.54266 .05808 L
.54148 .04837 L
.54034 .03866 L
.53923 .02894 L
.53815 .01923 L
.98077 .01923 L
F
0 g
.98077 .46185 m
.97106 .46077 L
.96134 .45966 L
.95163 .45852 L
.94192 .45734 L
.93345 .45629 L
.93221 .45613 L
.92249 .45487 L
.91278 .45356 L
.90307 .4522 L
.89336 .45079 L
.88364 .44932 L
.87393 .44779 L
.86652 .44658 L
.86422 .44619 L
.85451 .44451 L
.84479 .44275 L
.83508 .44089 L
.82537 .43894 L
.81566 .43688 L
.81562 .43687 L
.80594 .43469 L
.79623 .43237 L
.78652 .42991 L
.77681 .42729 L
.77634 .42716 L
.76709 .42448 L
.75738 .42147 L
.74767 .41824 L
.74538 .41744 L
.73796 .41476 L
.72824 .41099 L
.72045 .40773 L
.71853 .4069 L
.70882 .40244 L
.69999 .39802 L
.69911 .39756 L
.68939 .39219 L
.68292 .38831 L
.67968 .38626 L
.66997 .37968 L
.66846 .37859 L
.66026 .37232 L
.65608 .36888 L
.65054 .36404 L
.64535 .35917 L
.64083 .35465 L
.63596 .34946 L
.63112 .34392 L
.62768 .33974 L
Mistroke
.62141 .33154 L
.62032 .33003 L
.61374 .32032 L
.61169 .31708 L
.60781 .31061 L
.60244 .30089 L
.60198 .30001 L
.59756 .29118 L
.5931 .28147 L
.59227 .27955 L
.58901 .27176 L
.58524 .26204 L
.58256 .25462 L
.58176 .25233 L
.57853 .24262 L
.57552 .23291 L
.57284 .22366 L
.57271 .22319 L
.57009 .21348 L
.56763 .20377 L
.56531 .19406 L
.56313 .18438 L
.56312 .18434 L
.56106 .17463 L
.55911 .16492 L
.55725 .15521 L
.55549 .14549 L
.55381 .13578 L
.55342 .13348 L
.55221 .12607 L
.55068 .11636 L
.54921 .10664 L
.5478 .09693 L
.54644 .08722 L
.54513 .07751 L
.54387 .06779 L
.54371 .06655 L
.54266 .05808 L
.54148 .04837 L
.54034 .03866 L
.53923 .02894 L
.53815 .01923 L
Mfstroke
0 1 .88 r
.98077 .53815 m
.97106 .53923 L
.96134 .54034 L
.95163 .54148 L
.94192 .54266 L
.93345 .54371 L
.93221 .54387 L
.92249 .54513 L
.91278 .54644 L
.90307 .5478 L
.89336 .54921 L
.88364 .55068 L
.87393 .55221 L
.86652 .55342 L
.86422 .55381 L
.85451 .55549 L
.84479 .55725 L
.83508 .55911 L
.82537 .56106 L
.81566 .56312 L
.81562 .56313 L
.80594 .56531 L
.79623 .56763 L
.78652 .57009 L
.77681 .57271 L
.77634 .57284 L
.76709 .57552 L
.75738 .57853 L
.74767 .58176 L
.74538 .58256 L
.73796 .58524 L
.72824 .58901 L
.72045 .59227 L
.71853 .5931 L
.70882 .59756 L
.69999 .60198 L
.69911 .60244 L
.68939 .60781 L
.68292 .61169 L
.67968 .61374 L
.66997 .62032 L
.66846 .62141 L
.66026 .62768 L
.65608 .63112 L
.65054 .63596 L
.64535 .64083 L
.64083 .64535 L
.63596 .65054 L
.63112 .65608 L
.62768 .66026 L
.62141 .66846 L
.62032 .66997 L
.61374 .67968 L
.61169 .68292 L
.60781 .68939 L
.60244 .69911 L
.60198 .69999 L
.59756 .70882 L
.5931 .71853 L
.59227 .72045 L
.58901 .72824 L
.58524 .73796 L
.58256 .74538 L
.58176 .74767 L
.57853 .75738 L
.57552 .76709 L
.57284 .77634 L
.57271 .77681 L
.57009 .78652 L
.56763 .79623 L
.56531 .80594 L
.56313 .81562 L
.56312 .81566 L
.56106 .82537 L
.55911 .83508 L
.55725 .84479 L
.55549 .85451 L
.55381 .86422 L
.55342 .86652 L
.55221 .87393 L
.55068 .88364 L
.54921 .89336 L
.5478 .90307 L
.54644 .91278 L
.54513 .92249 L
.54387 .93221 L
.54371 .93345 L
.54266 .94192 L
.54148 .95163 L
.54034 .96134 L
.53923 .97106 L
.53815 .98077 L
.98077 .98077 L
F
0 g
.98077 .53815 m
.97106 .53923 L
.96134 .54034 L
.95163 .54148 L
.94192 .54266 L
.93345 .54371 L
.93221 .54387 L
.92249 .54513 L
.91278 .54644 L
.90307 .5478 L
.89336 .54921 L
.88364 .55068 L
.87393 .55221 L
.86652 .55342 L
.86422 .55381 L
.85451 .55549 L
.84479 .55725 L
.83508 .55911 L
.82537 .56106 L
.81566 .56312 L
.81562 .56313 L
.80594 .56531 L
.79623 .56763 L
.78652 .57009 L
.77681 .57271 L
.77634 .57284 L
.76709 .57552 L
.75738 .57853 L
.74767 .58176 L
.74538 .58256 L
.73796 .58524 L
.72824 .58901 L
.72045 .59227 L
.71853 .5931 L
.70882 .59756 L
.69999 .60198 L
.69911 .60244 L
.68939 .60781 L
.68292 .61169 L
.67968 .61374 L
.66997 .62032 L
.66846 .62141 L
.66026 .62768 L
.65608 .63112 L
.65054 .63596 L
.64535 .64083 L
.64083 .64535 L
.63596 .65054 L
.63112 .65608 L
.62768 .66026 L
Mistroke
.62141 .66846 L
.62032 .66997 L
.61374 .67968 L
.61169 .68292 L
.60781 .68939 L
.60244 .69911 L
.60198 .69999 L
.59756 .70882 L
.5931 .71853 L
.59227 .72045 L
.58901 .72824 L
.58524 .73796 L
.58256 .74538 L
.58176 .74767 L
.57853 .75738 L
.57552 .76709 L
.57284 .77634 L
.57271 .77681 L
.57009 .78652 L
.56763 .79623 L
.56531 .80594 L
.56313 .81562 L
.56312 .81566 L
.56106 .82537 L
.55911 .83508 L
.55725 .84479 L
.55549 .85451 L
.55381 .86422 L
.55342 .86652 L
.55221 .87393 L
.55068 .88364 L
.54921 .89336 L
.5478 .90307 L
.54644 .91278 L
.54513 .92249 L
.54387 .93221 L
.54371 .93345 L
.54266 .94192 L
.54148 .95163 L
.54034 .96134 L
.53923 .97106 L
.53815 .98077 L
Mfstroke
0 1 .64 r
.98077 .45705 m
.97352 .45629 L
.97106 .45603 L
.96134 .45497 L
.95163 .45388 L
.94192 .45274 L
.93221 .45155 L
.92249 .45032 L
.91278 .44904 L
.90307 .44771 L
.8952 .44658 L
.89336 .44631 L
.88364 .44485 L
.87393 .44331 L
.86422 .4417 L
.85451 .44001 L
.84479 .43823 L
.83773 .43687 L
.83508 .43635 L
.82537 .43436 L
.81566 .43225 L
.80594 .43002 L
.79623 .42764 L
.79433 .42716 L
.78652 .42511 L
.77681 .4224 L
.76709 .4195 L
.76058 .41744 L
.75738 .41639 L
.74767 .41305 L
.73796 .40943 L
.73365 .40773 L
.72824 .40551 L
.71853 .40125 L
.7117 .39802 L
.70882 .39659 L
.69911 .3915 L
.69347 .38831 L
.68939 .38588 L
.67968 .37967 L
.6781 .37859 L
.66997 .37276 L
.66497 .36888 L
.66026 .36502 L
.65361 .35917 L
.65054 .35629 L
.64371 .34946 L
.64083 .34639 L
.63498 .33974 L
.63112 .33503 L
.62724 .33003 L
.62141 .3219 L
.62033 .32032 L
.61412 .31061 L
.61169 .30653 L
.6085 .30089 L
.60341 .29118 L
.60198 .2883 L
.59875 .28147 L
.59449 .27176 L
.59227 .26635 L
.59057 .26204 L
.58695 .25233 L
.58361 .24262 L
.58256 .23942 L
.5805 .23291 L
.5776 .22319 L
.57489 .21348 L
.57284 .20567 L
.57236 .20377 L
.56998 .19406 L
.56775 .18434 L
.56564 .17463 L
.56365 .16492 L
.56313 .16227 L
.56177 .15521 L
.55999 .14549 L
.5583 .13578 L
.55669 .12607 L
.55515 .11636 L
.55369 .10664 L
.55342 .1048 L
.55229 .09693 L
.55096 .08722 L
.54968 .07751 L
.54845 .06779 L
.54726 .05808 L
.54612 .04837 L
.54503 .03866 L
.54397 .02894 L
.54371 .02648 L
.54295 .01923 L
.98077 .01923 L
F
0 g
.98077 .45705 m
.97352 .45629 L
.97106 .45603 L
.96134 .45497 L
.95163 .45388 L
.94192 .45274 L
.93221 .45155 L
.92249 .45032 L
.91278 .44904 L
.90307 .44771 L
.8952 .44658 L
.89336 .44631 L
.88364 .44485 L
.87393 .44331 L
.86422 .4417 L
.85451 .44001 L
.84479 .43823 L
.83773 .43687 L
.83508 .43635 L
.82537 .43436 L
.81566 .43225 L
.80594 .43002 L
.79623 .42764 L
.79433 .42716 L
.78652 .42511 L
.77681 .4224 L
.76709 .4195 L
.76058 .41744 L
.75738 .41639 L
.74767 .41305 L
.73796 .40943 L
.73365 .40773 L
.72824 .40551 L
.71853 .40125 L
.7117 .39802 L
.70882 .39659 L
.69911 .3915 L
.69347 .38831 L
.68939 .38588 L
.67968 .37967 L
.6781 .37859 L
.66997 .37276 L
.66497 .36888 L
.66026 .36502 L
.65361 .35917 L
.65054 .35629 L
.64371 .34946 L
.64083 .34639 L
.63498 .33974 L
.63112 .33503 L
Mistroke
.62724 .33003 L
.62141 .3219 L
.62033 .32032 L
.61412 .31061 L
.61169 .30653 L
.6085 .30089 L
.60341 .29118 L
.60198 .2883 L
.59875 .28147 L
.59449 .27176 L
.59227 .26635 L
.59057 .26204 L
.58695 .25233 L
.58361 .24262 L
.58256 .23942 L
.5805 .23291 L
.5776 .22319 L
.57489 .21348 L
.57284 .20567 L
.57236 .20377 L
.56998 .19406 L
.56775 .18434 L
.56564 .17463 L
.56365 .16492 L
.56313 .16227 L
.56177 .15521 L
.55999 .14549 L
.5583 .13578 L
.55669 .12607 L
.55515 .11636 L
.55369 .10664 L
.55342 .1048 L
.55229 .09693 L
.55096 .08722 L
.54968 .07751 L
.54845 .06779 L
.54726 .05808 L
.54612 .04837 L
.54503 .03866 L
.54397 .02894 L
.54371 .02648 L
.54295 .01923 L
Mfstroke
0 1 .64 r
.98077 .54295 m
.97352 .54371 L
.97106 .54397 L
.96134 .54503 L
.95163 .54612 L
.94192 .54726 L
.93221 .54845 L
.92249 .54968 L
.91278 .55096 L
.90307 .55229 L
.8952 .55342 L
.89336 .55369 L
.88364 .55515 L
.87393 .55669 L
.86422 .5583 L
.85451 .55999 L
.84479 .56177 L
.83773 .56313 L
.83508 .56365 L
.82537 .56564 L
.81566 .56775 L
.80594 .56998 L
.79623 .57236 L
.79433 .57284 L
.78652 .57489 L
.77681 .5776 L
.76709 .5805 L
.76058 .58256 L
.75738 .58361 L
.74767 .58695 L
.73796 .59057 L
.73365 .59227 L
.72824 .59449 L
.71853 .59875 L
.7117 .60198 L
.70882 .60341 L
.69911 .6085 L
.69347 .61169 L
.68939 .61412 L
.67968 .62033 L
.6781 .62141 L
.66997 .62724 L
.66497 .63112 L
.66026 .63498 L
.65361 .64083 L
.65054 .64371 L
.64371 .65054 L
.64083 .65361 L
.63498 .66026 L
.63112 .66497 L
.62724 .66997 L
.62141 .6781 L
.62033 .67968 L
.61412 .68939 L
.61169 .69347 L
.6085 .69911 L
.60341 .70882 L
.60198 .7117 L
.59875 .71853 L
.59449 .72824 L
.59227 .73365 L
.59057 .73796 L
.58695 .74767 L
.58361 .75738 L
.58256 .76058 L
.5805 .76709 L
.5776 .77681 L
.57489 .78652 L
.57284 .79433 L
.57236 .79623 L
.56998 .80594 L
.56775 .81566 L
.56564 .82537 L
.56365 .83508 L
.56313 .83773 L
.56177 .84479 L
.55999 .85451 L
.5583 .86422 L
.55669 .87393 L
.55515 .88364 L
.55369 .89336 L
.55342 .8952 L
.55229 .90307 L
.55096 .91278 L
.54968 .92249 L
.54845 .93221 L
.54726 .94192 L
.54612 .95163 L
.54503 .96134 L
.54397 .97106 L
.54371 .97352 L
.54295 .98077 L
.98077 .98077 L
F
0 g
.98077 .54295 m
.97352 .54371 L
.97106 .54397 L
.96134 .54503 L
.95163 .54612 L
.94192 .54726 L
.93221 .54845 L
.92249 .54968 L
.91278 .55096 L
.90307 .55229 L
.8952 .55342 L
.89336 .55369 L
.88364 .55515 L
.87393 .55669 L
.86422 .5583 L
.85451 .55999 L
.84479 .56177 L
.83773 .56313 L
.83508 .56365 L
.82537 .56564 L
.81566 .56775 L
.80594 .56998 L
.79623 .57236 L
.79433 .57284 L
.78652 .57489 L
.77681 .5776 L
.76709 .5805 L
.76058 .58256 L
.75738 .58361 L
.74767 .58695 L
.73796 .59057 L
.73365 .59227 L
.72824 .59449 L
.71853 .59875 L
.7117 .60198 L
.70882 .60341 L
.69911 .6085 L
.69347 .61169 L
.68939 .61412 L
.67968 .62033 L
.6781 .62141 L
.66997 .62724 L
.66497 .63112 L
.66026 .63498 L
.65361 .64083 L
.65054 .64371 L
.64371 .65054 L
.64083 .65361 L
.63498 .66026 L
.63112 .66497 L
Mistroke
.62724 .66997 L
.62141 .6781 L
.62033 .67968 L
.61412 .68939 L
.61169 .69347 L
.6085 .69911 L
.60341 .70882 L
.60198 .7117 L
.59875 .71853 L
.59449 .72824 L
.59227 .73365 L
.59057 .73796 L
.58695 .74767 L
.58361 .75738 L
.58256 .76058 L
.5805 .76709 L
.5776 .77681 L
.57489 .78652 L
.57284 .79433 L
.57236 .79623 L
.56998 .80594 L
.56775 .81566 L
.56564 .82537 L
.56365 .83508 L
.56313 .83773 L
.56177 .84479 L
.55999 .85451 L
.5583 .86422 L
.55669 .87393 L
.55515 .88364 L
.55369 .89336 L
.55342 .8952 L
.55229 .90307 L
.55096 .91278 L
.54968 .92249 L
.54845 .93221 L
.54726 .94192 L
.54612 .95163 L
.54503 .96134 L
.54397 .97106 L
.54371 .97352 L
.54295 .98077 L
Mfstroke
0 1 .4 r
.98077 .45233 m
.97106 .45136 L
.96134 .45035 L
.95163 .44929 L
.94192 .44819 L
.93221 .44704 L
.92842 .44658 L
.92249 .44584 L
.91278 .44458 L
.90307 .44326 L
.89336 .44187 L
.88364 .44041 L
.87393 .43887 L
.86422 .43725 L
.86201 .43687 L
.85451 .43554 L
.84479 .43374 L
.83508 .43183 L
.82537 .4298 L
.81566 .42765 L
.81351 .42716 L
.80594 .42536 L
.79623 .42292 L
.78652 .42032 L
.77681 .41753 L
.77651 .41744 L
.76709 .41454 L
.75738 .41133 L
.74767 .40785 L
.74734 .40773 L
.73796 .4041 L
.72824 .40003 L
.72374 .39802 L
.71853 .39559 L
.70882 .39074 L
.70426 .38831 L
.69911 .38542 L
.68939 .37955 L
.68791 .37859 L
.67968 .37304 L
.67398 .36888 L
.66997 .36579 L
.66198 .35917 L
.66026 .35766 L
.65152 .34946 L
.65054 .34848 L
.64234 .33974 L
.64083 .33802 L
.63421 .33003 L
.63112 .32602 L
.62696 .32032 L
.62141 .31209 L
.62045 .31061 L
.61458 .30089 L
.61169 .29574 L
.60926 .29118 L
.60441 .28147 L
.60198 .27626 L
.59997 .27176 L
.5959 .26204 L
.59227 .25266 L
.59215 .25233 L
.58867 .24262 L
.58546 .23291 L
.58256 .22349 L
.58247 .22319 L
.57968 .21348 L
.57708 .20377 L
.57464 .19406 L
.57284 .18649 L
.57235 .18434 L
.5702 .17463 L
.56817 .16492 L
.56626 .15521 L
.56446 .14549 L
.56313 .13799 L
.56275 .13578 L
.56113 .12607 L
.55959 .11636 L
.55813 .10664 L
.55674 .09693 L
.55542 .08722 L
.55416 .07751 L
.55342 .07158 L
.55296 .06779 L
.55181 .05808 L
.55071 .04837 L
.54965 .03866 L
.54864 .02894 L
.54767 .01923 L
.98077 .01923 L
F
0 g
.98077 .45233 m
.97106 .45136 L
.96134 .45035 L
.95163 .44929 L
.94192 .44819 L
.93221 .44704 L
.92842 .44658 L
.92249 .44584 L
.91278 .44458 L
.90307 .44326 L
.89336 .44187 L
.88364 .44041 L
.87393 .43887 L
.86422 .43725 L
.86201 .43687 L
.85451 .43554 L
.84479 .43374 L
.83508 .43183 L
.82537 .4298 L
.81566 .42765 L
.81351 .42716 L
.80594 .42536 L
.79623 .42292 L
.78652 .42032 L
.77681 .41753 L
.77651 .41744 L
.76709 .41454 L
.75738 .41133 L
.74767 .40785 L
.74734 .40773 L
.73796 .4041 L
.72824 .40003 L
.72374 .39802 L
.71853 .39559 L
.70882 .39074 L
.70426 .38831 L
.69911 .38542 L
.68939 .37955 L
.68791 .37859 L
.67968 .37304 L
.67398 .36888 L
.66997 .36579 L
.66198 .35917 L
.66026 .35766 L
.65152 .34946 L
.65054 .34848 L
.64234 .33974 L
.64083 .33802 L
.63421 .33003 L
.63112 .32602 L
Mistroke
.62696 .32032 L
.62141 .31209 L
.62045 .31061 L
.61458 .30089 L
.61169 .29574 L
.60926 .29118 L
.60441 .28147 L
.60198 .27626 L
.59997 .27176 L
.5959 .26204 L
.59227 .25266 L
.59215 .25233 L
.58867 .24262 L
.58546 .23291 L
.58256 .22349 L
.58247 .22319 L
.57968 .21348 L
.57708 .20377 L
.57464 .19406 L
.57284 .18649 L
.57235 .18434 L
.5702 .17463 L
.56817 .16492 L
.56626 .15521 L
.56446 .14549 L
.56313 .13799 L
.56275 .13578 L
.56113 .12607 L
.55959 .11636 L
.55813 .10664 L
.55674 .09693 L
.55542 .08722 L
.55416 .07751 L
.55342 .07158 L
.55296 .06779 L
.55181 .05808 L
.55071 .04837 L
.54965 .03866 L
.54864 .02894 L
.54767 .01923 L
Mfstroke
0 1 .4 r
.98077 .54767 m
.97106 .54864 L
.96134 .54965 L
.95163 .55071 L
.94192 .55181 L
.93221 .55296 L
.92842 .55342 L
.92249 .55416 L
.91278 .55542 L
.90307 .55674 L
.89336 .55813 L
.88364 .55959 L
.87393 .56113 L
.86422 .56275 L
.86201 .56313 L
.85451 .56446 L
.84479 .56626 L
.83508 .56817 L
.82537 .5702 L
.81566 .57235 L
.81351 .57284 L
.80594 .57464 L
.79623 .57708 L
.78652 .57968 L
.77681 .58247 L
.77651 .58256 L
.76709 .58546 L
.75738 .58867 L
.74767 .59215 L
.74734 .59227 L
.73796 .5959 L
.72824 .59997 L
.72374 .60198 L
.71853 .60441 L
.70882 .60926 L
.70426 .61169 L
.69911 .61458 L
.68939 .62045 L
.68791 .62141 L
.67968 .62696 L
.67398 .63112 L
.66997 .63421 L
.66198 .64083 L
.66026 .64234 L
.65152 .65054 L
.65054 .65152 L
.64234 .66026 L
.64083 .66198 L
.63421 .66997 L
.63112 .67398 L
.62696 .67968 L
.62141 .68791 L
.62045 .68939 L
.61458 .69911 L
.61169 .70426 L
.60926 .70882 L
.60441 .71853 L
.60198 .72374 L
.59997 .72824 L
.5959 .73796 L
.59227 .74734 L
.59215 .74767 L
.58867 .75738 L
.58546 .76709 L
.58256 .77651 L
.58247 .77681 L
.57968 .78652 L
.57708 .79623 L
.57464 .80594 L
.57284 .81351 L
.57235 .81566 L
.5702 .82537 L
.56817 .83508 L
.56626 .84479 L
.56446 .85451 L
.56313 .86201 L
.56275 .86422 L
.56113 .87393 L
.55959 .88364 L
.55813 .89336 L
.55674 .90307 L
.55542 .91278 L
.55416 .92249 L
.55342 .92842 L
.55296 .93221 L
.55181 .94192 L
.55071 .95163 L
.54965 .96134 L
.54864 .97106 L
.54767 .98077 L
.98077 .98077 L
F
0 g
.98077 .54767 m
.97106 .54864 L
.96134 .54965 L
.95163 .55071 L
.94192 .55181 L
.93221 .55296 L
.92842 .55342 L
.92249 .55416 L
.91278 .55542 L
.90307 .55674 L
.89336 .55813 L
.88364 .55959 L
.87393 .56113 L
.86422 .56275 L
.86201 .56313 L
.85451 .56446 L
.84479 .56626 L
.83508 .56817 L
.82537 .5702 L
.81566 .57235 L
.81351 .57284 L
.80594 .57464 L
.79623 .57708 L
.78652 .57968 L
.77681 .58247 L
.77651 .58256 L
.76709 .58546 L
.75738 .58867 L
.74767 .59215 L
.74734 .59227 L
.73796 .5959 L
.72824 .59997 L
.72374 .60198 L
.71853 .60441 L
.70882 .60926 L
.70426 .61169 L
.69911 .61458 L
.68939 .62045 L
.68791 .62141 L
.67968 .62696 L
.67398 .63112 L
.66997 .63421 L
.66198 .64083 L
.66026 .64234 L
.65152 .65054 L
.65054 .65152 L
.64234 .66026 L
.64083 .66198 L
.63421 .66997 L
.63112 .67398 L
Mistroke
.62696 .67968 L
.62141 .68791 L
.62045 .68939 L
.61458 .69911 L
.61169 .70426 L
.60926 .70882 L
.60441 .71853 L
.60198 .72374 L
.59997 .72824 L
.5959 .73796 L
.59227 .74734 L
.59215 .74767 L
.58867 .75738 L
.58546 .76709 L
.58256 .77651 L
.58247 .77681 L
.57968 .78652 L
.57708 .79623 L
.57464 .80594 L
.57284 .81351 L
.57235 .81566 L
.5702 .82537 L
.56817 .83508 L
.56626 .84479 L
.56446 .85451 L
.56313 .86201 L
.56275 .86422 L
.56113 .87393 L
.55959 .88364 L
.55813 .89336 L
.55674 .90307 L
.55542 .91278 L
.55416 .92249 L
.55342 .92842 L
.55296 .93221 L
.55181 .94192 L
.55071 .95163 L
.54965 .96134 L
.54864 .97106 L
.54767 .98077 L
Mfstroke
0 1 .16 r
.98077 .44757 m
.97106 .44666 L
.97025 .44658 L
.96134 .44571 L
.95163 .4447 L
.94192 .44364 L
.93221 .44252 L
.92249 .44134 L
.91278 .44009 L
.90307 .43878 L
.89336 .4374 L
.88977 .43687 L
.88364 .43594 L
.87393 .43441 L
.86422 .43278 L
.85451 .43105 L
.84479 .42922 L
.83508 .42728 L
.83449 .42716 L
.82537 .42522 L
.81566 .42302 L
.80594 .42068 L
.79623 .41818 L
.7935 .41744 L
.78652 .4155 L
.77681 .41263 L
.76709 .40954 L
.76171 .40773 L
.75738 .40622 L
.74767 .40263 L
.73796 .39873 L
.73626 .39802 L
.72824 .3945 L
.71853 .38989 L
.7154 .38831 L
.70882 .38483 L
.69911 .37928 L
.69798 .37859 L
.68939 .37314 L
.6832 .36888 L
.67968 .36633 L
.6705 .35917 L
.66997 .35873 L
.66026 .35019 L
.65948 .34946 L
.65054 .34052 L
.64981 .33974 L
.64127 .33003 L
.64083 .3295 L
.63367 .32032 L
.63112 .3168 L
.62686 .31061 L
.62141 .30202 L
.62072 .30089 L
.61517 .29118 L
.61169 .2846 L
.61011 .28147 L
.6055 .27176 L
.60198 .26374 L
.60127 .26204 L
.59737 .25233 L
.59378 .24262 L
.59227 .23829 L
.59046 .23291 L
.58737 .22319 L
.5845 .21348 L
.58256 .2065 L
.58182 .20377 L
.57932 .19406 L
.57698 .18434 L
.57478 .17463 L
.57284 .16551 L
.57272 .16492 L
.57078 .15521 L
.56895 .14549 L
.56722 .13578 L
.56559 .12607 L
.56406 .11636 L
.56313 .11023 L
.5626 .10664 L
.56122 .09693 L
.55991 .08722 L
.55866 .07751 L
.55748 .06779 L
.55636 .05808 L
.5553 .04837 L
.55429 .03866 L
.55342 .02975 L
.55334 .02894 L
.55243 .01923 L
.98077 .01923 L
F
0 g
.98077 .44757 m
.97106 .44666 L
.97025 .44658 L
.96134 .44571 L
.95163 .4447 L
.94192 .44364 L
.93221 .44252 L
.92249 .44134 L
.91278 .44009 L
.90307 .43878 L
.89336 .4374 L
.88977 .43687 L
.88364 .43594 L
.87393 .43441 L
.86422 .43278 L
.85451 .43105 L
.84479 .42922 L
.83508 .42728 L
.83449 .42716 L
.82537 .42522 L
.81566 .42302 L
.80594 .42068 L
.79623 .41818 L
.7935 .41744 L
.78652 .4155 L
.77681 .41263 L
.76709 .40954 L
.76171 .40773 L
.75738 .40622 L
.74767 .40263 L
.73796 .39873 L
.73626 .39802 L
.72824 .3945 L
.71853 .38989 L
.7154 .38831 L
.70882 .38483 L
.69911 .37928 L
.69798 .37859 L
.68939 .37314 L
.6832 .36888 L
.67968 .36633 L
.6705 .35917 L
.66997 .35873 L
.66026 .35019 L
.65948 .34946 L
.65054 .34052 L
.64981 .33974 L
.64127 .33003 L
.64083 .3295 L
.63367 .32032 L
Mistroke
.63112 .3168 L
.62686 .31061 L
.62141 .30202 L
.62072 .30089 L
.61517 .29118 L
.61169 .2846 L
.61011 .28147 L
.6055 .27176 L
.60198 .26374 L
.60127 .26204 L
.59737 .25233 L
.59378 .24262 L
.59227 .23829 L
.59046 .23291 L
.58737 .22319 L
.5845 .21348 L
.58256 .2065 L
.58182 .20377 L
.57932 .19406 L
.57698 .18434 L
.57478 .17463 L
.57284 .16551 L
.57272 .16492 L
.57078 .15521 L
.56895 .14549 L
.56722 .13578 L
.56559 .12607 L
.56406 .11636 L
.56313 .11023 L
.5626 .10664 L
.56122 .09693 L
.55991 .08722 L
.55866 .07751 L
.55748 .06779 L
.55636 .05808 L
.5553 .04837 L
.55429 .03866 L
.55342 .02975 L
.55334 .02894 L
.55243 .01923 L
Mfstroke
0 1 .16 r
.98077 .55243 m
.97106 .55334 L
.97025 .55342 L
.96134 .55429 L
.95163 .5553 L
.94192 .55636 L
.93221 .55748 L
.92249 .55866 L
.91278 .55991 L
.90307 .56122 L
.89336 .5626 L
.88977 .56313 L
.88364 .56406 L
.87393 .56559 L
.86422 .56722 L
.85451 .56895 L
.84479 .57078 L
.83508 .57272 L
.83449 .57284 L
.82537 .57478 L
.81566 .57698 L
.80594 .57932 L
.79623 .58182 L
.7935 .58256 L
.78652 .5845 L
.77681 .58737 L
.76709 .59046 L
.76171 .59227 L
.75738 .59378 L
.74767 .59737 L
.73796 .60127 L
.73626 .60198 L
.72824 .6055 L
.71853 .61011 L
.7154 .61169 L
.70882 .61517 L
.69911 .62072 L
.69798 .62141 L
.68939 .62686 L
.6832 .63112 L
.67968 .63367 L
.6705 .64083 L
.66997 .64127 L
.66026 .64981 L
.65948 .65054 L
.65054 .65948 L
.64981 .66026 L
.64127 .66997 L
.64083 .6705 L
.63367 .67968 L
.63112 .6832 L
.62686 .68939 L
.62141 .69798 L
.62072 .69911 L
.61517 .70882 L
.61169 .7154 L
.61011 .71853 L
.6055 .72824 L
.60198 .73626 L
.60127 .73796 L
.59737 .74767 L
.59378 .75738 L
.59227 .76171 L
.59046 .76709 L
.58737 .77681 L
.5845 .78652 L
.58256 .7935 L
.58182 .79623 L
.57932 .80594 L
.57698 .81566 L
.57478 .82537 L
.57284 .83449 L
.57272 .83508 L
.57078 .84479 L
.56895 .85451 L
.56722 .86422 L
.56559 .87393 L
.56406 .88364 L
.56313 .88977 L
.5626 .89336 L
.56122 .90307 L
.55991 .91278 L
.55866 .92249 L
.55748 .93221 L
.55636 .94192 L
.5553 .95163 L
.55429 .96134 L
.55342 .97025 L
.55334 .97106 L
.55243 .98077 L
.98077 .98077 L
F
0 g
.98077 .55243 m
.97106 .55334 L
.97025 .55342 L
.96134 .55429 L
.95163 .5553 L
.94192 .55636 L
.93221 .55748 L
.92249 .55866 L
.91278 .55991 L
.90307 .56122 L
.89336 .5626 L
.88977 .56313 L
.88364 .56406 L
.87393 .56559 L
.86422 .56722 L
.85451 .56895 L
.84479 .57078 L
.83508 .57272 L
.83449 .57284 L
.82537 .57478 L
.81566 .57698 L
.80594 .57932 L
.79623 .58182 L
.7935 .58256 L
.78652 .5845 L
.77681 .58737 L
.76709 .59046 L
.76171 .59227 L
.75738 .59378 L
.74767 .59737 L
.73796 .60127 L
.73626 .60198 L
.72824 .6055 L
.71853 .61011 L
.7154 .61169 L
.70882 .61517 L
.69911 .62072 L
.69798 .62141 L
.68939 .62686 L
.6832 .63112 L
.67968 .63367 L
.6705 .64083 L
.66997 .64127 L
.66026 .64981 L
.65948 .65054 L
.65054 .65948 L
.64981 .66026 L
.64127 .66997 L
.64083 .6705 L
.63367 .67968 L
Mistroke
.63112 .6832 L
.62686 .68939 L
.62141 .69798 L
.62072 .69911 L
.61517 .70882 L
.61169 .7154 L
.61011 .71853 L
.6055 .72824 L
.60198 .73626 L
.60127 .73796 L
.59737 .74767 L
.59378 .75738 L
.59227 .76171 L
.59046 .76709 L
.58737 .77681 L
.5845 .78652 L
.58256 .7935 L
.58182 .79623 L
.57932 .80594 L
.57698 .81566 L
.57478 .82537 L
.57284 .83449 L
.57272 .83508 L
.57078 .84479 L
.56895 .85451 L
.56722 .86422 L
.56559 .87393 L
.56406 .88364 L
.56313 .88977 L
.5626 .89336 L
.56122 .90307 L
.55991 .91278 L
.55866 .92249 L
.55748 .93221 L
.55636 .94192 L
.5553 .95163 L
.55429 .96134 L
.55342 .97025 L
.55334 .97106 L
.55243 .98077 L
Mfstroke
.08 1 0 r
.98077 .4427 m
.97106 .44185 L
.96134 .44094 L
.95163 .43997 L
.94192 .43895 L
.93221 .43786 L
.92382 .43687 L
.92249 .43671 L
.91278 .4355 L
.90307 .43422 L
.89336 .43285 L
.88364 .4314 L
.87393 .42985 L
.86422 .42821 L
.85827 .42716 L
.85451 .42647 L
.84479 .42463 L
.83508 .42266 L
.82537 .42056 L
.81566 .41831 L
.81205 .41744 L
.80594 .41592 L
.79623 .41336 L
.78652 .41061 L
.77707 .40773 L
.77681 .40765 L
.76709 .40447 L
.75738 .40103 L
.74946 .39802 L
.74767 .39731 L
.73796 .39328 L
.72824 .38888 L
.72704 .38831 L
.71853 .38408 L
.70882 .37882 L
.70843 .37859 L
.69911 .37302 L
.69272 .36888 L
.68939 .36661 L
.67968 .35948 L
.67928 .35917 L
.66997 .35151 L
.66764 .34946 L
.66026 .34253 L
.65747 .33974 L
.65054 .33236 L
.64849 .33003 L
.64083 .32072 L
.64052 .32032 L
.63339 .31061 L
.63112 .30728 L
.62698 .30089 L
.62141 .29157 L
.62118 .29118 L
.61592 .28147 L
.61169 .27296 L
.61112 .27176 L
.60672 .26204 L
.60269 .25233 L
.60198 .25054 L
.59897 .24262 L
.59553 .23291 L
.59235 .22319 L
.59227 .22293 L
.58939 .21348 L
.58664 .20377 L
.58408 .19406 L
.58256 .18795 L
.58169 .18434 L
.57944 .17463 L
.57734 .16492 L
.57537 .15521 L
.57353 .14549 L
.57284 .14173 L
.57179 .13578 L
.57015 .12607 L
.5686 .11636 L
.56715 .10664 L
.56578 .09693 L
.5645 .08722 L
.56329 .07751 L
.56313 .07618 L
.56214 .06779 L
.56105 .05808 L
.56003 .04837 L
.55906 .03866 L
.55815 .02894 L
.5573 .01923 L
.98077 .01923 L
F
0 g
.98077 .4427 m
.97106 .44185 L
.96134 .44094 L
.95163 .43997 L
.94192 .43895 L
.93221 .43786 L
.92382 .43687 L
.92249 .43671 L
.91278 .4355 L
.90307 .43422 L
.89336 .43285 L
.88364 .4314 L
.87393 .42985 L
.86422 .42821 L
.85827 .42716 L
.85451 .42647 L
.84479 .42463 L
.83508 .42266 L
.82537 .42056 L
.81566 .41831 L
.81205 .41744 L
.80594 .41592 L
.79623 .41336 L
.78652 .41061 L
.77707 .40773 L
.77681 .40765 L
.76709 .40447 L
.75738 .40103 L
.74946 .39802 L
.74767 .39731 L
.73796 .39328 L
.72824 .38888 L
.72704 .38831 L
.71853 .38408 L
.70882 .37882 L
.70843 .37859 L
.69911 .37302 L
.69272 .36888 L
.68939 .36661 L
.67968 .35948 L
.67928 .35917 L
.66997 .35151 L
.66764 .34946 L
.66026 .34253 L
.65747 .33974 L
.65054 .33236 L
.64849 .33003 L
.64083 .32072 L
.64052 .32032 L
.63339 .31061 L
Mistroke
.63112 .30728 L
.62698 .30089 L
.62141 .29157 L
.62118 .29118 L
.61592 .28147 L
.61169 .27296 L
.61112 .27176 L
.60672 .26204 L
.60269 .25233 L
.60198 .25054 L
.59897 .24262 L
.59553 .23291 L
.59235 .22319 L
.59227 .22293 L
.58939 .21348 L
.58664 .20377 L
.58408 .19406 L
.58256 .18795 L
.58169 .18434 L
.57944 .17463 L
.57734 .16492 L
.57537 .15521 L
.57353 .14549 L
.57284 .14173 L
.57179 .13578 L
.57015 .12607 L
.5686 .11636 L
.56715 .10664 L
.56578 .09693 L
.5645 .08722 L
.56329 .07751 L
.56313 .07618 L
.56214 .06779 L
.56105 .05808 L
.56003 .04837 L
.55906 .03866 L
.55815 .02894 L
.5573 .01923 L
Mfstroke
.08 1 0 r
.98077 .5573 m
.97106 .55815 L
.96134 .55906 L
.95163 .56003 L
.94192 .56105 L
.93221 .56214 L
.92382 .56313 L
.92249 .56329 L
.91278 .5645 L
.90307 .56578 L
.89336 .56715 L
.88364 .5686 L
.87393 .57015 L
.86422 .57179 L
.85827 .57284 L
.85451 .57353 L
.84479 .57537 L
.83508 .57734 L
.82537 .57944 L
.81566 .58169 L
.81205 .58256 L
.80594 .58408 L
.79623 .58664 L
.78652 .58939 L
.77707 .59227 L
.77681 .59235 L
.76709 .59553 L
.75738 .59897 L
.74946 .60198 L
.74767 .60269 L
.73796 .60672 L
.72824 .61112 L
.72704 .61169 L
.71853 .61592 L
.70882 .62118 L
.70843 .62141 L
.69911 .62698 L
.69272 .63112 L
.68939 .63339 L
.67968 .64052 L
.67928 .64083 L
.66997 .64849 L
.66764 .65054 L
.66026 .65747 L
.65747 .66026 L
.65054 .66764 L
.64849 .66997 L
.64083 .67928 L
.64052 .67968 L
.63339 .68939 L
.63112 .69272 L
.62698 .69911 L
.62141 .70843 L
.62118 .70882 L
.61592 .71853 L
.61169 .72704 L
.61112 .72824 L
.60672 .73796 L
.60269 .74767 L
.60198 .74946 L
.59897 .75738 L
.59553 .76709 L
.59235 .77681 L
.59227 .77707 L
.58939 .78652 L
.58664 .79623 L
.58408 .80594 L
.58256 .81205 L
.58169 .81566 L
.57944 .82537 L
.57734 .83508 L
.57537 .84479 L
.57353 .85451 L
.57284 .85827 L
.57179 .86422 L
.57015 .87393 L
.5686 .88364 L
.56715 .89336 L
.56578 .90307 L
.5645 .91278 L
.56329 .92249 L
.56313 .92382 L
.56214 .93221 L
.56105 .94192 L
.56003 .95163 L
.55906 .96134 L
.55815 .97106 L
.5573 .98077 L
.98077 .98077 L
F
0 g
.98077 .5573 m
.97106 .55815 L
.96134 .55906 L
.95163 .56003 L
.94192 .56105 L
.93221 .56214 L
.92382 .56313 L
.92249 .56329 L
.91278 .5645 L
.90307 .56578 L
.89336 .56715 L
.88364 .5686 L
.87393 .57015 L
.86422 .57179 L
.85827 .57284 L
.85451 .57353 L
.84479 .57537 L
.83508 .57734 L
.82537 .57944 L
.81566 .58169 L
.81205 .58256 L
.80594 .58408 L
.79623 .58664 L
.78652 .58939 L
.77707 .59227 L
.77681 .59235 L
.76709 .59553 L
.75738 .59897 L
.74946 .60198 L
.74767 .60269 L
.73796 .60672 L
.72824 .61112 L
.72704 .61169 L
.71853 .61592 L
.70882 .62118 L
.70843 .62141 L
.69911 .62698 L
.69272 .63112 L
.68939 .63339 L
.67968 .64052 L
.67928 .64083 L
.66997 .64849 L
.66764 .65054 L
.66026 .65747 L
.65747 .66026 L
.65054 .66764 L
.64849 .66997 L
.64083 .67928 L
.64052 .67968 L
.63339 .68939 L
Mistroke
.63112 .69272 L
.62698 .69911 L
.62141 .70843 L
.62118 .70882 L
.61592 .71853 L
.61169 .72704 L
.61112 .72824 L
.60672 .73796 L
.60269 .74767 L
.60198 .74946 L
.59897 .75738 L
.59553 .76709 L
.59235 .77681 L
.59227 .77707 L
.58939 .78652 L
.58664 .79623 L
.58408 .80594 L
.58256 .81205 L
.58169 .81566 L
.57944 .82537 L
.57734 .83508 L
.57537 .84479 L
.57353 .85451 L
.57284 .85827 L
.57179 .86422 L
.57015 .87393 L
.5686 .88364 L
.56715 .89336 L
.56578 .90307 L
.5645 .91278 L
.56329 .92249 L
.56313 .92382 L
.56214 .93221 L
.56105 .94192 L
.56003 .95163 L
.55906 .96134 L
.55815 .97106 L
.5573 .98077 L
Mfstroke
.32 1 0 r
.98077 .43743 m
.97357 .43687 L
.97106 .43668 L
.96134 .43588 L
.95163 .435 L
.94192 .43405 L
.93221 .43302 L
.92249 .43191 L
.91278 .43072 L
.90307 .42945 L
.89336 .4281 L
.88694 .42716 L
.88364 .42666 L
.87393 .42514 L
.86422 .42351 L
.85451 .42176 L
.84479 .41988 L
.83508 .41788 L
.83302 .41744 L
.82537 .41576 L
.81566 .41348 L
.80594 .41103 L
.79623 .4084 L
.79388 .40773 L
.78652 .40558 L
.77681 .40254 L
.76709 .39925 L
.76363 .39802 L
.75738 .39571 L
.74767 .39186 L
.73938 .38831 L
.73796 .38767 L
.72824 .38311 L
.71942 .37859 L
.71853 .37812 L
.70882 .37263 L
.70268 .36888 L
.69911 .36659 L
.68939 .35988 L
.68842 .35917 L
.67968 .35242 L
.67611 .34946 L
.66997 .34405 L
.66538 .33974 L
.66026 .33462 L
.65595 .33003 L
.65054 .32389 L
.64758 .32032 L
.64083 .31158 L
.64012 .31061 L
.63341 .30089 L
.63112 .29732 L
.62737 .29118 L
.62188 .28147 L
.62141 .28058 L
.61689 .27176 L
.61233 .26204 L
.61169 .26062 L
.60814 .25233 L
.60429 .24262 L
.60198 .23637 L
.60075 .23291 L
.59746 .22319 L
.59442 .21348 L
.59227 .20612 L
.5916 .20377 L
.58897 .19406 L
.58652 .18434 L
.58424 .17463 L
.58256 .16698 L
.58212 .16492 L
.58012 .15521 L
.57824 .14549 L
.57649 .13578 L
.57486 .12607 L
.57334 .11636 L
.57284 .11306 L
.5719 .10664 L
.57055 .09693 L
.56928 .08722 L
.56809 .07751 L
.56698 .06779 L
.56595 .05808 L
.565 .04837 L
.56412 .03866 L
.56332 .02894 L
.56313 .02643 L
.56257 .01923 L
.98077 .01923 L
F
0 g
.98077 .43743 m
.97357 .43687 L
.97106 .43668 L
.96134 .43588 L
.95163 .435 L
.94192 .43405 L
.93221 .43302 L
.92249 .43191 L
.91278 .43072 L
.90307 .42945 L
.89336 .4281 L
.88694 .42716 L
.88364 .42666 L
.87393 .42514 L
.86422 .42351 L
.85451 .42176 L
.84479 .41988 L
.83508 .41788 L
.83302 .41744 L
.82537 .41576 L
.81566 .41348 L
.80594 .41103 L
.79623 .4084 L
.79388 .40773 L
.78652 .40558 L
.77681 .40254 L
.76709 .39925 L
.76363 .39802 L
.75738 .39571 L
.74767 .39186 L
.73938 .38831 L
.73796 .38767 L
.72824 .38311 L
.71942 .37859 L
.71853 .37812 L
.70882 .37263 L
.70268 .36888 L
.69911 .36659 L
.68939 .35988 L
.68842 .35917 L
.67968 .35242 L
.67611 .34946 L
.66997 .34405 L
.66538 .33974 L
.66026 .33462 L
.65595 .33003 L
.65054 .32389 L
.64758 .32032 L
.64083 .31158 L
.64012 .31061 L
Mistroke
.63341 .30089 L
.63112 .29732 L
.62737 .29118 L
.62188 .28147 L
.62141 .28058 L
.61689 .27176 L
.61233 .26204 L
.61169 .26062 L
.60814 .25233 L
.60429 .24262 L
.60198 .23637 L
.60075 .23291 L
.59746 .22319 L
.59442 .21348 L
.59227 .20612 L
.5916 .20377 L
.58897 .19406 L
.58652 .18434 L
.58424 .17463 L
.58256 .16698 L
.58212 .16492 L
.58012 .15521 L
.57824 .14549 L
.57649 .13578 L
.57486 .12607 L
.57334 .11636 L
.57284 .11306 L
.5719 .10664 L
.57055 .09693 L
.56928 .08722 L
.56809 .07751 L
.56698 .06779 L
.56595 .05808 L
.565 .04837 L
.56412 .03866 L
.56332 .02894 L
.56313 .02643 L
.56257 .01923 L
Mfstroke
.32 1 0 r
.98077 .56257 m
.97357 .56313 L
.97106 .56332 L
.96134 .56412 L
.95163 .565 L
.94192 .56595 L
.93221 .56698 L
.92249 .56809 L
.91278 .56928 L
.90307 .57055 L
.89336 .5719 L
.88694 .57284 L
.88364 .57334 L
.87393 .57486 L
.86422 .57649 L
.85451 .57824 L
.84479 .58012 L
.83508 .58212 L
.83302 .58256 L
.82537 .58424 L
.81566 .58652 L
.80594 .58897 L
.79623 .5916 L
.79388 .59227 L
.78652 .59442 L
.77681 .59746 L
.76709 .60075 L
.76363 .60198 L
.75738 .60429 L
.74767 .60814 L
.73938 .61169 L
.73796 .61233 L
.72824 .61689 L
.71942 .62141 L
.71853 .62188 L
.70882 .62737 L
.70268 .63112 L
.69911 .63341 L
.68939 .64012 L
.68842 .64083 L
.67968 .64758 L
.67611 .65054 L
.66997 .65595 L
.66538 .66026 L
.66026 .66538 L
.65595 .66997 L
.65054 .67611 L
.64758 .67968 L
.64083 .68842 L
.64012 .68939 L
.63341 .69911 L
.63112 .70268 L
.62737 .70882 L
.62188 .71853 L
.62141 .71942 L
.61689 .72824 L
.61233 .73796 L
.61169 .73938 L
.60814 .74767 L
.60429 .75738 L
.60198 .76363 L
.60075 .76709 L
.59746 .77681 L
.59442 .78652 L
.59227 .79388 L
.5916 .79623 L
.58897 .80594 L
.58652 .81566 L
.58424 .82537 L
.58256 .83302 L
.58212 .83508 L
.58012 .84479 L
.57824 .85451 L
.57649 .86422 L
.57486 .87393 L
.57334 .88364 L
.57284 .88694 L
.5719 .89336 L
.57055 .90307 L
.56928 .91278 L
.56809 .92249 L
.56698 .93221 L
.56595 .94192 L
.565 .95163 L
.56412 .96134 L
.56332 .97106 L
.56313 .97357 L
.56257 .98077 L
.98077 .98077 L
F
0 g
.98077 .56257 m
.97357 .56313 L
.97106 .56332 L
.96134 .56412 L
.95163 .565 L
.94192 .56595 L
.93221 .56698 L
.92249 .56809 L
.91278 .56928 L
.90307 .57055 L
.89336 .5719 L
.88694 .57284 L
.88364 .57334 L
.87393 .57486 L
.86422 .57649 L
.85451 .57824 L
.84479 .58012 L
.83508 .58212 L
.83302 .58256 L
.82537 .58424 L
.81566 .58652 L
.80594 .58897 L
.79623 .5916 L
.79388 .59227 L
.78652 .59442 L
.77681 .59746 L
.76709 .60075 L
.76363 .60198 L
.75738 .60429 L
.74767 .60814 L
.73938 .61169 L
.73796 .61233 L
.72824 .61689 L
.71942 .62141 L
.71853 .62188 L
.70882 .62737 L
.70268 .63112 L
.69911 .63341 L
.68939 .64012 L
.68842 .64083 L
.67968 .64758 L
.67611 .65054 L
.66997 .65595 L
.66538 .66026 L
.66026 .66538 L
.65595 .66997 L
.65054 .67611 L
.64758 .67968 L
.64083 .68842 L
.64012 .68939 L
Mistroke
.63341 .69911 L
.63112 .70268 L
.62737 .70882 L
.62188 .71853 L
.62141 .71942 L
.61689 .72824 L
.61233 .73796 L
.61169 .73938 L
.60814 .74767 L
.60429 .75738 L
.60198 .76363 L
.60075 .76709 L
.59746 .77681 L
.59442 .78652 L
.59227 .79388 L
.5916 .79623 L
.58897 .80594 L
.58652 .81566 L
.58424 .82537 L
.58256 .83302 L
.58212 .83508 L
.58012 .84479 L
.57824 .85451 L
.57649 .86422 L
.57486 .87393 L
.57334 .88364 L
.57284 .88694 L
.5719 .89336 L
.57055 .90307 L
.56928 .91278 L
.56809 .92249 L
.56698 .93221 L
.56595 .94192 L
.565 .95163 L
.56412 .96134 L
.56332 .97106 L
.56313 .97357 L
.56257 .98077 L
Mfstroke
.56 1 0 r
.98077 .43173 m
.97106 .43108 L
.96134 .43036 L
.95163 .42956 L
.94192 .42868 L
.93221 .42772 L
.92686 .42716 L
.92249 .4267 L
.91278 .42561 L
.90307 .42441 L
.89336 .4231 L
.88364 .42169 L
.87393 .42016 L
.86422 .41852 L
.85816 .41744 L
.85451 .41678 L
.84479 .41492 L
.83508 .41291 L
.82537 .41074 L
.81566 .40842 L
.81293 .40773 L
.80594 .40593 L
.79623 .40325 L
.78652 .40035 L
.77923 .39802 L
.77681 .39722 L
.76709 .39384 L
.75738 .39017 L
.75273 .38831 L
.74767 .3862 L
.73796 .38186 L
.73118 .37859 L
.72824 .37712 L
.71853 .37193 L
.71324 .36888 L
.70882 .36621 L
.69911 .3599 L
.69805 .35917 L
.68939 .35289 L
.685 .34946 L
.67968 .34506 L
.67367 .33974 L
.66997 .33627 L
.66373 .33003 L
.66026 .32633 L
.65494 .32032 L
.65054 .315 L
.64711 .31061 L
.64083 .30195 L
.6401 .30089 L
.63379 .29118 L
.63112 .28676 L
.62807 .28147 L
.62288 .27176 L
.62141 .26882 L
.61814 .26204 L
.6138 .25233 L
.61169 .24727 L
.60983 .24262 L
.60616 .23291 L
.60278 .22319 L
.60198 .22077 L
.59965 .21348 L
.59675 .20377 L
.59407 .19406 L
.59227 .18707 L
.59158 .18434 L
.58926 .17463 L
.58709 .16492 L
.58508 .15521 L
.58322 .14549 L
.58256 .14184 L
.58148 .13578 L
.57984 .12607 L
.57831 .11636 L
.5769 .10664 L
.57559 .09693 L
.57439 .08722 L
.5733 .07751 L
.57284 .07314 L
.57228 .06779 L
.57132 .05808 L
.57044 .04837 L
.56964 .03866 L
.56892 .02894 L
.56827 .01923 L
.98077 .01923 L
F
0 g
.98077 .43173 m
.97106 .43108 L
.96134 .43036 L
.95163 .42956 L
.94192 .42868 L
.93221 .42772 L
.92686 .42716 L
.92249 .4267 L
.91278 .42561 L
.90307 .42441 L
.89336 .4231 L
.88364 .42169 L
.87393 .42016 L
.86422 .41852 L
.85816 .41744 L
.85451 .41678 L
.84479 .41492 L
.83508 .41291 L
.82537 .41074 L
.81566 .40842 L
.81293 .40773 L
.80594 .40593 L
.79623 .40325 L
.78652 .40035 L
.77923 .39802 L
.77681 .39722 L
.76709 .39384 L
.75738 .39017 L
.75273 .38831 L
.74767 .3862 L
.73796 .38186 L
.73118 .37859 L
.72824 .37712 L
.71853 .37193 L
.71324 .36888 L
.70882 .36621 L
.69911 .3599 L
.69805 .35917 L
.68939 .35289 L
.685 .34946 L
.67968 .34506 L
.67367 .33974 L
.66997 .33627 L
.66373 .33003 L
.66026 .32633 L
.65494 .32032 L
.65054 .315 L
.64711 .31061 L
.64083 .30195 L
.6401 .30089 L
Mistroke
.63379 .29118 L
.63112 .28676 L
.62807 .28147 L
.62288 .27176 L
.62141 .26882 L
.61814 .26204 L
.6138 .25233 L
.61169 .24727 L
.60983 .24262 L
.60616 .23291 L
.60278 .22319 L
.60198 .22077 L
.59965 .21348 L
.59675 .20377 L
.59407 .19406 L
.59227 .18707 L
.59158 .18434 L
.58926 .17463 L
.58709 .16492 L
.58508 .15521 L
.58322 .14549 L
.58256 .14184 L
.58148 .13578 L
.57984 .12607 L
.57831 .11636 L
.5769 .10664 L
.57559 .09693 L
.57439 .08722 L
.5733 .07751 L
.57284 .07314 L
.57228 .06779 L
.57132 .05808 L
.57044 .04837 L
.56964 .03866 L
.56892 .02894 L
.56827 .01923 L
Mfstroke
.56 1 0 r
.98077 .56827 m
.97106 .56892 L
.96134 .56964 L
.95163 .57044 L
.94192 .57132 L
.93221 .57228 L
.92686 .57284 L
.92249 .5733 L
.91278 .57439 L
.90307 .57559 L
.89336 .5769 L
.88364 .57831 L
.87393 .57984 L
.86422 .58148 L
.85816 .58256 L
.85451 .58322 L
.84479 .58508 L
.83508 .58709 L
.82537 .58926 L
.81566 .59158 L
.81293 .59227 L
.80594 .59407 L
.79623 .59675 L
.78652 .59965 L
.77923 .60198 L
.77681 .60278 L
.76709 .60616 L
.75738 .60983 L
.75273 .61169 L
.74767 .6138 L
.73796 .61814 L
.73118 .62141 L
.72824 .62288 L
.71853 .62807 L
.71324 .63112 L
.70882 .63379 L
.69911 .6401 L
.69805 .64083 L
.68939 .64711 L
.685 .65054 L
.67968 .65494 L
.67367 .66026 L
.66997 .66373 L
.66373 .66997 L
.66026 .67367 L
.65494 .67968 L
.65054 .685 L
.64711 .68939 L
.64083 .69805 L
.6401 .69911 L
.63379 .70882 L
.63112 .71324 L
.62807 .71853 L
.62288 .72824 L
.62141 .73118 L
.61814 .73796 L
.6138 .74767 L
.61169 .75273 L
.60983 .75738 L
.60616 .76709 L
.60278 .77681 L
.60198 .77923 L
.59965 .78652 L
.59675 .79623 L
.59407 .80594 L
.59227 .81293 L
.59158 .81566 L
.58926 .82537 L
.58709 .83508 L
.58508 .84479 L
.58322 .85451 L
.58256 .85816 L
.58148 .86422 L
.57984 .87393 L
.57831 .88364 L
.5769 .89336 L
.57559 .90307 L
.57439 .91278 L
.5733 .92249 L
.57284 .92686 L
.57228 .93221 L
.57132 .94192 L
.57044 .95163 L
.56964 .96134 L
.56892 .97106 L
.56827 .98077 L
.98077 .98077 L
F
0 g
.98077 .56827 m
.97106 .56892 L
.96134 .56964 L
.95163 .57044 L
.94192 .57132 L
.93221 .57228 L
.92686 .57284 L
.92249 .5733 L
.91278 .57439 L
.90307 .57559 L
.89336 .5769 L
.88364 .57831 L
.87393 .57984 L
.86422 .58148 L
.85816 .58256 L
.85451 .58322 L
.84479 .58508 L
.83508 .58709 L
.82537 .58926 L
.81566 .59158 L
.81293 .59227 L
.80594 .59407 L
.79623 .59675 L
.78652 .59965 L
.77923 .60198 L
.77681 .60278 L
.76709 .60616 L
.75738 .60983 L
.75273 .61169 L
.74767 .6138 L
.73796 .61814 L
.73118 .62141 L
.72824 .62288 L
.71853 .62807 L
.71324 .63112 L
.70882 .63379 L
.69911 .6401 L
.69805 .64083 L
.68939 .64711 L
.685 .65054 L
.67968 .65494 L
.67367 .66026 L
.66997 .66373 L
.66373 .66997 L
.66026 .67367 L
.65494 .67968 L
.65054 .685 L
.64711 .68939 L
.64083 .69805 L
.6401 .69911 L
Mistroke
.63379 .70882 L
.63112 .71324 L
.62807 .71853 L
.62288 .72824 L
.62141 .73118 L
.61814 .73796 L
.6138 .74767 L
.61169 .75273 L
.60983 .75738 L
.60616 .76709 L
.60278 .77681 L
.60198 .77923 L
.59965 .78652 L
.59675 .79623 L
.59407 .80594 L
.59227 .81293 L
.59158 .81566 L
.58926 .82537 L
.58709 .83508 L
.58508 .84479 L
.58322 .85451 L
.58256 .85816 L
.58148 .86422 L
.57984 .87393 L
.57831 .88364 L
.5769 .89336 L
.57559 .90307 L
.57439 .91278 L
.5733 .92249 L
.57284 .92686 L
.57228 .93221 L
.57132 .94192 L
.57044 .95163 L
.56964 .96134 L
.56892 .97106 L
.56827 .98077 L
Mfstroke
.8 1 0 r
.98077 .42536 m
.97106 .42475 L
.96134 .42415 L
.95163 .42349 L
.94192 .42275 L
.93221 .42191 L
.92249 .42097 L
.91278 .41993 L
.90307 .41879 L
.89336 .41754 L
.89266 .41744 L
.88364 .41624 L
.87393 .4148 L
.86422 .4132 L
.85451 .41147 L
.84479 .40958 L
.83589 .40773 L
.83508 .40756 L
.82537 .40541 L
.81566 .40307 L
.80594 .40052 L
.79708 .39802 L
.79623 .39777 L
.78652 .39483 L
.77681 .39161 L
.76758 .38831 L
.76709 .38813 L
.75738 .38436 L
.74767 .38023 L
.74403 .37859 L
.73796 .37575 L
.72824 .37082 L
.72465 .36888 L
.71853 .36542 L
.70882 .35945 L
.70838 .35917 L
.69911 .35286 L
.69448 .34946 L
.68939 .34551 L
.68246 .33974 L
.67968 .33729 L
.67196 .33003 L
.66997 .32804 L
.66271 .32032 L
.66026 .31754 L
.65449 .31061 L
.65054 .30552 L
.64714 .30089 L
.64083 .29162 L
.64055 .29118 L
.63458 .28147 L
.63112 .27535 L
.62918 .27176 L
.62425 .26204 L
.62141 .25597 L
.61977 .25233 L
.61564 .24262 L
.61187 .23291 L
.61169 .23242 L
.60839 .22319 L
.60517 .21348 L
.60223 .20377 L
.60198 .20292 L
.59948 .19406 L
.59693 .18434 L
.59459 .17463 L
.59244 .16492 L
.59227 .16411 L
.59042 .15521 L
.58853 .14549 L
.5868 .13578 L
.5852 .12607 L
.58376 .11636 L
.58256 .10734 L
.58246 .10664 L
.58121 .09693 L
.58007 .08722 L
.57903 .07751 L
.57809 .06779 L
.57725 .05808 L
.57651 .04837 L
.57585 .03866 L
.57525 .02894 L
.57464 .01923 L
.98077 .01923 L
F
0 g
.98077 .42536 m
.97106 .42475 L
.96134 .42415 L
.95163 .42349 L
.94192 .42275 L
.93221 .42191 L
.92249 .42097 L
.91278 .41993 L
.90307 .41879 L
.89336 .41754 L
.89266 .41744 L
.88364 .41624 L
.87393 .4148 L
.86422 .4132 L
.85451 .41147 L
.84479 .40958 L
.83589 .40773 L
.83508 .40756 L
.82537 .40541 L
.81566 .40307 L
.80594 .40052 L
.79708 .39802 L
.79623 .39777 L
.78652 .39483 L
.77681 .39161 L
.76758 .38831 L
.76709 .38813 L
.75738 .38436 L
.74767 .38023 L
.74403 .37859 L
.73796 .37575 L
.72824 .37082 L
.72465 .36888 L
.71853 .36542 L
.70882 .35945 L
.70838 .35917 L
.69911 .35286 L
.69448 .34946 L
.68939 .34551 L
.68246 .33974 L
.67968 .33729 L
.67196 .33003 L
.66997 .32804 L
.66271 .32032 L
.66026 .31754 L
.65449 .31061 L
.65054 .30552 L
.64714 .30089 L
.64083 .29162 L
.64055 .29118 L
Mistroke
.63458 .28147 L
.63112 .27535 L
.62918 .27176 L
.62425 .26204 L
.62141 .25597 L
.61977 .25233 L
.61564 .24262 L
.61187 .23291 L
.61169 .23242 L
.60839 .22319 L
.60517 .21348 L
.60223 .20377 L
.60198 .20292 L
.59948 .19406 L
.59693 .18434 L
.59459 .17463 L
.59244 .16492 L
.59227 .16411 L
.59042 .15521 L
.58853 .14549 L
.5868 .13578 L
.5852 .12607 L
.58376 .11636 L
.58256 .10734 L
.58246 .10664 L
.58121 .09693 L
.58007 .08722 L
.57903 .07751 L
.57809 .06779 L
.57725 .05808 L
.57651 .04837 L
.57585 .03866 L
.57525 .02894 L
.57464 .01923 L
Mfstroke
.8 1 0 r
.98077 .57464 m
.97106 .57525 L
.96134 .57585 L
.95163 .57651 L
.94192 .57725 L
.93221 .57809 L
.92249 .57903 L
.91278 .58007 L
.90307 .58121 L
.89336 .58246 L
.89266 .58256 L
.88364 .58376 L
.87393 .5852 L
.86422 .5868 L
.85451 .58853 L
.84479 .59042 L
.83589 .59227 L
.83508 .59244 L
.82537 .59459 L
.81566 .59693 L
.80594 .59948 L
.79708 .60198 L
.79623 .60223 L
.78652 .60517 L
.77681 .60839 L
.76758 .61169 L
.76709 .61187 L
.75738 .61564 L
.74767 .61977 L
.74403 .62141 L
.73796 .62425 L
.72824 .62918 L
.72465 .63112 L
.71853 .63458 L
.70882 .64055 L
.70838 .64083 L
.69911 .64714 L
.69448 .65054 L
.68939 .65449 L
.68246 .66026 L
.67968 .66271 L
.67196 .66997 L
.66997 .67196 L
.66271 .67968 L
.66026 .68246 L
.65449 .68939 L
.65054 .69448 L
.64714 .69911 L
.64083 .70838 L
.64055 .70882 L
.63458 .71853 L
.63112 .72465 L
.62918 .72824 L
.62425 .73796 L
.62141 .74403 L
.61977 .74767 L
.61564 .75738 L
.61187 .76709 L
.61169 .76758 L
.60839 .77681 L
.60517 .78652 L
.60223 .79623 L
.60198 .79708 L
.59948 .80594 L
.59693 .81566 L
.59459 .82537 L
.59244 .83508 L
.59227 .83589 L
.59042 .84479 L
.58853 .85451 L
.5868 .86422 L
.5852 .87393 L
.58376 .88364 L
.58256 .89266 L
.58246 .89336 L
.58121 .90307 L
.58007 .91278 L
.57903 .92249 L
.57809 .93221 L
.57725 .94192 L
.57651 .95163 L
.57585 .96134 L
.57525 .97106 L
.57464 .98077 L
.98077 .98077 L
F
0 g
.98077 .57464 m
.97106 .57525 L
.96134 .57585 L
.95163 .57651 L
.94192 .57725 L
.93221 .57809 L
.92249 .57903 L
.91278 .58007 L
.90307 .58121 L
.89336 .58246 L
.89266 .58256 L
.88364 .58376 L
.87393 .5852 L
.86422 .5868 L
.85451 .58853 L
.84479 .59042 L
.83589 .59227 L
.83508 .59244 L
.82537 .59459 L
.81566 .59693 L
.80594 .59948 L
.79708 .60198 L
.79623 .60223 L
.78652 .60517 L
.77681 .60839 L
.76758 .61169 L
.76709 .61187 L
.75738 .61564 L
.74767 .61977 L
.74403 .62141 L
.73796 .62425 L
.72824 .62918 L
.72465 .63112 L
.71853 .63458 L
.70882 .64055 L
.70838 .64083 L
.69911 .64714 L
.69448 .65054 L
.68939 .65449 L
.68246 .66026 L
.67968 .66271 L
.67196 .66997 L
.66997 .67196 L
.66271 .67968 L
.66026 .68246 L
.65449 .68939 L
.65054 .69448 L
.64714 .69911 L
.64083 .70838 L
.64055 .70882 L
Mistroke
.63458 .71853 L
.63112 .72465 L
.62918 .72824 L
.62425 .73796 L
.62141 .74403 L
.61977 .74767 L
.61564 .75738 L
.61187 .76709 L
.61169 .76758 L
.60839 .77681 L
.60517 .78652 L
.60223 .79623 L
.60198 .79708 L
.59948 .80594 L
.59693 .81566 L
.59459 .82537 L
.59244 .83508 L
.59227 .83589 L
.59042 .84479 L
.58853 .85451 L
.5868 .86422 L
.5852 .87393 L
.58376 .88364 L
.58256 .89266 L
.58246 .89336 L
.58121 .90307 L
.58007 .91278 L
.57903 .92249 L
.57809 .93221 L
.57725 .94192 L
.57651 .95163 L
.57585 .96134 L
.57525 .97106 L
.57464 .98077 L
Mfstroke
1 .96 0 r
.98077 .58788 m
.97106 .58611 L
.96134 .58539 L
.95163 .58516 L
.94192 .58368 L
.93221 .58474 L
.92249 .58562 L
.91278 .58654 L
.90307 .58755 L
.89336 .58869 L
.88364 .58995 L
.87393 .59134 L
.86794 .59227 L
.86422 .59282 L
.85451 .59443 L
.84479 .59625 L
.83508 .59826 L
.82537 .60045 L
.819 .60198 L
.81566 .60279 L
.80594 .60533 L
.79623 .60812 L
.78652 .61116 L
.78488 .61169 L
.77681 .61442 L
.76709 .61799 L
.75857 .62141 L
.75738 .6219 L
.74767 .62614 L
.73796 .63081 L
.73735 .63112 L
.72824 .63591 L
.71972 .64083 L
.71853 .64155 L
.70882 .64777 L
.7048 .65054 L
.69911 .65468 L
.69198 .66026 L
.68939 .6624 L
.68084 .66997 L
.67968 .67106 L
.67106 .67968 L
.66997 .68084 L
.6624 .68939 L
.66026 .69198 L
.65468 .69911 L
.65054 .7048 L
.64777 .70882 L
.64155 .71853 L
.64083 .71972 L
.63591 .72824 L
.63112 .73735 L
.63081 .73796 L
.62614 .74767 L
.6219 .75738 L
.62141 .75857 L
.61799 .76709 L
.61442 .77681 L
.61169 .78488 L
.61116 .78652 L
.60812 .79623 L
.60533 .80594 L
.60279 .81566 L
.60198 .819 L
.60045 .82537 L
.59826 .83508 L
.59625 .84479 L
.59443 .85451 L
.59282 .86422 L
.59227 .86794 L
.59134 .87393 L
.58995 .88364 L
.58869 .89336 L
.58755 .90307 L
.58654 .91278 L
.58562 .92249 L
.58474 .93221 L
.58368 .94192 L
.58516 .95163 L
.58539 .96134 L
.58611 .97106 L
.58788 .98077 L
.98077 .98077 L
F
0 g
.98077 .58788 m
.97106 .58611 L
.96134 .58539 L
.95163 .58516 L
.94192 .58368 L
.93221 .58474 L
.92249 .58562 L
.91278 .58654 L
.90307 .58755 L
.89336 .58869 L
.88364 .58995 L
.87393 .59134 L
.86794 .59227 L
.86422 .59282 L
.85451 .59443 L
.84479 .59625 L
.83508 .59826 L
.82537 .60045 L
.819 .60198 L
.81566 .60279 L
.80594 .60533 L
.79623 .60812 L
.78652 .61116 L
.78488 .61169 L
.77681 .61442 L
.76709 .61799 L
.75857 .62141 L
.75738 .6219 L
.74767 .62614 L
.73796 .63081 L
.73735 .63112 L
.72824 .63591 L
.71972 .64083 L
.71853 .64155 L
.70882 .64777 L
.7048 .65054 L
.69911 .65468 L
.69198 .66026 L
.68939 .6624 L
.68084 .66997 L
.67968 .67106 L
.67106 .67968 L
.66997 .68084 L
.6624 .68939 L
.66026 .69198 L
.65468 .69911 L
.65054 .7048 L
.64777 .70882 L
.64155 .71853 L
.64083 .71972 L
Mistroke
.63591 .72824 L
.63112 .73735 L
.63081 .73796 L
.62614 .74767 L
.6219 .75738 L
.62141 .75857 L
.61799 .76709 L
.61442 .77681 L
.61169 .78488 L
.61116 .78652 L
.60812 .79623 L
.60533 .80594 L
.60279 .81566 L
.60198 .819 L
.60045 .82537 L
.59826 .83508 L
.59625 .84479 L
.59443 .85451 L
.59282 .86422 L
.59227 .86794 L
.59134 .87393 L
.58995 .88364 L
.58869 .89336 L
.58755 .90307 L
.58654 .91278 L
.58562 .92249 L
.58474 .93221 L
.58368 .94192 L
.58516 .95163 L
.58539 .96134 L
.58611 .97106 L
.58788 .98077 L
Mfstroke
1 .96 0 r
.98077 .41212 m
.97106 .41389 L
.96134 .41461 L
.95163 .41484 L
.94192 .41632 L
.93221 .41526 L
.92249 .41438 L
.91278 .41346 L
.90307 .41245 L
.89336 .41131 L
.88364 .41005 L
.87393 .40866 L
.86794 .40773 L
.86422 .40718 L
.85451 .40557 L
.84479 .40375 L
.83508 .40174 L
.82537 .39955 L
.819 .39802 L
.81566 .39721 L
.80594 .39467 L
.79623 .39188 L
.78652 .38884 L
.78488 .38831 L
.77681 .38558 L
.76709 .38201 L
.75857 .37859 L
.75738 .3781 L
.74767 .37386 L
.73796 .36919 L
.73735 .36888 L
.72824 .36409 L
.71972 .35917 L
.71853 .35845 L
.70882 .35223 L
.7048 .34946 L
.69911 .34532 L
.69198 .33974 L
.68939 .3376 L
.68084 .33003 L
.67968 .32894 L
.67106 .32032 L
.66997 .31916 L
.6624 .31061 L
.66026 .30802 L
.65468 .30089 L
.65054 .2952 L
.64777 .29118 L
.64155 .28147 L
.64083 .28028 L
.63591 .27176 L
.63112 .26265 L
.63081 .26204 L
.62614 .25233 L
.6219 .24262 L
.62141 .24143 L
.61799 .23291 L
.61442 .22319 L
.61169 .21512 L
.61116 .21348 L
.60812 .20377 L
.60533 .19406 L
.60279 .18434 L
.60198 .181 L
.60045 .17463 L
.59826 .16492 L
.59625 .15521 L
.59443 .14549 L
.59282 .13578 L
.59227 .13206 L
.59134 .12607 L
.58995 .11636 L
.58869 .10664 L
.58755 .09693 L
.58654 .08722 L
.58562 .07751 L
.58474 .06779 L
.58368 .05808 L
.58516 .04837 L
.58539 .03866 L
.58611 .02894 L
.58788 .01923 L
.98077 .01923 L
F
0 g
.98077 .41212 m
.97106 .41389 L
.96134 .41461 L
.95163 .41484 L
.94192 .41632 L
.93221 .41526 L
.92249 .41438 L
.91278 .41346 L
.90307 .41245 L
.89336 .41131 L
.88364 .41005 L
.87393 .40866 L
.86794 .40773 L
.86422 .40718 L
.85451 .40557 L
.84479 .40375 L
.83508 .40174 L
.82537 .39955 L
.819 .39802 L
.81566 .39721 L
.80594 .39467 L
.79623 .39188 L
.78652 .38884 L
.78488 .38831 L
.77681 .38558 L
.76709 .38201 L
.75857 .37859 L
.75738 .3781 L
.74767 .37386 L
.73796 .36919 L
.73735 .36888 L
.72824 .36409 L
.71972 .35917 L
.71853 .35845 L
.70882 .35223 L
.7048 .34946 L
.69911 .34532 L
.69198 .33974 L
.68939 .3376 L
.68084 .33003 L
.67968 .32894 L
.67106 .32032 L
.66997 .31916 L
.6624 .31061 L
.66026 .30802 L
.65468 .30089 L
.65054 .2952 L
.64777 .29118 L
.64155 .28147 L
.64083 .28028 L
Mistroke
.63591 .27176 L
.63112 .26265 L
.63081 .26204 L
.62614 .25233 L
.6219 .24262 L
.62141 .24143 L
.61799 .23291 L
.61442 .22319 L
.61169 .21512 L
.61116 .21348 L
.60812 .20377 L
.60533 .19406 L
.60279 .18434 L
.60198 .181 L
.60045 .17463 L
.59826 .16492 L
.59625 .15521 L
.59443 .14549 L
.59282 .13578 L
.59227 .13206 L
.59134 .12607 L
.58995 .11636 L
.58869 .10664 L
.58755 .09693 L
.58654 .08722 L
.58562 .07751 L
.58474 .06779 L
.58368 .05808 L
.58516 .04837 L
.58539 .03866 L
.58611 .02894 L
.58788 .01923 L
Mfstroke
1 .72 0 r
.85451 .6015 m
.86422 .59996 L
.87393 .59857 L
.88364 .59732 L
.89336 .59611 L
.90307 .59652 L
.91278 .5964 L
.92249 .59686 L
.93221 .59859 L
.93715 .60198 L
.93221 .60984 L
.92976 .61169 L
.92249 .6163 L
.9134 .62141 L
.91278 .62182 L
.90307 .62858 L
.89466 .63112 L
.89336 .63185 L
.88364 .6373 L
.87562 .64083 L
.87393 .64176 L
.86422 .64707 L
.85718 .65054 L
.85451 .65204 L
.84479 .65752 L
.83968 .66026 L
.83508 .66298 L
.82537 .66875 L
.82327 .66997 L
.81566 .67479 L
.80797 .67968 L
.80594 .68104 L
.79623 .68768 L
.79375 .68939 L
.78652 .69465 L
.78052 .69911 L
.77681 .70198 L
.76821 .70882 L
.76709 .70974 L
.75738 .71799 L
.75676 .71853 L
.74767 .72677 L
.74609 .72824 L
.73796 .73611 L
.73611 .73796 L
.72824 .74609 L
.72677 .74767 L
.71853 .75676 L
.71799 .75738 L
.70974 .76709 L
.70882 .76821 L
.70198 .77681 L
.69911 .78052 L
.69465 .78652 L
.68939 .79375 L
.68768 .79623 L
.68104 .80594 L
.67968 .80797 L
.67479 .81566 L
.66997 .82327 L
.66875 .82537 L
.66298 .83508 L
.66026 .83968 L
.65752 .84479 L
.65204 .85451 L
.65054 .85718 L
.64707 .86422 L
.64176 .87393 L
.64083 .87562 L
.6373 .88364 L
.63185 .89336 L
.63112 .89466 L
.62858 .90307 L
.62182 .91278 L
.62141 .9134 L
.6163 .92249 L
.61169 .92976 L
.60984 .93221 L
.60198 .93715 L
.59859 .93221 L
.59686 .92249 L
.5964 .91278 L
.59652 .90307 L
.59611 .89336 L
.59732 .88364 L
.59857 .87393 L
.59996 .86422 L
.6015 .85451 L
.60198 .85167 L
.60306 .84479 L
.60491 .83508 L
.60704 .82537 L
.6094 .81566 L
.61169 .80698 L
.61196 .80594 L
.61469 .79623 L
.61776 .78652 L
.62113 .77681 L
.62141 .77606 L
.62475 .76709 L
.62877 .75738 L
.63112 .7521 L
.63315 .74767 L
.63798 .73796 L
.64083 .73265 L
.64329 .72824 L
.64918 .71853 L
.65054 .71642 L
.65569 .70882 L
.66026 .7026 L
.66295 .69911 L
.66997 .69068 L
.6711 .68939 L
.67968 .68027 L
.68027 .67968 L
.68939 .6711 L
.69068 .66997 L
.69911 .66295 L
.7026 .66026 L
.70882 .65569 L
.71642 .65054 L
.71853 .64918 L
.72824 .64329 L
.73265 .64083 L
.73796 .63798 L
.74767 .63315 L
.7521 .63112 L
.75738 .62877 L
.76709 .62475 L
.77606 .62141 L
.77681 .62113 L
.78652 .61776 L
.79623 .61469 L
.80594 .61196 L
.80698 .61169 L
.81566 .6094 L
.82537 .60704 L
.83508 .60491 L
.84479 .60306 L
.85167 .60198 L
F
0 g
.85451 .6015 m
.86422 .59996 L
.87393 .59857 L
.88364 .59732 L
.89336 .59611 L
.90307 .59652 L
.91278 .5964 L
.92249 .59686 L
.93221 .59859 L
.93715 .60198 L
.93221 .60984 L
.92976 .61169 L
.92249 .6163 L
.9134 .62141 L
.91278 .62182 L
.90307 .62858 L
.89466 .63112 L
.89336 .63185 L
.88364 .6373 L
.87562 .64083 L
.87393 .64176 L
.86422 .64707 L
.85718 .65054 L
.85451 .65204 L
.84479 .65752 L
.83968 .66026 L
.83508 .66298 L
.82537 .66875 L
.82327 .66997 L
.81566 .67479 L
.80797 .67968 L
.80594 .68104 L
.79623 .68768 L
.79375 .68939 L
.78652 .69465 L
.78052 .69911 L
.77681 .70198 L
.76821 .70882 L
.76709 .70974 L
.75738 .71799 L
.75676 .71853 L
.74767 .72677 L
.74609 .72824 L
.73796 .73611 L
.73611 .73796 L
.72824 .74609 L
.72677 .74767 L
.71853 .75676 L
.71799 .75738 L
.70974 .76709 L
Mistroke
.70882 .76821 L
.70198 .77681 L
.69911 .78052 L
.69465 .78652 L
.68939 .79375 L
.68768 .79623 L
.68104 .80594 L
.67968 .80797 L
.67479 .81566 L
.66997 .82327 L
.66875 .82537 L
.66298 .83508 L
.66026 .83968 L
.65752 .84479 L
.65204 .85451 L
.65054 .85718 L
.64707 .86422 L
.64176 .87393 L
.64083 .87562 L
.6373 .88364 L
.63185 .89336 L
.63112 .89466 L
.62858 .90307 L
.62182 .91278 L
.62141 .9134 L
.6163 .92249 L
.61169 .92976 L
.60984 .93221 L
.60198 .93715 L
.59859 .93221 L
.59686 .92249 L
.5964 .91278 L
.59652 .90307 L
.59611 .89336 L
.59732 .88364 L
.59857 .87393 L
.59996 .86422 L
.6015 .85451 L
.60198 .85167 L
.60306 .84479 L
.60491 .83508 L
.60704 .82537 L
.6094 .81566 L
.61169 .80698 L
.61196 .80594 L
.61469 .79623 L
.61776 .78652 L
.62113 .77681 L
.62141 .77606 L
.62475 .76709 L
Mistroke
.62877 .75738 L
.63112 .7521 L
.63315 .74767 L
.63798 .73796 L
.64083 .73265 L
.64329 .72824 L
.64918 .71853 L
.65054 .71642 L
.65569 .70882 L
.66026 .7026 L
.66295 .69911 L
.66997 .69068 L
.6711 .68939 L
.67968 .68027 L
.68027 .67968 L
.68939 .6711 L
.69068 .66997 L
.69911 .66295 L
.7026 .66026 L
.70882 .65569 L
.71642 .65054 L
.71853 .64918 L
.72824 .64329 L
.73265 .64083 L
.73796 .63798 L
.74767 .63315 L
.7521 .63112 L
.75738 .62877 L
.76709 .62475 L
.77606 .62141 L
.77681 .62113 L
.78652 .61776 L
.79623 .61469 L
.80594 .61196 L
.80698 .61169 L
.81566 .6094 L
.82537 .60704 L
.83508 .60491 L
.84479 .60306 L
.85167 .60198 L
.85451 .6015 L
Mfstroke
1 .72 0 r
.60198 .06285 m
.60984 .06779 L
.61169 .07024 L
.6163 .07751 L
.62141 .0866 L
.62182 .08722 L
.62858 .09693 L
.63112 .10534 L
.63185 .10664 L
.6373 .11636 L
.64083 .12438 L
.64176 .12607 L
.64707 .13578 L
.65054 .14282 L
.65204 .14549 L
.65752 .15521 L
.66026 .16032 L
.66298 .16492 L
.66875 .17463 L
.66997 .17673 L
.67479 .18434 L
.67968 .19203 L
.68104 .19406 L
.68768 .20377 L
.68939 .20625 L
.69465 .21348 L
.69911 .21948 L
.70198 .22319 L
.70882 .23179 L
.70974 .23291 L
.71799 .24262 L
.71853 .24324 L
.72677 .25233 L
.72824 .25391 L
.73611 .26204 L
.73796 .26389 L
.74609 .27176 L
.74767 .27323 L
.75676 .28147 L
.75738 .28201 L
.76709 .29026 L
.76821 .29118 L
.77681 .29802 L
.78052 .30089 L
.78652 .30535 L
.79375 .31061 L
.79623 .31232 L
.80594 .31896 L
.80797 .32032 L
.81566 .32521 L
.82327 .33003 L
.82537 .33125 L
.83508 .33702 L
.83968 .33974 L
.84479 .34248 L
.85451 .34796 L
.85718 .34946 L
.86422 .35293 L
.87393 .35824 L
.87562 .35917 L
.88364 .3627 L
.89336 .36815 L
.89466 .36888 L
.90307 .37142 L
.91278 .37818 L
.9134 .37859 L
.92249 .3837 L
.92976 .38831 L
.93221 .39016 L
.93715 .39802 L
.93221 .40141 L
.92249 .40314 L
.91278 .4036 L
.90307 .40348 L
.89336 .40389 L
.88364 .40268 L
.87393 .40143 L
.86422 .40004 L
.85451 .3985 L
.85167 .39802 L
.84479 .39694 L
.83508 .39509 L
.82537 .39296 L
.81566 .3906 L
.80698 .38831 L
.80594 .38804 L
.79623 .38531 L
.78652 .38224 L
.77681 .37887 L
.77606 .37859 L
.76709 .37525 L
.75738 .37123 L
.7521 .36888 L
.74767 .36685 L
.73796 .36202 L
.73265 .35917 L
.72824 .35671 L
.71853 .35082 L
.71642 .34946 L
.70882 .34431 L
.7026 .33974 L
.69911 .33705 L
.69068 .33003 L
.68939 .3289 L
.68027 .32032 L
.67968 .31973 L
.6711 .31061 L
.66997 .30932 L
.66295 .30089 L
.66026 .2974 L
.65569 .29118 L
.65054 .28358 L
.64918 .28147 L
.64329 .27176 L
.64083 .26735 L
.63798 .26204 L
.63315 .25233 L
.63112 .2479 L
.62877 .24262 L
.62475 .23291 L
.62141 .22394 L
.62113 .22319 L
.61776 .21348 L
.61469 .20377 L
.61196 .19406 L
.61169 .19302 L
.6094 .18434 L
.60704 .17463 L
.60491 .16492 L
.60306 .15521 L
.60198 .14833 L
.6015 .14549 L
.59996 .13578 L
.59857 .12607 L
.59732 .11636 L
.59611 .10664 L
.59652 .09693 L
.5964 .08722 L
.59686 .07751 L
.59859 .06779 L
F
0 g
.60198 .06285 m
.60984 .06779 L
.61169 .07024 L
.6163 .07751 L
.62141 .0866 L
.62182 .08722 L
.62858 .09693 L
.63112 .10534 L
.63185 .10664 L
.6373 .11636 L
.64083 .12438 L
.64176 .12607 L
.64707 .13578 L
.65054 .14282 L
.65204 .14549 L
.65752 .15521 L
.66026 .16032 L
.66298 .16492 L
.66875 .17463 L
.66997 .17673 L
.67479 .18434 L
.67968 .19203 L
.68104 .19406 L
.68768 .20377 L
.68939 .20625 L
.69465 .21348 L
.69911 .21948 L
.70198 .22319 L
.70882 .23179 L
.70974 .23291 L
.71799 .24262 L
.71853 .24324 L
.72677 .25233 L
.72824 .25391 L
.73611 .26204 L
.73796 .26389 L
.74609 .27176 L
.74767 .27323 L
.75676 .28147 L
.75738 .28201 L
.76709 .29026 L
.76821 .29118 L
.77681 .29802 L
.78052 .30089 L
.78652 .30535 L
.79375 .31061 L
.79623 .31232 L
.80594 .31896 L
.80797 .32032 L
.81566 .32521 L
Mistroke
.82327 .33003 L
.82537 .33125 L
.83508 .33702 L
.83968 .33974 L
.84479 .34248 L
.85451 .34796 L
.85718 .34946 L
.86422 .35293 L
.87393 .35824 L
.87562 .35917 L
.88364 .3627 L
.89336 .36815 L
.89466 .36888 L
.90307 .37142 L
.91278 .37818 L
.9134 .37859 L
.92249 .3837 L
.92976 .38831 L
.93221 .39016 L
.93715 .39802 L
.93221 .40141 L
.92249 .40314 L
.91278 .4036 L
.90307 .40348 L
.89336 .40389 L
.88364 .40268 L
.87393 .40143 L
.86422 .40004 L
.85451 .3985 L
.85167 .39802 L
.84479 .39694 L
.83508 .39509 L
.82537 .39296 L
.81566 .3906 L
.80698 .38831 L
.80594 .38804 L
.79623 .38531 L
.78652 .38224 L
.77681 .37887 L
.77606 .37859 L
.76709 .37525 L
.75738 .37123 L
.7521 .36888 L
.74767 .36685 L
.73796 .36202 L
.73265 .35917 L
.72824 .35671 L
.71853 .35082 L
.71642 .34946 L
.70882 .34431 L
Mistroke
.7026 .33974 L
.69911 .33705 L
.69068 .33003 L
.68939 .3289 L
.68027 .32032 L
.67968 .31973 L
.6711 .31061 L
.66997 .30932 L
.66295 .30089 L
.66026 .2974 L
.65569 .29118 L
.65054 .28358 L
.64918 .28147 L
.64329 .27176 L
.64083 .26735 L
.63798 .26204 L
.63315 .25233 L
.63112 .2479 L
.62877 .24262 L
.62475 .23291 L
.62141 .22394 L
.62113 .22319 L
.61776 .21348 L
.61469 .20377 L
.61196 .19406 L
.61169 .19302 L
.6094 .18434 L
.60704 .17463 L
.60491 .16492 L
.60306 .15521 L
.60198 .14833 L
.6015 .14549 L
.59996 .13578 L
.59857 .12607 L
.59732 .11636 L
.59611 .10664 L
.59652 .09693 L
.5964 .08722 L
.59686 .07751 L
.59859 .06779 L
.60198 .06285 L
Mfstroke
.8 1 0 r
.98077 .40023 m
.97729 .39802 L
.97106 .39462 L
.96134 .38989 L
.95792 .38831 L
.95163 .38478 L
.94192 .37977 L
.93581 .37859 L
.93221 .37689 L
.92249 .37236 L
.91344 .36888 L
.91278 .36858 L
.90307 .36406 L
.89336 .35977 L
.89188 .35917 L
.88364 .3552 L
.87393 .35055 L
.87155 .34946 L
.86422 .34571 L
.85451 .34071 L
.85261 .33974 L
.84479 .33545 L
.83508 .33004 L
.83507 .33003 L
.82537 .32427 L
.81886 .32032 L
.81566 .31829 L
.80594 .31197 L
.80388 .31061 L
.79623 .30531 L
.79004 .30089 L
.78652 .29828 L
.77723 .29118 L
.77681 .29085 L
.76709 .28293 L
.76536 .28147 L
.75738 .27451 L
.75432 .27176 L
.74767 .26554 L
.74405 .26204 L
.73796 .25595 L
.73446 .25233 L
.72824 .24568 L
.72549 .24262 L
.71853 .23464 L
.71707 .23291 L
.70915 .22319 L
.70882 .22277 L
.70172 .21348 L
.69911 .20996 L
.69469 .20377 L
.68939 .19612 L
.68803 .19406 L
.68171 .18434 L
.67968 .18114 L
.67573 .17463 L
.66997 .16493 L
.66996 .16492 L
.66455 .15521 L
.66026 .14739 L
.65929 .14549 L
.65429 .13578 L
.65054 .12845 L
.64945 .12607 L
.6448 .11636 L
.64083 .10812 L
.64023 .10664 L
.63594 .09693 L
.63142 .08722 L
.63112 .08656 L
.62764 .07751 L
.62311 .06779 L
.62141 .06419 L
.62023 .05808 L
.61522 .04837 L
.61169 .04208 L
.61011 .03866 L
.60538 .02894 L
.60198 .02271 L
.59977 .01923 L
.98077 .01923 L
F
0 g
.98077 .40023 m
.97729 .39802 L
.97106 .39462 L
.96134 .38989 L
.95792 .38831 L
.95163 .38478 L
.94192 .37977 L
.93581 .37859 L
.93221 .37689 L
.92249 .37236 L
.91344 .36888 L
.91278 .36858 L
.90307 .36406 L
.89336 .35977 L
.89188 .35917 L
.88364 .3552 L
.87393 .35055 L
.87155 .34946 L
.86422 .34571 L
.85451 .34071 L
.85261 .33974 L
.84479 .33545 L
.83508 .33004 L
.83507 .33003 L
.82537 .32427 L
.81886 .32032 L
.81566 .31829 L
.80594 .31197 L
.80388 .31061 L
.79623 .30531 L
.79004 .30089 L
.78652 .29828 L
.77723 .29118 L
.77681 .29085 L
.76709 .28293 L
.76536 .28147 L
.75738 .27451 L
.75432 .27176 L
.74767 .26554 L
.74405 .26204 L
.73796 .25595 L
.73446 .25233 L
.72824 .24568 L
.72549 .24262 L
.71853 .23464 L
.71707 .23291 L
.70915 .22319 L
.70882 .22277 L
.70172 .21348 L
.69911 .20996 L
Mistroke
.69469 .20377 L
.68939 .19612 L
.68803 .19406 L
.68171 .18434 L
.67968 .18114 L
.67573 .17463 L
.66997 .16493 L
.66996 .16492 L
.66455 .15521 L
.66026 .14739 L
.65929 .14549 L
.65429 .13578 L
.65054 .12845 L
.64945 .12607 L
.6448 .11636 L
.64083 .10812 L
.64023 .10664 L
.63594 .09693 L
.63142 .08722 L
.63112 .08656 L
.62764 .07751 L
.62311 .06779 L
.62141 .06419 L
.62023 .05808 L
.61522 .04837 L
.61169 .04208 L
.61011 .03866 L
.60538 .02894 L
.60198 .02271 L
.59977 .01923 L
Mfstroke
.8 1 0 r
.98077 .59977 m
.97729 .60198 L
.97106 .60538 L
.96134 .61011 L
.95792 .61169 L
.95163 .61522 L
.94192 .62023 L
.93581 .62141 L
.93221 .62311 L
.92249 .62764 L
.91344 .63112 L
.91278 .63142 L
.90307 .63594 L
.89336 .64023 L
.89188 .64083 L
.88364 .6448 L
.87393 .64945 L
.87155 .65054 L
.86422 .65429 L
.85451 .65929 L
.85261 .66026 L
.84479 .66455 L
.83508 .66996 L
.83507 .66997 L
.82537 .67573 L
.81886 .67968 L
.81566 .68171 L
.80594 .68803 L
.80388 .68939 L
.79623 .69469 L
.79004 .69911 L
.78652 .70172 L
.77723 .70882 L
.77681 .70915 L
.76709 .71707 L
.76536 .71853 L
.75738 .72549 L
.75432 .72824 L
.74767 .73446 L
.74405 .73796 L
.73796 .74405 L
.73446 .74767 L
.72824 .75432 L
.72549 .75738 L
.71853 .76536 L
.71707 .76709 L
.70915 .77681 L
.70882 .77723 L
.70172 .78652 L
.69911 .79004 L
.69469 .79623 L
.68939 .80388 L
.68803 .80594 L
.68171 .81566 L
.67968 .81886 L
.67573 .82537 L
.66997 .83507 L
.66996 .83508 L
.66455 .84479 L
.66026 .85261 L
.65929 .85451 L
.65429 .86422 L
.65054 .87155 L
.64945 .87393 L
.6448 .88364 L
.64083 .89188 L
.64023 .89336 L
.63594 .90307 L
.63142 .91278 L
.63112 .91344 L
.62764 .92249 L
.62311 .93221 L
.62141 .93581 L
.62023 .94192 L
.61522 .95163 L
.61169 .95792 L
.61011 .96134 L
.60538 .97106 L
.60198 .97729 L
.59977 .98077 L
.98077 .98077 L
F
0 g
.98077 .59977 m
.97729 .60198 L
.97106 .60538 L
.96134 .61011 L
.95792 .61169 L
.95163 .61522 L
.94192 .62023 L
.93581 .62141 L
.93221 .62311 L
.92249 .62764 L
.91344 .63112 L
.91278 .63142 L
.90307 .63594 L
.89336 .64023 L
.89188 .64083 L
.88364 .6448 L
.87393 .64945 L
.87155 .65054 L
.86422 .65429 L
.85451 .65929 L
.85261 .66026 L
.84479 .66455 L
.83508 .66996 L
.83507 .66997 L
.82537 .67573 L
.81886 .67968 L
.81566 .68171 L
.80594 .68803 L
.80388 .68939 L
.79623 .69469 L
.79004 .69911 L
.78652 .70172 L
.77723 .70882 L
.77681 .70915 L
.76709 .71707 L
.76536 .71853 L
.75738 .72549 L
.75432 .72824 L
.74767 .73446 L
.74405 .73796 L
.73796 .74405 L
.73446 .74767 L
.72824 .75432 L
.72549 .75738 L
.71853 .76536 L
.71707 .76709 L
.70915 .77681 L
.70882 .77723 L
.70172 .78652 L
.69911 .79004 L
Mistroke
.69469 .79623 L
.68939 .80388 L
.68803 .80594 L
.68171 .81566 L
.67968 .81886 L
.67573 .82537 L
.66997 .83507 L
.66996 .83508 L
.66455 .84479 L
.66026 .85261 L
.65929 .85451 L
.65429 .86422 L
.65054 .87155 L
.64945 .87393 L
.6448 .88364 L
.64083 .89188 L
.64023 .89336 L
.63594 .90307 L
.63142 .91278 L
.63112 .91344 L
.62764 .92249 L
.62311 .93221 L
.62141 .93581 L
.62023 .94192 L
.61522 .95163 L
.61169 .95792 L
.61011 .96134 L
.60538 .97106 L
.60198 .97729 L
.59977 .98077 L
Mfstroke
1 .48 0 r
.85451 .61093 m
.86422 .60995 L
.87393 .61039 L
.88291 .61169 L
.88364 .61194 L
.88462 .62141 L
.88364 .62237 L
.87393 .62992 L
.87219 .63112 L
.86422 .63624 L
.85682 .64083 L
.85451 .64248 L
.84479 .64916 L
.84085 .65054 L
.83508 .65438 L
.82537 .66015 L
.82513 .66026 L
.81566 .66657 L
.8101 .66997 L
.80594 .67283 L
.79623 .67944 L
.79586 .67968 L
.78652 .6865 L
.78249 .68939 L
.77681 .69377 L
.76994 .69911 L
.76709 .70143 L
.75819 .70882 L
.75738 .70951 L
.74767 .7181 L
.74719 .71853 L
.73796 .72721 L
.73689 .72824 L
.72824 .73689 L
.72721 .73796 L
.71853 .74719 L
.7181 .74767 L
.70951 .75738 L
.70882 .75819 L
.70143 .76709 L
.69911 .76994 L
.69377 .77681 L
.68939 .78249 L
.6865 .78652 L
.67968 .79586 L
.67944 .79623 L
.67283 .80594 L
.66997 .8101 L
.66657 .81566 L
.66026 .82513 L
.66015 .82537 L
.65438 .83508 L
.65054 .84085 L
.64916 .84479 L
.64248 .85451 L
.64083 .85682 L
.63624 .86422 L
.63112 .87219 L
.62992 .87393 L
.62237 .88364 L
.62141 .88462 L
.61194 .88364 L
.61169 .88291 L
.61039 .87393 L
.60995 .86422 L
.61093 .85451 L
.61169 .84703 L
.61216 .84479 L
.61302 .83508 L
.61503 .82537 L
.61732 .81566 L
.61985 .80594 L
.62141 .80036 L
.6225 .79623 L
.62548 .78652 L
.62889 .77681 L
.63112 .77091 L
.63257 .76709 L
.63664 .75738 L
.64083 .74845 L
.6412 .74767 L
.64616 .73796 L
.65054 .73023 L
.65171 .72824 L
.65785 .71853 L
.66026 .71501 L
.66469 .70882 L
.66997 .70204 L
.67237 .69911 L
.67968 .69083 L
.68103 .68939 L
.68939 .68103 L
.69083 .67968 L
.69911 .67237 L
.70204 .66997 L
.70882 .66469 L
.71501 .66026 L
.71853 .65785 L
.72824 .65171 L
.73023 .65054 L
.73796 .64616 L
.74767 .6412 L
.74845 .64083 L
.75738 .63664 L
.76709 .63257 L
.77091 .63112 L
.77681 .62889 L
.78652 .62548 L
.79623 .6225 L
.80036 .62141 L
.80594 .61985 L
.81566 .61732 L
.82537 .61503 L
.83508 .61302 L
.84479 .61216 L
.84703 .61169 L
F
0 g
.85451 .61093 m
.86422 .60995 L
.87393 .61039 L
.88291 .61169 L
.88364 .61194 L
.88462 .62141 L
.88364 .62237 L
.87393 .62992 L
.87219 .63112 L
.86422 .63624 L
.85682 .64083 L
.85451 .64248 L
.84479 .64916 L
.84085 .65054 L
.83508 .65438 L
.82537 .66015 L
.82513 .66026 L
.81566 .66657 L
.8101 .66997 L
.80594 .67283 L
.79623 .67944 L
.79586 .67968 L
.78652 .6865 L
.78249 .68939 L
.77681 .69377 L
.76994 .69911 L
.76709 .70143 L
.75819 .70882 L
.75738 .70951 L
.74767 .7181 L
.74719 .71853 L
.73796 .72721 L
.73689 .72824 L
.72824 .73689 L
.72721 .73796 L
.71853 .74719 L
.7181 .74767 L
.70951 .75738 L
.70882 .75819 L
.70143 .76709 L
.69911 .76994 L
.69377 .77681 L
.68939 .78249 L
.6865 .78652 L
.67968 .79586 L
.67944 .79623 L
.67283 .80594 L
.66997 .8101 L
.66657 .81566 L
.66026 .82513 L
Mistroke
.66015 .82537 L
.65438 .83508 L
.65054 .84085 L
.64916 .84479 L
.64248 .85451 L
.64083 .85682 L
.63624 .86422 L
.63112 .87219 L
.62992 .87393 L
.62237 .88364 L
.62141 .88462 L
.61194 .88364 L
.61169 .88291 L
.61039 .87393 L
.60995 .86422 L
.61093 .85451 L
.61169 .84703 L
.61216 .84479 L
.61302 .83508 L
.61503 .82537 L
.61732 .81566 L
.61985 .80594 L
.62141 .80036 L
.6225 .79623 L
.62548 .78652 L
.62889 .77681 L
.63112 .77091 L
.63257 .76709 L
.63664 .75738 L
.64083 .74845 L
.6412 .74767 L
.64616 .73796 L
.65054 .73023 L
.65171 .72824 L
.65785 .71853 L
.66026 .71501 L
.66469 .70882 L
.66997 .70204 L
.67237 .69911 L
.67968 .69083 L
.68103 .68939 L
.68939 .68103 L
.69083 .67968 L
.69911 .67237 L
.70204 .66997 L
.70882 .66469 L
.71501 .66026 L
.71853 .65785 L
.72824 .65171 L
.73023 .65054 L
Mistroke
.73796 .64616 L
.74767 .6412 L
.74845 .64083 L
.75738 .63664 L
.76709 .63257 L
.77091 .63112 L
.77681 .62889 L
.78652 .62548 L
.79623 .6225 L
.80036 .62141 L
.80594 .61985 L
.81566 .61732 L
.82537 .61503 L
.83508 .61302 L
.84479 .61216 L
.84703 .61169 L
.85451 .61093 L
Mfstroke
1 .48 0 r
.62141 .11538 m
.62237 .11636 L
.62992 .12607 L
.63112 .12781 L
.63624 .13578 L
.64083 .14318 L
.64248 .14549 L
.64916 .15521 L
.65054 .15915 L
.65438 .16492 L
.66015 .17463 L
.66026 .17487 L
.66657 .18434 L
.66997 .1899 L
.67283 .19406 L
.67944 .20377 L
.67968 .20414 L
.6865 .21348 L
.68939 .21751 L
.69377 .22319 L
.69911 .23006 L
.70143 .23291 L
.70882 .24181 L
.70951 .24262 L
.7181 .25233 L
.71853 .25281 L
.72721 .26204 L
.72824 .26311 L
.73689 .27176 L
.73796 .27279 L
.74719 .28147 L
.74767 .2819 L
.75738 .29049 L
.75819 .29118 L
.76709 .29857 L
.76994 .30089 L
.77681 .30623 L
.78249 .31061 L
.78652 .3135 L
.79586 .32032 L
.79623 .32056 L
.80594 .32717 L
.8101 .33003 L
.81566 .33343 L
.82513 .33974 L
.82537 .33985 L
.83508 .34562 L
.84085 .34946 L
.84479 .35084 L
.85451 .35752 L
.85682 .35917 L
.86422 .36376 L
.87219 .36888 L
.87393 .37008 L
.88364 .37763 L
.88462 .37859 L
.88364 .38806 L
.88291 .38831 L
.87393 .38961 L
.86422 .39005 L
.85451 .38907 L
.84703 .38831 L
.84479 .38784 L
.83508 .38698 L
.82537 .38497 L
.81566 .38268 L
.80594 .38015 L
.80036 .37859 L
.79623 .3775 L
.78652 .37452 L
.77681 .37111 L
.77091 .36888 L
.76709 .36743 L
.75738 .36336 L
.74845 .35917 L
.74767 .3588 L
.73796 .35384 L
.73023 .34946 L
.72824 .34829 L
.71853 .34215 L
.71501 .33974 L
.70882 .33531 L
.70204 .33003 L
.69911 .32763 L
.69083 .32032 L
.68939 .31897 L
.68103 .31061 L
.67968 .30917 L
.67237 .30089 L
.66997 .29796 L
.66469 .29118 L
.66026 .28499 L
.65785 .28147 L
.65171 .27176 L
.65054 .26977 L
.64616 .26204 L
.6412 .25233 L
.64083 .25155 L
.63664 .24262 L
.63257 .23291 L
.63112 .22909 L
.62889 .22319 L
.62548 .21348 L
.6225 .20377 L
.62141 .19964 L
.61985 .19406 L
.61732 .18434 L
.61503 .17463 L
.61302 .16492 L
.61216 .15521 L
.61169 .15297 L
.61093 .14549 L
.60995 .13578 L
.61039 .12607 L
.61169 .11709 L
.61194 .11636 L
F
0 g
.62141 .11538 m
.62237 .11636 L
.62992 .12607 L
.63112 .12781 L
.63624 .13578 L
.64083 .14318 L
.64248 .14549 L
.64916 .15521 L
.65054 .15915 L
.65438 .16492 L
.66015 .17463 L
.66026 .17487 L
.66657 .18434 L
.66997 .1899 L
.67283 .19406 L
.67944 .20377 L
.67968 .20414 L
.6865 .21348 L
.68939 .21751 L
.69377 .22319 L
.69911 .23006 L
.70143 .23291 L
.70882 .24181 L
.70951 .24262 L
.7181 .25233 L
.71853 .25281 L
.72721 .26204 L
.72824 .26311 L
.73689 .27176 L
.73796 .27279 L
.74719 .28147 L
.74767 .2819 L
.75738 .29049 L
.75819 .29118 L
.76709 .29857 L
.76994 .30089 L
.77681 .30623 L
.78249 .31061 L
.78652 .3135 L
.79586 .32032 L
.79623 .32056 L
.80594 .32717 L
.8101 .33003 L
.81566 .33343 L
.82513 .33974 L
.82537 .33985 L
.83508 .34562 L
.84085 .34946 L
.84479 .35084 L
.85451 .35752 L
Mistroke
.85682 .35917 L
.86422 .36376 L
.87219 .36888 L
.87393 .37008 L
.88364 .37763 L
.88462 .37859 L
.88364 .38806 L
.88291 .38831 L
.87393 .38961 L
.86422 .39005 L
.85451 .38907 L
.84703 .38831 L
.84479 .38784 L
.83508 .38698 L
.82537 .38497 L
.81566 .38268 L
.80594 .38015 L
.80036 .37859 L
.79623 .3775 L
.78652 .37452 L
.77681 .37111 L
.77091 .36888 L
.76709 .36743 L
.75738 .36336 L
.74845 .35917 L
.74767 .3588 L
.73796 .35384 L
.73023 .34946 L
.72824 .34829 L
.71853 .34215 L
.71501 .33974 L
.70882 .33531 L
.70204 .33003 L
.69911 .32763 L
.69083 .32032 L
.68939 .31897 L
.68103 .31061 L
.67968 .30917 L
.67237 .30089 L
.66997 .29796 L
.66469 .29118 L
.66026 .28499 L
.65785 .28147 L
.65171 .27176 L
.65054 .26977 L
.64616 .26204 L
.6412 .25233 L
.64083 .25155 L
.63664 .24262 L
.63257 .23291 L
Mistroke
.63112 .22909 L
.62889 .22319 L
.62548 .21348 L
.6225 .20377 L
.62141 .19964 L
.61985 .19406 L
.61732 .18434 L
.61503 .17463 L
.61302 .16492 L
.61216 .15521 L
.61169 .15297 L
.61093 .14549 L
.60995 .13578 L
.61039 .12607 L
.61169 .11709 L
.61194 .11636 L
.62141 .11538 L
Mfstroke
.56 1 0 r
.98077 .38779 m
.97106 .38388 L
.96134 .38026 L
.95587 .37859 L
.95163 .37692 L
.94192 .37309 L
.93221 .36944 L
.93062 .36888 L
.92249 .36554 L
.91278 .36153 L
.90691 .35917 L
.90307 .3575 L
.89336 .3532 L
.88493 .34946 L
.88364 .34886 L
.87393 .34423 L
.86469 .33974 L
.86422 .33951 L
.85451 .33449 L
.8461 .33003 L
.84479 .32931 L
.83508 .32383 L
.82904 .32032 L
.82537 .3181 L
.81566 .31206 L
.81337 .31061 L
.80594 .30569 L
.79895 .30089 L
.79623 .29897 L
.78652 .29184 L
.78565 .29118 L
.77681 .28426 L
.77337 .28147 L
.76709 .2762 L
.76198 .27176 L
.75738 .26761 L
.75141 .26204 L
.74767 .25843 L
.74157 .25233 L
.73796 .24859 L
.73239 .24262 L
.72824 .23802 L
.7238 .23291 L
.71853 .22663 L
.71574 .22319 L
.70882 .21435 L
.70816 .21348 L
.70103 .20377 L
.69911 .20105 L
.69431 .19406 L
.68939 .18663 L
.68794 .18434 L
.6819 .17463 L
.67968 .17096 L
.67617 .16492 L
.67069 .15521 L
.66997 .1539 L
.66551 .14549 L
.66049 .13578 L
.66026 .13531 L
.65577 .12607 L
.65114 .11636 L
.65054 .11507 L
.6468 .10664 L
.6425 .09693 L
.64083 .09309 L
.63847 .08722 L
.63446 .07751 L
.63112 .06938 L
.63056 .06779 L
.62691 .05808 L
.62308 .04837 L
.62141 .04413 L
.61974 .03866 L
.61612 .02894 L
.61221 .01923 L
.98077 .01923 L
F
0 g
.98077 .38779 m
.97106 .38388 L
.96134 .38026 L
.95587 .37859 L
.95163 .37692 L
.94192 .37309 L
.93221 .36944 L
.93062 .36888 L
.92249 .36554 L
.91278 .36153 L
.90691 .35917 L
.90307 .3575 L
.89336 .3532 L
.88493 .34946 L
.88364 .34886 L
.87393 .34423 L
.86469 .33974 L
.86422 .33951 L
.85451 .33449 L
.8461 .33003 L
.84479 .32931 L
.83508 .32383 L
.82904 .32032 L
.82537 .3181 L
.81566 .31206 L
.81337 .31061 L
.80594 .30569 L
.79895 .30089 L
.79623 .29897 L
.78652 .29184 L
.78565 .29118 L
.77681 .28426 L
.77337 .28147 L
.76709 .2762 L
.76198 .27176 L
.75738 .26761 L
.75141 .26204 L
.74767 .25843 L
.74157 .25233 L
.73796 .24859 L
.73239 .24262 L
.72824 .23802 L
.7238 .23291 L
.71853 .22663 L
.71574 .22319 L
.70882 .21435 L
.70816 .21348 L
.70103 .20377 L
.69911 .20105 L
.69431 .19406 L
Mistroke
.68939 .18663 L
.68794 .18434 L
.6819 .17463 L
.67968 .17096 L
.67617 .16492 L
.67069 .15521 L
.66997 .1539 L
.66551 .14549 L
.66049 .13578 L
.66026 .13531 L
.65577 .12607 L
.65114 .11636 L
.65054 .11507 L
.6468 .10664 L
.6425 .09693 L
.64083 .09309 L
.63847 .08722 L
.63446 .07751 L
.63112 .06938 L
.63056 .06779 L
.62691 .05808 L
.62308 .04837 L
.62141 .04413 L
.61974 .03866 L
.61612 .02894 L
.61221 .01923 L
Mfstroke
.56 1 0 r
.98077 .61221 m
.97106 .61612 L
.96134 .61974 L
.95587 .62141 L
.95163 .62308 L
.94192 .62691 L
.93221 .63056 L
.93062 .63112 L
.92249 .63446 L
.91278 .63847 L
.90691 .64083 L
.90307 .6425 L
.89336 .6468 L
.88493 .65054 L
.88364 .65114 L
.87393 .65577 L
.86469 .66026 L
.86422 .66049 L
.85451 .66551 L
.8461 .66997 L
.84479 .67069 L
.83508 .67617 L
.82904 .67968 L
.82537 .6819 L
.81566 .68794 L
.81337 .68939 L
.80594 .69431 L
.79895 .69911 L
.79623 .70103 L
.78652 .70816 L
.78565 .70882 L
.77681 .71574 L
.77337 .71853 L
.76709 .7238 L
.76198 .72824 L
.75738 .73239 L
.75141 .73796 L
.74767 .74157 L
.74157 .74767 L
.73796 .75141 L
.73239 .75738 L
.72824 .76198 L
.7238 .76709 L
.71853 .77337 L
.71574 .77681 L
.70882 .78565 L
.70816 .78652 L
.70103 .79623 L
.69911 .79895 L
.69431 .80594 L
.68939 .81337 L
.68794 .81566 L
.6819 .82537 L
.67968 .82904 L
.67617 .83508 L
.67069 .84479 L
.66997 .8461 L
.66551 .85451 L
.66049 .86422 L
.66026 .86469 L
.65577 .87393 L
.65114 .88364 L
.65054 .88493 L
.6468 .89336 L
.6425 .90307 L
.64083 .90691 L
.63847 .91278 L
.63446 .92249 L
.63112 .93062 L
.63056 .93221 L
.62691 .94192 L
.62308 .95163 L
.62141 .95587 L
.61974 .96134 L
.61612 .97106 L
.61221 .98077 L
.98077 .98077 L
F
0 g
.98077 .61221 m
.97106 .61612 L
.96134 .61974 L
.95587 .62141 L
.95163 .62308 L
.94192 .62691 L
.93221 .63056 L
.93062 .63112 L
.92249 .63446 L
.91278 .63847 L
.90691 .64083 L
.90307 .6425 L
.89336 .6468 L
.88493 .65054 L
.88364 .65114 L
.87393 .65577 L
.86469 .66026 L
.86422 .66049 L
.85451 .66551 L
.8461 .66997 L
.84479 .67069 L
.83508 .67617 L
.82904 .67968 L
.82537 .6819 L
.81566 .68794 L
.81337 .68939 L
.80594 .69431 L
.79895 .69911 L
.79623 .70103 L
.78652 .70816 L
.78565 .70882 L
.77681 .71574 L
.77337 .71853 L
.76709 .7238 L
.76198 .72824 L
.75738 .73239 L
.75141 .73796 L
.74767 .74157 L
.74157 .74767 L
.73796 .75141 L
.73239 .75738 L
.72824 .76198 L
.7238 .76709 L
.71853 .77337 L
.71574 .77681 L
.70882 .78565 L
.70816 .78652 L
.70103 .79623 L
.69911 .79895 L
.69431 .80594 L
Mistroke
.68939 .81337 L
.68794 .81566 L
.6819 .82537 L
.67968 .82904 L
.67617 .83508 L
.67069 .84479 L
.66997 .8461 L
.66551 .85451 L
.66049 .86422 L
.66026 .86469 L
.65577 .87393 L
.65114 .88364 L
.65054 .88493 L
.6468 .89336 L
.6425 .90307 L
.64083 .90691 L
.63847 .91278 L
.63446 .92249 L
.63112 .93062 L
.63056 .93221 L
.62691 .94192 L
.62308 .95163 L
.62141 .95587 L
.61974 .96134 L
.61612 .97106 L
.61221 .98077 L
Mfstroke
.32 1 0 r
.98077 .38035 m
.97501 .37859 L
.97106 .37723 L
.96134 .37382 L
.95163 .37043 L
.9471 .36888 L
.94192 .36696 L
.93221 .36329 L
.92249 .3596 L
.92136 .35917 L
.91278 .3557 L
.90307 .35167 L
.89781 .34946 L
.89336 .3475 L
.88364 .34313 L
.87631 .33974 L
.87393 .3386 L
.86422 .33383 L
.85672 .33003 L
.85451 .32887 L
.84479 .32364 L
.83883 .32032 L
.83508 .31816 L
.82537 .31238 L
.82248 .31061 L
.81566 .30628 L
.80749 .30089 L
.80594 .29984 L
.79623 .29301 L
.79371 .29118 L
.78652 .28577 L
.78102 .28147 L
.77681 .27806 L
.76929 .27176 L
.76709 .26985 L
.75842 .26204 L
.75738 .26107 L
.74833 .25233 L
.74767 .25167 L
.73893 .24262 L
.73796 .24158 L
.73015 .23291 L
.72824 .23071 L
.72194 .22319 L
.71853 .21898 L
.71423 .21348 L
.70882 .20629 L
.70699 .20377 L
.70016 .19406 L
.69911 .19251 L
.69372 .18434 L
.68939 .17752 L
.68762 .17463 L
.68184 .16492 L
.67968 .16117 L
.67636 .15521 L
.67113 .14549 L
.66997 .14328 L
.66617 .13578 L
.6614 .12607 L
.66026 .12369 L
.65687 .11636 L
.6525 .10664 L
.65054 .10219 L
.64833 .09693 L
.6443 .08722 L
.64083 .07864 L
.6404 .07751 L
.63671 .06779 L
.63304 .05808 L
.63112 .0529 L
.62957 .04837 L
.62618 .03866 L
.62277 .02894 L
.62141 .02499 L
.61965 .01923 L
.98077 .01923 L
F
0 g
.98077 .38035 m
.97501 .37859 L
.97106 .37723 L
.96134 .37382 L
.95163 .37043 L
.9471 .36888 L
.94192 .36696 L
.93221 .36329 L
.92249 .3596 L
.92136 .35917 L
.91278 .3557 L
.90307 .35167 L
.89781 .34946 L
.89336 .3475 L
.88364 .34313 L
.87631 .33974 L
.87393 .3386 L
.86422 .33383 L
.85672 .33003 L
.85451 .32887 L
.84479 .32364 L
.83883 .32032 L
.83508 .31816 L
.82537 .31238 L
.82248 .31061 L
.81566 .30628 L
.80749 .30089 L
.80594 .29984 L
.79623 .29301 L
.79371 .29118 L
.78652 .28577 L
.78102 .28147 L
.77681 .27806 L
.76929 .27176 L
.76709 .26985 L
.75842 .26204 L
.75738 .26107 L
.74833 .25233 L
.74767 .25167 L
.73893 .24262 L
.73796 .24158 L
.73015 .23291 L
.72824 .23071 L
.72194 .22319 L
.71853 .21898 L
.71423 .21348 L
.70882 .20629 L
.70699 .20377 L
.70016 .19406 L
.69911 .19251 L
Mistroke
.69372 .18434 L
.68939 .17752 L
.68762 .17463 L
.68184 .16492 L
.67968 .16117 L
.67636 .15521 L
.67113 .14549 L
.66997 .14328 L
.66617 .13578 L
.6614 .12607 L
.66026 .12369 L
.65687 .11636 L
.6525 .10664 L
.65054 .10219 L
.64833 .09693 L
.6443 .08722 L
.64083 .07864 L
.6404 .07751 L
.63671 .06779 L
.63304 .05808 L
.63112 .0529 L
.62957 .04837 L
.62618 .03866 L
.62277 .02894 L
.62141 .02499 L
.61965 .01923 L
Mfstroke
.32 1 0 r
.98077 .61965 m
.97501 .62141 L
.97106 .62277 L
.96134 .62618 L
.95163 .62957 L
.9471 .63112 L
.94192 .63304 L
.93221 .63671 L
.92249 .6404 L
.92136 .64083 L
.91278 .6443 L
.90307 .64833 L
.89781 .65054 L
.89336 .6525 L
.88364 .65687 L
.87631 .66026 L
.87393 .6614 L
.86422 .66617 L
.85672 .66997 L
.85451 .67113 L
.84479 .67636 L
.83883 .67968 L
.83508 .68184 L
.82537 .68762 L
.82248 .68939 L
.81566 .69372 L
.80749 .69911 L
.80594 .70016 L
.79623 .70699 L
.79371 .70882 L
.78652 .71423 L
.78102 .71853 L
.77681 .72194 L
.76929 .72824 L
.76709 .73015 L
.75842 .73796 L
.75738 .73893 L
.74833 .74767 L
.74767 .74833 L
.73893 .75738 L
.73796 .75842 L
.73015 .76709 L
.72824 .76929 L
.72194 .77681 L
.71853 .78102 L
.71423 .78652 L
.70882 .79371 L
.70699 .79623 L
.70016 .80594 L
.69911 .80749 L
.69372 .81566 L
.68939 .82248 L
.68762 .82537 L
.68184 .83508 L
.67968 .83883 L
.67636 .84479 L
.67113 .85451 L
.66997 .85672 L
.66617 .86422 L
.6614 .87393 L
.66026 .87631 L
.65687 .88364 L
.6525 .89336 L
.65054 .89781 L
.64833 .90307 L
.6443 .91278 L
.64083 .92136 L
.6404 .92249 L
.63671 .93221 L
.63304 .94192 L
.63112 .9471 L
.62957 .95163 L
.62618 .96134 L
.62277 .97106 L
.62141 .97501 L
.61965 .98077 L
.98077 .98077 L
F
0 g
.98077 .61965 m
.97501 .62141 L
.97106 .62277 L
.96134 .62618 L
.95163 .62957 L
.9471 .63112 L
.94192 .63304 L
.93221 .63671 L
.92249 .6404 L
.92136 .64083 L
.91278 .6443 L
.90307 .64833 L
.89781 .65054 L
.89336 .6525 L
.88364 .65687 L
.87631 .66026 L
.87393 .6614 L
.86422 .66617 L
.85672 .66997 L
.85451 .67113 L
.84479 .67636 L
.83883 .67968 L
.83508 .68184 L
.82537 .68762 L
.82248 .68939 L
.81566 .69372 L
.80749 .69911 L
.80594 .70016 L
.79623 .70699 L
.79371 .70882 L
.78652 .71423 L
.78102 .71853 L
.77681 .72194 L
.76929 .72824 L
.76709 .73015 L
.75842 .73796 L
.75738 .73893 L
.74833 .74767 L
.74767 .74833 L
.73893 .75738 L
.73796 .75842 L
.73015 .76709 L
.72824 .76929 L
.72194 .77681 L
.71853 .78102 L
.71423 .78652 L
.70882 .79371 L
.70699 .79623 L
.70016 .80594 L
.69911 .80749 L
Mistroke
.69372 .81566 L
.68939 .82248 L
.68762 .82537 L
.68184 .83508 L
.67968 .83883 L
.67636 .84479 L
.67113 .85451 L
.66997 .85672 L
.66617 .86422 L
.6614 .87393 L
.66026 .87631 L
.65687 .88364 L
.6525 .89336 L
.65054 .89781 L
.64833 .90307 L
.6443 .91278 L
.64083 .92136 L
.6404 .92249 L
.63671 .93221 L
.63304 .94192 L
.63112 .9471 L
.62957 .95163 L
.62618 .96134 L
.62277 .97106 L
.62141 .97501 L
.61965 .98077 L
Mfstroke
.08 1 0 r
.98077 .37443 m
.97106 .37132 L
.96346 .36888 L
.96134 .36817 L
.95163 .36482 L
.94192 .3614 L
.93568 .35917 L
.93221 .35787 L
.92249 .35417 L
.91278 .35035 L
.91053 .34946 L
.90307 .34636 L
.89336 .3422 L
.88777 .33974 L
.88364 .33787 L
.87393 .33331 L
.86716 .33003 L
.86422 .32856 L
.85451 .32356 L
.84844 .32032 L
.84479 .31831 L
.83508 .31278 L
.8314 .31061 L
.82537 .30694 L
.81583 .30089 L
.81566 .30078 L
.80594 .29424 L
.80158 .29118 L
.79623 .28731 L
.78847 .28147 L
.78652 .27995 L
.77681 .2721 L
.77639 .27176 L
.76709 .26372 L
.76523 .26204 L
.75738 .25475 L
.75488 .25233 L
.74767 .24512 L
.74525 .24262 L
.73796 .23477 L
.73628 .23291 L
.72824 .22361 L
.7279 .22319 L
.72005 .21348 L
.71853 .21153 L
.71269 .20377 L
.70882 .19842 L
.70576 .19406 L
.69922 .18434 L
.69911 .18417 L
.69306 .17463 L
.68939 .1686 L
.68722 .16492 L
.68169 .15521 L
.67968 .15156 L
.67644 .14549 L
.67144 .13578 L
.66997 .13284 L
.66669 .12607 L
.66213 .11636 L
.66026 .11223 L
.6578 .10664 L
.65364 .09693 L
.65054 .08947 L
.64965 .08722 L
.64583 .07751 L
.64213 .06779 L
.64083 .06432 L
.6386 .05808 L
.63518 .04837 L
.63183 .03866 L
.63112 .03654 L
.62868 .02894 L
.62557 .01923 L
.98077 .01923 L
F
0 g
.98077 .37443 m
.97106 .37132 L
.96346 .36888 L
.96134 .36817 L
.95163 .36482 L
.94192 .3614 L
.93568 .35917 L
.93221 .35787 L
.92249 .35417 L
.91278 .35035 L
.91053 .34946 L
.90307 .34636 L
.89336 .3422 L
.88777 .33974 L
.88364 .33787 L
.87393 .33331 L
.86716 .33003 L
.86422 .32856 L
.85451 .32356 L
.84844 .32032 L
.84479 .31831 L
.83508 .31278 L
.8314 .31061 L
.82537 .30694 L
.81583 .30089 L
.81566 .30078 L
.80594 .29424 L
.80158 .29118 L
.79623 .28731 L
.78847 .28147 L
.78652 .27995 L
.77681 .2721 L
.77639 .27176 L
.76709 .26372 L
.76523 .26204 L
.75738 .25475 L
.75488 .25233 L
.74767 .24512 L
.74525 .24262 L
.73796 .23477 L
.73628 .23291 L
.72824 .22361 L
.7279 .22319 L
.72005 .21348 L
.71853 .21153 L
.71269 .20377 L
.70882 .19842 L
.70576 .19406 L
.69922 .18434 L
.69911 .18417 L
Mistroke
.69306 .17463 L
.68939 .1686 L
.68722 .16492 L
.68169 .15521 L
.67968 .15156 L
.67644 .14549 L
.67144 .13578 L
.66997 .13284 L
.66669 .12607 L
.66213 .11636 L
.66026 .11223 L
.6578 .10664 L
.65364 .09693 L
.65054 .08947 L
.64965 .08722 L
.64583 .07751 L
.64213 .06779 L
.64083 .06432 L
.6386 .05808 L
.63518 .04837 L
.63183 .03866 L
.63112 .03654 L
.62868 .02894 L
.62557 .01923 L
Mfstroke
.08 1 0 r
.98077 .62557 m
.97106 .62868 L
.96346 .63112 L
.96134 .63183 L
.95163 .63518 L
.94192 .6386 L
.93568 .64083 L
.93221 .64213 L
.92249 .64583 L
.91278 .64965 L
.91053 .65054 L
.90307 .65364 L
.89336 .6578 L
.88777 .66026 L
.88364 .66213 L
.87393 .66669 L
.86716 .66997 L
.86422 .67144 L
.85451 .67644 L
.84844 .67968 L
.84479 .68169 L
.83508 .68722 L
.8314 .68939 L
.82537 .69306 L
.81583 .69911 L
.81566 .69922 L
.80594 .70576 L
.80158 .70882 L
.79623 .71269 L
.78847 .71853 L
.78652 .72005 L
.77681 .7279 L
.77639 .72824 L
.76709 .73628 L
.76523 .73796 L
.75738 .74525 L
.75488 .74767 L
.74767 .75488 L
.74525 .75738 L
.73796 .76523 L
.73628 .76709 L
.72824 .77639 L
.7279 .77681 L
.72005 .78652 L
.71853 .78847 L
.71269 .79623 L
.70882 .80158 L
.70576 .80594 L
.69922 .81566 L
.69911 .81583 L
.69306 .82537 L
.68939 .8314 L
.68722 .83508 L
.68169 .84479 L
.67968 .84844 L
.67644 .85451 L
.67144 .86422 L
.66997 .86716 L
.66669 .87393 L
.66213 .88364 L
.66026 .88777 L
.6578 .89336 L
.65364 .90307 L
.65054 .91053 L
.64965 .91278 L
.64583 .92249 L
.64213 .93221 L
.64083 .93568 L
.6386 .94192 L
.63518 .95163 L
.63183 .96134 L
.63112 .96346 L
.62868 .97106 L
.62557 .98077 L
.98077 .98077 L
F
0 g
.98077 .62557 m
.97106 .62868 L
.96346 .63112 L
.96134 .63183 L
.95163 .63518 L
.94192 .6386 L
.93568 .64083 L
.93221 .64213 L
.92249 .64583 L
.91278 .64965 L
.91053 .65054 L
.90307 .65364 L
.89336 .6578 L
.88777 .66026 L
.88364 .66213 L
.87393 .66669 L
.86716 .66997 L
.86422 .67144 L
.85451 .67644 L
.84844 .67968 L
.84479 .68169 L
.83508 .68722 L
.8314 .68939 L
.82537 .69306 L
.81583 .69911 L
.81566 .69922 L
.80594 .70576 L
.80158 .70882 L
.79623 .71269 L
.78847 .71853 L
.78652 .72005 L
.77681 .7279 L
.77639 .72824 L
.76709 .73628 L
.76523 .73796 L
.75738 .74525 L
.75488 .74767 L
.74767 .75488 L
.74525 .75738 L
.73796 .76523 L
.73628 .76709 L
.72824 .77639 L
.7279 .77681 L
.72005 .78652 L
.71853 .78847 L
.71269 .79623 L
.70882 .80158 L
.70576 .80594 L
.69922 .81566 L
.69911 .81583 L
Mistroke
.69306 .82537 L
.68939 .8314 L
.68722 .83508 L
.68169 .84479 L
.67968 .84844 L
.67644 .85451 L
.67144 .86422 L
.66997 .86716 L
.66669 .87393 L
.66213 .88364 L
.66026 .88777 L
.6578 .89336 L
.65364 .90307 L
.65054 .91053 L
.64965 .91278 L
.64583 .92249 L
.64213 .93221 L
.64083 .93568 L
.6386 .94192 L
.63518 .95163 L
.63183 .96134 L
.63112 .96346 L
.62868 .97106 L
.62557 .98077 L
Mfstroke
1 .24 0 r
.81566 .62921 m
.82537 .62933 L
.83325 .63112 L
.83088 .64083 L
.82537 .64592 L
.81984 .65054 L
.81566 .65376 L
.80708 .66026 L
.80594 .66125 L
.79623 .66833 L
.79409 .66997 L
.78652 .67622 L
.78133 .67968 L
.77681 .6834 L
.76903 .68939 L
.76709 .69102 L
.75738 .69906 L
.75732 .69911 L
.74767 .70763 L
.74628 .70882 L
.73796 .71655 L
.73582 .71853 L
.72824 .72592 L
.72592 .72824 L
.71853 .73582 L
.71655 .73796 L
.70882 .74628 L
.70763 .74767 L
.69911 .75732 L
.69906 .75738 L
.69102 .76709 L
.68939 .76903 L
.6834 .77681 L
.67968 .78133 L
.67622 .78652 L
.66997 .79409 L
.66833 .79623 L
.66125 .80594 L
.66026 .80708 L
.65376 .81566 L
.65054 .81984 L
.64592 .82537 L
.64083 .83088 L
.63112 .83325 L
.62933 .82537 L
.62921 .81566 L
.63112 .80655 L
.63123 .80594 L
.63197 .79623 L
.6354 .78652 L
.63891 .77681 L
.64083 .77159 L
.64232 .76709 L
.64647 .75738 L
.65054 .74892 L
.65112 .74767 L
.65619 .73796 L
.66026 .73107 L
.66194 .72824 L
.66845 .71853 L
.66997 .71641 L
.67566 .70882 L
.67968 .70393 L
.68386 .69911 L
.68939 .69321 L
.69321 .68939 L
.69911 .68386 L
.70393 .67968 L
.70882 .67566 L
.71641 .66997 L
.71853 .66845 L
.72824 .66194 L
.73107 .66026 L
.73796 .65619 L
.74767 .65112 L
.74892 .65054 L
.75738 .64647 L
.76709 .64232 L
.77159 .64083 L
.77681 .63891 L
.78652 .6354 L
.79623 .63197 L
.80594 .63123 L
.80655 .63112 L
F
0 g
.81566 .62921 m
.82537 .62933 L
.83325 .63112 L
.83088 .64083 L
.82537 .64592 L
.81984 .65054 L
.81566 .65376 L
.80708 .66026 L
.80594 .66125 L
.79623 .66833 L
.79409 .66997 L
.78652 .67622 L
.78133 .67968 L
.77681 .6834 L
.76903 .68939 L
.76709 .69102 L
.75738 .69906 L
.75732 .69911 L
.74767 .70763 L
.74628 .70882 L
.73796 .71655 L
.73582 .71853 L
.72824 .72592 L
.72592 .72824 L
.71853 .73582 L
.71655 .73796 L
.70882 .74628 L
.70763 .74767 L
.69911 .75732 L
.69906 .75738 L
.69102 .76709 L
.68939 .76903 L
.6834 .77681 L
.67968 .78133 L
.67622 .78652 L
.66997 .79409 L
.66833 .79623 L
.66125 .80594 L
.66026 .80708 L
.65376 .81566 L
.65054 .81984 L
.64592 .82537 L
.64083 .83088 L
.63112 .83325 L
.62933 .82537 L
.62921 .81566 L
.63112 .80655 L
.63123 .80594 L
.63197 .79623 L
.6354 .78652 L
Mistroke
.63891 .77681 L
.64083 .77159 L
.64232 .76709 L
.64647 .75738 L
.65054 .74892 L
.65112 .74767 L
.65619 .73796 L
.66026 .73107 L
.66194 .72824 L
.66845 .71853 L
.66997 .71641 L
.67566 .70882 L
.67968 .70393 L
.68386 .69911 L
.68939 .69321 L
.69321 .68939 L
.69911 .68386 L
.70393 .67968 L
.70882 .67566 L
.71641 .66997 L
.71853 .66845 L
.72824 .66194 L
.73107 .66026 L
.73796 .65619 L
.74767 .65112 L
.74892 .65054 L
.75738 .64647 L
.76709 .64232 L
.77159 .64083 L
.77681 .63891 L
.78652 .6354 L
.79623 .63197 L
.80594 .63123 L
.80655 .63112 L
.81566 .62921 L
Mfstroke
1 .24 0 r
.63112 .16675 m
.64083 .16912 L
.64592 .17463 L
.65054 .18016 L
.65376 .18434 L
.66026 .19292 L
.66125 .19406 L
.66833 .20377 L
.66997 .20591 L
.67622 .21348 L
.67968 .21867 L
.6834 .22319 L
.68939 .23097 L
.69102 .23291 L
.69906 .24262 L
.69911 .24268 L
.70763 .25233 L
.70882 .25372 L
.71655 .26204 L
.71853 .26418 L
.72592 .27176 L
.72824 .27408 L
.73582 .28147 L
.73796 .28345 L
.74628 .29118 L
.74767 .29237 L
.75732 .30089 L
.75738 .30094 L
.76709 .30898 L
.76903 .31061 L
.77681 .3166 L
.78133 .32032 L
.78652 .32378 L
.79409 .33003 L
.79623 .33167 L
.80594 .33875 L
.80708 .33974 L
.81566 .34624 L
.81984 .34946 L
.82537 .35408 L
.83088 .35917 L
.83325 .36888 L
.82537 .37067 L
.81566 .37079 L
.80655 .36888 L
.80594 .36877 L
.79623 .36803 L
.78652 .3646 L
.77681 .36109 L
.77159 .35917 L
.76709 .35768 L
.75738 .35353 L
.74892 .34946 L
.74767 .34888 L
.73796 .34381 L
.73107 .33974 L
.72824 .33806 L
.71853 .33155 L
.71641 .33003 L
.70882 .32434 L
.70393 .32032 L
.69911 .31614 L
.69321 .31061 L
.68939 .30679 L
.68386 .30089 L
.67968 .29607 L
.67566 .29118 L
.66997 .28359 L
.66845 .28147 L
.66194 .27176 L
.66026 .26893 L
.65619 .26204 L
.65112 .25233 L
.65054 .25108 L
.64647 .24262 L
.64232 .23291 L
.64083 .22841 L
.63891 .22319 L
.6354 .21348 L
.63197 .20377 L
.63123 .19406 L
.63112 .19345 L
.62921 .18434 L
.62933 .17463 L
F
0 g
.63112 .16675 m
.64083 .16912 L
.64592 .17463 L
.65054 .18016 L
.65376 .18434 L
.66026 .19292 L
.66125 .19406 L
.66833 .20377 L
.66997 .20591 L
.67622 .21348 L
.67968 .21867 L
.6834 .22319 L
.68939 .23097 L
.69102 .23291 L
.69906 .24262 L
.69911 .24268 L
.70763 .25233 L
.70882 .25372 L
.71655 .26204 L
.71853 .26418 L
.72592 .27176 L
.72824 .27408 L
.73582 .28147 L
.73796 .28345 L
.74628 .29118 L
.74767 .29237 L
.75732 .30089 L
.75738 .30094 L
.76709 .30898 L
.76903 .31061 L
.77681 .3166 L
.78133 .32032 L
.78652 .32378 L
.79409 .33003 L
.79623 .33167 L
.80594 .33875 L
.80708 .33974 L
.81566 .34624 L
.81984 .34946 L
.82537 .35408 L
.83088 .35917 L
.83325 .36888 L
.82537 .37067 L
.81566 .37079 L
.80655 .36888 L
.80594 .36877 L
.79623 .36803 L
.78652 .3646 L
.77681 .36109 L
.77159 .35917 L
Mistroke
.76709 .35768 L
.75738 .35353 L
.74892 .34946 L
.74767 .34888 L
.73796 .34381 L
.73107 .33974 L
.72824 .33806 L
.71853 .33155 L
.71641 .33003 L
.70882 .32434 L
.70393 .32032 L
.69911 .31614 L
.69321 .31061 L
.68939 .30679 L
.68386 .30089 L
.67968 .29607 L
.67566 .29118 L
.66997 .28359 L
.66845 .28147 L
.66194 .27176 L
.66026 .26893 L
.65619 .26204 L
.65112 .25233 L
.65054 .25108 L
.64647 .24262 L
.64232 .23291 L
.64083 .22841 L
.63891 .22319 L
.6354 .21348 L
.63197 .20377 L
.63123 .19406 L
.63112 .19345 L
.62921 .18434 L
.62933 .17463 L
.63112 .16675 L
Mfstroke
0 1 .16 r
.98077 .36906 m
.98019 .36888 L
.97106 .36602 L
.96134 .36287 L
.95163 .35964 L
.95022 .35917 L
.94192 .35626 L
.93221 .35275 L
.92338 .34946 L
.92249 .34911 L
.91278 .3453 L
.90307 .34133 L
.8993 .33974 L
.89336 .33718 L
.88364 .33283 L
.87762 .33003 L
.87393 .32827 L
.86422 .32349 L
.85803 .32032 L
.85451 .31846 L
.84479 .31316 L
.84028 .31061 L
.83508 .30758 L
.82537 .30167 L
.82412 .30089 L
.81566 .29542 L
.80937 .29118 L
.80594 .2888 L
.79623 .28176 L
.79584 .28147 L
.78652 .27426 L
.7834 .27176 L
.77681 .26626 L
.77193 .26204 L
.76709 .25771 L
.76131 .25233 L
.75738 .24854 L
.75146 .24262 L
.74767 .23869 L
.74229 .23291 L
.73796 .22807 L
.73374 .22319 L
.72824 .2166 L
.72574 .21348 L
.71853 .20416 L
.71824 .20377 L
.7112 .19406 L
.70882 .19063 L
.70458 .18434 L
.69911 .17588 L
.69833 .17463 L
.69242 .16492 L
.68939 .15972 L
.68684 .15521 L
.68154 .14549 L
.67968 .14197 L
.67651 .13578 L
.67173 .12607 L
.66997 .12238 L
.66717 .11636 L
.66282 .10664 L
.66026 .1007 L
.65867 .09693 L
.6547 .08722 L
.65089 .07751 L
.65054 .07662 L
.64725 .06779 L
.64374 .05808 L
.64083 .04978 L
.64036 .04837 L
.63713 .03866 L
.63398 .02894 L
.63112 .01981 L
.63094 .01923 L
.98077 .01923 L
F
0 g
.98077 .36906 m
.98019 .36888 L
.97106 .36602 L
.96134 .36287 L
.95163 .35964 L
.95022 .35917 L
.94192 .35626 L
.93221 .35275 L
.92338 .34946 L
.92249 .34911 L
.91278 .3453 L
.90307 .34133 L
.8993 .33974 L
.89336 .33718 L
.88364 .33283 L
.87762 .33003 L
.87393 .32827 L
.86422 .32349 L
.85803 .32032 L
.85451 .31846 L
.84479 .31316 L
.84028 .31061 L
.83508 .30758 L
.82537 .30167 L
.82412 .30089 L
.81566 .29542 L
.80937 .29118 L
.80594 .2888 L
.79623 .28176 L
.79584 .28147 L
.78652 .27426 L
.7834 .27176 L
.77681 .26626 L
.77193 .26204 L
.76709 .25771 L
.76131 .25233 L
.75738 .24854 L
.75146 .24262 L
.74767 .23869 L
.74229 .23291 L
.73796 .22807 L
.73374 .22319 L
.72824 .2166 L
.72574 .21348 L
.71853 .20416 L
.71824 .20377 L
.7112 .19406 L
.70882 .19063 L
.70458 .18434 L
.69911 .17588 L
Mistroke
.69833 .17463 L
.69242 .16492 L
.68939 .15972 L
.68684 .15521 L
.68154 .14549 L
.67968 .14197 L
.67651 .13578 L
.67173 .12607 L
.66997 .12238 L
.66717 .11636 L
.66282 .10664 L
.66026 .1007 L
.65867 .09693 L
.6547 .08722 L
.65089 .07751 L
.65054 .07662 L
.64725 .06779 L
.64374 .05808 L
.64083 .04978 L
.64036 .04837 L
.63713 .03866 L
.63398 .02894 L
.63112 .01981 L
.63094 .01923 L
Mfstroke
0 1 .16 r
.98077 .63094 m
.98019 .63112 L
.97106 .63398 L
.96134 .63713 L
.95163 .64036 L
.95022 .64083 L
.94192 .64374 L
.93221 .64725 L
.92338 .65054 L
.92249 .65089 L
.91278 .6547 L
.90307 .65867 L
.8993 .66026 L
.89336 .66282 L
.88364 .66717 L
.87762 .66997 L
.87393 .67173 L
.86422 .67651 L
.85803 .67968 L
.85451 .68154 L
.84479 .68684 L
.84028 .68939 L
.83508 .69242 L
.82537 .69833 L
.82412 .69911 L
.81566 .70458 L
.80937 .70882 L
.80594 .7112 L
.79623 .71824 L
.79584 .71853 L
.78652 .72574 L
.7834 .72824 L
.77681 .73374 L
.77193 .73796 L
.76709 .74229 L
.76131 .74767 L
.75738 .75146 L
.75146 .75738 L
.74767 .76131 L
.74229 .76709 L
.73796 .77193 L
.73374 .77681 L
.72824 .7834 L
.72574 .78652 L
.71853 .79584 L
.71824 .79623 L
.7112 .80594 L
.70882 .80937 L
.70458 .81566 L
.69911 .82412 L
.69833 .82537 L
.69242 .83508 L
.68939 .84028 L
.68684 .84479 L
.68154 .85451 L
.67968 .85803 L
.67651 .86422 L
.67173 .87393 L
.66997 .87762 L
.66717 .88364 L
.66282 .89336 L
.66026 .8993 L
.65867 .90307 L
.6547 .91278 L
.65089 .92249 L
.65054 .92338 L
.64725 .93221 L
.64374 .94192 L
.64083 .95022 L
.64036 .95163 L
.63713 .96134 L
.63398 .97106 L
.63112 .98019 L
.63094 .98077 L
.98077 .98077 L
F
0 g
.98077 .63094 m
.98019 .63112 L
.97106 .63398 L
.96134 .63713 L
.95163 .64036 L
.95022 .64083 L
.94192 .64374 L
.93221 .64725 L
.92338 .65054 L
.92249 .65089 L
.91278 .6547 L
.90307 .65867 L
.8993 .66026 L
.89336 .66282 L
.88364 .66717 L
.87762 .66997 L
.87393 .67173 L
.86422 .67651 L
.85803 .67968 L
.85451 .68154 L
.84479 .68684 L
.84028 .68939 L
.83508 .69242 L
.82537 .69833 L
.82412 .69911 L
.81566 .70458 L
.80937 .70882 L
.80594 .7112 L
.79623 .71824 L
.79584 .71853 L
.78652 .72574 L
.7834 .72824 L
.77681 .73374 L
.77193 .73796 L
.76709 .74229 L
.76131 .74767 L
.75738 .75146 L
.75146 .75738 L
.74767 .76131 L
.74229 .76709 L
.73796 .77193 L
.73374 .77681 L
.72824 .7834 L
.72574 .78652 L
.71853 .79584 L
.71824 .79623 L
.7112 .80594 L
.70882 .80937 L
.70458 .81566 L
.69911 .82412 L
Mistroke
.69833 .82537 L
.69242 .83508 L
.68939 .84028 L
.68684 .84479 L
.68154 .85451 L
.67968 .85803 L
.67651 .86422 L
.67173 .87393 L
.66997 .87762 L
.66717 .88364 L
.66282 .89336 L
.66026 .8993 L
.65867 .90307 L
.6547 .91278 L
.65089 .92249 L
.65054 .92338 L
.64725 .93221 L
.64374 .94192 L
.64083 .95022 L
.64036 .95163 L
.63713 .96134 L
.63398 .97106 L
.63112 .98019 L
.63094 .98077 L
Mfstroke
0 1 .4 r
.98077 .36392 m
.97106 .36095 L
.96538 .35917 L
.96134 .35787 L
.95163 .35466 L
.94192 .35132 L
.93666 .34946 L
.93221 .34784 L
.92249 .3442 L
.91278 .34041 L
.91112 .33974 L
.90307 .33644 L
.89336 .33228 L
.88828 .33003 L
.88364 .32792 L
.87393 .32334 L
.86777 .32032 L
.86422 .31853 L
.85451 .31346 L
.84926 .31061 L
.84479 .30811 L
.83508 .30246 L
.83248 .30089 L
.82537 .29648 L
.8172 .29118 L
.81566 .29015 L
.80594 .28342 L
.80323 .28147 L
.79623 .27626 L
.79042 .27176 L
.78652 .26863 L
.77862 .26204 L
.77681 .26048 L
.76773 .25233 L
.76709 .25174 L
.75763 .24262 L
.75738 .24237 L
.74826 .23291 L
.74767 .23227 L
.73952 .22319 L
.73796 .22138 L
.73137 .21348 L
.72824 .20958 L
.72374 .20377 L
.71853 .19677 L
.71658 .19406 L
.70985 .18434 L
.70882 .1828 L
.70352 .17463 L
.69911 .16752 L
.69754 .16492 L
.69189 .15521 L
.68939 .15074 L
.68654 .14549 L
.68147 .13578 L
.67968 .13223 L
.67666 .12607 L
.67208 .11636 L
.66997 .11172 L
.66772 .10664 L
.66356 .09693 L
.66026 .08888 L
.65959 .08722 L
.6558 .07751 L
.65216 .06779 L
.65054 .06334 L
.64868 .05808 L
.64534 .04837 L
.64213 .03866 L
.64083 .03462 L
.63905 .02894 L
.63608 .01923 L
.98077 .01923 L
F
0 g
.98077 .36392 m
.97106 .36095 L
.96538 .35917 L
.96134 .35787 L
.95163 .35466 L
.94192 .35132 L
.93666 .34946 L
.93221 .34784 L
.92249 .3442 L
.91278 .34041 L
.91112 .33974 L
.90307 .33644 L
.89336 .33228 L
.88828 .33003 L
.88364 .32792 L
.87393 .32334 L
.86777 .32032 L
.86422 .31853 L
.85451 .31346 L
.84926 .31061 L
.84479 .30811 L
.83508 .30246 L
.83248 .30089 L
.82537 .29648 L
.8172 .29118 L
.81566 .29015 L
.80594 .28342 L
.80323 .28147 L
.79623 .27626 L
.79042 .27176 L
.78652 .26863 L
.77862 .26204 L
.77681 .26048 L
.76773 .25233 L
.76709 .25174 L
.75763 .24262 L
.75738 .24237 L
.74826 .23291 L
.74767 .23227 L
.73952 .22319 L
.73796 .22138 L
.73137 .21348 L
.72824 .20958 L
.72374 .20377 L
.71853 .19677 L
.71658 .19406 L
.70985 .18434 L
.70882 .1828 L
.70352 .17463 L
.69911 .16752 L
Mistroke
.69754 .16492 L
.69189 .15521 L
.68939 .15074 L
.68654 .14549 L
.68147 .13578 L
.67968 .13223 L
.67666 .12607 L
.67208 .11636 L
.66997 .11172 L
.66772 .10664 L
.66356 .09693 L
.66026 .08888 L
.65959 .08722 L
.6558 .07751 L
.65216 .06779 L
.65054 .06334 L
.64868 .05808 L
.64534 .04837 L
.64213 .03866 L
.64083 .03462 L
.63905 .02894 L
.63608 .01923 L
Mfstroke
0 1 .4 r
.98077 .63608 m
.97106 .63905 L
.96538 .64083 L
.96134 .64213 L
.95163 .64534 L
.94192 .64868 L
.93666 .65054 L
.93221 .65216 L
.92249 .6558 L
.91278 .65959 L
.91112 .66026 L
.90307 .66356 L
.89336 .66772 L
.88828 .66997 L
.88364 .67208 L
.87393 .67666 L
.86777 .67968 L
.86422 .68147 L
.85451 .68654 L
.84926 .68939 L
.84479 .69189 L
.83508 .69754 L
.83248 .69911 L
.82537 .70352 L
.8172 .70882 L
.81566 .70985 L
.80594 .71658 L
.80323 .71853 L
.79623 .72374 L
.79042 .72824 L
.78652 .73137 L
.77862 .73796 L
.77681 .73952 L
.76773 .74767 L
.76709 .74826 L
.75763 .75738 L
.75738 .75763 L
.74826 .76709 L
.74767 .76773 L
.73952 .77681 L
.73796 .77862 L
.73137 .78652 L
.72824 .79042 L
.72374 .79623 L
.71853 .80323 L
.71658 .80594 L
.70985 .81566 L
.70882 .8172 L
.70352 .82537 L
.69911 .83248 L
.69754 .83508 L
.69189 .84479 L
.68939 .84926 L
.68654 .85451 L
.68147 .86422 L
.67968 .86777 L
.67666 .87393 L
.67208 .88364 L
.66997 .88828 L
.66772 .89336 L
.66356 .90307 L
.66026 .91112 L
.65959 .91278 L
.6558 .92249 L
.65216 .93221 L
.65054 .93666 L
.64868 .94192 L
.64534 .95163 L
.64213 .96134 L
.64083 .96538 L
.63905 .97106 L
.63608 .98077 L
.98077 .98077 L
F
0 g
.98077 .63608 m
.97106 .63905 L
.96538 .64083 L
.96134 .64213 L
.95163 .64534 L
.94192 .64868 L
.93666 .65054 L
.93221 .65216 L
.92249 .6558 L
.91278 .65959 L
.91112 .66026 L
.90307 .66356 L
.89336 .66772 L
.88828 .66997 L
.88364 .67208 L
.87393 .67666 L
.86777 .67968 L
.86422 .68147 L
.85451 .68654 L
.84926 .68939 L
.84479 .69189 L
.83508 .69754 L
.83248 .69911 L
.82537 .70352 L
.8172 .70882 L
.81566 .70985 L
.80594 .71658 L
.80323 .71853 L
.79623 .72374 L
.79042 .72824 L
.78652 .73137 L
.77862 .73796 L
.77681 .73952 L
.76773 .74767 L
.76709 .74826 L
.75763 .75738 L
.75738 .75763 L
.74826 .76709 L
.74767 .76773 L
.73952 .77681 L
.73796 .77862 L
.73137 .78652 L
.72824 .79042 L
.72374 .79623 L
.71853 .80323 L
.71658 .80594 L
.70985 .81566 L
.70882 .8172 L
.70352 .82537 L
.69911 .83248 L
Mistroke
.69754 .83508 L
.69189 .84479 L
.68939 .84926 L
.68654 .85451 L
.68147 .86422 L
.67968 .86777 L
.67666 .87393 L
.67208 .88364 L
.66997 .88828 L
.66772 .89336 L
.66356 .90307 L
.66026 .91112 L
.65959 .91278 L
.6558 .92249 L
.65216 .93221 L
.65054 .93666 L
.64868 .94192 L
.64534 .95163 L
.64213 .96134 L
.64083 .96538 L
.63905 .97106 L
.63608 .98077 L
Mfstroke
0 1 .64 r
.98077 .35891 m
.97106 .35599 L
.96134 .35294 L
.95163 .34976 L
.95072 .34946 L
.94192 .34645 L
.93221 .34298 L
.9235 .33974 L
.92249 .33936 L
.91278 .33557 L
.90307 .33159 L
.89937 .33003 L
.89336 .32743 L
.88364 .32305 L
.87783 .32032 L
.87393 .31844 L
.86422 .31359 L
.85849 .31061 L
.85451 .30848 L
.84479 .30307 L
.84103 .30089 L
.83508 .29736 L
.82537 .2913 L
.82519 .29118 L
.81566 .28487 L
.81075 .28147 L
.80594 .27804 L
.79753 .27176 L
.79623 .27075 L
.78652 .26298 L
.7854 .26204 L
.77681 .25466 L
.77421 .25233 L
.76709 .24573 L
.76386 .24262 L
.75738 .23614 L
.75427 .23291 L
.74767 .22579 L
.74534 .22319 L
.73796 .2146 L
.73702 .21348 L
.72925 .20377 L
.72824 .20247 L
.72196 .19406 L
.71853 .18925 L
.71513 .18434 L
.70882 .17481 L
.7087 .17463 L
.70264 .16492 L
.69911 .15897 L
.69693 .15521 L
.69152 .14549 L
.68939 .14151 L
.68641 .13578 L
.68156 .12607 L
.67968 .12217 L
.67695 .11636 L
.67257 .10664 L
.66997 .10063 L
.66841 .09693 L
.66443 .08722 L
.66064 .07751 L
.66026 .0765 L
.65702 .06779 L
.65355 .05808 L
.65054 .04928 L
.65024 .04837 L
.64706 .03866 L
.64401 .02894 L
.64109 .01923 L
.98077 .01923 L
F
0 g
.98077 .35891 m
.97106 .35599 L
.96134 .35294 L
.95163 .34976 L
.95072 .34946 L
.94192 .34645 L
.93221 .34298 L
.9235 .33974 L
.92249 .33936 L
.91278 .33557 L
.90307 .33159 L
.89937 .33003 L
.89336 .32743 L
.88364 .32305 L
.87783 .32032 L
.87393 .31844 L
.86422 .31359 L
.85849 .31061 L
.85451 .30848 L
.84479 .30307 L
.84103 .30089 L
.83508 .29736 L
.82537 .2913 L
.82519 .29118 L
.81566 .28487 L
.81075 .28147 L
.80594 .27804 L
.79753 .27176 L
.79623 .27075 L
.78652 .26298 L
.7854 .26204 L
.77681 .25466 L
.77421 .25233 L
.76709 .24573 L
.76386 .24262 L
.75738 .23614 L
.75427 .23291 L
.74767 .22579 L
.74534 .22319 L
.73796 .2146 L
.73702 .21348 L
.72925 .20377 L
.72824 .20247 L
.72196 .19406 L
.71853 .18925 L
.71513 .18434 L
.70882 .17481 L
.7087 .17463 L
.70264 .16492 L
.69911 .15897 L
Mistroke
.69693 .15521 L
.69152 .14549 L
.68939 .14151 L
.68641 .13578 L
.68156 .12607 L
.67968 .12217 L
.67695 .11636 L
.67257 .10664 L
.66997 .10063 L
.66841 .09693 L
.66443 .08722 L
.66064 .07751 L
.66026 .0765 L
.65702 .06779 L
.65355 .05808 L
.65054 .04928 L
.65024 .04837 L
.64706 .03866 L
.64401 .02894 L
.64109 .01923 L
Mfstroke
0 1 .64 r
.98077 .64109 m
.97106 .64401 L
.96134 .64706 L
.95163 .65024 L
.95072 .65054 L
.94192 .65355 L
.93221 .65702 L
.9235 .66026 L
.92249 .66064 L
.91278 .66443 L
.90307 .66841 L
.89937 .66997 L
.89336 .67257 L
.88364 .67695 L
.87783 .67968 L
.87393 .68156 L
.86422 .68641 L
.85849 .68939 L
.85451 .69152 L
.84479 .69693 L
.84103 .69911 L
.83508 .70264 L
.82537 .7087 L
.82519 .70882 L
.81566 .71513 L
.81075 .71853 L
.80594 .72196 L
.79753 .72824 L
.79623 .72925 L
.78652 .73702 L
.7854 .73796 L
.77681 .74534 L
.77421 .74767 L
.76709 .75427 L
.76386 .75738 L
.75738 .76386 L
.75427 .76709 L
.74767 .77421 L
.74534 .77681 L
.73796 .7854 L
.73702 .78652 L
.72925 .79623 L
.72824 .79753 L
.72196 .80594 L
.71853 .81075 L
.71513 .81566 L
.70882 .82519 L
.7087 .82537 L
.70264 .83508 L
.69911 .84103 L
.69693 .84479 L
.69152 .85451 L
.68939 .85849 L
.68641 .86422 L
.68156 .87393 L
.67968 .87783 L
.67695 .88364 L
.67257 .89336 L
.66997 .89937 L
.66841 .90307 L
.66443 .91278 L
.66064 .92249 L
.66026 .9235 L
.65702 .93221 L
.65355 .94192 L
.65054 .95072 L
.65024 .95163 L
.64706 .96134 L
.64401 .97106 L
.64109 .98077 L
.98077 .98077 L
F
0 g
.98077 .64109 m
.97106 .64401 L
.96134 .64706 L
.95163 .65024 L
.95072 .65054 L
.94192 .65355 L
.93221 .65702 L
.9235 .66026 L
.92249 .66064 L
.91278 .66443 L
.90307 .66841 L
.89937 .66997 L
.89336 .67257 L
.88364 .67695 L
.87783 .67968 L
.87393 .68156 L
.86422 .68641 L
.85849 .68939 L
.85451 .69152 L
.84479 .69693 L
.84103 .69911 L
.83508 .70264 L
.82537 .7087 L
.82519 .70882 L
.81566 .71513 L
.81075 .71853 L
.80594 .72196 L
.79753 .72824 L
.79623 .72925 L
.78652 .73702 L
.7854 .73796 L
.77681 .74534 L
.77421 .74767 L
.76709 .75427 L
.76386 .75738 L
.75738 .76386 L
.75427 .76709 L
.74767 .77421 L
.74534 .77681 L
.73796 .7854 L
.73702 .78652 L
.72925 .79623 L
.72824 .79753 L
.72196 .80594 L
.71853 .81075 L
.71513 .81566 L
.70882 .82519 L
.7087 .82537 L
.70264 .83508 L
.69911 .84103 L
Mistroke
.69693 .84479 L
.69152 .85451 L
.68939 .85849 L
.68641 .86422 L
.68156 .87393 L
.67968 .87783 L
.67695 .88364 L
.67257 .89336 L
.66997 .89937 L
.66841 .90307 L
.66443 .91278 L
.66064 .92249 L
.66026 .9235 L
.65702 .93221 L
.65355 .94192 L
.65054 .95072 L
.65024 .95163 L
.64706 .96134 L
.64401 .97106 L
.64109 .98077 L
Mfstroke
0 1 .88 r
.98077 .35387 m
.97106 .35097 L
.96611 .34946 L
.96134 .34797 L
.95163 .34483 L
.94192 .34153 L
.93684 .33974 L
.93221 .33809 L
.92249 .33448 L
.91278 .33068 L
.91117 .33003 L
.90307 .32671 L
.89336 .32252 L
.88844 .32032 L
.88364 .31813 L
.87393 .31349 L
.86815 .31061 L
.86422 .3086 L
.85451 .30343 L
.84993 .30089 L
.84479 .29797 L
.83508 .29218 L
.83346 .29118 L
.82537 .28604 L
.8185 .28147 L
.81566 .27952 L
.80594 .27256 L
.80486 .27176 L
.79623 .26515 L
.79235 .26204 L
.78652 .25722 L
.78084 .25233 L
.77681 .24873 L
.77023 .24262 L
.76709 .2396 L
.7604 .23291 L
.75738 .22977 L
.75127 .22319 L
.74767 .21916 L
.74278 .21348 L
.73796 .20765 L
.73485 .20377 L
.72824 .19514 L
.72744 .19406 L
.72048 .18434 L
.71853 .1815 L
.71396 .17463 L
.70882 .16654 L
.70782 .16492 L
.70203 .15521 L
.69911 .15007 L
.69657 .14549 L
.6914 .13578 L
.68939 .13185 L
.68651 .12607 L
.68187 .11636 L
.67968 .11156 L
.67748 .10664 L
.67329 .09693 L
.66997 .08883 L
.66932 .08722 L
.66552 .07751 L
.66191 .06779 L
.66026 .06316 L
.65847 .05808 L
.65517 .04837 L
.65203 .03866 L
.65054 .03389 L
.64903 .02894 L
.64613 .01923 L
.98077 .01923 L
F
0 g
.98077 .35387 m
.97106 .35097 L
.96611 .34946 L
.96134 .34797 L
.95163 .34483 L
.94192 .34153 L
.93684 .33974 L
.93221 .33809 L
.92249 .33448 L
.91278 .33068 L
.91117 .33003 L
.90307 .32671 L
.89336 .32252 L
.88844 .32032 L
.88364 .31813 L
.87393 .31349 L
.86815 .31061 L
.86422 .3086 L
.85451 .30343 L
.84993 .30089 L
.84479 .29797 L
.83508 .29218 L
.83346 .29118 L
.82537 .28604 L
.8185 .28147 L
.81566 .27952 L
.80594 .27256 L
.80486 .27176 L
.79623 .26515 L
.79235 .26204 L
.78652 .25722 L
.78084 .25233 L
.77681 .24873 L
.77023 .24262 L
.76709 .2396 L
.7604 .23291 L
.75738 .22977 L
.75127 .22319 L
.74767 .21916 L
.74278 .21348 L
.73796 .20765 L
.73485 .20377 L
.72824 .19514 L
.72744 .19406 L
.72048 .18434 L
.71853 .1815 L
.71396 .17463 L
.70882 .16654 L
.70782 .16492 L
.70203 .15521 L
Mistroke
.69911 .15007 L
.69657 .14549 L
.6914 .13578 L
.68939 .13185 L
.68651 .12607 L
.68187 .11636 L
.67968 .11156 L
.67748 .10664 L
.67329 .09693 L
.66997 .08883 L
.66932 .08722 L
.66552 .07751 L
.66191 .06779 L
.66026 .06316 L
.65847 .05808 L
.65517 .04837 L
.65203 .03866 L
.65054 .03389 L
.64903 .02894 L
.64613 .01923 L
Mfstroke
0 1 .88 r
.98077 .64613 m
.97106 .64903 L
.96611 .65054 L
.96134 .65203 L
.95163 .65517 L
.94192 .65847 L
.93684 .66026 L
.93221 .66191 L
.92249 .66552 L
.91278 .66932 L
.91117 .66997 L
.90307 .67329 L
.89336 .67748 L
.88844 .67968 L
.88364 .68187 L
.87393 .68651 L
.86815 .68939 L
.86422 .6914 L
.85451 .69657 L
.84993 .69911 L
.84479 .70203 L
.83508 .70782 L
.83346 .70882 L
.82537 .71396 L
.8185 .71853 L
.81566 .72048 L
.80594 .72744 L
.80486 .72824 L
.79623 .73485 L
.79235 .73796 L
.78652 .74278 L
.78084 .74767 L
.77681 .75127 L
.77023 .75738 L
.76709 .7604 L
.7604 .76709 L
.75738 .77023 L
.75127 .77681 L
.74767 .78084 L
.74278 .78652 L
.73796 .79235 L
.73485 .79623 L
.72824 .80486 L
.72744 .80594 L
.72048 .81566 L
.71853 .8185 L
.71396 .82537 L
.70882 .83346 L
.70782 .83508 L
.70203 .84479 L
.69911 .84993 L
.69657 .85451 L
.6914 .86422 L
.68939 .86815 L
.68651 .87393 L
.68187 .88364 L
.67968 .88844 L
.67748 .89336 L
.67329 .90307 L
.66997 .91117 L
.66932 .91278 L
.66552 .92249 L
.66191 .93221 L
.66026 .93684 L
.65847 .94192 L
.65517 .95163 L
.65203 .96134 L
.65054 .96611 L
.64903 .97106 L
.64613 .98077 L
.98077 .98077 L
F
0 g
.98077 .64613 m
.97106 .64903 L
.96611 .65054 L
.96134 .65203 L
.95163 .65517 L
.94192 .65847 L
.93684 .66026 L
.93221 .66191 L
.92249 .66552 L
.91278 .66932 L
.91117 .66997 L
.90307 .67329 L
.89336 .67748 L
.88844 .67968 L
.88364 .68187 L
.87393 .68651 L
.86815 .68939 L
.86422 .6914 L
.85451 .69657 L
.84993 .69911 L
.84479 .70203 L
.83508 .70782 L
.83346 .70882 L
.82537 .71396 L
.8185 .71853 L
.81566 .72048 L
.80594 .72744 L
.80486 .72824 L
.79623 .73485 L
.79235 .73796 L
.78652 .74278 L
.78084 .74767 L
.77681 .75127 L
.77023 .75738 L
.76709 .7604 L
.7604 .76709 L
.75738 .77023 L
.75127 .77681 L
.74767 .78084 L
.74278 .78652 L
.73796 .79235 L
.73485 .79623 L
.72824 .80486 L
.72744 .80594 L
.72048 .81566 L
.71853 .8185 L
.71396 .82537 L
.70882 .83346 L
.70782 .83508 L
.70203 .84479 L
Mistroke
.69911 .84993 L
.69657 .85451 L
.6914 .86422 L
.68939 .86815 L
.68651 .87393 L
.68187 .88364 L
.67968 .88844 L
.67748 .89336 L
.67329 .90307 L
.66997 .91117 L
.66932 .91278 L
.66552 .92249 L
.66191 .93221 L
.66026 .93684 L
.65847 .94192 L
.65517 .95163 L
.65203 .96134 L
.65054 .96611 L
.64903 .97106 L
.64613 .98077 L
Mfstroke
0 .88 1 r
.98077 .34862 m
.97106 .34581 L
.96134 .34283 L
.95175 .33974 L
.95163 .33971 L
.94192 .33647 L
.93221 .33304 L
.9241 .33003 L
.92249 .32943 L
.91278 .32566 L
.90307 .32167 L
.8999 .32032 L
.89336 .31748 L
.88364 .31306 L
.87849 .31061 L
.87393 .3084 L
.86422 .30346 L
.85938 .30089 L
.85451 .29825 L
.84479 .29272 L
.8422 .29118 L
.83508 .28686 L
.82665 .28147 L
.82537 .28063 L
.81566 .274 L
.81251 .27176 L
.80594 .26693 L
.79959 .26204 L
.79623 .25937 L
.78774 .25233 L
.78652 .25128 L
.77683 .24262 L
.77681 .24259 L
.76709 .23325 L
.76675 .23291 L
.75741 .22319 L
.75738 .22317 L
.74872 .21348 L
.74767 .21226 L
.74063 .20377 L
.73796 .20041 L
.73307 .19406 L
.72824 .18749 L
.726 .18434 L
.71937 .17463 L
.71853 .17335 L
.71314 .16492 L
.70882 .1578 L
.70728 .15521 L
.70175 .14549 L
.69911 .14062 L
.69654 .13578 L
.6916 .12607 L
.68939 .12151 L
.68694 .11636 L
.68252 .10664 L
.67968 .1001 L
.67833 .09693 L
.67434 .08722 L
.67057 .07751 L
.66997 .0759 L
.66696 .06779 L
.66353 .05808 L
.66029 .04837 L
.66026 .04825 L
.65717 .03866 L
.65419 .02894 L
.65138 .01923 L
.98077 .01923 L
F
0 g
.98077 .34862 m
.97106 .34581 L
.96134 .34283 L
.95175 .33974 L
.95163 .33971 L
.94192 .33647 L
.93221 .33304 L
.9241 .33003 L
.92249 .32943 L
.91278 .32566 L
.90307 .32167 L
.8999 .32032 L
.89336 .31748 L
.88364 .31306 L
.87849 .31061 L
.87393 .3084 L
.86422 .30346 L
.85938 .30089 L
.85451 .29825 L
.84479 .29272 L
.8422 .29118 L
.83508 .28686 L
.82665 .28147 L
.82537 .28063 L
.81566 .274 L
.81251 .27176 L
.80594 .26693 L
.79959 .26204 L
.79623 .25937 L
.78774 .25233 L
.78652 .25128 L
.77683 .24262 L
.77681 .24259 L
.76709 .23325 L
.76675 .23291 L
.75741 .22319 L
.75738 .22317 L
.74872 .21348 L
.74767 .21226 L
.74063 .20377 L
.73796 .20041 L
.73307 .19406 L
.72824 .18749 L
.726 .18434 L
.71937 .17463 L
.71853 .17335 L
.71314 .16492 L
.70882 .1578 L
.70728 .15521 L
.70175 .14549 L
Mistroke
.69911 .14062 L
.69654 .13578 L
.6916 .12607 L
.68939 .12151 L
.68694 .11636 L
.68252 .10664 L
.67968 .1001 L
.67833 .09693 L
.67434 .08722 L
.67057 .07751 L
.66997 .0759 L
.66696 .06779 L
.66353 .05808 L
.66029 .04837 L
.66026 .04825 L
.65717 .03866 L
.65419 .02894 L
.65138 .01923 L
Mfstroke
0 .88 1 r
.98077 .65138 m
.97106 .65419 L
.96134 .65717 L
.95175 .66026 L
.95163 .66029 L
.94192 .66353 L
.93221 .66696 L
.9241 .66997 L
.92249 .67057 L
.91278 .67434 L
.90307 .67833 L
.8999 .67968 L
.89336 .68252 L
.88364 .68694 L
.87849 .68939 L
.87393 .6916 L
.86422 .69654 L
.85938 .69911 L
.85451 .70175 L
.84479 .70728 L
.8422 .70882 L
.83508 .71314 L
.82665 .71853 L
.82537 .71937 L
.81566 .726 L
.81251 .72824 L
.80594 .73307 L
.79959 .73796 L
.79623 .74063 L
.78774 .74767 L
.78652 .74872 L
.77683 .75738 L
.77681 .75741 L
.76709 .76675 L
.76675 .76709 L
.75741 .77681 L
.75738 .77683 L
.74872 .78652 L
.74767 .78774 L
.74063 .79623 L
.73796 .79959 L
.73307 .80594 L
.72824 .81251 L
.726 .81566 L
.71937 .82537 L
.71853 .82665 L
.71314 .83508 L
.70882 .8422 L
.70728 .84479 L
.70175 .85451 L
.69911 .85938 L
.69654 .86422 L
.6916 .87393 L
.68939 .87849 L
.68694 .88364 L
.68252 .89336 L
.67968 .8999 L
.67833 .90307 L
.67434 .91278 L
.67057 .92249 L
.66997 .9241 L
.66696 .93221 L
.66353 .94192 L
.66029 .95163 L
.66026 .95175 L
.65717 .96134 L
.65419 .97106 L
.65138 .98077 L
.98077 .98077 L
F
0 g
.98077 .65138 m
.97106 .65419 L
.96134 .65717 L
.95175 .66026 L
.95163 .66029 L
.94192 .66353 L
.93221 .66696 L
.9241 .66997 L
.92249 .67057 L
.91278 .67434 L
.90307 .67833 L
.8999 .67968 L
.89336 .68252 L
.88364 .68694 L
.87849 .68939 L
.87393 .6916 L
.86422 .69654 L
.85938 .69911 L
.85451 .70175 L
.84479 .70728 L
.8422 .70882 L
.83508 .71314 L
.82665 .71853 L
.82537 .71937 L
.81566 .726 L
.81251 .72824 L
.80594 .73307 L
.79959 .73796 L
.79623 .74063 L
.78774 .74767 L
.78652 .74872 L
.77683 .75738 L
.77681 .75741 L
.76709 .76675 L
.76675 .76709 L
.75741 .77681 L
.75738 .77683 L
.74872 .78652 L
.74767 .78774 L
.74063 .79623 L
.73796 .79959 L
.73307 .80594 L
.72824 .81251 L
.726 .81566 L
.71937 .82537 L
.71853 .82665 L
.71314 .83508 L
.70882 .8422 L
.70728 .84479 L
.70175 .85451 L
Mistroke
.69911 .85938 L
.69654 .86422 L
.6916 .87393 L
.68939 .87849 L
.68694 .88364 L
.68252 .89336 L
.67968 .8999 L
.67833 .90307 L
.67434 .91278 L
.67057 .92249 L
.66997 .9241 L
.66696 .93221 L
.66353 .94192 L
.66029 .95163 L
.66026 .95175 L
.65717 .96134 L
.65419 .97106 L
.65138 .98077 L
Mfstroke
0 .64 1 r
.98077 .343 m
.97106 .3402 L
.96949 .33974 L
.96134 .33737 L
.95163 .33431 L
.94192 .33105 L
.93896 .33003 L
.93221 .3277 L
.92249 .32412 L
.91278 .32032 L
.91278 .32032 L
.90307 .31638 L
.89336 .31216 L
.88991 .31061 L
.88364 .30774 L
.87393 .30304 L
.86969 .30089 L
.86422 .29808 L
.85451 .2928 L
.85164 .29118 L
.84479 .28723 L
.8354 .28147 L
.83508 .28127 L
.82537 .27496 L
.82068 .27176 L
.81566 .26822 L
.8073 .26204 L
.80594 .26101 L
.79623 .2533 L
.79506 .25233 L
.78652 .24504 L
.78381 .24262 L
.77681 .23615 L
.77344 .23291 L
.76709 .22656 L
.76385 .22319 L
.75738 .21619 L
.75496 .21348 L
.74767 .20494 L
.7467 .20377 L
.73899 .19406 L
.73796 .1927 L
.73178 .18434 L
.72824 .17932 L
.72504 .17463 L
.71873 .16492 L
.71853 .1646 L
.71277 .15521 L
.70882 .14836 L
.7072 .14549 L
.70192 .13578 L
.69911 .13031 L
.69696 .12607 L
.69226 .11636 L
.68939 .11009 L
.68784 .10664 L
.68362 .09693 L
.67968 .08722 L
.67968 .08722 L
.67588 .07751 L
.6723 .06779 L
.66997 .06104 L
.66895 .05808 L
.66569 .04837 L
.66263 .03866 L
.66026 .03051 L
.6598 .02894 L
.657 .01923 L
.98077 .01923 L
F
0 g
.98077 .343 m
.97106 .3402 L
.96949 .33974 L
.96134 .33737 L
.95163 .33431 L
.94192 .33105 L
.93896 .33003 L
.93221 .3277 L
.92249 .32412 L
.91278 .32032 L
.91278 .32032 L
.90307 .31638 L
.89336 .31216 L
.88991 .31061 L
.88364 .30774 L
.87393 .30304 L
.86969 .30089 L
.86422 .29808 L
.85451 .2928 L
.85164 .29118 L
.84479 .28723 L
.8354 .28147 L
.83508 .28127 L
.82537 .27496 L
.82068 .27176 L
.81566 .26822 L
.8073 .26204 L
.80594 .26101 L
.79623 .2533 L
.79506 .25233 L
.78652 .24504 L
.78381 .24262 L
.77681 .23615 L
.77344 .23291 L
.76709 .22656 L
.76385 .22319 L
.75738 .21619 L
.75496 .21348 L
.74767 .20494 L
.7467 .20377 L
.73899 .19406 L
.73796 .1927 L
.73178 .18434 L
.72824 .17932 L
.72504 .17463 L
.71873 .16492 L
.71853 .1646 L
.71277 .15521 L
.70882 .14836 L
.7072 .14549 L
Mistroke
.70192 .13578 L
.69911 .13031 L
.69696 .12607 L
.69226 .11636 L
.68939 .11009 L
.68784 .10664 L
.68362 .09693 L
.67968 .08722 L
.67968 .08722 L
.67588 .07751 L
.6723 .06779 L
.66997 .06104 L
.66895 .05808 L
.66569 .04837 L
.66263 .03866 L
.66026 .03051 L
.6598 .02894 L
.657 .01923 L
Mfstroke
0 .64 1 r
.98077 .657 m
.97106 .6598 L
.96949 .66026 L
.96134 .66263 L
.95163 .66569 L
.94192 .66895 L
.93896 .66997 L
.93221 .6723 L
.92249 .67588 L
.91278 .67968 L
.91278 .67968 L
.90307 .68362 L
.89336 .68784 L
.88991 .68939 L
.88364 .69226 L
.87393 .69696 L
.86969 .69911 L
.86422 .70192 L
.85451 .7072 L
.85164 .70882 L
.84479 .71277 L
.8354 .71853 L
.83508 .71873 L
.82537 .72504 L
.82068 .72824 L
.81566 .73178 L
.8073 .73796 L
.80594 .73899 L
.79623 .7467 L
.79506 .74767 L
.78652 .75496 L
.78381 .75738 L
.77681 .76385 L
.77344 .76709 L
.76709 .77344 L
.76385 .77681 L
.75738 .78381 L
.75496 .78652 L
.74767 .79506 L
.7467 .79623 L
.73899 .80594 L
.73796 .8073 L
.73178 .81566 L
.72824 .82068 L
.72504 .82537 L
.71873 .83508 L
.71853 .8354 L
.71277 .84479 L
.70882 .85164 L
.7072 .85451 L
.70192 .86422 L
.69911 .86969 L
.69696 .87393 L
.69226 .88364 L
.68939 .88991 L
.68784 .89336 L
.68362 .90307 L
.67968 .91278 L
.67968 .91278 L
.67588 .92249 L
.6723 .93221 L
.66997 .93896 L
.66895 .94192 L
.66569 .95163 L
.66263 .96134 L
.66026 .96949 L
.6598 .97106 L
.657 .98077 L
.98077 .98077 L
F
0 g
.98077 .657 m
.97106 .6598 L
.96949 .66026 L
.96134 .66263 L
.95163 .66569 L
.94192 .66895 L
.93896 .66997 L
.93221 .6723 L
.92249 .67588 L
.91278 .67968 L
.91278 .67968 L
.90307 .68362 L
.89336 .68784 L
.88991 .68939 L
.88364 .69226 L
.87393 .69696 L
.86969 .69911 L
.86422 .70192 L
.85451 .7072 L
.85164 .70882 L
.84479 .71277 L
.8354 .71853 L
.83508 .71873 L
.82537 .72504 L
.82068 .72824 L
.81566 .73178 L
.8073 .73796 L
.80594 .73899 L
.79623 .7467 L
.79506 .74767 L
.78652 .75496 L
.78381 .75738 L
.77681 .76385 L
.77344 .76709 L
.76709 .77344 L
.76385 .77681 L
.75738 .78381 L
.75496 .78652 L
.74767 .79506 L
.7467 .79623 L
.73899 .80594 L
.73796 .8073 L
.73178 .81566 L
.72824 .82068 L
.72504 .82537 L
.71873 .83508 L
.71853 .8354 L
.71277 .84479 L
.70882 .85164 L
.7072 .85451 L
Mistroke
.70192 .86422 L
.69911 .86969 L
.69696 .87393 L
.69226 .88364 L
.68939 .88991 L
.68784 .89336 L
.68362 .90307 L
.67968 .91278 L
.67968 .91278 L
.67588 .92249 L
.6723 .93221 L
.66997 .93896 L
.66895 .94192 L
.66569 .95163 L
.66263 .96134 L
.66026 .96949 L
.6598 .97106 L
.657 .98077 L
Mfstroke
0 .4 1 r
.98077 .33674 m
.97106 .33403 L
.96134 .33115 L
.95766 .33003 L
.95163 .3283 L
.94192 .32516 L
.93221 .32177 L
.92819 .32032 L
.92249 .31831 L
.91278 .31457 L
.90319 .31061 L
.90307 .31056 L
.89336 .30644 L
.88364 .30195 L
.88142 .30089 L
.87393 .29729 L
.86422 .29225 L
.86222 .29118 L
.85451 .28698 L
.84509 .28147 L
.84479 .28129 L
.83508 .2753 L
.82967 .27176 L
.82537 .26888 L
.81572 .26204 L
.81566 .262 L
.80594 .25467 L
.80299 .25233 L
.79623 .2468 L
.79135 .24262 L
.78652 .23834 L
.78065 .23291 L
.77681 .22921 L
.77079 .22319 L
.76709 .21935 L
.76166 .21348 L
.75738 .20865 L
.7532 .20377 L
.74767 .19701 L
.74533 .19406 L
.738 .18434 L
.73796 .18428 L
.73112 .17463 L
.72824 .17033 L
.7247 .16492 L
.71871 .15521 L
.71853 .15491 L
.71302 .14549 L
.70882 .13778 L
.70775 .13578 L
.70271 .12607 L
.69911 .11858 L
.69805 .11636 L
.69356 .10664 L
.68944 .09693 L
.68939 .09681 L
.68543 .08722 L
.68169 .07751 L
.67968 .07181 L
.67823 .06779 L
.67484 .05808 L
.6717 .04837 L
.66997 .04234 L
.66885 .03866 L
.66597 .02894 L
.66326 .01923 L
.98077 .01923 L
F
0 g
.98077 .33674 m
.97106 .33403 L
.96134 .33115 L
.95766 .33003 L
.95163 .3283 L
.94192 .32516 L
.93221 .32177 L
.92819 .32032 L
.92249 .31831 L
.91278 .31457 L
.90319 .31061 L
.90307 .31056 L
.89336 .30644 L
.88364 .30195 L
.88142 .30089 L
.87393 .29729 L
.86422 .29225 L
.86222 .29118 L
.85451 .28698 L
.84509 .28147 L
.84479 .28129 L
.83508 .2753 L
.82967 .27176 L
.82537 .26888 L
.81572 .26204 L
.81566 .262 L
.80594 .25467 L
.80299 .25233 L
.79623 .2468 L
.79135 .24262 L
.78652 .23834 L
.78065 .23291 L
.77681 .22921 L
.77079 .22319 L
.76709 .21935 L
.76166 .21348 L
.75738 .20865 L
.7532 .20377 L
.74767 .19701 L
.74533 .19406 L
.738 .18434 L
.73796 .18428 L
.73112 .17463 L
.72824 .17033 L
.7247 .16492 L
.71871 .15521 L
.71853 .15491 L
.71302 .14549 L
.70882 .13778 L
.70775 .13578 L
Mistroke
.70271 .12607 L
.69911 .11858 L
.69805 .11636 L
.69356 .10664 L
.68944 .09693 L
.68939 .09681 L
.68543 .08722 L
.68169 .07751 L
.67968 .07181 L
.67823 .06779 L
.67484 .05808 L
.6717 .04837 L
.66997 .04234 L
.66885 .03866 L
.66597 .02894 L
.66326 .01923 L
Mfstroke
0 .4 1 r
.98077 .66326 m
.97106 .66597 L
.96134 .66885 L
.95766 .66997 L
.95163 .6717 L
.94192 .67484 L
.93221 .67823 L
.92819 .67968 L
.92249 .68169 L
.91278 .68543 L
.90319 .68939 L
.90307 .68944 L
.89336 .69356 L
.88364 .69805 L
.88142 .69911 L
.87393 .70271 L
.86422 .70775 L
.86222 .70882 L
.85451 .71302 L
.84509 .71853 L
.84479 .71871 L
.83508 .7247 L
.82967 .72824 L
.82537 .73112 L
.81572 .73796 L
.81566 .738 L
.80594 .74533 L
.80299 .74767 L
.79623 .7532 L
.79135 .75738 L
.78652 .76166 L
.78065 .76709 L
.77681 .77079 L
.77079 .77681 L
.76709 .78065 L
.76166 .78652 L
.75738 .79135 L
.7532 .79623 L
.74767 .80299 L
.74533 .80594 L
.738 .81566 L
.73796 .81572 L
.73112 .82537 L
.72824 .82967 L
.7247 .83508 L
.71871 .84479 L
.71853 .84509 L
.71302 .85451 L
.70882 .86222 L
.70775 .86422 L
.70271 .87393 L
.69911 .88142 L
.69805 .88364 L
.69356 .89336 L
.68944 .90307 L
.68939 .90319 L
.68543 .91278 L
.68169 .92249 L
.67968 .92819 L
.67823 .93221 L
.67484 .94192 L
.6717 .95163 L
.66997 .95766 L
.66885 .96134 L
.66597 .97106 L
.66326 .98077 L
.98077 .98077 L
F
0 g
.98077 .66326 m
.97106 .66597 L
.96134 .66885 L
.95766 .66997 L
.95163 .6717 L
.94192 .67484 L
.93221 .67823 L
.92819 .67968 L
.92249 .68169 L
.91278 .68543 L
.90319 .68939 L
.90307 .68944 L
.89336 .69356 L
.88364 .69805 L
.88142 .69911 L
.87393 .70271 L
.86422 .70775 L
.86222 .70882 L
.85451 .71302 L
.84509 .71853 L
.84479 .71871 L
.83508 .7247 L
.82967 .72824 L
.82537 .73112 L
.81572 .73796 L
.81566 .738 L
.80594 .74533 L
.80299 .74767 L
.79623 .7532 L
.79135 .75738 L
.78652 .76166 L
.78065 .76709 L
.77681 .77079 L
.77079 .77681 L
.76709 .78065 L
.76166 .78652 L
.75738 .79135 L
.7532 .79623 L
.74767 .80299 L
.74533 .80594 L
.738 .81566 L
.73796 .81572 L
.73112 .82537 L
.72824 .82967 L
.7247 .83508 L
.71871 .84479 L
.71853 .84509 L
.71302 .85451 L
.70882 .86222 L
.70775 .86422 L
Mistroke
.70271 .87393 L
.69911 .88142 L
.69805 .88364 L
.69356 .89336 L
.68944 .90307 L
.68939 .90319 L
.68543 .91278 L
.68169 .92249 L
.67968 .92819 L
.67823 .93221 L
.67484 .94192 L
.6717 .95163 L
.66997 .95766 L
.66885 .96134 L
.66597 .97106 L
.66326 .98077 L
Mfstroke
1 0 0 r
.74767 .66746 m
.75738 .66409 L
.76686 .66997 L
.75896 .67968 L
.75738 .68112 L
.74919 .68939 L
.74767 .69132 L
.73907 .69911 L
.73796 .7003 L
.72901 .70882 L
.72824 .7096 L
.71917 .71853 L
.71853 .71917 L
.7096 .72824 L
.70882 .72901 L
.7003 .73796 L
.69911 .73907 L
.69132 .74767 L
.68939 .74919 L
.68112 .75738 L
.67968 .75896 L
.66997 .76686 L
.66409 .75738 L
.66746 .74767 L
.66997 .74209 L
.67206 .73796 L
.67731 .72824 L
.67968 .7246 L
.68338 .71853 L
.68939 .71109 L
.69109 .70882 L
.69911 .70033 L
.70033 .69911 L
.70882 .69109 L
.71109 .68939 L
.71853 .68338 L
.7246 .67968 L
.72824 .67731 L
.73796 .67206 L
.74209 .66997 L
F
0 g
.74767 .66746 m
.75738 .66409 L
.76686 .66997 L
.75896 .67968 L
.75738 .68112 L
.74919 .68939 L
.74767 .69132 L
.73907 .69911 L
.73796 .7003 L
.72901 .70882 L
.72824 .7096 L
.71917 .71853 L
.71853 .71917 L
.7096 .72824 L
.70882 .72901 L
.7003 .73796 L
.69911 .73907 L
.69132 .74767 L
.68939 .74919 L
.68112 .75738 L
.67968 .75896 L
.66997 .76686 L
.66409 .75738 L
.66746 .74767 L
.66997 .74209 L
.67206 .73796 L
.67731 .72824 L
.67968 .7246 L
.68338 .71853 L
.68939 .71109 L
.69109 .70882 L
.69911 .70033 L
.70033 .69911 L
.70882 .69109 L
.71109 .68939 L
.71853 .68338 L
.7246 .67968 L
.72824 .67731 L
.73796 .67206 L
.74209 .66997 L
.74767 .66746 L
s
1 0 0 r
.66997 .23314 m
.67968 .24104 L
.68112 .24262 L
.68939 .25081 L
.69132 .25233 L
.69911 .26093 L
.7003 .26204 L
.70882 .27099 L
.7096 .27176 L
.71853 .28083 L
.71917 .28147 L
.72824 .2904 L
.72901 .29118 L
.73796 .2997 L
.73907 .30089 L
.74767 .30868 L
.74919 .31061 L
.75738 .31888 L
.75896 .32032 L
.76686 .33003 L
.75738 .33591 L
.74767 .33254 L
.74209 .33003 L
.73796 .32794 L
.72824 .32269 L
.7246 .32032 L
.71853 .31662 L
.71109 .31061 L
.70882 .30891 L
.70033 .30089 L
.69911 .29967 L
.69109 .29118 L
.68939 .28891 L
.68338 .28147 L
.67968 .2754 L
.67731 .27176 L
.67206 .26204 L
.66997 .25791 L
.66746 .25233 L
.66409 .24262 L
F
0 g
.66997 .23314 m
.67968 .24104 L
.68112 .24262 L
.68939 .25081 L
.69132 .25233 L
.69911 .26093 L
.7003 .26204 L
.70882 .27099 L
.7096 .27176 L
.71853 .28083 L
.71917 .28147 L
.72824 .2904 L
.72901 .29118 L
.73796 .2997 L
.73907 .30089 L
.74767 .30868 L
.74919 .31061 L
.75738 .31888 L
.75896 .32032 L
.76686 .33003 L
.75738 .33591 L
.74767 .33254 L
.74209 .33003 L
.73796 .32794 L
.72824 .32269 L
.7246 .32032 L
.71853 .31662 L
.71109 .31061 L
.70882 .30891 L
.70033 .30089 L
.69911 .29967 L
.69109 .29118 L
.68939 .28891 L
.68338 .28147 L
.67968 .2754 L
.67731 .27176 L
.67206 .26204 L
.66997 .25791 L
.66746 .25233 L
.66409 .24262 L
.66997 .23314 L
s
0 .16 1 r
.98077 .32811 m
.97106 .32694 L
.96134 .32385 L
.95163 .32086 L
.94985 .32032 L
.94192 .31833 L
.93221 .31501 L
.92249 .31144 L
.92026 .31061 L
.91278 .308 L
.90307 .30403 L
.89581 .30089 L
.89336 .2999 L
.88364 .29552 L
.87483 .29118 L
.87393 .29076 L
.86422 .28581 L
.85639 .28147 L
.85451 .28044 L
.84479 .27474 L
.83999 .27176 L
.83508 .26867 L
.82537 .2621 L
.82529 .26204 L
.81566 .25514 L
.81194 .25233 L
.80594 .24766 L
.79981 .24262 L
.79623 .23959 L
.78871 .23291 L
.78652 .23089 L
.77852 .22319 L
.77681 .22148 L
.76911 .21348 L
.76709 .21129 L
.76041 .20377 L
.75738 .20019 L
.75234 .19406 L
.74767 .18806 L
.74486 .18434 L
.73796 .17471 L
.7379 .17463 L
.73133 .16492 L
.72824 .16001 L
.72526 .15521 L
.71956 .14549 L
.71853 .14361 L
.71419 .13578 L
.70924 .12607 L
.70882 .12517 L
.70448 .11636 L
.7001 .10664 L
.69911 .10419 L
.69597 .09693 L
.692 .08722 L
.68939 .07974 L
.68856 .07751 L
.68499 .06779 L
.68167 .05808 L
.67968 .05015 L
.67914 .04837 L
.67615 .03866 L
.67306 .02894 L
.67189 .01923 L
.98077 .01923 L
F
0 g
.98077 .32811 m
.97106 .32694 L
.96134 .32385 L
.95163 .32086 L
.94985 .32032 L
.94192 .31833 L
.93221 .31501 L
.92249 .31144 L
.92026 .31061 L
.91278 .308 L
.90307 .30403 L
.89581 .30089 L
.89336 .2999 L
.88364 .29552 L
.87483 .29118 L
.87393 .29076 L
.86422 .28581 L
.85639 .28147 L
.85451 .28044 L
.84479 .27474 L
.83999 .27176 L
.83508 .26867 L
.82537 .2621 L
.82529 .26204 L
.81566 .25514 L
.81194 .25233 L
.80594 .24766 L
.79981 .24262 L
.79623 .23959 L
.78871 .23291 L
.78652 .23089 L
.77852 .22319 L
.77681 .22148 L
.76911 .21348 L
.76709 .21129 L
.76041 .20377 L
.75738 .20019 L
.75234 .19406 L
.74767 .18806 L
.74486 .18434 L
.73796 .17471 L
.7379 .17463 L
.73133 .16492 L
.72824 .16001 L
.72526 .15521 L
.71956 .14549 L
.71853 .14361 L
.71419 .13578 L
.70924 .12607 L
.70882 .12517 L
Mistroke
.70448 .11636 L
.7001 .10664 L
.69911 .10419 L
.69597 .09693 L
.692 .08722 L
.68939 .07974 L
.68856 .07751 L
.68499 .06779 L
.68167 .05808 L
.67968 .05015 L
.67914 .04837 L
.67615 .03866 L
.67306 .02894 L
.67189 .01923 L
Mfstroke
0 .16 1 r
.98077 .67189 m
.97106 .67306 L
.96134 .67615 L
.95163 .67914 L
.94985 .67968 L
.94192 .68167 L
.93221 .68499 L
.92249 .68856 L
.92026 .68939 L
.91278 .692 L
.90307 .69597 L
.89581 .69911 L
.89336 .7001 L
.88364 .70448 L
.87483 .70882 L
.87393 .70924 L
.86422 .71419 L
.85639 .71853 L
.85451 .71956 L
.84479 .72526 L
.83999 .72824 L
.83508 .73133 L
.82537 .7379 L
.82529 .73796 L
.81566 .74486 L
.81194 .74767 L
.80594 .75234 L
.79981 .75738 L
.79623 .76041 L
.78871 .76709 L
.78652 .76911 L
.77852 .77681 L
.77681 .77852 L
.76911 .78652 L
.76709 .78871 L
.76041 .79623 L
.75738 .79981 L
.75234 .80594 L
.74767 .81194 L
.74486 .81566 L
.73796 .82529 L
.7379 .82537 L
.73133 .83508 L
.72824 .83999 L
.72526 .84479 L
.71956 .85451 L
.71853 .85639 L
.71419 .86422 L
.70924 .87393 L
.70882 .87483 L
.70448 .88364 L
.7001 .89336 L
.69911 .89581 L
.69597 .90307 L
.692 .91278 L
.68939 .92026 L
.68856 .92249 L
.68499 .93221 L
.68167 .94192 L
.67968 .94985 L
.67914 .95163 L
.67615 .96134 L
.67306 .97106 L
.67189 .98077 L
.98077 .98077 L
F
0 g
.98077 .67189 m
.97106 .67306 L
.96134 .67615 L
.95163 .67914 L
.94985 .67968 L
.94192 .68167 L
.93221 .68499 L
.92249 .68856 L
.92026 .68939 L
.91278 .692 L
.90307 .69597 L
.89581 .69911 L
.89336 .7001 L
.88364 .70448 L
.87483 .70882 L
.87393 .70924 L
.86422 .71419 L
.85639 .71853 L
.85451 .71956 L
.84479 .72526 L
.83999 .72824 L
.83508 .73133 L
.82537 .7379 L
.82529 .73796 L
.81566 .74486 L
.81194 .74767 L
.80594 .75234 L
.79981 .75738 L
.79623 .76041 L
.78871 .76709 L
.78652 .76911 L
.77852 .77681 L
.77681 .77852 L
.76911 .78652 L
.76709 .78871 L
.76041 .79623 L
.75738 .79981 L
.75234 .80594 L
.74767 .81194 L
.74486 .81566 L
.73796 .82529 L
.7379 .82537 L
.73133 .83508 L
.72824 .83999 L
.72526 .84479 L
.71956 .85451 L
.71853 .85639 L
.71419 .86422 L
.70924 .87393 L
.70882 .87483 L
Mistroke
.70448 .88364 L
.7001 .89336 L
.69911 .89581 L
.69597 .90307 L
.692 .91278 L
.68939 .92026 L
.68856 .92249 L
.68499 .93221 L
.68167 .94192 L
.67968 .94985 L
.67914 .95163 L
.67615 .96134 L
.67306 .97106 L
.67189 .98077 L
Mfstroke
.08 0 1 r
.69911 .03687 m
.7007 .03866 L
.70848 .04837 L
.70882 .04879 L
.71539 .05808 L
.71853 .06252 L
.72347 .06779 L
.72824 .07675 L
.72883 .07751 L
.73668 .08722 L
.73796 .09072 L
.74278 .09693 L
.74767 .10436 L
.74943 .10664 L
.75673 .11636 L
.75738 .11747 L
.76424 .12607 L
.76709 .13001 L
.77185 .13578 L
.77681 .14202 L
.77976 .14549 L
.78652 .15348 L
.78804 .15521 L
.79623 .16441 L
.7967 .16492 L
.80577 .17463 L
.80594 .17482 L
.81526 .18434 L
.81566 .18474 L
.82518 .19406 L
.82537 .19423 L
.83508 .2033 L
.83559 .20377 L
.84479 .21196 L
.84652 .21348 L
.85451 .22024 L
.85798 .22319 L
.86422 .22815 L
.86999 .23291 L
.87393 .23576 L
.88253 .24262 L
.88364 .24327 L
.89336 .25057 L
.89564 .25233 L
.90307 .25722 L
.90928 .26204 L
.91278 .26332 L
.92249 .27117 L
.92325 .27176 L
.93221 .27653 L
.93748 .28147 L
.94192 .28461 L
.95121 .29118 L
.95163 .29152 L
.96134 .2993 L
.96313 .30089 L
.96134 .30984 L
.95163 .30927 L
.94192 .30749 L
.93221 .30511 L
.92249 .30255 L
.91778 .30089 L
.91278 .29907 L
.90307 .29607 L
.89336 .29166 L
.89227 .29118 L
.88364 .28782 L
.87393 .28289 L
.87113 .28147 L
.86422 .27817 L
.85451 .27262 L
.85302 .27176 L
.84479 .26705 L
.837 .26204 L
.83508 .26086 L
.82537 .2542 L
.82277 .25233 L
.81566 .24714 L
.80989 .24262 L
.80594 .23948 L
.79824 .23291 L
.79623 .23116 L
.78761 .22319 L
.78652 .22215 L
.77785 .21348 L
.77681 .21239 L
.76884 .20377 L
.76709 .20176 L
.76052 .19406 L
.75738 .19011 L
.75286 .18434 L
.74767 .17723 L
.7458 .17463 L
.73914 .16492 L
.73796 .163 L
.73295 .15521 L
.72824 .14698 L
.72738 .14549 L
.72183 .13578 L
.71853 .12887 L
.71711 .12607 L
.71218 .11636 L
.70882 .10773 L
.70834 .10664 L
.70393 .09693 L
.70093 .08722 L
.69911 .08222 L
.69745 .07751 L
.69489 .06779 L
.69251 .05808 L
.69073 .04837 L
.69016 .03866 L
F
0 g
.69911 .03687 m
.7007 .03866 L
.70848 .04837 L
.70882 .04879 L
.71539 .05808 L
.71853 .06252 L
.72347 .06779 L
.72824 .07675 L
.72883 .07751 L
.73668 .08722 L
.73796 .09072 L
.74278 .09693 L
.74767 .10436 L
.74943 .10664 L
.75673 .11636 L
.75738 .11747 L
.76424 .12607 L
.76709 .13001 L
.77185 .13578 L
.77681 .14202 L
.77976 .14549 L
.78652 .15348 L
.78804 .15521 L
.79623 .16441 L
.7967 .16492 L
.80577 .17463 L
.80594 .17482 L
.81526 .18434 L
.81566 .18474 L
.82518 .19406 L
.82537 .19423 L
.83508 .2033 L
.83559 .20377 L
.84479 .21196 L
.84652 .21348 L
.85451 .22024 L
.85798 .22319 L
.86422 .22815 L
.86999 .23291 L
.87393 .23576 L
.88253 .24262 L
.88364 .24327 L
.89336 .25057 L
.89564 .25233 L
.90307 .25722 L
.90928 .26204 L
.91278 .26332 L
.92249 .27117 L
.92325 .27176 L
.93221 .27653 L
Mistroke
.93748 .28147 L
.94192 .28461 L
.95121 .29118 L
.95163 .29152 L
.96134 .2993 L
.96313 .30089 L
.96134 .30984 L
.95163 .30927 L
.94192 .30749 L
.93221 .30511 L
.92249 .30255 L
.91778 .30089 L
.91278 .29907 L
.90307 .29607 L
.89336 .29166 L
.89227 .29118 L
.88364 .28782 L
.87393 .28289 L
.87113 .28147 L
.86422 .27817 L
.85451 .27262 L
.85302 .27176 L
.84479 .26705 L
.837 .26204 L
.83508 .26086 L
.82537 .2542 L
.82277 .25233 L
.81566 .24714 L
.80989 .24262 L
.80594 .23948 L
.79824 .23291 L
.79623 .23116 L
.78761 .22319 L
.78652 .22215 L
.77785 .21348 L
.77681 .21239 L
.76884 .20377 L
.76709 .20176 L
.76052 .19406 L
.75738 .19011 L
.75286 .18434 L
.74767 .17723 L
.7458 .17463 L
.73914 .16492 L
.73796 .163 L
.73295 .15521 L
.72824 .14698 L
.72738 .14549 L
.72183 .13578 L
.71853 .12887 L
Mistroke
.71711 .12607 L
.71218 .11636 L
.70882 .10773 L
.70834 .10664 L
.70393 .09693 L
.70093 .08722 L
.69911 .08222 L
.69745 .07751 L
.69489 .06779 L
.69251 .05808 L
.69073 .04837 L
.69016 .03866 L
.69911 .03687 L
Mfstroke
.08 0 1 r
.92249 .69745 m
.93221 .69489 L
.94192 .69251 L
.95163 .69073 L
.96134 .69016 L
.96313 .69911 L
.96134 .7007 L
.95163 .70848 L
.95121 .70882 L
.94192 .71539 L
.93748 .71853 L
.93221 .72347 L
.92325 .72824 L
.92249 .72883 L
.91278 .73668 L
.90928 .73796 L
.90307 .74278 L
.89564 .74767 L
.89336 .74943 L
.88364 .75673 L
.88253 .75738 L
.87393 .76424 L
.86999 .76709 L
.86422 .77185 L
.85798 .77681 L
.85451 .77976 L
.84652 .78652 L
.84479 .78804 L
.83559 .79623 L
.83508 .7967 L
.82537 .80577 L
.82518 .80594 L
.81566 .81526 L
.81526 .81566 L
.80594 .82518 L
.80577 .82537 L
.7967 .83508 L
.79623 .83559 L
.78804 .84479 L
.78652 .84652 L
.77976 .85451 L
.77681 .85798 L
.77185 .86422 L
.76709 .86999 L
.76424 .87393 L
.75738 .88253 L
.75673 .88364 L
.74943 .89336 L
.74767 .89564 L
.74278 .90307 L
.73796 .90928 L
.73668 .91278 L
.72883 .92249 L
.72824 .92325 L
.72347 .93221 L
.71853 .93748 L
.71539 .94192 L
.70882 .95121 L
.70848 .95163 L
.7007 .96134 L
.69911 .96313 L
.69016 .96134 L
.69073 .95163 L
.69251 .94192 L
.69489 .93221 L
.69745 .92249 L
.69911 .91778 L
.70093 .91278 L
.70393 .90307 L
.70834 .89336 L
.70882 .89227 L
.71218 .88364 L
.71711 .87393 L
.71853 .87113 L
.72183 .86422 L
.72738 .85451 L
.72824 .85302 L
.73295 .84479 L
.73796 .837 L
.73914 .83508 L
.7458 .82537 L
.74767 .82277 L
.75286 .81566 L
.75738 .80989 L
.76052 .80594 L
.76709 .79824 L
.76884 .79623 L
.77681 .78761 L
.77785 .78652 L
.78652 .77785 L
.78761 .77681 L
.79623 .76884 L
.79824 .76709 L
.80594 .76052 L
.80989 .75738 L
.81566 .75286 L
.82277 .74767 L
.82537 .7458 L
.83508 .73914 L
.837 .73796 L
.84479 .73295 L
.85302 .72824 L
.85451 .72738 L
.86422 .72183 L
.87113 .71853 L
.87393 .71711 L
.88364 .71218 L
.89227 .70882 L
.89336 .70834 L
.90307 .70393 L
.91278 .70093 L
.91778 .69911 L
F
0 g
.92249 .69745 m
.93221 .69489 L
.94192 .69251 L
.95163 .69073 L
.96134 .69016 L
.96313 .69911 L
.96134 .7007 L
.95163 .70848 L
.95121 .70882 L
.94192 .71539 L
.93748 .71853 L
.93221 .72347 L
.92325 .72824 L
.92249 .72883 L
.91278 .73668 L
.90928 .73796 L
.90307 .74278 L
.89564 .74767 L
.89336 .74943 L
.88364 .75673 L
.88253 .75738 L
.87393 .76424 L
.86999 .76709 L
.86422 .77185 L
.85798 .77681 L
.85451 .77976 L
.84652 .78652 L
.84479 .78804 L
.83559 .79623 L
.83508 .7967 L
.82537 .80577 L
.82518 .80594 L
.81566 .81526 L
.81526 .81566 L
.80594 .82518 L
.80577 .82537 L
.7967 .83508 L
.79623 .83559 L
.78804 .84479 L
.78652 .84652 L
.77976 .85451 L
.77681 .85798 L
.77185 .86422 L
.76709 .86999 L
.76424 .87393 L
.75738 .88253 L
.75673 .88364 L
.74943 .89336 L
.74767 .89564 L
.74278 .90307 L
Mistroke
.73796 .90928 L
.73668 .91278 L
.72883 .92249 L
.72824 .92325 L
.72347 .93221 L
.71853 .93748 L
.71539 .94192 L
.70882 .95121 L
.70848 .95163 L
.7007 .96134 L
.69911 .96313 L
.69016 .96134 L
.69073 .95163 L
.69251 .94192 L
.69489 .93221 L
.69745 .92249 L
.69911 .91778 L
.70093 .91278 L
.70393 .90307 L
.70834 .89336 L
.70882 .89227 L
.71218 .88364 L
.71711 .87393 L
.71853 .87113 L
.72183 .86422 L
.72738 .85451 L
.72824 .85302 L
.73295 .84479 L
.73796 .837 L
.73914 .83508 L
.7458 .82537 L
.74767 .82277 L
.75286 .81566 L
.75738 .80989 L
.76052 .80594 L
.76709 .79824 L
.76884 .79623 L
.77681 .78761 L
.77785 .78652 L
.78652 .77785 L
.78761 .77681 L
.79623 .76884 L
.79824 .76709 L
.80594 .76052 L
.80989 .75738 L
.81566 .75286 L
.82277 .74767 L
.82537 .7458 L
.83508 .73914 L
.837 .73796 L
Mistroke
.84479 .73295 L
.85302 .72824 L
.85451 .72738 L
.86422 .72183 L
.87113 .71853 L
.87393 .71711 L
.88364 .71218 L
.89227 .70882 L
.89336 .70834 L
.90307 .70393 L
.91278 .70093 L
.91778 .69911 L
.92249 .69745 L
Mfstroke
0 .4 1 r
.98077 .2954 m
.97155 .29118 L
.97106 .29087 L
.96134 .28471 L
.95474 .28147 L
.95163 .27952 L
.94192 .27351 L
.93851 .27176 L
.93221 .2677 L
.92297 .26204 L
.92249 .26173 L
.91278 .25523 L
.90821 .25233 L
.90307 .24875 L
.89418 .24262 L
.89336 .24202 L
.88364 .23488 L
.88091 .23291 L
.87393 .22753 L
.86832 .22319 L
.86422 .21988 L
.85639 .21348 L
.85451 .21189 L
.84507 .20377 L
.84479 .20352 L
.83508 .19472 L
.83436 .19406 L
.82537 .1855 L
.82418 .18434 L
.81566 .17582 L
.8145 .17463 L
.80594 .16564 L
.80528 .16492 L
.79648 .15521 L
.79623 .15493 L
.78811 .14549 L
.78652 .14361 L
.78012 .13578 L
.77681 .13168 L
.77247 .12607 L
.76709 .11909 L
.76512 .11636 L
.75798 .10664 L
.75738 .10582 L
.75125 .09693 L
.74767 .09179 L
.74477 .08722 L
.73827 .07751 L
.73796 .07703 L
.7323 .06779 L
.72824 .06149 L
.72649 .05808 L
.72048 .04837 L
.71853 .04526 L
.71529 .03866 L
.70913 .02894 L
.70882 .02845 L
.7046 .01923 L
.98077 .01923 L
F
0 g
.98077 .2954 m
.97155 .29118 L
.97106 .29087 L
.96134 .28471 L
.95474 .28147 L
.95163 .27952 L
.94192 .27351 L
.93851 .27176 L
.93221 .2677 L
.92297 .26204 L
.92249 .26173 L
.91278 .25523 L
.90821 .25233 L
.90307 .24875 L
.89418 .24262 L
.89336 .24202 L
.88364 .23488 L
.88091 .23291 L
.87393 .22753 L
.86832 .22319 L
.86422 .21988 L
.85639 .21348 L
.85451 .21189 L
.84507 .20377 L
.84479 .20352 L
.83508 .19472 L
.83436 .19406 L
.82537 .1855 L
.82418 .18434 L
.81566 .17582 L
.8145 .17463 L
.80594 .16564 L
.80528 .16492 L
.79648 .15521 L
.79623 .15493 L
.78811 .14549 L
.78652 .14361 L
.78012 .13578 L
.77681 .13168 L
.77247 .12607 L
.76709 .11909 L
.76512 .11636 L
.75798 .10664 L
.75738 .10582 L
.75125 .09693 L
.74767 .09179 L
.74477 .08722 L
.73827 .07751 L
.73796 .07703 L
.7323 .06779 L
Mistroke
.72824 .06149 L
.72649 .05808 L
.72048 .04837 L
.71853 .04526 L
.71529 .03866 L
.70913 .02894 L
.70882 .02845 L
.7046 .01923 L
Mfstroke
0 .4 1 r
.98077 .7046 m
.97155 .70882 L
.97106 .70913 L
.96134 .71529 L
.95474 .71853 L
.95163 .72048 L
.94192 .72649 L
.93851 .72824 L
.93221 .7323 L
.92297 .73796 L
.92249 .73827 L
.91278 .74477 L
.90821 .74767 L
.90307 .75125 L
.89418 .75738 L
.89336 .75798 L
.88364 .76512 L
.88091 .76709 L
.87393 .77247 L
.86832 .77681 L
.86422 .78012 L
.85639 .78652 L
.85451 .78811 L
.84507 .79623 L
.84479 .79648 L
.83508 .80528 L
.83436 .80594 L
.82537 .8145 L
.82418 .81566 L
.81566 .82418 L
.8145 .82537 L
.80594 .83436 L
.80528 .83508 L
.79648 .84479 L
.79623 .84507 L
.78811 .85451 L
.78652 .85639 L
.78012 .86422 L
.77681 .86832 L
.77247 .87393 L
.76709 .88091 L
.76512 .88364 L
.75798 .89336 L
.75738 .89418 L
.75125 .90307 L
.74767 .90821 L
.74477 .91278 L
.73827 .92249 L
.73796 .92297 L
.7323 .93221 L
.72824 .93851 L
.72649 .94192 L
.72048 .95163 L
.71853 .95474 L
.71529 .96134 L
.70913 .97106 L
.70882 .97155 L
.7046 .98077 L
.98077 .98077 L
F
0 g
.98077 .7046 m
.97155 .70882 L
.97106 .70913 L
.96134 .71529 L
.95474 .71853 L
.95163 .72048 L
.94192 .72649 L
.93851 .72824 L
.93221 .7323 L
.92297 .73796 L
.92249 .73827 L
.91278 .74477 L
.90821 .74767 L
.90307 .75125 L
.89418 .75738 L
.89336 .75798 L
.88364 .76512 L
.88091 .76709 L
.87393 .77247 L
.86832 .77681 L
.86422 .78012 L
.85639 .78652 L
.85451 .78811 L
.84507 .79623 L
.84479 .79648 L
.83508 .80528 L
.83436 .80594 L
.82537 .8145 L
.82418 .81566 L
.81566 .82418 L
.8145 .82537 L
.80594 .83436 L
.80528 .83508 L
.79648 .84479 L
.79623 .84507 L
.78811 .85451 L
.78652 .85639 L
.78012 .86422 L
.77681 .86832 L
.77247 .87393 L
.76709 .88091 L
.76512 .88364 L
.75798 .89336 L
.75738 .89418 L
.75125 .90307 L
.74767 .90821 L
.74477 .91278 L
.73827 .92249 L
.73796 .92297 L
.7323 .93221 L
Mistroke
.72824 .93851 L
.72649 .94192 L
.72048 .95163 L
.71853 .95474 L
.71529 .96134 L
.70913 .97106 L
.70882 .97155 L
.7046 .98077 L
Mfstroke
0 .64 1 r
.98077 .2876 m
.97106 .28251 L
.96891 .28147 L
.96134 .27725 L
.95163 .27191 L
.95133 .27176 L
.94192 .26622 L
.93474 .26204 L
.93221 .26048 L
.92249 .25443 L
.91911 .25233 L
.91278 .24819 L
.90438 .24262 L
.90307 .24172 L
.89336 .23492 L
.89051 .23291 L
.88364 .22785 L
.87743 .22319 L
.87393 .22048 L
.86509 .21348 L
.86422 .21277 L
.85451 .20467 L
.85344 .20377 L
.84479 .19618 L
.84243 .19406 L
.83508 .18726 L
.83201 .18434 L
.82537 .17788 L
.82212 .17463 L
.81566 .16799 L
.81274 .16492 L
.80594 .15757 L
.80382 .15521 L
.79623 .14656 L
.79533 .14549 L
.78723 .13578 L
.78652 .13491 L
.77952 .12607 L
.77681 .12257 L
.77215 .11636 L
.76709 .10949 L
.76508 .10664 L
.75828 .09693 L
.75738 .09562 L
.75181 .08722 L
.74767 .08089 L
.74557 .07751 L
.73952 .06779 L
.73796 .06526 L
.73378 .05808 L
.72824 .04867 L
.72809 .04837 L
.72275 .03866 L
.71853 .03109 L
.71749 .02894 L
.7124 .01923 L
.98077 .01923 L
F
0 g
.98077 .2876 m
.97106 .28251 L
.96891 .28147 L
.96134 .27725 L
.95163 .27191 L
.95133 .27176 L
.94192 .26622 L
.93474 .26204 L
.93221 .26048 L
.92249 .25443 L
.91911 .25233 L
.91278 .24819 L
.90438 .24262 L
.90307 .24172 L
.89336 .23492 L
.89051 .23291 L
.88364 .22785 L
.87743 .22319 L
.87393 .22048 L
.86509 .21348 L
.86422 .21277 L
.85451 .20467 L
.85344 .20377 L
.84479 .19618 L
.84243 .19406 L
.83508 .18726 L
.83201 .18434 L
.82537 .17788 L
.82212 .17463 L
.81566 .16799 L
.81274 .16492 L
.80594 .15757 L
.80382 .15521 L
.79623 .14656 L
.79533 .14549 L
.78723 .13578 L
.78652 .13491 L
.77952 .12607 L
.77681 .12257 L
.77215 .11636 L
.76709 .10949 L
.76508 .10664 L
.75828 .09693 L
.75738 .09562 L
.75181 .08722 L
.74767 .08089 L
.74557 .07751 L
.73952 .06779 L
.73796 .06526 L
.73378 .05808 L
Mistroke
.72824 .04867 L
.72809 .04837 L
.72275 .03866 L
.71853 .03109 L
.71749 .02894 L
.7124 .01923 L
Mfstroke
0 .64 1 r
.98077 .7124 m
.97106 .71749 L
.96891 .71853 L
.96134 .72275 L
.95163 .72809 L
.95133 .72824 L
.94192 .73378 L
.93474 .73796 L
.93221 .73952 L
.92249 .74557 L
.91911 .74767 L
.91278 .75181 L
.90438 .75738 L
.90307 .75828 L
.89336 .76508 L
.89051 .76709 L
.88364 .77215 L
.87743 .77681 L
.87393 .77952 L
.86509 .78652 L
.86422 .78723 L
.85451 .79533 L
.85344 .79623 L
.84479 .80382 L
.84243 .80594 L
.83508 .81274 L
.83201 .81566 L
.82537 .82212 L
.82212 .82537 L
.81566 .83201 L
.81274 .83508 L
.80594 .84243 L
.80382 .84479 L
.79623 .85344 L
.79533 .85451 L
.78723 .86422 L
.78652 .86509 L
.77952 .87393 L
.77681 .87743 L
.77215 .88364 L
.76709 .89051 L
.76508 .89336 L
.75828 .90307 L
.75738 .90438 L
.75181 .91278 L
.74767 .91911 L
.74557 .92249 L
.73952 .93221 L
.73796 .93474 L
.73378 .94192 L
.72824 .95133 L
.72809 .95163 L
.72275 .96134 L
.71853 .96891 L
.71749 .97106 L
.7124 .98077 L
.98077 .98077 L
F
0 g
.98077 .7124 m
.97106 .71749 L
.96891 .71853 L
.96134 .72275 L
.95163 .72809 L
.95133 .72824 L
.94192 .73378 L
.93474 .73796 L
.93221 .73952 L
.92249 .74557 L
.91911 .74767 L
.91278 .75181 L
.90438 .75738 L
.90307 .75828 L
.89336 .76508 L
.89051 .76709 L
.88364 .77215 L
.87743 .77681 L
.87393 .77952 L
.86509 .78652 L
.86422 .78723 L
.85451 .79533 L
.85344 .79623 L
.84479 .80382 L
.84243 .80594 L
.83508 .81274 L
.83201 .81566 L
.82537 .82212 L
.82212 .82537 L
.81566 .83201 L
.81274 .83508 L
.80594 .84243 L
.80382 .84479 L
.79623 .85344 L
.79533 .85451 L
.78723 .86422 L
.78652 .86509 L
.77952 .87393 L
.77681 .87743 L
.77215 .88364 L
.76709 .89051 L
.76508 .89336 L
.75828 .90307 L
.75738 .90438 L
.75181 .91278 L
.74767 .91911 L
.74557 .92249 L
.73952 .93221 L
.73796 .93474 L
.73378 .94192 L
Mistroke
.72824 .95133 L
.72809 .95163 L
.72275 .96134 L
.71853 .96891 L
.71749 .97106 L
.7124 .98077 L
Mfstroke
0 .88 1 r
.98077 .2809 m
.97106 .27587 L
.96317 .27176 L
.96134 .27077 L
.95163 .26539 L
.94565 .26204 L
.94192 .25987 L
.93221 .25411 L
.92925 .25233 L
.92249 .24812 L
.91388 .24262 L
.91278 .2419 L
.90307 .23537 L
.89947 .23291 L
.89336 .22858 L
.88593 .22319 L
.88364 .22149 L
.87393 .21405 L
.87321 .21348 L
.86422 .20624 L
.86123 .20377 L
.85451 .19805 L
.84993 .19406 L
.84479 .18944 L
.83926 .18434 L
.83508 .18038 L
.82918 .17463 L
.82537 .17082 L
.81962 .16492 L
.81566 .16074 L
.81056 .15521 L
.80594 .15007 L
.80195 .14549 L
.79623 .13877 L
.79376 .13578 L
.78652 .12679 L
.78595 .12607 L
.77851 .11636 L
.77681 .11407 L
.77142 .10664 L
.76709 .10053 L
.76463 .09693 L
.7581 .08722 L
.75738 .08612 L
.75188 .07751 L
.74767 .07075 L
.74589 .06779 L
.74013 .05808 L
.73796 .05435 L
.73461 .04837 L
.72923 .03866 L
.72824 .03683 L
.72413 .02894 L
.7191 .01923 L
.98077 .01923 L
F
0 g
.98077 .2809 m
.97106 .27587 L
.96317 .27176 L
.96134 .27077 L
.95163 .26539 L
.94565 .26204 L
.94192 .25987 L
.93221 .25411 L
.92925 .25233 L
.92249 .24812 L
.91388 .24262 L
.91278 .2419 L
.90307 .23537 L
.89947 .23291 L
.89336 .22858 L
.88593 .22319 L
.88364 .22149 L
.87393 .21405 L
.87321 .21348 L
.86422 .20624 L
.86123 .20377 L
.85451 .19805 L
.84993 .19406 L
.84479 .18944 L
.83926 .18434 L
.83508 .18038 L
.82918 .17463 L
.82537 .17082 L
.81962 .16492 L
.81566 .16074 L
.81056 .15521 L
.80594 .15007 L
.80195 .14549 L
.79623 .13877 L
.79376 .13578 L
.78652 .12679 L
.78595 .12607 L
.77851 .11636 L
.77681 .11407 L
.77142 .10664 L
.76709 .10053 L
.76463 .09693 L
.7581 .08722 L
.75738 .08612 L
.75188 .07751 L
.74767 .07075 L
.74589 .06779 L
.74013 .05808 L
.73796 .05435 L
.73461 .04837 L
Mistroke
.72923 .03866 L
.72824 .03683 L
.72413 .02894 L
.7191 .01923 L
Mfstroke
0 .88 1 r
.98077 .7191 m
.97106 .72413 L
.96317 .72824 L
.96134 .72923 L
.95163 .73461 L
.94565 .73796 L
.94192 .74013 L
.93221 .74589 L
.92925 .74767 L
.92249 .75188 L
.91388 .75738 L
.91278 .7581 L
.90307 .76463 L
.89947 .76709 L
.89336 .77142 L
.88593 .77681 L
.88364 .77851 L
.87393 .78595 L
.87321 .78652 L
.86422 .79376 L
.86123 .79623 L
.85451 .80195 L
.84993 .80594 L
.84479 .81056 L
.83926 .81566 L
.83508 .81962 L
.82918 .82537 L
.82537 .82918 L
.81962 .83508 L
.81566 .83926 L
.81056 .84479 L
.80594 .84993 L
.80195 .85451 L
.79623 .86123 L
.79376 .86422 L
.78652 .87321 L
.78595 .87393 L
.77851 .88364 L
.77681 .88593 L
.77142 .89336 L
.76709 .89947 L
.76463 .90307 L
.7581 .91278 L
.75738 .91388 L
.75188 .92249 L
.74767 .92925 L
.74589 .93221 L
.74013 .94192 L
.73796 .94565 L
.73461 .95163 L
.72923 .96134 L
.72824 .96317 L
.72413 .97106 L
.7191 .98077 L
.98077 .98077 L
F
0 g
.98077 .7191 m
.97106 .72413 L
.96317 .72824 L
.96134 .72923 L
.95163 .73461 L
.94565 .73796 L
.94192 .74013 L
.93221 .74589 L
.92925 .74767 L
.92249 .75188 L
.91388 .75738 L
.91278 .7581 L
.90307 .76463 L
.89947 .76709 L
.89336 .77142 L
.88593 .77681 L
.88364 .77851 L
.87393 .78595 L
.87321 .78652 L
.86422 .79376 L
.86123 .79623 L
.85451 .80195 L
.84993 .80594 L
.84479 .81056 L
.83926 .81566 L
.83508 .81962 L
.82918 .82537 L
.82537 .82918 L
.81962 .83508 L
.81566 .83926 L
.81056 .84479 L
.80594 .84993 L
.80195 .85451 L
.79623 .86123 L
.79376 .86422 L
.78652 .87321 L
.78595 .87393 L
.77851 .88364 L
.77681 .88593 L
.77142 .89336 L
.76709 .89947 L
.76463 .90307 L
.7581 .91278 L
.75738 .91388 L
.75188 .92249 L
.74767 .92925 L
.74589 .93221 L
.74013 .94192 L
.73796 .94565 L
.73461 .95163 L
Mistroke
.72923 .96134 L
.72824 .96317 L
.72413 .97106 L
.7191 .98077 L
Mfstroke
.32 0 1 r
.88364 .7258 m
.89336 .72365 L
.89765 .72824 L
.89336 .73362 L
.89007 .73796 L
.88364 .74287 L
.87859 .74767 L
.87393 .75133 L
.86706 .75738 L
.86422 .76047 L
.85565 .76709 L
.85451 .76821 L
.84479 .77657 L
.8446 .77681 L
.83508 .78533 L
.83406 .78652 L
.82537 .79509 L
.82377 .79623 L
.81566 .80426 L
.81383 .80594 L
.80594 .81383 L
.80426 .81566 L
.79623 .82377 L
.79509 .82537 L
.78652 .83406 L
.78533 .83508 L
.77681 .8446 L
.77657 .84479 L
.76821 .85451 L
.76709 .85565 L
.76047 .86422 L
.75738 .86706 L
.75133 .87393 L
.74767 .87859 L
.74287 .88364 L
.73796 .89007 L
.73362 .89336 L
.72824 .89765 L
.72365 .89336 L
.7258 .88364 L
.72824 .87664 L
.72932 .87393 L
.73361 .86422 L
.73796 .85551 L
.73849 .85451 L
.74312 .84479 L
.74767 .83792 L
.74863 .83508 L
.75646 .82537 L
.75738 .8241 L
.76338 .81566 L
.76709 .81105 L
.77109 .80594 L
.77681 .7995 L
.77967 .79623 L
.78652 .78912 L
.78912 .78652 L
.79623 .77967 L
.7995 .77681 L
.80594 .77109 L
.81105 .76709 L
.81566 .76338 L
.8241 .75738 L
.82537 .75646 L
.83508 .74863 L
.83792 .74767 L
.84479 .74312 L
.85451 .73849 L
.85551 .73796 L
.86422 .73361 L
.87393 .72932 L
.87664 .72824 L
F
0 g
.88364 .7258 m
.89336 .72365 L
.89765 .72824 L
.89336 .73362 L
.89007 .73796 L
.88364 .74287 L
.87859 .74767 L
.87393 .75133 L
.86706 .75738 L
.86422 .76047 L
.85565 .76709 L
.85451 .76821 L
.84479 .77657 L
.8446 .77681 L
.83508 .78533 L
.83406 .78652 L
.82537 .79509 L
.82377 .79623 L
.81566 .80426 L
.81383 .80594 L
.80594 .81383 L
.80426 .81566 L
.79623 .82377 L
.79509 .82537 L
.78652 .83406 L
.78533 .83508 L
.77681 .8446 L
.77657 .84479 L
.76821 .85451 L
.76709 .85565 L
.76047 .86422 L
.75738 .86706 L
.75133 .87393 L
.74767 .87859 L
.74287 .88364 L
.73796 .89007 L
.73362 .89336 L
.72824 .89765 L
.72365 .89336 L
.7258 .88364 L
.72824 .87664 L
.72932 .87393 L
.73361 .86422 L
.73796 .85551 L
.73849 .85451 L
.74312 .84479 L
.74767 .83792 L
.74863 .83508 L
.75646 .82537 L
.75738 .8241 L
Mistroke
.76338 .81566 L
.76709 .81105 L
.77109 .80594 L
.77681 .7995 L
.77967 .79623 L
.78652 .78912 L
.78912 .78652 L
.79623 .77967 L
.7995 .77681 L
.80594 .77109 L
.81105 .76709 L
.81566 .76338 L
.8241 .75738 L
.82537 .75646 L
.83508 .74863 L
.83792 .74767 L
.84479 .74312 L
.85451 .73849 L
.85551 .73796 L
.86422 .73361 L
.87393 .72932 L
.87664 .72824 L
.88364 .7258 L
Mfstroke
.32 0 1 r
.72824 .10235 m
.73362 .10664 L
.73796 .10993 L
.74287 .11636 L
.74767 .12141 L
.75133 .12607 L
.75738 .13294 L
.76047 .13578 L
.76709 .14435 L
.76821 .14549 L
.77657 .15521 L
.77681 .1554 L
.78533 .16492 L
.78652 .16594 L
.79509 .17463 L
.79623 .17623 L
.80426 .18434 L
.80594 .18617 L
.81383 .19406 L
.81566 .19574 L
.82377 .20377 L
.82537 .20491 L
.83406 .21348 L
.83508 .21467 L
.8446 .22319 L
.84479 .22343 L
.85451 .23179 L
.85565 .23291 L
.86422 .23953 L
.86706 .24262 L
.87393 .24867 L
.87859 .25233 L
.88364 .25713 L
.89007 .26204 L
.89336 .26638 L
.89765 .27176 L
.89336 .27635 L
.88364 .2742 L
.87664 .27176 L
.87393 .27068 L
.86422 .26639 L
.85551 .26204 L
.85451 .26151 L
.84479 .25688 L
.83792 .25233 L
.83508 .25137 L
.82537 .24354 L
.8241 .24262 L
.81566 .23662 L
.81105 .23291 L
.80594 .22891 L
.7995 .22319 L
.79623 .22033 L
.78912 .21348 L
.78652 .21088 L
.77967 .20377 L
.77681 .2005 L
.77109 .19406 L
.76709 .18895 L
.76338 .18434 L
.75738 .1759 L
.75646 .17463 L
.74863 .16492 L
.74767 .16208 L
.74312 .15521 L
.73849 .14549 L
.73796 .14449 L
.73361 .13578 L
.72932 .12607 L
.72824 .12336 L
.7258 .11636 L
.72365 .10664 L
F
0 g
.72824 .10235 m
.73362 .10664 L
.73796 .10993 L
.74287 .11636 L
.74767 .12141 L
.75133 .12607 L
.75738 .13294 L
.76047 .13578 L
.76709 .14435 L
.76821 .14549 L
.77657 .15521 L
.77681 .1554 L
.78533 .16492 L
.78652 .16594 L
.79509 .17463 L
.79623 .17623 L
.80426 .18434 L
.80594 .18617 L
.81383 .19406 L
.81566 .19574 L
.82377 .20377 L
.82537 .20491 L
.83406 .21348 L
.83508 .21467 L
.8446 .22319 L
.84479 .22343 L
.85451 .23179 L
.85565 .23291 L
.86422 .23953 L
.86706 .24262 L
.87393 .24867 L
.87859 .25233 L
.88364 .25713 L
.89007 .26204 L
.89336 .26638 L
.89765 .27176 L
.89336 .27635 L
.88364 .2742 L
.87664 .27176 L
.87393 .27068 L
.86422 .26639 L
.85551 .26204 L
.85451 .26151 L
.84479 .25688 L
.83792 .25233 L
.83508 .25137 L
.82537 .24354 L
.8241 .24262 L
.81566 .23662 L
.81105 .23291 L
Mistroke
.80594 .22891 L
.7995 .22319 L
.79623 .22033 L
.78912 .21348 L
.78652 .21088 L
.77967 .20377 L
.77681 .2005 L
.77109 .19406 L
.76709 .18895 L
.76338 .18434 L
.75738 .1759 L
.75646 .17463 L
.74863 .16492 L
.74767 .16208 L
.74312 .15521 L
.73849 .14549 L
.73796 .14449 L
.73361 .13578 L
.72932 .12607 L
.72824 .12336 L
.7258 .11636 L
.72365 .10664 L
.72824 .10235 L
Mfstroke
0 1 .88 r
.98077 .27481 m
.97465 .27176 L
.97106 .26991 L
.96134 .26479 L
.95623 .26204 L
.95163 .2595 L
.94192 .25397 L
.93908 .25233 L
.93221 .24823 L
.92309 .24262 L
.92249 .24224 L
.91278 .23598 L
.90814 .23291 L
.90307 .22945 L
.89416 .22319 L
.89336 .22261 L
.88364 .21544 L
.88105 .21348 L
.87393 .20793 L
.86874 .20377 L
.86422 .20004 L
.85716 .19406 L
.85451 .19174 L
.84625 .18434 L
.84479 .18301 L
.83595 .17463 L
.83508 .17379 L
.82621 .16492 L
.82537 .16405 L
.81699 .15521 L
.81566 .15375 L
.80826 .14549 L
.80594 .14284 L
.79996 .13578 L
.79623 .13126 L
.79207 .12607 L
.78652 .11895 L
.78456 .11636 L
.77739 .10664 L
.77681 .10584 L
.77055 .09693 L
.76709 .09186 L
.76402 .08722 L
.75776 .07751 L
.75738 .07691 L
.75177 .06779 L
.74767 .06092 L
.74603 .05808 L
.7405 .04837 L
.73796 .04377 L
.73521 .03866 L
.73009 .02894 L
.72824 .02535 L
.72519 .01923 L
.98077 .01923 L
F
0 g
.98077 .27481 m
.97465 .27176 L
.97106 .26991 L
.96134 .26479 L
.95623 .26204 L
.95163 .2595 L
.94192 .25397 L
.93908 .25233 L
.93221 .24823 L
.92309 .24262 L
.92249 .24224 L
.91278 .23598 L
.90814 .23291 L
.90307 .22945 L
.89416 .22319 L
.89336 .22261 L
.88364 .21544 L
.88105 .21348 L
.87393 .20793 L
.86874 .20377 L
.86422 .20004 L
.85716 .19406 L
.85451 .19174 L
.84625 .18434 L
.84479 .18301 L
.83595 .17463 L
.83508 .17379 L
.82621 .16492 L
.82537 .16405 L
.81699 .15521 L
.81566 .15375 L
.80826 .14549 L
.80594 .14284 L
.79996 .13578 L
.79623 .13126 L
.79207 .12607 L
.78652 .11895 L
.78456 .11636 L
.77739 .10664 L
.77681 .10584 L
.77055 .09693 L
.76709 .09186 L
.76402 .08722 L
.75776 .07751 L
.75738 .07691 L
.75177 .06779 L
.74767 .06092 L
.74603 .05808 L
.7405 .04837 L
.73796 .04377 L
Mistroke
.73521 .03866 L
.73009 .02894 L
.72824 .02535 L
.72519 .01923 L
Mfstroke
0 1 .88 r
.98077 .72519 m
.97465 .72824 L
.97106 .73009 L
.96134 .73521 L
.95623 .73796 L
.95163 .7405 L
.94192 .74603 L
.93908 .74767 L
.93221 .75177 L
.92309 .75738 L
.92249 .75776 L
.91278 .76402 L
.90814 .76709 L
.90307 .77055 L
.89416 .77681 L
.89336 .77739 L
.88364 .78456 L
.88105 .78652 L
.87393 .79207 L
.86874 .79623 L
.86422 .79996 L
.85716 .80594 L
.85451 .80826 L
.84625 .81566 L
.84479 .81699 L
.83595 .82537 L
.83508 .82621 L
.82621 .83508 L
.82537 .83595 L
.81699 .84479 L
.81566 .84625 L
.80826 .85451 L
.80594 .85716 L
.79996 .86422 L
.79623 .86874 L
.79207 .87393 L
.78652 .88105 L
.78456 .88364 L
.77739 .89336 L
.77681 .89416 L
.77055 .90307 L
.76709 .90814 L
.76402 .91278 L
.75776 .92249 L
.75738 .92309 L
.75177 .93221 L
.74767 .93908 L
.74603 .94192 L
.7405 .95163 L
.73796 .95623 L
.73521 .96134 L
.73009 .97106 L
.72824 .97465 L
.72519 .98077 L
.98077 .98077 L
F
0 g
.98077 .72519 m
.97465 .72824 L
.97106 .73009 L
.96134 .73521 L
.95623 .73796 L
.95163 .7405 L
.94192 .74603 L
.93908 .74767 L
.93221 .75177 L
.92309 .75738 L
.92249 .75776 L
.91278 .76402 L
.90814 .76709 L
.90307 .77055 L
.89416 .77681 L
.89336 .77739 L
.88364 .78456 L
.88105 .78652 L
.87393 .79207 L
.86874 .79623 L
.86422 .79996 L
.85716 .80594 L
.85451 .80826 L
.84625 .81566 L
.84479 .81699 L
.83595 .82537 L
.83508 .82621 L
.82621 .83508 L
.82537 .83595 L
.81699 .84479 L
.81566 .84625 L
.80826 .85451 L
.80594 .85716 L
.79996 .86422 L
.79623 .86874 L
.79207 .87393 L
.78652 .88105 L
.78456 .88364 L
.77739 .89336 L
.77681 .89416 L
.77055 .90307 L
.76709 .90814 L
.76402 .91278 L
.75776 .92249 L
.75738 .92309 L
.75177 .93221 L
.74767 .93908 L
.74603 .94192 L
.7405 .95163 L
.73796 .95623 L
Mistroke
.73521 .96134 L
.73009 .97106 L
.72824 .97465 L
.72519 .98077 L
Mfstroke
0 1 .64 r
.98077 .26909 m
.97106 .2642 L
.96687 .26204 L
.96134 .25912 L
.95163 .25383 L
.94895 .25233 L
.94192 .24831 L
.9323 .24262 L
.93221 .24256 L
.92249 .23655 L
.91681 .23291 L
.91278 .23027 L
.90307 .22369 L
.90236 .22319 L
.89336 .21679 L
.88885 .21348 L
.88364 .20956 L
.8762 .20377 L
.87393 .20196 L
.86432 .19406 L
.86422 .19397 L
.85451 .18555 L
.85316 .18434 L
.84479 .17667 L
.84264 .17463 L
.83508 .16729 L
.83271 .16492 L
.82537 .15736 L
.82333 .15521 L
.81566 .14684 L
.81445 .14549 L
.80603 .13578 L
.80594 .13568 L
.79804 .12607 L
.79623 .1238 L
.79044 .11636 L
.78652 .11115 L
.78321 .10664 L
.77681 .09764 L
.77631 .09693 L
.76973 .08722 L
.76709 .08319 L
.76345 .07751 L
.75744 .06779 L
.75738 .0677 L
.75169 .05808 L
.74767 .05105 L
.74617 .04837 L
.74088 .03866 L
.73796 .03313 L
.7358 .02894 L
.73091 .01923 L
.98077 .01923 L
F
0 g
.98077 .26909 m
.97106 .2642 L
.96687 .26204 L
.96134 .25912 L
.95163 .25383 L
.94895 .25233 L
.94192 .24831 L
.9323 .24262 L
.93221 .24256 L
.92249 .23655 L
.91681 .23291 L
.91278 .23027 L
.90307 .22369 L
.90236 .22319 L
.89336 .21679 L
.88885 .21348 L
.88364 .20956 L
.8762 .20377 L
.87393 .20196 L
.86432 .19406 L
.86422 .19397 L
.85451 .18555 L
.85316 .18434 L
.84479 .17667 L
.84264 .17463 L
.83508 .16729 L
.83271 .16492 L
.82537 .15736 L
.82333 .15521 L
.81566 .14684 L
.81445 .14549 L
.80603 .13578 L
.80594 .13568 L
.79804 .12607 L
.79623 .1238 L
.79044 .11636 L
.78652 .11115 L
.78321 .10664 L
.77681 .09764 L
.77631 .09693 L
.76973 .08722 L
.76709 .08319 L
.76345 .07751 L
.75744 .06779 L
.75738 .0677 L
.75169 .05808 L
.74767 .05105 L
.74617 .04837 L
.74088 .03866 L
.73796 .03313 L
Mistroke
.7358 .02894 L
.73091 .01923 L
Mfstroke
0 1 .64 r
.98077 .73091 m
.97106 .7358 L
.96687 .73796 L
.96134 .74088 L
.95163 .74617 L
.94895 .74767 L
.94192 .75169 L
.9323 .75738 L
.93221 .75744 L
.92249 .76345 L
.91681 .76709 L
.91278 .76973 L
.90307 .77631 L
.90236 .77681 L
.89336 .78321 L
.88885 .78652 L
.88364 .79044 L
.8762 .79623 L
.87393 .79804 L
.86432 .80594 L
.86422 .80603 L
.85451 .81445 L
.85316 .81566 L
.84479 .82333 L
.84264 .82537 L
.83508 .83271 L
.83271 .83508 L
.82537 .84264 L
.82333 .84479 L
.81566 .85316 L
.81445 .85451 L
.80603 .86422 L
.80594 .86432 L
.79804 .87393 L
.79623 .8762 L
.79044 .88364 L
.78652 .88885 L
.78321 .89336 L
.77681 .90236 L
.77631 .90307 L
.76973 .91278 L
.76709 .91681 L
.76345 .92249 L
.75744 .93221 L
.75738 .9323 L
.75169 .94192 L
.74767 .94895 L
.74617 .95163 L
.74088 .96134 L
.73796 .96687 L
.7358 .97106 L
.73091 .98077 L
.98077 .98077 L
F
0 g
.98077 .73091 m
.97106 .7358 L
.96687 .73796 L
.96134 .74088 L
.95163 .74617 L
.94895 .74767 L
.94192 .75169 L
.9323 .75738 L
.93221 .75744 L
.92249 .76345 L
.91681 .76709 L
.91278 .76973 L
.90307 .77631 L
.90236 .77681 L
.89336 .78321 L
.88885 .78652 L
.88364 .79044 L
.8762 .79623 L
.87393 .79804 L
.86432 .80594 L
.86422 .80603 L
.85451 .81445 L
.85316 .81566 L
.84479 .82333 L
.84264 .82537 L
.83508 .83271 L
.83271 .83508 L
.82537 .84264 L
.82333 .84479 L
.81566 .85316 L
.81445 .85451 L
.80603 .86422 L
.80594 .86432 L
.79804 .87393 L
.79623 .8762 L
.79044 .88364 L
.78652 .88885 L
.78321 .89336 L
.77681 .90236 L
.77631 .90307 L
.76973 .91278 L
.76709 .91681 L
.76345 .92249 L
.75744 .93221 L
.75738 .9323 L
.75169 .94192 L
.74767 .94895 L
.74617 .95163 L
.74088 .96134 L
.73796 .96687 L
Mistroke
.7358 .97106 L
.73091 .98077 L
Mfstroke
0 1 .4 r
.98077 .26343 m
.97796 .26204 L
.97106 .25858 L
.96134 .2535 L
.95916 .25233 L
.95163 .24822 L
.94192 .24269 L
.9418 .24262 L
.93221 .23692 L
.9257 .23291 L
.92249 .23089 L
.91278 .22456 L
.91075 .22319 L
.90307 .21793 L
.8968 .21348 L
.89336 .21098 L
.88379 .20377 L
.88364 .20366 L
.87393 .19597 L
.87159 .19406 L
.86422 .18787 L
.86015 .18434 L
.85451 .17932 L
.84939 .17463 L
.84479 .17029 L
.83927 .16492 L
.83508 .16073 L
.82971 .15521 L
.82537 .15061 L
.82068 .14549 L
.81566 .13985 L
.81213 .13578 L
.80594 .12841 L
.80403 .12607 L
.79634 .11636 L
.79623 .11621 L
.78902 .10664 L
.78652 .1032 L
.78207 .09693 L
.77681 .08925 L
.77544 .08722 L
.76911 .07751 L
.76709 .0743 L
.76308 .06779 L
.75738 .0582 L
.75731 .05808 L
.75178 .04837 L
.74767 .04084 L
.7465 .03866 L
.74142 .02894 L
.73796 .02204 L
.73657 .01923 L
.98077 .01923 L
F
0 g
.98077 .26343 m
.97796 .26204 L
.97106 .25858 L
.96134 .2535 L
.95916 .25233 L
.95163 .24822 L
.94192 .24269 L
.9418 .24262 L
.93221 .23692 L
.9257 .23291 L
.92249 .23089 L
.91278 .22456 L
.91075 .22319 L
.90307 .21793 L
.8968 .21348 L
.89336 .21098 L
.88379 .20377 L
.88364 .20366 L
.87393 .19597 L
.87159 .19406 L
.86422 .18787 L
.86015 .18434 L
.85451 .17932 L
.84939 .17463 L
.84479 .17029 L
.83927 .16492 L
.83508 .16073 L
.82971 .15521 L
.82537 .15061 L
.82068 .14549 L
.81566 .13985 L
.81213 .13578 L
.80594 .12841 L
.80403 .12607 L
.79634 .11636 L
.79623 .11621 L
.78902 .10664 L
.78652 .1032 L
.78207 .09693 L
.77681 .08925 L
.77544 .08722 L
.76911 .07751 L
.76709 .0743 L
.76308 .06779 L
.75738 .0582 L
.75731 .05808 L
.75178 .04837 L
.74767 .04084 L
.7465 .03866 L
.74142 .02894 L
Mistroke
.73796 .02204 L
.73657 .01923 L
Mfstroke
0 1 .4 r
.98077 .73657 m
.97796 .73796 L
.97106 .74142 L
.96134 .7465 L
.95916 .74767 L
.95163 .75178 L
.94192 .75731 L
.9418 .75738 L
.93221 .76308 L
.9257 .76709 L
.92249 .76911 L
.91278 .77544 L
.91075 .77681 L
.90307 .78207 L
.8968 .78652 L
.89336 .78902 L
.88379 .79623 L
.88364 .79634 L
.87393 .80403 L
.87159 .80594 L
.86422 .81213 L
.86015 .81566 L
.85451 .82068 L
.84939 .82537 L
.84479 .82971 L
.83927 .83508 L
.83508 .83927 L
.82971 .84479 L
.82537 .84939 L
.82068 .85451 L
.81566 .86015 L
.81213 .86422 L
.80594 .87159 L
.80403 .87393 L
.79634 .88364 L
.79623 .88379 L
.78902 .89336 L
.78652 .8968 L
.78207 .90307 L
.77681 .91075 L
.77544 .91278 L
.76911 .92249 L
.76709 .9257 L
.76308 .93221 L
.75738 .9418 L
.75731 .94192 L
.75178 .95163 L
.74767 .95916 L
.7465 .96134 L
.74142 .97106 L
.73796 .97796 L
.73657 .98077 L
.98077 .98077 L
F
0 g
.98077 .73657 m
.97796 .73796 L
.97106 .74142 L
.96134 .7465 L
.95916 .74767 L
.95163 .75178 L
.94192 .75731 L
.9418 .75738 L
.93221 .76308 L
.9257 .76709 L
.92249 .76911 L
.91278 .77544 L
.91075 .77681 L
.90307 .78207 L
.8968 .78652 L
.89336 .78902 L
.88379 .79623 L
.88364 .79634 L
.87393 .80403 L
.87159 .80594 L
.86422 .81213 L
.86015 .81566 L
.85451 .82068 L
.84939 .82537 L
.84479 .82971 L
.83927 .83508 L
.83508 .83927 L
.82971 .84479 L
.82537 .84939 L
.82068 .85451 L
.81566 .86015 L
.81213 .86422 L
.80594 .87159 L
.80403 .87393 L
.79634 .88364 L
.79623 .88379 L
.78902 .89336 L
.78652 .8968 L
.78207 .90307 L
.77681 .91075 L
.77544 .91278 L
.76911 .92249 L
.76709 .9257 L
.76308 .93221 L
.75738 .9418 L
.75731 .94192 L
.75178 .95163 L
.74767 .95916 L
.7465 .96134 L
.74142 .97106 L
Mistroke
.73796 .97796 L
.73657 .98077 L
Mfstroke
0 1 .16 r
.98077 .25767 m
.97106 .2528 L
.97014 .25233 L
.96134 .24777 L
.95192 .24262 L
.95163 .24246 L
.94192 .23694 L
.93512 .23291 L
.93221 .23115 L
.92249 .22507 L
.91958 .22319 L
.91278 .21872 L
.90515 .21348 L
.90307 .21203 L
.89336 .20499 L
.89171 .20377 L
.88364 .19761 L
.87916 .19406 L
.87393 .18982 L
.86742 .18434 L
.86422 .18159 L
.8564 .17463 L
.85451 .1729 L
.84605 .16492 L
.84479 .1637 L
.8363 .15521 L
.83508 .15395 L
.8271 .14549 L
.82537 .1436 L
.81841 .13578 L
.81566 .13258 L
.81018 .12607 L
.80594 .12084 L
.80239 .11636 L
.79623 .10829 L
.79501 .10664 L
.78797 .09693 L
.78652 .09485 L
.78128 .08722 L
.77681 .08042 L
.77493 .07751 L
.76885 .06779 L
.76709 .06488 L
.76306 .05808 L
.75754 .04837 L
.75738 .04808 L
.75223 .03866 L
.74767 .02986 L
.7472 .02894 L
.74233 .01923 L
.98077 .01923 L
F
0 g
.98077 .25767 m
.97106 .2528 L
.97014 .25233 L
.96134 .24777 L
.95192 .24262 L
.95163 .24246 L
.94192 .23694 L
.93512 .23291 L
.93221 .23115 L
.92249 .22507 L
.91958 .22319 L
.91278 .21872 L
.90515 .21348 L
.90307 .21203 L
.89336 .20499 L
.89171 .20377 L
.88364 .19761 L
.87916 .19406 L
.87393 .18982 L
.86742 .18434 L
.86422 .18159 L
.8564 .17463 L
.85451 .1729 L
.84605 .16492 L
.84479 .1637 L
.8363 .15521 L
.83508 .15395 L
.8271 .14549 L
.82537 .1436 L
.81841 .13578 L
.81566 .13258 L
.81018 .12607 L
.80594 .12084 L
.80239 .11636 L
.79623 .10829 L
.79501 .10664 L
.78797 .09693 L
.78652 .09485 L
.78128 .08722 L
.77681 .08042 L
.77493 .07751 L
.76885 .06779 L
.76709 .06488 L
.76306 .05808 L
.75754 .04837 L
.75738 .04808 L
.75223 .03866 L
.74767 .02986 L
.7472 .02894 L
.74233 .01923 L
Mistroke
Mfstroke
0 1 .16 r
.98077 .74233 m
.97106 .7472 L
.97014 .74767 L
.96134 .75223 L
.95192 .75738 L
.95163 .75754 L
.94192 .76306 L
.93512 .76709 L
.93221 .76885 L
.92249 .77493 L
.91958 .77681 L
.91278 .78128 L
.90515 .78652 L
.90307 .78797 L
.89336 .79501 L
.89171 .79623 L
.88364 .80239 L
.87916 .80594 L
.87393 .81018 L
.86742 .81566 L
.86422 .81841 L
.8564 .82537 L
.85451 .8271 L
.84605 .83508 L
.84479 .8363 L
.8363 .84479 L
.83508 .84605 L
.8271 .85451 L
.82537 .8564 L
.81841 .86422 L
.81566 .86742 L
.81018 .87393 L
.80594 .87916 L
.80239 .88364 L
.79623 .89171 L
.79501 .89336 L
.78797 .90307 L
.78652 .90515 L
.78128 .91278 L
.77681 .91958 L
.77493 .92249 L
.76885 .93221 L
.76709 .93512 L
.76306 .94192 L
.75754 .95163 L
.75738 .95192 L
.75223 .96134 L
.74767 .97014 L
.7472 .97106 L
.74233 .98077 L
.98077 .98077 L
F
0 g
.98077 .74233 m
.97106 .7472 L
.97014 .74767 L
.96134 .75223 L
.95192 .75738 L
.95163 .75754 L
.94192 .76306 L
.93512 .76709 L
.93221 .76885 L
.92249 .77493 L
.91958 .77681 L
.91278 .78128 L
.90515 .78652 L
.90307 .78797 L
.89336 .79501 L
.89171 .79623 L
.88364 .80239 L
.87916 .80594 L
.87393 .81018 L
.86742 .81566 L
.86422 .81841 L
.8564 .82537 L
.85451 .8271 L
.84605 .83508 L
.84479 .8363 L
.8363 .84479 L
.83508 .84605 L
.8271 .85451 L
.82537 .8564 L
.81841 .86422 L
.81566 .86742 L
.81018 .87393 L
.80594 .87916 L
.80239 .88364 L
.79623 .89171 L
.79501 .89336 L
.78797 .90307 L
.78652 .90515 L
.78128 .91278 L
.77681 .91958 L
.77493 .92249 L
.76885 .93221 L
.76709 .93512 L
.76306 .94192 L
.75754 .95163 L
.75738 .95192 L
.75223 .96134 L
.74767 .97014 L
.7472 .97106 L
.74233 .98077 L
Mistroke
Mfstroke
.08 1 0 r
.98077 .2515 m
.97106 .24671 L
.96322 .24262 L
.96134 .24165 L
.95163 .23638 L
.94552 .23291 L
.94192 .23085 L
.93221 .22502 L
.92925 .22319 L
.92249 .21895 L
.91422 .21348 L
.91278 .21252 L
.90307 .20578 L
.90028 .20377 L
.89336 .19869 L
.8873 .19406 L
.88364 .1912 L
.87521 .18434 L
.87393 .18328 L
.86422 .17491 L
.8639 .17463 L
.85451 .16607 L
.85329 .16492 L
.84479 .15669 L
.84331 .15521 L
.83508 .14671 L
.83393 .14549 L
.82537 .1361 L
.82509 .13578 L
.81672 .12607 L
.81566 .12479 L
.8088 .11636 L
.80594 .1127 L
.80131 .10664 L
.79623 .09972 L
.79422 .09693 L
.78748 .08722 L
.78652 .08578 L
.78105 .07751 L
.77681 .07075 L
.77498 .06779 L
.76915 .05808 L
.76709 .05448 L
.76362 .04837 L
.75835 .03866 L
.75738 .03678 L
.75329 .02894 L
.7485 .01923 L
.98077 .01923 L
F
0 g
.98077 .2515 m
.97106 .24671 L
.96322 .24262 L
.96134 .24165 L
.95163 .23638 L
.94552 .23291 L
.94192 .23085 L
.93221 .22502 L
.92925 .22319 L
.92249 .21895 L
.91422 .21348 L
.91278 .21252 L
.90307 .20578 L
.90028 .20377 L
.89336 .19869 L
.8873 .19406 L
.88364 .1912 L
.87521 .18434 L
.87393 .18328 L
.86422 .17491 L
.8639 .17463 L
.85451 .16607 L
.85329 .16492 L
.84479 .15669 L
.84331 .15521 L
.83508 .14671 L
.83393 .14549 L
.82537 .1361 L
.82509 .13578 L
.81672 .12607 L
.81566 .12479 L
.8088 .11636 L
.80594 .1127 L
.80131 .10664 L
.79623 .09972 L
.79422 .09693 L
.78748 .08722 L
.78652 .08578 L
.78105 .07751 L
.77681 .07075 L
.77498 .06779 L
.76915 .05808 L
.76709 .05448 L
.76362 .04837 L
.75835 .03866 L
.75738 .03678 L
.75329 .02894 L
.7485 .01923 L
s
.08 1 0 r
.98077 .7485 m
.97106 .75329 L
.96322 .75738 L
.96134 .75835 L
.95163 .76362 L
.94552 .76709 L
.94192 .76915 L
.93221 .77498 L
.92925 .77681 L
.92249 .78105 L
.91422 .78652 L
.91278 .78748 L
.90307 .79422 L
.90028 .79623 L
.89336 .80131 L
.8873 .80594 L
.88364 .8088 L
.87521 .81566 L
.87393 .81672 L
.86422 .82509 L
.8639 .82537 L
.85451 .83393 L
.85329 .83508 L
.84479 .84331 L
.84331 .84479 L
.83508 .85329 L
.83393 .85451 L
.82537 .8639 L
.82509 .86422 L
.81672 .87393 L
.81566 .87521 L
.8088 .88364 L
.80594 .8873 L
.80131 .89336 L
.79623 .90028 L
.79422 .90307 L
.78748 .91278 L
.78652 .91422 L
.78105 .92249 L
.77681 .92925 L
.77498 .93221 L
.76915 .94192 L
.76709 .94552 L
.76362 .95163 L
.75835 .96134 L
.75738 .96322 L
.75329 .97106 L
.7485 .98077 L
.98077 .98077 L
F
0 g
.98077 .7485 m
.97106 .75329 L
.96322 .75738 L
.96134 .75835 L
.95163 .76362 L
.94552 .76709 L
.94192 .76915 L
.93221 .77498 L
.92925 .77681 L
.92249 .78105 L
.91422 .78652 L
.91278 .78748 L
.90307 .79422 L
.90028 .79623 L
.89336 .80131 L
.8873 .80594 L
.88364 .8088 L
.87521 .81566 L
.87393 .81672 L
.86422 .82509 L
.8639 .82537 L
.85451 .83393 L
.85329 .83508 L
.84479 .84331 L
.84331 .84479 L
.83508 .85329 L
.83393 .85451 L
.82537 .8639 L
.82509 .86422 L
.81672 .87393 L
.81566 .87521 L
.8088 .88364 L
.80594 .8873 L
.80131 .89336 L
.79623 .90028 L
.79422 .90307 L
.78748 .91278 L
.78652 .91422 L
.78105 .92249 L
.77681 .92925 L
.77498 .93221 L
.76915 .94192 L
.76709 .94552 L
.76362 .95163 L
.75835 .96134 L
.75738 .96322 L
.75329 .97106 L
.7485 .98077 L
s
.32 1 0 r
.98077 .24452 m
.97698 .24262 L
.97106 .23989 L
.96134 .2348 L
.95783 .23291 L
.95163 .22964 L
.94192 .22402 L
.94052 .22319 L
.93221 .21829 L
.92462 .21348 L
.92249 .21215 L
.91278 .20568 L
.91002 .20377 L
.90307 .19892 L
.89648 .19406 L
.89336 .19174 L
.88395 .18434 L
.88364 .1841 L
.87393 .17608 L
.87225 .17463 L
.86422 .16758 L
.8613 .16492 L
.85451 .15855 L
.85106 .15521 L
.84479 .14894 L
.84145 .14549 L
.83508 .1387 L
.83242 .13578 L
.82537 .12775 L
.82392 .12607 L
.8159 .11636 L
.81566 .11605 L
.80826 .10664 L
.80594 .10352 L
.80108 .09693 L
.79623 .08998 L
.79432 .08722 L
.78785 .07751 L
.78652 .07538 L
.78171 .06779 L
.77681 .05948 L
.77598 .05808 L
.77036 .04837 L
.76709 .04217 L
.7652 .03866 L
.76011 .02894 L
.75738 .02302 L
.75548 .01923 L
.98077 .01923 L
F
0 g
.98077 .24452 m
.97698 .24262 L
.97106 .23989 L
.96134 .2348 L
.95783 .23291 L
.95163 .22964 L
.94192 .22402 L
.94052 .22319 L
.93221 .21829 L
.92462 .21348 L
.92249 .21215 L
.91278 .20568 L
.91002 .20377 L
.90307 .19892 L
.89648 .19406 L
.89336 .19174 L
.88395 .18434 L
.88364 .1841 L
.87393 .17608 L
.87225 .17463 L
.86422 .16758 L
.8613 .16492 L
.85451 .15855 L
.85106 .15521 L
.84479 .14894 L
.84145 .14549 L
.83508 .1387 L
.83242 .13578 L
.82537 .12775 L
.82392 .12607 L
.8159 .11636 L
.81566 .11605 L
.80826 .10664 L
.80594 .10352 L
.80108 .09693 L
.79623 .08998 L
.79432 .08722 L
.78785 .07751 L
.78652 .07538 L
.78171 .06779 L
.77681 .05948 L
.77598 .05808 L
.77036 .04837 L
.76709 .04217 L
.7652 .03866 L
.76011 .02894 L
.75738 .02302 L
.75548 .01923 L
s
.32 1 0 r
.98077 .75548 m
.97698 .75738 L
.97106 .76011 L
.96134 .7652 L
.95783 .76709 L
.95163 .77036 L
.94192 .77598 L
.94052 .77681 L
.93221 .78171 L
.92462 .78652 L
.92249 .78785 L
.91278 .79432 L
.91002 .79623 L
.90307 .80108 L
.89648 .80594 L
.89336 .80826 L
.88395 .81566 L
.88364 .8159 L
.87393 .82392 L
.87225 .82537 L
.86422 .83242 L
.8613 .83508 L
.85451 .84145 L
.85106 .84479 L
.84479 .85106 L
.84145 .85451 L
.83508 .8613 L
.83242 .86422 L
.82537 .87225 L
.82392 .87393 L
.8159 .88364 L
.81566 .88395 L
.80826 .89336 L
.80594 .89648 L
.80108 .90307 L
.79623 .91002 L
.79432 .91278 L
.78785 .92249 L
.78652 .92462 L
.78171 .93221 L
.77681 .94052 L
.77598 .94192 L
.77036 .95163 L
.76709 .95783 L
.7652 .96134 L
.76011 .97106 L
.75738 .97698 L
.75548 .98077 L
.98077 .98077 L
F
0 g
.98077 .75548 m
.97698 .75738 L
.97106 .76011 L
.96134 .7652 L
.95783 .76709 L
.95163 .77036 L
.94192 .77598 L
.94052 .77681 L
.93221 .78171 L
.92462 .78652 L
.92249 .78785 L
.91278 .79432 L
.91002 .79623 L
.90307 .80108 L
.89648 .80594 L
.89336 .80826 L
.88395 .81566 L
.88364 .8159 L
.87393 .82392 L
.87225 .82537 L
.86422 .83242 L
.8613 .83508 L
.85451 .84145 L
.85106 .84479 L
.84479 .85106 L
.84145 .85451 L
.83508 .8613 L
.83242 .86422 L
.82537 .87225 L
.82392 .87393 L
.8159 .88364 L
.81566 .88395 L
.80826 .89336 L
.80594 .89648 L
.80108 .90307 L
.79623 .91002 L
.79432 .91278 L
.78785 .92249 L
.78652 .92462 L
.78171 .93221 L
.77681 .94052 L
.77598 .94192 L
.77036 .95163 L
.76709 .95783 L
.7652 .96134 L
.76011 .97106 L
.75738 .97698 L
.75548 .98077 L
s
.56 1 0 r
.98077 .23582 m
.97534 .23291 L
.97106 .23198 L
.96134 .22649 L
.95536 .22319 L
.95163 .22166 L
.94192 .21586 L
.93796 .21348 L
.93221 .21034 L
.92249 .20391 L
.92228 .20377 L
.91278 .19763 L
.90774 .19406 L
.90307 .19087 L
.89451 .18434 L
.89336 .18351 L
.88364 .17574 L
.8823 .17463 L
.87393 .16762 L
.87087 .16492 L
.86422 .15894 L
.86024 .15521 L
.85451 .14968 L
.85032 .14549 L
.84479 .13976 L
.84106 .13578 L
.83508 .12913 L
.83238 .12607 L
.82537 .1177 L
.82426 .11636 L
.81649 .10664 L
.81566 .10549 L
.80913 .09693 L
.80594 .09226 L
.80237 .08722 L
.79623 .07772 L
.79609 .07751 L
.78966 .06779 L
.78652 .06204 L
.78414 .05808 L
.77834 .04837 L
.77681 .04464 L
.77351 .03866 L
.76802 .02894 L
.76709 .02466 L
.76418 .01923 L
.98077 .01923 L
F
0 g
.98077 .23582 m
.97534 .23291 L
.97106 .23198 L
.96134 .22649 L
.95536 .22319 L
.95163 .22166 L
.94192 .21586 L
.93796 .21348 L
.93221 .21034 L
.92249 .20391 L
.92228 .20377 L
.91278 .19763 L
.90774 .19406 L
.90307 .19087 L
.89451 .18434 L
.89336 .18351 L
.88364 .17574 L
.8823 .17463 L
.87393 .16762 L
.87087 .16492 L
.86422 .15894 L
.86024 .15521 L
.85451 .14968 L
.85032 .14549 L
.84479 .13976 L
.84106 .13578 L
.83508 .12913 L
.83238 .12607 L
.82537 .1177 L
.82426 .11636 L
.81649 .10664 L
.81566 .10549 L
.80913 .09693 L
.80594 .09226 L
.80237 .08722 L
.79623 .07772 L
.79609 .07751 L
.78966 .06779 L
.78652 .06204 L
.78414 .05808 L
.77834 .04837 L
.77681 .04464 L
.77351 .03866 L
.76802 .02894 L
.76709 .02466 L
.76418 .01923 L
s
.56 1 0 r
.98077 .76418 m
.97534 .76709 L
.97106 .76802 L
.96134 .77351 L
.95536 .77681 L
.95163 .77834 L
.94192 .78414 L
.93796 .78652 L
.93221 .78966 L
.92249 .79609 L
.92228 .79623 L
.91278 .80237 L
.90774 .80594 L
.90307 .80913 L
.89451 .81566 L
.89336 .81649 L
.88364 .82426 L
.8823 .82537 L
.87393 .83238 L
.87087 .83508 L
.86422 .84106 L
.86024 .84479 L
.85451 .85032 L
.85032 .85451 L
.84479 .86024 L
.84106 .86422 L
.83508 .87087 L
.83238 .87393 L
.82537 .8823 L
.82426 .88364 L
.81649 .89336 L
.81566 .89451 L
.80913 .90307 L
.80594 .90774 L
.80237 .91278 L
.79623 .92228 L
.79609 .92249 L
.78966 .93221 L
.78652 .93796 L
.78414 .94192 L
.77834 .95163 L
.77681 .95536 L
.77351 .96134 L
.76802 .97106 L
.76709 .97534 L
.76418 .98077 L
.98077 .98077 L
F
0 g
.98077 .76418 m
.97534 .76709 L
.97106 .76802 L
.96134 .77351 L
.95536 .77681 L
.95163 .77834 L
.94192 .78414 L
.93796 .78652 L
.93221 .78966 L
.92249 .79609 L
.92228 .79623 L
.91278 .80237 L
.90774 .80594 L
.90307 .80913 L
.89451 .81566 L
.89336 .81649 L
.88364 .82426 L
.8823 .82537 L
.87393 .83238 L
.87087 .83508 L
.86422 .84106 L
.86024 .84479 L
.85451 .85032 L
.85032 .85451 L
.84479 .86024 L
.84106 .86422 L
.83508 .87087 L
.83238 .87393 L
.82537 .8823 L
.82426 .88364 L
.81649 .89336 L
.81566 .89451 L
.80913 .90307 L
.80594 .90774 L
.80237 .91278 L
.79623 .92228 L
.79609 .92249 L
.78966 .93221 L
.78652 .93796 L
.78414 .94192 L
.77834 .95163 L
.77681 .95536 L
.77351 .96134 L
.76802 .97106 L
.76709 .97534 L
.76418 .98077 L
s
.8 1 0 r
.79623 .94677 m
.80413 .95163 L
.79623 .95775 L
.7946 .95163 L
F
0 g
.79623 .94677 m
.80413 .95163 L
.79623 .95775 L
.7946 .95163 L
.79623 .94677 L
s
.8 1 0 r
.79623 .04225 m
.80413 .04837 L
.80594 .05013 L
.81347 .05808 L
.81566 .06046 L
.82225 .06779 L
.82537 .06953 L
.83095 .07751 L
.83508 .08107 L
.83972 .08722 L
.84479 .09206 L
.85014 .09693 L
.85451 .10265 L
.85862 .10664 L
.86422 .11289 L
.86772 .11636 L
.87393 .12277 L
.87723 .12607 L
.88364 .13228 L
.88711 .13578 L
.89336 .14138 L
.89735 .14549 L
.90307 .14986 L
.90794 .15521 L
.91278 .16028 L
.91893 .16492 L
.92249 .16905 L
.93047 .17463 L
.93221 .17775 L
.93954 .18434 L
.94192 .18653 L
.94987 .19406 L
.95163 .19587 L
.95775 .20377 L
.95163 .2054 L
.94677 .20377 L
.94192 .20172 L
.93221 .19691 L
.92701 .19406 L
.92249 .19142 L
.91278 .1857 L
.91084 .18434 L
.90307 .17871 L
.89743 .17463 L
.89336 .17155 L
.88496 .16492 L
.88364 .16383 L
.87393 .15557 L
.87353 .15521 L
.86422 .14672 L
.86295 .14549 L
.85451 .13705 L
.85328 .13578 L
.84479 .12647 L
.84443 .12607 L
.83617 .11636 L
.83508 .11504 L
.82845 .10664 L
.82537 .10257 L
.82129 .09693 L
.81566 .08916 L
.8143 .08722 L
.80858 .07751 L
.80594 .07299 L
.80309 .06779 L
.79828 .05808 L
.79623 .05323 L
.7946 .04837 L
F
0 g
.79623 .04225 m
.80413 .04837 L
.80594 .05013 L
.81347 .05808 L
.81566 .06046 L
.82225 .06779 L
.82537 .06953 L
.83095 .07751 L
.83508 .08107 L
.83972 .08722 L
.84479 .09206 L
.85014 .09693 L
.85451 .10265 L
.85862 .10664 L
.86422 .11289 L
.86772 .11636 L
.87393 .12277 L
.87723 .12607 L
.88364 .13228 L
.88711 .13578 L
.89336 .14138 L
.89735 .14549 L
.90307 .14986 L
.90794 .15521 L
.91278 .16028 L
.91893 .16492 L
.92249 .16905 L
.93047 .17463 L
.93221 .17775 L
.93954 .18434 L
.94192 .18653 L
.94987 .19406 L
.95163 .19587 L
.95775 .20377 L
.95163 .2054 L
.94677 .20377 L
.94192 .20172 L
.93221 .19691 L
.92701 .19406 L
.92249 .19142 L
.91278 .1857 L
.91084 .18434 L
.90307 .17871 L
.89743 .17463 L
.89336 .17155 L
.88496 .16492 L
.88364 .16383 L
.87393 .15557 L
.87353 .15521 L
.86422 .14672 L
Mistroke
.86295 .14549 L
.85451 .13705 L
.85328 .13578 L
.84479 .12647 L
.84443 .12607 L
.83617 .11636 L
.83508 .11504 L
.82845 .10664 L
.82537 .10257 L
.82129 .09693 L
.81566 .08916 L
.8143 .08722 L
.80858 .07751 L
.80594 .07299 L
.80309 .06779 L
.79828 .05808 L
.79623 .05323 L
.7946 .04837 L
.79623 .04225 L
Mfstroke
.8 1 0 r
.93221 .80309 m
.94192 .79828 L
.94987 .80594 L
.94192 .81347 L
.93954 .81566 L
.93221 .82225 L
.93047 .82537 L
.92249 .83095 L
.91893 .83508 L
.91278 .83972 L
.90794 .84479 L
.90307 .85014 L
.89735 .85451 L
.89336 .85862 L
.88711 .86422 L
.88364 .86772 L
.87723 .87393 L
.87393 .87723 L
.86772 .88364 L
.86422 .88711 L
.85862 .89336 L
.85451 .89735 L
.85014 .90307 L
.84479 .90794 L
.83972 .91278 L
.83508 .91893 L
.83095 .92249 L
.82537 .93047 L
.82225 .93221 L
.81566 .93954 L
.81347 .94192 L
.80594 .94987 L
.79828 .94192 L
.80309 .93221 L
.80594 .92701 L
.80858 .92249 L
.8143 .91278 L
.81566 .91084 L
.82129 .90307 L
.82537 .89743 L
.82845 .89336 L
.83508 .88496 L
.83617 .88364 L
.84443 .87393 L
.84479 .87353 L
.85328 .86422 L
.85451 .86295 L
.86295 .85451 L
.86422 .85328 L
.87353 .84479 L
.87393 .84443 L
.88364 .83617 L
.88496 .83508 L
.89336 .82845 L
.89743 .82537 L
.90307 .82129 L
.91084 .81566 L
.91278 .8143 L
.92249 .80858 L
.92701 .80594 L
F
0 g
.93221 .80309 m
.94192 .79828 L
.94987 .80594 L
.94192 .81347 L
.93954 .81566 L
.93221 .82225 L
.93047 .82537 L
.92249 .83095 L
.91893 .83508 L
.91278 .83972 L
.90794 .84479 L
.90307 .85014 L
.89735 .85451 L
.89336 .85862 L
.88711 .86422 L
.88364 .86772 L
.87723 .87393 L
.87393 .87723 L
.86772 .88364 L
.86422 .88711 L
.85862 .89336 L
.85451 .89735 L
.85014 .90307 L
.84479 .90794 L
.83972 .91278 L
.83508 .91893 L
.83095 .92249 L
.82537 .93047 L
.82225 .93221 L
.81566 .93954 L
.81347 .94192 L
.80594 .94987 L
.79828 .94192 L
.80309 .93221 L
.80594 .92701 L
.80858 .92249 L
.8143 .91278 L
.81566 .91084 L
.82129 .90307 L
.82537 .89743 L
.82845 .89336 L
.83508 .88496 L
.83617 .88364 L
.84443 .87393 L
.84479 .87353 L
.85328 .86422 L
.85451 .86295 L
.86295 .85451 L
.86422 .85328 L
.87353 .84479 L
Mistroke
.87393 .84443 L
.88364 .83617 L
.88496 .83508 L
.89336 .82845 L
.89743 .82537 L
.90307 .82129 L
.91084 .81566 L
.91278 .8143 L
.92249 .80858 L
.92701 .80594 L
.93221 .80309 L
Mfstroke
.32 1 0 r
.98077 .201 m
.97106 .19426 L
.9708 .19406 L
.96134 .18644 L
.95789 .18434 L
.95163 .17925 L
.94536 .17463 L
.94192 .17179 L
.93329 .16492 L
.93221 .164 L
.92249 .15577 L
.92175 .15521 L
.91278 .14725 L
.9107 .14549 L
.90307 .13849 L
.90006 .13578 L
.89336 .12941 L
.88984 .12607 L
.88364 .11998 L
.88002 .11636 L
.87393 .11016 L
.87059 .10664 L
.86422 .09994 L
.86151 .09693 L
.85451 .0893 L
.85275 .08722 L
.84479 .07825 L
.84423 .07751 L
.836 .06779 L
.83508 .06671 L
.82821 .05808 L
.82537 .05464 L
.82075 .04837 L
.81566 .04211 L
.81356 .03866 L
.80594 .0292 L
.80574 .02894 L
.799 .01923 L
.98077 .01923 L
F
0 g
.98077 .201 m
.97106 .19426 L
.9708 .19406 L
.96134 .18644 L
.95789 .18434 L
.95163 .17925 L
.94536 .17463 L
.94192 .17179 L
.93329 .16492 L
.93221 .164 L
.92249 .15577 L
.92175 .15521 L
.91278 .14725 L
.9107 .14549 L
.90307 .13849 L
.90006 .13578 L
.89336 .12941 L
.88984 .12607 L
.88364 .11998 L
.88002 .11636 L
.87393 .11016 L
.87059 .10664 L
.86422 .09994 L
.86151 .09693 L
.85451 .0893 L
.85275 .08722 L
.84479 .07825 L
.84423 .07751 L
.836 .06779 L
.83508 .06671 L
.82821 .05808 L
.82537 .05464 L
.82075 .04837 L
.81566 .04211 L
.81356 .03866 L
.80594 .0292 L
.80574 .02894 L
.799 .01923 L
s
.32 1 0 r
.98077 .799 m
.97106 .80574 L
.9708 .80594 L
.96134 .81356 L
.95789 .81566 L
.95163 .82075 L
.94536 .82537 L
.94192 .82821 L
.93329 .83508 L
.93221 .836 L
.92249 .84423 L
.92175 .84479 L
.91278 .85275 L
.9107 .85451 L
.90307 .86151 L
.90006 .86422 L
.89336 .87059 L
.88984 .87393 L
.88364 .88002 L
.88002 .88364 L
.87393 .88984 L
.87059 .89336 L
.86422 .90006 L
.86151 .90307 L
.85451 .9107 L
.85275 .91278 L
.84479 .92175 L
.84423 .92249 L
.836 .93221 L
.83508 .93329 L
.82821 .94192 L
.82537 .94536 L
.82075 .95163 L
.81566 .95789 L
.81356 .96134 L
.80594 .9708 L
.80574 .97106 L
.799 .98077 L
.98077 .98077 L
F
0 g
.98077 .799 m
.97106 .80574 L
.9708 .80594 L
.96134 .81356 L
.95789 .81566 L
.95163 .82075 L
.94536 .82537 L
.94192 .82821 L
.93329 .83508 L
.93221 .836 L
.92249 .84423 L
.92175 .84479 L
.91278 .85275 L
.9107 .85451 L
.90307 .86151 L
.90006 .86422 L
.89336 .87059 L
.88984 .87393 L
.88364 .88002 L
.88002 .88364 L
.87393 .88984 L
.87059 .89336 L
.86422 .90006 L
.86151 .90307 L
.85451 .9107 L
.85275 .91278 L
.84479 .92175 L
.84423 .92249 L
.836 .93221 L
.83508 .93329 L
.82821 .94192 L
.82537 .94536 L
.82075 .95163 L
.81566 .95789 L
.81356 .96134 L
.80594 .9708 L
.80574 .97106 L
.799 .98077 L
s
.08 1 0 r
.98077 .19166 m
.97106 .18485 L
.9703 .18434 L
.96134 .17772 L
.95708 .17463 L
.95163 .17045 L
.94444 .16492 L
.94192 .16291 L
.93237 .15521 L
.93221 .15507 L
.92249 .14686 L
.92088 .14549 L
.91278 .13835 L
.9099 .13578 L
.90307 .1295 L
.8994 .12607 L
.89336 .12027 L
.88935 .11636 L
.88364 .11065 L
.87973 .10664 L
.87393 .1006 L
.8705 .09693 L
.86422 .0901 L
.86165 .08722 L
.85451 .07912 L
.85314 .07751 L
.84493 .06779 L
.84479 .06763 L
.83709 .05808 L
.83508 .05556 L
.82955 .04837 L
.82537 .04292 L
.82228 .03866 L
.81566 .0297 L
.81515 .02894 L
.80834 .01923 L
.98077 .01923 L
F
0 g
.98077 .19166 m
.97106 .18485 L
.9703 .18434 L
.96134 .17772 L
.95708 .17463 L
.95163 .17045 L
.94444 .16492 L
.94192 .16291 L
.93237 .15521 L
.93221 .15507 L
.92249 .14686 L
.92088 .14549 L
.91278 .13835 L
.9099 .13578 L
.90307 .1295 L
.8994 .12607 L
.89336 .12027 L
.88935 .11636 L
.88364 .11065 L
.87973 .10664 L
.87393 .1006 L
.8705 .09693 L
.86422 .0901 L
.86165 .08722 L
.85451 .07912 L
.85314 .07751 L
.84493 .06779 L
.84479 .06763 L
.83709 .05808 L
.83508 .05556 L
.82955 .04837 L
.82537 .04292 L
.82228 .03866 L
.81566 .0297 L
.81515 .02894 L
.80834 .01923 L
s
.08 1 0 r
.98077 .80834 m
.97106 .81515 L
.9703 .81566 L
.96134 .82228 L
.95708 .82537 L
.95163 .82955 L
.94444 .83508 L
.94192 .83709 L
.93237 .84479 L
.93221 .84493 L
.92249 .85314 L
.92088 .85451 L
.91278 .86165 L
.9099 .86422 L
.90307 .8705 L
.8994 .87393 L
.89336 .87973 L
.88935 .88364 L
.88364 .88935 L
.87973 .89336 L
.87393 .8994 L
.8705 .90307 L
.86422 .9099 L
.86165 .91278 L
.85451 .92088 L
.85314 .92249 L
.84493 .93221 L
.84479 .93237 L
.83709 .94192 L
.83508 .94444 L
.82955 .95163 L
.82537 .95708 L
.82228 .96134 L
.81566 .9703 L
.81515 .97106 L
.80834 .98077 L
.98077 .98077 L
F
0 g
.98077 .80834 m
.97106 .81515 L
.9703 .81566 L
.96134 .82228 L
.95708 .82537 L
.95163 .82955 L
.94444 .83508 L
.94192 .83709 L
.93237 .84479 L
.93221 .84493 L
.92249 .85314 L
.92088 .85451 L
.91278 .86165 L
.9099 .86422 L
.90307 .8705 L
.8994 .87393 L
.89336 .87973 L
.88935 .88364 L
.88364 .88935 L
.87973 .89336 L
.87393 .8994 L
.8705 .90307 L
.86422 .9099 L
.86165 .91278 L
.85451 .92088 L
.85314 .92249 L
.84493 .93221 L
.84479 .93237 L
.83709 .94192 L
.83508 .94444 L
.82955 .95163 L
.82537 .95708 L
.82228 .96134 L
.81566 .9703 L
.81515 .97106 L
.80834 .98077 L
s
0 1 .16 r
.98077 .18405 m
.97106 .17723 L
.96739 .17463 L
.96134 .1702 L
.95426 .16492 L
.95163 .1629 L
.94192 .15531 L
.94179 .15521 L
.93221 .14738 L
.92993 .14549 L
.92249 .13914 L
.91864 .13578 L
.91278 .13055 L
.90787 .12607 L
.90307 .12158 L
.8976 .11636 L
.89336 .1122 L
.8878 .10664 L
.88364 .1024 L
.87842 .09693 L
.87393 .09213 L
.86945 .08722 L
.86422 .08136 L
.86086 .07751 L
.85451 .07007 L
.85262 .06779 L
.84479 .05821 L
.84469 .05808 L
.8371 .04837 L
.83508 .04574 L
.8298 .03866 L
.82537 .03261 L
.82277 .02894 L
.81595 .01923 L
.98077 .01923 L
F
0 g
.98077 .18405 m
.97106 .17723 L
.96739 .17463 L
.96134 .1702 L
.95426 .16492 L
.95163 .1629 L
.94192 .15531 L
.94179 .15521 L
.93221 .14738 L
.92993 .14549 L
.92249 .13914 L
.91864 .13578 L
.91278 .13055 L
.90787 .12607 L
.90307 .12158 L
.8976 .11636 L
.89336 .1122 L
.8878 .10664 L
.88364 .1024 L
.87842 .09693 L
.87393 .09213 L
.86945 .08722 L
.86422 .08136 L
.86086 .07751 L
.85451 .07007 L
.85262 .06779 L
.84479 .05821 L
.84469 .05808 L
.8371 .04837 L
.83508 .04574 L
.8298 .03866 L
.82537 .03261 L
.82277 .02894 L
.81595 .01923 L
s
0 1 .16 r
.98077 .81595 m
.97106 .82277 L
.96739 .82537 L
.96134 .8298 L
.95426 .83508 L
.95163 .8371 L
.94192 .84469 L
.94179 .84479 L
.93221 .85262 L
.92993 .85451 L
.92249 .86086 L
.91864 .86422 L
.91278 .86945 L
.90787 .87393 L
.90307 .87842 L
.8976 .88364 L
.89336 .8878 L
.8878 .89336 L
.88364 .8976 L
.87842 .90307 L
.87393 .90787 L
.86945 .91278 L
.86422 .91864 L
.86086 .92249 L
.85451 .92993 L
.85262 .93221 L
.84479 .94179 L
.84469 .94192 L
.8371 .95163 L
.83508 .95426 L
.8298 .96134 L
.82537 .96739 L
.82277 .97106 L
.81595 .98077 L
.98077 .98077 L
F
0 g
.98077 .81595 m
.97106 .82277 L
.96739 .82537 L
.96134 .8298 L
.95426 .83508 L
.95163 .8371 L
.94192 .84469 L
.94179 .84479 L
.93221 .85262 L
.92993 .85451 L
.92249 .86086 L
.91864 .86422 L
.91278 .86945 L
.90787 .87393 L
.90307 .87842 L
.8976 .88364 L
.89336 .8878 L
.8878 .89336 L
.88364 .8976 L
.87842 .90307 L
.87393 .90787 L
.86945 .91278 L
.86422 .91864 L
.86086 .92249 L
.85451 .92993 L
.85262 .93221 L
.84479 .94179 L
.84469 .94192 L
.8371 .95163 L
.83508 .95426 L
.8298 .96134 L
.82537 .96739 L
.82277 .97106 L
.81595 .98077 L
s
0 1 .4 r
.98077 .17706 m
.97722 .17463 L
.97106 .17028 L
.96363 .16492 L
.96134 .16323 L
.95163 .15588 L
.95076 .15521 L
.94192 .14822 L
.93855 .14549 L
.93221 .14024 L
.92695 .13578 L
.92249 .13191 L
.91593 .12607 L
.91278 .1232 L
.90544 .11636 L
.90307 .1141 L
.89544 .10664 L
.89336 .10456 L
.8859 .09693 L
.88364 .09456 L
.8768 .08722 L
.87393 .08407 L
.86809 .07751 L
.86422 .07305 L
.85976 .06779 L
.85451 .06145 L
.85178 .05808 L
.84479 .04924 L
.84412 .04837 L
.83677 .03866 L
.83508 .03637 L
.82972 .02894 L
.82537 .02278 L
.82294 .01923 L
.98077 .01923 L
F
0 g
.98077 .17706 m
.97722 .17463 L
.97106 .17028 L
.96363 .16492 L
.96134 .16323 L
.95163 .15588 L
.95076 .15521 L
.94192 .14822 L
.93855 .14549 L
.93221 .14024 L
.92695 .13578 L
.92249 .13191 L
.91593 .12607 L
.91278 .1232 L
.90544 .11636 L
.90307 .1141 L
.89544 .10664 L
.89336 .10456 L
.8859 .09693 L
.88364 .09456 L
.8768 .08722 L
.87393 .08407 L
.86809 .07751 L
.86422 .07305 L
.85976 .06779 L
.85451 .06145 L
.85178 .05808 L
.84479 .04924 L
.84412 .04837 L
.83677 .03866 L
.83508 .03637 L
.82972 .02894 L
.82537 .02278 L
.82294 .01923 L
s
0 1 .4 r
.98077 .82294 m
.97722 .82537 L
.97106 .82972 L
.96363 .83508 L
.96134 .83677 L
.95163 .84412 L
.95076 .84479 L
.94192 .85178 L
.93855 .85451 L
.93221 .85976 L
.92695 .86422 L
.92249 .86809 L
.91593 .87393 L
.91278 .8768 L
.90544 .88364 L
.90307 .8859 L
.89544 .89336 L
.89336 .89544 L
.8859 .90307 L
.88364 .90544 L
.8768 .91278 L
.87393 .91593 L
.86809 .92249 L
.86422 .92695 L
.85976 .93221 L
.85451 .93855 L
.85178 .94192 L
.84479 .95076 L
.84412 .95163 L
.83677 .96134 L
.83508 .96363 L
.82972 .97106 L
.82537 .97722 L
.82294 .98077 L
.98077 .98077 L
F
0 g
.98077 .82294 m
.97722 .82537 L
.97106 .82972 L
.96363 .83508 L
.96134 .83677 L
.95163 .84412 L
.95076 .84479 L
.94192 .85178 L
.93855 .85451 L
.93221 .85976 L
.92695 .86422 L
.92249 .86809 L
.91593 .87393 L
.91278 .8768 L
.90544 .88364 L
.90307 .8859 L
.89544 .89336 L
.89336 .89544 L
.8859 .90307 L
.88364 .90544 L
.8768 .91278 L
.87393 .91593 L
.86809 .92249 L
.86422 .92695 L
.85976 .93221 L
.85451 .93855 L
.85178 .94192 L
.84479 .95076 L
.84412 .95163 L
.83677 .96134 L
.83508 .96363 L
.82972 .97106 L
.82537 .97722 L
.82294 .98077 L
s
0 1 .64 r
.98077 .1703 m
.97308 .16492 L
.97106 .16349 L
.96134 .15639 L
.95977 .15521 L
.95163 .14899 L
.94718 .14549 L
.94192 .14127 L
.93526 .13578 L
.93221 .1332 L
.92397 .12607 L
.92249 .12476 L
.91324 .11636 L
.91278 .11593 L
.90307 .10667 L
.90304 .10664 L
.89336 .09696 L
.89333 .09693 L
.88407 .08722 L
.88364 .08676 L
.87524 .07751 L
.87393 .07603 L
.8668 .06779 L
.86422 .06474 L
.85873 .05808 L
.85451 .05282 L
.85101 .04837 L
.84479 .04023 L
.84361 .03866 L
.83651 .02894 L
.83508 .02692 L
.8297 .01923 L
.98077 .01923 L
F
0 g
.98077 .1703 m
.97308 .16492 L
.97106 .16349 L
.96134 .15639 L
.95977 .15521 L
.95163 .14899 L
.94718 .14549 L
.94192 .14127 L
.93526 .13578 L
.93221 .1332 L
.92397 .12607 L
.92249 .12476 L
.91324 .11636 L
.91278 .11593 L
.90307 .10667 L
.90304 .10664 L
.89336 .09696 L
.89333 .09693 L
.88407 .08722 L
.88364 .08676 L
.87524 .07751 L
.87393 .07603 L
.8668 .06779 L
.86422 .06474 L
.85873 .05808 L
.85451 .05282 L
.85101 .04837 L
.84479 .04023 L
.84361 .03866 L
.83651 .02894 L
.83508 .02692 L
.8297 .01923 L
s
0 1 .64 r
.98077 .8297 m
.97308 .83508 L
.97106 .83651 L
.96134 .84361 L
.95977 .84479 L
.95163 .85101 L
.94718 .85451 L
.94192 .85873 L
.93526 .86422 L
.93221 .8668 L
.92397 .87393 L
.92249 .87524 L
.91324 .88364 L
.91278 .88407 L
.90307 .89333 L
.90304 .89336 L
.89336 .90304 L
.89333 .90307 L
.88407 .91278 L
.88364 .91324 L
.87524 .92249 L
.87393 .92397 L
.8668 .93221 L
.86422 .93526 L
.85873 .94192 L
.85451 .94718 L
.85101 .95163 L
.84479 .95977 L
.84361 .96134 L
.83651 .97106 L
.83508 .97308 L
.8297 .98077 L
.98077 .98077 L
F
0 g
.98077 .8297 m
.97308 .83508 L
.97106 .83651 L
.96134 .84361 L
.95977 .84479 L
.95163 .85101 L
.94718 .85451 L
.94192 .85873 L
.93526 .86422 L
.93221 .8668 L
.92397 .87393 L
.92249 .87524 L
.91324 .88364 L
.91278 .88407 L
.90307 .89333 L
.90304 .89336 L
.89336 .90304 L
.89333 .90307 L
.88407 .91278 L
.88364 .91324 L
.87524 .92249 L
.87393 .92397 L
.8668 .93221 L
.86422 .93526 L
.85873 .94192 L
.85451 .94718 L
.85101 .95163 L
.84479 .95977 L
.84361 .96134 L
.83651 .97106 L
.83508 .97308 L
.8297 .98077 L
s
0 1 .88 r
.98077 .16335 m
.97106 .15649 L
.96928 .15521 L
.96134 .14937 L
.95626 .14549 L
.95163 .14191 L
.94398 .13578 L
.94192 .1341 L
.93238 .12607 L
.93221 .12592 L
.92249 .11737 L
.92138 .11636 L
.91278 .10839 L
.91094 .10664 L
.90307 .09897 L
.90103 .09693 L
.89336 .08906 L
.89161 .08722 L
.88364 .07862 L
.88263 .07751 L
.87408 .06779 L
.87393 .06762 L
.8659 .05808 L
.86422 .05602 L
.85809 .04837 L
.85451 .04374 L
.85063 .03866 L
.84479 .03072 L
.84351 .02894 L
.83665 .01923 L
.98077 .01923 L
F
0 g
.98077 .16335 m
.97106 .15649 L
.96928 .15521 L
.96134 .14937 L
.95626 .14549 L
.95163 .14191 L
.94398 .13578 L
.94192 .1341 L
.93238 .12607 L
.93221 .12592 L
.92249 .11737 L
.92138 .11636 L
.91278 .10839 L
.91094 .10664 L
.90307 .09897 L
.90103 .09693 L
.89336 .08906 L
.89161 .08722 L
.88364 .07862 L
.88263 .07751 L
.87408 .06779 L
.87393 .06762 L
.8659 .05808 L
.86422 .05602 L
.85809 .04837 L
.85451 .04374 L
.85063 .03866 L
.84479 .03072 L
.84351 .02894 L
.83665 .01923 L
s
0 1 .88 r
.98077 .83665 m
.97106 .84351 L
.96928 .84479 L
.96134 .85063 L
.95626 .85451 L
.95163 .85809 L
.94398 .86422 L
.94192 .8659 L
.93238 .87393 L
.93221 .87408 L
.92249 .88263 L
.92138 .88364 L
.91278 .89161 L
.91094 .89336 L
.90307 .90103 L
.90103 .90307 L
.89336 .91094 L
.89161 .91278 L
.88364 .92138 L
.88263 .92249 L
.87408 .93221 L
.87393 .93238 L
.8659 .94192 L
.86422 .94398 L
.85809 .95163 L
.85451 .95626 L
.85063 .96134 L
.84479 .96928 L
.84351 .97106 L
.83665 .98077 L
.98077 .98077 L
F
0 g
.98077 .83665 m
.97106 .84351 L
.96928 .84479 L
.96134 .85063 L
.95626 .85451 L
.95163 .85809 L
.94398 .86422 L
.94192 .8659 L
.93238 .87393 L
.93221 .87408 L
.92249 .88263 L
.92138 .88364 L
.91278 .89161 L
.91094 .89336 L
.90307 .90103 L
.90103 .90307 L
.89336 .91094 L
.89161 .91278 L
.88364 .92138 L
.88263 .92249 L
.87408 .93221 L
.87393 .93238 L
.8659 .94192 L
.86422 .94398 L
.85809 .95163 L
.85451 .95626 L
.85063 .96134 L
.84479 .96928 L
.84351 .97106 L
.83665 .98077 L
s
0 .88 1 r
.98077 .15572 m
.98005 .15521 L
.97106 .1489 L
.96643 .14549 L
.96134 .14174 L
.95368 .13578 L
.95163 .13419 L
.94192 .12625 L
.9417 .12607 L
.93221 .11799 L
.93035 .11636 L
.92249 .1093 L
.91962 .10664 L
.91278 .10017 L
.90946 .09693 L
.90307 .09054 L
.89983 .08722 L
.89336 .08038 L
.8907 .07751 L
.88364 .06965 L
.88201 .06779 L
.87393 .0583 L
.87375 .05808 L
.86581 .04837 L
.86422 .04632 L
.85826 .03866 L
.85451 .03357 L
.8511 .02894 L
.84479 .01995 L
.84428 .01923 L
.98077 .01923 L
F
0 g
.98077 .15572 m
.98005 .15521 L
.97106 .1489 L
.96643 .14549 L
.96134 .14174 L
.95368 .13578 L
.95163 .13419 L
.94192 .12625 L
.9417 .12607 L
.93221 .11799 L
.93035 .11636 L
.92249 .1093 L
.91962 .10664 L
.91278 .10017 L
.90946 .09693 L
.90307 .09054 L
.89983 .08722 L
.89336 .08038 L
.8907 .07751 L
.88364 .06965 L
.88201 .06779 L
.87393 .0583 L
.87375 .05808 L
.86581 .04837 L
.86422 .04632 L
.85826 .03866 L
.85451 .03357 L
.8511 .02894 L
.84479 .01995 L
.84428 .01923 L
s
0 .88 1 r
.98077 .84428 m
.98005 .84479 L
.97106 .8511 L
.96643 .85451 L
.96134 .85826 L
.95368 .86422 L
.95163 .86581 L
.94192 .87375 L
.9417 .87393 L
.93221 .88201 L
.93035 .88364 L
.92249 .8907 L
.91962 .89336 L
.91278 .89983 L
.90946 .90307 L
.90307 .90946 L
.89983 .91278 L
.89336 .91962 L
.8907 .92249 L
.88364 .93035 L
.88201 .93221 L
.87393 .9417 L
.87375 .94192 L
.86581 .95163 L
.86422 .95368 L
.85826 .96134 L
.85451 .96643 L
.8511 .97106 L
.84479 .98005 L
.84428 .98077 L
.98077 .98077 L
F
0 g
.98077 .84428 m
.98005 .84479 L
.97106 .8511 L
.96643 .85451 L
.96134 .85826 L
.95368 .86422 L
.95163 .86581 L
.94192 .87375 L
.9417 .87393 L
.93221 .88201 L
.93035 .88364 L
.92249 .8907 L
.91962 .89336 L
.91278 .89983 L
.90946 .90307 L
.90307 .90946 L
.89983 .91278 L
.89336 .91962 L
.8907 .92249 L
.88364 .93035 L
.88201 .93221 L
.87393 .9417 L
.87375 .94192 L
.86581 .95163 L
.86422 .95368 L
.85826 .96134 L
.85451 .96643 L
.8511 .97106 L
.84479 .98005 L
.84428 .98077 L
s
0 .64 1 r
.98077 .14652 m
.97942 .14549 L
.97106 .13985 L
.96565 .13578 L
.96134 .13273 L
.95304 .12607 L
.95163 .12504 L
.94192 .11687 L
.94131 .11636 L
.93221 .10852 L
.93012 .10664 L
.92249 .09968 L
.9196 .09693 L
.91278 .09032 L
.90968 .08722 L
.90307 .0804 L
.90032 .07751 L
.89336 .06988 L
.89148 .06779 L
.88364 .05869 L
.88313 .05808 L
.87496 .04837 L
.87393 .04696 L
.86727 .03866 L
.86422 .03435 L
.86015 .02894 L
.85451 .02058 L
.85348 .01923 L
.98077 .01923 L
F
0 g
.98077 .14652 m
.97942 .14549 L
.97106 .13985 L
.96565 .13578 L
.96134 .13273 L
.95304 .12607 L
.95163 .12504 L
.94192 .11687 L
.94131 .11636 L
.93221 .10852 L
.93012 .10664 L
.92249 .09968 L
.9196 .09693 L
.91278 .09032 L
.90968 .08722 L
.90307 .0804 L
.90032 .07751 L
.89336 .06988 L
.89148 .06779 L
.88364 .05869 L
.88313 .05808 L
.87496 .04837 L
.87393 .04696 L
.86727 .03866 L
.86422 .03435 L
.86015 .02894 L
.85451 .02058 L
.85348 .01923 L
s
0 .64 1 r
.98077 .85348 m
.97942 .85451 L
.97106 .86015 L
.96565 .86422 L
.96134 .86727 L
.95304 .87393 L
.95163 .87496 L
.94192 .88313 L
.94131 .88364 L
.93221 .89148 L
.93012 .89336 L
.92249 .90032 L
.9196 .90307 L
.91278 .90968 L
.90968 .91278 L
.90307 .9196 L
.90032 .92249 L
.89336 .93012 L
.89148 .93221 L
.88364 .94131 L
.88313 .94192 L
.87496 .95163 L
.87393 .95304 L
.86727 .96134 L
.86422 .96565 L
.86015 .97106 L
.85451 .97942 L
.85348 .98077 L
.98077 .98077 L
F
0 g
.98077 .85348 m
.97942 .85451 L
.97106 .86015 L
.96565 .86422 L
.96134 .86727 L
.95304 .87393 L
.95163 .87496 L
.94192 .88313 L
.94131 .88364 L
.93221 .89148 L
.93012 .89336 L
.92249 .90032 L
.9196 .90307 L
.91278 .90968 L
.90968 .91278 L
.90307 .9196 L
.90032 .92249 L
.89336 .93012 L
.89148 .93221 L
.88364 .94131 L
.88313 .94192 L
.87496 .95163 L
.87393 .95304 L
.86727 .96134 L
.86422 .96565 L
.86015 .97106 L
.85451 .97942 L
.85348 .98077 L
s
0 .4 1 r
.98077 .12841 m
.97606 .12607 L
.97106 .12304 L
.96134 .11659 L
.96102 .11636 L
.95163 .10936 L
.94818 .10664 L
.94192 .1015 L
.93658 .09693 L
.93221 .09303 L
.92594 .08722 L
.92249 .0839 L
.9161 .07751 L
.91278 .07406 L
.90697 .06779 L
.90307 .06342 L
.8985 .05808 L
.89336 .05182 L
.89064 .04837 L
.88364 .03898 L
.88341 .03866 L
.87696 .02894 L
.87393 .02394 L
.87159 .01923 L
.98077 .01923 L
F
0 g
.98077 .12841 m
.97606 .12607 L
.97106 .12304 L
.96134 .11659 L
.96102 .11636 L
.95163 .10936 L
.94818 .10664 L
.94192 .1015 L
.93658 .09693 L
.93221 .09303 L
.92594 .08722 L
.92249 .0839 L
.9161 .07751 L
.91278 .07406 L
.90697 .06779 L
.90307 .06342 L
.8985 .05808 L
.89336 .05182 L
.89064 .04837 L
.88364 .03898 L
.88341 .03866 L
.87696 .02894 L
.87393 .02394 L
.87159 .01923 L
s
0 .4 1 r
.87893 .98077 m
.87393 .97606 L
.87159 .98077 L
F
0 g
.87893 .98077 m
.87393 .97606 L
.87159 .98077 L
s
0 .4 1 r
.96134 .88341 m
.97106 .87696 L
.97637 .88364 L
.97106 .88889 L
.96651 .89336 L
.96134 .89836 L
.95649 .90307 L
.95163 .90776 L
.94648 .91278 L
.94192 .91724 L
.93659 .92249 L
.93221 .92684 L
.92684 .93221 L
.92249 .93659 L
.91724 .94192 L
.91278 .94648 L
.90776 .95163 L
.90307 .95649 L
.89836 .96134 L
.89336 .96651 L
.88889 .97106 L
.88364 .97637 L
.87696 .97106 L
.88341 .96134 L
.88364 .96102 L
.89064 .95163 L
.89336 .94818 L
.8985 .94192 L
.90307 .93658 L
.90697 .93221 L
.91278 .92594 L
.9161 .92249 L
.92249 .9161 L
.92594 .91278 L
.93221 .90697 L
.93658 .90307 L
.94192 .8985 L
.94818 .89336 L
.95163 .89064 L
.96102 .88364 L
F
0 g
.96134 .88341 m
.97106 .87696 L
.97637 .88364 L
.97106 .88889 L
.96651 .89336 L
.96134 .89836 L
.95649 .90307 L
.95163 .90776 L
.94648 .91278 L
.94192 .91724 L
.93659 .92249 L
.93221 .92684 L
.92684 .93221 L
.92249 .93659 L
.91724 .94192 L
.91278 .94648 L
.90776 .95163 L
.90307 .95649 L
.89836 .96134 L
.89336 .96651 L
.88889 .97106 L
.88364 .97637 L
.87696 .97106 L
.88341 .96134 L
.88364 .96102 L
.89064 .95163 L
.89336 .94818 L
.8985 .94192 L
.90307 .93658 L
.90697 .93221 L
.91278 .92594 L
.9161 .92249 L
.92249 .9161 L
.92594 .91278 L
.93221 .90697 L
.93658 .90307 L
.94192 .8985 L
.94818 .89336 L
.95163 .89064 L
.96102 .88364 L
.96134 .88341 L
s
0 .64 1 r
.98077 .12107 m
.97637 .11636 L
.97106 .11111 L
.96651 .10664 L
.96134 .10164 L
.95649 .09693 L
.95163 .09224 L
.94648 .08722 L
.94192 .08276 L
.93659 .07751 L
.93221 .07316 L
.92684 .06779 L
.92249 .06341 L
.91724 .05808 L
.91278 .05352 L
.90776 .04837 L
.90307 .04351 L
.89836 .03866 L
.89336 .03349 L
.88889 .02894 L
.88364 .02363 L
.87893 .01923 L
.98077 .01923 L
F
0 g
.98077 .12107 m
.97637 .11636 L
.97106 .11111 L
.96651 .10664 L
.96134 .10164 L
.95649 .09693 L
.95163 .09224 L
.94648 .08722 L
.94192 .08276 L
.93659 .07751 L
.93221 .07316 L
.92684 .06779 L
.92249 .06341 L
.91724 .05808 L
.91278 .05352 L
.90776 .04837 L
.90307 .04351 L
.89836 .03866 L
.89336 .03349 L
.88889 .02894 L
.88364 .02363 L
.87893 .01923 L
s
0 .88 1 r
.98077 .10192 m
.97408 .09693 L
.97106 .09349 L
.964 .08722 L
.96134 .08475 L
.95353 .07751 L
.95163 .07569 L
.94344 .06779 L
.94192 .0663 L
.9337 .05808 L
.93221 .05656 L
.92431 .04837 L
.92249 .04647 L
.91525 .03866 L
.91278 .036 L
.90651 .02894 L
.90307 .02592 L
.89808 .01923 L
.98077 .01923 L
F
0 g
.98077 .10192 m
.97408 .09693 L
.97106 .09349 L
.964 .08722 L
.96134 .08475 L
.95353 .07751 L
.95163 .07569 L
.94344 .06779 L
.94192 .0663 L
.9337 .05808 L
.93221 .05656 L
.92431 .04837 L
.92249 .04647 L
.91525 .03866 L
.91278 .036 L
.90651 .02894 L
.90307 .02592 L
.89808 .01923 L
s
0 .88 1 r
.98077 .89808 m
.97408 .90307 L
.97106 .90651 L
.964 .91278 L
.96134 .91525 L
.95353 .92249 L
.95163 .92431 L
.94344 .93221 L
.94192 .9337 L
.9337 .94192 L
.93221 .94344 L
.92431 .95163 L
.92249 .95353 L
.91525 .96134 L
.91278 .964 L
.90651 .97106 L
.90307 .97408 L
.89808 .98077 L
.98077 .98077 L
F
0 g
.98077 .89808 m
.97408 .90307 L
.97106 .90651 L
.964 .91278 L
.96134 .91525 L
.95353 .92249 L
.95163 .92431 L
.94344 .93221 L
.94192 .9337 L
.9337 .94192 L
.93221 .94344 L
.92431 .95163 L
.92249 .95353 L
.91525 .96134 L
.91278 .964 L
.90651 .97106 L
.90307 .97408 L
.89808 .98077 L
s
0 1 .88 r
.98077 .09183 m
.97531 .08722 L
.97106 .08328 L
.96468 .07751 L
.96134 .07441 L
.95432 .06779 L
.95163 .06521 L
.94436 .05808 L
.94192 .05564 L
.93479 .04837 L
.93221 .04568 L
.92559 .03866 L
.92249 .03532 L
.91672 .02894 L
.91278 .02469 L
.90817 .01923 L
.98077 .01923 L
F
0 g
.98077 .09183 m
.97531 .08722 L
.97106 .08328 L
.96468 .07751 L
.96134 .07441 L
.95432 .06779 L
.95163 .06521 L
.94436 .05808 L
.94192 .05564 L
.93479 .04837 L
.93221 .04568 L
.92559 .03866 L
.92249 .03532 L
.91672 .02894 L
.91278 .02469 L
.90817 .01923 L
s
0 1 .88 r
.98077 .90817 m
.97531 .91278 L
.97106 .91672 L
.96468 .92249 L
.96134 .92559 L
.95432 .93221 L
.95163 .93479 L
.94436 .94192 L
.94192 .94436 L
.93479 .95163 L
.93221 .95432 L
.92559 .96134 L
.92249 .96468 L
.91672 .97106 L
.91278 .97531 L
.90817 .98077 L
.98077 .98077 L
F
0 g
.98077 .90817 m
.97531 .91278 L
.97106 .91672 L
.96468 .92249 L
.96134 .92559 L
.95432 .93221 L
.95163 .93479 L
.94436 .94192 L
.94192 .94436 L
.93479 .95163 L
.93221 .95432 L
.92559 .96134 L
.92249 .96468 L
.91672 .97106 L
.91278 .97531 L
.90817 .98077 L
s
0 1 .64 r
.98077 .08297 m
.97456 .07751 L
.97106 .07432 L
.96397 .06779 L
.96134 .06533 L
.95379 .05808 L
.95163 .05596 L
.94404 .04837 L
.94192 .04621 L
.93467 .03866 L
.93221 .03603 L
.92568 .02894 L
.92249 .02544 L
.91703 .01923 L
.98077 .01923 L
F
0 g
.98077 .08297 m
.97456 .07751 L
.97106 .07432 L
.96397 .06779 L
.96134 .06533 L
.95379 .05808 L
.95163 .05596 L
.94404 .04837 L
.94192 .04621 L
.93467 .03866 L
.93221 .03603 L
.92568 .02894 L
.92249 .02544 L
.91703 .01923 L
s
0 1 .64 r
.98077 .91703 m
.97456 .92249 L
.97106 .92568 L
.96397 .93221 L
.96134 .93467 L
.95379 .94192 L
.95163 .94404 L
.94404 .95163 L
.94192 .95379 L
.93467 .96134 L
.93221 .96397 L
.92568 .97106 L
.92249 .97456 L
.91703 .98077 L
.98077 .98077 L
F
0 g
.98077 .91703 m
.97456 .92249 L
.97106 .92568 L
.96397 .93221 L
.96134 .93467 L
.95379 .94192 L
.95163 .94404 L
.94404 .95163 L
.94192 .95379 L
.93467 .96134 L
.93221 .96397 L
.92568 .97106 L
.92249 .97456 L
.91703 .98077 L
s
0 1 .4 r
.98077 .07434 m
.9735 .06779 L
.97106 .06557 L
.96308 .05808 L
.96134 .05642 L
.95312 .04837 L
.95163 .04688 L
.94358 .03866 L
.94192 .03692 L
.93443 .02894 L
.93221 .0265 L
.92566 .01923 L
.98077 .01923 L
F
0 g
.98077 .07434 m
.9735 .06779 L
.97106 .06557 L
.96308 .05808 L
.96134 .05642 L
.95312 .04837 L
.95163 .04688 L
.94358 .03866 L
.94192 .03692 L
.93443 .02894 L
.93221 .0265 L
.92566 .01923 L
s
0 1 .4 r
.98077 .92566 m
.9735 .93221 L
.97106 .93443 L
.96308 .94192 L
.96134 .94358 L
.95312 .95163 L
.95163 .95312 L
.94358 .96134 L
.94192 .96308 L
.93443 .97106 L
.93221 .9735 L
.92566 .98077 L
.98077 .98077 L
F
0 g
.98077 .92566 m
.9735 .93221 L
.97106 .93443 L
.96308 .94192 L
.96134 .94358 L
.95312 .95163 L
.95163 .95312 L
.94358 .96134 L
.94192 .96308 L
.93443 .97106 L
.93221 .9735 L
.92566 .98077 L
s
0 1 .16 r
.98077 .06516 m
.9731 .05808 L
.97106 .05623 L
.96284 .04837 L
.96134 .04691 L
.95309 .03866 L
.95163 .03716 L
.94377 .02894 L
.94192 .0269 L
.93484 .01923 L
.98077 .01923 L
F
0 g
.98077 .06516 m
.9731 .05808 L
.97106 .05623 L
.96284 .04837 L
.96134 .04691 L
.95309 .03866 L
.95163 .03716 L
.94377 .02894 L
.94192 .0269 L
.93484 .01923 L
s
0 1 .16 r
.98077 .93484 m
.9731 .94192 L
.97106 .94377 L
.96284 .95163 L
.96134 .95309 L
.95309 .96134 L
.95163 .96284 L
.94377 .97106 L
.94192 .9731 L
.93484 .98077 L
.98077 .98077 L
F
0 g
.98077 .93484 m
.9731 .94192 L
.97106 .94377 L
.96284 .95163 L
.96134 .95309 L
.95309 .96134 L
.95163 .96284 L
.94377 .97106 L
.94192 .9731 L
.93484 .98077 L
s
.08 1 0 r
.98077 .05419 m
.97494 .04837 L
.97106 .04509 L
.96449 .03866 L
.96134 .03551 L
.95491 .02894 L
.95163 .02506 L
.94581 .01923 L
.98077 .01923 L
F
0 g
.98077 .05419 m
.97494 .04837 L
.97106 .04509 L
.96449 .03866 L
.96134 .03551 L
.95491 .02894 L
.95163 .02506 L
.94581 .01923 L
s
.08 1 0 r
.98077 .94581 m
.97494 .95163 L
.97106 .95491 L
.96449 .96134 L
.96134 .96449 L
.95491 .97106 L
.95163 .97494 L
.94581 .98077 L
.98077 .98077 L
F
0 g
.98077 .94581 m
.97494 .95163 L
.97106 .95491 L
.96449 .96134 L
.96134 .96449 L
.95491 .97106 L
.95163 .97494 L
.94581 .98077 L
s
.8 1 0 r
.95163 .7946 m
.95775 .79623 L
.95163 .80413 L
.94677 .79623 L
F
0 g
.95163 .7946 m
.95775 .79623 L
.95163 .80413 L
.94677 .79623 L
.95163 .7946 L
s
0 .4 1 r
.98077 .87893 m
.97606 .87393 L
.98077 .87159 L
F
0 g
.98077 .87893 m
.97606 .87393 L
.98077 .87159 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00;@3oool0103ooml0W4P00018W03OoomH0?ooo`030;md000000000000
00@0000000<0B9cO0?ooo`3oool0DP3oool01@3ogi`0B0000000000007@0_ooo05P0oooo00<0_g@0
000000000000100000000`18W=l0oooo0?ooo`1J0?ooo`040?oog`2LB00004RL0=ooo`H0oooo000]
0?ooo`040?oog`2LB00004RL0=oooeT0oooo00<0gia80000000000000`0000000`00B9`0gooo0?oo
o`1A0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`1H0?ooo`030=nLB0000000000000<0
000000<004RL0=ooo`3oool0FP3oool0103ooml0W4P00018W03Oool60?ooo`00;@3oool0103ooml0
W4P00018W03OoomJ0?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0D@3oool00`2o
M00004RL0=ooo`020?ooo`030;md0000M;l0oooo05P0oooo00<0gia804RLg`3oool00P3oool00`3O
W7@0W=oo0?ooo`1J0?ooo`040?oog`2LB00004RL0=ooo`H0oooo000Y0?ooo`030;md000000000000
00D0000000<0M;oo0?ooo`3oool0F03oool00`3OW4P0M;oo0?ooo`1D0?ooo`030?oo_`1d0000B9cO
00<0oooo00<0gia80000M02oool0F@3oool00`3OW4P0M;oo0?ooo`1J0?ooo`030;md000000000000
00D0000000<0M;oo0?ooo`3oool00`3oool001d0oooo00<0oooO09a8000000001`0000001018W=l0
oooo0?oOW01dW=l20?ooo`040?oog`2LB00004RL0=oood`0oooo00<0oooO09a8000000001`000000
0`18W=l0oooo0?ooo`030?ooo`030;mdB02Lgol0oooo05<0oooo00<0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`1J0?ooo`030;mdB02Lgol0oooo05T0oooo00P0omnL07BLg`3oool0oooo0?oo
g`2LB00004RL0=ooo`H0oooo000Z0?ooo`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`1L0?oo
o`030?oog`2LB7@0_ooo05<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`1J0?oo
o`030?oog`2LB7@0_ooo05X0oooo00L0gia807Boo`3oool0oooO09a80000B9`0gooo00H0oooo000Z
0?ooo`070?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`1M0?ooo`030?noM000M;l0oooo0580oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`1K0?ooo`030?noM000M;l0oooo05T0oooo
00L0oooO09a8M02oool0oooO09a80000B9`0gooo00H0oooo000[0?ooo`060?oOW018M;l0oooO09a8
0000B9`0goooG@3oool00`3ookl0M01809cOo`1B0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol0F`3oool00`3ookl0M01809cOo`1J0?ooo`060?oOW018M;l0oooO09a80000B9`0gooo
1P3oool002`0oooo00D0gia807Bog`2LB00004RL0=ooo`1M0?ooo`030?oo_`1d0000B9cO0580oooo
00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`1K0?ooo`030?oo_`1d0000B9cO05/0oooo
00D0gia807Bog`2LB00004RL0=ooo`060?ooo`00;03oool01@3ooml0W4Q804Q80000B9`0gooo05P0
oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`1C0?ooo`030=nLB000B9`0gooo0080
oooo00<0_g@0001d_`3oool0EP3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO05/0
oooo00D0oooO09a8B018B00004RL0=ooo`060?ooo`00;@3oool0103ogi`0B0000018W03OoomH0?oo
o`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02ooomC0?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`1G0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02ooomL0?ooo`040?oO
W018000004RL0=ooo`H0oooo000^0?ooo`030=nLB000B9`0gooo05T0oooo00H0oooO09a800000000
00000000M02ooomE0?ooo`050?oo_`1d000000000000M02oool0F@3oool01P3ooml0W4P000000000
0000001d0;oooeh0oooo00<0gia80018W03Oool01P3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00200ooooo`00002@00000040oooo000P
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0D
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`04
0?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo01@0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo00@0oooo0@0000010?ooo`004`3oool0103ooml0W4P00018W03Oool90?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`1M0?oo
o`030000003oool0oooo05d0oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`040?oo
o`4000000@3oool001<0oooo00@0oooO09a80000B9`0gooo2@3oool00`000000oooo0?ooo`3o0?oo
oh`0oooo0@0000010?ooo`004`3oool0103ooml0W4P00018W03Oool90?ooo`030000003oool0oooo
0?l0ooooS03oool100000040oooo000?0?ooo`030;md00000000000000D0000000<0M;oo0?ooo`3o
ool01P3oool00`000000oooo0?ooo`3o0?oooh`0oooo0@0000010?ooo`000`3oool00`3ooml0W4P0
000000070000000404RLg`3oool0omnL07BLg`80oooo00@0oooO09a80000B9`0gooo2@3oool00`00
0000oooo0?ooo`3o0?oooh`0oooo0@0000010?ooo`00403oool01`3OW4P0M;oo0?ooo`3ooml0W4P0
0018W03Oool02@3oool3000001<0oooo00D000000?ooo`3oool0oooo000000030?ooo`040000003o
ool0oooo000000<0oooo00D000000?ooo`3oool0oooo000000060?ooo`040000003oool0oooo0000
00H0oooo00D000000?ooo`3oool0oooo000000020?ooo`060000003oool0oooo0000003oool00000
0P3oool010000000oooo0?ooo`0000030?ooo`030000003oool0oooo00/0oooo00@000000?ooo`3o
ool000000P3oool02P000000oooo0?ooo`000000oooo0000003oool000000?ooo`0000020?ooo`04
0000003oool0oooo00000080oooo00<000000?ooo`3oool02P3oool010000000oooo0?ooo`000002
0?ooo`0?0000003oool000000?ooo`000000oooo0000003oool000000?ooo`000000oooo0000003o
ool000000080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo0080oooo01@000000?ooo`3oool000000?ooo`000000oooo0000003oool000000?ooo`00
0000oooo0000003oool000000?ooo`000000oooo00000080oooo00@000000?ooo`3oool000002@3o
ool05`000000oooo0?ooo`3oool000000?ooo`000000oooo0000003oool000000?ooo`000000oooo
0000003oool000000?ooo`000000oooo0000003oool000000080oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo010000000?ooo`3oool00000
0?ooo`000000oooo0000003oool000000?ooo`000000oooo0000003oool000000P3oool00`000000
oooo000000030?ooo`030000003oool0oooo00X0oooo00H000000?ooo`3oool000000?ooo`000002
0?ooo`050000003oool000000?ooo`0000000P3oool00`000000oooo000000020?ooo`040000003o
ool0oooo000000d0oooo00D000000?ooo`3oool0oooo000000020?ooo`060000003oool0oooo0000
003oool000000P3oool010000000oooo0?ooo`0000030?ooo`030000003oool0oooo00@0oooo00@0
00000?ooo`3oool000001`3oool01@000000oooo0?ooo`3oool000000080oooo00D000000?ooo`3o
ool0oooo000000020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0403oool30000
0040oooo000@0?ooo`070?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`090?ooo`030000003oool0
oooo00D0oooo300Do`02000000<00?lX00D00000003oIP00ofH00?mV00000002003oX`0500000000
on400?oQ003oh@0000000`00hOl00`0000000:?o002So`03002So`0300000000Xol00:?o00800000
1P00Xol01@0000000>7o003Qo`00hOl0000000800?oQ00@00000003oX`00oj<000000P00ofH00`00
00000?lX0000000201Co00050000001Ao`00DOl0057o000000003@2?o`001`000000DOl0057o001A
o`00000001Co000000000P00obP01`0000000?mV00000000oj<00000003oh@0000000P00hOl01000
00000:?o002So`000002001Vo`0300000000:?l002So00/002So00H00000001Vo`00I_l00000002S
o`000002003Qo`0=00000000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl0
0000000208oo00030000003<o`00c?l00080c?l000<000000?oe003om@000P3om@000`000000c?l0
0<co00020<co000D0000002?o`00Sol00000001Ao`00000001Co000000000?lX00000000ofH00000
003oX`0000000?oQ00000000hOl00000002So`000002001Vo`0400000000:?l002So000000T0503o
01L00000000Xo`00:?l002So00000000I_l00000002So`0000000>7o00000000on400000003oX`00
00000?mV00000000obP0000001Co00000000DOl00000000208oo00030000003<o`00c?l00080c?l0
00<000000?oe003om@000`3om@001@000000c?l00<co003<o`0000000080Sol000d00000057o0000
00005?l000000000obP00000003oIP0000000?nS00000000on40000000800>7o00<00000002So`00
00000`00I_l00`00000002So000Xo`0:000Xo`0600000000I_l006Ko00000000Xol000000P00hOl0
0`0000000?oQ00000002003oX`0500000000ofH00000003o:00000000`0Do`0010000000DOl0057o
0000000=08oo00050000001Ao`00DOl0057o000000000P0Do`001P0000000?lX003o:00000000?mV
000000800?nS00@00000003oh@00on4000000`00hOl00`0000000:?o002So`03002So`0400000000
I_l006Ko001Vo`8000001P00Xol01@0000000>7o003Qo`00hOl0000000800?oQ00D00000003oX`00
oj<00?nS00000002003oIP0300000000obP00?lX00800?lX00<0000001Co000Do`002P0Do`070?oo
o`4000000@3oool00140oooo00H0omnL04Qd_`3ooml0W4P00018W03Oool90?ooo`030000003oool0
oooo00D0oooo2`0Do`000`0000000?lX003o:002003o:00500000000ofH00?mV003oIP0000000P00
oj<01@0000000?oQ003oh@00on40000000<00>7o00<00000002So`00Xol00`00Xol00`00000006Ko
00000002002So`0300000000Xol00:?o00@00:?o00@00000003Qo`00hOl000000P00on401`000000
0?nS003oX`00oj<00000003oIP0000000P00obP0100000005?l001Co00000002057o00030000002?
o`00Sol000`0Sol000L00000057o001Ao`00DOl00000000Do`00000000800?lX00L00000003oIP00
00000?nS00000000on40000000800>7o00<00000002So`0000000`00I_l00`00000002So000Xo`0:
000Xo`0400000000I_l006Ko000000800:?o00d00000003Qo`0000000?oQ00000000oj<00000003o
IP0000000?lX0000000Do`0000000080DOl000<0000008oo00000000103<o`000`000000ooD00?oe
00030?oe00030000003<o`00c?l00080c?l001H0000008oo002?o`000000057o000000005?l00000
0000obP00000003oIP0000000?nS00000000on400000003Qo`0000000:?o00000000I_l000000P00
:?l00`000000503o01@0o`0801@0o`0700000000:?l002So000Xo`00000006Ko00000002002So`80
00000P00on4200000008003oIP0000000?lX0000000Do`000000057o0000000208oo00050000003<
o`00c?l00<co000000001P3om@001@000000c?l00<co003<o`0000000080Sol000<00000057o0000
00000P0Do`00100000000?lX00000000ofH2000000800?oQ00D00000003Qo`0000000:?o00000003
001Vo`0300000000:?l002So00/002So00H00000001Vo`00I_l00000002So`000002003Qo`050000
0000on400000003oX`0000000P00ofH00`0000000?lX0000000201Co00050000001Ao`00DOl0057o
00000000302?o`001@000000DOl0057o001Ao`00000000805?l000@00000003o:000obP000000P00
ofH00`0000000?nS00000002003oh@0500000000hOl00>7o003Qo`0000001P00Xol00`00000006Ko
001Vo`02001Vo`0300000000Xol00:?o00@00:?o00D00000003Qo`00hOl00>7o00000002003oh@04
00000000oj<00?nS000000<00?mV00<00000003o:000obP00P00obP00`0000005?l001Co000901Co
00L0oooo0@0000010?ooo`004P3oool01@3OW4P0M;oO09a80000B9`0gooo00T0oooo00<000000?oo
o`3oool01@3oool:01Co000300000000obP00?lX00800?lX00D00000003oIP00ofH00?mV00000002
003oX`0500000000on400?oQ003oh@0000000`00hOl00`0000000:?o002So`04002So`0400000000
I_l006Ko000000L00:?o00D00000003Qo`00hOl00>7o00000002003oh@0400000000oj<00?nS0000
00800?mV00<00000003o:00000000P0Do`001@000000DOl0057o001Ao`00000000d0Sol000L00000
057o001Ao`00DOl00000000Do`00000000800?lX00L00000003oIP0000000?nS00000000on400000
00800>7o00@00000002So`00Xol000000P00I_l00`00000002So000Xo`0;000Xo`0600000000I_l0
06Ko00000000Xol000000P00hOl03@0000000?oQ00000000oj<00000003oIP0000000?lX0000000D
o`000000057o000000000P2?o`000`000000c?l00<co00020<co00030000003om@00ooD000@0ooD0
00D000000<co003<o`00c?l00000000208oo000C0000001Ao`00000001Co000000000?lX00000000
ofH00000003oX`0000000?oQ00000000hOl00000002So`00000006Ko00000002000Xo`030000000D
0?l0503o00T0503o00@00000000Xo`00:?l000000P00I_l0200000000:?o00000000hOl00000003o
h@0000000?nS0P0000001000obP0000001Co00000002057o00030000002?o`00000000<0c?l000<0
00000?oe003om@001@3om@001`000000c?l00<co003<o`00000008oo000000000P1Ao`002`000000
5?l000000000obP00000003oIP0000000?nS00000000on40000000800>7o00<00000002So`000000
0P00I_l00`00000002So000Xo`0;000Xo`0600000000I_l006Ko00000000Xol000000P00hOl00`00
00000?oQ00000002003oX`0500000000ofH00000003o:00000000`0Do`0010000000DOl0057o0000
000=08oo00050000001Ao`00DOl0057o000000000P0Do`00100000000?lX003o:0000002003oIP03
00000000oj<0000000800?oQ00D00000003Qo`00hOl00>7o00000005002So`0300000000I_l006Ko
008006Ko00<00000002So`00Xol01000Xol01@0000000>7o003Qo`00hOl0000000<00?oQ00@00000
003oX`00oj<000000`00ofH01@0000000?lX003o:000obP0000000/05?l01`3oool100000040oooo
000B0?ooo`050?oog`2LB4P0B4P00018W03Oool02@3oool00`000000oooo0?ooo`050?ooo`X05?l0
00D00000003o:000obP00?lX00000003003oIP0500000000oj<00?nS003oX`0000000P00on401@00
00000>7o003Qo`00hOl0000000H00:?o00<00000001Vo`00I_l00P00I_l00`0000000:?o002So`04
002So`0400000000hOl00>7o000000800?oQ00@00000003oX`00oj<000000P00ofH01@0000000?lX
003o:00000005?l0008000000P1Ao`000`000000Sol008oo000<08oo00080000001Ao`00DOl0057o
000000005?l000000000obP200000006003oIP0000000?nS00000000on4000000P00hOl00`000000
0:?o00000003001Vo`0300000000:?l002So00/002So00H00000001Vo`00I_l00000002So`000002
003Qo`0=00000000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl000000002
08oo00050000003<o`00c?l00<co000000001`3om@001@000000c?l00<co003<o`0000000080Sol0
00`00000057o000000005?l000000000obP00000003oIP0000000?nS00000000on42000000800:?o
00<00000001Vo`0000000P00:?l00`000000503o01@0o`0901@0o`0400000000:?l002So00000080
06Ko00P00000002So`0000000>7o00000000on400000003oX`80000000@00?lX0000000Do`000000
0P1Ao`000`000000Sol0000000030<co00030000003om@00ooD000D0ooD000D000000<co003<o`00
c?l00000000208oo00030000001Ao`00000000805?l000@00000003o:00000000?mV0P000002003o
h@0500000000hOl00000002So`0000000P00I_l00`00000002So000Xo`0;000Xo`0700000000I_l0
06Ko001Vo`0000000:?o00000002003Qo`0500000000on400000003oX`0000000P00ofH00`000000
0?lX0000000201Co00040000001Ao`00DOl0000000h0Sol000D00000057o001Ao`00DOl000000002
01Co000300000000obP0000000800?mV00@00000003oX`00oj<000000P00on401@0000000>7o003Q
o`00hOl0000000D00:?o00<00000001Vo`00I_l00P00I_l00`0000000:?o002So`04002So`050000
0000hOl00>7o003Qo`0000000`00on40100000000?nS003oX`000003003oIP0500000000obP00?lX
003o:00000002P0Do`070?ooo`4000000@3oool001<0oooo00@0omnL04P00000B9`0gooo2@3oool0
0`000000oooo0?ooo`050?ooo`P05?l00P000003003o:00500000000ofH00?mV003oIP0000000P00
oj<2000000800?oQ00D00000003Qo`00hOl00>7o00000005002So`8000001000I_l00`0000000:?o
002So`04002So`0400000000hOl00>7o000000800?oQ00@00000003oX`00oj<000000P00ofH01P00
00000?lX003o:00000005?l0000000<0DOl000<0000008oo002?o`00302?o`001@000000DOl0057o
001Ao`00000000805?l000T00000003o:00000000?mV00000000oj<00000003oh@0000000P00hOl0
0`0000000:?o00000003001Vo`0300000000:?l002So00`002So00H00000001Vo`00I_l00000002S
o`000002003Qo`8000000P00oj<2000000800?lX00L0000001Co00000000DOl00000002?o`000000
00<0c?l000<000000?oe003om@001P3om@001@000000c?l00<co003<o`0000000080Sol000`00000
057o000000005?l000000000obP00000003oIP0000000?nS00000000on42000000800:?o00<00000
001Vo`0000000P00:?l00`000000503o01@0o`0901@0o`0400000000:?l002So0000008006Ko00P0
0000002So`0000000>7o00000000on400000003oX`80000000@00?lX0000000Do`0000000P1Ao`00
0`000000Sol0000000030<co00030000003om@00ooD000H0ooD000T000000<co003<o`00c?l00000
002?o`000000057o000000000P0Do`00100000000?lX00000000ofH2000000800?oQ00D00000003Q
o`0000000:?o00000002001Vo`0300000000:?l002So00`002So00L00000001Vo`00I_l006Ko0000
0000Xol0000000800>7o00D00000003oh@0000000?nS00000002003oIP0300000000obP000000080
5?l000@00000057o001Ao`0000003P2?o`001@000000DOl0057o001Ao`00000000805?l000<00000
003o:00000000P00ofH0100000000?nS003oX`000002003oh@0400000000hOl00>7o000000H00:?o
00<00000001Vo`00I_l00P00I_l2000000D00:?o0P000002003Qo`0500000000on400?oQ003oh@00
00000P00oj<01@0000000?mV003oIP00ofH0000000<00?lX0P00000801Co00L0oooo0@0000010?oo
o`00503oool00`3OW4P004RL0=ooo`090?ooo`030000003oool0oooo00D0oooo1`0Do`000`000000
0?lX003o:002003o:00500000000ofH00?mV003oIP0000000P00oj<01@0000000?oQ003oh@00on40
000000<00>7o00<00000002So`00Xol00`00Xol00`00000006Ko001Vo`03001Vo`0300000000Xol0
0:?o00@00:?o00@00000003Qo`00hOl000000P00on40100000000?nS003oX`000003003oIP030000
0000obP0000000805?l000@00000057o001Ao`0000003`2?o`0010000000DOl0057o0000000201Co
000300000000obP0000000800?mV0P000002003oh@0600000000hOl00>7o00000000Xol000000P00
I_l00`00000002So000Xo`0<000Xo`0600000000I_l006Ko00000000Xol000000P00hOl03@000000
0?oQ00000000oj<00000003oIP0000000?lX0000000Do`000000057o000000000P2?o`001@000000
c?l00<co003<o`00000000P0ooD000L000000<co003<o`00c?l00000002?o`0000000080DOl000@0
000001Co000000000?lX0P0000002000oj<00000003oh@0000000>7o00000000Xol000000P00I_l0
0`00000002So0000000=01@0o`0H00000000:?l002So00000000I_l00000002So`0000000>7o0000
0000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`0000000`3<
o`000`000000ooD00?oe00060?oe00050000003<o`00c?l00<co000000000P2?o`000`000000DOl0
0000000201Co008000000P00ofH2000000800?oQ00D00000003Qo`0000000:?o00000002001Vo`03
00000000:?l002So00`002So00H00000001Vo`00I_l00000002So`000002003Qo`0500000000on40
0000003oX`0000000P00ofH00`0000000?lX0000000201Co00040000001Ao`00DOl0000000l0Sol0
00D00000057o001Ao`00DOl00000000201Co000300000000obP0000000800?mV00@00000003oX`00
oj<000000P00on40100000000>7o003Qo`000006002So`0300000000I_l006Ko00<006Ko00<00000
002So`00Xol01000Xol0100000000>7o003Qo`000003003oh@0400000000oj<00?nS000000<00?mV
00<00000003o:000obP00P00obP00`0000005?l001Co000501Co00L0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooo`H05?l000<00000003o:000obP00P00obP00`0000000?mV003o
IP02000000800?nS00D00000003oh@00on400?oQ00000003003Qo`0300000000Xol00:?o00<00:?o
00<00000001Vo`00I_l00`00I_l00`0000000:?o002So`04002So`0400000000hOl00>7o00000080
0?oQ00@00000003oX`00oj<000000P00ofH200000005003o:00000005?l001Co000000000P1Ao`00
0`000000Sol008oo000=08oo00040000001Ao`00DOl0000000805?l00P0000002000obP00000003o
IP0000000?nS00000000on4000000P00hOl00`0000000:?o00000002001Vo`0300000000:?l002So
00<002So1@000005000Xo`0600000000I_l006Ko00000000Xol000000P00hOl03`0000000?oQ0000
0000oj<00000003oIP0000000?lX0000000Do`000000057o00000000Sol0000000030<co00030000
003om@00ooD000L0ooD000L000000<co003<o`00c?l00000002?o`0000000080DOl000@0000001Co
000000000?lX0P0000002000oj<00000003oh@0000000>7o00000000Xol000000P00I_l00`000000
02So0000000=01@0o`0H00000000:?l002So00000000I_l00000002So`0000000>7o00000000on40
0000003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`0000000`3<o`000`00
0000ooD00?oe00070?oe00080000003<o`00c?l00<co00000000Sol00000001Ao`020000000>01Co
000000000?lX00000000ofH00000003oX`0000000?oQ00000000hOl00000002So`000002001Vo`03
00000000:?l002So00<002So1@000005000Xo`0600000000I_l006Ko00000000Xol000000P00hOl0
1@0000000?oQ00000000oj<0000000800?mV00<00000003o:00000000P0Do`0010000000DOl0057o
0000000?08oo00050000001Ao`00DOl0057o000000000P0Do`000`0000000?lX00000002003oIP04
00000000oj<00?nS000000800?oQ00@00000003Qo`00hOl000001P00Xol00`00000006Ko001Vo`03
001Vo`0300000000Xol00:?o00@00:?o00<00000003Qo`00hOl00P000002003oh@0300000000oj<0
0?nS008000000P00ofH2000000<00?lX00<0000001Co000Do`00100Do`070?ooo`4000000@3oool0
0200oooo00<000000?ooo`3oool01@3oool501Co000300000000obP00?lX00800?lX00@00000003o
IP00ofH000000`00oj<01@0000000?oQ003oh@00on40000000<00>7o00<00000002So`00Xol00`00
Xol00`00000006Ko001Vo`04001Vo`0300000000Xol00:?o00<00:?o00@00000003Qo`00hOl00000
0P00on401`0000000?nS003oX`00oj<00000003oIP0000000P00obP0100000005?l001Co00000002
057o00030000002?o`00Sol000h0Sol000H00000057o001Ao`00000001Co00000002003o:0070000
0000ofH00000003oX`0000000?oQ00000002003Qo`0300000000Xol0000000<006Ko00<00000000X
o`00:?l00P00:?l00`000000503o01@0o`0201@0o`0300000000:?l002So00<002So00<00000001V
o`0000000P00Xol03@0000000>7o00000000on400000003oX`0000000?mV00000000obP0000001Co
000000000P1Ao`000`000000Sol0000000030<co00030000003om@00ooD000L0ooD000L000000<co
003<o`00c?l00000002?o`0000000080DOl000@0000001Co000000000?lX0P0000002P00oj<00000
003oh@0000000>7o00000000Xol00000001Vo`000002000Xo`030000000D0?l0503o00/0503o01P0
0000000Xo`00:?l00000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH0
0000003o:00000005?l00000001Ao`00000008oo000000030<co00030000003om@00ooD000L0ooD0
00L000000<co003<o`00c?l00000002?o`0000000080DOl000l0000001Co000000000?lX00000000
ofH00000003oX`0000000?oQ00000000hOl00000002So`0000000P00I_l00`00000002So000Xo`03
000Xo`030000000D0?l0503o0080503o00<00000000Xo`00:?l00P00:?l01000000006Ko001Vo`00
0002002So`0500000000hOl00000003oh@0000000P00oj<01@0000000?mV00000000obP000000080
5?l000@00000057o001Ao`0000003`2?o`001@000000DOl0057o001Ao`00000000<05?l000D00000
003o:00000000?mV00000002003oX`0500000000on400?oQ003oh@0000000P00hOl00`0000000:?o
002So`03002So`0300000000I_l006Ko00@006Ko00<00000002So`00Xol00`00Xol00`0000000>7o
003Qo`02003Qo`0400000000on400?oQ000000<00?nS00D00000003oIP00ofH00?mV00000003003o
:0030000000Do`005?l000<05?l01`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0
oooo100Do`000`0000000?lX003o:002003o:00400000000ofH00?mV000000<00?nS00D00000003o
h@00on400?oQ00000003003Qo`0300000000Xol00:?o00@00:?o00<00000001Vo`00I_l00`00I_l0
0`0000000:?o002So`03002So`0500000000hOl00>7o003Qo`0000000P00on401P0000000?nS003o
X`0000000?mV000000<00?lX00@0000001Co000Do`0000000P1Ao`000`000000Sol008oo000508oo
00030000002?o`00Sol000D0Sol000L00000057o001Ao`00DOl00000000Do`00000000800?lX00L0
0000003oIP0000000?nS00000000on40000000800>7o00<00000002So`0000000P00I_l00`000000
02So000Xo`02000Xo`030000000D0?l0503o00<0503o00<00000000Xo`00:?l00P00:?l01P000000
06Ko001Vo`0000000:?o000000800>7o00d00000003oh@0000000?nS00000000ofH00000003o:000
00005?l00000001Ao`0000000080Sol000D000000<co003<o`00c?l0000000090?oe000I0000003<
o`00c?l00<co00000000Sol00000001Ao`00000001Co000000000?lX00000000ofH00000003oX`00
00000?oQ00000000hOl00000002So`00000006Ko00000002000Xo`030000000D0?l0503o00/0503o
01P00000000Xo`00:?l00000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000
ofH00000003o:00000005?l00000001Ao`00000008oo000000030<co00030000003om@00ooD000P0
ooD000@000000<co003<o`0000000P2?o`003`000000DOl00000000Do`000000003o:00000000?mV
00000000oj<00000003oh@0000000>7o00000002002So`0300000000I_l0000000D002So00<00000
01@0o`0D0?l00P0D0?l00`00000002So000Xo`03000Xo`0600000000I_l006Ko00000000Xol00000
0P00hOl01@0000000?oQ00000000oj<0000000800?mV00<00000003o:00000000P0Do`0010000000
DOl0057o0000000708oo00030000002?o`00Sol000D0Sol000D00000057o001Ao`00DOl000000002
01Co000300000000obP0000000800?mV00@00000003oX`00oj<000000P00on40100000000>7o003Q
o`000006002So`0300000000I_l006Ko00@006Ko00<00000002So`00Xol00`00Xol01@0000000>7o
003Qo`00hOl0000000<00?oQ00@00000003oX`00oj<000000`00ofH00`0000000?lX003o:002003o
:0030000000Do`005?l000805?l01`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0
oooo0`0Do`000`0000000?lX003o:002003o:00500000000ofH00?mV003oIP0000000P00oj<01@00
00000?oQ003oh@00on40000000@00>7o00<00000002So`00Xol00`00Xol00`00000006Ko001Vo`04
001Vo`0300000000Xol00:?o00<00:?o00@00000003Qo`00hOl000000P00on40100000000?nS003o
X`000002003oIP0400000000obP00?lX000000805?l000@00000057o001Ao`0000001P2?o`030000
00L0Sol000H00000057o001Ao`00000001Co00000002003o:00700000000ofH00000003oX`000000
0?oQ00000002003Qo`0300000000Xol0000000<006Ko00<00000000Xo`00:?l00P00:?l00`000000
503o01@0o`0301@0o`0300000000:?l002So008002So01800000001Vo`00I_l00000002So`000000
0>7o00000000on400000003oX`0000000?mV00000000obP0000001Co00000002057o00040000002?
o`00Sol000000080c?l000<000000?oe003om@00203om@006@000000c?l00<co003<o`00000008oo
00000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o0000
0000Xol00000001Vo`0000000P00:?l00`000000503o01@0o`0201@0o`D000001@0D0?l00`000000
02So00000002001Vo`0600000000Xol00000003Qo`0000000?oQ0P0000002000ofH00000003o:000
00005?l00000001Ao`0000000P2?o`0010000000c?l00<co0000000:0?oe00040000003<o`00c?l0
00000080Sol001800000057o001Ao`00000001Co000000000?lX00000000ofH00000003oX`000000
0?oQ00000000hOl00000002So`000002001Vo`0300000000:?l002So008002So00<0000001@0o`0D
0?l00`0D0?l00`00000002So000Xo`02000Xo`0400000000I_l006Ko000000800:?o00D00000003Q
o`0000000?oQ00000002003oX`0500000000ofH00000003o:00000000P0Do`0010000000DOl0057o
0000000608oo00040000003<o`000000000000H0Sol000@00000057o001Ao`0000000`0Do`000`00
00000?lX00000002003oIP0400000000oj<00?nS000000800?oQ00@00000003Qo`00hOl000001P00
Xol00`00000006Ko001Vo`04001Vo`0300000000Xol00:?o00800:?o00<00000003Qo`00hOl00P00
hOl01@0000000?oQ003oh@00on40000000800?nS00D00000003oIP00ofH00?mV00000004003o:004
0000000Do`005?l001Co00L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`80
5?l000<00000003o:000obP00P00obP01@0000000?mV003oIP00ofH0000000800?nS00D00000003o
h@00on400?oQ00000003003Qo`8000001@00Xol00`00000006Ko001Vo`04001Vo`0300000000Xol0
0:?o00<00:?o00@00000003Qo`00hOl000000P00on40100000000?nS003oX`000002003oIP040000
0000obP00?lX000000805?l000@00000057o001Ao`0000001P2?o`001@000000c?l00<co003<o`00
000000D0Sol000@00000057o001Ao`0000000P0Do`002P0000000?lX003o:00000000?mV00000000
oj<00000003oh@000002003Qo`0300000000Xol00000008006Ko00<00000000Xo`00:?l00P00:?l0
0`000000503o01@0o`0401@0o`0300000000:?l002So008002So01800000001Vo`00I_l00000002S
o`0000000>7o00000000on400000003oX`0000000?mV00000000obP0000001Co00000002057o0003
0000002?o`0000000080c?l000<000000?oe003om@002@3om@006@000000c?l00<co003<o`000000
08oo00000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o
00000000Xol00000001Vo`0000000P00:?l00`000000503o01@0o`0201@0o`030000001A0?l0D@3o
0080D@3o00<0000001@0o`0D0?l00P0D0?l00`00000002So00000002001Vo`0600000000Xol00000
003Qo`0000000?oQ0P0000002000ofH00000003o:00000005?l00000001Ao`0000000P2?o`001000
0000c?l00<co0000000;0?oe00040000003<o`00c?l000000080Sol000l00000057o000000005?l0
00000000obP00000003oIP0000000?nS00000000on400000003Qo`0000000P00Xol00`00000006Ko
00000004000Xo`030000000D0?l0503o00@0503o00<00000000Xo`00:?l00P00:?l01000000006Ko
001Vo`000002002So`0700000000hOl00000003oh@0000000?nS00000002003oIP0300000000obP0
000000805?l000@00000057o001Ao`0000001@2?o`001@000000c?l00<co003<o`00000000H0Sol0
00@00000057o001Ao`0000000P0Do`00100000000?lX003o:0000002003oIP0400000000oj<00?nS
000000800?oQ00<00000003Qo`00hOl00P000005002So`0300000000I_l006Ko00@006Ko00<00000
002So`00Xol00P00Xol2000000<00>7o00D00000003oh@00on400?oQ00000002003oX`8000000P00
ofH2000000<00?lX00<0000001Co000Do`001`3oool100000040oooo000P0?ooo`030000003oool0
oooo00D0oooo00<05?l000000000obP00`00obP01@0000000?mV003oIP00ofH0000000800?nS00D0
0000003oh@00on400?oQ00000003003Qo`0300000000Xol00:?o00@00:?o00<00000001Vo`00I_l0
1000I_l00`0000000:?o002So`03002So`0400000000hOl00>7o000000800?oQ00@00000003oX`00
oj<000000`00ofH00`0000000?lX0000000201Co00050000001Ao`00DOl0057o000000001@2?o`00
0`000000Sol000000002000000D0Sol000@00000057o001Ao`0000000`0Do`002@0000000?lX0000
0000ofH00000003oX`0000000?oQ00000002003Qo`0300000000Xol00000008006Ko00<00000000X
o`00:?l00P00:?l00`000000503o01@0o`0501@0o`0300000000:?l002So008002So00<00000001V
o`0000000P00Xol03`0000000>7o00000000on400000003oX`0000000?mV00000000obP0000001Co
00000000DOl00000000208oo00030000003<o`00000000`0ooD001T000000<co003<o`00c?l00000
002?o`000000057o000000005?l000000000obP00000003oIP0000000?nS00000000on400000003Q
o`0000000:?o00000000I_l00000008002So00D0000001@0o`0D0?l0503o000000050540o`030000
000D0?l0503o0080503o00<00000000Xo`0000000P00I_l01P0000000:?o00000000hOl00000003o
h@80000000P00?mV00000000obP0000001Co00000000DOl000000080Sol000@000000<co003<o`00
00002`3om@0010000000c?l00<co0000000208oo000?0000001Ao`00000001Co000000000?lX0000
0000ofH00000003oX`0000000?oQ00000000hOl0000000800:?o00@00000001Vo`00I_l000000`00
:?l00`000000503o01@0o`0501@0o`0500000000:?l002So000Xo`0000000P00I_l01P0000000:?o
002So`0000000>7o000000800?oQ00D00000003oX`0000000?mV00000002003o:0030000000Do`00
00000080DOl000<0000008oo002?o`000`2?o`000`000000c?l00<co00020<co00030000002?o`00
Sol000@0Sol000@00000057o001Ao`0000000P0Do`000`0000000?lX00000002003oIP0500000000
oj<00?nS003oX`0000000P00on401@0000000>7o003Qo`00hOl0000000@00:?o00<00000001Vo`00
I_l01@00I_l00`0000000:?o002So`03002So`0500000000hOl00>7o003Qo`0000000`00on401@00
00000?nS003oX`00oj<0000000<00?mV00H00000003o:000obP00?lX0000000Do`070?ooo`400000
0@3oool00200oooo00<000000?ooo`3oool01@3oool0100Do`000000003o:000obP2000000800?mV
0P000002003oX`0500000000on400?oQ003oh@0000000`00hOl00`0000000:?o002So`04002So`03
00000000I_l006Ko00@006Ko00<00000002So`00Xol00`00Xol0100000000>7o003Qo`000003003o
h@0400000000oj<00?nS000000800?mV00<00000003o:00000000`0Do`0010000000DOl0057o0000
000508oo00030000003<o`0000000080Sol000<0000008oo002?o`000`2?o`0010000000DOl0057o
0000000201Co000300000000obP0000000800?mV00D00000003oX`0000000?oQ00000002003Qo`03
00000000Xol00000008006Ko00D00000000Xo`00:?l002So0000000701@0o`0300000000:?l002So
00<002So00T00000001Vo`0000000:?o00000000hOl00000003oh@0000000P00oj<2000000800?lX
0P000002057o00030000002?o`0000000080c?l000<000000?oe003om@002P3om@0010000000c?l0
0<co0000000208oo00060000001Ao`00000001Co000000000?lX0P0000002P00oj<00000003oh@00
00000>7o00000000Xol00000001Vo`000002000Xo`030000000D0?l0503o0080503o00<000000540
o`1A0?l00`1A0?l00`000000503o01@0o`0201@0o`0300000000:?l00000008006Ko00H00000002S
o`0000000>7o00000000on4200000008003oIP0000000?lX0000000Do`000000057o0000000208oo
00030000003<o`00000000d0ooD001@000000<co003<o`00000008oo00000000DOl00000000Do`00
0000003o:00000000?mV00000000oj<00000003oh@0000000>7o000000800:?o00@00000001Vo`00
I_l000000`00:?l00`000000503o01@0o`0501@0o`0300000000:?l002So008002So00@00000001V
o`00I_l000000P00Xol01`0000000>7o00000000on400000003oX`0000000P00ofH00`0000000?lX
0000000201Co00040000001Ao`00DOl0000000D0Sol000<000000<co003<o`000P3<o`000`000000
Sol008oo000308oo00070000001Ao`00DOl0057o000000005?l000000002003o:00400000000ofH0
0?mV000000800?nS00@00000003oh@00on4000000`00hOl00`0000000:?o002So`03002So`030000
0000I_l006Ko00D006Ko00<00000002So`00Xol00P00Xol00`0000000>7o003Qo`02003Qo`050000
0000on400?oQ003oh@0000000`00oj<01@0000000?mV003oIP00ofH0000000<00?lX00<000000?oo
o`3oool01@3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003o:000
obP000000`00ofH01@0000000?nS003oX`00oj<0000000<00?oQ00<00000003Qo`00hOl00P00hOl0
0`0000000:?o002So`03002So`0300000000I_l006Ko00@006Ko00<00000002So`00Xol01000Xol0
100000000>7o003Qo`000002003oh@0400000000oj<00?nS000000800?mV00@00000003o:000obP0
00000P0Do`001@000000DOl0057o001Ao`00000000@0Sol000D000000<co003<o`00c?l000000006
08oo00040000001Ao`00DOl0000000<05?l000T00000003o:00000000?mV00000000oj<00000003o
h@0000000P00hOl00`0000000:?o00000002001Vo`0300000000:?l002So008002So00<0000001@0
o`0D0?l01@0D0?l00`00000002So000Xo`02000Xo`0300000000I_l0000000800:?o00l00000003Q
o`0000000?oQ00000000oj<00000003oIP0000000?lX0000000Do`000000057o000000000P2?o`00
0`000000c?l00000000>0?oe00030000003<o`0000000080Sol000H00000057o000000005?l00000
0000obP20000000:003oX`0000000?oQ00000000hOl00000002So`00000006Ko0000008002So00D0
000001@0o`0D0?l0503o000000070540o`050000000D0?l0503o01@0o`0000000P00:?l030000000
06Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000003oIP80000000T05?l00000001A
o`00DOl00000002?o`0000000<co000000003@3om@001P000000c?l00<co00000000Sol000000080
DOl000d0000001Co000000000?lX00000000ofH00000003oX`0000000?oQ00000000hOl000000080
0:?o00@00000001Vo`00I_l000000`00:?l00`000000503o01@0o`0501@0o`0500000000:?l002So
000Xo`0000000P00I_l01P0000000:?o002So`0000000>7o000000800?oQ00D00000003oX`000000
0?mV00000002003o:0030000000Do`0000000080DOl000<0000008oo002?o`000`2?o`000`000000
c?l00<co00030<co00030000002?o`00Sol000<0Sol000L00000057o001Ao`00DOl00000000Do`00
000000800?lX00@00000003oIP00ofH000000P00oj<0100000000?oQ003oh@000003003Qo`030000
0000Xol00:?o00<00:?o00<00000001Vo`00I_l01000I_l00`0000000:?o002So`03002So`030000
0000hOl00>7o00800>7o00D00000003oh@00on400?oQ00000003003oX`0500000000ofH00?mV003o
IP0000000`00obP00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3o
ool01@3oool3003o:00500000000ofH00?mV003oIP0000000P00oj<01@0000000?oQ003oh@00on40
000000<00>7o0P000005002So`0300000000I_l006Ko00@006Ko00<00000002So`00Xol00`00Xol2
000000800>7o00@00000003oh@00on4000000P00oj<0100000000?mV003oIP000002003o:0040000
000Do`005?l000000080DOl00P00000408oo00030000003<o`00c?l000<0c?l000<0000008oo002?
o`000`2?o`0010000000DOl0057o0000000201Co000300000000obP0000000800?mV00L00000003o
X`0000000?oQ00000000hOl0000000800:?o00@00000001Vo`00I_l000000`00:?l00`000000503o
01@0o`0601@0o`0300000000:?l002So008002So00<00000001Vo`0000000P00Xol03`0000000>7o
00000000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl00000000208oo0003
0000003<o`00000000h0ooD000<000000<co000000000P2?o`001P000000DOl00000000Do`000000
003o:080000000X00?nS00000000on400000003Qo`0000000:?o00000000I_l000000P00:?l01@00
0000503o01@0o`0D0?l0000000L0D@3o00D0000001@0o`0D0?l0503o00000002000Xo`0<00000000
I_l00000002So`0000000>7o00000000on400000003oX`0000000?mV0P0000002@0Do`000000057o
001Ao`00000008oo00000000c?l00000000>0?oe000D0000003<o`00c?l00000002?o`000000057o
000000005?l000000000obP00000003oIP0000000?nS00000000on400000003Qo`000002002So`04
00000000I_l006Ko000000<002So00<0000001@0o`0D0?l01P0D0?l01@00000002So000Xo`00:?l0
0000008006Ko00P00000002So`00Xol00000003Qo`0000000?oQ000000800?nS00D00000003oIP00
00000?lX0000000201Co00040000001Ao`00DOl0000000D0Sol000<000000<co003<o`000`3<o`00
0`000000Sol008oo000308oo00040000001Ao`00DOl0000000805?l000@00000003o:000obP00000
0P00ofH0100000000?nS003oX`000002003oh@0500000000hOl00>7o003Qo`0000001@00Xol00`00
000006Ko001Vo`04001Vo`0300000000Xol00:?o00<00:?o0P000003003Qo`0500000000on400?oQ
003oh@0000000`00oj<01@0000000?mV003oIP00ofH0000000800?lX1`3oool100000040oooo000P
0?ooo`030000003oool0oooo00D0oooo0P00obP01@0000000?mV003oIP00ofH0000000800?nS00D0
0000003oh@00on400?oQ00000003003Qo`0300000000Xol00:?o00@00:?o00<00000001Vo`00I_l0
1000I_l00`0000000:?o002So`03002So`0500000000hOl00>7o003Qo`0000000P00on4010000000
0?nS003oX`000002003oIP0400000000obP00?lX000000805?l000@00000057o001Ao`0000001@2?
o`000`000000c?l00<co00030<co00030000002?o`00Sol000<0Sol000L00000057o001Ao`00DOl0
0000000Do`00000000800?lX00T00000003oIP0000000?nS00000000on400000003Qo`0000000P00
Xol01000000006Ko001Vo`000003000Xo`030000000D0?l0503o00L0503o00D00000000Xo`00:?l0
02So00000002001Vo`0300000000Xol0000000800>7o0P000002003oX`80000000T00?lX0000000D
o`005?l00000001Ao`00000008oo000000000P3<o`000`000000ooD00?oe000<0?oe00030000003<
o`0000000080Sol000H00000057o000000005?l000000000obP20000000:003oX`0000000?oQ0000
0000hOl00000002So`00000006Ko0000008002So00D0000001@0o`0D0?l0503o000000070540o`05
0000000D0?l0503o01@0o`0000000P00:?l03000000006Ko00000000Xol00000003Qo`0000000?oQ
00000000oj<00000003oIP80000000T05?l00000001Ao`00DOl00000002?o`0000000<co00000000
3P3om@004P000000c?l00<co00000000Sol00000001Ao`00000001Co000000000?lX00000000ofH0
0000003oX`0000000?oQ000000800>7o00<00000002So`0000000P00I_l01@00000002So000Xo`00
:?l0000000P0503o00D00000000Xo`00:?l002So00000003001Vo`0300000000Xol0000000800>7o
00D00000003oh@0000000?nS00000002003oIP0500000000obP0000001Co000000000P1Ao`000`00
0000Sol008oo000308oo00030000003<o`00c?l000@0c?l000<0000008oo002?o`000`2?o`001000
0000DOl0057o0000000201Co000400000000obP00?lX000000800?mV00@00000003oX`00oj<00000
0P00on401@0000000>7o003Qo`00hOl0000000D00:?o00<00000001Vo`00I_l01000I_l00`000000
0:?o002So`04002So`0500000000hOl00>7o003Qo`0000000`00on401@0000000?nS003oX`00oj<0
000000<00?mV00<00000003o:03oool01P3oool100000040oooo000P0?ooo`030000003oool0oooo
00D0oooo00@00?lX00000000ofH00?mV0P000002003oX`0500000000on400?oQ003oh@0000000`00
hOl00`0000000:?o002So`04002So`0300000000I_l006Ko00D006Ko00<00000002So`00Xol00P00
Xol01@0000000>7o003Qo`00hOl0000000800?oQ00L00000003oX`00oj<00?nS00000000ofH00000
00800?lX00@0000001Co000Do`0000000`1Ao`000`000000Sol008oo000208oo00030000003<o`00
c?l000@0c?l000<0000008oo002?o`000`2?o`0010000000DOl0057o0000000201Co000500000000
obP00000003oIP0000000P00oj<01@0000000?oQ00000000hOl0000000800:?o00@00000001Vo`00
I_l000000`00:?l00`000000503o01@0o`0601@0o`0300000000:?l002So008002So00<00000001V
o`0000000P00Xol03`0000000>7o00000000on400000003oX`0000000?mV00000000obP0000001Co
00000000DOl00000000208oo00040000003<o`00c?l0000000L0ooD000<000000?oe003om@000`3o
m@006P000000c?l00<co00000000Sol00000001Ao`00000001Co000000000?lX00000000ofH00000
003oX`0000000?oQ00000000hOl00000002So`00000006Ko00000000:?l000000`0D0?l00`000000
D@3o0540o`060540o`050000000D0?l0503o01@0o`0000000P00:?l03000000006Ko00000000Xol0
0000003Qo`0000000?oQ00000000oj<00000003oIP80000000T05?l00000001Ao`00DOl00000002?
o`0000000<co000000001`3om@000`000000ooD00?oe00040?oe00040000003<o`00c?l000000080
Sol001400000057o000000005?l000000000obP00000003oIP0000000?nS00000000on400000003Q
o`0000000:?o00000002001Vo`0500000000:?l002So000Xo`0000002@0D0?l01@00000002So000X
o`00:?l00000008006Ko00P00000002So`00Xol00000003Qo`0000000?oQ000000800?nS00D00000
003oIP0000000?lX0000000201Co00040000001Ao`00DOl0000000D0Sol000<000000<co003<o`00
103<o`000`000000Sol008oo000208oo00040000001Ao`00DOl0000000<05?l000H00000003o:000
obP00000003oIP000002003oX`0400000000on400?oQ000000<00>7o00<00000002So`00Xol01000
Xol00`00000006Ko001Vo`04001Vo`0300000000Xol00:?o00@00:?o00D00000003Qo`00hOl00>7o
00000003003oh@0500000000oj<00?nS003oX`0000000`00ofH00`000000oooo0?ooo`050?ooo`40
00000@3oool00200oooo00<000000?ooo`3oool01@3oool0100000000?mV003oIP000003003oX`05
00000000on400?oQ003oh@0000000`00hOl00`0000000:?o002So`04002So`0300000000I_l006Ko
00D006Ko00<00000002So`00Xol00`00Xol01@0000000>7o003Qo`00hOl0000000800?oQ00@00000
003oX`00oj<000000P00ofH0100000000?lX003o:000000201Co00040000001Ao`00DOl0000000D0
Sol000<000000<co003<o`000`3<o`000`000000Sol008oo000308oo00070000001Ao`00DOl0057o
000000005?l000000002003o:00900000000ofH00000003oX`0000000?oQ00000000hOl000000080
0:?o00@00000001Vo`00I_l000000`00:?l00`000000503o01@0o`0701@0o`0500000000:?l002So
000Xo`0000000P00I_l00`0000000:?o00000002003Qo`0600000000on400000003oX`0000000?mV
0P00000201Co00050000001Ao`00000008oo000000000`3<o`000`000000ooD00?oe00030?oe00<0
00001@3om@006P000000c?l00<co00000000Sol00000001Ao`00000001Co000000000?lX00000000
ofH00000003oX`0000000?oQ00000000hOl00000002So`00000006Ko00000000:?l000000`0D0?l0
0`000000D@3o0540o`070540o`050000000D0?l0503o01@0o`0000000P00:?l01P00000006Ko0000
0000Xol00000003Qo`80000000`00?nS00000000ofH00000003o:00000005?l00000001Ao`000000
08oo000000020<co00030000003om@00ooD000<0ooD000@000000?nh000000000000103om@004`00
0000c?l00<co003<o`00000008oo00000000DOl00000000Do`000000003o:00000000?mV00000000
oj<00000003oh@0000000P00hOl00`0000000:?o00000002001Vo`0500000000:?l002So000Xo`00
00002@0D0?l01@00000002So000Xo`00:?l00000008006Ko00<00000002So`0000000P00hOl01@00
00000?oQ00000000oj<0000000800?mV00D00000003o:00000005?l000000003057o00030000002?
o`00Sol00080Sol000<000000<co003<o`001@3<o`000`000000Sol008oo000208oo00040000001A
o`00DOl0000000<05?l000<00000003o:00000000P00ofH0100000000?nS003oX`000002003oh@05
00000000hOl00>7o003Qo`0000001@00Xol00`00000006Ko001Vo`05001Vo`0300000000Xol00:?o
00@00:?o00D00000003Qo`00hOl00>7o00000003003oh@0500000000oj<00?nS003oX`0000000`00
ofH00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool3
003oIP0400000000oj<00?nS000000<00?oQ00D00000003Qo`00hOl00>7o00000006002So`030000
0000I_l006Ko00D006Ko00<00000002So`00Xol00`00Xol01@0000000>7o003Qo`00hOl000000080
0?oQ00@00000003oX`00oj<000000P00ofH0100000000?lX003o:000000201Co00030000001Ao`00
DOl000800000102?o`000`000000c?l00<co00040<co00030000002?o`00Sol000<0Sol000@00000
057o001Ao`0000000P0Do`001@0000000?lX00000000ofH0000000800?nS00D00000003oh@000000
0>7o00000002002So`0400000000I_l006Ko0000008002So00<0000001@0o`0D0?l0200D0?l01@00
000002So000Xo`00:?l00000008006Ko00<00000002So`0000000P00hOl01P0000000?oQ00000000
oj<00000003oIP8000000P0Do`001@000000DOl00000002?o`0000000080c?l000<000000?oe003o
m@000P3om@02000000<0okP000<000000?oe003om@000P3om@006P000000c?l00<co00000000Sol0
0000001Ao`00000001Co000000000?lX00000000ofH00000003oX`0000000?oQ00000000hOl00000
002So`00000006Ko00000000:?l000000`0D0?l00`000000D@3o0540o`070540o`050000000D0?l0
503o01@0o`0000000P00:?l01P00000006Ko00000000Xol00000003Qo`80000000`00?nS00000000
ofH00000003o:00000005?l00000001Ao`00000008oo000000020<co00030000003om@00ooD00080
ooD000<000000?nh003o^0000P3o^0000`000000ooD00?oe00020?oe00040000003<o`00c?l00000
0080Sol001400000057o000000005?l000000000obP00000003oIP0000000?nS00000000on400000
003Qo`0000000:?o00000002001Vo`0500000000:?l002So000Xo`0000002P0D0?l01@00000002So
000Xo`00:?l00000008006Ko00D00000002So`0000000>7o00000002003oh@0500000000oj<00000
003oIP0000000P00obP00`0000005?l000000003057o00030000002?o`00Sol00080Sol000<00000
0<co003<o`001@3<o`000`000000Sol008oo000208oo00030000001Ao`00DOl0008000000P0Do`00
0`0000000?lX00000002003oIP0400000000oj<00?nS000000800?oQ00D00000003Qo`00hOl00>7o
00000005002So`0300000000I_l006Ko00D006Ko00<00000002So`00Xol01000Xol01@0000000>7o
003Qo`00hOl0000000<00?oQ00D00000003oX`00oj<00?nS00000002003oIPL0oooo0@0000010?oo
o`00803oool00`000000oooo0?ooo`050?ooo`800?mV00@00000003oX`00oj<000000`00on401@00
00000>7o003Qo`00hOl0000000H00:?o00<00000001Vo`00I_l01@00I_l00`0000000:?o002So`03
002So`0500000000hOl00>7o003Qo`0000000P00on40100000000?nS003oX`000002003oIP040000
0000obP00?lX000000805?l000@00000057o001Ao`0000001@2?o`000`000000c?l00<co00040<co
00030000002?o`00Sol000<0Sol000@00000057o001Ao`0000000P0Do`000`0000000?lX00000002
003oIP0700000000oj<00000003oh@0000000>7o00000002002So`0400000000I_l006Ko00000080
02So00<0000001@0o`0D0?l02@0D0?l01@00000002So000Xo`00:?l00000008006Ko01400000002S
o`0000000>7o00000000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl00000
000208oo00030000003<o`00000000D0ooD000<000000?nh003o^0000P3o^0000`000000ooD00?oe
00020?oe000J0000003<o`00c?l00000002?o`000000057o000000005?l000000000obP00000003o
IP0000000?nS00000000on400000003Qo`0000000:?o00000000I_l00000000Xo`00000201@0o`03
0000001A0?l0D@3o00P0D@3o00D0000001@0o`0D0?l0503o00000002000Xo`0600000000I_l00000
002So`0000000>7o0P0000003000oj<00000003oIP0000000?lX0000000Do`000000057o00000000
Sol000000080c?l000<000000?oe003om@000P3om@000`000000okP00?nh00030?nh00050000003o
m@00ooD00?oe000000000P3<o`0050000000Sol008oo00000000DOl00000000Do`000000003o:000
00000?mV00000000oj<00000003oh@0000000>7o00000000Xol000000P00I_l01@00000002So000X
o`00:?l0000000X0503o00D00000000Xo`00:?l002So00000002001Vo`0800000000Xol00:?o0000
0000hOl00000003oh@000002003oX`0500000000ofH00000003o:00000000P0Do`0010000000DOl0
057o0000000408oo00030000003<o`00c?l000H0c?l000<0000008oo002?o`000P2?o`001`000000
DOl0057o001Ao`00000001Co000000000P00obP0100000000?mV003oIP000002003oX`0400000000
on400?oQ000000<00>7o00<00000002So`00Xol00`00Xol00`00000006Ko001Vo`05001Vo`030000
0000Xol00:?o00@00:?o00D00000003Qo`00hOl00>7o00000003003oh@0600000000oj<00?nS003o
X`0000000?mV1`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00D00?mV0000
0000oj<00?nS00000003003oh@0500000000hOl00>7o003Qo`0000001P00Xol00`00000006Ko001V
o`06001Vo`0300000000Xol00:?o00800:?o00D00000003Qo`00hOl00>7o00000002003oh@040000
0000oj<00?nS000000800?mV00@00000003o:000obP000000P0Do`0010000000DOl0057o00000005
08oo00030000003<o`00c?l000D0c?l000<0000008oo002?o`000P2?o`001`000000DOl0057o001A
o`00000001Co000000000P00obP01@0000000?mV00000000oj<0000000800?oQ00D00000003Qo`00
00000:?o00000002001Vo`0500000000:?l002So000Xo`0000002P0D0?l01@00000002So000Xo`00
:?l0000000<006Ko0P000002003Qo`0600000000on400000003oX`0000000?mV0P00000201Co0003
0000001Ao`0000000080Sol000<000000<co00000000103om@000`000000okP00?nh00030?nh0003
0000003om@00ooD00080ooD000d000000<co003<o`00000008oo00000000DOl00000000Do`000000
003o:00000000?mV0080000000P00?oQ00000000hOl00000002So`00000006Ko0000008002So00@0
000001@0o`0D0?l000002`1A0?l010000000503o01@0o`000002000Xo`0600000000I_l00000002S
o`0000000>7o0P0000003000oj<00000003oIP0000000?lX0000000Do`000000057o00000000Sol0
00000080c?l000<000000?oe003om@000P3om@000`000000okP00?nh00030?nh00030000003om@00
ooD00080ooD000`000000<co003<o`00000008oo00000000DOl00000000Do`000000003o:0000002
003oIP8000000P00on42000000800:?o00<00000001Vo`0000000`00:?l00`000000503o01@0o`09
01@0o`0500000000:?l002So000Xo`0000000P00I_l01@0000000:?o00000000hOl0000000800?oQ
00D00000003oX`0000000?mV00000002003o:0030000000Do`00000000<0DOl000<0000008oo002?
o`000P2?o`000`000000c?l00<co00060<co00050000002?o`00Sol008oo000000000`1Ao`001000
00005?l001Co00000002003o:00400000000ofH00?mV000000800?nS00@00000003oh@00on400000
0`00hOl00`0000000:?o002So`03002So`0300000000I_l006Ko00D006Ko00<00000002So`00Xol0
1000Xol01@0000000>7o003Qo`00hOl0000000<00?oQ00D00000003oX`00oj<00?nS000000070?oo
o`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool0100000000?nS003oX`000003003o
h@0500000000hOl00>7o003Qo`0000001P00Xol00`00000006Ko001Vo`06001Vo`0300000000Xol0
0:?o00<00:?o00@00000003Qo`00hOl000000`00on400`0000000?nS00000002003oIP0400000000
obP00?lX000000805?l000D00000057o001Ao`00DOl00000000408oo00030000003<o`00c?l000D0
c?l000<0000008oo002?o`000`2?o`0010000000DOl0057o0000000201Co000300000000obP00000
00800?mV00L00000003oX`0000000?oQ00000000hOl0000000800:?o00@00000001Vo`00I_l00000
0P00:?l00`000000503o01@0o`0901@0o`0500000000:?l002So000Xo`0000000P00I_l04@000000
0:?o00000000hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`00
00000080Sol000<000000<co000000001@3om@000`000000okP00?nh00040?nh00050000003om@00
ooD00?oe000000000P3<o`002P000000Sol00000001Ao`00000001Co000000000?lX00000000ofH2
00000008003oh@0000000>7o00000000Xol00000001Vo`000002000Xo`040000000D0?l0503o0000
00/0D@3o00h0000001@0o`0D0?l0503o00000000:?l00000001Vo`0000000:?o00000000hOl00000
003oh@8000000P00ofH20000000401Co00000000DOl000000080Sol000<000000<co00000000103o
m@000`000000okP00?nh00040?nh00050000003om@00ooD00?oe000000000P3<o`0050000000Sol0
08oo00000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o
00000000Xol000000P00I_l01@00000002So000Xo`00:?l0000000X0503o00D00000000Xo`00:?l0
02So00000002001Vo`0800000000Xol00:?o00000000hOl00000003oh@000002003oX`0500000000
ofH00000003o:00000000P0Do`0010000000DOl0057o0000000408oo00030000003<o`00c?l000H0
c?l000<0000008oo002?o`000P2?o`001@000000DOl0057o001Ao`00000000805?l000@00000003o
:000obP000000P00ofH0100000000?nS003oX`000002003oh@0500000000hOl00>7o003Qo`000000
1@00Xol00`00000006Ko001Vo`05001Vo`0300000000Xol00:?o00@00:?o00D00000003Qo`00hOl0
0>7o00000003003oh@0500000000oj<00?nS003oX`0000001P3oool100000040oooo000P0?ooo`03
0000003oool0oooo00D0oooo0P00oj<01@0000000?oQ003oh@00on40000000<00>7o00<00000002S
o`00Xol01000Xol00`00000006Ko001Vo`05001Vo`8000001@00Xol0100000000>7o003Qo`000002
003oh@8000000P00oj<00`0000000?mV00000002003o:0040000000Do`005?l0000000<0DOl000<0
000008oo002?o`000P2?o`000`000000c?l00<co00060<co00030000002?o`00Sol00080Sol000@0
0000057o001Ao`0000000P0Do`002@0000000?lX003o:00000000?mV00000000oj<00000003oh@02
00000004003Qo`0000000:?o0000008006Ko00D00000000Xo`00:?l002So0000000;01@0o`050000
0000:?l002So000Xo`0000000P00I_l04@0000000:?o00000000hOl00000003oh@0000000?nS0000
0000ofH00000003o:00000005?l00000001Ao`0000000080Sol000<000000<co00000000103om@00
0`000000okP00?nh00050?nh00050000003om@00ooD00?oe000000000P3<o`002P000000Sol00000
001Ao`00000001Co000000000?lX00000000ofH200000008003oh@0000000>7o00000000Xol00000
001Vo`000002000Xo`040000000D0?l0503o000000/0D@3o0180000001@0o`0D0?l0503o00000000
:?l00000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH20000000401Co
00000000DOl000000080Sol000<000000<co00000000103om@000`000000okP00?nh00040?nh0003
0000003om@00ooD00080ooD001@000000<co003<o`00000008oo00000000DOl00000000Do`000000
003o:00000000?mV00000000oj<00000003oh@0000000>7o000000800:?o00<00000001Vo`000000
0`00:?l00`000000503o01@0o`0901@0o`0500000000:?l002So000Xo`0000000P00I_l00`000000
0:?o00000002003Qo`0500000000on400000003oX`0000000P00ofH00`0000000?lX0000000201Co
00040000001Ao`00DOl0000000@0Sol000<000000<co003<o`001P3<o`000`000000Sol008oo0002
08oo00050000001Ao`00DOl0057o000000000P0Do`00100000000?lX003o:0000002003oIP040000
0000oj<00?nS000000800?oQ00D00000003Qo`00hOl00>7o00000005002So`0300000000I_l006Ko
00D006Ko00<00000002So`00Xol01000Xol01@0000000>7o003Qo`00hOl0000000<00?oQ00<00000
003oX`00oj<01`3oool100000040oooo000P0?ooo`8000001P3oool01000oj<00000003oh@00on42
000000<00>7o00<00000002So`00Xol01000Xol00`00000006Ko001Vo`05001Vo`0300000000Xol0
0:?o00@00:?o00@00000003Qo`00hOl000000P00on401`0000000?nS003oX`00oj<00000003oIP00
00000P00obP01@0000005?l001Co000Do`0000000080DOl000<0000008oo002?o`000P2?o`000`00
0000c?l00<co00060<co00030000002?o`00Sol00080Sol000@00000057o001Ao`0000000P0Do`00
2P0000000?lX003o:00000000?mV00000000oj<00000003oh@000002003Qo`0300000000Xol00000
008006Ko00D00000000Xo`00:?l002So0000000<01@0o`0700000000:?l002So000Xo`00000006Ko
00000002002So`0A00000000hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l0
0000001Ao`00000008oo000000000P3<o`001@000000ooD00?oe003om@00000000P0okP000D00000
0?oe003om@00ooD0000000020<co000:0000002?o`000000057o000000005?l000000000obP00000
003oIP80000000X00?oQ00000000hOl00000002So`00000006Ko00000000:?l000000`0D0?l00`00
0000D@3o0540o`090540o`0B0000000D0?l0503o01@0o`00000002So00000000I_l00000002So`00
00000>7o00000000on400000003oX`0000000?mV0P000000100Do`000000057o0000000208oo0003
0000003<o`00000000@0ooD000<000000?nh003o^000103o^0000`000000ooD00?oe00020?oe000D
0000003<o`00c?l00000002?o`000000057o000000005?l000000000obP00000003oIP0000000?nS
00000000on400000003Qo`000002002So`0400000000I_l006Ko0000008002So00<0000001@0o`0D
0?l02P0D0?l01@00000002So000Xo`00:?l00000008006Ko00D00000002So`0000000>7o00000002
003oh@0500000000oj<00000003oIP0000000P00obP00`0000005?l000000003057o00050000002?
o`00Sol008oo000000002@3<o`02000000<0Sol000@00000057o001Ao`0000000`0Do`0010000000
0?lX003o:0000002003oIP0400000000oj<00?nS000000800?oQ00D00000003Qo`00hOl00>7o0000
0005002So`0300000000I_l006Ko00D006Ko00<00000002So`00Xol01000Xol0100000000>7o003Q
o`00hOl2000000800?oQ0P0000060?ooo`8000000@3oool00200oooo00<000000?ooo`3oool01@3o
ool0100000000?oQ003oh@000004003Qo`0300000000Xol00:?o00@00:?o00<00000001Vo`00I_l0
1@00I_l00`0000000:?o002So`04002So`0400000000hOl00>7o000000800?oQ00D00000003oX`00
oj<00?nS00000002003oIP0300000000obP0000000<05?l000@00000057o001Ao`000000102?o`00
0`000000c?l00<co00070<co00030000002?o`00Sol00080Sol000@00000057o001Ao`0000000P0D
o`001@0000000?lX00000000ofH0000000800?nS00<00000003oh@0000000P00hOl00`0000000:?o
00000002001Vo`0400000000:?l002So000000d0503o01L00000000Xo`00:?l002So00000000I_l0
0000002So`0000000>7o00000000on400000003oX`0000000?mV00000000obP0000001Co00000000
DOl00000000208oo00030000003<o`00000000@0ooD000<000000?nh003o^0001P3o^0002@000000
ooD00?oe003om@0000000<co00000000Sol000000002057o0080000000L00?lX00000000ofH00000
003oX`0000000?oQ0080000000L00:?o00000000I_l006Ko00000000:?l000000080503o00<00000
0540o`1A0?l02P1A0?l04P000000503o01@0o`0D0?l00000000Xo`00000006Ko00000000Xol00000
003Qo`0000000?oQ00000000oj<00000003oIP80000000@05?l00000001Ao`0000000P2?o`000`00
0000c?l0000000040?oe00030000003o^000okP000D0okP000D000000?oe003om@00ooD000000002
0<co000D0000002?o`00Sol00000001Ao`00000001Co000000000?lX00000000ofH00000003oX`00
00000?oQ00000000hOl00000002So`000002001Vo`0400000000:?l002So000000`0503o00D00000
000Xo`00:?l002So00000002001Vo`0600000000Xol00:?o00000000hOl000000P00on4200000080
0?mV00<00000003o:00000000P0Do`0010000000DOl0057o0000000408oo00030000003<o`00c?l0
00L0c?l000<0000008oo002?o`000P2?o`0010000000DOl0057o0000000301Co000300000000obP0
000000800?mV00@00000003oX`00oj<000000`00on401@0000000>7o003Qo`00hOl0000000D00:?o
00<00000001Vo`00I_l01@00I_l00`0000000:?o002So`04002So`0300000000hOl00>7o00800>7o
00D00000003oh@00on400?oQ000000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool0
1@3oool2003oh@0300000000hOl00>7o00800>7o00<00000002So`00Xol01000Xol00`00000006Ko
001Vo`05001Vo`0300000000Xol00:?o00@00:?o00D00000003Qo`00hOl00>7o00000002003oh@04
00000000oj<00?nS000000800?mV00@00000003o:000obP000000P0Do`001@000000DOl0057o001A
o`00000000@0Sol000<000000<co003<o`001P3<o`000`000000Sol008oo000208oo00040000001A
o`00DOl0000000805?l000X00000003o:000obP00000003oIP0000000?nS00000000on4000000P00
hOl00`0000000:?o00000002001Vo`0500000000:?l002So000Xo`000000300D0?l01`00000002So
000Xo`00:?l00000001Vo`0000000P00Xol04@0000000>7o00000000on400000003oX`0000000?mV
00000000obP0000001Co00000000DOl00000002?o`0000000080c?l000D000000?oe003om@00ooD0
000000090?nh00090000003om@00ooD00?oe00000000c?l00000002?o`0000000080DOl00P000000
1`00obP00000003oIP0000000?nS00000000on400P0000001`00Xol00000001Vo`00I_l00000000X
o`0000000P0D0?l00`000000D@3o0540o`0;0540o`040000000D0?l0503o0000008002So00H00000
001Vo`0000000:?o00000000hOl20000000<003oX`0000000?mV00000000obP0000001Co00000000
DOl00000002?o`0000000P3<o`001@000000ooD00?oe003om@00000000L0okP000<000000?oe003o
m@000P3om@0050000000c?l00<co00000000Sol00000001Ao`00000001Co000000000?lX00000000
ofH00000003oX`0000000?oQ00000000hOl000000P00Xol01P00000006Ko001Vo`00000002So0000
00d0503o00D00000000Xo`00:?l002So00000002001Vo`0300000000Xol00:?o008000000P00on40
1@0000000?nS00000000ofH0000000800?lX00<0000001Co000000000`1Ao`001@000000Sol008oo
002?o`00000000X0c?l000<0000008oo002?o`000P2?o`0010000000DOl0057o0000000201Co0004
00000000obP00?lX000000800?mV00@00000003oX`00oj<000000P00on401@0000000>7o003Qo`00
hOl0000000H00:?o00<00000001Vo`00I_l01@00I_l00`0000000:?o002So`04002So`0500000000
hOl00>7o003Qo`0000000`00on470?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3o
ool2003oh@0500000000hOl00>7o003Qo`0000001P00Xol00`00000006Ko001Vo`05001Vo`030000
0000Xol00:?o00@00:?o00H00000003Qo`00hOl00>7o00000000on42000000800?nS00@00000003o
IP00ofH000000P00obP0100000005?l001Co00000003057o00030000002?o`00Sol00080Sol000<0
00000<co003<o`001`3<o`000`000000Sol008oo000208oo00040000001Ao`00DOl0000000805?l0
00D00000003o:00000000?mV00000002003oX`0500000000on400000003Qo`0000000P00Xol01000
000006Ko001Vo`000002000Xo`030000000D0?l0503o00/0503o00L00000000Xo`00:?l002So0000
0000I_l0000000800:?o00<00000003Qo`0000000P0000003000oj<000000000000000000?lX0000
000Do`000000057o00000000Sol000000080c?l000D000000?oe003om@00ooD0000000090?nh0009
0000003om@00ooD00?oe00000000c?l00000002?o`0000000080DOl00P0000001`00obP00000003o
IP0000000?nS00000000on400P0000001`00Xol00000001Vo`00I_l00000000Xo`0000000P0D0?l0
0`000000D@3o0540o`0;0540o`040000000D0?l0503o0000008002So00H00000001Vo`0000000:?o
00000000hOl20000000<003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`00
00000P3<o`001@000000ooD00?oe003om@00000000P0okP00P0000020?oe00040000003<o`00c?l0
00000080Sol001400000057o000000005?l000000000obP00000003oIP0000000?nS00000000on40
0000003Qo`0000000:?o00000002001Vo`0400000000:?l002So000000d0503o00L00000000Xo`00
:?l002So00000000I_l0000000800:?o00<00000003Qo`0000000P00on401@0000000?nS00000000
ofH0000000800?lX00<0000001Co000000000`1Ao`001@000000Sol008oo002?o`00000000T0c?l0
00<0000008oo002?o`000`2?o`0010000000DOl0057o0000000201Co000400000000obP00?lX0000
00800?mV00<00000003oX`00oj<00P0000000`00on400000003Qo`02003Qo`0300000000Xol00:?o
00@00:?o00<00000001Vo`00I_l01@00I_l00`0000000:?o002So`04002So`0500000000hOl00>7o
003Qo`0000000P00on470?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool01000
on400000003Qo`00hOl2000000H00:?o00<00000001Vo`00I_l01@00I_l00`0000000:?o002So`04
002So`0700000000hOl00>7o003Qo`0000000?oQ00000003003oX`0400000000ofH00?mV00000080
0?lX00@0000001Co000Do`0000000`1Ao`000`000000Sol008oo000208oo00030000003<o`00c?l0
00L0c?l000<0000008oo002?o`000P2?o`0010000000DOl0057o0000000201Co000300000000obP0
000000800?mV00L00000003oX`0000000?oQ00000000hOl0000000800:?o00@00000001Vo`00I_l0
00000P00:?l00`000000503o01@0o`0<01@0o`0400000000:?l002So0000008006Ko01<00000002S
o`0000000>7o00000000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl00000
002?o`0000000080c?l000D000000?oe003om@00ooD00000000:0?nh00090000003om@00ooD00?oe
00000000c?l00000002?o`0000000080DOl00P0000001`00obP00000003oIP0000000?nS00000000
on400P0000001`00Xol00000001Vo`00I_l00000000Xo`0000000P0D0?l00`000000D@3o0540o`02
0540o`D00000101A0?l010000000503o01@0o`000002000Xo`0600000000I_l00000002So`000000
0>7o0P0000003000oj<00000003oIP0000000?lX0000000Do`000000057o00000000Sol000000080
c?l000D000000?oe003om@00ooD0000000090?nh00040000003om@00ooD000000080c?l001@00000
08oo002?o`000000057o000000005?l000000000obP00000003oIP0000000?nS00000000on400000
003Qo`0000000:?o0000008006Ko00@00000000Xo`00:?l000003P0D0?l01000000002So000Xo`00
0002001Vo`0:00000000Xol00:?o00000000hOl00000003oh@0000000?nS000000800?mV00<00000
003o:00000000P0Do`0010000000DOl0057o0000000408oo00030000003<o`00c?l000L0c?l000<0
000008oo002?o`000P2?o`001@000000DOl0057o001Ao`00000000805?l000@00000003o:000obP0
00000P00ofH01P0000000?nS003oX`00oj<00000003oh@8000000P00hOl2000000D00:?o00<00000
001Vo`00I_l01@00I_l00`0000000:?o002So`04002So`8000000P00hOl2000000L0oooo0@000001
0?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000hOl00>7o000000L00:?o00<00000
001Vo`00I_l01@00I_l00`0000000:?o002So`04002So`0700000000hOl00>7o003Qo`0000000?oQ
00000003003oX`0400000000ofH00?mV000000800?lX00@0000001Co000Do`0000000`1Ao`000`00
0000Sol008oo000208oo00030000003<o`00c?l000L0c?l000<0000008oo002?o`000P2?o`001000
0000DOl0057o0000000201Co000800000000obP00?lX00000000ofH00000003oX`000002003oh@03
00000000hOl0000000800:?o00<00000001Vo`0000000P00:?l00`000000503o01@0o`0<01@0o`07
00000000:?l002So000Xo`00000006Ko00000002002So`0400000000hOl00000000000800?nS0P00
0002003o:0070000000Do`000000057o00000000Sol0000000020<co00040000003om@00ooD00000
00/0okP000@000000?oe003om@0000000P3<o`002P000000Sol00000001Ao`00000001Co00000000
0?lX00000000ofH20000000:003oh@0000000>7o00000000Xol00000001Vo`00000002So00000080
503o00<000000540o`1A0?l00`1A0?l00`000000S`3o08l0o`0208l0o`050000001A0?l0D@3o0540
o`0000000P0D0?l02@00000002So000Xo`00000006Ko00000000Xol00000003Qo`020000000<003o
X`0000000?mV00000000obP0000001Co00000000DOl00000002?o`0000000P3<o`001@000000ooD0
0?oe003om@00000000X0okP000@000000?oe003om@0000000P3<o`004@000000Sol00000001Ao`00
000001Co000000000?lX00000000ofH00000003oX`0000000?oQ00000000hOl0000000800:?o00<0
0000001Vo`0000000P00:?l00`000000503o01@0o`0=01@0o`0400000000:?l002So0000008006Ko
00<00000002So`0000000P00hOl01`0000000?oQ00000000oj<00000003oIP0000000P00obP01000
00005?l001Co00000002057o00030000002?o`00Sol00080Sol000<000000<co003<o`001`3<o`00
0`000000Sol008oo000208oo00050000001Ao`00DOl0057o000000000P0Do`00100000000?lX003o
:0000002003oIP0400000000oj<00?nS000000<00?oQ00D00000003Qo`00hOl00>7o00000005002S
o`0300000000I_l006Ko00D006Ko00<00000002So`00Xol01@00Xol01@0000000>7o003Qo`00hOl0
000000H0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`800>7o00<00000002S
o`00Xol01@00Xol00`00000006Ko001Vo`05001Vo`0300000000Xol00:?o00@00:?o00D00000003Q
o`00hOl00>7o00000002003oh@0400000000oj<00?nS000000800?mV00@00000003o:000obP00000
0`0Do`0010000000DOl0057o0000000408oo00030000003<o`00c?l000P0c?l000D0000008oo002?
o`00Sol000000003057o00060000000Do`005?l000000000obP000000P00ofH01`0000000?nS0000
0000on400000003Qo`0000000P00Xol00`00000006Ko00000003000Xo`030000000D0?l0503o00`0
503o00@00000000Xo`00:?l000000P00I_l04`0000000:?o00000000hOl00000003oh@0000000?nS
00000000ofH00000003o:00000005?l00000001Ao`00000008oo000000000P3<o`001@000000ooD0
0?oe003om@00000000/0okP000@000000?oe003om@0000000P3<o`002P000000Sol00000001Ao`00
000001Co000000000?lX00000000ofH20000000:003oh@0000000>7o00000000Xol00000001Vo`00
000002So00000080503o00<000000540o`1A0?l00P1A0?l00`000000S`3o08l0o`0308l0o`030000
001A0?l0D@3o0080D@3o0140000001@0o`0D0?l00000000Xo`00000006Ko00000000Xol00000003Q
o`0000000?oQ00000000oj<00000003oIP020000000601Co00000000DOl00000002?o`0000000P3<
o`001@000000ooD00?oe003om@00000000X0okP000@000000?oe003om@0000000P3<o`0050000000
Sol008oo00000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@000000
0>7o00000000Xol000000P00I_l01000000002So000Xo`00000>01@0o`0400000000:?l002So0000
00<006Ko00T00000002So`0000000>7o00000000on400000003oX`0000000P00ofH00`0000000?lX
0000000201Co00040000001Ao`00DOl0000000@0Sol000<000000<co003<o`00203<o`02000000<0
Sol000D00000057o001Ao`00DOl00000000201Co000400000000obP00?lX000000800?mV00@00000
003oX`00oj<000000P00on401@0000000>7o003Qo`00hOl0000000H00:?o00<00000001Vo`00I_l0
1@00I_l00`0000000:?o002So`04002So`0400000000hOl00>7o003Qo`L0oooo0@0000010?ooo`00
803oool00`000000oooo0?ooo`050?ooo`800>7o00<00000002So`00Xol01000Xol00`00000006Ko
001Vo`05001Vo`0300000000Xol00:?o00@00:?o00D00000003Qo`00hOl00>7o00000002003oh@04
00000000oj<00?nS000000800?mV00@00000003o:000obP000000`0Do`0010000000DOl0057o0000
000408oo00030000003<o`00c?l000P0c?l000<0000008oo002?o`000P2?o`0010000000DOl0057o
0000000201Co000300000000obP0000000800?mV00D00000003oX`0000000?oQ00000002003Qo`03
00000000Xol00000008006Ko00@00000000Xo`00:?l000003`0D0?l01000000002So000Xo`000002
001Vo`0C00000000Xol00000003Qo`0000000?oQ00000000oj<00000003oIP0000000?lX0000000D
o`000000057o00000000Sol0000000020<co00040000003om@00ooD0000000`0okP000@000000?oe
003om@0000000P3<o`002P000000Sol00000001Ao`00000001Co000000000?lX00000000ofH20000
000:003oh@0000000>7o00000000Xol00000001Vo`00000002So00000080503o00<000000540o`1A
0?l00P1A0?l00`000000S`3o08l0o`0408l0o`050000001A0?l0D@3o0540o`0000000P0D0?l03P00
000002So00000000I_l00000002So`0000000>7o00000000on400000003oX`0000000?mV0P000000
1P0Do`000000057o00000000Sol000000080c?l000D000000?oe003om@00ooD00000000;0?nh0004
0000003om@00ooD000000080c?l00140000008oo00000000DOl00000000Do`000000003o:0000000
0?mV00000000oj<00000003oh@0000000>7o00000002002So`0300000000I_l00000008002So00<0
000001@0o`0D0?l03@0D0?l01000000002So000Xo`000002001Vo`0:00000000Xol00:?o00000000
hOl00000003oh@0000000?nS000000800?mV00<00000003o:00000000P0Do`0010000000DOl0057o
0000000408oo00030000003<o`00c?l000T0c?l000D0000008oo002?o`00Sol000000003057o0004
0000000Do`005?l0000000800?lX00@00000003oIP00ofH000000P00oj<200000003003oh@000000
0>7o00800>7o00<00000002So`00Xol01000Xol00`00000006Ko001Vo`05001Vo`0300000000Xol0
0:?o00@00:?o00<00000003Qo`00hOl01`3oool100000040oooo000P0?ooo`030000003oool0oooo
00D0oooo00<00>7o00000000Xol01@00Xol00`00000006Ko001Vo`04001Vo`8000001P00Xol01@00
00000>7o003Qo`00hOl0000000800?oQ00@00000003oX`00oj<000000P00ofH0100000000?lX003o
:000000201Co008000000P1Ao`000`000000Sol008oo000208oo00030000003<o`00c?l000P0c?l0
00<0000008oo002?o`000P2?o`0010000000DOl0057o0000000201Co000300000000obP000000080
0?mV00D00000003oX`0000000?oQ00000002003Qo`0300000000Xol00000008006Ko00@00000000X
o`00:?l000003`0D0?l02@00000002So000Xo`00:?l00000001Vo`0000000:?o00000002003Qo`80
000000X00?nS00000000ofH00000003o:00000005?l00000001Ao`0000000P2?o`000`000000c?l0
000000030?oe00030000003o^000okP000X0okP000@000000?oe003om@0000000P3<o`002P000000
Sol00000001Ao`00000001Co000000000?lX00000000ofH20000000:003oh@0000000>7o00000000
Xol00000001Vo`00000002So00000080503o00D000000540o`1A0?l0D@3o0000000708l0o`050000
001A0?l0D@3o0540o`0000000P0D0?l01@00000002So00000000I_l0000000800:?o0P0000001@00
on400000003oX`0000000?mV0080000000H05?l00000001Ao`00000008oo000000020<co00050000
003om@00ooD00?oe000000002`3o^00010000000ooD00?oe000000020<co00050000002?o`000000
057o000000000P0Do`02000000800?mV0P000002003oh@8000000P00Xol00`00000006Ko00000002
000Xo`030000000D0?l0503o00h0503o00@00000000Xo`00:?l000000P00I_l02P0000000:?o002S
o`0000000>7o00000000on400000003oX`000002003oIP0300000000obP0000000805?l000@00000
057o001Ao`000000102?o`000`000000c?l00<co00080<co00030000002?o`00Sol00080Sol000D0
0000057o001Ao`00DOl00000000201Co000400000000obP00?lX000000800?mV00H00000003oX`00
oj<00?nS00000000on42000000800>7o00<00000002So`00Xol01000Xol2000000H006Ko00<00000
002So`00Xol01000Xol2000000L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oo
o`0300000000Xol00:?o00@00:?o00<00000001Vo`00I_l01000I_l00`0000000:?o002So`05002S
o`0500000000hOl00>7o003Qo`0000000P00on40100000000?nS003oX`000002003oIP0400000000
obP00?lX000000805?l000D00000057o001Ao`00DOl00000000408oo00030000003<o`00c?l000T0
c?l000D0000008oo002?o`00Sol000000002057o00040000000Do`005?l0000000800?lX00D00000
003oIP0000000?nS00000002003oh@0300000000hOl0000000800:?o00@00000001Vo`00I_l00000
0P00:?l00`000000503o01@0o`0401@0o`030000000D0?l0503o00H0503o00@00000000Xo`00:?l0
00000P00I_l04`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH00000003o:000
00005?l00000001Ao`00000008oo000000000P3<o`0010000000ooD00?oe0000000=0?nh000E0000
003om@00ooD00000003<o`00000008oo00000000DOl00000000Do`000000003o:00000000?mV0000
0000oj<00000003oh@0000000>7o008000000P00I_l01@00000002So0000000D0?l0000000@0D@3o
00<0000008l0o`2?0?l01@2?0?l01@000000D@3o0540o`1A0?l000000080503o00<00000000Xo`00
:?l00P000002002So`80000000D00?oQ00000000oj<00000003oIP020000000601Co00000000DOl0
0000002?o`0000000P3<o`0010000000ooD00?oe0000000=0?nh00040000003om@00ooD000000080
c?l001<0000008oo00000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000003o
h@0000000>7o00000000Xol00000008006Ko00@00000000Xo`00:?l00000200D0?l00`000000503o
01@0o`0401@0o`0700000000:?l002So000Xo`00000006Ko00000002002So`0500000000hOl00000
003oh@0000000P00oj<00`0000000?mV00000002003o:0040000000Do`005?l000000080DOl000<0
000008oo002?o`000P2?o`000`000000c?l00<co00080<co00030000002?o`00Sol00080Sol000D0
0000057o001Ao`00DOl00000000201Co000400000000obP00?lX000000800?mV00@00000003oX`00
oj<000000`00on40100000000>7o003Qo`000007002So`0300000000I_l006Ko00@006Ko00<00000
002So`00Xol01@00Xol00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?oo
o`3oool01@3oool6002So`0300000000I_l006Ko00D006Ko00<00000002So`00Xol01000Xol01@00
00000>7o003Qo`00hOl0000000800?oQ00<00000003oX`00oj<00P0000001@00ofH00000003o:000
obP0000000<05?l000D00000057o001Ao`00DOl00000000308oo00030000003<o`00c?l000T0c?l0
00D0000008oo002?o`00Sol000000002057o00090000000Do`005?l001Co000000000?lX00000000
ofH0000000800?nS00<00000003oh@0000000P00hOl00`0000000:?o00000002001Vo`0400000000
:?l002So000000H0503o00<000000540o`0000001P0D0?l02@00000002So000Xo`00:?l00000001V
o`0000000:?o00000002003Qo`80000000X00?nS00000000ofH00000003o:00000005?l00000001A
o`0000000P2?o`000`000000c?l0000000030?oe00030000003o^000okP000/0okP001<000000?oe
003om@0000000<co00000000Sol00000001Ao`00000001Co000000000?lX00000000ofH00000003o
X`0000000?oQ0080000000@00:?o00000000I_l000000P00:?l00`000000503o000000030540o`03
0000002?0?l0S`3o00L0S`3o00D000000540o`1A0?l0D@3o0000000201@0o`0800000000:?l00000
001Vo`0000000:?o00000000hOl20000000>003oX`0000000?mV00000000obP0000001Co00000000
DOl00000002?o`0000000<co000000020?oe00030000003o^000okP000/0okP000@000000?oe003o
m@0000000P3<o`001@000000Sol00000001Ao`00000000805?l00P000002003oIP8000000P00on42
000000800:?o00<00000001Vo`0000000P00:?l00`000000503o01@0o`0501@0o`030000001A0?l0
000000H0503o00L00000000Xo`00:?l002So00000000I_l0000000800:?o00L00000003Qo`000000
0?oQ00000000oj<0000000800?mV00<00000003o:00000000P0Do`001@000000DOl0057o001Ao`00
000000<0Sol000<000000<co003<o`002@3<o`000`000000Sol008oo000208oo00040000001Ao`00
DOl0000000<05?l000@00000003o:000obP000000P00ofH0100000000?nS003oX`000002003oh@05
00000000hOl00>7o003Qo`0000001`00Xol00`00000006Ko001Vo`04001Vo`0300000000Xol00:?o
00@00:?o1`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo1@00Xol00`000000
06Ko001Vo`05001Vo`0300000000Xol00:?o00<00:?o0P000003003Qo`0400000000on400?oQ0000
00<00?nS00@00000003oIP00ofH000000P00obP0100000005?l001Co00000003057o00050000002?
o`00Sol008oo000000002`3<o`000`000000Sol008oo000208oo00040000001Ao`00DOl000000080
5?l000P00000003o:00000000000003oIP0000000?nS000000800?oQ00<00000003Qo`0000000P00
Xol00`00000006Ko00000003000Xo`030000000D0?l0503o00<0503o00D000000540o`1A0?l0D@3o
0000000501@0o`0400000000:?l002So0000008006Ko01400000002So`0000000>7o00000000on40
0000003oX`0000000?mV00000000obP0000001Co00000000DOl00000000208oo00030000003<o`00
000000<0ooD000<000000?nh003o^0002`3o^0004`000000ooD00?oe00000000c?l00000002?o`00
0000057o000000005?l000000000obP00000003oIP0000000?nS00000000on400P0000001000Xol0
0000001Vo`000002000Xo`030000000D0?l0000000<0D@3o00<0000008l0o`2?0?l01`2?0?l01@00
0000D@3o0540o`1A0?l000000080503o00X00000000Xo`00000006Ko00000000Xol00000003Qo`00
00000?oQ0P0000003000ofH00000003o:00000005?l00000001Ao`00000008oo00000000c?l00000
0080ooD000<000000?nh003o^000303o^00010000000ooD00?oe000000020<co000C0000002?o`00
0000057o000000005?l000000000obP00000003oIP0000000?nS00000000on400000003Qo`000000
0:?o00000002001Vo`0400000000:?l002So000000D0503o00D000000540o`1A0?l0D@3o00000005
01@0o`0500000000:?l002So000Xo`0000000P00I_l00`0000000:?o00000002003Qo`0500000000
on400000003oX`0000000P00ofH00`0000000?lX0000000201Co00040000001Ao`00DOl0000000@0
Sol000<000000<co003<o`002@3<o`000`000000Sol008oo000208oo00040000001Ao`00DOl00000
00<05?l000@00000003o:000obP000000P00ofH00`0000000?nS003oX`0200000003003oh@000000
0>7o00800>7o0P000006002So`0300000000I_l006Ko00@006Ko00<00000002So`00Xol00`00Xol7
0?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool3002So`8000001P00I_l20000
00D00:?o00<00000003Qo`00hOl00P00hOl0100000000?oQ003oh@000003003oX`0500000000ofH0
000000000000obP00P0000000`0Do`00000000000002057o008000000`2?o`000`000000c?l00<co
00090<co00030000002?o`00Sol00080Sol000@00000057o001Ao`0000000P0Do`000`0000000?lX
00000002003oIP0300000000oj<0000000800?oQ00<00000003Qo`0000000P00Xol00`00000006Ko
00000003000Xo`030000000D0?l0503o0080503o0P0000040540o`030000000D0?l0503o00<0503o
00H00000000Xo`00:?l00000001Vo`000002002So`0600000000hOl00000003oh@0000000?nS0P00
0002003o:08000000P1Ao`001@000000Sol00000003<o`00000000<0ooD000<000000?nh003o^000
303o^0004`000000ooD00?oe00000000c?l00000002?o`000000057o000000005?l000000000obP0
0000003oIP0000000?nS00000000on400P0000001000Xol00000001Vo`000002000Xo`030000000D
0?l0000000<0D@3o00<0000008l0o`2?0?l01`2?0?l01@000000D@3o0540o`1A0?l000000080503o
00X00000000Xo`00000006Ko00000000Xol00000003Qo`0000000?oQ0P0000003000ofH00000003o
:00000005?l00000001Ao`00000008oo00000000c?l000000080ooD000<000000?nh003o^0003@3o
^0001P000000ooD00?oe00000000c?l000000080Sol00P00000201Co008000000P00ofH02@000000
0?nS00000000on400000003Qo`0000000:?o00000002001Vo`0400000000:?l002So000000@0503o
00<000000540o`1A0?l00P1A0?l2000000@0503o00D00000000Xo`00:?l002So00000002001Vo`03
00000000Xol0000000800>7o00D00000003oh@0000000?nS00000002003oIP0300000000obP00000
00805?l000@00000057o001Ao`000000102?o`000`000000c?l00<co00090<co00030000002?o`00
Sol00080Sol000<00000057o001Ao`000P00000201Co000300000000obP00?lX0080000000<00?mV
00000000oj<00P00oj<00`0000000?oQ00000004003Qo`0300000000Xol00:?o00@00:?o00<00000
001Vo`00I_l01000I_l2000000<00:?o1`3oool100000040oooo000P0?ooo`030000003oool0oooo
00D0oooo0P00Xol00`00000006Ko001Vo`04001Vo`8000001@00Xol2000000<00>7o0P0000000`00
on40000000000002003oX`80000000D00?mV00000000obP00?lX0000000201Co00050000001Ao`00
DOl0057o00000000102?o`000`000000c?l00<co000:0<co00050000002?o`00Sol008oo00000000
0P1Ao`00100000005?l001Co00000002003o:00500000000ofH00000003oX`0000000P00on400`00
00000>7o00000002002So`0400000000I_l006Ko0000008002So00<0000001@0o`0D0?l00P0D0?l0
0`000000D@3o0540o`030540o`030000000D0?l0503o00<0503o00@00000000Xo`00:?l000000P00
I_l04@0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l0
0000001Ao`0000000080Sol000<000000<co000000000P3om@000`000000okP00?nh000<0?nh0004
0000003om@00ooD000000080c?l000P0000008oo00000000DOl00000000Do`000000003o:0800000
00`00?nS00000000on400000003Qo`0000000:?o00000000I_l00000000Xo`00000201@0o`040000
001A0?l0D@3o000000/0S`3o00L000000540o`1A0?l0D@3o0000000D0?l00000008002So00T00000
001Vo`0000000000003Qo`0000000?oQ00000000oj<00P0000002000obP0000001Co00000000DOl0
0000002?o`0000000P3<o`0010000000ooD00?oe0000000>0?nh00040000003om@00ooD000000080
c?l001<0000008oo00000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000003o
h@0000000>7o00000000Xol00000008006Ko00@00000000Xo`00:?l000001@0D0?l00`000000D@3o
0540o`030540o`030000000D0?l0503o0080503o00D00000000Xo`00:?l002So00000002001Vo`05
00000000Xol00000003Qo`0000000P00on400`0000000?nS00000002003oIP0300000000obP00000
00805?l000@00000057o001Ao`000000102?o`000`000000c?l00<co00090<co00030000002?o`00
Sol00080Sol00P000002057o00040000000Do`005?l0000000<00?lX00@00000003oIP0000000000
0P00oj<0100000000?oQ000000000003003Qo`8000001@00Xol2000000H006Ko00<00000002So`00
Xol01`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00<00:?o00000000I_l0
1000I_l2000000H00:?o00<00000003Qo`00hOl00P00hOl0100000000?oQ003oh@000003003oX`04
00000000ofH00?mV000000800?lX00@0000001Co000Do`0000000`1Ao`000`000000Sol008oo0002
08oo00030000003<o`00c?l000T0c?l00P00000308oo00040000001Ao`00DOl0000000805?l000P0
0000003o:000obP00000003oIP0000000?nS000000800?oQ00<00000003Qo`0000000P00Xol01000
000006Ko001Vo`000002000Xo`030000000D0?l0503o0080503o00<000000540o`1A0?l0101A0?l0
0`000000503o01@0o`0301@0o`0600000000:?l002So00000000I_l000000P00Xol04`0000000>7o
00000000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`000000
0<co000000000P3om@000`000000okP00?nh00030?nh00H00000103o^00010000000ooD00?oe0000
00020<co00080000002?o`000000057o000000005?l000000000obP20000000<003oX`0000000?oQ
00000000hOl00000002So`00000006Ko00000000:?l000000P0D0?l010000000D@3o0540o`00000;
08l0o`070000001A0?l0D@3o0540o`000000503o00000002000Xo`0;00000000I_l0000000000000
hOl00000003oh@0000000?nS00000000ofH00P0000001P0Do`000000057o00000000Sol000000080
c?l000@000000?oe003om@0000001@3o^005000000D0okP000H000000?oe003om@0000000<co0000
000208oo000A0000001Ao`00000001Co000000000?lX00000000ofH00000003oX`0000000?oQ0000
0000hOl00000002So`0000000P00I_l01000000002So000Xo`00000401@0o`030000001A0?l0D@3o
00@0D@3o00<0000001@0o`0D0?l00P0D0?l01@00000002So000Xo`00:?l00000008006Ko00D00000
002So`0000000>7o00000002003oh@0300000000oj<0000000800?mV00<00000003o:00000000P0D
o`0010000000DOl0057o0000000408oo00030000003<o`00c?l000T0c?l00P00000408oo00040000
001Ao`00DOl0000000805?l00P000002003o:00400000000ofH00?mV000000800?nS00<00000003o
h@00on400P000003003Qo`0300000000Xol00:?o00@00:?o0P000005001Vo`0300000000Xol0oooo
00H0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`800:?o00D00000001Vo`00
I_l006Ko00000008002So`0500000000hOl00>7o003Qo`0000000`00on40100000000?nS003oX`00
0002003oIP0400000000obP00?lX000000805?l000D00000057o001Ao`00DOl00000000408oo0003
0000003<o`00c?l000T0c?l000<0000008oo002?o`000`2?o`000`000000DOl00000000201Co0008
00000000obP00?lX00000000ofH00000003oX`000002003oh@0300000000hOl0000000800:?o00@0
0000001Vo`00I_l000000`00:?l01@000000503o01@0o`0D0?l0000000L0D@3o00<0000001@0o`0D
0?l00P0D0?l01000000002So000Xo`000002001Vo`0700000000Xol00000003Qo`0000000?oQ0000
0002003oX`80000000P00?lX0000000Do`000000057o00000000Sol000000080c?l000@000000?oe
003om@000000103o^0000`000000ogX00?mj00030?mj00030000003o^000okP00080okP000@00000
0?oe003om@0000000P3<o`0020000000Sol00000001Ao`00000001Co000000000?lX0P0000003P00
oj<00000003oh@0000000>7o00000000Xol00000001Vo`00000002So0000000D0?l000000`1A0?l0
0`000000S`3o08l0o`0908l0o`070000001A0?l0D@3o0540o`000000503o00000002000Xo`0;0000
0000I_l0000000000000hOl00000003oh@0000000?nS00000000ofH00P0000001P0Do`000000057o
00000000Sol000000080c?l000@000000?oe003om@0000001@3o^0000`000000ogX00?mj00020?mj
00030000003o^000okP00080okP000H000000?oe003om@0000000<co0000000208oo000A0000001A
o`00000001Co000000000?lX00000000ofH00000003oX`0000000?oQ00000000hOl00000002So`00
00000P00I_l01000000002So000Xo`00000401@0o`030000001A0?l0D@3o00D0D@3o00<0000001@0
o`0D0?l00P0D0?l01`00000002So000Xo`00:?l00000001Vo`0000000P00Xol00`0000000>7o0000
0002003oh@0500000000oj<00000003oIP0000000P00obP0100000005?l001Co00000002057o0003
0000002?o`00Sol00080Sol000<000000<co003<o`002P3<o`001@000000Sol008oo002?o`000000
00<0DOl000D0000001Co000Do`005?l000000002003o:00600000000ofH00?mV00000000oj<00000
0`00on401@0000000>7o003Qo`00hOl0000000P00:?o00<00000001Vo`00I_l00`00I_l00`000000
oooo0?ooo`050?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool0100000000:?o
00000000I_l2000000P00:?o00D00000003Qo`00hOl00>7o00000003003oh@0400000000oj<00?nS
000000800?mV00@00000003o:000obP000000P0Do`001@000000DOl0057o001Ao`00000000@0Sol0
00<000000<co003<o`002@3<o`000`000000Sol008oo000308oo00030000001Ao`00000000805?l0
00H00000003o:000obP00000003oIP000002003oX`0500000000on400000003Qo`0000000`00Xol0
0`00000006Ko00000003000Xo`050000000D0?l0503o01@0o`0000001`1A0?l00`000000503o01@0
o`0301@0o`0600000000:?l002So00000000I_l000000P00Xol04`0000000>7o00000000on400000
003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`0000000<co000000000P3o
m@000`000000okP00?nh00020?nh00030000003oNP00ogX000@0ogX000<000000?nh003o^0000P3o
^0004@000000ooD00?oe00000000c?l00000002?o`000000057o000000005?l000000000obP00000
003oIP0000000?nS0080000000P00>7o00000000Xol00000001Vo`00000002So00000080503o00@0
00000540o`1A0?l00000302?0?l01`000000D@3o0540o`1A0?l0000001@0o`0000000P00:?l02`00
000006Ko0000000000000>7o00000000on400000003oX`0000000?mV0080000000H05?l00000001A
o`00000008oo000000020<co00040000003om@00ooD0000000D0okP000<000000?mj003oNP000P3o
NP000`000000okP00?nh00020?nh00070000003om@00ooD00?oe00000000c?l00000000208oo0008
0000001Ao`00000001Co000000000?lX00000000ofH2000000800?oQ00L00000003Qo`0000000:?o
00000000I_l0000000<002So00<0000001@0o`0D0?l00P0D0?l00`000000D@3o0540o`050540o`03
0000000D0?l0503o0080503o00T00000000Xo`00:?l002So00000000I_l00000002So`0000000P00
hOl01@0000000?oQ00000000oj<0000000800?mV00<00000003o:00000000P0Do`001@000000DOl0
057o001Ao`00000000<0Sol000<000000<co003<o`002P3<o`000`000000Sol008oo000208oo0005
0000001Ao`00DOl0057o000000000`0Do`001P0000000?lX003o:00000000?mV000000800?nS00D0
0000003oh@00on400?oQ00000003003Qo`0300000000Xol00:?o00H00:?o0P000004001Vo`030000
003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000
Xol00:?o000000P00:?o0P000002003Qo`8000000P00on42000000800?nS00@00000003oIP00ofH0
00000P00obP0100000005?l001Co00000003057o00030000002?o`00Sol00080Sol000<000000<co
003<o`002P3<o`000`000000Sol008oo000208oo00040000001Ao`00DOl0000000805?l000<00000
003o:00000000P00ofH00`0000000?nS00000002003oh@0300000000hOl0000000800:?o00@00000
001Vo`00I_l000000P00:?l01@000000503o01@0o`0D0?l0000000P0D@3o00<0000001@0o`0D0?l0
0P0D0?l01000000002So000Xo`000002001Vo`0700000000Xol00000003Qo`0000000?oQ00000002
003oX`80000000P00?lX0000000Do`000000057o00000000Sol000000080c?l000@000000?oe003o
m@000000103o^0000`000000ogX00?mj00040?mj00030000003o^000okP00080okP0014000000?oe
003om@0000000<co00000000Sol00000001Ao`00000001Co000000000?lX00000000ofH00000003o
X`0200000008003Qo`0000000:?o00000000I_l00000000Xo`00000201@0o`040000001A0?l0D@3o
000000d0S`3o00@000000540o`1A0?l000000P0D0?l02P00000002So00000000I_l00000002So`00
00000>7o00000000on420000000:003oIP0000000?lX0000000Do`000000057o00000000Sol00000
0080c?l000<000000?oe00000000103o^0000`000000ogX00?mj00040?mj00030000003o^000okP0
0080okP000H000000?oe003om@0000000<co0000000208oo000A0000001Ao`00000001Co00000000
0?lX00000000ofH00000003oX`0000000?oQ00000000hOl00000002So`0000000P00I_l010000000
02So000Xo`00000401@0o`030000001A0?l0D@3o00D0D@3o00<0000001@0o`0D0?l00P0D0?l01`00
000002So000Xo`00:?l00000001Vo`0000000P00Xol00`0000000>7o00000002003oh@0300000000
oj<0000000800?mV00<00000003o:00000000P0Do`001@000000DOl0057o001Ao`00000000<0Sol0
00<000000<co003<o`002P3<o`000`000000Sol008oo000208oo00050000001Ao`00DOl0057o0000
00000`0Do`001P0000000?lX003o:00000000?mV000000800?nS0P000002003oh@8000000P00hOl2
000000P00:?o00@00000001Vo`00I_l006Ko1`3oool100000040oooo000P0?ooo`030000003oool0
oooo00D0oooo00<006Ko00000000Xol02000Xol01@0000000>7o003Qo`00hOl0000000<00?oQ00<0
0000003oX`00oj<00P0000000`00ofH0000000000002003o:0040000000Do`005?l0000000<0DOl0
00<0000008oo002?o`000P2?o`000`000000c?l00<co000:0<co00030000002?o`00Sol00080Sol0
00@00000057o001Ao`0000000P0Do`00200000000?lX003o:00000000?mV00000000oj<000000P00
on400`0000000>7o00000002002So`0400000000I_l006Ko0000008002So00<0000001@0o`0D0?l0
0P0D0?l00`000000D@3o0540o`060540o`030000000D0?l0503o0080503o00H00000000Xo`00:?l0
0000001Vo`000002002So`0A00000000hOl00000003oh@0000000?nS00000000ofH00000003o:000
00005?l00000001Ao`00000008oo000000000P3<o`0010000000ooD00?oe000000030?nh00030000
003oNP00ogX000D0ogX000<000000?nh003o^0000P3o^0004@000000ooD00?oe00000000c?l00000
002?o`000000057o000000005?l000000000obP00000003oIP0000000?nS0080000000P00>7o0000
0000Xol00000001Vo`00000002So00000080503o00@000000540o`1A0?l000003@2?0?l010000000
D@3o0540o`00000201@0o`0:00000000:?l00000001Vo`0000000:?o00000000hOl00000003oh@80
000000X00?mV00000000obP0000001Co00000000DOl00000002?o`0000000P3<o`000`000000ooD0
000000040?nh00030000003oNP00ogX000D0ogX000@000000?nh003o^000okP00P0000020?oe0005
0000003<o`00000008oo000000000P1Ao`00100000005?l000000000obP200000004003oX`000000
0?oQ000000800>7o00D00000002So`00000006Ko00000003000Xo`050000000D0?l0503o01@0o`00
0000201A0?l00`000000503o01@0o`0201@0o`0700000000:?l002So000Xo`00000006Ko00000002
002So`0300000000hOl0000000800?oQ00<00000003oX`0000000P00ofH00`0000000?lX00000002
01Co00050000001Ao`00DOl0057o000000000`2?o`000`000000c?l00<co000:0<co00030000002?
o`00Sol00080Sol000D00000057o001Ao`00DOl00000000301Co000500000000obP00?lX00000000
ofH00P000002003oX`0500000000on400?oQ003oh@0000000`00hOl2000000L00:?o0P0000000`00
I_l0oooo0?ooo`050?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool00`000000
0:?o002So`07002So`0500000000hOl00>7o003Qo`0000000`00on40100000000?nS003oX`000002
003oIP0500000000obP00?lX003o:00000000P0Do`001@000000DOl0057o001Ao`00000000@0Sol0
00<000000<co003<o`002P3<o`000`000000Sol008oo000208oo00070000001Ao`00DOl0057o0000
00005?l000000002003o:00500000000ofH00000003oX`0000000P00on401P0000000>7o003Qo`00
00000:?o0000008006Ko00@00000000Xo`00:?l00000100D0?l00`000000D@3o0540o`060540o`03
0000000D0?l0503o0080503o00H00000000Xo`00:?l00000001Vo`000002002So`0300000000hOl0
000000800?oQ0P000002003oIP80000000H05?l00000001Ao`00000008oo000000020<co00040000
003om@00ooD0000000@0okP000<000000?mj003oNP001@3oNP000`000000okP00?nh00020?nh000A
0000003om@00ooD00000003<o`00000008oo00000000DOl00000000Do`000000003o:00000000?mV
00000000oj<00P0000002000hOl00000002So`00000006Ko00000000:?l000000P0D0?l010000000
D@3o0540o`00000=08l0o`040000001A0?l0D@3o00000080503o00X00000000Xo`00000006Ko0000
0000Xol00000003Qo`0000000?oQ0P0000002P00ofH00000003o:00000005?l00000001Ao`000000
08oo000000020<co00030000003om@00000000@0okP000<000000?mj003oNP001@3oNP000`000000
okP00?nh00020?nh00080000003om@00ooD00000003<o`00000008oo00000002057o00040000000D
o`000000003o:08000000P00oj<2000000800>7o00D00000002So`00000006Ko00000003000Xo`05
0000000D0?l0503o01@0o`0000002@1A0?l00`000000503o01@0o`0201@0o`0400000000:?l002So
0000008006Ko00<00000002So`0000000P00hOl01P0000000?oQ003oh@0000000?nS000000800?mV
00<00000003o:00000000P0Do`001@000000DOl0057o001Ao`00000000<0Sol000<000000<co003<
o`002P3<o`000`000000Sol008oo000208oo00050000001Ao`00DOl0057o000000000`0Do`001000
00000?lX003o:0000002003oIP0400000000oj<00?nS000000<00?oQ00<00000003Qo`00hOl00P00
hOl00`0000000:?o002So`06002So`8000001P3oool100000040oooo000P0?ooo`030000003oool0
oooo00D0oooo2000Xol2000000800>7o0P000002003oh@8000000P00oj<01@0000000?mV003oIP00
ofH0000000800?lX00D0000001Co000Do`005?l000000002057o00030000002?o`00Sol00080Sol0
00<000000<co003<o`002P3<o`000`000000Sol008oo000208oo00070000001Ao`00DOl0057o0000
00005?l000000002003o:00500000000ofH00000003oX`0000000P00on401P0000000>7o003Qo`00
00000:?o0000008006Ko00D00000000Xo`00:?l002So0000000401@0o`030000001A0?l0D@3o00H0
D@3o00<0000001@0o`0D0?l00P0D0?l01P00000002So000Xo`00000006Ko000000800:?o01400000
003Qo`0000000?oQ00000000oj<00000003oIP0000000?lX0000000Do`000000057o00000000Sol0
000000020<co00040000003om@00ooD0000000<0okP000<000000?mj003oNP001P3oNP001@000000
okP00?nh003o^00000000080ooD000T000000<co003<o`00000008oo00000000DOl00000000Do`00
0P0000002@00ofH00000003oX`0000000?oQ00000000hOl00000002So`020000008002So00<00000
01@0o`0000000P1A0?l00`000000S`3o08l0o`0<08l0o`040000001A0?l0D@3o00000080503o00<0
0000000Xo`0000000P00I_l200000003003Qo`0000000?oQ0080000000X00?mV00000000obP00000
01Co00000000DOl00000002?o`0000000P3<o`000`000000ooD0000000040?nh00030000003oNP00
ogX000H0ogX000<000000?nh003o^0000P3o^00020000000ooD00?oe00000000c?l00000002?o`00
00000P1Ao`02000000800?lX0P0000001P00oj<00000003oh@0000000>7o000000800:?o00<00000
001Vo`0000000`00:?l01@000000503o01@0o`0D0?l0000000T0D@3o00<0000001@0o`0D0?l00P0D
0?l01P00000002So000Xo`00000006Ko000000800:?o00H00000003Qo`00hOl00000003oh@000002
003oX`0300000000ofH0000000800?lX00@0000001Co000Do`0000000`1Ao`001@000000Sol008oo
002?o`00000000`0c?l000<0000008oo002?o`000P2?o`001@000000DOl0057o001Ao`00000000<0
5?l000<00000003o:00000000`00ofH00`0000000?nS003oX`02000000800?oQ0P000003003Qo`03
00000000Xol00:?o00H00:?o1`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo
1`00Xol01@0000000>7o003Qo`00hOl0000000<00?oQ00D00000003oX`00oj<00?nS00000003003o
IP0400000000obP00000000000805?l00P000002057o00030000002?o`00Sol00080Sol000<00000
0<co003<o`002P3<o`000`000000Sol008oo000208oo00050000001Ao`00DOl0057o000000000P0D
o`000`0000000?lX00000002003oIP0500000000oj<00000003oh@0000000P00hOl01P0000000:?o
002So`00000006Ko000000<002So00<0000001@0o`0D0?l00P0D0?l00`000000D@3o0540o`060540
o`030000000D0?l0503o0080503o00H00000000Xo`00:?l00000001Vo`000002002So`0300000000
hOl0000000800?oQ0P000002003oIP80000000H05?l00000001Ao`00000008oo000000020<co0004
0000003om@00ooD0000000@0okP000<000000?mj003oNP001P3oNP001@000000okP00?nh003o^000
00000080ooD000T000000<co003<o`00000008oo00000000DOl00000000Do`000P0000001`00ofH0
0000003oX`0000000?oQ00000000hOl00P0000001`00I_l00000000Xo`00:?l0000001@0o`000000
0P1A0?l00`000000S`3o08l0o`0308l0o`D000001@2?0?l010000000D@3o0540o`00000201@0o`06
00000000:?l00000001Vo`0000000:?o0P0000004000on400000003oX`0000000?mV00000000obP0
000001Co00000000DOl00000002?o`0000000<co000000020?oe00050000003o^000okP00?nh0000
0000203oNP000`000000okP00?nh00020?nh00080000003om@00ooD00000003<o`00000008oo0000
0002057o00040000000Do`000000003o:08000000P00oj<2000000800>7o00D00000002So`000000
06Ko00000003000Xo`050000000D0?l0503o01@0o`0000002P1A0?l01@000000503o01@0o`0D0?l0
000000<002So00<00000001Vo`0000000P00Xol0200000000>7o003Qo`0000000?oQ00000000oj<0
00000P00ofH0100000000?lX003o:000000201Co00040000001Ao`00DOl0000000@0Sol000<00000
0<co003<o`002P3<o`000`000000Sol008oo000208oo008000000P1Ao`001`0000005?l001Co000D
o`000000003o:00000000`00ofH01@0000000?nS003oX`00oj<0000000<00?oQ00D00000003Qo`00
hOl00>7o00000007002So`L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`H0
0:?o00D00000003Qo`00hOl00>7o00000003003oh@0500000000oj<00?nS003oX`0000000P00ofH2
00000003003o:00000005?l000805?l000D00000057o001Ao`00DOl00000000408oo00030000003<
o`00c?l000X0c?l000<0000008oo002?o`000P2?o`001@000000DOl0057o001Ao`00000000805?l0
00<00000003o:00000000P00ofH00`0000000?nS00000002003oh@0300000000hOl0000000800:?o
00@00000001Vo`00I_l000000P00:?l010000000503o01@0o`0D0?l2000000T0D@3o00<0000001@0
o`0D0?l00P0D0?l01P00000002So000Xo`00000006Ko000000800:?o01400000003Qo`0000000?oQ
00000000oj<00000003oIP0000000?lX0000000Do`000000057o00000000Sol0000000020<co0004
0000003om@00ooD0000000<0okP000<000000?mj003oNP001`3oNP001@000000okP00?nh003o^000
00000080ooD000T000000<co003<o`00000008oo00000000DOl00000000Do`000P0000001`00ofH0
0000003oX`0000000?oQ00000000hOl00P0000001`00I_l00000000Xo`00:?l0000001@0o`000000
0P1A0?l00`000000S`3o08l0o`0308l0o`030000003<0?l0c03o0080c03o00<0000008l0o`2?0?l0
0P2?0?l010000000D@3o0540o`00000201@0o`0800000000:?l00000001Vo`0000000:?o00000000
hOl20000000>003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`0000000<co
000000020?oe00050000003o^000okP00?nh000000002@3oNP000`000000okP00?nh00020?nh000D
0000003om@00ooD00000003<o`00000008oo00000000DOl00000000Do`000000003o:00000000?mV
00000000oj<00000003oh@000002003Qo`0500000000Xol00000001Vo`0000000`00:?l01@000000
503o01@0o`0D0?l0000000X0D@3o00D0000001@0o`0D0?l0503o00000003000Xo`0600000000I_l0
000000000000Xol000000P00hOl01@0000000?oQ00000000oj<0000000800?mV00@00000003o:000
obP000000P0Do`0010000000DOl0057o0000000408oo00030000003<o`00c?l000X0c?l000<00000
08oo002?o`000`2?o`0010000000DOl0057o0000000301Co000400000000obP00000000000800?mV
0P000002003oX`0500000000on400?oQ003oh@0000000`00hOl2000000D00:?o1`3oool100000040
oooo000P0?ooo`030000003oool0oooo00D0oooo1000Xol2000000800>7o0P000003003oh@030000
0000oj<00?nS008000000P00ofH0100000000?lX003o:000000301Co00050000001Ao`00DOl0057o
000000000`2?o`02000000d0c?l000D0000008oo002?o`00Sol000000003057o00060000000Do`00
5?l000000000obP000000P00ofH00`0000000?nS00000002003oh@0300000000hOl0000000800:?o
00@00000001Vo`00I_l000000P00:?l01@000000503o01@0o`0D0?l0000000X0D@3o00<0000001@0
o`0D0?l00P0D0?l01000000002So000Xo`000002001Vo`0;00000000Xol00000003Qo`0000000?oQ
00000000oj<00000003oIP0000000P00obP200000004057o00000000Sol000000080c?l000@00000
0?oe003om@0000000`3o^0000`000000ogX00?mj00080?mj00050000003o^000okP00?nh00000000
0P3om@001`000000c?l00<co00000000Sol00000001Ao`000P0000002@00obP00000003oIP000000
0?nS00000000on400000003Qo`0200000007001Vo`00000002So000Xo`000000503o000000020540
o`030000002?0?l0S`3o0080S`3o00<000000<`0o`3<0?l00`3<0?l00`000000S`3o08l0o`0208l0
o`040000001A0?l0D@3o00000080503o00P00000000Xo`00000006Ko00000000Xol00000003Qo`80
000000h00?nS00000000ofH00000003o:00000005?l00000001Ao`00000008oo00000000c?l00000
0080ooD000D000000?nh003o^000okP00000000:0?mj00050000003o^000okP00?nh000000000P3o
m@000`000000c?l00000000208oo008000000P0Do`02000000800?mV0P0000001`00on400000003Q
o`00hOl00000002So`0000000P00I_l01000000002So000Xo`00000301@0o`030000001A0?l0D@3o
00T0D@3o00D0000001@0o`0D0?l0503o00000003000Xo`0600000000I_l006Ko00000000Xol00000
0P00hOl01@0000000?oQ00000000oj<0000000800?mV00@00000003o:000obP000000P0Do`000`00
0000DOl00000000408oo00030000003<o`00c?l000/0c?l00P00000408oo00030000001Ao`00DOl0
008000000P0Do`00100000000?lX003o:0000003003oIP0400000000oj<00?nS000000<00?oQ0P00
0003003Qo`0300000000Xol00:?o00800:?o1`3oool100000040oooo000P0?ooo`8000001P3oool3
002So`0500000000hOl00>7o003Qo`0000001000on40100000000?nS003oX`000003003oIP040000
0000obP00?lX000000<05?l000D00000057o001Ao`00DOl00000000308oo00030000003<o`00c?l0
00`0c?l000<0000008oo002?o`000P2?o`0010000000DOl0057o0000000201Co000800000000obP0
0?lX00000000ofH00000003oX`000002003oh@0300000000hOl0000000800:?o00@00000001Vo`00
I_l000000`00:?l01@000000503o01@0o`0D0?l0000000X0D@3o00<0000001@0o`0D0?l00P0D0?l0
1P00000002So000Xo`00000006Ko000000800:?o01400000003Qo`0000000?oQ00000000oj<00000
003oIP0000000?lX0000000Do`000000057o00000000Sol0000000020<co00040000003om@00ooD0
000000<0okP000<000000?mj003oNP00203oNP001@000000okP00?nh003o^00000000080ooD000`0
00000<co00000000Sol00000001Ao`00000001Co000000000?lX00000000ofH20000000<003oh@00
00000>7o00000000Xol00000001Vo`00000002So0000000D0?l000000P1A0?l00`000000S`3o08l0
o`0308l0o`030000003<0?l0c03o00@0c03o00D0000008l0o`2?0?l0S`3o000000020540o`0;0000
000D0?l0503o00000000:?l00000001Vo`0000000:?o00000000hOl00P0000003P00oj<00000003o
IP0000000?lX0000000Do`000000057o00000000Sol00000003<o`0000000P3om@001@000000okP0
0?nh003o^000000000X0ogX000<000000?nh003o^0000P3o^0003P000000ooD00?oe00000000c?l0
0000002?o`000000057o000000005?l000000000obP000000P00ofH200000009003oh@0000000>7o
003Qo`0000000:?o00000000I_l0000000<002So00D0000001@0o`0D0?l0503o0000000:0540o`05
0000000D0?l0503o01@0o`0000000`00:?l01000000006Ko001Vo`000002002So`0500000000hOl0
0000003oh@0000000P00oj<01P0000000?mV003oIP0000000?lX000000<05?l000<00000057o0000
0000102?o`000`000000c?l00<co000<0<co00030000002?o`00Sol00080Sol000D00000057o001A
o`00DOl00000000201Co000400000000obP00?lX000000<00?mV00@00000003oX`00oj<000001000
on401@0000000>7o003Qo`00hOl0000000<00:?o1P3oool200000040oooo000P0?ooo`030000003o
ool0oooo00D0oooo00<00:?o0000000000000`00hOl00`0000000?oQ003oh@02003oh@0400000000
oj<00?nS000000800?mV0P000002003o:0030000000Do`005?l0008000000`1Ao`001@000000Sol0
08oo002?o`00000000h0c?l000<0000008oo002?o`000P2?o`0010000000DOl0057o0000000201Co
000600000000obP00?lX00000000ofH000000P00oj<00`0000000?oQ00000002003Qo`0300000000
Xol00000008006Ko00D00000000Xo`00:?l002So0000000301@0o`030000001A0?l0D@3o00P0D@3o
00<0000001@0o`0D0?l00P0D0?l01000000002So000Xo`000002001Vo`0;00000000Xol00000003Q
o`0000000?oQ00000000oj<00000003oIP0000000P00obP200000004057o00000000Sol000000080
c?l000@000000?oe003om@0000000`3o^0000`000000ogX00?mj00090?mj00050000003o^000okP0
0?nh000000000P3om@002P000000c?l00000002?o`000000057o000000005?l000000000obP20000
000>003oX`0000000?oQ00000000hOl00000002So`00000006Ko00000000:?l0000001@0o`000002
0540o`030000002?0?l0S`3o0080S`3o00<000000<`0o`3<0?l01@3<0?l00`000000S`3o08l0o`02
08l0o`060000001A0?l0D@3o0000000D0?l000000P00:?l02`00000006Ko0000000000000>7o0000
0000on400000003oX`0000000?mV0080000000H05?l00000001Ao`00000008oo000000020<co0004
0000003om@00ooD000000080okP000<000000?mj003oNP002@3oNP001@000000okP00?nh003o^000
00000080ooD000<000000<co000000000P2?o`02000000805?l00P000002003oIP8000000P00on40
1@0000000>7o00000000Xol00000008006Ko00@00000000Xo`00:?l000000`0D0?l00`000000D@3o
0540o`090540o`050000000D0?l0503o01@0o`0000000`00:?l01P00000006Ko001Vo`0000000:?o
000000800>7o00<00000003oh@0000000P00oj<01P0000000?mV003oIP0000000?lX000000805?l0
00<00000057o001Ao`000P00000308oo00800000303<o`000`000000Sol008oo000308oo00050000
001Ao`00DOl0057o000000000P0Do`000`0000000?lX003o:002000000800?mV00@00000003oX`00
oj<000001000on40100000000>7o003Qo`00hOl200000003002So`3oool0oooo00D0oooo0@000001
0?ooo`00803oool00`000000oooo0?ooo`050?ooo`0300000000hOl00>7o00800>7o00@00000003o
h@00on400?oQ0P000002003oX`0400000000ofH00?mV000000<00?lX00@0000001Co000Do`000000
101Ao`000`000000Sol008oo000208oo00030000003<o`00c?l000X0c?l00P00000408oo00040000
001Ao`00DOl0000000805?l000H00000003o:000obP00000003oIP000002003oX`0300000000on40
000000800>7o00<00000002So`0000000P00I_l01@00000002So000Xo`00:?l0000000<0503o0P00
000:0540o`030000000D0?l0503o0080503o00H00000000Xo`00:?l00000001Vo`000002002So`0A
00000000hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`000000
08oo000000000P3<o`0010000000ooD00?oe000000020?nh00030000003oNP00ogX000X0ogX000D0
00000?nh003o^000okP0000000020?oe000:0000003<o`00000008oo00000000DOl00000000Do`00
0000003o:080000000h00?nS00000000on400000003Qo`0000000:?o00000000I_l00000000Xo`00
0000503o00000080D@3o00<0000008l0o`2?0?l00P2?0?l00`000000c03o0<`0o`060<`0o`050000
002?0?l0S`3o08l0o`0000000P1A0?l00`000000503o00000002000Xo`0;00000000I_l000000000
0000hOl00000003oh@0000000?nS00000000ofH00P0000001P0Do`000000057o00000000Sol00000
0080c?l000@000000?oe003om@0000000P3o^0000`000000ogX00?mj000:0?mj00050000003o^000
okP00?nh000000000P3om@003`000000c?l00000002?o`000000057o000000005?l000000000obP0
0000003oIP0000000?nS00000002003oh@0500000000hOl00000002So`0000000P00I_l010000000
02So000Xo`00000301@0o`030000001A0?l0D@3o00T0D@3o00@0000001@0o`0D0?l0503o0P000002
000Xo`0600000000I_l006Ko00000000Xol000000P00hOl00`0000000?oQ00000002003oX`030000
0000ofH0000000800?lX0P0000000`0Do`000000057o0002057o00030000002?o`00Sol00080Sol0
00<000000<co003<o`002P3<o`000`000000Sol008oo000308oo00050000001Ao`00DOl0057o0000
00000P0Do`001@0000000?lX003o:000obP0000000800?mV00<00000003oX`00oj<00P000003003o
h@0300000000hOl00>7o00800>7o0P0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3o
ool01@3oool3003Qo`8000000`00on400`0000000?nS003oX`02000000800?mV00D00000003o:000
obP00?lX0000000301Co00050000001Ao`00DOl0057o00000000102?o`000`000000c?l00<co000:
0<co00030000002?o`00Sol000<0Sol000@00000057o001Ao`0000000P0Do`001P0000000?lX003o
:00000000?mV000000800?nS00<00000003oh@0000000P00hOl00`0000000:?o00000002001Vo`05
00000000:?l002So000Xo`0000000`0D0?l00`000000D@3o0540o`090540o`030000000D0?l0503o
0080503o00H00000000Xo`00:?l00000001Vo`000002002So`0A00000000hOl00000003oh@000000
0?nS00000000ofH00000003o:00000005?l00000001Ao`00000008oo000000000P3<o`0010000000
ooD00?oe000000020?nh00030000003oNP00ogX000/0ogX000D000000?nh003o^000okP000000002
0?oe000:0000003<o`00000008oo00000000DOl00000000Do`000000003o:080000000h00?nS0000
0000on400000003Qo`0000000:?o00000000I_l00000000Xo`000000503o00000080D@3o00D00000
08l0o`2?0?l0S`3o000000090<`0o`050000002?0?l0S`3o08l0o`0000000P1A0?l00`000000503o
00000002000Xo`0;00000000I_l0000000000000hOl00000003oh@0000000?nS00000000ofH00P00
00001P0Do`000000057o00000000Sol000000080c?l000@000000?oe003om@0000000P3o^0000`00
0000ogX00?mj000;0?mj00040000003o^000okP000000080ooD000<000000<co003<o`000P000002
057o00060000000Do`000000003o:00000000?mV0P000002003oh@0500000000hOl00000002So`00
00000P00I_l01000000002So000Xo`00000301@0o`030000001A0?l0D@3o00X0D@3o00<0000001@0
o`0D0?l00P0D0?l01000000002So000Xo`000002001Vo`0300000000Xol0000000800>7o00<00000
003oh@0000000P00oj<00`0000000?mV00000003003o:0030000000Do`0000000080DOl000<00000
08oo002?o`000`2?o`000`000000c?l00<co000:0<co00800000102?o`0010000000DOl0057o0000
000301Co000500000000obP00?lX003o:00000000P00ofH2000000800?nS00@00000003oh@00on40
0?oQ0P000003003Qo`L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`800>7o
00<00000003oh@00on400P00on40100000000?nS003oX`000003003oIP0500000000obP00?lX003o
:00000000`0Do`001@000000DOl0057o001Ao`00000000@0Sol000<000000<co003<o`002P3<o`00
0`000000Sol008oo000308oo00040000001Ao`00DOl0000000805?l000@00000003o:000obP00000
0P00ofH01@0000000?nS00000000on40000000800>7o00<00000002So`0000000P00I_l01@000000
02So000Xo`00:?l0000000<0503o00<000000540o`1A0?l02P1A0?l01@000000503o01@0o`0D0?l0
0000008002So00H00000001Vo`00I_l00000002So`000002003Qo`0?00000000on400000003oX`00
00000?mV00000000obP0000001Co00000000DOl00000002?o`0000000080c?l000<000000?oe0000
00000`3o^0000`000000ogX00?mj000;0?mj000D0000003o^000okP00?nh00000000ooD00000003<
o`00000008oo00000000DOl00000000Do`000000003o:00000000?mV00000000oj<200000008003Q
o`0000000:?o00000000I_l00000000Xo`00000201@0o`030000001A0?l0000000@0S`3o00<00000
0<`0o`3<0?l01`3<0?l01@000000S`3o08l0o`2?0?l000000080D@3o00<0000001@0o`0000000P00
:?l02`00000006Ko0000000000000>7o00000000on400000003oX`0000000?mV0080000000H05?l0
0000001Ao`00000008oo000000020<co00040000003om@00ooD0000000<0okP000<000000?mj003o
NP002P3oNP001@000000okP00?nh003o^00000000080ooD000D000000<co00000000Sol000000002
057o008000000P00obP200000009003oX`0000000?oQ003oh@0000000>7o00000000Xol000000080
06Ko00@00000000Xo`00:?l000000`0D0?l00`000000D@3o0540o`090540o`030000000D0?l0503o
0080503o00@00000000Xo`00:?l000000P00I_l0100000000:?o002So`000002003Qo`0300000000
on40000000800?nS00H00000003oIP0000000000003o:000000201Co00040000001Ao`00DOl00000
00D0Sol000<000000<co003<o`002`3<o`000`000000Sol008oo000208oo00040000001Ao`00DOl0
000000<05?l000D00000003o:000obP00?lX00000003003oIP0400000000oj<00?nS000000@00?oQ
00<00000003Qo`00hOl01`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00<0
0>7o00000000on400P00on42000000800?nS00D00000003oIP00ofH00?mV00000002003o:0800000
0P0Do`0200000080DOl00P00000408oo00030000003<o`00c?l000X0c?l000<0000008oo002?o`00
0`2?o`0010000000DOl0057o0000000201Co000400000000obP00?lX000000800?mV00D00000003o
X`0000000?oQ00000002003Qo`0400000000Xol00:?o0000008006Ko00@00000000Xo`00:?l00000
100D0?l00`000000D@3o0540o`090540o`030000000D0?l0503o0080503o00H00000000Xo`00:?l0
0000001Vo`000002002So`0A00000000hOl00000003oh@0000000?nS00000000ofH00000003o:000
00005?l00000001Ao`00000008oo000000000P3<o`0010000000ooD00?oe000000020?nh00030000
003oNP00ogX000`0ogX000@000000?nh003o^00000000P3om@003P000000c?l00000002?o`000000
057o000000005?l000000000obP00000003oIP0000000?nS0P0000002000hOl00000002So`000000
06Ko00000000:?l000000P0D0?l00`000000D@3o0000000308l0o`030000003<0?l0c03o00T0c03o
00D0000008l0o`2?0?l0S`3o000000020540o`0>0000000D0?l00000000Xo`00000006Ko00000000
Xol00000003Qo`0000000?oQ00000000oj<20000000:003o:00000005?l00000001Ao`00000008oo
00000000c?l000000080ooD000D000000?nh003o^000okP00000000=0?mj00040000003o^000okP0
00000080ooD000<000000<co003<o`000P000002057o000?0000000Do`000000003o:00000000?mV
00000000oj<00000003oh@0000000>7o00000000Xol00000008006Ko00@00000000Xo`00:?l00000
0`0D0?l00`000000D@3o0540o`0:0540o`030000000D0?l0503o0080503o00@00000000Xo`00:?l0
00000P00I_l0100000000:?o002So`000002003Qo`0500000000on400000003oX`0000000`00ofH0
0`0000000?lX0000000201Co00030000001Ao`00DOl000800000102?o`000`000000c?l00<co000;
0<co00030000002?o`00Sol00080Sol000<00000057o001Ao`000P00000201Co008000000P00obP2
000000800?mV00<00000003oX`00oj<00P000003003oh@0300000000hOl0oooo00H0oooo0@000001
0?ooo`00803oool00`000000oooo0?ooo`050?ooo`0300000000on400?oQ008000000P00oj<20000
00800?mV0P000002003o:0050000000Do`005?l001Co000000000`1Ao`000`000000Sol008oo0003
08oo00030000003<o`00c?l000X0c?l000<0000008oo002?o`000`2?o`0010000000DOl0057o0000
000201Co000400000000obP00?lX000000800?mV00<00000003oX`0000000P00on400`0000000>7o
00000002002So`0400000000I_l006Ko0000008002So00<0000001@0o`0D0?l00P0D0?l00`000000
D@3o0540o`0:0540o`050000000D0?l0503o01@0o`0000000P00:?l01P00000006Ko001Vo`000000
0:?o000000800>7o00l00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001A
o`00000008oo000000000P3<o`000`000000ooD0000000030?nh00030000003oNP00ogX000/0ogX0
00D000000?nh003o^000okP0000000020?oe000>0000003<o`00000008oo00000000DOl00000000D
o`000000003o:00000000?mV00000000oj<200000008003Qo`0000000:?o00000000I_l00000000X
o`00000201@0o`030000001A0?l0000000<0S`3o00<000000<`0o`3<0?l02@3<0?l01@000000S`3o
08l0o`2?0?l000000080D@3o00h0000001@0o`00000002So00000000I_l00000002So`0000000>7o
00000000on400000003oX`80000000X00?lX0000000Do`000000057o00000000Sol00000003<o`00
00000P3om@001@000000okP00?nh003o^000000000h0ogX000@000000?nh003o^00000000P3om@00
1@000000c?l00000002?o`0000000080DOl000l0000001Co000000000?lX00000000ofH00000003o
X`0000000?oQ00000000hOl00000002So`0000000P00I_l01000000002So000Xo`00000301@0o`03
0000001A0?l0D@3o00X0D@3o00<0000001@0o`0D0?l00P0D0?l01000000002So000Xo`000002001V
o`0600000000Xol00:?o00000000hOl000000P00on400`0000000?nS00000002003oIP0400000000
obP00?lX000000805?l000D00000057o001Ao`00DOl00000000408oo00030000003<o`00c?l000/0
c?l000<0000008oo002?o`000P2?o`001@000000DOl0057o001Ao`00000000<05?l000D00000003o
:000obP00?lX00000002003oIP8000000P00oj<2000000800?oQ0P0000060?ooo`4000000@3oool0
0200oooo00<000000?ooo`3oool01@3oool2003oh@0500000000oj<00?nS003oX`0000000`00ofH0
1@0000000?lX003o:000obP0000000<05?l000D00000057o001Ao`00DOl00000000508oo00030000
003<o`00c?l000X0c?l000<0000008oo002?o`000P2?o`0200000080DOl000@0000001Co000Do`00
00000P00obP0100000000?mV003oIP000002003oX`0500000000on400000003Qo`0000000P00Xol0
1000000006Ko001Vo`000002000Xo`030000000D0?l0503o0080503o00<000000540o`1A0?l02P1A
0?l01@000000503o01@0o`0D0?l00000008002So00H00000001Vo`00I_l00000002So`000002003Q
o`0?00000000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`00
00000080c?l000<000000?oe000000000`3o^0000`000000ogX00?mj000<0?mj00050000003o^000
okP00?nh000000000P3om@003P000000c?l00000002?o`000000057o000000005?l000000000obP0
0000003oIP0000000?nS0P0000002000hOl00000002So`00000006Ko00000000:?l000000P0D0?l0
0`000000D@3o0000000308l0o`030000003<0?l0c03o00T0c03o00D0000008l0o`2?0?l0S`3o0000
00020540o`0>0000000D0?l00000000Xo`00000006Ko00000000Xol00000003Qo`0000000?oQ0000
0000oj<20000000:003o:00000005?l00000001Ao`00000008oo00000000c?l000000080ooD000D0
00000?nh003o^000okP00000000>0?mj00040000003o^000okP000000080ooD000<000000<co0000
00000P2?o`003`000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@00
00000>7o00000002002So`0300000000I_l0000000<002So00D0000001@0o`0D0?l0503o0000000<
0540o`030000000D0?l0503o0080503o00@00000000Xo`00:?l000000P00I_l01P0000000:?o002S
o`0000000>7o000000800?oQ00H00000003oX`0000000000003oIP000002003o:0040000000Do`00
5?l0000000<0DOl000<0000008oo002?o`000P2?o`000`000000c?l00<co000;0<co00030000002?
o`00Sol00080Sol000D00000057o001Ao`00DOl00000000301Co000500000000obP00?lX003o:000
00000`00ofH01@0000000?nS003oX`00oj<0000000800?oQ1`3oool100000040oooo000P0?ooo`03
0000003oool0oooo00D0oooo00@00?oQ00000000oj<00?nS0P000002003oIP8000000P00obP20000
00805?l00P000002057o008000001@2?o`000`000000c?l00<co00090<co00800000102?o`001@00
0000DOl0057o001Ao`00000000805?l000@00000003o:000obP000000P00ofH0200000000?nS003o
X`0000000?oQ00000000hOl000000P00Xol01000000006Ko001Vo`000002000Xo`030000000D0?l0
503o0080503o00<000000540o`1A0?l02`1A0?l01@000000503o01@0o`0D0?l00000008002So00h0
0000001Vo`00I_l00000002So`0000000>7o00000000on400000003oX`0000000?mV000000800?lX
00T0000001Co00000000DOl00000002?o`0000000<co000000000P3om@001@000000okP00?nh003o
^000000000h0ogX000@000000?nh003o^00000000P3om@004P000000c?l00000002?o`000000057o
000000005?l000000000obP00000003oIP0000000?nS00000000on400000003Qo`8000000P00I_l2
00000080503o00@000000540o`1A0?l000000P2?0?l00`000000c03o0<`0o`0:0<`0o`050000002?
0?l0S`3o08l0o`0000000P1A0?l00`000000503o00000002000Xo`80000000L00:?o00000000hOl0
0000003oh@0000000?nS0080000000H00?lX0000000Do`000000057o0000000208oo008000000P3o
m@001@000000okP00?nh003o^000000000l0ogX000@000000?nh003o^00000000P3om@000`000000
c?l00000000208oo000A0000001Ao`00000001Co000000000?lX00000000ofH00000003oX`000000
0?oQ00000000hOl00000002So`0000000P00I_l01000000002So000Xo`00000301@0o`030000001A
0?l0D@3o00/0D@3o00D0000001@0o`0D0?l0503o00000003000Xo`0400000000I_l006Ko00000080
0:?o00H00000003Qo`0000000000003oh@000002003oX`0300000000ofH0000000800?lX00@00000
01Co000Do`0000000`1Ao`000`000000Sol008oo000208oo00800000303<o`000`000000Sol008oo
000208oo008000000P1Ao`02000000805?l00P000002003o:08000000P00ofH2000000800?nS0P00
00070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool0100000000?nS003oX`00
0003003oIP0500000000obP00?lX003o:00000000`0Do`001@000000DOl0057o001Ao`00000000D0
Sol00P00000;0<co00030000002?o`00Sol000<0Sol000D00000057o001Ao`00DOl00000000201Co
000400000000obP00?lX000000800?mV00P00000003oX`00oj<00000003oh@0000000>7o000000<0
0:?o00@00000001Vo`00I_l000000P00:?l01@000000503o01@0o`0D0?l0000000d0D@3o00D00000
01@0o`0D0?l0503o00000002000Xo`0600000000I_l006Ko00000000Xol000000P00hOl03`000000
0?oQ00000000oj<00000003oIP0000000?lX0000000Do`000000057o00000000Sol0000000020<co
00030000003om@00000000<0okP000<000000?mj003oNP003@3oNP0010000000okP00?nh00000002
0?oe000@0000003<o`00000008oo00000000DOl00000000Do`000000003o:00000000?mV00000000
oj<00000003oh@80000000H00:?o00000000I_l00000000Xo`00000201@0o`040000001A0?l0D@3o
00000080S`3o00<000000<`0o`3<0?l02`3<0?l01`000000S`3o08l0o`2?0?l000000540o`000000
0P0D0?l01000000002So00000000I_l2000000800>7o0P0000003P00oj<00000003oIP0000000?lX
0000000Do`000000057o00000000Sol00000003<o`0000000P3om@0010000000okP00?nh0000000?
0?mj00040000003o^000okP000000080ooD000<000000<co000000000P2?o`003`000000DOl00000
000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o00000002002So`030000
0000I_l0000000<002So00D0000001@0o`0D0?l0503o0000000=0540o`050000000D0?l0503o01@0
o`0000000`00:?l01P00000006Ko001Vo`0000000:?o000000<00>7o00D00000003oh@0000000?nS
00000002003oIP0400000000obP00?lX000000805?l000D00000057o001Ao`00DOl00000000508oo
00030000003<o`00c?l000X0c?l000<0000008oo002?o`000`2?o`001@000000DOl0057o001Ao`00
000000<05?l000D00000003o:000obP00?lX00000003003oIP0500000000oj<00?nS003oX`000000
1P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo0`00oj<0100000000?mV003o
IP000003003o:0050000000Do`005?l001Co00000000101Ao`000`000000Sol008oo000208oo0003
0000003<o`00c?l000/0c?l000<0000008oo002?o`000P2?o`001@000000DOl0057o001Ao`000000
00805?l000@00000003o:000obP000000P00ofH01P0000000?nS003oX`0000000?oQ000000800>7o
00D00000002So`00Xol00000001Vo`020000008002So00D0000001@0o`0D0?l0503o0000000>0540
o`050000000D0?l0503o01@0o`0000000P00:?l04000000006Ko001Vo`0000000:?o00000000hOl0
0000003oh@0000000?nS00000000ofH00000003o:000000201Co00060000001Ao`00000008oo0000
0000c?l00P0000000`3om@0000000?nh00020?nh00030000003oNP00ogX000D0ogX00P0000060?mj
00040000003o^000okP000000080ooD0010000000<co00000000Sol00000001Ao`00000001Co0000
00000?lX00000000ofH00000003oX`0000000?oQ0P0000001P00Xol00000001Vo`00000002So0000
0080503o00@000000540o`1A0?l000000P2?0?l00`000000c03o0<`0o`0;0<`0o`070000002?0?l0
S`3o08l0o`000000D@3o0000000201@0o`0800000000:?l00000001Vo`0000000:?o00000000hOl2
0000000>003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`0000000<co0000
00020?oe00040000003o^000okP0000000H0ogX00P0000080?mj00040000003o^000okP000000080
ooD000<000000<co000000000P2?o`003`000000DOl00000000Do`000000003o:00000000?mV0000
0000oj<00000003oh@0000000>7o00000002002So`0300000000I_l00000008002So00D0000001@0
o`0D0?l0503o0000000>0540o`050000000D0?l0503o01@0o`0000000`00:?l01P00000006Ko001V
o`0000000:?o000000800>7o00H00000003oh@00on400000003oX`000002003oIP0400000000obP0
0?lX000000805?l000D00000057o001Ao`00DOl00000000408oo00030000003<o`00c?l000/0c?l0
00<0000008oo002?o`000`2?o`001@000000DOl0057o001Ao`00000000<05?l000D00000003o:000
obP00?lX00000003003oIP0300000000oj<00?nS00L0oooo0@0000010?ooo`00803oool00`000000
oooo0?ooo`050?ooo`04003oX`0000000000003oIP8000000P00obP2000000805?l00P000003057o
00800000102?o`000`000000c?l00<co000:0<co00800000102?o`001@000000DOl0057o001Ao`00
000000805?l000@00000003o:000obP000000P00ofH01P0000000?nS003oX`0000000?oQ00000080
0>7o00H00000002So`0000000000001Vo`000003000Xo`050000000D0?l0503o01@0o`0000003P1A
0?l01@000000503o01@0o`0D0?l00000008002So00`00000001Vo`00I_l00000002So`0000000>7o
00000000on400000003oX`000002003oIP8000000P0Do`001`000000DOl00000002?o`0000000<co
000000000P3om@001@000000okP00?nh003o^000000000D0ogX00`0000000`3o?@0000000?mj0005
0?mj00040000003o^000okP000000080ooD0010000000<co00000000Sol00000001Ao`00000001Co
000000000?lX00000000ofH00000003oX`0000000?oQ0P0000001P00Xol00000001Vo`00000002So
00000080503o00@000000540o`1A0?l000000P2?0?l00`000000c03o0<`0o`0;0<`0o`050000002?
0?l0S`3o08l0o`0000000P1A0?l02P000000503o00000000:?l00000001Vo`0000000:?o00000000
hOl20000000>003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`0000000<co
000000020?oe00040000003o^000okP0000000H0ogX000<000000?lm000000000P0000050?mj0004
0000003o^000okP000000080ooD000H000000<co003<o`00000008oo00000002057o0080000000H0
0?lX00000000ofH00000003oX`000002003oh@0500000000hOl00000002So`0000000P00I_l01000
000002So000Xo`00000301@0o`030000001A0?l0D@3o00`0D@3o00D0000001@0o`0D0?l0503o0000
0003000Xo`0500000000I_l006Ko00000000Xol00P0000001`00hOl00000003oh@00on400000003o
X`0000000P00ofH0100000000?lX003o:000000201Co00050000001Ao`00DOl0057o00000000102?
o`02000000`0c?l000<0000008oo002?o`000`2?o`0200000080DOl00P00000201Co008000000P00
obP2000000800?mV0P0000070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool0
100000000?mV003oIP000003003o:0050000000Do`005?l001Co000000000`1Ao`02000000D0Sol0
00<000000<co003<o`002@3<o`02000000D0Sol000D00000057o001Ao`00DOl00000000201Co0004
00000000obP00?lX000000800?mV00H00000003oX`00oj<00000003oh@000002003Qo`0300000000
Xol00000008006Ko00D00000000Xo`00:?l002So0000000301@0o`030000001A0?l0D@3o00`0D@3o
00D0000001@0o`0D0?l0503o00000002000Xo`0600000000I_l006Ko00000000Xol000000P00hOl0
3@0000000?oQ00000000oj<00000003oIP0000000?lX0000000Do`000000057o000000000P2?o`00
0`000000c?l0000000020?oe00040000003o^000okP0000000D0ogX000<000000?lm003o?@000P3o
?@000`000000ogX00?mj00030?mj00040000003o^000okP000000080ooD000<000000<co00000000
0P2?o`0050000000DOl00000000000000?lX00000000ofH00000003oX`0000000?oQ00000000hOl0
0000002So`00000006Ko00000000:?l000000P0D0?l00`000000D@3o0000000208l0o`030000003<
0?l0c03o00`0c03o00D0000008l0o`2?0?l0S`3o000000020540o`0:0000000D0?l00000000Xo`00
000006Ko00000000Xol00000003Qo`80000000h00?nS00000000ofH00000003o:00000005?l00000
001Ao`00000008oo00000000c?l000000080ooD000@000000?nh003o^00000001P3oNP000`000000
ocd00?lm00020?lm00030000003oNP00ogX000<0ogX000@000000?nh003o^00000000P3om@005000
0000c?l00<co00000000Sol00000001Ao`00000001Co000000000?lX00000000ofH00000003oX`00
00000?oQ00000000hOl000000P00Xol00`00000006Ko00000002000Xo`050000000D0?l0503o01@0
o`0000003`1A0?l01@000000503o01@0o`0D0?l00000008002So00D00000001Vo`00I_l006Ko0000
0002002So`0800000000hOl0000000000000on400000003oX`000002003oIP0400000000obP00?lX
000000805?l000D00000057o001Ao`00DOl00000000508oo008000002`3<o`000`000000Sol008oo
000408oo00050000001Ao`00DOl0057o000000000`0Do`001@0000000?lX003o:000obP0000000<0
0?mV00<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo
0P00ofH01@0000000?lX003o:000obP0000000<05?l000D00000057o001Ao`00DOl00000000608oo
00030000003<o`00c?l000T0c?l000<0000008oo002?o`00102?o`001@000000DOl0057o001Ao`00
000000805?l000@00000003o:000obP000000P00ofH01P0000000?nS003oX`0000000?oQ00000080
0>7o00@00000002So`00Xol000000P00I_l01000000002So000Xo`00000401@0o`030000001A0?l0
D@3o00`0D@3o00H0000001@0o`0D0?l0503o00000000:?l20000000;001Vo`0000000:?o002So`00
00000>7o00000000on400000003oX`0000000P00ofH2000000805?l000L00000057o00000000Sol0
0000003<o`0000000080ooD000D000000?nh003o^000okP0000000050?mj00030000003o?@00ocd0
00<0ocd000<000000?mj003oNP000P3oNP0010000000okP00?nh000000020?oe00030000003<o`00
00000080Sol001H00000057o000000000000003o:00000000?mV00000000oj<00000003oh@000000
0>7o00000000Xol00000001Vo`00000002So0000000D0?l000000P1A0?l010000000S`3o08l0o`00
000?0<`0o`070000002?0?l0S`3o08l0o`000000D@3o0000000201@0o`0600000000:?l00000001V
o`0000000:?o0P0000004000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl0
0000002?o`0000000<co000000020?oe00040000003o^000okP0000000D0ogX000<000000?lm003o
?@000P3o?@000`000000ogX00?mj00030?mj00040000003o^000okP000000080ooD000H000000<co
003<o`00000008oo00000002057o0080000000H00?lX00000000ofH00000003oX`000002003oh@05
00000000hOl00000002So`0000000P00I_l200000003000Xo`000000503o0080503o00<000000540
o`1A0?l03@1A0?l01@000000503o01@0o`0D0?l00000008002So00@00000001Vo`00I_l000000P00
Xol02@0000000>7o003Qo`00hOl00000003oh@0000000?nS00000002003oIP0400000000obP00?lX
000000805?l000D00000057o001Ao`00DOl00000000508oo00030000003<o`00c?l000X0c?l00P00
000508oo00050000001Ao`00DOl0057o000000000`0Do`001@0000000?lX003o:000obP000000080
0?mV1`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00<00?mV00000000obP0
0P00obP01@0000005?l001Co000Do`00000000<0DOl00P00000508oo00030000003<o`00c?l000X0
c?l000<0000008oo002?o`000P2?o`0200000080DOl00P00000201Co000400000000obP00?lX0000
00800?mV00H00000003oX`0000000000003oh@000002003Qo`0500000000Xol00:?o00000000I_l0
0P000002000Xo`040000000D0?l0503o01@0o`8000003P1A0?l01@000000503o01@0o`0D0?l00000
008002So00H00000001Vo`00I_l00000002So`000002003Qo`0=00000000on400000003oX`000000
0?mV00000000obP0000001Co00000000DOl00000000208oo00040000003<o`0000000?oe00800000
0P3o^0000`000000ogX00?mj00030?mj00030000003o?@00ocd000@0ocd000<000000?mj003oNP00
0P3oNP0010000000okP00?nh000000020?oe00030000003<o`0000000080Sol001H00000057o0000
00000000003o:00000000?mV00000000oj<00000003oh@0000000>7o00000000Xol00000001Vo`00
000002So0000000D0?l000000P1A0?l010000000S`3o08l0o`00000?0<`0o`070000002?0?l0S`3o
08l0o`000000D@3o0000000201@0o`0600000000:?l00000001Vo`0000000:?o0P0000004000on40
0000003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`0000000<co00000002
0?oe00040000003o^000okP0000000D0ogX000<000000?lm003o?@000`3o?@000`000000ogX00?mj
00030?mj00030000003o^000okP0008000000140ooD00000003<o`00c?l00000002?o`000000057o
000000005?l000000000obP00000003oIP0000000?nS00000002003oh@0500000000hOl00000002S
o`0000000P00I_l01000000002So000Xo`00000301@0o`8000003P1A0?l01@000000503o01@0o`0D
0?l00000008002So0P0000001@00I_l00000002So`00Xol0000000800>7o00H00000003oh@00on40
0000003oX`000002003oIP0400000000obP00?lX000000805?l00P000002057o00800000102?o`02
000000`0c?l000<0000008oo002?o`000`2?o`001@000000DOl0057o001Ao`00000000<05?l000H0
0000003o:000obP00?lX00000000ofH70?ooo`4000000@3oool00200oooo00<000000?ooo`3oool0
1@3oool00`0000000?lX003o:002000000805?l00P000002057o008000001P2?o`000`000000c?l0
0<co00090<co00800000102?o`001@000000DOl0057o001Ao`00000000<05?l000D00000003o:000
00000000003oIP0200000005003oX`0000000?oQ003oh@0000000P00hOl01P0000000:?o002So`00
000006Ko000000<002So00D0000001@0o`0D0?l0503o0000000?0540o`050000000D0?l0503o01@0
o`0000000P00:?l01P00000006Ko001Vo`0000000:?o000000800>7o00/00000003oh@0000000?nS
00000000ofH00000003o:00000005?l000000002057o00050000002?o`0000000<co000000000P3o
m@0010000000okP00?nh000000050?mj00030000003o?@00ocd000D0ocd000<000000?mj003oNP00
0P3oNP0010000000okP00?nh000000020?oe00030000003<o`0000000080Sol001H00000057o0000
00000000003o:00000000?mV00000000oj<00000003oh@0000000>7o00000000Xol00000001Vo`00
000002So0000000D0?l000000P1A0?l010000000S`3o08l0o`0000050<`0o`D000001@3<0?l01`00
0000S`3o08l0o`2?0?l000000540o`0000000P0D0?l01P00000002So00000000I_l00000002So`80
000001000?oQ00000000oj<00000003oIP0000000?lX0000000Do`000000057o00000000Sol00000
003<o`0000000P3om@0010000000okP00?nh000000050?mj00030000003o?@00ocd000@0ocd000<0
00000?mj003oNP000`3oNP001`000000okP00?nh003o^00000000?oe000000000P3<o`003@000000
Sol00000001Ao`00000001Co000000000?lX00000000ofH00000003oX`0000000P00on401@000000
0>7o00000000Xol00000008006Ko00@00000000Xo`00:?l00000100D0?l00`000000D@3o0540o`0<
0540o`050000000D0?l0503o01@0o`0000000`00:?l01@00000006Ko0000000000000:?o00800000
00H00>7o00000000on400?oQ00000000oj<200000005003oIP0000000000003o:00000000`0Do`00
1@000000DOl0057o001Ao`00000000D0Sol00P00000:0<co00030000002?o`00Sol000@0Sol00P00
0002057o008000000P0Do`02000000800?lX0P0000060?ooo`4000000@3oool00200oooo00<00000
0?ooo`3oool01@3oool00`00obP000000000000201Co008000000`1Ao`000`000000Sol008oo0004
08oo008000002P3<o`02000000@0Sol00P000002057o008000000P0Do`0200000006003o:0000000
0?mV003oIP0000000?nS0P000002003oh@0500000000hOl00>7o00000000Xol00P0000000`00I_l0
0000000Xo`02000Xo`050000000D0?l0503o01@0o`0000003`1A0?l01@000000503o01@0o`0D0?l0
000000<002So00D00000001Vo`0000000:?o00000002003Qo`0500000000on400000003oX`000000
0P00ofH02@0000000?lX0000000Do`000000057o00000000Sol0000000020<co00030000003om@00
000000<0okP000<000000?mj003oNP000P3oNP000`000000ocd00?lm00060?lm00050000003oNP00
ogX00?mj000000000`3o^0002`000000ooD00?oe00000000c?l00000002?o`000000057o00000000
5?l00080000001000?mV00000000oj<00000003oh@0000000>7o00000000Xol00000001Vo`000000
02So0000000D0?l000000P1A0?l010000000S`3o08l0o`0000050<`0o`030000003o0?D0o`3e00<0
o`3e00<000000<`0o`3<0?l00`3<0?l010000000S`3o08l0o`0000020540o`0K0000000D0?l00000
000Xo`00000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000003oIP0000000?lX
0000000Do`000000057o00000000Sol00000003<o`0000000080ooD000@000000?nh003o^0000000
103oNP000`000000ocd00?lm00050?lm00030000003oNP00ogX00080ogX000D000000?nh003o^000
okP0000000020?oe00070000003<o`00000008oo00000000DOl00000000201Co008000000P00ofH0
1`0000000?nS00000000on400000003Qo`0000000P00Xol01000000006Ko001Vo`000002000Xo`05
0000000D0?l0503o01@0o`0000003`1A0?l010000000503o01@0o`0D0?l20000008002So00@00000
001Vo`00I_l000000P00Xol00`0000000>7o00000002003oh@80000000H00?nS0000000000000?mV
00000000obP2000000805?l00P000002057o008000001@2?o`02000000T0c?l00P00000608oo0080
00000P1Ao`02000000805?l00P0000000`00obP0oooo0?ooo`050?ooo`4000000@3oool00200oooo
00<000000?ooo`3oool01@3oool01@0000005?l001Co000Do`00000000@0DOl000<0000008oo002?
o`00102?o`000`000000c?l00<co00080<co008000001@2?o`001@000000DOl0057o001Ao`000000
00<05?l000@00000003o:000obP000000P00ofH00`0000000?nS00000002003oh@80000000D00>7o
0000000000000:?o00000002001Vo`0500000000:?l002So000Xo`0000000`0D0?l00`000000D@3o
0540o`0=0540o`050000000D0?l0503o01@0o`0000000`00:?l00`00000006Ko00000002002So`07
00000000hOl00000003oh@0000000?nS00000002003oIP0900000000obP0000001Co00000000DOl0
0000002?o`0000000080c?l000@000000?oe003om@0000000P3o^0000`000000ogX00?mj00020?mj
00030000003o?@00ocd000L0ocd000D000000?mj003oNP00ogX0000000030?nh00030000003om@00
00000080c?l000H0000008oo00000000DOl00000000Do`020000000@003oIP0000000?nS00000000
on400000003Qo`0000000:?o00000000I_l00000000Xo`000000503o00000080D@3o00@0000008l0
o`2?0?l00000103<0?l00`000000o`3e0?l0m@050?l0m@030000003<0?l0c03o0080c03o00@00000
08l0o`2?0?l000000P1A0?l06`000000503o00000000:?l00000001Vo`0000000:?o00000000hOl0
0000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`00000008oo00000000
c?l0000000020?oe00040000003o^000okP0000000@0ogX000<000000?lm003o?@001P3o?@000`00
0000ogX00?mj00020?mj00050000003o^000okP00?nh000000000P3om@001`000000c?l00000002?
o`000000057o000000000P0Do`0300000008003oIP0000000?nS00000000on400000003Qo`000002
002So`0400000000I_l006Ko0000008002So00D0000001@0o`0D0?l0503o0000000?0540o`030000
000D0?l0503o0080503o00@00000000Xo`00:?l000000P00I_l01@0000000:?o002So`0000000>7o
008000000P00on401P0000000?nS003oX`0000000?mV000000800?lX0P00000201Co00050000001A
o`00DOl0057o000000001P2?o`000`000000c?l00<co00080<co00030000002?o`00Sol000D0Sol0
00D00000057o001Ao`00DOl00000000301Co008000001P3oool100000040oooo000P0?ooo`030000
003oool0oooo00D0oooo0`0Do`0010000000DOl0057o001Ao`02000000H0Sol000<000000<co003<
o`002@3<o`000`000000Sol008oo000308oo00050000001Ao`00DOl0057o000000000`0Do`001000
00000?lX003o:0000002003oIP0600000000oj<00?nS00000000on4000000P00hOl0100000000:?o
002So`000003001Vo`0400000000:?l002So000000<0503o00<000000540o`1A0?l03@1A0?l00`00
0000503o01@0o`0201@0o`0600000000:?l002So00000000I_l000000P00Xol01`0000000>7o0000
0000on400000003oX`0000000P00ofH01`0000000?lX0000000Do`000000057o000000000P2?o`00
0`000000c?l0000000020?oe00040000003o^000okP0000000@0ogX000<000000?lm003o?@00203o
?@001@000000ogX00?mj003oNP00000000<0okP000<000000?oe000000000P3<o`0200000080DOl0
0P0000004P00obP00000003oIP0000000?nS00000000on400000003Qo`0000000:?o00000000I_l0
0000000Xo`000000503o00000080D@3o00@0000008l0o`2?0?l00000103<0?l00`000000o`3e0?l0
m@050?l0m@030000003<0?l0c03o0080c03o00@0000008l0o`2?0?l000000P1A0?l00`000000503o
01@0o`020000000F001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH00000
003o:00000005?l00000001Ao`00000008oo00000000c?l000000080ooD000@000000?nh003o^000
0000103oNP000`000000ocd00?lm00070?lm00030000003oNP00ogX00080ogX000@000000?nh003o
^00000000P3om@000`000000c?l00<co000200000080DOl000L0000001Co000000000?lX00000000
ofH0000000800?nS0P000002003Qo`0600000000Xol00:?o00000000I_l000000`00:?l01@000000
503o01@0o`0D0?l0000000l0D@3o00D0000001@0o`0D0?l0503o00000003000Xo`0400000000I_l0
06Ko000000800:?o00H00000003Qo`00hOl00000003oh@000002003oX`0400000000ofH00?mV0000
00<00?lX00@0000001Co000Do`0000000`1Ao`000`000000Sol008oo000408oo00030000003<o`00
c?l000P0c?l000<0000008oo002?o`001@2?o`001@000000DOl0057o001Ao`00000000<05?l01`3o
ool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00<05?l00000000000000`1Ao`00
0`000000Sol008oo000508oo00030000003<o`00c?l000P0c?l00P00000508oo00030000001Ao`00
DOl0008000000P0Do`02000000800?lX00@00000003oIP00ofH000000P00oj<00`0000000?oQ0000
0002003Qo`0500000000Xol00:?o002So`0000000P00I_l01000000002So000Xo`00000301@0o`03
0000001A0?l0D@3o00h0D@3o00D0000001@0o`0D0?l0503o00000002000Xo`0300000000I_l00000
00800:?o00X00000003Qo`00hOl00000003oh@0000000?nS00000000ofH000000P00obP200000080
DOl000D0000008oo00000000c?l0000000030?oe00040000003o^000okP0000000<0ogX000<00000
0?lm003o?@00203o?@001@000000ogX00?mj003oNP00000000<0okP000<000000?oe000000000P3<
o`002P000000Sol00000001Ao`00000001Co000000000?lX00000000ofH200000008003oh@000000
0>7o00000000Xol00000001Vo`000002000Xo`8000000P1A0?l010000000S`3o08l0o`0000040<`0
o`030000003o0?D0o`3e00L0o`3e00<000000<`0o`3<0?l00P3<0?l00`000000S`3o000000020540
o`030000000D0?l0503o0080000001H006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<0
0000003oIP0000000?lX0000000Do`000000057o00000000Sol00000003<o`0000000P3om@001000
0000okP00?nh000000040?mj00030000003o?@00ocd000L0ocd000<000000?mj003oNP000P3oNP00
1@000000okP00?nh003o^00000000080ooD000<000000<co003<o`000P0000003`1Ao`00000001Co
000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o00000002002So`040000
0000I_l006Ko0000008002So00<0000001@0o`0D0?l00P0D0?l00`000000D@3o0540o`0=0540o`05
0000000D0?l0503o01@0o`0000000`00:?l01000000006Ko001Vo`000002002So`0600000000hOl0
0>7o00000000on4000000P00oj<0100000000?mV003oIP000003003o:0030000000Do`005?l00080
00000P1Ao`02000000D0Sol00P0000090<co00030000002?o`00Sol000D0Sol000@00000057o001A
o`00DOl00P0000000`0Do`00oooo0?ooo`050?ooo`4000000@3oool00200oooo00<000000?ooo`3o
ool01@3oool00`000000DOl0057o0002057o00030000002?o`00Sol000D0Sol000<000000<co003<
o`001`3<o`02000000H0Sol000@00000057o001Ao`0000000`0Do`001P0000000?lX003o:000obP0
0000003oIP80000000D00?nS0000000000000?oQ00000002003Qo`0300000000Xol00:?o00800000
00<006Ko0000000000000P00:?l01@000000503o01@0o`0D0?l000000100D@3o00<0000001@0o`0D
0?l00P0D0?l00`00000002So00000002001Vo`0:00000000Xol00:?o00000000hOl00000003oh@00
00000?nS000000800?mV00L00000003o:00000005?l00000001Ao`0000000080Sol000<000000<co
000000000P3om@0010000000okP00?nh000000040?mj00030000003o?@00ocd000P0ocd000D00000
0?mj003oNP00ogX0000000030?nh00030000003om@0000000080c?l000P0000008oo00000000DOl0
0000000Do`000000003o:080000000h00?nS00000000on400000003Qo`0000000:?o00000000I_l0
0000000Xo`000000503o00000080D@3o00@0000008l0o`2?0?l00000103<0?l00`000000o`3e0?l0
m@070?l0m@030000003<0?l0c03o0080c03o00@0000008l0o`2?0?l000000P1A0?l040000000503o
00000000:?l00000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH20000
00805?l00P00000208oo00030000003<o`0000000080ooD000@000000?nh003o^00000000`3oNP00
0`000000ocd00?lm00080?lm00030000003oNP00ogX00080ogX000@000000?nh003o^00000000P3o
m@000`000000c?l00<co000200000080DOl000D0000001Co000000000?lX00000002003oIP070000
0000oj<00000003oh@0000000>7o00000002002So`0400000000I_l006Ko0000008002So00D00000
01@0o`0D0?l0503o0000000@0540o`040000000D0?l0503o01@0o`8000000P00:?l00`00000006Ko
001Vo`0200000006002So`0000000000003Qo`0000000?oQ0P0000001@00oj<0000000000000ofH0
000000<00?lX00D0000001Co000Do`005?l000000003057o00030000002?o`00Sol000@0Sol000<0
00000<co003<o`001`3<o`000`000000Sol008oo000508oo008000000`1Ao`02000000H0oooo0@00
00010?ooo`00803oool00`000000oooo0?ooo`050?ooo`<0DOl00P00000708oo00030000003<o`00
c?l000H0c?l00P00000608oo008000000P1Ao`001@0000005?l001Co000Do`00000000800?lX0P00
00001@00ofH00000003oX`00oj<0000000800?oQ00@00000003Qo`00hOl000000P00Xol010000000
06Ko001Vo`000003000Xo`050000000D0?l0503o01@0o`000000401A0?l010000000503o01@0o`0D
0?l20000008002So00<00000001Vo`0000000P00Xol00`0000000>7o00000002003oh@0:00000000
oj<00000003oIP0000000?lX0000000Do`000000057o0080000000@0Sol00000003<o`0000000P3o
m@02000000030?nh00000000ogX000<0ogX000<000000?lm003o?@002@3o?@001@000000ogX00?mj
003oNP00000000<0okP000<000000?oe000000000P3<o`0020000000Sol00000001Ao`00000001Co
000000000?lX0P0000003P00oj<00000003oh@0000000>7o00000000Xol00000001Vo`00000002So
0000000D0?l000000P1A0?l010000000S`3o08l0o`0000040<`0o`030000003o0?D0o`3e00P0o`3e
00D000000<`0o`3<0?l0c03o0000000208l0o`0N0000001A0?l0D@3o0000000D0?l00000000Xo`00
000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000003oIP0000000?lX0000000D
o`000000057o00000000Sol00000003<o`0000000P3om@0010000000okP00?nh000000030?mj0003
0000003o?@00ocd000T0ocd000<000000?mj003oNP000P3oNP0010000000okP00?nh000000020?oe
00030000003<o`00c?l000<0000000H0DOl00000000Do`000000003o:0000002003oIP0700000000
oj<00000003oh@0000000>7o00000002002So`0400000000I_l006Ko0000008002So00D0000001@0
o`0D0?l0503o0000000@0540o`030000000D0?l0503o0080503o00@00000000Xo`00:?l000000`00
I_l01P0000000:?o002So`0000000>7o000000800?oQ00D00000003oX`00oj<00000003oIP020000
00800?lX00D0000001Co000Do`005?l000000003057o008000001@2?o`02000000P0c?l000<00000
08oo002?o`001P2?o`0010000000DOl0057o001Ao`070?ooo`4000000@3oool00200oooo00<00000
0?ooo`3oool01@3oool2057o00030000002?o`00Sol000H0Sol000<0000008oo000000001@3<o`02
000000L0Sol000D00000057o001Ao`00DOl00000000201Co008000000P00obP0100000000?mV003o
IP000002003oX`0400000000on400?oQ000000800>7o00@00000002So`00Xol000000`00I_l01000
000002So000Xo`00000301@0o`030000001A0?l0D@3o00h0D@3o00D0000001@0o`0D0?l0503o0000
0003000Xo`0300000000I_l0000000800:?o00<00000003Qo`0000000P00on402`0000000?nS0000
0000ofH00000003o:00000005?l00000001Ao`0000000080Sol000H000000<co003<o`0000000?oe
000000020?nh00030000003oNP00ogX00080ogX000<000000?lm003o?@002P3o?@001@000000ogX0
0?mj003oNP0000000080okP000@000000?oe003om@0000000P3<o`0020000000Sol00000001Ao`00
000001Co000000000?lX0P0000003P00oj<00000003oh@0000000>7o00000000Xol00000001Vo`00
000002So0000000D0?l000000P1A0?l010000000S`3o08l0o`0000030<`0o`030000003o0?D0o`3e
00T0o`3e00D000000<`0o`3<0?l0c03o0000000208l0o`0N0000001A0?l0D@3o0000000D0?l00000
000Xo`00000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000003oIP0000000?lX
0000000Do`000000057o00000000Sol00000003<o`0000000P3om@0010000000okP00?nh00000003
0?mj00030000003o?@00ocd000X0ocd000<000000?mj003oNP000P3oNP0010000000okP00?nh0000
00020?oe00030000003<o`0000000080Sol00P00000201Co000700000000obP00000003oIP000000
0?nS00000002003oh@0500000000hOl00000002So`0000000`00I_l01000000002So000Xo`000003
01@0o`030000001A0?l0D@3o00h0D@3o00<0000001@0o`0D0?l00P0D0?l01000000002So000Xo`00
0002001Vo`0400000000Xol00:?o000000800>7o00@00000003oh@00on4000000P00oj<010000000
0?mV003oIP000002003o:08000000P0Do`02000000<0DOl000<0000008oo002?o`00102?o`020000
00L0c?l000<0000008oo002?o`001P2?o`000`000000DOl0057o00070?ooo`4000000@3oool00200
oooo00<000000?ooo`3oool01@3oool00`1Ao`00000008oo000708oo00050000003<o`00000008oo
000000000P3<o`02000000P0Sol000D00000057o001Ao`00DOl00000000201Co000500000000obP0
0?lX003o:00000000P00ofH0100000000?nS000000000002003oh@0400000000hOl00>7o00000080
0:?o00D00000001Vo`00I_l006Ko00000002000Xo`050000000D0?l0503o01@0o`0000004@1A0?l0
1@000000503o01@0o`0D0?l0000000<002So00D00000001Vo`0000000:?o00000002003Qo`090000
0000on400000003oX`0000000?mV00000000obP0000000805?l000<00000057o000000000P2?o`00
1@000000c?l00000003om@00000000<0okP000D000000?mj003oNP00ogX00000000=0?lm00050000
003oNP00ogX00?mj000000000P3o^0000`000000ooD0000000020<co00080000002?o`000000057o
000000005?l000000000obP20000000<003oX`0000000?oQ00000000hOl00000002So`00000006Ko
00000000:?l000000P0D0?l01P000000D@3o0540o`000000S`3o000000@0c03o00<000000?l0m@3o
0?D02@3o0?D01@000000c03o0<`0o`3<0?l0000000<0S`3o01d000000540o`000000503o00000000
:?l00000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH00000003o:000
00005?l00000001Ao`00000008oo00000000c?l0000000020?oe00040000003o^000okP0000000<0
ogX000<000000?lm003o?@002P3o?@000`000000ogX00?mj00020?mj00040000003o^000okP00000
00<0ooD000T000000<co00000000Sol00000001Ao`00000001Co000000000P00obP02@0000000?mV
00000000oj<00000003oh@0000000>7o00000002002So`0400000000I_l006Ko000000<002So00@0
000001@0o`0D0?l000004@1A0?l00`000000503o01@0o`0201@0o`0400000000:?l002So00000080
06Ko00@00000002So`00Xol000000P00hOl00`0000000?oQ003oh@0200000005003oX`0000000?mV
003oIP0000000`00obP01@0000005?l001Co000Do`00000000<0DOl000<0000008oo002?o`001@2?
o`02000000H0c?l000<0000008oo002?o`001P2?o`02000000L0oooo0@0000010?ooo`00803oool2
000000H0oooo00<0000008oo002?o`001`2?o`000`000000c?l00000000208oo00800000202?o`02
00000080DOl00P00000201Co000300000000obP00?lX0080000000D00?mV0000000000000?nS0000
0003003oh@0500000000hOl0000000000000Xol00P000002001Vo`8000000P00:?l01@000000503o
01@0o`0D0?l000000140D@3o00D0000001@0o`0D0?l0503o00000003000Xo`0300000000I_l00000
00800:?o00<00000003Qo`0000000P00on402`0000000?nS00000000ofH00000003o:00000005?l0
0000001Ao`0000000080Sol000H000000<co003<o`0000000?oe000000020?nh00030000003oNP00
ogX00080ogX000<000000?lm003o?@002`3o?@001`000000ogX00?mj003oNP0000000?nh00000000
0P3om@002`000000c?l00<co00000000Sol00000001Ao`00000001Co000000000?lX0080000000`0
0?nS00000000on400000003Qo`0000000:?o00000000I_l00000000Xo`00000201@0o`030000001A
0?l000000080S`3o00D000000<`0o`3<0?l0c03o0000000=0?l0m@050000003<0?l0c03o0<`0o`00
00000P2?0?l07P000000D@3o0540o`000000503o00000000:?l00000001Vo`0000000:?o00000000
hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`00000008oo0000
0000c?l000000080ooD000<000000?nh00000000103oNP000`000000ocd00?lm000:0?lm00030000
003oNP00ogX00080ogX000@000000?nh003o^00000000P3om@000`000000c?l00000000208oo0009
0000001Ao`00000001Co000000000?lX00000000ofH0000000800?nS00D00000003oh@0000000>7o
00000002002So`0400000000I_l006Ko000000<002So00@0000001@0o`0D0?l000004@1A0?l00`00
0000503o01@0o`0201@0o`0400000000:?l002So0000008006Ko0P0000001@00Xol0000000000000
hOl0000000<00?oQ00D00000003oX`0000000000003oIP02000000800?lX0P00000201Co00040000
001Ao`00DOl0057o008000001`2?o`02000000@0c?l000<0000008oo002?o`00202?o`000`000000
oooo0?ooo`030?ooo`8000000@3oool00200oooo00<000000?ooo`3oool01@3oool908oo00050000
003<o`00c?l00<co000000002@2?o`02000000<0DOl000D0000001Co000Do`005?l000000002003o
:00400000000ofH00?mV000000800?nS00<00000003oh@00on400P0000001@00hOl00000002So`00
Xol0000000<006Ko00<00000000Xo`00:?l00P00000301@0o`030000001A0?l0D@3o00l0D@3o00D0
000001@0o`0D0?l0503o00000002000Xo`80000000L006Ko00000000Xol00:?o00000000hOl00000
00800?oQ00T00000003oX`0000000?mV00000000obP0000001Co000000000P1Ao`000`000000Sol0
000000020<co00030000003om@00000000<0okP000D000000?mj003oNP00ogX00000000>0?lm0004
0000003oNP00ogX000000080okP000D000000?oe003om@0000000<co00020000000708oo00000000
DOl00000000Do`000000003o:0020000000<003oX`0000000?oQ00000000hOl00000002So`000000
06Ko00000000:?l000000P0D0?l00`000000D@3o0000000208l0o`050000003<0?l0c03o0<`0o`00
00003@3o0?D01@000000c03o0<`0o`3<0?l000000080S`3o01h000000540o`1A0?l0000001@0o`00
000002So00000000I_l00000002So`0000000>7o00000000on400000003oX`0000000?mV00000000
obP0000001Co00000000DOl00000002?o`0000000<co000000020?oe00040000003o^000okP00000
00<0ogX000<000000?lm003o?@002`3o?@000`000000ogX00?mj00020?mj00040000003o^000okP0
00000080ooD000<000000<co000000000P2?o`002@000000DOl00000000Do`000000003o:0000000
0?mV00000002003oX`0500000000on400000003Qo`0000000P00Xol01000000006Ko001Vo`000002
000Xo`040000000D0?l0503o01@0o`800000401A0?l00`000000503o01@0o`0201@0o`0300000000
:?l002So008000000P00I_l01@0000000:?o002So`0000000>7o008000000P00on40100000000?nS
003oX`000002003oIP0500000000obP00?lX003o:00000000P0Do`000`000000DOl0057o0002057o
00030000002?o`00Sol000H0Sol00P0000030<co00030000002?o`00Sol000L0Sol01`3oool10000
0040oooo000P0?ooo`030000003oool0oooo00D0oooo2P2?o`03000000T0Sol000<00000057o001A
o`000P1Ao`001@0000005?l001Co000Do`00000000800?lX00@00000003oIP00ofH000000P00oj<0
100000000?oQ003oh@000002003Qo`0500000000Xol00:?o002So`0000000P00I_l01000000002So
000Xo`00000501@0o`030000001A0?l0D@3o00h0D@3o00D0000001@0o`0D0?l0503o00000002000X
o`0400000000I_l006Ko000000800:?o00<00000003Qo`0000000P00on402@0000000?nS00000000
ofH00000003o:00000005?l000000002057o00030000002?o`0000000080c?l000@000000?oe003o
m@0000000P3o^0001@000000ogX00?mj003oNP00000000h0ocd000D000000?mj003oNP00ogX00000
00020?nh000:0000003om@00ooD00000003<o`00000008oo00000000DOl0000000805?l00P000000
3P00ofH00000003oX`0000000?oQ00000000hOl00000002So`00000006Ko00000000:?l000000P0D
0?l00`000000D@3o0000000208l0o`050000003<0?l0c03o0<`0o`0000003@3o0?D01@000000c03o
0<`0o`3<0?l000000080S`3o01h000000540o`1A0?l0000001@0o`00000002So00000000I_l00000
002So`0000000>7o00000000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl0
0000002?o`0000000<co000000020?oe00040000003o^000okP0000000<0ogX000<000000?lm003o
?@00303o?@001@000000ogX00?mj003oNP00000000<0okP000<000000?oe000000000P3<o`000`00
0000Sol000000002057o00090000000Do`000000003o:00000000?mV00000000oj<0000000800?oQ
00<00000003Qo`0000000P00Xol01000000006Ko001Vo`000002000Xo`030000000D0?l0503o0080
503o00<000000540o`1A0?l03P1A0?l00`000000503o01@0o`0201@0o`0700000000:?l002So000X
o`00000006Ko00000003002So`0400000000hOl00>7o000000800?oQ00@00000003oX`00oj<00000
0P00ofH01@0000000?lX003o:000obP0000000805?l00P000003057o00030000002?o`00Sol000L0
Sol00`00000:08oo00L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oooa@0Sol0
0P000004057o00060000000Do`005?l001Co000000000?lX0P0000000`00ofH0000000000002003o
X`0400000000on400?oQ000000800>7o00H00000002So`00Xol00:?o00000000I_l20000008002So
00<0000001@0o`0D0?l00`0D0?l00`000000D@3o0540o`0>0540o`050000000D0?l0503o01@0o`00
00000`00:?l02000000006Ko001Vo`0000000:?o00000000hOl000000P00on400`0000000?nS0000
0002003oIP8000000P0Do`001@000000DOl00000002?o`00000000<0c?l000<000000?oe00000000
0P3o^0001@000000ogX00?mj003oNP00000000l0ocd000@000000?mj003oNP0000000P3o^0001P00
0000ooD00?oe00000000c?l000000080Sol000H00000057o000000005?l000000000obP200000080
0?nS0P000002003Qo`8000000P00I_l200000080503o00<000000540o`0000000P2?0?l01@000000
c03o0<`0o`3<0?l0000000h0o`3e00D000000<`0o`3<0?l0c03o0000000208l0o`0N0000001A0?l0
D@3o0000000D0?l00000000Xo`00000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<0
0000003oIP0000000?lX0000000Do`000000057o00000000Sol00000003<o`0000000P3om@001000
0000okP00?nh000000030?mj00030000003o?@00ocd000d0ocd000D000000?mj003oNP00ogX00000
00030?nh00050000003om@0000000<co000000000P2?o`001`000000DOl00000000Do`000000003o
:00000000P00ofH01@0000000?nS00000000on40000000800>7o00@00000002So`00Xol000000P00
I_l00`00000002So0000000401@0o`030000001A0?l0D@3o00l0D@3o00<0000001@0o`0D0?l00P0D
0?l01`00000002So000Xo`00:?l00000001Vo`0000000`00Xol0100000000>7o003Qo`000002003o
h@0300000000oj<00?nS0080000000<00?mV0000000000000P00obP01@0000005?l001Co000Do`00
000000<0DOl00P00000D08oo00L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oo
oa<0Sol000<00000057o001Ao`000P1Ao`02000000805?l00P0000001@00obP00000003oIP00ofH0
000000<00?nS00D00000003oh@0000000000003Qo`02000000800:?o0P0000000`00I_l00000000X
o`02000Xo`030000000D0?l0503o0080503o0P00000@0540o`030000000D0?l0503o0080503o00@0
0000000Xo`00:?l000000P00I_l0200000000:?o002So`0000000>7o00000000on4000000P00oj<0
1`0000000?mV00000000obP0000001Co000000000P1Ao`000`000000Sol0000000020<co00040000
003om@00ooD000000080okP000D000000?mj003oNP00ogX00000000>0?lm00050000003oNP00ogX0
0?mj000000000P3o^00020000000ooD00?oe00000000c?l00000002?o`0000000P1Ao`004P000000
5?l00000000000000?mV00000000oj<00000003oh@0000000>7o00000000Xol00000001Vo`000000
02So00000080503o00<000000540o`0000000P2?0?l01@000000c03o0<`0o`3<0?l0000000l0o`3e
00D000000<`0o`3<0?l0c03o0000000208l0o`0N0000001A0?l0D@3o0000000D0?l00000000Xo`00
000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000003oIP0000000?lX0000000D
o`000000057o00000000Sol00000003<o`0000000P3om@0010000000okP00?nh000000030?mj0003
0000003o?@00ocd000`0ocd000D000000?mj003oNP00ogX0000000030?nh00030000003om@000000
0080c?l000<0000008oo000000000P1Ao`001@0000005?l000000000obP0000000800?mV00D00000
003oX`0000000?oQ00000002003Qo`0300000000Xol0000000<006Ko00<00000000Xo`000000100D
0?l00`000000D@3o0540o`0?0540o`030000000D0?l0503o0080503o00H00000000Xo`00:?l002So
00000000I_l2000000800:?o0P0000001@00hOl0000000000000on40000000<00?nS00<00000003o
IP00ofH00P0000000`00obP000000000000201Co00030000001Ao`00DOl00080DOl00P00000B08oo
00L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oooa40Sol00P000003057o0080
00000P0Do`02000000800?lX00@00000003oIP00ofH000000P00oj<200000005003oh@0000000>7o
003Qo`0000000`00Xol01000000006Ko001Vo`000003000Xo`030000000D0?l0503o0080503o00<0
00000540o`1A0?l03`1A0?l010000000503o01@0o`0D0?l20000008002So00@00000001Vo`00I_l0
00000P00Xol01@0000000>7o00000000on40000000800?nS00L00000003oIP0000000?lX0000000D
o`0000000080DOl000<0000008oo000000000P3<o`0010000000ooD00?oe000000020?nh00050000
003oNP00ogX00?mj000000003`3o?@001@000000ogX00?mj003oNP0000000080okP000P000000?oe
003om@0000000<co00000000Sol000000080DOl00180000001Co000000000000003oIP0000000?nS
00000000on400000003Qo`0000000:?o00000000I_l00000000Xo`00000201@0o`030000001A0?l0
00000080S`3o00D000000<`0o`3<0?l0c03o0000000?0?l0m@050000003<0?l0c03o0<`0o`000000
0P2?0?l07P000000D@3o0540o`000000503o00000000:?l00000001Vo`0000000:?o00000000hOl0
0000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`00000008oo00000000
c?l000000080ooD000@000000?nh003o^00000000`3oNP000`000000ocd00?lm000=0?lm00050000
003oNP00ogX00?mj000000000`3o^0000`000000ooD0000000020<co00030000002?o`0000000080
DOl000D0000001Co000000000?lX00000002003oIP0500000000oj<00000003oh@0000000P00hOl0
0`0000000:?o00000002001Vo`0300000000:?l002So008000000`0D0?l00`000000D@3o0540o`0?
0540o`030000000D0?l0503o0080503o00D00000000Xo`00:?l002So00000002001Vo`8000000P00
Xol01@0000000>7o003Qo`0000000?oQ008000000P00oj<01@0000000?mV003oIP00ofH000000080
0?lX00<0000001Co000Do`000P000004057o00800000402?o`070?ooo`4000000@3oool00200oooo
00<000000?ooo`3oool01@3oool?08oo008000000`1Ao`02000000805?l00P000002003o:0800000
0P00ofH0100000000?nS000000000002003oh@0400000000hOl00>7o000000<00:?o00@00000001V
o`00I_l000000`00:?l00`000000503o01@0o`0301@0o`030000001A0?l0D@3o00h0D@3o00D00000
01@0o`0D0?l0503o00000003000Xo`0400000000I_l006Ko000000800:?o00D00000003Qo`000000
0?oQ00000002003oX`0700000000ofH00000003o:00000005?l000000002057o00030000002?o`00
00000080c?l000@000000?oe003om@0000000P3o^0001@000000ogX00?mj003oNP0000000100ocd0
00D000000?mj003oNP00ogX0000000020?nh000L0000003om@00ooD00000003<o`00000008oo0000
0000DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o00000000
Xol00000001Vo`00000002So00000080503o00<000000540o`0000000P2?0?l01@000000c03o0<`0
o`3<0?l0000000l0o`3e00D000000<`0o`3<0?l0c03o0000000208l0o`0L0000001A0?l0D@3o0000
000D0?l00000000Xo`00000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000003o
IP0000000?lX0000000Do`000000057o00000000Sol000000080c?l000<000000?oe000000000P3o
^0001@000000ogX00?mj003oNP0000000100ocd000D000000?mj003oNP00ogX0000000020?nh0004
0000003om@00ooD000000080c?l000<0000008oo000000000P1Ao`001`0000005?l000000000obP0
0000003oIP0000000P00oj<00`0000000?oQ00000002003Qo`0600000000Xol0000000000000I_l0
00000`00:?l01@000000503o01@0o`0D0?l000000140D@3o00<0000001@0o`0D0?l00P0D0?l01@00
000002So000Xo`00:?l0000000<006Ko00@00000002So`00Xol000000P00hOl00`0000000?oQ003o
h@0200000003003oX`000000000000800?mV0P0000000`00obP000000000000201Co00800000101A
o`000`000000Sol008oo000=08oo00L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`05
0?ooo`d0Sol00P000004057o00050000000Do`005?l001Co000000000`00obP01`0000000?mV003o
IP00ofH00000003oX`0000000`00on40100000000>7o000000000002002So`8000000P00I_l01@00
000002So000Xo`00:?l0000000D0503o00<000000540o`1A0?l03P1A0?l01@000000503o01@0o`0D
0?l0000000<002So00@00000001Vo`00I_l000000P00Xol01@0000000>7o00000000on4000000080
0?nS00L00000003oIP0000000?lX0000000Do`0000000080DOl000<0000008oo000000000P3<o`00
10000000ooD00?oe000000030?nh00050000003oNP00ogX00?mj000000003`3o?@001@000000ogX0
0?mj003oNP0000000080okP000@000000?oe003om@00ooD00P00000208oo00040000001Ao`000000
01Co0080000000h00?mV00000000oj<00000003oh@0000000>7o00000000Xol00000001Vo`000000
02So00000080503o00<000000540o`0000000P2?0?l01@000000c03o0<`0o`3<0?l000000100o`3e
00D000000<`0o`3<0?l0c03o0000000208l0o`0H0000001A0?l0D@3o0000000D0?l00000000Xo`00
000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000003oIP0000000?lX0000000D
o`0000000P1Ao`0200000080c?l000<000000?oe000000000P3o^0001@000000ogX00?mj003oNP00
00000140ocd000D000000?mj003oNP00ogX0000000020?nh00030000003om@0000000080c?l000H0
000008oo002?o`000000057o0000000201Co000500000000obP00000003oIP0000000P00oj<00`00
00000?oQ00000002003Qo`0600000000Xol00:?o00000000I_l000000P00:?l00`000000503o01@0
o`0201@0o`030000001A0?l0D@3o00l0D@3o00<0000001@0o`0D0?l00P0D0?l01@00000002So000X
o`00:?l0000000<006Ko00<00000002So`00Xol00P0000000`00hOl00000003oh@02003oh@040000
0000oj<00?nS000000<00?mV00@00000003o:000obP000000`0Do`000`000000DOl0057o0002057o
008000003@2?o`070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool;08oo0080
0000101Ao`02000000805?l00P000002003o:08000000P00ofH200000004003oX`0000000?oQ003o
h@80000000<00>7o00000000Xol00P00Xol01@00000006Ko001Vo`00I_l0000000<002So00<00000
01@0o`0D0?l00P0D0?l200000100D@3o00<0000001@0o`0D0?l00P0D0?l01000000002So000Xo`00
0002001Vo`0400000000Xol00:?o000000800>7o0P000002003oX`0300000000ofH0000000800?lX
00D0000001Co00000000DOl00000000208oo00060000003<o`00c?l00000003om@0000000`3o^000
1@000000ogX00?mj003oNP0000000100ocd000D000000?mj003oNP00ogX0000000020?nh00060000
003om@00ooD00000003<o`0000000P2?o`020000000B01Co000000000?lX00000000ofH00000003o
X`0000000?oQ00000000hOl00000002So`00000006Ko00000000:?l000000P0D0?l00`000000D@3o
0000000208l0o`050000003<0?l0c03o0<`0o`0000004@3o0?D01@000000c03o0<`0o`3<0?l00000
0080S`3o01`000000540o`1A0?l0000001@0o`00000002So00000000I_l00000002So`0000000>7o
00000000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`000000
0P3<o`000`000000ooD0000000020?nh00050000003oNP00ogX00?mj00000000403o?@001@000000
ogX00?mj003oNP00000000<0okP000<000000?oe000000000P3<o`0020000000Sol008oo00000000
DOl00000000Do`0000000P00obP01@0000000?mV00000000oj<0000000800?oQ00<00000003Qo`00
00000P00Xol01000000006Ko001Vo`000002000Xo`030000000D0?l0503o0080503o00<000000540
o`1A0?l03`1A0?l00`000000503o01@0o`0201@0o`0500000000:?l002So000Xo`0000000`00I_l0
1P0000000:?o002So`00Xol00000003Qo`8000000P00on400`0000000?nS003oX`02000000800?mV
00<00000003o:000obP00P00000201Co00800000101Ao`02000000/0Sol01`3oool100000040oooo
000P0?ooo`030000003oool0oooo00D0oooo2@2?o`02000000@0DOl00P00000301Co000500000000
obP00?lX003o:00000000P00ofH2000000800?nS00@00000003oh@00on4000000P00hOl01@000000
0:?o002So`00Xol00000008006Ko0P000003000Xo`030000000D0?l0503o0080503o00<000000540
o`1A0?l03`1A0?l00`000000503o01@0o`0201@0o`0400000000:?l002So0000008006Ko00@00000
002So`00Xol000000P00hOl01@0000000?oQ00000000oj<0000000800?mV00D00000003o:0000000
5?l000000002057o00030000002?o`0000000080c?l000@000000?oe003om@0000000P3o^0001@00
0000ogX00?mj003oNP0000000140ocd000D000000?mj003oNP00ogX0000000020?nh000L0000003o
m@00ooD00000003<o`00000008oo00000000DOl00000000Do`000000003o:00000000?mV00000000
oj<00000003oh@0000000>7o00000000Xol00000001Vo`00000002So00000080503o00<000000540
o`0000000P2?0?l01@000000c03o0<`0o`3<0?l0000000H0o`3e1@0000060?l0m@050000003<0?l0
c03o0<`0o`0000000P2?0?l070000000D@3o0540o`000000503o00000000:?l00000001Vo`000000
0:?o00000000hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`00
000008oo000000020<co00030000003om@0000000080okP000D000000?mj003oNP00ogX00000000A
0?lm00050000003oNP00ogX00?mj000000000`3o^0000`000000ooD0000000020<co00080000002?
o`00Sol00000001Ao`00000001Co00000002003o:00500000000ofH00000003oX`0000000P00on40
0`0000000>7o00000002002So`0400000000I_l006Ko0000008002So0P00000301@0o`030000001A
0?l0D@3o00l0D@3o00<0000001@0o`0D0?l00P0D0?l20000008002So0P000002001Vo`0500000000
Xol00:?o002So`0000000P00hOl00`0000000?oQ003oh@02000000800?nS0P0000000`00ofH00000
003o:002003o:0040000000Do`005?l001Co00800000101Ao`02000000T0Sol01`3oool100000040
oooo000P0?ooo`030000003oool0oooo00D0oooo1`2?o`02000000D0DOl000<0000001Co000Do`00
0P0Do`001@0000000?lX003o:000obP0000000800?mV00H00000003oX`00oj<00?nS00000000on42
00000003003Qo`000000000000800:?o0P000002001Vo`0400000000:?l002So000Xo`800000100D
0?l00`000000D@3o0540o`0>0540o`800000100D0?l01000000002So000Xo`000002001Vo`040000
0000Xol00:?o000000800>7o00D00000003oh@0000000?nS00000002003oIP0500000000obP00000
01Co000000000P1Ao`000`000000Sol0000000020<co00040000003om@00ooD000000080okP000D0
00000?mj003oNP00ogX00000000A0?lm00050000003oNP00ogX00?mj000000000P3o^00010000000
ooD00?oe000000020<co000G0000002?o`000000057o000000005?l000000000obP00000003oIP00
00000?nS00000000on400000003Qo`0000000:?o00000000I_l00000000Xo`0000000P0D0?l00`00
0000D@3o0000000208l0o`040000003<0?l0c03o000000H0o`3e00<000000?l0^03o0;P00`3o0;P0
0`000000o`3e0?l0m@030?l0m@050000003<0?l0c03o0<`0o`0000000P2?0?l070000000D@3o0540
o`000000503o00000000:?l00000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS0000
0000ofH00000003o:00000005?l00000001Ao`00000008oo000000020<co00030000003om@000000
0080okP000D000000?mj003oNP00ogX00000000B0?lm00050000003oNP00ogX00?mj000000000`3o
^0000`000000ooD0000000020<co00080000002?o`00Sol00000001Ao`00000001Co00000002003o
:00500000000ofH00000003oX`0000000P00on401P0000000>7o0000000000000:?o0000008006Ko
00D00000000Xo`00:?l002So0000000301@0o`800000401A0?l00`000000503o01@0o`0301@0o`05
00000000:?l002So000Xo`0000000P00I_l2000000800:?o0P0000000`00hOl0000000000002003o
h@0700000000oj<00?nS003oX`0000000?mV00000003003o:0030000000Do`005?l000805?l000<0
0000057o001Ao`000`1Ao`02000000L0Sol01`3oool100000040oooo000P0?ooo`030000003oool0
oooo00D0oooo1@2?o`02000000D0DOl00P00000301Co008000000P00obP200000003003oIP000000
000000800?nS0P0000001@00on400000003Qo`00hOl0000000<00:?o00D00000001Vo`00I_l006Ko
00000003000Xo`030000000D0?l0503o0080503o0P00000@0540o`030000000D0?l0503o0080503o
0P000002000Xo`0400000000I_l006Ko000000800:?o00@00000003Qo`00hOl000000P00on420000
00800?mV00D00000003o:00000005?l000000002057o00030000002?o`0000000080c?l000@00000
0?oe003om@0000000`3o^0001@000000ogX00?mj003oNP0000000140ocd000D000000?mj003oNP00
ogX0000000020?nh000F0000003om@00ooD00000003<o`00000008oo00000000DOl00000000Do`00
0000003o:00000000?mV00000000oj<00000003oh@0000000>7o000000800:?o0P000002000Xo`05
0000000D0?l000000540o`0000000`2?0?l010000000c03o0<`0o`0000060?l0m@030000003o0;P0
o`2h00<0o`2h00<000000?l0m@3o0?D0103o0?D01`000000c03o0<`0o`3<0?l0000008l0o`000000
0P1A0?l05`000000503o00000000:?l00000001Vo`0000000:?o00000000hOl00000003oh@000000
0?nS00000000ofH00000003o:00000005?l00000001Ao`0000000080Sol000<000000<co00000000
0P3om@0010000000okP00?nh000000030?mj00030000003o?@00ocd00100ocd000@000000?mj003o
NP0000000`3o^00010000000ooD00?oe000000020<co00050000002?o`000000057o000000000P0D
o`000`0000000?lX00000002003oIP0500000000oj<00000003oh@0000000`00hOl00`0000000:?o
00000002001Vo`0500000000:?l002So000Xo`000000100D0?l00`000000D@3o0540o`0>0540o`80
0000100D0?l01@00000002So000Xo`00:?l0000000<006Ko00D00000002So`00Xol00:?o00000002
003Qo`80000000<00?oQ0000000000000P00oj<0100000000?mV000000000002003o:08000000`0D
o`02000000D0DOl00P00000508oo00L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`05
0?ooo`<0Sol00P000006057o00040000000Do`005?l001Co008000000P00obP2000000800?mV00D0
0000003oX`00oj<00?nS00000002003oh@0400000000hOl00>7o000000<00:?o00<00000001Vo`00
I_l00P000003000Xo`030000000D0?l0503o0080503o00<000000540o`1A0?l03`1A0?l00`000000
503o01@0o`0201@0o`0500000000:?l002So000Xo`0000000P00I_l0100000000:?o002So`000002
003Qo`0:00000000on400?oQ00000000oj<00000003oIP0000000?lX000000805?l000<00000057o
000000000P2?o`001P000000c?l00<co00000000ooD0000000<0okP000D000000?mj003oNP00ogX0
0000000A0?lm00050000003oNP00ogX00?mj000000000P3o^00010000000ooD00?oe000000020<co
000G0000002?o`000000057o000000005?l000000000obP00000003oIP0000000?nS00000000on40
0000003Qo`0000000:?o00000000I_l00000000Xo`0000000P0D0?l00`000000D@3o0000000308l0
o`040000003<0?l0c03o000000D0o`3e00<000000?l0^03o0;P01@3o0;P00`000000o`3e0?l0m@03
0?l0m@050000003<0?l0c03o0<`0o`0000000P2?0?l01@000000D@3o0000000D0?l00000008002So
01D00000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH00000003o:000
00005?l00000001Ao`00000008oo000000000P3<o`000`000000ooD0000000020?nh00050000003o
NP00ogX00?mj000000004`3o?@0010000000ogX00?mj000000030?nh00060000003om@00ooD00000
003<o`0000000P2?o`000`000000DOl00000000201Co000500000000obP00000003oIP0000000P00
oj<00`0000000?oQ00000002003Qo`0400000000Xol00:?o0000008006Ko00D00000000Xo`00:?l0
02So0000000401@0o`030000001A0?l0D@3o00l0D@3o00<0000001@0o`0D0?l00P0D0?l200000080
02So0P000002001Vo`8000000P00Xol01@0000000>7o003Qo`00hOl0000000800?oQ00@00000003o
X`00oj<000000P00ofH2000000800?lX0P00000301Co008000001@1Ao`02000000<0Sol01`3oool1
00000040oooo000P0?ooo`030000003oool0oooo00D0oooo0P2?o`000`000000DOl0057o0004057o
008000000P0Do`02000000800?lX0P000002003oIP8000000P00oj<2000000800?oQ00@00000003Q
o`00hOl000000`00Xol01000000006Ko001Vo`000004000Xo`030000000D0?l0503o0080503o00<0
00000540o`1A0?l03`1A0?l00`000000503o01@0o`0201@0o`0500000000:?l002So000Xo`000000
0P00I_l0100000000:?o002So`000002003Qo`0800000000on40000000000000oj<00000003oIP00
0002003o:0050000000Do`000000057o000000000P2?o`0010000000c?l00000000000020?oe0004
0000003o^000okP0000000<0ogX000<000000?lm003o?@001@3o?@000`000000ocd00?lm00080?lm
00050000003oNP00ogX00?mj000000000P3o^00010000000ooD00?oe000000020<co000G0000002?
o`000000057o000000005?l000000000obP00000003oIP0000000?nS00000000on400000003Qo`00
00000:?o00000000I_l00000000Xo`0000000P0D0?l00`000000D@3o0000000208l0o`050000003<
0?l0c03o0<`0o`0000001@3o0?D00`000000o`2h0?l0^0060?l0^0030000003o0?D0o`3e0080o`3e
00D000000<`0o`3<0?l0c03o0000000208l0o`030000001A0?l000000080503o01D00000000Xo`00
000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000003oIP0000000?lX0000000D
o`000000057o000000000P2?o`000`000000c?l0000000020?oe00040000003o^000okP0000000<0
ogX000<000000?lm003o?@00203o?@000`000000ocd00?lm00050?lm00050000003oNP00ogX00?mj
000000000`3o^0001P000000ooD00?oe00000000c?l000000080Sol000<00000057o000000000P0D
o`001@0000000?lX00000000ofH0000000800?nS00H00000003oh@0000000000003Qo`000002002S
o`0400000000I_l006Ko000000<002So00<0000001@0o`0D0?l00P0D0?l00`000000D@3o0540o`0?
0540o`030000000D0?l0503o00<0503o00D00000000Xo`00:?l002So00000003001Vo`0400000000
Xol00:?o000000<00>7o00@00000003oh@00on4000000P00oj<2000000800?mV0P000002003o:080
00000`0Do`02000000D0DOl00P0000000`2?o`00oooo0?ooo`050?ooo`4000000@3oool00200oooo
00<000000?ooo`3oool01@3oool2000000D0DOl00P00000201Co008000000P00obP2000000800?mV
0P000002003oX`8000000P00on42000000800>7o00<00000002So`00Xol00P000002001Vo`040000
0000:?l002So000Xo`8000000`0D0?l200000100D@3o0P00000301@0o`8000000P00:?l200000080
06Ko00@00000002So`00Xol000000P00hOl00`0000000?oQ00000002003oX`0400000000ofH00000
003o:080000000@05?l00000001Ao`0000000P2?o`000`000000c?l0000000030?oe00040000003o
^000okP0000000<0ogX000<000000?lm003o?@001@3o?@000`000000o`00000000090?lm00050000
003oNP00ogX00?mj000000000P3o^00020000000ooD00?oe00000000c?l00000002?o`0000000P1A
o`00100000005?l0000000000002003oIP8000000P00on402@0000000>7o00000000Xol00000001V
o`00000002So0000000201@0o`030000001A0?l000000080S`3o00D000000<`0o`3<0?l0c03o0000
00040?l0m@030000003o0;P0o`2h00L0o`2h00<000000?l0m@3o0?D00P3o0?D01@000000c03o0<`0
o`3<0?l000000080S`3o00<000000540o`0000000P0D0?l04@00000002So00000000I_l00000002S
o`0000000>7o00000000on400000003oX`0000000?mV00000000obP0000000805?l00P00000208oo
00030000003<o`0000000080ooD000@000000?nh003o^00000000`3oNP000`000000ocd00?lm0007
0?lm00030000003o0000000000L0ocd000D000000?mj003oNP00ogX0000000030?nh00060000003o
m@00ooD00000003<o`0000000P2?o`000`000000DOl00000000201Co000400000000obP00000003o
IP80000000L00?nS0000000000000?oQ00000000hOl0000000800:?o00<00000001Vo`00I_l00P00
0002000Xo`8000000`0D0?l200000100D@3o0P00000401@0o`0500000000:?l002So000Xo`000000
0`00I_l00`0000000:?o002So`02000000800>7o0P0000000`00on40000000000002003oX`800000
0P00ofH2000000800?lX0P00000301Co00030000001Ao`00DOl000@0DOl00P0000060?ooo`400000
0@3oool00200oooo00<000000?ooo`3oool01@3oool5057o008000000P0Do`02000000<00?lX00D0
0000003oIP00ofH00?mV00000003003oX`0500000000on400?oQ003oh@0000000P00hOl200000003
002So`0000000000008006Ko0P000003000Xo`030000000D0?l0503o0080503o00<000000540o`1A
0?l03`1A0?l00`000000503o01@0o`0201@0o`0500000000:?l002So000Xo`0000000`00I_l01000
00000:?o002So`000002003Qo`0300000000on40000000800?nS00D00000003oIP0000000?lX0000
000201Co00030000001Ao`0000000080Sol000<000000<co000000000`3om@0010000000okP00?nh
000000030?mj00030000003o?@00ocd000D0ocd000@000000?l0003o00000000203o?@001@000000
ogX00?mj003oNP0000000080okP000@000000?oe003om@0000000P3<o`000`000000Sol000000002
057o0080000001000?lX00000000ofH00000003oX`0000000?oQ00000000hOl00000002So`000000
06Ko00000000:?l000000P0D0?l01P000000D@3o0540o`000000S`3o000000<0c03o00<000000?l0
m@3o0?D00P3o0?D00`000000o`2h0?l0^0090?l0^0030000003o0?D0o`3e0080o`3e00@000000<`0
o`3<0?l000000`2?0?l02@000000D@3o0000000D0?l00000000Xo`00000006Ko00000002002So`80
00000P00on42000000800?mV00T00000003o:00000005?l00000001Ao`00000008oo000000000P3<
o`000`000000ooD0000000030?nh00040000003oNP00ogX0000000P0ocd000@000000?l0003o0000
o`000P0000050?lm00030000003oNP00ogX00080ogX000@000000?nh003o^00000000P3om@001000
0000c?l00<co0000000208oo00030000001Ao`00000000805?l000<00000003o:00000000P00ofH0
0`0000000?nS003oX`02000000800>7o00@00000002So`00Xol000000`00I_l01@00000002So000X
o`00:?l0000000@0503o00<000000540o`1A0?l03`1A0?l00`000000503o01@0o`0201@0o`040000
0000:?l002So000Xo`8000000P00I_l2000000800:?o0P000002003Qo`0400000000on400?oQ0000
00<00?nS00D00000003oIP00ofH00?mV00000003003o:08000000P0Do`02000000D0DOl01`3oool1
00000040oooo000P0?ooo`030000003oool0oooo00D0oooo0`1Ao`02000000805?l00P000003003o
:08000000P00ofH2000000800?nS0P000002003oh@80000000<00>7o0000000000000P00Xol00`00
000006Ko001Vo`02000000@002So00<0000001@0o`0D0?l00P0D0?l00`000000D@3o0540o`0?0540
o`030000000D0?l0503o0080503o00D00000000Xo`00:?l002So00000003001Vo`0400000000Xol0
0:?o000000800>7o00<00000003oh@0000000P00oj<00`0000000?mV003oIP02000000805?l000<0
0000057o000000000P2?o`0010000000c?l00<co000000020?oe00040000003o^000okP0000000@0
ogX000<000000?lm003o?@00103o?@000`000000o`000?l000020?l000030000003o?@00ocd000D0
ocd000D000000?mj003oNP00ogX0000000020?nh00080000003om@00ooD00000003<o`00000008oo
00000002057o000C0000000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o
00000000Xol00000001Vo`00000002So0000000201@0o`030000001A0?l000000080S`3o00D00000
0<`0o`3<0?l0c03o000000040?l0m@030000003o0;P0o`2h00T0o`2h00<000000?l0m@3o0?D00P3o
0?D01@000000c03o0<`0o`3<0?l000000080S`3o00<000000540o`0000000P0D0?l04@00000002So
00000000I_l00000002So`0000000>7o00000000on400000003oX`0000000?mV00000000obP00000
00805?l00P00000208oo00030000003<o`0000000080ooD000@000000?nh003o^00000000`3oNP00
0`000000ocd00?lm00050?lm00030000003o0000o`0000<0o`0000<000000?lm003o?@000`3o?@00
0`000000ogX00?mj00020?mj00040000003o^000okP000000080ooD000@000000<co003<o`000000
0P2?o`001@000000DOl00000000Do`00000000800?lX00<00000003oIP0000000P00oj<00`000000
0?oQ00000002003Qo`0400000000Xol00:?o000000<006Ko00D00000000Xo`00:?l002So00000004
01@0o`030000001A0?l0D@3o00h0D@3o00<0000001@0o`0D0?l00`0D0?l2000000<002So0P000002
001Vo`8000000P00Xol200000004003Qo`0000000?oQ003oh@8000000P00oj<2000000800?mV0P00
0003003o:08000000P0Do`02000000<0DOl01`3oool100000040oooo000P0?ooo`030000003oool0
oooo00D0oooo0`00000301Co008000000`00obP01@0000000?mV003oIP00ofH0000000<00?nS00D0
0000003oh@00on400?oQ00000002003Qo`0500000000Xol00:?o002So`0000000`00I_l010000000
02So000Xo`00:?l2000000D0503o00<000000540o`1A0?l03P1A0?l00`000000503o01@0o`0201@0
o`8000000`00:?l01000000006Ko001Vo`000002002So`0600000000hOl00>7o00000000on400000
0P00oj<0200000000?mV003oIP0000000?lX0000000Do`0000000P1Ao`001P000000Sol008oo0000
0000c?l000000080ooD000D000000?nh003o^000okP0000000030?mj00030000003o?@00ocd000@0
ocd000<000000?l0003o0000103o00000`000000ocd00?lm00040?lm00050000003oNP00ogX00?mj
000000000P3o^00020000000ooD00?oe00000000c?l00000002?o`0000000P1Ao`004`0000005?l0
00000000obP00000003oIP0000000?nS00000000on400000003Qo`0000000:?o00000000I_l00000
000Xo`0000000P0D0?l00`000000D@3o0000000208l0o`050000003<0?l0c03o0<`0o`000000103o
0?D00`000000o`2h0?l0^00:0?l0^0030000003o0?D0o`3e0080o`3e00@000000<`0o`3<0?l00000
0P2?0?l050000000D@3o0540o`000000503o00000000:?l00000001Vo`0000000:?o00000000hOl0
0000003oh@0000000?nS00000000ofH000000P00obP01`0000005?l00000001Ao`00000008oo0000
00000P3<o`000`000000ooD0000000030?nh00040000003oNP00ogX0000000H0ocd000<000000?l0
003o0000103o00000`000000ocd00?lm00040?lm00030000003oNP00ogX00080ogX000@000000?nh
003o^00000000P3om@001P000000c?l00<co00000000Sol000000080DOl000<0000001Co00000000
0P00obP00`0000000?mV00000002003oX`0600000000on40000000000000hOl000000P00Xol01@00
000006Ko001Vo`00I_l0000000<002So00<0000001@0o`0D0?l00P0D0?l2000000l0D@3o00<00000
01@0o`0D0?l0100D0?l01@00000002So000Xo`00:?l0000000<006Ko00D00000002So`00Xol00:?o
00000002003Qo`8000000P00on401@0000000?nS003oX`00oj<0000000<00?mV00<00000003o:000
obP00P00obP2000000805?l0100000060?ooo`4000000@3oool000l0oooo00<0_g@0000000000000
100000000`18W=l0oooo0?ooo`070?ooo`030000003oool0oooo00D0oooo0`0Do`03000000<00?lX
0P000002003oIP8000000P00oj<2000000800?oQ0P0000000`00hOl0000000000002002So`800000
0P00I_l2000000<002So00<0000001@0o`0D0?l00`0D0?l200000100D@3o00@0000001@0o`0D0?l0
503o0P000003000Xo`80000000<006Ko0000000000000P00Xol01P0000000>7o003Qo`0000000?oQ
000000800?nS00P00000003oIP00ofH00000003o:00000005?l000000080DOl000H0000008oo0000
000000000<co000000020?oe00050000003o^000okP00?nh000000000`3oNP000`000000ocd00?lm
00040?lm00030000003o0000o`0000@0o`0000<000000?lm003o?@00103o?@001@000000ogX00?mj
003oNP0000000080okP000H000000?oe003om@0000000<co0000000208oo00030000001Ao`000000
00805?l00P0000001000ofH00000003oX`000002003oh@0900000000hOl00000002So`00000006Ko
00000000:?l000000080503o00<000000540o`0000000P2?0?l01@000000c03o0<`0o`3<0?l00000
00<0o`3e00<000000?l0^03o0;P02`3o0;P00`000000o`3e0?l0m@020?l0m@040000003<0?l0c03o
00000080S`3o01@000000540o`1A0?l0000001@0o`00000002So00000000I_l00000002So`000000
0>7o00000000on400000003oX`0000000?mV000000800?lX00L0000001Co00000000DOl00000002?
o`0000000080c?l000<000000?oe000000000`3o^0001@000000ogX00?mj003oNP00000000D0ocd0
00<000000?l0003o00001@3o00000`000000ocd00?lm00040?lm00030000003oNP00ogX00080ogX0
00@000000?nh003o^00000000P3om@001P000000c?l00<co00000000Sol000000080DOl000L00000
01Co000000000000003o:00000000?mV0080000000@00?nS00000000on4000000P00hOl00`000000
0:?o002So`020000008006Ko00@00000000Xo`00:?l002So0P00000401@0o`030000001A0?l0D@3o
00d0D@3o0P00000501@0o`0400000000:?l002So000Xo`8000000P00I_l2000000800:?o0P000002
003Qo`80000000<00?oQ0000000000000P00oj<2000000800?mV0P000004003o:08000000`0Do`07
0?ooo`4000000@3oool00100oooo00<0gia80000000000000`0000000`00B9`0gooo0?ooo`070?oo
o`030000003oool0oooo00D0oooo00<05?l00000000000001000obP2000000800?mV0P000002003o
X`8000000P00on42000000800>7o00D00000002So`00Xol00:?o00000003001Vo`0400000000:?l0
02So000Xo`8000001@0D0?l00`000000D@3o0540o`0>0540o`8000000`0D0?l00`00000002So000X
o`02000Xo`0400000000I_l006Ko000000<00:?o00H00000003Qo`0000000000003oh@000002003o
X`0900000000ofH0000000000000obP0000001Co00000000DOl00080000000D0Sol00000003<o`00
c?l0000000020?oe00050000003o^000okP00?nh000000000`3oNP000`000000ocd00?lm00040?lm
00030000003o0000o`0000D0o`0000<000000?lm003o?@00103o?@001@000000ogX00?mj003oNP00
00000080okP000H000000?oe003om@0000000<co0000000208oo000E0000001Ao`00000001Co0000
00000?lX00000000ofH00000003oX`0000000?oQ00000000hOl00000002So`00000006Ko00000000
:?l000000080503o00<000000540o`0000000P2?0?l01@000000c03o0<`0o`3<0?l0000000@0o`3e
00<000000?l0^03o0;P02`3o0;P00`000000o`3e0?l0m@020?l0m@040000003<0?l0c03o00000080
S`3o00X000000540o`1A0?l0000001@0o`00000002So00000000I_l000000P00Xol02@0000000>7o
00000000on400000003oX`0000000?mV00000002003o:08000000P1Ao`001@000000Sol00000003<
o`0000000080ooD000D000000?nh003o^000okP0000000020?mj00030000003o?@00ocd000@0ocd0
00<000000?l0003o00001@3o0002000000D0ocd000D000000?mj003oNP00ogX0000000020?nh0005
0000003om@00ooD00?oe000000000P3<o`001P000000Sol0000000000000DOl0000000805?l000T0
0000003o:00000000000003oIP0000000?nS00000000on400P0000000`00hOl00000002So`02002S
o`0400000000I_l006Ko000000@002So00<0000001@0o`0D0?l00P0D0?l2000000l0D@3o00<00000
01@0o`0D0?l00`0D0?l2000000<002So0P000002001Vo`8000000P00Xol01@0000000>7o003Qo`00
hOl0000000800?oQ0P000002003oX`8000000P00ofH2000000@00?lX0P0000000`0Do`00oooo0?oo
o`050?ooo`4000000@3oool00140oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`07
0?ooo`030000003oool0oooo00D0oooo00<00000003o:000obP00P00obP2000000800?mV0P000002
003oX`8000000P00on42000000800>7o0P000002002So`8000000`00I_l01@00000002So000Xo`00
:?l0000000H0503o00<000000540o`1A0?l03P1A0?l00`000000503o01@0o`0301@0o`0500000000
:?l002So000Xo`0000000P00I_l01`0000000:?o002So`00Xol00000003Qo`0000000P00on401P00
00000?nS003oX`0000000?mV000000800?lX00<0000001Co000Do`000P00000208oo00050000003<
o`00c?l00<co000000000P3om@0010000000okP00?nh000000030?mj00030000003o?@00ocd000@0
ocd000<000000?l0003o00001@3o00000`000000ocd00?lm00040?lm00050000003oNP00ogX00?mj
000000000P3o^0001P000000ooD00?oe00000000c?l000000080Sol000<00000057o000000000P0D
o`0200000004003oIP0000000?nS000000800?oQ00T00000003Qo`0000000:?o00000000I_l00000
000Xo`0000000P0D0?l00`000000D@3o0000000208l0o`050000003<0?l0c03o0<`0o`0000000`3o
0?D00`000000o`2h0?l0^00=0?l0^0030000003o0?D0o`3e0080o`3e00@000000<`0o`3<0?l00000
0P2?0?l00`000000D@3o0000000201@0o`0?00000000:?l00000001Vo`0000000:?o00000000hOl0
0000003oh@0000000?nS00000000ofH0000000800?lX00L0000001Co00000000DOl00000002?o`00
00000080c?l000<000000?oe000000000`3o^0001@000000ogX00?mj003oNP00000000D0ocd000<0
00000?l0003o00001`3o00000`000000ocd00?lm00030?lm00050000003oNP00ogX00?mj00000000
0P3o^0001`000000ooD00?oe003om@0000000<co000000000`2?o`001@000000DOl00000000Do`00
000000<00?lX0P000002003oX`0500000000on400?oQ00000000hOl00P000002002So`0400000000
I_l006Ko000000@002So00<0000001@0o`0D0?l00`0D0?l00`000000D@3o0540o`0=0540o`030000
000D0?l0503o00@0503o00<00000000Xo`00:?l00P00:?l01@00000006Ko001Vo`00I_l000000080
0:?o0P000002003Qo`8000000P00on42000000800?nS0P000002003oIP8000001000obP2000000H0
oooo0@0000010?ooo`004P3oool00`3OW4P0M;oo0?ooo`0;0?ooo`030000003oool0oooo00D0oooo
0`00obP2000000800?mV0P000002003oX`8000000P00on42000000800>7o0P000002002So`800000
1000I_l01000000002So000Xo`00:?l2000000D0503o00<000000540o`1A0?l03@1A0?l2000000@0
503o0P000003000Xo`0400000000I_l00000000000800:?o0P000002003Qo`0400000000on400000
003oX`80000000D00?mV00000000obP00?lX0000000201Co00030000001Ao`0000000080Sol000D0
00000<co003<o`0000000?oe000200000080okP000D000000?mj003oNP00ogX0000000060?lm0003
0000003o0000o`0000H0o`0000<000000?lm003o?@00103o?@001@000000ogX00?mj003oNP000000
0080okP000H000000?oe003om@0000000<co0000000208oo000<0000001Ao`00000001Co00000000
0?lX00000000ofH00000003oX`0000000?oQ0P0000002000hOl00000002So`00000006Ko00000000
:?l000000P0D0?l00`000000D@3o0000000208l0o`050000003<0?l0c03o0<`0o`0000000`3o0?D0
0`000000o`2h0?l0^00=0?l0^0030000003o0?D0o`3e0080o`3e00@000000<`0o`3<0?l000000P2?
0?l02P000000D@3o0540o`000000503o00000000:?l00000001Vo`000002002So`0?00000000hOl0
0000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`0000000080Sol000<0
00000<co000000000P3om@0010000000okP00?nh000000030?mj00030000003o?@00ocd000@0ocd0
00<000000?l0003o00001P3o00000`000000ocd00?lm00040?lm00050000003oNP00ogX00?mj0000
00000P3o^00200000080ooD000<000000<co000000000P2?o`001P000000DOl0057o000000005?l0
000000800?lX00H00000003oIP0000000000003oX`000002003oh@0400000000hOl00>7o00000080
0:?o0P0000000`00I_l00000000Xo`03000Xo`800000100D0?l00`000000D@3o0540o`0=0540o`03
0000000D0?l0503o00@0503o0P000003000Xo`0400000000I_l006Ko001Vo`8000000P00Xol20000
00800>7o0P000002003oh@8000000P00oj<2000000800?mV0P000003003o:0L0oooo0@0000010?oo
o`000`3oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00<0oooo00<0_gA809cOo`3o
ool02P3oool00`000000oooo0?ooo`050?ooo`03003o:0000000000000800?mV0P000002003oX`80
00000P00on42000000<00>7o00D00000002So`00Xol00:?o00000004001Vo`8000000P00:?l20000
00D0503o0P00000>0540o`8000001@0D0?l01000000002So000Xo`00:?l200000003001Vo`000000
0:?o00800:?o00<00000003Qo`00hOl0100000001@00oj<00000003oIP00ofH0000000800?lX00L0
000001Co000Do`000000057o00000000Sol00080000000D0c?l0000000000000ooD0000000030?nh
00050000003oNP00ogX00?mj000000001P3o?@000`000000o`000?l000060?l000030000003o?@00
ocd000@0ocd000D000000?mj003oNP00ogX0000000020?nh00040000003om@00ooD000000080c?l0
00<0000008oo000000000P1Ao`003`0000005?l000000000obP00000003oIP0000000?nS00000000
on400000003Qo`0000000:?o00000002001Vo`0300000000:?l000000080503o00<000000540o`00
00000P2?0?l01@000000c03o0<`0o`3<0?l0000000<0o`3e00<000000?l0^03o0;P03@3o0;P00`00
0000o`3e0?l0m@020?l0m@040000003<0?l0c03o00000080S`3o00H000000540o`1A0?l0000001@0
o`000002000Xo`8000000P00Xol02@0000000>7o00000000on400000003oX`0000000?mV00000002
003o:0070000000Do`000000057o00000000Sol0000000020<co00040000003om@00ooD000000080
okP000D000000?mj003oNP00ogX0000000050?lm00030000003o0000o`0000L0o`0000<000000?lm
003o?@00103o?@001@000000ogX00?mj003oNP00000000<0okP000D000000?oe003om@0000000<co
00020000000608oo00000000DOl0057o000000005?l00P0000001P00obP00000003oIP00ofH00000
003oX`80000000D00?oQ0000000000000>7o00000003002So`0400000000I_l00000000000@002So
00<0000001@0o`0D0?l00P0D0?l2000000h0D@3o0P00000601@0o`8000000P00:?l2000000<006Ko
00D00000002So`00Xol00:?o00000003003Qo`8000000P00on42000000800?nS0P000002003oIP80
000000<00?lX0?ooo`3oool01@3oool100000040oooo000C0?ooo`030?oog`2LB7@0_ooo00X0oooo
0`0000050?ooo`0300000000ofH00?mV008000000P00oj<2000000800?oQ0P000003003Qo`800000
0P00Xol2000000<006Ko0P000002000Xo`8000001P0D0?l00`000000D@3o0540o`0<0540o`800000
1@0D0?l20000008002So0P0000000`00I_l0000000000002002So`80000000H00>7o000000000000
0?oQ00000000oj<2000000800?mV00D00000003o:000obP0000001Co000200000004057o00000000
Sol000000080c?l000@000000?oe003om@0000000`3o^0001@000000ogX00?mj003oNP00000000H0
ocd000<000000?l0003o00001`3o00000`000000ocd00?lm00040?lm00050000003oNP00ogX00?mj
000000000P3o^0001P000000ooD00?oe00000000c?l000000080Sol000d00000057o000000005?l0
00000000obP00000003oIP0000000?nS00000000on40000000800>7o00L00000002So`00000006Ko
00000000:?l000000080503o00<000000540o`0000000P2?0?l01@000000c03o0<`0o`3<0?l00000
00<0o`3e00<000000?l0^03o0;P03`3o0;P00`000000o`3e0?l0m@020?l0m@040000003<0?l0c03o
00000080S`3o00X000000540o`1A0?l0000001@0o`00000002So00000000I_l000000P00Xol03`00
00000>7o00000000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl000000002
08oo00030000003<o`0000000080ooD000@000000?nh003o^00000000`3oNP000`000000ocd00?lm
00040?lm00030000003o0000o`0000L0o`0000<000000?lm003o?@00103o?@0010000000ogX00?mj
003oNP0200000080okP000@000000?oe003om@0000000P3<o`001@000000Sol0000000000000DOl0
0080000000@05?l000000000obP000000P00ofH200000006003oX`0000000000003oh@0000000>7o
0P000002002So`80000000<006Ko0000000000000`00:?l2000000@0503o0P00000>0540o`800000
1P0D0?l01000000002So000Xo`00:?l20000008006Ko0P000002002So`8000000`00hOl200000080
0?oQ0P000002003oX`<0000000<00?mV000000000000103oool300000040oooo000D0?ooo`030?no
M000M;l0oooo00T0oooo00<000000?ooo`3oool01@3oool00`00ofH0000000000002003oX`800000
0P00on42000000<00>7o0P000002002So`8000000`00I_l2000000<002So00<0000001@0o`0D0?l0
1@0D0?l00`000000D@3o0540o`0;0540o`8000001@0D0?l2000000<002So00@00000001Vo`00I_l0
00000`00Xol0100000000>7o003Qo`000002003oh@0300000000oj<0000000800?mV0P0000000`00
obP000000000000200000005057o00000000000008oo000000000P3<o`0010000000ooD00?oe0000
00030?nh00050000003oNP00ogX00?mj000000001@3o?@02000000X0o`0000<000000?lm003o?@00
0`3o?@001@000000ogX00?mj003oNP0000000080okP000@000000?oe003om@0000000P3<o`001000
0000Sol00000001Ao`020000000>01Co000000000?lX00000000ofH00000003oX`0000000?oQ0000
0000hOl00000002So`000002001Vo`0300000000:?l000000080503o00<000000540o`0000000P2?
0?l01@000000c03o0<`0o`3<0?l0000000<0o`3e00<000000?l0^03o0;P03`3o0;P00`000000o`3e
0?l0m@020?l0m@040000003<0?l0c03o00000080S`3o00H000000540o`1A0?l0000001@0o`000002
000Xo`0=00000000I_l00000002So`0000000>7o00000000on400000003oX`0000000?mV00000002
003o:0050000000Do`000000057o000000000P2?o`000`000000c?l0000000020?oe00040000003o
^000okP0000000<0ogX000<000000?lm003o?@00103o?@000`000000o`000?l000070?l000800000
1@3o?@000`000000ogX00?mj00020?mj00040000003o^000okP000000080ooD000@000000<co003<
o`0000000P2?o`001`000000DOl0057o000000005?l000000000obP00P000002003oIP0600000000
oj<00?nS00000000on4000000P00hOl2000000800:?o00<00000001Vo`00I_l00P000003000Xo`03
0000000D0?l0503o00<0503o0`00000=0540o`030000000D0?l0503o00@0503o00<00000000Xo`00
:?l00P00:?l20000008006Ko0P000002002So`<000000P00hOl2000000800?oQ0P000003003oX`80
00001`3oool100000040oooo000D0?ooo`030?oo_`1d04P0W=oo00T0oooo00<000000?ooo`3oool0
1@3oool00`0000000?nS003oX`02000000<00?oQ00@00000003Qo`00hOl00>7o0P000002002So`80
00001000I_l00`00000002So000Xo`02000Xo`030000000D0?l0503o00D0503o00<000000540o`1A
0?l02P1A0?l2000000D0503o0P000004000Xo`0400000000I_l006Ko000000<00:?o00@00000003Q
o`00000000000P00on401P0000000?nS003oX`0000000?mV000000800?lX00H0000001Co000Do`00
0000057o0000000208oo00040000003<o`00c?l000000080ooD000D000000?nh003o^000okP00000
00030?mj00030000003o?@00ocd000<0ocd000<000000?l0003o00002@3o00000`000000ocd00?lm
00030?lm00050000003oNP00ogX00?mj000000000`3o^0001P000000ooD00?oe00000000c?l00000
0080Sol000D00000057o000000005?l000000002003o:080000000/00?nS00000000on400?oQ0000
0000hOl00000002So`00000006Ko00000002000Xo`030000000D0?l000000080D@3o00@0000008l0
o`2?0?l000000P3<0?l01@000000o`3e0?l0m@3o0?D000000180o`2h00<000000?l0m@3o0?D00P3o
0?D010000000c03o0<`0o`00000208l0o`060000001A0?l0D@3o0000000D0?l000000P00:?l02`00
000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<0000000800?mV00L00000003o:000
00005?l00000001Ao`0000000080Sol000<000000<co000000000P3om@001@000000okP00?nh003o
^000000000<0ogX000<000000?lm003o?@000`3o?@000`000000o`000?l000090?l000030000003o
?@00ocd000<0ocd000D000000?mj003oNP00ogX0000000030?nh00040000003om@00ooD000000080
c?l000@0000008oo002?o`0000000P1Ao`000`0000005?l000000002003o:00300000000ofH00000
00800?nS00<00000003oh@00on400P000002003Qo`0400000000Xol00:?o000000<006Ko00@00000
000Xo`00:?l002So0P00000601@0o`8000002`1A0?l00`000000503o01@0o`0501@0o`8000001000
:?l01000000006Ko001Vo`00I_l2000000<00:?o00D00000003Qo`00hOl00>7o00000003003oh@80
00000`00oj<00`000000oooo0?ooo`040?ooo`4000000@3oool001@0oooo00<0oono07@00018W=l0
2@3oool00`000000oooo0?ooo`050?ooo`03003oX`000000000000@00?oQ00D00000003Qo`00hOl0
0>7o00000003002So`0300000000I_l006Ko008006Ko0P000004000Xo`030000000D0?l0503o00D0
503o00<000000540o`1A0?l02@1A0?l2000000H0503o00<00000000Xo`00:?l00P00:?l200000080
06Ko00<00000002So`00Xol00P0000000`00hOl00000003oh@02003oh@0600000000oj<00?nS0000
0000ofH000000P00obP01P0000005?l001Co00000000DOl0000000<0Sol000<000000<co00000000
0P3om@001@000000okP00?nh003o^000000000<0ogX000<000000?lm003o?@000`3o?@000`000000
o`000?l000090?l000030000003o?@00ocd000@0ocd000D000000?mj003oNP00ogX0000000020?nh
00040000003om@00ooD000000080c?l000D0000008oo00000000DOl00000000201Co000=00000000
obP00000003oIP0000000?nS00000000on400000003Qo`0000000:?o00000002001Vo`0800000000
:?l002So0000000D0?l000000540o`00000308l0o`040000003<0?l0c03o000000<0o`3e00<00000
0?l0^03o0;P04@3o0;P01@000000o`3e0?l0m@3o0?D0000000<0c03o00<0000008l0o`0000000P1A
0?l00`000000503o00000002000Xo`0300000000I_l0000000800:?o00/00000003Qo`0000000?oQ
00000000oj<00000003oIP0000000?lX0000000201Co00030000001Ao`0000000080Sol000<00000
0<co000000000P3om@0010000000okP00?nh000000030?mj00030000003o?@00ocd000@0ocd000<0
00000?l0003o0000203o00000`000000ocd00?lm00040?lm00050000003oNP00ogX00?mj00000000
0`3o^00010000000ooD00?oe000000020<co00040000002?o`00Sol000000080DOl000<0000001Co
000000000P00obP00`0000000?mV00000002003oX`0500000000on400?oQ003oh@0000000P00hOl0
0`0000000:?o002So`020000008006Ko0P000003000Xo`030000000D0?l0503o00D0503o00<00000
0540o`1A0?l02@1A0?l00`000000503o01@0o`0601@0o`0300000000:?l002So008002So00<00000
001Vo`00I_l00P00I_l01@0000000:?o002So`00Xol0000000<00>7o00<00000003oh@00on400P00
on4200000003003oX`3oool0oooo00D0oooo0@0000010?ooo`003`3oool00`3o_i`0W=oo0?ooo`02
0?ooo`030?oOW0180000B9cO00T0oooo00<000000?ooo`3oool01@3oool00`0000000?oQ003oh@02
003oh@8000000P00hOl2000000800:?o0P000003001Vo`8000001000:?l2000000L0503o00<00000
0540o`1A0?l0201A0?l2000000H0503o0P000004000Xo`0500000000I_l006Ko001Vo`0000000P00
Xol0100000000>7o003Qo`000002003oh@80000000H00?nS0000000000000?mV00000000obP20000
000301Co0000000000000080DOl000H0000008oo002?o`0000000<co000000020?oe00050000003o
^000okP00?nh000000000`3oNP000`000000ocd00?lm00040?lm00030000003o0000o`0000T0o`00
00<000000?lm003o?@000`3o?@001@000000ogX00?mj003oNP00000000<0okP000H000000?oe003o
m@0000000<co0000000208oo00050000001Ao`00000001Co000000000P00obP20000000;003oX`00
00000?oQ003oh@0000000>7o00000000Xol00000001Vo`0000000P00:?l00`000000503o00000002
0540o`040000002?0?l0S`3o00000080c03o00<000000?l0m@3o0?D00P3o0?D00`000000o`2h0?l0
^00A0?l0^0030000003o0?D0o`3e0080o`3e00@000000<`0o`3<0?l000000P2?0?l01P000000D@3o
0540o`000000503o0000008002So00/00000001Vo`0000000:?o00000000hOl00000003oh@000000
0?nS00000002003oIP0700000000obP0000001Co00000000DOl00000000208oo00030000003<o`00
00000080ooD000D000000?nh003o^000okP0000000030?mj00030000003o?@00ocd000<0ocd000<0
00000?l0003o00002@3o00000`000000ocd00?lm00040?lm00050000003oNP00ogX00?mj00000000
0`3o^00010000000ooD00?oe000000020<co00040000002?o`00Sol000000080DOl000L0000001Co
000000000000003o:00000000?mV0080000000<00?nS0000000000000P00on4200000003003Qo`00
00000:?o00800:?o00D00000001Vo`00I_l006Ko00000003000Xo`8000001P0D0?l2000000X0D@3o
00<0000001@0o`0D0?l01P0D0?l00`00000002So000Xo`02000Xo`8000000`00I_l2000000800:?o
0P000002003Qo`8000001000on42000000H0oooo0@0000010?ooo`003`3oool0203ooml0W4Q809cO
o`3oool0oono07@0000007@0_ooo2@3oool00`000000oooo0?ooo`050?ooo`800?oQ0`000002003Q
o`8000000P00Xol2000000<006Ko0P000005000Xo`030000000D0?l0503o00L0503o00<000000540
o`1A0?l0101A0?l4000000L0503o00<00000000Xo`00:?l00P00:?l20000008006Ko0P0000000`00
Xol0000000000002003Qo`0400000000on400?oQ000000800?nS00H00000003oIP0000000000003o
:000000201Co00030000001Ao`00DOl00080000000D0Sol0000000000000c?l0000000020?oe0005
0000003o^000okP00?nh000000000`3oNP000`000000ocd00?lm00040?lm00030000003o0000o`00
00T0o`0000<000000?lm003o?@00103o?@001@000000ogX00?mj003oNP0000000080okP000@00000
0?oe003om@0000000P3<o`000`000000Sol000000002057o00090000000Do`000000003o:0000000
0?mV00000000oj<0000000800?oQ00L00000003Qo`0000000:?o00000000I_l00000008002So00<0
000001@0o`0000000P1A0?l010000000S`3o08l0o`0000020<`0o`030000003o0?D0o`3e0080o`3e
00<000000?l0^03o0;P01@3o0;P5000000P0o`2h00D000000?l0m@3o0?D0o`3e000000020<`0o`04
0000002?0?l0S`3o00000080D@3o00<0000001@0o`0000000P00:?l01@00000006Ko00000000Xol0
000000800>7o00T00000003oh@0000000?nS00000000ofH00000003o:00000000P0Do`000`000000
DOl00000000208oo00030000003<o`0000000080ooD000D000000?nh003o^000okP0000000030?mj
00030000003o?@00ocd000<0ocd000<000000?l0003o00002@3o0002000000D0ocd000D000000?mj
003oNP00ogX0000000030?nh00040000003om@00ooD000000080c?l000@0000008oo002?o`000000
0P1Ao`001@0000005?l001Co000000000?lX0080000000@00?mV0000000000000?nS0P000002003o
h@0400000000hOl00000000000800:?o00@00000001Vo`00I_l006Ko0P000003000Xo`8000001P0D
0?l4000000L0D@3o00<0000001@0o`0D0?l01P0D0?l2000000@002So0P000003001Vo`8000000P00
Xol300000003003Qo`00000000000080000000<00?oQ0?ooo`3oool01@3oool100000040oooo000@
0?ooo`060?oog`2LB00000000000000007@0_ooo2P3oool00`000000oooo0?ooo`050?ooo`800000
0`00hOl2000000800:?o0P000004001Vo`0300000000:?l002So00<002So0P00000:01@0o`040000
001A0?l0D@3o0540o`<000002P0D0?l01000000002So000Xo`00:?l2000000<006Ko00@00000002S
o`00Xol000000P00hOl2000000800?oQ00H00000003oX`00oj<00000003oIP000002003o:0040000
000Do`005?l000000080DOl000@0000008oo002?o`0000000P3<o`0010000000ooD00?oe00000003
0?nh00050000003oNP00ogX00?mj000000001P3o?@000`000000o`000?l0000:0?l000030000003o
?@00ocd000<0ocd000D000000?mj003oNP00ogX0000000020?nh00040000003om@00ooD000000080
c?l000<0000008oo000000000P1Ao`000`0000005?l001Co0002000000800?mV00L00000003oX`00
00000?oQ00000000hOl0000000800:?o00D00000001Vo`00000002So0000000201@0o`040000001A
0?l0D@3o00000080S`3o00@000000<`0o`3<0?l000000`3o0?D00`000000o`2h0?l0^0050?l0^003
0000003o07X0o`1j00<0o`1j00<000000?l0^03o0;P01@3o0;P01@000000o`3e0?l0m@3o0?D00000
0080c03o00@0000008l0o`2?0?l000000P1A0?l00`000000503o00000002000Xo`0500000000I_l0
0000002So`0000000P00hOl01`0000000?oQ00000000oj<00000003oIP0000000P00obP01@000000
5?l00000001Ao`0000000080Sol000<000000<co000000000`3om@0010000000okP00?nh00000003
0?mj00030000003o?@00ocd000<0ocd000<000000?l0003o00002`3o00000`000000ocd00?lm0003
0?lm00050000003oNP00ogX00?mj000000000`3o^0000`000000ooD00?oe0002000000050<co0000
0000000008oo000000000P1Ao`00100000005?l001Co00000002003o:00600000000ofH00?mV0000
0000oj<000000`00on40100000000>7o003Qo`000002002So`8000000`00I_l01@00000002So000X
o`00:?l0000000X0503o0`0000030540o`030000000D0?l0503o00T0503o0P000004000Xo`030000
0000I_l006Ko008006Ko0P000003002So`8000000`00hOl2000000H0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooo`<00>7o0P000002002So`8000001000I_l2000000@002So0P00
000=01@0o`<000002`0D0?l2000000<002So00<00000001Vo`00I_l00P00I_l01`0000000:?o002S
o`00Xol00000003Qo`0000000P00on4200000005003oX`0000000000003oIP0000000P00obP01000
00005?l001Co00000002057o00040000002?o`00Sol000000080c?l000@000000?oe003om@000000
0`3o^0001@000000ogX00?mj003oNP00000000H0ocd000<000000?l0003o00002`3o00000`000000
ocd00?lm00020?lm00050000003oNP00ogX00?mj000000000`3o^00010000000ooD00?oe00000002
0<co00030000002?o`0000000080DOl000D0000001Co000000000?lX00000002003oIP8000000P00
on401@0000000>7o00000000Xol00000008006Ko00D00000000Xo`000000503o000000030540o`03
0000002?0?l000000080c03o00<000000?l0m@3o0?D00P3o0?D00`000000o`2h0?l0^0040?l0^003
0000003o07X0o`1j00@0o`1j00<000000?l0^03o0;P01@3o0;P01@000000o`3e0?l0m@3o0?D00000
00<0c03o00H0000008l0o`2?0?l000000540o`00000201@0o`0300000000:?l00000008006Ko00<0
0000002So`0000000P00hOl2000000800?nS00L00000003oIP0000000?lX0000000Do`0000000080
DOl000<0000008oo000000000P3<o`0010000000ooD00?oe000000030?nh00050000003oNP00ogX0
0?mj000000001@3o?@000`000000o`000?l0000:0?l000030000003o?@00ocd000@0ocd000D00000
0?mj003oNP00ogX0000000030?nh00050000003om@00ooD00?oe000000000P3<o`000`000000Sol0
00000002057o00040000000Do`005?l0000000800?lX00D00000003oIP00ofH00000003oX`020000
00800?oQ00@00000003Qo`00hOl000000`00Xol01@00000006Ko001Vo`00I_l0000000<002So0P00
000;01@0o`<000003P0D0?l00`00000002So000Xo`02000Xo`8000001000I_l2000000<00:?o00@0
0000003Qo`00hOl00>7o1`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00<0
0>7o0000000000000P00Xol2000000@006Ko0P000005000Xo`030000000D0?l0503o01P0503o0`00
0003000Xo`8000000`00I_l2000000800:?o0P0000000`00hOl00000003oh@02000000800?nS00D0
0000003oIP0000000000003o:002000000805?l000D00000057o001Ao`00000008oo000200000003
0<co0000000000000080ooD000D000000?nh003o^000okP0000000030?mj00030000003o?@00ocd0
00@0ocd000<000000?l0003o00002`3o00000`000000ocd00?lm00030?lm00050000003oNP00ogX0
0?mj000000000P3o^00010000000ooD00?oe000000020<co00030000002?o`0000000080DOl000<0
000001Co000Do`000P000002003oIP0700000000oj<00000003oh@0000000>7o00000002002So`05
00000000I_l00000000Xo`0000000P0D0?l010000000D@3o0540o`00000208l0o`040000003<0?l0
c03o000000<0o`3e00<000000?l0^03o0;P01@3o0;P00`000000o`1j0?l0NP050?l0NP030000003o
0;P0o`2h00D0o`2h00D000000?l0m@3o0?D0o`3e000000020<`0o`040000002?0?l0S`3o00000080
D@3o00<0000001@0o`0000000P00:?l01@00000006Ko00000000Xol0000000800>7o00L00000003o
h@0000000?nS00000000ofH0000000800?lX00D0000001Co00000000DOl00000000208oo00030000
003<o`00000000<0ooD000@000000?nh003o^00000000`3oNP000`000000ocd00?lm00030?lm0003
0000003o0000o`0000/0o`0000<000000?lm003o?@00103o?@0010000000ogX00?mj003oNP020000
0080okP000D000000?oe003om@00ooD0000000020<co00030000002?o`0000000080DOl000<00000
01Co000Do`000P0000001@00obP0000000000000ofH0000000800?nS0P0000001000on4000000000
0000hOl2000000800:?o0P000002001Vo`8000000`00:?l3000001X0503o00<00000000Xo`00:?l0
0`00:?l2000000@006Ko0P000002002So`80000000<00>7o0?ooo`3oool01@3oool100000040oooo
000P0?ooo`030000003oool0oooo00D0oooo00D00000002So`00Xol00:?o00000004001Vo`800000
1P00:?l00`000000503o01@0o`0G01@0o`8000001000:?l2000000<006Ko0P000002002So`800000
00D00>7o0000000000000?oQ00000002003oX`80000000D00?mV00000000obP00?lX0000000201Co
0080000000D0DOl0000000000000Sol0000000020<co00050000003om@00ooD00?oe000000000P3o
^002000000<0ogX000<000000?lm003o?@00103o?@000`000000o`000?l0000;0?l000030000003o
?@00ocd000<0ocd000D000000?mj003oNP00ogX0000000030?nh00040000003om@00ooD000000080
c?l000<0000008oo000000000P1Ao`001@0000005?l000000000obP0000000800?mV00L00000003o
X`0000000?oQ00000000hOl0000000800:?o00D00000001Vo`00000002So0000000201@0o`040000
001A0?l0D@3o00000080S`3o00@000000<`0o`3<0?l000000`3o0?D00`000000o`2h0?l0^0040?l0
^0030000003o07X0o`1j00L0o`1j00<000000?l0^03o0;P0103o0;P01@000000o`3e0?l0m@3o0?D0
000000<0c03o00H0000008l0o`2?0?l000000540o`00000201@0o`0300000000:?l00000008006Ko
00L00000002So`0000000>7o00000000on40000000800?nS00H00000003oIP0000000?lX0000000D
o`0200000006057o00000000Sol008oo00000000c?l00P0000020?oe00040000003o^000okP00000
00<0ogX000<000000?lm003o?@000`3o?@000`000000o`000?l0000;0?l000030000003o?@00ocd0
00@0ocd000<000000?mj003oNP000P3oNP000`000000okP00?nh000200000080ooD000D000000<co
003<o`00000008oo000200000003057o00000000000000805?l000D00000003o:000obP00000003o
IP02000000800?nS00@00000003oh@00on4000000P00hOl2000000800:?o0P000002001Vo`800000
1000:?l2000001T0503o00<00000000Xo`00:?l01000:?l2000000@006Ko00@00000002So`00Xol0
0:?o0P0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool3002So`040000
0000I_l006Ko001Vo`8000001P00:?l2000001T0503o00<00000000Xo`00:?l00P00:?l2000000<0
06Ko0P000002002So`80000000<00>7o0000000000000P00on40100000000?nS003oX`000002003o
IP0500000000obP00?lX0000000Do`000P000002057o00040000002?o`00Sol0000000<0c?l000@0
00000?oe003om@0000000P3o^0000`000000ogX00?mj00020?mj00030000003o?@00ocd000@0ocd0
00<000000?l0003o00002`3o00000`000000ocd00?lm00030?lm00050000003oNP00ogX00?mj0000
00000`3o^00010000000ooD00?oe000000020<co00030000002?o`0000000080DOl000<0000001Co
000000000P00obP00`0000000?mV00000002003oX`8000000P00hOl01@0000000:?o00000000I_l0
0000008002So00H0000001@0o`0D0?l000000540o`00000208l0o`050000003<0?l0c03o0<`0o`00
00000`3o0?D00`000000o`2h0?l0^0030?l0^0030000003o07X0o`1j00P0o`1j00<000000?l0^03o
0;P0103o0;P01@000000o`3e0?l0m@3o0?D0000000<0c03o00H0000008l0o`2?0?l000000540o`00
000201@0o`0300000000:?l00000008006Ko00D00000002So`0000000>7o00000002003oh@030000
0000oj<0000000800?mV0P00000201Co00030000001Ao`0000000080Sol000@000000<co003<o`00
00000P3om@0010000000okP00?nh000000040?mj00030000003o?@00ocd000<0ocd000<000000?l0
003o00002`3o00000`000000ocd00?lm00040?lm00030000003oNP00ogX00080ogX000D000000?nh
003o^000okP0000000020?oe00040000003<o`00c?l000000080Sol000<00000057o001Ao`000P00
00001@0Do`000000003o:000obP0000000800?mV00<00000003oX`00oj<00P0000000`00on400000
00000002003Qo`8000000P00Xol20000008006Ko0P000004000Xo`030000000D0?l0503o01L0503o
0P000006000Xo`8000000`00I_l0100000000:?o002So`00Xol70?ooo`4000000@3oool00200oooo
00<000000?ooo`3oool01@3oool00`00Xol0000000000003001Vo`0300000000:?l002So00D002So
00<0000001@0o`0D0?l05`0D0?l2000000@002So00@00000001Vo`00I_l006Ko0P000002002So`80
00000P00hOl01@0000000?oQ003oh@00on40000000800?nS00D00000003oIP00ofH00000003o:002
0000000301Co00000000DOl00080DOl000@0000008oo002?o`0000000`3<o`0010000000ooD00?oe
000000020?nh00030000003oNP00ogX00080ogX000<000000?lm003o?@00103o?@000`000000o`00
0?l0000<0?l000030000003o?@00ocd00080ocd000<000000?mj003oNP000P3oNP0010000000okP0
0?nh000000020?oe00040000003<o`00c?l000000080Sol000<00000057o000000000P0Do`000`00
00000?lX00000002003oIP0700000000oj<00000003oh@0000000>7o00000002002So`0500000000
I_l00000000Xo`0000000P0D0?l010000000D@3o0540o`00000208l0o`040000003<0?l0c03o0000
00<0o`3e00<000000?l0^03o0;P0103o0;P00`000000o`1j0?l0NP090?l0NP030000003o0;P0o`2h
00<0o`2h00D000000?l0m@3o0?D0o`3e000000030<`0o`040000002?0?l0S`3o00000080D@3o00<0
000001@0o`0000000P00:?l00`00000006Ko00000002002So`0500000000hOl00000003oh@000000
0P00oj<01@0000000?mV00000000obP0000000805?l000<00000057o000000000P2?o`0010000000
c?l00<co000000020?oe00040000003o^000okP0000000@0ogX000<000000?lm003o?@000`3o?@00
0`000000o`000?l0000;0?l000030000003o?@00ocd000@0ocd000<000000?mj003oNP000P3oNP00
1@000000okP00?nh003o^00000000080ooD000@000000<co003<o`0000000P2?o`001P000000DOl0
057o001Ao`00000001Co0080000000D00?lX00000000ofH00?mV00000003003oX`0300000000on40
0?oQ008000000P00hOl01@0000000:?o002So`00Xol0000000<006Ko00<00000000Xo`00:?l00P00
:?l2000001T0503o00<00000000Xo`00:?l01@00:?l01000000006Ko001Vo`00I_l200000003002S
o`3oool0oooo00D0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0400000000
I_l006Ko001Vo`8000001P00:?l2000001L0503o0`000004000Xo`8000000P00I_l2000000800:?o
0P000002003Qo`8000000P00on4200000004003oX`0000000000003oIP80000000D00?lX0000000D
o`005?l000000002057o0080000000<0Sol00000000000000P3<o`0200000080ooD000D000000?nh
003o^000okP0000000030?mj00030000003o?@00ocd000@0ocd000<000000?l0003o0000303o0000
0`000000ocd00?lm00030?lm00050000003oNP00ogX00?mj000000000`3o^00010000000ooD00?oe
000000020<co00030000002?o`0000000080DOl000<0000001Co000000000P00obP00`0000000?mV
00000002003oX`8000000P00hOl01@0000000:?o00000000I_l00000008002So00H0000001@0o`0D
0?l000000540o`00000208l0o`050000003<0?l0c03o0<`0o`0000000`3o0?D00`000000o`2h0?l0
^0030?l0^0030000003o07X0o`1j00X0o`1j00<000000?l0^03o0;P0103o0;P01@000000o`3e0?l0
m@3o0?D0000000<0c03o00H0000008l0o`2?0?l000000540o`00000201@0o`0300000000:?l00000
008006Ko00D00000002So`0000000>7o00000002003oh@0300000000oj<0000000800?mV0P000002
01Co00030000001Ao`0000000080Sol000@000000<co003<o`0000000P3om@001@000000okP00?nh
003o^000000000<0ogX000<000000?lm003o?@00103o?@000`000000o`000?l0000;0?l000030000
003o?@00ocd000@0ocd000<000000?mj003oNP000P3oNP001@000000okP00?nh003o^00000000080
ooD000<000000<co003<o`000P0000000`2?o`00000000000002057o00050000000Do`005?l00000
0000obP00P0000000`00ofH0000000000002003oX`8000000P00on400`0000000>7o003Qo`020000
00800:?o0P000002001Vo`8000001000:?l3000001L0503o0P000006000Xo`8000000`00I_l20000
00H0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`<006Ko00<00000000Xo`00
:?l01@00:?l00`000000503o01@0o`0D01@0o`<000001@00:?l20000008006Ko0P000003002So`05
00000000hOl00>7o003Qo`0000000P00on42000000800?nS00@00000003oIP00ofH000000P00obP0
100000005?l001Co00000002057o00040000002?o`00Sol0000000<0c?l000<000000?oe003om@00
0P0000020?nh008000000P3oNP02000000H0ocd000<000000?l0003o0000303o00000`000000ocd0
0?lm00030?lm00030000003oNP00ogX00080ogX000@000000?nh003o^00000000P3om@0010000000
c?l00<co0000000208oo00030000001Ao`00000000805?l000<00000003o:00000000P00ofH01@00
00000?nS00000000on40000000800>7o00D00000002So`00000006Ko00000002000Xo`060000000D
0?l0503o0000001A0?l000000P2?0?l01@000000c03o0<`0o`3<0?l0000000<0o`3e00<000000?l0
^03o0;P00`3o0;P00`000000o`1j0?l0NP0;0?l0NP030000003o0;P0o`2h00<0o`2h00D000000?l0
m@3o0?D0o`3e000000030<`0o`040000002?0?l0S`3o00000080D@3o00<0000001@0o`0000000P00
:?l00`00000006Ko00000002002So`0500000000hOl00000003oh@0000000P00oj<01@0000000?mV
00000000obP0000000805?l000<00000057o000000000P2?o`0010000000c?l00<co000000020?oe
00050000003o^000okP00?nh000000000`3oNP000`000000ocd00?lm00030?lm00030000003o0000
o`0000`0o`0000<000000?lm003o?@00103o?@02000000<0ogX000D000000?nh003o^000okP00000
00020?oe008000000P3<o`0010000000Sol008oo00000002057o00040000000Do`005?l000000080
0?lX00<00000003oIP00ofH00P000002003oX`80000000<00?oQ00000000hOl00P00hOl010000000
0:?o002So`00Xol20000008006Ko0P000005000Xo`<000005P0D0?l00`00000002So000Xo`05000X
o`0400000000I_l006Ko001Vo`L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oo
o`03001Vo`000000000000L002So00<0000001@0o`0D0?l04`0D0?l2000000H002So0P000002001V
o`8000001000Xol01@0000000>7o003Qo`00hOl0000000800?oQ00<00000003oX`00oj<00P000000
1000ofH0000000000000obP20000000301Co0000000000000080DOl000@0000008oo002?o`000000
0`3<o`0010000000ooD00?oe000000030?nh00050000003oNP00ogX00?mj000000001`3o?@000`00
0000o`000?l0000<0?l000030000003o?@00ocd000<0ocd000<000000?mj003oNP000P3oNP001000
0000okP00?nh000000020?oe00040000003<o`00c?l000000080Sol000<00000057o000000000P0D
o`000`0000000?lX00000002003oIP0300000000oj<0000000800?oQ00D00000003Qo`0000000:?o
00000002001Vo`0300000000:?l000000080503o00@000000540o`1A0?l000000P2?0?l01@000000
c03o0<`0o`3<0?l0000000<0o`3e00<000000?l0^03o0;P00P3o0;P00`000000o`1j0?l0NP0<0?l0
NP030000003o0;P0o`2h00<0o`2h00<000000?l0m@3o0?D00P3o0?D010000000c03o0<`0o`000002
08l0o`060000001A0?l0D@3o0000000D0?l000000P00:?l00`00000006Ko00000002002So`050000
0000hOl00000003oh@0000000P00oj<00`0000000?mV00000002003o:0030000000Do`0000000080
DOl000H0000008oo002?o`0000000<co000000030?oe00050000003o^000okP00?nh000000000`3o
NP000`000000ocd00?lm00030?lm00030000003o0000o`0000`0o`0000<000000?lm003o?@001@3o
?@001@000000ogX00?mj003oNP00000000<0okP000D000000?oe003om@00ooD0000000020<co0004
0000002?o`00Sol000000080DOl00P000000100Do`00000000000000obP2000000800?mV0P000002
003oX`0300000000on40000000<00>7o00<00000002So`00Xol00P00Xol20000008006Ko0P000006
000Xo`8000005@0D0?l00`00000002So000Xo`05000Xo`80000000<006Ko0?ooo`3oool01@3oool1
00000040oooo000P0?ooo`030000003oool0oooo00D0oooo00<00000000Xo`00:?l01@00:?l20000
01@0503o0P000005000Xo`<000000P00I_l2000000@00:?o0P000002003Qo`80000000@00?oQ0000
000000000?nS0P000002003oIP0400000000obP00?lX000000805?l000<00000057o001Ao`000P00
00000`2?o`000000000000020<co008000000P3om@001@000000okP00?nh003o^000000000<0ogX0
00<000000?lm003o?@00103o?@02000000h0o`0000<000000?lm003o?@00103o?@001@000000ogX0
0?mj003oNP0000000080okP000D000000?oe003om@00ooD0000000020<co00030000002?o`000000
0080DOl000H0000001Co000Do`000000003o:0000002003oIP0500000000oj<00000003oh@000000
0P00hOl01@0000000:?o00000000I_l00000008002So00H0000001@0o`0D0?l000000540o`000003
08l0o`050000003<0?l0c03o0<`0o`0000000`3o0?D00`000000o`2h0?l0^0020?l0^0030000003o
07X0o`1j00d0o`1j00<000000?l0^03o0;P00`3o0;P01@000000o`3e0?l0m@3o0?D000000080c03o
00@0000008l0o`2?0?l000000P1A0?l010000000503o01@0o`000002000Xo`0300000000I_l00000
00800:?o00D00000003Qo`0000000?oQ00000002003oX`0500000000ofH00000003o:00000000P0D
o`000`000000DOl00000000208oo00040000003<o`00c?l0000000<0ooD000@000000?nh003o^000
0000103oNP000`000000ocd00?lm00030?lm00030000003o0000o`0000`0o`000P0000060?lm0004
0000003oNP00ogX00?mj008000000P3o^0001@000000ooD00?oe003om@0000000080c?l00P000000
0`2?o`00000000000002057o00040000000Do`005?l0000000800?lX0P000002003oIP80000000<0
0?nS00000000on400P000002003Qo`8000001000Xol20000008006Ko0`000005000Xo`<000004`0D
0?l2000000L002So0P0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool7
000Xo`030000000D0?l0503o0140503o0`000004000Xo`<000000`00I_l2000000@00:?o0P000002
003Qo`8000000P00on40100000000?nS000000000002003oIP80000000@00?lX0000000000005?l0
0P0000000`1Ao`0000000000000208oo00050000003<o`00c?l00<co000000000`3om@001@000000
okP00?nh003o^000000000@0ogX000<000000?lm003o?@000`3o?@000`000000o`000?l0000=0?l0
00030000003o?@00ocd000@0ocd000D000000?mj003oNP00ogX0000000020?nh00050000003om@00
ooD00?oe000000000P3<o`001P000000Sol008oo00000000DOl0000000805?l000<00000003o:000
00000P00ofH00`0000000?nS00000002003oh@0500000000hOl00000002So`0000000P00I_l00`00
000002So0000000201@0o`040000001A0?l0D@3o00000080S`3o00D000000<`0o`3<0?l0c03o0000
00030?l0m@030000003o0;P0o`2h0080o`2h00<000000?l0NP3o07X03P3o07X00`000000o`2h0?l0
^0030?l0^0030000003o0?D0o`3e0080o`3e00<000000<`0o`0000000`2?0?l01P000000D@3o0540
o`000000503o0000008002So00<00000001Vo`0000000P00Xol01@0000000>7o00000000on400000
00800?nS00<00000003oIP0000000P00obP00`0000005?l000000002057o00040000002?o`00Sol0
00000080c?l000@000000?oe003om@0000000`3o^0001@000000ogX00?mj003oNP00000000H0ocd0
00<000000?l0003o00003@3o00000`000000ocd00?lm00040?lm00030000003oNP00ogX00080ogX0
00@000000?nh003o^00000000`3om@001@000000c?l00<co003<o`0000000080Sol00P000000101A
o`0000000000000Do`02000000800?lX0P000002003oIP0500000000oj<0000000000000on400P00
0002003Qo`8000001000Xol2000000<006Ko0`000005000Xo`<000004P0D0?l00`00000002So000X
o`05000Xo`L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`L002So00<00000
01@0o`0D0?l03`0D0?l2000000H002So00<00000001Vo`00I_l00P00I_l2000000@00:?o0P000002
003Qo`8000000P00on4200000003003oX`0000000?mV00800?mV00@00000003o:000obP000000P0D
o`001@000000DOl0057o001Ao`0000000080Sol000D000000<co003<o`00c?l0000000020?oe0080
00000`3o^0000`000000ogX00?mj00020?mj00030000003o?@00ocd000@0ocd000<000000?l0003o
0000303o00000`000000ocd00?lm00040?lm00050000003oNP00ogX00?mj000000000`3o^0001@00
0000ooD00?oe003om@0000000080c?l000<0000008oo000000000P1Ao`00200000005?l001Co0000
00000?lX00000000ofH000000P00oj<0200000000?oQ0000000000000>7o00000000Xol000000P00
I_l00`00000002So0000000201@0o`040000001A0?l0D@3o00000080S`3o00D000000<`0o`3<0?l0
c03o000000030?l0m@030000003o0;P0o`2h0080o`2h00<000000?l0NP3o07X03`3o07X00`000000
o`2h0?l0^0020?l0^0030000003o0?D0o`3e0080o`3e00@000000<`0o`3<0?l000000P2?0?l01000
0000D@3o0540o`00000201@0o`0600000000:?l002So00000000I_l000000P00Xol01@0000000>7o
00000000on40000000800?nS00<00000003oIP0000000P00obP00`0000005?l000000002057o0004
0000002?o`00Sol000000080c?l000@000000?oe003om@0000000`3o^0001@000000ogX00?mj003o
NP00000000H0ocd000<000000?l0003o0000303o00000`000000ocd00?lm00050?lm00030000003o
NP00ogX00080ogX000<000000?nh003o^0000P0000020?oe008000000P3<o`001@000000Sol008oo
002?o`0000000080DOl000<0000001Co000Do`000P000002003o:00400000000ofH00?mV00000080
0?nS0P0000000`00on40000000000002003Qo`8000001000Xol2000000@006Ko0P000006000Xo`80
00004@0D0?l00`00000002So000Xo`04000Xo`L0oooo0@0000010?ooo`00803oool00`000000oooo
0?ooo`050?ooo`L002So00<0000001@0o`0D0?l0300D0?l3000000H002So0P000003001Vo`800000
1000Xol2000000800>7o0P000002003oh@8000000P00oj<01@0000000?mV003oIP00ofH000000080
0?lX00@0000001Co0000000000000P1Ao`020000000308oo0000000000000080c?l00P0000020?oe
00040000003o^000okP00?nh00800000103oNP000`000000ocd00?lm00040?lm00030000003o0000
o`0000`0o`0000<000000?lm003o?@00103o?@001@000000ogX00?mj003oNP00000000<0okP000D0
00000?oe003om@00ooD0000000020<co00030000002?o`0000000080DOl000P0000001Co000Do`00
0000003o:00000000?mV000000800?nS00<00000003oh@0000000P00hOl00`0000000:?o00000002
001Vo`0600000000:?l002So0000000D0?l000000P1A0?l01@000000S`3o08l0o`2?0?l000000080
c03o00D000000?l0m@3o0?D0o`3e000000050?l0^0030000003o07X0o`1j00l0o`1j00<000000?l0
^03o0;P00P3o0;P00`000000o`3e0?l0m@020?l0m@040000003<0?l0c03o00000080S`3o00@00000
0540o`1A0?l000000P0D0?l01P00000002So000Xo`00000006Ko000000800:?o00<00000003Qo`00
00000P00on400`0000000?nS00000002003oIP0300000000obP0000000805?l000H00000057o001A
o`00000008oo000000030<co00040000003om@00ooD0000000<0okP000@000000?mj003oNP00ogX0
0P0000050?lm00030000003o0000o`0000`0o`0000<000000?lm003o?@001@3o?@000`000000ogX0
0?mj00020?mj00050000003o^000okP00?nh000000000`3om@000`000000c?l00<co000200000080
Sol00P0000000`1Ao`0000000000000201Co000400000000obP00?lX000000800?mV00D00000003o
X`00oj<00?nS00000002003oh@8000000P00hOl2000000@00:?o0P000004001Vo`8000001P00:?l2
000000l0503o00<00000000Xo`00:?l01000:?l70?ooo`4000000@3oool00200oooo00<000000?oo
o`3oool01@3oool7000Xo`030000000D0?l0503o00P0503o10000007000Xo`8000000`00I_l20000
00@00:?o0P000002003Qo`8000000P00on42000000800?nS0P000002003oIP80000000D00?lX0000
000000005?l000000002057o008000000P2?o`001@000000c?l00<co003<o`00000000<0ooD000D0
00000?nh003o^000okP0000000040?mj008000001@3o?@02000000h0o`0000<000000?lm003o?@00
103o?@000`000000ogX00?mj00020?mj00040000003o^000okP0000000<0ooD000@000000<co003<
o`0000000P2?o`000`000000DOl00000000201Co000600000000obP00?lX00000000ofH000000P00
oj<00`0000000?oQ00000002003Qo`0300000000Xol00000008006Ko00<00000000Xo`0000000P0D
0?l010000000D@3o0540o`00000208l0o`050000003<0?l0c03o0<`0o`0000000`3o0?D00`000000
o`2h0?l0^0020?l0^0030000003o07X0o`1j0140o`1j00<000000?l0^03o0;P00P3o0;P01@000000
o`3e0?l0m@3o0?D0000000<0c03o00@0000008l0o`2?0?l000000P1A0?l01P000000503o01@0o`00
000002So0000008006Ko00H00000002So`00Xol00000003Qo`000002003oh@0500000000oj<00000
003oIP0000000P00obP00`0000005?l000000002057o00040000002?o`00Sol000000080c?l000D0
00000?oe003om@00ooD0000000030?nh00030000003oNP00ogX00080ogX000<000000?lm003o?@00
0`3o?@000`000000o`000?l0000<0?l0008000001P3o?@02000000<0ogX000D000000?nh003o^000
okP0000000030?oe00050000003<o`00c?l00<co000000000`2?o`000`000000DOl0057o00020000
000401Co000000000000003o:080000000<00?mV0000000000000P00oj<2000000800?oQ0P000002
003Qo`8000001000Xol2000000@006Ko0P000006000Xo`@000002P0D0?l00`00000002So000Xo`05
000Xo`L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`L002So2`000008000X
o`<000000`00I_l2000000@00:?o0P000002003Qo`8000000P00on42000000800?nS0P000002003o
IP8000000P00obP0100000005?l0000000000002057o00030000002?o`00Sol0008000000P3<o`02
00000080ooD00P0000030?nh00030000003oNP00ogX00080ogX000<000000?lm003o?@00103o?@00
0`000000o`000?l0000=0?l000030000003o?@00ocd000@0ocd000<000000?mj003oNP000P3oNP00
10000000okP00?nh000000030?oe00040000003<o`00c?l000000080Sol000H00000057o001Ao`00
000001Co00000002003o:00300000000ofH0000000800?nS00<00000003oh@0000000P00hOl00`00
00000:?o00000002001Vo`0600000000:?l002So0000000D0?l000000P1A0?l01@000000S`3o08l0
o`2?0?l000000080c03o00D000000?l0m@3o0?D0o`3e000000050?l0^0030000003o07X0o`1j0140
o`1j00<000000?l0^03o0;P00P3o0;P00`000000o`3e0?l0m@020?l0m@040000003<0?l0c03o0000
0080S`3o00@000000540o`1A0?l000000P0D0?l01P00000002So000Xo`00000006Ko000000800:?o
00<00000003Qo`0000000P00on400`0000000?nS00000002003oIP0300000000obP0000000805?l0
00@00000057o001Ao`0000000P2?o`0010000000c?l00<co000000030?oe00050000003o^000okP0
0?nh000000000`3oNP000`000000ocd00?lm00040?lm00030000003o0000o`0000d0o`0000<00000
0?lm003o?@001@3o?@0010000000ogX00?mj003oNP0200000080okP00P0000020?oe008000000P3<
o`0200000080Sol00P000002057o00040000000Do`005?l0000000800?lX00<00000003oIP00ofH0
0P000002003oX`8000000P00on42000000800>7o0P000004002So`8000001000I_l3000000L002So
2`000007000Xo`L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oooaL002So0`00
0004001Vo`8000000P00Xol4000000800>7o0P000002003oh@8000000P00oj<2000000800?mV0P00
0002003o:080000000@05?l00000001Ao`00DOl00P0000000`2?o`000000000000020<co00800000
0P3om@02000000@0okP000<000000?mj003oNP000P3oNP000`000000ocd00?lm00050?lm00030000
003o0000o`0000`0o`0000<000000?lm003o?@00103o?@000`000000ogX00?mj00020?mj00050000
003o^000okP00?nh000000000P3om@001@000000c?l00<co003<o`0000000080Sol000<00000057o
000000000P0Do`001P0000000?lX003o:00000000?mV000000800?nS00<00000003oh@0000000P00
hOl00`0000000:?o00000002001Vo`0600000000:?l000000000000D0?l000000P1A0?l01@000000
S`3o08l0o`2?0?l000000080c03o00D000000?l0m@3o0?D0o`3e000000050?l0^0030000003o07X0
o`1j0140o`1j00<000000?l0^03o0;P00`3o0;P01@000000o`3e0?l0m@3o0?D0000000<0c03o00@0
000008l0o`2?0?l000000P1A0?l01P000000503o01@0o`00000002So0000008006Ko00H00000002S
o`00Xol00000003Qo`000002003oh@0300000000oj<0000000800?mV00<00000003o:00000000P0D
o`0010000000DOl0057o0000000208oo00040000003<o`00c?l000000080ooD000D000000?nh003o
^000okP0000000040?mj00030000003o?@00ocd000@0ocd000<000000?l0003o0000303o00000`00
0000ocd00?lm00050?lm00030000003oNP00ogX000<0ogX000@000000?nh003o^000okP00P000002
0?oe008000000P3<o`0200000080Sol00P000000101Ao`00000001Co000Do`0200000003003o:000
0000000000800?mV0P000002003oX`8000000P00on42000000800>7o10000002002So`<000001000
I_l2000001L002So1`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo5@00:?l2
000000@006Ko0`000002002So`8000000`00hOl3000000800?oQ0P000002003oX`8000000P00ofH2
000000800?lX0P00000201Co00040000001Ao`00DOl000000080Sol000D000000<co003<o`00c?l0
000000030?oe00030000003o^000okP00080okP00P0000030?mj008000001`3o?@000`000000o`00
0?l0000;0?l0008000001P3o?@000`000000ogX00?mj00020?mj00050000003o^000okP00?nh0000
00000P3om@001@000000c?l00<co003<o`0000000080Sol000<00000057o000000000P0Do`001P00
00000?lX003o:00000000?mV000000800?nS00<00000003oh@0000000P00hOl00`0000000:?o0000
0002001Vo`0400000000:?l002So00000080503o00@000000540o`1A0?l000000P2?0?l01@000000
c03o0<`0o`3<0?l0000000<0o`3e00<000000?l0^03o0;P00P3o0;P00`000000o`1j0?l0NP0B0?l0
NP030000003o0;P0o`2h00<0o`2h00D000000?l0m@3o0?D0o`3e000000030<`0o`040000002?0?l0
S`3o00000080D@3o00@0000001@0o`0D0?l000000P00:?l00`00000006Ko00000002002So`060000
0000hOl00>7o00000000on4000000P00oj<00`0000000?mV00000002003o:0060000000Do`005?l0
0000001Ao`0000000`2?o`0010000000c?l00<co000000020?oe00050000003o^000okP00?nh0000
0000103oNP000`000000ocd00?lm00040?lm008000003@3o00000`000000ocd00?lm00050?lm0080
0000103oNP000`000000okP00?nh00020?nh00050000003om@00ooD00?oe000000000`3<o`001P00
0000Sol008oo002?o`000000057o008000000P0Do`000`0000000?lX003o:002000000800?mV0P00
0002003oX`8000000P00on43000000<00>7o0P000003002So`<000000`00I_l2000001D002So1`3o
ool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo4P00:?l3000000@006Ko0P000004
002So`0300000000hOl00>7o00<000000`00on42000000800?nS0P000002003oIP8000000P00obP2
000000805?l00P0000000`1Ao`0000000000000208oo00030000003<o`00c?l0008000000P3om@02
000000@0okP000<000000?mj003oNP000P3oNP000`000000ocd00?lm00060?lm00030000003o0000
o`0000/0o`0000<000000?lm003o?@001@3o?@000`000000ogX00?mj00020?mj00050000003o^000
okP00?nh000000000P3om@001@000000c?l00<co003<o`0000000080Sol000<00000057o00000000
0P0Do`00100000000?lX003o:0000002003oIP0300000000oj<0000000800?oQ00<00000003Qo`00
00000P00Xol01P00000006Ko001Vo`00000002So00000080503o00@000000540o`1A0?l000000`2?
0?l010000000c03o0<`0o`0000030?l0m@030000003o0;P0o`2h00<0o`2h00<000000?l0NP3o07X0
4`3o07X00`000000o`2h0?l0^0020?l0^0030000003o0?D0o`3e0080o`3e00@000000<`0o`3<0?l0
00000`2?0?l010000000D@3o0540o`00000201@0o`0300000000:?l00000008006Ko00@00000002S
o`00Xol000000P00hOl01@0000000?oQ00000000oj<0000000800?mV00<00000003o:00000000P0D
o`0010000000DOl0057o0000000208oo00050000003<o`00c?l00<co000000000P3om@0200000080
okP00P0000030?mj00030000003o?@00ocd000D0ocd000<000000?l0003o00002`3o0002000000L0
ocd000<000000?mj003oNP000P3oNP000`000000okP00?nh00020?nh008000000P3om@0200000080
c?l00P00000208oo0080000000@0DOl00000000000005?l00P000002003o:08000000P00ofH20000
00800?nS0P000003003oh@<000000P00hOl2000000@00:?o0P000003001Vo`<000004P00:?l70?oo
o`4000000@3oool00200oooo0P0000060?ooo`h002So10000005001Vo`8000000P00Xol010000000
0:?o000000000002003Qo`0300000000on400?oQ00800?oQ0P000002003oX`8000000P00ofH20000
00800?lX0P00000201Co008000000P1Ao`001@000000Sol008oo002?o`0000000080c?l000D00000
0?oe003om@00ooD0000000040?nh00800000103oNP000`000000ocd00?lm00060?lm00030000003o
0000o`0000/0o`0000<000000?lm003o?@001@3o?@000`000000ogX00?mj00020?mj00050000003o
^000okP00?nh000000000P3om@001@000000c?l00<co003<o`0000000080Sol000@00000057o001A
o`0000000P0Do`000`0000000?lX00000002003oIP0600000000oj<00?nS00000000on4000000P00
hOl00`0000000:?o00000002001Vo`0400000000:?l002So00000080503o00@000000540o`1A0?l0
00000P2?0?l01@000000c03o0<`0o`3<0?l0000000<0o`3e00<000000?l0^03o0;P00P3o0;P00`00
0000o`1j0?l0NP0D0?l0NP030000003o0;P0o`2h00<0o`2h00D000000?l0m@3o0?D0o`3e00000003
0<`0o`040000002?0?l0S`3o00000080D@3o00@0000001@0o`0D0?l000000P00:?l00`00000006Ko
00000002002So`0600000000hOl00>7o00000000on4000000P00oj<00`0000000?mV00000002003o
:0040000000Do`005?l000000080DOl000@0000008oo002?o`0000000`3<o`001@000000ooD00?oe
003om@00000000<0okP000D000000?mj003oNP00ogX0000000070?lm008000003@3o00000`000000
ocd00?lm00050?lm00030000003oNP00ogX00080ogX00P0000040?nh00050000003om@00ooD00?oe
000000000`3<o`001@000000Sol008oo002?o`0000000080DOl000<0000001Co000Do`000P000002
003o:08000000P00ofH2000000800?nS0P000004003oh@0400000000hOl00>7o003Qo`80000000@0
0:?o00000000Xol00:?o0P000004001Vo`@000003P00:?l60?ooo`8000000@3oool00200oooo00<0
00000?ooo`3oool01@3oool:000Xo`@000001@00I_l4000000<00:?o0`000002003Qo`80000000<0
0?oQ0000000000000P000002003oX`8000000P00ofH2000000800?lX0P00000201Co008000000P1A
o`0200000080Sol00P0000020<co00050000003om@00ooD00?oe00000000103o^0000`000000ogX0
0?mj00030?mj00030000003o?@00ocd000H0ocd000<000000?l0003o00002`3o00000`000000ocd0
0?lm00050?lm00030000003oNP00ogX00080ogX000D000000?nh003o^000okP0000000030?oe0004
0000003<o`00c?l000000080Sol000@00000057o001Ao`0000000P0Do`00100000000?lX003o:000
0002003oIP0300000000oj<0000000800?oQ00<00000003Qo`0000000P00Xol01000000006Ko001V
o`000002000Xo`040000000D0?l0503o00000080D@3o00@0000008l0o`2?0?l000000`3<0?l01@00
0000o`3e0?l0m@3o0?D0000000@0o`2h00<000000?l0NP3o07X05@3o07X00`000000o`2h0?l0^002
0?l0^08000000`3o0?D010000000c03o0<`0o`00000308l0o`040000001A0?l0D@3o00000080503o
00<00000000Xo`0000000P00I_l01P0000000:?o002So`0000000>7o000000800?oQ00H00000003o
X`00oj<00000003oIP000002003o:0040000000Do`005?l000000080DOl000@0000008oo002?o`00
00000P3<o`001@000000ooD00?oe003om@00000000<0okP000<000000?mj003oNP000P3oNP000`00
0000ocd00?lm00060?lm00030000003o0000o`0000X0o`0000<000000?lm003o?@001P3o?@000`00
0000ogX00?mj00030?mj00030000003o^000okP00080okP000@000000?oe003om@00ooD00P000002
0<co008000000P2?o`000`000000DOl0057o0002000000805?l00P000002003o:08000000P00ofH2
000000800?nS100000001000on400000000000000002003Qo`<000000`00Xol3000000D006Ko1000
000:000Xo`L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`H002So10000007
001Vo`8000001000Xol3000000<00>7o0P0000000`00on40000000000003003oX`<000000P00ofH2
000000800?lX0P00000201Co008000000P1Ao`0200000080Sol00P0000020<co008000000P3om@02
000000<0okP00P0000040?mj00800000203o?@000`000000o`000?l0000:0?l0008000001P3o?@02
000000@0ogX000D000000?nh003o^000okP0000000030?oe00040000003<o`00c?l000000080Sol0
00@00000057o001Ao`0000000P0Do`00100000000?lX003o:0000002003oIP0300000000oj<00000
00800?oQ00<00000003Qo`0000000`00Xol00`00000006Ko00000002000Xo`040000000D0?l0503o
00000080D@3o00@0000008l0o`2?0?l000000`3<0?l01@000000o`3e0?l0m@3o0?D0000000D0o`2h
00<000000?l0NP3o07X05@3o07X00`000000o`2h0?l0^0030?l0^0050000003o0?D0o`3e0?l0m@00
00000`3<0?l010000000S`3o08l0o`0000020540o`070000000D0?l0503o01@0o`00000002So0000
0002001Vo`0600000000Xol00:?o00000000hOl000000P00on400`0000000?nS00000002003oIP04
00000000obP00?lX000000805?l000@00000057o001Ao`0000000P2?o`0010000000c?l00<co0000
00030?oe00050000003o^000okP00?nh00000000103oNP02000000L0ocd000<000000?l0003o0000
2P3o00000`000000ocd00?lm00060?lm00800000103oNP02000000<0okP00P0000030?oe00800000
0P3<o`020000000308oo0000000000000080DOl00P00000201Co008000000P00obP2000000800?mV
0`000003003oX`8000000P00on42000000<00>7o0P000004002So`8000001`00I_l4000000H002So
1`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo0P00:?l4000000P006Ko0`00
0004002So`8000000`00hOl300000003003oh@000000000000800?nS0`000003003oIP8000000P00
obP2000000805?l00P000002057o008000000P2?o`0200000080c?l00P0000020?oe008000000`3o
^002000000@0ogX00P0000090?lm00030000003o0000o`0000T0o`000P0000060?lm008000001@3o
NP001@000000okP00?nh003o^000000000<0ooD000@000000<co003<o`0000000P2?o`0010000000
DOl0057o0000000201Co000400000000obP00?lX000000800?mV00<00000003oX`0000000P00on40
100000000>7o003Qo`000002002So`0400000000I_l006Ko0000008002So00<0000001@0o`000000
0P1A0?l01@000000S`3o08l0o`2?0?l0000000<0c03o00D000000?l0m@3o0?D0o`3e000000040?l0
^0030000003o07X0o`1j01L0o`1j00<000000?l0^03o0;P00`3o0;P01@000000o`3e0?l0m@3o0?D0
00000080c03o00D0000008l0o`2?0?l0S`3o000000020540o`040000000D0?l0503o0000008002So
00@00000001Vo`00I_l000000P00Xol00`0000000>7o00000002003oh@0300000000oj<000000080
0?mV00@00000003o:000obP000000P0Do`0010000000DOl0057o0000000208oo00030000003<o`00
c?l0008000000P3om@0200000080okP000<000000?mj003oNP000`3oNP02000000H0ocd00P00000;
0?l000030000003o?@00ocd000L0ocd00P0000040?mj008000000`3o^002000000<0ooD00P000002
0<co0080000000<0Sol00000000000000P1Ao`02000000805?l00P000002003o:08000000`00ofH3
000000800?nS0`0000001000on400000000000000002003Qo`8000001000Xol3000000P006Ko1000
0002000Xo`L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`8000001`00I_l5
000000800:?o1@0000000`00hOl000000000000200000004003oh@0000000000000000800?nS0P00
0002003oIP@0000000D00?lX000000000000000001Co000300000005057o0000000000000000002?
o`000`0000020<co008000000P3om@02000000@0okP000<000000?mj003oNP000`3oNP000`000000
ocd00?lm00080?lm00030000003o0000o`0000H0o`00100000070?lm00030000003oNP00ogX000<0
ogX00P0000020?nh008000000P3om@0200000080c?l000@0000008oo002?o`0000000P1Ao`001000
00005?l001Co00000002003o:00600000000ofH00?mV00000000oj<000000P00on40100000000>7o
003Qo`000002002So`0400000000I_l006Ko0000008002So00@0000001@0o`0D0?l000000P1A0?l0
10000000S`3o08l0o`0000030<`0o`050000003o0?D0o`3e0?l0m@0000001@3o0;P00`000000o`1j
0?l0NP080?l0NPD000002P3o07X00`000000o`2h0?l0^0030?l0^0050000003o0?D0o`3e0?l0m@00
00000`3<0?l01@000000S`3o08l0o`2?0?l000000080D@3o00@0000001@0o`0D0?l000000P00:?l0
1P00000006Ko001Vo`0000000:?o000000800>7o00D00000003oh@00on400000003oX`0200000005
003oIP0000000?lX003o:00000000P0Do`0010000000DOl0057o0000000208oo008000000P3<o`00
1@000000ooD00?oe003om@0000000080okP00`0000040?mj00030000003o?@00ocd000D0ocd01000
00080?l000030000003o?@00ocd000P0ocd000<000000?mj003oNP000`3oNP02000000<0okP00P00
00030?oe008000000P3<o`020000000508oo0000000000000000001Ao`000`0000000`0Do`000000
00000006000000800?mV0P000003003oX`8000000P00on4300000003003Qo`000000000000<00000
0P00Xol5000000L006Ko0`0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3o
ool6001Vo`<000001@00Xol2000000@00>7o0P000004003oh@8000000`00oj<00`0000000?mV003o
IP02000000<00?lX0P00000201Co008000000P1Ao`0200000080Sol00P0000040<co00030000003o
m@00ooD00080ooD000@000000?nh003o^000okP00P0000050?mj00030000003o?@00ocd000X0ocd0
00<000000?l0003o00000`3o0002000000X0ocd000<000000?mj003oNP000`3oNP001@000000okP0
0?nh003o^000000000<0ooD000D000000<co003<o`00c?l00000000208oo00040000001Ao`00DOl0
000000805?l000@00000003o:000obP000000P00ofH00`0000000?nS00000003003oh@0300000000
hOl0000000800:?o00@00000001Vo`00I_l000000P00:?l010000000503o01@0o`0000020540o`04
0000002?0?l0S`3o000000<0c03o00<000000?l0m@3o0?D00P3o0?D00`000000o`2h0?l0^0030?l0
^0030000003o07X0o`1j00L0o`1j00<000000?l0?@3o03d00`3o03d00`000000o`1j0?l0NP080?l0
NP030000003o0;P0o`2h0080o`2h00D000000?l0m@3o0?D0o`3e000000030<`0o`050000002?0?l0
S`3o08l0o`0000000P1A0?l010000000503o01@0o`000002000Xo`0400000000I_l006Ko00000080
0:?o00<00000003Qo`0000000P00on40100000000?nS003oX`000002003oIP0400000000obP00?lX
000000805?l000@00000057o001Ao`0000000`2?o`0010000000c?l00<co000000030?oe00030000
003o^000okP00080okP000<000000?mj003oNP000P3oNP000`000000ocd00?lm00080?lm00800000
1@3o00000`000000ocd00?lm000:0?lm00030000003oNP00ogX000@0ogX000<000000?nh003o^000
0P3o^0000`000000ooD00?oe00020?oe00040000003<o`00c?l00<co008000000P2?o`0200000080
DOl00P00000301Co000400000000obP00?lX003o:08000000P00ofH2000000<00?nS0P000003003o
h@8000001000hOl2000000D00:?o0`000006001Vo`L0oooo0@0000010?ooo`00803oool00`000000
oooo0?ooo`050?ooo`8006Ko10000004002So`@000001000hOl2000000<00?oQ0`000003003oX`80
000000<00?mV0000000000000`00obP2000000<05?l000D00000057o001Ao`00DOl00000000308oo
00030000003<o`00c?l00080c?l00P0000030?oe008000000`3o^0000`000000ogX00?mj00040?mj
00030000003o?@00ocd000`0ocd00P0000000`3o000000000000000;0?lm00030000003oNP00ogX0
00<0ogX000D000000?nh003o^000okP0000000030?oe00050000003<o`00c?l00<co000000000P2?
o`0010000000DOl0057o0000000201Co000400000000obP00?lX000000800?mV00@00000003oX`00
oj<000000P00on400`0000000>7o00000003002So`0600000000I_l006Ko00000000:?l000000P0D
0?l01@000000D@3o0540o`1A0?l000000080S`3o00D000000<`0o`3<0?l0c03o000000030?l0m@03
0000003o0;P0o`2h00<0o`2h00<000000?l0NP3o07X01`3o07X00`000000o`0m0?l0?@050?l0?@03
0000003o07X0o`1j00L0o`1j00<000000?l0^03o0;P00`3o0;P01@000000o`3e0?l0m@3o0?D00000
00<0c03o00D0000008l0o`2?0?l0S`3o000000020540o`040000000D0?l0503o0000008002So00H0
0000001Vo`00I_l00000002So`000002003Qo`0400000000on400?oQ000000800?nS00@00000003o
IP00ofH000000P00obP0100000005?l001Co00000002057o00050000002?o`00Sol008oo00000000
0P3<o`001@000000ooD00?oe003om@00000000<0okP000<000000?mj003oNP000`3oNP000`000000
ocd00?lm00090?lm008000000P3o00000`000000ocd00?lm000<0?lm00030000003oNP00ogX000@0
ogX000<000000?nh003o^0000P3o^002000000<0ooD000<000000<co003<o`000P3<o`001@000000
Sol008oo002?o`00000000<0DOl000@0000001Co000Do`005?l00P000003003o:08000000P00ofH2
000000<00?nS0P000003003oh@8000001000hOl4000000@00:?o10000002001Vo`L0oooo0@000001
0?ooo`00803oool00`000000oooo0?ooo`050?ooo`8000001000Xol4000000@00>7o10000003003o
h@8000000`00oj<300000003003oIP000000000000800?lX0`00000301Co008000000P1Ao`020000
0080Sol00P0000030<co008000000`3om@02000000<0okP00P0000050?mj008000004@3o?@000`00
0000ocd00?lm00090?lm00800000103oNP02000000<0okP000<000000?oe003om@000P3om@001000
0000c?l00<co0000000208oo00040000001Ao`00DOl0000000805?l000@00000003o:000obP00000
0`00ofH00`0000000?nS00000002003oh@0400000000hOl00>7o000000800:?o00@00000001Vo`00
I_l000000P00:?l010000000503o01@0o`0000020540o`040000002?0?l0S`3o000000<0c03o00<0
00000?l0m@3o0?D00P3o0?D00`000000o`2h0?l0^0030?l0^0030000003o07X0o`1j00L0o`1j00<0
00000?l0?@3o03d01P3o03d00`000000o`1j0?l0NP070?l0NP030000003o0;P0o`2h0080o`2h00D0
00000?l0m@3o0?D0o`3e000000030<`0o`050000002?0?l0S`3o08l0o`0000000P1A0?l010000000
503o01@0o`000002000Xo`0400000000I_l006Ko000000800:?o00@00000003Qo`00hOl000000P00
on400`0000000?nS00000003003oIP0400000000obP00?lX000000805?l000@00000057o001Ao`00
00000`2?o`0010000000c?l00<co000000030?oe00040000003o^000okP00?nh00800000103oNP02
000000/0ocd00P00000@0?lm008000001@3oNP02000000@0okP00P0000020?oe008000000`3<o`02
00000080Sol00P000002057o008000000`0Do`03000000800?lX0`0000001000ofH0000000000000
0002003oX`8000000`00on44000000@00>7o10000004002So`<000001P3oool100000040oooo000P
0?ooo`030000003oool0oooo00D0oooo1000Xol2000000D00>7o0`000003003oh@@0000000<00?nS
0000000000000P0000001000ofH00000000000000002003o:08000000P0Do`0400000004057o0000
0000000000000080Sol00P0000030<co008000000`3om@02000000@0okP000<000000?mj003oNP00
103oNP000`000000ocd00?lm000K0?lm00800000103oNP02000000<0okP00P0000030?oe00800000
00<0c?l00000000000000P2?o`0010000000DOl0057o0000000201Co000400000000obP00?lX0000
00800?mV0P0000001@00oj<00000003oh@00on40000000800>7o00<00000002So`00Xol00P000000
1@00I_l00000000Xo`00:?l000000080503o00D000000540o`1A0?l0D@3o0000000208l0o`050000
003<0?l0c03o0<`0o`000000103o0?D00`000000o`2h0?l0^0020?l0^0030000003o07X0o`1j00L0
o`1j00<000000?l0?@3o03d0203o03d00`000000o`1j0?l0NP060?l0NP030000003o0;P0o`2h00<0
o`2h00D000000?l0m@3o0?D0o`3e000000030<`0o`050000002?0?l0S`3o08l0o`0000000P1A0?l0
10000000503o01@0o`000002000Xo`0400000000I_l006Ko000000800:?o00@00000003Qo`00hOl0
00000P00on40100000000?nS000000000002003oIP0400000000obP00?lX000000805?l000@00000
057o001Ao`0000000`2?o`000`000000c?l00<co000200000080ooD00P0000030?nh00800000103o
NP02000001d0ocd000<000000?mj003oNP00103oNP02000000@0okP000@000000?oe003om@00ooD0
0`0000020<co008000000P2?o`0300000003057o000000000000008000000P0Do`02000000<00?lX
0P000002003oIP<0000000<00?nS0000000000000P000003003oh@<000001@00hOl2000000@00:?o
1`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo0P00Xol2000000D00>7o0P00
0003003oh@<000000`00oj<2000000@00?mV0P000003003o:0030000000Do`005?l0008000000`1A
o`02000000<0Sol00P0000040<co00040000003om@00ooD00?oe008000001@3o^0000`000000ogX0
0?mj00040?mj00030000003o?@00ocd001X0ocd00P0000050?mj00030000003o^000okP00080okP0
00<000000?oe003om@000P3om@0010000000c?l00<co0000000308oo00040000001Ao`00DOl00000
00805?l000L00000003o:000obP00?lX00000000ofH0000000800?nS00@00000003oh@00on400000
0`00hOl00`0000000:?o00000002001Vo`0700000000:?l002So000Xo`000000503o000000030540
o`050000002?0?l0S`3o08l0o`0000000`3<0?l01@000000o`3e0?l0m@3o0?D0000000@0o`2h00<0
00000?l0NP3o07X01`3o07X00`000000o`0m0?l0?@090?l0?@030000003o07X0o`1j00H0o`1j00<0
00000?l0^03o0;P00`3o0;P01@000000o`3e0?l0m@3o0?D0000000<0c03o00D0000008l0o`2?0?l0
S`3o000000020540o`050000000D0?l0503o01@0o`0000000P00:?l01P00000006Ko001Vo`000000
0:?o000000800>7o00D00000003oh@00on400?oQ00000002003oX`0300000000ofH0000000<00?lX
00@0000001Co000Do`0000000P1Ao`0200000080Sol000D000000<co003<o`00c?l0000000030?oe
00030000003o^000okP00080okP000<000000?mj003oNP000`3oNP02000001`0ocd000<000000?mj
003oNP001@3oNP000`000000okP00?nh00020?nh008000000`3om@000`000000c?l00<co00020<co
008000000`2?o`02000000<0DOl00P00000201Co008000000`00obP2000000<00?mV0P000003003o
X`<000000`00on42000000D00>7o0P000002002So`L0oooo0@0000010?ooo`00803oool00`000000
oooo0?ooo`050?ooo`8000001000hOl3000000800?oQ0`000004003oX`8000000`00ofH3000000<0
0?lX0P0000000`0Do`00000000000003057o008000000`2?o`02000000@0c?l00P0000030?oe0003
0000003o^000okP000<0okP00P0000050?mj00800000703o?@000`000000ogX00?mj00040?mj0003
0000003o^000okP00080okP000<000000?oe003om@000P3om@001@000000c?l00<co003<o`000000
0080Sol000@00000057o001Ao`0000000P0Do`001`0000000?lX003o:000obP00000003oIP000000
0`00oj<0100000000?oQ003oh@000002003Qo`0300000000Xol0000000<006Ko00@00000000Xo`00
:?l000000P0D0?l010000000D@3o0540o`00000308l0o`050000003<0?l0c03o0<`0o`000000103o
0?D00`000000o`2h0?l0^0020?l0^0030000003o07X0o`1j00H0o`1j00<000000?l0?@3o03d02`3o
03d00`000000o`1j0?l0NP060?l0NP030000003o0;P0o`2h0080o`2h00<000000?l0m@3o0?D00P3o
0?D01@000000c03o0<`0o`3<0?l0000000<0S`3o00@000000540o`1A0?l000000P0D0?l010000000
02So000Xo`000002001Vo`0400000000Xol00:?o000000800>7o00@00000003oh@00on4000000P00
oj<0100000000?mV003oIP000003003o:0040000000Do`005?l0000000<0DOl000@0000008oo002?
o`0000000P3<o`000`000000ooD00?oe00020?oe00030000003o^000okP00080okP000<000000?mj
003oNP00103oNP000`000000ocd00?lm000J0?lm008000001P3oNP02000000@0okP000@000000?oe
003om@00ooD00P0000040<co008000000`2?o`02000000<0DOl00P00000201Co008000000`00obP2
000000<00?mV0P000004003oX`<000000P00on43000000@00>7o0`0000060?ooo`4000000@3oool0
0200oooo00<000000?ooo`3oool01@3oool2003Qo`@000000`00on42000000<00?nS10000003003o
IP8000000`00obP30000000301Co0000000000000080DOl00`00000308oo008000000`3<o`030000
00<0ooD00P0000040?nh008000001P3oNP000`000000ocd00?lm000J0?lm008000001@3oNP020000
00<0okP00P0000040?oe00060000003<o`00c?l00<co00000000Sol00P000002057o00040000000D
o`005?l0000000<00?lX00@00000003oIP00ofH000000P00oj<0100000000?oQ003oh@000002003Q
o`0400000000Xol00:?o0000008006Ko00L00000000Xo`00:?l002So0000000D0?l0000000<0D@3o
00D0000008l0o`2?0?l0S`3o000000030<`0o`050000003o0?D0o`3e0?l0m@000000103o0;P00`00
0000o`1j0?l0NP070?l0NP030000003o03d0o`0m00/0o`0m00<000000?l0NP3o07X01P3o07X00`00
0000o`2h0?l0^0020?l0^0030000003o0?D0o`3e0080o`3e00D000000<`0o`3<0?l0c03o00000003
08l0o`040000001A0?l0D@3o000000<0503o00@00000000Xo`00:?l000000P00I_l0100000000:?o
002So`000002003Qo`0400000000on400?oQ000000800?nS00<00000003oIP00ofH00P000002003o
:0040000000Do`005?l0000000<0DOl000@0000008oo002?o`0000000P3<o`02000000<0ooD00P00
00030?nh008000001@3oNP02000001`0ocd000<000000?mj003oNP001@3oNP000`000000okP00?nh
00020?nh008000000`3om@03000000<0c?l00P00000308oo00<000000P1Ao`030000000401Co0000
00000000000000800?lX0P000003003oIP@000000`00oj<2000000<00?oQ10000002003Qo`L0oooo
0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`8000000`00on44000000800?nS0`00
0003003oIP@0000000<00?lX0000000000000P00000201Co008000000`1Ao`000`000000Sol008oo
000208oo008000000`3<o`02000000@0ooD00P0000040?nh008000001`3oNP000`000000ocd00?lm
000I0?lm008000001@3oNP02000000@0okP000<000000?oe003om@000P3om@0200000080c?l00P00
00000`2?o`000000057o0002057o00040000000Do`005?l0000000<00?lX00@00000003oIP00ofH0
00000P00oj<0100000000?oQ003oh@000002003Qo`0400000000Xol00:?o0000008006Ko0P000002
000Xo`040000000D0?l0503o00000080D@3o00D0000008l0o`2?0?l0S`3o000000030<`0o`030000
003o0?D0o`3e0080o`3e00<000000?l0^03o0;P00P3o0;P00`000000o`1j0?l0NP060?l0NP030000
003o03d0o`0m00d0o`0m00<000000?l0NP3o07X01P3o07X00`000000o`2h0?l0^0020?l0^0030000
003o0?D0o`3e0080o`3e00D000000<`0o`3<0?l0c03o0000000308l0o`040000001A0?l0D@3o0000
00<0503o00@00000000Xo`00:?l000000P00I_l0100000000:?o002So`000002003Qo`0400000000
on400?oQ000000800?nS00D00000003oIP00ofH00?mV00000002003o:0040000000Do`005?l00000
00<0DOl000<0000008oo002?o`000P0000020<co008000000`3om@000`000000okP00?nh00020?nh
008000001@3oNP02000001/0ocd00P0000060?mj00800000103o^002000000@0ooD00P0000030<co
00800000102?o`02000000<0DOl00P00000201Co00<0000000<00?lX0000000000000P000003003o
IP<000000P00oj<4000000<00?oQ0`0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3o
ool01@3oool2003oh@<000001000oj<2000000<00?mV0`000003003o:0800000100Do`02000000<0
DOl00P00000308oo00800000103<o`000`000000ooD00?oe00020?oe008000001@3o^0000`000000
ogX00?mj00060?mj00030000003o?@00ocd001P0ocd00P0000050?mj008000001@3o^00010000000
ooD00?oe003om@02000000<0c?l000@0000008oo002?o`0000000`1Ao`00100000005?l001Co0000
0003003o:00400000000ofH00?mV000000800?nS00@00000003oh@00on4000000P00hOl01`000000
0:?o002So`00Xol00000001Vo`0000000`00:?l010000000503o01@0o`0000020540o`050000002?
0?l0S`3o08l0o`0000000`3<0?l00`000000o`3e0?l0m@020?l0m@030000003o0;P0o`2h00<0o`2h
00<000000?l0NP3o07X01@3o07X00`000000o`0m0?l0?@0?0?l0?@030000003o07X0o`1j00D0o`1j
00<000000?l0^03o0;P00P3o0;P00`000000o`3e0?l0m@020?l0m@030000003<0?l0c03o0080c03o
00D0000008l0o`2?0?l0S`3o000000020540o`050000000D0?l0503o01@0o`0000000P00:?l00`00
000006Ko00000002002So`0500000000hOl00>7o003Qo`0000000P00on40100000000?nS003oX`00
0002003oIP0500000000obP00?lX003o:00000000P0Do`0200000080DOl000@0000008oo002?o`00
0000103<o`001@000000ooD00?oe003om@00000000D0okP00P0000050?mj008000006`3o?@000`00
0000ogX00?mj00050?mj00030000003o^000okP000<0okP00P0000040?oe00030000003<o`00c?l0
0080c?l00P00000408oo008000000`1Ao`02000000<05?l00P000003003o:0<000000`00ofH20000
00@00?nS0`000002003oh@L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`80
00001000oj<3000000800?mV0`000004003o:08000000`0Do`03000000<0DOl00P00000208oo00<0
0000103<o`02000000@0ooD000<000000?nh003o^0000`3o^002000000P0ogX000<000000?lm003o
?@00603o?@000`000000ogX00?mj00040?mj00030000003o^000okP000<0okP00P0000030?oe0003
0000003<o`00c?l00080c?l000D0000008oo002?o`00Sol000000002057o00040000000Do`005?l0
000000<00?lX00@00000003oIP00000000000P00oj<0100000000?oQ003oh@000003003Qo`060000
0000Xol00:?o00000000I_l000000`00:?l01@000000503o01@0o`0D0?l000000080D@3o00D00000
08l0o`2?0?l0S`3o000000030<`0o`030000003o0?D0o`3e0080o`3e00<000000?l0^03o0;P00`3o
0;P00`000000o`1j0?l0NP050?l0NP030000003o03d0o`0m00l0o`0m00<000000?l0NP3o07X01P3o
07X00`000000o`2h0?l0^0020?l0^0030000003o0?D0o`3e0080o`3e00D000000<`0o`3<0?l0c03o
0000000308l0o`040000001A0?l0D@3o000000<0503o00@00000000Xo`00:?l000000P00I_l01000
00000:?o002So`000003003Qo`0400000000on400?oQ000000800?nS0P0000000`00ofH00000003o
:002003o:0050000000Do`005?l001Co000000000P1Ao`0010000000Sol008oo000000040<co0004
0000003om@00ooD00?oe008000001@3o^0000`000000ogX00?mj00040?mj008000006P3o?@000`00
0000ogX00?mj00050?mj008000001@3o^0000`000000ooD00?oe00020?oe00800000103<o`030000
00<0Sol00P000003057o008000000`0Do`02000000@00?lX0`000002003oIP<000001000oj<30000
00H0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`800?nS10000003003oIP80
00000`00obP4000000<05?l00P000003057o00<000000P2?o`02000000@0c?l00`0000040?oe0080
0000103o^002000000P0ogX00P00000I0?lm008000001@3oNP02000000@0okP00P0000030?oe0080
00000`3<o`0200000080Sol00P0000000`1Ao`0000000000000201Co000300000000obP00?lX0080
000000<00?mV00000000oj<00P00oj<0100000000?oQ003oh@000003003Qo`0400000000Xol00:?o
0000008006Ko00D00000000Xo`00:?l002So0000000201@0o`040000001A0?l0D@3o000000<0S`3o
00D000000<`0o`3<0?l0c03o000000040?l0m@030000003o0;P0o`2h00<0o`2h00<000000?l0NP3o
07X01@3o07X00`000000o`0m0?l0?@0A0?l0?@030000003o07X0o`1j00D0o`1j00<000000?l0^03o
0;P00P3o0;P00`000000o`3e0?l0m@020?l0m@030000003<0?l0c03o0080c03o00D0000008l0o`2?
0?l0S`3o000000020540o`050000000D0?l0503o01@0o`0000000P00:?l01000000006Ko001Vo`00
0002002So`8000000P00hOl0100000000?oQ003oh@000003003oX`0400000000ofH0000000000080
0?lX0P00000201Co00030000001Ao`00DOl00080000000<0Sol00000000000000`3<o`02000000<0
ooD00P0000040?nh008000001P3oNP02000001T0ocd000<000000?mj003oNP001P3oNP02000000@0
okP00P0000040?oe00<00000103<o`0300000080Sol00`000002057o008000000`0Do`04000000<0
0?lX0P000003003oIP@000000P00oj<70?ooo`4000000@3oool00200oooo00<000000?ooo`3oool0
1@3oool2000000<00?mV10000002003o:0<000000`0Do`0400000080DOl00`00000308oo00800000
0`3<o`03000000D0ooD00P0000050?nh00030000003oNP00ogX000L0ogX000<000000?lm003o?@00
5@3o?@04000000D0ogX00P0000040?nh00800000103om@0010000000c?l00<co003<o`0200000080
Sol00P000002057o00050000000Do`005?l001Co000000000P00obP01@0000000?mV003oIP00ofH0
000000800?nS00@00000003oh@00on4000000`00hOl0100000000:?o002So`000002001Vo`050000
0000:?l002So000Xo`0000000P0D0?l200000080D@3o00D0000008l0o`2?0?l0S`3o000000030<`0
o`030000003o0?D0o`3e0080o`3e00<000000?l0^03o0;P00`3o0;P00`000000o`1j0?l0NP050?l0
NP030000003o03d0o`0m0140o`0m00<000000?l0NP3o07X01P3o07X00`000000o`2h0?l0^0020?l0
^0030000003o0?D0o`3e0080o`3e00<000000<`0o`3<0?l00P3<0?l010000000S`3o08l0o`000003
0540o`050000000D0?l0503o01@0o`0000000P00:?l01000000006Ko001Vo`000003002So`040000
0000hOl00>7o000000800?oQ00D00000003oX`00oj<00?nS00000002003oIP0500000000obP00?lX
003o:00000000P0Do`0200000080DOl000<0000008oo002?o`000P0000030<co008000000`3om@02
000000@0okP00P0000060?mj00<000005`3o?@000`000000ogX00?mj00070?mj00030000003o^000
okP000<0okP00`0000040?oe00<00000103<o`02000000<0Sol00P000002057o00@000000`0Do`03
000000800?lX10000003003oIP<000001P3oool100000040oooo000P0?ooo`030000003oool0oooo
00D0oooo0P00ofH3000000@00?lX0P00000301Co00<00000101Ao`02000000@0Sol00P0000030<co
008000001P3om@02000000D0okP00P00000:0?mj00030000003o?@00ocd00180ocd00`0000070?mj
00800000103o^002000000@0ooD00P0000030<co00050000002?o`00Sol008oo000000000`1Ao`00
1@0000005?l001Co000Do`00000000800?lX00H00000003oIP00ofH00?mV00000000oj<200000080
0?oQ00D00000003Qo`00hOl00>7o00000002002So`0500000000I_l006Ko001Vo`0000000P00:?l0
10000000503o01@0o`0000030540o`050000002?0?l0S`3o08l0o`0000000`3<0?l00`000000o`3e
0?l0m@020?l0m@030000003o0;P0o`2h00<0o`2h00<000000?l0NP3o07X01@3o07X00`000000o`0m
0?l0?@0B0?l0?@030000003o07X0o`1j00H0o`1j00<000000?l0^03o0;P00P3o0;P00`000000o`3e
0?l0m@030?l0m@050000003<0?l0c03o0<`0o`0000000`2?0?l01@000000D@3o0540o`1A0?l00000
00<0503o00<00000000Xo`0000000`00I_l0100000000:?o002So`000003003Qo`0300000000on40
0?oQ008000000P00oj<0100000000?mV003oIP000003003o:0050000000Do`005?l001Co00000000
0P1Ao`001@000000Sol008oo002?o`00000000@0c?l00P0000030?oe00800000103o^002000000L0
ogX00`00000D0?lm00030000003oNP00ogX000P0ogX00P0000060?nh008000001@3om@02000000@0
c?l00P00000308oo00800000101Ao`03000000<05?l00P000004003o:0<000000P00ofH70?ooo`40
00000@3oool00200oooo00<000000?ooo`3oool01@3oool2000000@00?lX0`00000201Co00<00000
101Ao`03000000@0Sol00P0000040<co00030000003om@00ooD000@0ooD00P0000060?nh00030000
003oNP00ogX000T0ogX000<000000?lm003o?@00403o?@03000000P0ogX00P0000040?nh00800000
103om@02000000<0c?l00P00000208oo008000000P1Ao`02000000805?l00P000002003o:0030000
0000ofH00?mV0080000000<00?nS00000000on400P00on401@0000000>7o003Qo`00hOl000000080
0:?o00D00000001Vo`00I_l006Ko00000002000Xo`040000000D0?l0503o000000<0D@3o00D00000
08l0o`2?0?l0S`3o000000040<`0o`030000003o0?D0o`3e0080o`3e00<000000?l0^03o0;P00`3o
0;P00`000000o`1j0?l0NP050?l0NP030000003o03d0o`0m01<0o`0m00<000000?l0NP3o07X01P3o
07X00`000000o`2h0?l0^0020?l0^0030000003o0?D0o`3e0080o`3e00<000000<`0o`3<0?l00P3<
0?l010000000S`3o08l0o`0000030540o`050000000D0?l0503o01@0o`0000000P00:?l01@000000
06Ko001Vo`00I_l0000000800:?o00D00000003Qo`00hOl00>7o00000003003oh@0400000000oj<0
0?nS000000800?mV0P000002003o:08000000P0Do`000`000000DOl0057o000200000080Sol00P00
00040<co00030000003om@00ooD00080ooD00P0000040?nh00<000001`3oNP0300000140ocd000<0
00000?mj003oNP002P3oNP000`000000okP00?nh00050?nh00030000003om@00ooD000@0ooD00P00
00040<co008000000`2?o`03000000@0DOl00`00000201Co00<000001000obP3000000H0oooo0@00
00010?ooo`00803oool00`000000oooo0?ooo`050?ooo`800?lX1000000301Co00800000101Ao`03
000000@0Sol00`0000040<co008000001@3om@02000000H0okP00P00000<0?mj00030000003o?@00
ocd000h0ocd00P0000080?mj00<00000103o^002000000@0ooD00P0000040<co00050000002?o`00
Sol008oo000000000`1Ao`001@0000005?l001Co000Do`00000000<00?lX00@00000003oIP00ofH0
00000P00oj<01@0000000?oQ003oh@00on40000000800>7o0P000002002So`0500000000I_l006Ko
001Vo`0000000P00:?l01@000000503o01@0o`0D0?l0000000<0D@3o00D0000008l0o`2?0?l0S`3o
000000030<`0o`030000003o0?D0o`3e0080o`3e00<000000?l0^03o0;P00`3o0;P00`000000o`1j
0?l0NP050?l0NP030000003o03d0o`0m01@0o`0m00<000000?l0NP3o07X01P3o07X00`000000o`2h
0?l0^0030?l0^0030000003o0?D0o`3e0080o`3e00D000000<`0o`3<0?l0c03o0000000308l0o`05
0000001A0?l0D@3o0540o`0000000`0D0?l00`00000002So000Xo`020000008006Ko00<00000002S
o`00Xol00P000002003Qo`0500000000on400?oQ003oh@0000000P00oj<01@0000000?mV003oIP00
ofH0000000<00?lX00@0000001Co000Do`0000000`1Ao`0010000000Sol008oo002?o`02000000<0
c?l00P0000040?oe008000001@3o^003000000L0ogX00P00000@0?lm00030000003oNP00ogX000X0
ogX00P0000060?nh008000001P3om@02000000@0c?l00P00000408oo00<00000101Ao`02000000<0
5?l010000002003o:0L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`800000
100Do`03000000@0DOl00P00000408oo00<000000`3<o`04000000@0ooD00`0000060?nh00800000
3P3oNP02000000P0ocd01`0000080?mj008000001P3o^0000`000000ooD00?oe00020?oe00800000
103<o`0200000080Sol00P000002057o008000000P0Do`02000000800?lX0P000002003oIP040000
0000oj<00?nS000000<00?oQ00@00000003Qo`00hOl000000`00Xol01@00000006Ko001Vo`00I_l0
0000008002So00D0000001@0o`0D0?l0503o000000030540o`050000002?0?l0S`3o08l0o`000000
103<0?l00`000000o`3e0?l0m@020?l0m@030000003o0;P0o`2h00<0o`2h00<000000?l0NP3o07X0
1@3o07X00`000000o`0m0?l0?@0E0?l0?@030000003o07X0o`1j00H0o`1j00<000000?l0^03o0;P0
0`3o0;P00`000000o`3e0?l0m@020?l0m@050000003<0?l0c03o0<`0o`0000000`2?0?l200000080
D@3o0P00000201@0o`0500000000:?l002So000Xo`0000000P00I_l01@0000000:?o002So`00Xol0
000000800>7o00D00000003oh@00on400?oQ00000002003oX`0500000000ofH00?mV003oIP000000
0`00obP00`0000005?l001Co000200000080DOl00P00000308oo008000000`3<o`02000000@0ooD0
0P0000060?nh00030000003oNP00ogX000H0ogX01`0000080?lm008000003P3oNP02000000H0okP0
0`0000050?oe00<000000`3<o`03000000@0Sol00P000004057o00@000000`0Do`03000000H0oooo
0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`805?l010000004057o00<00000102?
o`02000000@0c?l00P0000060?oe00800000203o^0000`000000ogX00?mj000@0?mj00P00000303o
NP03000000H0okP00P0000030?oe00800000103<o`0200000080Sol00P000002057o008000000`0D
o`001@0000000?lX003o:000obP0000000<00?mV00@00000003oX`00oj<000000P00on4200000080
0>7o00D00000002So`00Xol00:?o00000003001Vo`0400000000:?l002So000000<0503o00D00000
0540o`1A0?l0D@3o0000000308l0o`030000003<0?l0c03o0080c03o00<000000?l0m@3o0?D00P3o
0?D00`000000o`2h0?l0^0030?l0^0030000003o07X0o`1j00D0o`1j00<000000?l0?@3o03d05`3o
03d00`000000o`1j0?l0NP050?l0NP030000003o0;P0o`2h00<0o`2h00<000000?l0m@3o0?D00P3o
0?D00`000000c03o0<`0o`020<`0o`030000002?0?l0S`3o0080S`3o00D000000540o`1A0?l0D@3o
0000000201@0o`0400000000:?l002So000000<006Ko00D00000002So`00Xol00:?o00000002003Q
o`8000000P00on400`0000000?nS003oX`02000000800?mV0P000002003o:0040000000Do`005?l0
01Co008000000P1Ao`02000000<0Sol000<000000<co003<o`000P3<o`02000000@0ooD00P000005
0?nh00<00000303oNP0800000180ogX000<000000?nh003o^0001P3o^003000000D0ooD00P000004
0<co00800000102?o`03000000D0DOl00`00000201Co00L0oooo0@0000010?ooo`00803oool00`00
0000oooo0?ooo`050?ooo`800000101Ao`04000000@0Sol00`0000040<co008000001`3om@000`00
0000okP00?nh00060?nh008000008`3oNP04000000L0okP00P0000030?oe00800000103<o`020000
0080Sol00P000003057o00030000000Do`005?l000805?l000D00000003o:000obP00?lX00000003
003oIP0400000000oj<00?nS000000800?oQ00D00000003Qo`00hOl00>7o00000003002So`050000
0000I_l006Ko001Vo`0000000P00:?l01@000000503o01@0o`0D0?l0000000<0D@3o00D0000008l0
o`2?0?l0S`3o000000040<`0o`030000003o0?D0o`3e0080o`3e00<000000?l0^03o0;P00`3o0;P0
0`000000o`1j0?l0NP060?l0NP030000003o03d0o`0m01P0o`0m00<000000?l0NP3o07X01@3o07X0
0`000000o`2h0?l0^0030?l0^0030000003o0?D0o`3e0080o`3e00<000000<`0o`3<0?l00P3<0?l0
1@000000S`3o08l0o`2?0?l0000000<0D@3o00D0000001@0o`0D0?l0503o00000002000Xo`050000
0000I_l006Ko001Vo`0000000`00Xol01@0000000>7o003Qo`00hOl0000000800?oQ00D00000003o
X`00oj<00?nS00000003003oIP0400000000obP00?lX000000@05?l000@00000057o001Ao`00DOl0
0P00000208oo00800000103<o`02000000@0ooD00P0000060?nh00@000008`3oNP02000000T0okP0
00<000000?oe003om@00103om@02000000@0c?l00`00000408oo00@00000101Ao`03000000H0oooo
0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`<0DOl00`00000508oo00<000001@3<
o`02000000L0ooD00P0000080?nh00030000003oNP00ogX00200ogX00`0000080?nh00<000000`3o
m@02000000@0c?l00P00000208oo008000000`1Ao`02000000<05?l00P000002003o:08000000P00
ofH200000003003oX`000000000000800?oQ00D00000003Qo`00hOl00>7o00000003002So`030000
0000I_l006Ko008000000P00:?l01@000000503o01@0o`0D0?l0000000@0D@3o00D0000008l0o`2?
0?l0S`3o000000030<`0o`030000003o0?D0o`3e00<0o`3e00<000000?l0^03o0;P00`3o0;P00`00
0000o`1j0?l0NP050?l0NP030000003o03d0o`0m01T0o`0m00<000000?l0NP3o07X01@3o07X00`00
0000o`2h0?l0^0030?l0^0030000003o0?D0o`3e00<0o`3e00D000000<`0o`3<0?l0c03o00000004
08l0o`050000001A0?l0D@3o0540o`0000000`0D0?l00`00000002So000Xo`020000008006Ko0P00
0002002So`0500000000hOl00>7o003Qo`0000000P00on42000000800?nS0P000002003oIP800000
00<00?lX0000000000000`0Do`02000000<0DOl00P00000208oo00800000103<o`02000000@0ooD0
0`0000070?nh00<000008P3oNP000`000000okP00?nh00070?nh008000001P3om@02000000D0c?l0
0`00000508oo00<000000`1Ao`070?ooo`4000000@3oool00200oooo0P0000060?ooo`80DOl000<0
000008oo002?o`00102?o`02000000@0c?l0100000070?oe008000002P3o^0000`000000ogX00?mj
000K0?mj00D000001`3o^004000000<0ooD00`0000020<co00@000000P2?o`02000000@0DOl000@0
000001Co000Do`005?l00P000002003o:08000000P00ofH2000000800?nS00<00000003oh@00on40
0P000002003Qo`8000000P00Xol20000008006Ko00D00000000Xo`00:?l002So0000000301@0o`03
0000001A0?l0D@3o0080D@3o00D0000008l0o`2?0?l0S`3o000000040<`0o`030000003o0?D0o`3e
0080o`3e00<000000?l0^03o0;P00`3o0;P00`000000o`1j0?l0NP060?l0NP030000003o03d0o`0m
01T0o`0m00<000000?l0NP3o07X01P3o07X00`000000o`2h0?l0^0030?l0^0030000003o0?D0o`3e
00<0o`3e00D000000<`0o`3<0?l0c03o0000000408l0o`050000001A0?l0D@3o0540o`0000000`0D
0?l01@00000002So000Xo`00:?l0000000<006Ko00<00000002So`00Xol00P000002003Qo`800000
0P00on42000000800?nS0P000002003oIP0300000000obP00?lX008000000`0Do`02000000<0DOl0
0P00000208oo00@000000P3<o`03000000@0ooD00`0000070?nh00D000007@3oNP000`000000okP0
0?nh00090?nh00030000003om@00ooD000D0ooD0100000040<co008000001P2?o`0200000003057o
003oool0oooo00@0oooo0P0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`800000102?
o`03000000@0c?l00P0000090?oe00800000303o^0000`000000ogX00?mj000G0?mj00@000002@3o
^003000000D0ooD00P0000030<co00800000102?o`02000000@0DOl00P00000201Co008000000P00
obP2000000<00?mV00D00000003oX`00oj<00?nS00000002003oh@0500000000hOl00>7o003Qo`00
00000`00Xol01@00000006Ko001Vo`00I_l0000000<002So00D0000001@0o`0D0?l0503o00000004
0540o`050000002?0?l0S`3o08l0o`000000103<0?l00`000000o`3e0?l0m@020?l0m@030000003o
0;P0o`2h00<0o`2h00<000000?l0NP3o07X01P3o07X00`000000o`0m0?l0?@0K0?l0?@030000003o
07X0o`1j00D0o`1j00<000000?l0^03o0;P0103o0;P00`000000o`3e0?l0m@030?l0m@050000003<
0?l0c03o0<`0o`000000102?0?l01@000000D@3o0540o`1A0?l0000000<0503o00D00000000Xo`00
:?l002So00000003001Vo`0500000000Xol00:?o002So`0000000`00hOl01@0000000?oQ003oh@00
on40000000<00?nS00<00000003oIP00ofH00P000002003o:08000000`0Do`02000000<0DOl00P00
000408oo008000000`3<o`03000000@0ooD00`0000090?nh00@000006@3oNP000`000000okP00?nh
000:0?nh008000002@3om@02000000@0c?l00`00000508oo008000001P3oool100000040oooo000P
0?ooo`030000003oool0oooo00D0oooo0`2?o`03000000D0c?l00P00000:0?oe00030000003o^000
okP000`0okP000<000000?mj003oNP00503oNP03000000/0okP00P0000060?oe00800000103<o`00
0`000000Sol008oo000208oo008000000`1Ao`03000000805?l00P000002003o:08000000`00ofH2
000000800?nS0P000002003oh@0300000000hOl00>7o008000000P00Xol20000008006Ko0P000003
000Xo`050000000D0?l0503o01@0o`000000101A0?l00`000000S`3o08l0o`0208l0o`050000003<
0?l0c03o0<`0o`000000103o0?D00`000000o`2h0?l0^0040?l0^0030000003o07X0o`1j00H0o`1j
00<000000?l0?@3o03d06`3o03d00`000000o`1j0?l0NP060?l0NP030000003o0;P0o`2h00<0o`2h
00<000000?l0m@3o0?D00`3o0?D00`000000c03o0<`0o`020<`0o`030000002?0?l0S`3o0080S`3o
00D000000540o`1A0?l0D@3o0000000301@0o`0400000000:?l002So000Xo`8000000P00I_l20000
00800:?o0P000002003Qo`8000000P00on42000000800?nS00@00000003oIP00ofH00?mV0P000002
003o:08000000`0Do`02000000<0DOl00P00000408oo00800000103<o`02000000D0ooD00P00000;
0?nh00@000005@3oNP000`000000okP00?nh000<0?nh00030000003om@00ooD000P0ooD00P000005
0<co00<000000`2?o`070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool00`2?
o`000000000000070<co00030000003om@00ooD000P0ooD00P00000?0?nh00030000003oNP00ogX0
0100ogX01000000;0?nh00<000001@3om@03000000@0c?l00P00000308oo008000000P1Ao`030000
00<05?l00P000002003o:08000001000ofH01@0000000?nS003oX`00oj<0000000<00?oQ00@00000
003Qo`00hOl000000`00Xol01@00000006Ko001Vo`00I_l0000000@002So00D0000001@0o`0D0?l0
503o000000040540o`030000002?0?l0S`3o0080S`3o00D000000<`0o`3<0?l0c03o000000050?l0
m@030000003o0;P0o`2h00<0o`2h00<000000?l0NP3o07X01P3o07X00`000000o`0m0?l0?@0M0?l0
?@030000003o07X0o`1j00D0o`1j00<000000?l0^03o0;P0103o0;P00`000000o`3e0?l0m@030?l0
m@030000003<0?l0c03o0080c03o00D0000008l0o`2?0?l0S`3o000000040540o`050000000D0?l0
503o01@0o`0000001000:?l01@00000006Ko001Vo`00I_l0000000<00:?o00D00000003Qo`00hOl0
0>7o00000003003oh@0300000000oj<00?nS008000000`00ofH2000000800?lX0P00000301Co00<0
00000P1Ao`02000000@0Sol00P0000040<co008000001@3om@03000000`0okP01000000A0?mj0003
0000003o^000okP000d0okP00P00000:0?oe008000001P3<o`020000000308oo003oool0oooo00D0
oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`030000003<o`00c?l000@0c?l0
0`0000090?oe008000004@3o^00C000000/0okP0100000040?oe00@00000103<o`0300000080Sol0
0`000003057o008000000`0Do`02000000800?lX0P000004003oIP8000000P00oj<2000000800?oQ
0P000002003Qo`0300000000Xol00:?o00800:?o00D00000001Vo`00I_l006Ko00000003000Xo`05
0000000D0?l0503o01@0o`000000101A0?l00`000000S`3o08l0o`0208l0o`030000003<0?l0c03o
0080c03o00<000000?l0m@3o0?D00P3o0?D00`000000o`2h0?l0^0040?l0^0030000003o07X0o`1j
00H0o`1j00<000000?l0?@3o03d07P3o03d00`000000o`1j0?l0NP050?l0NP030000003o0;P0o`2h
00@0o`2h00<000000?l0m@3o0?D00`3o0?D00`000000c03o0<`0o`020<`0o`050000002?0?l0S`3o
08l0o`000000101A0?l01@000000503o01@0o`0D0?l0000000@002So00D00000001Vo`00I_l006Ko
00000003002So`0400000000hOl00>7o003Qo`8000000P00on40100000000?nS003oX`00oj<20000
00<00?mV0P000002003o:08000000`0Do`02000000<0DOl01000000208oo00<000000`3<o`040000
00@0ooD01000000<0?nh018000004@3o^002000000X0ooD00P0000060<co008000001P3oool10000
0040oooo000P0?ooo`030000003oool0oooo00D0oooo1@3<o`02000000/0ooD000<000000?nh003o
^000:`3o^004000000@0ooD0100000040<co00@000000`2?o`02000000<0DOl00`00000301Co0080
00000P00obP2000000@00?mV0P000002003oX`8000000P00on42000000<00>7o00@00000002So`00
Xol00:?o0P000002001Vo`8000000`00:?l01@000000503o01@0o`0D0?l0000000@0D@3o00<00000
08l0o`2?0?l00P2?0?l00`000000c03o0<`0o`020<`0o`030000003o0?D0o`3e0080o`3e00<00000
0?l0^03o0;P0103o0;P00`000000o`1j0?l0NP060?l0NP030000003o03d0o`0m0200o`0m00<00000
0?l0NP3o07X01@3o07X00`000000o`2h0?l0^0040?l0^0030000003o0?D0o`3e00<0o`3e00<00000
0<`0o`3<0?l00P3<0?l01@000000S`3o08l0o`2?0?l0000000@0D@3o00D0000001@0o`0D0?l0503o
00000004000Xo`8000000P00I_l2000000800:?o0P000003003Qo`80000000<00?oQ000000000000
0`00oj<00`0000000?mV003oIP02003oIP8000000P00obP2000000<05?l00`000004057o00800000
0`2?o`03000000@0c?l0100000040?oe00@00000;03o^0000`000000ooD00?oe000:0?oe00800000
1@3<o`070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool30<co008000003@3o
m@000`000000okP00?nh000T0?nh00L00000103om@04000000@0c?l01000000408oo00<000000`1A
o`02000000<05?l00`000002003o:08000001000ofH2000000800?nS0P000002003oh@8000001000
hOl00`0000000:?o002So`020000008006Ko0P000003000Xo`8000000P0D0?l2000000@0D@3o00<0
000008l0o`2?0?l00P2?0?l00`000000c03o0<`0o`020<`0o`030000003o0?D0o`3e0080o`3e00<0
00000?l0^03o0;P0103o0;P00`000000o`1j0?l0NP060?l0NP030000003o03d0o`0m0240o`0m00<0
00000?l0NP3o07X01@3o07X00`000000o`2h0?l0^0050?l0^0030000003o0?D0o`3e00<0o`3e00<0
00000<`0o`3<0?l00P3<0?l01@000000S`3o08l0o`2?0?l0000000@0D@3o0P00000201@0o`800000
1000:?l01@00000006Ko001Vo`00I_l0000000<00:?o00<00000003Qo`00hOl00P00hOl200000003
003oh@000000000000800?nS0P000004003oIP8000000P00obP3000000<05?l00P000004057o00<0
00000`2?o`04000000@0c?l0100000040?oe00P000009@3o^0000`000000ooD00?oe000;0?oe0080
00000`3<o`070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool20<co00030000
003om@00ooD000`0ooD000<000000?nh003o^0008@3o^004000000P0ooD00`0000050<co00<00000
1@2?o`03000000@0DOl00P00000201Co00<000000`00obP2000000@00?mV0P000002003oX`800000
0P00on42000000@00>7o0P000002002So`0500000000I_l006Ko001Vo`0000001000:?l01@000000
503o01@0o`0D0?l0000000@0D@3o0P00000308l0o`800000103<0?l00`000000o`3e0?l0m@030?l0
m@030000003o0;P0o`2h00@0o`2h00<000000?l0NP3o07X01P3o07X00`000000o`0m0?l0?@0R0?l0
?@030000003o07X0o`1j00D0o`1j00<000000?l0^03o0;P0103o0;P00`000000o`3e0?l0m@040?l0
m@030000003<0?l0c03o0080c03o00@0000008l0o`2?0?l0S`3o0P0000040540o`050000000D0?l0
503o01@0o`0000001000:?l01000000006Ko001Vo`00I_l2000000800:?o0P000004003Qo`030000
0000on400?oQ008000000P00oj<2000000@00?mV0P000003003o:0<000000P0Do`02000000D0DOl0
0P00000508oo00<000001@3<o`03000000T0ooD00`00000R0?nh00030000003om@00ooD000d0ooD0
00<000000<co003<o`001`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo0P00
000?0?oe00030000003o^000okP001T0okP0200000080?oe00@000001`3<o`02000000D0Sol00P00
0003057o00@000000`0Do`02000000800?lX0`000002003oIP@000000P00oj<2000000800?oQ0P00
0004003Qo`8000000P00Xol2000000<006Ko00@00000000Xo`00:?l002So0P00000301@0o`030000
001A0?l0D@3o0080D@3o00<0000008l0o`2?0?l00P2?0?l00`000000c03o0<`0o`030<`0o`030000
003o0?D0o`3e00<0o`3e00<000000?l0^03o0;P0103o0;P00`000000o`1j0?l0NP060?l0NP030000
003o03d0o`0m02<0o`0m00<000000?l0NP3o07X01P3o07X00`000000o`2h0?l0^0040?l0^0030000
003o0?D0o`3e00<0o`3e00<000000<`0o`3<0?l00`3<0?l00`000000S`3o08l0o`0208l0o`030000
001A0?l0D@3o0080D@3o00@0000001@0o`0D0?l0503o0P000003000Xo`0300000000I_l006Ko0080
06Ko0P000002002So`8000000`00hOl2000000800?oQ0P000002003oX`@000000P00ofH3000000<0
0?lX0P00000201Co00D000000P1Ao`02000000D0Sol00P0000070<co00D000001`3om@08000001/0
okP000<000000?oe003om@003@3om@03000000H0oooo0@0000010?ooo`00803oool00`000000oooo
0?ooo`050?oooa80ooD00`00000@0?nh00P00000303om@04000000L0c?l01000000408oo00<00000
0P1Ao`03000000@05?l00`0000001000obP00000000000000003003oIP8000001000oj<200000080
0?oQ0P000004003Qo`8000000P00Xol2000000<006Ko0P000002000Xo`8000000`0D0?l2000000@0
D@3o00<0000008l0o`2?0?l00P2?0?l00`000000c03o0<`0o`030<`0o`030000003o0?D0o`3e0080
o`3e0P0000060?l0^0030000003o07X0o`1j00H0o`1j00<000000?l0?@3o03d09@3o03d00`000000
o`1j0?l0NP060?l0NP030000003o0;P0o`2h00@0o`2h0P0000040?l0m@030000003<0?l0c03o00<0
c03o00<0000008l0o`2?0?l00P2?0?l00`000000D@3o0540o`020540o`8000000`0D0?l200000080
02So0P000004001Vo`8000000P00Xol2000000<00>7o0P000002003oh@8000001000oj<2000000<0
0?mV0`000002003o:0<00000100Do`0200000080DOl00`00000408oo00@00000203<o`04000000/0
ooD02000000@0?nh00<000004P3om@070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool0
1@3ooolE0?oe01000000403om@04000000L0c?l01000000408oo00@000000`1Ao`02000000<05?l0
10000002003o:08000000`00ofH3000000@00?nS0P000002003oh@8000000`00hOl3000000800:?o
0P000003001Vo`8000000P00:?l2000000@0503o00<000000540o`1A0?l00P1A0?l2000000<0S`3o
0P0000050<`0o`030000003o0?D0o`3e0080o`3e00<000000?l0^03o0;P01@3o0;P00`000000o`1j
0?l0NP070?l0NP030000003o03d0o`0m02D0o`0m00<000000?l0NP3o07X01`3o07X00`000000o`2h
0?l0^0050?l0^0030000003o0?D0o`3e0080o`3e0P0000040<`0o`8000000`2?0?l2000000@0D@3o
00<0000001@0o`0D0?l00P0D0?l20000008002So0P000004001Vo`8000000P00Xol2000000<00>7o
0P000002003oh@8000001000oj<3000000<00?mV0P000003003o:0@000000P0Do`02000000<0DOl0
1000000408oo00@00000203<o`0300000100ooD04000000E0?oe00L0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooob`0ooD02@0000030<co00P000000P2?o`0600000080DOl01@00
000201Co00<000000`00obP3000000800?mV0`000003003oX`@0000000@00?oQ0000000000000000
0P00hOl3000000<00:?o0P000003001Vo`8000000`00:?l00`000000503o01@0o`0201@0o`800000
0`1A0?l2000000@0S`3o00<000000<`0o`3<0?l00`3<0?l2000000@0o`3e00<000000?l0^03o0;P0
1@3o0;P00`000000o`1j0?l0NP060?l0NP800000:@3o03d2000000P0o`1j00<000000?l0^03o0;P0
1@3o0;P00`000000o`3e0?l0m@030?l0m@800000103<0?l00`000000S`3o08l0o`0208l0o`800000
0`1A0?l2000000@0503o0P000002000Xo`<000000`00I_l2000000800:?o0`000002003Qo`<00000
00<00?oQ0000000000000P000003003oX`<000000P00ofH4000000<00?lX0`0000000`0Do`000000
0000000300000080DOl01P00000208oo00P000000`3<o`09000002`0ooD01`3oool100000040oooo
000P0?ooo`030000003oool0oooo00D0oooo9@3om@07000000T0c?l00`00000808oo008000001P1A
o`02000000D05?l00P000003003o:0<000000`00ofH2000000@00?nS0P000003003oh@8000001000
hOl2000000<00:?o0P000003001Vo`8000001000:?l010000000503o01@0o`0D0?l2000000<0D@3o
0P00000508l0o`030000003<0?l0c03o00<0c03o00<000000?l0m@3o0?D00`3o0?D00`000000o`2h
0?l0^0050?l0^0030000003o07X0o`1j00H0o`1j00<000000?l0?@3o03d0:`3o03d00`000000o`1j
0?l0NP060?l0NP030000003o0;P0o`2h00D0o`2h00<000000?l0m@3o0?D0103o0?D00`000000c03o
0<`0o`020<`0o`030000002?0?l0S`3o00<0S`3o0P0000030540o`8000000`0D0?l00`00000002So
000Xo`03000Xo`8000000`00I_l0100000000:?o002So`00Xol2000000@00>7o0P000003003oh@80
00001000oj<2000000@00?mV0`000003003o:0800000100Do`02000000H0DOl00P00000808oo00<0
00002@3<o`07000002D0ooD01`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo
0P00000L0?oe00L00000303<o`04000000P0Sol00`000004057o00@00000100Do`03000000800?lX
0`000004003oIP8000001000oj<2000000<00?oQ0P000003003Qo`<000000`00Xol2000000@006Ko
00<00000000Xo`00:?l00P00:?l2000000<0503o00<000000540o`1A0?l00P1A0?l00`000000S`3o
08l0o`0408l0o`030000003<0?l0c03o00<0c03o00<000000?l0m@3o0?D0103o0?D00`000000o`2h
0?l0^0040?l0^0030000003o07X0o`1j00H0o`1j00<000000?l0?@3o03d04`3o03d5000001@0o`0m
00<000000?l0NP3o07X01`3o07X00`000000o`2h0?l0^0050?l0^0030000003o0?D0o`3e00<0o`3e
00<000000<`0o`3<0?l00`3<0?l00`000000S`3o08l0o`0408l0o`030000001A0?l0D@3o0080D@3o
00@0000001@0o`0D0?l0503o0P000005000Xo`0400000000I_l006Ko001Vo`8000000`00Xol30000
00<00>7o0P000003003oh@8000001000oj<3000000@00?mV0P000003003o:0800000100Do`040000
00@0DOl01000000708oo00@00000303<o`08000001`0ooD00P0000060?ooo`4000000@3oool00200
oooo00<000000?ooo`3oool01@3oool20<co00/000001@3om@0<000000/0c?l02000000408oo00P0
00000`1Ao`04000000@05?l010000003003o:08000000`00ofH4000000<00?nS0`0000000`00on40
000000000002000000<00>7o0P000003002So`<000001000I_l2000000<002So0P00000301@0o`80
00000`1A0?l2000000D0S`3o0P0000050<`0o`030000003o0?D0o`3e00<0o`3e0P0000060?l0^003
0000003o07X0o`1j00L0o`1j00<000000?l0?@3o03d04@3o03d2000000D0o`000P00000C0?l0?@03
0000003o07X0o`1j00L0o`1j00<000000?l0^03o0;P01@3o0;P00`000000o`3e0?l0m@030?l0m@80
0000103<0?l2000000D0S`3o0P0000030540o`8000000`0D0?l3000000<002So0P000003001Vo`<0
00000`00Xol2000000<00>7o100000001000on400000000000000004003oX`@000000P00ofH20000
00<00?lX1000000401Co00@00000101Ao`07000000@0Sol02000000<0<co00/000001@3om@0<0000
00030<co003oool0oooo00D0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`d0
c?l01@00000?0<co00P000001`2?o`05000000@0DOl01`00000301Co00D000000P00obP500000003
003oIP0000000000008000000`00oj<4000000800?oQ0P000003003Qo`@000000`00Xol2000000<0
06Ko10000002000Xo`<000000`0D0?l2000000@0D@3o00<0000008l0o`2?0?l00`2?0?l2000000D0
c03o0P0000050?l0m@030000003o0;P0o`2h00D0o`2h00<000000?l0NP3o07X01`3o07X00`000000
o`0m0?l0?@0A0?l0?@030000003o0000o`0000L0o`000P00000B0?l0?@030000003o07X0o`1j00L0
o`1j00<000000?l0^03o0;P01@3o0;P00`000000o`3e0?l0m@040?l0m@800000103<0?l2000000D0
S`3o00<000000540o`1A0?l00P1A0?l2000000@0503o0P000003000Xo`<000000`00I_l2000000<0
0:?o10000003003Qo`8000000P00on45000000<00?nS0`0000000`00ofH000000000000300000080
0?lX1@00000301Co00P000000`1Ao`05000000L0Sol02000000?0<co00D000003@3<o`070?ooo`40
00000@3oool00200oooo00<000000?ooo`3oool01@3ooolJ0<co00L00000202?o`07000000D0DOl0
1000000701Co00<000001@00obP2000000800?mV10000004003oX`<000001000on42000000<00>7o
0P000005002So`8000000`00I_l2000000D002So0P00000401@0o`030000001A0?l0D@3o00<0D@3o
00<0000008l0o`2?0?l00P2?0?l2000000H0c03o00<000000?l0m@3o0?D0103o0?D00`000000o`2h
0?l0^0050?l0^0030000003o07X0o`1j00P0o`1j00<000000?l0?@3o03d0403o03d00`000000o`00
0?l000090?l000030000003o03d0o`0m0100o`0m00<000000?l0NP3o07X0203o07X00`000000o`2h
0?l0^0050?l0^08000001@3o0?D00`000000c03o0<`0o`040<`0o`800000102?0?l00`000000D@3o
0540o`030540o`800000100D0?l2000000@002So0P000003001Vo`8000001@00Xol2000000<00>7o
0P000005003oh@<000000`00oj<4000000800?mV0P000005003o:0<00000200Do`03000000D0DOl0
1`00000808oo00P000006@3<o`070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3o
ool>0<co00`000002`2?o`04000000P0DOl01000000401Co00L000001000obP8000000@00?nS1000
0004003oh@<000000`00hOl2000000@00:?o0`000003001Vo`8000001000:?l3000000@0503o0P00
00040540o`800000102?0?l00`000000c03o0<`0o`050<`0o`030000003o0?D0o`3e00@0o`3e00<0
00000?l0^03o0;P01@3o0;P00`000000o`1j0?l0NP080?l0NP030000003o03d0o`0m0100o`0m00<0
00000?l0003o00002`3o00000`000000o`0m0?l0?@0@0?l0?@030000003o07X0o`1j00P0o`1j00<0
00000?l0^03o0;P01P3o0;P00`000000o`3e0?l0m@030?l0m@030000003<0?l0c03o00D0c03o00<0
000008l0o`2?0?l00P2?0?l2000000D0D@3o0P00000401@0o`8000001000:?l2000000<006Ko0`00
0004002So`8000000`00hOl4000000@00?oQ0`000004003oX`P000001000obP8000000<05?l01000
0008057o00@00000302?o`0;000000h0c?l01`3oool100000040oooo000P0?ooo`030000003oool0
oooo00D0oooo3P00000@08oo00L00000101Ao`08000000@05?l010000004003o:0/000001000oj<4
000000@00?oQ10000003003Qo`<000001000Xol2000000<006Ko0`000004000Xo`800000100D0?l3
000000@0D@3o0P00000408l0o`8000001P3<0?l2000000D0o`3e0P0000070?l0^0030000003o07X0
o`1j00P0o`1j00<000000?l0?@3o03d0403o03d00`000000o`000?l0000=0?l0008000004@3o03d2
000000T0o`1j0P0000070?l0^0030000003o0?D0o`3e00<0o`3e0P0000060<`0o`800000102?0?l3
000000@0D@3o0`00000301@0o`8000001000:?l3000000<006Ko0P000004002So`<000001000hOl4
000000<00?oQ10000004003oX``000001000obP3000000@05?l020000004057o00P000003`2?o`0?
000000H0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oooa80Sol030000006057o
00D000001`0Do`05000000@00?lX10000003003oIP@000001000oj<4000000@00?oQ10000003003Q
o`@000000`00Xol40000008006Ko0`000003000Xo`@00000100D0?l2000000D0D@3o0P00000408l0
o`8000001P3<0?l2000000H0o`3e00<000000?l0^03o0;P01P3o0;P00`000000o`1j0?l0NP080?l0
NP030000003o03d0o`0m0100o`0m00<000000?l0003o0000403o00000`000000o`0m0?l0?@0@0?l0
?@030000003o07X0o`1j00P0o`1j00<000000?l0^03o0;P01@3o0;P2000000D0o`3e0P0000060<`0
o`8000001@2?0?l2000000D0D@3o00<0000001@0o`0D0?l00P0D0?l4000000<002So0`000002001V
o`@000000`00Xol4000000@00>7o0`000004003oh@@000001000oj<4000000@00?mV10000003003o
:0D000001`0Do`05000000L0DOl02`00000B08oo00L0oooo0@0000010?ooo`00803oool00`000000
oooo0?ooo`050?ooo`H0Sol03000000<057o00H000001@0Do`07000000D00?lX10000004003oIP<0
00001000oj<4000000@00?oQ10000004003Qo`<000001000Xol3000000@006Ko0P000004000Xo`80
00001P0D0?l2000000D0D@3o0P00000408l0o`8000001P3<0?l2000000L0o`3e00<000000?l0^03o
0;P01P3o0;P00`000000o`1j0?l0NP080?l0NP030000003o03d0o`0m0100o`0m00<000000?l0003o
00004@3o00000`000000o`0m0?l0?@0@0?l0?@030000003o07X0o`1j00T0o`1j00<000000?l0^03o
0;P01P3o0;P00`000000o`3e0?l0m@040?l0m@8000001P3<0?l2000000D0S`3o0P0000040540o`80
00001P0D0?l2000000@002So0P000004001Vo`<000001000Xol4000000<00>7o10000004003oh@@0
00001000oj<4000000@00?mV0`000005003o:0L000001@0Do`07000000/0DOl03000000608oo00L0
oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`H00000401Ao`08000000<05?l0
20000004003o:0P000001000ofH4000000<00?nS10000004003oh@@000001000hOl4000000@00:?o
0`000004001Vo`<000001000:?l2000000H0503o0P0000040540o`<00000102?0?l2000000H0c03o
0P0000070?l0m@800000203o0;P00`000000o`1j0?l0NP080?l0NP030000003o03d0o`0m0100o`0m
00<000000?l0003o00004`3o00000`000000o`0m0?l0?@0@0?l0?@030000003o07X0o`1j00T0o`1j
00<000000?l0^03o0;P01P3o0;P00`000000o`3e0?l0m@050?l0m@8000001P3<0?l2000000D0S`3o
0`0000030540o`8000001P0D0?l2000000@002So0`000004001Vo`@000001000Xol3000000@00>7o
10000004003oh@@000001000oj<4000000<00?mV20000004003o:0P00000100Do`0700000100DOl0
1`0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool:057o00`00000200D
o`03000000P00?lX10000008003oIP@000001000oj<3000000@00?oQ10000004003Qo`@000001000
Xol4000000<006Ko10000004000Xo`<000001@0D0?l3000000@0D@3o0P00000508l0o`8000001P3<
0?l2000000P0o`3e00<000000?l0^03o0;P01P3o0;P2000000X0o`1j00<000000?l0?@3o03d0403o
03d00`000000o`000?l0000E0?l000030000003o03d0o`0m0100o`0m0P00000:0?l0NP8000001`3o
0;P2000000L0o`3e0P0000060<`0o`8000001P2?0?l2000000<0D@3o0`00000501@0o`<000001000
:?l4000000@006Ko10000003002So`@000001000hOl4000000@00?oQ10000004003oX`<000002000
ofH4000000P00?lX1000000701Co00`000002P1Ao`070?ooo`4000000@3oool00200oooo00<00000
0?ooo`3oool01@3oool:000000P05?l030000003003o:0P000001000ofH8000000<00?nS1@000003
003oh@@000001000hOl4000000@00:?o10000004001Vo`<000001000:?l4000000@0503o10000004
0540o`<000001@2?0?l2000000H0c03o0P0000080?l0m@8000001`3o0;P2000000X0o`1j0P00000B
0?l0?@030000003o0000o`0001L0o`0000<000000?l0?@3o03d04@3o03d00`000000o`1j0?l0NP09
0?l0NP030000003o0;P0o`2h00H0o`2h0P0000070?l0m@8000001P3<0?l2000000H0S`3o0`000003
0540o`@00000100D0?l4000000@002So10000004001Vo`<000001000Xol4000000@00>7o10000004
003oh@@000000`00oj<8000000@00?mV20000004003o:0/00000200Do`0;000000H0oooo0@000001
0?ooo`00803oool2000000H0oooo2P0Do`08000000P00?lX1`000004003oIPP000001000oj<70000
00D00?oQ0`000004003Qo`@000001000Xol4000000@006Ko10000003000Xo`@000001@0D0?l30000
00D0D@3o0`00000608l0o`8000001P3<0?l2000000P0o`3e0P0000070?l0^08000002`3o07X00`00
0000o`0m0?l0?@0B0?l0?@030000003o0000o`0001P0o`0000<000000?l0?@3o03d04@3o03d00`00
0000o`1j0?l0NP090?l0NP800000203o0;P00`000000o`3e0?l0m@060?l0m@8000001P3<0?l20000
00L0S`3o0P0000050540o`<000001@0D0?l4000000@002So10000003001Vo`@000001000Xol40000
00@00>7o10000004003oh@L000001000oj<8000000@00?mV20000007003o:0P000002P0Do`060?oo
o`8000000@3oool00200oooo00<000000?ooo`3oool01@3oool:000000P00?lX20000007003oIP@0
00002000oj<4000000@00?oQ20000003003Qo`@000001000Xol4000000@006Ko10000004000Xo`<0
00001`0D0?l2000000H0D@3o0P00000708l0o`8000001P3<0?l2000000P0o`3e0P0000080?l0^003
0000003o07X0o`1j00T0o`1j0P00000D0?l0?@030000003o0000o`0001X0o`0000<000000?l0?@3o
03d04@3o03d2000000/0o`1j00<000000?l0^03o0;P01P3o0;P2000000P0o`3e0P0000060<`0o`<0
00001P2?0?l2000000H0D@3o0P00000701@0o`@000001000:?l3000000@006Ko10000004002So`@0
00001000hOl7000000@00?oQ10000008003oX`@000002000ofH7000000P00?lX2`0000060?ooo`40
00000@3oool00200oooo00<000000?ooo`3oool01@3oool:003o:0P000002000ofH7000000@00?nS
20000004003oh@@000002000hOl3000000@00:?o10000004001Vo`@000001@00:?l3000000L0503o
0`0000050540o`<000001P2?0?l3000000H0c03o0P0000080?l0m@800000203o0;P2000000/0o`1j
00<000000?l0?@3o03d04`3o03d00`000000o`000?l0000K0?l000030000003o03d0o`0m01<0o`0m
00<000000?l0NP3o07X02@3o07X2000000P0o`2h0P0000080?l0m@8000001`3<0?l3000000D0S`3o
0`0000050540o`<00000200D0?l3000000@002So10000004001Vo`@000001000Xol4000000L00>7o
10000004003oh@P000001000oj<8000000L00?mV2000000:003o:0L0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooo`X000002000ofH8000000@00?nS1`000004003oh@P000001000
hOl8000000<00:?o10000004001Vo`@000001000:?l5000000H0503o100000040540o`@000001`2?
0?l2000000H0c03o0`0000080?l0m@8000002@3o0;P00`000000o`1j0?l0NP0;0?l0NP030000003o
03d0o`0m0180o`0m00<000000?l0003o00007@3o00000`000000o`0m0?l0?@0C0?l0?@030000003o
07X0o`1j00X0o`1j00<000000?l0^03o0;P01`3o0;P2000000P0o`3e0`0000070<`0o`8000001P2?
0?l4000000@0D@3o1000000701@0o`@000001000:?l4000000@006Ko10000004002So`L000001000
hOl8000000@00?oQ20000004003oX`L000002000ofH;000000H0oooo0@0000010?ooo`00803oool0
0`000000oooo0?ooo`050?ooo`H00?mV30000004003oX`P000001`00on44000000P00>7o10000004
002So`L000001000I_l4000000@002So1000000801@0o`<00000101A0?l4000000P0S`3o0`000006
0<`0o`800000203o0?D3000000X0o`2h00<000000?l0NP3o07X02P3o07X2000001@0o`0m00<00000
0?l0003o00007`3o0002000001@0o`0m0P00000;0?l0NP030000003o0;P0o`2h00P0o`2h0`000008
0?l0m@<000001P3<0?l2000000P0S`3o100000040540o`@000001`0D0?l4000000@002So10000004
001Vo`P000000`00Xol4000000P00>7o10000008003oh@L000001000oj<<000000H00?mV1`3oool1
00000040oooo000P0?ooo`030000003oool0oooo00D0oooo1P000008003oX`P000002000on470000
00@00>7o20000004002So`@000001`00I_l4000000@002So1000000801@0o`@000000`1A0?l40000
00P0S`3o100000070<`0o`8000001`3o0?D3000000/0o`2h0P00000<0?l0NP030000003o03d0o`0m
01<0o`0m00<000000?l0003o00008P3o00000`000000o`0m0?l0?@0C0?l0?@030000003o07X0o`1j
00T0o`1j0P00000;0?l0^0<00000203o0?D2000000H0c03o1000000808l0o`@00000101A0?l40000
00L0503o10000004000Xo`@000002000I_l3000000@00:?o20000004003Qo`P000001`00on480000
00P00?nS1`0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool6003oX`P0
00002000on48000000@00>7o1`000005002So`L000001@00I_l6000000@002So1000000801@0o`@0
00001@1A0?l2000000T0S`3o0`0000080<`0o`<000001`3o0?D2000000`0o`2h0P00000<0?l0NP80
00005@3o03d00`000000o`000?l0000S0?l000030000003o03d0o`0m01@0o`0m00<000000?l0NP3o
07X02P3o07X3000000/0o`2h0P0000080?l0m@800000203<0?l4000000P0S`3o0`0000050540o`<0
0000200D0?l4000000@002So1`000004001Vo`L000001@00Xol7000000D00>7o1`000008003oh@P0
00001P00oj<70?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool6000000P00?oQ
20000008003Qo`@000001`00Xol5000000L006Ko1@000006000Xo`@00000200D0?l4000000@0D@3o
1@00000608l0o`D000001`3<0?l4000000P0o`3e0P00000<0?l0^08000003@3o07X00`000000o`0m
0?l0?@0D0?l0?@030000003o0000o`0002D0o`0000<000000?l0?@3o03d0503o03d00`000000o`1j
0?l0NP0<0?l0NP8000002`3o0;P3000000L0o`3e100000080<`0o`@000001`2?0?l5000000<0D@3o
1000000801@0o`@000001`00:?l4000000L006Ko1@000007002So`D000001`00hOl8000000P00?oQ
1`0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool6003oh@P000002000
hOl8000000<00:?o20000004001Vo`P000002000:?l3000000@0503o200000040540o`@000001`2?
0?l4000000P0c03o100000090?l0m@<000002`3o0;P3000000h0o`1j00<000000?l0?@3o03d0503o
03d00`000000o`000?l0000W0?l000030000003o03d0o`0m01@0o`0m0P00000>0?l0NP800000303o
0;P3000000P0o`3e100000080<`0o`@00000202?0?l3000000@0D@3o2000000401@0o`@000001`00
:?l8000000@006Ko20000004002So`L000002000hOl8000000H00?oQ1`3oool100000040oooo000P
0?ooo`030000003oool0oooo00D0oooo1P000008003Qo`P000002000Xol3000000P006Ko10000008
000Xo`P000000`0D0?l4000000P0D@3o1000000808l0o`<000002@3<0?l3000000/0o`3e0P00000;
0?l0^0<000003`3o07X2000001H0o`0m00<000000?l0003o0000:@3o00000`000000o`0m0?l0?@0E
0?l0?@8000003P3o07X3000000`0o`2h0P00000:0?l0m@<000002@3<0?l4000000L0S`3o10000008
0540o`@00000100D0?l7000000P002So10000008001Vo`@000001`00Xol8000000P00>7o1`000006
0?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool6003Qo`P000002000Xol80000
00@006Ko1`000008000Xo`@00000200D0?l3000000T0D@3o0`00000908l0o`<000002P3<0?l20000
00`0o`3e0P00000;0?l0^0800000403o07X2000001L0o`0m00<000000?l0003o0000:`3o00000`00
0000o`0m0?l0?@0F0?l0?@030000003o07X0o`1j00h0o`1j0P00000<0?l0^08000002`3o0?D20000
00/0c03o0`00000808l0o`<000002@1A0?l3000000P0503o10000008000Xo`L000001@00I_l70000
00P00:?o20000006003Qo`L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`H0
00002000Xol8000000P006Ko10000007000Xo`P00000200D0?l4000000L0D@3o1@00000708l0o`D0
00002P3<0?l3000000T0o`3e1@00000:0?l0^0<00000403o07X2000001L0o`0m0P00000_0?l00080
00005`3o03d200000100o`1j0`00000;0?l0^0@000002@3o0?D3000000/0c03o1000000708l0o`D0
00001`1A0?l5000000L0503o20000007000Xo`D000001`00I_l8000000P00:?o1`0000060?ooo`40
00000@3oool00200oooo00<000000?ooo`3oool01@3oool6002So`P000002000I_l8000000L002So
1000000801@0o`P000001`1A0?l4000000P0S`3o1000000;0<`0o`@000002@3o0?D3000000d0o`2h
0P00000A0?l0NP800000603o03d00`000000o`000?l0000a0?l000030000003o03d0o`0m01H0o`0m
0P00000A0?l0NP<00000303o0;P3000000T0o`3e1000000;0<`0o`@00000202?0?l4000000P0D@3o
1`00000801@0o`@000002000:?l7000000P006Ko20000006002So`L0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooo`H000002000I_l8000000P002So1`00000801@0o`@00000301A
0?l3000000P0S`3o1000000<0<`0o`<000002`3o0?D2000000h0o`2h0P00000A0?l0NP8000006@3o
03d00`000000o`000?l0000c0?l000030000003o03d0o`0m01L0o`0m0P00000B0?l0NP8000003@3o
0;P2000000/0o`3e1000000;0<`0o`@00000202?0?l4000000/0D@3o1000000801@0o`P000001`00
:?l8000000P006Ko1`0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool:
001Vo`@000003000:?l4000000/0503o1000000<0540o`@00000202?0?l3000000d0c03o0`00000<
0?l0m@8000003P3o0;P200000140o`1j0P00000K0?l0?@030000003o0000o`0003@0o`0000<00000
0?l0?@3o03d0603o03d200000180o`1j0P00000=0?l0^08000003@3o0?D3000000`0c03o0`000009
08l0o`@000002`1A0?l4000000`0503o0`00000<000Xo`@000002P00I_l70?ooo`4000000@3oool0
0200oooo00<000000?ooo`3oool01@3oool2001Vo`P000002000:?l8000000L0503o200000080540
o`P000001`2?0?l5000000/0c03o1@00000:0?l0m@D000003P3o0;P200000140o`1j0P00000K0?l0
?@800000>03o00000`000000o`0m0?l0?@0I0?l0?@8000004P3o07X00`000000o`2h0?l0^00<0?l0
^0D000002`3o0?D4000000/0c03o1@00000808l0o`L00000201A0?l8000000L0503o20000008000X
o`P000000P00I_l70?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool2000000`0
02So1000000<01@0o`<00000301A0?l4000000`0S`3o0`00000<0<`0o`@00000303o0?D300000100
o`2h0`00000A0?l0NP800000703o03d00`000000o`000?l0000i0?l000030000003o03d0o`0m01X0
o`0m0P00000A0?l0NP<00000403o0;P3000000`0o`3e1000000<0<`0o`@000002`2?0?l4000000`0
D@3o1000000;01@0o`@000003000:?l3000000H0oooo0@0000010?ooo`00803oool00`000000oooo
0?ooo`050?ooo`H002So2000000<01@0o`@000002`1A0?l4000000`0S`3o1000000;0<`0o`@00000
3P3o0?D200000100o`2h0`00000B0?l0NP8000007@3o03d00`000000o`000?l0000k0?l000800000
703o03d200000180o`1j0`00000@0?l0^08000003P3o0?D4000000`0c03o1000000;08l0o`@00000
301A0?l4000000/0503o20000006000Xo`L0oooo0@0000010?ooo`00803oool00`000000oooo0?oo
o`050?ooo`8002So1000000?01@0o`D000002`1A0?l4000000d0S`3o0`00000=0<`0o`8000003`3o
0?D300000100o`2h0P00000D0?l0NP030000003o03d0o`0m01`0o`0m00<000000?l0003o0000?P3o
00000`000000o`0m0?l0?@0K0?l0?@8000004`3o07X200000100o`2h0`00000?0?l0m@8000003P3<
0?l2000000d0S`3o1@00000;0540o`@000003`0D0?l40000008002So1`3oool100000040oooo000A
0?ooo`050?oOW018000000000000M02oool02P3oool00`000000oooo0?ooo`050?ooo`800000400D
0?l3000000d0D@3o0`00000?08l0o`8000003P3<0?l200000100o`3e00<000000?l0^03o0;P03`3o
0;P2000001D0o`1j00<000000?l0?@3o03d0703o03d00`000000o`000?l0000o0?l000030000003o
03d0o`0m01`0o`0m00<000000?l0NP3o07X04`3o07X200000140o`2h0P00000?0?l0m@8000003P3<
0?l200000100S`3o0P00000=0540o`<00000400D0?l3000000H0oooo0@0000010?ooo`00403oool0
1`3ookl0M00007Boo`3oool0okmd0000M02oool02@3oool00`000000oooo0?ooo`050?ooo`l0503o
0`00000>0540o`800000402?0?l2000000d0c03o0`00000A0?l0m@030000003o0;P0o`2h00l0o`2h
00<000000?l0NP3o07X0503o07X00`000000o`0m0?l0?@0L0?l0?@030000003o0000o`000440o`00
00<000000?l0?@3o03d0703o03d00`000000o`1j0?l0NP0D0?l0NP030000003o0;P0o`2h0100o`2h
00<000000?l0m@3o0?D03P3o0?D3000000d0c03o0P00000@08l0o`8000003P1A0?l3000000l0503o
1`3oool100000040oooo000@0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool0203o
ool00`000000oooo0?ooo`050?ooo`d0503o0P00000?0540o`800000402?0?l2000000h0c03o00<0
00000?l0m@3o0?D0403o0?D200000100o`2h0P00000E0?l0NP8000007`3o03d00`000000o`000?l0
00110?l000030000003o03d0o`0m01d0o`0m0P00000E0?l0NP8000004@3o0;P200000140o`3e0P00
000=0<`0o`800000402?0?l3000000h0D@3o0P00000=01@0o`L0oooo0@0000010?ooo`003`3oool0
0`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00P0oooo00<000000?ooo`3oool01@3oool:
01@0o`<000003`1A0?l2000000l0S`3o0`00000?0<`0o`030000003o0?D0o`3e00l0o`3e0P00000A
0?l0^0030000003o07X0o`1j01@0o`1j00<000000?l0?@3o03d07P3o03d00`000000o`000?l00013
0?l000030000003o03d0o`0m01h0o`0m00<000000?l0NP3o07X0503o07X00`000000o`2h0?l0^00@
0?l0^08000004@3o0?D00`000000c03o0<`0o`0<0<`0o`<00000402?0?l2000000h0D@3o0`00000:
01@0o`L0oooo0@0000010?ooo`003`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00P0oooo00<000000?ooo`3oool01@3oool:01@0o`030000001A0?l0D@3o00l0D@3o00<0000008l0
o`2?0?l03P2?0?l00`000000c03o0<`0o`0?0<`0o`030000003o0?D0o`3e00l0o`3e00<000000?l0
^03o0;P0403o0;P00`000000o`1j0?l0NP0D0?l0NP030000003o03d0o`0m01h0o`0m00<000000?l0
003o0000@`3o00000`000000o`0m0?l0?@0N0?l0?@030000003o07X0o`1j01@0o`1j00<000000?l0
^03o0;P04@3o0;P00`000000o`3e0?l0m@0?0?l0m@030000003<0?l0c03o00h0c03o00<0000008l0
o`2?0?l03`2?0?l00`000000D@3o0540o`0>0540o`030000000D0?l0503o00P0503o1`3oool10000
0040oooo000?0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0203oool3000000D0
oooo2P0D0?l00`000000D@3o0540o`0?0540o`030000002?0?l0S`3o00h0S`3o00<000000<`0o`3<
0?l03`3<0?l00`000000o`3e0?l0m@0?0?l0m@030000003o0;P0o`2h0100o`2h00<000000?l0NP3o
07X0503o07X00`000000o`0m0?l0?@0N0?l0?@030000003o0000o`0004<0o`0000<000000?l0?@3o
03d07P3o03d00`000000o`1j0?l0NP0D0?l0NP030000003o0;P0o`2h0140o`2h00<000000?l0m@3o
0?D03`3o0?D00`000000c03o0<`0o`0>0<`0o`030000002?0?l0S`3o00l0S`3o00<000000540o`1A
0?l03P1A0?l00`000000503o01@0o`0801@0o`D0oooo0`0000010?ooo`003`3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00P0oooo00<000000?ooo`3oool01@3oool:01@0o`030000
001A0?l0D@3o00l0D@3o00<0000008l0o`2?0?l03P2?0?l00`000000c03o0<`0o`0?0<`0o`030000
003o0?D0o`3e00l0o`3e00<000000?l0^03o0;P0403o0;P00`000000o`1j0?l0NP0D0?l0NP030000
003o03d0o`0m01h0o`0m00<000000?l0003o0000@`3o00000`000000o`0m0?l0?@0N0?l0?@030000
003o07X0o`1j01@0o`1j00<000000?l0^03o0;P04@3o0;P00`000000o`3e0?l0m@0?0?l0m@030000
003<0?l0c03o00h0c03o00<0000008l0o`2?0?l03`2?0?l00`000000D@3o0540o`0>0540o`030000
000D0?l0503o00P0503o1`3oool100000040oooo000?0?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol0203oool00`000000oooo0?ooo`050?ooo`X0503o00<000000540o`1A0?l03`1A
0?l00`000000S`3o08l0o`0>08l0o`030000003<0?l0c03o00l0c03o00<000000?l0m@3o0?D03`3o
0?D00`000000o`2h0?l0^00@0?l0^0030000003o07X0o`1j01@0o`1j00<000000?l0?@3o03d07P3o
03d00`000000o`000?l000130?l000030000003o03d0o`0m01h0o`0m00<000000?l0NP3o07X0503o
07X00`000000o`2h0?l0^00@0?l0^08000004@3o0?D00`000000c03o0<`0o`0>0<`0o`030000002?
0?l0S`3o00l0S`3o00<000000540o`1A0?l03P1A0?l00`000000503o01@0o`0801@0o`L0oooo0@00
00010?ooo`003`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00P0oooo00<00000
0?ooo`3oool01@3oool;01@0o`800000401A0?l2000000l0S`3o00<000000<`0o`3<0?l03`3<0?l0
0`000000o`3e0?l0m@0?0?l0m@030000003o0;P0o`2h0100o`2h00<000000?l0NP3o07X0503o07X0
0`000000o`0m0?l0?@0M0?l0?@030000003o0000o`000480o`0000<000000?l0?@3o03d07P3o03d0
0`000000o`1j0?l0NP0D0?l0NP030000003o0;P0o`2h0100o`2h00<000000?l0m@3o0?D03`3o0?D2
00000100c03o00<0000008l0o`2?0?l03P2?0?l2000000l0D@3o0P00000;01@0o`L0oooo0@000001
0?ooo`00403oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00P0oooo00<000000?oo
o`3oool01@3oool=01@0o`@000003P1A0?l2000000h0S`3o0P00000@0<`0o`030000003o0?D0o`3e
00l0o`3e0P00000A0?l0^08000005@3o07X00`000000o`0m0?l0?@0M0?l0?@030000003o0000o`00
0440o`0000<000000?l0?@3o03d07@3o03d00`000000o`1j0?l0NP0C0?l0NP8000004P3o0;P00`00
0000o`3e0?l0m@0?0?l0m@030000003<0?l0c03o00h0c03o0P00000?08l0o`8000003@1A0?l40000
00d0503o1`3oool100000040oooo000@0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`09
0?ooo`030000003oool0oooo00D0oooo0P00000?01@0o`<000003@1A0?l3000000d0S`3o0`00000>
0<`0o`8000004@3o0?D00`000000o`2h0?l0^00@0?l0^0030000003o07X0o`1j01<0o`1j00<00000
0?l0?@3o03d0703o03d00`000000o`000?l000100?l000030000003o03d0o`0m01d0o`0m00<00000
0?l0NP3o07X04`3o07X00`000000o`2h0?l0^00@0?l0^0800000403o0?D2000000h0c03o0`00000=
08l0o`@00000301A0?l300000100503o0P0000060?ooo`4000000@3oool00140oooo00D0oono07@0
00000000001d0;ooo`0:0?ooo`030000003oool0oooo00D0oooo0P00:?l4000000h0503o0P00000>
0540o`@00000302?0?l4000000`0c03o1000000>0?l0m@8000004@3o0;P2000001@0o`1j0P00000M
0?l0?@800000@03o00000`000000o`0m0?l0?@0L0?l0?@030000003o07X0o`1j0180o`1j0P00000A
0?l0^08000003P3o0?D4000000`0c03o1000000<08l0o`@000003P1A0?l2000000l0503o10000000
0`00:?l0oooo0?ooo`050?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool6000X
o`<000003@0D0?l3000000l0D@3o1000000<08l0o`@00000303<0?l3000000d0o`3e0`00000@0?l0
^0800000503o07X2000001d0o`0m00<000000?l0003o0000?03o00000`000000o`0m0?l0?@0K0?l0
?@8000004`3o07X200000140o`2h0P00000=0?l0m@<000003@3<0?l3000000`0S`3o1000000?0540
o`<000003@0D0?l4000000D002So1`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0
oooo2@00:?l8000000P0503o2000000;0540o`@00000302?0?l4000000/0c03o1000000<0?l0m@<0
00003`3o0;P2000001@0o`1j00<000000?l0?@3o03d06P3o03d00`000000o`000?l0000k0?l00003
0000003o03d0o`0m01X0o`0m0P00000C0?l0NP8000003`3o0;P4000000/0o`3e1000000<0<`0o`@0
00002`2?0?l4000000`0D@3o1`00000801@0o`P000002@00:?l70?ooo`4000000@3oool00200oooo
00<000000?ooo`3oool01@3oool5000000`002So1000000<01@0o`<00000301A0?l4000000`0S`3o
0`00000<0<`0o`<00000303o0?D3000000h0o`2h0`00000B0?l0NP8000006`3o03d00`000000o`00
0?l0000i0?l000030000003o03d0o`0m01T0o`0m0P00000C0?l0NP8000003P3o0;P3000000`0o`3e
0`00000<0<`0o`@000002`2?0?l4000000`0D@3o1000000;01@0o`@000003000:?l6000000H0oooo
0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`D006Ko20000008000Xo`P000001`0D
0?l8000000P0D@3o1000000;08l0o`@000002`3<0?l4000000/0o`3e0P00000?0?l0^0<000004@3o
07X2000001X0o`0m0P00000h0?l000030000003o03d0o`0m01P0o`0m0P00000B0?l0NP<000003P3o
0;P2000000`0o`3e0`00000;0<`0o`@000002`2?0?l4000000P0D@3o2000000801@0o`L000002000
:?l8000000D006Ko1`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo0P00000;
001Vo`@000003000:?l4000000/0503o100000080540o`@000002`2?0?l4000000/0c03o0P00000;
0?l0m@<000003`3o0;P200000140o`1j0P00000J0?l0?@030000003o0000o`0003@0o`0000<00000
0?l0?@3o03d05`3o03d200000180o`1j0P00000>0?l0^0<00000303o0?D2000000X0c03o1000000<
08l0o`<00000201A0?l4000000`0503o1000000;000Xo`@000003000I_l2000000H0oooo0@000001
0?ooo`00803oool00`000000oooo0?ooo`050?ooo`800:?o1`000008001Vo`P000002000:?l70000
00P0503o100000080540o`L00000202?0?l4000000T0c03o0`00000;0?l0m@@000003@3o0;P30000
0100o`1j0P00000I0?l0?@030000003o0000o`000340o`000P00000H0?l0?@8000004@3o07X30000
00`0o`2h1000000<0?l0m@<00000203<0?l4000000P0S`3o200000070540o`@00000200D0?l80000
00L002So20000008001Vo`P0000000<00:?o0?ooo`3oool01@3oool100000040oooo000P0?ooo`03
0000003oool0oooo00D0oooo0P000007002So`P000002000I_l7000000P002So1000000801@0o`@0
00002`1A0?l3000000T0S`3o100000080<`0o`<00000303o0?D3000000d0o`2h0P00000@0?l0NP03
0000003o03d0o`0m01L0o`0m00<000000?l0003o0000;`3o00000`000000o`0m0?l0?@0G0?l0?@03
0000003o07X0o`1j00h0o`1j0`00000<0?l0^0<00000303o0?D4000000P0c03o0`00000908l0o`<0
00002`1A0?l4000000P0503o10000008000Xo`L000002000I_l8000000P00:?o0P0000060?ooo`40
00000@3oool00200oooo00<000000?ooo`3oool01@3oool2003Qo`L000002000Xol8000000L006Ko
10000008000Xo`P00000100D0?l8000000H0D@3o1@00000808l0o`@000001`3<0?l4000000/0o`3e
0P00000=0?l0^08000003`3o07X00`000000o`0m0?l0?@0G0?l0?@030000003o0000o`0002d0o`00
00<000000?l0?@3o03d05`3o03d00`000000o`1j0?l0NP0=0?l0NP8000003@3o0;P2000000/0o`3e
100000080<`0o`@000001`2?0?l5000000L0D@3o1`00000401@0o`P000002000:?l4000000L006Ko
20000008002So`P0000000<00>7o0?ooo`3oool01@3oool100000040oooo000P0?ooo`030000003o
ool0oooo00D0oooo0P000007003Qo`P000002000Xol8000000<006Ko20000008000Xo`@00000200D
0?l3000000P0D@3o1000000808l0o`@000001`3<0?l4000000T0o`3e0`00000<0?l0^08000003P3o
07X2000001P0o`0m00<000000?l0003o0000:`3o00000`000000o`0m0?l0?@0F0?l0?@8000003P3o
07X2000000`0o`2h0`0000090?l0m@@00000203<0?l4000000L0S`3o100000080540o`@000001`0D
0?l4000000P002So20000004001Vo`L000002000Xol8000000P00>7o0P0000060?ooo`4000000@3o
ool00200oooo00<000000?ooo`3oool01@3oool2003oh@L000002000hOl8000000P00:?o0`000008
001Vo`@000002000:?l4000000L0503o100000080540o`@00000202?0?l3000000P0c03o10000008
0?l0m@@000002P3o0;P2000000h0o`1j0P00000G0?l0?@030000003o0000o`0002T0o`0000<00000
0?l0?@3o03d05@3o03d2000000h0o`1j0P00000;0?l0^0<00000203o0?D4000000P0c03o10000007
08l0o`@00000201A0?l4000000P0503o0`000008000Xo`@000002000I_l4000000L00:?o20000008
003Qo`P0000000<00?oQ0?ooo`3oool01@3oool100000040oooo000P0?ooo`030000003oool0oooo
00D0oooo0P000007003oh@P000002000hOl7000000@00:?o1`000005001Vo`L000001@00:?l60000
00D0503o100000080540o`<00000202?0?l3000000T0c03o0`0000090?l0m@<000002@3o0;P20000
00h0o`1j0P00000F0?l0?@030000003o0000o`0002L0o`0000<000000?l0?@3o03d05@3o03d00`00
0000o`1j0?l0NP0<0?l0NP8000002P3o0;P3000000P0o`3e0`0000090<`0o`<00000202?0?l30000
00P0D@3o1000000501@0o`L000001000:?l7000000D006Ko1`000005002So`H000002000hOl80000
00P00?oQ0P0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool2003oX`L0
00002000on48000000L00>7o10000007002So`D000001`00I_l5000000H002So1@00000401@0o`@0
00001`1A0?l5000000H0S`3o100000080<`0o`8000002@3o0?D2000000X0o`2h0P00000>0?l0NP80
00005@3o03d00`000000o`000?l0000U0?l000030000003o03d0o`0m01@0o`0m0P00000=0?l0NP80
00002P3o0;P2000000X0o`3e0P0000070<`0o`@000001`2?0?l4000000L0D@3o1000000401@0o`D0
00001`00:?l4000000L006Ko1@000007002So`D000001P00hOl8000000P00?oQ200000000`00oj<0
oooo0?ooo`050?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool2000000L00?nS
20000008003oh@P000000`00hOl8000000@00:?o20000004001Vo`@000001`00:?l4000000@0503o
100000080540o`<000001`2?0?l2000000P0c03o0`0000080?l0m@8000002P3o0;P2000000h0o`1j
0P00000D0?l0?@800000903o00000`000000o`0m0?l0?@0C0?l0?@8000003@3o07X2000000X0o`2h
0P0000080?l0m@@000001`3<0?l2000000L0S`3o100000070540o`@00000100D0?l4000000P002So
10000003001Vo`P000001000Xol8000000@00>7o1`000008003oh@P000002000oj<2000000H0oooo
0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`800?mV2`000004003oX`P000002000
on47000000@00>7o20000004002So`@000001000I_l7000000@002So1000000401@0o`@000001`1A
0?l4000000D0S`3o0`0000080<`0o`@000001P3o0?D2000000X0o`2h0P00000>0?l0NP800000503o
03d00`000000o`000?l0000P0?l000030000003o03d0o`0m0180o`0m0P00000=0?l0NP8000002P3o
0;P2000000L0o`3e0`0000080<`0o`<000001@2?0?l4000000P0D@3o0`00000401@0o`@000001000
:?l8000000@006Ko0`000004002So`P000001000hOl8000000L00?oQ20000004003oX``0000000<0
0?mV0?ooo`3oool01@3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo1P000007
003oIPT000000`00oj<:000000D00?oQ1P000006003Qo`D000000`00Xol6000000D006Ko1@000003
000Xo`D000000`0D0?l6000000D0D@3o1000000408l0o`D000001`3<0?l2000000H0o`3e10000008
0?l0^0<000003@3o07X00`000000o`0m0?l0?@0B0?l0?@030000003o0000o`0001d0o`000P00000D
0?l0?@030000003o07X0o`1j00X0o`1j0`0000080?l0^0@000001P3o0?D3000000H0c03o1@000005
08l0o`<000001@1A0?l700000080503o1@000003000Xo`D000001@00I_l7000000<00:?o10000006
003Qo`H000001P00on49000000<00?nS2@000007003oIPL000001P3oool100000040oooo000P0?oo
o`030000003oool0oooo00D0oooo1P00obP7000000T00?mV1`000005003oX`H000001P00on460000
00D00>7o0`000005002So`H000001@00I_l3000000D002So0`00000501@0o`800000201A0?l20000
00L0S`3o0P0000070<`0o`800000203o0?D00`000000o`2h0?l0^0080?l0^08000002`3o07X00`00
0000o`0m0?l0?@0C0?l0?@030000003o0000o`0001/0o`0000<000000?l0?@3o03d04`3o03d00`00
0000o`1j0?l0NP090?l0NP8000002P3o0;P00`000000o`3e0?l0m@070?l0m@8000001P3<0?l20000
00P0S`3o0P0000070540o`8000001P0D0?l2000000D002So0`000005001Vo`H000001P00Xol30000
00@00>7o1P000006003oh@H000001P00oj<6000000T00?mV1`000006003o:0L0oooo0@0000010?oo
o`00803oool00`000000oooo0?ooo`050?ooo`D000002000obP8000000P00?mV1`000004003oX`P0
00001000on44000000@00>7o1`000004002So`@000001000I_l4000000@002So1000000301@0o`@0
00001P1A0?l2000000L0S`3o0P0000070<`0o`8000001`3o0?D00`000000o`2h0?l0^0090?l0^003
0000003o07X0o`1j00T0o`1j00<000000?l0?@3o03d04`3o03d00`000000o`000?l0000J0?l00003
0000003o03d0o`0m0180o`0m00<000000?l0NP3o07X02@3o07X00`000000o`2h0?l0^0090?l0^003
0000003o0?D0o`3e00D0o`3e0`0000060<`0o`8000001`2?0?l3000000D0D@3o1000000401@0o`@0
00000`00:?l4000000@006Ko10000004002So`P000000`00hOl4000000@00?oQ20000004003oX`P0
00001`00ofH8000000P00?lX1P0000060?ooo`4000000@3oool00200oooo0P0000060?ooo`D05?l0
20000008003o:0P000001`00ofH4000000P00?nS10000004003oh@@000001`00hOl4000000@00:?o
10000004001Vo`@000001000:?l3000000@0503o100000040540o`@000001@2?0?l3000000H0c03o
0P0000060?l0m@8000002P3o0;P2000000X0o`1j0P00000D0?l0?@030000003o0000o`0001P0o`00
00<000000?l0?@3o03d04@3o03d2000000X0o`1j0P00000:0?l0^08000001P3o0?D2000000L0c03o
0P00000508l0o`@00000101A0?l4000000@0503o10000004000Xo`<000001000I_l4000000@00:?o
10000008003Qo`<000001000on44000000P00?nS10000008003oIPL000002000obP8000000D05?l0
1P3oool200000040oooo000P0?ooo`030000003oool0oooo00D0oooo1@00000801Co00`000001000
obP8000000<00?mV2@000003003oX`H000000P00on47000000@00>7o1@000003002So`@000001000
I_l4000000<002So1000000401@0o`<000001@1A0?l3000000D0S`3o0`0000050<`0o`<000001@3o
0?D4000000P0o`2h0P00000:0?l0NP8000004P3o03d00`000000o`000?l0000G0?l000030000003o
03d0o`0m0140o`0m00<000000?l0NP3o07X02@3o07X00`000000o`2h0?l0^0070?l0^0<000001P3o
0?D2000000H0c03o0`00000408l0o`<000001@1A0?l3000000D0503o0`000003000Xo`D000000`00
I_l4000000@00:?o10000004003Qo`P000000P00on45000000<00?nS2@000003003oIPT000000`00
obP<000000P05?l01P0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool5
057o00`00000200Do`08000000@00?lX1`000005003oIP<000001P00oj<6000000<00?oQ10000005
003Qo`<000001000Xol4000000@006Ko0P000005000Xo`8000001@0D0?l2000000H0D@3o0P000006
08l0o`8000001P3<0?l2000000L0o`3e00<000000?l0^03o0;P01`3o0;P00`000000o`1j0?l0NP09
0?l0NP030000003o03d0o`0m0100o`0m00<000000?l0003o00005@3o00000`000000o`0m0?l0?@0A
0?l0?@030000003o07X0o`1j00T0o`1j00<000000?l0^03o0;P01P3o0;P2000000L0o`3e0P000006
0<`0o`8000001@2?0?l2000000H0D@3o0P00000601@0o`8000001000:?l2000000D006Ko0`000004
002So`@000001000hOl4000000@00?oQ1P000005003oX`<000001@00ofH7000000D00?lX1`000008
01Co00`000001@1Ao`070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool20000
00l0DOl02000000801Co00<000002000obP4000000@00?mV20000004003oX`<000001000on440000
00@00>7o10000004002So`<000000`00I_l2000000D002So0P00000501@0o`<000001@1A0?l20000
00H0S`3o00<000000<`0o`3<0?l01@3<0?l00`000000o`3e0?l0m@050?l0m@030000003o0;P0o`2h
00L0o`2h0P0000090?l0NP030000003o03d0o`0m0100o`0m00<000000?l0003o0000503o00000`00
0000o`0m0?l0?@0A0?l0?@030000003o07X0o`1j00P0o`1j0P0000080?l0^0030000003o0?D0o`3e
00D0o`3e0P0000060<`0o`8000001P2?0?l00`000000D@3o0540o`030540o`<000001@0D0?l30000
00@002So0P000004001Vo`<000000`00Xol4000000@00>7o10000004003oh@@000001000oj<70000
00@00?mV10000008003o:0@000001`0Do`0800000100DOl00P0000060?ooo`4000000@3oool00200
oooo00<000000?ooo`3oool01@3oool208oo00/00000301Ao`08000000<05?l020000004003o:0P0
00001000ofH4000000<00?nS10000004003oh@@000001000hOl4000000<00:?o0P000003001Vo`<0
00001000:?l4000000@0503o0`0000040540o`8000001@2?0?l2000000H0c03o0P0000060?l0m@80
00002@3o0;P00`000000o`1j0?l0NP070?l0NP030000003o03d0o`0m0100o`0m00<000000?l0003o
00004`3o00000`000000o`0m0?l0?@0@0?l0?@030000003o07X0o`1j00P0o`1j00<000000?l0^03o
0;P01P3o0;P2000000H0o`3e0P0000060<`0o`8000001P2?0?l2000000<0D@3o0`00000401@0o`@0
00001000:?l3000000@006Ko0P000003002So`<000001000hOl4000000@00?oQ10000004003oX`@0
00000`00ofH8000000@00?lX2000000401Co00L00000301Ao`0<0000000308oo003oool0oooo00D0
oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`d0Sol030000008057o00L00000
100Do`08000000@00?lX10000004003oIP@000000`00oj<5000000<00?oQ10000004003Qo`<00000
0P00Xol40000008006Ko1@000003000Xo`<00000100D0?l2000000@0D@3o0`00000408l0o`<00000
1@3<0?l2000000H0o`3e0P0000080?l0^0030000003o07X0o`1j00L0o`1j0P00000A0?l0?@800000
4P3o00000`000000o`0m0?l0?@0@0?l0?@030000003o07X0o`1j00P0o`1j00<000000?l0^03o0;P0
1@3o0;P2000000H0o`3e0P0000060<`0o`8000001@2?0?l3000000<0D@3o0P00000401@0o`<00000
0`00:?l5000000<006Ko10000002002So`<000000`00hOl4000000<00?oQ1@000003003oX`D00000
0`00ofH4000000@00?lX2000000401Co00P000001`1Ao`0<000000d0Sol01`3oool100000040oooo
000P0?ooo`030000003oool0oooo00D0oooo2@00000@08oo00P000001`1Ao`04000000P05?l01000
0004003o:0@000001000ofH3000000D00?nS0`000004003oh@@000000`00hOl2000000@00:?o0P00
0005001Vo`8000001000:?l2000000@0503o0P0000050540o`030000002?0?l0S`3o00@0S`3o0P00
00050<`0o`030000003o0?D0o`3e00D0o`3e00<000000?l0^03o0;P01P3o0;P00`000000o`1j0?l0
NP080?l0NP030000003o03d0o`0m0100o`0m00<000000?l0003o00003P3o00000`000000o`0m0?l0
?@0@0?l0?@030000003o07X0o`1j00P0o`1j00<000000?l0^03o0;P01@3o0;P00`000000o`3e0?l0
m@050?l0m@030000003<0?l0c03o00@0c03o0P00000508l0o`8000001@1A0?l00`000000503o01@0
o`0201@0o`8000001000:?l2000000D006Ko0`000004002So`8000000`00hOl3000000@00?oQ0`00
0005003oX`<000001@00ofH3000000@00?lX1000000801Co00@00000201Ao`0700000100Sol02P00
00060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool90<co00`00000302?o`07
000000@0DOl02000000401Co00@000001000obP3000000@00?mV10000004003oX`@000001000on43
000000800>7o0P000004002So`<000001000I_l2000000@002So0P00000401@0o`800000101A0?l0
0`000000S`3o08l0o`0508l0o`030000003<0?l0c03o00<0c03o0P0000060?l0m@8000001P3o0;P0
0`000000o`1j0?l0NP080?l0NP030000003o03d0o`0m0100o`0m00<000000?l0003o00003@3o0000
0`000000o`0m0?l0?@0@0?l0?@030000003o07X0o`1j00P0o`1j00<000000?l0^03o0;P01@3o0;P0
0`000000o`3e0?l0m@040?l0m@8000001@3<0?l2000000H0S`3o00<000000540o`1A0?l00`1A0?l2
000000<0503o0P000004000Xo`8000001000I_l3000000@00:?o0`000002003Qo`<000000`00on44
000000@00?nS10000004003oIP@000001000obP3000000@05?l020000004057o00P000002`2?o`0<
000000T0c?l01`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo5@3<o`080000
00/0Sol010000008057o00@00000100Do`04000000<00?lX10000004003oIP@000001000oj<30000
00@00?oQ0P000002003Qo`@000000`00Xol2000000@006Ko0`000003000Xo`800000100D0?l20000
00<0D@3o0P00000608l0o`8000001@3<0?l00`000000o`3e0?l0m@050?l0m@030000003o0;P0o`2h
00@0o`2h00<000000?l0NP3o07X0203o07X00`000000o`0m0?l0?@0@0?l0?@030000003o0000o`00
00/0o`0000<000000?l0?@3o03d0403o03d00`000000o`1j0?l0NP080?l0NP030000003o0;P0o`2h
00D0o`2h00<000000?l0m@3o0?D0103o0?D00`000000c03o0<`0o`030<`0o`8000001P2?0?l20000
00<0D@3o0`00000301@0o`8000001000:?l2000000@006Ko0P000003002So`@000000`00hOl20000
00<00?oQ0`000004003oX`@000001000ofH4000000@00?lX1000000301Co00@00000201Ao`040000
00`0Sol01`00000E0<co00L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oooad0
c?l02000000708oo00P00000101Ao`04000000@05?l010000003003o:0D000000`00ofH4000000<0
0?nS0P000004003oh@<000000`00hOl3000000800:?o10000003001Vo`8000000`00:?l2000000@0
503o0P0000030540o`<000001@2?0?l00`000000c03o0<`0o`030<`0o`8000001P3o0?D00`000000
o`2h0?l0^0040?l0^08000002@3o07X00`000000o`0m0?l0?@0@0?l0?@8000002P3o00000`000000
o`0m0?l0?@0@0?l0?@030000003o07X0o`1j00L0o`1j0P0000060?l0^08000001P3o0?D00`000000
c03o0<`0o`030<`0o`030000002?0?l0S`3o00<0S`3o0`0000030540o`800000100D0?l2000000@0
02So0P000002001Vo`@000000P00Xol3000000<00>7o10000003003oh@8000000`00oj<4000000<0
0?mV1@000003003o:0D000000`0Do`04000000@0DOl02000000808oo00P00000703<o`070?ooo`40
00000@3oool00200oooo00<000000?ooo`3oool01@3oool60<co01<00000303<o`07000000P0Sol0
10000004057o00@00000100Do`03000000D00?lX0`000004003oIP<000000P00oj<3000000@00?oQ
0P000004003Qo`8000001000Xol2000000<006Ko0P000003000Xo`800000100D0?l00`000000D@3o
0540o`030540o`800000102?0?l00`000000c03o0<`0o`040<`0o`030000003o0?D0o`3e00@0o`3e
0P0000060?l0^0030000003o07X0o`1j00L0o`1j00<000000?l0?@3o03d04@3o03d2000000H0o`00
0P00000B0?l0?@030000003o07X0o`1j00L0o`1j00<000000?l0^03o0;P01@3o0;P00`000000o`3e
0?l0m@040?l0m@8000001@3<0?l00`000000S`3o08l0o`0208l0o`8000001@1A0?l00`000000503o
01@0o`0201@0o`8000000`00:?l30000008006Ko0P000004002So`8000001000hOl2000000D00?oQ
0P000003003oX`8000001000ofH3000000D00?lX0`00000501Co00<00000101Ao`04000000P0Sol0
2000000;0<co01<000001P3<o`070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3o
ool6000001<0ooD02000000;0<co00P00000102?o`04000000@0DOl00`00000401Co00@000001000
obP4000000<00?mV0P000003003oX`<000000`00on42000000@00>7o0P000004002So`8000000`00
I_l30000008002So0P00000301@0o`8000001@1A0?l00`000000S`3o08l0o`0208l0o`8000001@3<
0?l00`000000o`3e0?l0m@050?l0m@030000003o0;P0o`2h00@0o`2h00<000000?l0NP3o07X01P3o
07X00`000000o`0m0?l0?@0C0?l0?@H000004`3o03d00`000000o`1j0?l0NP070?l0NP030000003o
0;P0o`2h00D0o`2h00<000000?l0m@3o0?D0103o0?D00`000000c03o0<`0o`030<`0o`800000102?
0?l00`000000D@3o0540o`030540o`800000100D0?l01000000002So000Xo`00:?l2000000<006Ko
0P000004002So`8000001000hOl2000000<00?oQ10000003003oX`8000000P00ofH4000000@00?lX
1000000401Co00@00000101Ao`03000000@0Sol02000000<0<co00L000004`3om@07000000H0oooo
0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooob40ooD01`00000<0<co00@00000102?
o`04000000<0DOl01000000401Co00@000001000obP3000000800?mV0`000003003oX`<000000P00
on43000000<00>7o0P000004002So`8000001000I_l20000008002So0P00000301@0o`800000101A
0?l2000000@0S`3o0P0000040<`0o`8000001@3o0?D00`000000o`2h0?l0^0050?l0^0030000003o
07X0o`1j00H0o`1j00<000000?l0?@3o03d0:`3o03d00`000000o`1j0?l0NP060?l0NP030000003o
0;P0o`2h00D0o`2h00<000000?l0m@3o0?D0103o0?D00`000000c03o0<`0o`020<`0o`800000102?
0?l2000000@0D@3o0P00000401@0o`8000000P00:?l2000000<006Ko0P000004002So`8000000`00
hOl3000000800?oQ0`000004003oX`<0000000@00?mV00000000000000001000obP4000000@05?l0
10000004057o00@000000`2?o`04000000`0c?l02000000P0?oe00L0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooobP0ooD0200000080<co00@00000102?o`04000000<0DOl01@00
000301Co00@000000`00obP2000000<00?mV10000002003oX`8000000`00on43000000800>7o0P00
0004002So`8000001000I_l20000008002So0P00000301@0o`800000101A0?l00`000000S`3o08l0
o`0308l0o`030000003<0?l0c03o00<0c03o00<000000?l0m@3o0?D00`3o0?D00`000000o`2h0?l0
^0050?l0^0030000003o07X0o`1j00H0o`1j00<000000?l0?@3o03d0:@3o03d00`000000o`1j0?l0
NP060?l0NP030000003o0;P0o`2h00D0o`2h00<000000?l0m@3o0?D0103o0?D00`000000c03o0<`0
o`020<`0o`030000002?0?l0S`3o00<0S`3o00<000000540o`1A0?l00P1A0?l2000000@0503o0P00
0002000Xo`8000001000I_l00`0000000:?o002So`02002So`8000000P00hOl3000000<00?oQ0P00
0003003oX`@000000P00ofH2000000<00?lX1000000301Co00D000000`1Ao`05000000@0Sol00`00
00080<co00P00000:03om@070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooolQ
0?oe00<00000303om@04000000P0c?l01000000408oo00<000001@1Ao`03000000@05?l00`000002
003o:0<000001000ofH2000000800?nS0P000004003oh@8000000P00hOl2000000@00:?o0P000004
001Vo`0400000000:?l002So000Xo`8000000`0D0?l2000000<0D@3o00<0000008l0o`2?0?l00`2?
0?l2000000@0c03o00<000000?l0m@3o0?D00`3o0?D00`000000o`2h0?l0^0050?l0^0030000003o
07X0o`1j00H0o`1j00<000000?l0?@3o03d09`3o03d00`000000o`1j0?l0NP060?l0NP030000003o
0;P0o`2h00D0o`2h00<000000?l0m@3o0?D0103o0?D010000000c03o0<`0o`3<0?l2000000D0S`3o
00@000000540o`1A0?l0D@3o0P00000401@0o`8000000`00:?l00`00000006Ko001Vo`02001Vo`80
00000`00Xol2000000800>7o0P000004003oh@8000000P00oj<3000000@00?mV0P000003003o:080
0000100Do`03000000D0DOl00`00000508oo00@000001`3<o`04000000`0ooD00`00000Q0?oe00L0
oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oooa<0ooD03P0000030?nh00P00000
203om@08000000@0c?l00`00000408oo00@00000101Ao`04000000<05?l00P000003003o:0800000
1000ofH2000000800?nS0P000004003oh@8000000P00hOl2000000@00:?o0P000003001Vo`030000
0000:?l002So008002So00<0000001@0o`0D0?l00P0D0?l010000000D@3o0540o`1A0?l2000000D0
S`3o00<000000<`0o`3<0?l00P3<0?l00`000000o`3e0?l0m@030?l0m@030000003o0;P0o`2h00@0
o`2h00<000000?l0NP3o07X01`3o07X00`000000o`0m0?l0?@0U0?l0?@030000003o07X0o`1j00L0
o`1j00<000000?l0^03o0;P0103o0;P00`000000o`3e0?l0m@040?l0m@050000003<0?l0c03o0<`0
o`0000001@2?0?l2000000<0D@3o00<0000001@0o`0D0?l00P0D0?l2000000<002So0P000003001V
o`8000000`00Xol2000000800>7o0P000004003oh@8000000P00oj<2000000@00?mV0`000003003o
:08000000P0Do`04000000@0DOl01000000408oo00@00000103<o`07000000P0ooD0200000030?nh
00h000004`3om@070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool00`000000
ooD00?oe000?0?oe00030000003o^000okP001L0okP01000000<0?oe00<00000103<o`04000000@0
Sol010000004057o00<000000P0Do`03000000800?lX0P000004003oIP8000000P00oj<3000000<0
0?oQ0P000002003Qo`8000001000Xol20000008006Ko00<00000000Xo`00:?l00P00:?l2000000<0
503o00<000000540o`1A0?l00P1A0?l2000000<0S`3o00<000000<`0o`3<0?l00`3<0?l00`000000
o`3e0?l0m@030?l0m@030000003o0;P0o`2h00@0o`2h00<000000?l0NP3o07X01P3o07X00`000000
o`0m0?l0?@0U0?l0?@030000003o07X0o`1j00H0o`1j00<000000?l0^03o0;P0103o0;P00`000000
o`3e0?l0m@030?l0m@8000000`3<0?l00`000000S`3o08l0o`0208l0o`800000101A0?l010000000
503o01@0o`0D0?l2000000@002So00@00000001Vo`00I_l006Ko0P000003002So`8000000P00hOl2
000000<00?oQ0`000002003oX`8000001000ofH2000000<00?lX0`00000201Co00800000101Ao`04
000000@0Sol0100000040<co00@000002`3om@04000001T0okP000<000000?oe003om@00403om@00
0`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool00`3<
o`0000000?oe000>0?oe00030000003o^000okP001`0okP0200000070?oe00@00000103<o`040000
00@0Sol010000003057o008000000`0Do`02000000800?lX10000002003oIP8000000`00oj<30000
00800?oQ0P000002003Qo`8000001000Xol00`00000006Ko001Vo`02000000@002So00@0000001@0
o`0D0?l0503o0P0000030540o`030000002?0?l0S`3o0080S`3o00<000000<`0o`3<0?l00`3<0?l2
000000@0o`3e00<000000?l0^03o0;P0103o0;P00`000000o`1j0?l0NP060?l0NP030000003o03d0
o`0m02<0o`0m00<000000?l0NP3o07X01P3o07X00`000000o`2h0?l0^0040?l0^0030000003o0?D0
o`3e00<0o`3e00<000000<`0o`3<0?l00`3<0?l00`000000S`3o08l0o`0208l0o`040000001A0?l0
D@3o0540o`8000000`0D0?l00`00000002So000Xo`02000Xo`8000000`00I_l0100000000:?o002S
o`00Xol2000000800>7o0P000002003oh@<000000`00oj<2000000800?mV10000002003o:0<00000
0`0Do`0200000080DOl01000000408oo00@00000103<o`04000000L0ooD02000000N0?nh00030000
003om@00ooD000d0ooD00P0000070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3o
ool20<co008000003P3om@000`000000okP00?nh000S0?nh00@000001`3om@04000000@0c?l01000
000408oo00<000000P1Ao`03000000805?l00P000004003o:08000000P00ofH2000000@00?nS0P00
0002003oh@8000000P00hOl2000000<00:?o00@00000001Vo`00I_l006Ko0P000003000Xo`030000
000D0?l0503o0080503o00D000000540o`1A0?l0D@3o0000000408l0o`8000001@3<0?l00`000000
o`3e0?l0m@020?l0m@030000003o0;P0o`2h00@0o`2h00<000000?l0NP3o07X01P3o07X00`000000
o`0m0?l0?@0Q0?l0?@030000003o07X0o`1j00H0o`1j00<000000?l0^03o0;P0103o0;P00`000000
o`3e0?l0m@030?l0m@030000003<0?l0c03o00<0c03o00<0000008l0o`2?0?l00P2?0?l01@000000
D@3o0540o`1A0?l0000000@0503o00@00000000Xo`00:?l002So0P000003001Vo`8000000P00Xol2
000000800>7o0P000002003oh@8000001000oj<2000000800?mV0P000004003o:08000000`0Do`03
00000080DOl00P00000408oo00@00000103<o`04000000P0ooD00`00000U0?nh00030000003om@00
ooD000`0ooD00P0000020<co00L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oo
o`@0c?l00P00000<0?oe00030000003o^000okP002L0okP0100000070?oe00@00000103<o`030000
00@0Sol00P000003057o008000000P0Do`03000000<00?lX0P000002003oIP8000001000oj<20000
00800?oQ0P000002003Qo`8000000P00Xol00`00000006Ko001Vo`02001Vo`0500000000:?l002So
000Xo`000000100D0?l200000080D@3o0P00000408l0o`030000003<0?l0c03o0080c03o00<00000
0?l0m@3o0?D00`3o0?D00`000000o`2h0?l0^0030?l0^0030000003o07X0o`1j00L0o`1j00<00000
0?l0?@3o03d0803o03d00`000000o`1j0?l0NP050?l0NP030000003o0;P0o`2h00@0o`2h00<00000
0?l0m@3o0?D00`3o0?D00`000000c03o0<`0o`030<`0o`040000002?0?l0S`3o08l0o`8000000P1A
0?l2000000@0503o00D00000000Xo`00:?l002So00000004001Vo`0500000000Xol00:?o002So`00
00000`00hOl00`0000000?oQ003oh@02000000@00?nS0P000002003oIP8000000`00obP300000080
5?l00`000003057o008000000`2?o`03000000@0c?l0100000080?oe00@00000:03o^0000`000000
ooD00?oe000:0?oe00800000103<o`070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool0
1@3oool60<co008000002`3om@000`000000okP00?nh000C0?nh00/00000303o^003000000P0ooD0
0`0000040<co00800000102?o`02000000<0DOl00P00000301Co00<000000P00obP2000000800?mV
0P000004003oX`0500000000on400?oQ003oh@0000000`00hOl00`0000000:?o002So`02000000<0
06Ko0P000002000Xo`800000100D0?l01@000000D@3o0540o`1A0?l0000000@0S`3o00<000000<`0
o`3<0?l00P3<0?l00`000000o`3e0?l0m@020?l0m@030000003o0;P0o`2h00@0o`2h00<000000?l0
NP3o07X01P3o07X00`000000o`0m0?l0?@0O0?l0?@030000003o07X0o`1j00H0o`1j00<000000?l0
^03o0;P00`3o0;P00`000000o`3e0?l0m@040?l0m@030000003<0?l0c03o0080c03o00D0000008l0
o`2?0?l0S`3o000000030540o`030000000D0?l0503o0080503o0P000003000Xo`0400000000I_l0
06Ko001Vo`8000000P00Xol2000000800>7o0P000002003oh@0300000000oj<00?nS00800?nS0P00
0002003oIP8000000P00obP3000000<05?l00P000004057o008000000`2?o`02000000@0c?l00`00
00080?oe00@000002`3o^00;000001D0okP000<000000?oe003om@002@3om@02000000H0c?l01`3o
ool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo0P0000060<co008000002P3om@00
0`000000okP00?nh000?0?nh00<000002`3oNP08000000L0okP0100000070?oe00800000103<o`03
000000<0Sol000<00000057o001Ao`000P1Ao`02000000@05?l00P000002003o:08000000P00ofH2
000000<00?nS00D00000003oh@00on400?oQ00000003003Qo`8000000P00Xol2000000<006Ko00D0
0000000Xo`00:?l002So0000000301@0o`030000001A0?l0D@3o0080D@3o00<0000008l0o`2?0?l0
0P2?0?l00`000000c03o0<`0o`020<`0o`030000003o0?D0o`3e0080o`3e00<000000?l0^03o0;P0
103o0;P00`000000o`1j0?l0NP060?l0NP030000003o03d0o`0m01h0o`0m00<000000?l0NP3o07X0
1@3o07X00`000000o`2h0?l0^0040?l0^0030000003o0?D0o`3e00<0o`3e00<000000<`0o`3<0?l0
0P3<0?l01@000000S`3o08l0o`2?0?l0000000@0D@3o00D0000001@0o`0D0?l0503o00000003000X
o`8000000P00I_l2000000800:?o0P000003003Qo`0500000000on400?oQ003oh@0000000`00oj<2
000000800?mV0P000003003o:08000000`0Do`02000000@0DOl00P00000308oo008000001@3<o`02
000000H0ooD0100000080?nh00L000002`3oNP0300000140okP000<000000?oe003om@00203om@02
000000L0c?l00P0000060?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool208oo
008000001P3<o`02000000T0ooD00P00000?0?nh00030000003oNP00ogX001<0ogX0100000070?nh
00@000001@3om@03000000@0c?l00`000000102?o`000000000000000003057o00800000100Do`02
000000800?lX0P000002003oIP8000000P00oj<2000000800?oQ0P000003003Qo`0500000000Xol0
0:?o002So`0000000`00I_l01@00000002So000Xo`00:?l0000000<0503o00<000000540o`1A0?l0
0P1A0?l00`000000S`3o08l0o`0208l0o`050000003<0?l0c03o0<`0o`0000001@3o0?D00`000000
o`2h0?l0^0030?l0^0030000003o07X0o`1j00H0o`1j00<000000?l0?@3o03d07@3o03d00`000000
o`1j0?l0NP050?l0NP030000003o0;P0o`2h00@0o`2h00<000000?l0m@3o0?D00`3o0?D00`000000
c03o0<`0o`020<`0o`050000002?0?l0S`3o08l0o`000000101A0?l01@000000503o01@0o`0D0?l0
000000<002So00D00000001Vo`00I_l006Ko00000003002So`0300000000hOl00>7o00800>7o00D0
0000003oh@00on400?oQ00000002003oX`8000000P00ofH2000000800?lX0`00000301Co00800000
0`1Ao`0300000080Sol00`0000030<co00@00000103om@04000000P0okP01000000D0?mj00030000
003o^000okP000d0okP00P0000090?oe008000001P3<o`030000000308oo003oool0oooo00D0oooo
0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`@0Sol00P0000060<co008000002@3o
m@000`000000okP00?nh000<0?nh00030000003oNP00ogX001L0ogX00`0000080?nh00@00000103o
m@03000000@0c?l00P00000208oo00<000000P1Ao`02000000@05?l00P000002003o:08000000P00
ofH2000000800?nS0P000002003oh@8000000P00hOl2000000800:?o0P000002001Vo`8000000P00
:?l200000080503o0P0000030540o`030000002?0?l0S`3o0080S`3o00D000000<`0o`3<0?l0c03o
000000040?l0m@030000003o0;P0o`2h00@0o`2h00<000000?l0NP3o07X01P3o07X00`000000o`0m
0?l0?@0K0?l0?@030000003o07X0o`1j00H0o`1j00<000000?l0^03o0;P00`3o0;P00`000000o`3e
0?l0m@030?l0m@030000003<0?l0c03o0080c03o00<0000008l0o`2?0?l00P2?0?l00`000000D@3o
0540o`0200000080503o0P000002000Xo`8000000P00I_l2000000800:?o0P000003003Qo`800000
0P00on4200000003003oX`000000000000800?mV0P000002003o:0800000100Do`0200000080DOl0
0`00000308oo00800000103<o`02000000@0ooD0100000080?nh00@00000603oNP000`000000okP0
0?nh000<0?nh00030000003om@00ooD000L0ooD00P0000060<co00800000102?o`070?ooo`400000
0@3oool00200oooo0P0000060?ooo`030000002?o`00Sol000<0Sol00`0000050<co00<000001`3o
m@02000000`0okP000<000000?mj003oNP006P3oNP03000000T0okP00`0000040?oe00800000103<
o`000`000000Sol008oo000208oo008000000P1Ao`02000000@05?l00P000002003o:08000000P00
ofH2000000800?nS0P000002003oh@0500000000hOl00>7o003Qo`0000000`00Xol01@00000006Ko
001Vo`00I_l0000000<002So00D0000001@0o`0D0?l0503o000000030540o`050000002?0?l0S`3o
08l0o`000000103<0?l00`000000o`3e0?l0m@020?l0m@800000103o0;P00`000000o`1j0?l0NP06
0?l0NP030000003o03d0o`0m01/0o`0m00<000000?l0NP3o07X01@3o07X00`000000o`2h0?l0^003
0?l0^08000001@3o0?D01@000000c03o0<`0o`3<0?l0000000@0S`3o00@000000540o`1A0?l00000
0`0D0?l01@00000002So000Xo`00:?l0000000<006Ko00D00000002So`00Xol00:?o00000003003Q
o`8000000P00on42000000800?nS00<00000003oIP00ofH00P000002003o:0800000100Do`020000
0080DOl00P00000408oo00800000103<o`02000000<0ooD00`0000090?nh00<00000703oNP000`00
0000okP00?nh000:0?nh00800000203om@02000000H0c?l00P00000508oo008000001@3oool20000
0040oooo000P0?ooo`030000003oool0oooo00D0oooo00<0DOl00000000000000P00000408oo00<0
00001@3<o`03000000H0ooD00P00000:0?nh00030000003oNP00ogX001d0ogX01@0000060?nh0080
00001@3om@0300000080c?l00P00000408oo008000000P1Ao`02000000<05?l000D00000003o:000
obP00?lX00000003003oIP0500000000oj<00?nS003oX`0000000`00on401@0000000>7o003Qo`00
hOl0000000<00:?o00@00000001Vo`00I_l000000`00:?l01@000000503o01@0o`0D0?l0000000@0
D@3o00D0000008l0o`2?0?l0S`3o000000040<`0o`030000003o0?D0o`3e0080o`3e00<000000?l0
^03o0;P00`3o0;P00`000000o`1j0?l0NP060?l0NP030000003o03d0o`0m01T0o`0m00<000000?l0
NP3o07X01P3o07X00`000000o`2h0?l0^0030?l0^0030000003o0?D0o`3e00<0o`3e00D000000<`0
o`3<0?l0c03o0000000408l0o`050000001A0?l0D@3o0540o`0000000`0D0?l01@00000002So000X
o`00:?l0000000<006Ko00@00000002So`00Xol000000`00hOl01@0000000?oQ003oh@00on400000
00<00?nS00D00000003oIP00ofH00?mV00000003003o:0040000000Do`005?l001Co008000000P1A
o`02000000@0Sol00P0000030<co00<00000103om@02000000H0okP01@00000O0?mj00030000003o
^000okP000T0okP000<000000?oe003om@00103om@04000000@0c?l01000000308oo00@0000000<0
DOl00?ooo`3oool01@3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo0P000003
057o00<00000102?o`02000000H0c?l00P0000060?oe008000002@3o^0000`000000ogX00?mj000Q
0?mj00<000001@3o^004000000@0ooD00P0000020<co00800000102?o`0200000080DOl00P000002
01Co008000000P00obP2000000800?mV0P000002003oX`8000000P00on42000000800>7o0P000002
002So`0400000000I_l006Ko000000<002So00D0000001@0o`0D0?l0503o000000040540o`050000
002?0?l0S`3o08l0o`0000000`3<0?l00`000000o`3e0?l0m@030?l0m@030000003o0;P0o`2h00<0
o`2h00<000000?l0NP3o07X01P3o07X00`000000o`0m0?l0?@0H0?l0?@030000003o07X0o`1j00D0
o`1j00<000000?l0^03o0;P00`3o0;P00`000000o`3e0?l0m@030?l0m@050000003<0?l0c03o0<`0
o`000000102?0?l01@000000D@3o0540o`1A0?l0000000<0503o00D00000000Xo`00:?l002So0000
0003001Vo`0400000000Xol00000000000800>7o0P000002003oh@8000000P00oj<2000000800?mV
0P000002003o:08000000P0Do`0200000080DOl00P00000408oo008000000P3<o`03000000@0ooD0
0`0000050?nh00<000008`3oNP000`000000okP00?nh00080?nh008000001@3om@02000000H0c?l0
0P00000408oo00<00000101Ao`02000000H0oooo0@0000010?ooo`00803oool00`000000oooo0?oo
o`050?ooo`805?l00`000003057o00800000102?o`03000000D0c?l00`0000050?oe00<000001`3o
^002000002@0ogX00P0000070?nh00<000000`3om@0200000080c?l00P00000408oo008000000P1A
o`02000000805?l00P000002003o:08000000P00ofH2000000800?nS00D00000003oh@00on400?oQ
00000003003Qo`0400000000Xol00:?o0000008006Ko0P000002000Xo`8000000P0D0?l2000000<0
D@3o00@0000008l0o`2?0?l00000103<0?l00`000000o`3e0?l0m@030?l0m@030000003o0;P0o`2h
00<0o`2h00<000000?l0NP3o07X01@3o07X00`000000o`0m0?l0?@0G0?l0?@030000003o07X0o`1j
00D0o`1j00<000000?l0^03o0;P0103o0;P00`000000o`3e0?l0m@020?l0m@030000003<0?l0c03o
0080c03o00D0000008l0o`2?0?l0S`3o000000030540o`030000000D0?l0503o008000000P00:?l2
0000008006Ko0P0000000`00Xol00000003Qo`02003Qo`0500000000on400?oQ003oh@0000000P00
oj<2000000800?mV0P000002003o:08000000P0Do`0200000080DOl00P00000408oo008000000P3<
o`02000000@0ooD00`0000060?nh00800000903oNP02000000P0okP00`0000040?oe00<000001@3<
o`03000000@0Sol00P000003057o00@0000000<05?l00?ooo`3oool01@3oool100000040oooo000P
0?ooo`030000003oool0oooo00D0oooo00<0000001Co000Do`000P0Do`0300000080DOl010000003
08oo00@00000103<o`02000000H0ooD00P0000070?nh008000003P3oNP0=000000T0ogX010000006
0?nh008000000`3om@03000000030<co000000000000008000000P2?o`0200000080DOl00P000002
01Co008000000P00obP2000000800?mV0P0000000`00oj<0000000000002003oh@8000000P00hOl0
0`0000000:?o002So`020000008006Ko0P000002000Xo`050000000D0?l0503o01@0o`0000000`1A
0?l00`000000S`3o08l0o`02000000<0c03o0P0000040?l0m@030000003o0;P0o`2h00<0o`2h00<0
00000?l0NP3o07X01@3o07X00`000000o`0m0?l0?@0E0?l0?@030000003o07X0o`1j00H0o`1j00<0
00000?l0^03o0;P00`3o0;P00`000000o`3e0?l0m@020?l0m@040000003<0?l0c03o0<`0o`800000
0P2?0?l200000080D@3o0P00000201@0o`0500000000:?l002So000Xo`0000000P00I_l200000080
0:?o00<00000003Qo`00hOl00P000002003oh@80000000<00?nS0000000000000P00ofH200000080
0?lX0P00000201Co008000000P1Ao`02000000<0Sol00`0000020<co00800000103om@02000000D0
okP0100000090?mj00d000003P3oNP02000000L0okP00`0000050?oe008000001@3<o`03000000<0
Sol010000003057o00800000100Do`02000000H0oooo0@0000010?ooo`00803oool00`000000oooo
0?ooo`050?ooo`03003o:0000000000000@05?l00P000005057o008000001@2?o`02000000@0c?l0
0P0000050?oe00030000003o^000okP000L0okP00P00000<0?mj00030000003o?@00ocd000X0ocd0
100000090?mj008000001P3o^0000`000000ooD00?oe00030?oe008000000`3<o`0010000000Sol0
08oo002?o`0200000080DOl00P00000201Co008000000P00obP200000003003oIP0000000?nS0080
0?nS00D00000003oh@00on400?oQ00000002003Qo`0500000000Xol00:?o002So`0000000`00I_l0
1000000002So000Xo`00000301@0o`050000001A0?l0D@3o0540o`0000000`2?0?l01@000000c03o
0<`0o`3<0?l0000000@0o`3e00<000000?l0^03o0;P00`3o0;P00`000000o`1j0?l0NP050?l0NP03
0000003o03d0o`0m01@0o`0m00<000000?l0NP3o07X01P3o07X00`000000o`2h0?l0^0030?l0^003
0000003o0?D0o`3e0080o`3e00D000000<`0o`3<0?l0c03o0000000308l0o`050000001A0?l0D@3o
0540o`0000000`0D0?l01@00000002So000Xo`00:?l00000008006Ko00D00000002So`00Xol00:?o
00000002003Qo`0500000000on400?oQ003oh@0000000P00oj<01@0000000?mV003oIP00ofH00000
00<00?lX00<0000001Co000Do`000P000002057o008000000`2?o`02000000<0c?l00P0000040?oe
008000001P3o^0000`000000ogX00?mj00070?mj00@00000303o?@000`000000ogX00?mj000:0?mj
008000002@3o^0000`000000ooD00?oe00040?oe00800000103<o`02000000@0Sol00P000005057o
00800000100Do`0200000003003o:03oool0oooo00D0oooo0@0000010?ooo`00803oool00`000000
oooo0?ooo`050?ooo`0300000000obP00?lX00800000100Do`04000000<0DOl00P00000508oo0080
0000103<o`03000000<0ooD00P0000090?nh00030000003oNP00ogX000T0ogX000<000000?lm003o
?@003P3o?@03000000P0ogX00P0000050?nh008000001@3om@001@000000c?l00<co003<o`000000
00@0Sol000D00000057o001Ao`00DOl00000000301Co000700000000obP00?lX003o:00000000?mV
00000003003oX`0500000000on400?oQ003oh@0000000P00hOl01@0000000:?o002So`00Xol00000
008006Ko00D00000000Xo`00:?l002So0000000201@0o`050000001A0?l0D@3o0540o`0000000`2?
0?l00`000000c03o0<`0o`020<`0o`030000003o0?D0o`3e0080o`3e00<000000?l0^03o0;P00`3o
0;P00`000000o`1j0?l0NP050?l0NP030000003o03d0o`0m01<0o`0m00<000000?l0NP3o07X01P3o
07X00`000000o`2h0?l0^0030?l0^0050000003o0?D0o`3e0?l0m@000000103<0?l010000000S`3o
08l0o`0000030540o`050000000D0?l0503o01@0o`0000000`00:?l01000000006Ko001Vo`000003
002So`0400000000hOl00>7o000000<00?oQ00D00000003oX`00oj<00?nS00000002003oIP050000
0000obP00?lX003o:00000000P0Do`001@000000DOl0057o001Ao`00000000@0Sol000@000000<co
003<o`00c?l00P0000040?oe008000001P3o^002000000L0ogX00`00000?0?lm00030000003oNP00
ogX000X0ogX000<000000?nh003o^0001`3o^002000000@0ooD00`0000030<co00<00000102?o`02
000000<0DOl01000000401Co008000000`00obP00`000000oooo0?ooo`040?ooo`4000000@3oool0
0200oooo00<000000?ooo`3oool01@3oool00`00ofH0000000000002003o:0@00000100Do`040000
0004057o000000000000000000@0Sol00`0000040<co008000000`3om@03000000L0okP00P00000:
0?mj00030000003o?@00ocd00100ocd00P0000080?mj00<00000103o^002000000@0ooD00P000002
0<co008000000`2?o`0200000080DOl00P00000201Co008000000P00obP0100000000?mV00000000
0002003oX`8000000P00on40100000000>7o003Qo`000003002So`0400000000I_l006Ko000000<0
02So00@0000001@0o`0D0?l000000`1A0?l01@000000S`3o08l0o`2?0?l0000000<0c03o00<00000
0?l0m@3o0?D00P3o0?D00`000000o`2h0?l0^0030?l0^0030000003o07X0o`1j00D0o`1j00<00000
0?l0?@3o03d04P3o03d00`000000o`1j0?l0NP060?l0NP030000003o0;P0o`2h00<0o`2h00<00000
0?l0m@3o0?D00P3o0?D01@000000c03o0<`0o`3<0?l0000000<0S`3o00D000000540o`1A0?l0D@3o
0000000301@0o`0400000000:?l002So0000008006Ko00D00000002So`00Xol00:?o00000002003Q
o`0300000000on400?oQ008000000P00oj<200000003003oIP000000000000800?lX0P00000201Co
00030000001Ao`00DOl0008000000`2?o`0200000080c?l00P0000040?oe008000001@3o^0030000
00L0ogX00P00000B0?lm00030000003oNP00ogX000P0ogX00P0000070?nh00<00000103om@020000
00<0c?l00`00000408oo00<000000P1Ao`03000000@05?l010000003003o:08000001`3oool10000
0040oooo000P0?ooo`030000003oool0oooo00D0oooo00<00000003oIP00ofH010000002003o:0D0
00000`0Do`0200000080DOl00`00000408oo00800000103<o`02000000@0ooD00`0000060?nh0003
0000003oNP00ogX000L0ogX000<000000?lm003o?@004P3o?@04000000L0ogX00P0000040?nh0080
0000103om@0200000080c?l00P00000308oo00050000001Ao`00DOl0057o000000000`0Do`001000
00000?lX003o:0000002003oIP8000000P00oj<0100000000?oQ003oh@000002003Qo`8000000P00
Xol01000000006Ko001Vo`000003000Xo`040000000D0?l0503o000000<0D@3o00D0000008l0o`2?
0?l0S`3o000000030<`0o`030000003o0?D0o`3e0080o`3e00<000000?l0^03o0;P00`3o0;P00`00
0000o`1j0?l0NP050?l0NP030000003o03d0o`0m0140o`0m00<000000?l0NP3o07X01P3o07X00`00
0000o`2h0?l0^0020?l0^0030000003o0?D0o`3e0080o`3e00<000000<`0o`3<0?l00P3<0?l01000
0000S`3o08l0o`0000030540o`050000000D0?l0503o01@0o`0000000P00:?l01000000006Ko001V
o`000003002So`0400000000hOl00>7o000000800?oQ00D00000003oX`00oj<00?nS00000002003o
IP0500000000obP00?lX003o:00000000P0Do`0200000080DOl000@0000008oo002?o`00Sol00P00
00030<co00030000003om@00ooD00080ooD00P0000040?nh00<000001P3oNP04000001@0ocd000<0
00000?mj003oNP001`3oNP000`000000okP00?nh00050?nh00800000103om@03000000<0c?l00P00
000408oo00<000000`1Ao`02000000805?l01@000003003o:0@000000P00ofH00`000000oooo0?oo
o`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool00`00oj<0000000000004
003oIP8000001@00obP2000000<05?l00P000003057o00800000102?o`02000000@0c?l00P000004
0?oe00030000003o^000okP000D0okP000<000000?mj003oNP001`3oNP02000001H0ocd00P000007
0?mj00800000103o^002000000<0ooD000<000000<co003<o`000P3<o`001@000000Sol008oo002?
o`00000000<0DOl000D0000001Co000Do`005?l000000002003o:00400000000ofH00?mV000000<0
0?nS00@00000003oh@00on4000000`00hOl0100000000:?o002So`000002001Vo`0500000000:?l0
02So000Xo`0000000P0D0?l010000000D@3o0540o`00000308l0o`050000003<0?l0c03o0<`0o`00
0000103o0?D00`000000o`2h0?l0^0030?l0^0030000003o07X0o`1j00D0o`1j00<000000?l0?@3o
03d04@3o03d00`000000o`1j0?l0NP060?l0NP050000003o0;P0o`2h0?l0^00000001@3o0?D01@00
0000c03o0<`0o`3<0?l0000000<0S`3o00D000000540o`1A0?l0D@3o0000000201@0o`0400000000
:?l002So0000008006Ko00D00000002So`00Xol00:?o00000002003Qo`0400000000on400?oQ0000
00<00?nS00@00000003oIP00ofH000000`00obP0100000005?l001Co00000003057o00050000002?
o`00Sol008oo00000000103<o`0010000000ooD00?oe003om@02000000@0okP00P0000080?mj0003
0000003o?@00ocd001D0ocd000<000000?mj003oNP001`3oNP000`000000okP00?nh00050?nh0003
0000003om@00ooD00080ooD00P0000050<co00030000002?o`00Sol00080Sol00P000003057o00<0
00000P0Do`02000000D00?lX0`000004003oIP8000001`3oool100000040oooo000P0?ooo`030000
003oool0oooo00D0oooo00<00000003oX`00oj<00P000004003oIP@000000`00obP2000000<05?l0
0`000002057o00<000000`2?o`02000000@0c?l000<000000?oe003om@000P3om@02000000H0okP0
00<000000?mj003oNP00203oNP000`000000ocd00?lm000E0?lm008000001`3oNP02000000@0okP0
00@000000?oe003om@00ooD00P0000030<co008000000P2?o`0200000080DOl00P00000201Co0004
00000000obP00?lX000000800?mV00D00000003oX`00oj<00?nS00000002003oh@0400000000hOl0
0>7o000000800:?o00@00000001Vo`00I_l000000`00:?l01@000000503o01@0o`0D0?l000000080
D@3o00D0000008l0o`2?0?l0S`3o000000030<`0o`030000003o0?D0o`3e0080o`3e00<000000?l0
^03o0;P00`3o0;P00`000000o`1j0?l0NP050?l0NP030000003o03d0o`0m00l0o`0m00<000000?l0
NP3o07X01P3o07X00`000000o`2h0?l0^0020?l0^0030000003o0?D0o`3e0080o`3e00D000000<`0
o`3<0?l0c03o0000000308l0o`050000001A0?l0D@3o0540o`0000000P0D0?l01000000002So000X
o`000002001Vo`0500000000Xol00:?o002So`0000000P00hOl0100000000?oQ003oh@000003003o
X`0400000000ofH00?mV000000800?lX0P00000201Co00050000001Ao`00DOl0057o000000000P2?
o`02000000<0c?l00P0000030?oe00030000003o^000okP00080okP00P0000080?mj008000005`3o
?@000`000000ogX00?mj00070?mj00030000003o^000okP000@0okP00P0000040?oe00030000003<
o`00c?l000<0c?l00P00000308oo008000000P1Ao`03000000<05?l00P000003003o:0@000001000
ofH3000000800?nS00<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0
oooo00D0oooo00<00?oQ0000000000000P00oj<4000000@00?mV0`000002003o:0<000000`0Do`02
000000<0DOl00`00000208oo008000000`3<o`02000000@0ooD00`0000040?nh008000002@3oNP00
0`000000ocd00?lm000F0?lm00<000001P3oNP02000000<0okP00P0000030?oe008000000`3<o`00
1@000000Sol008oo002?o`00000000<0DOl000@0000001Co000Do`0000000P00obP00`0000000?mV
003oIP02000000800?nS00@00000003oh@00on4000000P00hOl0100000000:?o002So`000002001V
o`0500000000:?l002So000Xo`0000000P0D0?l010000000D@3o0540o`00000308l0o`050000003<
0?l0c03o0<`0o`000000103o0?D00`000000o`2h0?l0^0030?l0^0030000003o07X0o`1j00D0o`1j
00<000000?l0?@3o03d03`3o03d00`000000o`1j0?l0NP060?l0NP050000003o0;P0o`2h0?l0^000
00001@3o0?D01@000000c03o0<`0o`3<0?l0000000<0S`3o00@000000540o`1A0?l000000`0D0?l0
1000000002So000Xo`000002001Vo`0400000000Xol00:?o000000800>7o00@00000003oh@00on40
00000P00oj<2000000800?mV00@00000003o:000obP000000`0Do`001@000000DOl0057o001Ao`00
00000080Sol000@000000<co003<o`00c?l00P0000030?oe008000000`3o^002000000L0ogX00`00
000H0?lm00030000003oNP00ogX000H0ogX00P0000040?nh00<00000103om@02000000<0c?l00`00
000208oo00<000000P1Ao`02000000<05?l00`000002003o:0<000001000ofH4000000<00?nS0P00
00070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool3003oh@<000000`00oj<5
000000800?mV0P000003003o:0<000000P0Do`0400000080DOl00P00000208oo00<000000P3<o`03
000000@0ooD00P0000040?nh00800000203oNP000`000000ocd00?lm000H0?lm008000001P3oNP00
0`000000okP00?nh00020?nh008000000`3om@0200000080c?l00P00000208oo00050000001Ao`00
DOl0057o000000000P0Do`000`0000000?lX003o:002000000800?mV00@00000003oX`00oj<00000
0P00on40100000000>7o003Qo`000002002So`0400000000I_l006Ko000000<002So00@0000001@0
o`0D0?l000000P1A0?l01@000000S`3o08l0o`2?0?l0000000<0c03o00<000000?l0m@3o0?D00P3o
0?D00`000000o`2h0?l0^0020?l0^0030000003o07X0o`1j00H0o`1j00<000000?l0?@3o03d03@3o
03d00`000000o`1j0?l0NP060?l0NP030000003o0;P0o`2h0080o`2h00<000000?l0m@3o0?D00P3o
0?D01@000000c03o0<`0o`3<0?l0000000<0S`3o00@000000540o`1A0?l000000`0D0?l010000000
02So000Xo`000002001Vo`0400000000Xol00:?o000000800>7o00@00000003oh@00on4000000P00
oj<01@0000000?mV003oIP00ofH0000000800?lX00<0000001Co000Do`000P000002057o00800000
0P2?o`001@000000c?l00<co003<o`00000000<0ooD00P0000040?nh00030000003oNP00ogX000@0
ogX00`00000J0?lm00030000003oNP00ogX000D0ogX00P0000040?nh00800000103om@0300000080
c?l00`00000308oo008000000P1Ao`03000000805?l00`000003003o:08000000`00ofH4000000<0
0?nS10000002003oh@L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`800000
1000on43000000D00?nS0P000002003oIP<000000`00obP2000000@05?l00P000002057o00030000
002?o`00Sol00080Sol00P0000030<co00800000103om@000`000000okP00?nh00030?nh00800000
1`3oNP000`000000ocd00?lm000I0?lm008000001@3oNP000`000000okP00?nh00030?nh00800000
0`3om@001@000000c?l00<co003<o`0000000080Sol00P000002057o00040000000Do`005?l00000
00<00?lX00<00000003oIP0000000`00oj<0100000000?oQ003oh@000002003Qo`0400000000Xol0
0:?o0000008006Ko00@00000000Xo`00:?l000000P0D0?l01@000000D@3o0540o`1A0?l000000080
S`3o00D000000<`0o`3<0?l0c03o000000040?l0m@030000003o0;P0o`2h0080o`2h00<000000?l0
NP3o07X01P3o07X00`000000o`0m0?l0?@0<0?l0?@030000003o07X0o`1j00H0o`1j00<000000?l0
^03o0;P00P3o0;P00`000000o`3e0?l0m@020?l0m@050000003<0?l0c03o0<`0o`0000000`2?0?l0
10000000D@3o0540o`00000301@0o`0700000000:?l002So000Xo`00000006Ko00000002002So`04
00000000hOl00>7o000000800?oQ00D00000003oX`00oj<00?nS00000002003oIP0400000000obP0
0?lX000000805?l000D00000057o001Ao`00DOl00000000308oo00030000003<o`00c?l000800000
0P3om@02000000D0okP000<000000?mj003oNP000`3oNP02000001/0ocd00P0000070?mj00030000
003o^000okP000<0okP000<000000?oe003om@000P3om@02000000<0c?l00P00000408oo00800000
0P1Ao`02000000<05?l00P000003003o:0<000000P00ofH3000000@00?nS0`000005003oh@800000
1P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo0P00hOl3000000@00?oQ1000
0003003oX`8000000`00ofH3000000800?lX0`00000301Co0080000000<0DOl0000000000000102?
o`02000000<0c?l000<000000?oe003om@000P3om@02000000D0okP000<000000?mj003oNP001@3o
NP000`000000ocd00?lm000J0?lm00800000103oNP02000000D0okP000D000000?oe003om@00ooD0
000000030<co00050000002?o`00Sol008oo000000000P1Ao`00100000005?l001Co00000002003o
:00400000000ofH00?mV000000<00?nS00H00000003oh@00on400000003Qo`000002002So`050000
0000I_l006Ko001Vo`0000000P00:?l010000000503o01@0o`0000020540o`050000002?0?l0S`3o
08l0o`0000000`3<0?l00`000000o`3e0?l0m@020?l0m@030000003o0;P0o`2h0080o`2h00<00000
0?l0NP3o07X01P3o07X00`000000o`0m0?l0?@0;0?l0?@030000003o07X0o`1j00H0o`1j00<00000
0?l0^03o0;P00P3o0;P00`000000o`3e0?l0m@020?l0m@050000003<0?l0c03o0<`0o`0000000`2?
0?l010000000D@3o0540o`00000301@0o`0600000000:?l002So00000000I_l000000P00Xol01000
00000>7o003Qo`000002003oh@0500000000oj<00?nS003oX`0000000P00ofH0100000000?lX003o
:000000201Co00050000001Ao`00DOl0057o000000000`2?o`0010000000c?l00<co000000030?oe
00030000003o^000okP000<0okP00P0000040?mj00800000703o?@000`000000ogX00?mj00060?mj
00030000003o^000okP00080okP00P0000040?oe00040000003<o`00c?l00<co00800000102?o`02
00000080DOl00P00000301Co008000000P00obP3000000<00?mV0P000003003oX`@000001000on44
00000003003Qo`3oool0oooo00D0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oo
o`8000000`00hOl4000000@00?oQ0`000002003oX`<000000`00ofH2000000<00?lX0`00000201Co
0080000000@0DOl000000000000000000`2?o`0200000080c?l00P0000040?oe00800000103o^002
000000H0ogX00P00000L0?lm00800000103oNP02000000@0okP000@000000?oe003om@00ooD00P00
00020<co00050000002?o`00Sol008oo000000000P1Ao`00100000005?l001Co00000002003o:004
00000000ofH00?mV000000800?nS00@00000003oh@00on4000000P00hOl0100000000:?o002So`00
0002001Vo`0400000000:?l002So00000080503o00D000000540o`1A0?l0D@3o0000000208l0o`05
0000003<0?l0c03o0<`0o`000000103o0?D00`000000o`2h0?l0^0020?l0^0030000003o07X0o`1j
00H0o`1j00<000000?l0?@3o03d02P3o03d00`000000o`1j0?l0NP060?l0NP030000003o0;P0o`2h
00<0o`2h00D000000?l0m@3o0?D0o`3e000000030<`0o`050000002?0?l0S`3o08l0o`0000000P1A
0?l01@000000503o01@0o`0D0?l00000008002So00@00000001Vo`00I_l000000P00Xol00`000000
0>7o00000003003oh@0400000000oj<00?nS000000800?mV00@00000003o:000obP000000P0Do`00
0`000000DOl0057o000200000080Sol00P0000020<co00050000003om@00ooD00?oe00000000103o
^002000000@0ogX00P00000L0?lm008000001`3oNP02000000<0okP00P0000040?oe008000000P3<
o`02000000<0Sol00`000002057o008000000P0Do`03000000800?lX0P000003003oIP<000000P00
oj<3000000@00?oQ10000004003Qo`8000001P3oool100000040oooo000P0?ooo`030000003oool0
oooo00D0oooo0P00Xol4000000<00>7o10000003003oh@8000000`00oj<3000000800?mV0`000003
003o:0040000000Do`005?l001Co008000000P1Ao`0300000080Sol00P0000020<co00<000000`3o
m@02000000@0okP00P0000060?mj00030000003o?@00ocd001/0ocd000<000000?mj003oNP000`3o
NP000`000000okP00?nh00020?nh00030000003om@00ooD00080ooD000@000000<co003<o`000000
0`2?o`0010000000DOl0057o0000000201Co000400000000obP00?lX000000800?mV00@00000003o
X`00oj<000000P00on40100000000>7o003Qo`000002002So`0400000000I_l006Ko0000008002So
00@0000001@0o`0D0?l000000`1A0?l010000000S`3o08l0o`0000030<`0o`030000003o0?D0o`3e
0080o`3e00<000000?l0^03o0;P00P3o0;P00`000000o`1j0?l0NP060?l0NP8000002P3o03d00`00
0000o`1j0?l0NP060?l0NP030000003o0;P0o`2h00<0o`2h00D000000?l0m@3o0?D0o`3e00000003
0<`0o`050000002?0?l0S`3o08l0o`0000000P1A0?l010000000503o01@0o`000002000Xo`040000
0000I_l006Ko000000800:?o00@00000003Qo`00hOl000000P00on40100000000?nS003oX`000002
003oIP0400000000obP00?lX000000805?l000@00000057o001Ao`0000000`2?o`001@000000c?l0
0<co003<o`00000000<0ooD000<000000?nh003o^0000P3o^0000`000000ogX00?mj00030?mj0003
0000003o?@00ocd001/0ocd000<000000?mj003oNP001@3oNP02000000<0okP00P0000040?oe0080
00000P3<o`0200000080Sol00`000003057o008000000P0Do`02000000800?lX0`000002003oIP<0
00000`00oj<2000000<00?oQ10000003003Qo`D0000000<00:?o0?ooo`3oool01@3oool100000040
oooo000P0?ooo`030000003oool0oooo00D0oooo0P000004002So`<000001000hOl3000000800?oQ
0`000003003oX`8000000`00ofH300000003003o:0000000000000<05?l000<00000057o001Ao`00
0P1Ao`0200000080Sol00P0000030<co008000000`3om@02000000@0okP000<000000?mj003oNP00
103oNP000`000000ocd00?lm000K0?lm00030000003oNP00ogX000<0ogX00P0000030?nh00800000
0`3om@000`000000c?l00<co000200000080Sol000@00000057o001Ao`0000000P0Do`0010000000
0?lX003o:0000002003oIP0400000000oj<00?nS000000800?oQ00@00000003Qo`00hOl000000P00
Xol01P00000006Ko001Vo`00000002So000000<0503o00@000000540o`1A0?l000000P2?0?l01@00
0000c03o0<`0o`3<0?l0000000@0o`3e00<000000?l0^03o0;P00`3o0;P00`000000o`1j0?l0NP07
0?l0NP030000003o03d0o`0m00H0o`0m00<000000?l0NP3o07X01`3o07X00`000000o`2h0?l0^002
0?l0^0050000003o0?D0o`3e0?l0m@0000000`3<0?l01@000000S`3o08l0o`2?0?l000000080D@3o
00@0000001@0o`0D0?l000000P00:?l01`00000006Ko001Vo`00I_l00000002So`0000000P00hOl0
100000000?oQ003oh@000002003oX`0400000000ofH00?mV000000800?lX00@0000001Co000Do`00
00000P1Ao`001@000000Sol008oo002?o`00000000<0c?l000<000000?oe003om@000P0000030?nh
008000001@3oNP000`000000ocd00?lm000K0?lm00030000003oNP00ogX000D0ogX000<000000?nh
003o^0000P3o^0000`000000ooD00?oe00020?oe008000000P3<o`0200000080Sol00P000004057o
008000000P0Do`02000000800?lX0P000003003oIP8000000`00oj<3000000800?oQ0`000004003Q
o`<000001@00Xol2000000H0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`80
06Ko0`000004002So`<000001000hOl2000000<00?oQ0`000002003oX`<000000`00ofH200000003
003o:0000000000000805?l00P000004057o008000000P2?o`02000000<0c?l000<000000?oe003o
m@000P3om@02000000<0okP000<000000?mj003oNP00103oNP0200000100ocd00P00000:0?lm0080
00001@3oNP000`000000okP00?nh00020?nh00050000003om@00ooD00?oe000000000`3<o`001000
0000Sol008oo00000002057o00040000000Do`005?l0000000800?lX00@00000003oIP00ofH00000
0P00oj<01P0000000?oQ003oh@0000000>7o000000<00:?o00H00000001Vo`00I_l00000000Xo`00
000201@0o`050000001A0?l0D@3o0540o`0000000P2?0?l01@000000c03o0<`0o`3<0?l0000000<0
o`3e00<000000?l0^03o0;P00`3o0;P00`000000o`1j0?l0NP070?l0NP030000003o03d0o`0m00D0
o`0m00<000000?l0NP3o07X01`3o07X00`000000o`2h0?l0^0030?l0^0050000003o0?D0o`3e0?l0
m@0000000`3<0?l01@000000S`3o08l0o`2?0?l000000080D@3o00@0000001@0o`0D0?l000000P00
:?l01P00000006Ko001Vo`0000000:?o000000800>7o00@00000003oh@00on4000000P00oj<01000
00000?mV003oIP000002003o:0040000000Do`005?l000000080DOl000D0000008oo002?o`00Sol0
000000030<co00040000003om@00ooD0000000@0okP000<000000?mj003oNP000`3oNP02000000/0
ocd00P00000?0?lm008000001P3oNP02000000<0okP00P0000030?oe008000000P3<o`0200000080
Sol00P000004057o008000000P0Do`02000000800?lX0P000003003oIP8000000P00oj<3000000<0
0?oQ0P000004003Qo`<000001000Xol400000003001Vo`3oool0oooo00D0oooo0@0000010?ooo`00
803oool00`000000oooo0?ooo`050?ooo`D006Ko0`000004002So`8000001000hOl3000000<00?oQ
0P000003003oX`<000000P00ofH00`0000000?lX003o:0030000000401Co000000000000000000<0
DOl00P00000208oo008000000P3<o`02000000@0ooD00P0000020?nh008000001P3oNP000`000000
ocd00?lm000;0?lm008000000P3o0003000000T0ocd00P0000040?mj008000000`3o^0001@000000
ooD00?oe003om@00000000<0c?l000@0000008oo002?o`0000000P1Ao`00100000005?l001Co0000
0002003o:00600000000ofH00?mV00000000oj<000000`00on400`0000000>7o00000002002So`06
00000000I_l006Ko00000000:?l000000`0D0?l010000000D@3o0540o`00000208l0o`050000003<
0?l0c03o0<`0o`000000103o0?D00`000000o`2h0?l0^0030?l0^0030000003o07X0o`1j00L0o`1j
00<000000?l0?@3o03d00`3o03d00`000000o`1j0?l0NP080?l0NP030000003o0;P0o`2h0080o`2h
00D000000?l0m@3o0?D0o`3e000000030<`0o`050000002?0?l0S`3o08l0o`0000000P1A0?l01000
0000503o01@0o`000002000Xo`0900000000I_l006Ko001Vo`0000000:?o00000000hOl000000080
0?oQ00L00000003oX`00oj<00?nS00000000ofH0000000800?lX00@0000001Co000Do`0000000P1A
o`001@000000Sol008oo002?o`0000000080c?l00P0000020?oe00030000003o^000okP00080okP0
00<000000?mj003oNP000P3oNP02000000X0ocd00`0000020?l000800000303o?@000`000000ogX0
0?mj00050?mj00040000003o^000okP00?nh008000000`3om@0200000080c?l00P00000208oo0080
00000`1Ao`03000000805?l00P000002003o:08000000P00ofH3000000800?nS0P000003003oh@<0
00001000hOl2000000@00:?o0`000005001Vo`L0oooo0@0000010?ooo`00803oool00`000000oooo
0?ooo`050?ooo`P006Ko0P000004002So`@000000`00hOl3000000800?oQ0`000003003oX`030000
0000ofH00?mV008000000`00obP2000000805?l00`000002057o008000000P2?o`0200000080c?l0
0P0000030?oe00030000003o^000okP00080okP00P0000050?mj00030000003o?@00ocd000P0ocd0
0P0000070?l0008000002@3o?@000`000000ogX00?mj00030?mj00040000003o^000okP00?nh0080
00000P3om@0200000080c?l000@0000008oo002?o`0000000P1Ao`00100000005?l001Co00000002
003o:00600000000ofH00?mV00000000oj<000000P00on40100000000>7o003Qo`000002002So`05
00000000I_l006Ko00000000:?l00P00000201@0o`040000001A0?l0D@3o00000080S`3o00D00000
0<`0o`3<0?l0c03o000000030?l0m@030000003o0;P0o`2h00<0o`2h00<000000?l0NP3o07X0203o
07X5000000X0o`1j00<000000?l0^03o0;P00`3o0;P01@000000o`3e0?l0m@3o0?D0000000<0c03o
00D0000008l0o`2?0?l0S`3o000000020540o`040000000D0?l0503o0000008002So00H00000001V
o`00I_l00000002So`000002003Qo`0400000000on400?oQ000000800?nS00<00000003oIP000000
0P00obP0100000005?l001Co00000002057o00050000002?o`00Sol008oo000000000P3<o`001@00
0000ooD00?oe003om@00000000@0okP000<000000?mj003oNP000P3oNP000`000000ocd00?lm0008
0?lm008000001`3o00000`000000ocd00?lm00080?lm00030000003oNP00ogX000@0ogX00P000003
0?nh00030000003om@00ooD00080ooD000<000000<co003<o`000P00000208oo008000000P1Ao`03
000000<05?l00P000002003o:08000000P00ofH2000000800?nS0`000002003oh@<000000`00hOl4
000000@00:?o0P000008001Vo`L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?oo
o`D000001@00I_l3000000D00:?o0`000003003Qo`8000000`00on4300000003003oX`0000000000
00800?mV0P000003003o:0030000000Do`005?l000805?l00P000002057o008000000P2?o`020000
0080c?l00P0000020?oe00800000103o^0000`000000ogX00?mj00030?mj008000002@3o?@000`00
0000o`000?l000070?l000030000003o?@00ocd000L0ocd00P0000040?mj00030000003o^000okP0
0080okP000L000000?oe003om@00ooD00000003<o`00000000<0Sol000@00000057o001Ao`000000
0P0Do`001P0000000?lX003o:00000000?mV000000800?nS00@00000003oh@00on4000000P00hOl0
1P0000000:?o002So`00000006Ko0000008002So00@0000001@0o`0D0?l000000P1A0?l010000000
S`3o08l0o`0000030<`0o`030000003o0?D0o`3e0080o`3e00<000000?l0^03o0;P00P3o0;P00`00
0000o`1j0?l0NP0G0?l0NP030000003o0;P0o`2h00<0o`2h00D000000?l0m@3o0?D0o`3e00000002
0<`0o`050000002?0?l0S`3o08l0o`0000000P1A0?l010000000503o01@0o`000002000Xo`040000
0000I_l006Ko000000800:?o00<00000003Qo`0000000P00on401P0000000?nS003oX`0000000?mV
000000800?lX00@0000001Co000Do`0000000P1Ao`001@000000Sol008oo002?o`0000000080c?l0
00D000000?oe003om@00ooD0000000030?nh008000000`3oNP02000000T0ocd00P0000080?l00003
0000003o?@00ocd000L0ocd00P0000050?mj008000000`3o^002000000<0ooD00P0000000`3<o`00
00000000000208oo008000000P1Ao`02000000@05?l00P000002003o:08000000P00ofH200000080
0?nS0P000003003oh@8000000`00hOl3000000D00:?o0`000005001Vo`H000001P3oool100000040
oooo000P0?ooo`030000003oool0oooo00D0oooo1@00:?l3000000D006Ko10000004002So`<00000
0P00hOl3000000<00?oQ0P0000000`00oj<0000000000002003oIP8000000P00obP2000000@05?l0
0P000002057o008000000P2?o`0200000080c?l00P0000020?oe00030000003o^000okP00080okP0
00<000000?mj003oNP00103oNP000`000000ocd00?lm00060?lm00030000003o0000o`0000P0o`00
0P0000090?lm00030000003oNP00ogX00080ogX000D000000?nh003o^000okP0000000030?oe0004
0000003<o`00c?l0000000<0Sol000H00000057o001Ao`00000001Co00000002003o:00400000000
ofH00?mV000000800?nS00<00000003oh@0000000P00hOl0100000000:?o002So`000002001Vo`04
00000000:?l002So00000080503o00@000000540o`1A0?l000000P2?0?l01@000000c03o0<`0o`3<
0?l0000000<0o`3e00<000000?l0^03o0;P00P3o0;P00`000000o`1j0?l0NP0F0?l0NP030000003o
0;P0o`2h00<0o`2h00D000000?l0m@3o0?D0o`3e000000030<`0o`050000002?0?l0S`3o08l0o`00
00000P1A0?l01P000000503o01@0o`00000002So000000<006Ko00D00000002So`0000000>7o0000
0002003oh@0600000000oj<00?nS00000000ofH000000P00obP0100000005?l001Co00000002057o
00050000002?o`00Sol008oo000000000P3<o`001@000000ooD00?oe003om@00000000<0okP000<0
00000?mj003oNP000P3oNP000`000000ocd00?lm00070?lm008000002@3o00000`000000ocd00?lm
00070?lm00030000003oNP00ogX000@0ogX000<000000?nh003o^0000P3o^00010000000ooD00?oe
003om@0200000080c?l000<0000008oo002?o`000P000002057o00800000100Do`02000000800?lX
0P000002003oIP8000000P00oj<2000000<00?oQ0P000002003Qo`<000001000Xol4000000D006Ko
0`000005000Xo`L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`P002So0P00
0007001Vo`<000001000Xol2000000<00>7o0`000002003oh@0300000000oj<00?nS00<0000000D0
0?mV0000000000000000003o:003000000<05?l000D00000057o001Ao`00DOl00000000308oo0005
0000003<o`00c?l00<co000000000P3om@02000000<0okP00P0000050?mj00030000003o?@00ocd0
00H0ocd000<000000?l0003o00002@3o0002000000P0ocd000<000000?mj003oNP000P3oNP001@00
0000okP00?nh003o^000000000<0ooD000@000000<co003<o`0000000P2?o`0010000000DOl0057o
0000000201Co000400000000obP00?lX000000800?mV00H00000003oX`00oj<00000003oh@000002
003Qo`0600000000Xol00:?o00000000I_l000000P00:?l010000000503o01@0o`0000020540o`04
0000002?0?l0S`3o000000<0c03o00<000000?l0m@3o0?D00P3o0?D00`000000o`2h0?l0^0020?l0
^0030000003o07X0o`1j01D0o`1j00<000000?l0^03o0;P00`3o0;P01@000000o`3e0?l0m@3o0?D0
00000080c03o00D0000008l0o`2?0?l0S`3o000000020540o`040000000D0?l0503o0000008002So
00@00000001Vo`00I_l000000P00Xol01@0000000>7o00000000on40000000800?nS00@00000003o
IP00ofH000000P00obP00`0000005?l000000003057o00040000002?o`00Sol000000080c?l000D0
00000?oe003om@00ooD0000000030?nh00030000003oNP00ogX00080ogX000<000000?lm003o?@00
1P3o?@02000000/0o`0000<000000?lm003o?@001P3o?@000`000000ogX00?mj00030?mj00800000
0`3o^00200000080ooD00P0000020<co008000000P2?o`001@000000DOl0057o001Ao`00000000@0
5?l00P000002003o:08000000P00ofH2000000800?nS0P000002003oh@<000000P00hOl2000000@0
0:?o0`000007001Vo`8000002000:?l70?ooo`4000000@3oool00200oooo0P0000060?ooo`X002So
10000006001Vo`8000001000Xol3000000<00>7o00<00000003oh@00on400P000003003oX`800000
0P00ofH300000003003o:0000000000000805?l000D00000057o001Ao`00DOl00000000308oo0005
0000003<o`00c?l00<co000000000`3om@02000000<0okP000<000000?mj003oNP000`3oNP000`00
0000ocd00?lm00060?lm008000002`3o0002000000L0ocd000<000000?mj003oNP000P3oNP020000
0080okP00P0000020?oe00040000003<o`00c?l000000080Sol000@00000057o001Ao`0000000P0D
o`00100000000?lX003o:0000002003oIP0300000000oj<0000000800?oQ00<00000003Qo`000000
0P00Xol01000000006Ko001Vo`000002000Xo`040000000D0?l0503o00000080D@3o00@0000008l0
o`2?0?l000000`3<0?l01@000000o`3e0?l0m@3o0?D0000000@0o`2h00<000000?l0NP3o07X05@3o
07X00`000000o`2h0?l0^0020?l0^0040000003o0?D0o`3e0?l0m@8000000P3<0?l01@000000S`3o
08l0o`2?0?l000000080D@3o00H0000001@0o`0D0?l00000000Xo`000002001Vo`0600000000Xol0
0:?o00000000hOl000000P00on401P0000000?nS003oX`0000000?mV000000800?lX00@0000001Co
000Do`0000000P1Ao`0010000000Sol008oo000000020<co00050000003om@00ooD00?oe00000000
0`3o^0000`000000ogX00?mj00020?mj00030000003o?@00ocd000D0ocd00P00000<0?l000030000
003o?@00ocd000H0ocd000<000000?mj003oNP000`3oNP0010000000okP00?nh003o^002000000<0
ooD000D000000<co003<o`00c?l00000000308oo00050000001Ao`00DOl0057o000000000`0Do`06
000000800?mV0P000002003oX`8000000P00on42000000<00>7o0P000004002So`8000001P00I_l4
000000X002So1P3oool200000040oooo000P0?ooo`030000003oool0oooo00D0oooo3P00:?l30000
00D006Ko0`000004002So`<0000000<00>7o0000000000000P00on43000000800?nS0P000003003o
IP0400000000obP00?lX000000805?l00P000002057o008000000P2?o`0200000080c?l000<00000
0?oe003om@000P3om@0010000000okP00?nh003o^002000000@0ogX00P0000080?lm00030000003o
0000o`0000X0o`000P0000060?lm00800000103oNP001@000000okP00?nh003o^00000000080ooD0
00@000000<co003<o`0000000P2?o`0010000000DOl0057o0000000201Co000400000000obP00?lX
000000800?mV00<00000003oX`0000000P00on401@0000000>7o00000000Xol00000008006Ko00@0
0000000Xo`00:?l000000P0D0?l010000000D@3o0540o`00000208l0o`050000003<0?l0c03o0<`0
o`0000000`3o0?D00`000000o`2h0?l0^0020?l0^0030000003o07X0o`1j01@0o`1j00<000000?l0
^03o0;P00`3o0;P01@000000o`3e0?l0m@3o0?D0000000<0c03o00@0000008l0o`2?0?l000000P1A
0?l010000000503o01@0o`000002000Xo`0300000000I_l0000000800:?o00H00000003Qo`00hOl0
0000003oh@000002003oX`0300000000ofH0000000800?lX00@0000001Co000Do`0000000P1Ao`00
10000000Sol008oo000000020<co00050000003om@00ooD00?oe000000000`3o^00010000000ogX0
0?mj003oNP02000000H0ocd00P00000=0?l000030000003o?@00ocd000D0ocd00P0000040?mj0080
00000`3o^0000`000000ooD00?oe00020?oe00050000003<o`00c?l00<co000000000`2?o`000`00
0000DOl0057o0002000000805?l00P0000000`00obP00000003oIP02003oIP<0000000@00?nS0000
0000000000000P00on42000000800>7o0`000003002So`<000001@00I_l3000000h002So1`3oool1
00000040oooo000P0?ooo`030000003oool0oooo00D0oooo4@00:?l4000000@006Ko0`000004002S
o`80000000<00>7o0000000000000`00on4300000005003oX`000000000000000000ofH00P000000
0`00obP000000000000201Co008000000P1Ao`0200000080Sol00P0000000`3<o`00000000000003
0?oe008000000`3o^002000000@0ogX000<000000?lm003o?@001@3o?@000`000000o`000?l0000<
0?l000030000003o?@00ocd000D0ocd000D000000?mj003oNP00ogX0000000030?nh00040000003o
m@00ooD0000000<0c?l000@0000008oo002?o`0000000P1Ao`000`0000005?l000000003003o:005
00000000ofH00000003oX`0000000P00on401P0000000>7o003Qo`0000000:?o0000008006Ko00H0
0000000Xo`00:?l0000001@0o`0000020540o`050000002?0?l0S`3o08l0o`0000000P3<0?l01@00
0000o`3e0?l0m@3o0?D0000000D0o`2h00<000000?l0NP3o07X04`3o07X00`000000o`2h0?l0^002
0?l0^0050000003o0?D0o`3e0?l0m@0000000`3<0?l01@000000S`3o08l0o`2?0?l000000080D@3o
00H0000001@0o`0D0?l00000000Xo`000002001Vo`0600000000Xol00:?o00000000hOl000000P00
on400`0000000?nS00000002003oIP0300000000obP0000000805?l000@00000057o001Ao`000000
0P2?o`0010000000c?l00<co000000030?oe00050000003o^000okP00?nh000000000`3oNP000`00
0000ocd00?lm00050?lm00030000003o0000o`0000`0o`0000<000000?lm003o?@001@3o?@000`00
0000ogX00?mj00020?mj008000000`3o^002000000<0ooD00P0000020<co008000000P2?o`020000
0003057o00000000000000805?l00P0000001000obP0000000000000ofH300000004003oX`000000
0000000000<00?oQ0P000002003Qo`8000001000Xol2000000@006Ko1000000A000Xo`L0oooo0@00
00010?ooo`00803oool00`000000oooo0?ooo`050?oooaD002So0`000004001Vo`8000001000Xol0
0`0000000>7o003Qo`02000000@00?oQ00D00000003oX`00oj<00?nS00000002003oIP0300000000
obP00?lX008000000P0Do`0200000080DOl00P00000208oo00030000003<o`00c?l0008000000`3o
m@000`000000okP00?nh00020?nh00030000003oNP00ogX00080ogX000<000000?lm003o?@001@3o
?@000`000000o`000?l0000<0?l000030000003o?@00ocd000@0ocd000<000000?mj003oNP000P3o
NP001@000000okP00?nh003o^00000000080ooD000D000000<co003<o`00c?l00000000208oo0006
0000001Ao`00DOl00000000Do`0000000P00obP01P0000000?mV003oIP0000000?nS000000800?oQ
00<00000003Qo`00hOl00P000002001Vo`0400000000:?l002So00000080503o00@000000540o`1A
0?l000000P2?0?l01@000000c03o0<`0o`3<0?l0000000<0o`3e00<000000?l0^03o0;P00P3o0;P0
0`000000o`1j0?l0NP0B0?l0NP030000003o0;P0o`2h00<0o`2h00D000000?l0m@3o0?D0o`3e0000
00030<`0o`040000002?0?l0S`3o00000080D@3o00@0000001@0o`0D0?l000000P00:?l00`000000
06Ko00000002002So`0800000000hOl00>7o00000000on400000003oX`000002003oIP0600000000
obP00?lX0000000Do`0000000P1Ao`001@000000Sol008oo002?o`0000000080c?l000@000000?oe
003om@0000000`3o^0000`000000ogX00?mj00020?mj00030000003o?@00ocd000@0ocd000<00000
0?l0003o0000303o00000`000000ocd00?lm00050?lm00030000003oNP00ogX00080ogX000<00000
0?nh003o^0000P3o^00010000000ooD00?oe003om@0200000080c?l00P00000208oo008000000P1A
o`000`0000005?l001Co0002000000800?lX00@00000003oIP00ofH000000`00oj<00`0000000?oQ
003oh@02003oh@8000000P00hOl2000000@00:?o0P000003001Vo`<000005@00:?l70?ooo`400000
0@3oool00200oooo00<000000?ooo`3oool01@3oool7000Xo`H000002`00:?l2000000@006Ko0`00
0002002So`8000000P00hOl2000000<00?oQ0`0000000`00oj<0000000000003000000800?lX0P00
000201Co008000000P1Ao`0010000000Sol008oo000000030<co00050000003om@00ooD00?oe0000
0000103o^0000`000000ogX00?mj00020?mj008000001P3o?@000`000000o`000?l0000<0?l00080
00001@3o?@000`000000ogX00?mj00020?mj008000000P3o^00010000000ooD00?oe000000030<co
00060000002?o`00Sol00000001Ao`0000000P0Do`001P0000000?lX003o:00000000?mV00000080
0?nS00<00000003oh@0000000P00hOl00`0000000:?o00000002001Vo`0600000000:?l002So0000
000D0?l000000P1A0?l01@000000S`3o08l0o`2?0?l000000080c03o00D000000?l0m@3o0?D0o`3e
000000050?l0^0030000003o07X0o`1j0140o`1j00<000000?l0^03o0;P00`3o0;P01@000000o`3e
0?l0m@3o0?D0000000<0c03o00@0000008l0o`2?0?l000000P1A0?l01P000000503o01@0o`000000
02So0000008006Ko00H00000002So`00Xol00000003Qo`000002003oh@0300000000oj<000000080
0?mV00<00000003o:00000000P0Do`0010000000DOl0057o0000000208oo00040000003<o`00c?l0
00000080ooD000D000000?nh003o^000okP0000000040?mj00030000003o?@00ocd000<0ocd00P00
000>0?l000030000003o?@00ocd000D0ocd000<000000?mj003oNP000P3oNP02000000<0okP000D0
00000?oe003om@00ooD0000000030<co00030000002?o`00Sol0008000000P1Ao`02000000805?l0
00<00000003o:000obP01@0000001000oj<00000000000000003003oh@8000000P00hOl2000000<0
0:?o0`000004001Vo`8000002P00:?l6000000L002So1`3oool100000040oooo000P0?ooo`030000
003oool0oooo00D0oooo1`00:?l00`000000503o01@0o`0301@0o`@000002@00:?l3000000@006Ko
0P000002002So`8000000P00hOl3000000<00?oQ00H00000003oX`00oj<00?nS00000000ofH20000
00800?lX0P00000201Co0080000000@0DOl0000000000000Sol00P0000020<co008000000P3om@02
000000<0okP00P0000040?mj00030000003o?@00ocd000@0ocd000<000000?l0003o00003@3o0000
0`000000ocd00?lm00030?lm008000000`3oNP001@000000okP00?nh003o^00000000080ooD000@0
00000<co003<o`0000000`2?o`000`000000DOl00000000201Co000300000000obP00?lX00800000
0`00oj<2000000800>7o00<00000002So`0000000P00I_l01000000002So000Xo`00000201@0o`03
0000001A0?l0000000<0S`3o00@000000<`0o`3<0?l000000`3o0?D00`000000o`2h0?l0^0030?l0
^0030000003o07X0o`1j0140o`1j00<000000?l0^03o0;P00P3o0;P00`000000o`3e0?l0m@020?l0
m@040000003<0?l0c03o00000080S`3o00@000000540o`1A0?l000000P0D0?l01P00000002So000X
o`00000006Ko000000800:?o00<00000003Qo`0000000P00on400`0000000?nS00000002003oIP03
00000000obP0000000805?l000@00000057o001Ao`0000000P2?o`0010000000c?l00<co00000002
0?oe00050000003o^000okP00?nh00000000103oNP000`000000ocd00?lm00030?lm00030000003o
0000o`0000d0o`0000<000000?lm003o?@001@3o?@0010000000ogX00?mj003oNP02000000<0okP0
0P0000020?oe008000000P3<o`020000000308oo0000000000000080DOl00P00000201Co00800000
00D00?lX0000000000000?mV00000003003oX`0400000000on400?oQ003oh@<000000P00hOl20000
00800:?o0`000004001Vo`<000002000:?l4000000D0503o00<00000000Xo`00:?l01@00:?l70?oo
o`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool7000Xo`030000000D0?l0503o00L0
503o10000008000Xo`<000000`00I_l2000000800:?o0P000003003Qo`<0000000<00?oQ00000000
00000P00oj<200000003003oIP000000000000800?lX0P00000201Co00040000001Ao`00DOl00000
0080Sol000D000000<co003<o`00c?l0000000030?oe00030000003o^000okP00080okP000<00000
0?mj003oNP000P3oNP02000000D0ocd000<000000?l0003o0000303o00000`000000ocd00?lm0005
0?lm00050000003oNP00ogX00?mj000000000P3o^0001@000000ooD00?oe003om@0000000080c?l0
00H0000008oo002?o`000000057o0000000301Co000500000000obP00000003oIP0000000P00oj<0
0`0000000?oQ00000002003Qo`0300000000Xol00000008006Ko00<00000000Xo`0000000P0D0?l0
10000000D@3o0540o`00000208l0o`050000003<0?l0c03o0<`0o`0000000`3o0?D00`000000o`2h
0?l0^0020?l0^0030000003o07X0o`1j0140o`1j00<000000?l0^03o0;P00P3o0;P01@000000o`3e
0?l0m@3o0?D0000000<0c03o00@0000008l0o`2?0?l000000P1A0?l01P000000503o01@0o`000000
02So0000008006Ko00P00000002So`00Xol00000003Qo`0000000?oQ000000800?nS00<00000003o
IP0000000P00obP00`0000005?l000000002057o00040000002?o`00Sol000000080c?l000@00000
0?oe003om@0000000`3o^0000`000000ogX00?mj00020?mj00030000003o?@00ocd000<0ocd000<0
00000?l0003o00003@3o00000`000000ocd00?lm00040?lm008000000`3oNP000`000000okP00?nh
00020?nh00050000003om@00ooD00?oe000000000`3<o`0010000000Sol008oo00000003057o0003
0000000Do`005?l00080000000@00?lX0000000000000?mV0P000002003oX`80000000@00?oQ0000
0000000000000`00hOl2000000800:?o0P000004001Vo`<000001`00:?l4000000T0503o00<00000
000Xo`00:?l01@00:?l70?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool7000X
o`030000000D0?l0503o00/0503o0`000008000Xo`0300000000I_l006Ko008006Ko0P000002002S
o`8000001000hOl0100000000?oQ003oh@000003003oX`0300000000ofH00?mV008000000P00obP2
0000000401Co00000000DOl0057o0080000000<0Sol00000000000000P3<o`0200000080ooD000<0
00000?nh003o^0000P3o^0000`000000ogX00?mj00030?mj00030000003o?@00ocd000<0ocd000<0
00000?l0003o0000303o00000`000000ocd00?lm00040?lm00050000003oNP00ogX00?mj00000000
0`3o^0001@000000ooD00?oe003om@0000000080c?l000H0000008oo002?o`000000057o00000002
01Co000300000000obP00?lX008000000`00oj<2000000800>7o00<00000002So`0000000P00I_l0
1000000002So000Xo`00000201@0o`030000001A0?l0000000<0S`3o00@000000<`0o`3<0?l00000
0`3o0?D00`000000o`2h0?l0^0030?l0^0030000003o07X0o`1j00l0o`1j00<000000?l0^03o0;P0
0P3o0;P00`000000o`3e0?l0m@020?l0m@040000003<0?l0c03o00000080S`3o00@000000540o`1A
0?l000000P0D0?l01P00000002So000Xo`00000006Ko000000800:?o00<00000003Qo`0000000P00
on401@0000000?nS00000000ofH0000000800?lX00H0000001Co000Do`000000057o0000000208oo
00050000003<o`00c?l00<co000000000P3om@001@000000okP00?nh003o^000000000<0ogX000<0
00000?lm003o?@00103o?@000`000000o`000?l0000<0?l000030000003o?@00ocd000@0ocd000<0
00000?mj003oNP000`3oNP001@000000okP00?nh003o^000000000<0ooD000<000000<co003<o`00
0P00000208oo00030000001Ao`00DOl00080000000<05?l00000000000000P00obP0100000000?mV
003oIP000003003oX`0400000000on400?oQ000000@00>7o0P000002002So`8000001000I_l20000
00L002So0`00000>01@0o`0300000000:?l002So00@002So1`3oool100000040oooo000P0?ooo`03
0000003oool0oooo00D0oooo1`00:?l00`000000503o01@0o`0>01@0o`D000001000:?l2000000@0
06Ko0P000002002So`8000000`00hOl00`0000000?oQ003oh@02000000800?nS0P000002003oIP80
000000D00?lX0000000Do`005?l000000003057o00030000002?o`00Sol0008000000P3<o`001000
0000ooD00?oe000000040?nh008000000`3oNP000`000000ocd00?lm00040?lm00030000003o0000
o`0000`0o`000P0000050?lm00050000003oNP00ogX00?mj000000000`3o^0001@000000ooD00?oe
003om@0000000080c?l000<0000008oo000000000P1Ao`00200000005?l001Co000000000?lX0000
0000ofH000000P00oj<0200000000?oQ0000000000000>7o00000000Xol000000P00I_l00`000000
02So0000000201@0o`040000001A0?l0D@3o00000080S`3o00D000000<`0o`3<0?l0c03o00000003
0?l0m@030000003o0;P0o`2h0080o`2h00<000000?l0NP3o07X03`3o07X00`000000o`2h0?l0^002
0?l0^0030000003o0?D0o`3e0080o`3e00@000000<`0o`3<0?l000000P2?0?l010000000D@3o0540
o`00000201@0o`0600000000:?l002So00000000I_l000000P00Xol01@0000000>7o00000000on40
000000800?nS00<00000003oIP0000000P00obP00`0000005?l000000002057o00040000002?o`00
Sol000000080c?l000@000000?oe003om@0000000`3o^0001@000000ogX00?mj003oNP00000000D0
ocd00P00000>0?l000030000003o?@00ocd000D0ocd000@000000?mj003oNP00ogX00P0000030?nh
00060000003om@00ooD00?oe00000000c?l00P00000208oo008000000P1Ao`00100000005?l001Co
00000002003o:08000000P00ofH00`0000000?nS003oX`02000000800?oQ00@00000003Qo`00hOl0
0>7o0P000002002So`8000001000I_l2000000D002So1000000@01@0o`8000001P00:?l70?ooo`40
00000@3oool00200oooo00<000000?ooo`3oool01@3oool8000Xo`030000000D0?l0503o0180503o
0`000003000Xo`8000001000I_l2000000800:?o0P000002003Qo`8000000P00on42000000800?nS
0P000002003oIP@0000000<05?l00000000000000P1Ao`001@000000Sol008oo002?o`0000000080
c?l000<000000?oe003om@000P0000040?nh00050000003oNP00ogX00?mj000000001P3o?@000`00
0000o`000?l0000=0?l000030000003o?@00ocd000<0ocd00P0000020?mj00050000003o^000okP0
0?nh000000000P3om@001`000000c?l00<co003<o`00000008oo000000000P1Ao`00200000005?l0
01Co000000000?lX00000000ofH000000P00oj<01@0000000?oQ00000000hOl0000000800:?o00<0
0000001Vo`0000000P00:?l01P000000503o01@0o`000000D@3o00000080S`3o00D000000<`0o`3<
0?l0c03o000000030?l0m@030000003o0;P0o`2h00<0o`2h00<000000?l0NP3o07X03@3o07X00`00
0000o`2h0?l0^0030?l0^0030000003o0?D0o`3e0080o`3e00<000000<`0o`0000000`2?0?l01P00
0000D@3o0540o`000000503o0000008002So00<00000001Vo`0000000P00Xol01@0000000>7o0000
0000on40000000800?nS00<00000003oIP0000000P00obP00`0000005?l000000002057o00040000
002?o`00Sol000000080c?l000@000000?oe003om@0000000`3o^0001@000000ogX00?mj003oNP00
000000D0ocd000<000000?l0003o00003@3o00000`000000ocd00?lm00050?lm00050000003oNP00
ogX00?mj000000000`3o^00200000080ooD00P0000000`3<o`00000008oo000208oo00030000001A
o`00DOl00080000000@05?l00000000000000?lX0P000002003oIP80000000<00?nS000000000000
0P00on42000000800>7o0P000002002So`8000001000I_l2000000@002So0`00000C01@0o`030000
0000:?l002So00H002So1`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo0P00
0007000Xo`8000005@0D0?l2000000<002So0`000003001Vo`8000000P00Xol2000000800>7o0P00
0002003oh@8000000P00oj<2000000800?mV10000000100Do`0000000000001Ao`0200000080Sol0
0P0000000`3<o`000000000000020?oe008000000`3o^0001@000000ogX00?mj003oNP00000000H0
ocd00P00000>0?l000030000003o?@00ocd000<0ocd000@000000?mj003oNP00ogX00P0000020?nh
00040000003om@00ooD000000080c?l000H0000008oo002?o`000000057o0000000201Co00030000
0000obP0000000800?mV00<00000003oX`0000000P00on401`0000000>7o00000000Xol00000001V
o`0000000P00:?l010000000503o01@0o`0000020540o`040000002?0?l0S`3o000000<0c03o00D0
00000?l0m@3o0?D0o`3e000000040?l0^0030000003o07X0o`1j00d0o`1j00<000000?l0^03o0;P0
0`3o0;P01@000000o`3e0?l0m@3o0?D000000080c03o00@0000008l0o`2?0?l000000P1A0?l01P00
0000503o01@0o`00000002So0000008006Ko00P00000002So`00Xol00000003Qo`0000000?oQ0000
00800?nS00L00000003oIP0000000?lX0000000Do`0000000080DOl000@0000008oo002?o`000000
0P3<o`0010000000ooD00?oe000000030?nh00050000003oNP00ogX00?mj000000001@3o?@000`00
0000o`000?l0000<0?l0008000001P3o?@0200000080ogX00P0000020?nh008000000P3om@020000
00030<co0000000000000080Sol00P000000101Ao`0000000000000Do`0200000003003o:0000000
000000800?mV0P0000000`00oj<0000000000002003oh@8000000P00hOl2000000800:?o0P000003
001Vo`<000001000:?l2000001@0503o0P000008000Xo`8000001P3oool100000040oooo000P0?oo
o`030000003oool0oooo00D0oooo0P00I_l2000000L002So00<0000001@0o`0D0?l0500D0?l20000
00@002So0`000002001Vo`<0000000D00:?o0000000000000000003Qo`0300000003003oh@000000
000000800?nS0P000002003oIP0500000000obP000000000000Do`000P0000000`1Ao`0000000000
000208oo00030000003<o`00c?l0008000000P3om@001@000000okP00?nh003o^000000000<0ogX0
0P0000060?lm00030000003o0000o`0000`0o`0000<000000?lm003o?@000`3o?@000`000000ogX0
0?mj00020?mj00040000003o^000okP000000080ooD000@000000<co003<o`0000000P2?o`000`00
0000DOl00000000201Co000300000000obP0000000800?mV00H00000003oX`0000000?oQ00000000
hOl200000004002So`00000006Ko0000008002So00H0000001@0o`0D0?l000000540o`00000208l0
o`050000003<0?l0c03o0<`0o`0000000`3o0?D00`000000o`2h0?l0^0030?l0^0030000003o07X0
o`1j00/0o`1j00<000000?l0^03o0;P00`3o0;P010000000o`3e0?l0m@3o0?D200000080c03o00@0
000008l0o`2?0?l000000P1A0?l00`000000503o00000002000Xo`0300000000I_l0000000800:?o
00D00000003Qo`0000000?oQ00000002003oX`0500000000ofH00000003o:00000000P0Do`000`00
0000DOl00000000208oo00040000003<o`00c?l000000080ooD000D000000?nh003o^000okP00000
00030?mj00030000003o?@00ocd000<0ocd000<000000?l0003o0000303o00000`000000ocd00?lm
00040?lm008000000`3oNP001@000000okP00?nh003o^000000000<0ooD000@000000<co003<o`00
00000P2?o`0200000004057o00000000000001Co0080000000<00?lX0000000000000P00ofH20000
0004003oX`0000000000003oh@<0000000D00>7o0000000000000000002So`030000008006Ko0`00
0004000Xo`<000005@0D0?l00`00000002So000Xo`05000Xo`<0000000<006Ko0?ooo`3oool01@3o
ool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo0`00I_l00`00000002So000Xo`06
000Xo`030000000D0?l0503o01D0503o0P000004000Xo`0300000000I_l006Ko00<006Ko0P000002
002So`8000000P00hOl200000003003oh@000000000000800?nS00<00000003oIP0000000`00obP0
100000005?l001Co00000002057o00040000002?o`00Sol000000080c?l000D000000?oe003om@00
ooD0000000030?nh00030000003oNP00ogX00080ogX000<000000?lm003o?@000`3o?@000`000000
o`000?l0000=0?l000030000003o?@00ocd000<0ocd000D000000?mj003oNP00ogX0000000030?nh
00040000003om@00ooD000000080c?l000<0000008oo000000000P1Ao`000`0000005?l000000002
003o:00700000000ofH00000003oX`0000000?oQ00000002003Qo`0500000000Xol00000001Vo`00
00000P00:?l01P000000503o01@0o`000000D@3o00000080S`3o00D000000<`0o`3<0?l0c03o0000
00030?l0m@030000003o0;P0o`2h00<0o`2h00<000000?l0NP3o07X02P3o07X00`000000o`2h0?l0
^0040?l0^0050000003o0?D0o`3e0?l0m@0000000`3<0?l01P000000S`3o08l0o`000000D@3o0000
0080503o00<00000000Xo`0000000P00I_l01@0000000:?o00000000hOl0000000800?oQ00<00000
003oX`0000000P00ofH2000000805?l000<00000057o000000000P2?o`0010000000c?l00<co0000
00020?oe00050000003o^000okP00?nh000000000`3oNP000`000000ocd00?lm00040?lm00030000
003o0000o`0000`0o`0000<000000?lm003o?@000`3o?@000`000000ogX00?mj00030?mj00050000
003o^000okP00?nh000000000P3om@0010000000c?l00<co0000000208oo00040000001Ao`00DOl0
000000805?l000D00000003o:000obP00?lX00000002003oIP0400000000oj<00000000000800?oQ
00<00000003Qo`00hOl00P000002002So`8000001@00I_l00`00000002So000Xo`02000Xo`800000
5`0D0?l00`00000002So000Xo`06000Xo`0400000000I_l006Ko001Vo`L0oooo0@0000010?ooo`00
803oool00`000000oooo0?ooo`050?ooo`0400000000I_l006Ko001Vo`8000001`00:?l2000001L0
503o00<00000000Xo`00:?l00P00:?l2000000D006Ko00D00000002So`00Xol00:?o00000003003Q
o`0300000000on400?oQ0080000000<00?nS00000000ofH00P000002003o:0040000000Do`005?l0
00000080DOl000<0000008oo002?o`000P0000000`3<o`0000000?oe00020?oe00050000003o^000
okP00?nh00000000103oNP000`000000ocd00?lm00030?lm00030000003o0000o`0000d0o`0000<0
00000?lm003o?@000P3o?@000`000000ogX00?mj00020?mj00040000003o^000okP000000080ooD0
00@000000<co003<o`0000000P2?o`000`000000DOl00000000201Co000300000000obP000000080
0?mV00L00000003oX`0000000?oQ00000000hOl0000000800:?o00D00000001Vo`00000002So0000
000201@0o`040000001A0?l0D@3o00000080S`3o00@000000<`0o`3<0?l000000`3o0?D00`000000
o`2h0?l0^0040?l0^0030000003o07X0o`1j00T0o`1j00<000000?l0^03o0;P00`3o0;P01@000000
o`3e0?l0m@3o0?D0000000<0c03o00@0000008l0o`2?0?l000000P1A0?l020000000503o01@0o`00
000002So00000000I_l000000P00Xol00`0000000>7o00000002003oh@0700000000oj<00000003o
IP0000000?lX0000000201Co00030000001Ao`0000000080Sol000@000000<co003<o`0000000P3o
m@001@000000okP00?nh003o^000000000<0ogX000<000000?lm003o?@000`3o?@000`000000o`00
0?l0000<0?l000030000003o?@00ocd000<0ocd000<000000?mj003oNP000`3oNP001@000000okP0
0?nh003o^00000000080ooD000@000000<co0000000000000P2?o`0010000000DOl0057o00000002
01Co000300000000obP00?lX0080000000D00?mV0000000000000?nS00000003003oh@0400000000
hOl00>7o000000<00:?o00<00000001Vo`00I_l00`00I_l2000000@002So00<0000001@0o`0D0?l0
5@0D0?l2000000L002So0P000004001Vo`030000003oool0oooo00@0oooo0@0000010?ooo`00803o
ool00`000000oooo0?ooo`050?ooo`03002So`000000000000<006Ko0`000006000Xo`030000000D
0?l0503o01D0503o0P000004000Xo`<000000`00I_l2000000800:?o0P000002003Qo`8000000P00
on400`0000000?nS00000002003oIP80000000@00?lX0000000000005?l00P0000000`1Ao`000000
08oo000208oo00040000003<o`00000000000080ooD00P0000020?nh00030000003oNP00ogX00080
ogX000<000000?lm003o?@000`3o?@02000000d0o`0000<000000?lm003o?@000`3o?@001@000000
ogX00?mj003oNP00000000<0okP000@000000?oe003om@0000000P3<o`000`000000Sol000000002
057o00030000000Do`00000000800?lX00L00000003oIP0000000?nS00000000on40000000800>7o
00D00000002So`00000006Ko00000002000Xo`060000000D0?l0503o0000001A0?l000000P2?0?l0
1@000000c03o0<`0o`3<0?l0000000<0o`3e00<000000?l0^03o0;P00`3o0;P00`000000o`1j0?l0
NP080?l0NP030000003o0;P0o`2h00@0o`2h00D000000?l0m@3o0?D0o`3e000000030<`0o`060000
002?0?l0S`3o0000001A0?l000000P0D0?l00`00000002So00000002001Vo`0500000000Xol00000
003Qo`0000000P00on400`0000000?nS00000002003oIP8000000P0Do`000`000000DOl000000002
08oo00040000003<o`00c?l000000080ooD000D000000?nh003o^000okP0000000030?mj00030000
003o?@00ocd000<0ocd000<000000?l0003o00002`3o0002000000D0ocd000<000000?mj003oNP00
0P3oNP0200000080okP00P0000020?oe00030000003<o`00000000<0Sol000D00000057o00000000
000001Co000200000003003o:0000000000000800?mV00@00000003oX`00000000000P00on420000
0003003Qo`000000000000800:?o0P000003001Vo`<000001000:?l2000001L0503o00<00000000X
o`00:?l01000:?l3000000@006Ko0P0000070?ooo`4000000@3oool00200oooo00<000000?ooo`3o
ool01@3oool00`0000000:?o002So`02000000@006Ko0P000005000Xo`8000005`0D0?l3000000@0
02So0P000003001Vo`8000000P00Xol2000000800>7o00D00000003oh@00on400000003oX`020000
00800?mV00@00000003o:000obP000000P0Do`0010000000DOl000000000000208oo00040000003<
o`00c?l0000000<0ooD000<000000?nh003o^0000P0000030?mj00030000003o?@00ocd000@0ocd0
00<000000?l0003o00002`3o00000`000000ocd00?lm00030?lm00050000003oNP00ogX00?mj0000
00000`3o^00010000000ooD00?oe000000020<co00030000002?o`0000000080DOl000D0000001Co
000000000?lX00000002003oIP0700000000oj<00000003oh@0000000>7o00000002002So`050000
0000I_l00000000Xo`0000000P0D0?l010000000D@3o0540o`00000208l0o`040000003<0?l0c03o
000000<0o`3e00<000000?l0^03o0;P0103o0;P00`000000o`1j0?l0NP070?l0NP030000003o0;P0
o`2h00@0o`2h00D000000?l0m@3o0?D0o`3e000000030<`0o`060000002?0?l0S`3o0000001A0?l0
00000P0D0?l00`00000002So00000002001Vo`0700000000Xol00000003Qo`0000000?oQ00000002
003oX`0500000000ofH00000003o:00000000P0Do`000`000000DOl00000000208oo00040000003<
o`00c?l000000080ooD000@000000?nh003o^00000000`3oNP000`000000ocd00?lm00030?lm0003
0000003o0000o`0000/0o`0000<000000?lm003o?@00103o?@000`000000ogX00?mj00020?mj0003
0000003o^000okP0008000000P3om@02000000040<co00000000Sol008oo0080000000D0DOl00000
000Do`005?l000000002003o:00300000000ofH00?mV0080000000@00?nS00000000on400?oQ0P00
0002003Qo`0300000000Xol00:?o008000000`00I_l2000000@002So0`00000G01@0o`8000001@00
:?l2000000@006Ko0`000002002So`030000003oool0oooo00@0oooo0@0000010?ooo`00803oool0
0`000000oooo0?ooo`050?ooo`03003Qo`000000000000800:?o0P000004001Vo`0300000000:?l0
02So00@002So00<0000001@0o`0D0?l05`0D0?l2000000@002So00<00000001Vo`00I_l00P00I_l0
1@0000000:?o002So`00Xol0000000800>7o00@00000003oh@00on4000000P00oj<00`0000000?mV
003oIP0200000005003o:00000005?l001Co000000000P1Ao`000`000000Sol008oo000200000003
0<co00000000ooD00080ooD000D000000?nh003o^000okP0000000030?mj00030000003o?@00ocd0
00<0ocd000<000000?l0003o0000303o00000`000000ocd00?lm00030?lm00050000003oNP00ogX0
0?mj000000000P3o^00010000000ooD00?oe000000020<co00030000002?o`0000000080DOl000D0
000001Co000000000?lX00000002003oIP0700000000oj<00000003oh@0000000>7o00000002002S
o`0500000000I_l00000000Xo`0000000P0D0?l010000000D@3o0540o`00000208l0o`040000003<
0?l0c03o000000<0o`3e00<000000?l0^03o0;P01@3o0;P00`000000o`1j0?l0NP050?l0NP030000
003o0;P0o`2h00D0o`2h00D000000?l0m@3o0?D0o`3e000000020<`0o`040000002?0?l0S`3o0000
0080D@3o00<0000001@0o`0000000P00:?l01@00000006Ko00000000Xol0000000800>7o00L00000
003oh@0000000?nS00000000ofH0000000800?lX00D0000001Co00000000DOl00000000208oo0004
0000003<o`00c?l000000080ooD000@000000?nh003o^00000000`3oNP000`000000ocd00?lm0004
0?lm00030000003o0000o`0000/0o`0000<000000?lm003o?@000`3o?@000`000000ogX00?mj0002
0?mj00040000003o^000okP0000000<0ooD000@000000<co003<o`0000000P2?o`0010000000DOl0
057o0000000201Co000400000000obP00000000000800?mV00@00000003oX`00oj<000000`00on40
100000000>7o003Qo`000002002So`0300000000I_l006Ko008006Ko00<00000000Xo`00:?l00P00
:?l2000001T0503o00<00000000Xo`00:?l01000:?l00`00000006Ko001Vo`02001Vo`8000000`00
Xol2000000L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`<00>7o00D00000
002So`00Xol00:?o00000004001Vo`8000001@00:?l00`000000503o01@0o`0<01@0o`8000002P0D
0?l00`00000002So000Xo`02000Xo`8000000`00I_l2000000800:?o0P0000001@00hOl000000000
0000on40000000800?nS00H00000003oIP00ofH00?mV00000000obP20000000501Co00000000DOl0
057o000000000`2?o`001P000000c?l0000000000000ooD0000000<0okP000<000000?mj003oNP00
0P3oNP000`000000ocd00?lm00030?lm00030000003o0000o`0000/0o`0000<000000?lm003o?@00
0`3o?@001@000000ogX00?mj003oNP00000000<0okP000@000000?oe003om@0000000P3<o`000`00
0000Sol000000002057o00050000000Do`000000003o:00000000P00ofH2000000800?oQ00D00000
003Qo`0000000:?o00000002001Vo`0500000000:?l0000001@0o`0000000P1A0?l010000000S`3o
08l0o`0000020<`0o`030000003o0?D0o`3e0080o`3e00<000000?l0^03o0;P0103o0;P00`000000
o`1j0?l0NP040?l0NP030000003o0;P0o`2h00D0o`2h00D000000?l0m@3o0?D0o`3e000000030<`0
o`060000002?0?l0S`3o0000001A0?l000000P0D0?l00`00000002So00000002001Vo`0300000000
Xol0000000800>7o0P000002003oX`0500000000ofH00000003o:00000000P0Do`000`000000DOl0
0000000208oo00030000003<o`0000000080ooD000D000000?nh003o^000okP0000000030?mj0003
0000003o?@00ocd000<0ocd000<000000?l0003o00002`3o00000`000000ocd00?lm00030?lm0003
0000003oNP00ogX00080ogX000@000000?nh003o^00000000`3om@0010000000c?l00<co00000002
08oo00050000001Ao`00DOl00000000Do`000P0000000`00obP00000003oIP02003oIP0400000000
oj<00?nS000000800?oQ0P0000000`00hOl0000000000002002So`0400000000I_l006Ko001Vo`80
00001000:?l00`000000503o01@0o`0801@0o`8000003P0D0?l00`00000002So000Xo`03000Xo`80
00001000I_l0100000000:?o002So`00Xol2000000800>7o1`3oool100000040oooo000P0?ooo`03
0000003oool0oooo00D0oooo00@00000003Qo`00hOl00>7o0P000002002So`8000001000I_l20000
00@002So0P00000;01@0o`030000001A0?l0D@3o00@000001`0D0?l2000000@002So0P000003001V
o`0500000000Xol00:?o002So`0000000P00hOl01@0000000?oQ0000000000000?nS008000000P00
ofH01@0000000?lX003o:00000005?l00080000000<0DOl00000000000000P2?o`001P000000c?l0
0<co00000000ooD0000000<0okP000<000000?mj003oNP000P3oNP000`000000ocd00?lm00030?lm
00800000303o00000`000000ocd00?lm00030?lm00050000003oNP00ogX00?mj000000000P3o^000
10000000ooD00?oe000000020<co00030000002?o`0000000080DOl000D0000001Co000000000?lX
00000002003oIP0700000000oj<00000003oh@0000000>7o00000002002So`0500000000I_l00000
000Xo`0000000P0D0?l010000000D@3o0540o`00000208l0o`040000003<0?l0c03o000000<0o`3e
00<000000?l0^03o0;P01@3o0;P00`000000o`1j0?l0NP030?l0NP030000003o0;P0o`2h00D0o`2h
00D000000?l0m@3o0?D0o`3e000000020<`0o`040000002?0?l0S`3o00000080D@3o00<0000001@0
o`0000000P00:?l01@00000006Ko00000000Xol0000000800>7o00L00000003oh@0000000?nS0000
0000ofH0000000800?lX00D0000001Co00000000DOl00000000208oo00040000003<o`00c?l00000
0080ooD000@000000?nh003o^00000000`3oNP000`000000ocd00?lm00040?lm00030000003o0000
o`0000X0o`0000<000000?lm003o?@000`3o?@000`000000ogX00?mj00020?mj00050000003o^000
okP00?nh000000000P3om@001@000000c?l0000000000000Sol00080000000D0DOl0000000000000
5?l000000002003o:00300000000ofH00?mV0080000000@00?nS0000000000000?oQ0P000002003Q
o`0300000000Xol00:?o008000000P00I_l2000000@002So0P00000701@0o`@000000P1A0?l00`00
0000503o01@0o`0901@0o`8000001000:?l2000000@006Ko0P000002002So`8000000`00hOl20000
00H0oooo0@0000010?ooo`003`3oool00`2oM0000000000000040000000304RLg`3oool0oooo00L0
oooo00<000000?ooo`3oool01@3oool00`00on40000000000003003Qo`8000000P00Xol2000000@0
06Ko00<00000000Xo`00:?l00`00:?l2000000T0503o00<000000540o`1A0?l0101A0?l2000000L0
503o0`000003000Xo`8000000P00I_l2000000800:?o00<00000003Qo`00hOl00P0000001000on40
000000000000oj<200000005003oIP0000000?lX003o:00000000P0Do`0010000000DOl0057o0000
000208oo00050000003<o`00c?l00000003om@000P0000020?nh008000000`3oNP000`000000ocd0
0?lm00040?lm00030000003o0000o`0000T0o`0000<000000?lm003o?@00103o?@001@000000ogX0
0?mj003oNP0000000080okP000@000000?oe003om@0000000P3<o`000`000000Sol000000002057o
00090000000Do`000000003o:00000000?mV00000000oj<0000000800?oQ00L00000003Qo`000000
0:?o00000000I_l00000008002So00<0000001@0o`0000000P1A0?l010000000S`3o08l0o`000002
0<`0o`030000003o0?D0o`3e0080o`3e00<000000?l0^03o0;P01@3o0;P5000000P0o`2h00D00000
0?l0m@3o0?D0o`3e000000020<`0o`040000002?0?l0S`3o00000080D@3o00<0000001@0o`000000
0P00:?l01@00000006Ko00000000Xol0000000800>7o00P00000003oh@0000000?nS00000000ofH0
0000003o:080000000@05?l00000001Ao`0000000P2?o`000`000000c?l0000000020?oe00050000
003o^000okP00?nh000000000`3oNP000`000000ocd00?lm00030?lm00030000003o0000o`0000T0
o`000P0000050?lm00040000003oNP00ogX00?mj008000000P3o^002000000050?oe000000000000
0<co000000000P2?o`0010000000DOl0057o0000000201Co000400000000obP00?lX000000800?mV
00@00000003oX`00oj<000000P00on400`0000000>7o003Qo`0200000003002So`000000000000<0
06Ko00<00000000Xo`00:?l00P00:?l2000000L0503o0P0000070540o`030000000D0?l0503o00H0
503o0P000005000Xo`0300000000I_l006Ko008006Ko0P000003002So`0400000000hOl00>7o003Q
o`80000000<00?oQ0?ooo`3oool01@3oool100000040oooo000@0?ooo`030=nLB0000000000000<0
000000<004RL0=ooo`3oool01`3oool00`000000oooo0?ooo`050?ooo`0400000000on400?oQ0000
00<00>7o00<00000002So`00Xol00P00Xol2000000<006Ko0P000005000Xo`030000000D0?l0503o
00H0503o00<000000540o`1A0?l01P1A0?l2000000P0503o0P000003000Xo`8000000P00I_l20000
0003002So`0000000>7o00800>7o00H00000003oh@00on400000003oX`000002003oIP80000000D0
0?lX0000000000005?l000000002057o00040000002?o`00Sol000000080c?l000@000000?oe003o
m@0000000`3o^0001@000000ogX00?mj003oNP00000000H0ocd000<000000?l0003o00002@3o0000
0`000000ocd00?lm00030?lm00050000003oNP00ogX00?mj000000000P3o^00010000000ooD00?oe
000000020<co00080000002?o`00Sol00000001Ao`00000001Co00000002003o:080000000/00?nS
00000000on400?oQ00000000hOl00000002So`00000006Ko00000002000Xo`030000000D0?l00000
0080D@3o00@0000008l0o`2?0?l000000P3<0?l00`000000o`3e0?l0m@020?l0m@030000003o0;P0
o`2h0140o`2h00<000000?l0m@3o0?D00P3o0?D010000000c03o0<`0o`00000208l0o`060000001A
0?l0D@3o0000000D0?l000000P00:?l02`00000006Ko00000000Xol00000003Qo`0000000?oQ0000
0000oj<0000000800?mV00L00000003o:00000005?l00000001Ao`0000000080Sol000<000000<co
000000000P3om@001@000000okP00?nh003o^000000000<0ogX000<000000?lm003o?@000`3o?@00
0`000000o`000?l000090?l000030000003o?@00ocd000@0ocd000D000000?mj003oNP00ogX00000
00030?nh00040000003om@00ooD000000080c?l000P0000008oo002?o`00Sol00000001Ao`000000
01Co0080000000<00?lX0000000000000P00ofH01@0000000?nS003oX`0000000?oQ008000000P00
hOl0100000000:?o002So`000003001Vo`8000000`00:?l2000000L0503o0P0000080540o`030000
000D0?l0503o00H0503o00<00000000Xo`00:?l00`00:?l2000000<006Ko0P000004002So`050000
0000hOl00>7o003Qo`0000000`00on400`000000oooo0?ooo`040?ooo`4000000@3oool00140oooo
00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`070?ooo`030000003oool0oooo00D0oooo
00@00?nS0000000000000?oQ0P000002003Qo`8000001000Xol2000000<006Ko00<00000000Xo`00
:?l00`00:?l2000000L0503o00<000000540o`1A0?l01`1A0?l3000000L0503o00<00000000Xo`00
:?l00P00:?l01`00000006Ko001Vo`00I_l00000002So`0000000`00hOl01P0000000?oQ003oh@00
00000?nS000000<00?mV00D00000003o:000obP0000001Co000200000007057o00000000Sol008oo
00000000c?l0000000030?oe00040000003o^000okP0000000@0ogX000<000000?lm003o?@000`3o
?@000`000000o`000?l000090?l000030000003o?@00ocd000@0ocd000D000000?mj003oNP00ogX0
000000020?nh00040000003om@00ooD000000080c?l000D0000008oo00000000DOl00000000201Co
000=00000000obP00000003oIP0000000?nS00000000on400000003Qo`0000000:?o00000002001V
o`0300000000:?l000000080503o00<000000540o`0000000`2?0?l010000000c03o0<`0o`000003
0?l0m@030000003o0;P0o`2h0140o`2h00D000000?l0m@3o0?D0o`3e000000030<`0o`030000002?
0?l000000080D@3o00<0000001@0o`0000000P00:?l00`00000006Ko00000002002So`0;00000000
hOl00000003oh@0000000?nS00000000ofH00000003o:00000000P0Do`000`000000DOl000000002
08oo00030000003<o`0000000080ooD000@000000?nh003o^00000000`3oNP000`000000ocd00?lm
00040?lm00030000003o0000o`0000P0o`0000<000000?lm003o?@00103o?@001@000000ogX00?mj
003oNP00000000<0okP000D000000?oe003om@00ooD0000000020<co00080000002?o`00Sol00000
001Ao`00000001Co00000002003o:00600000000ofH00?mV003oIP0000000?nS0P0000000`00on40
0000003Qo`02003Qo`0400000000Xol00:?o000000<006Ko00@00000000Xo`00:?l002So0P000006
01@0o`<000002@1A0?l00`000000503o01@0o`0601@0o`0300000000:?l002So00<002So00@00000
001Vo`00I_l006Ko0P000004002So`8000000P00hOl2000000800?oQ0P0000070?ooo`4000000@3o
ool00180oooo00<0gia807Boo`3oool02`3oool00`000000oooo0?ooo`050?ooo`0400000000oj<0
0?nS000000800?oQ0P000002003Qo`<000000`00Xol20000008006Ko0P000005000Xo`030000000D
0?l0503o00D0503o00<000000540o`1A0?l02@1A0?l2000000H0503o0P000003000Xo`8000000P00
I_l0100000000:?o000000000002003Qo`80000000<00?oQ00000000oj<00P000002003oIP040000
0000obP00?lX000000805?l000<00000057o000000000P2?o`000`000000c?l0000000030?oe0004
0000003o^000okP0000000@0ogX000<000000?lm003o?@000`3o?@000`000000o`000?l000090?l0
00030000003o?@00ocd000<0ocd000D000000?mj003oNP00ogX0000000020?nh00040000003om@00
ooD000000080c?l000P0000008oo002?o`000000057o000000005?l0000000800?lX0P0000002`00
oj<00000003oh@00on400000003Qo`0000000:?o00000000I_l00000008002So00<0000001@0o`00
00000P1A0?l010000000S`3o08l0o`0000020<`0o`050000003o0?D0o`3e0?l0m@0000004P3o0;P0
0`000000o`3e0?l0m@020?l0m@070000003<0?l0c03o0<`0o`000000S`3o000000020540o`030000
000D0?l00000008002So00/00000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS0000
0002003oIP0700000000obP0000001Co00000000DOl00000000208oo00030000003<o`0000000080
ooD000D000000?nh003o^000okP0000000030?mj00030000003o?@00ocd000<0ocd000<000000?l0
003o00002@3o00000`000000ocd00?lm00030?lm00030000003oNP00ogX00080ogX000@000000?nh
003o^00000000`3om@001@000000c?l0000000000000Sol00080000000@0DOl00000000Do`000000
0P00obP00`0000000?mV003oIP0200000003003oX`0000000?oQ008000000P00hOl200000003002S
o`0000000000008006Ko0P000002000Xo`8000001P0D0?l2000000/0D@3o00<0000001@0o`0D0?l0
1P0D0?l00`00000002So000Xo`02000Xo`8000000P00I_l2000000@00:?o0P000002003Qo`800000
0P00on42000000800?nS00<000000?ooo`3oool0103oool100000040oooo000C0?ooo`030;mdB02L
gol0oooo00X0oooo00<000000?ooo`3oool01@3oool01000ofH00000003oX`00oj<2000000800?oQ
0P000003003Qo`8000000`00Xol01000000006Ko001Vo`00I_l2000000@002So00<0000001@0o`0D
0?l01@0D0?l00`000000D@3o0540o`0:0540o`8000001P0D0?l2000000<002So00@00000001Vo`00
I_l000000P00Xol2000000800>7o00H00000003oh@0000000000003oX`000002003oIP0600000000
obP00?lX0000000Do`0000000P1Ao`020000000308oo00000000c?l0008000000P3om@000`000000
okP00?nh0002000000<0ogX000<000000?lm003o?@000`3o?@02000000X0o`0000<000000?lm003o
?@000`3o?@001@000000ogX00?mj003oNP0000000080okP000@000000?oe003om@0000000P3<o`00
1@000000Sol00000001Ao`00000000805?l000d00000003o:00000000?mV00000000oj<00000003o
h@0000000>7o00000000Xol00000008006Ko00<00000000Xo`0000000P0D0?l00`000000D@3o0000
000208l0o`050000003<0?l0c03o0<`0o`0000000`3o0?D00`000000o`2h0?l0^00?0?l0^0030000
003o0?D0o`3e0080o`3e00@000000<`0o`3<0?l000000P2?0?l01P000000D@3o0540o`000000503o
0000008002So00d00000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH0
000000800?lX00D0000001Co00000000DOl00000000208oo00030000003<o`0000000080ooD000@0
00000?nh003o^00000000`3oNP000`000000ocd00?lm00040?lm00030000003o0000o`0000L0o`00
0P0000050?lm00030000003oNP00ogX00080ogX000@000000?nh003o^00000000P3om@0200000005
0<co00000000Sol008oo000000000P1Ao`000`0000005?l000000002003o:00400000000ofH00?mV
000000800?nS00<00000003oh@0000000P00hOl2000000800:?o00D00000001Vo`00I_l006Ko0000
0003000Xo`030000000D0?l0503o00@0503o0P00000=0540o`030000000D0?l0503o00@0503o0P00
0003000Xo`8000000`00I_l00`0000000:?o002So`02002So`8000000P00hOl2000000800?oQ0P00
0002003oX`8000001`3oool100000040oooo000C0?ooo`030?oog`2LB7@0_ooo00X0oooo0`000005
0?ooo`800?mV0P000002003oX`8000000P00on42000000<00>7o00D00000002So`00Xol00:?o0000
0004001Vo`0300000000:?l002So008002So0P00000601@0o`030000001A0?l0D@3o00/0D@3o0P00
000601@0o`0500000000:?l002So000Xo`0000000P00I_l01@0000000:?o002So`00Xol000000080
0>7o00H00000003oh@00on400000003oX`000002003oIP0600000000obP00?lX0000000Do`000000
0`1Ao`000`000000Sol0000000020<co00040000003om@00ooD0000000<0okP000D000000?mj003o
NP00ogX0000000060?lm00030000003o0000o`0000L0o`0000<000000?lm003o?@000`3o?@001@00
0000ogX00?mj003oNP0000000080okP000@000000?oe003om@0000000P3<o`000`000000Sol00000
0002057o000?0000000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o0000
0000Xol00000008006Ko00<00000000Xo`0000000P0D0?l00`000000D@3o0000000208l0o`050000
003<0?l0c03o0<`0o`0000000`3o0?D00`000000o`2h0?l0^00?0?l0^0030000003o0?D0o`3e0080
o`3e00@000000<`0o`3<0?l000000P2?0?l02P000000D@3o0540o`000000503o00000000:?l00000
001Vo`000002002So`0;00000000hOl00000003oh@0000000?nS00000000ofH00000003o:0000000
0P0Do`0200000080Sol000<000000<co000000000P3om@0010000000okP00?nh000000030?mj0003
0000003o?@00ocd000@0ocd000<000000?l0003o00001`3o00000`000000ocd00?lm00040?lm0003
0000003oNP00ogX00080ogX000@000000?nh003o^00000000P3om@0010000000c?l00<co00000002
08oo00060000001Ao`00DOl00000000Do`0000000P00obP01@0000000?mV003oIP0000000?nS0080
000000D00?oQ00000000hOl00>7o00000003002So`0500000000I_l006Ko001Vo`0000000`00:?l0
0`000000503o01@0o`0301@0o`8000003@1A0?l2000000H0503o00<00000000Xo`00:?l00P00:?l0
0`00000006Ko001Vo`02001Vo`0400000000Xol00:?o002So`8000000P00hOl2000000800?oQ0P00
0002003oX`8000000P00ofH50?ooo`<000000@3oool001@0oooo00<0okmd001d_`3oool02@3oool0
0`000000oooo0?ooo`050?ooo`0400000000ofH00?mV003oIP8000000P00oj<2000000800?oQ0P00
0002003Qo`8000000P00Xol2000000<006Ko0P000004000Xo`030000000D0?l0503o00@0503o00<0
00000540o`1A0?l0301A0?l2000000D0503o0P000002000Xo`0300000000I_l006Ko008000000P00
Xol200000006003Qo`0000000?oQ003oh@0000000?nS0P0000001P00ofH00000003o:000obP00000
01Co0080000000D0DOl00000002?o`00Sol0000000020<co00040000003om@00ooD000000080okP0
00<000000?mj003oNP000P3oNP000`000000ocd00?lm00030?lm00030000003o0000o`0000L0o`00
00<000000?lm003o?@00103o?@001@000000ogX00?mj003oNP0000000080okP000@000000?oe003o
m@0000000P3<o`003@000000Sol00000001Ao`00000001Co000000000?lX00000000ofH00000003o
X`0000000P00on402@0000000>7o00000000Xol00000001Vo`00000002So0000000201@0o`030000
001A0?l000000080S`3o00D000000<`0o`3<0?l0c03o000000030?l0m@030000003o0;P0o`2h00h0
o`2h00<000000?l0m@3o0?D00P3o0?D010000000c03o0<`0o`00000208l0o`060000001A0?l0D@3o
0000000D0?l000000P00:?l03@00000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<0
0000003oIP0000000P00obP01`0000005?l00000001Ao`00000008oo000000000P3<o`0010000000
ooD00?oe000000020?nh00050000003oNP00ogX00?mj000000001P3o?@000`000000o`000?l00006
0?l000030000003o?@00ocd000@0ocd000<000000?mj003oNP000P3oNP0010000000okP00?nh0000
00020?oe00040000003<o`00c?l000000080Sol000@00000057o001Ao`0000000P0Do`0010000000
0?lX00000000ofH200000006003oX`0000000?oQ003oh@0000000>7o0P000002002So`8000000P00
I_l20000008002So0P00000401@0o`8000003P1A0?l00`000000503o01@0o`0501@0o`0300000000
:?l002So008002So00@00000001Vo`00I_l006Ko0P000002002So`8000000P00hOl2000000800?oQ
0P000002003oX`8000001000ofH00`000000oooo0?ooo`040?ooo`4000000@3oool001@0oooo00<0
oono07@0B02Lgol02@3oool00`000000oooo0?ooo`050?ooo`03003o:0000000000000<00?mV0P00
0002003oX`8000000P00on42000000800>7o0P000002002So`8000000`00I_l00`00000002So000X
o`02000Xo`8000001@0D0?l00`000000D@3o0540o`0=0540o`8000001@0D0?l01000000002So000X
o`000003001Vo`0600000000Xol00:?o002So`0000000>7o0P0000001`00on400000003oX`00oj<0
0000003oIP0000000P00obP01P0000005?l001Co00000000DOl000000080Sol000@000000<co003<
o`0000000P3om@0010000000okP00?nh000000040?mj00030000003o?@00ocd000<0ocd000<00000
0?l0003o00001P3o00000`000000ocd00?lm00040?lm00050000003oNP00ogX00?mj000000000P3o
^00010000000ooD00?oe000000020<co00030000002?o`0000000080DOl000l0000001Co00000000
0?lX00000000ofH00000003oX`0000000?oQ00000000hOl00000002So`0000000P00I_l00`000000
02So0000000201@0o`030000001A0?l000000080S`3o00D000000<`0o`3<0?l0c03o000000030?l0
m@030000003o0;P0o`2h00d0o`2h00<000000?l0m@3o0?D00P3o0?D010000000c03o0<`0o`000002
08l0o`0:0000001A0?l0D@3o0000000D0?l00000000Xo`00000006Ko000000800:?o00/00000003Q
o`0000000?oQ00000000oj<00000003oIP0000000?lX0000000201Co008000000P2?o`000`000000
c?l0000000020?oe00040000003o^000okP0000000<0ogX000<000000?lm003o?@00103o?@000`00
0000o`000?l000070?l000030000003o?@00ocd000<0ocd000<000000?mj003oNP000P3oNP001000
0000okP00?nh000000020?oe00040000003<o`00c?l000000080Sol000D00000057o001Ao`000000
01Co000200000004003o:00000000?mV000000800?nS00H00000003oh@00on400000003Qo`000003
002So`0500000000I_l006Ko001Vo`0000000`00:?l00`000000503o01@0o`0301@0o`030000001A
0?l0D@3o00d0D@3o00<0000001@0o`0D0?l0100D0?l2000000@002So00<00000001Vo`00I_l00P00
0002002So`8000000P00hOl2000000800?oQ0P000002003oX`8000001000ofH2000000L0oooo0@00
00010?ooo`00503oool00`3ookl0M00004RLg`090?ooo`030000003oool0oooo00D0oooo00<00000
003o:000obP00P000003003oIP8000000P00oj<2000000800?oQ0P000002003Qo`0500000000Xol0
0:?o002So`0000000`00I_l2000000@002So00<0000001@0o`0D0?l00`0D0?l00`000000D@3o0540
o`0>0540o`800000100D0?l00`00000002So000Xo`020000008006Ko00D00000002So`00Xol00:?o
00000002003Qo`0300000000on40000000800?nS00H00000003oIP0000000000003o:000000201Co
00030000001Ao`0000000080Sol000@000000<co003<o`0000000P3om@0010000000okP00?nh0000
00030?mj00030000003o?@00ocd000@0ocd00P0000070?l000030000003o?@00ocd000@0ocd000D0
00000?mj003oNP00ogX0000000020?nh00060000003om@00ooD00000003<o`0000000P2?o`002`00
0000DOl00000000Do`000000003o:00000000?mV00000000oj<0000000800?oQ00T00000003Qo`00
00000:?o00000000I_l00000000Xo`0000000P0D0?l00`000000D@3o0000000208l0o`050000003<
0?l0c03o0<`0o`0000000`3o0?D00`000000o`2h0?l0^00=0?l0^0030000003o0?D0o`3e0080o`3e
00@000000<`0o`3<0?l000000P2?0?l02P000000D@3o000000000000503o00000000:?l00000001V
o`000002002So`0?00000000hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l0
0000001Ao`0000000080Sol000<000000<co000000000P3om@0010000000okP00?nh000000030?mj
00030000003o?@00ocd000@0ocd000<000000?l0003o00001P3o00000`000000ocd00?lm00040?lm
00050000003oNP00ogX00?mj000000000P3o^0001`000000ooD00?oe003om@0000000<co00000000
0P2?o`001P000000DOl0057o000000005?l0000000800?lX00<00000003oIP0000000P00oj<01000
00000?oQ000000000002003Qo`0400000000Xol00:?o0000008006Ko0P000003000Xo`030000000D
0?l0503o0080503o0P00000?0540o`030000000D0?l0503o00@0503o00<00000000Xo`00:?l00P00
:?l20000008006Ko00D00000002So`00Xol00:?o00000002003Qo`8000000P00on42000000800?nS
0P000003003oIP<000000P00obP00`000000oooo0?ooo`040?ooo`4000000@3oool000l0oooo00<0
oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`090?ooo`030000003oool0oooo00D0oooo00<0
5?l00000000000000P00obP2000000<00?mV0P000002003oX`8000000P00on4200000003003Qo`00
0000000000800:?o0P000003001Vo`8000000P00:?l00`000000503o01@0o`0401@0o`800000401A
0?l00`000000503o01@0o`0201@0o`0500000000:?l002So000Xo`0000000P00I_l2000000800:?o
00D00000003Qo`00hOl00000003oh@0200000006003oX`0000000000003oIP0000000?lX0P000000
100Do`000000057o0000000208oo00040000003<o`00c?l000000080ooD000@000000?nh003o^000
00000`3oNP000`000000ocd00?lm00050?lm00030000003o0000o`0000@0o`0000<000000?lm003o
?@00103o?@001@000000ogX00?mj003oNP0000000080okP000H000000?oe003om@0000000<co0000
000208oo00030000001Ao`00000000805?l00P0000001000ofH00000003oX`000002003oh@090000
0000hOl00000002So`00000006Ko00000000:?l000000080503o00<000000540o`0000000P2?0?l0
1@000000c03o0<`0o`3<0?l0000000<0o`3e00<000000?l0^03o0;P03@3o0;P00`000000o`3e0?l0
m@020?l0m@040000003<0?l0c03o00000080S`3o00<000000540o`0000000P0D0?l03`00000002So
00000000I_l00000002So`0000000>7o00000000on400000003oX`0000000?mV00000002003o:007
0000000Do`000000057o00000000Sol0000000020<co00030000003om@00000000<0okP000@00000
0?mj003oNP0000001P3o?@000`000000o`000?l000060?l000030000003o?@00ocd000@0ocd000D0
00000?mj003oNP00ogX0000000020?nh00070000003om@00ooD00?oe00000000c?l00000000208oo
00060000001Ao`00DOl00000000Do`0000000P00obP0100000000?mV00000000oj<200000003003o
h@0000000>7o00800>7o00@00000002So`00000000000P00I_l01000000002So000Xo`00:?l20000
00@0503o00<000000540o`1A0?l03@1A0?l2000000H0503o00@00000000Xo`00:?l002So0P000003
001Vo`0300000000Xol00:?o0080000000<00>7o0000000000000P00on42000000800?nS0P000003
003oIP8000000`00obP2000000L0oooo0@0000010?ooo`003`3oool0203ooml0W4Q809cOo`3oool0
oono07@0000007@0_ooo2@3oool00`000000oooo0?ooo`050?ooo`<05?l000D00000003o:000obP0
0?lX00000004003oIP0500000000oj<00?nS003oX`0000000`00on40100000000>7o003Qo`000003
002So`0300000000I_l006Ko008006Ko00<00000000Xo`00:?l00P00000601@0o`030000001A0?l0
D@3o00h0D@3o0P00000301@0o`0500000000:?l002So000Xo`0000000`00I_l0100000000:?o002S
o`000002003Qo`0400000000on400?oQ000000800?nS00<00000003oIP0000000P00obP200000080
DOl000H0000008oo002?o`0000000<co000000030?oe00040000003o^000okP0000000<0ogX000<0
00000?lm003o?@00103o?@000`000000o`000?l000050?l000030000003o?@00ocd000@0ocd000D0
00000?mj003oNP00ogX0000000020?nh00060000003om@00ooD00000003<o`0000000P2?o`004@00
0000DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o00000000
Xol00000008006Ko0P00000201@0o`030000001A0?l000000080S`3o00D000000<`0o`3<0?l0c03o
000000040?l0m@030000003o0;P0o`2h00/0o`2h00<000000?l0m@3o0?D00P3o0?D010000000c03o
0<`0o`00000208l0o`0:0000001A0?l0D@3o0000000D0?l00000000Xo`00000006Ko000000800:?o
00T00000003Qo`0000000?oQ00000000oj<00000003oIP0000000P00obP200000080DOl000D00000
08oo00000000c?l0000000020?oe00040000003o^000okP0000000<0ogX000<000000?lm003o?@00
103o?@000`000000o`000?l000050?l000030000003o?@00ocd000@0ocd000D000000?mj003oNP00
ogX0000000020?nh00070000003om@00ooD00?oe00000000c?l00000000208oo00060000001Ao`00
DOl00000000Do`0000000P00obP01@0000000?mV00000000oj<0000000800?oQ00<00000003Qo`00
hOl00P0000000`00Xol00000001Vo`02001Vo`0500000000:?l002So000Xo`000000100D0?l20000
00l0D@3o00<0000001@0o`0D0?l0100D0?l2000000<002So00<00000001Vo`00I_l00P00I_l01000
00000:?o002So`000002003Qo`0500000000on400?oQ003oh@0000000`00oj<00`0000000?mV003o
IP02003oIP0400000000obP00?lX003o:08000000P0Do`070?ooo`4000000@3oool00100oooo00H0
oooO09a80000000000000000M02oool:0?ooo`030000003oool0oooo00D0oooo00@0000001Co000D
o`005?l00P000002003o:08000000`00ofH2000000800?nS0P000002003oh@80000000<00>7o0000
000000000P00Xol2000000<006Ko00@00000000Xo`00:?l002So0P00000501@0o`030000001A0?l0
D@3o00l0D@3o00@0000001@0o`0D0?l0503o0P000002000Xo`8000000P00I_l0100000000:?o002S
o`000002003Qo`0300000000on40000000800?nS00P00000003oIP00ofH00000003o:00000005?l0
00000080DOl000H0000008oo002?o`0000000<co000000030?oe00040000003o^000okP0000000<0
ogX000<000000?lm003o?@00103o?@000`000000o`000?l000040?l000030000003o?@00ocd000@0
ocd000D000000?mj003oNP00ogX0000000020?nh00060000003om@00ooD00000003<o`0000000P2?
o`000`000000DOl00000000201Co0080000000@00?mV00000000oj<000000P00on402@0000000>7o
00000000Xol00000001Vo`00000002So0000000201@0o`030000001A0?l000000080S`3o00D00000
0<`0o`3<0?l0c03o000000030?l0m@030000003o0;P0o`2h00/0o`2h00<000000?l0m@3o0?D00P3o
0?D010000000c03o0<`0o`00000208l0o`0D0000001A0?l0D@3o0000000D0?l00000000Xo`000000
06Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000003oIP000002003o:0070000000D
o`000000057o00000000Sol0000000020<co00030000003om@00000000<0okP000@000000?mj003o
NP0000001P3o?@000`000000o`000?l000050?l000030000003o?@00ocd000@0ocd000<000000?mj
003oNP000P3oNP000`000000okP0000000030?oe00060000003<o`00c?l00000002?o`0000000P1A
o`000`0000005?l000000002003o:00300000000ofH0000000800?nS00<00000003oh@0000000P00
hOl0100000000:?o002So`000003001Vo`0500000000:?l002So000Xo`000000100D0?l00`000000
D@3o0540o`0>0540o`030000000D0?l0503o00<0503o0P000003000Xo`8000000`00I_l200000003
002So`000000000000800>7o00<00000003oh@00on400P000002003oX`8000000`00ofH200000080
0?lX0P00000401Co00030000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo
0?ooo`050?ooo`03057o000000005?l000<05?l00P000002003o:08000000`00ofH01@0000000?nS
003oX`00oj<0000000<00?oQ00<00000003Qo`00hOl00P000002002So`8000000P00I_l00`000000
02So000Xo`02000Xo`030000000D0?l0503o00<0503o00<000000540o`1A0?l03P1A0?l00`000000
503o01@0o`0301@0o`0500000000:?l002So000Xo`0000000P00I_l0100000000:?o002So`000002
003Qo`0300000000on40000000800?nS00P00000003oIP00ofH00000003o:00000005?l000000080
DOl000H0000008oo002?o`0000000<co000000020?oe00050000003o^000okP00?nh000000000`3o
NP000`000000ocd00?lm00040?lm008000001@3o00000`000000ocd00?lm00040?lm00050000003o
NP00ogX00?mj000000000P3o^00020000000ooD00?oe00000000c?l00000002?o`0000000P1Ao`00
4`0000005?l000000000obP00000003oIP0000000?nS00000000on400000003Qo`0000000:?o0000
0000I_l00000000Xo`0000000P0D0?l00`000000D@3o0000000208l0o`050000003<0?l0c03o0<`0
o`000000103o0?D00`000000o`2h0?l0^00:0?l0^0030000003o0?D0o`3e0080o`3e00@000000<`0
o`3<0?l000000P2?0?l050000000D@3o0540o`000000503o00000000:?l00000001Vo`0000000:?o
00000000hOl00000003oh@0000000?nS00000000ofH000000P00obP0200000005?l00000001Ao`00
000008oo00000000c?l00P0000000`3om@0000000?nh00020?nh00040000003oNP00ogX0000000H0
ocd000<000000?l0003o0000103o00000`000000ocd00?lm00040?lm00040000003oNP00ogX00?mj
008000000P3o^00010000000ooD00?oe000000020<co00030000002?o`0000000080DOl000<00000
01Co000000000P00obP00`0000000?mV00000002003oX`0600000000on40000000000000hOl00000
0P00Xol01@00000006Ko001Vo`00I_l0000000<002So00<0000001@0o`0D0?l00P0D0?l00`000000
D@3o0540o`0>0540o`030000000D0?l0503o00@0503o00D00000000Xo`00:?l002So00000003001V
o`8000000P00Xol00`0000000>7o003Qo`02000000800?oQ00D00000003oX`00oj<00?nS00000003
003oIP8000000P00obP2000000@05?l00P0000070?ooo`4000000@3oool00200oooo00<000000?oo
o`3oool01@3oool2057o00800000100Do`02000000800?lX0P000002003oIP8000000P00oj<20000
00800?oQ0P000002003Qo`0500000000Xol00:?o002So`0000000P00I_l00`00000002So000Xo`02
000Xo`800000100D0?l00`000000D@3o0540o`0>0540o`800000100D0?l01@00000002So000Xo`00
:?l00000008006Ko0P0000001P00Xol0000000000000hOl00000003oh@80000000T00?nS00000000
00000?mV00000000obP0000001Co000000000P1Ao`000`000000Sol0000000020<co00040000003o
m@00ooD0000000<0okP000D000000?mj003oNP00ogX0000000070?lm00050000003o0000o`000?l0
000000001`3o?@001@000000ogX00?mj003oNP0000000080okP000P000000?oe003om@0000000<co
00000000Sol000000080DOl001<0000001Co000000000?lX00000000ofH00000003oX`0000000?oQ
00000000hOl00000002So`00000006Ko00000000:?l000000080503o00H000000540o`1A0?l00000
08l0o`0000030<`0o`030000003o0?D0o`3e0080o`3e00<000000?l0^03o0;P02@3o0;P00`000000
o`3e0?l0m@020?l0m@050000003<0?l0c03o0<`0o`0000000P2?0?l00`000000D@3o0000000201@0
o`0E00000000:?l00000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH0
0000003o:00000005?l00000001Ao`0000000080Sol000<000000<co000000000P3om@0010000000
okP00?nh000000030?mj00030000003o?@00ocd000D0ocd000<000000?l0003o00000P3o00000`00
0000ocd00?lm00040?lm00050000003oNP00ogX00?mj000000000`3o^00010000000ooD00?oe0000
00020<co00030000002?o`0000000080DOl000<0000001Co000000000P00obP00`0000000?mV0000
0002003oX`0400000000on400000003Qo`80000000<00:?o0000000000000P00I_l20000008002So
0P00000301@0o`800000401A0?l00`000000503o01@0o`0301@0o`8000000`00:?l01@00000006Ko
001Vo`00I_l0000000<00:?o00@00000003Qo`00000000000P00on42000000800?nS0P000002003o
IP8000000P00obP2000000@05?l00P000002057o00L0oooo0@0000010?ooo`00803oool00`000000
oooo0?ooo`050?ooo`040000001Ao`00DOl0057o00800000100Do`02000000800?lX0P000002003o
IP8000000P00oj<2000000800?oQ0P0000000`00hOl0000000000002002So`0300000000I_l006Ko
008000001000:?l00`000000503o01@0o`0201@0o`030000001A0?l0D@3o00l0D@3o00<0000001@0
o`0D0?l00P0D0?l01@00000002So000Xo`00:?l0000000<006Ko00D00000002So`00Xol00000003Q
o`0200000004003oh@0000000?nS003oX`8000000P00obP01@0000005?l00000001Ao`0000000080
Sol000@000000<co003<o`0000000P3om@0010000000okP00?nh000000030?mj00030000003o?@00
ocd000D0ocd000@000000?l0003o00000000203o?@001@000000ogX00?mj003oNP0000000080okP0
00@000000?oe003om@0000000P3<o`000`000000Sol000000002057o0080000001000?lX00000000
ofH00000003oX`0000000?oQ00000000hOl00000002So`00000006Ko00000000:?l000000P0D0?l0
1P000000D@3o0540o`000000S`3o000000<0c03o00<000000?l0m@3o0?D00`3o0?D00`000000o`2h
0?l0^0080?l0^0030000003o0?D0o`3e0080o`3e00@000000<`0o`3<0?l000000`2?0?l03@000000
D@3o0000000D0?l00000000Xo`00000006Ko00000000Xol00000003Qo`0000000P00on4200000080
0?mV00/00000003o:00000005?l00000001Ao`00000008oo00000000c?l0000000020?oe00050000
003o^000okP00?nh000000000P3oNP000`000000ocd00?lm00060?lm00050000003o0000o`000?l0
000000001P3o?@001@000000ogX00?mj003oNP00000000<0okP000@000000?oe003om@0000000P3<
o`000`000000Sol000000002057o00030000000Do`00000000800?lX00<00000003oIP0000000P00
oj<01@0000000?oQ00000000hOl0000000800:?o00D00000001Vo`00I_l006Ko00000003000Xo`03
0000000D0?l0503o0080503o00<000000540o`1A0?l03`1A0?l00`000000503o01@0o`0201@0o`80
00000`00:?l20000008006Ko0P000002002So`80000000@00>7o00000000on400?oQ0P000002003o
X`8000000P00ofH2000000800?lX0P00000401Co00800000101Ao`000`000000oooo0?ooo`040?oo
o`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool00`2?o`000000057o0003057o0080
0000100Do`02000000800?lX0P000002003oIP8000000P00oj<01@0000000?oQ003oh@00on400000
00800>7o00@00000002So`00Xol000000`00I_l00`00000002So000Xo`02000Xo`030000000D0?l0
503o0080503o00<000000540o`1A0?l03`1A0?l00`000000503o01@0o`0201@0o`0500000000:?l0
02So000Xo`0000000P00I_l0100000000:?o002So`000003003Qo`8000000P00oj<00`0000000?mV
00000002003o:0050000000Do`000000057o000000000P2?o`0010000000c?l00<co000000020?oe
00040000003o^000okP0000000<0ogX000<000000?lm003o?@001@3o?@000`000000o`0000000009
0?lm00050000003oNP00ogX00?mj000000000P3o^00020000000ooD00?oe00000000c?l00000002?
o`0000000P1Ao`004`0000005?l000000000obP00000003oIP0000000?nS00000000on400000003Q
o`0000000:?o00000000I_l00000000Xo`0000000P0D0?l00`000000D@3o0000000208l0o`050000
003<0?l0c03o0<`0o`000000103o0?D00`000000o`2h0?l0^0070?l0^0030000003o0?D0o`3e0080
o`3e00D000000<`0o`3<0?l0c03o0000000208l0o`030000001A0?l000000080503o01D00000000X
o`00000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000003oIP0000000?lX0000
000Do`000000057o000000000P2?o`000`000000c?l0000000020?oe00040000003o^000okP00000
00<0ogX000<000000?lm003o?@001`3o?@000`000000o`00000000070?lm00050000003oNP00ogX0
0?mj000000000`3o^0000`000000ooD0000000020<co00060000002?o`00Sol00000001Ao`000000
0P0Do`001@0000000?lX00000000ofH0000000800?nS00D00000003oh@0000000>7o00000003002S
o`0400000000I_l006Ko000000<002So00<0000001@0o`0D0?l00P0D0?l00`000000D@3o0540o`0?
0540o`030000000D0?l0503o0080503o00<00000000Xo`00:?l00P00:?l01@00000006Ko001Vo`00
I_l0000000<00:?o00@00000003Qo`00hOl000000P00on401@0000000?nS003oX`00oj<000000080
0?mV0P000003003o:0030000000Do`005?l000805?l00P000004057o008000001`3oool100000040
oooo000P0?ooo`030000003oool0oooo00D0oooo0P2?o`02000000@0DOl00P00000401Co00800000
0P00obP2000000800?mV0P0000000`00oj<00000003oh@02003oh@0400000000hOl00>7o00000080
0:?o0P000002001Vo`0300000000:?l002So008002So00<0000001@0o`0D0?l00P0D0?l200000100
D@3o00<0000001@0o`0D0?l00P0D0?l20000008002So0P0000001@00I_l00000002So`00Xol00000
00800>7o00<00000003oh@0000000P00oj<00`0000000?mV00000002003o:0040000000Do`000000
057o0080000000D0Sol00000003<o`00c?l0000000020?oe00040000003o^000okP0000000<0ogX0
0P0000060?lm00030000003o?@00ocd000P0ocd000D000000?mj003oNP00ogX0000000020?nh0004
0000003om@00ooD000000080c?l001L0000008oo00000000DOl00000000Do`000000003o:0000000
0?mV00000000oj<00000003oh@0000000>7o00000000Xol00000001Vo`00000002So0000000201@0
o`030000001A0?l000000080S`3o00D000000<`0o`3<0?l0c03o000000050?l0m@030000003o0;P0
o`2h00H0o`2h00<000000?l0m@3o0?D00P3o0?D01@000000c03o0<`0o`3<0?l000000080S`3o00<0
00000540o`0000000P0D0?l05@00000002So00000000I_l00000002So`0000000>7o00000000on40
0000003oX`0000000?mV00000000obP0000001Co00000000DOl00000000208oo00030000003<o`00
00000080ooD000@000000?nh003o^00000000`3oNP000`000000ocd00?lm00080?lm00030000003o
?@00ocd000D0ocd000D000000?mj003oNP00ogX0000000020?nh008000000P3om@0010000000c?l0
0000002?o`0200000009057o000000005?l001Co000000000?lX00000000ofH0000000800?nS00<0
0000003oh@0000000P00hOl01@0000000:?o002So`00000006Ko008000000P00:?l2000000@0503o
00<000000540o`1A0?l03P1A0?l2000000@0503o0P000003000Xo`0500000000I_l006Ko001Vo`00
00000P00Xol2000000800>7o00D00000003oh@00on400000003oX`0300000003003oIP0000000000
00800?lX0`00000301Co00800000101Ao`0200000080Sol01`3oool100000040oooo000P0?ooo`03
0000003oool0oooo00D0oooo102?o`02000000@0DOl00P00000401Co000500000000obP00?lX003o
:00000000`00ofH0100000000?nS000000000002003oh@80000000<00>7o0000000000000P00Xol0
0`00000006Ko001Vo`02000000<002So0P00000401@0o`030000001A0?l0D@3o00h0D@3o0P000004
01@0o`0600000000:?l002So000Xo`00000006Ko0P0000001P00Xol0000000000000hOl00000003o
h@80000000H00?nS00000000ofH00000003o:000000201Co00060000001Ao`00DOl00000002?o`00
00000P3<o`0010000000ooD00?oe000000020?nh00030000003oNP00ogX00080ogX000<000000?lm
003o?@003`3o?@001@000000ogX00?mj003oNP0000000080okP001X000000?oe003om@0000000<co
00000000Sol00000001Ao`00000001Co000000000?lX00000000ofH00000003oX`0000000?oQ0000
0000hOl00000002So`00000006Ko0000008002So00D0000001@0o`000000D@3o0000000308l0o`04
0000003<0?l0c03o000000H0o`3e00<000000?l0^03o0;P0103o0;P00`000000o`3e0?l0m@030?l0
m@050000003<0?l0c03o0<`0o`0000000P2?0?l01@000000D@3o0000000D0?l00000008002So01D0
0000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH00000003o:0000000
5?l00000001Ao`00000008oo000000000P3<o`000`000000ooD0000000020?nh00050000003oNP00
ogX00?mj000000004P3o?@001@000000ogX00?mj003oNP0000000080okP000T000000?oe003om@00
ooD00000003<o`00000008oo000000000P1Ao`00200000005?l001Co000000000?lX00000000ofH0
00000P00oj<00`0000000?oQ00000002003Qo`0600000000Xol0000000000000I_l000000`00:?l0
0`000000503o01@0o`0201@0o`800000401A0?l00`000000503o01@0o`0201@0o`8000000`00:?l2
0000008006Ko0P0000000`00Xol0000000000002003Qo`8000000P00on400`0000000?nS00000003
003oIP0500000000obP00?lX003o:0000000100Do`02000000@0DOl00P00000408oo00L0oooo0@00
00010?ooo`00803oool00`000000oooo0?ooo`050?ooo`H0Sol00`000003057o008000000`0Do`02
000000800?lX0P000002003oIP80000000<00?nS0000000000000P00on4200000004003Qo`000000
0000002So`8000000P00I_l00`00000002So000Xo`02000Xo`8000000`0D0?l200000100D@3o0P00
000301@0o`8000000P00:?l01000000006Ko001Vo`000002002So`0900000000hOl0000000000000
on400000003oX`0000000?mV00<000000P0Do`001P000000DOl0057o00000000Sol000000080c?l0
00@000000?oe003om@0000000P3o^0001@000000ogX00?mj003oNP0000000140ocd000D000000?mj
003oNP00ogX0000000020?nh00040000003om@00ooD000000080c?l001L0000008oo00000000DOl0
0000000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o00000000Xol00000
001Vo`00000002So0000000201@0o`030000001A0?l000000080S`3o00@000000<`0o`3<0?l00000
1P3o0?D00`000000o`2h0?l0^0030?l0^0030000003o0?D0o`3e00@0o`3e00@000000<`0o`3<0?l0
00000P2?0?l06P000000D@3o0540o`000000503o00000000:?l00000001Vo`0000000:?o00000000
hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`0000000P2?o`00
0`000000c?l0000000020?oe00030000003o^000000000<0ogX000<000000?lm003o?@00403o?@00
1@000000ogX00?mj003oNP00000000<0okP000H000000?oe0000000000000<co0000000208oo0003
0000001Ao`00000000805?l000H00000003o:00000000?mV00000000oj<200000007003oh@000000
0>7o003Qo`0000000:?o00000002001Vo`0500000000:?l002So000Xo`0000000`0D0?l200000100
D@3o0P00000401@0o`0300000000:?l002So008002So00<00000001Vo`00I_l00P0000000`00Xol0
000000000002003Qo`8000000P00on4200000004003oX`0000000?mV003oIP8000000P00obP20000
00<05?l00P000004057o008000001P2?o`070?ooo`4000000@3oool00200oooo00<000000?ooo`3o
ool01@3oool908oo008000000`1Ao`03000000805?l00P000002003o:08000000P00ofH200000003
003oX`000000000000800?oQ00@00000003Qo`00hOl000000P00Xol200000003001Vo`0000000000
00@002So00<0000001@0o`0D0?l00P0D0?l00`000000D@3o0540o`0?0540o`030000000D0?l0503o
0080503o00@00000000Xo`00:?l000000P00I_l0100000000:?o002So`000002003Qo`0500000000
on400000003oX`0000000P00ofH30000000701Co00000000DOl0057o00000000Sol0000000020<co
00040000003om@00ooD000000080okP000D000000?mj003oNP00ogX00000000A0?lm00050000003o
NP00ogX00?mj000000000P3o^00070000000ooD00?oe00000000c?l00000002?o`000000057o0000
00005?l000000000obP00000003oIP0000000?nS00000000on400000003Qo`0000000:?o00000000
I_l00000000Xo`00000201@0o`030000001A0?l000000080S`3o00D000000<`0o`3<0?l0c03o0000
00060?l0m@050000003o0;P0o`2h0?l0^00000001P3o0?D01@000000c03o0<`0o`3<0?l000000080
S`3o01`000000540o`1A0?l0000001@0o`00000002So00000000I_l00000002So`0000000>7o0000
0000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`0000000P3<
o`000`000000ooD0000000020?nh00050000003oNP00ogX00?mj000000004@3o?@001@000000ogX0
0?mj003oNP00000000<0okP000<000000?oe000000000P3<o`001`000000Sol008oo00000000DOl0
0000000Do`000P0000001P00obP00000003oIP0000000?nS000000800?oQ00H00000003Qo`000000
0000002So`000002001Vo`0500000000:?l002So000Xo`0000000`0D0?l00`000000D@3o0540o`0?
0540o`030000000D0?l0503o00<0503o00@00000000Xo`00:?l002So0P0000000`00I_l000000000
0002002So`0300000000hOl00>7o0080000000D00?oQ0000000000000000003oX`0200000003003o
IP000000000000800?lX0P00000301Co00800000101Ao`02000000P0Sol01`3oool100000040oooo
000P0?ooo`030000003oool0oooo00D0oooo2`2?o`000`000000DOl0057o0003057o008000000P0D
o`02000000800?lX0P000002003oIP0600000000oj<00?nS00000000on4000000`00hOl010000000
0:?o002So`000003001Vo`0500000000:?l002So000Xo`0000001@0D0?l00`000000D@3o0540o`0>
0540o`030000000D0?l0503o0080503o00D00000000Xo`00:?l002So00000002001Vo`0400000000
Xol00:?o000000800>7o0P000002003oX`0300000000ofH0000000800?lX00D0000001Co00000000
DOl00000000208oo00030000003<o`0000000080ooD000D000000?nh003o^000okP0000000030?mj
00030000003o?@00ocd000h0ocd000D000000?mj003oNP00ogX0000000020?nh00060000003om@00
ooD00000003<o`0000000P2?o`020000000B01Co000000000?lX00000000ofH00000003oX`000000
0?oQ00000000hOl00000002So`00000006Ko00000000:?l000000P0D0?l00`000000D@3o00000002
08l0o`050000003<0?l0c03o0<`0o`0000001P3o0?D5000000H0o`3e00D000000<`0o`3<0?l0c03o
0000000208l0o`0L0000001A0?l0D@3o0000000D0?l00000000Xo`00000006Ko00000000Xol00000
003Qo`0000000?oQ00000000oj<00000003oIP0000000?lX0000000Do`000000057o00000000Sol0
00000080c?l000<000000?oe000000000P3o^0001@000000ogX00?mj003oNP0000000100ocd000<0
00000?mj003oNP000P3oNP0010000000okP00?nh000000020?oe00030000003<o`0000000080Sol0
00D00000057o000000005?l000000002003o:00500000000ofH00000003oX`0000000P00on400`00
00000>7o00000002002So`0400000000I_l006Ko000000<002So00D0000001@0o`0D0?l0503o0000
000A0540o`030000000D0?l0503o00<0503o00D00000000Xo`00:?l002So00000002001Vo`050000
0000Xol00:?o002So`0000000`00hOl00`0000000?oQ00000003003oX`0400000000ofH00?mV0000
00800?lX0P00000301Co00800000101Ao`02000000X0Sol01`3oool100000040oooo000P0?ooo`03
0000003oool0oooo00D0oooo302?o`02000000D0DOl000D0000001Co000Do`005?l000000003003o
:00400000000ofH00?mV000000800?nS00<00000003oh@0000000`00hOl0100000000:?o002So`00
0003001Vo`0500000000:?l002So000Xo`0000001@0D0?l00`000000D@3o0540o`0>0540o`030000
000D0?l0503o0080503o00D00000000Xo`00:?l002So00000002001Vo`0:00000000Xol00:?o0000
0000hOl00000003oh@0000000?nS000000800?mV00D00000003o:00000005?l000000002057o0003
0000002?o`0000000080c?l000@000000?oe003om@0000000P3o^0000`000000ogX00?mj00020?mj
00030000003o?@00ocd000d0ocd000D000000?mj003oNP00ogX0000000020?nh00060000003om@00
ooD00000003<o`0000000P2?o`0010000000DOl00000000Do`020000000>003oIP0000000?nS0000
0000on400000003Qo`0000000:?o00000000I_l00000000Xo`00000201@0o`030000001A0?l00000
0080S`3o00D000000<`0o`3<0?l0c03o0000000@0?l0m@050000003<0?l0c03o0<`0o`0000000P2?
0?l050000000D@3o0540o`000000503o00000000:?l00000001Vo`0000000:?o00000000hOl00000
003oh@0000000?nS00000000ofH000000P00obP200000080DOl00P0000020<co00030000003om@00
00000080okP000D000000?mj003oNP00ogX00000000A0?lm00050000003oNP00ogX00?mj00000000
0P3o^0001P000000ooD00?oe00000000c?l000000080Sol000<00000057o000000000P0Do`001@00
00000?lX00000000ofH0000000800?nS00<00000003oh@0000000P00hOl01P0000000:?o002So`00
000006Ko000000<002So00<0000001@0o`0D0?l00P0D0?l00`000000D@3o0540o`0>0540o`030000
000D0?l0503o00<0503o00D00000000Xo`00:?l002So00000003001Vo`0400000000Xol00:?o0000
00<00>7o00<00000003oh@0000000`00oj<0100000000?mV003oIP000002003o:0040000000Do`00
5?l001Co00800000101Ao`02000000`0Sol01`3oool100000040oooo000P0?ooo`030000003oool0
oooo00D0oooo3P2?o`02000000@0DOl00P00000201Co008000000P00obP200000006003oIP000000
0?nS003oX`0000000?oQ0P000002003Qo`80000000<00:?o0000000000000P00I_l20000008002So
0P00000401@0o`030000001A0?l0D@3o00h0D@3o0P00000301@0o`0500000000:?l002So000Xo`00
00000P00I_l02P0000000:?o002So`0000000>7o00000000on400000003oX`000002003oIP050000
0000obP0000001Co000000000P1Ao`000`000000Sol0000000020<co00040000003om@00ooD00000
0080okP000D000000?mj003oNP00ogX00000000@0?lm00050000003oNP00ogX00?mj000000000P3o
^00070000000ooD00?oe00000000c?l00000002?o`000000057o000000005?l000000000obP00000
003oIP0000000?nS00000000on400000003Qo`0000000:?o00000000I_l00000000Xo`00000201@0
o`030000001A0?l000000080S`3o00D000000<`0o`3<0?l0c03o0000000?0?l0m@050000003<0?l0
c03o0<`0o`0000000P2?0?l070000000D@3o0540o`000000503o00000000:?l00000001Vo`000000
0:?o00000000hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`00
000008oo000000020<co00030000003om@0000000080okP000D000000?mj003oNP00ogX00000000@
0?lm00030000003oNP00ogX00080ogX000<000000?nh000000000P3om@001P000000c?l00<co0000
0000Sol000000080DOl000D0000001Co000000000?lX00000002003oIP0500000000oj<00000003o
h@0000000P00hOl01P0000000:?o00000000000006Ko000000<002So00@0000001@0o`0D0?l0503o
0P00000A0540o`040000000D0?l0503o01@0o`8000000P00:?l20000008006Ko0P0000000`00Xol0
000000000002003Qo`80000000@00?oQ00000000oj<00?nS0P0000001000ofH0000000000000obP2
000000805?l00P000004057o008000003P2?o`070?ooo`4000000@3oool00200oooo00<000000?oo
o`3oool01@3oool@08oo00800000101Ao`001@0000005?l001Co000Do`00000000<00?lX00D00000
003oIP0000000000003oX`0200000003003oh@000000000000800>7o00<00000002So`00Xol00P00
0002001Vo`0500000000:?l002So000Xo`000000100D0?l00`000000D@3o0540o`0?0540o`050000
000D0?l0503o01@0o`0000000`00:?l01000000006Ko001Vo`000002002So`0600000000hOl00000
003oh@0000000?nS0P0000001P00ofH00000003o:00000005?l000000080DOl000<0000008oo0000
00000P3<o`0010000000ooD00?oe000000020?nh00050000003oNP00ogX00?mj000000003`3o?@00
1@000000ogX00?mj003oNP0000000080okP000P000000?oe003om@0000000<co00000000Sol00000
0080DOl00180000001Co000000000000003oIP0000000?nS00000000on400000003Qo`0000000:?o
00000000I_l00000000Xo`00000201@0o`030000001A0?l000000080S`3o00D000000<`0o`3<0?l0
c03o0000000?0?l0m@050000003<0?l0c03o0<`0o`0000000P2?0?l07@000000D@3o0540o`000000
503o00000000:?l00000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH0
0000003o:00000005?l00000001Ao`00000008oo00000000c?l00080000000D0ooD00000003o^000
okP0000000030?mj00030000003o?@00ocd000d0ocd000@000000?mj003oNP00ogX00P0000020?nh
00030000003om@0000000080c?l000<0000008oo000000000P1Ao`001@0000005?l000000000obP0
000000800?mV00D00000003oX`0000000?oQ00000002003Qo`0300000000Xol00000008006Ko00D0
0000000Xo`00:?l002So0000000301@0o`030000001A0?l0D@3o00l0D@3o0P00000301@0o`050000
0000:?l002So000Xo`0000000`00I_l0100000000:?o002So`000002003Qo`80000000@00?oQ0000
000000000?nS0P000002003oIP0400000000obP00?lX000000<05?l000<00000057o001Ao`000`1A
o`000`000000Sol008oo000>08oo00L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`05
0?oooa80Sol000<00000057o001Ao`000P1Ao`001@0000005?l001Co000Do`00000000<00?lX00@0
0000003oIP00ofH000000P00oj<0100000000?oQ003oh@000002003Qo`0400000000Xol00:?o0000
00<006Ko00D00000000Xo`00:?l002So0000000401@0o`030000001A0?l0D@3o00l0D@3o00D00000
01@0o`0D0?l0503o00000003000Xo`0800000000I_l006Ko00000000Xol00000003Qo`000002003o
h@0300000000oj<00?nS008000000P00obP00`0000005?l000000002057o00050000002?o`000000
0<co000000000`3om@0010000000okP00?nh000000030?mj00030000003o?@00ocd000`0ocd000D0
00000?mj003oNP00ogX0000000020?nh00080000003om@00ooD00000003<o`00000008oo00000002
057o000B0000000Do`00000000000000ofH00000003oX`0000000?oQ00000000hOl00000002So`00
000006Ko00000000:?l000000P0D0?l00`000000D@3o0000000208l0o`050000003<0?l0c03o0<`0
o`0000003`3o0?D01@000000c03o0<`0o`3<0?l000000080S`3o01h000000540o`1A0?l0000001@0
o`00000002So00000000I_l00000002So`0000000>7o00000000on400000003oX`0000000?mV0000
0000obP0000001Co00000000DOl00000002?o`0000000<co000000020?oe00040000003o^000okP0
000000<0ogX000<000000?lm003o?@00303o?@001@000000ogX00?mj003oNP00000000<0okP000<0
00000?oe000000000P3<o`000`000000Sol000000002057o00050000000Do`000000003o:0000000
0P00ofH00`0000000?nS00000002003oh@0500000000hOl00000002So`0000000`00I_l010000000
02So000Xo`00000301@0o`030000001A0?l0D@3o00l0D@3o00<0000001@0o`0D0?l00P0D0?l01@00
000002So000Xo`00:?l0000000<006Ko00@00000002So`00Xol000000P00hOl0100000000?oQ003o
h@000002003oX`0500000000ofH00?mV003oIP0000000`00obP0100000005?l001Co00000004057o
008000004@2?o`070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3ooolC08oo0003
0000001Ao`00DOl00080DOl00P00000201Co008000000P00obP0100000000?mV003oIP000002003o
X`0400000000on400?oQ000000800>7o0P0000000`00Xol00000001Vo`02001Vo`0500000000:?l0
02So000Xo`000000100D0?l00`000000D@3o0540o`0?0540o`050000000D0?l0503o01@0o`000000
0`00:?l02P00000006Ko001Vo`0000000:?o00000000hOl00000003oh@000002003oX`0500000000
ofH00000003o:00000000P0Do`001@000000DOl00000002?o`0000000080c?l000@000000?oe003o
m@0000000P3o^0001@000000ogX00?mj003oNP00000000l0ocd000@000000?mj003oNP0000000P3o
^0001P000000ooD00?oe00000000c?l000000080Sol000H00000057o000000005?l000000000obP2
000000800?nS0P0000001000hOl00000002So`000002001Vo`8000000P0D0?l00`000000D@3o0000
000208l0o`050000003<0?l0c03o0<`0o`0000003P3o0?D01@000000c03o0<`0o`3<0?l000000080
S`3o01h000000540o`1A0?l0000001@0o`00000002So00000000I_l00000002So`0000000>7o0000
0000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`0000000<co
000000020?oe00040000003o^000okP0000000<0ogX000<000000?lm003o?@003@3o?@001@000000
ogX00?mj003oNP00000000<0okP000D000000?oe00000000c?l00000000208oo00050000001Ao`00
000001Co000000000P00obP00`0000000?mV00000002003oX`0500000000on400000003Qo`000000
0P00Xol01000000006Ko001Vo`000002000Xo`050000000D0?l0503o01@0o`0000004@1A0?l00`00
0000503o01@0o`0201@0o`0500000000:?l002So000Xo`0000000`00I_l0100000000:?o00000000
0002003Qo`0400000000on400?oQ000000800?nS00D00000003oIP00ofH00?mV00000002003o:080
000000<05?l0000000000000101Ao`000`000000Sol008oo000A08oo00L0oooo0@0000010?ooo`00
803oool00`000000oooo0?ooo`050?oooa@0Sol00P000004057o008000000P0Do`0200000004003o
:00000000?mV003oIP80000000D00?nS0000000000000?oQ00000003003Qo`0400000000Xol00000
0000008006Ko0P000002000Xo`030000000D0?l0503o0080503o00<000000540o`1A0?l03`1A0?l0
1@000000503o01@0o`0D0?l00000008002So00H00000001Vo`00I_l00000002So`000002003Qo`03
00000000on40000000800?nS00L00000003oIP0000000?lX0000000Do`0000000080DOl000<00000
08oo000000000P3<o`0010000000ooD00?oe000000020?nh00050000003oNP00ogX00?mj00000000
3P3o?@001@000000ogX00?mj003oNP0000000080okP000X000000?oe003om@0000000<co00000000
Sol00000001Ao`0000000P0Do`020000000>003oIP0000000?nS00000000on400000003Qo`000000
0:?o00000000I_l00000000Xo`00000201@0o`030000001A0?l000000080S`3o00D000000<`0o`3<
0?l0c03o0000000=0?l0m@050000003<0?l0c03o0<`0o`0000000P2?0?l07P000000D@3o0540o`00
0000503o00000000:?l00000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000
ofH00000003o:00000005?l00000001Ao`00000008oo00000000c?l000000080ooD000@000000?nh
003o^00000000`3oNP000`000000ocd00?lm000<0?lm00050000003oNP00ogX00?mj000000000`3o
^0000`000000ooD0000000020<co00030000002?o`0000000080DOl000L0000001Co000000000?lX
00000000ofH0000000800?nS00<00000003oh@0000000P00hOl00`0000000:?o00000002001Vo`04
00000000:?l002So000000@0503o00<000000540o`1A0?l03`1A0?l01@000000503o01@0o`0D0?l0
000000<002So00<00000001Vo`00I_l00P0000000`00Xol00000003Qo`02003Qo`0500000000on40
000000000000oj<00P000002003oIP80000000<00?lX0000000000000P0Do`000`000000DOl0057o
0002057o00800000502?o`070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool9
08oo00D00000202?o`02000000@0DOl000H0000001Co000Do`005?l000000000obP2000000800?mV
00D00000003oX`00oj<00000003oh@02000000800>7o00@00000002So`00Xol000000`00I_l00`00
000002So000Xo`02000000<0503o00<000000540o`1A0?l03`1A0?l01@000000503o01@0o`0D0?l0
0000008002So00D00000001Vo`00I_l00000002So`0200000003003Qo`0000000?oQ0080000000P0
0?nS00000000ofH00000003o:00000005?l000000080DOl000<0000008oo000000000P3<o`000`00
0000ooD0000000030?nh00050000003oNP00ogX00?mj000000003P3o?@0010000000ogX00?mj0000
00020?nh00060000003om@00ooD00000003<o`0000000P2?o`001P000000DOl00000000Do`000000
003o:080000000`00?nS00000000on400000003Qo`0000000:?o00000000I_l00000000Xo`000002
01@0o`030000001A0?l000000080S`3o00D000000<`0o`3<0?l0c03o0000000=0?l0m@050000003<
0?l0c03o0<`0o`0000000P2?0?l07P000000D@3o0540o`000000503o00000000:?l00000001Vo`00
00000:?o00000000hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001A
o`00000008oo00000000c?l000000080ooD000@000000?nh003o^00000000`3oNP000`000000ocd0
0?lm000;0?lm00030000003oNP00ogX00080ogX000@000000?nh003o^00000000P3om@000`000000
c?l00000000208oo00090000001Ao`00000001Co000000000?lX00000000ofH0000000800?nS00@0
0000003oh@0000000>7o0P0000001@00Xol00000001Vo`00I_l00000008002So00<0000001@0o`0D
0?l00P0D0?l00`000000D@3o0540o`0>0540o`8000000`0D0?l00`00000002So000Xo`0200000080
06Ko00@00000002So`00Xol000000P00hOl200000005003oh@0000000?nS003oX`0000000P00ofH2
000000800?lX00<0000001Co000Do`000P000003057o00800000202?o`05000000T0Sol01`3oool1
00000040oooo000P0?ooo`030000003oool0oooo00D0oooo00<0000008oo002?o`001`2?o`001000
0000c?l00<co003<o`02000000P0Sol000<00000057o001Ao`000P1Ao`02000000805?l000@00000
003o:000obP000000P00ofH0100000000?nS003oX`000002003oh@0300000000hOl00>7o00800000
00<00:?o00000000I_l00P00I_l01@00000002So000Xo`00:?l0000000<0503o00<000000540o`1A
0?l03P1A0?l00`000000503o01@0o`0201@0o`0400000000:?l002So0000008006Ko00<00000002S
o`00Xol00P000002003oh@0300000000oj<0000000800?mV00L00000003o:00000005?l00000001A
o`0000000080Sol000H000000<co003<o`0000000?oe000000020?nh00030000003oNP00ogX00080
ogX000<000000?lm003o?@002`3o?@0010000000ogX00?mj000000020?nh00040000003om@00ooD0
00000080c?l000P0000008oo00000000DOl00000000Do`000000003o:080000000`00?nS00000000
on400000003Qo`0000000:?o00000000I_l00000000Xo`00000201@0o`030000001A0?l000000080
S`3o00D000000<`0o`3<0?l0c03o0000000=0?l0m@050000003<0?l0c03o0<`0o`0000000P2?0?l0
7P000000D@3o0540o`000000503o00000000:?l00000001Vo`0000000:?o00000000hOl00000003o
h@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`00000008oo00000000c?l00000
0080ooD000<000000?nh00000000103oNP000`000000ocd00?lm000:0?lm00030000003oNP00ogX0
0080ogX000@000000?nh003o^00000000P3om@000`000000c?l00000000208oo00070000001Ao`00
000001Co000000000?lX00000002003oIP0700000000oj<00000003oh@0000000>7o00000002002S
o`0400000000I_l006Ko000000<002So00D0000001@0o`0D0?l0503o0000000@0540o`030000000D
0?l0503o0080503o00@00000000Xo`00:?l000000`00I_l0100000000:?o000000000002003Qo`04
00000000on400?oQ000000800?nS00@00000003oIP00ofH000000P00obP2000000805?l000<00000
057o001Ao`000P1Ao`000`000000Sol008oo000608oo00060000002?o`00Sol00000003<o`000000
2P2?o`000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo0P0000060?ooo`03057o0000
0000Sol000L0Sol000<000000<co003<o`000`3<o`02000000L0Sol00P000004057o00040000000D
o`005?l0000000800?lX00<00000003oIP00ofH00P0000001@00oj<00000003oh@00on40000000<0
0>7o00@00000002So`00000000000P00I_l01@00000002So000Xo`00:?l0000000<0503o00<00000
0540o`1A0?l03P1A0?l00`000000503o01@0o`0201@0o`0600000000:?l002So00000000I_l00000
0P00Xol00`0000000>7o00000002003oh@0700000000oj<00000003oIP0000000?lX0000000201Co
00030000001Ao`0000000080Sol000D000000<co00000000ooD0000000030?nh00050000003oNP00
ogX00?mj000000003@3o?@001@000000ogX00?mj003oNP0000000080okP000<000000?oe00000000
0P3<o`0020000000Sol00000001Ao`00000001Co000000000?lX0P0000003000oj<00000003oh@00
00000>7o00000000Xol00000001Vo`00000002So00000080503o00H000000540o`1A0?l0000008l0
o`0000040<`0o`030000003o0?D0o`3e00T0o`3e00D000000<`0o`3<0?l0c03o0000000308l0o`0I
0000001A0?l0000001@0o`00000002So00000000I_l00000002So`0000000>7o00000000on400000
003oX`0000000?mV00000000obP0000001Co00000000DOl00000000208oo008000000P3om@001000
0000okP00?nh000000030?mj00030000003o?@00ocd000X0ocd000<000000?mj003oNP000P3oNP00
10000000okP00?nh000000030?oe00090000003<o`00000008oo00000000DOl00000000Do`000000
00800?lX00D00000003oIP0000000?nS00000002003oh@8000000`00Xol01000000006Ko001Vo`00
0002000Xo`050000000D0?l0503o01@0o`000000401A0?l00`000000503o01@0o`0201@0o`040000
0000:?l002So0000008006Ko0P0000000`00Xol00000003Qo`02003Qo`0500000000on400?oQ0000
0000oj<00P000002003oIP0400000000obP00?lX000000<05?l000<00000057o001Ao`000P1Ao`00
0`000000Sol008oo000508oo0080000000H0c?l00000002?o`0000000<co0000000908oo00030000
003oool0oooo00@0oooo0P0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`80DOl000<0
000008oo002?o`001P2?o`000`000000c?l00<co00040<co008000001`2?o`000`000000DOl0057o
0002057o00030000000Do`005?l00080000000<00?lX00000000ofH00P00ofH01@0000000?nS0000
000000000?oQ008000000P00hOl0100000000:?o002So`000002001Vo`0500000000:?l002So000X
o`0000000`0D0?l00`000000D@3o0540o`0>0540o`050000000D0?l0503o01@0o`0000000`00:?l0
0`00000006Ko00000002002So`0500000000hOl00000003oh@0000000P00oj<02@0000000?mV0000
0000obP0000001Co00000000DOl00000000208oo00060000003<o`00c?l00000003om@0000000P3o
^0000`000000ogX00?mj00020?mj00030000003o?@00ocd000X0ocd000D000000?mj003oNP00ogX0
000000020?nh00040000003om@00ooD000000080c?l000P0000008oo00000000DOl00000000Do`00
0000003o:080000000h00?nS00000000on400000003Qo`0000000:?o00000000I_l00000000Xo`00
0000503o00000080D@3o00@0000008l0o`2?0?l000000`3<0?l00`000000o`3e0?l0m@090?l0m@05
0000003<0?l0c03o0<`0o`0000000P2?0?l07P000000D@3o0540o`000000503o00000000:?l00000
001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l0
0000001Ao`00000008oo00000000c?l000000080ooD000@000000?nh003o^00000000`3oNP000`00
0000ocd00?lm000:0?lm00030000003oNP00ogX00080ogX000@000000?nh003o^00000000P3om@00
0`000000c?l00000000208oo008000000P0Do`000`0000000?lX00000002003oIP0700000000oj<0
0000003oh@0000000>7o00000002002So`0400000000I_l006Ko0000008002So00D0000001@0o`0D
0?l0503o0000000@0540o`030000000D0?l0503o0080503o00L00000000Xo`00:?l002So00000000
I_l0000000800:?o00<00000003Qo`00hOl00P0000001@00on40000000000000oj<0000000<00?mV
00@00000003o:000000000000P0Do`02000000@0DOl000<0000008oo002?o`00102?o`02000000@0
c?l000<0000008oo00000000202?o`0200000003057o003oool0oooo00D0oooo0@0000010?ooo`00
803oool00`000000oooo0?ooo`050?ooo`<0DOl00P00000708oo00030000003<o`00c?l000D0c?l0
0P00000608oo008000000`1Ao`001P0000005?l001Co000Do`000000003o:08000000P00ofH01000
00000?nS003oX`000002003oh@0400000000hOl00>7o000000800:?o00@00000001Vo`00I_l00000
0`00:?l01@000000503o01@0o`0D0?l000000100D@3o00D0000001@0o`0D0?l0503o00000003000X
o`0300000000I_l0000000800:?o00@00000003Qo`0000000?oQ0P0000002P00oj<00000003oIP00
00000?lX0000000Do`000000057o0000000208oo00030000003<o`0000000080ooD000@000000?nh
003o^0000000103oNP000`000000ocd00?lm00090?lm00050000003oNP00ogX00?mj000000000`3o
^0000`000000ooD0000000020<co00080000002?o`000000057o000000005?l000000000obP20000
000>003oX`0000000?oQ00000000hOl00000002So`00000006Ko00000000:?l0000001@0o`000002
0540o`040000002?0?l0S`3o000000@0c03o00<000000?l0m@3o0?D0203o0?D01@000000c03o0<`0
o`3<0?l000000080S`3o01h000000540o`1A0?l0000001@0o`00000002So00000000I_l00000002S
o`0000000>7o00000000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl00000
002?o`0000000<co000000020?oe00040000003o^000okP0000000<0ogX000<000000?lm003o?@00
2@3o?@000`000000ogX00?mj00020?mj00040000003o^000okP000000080ooD000`000000<co003<
o`00000008oo00000000DOl00000000Do`000000003o:0000002003oIP0700000000oj<00000003o
h@0000000>7o00000002002So`0400000000I_l006Ko0000008002So00D0000001@0o`0D0?l0503o
0000000@0540o`030000000D0?l0503o0080503o00<00000000Xo`00:?l00P000002001Vo`060000
0000Xol00:?o00000000hOl000000P00on40100000000?nS003oX`000002003oIP80000000<00?lX
0000000Do`000P0Do`000`000000DOl0057o0002057o008000001@2?o`02000000L0c?l000<00000
08oo002?o`001P2?o`0010000000DOl0057o001Ao`070?ooo`4000000@3oool00200oooo00<00000
0?ooo`3oool01@3oool00`000000DOl0057o0002057o00030000002?o`00Sol000D0Sol000<00000
0<co003<o`001P3<o`02000000H0Sol00P000002057o00050000000Do`005?l001Co000000000P00
obP0100000000?mV003oIP000002003oX`0400000000on400?oQ000000800>7o00@00000002So`00
Xol000000P00I_l01@00000002So000Xo`00:?l0000000<0503o00<000000540o`1A0?l03P1A0?l0
1@000000503o01@0o`0D0?l00000008002So00H00000001Vo`00I_l00000002So`000002003Qo`05
00000000on400000003oX`0000000P00ofH01`0000000?lX0000000Do`000000057o000000000P2?
o`000`000000c?l0000000020?oe00030000003o^000000000@0ogX000<000000?lm003o?@002@3o
?@001@000000ogX00?mj003oNP00000000<0okP000<000000?oe000000000P3<o`0020000000Sol0
0000001Ao`00000001Co000000000?lX0P0000003P00oj<00000003oh@0000000>7o00000000Xol0
0000001Vo`00000002So0000000D0?l000000P1A0?l010000000S`3o08l0o`0000040<`0o`030000
003o0?D0o`3e00L0o`3e00<000000<`0o`3<0?l00P3<0?l010000000S`3o08l0o`0000020540o`0K
0000000D0?l00000000Xo`00000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000
003oIP0000000?lX0000000Do`000000057o00000000Sol00000003<o`0000000080ooD000@00000
0?nh003o^00000000`3oNP000`000000ocd00?lm00090?lm00050000003oNP00ogX00?mj00000000
0`3o^0000`000000ooD0000000020<co00090000002?o`000000057o000000005?l000000000obP0
000000800?mV00D00000003oX`0000000?oQ00000002003Qo`0300000000Xol00000008006Ko00@0
0000000Xo`00:?l000000`0D0?l00`000000D@3o0540o`0>0540o`030000000D0?l0503o0080503o
00@00000000Xo`00:?l000000`00I_l01P0000000:?o0000000000000>7o000000800?oQ00@00000
003oX`00oj<000000P00ofH0100000000?lX003o:000000301Co00040000001Ao`00DOl0057o0080
00001@2?o`02000000P0c?l000<0000008oo002?o`001P2?o`000`000000DOl0057o0002057o0003
0000003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0301Co
00000000DOl000<0DOl000<0000008oo002?o`001@2?o`000`000000c?l00<co00070<co00800000
1P2?o`0010000000DOl0057o0000000301Co000400000000obP00?lX000000800?mV00<00000003o
X`00oj<00P0000001@00on400000003Qo`00hOl0000000800:?o00<00000001Vo`00I_l00P000000
0`00:?l0000001@0o`0301@0o`030000001A0?l0D@3o00h0D@3o00D0000001@0o`0D0?l0503o0000
0002000Xo`0300000000I_l0000000800:?o00<00000003Qo`0000000P00on401@0000000?nS0000
0000ofH0000000800?lX0P000002057o00050000002?o`0000000<co000000000`3om@000`000000
okP0000000040?mj00030000003o?@00ocd000P0ocd000D000000?mj003oNP00ogX0000000030?nh
00030000003om@0000000080c?l000X0000008oo00000000DOl00000000Do`000000003o:0000000
0?mV0P0000002000on400000003Qo`0000000:?o00000000I_l000000P00:?l200000080D@3o00@0
000008l0o`2?0?l00000103<0?l00`000000o`3e0?l0m@070?l0m@030000003<0?l0c03o0080c03o
00<0000008l0o`0000000P1A0?l00`000000503o01@0o`020000000F001Vo`0000000:?o00000000
hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`00000008oo0000
0000c?l000000080ooD000@000000?nh003o^0000000103oNP000`000000ocd00?lm00080?lm0003
0000003oNP00ogX00080ogX000@000000?nh003o^00000000P3om@000`000000c?l00<co00020000
000;057o000000005?l001Co000000000?lX00000000ofH00000003oX`0000000P00on401@000000
0>7o00000000Xol00000008006Ko00D00000000Xo`00:?l002So0000000301@0o`030000001A0?l0
D@3o00d0D@3o00<0000001@0o`0D0?l00P0D0?l01000000002So000Xo`000003001Vo`0300000000
Xol0000000800>7o00@00000003oh@00000000000P00oj<0100000000?mV003oIP000002003o:005
0000000Do`005?l001Co000000000`1Ao`000`000000Sol008oo000308oo008000002@3<o`000`00
0000Sol008oo000508oo00800000101Ao`000`000000oooo0?ooo`050?ooo`4000000@3oool00200
oooo00<000000?ooo`3oool01@3oool201Co00030000001Ao`00DOl00080DOl000<0000008oo002?
o`001@2?o`000`000000c?l00<co00080<co008000001@2?o`000`000000DOl0057o000200000080
5?l00P0000001@00obP0000000000000ofH0000000<00?nS00<00000003oh@0000000P00hOl01000
00000:?o002So`000003001Vo`0300000000:?l0000000@0503o00<000000540o`1A0?l03@1A0?l0
1@000000503o01@0o`0D0?l0000000<002So00<00000001Vo`0000000P00Xol01`0000000>7o0000
0000on400000003oX`0000000P00ofH01`0000000?lX0000000Do`000000057o000000000P2?o`00
0`000000c?l0000000020?oe00040000003o^000okP0000000@0ogX000<000000?lm003o?@00203o
?@001@000000ogX00?mj003oNP00000000<0okP000<000000?oe000000000P3<o`0200000080DOl0
0P0000004P00obP00000003oIP0000000?nS00000000on400000003Qo`0000000:?o00000000I_l0
0000000Xo`000000503o00000080D@3o00@0000008l0o`2?0?l00000103<0?l00`000000o`3e0?l0
m@050?l0m@030000003<0?l0c03o0080c03o00@0000008l0o`2?0?l000000P1A0?l00`000000503o
01@0o`020000000F001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH00000
003o:00000005?l00000001Ao`00000008oo00000000c?l000000080ooD000@000000?nh003o^000
0000103oNP000`000000ocd00?lm00070?lm00030000003oNP00ogX00080ogX000L000000?nh003o
^000okP00000003om@0000000080c?l000/0000008oo00000000DOl00000000Do`000000003o:000
00000?mV00000002003oX`0500000000on400000003Qo`0000000P00Xol00`00000006Ko00000003
000Xo`050000000D0?l0503o01@0o`0000003`1A0?l00`000000503o01@0o`0201@0o`0400000000
:?l002So000000<006Ko00<00000002So`0000000P00hOl00`0000000?oQ00000003003oX`040000
0000ofH00000000000800?lX00<0000001Co000Do`000P000002057o008000001@2?o`000`000000
c?l00<co00080<co00030000002?o`00Sol000D0Sol000<00000057o001Ao`000P1Ao`0200000003
01Co003oool0oooo00D0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`030000
000Do`005?l0008000000`1Ao`02000000H0Sol000<000000<co003<o`00203<o`000`000000Sol0
08oo000408oo00050000001Ao`00DOl0057o000000000`0Do`001@0000000?lX003o:00000000?mV
0080000000@00?nS00000000on400?oQ0P0000001@00hOl00000002So`00Xol00000008006Ko00<0
0000000Xo`00:?l00P00000301@0o`030000001A0?l0D@3o00d0D@3o00D0000001@0o`0D0?l0503o
00000003000Xo`0300000000I_l0000000800:?o00L00000003Qo`0000000?oQ00000000oj<00000
00800?mV00T00000003o:00000005?l00000001Ao`00000008oo000000000P3<o`0010000000ooD0
0?oe000000020?nh00030000003oNP00ogX00080ogX000<000000?lm003o?@001`3o?@001@000000
ogX00?mj003oNP00000000<0okP000<000000?oe000000000P3<o`001P000000Sol00000001Ao`00
000001Co0080000001000?mV00000000oj<00000003oh@0000000>7o00000000Xol00000001Vo`00
000002So0000000D0?l000000P1A0?l010000000S`3o08l0o`0000040<`0o`030000003o0?D0o`3e
00D0o`3e00<000000<`0o`3<0?l00P3<0?l010000000S`3o08l0o`0000020540o`0K0000000D0?l0
0000000Xo`00000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000003oIP000000
0?lX0000000Do`000000057o00000000Sol00000003<o`0000000080ooD000@000000?nh003o^000
0000103oNP000`000000ocd00?lm00060?lm00030000003oNP00ogX00080ogX000D000000?nh003o
^000okP0000000020?oe00070000003<o`00000008oo00000000DOl00000000201Co000;00000000
obP00000003oIP0000000?nS00000000on400000003Qo`0000000P00Xol01000000006Ko001Vo`00
0002000Xo`050000000D0?l0503o01@0o`0000003`1A0?l00`000000503o01@0o`0201@0o`040000
0000:?l002So0000008006Ko0P0000000`00Xol00000003Qo`02000000800?oQ00H00000003oX`00
oj<00000003oIP000002003o:08000000P0Do`001@000000DOl0057o001Ao`00000000H0Sol000<0
00000<co003<o`00203<o`000`000000Sol008oo000508oo00040000001Ao`00DOl0057o00800000
0`0Do`000`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3o
ool00`00obP0000001Co000201Co00030000001Ao`00DOl00080DOl000<0000008oo002?o`00102?
o`000`000000c?l00<co00080<co008000001@2?o`001@000000DOl0057o001Ao`00000000<05?l0
00@00000003o:000obP000000P00ofH00`0000000?nS00000003003oh@0300000000hOl000000080
0:?o00@00000001Vo`00I_l000000`00:?l01@000000503o01@0o`0D0?l0000000l0D@3o00D00000
01@0o`0D0?l0503o00000002000Xo`0600000000I_l006Ko00000000Xol000000P00hOl01`000000
0?oQ00000000oj<00000003oIP0000000P00obP200000080DOl000<0000008oo000000000P3<o`00
0`000000ooD0000000020?nh00030000003oNP00ogX00080ogX000<000000?lm003o?@001`3o?@00
1@000000ogX00?mj003oNP00000000<0okP000/000000?oe003om@0000000<co00000000Sol00000
001Ao`00000001Co00020000000@003oIP0000000?nS00000000on400000003Qo`0000000:?o0000
0000I_l00000000Xo`000000503o00000080D@3o00@0000008l0o`2?0?l000001@3<0?l00`000000
o`3e0?l0m@030?l0m@030000003<0?l0c03o00<0c03o00@0000008l0o`2?0?l000000P1A0?l06`00
0000503o00000000:?l00000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000
ofH00000003o:00000005?l00000001Ao`00000008oo00000000c?l0000000020?oe00040000003o
^000okP0000000@0ogX000<000000?lm003o?@001@3o?@000`000000ogX00?mj00020?mj00050000
003o^000okP00?nh000000000P3om@000`000000c?l00000000208oo008000000P0Do`002`000000
0?lX00000000ofH00000003oX`0000000?oQ00000000hOl0000000800:?o00@00000001Vo`00I_l0
00000P00:?l01@000000503o01@0o`0D0?l0000000l0D@3o00<0000001@0o`0D0?l00P0D0?l01000
000002So000Xo`000002001Vo`0600000000Xol00:?o00000000hOl000000`00on401P0000000?nS
0000000000000?mV000000800?lX00D0000001Co000Do`005?l000000003057o00030000002?o`00
Sol000<0Sol00P00000:0<co00030000002?o`00Sol000@0Sol00P000003057o00030000000Do`00
5?l000805?l000<000000?ooo`3oool01@3oool100000040oooo000P0?ooo`030000003oool0oooo
00D0oooo0P00obP01@0000005?l001Co000Do`00000000@0DOl000<0000008oo002?o`00102?o`00
0`000000c?l00<co00090<co00800000102?o`0200000080DOl00P00000201Co000400000000obP0
0?lX000000800?mV00H00000003oX`0000000000003oh@000002003Qo`0400000000Xol00:?o0000
008006Ko00@00000000Xo`00:?l00000100D0?l00`000000D@3o0540o`0=0540o`040000000D0?l0
503o000000<002So00<00000001Vo`0000000P00Xol00`0000000>7o00000002003oh@8000000P00
ofH02@0000000?lX0000000Do`000000057o00000000Sol0000000020<co00040000003om@00ooD0
00000080okP000<000000?mj003oNP000P3oNP000`000000ocd00?lm00060?lm00050000003oNP00
ogX00?mj000000000`3o^0002`000000ooD00?oe00000000c?l00000002?o`000000057o00000000
5?l00080000001000?mV00000000oj<00000003oh@0000000>7o00000000Xol00000001Vo`000000
02So0000000D0?l000000P1A0?l010000000S`3o08l0o`0000050<`0o`030000003o0?D0o`3e00<0
o`3e00<000000<`0o`3<0?l00P3<0?l01`000000S`3o08l0o`2?0?l000000540o`0000000P0D0?l0
1P00000002So00000000I_l00000002So`80000000h00?oQ00000000oj<00000003oIP0000000?lX
0000000Do`000000057o00000000Sol000000080c?l000<000000?oe000000000P3o^0000`000000
ogX00?mj00030?mj00030000003o?@00ocd000D0ocd000<000000?mj003oNP000P3oNP001`000000
okP00?nh003o^00000000?oe000000000P3<o`003@000000Sol00000001Ao`00000001Co00000000
0?lX00000000ofH00000003oX`0000000P00on401@0000000>7o00000000Xol00000008006Ko00D0
0000000Xo`00:?l002So0000000301@0o`030000001A0?l0D@3o00`0D@3o00<0000001@0o`0D0?l0
0P0D0?l01000000002So000Xo`000003001Vo`0300000000Xol00:?o008000000P00on4200000005
003oX`0000000?mV003oIP0000000P00obP00`0000005?l001Co000200000080DOl00P00000508oo
00030000003<o`00c?l000T0c?l000<0000008oo002?o`00102?o`000`000000DOl0057o0002057o
00040000000Do`005?l001Co0080000000<00?lX0?ooo`3oool01@3oool100000040oooo000P0?oo
o`030000003oool0oooo00D0oooo00@00000003o:000obP000000`0Do`000`000000DOl0057o0002
057o008000001@2?o`000`000000c?l00<co000:0<co00030000002?o`00Sol000<0Sol000D00000
057o001Ao`00DOl00000000201Co000400000000obP00?lX000000800?mV00H00000003oX`00oj<0
0000003oh@000002003Qo`0400000000Xol00:?o0000008006Ko00@00000000Xo`00:?l00000100D
0?l00`000000D@3o0540o`0=0540o`040000000D0?l0503o0000008002So00H00000001Vo`00I_l0
0000002So`000002003Qo`0700000000on400000003oX`0000000?mV00000002003o:08000000P1A
o`000`000000Sol0000000020<co00030000003om@0000000080okP000<000000?mj003oNP000`3o
NP000`000000ocd00?lm00050?lm00030000003oNP00ogX00080ogX000@000000?nh003o^0000000
0P3om@000`000000c?l00000000208oo000F0000001Ao`00000000000000obP00000003oIP000000
0?nS00000000on400000003Qo`0000000:?o00000000I_l00000000Xo`000000503o00000080D@3o
00@0000008l0o`2?0?l000001@3<0?l5000000D0c03o00L0000008l0o`2?0?l0S`3o0000001A0?l0
00000080503o00H00000000Xo`00000006Ko00000000Xol20000000@003oh@0000000?nS00000000
ofH00000003o:00000005?l00000001Ao`00000008oo00000000c?l000000080ooD000@000000?nh
003o^00000001@3oNP000`000000ocd00?lm00040?lm00030000003oNP00ogX00080ogX000D00000
0?nh003o^000okP0000000020?oe00030000003<o`0000000080Sol00P0000003@0Do`000000003o
:000obP00000003oIP0000000?nS00000000on400000003Qo`0000000P00Xol01000000006Ko001V
o`000002000Xo`050000000D0?l0503o01@0o`0000003P1A0?l00`000000503o01@0o`0201@0o`04
00000000:?l002So000000<006Ko00P00000002So`0000000000003Qo`0000000?oQ000000800?nS
00@00000003oIP00ofH000000P00obP0100000005?l001Co00000003057o00030000002?o`00Sol0
00<0Sol00P00000;0<co00030000002?o`00Sol000@0Sol000<00000057o001Ao`000P1Ao`001@00
00005?l001Co000Do`00000000<00?lX00<000000?ooo`3oool0103oool100000040oooo000P0?oo
o`030000003oool0oooo00D0oooo00@00?mV00000000obP00?lX0P00000201Co00800000101Ao`00
0`000000Sol008oo000308oo00030000003<o`00c?l000X0c?l000<0000008oo002?o`000`2?o`00
1@000000DOl0057o001Ao`00000000805?l000@00000003o:000obP000000P00ofH01@0000000?nS
003oX`0000000?oQ0080000000D00>7o00000000Xol00:?o00000002001Vo`0400000000:?l002So
000000@0503o00<000000540o`1A0?l0301A0?l01@000000503o01@0o`0D0?l00000008002So00H0
0000001Vo`00I_l00000002So`000002003Qo`0=00000000on400000003oX`0000000?mV00000000
obP0000001Co00000000DOl00000000208oo00030000003<o`0000000080ooD000@000000?nh003o
^00000001@3oNP000`000000ocd00?lm00040?lm00030000003oNP00ogX00080ogX000@000000?nh
003o^00000000P3om@000`000000c?l00000000208oo000F0000001Ao`00000000000000obP00000
003oIP0000000?nS00000000on400000003Qo`0000000:?o00000000I_l00000000Xo`000000503o
00000080D@3o00@0000008l0o`2?0?l000003`3<0?l01`000000S`3o08l0o`2?0?l000000540o`00
00000P0D0?l01P00000002So00000000I_l00000002So`80000001000?oQ00000000oj<00000003o
IP0000000?lX0000000Do`000000057o00000000Sol00000003<o`0000000P3om@0010000000okP0
0?nh000000050?mj00030000003o?@00ocd000<0ocd000<000000?mj003oNP000`3oNP0010000000
okP00?nh000000020?oe000A0000003<o`00c?l00000002?o`000000057o000000005?l000000000
obP00000003oIP0000000?nS00000000on400P0000001000hOl00000002So`000002001Vo`040000
0000:?l002So000000<0503o00<000000540o`1A0?l03@1A0?l01@000000503o01@0o`0D0?l00000
008002So00<00000001Vo`00I_l00P0000001@00Xol00000003Qo`00hOl0000000800?oQ00<00000
003oX`0000000P00ofH0100000000?lX003o:000000201Co00050000001Ao`00DOl0057o00000000
1@2?o`000`000000c?l00<co000;0<co00030000002?o`00Sol000<0Sol000@00000057o001Ao`00
DOl00P00000201Co008000000P00obP2000000L0oooo0@0000010?ooo`00803oool00`000000oooo
0?ooo`050?ooo`800?mV00D00000003o:000obP00?lX0000000301Co008000000`1Ao`000`000000
Sol008oo000308oo008000002`3<o`000`000000Sol008oo000308oo00050000001Ao`00DOl0057o
000000000P0Do`00100000000?lX003o:0000002003oIP0400000000oj<00?nS000000800?oQ00H0
0000003Qo`0000000000002So`000002001Vo`0400000000:?l002So000000@0503o00<000000540
o`1A0?l0301A0?l01@000000503o01@0o`0D0?l00000008002So00<00000001Vo`0000000P00Xol0
0`0000000>7o00000002003oh@8000000P00ofH2000000805?l000<00000057o000000000P2?o`02
00000080ooD000D000000?nh003o^000okP0000000050?mj00030000003o?@00ocd000<0ocd000<0
00000?mj003oNP000P3oNP0010000000okP00?nh000000020?oe00030000003<o`0000000080Sol0
01H00000057o000000000000003o:00000000?mV00000000oj<00000003oh@0000000>7o00000000
Xol00000001Vo`00000002So0000000D0?l000000P1A0?l010000000S`3o08l0o`00000?0<`0o`07
0000002?0?l0S`3o08l0o`000000D@3o0000000201@0o`0600000000:?l00000001Vo`0000000:?o
0P0000004000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`00
00000<co000000020?oe00040000003o^000okP0000000D0ogX000<000000?lm003o?@000P3o?@00
0`000000ogX00?mj00030?mj00040000003o^000okP000000080ooD0018000000<co003<o`000000
08oo00000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@000002003Q
o`0300000000Xol00000008006Ko00@00000000Xo`00:?l000000`0D0?l00`000000D@3o0540o`0=
0540o`050000000D0?l0503o01@0o`0000000P00:?l01000000006Ko001Vo`000002002So`040000
0000hOl00>7o000000800?oQ00<00000003oX`0000000P00ofH0100000000?lX003o:000000201Co
00050000001Ao`00DOl0057o000000001@2?o`000`000000c?l00<co000:0<co00800000102?o`02
00000080DOl00P00000301Co000500000000obP00?lX003o:00000000P00ofH70?ooo`4000000@3o
ool00200oooo00<000000?ooo`3oool01@3oool0100000000?mV003oIP000003003o:0030000000D
o`005?l000805?l000D00000057o001Ao`00DOl00000000608oo00030000003<o`00c?l000T0c?l0
00<0000008oo002?o`000`2?o`001@000000DOl0057o001Ao`00000000805?l00P0000001@00obP0
0000003oIP00ofH0000000800?nS00<00000003oh@0000000`00hOl00`0000000:?o00000002001V
o`0400000000:?l002So000000@0503o00<000000540o`1A0?l02`1A0?l01@000000503o01@0o`0D
0?l00000008002So00H00000001Vo`00I_l00000002So`000002003Qo`0900000000on400000003o
X`0000000?mV00000000obP0000000805?l00P00000208oo00030000003<o`0000000080ooD000@0
00000?nh003o^00000001P3oNP0200000080ocd000<000000?mj003oNP000`3oNP0010000000okP0
0?nh000000020?oe00030000003<o`0000000080Sol000@00000057o0000000000000P00obP20000
000<003oX`0000000?oQ00000000hOl00000002So`00000006Ko00000000:?l000000P0D0?l00`00
0000D@3o0000000208l0o`030000003<0?l0c03o00`0c03o00D0000008l0o`2?0?l0S`3o00000002
0540o`0:0000000D0?l00000000Xo`00000006Ko00000000Xol00000003Qo`80000000h00?nS0000
0000ofH00000003o:00000005?l00000001Ao`00000008oo00000000c?l000000080ooD000@00000
0?nh003o^00000001P3oNP000`000000ocd00?lm0002000000H0ogX000@000000?nh003o^0000000
0P3om@000`000000c?l00000000208oo000?0000001Ao`00000001Co000000000?lX00000000ofH0
0000003oX`0000000?oQ00000000hOl0000000800:?o00<00000001Vo`0000000P00:?l01@000000
503o01@0o`0D0?l0000000h0D@3o0P00000301@0o`0400000000:?l002So0000008006Ko00@00000
002So`00Xol000000P00hOl01P0000000?oQ003oh@0000000?nS000000800?mV00@00000003o:000
000000000P0Do`001@000000DOl0057o001Ao`00000000D0Sol000<000000<co003<o`002@3<o`02
000000D0Sol000D00000057o001Ao`00DOl00000000401Co000500000000obP00?lX003o:0000000
0`00ofH00`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3o
ool01000oj<0000000000000ofH2000000800?lX0P00000301Co00040000001Ao`00DOl0057o0080
0000102?o`000`000000c?l00<co000:0<co00800000102?o`001@000000DOl0057o001Ao`000000
00<05?l000<00000003o:00000000P00ofH01P0000000?nS003oX`0000000?oQ000000800>7o00@0
0000002So`00Xol000000P00I_l01000000002So000Xo`00000401@0o`030000001A0?l0D@3o00/0
D@3o00D0000001@0o`0D0?l0503o00000002000Xo`0800000000I_l006Ko00000000Xol00000003Q
o`000002003oh@8000000P00ofH2000000805?l000<00000057o000000000P2?o`0200000080ooD0
00D000000?nh003o^000okP0000000070?mj00<000001P3oNP0010000000okP00?nh000000020?oe
000@0000003<o`00000008oo00000000DOl00000000Do`000000003o:00000000?mV00000000oj<0
0000003oh@80000000H00:?o00000000I_l00000000Xo`00000201@0o`040000001A0?l0D@3o0000
0080S`3o00<000000<`0o`3<0?l02`3<0?l01`000000S`3o08l0o`2?0?l000000540o`0000000P0D
0?l02000000002So00000000I_l00000002So`0000000>7o0P0000003P00oj<00000003oIP000000
0?lX0000000Do`000000057o00000000Sol00000003<o`0000000P3om@0010000000okP00?nh0000
00060?mj00<000001`3oNP0010000000okP00?nh000000020?oe000B0000003<o`00c?l00000002?
o`000000057o000000005?l000000000obP00000003oIP0000000?nS00000000on4000000P00hOl0
0`0000000:?o00000002001Vo`0400000000:?l002So000000<0503o00<000000540o`1A0?l02`1A
0?l00`000000503o01@0o`0201@0o`0400000000:?l002So0000008006Ko00@00000002So`00Xol0
00000P00hOl01P0000000?oQ003oh@0000000?nS000000800?mV00<00000003o:00000000`0Do`00
1@000000DOl0057o001Ao`00000000@0Sol00P00000;0<co00030000002?o`00Sol000@0Sol000D0
0000057o001Ao`00DOl00000000301Co008000000P00obP2000000800?mV0P0000070?ooo`400000
0@3oool00200oooo00<000000?ooo`3oool01@3oool0100000000?nS003oX`000002003oIP800000
0P00obP2000000805?l00P000003057o00030000002?o`00Sol00080Sol000<000000<co003<o`00
2`3<o`000`000000Sol008oo000208oo00050000001Ao`00DOl0057o000000000P0Do`000`000000
0?lX003o:00200000006003oIP0000000?nS003oX`0000000?oQ0P0000001@00hOl00000002So`00
Xol00000008006Ko00@00000000Xo`00:?l000000`0D0?l00`000000D@3o0540o`0<0540o`050000
000D0?l0503o01@0o`0000000P00:?l04000000006Ko001Vo`0000000:?o00000000hOl00000003o
h@0000000?nS00000000ofH00000003o:000000201Co00060000001Ao`00000008oo00000000c?l0
0P0000000`3om@0000000?nh00020?nh00030000003oNP00ogX000d0ogX000@000000?nh003o^000
00000P3om@0040000000c?l00000002?o`000000057o000000005?l000000000obP00000003oIP00
00000?nS00000000on4200000006002So`00000006Ko00000000:?l000000P0D0?l010000000D@3o
0540o`00000208l0o`030000003<0?l0c03o00/0c03o00L0000008l0o`2?0?l0S`3o0000001A0?l0
00000080503o00P00000000Xo`00000006Ko00000000Xol00000003Qo`80000000h00?nS00000000
ofH00000003o:00000005?l00000001Ao`00000008oo00000000c?l000000080ooD000@000000?nh
003o^0000000403oNP0010000000okP00?nh000000020?oe00030000003<o`0000000080Sol000l0
0000057o000000005?l000000000obP00000003oIP0000000?nS00000000on400000003Qo`000000
0P00Xol00`00000006Ko00000002000Xo`050000000D0?l0503o01@0o`0000003P1A0?l01@000000
503o01@0o`0D0?l0000000<002So00H00000001Vo`00I_l00000002So`000002003Qo`0700000000
on400?oQ00000000oj<00000003oIP02000000800?lX00@0000001Co000Do`0000000`1Ao`000`00
0000Sol008oo000208oo00030000003<o`00c?l000/0c?l000<0000008oo002?o`000P2?o`020000
0080DOl00P00000201Co008000000P00obP2000000800?mV0P000002003oX`030000003oool0oooo
00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`05003oh@0000000?nS003o
X`0000000`00ofH01@0000000?lX003o:000obP0000000<05?l000D00000057o001Ao`00DOl00000
000408oo00800000303<o`000`000000Sol008oo000208oo008000000P1Ao`020000000301Co0000
00000?lX00800?lX00H00000003oIP0000000000003oX`000002003oh@0300000000hOl000000080
0:?o00@00000001Vo`00I_l000000P00:?l01@000000503o01@0o`0D0?l0000000d0D@3o00D00000
01@0o`0D0?l0503o00000002000Xo`0600000000I_l006Ko00000000Xol000000P00hOl02`000000
0?oQ00000000oj<00000003oIP0000000?lX0000000Do`0000000080DOl00P0000020<co00030000
003om@00000000<0okP000<000000?mj003oNP003@3oNP0010000000okP00?nh000000020?oe000@
0000003<o`00000008oo00000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000
003oh@80000000H00:?o00000000I_l00000000Xo`00000201@0o`040000001A0?l0D@3o00000080
S`3o00<000000<`0o`3<0?l02`3<0?l01`000000S`3o08l0o`2?0?l000000540o`0000000P0D0?l0
1000000002So00000000I_l2000000800>7o0P0000003P00oj<00000003oIP0000000?lX0000000D
o`000000057o00000000Sol00000003<o`0000000P3om@0010000000okP00?nh0000000?0?mj0004
0000003o^000okP000000080ooD000<000000<co000000000P2?o`003`000000DOl00000000Do`00
0000003o:00000000?mV00000000oj<00000003oh@0000000>7o00000002002So`0300000000I_l0
0000008002So00<0000001@0o`0D0?l00P0D0?l00`000000D@3o0540o`0;0540o`050000000D0?l0
503o01@0o`0000000`00:?l01P00000006Ko001Vo`0000000:?o000000800>7o00P00000003oh@00
on400000003oX`0000000?mV000000<00?lX00@0000001Co0000000000000P1Ao`02000000@0Sol0
00<000000<co003<o`002`3<o`000`000000Sol008oo000208oo00050000001Ao`00DOl0057o0000
00000`0Do`001@0000000?lX003o:000obP0000000<00?mV00D00000003oX`00oj<00?nS00000007
0?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool2003oh@0300000000oj<00?nS
008000000P00ofH2000000800?lX00D0000001Co000Do`005?l000000003057o00800000102?o`00
0`000000c?l00<co000:0<co00800000102?o`001P000000DOl0057o001Ao`00000001Co00800000
00<00?lX00000000ofH00P00ofH01@0000000?nS00000000on40000000800>7o00<00000002So`00
00000P00I_l01000000002So000Xo`00000401@0o`800000301A0?l01@000000503o01@0o`0D0?l0
0000008002So00X00000001Vo`00I_l00000002So`0000000>7o00000000on4000000P00oj<20000
00800?lX00T0000001Co00000000DOl00000002?o`0000000<co000000000P3om@001@000000okP0
0?nh003o^000000000h0ogX000@000000?nh003o^00000000P3om@004P000000c?l00000002?o`00
0000057o000000005?l000000000obP00000003oIP0000000?nS00000000on400000003Qo`800000
0P00I_l200000080503o00@000000540o`1A0?l000000`2?0?l00`000000c03o0<`0o`090<`0o`05
0000002?0?l0S`3o08l0o`0000000P1A0?l00`000000503o00000002000Xo`80000000L00:?o0000
0000hOl00000003oh@0000000?nS0080000000X00?lX0000000Do`000000057o00000000Sol00000
003<o`0000000P3om@001@000000okP00?nh003o^000000000h0ogX000D000000?nh003o^000okP0
000000020?oe00050000003<o`00000008oo000000000P1Ao`00100000005?l000000000obP20000
00800?nS00L00000003oh@0000000>7o00000000Xol00000008006Ko00@00000000Xo`00:?l00000
0`0D0?l00`000000D@3o0540o`0:0540o`8000000`0D0?l01@00000002So000Xo`00:?l000000080
06Ko00<00000002So`0000000P00hOl0200000000?oQ003oh@0000000?nS00000000ofH000000P00
obP20000000301Co00000000DOl00080DOl000<0000008oo002?o`000P2?o`02000000d0c?l000@0
000008oo002?o`00Sol00P000003057o00050000000Do`005?l001Co000000000P00obP200000080
0?mV0P000002003oX`80000000<00?oQ0?ooo`3oool01@3oool100000040oooo000P0?ooo`030000
003oool0oooo00D0oooo00@00000003oh@00on4000000`00oj<01@0000000?mV003oIP00ofH00000
00800?lX00D0000001Co000Do`005?l000000004057o00030000002?o`00Sol00080Sol000<00000
0<co003<o`002`3<o`000`000000Sol008oo000208oo00040000001Ao`00DOl0000000<05?l000<0
0000003o:00000000P00ofH0200000000?nS003oX`0000000?oQ00000000hOl000000P00Xol01000
000006Ko001Vo`000002000Xo`030000000D0?l0503o00<0503o00<000000540o`1A0?l02@1A0?l0
1@000000503o01@0o`0D0?l00000008002So00H00000001Vo`00I_l00000002So`000002003Qo`0;
00000000on400000003oX`0000000?mV00000000obP0000001Co000000000P1Ao`0200000080c?l0
00<000000?oe000000000`3o^0000`000000ogX00?mj000<0?mj00050000003o^000okP00?nh0000
00000P3om@003P000000c?l00000002?o`000000057o000000005?l000000000obP00000003oIP00
00000?nS0P0000002000hOl00000002So`00000006Ko00000000:?l000000P0D0?l00`000000D@3o
0000000308l0o`030000003<0?l0c03o00T0c03o00D0000008l0o`2?0?l0S`3o000000020540o`0>
0000000D0?l00000000Xo`00000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<20000
000:003o:00000005?l00000001Ao`00000008oo00000000c?l000000080ooD000D000000?nh003o
^000okP00000000>0?mj00040000003o^000okP000000080ooD000<000000<co000000000P2?o`00
3`000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o0000
0002002So`0300000000I_l00000008002So00<0000001@0o`0D0?l00P0D0?l00`000000D@3o0540
o`090540o`030000000D0?l0503o00<0503o00@00000000Xo`00:?l000000P00I_l01P0000000:?o
002So`0000000>7o000000800?oQ00<00000003oX`0000000P00ofH00`0000000?lX0000000201Co
00050000001Ao`00DOl0057o00000000102?o`000`000000c?l00<co000<0<co00050000002?o`00
Sol008oo00000000101Ao`001@0000005?l001Co000Do`00000000800?lX00D00000003oIP00ofH0
0?mV00000003003oX`0500000000on400?oQ003oh@0000001P3oool100000040oooo000P0?ooo`03
0000003oool0oooo00D0oooo00D00000003oh@00on400?oQ00000003003oX`0500000000ofH00?mV
003oIP0000000P00obP01@0000005?l001Co000Do`00000000<0DOl000<0000008oo002?o`000`2?
o`000`000000c?l00<co000:0<co00030000002?o`00Sol000<0Sol000@00000057o001Ao`000000
0P0Do`000`0000000?lX003o:00200000004003oIP0000000?nS000000800?oQ00<00000003Qo`00
00000P00Xol01000000006Ko001Vo`000002000Xo`8000000`0D0?l00`000000D@3o0540o`0:0540
o`050000000D0?l0503o01@0o`0000000P00:?l01P00000006Ko001Vo`0000000:?o000000800>7o
00l00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`00000008oo0000
00000P3<o`000`000000ooD0000000030?nh00030000003oNP00ogX000/0ogX00P0000020?nh000A
0000003om@00ooD00000003<o`00000008oo00000000DOl00000000Do`000000003o:00000000?mV
00000000oj<00P0000002000hOl00000002So`00000006Ko00000000:?l000000P0D0?l00`000000
D@3o0000000308l0o`030000003<0?l0c03o00T0c03o00D0000008l0o`2?0?l0S`3o000000020540
o`0>0000000D0?l00000000Xo`00000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<2
0000000:003o:00000005?l00000001Ao`00000008oo00000000c?l000000080ooD000D000000?nh
003o^000okP00000000>0?mj00040000003o^000okP000000080ooD000D000000<co00000000Sol0
00000002057o000?0000000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o
00000000Xol00000008006Ko00@00000000Xo`00:?l000000`0D0?l00`000000D@3o0540o`0:0540
o`040000000D0?l0503o01@0o`8000000P00:?l01000000006Ko001Vo`000002002So`0300000000
hOl0000000800?oQ00<00000003oX`0000000P00ofH0100000000?lX003o:000000201Co00050000
001Ao`00DOl0057o00000000102?o`000`000000c?l00<co000;0<co00030000002?o`00Sol00080
Sol000D00000057o001Ao`00DOl00000000301Co000500000000obP00?lX003o:00000000P00ofH0
1@0000000?nS003oX`00oj<0000000<00?oQ00<000000?ooo`3oool01@3oool100000040oooo000P
0?ooo`030000003oool0oooo00D0oooo00<00>7o0000000000000P00on42000000800?nS0P000002
003oIP80000000<00?lX0000000Do`000P0Do`001@000000DOl0057o001Ao`00000000D0Sol000<0
00000<co003<o`002P3<o`000`000000Sol008oo000308oo00030000001Ao`00DOl00080000000<0
5?l000000000obP00P00obP00`0000000?mV00000002000000800?oQ00<00000003Qo`0000000P00
Xol01000000006Ko001Vo`000003000Xo`050000000D0?l0503o01@0o`000000301A0?l01@000000
503o01@0o`0D0?l00000008002So00H00000001Vo`00I_l00000002So`000002003Qo`8000000P00
oj<03@0000000?mV00000000obP0000001Co00000000DOl00000002?o`0000000<co000000000P3o
m@001@000000okP00?nh003o^000000000d0ogX000@000000?nh003o^00000000P3om@003P000000
c?l00000002?o`000000057o000000005?l000000000obP00000003oIP0000000?nS0P0000002000
hOl00000002So`00000006Ko00000000:?l000000P0D0?l00`000000D@3o0000000408l0o`030000
003<0?l0c03o00P0c03o00D0000008l0o`2?0?l0S`3o000000020540o`0>0000000D0?l00000000X
o`00000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<20000000:003o:00000005?l0
0000001Ao`00000008oo00000000c?l000000080ooD000D000000?nh003o^000okP00000000=0?mj
00040000003o^000okP000000080ooD000D000000<co00000000Sol000000002057o000?0000000D
o`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o00000000Xol00000008006Ko
00@00000000Xo`00:?l000000`0D0?l00`000000D@3o0540o`0:0540o`050000000D0?l0503o01@0
o`0000000`00:?l01000000006Ko001Vo`000002002So`0300000000hOl0000000800?oQ00<00000
003oX`0000000P00ofH01@0000000?lX003o:00000005?l0008000000P1Ao`02000000@0Sol000<0
00000<co003<o`002`3<o`000`000000Sol008oo000208oo00050000001Ao`00DOl0057o00000000
0`0Do`000`0000000?lX003o:00200000003003oIP000000000000800?nS0P000002003oh@800000
00<00>7o0?ooo`3oool01@3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo0`00
hOl01@0000000?oQ003oh@00on40000000<00?nS00H00000003oIP00ofH00?mV00000000obP20000
00805?l00P000002057o00800000102?o`000`000000c?l00<co000:0<co00800000102?o`001P00
0000DOl0057o001Ao`00000001Co008000000P00obP01@0000000?mV00000000oj<0000000800?oQ
00<00000003Qo`0000000P00Xol01000000006Ko001Vo`000003000Xo`050000000D0?l0503o01@0
o`0000002`1A0?l01@000000503o01@0o`0D0?l0000000<002So00<00000001Vo`0000000P00Xol0
4@0000000>7o00000000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl00000
002?o`0000000080c?l000@000000?oe003om@0000000P3o^0000`000000ogX00?mj000;0?mj0005
0000003o^000okP00?nh000000000P3om@002P000000c?l00000002?o`000000057o000000005?l0
00000000obP20000000>003oX`0000000?oQ00000000hOl00000002So`00000006Ko00000000:?l0
000001@0o`0000020540o`050000002?0?l0S`3o08l0o`0000002@3<0?l01@000000S`3o08l0o`2?
0?l000000080D@3o00<0000001@0o`0000000P00:?l02`00000006Ko0000000000000>7o00000000
on400000003oX`0000000?mV0080000000H05?l00000001Ao`00000008oo000000020<co00040000
003om@00ooD000000080okP000<000000?mj003oNP002`3oNP001@000000okP00?nh003o^0000000
0080ooD000l000000<co00000000Sol00000001Ao`00000001Co000000000?lX00000000ofH00000
003oX`0000000P00on42000000800:?o00@00000001Vo`00I_l000000P00:?l01@000000503o01@0
o`0D0?l0000000/0D@3o00<0000001@0o`0D0?l00P0D0?l01000000002So000Xo`000002001Vo`06
00000000Xol00:?o00000000hOl000000P00on40100000000?nS00000000ofH2000000800?lX00<0
000001Co000000000`1Ao`000`000000Sol008oo000208oo00800000303<o`02000000@0Sol000<0
0000057o001Ao`000P00000201Co0080000000<00?lX0000000000000P00ofH01@0000000?nS003o
X`00oj<0000000<00?oQ00@00000003Qo`00hOl00>7o1`3oool100000040oooo000P0?ooo`030000
003oool0oooo00D0oooo00@00000003Qo`00hOl00>7o0P000002003oh@8000000P00oj<200000080
0?mV00<00000003o:000obP00P00000201Co00040000001Ao`00DOl0057o008000000`2?o`020000
00`0c?l00P00000308oo008000000P1Ao`00100000005?l001Co00000002003o:00400000000ofH0
0000003oX`80000000<00?oQ00000000hOl00P0000001@00Xol00000001Vo`00I_l0000000<002So
00D0000001@0o`0D0?l0503o0000000;0540o`040000000D0?l0503o01@0o`8000000P00:?l00`00
000006Ko00000002002So`0B00000000hOl00000003oh@0000000?nS00000000ofH00000003o:000
00005?l00000001Ao`00000008oo00000000c?l00P0000020?oe00040000003o^000okP0000000`0
ogX000D000000?nh003o^000okP0000000020?oe000:0000003<o`00000008oo00000000DOl00000
000Do`000000003o:080000000h00?nS00000000on400000003Qo`0000000:?o00000000I_l00000
000Xo`000000503o00000080D@3o00<0000008l0o`2?0?l00P2?0?l00`000000c03o0<`0o`060<`0
o`050000002?0?l0S`3o08l0o`0000000P1A0?l00`000000503o00000002000Xo`0;00000000I_l0
000000000000hOl00000003oh@0000000?nS00000000ofH00P0000001P0Do`000000057o00000000
Sol000000080c?l000@000000?oe003om@0000000P3o^0000`000000ogX00?mj000:0?mj00050000
003o^000okP00?nh000000000P3om@003`000000c?l00000002?o`000000057o000000005?l00000
0000obP00000003oIP0000000?nS00000002003oh@0500000000hOl00000002So`0000000P00I_l0
1000000002So000Xo`00000301@0o`030000001A0?l0D@3o00T0D@3o00<0000001@0o`0D0?l00P0D
0?l01000000002So000Xo`000002001Vo`0700000000Xol0000000000000hOl00000003oh@020000
0004003oX`0000000?mV000000800?lX0P0000000`0Do`000000057o0002057o00030000002?o`00
Sol00080Sol000<000000<co003<o`002P3<o`02000000@0Sol00P000002057o00030000000Do`00
5?l0008000000P00obP00`0000000?mV003oIP02000000800?nS0P000002003oh@8000000`00hOl2
000000H0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`03002So`0000000>7o
00<00>7o00D00000003oh@00on400?oQ00000003003oX`0400000000ofH00?mV000000<00?lX00@0
000001Co000Do`000000101Ao`001@000000Sol008oo002?o`00000000h0c?l000<0000008oo002?
o`000P2?o`0010000000DOl0057o0000000201Co000600000000obP00?lX00000000ofH000000P00
oj<00`0000000?oQ00000002003Qo`0300000000Xol00000008006Ko00@00000000Xo`00:?l00000
100D0?l00`000000D@3o0540o`080540o`030000000D0?l0503o0080503o00@00000000Xo`00:?l0
00000P00I_l04`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH00000003o:000
00005?l00000001Ao`00000008oo000000000P3<o`0010000000ooD00?oe000000030?nh00030000
003oNP00ogX000T0ogX000D000000?nh003o^000okP0000000020?oe000:0000003<o`00000008oo
00000000DOl00000000Do`000000003o:080000000h00?nS00000000on400000003Qo`0000000:?o
00000000I_l00000000Xo`000000503o00000080D@3o00<0000008l0o`2?0?l00P2?0?l00`000000
c03o0<`0o`050<`0o`030000002?0?l0S`3o0080S`3o00H000000540o`1A0?l0000001@0o`000002
000Xo`0;00000000I_l0000000000000hOl00000003oh@0000000?nS00000000ofH00P0000001P0D
o`000000057o00000000Sol000000080c?l000@000000?oe003om@0000000P3o^0000`000000ogX0
0?mj00090?mj00050000003o^000okP00?nh000000000P3om@000`000000c?l00000000208oo0080
00000P0Do`02000000800?mV0P000002003oh@0500000000hOl00000002So`0000000P00I_l01000
000002So000Xo`00000401@0o`030000001A0?l0D@3o00T0D@3o00D0000001@0o`0D0?l0503o0000
0002000Xo`0600000000I_l006Ko00000000Xol000000P00hOl00`0000000?oQ00000002003oX`06
00000000ofH00?mV00000000obP000000P0Do`001@000000DOl0057o001Ao`00000000@0Sol000<0
00000<co003<o`002P3<o`000`000000Sol008oo000308oo00050000001Ao`00DOl0057o00000000
0P0Do`001@0000000?lX003o:000obP0000000800?mV00D00000003oX`00oj<00?nS00000003003o
h@0300000000hOl00>7o00800>7o00<00000002So`3oool01P3oool100000040oooo000P0?ooo`03
0000003oool0oooo00D0oooo0P00Xol00`0000000>7o003Qo`02003Qo`0500000000on400?oQ003o
h@0000000`00oj<0100000000?mV003oIP000003003o:0030000000Do`005?l0008000000`1Ao`00
1@000000Sol008oo002?o`00000000d0c?l000<0000008oo002?o`000P2?o`001@000000DOl0057o
001Ao`00000000805?l000<00000003o:000obP00P000002003oX`0600000000on400?oQ00000000
hOl000000P00Xol01000000006Ko001Vo`000002000Xo`030000000D0?l0503o0080503o00<00000
0540o`1A0?l0201A0?l00`000000503o01@0o`0201@0o`0600000000:?l002So00000000I_l00000
0P00Xol04@0000000>7o00000000on400000003oX`0000000?mV00000000obP0000001Co00000000
DOl00000002?o`0000000080c?l000@000000?oe003om@0000000`3o^0000`000000ogX00?mj0008
0?mj00050000003o^000okP00?nh000000000P3om@0030000000c?l00000002?o`000000057o0000
00005?l000000000obP00000003oIP80000000`00?oQ00000000hOl00000002So`00000006Ko0000
0000:?l0000001@0o`0000020540o`030000002?0?l0S`3o00<0S`3o00<000000<`0o`3<0?l0103<
0?l01@000000S`3o08l0o`2?0?l000000080D@3o00/0000001@0o`0D0?l00000000Xo`00000006Ko
00000000Xol00000003Qo`020000000>003oX`0000000?mV00000000obP0000001Co00000000DOl0
0000002?o`0000000<co000000020?oe00050000003o^000okP00?nh000000002P3oNP000`000000
okP00?nh00020?nh000:0000003om@00ooD00000003<o`00000008oo00000000DOl0000000805?l0
0P000002003oIP80000000T00?oQ00000000hOl00>7o00000000Xol00000001Vo`0000000`00:?l0
1@000000503o01@0o`0D0?l0000000/0D@3o00D0000001@0o`0D0?l0503o00000002000Xo`040000
0000I_l006Ko000000800:?o00D00000003Qo`0000000?oQ00000003003oX`0500000000ofH00000
003o:00000000`0Do`0010000000DOl0057o0000000408oo00030000003<o`00c?l000/0c?l000<0
000008oo002?o`000P2?o`001@000000DOl0057o001Ao`00000000805?l000D00000003o:000obP0
0?lX00000002003oIP0500000000oj<00?nS003oX`0000000`00on400`0000000>7o003Qo`02003Q
o`0300000000Xol00:?o00L0oooo0@0000010?ooo`00803oool2000000H0oooo0`00Xol2000000<0
0>7o0P000002003oh@8000000P00oj<200000003003oIP0000000?lX00800?lX00D0000001Co000D
o`005?l000000003057o00040000002?o`00Sol008oo00800000303<o`000`000000Sol008oo0002
08oo00050000001Ao`00DOl0057o000000000P0Do`001@0000000?lX00000000ofH0000000800?nS
00H00000003oh@00on400000003Qo`000002002So`0400000000I_l006Ko0000008002So00D00000
01@0o`0D0?l0503o0000000:0540o`030000000D0?l0503o0080503o00@00000000Xo`00:?l00000
0P00I_l04`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH00000003o:0000000
5?l00000001Ao`00000008oo000000000P3<o`0010000000ooD00?oe000000030?nh00030000003o
NP00ogX000P0ogX000D000000?nh003o^000okP0000000020?oe00030000003<o`00c?l000800000
0P1Ao`0200000009003o:00000000?mV00000000oj<00000003oh@0000000>7o0080000000L006Ko
00000000:?l002So0000000D0?l000000080D@3o00<0000008l0o`2?0?l00P2?0?l00`000000c03o
0<`0o`030<`0o`030000002?0?l0S`3o0080S`3o00@000000540o`1A0?l000000P0D0?l020000000
02So00000000I_l00000002So`0000000>7o0P0000003P00oj<00000003oIP0000000?lX0000000D
o`000000057o00000000Sol00000003<o`0000000P3om@001@000000okP00?nh003o^000000000X0
ogX000D000000?nh003o^000okP0000000020?oe00030000003<o`0000000080Sol00P00000201Co
008000000P00ofH200000007003oh@0000000>7o003Qo`0000000:?o00000002001Vo`0400000000
:?l002So000000@0503o00<000000540o`1A0?l0201A0?l01@000000503o01@0o`0D0?l0000000<0
02So00H00000001Vo`00I_l00000002So`000002003Qo`0300000000on40000000800?nS00<00000
003oIP0000000P00obP0100000005?l001Co00000002057o00030000002?o`00Sol00080Sol000<0
00000<co003<o`002P3<o`02000000@0Sol000D00000057o001Ao`00DOl00000000201Co00050000
0000obP00?lX003o:00000000P00ofH00`0000000?nS003oX`02000000800?oQ0P000004003Qo`04
00000000Xol00:?o002So`H0oooo0P0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`D0
0:?o0P000003003Qo`0500000000on400?oQ003oh@0000000`00oj<0100000000?mV000000000002
003o:0050000000Do`005?l001Co000000000`1Ao`000`000000Sol008oo000208oo00030000003<
o`00c?l000X0c?l000<0000008oo002?o`000P2?o`001@000000DOl0057o001Ao`00000000805?l0
00@00000003o:00000000?mV0P0000001`00oj<00000003oh@00on400000003Qo`0000000P00Xol0
1000000006Ko001Vo`000002000Xo`050000000D0?l0503o01@0o`0000002P1A0?l00`000000503o
01@0o`0201@0o`0600000000:?l002So00000000I_l000000P00Xol04@0000000>7o00000000on40
0000003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`0000000080c?l000@0
00000?oe003om@0000000`3o^0000`000000ogX00?mj00070?mj00050000003o^000okP00?nh0000
00000P3om@002@000000c?l00<co00000000Sol00000001Ao`00000001Co000200000007003oIP00
00000?nS00000000on400000003Qo`0200000007001Vo`00000002So000Xo`000000503o00000002
0540o`030000002?0?l0S`3o00<0S`3o00<000000<`0o`3<0?l00P3<0?l00`000000S`3o08l0o`02
08l0o`040000001A0?l0D@3o00000080503o00P00000000Xo`00000006Ko00000000Xol00000003Q
o`80000000h00?nS00000000ofH00000003o:00000005?l00000001Ao`00000008oo00000000c?l0
00000080ooD000D000000?nh003o^000okP0000000090?mj00030000003o^000okP00080okP001@0
00000?oe003om@0000000<co00000000Sol00000001Ao`00000001Co000000000?lX00000000ofH0
0000003oX`0000000?oQ000000800>7o00D00000002So`00000006Ko00000003000Xo`050000000D
0?l0503o01@0o`0000002P1A0?l01@000000503o01@0o`0D0?l0000000<002So00H00000001Vo`00
I_l00000002So`000002003Qo`0400000000on400000003oX`80000000D00?mV00000000obP00?lX
0000000201Co00040000001Ao`00DOl0000000@0Sol000<000000<co003<o`002P3<o`000`000000
Sol008oo000308oo00030000001Ao`00DOl0008000000P0Do`000`0000000?lX003o:00200000080
0?mV00@00000003oX`00oj<000000`00on400`0000000>7o003Qo`02003Qo`8000001000Xol70?oo
o`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool6002So`0300000000hOl00>7o0080
0>7o00D00000003oh@00on400?oQ00000003003oX`0400000000ofH00?mV000000800?lX0P000002
01Co00050000001Ao`00DOl0057o000000000`2?o`000`000000c?l00<co000;0<co00030000002?
o`00Sol00080Sol000D00000057o001Ao`00DOl00000000201Co008000000`00ofH01@0000000?nS
00000000on40000000800>7o00<00000002So`0000000P00I_l01@00000002So000Xo`00:?l00000
00<0503o0P0000080540o`030000000D0?l0503o0080503o00H00000000Xo`00:?l00000001Vo`00
0002002So`0A00000000hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000
001Ao`00000008oo000000000P3<o`0010000000ooD00?oe000000040?nh00030000003oNP00ogX0
00H0ogX000D000000?nh003o^000okP0000000020?oe00090000003<o`00c?l00000002?o`000000
057o000000005?l00080000000L00?mV00000000oj<00000003oh@0000000>7o0080000000L006Ko
00000000:?l002So0000000D0?l000000080D@3o00<0000008l0o`2?0?l00`2?0?l5000000D0S`3o
00@000000540o`1A0?l000000P0D0?l01P00000002So00000000I_l00000002So`80000001000?oQ
00000000oj<00000003oIP0000000?lX0000000Do`000000057o00000000Sol00000003<o`000000
0P3om@001@000000okP00?nh003o^000000000P0ogX000<000000?nh003o^0000P3o^00020000000
ooD00?oe00000000c?l00000002?o`0000000P1Ao`00100000005?l000000000obP2000000800?nS
0P000002003Qo`0500000000Xol00000001Vo`0000000`00:?l01@000000503o01@0o`0D0?l00000
00X0D@3o00<0000001@0o`0D0?l00P0D0?l01000000002So000Xo`000002001Vo`0300000000Xol0
000000800>7o00D00000003oh@0000000?nS00000002003oIP0400000000obP00?lX000000805?l0
00@00000057o001Ao`000000102?o`000`000000c?l00<co000:0<co00030000002?o`00Sol000<0
Sol000@00000057o001Ao`0000000`0Do`00100000000?lX003o:0000002003oIP8000000P00oj<0
1@0000000?oQ003oh@00on40000000@00>7o00<00000002So`00Xol01000Xol70?ooo`4000000@3o
ool00200oooo00<000000?ooo`3oool01@3oool7002So`0300000000hOl00>7o00800>7o00D00000
003oh@00on400?oQ00000003003oX`0400000000ofH00?mV000000<00?lX00@0000001Co000Do`00
00000P1Ao`000`000000Sol008oo000208oo00030000003<o`00c?l000/0c?l000<0000008oo002?
o`000P2?o`002@000000DOl0057o001Ao`00000001Co000000000?lX00000002003oIP0300000000
oj<0000000800?oQ00<00000003Qo`0000000P00Xol01000000006Ko001Vo`000002000Xo`030000
000D0?l0503o00<0503o00<000000540o`1A0?l01P1A0?l00`000000503o01@0o`0201@0o`060000
0000:?l002So00000000I_l000000P00Xol04@0000000>7o00000000on400000003oX`0000000?mV
00000000obP0000001Co00000000DOl00000002?o`0000000080c?l000@000000?oe003om@000000
0`3o^0000`000000ogX00?mj00060?mj00050000003o^000okP00?nh000000000P3om@002@000000
c?l00<co00000000Sol00000001Ao`00000001Co000200000009003oIP0000000?nS00000000on40
0000003Qo`0000000:?o008000000P00:?l00`000000503o000000020540o`030000002?0?l0S`3o
00`0S`3o00@000000540o`1A0?l000000P0D0?l00`00000002So00000002001Vo`80000000<00>7o
00000000on400P0000002P00ofH00000003o:00000005?l00000001Ao`00000008oo000000020<co
00030000003om@00000000@0okP000<000000?mj003oNP001P3oNP000`000000okP00?nh00020?nh
00080000003om@00ooD00000003<o`00000008oo00000002057o008000000P00obP2000000800?nS
0P0000001`00hOl00000002So`00Xol00000001Vo`0000000`00:?l01@000000503o01@0o`0D0?l0
000000T0D@3o00<0000001@0o`0D0?l00P0D0?l01000000002So000Xo`000002001Vo`0300000000
Xol0000000800>7o00<00000003oh@0000000P00oj<01P0000000?mV003oIP0000000?lX00000080
5?l000D00000057o001Ao`00DOl00000000308oo00030000003<o`00c?l000/0c?l000<0000008oo
002?o`000P2?o`001@000000DOl0057o001Ao`00000000805?l000@00000003o:000obP000000P00
ofH01@0000000?nS003oX`00oj<0000000<00?oQ00<00000003Qo`00hOl00P00hOl00`0000000:?o
002So`05002So`L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0300000000
Xol00:?o00D00:?o0P000003003Qo`8000000P00on42000000800?nS00@00000003oIP00ofH00000
0`00obP0100000005?l001Co00000002057o00030000002?o`00Sol00080Sol00P00000<0<co0003
0000002?o`00Sol00080Sol000P00000057o001Ao`00DOl00000000Do`000000003o:080000000@0
0?mV00000000oj<000000P00on400`0000000>7o00000002002So`0400000000I_l006Ko00000080
02So00<0000001@0o`0D0?l00P0D0?l00`000000D@3o0540o`060540o`030000000D0?l0503o0080
503o00H00000000Xo`00:?l00000001Vo`000002002So`0A00000000hOl00000003oh@0000000?nS
00000000ofH00000003o:00000005?l00000001Ao`00000008oo000000000P3<o`0010000000ooD0
0?oe000000040?nh00030000003oNP00ogX000D0ogX000<000000?nh003o^0000P3o^0000`000000
ooD0000000020<co000<0000002?o`000000057o000000005?l000000000obP00000003oIP000000
0?nS0P0000002000hOl00000002So`00000006Ko00000000:?l000000P0D0?l010000000D@3o0540
o`00000=08l0o`040000001A0?l0D@3o00000080503o00X00000000Xo`00000006Ko00000000Xol0
0000003Qo`0000000?oQ0P0000002P00ofH00000003o:00000005?l00000001Ao`00000008oo0000
00020<co00030000003om@00000000@0okP000<000000?mj003oNP001@3oNP000`000000okP00?nh
00020?nh00080000003om@00ooD00000003<o`00000008oo00000002057o00040000000Do`000000
003o:08000000P00oj<2000000800>7o00D00000002So`00000006Ko00000003000Xo`050000000D
0?l0503o01@0o`0000002@1A0?l00`000000503o01@0o`0301@0o`0300000000:?l00000008006Ko
00H00000002So`00Xol00000003Qo`000002003oh@0300000000oj<0000000800?mV00<00000003o
:00000000P0Do`001@000000DOl0057o001Ao`00000000<0Sol000<000000<co003<o`002P3<o`02
000000@0Sol000D00000057o001Ao`00DOl00000000201Co000400000000obP00?lX000000800?mV
00<00000003oX`00oj<00P000002003oh@8000000`00hOl2000000P00:?o00<000000?ooo`3oool0
103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00<006Ko00000000Xol01`00
Xol2000000<00>7o00D00000003oh@00on400?oQ00000002003oX`0400000000ofH00?mV000000<0
0?lX00@0000001Co000Do`0000000P1Ao`02000000@0Sol000<000000<co003<o`002P3<o`000`00
0000Sol008oo000208oo00040000001Ao`00DOl0000000805?l000P00000003o:000obP00000003o
IP0000000?nS000000800?oQ00<00000003Qo`0000000P00Xol01000000006Ko001Vo`000002000X
o`030000000D0?l0503o0080503o00<000000540o`1A0?l01P1A0?l00`000000503o01@0o`0201@0
o`0600000000:?l002So00000000I_l000000P00Xol04@0000000>7o00000000on400000003oX`00
00000?mV00000000obP0000001Co00000000DOl00000002?o`0000000080c?l000@000000?oe003o
m@0000000`3o^0000`000000ogX00?mj00050?mj00030000003o^000okP00080okP0014000000?oe
003om@0000000<co00000000Sol00000001Ao`00000001Co000000000?lX00000000ofH00000003o
X`0200000008003Qo`0000000:?o00000000I_l00000000Xo`00000201@0o`040000001A0?l0D@3o
000000d0S`3o00@000000540o`1A0?l000000P0D0?l02P00000002So00000000I_l00000002So`00
00000>7o00000000on420000000:003oIP0000000?lX0000000Do`000000057o00000000Sol00000
0080c?l000<000000?oe00000000103o^0000`000000ogX00?mj00050?mj00030000003o^000okP0
0080okP000P000000?oe003om@0000000<co00000000Sol000000080DOl000@0000001Co00000000
0?lX0P0000001000oj<00000003oh@000002003Qo`0500000000Xol00000001Vo`0000000`00:?l0
1@000000503o01@0o`0D0?l0000000P0D@3o00<0000001@0o`0D0?l00P0D0?l20000008002So00<0
0000001Vo`0000000P00Xol00`0000000>7o00000002003oh@0300000000oj<0000000800?mV00<0
0000003o:00000000P0Do`001@000000DOl0057o001Ao`00000000<0Sol000<000000<co003<o`00
2P3<o`000`000000Sol008oo000208oo008000000P1Ao`02000000805?l000@00000003o:000obP0
00000P00ofH0100000000?nS003oX`000003003oh@0300000000hOl00>7o00800>7o00<00000002S
o`00Xol01P00Xol2000000L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`80
06Ko0P000007002So`0300000000hOl00>7o00800>7o00D00000003oh@00on400?oQ00000002003o
X`0300000000ofH00?mV008000000P00obP0100000005?l001Co00000003057o00030000002?o`00
Sol00080Sol000<000000<co003<o`002P3<o`000`000000Sol008oo000208oo00060000001Ao`00
DOl00000000Do`0000000P00obP01P0000000?mV003oIP0000000?nS000000800?oQ00D00000003Q
o`0000000:?o00000003001Vo`0400000000:?l002So000000<0503o00<000000540o`1A0?l01P1A
0?l00`000000503o01@0o`0201@0o`0400000000:?l002So0000008006Ko01<00000002So`000000
0>7o00000000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`00
00000080c?l000@000000?oe003om@000000103o^0000`000000ogX00?mj00040?mj00030000003o
^000okP00080okP0014000000?oe003om@0000000<co00000000Sol00000001Ao`00000001Co0000
00000?lX00000000ofH00000003oX`0200000008003Qo`0000000:?o00000000I_l00000000Xo`00
000201@0o`040000001A0?l0D@3o000000d0S`3o00@000000540o`1A0?l000000P0D0?l02P000000
02So00000000I_l00000002So`0000000>7o00000000on420000000:003oIP0000000?lX0000000D
o`000000057o00000000Sol000000080c?l000<000000?oe00000000103o^0000`000000ogX00?mj
00040?mj00030000003o^000okP00080okP000H000000?oe003om@0000000<co0000000208oo000A
0000001Ao`00000001Co000000000?lX00000000ofH00000003oX`0000000?oQ00000000hOl00000
002So`0000000P00I_l01000000002So000Xo`00000401@0o`030000001A0?l0D@3o00D0D@3o00<0
000001@0o`0D0?l00P0D0?l01`00000002So000Xo`00:?l00000001Vo`0000000P00Xol00`000000
0>7o00000002003oh@0300000000oj<0000000800?mV00H00000003o:000obP0000001Co00000003
057o00050000002?o`00Sol008oo00000000303<o`000`000000Sol008oo000208oo00050000001A
o`00DOl0057o000000000`0Do`00100000000?lX003o:0000002003oIP0400000000oj<00?nS0000
00<00?oQ00<00000003Qo`00hOl00P00hOl00`0000000:?o002So`05002So`0400000000Xol00:?o
000000L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`0500000000I_l006Ko
001Vo`0000001`00Xol00`0000000>7o003Qo`02003Qo`8000000P00on40100000000?nS003oX`00
0003003oIP0400000000obP00?lX000000805?l000@00000057o001Ao`0000001@2?o`000`000000
c?l00<co00090<co00030000002?o`00Sol000<0Sol000<00000057o000000000P0Do`001P000000
0?lX003o:00000000?mV000000800?nS00D00000003oh@0000000>7o00000002002So`0400000000
I_l006Ko0000008002So00<0000001@0o`0D0?l00P0D0?l00`000000D@3o0540o`050540o`030000
000D0?l0503o00<0503o00H00000000Xo`00:?l00000001Vo`000002002So`0800000000hOl00000
003oh@0000000?nS00000000ofH2000000805?l000L00000057o00000000Sol00000003<o`000000
0080ooD000<000000?nh003o^0000P3o^0000`000000ogX00?mj00040?mj00030000003o^000okP0
0080okP0014000000?oe003om@0000000<co00000000Sol00000001Ao`00000001Co000000000?lX
00000000ofH00000003oX`0200000008003Qo`0000000:?o00000000I_l00000000Xo`00000201@0
o`040000001A0?l0D@3o000000`0S`3o00L000000540o`1A0?l0D@3o0000000D0?l00000008002So
00/00000001Vo`0000000000003Qo`0000000?oQ00000000oj<00000003oIP020000000601Co0000
0000DOl00000002?o`0000000P3<o`0010000000ooD00?oe000000040?nh00030000003oNP00ogX0
00<0ogX000<000000?nh003o^0000`3o^0001P000000ooD00?oe00000000c?l000000080Sol000P0
0000057o000000005?l000000000obP00000003oIP8000000P00on401@0000000>7o00000000Xol0
0000008006Ko00@00000000Xo`00:?l00000100D0?l00`000000D@3o0540o`050540o`030000000D
0?l0503o0080503o00T00000000Xo`00:?l002So00000000I_l00000002So`0000000P00hOl01@00
00000?oQ00000000oj<0000000800?mV00H00000003o:000obP0000001Co00000003057o00050000
002?o`00Sol008oo00000000303<o`000`000000Sol008oo000308oo00040000001Ao`00DOl00000
00<05?l000@00000003o:000obP000000P00ofH0100000000?nS003oX`000002003oh@8000001000
hOl00`0000000:?o002So`05002So`0300000000I_l0000000800:?o00<000000?ooo`3oool0103o
ool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00<00000001Vo`00I_l00P00I_l2
000000H00:?o0P000004003Qo`0400000000on400?oQ000000800?nS0P000002003oIP0400000000
obP00?lX000000805?l000@00000057o001Ao`0000001@2?o`000`000000c?l00<co00090<co0003
0000002?o`00Sol000<0Sol000<00000057o000000000P0Do`00200000000?lX003o:00000000?mV
00000000oj<000000P00on400`0000000>7o00000002002So`0400000000I_l006Ko0000008002So
00<0000001@0o`0D0?l00P0D0?l00`000000D@3o0540o`050540o`030000000D0?l0503o0080503o
00@00000000Xo`00:?l000000P00I_l04`0000000:?o00000000hOl00000003oh@0000000?nS0000
0000ofH00000003o:00000005?l00000001Ao`00000008oo000000000P3<o`0010000000ooD00?oe
000000040?nh00030000003oNP00ogX000<0ogX000<000000?nh003o^0000P3o^00010000000ooD0
0?oe000000020<co00080000002?o`000000057o000000005?l000000000obP20000000>003oX`00
00000?oQ00000000hOl00000002So`00000006Ko00000000:?l0000001@0o`0000030540o`030000
002?0?l0S`3o00T0S`3o00L000000540o`1A0?l0D@3o0000000D0?l00000008002So00/00000001V
o`0000000000003Qo`0000000?oQ00000000oj<00000003oIP020000000601Co00000000DOl00000
002?o`0000000P3<o`0010000000ooD00?oe000000050?nh00030000003oNP00ogX00080ogX000<0
00000?nh003o^0000P3o^0001P000000ooD00?oe00000000c?l000000080Sol001400000057o0000
00005?l000000000obP00000003oIP0000000?nS00000000on400000003Qo`0000000:?o00000002
001Vo`0400000000:?l002So000000@0503o00<000000540o`1A0?l01@1A0?l00`000000503o01@0
o`0201@0o`0700000000:?l002So000Xo`00000006Ko00000002002So`0300000000hOl000000080
0?oQ00D00000003oX`0000000?mV00000002003o:0040000000Do`005?l000000080DOl000<00000
08oo002?o`000P2?o`000`000000c?l00<co000:0<co00030000002?o`00Sol00080Sol000@00000
057o001Ao`0000000`0Do`00100000000?lX003o:0000002003oIP0400000000oj<00?nS00000080
0?oQ00<00000003Qo`00hOl00P00hOl2000000H00:?o0P000002001Vo`0300000000Xol00:?o00L0
oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`03002So`00000006Ko00@006Ko
00<00000002So`00Xol01@00Xol00`0000000>7o003Qo`02003Qo`0400000000on400?oQ000000<0
0?nS00@00000003oIP00ofH000000P00obP0100000005?l001Co00000002057o00800000102?o`00
0`000000c?l00<co00090<co00030000002?o`00Sol00080Sol000@00000057o001Ao`0000000P0D
o`00200000000?lX003o:00000000?mV00000000oj<000000P00on400`0000000>7o00000002002S
o`0400000000I_l006Ko0000008002So00<0000001@0o`0D0?l00P0D0?l00`000000D@3o0540o`04
0540o`030000000D0?l0503o00<0503o00H00000000Xo`00:?l00000001Vo`000002002So`0C0000
0000hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`00000008oo
00000000c?l0000000020?oe00030000003o^000okP000<0okP01P0000040?nh00040000003om@00
ooD000000080c?l000P0000008oo00000000DOl00000000Do`000000003o:080000000`00?nS0000
0000on400000003Qo`0000000:?o00000000I_l00000000Xo`00000201@0o`040000001A0?l0D@3o
000000/0S`3o00L000000540o`1A0?l0D@3o0000000D0?l00000008002So00/00000001Vo`000000
0000003Qo`0000000?oQ00000000oj<00000003oIP020000000601Co00000000DOl00000002?o`00
00000P3<o`0010000000ooD00?oe000000050?nh00D000001@3o^0001P000000ooD00?oe00000000
c?l000000080Sol001400000057o000000005?l000000000obP00000003oIP0000000?nS00000000
on400000003Qo`0000000:?o00000002001Vo`0400000000:?l002So000000@0503o00<000000540
o`1A0?l0101A0?l00`000000503o01@0o`0201@0o`8000000P00:?l02000000006Ko001Vo`000000
0:?o00000000hOl000000P00on400`0000000?nS00000002003oIP0300000000obP0000000805?l0
00@00000057o001Ao`000000102?o`000`000000c?l00<co000:0<co00030000002?o`00Sol00080
Sol000@00000057o001Ao`0000000`0Do`00100000000?lX003o:0000002003oIP0400000000oj<0
0?nS000000<00?oQ00D00000003Qo`00hOl00>7o00000007002So`0300000000I_l006Ko00<006Ko
00<00000002So`3oool01P3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo0P00
Xol00`00000006Ko001Vo`03001Vo`0300000000Xol00:?o00D00:?o00<00000003Qo`00hOl00P00
hOl0100000000?oQ003oh@000003003oX`0600000000ofH00?mV00000000obP000000`0Do`001@00
0000DOl0057o001Ao`00000000@0Sol000<000000<co003<o`002@3<o`000`000000Sol008oo0002
08oo00040000001Ao`00DOl0000000805?l000P00000003o:000obP00000003oIP0000000?nS0000
00800?oQ00<00000003Qo`0000000P00Xol00`00000006Ko00000003000Xo`030000000D0?l0503o
0080503o00<000000540o`1A0?l00`1A0?l00`000000503o01@0o`0301@0o`0600000000:?l002So
00000000I_l000000P00Xol04@0000000>7o00000000on400000003oX`0000000?mV00000000obP0
000001Co00000000DOl00000002?o`0000000080c?l000@000000?oe003om@0000003P3o^0001000
0000ooD00?oe000000020<co00080000002?o`000000057o000000005?l000000000obP20000000<
003oX`0000000?oQ00000000hOl00000002So`00000006Ko00000000:?l000000P0D0?l010000000
D@3o0540o`00000;08l0o`070000001A0?l0D@3o0540o`000000503o00000002000Xo`0900000000
I_l0000000000000hOl00000003oh@0000000?nS0080000000P00?lX0000000Do`000000057o0000
0000Sol000000080c?l000@000000?oe003om@0000003P3o^00010000000ooD00?oe000000020<co
000?0000002?o`000000057o000000005?l000000000obP00000003oIP0000000?nS00000000on40
000000800>7o0P000002001Vo`0400000000:?l002So000000D0503o00<000000540o`1A0?l00`1A
0?l00`000000503o01@0o`0201@0o`0500000000:?l002So000Xo`0000000P00I_l01@0000000:?o
00000000hOl0000000800?oQ00<00000003oX`0000000P00ofH00`0000000?lX0000000201Co0004
0000001Ao`00DOl0000000D0Sol000<000000<co003<o`00203<o`02000000<0Sol00P000002057o
00050000000Do`005?l001Co000000000P00obP0100000000?mV003oIP000002003oX`0500000000
on400?oQ003oh@0000000`00hOl00`0000000:?o002So`05002So`0300000000I_l006Ko00<006Ko
00<00000002So`00Xol01`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo0`00
Xol00`00000006Ko001Vo`03001Vo`8000001P00Xol2000000<00>7o00<00000003oh@00on400P00
0002003oX`0600000000ofH00?mV00000000obP000000`0Do`001@000000DOl0057o001Ao`000000
00@0Sol000<000000<co003<o`002@3<o`001@000000Sol008oo002?o`0000000080DOl000T00000
01Co000Do`005?l000000000obP00000003oIP0000000P00oj<00`0000000?oQ00000002003Qo`03
00000000Xol00000008006Ko00@00000000Xo`00:?l00000100D0?l00`000000D@3o0540o`030540
o`030000000D0?l0503o00@0503o00<00000000Xo`0000000P00I_l00`0000000:?o002So`020000
00800?oQ0P0000002000ofH00000003o:00000005?l00000001Ao`0000000P2?o`000`000000c?l0
000000020?oe00030000003o^000okP000`0okP000@000000?oe003om@0000000P3<o`0020000000
Sol00000001Ao`00000001Co000000000?lX0P0000003000oj<00000003oh@0000000>7o00000000
Xol00000001Vo`00000002So00000080503o00@000000540o`1A0?l000002`2?0?l010000000D@3o
0540o`00000201@0o`0:00000000:?l00000001Vo`0000000:?o00000000hOl00000003oh@800000
00`00?mV00000000obP0000001Co00000000DOl00000002?o`0000000<co000000020?oe00030000
003o^000okP000d0okP000@000000?oe003om@0000000P3<o`004`000000Sol00000001Ao`000000
01Co000000000?lX00000000ofH00000003oX`0000000?oQ00000000hOl00000002So`0000000P00
I_l01000000002So000Xo`00000401@0o`030000001A0?l0D@3o00@0D@3o00D0000001@0o`0D0?l0
503o00000003000Xo`0400000000I_l006Ko001Vo`8000000P00hOl01@0000000?oQ00000000oj<0
000000800?mV00<00000003o:00000000P0Do`0010000000DOl0057o0000000508oo00030000003<
o`00c?l000P0c?l000<0000008oo002?o`000P2?o`001@000000DOl0057o001Ao`00000000<05?l0
00@00000003o:000obP000000P00ofH0100000000?nS003oX`000002003oh@8000000P00hOl20000
00H00:?o0P000005001Vo`0400000000Xol00:?o002So`L0oooo0@0000010?ooo`00803oool00`00
0000oooo0?ooo`050?ooo`@00:?o00<00000001Vo`00I_l01000I_l00`0000000:?o002So`05002S
o`0500000000hOl00>7o003Qo`0000000`00on40100000000?nS003oX`000002003oIP0400000000
obP00000000000805?l000D00000057o001Ao`00DOl00000000308oo00030000003<o`00c?l000T0
c?l000<0000008oo002?o`000P2?o`0010000000DOl0057o0000000201Co000800000000obP00?lX
00000000ofH00000003oX`000002003oh@0300000000hOl0000000800:?o00<00000001Vo`000000
0`00:?l00`000000503o01@0o`0201@0o`030000001A0?l0D@3o00<0D@3o00<0000001@0o`0D0?l0
0`0D0?l01P00000002So000Xo`00000006Ko000000800:?o00H00000003Qo`0000000?oQ00000000
oj<2000000800?lX0P000002057o00030000002?o`0000000080c?l000@000000?oe003om@000000
3P3o^0004`000000ooD00?oe00000000c?l00000002?o`000000057o000000005?l000000000obP0
0000003oIP0000000?nS00000000on400P0000001000Xol00000001Vo`000002000Xo`030000000D
0?l0000000<0D@3o00<0000008l0o`2?0?l01`2?0?l01@000000D@3o0540o`1A0?l000000080503o
00X00000000Xo`00000006Ko00000000Xol00000003Qo`0000000?oQ0P0000003000ofH00000003o
:00000005?l00000001Ao`00000008oo00000000c?l000000080ooD000<000000?nh003o^000303o
^00010000000ooD00?oe000000020<co000?0000002?o`000000057o000000005?l000000000obP0
0000003oIP0000000?nS00000000on40000000800>7o0P000002001Vo`0400000000:?l002So0000
00D0503o00<000000540o`1A0?l00`1A0?l00`000000503o01@0o`0201@0o`0500000000:?l002So
000Xo`0000000P00I_l01`0000000:?o00000000hOl00000003oh@0000000P00oj<00`0000000?mV
00000002003o:0040000000Do`005?l000000080DOl000<0000008oo002?o`000P2?o`000`000000
c?l00<co00090<co00050000002?o`00Sol008oo000000000`1Ao`001@0000005?l001Co000Do`00
000000800?lX00@00000003oIP00ofH000000P00oj<0100000000?oQ003oh@000003003Qo`030000
0000Xol00:?o00D00:?o00<00000001Vo`00I_l01000I_l00`0000000:?o002So`02002So`L0oooo
0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`D00:?o00<00000001Vo`00I_l01000
I_l2000000H00:?o00D00000003Qo`00hOl00>7o00000002003oh@0500000000oj<00?nS003oX`00
00000P00ofH0100000000?lX003o:000000201Co00050000001Ao`00DOl0057o000000000`2?o`00
0`000000c?l00<co00090<co00030000002?o`00Sol00080Sol000@00000057o001Ao`0000000P0D
o`00200000000?lX003o:00000000?mV00000000oj<000000P00on400`0000000>7o00000002002S
o`0400000000I_l000000000008002So00<0000001@0o`0D0?l00P0D0?l2000000<0D@3o00<00000
01@0o`0D0?l00`0D0?l01000000002So000Xo`000002001Vo`0A00000000Xol00000003Qo`000000
0?oQ00000000oj<00000003oIP0000000?lX0000000Do`000000057o000000000P2?o`000`000000
c?l0000000030?oe00030000003o^000okP000/0okP001<000000?oe003om@0000000<co00000000
Sol00000001Ao`00000001Co000000000?lX00000000ofH00000003oX`0000000?oQ0080000000@0
0:?o00000000I_l000000P00:?l00`000000503o000000030540o`030000002?0?l0S`3o00L0S`3o
00D000000540o`1A0?l0D@3o0000000201@0o`0:00000000:?l00000001Vo`0000000:?o00000000
hOl00000003oh@80000000`00?mV00000000obP0000001Co00000000DOl00000002?o`0000000<co
000000020?oe00030000003o^000okP000/0okP00P0000020?oe000F0000003<o`00c?l00000002?
o`000000057o000000005?l000000000obP00000003oIP0000000?nS00000000on400000003Qo`00
00000:?o0000008006Ko00@00000000Xo`00:?l000001@0D0?l010000000D@3o0540o`1A0?l20000
00@0503o00D00000000Xo`00:?l002So00000002001Vo`0400000000Xol00000003Qo`80000000@0
0?oQ00000000oj<000000P00ofH00`0000000?lX0000000201Co00040000001Ao`00DOl0000000@0
Sol000<000000<co003<o`002@3<o`000`000000Sol008oo000208oo00040000001Ao`00DOl00000
00<05?l000@00000003o:000obP000000P00ofH0100000000?nS003oX`000002003oh@0500000000
hOl00>7o003Qo`0000001`00Xol00`00000006Ko001Vo`04001Vo`0300000000Xol00:?o00<00:?o
1`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo1P00Xol00`00000006Ko001V
o`05001Vo`0300000000Xol00:?o00@00:?o00@00000003Qo`00hOl00>7o0P0000000`00on400000
003oX`02003oX`0300000000ofH0000000800?lX00D0000001Co000Do`005?l000000003057o0004
0000002?o`00Sol008oo008000002P3<o`000`000000Sol008oo000208oo00040000001Ao`00DOl0
000000805?l000<00000003o:00000000P00ofH01@0000000?nS00000000on40000000800>7o00<0
0000002So`0000000P00I_l01000000002So000Xo`00000601@0o`030000001A0?l0000000H0503o
00T00000000Xo`00:?l002So00000000I_l00000002So`0000000P00hOl20000000:003oX`000000
0?mV00000000obP0000001Co00000000DOl000000080Sol000<000000<co000000000`3om@000`00
0000okP00?nh000;0?nh000C0000003om@00ooD00000003<o`00000008oo00000000DOl00000000D
o`000000003o:00000000?mV00000000oj<00000003oh@0200000009002So`00000006Ko001Vo`00
000002So0000000D0?l0000000<0D@3o00<0000008l0o`2?0?l01`2?0?l01@000000D@3o0540o`1A
0?l000000080503o00P00000000Xo`00000006Ko00000000Xol00000003Qo`80000000h00?nS0000
0000ofH00000003o:00000005?l00000001Ao`00000008oo00000000c?l000000080ooD000<00000
0?nh003o^0002`3o^00010000000ooD00?oe000000020<co000=0000002?o`000000057o00000000
5?l000000000obP00000003oIP0000000?nS00000002003oh@8000000P00Xol00`00000006Ko0000
0002000Xo`030000000D0?l0503o00D0503o00<000000540o`0000001P0D0?l01000000002So000X
o`000002001Vo`0500000000Xol00000003Qo`0000000P00on400`0000000?nS00000002003oIP03
00000000obP0000000805?l000@00000057o001Ao`000000102?o`000`000000c?l00<co00080<co
00800000102?o`0010000000DOl0057o0000000301Co000400000000obP00?lX000000800?mV00@0
0000003oX`00oj<000000P00on401@0000000>7o003Qo`00hOl0000000H00:?o0P000006001Vo`03
00000000Xol00:?o00@00:?o1`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo
00<00000002So`00Xol01000Xol00`00000006Ko001Vo`05001Vo`0300000000Xol00:?o00@00:?o
00<00000003Qo`00hOl00P00hOl01P0000000?oQ0000000000000?nS000000800?mV00@00000003o
:000obP000000`0Do`0010000000DOl0057o0000000508oo00030000003<o`00c?l000P0c?l000D0
000008oo002?o`00Sol000000003057o00030000000Do`00000000800?lX00P00000003oIP00ofH0
0000003oX`0000000?oQ000000800>7o00<00000002So`0000000P00I_l01000000002So000Xo`00
000601@0o`030000000D0?l0503o00H0503o00@00000000Xo`00:?l000000P00I_l04`0000000:?o
00000000hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`000000
08oo000000000P3<o`0010000000ooD00?oe0000000=0?nh000E0000003om@00ooD00000003<o`00
000008oo00000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@000000
0>7o008000000P00I_l01@00000002So0000000D0?l0000000@0D@3o00<0000008l0o`2?0?l01@2?
0?l01@000000D@3o0540o`1A0?l000000080503o00<00000000Xo`00:?l00P000002002So`800000
00D00?oQ00000000oj<00000003oIP020000000601Co00000000DOl00000002?o`0000000P3<o`00
10000000ooD00?oe0000000<0?nh00050000003om@00ooD00?oe000000000P3<o`004`000000Sol0
0000001Ao`00000001Co000000000?lX00000000ofH00000003oX`0000000?oQ00000000hOl00000
002So`0000000P00I_l01000000002So000Xo`00000801@0o`030000000D0?l0503o00@0503o00@0
0000000Xo`00:?l000000P00I_l02P0000000:?o002So`0000000>7o00000000on400000003oX`00
0002003oIP0600000000obP00?lX0000000Do`0000000P1Ao`000`000000Sol008oo000308oo0003
0000003<o`00c?l000P0c?l000<0000008oo002?o`000P2?o`0010000000DOl0057o0000000301Co
000400000000obP00?lX000000800?mV00@00000003oX`00oj<000000P00on401@0000000>7o003Q
o`00hOl0000000H00:?o00<00000001Vo`00I_l01@00I_l00`0000000:?o002So`05002So`030000
003oool0oooo00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`03003Qo`00
00000:?o00D00:?o00<00000001Vo`00I_l01@00I_l00`0000000:?o002So`04002So`0500000000
hOl00>7o003Qo`0000000`00on400`0000000?nS00000002003oIP0400000000obP00?lX000000<0
5?l000@00000057o001Ao`000000102?o`000`000000c?l00<co00080<co00030000002?o`00Sol0
0080Sol000@00000057o001Ao`0000000P0Do`000`0000000?lX00000002003oIP0300000000oj<0
0?nS008000000P00hOl00`0000000:?o00000002001Vo`0400000000:?l002So000000l0503o00T0
0000000Xo`00:?l002So00000000I_l00000002So`0000000P00hOl20000000:003oX`0000000?mV
00000000obP0000001Co00000000DOl000000080Sol000<000000<co000000000`3om@000`000000
okP00?nh000:0?nh00040000003om@00ooD000000080c?l000X0000008oo00000000DOl00000000D
o`000000003o:00000000?mV0P0000002P00on400000003Qo`0000000:?o00000000I_l00000000X
o`00000201@0o`050000001A0?l0D@3o0540o`0000001`2?0?l01@000000D@3o0540o`1A0?l00000
0080503o00h00000000Xo`00000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000
003oIP80000000H05?l00000001Ao`00000008oo000000020<co00050000003om@00ooD00?oe0000
00002`3o^00010000000ooD00?oe000000020<co000=0000002?o`000000057o000000005?l00000
0000obP00000003oIP0000000?nS00000002003oh@8000000P00Xol00`00000006Ko00000002000X
o`030000000D0?l0503o00h0503o00@00000000Xo`00:?l000000P00I_l01`0000000:?o00000000
hOl00000003oh@0000000P00oj<01P0000000?mV003oIP0000000?lX000000805?l000@00000057o
001Ao`000000102?o`000`000000c?l00<co00080<co00030000002?o`00Sol00080Sol000@00000
057o001Ao`0000000`0Do`00100000000?lX003o:0000002003oIP0400000000oj<00?nS00000080
0?oQ00D00000003Qo`00hOl00>7o00000006002So`0300000000I_l006Ko00D006Ko00<00000002S
o`00Xol01000Xol2000000L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`80
0>7o00<00000002So`00Xol01000Xol00`00000006Ko001Vo`05001Vo`0300000000Xol00:?o00@0
0:?o00D00000003Qo`00hOl00>7o00000002003oh@0400000000oj<00?nS000000800?mV00@00000
003o:000obP000000`0Do`0010000000DOl0057o0000000408oo00030000003<o`00c?l000P0c?l0
00<0000008oo002?o`000P2?o`0010000000DOl0057o0000000201Co000300000000obP000000080
0?mV00D00000003oX`0000000?oQ00000002003Qo`0600000000Xol0000000000000I_l000000P00
:?l00`000000503o01@0o`0=01@0o`0400000000:?l002So0000008006Ko01<00000002So`000000
0>7o00000000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`00
00000080c?l000@000000?oe003om@000000303o^00010000000ooD00?oe000000020<co000:0000
002?o`000000057o000000005?l000000000obP00000003oIP80000000X00?oQ00000000hOl00000
002So`00000006Ko00000000:?l000000P0D0?l00`000000D@3o0540o`020540o`030000002?0?l0
S`3o00@0S`3o00D000000540o`1A0?l0D@3o0000000201@0o`0>00000000:?l00000001Vo`000000
0:?o00000000hOl00000003oh@0000000?nS00000000ofH20000000601Co00000000DOl00000002?
o`0000000P3<o`001@000000ooD00?oe003om@00000000/0okP000@000000?oe003om@0000000P3<
o`004@000000Sol00000001Ao`00000001Co000000000?lX00000000ofH00000003oX`0000000?oQ
00000000hOl0000000800:?o00<00000001Vo`0000000P00:?l00`000000503o01@0o`0=01@0o`04
00000000:?l002So0000008006Ko00X00000002So`00Xol00000003Qo`0000000?oQ00000000oj<0
00000P00ofH00`0000000?lX0000000201Co00040000001Ao`00DOl0000000@0Sol000<000000<co
003<o`002@3<o`001@000000Sol008oo002?o`00000000<0DOl000@0000001Co000Do`0000000P00
obP0100000000?mV003oIP000002003oX`0400000000on400?oQ000000<00>7o00<00000002So`00
Xol01000Xol00`00000006Ko001Vo`05001Vo`0300000000Xol00:?o00@00:?o00<00000003Qo`00
hOl01`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo0`00hOl00`0000000:?o
002So`04002So`0300000000I_l006Ko00D006Ko0P000005002So`8000000P00hOl200000005003o
h@0000000?nS003oX`0000000P00ofH0100000000?lX003o:000000301Co00040000001Ao`00DOl0
000000@0Sol000<000000<co003<o`00203<o`001@000000Sol008oo002?o`00000000<0DOl000H0
000001Co000Do`000000003o:0000002003oIP0700000000oj<00000003oh@0000000>7o00000002
002So`0300000000I_l0000000<002So00<0000001@0o`0D0?l0300D0?l01000000002So000Xo`00
0002001Vo`0A00000000Xol00000003Qo`0000000?oQ00000000oj<00000003oIP0000000?lX0000
000Do`000000057o000000000P2?o`000`000000c?l0000000030?oe00030000003o^000okP000T0
okP000@000000?oe003om@0000000P3<o`002P000000Sol00000001Ao`00000001Co000000000?lX
00000000ofH20000000:003oh@0000000>7o00000000Xol00000001Vo`00000002So00000080503o
00<000000540o`1A0?l00P1A0?l00`000000S`3o08l0o`0308l0o`030000001A0?l0D@3o0080D@3o
0140000001@0o`0D0?l00000000Xo`00000006Ko00000000Xol00000003Qo`0000000?oQ00000000
oj<00000003oIP020000000601Co00000000DOl00000002?o`0000000P3<o`001@000000ooD00?oe
003om@00000000X0okP000@000000?oe003om@0000000P3<o`0050000000Sol008oo00000000DOl0
0000000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o00000000Xol00000
0P00I_l01000000002So000Xo`00000>01@0o`0400000000:?l002So0000008006Ko00X00000002S
o`00Xol00000003Qo`0000000?oQ00000000oj<000000P00ofH00`0000000?lX0000000201Co0004
0000001Ao`00DOl0000000@0Sol000<000000<co003<o`00203<o`02000000<0Sol000<00000057o
001Ao`000P00000201Co000400000000obP00?lX000000800?mV00@00000003oX`00oj<000000P00
on401@0000000>7o003Qo`00hOl0000000D00:?o0P000007001Vo`0300000000Xol00:?o00@00:?o
00@00000003Qo`00hOl00>7o1`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo
00D00000003Qo`00hOl00>7o00000006002So`0300000000I_l006Ko00H006Ko00<00000002So`00
Xol01000Xol01`0000000>7o003Qo`00hOl00000003oh@0000000P00oj<0100000000?mV003oIP00
0002003o:0050000000Do`005?l001Co000000000P1Ao`000`000000Sol008oo000208oo00030000
003<o`00c?l000L0c?l000<0000008oo002?o`000P2?o`0010000000DOl0057o0000000201Co0008
00000000obP00?lX00000000ofH00000003oX`000002003oh@0300000000hOl0000000800:?o00<0
0000001Vo`0000000`00:?l00`000000503o01@0o`0;01@0o`0700000000:?l002So000Xo`000000
06Ko00000002002So`0A00000000hOl00000003oh@0000000?nS00000000ofH00000003o:0000000
5?l00000001Ao`00000008oo000000000P3<o`0010000000ooD00?oe0000000;0?nh00040000003o
m@00ooD000000080c?l000X0000008oo00000000DOl00000000Do`000000003o:00000000?mV0P00
00002P00on400000003Qo`0000000:?o00000000I_l00000000Xo`00000201@0o`030000001A0?l0
D@3o00<0D@3o00<0000008l0o`2?0?l00P2?0?l01@000000D@3o0540o`1A0?l000000080503o00T0
0000000Xo`00:?l00000001Vo`0000000:?o00000000hOl00P0000003000oj<00000003oIP000000
0?lX0000000Do`000000057o00000000Sol000000080c?l000D000000?oe003om@00ooD00000000:
0?nh00040000003om@00ooD000000080c?l00140000008oo00000000DOl00000000Do`000000003o
:00000000?mV00000000oj<00000003oh@0000000>7o00000002002So`0300000000I_l000000080
02So00<0000001@0o`0D0?l0300D0?l01`00000002So000Xo`00:?l00000001Vo`0000000P00Xol0
2P0000000>7o003Qo`0000000?oQ00000000oj<00000003oIP000002003o:0040000000Do`005?l0
00000080DOl000<0000008oo002?o`000P2?o`000`000000c?l00<co00070<co00030000002?o`00
Sol00080Sol000@00000057o001Ao`0000000`0Do`00100000000?lX003o:0000002003oIP040000
0000oj<00?nS000000800?oQ00D00000003Qo`00hOl00>7o00000005002So`0300000000I_l006Ko
00H006Ko00<00000002So`00Xol01000Xol00`0000000>7o003Qo`02003Qo`030000003oool0oooo
00@0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`03003oh@0000000>7o0080
0>7o00<00000002So`00Xol01000Xol00`00000006Ko001Vo`05001Vo`0300000000Xol00:?o00@0
0:?o00D00000003Qo`00hOl00>7o00000002003oh@0400000000oj<00?nS000000800?mV00@00000
003o:000obP000000P0Do`001@000000DOl0057o001Ao`00000000@0Sol000<000000<co003<o`00
1`3<o`000`000000Sol008oo000208oo00040000001Ao`00DOl0000000805?l000<00000003o:000
00000P00ofH01`0000000?nS00000000on400000003Qo`0000000P00Xol01000000006Ko001Vo`00
0002000Xo`030000000D0?l0503o00`0503o00@00000000Xo`00:?l000000P00I_l04@0000000:?o
00000000hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`000000
0080Sol000<000000<co000000000`3om@000`000000okP00?nh00080?nh00090000003om@00ooD0
0?oe00000000c?l00000002?o`0000000080DOl00P0000001`00obP00000003oIP0000000?nS0000
0000on400P0000001`00Xol00000001Vo`00I_l00000000Xo`0000000P0D0?l00`000000D@3o0540
o`020540o`D00000101A0?l010000000503o01@0o`000002000Xo`0600000000I_l00000002So`00
00000>7o0P0000003000oj<00000003oIP0000000?lX0000000Do`000000057o00000000Sol00000
0080c?l000D000000?oe003om@00ooD0000000090?nh00040000003om@00ooD000000080c?l001@0
000008oo002?o`000000057o000000005?l000000000obP00000003oIP0000000?nS00000000on40
0000003Qo`0000000:?o0000008006Ko00@00000000Xo`00:?l000003@0D0?l01@00000002So000X
o`00:?l00000008006Ko00X00000002So`00Xol00000003Qo`0000000?oQ00000000oj<000000P00
ofH00`0000000?lX0000000201Co00040000001Ao`00DOl0000000@0Sol000<000000<co003<o`00
1`3<o`000`000000Sol008oo000208oo00040000001Ao`00DOl0000000<05?l000@00000003o:000
obP000000P00ofH0100000000?nS003oX`000002003oh@0500000000hOl00>7o003Qo`0000001P00
Xol00`00000006Ko001Vo`05001Vo`0300000000Xol00:?o00@00:?o00@00000003Qo`00hOl00>7o
0P0000070?ooo`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool2003oh@0500000000
hOl00>7o003Qo`0000001P00Xol00`00000006Ko001Vo`05001Vo`0300000000Xol00:?o00@00:?o
00D00000003Qo`00hOl00>7o00000002003oh@0400000000oj<00?nS000000800?mV00@00000003o
:000obP000000P0Do`001@000000DOl0057o001Ao`00000000@0Sol000<000000<co003<o`001`3<
o`000`000000Sol008oo000208oo00040000001Ao`00DOl0000000805?l000D00000003o:0000000
0?mV00000002003oX`0400000000on400000003Qo`80000000D00:?o00000000I_l006Ko00000002
000Xo`030000000D0?l0503o00/0503o00L00000000Xo`00:?l002So00000000I_l0000000800:?o
01400000003Qo`0000000?oQ00000000oj<00000003oIP0000000?lX0000000Do`000000057o0000
0000Sol0000000020<co00050000003om@00ooD00?oe000000002@3o^0002@000000ooD00?oe003o
m@0000000<co00000000Sol000000002057o0080000000L00?lX00000000ofH00000003oX`000000
0?oQ0080000000L00:?o00000000I_l006Ko00000000:?l000000080503o00<000000540o`1A0?l0
2`1A0?l010000000503o01@0o`000002000Xo`0600000000I_l00000002So`0000000>7o0P000000
3000oj<00000003oIP0000000?lX0000000Do`000000057o00000000Sol000000080c?l000D00000
0?oe003om@00ooD0000000080?nh008000000P3om@0010000000c?l00<co0000000208oo000A0000
001Ao`00000001Co000000000?lX00000000ofH00000003oX`0000000?oQ00000000hOl00000002S
o`0000000P00I_l01000000002So000Xo`00000=01@0o`0700000000:?l002So000Xo`00000006Ko
00000002002So`0300000000hOl0000000800?oQ00H00000003oX`0000000?mV00000000obP20000
000301Co00000000DOl00080DOl000D0000008oo002?o`00Sol0000000090<co00030000002?o`00
Sol000<0Sol000@00000057o001Ao`0000000P0Do`00100000000?lX003o:0000002003oIP040000
0000oj<00?nS000000800?oQ00D00000003Qo`00hOl00>7o00000006002So`0300000000I_l006Ko
00D006Ko00<00000002So`00Xol01000Xol01@0000000>7o003Qo`00hOl0000000800?oQ1`3oool1
00000040oooo000P0?ooo`030000003oool0oooo00D0oooo00@00000003oh@00on4000000`00hOl0
0`0000000:?o002So`04002So`8000001P00I_l2000000D00:?o0P000002003Qo`0400000000on40
0?oQ000000800?nS00@00000003oIP00ofH000000P00obP0100000005?l001Co00000003057o0003
0000002?o`00Sol00080Sol000<000000<co003<o`001P3<o`000`000000Sol008oo000308oo0003
0000001Ao`00000000805?l000H00000003o:000obP00000003oIP000002003oX`8000000P00hOl0
0`0000000:?o00000002001Vo`0500000000:?l002So000Xo`000000300D0?l01`00000002So000X
o`00:?l00000001Vo`0000000P00Xol04@0000000>7o00000000on400000003oX`0000000?mV0000
0000obP0000001Co00000000DOl00000002?o`0000000080c?l000D000000?oe003om@00ooD00000
00090?nh00090000003om@00ooD00?oe00000000c?l00000002?o`0000000080DOl00P0000001`00
obP00000003oIP0000000?nS00000000on400P0000001`00Xol00000001Vo`00I_l00000000Xo`00
00000P0D0?l00`000000D@3o0540o`0;0540o`040000000D0?l0503o0000008002So00H00000001V
o`0000000:?o00000000hOl20000000<003oX`0000000?mV00000000obP0000001Co00000000DOl0
0000002?o`0000000P3<o`001@000000ooD00?oe003om@00000000L0okP000<000000?oe003om@00
0P3om@0050000000c?l00<co00000000Sol00000001Ao`00000001Co000000000?lX00000000ofH0
0000003oX`0000000?oQ00000000hOl000000P00Xol01P00000006Ko001Vo`00000002So000000d0
503o00L00000000Xo`00:?l002So00000000I_l0000000800:?o00<00000003Qo`0000000P00on40
1`0000000?nS00000000ofH00000003o:00000000P0Do`001@000000DOl0057o001Ao`00000000<0
Sol000<000000<co003<o`001`3<o`000`000000Sol008oo000308oo00040000001Ao`00DOl00000
00805?l000@00000003o:000obP000000P00ofH0100000000?nS003oX`000002003oh@0500000000
hOl00>7o003Qo`0000001@00Xol2000000H006Ko0P000006002So`0500000000hOl00>7o003Qo`00
00000`00on400`000000oooo0?ooo`040?ooo`4000000@3oool00200oooo00<000000?ooo`3oool0
1@3oool01@00oj<00000003oh@00on40000000<00>7o00<00000002So`00Xol01@00Xol00`000000
06Ko001Vo`05001Vo`0300000000Xol00:?o00@00:?o00@00000003Qo`00hOl000000P00on401000
00000?nS003oX`000002003oIP0400000000obP00?lX000000805?l000D00000057o001Ao`00DOl0
0000000408oo00030000003<o`00c?l000H0c?l000<0000008oo002?o`000P2?o`0010000000DOl0
057o0000000201Co000300000000obP00?lX008000000P00oj<00`0000000?oQ00000002003Qo`03
00000000Xol00000008006Ko00@00000000Xo`00:?l000003@0D0?l05@00000002So000Xo`00:?l0
0000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH00000003o:0000000
5?l000000002057o00050000002?o`0000000<co00000000103om@000`000000okP00?nh00060?nh
00090000003om@00ooD00?oe00000000c?l00000002?o`0000000080DOl00P0000001`00obP00000
003oIP0000000?nS00000000on400P0000001`00Xol00000001Vo`00I_l00000000Xo`0000000P0D
0?l00`000000D@3o0540o`0:0540o`0B0000000D0?l0503o01@0o`00000002So00000000I_l00000
002So`0000000>7o00000000on400000003oX`0000000?mV0P000000100Do`000000057o00000002
08oo00030000003<o`00000000@0ooD000<000000?nh003o^0001@3o^0001@000000ooD00?oe003o
m@0000000080c?l001@0000008oo002?o`000000057o000000005?l000000000obP00000003oIP00
00000?nS00000000on400000003Qo`0000000:?o0000008006Ko00@00000000Xo`00:?l00000300D
0?l01@00000002So000Xo`00:?l00000008006Ko00<00000002So`0000000P00hOl01`0000000?oQ
00000000oj<00000003oIP0000000P00obP0100000005?l001Co00000002057o00030000002?o`00
Sol00080Sol000<000000<co003<o`001`3<o`0010000000Sol008oo002?o`0200000080DOl000@0
000001Co000Do`0000000P00obP0100000000?mV003oIP000002003oX`0400000000on400?oQ0000
00<00>7o00<00000002So`00Xol00`00Xol00`00000006Ko001Vo`05001Vo`0300000000Xol00:?o
00D00:?o00D00000003Qo`00hOl00>7o00000003003oh@030000003oool0oooo00D0oooo0@000001
0?ooo`00803oool2000000H0oooo00<00?nS00000000on400P00on401@0000000>7o003Qo`00hOl0
000000H00:?o00<00000001Vo`00I_l01@00I_l00`0000000:?o002So`04002So`0400000000hOl0
0>7o000000<00?oQ00H00000003oX`00oj<00000003oIP000002003o:0050000000Do`005?l001Co
000000000`1Ao`001@000000Sol008oo002?o`00000000P0c?l000<0000008oo002?o`000`2?o`00
0`000000DOl00000000201Co000600000000obP00?lX00000000ofH000000P00oj<2000000800>7o
00<00000002So`0000000P00I_l01@00000002So000Xo`00:?l0000000`0503o00L00000000Xo`00
:?l002So00000000I_l0000000800:?o01400000003Qo`0000000?oQ00000000oj<00000003oIP00
00000?lX0000000Do`000000057o00000000Sol0000000020<co00050000003om@00ooD00?oe0000
0000203o^0001@000000ooD00?oe003om@0000000080c?l000X0000008oo00000000DOl00000000D
o`000000003o:00000000?mV0P0000002P00on400000003Qo`0000000:?o00000000I_l00000000X
o`00000301@0o`030000001A0?l0D@3o00T0D@3o0180000001@0o`0D0?l0503o00000000:?l00000
001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000ofH20000000401Co00000000
DOl000000080Sol000<000000<co00000000103om@000`000000okP00?nh00040?nh00030000003o
m@00ooD00080ooD001@000000<co003<o`00000008oo00000000DOl00000000Do`000000003o:000
00000?mV00000000oj<00000003oh@0000000>7o000000800:?o00@00000001Vo`00I_l000000P00
:?l00`000000503o01@0o`0:01@0o`0700000000:?l002So000Xo`00000006Ko00000002002So`03
00000000hOl0000000800?oQ00L00000003oX`0000000?mV00000000obP0000000805?l000@00000
057o001Ao`000000102?o`000`000000c?l00<co00070<co00050000002?o`00Sol008oo00000000
0`1Ao`00100000005?l001Co00000003003o:00600000000ofH00?mV00000000oj<000000P00on40
1@0000000>7o003Qo`00hOl0000000H00:?o00<00000001Vo`00I_l01@00I_l00`0000000:?o002S
o`04002So`0500000000hOl00>7o003Qo`0000000`00on400`0000000?nS0?ooo`050?ooo`800000
0@3oool00200oooo00<000000?ooo`3oool01@3oool2003oX`0500000000on400?oQ003oh@000000
0`00hOl2000000D00:?o00<00000001Vo`00I_l01@00I_l00`0000000:?o002So`04002So`040000
0000hOl00>7o000000<00?oQ00H00000003oX`00oj<00000003oIP000002003o:0040000000Do`00
5?l0000000<0DOl000<0000008oo002?o`000P2?o`000`000000c?l00<co00060<co00030000002?
o`00Sol00080Sol000<00000057o001Ao`000P0000002P0Do`000000003o:000obP00000003oIP00
00000?nS00000000on4200000004003Qo`0000000:?o0000008006Ko00D00000000Xo`00:?l002So
0000000;01@0o`0500000000:?l002So000Xo`0000000P00I_l04@0000000:?o00000000hOl00000
003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`0000000080Sol000<00000
0<co00000000103om@000`000000okP00?nh00050?nh00050000003om@00ooD00?oe000000000P3<
o`002P000000Sol00000001Ao`00000001Co000000000?lX00000000ofH200000008003oh@000000
0>7o00000000Xol00000001Vo`000002000Xo`040000000D0?l0503o000000/0D@3o0180000001@0
o`0D0?l0503o00000000:?l00000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS0000
0000ofH20000000401Co00000000DOl000000080Sol000<000000<co00000000103om@000`000000
okP00?nh00040?nh00030000003om@00ooD00080ooD001@000000<co003<o`00000008oo00000000
DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o000000800:?o
00<00000001Vo`0000000`00:?l00`000000503o01@0o`0901@0o`0500000000:?l002So000Xo`00
00000P00I_l00`0000000:?o00000002003Qo`0500000000on400000003oX`0000000P00ofH01000
00000?lX0000000Do`0200000080DOl000<0000008oo002?o`000P2?o`000`000000c?l00<co0006
0<co00030000002?o`00Sol00080Sol000D00000057o001Ao`00DOl00000000201Co000500000000
obP00?lX00000000ofH00P000002003oX`0400000000on400?oQ000000<00>7o00<00000002So`00
Xol00`00Xol00`00000006Ko001Vo`05001Vo`0300000000Xol00:?o00<00:?o0P000003003Qo`05
00000000on400?oQ003oh@0000000P00oj<70?ooo`4000000@3oool00200oooo00<000000?ooo`3o
ool01@3oool0100000000?nS003oX`000003003oh@0300000000hOl00>7o00800>7o00<00000002S
o`00Xol00`00Xol2000000H006Ko0P000005002So`0400000000hOl00>7o000000<00?oQ00D00000
003oX`00oj<00000003oIP0200000005003o:0000000000001Co000000000`1Ao`000`000000Sol0
08oo000208oo00030000003<o`00c?l000H0c?l000<0000008oo002?o`000P2?o`001`000000DOl0
057o001Ao`00000001Co000000000P00obP01@0000000?mV00000000oj<0000000800?oQ0P000002
002So`0400000000I_l006Ko000000<002So00<0000001@0o`0D0?l0200D0?l01@00000002So000X
o`00:?l00000008006Ko00<00000002So`0000000P00hOl01P0000000?oQ00000000oj<00000003o
IP8000000P0Do`001@000000DOl00000002?o`0000000080c?l000<000000?oe003om@000P3om@00
0`000000okP00?nh00040?nh00050000003om@00ooD00?oe000000000P3<o`002P000000Sol00000
001Ao`00000001Co000000000?lX00000000ofH200000008003oh@0000000>7o00000000Xol00000
001Vo`000002000Xo`040000000D0?l0503o000000/0D@3o00h0000001@0o`0D0?l0503o00000000
:?l00000001Vo`0000000:?o00000000hOl00000003oh@8000000P00ofH20000000401Co00000000
DOl000000080Sol000<000000<co00000000103om@000`000000okP00?nh00030?nh00030000003o
m@00ooD00080ooD000@000000<co003<o`0000000P2?o`004@000000DOl00000000Do`000000003o
:00000000?mV00000000oj<00000003oh@0000000>7o00000000Xol00000008006Ko00D00000000X
o`00:?l002So0000000:01@0o`0500000000:?l002So000Xo`0000000P00I_l00`0000000:?o0000
0002003Qo`0300000000on40000000800?nS00L00000003oIP0000000?lX0000000Do`00000000<0
DOl000<0000008oo002?o`000P2?o`000`000000c?l00<co00060<co00030000002?o`00Sol00080
Sol000H00000057o001Ao`00DOl00000000Do`0200000005003o:00000000000003oIP0000000`00
oj<0100000000?oQ003oh@000003003Qo`0300000000Xol00:?o00<00:?o00<00000001Vo`00I_l0
1000I_l2000000D00:?o00<00000003Qo`00hOl00P00hOl01@0000000?oQ003oh@00on40000000<0
0?nS00<000000?ooo`3oool0103oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo
00@00?mV0000000000000?nS0P000002003oh@8000000`00hOl00`0000000:?o002So`04002So`03
00000000I_l006Ko00D006Ko00<00000002So`00Xol00P00Xol01@0000000>7o003Qo`00hOl00000
00<00?oQ00@00000003oX`00oj<000000P00ofH01@0000000?lX003o:00000005?l0008000000P1A
o`000`000000Sol008oo000208oo00030000003<o`00c?l000D0c?l000<0000008oo002?o`000`2?
o`0010000000DOl0057o0000000201Co000500000000obP00000003oIP0000000P00oj<01@000000
0?oQ00000000hOl0000000800:?o00@00000001Vo`00I_l000000P00:?l00`000000503o01@0o`09
01@0o`0500000000:?l002So000Xo`0000000P00I_l04@0000000:?o00000000hOl00000003oh@00
00000?nS00000000ofH00000003o:00000005?l00000001Ao`0000000080Sol000<000000<co0000
00001@3om@000`000000okP00?nh00020?nh00030000003om@00ooD00080ooD001X000000<co003<
o`00000008oo00000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@00
00000>7o00000000Xol00000001Vo`00000002So00000080503o00<000000540o`1A0?l0201A0?l0
1@000000503o01@0o`0D0?l00000008002So00H00000001Vo`0000000:?o00000000hOl20000000<
003oX`0000000?mV00000000obP0000001Co00000000DOl00000002?o`0000000P3<o`000`000000
ooD00?oe00020?oe00030000003o^000okP000<0okP000<000000?oe003om@000P3om@0050000000
c?l00<co00000000Sol00000001Ao`00000001Co000000000?lX00000000ofH00000003oX`000000
0?oQ00000000hOl000000P00Xol00`00000006Ko00000003000Xo`030000000D0?l0503o00T0503o
00D00000000Xo`00:?l002So00000002001Vo`0500000000Xol00000003Qo`0000000P00on401`00
00000?nS00000000ofH00000003o:00000000P0Do`001@000000DOl0057o001Ao`00000000<0Sol0
00<000000<co003<o`001`3<o`001@000000Sol008oo002?o`00000000<0DOl000<0000001Co0000
00000P00obP0100000000?mV003oIP000002003oX`8000000P00on401@0000000>7o003Qo`00hOl0
000000@00:?o0P000006001Vo`0300000000Xol00:?o00@00:?o00@00000003Qo`00hOl00>7o0P00
0002003oh@8000000P00oj<2000000L0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`05
0?ooo`<00?mV00@00000003oX`00oj<000000`00on401@0000000>7o003Qo`00hOl0000000H00:?o
00<00000001Vo`00I_l01@00I_l00`0000000:?o002So`02002So`8000000P00hOl2000000800?oQ
00@00000003oX`00oj<000000P00ofH0100000000?lX003o:000000201Co00040000001Ao`00DOl0
000000@0Sol00P0000060<co00030000002?o`00Sol000<0Sol000@00000057o001Ao`0000000P0D
o`001@0000000?lX00000000ofH0000000800?nS00D00000003oh@0000000>7o00000002002So`04
00000000I_l006Ko0000008002So00<0000001@0o`0D0?l0200D0?l01@00000002So000Xo`00:?l0
0000008006Ko00<00000002So`0000000P00hOl01P0000000?oQ00000000oj<00000003oIP800000
0P0Do`001@000000DOl00000002?o`0000000080c?l000<000000?oe003om@000P3om@000`000000
okP00?nh00020?nh00030000003om@00ooD00080ooD001X000000<co003<o`00000008oo00000000
DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o00000000Xol0
0000001Vo`00000002So000000<0503o00<000000540o`1A0?l01`1A0?l01@000000503o01@0o`0D
0?l00000008002So00H00000001Vo`0000000:?o00000000hOl20000000<003oX`0000000?mV0000
0000obP0000001Co00000000DOl00000002?o`0000000P3<o`000`000000ooD00?oe00020?oe0003
0000003o^000okP00080okP000<000000?oe003om@000P3om@0010000000c?l00<co0000000208oo
000A0000001Ao`00000001Co000000000?lX00000000ofH00000003oX`0000000?oQ00000000hOl0
0000002So`0000000P00I_l01@00000002So000Xo`00:?l0000000X0503o00D00000000Xo`00:?l0
02So00000002001Vo`0500000000Xol00000003Qo`0000000P00on401@0000000?nS00000000ofH0
000000800?lX00<0000001Co000000000`1Ao`000`000000Sol008oo000208oo00030000003<o`00
c?l000D0c?l00P00000308oo00030000001Ao`00DOl0008000000P0Do`000`0000000?lX00000002
003oIP0400000000oj<00?nS000000800?oQ0P000002003Qo`8000001000Xol00`00000006Ko001V
o`05001Vo`0300000000Xol00:?o00@00:?o00D00000003Qo`00hOl00>7o00000003003oh@050000
0000oj<00?nS003oX`0000000P00ofH70?ooo`4000000@3oool00200oooo00<000000?ooo`3oool0
1@3oool0100000000?mV003oIP000003003oX`0500000000on400?oQ003oh@0000000`00hOl00`00
00000:?o002So`04002So`0300000000I_l006Ko00D006Ko00<00000002So`00Xol00`00Xol01@00
00000>7o003Qo`00hOl0000000800?oQ00<00000003oX`0000000`00ofH0100000000?lX003o:000
000201Co00040000001Ao`00DOl0000000D0Sol000<000000<co003<o`000`3<o`000`000000Sol0
08oo000308oo00040000001Ao`00DOl0000000805?l000<00000003o:00000000P00ofH01`000000
0?nS00000000on400000003Qo`0000000P00Xol01000000006Ko001Vo`000003000Xo`030000000D
0?l0503o00L0503o00D00000000Xo`00:?l002So00000002001Vo`0300000000Xol0000000800>7o
00H00000003oh@0000000?nS00000000ofH2000000805?l000D00000057o00000000Sol000000002
0<co00030000003om@00ooD000<0ooD00P0000000`3o^00000000?oe00040?oe000J0000003<o`00
c?l00000002?o`000000057o000000005?l000000000obP00000003oIP0000000?nS00000000on40
0000003Qo`0000000:?o00000000I_l00000000Xo`00000301@0o`030000001A0?l0D@3o00L0D@3o
00D0000001@0o`0D0?l0503o00000002000Xo`0600000000I_l00000002So`0000000>7o0P000000
3000oj<00000003oIP0000000?lX0000000Do`000000057o00000000Sol000000080c?l000<00000
0?oe003om@000`3om@0010000000okP00000000000040?oe000C0000003<o`00c?l00<co00000000
Sol00000001Ao`00000001Co000000000?lX00000000ofH00000003oX`0000000?oQ00000002003Q
o`0300000000Xol00000008006Ko00D00000000Xo`00:?l002So0000000901@0o`0500000000:?l0
02So000Xo`0000000P00I_l00`0000000:?o00000002003Qo`0500000000on400000003oX`000000
0P00ofH01@0000000?lX0000000Do`00000000<0DOl000<0000008oo002?o`000P2?o`000`000000
c?l00<co00050<co00030000002?o`00Sol000<0Sol000<00000057o000000000`0Do`000`000000
0?lX00000002003oIP0400000000oj<00?nS000000800?oQ00D00000003Qo`00hOl00>7o00000005
002So`0300000000I_l006Ko00D006Ko00<00000002So`00Xol01000Xol01@0000000>7o003Qo`00
hOl0000000<00?oQ00D00000003oX`00oj<00?nS00000003003oIP030000003oool0oooo00@0oooo
0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`05003o:00000000?mV003oIP000000
0`00oj<01@0000000?oQ003oh@00on40000000<00>7o00<00000002So`00Xol01000Xol00`000000
06Ko001Vo`05001Vo`0300000000Xol00:?o00800:?o00D00000003Qo`00hOl00>7o00000002003o
h@0400000000oj<00?nS000000800?mV00@00000003o:000obP000000P0Do`001@000000DOl0057o
001Ao`00000000@0Sol000<000000<co003<o`00103<o`000`000000Sol008oo000308oo00040000
001Ao`00DOl0000000805?l000D00000003o:00000000?mV00000002003oX`0500000000on400000
003Qo`0000000P00Xol01000000006Ko001Vo`000003000Xo`030000000D0?l0503o00H0503o00<0
0000000Xo`00:?l00P00:?l00`00000006Ko00000002002So`0?00000000hOl00000003oh@000000
0?nS00000000ofH00000003o:00000005?l00000001Ao`0000000080Sol000@000000<co003<o`00
00001P3om@02000000D0ooD001X000000<co003<o`00000008oo00000000DOl00000000Do`000000
003o:00000000?mV00000000oj<00000003oh@0000000>7o00000000Xol00000001Vo`00000002So
000000<0503o00<000000540o`1A0?l01P1A0?l01@000000503o01@0o`0D0?l00000008002So00<0
0000001Vo`00I_l00P000002003Qo`80000000<00?nS00000000ofH00P0000002@0Do`000000057o
001Ao`00000008oo00000000c?l0000000070?oe00030000003om@00ooD000@0ooD000@000000<co
003<o`0000000P2?o`004@000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000
003oh@0000000>7o00000000Xol00000008006Ko00D00000000Xo`00:?l002So0000000901@0o`03
00000000:?l002So008002So00<00000001Vo`0000000P00Xol01@0000000>7o00000000on400000
00800?nS00<00000003oIP00ofH00P00000201Co00050000001Ao`00DOl0057o00000000102?o`00
0`000000c?l00<co00040<co00030000002?o`00Sol000<0Sol000<00000057o000000000`0Do`00
100000000?lX003o:0000002003oIP0300000000oj<0000000800?oQ00D00000003Qo`00hOl00>7o
00000006002So`0300000000I_l006Ko00@006Ko00<00000002So`00Xol01000Xol01@0000000>7o
003Qo`00hOl0000000<00?oQ00D00000003oX`00oj<00?nS00000003003oIP030000003oool0oooo
00D0oooo0@0000010?ooo`00803oool00`000000oooo0?ooo`050?ooo`800?lX00<00000003oIP00
ofH00P000002003oX`0500000000on400?oQ003oh@0000000`00hOl00`0000000:?o002So`04002S
o`0300000000I_l006Ko00@006Ko00<00000002So`00Xol00`00Xol01@0000000>7o003Qo`00hOl0
000000800?oQ00@00000003oX`00oj<000000P00ofH0100000000?lX003o:000000201Co00040000
001Ao`00DOl0000000D0Sol000<000000<co003<o`000`3<o`000`000000Sol008oo000308oo0004
0000001Ao`00DOl0000000805?l000<00000003o:00000000P00ofH01`0000000?nS00000000on40
0000003Qo`0000000P00Xol01000000006Ko001Vo`000003000Xo`030000000D0?l0503o00L0503o
00D00000000Xo`00:?l002So00000002001Vo`0300000000Xol0000000800>7o0P000002003oX`80
000000@00?lX0000000Do`0000000P1Ao`000`000000Sol0000000020<co00030000003om@00ooD0
00`0ooD000<000000<co000000000P2?o`001P000000DOl00000000Do`000000003o:080000000X0
0?nS00000000on400000003Qo`0000000:?o00000000I_l000000P00:?l01@000000503o01@0o`0D
0?l0000000L0D@3o00D0000001@0o`0D0?l0503o00000002000Xo`0<00000000I_l00000002So`00
00000>7o00000000on400000003oX`0000000?mV0P0000002@0Do`000000057o001Ao`00000008oo
00000000c?l00000000>0?oe000B0000003<o`00c?l00000002?o`000000057o000000005?l00000
0000obP00000003oIP0000000?nS00000000on4000000P00hOl00`0000000:?o00000002001Vo`05
00000000:?l002So000Xo`000000200D0?l00`00000002So000Xo`02000Xo`0600000000I_l006Ko
00000000Xol000000P00hOl01@0000000?oQ00000000oj<0000000800?mV00D00000003o:0000000
5?l000000003057o00030000002?o`00Sol00080Sol000<000000<co003<o`00103<o`000`000000
Sol008oo000308oo00040000001Ao`00DOl0000000805?l000@00000003o:000obP000000P00ofH0
100000000?nS003oX`000002003oh@0500000000hOl00>7o003Qo`0000001@00Xol00`00000006Ko
001Vo`04001Vo`0300000000Xol00:?o00@00:?o00D00000003Qo`00hOl00>7o00000003003oh@05
00000000oj<00?nS003oX`0000000`00ofH00`0000000?lX0?ooo`060?ooo`4000000@3oool00200
oooo00<000000?ooo`3oool01@3oool3003o:00500000000ofH00?mV003oIP0000000P00oj<01@00
00000?oQ003oh@00on40000000<00>7o00<00000002So`00Xol01000Xol00`00000006Ko001Vo`04
001Vo`0300000000Xol00:?o00<00:?o0P000002003Qo`0400000000on400?oQ000000800?nS00@0
0000003oIP00ofH000000P00obP0100000005?l001Co00000002057o00030000002?o`00Sol000<0
Sol000<000000<co003<o`000`3<o`000`000000Sol008oo000308oo00040000001Ao`00DOl00000
00805?l000<00000003o:00000000P00ofH01`0000000?nS00000000on400000003Qo`0000000P00
Xol01000000006Ko001Vo`000003000Xo`030000000D0?l0503o00H0503o00<00000000Xo`00:?l0
0P00:?l00`00000006Ko00000002002So`0?00000000hOl00000003oh@0000000?nS00000000ofH0
0000003o:00000005?l00000001Ao`0000000080Sol000<000000<co000000003P3om@000`000000
c?l00000000208oo00060000001Ao`00000001Co000000000?lX0P0000002P00oj<00000003oh@00
00000>7o00000000Xol00000001Vo`000002000Xo`050000000D0?l0503o01@0o`0000001`1A0?l0
1@000000503o01@0o`0D0?l00000008002So00`00000001Vo`0000000:?o00000000hOl00000003o
h@0000000?nS00000000ofH20000000901Co00000000DOl0057o00000000Sol00000003<o`000000
00h0ooD001@000000<co003<o`00000008oo00000000DOl00000000Do`000000003o:00000000?mV
00000000oj<00000003oh@0000000>7o000000800:?o00@00000001Vo`00I_l000000`00:?l00`00
0000503o01@0o`0601@0o`0500000000:?l002So000Xo`0000000P00I_l0200000000:?o002So`00
00000>7o00000000on4000000P00oj<01@0000000?mV00000000obP0000000805?l000@00000057o
001Ao`0000001@2?o`000`000000c?l00<co00030<co00030000002?o`00Sol000<0Sol000@00000
057o001Ao`0000000P0Do`0200000005003o:00000000?mV003oIP0000000P00oj<0100000000?oQ
003oh@000003003Qo`0300000000Xol00:?o00<00:?o00<00000001Vo`00I_l01000I_l00`000000
0:?o002So`04002So`0500000000hOl00>7o003Qo`0000000`00on401@0000000?nS003oX`00oj<0
000000<00?mV00<00000003o:000obP01`3oool100000040oooo000P0?ooo`030000003oool0oooo
00D0oooo0`00obP01@0000000?mV003oIP00ofH0000000<00?nS00D00000003oh@00on400?oQ0000
0003003Qo`0300000000Xol00:?o00@00:?o00<00000001Vo`00I_l01000I_l00`0000000:?o002S
o`04002So`0400000000hOl00>7o000000800?oQ00@00000003oX`00oj<000000P00ofH010000000
0?lX003o:000000201Co00040000001Ao`00DOl0000000D0Sol000<000000<co003<o`000P3<o`00
0`000000Sol008oo000308oo00040000001Ao`00DOl0000000805?l000`00000003o:000obP00000
003oIP0000000?nS00000000on400000003Qo`000002002So`0400000000I_l006Ko000000@002So
00<0000001@0o`0D0?l01@0D0?l00`00000002So000Xo`02000Xo`0300000000I_l0000000800:?o
00l00000003Qo`0000000?oQ00000000oj<00000003oIP0000000?lX0000000Do`000000057o0000
00000P2?o`000`000000c?l00000000>0?oe00030000003<o`0000000080Sol000H00000057o0000
00005?l000000000obP20000000:003oX`0000000?oQ00000000hOl00000002So`00000006Ko0000
008002So00D0000001@0o`0D0?l0503o000000070540o`050000000D0?l0503o01@0o`0000000P00
:?l03000000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000003oIP80000000T0
5?l00000001Ao`00DOl00000002?o`0000000<co000000003@3om@001P000000c?l00<co00000000
Sol000000080DOl000d0000001Co000000000?lX00000000ofH00000003oX`0000000?oQ00000000
hOl0000000800:?o00@00000001Vo`00I_l000000`00:?l00`000000503o01@0o`0501@0o`050000
0000:?l002So000Xo`0000000P00I_l01P0000000:?o002So`0000000>7o000000800?oQ00D00000
003oX`0000000?mV00000002003o:0030000000Do`0000000080DOl000<0000008oo002?o`001@2?
o`001@000000c?l00<co003<o`00000000D0Sol000L00000057o001Ao`00DOl00000000Do`000000
00800?lX00@00000003oIP00ofH000000P00oj<0100000000?oQ003oh@000003003Qo`0300000000
Xol00:?o00<00:?o00<00000001Vo`00I_l01000I_l00`0000000:?o002So`04002So`0500000000
hOl00>7o003Qo`0000000`00on401@0000000?nS003oX`00oj<0000000<00?mV00@00000003o:000
obP00?lX1`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo00D00000003o:000
obP00?lX00000003003oIP0500000000oj<00?nS003oX`0000000`00on401@0000000>7o003Qo`00
hOl0000000H00:?o00<00000001Vo`00I_l00`00I_l00`0000000:?o002So`04002So`0400000000
hOl00>7o000000<00?oQ00H00000003oX`00oj<00000003oIP000002003o:0050000000Do`005?l0
01Co000000000P1Ao`000`000000Sol008oo000308oo00030000003<o`00c?l00080c?l000<00000
08oo002?o`000`2?o`0010000000DOl0057o0000000201Co000300000000obP0000000800?mV00L0
0000003oX`0000000?oQ00000000hOl0000000800:?o00@00000001Vo`00I_l000000`00:?l00`00
0000503o01@0o`0501@0o`0300000000:?l002So00<002So01400000001Vo`0000000:?o00000000
hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l000000002057o00030000002?
o`0000000080c?l000<000000?oe003om@002P3om@0010000000c?l00<co0000000208oo00060000
001Ao`00000001Co000000000?lX0P0000002P00oj<00000003oh@0000000>7o00000000Xol00000
001Vo`000002000Xo`030000000D0?l0503o0080503o00<000000540o`1A0?l00`1A0?l00`000000
503o01@0o`0201@0o`0300000000:?l00000008006Ko00H00000002So`0000000>7o00000000on42
00000008003oIP0000000?lX0000000Do`000000057o0000000208oo00030000003<o`00000000d0
ooD001@000000<co003<o`00000008oo00000000DOl00000000Do`000000003o:00000000?mV0000
0000oj<00000003oh@0000000>7o000000800:?o00@00000001Vo`00I_l000000`00:?l00`000000
503o01@0o`0501@0o`0300000000:?l002So008002So00@00000001Vo`00I_l000000P00Xol01`00
00000>7o00000000on400000003oX`0000000P00ofH01P0000000?lX003o:00000005?l000000080
DOl000<0000008oo002?o`000`2?o`001P000000Sol008oo00000000c?l0000000H0Sol000H00000
057o001Ao`00000001Co00000002003o:00400000000ofH00?mV000000800?nS00@00000003oh@00
on4000000`00hOl00`0000000:?o002So`04002So`0300000000I_l006Ko00@006Ko00<00000002S
o`00Xol00`00Xol01@0000000>7o003Qo`00hOl0000000<00?oQ00D00000003oX`00oj<00?nS0000
0003003oIP0300000000obP00?lX00800?lX00<000000?ooo`3oool0103oool100000040oooo000P
0?ooo`030000003oool0oooo00D0oooo00<05?l000000000obP00P00obP2000000800?mV0P000002
003oX`0500000000on400?oQ003oh@0000000`00hOl00`0000000:?o002So`04002So`0300000000
I_l006Ko00<006Ko00<00000002So`00Xol01000Xol0100000000>7o003Qo`000002003oh@040000
0000oj<00?nS000000800?mV00@00000003o:000obP000000P0Do`0010000000DOl0057o00000006
08oo00050000003<o`00c?l00<co000000001@2?o`0010000000DOl0057o0000000201Co000:0000
0000obP00?lX00000000ofH00000003oX`0000000?oQ000000800>7o00<00000002So`0000000P00
I_l00`00000002So000Xo`02000Xo`030000000D0?l0503o00D0503o00<00000000Xo`00:?l00P00
:?l00`00000006Ko00000002002So`0?00000000hOl00000003oh@0000000?nS00000000ofH00000
003o:00000005?l00000001Ao`0000000080Sol000<000000<co00000000303om@006@000000c?l0
0<co003<o`00000008oo00000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000
003oh@0000000>7o00000000Xol00000001Vo`0000000P00:?l01@000000503o01@0o`0D0?l00000
00D0D@3o00<0000001@0o`0D0?l00P0D0?l00`00000002So00000002001Vo`0600000000Xol00000
003Qo`0000000?oQ0P0000002000ofH00000003o:00000005?l00000001Ao`0000000P2?o`001000
0000c?l00<co0000000;0?oe00040000003<o`00c?l000000080Sol000l00000057o000000005?l0
00000000obP00000003oIP0000000?nS00000000on400000003Qo`0000000P00Xol01000000006Ko
001Vo`000003000Xo`030000000D0?l0503o00D0503o00D00000000Xo`00:?l002So00000002001V
o`0600000000Xol00:?o00000000hOl000000P00on401@0000000?nS00000000ofH0000000800?lX
00<0000001Co000000000P1Ao`000`000000Sol008oo000308oo00060000003<o`0000000000002?
o`0000001P2?o`0010000000DOl0057o0000000201Co000600000000obP00?lX00000000ofH00000
0`00oj<0100000000?oQ003oh@000003003Qo`0300000000Xol00:?o00<00:?o00<00000001Vo`00
I_l01000I_l00`0000000:?o002So`03002So`0500000000hOl00>7o003Qo`0000000`00on401@00
00000?nS003oX`00oj<0000000<00?mV00@00000003o:000obP00?lX0P0000070?ooo`4000000@3o
ool00200oooo00<000000?ooo`3oool01@3oool201Co000300000000obP00?lX00800?lX00D00000
003oIP00ofH00?mV00000002003oX`0500000000on400?oQ003oh@0000000`00hOl00`0000000:?o
002So`04002So`0300000000I_l006Ko00<006Ko0P000005002So`0400000000hOl00>7o00000080
0?oQ00@00000003oX`00oj<000000P00ofH0100000000?lX003o:000000201Co00040000001Ao`00
DOl0000000H0Sol000D000000<co003<o`00c?l00000000508oo00040000001Ao`00DOl000000080
5?l000X00000003o:000obP00000003oIP0000000?nS00000000on4000000P00hOl00`0000000:?o
00000002001Vo`0300000000:?l002So008002So00<0000001@0o`0D0?l0100D0?l00`00000002So
000Xo`02000Xo`0B00000000I_l006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000
003oIP0000000?lX0000000Do`0000000P1Ao`000`000000Sol0000000020<co00030000003om@00
ooD000T0ooD001T000000<co003<o`00c?l00000002?o`000000057o000000005?l000000000obP0
0000003oIP0000000?nS00000000on400000003Qo`0000000:?o00000000I_l00000008002So00<0
000001@0o`0D0?l00P0D0?l00`000000D@3o0540o`020540o`030000000D0?l0503o0080503o00<0
0000000Xo`0000000P00I_l01P0000000:?o00000000hOl00000003oh@80000000P00?mV00000000
obP0000001Co00000000DOl000000080Sol000@000000<co003<o`0000002`3om@0010000000c?l0
0<co0000000208oo000?0000001Ao`00000001Co000000000?lX00000000ofH00000003oX`000000
0?oQ00000000hOl0000000800:?o00<00000001Vo`0000001000:?l00`000000503o01@0o`0401@0
o`0300000000:?l002So008002So00@00000001Vo`00I_l000000P00Xol01`0000000>7o00000000
on400000003oX`0000000P00ofH00`0000000?lX0000000201Co00040000001Ao`00DOl0000000D0
Sol000D000000<co003<o`00c?l00000000608oo00040000001Ao`00DOl0000000805?l000@00000
003o:000obP000000P00ofH0100000000?nS003oX`000002003oh@0500000000hOl00>7o003Qo`00
00001@00Xol00`00000006Ko001Vo`04001Vo`0300000000Xol00:?o00<00:?o00D00000003Qo`00
hOl00>7o00000003003oh@0500000000oj<00?nS003oX`0000000`00ofH01@0000000?lX003o:000
obP0000000805?l01`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo0`0Do`00
1@0000000?lX003o:000obP0000000<00?mV00D00000003oX`00oj<00?nS00000003003oh@050000
0000hOl00>7o003Qo`0000001P00Xol00`00000006Ko001Vo`03001Vo`0300000000Xol00:?o00<0
0:?o00D00000003Qo`00hOl00>7o00000002003oh@0400000000oj<00?nS000000800?mV00@00000
003o:000obP000000P0Do`0010000000DOl0057o0000000608oo00040000003<o`00c?l0000000D0
Sol000L00000057o001Ao`00DOl00000000Do`00000000800?lX00L00000003oIP0000000?nS0000
0000on40000000800>7o00<00000002So`0000000`00I_l00`00000002So000Xo`02000Xo`030000
000D0?l0503o00<0503o00<00000000Xo`00:?l00P00:?l04P00000006Ko001Vo`0000000:?o0000
0000hOl00000003oh@0000000?nS00000000ofH00000003o:00000005?l000000080DOl000<00000
08oo000000000`3<o`000`000000ooD00?oe00080?oe000I0000003<o`00c?l00<co00000000Sol0
0000001Ao`00000001Co000000000?lX00000000ofH00000003oX`0000000?oQ00000000hOl00000
002So`00000006Ko00000002000Xo`030000000D0?l0503o0080503o1@00000501@0o`0300000000
:?l00000008006Ko00H00000002So`0000000>7o00000000on4200000008003oIP0000000?lX0000
000Do`000000057o0000000208oo00040000003<o`00c?l0000000/0ooD000<000000<co00000000
0P2?o`0040000000DOl0057o000000005?l000000000obP00000003oIP0000000?nS00000000on40
0000003Qo`000002002So`0300000000I_l0000000@002So00<0000001@0o`0D0?l00`0D0?l00`00
000002So000Xo`02000Xo`0400000000I_l006Ko000000800:?o00D00000003Qo`0000000?oQ0000
0002003oX`0500000000ofH00000003o:00000000`0Do`000`000000DOl00000000608oo00040000
003<o`00c?l0000000H0Sol000@00000057o001Ao`0000000`0Do`000`0000000?lX00000002003o
IP0400000000oj<00?nS000000800?oQ00D00000003Qo`00hOl00>7o00000005002So`0300000000
I_l006Ko00@006Ko00<00000002So`00Xol00`00Xol01@0000000>7o003Qo`00hOl0000000<00?oQ
00D00000003oX`00oj<00?nS00000003003oIP0500000000obP00?lX003o:00000000`0Do`070?oo
o`4000000@3oool00200oooo00<000000?ooo`3oool01@3oool401Co000500000000obP00?lX003o
:00000000`00ofH01@0000000?nS003oX`00oj<0000000<00?oQ00D00000003Qo`00hOl00>7o0000
0006002So`0300000000I_l006Ko00<006Ko00<00000002So`00Xol00`00Xol01@0000000>7o003Q
o`00hOl0000000800?oQ00H00000003oX`00oj<00000003oIP000003003o:0060000000Do`005?l0
0000001Ao`0000001`2?o`02000000L0Sol000@00000057o001Ao`0000000P0Do`002P0000000?lX
003o:00000000?mV00000000oj<00000003oh@000002003Qo`0300000000Xol00000008006Ko00<0
0000000Xo`00:?l00P00:?l00`000000503o01@0o`0301@0o`0300000000:?l002So008002So00H0
0000001Vo`00I_l00000002So`000002003Qo`0=00000000on400000003oX`0000000?mV00000000
obP0000001Co00000000DOl00000000208oo00040000003<o`00c?l0000000X0ooD001T000000<co
003<o`00c?l00000002?o`000000057o000000005?l000000000obP00000003oIP0000000?nS0000
0000on400000003Qo`0000000:?o00000000I_l00000008002So00<0000001@0o`0D0?l02`0D0?l0
6000000002So000Xo`00000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000003o
IP0000000?lX0000000Do`000000057o00000000Sol0000000<0c?l000<000000?oe003om@00203o
m@0010000000c?l00<co0000000208oo000?0000001Ao`00000001Co000000000?lX00000000ofH0
0000003oX`0000000?oQ00000000hOl0000000800:?o00<00000001Vo`0000001@00:?l00`000000
503o01@0o`0201@0o`0300000000:?l002So00<002So00H00000001Vo`00I_l00000002So`000002
003Qo`0500000000on400000003oX`0000000P00ofH00`0000000?lX0000000201Co00040000001A
o`00DOl0000000L0Sol00P00000708oo00040000001Ao`00DOl0000000805?l000<00000003o:000
00000P00ofH0100000000?nS003oX`000002003oh@0500000000hOl00>7o003Qo`0000001@00Xol0
0`00000006Ko001Vo`04001Vo`0300000000Xol00:?o00<00:?o00D00000003Qo`00hOl00>7o0000
0003003oh@0500000000oj<00?nS003oX`0000000`00ofH01@0000000?lX003o:000obP0000000@0
5?l01`3oool100000040oooo000P0?ooo`030000003oool0oooo00D0oooo1@0Do`00100000000?lX
003o:000obP2000000800?mV00D00000003oX`00oj<00?nS00000003003oh@0500000000hOl00>7o
003Qo`0000001@00Xol00`00000006Ko001Vo`03001Vo`0300000000Xol00:?o00@00:?o00@00000
003Qo`00hOl000000P00on401`0000000?nS003oX`00oj<00000003oIP0000000P00obP010000000
5?l001Co00000002057o00030000002?o`00Sol000D0Sol000<0000008oo002?o`001@2?o`001`00
0000DOl0057o001Ao`00000001Co000000000P00obP01`0000000?mV00000000oj<00000003oh@00
00000P00hOl00`0000000:?o00000003001Vo`0300000000:?l002So008002So00<0000001@0o`0D
0?l00P0D0?l00`00000002So000Xo`03000Xo`0300000000I_l0000000800:?o00d00000003Qo`00
00000?oQ00000000oj<00000003oIP0000000?lX0000000Do`0000000080DOl000<0000008oo0000
00000`3<o`000`000000ooD00?oe00070?oe00070000003<o`00c?l00<co00000000Sol000000002
057o00040000000Do`000000003o:080000000X00?nS00000000on400000003Qo`0000000:?o0000
0000I_l000000P00:?l00`000000503o01@0o`0;01@0o`0H00000000:?l002So00000000I_l00000
002So`0000000>7o00000000on400000003oX`0000000?mV00000000obP0000001Co00000000DOl0
0000002?o`0000000`3<o`000`000000ooD00?oe00080?oe00060000003<o`00c?l00000002?o`00
00000P1Ao`003@0000005?l000000000obP00000003oIP0000000?nS00000000on400000003Qo`00
00000P00Xol00`00000006Ko00000005000Xo`030000000D0?l0503o0080503o00<00000000Xo`00
:?l00P00:?l01000000006Ko001Vo`000002002So`0500000000hOl00000003oh@0000000P00oj<0
1@0000000?mV00000000obP0000000<05?l000<00000057o00000000202?o`000`000000Sol008oo
000508oo00040000001Ao`00DOl0000000<05?l000D00000003o:00000000?mV00000002003oX`05
00000000on400?oQ003oh@0000000P00hOl00`0000000:?o002So`04002So`0300000000I_l006Ko
00<006Ko00<00000002So`00Xol01000Xol01@0000000>7o003Qo`00hOl0000000800?oQ00D00000
003oX`00oj<00?nS00000003003oIP0500000000obP00?lX003o:00000001@0Do`070?ooo`400000
0@3oool00200oooo00<000000?ooo`3oool01@3oool601Co000300000000obP00?lX00800?lX00@0
0000003oIP00ofH000000`00oj<01@0000000?oQ003oh@00on40000000<00>7o00<00000002So`00
Xol00`00Xol00`00000006Ko001Vo`03001Vo`0300000000Xol00:?o00@00:?o00@00000003Qo`00
hOl000000P00on40100000000?nS003oX`000002003oIP0400000000obP00?lX000000805?l000@0
0000057o001Ao`0000003`2?o`0010000000DOl0057o0000000201Co000:00000000obP00?lX0000
0000ofH00000003oX`0000000?oQ000000800>7o00<00000002So`0000000P00I_l00`00000002So
000Xo`03000Xo`030000000D0?l0503o008000001@00:?l01P00000006Ko001Vo`0000000:?o0000
00800>7o00l00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`000000
08oo000000000`3<o`000`000000ooD00?oe00070?oe00070000003<o`00c?l00<co00000000Sol0
00000002057o00040000000Do`000000003o:080000000T00?nS00000000on400000003Qo`000000
0:?o00000000I_l00P0000000`00:?l0000001@0o`0<01@0o`0H00000000:?l002So00000000I_l0
0000002So`0000000>7o00000000on400000003oX`0000000?mV00000000obP0000001Co00000000
DOl00000002?o`0000000`3<o`000`000000ooD00?oe00070?oe00070000003<o`00c?l00<co0000
0000Sol000000002057o000?0000000Do`000000003o:00000000?mV00000000oj<00000003oh@00
00000>7o00000000Xol00000008006Ko00<00000000Xo`00:?l00`00:?l200000080503o00<00000
000Xo`00:?l00`00:?l01P00000006Ko001Vo`0000000:?o000000800>7o00D00000003oh@000000
0?nS00000002003oIP0300000000obP0000000805?l000@00000057o001Ao`0000003`2?o`001@00
0000DOl0057o001Ao`00000000805?l000<00000003o:00000000P00ofH0100000000?nS003oX`00
0002003oh@0400000000hOl00>7o000000H00:?o00<00000001Vo`00I_l00`00I_l00`0000000:?o
002So`04002So`0500000000hOl00>7o003Qo`0000000P00on401@0000000?nS003oX`00oj<00000
00<00?mV00D00000003o:000obP00?lX0000000601Co00L0oooo0@0000010?ooo`00803oool00`00
0000oooo0?ooo`050?ooo`L05?l000D00000003o:000obP00?lX00000003003oIP0500000000oj<0
0?nS003oX`0000000`00on401@0000000>7o003Qo`00hOl0000000D00:?o00<00000001Vo`00I_l0
0`00I_l00`0000000:?o002So`03002So`0500000000hOl00>7o003Qo`0000000P00on4010000000
0?nS003oX`000002003oIP0600000000obP00?lX0000000Do`0000000`1Ao`000`000000Sol008oo
000<08oo00050000001Ao`00DOl0057o000000000P0Do`002@0000000?lX00000000ofH00000003o
X`0000000?oQ00000002003Qo`0400000000Xol00:?o0000008006Ko00<00000000Xo`00:?l00`00
:?l2000000L002So00H00000001Vo`00I_l00000002So`000002003Qo`0?00000000on400000003o
X`0000000?mV00000000obP0000001Co00000000DOl00000002?o`00000000<0c?l000<000000?oe
003om@001`3om@001`000000c?l00<co003<o`00000008oo000000000P1Ao`00100000005?l00000
0000obP200000008003oX`0000000?oQ00000000hOl00000002So`000002001Vo`0300000000:?l0
000000d0503o01P00000000Xo`00:?l00000001Vo`0000000:?o00000000hOl00000003oh@000000
0?nS00000000ofH00000003o:00000005?l00000001Ao`00000008oo000000030<co00030000003o
m@00ooD000H0ooD000D000000<co003<o`00c?l00000000208oo00030000001Ao`00000000805?l0
0P000002003oIP8000000P00on401@0000000>7o00000000Xol00000008006Ko00<00000000Xo`00
:?l01@00:?l2000000D002So00@00000001Vo`00I_l000000P00Xol01`0000000>7o00000000on40
0000003oX`0000000P00ofH01P0000000?lX003o:00000005?l000000080DOl000<0000008oo002?
o`003@2?o`001@000000DOl0057o001Ao`00000000805?l000H00000003o:000obP00000003oIP00
0002003oX`0400000000on400?oQ000000800>7o00<00000002So`00Xol01000Xol00`00000006Ko
001Vo`03001Vo`0300000000Xol00:?o00@00:?o00<00000003Qo`00hOl00P000002003oh@050000
0000oj<00?nS003oX`0000000`00ofH01@0000000?lX003o:000obP0000000L05?l01`3oool10000
0040oooo000P0?ooo`030000003oool0oooo00D0oooo200Do`001@0000000?lX003o:000obP00000
00<00?mV00D00000003oX`00oj<00?nS00000003003oh@0500000000hOl00>7o003Qo`0000001@00
Xol00`00000006Ko001Vo`02001Vo`0300000000Xol00:?o00@00:?o00D00000003Qo`00hOl00>7o
00000002003oh@0300000000oj<0000000<00?mV00H00000003o:000obP0000001Co00000002057o
00030000002?o`00Sol000d0Sol000@00000057o001Ao`0000000P0Do`002P0000000?lX003o:000
00000?mV00000000oj<00000003oh@000002003Qo`0300000000Xol00000008006Ko00<00000000X
o`00:?l03000:?l01000000006Ko001Vo`000002002So`0?00000000hOl00000003oh@0000000?nS
00000000ofH00000003o:00000005?l00000001Ao`0000000080Sol000D000000<co003<o`00c?l0
000000080?oe00070000003<o`00c?l00<co00000000Sol000000002057o00040000000Do`000000
003o:080000000P00?nS00000000on400000003Qo`0000000:?o0000008006Ko00<00000000Xo`00
0000300D0?l01000000002So000Xo`000002001Vo`0800000000Xol00000003Qo`0000000?oQ0000
0000oj<200000004003o:00000005?l000000080DOl000<0000008oo000000000`3<o`000`000000
ooD00?oe00060?oe00070000003<o`00c?l00<co00000000Sol000000002057o000?0000000Do`00
0000003o:00000000?mV00000000oj<00000003oh@0000000>7o00000000Xol00000008006Ko00<0
0000000Xo`00:?l03000:?l01`00000006Ko001Vo`00I_l00000002So`0000000P00hOl01@000000
0?oQ00000000oj<0000000800?mV00<00000003o:00000000P0Do`0010000000DOl0057o0000000>
08oo00050000001Ao`00DOl0057o000000000P0Do`001P0000000?lX003o:00000000?mV00000080
0?nS00@00000003oh@00on4000000`00hOl00`0000000:?o002So`03002So`0300000000I_l006Ko
00<006Ko00<00000002So`00Xol01000Xol0100000000>7o003Qo`000003003oh@0500000000oj<0
0?nS003oX`0000000`00ofH01@0000000?lX003o:000obP0000000P05?l01`3oool100000040oooo
000C0?ooo`040?oog`2LB00004RL0=ooo`T0oooo00<000000?ooo`3oool01@3oool901Co00050000
0000obP00?lX003o:00000000`00ofH01@0000000?nS003oX`00oj<0000000800?oQ00D00000003Q
o`00hOl00>7o00000006002So`0300000000I_l006Ko008006Ko00<00000002So`00Xol01000Xol0
100000000>7o003Qo`000003003oh@0300000000oj<0000000800?mV00H00000003o:000obP00000
01Co00000003057o00030000002?o`00Sol000`0Sol000D00000057o001Ao`00DOl00000000201Co
000900000000obP00000003oIP0000000?nS00000000on40000000800>7o00@00000002So`00Xol0
00000P00I_l00`00000002So000Xo`0<000Xo`0600000000I_l006Ko00000000Xol000000P00hOl2
000000800?nS0P000002003o:0070000000Do`000000057o00000000Sol0000000030<co00030000
003om@00ooD000H0ooD000D000000<co003<o`00c?l00000000208oo000<0000001Ao`00000001Co
000000000?lX00000000ofH00000003oX`0000000?oQ0P000002002So`0300000000I_l000000080
02So00<0000001@0o`0D0?l02@0D0?l01000000002So000Xo`000002001Vo`0800000000Xol00000
003Qo`0000000?oQ00000000oj<200000004003o:00000005?l000000080DOl000<0000008oo0000
00000`3<o`000`000000ooD00?oe00050?oe00050000003<o`00c?l00<co000000000P2?o`000`00
0000DOl00000000201Co000400000000obP00000003oIP8000000P00on401@0000000>7o00000000
Xol00000008006Ko00<00000000Xo`00:?l03000:?l01000000006Ko001Vo`000002002So`070000
0000hOl00000003oh@0000000?nS00000002003oIP0600000000obP00?lX0000000Do`0000000P1A
o`000`000000Sol008oo000=08oo00070000001Ao`00DOl0057o000000005?l000000002003o:006
00000000ofH00?mV00000000oj<000000P00on401@0000000>7o003Qo`00hOl0000000H00:?o00<0
0000001Vo`00I_l00P00I_l00`0000000:?o002So`04002So`0500000000hOl00>7o003Qo`000000
0`00on40100000000?nS003oX`000003003oIP0500000000obP00?lX003o:00000002@0Do`070?oo
o`4000000@3oool001<0oooo00@0oooO09a80000B9`0gooo2@3oool00`000000oooo0?ooo`050?oo
o`X05?l000D00000003o:000obP00?lX00000003003oIP0500000000oj<00?nS003oX`0000000P00
on401@0000000>7o003Qo`00hOl0000000H00:?o00<00000001Vo`00I_l00P00I_l00`0000000:?o
002So`04002So`0400000000hOl00>7o000000800?oQ00@00000003oX`00oj<000000P00ofH01@00
00000?lX003o:00000005?l0008000000P1Ao`000`000000Sol008oo000<08oo00070000001Ao`00
DOl0057o000000005?l000000002003o:00700000000ofH00000003oX`0000000?oQ00000002003Q
o`0300000000Xol0000000<006Ko00<00000000Xo`00:?l02`00:?l01P00000006Ko001Vo`000000
0:?o000000800>7o00d00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001A
o`0000000080Sol000D000000<co003<o`00c?l0000000070?oe00050000003<o`00c?l00<co0000
00000P2?o`0030000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@80
00000P00Xol00`00000006Ko00000002000Xo`030000000D0?l0503o00T0503o00@00000000Xo`00
:?l000000P00I_l0200000000:?o00000000hOl00000003oh@0000000?nS0P0000001000obP00000
01Co00000002057o00030000002?o`00000000<0c?l000<000000?oe003om@001@3om@001@000000
c?l00<co003<o`0000000080Sol000<00000057o000000000P0Do`00100000000?lX00000000ofH2
000000800?oQ00D00000003Qo`0000000:?o00000002001Vo`0300000000:?l002So00/002So00L0
0000001Vo`00I_l006Ko00000000Xol0000000800>7o00D00000003oh@0000000?nS00000002003o
IP0300000000obP0000000805?l000@00000057o001Ao`0000003P2?o`001@000000DOl0057o001A
o`00000000805?l000<00000003o:00000000P00ofH0100000000?nS003oX`000002003oh@050000
0000hOl00>7o003Qo`0000001@00Xol00`00000006Ko001Vo`02001Vo`0300000000Xol00:?o00@0
0:?o00D00000003Qo`00hOl00>7o00000003003oh@0400000000oj<00?nS000000<00?mV00D00000
003o:000obP00?lX0000000:01Co00L0oooo0@0000010?ooo`004`3oool0103ooml0W4P00018W03O
ool90?ooo`030000003oool0oooo00D0oooo2`0Do`001@0000000?lX003o:000obP0000000<00?mV
00D00000003oX`00oj<00?nS00000002003oh@0400000000hOl00>7o003Qo`8000001@00Xol01@00
000006Ko001Vo`00I_l0000000L00:?o00@00000003Qo`00hOl000000P00on40100000000?nS003o
X`000002003oIP0300000000obP0000000805?l000D00000057o001Ao`00DOl00000000=08oo0007
0000001Ao`00DOl0057o000000005?l000000002003o:00500000000ofH00000003oX`0000000P00
on400`0000000>7o00000002002So`0400000000I_l006Ko000000d002So00H00000001Vo`00I_l0
0000002So`000002003Qo`0;00000000on400000003oX`0000000?mV00000000obP0000001Co0000
00000P1Ao`000`000000Sol0000000040<co00030000003om@00ooD000@0ooD000D000000<co003<
o`00c?l00000000208oo000C0000001Ao`00000001Co000000000?lX00000000ofH00000003oX`00
00000?oQ00000000hOl00000002So`00000006Ko00000002000Xo`030000000D0?l0503o00T0503o
00@00000000Xo`00:?l000000P00I_l0200000000:?o00000000hOl00000003oh@0000000?nS0P00
00001000obP0000001Co00000002057o00030000002?o`00000000<0c?l000<000000?oe003om@00
103om@000`000000c?l00<co00020<co00030000002?o`0000000080DOl000/0000001Co00000000
0?lX00000000ofH00000003oX`0000000?oQ00000002003Qo`0300000000Xol00000008006Ko00<0
0000000Xo`00:?l02`00:?l01P00000006Ko001Vo`0000000:?o000000800>7o00<00000003oh@00
00000P00oj<01@0000000?mV00000000obP0000000805?l000D00000057o001Ao`00DOl00000000=
08oo00050000001Ao`00DOl0057o000000000P0Do`000`0000000?lX00000003003oIP0300000000
oj<0000000800?oQ00D00000003Qo`00hOl00>7o00000007002So`80000000<006Ko00000000Xol0
1@00Xol01@0000000>7o003Qo`00hOl0000000800?oQ0P000002003oX`0500000000ofH00?mV003o
IP0000000`00obP00`0000005?l001Co000901Co00L0oooo0@0000010?ooo`003`3oool00`2oM000
0000000000050000000307Boo`3oool0oooo00H0oooo00<000000?ooo`3oool01@3oool<01Co0005
00000000obP00?lX003o:00000000`00ofH01@0000000?nS003oX`00oj<0000000800?oQ00<00000
003Qo`00hOl00P00hOl00`0000000:?o002So`03002So`0500000000I_l006Ko001Vo`0000001P00
Xol01`0000000>7o003Qo`00hOl00000003oh@0000000`00oj<01P0000000?mV003oIP0000000?lX
000000805?l000@00000057o001Ao`0000003P2?o`001`000000DOl0057o001Ao`00000001Co0000
00000P00obP01`0000000?mV00000000oj<00000003oh@0000000P00hOl00`0000000:?o00000003
001Vo`0300000000:?l002So00X002So00@00000001Vo`00I_l000000P00Xol02@0000000>7o0000
0000on400000003oX`0000000?mV00000002003o:08000000P1Ao`000`000000Sol0000000050<co
00030000003om@00ooD00080ooD000<000000<co003<o`000P3<o`005P000000Sol008oo00000000
DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o00000000Xol0
0000001Vo`000002000Xo`030000000D0?l0503o00P0503o00L00000000Xo`00:?l002So00000000
I_l0000000800:?o0P000002003oh@80000000P00?mV00000000obP0000001Co00000000DOl00000
0080Sol000D000000<co003<o`00c?l0000000060?oe00050000003<o`00c?l00<co000000000P2?
o`000`000000DOl00000000201Co000400000000obP00000003oIP8000000P00on401@0000000>7o
00000000Xol0000000<006Ko00<00000000Xo`00:?l02P00:?l01`00000006Ko001Vo`00I_l00000
002So`0000000P00hOl01@0000000?oQ00000000oj<0000000800?mV00<00000003o:00000000P0D
o`001@000000DOl0057o001Ao`00000000`0Sol000D00000057o001Ao`00DOl00000000201Co0004
00000000obP00?lX000000800?mV00<00000003oX`0000000P00on401@0000000>7o003Qo`00hOl0
000000H00:?o00D00000002So`00Xol00:?o00000006002So`0300000000hOl00>7o00800>7o00<0
0000003oh@0000000`00oj<01@0000000?mV003oIP00ofH0000000<00?lX00<0000001Co000Do`00
2P0Do`070?ooo`4000000@3oool000l0oooo00P0omnL07BLg`3oool0oooo0?oog`2LB00004RL0=oo
o`T0oooo00<000000?ooo`3oool01@3oool=01Co000500000000obP00?lX003o:00000000`00ofH0
100000000?nS003oX`000003003oh@0500000000hOl00>7o003Qo`0000001P00Xol01000000006Ko
001Vo`000007002So`0700000000hOl00>7o003Qo`0000000?oQ00000002003oX`0600000000ofH0
0?mV00000000obP000000P0Do`001@000000DOl0057o001Ao`00000000d0Sol000L00000057o001A
o`00DOl00000000Do`00000000800?lX00D00000003oIP0000000?nS00000002003oh@0300000000
hOl0000000800:?o00@00000001Vo`00I_l000003@00:?l01P00000006Ko001Vo`0000000:?o0000
00800>7o00d00000003oh@0000000?nS00000000ofH00000003o:00000005?l00000001Ao`000000
0080Sol000<000000<co003<o`000P3<o`000`000000ooD00?oe00020?oe00030000003<o`00c?l0
0080c?l001@0000008oo002?o`000000057o000000005?l000000000obP00000003oIP0000000?nS
00000000on400000003Qo`0000000:?o0000008006Ko00@00000000Xo`00:?l000002@0D0?l05`00
000002So000Xo`00:?l00000001Vo`0000000:?o00000000hOl00000003oh@0000000?nS00000000
ofH00000003o:00000005?l00000001Ao`0000000080Sol000<000000<co003<o`000P3<o`000`00
0000ooD00?oe00020?oe00030000003<o`00c?l00080c?l00100000008oo002?o`000000057o0000
00005?l000000000obP00000003oIP0000000?nS00000000on4000000P00hOl00`0000000:?o0000
0003001Vo`0300000000:?l002So00X002So00H00000001Vo`00I_l00000002So`000002003Qo`03
00000000on40000000800?nS00D00000003oIP0000000?lX0000000201Co00050000001Ao`00DOl0
057o000000003@2?o`001@000000DOl0057o001Ao`00000000805?l000<00000003o:00000000P00
ofH01P0000000?nS003oX`0000000?oQ000000<00>7o00<00000002So`00Xol01000Xol00`000000
06Ko00000009002So`0500000000hOl00>7o003Qo`0000000P00on401@0000000?nS003oX`00oj<0
000000800?mV00<00000003o:000obP00P00obP00`0000005?l001Co000:01Co00L0oooo0@000001
0?ooo`00403oool01`3OW4P0M;oo0?ooo`3ooml0W4P00018W03Oool02@3oool3000000D0oooo3P0D
o`001@0000000?lX003o:000obP0000000<00?mV00@00000003oX`00oj<000000`00on401@000000
0>7o003Qo`00hOl0000000H00:?o00@00000001Vo`00I_l000001P00Xol01@0000000>7o003Qo`00
hOl0000000800?oQ00H00000003oX`00oj<00000003oIP000002003o:0040000000Do`005?l00000
00<0DOl000<0000008oo002?o`002`2?o`0010000000DOl0057o0000000201Co000:00000000obP0
0?lX00000000ofH00000003oX`0000000?oQ000000800>7o00@00000002So`00Xol000000P00I_l0
0`00000002So000Xo`0:000Xo`0400000000I_l006Ko000000800:?o00l00000003Qo`0000000?oQ
00000000oj<00000003oIP0000000?lX0000000Do`000000057o000000000P2?o`000`000000c?l0
0<co00020<co00030000003om@00ooD00080ooD000<000000<co003<o`000P3<o`0050000000Sol0
08oo00000000DOl00000000Do`000000003o:00000000?mV00000000oj<00000003oh@0000000>7o
00000000Xol000000P00I_l01000000002So000Xo`00000901@0o`0G00000000:?l002So000Xo`00
000006Ko00000000Xol00000003Qo`0000000?oQ00000000oj<00000003oIP0000000?lX0000000D
o`000000057o000000000P2?o`000`000000c?l00<co00020<co00030000003om@00ooD00080ooD0
00<000000<co003<o`000P3<o`0040000000Sol008oo00000000DOl00000000Do`000000003o:000
00000?mV00000000oj<00000003oh@000002003Qo`0300000000Xol0000000<006Ko00<00000000X
o`00:?l02P00:?l01P00000006Ko001Vo`0000000:?o000000800>7o00D00000003oh@0000000?nS
00000002003oIP0300000000obP0000000805?l000@00000057o001Ao`0000003@2?o`001@000000
DOl0057o001Ao`00000000805?l000H00000003o:000obP00000003oIP000002003oX`0400000000
on400?oQ000000<00>7o00<00000002So`00Xol01000Xol01000000006Ko001Vo`000007002So`05
00000000hOl00>7o003Qo`0000000P00on401@0000000?nS003oX`00oj<0000000800?mV00<00000
003o:000obP00P00obP00`0000005?l001Co000;01Co00D0oooo0`0000010?ooo`00403oool01`3o
oml0W4Qd0;ooo`3ooml0W4P00018W03Oool02@3oool00`000000oooo0?ooo`3o0?oooh`0oooo0@00
00010?ooo`004@3oool01P3ogi`0B7Bo0?oog`2LB00004RL0=ooo`T0oooo00<000000?ooo`3oool0
o`3ooon<0?ooo`4000000@3oool00180oooo00D0gia807Bog`2LB00004RL0=ooo`090?ooo`030000
003oool0oooo0?l0ooooS03oool100000040oooo000B0?ooo`050?oog`2LB4P0B4P00018W03Oool0
2@3oool00`000000oooo0?ooo`3o0?oooh`0oooo0@0000010?ooo`004`3oool0103ogi`0B0000018
W03Oool90?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0G03oool00`000000oooo
0?ooo`1M0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool0G@3oool00`000000oooo
0?ooo`040?ooo`4000000@3oool001@0oooo00<0gia80018W03Oool02@3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0503oool00`000000oooo0?oo
o`0E0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?oo
o`0E0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?oo
o`0E0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?oo
o`0E0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?oo
o`0E0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0103oool100000040oooo000P
0?ooool00000T00000010?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3o
oona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-4.83776, -4.75573, 0.0313285, \
0.0313285}}]
}, Closed]],

Cell[TextData[{
  "You can also use ",
  StyleBox["ContourPlot", "InlineInput",
    FontWeight->"Bold"],
  " to find starting values for numerical solutions of simultaneous \
equations.  For example, suppose you wish to solve the following system of \
equations."
}], "Text"],

Cell[BoxData[
    \(\(eqs = {\(x\^3\) y\^3 \[Equal] 
            2, \(\(\ \)\(x\ y\)\)\/8 + \((x\/2)\)\^2\  + \ \((y\/3)\)\^2 \
\[Equal] 2};\)\)], "Input"],

Cell["\<\
If convert these equations to a pair of expressions which should both have \
the value zero\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(expressions = eqs /. Equal \[Rule] Subtract\)], "Input"],

Cell[BoxData[
    \({\(-2\) + x\^3\ y\^3, \(-2\) + x\^2\/4 + \(x\ y\)\/8 + 
        y\^2\/9}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "and then superimpose contour plots for each that only show the special \
contour ",
  StyleBox["Contours\[Rule]{0}", "InlineInput",
    FontWeight->"Bold"]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[
        ContourPlot[
          expressions\[LeftDoubleBracket]1\[RightDoubleBracket], {x, \(-5\), 
            5}, {y, \(-5\), 5}, Contours \[Rule] {0}, PlotPoints \[Rule] 50, 
          ContourShading \[Rule] False, 
          ContourStyle \[Rule] Dashing[{0.02, 0.01}]], 
        ContourPlot[
          expressions\[LeftDoubleBracket]2\[RightDoubleBracket], {x, \(-5\), 
            5}, {y, \(-5\), 5}, Contours \[Rule] {0}, PlotPoints \[Rule] 50, 
          ContourShading \[Rule] False], 
        AspectRatio \[Rule] Automatic];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.0952381 0.5 0.0952381 [
[.11905 -0.0125 -9.15625 -15.6875 ]
[.11905 -0.0125 9.15625 0 ]
[.30952 -0.0125 -9.15625 -15.6875 ]
[.30952 -0.0125 9.15625 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.69048 -0.0125 -5 -15.6875 ]
[.69048 -0.0125 5 0 ]
[.88095 -0.0125 -5 -15.6875 ]
[.88095 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .11905 -18.3125 -7.84375 ]
[-0.0125 .11905 0 7.84375 ]
[-0.0125 .30952 -18.3125 -7.84375 ]
[-0.0125 .30952 0 7.84375 ]
[-0.0125 .5 -10 -7.84375 ]
[-0.0125 .5 0 7.84375 ]
[-0.0125 .69048 -10 -7.84375 ]
[-0.0125 .69048 0 7.84375 ]
[-0.0125 .88095 -10 -7.84375 ]
[-0.0125 .88095 0 7.84375 ]
[ 0 0 -0.125 0 ]
[ 0 1 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.11905 0 m
.11905 .00625 L
s
gsave
.11905 -0.0125 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.30952 0 m
.30952 .00625 L
s
gsave
.30952 -0.0125 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.69048 0 m
.69048 .00625 L
s
gsave
.69048 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.88095 0 m
.88095 .00625 L
s
gsave
.88095 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.2619 0 m
.2619 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.7381 0 m
.7381 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.07143 0 m
.07143 .00375 L
s
.02381 0 m
.02381 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.97619 0 m
.97619 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .11905 m
.00625 .11905 L
s
gsave
-0.0125 .11905 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
0 .30952 m
.00625 .30952 L
s
gsave
-0.0125 .30952 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .5 m
.00625 .5 L
s
gsave
-0.0125 .5 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .69048 m
.00625 .69048 L
s
gsave
-0.0125 .69048 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .88095 m
.00625 .88095 L
s
gsave
-0.0125 .88095 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .16667 m
.00375 .16667 L
s
0 .21429 m
.00375 .21429 L
s
0 .2619 m
.00375 .2619 L
s
0 .35714 m
.00375 .35714 L
s
0 .40476 m
.00375 .40476 L
s
0 .45238 m
.00375 .45238 L
s
0 .54762 m
.00375 .54762 L
s
0 .59524 m
.00375 .59524 L
s
0 .64286 m
.00375 .64286 L
s
0 .7381 m
.00375 .7381 L
s
0 .78571 m
.00375 .78571 L
s
0 .83333 m
.00375 .83333 L
s
0 .07143 m
.00375 .07143 L
s
0 .02381 m
.00375 .02381 L
s
0 .92857 m
.00375 .92857 L
s
0 .97619 m
.00375 .97619 L
s
.25 Mabswid
0 0 m
0 1 L
s
.11905 .99375 m
.11905 1 L
s
.30952 .99375 m
.30952 1 L
s
.5 .99375 m
.5 1 L
s
.69048 .99375 m
.69048 1 L
s
.88095 .99375 m
.88095 1 L
s
.125 Mabswid
.16667 .99625 m
.16667 1 L
s
.21429 .99625 m
.21429 1 L
s
.2619 .99625 m
.2619 1 L
s
.35714 .99625 m
.35714 1 L
s
.40476 .99625 m
.40476 1 L
s
.45238 .99625 m
.45238 1 L
s
.54762 .99625 m
.54762 1 L
s
.59524 .99625 m
.59524 1 L
s
.64286 .99625 m
.64286 1 L
s
.7381 .99625 m
.7381 1 L
s
.78571 .99625 m
.78571 1 L
s
.83333 .99625 m
.83333 1 L
s
.07143 .99625 m
.07143 1 L
s
.02381 .99625 m
.02381 1 L
s
.92857 .99625 m
.92857 1 L
s
.97619 .99625 m
.97619 1 L
s
.25 Mabswid
0 1 m
1 1 L
s
.99375 .11905 m
1 .11905 L
s
.99375 .30952 m
1 .30952 L
s
.99375 .5 m
1 .5 L
s
.99375 .69048 m
1 .69048 L
s
.99375 .88095 m
1 .88095 L
s
.125 Mabswid
.99625 .16667 m
1 .16667 L
s
.99625 .21429 m
1 .21429 L
s
.99625 .2619 m
1 .2619 L
s
.99625 .35714 m
1 .35714 L
s
.99625 .40476 m
1 .40476 L
s
.99625 .45238 m
1 .45238 L
s
.99625 .54762 m
1 .54762 L
s
.99625 .59524 m
1 .59524 L
s
.99625 .64286 m
1 .64286 L
s
.99625 .7381 m
1 .7381 L
s
.99625 .78571 m
1 .78571 L
s
.99625 .83333 m
1 .83333 L
s
.99625 .07143 m
1 .07143 L
s
.99625 .02381 m
1 .02381 L
s
.99625 .92857 m
1 .92857 L
s
.99625 .97619 m
1 .97619 L
s
.25 Mabswid
1 0 m
1 1 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
.5 Mabswid
[ .02 .01 ] 0 setdash
.97619 .524 m
.95675 .52502 L
.93732 .52613 L
.91788 .52735 L
.89845 .52868 L
.89198 .52915 L
.87901 .53015 L
.85957 .53178 L
.84014 .5336 L
.8207 .53563 L
.80126 .53793 L
.78183 .54055 L
.76239 .54355 L
.74295 .54704 L
.73521 .54859 L
.72352 .55164 L
.70408 .55732 L
.68465 .56333 L
.66801 .56803 L
.66521 .56927 L
.64577 .57901 L
.63073 .58746 L
.62634 .59059 L
.6069 .6069 L
.6069 .6069 L
.59059 .62634 L
.58746 .63073 L
.57901 .64577 L
.56927 .66521 L
.56803 .66801 L
.56333 .68465 L
.55732 .70408 L
.55164 .72352 L
.54859 .73521 L
.54704 .74295 L
.54355 .76239 L
.54055 .78183 L
.53793 .80126 L
.53563 .8207 L
.5336 .84014 L
.53178 .85957 L
.53015 .87901 L
.52915 .89198 L
.52868 .89845 L
.52735 .91788 L
.52613 .93732 L
.52502 .95675 L
.524 .97619 L
s
.02381 .476 m
.04325 .47498 L
.06268 .47387 L
.08212 .47265 L
.10155 .47132 L
.10802 .47085 L
.12099 .46985 L
.14043 .46822 L
.15986 .4664 L
.1793 .46437 L
.19874 .46207 L
.21817 .45945 L
.23761 .45645 L
.25705 .45296 L
.26479 .45141 L
.27648 .44836 L
.29592 .44268 L
.31535 .43667 L
.33199 .43197 L
.33479 .43073 L
.35423 .42099 L
.36927 .41254 L
.37366 .40941 L
.3931 .3931 L
.3931 .3931 L
.40941 .37366 L
.41254 .36927 L
.42099 .35423 L
.43073 .33479 L
.43197 .33199 L
.43667 .31535 L
.44268 .29592 L
.44836 .27648 L
.45141 .26479 L
.45296 .25705 L
.45645 .23761 L
.45945 .21817 L
.46207 .19874 L
.46437 .1793 L
.4664 .15986 L
.46822 .14043 L
.46985 .12099 L
.47085 .10802 L
.47132 .10155 L
.47265 .08212 L
.47387 .06268 L
.47498 .04325 L
.476 .02381 L
s
[ ] 0 setdash
.52915 .08158 m
.54859 .0743 L
.56803 .0689 L
.58746 .06548 L
.6069 .06414 L
.62634 .06506 L
.64577 .06846 L
.66521 .07467 L
.68115 .08212 L
.68465 .08414 L
.70408 .09757 L
.70884 .10155 L
.72352 .11608 L
.72779 .12099 L
.74215 .14043 L
.74295 .14168 L
.75349 .15986 L
.76239 .1788 L
.7626 .1793 L
.76999 .19874 L
.77595 .21817 L
.78069 .23761 L
.78183 .243 L
.78439 .25705 L
.78714 .27648 L
.78905 .29592 L
.79017 .31535 L
.79058 .33479 L
.79031 .35423 L
.7894 .37366 L
.78787 .3931 L
.78576 .41254 L
.78308 .43197 L
.78183 .43994 L
.77984 .45141 L
.77606 .47085 L
.77174 .49028 L
.76688 .50972 L
.76239 .52602 L
.76148 .52915 L
.75555 .54859 L
.74907 .56803 L
.74295 .585 L
.74203 .58746 L
.73442 .6069 L
.72623 .62634 L
.72352 .63246 L
.71742 .64577 L
.70797 .66521 L
.70408 .67284 L
Mistroke
.69785 .68465 L
.687 .70408 L
.68465 .70813 L
.67538 .72352 L
.66521 .73947 L
.66291 .74295 L
.6495 .76239 L
.64577 .76754 L
.63504 .78183 L
.62634 .79281 L
.61936 .80126 L
.6069 .81559 L
.60226 .8207 L
.58746 .83613 L
.58342 .84014 L
.56803 .85457 L
.56237 .85957 L
.54859 .87104 L
.5383 .87901 L
.52915 .88562 L
.50972 .89837 L
.50959 .89845 L
.49028 .9093 L
.47213 .91788 L
.47085 .91842 L
.45141 .9257 L
.43197 .9311 L
.41254 .93452 L
.3931 .93586 L
.37366 .93494 L
.35423 .93154 L
.33479 .92533 L
.31885 .91788 L
.31535 .91586 L
.29592 .90243 L
.29116 .89845 L
.27648 .88392 L
.27221 .87901 L
.25785 .85957 L
.25705 .85832 L
.24651 .84014 L
.23761 .8212 L
.2374 .8207 L
.23001 .80126 L
.22405 .78183 L
.21931 .76239 L
.21817 .757 L
.21561 .74295 L
.21286 .72352 L
.21095 .70408 L
Mistroke
.20983 .68465 L
.20942 .66521 L
.20969 .64577 L
.2106 .62634 L
.21213 .6069 L
.21424 .58746 L
.21692 .56803 L
.21817 .56006 L
.22016 .54859 L
.22394 .52915 L
.22826 .50972 L
.23312 .49028 L
.23761 .47398 L
.23852 .47085 L
.24445 .45141 L
.25093 .43197 L
.25705 .415 L
.25797 .41254 L
.26558 .3931 L
.27377 .37366 L
.27648 .36754 L
.28258 .35423 L
.29203 .33479 L
.29592 .32716 L
.30215 .31535 L
.313 .29592 L
.31535 .29187 L
.32462 .27648 L
.33479 .26053 L
.33709 .25705 L
.3505 .23761 L
.35423 .23246 L
.36496 .21817 L
.37366 .20719 L
.38064 .19874 L
.3931 .18441 L
.39774 .1793 L
.41254 .16387 L
.41658 .15986 L
.43197 .14543 L
.43763 .14043 L
.45141 .12896 L
.4617 .12099 L
.47085 .11438 L
.49028 .10163 L
.49041 .10155 L
.50972 .0907 L
.52787 .08212 L
.52915 .08158 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{287.875, 287.875},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006_0001[b000`400?l00000o`00003oo`3ooon`
0?ooo`00o`3ooon`0?ooo`00o`3ooon`0?ooo`00o`3ooon`0?ooo`00o`3ooon`0?ooo`00o`3ooon`
0?ooo`00o`3ooon`0?ooo`00o`3ooon`0?ooo`00o`3ooon`0?ooo`00E@3oool0103ooml0W4P00018
W03Ooom40?ooo`030;md00000000000000@0000000<0B9cO0?ooo`3oool0?P3oool01@3ogi`0B000
0000000007@0_ooo04@0oooo00<0_g@0000000000000100000000`18W=l0oooo0?ooo`160?ooo`04
0?oog`2LB00004RL0=ooobd0oooo001E0?ooo`040?oog`2LB00004RL0=ooodD0oooo00<0gia80000
000000000`0000000`00B9`0gooo0?ooo`0m0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oo
o`140?ooo`030=nLB0000000000000<0000000<004RL0=ooo`3oool0AP3oool0103ooml0W4P00018
W03Oool]0?ooo`00E@3oool0103ooml0W4P00018W03Ooom60?ooo`030=nLB018W=l0oooo0080oooo
00<0giad09cOo`3oool0?@3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo04@0oooo
00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`160?ooo`040?oog`2LB00004RL0=ooobd0
oooo001A0?ooo`030;md00000000000000D0000000<0M;oo0?ooo`3oool0A03oool00`3OW4P0M;oo
0?ooo`100?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0A@3oool00`3OW4P0M;oo
0?ooo`160?ooo`030;md00000000000000D0000000<0M;oo0?ooo`3oool0:P3oool004D0oooo00<0
oooO09a8000000001`0000001018W=l0oooo0?oOW01dW=l20?ooo`040?oog`2LB00004RL0=ooocL0
oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`040?ooo`030;mdB02Lgol0oooo03l0
oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`160?ooo`030;mdB02Lgol0oooo04D0
oooo00P0omnL07BLg`3oool0oooo0?oog`2LB00004RL0=ooobd0oooo001B0?ooo`070=nLB01d_ol0
oooo0?oog`2LB00004RL0=ooo`180?ooo`030?oog`2LB7@0_ooo03l0oooo00<0omnL04P0001d_ol0
0`3oool00`3o_g@0001809cOo`160?ooo`030?oog`2LB7@0_ooo04H0oooo00L0gia807Boo`3oool0
oooO09a80000B9`0gooo02d0oooo001B0?ooo`070?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`19
0?ooo`030?noM000M;l0oooo03h0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`17
0?ooo`030?noM000M;l0oooo04D0oooo00L0oooO09a8M02oool0oooO09a80000B9`0gooo02d0oooo
001C0?ooo`060?oOW018M;l0oooO09a80000B9`0goooB@3oool00`3ookl0M01809cOo`0n0?ooo`03
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0A`3oool00`3ookl0M01809cOo`160?ooo`06
0?oOW018M;l0oooO09a80000B9`0gooo;@3oool005@0oooo00D0gia807Bog`2LB00004RL0=ooo`19
0?ooo`030?oo_`1d0000B9cO03h0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`17
0?ooo`030?oo_`1d0000B9cO04L0oooo00D0gia807Bog`2LB00004RL0=ooo`0]0?ooo`00E03oool0
1@3ooml0W4Q804Q80000B9`0gooo04@0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RL
g`0o0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0@P3oool00`3o_i`0W=oo0?oo
o`020?ooo`030?oOW0180000B9cO04L0oooo00D0oooO09a8B018B00004RL0=ooo`0]0?ooo`00E@3o
ool0103ogi`0B0000018W03Ooom40?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02ooolo
0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`130?ooo`080?oog`2LB4P0W=oo0?ooo`3o
okl0M0000000M02ooom80?ooo`040?oOW018000004RL0=ooobd0oooo001F0?ooo`030=nLB000B9`0
gooo04D0oooo00H0oooO09a80000000000000000M02ooom10?ooo`050?oo_`1d000000000000M02o
ool0A@3oool01P3ooml0W4P0000000000000001d0;ooodX0oooo00<0gia80018W03Oool0;@3oool0
0?l0oooo/03oool00?l0oooo/03oool00?l0oooo/03oool00?l0oooo/03oool00?l0oooo/03oool0
0?l0oooo/03oool00?l0oooo/03oool00?l0oooo/03oool00?l0oooo/03oool00200ooooo`00002@
00000000803oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`060?ooo`400000000P0?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00H0
oooo0@0000000200oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`190?ooo`030000
003oool0oooo04T0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`190?ooo`030000
003oool0oooo02`0oooo0@0000000200oooo00<000000?ooo`3oool0o`3ooon<0?ooo`400000000P
0?ooo`030000003oool0oooo0?l0ooooS03oool100000000803oool00`000000oooo0?ooo`3o0?oo
oh`0oooo0@0000000200oooo00<000000?ooo`3oool0o`3ooon<0?ooo`400000000P0?ooo`030000
003oool0oooo0?l0ooooS03oool100000000803oool00`000000oooo0?ooo`3o0?oooh`0oooo0@00
00000200oooo0P00002k0?ooo`030000003oool0oooo0<d0oooo0P0000000200oooo00<000000?oo
o`3oool0^P3oool00`000000oooo0?ooo`3>0?ooo`400000000P0?ooo`030000003oool0oooo0;X0
oooo00<000000?ooo`3oool0cP3oool100000000803oool00`000000oooo0?ooo`2j0?ooo`030000
003oool0oooo0<h0oooo0@0000000200oooo00<000000?ooo`3oool0^P3oool00`000000oooo0?oo
o`3>0?ooo`400000000P0?ooo`030000003oool0oooo0;X0oooo00<000000?ooo`3oool0cP3oool1
00000000803oool00`000000oooo0?ooo`2j0?ooo`030000003oool0oooo0<h0oooo0@0000000200
oooo00<000000?ooo`3oool0^P3oool00`000000oooo0?ooo`3>0?ooo`400000000P0?ooo`030000
003oool0oooo0;X0oooo00<000000?ooo`3oool0cP3oool100000000803oool00`000000oooo0?oo
o`3o0?oooh`0oooo0@0000000200oooo00<000000?ooo`3oool0o`3ooon<0?ooo`400000000P0?oo
o`030000003oool0oooo0?l0ooooS03oool100000000803oool00`000000oooo0?ooo`2j0?ooo`03
0000003oool0oooo0<h0oooo0@0000000200oooo00<000000?ooo`3oool0^P3oool00`000000oooo
0?ooo`3>0?ooo`400000000P0?ooo`030000003oool0oooo0;X0oooo00<000000?ooo`3oool0cP3o
ool100000000803oool00`000000oooo0?ooo`2j0?ooo`030000003oool0oooo0<h0oooo0@000000
0200oooo00<000000?ooo`3oool0^P3oool00`000000oooo0?ooo`0Y0?oooa800000T`3oool10000
0000803oool00`000000oooo0?ooo`2j0?ooo`030000003oool0oooo02D0oooo1000000B0?ooo`@0
0000S`3oool100000000803oool00`000000oooo0?ooo`2i0?ooo`030000003oool0oooo0280oooo
1000000J0?ooo`@00000R`3oool100000000803oool200000;X0oooo00<000000?ooo`3oool07P3o
ool400000280oooo100000260?ooo`800000000P0?ooo`030000003oool0oooo0;T0oooo00<00000
0?ooo`3oool06P3oool4000002X0oooo0`0000240?ooo`400000000P0?ooo`030000003oool0oooo
0=@0oooo0P00000a0?ooo`800000PP3oool100000000803oool00`000000oooo0?ooo`3B0?ooo`80
0000=@3oool200000800oooo0@0000000200oooo00<000000?ooo`3oool0cP3oool4000003T0oooo
0P00001n0?ooo`400000000P0?ooo`030000003oool0oooo0;T0oooo00<000000?ooo`3oool0403o
ool2000003l0oooo0P00001l0?ooo`400000000P0?ooo`030000003oool0oooo0;T0oooo00<00000
0?ooo`3oool03P3oool2000004<0oooo0P00001j0?ooo`400000000P0?ooo`030000003oool0oooo
0;T0oooo00<000000?ooo`3oool0303oool2000004L0oooo00<000000?ooo`3oool0M`3oool10000
0000803oool00`000000oooo0?ooo`2i0?ooo`030000003oool0oooo00X0oooo0P00001:0?ooo`03
0000003oool0oooo07H0oooo0@0000000200oooo00<000000?ooo`3oool0^@3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo04/0oooo0P00001f0?ooo`400000000P0?ooo`030000003o
ool0oooo0;T0oooo00<000000?ooo`3oool01`3oool200000500oooo00<000000?ooo`3oool0L`3o
ool100000000803oool00`000000oooo0?ooo`2i0?ooo`030000003oool0oooo00D0oooo0P00001C
0?ooo`030000003oool0oooo0780oooo0@0000000200oooo00<000000?ooo`3oool0^@3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0L@3oool10000
0000803oool00`000000oooo0?ooo`2h0?ooo`030000003oool0oooo00<0oooo0P00001H0?ooo`03
0000003oool0oooo0700oooo0@00000001<0oooo00@0oooO09a80000B9`0gooo2@3oool00`000000
oooo0?ooo`2m0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3oool0K`3oool100000000
4`3oool0103ooml0W4P00018W03Oool90?ooo`030000003oool0oooo0;/0oooo0P00001M0?ooo`80
0000K`3oool1000000004`3oool0103ooml0W4P00018W03Oool90?ooo`030000003oool0oooo0;T0
oooo0P00001Q0?ooo`030000003oool0oooo06`0oooo0@00000000l0oooo00<0_g@0000000000000
1@0000000`1d_ol0oooo0?ooo`060?ooo`030000003oool0oooo0;P0oooo00<000000?ooo`3oool0
HP3oool00`000000oooo0?ooo`1[0?ooo`40000000030?ooo`030?oog`2LB000000000L0000000@0
B9cO0?ooo`3ogi`0M9cO0P3oool0103ooml0W4P00018W03Oool90?ooo`030000003oool0oooo0;H0
oooo0`00001U0?ooo`030000003oool0oooo06X0oooo0@0000000100oooo00L0gia807Boo`3oool0
oooO09a80000B9`0gooo00T0oooo0`00002e0?ooo`040000003oool0oooo000006H0oooo00<00000
0?ooo`3oool0I`3oool300000000403oool01`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool02@3o
ool00`000000oooo0?ooo`2d0?ooo`050000003oool0oooo0?ooo`000000I`3oool00`000000oooo
0?ooo`1X0?ooo`400000000A0?ooo`060?oOW018M;l0oooO09a80000B9`0gooo2@3oool00`000000
oooo0?ooo`2c0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0I@3oool00`000000
oooo0?ooo`1X0?ooo`400000000B0?ooo`050=nLB01d_ml0W4P00018W03Oool02@3oool00`000000
oooo0?ooo`2b0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0IP3oool00`000000
oooo0?ooo`1W0?ooo`400000000B0?ooo`050?oog`2LB4P0B4P00018W03Oool02@3oool00`000000
oooo0?ooo`2a0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0IP3oool00`000000
oooo0?ooo`1W0?ooo`400000000C0?ooo`040?oOW018000004RL0=ooo`T0oooo00<000000?ooo`3o
ool0[`3oool2000000L0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`1V0?ooo`40
0000000D0?ooo`030=nLB000B9`0gooo00T0oooo00<000000?ooo`3oool0[P3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool0I@3oool100000000803o
ool00`000000oooo0?ooo`2/0?ooo`800000M@3oool00`000000oooo0?ooo`1U0?ooo`400000000P
0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool0MP3oool00`000000oooo0?ooo`1T
0?ooo`400000000P0?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0N03oool00`00
0000oooo0?ooo`1S0?ooo`400000000P0?ooo`030000003oool0oooo0:T0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`1/0?ooo`030000003oool0oooo0680oooo0@0000000200oooo
00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo06`0oooo
00<000000?ooo`3oool0HP3oool100000000803oool00`000000oooo0?ooo`2W0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool0K@3oool00`000000oooo0?ooo`1Q0?ooo`400000000P
0?ooo`030000003oool0oooo0:H0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1]
0?ooo`030000003oool0oooo0640oooo0@0000000200oooo00<000000?ooo`3oool0Y@3oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool0H03oool10000
0000803oool00`000000oooo0?ooo`2T0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool0KP3oool00`000000oooo0?ooo`1P0?ooo`400000000P0?ooo`030000003oool0oooo0:<0oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo05l0oooo
0@0000000200oooo00<000000?ooo`3oool0XP3oool00`000000oooo0?ooo`0A0?ooo`030000003o
ool0oooo0740oooo00<000000?ooo`3oool0GP3oool100000000803oool00`000000oooo0?ooo`2Q
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`1N
0?ooo`400000000P0?ooo`800000X@3oool00`000000oooo0?ooo`280?ooo`030000003oool0oooo
05`0oooo0P0000000200oooo00<000000?ooo`3oool0W`3oool00`000000oooo0?ooo`290?ooo`03
0000003oool0oooo05d0oooo0@0000000200oooo00<000000?ooo`3oool0WP3oool00`000000oooo
0?ooo`2;0?ooo`030000003oool0oooo05`0oooo0@0000000200oooo00<000000?ooo`3oool0W@3o
ool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool0G03o
ool100000000803oool00`000000oooo0?ooo`2L0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool0M03oool00`000000oooo0?ooo`1K0?ooo`400000000P0?ooo`030000003oool0oooo
09/0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1d0?ooo`030000003oool0oooo
05/0oooo0@0000000200oooo00<000000?ooo`3oool0VP3oool00`000000oooo0?ooo`0I0?ooo`03
0000003oool0oooo07@0oooo00<000000?ooo`3oool0F`3oool100000000803oool00`000000oooo
0?ooo`2I0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0M@3oool00`000000oooo
0?ooo`1J0?ooo`400000000P0?ooo`030000003oool0oooo09P0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`1f0?ooo`030000003oool0oooo05X0oooo0@0000000200oooo00<00000
0?ooo`3oool0U`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo07L0oooo00<00000
0?ooo`3oool0F@3oool100000000803oool00`000000oooo0?ooo`2G0?ooo`030000003oool0oooo
01/0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`1I0?ooo`400000000P0?ooo`03
0000003oool0oooo09H0oooo00<000000?ooo`3oool0U`3oool00`000000oooo0?ooo`1H0?ooo`40
0000000P0?ooo`030000003oool0oooo09D0oooo00<000000?ooo`3oool0V03oool00`000000oooo
0?ooo`1H0?ooo`400000000P0?ooo`030000003oool0oooo09@0oooo00<000000?ooo`3oool0V@3o
ool00`000000oooo0?ooo`1H0?ooo`400000000P0?ooo`030000003oool0oooo09<0oooo00<00000
0?ooo`3oool07`3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo05P0oooo0@000000
0200oooo00<000000?ooo`3oool0T`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo
07T0oooo00<000000?ooo`3oool0E`3oool100000000803oool00`000000oooo0?ooo`2B0?ooo`03
0000003oool0oooo0200oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`1G0?ooo`40
0000000P0?ooo`030000003oool0oooo0940oooo00<000000?ooo`3oool0803oool00`000000oooo
0?ooo`1j0?ooo`030000003oool0oooo05L0oooo0@0000000200oooo00<000000?ooo`3oool0T03o
ool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0E`3o
ool100000000803oool200000900oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`1k
0?ooo`030000003oool0oooo05D0oooo0P0000000200oooo00<000000?ooo`3oool0SP3oool00`00
0000oooo0?ooo`0S0?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool0EP3oool10000
0000803oool00`000000oooo0?ooo`2=0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3o
ool0N`3oool00`000000oooo0?ooo`1F0?ooo`400000000P0?ooo`030000003oool0oooo08d0oooo
00<000000?ooo`3oool0903oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo05H0oooo
0@0000000200oooo00<000000?ooo`3oool0S03oool00`000000oooo0?ooo`2T0?ooo`030000003o
ool0oooo05D0oooo0@0000000200oooo00<000000?ooo`3oool0R`3oool00`000000oooo0?ooo`2U
0?ooo`030000003oool0oooo05D0oooo0@0000000200oooo00<000000?ooo`3oool0R`3oool00`00
0000oooo0?ooo`2U0?ooo`030000003oool0oooo05D0oooo0@0000000200oooo00<000000?ooo`3o
ool0RP3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo07d0oooo00<000000?ooo`3o
ool0E@3oool100000000803oool00`000000oooo0?ooo`290?ooo`030000003oool0oooo02L0oooo
00<000000?ooo`3oool0OP3oool00`000000oooo0?ooo`1D0?ooo`400000000P0?ooo`030000003o
ool0oooo08T0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`1n0?ooo`030000003o
ool0oooo05@0oooo0@0000000200oooo00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`0X
0?ooo`030000003oool0oooo07h0oooo00<000000?ooo`3oool0E03oool100000000803oool00`00
0000oooo0?ooo`270?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0OP3oool00`00
0000oooo0?ooo`1D0?ooo`400000000P0?ooo`030000003oool0oooo08L0oooo00<000000?ooo`3o
ool0:03oool00`000000oooo0?ooo`1o0?ooo`030000003oool0oooo05@0oooo0@0000000200oooo
00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo07l0oooo
00<000000?ooo`3oool0E03oool100000000803oool00`000000oooo0?ooo`250?ooo`030000003o
ool0oooo02X0oooo00<000000?ooo`3oool0O`3oool00`000000oooo0?ooo`1D0?ooo`400000000P
0?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`20
0?ooo`030000003oool0oooo05<0oooo0@0000000200oooo00<000000?ooo`3oool0Q03oool00`00
0000oooo0?ooo`2^0?ooo`030000003oool0oooo05<0oooo0@0000000200oooo00<000000?ooo`3o
ool0P`3oool00`000000oooo0?ooo`2_0?ooo`030000003oool0oooo05<0oooo0@0000000200oooo
00<000000?ooo`3oool0PP3oool00`000000oooo0?ooo`2`0?ooo`030000003oool0oooo05<0oooo
0@0000000200oooo0P0000220?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0P@3o
ool00`000000oooo0?ooo`1B0?ooo`800000000P0?ooo`030000003oool0oooo0840oooo00<00000
0?ooo`3oool0;@3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo05<0oooo0@000000
0200oooo00<000000?ooo`3oool0P03oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo
08<0oooo00<000000?ooo`3oool0DP3oool100000000803oool00`000000oooo0?ooo`1o0?ooo`03
0000003oool0oooo02h0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`1B0?ooo`40
0000000P0?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool0;@3oool00`000000oooo
0?ooo`240?ooo`030000003oool0oooo0580oooo0@0000000200oooo00<000000?ooo`3oool0OP3o
ool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool0DP3o
ool100000000803oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo02l0oooo00<00000
0?ooo`3oool0Q03oool00`000000oooo0?ooo`1B0?ooo`400000000P0?ooo`030000003oool0oooo
07d0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`240?ooo`030000003oool0oooo
0580oooo0@0000000200oooo00<000000?ooo`3oool0O03oool00`000000oooo0?ooo`0`0?ooo`03
0000003oool0oooo08@0oooo00<000000?ooo`3oool0DP3oool100000000803oool00`000000oooo
0?ooo`1l0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool0DP3oool100000000803o
ool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo0;T0oooo00<000000?ooo`3oool0D@3o
ool100000000803oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo0300oooo00<00000
0?ooo`3oool0QP3oool00`000000oooo0?ooo`1A0?ooo`400000000P0?ooo`030000003oool0oooo
07X0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`270?ooo`030000003oool0oooo
0540oooo0@0000000200oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`0`0?ooo`03
0000003oool0oooo08L0oooo00<000000?ooo`3oool0D@3oool1000000003`3oool00`2oM0000000
000000040000000304RLg`3oool0oooo00L0oooo00<000000?ooo`3oool0N@3oool00`000000oooo
0?ooo`0a0?ooo`030000003oool0oooo08L0oooo00<000000?ooo`3oool0D@3oool100000000403o
ool00`3OW4P0000000000003000000030018W03Oool0oooo00L0oooo00<000000?ooo`3oool0N@3o
ool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo08P0oooo00<000000?ooo`3oool0D@3o
ool1000000004@3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo00L0oooo00<00000
0?ooo`3oool0N03oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo08P0oooo00<00000
0?ooo`3oool0D@3oool1000000004P3oool00`3OW4P0M;oo0?ooo`0;0?ooo`030000003oool0oooo
07L0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`280?ooo`030000003oool0oooo
0540oooo0@00000000<0oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`030?ooo`03
0;mdB02Lgol0oooo00X0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0a0?ooo`03
0000003oool0oooo08X0oooo00<000000?ooo`3oool0D03oool1000000004`3oool00`3ooml0W4Qd
0;ooo`0:0?ooo`<00000MP3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo08X0oooo
00<000000?ooo`3oool0CP3oool300000000503oool00`3o_g@007Bo0?ooo`090?ooo`030000003o
ool0oooo07H0oooo00<000000?ooo`3oool0_`3oool00`000000oooo0?ooo`1@0?ooo`400000000D
0?ooo`030?oo_`1d04P0W=oo00T0oooo00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`30
0?ooo`030000003oool0oooo0500oooo0@00000001@0oooo00<0oono07@00018W=l02@3oool00`00
0000oooo0?ooo`1e0?ooo`030000003oool0oooo0<00oooo00<000000?ooo`3oool0D03oool10000
00003`3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO00T0oooo00<000000?ooo`3o
ool0M03oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo08/0oooo00<000000?ooo`3o
ool0D03oool1000000003`3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo2@3oool0
0`000000oooo0?ooo`1d0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0R`3oool0
0`000000oooo0?ooo`1@0?ooo`400000000@0?ooo`060?oog`2LB00000000000000007@0_ooo2P3o
ool00`000000oooo0?ooo`1c0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0S03o
ool00`000000oooo0?ooo`1@0?ooo`400000000P0?ooo`030000003oool0oooo07<0oooo00<00000
0?ooo`3oool0<`3oool00`000000oooo0?ooo`2<0?ooo`030000003oool0oooo0500oooo0@000000
0200oooo00<000000?ooo`3oool0LP3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo
08`0oooo00<000000?ooo`3oool0D03oool100000000803oool00`000000oooo0?ooo`1b0?ooo`03
0000003oool0oooo03<0oooo00<000000?ooo`3oool0S@3oool00`000000oooo0?ooo`1@0?ooo`40
0000000P0?ooo`030000003oool0oooo0740oooo00<000000?ooo`3oool0=03oool00`000000oooo
0?ooo`2=0?ooo`030000003oool0oooo0500oooo0@0000000200oooo00<000000?ooo`3oool0L03o
ool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo08d0oooo00<000000?ooo`3oool0D03o
ool100000000803oool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo0<D0oooo00<00000
0?ooo`3oool0D03oool100000000803oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo
0<H0oooo00<000000?ooo`3oool0D03oool100000000803oool00`000000oooo0?ooo`1_0?ooo`03
0000003oool0oooo0<H0oooo00<000000?ooo`3oool0D03oool100000000803oool00`000000oooo
0?ooo`1^0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0S`3oool00`000000oooo
0?ooo`1@0?ooo`400000000P0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool0=03o
ool00`000000oooo0?ooo`2@0?ooo`030000003oool0oooo0500oooo0@0000000200oooo00<00000
0?ooo`3oool0K@3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo0900oooo00<00000
0?ooo`3oool0D03oool100000000803oool2000006h0oooo00<000000?ooo`3oool0=03oool00`00
0000oooo0?ooo`2A0?ooo`030000003oool0oooo04l0oooo0P0000000200oooo00<000000?ooo`3o
ool0K03oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo0940oooo00<000000?ooo`3o
ool0D03oool100000000803oool00`000000oooo0?ooo`1/0?ooo`030000003oool0oooo03@0oooo
00<000000?ooo`3oool0TP3oool00`000000oooo0?ooo`1@0?ooo`400000000P0?ooo`030000003o
ool0oooo06`0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`2A0?ooo`030000003o
ool0oooo0540oooo0@0000000200oooo00<000000?ooo`3oool0J`3oool00`000000oooo0?ooo`0d
0?ooo`030000003oool0oooo0980oooo00<000000?ooo`3oool0D@3oool100000000803oool00`00
0000oooo0?ooo`1[0?ooo`030000003oool0oooo0<T0oooo00<000000?ooo`3oool0D@3oool10000
0000803oool00`000000oooo0?ooo`1Z0?ooo`030000003oool0oooo0<X0oooo00<000000?ooo`3o
ool0D@3oool100000000803oool00`000000oooo0?ooo`1Z0?ooo`030000003oool0oooo03<0oooo
00<000000?ooo`3oool0U03oool00`000000oooo0?ooo`1A0?ooo`400000000P0?ooo`030000003o
ool0oooo06X0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`2E0?ooo`030000003o
ool0oooo0540oooo0@0000000200oooo00<000000?ooo`3oool0J@3oool00`000000oooo0?ooo`0b
0?ooo`030000003oool0oooo09H0oooo00<000000?ooo`3oool0D@3oool100000000803oool00`00
0000oooo0?ooo`1Y0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0U`3oool00`00
0000oooo0?ooo`1A0?ooo`400000000P0?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3o
ool0<P3oool00`000000oooo0?ooo`2G0?ooo`030000003oool0oooo0540oooo0@0000000200oooo
00<000000?ooo`3oool0J03oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo09P0oooo
00<000000?ooo`3oool0D@3oool100000000803oool00`000000oooo0?ooo`1X0?ooo`030000003o
ool0oooo0300oooo00<000000?ooo`3oool0V@3oool00`000000oooo0?ooo`1A0?ooo`400000000P
0?ooo`030000003oool0oooo06L0oooo00<000000?ooo`3oool0c@3oool00`000000oooo0?ooo`1A
0?ooo`400000000P0?ooo`030000003oool0oooo06L0oooo00<000000?ooo`3oool0c@3oool00`00
0000oooo0?ooo`1A0?ooo`400000000P0?ooo`030000003oool0oooo06L0oooo00<000000?ooo`3o
ool0;P3oool00`000000oooo0?ooo`2L0?ooo`030000003oool0oooo0540oooo0@0000000200oooo
00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo09d0oooo
00<000000?ooo`3oool0D@3oool100000000803oool00`000000oooo0?ooo`1V0?ooo`030000003o
ool0oooo02d0oooo00<000000?ooo`3oool0WP3oool00`000000oooo0?ooo`1A0?ooo`400000000P
0?ooo`800000IP3oool00`000000oooo0?ooo`0/0?ooo`800000X03oool00`000000oooo0?ooo`1A
0?ooo`800000000P0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3oool0:`3oool00`00
0000oooo0?ooo`2P0?ooo`030000003oool0oooo0580oooo0@0000000200oooo00<000000?ooo`3o
ool0I@3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool0DP3oool100000000803oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo0<l0oooo
00<000000?ooo`3oool0DP3oool100000000803oool00`000000oooo0?ooo`1T0?ooo`030000003o
ool0oooo02L0oooo00<000000?ooo`3oool0Y@3oool00`000000oooo0?ooo`1B0?ooo`400000000P
0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`2V
0?ooo`030000003oool0oooo0580oooo0@0000000200oooo00<000000?ooo`3oool0H`3oool00`00
0000oooo0?ooo`0U0?ooo`800000Z@3oool00`000000oooo0?ooo`1B0?ooo`400000000P0?ooo`03
0000003oool0oooo0680oooo00<000000?ooo`3oool0903oool200000:/0oooo00<000000?ooo`3o
ool0DP3oool100000000803oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo0280oooo
0P00002/0?ooo`030000003oool0oooo05<0oooo0@0000000200oooo00<000000?ooo`3oool0HP3o
ool00`000000oooo0?ooo`3@0?ooo`030000003oool0oooo05<0oooo0@0000000200oooo00<00000
0?ooo`3oool0H@3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0:h0oooo00<00000
0?ooo`3oool0D`3oool100000000803oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo
01`0oooo1000002a0?ooo`030000003oool0oooo05<0oooo0@0000000200oooo00<000000?ooo`3o
ool0H@3oool00`000000oooo0?ooo`0J0?ooo`800000]@3oool00`000000oooo0?ooo`1C0?ooo`40
0000000P0?ooo`030000003oool0oooo0600oooo00<000000?ooo`3oool06@3oool200000;H0oooo
00<000000?ooo`3oool0E03oool100000000803oool00`000000oooo0?ooo`1P0?ooo`030000003o
ool0oooo01D0oooo0P00002j0?ooo`030000003oool0oooo05@0oooo0@0000000200oooo00<00000
0?ooo`3oool0G`3oool00`000000oooo0?ooo`0C0?ooo`<00000_03oool00`000000oooo0?ooo`1D
0?ooo`400000000P0?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3oool03`3oool40000
0;l0oooo00<000000?ooo`3oool0E03oool100000000803oool00`000000oooo0?ooo`1O0?ooo`03
0000003oool0oooo00/0oooo00<000000?ooo`3oool0``3oool00`000000oooo0?ooo`1E0?ooo`40
0000000P0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool0203oool400000<H0oooo
00<000000?ooo`3oool0E@3oool100000000803oool2000005l0oooo00<000000?ooo`3oool0103o
ool400000<X0oooo00<000000?ooo`3oool0E03oool200000000803oool00`000000oooo0?ooo`1N
0?ooo`@00000d@3oool00`000000oooo0?ooo`1E0?ooo`400000000P0?ooo`030000003oool0oooo
05X0oooo1@00003D0?ooo`030000003oool0oooo05D0oooo0@0000000200oooo00<000000?ooo`3o
ool0DP3oool5000000H0oooo00<000000?ooo`3oool0d`3oool00`000000oooo0?ooo`1E0?ooo`40
0000000P0?ooo`030000003oool0oooo04h0oooo1000000;0?ooo`030000003oool0oooo0=80oooo
00<000000?ooo`3oool0EP3oool100000000803oool00`000000oooo0?ooo`120?ooo`T000004P3o
ool00`000000oooo0?ooo`3B0?ooo`030000003oool0oooo05H0oooo0@0000000200oooo00<00000
0?ooo`3oool0>P3oool5000001d0oooo00<000000?ooo`3oool0d`3oool00`000000oooo0?ooo`1F
0?ooo`400000000P0?ooo`030000003oool0oooo02`0oooo1`0000030?ooo`@000008P3oool00`00
0000oooo0?ooo`3C0?ooo`030000003oool0oooo05H0oooo0@0000000200oooo00<000000?ooo`3o
ool07P3oool9000000<0oooo0P00000`0?ooo`030000003oool0oooo0=<0oooo00<000000?ooo`3o
ool0EP3oool100000000803oool00`000000oooo0?ooo`0;0?ooo`@000000`3oool900000440oooo
00<000000?ooo`3oool0d`3oool00`000000oooo0?ooo`1F0?ooo`400000000P0?ooo`030000003o
ool0oooo00H0oooo1@00001A0?ooo`030000003oool0oooo0=80oooo00<000000?ooo`3oool0E`3o
ool100000000803oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo0=<0oooo00<00000
0?ooo`3oool0E`3oool100000000803oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo
0=<0oooo00<000000?ooo`3oool0E`3oool100000000803oool00`000000oooo0?ooo`1K0?ooo`03
0000003oool0oooo0=<0oooo00<000000?ooo`3oool0E`3oool1000000004@3oool01@3ogi`0B000
0000000007@0_ooo00X0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`3C0?ooo`03
0000003oool0oooo05P0oooo0@0000000100oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo
00T0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`3C0?ooo`030000003oool0oooo
05P0oooo0@0000000100oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`080?ooo`03
0000003oool0oooo05X0oooo00<000000?ooo`3oool0d`3oool00`000000oooo0?ooo`1H0?ooo`40
0000000?0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool00`000000oooo
0?ooo`1I0?ooo`030000003oool0oooo0=<0oooo00<000000?ooo`3oool0F@3oool1000000003`3o
ool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00P0oooo00<000000?ooo`3oool0F@3o
ool00`000000oooo0?ooo`3C0?ooo`030000003oool0oooo05T0oooo0@00000000l0oooo00<0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`080?ooo`<00000F@3oool00`000000oooo0?ooo`3C
0?ooo`030000003oool0oooo05L0oooo0`00000000l0oooo00<0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`080?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3oool0d`3oool00`00
0000oooo0?ooo`1I0?ooo`400000000?0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol0203oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo0=<0oooo00<000000?ooo`3o
ool0FP3oool1000000003`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00P0oooo
00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`3C0?ooo`030000003oool0oooo05X0oooo
0@0000000100oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`080?ooo`030000003o
ool0oooo05P0oooo00<000000?ooo`3oool0d`3oool00`000000oooo0?ooo`1J0?ooo`400000000@
0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`030000003oool0oooo05P0oooo
00<000000?ooo`3oool0d`3oool00`000000oooo0?ooo`1J0?ooo`400000000A0?ooo`050?oo_`1d
000000000000M02oool02P3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo0=@0oooo
00<000000?ooo`3oool0FP3oool100000000803oool00`000000oooo0?ooo`1G0?ooo`030000003o
ool0oooo0=@0oooo00<000000?ooo`3oool0FP3oool100000000803oool00`000000oooo0?ooo`1G
0?ooo`030000003oool0oooo0=<0oooo00<000000?ooo`3oool0F`3oool100000000803oool00`00
0000oooo0?ooo`1G0?ooo`030000003oool0oooo0=<0oooo00<000000?ooo`3oool0F`3oool10000
0000803oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo0=@0oooo00<000000?ooo`3o
ool0BP3oool9000000P0oooo0@0000000200oooo00<000000?ooo`3oool0EP3oool00`000000oooo
0?ooo`3D0?ooo`030000003oool0oooo03L0oooo100000030?ooo`T00000503oool100000000803o
ool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo0=<0oooo00<000000?ooo`3oool0:P3o
ool6000000<0oooo1@00000T0?ooo`400000000P0?ooo`030000003oool0oooo05H0oooo00<00000
0?ooo`3oool0d`3oool00`000000oooo0?ooo`0P0?ooo`@000000`3oool300000380oooo0@000000
0200oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`3B0?ooo`030000003oool0oooo
01P0oooo00@000000?ooo`3oool0oooo1@00000l0?ooo`400000000P0?ooo`030000003oool0oooo
05H0oooo00<000000?ooo`3oool0dP3oool00`000000oooo0?ooo`0@0?ooo`P00000A@3oool10000
0000803oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo0=<0oooo00<000000?ooo`3o
ool01P3oool700000500oooo0@0000000200oooo00<000000?ooo`3oool0E@3oool00`000000oooo
0?ooo`3C0?ooo`040000003oool0oooo000000<0oooo0P00001G0?ooo`400000000P0?ooo`030000
003oool0oooo05D0oooo00<000000?ooo`3oool0dP3oool4000005d0oooo0@0000000200oooo0P00
001F0?ooo`030000003oool0oooo0<l0oooo1000001O0?ooo`800000000P0?ooo`030000003oool0
oooo05D0oooo00<000000?ooo`3oool0b03oool4000000H0oooo00<000000?ooo`3oool0GP3oool1
00000000803oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo0<D0oooo0`00000:0?oo
o`030000003oool0oooo05h0oooo0@0000000200oooo00<000000?ooo`3oool0E03oool00`000000
oooo0?ooo`340?ooo`800000303oool00`000000oooo0?ooo`1O0?ooo`400000000P0?ooo`030000
003oool0oooo05@0oooo00<000000?ooo`3oool0_@3oool400000140oooo00<000000?ooo`3oool0
G`3oool100000000803oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo0;X0oooo0`00
000D0?ooo`030000003oool0oooo0600oooo0@0000000200oooo00<000000?ooo`3oool0E03oool0
0`000000oooo0?ooo`2h0?ooo`8000005`3oool00`000000oooo0?ooo`1P0?ooo`400000000P0?oo
o`030000003oool0oooo05<0oooo00<000000?ooo`3oool0]03oool3000001X0oooo00<000000?oo
o`3oool0H@3oool100000000803oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo0;40
oooo0`00000M0?ooo`030000003oool0oooo0640oooo0@0000000200oooo00<000000?ooo`3oool0
D`3oool00`000000oooo0?ooo`2`0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0
H@3oool100000000803oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo0:h0oooo0P00
000P0?ooo`030000003oool0oooo0680oooo0@0000000200oooo00<000000?ooo`3oool0D`3oool0
0`000000oooo0?ooo`3@0?ooo`030000003oool0oooo0680oooo0@0000000200oooo00<000000?oo
o`3oool0DP3oool00`000000oooo0?ooo`2[0?ooo`800000903oool00`000000oooo0?ooo`1R0?oo
o`400000000P0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0Z@3oool2000002H0
oooo00<000000?ooo`3oool0HP3oool100000000803oool00`000000oooo0?ooo`1B0?ooo`030000
003oool0oooo0:L0oooo0P00000W0?ooo`030000003oool0oooo06<0oooo0@0000000200oooo00<0
00000?ooo`3oool0DP3oool00`000000oooo0?ooo`2U0?ooo`800000:@3oool00`000000oooo0?oo
o`1S0?ooo`400000000P0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0c`3oool0
0`000000oooo0?ooo`1T0?ooo`400000000P0?ooo`030000003oool0oooo0580oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo06@0oooo0@0000000200
oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`2P0?ooo`030000003oool0oooo02/0
oooo00<000000?ooo`3oool0I@3oool100000000803oool2000005<0oooo00<000000?ooo`3oool0
W`3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo06@0oooo0P0000000200oooo00<0
00000?ooo`3oool0D@3oool00`000000oooo0?ooo`2O0?ooo`030000003oool0oooo02`0oooo00<0
00000?ooo`3oool0IP3oool100000000803oool00`000000oooo0?ooo`1A0?ooo`030000003oool0
oooo09h0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`1V0?ooo`400000000P0?oo
o`030000003oool0oooo0540oooo00<000000?ooo`3oool0W@3oool00`000000oooo0?ooo`0^0?oo
o`030000003oool0oooo06H0oooo0@0000000200oooo00<000000?ooo`3oool0D@3oool00`000000
oooo0?ooo`2L0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0I`3oool100000000
803oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo0<d0oooo00<000000?ooo`3oool0
I`3oool100000000803oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo09T0oooo00<0
00000?ooo`3oool0<@3oool00`000000oooo0?ooo`1W0?ooo`400000000P0?ooo`030000003oool0
oooo0540oooo00<000000?ooo`3oool0V03oool00`000000oooo0?ooo`0a0?ooo`030000003oool0
oooo06P0oooo0@0000000200oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`2G0?oo
o`030000003oool0oooo0380oooo00<000000?ooo`3oool0J03oool100000000803oool00`000000
oooo0?ooo`1A0?ooo`030000003oool0oooo09H0oooo00<000000?ooo`3oool0<P3oool00`000000
oooo0?ooo`1Y0?ooo`400000000P0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool0
UP3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo06T0oooo0@0000000200oooo00<0
00000?ooo`3oool0D@3oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo03<0oooo00<0
00000?ooo`3oool0J@3oool100000000803oool00`000000oooo0?ooo`1A0?ooo`030000003oool0
oooo09@0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`1Z0?ooo`400000000P0?oo
o`030000003oool0oooo0540oooo00<000000?ooo`3oool0bP3oool00`000000oooo0?ooo`1Z0?oo
o`400000000P0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool0b@3oool00`000000
oooo0?ooo`1[0?ooo`400000000P0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool0
b@3oool00`000000oooo0?ooo`1[0?ooo`400000000P0?ooo`030000003oool0oooo0540oooo00<0
00000?ooo`3oool0TP3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo06/0oooo0@00
00000200oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`2B0?ooo`030000003oool0
oooo03@0oooo00<000000?ooo`3oool0K03oool100000000803oool00`000000oooo0?ooo`1@0?oo
o`030000003oool0oooo0940oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`1/0?oo
o`400000000P0?ooo`800000D@3oool00`000000oooo0?ooo`2A0?ooo`030000003oool0oooo03@0
oooo00<000000?ooo`3oool0K03oool200000000803oool00`000000oooo0?ooo`1@0?ooo`030000
003oool0oooo0900oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`1]0?ooo`400000
000P0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0T03oool00`000000oooo0?oo
o`0d0?ooo`030000003oool0oooo06h0oooo0@0000000200oooo00<000000?ooo`3oool0D03oool0
0`000000oooo0?ooo`2?0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0KP3oool1
00000000803oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo08l0oooo00<000000?oo
o`3oool0=03oool00`000000oooo0?ooo`1_0?ooo`400000000P0?ooo`030000003oool0oooo0500
oooo00<000000?ooo`3oool0aP3oool00`000000oooo0?ooo`1_0?ooo`400000000P0?ooo`030000
003oool0oooo0500oooo00<000000?ooo`3oool0a@3oool00`000000oooo0?ooo`1`0?ooo`400000
000P0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0S@3oool00`000000oooo0?oo
o`0e0?ooo`030000003oool0oooo0700oooo0@0000000200oooo00<000000?ooo`3oool0D03oool0
0`000000oooo0?ooo`2=0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0L@3oool1
00000000803oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo08d0oooo00<000000?oo
o`3oool0=03oool00`000000oooo0?ooo`1a0?ooo`400000000P0?ooo`030000003oool0oooo0500
oooo00<000000?ooo`3oool0S03oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo0780
oooo0@0000000200oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`2<0?ooo`030000
003oool0oooo03@0oooo00<000000?ooo`3oool0LP3oool100000000803oool00`000000oooo0?oo
o`1@0?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?oo
o`1c0?ooo`400000000P0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0R`3oool0
0`000000oooo0?ooo`0d0?ooo`030000003oool0oooo07<0oooo0@00000000l0oooo00<0_g@00000
00000000100000000`18W=l0oooo0?ooo`070?ooo`030000003oool0oooo0500oooo00<000000?oo
o`3oool0R`3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo07@0oooo0@0000000100
oooo00<0gia80000000000000`0000000`00B9`0gooo0?ooo`070?ooo`030000003oool0oooo0500
oooo00<000000?ooo`3oool0R`3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo07@0
oooo0@0000000140oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`070?ooo`030000
003oool0oooo0500oooo00<000000?ooo`3oool0`03oool00`000000oooo0?ooo`1e0?ooo`400000
000B0?ooo`030=nLB01d_ol0oooo00/0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?oo
o`300?ooo`030000003oool0oooo07D0oooo0@00000001<0oooo00<0_gA809cOo`3oool02P3oool0
0`000000oooo0?ooo`1@0?ooo`030000003oool0oooo0;l0oooo00<000000?ooo`3oool0MP3oool1
000000004`3oool00`3ooml0W4Qd0;ooo`0:0?ooo`<00000D03oool00`000000oooo0?ooo`2:0?oo
o`030000003oool0oooo0380oooo00<000000?ooo`3oool0M03oool300000000503oool00`3o_g@0
07Bo0?ooo`090?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0RP3oool00`000000
oooo0?ooo`0a0?ooo`030000003oool0oooo07L0oooo0@00000001@0oooo00<0oono07@0B02Lgol0
2@3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo08T0oooo00<000000?ooo`3oool0
<@3oool00`000000oooo0?ooo`1h0?ooo`400000000D0?ooo`030?oo_`1d0000B9cO00T0oooo00<0
00000?ooo`3oool0D@3oool00`000000oooo0?ooo`280?ooo`030000003oool0oooo0340oooo00<0
00000?ooo`3oool0N03oool1000000003`3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000
B9cO00T0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`280?ooo`030000003oool0
oooo0300oooo00<000000?ooo`3oool0N@3oool1000000003`3oool0203ooml0W4Q809cOo`3oool0
oono07@0000007@0_ooo2@3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo08L0oooo
00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`1i0?ooo`400000000@0?ooo`060?oog`2L
B00000000000000007@0_ooo2P3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo08L0
oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`1j0?ooo`400000000P0?ooo`030000
003oool0oooo0540oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`0a0?ooo`030000
003oool0oooo07X0oooo0@0000000200oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?oo
o`2i0?ooo`030000003oool0oooo07/0oooo0@0000000200oooo00<000000?ooo`3oool0D@3oool0
0`000000oooo0?ooo`2i0?ooo`030000003oool0oooo07/0oooo0@0000000200oooo00<000000?oo
o`3oool0DP3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo07`0oooo0@0000000200
oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`240?ooo`030000003oool0oooo02l0
oooo00<000000?ooo`3oool0O@3oool100000000803oool00`000000oooo0?ooo`1B0?ooo`030000
003oool0oooo08@0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`1m0?ooo`400000
000P0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?oo
o`0^0?ooo`030000003oool0oooo07h0oooo0@0000000200oooo00<000000?ooo`3oool0DP3oool0
0`000000oooo0?ooo`240?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0O`3oool1
00000000803oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo08<0oooo00<000000?oo
o`3oool0;P3oool00`000000oooo0?ooo`1o0?ooo`400000000P0?ooo`030000003oool0oooo0580
oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo0800
oooo0@0000000200oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`220?ooo`030000
003oool0oooo02h0oooo00<000000?ooo`3oool0P03oool100000000803oool00`000000oooo0?oo
o`1C0?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?oo
o`210?ooo`400000000P0?ooo`800000E03oool00`000000oooo0?ooo`210?ooo`030000003oool0
oooo02d0oooo00<000000?ooo`3oool0P03oool200000000803oool00`000000oooo0?ooo`1C0?oo
o`030000003oool0oooo0;00oooo00<000000?ooo`3oool0PP3oool100000000803oool00`000000
oooo0?ooo`1C0?ooo`030000003oool0oooo0:l0oooo00<000000?ooo`3oool0P`3oool100000000
803oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo0:h0oooo00<000000?ooo`3oool0
Q03oool100000000803oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo0800oooo00<0
00000?ooo`3oool0:`3oool00`000000oooo0?ooo`240?ooo`400000000P0?ooo`030000003oool0
oooo05<0oooo00<000000?ooo`3oool0P03oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0
oooo08D0oooo0@0000000200oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`1o0?oo
o`030000003oool0oooo02T0oooo00<000000?ooo`3oool0QP3oool100000000803oool00`000000
oooo0?ooo`1D0?ooo`030000003oool0oooo07h0oooo00<000000?ooo`3oool0:P3oool00`000000
oooo0?ooo`260?ooo`400000000P0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0
OP3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo08L0oooo0@0000000200oooo00<0
00000?ooo`3oool0E03oool00`000000oooo0?ooo`1n0?ooo`030000003oool0oooo02T0oooo00<0
00000?ooo`3oool0Q`3oool100000000803oool00`000000oooo0?ooo`1D0?ooo`030000003oool0
oooo07h0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`280?ooo`400000000P0?oo
o`030000003oool0oooo05@0oooo00<000000?ooo`3oool0OP3oool00`000000oooo0?ooo`0W0?oo
o`030000003oool0oooo08T0oooo0@0000000200oooo00<000000?ooo`3oool0E03oool00`000000
oooo0?ooo`1n0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0RP3oool100000000
803oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo0:D0oooo00<000000?ooo`3oool0
R`3oool100000000803oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo0:D0oooo00<0
00000?ooo`3oool0R`3oool100000000803oool00`000000oooo0?ooo`1E0?ooo`030000003oool0
oooo07`0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`2<0?ooo`400000000P0?oo
o`030000003oool0oooo05D0oooo00<000000?ooo`3oool0O03oool00`000000oooo0?ooo`0T0?oo
o`030000003oool0oooo08d0oooo0@0000000200oooo00<000000?ooo`3oool0EP3oool00`000000
oooo0?ooo`1k0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0S@3oool100000000
803oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool0
8`3oool00`000000oooo0?ooo`2>0?ooo`400000000P0?ooo`800000E`3oool00`000000oooo0?oo
o`1k0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0SP3oool200000000803oool0
0`000000oooo0?ooo`1G0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool08@3oool0
0`000000oooo0?ooo`2@0?ooo`400000000P0?ooo`030000003oool0oooo05L0oooo00<000000?oo
o`3oool0N@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0940oooo0@0000000200
oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo0200
oooo00<000000?ooo`3oool0TP3oool100000000803oool00`000000oooo0?ooo`1H0?ooo`030000
003oool0oooo07P0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`2C0?ooo`400000
000P0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3oool0V@3oool00`000000oooo0?oo
o`2C0?ooo`400000000P0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3oool0V03oool0
0`000000oooo0?ooo`2D0?ooo`400000000P0?ooo`030000003oool0oooo05T0oooo00<000000?oo
o`3oool0U`3oool00`000000oooo0?ooo`2E0?ooo`400000000P0?ooo`030000003oool0oooo05T0
oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo09H0
oooo0@0000000200oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`1f0?ooo`030000
003oool0oooo01/0oooo00<000000?ooo`3oool0U`3oool100000000803oool00`000000oooo0?oo
o`1J0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?oo
o`2G0?ooo`400000000P0?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool0M@3oool0
0`000000oooo0?ooo`0K0?ooo`030000003oool0oooo09P0oooo0@0000000200oooo00<000000?oo
o`3oool0FP3oool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool0V@3oool100000000803oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo07@0
oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`2J0?ooo`400000000P0?ooo`030000
003oool0oooo05/0oooo0P00001e0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0
V`3oool100000000803oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo07<0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`2L0?ooo`400000000P0?ooo`030000003oool0
oooo05d0oooo00<000000?ooo`3oool0LP3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0
oooo09d0oooo0@0000000200oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`2:0?oo
o`030000003oool0oooo09h0oooo0@0000000200oooo00<000000?ooo`3oool0G@3oool00`000000
oooo0?ooo`290?ooo`030000003oool0oooo09l0oooo0@0000000200oooo0P00001O0?ooo`030000
003oool0oooo08L0oooo00<000000?ooo`3oool0W`3oool200000000803oool00`000000oooo0?oo
o`1N0?ooo`030000003oool0oooo0740oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`2Q0?ooo`400000000P0?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3oool0K`3oool0
0`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0:80oooo0@0000000200oooo00<000000?oo
o`3oool0G`3oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool0X`3oool100000000803oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo06h0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`2T0?ooo`400000000P0?ooo`030000
003oool0oooo0600oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`0?0?ooo`030000
003oool0oooo0:D0oooo0@0000000200oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?oo
o`1]0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0YP3oool100000000803oool0
0`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`2W0?ooo`400000000P0?ooo`030000003oool0oooo0680oooo00<000000?oo
o`3oool0K03oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0:P0oooo0@0000000200
oooo00<000000?ooo`3oool0HP3oool00`000000oooo0?ooo`1/0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool0Z@3oool100000000803oool00`000000oooo0?ooo`1S0?ooo`030000
003oool0oooo07P0oooo00<000000?ooo`3oool0ZP3oool100000000803oool00`000000oooo0?oo
o`1T0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3oool0Z`3oool100000000803oool0
0`000000oooo0?ooo`1U0?ooo`030000003oool0oooo07<0oooo0P00002^0?ooo`400000000P0?oo
o`030000003oool0oooo06D0oooo00<000000?ooo`3oool0J03oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo0:h0oooo0@00000001<0oooo00@0oooO09a80000B9`0gooo2@3oool00`00
0000oooo0?ooo`1V0?ooo`030000003oool0oooo06L0oooo00<000000?ooo`3oool01@3oool20000
0;40oooo0@00000001<0oooo00@0oooO09a80000B9`0gooo2@3oool00`000000oooo0?ooo`1W0?oo
o`030000003oool0oooo06H0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`2a0?oo
o`400000000C0?ooo`040?oog`2LB00004RL0=ooo`T0oooo00<000000?ooo`3oool0J03oool00`00
0000oooo0?ooo`1U0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0/P3oool10000
00003`3oool00`2oM0000000000000050000000307Boo`3oool0oooo00H0oooo00<000000?ooo`3o
ool0J03oool00`000000oooo0?ooo`1U0?ooo`040000003oool0oooo0?ooo`800000]@3oool10000
00003`3oool0203ogi`0M9cO0?ooo`3oool0oooO09a80000B9`0gooo2@3oool00`000000oooo0?oo
o`1Y0?ooo`030000003oool0oooo06@0oooo00@000000?ooo`3oool00000]`3oool100000000403o
ool01`3OW4P0M;oo0?ooo`3ooml0W4P00018W03Oool02@3oool3000006X0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0000002f0?ooo`<00000000@0?ooo`070?oog`2LB7@0_ooo0?oog`2L
B00004RL0=ooo`090?ooo`030000003oool0oooo06/0oooo00<000000?ooo`3oool0HP3oool20000
0;T0oooo0@0000000140oooo00H0omnL04Qd_`3ooml0W4P00018W03Oool90?ooo`030000003oool0
oooo06`0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`2h0?ooo`400000000B0?oo
o`050=nLB01d_ml0W4P00018W03Oool02@3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0
oooo05h0oooo0P00002k0?ooo`400000000B0?ooo`050?oog`2LB4P0B4P00018W03Oool02@3oool0
0`000000oooo0?ooo`1^0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3oool0^`3oool1
000000004`3oool0103ogi`0B0000018W03Oool90?ooo`030000003oool0oooo06l0oooo00<00000
0?ooo`3oool0F@3oool200000;h0oooo0@00000001@0oooo00<0gia80018W03Oool02@3oool00`00
0000oooo0?ooo`1`0?ooo`030000003oool0oooo05H0oooo0P0000050?ooo`030000003oool0oooo
0;P0oooo0@0000000200oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`1D0?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool0^@3oool100000000803oool00`000000oooo
0?ooo`1b0?ooo`030000003oool0oooo0540oooo0P0000070?ooo`030000003oool0oooo0;T0oooo
0@0000000200oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`1>0?ooo`8000002@3o
ool00`000000oooo0?ooo`2i0?ooo`400000000P0?ooo`030000003oool0oooo07@0oooo0P00001<
0?ooo`8000002`3oool00`000000oooo0?ooo`2i0?ooo`400000000P0?ooo`030000003oool0oooo
07H0oooo00<000000?ooo`3oool0A`3oool2000000d0oooo00<000000?ooo`3oool0^@3oool10000
0000803oool00`000000oooo0?ooo`1g0?ooo`800000A@3oool2000000l0oooo00<000000?ooo`3o
ool0^@3oool100000000803oool00`000000oooo0?ooo`1i0?ooo`800000@@3oool200000140oooo
00<000000?ooo`3oool0^@3oool100000000803oool00`000000oooo0?ooo`1k0?ooo`030000003o
ool0oooo03`0oooo0P00000C0?ooo`030000003oool0oooo0;T0oooo0@0000000200oooo00<00000
0?ooo`3oool0O03oool00`000000oooo0?ooo`0h0?ooo`<00000d@3oool100000000803oool00`00
0000oooo0?ooo`1m0?ooo`800000=P3oool200000=@0oooo0@0000000200oooo00<000000?ooo`3o
ool0O`3oool200000380oooo0P00003F0?ooo`400000000P0?ooo`030000003oool0oooo0840oooo
0`00000/0?ooo`<00000703oool00`000000oooo0?ooo`2i0?ooo`400000000P0?ooo`800000Q@3o
ool4000002@0oooo1000000O0?ooo`030000003oool0oooo0;P0oooo0P0000000200oooo00<00000
0?ooo`3oool0R03oool4000001X0oooo1P00000R0?ooo`030000003oool0oooo0;X0oooo0@000000
0200oooo00<000000?ooo`3oool0S03oool4000000h0oooo2000000X0?ooo`030000003oool0oooo
0;X0oooo0@0000000200oooo00<000000?ooo`3oool0T03oool>00000300oooo00<000000?ooo`3o
ool0^P3oool100000000803oool00`000000oooo0?ooo`3>0?ooo`030000003oool0oooo0;X0oooo
0@0000000200oooo00<000000?ooo`3oool0cP3oool00`000000oooo0?ooo`2j0?ooo`400000000P
0?ooo`030000003oool0oooo0<h0oooo00<000000?ooo`3oool0^P3oool100000000803oool00`00
0000oooo0?ooo`3>0?ooo`030000003oool0oooo0;X0oooo0@0000000200oooo00<000000?ooo`3o
ool0o`3ooon<0?ooo`400000000P0?ooo`030000003oool0oooo0?l0ooooS03oool100000000803o
ool00`000000oooo0?ooo`3o0?oooh`0oooo0@0000000200oooo00<000000?ooo`3oool0cP3oool0
0`000000oooo0?ooo`2j0?ooo`400000000P0?ooo`030000003oool0oooo0<h0oooo00<000000?oo
o`3oool0^P3oool100000000803oool00`000000oooo0?ooo`3>0?ooo`030000003oool0oooo0;X0
oooo0@0000000200oooo00<000000?ooo`3oool0cP3oool00`000000oooo0?ooo`2j0?ooo`400000
000P0?ooo`030000003oool0oooo0<h0oooo00<000000?ooo`3oool0^P3oool100000000803oool0
0`000000oooo0?ooo`3>0?ooo`030000003oool0oooo0;X0oooo0@0000000200oooo00<000000?oo
o`3oool0cP3oool00`000000oooo0?ooo`2j0?ooo`400000000P0?ooo`030000003oool0oooo0<h0
oooo00<000000?ooo`3oool0^P3oool100000000803oool200000<l0oooo00<000000?ooo`3oool0
^@3oool200000000803oool00`000000oooo0?ooo`3o0?oooh`0oooo0@0000000200oooo00<00000
0?ooo`3oool0o`3ooon<0?ooo`400000000P0?ooo`030000003oool0oooo0?l0ooooS03oool10000
0000803oool00`000000oooo0?ooo`3o0?oooh`0oooo0@0000000200oooo00<000000?ooo`3oool0
o`3ooon<0?ooo`400000000P0?ooo`030000003oool0oooo0?l0ooooS03oool100000000803oool0
0`000000oooo0?ooo`0/0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0B@3oool0
0`000000oooo0?ooo`190?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0;03oool1
00000000803oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`060?ooo`400000000P0?ooool00000T00000000?l0oooo
/03oool00?l0oooo/03oool00001\
\>"],
  ImageRangeCache->{{{0, 430.813}, {430.813, 0}} -> {-6.10566, -6.05622, \
0.0395555, 0.0395555}}]
}, Closed]],

Cell[TextData[{
  "each of the intersections provides a solution.  By visual inspection or by \
using the cursor, we can now find approximate values to supply ",
  StyleBox["FindRoot", "InlineInput",
    FontWeight->"Bold"],
  " as starting conditions."
}], "Text"],

Cell[TextData[{
  "Plots of this type can also be made for data in the form of lists using ",
  ButtonBox["ListDensityPlot",
    ButtonStyle->"RefGuideLink"],
  " or ",
  ButtonBox["ListContourPlot",
    ButtonStyle->"RefGuideLink"],
  ".  Three-dimensional contour plots can be made using ",
  ButtonBox["ContourPlot3D",
    ButtonStyle->"RefGuideLink"],
  " or ",
  ButtonBox["ListContourPlot3D",
    ButtonStyle->"RefGuideLink"],
  " in the ",
  StyleBox["Graphics`ContourPlot3D`", "InlineInput",
    FontWeight->"Bold"],
  " package."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Creating a legend", "Section"],

Cell[CellGroupData[{

Cell["Using the Legend package", "Subsection"],

Cell[TextData[{
  "The simplest method for adding a legend to a figure is to use the ",
  StyleBox["Graphics`Legend`", "InlineInput",
    FontWeight->"Bold"],
  " package.  The option ",
  StyleBox["PlotLegend\[Rule]{", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["text1,text2,\[CenterEllipsis]", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["}", "InlineInput",
    FontWeight->"Bold"],
  " initiates the legend and assigns text to each line style cyclically.  The \
appearance of the legend can then be modified using ",
  StyleBox["LegendPosition", "InlineInput",
    FontWeight->"Bold"],
  ", ",
  StyleBox["LegendSize", "InlineInput",
    FontWeight->"Bold"],
  ", and other options.  The position and size variables are supposed to use \
the longest plot dimensions and scale that dimension to be in the range ",
  StyleBox["{-1,1}", "InlineInput"],
  " with 0 at the center.  An example is provided below and more examples and \
options can be found in the Help Browser."
}], "Text"],

Cell[BoxData[{
    \(\(labelJ = 
        StyleForm[\*"\"\<\!\(J\_2\)\>\"", FontFamily \[Rule] "\<Helvetica\>", 
          FontColor \[Rule] RGBColor[1, 0, 0], 
          FontSize \[Rule] 12];\)\), "\n", 
    \(\(labelY = 
        StyleForm[\*"\"\<\!\(Y\_2\)\>\"", FontFamily \[Rule] "\<Helvetica\>", 
          FontColor \[Rule] RGBColor[0, 1, 0], 
          FontSize \[Rule] 12];\)\), "\n", 
    \(\(labelLegend = 
        StyleForm["\<Bessel  \[NonBreakingSpace]functions\>", 
          FontFamily \[Rule] "\<Helvetica\>", 
          FontColor \[Rule] RGBColor[0, 1, 0], 
          FontSize \[Rule] 12];\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Evaluate[{BesselJ[2, x], BesselY[2, x]}], {x, 0, 20}, 
        PlotStyle \[Rule] {{RGBColor[1, 0, 0]}, {Dashing[{0.02, 0.02}], 
              RGBColor[0, 1, 0]}}, PlotLegend \[Rule] {labelJ, labelY}, 
        LegendPosition \[Rule] {1, 0}, LegendSize \[Rule] {1. , 0.5}, 
        LegendLabel \[Rule] labelLegend];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .40527 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.336066 0.312256 0.202634 0.312256 [
[ 0 0 0 0 ]
[ 1 .40527 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
% Start of sub-graphic
p
0.0238095 0.00964924 0.648322 0.395619 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.047619 0.420092 0.376624 [
[.2619 .40759 -5 -15.6875 ]
[.2619 .40759 5 0 ]
[.5 .40759 -8 -15.6875 ]
[.5 .40759 8 0 ]
[.7381 .40759 -8 -15.6875 ]
[.7381 .40759 8 0 ]
[.97619 .40759 -8 -15.6875 ]
[.97619 .40759 8 0 ]
[.01131 .04347 -18.3125 -7.84375 ]
[.01131 .04347 0 7.84375 ]
[.01131 .13762 -33.3125 -7.84375 ]
[.01131 .13762 0 7.84375 ]
[.01131 .23178 -27.3125 -7.84375 ]
[.01131 .23178 0 7.84375 ]
[.01131 .32594 -33.3125 -7.84375 ]
[.01131 .32594 0 7.84375 ]
[.01131 .51425 -25 -7.84375 ]
[.01131 .51425 0 7.84375 ]
[.01131 .6084 -19 -7.84375 ]
[.01131 .6084 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2619 .42009 m
.2619 .42634 L
s
gsave
.2619 .40759 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.5 .42009 m
.5 .42634 L
s
gsave
.5 .40759 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.7381 .42009 m
.7381 .42634 L
s
gsave
.7381 .40759 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(15) show
1.000 setlinewidth
grestore
.97619 .42009 m
.97619 .42634 L
s
gsave
.97619 .40759 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .42009 m
.07143 .42384 L
s
.11905 .42009 m
.11905 .42384 L
s
.16667 .42009 m
.16667 .42384 L
s
.21429 .42009 m
.21429 .42384 L
s
.30952 .42009 m
.30952 .42384 L
s
.35714 .42009 m
.35714 .42384 L
s
.40476 .42009 m
.40476 .42384 L
s
.45238 .42009 m
.45238 .42384 L
s
.54762 .42009 m
.54762 .42384 L
s
.59524 .42009 m
.59524 .42384 L
s
.64286 .42009 m
.64286 .42384 L
s
.69048 .42009 m
.69048 .42384 L
s
.78571 .42009 m
.78571 .42384 L
s
.83333 .42009 m
.83333 .42384 L
s
.88095 .42009 m
.88095 .42384 L
s
.92857 .42009 m
.92857 .42384 L
s
.25 Mabswid
0 .42009 m
1 .42009 L
s
.02381 .04347 m
.03006 .04347 L
s
gsave
.01131 .04347 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .13762 m
.03006 .13762 L
s
gsave
.01131 .13762 -94.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.75) show
1.000 setlinewidth
grestore
.02381 .23178 m
.03006 .23178 L
s
gsave
.01131 .23178 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .32594 m
.03006 .32594 L
s
gsave
.01131 .32594 -94.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.25) show
1.000 setlinewidth
grestore
.02381 .51425 m
.03006 .51425 L
s
gsave
.01131 .51425 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.25) show
1.000 setlinewidth
grestore
.02381 .6084 m
.03006 .6084 L
s
gsave
.01131 .6084 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .0623 m
.02756 .0623 L
s
.02381 .08113 m
.02756 .08113 L
s
.02381 .09996 m
.02756 .09996 L
s
.02381 .11879 m
.02756 .11879 L
s
.02381 .15646 m
.02756 .15646 L
s
.02381 .17529 m
.02756 .17529 L
s
.02381 .19412 m
.02756 .19412 L
s
.02381 .21295 m
.02756 .21295 L
s
.02381 .25061 m
.02756 .25061 L
s
.02381 .26944 m
.02756 .26944 L
s
.02381 .28827 m
.02756 .28827 L
s
.02381 .30711 m
.02756 .30711 L
s
.02381 .34477 m
.02756 .34477 L
s
.02381 .3636 m
.02756 .3636 L
s
.02381 .38243 m
.02756 .38243 L
s
.02381 .40126 m
.02756 .40126 L
s
.02381 .43892 m
.02756 .43892 L
s
.02381 .45775 m
.02756 .45775 L
s
.02381 .47659 m
.02756 .47659 L
s
.02381 .49542 m
.02756 .49542 L
s
.02381 .53308 m
.02756 .53308 L
s
.02381 .55191 m
.02756 .55191 L
s
.02381 .57074 m
.02756 .57074 L
s
.02381 .58957 m
.02756 .58957 L
s
.02381 .02464 m
.02756 .02464 L
s
.02381 .00581 m
.02756 .00581 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
1 0 0 r
.5 Mabswid
.02381 .42009 m
.02499 .42012 L
.02605 .4202 L
.02729 .42034 L
.02846 .42054 L
.03053 .42103 L
.03279 .42176 L
.03527 .4228 L
.0379 .42418 L
.04262 .42734 L
.05205 .43617 L
.06244 .44942 L
.10458 .52585 L
.12562 .56417 L
.1356 .57913 L
.14487 .59025 L
.14985 .59493 L
.15437 .59833 L
.15678 .5998 L
.15939 .60111 L
.16083 .6017 L
.16215 .60217 L
.1634 .60254 L
.16474 .60285 L
.16593 .60307 L
.16701 .6032 L
.1681 .60329 L
.16912 .60332 L
.17032 .60329 L
.17143 .60321 L
.17264 .60305 L
.17393 .60281 L
.17508 .60253 L
.17631 .60216 L
.17855 .60131 L
.1812 .59999 L
.18364 .59849 L
.18838 .59479 L
.19344 .5897 L
.20256 .57767 L
.21341 .55894 L
.22326 .53828 L
.26325 .43396 L
.28156 .38593 L
.29197 .36166 L
.30171 .34193 L
.3121 .32477 L
.32159 .31307 L
.32679 .30839 L
.33173 .30512 L
Mistroke
.33444 .30382 L
.33695 .30293 L
.33839 .30255 L
.3397 .30229 L
.34043 .30219 L
.34121 .3021 L
.34195 .30204 L
.34263 .30202 L
.3439 .30202 L
.3451 .3021 L
.34617 .30223 L
.34734 .30243 L
.34862 .30271 L
.34997 .3031 L
.35239 .30401 L
.35672 .3063 L
.36149 .30978 L
.36663 .31461 L
.37216 .32098 L
.38217 .33528 L
.42336 .41737 L
.44393 .45951 L
.45395 .47725 L
.46304 .49097 L
.47254 .5024 L
.4769 .50654 L
.4815 .51013 L
.48385 .51164 L
.48643 .51304 L
.4888 .5141 L
.49096 .51486 L
.49218 .51521 L
.49351 .51552 L
.49424 .51566 L
.49492 .51577 L
.49561 .51586 L
.49625 .51593 L
.49739 .51601 L
.49865 .51603 L
.49988 .516 L
.50057 .51595 L
.5012 .51589 L
.50247 .51572 L
.50368 .5155 L
.50593 .51494 L
.50835 .5141 L
.51099 .51293 L
.51573 .51017 L
.5201 .50688 L
.53015 .49684 L
Mistroke
.54079 .48289 L
.57865 .41732 L
.5997 .38089 L
.60969 .36625 L
.61895 .3549 L
.62847 .34592 L
.63349 .34237 L
.63606 .34088 L
.63885 .33954 L
.64151 .33851 L
.64395 .33779 L
.64516 .33751 L
.64629 .3373 L
.64747 .33712 L
.64876 .33699 L
.64993 .33692 L
.65103 .3369 L
.65202 .33692 L
.6531 .33698 L
.65427 .3371 L
.65551 .33727 L
.65657 .33746 L
.65775 .33772 L
.6601 .33838 L
.66267 .33933 L
.66785 .34191 L
.67265 .34509 L
.67711 .3487 L
.69797 .37263 L
.73703 .43407 L
.75696 .46321 L
.76745 .47558 L
.77319 .4812 L
.77855 .48563 L
.78373 .48911 L
.78847 .49156 L
.79116 .49265 L
.79372 .49346 L
.79643 .49409 L
.79795 .49434 L
.79869 .49443 L
.79937 .4945 L
.80058 .49459 L
.80188 .49464 L
.80303 .49463 L
.80425 .49458 L
.80534 .49449 L
.80635 .49437 L
.80747 .4942 L
.80866 .49398 L
Mistroke
.81132 .49332 L
.81384 .49249 L
.81855 .49041 L
.82316 .48774 L
.82824 .48409 L
.83852 .47464 L
.85696 .45215 L
.87798 .42166 L
.89723 .394 L
.90673 .38185 L
.9171 .37044 L
.92689 .3619 L
.93163 .35865 L
.93599 .3562 L
.94106 .35404 L
.94373 .3532 L
.94509 .35285 L
.94655 .35254 L
.94783 .35232 L
.949 .35216 L
.95018 .35204 L
.95129 .35196 L
.95259 .35192 L
.95379 .35193 L
.9551 .35199 L
.95577 .35203 L
.9565 .3521 L
.95768 .35224 L
.95891 .35244 L
.96112 .3529 L
.9635 .35355 L
.9661 .35445 L
.97076 .35653 L
.97511 .35901 L
.97619 .3597 L
Mfstroke
0 1 0 r
[ .02 .02 ] 0 setdash
.08687 0 m
.09309 .05439 L
.09906 .09268 L
.10458 .12276 L
.14385 .28 L
.1816 .40935 L
.20257 .47203 L
.21175 .49548 L
.2218 .5175 L
.23091 .53374 L
.24065 .54677 L
.24603 .55196 L
.25092 .55538 L
.25328 .55659 L
.25453 .55711 L
.25584 .55758 L
.25699 .55791 L
.25826 .5582 L
.25941 .55839 L
.26049 .5585 L
.26174 .55856 L
.26244 .55856 L
.26307 .55854 L
.26431 .55843 L
.26547 .55826 L
.26613 .55813 L
.26683 .55798 L
.26808 .55763 L
.27089 .55657 L
.27333 .55533 L
.27561 .55391 L
.28077 .54978 L
.28973 .53978 L
.29931 .52544 L
.32059 .48317 L
.34072 .43586 L
.35898 .39318 L
.36937 .37115 L
.37911 .353 L
.38945 .33707 L
.39892 .32602 L
.40385 .32171 L
.40904 .3183 L
.41174 .31698 L
.41425 .31605 L
.41569 .31564 L
.417 .31535 L
.41773 .31522 L
.41852 .3151 L
.41926 .31502 L
Mistroke
.41994 .31496 L
.4212 .31491 L
.42239 .31493 L
.42346 .315 L
.42463 .31513 L
.4259 .31534 L
.42724 .31564 L
.42966 .31637 L
.43189 .31726 L
.43398 .31828 L
.43874 .32129 L
.44392 .32558 L
.44868 .33039 L
.4594 .34405 L
.4783 .37559 L
.49899 .41588 L
.52021 .45612 L
.53952 .48564 L
.54914 .49639 L
.55457 .50107 L
.55958 .50445 L
.56396 .50665 L
.56645 .50757 L
.56878 .50822 L
.57004 .50849 L
.57123 .50869 L
.57231 .50882 L
.57348 .50891 L
.57475 .50895 L
.57546 .50895 L
.5761 .50893 L
.57735 .50884 L
.57853 .50871 L
.57969 .50853 L
.58094 .50828 L
.58319 .50768 L
.58586 .50672 L
.5883 .50562 L
.5936 .5025 L
.59862 .49866 L
.60748 .48994 L
.6171 .47799 L
.6579 .41143 L
.67817 .37918 L
.68809 .36603 L
.69719 .35613 L
.70605 .3488 L
.7105 .34604 L
.71537 .34377 L
.71794 .34289 L
Mistroke
.72028 .34229 L
.72157 .34204 L
.72225 .34192 L
.72298 .34182 L
.72361 .34175 L
.7243 .34169 L
.72552 .34161 L
.72675 .34159 L
.72787 .34161 L
.72917 .34169 L
.7304 .34182 L
.73153 .34198 L
.73258 .34217 L
.73496 .34274 L
.73762 .34359 L
.74006 .34458 L
.74556 .3475 L
.75042 .35085 L
.75554 .35512 L
.77416 .37612 L
.8122 .43308 L
.83341 .46248 L
.84344 .4736 L
.8527 .48169 L
.85736 .48489 L
.86231 .48758 L
.8647 .48863 L
.86727 .48956 L
.86874 .49 L
.87009 .49034 L
.87137 .49062 L
.87274 .49086 L
.874 .49103 L
.87514 .49114 L
.87629 .49121 L
.87738 .49123 L
.87865 .49122 L
.87981 .49117 L
.88109 .49106 L
.88174 .49098 L
.88245 .49088 L
.88475 .49046 L
.88722 .48983 L
.89168 .48824 L
.89696 .48559 L
.90185 .48246 L
.91291 .47309 L
.93303 .44966 L
.97228 .39524 L
.97619 .39027 L
Mistroke
Mfstroke
MathSubEnd
P
% End of sub-graphic
0 g
.66393 .18702 m
.66393 .34315 L
.97619 .34315 L
.97619 .18702 L
F
1 g
.64832 .20263 m
.64832 .35876 L
.96058 .35876 L
.96058 .20263 L
F
0 g
.001 w
[ ] 0 setdash
.64832 .20263 m
.96058 .20263 L
.96058 .35876 L
.64832 .35876 L
.64832 .20263 L
s
% Start of sub-graphic
p
0.648322 0.202634 0.960578 0.358762 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0297619 0.297619 0.0142045 0.142045 [
[ 0 0 0 0 ]
[ 1 .5 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .5 L
0 .5 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0535714 0.0255682 0.35119 0.167614 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.309017 0.294302 [
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
0 1 0 r
.5 Mabswid
[ .13333 .13333 ] 0 setdash
.02381 .30902 m
.97619 .30902 L
s
MathSubEnd
P
% End of sub-graphic
0 g
gsave
.38988 .09659 -61 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
%%BeginResource: font Helvetica-MISO
%%BeginFont: Helvetica-MISO
/Helvetica /Helvetica-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(Y) show
70.938 18.000 moveto
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(2) show
76.125 15.813 moveto
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
% Start of sub-graphic
p
0.0535714 0.190341 0.35119 0.332386 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.309017 0.294302 [
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
1 0 0 r
.5 Mabswid
[ ] 0 setdash
.02381 .30902 m
.97619 .30902 L
s
MathSubEnd
P
% End of sub-graphic
0 g
gsave
.38988 .26136 -61 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
%%BeginResource: font Helvetica-MISO
%%BeginFont: Helvetica-MISO
/Helvetica /Helvetica-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(J) show
69.000 18.000 moveto
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(2) show
74.188 15.813 moveto
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.5 .41477 -110.031 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
%%BeginResource: font Helvetica-MISO
%%BeginFont: Helvetica-MISO
/Helvetica /Helvetica-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(Bessel) show
99.000 15.813 moveto
(  \\240) show
108.938 15.813 moveto
(functions) show
157.063 15.813 moveto
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
MathSubEnd
P
% End of sub-graphic
0 0 m
1 0 L
1 .40527 L
0 .40527 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{329.25, 133.375},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40007]0000b2000`400?l00000o`00003oo`3oooo^
0?ooo`00o`3oooo^0?ooo`00o`3oooo^0?ooo`00o`3oooo^0?ooo`00o`3oooo^0?ooo`00o`3oooo^
0?ooo`00o`3oooo^0?ooo`00o`3oooo^0?ooo`00o`3oooo^0?ooo`00o`3oooo^0?ooo`00o`3oooo^
0?ooo`00o`3oooo^0?ooo`00o`3oooo^0?ooo`00o`3oooo^0?ooo`00o`3oooo^0?ooo`00o`3oooo^
0?ooo`00o`3oooo^0?ooo`00@@3oool00`000000oooo0?ooo`0=0?ooo`03003o003oool0oooo0?l0
ooooVP3oool00440oooo00<000000?ooo`3oool03@3oool00`00o`00oooo0?ooo`3o0?oooiX0oooo
00110?ooo`800000o`3ooon[0?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooojX0oooo00110?oo
o`030000003oool0oooo0?l0ooooZP3oool00440oooo00<000000?ooo`3oool0o`3ooonZ0?ooo`00
=03oool01P2oM000000000000000000004RL0=ooo`L0oooo00<000000?ooo`3oool03P3oool00`00
o`00oooo0?ooo`3o0?oooiT0oooo000e0?ooo`030?oOW0180000M;oo00T0oooo0P00000?0?ooo`03
003o003oool0oooo0?l0ooooV@3oool003D0oooo00<0omnL04P0001d_ol02@3oool00`000000oooo
0?ooo`0>0?ooo`03003o003oool0oooo0?l0ooooV@3oool003D0oooo00<0omnL04P0001d_ol02@3o
ool00`000000oooo0?ooo`0>0?ooo`03003o003oool0oooo0?l0ooooV@3oool002H0oooo00<0oooO
09a8000000001`0000000`18W=l0oooo0?ooo`020?ooo`030?oOW0180000M;oo00T0oooo00<00000
0?ooo`3oool03P3oool00`00o`00oooo0?ooo`3o0?oooiT0oooo000e0?ooo`030?oOW0180000M;oo
00T0oooo0`00000>0?ooo`03003o003oool0oooo0?l0ooooV@3oool003D0oooo00<0omnL04P0001d
_ol02@3oool00`000000oooo0?ooo`3o0?ooojX0oooo000e0?ooo`030?oOW0180000M;oo00T0oooo
00<000000?ooo`3oool0o`3ooonZ0?ooo`00=@3oool00`3ogi`0B00007Boo`090?ooo`030000003o
ool0oooo0?l0ooooZP3oool003D0oooo00<0omnL04P0001d_ol02@3oool00`000000oooo0?ooo`3o
0?ooojX0oooo000c0?ooo`050?oog`2LB00000000000001d_ol02@3oool200000100oooo00<00?l0
0?ooo`3oool0o`3ooonH0?ooo`00=@3oool00`3o_g@0000007Boo`090?ooo`030000003oool0oooo
00l0oooo00<00?l00?ooo`3oool0o`3ooonH0?ooo`00@@3oool00`000000oooo0?ooo`0?0?ooo`03
003o003oool0oooo0?l0ooooV03oool00440oooo00<000000?ooo`3oool03`3oool00`00o`00oooo
0?ooo`3o0?oooiP0oooo00110?ooo`030000003oool0oooo0100oooo00<00?l00?ooo`3oool0o`3o
oonG0?ooo`00@@3oool200000140oooo00<00?l00?ooo`3oool0o`3ooonG0?ooo`00@@3oool00`00
0000oooo0?ooo`0@0?ooo`03003o003oool0oooo0?l0ooooU`3oool00440oooo00<000000?ooo`3o
ool0o`3ooonZ0?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooojX0oooo00110?ooo`030000003o
ool0oooo0?l0ooooZP3oool00440oooo0P00003o0?oooj/0oooo00110?ooo`030000003oool0oooo
0140oooo00<00?l00?ooo`3oool0o`3ooonF0?ooo`00@@3oool00`000000oooo0?ooo`0A0?ooo`03
003o003oool0oooo0?l0ooooUP3oool00440oooo00<000000?ooo`3oool04@3oool00`00o`00oooo
0?ooo`3o0?oooiH0oooo000M0?ooo`050?oOW018000000000000M02oool00`3oool00`3o_g@00000
07Boo`040?ooo`030=nLB018W=l0oooo00<0oooo00<0oooO09a8000000000P0000000`1d_ol0oooo
0?ooo`070?ooo`8000004`3oool00`00o`00oooo0?ooo`3o0?oooiD0oooo000L0?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo00@0oooo00<0omnL04Q8W03O
ool00`3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00L0oooo00<000000?ooo`3o
ool04P3oool00`00o`00oooo0?ooo`3o0?oooiD0oooo000L0?ooo`030;md0000B9`0gooo0080oooo
00<0_g@0001d_`3oool0203oool00`3ooml0W4Q809cOo`080?ooo`030?oog`2LB7@0_ooo00L0oooo
00<000000?ooo`3oool0o`3ooonZ0?ooo`006`3oool00`3ookl0M00004RLg`030?ooo`030=nLB000
07@0_ooo00T0oooo00<0gia804RLg`3oool0203oool00`2oM000M;oo0?ooo`060?ooo`030000003o
ool0oooo0?l0ooooZP3oool000l0oooo00<0oooO09a8000000001`0000001@18W=l0oooo0?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol02@3oool00`3ogi`0B4RL0=ooo`070?ooo`030?oog`2L
B000B9cO00L0oooo00<000000?ooo`3oool0o`3ooonZ0?ooo`006`3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo00T0oooo00<0oooO09a8B02Lgol01`3oool00`3OW4P0000007Boo`07
0?ooo`<00000o`3ooonZ0?ooo`006`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00X0oooo00<0gia8001d_`3oool00P3oool01`3Ogml0oooo0?oo_`1d000000000000M02oool01`3o
ool00`000000oooo0?ooo`0C0?ooo`03003o003oool0oooo0?l0ooooU03oool001/0oooo00<0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030?oo_`1d07@0_ooo0080oooo00H0gia8
00000000000000000000B02Lgol80?ooo`030000003oool0oooo01<0oooo00<00?l00?ooo`3oool0
o`3ooonD0?ooo`006`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00D0oooo00<0
omoO0=ooo`3oool00`3oool0202oM000M;oo0?ooo`3oool0oono07@000000000B9cO2P3oool00`00
0000oooo0?ooo`0C0?ooo`03003o003oool0oooo0?l0ooooU03oool001`0oooo00<0gia80018W03O
ool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?oo_`1dB9`0gooo00<0oooo00<0okmd001d_`3o
ool00P3oool00`2oM000M;oo0?ooo`0:0?ooo`030000003oool0oooo01@0oooo00<00?l00?ooo`3o
ool0o`3ooonC0?ooo`00703oool01`3ookl0M00007Boo`3oool0okmd0000M02oool01`3oool00`2o
M000000000000003000000060000M02oool0oooo0?ooo`3ogi`0B0000P0000000`0004P0W=oo0?oo
o`060?ooo`8000005@3oool00`00o`00oooo0?ooo`3o0?oooi<0oooo000M0?ooo`050?oo_`1d0000
00000000M02oool0203oool00`3o_g@00000000000040000000507Boo`3oool0oooo0?ooo`2oM000
0`0000000`18W=l0oooo0?ooo`050?ooo`030000003oool0oooo01@0oooo00<00?l00?ooo`3oool0
o`3ooonC0?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooojX0oooo00110?ooo`030000003oool0
oooo0?l0ooooZP3oool00440oooo00<000000?ooo`3oool0o`3ooonZ0?ooo`00@@3oool200000?l0
ooooZ`3oool00440oooo00<000000?ooo`3oool05@3oool00`00o`00oooo0?ooo`3o0?oooi80oooo
00110?ooo`030000003oool0oooo01D0oooo00<00?l00?ooo`3oool0o`3ooonB0?ooo`00@@3oool0
0`000000oooo0?ooo`0F0?ooo`03003o003oool0oooo0?l0ooooT@3oool00440oooo00<000000?oo
o`3oool05P3oool00`00o`00oooo0?ooo`3o0?oooi40oooo00110?ooo`800000603oool00`00o`00
oooo0?ooo`3o0?oooi00oooo00110?ooo`030000003oool0oooo01L0oooo00<00?l00?ooo`3oool0
o`3ooon@0?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooojX0oooo00110?ooo`030000003oool0
oooo0?l0ooooZP3oool002H0oooo00D0omnL04P000000000001d0;ooo`030?ooo`060?noM0000000
M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`070?ooo`800000o`3ooon[0?ooo`009@3o
ool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@0000007Boo`3oool0oooO
09a8M02oool00P3oool00`2oM000M;oo0?ooo`070?ooo`030000003oool0oooo0?l0ooooZP3oool0
02D0oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`0:0?ooo`030?oog`2LB7@0_ooo
00L0oooo00<000000?ooo`3oool0603oool00`00o`00oooo0?ooo`3o0?ooohl0oooo000T0?ooo`03
0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02`3oool00`2oM000M;oo0?ooo`060?ooo`03
0000003oool0oooo01P0oooo00<00?l00?ooo`3oool0o`3ooon?0?ooo`00603oool00`3ooml0W4P0
000000070000000504RLg`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`03
0?oog`2LB000B9cO00L0oooo00<000000?ooo`3oool0603oool00`00o`00oooo0?ooo`3o0?ooohl0
oooo000T0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3OW4P00000
07Boo`070?ooo`<000006@3oool00`00o`00oooo0?ooo`3o0?ooohh0oooo000T0?ooo`030?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01`3Ogml0oooo0?oo_`1d000000000000M02o
ool01`3oool00`000000oooo0?ooo`0I0?ooo`03003o003oool0oooo0?l0ooooSP3oool002@0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`060=nLB0000000000000000000
04P0W=oo203oool00`000000oooo0?ooo`0I0?ooo`03003o003oool0oooo0?l0ooooSP3oool002@0
oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d0000000004RL
g`X0oooo00<000000?ooo`3oool0o`3ooonZ0?ooo`009@3oool00`3OW4P004RL0=ooo`020?ooo`03
0;md0000M;l0oooo00L0oooo00<0_g@007Boo`3oool02P3oool00`000000oooo0?ooo`3o0?ooojX0
oooo000U0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`060?oOW01800000000
0000000004P0W=oo1`3oool200000?l0ooooZ`3oool002H0oooo00D0oono07@000000000001d0;oo
o`0:0?ooo`050;md00000000000000000018W=l01`3oool00`000000oooo0?ooo`3o0?ooojX0oooo
00110?ooo`030000003oool0oooo01X0oooo00<00?l00?ooo`3oool0o`3ooon=0?ooo`00@@3oool0
0`000000oooo0?ooo`0J0?ooo`03003o003oool0oooo0?l0ooooS@3oool00440oooo00<000000?oo
o`3oool06P3oool00`00o`00oooo0?ooo`3o0?ooohd0oooo00110?ooo`800000703oool00`00o`00
oooo0?ooo`3o0?oooh`0oooo00110?ooo`030000003oool0oooo01/0oooo00<00?l00?ooo`3oool0
o`3ooon<0?ooo`00@@3oool00`000000oooo0?ooo`0K0?ooo`03003o003oool0oooo0?l0ooooS03o
ool00440oooo00<000000?ooo`3oool0o`3ooonZ0?ooo`00@@3oool200000?l0ooooZ`3oool00440
oooo00<000000?ooo`3oool0o`3ooonZ0?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooojX0oooo
00110?ooo`030000003oool0oooo01d0oooo00<00?l00?ooo`3oool0hP3ooonJ000000d0oooo0011
0?ooo`030000003oool0oooo01d0oooo00<00?l00?ooo`3oool0;03oool50?l00;40ooooVP00000=
0?ooo`007@3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00D0okmd0000001d_ol0oooo0;md
00060000000404RLg`3oool0oooO09a800<0000000<0M;oo0?ooo`3oool01`3oool2000001h0oooo
00<00?l00?ooo`3oool0:`3oool00`3o0000oooo0?ooo`030?ooo`80o`00[`3ooonJ000000d0oooo
000L0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo0080
oooo00<0gia80000000000000`0000001@00B9`0gooo0?oog`2LB7@0_ooo0080oooo00<0_g@007Bo
o`3oool01`3oool00`000000oooo0?ooo`0N0?ooo`03003o003oool0oooo02T0oooo00<0o`000?oo
o`3oool01P3oool00`3o0000oooo0?ooo`2/0?oooiX000003@3oool001`0oooo00<0_g@00018W03O
ool00P3oool00`2oM00007Bo0?ooo`070?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3o
ool0103oool00`3ooml0W4Qd0;ooo`070?ooo`030000003oool0oooo01h0oooo00<00?l00?ooo`3o
ool0:03oool00`3o0000oooo0?ooo`080?ooo`030?l0003oool0oooo00X0oooo0P00o`2O0?oooiX0
00003@3oool001/0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`080?ooo`030=nL
B01d_ol0oooo00T0oooo00<0_g@007Boo`3oool01P3oool00`000000oooo0?ooo`0N0?ooo`03003o
003oool0oooo02L0oooo00<0o`000?ooo`3oool02P3oool00`3o0000oooo0?ooo`070?ooo`03003o
003oool0oooo0:00ooooVP00000=0?ooo`003`3oool00`3ooml0W4P0000000070000000504RLg`3o
ool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`090?ooo`030;mdB02Lgol0oooo00L0oooo
00<0oooO09a80018W=l01`3oool00`000000oooo0?ooo`170?ooo`030?l0003oool0oooo00`0oooo
00<0o`000?ooo`3oool01@3oool00`00o`00oooo0?ooo`050?ooo`03003o003oool0oooo09T0oooo
VP00000=0?ooo`006`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00<0
oooO09a8M02oool01`3oool00`3OW4P0000007Boo`070?ooo`<00000AP3oool00`3o0000oooo0?oo
o`0>0?ooo`030?l0003oool0oooo0080oooo0P00o`090?ooo`800?l0TP3ooonQ000000d0oooo000K
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3o_g@007Bo0?ooo`02
0?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`070?ooo`030000003oool0oooo04D0oooo
00<0o`000?ooo`3oool03`3oool01@3o0000oooo0?ooo`3oool00?l000d0oooo00<00?l00?ooo`3o
ool0S`3oool00`000000oooo0?ooo`2F0?ooo`P000003@3oool001/0oooo00<0omnL04P0001d_ol0
0`3oool00`3o_g@0001809cOo`0:0?ooo`030?oo_`1d04P0W=oo0080oooo00H0gia8000000000000
00000000B02Lgol80?ooo`030000003oool0oooo0200oooo00<00?l00?ooo`3oool08P3oool00`3o
0000oooo0?ooo`0@0?ooo`040?l0003oool0oooo003o00h0oooo00<00?l00?ooo`3oool0SP3oool0
0`000000oooo0?ooo`2F0?ooo`P000003@3oool001/0oooo00<0oono07@00018W=l00`3oool00`3O
W4P0001d0;ooo`0:0?ooo`030?oo_`1d0000B9cO0080oooo00@0oono07@000000000B9cO2P3oool0
0`000000oooo0?ooo`0P0?ooo`03003o003oool0oooo0240oooo00<0o`000?ooo`3oool04P3oool0
0`3o0000oooo0?ooo`0m0?ooo`L0o`00F`3oool00`000000oooo0?ooo`2F0?ooo`P000003@3oool0
01`0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?noW02Lgol0oooo
0080oooo00<0omnL04P00018W=l00`3oool00`2oM000M;oo0?ooo`0:0?ooo`030000003oool0oooo
0200oooo00<00?l00?ooo`3oool04P3oool00`3ooml0W4P0000000020000000307Boo`3oool0oooo
00L0oooo00<0o`000?ooo`3oool04P3oool00`3o0000oooo0?ooo`0C0?ooo`060;md000000000000
00000000B9`0gooo0`3oool01@3ogi`0B0000000000007@0_ooo01/0oooo00<0o`000?ooo`3oool0
1@3oool20?l000T0oooo00@0_g@000000000000000001000o`001@2LB0000000000000000000M;oo
02h0oooo00<0_g@0000000000000100000002@18W=l0oooo0?ooo`3oool0omnL04P000000000001d
0;ooo`050?ooo`030000003oool0oooo09H0oooo2000000=0?ooo`00703oool01`3ookl0M00007Bo
o`3oool0okmd0000M02oool01P3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo0`3o
ool01P3ogi`0B000000000000000001809cOo`L0oooo0P00000R0?ooo`03003o003oool0oooo0140
oooo00<0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`060?ooo`030?l0003oool0oooo01@0
oooo00<0o`000?ooo`3oool04`3oool00`3ogi`0B00007Boo`040?ooo`070?oo_`1d0000M;oo0?oo
o`3o_g@0001d0;ooo`0I0?ooo`030?l0003oool0oooo00P0oooo00<0o`000?ooo`3oool01P3oool3
003o000507Boo`3oool0oooo0?ooo`3ooml00P00o`020?ooo`030;md001d_ol0oooo02d0oooo00<0
gia80000000000000`0000002P00B9`0gooo0?ooo`3ookl0M00007Boo`3oool0okmd0000M02oool4
0?ooo`030000003oool0oooo09H0oooo2000000=0?ooo`007@3oool01@3ookl0M0000000000007@0
_ooo00P0oooo00H0oooO09a80000000000000000M02oool50?ooo`050;md00000000000000000018
W=l01`3oool00`000000oooo0?ooo`0Q0?ooo`03003o003oool0oooo01H0oooo00<0oooO09a8M02o
ool01P3oool00`3o0000oooo0?ooo`0B0?ooo`03003o003oool0o`0001D0oooo00<0omnL04P0001d
_ol0103oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo01H0oooo0P3o000<0?ooo`03
0?l0003oool0oooo00@0oooo00D00?l00?ooo`3ogi`0B00007Boo`080?ooo`030?oog`2LB7@0_ooo
02h0oooo00<0gia804RLg`3oool00P3oool01P3OW7@0W=oo0?ooo`2oM00004RL0=ooo`80oooo00<0
_g@0001d_`3oool00`3oool00`000000oooo0?ooo`2F0?ooo`P000003@3oool00440oooo00<00000
0?ooo`3oool08@3oool00`00o`00oooo0?ooo`0G0?ooo`030;md001d_ol0oooo00@0oooo00<0o`00
0?ooo`3oool04P3oool01@00o`00oooo0?ooo`3oool0o`000140oooo00H00?l00?ooo`3oool0omnL
04P0001d_ol30?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool05@3oool00`3o0000
oooo0?ooo`0=0?ooo`040?l0003oool0oooo0?ooo`800?l00P3oool00`3ogi`0B00007Boo`090?oo
o`030;md001d_ol0oooo02/0oooo1`3o00020?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000
M02oool00`3oool00`000000oooo0?ooo`170?ooo`030;oo0000o`000?l000<00?l000<0M?oo0?oo
o`3oool0AP3oool8000000d0oooo00110?ooo`030000003oool0oooo03X0oooo00<0oooO09a80018
W=l01@3oool00`3o0000oooo0?ooo`0B0?ooo`05003o003oool0oooo0?ooo`3o00004@3oool01P00
o`00oooo0?ooo`3ogi`0B00007Boo`<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`0D0?ooo`030?l0003oool0oooo00l0oooo00<0o`000?ooo`3oool0103oool00`3ogi`0B00007Bo
o`080?ooo`030?oog`00o`00B9cO02T0oooo0`3o00040?ooo`030;mdB02Lgol0oooo0080o`0000<0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`030?ooo`030000003oool0oooo04L0oooo00<0
oonL04SoB02Lool0C03oool8000000d0oooo00110?ooo`030000003oool0oooo03X0oooo00<0gia8
0000001d_ol0103oool00`3o0000oooo0?ooo`0B0?ooo`03003o003oool0oooo00<0oooo00<0o`00
0?ooo`3oool03`3oool01@00o`00oooo0?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3o
ool00`3o_g@0001809cOo`0D0?ooo`030?l0003oool0oooo00l0oooo00<0o`000?ooo`3oool0103o
ool00`3ogi`0B00007Boo`080?ooo`030=nLB00000000?l002P0oooo00<0o`000?ooo`3oool01@3o
ool00`3ooml0W4Qd0;ooo`020?ooo`030?l000180000M;oo00<0oooo00<0okmd0000B02Lgol00`3o
ool00`000000oooo0?ooo`180?ooo`030=ooB000og@0_ooo04/0oooo2000000=0?ooo`00@@3oool2
000003L0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00@0oooo00<0o`000?ooo`3oool0
4P3oool00`00o`00oooo0?ooo`030?ooo`030?l0003oool0oooo00l0oooo00D00?l00?ooo`3ogi`0
B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol04`3oool00`3o0000
oooo0?ooo`0A0?ooo`030?l0003oool0oooo00<0oooo00<0omnL04P0001d_ol0103oool01`3Ogml0
oooo0?oo_`1d000000000000M000o`009`3oool00`3o0000oooo0?ooo`070?ooo`070?noM000M;l0
oooo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00<0oooo00<000000?ooo`3oool0
@03oool00`3ookl0M?l0003o_`060?ooo`030=ooB000odP0W?oo04X0oooo2000000=0?ooo`00@@3o
ool00`000000oooo0?ooo`0f0?ooo`060=nLB000000000000000000004P0W=oo103oool00`3o0000
oooo0?ooo`0J0?ooo`030?l0003oool0oooo00l0oooo00@00?l00?oOW0180000M;oo0`3oool00`3o
gi`0B00007Boo`030?ooo`030?noM00004P0W=oo0180oooo00<0o`000?ooo`3oool04@3oool00`00
o`00oooo0?l000040?ooo`030?oOW0180000M;oo00@0oooo00P0gia800000000000000000000B02L
gol0oooo003o02D0oooo00<0o`000?ooo`3oool0203oool01`3ookl0M01809cOo`3oool0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`030?ooo`030000003oool0oooo0400oooo00<0oono07Co
0000okl01`3oool00`3OodP00?md0;ooo`190?ooo`P000003@3oool00440oooo00<000000?ooo`3o
ool08`3oool00`00o`00oooo0?ooo`0@0?ooo`040?oo_`1d0000000004RLg`H0oooo00<0o`000?oo
o`3oool06P3oool00`3o0000oooo0?ooo`0@0?ooo`030?oOW0180000M;oo00<0oooo00<0oono07@0
0018W=l00`3oool00`3OW4P0001d0;ooo`0A0?ooo`030?l0003oool0oooo0140oooo00D00?l00?oo
o`3oool0oooo0?l000030?ooo`030?oOW0180000M;oo00@0oooo00@0oono07@000000000B9cO103o
ool00`00o`00oooo0?ooo`0Q0?ooo`030?l0003oool0oooo00T0oooo00L0oono07@00018W=l0oooo
0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool00`3oool00`000000oooo0?ooo`100?ooo`03
0?oo_`1do`000?no00P0oooo00<0gom8003oW03Oool0B03oool8000000d0oooo00110?ooo`030000
003oool0oooo02<0oooo00<00?l00?ooo`3oool04@3oool00`2oM000M;oo0?ooo`060?ooo`030?l0
003oool0oooo01/0oooo00<0o`000?ooo`3oool03`3oool00`3ogi`0B00007Boo`040?ooo`030=nL
B000B9`0gooo0080oooo00<0_g@0001d_`3oool04@3oool00`3o0000oooo0?ooo`0A0?ooo`03003o
003oool0oooo0080oooo00H0o`000?ooo`3oool0omnL04P0001d_ol50?ooo`030;md001d_ol0oooo
02L0oooo00<0o`000?ooo`3oool01@3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO
0080oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`030?ooo`030000003oool0oooo
0400oooo00<0oono07Co0000okl0203oool00`3ookl0M?l007Coo`180?ooo`P000003@3oool00440
oooo0P00000T0?ooo`03003o003oool0oooo0140oooo00H0omnL04P00000000000000000B02Lgol2
0?ooo`030?l0003oool0oooo01`0oooo00<0o`000?ooo`3oool03@3oool01@3ooml0W4P000000000
0000M;oo00@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0140oooo00<0o`000?ooo`3o
ool04@3oool00`00o`00oooo0?ooo`030?ooo`030?l0003ooml0W4P00080000000<0M;oo0?ooo`3o
ool00`3oool01P3ogi`0B000000000000000001809cOob<0oooo00<0o`000?ooo`3oool01P3oool0
203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo0P3oool01`3ookl0M00007Boo`3oool0okmd
0000M02oool0103oool00`000000oooo0?ooo`100?ooo`030?oo_`1do`000?no00@0oooo00L0oomd
003oW03Oool0oooo0?oog`2Lo`00M?oo04P0oooo2000000=0?ooo`00@@3oool00`000000oooo0?oo
o`0T0?ooo`03003o003oool0oooo0140oooo00D0_g@0000000000000000004RLg`020?ooo`030?l0
003oool0oooo0140oooo00<00?l00?ooo`3oool02@3oool00`3o0000oooo0?ooo`0>0?ooo`030?no
M00000000?l000D0oooo00D0oono07@000000000001d0;ooo`0A0?ooo`030?l0003oool0oooo0140
oooo00<00?l00?ooo`3oool01@3oool01@3o0000oooo0?noM0000000M;oo00H0oooo00D0_g@00000
00000000000004RLg`0R0?ooo`030?l0003oool0oooo00P0oooo00H0oooO09a80000000000000000
M02oool40?ooo`050?oo_`1d000000000000M02oool01@3oool00`000000oooo0?ooo`0;0?ooo`H0
0?l00`3oool6003o00@0oooo1@00o`040?ooo`H00?l04`3oool00`3ookl0M?l0003o_`040?ooo`07
0?oo_`1do`00M?oo0?ooo`3oog@00?m809coo`180?ooo`P000003@3oool00440oooo00<000000?oo
o`3oool0903oool00`00o`00oooo0?ooo`0H0?ooo`030?l0003oool0oooo0100oooo00<00?l00?oo
o`3oool02P3oool00`3o0000oooo0?ooo`0@0?ooo`03003o003oool0oooo01T0oooo00<0o`000?oo
o`3oool06P3oool00`3o0000oooo0?ooo`0@0?ooo`03003o003oool0oooo01/0oooo00<0o`000?oo
o`3oool0403oool00`00o`00oooo0?ooo`090?ooo`030000003oool0oooo0400oooo00@0_ol0003o
0000o`00M?oo103oool01@3ooi`0B?l0003o0000o`00M?oo04T0oooo2000000=0?ooo`00@@3oool0
0`000000oooo0?ooo`0T0?ooo`03003o003oool0oooo01L0oooo00<0o`000?ooo`3oool04@3oool0
0`00o`00oooo0?ooo`0;0?ooo`030?l0003oool0oooo0100oooo00<00?l00?ooo`3oool05`3oool0
0`3o0000oooo0?ooo`0K0?ooo`030?l0003oool0oooo0140oooo00<00?l00?ooo`3oool06@3oool0
0`3o0000oooo0?ooo`0@0?ooo`03003o003oool0oooo00X0oooo00<000000?ooo`3oool0?`3oool0
1P3oog@00?l0003oW02Lo`000?md0;oooe40oooo2000000=0?ooo`00@@3oool00`000000oooo0?oo
o`0n0?ooo`030?l0003oool0oooo0100oooo00<00?l00?ooo`3oool0303oool00`3o0000oooo0?oo
o`0@0?ooo`03003o003oool0oooo01L0oooo00<0o`000?ooo`3oool0703oool00`3o0000oooo0?oo
o`0@0?ooo`03003o003oool0oooo01P0oooo00<0o`000?ooo`3oool04@3oool00`00o`00oooo0?oo
o`0:0?ooo`030000003oool0oooo03h0oooo00L0oono07Co0000og@0_ooo0?ooW018o`000?no0540
oooo2000000=0?ooo`00@@3oool2000003h0oooo00<0o`000?ooo`3oool04@3oool00`00o`00oooo
0?ooo`0=0?ooo`030?l0003oool0oooo0100oooo00<00?l00?ooo`3oool05@3oool00`3o0000oooo
0?ooo`0N0?ooo`030?l0003oool0oooo0100oooo00<00?l00?ooo`3oool05`3oool00`3o0000oooo
0?ooo`0N0?ooo`030000003oool0oooo03h0oooo00<0_ol0003o001dool00P3oool00`3OodP00?l0
07Coo`1@0?ooo`P000003@3oool00440oooo00<000000?ooo`3oool0?@3oool00`3o0000oooo0?oo
o`0Q0?ooo`030?l0003oool0oooo0100oooo00<00?l00?ooo`3oool05@3oool00`3o0000oooo0?oo
o`0@0?ooo`03003o003oool0oooo00/0oooo00<0o`000?ooo`3oool0403oool00`00o`00oooo0?oo
o`0F0?ooo`030?l0003oool0oooo01l0oooo00<000000?ooo`3oool0?@3oool00`3oog@00?l0003o
_`030?ooo`040?oog`2Lo`000?md0;ooodl0oooo2000000=0?ooo`00@@3oool00`000000oooo0?oo
o`0l0?ooo`030?l0003oool0oooo02<0oooo00<0o`000?ooo`3oool09P3oool00`3o0000oooo0?oo
o`0@0?ooo`03003o003oool0oooo00d0oooo00<0o`000?ooo`3oool0403oool00`00o`00oooo0?oo
o`0D0?ooo`030?l0003oool0oooo0200oooo00<000000?ooo`3oool0?03oool0103ookl0M?l0003o
M02oool40?ooo`030?ooM000o`000?no04l0oooo2000000=0?ooo`00@@3oool00`000000oooo0?oo
o`0U0?ooo`03003o003oool0oooo01@0oooo00<0o`000?ooo`3oool08`3oool00`3o0000oooo0?oo
o`0V0?ooo`030?l0003oool0oooo0100oooo00<00?l00?ooo`3oool03P3oool00`3o0000oooo0?oo
o`0U0?ooo`030?l0003oool0oooo0140oooo00<00?l00?ooo`3oool03@3oool00`000000oooo0?oo
o`0l0?ooo`030;oo0000o`00M?oo00H0oooo00<0_ol0003o001dool0CP3oool8000000d0oooo0011
0?ooo`030000003oool0oooo02D0oooo00<00?l00?ooo`3oool0503oool00`3o0000oooo0?ooo`0T
0?ooo`030?l0003oool0oooo02@0oooo00<0o`000?ooo`3oool0403oool00`00o`00oooo0?ooo`0?
0?ooo`030?l0003oool0oooo02D0oooo00<0o`000?ooo`3oool0403oool00`00o`00oooo0?ooo`0>
0?ooo`030000003oool0oooo09H0oooo2000000=0?ooo`00>`3oool6000000<0o`009P0000000`00
o`0000000000000B000000030?l00000000000000140000000<00?l00000000000004@0000000`3o
000000000000000@00000003003o0000000000000100000000<0o`000000000000004@0000000`00
o`0000000000000@000000030?l000000000000002<0000000<0o`00000000000000400000000`00
o`0000000000000:000000D0oooo00<000000?ooo`3oool0UP3oool8000000d0oooo00110?ooo`03
0000003oool0oooo0080o`001`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool00P3oool00`00o`00oooo0?ooo`040?ooo`030000003oool0oooo00X0
oooo00<000000?l0003oool02@3oool00`000000oooo0?ooo`060?ooo`04003o003oool0oooo0000
00/0oooo00<000000?ooo`3oool00`3oool00`3o0000oooo0?ooo`030?ooo`030000003oool0oooo
00T0oooo00<00000003o003oool02@3oool00`000000oooo0?ooo`040?ooo`030?l0003oool0oooo
0080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo
00P0oooo00<0o`000000003oool02P3oool00`000000oooo0?ooo`030?ooo`03003o003oool0oooo
00@0oooo00<000000?ooo`3oool01`3oool00`3o0000oooo0000000;0?ooo`030000003oool0oooo
00<0oooo00<00?l00?ooo`3oool00`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo
09H0oooo2000000=0?ooo`00@@3oool00`000000oooo0?ooo`020?ooo`030?l0003oool0oooo0280
oooo00<00?l00?ooo`3oool0403oool00`3o0000oooo0?ooo`0A0?ooo`03003o003oool0oooo01<0
oooo00<0o`000?ooo`3oool0403oool00`00o`00oooo0?ooo`0>0?ooo`030?l0003oool0oooo02H0
oooo00<0o`000?ooo`3oool0403oool00`00o`00oooo0?ooo`0>0?ooo`030?l0003oool0oooo0100
oooo00<00?l00?ooo`3oool0403oool00`000000oooo0?ooo`2F0?ooo`P000003@3oool00440oooo
00<000000?ooo`3oool00`3oool00`3o0000oooo0?ooo`0Q0?ooo`03003o003oool0oooo0100oooo
00<0o`000?ooo`3oool04@3oool00`00o`00oooo0?ooo`0D0?ooo`030?l0003oool0oooo0100oooo
00<00?l00?ooo`3oool03@3oool00`3o0000oooo0?ooo`0W0?ooo`030?l0003oool0oooo0100oooo
00<00?l00?ooo`3oool0303oool00`3o0000oooo0?ooo`0T0?ooo`030000003oool0oooo09H0oooo
2000000=0?ooo`00@@3oool00`000000oooo0?ooo`040?ooo`030?l0003oool0oooo03<0oooo00<0
o`000?ooo`3oool04@3oool00`00o`00oooo0?ooo`0D0?ooo`030?l0003oool0oooo0100oooo00<0
0?l00?ooo`3oool0303oool00`3o0000oooo0?ooo`0@0?ooo`03003o003oool0oooo01H0oooo00<0
o`000?ooo`3oool03`3oool00`00o`00oooo0?ooo`0;0?ooo`030?l0003oool0oooo02D0oooo00<0
00000?ooo`3oool0UP3oool8000000d0oooo00110?ooo`8000001P3oool00`3o0000oooo0?ooo`0a
0?ooo`030?l0003oool0oooo0140oooo00<00?l00?ooo`3oool05P3oool00`3o0000oooo0?ooo`0@
0?ooo`03003o003oool0oooo00X0oooo00<0o`000?ooo`3oool0403oool00`00o`00oooo0?ooo`0G
0?ooo`030?l0003oool0oooo0100oooo00<00?l00?ooo`3oool02P3oool00`3o0000oooo0?ooo`0U
0?ooo`030000003oool0oooo09H0oooo2000000=0?ooo`00@@3oool00`000000oooo0?ooo`060?oo
o`030?l0003oool0oooo0300oooo00<0o`000?ooo`3oool0:P3oool00`3o0000oooo0?ooo`0M0?oo
o`030?l0003oool0oooo0100oooo00<00?l00?ooo`3oool0603oool00`3o0000oooo0?ooo`0K0?oo
o`030?l0003oool0oooo0100oooo00<00?l00?ooo`3oool04`3oool00`000000oooo0?ooo`2F0?oo
o`P000003@3oool00440oooo00<000000?ooo`3oool01`3oool00`3o0000oooo0?ooo`0_0?ooo`03
0?l0003oool0oooo02/0oooo00<0o`000?ooo`3oool06`3oool00`3o0000oooo0?ooo`0@0?ooo`03
003o003oool0oooo01T0oooo00<0o`000?ooo`3oool06`3oool00`3o0000oooo0?ooo`0?0?ooo`03
003o003oool0oooo01@0oooo00<000000?ooo`3oool0UP3oool8000000d0oooo00110?ooo`030000
003oool0oooo00P0oooo00<0o`000?ooo`3oool07P3oool00`00o`00oooo0?ooo`0<0?ooo`030?l0
003oool0oooo02`0oooo00<0o`000?ooo`3oool06`3oool00`3o0000oooo0?ooo`0@0?ooo`03003o
003oool0oooo01X0oooo00<0o`000?ooo`3oool06@3oool00`3o0000oooo0?ooo`0?0?ooo`03003o
003oool0oooo01D0oooo00<000000?ooo`3oool0A@3oool00`3oM000o`000?l000030?l000030?no
o`3oool0oooo04P0oooo2000000=0?ooo`00@@3oool2000000T0oooo00<0o`000?ooo`3oool07P3o
ool00`00o`00oooo0?ooo`0<0?ooo`030?l0003oool0oooo0100oooo00<00?l00?ooo`3oool06P3o
ool00`3o0000oooo0?ooo`0@0?ooo`03003o003oool0oooo00H0oooo00<0o`000?ooo`3oool0403o
ool00`00o`00oooo0?ooo`0L0?ooo`030?l0003oool0oooo0100oooo00<00?l00?ooo`3oool0103o
ool00`3o0000oooo0?ooo`0@0?ooo`03003o003oool0oooo01D0oooo00<000000?ooo`3oool0A@3o
ool00`3ogi`0o`180?oOo`1>0?ooo`P000003@3oool00440oooo00<000000?ooo`3oool02@3oool0
0`3o0000oooo0?ooo`0M0?ooo`03003o003oool0oooo00/0oooo00<0o`000?ooo`3oool04@3oool0
0`00o`00oooo0?ooo`0J0?ooo`030?l0003oool0oooo0140oooo00<00?l00?ooo`3oool01@3oool0
0`3o0000oooo0?ooo`0_0?ooo`030?l0003oool0oooo0140oooo00<00?l00?ooo`3oool00P3oool0
0`3o0000oooo0?ooo`0@0?ooo`03003o003oool0oooo01H0oooo00<000000?ooo`3oool0AP3oool0
0`3oW4P0o`1d0?ooo`1=0?ooo`P000003@3oool00440oooo00<000000?ooo`3oool02@3oool00`3o
0000oooo0?ooo`0N0?ooo`03003o003oool0oooo00X0oooo00<0o`000?ooo`3oool0403oool00`00
o`00oooo0?ooo`0K0?ooo`030?l0003oool0oooo0140oooo00<00?l00?ooo`3oool0103oool00`3o
0000oooo0?ooo`0a0?ooo`030?l0003oool0oooo0140oooo00@00?l00?ooo`3oool0o`00;03oool0
0`000000oooo0?ooo`0m0?ooo`050?oo_`3o0000o`000?l0003o07@01@3oool00`3oW4P0o`180?oO
o`1<0?ooo`P000003@3oool00440oooo00<000000?ooo`3oool02P3oool00`3o0000oooo0?ooo`0M
0?ooo`03003o003oool0oooo00X0oooo00<0o`000?ooo`3oool0403oool00`00o`00oooo0?ooo`0L
0?ooo`030?l0003oool0oooo0140oooo00<00?l00?ooo`3oool00P3oool00`3o0000oooo0?ooo`0c
0?ooo`030?l0003oool0oooo0140oooo00<00?l00?ooo`3o0000;03oool00`000000oooo0?ooo`0l
0?ooo`030?oo_`3o0000o`0000<0o`0000<0o`1d0?ooo`3oool00`3oool00`3oW4P0o`1d0?ooo`1;
0?ooo`P000003@3oool00440oooo00<000000?ooo`3oool02P3oool00`3o0000oooo0?ooo`0Y0?oo
o`030?l0003oool0oooo0100oooo00<00?l00?ooo`3oool07@3oool00`3o0000oooo0?ooo`0A0?oo
o`03003o003oool0oooo0080oooo00<0o`000?ooo`3oool0403oool00`00o`00oooo0?ooo`0Q0?oo
o`030?l0003oool0oooo0140oooo00<00?l00?ooo`3oool0:`3oool00`000000oooo0?ooo`0l0?oo
o`030?noM03o0000ogBo0080oooo00<0oia80?l0003o_ol01@3oool00`3oW4P0odRL0?ooo`1:0?oo
o`P000003@3oool00440oooo0P00000<0?ooo`030?l0003oool0oooo02P0oooo00<0o`000?ooo`3o
ool0403oool00`00o`00oooo0?ooo`0N0?ooo`030?l0003oool0oooo0140oooo00@00?l00?ooo`3o
ool0o`004P3oool00`00o`00oooo0?ooo`0S0?ooo`030?l0003oool0oooo00l0oooo00<0o`000?oo
o`3oool03`3oool00`00o`00oooo0?ooo`0J0?ooo`030000003oool0oooo03`0oooo00<0oia80?l0
B03ogol00P3oool00`3ogi`0o`000?nLg`050?ooo`030?oo_`3o0000okoo04X0oooo2000000=0?oo
o`00@@3oool00`000000oooo0?ooo`0;0?ooo`030?l0003oool0oooo02P0oooo00<0o`000?ooo`3o
ool0<P3oool00`3o0000oooo0?ooo`0C0?ooo`030?l0003oool0oooo0100oooo00<00?l00?ooo`3o
ool0903oool00`3o0000oooo0?ooo`0<0?ooo`80o`004@3oool00`00o`00oooo0?ooo`0K0?ooo`03
0000003oool0oooo0440oooo00H0oono0?l0003oW=l0oooo0?noM03oB9`20?ooo`030?oog`3oB000
okoo04X0oooo2000000=0?ooo`00@@3oool00`000000oooo0?ooo`0<0?ooo`030?l0003oool0oooo
02H0oooo00<0o`000?ooo`3oool0<`3oool00`3o0000oooo0?ooo`0B0?ooo`030?l0003oool0oooo
0100oooo00<00?l00?ooo`3oool09P3oool00`3o0000oooo0?ooo`0:0?ooo`030?l0003oool0oooo
00@0oooo0P00o`0:0?ooo`03003o003oool0oooo01`0oooo00<000000?ooo`3oool02`3ooolS0?l0
01<0oooo00/0oono0?l0003oW=l0oooo0?oo_`3o0000okoo0?ooo`3o_g@0o`180?oOo`1:0?ooo`P0
00003@3oool00440oooo00<000000?ooo`3oool0303oool00`3o0000oooo0?ooo`0M0?ooo`03003o
003oool0oooo00H0oooo00<0o`000?ooo`3oool0=03oool00`3o0000oooo0?ooo`0@0?ooo`030?l0
003oool0oooo0100oooo00<00?l00?ooo`3oool0:03oool20?l000T0oooo00<0o`000?ooo`3oool0
1`3oool00`00o`00oooo0?ooo`050?ooo`800?l07`3oool00`000000oooo0?ooo`110?ooo`030?oo
_`3o0000oicO0080oooo00D0omnL0?l0003o0000o`000?noo`1;0?ooo`P000003@3oool00440oooo
00<000000?ooo`3oool03@3oool00`3o0000oooo0?ooo`0L0?ooo`03003o003oool0oooo00D0oooo
00<0o`000?ooo`3oool0=P3oool00`3o0000oooo0?ooo`0?0?ooo`030?l0003oool00?l003d0oooo
0P3o00040?ooo`<0o`002`3oool3003o02D0oooo00<000000?ooo`3oool0@@3oool00`3ookl0o`00
0?nLg`1B0?ooo`P000003@3oool001h0oooo00D0omnL04P000000000001d0;ooo`030?ooo`050?no
M0000000M;oo0?ooo`2oM0001P0000001018W=l0oooo0?oog`2LB0030000000307Boo`3oool0oooo
00H0oooo0P00000?0?ooo`030?l0003oool0oooo01`0oooo00<00?l00?ooo`3oool0103oool00`3o
0000oooo0?ooo`0@0?ooo`03003o003oool0oooo02@0oooo00<0o`000?ooo`3oool03@3oool01@3o
0000oooo0?ooo`3oool00?l003h0oooo103o000f0?ooo`030000003oool0oooo0440oooo00<0oono
0?l0003oW=l0DP3oool8000000d0oooo000M0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oo
o`020?ooo`030?noM0000000M;oo0080oooo00<0gia80000000000000`0000001@00B9`0gooo0?oo
g`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool01P3oool00`000000oooo0?ooo`0>0?ooo`030?l0
003oool0oooo01`0oooo00<00?l00?ooo`3oool0103oool00`3o0000oooo0?ooo`0?0?ooo`03003o
003oool0oooo02D0oooo00<0o`000?ooo`3oool0303oool00`3o0000oooo0?ooo`030?ooo`800?l0
MP3oool00`000000oooo0?ooo`110?ooo`030?oo_`3o0000oicO0580oooo2000000=0?ooo`007@3o
ool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0gia804RLg`3oool00P3o
ool00`3OW7@0W=oo0?ooo`040?ooo`030?oog`2LB7@0_ooo00H0oooo00<000000?ooo`3oool03`3o
ool00`3o0000oooo0?ooo`0L0?ooo`03003o003oool0oooo0080oooo00<0o`000?ooo`3oool0403o
ool00`00o`00oooo0?ooo`0V0?ooo`030?l0003oool0oooo00X0oooo00<0o`000?ooo`3oool01P3o
ool00`00o`00oooo0?ooo`060?ooo`03003o003oool0oooo06X0oooo00<000000?ooo`3oool0@@3o
ool00`3ookl0o`000?nLg`1B0?ooo`P000003@3oool001`0oooo00<0oono07@00018W=l00`3oool0
0`3OW4P0001d0;ooo`080?ooo`030=nLB01d_ol0oooo00T0oooo00<0_g@007Boo`3oool01@3oool0
0`000000oooo0?ooo`0?0?ooo`030?l0003oool0oooo01`0oooo00<00?l00?ooo`3oool00P3oool0
0`3o0000oooo0?ooo`0?0?ooo`03003o003oool0oooo02P0oooo00<0o`000?ooo`3oool0203oool0
0`3o0000oooo0?ooo`0;0?ooo`D00?l0K@3oool00`000000oooo0?ooo`110?ooo`030?oo_`3o0000
oicO0580oooo2000000=0?ooo`00703oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00T0oooo00<0_gA809cOo`3oool01`3oool00`3ooml0W4P004RLg`060?ooo`030000003oool0oooo
0100oooo00<0o`000?ooo`3oool0803oool00`3o0000oooo0?ooo`0?0?ooo`03003o003oool0oooo
02T0oooo0`3o00050?ooo`80o`00P03oool00`000000oooo0?ooo`110?ooo`030?oo_`3o0000oicO
0580oooo2000000=0?ooo`00703oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0
oooo00<0oooO09a8M02oool01`3oool00`3OW4P0000007Boo`060?ooo`<00000403oool00`3o0000
oooo0?ooo`0O0?ooo`030?l0003oool0oooo03h0oooo1P3o00220?ooo`030000003oool0oooo09H0
oooo2000000=0?ooo`00703oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo
00<0okmd001d_`3oool00P3oool01`3Ogml0oooo0?oo_`1d000000000000M02oool01P3oool00`00
0000oooo0?ooo`0A0?ooo`030?l0003oool0oooo01h0oooo00<0o`000?ooo`3oool0aP3oool00`00
0000oooo0?ooo`2F0?ooo`P000003@3oool001`0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`0:0?ooo`030?oo_`1d04P0W=oo0080oooo00H0gia800000000000000000000B02Lgol7
0?ooo`030000003oool0oooo0140oooo00<0o`000?ooo`3oool0703oool00`00o`00o`000?ooo`38
0?ooo`030000003oool0oooo09H0oooo2000000=0?ooo`00703oool00`3ookl0M00004RLg`030?oo
o`030=nLB00007@0_ooo00X0oooo00<0oono07@00018W=l00P3oool0103ookl0M00000000018W=l9
0?ooo`030000003oool0oooo0180oooo00<0o`000?ooo`3oool0703oool00`00o`00oooo0?ooo`37
0?ooo`030000003oool0oooo09H0oooo2000000=0?ooo`007@3oool00`3OW4P004RL0=ooo`020?oo
o`030;md0000M;l0oooo00D0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`030?oo
o`030;md001d_ol0oooo00T0oooo00<000000?ooo`3oool04`3oool00`3o0000oooo0?ooo`0K0?oo
o`03003o003oool0oooo00h0oooo00<00?l00?ooo`3oool0]P3oool00`000000oooo0?ooo`2F0?oo
o`P000003@3oool001d0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00H0oooo00P0oooO
09a8B02Lgol0oooo0?oo_`1d0000001d0;ooo`<0oooo00H0omnL04P00000000000000000B02Lgol6
0?ooo`800000503oool00`3o0000oooo0?ooo`0J0?ooo`030?l0003oool00?l000h0oooo00<00?l0
0?ooo`3oool0]`3oool00`000000oooo0?ooo`2F0?ooo`P000003@3oool001h0oooo00D0oono07@0
00000000001d0;ooo`080?ooo`060?oog`2LB00000000000000007@0_ooo1@3oool01@2oM0000000
000000000000B9cO00H0oooo00<000000?ooo`3oool0503oool00`3o0000oooo0?ooo`0I0?ooo`03
0?l0003oool00?l000h0oooo00<00?l00?ooo`3oool0]`3oool00`000000oooo0?ooo`2F0?ooo`P0
00003@3oool00440oooo00<000000?ooo`3oool0503oool00`3o0000oooo0?ooo`0H0?ooo`050?l0
003oool0oooo0?ooo`00o`00303oool00`00o`00oooo0?ooo`2h0?ooo`030000003oool0oooo09H0
oooo2000000=0?ooo`00@@3oool00`000000oooo0?ooo`0E0?ooo`030?l0003oool0oooo01L0oooo
00<0o`000?ooo`3oool03@3oool00`00o`00oooo0?ooo`2i0?ooo`030000003oool0oooo09H0oooo
2000000=0?ooo`00@@3oool2000001H0oooo00<0o`000?ooo`3oool05P3oool00`3o0000oooo0?oo
o`3:0?ooo`030000003oool0oooo09H0oooo2000000=0?ooo`00@@3oool00`000000oooo0?ooo`0F
0?ooo`030?l0003oool0oooo01D0oooo00<0o`000?ooo`3oool01P3oool00`00o`00oooo0?ooo`31
0?ooo`030000003oool0oooo09H0oooo2000000=0?ooo`00@@3oool00`000000oooo0?ooo`0F0?oo
o`030?l0003oool0oooo01@0oooo00<0o`000?ooo`3oool01`3oool5003o0;l0oooo00<000000?oo
o`3oool0103oool00`3OodP00?l0003o0004003o0003003oM02oool0oooo00<0oooo00<0oonL04So
0000o`000P00o`001P1dool0oooo0?ooo`3oool0oooO09co00<00?l000H00?no0?ooo`3oool0oooo
0?oog`2Lo`03003o0003003o_`3oool0oooo0080oooo00<0oonL04So0000o`000P00o`001`1dool0
oooo0?ooo`3ooml0W?l0003oW03Oool0403oool00`3oog@00?l007Coo`030?ooo`0<0=ooB000o`00
0?l0003oW03Ooml0W?l0003o_`3oool0oooO09co0000oi`0gooo0P3oool00`3ooml0W?l0003o_`03
0?ooo`050?ooM000o`000?l0003o001dool00`3oool0203ookl0M?l0003o0000oi`0gooO09co0000
oi`0gooo0P3oool01P3ooi`0B?l0003o0000o`000?md0;ooo`80oooo00@0oooO09co0000oi`0gooo
0P3oool00`3ooml0W?l0003o_`020?ooo`030?oog`2Lo`000?l000800?l000<00?no0?ooo`3oool0
1P3oool8000000d0oooo00110?ooo`030000003oool0oooo01L0oooo00<0o`000?ooo`3oool04P3o
ool00`3o0000oooo0?ooo`3<0?ooo`030000003oool0oooo00@0oooo00<0gom8003o0000o`001@00
o`001@00odP0W?oo0?ooo`3oool0oomd00H00?l000@0B?oO0?ooo`3ooml0W?l01@00o`001000okl0
oooo0?oog`2Lo`05003o0004003o_`3oool0oooo0?ooM0H00?l000H0B?oO0?ooo`3ooml0W?l0003o
W03Oool@0?ooo`030?ooM000o`00M?oo0080oooo00<0oomd003o0000o`000P00o`002000og@0M?l0
003o_`3oool0oooO09co0000oi`0gooo0P3oool00`3ooml0W?l0003o_`020?ooo`030?ooM000o`00
0?l000<00?l000@0M?oo0?ooo`3oool0gom80P00o`001`18oml0oooO09co0000oi`0gooo0?ooo`3o
og@01@00o`001`00odP0W?oo0?ooo`3ooml0W?l0003oW03Oool00P3oool01P3ooml0W?l0003o_`3o
ool0oooO09co00D00?l000<00?no0?ooo`3oool01@3oool8000000d0oooo00110?ooo`030000003o
ool0oooo01P0oooo00<0o`000?ooo`3oool04@3oool00`3o0000oooo0?ooo`3<0?ooo`030000003o
ool0oooo00@0oooo00<0gom8003oM02oool0103oool0203OodP00?l0003o_`3oool0oono07Co0000
odP0W?oo0P3oool01`3ooml0W?l0003oM02oool0oomd003o0018oml00`3oool01P2oo`000?m809co
o`3oog@00?l004Sog`<0oooo00L0_ol0003oB02Lool0oono07Co0000odP0W?oo0080oooo00P0oooO
09co0000og@0_ooo0?oog`2Lo`000?nL0=oooa00oooo00<0oomd003o001dool00P3oool03@2oo`00
0?l007Coo`3oool0oooO09co0000o`000?no0?ooo`3ooml0W?l0003oW03Oool00P3oool06P3ooml0
W?l0003o_`3oool0oooO09co0000odP0W?oo0?ooo`3ookl0M?l0003oW03Oool0oooo0=ooB000odP0
W?oo0?ooo`3ooml0W?l0003oW03Oool0oooO09co0000o`00M?oo0P3oool0203OodP00?l0003oW03O
ool0oooO09co0000oi`0gooo0P3oool01`3ooml0W?l0003o_`3oool0oomd003o0018oml00`3oool0
0`2oo`000?m809coo`060?ooo`P000003@3oool00440oooo0P00000I0?ooo`030?l0003oool0oooo
0100oooo00<0o`000?ooo`3oool0c@3oool00`000000oooo0?ooo`040?ooo`030=ooB000og@0_ooo
00@0oooo00L0oooO09co0000oi`0gooo0?ooW018o`00M?oo00`0oooo00@0oooO09co0000odP0W?oo
1@3oool01`3ooml0W?l0003oB02Lool0oonL04So001dool01`3oool0103ooml0W?l0003oW03Oool@
0?ooo`030?ooM000o`00M?oo0080oooo00<0_ol0003oM02oool00P3oool0203ookl0M?l0003o_`3o
ool0oooO09co0000oi`0gooo0P3oool01`3ooml0W?l0003o_`3oool0oonL04So0018oml01`3oool0
2`3OodP00?md0;ooo`3oool0oooO09co0000oi`0gooo0?ooW018o`000?no00@0oooo00L0_ol0003o
B02Lool0oooO09co0000oi`0gooo0080oooo00<0oooO09co0000okl01P3oool0103ooml0W?l0003o
B02Lool60?ooo`P000003@3oool00440oooo00<000000?ooo`3oool06@3oool00`3o0000oooo0?oo
o`0?0?ooo`030?l0003oool0oooo0<d0oooo00<000000?ooo`3oool0103oool00`3OodP00?md0;oo
o`040?ooo`050?oog`2Lo`000?nL0=ooo`3oog@01`00o`001P00odP0W?oo0?ooo`3oool0oonL04So
00<00?l000H00?md0;ooo`3oool0oooo0?ooW018o`03003o0003003oM02oool0oomd00L00?l000H0
0?m809coo`3ooml0W?l0003oW03Oool@0?ooo`080?ooM000o`00M?oo0?ooo`3ooml0W?l0003oW03O
ool20?ooo`080?oog`2Lo`000?no0?ooo`3ooml0W?l0003oW03Oool20?ooo`070?oog`2Lo`000?no
0?ooo`3ooi`0B?l007Coo`070?ooo`0;0=ooB000og@0_ooo0?ooo`3ooml0W?l0003oW03Oool0oonL
04So001dool0103oool01`3oog@00?m809coo`3ooml0W?l0003oW03Oool00P3oool00`3ooml0W?l0
003o_`030?ooo`030?ooW018o`000?l000800?l000<00?md0;ooo`3oool01@3oool8000000d0oooo
00110?ooo`030000003oool0oooo01X0oooo00<0o`000?ooo`3oool03@3oool00`3o0000oooo0?oo
o`3>0?ooo`030000003oool0oooo00@0oooo00<0gom8003oM02oool0103oool01@3OodP00?l004So
g`3oool0oomd00L00?l000@00?m809coo`3oool0_ol01000o`001@00oi`0gooo0?ooo`3oool0_ol0
00@00?l000@00?nL0=ooo`3oool0oomd1`00o`001P00odP0W?oo0?oog`2Lo`000?nL0=oooa00oooo
00P0oomd003o001dool0oooo0?oog`2Lo`000?nL0=ooo`80oooo00P0oooO09co0000okl0oooo0?oo
g`2Lo`000?nL0=ooo`80oooo00L0oooO09co0000okl0oooo0?ooW018o`00M?oo00L0oooo00/0gom8
003oM02oool0oooo0?oog`2Lo`000?nL0=ooo`3ooi`0B?l007Coo`040?ooo`070?ooM000odP0W?oo
0?oog`2Lo`000?nL0=ooo`020?ooo`030?oog`2Lo`000?no0080oooo00<0_ol0003o0000o`000P00
o`000`00oi`0gooo0?ooo`060?ooo`P000003@3oool00440oooo00<000000?ooo`3oool06`3oool0
0`3o0000oooo0?ooo`0;0?ooo`030?l0003oool0oooo0<l0oooo00<000000?ooo`3oool0103oool0
0`3OodP00?l0003o0005003o0006003o_`3oool0oooo0?ooW018o`00M?oo103oool01`3OodP00?md
0;ooo`3ookl0M?l0003o0000okl01@3oool0103ookl0M?l0003o0000okl50?ooo`030?ooW018o`00
M?oo00@0oooo00L0gom8003oM02oool0oooO09co0000oi`0gooo0100oooo00P0oomd003o001dool0
oooo0?oog`2Lo`000?nL0=ooo`80oooo00P0oooO09co0000okl0oooo0?oog`2Lo`000?md0;ooo`80
oooo00L0oono07Co0000okl0oooo0?ooW018o`00B?oO00L0oooo00/0gom8003oM02oool0oooo0?oo
g`2Lo`000?nL0=ooo`3ooi`0B?l0003o_`040?ooo`070;oo0000odP0W?oo0?oog`2Lo`000?md0;oo
o`020?ooo`080?oo_`1do`000?no0?ooo`3ookl0M?l0003o0000okl;0?ooo`P000003@3oool00440
oooo00<000000?ooo`3oool06`3oool00`3o0000oooo0?ooo`0;0?ooo`030?l0003oool0oooo0<l0
oooo00<000000?ooo`3oool0103oool00`3OodP00?l0003o0005003o000707Coo`3oool0oooo0?oo
g`2Lo`000?m809coo`020?ooo`070?ooW018o`000?no0?ooo`3ooi`0B?l0003o_`020?ooo`070?oo
W018o`000?no0?ooo`3ooi`0B?l0003o_`020?ooo`080?ooW018o`000?no0?ooo`3ooml0W?l0003o
B02Lool20?ooo`080?ooW018o`000?no0?ooo`3ooml0W?l0003oW03Oool@0?ooo`080?ooM000o`00
M?oo0?ooo`3ooml0W?l0003oW03Oool20?ooo`080?oog`2Lo`000?no0?ooo`3ooml0W?l0003o0000
okl20?ooo`0<0=ooB000o`000?no0?ooo`3ooml0W?l0003oB02Lool0oooo0?ooM000o`00B?oO0P3o
ool0303OodP00?md0;ooo`3oool0oooO09co0000oi`0gooo0?oog`2Lo`000?l007Coo`80oooo00P0
gom8003o0000oi`0gooo0?oog`2Lo`000?l0003o_`80oooo00L0gom8003o0000okl0oooo0?ooW018
o`000?no0080oooo00<0oonL04So0000okl01`3oool8000000d0oooo000W0?ooo`050?oOW0180000
00000000M02oool00`3oool01P3o_g@0000007Boo`3oool0oooO09a800<0000000<0M;oo0?ooo`3o
ool01P3oool2000001d0oooo00<0o`000?ooo`3oool02@3oool00`3o0000oooo0?ooo`3@0?ooo`03
0000003oool0oooo00@0oooo00<0gom8003oM02oool0103oool00`2oo`000?m809coo`020?ooo`03
0?ooM000o`000?l000<00?l000D00?m809coo`3oool0oooO09co0005003o000407Coo`3oool0oooO
09co00D00?l000@0M?oo0?ooo`3oool0oomd1@00o`001`00odP0W?oo0?ooo`3ooml0W?l0003oW03O
ool03`3oool02@2oo`000?l0003o0000o`000?md0;oog`2Lo`000?nL0=ooo`020?ooo`060?oog`2L
o`000?no0?ooo`3ooml0W?l01P00o`00101dool0oooo0?ooo`3OodP4003o0004003oB02Lool0oooO
09co00<00?l000L0B?oO0?oog`2Lo`000?nL0=ooo`3oool0oomd00D00?l000D00?m809coo`3oool0
oooO09co0006003o000407Coo`3oool0oooO09co00D00?l000<0M?oo0?ooo`3oool01@3oool80000
00d0oooo000V0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000
M;oo0?ooo`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00H0oooo00<000000?ooo`3oool0
7@3oool00`3o0000oooo0?ooo`070?ooo`030?l0003oool0oooo0=40oooo00<000000?ooo`3oool0
103oool00`3OodP00?md0;ooo`040?ooo`030?ooM000o`00M?oo00<0oooo00H0oonL04So0000o`00
0?l0003oM02oool30?ooo`030?oog`2Lo`000?l000800?l000H0M?oo0?ooo`3oool0oooo0?oog`2L
o`03003o000307Coo`3oool0oooo0080oooo00H0oonL04So0000o`000?l0003oM02oool20?ooo`04
0?oog`2Lo`000?nL0=ooo`l0oooo00T0_ol0003o0000o`000?l0003oM02ooml0W?l0003oW03Oool0
0P3oool0203ooml0W?l0003o_`3oool0oooO09co0000oi`0gomd0`00o`000`18oml0oooo0?ooo`02
0?ooo`080?ooM000o`000?l0003oB02Lool0oooo0?oog`2Lo`03003o000504Sog`3ooml0W?l0003o
W03Oool00P3oool01P3ookl0M?l0003o0000o`000?md0;ooo`80oooo00@0oooO09co0000oi`0gomd
0`00o`001P18oml0oooo0?ooo`3oool0oooO09co00<00?l000<0M?oo0?ooo`3oool01P3oool80000
00d0oooo000V0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3ooml0
W4Qd0;ooo`060?ooo`030000003oool0oooo01h0oooo0P3o00060?ooo`030?l0003oool0oooo0=80
oooo00<000000?ooo`3oool0103oool00`3OodP00?md0;ooo`040?ooo`030;oo0000o`00M?oo02L0
oooo00@0oooO09co0000oi`0gooo403oool00`3oog@00?l007Coo`0O0?ooo`030=ooB000og@0_ooo
02P0oooo2000000=0?ooo`009@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00/0
oooo00<0_g@007Boo`3oool01@3oool00`000000oooo0?ooo`0P0?ooo`H0o`00e@3oool00`000000
oooo0?ooo`040?ooo`030=ooB000o`000?l000D00?l000<00?nL0=ooo`3oool09P3oool0103ooml0
W?l0003oW03Oool@0?ooo`050?ooM000o`000?l0003o0000okl07@3oool0203OodP00?md0;ooo`3o
ool0oooO09co0000oi`0gooo8`3oool00`000000oooo0?ooo`0B0?ooo`009@3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0oooO09a80018W=l01P3oool00`000000oooo
0?ooo`3k0?ooo`030000003oool0oooo00@0oooo00<0gom8003o0000o`001000o`000`00okl0oooo
0?ooo`0W0?ooo`040?oog`2Lo`000?nL0=oooa40oooo00D0_ol0003o0000o`000?nL0=ooo`0L0?oo
o`080?ooW018og@0_ooo0?ooo`3ooml0W?l0003oW03OoolS0?ooo`030000003oool0oooo0180oooo
000U0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3OW4P0000007Bo
o`060?ooo`<00000n`3oool00`000000oooo0?ooo`2F0?ooo`030000003oool0oooo0180oooo000U
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01`3Ogml0oooo0?oo_`1d
000000000000M02oool01P3oool00`000000oooo0?ooo`3k0?ooo`030000003oool0oooo09H0oooo
00<000000?ooo`3oool04P3oool002D0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`060?ooo`060=nLB000000000000000000004P0W=oo1`3oool00`000000oooo0?ooo`3k0?ooo`03
0000003oool0oooo09H0oooo00<000000?ooo`3oool04P3oool002D0oooo00<0oono07@00018W=l0
0`3oool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d0000000004RLgol0oooo203oool00`000000
oooo0?ooo`2F0?ooo`030000003oool0oooo0180oooo000V0?ooo`030=nLB000B9`0gooo0080oooo
00<0_g@0001d_`3oool01`3oool00`2oM000M;oo0?ooo`3o0?ooo`P0oooo00<000000?ooo`3oool0
UP3oool00`000000oooo0?ooo`0B0?ooo`009P3oool01`3ookl0M00007Boo`3oool0okmd0000M02o
ool0203oool01P3ogi`0B000000000000000001809cOool0oooo1@3ooonJ000001@0oooo000W0?oo
o`050?oo_`1d000000000000M02oool02P3oool01@2oM0000000000000000000B9cO0?l0oooo/`3o
ool00?l0ooookP3oool00?l0ooookP3oool00?l0ooookP3oool00?l0ooookP3oool00?l0ooookP3o
ool00?l0ooookP3oool00?l0ooookP3oool00?l0ooookP3oool00?l0ooookP3oool00?l0ooookP3o
ool00?l0ooookP3oool00?l0ooookP3oool00?l0ooookP3oool00?l0ooookP3oool00?l0ooookP3o
ool00?l0ooookP3oool00?l0ooookP3oool00?l0ooookP3oool00?l0ooookP3oool00?l0ooookP3o
ool00?l0ooookP3oool00001\
\>"],
  ImageRangeCache->{{{0, 492.875}, {199.063, 0}} -> {-1.08178, -0.648942, \
0.00978002, 0.00978002}, {{12.5, 319.25}, {194.313, 4.6875}} -> {-5.43368, \
-1.29205, 0.124168, 0.0156994}, {{319.25, 472.625}, {99.5, 22.8125}} -> \
{-7.09393, -4.67005, 0.032861, 0.0688518}, {{327.438, 373.063}, {95.5625, \
73.75}} -> {-9.9202, -11.0145, 0.0446261, 0.144413}, {{327.438, 373.063}, \
{70.25, 48.5}} -> {-9.95014, -13.4871, 0.0447543, 0.144828}}]
}, Closed]],

Cell[TextData[{
  "However, the actual behavior of the position and size options is \
problematical and much trial-and-error may be required to achieve the desired \
placement.  Furthermore, ",
  StyleBox["Legend", "InlineInput",
    FontWeight->"Bold"],
  " interferes with ",
  StyleBox["AspectRatio", "InlineInput",
    FontWeight->"Bold"],
  "; using both together can have unpredictable results.  We also find that \
nonalphanumeric characters in legend text, including spaces, are sometimes \
rendered poorly.  Although this package has many useful features, we feel \
that it could still use some improvement."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Using graphics primitives", "Subsection"],

Cell[TextData[{
  "Another method, which is more cumbersome but more reliable, is to add text \
using graphics primitives.  The graphics primitive ",
  ButtonBox["Text",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["expression,position", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " prints ",
  StyleBox["expression", "InlineInput"],
  " at a position specified as ",
  StyleBox["{x,y}", "InlineInput"],
  " in the units of the plot.  By default, ",
  StyleBox["position", "InlineInput"],
  " specifies the lower left corner but other choices and offsets can be \
chosen by including appropriate options.  The expression can be a string or \
text formatted using ",
  StyleBox["StyleForm", "InlineInput",
    FontWeight->"Bold"],
  ".  A graphics object based upon a graphics primitive, such as ",
  StyleBox["Text", "InlineInput",
    FontWeight->"Bold"],
  ", is created using ",
  ButtonBox["Graphics",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["primitive", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " where in this case ",
  StyleBox["primitive", "InlineInput"],
  " stands for the output of ",
  StyleBox["Text", "InlineInput",
    FontWeight->"Bold"],
  ".  ",
  StyleBox["Graphics", "InlineInput",
    FontWeight->"Bold"],
  " can also accept a list of primitives."
}], "Text"],

Cell[TextData[{
  "The example below places ",
  StyleBox["labelJ", "InlineInput",
    FontWeight->"Bold"],
  " and ",
  StyleBox["labelY", "InlineInput",
    FontWeight->"Bold"],
  ", defined in the preceding section using ",
  StyleBox["StyleForm", "InlineInput",
    FontWeight->"Bold"],
  ", as locations near the associated curves.  ",
  StyleBox["DisplayTogether", "InlineInput",
    FontWeight->"Bold"],
  " is used to combine the basic plot with the graphics objects that \
represent the labels.  One could include arrows or other graphics objects \
also."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[
        Plot[Evaluate[{BesselJ[2, x], BesselY[2, x]}], {x, 0, 20}, 
          PlotStyle \[Rule] {{RGBColor[1, 0, 0]}, {Dashing[{0.02, 0.02}], 
                RGBColor[0, 1, 0]}}], 
        Graphics[{Text[labelJ, \ {2.5, 0.6}], 
            Text[labelY, \ {2.5, \(-0.8\)}]}]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.047619 0.392698 0.351035 [
[.2619 .3802 -5 -15.6875 ]
[.2619 .3802 5 0 ]
[.5 .3802 -8 -15.6875 ]
[.5 .3802 8 0 ]
[.7381 .3802 -8 -15.6875 ]
[.7381 .3802 8 0 ]
[.97619 .3802 -8 -15.6875 ]
[.97619 .3802 8 0 ]
[.01131 .04166 -18.3125 -7.84375 ]
[.01131 .04166 0 7.84375 ]
[.01131 .12942 -33.3125 -7.84375 ]
[.01131 .12942 0 7.84375 ]
[.01131 .21718 -27.3125 -7.84375 ]
[.01131 .21718 0 7.84375 ]
[.01131 .30494 -33.3125 -7.84375 ]
[.01131 .30494 0 7.84375 ]
[.01131 .48046 -25 -7.84375 ]
[.01131 .48046 0 7.84375 ]
[.01131 .56822 -19 -7.84375 ]
[.01131 .56822 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2619 .3927 m
.2619 .39895 L
s
gsave
.2619 .3802 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.5 .3927 m
.5 .39895 L
s
gsave
.5 .3802 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.7381 .3927 m
.7381 .39895 L
s
gsave
.7381 .3802 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(15) show
1.000 setlinewidth
grestore
.97619 .3927 m
.97619 .39895 L
s
gsave
.97619 .3802 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .3927 m
.07143 .39645 L
s
.11905 .3927 m
.11905 .39645 L
s
.16667 .3927 m
.16667 .39645 L
s
.21429 .3927 m
.21429 .39645 L
s
.30952 .3927 m
.30952 .39645 L
s
.35714 .3927 m
.35714 .39645 L
s
.40476 .3927 m
.40476 .39645 L
s
.45238 .3927 m
.45238 .39645 L
s
.54762 .3927 m
.54762 .39645 L
s
.59524 .3927 m
.59524 .39645 L
s
.64286 .3927 m
.64286 .39645 L
s
.69048 .3927 m
.69048 .39645 L
s
.78571 .3927 m
.78571 .39645 L
s
.83333 .3927 m
.83333 .39645 L
s
.88095 .3927 m
.88095 .39645 L
s
.92857 .3927 m
.92857 .39645 L
s
.25 Mabswid
0 .3927 m
1 .3927 L
s
.02381 .04166 m
.03006 .04166 L
s
gsave
.01131 .04166 -79.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .12942 m
.03006 .12942 L
s
gsave
.01131 .12942 -94.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.75) show
1.000 setlinewidth
grestore
.02381 .21718 m
.03006 .21718 L
s
gsave
.01131 .21718 -88.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.02381 .30494 m
.03006 .30494 L
s
gsave
.01131 .30494 -94.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.25) show
1.000 setlinewidth
grestore
.02381 .48046 m
.03006 .48046 L
s
gsave
.01131 .48046 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.25) show
1.000 setlinewidth
grestore
.02381 .56822 m
.03006 .56822 L
s
gsave
.01131 .56822 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .05921 m
.02756 .05921 L
s
.02381 .07677 m
.02756 .07677 L
s
.02381 .09432 m
.02756 .09432 L
s
.02381 .11187 m
.02756 .11187 L
s
.02381 .14697 m
.02756 .14697 L
s
.02381 .16452 m
.02756 .16452 L
s
.02381 .18208 m
.02756 .18208 L
s
.02381 .19963 m
.02756 .19963 L
s
.02381 .23473 m
.02756 .23473 L
s
.02381 .25228 m
.02756 .25228 L
s
.02381 .26984 m
.02756 .26984 L
s
.02381 .28739 m
.02756 .28739 L
s
.02381 .32249 m
.02756 .32249 L
s
.02381 .34004 m
.02756 .34004 L
s
.02381 .35759 m
.02756 .35759 L
s
.02381 .37515 m
.02756 .37515 L
s
.02381 .41025 m
.02756 .41025 L
s
.02381 .4278 m
.02756 .4278 L
s
.02381 .44535 m
.02756 .44535 L
s
.02381 .4629 m
.02756 .4629 L
s
.02381 .49801 m
.02756 .49801 L
s
.02381 .51556 m
.02756 .51556 L
s
.02381 .53311 m
.02756 .53311 L
s
.02381 .55066 m
.02756 .55066 L
s
.02381 .02411 m
.02756 .02411 L
s
.02381 .00656 m
.02756 .00656 L
s
.02381 .58577 m
.02756 .58577 L
s
.02381 .60332 m
.02756 .60332 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
1 0 0 r
.5 Mabswid
.02381 .3927 m
.02499 .39272 L
.02605 .3928 L
.02729 .39293 L
.02846 .39312 L
.03053 .39357 L
.03279 .39425 L
.03527 .39523 L
.0379 .39651 L
.04262 .39945 L
.05205 .40769 L
.06244 .42003 L
.10458 .49127 L
.12562 .52699 L
.1356 .54093 L
.14487 .55129 L
.14985 .55566 L
.15437 .55883 L
.15678 .5602 L
.15939 .56142 L
.16083 .56197 L
.16215 .5624 L
.1634 .56275 L
.16474 .56304 L
.16593 .56324 L
.16701 .56337 L
.1681 .56345 L
.16912 .56348 L
.17032 .56345 L
.17143 .56337 L
.17264 .56323 L
.17393 .563 L
.17508 .56274 L
.17631 .5624 L
.17855 .5616 L
.1812 .56037 L
.18364 .55898 L
.18838 .55553 L
.19344 .55078 L
.20256 .53957 L
.21341 .52211 L
.22326 .50286 L
.26325 .40563 L
.28156 .36086 L
.29197 .33823 L
.30171 .31985 L
.3121 .30386 L
.32159 .29295 L
.32679 .28858 L
.33173 .28554 L
Mistroke
.33444 .28432 L
.33695 .28349 L
.33839 .28314 L
.3397 .2829 L
.34043 .2828 L
.34121 .28272 L
.34195 .28267 L
.34263 .28264 L
.3439 .28265 L
.3451 .28272 L
.34617 .28284 L
.34734 .28302 L
.34862 .28329 L
.34997 .28366 L
.35239 .28451 L
.35672 .28664 L
.36149 .28988 L
.36663 .29438 L
.37216 .30032 L
.38217 .31365 L
.42336 .39016 L
.44393 .42944 L
.45395 .44597 L
.46304 .45876 L
.47254 .46941 L
.4769 .47327 L
.4815 .47662 L
.48385 .47803 L
.48643 .47933 L
.4888 .48032 L
.49096 .48103 L
.49218 .48135 L
.49351 .48164 L
.49424 .48177 L
.49492 .48187 L
.49561 .48196 L
.49625 .48202 L
.49739 .4821 L
.49865 .48212 L
.49988 .48209 L
.50057 .48204 L
.5012 .48199 L
.50247 .48183 L
.50368 .48163 L
.50593 .4811 L
.50835 .48032 L
.51099 .47923 L
.51573 .47665 L
.5201 .47359 L
.53015 .46423 L
Mistroke
.54079 .45123 L
.57865 .39011 L
.5997 .35616 L
.60969 .34251 L
.61895 .33194 L
.62847 .32356 L
.63349 .32025 L
.63606 .31887 L
.63885 .31762 L
.64151 .31666 L
.64395 .31598 L
.64516 .31572 L
.64629 .31553 L
.64747 .31537 L
.64876 .31524 L
.64993 .31518 L
.65103 .31516 L
.65202 .31518 L
.6531 .31524 L
.65427 .31534 L
.65551 .3155 L
.65657 .31568 L
.65775 .31592 L
.6601 .31654 L
.66267 .31742 L
.66785 .31983 L
.67265 .32279 L
.67711 .32615 L
.69797 .34846 L
.73703 .40572 L
.75696 .43288 L
.76745 .44442 L
.77319 .44966 L
.77855 .45378 L
.78373 .45702 L
.78847 .45931 L
.79116 .46032 L
.79372 .46108 L
.79643 .46167 L
.79795 .4619 L
.79869 .46199 L
.79937 .46205 L
.80058 .46214 L
.80188 .46218 L
.80303 .46217 L
.80425 .46212 L
.80534 .46204 L
.80635 .46193 L
.80747 .46177 L
.80866 .46157 L
Mistroke
.81132 .46095 L
.81384 .46018 L
.81855 .45824 L
.82316 .45575 L
.82824 .45235 L
.83852 .44354 L
.85696 .42258 L
.87798 .39416 L
.89723 .36838 L
.90673 .35705 L
.9171 .34642 L
.92689 .33846 L
.93163 .33543 L
.93599 .33315 L
.94106 .33113 L
.94373 .33035 L
.94509 .33002 L
.94655 .32973 L
.94783 .32953 L
.949 .32938 L
.95018 .32927 L
.95129 .3292 L
.95259 .32916 L
.95379 .32917 L
.9551 .32922 L
.95577 .32926 L
.9565 .32933 L
.95768 .32946 L
.95891 .32964 L
.96112 .33007 L
.9635 .33067 L
.9661 .33151 L
.97076 .33346 L
.97511 .33576 L
.97619 .33641 L
Mfstroke
0 1 0 r
[ .02 .02 ] 0 setdash
.08672 0 m
.09309 .05185 L
.09906 .08753 L
.10458 .11557 L
.14385 .26212 L
.1816 .38269 L
.20257 .44111 L
.21175 .46296 L
.2218 .48349 L
.23091 .49862 L
.24065 .51077 L
.24603 .5156 L
.25092 .5188 L
.25328 .51992 L
.25453 .52041 L
.25584 .52084 L
.25699 .52115 L
.25826 .52142 L
.25941 .5216 L
.26049 .5217 L
.26174 .52176 L
.26244 .52176 L
.26307 .52173 L
.26431 .52164 L
.26547 .52148 L
.26613 .52136 L
.26683 .52121 L
.26808 .52089 L
.27089 .5199 L
.27333 .51875 L
.27561 .51743 L
.28077 .51358 L
.28973 .50425 L
.29931 .49089 L
.32059 .45149 L
.34072 .4074 L
.35898 .36761 L
.36937 .34708 L
.37911 .33016 L
.38945 .31532 L
.39892 .30501 L
.40385 .301 L
.40904 .29782 L
.41174 .2966 L
.41425 .29573 L
.41569 .29534 L
.417 .29507 L
.41773 .29495 L
.41852 .29484 L
.41926 .29476 L
Mistroke
.41994 .29471 L
.4212 .29466 L
.42239 .29468 L
.42346 .29474 L
.42463 .29487 L
.4259 .29506 L
.42724 .29534 L
.42966 .29602 L
.43189 .29685 L
.43398 .29781 L
.43874 .30061 L
.44392 .3046 L
.44868 .30909 L
.4594 .32182 L
.4783 .35122 L
.49899 .38877 L
.52021 .42628 L
.53952 .45379 L
.54914 .46381 L
.55457 .46817 L
.55958 .47133 L
.56396 .47337 L
.56645 .47423 L
.56878 .47484 L
.57004 .47509 L
.57123 .47527 L
.57231 .4754 L
.57348 .47548 L
.57475 .47552 L
.57546 .47552 L
.5761 .4755 L
.57735 .47542 L
.57853 .4753 L
.57969 .47513 L
.58094 .47489 L
.58319 .47433 L
.58586 .47344 L
.5883 .47241 L
.5936 .4695 L
.59862 .46592 L
.60748 .4578 L
.6171 .44667 L
.6579 .38462 L
.67817 .35457 L
.68809 .34231 L
.69719 .33309 L
.70605 .32625 L
.7105 .32368 L
.71537 .32156 L
.71794 .32074 L
Mistroke
.72028 .32018 L
.72157 .31994 L
.72225 .31984 L
.72298 .31975 L
.72361 .31968 L
.7243 .31962 L
.72552 .31955 L
.72675 .31953 L
.72787 .31955 L
.72917 .31962 L
.7304 .31974 L
.73153 .31989 L
.73258 .32007 L
.73496 .3206 L
.73762 .32139 L
.74006 .32231 L
.74556 .32504 L
.75042 .32816 L
.75554 .33214 L
.77416 .35171 L
.8122 .4048 L
.83341 .4322 L
.84344 .44257 L
.8527 .45011 L
.85736 .45309 L
.86231 .4556 L
.8647 .45658 L
.86727 .45744 L
.86874 .45785 L
.87009 .45818 L
.87137 .45843 L
.87274 .45865 L
.874 .45881 L
.87514 .45891 L
.87629 .45898 L
.87738 .45901 L
.87865 .45899 L
.87981 .45894 L
.88109 .45884 L
.88174 .45877 L
.88245 .45868 L
.88475 .45829 L
.88722 .4577 L
.89168 .45621 L
.89696 .45375 L
.90185 .45082 L
.91291 .4421 L
.93303 .42025 L
.97228 .36953 L
.97619 .3649 L
Mistroke
Mfstroke
0 g
gsave
.14286 .60332 -68.5938 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
%%BeginResource: font Helvetica-MISO
%%BeginFont: Helvetica-MISO
/Helvetica /Helvetica-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(J) show
69.000 18.000 moveto
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
(2) show
74.188 15.813 moveto
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
1.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.14286 .11187 -69.5625 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
%%BeginResource: font Helvetica-MISO
%%BeginFont: Helvetica-MISO
/Helvetica /Helvetica-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(Y) show
70.938 18.000 moveto
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
(2) show
76.125 15.813 moveto
%%IncludeResource: font Helvetica-MISO
%%IncludeFont: Helvetica-MISO
/Helvetica-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 1.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00=P3oool00`000000oooo0?oo
o`0E0?ooo`03003o003oool0oooo0?l0ooooH03oool003H0oooo00<000000?ooo`3oool05@3oool0
0`00o`00oooo0?ooo`3o0?ooof00oooo000f0?ooo`030000003oool0oooo01D0oooo00<00?l00?oo
o`3oool0o`3ooomP0?ooo`00=P3oool3000001H0oooo00<00?l00?ooo`3oool0o`3ooomO0?ooo`00
=P3oool00`000000oooo0?ooo`0F0?ooo`03003o003oool0oooo0?l0ooooG`3oool003H0oooo00<0
00000?ooo`3oool05P3oool00`00o`00oooo0?ooo`3o0?oooel0oooo000f0?ooo`030000003oool0
oooo0?l0ooooN03oool003H0oooo00<000000?ooo`3oool0o`3ooomh0?ooo`00=P3oool00`000000
oooo0?ooo`3o0?ooogP0oooo000f0?ooo`030000003oool0oooo0?l0ooooN03oool003H0oooo0`00
003o0?ooogP0oooo000W0?ooo`060;md00000000000000000000B9`0gooo2@3oool00`000000oooo
0?ooo`3o0?ooogP0oooo000X0?ooo`030?oOW0180000M;oo00/0oooo00<000000?ooo`3oool0o`3o
oomh0?ooo`00:03oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo01L0oooo00<00?l0
0?ooo`3oool0o`3ooomN0?ooo`00:03oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo
01L0oooo00<00?l00?ooo`3oool0o`3ooomN0?ooo`006@3oool00`3ooml0W4P00000000700000003
04RLg`3oool0oooo0080oooo00<0omnL04P0001d_ol02`3oool00`000000oooo0?ooo`0G0?ooo`03
003o003oool0oooo0?l0ooooGP3oool002P0oooo00<0omnL04P0001d_ol02`3oool4000001H0oooo
00<00?l00?ooo`3oool0o`3ooomN0?ooo`00:03oool00`3ogi`0B00007Boo`0;0?ooo`030000003o
ool0oooo01P0oooo00<00?l00?ooo`3oool0o`3ooomM0?ooo`00:03oool00`3ogi`0B00007Boo`0;
0?ooo`030000003oool0oooo01P0oooo00<00?l00?ooo`3oool0o`3ooomM0?ooo`00:03oool00`3o
gi`0B00007Boo`0;0?ooo`030000003oool0oooo01P0oooo00<00?l00?ooo`3oool0o`3ooomM0?oo
o`00:03oool00`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo01P0oooo00<00?l00?ooo`3o
ool0o`3ooomM0?ooo`009P3oool01@3ooml0W4P0000000000000M;oo00/0oooo00<000000?ooo`3o
ool0603oool00`00o`00oooo0?ooo`3o0?oooed0oooo000X0?ooo`030?noM0000000M;oo00/0oooo
00<000000?ooo`3oool0o`3ooomh0?ooo`00=P3oool300000?l0ooooN03oool003H0oooo00<00000
0?ooo`3oool0o`3ooomh0?ooo`00=P3oool00`000000oooo0?ooo`3o0?ooogP0oooo000f0?ooo`03
0000003oool0oooo0?l0ooooN03oool003H0oooo00<000000?ooo`3oool0o`3ooomh0?ooo`00=P3o
ool00`000000oooo0?ooo`0I0?ooo`03003o003oool0oooo0?l0ooooG03oool003H0oooo00<00000
0?ooo`3oool06@3oool00`00o`00oooo0?ooo`3o0?oooe`0oooo000f0?ooo`<000006@3oool00`00
o`00oooo0?ooo`3o0?oooe`0oooo000f0?ooo`030000003oool0oooo01T0oooo00<00?l00?ooo`3o
ool0o`3ooomL0?ooo`00=P3oool00`000000oooo0?ooo`0J0?ooo`03003o003oool0oooo0?l0oooo
F`3oool003H0oooo00<000000?ooo`3oool06P3oool00`00o`00oooo0?ooo`3o0?oooe/0oooo000f
0?ooo`030000003oool0oooo01X0oooo00<00?l00?ooo`3oool0o`3ooomK0?ooo`00=P3oool00`00
0000oooo0?ooo`0J0?ooo`03003o003oool0oooo0100oooo00<0_ol0003o0000o`000`00o`000`1d
ool0oooo0?ooo`3o0?oood80oooo000f0?ooo`<000006P3oool00`00o`00oooo0?ooo`0@0?ooo`03
0?ooW018odP0W?oo0?l0ooooB03oool003H0oooo00<000000?ooo`3oool0;P3oool00`3OodP00?md
0;ooo`3o0?ooodL0oooo000f0?ooo`030000003oool0oooo02D0oooo00<0oono07Co0000okl01`3o
ool00`3OodP00?m809coo`3o0?ooodH0oooo000f0?ooo`030000003oool0oooo02D0oooo00<0oono
07Co0000okl0203oool00`3OodP00?md0;ooo`3o0?ooodD0oooo000f0?ooo`030000003oool0oooo
02D0oooo00<0oono07Co0000okl02@3oool00`3OodP00?nL0=ooo`3o0?oood@0oooo000f0?ooo`03
0000003oool0oooo02D0oooo00<0oono07Co0000okl02@3oool00`3ookl0M?l007Coo`3o0?oood@0
oooo000f0?ooo`030000003oool0oooo02D0oooo00<0oono07Co0000okl01@3oool01`3oog@00?nL
0=ooo`3oool0oooO09co001dool0o`3ooom40?ooo`00=P3oool3000001`0oooo00<00?l00?ooo`3o
ool01P3oool00`3ookl0M?l0003o_`050?ooo`070?oo_`1do`00M?oo0?ooo`3oog@00?m809coo`3o
0?oood@0oooo000f0?ooo`030000003oool0oooo01`0oooo00<00?l00?ooo`3oool01P3oool0102o
o`000?l0003o001dool50?ooo`050?ooW018o`000?l0003o001dool0o`3ooom50?ooo`004@3oool0
1@3ogi`0B0000000000007@0_ooo00<0oooo00<0okmd0000001d_ol0103oool00`3OW4P0B9cO0?oo
o`030?ooo`030?oog`2LB00000000080000000<0M;oo0?ooo`3oool0203oool00`000000oooo0?oo
o`0L0?ooo`03003o003oool0oooo00D0oooo00H0oomd003o0000oi`0W?l0003oM02ooooo0?ooodh0
oooo000@0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo
00@0oooo00<0omnL04Q8W03Oool00`3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo
00P0oooo00<000000?ooo`3oool0703oool00`00o`00oooo0?ooo`040?ooo`070?oo_`1do`000?md
0;ooo`3ooi`0B?l0003o_`3o0?ooodh0oooo000@0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0
001d_`3oool0203oool00`3ooml0W4Q809cOo`080?ooo`030?oog`2LB7@0_ooo00P0oooo00<00000
0?ooo`3oool07@3oool00`00o`00oooo0?ooo`030?ooo`030;oo0000o`00M?oo0080oooo00<0gom8
003o001dool0o`3ooom=0?ooo`003`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo
00T0oooo00<0gia804RLg`3oool0203oool00`2oM000M;oo0?ooo`070?ooo`030000003oool0oooo
01d0oooo00<00?l00?ooo`3oool00P3oool00`3oog@00?l0003o_`030?ooo`040?oog`2Lo`000?md
0;ooool0ooooC03oool000<0oooo00<0oooO09a8000000001`0000001@18W=l0oooo0?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol02@3oool00`3ogi`0B4RL0=ooo`070?ooo`030?oog`2LB000
B9cO00P0oooo00<000000?ooo`3oool07@3oool02000o`00oooo0?ooo`3oool0oono07Co0000og@0
_ooo103oool00`3oog@00?l0003o_`3o0?oood`0oooo000?0?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol02@3oool00`3ooml0W4Q809cOo`070?ooo`030=nLB0000000M;oo00P0oooo
1000000M0?ooo`06003o003oool0oooo0;oo0000o`00M?oo1P3oool00`2oo`000?l007Coo`3o0?oo
od/0oooo000?0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3OW4P0
07Bo0?ooo`020?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`080?ooo`030000003oool0
oooo01h0oooo00<00?l00?ooo`3oool0o`3ooomG0?ooo`003`3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00X0oooo00<0oono07@0M02oool00P3oool01P3OW4P00000000000000000
001809cOo`T0oooo00<000000?ooo`3oool0o`3ooomh0?ooo`003`3oool00`3ookl0M00004RLg`03
0?ooo`030=nLB00007@0_ooo00D0oooo00<0omoO0=ooo`3oool00`3oool0202oM000M;oo0?ooo`3o
ool0oono07@000000000B9cO2`3oool00`000000oooo0?ooo`3o0?ooogP0oooo000@0?ooo`030=nL
B000B9`0gooo0080oooo00<0_g@0001d_`3oool01@3oool00`3ookl0M4RL0=ooo`030?ooo`030?no
M000M;l0oooo0080oooo00<0_g@007Boo`3oool02`3oool00`000000oooo0?ooo`3o0?ooogP0oooo
000@0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`070?ooo`030;md00000000000000<0
000000H0001d0;ooo`3oool0oooo0?oOW0180002000000030000B02Lgol0oooo00L0oooo00<00000
0?ooo`3oool0o`3ooomh0?ooo`004@3oool01@3ookl0M0000000000007@0_ooo00P0oooo00<0okmd
000000000000100000001@1d_ol0oooo0?ooo`3oool0_g@000<0000000<0B9cO0?ooo`3oool01P3o
ool00`000000oooo0?ooo`3o0?ooogP0oooo000f0?ooo`<00000o`3ooomh0?ooo`00=P3oool00`00
0000oooo0?ooo`0P0?ooo`03003o003oool0oooo0?l0ooooE@3oool003H0oooo00<000000?ooo`3o
ool0803oool00`00o`00oooo0?ooo`3o0?oooeD0oooo000f0?ooo`030000003oool0oooo0240oooo
00<00?l00?ooo`3oool0o`3ooomD0?ooo`00=P3oool00`000000oooo0?ooo`0Q0?ooo`03003o003o
ool0oooo0?l0ooooE03oool003H0oooo00<000000?ooo`3oool08@3oool00`00o`00oooo0?ooo`3o
0?oooe@0oooo000f0?ooo`<000008@3oool00`00o`00oooo0?ooo`3o0?oooe@0oooo000f0?ooo`03
0000003oool0oooo0280oooo00<00?l00?ooo`3oool0o`3ooomC0?ooo`00=P3oool00`000000oooo
0?ooo`0R0?ooo`03003o003oool0oooo0?l0ooooD`3oool003H0oooo00<000000?ooo`3oool0o`3o
oomh0?ooo`00=P3oool00`000000oooo0?ooo`3o0?ooogP0oooo000f0?ooo`030000003oool0oooo
0?l0ooooN03oool003H0oooo00<000000?ooo`3oool0o`3ooomh0?ooo`00=P3oool300000?l0oooo
N03oool003H0oooo00<000000?ooo`3oool0o`3ooomh0?ooo`00=P3oool00`000000oooo0?ooo`3o
0?ooogP0oooo000f0?ooo`030000003oool0oooo02@0oooo00<00?l00?ooo`3oool0o`3ooomA0?oo
o`00=P3oool00`000000oooo0?ooo`0T0?ooo`03003o003oool0oooo0?l0ooooD@3oool003H0oooo
00<000000?ooo`3oool09@3oool00`00o`00oooo0?ooo`3o0?oooe00oooo000f0?ooo`030000003o
ool0oooo02D0oooo00<00?l00?ooo`3oool0o`3ooom@0?ooo`00=P3oool3000002D0oooo00<00?l0
0?ooo`3oool0o`3ooom@0?ooo`006P3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00H0okmd
0000001d_ol0oooo0?oog`2LB0030000000307Boo`3oool0oooo00P0oooo00<000000?ooo`3oool0
9@3oool00`00o`00oooo0?ooo`3o0?oooe00oooo000I0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0
001d0;ooo`020?ooo`070?noM0000000M;oo0?ooo`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0
oooo00P0oooo00<000000?ooo`3oool09P3oool00`00o`00oooo0?ooo`3o0?ooodl0oooo000I0?oo
o`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd0;ooo`080?oo
o`030000003oool0oooo02H0oooo00<00?l00?ooo`3oool0o`3ooom?0?ooo`00603oool00`3ookl0
M00004RLg`030?ooo`030=nLB00007@0_ooo00/0oooo00<0_g@007Boo`3oool01`3oool00`000000
oooo0?ooo`3o0?ooogP0oooo000<0?ooo`030?oog`2LB000000000L0000000D0B9cO0?ooo`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0oooO09a80018W=l0203oool00`000000
oooo0?ooo`3o0?ooogP0oooo000H0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
2P3oool00`3OW4P0000007Boo`080?ooo`@00000o`3ooomg0?ooo`00603oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00H0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00P0
oooo00<000000?ooo`3oool0o`3ooomh0?ooo`00603oool00`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00H0oooo00H0gia800000000000000000000B02Lgol90?ooo`030000003oool0oooo
0?l0ooooN03oool001P0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`04
0?oo_`1d0000000004RLg`/0oooo00<000000?ooo`3oool0o`3ooomh0?ooo`006@3oool00`3OW4P0
04RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0_g@007Boo`3oool02`3oool00`000000
oooo0?ooo`3o0?ooogP0oooo000I0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?oo
o`060?oOW018000000000000000004P0W=oo203oool00`000000oooo0?ooo`0X0?ooo`03003o003o
ool0oooo0?l0ooooC@3oool001X0oooo00D0oono07@000000000001d0;ooo`0:0?ooo`050;md0000
0000000000000018W=l0203oool00`000000oooo0?ooo`0X0?ooo`03003o003oool0oooo0?l0oooo
C@3oool003H0oooo0`00000Y0?ooo`03003o003oool0oooo0?l0ooooC03oool003H0oooo00<00000
0?ooo`3oool0:@3oool00`00o`00oooo0?ooo`3o0?oood`0oooo000f0?ooo`030000003oool0oooo
02T0oooo00<00?l00?ooo`3oool0o`3ooom<0?ooo`00=P3oool00`000000oooo0?ooo`0Y0?ooo`03
003o003oool0oooo0?l0ooooC03oool003H0oooo00<000000?ooo`3oool0:P3oool00`00o`00oooo
0?ooo`3o0?oood/0oooo000f0?ooo`030000003oool0oooo02X0oooo00<00?l00?ooo`3oool0o`3o
oom;0?ooo`00=P3oool00`000000oooo0?ooo`3o0?ooogP0oooo000f0?ooo`<00000o`3ooomh0?oo
o`00=P3oool00`000000oooo0?ooo`3o0?ooogP0oooo000f0?ooo`030000003oool0oooo0?l0oooo
N03oool003H0oooo00<000000?ooo`3oool0o`3ooomh0?ooo`00=P3oool00`000000oooo0?ooo`3o
0?ooogP0oooo000f0?ooo`030000003oool0oooo02`0oooo00<00?l00?ooo`3oool0o`3ooom90?oo
o`00=P3oool00`000000oooo0?ooo`0/0?ooo`03003o003oool0oooo0?l0ooooB@3oool003H0oooo
0`00000/0?ooo`03003o003oool0oooo0?l0ooooB@3oool003H0oooo00<000000?ooo`3oool0;@3o
ool00`00o`00oooo0?ooo`3o0?ooodP0oooo000f0?ooo`030000003oool0oooo02d0oooo00<00?l0
0?ooo`3oool0o`3ooom80?ooo`00=P3oool00`000000oooo0?ooo`0]0?ooo`03003o003oool0oooo
0?l0ooooB03oool003H0oooo00<000000?ooo`3oool0;@3oool00`00o`00oooo0?ooo`170?ooo`80
o`00oP3oool003H0oooo00<000000?ooo`3oool0;P3oool00`00o`00oooo0?ooo`130?ooo`P0o`00
n`3oool003H0oooo0`00000^0?ooo`03003o003oool0oooo0480oooo00<0o`000?ooo`3oool01P3o
ool20?l00?T0oooo000f0?ooo`030000003oool0oooo0740oooo0P3o000;0?ooo`030?l0003oool0
oooo0?H0oooo000A0?ooo`050?oOW018000000000000M02oool00`3oool01@3o_g@0000007Boo`3o
ool0_g@000H0000000@0B9cO0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`080?ooo`030000
003oool0oooo06l0oooo0P3o000>0?ooo`030?l0003oool0oooo0?D0oooo000@0?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo0080oooo00<0gia800000000
00000`0000001@00B9`0gooo0?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool0203oool00`00
0000oooo0?ooo`1^0?ooo`030?l0003oool0oooo00l0oooo00<0o`000?ooo`3oool0403oool2003o
00<0oooo00<00?l00?ooo`3oool0g03oool00100oooo00<0_g@00018W03Oool00P3oool00`2oM000
07Bo0?ooo`070?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0103oool00`3ooml0
W4Qd0;ooo`080?ooo`030000003oool0oooo06h0oooo00<0o`000?ooo`3oool0403oool00`3o0000
oooo0?ooo`0=0?ooo`800?l01@3oool3003o0=`0oooo000?0?ooo`030?oo_`1d0000B9cO00<0oooo
00<0gia80000M02oool0203oool00`3OW4P0M;oo0?ooo`090?ooo`030;md001d_ol0oooo00L0oooo
00<000000?ooo`3oool0K@3oool00`3o0000oooo0?ooo`0B0?ooo`030?l0003oool0oooo00/0oooo
00<00?l00?ooo`3oool0203oool2003o0=X0oooo00030?ooo`030?oog`2LB000000000L0000000D0
B9cO0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00<0_gA809cOo`3oool0
1`3oool00`3ooml0W4P004RLg`080?ooo`030000003oool0oooo06d0oooo00<0o`000?ooo`3oool0
4`3oool00`3o0000oooo0?ooo`0G0?ooo`03003o003oool0oooo0=L0oooo000?0?ooo`030?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol02@3oool00`3ooml0W4Qd0;ooo`070?ooo`030=nLB000
0000M;oo00P0oooo1000000_0?ooo`03003o003oool0oooo03T0oooo00<0o`000?ooo`3oool05@3o
ool00`3o0000oooo0?ooo`0G0?ooo`03003o003oool0oooo0=H0oooo000?0?ooo`030?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3o_g@007Bo0?ooo`020?ooo`070=oOg`3oool0
oono07@000000000001d0;ooo`080?ooo`030000003oool0oooo0300oooo00<00?l00?ooo`3oool0
>03oool00`3o0000oooo0?ooo`0F0?ooo`030?l0003oool0oooo0?00oooo000?0?ooo`030?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3ookl0M01809cOo`020?ooo`060=nLB000
000000000000000004P0W=oo2@3oool00`000000oooo0?ooo`0a0?ooo`03003o003oool0oooo03L0
oooo00<0o`000?ooo`3oool05`3oool00`3o0000oooo0?ooo`3_0?ooo`003`3oool00`3ookl0M000
04RLg`030?ooo`030=nLB00007@0_ooo00X0oooo00<0oono07@00018W=l00P3oool0103ookl0M000
00000018W=l;0?ooo`030000003oool0oooo0340oooo00<00?l00?ooo`3oool0=P3oool00`3o0000
oooo0?ooo`0I0?ooo`050?l0003oool0oooo0?ooo`00o`00k03oool00100oooo00<0gia80018W03O
ool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P00018
W=l00`3oool00`2oM000M;oo0?ooo`0;0?ooo`030000003oool0oooo0340oooo00<00?l00?ooo`3o
ool0=@3oool00`3o0000oooo0?ooo`0K0?ooo`030?l0003oool00?l005l0oooo2@3o00250?ooo`00
403oool01`3ookl0M00007Boo`3oool0okmd0000M02oool01P3oool0203ooml0W4Q809cOo`3oool0
oono07@0000007@0_ooo0`3oool01P3ogi`0B000000000000000001809cOo`P0oooo00<000000?oo
o`3oool0<@3oool00`00o`00oooo0?ooo`0e0?ooo`030?l0003oool0oooo01/0oooo00<0o`00003o
003oool0G@3oool20?l000T0oooo00<0o`000?ooo`3oool0PP3oool00140oooo00D0oono07@00000
0000001d0;ooo`080?ooo`060?oog`2LB00000000000000007@0_ooo1@3oool01@2oM00000000000
00000000B9cO00P0oooo00<000000?ooo`3oool0<P3oool00`00o`00oooo0?ooo`0c0?ooo`030?l0
003oool0oooo01d0oooo00<00?l00?ooo`3oool0603oool2003o0440oooo00<0o`000?ooo`3oool0
2P3oool20?l00140oooo0P00o`020?ooo`D00?l0J03oool003H0oooo0`00000b0?ooo`03003o003o
ool0oooo0380oooo00<0o`000?ooo`3oool07@3oool00`00o`00o`000?ooo`0K0?ooo`03003o003o
ool0oooo03d0oooo00<0o`000?ooo`3oool03@3oool00`3o0000oooo0?ooo`0<0?ooo`800?l02@3o
ool00`00o`00oooo0?ooo`1U0?ooo`00=P3oool00`000000oooo0?ooo`1W0?ooo`030?l0003oool0
oooo01d0oooo00<00?l00?ooo`3o00006`3oool00`00o`00oooo0?ooo`0l0?ooo`030?l0003oool0
oooo00l0oooo0P3o000:0?ooo`800?l0303oool2003o06D0oooo000f0?ooo`030000003oool0oooo
06H0oooo00<0o`000?ooo`3oool07@3oool01000o`00oooo0?ooo`3o000L0?ooo`03003o003oool0
oooo03T0oooo0P3o000D0?ooo`030?l0003oool0oooo00H0oooo00<00?l00?ooo`3oool0FP3oool6
0?l001<0oooo000f0?ooo`030000003oool0oooo06H0oooo00<0o`000?ooo`3oool08@3oool00`3o
0000oooo0?ooo`0J0?ooo`03003o003oool0oooo03L0oooo00<0o`000?ooo`3oool05@3oool00`3o
0000oooo0?ooo`1P0?ooo`<0o`001@3oool30?l00100oooo000f0?ooo`030000003oool0oooo06D0
oooo00<0o`000?ooo`3oool08P3oool00`3o0000oooo0?ooo`0J0?ooo`03003o003oool0oooo03H0
oooo00<0o`000?ooo`3oool05`3oool00`3o0000oooo0?ooo`1M0?ooo`80o`002`3oool00`3o0000
oooo0?ooo`0=0?ooo`00=P3oool00`000000oooo0?ooo`1D0?ooo`030?oog`2LB000000000800000
00<0M;oo0?ooo`3oool02@3oool00`3o0000oooo0?ooo`0S0?ooo`030?l0003oool0oooo01X0oooo
00D00?l00?ooo`3oool0oooo0;md0003000000030018W03Oool0oooo0080oooo00D0omnL04P00000
0000001d0;ooo`0U0?ooo`030?l0003oool0oooo01T0oooo00<0o`000?ooo`3oool02@3oool02@2o
M000000000000000000004RL0=ooo`3oool0oooO09a800030000000507Boo`3oool0oooo0?ooo`00
o`00@03oool20?l000L0oooo00<0_g@0000000000000100000030?l000060?ooo`3ogi`0B0000000
000007@0_ooo1P3oool003H0oooo0`00001D0?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@007Bo
o`3oool0203oool00`3o0000oooo0?ooo`0T0?ooo`030?l0003oool0oooo01l0oooo00<0omnL04P0
001d_ol0103oool01`3ookl0M00007Boo`3oool0okmd0000M02oool08`3oool00`3o0000oooo0?oo
o`0J0?ooo`030?l0003oool0oooo00X0oooo00<0omnL04P0001d_ol00`3oool00`3ooml0W4Qd0;oo
o`020?ooo`030;md001d_ol0oooo0080oooo00<00?l00?ooo`3oool0?03oool00`3o0000oooo0?oo
o`080?ooo`030=nLB0000000000000<0000000X004RL0=ooo`3oool0oono07@0001d_ol0oooo0?no
M00007@0_ooo1@3oool003H0oooo00<000000?ooo`3oool0=@3oool00`00o`00oooo0?ooo`0Q0?oo
o`030?oog`2LB7@0_ooo00L0oooo00<0o`000?ooo`3oool09P3oool00`3o0000oooo0?ooo`0N0?oo
o`030?oOW0180000M;oo00@0oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`0Q0?oo
o`030?l0003oool0oooo01/0oooo0P00o`0;0?ooo`030?oOW0180000M;oo00P0oooo00<0oooO09a8
M02oool00`3oool00`00o`00oooo0?ooo`0j0?ooo`030?l0003oool0oooo00X0oooo00<0gia804RL
g`3oool00P3oool01P3OW7@0W=oo0?ooo`2oM00004RL0=ooo`80oooo00<0_g@0001d_`3oool0103o
ool003H0oooo00<000000?ooo`3oool0=@3oool00`00o`00oooo0?ooo`0R0?ooo`030;md001d_ol0
oooo00H0oooo00<0o`000?ooo`3oool09P3oool00`3o0000oooo0?ooo`0N0?ooo`030?oOW0180000
M;oo00<0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0P0?ooo`030?l0003oool0
oooo01/0oooo00@00?l00?ooo`3oool0o`002P3oool00`3ogi`0B00007Boo`090?ooo`030;md001d
_ol0oooo0080oooo00<00?l00?ooo`3oool0>03oool20?l000h0oooo00<0gia807Boo`3oool00`3o
ool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00@0oooo000f0?ooo`030000003oool0
oooo03D0oooo00<00?l00?ooo`3oool08@3oool00`3ooml0W4P004RLg`060?ooo`030?l0003oool0
oooo01`0oooo0P00o`0:0?ooo`030?l0003oool0oooo01d0oooo00<0omnL04P0001d_ol00`3oool0
0`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo0200oooo00<0o`000?ooo`3oool06`3oool0
1@00o`00oooo0?ooo`3oool0o`0000T0oooo00<0omnL04P0001d_ol0203oool00`3ooml0W4P004RL
g`040?ooo`03003o003oool0oooo03H0oooo00<0o`000?ooo`3oool03`3oool00`2oM4P0W=oo0?oo
o`020?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0103oool003H0oooo00<00000
0?ooo`3oool0=P3oool00`00o`00oooo0?ooo`0P0?ooo`030=nLB0000000M;oo00H0oooo00<0o`00
0?ooo`3oool0703oool00`00o`00oooo0?ooo`090?ooo`030?l0003oool0oooo01d0oooo00<0omnL
04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo01l0oooo00<0o`00
0?ooo`3oool06`3oool00`00o`00oooo0?ooo`030?ooo`030?l0003oool0oooo00H0oooo00<0omnL
04P0001d_ol0203oool00`3OW4P0000007Boo`050?ooo`03003o003oool0oooo03@0oooo00<0o`00
0?ooo`3oool0403oool00`3ooml0W4Qd0;ooo`020?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol0103oool003H0oooo00<000000?ooo`3oool0=P3oool00`00o`00oooo0?ooo`0L0?oo
o`070=oOg`3oool0oono07@000000000001d0;ooo`050?ooo`030?l0003oool0oooo01`0oooo00<0
0?l00?ooo`3oool02`3oool00`3o0000oooo0?ooo`0L0?ooo`030?oOW0180000M;oo00<0oooo00<0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0O0?ooo`030?l0003oool0oooo01/0oooo00<0
0?l00?ooo`3oool0103oool00`3o0000oooo0?ooo`050?ooo`030?oOW0180000M;oo00@0oooo00L0
gmoO0?ooo`3ookl0M0000000000007@0_ooo00H0oooo00<00?l00?ooo`3oool0<P3oool00`3o0000
oooo0?ooo`0B0?ooo`070?noM000M;l0oooo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo00@0oooo000f0?ooo`030000003oool0oooo03H0oooo00<00?l00?ooo`3oool0703oool01P3O
W4P00000000000000000001809cOo`H0oooo00<0o`000?ooo`3oool0703oool00`00o`00oooo0?oo
o`0<0?ooo`030?l0003oool0oooo01X0oooo00@00?l00?oOW0180000M;oo0`3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo01h0oooo00<0o`000?ooo`3oool06`3oool00`00o`00oooo
0?ooo`050?ooo`030?l0003oool0oooo00D0oooo00<0omnL04P0001d_ol0103oool01P3OW4P00000
000000000000001809cOoc/0oooo00<0o`000?ooo`3oool04`3oool01`3ookl0M01809cOo`3oool0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`040?ooo`00=P3oool3000003H0oooo00<00?l0
0?ooo`3oool0703oool0103ookl0M00000000018W=l70?ooo`030?l0003oool0oooo01`0oooo00<0
0?l00?ooo`3oool03@3oool00`3o0000oooo0?ooo`0J0?ooo`04003o003ogi`0B00007Boo`<0oooo
00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0N0?ooo`030?l0003oool0oooo01X0oooo
00<00?l00?ooo`3oool01`3oool00`3o0000oooo0?ooo`040?ooo`030?oOW0180000M;oo00@0oooo
00@0oono07@000000000B9cO?03oool00`3o0000oooo0?ooo`0D0?ooo`070?oo_`1d0000B9cO0?oo
o`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00@0oooo000f0?ooo`030000003oool0oooo
03L0oooo00<00?l00?ooo`3oool0703oool00`2oM000M;oo0?ooo`070?ooo`030?l0003oool0oooo
01`0oooo00<00?l00?ooo`3oool03P3oool00`3o0000oooo0?ooo`0J0?ooo`03003o00180000M;oo
00@0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`0M0?ooo`030?l0003oool0oooo
02H0oooo00<0o`000?ooo`3oool00`3oool00`3ogi`0B00007Boo`050?ooo`030;md001d_ol0oooo
03/0oooo00<0o`000?ooo`3oool0403oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO
0080oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`040?ooo`00=P3oool00`000000
oooo0?ooo`0g0?ooo`03003o003oool0oooo01`0oooo00H0omnL04P00000000000000000B02Lgol4
0?ooo`030?l0003oool0oooo01/0oooo00<00?l00?ooo`3oool03`3oool00`3o0000oooo0?ooo`0H
0?ooo`050?oog`2LB0000?l00000001d_ol0103oool01`3ookl0M00007Boo`3oool0okmd0000M02o
ool07P3oool00`3o0000oooo0?ooo`0V0?ooo`060?l0003oool0oooo0?ooo`3ooml0W4P00P000000
0`1d_ol0oooo0?ooo`030?ooo`060?oOW018000000000000000004P0W=oo=`3oool00`3o0000oooo
0?ooo`0A0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool20?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`050?ooo`00=P3oool00`000000oooo0?ooo`1G0?ooo`050;md0000
0000000000000018W=l00`3oool00`3o0000oooo0?ooo`0L0?ooo`03003o003oool0oooo0100oooo
00<0o`000?ooo`3oool06@3oool00`3o_g@00?l007Boo`050?ooo`050?oo_`1d000000000000M02o
ool07P3oool00`3o0000oooo0?ooo`0X0?ooo`030?l0003oool0oooo0080oooo00<0okmd0000001d
_ol01P3oool01@2oM0000000000000000000B9cO03H0oooo00<0o`000?ooo`3oool04`3oool01P3o
oml0W4P0000000000000001d0;ooo`@0oooo00D0oono07@000000000001d0;ooo`060?ooo`00=P3o
ool00`000000oooo0?ooo`1O0?ooo`030?l0003oool0oooo02l0oooo00<0o`000?ooo`3oool06P3o
ool00`00o`00oooo0?ooo`0V0?ooo`030?l0003oool0oooo02X0oooo00<0o`000?ooo`3oool06P3o
ool00`00o`00oooo0?ooo`0W0?ooo`030?l0003oool0oooo02T0oooo000f0?ooo`030000003oool0
oooo05h0oooo00<0o`000?ooo`3oool0<@3oool00`3o0000oooo0?ooo`0J0?ooo`03003o003oool0
oooo02D0oooo00<0o`000?ooo`3oool0:P3oool00`3o0000oooo0?ooo`0K0?ooo`03003o003oool0
oooo02D0oooo00<0o`000?ooo`3oool0:P3oool003H0oooo00<000000?ooo`3oool0GP3oool00`3o
0000oooo0?ooo`0a0?ooo`030?l0003oool0oooo01X0oooo00<00?l00?ooo`3oool0903oool00`3o
0000oooo0?ooo`0J0?ooo`03003o003oool0oooo00l0oooo00<0o`000?ooo`3oool06P3oool00`00
o`00oooo0?ooo`0U0?ooo`030?l0003oool0oooo02X0oooo000f0?ooo`<00000GP3oool00`3o0000
oooo0?ooo`0b0?ooo`030?l0003oool0oooo03l0oooo00<0o`000?ooo`3oool06P3oool00`00o`00
oooo0?ooo`0A0?ooo`030?l0003oool0oooo01X0oooo00<00?l00?ooo`3oool08`3oool00`3o0000
oooo0?ooo`0J0?ooo`03003o003oool0oooo00h0oooo000f0?ooo`030000003oool0oooo05d0oooo
00<0o`000?ooo`3oool0<`3oool00`3o0000oooo0?ooo`0n0?ooo`030?l0003oool0oooo01X0oooo
00<00?l00?ooo`3oool04P3oool00`3o0000oooo0?ooo`0K0?ooo`03003o003oool0oooo0240oooo
00<0o`000?ooo`3oool06P3oool00`00o`00oooo0?ooo`0?0?ooo`00=P3oool00`000000oooo0?oo
o`0i0?ooo`03003o003oool0oooo0240oooo00<0o`000?ooo`3oool06`3oool00`00o`00oooo0?oo
o`0F0?ooo`030?l0003oool0oooo03d0oooo00<0o`000?ooo`3oool06P3oool00`00o`00oooo0?oo
o`0C0?ooo`030?l0003oool0oooo01X0oooo00<00?l00?ooo`3oool0803oool00`3o0000oooo0?oo
o`0K0?ooo`03003o003oool0oooo00l0oooo000f0?ooo`030000003oool0oooo03X0oooo00<00?l0
0?ooo`3oool07`3oool00`3o0000oooo0?ooo`0L0?ooo`03003o003oool0oooo01H0oooo00<0o`00
0?ooo`3oool0?03oool00`3o0000oooo0?ooo`0J0?ooo`03003o003oool0oooo01D0oooo00<0o`00
0?ooo`3oool06P3oool00`00o`00oooo0?ooo`0O0?ooo`030?l0003oool0oooo01X0oooo00<00?l0
0?ooo`3oool0403oool003H0oooo00<000000?ooo`3oool0>P3oool00`00o`00oooo0?ooo`0O0?oo
o`030?l0003oool0oooo01/0oooo00<00?l00?ooo`3oool0603oool00`3o0000oooo0?ooo`0j0?oo
o`030?l0003oool0oooo01X0oooo00<00?l00?ooo`3oool05`3oool00`3o0000oooo0?ooo`0j0?oo
o`030?l0003oool0oooo01X0oooo00<00?l00?ooo`3oool04@3oool003H0oooo00<000000?ooo`3o
ool0>`3oool00`00o`00oooo0?ooo`0N0?ooo`030?l0003oool0oooo01/0oooo00<00?l00?ooo`3o
ool0603oool00`3o0000oooo0?ooo`0K0?ooo`03003o003oool0oooo01`0oooo00<0o`000?ooo`3o
ool06@3oool00`00o`00oooo0?ooo`0H0?ooo`030?l0003oool0oooo03T0oooo00<0o`000?ooo`3o
ool06`3oool00`00o`00oooo0?ooo`0A0?ooo`00=P3oool00`000000oooo0?ooo`0k0?ooo`03003o
003oool0oooo01d0oooo00<0o`000?ooo`3oool06`3oool00`00o`00oooo0?ooo`0J0?ooo`030?l0
003oool0oooo01/0oooo00<00?l00?ooo`3oool06P3oool00`3o0000oooo0?ooo`0f0?ooo`030?l0
003oool0oooo03P0oooo00<0o`000?ooo`3oool06P3oool00`00o`00oooo0?ooo`0B0?ooo`00;@3o
ool9000000D0o`00>@0000000`00o`0000000000000M000000030?l000000000000001/0000000<0
0?l00000000000006`0000000`3o000000000000000J00000003003o00000000000001T0000000<0
o`00000000000000>00000000`3o000000000000000f000000030?l000000000000002d000000`3o
ool003H0oooo00<000000?ooo`3oool00P3oool20?l000/0oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00D0oooo00<00?l00?ooo`3oool01`3oool0
0`000000oooo0?ooo`0?0?ooo`040000003oool0oooo0?l000l0oooo00<000000?ooo`3oool02P3o
ool00`00o`00oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool01@3o
ool00`3o0000oooo0?ooo`070?ooo`030000003oool0oooo00l0oooo00<000000?ooo`00o`00403o
ool00`000000oooo0?ooo`070?ooo`030?l0003oool0oooo00D0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00`0oooo00@0o`000?ooo`3oool00000
4P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00d0oooo00<0o`000?ooo`000000
4@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00X0oooo000f0?ooo`030000003o
ool0oooo00@0oooo00<0o`000?ooo`3oool0=@3oool00`00o`00oooo0?ooo`0I0?ooo`030000003o
ool0o`0001d0oooo00<00?l00?ooo`3oool07@3oool00`3o0000oooo0?ooo`0H0?ooo`030000003o
ool00?l001T0oooo00<0o`000?ooo`3oool0>P3oool00`3o0000oooo0000000J0?ooo`03003o003o
ool0oooo01L0oooo00<0o`000?ooo`3oool0903oool00`000000oooo0?ooo`0:0?ooo`00=P3oool0
0`000000oooo0?ooo`050?ooo`030?l0003oool0oooo0580oooo00<0o`000?ooo`3oool0>`3oool0
0`3o0000oooo0?ooo`0K0?ooo`03003o003oool0oooo01H0oooo00<0o`000?ooo`3oool0>`3oool0
0`3o0000oooo0?ooo`0J0?ooo`03003o003oool0oooo01D0oooo00<0o`000?ooo`3oool0<P3oool0
03H0oooo00<000000?ooo`3oool01P3oool20?l00540oooo00<0o`000?ooo`3oool0?@3oool00`3o
0000oooo0?ooo`0J0?ooo`03003o003oool0oooo01D0oooo00<0o`000?ooo`3oool0?03oool00`3o
0000oooo0?ooo`0K0?ooo`03003o003oool0oooo01<0oooo00<0o`000?ooo`3oool0<`3oool003H0
oooo00<000000?ooo`3oool0203oool00`3o0000oooo0?ooo`1>0?ooo`030?l0003oool0oooo03d0
oooo00<0o`000?ooo`3oool06`3oool00`00o`00oooo0?ooo`0C0?ooo`030?l0003oool0oooo01X0
oooo00<00?l00?ooo`3oool08@3oool00`3o0000oooo0?ooo`0K0?ooo`03003o003oool0oooo0140
oooo00<0o`000?ooo`3oool06P3oool00`00o`00oooo0?ooo`0G0?ooo`00=P3oool3000000T0oooo
00<0o`000?ooo`3oool0C@3oool00`3o0000oooo0?ooo`0n0?ooo`030?l0003oool0oooo0300oooo
00<0o`000?ooo`3oool06@3oool00`00o`00oooo0?ooo`0S0?ooo`030?l0003oool0oooo01X0oooo
00<00?l00?ooo`3oool04@3oool00`3o0000oooo0?ooo`0I0?ooo`03003o003oool0oooo01P0oooo
000f0?ooo`030000003oool0oooo00X0oooo00<0o`000?ooo`3oool0B`3oool00`3o0000oooo0?oo
o`0o0?ooo`030?l0003oool0oooo02l0oooo00<0o`000?ooo`3oool06P3oool00`00o`00oooo0?oo
o`0T0?ooo`030?l0003oool0oooo01X0oooo00<00?l00?ooo`3oool03`3oool00`3o0000oooo0?oo
o`0I0?ooo`03003o003oool0oooo01T0oooo000f0?ooo`030000003oool0oooo00X0oooo00<0o`00
0?ooo`3oool0B`3oool00`3o0000oooo0?ooo`100?ooo`030?l0003oool0oooo02h0oooo00<0o`00
0?ooo`3oool06@3oool00`00o`00oooo0?ooo`0U0?ooo`030?l0003oool0oooo01X0oooo00<00?l0
0?ooo`3oool03P3oool00`3o0000oooo0?ooo`0J0?ooo`03003o003oool0oooo01T0oooo000f0?oo
o`030000003oool0oooo00/0oooo00<0o`000?ooo`3oool0<03oool00`00o`00oooo0?ooo`0F0?oo
o`030?l0003oool0oooo01/0oooo00<00?l00?ooo`3oool08`3oool00`3o0000oooo0?ooo`0]0?oo
o`030?l0003oool0oooo01T0oooo00<00?l00?ooo`3oool09`3oool00`3o0000oooo0?ooo`0J0?oo
o`03003o003oool0oooo00`0oooo00<0o`000?ooo`3oool06P3oool00`00o`00oooo0?ooo`0J0?oo
o`00=P3oool00`000000oooo0?ooo`0<0?ooo`030?l0003oool0oooo02l0oooo00<00?l00?ooo`3o
ool05P3oool00`3o0000oooo0?ooo`0J0?ooo`03003o003oool0oooo02D0oooo00<0o`000?ooo`3o
ool0:`3oool00`3o0000oooo0?ooo`0J0?ooo`03003o003oool0oooo02P0oooo00<0o`000?ooo`3o
ool0:03oool00`3o0000oooo0?ooo`0I0?ooo`03003o003oool0oooo01/0oooo000f0?ooo`030000
003oool0oooo00d0oooo00<0o`000?ooo`3oool0;`3oool00`00o`00oooo0?ooo`0E0?ooo`030?l0
003oool0oooo01X0oooo00<00?l00?ooo`3oool09@3oool00`3o0000oooo0?ooo`0[0?ooo`030?l0
003oool0oooo01T0oooo00<00?l00?ooo`3oool0:P3oool00`3o0000oooo0?ooo`0V0?ooo`030?l0
003oool0oooo01T0oooo00<00?l00?ooo`3oool0703oool003H0oooo00<000000?ooo`3oool03@3o
ool00`3o0000oooo0?ooo`0_0?ooo`03003o003oool0oooo01@0oooo00<0o`000?ooo`3oool06P3o
ool00`00o`00oooo0?ooo`0W0?ooo`030?l0003oool0oooo01/0oooo00<00?l00?ooo`3oool02`3o
ool00`3o0000oooo0?ooo`180?ooo`030?l0003oool0oooo02@0oooo00<0o`000?ooo`3oool0>@3o
ool003H0oooo0`00000>0?ooo`030?l0003oool0oooo02h0oooo00<00?l00?ooo`3oool0503oool0
0`3o0000oooo0?ooo`0J0?ooo`03003o003oool0oooo02L0oooo00<0o`000?ooo`3oool0703oool0
0`00o`00oooo0?ooo`090?ooo`030?l0003oool0oooo04T0oooo00<0o`000?ooo`3oool08`3oool0
0`3o0000oooo0?ooo`0j0?ooo`00=P3oool00`000000oooo0?ooo`0>0?ooo`030?l0003oool0oooo
02l0oooo00<00?l00?ooo`3oool04P3oool00`3o0000oooo0?ooo`0J0?ooo`03003o003oool0oooo
02T0oooo00<0o`000?ooo`3oool06`3oool00`00o`00oooo0?ooo`090?ooo`030?l0003oool0oooo
04X0oooo00<0o`000?ooo`3oool08@3oool00`3o0000oooo0?ooo`0k0?ooo`00=P3oool00`000000
oooo0?ooo`0?0?ooo`030?l0003oool0oooo02h0oooo00<00?l00?ooo`3oool04P3oool00`3o0000
oooo0?ooo`0J0?ooo`03003o003oool0oooo02X0oooo00<0o`000?ooo`3oool06`3oool00`00o`00
oooo0?ooo`070?ooo`030?l0003oool0oooo04`0oooo00<0o`000?ooo`3oool06P3oool00`00o`00
oooo0?ooo`020?ooo`030?l0003oool0oooo03`0oooo000f0?ooo`030000003oool0oooo0100oooo
00<0o`000?ooo`3oool0;P3oool00`00o`00oooo0?ooo`0A0?ooo`030?l0003oool0oooo01T0oooo
00<00?l00?ooo`3oool0:`3oool00`3o0000oooo0?ooo`0L0?ooo`03003o003oool0oooo00H0oooo
00<0o`000?ooo`3oool0C@3oool00`3o0000oooo0?ooo`0J0?ooo`05003o003oool0oooo0?ooo`3o
0000?P3oool003H0oooo00<000000?ooo`3oool0403oool00`3o0000oooo0?ooo`0^0?ooo`03003o
003oool0oooo0100oooo00<0o`000?ooo`3oool0B@3oool00`3o0000oooo0?ooo`0K0?ooo`03003o
003oool0oooo00D0oooo00<0o`000?ooo`3oool0C`3oool00`3o0000oooo0?ooo`0J0?ooo`03003o
003oool0o`0001/0oooo00<00?l00?ooo`3oool08@3oool003H0oooo00<000000?ooo`3oool04@3o
ool00`3o0000oooo0?ooo`100?ooo`030?l0003oool0oooo04X0oooo00<0o`000?ooo`3oool06`3o
ool00`00o`00oooo0?ooo`030?ooo`030?l0003oool0oooo01X0oooo00<00?l00?ooo`3oool0=03o
ool00`3o0000oooo0?ooo`0J0?ooo`03003o003oool0oooo01P0oooo0P00o`0T0?ooo`00=P3oool0
0`000000oooo0?ooo`0A0?ooo`030?l0003oool0oooo03l0oooo00<0o`000?ooo`3oool0B`3oool0
0`3o0000oooo0?ooo`0Q0?ooo`030?l0003oool0oooo01T0oooo00<00?l00?ooo`3oool0=P3oool0
0`3o0000oooo0?ooo`0H0?ooo`030?l0003oool00?l001P0oooo00<00?l00?ooo`3oool0903oool0
03H0oooo0`00000B0?ooo`030?l0003oool0oooo03h0oooo00<0o`000?ooo`3oool0C03oool00`3o
0000oooo0?ooo`0O0?ooo`030?l0003oool0oooo01T0oooo00<00?l00?ooo`3oool0>03oool00`3o
0000oooo0?ooo`0F0?ooo`050?l0003oool0oooo0?ooo`00o`005P3oool00`00o`00oooo0?ooo`0U
0?ooo`00=P3oool00`000000oooo0?ooo`0C0?ooo`030?l0003oool0oooo03d0oooo00<0o`000?oo
o`3oool0C@3oool00`3o0000oooo0?ooo`0N0?ooo`030?l0003oool0oooo01P0oooo00<00?l00?oo
o`3oool0>P3oool00`3o0000oooo0?ooo`0C0?ooo`80o`006P3oool00`00o`00oooo0?ooo`0V0?oo
o`00=P3oool00`000000oooo0?ooo`0C0?ooo`030?l0003oool0oooo03`0oooo00<0o`000?ooo`3o
ool0CP3oool00`3o0000oooo0?ooo`0M0?ooo`030?l0003oool0oooo01P0oooo00<00?l00?ooo`3o
ool0?03oool00`3o0000oooo0?ooo`0A0?ooo`030?l0003oool0oooo01T0oooo00<00?l00?ooo`3o
ool09`3oool003H0oooo00<000000?ooo`3oool0503oool00`3o0000oooo0?ooo`0k0?ooo`030?l0
003oool0oooo01T0oooo00<00?l00?ooo`3oool0<`3oool00`3o0000oooo0?ooo`0K0?ooo`030?l0
003oool0oooo01T0oooo00<00?l00?ooo`3oool0?@3oool00`3o0000oooo0?ooo`0?0?ooo`030?l0
003oool0oooo04@0oooo000f0?ooo`030000003oool0oooo01@0oooo00<0o`000?ooo`3oool0;@3o
ool00`00o`00oooo0?ooo`0:0?ooo`030?l0003oool0oooo01T0oooo00<00?l00?ooo`3oool0=03o
ool00`3o0000oooo0?ooo`0J0?ooo`030?l0003oool00?l001T0oooo00<00?l00?ooo`3oool0?`3o
ool20?l000h0oooo00<0o`000?ooo`3oool03@3oool2003o03H0oooo000f0?ooo`030000003oool0
oooo01D0oooo00<0o`000?ooo`3oool0;03oool00`00o`00oooo0?ooo`0:0?ooo`030?l0003oool0
oooo01T0oooo00<00?l00?ooo`3oool0=@3oool00`3o0000oooo0?ooo`0I0?ooo`040?l0003oool0
oooo003o05`0oooo0`3o00070?ooo`@0o`004P3oool6003o0300oooo000f0?ooo`<000005@3oool0
0`3o0000oooo0?ooo`0]0?ooo`03003o003oool0oooo00P0oooo00<0o`000?ooo`3oool06@3oool0
0`00o`00oooo0?ooo`0g0?ooo`030?l0003oool0oooo01L0oooo00<0o`000?ooo`3oool00P3oool0
0`00o`00oooo0?ooo`1L0?ooo`L0o`00C03oool003H0oooo00<000000?ooo`3oool05P3oool00`3o
0000oooo0?ooo`0/0?ooo`03003o003oool0oooo00P0oooo00<0o`000?ooo`3oool06@3oool00`00
o`00oooo0?ooo`0h0?ooo`030?l0003oool0oooo01D0oooo00<0o`000?ooo`3oool0103oool00`00
o`00oooo0?ooo`2^0?ooo`004@3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00D0okmd0000
001d_ol0oooo0;md00060000000404RLg`3oool0oooO09a800<0000000<0M;oo0?ooo`3oool0203o
ool00`000000oooo0?ooo`0G0?ooo`030?l0003oool0oooo02`0oooo00<00?l00?ooo`3oool01`3o
ool00`3o0000oooo0?ooo`0H0?ooo`03003o003oool0oooo03X0oooo00<0o`000?ooo`3oool04`3o
ool00`3o0000oooo0?ooo`060?ooo`800?l0[P3oool00100oooo00L0oono07@0001d_ol0oooo0?no
M00007@0_ooo0080oooo00<0okmd0000001d_ol00P3oool00`3OW4P0000000000003000000050018
W03Oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`080?ooo`030000003oool0oooo01L0
oooo00<0o`000?ooo`3oool0;03oool00`00o`00oooo0?ooo`060?ooo`030?l0003oool0oooo01T0
oooo00<00?l00?ooo`3oool0>`3oool00`3o0000oooo0?ooo`0A0?ooo`030?l0003oool0oooo00T0
oooo00<00?l00?ooo`3oool02`3oool00`00o`00oooo0?ooo`2M0?ooo`00403oool00`2oM00004RL
0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo
0?ooo`040?ooo`030?oog`2LB7@0_ooo00P0oooo00<000000?ooo`3oool0603oool00`3o0000oooo
0?ooo`0/0?ooo`03003o003oool0oooo00D0oooo00<0o`000?ooo`3oool0603oool00`00o`00oooo
0?ooo`0m0?ooo`80o`00403oool00`3o0000oooo0?ooo`0E0?ooo`<00?l0X03oool000l0oooo00<0
oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`080?ooo`030=nLB01d_ol0oooo00T0oooo00<0
_g@007Boo`3oool01`3oool00`000000oooo0?ooo`0H0?ooo`030?l0003oool0oooo02`0oooo00<0
0?l00?ooo`3oool0103oool00`3o0000oooo0?ooo`1J0?ooo`030?l0003oool0oooo00d0oooo00<0
o`000?ooo`3oool04P3oool4003o0:<0oooo000?0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol02@3oool00`2oM4P0W=oo0?ooo`070?ooo`030?oog`2LB000B9cO00P0oooo00<00000
0?ooo`3oool06@3oool00`3o0000oooo0?ooo`0b0?ooo`030?l0003oool0oooo05/0oooo0P3o000;
0?ooo`80o`00_03oool000l0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`090?oo
o`030?oog`2LB7@0_ooo00L0oooo00<0gia80000001d_ol0203oool4000001T0oooo00<0o`000?oo
o`3oool0<@3oool00`3o0000oooo0?ooo`1M0?ooo`<0o`001@3oool30?l00;h0oooo000?0?ooo`03
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3o_g@007Bo0?ooo`020?ooo`07
0=oOg`3oool0oono07@000000000001d0;ooo`080?ooo`030000003oool0oooo01X0oooo00<0o`00
0?ooo`3oool0<03oool00`3o0000oooo0?ooo`1Q0?ooo`D0o`00`@3oool000l0oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030?oo_`1d04P0W=oo0080oooo00H0gia80000
0000000000000000B02Lgol90?ooo`030000003oool0oooo01/0oooo00<0o`000?ooo`3oool0;`3o
ool00`3o0000oooo0?ooo`3o0?ooobP0oooo000?0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia8
0000M02oool02P3oool00`3ookl0M00004RLg`020?ooo`040?oo_`1d0000000004RLg`/0oooo00<0
00000?ooo`3oool06`3oool00`3o0000oooo0?ooo`0^0?ooo`030?l0003oool0oooo0?l0oooo:@3o
ool00100oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?noW02Lgol0
oooo0080oooo00<0omnL04P00018W=l00`3oool00`2oM000M;oo0?ooo`0;0?ooo`030000003oool0
oooo01`0oooo00<0o`000?ooo`3oool0;03oool00`00o`00o`000?ooo`0I0?ooo`03003o003oool0
oooo0?l0oooo3P3oool00100oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00H0oooo00P0
oooO09a8B02Lgol0oooo0?oo_`1d0000001d0;ooo`<0oooo00H0omnL04P00000000000000000B02L
gol80?ooo`030000003oool0oooo01d0oooo00<0o`000?ooo`3oool0:`3oool00`00o`00o`000?oo
o`0H0?ooo`03003o003oool0oooo0?l0oooo3`3oool00140oooo00D0oono07@000000000001d0;oo
o`080?ooo`060?oog`2LB00000000000000007@0_ooo1@3oool01@2oM0000000000000000000B9cO
00P0oooo00<000000?ooo`3oool07@3oool00`3o0000oooo0?ooo`0[0?ooo`030?l00000o`00oooo
01L0oooo00<00?l00?ooo`3oool0o`3oool@0?ooo`00=P3oool3000001h0oooo00<0o`000?ooo`3o
ool0:P3oool00`3o00000?l00?ooo`0G0?ooo`03003o003oool0oooo0?l0oooo403oool003H0oooo
00<000000?ooo`3oool07P3oool00`3o0000oooo0?ooo`0Y0?ooo`040?l0003oool0oooo003o01H0
oooo00<00?l00?ooo`3oool0o`3ooolA0?ooo`00=P3oool00`000000oooo0?ooo`0O0?ooo`030?l0
003oool0oooo02P0oooo00@0o`000?ooo`3oool00?l05@3oool00`00o`00oooo0?ooo`3o0?oooa80
oooo000f0?ooo`030000003oool0oooo01l0oooo00<0o`000?ooo`3oool09`3oool00`3o0000oooo
0?ooo`020?ooo`03003o003oool0oooo0140oooo00<00?l00?ooo`3oool0o`3ooolC0?ooo`00=P3o
ool00`000000oooo0?ooo`0P0?ooo`030?l0003oool0oooo02H0oooo00<0o`000?ooo`3oool00`3o
ool00`00o`00oooo0?ooo`3o0?ooobH0oooo000f0?ooo`030000003oool0oooo0240oooo00<0o`00
0?ooo`3oool0903oool00`3o0000oooo0?ooo`3o0?ooobd0oooo000f0?ooo`030000003oool0oooo
0240oooo00<0o`000?ooo`3oool0903oool00`3o0000oooo0?ooo`3o0?ooobd0oooo000f0?ooo`<0
00008P3oool00`3o0000oooo0?ooo`0R0?ooo`030?l0003oool0oooo0?l0oooo;P3oool003H0oooo
00<000000?ooo`3oool08P3oool00`3o0000oooo0?ooo`0R0?ooo`030?l0003oool0oooo00/0oooo
00<00?l00?ooo`3oool0103oool2003o0?l0oooo6P3oool003H0oooo00<000000?ooo`3oool08`3o
ool00`3o0000oooo0?ooo`0P0?ooo`030?l0003oool0oooo00d0oooo1P00o`3o0?oooa`0oooo000f
0?ooo`030000003oool0oooo02<0oooo00<0o`000?ooo`3oool07`3oool00`3o0000oooo0?ooo`3o
0?oooc00oooo000f0?ooo`030000003oool0oooo02@0oooo00<0o`000?ooo`3oool07P3oool00`3o
0000oooo0?ooo`3o0?oooc00oooo000f0?ooo`030000003oool0oooo02D0oooo00<0o`000?ooo`3o
ool0703oool00`3o0000oooo0?ooo`3o0?oooc40oooo000f0?ooo`<000009P3oool00`3o0000oooo
0?ooo`0K0?ooo`030?l0003oool0oooo0?l0oooo<@3oool003H0oooo00<000000?ooo`3oool09P3o
ool00`3o0000oooo0?ooo`0J0?ooo`030?l0003oool0oooo0?l0oooo<P3oool003H0oooo00<00000
0?ooo`3oool09`3oool00`3o0000oooo0?ooo`0I0?ooo`030?l0003oool0oooo0?l0oooo<P3oool0
03H0oooo00<000000?ooo`3oool0:03oool00`3o0000oooo0?ooo`0G0?ooo`030?l0003oool0oooo
0?l0oooo<`3oool003H0oooo00<000000?ooo`3oool0:@3oool00`3o0000oooo0?ooo`0E0?ooo`03
0?l0003oool0oooo0?l0oooo=03oool003H0oooo00<000000?ooo`3oool0:P3oool00`3o0000oooo
0?ooo`0C0?ooo`030?l0003oool0oooo0?l0oooo=@3oool003H0oooo00<000000?ooo`3oool0:`3o
ool00`3o0000oooo0?ooo`0A0?ooo`030?l0003oool0oooo0?l0oooo=P3oool003H0oooo0`00000/
0?ooo`030?l0003oool0oooo00l0oooo00<0o`000?ooo`3oool0o`3ooolg0?ooo`00=P3oool00`00
0000oooo0?ooo`0/0?ooo`030?l0003oool0oooo00h0oooo00<0o`000?ooo`3oool0o`3ooolh0?oo
o`006P3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00H0okmd0000001d_ol0oooo0?oog`2L
B0030000000307Boo`3oool0oooo00P0oooo00<000000?ooo`3oool0;@3oool00`3o0000oooo0?oo
o`0<0?ooo`030?l0003oool0oooo0?l0oooo>@3oool001T0oooo00L0oono07@0001d_ol0oooo0?no
M00007@0_ooo0080oooo00L0okmd0000001d_ol0oooo0?oog`2LB7@0_ooo0080oooo00<0_g@007Bo
o`3oool0203oool00`000000oooo0?ooo`0^0?ooo`80o`002P3oool20?l00?l0oooo?03oool001T0
oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`0:0?ooo`030?oog`2LB7@0_ooo00P0
oooo00<000000?ooo`3oool0<03oool30?l000D0oooo0P3o003o0?oooch0oooo000H0?ooo`030?oo
_`1d0000B9cO00<0oooo00<0gia80000M02oool02`3oool00`2oM000M;oo0?ooo`070?ooo`030000
003oool0oooo03<0oooo1@3o003o0?oood00oooo000H0?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol02P3oool00`3ooml0W4P004RLg`080?ooo`030000003oool0oooo0?l0ooooN03o
ool001P0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030=nLB0000000
M;oo00P0oooo1000003o0?ooogL0oooo000H0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol01P3oool01`3Ogml0oooo0?oo_`1d000000000000M02oool0203oool00`000000oooo0?oo
o`3o0?ooogP0oooo000H0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool0
1P3OW4P00000000000000000001809cOo`T0oooo00<000000?ooo`3oool0o`3ooomh0?ooo`00603o
ool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00@0oono07@000000000B9cO
2`3oool00`000000oooo0?ooo`3o0?ooogP0oooo000I0?ooo`030=nLB000B9`0gooo0080oooo00<0
_g@0001d_`3oool01`3oool00`2oM000M;oo0?ooo`0;0?ooo`030000003oool0oooo0?l0ooooN03o
ool001T0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00H0omnL04P000000000
00000000B02Lgol80?ooo`030000003oool0oooo02/0oooo00<0og@00?l0003o00000`3o00000`3o
_ol0oooo0?ooo`3o0?oood@0oooo000J0?ooo`050?oo_`1d000000000000M02oool02P3oool01@2o
M0000000000000000000B9cO00P0oooo0`00000[0?ooo`030?oOW03o04P0omoo0?l0ooooBP3oool0
03H0oooo00<000000?ooo`3oool0;03oool00`3oW4P0o`1d0?ooo`3o0?ooodT0oooo000f0?ooo`03
0000003oool0oooo02<0oooo00D0oono0?l0003o0000o`000?l0M0050?ooo`030?nLB03o04P0omoo
0?l0ooooB03oool003H0oooo00<000000?ooo`3oool08P3oool00`3ookl0o`000?l000030?l00003
0?l0M03oool0oooo00<0oooo00<0oia80?l0M03oool0o`3ooom70?ooo`00=P3oool00`000000oooo
0?ooo`0R0?ooo`030?noM03o0000ogBo0080oooo00<0oia80?l0003o_ol01@3oool00`3oW4P0odRL
0?ooo`3o0?ooodH0oooo000f0?ooo`030000003oool0oooo0280oooo00<0oia80?l0B03ogol00P3o
ool00`3ogi`0o`000?nLg`050?ooo`030?oo_`3o0000okoo0?l0ooooAP3oool003H0oooo00<00000
0?ooo`3oool09`3oool01P3ookl0o`000?nLg`3oool0okmd0?m8W080oooo00<0oooO0?m8003o_ol0
o`3ooom60?ooo`00=P3oool3000002L0oooo00/0oono0?l0003oW=l0oooo0?oo_`3o0000okoo0?oo
o`3o_g@0o`180?oOo`3o0?ooodH0oooo000f0?ooo`030000003oool0oooo02L0oooo00<0oono0?l0
003oW=l00P3oool01@3ogi`0o`000?l0003o0000okoo0?l0ooooA`3oool003H0oooo00<000000?oo
o`3oool09`3oool00`3ookl0o`000?nLg`3o0?ooodh0oooo000f0?ooo`030000003oool0oooo02L0
oooo00<0oono0?l0003oW=l0o`3ooom>0?ooo`00=P3oool00`000000oooo0?ooo`0W0?ooo`030?oo
_`3o0000oicO0?l0ooooCP3oool003H0oooo00<000000?ooo`3oool09`3oool00`3ookl0o`000?nL
g`3o0?ooodh0oooo000f0?ooo`030000003oool0oooo02L0oooo00<0oono0?l0003oW=l0o`3ooom>
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-3.00627, -1.22856, \
0.0823617, 0.0111726}}]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Drawing with primitives", "Section"],

Cell[CellGroupData[{

Cell["Primitives", "Subsection"],

Cell[TextData[{
  "Relatively simple illustrations can often be very helpful in clarifying a \
discussion.  Although ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " does not provide drawing tools, simple illustrations can be made fairly \
easily using graphics primitives.  A list of basic primitives for drawing \
lines, points, circle, polygons, text and other simple graphics is found at \
",
  ButtonBox["Graphics",
    ButtonStyle->"RefGuideLink"],
  ".  Packages, such as ",
  StyleBox["Arrow", "InlineInput",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Shapes", "InlineInput",
    FontWeight->"Bold"],
  ", and ",
  StyleBox["Polyhedra", "InlineInput",
    FontWeight->"Bold"],
  ", expand the selection of primitives."
}], "Text"],

Cell["\<\
Here we offer a couple of simple examples, but do not attempt to cover \
drawing techniques in much depth.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Example: labelled triangle", "Subsection"],

Cell[TextData[{
  "One of the exercises at the end of ",
  StyleBox["getting started .nb",
    FontWeight->"Bold",
    FontSlant->"Italic"],
  " asked for a derivation of Heron's formula for the area of a triangle in \
terms of the lengths of its three sides using basic geometry and \
trigonometry.  The code below draws and labels a triangle.  ",
  StyleBox["Module", "InlineInput",
    FontWeight->"Bold"],
  " is used to isolate the internal coordinates, which are explained by \
embedded comments.  Notice that we use symbolic coordinates in ",
  StyleBox["Line", "InlineInput",
    FontWeight->"Bold"],
  " and ",
  StyleBox["Text", "InlineInput",
    FontWeight->"Bold"],
  " functions computed using simple relationships between a few basic \
vertices rather than employ numerical values in these functions; in that way \
the integrity of the figure can be maintained easily if we later decide to \
alter some of the details of its shape.  Also notice that because the figure \
can be resized easily, we place an important vertex at {0,0} and choose one \
important length, here the base, to have unit length and express all other \
lengths relative to that base.  The best positions for text elements depend \
somewhat upon the final size of the figure and require some tweaking, but do \
not take long to adjust.  Finally, although not really needed, we included \
graphics directives and options to modify the appearance of some elements of \
the figure."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(Module[\n\t\n\t (*\ \n\t\ta\  = \ 
            lower - left\ corner, \ \n\t\tb\  = \ 
            lower - right\ corner, \n\t\tx\  = \ 
            fraction\ of\ base\ where\ altitude\ intersects, \n\t\th\  = \ 
            height\ of\ triangle, \ \n\t\tstyle\  = \ 
            options\ for\ Text\n\t*) \n\t\ \n\t{a = {0, 0}, b = {1, 0}, 
          x = 0.7, h = 0.6, \n\t
          style = {FontSize \[Rule] 12, FontFamily \[Rule] "\<Arial\>", 
              FontColor \[Rule] RGBColor[0, 0, 1]}}, \n\t\n\t (*\n\t\tc\  = \ 
            upper\ vertex\ of\ triangle, \n\t\td\  = \ 
            lower\ end\ of\ altitude\n\t*) \n\t\n\td = x\ \((b - a)\); 
        c = d + {0, h}; \n\t
        Show[{\n\t\t\t\n\t\t\t (*\ 
              draw\ triangle\ *) \ \ \n\t\t\t\t\t\tGraphics[
              Line[{a, b, c, a}]], \n\t\t\t (*\ 
              draw\ altitude\ *) \n\t\t\t\t\t\tGraphics[{Dashing[{0.02, 
                    0.01}], Line[{d, c}]}], \n\t\t\t\n\t\t\t (*\ 
              label\ lengths\ and\ angles\ *) \n\t\ \ \ Graphics[{\n\t\t\t\t\t\
Text["\<a\>", \(a + c\)\/2 + {\(-0.05\), 0}, 
                  TextStyle \[Rule] 
                    style], \n\t\t\t\t\tText["\<b\>", \(a + b\)\/2 + {0.0, \
\(-0.05\)}, TextStyle \[Rule] 
                    style], \n\t\t\t\t\tText["\<c\>", \(b + c\)\/2 + {0.05, 
                      0}, TextStyle \[Rule] 
                    style], \n\t\t\t\t\tText["\<h\>", \(d + c\)\/2 + {\(-0.05\
\), 0}, TextStyle \[Rule] style], \n\t\t\t\t\tText["\<\[Alpha]\>", 
                  b + {0.03, \(-0.03\)}, 
                  TextStyle \[Rule] style], \n\t\t\t\t\tText["\<\[Beta]\>", 
                  c + {0.0, 0.03}, 
                  TextStyle \[Rule] style], \n\t\t\t\t\tText["\<\[Gamma]\>", 
                  a + {\(-0.03\), \(-0.03\)}, 
                  TextStyle \[Rule] style]\n\t\t\t\t}]\n\t}, 
          PlotRange \[Rule] All, 
          AspectRatio \[Rule] Automatic]\n\t];\)\), "\n", 
    \(\t\)}], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .64151 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0507637 0.898473 0.0601977 0.898473 [
[.32031 .32974 -5.34375 -7.71875 ]
[.32031 .32974 5.34375 7.71875 ]
[.5 .01527 -5.34375 -7.71875 ]
[.5 .01527 5.34375 7.71875 ]
[.85939 .32974 -5 -7.71875 ]
[.85939 .32974 5 7.71875 ]
[.63477 .32974 -5.34375 -7.71875 ]
[.63477 .32974 5.34375 7.71875 ]
[.97619 .03324 -5.53125 -7.71875 ]
[.97619 .03324 5.53125 7.71875 ]
[.67969 .62624 -5.65625 -7.71875 ]
[.67969 .62624 5.65625 7.71875 ]
[.02381 .03324 -5.21875 -7.71875 ]
[.02381 .03324 5.21875 7.71875 ]
[ 0 0 0 0 ]
[ 1 .64151 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.05076 .0602 m
.94924 .0602 L
.67969 .59928 L
.05076 .0602 L
s
[ .02 .01 ] 0 setdash
.67969 .0602 m
.67969 .59928 L
s
gsave
.32031 .32974 -66.3438 -11.7188 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.4375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Arial
%%IncludeFont: Arial
/Arial findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Arial
%%IncludeFont: Arial
/Arial findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(a) show
69.688 15.813 moveto
%%IncludeResource: font Arial
%%IncludeFont: Arial
/Arial findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.5 .01527 -66.3438 -11.7188 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.4375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Arial
%%IncludeFont: Arial
/Arial findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Arial
%%IncludeFont: Arial
/Arial findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(b) show
69.688 15.813 moveto
%%IncludeResource: font Arial
%%IncludeFont: Arial
/Arial findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.85939 .32974 -66 -11.7188 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.4375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Arial
%%IncludeFont: Arial
/Arial findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Arial
%%IncludeFont: Arial
/Arial findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(c) show
69.000 15.813 moveto
%%IncludeResource: font Arial
%%IncludeFont: Arial
/Arial findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.63477 .32974 -66.3438 -11.7188 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.4375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Arial
%%IncludeFont: Arial
/Arial findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Arial
%%IncludeFont: Arial
/Arial findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(h) show
69.688 15.813 moveto
%%IncludeResource: font Arial
%%IncludeFont: Arial
/Arial findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.97619 .03324 -66.5313 -11.7188 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.4375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Arial
%%IncludeFont: Arial
%%BeginResource: font Arial-MISO
%%BeginFont: Arial-MISO
/Arial /Arial-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Arial-MISO
%%IncludeFont: Arial-MISO
/Arial-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(a) show
70.063 15.813 moveto
%%IncludeResource: font Arial-MISO
%%IncludeFont: Arial-MISO
/Arial-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.67969 .62624 -66.6563 -11.7188 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.4375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Arial
%%IncludeFont: Arial
%%BeginResource: font Arial-MISO
%%BeginFont: Arial-MISO
/Arial /Arial-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Arial-MISO
%%IncludeFont: Arial-MISO
/Arial-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
64.188 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(b) show
70.313 15.813 moveto
%%IncludeResource: font Arial-MISO
%%IncludeFont: Arial-MISO
/Arial-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.02381 .03324 -66.2188 -11.7188 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.4375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Arial
%%IncludeFont: Arial
%%BeginResource: font Arial-MISO
%%BeginFont: Arial-MISO
/Arial /Arial-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Arial-MISO
%%IncludeFont: Arial-MISO
/Arial-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(g) show
69.438 15.813 moveto
%%IncludeResource: font Arial-MISO
%%IncludeFont: Arial-MISO
/Arial-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .64151 L
0 .64151 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282.313, 181.063},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006W00013b000`400?l00000o`00003oo`3ooonX
0?ooo`00o`3ooonX0?ooo`00o`3ooonX0?ooo`00o`3ooonX0?ooo`00o`3ooonX0?ooo`00o`3ooonX
0?ooo`00c`3oool0102LB?l004So0=ooo`1d0?l20000o`0309cOo`3oool0oooo0<l0oooo003?0?oo
o`0309a8o`00B?l0M4So00@0003o00<0M;oo0?ooo`3oool0cP3oool00<l0oooo00<0W4So0000o`00
M?l00P3oool0103o_ol0003o0018o`3Oooo?0?ooo`00c`3oool00`2LB?l0003o0;ooo`030?ooo`03
0=nLo`000?l0W=oo0<l0oooo000?0?ooo`0307@0o`18W?l0oooo0;d0oooo00<0W4So0018o`3Oool0
0`3oool00`3o_ol0003o07Boo`3?0?ooo`003`3oool00`2oM?l0003o0;ooo`2m0?ooo`0309a8o`00
B?l0gooo00<0oooo00<0okoo0000o`1d_ol0c`3oool000l0oooo00<0omoo04P0o`1d_ol0_@3oool0
0`2LB?l0003o0;ooo`030?ooo`030=nLo`000?l0W=oo0<l0oooo000@0?ooo`030?noo`18W?l0oooo
0;`0oooo00T0W4So0000o`00B?l0gooo0?ooo`3ogol0B03o0000o`2oool0c`3oool00140oooo00<0
W4So0;ooo`3oool0^`3oool00`2LB?l004So07A8o`040000o`0307Boo`3oool0oooo0;H0oooo00P0
gico0000o`000?l0003o0;ooo`1d0?l0003o001doa00oooo000A0?ooo`0309a8o`00M?l0oooo0;/0
oooo00@0W4So0018o`3Oool0M03o0P000?l00`1d_ol0oooo0?ooo`2f0?ooo`030?noo`000?l0M;oo
0080oooo00D0M03o0000o`2Lgol0okoo07Boo`0?0?ooo`004@3oool00`1d0?l0M7Co07Boo`2k0?oo
o`0309a8o`00B?l0gooo0;`0oooo00<0gico0000o`2oool00`3oool00`1d0?l0M;oo0?ooo`0@0?oo
o`004@3oool0101d0?l0W=oo07@0o`2ooonj0?ooo`0309a8o`00B?l0gooo0;`0oooo00<0gico0000
o`2oool00`3oool00`2oM?l0003o0;ooo`0@0?ooo`003@3oool0201dM?l0oooo0?ooo`3ogol0B03o
09cOo`3OW?l0B9co^P3oool00`2LB?l004So0=ooo`2l0?ooo`030?oOo`180?l0M;oo00<0oooo00@0
gico0018o`18B?l0gooo3`3oool000d0oooo00T0W4So09cOo`3oool0_gCo0000o`2oool0oooo07@0
o`2Lgol0o`3ooomj0?ooo`030;mdo`00B?l0gooo0080oooo00@0gico0018o`3O_ol007Co3`3oool0
00d0oooo00D0omoo04P0o`000?l0003o001do`020?ooo`030;mdo`00M?l0oooo0?l0ooooNP3oool0
2@2LB?l007Co0?ooo`3oool0_gCo0018o`3Oool0_gCo001do`0>0?ooo`003P3oool0103o_ol0003o
0000o`2oool20?ooo`030?oOo`180?l0W=oo0?l0ooooN`3oool01@3OW?l0003o0000o`000?l0W=oo
0140oooo003o0?ooojP0oooo003o0?ooojP0oooo003o0?ooojP0oooo003o0?ooojP0oooo003o0?oo
ojP0oooo003o0?ooojP0oooo003o0?ooojP0oooo003o0?ooojP0oooo000M0?ooool00000KP00000M
0?ooo`007P3oool00`000000oooo0?ooo`3j0?ooo`030000003oool0oooo06X0oooo00<000000?oo
o`3oool0703oool001l0oooo00<000000?ooo`3oool0n@3oool00`000000oooo0?ooo`1Z0?ooo`03
0000003oool0oooo01`0oooo000P0?ooo`800000n@3oool00`000000oooo0?ooo`1Y0?ooo`030000
003oool0oooo01d0oooo000R0?ooo`030000003oool0oooo0?H0oooo00<000000?ooo`3oool0J@3o
ool00`000000oooo0?ooo`0M0?ooo`008`3oool00`000000oooo0?ooo`3e0?ooo`030000003oool0
oooo06P0oooo00<000000?ooo`3oool07P3oool002@0oooo00<000000?ooo`3oool0m03oool00`00
0000oooo0?ooo`1X0?ooo`030000003oool0oooo01h0oooo000U0?ooo`030000003oool0oooo0?<0
oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`0O0?ooo`009P3oool00`000000oooo
0?ooo`3b0?ooo`030000003oool0oooo06L0oooo00<000000?ooo`3oool07`3oool002L0oooo0P00
003o0?oooe`0oooo00<000000?ooo`3oool0803oool002T0oooo00<000000?ooo`3oool0o`3ooomI
0?ooo`030000003oool0oooo0200oooo000Z0?ooo`030000003oool0oooo0?l0ooooE`3oool00`00
0000oooo0?ooo`0Q0?ooo`00:`3oool00`000000oooo0?ooo`3]0?ooo`030000003oool0oooo06D0
oooo00<000000?ooo`3oool08@3oool002`0oooo00<000000?ooo`3oool0k03oool00`000000oooo
0?ooo`1T0?ooo`030000003oool0oooo0280oooo000]0?ooo`030000003oool0oooo0>/0oooo00<0
00000?ooo`3oool0I03oool00`000000oooo0?ooo`0R0?ooo`00;P3oool200000>/0oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`0S0?ooo`00<03oool00`000000oooo0?ooo`3X0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool08`3oool00340oooo00<000000?ooo`3o
ool0i`3oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo02@0oooo000b0?ooo`030000
003oool0oooo0>H0oooo00<000000?ooo`3oool0HP3oool00`000000oooo0?ooo`0T0?ooo`00<`3o
ool00`000000oooo0?ooo`3U0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool09@3o
ool003@0oooo00<000000?ooo`3oool0i03oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0
oooo02D0oooo000e0?ooo`800000o`3ooom80?ooo`030000003oool0oooo02H0oooo000g0?ooo`03
0000003oool0oooo0?l0ooooA@3oool00`000000oooo0?ooo`0V0?ooo`00>03oool00`000000oooo
0?ooo`3o0?oood<0oooo00<000000?ooo`3oool09`3oool003T0oooo00<000000?ooo`3oool0g`3o
ool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo02L0oooo000j0?ooo`030000003oool0
oooo0=h0oooo00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`0X0?ooo`00>`3oool00`00
0000oooo0?ooo`3M0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool0:03oool003`0
oooo0P00003M0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool0:@3oool003h0oooo
00<000000?ooo`3oool0fP3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo02T0oooo
000o0?ooo`030000003oool0oooo0=T0oooo00<000000?ooo`3oool0G03oool00`000000oooo0?oo
o`0Z0?ooo`00@03oool00`000000oooo0?ooo`3H0?ooo`030000003oool0oooo05`0oooo00<00000
0?ooo`3oool0:P3oool00440oooo00<000000?ooo`3oool0e`3oool00`000000oooo0?ooo`1K0?oo
o`030000003oool0oooo02/0oooo00120?ooo`030000003oool0oooo0=H0oooo00<000000?ooo`3o
ool0F`3oool00`000000oooo0?ooo`0[0?ooo`00@`3oool200000?l0oooo=03oool00`000000oooo
0?ooo`0/0?ooo`00A@3oool00`000000oooo0?ooo`3o0?oooc40oooo00<000000?ooo`3oool0;03o
ool004H0oooo00<000000?ooo`3oool0o`3oool_0?ooo`030000003oool0oooo02d0oooo00170?oo
o`030000003oool0oooo0=40oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`0]0?oo
o`00B03oool00`000000oooo0?ooo`3@0?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3o
ool0;P3oool004T0oooo00<000000?ooo`3oool0c`3oool00`000000oooo0?ooo`1H0?ooo`030000
003oool0oooo02h0oooo001:0?ooo`800000c`3oool00`000000oooo0?ooo`1G0?ooo`030000003o
ool0oooo02l0oooo001<0?ooo`030000003oool0oooo0<`0oooo00<000000?ooo`3oool0E`3oool0
0`000000oooo0?ooo`0_0?ooo`00C@3oool00`000000oooo0?ooo`3;0?ooo`030000003oool0oooo
05H0oooo00<000000?ooo`3oool0<03oool004h0oooo00<000000?ooo`3oool0bP3oool00`000000
oooo0?ooo`1F0?ooo`030000003oool0oooo0300oooo001?0?ooo`030000003oool0oooo0<T0oooo
00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`0a0?ooo`00D03oool00`000000oooo0?oo
o`380?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0<@3oool00540oooo0P00003o
0?ooob00oooo00<000000?ooo`3oool0<P3oool005<0oooo00<000000?ooo`3oool0o`3ooolM0?oo
o`030000003oool0oooo0380oooo001D0?ooo`030000003oool0oooo0?l0oooo6`3oool00`000000
oooo0?ooo`0c0?ooo`00E@3oool00`000000oooo0?ooo`330?ooo`030000003oool0oooo05<0oooo
00<000000?ooo`3oool0<`3oool005H0oooo00<000000?ooo`3oool0`P3oool00`000000oooo0?oo
o`1B0?ooo`030000003oool0oooo03@0oooo001G0?ooo`030000003oool0oooo0<40oooo00<00000
0?ooo`3oool0DP3oool00`000000oooo0?ooo`0d0?ooo`00F03oool200000<40oooo00<000000?oo
o`3oool0D@3oool00`000000oooo0?ooo`0e0?ooo`00FP3oool00`000000oooo0?ooo`2n0?ooo`03
0000003oool0oooo0540oooo00<000000?ooo`3oool0=@3oool005/0oooo00<000000?ooo`3oool0
_@3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo03H0oooo001L0?ooo`030000003o
ool0oooo0;`0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`0f0?ooo`00G@3oool0
0`000000oooo0?ooo`2k0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0=`3oool0
05h0oooo00<000000?ooo`3oool0^P3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo
03L0oooo001O0?ooo`030000003oool0oooo0?l0oooo2`3oool00`000000oooo0?ooo`0h0?ooo`00
H03oool200000?l0oooo2`3oool00`000000oooo0?ooo`0h0?ooo`00HP3oool00`000000oooo0?oo
o`3o0?ooo`L0oooo00<000000?ooo`3oool0>@3oool006<0oooo00<000000?ooo`3oool0]@3oool0
0`000000oooo0?ooo`1=0?ooo`030000003oool0oooo03T0oooo001T0?ooo`030000003oool0oooo
0;@0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`0j0?ooo`00I@3oool00`000000
oooo0?ooo`2c0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0>P3oool006H0oooo
00<000000?ooo`3oool0/P3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo03/0oooo
001W0?ooo`800000/P3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo03/0oooo001Y
0?ooo`030000003oool0oooo0:l0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0l
0?ooo`00JP3oool00`000000oooo0?ooo`2^0?ooo`030000003oool0oooo04X0oooo00<000000?oo
o`3oool0?03oool006/0oooo00<000000?ooo`3oool0[@3oool00`000000oooo0?ooo`190?ooo`03
0000003oool0oooo03d0oooo001/0?ooo`030000003oool0oooo0:`0oooo00<000000?ooo`3oool0
B@3oool00`000000oooo0?ooo`0m0?ooo`00K@3oool00`000000oooo0?ooo`2[0?ooo`030000003o
ool0oooo04P0oooo00<000000?ooo`3oool0?P3oool006h0oooo0P00003f0?ooo`030000003oool0
oooo03h0oooo001`0?ooo`030000003oool0oooo0?80oooo00<000000?ooo`3oool0?`3oool00740
oooo00<000000?ooo`3oool0l@3oool00`000000oooo0?ooo`0o0?ooo`00LP3oool00`000000oooo
0?ooo`2V0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0@03oool007<0oooo00<0
00000?ooo`3oool0Y@3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo0400oooo001d
0?ooo`030000003oool0oooo0:@0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`11
0?ooo`00M@3oool200000:@0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`110?oo
o`00M`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3o
ool0@P3oool007P0oooo00<000000?ooo`3oool0X03oool00`000000oooo0?ooo`140?ooo`030000
003oool0oooo0480oooo001i0?ooo`030000003oool0oooo09l0oooo00<000000?ooo`3oool0@`3o
ool00`000000oooo0?ooo`130?ooo`00NP3oool00`000000oooo0?ooo`2N0?ooo`030000003oool0
oooo04<0oooo00<000000?ooo`3oool0@`3oool007/0oooo00<000000?ooo`3oool0W@3oool00`00
0000oooo0?ooo`120?ooo`030000003oool0oooo04@0oooo001l0?ooo`800000hP3oool00`000000
oooo0?ooo`140?ooo`00OP3oool00`000000oooo0?ooo`3N0?ooo`030000003oool0oooo04D0oooo
001o0?ooo`030000003oool0oooo0=d0oooo00<000000?ooo`3oool0A@3oool00800oooo00<00000
0?ooo`3oool0V03oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo04H0oooo00210?oo
o`030000003oool0oooo09L0oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`160?oo
o`00PP3oool00`000000oooo0?ooo`2F0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3o
ool0A`3oool008<0oooo0P00002F0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0
A`3oool008D0oooo00<000000?ooo`3oool0T`3oool00`000000oooo0?ooo`0n0?ooo`030000003o
ool0oooo04P0oooo00260?ooo`030000003oool0oooo0980oooo00<000000?ooo`3oool0?P3oool0
0`000000oooo0?ooo`180?ooo`00Q`3oool00`000000oooo0?ooo`2A0?ooo`030000003oool0oooo
03d0oooo00<000000?ooo`3oool0B@3oool008P0oooo00<000000?ooo`3oool0T03oool00`000000
oooo0?ooo`0m0?ooo`030000003oool0oooo04T0oooo00290?ooo`030000003oool0oooo08l0oooo
00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`1:0?ooo`00RP3oool200000<h0oooo00<0
00000?ooo`3oool0BP3oool008`0oooo00<000000?ooo`3oool0bP3oool00`000000oooo0?ooo`1;
0?ooo`00S@3oool00`000000oooo0?ooo`390?ooo`030000003oool0oooo04/0oooo002>0?ooo`03
0000003oool0oooo08X0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`1<0?ooo`00
S`3oool00`000000oooo0?ooo`290?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0
C03oool00900oooo00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`0i0?ooo`030000003o
ool0oooo04d0oooo002A0?ooo`800000R03oool00`000000oooo0?ooo`0i0?ooo`030000003oool0
oooo04d0oooo002C0?ooo`030000003oool0oooo08D0oooo00<000000?ooo`3oool0>03oool00`00
0000oooo0?ooo`1>0?ooo`00U03oool00`000000oooo0?ooo`240?ooo`030000003oool0oooo03P0
oooo00<000000?ooo`3oool0CP3oool009D0oooo00<000000?ooo`3oool0P`3oool00`000000oooo
0?ooo`0g0?ooo`030000003oool0oooo04l0oooo00270?ooo`0807@0o`000?l0003o0000o`2Lgol0
_gCo0000o`2oool70?ooo`030000003oool0oooo06`0oooo00<0W4So0018o`3Oool00`3oool00`2L
B?l007Co0?ooo`0=0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool02P3oool01@3o
_ol0003o0000o`000?l0M;oo0400oooo00260?ooo`0309a8o`000?l0003o00@0003o00<004So0=oo
o`3oool01`3oool00`000000oooo0?ooo`1[0?ooo`0309a8o`00B?l0gooo00<0oooo00<0W4So001d
o`3oool03@3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo00X0oooo00<0okoo0000
o`000?l00`000?l00`1d_ol0oooo0?ooo`0m0?ooo`00Q@3oool00`3ogol0B03o001do`030?ooo`04
07@0o`000?l004So0=ooo`T0oooo0P00001[0?ooo`0309a8o`00B?l0gooo00<0oooo00<0W4So001d
o`3oool0AP3oool00`000000oooo0?ooo`0:0?ooo`0309a8o`000?l0W=oo0080oooo00<0M03o0018
o`3Oool0?P3oool008D0oooo00<0omoo04P0o`1d_ol00`3oool00`3ogol0B03o001do`0<0?ooo`03
0000003oool0oooo06P0oooo00<0W4So0018o`3Oool00`3oool00`2LB?l007Co0?ooo`150?ooo`03
0000003oool0oooo00X0oooo00<0omoo04P0o`18W?l0A03oool008H0oooo00@0M03o0000o`000?l0
M;oo0P3oool00`1d0?l007Co0?ooo`0<0?ooo`030000003oool0oooo06L0oooo00<0W4So0018o`3O
ool00`3oool00`2LB?l007Co0?ooo`150?ooo`030000003oool0oooo00X0oooo00<0omoo04P0o`1d
_ol0A03oool008H0oooo00<0omoo04P0o`000?l010000?l00`00M?l0oooo0?ooo`0<0?ooo`030000
003oool0oooo06H0oooo00<0W4So0018o`3Oool00`3oool00`2LB?l007Co0?ooo`0=0?ooo`030000
003oool0oooo03@0oooo00<000000?ooo`3oool02`3oool00`3ogol0B03o07Boo`140?ooo`00RP3o
ool0101d0?l0003o0000o`00M?l?0?ooo`030000003oool0oooo06D0oooo00<0W4So0000o`2oool0
0`3oool00`1d0?l007Co0?ooo`0=0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0
303oool00`3ogol0B03o04RLo`140?ooo`00QP3oool00`2LB?l0003o0;ooo`020?ooo`030?oOo`18
0?l007Co0100oooo00<000000?ooo`3oool0I03oool0202LB?l0003o0018o`3Oool0oooo0?noo`00
0?l007Co3P3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo00d0oooo00L0W4So0000
o`2Lgol0oooo0?noo`000?l0B9co03l0oooo00260?ooo`030=nLo`000?l0003o00@0003o00<0M;oo
0?ooo`3oool03`3oool2000006@0oooo00<0W4So0018o`1dB?l010000?l00`1d_ol0oooo0?ooo`0<
0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool03P3oool00`3OW?l0003o0000o`03
0000o`0309cOo`3oool0oooo03d0oooo00270?ooo`030?noo`000?l0003o0080003o00<0B9co0?oo
o`3oool04P3oool00`000000oooo0?ooo`1Q0?ooo`0409a8o`00B?l0gooo07@0o`80003o00<0B9co
0?ooo`3oool03@3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo00l0oooo00D0okoo
0000o`000?l0003o09cOo`100?ooo`00XP3oool00`000000oooo0?ooo`1P0?ooo`0309a8o`00B?l0
gooo01<0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`1E0?ooo`00X`3oool00`00
0000oooo0?ooo`1O0?ooo`0309a8o`00B?l0gooo01<0oooo00<000000?ooo`3oool0<@3oool00`00
0000oooo0?ooo`1E0?ooo`00Y03oool00`000000oooo0?ooo`1N0?ooo`0309a8o`00B?l0gooo01<0
oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`1F0?ooo`00Y@3oool00`000000oooo
0?ooo`1c0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0EP3oool00:H0oooo0P00
002U0?ooo`030000003oool0oooo05L0oooo002X0?ooo`030000003oool0oooo0:80oooo00<00000
0?ooo`3oool0E`3oool00:T0oooo00<000000?ooo`3oool0X03oool00`000000oooo0?ooo`1H0?oo
o`00ZP3oool00`000000oooo0?ooo`1^0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3o
ool0F03oool00:/0oooo00<000000?ooo`3oool0K@3oool00`000000oooo0?ooo`0]0?ooo`030000
003oool0oooo05T0oooo002/0?ooo`030000003oool0oooo06`0oooo00<000000?ooo`3oool0;@3o
ool00`000000oooo0?ooo`1I0?ooo`00[@3oool2000006`0oooo00<000000?ooo`3oool0;03oool0
0`000000oooo0?ooo`1J0?ooo`00[`3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo
02`0oooo00<000000?ooo`3oool0FP3oool00;00oooo00<000000?ooo`3oool0J03oool00`000000
oooo0?ooo`0[0?ooo`030000003oool0oooo05/0oooo002a0?ooo`030000003oool0oooo06L0oooo
00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`1K0?ooo`00/P3oool00`000000oooo0?oo
o`1V0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0G03oool00;<0oooo00<00000
0?ooo`3oool0I@3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo05`0oooo002d0?oo
o`800000T@3oool00`000000oooo0?ooo`1M0?ooo`00]P3oool00`000000oooo0?ooo`2>0?ooo`03
0000003oool0oooo05d0oooo002g0?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool0
GP3oool00;P0oooo00<000000?ooo`3oool0H03oool00`000000oooo0?ooo`0X0?ooo`030000003o
ool0oooo05h0oooo002i0?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3oool09`3oool0
0`000000oooo0?ooo`1O0?ooo`00^P3oool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo
02L0oooo00<000000?ooo`3oool0G`3oool00;/0oooo0P00001N0?ooo`030000003oool0oooo02H0
oooo00<000000?ooo`3oool0H03oool00;d0oooo00<000000?ooo`3oool0F`3oool00`000000oooo
0?ooo`0V0?ooo`030000003oool0oooo0600oooo002n0?ooo`030000003oool0oooo05X0oooo00<0
00000?ooo`3oool09@3oool00`000000oooo0?ooo`1Q0?ooo`00_`3oool00`000000oooo0?ooo`1I
0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0H@3oool00<00oooo00<000000?oo
o`3oool0F03oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo0680oooo00310?ooo`03
0000003oool0oooo05L0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`1R0?ooo`00
`P3oool2000007d0oooo00<000000?ooo`3oool0H`3oool00<@0oooo00<000000?ooo`3oool0NP3o
ool00`000000oooo0?ooo`1S0?ooo`00a@3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0
oooo06@0oooo00360?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool0I03oool00<L0
oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo06D0
oooo00380?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool08@3oool00`000000oooo
0?ooo`1U0?ooo`00b@3oool200000500oooo00<000000?ooo`3oool0803oool00`000000oooo0?oo
o`1V0?ooo`00b`3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo0200oooo00<00000
0?ooo`3oool0IP3oool00<`0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`0O0?oo
o`030000003oool0oooo06L0oooo003=0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3o
ool07`3oool00`000000oooo0?ooo`1W0?ooo`00cP3oool00`000000oooo0?ooo`1:0?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool0J03oool00<l0oooo00<000000?ooo`3oool0B@3o
ool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo06P0oooo003@0?ooo`800000B@3oool0
0`000000oooo0?ooo`0M0?ooo`030000003oool0oooo06T0oooo003B0?ooo`030000003oool0oooo
06H0oooo00<000000?ooo`3oool0J@3oool00=<0oooo00<000000?ooo`3oool0I03oool00`000000
oooo0?ooo`1Z0?ooo`00e03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo06X0oooo
003E0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?oo
o`1[0?ooo`00eP3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo01/0oooo00<00000
0?ooo`3oool0J`3oool00=L0oooo0P0000120?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool0K03oool00=T0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo06`0oooo003J0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`1]0?ooo`00f`3oool00`000000oooo0?ooo`0m0?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool0K@3oool00=`0oooo00<000000?ooo`3oool0?03oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo06h0oooo003M0?ooo`030000003oool0oooo
03/0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1^0?ooo`00gP3oool00`000000
oooo0?ooo`0j0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0K`3oool00=l0oooo
0P00001D0?ooo`030000003oool0oooo06l0oooo003Q0?ooo`030000003oool0oooo0500oooo00<0
00000?ooo`3oool0L03oool00>80oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`1`
0?ooo`00h`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo01D0oooo00<000000?oo
o`3oool0L@3oool00>@0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0E0?ooo`03
0000003oool0oooo0740oooo003U0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0
503oool00`000000oooo0?ooo`1b0?ooo`00iP3oool2000003<0oooo00<000000?ooo`3oool0503o
ool00`000000oooo0?ooo`1b0?ooo`00j03oool00`000000oooo0?ooo`0`0?ooo`030000003oool0
oooo01<0oooo00<000000?ooo`3oool0L`3oool00>T0oooo00<000000?ooo`3oool0;`3oool00`00
0000oooo0?ooo`0C0?ooo`030000003oool0oooo07<0oooo003Z0?ooo`030000003oool0oooo02h0
oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`1d0?ooo`00j`3oool00`000000oooo
0?ooo`0]0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0M03oool00>`0oooo00<0
00000?ooo`3oool0;03oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo07D0oooo003]
0?ooo`800000@03oool00`000000oooo0?ooo`1e0?ooo`00k`3oool00`000000oooo0?ooo`0l0?oo
o`030000003oool0oooo07H0oooo003`0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3o
ool0MP3oool00?40oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0?0?ooo`030000
003oool0oooo07L0oooo003b0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`1g0?ooo`00l`3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool0N03oool00?@0oooo0P00000U0?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool0N03oool00?H0oooo00<000000?ooo`3oool08P3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo07T0oooo003g0?ooo`030000003oool0oooo0240oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1i0?ooo`00n03oool00`000000oooo0?oo
o`0P0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0NP3oool00?T0oooo00<00000
0?ooo`3oool07`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo07X0oooo003j0?oo
o`030000003oool0oooo01h0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1k0?oo
o`00n`3oool2000002`0oooo00<000000?ooo`3oool0N`3oool00?d0oooo00<000000?ooo`3oool0
:03oool00`000000oooo0?ooo`1l0?ooo`00oP3oool00`000000oooo0?ooo`0W0?ooo`030000003o
ool0oooo07`0oooo003o0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`1m0?ooo`00o`3oool10?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`1m0?ooo`00o`3oool20?ooo`030000003oool0oooo01L0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`1n0?ooo`00o`3oool30?ooo`800000
5`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo07h0oooo003o0?ooo`D0oooo00<0
00000?ooo`3oool0503oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo07l0oooo003o
0?ooo`H0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo07l0oooo003o0?ooo`L0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo0800oooo003o0?ooo`P0oooo00<000000?ooo`3oool04@3oool00`000000
oooo0?ooo`060?ooo`030000003oool0oooo0800oooo003o0?ooo`T0oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0840oooo003o0?ooo`X0oooo0P00
000H0?ooo`030000003oool0oooo0840oooo003o0?ooo``0oooo00<000000?ooo`3oool0503oool0
0`000000oooo0?ooo`220?ooo`00o`3oool=0?ooo`030000003oool0oooo01<0oooo00<000000?oo
o`3oool0PP3oool00?l0oooo3P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0P`3oool00?l0oooo3`3oool00`000000oooo0?ooo`0:0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool0P`3oool00?l0oooo403oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0Q03oool00?l0oooo4@3oool2
000000T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`240?ooo`00o`3ooolC0?oo
o`030000003oool0oooo00H0oooo00D000000?ooo`3oool0oooo000000270?ooo`00o`3ooolD0?oo
o`030000003oool0oooo00D0oooo00D000000?ooo`3oool0oooo000000270?ooo`00o`3ooolE0?oo
o`030000003oool0oooo00@0oooo00@000000?ooo`3oool00000R03oool00?l0oooo5P3oool00`00
0000oooo0?ooo`030?ooo`040000003oool0oooo000008P0oooo003o0?oooaL0oooo00<000000?oo
o`3oool00P3oool00`000000oooo000000290?ooo`00o`3ooolH0?ooo`800000103oool00`000000
oooo0?ooo`270?ooo`00o`3ooolJ0?ooo`040000003oool0oooo000008X0oooo003o0?oooa/0oooo
00<000000?ooo`000000RP3oool00?l0oooo703oool00`000000oooo0?ooo`290?ooo`00o`3ooonX
0?ooo`00o`3ooolG0?ooo`030;mdo`000?l0W=oo08h0oooo003o0?oooaP0oooo00<0M03o04RLo`3o
ool0S@3oool00?l0oooo603oool00`2oM?l007Co0?ooo`2=0?ooo`00o`3ooolH0?ooo`030=nLo`00
B?l0gooo08d0oooo003o0?oooaP0oooo00@0omoo04P0o`2Lgol0W4So0P000?l00`00B?l0gooo0?oo
o`270?ooo`00o`3ooolI0?ooo`0307@0o`000?l0M;oo0080oooo00<0_gCo0000o`2Lgol0Q`3oool0
0?l0oooo6@3oool00`2oM?l004So0=ooo`030?ooo`0309a8o`00B?l0gooo08H0oooo003o0?oooaT0
oooo00<0gico0018o`3Oool00`3oool00`3OW?l0003o09cOo`260?ooo`00o`3ooolI0?ooo`030?oO
o`180?l0W=oo00<0oooo00<0okoo0000o`1d_ol0QP3oool00?l0oooo6P3oool00`1d0?l0M;oo0?oo
o`020?ooo`030?noo`000?l0W=oo08H0oooo003o0?oooaX0oooo00<0_gCo001do`3oool00P3oool0
0`2oM?l007Co0?ooo`260?ooo`00o`3ooolJ0?ooo`070?noo`00B?l0gooo07@0o`000?l004So0=oo
o`270?ooo`00o`3ooolJ0?ooo`030?oOo`180?l0W=oo0080oooo00<0omoo04P0o`2oool0QP3oool0
0?l0oooo6`3oool00`2LB?l0M;oo0?ooo`020?ooo`030;mdo`00M?l0oooo08D0oooo003o0?oooa/0
oooo00<0gico0018o`3Oool00P3oool00`3OW?l0003o0;ooo`250?ooo`00o`3ooolL0?ooo`0709a8
o`00M?l0oooo0?ooo`2oM?l0003o0;ooo`250?ooo`00o`3ooolM0?ooo`050;mdo`000?l0003o0000
o`2Lgol0QP3oool00?l0ooooZ03oool00?l0ooooZ03oool00?l0ooooZ03oool00?l0ooooZ03oool0
0001\
\>"],
  ImageRangeCache->{{{0, 422.438}, {270.563, 0}} -> {-0.0805521, -0.0818313, \
0.00412287, 0.00412287}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Example: quark model", "Subsection"],

Cell[TextData[{
  "The quark model suggests that protons (and other particles in the hadron \
family) contain three constituents which each carry intrinsic angular \
momentum, which is called ",
  StyleBox["spin",
    FontSlant->"Italic"],
  " and is displayed as an arrow.  In addition, quarks have two internal \
coordinates whimsically known as ",
  StyleBox["flavor",
    FontSlant->"Italic"],
  " and ",
  StyleBox["color",
    FontSlant->"Italic"],
  ".  Two flavors are needed for the proton, two ",
  StyleBox["u",
    FontSlant->"Italic"],
  " or up quarks and one ",
  StyleBox["d",
    FontSlant->"Italic"],
  " or down quark, and three primary colors must be combined to form a net \
colorless (white) object.  The module ",
  StyleBox["DrawQuark", "InlineInput",
    FontWeight->"Bold"],
  " displays a single quark as a colored circle labeled by flavor \
superimposed upon a spin arrow.  The module ",
  StyleBox["DrawHadron", "InlineInput",
    FontWeight->"Bold"],
  " places three quarks within a circular or elliptical outline.  Our purpose \
here is not to explain the physics of this model, but to demonstrate that \
some relatively sophisticated illustrations can be created fairly easily \
using graphics primitives."
}], "Text"],

Cell[TextData[{
  "This somewhat more complicated example employs the graphics primitives ",
  StyleBox["Circle", "InlineInput",
    FontWeight->"Bold"],
  " and ",
  StyleBox["Arrow", "InlineInput",
    FontWeight->"Bold"],
  " in addition to ",
  StyleBox["Text", "InlineInput",
    FontWeight->"Bold"],
  ".  Note that ",
  StyleBox["Arrow", "InlineInput",
    FontWeight->"Bold"],
  " is found within a package by the same name.  These modules are little too \
long to explain in much detail \[LongDash] please study them until their \
operation becomes apparent."
}], "Text"],

Cell[BoxData[
    \(Needs["\<Graphics`Arrow`\>"]\)], "Input"],

Cell[BoxData[{
    \(\(Clear[DrawQuark, u, d];\)\), "\n", 
    \(\(DrawQuark[position_, flavor_, color_, spin_] := 
        Module[{a = {0, 0}, 
            labelStyle = 
              Sequence[FontSize \[Rule] 18, FontWeight \[Rule] "\<Bold\>", 
                FontSlant \[Rule] "\<Italic\>", 
                FontFamily \[Rule] "\<Times\>"]}, 
          Show[{Graphics[{Thickness[0.02], 
                  Arrow[position - {0.01, 0.02}, position + {0.01, 0.02}, 
                    HeadScaling \[Rule] Relative]}], 
              Graphics[{PointSize[1], 
                  RGBColor[\((color /. List \[Rule] Sequence)\)], 
                  Point[position]}], 
              Graphics[Text[StyleForm[flavor, labelStyle], position]]}, 
            AspectRatio \[Rule] Automatic, 
            ImageSize \[Rule] 60]];\)\)}], "Input"],

Cell[BoxData[{
    \(\(Clear[DrawHadron];\)\), "\n", 
    \(\(DrawHadron[name_, positions_, flavors_, colors_, spins_, 
          deformation_, opts___Rule] := 
        Module[{n, quarks = {}, radius = 0.5, arrowLength, begin, end, 
            labelStyle = 
              Sequence[FontSize \[Rule] 18, FontColor \[Rule] GrayLevel[0], 
                FontWeight \[Rule] "\<Bold\>", 
                FontSlant \[Rule] "\<Italic\>", 
                FontFamily \[Rule] "\<Times\>"]}, n = Length[positions]; 
          arrowLength = 0.75\ radius; \[IndentingNewLine]Do[
            begin = positions\[LeftDoubleBracket]i\[RightDoubleBracket] - 
                0.5  arrowLength {Cos[
                      spins\[LeftDoubleBracket]i\[RightDoubleBracket]], 
                    Sin[spins\[LeftDoubleBracket]i\[RightDoubleBracket]]}; 
            end = positions\[LeftDoubleBracket]i\[RightDoubleBracket] + 
                0.5  arrowLength {Cos[
                      spins\[LeftDoubleBracket]i\[RightDoubleBracket]], 
                    Sin[spins\[LeftDoubleBracket]i\[RightDoubleBracket]]}; 
            AppendTo[quarks, 
              Graphics[{Thickness[0.01], 
                  Arrow[begin, end, HeadScaling \[Rule] Relative], 
                  PointSize[0.175], 
                  RGBColor[
                    colors\[LeftDoubleBracket]i\[RightDoubleBracket] /. 
                      List \[Rule] Sequence], 
                  Point[positions\[LeftDoubleBracket]i\[RightDoubleBracket]], 
                  Text[StyleForm[
                      flavors\[LeftDoubleBracket]i\[RightDoubleBracket], 
                      labelStyle], 
                    positions\[LeftDoubleBracket]
                      i\[RightDoubleBracket]]}]], {i, 1, 
              n}]; \[IndentingNewLine]Show[{Graphics[
                Circle[{0, 0}, {\((1 + deformation)\) radius, radius}]], 
              quarks, Graphics[
                Text[StyleForm[name, labelStyle], {0, \(-1.2\)\ radius}]]}, 
            opts, AspectRatio \[Rule] Automatic]];\)\)}], "Input"],

Cell[TextData[{
  "The following statement creates a graphics array that compares quark \
models for the proton, neutron, and ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[CapitalDelta]\^+\)\)]],
  " particles.  Notice that the last is shown as an ellipse because it is \
believed to be slightly deformed (nonspherical)."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(range = 
        PlotRange \[Rule] {Automatic, {\(-0.7\), 
              0.6}};\)\), "\[IndentingNewLine]", 
    \(\(DisplayTogetherArray[{DrawHadron[
            proton, {{0. , 0.2}, {\(-0.2\), \(-0.2\)}, {0.3, \(-0.1\)}}, {u, 
              u, d}, {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}}, {60\ Degree, 
              120\ Degree, \(-60\) Degree}, 0. , 
            range], \[IndentingNewLine]DrawHadron[
            neutron, {{0. , 0.2}, {\(-0.2\), \(-0.2\)}, {0.3, \(-0.1\)}}, {d, 
              d, u}, {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}}, {60\ Degree, 
              120\ Degree, \(-60\) Degree}, 0. , 
            range], \[IndentingNewLine]DrawHadron[\*"\"\<\!\(\(\[CapitalDelta]\
\^+\)\)\>\"", {{0. , 0.2}, {\(-0.2\), \(-0.2\)}, {0.3, \(-0.1\)}}, {u, u, 
              d}, {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}}, {60\ Degree, 
              120\ Degree, 60  Degree}, \(-0.1\), range]}];\)\)}], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .3869 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.31746 0.00921202 0.31746 [
[ 0 0 0 0 ]
[ 1 .3869 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .3869 L
0 .3869 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
arrl arrl scale
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-0.2 0.05 moveto 0 0 lineto
-0.2 -0.05 lineto fill
-0.2 0.05 moveto 0 0 lineto
-0.2 -0.05 lineto
-0.2 0.05 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
% Start of sub-graphic
p
0.0238095 0.00921202 0.321429 0.377693 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.952381 0.666667 0.952381 [
[ 0 0 0 0 ]
[ 1 1.2381 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.2381 L
0 1.2381 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
newpath
.5 .66667 .47619 0 365.73 arc
s
.01 w
.41071 .7025 m
.58929 1.01179 L
s
% Start of user PostScript
-0.09375 0.037620237 0.09375 0.36237976 MAarrowhead1
% End of user PostScript
1 0 0 r
.175 w
.5 .85714 Mdot
gsave
.5 .85714 -67.9688 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(u) show
1.000 setlinewidth
grestore
0 g
.01 w
.39881 .32154 m
.22024 .63084 L
s
% Start of user PostScript
-0.10625 -0.36237976 -0.29375 -0.037620237 MAarrowhead1
% End of user PostScript
0 1 0 r
.175 w
.30952 .47619 Mdot
gsave
.30952 .47619 -67.9688 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(u) show
1.000 setlinewidth
grestore
0 g
.01 w
.69643 .72608 m
.875 .41678 L
s
% Start of user PostScript
0.20625 0.062379763 0.39375 -0.26237976 MAarrowhead1
% End of user PostScript
0 0 1 r
.175 w
.78571 .57143 Mdot
gsave
.78571 .57143 -67.9688 -12.5 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 17.125 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(d) show
1.000 setlinewidth
grestore
0 g
gsave
.5 .09524 -88.5625 -12.3125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 24.625 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
65.250 15.813 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(proton) show
1.000 setlinewidth
grestore
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.35119 0.00921202 0.64881 0.377693 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.952381 0.666667 0.952381 [
[ 0 0 0 0 ]
[ 1 1.2381 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.2381 L
0 1.2381 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
newpath
.5 .66667 .47619 0 365.73 arc
s
.01 w
.41071 .7025 m
.58929 1.01179 L
s
% Start of user PostScript
-0.09375 0.037620237 0.09375 0.36237976 MAarrowhead1
% End of user PostScript
1 0 0 r
.175 w
.5 .85714 Mdot
gsave
.5 .85714 -67.9688 -12.5 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 17.125 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(d) show
1.000 setlinewidth
grestore
0 g
.01 w
.39881 .32154 m
.22024 .63084 L
s
% Start of user PostScript
-0.10625 -0.36237976 -0.29375 -0.037620237 MAarrowhead1
% End of user PostScript
0 1 0 r
.175 w
.30952 .47619 Mdot
gsave
.30952 .47619 -67.9688 -12.5 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 17.125 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(d) show
1.000 setlinewidth
grestore
0 g
.01 w
.69643 .72608 m
.875 .41678 L
s
% Start of user PostScript
0.20625 0.062379763 0.39375 -0.26237976 MAarrowhead1
% End of user PostScript
0 0 1 r
.175 w
.78571 .57143 Mdot
gsave
.78571 .57143 -67.9688 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(u) show
1.000 setlinewidth
grestore
0 g
gsave
.5 .09524 -92.375 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(neutron) show
1.000 setlinewidth
grestore
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.678571 0.00921202 0.97619 0.377693 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 1.0582 0.740741 1.0582 [
[ 0 0 0 0 ]
[ 1 1.37566 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.37566 L
0 1.37566 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
newpath
matrix currentmatrix
0.47619 0.529101 scale
1.05 1.4 1 0 365.73 arc
setmatrix s
.01 w
.40079 .78055 m
.59921 1.12421 L
s
% Start of user PostScript
-0.09375 0.037620237 0.09375 0.36237976 MAarrowhead1
% End of user PostScript
1 0 0 r
.175 w
.5 .95238 Mdot
gsave
.5 .95238 -67.9688 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(u) show
1.000 setlinewidth
grestore
0 g
.01 w
.38757 .35727 m
.18915 .70093 L
s
% Start of user PostScript
-0.10625 -0.36237976 -0.29375 -0.037620237 MAarrowhead1
% End of user PostScript
0 1 0 r
.175 w
.28836 .5291 Mdot
gsave
.28836 .5291 -67.9688 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(u) show
1.000 setlinewidth
grestore
0 g
.01 w
.71825 .46309 m
.91667 .80675 L
s
% Start of user PostScript
0.20625 -0.26237976 0.39375 0.062379763 MAarrowhead1
% End of user PostScript
0 0 1 r
.175 w
.81746 .63492 Mdot
gsave
.81746 .63492 -67.9688 -12.5 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 25 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 17.125 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(d) show
1.000 setlinewidth
grestore
0 g
gsave
.5 .10582 -73.9375 -13.4688 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 26.9375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 19.063 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
%%BeginResource: font Times-BoldItalic-MISO
%%BeginFont: Times-BoldItalic-MISO
/Times-BoldItalic /Times-BoldItalic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-BoldItalic-MISO
%%IncludeFont: Times-BoldItalic-MISO
/Times-BoldItalic-MISO findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 19.063 moveto
%%IncludeResource: font Mathematica1-Bold
%%IncludeFont: Mathematica1-Bold
/Mathematica1-Bold findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(D) show
75.250 11.688 moveto
%%IncludeResource: font Mathematica1-Bold
%%IncludeFont: Mathematica1-Bold
/Mathematica1-Bold findfont 12.750 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(+) show
84.875 19.063 moveto
%%IncludeResource: font Times-BoldItalic-MISO
%%IncludeFont: Times-BoldItalic-MISO
/Times-BoldItalic-MISO findfont 18.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{409.438, 158.375},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40009V0000kB000`400?l00000o`00003oo`3ooooo
0?ooofP0oooo003o0?ooool0ooooJ03oool00?l0ooooo`3ooomX0?ooo`00o`3ooooo0?ooofP0oooo
003o0?ooool0ooooJ03oool00?l0ooooo`3ooomX0?ooo`00o`3ooooo0?ooofP0oooo003o0?ooool0
ooooJ03oool00?l0ooooo`3ooomX0?ooo`00o`3ooooo0?ooofP0oooo003o0?ooool0ooooJ03oool0
0?l0ooooo`3ooomX0?ooo`00o`3ooooo0?ooofP0oooo00130?ooo`030?noM0000000000000@00000
00<007Bo0?ooo`3oool0o`3ooooo0?oooa/0oooo00150?ooo`050;md0000000000000000001d_ol0
o`3ooooo0?oooah0oooo00150?ooo`050?oo_`1d000000000000B02Lgol0o`3ooooo0?oooah0oooo
00160?ooo`040;md00000000000007Boool0ooooX@3oool00`3ogi`0B0000000000>0000000307Bo
o`3oool0oooo06P0oooo00160?ooo`040?noM0000000000004RLgol0ooooXP3oool00`2oM0000000
0000000<000000030018W03Oool0oooo06T0oooo00160?ooo`050?oo_`1d000000000018W03Oool0
^`3oool01@3OW4P0000000000000B9`0gooo00<0oooo00D0gia800000000000004RL0=ooo`030?oo
o`030?oog`2LB000000000<0000000<0M;oo0?ooo`3oool00`3oool0103ogi`0B00000000000M;l2
0?ooo`050?oog`2LB000000000000000M;l00`3oool01P3ookl0M00000000000000004RL0=ooo`80
oooo00@0gia800000000000007Bo2@3oool00`3OW4P00000000000020000000307Boo`3oool0oooo
00@0oooo00D0gia800000000000004RL0=ooo`030?ooo`050=nLB000000000000018W03Oool0S@3o
ool00`3o_g@004RL0=ooo`080?ooo`050;md0000000000000000001d_ol0J`3oool004L0oooo00D0
_g@00000000004P0W;nL04P0000200000003001d_`3oool0oooo00@0oooo00@0gia8000000000000
07Bo2@3oool00`3OW4P00000000000020000000307Boo`3oool0oooo00@0oooo00H0oono07@00000
000000000018W03Oool40?ooo`030=nLB000000000000080000000<0M;oo0?ooo`3oool0103oool0
1@3OW4P0000000000000B9`0gooo00<0oooo00D0gia800000000000004RL0=ooo`1e0?ooo`050?oO
W018000000000000B02Lgol00P3oool02P3ooml0W4P000000000000004Q804P0M02oool0oooo0?oo
g`2LB006000000030000B02Lgol0oooo0080oooo00P0_g@000000000000000000000B02Lgol0oono
07@000<0000000@004RL0=ooo`3oool0okmd0P0000002P0007@0_mnL04P0M02oool0oooo0?oOW018
000000000000M02oool70?ooo`090;md00000000000007Boo`3oool0gia80000000004P0W=oo00@0
oooo00D0omnL04P000000000001809cOo`020?ooo`070?oog`2LB000000000000000B4P0B01d0;oo
o`2<0?ooo`030?oog`2LB000M;oo00L0oooo00H0omnL04P00000000000000018W03Ooom[0?ooo`00
A`3oool01`3OW4P000000000000007@0_ooo0?ooo`3o_g@00P0000000`1d_ol0oooo0?ooo`020?oo
o`050?oOW018000000000000M02oool01`3oool02@2oM00000000000001d_ol0oooo0=nLB0000000
001809cOo`040?ooo`070?noM000000000000000M02ogi`0B01d0;ooo`020?ooo`090;md00000000
000007Boo`3oool0gia80000000004P0W=oo00@0oooo00D0omnL04P000000000001809cOo`020?oo
o`070?oog`2LB000000000000000B4P0B01d0;ooo`1d0?ooo`050?oo_`1d0000000000000018W=l0
0`3oool02@2oM000000000000018W=l0oono07A8W03Oool0omnL04P0000200000007001d_`3oool0
oooo0?ooo`3ogi`0B00007Boo`020?ooo`080=nLB000000000000018W03Ooml0W4Q809cOg`2LB002
0000000A04RLg`3ogi`0B7Bo0?ooo`3ogi`0B0000000000004P0W=oo0?oo_`1dB9`0gooo0?oog`2L
B00000000000B02Lgol01P3oool01@3ogi`0B0000000000007@0_ooo0080oooo00@0_g@000000000
0000M;oo0`3oool01@3ookl0M000000000000000B9cO00<0oooo00L0_g@0000000000000B9cO0?oo
_`1dB9`0gooo08`0oooo00<0gia8001d_`3oool01P3oool01@2oM0000000000000000000M;oo06`0
oooo00170?ooo`050?oOW018000000000018W03Oool00P3oool0103o_g@0000000000018W=l30?oo
o`050?oog`2LB00000000000B02Lgol01P3oool01@3ogi`0B0000000000007@0_ooo0080oooo00@0
_g@0000000000000M;oo0`3oool03@3ogi`0B0000000000004P0W=oo0?oo_`1dB9`0gooo0?oOW018
000000000000M02oool00P3oool0102oM00000000000001d_ol30?ooo`050?oo_`1d000000000000
0018W=l00`3oool01`2oM000000000000018W=l0oono07A8W03Oool0M03oool0102oM00000000000
0000M;l30?ooo`050=nLB000000000000018W03Oool00P3oool01@3o_g@000000000000004P0W=oo
00P0oooo00P0omnL04P000000000001809cOo`3ooml0W7BL0;nLB080000000H007Bo0?ooo`3oool0
oooo0?oog`2LB0020000000304RLg`3oool0oooo0080oooo00@0gia8000000000000B9cO1P3oool0
1@3ogi`0B0000000000004P0W=oo0080oooo00D0oono07@000000000000007Boo`030?ooo`040;md
000000000000001d_`<0oooo00D0gia800000000000004RL0=ooo`2>0?ooo`030?oo_`1d04P0W=oo
00D0oooo00H0oono07@00000000000000000M02ooom/0?ooo`00A`3oool01@3ooml0W4P000000000
04P0W=oo00<0oooo00@0gia8000000000000M;oo0`3oool0103OW4P0000000000018W=l60?ooo`05
0?oOW018000000000000B02Lgol00P3oool01@3ookl0M000000000000000M;oo0080oooo00D0oooO
09a800000000000004RLg`030?ooo`050?oOW018000000000000B02Lgol00P3oool01@3ookl0M000
000000000000M;oo00<0oooo00@0_g@000000000000007Bo0`3oool01@3OW4P0000000000000B9`0
gooo07H0oooo00D0okmd000000000000001809cOo`020?ooo`050?oOW018000000000000B02Lgol0
0P3oool01@3ogi`0B0000000000004P0W=oo00P0oooo00`0oooO09a800000000000004RLg`3oool0
omnL07AdM018000000000000M02oool30?ooo`050;md0000000000000018W03Oool00`3oool01@3o
gi`0B0000000000007@0_ooo00D0oooo00D0omnL04P000000000000007Boo`030?ooo`050=nLB000
000000000000M02oool00P3oool01@3o_g@000000000000004P0W=oo0080oooo00D0omnL04P00000
0000001809cOo`2?0?ooo`030=nLB000M;l0oooo00@0oooo00D0gia8000000000000000004RLg`1]
0?ooo`00B03oool0103OW4P0000000000018W=l30?ooo`050?oo_`1d000000000000M02oool00P3o
ool01@3ogi`0B0000000000007@0_ooo00D0oooo00D0omnL04P000000000000007Boo`030?ooo`05
0=nLB000000000000000M02oool00P3oool01@2oM000000000000000B9`0gooo0080oooo00D0omnL
04P000000000000007Boo`030?ooo`050=nLB000000000000000M02oool00P3oool01@3o_g@00000
0000000004P0W=oo0080oooo00D0omnL04P000000000001809cOo`1f0?ooo`050?oo_`1d00000000
00000000M;l00P3oool01@3ooml0W4P0000000000000M;oo0080oooo00<0oono07@0000000001000
00000`1d_ol0oooo0?ooo`040?ooo`070=nLB000000000000018W03Oool0oooo0=nLB00300000005
07Boo`3oool0oooo0?ooo`3o_g@00P0000000`0004P0W=oo0?ooo`020?ooo`050?oog`2LB0000000
00000018W=l01@3oool01@3ookl0M00000000000000007Bo00<0oooo00D0oono07@0000000000000
04RLg`020?ooo`050?oo_`1d0000000000000000M;l00P3oool01@3ooml0W4P0000000000000M;oo
08l0oooo00<0omnL04P0M02oool00`3oool01P3ooml0W4P0000000000000001d0;ooofd0oooo0018
0?ooo`040?oOW01800000000001d_`@0oooo00@0_g@0000000000000B9cO0P3oool01@3ooml0W4P0
000000000000B9cO00D0oooo00D0oono07@0000000000000001d_`030?ooo`050?oo_`1d00000000
00000018W=l00P3oool01@3o_g@000000000000004P0W=oo0080oooo00D0oono07@0000000000000
001d_`030?ooo`050?oo_`1d0000000000000018W=l00P3oool01@3ookl0M00000000000000007Bo
0080oooo00D0oooO09a800000000000007Boo`1f0?ooo`030?oog`2LB00000000080000000@0M;oo
0?ooo`3oool0gia80P0000001@00M;l0oooo0?ooo`3oool0gia80080000000<007Bo0?ooo`2oM000
0P0000000`1d_ol0oooo0?ooo`020?ooo`080?oOW018000000000000M02oool0oooo0?oo_`1d0002
0000000604RLg`3oool0oooo0?ooo`3ookl0M0000P0000000`1d_ol0oooo0?ooo`020?ooo`050;md
0000000004Qd07A8M02oool01@3oool01@2oM00000000000000007@0_ooo00<0oooo00P0_g@00000
0000000004RL0=ooo`3oool0oooO09a800<0000000@0M;oo0?ooo`3oool0gia80P0000000`00M;l0
oooo0?ooo`2>0?ooo`030;md0018W=l0oooo0080oooo00D0okmd000000000000000004RLg`1^0?oo
o`00B03oool01@3ooml0W4P00000000007@0_ooo00<0oooo00D0okmd00000000000004RL0=ooo`02
0?ooo`050;md0000000004Qd07A8M02oool01@3oool01@2oM00000000000000007@0_ooo00<0oooo
00P0_g@000000000000004RL0=ooo`3oool0oono07@00080000000D0M;oo0?ooo`3oool0oooo0;md
0002000000030000M02oool0oooo0080oooo00P0_g@000000000000004RL0=ooo`3oool0oooO09a8
00<0000000@0M;oo0?ooo`3oool0gia80P0000000`00M;l0oooo0?ooo`1e0?ooo`0<0=nLB0000000
00000018W02LM7@0_ooo0?ooo`3ogi`0B0000000000007@0_ooo0P3oool01`3ookl0M00000000000
07@0_ooo0?ooo`3o_g@00P0000001P00M;l0oooo0?ooo`3oool0oooO09a80080000000@0M;oo0?oo
o`3oool0gia80P0000000`00B9`0gooo0?ooo`020?ooo`040;md000000000000001d_`@0oooo00D0
okmd0000000004P0W9ad04RLg`050?ooo`050?oOW018000000000000001d_ol00`3oool01@3OW4P0
00000000000007@0_ooo0080oooo00`0gia800000000000004RL09adM02oool0oooo0?oOW0180000
00000000M02ooon?0?ooo`060?oOW018B9`0gooo0?ooo`3ooml0W4P00P0000000`0007@0_ooo0?oo
o`1]0?ooo`00B@3oool0103OW4P000000000001d_ol30?ooo`050?oOW018000000000000M02oool0
0P3oool01@3o_g@000000000B02LW7@0B9cO00D0oooo00D0omnL04P000000000000007Boo`030?oo
o`050=nLB000000000000000M02oool00P3oool0102oM000000000000000M;l30?ooo`050?oOW018
000000000000001d_ol00`3oool01@3OW4P000000000000007@0_ooo0080oooo00`0gia800000000
000004RL09adM02oool0oooo0?oOW018000000000000M02ooomf0?ooo`090?oOW018000000000000
B02Lgkl0M4RL0=ooo`3ookl0M0000080000000D0M;oo0?ooo`3oool0oooo0?noM0020000000404RL
g`3oool0oooo0;md0080000000D0M;oo0?ooo`3oool0oooo0=nLB00200000008001d_`3oool0oooo
0?oo_`1d000000000000B02Lgol30?ooo`050?noM000000000000000M02oool00`3oool02@3ookl0
M00000000018W=l0_g@007Boo`3o_g@0000007Boo`020?ooo`050=nLB000000000000018W03Oool0
0P3oool01@3ogi`0B0000000000004P0W=oo0080oooo00T0omnL04P000000000001809cO_`1dB9`0
gooo0?oo_`1d00000P0000000`1d_ol0oooo0?ooo`2=0?ooo`060?oog`2LB000M;oo0?ooo`3ogi`0
B0000P0000000`00M;l0oooo0?ooo`1]0?ooo`00B@3oool0103ogi`0B00000000000M;l30?ooo`05
0?oOW018000000000000B02Lgol00P3oool02@3ookl0M00000000018W=l0_g@007Boo`3o_g@00000
07Boo`020?ooo`050=nLB000000000000018W03Oool00P3oool01@3ogi`0B0000000000004P0W=oo
0080oooo00D0okmd000000000000001d0;ooo`030?ooo`050=nLB000000000000018W03Oool00P3o
ool01@3ogi`0B0000000000004P0W=oo0080oooo00T0omnL04P000000000001809cO_`1dB9`0gooo
0?oo_`1d00000P0000000`1d_ol0oooo0?ooo`1d0?ooo`090?oo_`1d0000000000000018W=l0omnL
04Q8W03Oool0_g@00080000000<007Bo0?ooo`3oool00P3oool01P3o_g@000000000B02Lgol0oooo
0?noM080000000D007Bo0?ooo`3oool0oooo0?noM002000000050000M02oool0oooo0?oog`2LB000
0P0000001P1d_ol0oooo0?ooo`3oool0oono07@00080000000X0M;oo0?ooo`3oool0oooo0?oo_`1d
000000000018W03Ogi`0B0000P0000001@00B9`0gooo0?ooo`3oool0gia80080000000P0B9cO0?oo
o`3oool0omnL04P000000000001d0;ooo`80oooo00T0oono07@000000000000004RLg`3ogi`0B4RL
0=ooo`2oM0000P0000000`00M;l0oooo0?ooo`2>0?ooo`040?noM000B9`0gooo0;md0080000000<0
001809cOo`3oool02@3oool00`3o_g@0000004RLg`1R0?ooo`00B@3oool0203ogi`0B00000000000
000004RL0=ooo`3oool0gia80P0000002P0007@0_ooo0?ooo`3oool0oono07@00000000004RL0=oO
W0180002000000050018W03Oool0oooo0?ooo`3OW4P00P0000002018W=l0oooo0?ooo`3ogi`0B000
0000000007@0_ooo0P3oool01@3ookl0M000000000000000M;oo00@0oooo00@0gia8000000000000
B9cO0P3oool01@3ogi`0B0000000000007@0_ooo0080oooo00T0oono07@000000000000004RLg`3o
gi`0B4RL0=ooo`2oM0000P0000000`00M;l0oooo0?ooo`1c0?ooo`030?oOW0180000000000800000
00@007Bo0?ooo`3ogi`0B0000`0000000`00M;l0oooo0?ooo`030?ooo`0;0?oOW0180000001d0;oo
o`3ogi`0B00000000000B9`0gooo0?ooo`3OW4P010000000101d_ol0oooo0?ooo`3OW4P200000004
001d_`3oool0oooo0=nLB0H0000000<0B9cO0?noM00000000P0000000`0004P0W=oo0=nLB0020000
00030000M02oool0oooo0080oooo00H0okmd000000000000M;oo0?ooo`3o_g@20000000504RLg`3o
ool0oooo0?oOW01800000`0000001000M;l0oooo0?oOW018000300000003001d_`3oool0oooo08h0
oooo00@0oooO09a8B02Lgkl0M0000P0000000`00M;l0oooo0?ooo`090?ooo`030?noM0000000B9cO
0680oooo00180?ooo`070;md0000000000000000000004P0W;nL04P0000400000004001d_`3oool0
oooo0?noM0<0000000<0001809cOo`3OW4P00P0000000`0007@0_ooo0?ooo`020?ooo`060?noM000
0000000007Boo`3oool0okmd0P0000001018W=l0oooo0?ooo`3OW4P60000000504RLg`3oool0oooo
0?ooo`3o_g@00P0000000`1d_ol0oooo0?noM0020000000504RLg`3oool0oooo0?oOW01800000`00
00001000M;l0oooo0?oOW018000300000003001d_`3oool0oooo07D0oooo00/0oooO09a800000000
001d0;ooo`3oool0oooO09a800000000001809cOo`070?ooo`030?oOW018000000000080000000<0
M;oo0?ooo`3oool00P3oool0103ogi`0B00000000000M;l20?ooo`050?oOW018000000000000M02o
ool00P3oool00`3OW4P00000000000030000000<0018W03Oool0oooo0?oOW0180000000004RLg`3o
ool0gia800000000B9`0gooo1@3oool00`3OW4P00000000000020000000307Boo`3oool0oooo00<0
oooo00/0oooO09a800000000001d0;ooo`3oool0oooO09a800000000001809cOo`2A0?ooo`060=nL
B000000000000000000004P0W=oo2`3oool00`3o_g@0000004RLg`1R0?ooo`00BP3oool01P3o_g@0
000000000018W=l0oooo0;md0080000000<0001d0;ooo`3oool00`3oool02@3ogi`0B00000000018
W=l0oooo0=nLB000000004RL0=ooo`050?ooo`030=nLB000000000000080000000<0M;oo0?ooo`3o
ool00P3oool00`3OW4P0000000000003000000030018W03Oool0oooo00<0oooo00<0gia800000000
00000P0000000`1d_ol0oooo0?ooo`030?ooo`0;0?oog`2LB00000000000M02oool0oooo0?oog`2L
B00000000000B02Lgol0X@3oool01@3ookl0M0000000000004P0W=oo0;P0oooo00D0oono07@00000
00000000001d_`0<0?ooo`030?noM0000000B9cO0680oooo001a0?ooo`050?oo_`1d000000000000
B02Lgol0_P3oool0103ooml0W4P000000018W=ni0?ooo`040;md00000000000007Boo`L0oooo00<0
oooO09a800000000200000000`00B9`0gooo0?ooo`1L0?ooo`00LP3oool0103ooml0W4P000000018
W=no0?ooo`030?oog`2LB00007Bo0;T0oooo00@0omnL04P00000B9`0gooo1`3oool00`3ooml0W4P0
00000008000000030018W03Oool0oooo05`0oooo001c0?ooo`030?oog`2LB00007Bo0<00oooo00<0
oono07@0M02oool0^@3oool00`2oM000M;oo0?ooo`0<0?ooo`030?noM0000000B9cO0680oooo001d
0?ooo`030?oo_`1d07@0_ooo0?l0ooooR`3oool00`3o_g@0000004RLg`1R0?ooo`00o`3ooooo0?oo
o`<0oooo00<0okmd00000018W=l0HP3oool00?l0ooooo`3oool30?ooo`030?noM0000000B9cO0680
oooo003o0?ooool0ooooJ03oool00?l0ooooo`3ooomX0?ooo`00o`3ooooo0?ooofP0oooo001U0?oo
o`X00000_`3oool:00000;h0oooo2@00001W0?ooo`00G@3oool8000000X0oooo2@00002]0?ooo`T0
00002P3oool800000:l0oooo1`0000090?ooo`P00000G`3oool005L0oooo1P00000K0?ooo`H00000
X@3oool6000001/0oooo1P00002S0?ooo`H00000603oool6000005T0oooo001D0?ooo`<000009`3o
ool2000009d0oooo0P00000W0?ooo`<00000WP3oool2000002@0oooo00<000000?ooo`3oool0EP3o
ool00500oooo1000000/0?ooo`@00000U03oool5000002`0oooo1000002F0?ooo`@000009`3oool4
000005@0oooo001<0?ooo`@00000=03oool3000008d0oooo1000000e0?ooo`800000T@3oool30000
02l0oooo0P00001B0?ooo`00BP3oool2000003/0oooo100000270?ooo`800000>`3oool4000008/0
oooo0P00000d0?ooo`@00000CP3oool004P0oooo0P0000110?ooo`800000P`3oool200000440oooo
0P0000270?ooo`800000>P3oool00`000000oooo0?ooo`1;0?ooo`00AP3oool2000004D0oooo0P00
001o0?ooo`800000A@3oool2000008<0oooo0P00000m0?ooo`800000B`3oool004@0oooo0P000019
0?ooo`030000003oool0oooo07X0oooo0P0000190?ooo`800000O`3oool200000440oooo0P000019
0?ooo`00@P3oool2000004`0oooo0P00001i0?ooo`030000003oool0oooo04/0oooo0P00001l0?oo
o`030000003oool0oooo04<0oooo0P0000170?ooo`00@03oool200000500oooo0P00001e0?ooo`80
0000D03oool00`000000oooo0?ooo`1g0?ooo`800000B03oool00`000000oooo0?ooo`140?ooo`00
?`3oool00`000000oooo0?ooo`1B0?ooo`800000L@3oool2000005<0oooo0P00001e0?ooo`800000
B`3oool2000004@0oooo000m0?ooo`800000E`3oool2000006d0oooo0P00001G0?ooo`800000LP3o
ool00`000000oooo0?ooo`1=0?ooo`800000@P3oool003/0oooo0P00001K0?ooo`800000J@3oool2
000005/0oooo0P00001^0?ooo`800000DP3oool200000400oooo000i0?ooo`800000G`3oool20000
06D0oooo0P00001O0?ooo`800000JP3oool2000005H0oooo00<000000?ooo`3oool0?@3oool003P0
oooo00<000000?ooo`3oool0H03oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo0640
oooo00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo03`0
oooo000g0?ooo`030000003oool0oooo0680oooo00<000000?ooo`3oool0H03oool00`000000oooo
0?ooo`1S0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3oool0F03oool00`000000oooo
0?ooo`0k0?ooo`00=P3oool00`000000oooo0?ooo`1T0?ooo`800000G`3oool00`000000oooo0?oo
o`1U0?ooo`030000003oool0oooo0680oooo0P00001L0?ooo`030000003oool0oooo03X0oooo000d
0?ooo`800000J@3oool2000005/0oooo0P00001Y0?ooo`800000H@3oool00`000000oooo0?ooo`1M
0?ooo`800000>P3oool003<0oooo00<000000?ooo`3oool0J`3oool00`000000oooo0?ooo`1G0?oo
o`030000003oool0oooo06/0oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`1P0?oo
o`030000003oool0oooo03L0oooo000b0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3o
ool0E@3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3o
ool0HP3oool00`000000oooo0?ooo`0f0?ooo`00<@3oool00`000000oooo0?ooo`1_0?ooo`030000
003oool0oooo05<0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`1I0?ooo`030000
003oool0oooo06@0oooo00<000000?ooo`3oool0=@3oool002l0oooo0P00001c0?ooo`800000D@3o
ool2000007<0oooo0P00001H0?ooo`030000003oool0oooo06H0oooo00<000000?ooo`3oool0=03o
ool002h0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0
oooo04h0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0
oooo05@0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`150?ooo`030000003oool0
oooo03<0oooo000]0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0C@3oool00`00
0000oooo0?ooo`1=0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0C@3oool00`00
0000oooo0?ooo`1B0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0A`3oool00`00
0000oooo0?ooo`0b0?ooo`00;03oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo04h0
oooo0P00001;0?ooo`800000:@3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo04l0
oooo0P00000S0?ooo`030000003oool0oooo04P0oooo00<000000?ooo`3oool0<@3oool002/0oooo
00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo04P0oooo
00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo04d0oooo
00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo0340oooo
000Z0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?oo
o`170?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?oo
o`1<0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?oo
o`0`0?ooo`00:@3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo05<0oooo00<00000
0?ooo`3oool0A@3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo0580oooo00<00000
0?ooo`3oool0B`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo04`0oooo00<00000
0?ooo`3oool0;`3oool002T0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`1D0?oo
o`030000003oool0oooo04<0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`1D0?oo
o`030000003oool0oooo04X0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`1>0?oo
o`030000003oool0oooo02h0oooo000X0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3o
ool0EP3oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3o
ool0E@3oool00`000000oooo0?ooo`170?ooo`800000903oool00`000000oooo0?ooo`1?0?ooo`03
0000003oool0oooo02d0oooo000W0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0
F03oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0
E`3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0
D03oool00`000000oooo0?ooo`0]0?ooo`009P3oool00`000000oooo0?ooo`0Y0?ooo`030000003o
ool0oooo05T0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`0Z0?ooo`030000003o
ool0oooo05P0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`0S0?ooo`030000003o
ool0oooo0540oooo00<000000?ooo`3oool0;03oool002D0oooo00<000000?ooo`3oool0:@3oool0
0`000000oooo0?ooo`1J0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0:P3oool0
0`000000oooo0?ooo`1I0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool08`3oool0
0`000000oooo0?ooo`1B0?ooo`030000003oool0oooo02`0oooo000U0?ooo`030000003oool0oooo
02T0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo
02X0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo
02<0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`0[0?ooo`00903oool00`000000
oooo0?ooo`0Y0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool0>@3oool00`000000
oooo0?ooo`0Z0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3oool0@03oool00`000000
oooo0?ooo`0S0?ooo`030000003oool0oooo05D0oooo0P00000[0?ooo`008`3oool00`000000oooo
0?ooo`0Z0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool0>03oool00`000000oooo
0?ooo`0Z0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool0?P3oool00`000000oooo
0?ooo`0S0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0:@3oool00280oooo00<0
00000?ooo`3oool0803oool5003o00D0oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?oo
o`0f0?ooo`030000003oool0oooo0200oooo1@00o`060?ooo`030000003oool0oooo05l0oooo00<0
00000?ooo`3oool0?@3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo05P0oooo00<0
00000?ooo`3oool0:03oool00240oooo00<000000?ooo`3oool07@3oool=003o00030000003oool0
oooo0640oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0M0?ooo`d00?l000<0oooo
0000003oool0H@3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo01X0oooo1000o`05
0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3oool0:03oool00240oooo00<000000?oo
o`3oool0703oool?003o06@0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0L0?oo
o`l00?l000<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo
01H0oooo3000o`000`3oool000000?ooo`1K0?ooo`030000003oool0oooo02L0oooo000P0?ooo`03
0000003oool0oooo01/0oooo4`00o`1M0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0<P3oool00`000000oooo0?ooo`0K0?oooa<00?l0G@3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo03X0oooo00<000000?ooo`3oool05P3oool>003o03l0oooo00<000000?ooo`3o
ool06`3oool00`000000oooo0?ooo`0W0?ooo`00803oool00`000000oooo0?ooo`0J0?oooaD00?l0
F`3oool2000000D0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0J0?oooaD00?l0
F`3oool2000000D0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0D0?oooa800?l0
?P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo02H0oooo000O0?ooo`030000003o
ool0oooo01X0oooo5`00o`1I0?ooo`8000001P3oool00`000000oooo0?ooo`0`0?ooo`030000003o
ool0oooo01X0oooo5`00o`1I0?ooo`<000001@3oool00`000000oooo0?ooo`0h0?ooo`030000003o
ool0oooo01@0oooo5000o`0m0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool09P3o
ool001l0oooo00<000000?ooo`3oool06@3ooolI003o05L0oooo0`0000070?ooo`030000003oool0
oooo02l0oooo00<000000?ooo`3oool06@3ooolI003o05L0oooo0`0000070?ooo`030000003oool0
oooo03L0oooo00<000000?ooo`3oool04`3ooolF003o03d0oooo00<000000?ooo`3oool06`3oool0
0`000000oooo0?ooo`0U0?ooo`007P3oool00`000000oooo0?ooo`0J0?oooaT00?l0EP3oool40000
00P0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0J0?oooaT00?l0EP3oool40000
00P0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0D0?oooaH00?l0?@3oool00`00
0000oooo0?ooo`0L0?ooo`030000003oool0oooo02@0oooo000M0?ooo`030000003oool0oooo01X0
oooo6`00o`1D0?ooo`@000002@3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo01X0
oooo1`00o`001000g`00000000000000B002003o0004002L000000000000001800X00?l0E03oool5
000000P0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0D0?oooaP00?l0?@3oool0
0`000000oooo0?ooo`0K0?ooo`030000003oool0oooo02@0oooo000M0?ooo`030000003oool0oooo
01T0oooo2000o`001000g`00000000000000M003003o00040018000000000000001d00X00?l0D`3o
ool4000000X0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`0I0?ooo`P00?l01@00
00000`00o`0004P00000000300000003003O0000o`000?l000L00?l0D`3oool5000000T0oooo00<0
00000?ooo`3oool0<`3oool00`000000oooo0?ooo`0C0?ooo`H00?l000@00=l000000000000007@0
0`00o`001000B000000000000000M009003o03`0oooo00<000000?ooo`3oool0703oool00`000000
oooo0?ooo`0S0?ooo`00703oool00`000000oooo0?ooo`0J0?ooo`P00?l000<007@0000000000000
0P0000000`00g`000?l0000000030000000300180000o`000?l000L00?l0DP3oool5000000X0oooo
00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0J0?ooo`L00?l000P00=l0000000000000
0000003O0000o`0004P0002L0080000000<007@0003o0000B0002@00o`1B0?ooo`D000002P3oool0
0`000000oooo0?ooo`0b0?ooo`030000003oool0oooo01@0oooo1P00o`000`00M000000000000002
00000003003O0000o`00000000<0000000<004P0003o0000o`001P00o`0m0?ooo`030000003oool0
oooo01/0oooo00<000000?ooo`3oool08`3oool001`0oooo00<000000?ooo`3oool06P3oool8003o
0008002L00000000000000180000o`0004P0003O0000B00200000003002L0000g`0007@000P00?l0
D@3oool6000000/0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0J0?ooo`L00?l0
00@00=l000000000000000000P00o`001@00g`0007@000000000000004P000/00?l0D@3oool60000
00/0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0D0?ooo`H00?l000P009`00000
0000000004P0003o0000B0000=l000180080000000<009`0003O0000M0001`00o`0m0?ooo`030000
003oool0oooo01/0oooo00<000000?ooo`3oool08`3oool001/0oooo00<000000?ooo`3oool06`3o
ool8003o0008003O00000000000000000000g`000?l0001d0000W00200000003001d0000o`000?l0
00P00?l0D03oool6000000`0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0K0?oo
o`L00?l000D00=l00000000000000000003O0002003o0005002L00000000000000000000g`002P00
o`1@0?ooo`L000002`3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo01D0oooo1P00
o`002000g`0000000000000000000=l0003o0000M00009`00P0000000`00M0000?l0003o0007003o
03h0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0S0?ooo`006`3oool00`000000
oooo0?ooo`0J0?ooo`X00?l000L004P000000000000009`0003o0000g`0007@000<000002`00o`1>
0?ooo`L000003@3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo2@00o`00
1000B000000000000000W003003o00040018000000000000002L00/00?l0CP3oool8000000`0oooo
00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0D0?ooo`P00?l000L004P0000000000000
09`0003o0000g`0007@000<000002P00o`0n0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool08P3oool001X0oooo00<000000?ooo`3oool06`3oool:003o0004002L0000000000000018
00800?l000D009`00000000000000000002o000:003o04l0oooo1P00000=0?ooo`030000003oool0
oooo02H0oooo00<000000?ooo`3oool06`3oool9003o0004002L000000000000001800<00?l000@0
09`000000000000004P02`00o`1?0?ooo`H000003@3oool00`000000oooo0?ooo`0^0?ooo`030000
003oool0oooo01D0oooo2000o`001000W000000000000000B002003o0005002L0000000000000000
0000_`002@00o`0n0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool08P3oool001X0
oooo00<000000?ooo`3oool06`3oool:003o0004003O000000000000000000<00?l00`0000000`00
W0000?l0003o0008003o0500oooo1000000?0?ooo`030000003oool0oooo02D0oooo00<000000?oo
o`3oool06`3oool:003o00<0000000@00=l0003o0000o`000?l00`00000;003o0500oooo1@00000>
0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool05@3oool8003o0004003O00000000
0000000000<00?l00`0000000`00W0000?l0003o0007003o03l0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0Q0?ooo`006@3oool00`000000oooo0?ooo`0L0?ooo`/00?l000@004P0
0000000000000;l00P00o`001000W000000000000000B00:003o0500oooo00@000000?ooo`3oool0
00003`3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo01`0oooo2P00o`001000_`00
000000000000M003003o0004001d000000000000002o00X00?l0D03oool00`000000oooo0?ooo`02
000000h0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0E0?ooo`T00?l000@004P0
0000000000000;l00P00o`001000W000000000000000B009003o03l0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`0Q0?ooo`006@3oool00`000000oooo0?ooo`0L0?ooo`/00?l000@0
09`000000000000007@00`00o`0300000003003O0000o`000?l000L00?l0A03oool50000o`H0oooo
00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo01`0oooo
2`00o`001000W000000000000000_`02003o0004001d000000000000001d00X00?l0A@3oool50000
o`D0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo
01H0oooo2000o`001000W000000000000000M003003o00<0000000<00=l0003o0000o`001@00o`11
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0803oool001P0oooo00<000000?oo
o`3oool07P3oool:003o0004002o000000000000000000<00?l000@004P00000000000000;l02000
o`110?ooo`d0003o0P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo02<0oooo00<0
00000?ooo`3oool07@3oool;003o0006002L000000000000003O0000o`0004P00`000009003o0480
oooo3@000?l00`3oool000000?ooo`0C0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3o
ool05`3oool8003o0004002o000000000000000000<00?l000@004P00000000000000;l01`00o`11
0?ooo`030000003oool0oooo00<0oooo10000?lC0?ooo`030000003oool0oooo0200oooo000H0?oo
o`030000003oool0oooo01h0oooo2@00o`000`00W00000000000000200000004002o0000o`000?l0
002L0080000000<007@0003o0000o`001P00o`100?ooo`l0003o00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`0S0?ooo`030000003oool0oooo01d0oooo3@00o`0300000005002L00000000
000000000000g`002000o`110?ooo`l0003o5@3oool00`000000oooo0?ooo`0[0?ooo`030000003o
ool0oooo01L0oooo1`00o`000`00W00000000000000200000004002o0000o`000?l0002L00800000
00<007@0003o0000o`001@00o`120?ooo`03000000000?l0003o00X0003o3`3oool00`000000oooo
0?ooo`0P0?ooo`00603oool00`000000oooo0?ooo`0N0?ooo`/00?l000@00=l000000000000007@0
0P00o`001000g`000000000000000008003o03h0oooo4`000?lD0?ooo`030000003oool0oooo0280
oooo00<000000?ooo`3oool07P3ooolA003o0004001d000000000000002L00P00?l0?`3ooolC0000
oa<0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0H0?ooo`T00?l000@00=l00000
0000000007@00P00o`001000g`000000000000000007003o0480oooo3P000?l>0?ooo`030000003o
ool0oooo0200oooo000H0?ooo`030000003oool0oooo01h0oooo7@00o`0m0?oooaD0003o503oool0
0`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo01h0oooo4@00o`001000W000000000000000
B008003o03h0oooo5@000?lC0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool06@3o
oolH003o0440oooo4P000?l=0?ooo`030000003oool0oooo01l0oooo000G0?ooo`030000003oool0
oooo0200oooo6`00o`0m0?oooaL0003o4`3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0
oooo0200oooo4000o`001@00g`0000000000000000000=l000H00?l0?P3ooolG0000oa80oooo00<0
00000?ooo`3oool0:@3oool00`000000oooo0?ooo`0J0?oooaH00?l0@@3ooolD0000o``0oooo00<0
00000?ooo`3oool07`3oool001L0oooo00<000000?ooo`3oool08@3ooolI003o03d0oooo6@000?lC
0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool08@3oool@003o0004001d00000000
0000002o00D00?l0?P3ooolI0000oa80oooo00<000000?ooo`3oool0:03oool00`000000oooo0?oo
o`0J0?oooaH00?l0@03ooolF0000o``0oooo00<000000?ooo`3oool07P3oool001H0oooo00<00000
0?ooo`3oool08P3ooolI003o03d0oooo6@000?lC0?ooo`030000003oool0oooo01l0oooo00<00000
0?ooo`3oool08@3oool>003o0003003O0000000000000080000000<007@0003o0000o`000`00o`0n
0?oooaT0003o4P3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo01`0oooo5000o`11
0?oooaH0003o303oool00`000000oooo0?ooo`0N0?ooo`005P3oool00`000000oooo0?ooo`0S0?oo
oaL00?l0?@3ooolK0000oa80oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0R0?oo
oa000?l000<004P00000000000001000o`0n0?oooa/0003o4@3oool00`000000oooo0?ooo`0W0?oo
o`030000003oool0oooo01d0oooo4P00o`110?oooaP0003o2`3oool00`000000oooo0?ooo`0N0?oo
o`005P3oool00`000000oooo0?ooo`0T0?oooaD00?l0?@3oool80000o`040000W000000000000000
W080003o00@00018000000000000002L2`000?lA0?ooo`030000003oool0oooo01l0oooo00<00000
0?ooo`3oool08`3ooolE003o03h0oooo7@000?l@0?ooo`030000003oool0oooo02L0oooo00<00000
0?ooo`3oool07P3oool00`0000000?l0003o000<003o0480oooo1`000?l0100009`0000000000000
09`20000o`040000B000000000000000W0T0003o2P3oool00`000000oooo0?ooo`0N0?ooo`005P3o
ool00`000000oooo0?ooo`0U0?oooa<00?l0?P3oool70000o`030000_`00000000000080000000<0
001d0000g`0000000`0000000`0004P0003o0000o`080000oa80oooo00<000000?ooo`3oool07@3o
ool00`000000oooo0?ooo`0U0?oooa<00?l0?`3oool80000o`040000W000000000000000_`80003o
00D0003O00000000000000000000_`0:0000oa40oooo00<000000?ooo`3oool09P3oool00`000000
oooo0?ooo`0N0?ooo`030000003oool00?l000/00?l0@`3oool60000o`030000_`00000000000080
000000<0001d0000g`0000000`0000000`0004P0003o0000o`060000o`/0oooo00<000000?ooo`3o
ool07@3oool001H0oooo00<000000?ooo`3oool09P3oool00`0000000?l0003o000=003o0400oooo
1`000?l0200009`000000000000004P0003o0000_`0009`000180P0000000`000;l0003O0000M009
0000oa80oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0V0?ooo`0300000000o`00
0?l000d00?l0@@3oool80000o`@0000000<000180000o`000;l0100000000`0009`0003o0000o`07
0000oa40oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0
oooo00@0oooo1000o`170?ooo`H0003o00P0002L00000000000000180000o`000;l0002L0000B080
000000<0002o0000g`0007@01`000?l;0?ooo`030000003oool0oooo01d0oooo000E0?ooo`030000
003oool0oooo02H0oooo00<000000?ooo`3oool03@00o`110?ooo`L0003o00@0002L000000000000
001d0P000?l01@0009`0001d000000000000002L00/0003o4P3oool00`000000oooo0?ooo`0M0?oo
o`030000003oool0oooo02H0oooo00<000000?ooo`00o`003000o`120?ooo`P0003o00L000180000
00000000002L0000g`0004P0003O00<0000000<0003O0000W0000;l020000?lA0?ooo`030000003o
ool0oooo02D0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`1?0?ooo`H0003o00@0
002L000000000000001d0P000?l01@0009`0001d000000000000002L00T0003o2`3oool00`000000
oooo0?ooo`0M0?ooo`005@3oool00`000000oooo0?ooo`0S0?ooo`040000003oool0oooo000000H0
oooo1@00o`140?ooo`P0003o00@0002L00000000000000180`000?l0100004P000000000000004P<
0000oa40oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0R0?ooo`040000003oool0
oooo000000H0oooo1@00o`150?ooo`T0003o00P0002L00000000000000180000o`000=l0002L0000
B080000000<0002o0000o`000?l02@000?l@0?ooo`030000003oool0oooo02D0oooo00<000000?oo
o`3oool06P3oool010000000oooo0?ooo`00001A0?ooo`L0003o00@0002L00000000000000180`00
0?l0100004P000000000000004P:0000o`X0oooo00<000000?ooo`3oool07@3oool001D0oooo00<0
00000?ooo`3oool08`3oool300000500oooo20000?l01@000=l00000000000000000003O0080003o
00D0003O00000000000000000000g`0;0000oa40oooo00<000000?ooo`3oool0703oool00`000000
oooo0?ooo`0S0?ooo`<00000D@3oool90000o`080000g`0000000000000000000=l0003o0000W000
07@2000000030000M0000?l0003o00T0003o403oool00`000000oooo0?ooo`0U0?ooo`030000003o
ool0oooo01X0oooo0`00001B0?ooo`L0003o00D0003O00000000000000000000g`020000o`050000
g`0000000000000000000=l02@000?l:0?ooo`030000003oool0oooo01d0oooo000E0?ooo`030000
003oool0oooo0280oooo1P00001>0?ooo`T0003o00@00018000000000000002L0`000?l0100004P0
00000000000009`;0000oa80oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0R0?oo
o`H00000C`3oool:0000o`040000B000000000000000W080003o00@0001800000000000000002`00
0?lA0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool06@3oool600000500oooo2000
0?l0100004P000000000000009`30000o`040000B000000000000000W0T0003o2`3oool00`000000
oooo0?ooo`0L0?ooo`005@3oool00`000000oooo0?ooo`0R0?ooo`L00000C@3oool90000o`040000
_`00000000000000B0<0003o00@0002o000000000000001d2`000?lB0?ooo`030000003oool0oooo
01/0oooo00<000000?ooo`3oool08P3oool7000004h0oooo2P000?l0100009`000000000000007@2
0000o`050000_`0000000000000000000=l02P000?lA0?ooo`030000003oool0oooo02<0oooo00<0
00000?ooo`3oool06P3oool8000004h0oooo20000?l010000;l000000000000004P30000o`040000
_`00000000000000M0T0003o2`3oool00`000000oooo0?ooo`0L0?ooo`005@3oool00`000000oooo
0?ooo`0R0?ooo`H00000CP3oool:0000o`040000M000000000000000_`<0003o0`00000;0000oa80
oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0R0?ooo`H00000C`3oool:0000o`04
0000g`0000000000000000<0003o00@00018000000000000002L2P000?lA0?ooo`030000003oool0
oooo02<0oooo00<000000?ooo`3oool06P3oool700000500oooo20000?l0100007@0000000000000
0;l30000o`<0000020000?l<0?ooo`030000003oool0oooo01`0oooo000D0?ooo`030000003oool0
oooo0280oooo1P00001@0?ooo`X0003o00<000180000000000000`000?l3000000030000_`000?l0
003o00L0003o4`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0280oooo1P00001A
0?ooo`X0003o00@00018000000000000002o0P000?l010000;l000000000000004P90000oa80oooo
00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0I0?ooo`L00000D@3oool90000o`030000
B0000000000000<0003o0`0000000`000;l0003o0000o`050000o``0oooo00<000000?ooo`3oool0
703oool001@0oooo00<000000?ooo`3oool08P3oool500000540oooo2`000?l01@0004P000000000
B0000?l0003O00<0000000<0001d0000o`000?l01`000?lD0?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool08P3oool500000580oooo2P000?l0100007@000000000000007@20000o`04
0000g`0000000000000000T0003o4`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo
01T0oooo1P00001B0?ooo`X0003o00D000180000000004P0003o0000g`03000000030000M0000?l0
003o00D0003o303oool00`000000oooo0?ooo`0L0?ooo`00503oool00`000000oooo0?ooo`0R0?oo
o`@00000DP3oool<0000o`050000_`00000000000000000009`00P0000000`0004P0003o0000o`07
0000oa@0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0R0?ooo`@00000D`3oool9
0000o`030000B00000000000008000000`000?l0100004P00000000000000;l80000oa<0oooo00<0
00000?ooo`3oool08P3oool00`000000oooo0?ooo`0I0?ooo`D00000D`3oool;0000o`050000_`00
000000000000000009`00P0000000`0004P0003o0000o`050000o``0oooo00<000000?ooo`3oool0
703oool001@0oooo00<000000?ooo`3oool08@3oool4000005<0oooo4@000?l3000000030000g`00
0?l0003o00H0003o503oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0240oooo1000
001D0?ooo`/0003o00@0002L000000000000002o0P000?l0100009`000000000000007@80000oa<0
oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0H0?ooo`D00000E@3oool?0000o`<0
000000<0003O0000o`000?l00`000?l>0?ooo`030000003oool0oooo01/0oooo000D0?ooo`030000
003oool0oooo0240oooo0`00001E0?oooa00003o00@00018000000000000002L1`000?lE0?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool08@3oool3000005H0oooo6`000?lD0?ooo`03
0000003oool0oooo0280oooo00<000000?ooo`3oool0603oool3000005P0oooo3P000?l0100004P0
00000000000009`40000o`l0oooo00<000000?ooo`3oool06`3oool001@0oooo00<000000?ooo`3o
ool08@3oool2000005L0oooo3`000?l0100009`000000000000004P60000oaH0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`0Q0?ooo`800000F03ooolI0000oaD0oooo00<000000?oo
o`3oool08P3oool00`000000oooo0?ooo`0H0?ooo`800000F@3oool>0000o`040000W00000000000
0000B0@0003o3`3oool00`000000oooo0?ooo`0K0?ooo`004`3oool00`000000oooo0?ooo`0Q0?oo
o`800000F03oool@0000o`<000001P000?lF0?ooo`030000003oool0oooo01T0oooo00<000000?oo
o`3oool0803oool2000005T0oooo6@000?lE0?ooo`030000003oool0oooo0240oooo00<000000?oo
o`3oool0603oool2000005/0oooo3P000?l3000000<0003o403oool00`000000oooo0?ooo`0K0?oo
o`004`3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo05P0oooo3@000?l00`0009`0
000000000002000000030000_`000?l0003o0080003o5`3oool00`000000oooo0?ooo`0I0?ooo`03
0000003oool0oooo0200oooo00<000000?ooo`3oool0F@3ooolG0000oaH0oooo00<000000?ooo`3o
ool08@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo05/0oooo2`000?l00`0009`0
000000000002000000030000_`000?l0oooo0100oooo00<000000?ooo`3oool06`3oool001<0oooo
00<000000?ooo`3oool0O@3oool>0000o`040000g`00000000000000M0<0003o603oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo07d0oooo5@000?lG0?ooo`030000003oool0oooo0240
oooo00<000000?ooo`3oool0N03oool;0000o`050000g`00000000000000M02oool04@3oool00`00
0000oooo0?ooo`0K0?ooo`004`3oool00`000000oooo0?ooo`1n0?oooa<0003o6@3oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo07h0oooo4`000?lI0?ooo`030000003oool0oooo0200
oooo00<000000?ooo`3oool0N@3oool<0000o`030?ooo`000000oooo0140oooo00<000000?ooo`3o
ool06`3oool001<0oooo00<000000?ooo`3oool0P03oool?0000oa/0oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`1o0?ooo`03000000000?l0003o00d0003o6`3oool00`000000oooo
0?ooo`0P0?ooo`030000003oool0oooo07d0oooo10000?l60?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool06`3oool001<0oooo00<000000?ooo`3oool0P03oool00`000000003o0000
o`0;0000oa`0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`200?ooo`030000003o
ool0003o00`0003o703oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo08L0oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`0K0?ooo`004`3oool00`000000oooo0?ooo`1o
0?ooo`030000003oool0oooo00<0oooo1@000?lP0?ooo`030000003oool0oooo01P0oooo00<00000
0?ooo`3oool0O`3oool00`000000oooo0?ooo`040?ooo`D0003o803oool00`000000oooo0?ooo`0P
0?ooo`030000003oool0oooo08P0oooo00D000000?ooo`3oool0oooo0000000<0?ooo`030000003o
ool0oooo01/0oooo000C0?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool0:03oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool0:@3oool0
0`000000oooo0?ooo`0P0?ooo`030000003oool0oooo08P0oooo00<000000?ooo`0000000P00000<
0?ooo`030000003oool0oooo01/0oooo000C0?ooo`030000003oool0oooo07h0oooo00<000000?oo
o`3oool0:P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo07h0oooo00<000000?oo
o`3oool0:P3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo08P0oooo1@00000<0?oo
o`030000003oool0oooo01/0oooo000C0?ooo`030000003oool0oooo07h0oooo00<000000?ooo`3o
ool0:@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo07h0oooo00<000000?ooo`3o
ool0:P3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo08H0oooo2000000;0?ooo`03
0000003oool0oooo01/0oooo000C0?ooo`030000003oool0oooo07d0oooo00<000000?ooo`3oool0
:P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo07d0oooo00<000000?ooo`3oool0
:`3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo08L0oooo1`00000;0?ooo`030000
003oool0oooo01/0oooo000C0?ooo`030000003oool0oooo07`0oooo00<000000?ooo`3oool0:`3o
ool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo07`0oooo00<000000?ooo`3oool0;03o
ool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo08P0oooo1P00000;0?ooo`030000003o
ool0oooo01/0oooo000C0?ooo`030000003oool0oooo07`0oooo00<000000?ooo`3oool0:`3oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool0;03oool0
0`000000oooo0?ooo`0P0?ooo`030000003oool0oooo08T0oooo1@00000;0?ooo`030000003oool0
oooo01/0oooo000C0?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool0;03oool00`00
0000oooo0?ooo`0I0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0;03oool00`00
0000oooo0?ooo`0Q0?ooo`030000003oool0oooo08X0oooo1@00000:0?ooo`030000003oool0oooo
01/0oooo000C0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0=03oool00`000000
oooo0?ooo`0/0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0@`3oool00`000000
oooo0?ooo`0d0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool08@3oool00`000000
oooo0?ooo`0k0?ooo`030000003oool0oooo04`0oooo1@00000:0?ooo`030000003oool0oooo01/0
oooo000C0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0<P3oool00`000000oooo
0?ooo`0]0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0A03oool00`000000oooo
0?ooo`0b0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool08@3oool00`000000oooo
0?ooo`0l0?ooo`030000003oool0oooo04`0oooo1000000:0?ooo`030000003oool0oooo01/0oooo
000D0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?oo
o`0]0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?oo
o`0b0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?oo
o`0k0?ooo`030000003oool0oooo04d0oooo0`00000:0?ooo`030000003oool0oooo01/0oooo000D
0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0^
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0`
0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0l
0?ooo`030000003oool0oooo04d0oooo0`0000090?ooo`030000003oool0oooo01/0oooo000D0?oo
o`030000003oool0oooo04D0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0^0?oo
o`030000003oool0oooo01T0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0`0?oo
o`030000003oool0oooo02h0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0l0?oo
o`030000003oool0oooo04h0oooo0P0000090?ooo`030000003oool0oooo01/0oooo000D0?ooo`03
0000003oool0oooo04H0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0_0?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0^0?ooo`03
0000003oool0oooo02l0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0m0?ooo`03
0000003oool0oooo04h0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0K0?ooo`00
503oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo0600oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo0600oooo00<000000?ooo`3oool0
8P3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0
703oool001@0oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`1O0?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`1O0?ooo`030000003o
ool0oooo0280oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`1F0?ooo`030000003o
ool0oooo01`0oooo000D0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0GP3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0GP3oool0
0`000000oooo0?ooo`0S0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool0EP3oool0
0`000000oooo0?ooo`0L0?ooo`005@3oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo
05d0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo
05d0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo
05D0oooo00<000000?ooo`3oool0703oool001D0oooo00<000000?ooo`3oool0B03oool00`000000
oooo0?ooo`1L0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0B03oool00`000000
oooo0?ooo`1L0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0@03oool00`000000
oooo0?ooo`1D0?ooo`030000003oool0oooo01`0oooo000E0?ooo`030000003oool0oooo04P0oooo
00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo04P0oooo
00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo03l0oooo
00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`0L0?ooo`005@3oool00`000000oooo0?oo
o`190?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool0703oool00`000000oooo0?oo
o`190?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?oo
o`100?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool07@3oool001D0oooo00<00000
0?ooo`3oool0B@3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo01d0oooo00<00000
0?ooo`3oool0B03oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo02D0oooo00<00000
0?ooo`3oool0@03oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo01d0oooo000E0?oo
o`030000003oool0oooo04X0oooo00<000000?ooo`3oool00`3oool50?l00540oooo00<000000?oo
o`3oool07@3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo00<0oooo1@3o001A0?oo
o`030000003oool0oooo02D0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`1A0?oo
o`030000003oool0oooo01d0oooo000F0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3o
0000303o001=0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0B@3oool00`000000
oooo0?l0000<0?l004d0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`100?ooo`03
0000003oool0oooo00<0oooo103o001:0?ooo`030000003oool0oooo01d0oooo000F0?ooo`030000
003oool0oooo04X0oooo3`3o001;0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0
BP3oool?0?l004/0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`110?ooo`030000
003o0000o`0000X0o`00A@3oool00`000000oooo0?ooo`0N0?ooo`005P3oool00`000000oooo0?oo
o`180?oooa<0o`00B@3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo04L0oooo4`3o
00190?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0@@3oool>0?l004@0oooo00<0
00000?ooo`3oool07P3oool001H0oooo00<000000?ooo`3oool0A`3ooolE0?l004P0oooo00<00000
0?ooo`3oool07`3oool00`000000oooo0?ooo`160?oooaD0o`00B03oool00`000000oooo0?ooo`0W
0?ooo`030000003oool0oooo03l0oooo4P3o00120?ooo`030000003oool0oooo01h0oooo000G0?oo
o`030000003oool0oooo04D0oooo5`3o00170?ooo`030000003oool0oooo01l0oooo00<000000?oo
o`3oool0A@3ooolG0?l004L0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0m0?oo
oa@0o`00@@3oool00`000000oooo0?ooo`0N0?ooo`005`3oool00`000000oooo0?ooo`140?oooaT0
o`00AP3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo04@0oooo6@3o00160?ooo`03
0000003oool0oooo02P0oooo00<000000?ooo`3oool0?03ooolF0?l00400oooo00<000000?ooo`3o
ool07P3oool001P0oooo00<000000?ooo`3oool0@`3ooolI0?l004D0oooo00<000000?ooo`3oool0
8@3oool00`000000oooo0?ooo`130?oooaT0o`00A@3oool00`000000oooo0?ooo`0Y0?ooo`030000
003oool0oooo03`0oooo5P3o000o0?ooo`030000003oool0oooo01l0oooo000H0?ooo`030000003o
ool0oooo0480oooo6`3o00140?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0@P3o
oolK0?l004@0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0k0?oooaP0o`00?P3o
ool00`000000oooo0?ooo`0O0?ooo`00603oool00`000000oooo0?ooo`110?oooad0o`00@P3oool0
0`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0400oooo2@3o00001P18000000000000003O
0000o`000=l000<0000000<0g`000?l0003o0000203o00120?ooo`030000003oool0oooo02/0oooo
00<000000?ooo`3oool0>@3ooolJ0?l003`0oooo00<000000?ooo`3oool0803oool001P0oooo00<0
00000?ooo`3oool0@@3oool80?l0000304P000000000000000<0o`0000@0W0000000000000000000
2`3o00120?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0@03oool80?l0000307@0
0000000000000080000000<0_`0009`0000000000`0000000`2L0000o`000?l000070?l00480oooo
00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`0i0?ooo`P0o`0000<0B000000000000000
0`3o0000102L000000000000000000080?l003`0oooo00<000000?ooo`3oool0803oool001P0oooo
00<000000?ooo`3oool0@@3oool70?l000030;l00000000000000080000000<0W0000?l0001d0000
100000000`3O0000o`000?l000070?l00480oooo00<000000?ooo`3oool08`3oool00`000000oooo
0?ooo`100?ooo`P0o`0000L0B000000000000000W0000?l0001d0000_`0000<0000000<0o`0009`0
002o0000203o00120?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0>@3oool70?l0
00030;l00000000000000080000000<0W0000?l0001d0000100000000`3O0000o`000?l000040?l0
03`0oooo00<000000?ooo`3oool0803oool001T0oooo00<000000?ooo`3oool0@03oool70?l00005
0=l00000000000000000003O00000`2L00020000000304P0003o0000B00000T0o`00@P3oool00`00
0000oooo0?ooo`0S0?ooo`030000003oool0oooo0400oooo203o00001018000000000000002o0002
0?l0000507@00018000000000000003O00002P3o00120?ooo`030000003oool0oooo02`0oooo00<0
00000?ooo`3oool0>03oool70?l000050=l00000000000000000003O00000`2L00020000000304P0
003o0000B00000H0o`00?03oool00`000000oooo0?ooo`0P0?ooo`006@3oool00`000000oooo0?oo
o`0o0?ooo`T0o`0000L0B000000000000000W0000?l0002L0000_`0000<00000303o00100?ooo`03
0000003oool0oooo02@0oooo00<000000?ooo`3oool0?`3oool90?l0000404P000000000000009`0
0080o`0000D0g`00000000000000000009`0000;0?l00400oooo00<000000?ooo`3oool0;@3oool0
0`000000oooo0?ooo`0g0?ooo`T0o`0000L0B000000000000000W0000?l0002L0000_`0000<00000
2@3o000j0?ooo`030000003oool0oooo0240oooo000J0?ooo`030000003oool0oooo03h0oooo2@3o
0000102L000000000000001800020?l0000507@00018000000000000002o00002`3o00100?ooo`03
0000003oool0oooo02D0oooo00<000000?ooo`3oool0?P3oool90?l0000409`000000000000004P0
00<0o`0000@0W000000000000000B0002`3o00100?ooo`030000003oool0oooo02d0oooo00<00000
0?ooo`3oool0=`3oool90?l0000409`000000000000004P00080o`0000D0M00004P0000000000000
0;l000080?l003X0oooo00<000000?ooo`3oool08@3oool001X0oooo00<000000?ooo`3oool0?P3o
ool90?l000070=l00000000000000000003O0000o`000=l000030000000307@0003o0000o`0000T0
o`00?`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo03h0oooo2@3o00001@3O0000
0000000000000000g`000080o`0000D0g`0000000000000000000=l0000:0?l003l0oooo00<00000
0?ooo`3oool0;P3oool00`000000oooo0?ooo`0g0?ooo`T0o`0000L0g`0000000000000000000=l0
003o0000g`0000<0000000<0M0000?l0003o00001P3o000i0?ooo`030000003oool0oooo0280oooo
000K0?ooo`030000003oool0oooo03d0oooo2P3o00001018000000000000002o00020?l0000407@0
00000000000004P000/0o`00?`3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo03d0
oooo2P3o0000101d000000000000002L00030?l0000407@00000000000000;l000X0o`00?`3oool0
0`000000oooo0?ooo`0_0?ooo`030000003oool0oooo03H0oooo2P3o00001018000000000000002o
00020?l0000407@000000000000004P000P0o`00>@3oool00`000000oooo0?ooo`0R0?ooo`00703o
ool00`000000oooo0?ooo`0l0?ooo`X0o`0000@0W000000000000000M0000P3o00001@3O00000000
000000000000g`0000X0o`00?P3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo03`0
oooo2`3o0003000000<0o`0000@0_`00000000000000M0002P3o000n0?ooo`030000003oool0oooo
0340oooo00<000000?ooo`3oool0=P3oool90?l0000409`000000000000007@00080o`0000D0g`00
00000000000000000=l000060?l003T0oooo00<000000?ooo`3oool08`3oool001`0oooo00<00000
0?ooo`3oool0?@3oool90?l000040=l0000000000000000000<0o`0000@0M000000000000000W000
2@3o000o0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0?@3oool:0?l000040=l0
00000000000007@00080o`0000@0_`0000000000000000002@3o000o0?ooo`030000003oool0oooo
0340oooo00<000000?ooo`3oool0=P3oool90?l000040=l0000000000000000000<0o`0000@0M000
000000000000W0001P3o000i0?ooo`030000003oool0oooo02<0oooo000M0?ooo`030000003oool0
oooo03`0oooo2P3o0003000000040;l0003o0000o`0009`00080000000<0M0000?l0003o00001`3o
000n0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0?03oool;0?l000060=l00000
0000000009`0003o0000W0000`0000000`2o0000o`000?l000060?l003h0oooo00<000000?ooo`3o
ool0<`3oool00`000000oooo0?ooo`0e0?ooo`X0o`000`000000102o0000o`000?l0002L00020000
000307@0003o0000o`0000@0o`00>@3oool00`000000oooo0?ooo`0S0?ooo`007@3oool00`000000
oooo0?ooo`0l0?ooo`P0o`0000<0g`000000000000000P000000101d0000o`000?l0003O00030000
00T0o`00?@3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo03`0oooo3@3o00001@1d
00000000000000180000M0000080000000<0W0000?l0003o00001P3o000m0?ooo`030000003oool0
oooo03@0oooo00<000000?ooo`3oool0=@3oool80?l000030=l00000000000000080000000@0M000
0?l0003o0000g`000`0000060?l003P0oooo00<000000?ooo`3oool0903oool001h0oooo00<00000
0?ooo`3oool0>`3oool;0?l0000304P000000000000000<0o`0000@0B000000000000000_`00203o
000m0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0>`3ooolA0?l000040;l00000
0000000004P000P0o`00?@3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo03D0oooo
2P3o00000`1800000000000000030?l0000404P00000000000000;l000@0o`00>@3oool00`000000
oooo0?ooo`0T0?ooo`007P3oool00`000000oooo0?ooo`0l0?oooa/0o`00?@3oool00`000000oooo
0?ooo`0^0?ooo`030000003oool0oooo03`0oooo403o00001@3O00000000000000000000g`0000H0
o`00?@3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo03H0oooo5P3o000i0?ooo`03
0000003oool0oooo02D0oooo000O0?ooo`030000003oool0oooo03`0oooo6@3o000n0?ooo`030000
003oool0oooo02l0oooo00<000000?ooo`3oool0?03oool@0?l0000404P000000000000009`000D0
o`00?P3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo03D0oooo5P3o000i0?ooo`03
0000003oool0oooo02D0oooo000O0?ooo`030000003oool0oooo03`0oooo6@3o000m0?ooo`030000
003oool0oooo0300oooo00<000000?ooo`3oool0?03oool@0?l000040;l000000000000007@000D0
o`00?@3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo03H0oooo503o000i0?ooo`03
0000003oool0oooo02H0oooo000P0?ooo`030000003oool0oooo03`0oooo5`3o000n0?ooo`030000
003oool0oooo0340oooo00<000000?ooo`3oool0?03oool>0?l0000304P000000000000000800000
103o000n0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0=P3ooolB0?l003X0oooo
00<000000?ooo`3oool09P3oool00200oooo00<000000?ooo`3oool0?@3ooolE0?l003h0oooo00<0
00000?ooo`3oool0<P3oool00`000000oooo0?ooo`0m0?ooo`l0o`0000@0W000000000000000_`00
0P3o000n0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0>03oool>0?l003/0oooo
00<000000?ooo`3oool09`3oool00240oooo00<000000?ooo`3oool0?@3ooolC0?l003l0oooo00<0
00000?ooo`3oool0<`3oool00`000000oooo0?ooo`0m0?oooa<0o`00?`3oool00`000000oooo0?oo
o`0k0?ooo`030000003oool0oooo03P0oooo303o00000`3oool000000?ooo`0i0?ooo`030000003o
ool0oooo02L0oooo000Q0?ooo`030000003oool0oooo03l0oooo3`3o00000`000000oooo0?ooo`0m
0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0?`3oool?0?l000030000003oool0
oooo03d0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`0l0?ooo`@0o`001P3oool0
0`000000oooo0?ooo`0f0?ooo`030000003oool0oooo02P0oooo000R0?ooo`030000003oool0oooo
03l0oooo3@3o00000`3oool000000?ooo`0m0?ooo`030000003oool0oooo03H0oooo00<000000?oo
o`3oool0?`3oool=0?l000030?ooo`000000oooo03h0oooo00<000000?ooo`3oool0?@3oool00`00
0000oooo0?ooo`150?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0:03oool002<0
oooo0P0000130?ooo`D0o`001P3oool010000000oooo0?ooo`00000i0?ooo`030000003oool0oooo
03P0oooo0P0000130?ooo`D0o`001P3oool01@000000oooo0?ooo`3oool0000003P0oooo0P000011
0?ooo`030000003oool0oooo04D0oooo00D000000?ooo`3oool0oooo0000000a0?ooo`800000:`3o
ool002@0oooo00<000000?ooo`3oool0C03oool4000003P0oooo00<000000?ooo`3oool0>P3oool0
0`000000oooo0?ooo`1<0?ooo`030000003oool0000000800000=`3oool00`000000oooo0?ooo`11
0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`0000000P00000`0?ooo`030000003oool0
oooo02/0oooo000U0?ooo`030000003oool0oooo04X0oooo1P00000g0?ooo`030000003oool0oooo
03/0oooo00<000000?ooo`3oool0B`3oool5000003L0oooo00<000000?ooo`3oool0@P3oool00`00
0000oooo0?ooo`140?ooo`D00000<03oool00`000000oooo0?ooo`0[0?ooo`009@3oool00`000000
oooo0?ooo`190?ooo`L00000=P3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo04T0
oooo2000000e0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0@P3oool8000002h0
oooo00<000000?ooo`3oool0;03oool002H0oooo00<000000?ooo`3oool0B@3oool6000003H0oooo
00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`190?ooo`L00000=@3oool00`000000oooo
0?ooo`140?ooo`030000003oool0oooo0480oooo1`00000^0?ooo`030000003oool0oooo02`0oooo
000W0?ooo`030000003oool0oooo04T0oooo1P00000d0?ooo`030000003oool0oooo03l0oooo00<0
00000?ooo`3oool0B@3oool6000003@0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?oo
o`130?ooo`H00000;@3oool00`000000oooo0?ooo`0]0?ooo`00:03oool00`000000oooo0?ooo`19
0?ooo`D00000<`3oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo04T0oooo1@00000c
0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0@`3oool5000002d0oooo00<00000
0?ooo`3oool0;@3oool002T0oooo00<000000?ooo`3oool0B@3oool400000380oooo00<000000?oo
o`3oool0@`3oool00`000000oooo0?ooo`190?ooo`D00000<@3oool00`000000oooo0?ooo`190?oo
o`800000A03oool5000002/0oooo00<000000?ooo`3oool0;P3oool002T0oooo00<000000?ooo`3o
ool0B@3oool400000340oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`180?ooo`D0
0000<03oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo0480oooo1@00000Z0?ooo`03
0000003oool0oooo02l0oooo000Z0?ooo`030000003oool0oooo04T0oooo1000000_0?ooo`030000
003oool0oooo04L0oooo00<000000?ooo`3oool0B03oool400000300oooo00<000000?ooo`3oool0
C03oool00`000000oooo0?ooo`120?ooo`@00000:@3oool00`000000oooo0?ooo`0`0?ooo`00:`3o
ool2000004X0oooo0`00000^0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0B03o
ool3000002h0oooo0P00001@0?ooo`030000003oool0oooo0480oooo0`00000X0?ooo`030000003o
ool0oooo0340oooo000/0?ooo`030000003oool0oooo04T0oooo0P00000]0?ooo`030000003oool0
oooo04/0oooo00<000000?ooo`3oool0B03oool3000002`0oooo00<000000?ooo`3oool0D@3oool0
0`000000oooo0?ooo`120?ooo`<000009P3oool00`000000oooo0?ooo`0b0?ooo`00;@3oool00`00
0000oooo0?ooo`190?ooo`800000:`3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo
04P0oooo0P00000/0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0@P3oool20000
02H0oooo00<000000?ooo`3oool0<P3oool002h0oooo0P00001:0?ooo`030000003oool0oooo02P0
oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo02P0
oooo0P00001F0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool08`3oool00`000000
oooo0?ooo`0c0?ooo`00<03oool00`000000oooo0?ooo`1a0?ooo`030000003oool0oooo0540oooo
00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo06H0oooo
00<000000?ooo`3oool0=03oool00340oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?oo
o`1C0?ooo`030000003oool0oooo06l0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?oo
o`1T0?ooo`030000003oool0oooo03D0oooo000b0?ooo`030000003oool0oooo06d0oooo00<00000
0?ooo`3oool0E@3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo05/0oooo00<00000
0?ooo`3oool0HP3oool00`000000oooo0?ooo`0f0?ooo`00<`3oool00`000000oooo0?ooo`1[0?oo
o`030000003oool0oooo05L0oooo00<000000?ooo`3oool0J`3oool00`000000oooo0?ooo`1M0?oo
o`030000003oool0oooo0600oooo00<000000?ooo`3oool0=`3oool003@0oooo00<000000?ooo`3o
ool0J@3oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo06T0oooo00<000000?ooo`3o
ool0G`3oool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo03P0oooo000e0?ooo`800000
I`3oool2000005d0oooo0P00001W0?ooo`800000H`3oool00`000000oooo0?ooo`1K0?ooo`800000
>`3oool003L0oooo00<000000?ooo`3oool0HP3oool200000640oooo00<000000?ooo`3oool0H`3o
ool00`000000oooo0?ooo`1T0?ooo`800000FP3oool00`000000oooo0?ooo`0k0?ooo`00>03oool2
00000640oooo00<000000?ooo`3oool0HP3oool200000640oooo0P00001Y0?ooo`030000003oool0
oooo05D0oooo0P00000n0?ooo`00>P3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo
06D0oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`1Z0?ooo`030000003oool0oooo
05<0oooo00<000000?ooo`3oool0?P3oool003/0oooo00<000000?ooo`3oool0F`3oool00`000000
oooo0?ooo`1W0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0K03oool00`000000
oooo0?ooo`1A0?ooo`030000003oool0oooo03l0oooo000l0?ooo`800000F@3oool2000006/0oooo
0P00001I0?ooo`800000L03oool200000500oooo00<000000?ooo`3oool0@03oool003h0oooo0P00
001E0?ooo`800000K`3oool2000005D0oooo0P00001d0?ooo`030000003oool0oooo04/0oooo0P00
00130?ooo`00@03oool00`000000oooo0?ooo`1@0?ooo`800000L`3oool200000540oooo0P00001g
0?ooo`800000B@3oool2000004D0oooo00110?ooo`800000CP3oool2000007L0oooo0P00001>0?oo
o`030000003oool0oooo07T0oooo0P0000160?ooo`030000003oool0oooo04D0oooo00130?ooo`80
0000BP3oool2000007/0oooo00<000000?ooo`3oool0B@3oool2000007h0oooo00<000000?ooo`3o
ool0@@3oool2000004P0oooo00150?ooo`800000A`3oool00`000000oooo0?ooo`1l0?ooo`800000
A`3oool200000840oooo0P00000o0?ooo`800000BP3oool004L0oooo0P0000130?ooo`800000P@3o
ool2000004<0oooo0P0000250?ooo`800000>`3oool2000004`0oooo00190?ooo`@00000>P3oool5
000008D0oooo1000000k0?ooo`@00000R@3oool4000003@0oooo0`00001>0?ooo`00C@3oool20000
03H0oooo0P00002>0?ooo`800000=`3oool200000940oooo00<000000?ooo`3oool0;`3oool20000
0540oooo001?0?ooo`@00000;P3oool400000980oooo1@00000]0?ooo`D00000U03oool4000002T0
oooo1000001C0?ooo`00D`3oool5000002D0oooo1000002K0?ooo`@000009@3oool4000009d0oooo
1000000R0?ooo`<00000E`3oool005P0oooo1000000M0?ooo`@00000X`3oool4000001d0oooo1000
002U0?ooo`@000006P3oool4000005X0oooo001L0?ooo`/000001`3oool;00000:/0oooo2`000007
0?ooo`/00000[@3oool9000000L0oooo2P00001N0?ooo`00I`3oool700000<40oooo1`0000310?oo
o`L00000J03oool00?l0ooooo`3ooomX0?ooo`00o`3ooooo0?ooofP0oooo003o0?ooool0ooooJ03o
ool00?l0ooooo`3ooomX0?ooo`00o`3ooooo0?ooofP0oooo003o0?ooool0ooooJ03oool00?l0oooo
o`3ooomX0?ooo`00o`3ooooo0?ooofP0oooo003o0?ooool0ooooJ03oool00?l0ooooo`3ooomX0?oo
o`00o`3ooooo0?ooofP0oooo003o0?ooool0ooooJ03oool00?l0ooooo`3ooomX0?ooo`00o`3ooooo
0?ooofP0oooo003o0?ooool0ooooJ03oool00?l0ooooo`3ooomX0?ooo`00o`3ooooo0?ooofP0oooo
003o0?ooool0ooooJ03oool00?l0ooooo`3ooomX0?ooo`00o`3ooooo0?ooofP0oooo003o0?ooool0
ooooJ03oool00?l0ooooo`3ooomX0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 613.125}, {236.563, 0}} -> {-0.0793804, -0.029024, \
0.00772786, 0.00772786}, {{15.375, 197.375}, {230.875, 5.625}} -> {-0.613937, \
-0.732832, 0.00865717, 0.00865717}, {{215.563, 397.5}, {230.875, 5.625}} -> \
{-1.76913, -0.732832, 0.00865717, 0.00865717}, {{415.688, 597.688}, {230.875, \
5.625}} -> {-2.92431, -0.732832, 0.00865713, 0.00865713}}]
}, Closed]],

Cell[TextData[{
  "Of course, this figure could also be created, perhaps more easily, using \
an independent drawing program and then imported into a ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " notebook."
}], "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Exercises", "ExerciseMain"],

Cell[CellGroupData[{

Cell["Fourier-sine approximation to square wave", "ExerciseTitle"],

Cell[TextData[{
  "A square wave with unit amplitude and period ",
  Cell[BoxData[
      \(TraditionalForm\`2  \[Pi]\)]],
  " is defined by ",
  Cell[BoxData[
      \(TraditionalForm\`f[x + m\ \[Pi]] = f[x]\)]],
  " for any integer ",
  Cell[BoxData[
      \(TraditionalForm\`m\)]],
  StyleBox[" ",
    FontSlant->"Italic"],
  "and ",
  Cell[BoxData[
      \(TraditionalForm\`f[x] = Sign[x]\)]],
  " for ",
  Cell[BoxData[
      \(TraditionalForm\`\[LeftBracketingBar]x\[RightBracketingBar] < \
\[Pi]\)]],
  ".  This square wave can be approximated by a Fourier-sine series of the \
form"
}], "ExerciseText"],

Cell[BoxData[
    \(TraditionalForm\`\(f\&~\)\_n[x] = \ 
      4\/\[Pi]\ \ \(\[Sum]\+\(k = 1\)\%n Sin[\((2  k - 1)\) x]\/\(2  k - \
1\)\)\)], "DisplayFormula",
  TextAlignment->Left,
  TextJustification->0],

Cell[TextData[{
  "where ",
  StyleBox["n",
    FontSlant->"Italic"],
  " is the order of approximation.  Produce a 2\[Times]3 graphics array which \
compares the first 6 approximations with the square-wave itself."
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Antenna patterns", "ExerciseTitle"],

Cell[TextData[{
  "The angular distribution  of intensity produced by an array of ",
  Cell[BoxData[
      \(TraditionalForm\`n\)]],
  " identical equally-spaced antennas which broadcast in phase at wave length \
",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]\)]],
  " has the form"
}], "ExerciseText"],

Cell[BoxData[
    \(TraditionalForm\`f[n, d, \[Lambda], \[Theta]] = 
      Sin[n\ \[Pi]\ \(d\/\[Lambda]\) Sin[\[Theta]]]\^2\/Sin[\[Pi]\ \(d\/\
\[Lambda]\) Sin[\[Theta]]]\^2\)], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->-0.5],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`d\)]],
  " is the spacing and ",
  Cell[BoxData[
      \(TraditionalForm\`\[Theta]\)]],
  " the polar angle.  Use ",
  StyleBox["PolarPlot", "InlineInput",
    FontWeight->"Bold"],
  " to display antenna patterns for ",
  Cell[BoxData[
      \(TraditionalForm\`n = 4\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`d/\[Lambda] = {1, 4}\)]],
  "."
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Multiple-slit diffraction", "ExerciseTitle"],

Cell[TextData[{
  "The illumination of a distant screen by light that passes through ",
  Cell[BoxData[
      \(TraditionalForm\`n\)]],
  " slits of width ",
  Cell[BoxData[
      \(TraditionalForm\`a\)]],
  " and spacing ",
  Cell[BoxData[
      \(TraditionalForm\`d\)]],
  " is described by a  diffraction pattern of the form"
}], "ExerciseText"],

Cell[BoxData[
    \(TraditionalForm\`I[\[Theta]]\  = \ 
      I\_0\ f\_n[d, \[Lambda], \[Theta]]\ g[
          a, \[Lambda], \[Theta]]\)], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->-0.5],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`I\_0\)]],
  " is the peak intensity,"
}], "ExerciseText"],

Cell[BoxData[
    \(TraditionalForm\`f\_n[d, \[Lambda], \[Theta]] = 
      Sin[n\ \[Pi]\ \(d\/\[Lambda]\) Sin[\[Theta]]]\^2\/Sin[\[Pi]\ \(d\/\
\[Lambda]\) Sin[\[Theta]]]\^2\)], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->-0.5],

Cell[TextData[{
  "is the ",
  StyleBox["n",
    FontSlant->"Italic"],
  "-beam interference pattern, and "
}], "ExerciseText",
  CellDingbat->None],

Cell[BoxData[
    \(TraditionalForm\`g[a, \[Lambda], \[Theta]] = 
      Sin[\[Pi]\ \(a\/\[Lambda]\) Sin[\[Theta]]]\^2\/\((\[Pi]\ \(a\/\[Lambda]\
\) Sin[\[Theta]])\)\^2\)], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->-0.5],

Cell["is the single-slit diffraction pattern.  ", "ExerciseText",
  CellDingbat->None],

Cell[TextData[{
  "a) Produce a function which displays ",
  Cell[BoxData[
      \(TraditionalForm\`f\)]],
  ", ",
  Cell[BoxData[
      \(TraditionalForm\`g\)]],
  ", and ",
  Cell[BoxData[
      \(TraditionalForm\`I[\[Theta]]\)]],
  " in column form.  The arguments of your function should be based upon ",
  Cell[BoxData[
      \(TraditionalForm\`d/\[Lambda]\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`a/\[Lambda]\)]],
  ".  The horizontal axis should be ",
  Cell[BoxData[
      \(TraditionalForm\`Sin[\[Theta]]\)]],
  " in units of ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]/d\)]],
  ".  Design your function according to the following specifications."
}], "ExerciseText",
  CellDingbat->None],

Cell["\<\
Your function should accept arguments either in conventional units or in \
dimensionless form.  Use patterns with predicates and/or defaults as \
needed.\
\>", "ExerciseText",
  CellDingbat->"\[FilledSmallCircle]"],

Cell[TextData[{
  "Your function should accept options and you should establish default \
options for ",
  StyleBox["PlotRange", "InlineInput",
    FontWeight->"Bold"],
  ", ",
  StyleBox["AspectRatio", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[", ",
    FontWeight->"Bold"],
  StyleBox["PlotPoints", "InlineInput",
    FontWeight->"Bold"],
  ", etc.  "
}], "ExerciseText",
  CellDingbat->"\[FilledSmallCircle]"],

Cell["\<\
The entire figure should be enclosed in a frame, with frame labels for the \
axes and a descriptive plot label.  Each plot within the frame should be \
labeled appropriately also.  \
\>", "ExerciseText",
  CellDingbat->"\[FilledSmallCircle]"],

Cell["\<\
Provide versions which produce either linear or semilog plots.  (What is the \
advantage of semilog plots here?)  \
\>", "ExerciseText",
  CellDingbat->"\[FilledSmallCircle]"],

Cell[TextData[{
  "b) Write another function which produces a figure in which the single-slit \
diffraction pattern is shown as a dashed line while the total pattern is \
shown as a solid line.  In this representation the single-slit pattern should \
look like an envelope modulating the ",
  Cell[BoxData[
      \(TraditionalForm\`n\)]],
  "-beam interference pattern. "
}], "ExerciseText"],

Cell[TextData[{
  "c) Display figures for several interesting choices of ",
  Cell[BoxData[
      \(TraditionalForm\`n\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`a/d\)]],
  ".  For example, show both linear and semilog figures for ",
  Cell[BoxData[
      \(TraditionalForm\`n = 4\)]],
  ", ",
  Cell[BoxData[
      \(TraditionalForm\`a/d = 0.5\)]],
  ". You should be able to find similar figures in practically any elementary \
physics textbook, but your goal is to write concise functions which produce \
figures for arbitrary parameter sets. "
}], "ExerciseText"],

Cell["\<\
Here is an example of a column figure with linear vertical scale\
\>", "ExerciseText"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .8 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0190476 0.952381 [
[ 0 0 -0.125 0 ]
[.5 -0.0125 -23.75 -27.375 ]
[.5 -0.0125 23.75 0 ]
[ 0 0 -0.125 0 ]
[-0.01875 .4 -15.6875 -11.0938 ]
[-0.01875 .4 0 11.0938 ]
[ 0 .8 .125 0 ]
[ 1 0 .125 0 ]
[.5 .8125 -82.8125 0 ]
[.5 .8125 82.8125 15.6875 ]
[ 0 0 0 0 ]
[ 1 .8 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -84.75 -31.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 35.375 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 31.375000 moveto
 460.000000 31.375000 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 31.375000 lineto
closepath
clip newpath
63.000000 21.625000 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
65.937500 13.375000 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(d) show
%%IncludeResource: font Math1
%%IncludeFont: Math1
/Math1 findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
64.625000 21.625000 moveto
(\\200) show
66.125000 21.625000 moveto
(\\200) show
67.625000 21.625000 moveto
(\\200) show
69.125000 21.625000 moveto
(\\200) show
70.625000 21.625000 moveto
(\\200) show
72.125000 21.625000 moveto
(\\200) show
72.500000 21.625000 moveto
(\\200) show
66.062500 30.312500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(l) show
75.625000 21.625000 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Sin) show
%%IncludeResource: font Math2
%%IncludeFont: Math2
/Math2 findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
90.937500 21.625000 moveto
(@) show
94.875000 21.625000 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(q) show
%%IncludeResource: font Math2
%%IncludeFont: Math2
/Math2 findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
102.562500 21.625000 moveto
(D) show
106.500000 21.625000 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math1
%%+ font Times-Italic
%%+ font Math1-Italic
%%+ font Math2
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math1
%%+ Times-Italic
%%+ Math1-Italic
%%+ Math2
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math1
%%+ font Times-Italic
%%+ font Math1-Italic
%%+ font Math2
grestore
0 0 m
0 .8 L
s
gsave
-0.01875 .4 -79.9375 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
72.0938 11.8438 translate
90 rotate
-72.0938 -11.8438 translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.187500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(I) show
%%IncludeResource: font Math2
%%IncludeFont: Math2
/Math2 findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
68.187500 15.812500 moveto
(\\220) show
72.187500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(I) show
76.125000 17.750000 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 8.500000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0) show
81.187500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%+ font Math2
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%+ Math2
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
%%+ font Math2
grestore
0 .8 m
1 .8 L
s
1 0 m
1 .8 L
s
gsave
.5 .8125 -143.813 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.062500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Multiple) show
103.500000 15.812500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(-) show
111.625000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(slit) show
129.562500 15.812500 moveto
(Diffraction) show
188.750000 15.812500 moveto
(Pattern) show
224.625000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%+ font Math1-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%+ Math1-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
%%+ font Math1-Italic
grestore
0 0 m
1 0 L
1 .8 L
0 .8 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0190476 0.97619 0.257143 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.119048 0.00595238 0.238095 [
[.02381 -0.00655 -9.34375 -15.6875 ]
[.02381 -0.00655 9.34375 0 ]
[.2619 -0.00655 -9.34375 -15.6875 ]
[.2619 -0.00655 9.34375 0 ]
[.7381 -0.00655 -5 -15.6875 ]
[.7381 -0.00655 5 0 ]
[.97619 -0.00655 -5 -15.6875 ]
[.97619 -0.00655 5 0 ]
[.5 .2625 -60.3438 0 ]
[.5 .2625 60.3438 15.6875 ]
[ 0 0 0 0 ]
[ 1 .25 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
.02381 .00595 m
.02381 .0122 L
s
gsave
.02381 -0.00655 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(-) show
71.687500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%+ font Math1-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%+ Math1-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
%%+ font Math1-Italic
grestore
.2619 .00595 m
.2619 .0122 L
s
gsave
.2619 -0.00655 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(-) show
71.687500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%+ font Math1-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%+ Math1-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
%%+ font Math1-Italic
grestore
.7381 .00595 m
.7381 .0122 L
s
gsave
.7381 -0.00655 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.97619 .00595 m
.97619 .0122 L
s
gsave
.97619 -0.00655 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.125 Mabswid
.08333 .00595 m
.08333 .0097 L
s
.14286 .00595 m
.14286 .0097 L
s
.20238 .00595 m
.20238 .0097 L
s
.32143 .00595 m
.32143 .0097 L
s
.38095 .00595 m
.38095 .0097 L
s
.44048 .00595 m
.44048 .0097 L
s
.55952 .00595 m
.55952 .0097 L
s
.61905 .00595 m
.61905 .0097 L
s
.67857 .00595 m
.67857 .0097 L
s
.79762 .00595 m
.79762 .0097 L
s
.85714 .00595 m
.85714 .0097 L
s
.91667 .00595 m
.91667 .0097 L
s
.25 Mabswid
0 .00595 m
1 .00595 L
s
.5 0 m
.5 .25 L
s
gsave
.5 .2625 -121.344 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4) show
69.000000 15.812500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(-) show
77.125000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(slit) show
95.000000 15.812500 moveto
(diffaction) show
145.125000 15.812500 moveto
(pattern) show
179.687500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%+ font Math1-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%+ Math1-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
%%+ font Math1-Italic
grestore
.5 Mabswid
.02381 .00595 m
.02395 .00595 L
.02408 .00595 L
.02423 .00595 L
.02437 .00595 L
.02462 .00595 L
.02489 .00595 L
.02519 .00595 L
.02551 .00596 L
.02608 .00596 L
.02722 .00596 L
.02847 .00597 L
.03092 .006 L
.03355 .00603 L
.03485 .00604 L
.03609 .00605 L
.03663 .00605 L
.0372 .00605 L
.03753 .00605 L
.03783 .00606 L
.03813 .00606 L
.03841 .00606 L
.03855 .00606 L
.0387 .00606 L
.03884 .00606 L
.03897 .00606 L
.03912 .00606 L
.03921 .00606 L
.03929 .00606 L
.03943 .00606 L
.03959 .00606 L
.03973 .00606 L
.03989 .00606 L
.04018 .00606 L
.04052 .00605 L
.04083 .00605 L
.04135 .00605 L
.04192 .00605 L
.04306 .00604 L
.04542 .00602 L
.04795 .00599 L
.04919 .00598 L
.05052 .00596 L
.05114 .00596 L
.05179 .00596 L
.05234 .00595 L
.05266 .00595 L
.05296 .00595 L
.05311 .00595 L
.05319 .00595 L
Mistroke
.05327 .00595 L
.05342 .00595 L
.05356 .00595 L
.05364 .00595 L
.05372 .00595 L
.05387 .00595 L
.05395 .00595 L
.05403 .00595 L
.05421 .00595 L
.05437 .00595 L
.05451 .00595 L
.05479 .00595 L
.0554 .00596 L
.05596 .00596 L
.05647 .00596 L
.05764 .00598 L
.05989 .00601 L
.06238 .00605 L
.06491 .00609 L
.06602 .0061 L
.06661 .00611 L
.06723 .00611 L
.06785 .00612 L
.06815 .00612 L
.06842 .00612 L
.06869 .00612 L
.06883 .00612 L
.06898 .00612 L
.06911 .00612 L
.06923 .00612 L
.06937 .00612 L
.06951 .00612 L
.06959 .00612 L
.06967 .00612 L
.06982 .00612 L
.06998 .00612 L
.07015 .00612 L
.07047 .00612 L
.07076 .00612 L
.07131 .00611 L
.07191 .00611 L
.07306 .0061 L
.07432 .00608 L
.07686 .00603 L
.07818 .00601 L
.07942 .00599 L
.08063 .00597 L
.08117 .00596 L
.08175 .00596 L
.08202 .00596 L
Mistroke
.08232 .00596 L
.08259 .00595 L
.08272 .00595 L
.08285 .00595 L
.083 .00595 L
.08313 .00595 L
.08322 .00595 L
.0833 .00595 L
.08344 .00595 L
.08358 .00595 L
.08372 .00595 L
.08386 .00595 L
.08401 .00595 L
.0843 .00596 L
.08461 .00596 L
.08518 .00596 L
.08585 .00597 L
.08646 .00598 L
.08765 .00601 L
.08876 .00604 L
.09127 .00613 L
.09371 .00623 L
.096 .00632 L
.09707 .00635 L
.0982 .00638 L
.09878 .00639 L
.09912 .00639 L
.09943 .0064 L
.09958 .0064 L
.09971 .0064 L
.09984 .0064 L
.09998 .0064 L
.10012 .0064 L
.10028 .0064 L
.10041 .0064 L
.10056 .0064 L
.10071 .0064 L
.10086 .0064 L
.10102 .0064 L
.10119 .0064 L
.1015 .00639 L
.10179 .00639 L
.10236 .00638 L
.10291 .00637 L
.10412 .00633 L
.10543 .00627 L
.10779 .00615 L
.10904 .00608 L
.11038 .00602 L
.11104 .00599 L
Mistroke
.11167 .00597 L
.11195 .00597 L
.11224 .00596 L
.11239 .00596 L
.11256 .00596 L
.11271 .00595 L
.11285 .00595 L
.11299 .00595 L
.11315 .00595 L
.11329 .00595 L
.11343 .00595 L
.11356 .00595 L
.11368 .00596 L
.11396 .00596 L
.11411 .00596 L
.11425 .00597 L
.11457 .00598 L
.11515 .006 L
.11572 .00604 L
.11632 .00609 L
.11742 .00621 L
.11872 .00642 L
.11991 .00668 L
.12221 .00734 L
.12472 .00834 L
.12963 .01096 L
.13437 .01375 L
.13674 .01497 L
.13892 .01586 L
.14005 .01621 L
.14125 .01648 L
.14187 .01658 L
.14217 .01661 L
.14245 .01664 L
.14261 .01666 L
.14278 .01667 L
.14293 .01668 L
.14309 .01669 L
.14326 .01669 L
.14335 .01669 L
.14344 .0167 L
.1436 .0167 L
.14368 .0167 L
.14377 .0167 L
.14392 .01669 L
.14407 .01669 L
.1442 .01668 L
.14434 .01668 L
.14466 .01666 L
.14495 .01663 L
Mistroke
.14549 .01656 L
.14606 .01647 L
.14725 .01619 L
.14856 .01577 L
.15092 .01475 L
.15344 .01337 L
.15843 .01033 L
.16093 .00892 L
.16324 .00782 L
.16549 .00698 L
.16662 .00666 L
.16787 .00638 L
.16846 .00627 L
.16909 .00618 L
.16963 .00611 L
.17021 .00605 L
.17053 .00603 L
.17083 .00601 L
.17113 .00599 L
.17141 .00598 L
.17173 .00596 L
.1719 .00596 L
.17209 .00596 L
.17224 .00595 L
.17239 .00595 L
.17247 .00595 L
.17255 .00595 L
.17264 .00595 L
.17272 .00595 L
.17286 .00595 L
.17301 .00595 L
.17315 .00596 L
.17328 .00596 L
.17357 .00597 L
.17388 .00597 L
.17447 .006 L
.17512 .00603 L
.17734 .0062 L
.1799 .00641 L
.18111 .0065 L
.18223 .00657 L
.18339 .00663 L
.184 .00665 L
.18431 .00666 L
.18465 .00667 L
.18494 .00667 L
.18521 .00668 L
.18534 .00668 L
.18548 .00668 L
.18561 .00668 L
Mistroke
.18573 .00668 L
.18589 .00668 L
.18604 .00668 L
.18617 .00668 L
.18631 .00668 L
.18647 .00668 L
.18663 .00668 L
.18693 .00667 L
.18724 .00667 L
.18753 .00666 L
.18807 .00665 L
.1893 .0066 L
.19036 .00654 L
.19151 .00647 L
.19385 .00631 L
.19643 .00614 L
.19761 .00608 L
.19887 .00602 L
.19946 .006 L
.20007 .00598 L
.20059 .00597 L
.20089 .00596 L
.20117 .00596 L
.20147 .00596 L
.20163 .00596 L
.2018 .00595 L
.20195 .00595 L
.20209 .00595 L
.20224 .00595 L
.20233 .00595 L
.20241 .00595 L
.20255 .00595 L
.20268 .00595 L
.20283 .00595 L
.20298 .00595 L
.20327 .00596 L
.20353 .00596 L
.20416 .00597 L
.20483 .00598 L
.20606 .00602 L
.20842 .00611 L
.21101 .00621 L
.2122 .00626 L
.21348 .0063 L
.2141 .00631 L
.21468 .00632 L
.21521 .00633 L
.21551 .00633 L
.21578 .00634 L
.21606 .00634 L
Mistroke
.21622 .00634 L
.21637 .00634 L
.2165 .00634 L
.21663 .00634 L
.21677 .00634 L
.21692 .00634 L
.217 .00634 L
.21708 .00634 L
.21717 .00634 L
.21726 .00634 L
.21741 .00634 L
.21758 .00634 L
.21788 .00633 L
.21817 .00633 L
.21881 .00632 L
.21939 .00631 L
.22071 .00628 L
.22301 .0062 L
.22551 .0061 L
.22785 .00602 L
.22849 .006 L
.22918 .00599 L
.22977 .00597 L
.23009 .00597 L
.23043 .00596 L
.23075 .00596 L
.23105 .00596 L
.23119 .00596 L
.23135 .00595 L
.23149 .00595 L
.23163 .00595 L
.23178 .00595 L
.23186 .00595 L
.23195 .00595 L
.23204 .00595 L
.23212 .00595 L
.2322 .00595 L
.23229 .00595 L
.23244 .00595 L
.23252 .00595 L
.2326 .00595 L
.23275 .00595 L
.23289 .00595 L
.23317 .00596 L
.23346 .00596 L
.23398 .00597 L
.23454 .00597 L
.23516 .00599 L
.23751 .00605 L
.24004 .00614 L
Mistroke
.24243 .00622 L
.24378 .00626 L
.24442 .00627 L
.24503 .00628 L
.24556 .00629 L
.24584 .00629 L
.24614 .0063 L
.24645 .0063 L
.24662 .0063 L
.24678 .0063 L
.24692 .0063 L
.24706 .0063 L
.24721 .0063 L
.24736 .0063 L
.2475 .0063 L
.24766 .0063 L
.24783 .0063 L
.24798 .0063 L
.24825 .00629 L
.24855 .00629 L
.24922 .00628 L
.24984 .00627 L
.25103 .00625 L
.25213 .00621 L
.25464 .00613 L
.25691 .00604 L
.2581 .00601 L
.25936 .00598 L
.26001 .00597 L
.26035 .00596 L
.26071 .00596 L
.26089 .00596 L
.26105 .00596 L
.26121 .00595 L
.26136 .00595 L
.2615 .00595 L
.26164 .00595 L
.26179 .00595 L
.26187 .00595 L
.26195 .00595 L
.2621 .00595 L
.26219 .00595 L
.26227 .00595 L
.26241 .00595 L
.26257 .00595 L
.26286 .00596 L
.26313 .00596 L
.26378 .00597 L
.26439 .00598 L
.26557 .00601 L
Mistroke
.26681 .00605 L
.26904 .00614 L
.27143 .00625 L
.27362 .00633 L
.27419 .00635 L
.27482 .00637 L
.27537 .00638 L
.27597 .00639 L
.2763 .00639 L
.2766 .0064 L
.27677 .0064 L
.27693 .0064 L
.27702 .0064 L
.2771 .0064 L
.27719 .0064 L
.27728 .0064 L
.27744 .0064 L
.27759 .0064 L
.27772 .0064 L
.27786 .0064 L
.27802 .0064 L
.27819 .0064 L
.27849 .00639 L
.2788 .00639 L
.27909 .00639 L
.27963 .00638 L
.28085 .00634 L
.28191 .00631 L
.28306 .00626 L
.28564 .00613 L
.28797 .00603 L
.28914 .00599 L
.2898 .00597 L
.29011 .00597 L
.2904 .00596 L
.29068 .00596 L
.29093 .00596 L
.29108 .00595 L
.29122 .00595 L
.29136 .00595 L
.29152 .00595 L
.29166 .00595 L
.29181 .00595 L
.29189 .00595 L
.29198 .00595 L
.29213 .00595 L
.29227 .00595 L
.29242 .00596 L
.29269 .00596 L
.293 .00596 L
Mistroke
.2933 .00597 L
.29385 .00598 L
.29444 .006 L
.29508 .00603 L
.29729 .00616 L
.29965 .00634 L
.30223 .00655 L
.30342 .00663 L
.30469 .00671 L
.30531 .00674 L
.3059 .00676 L
.30642 .00678 L
.30672 .00678 L
.30699 .00679 L
.30713 .00679 L
.30728 .00679 L
.30741 .00679 L
.30755 .00679 L
.3077 .00679 L
.30778 .00679 L
.30786 .00679 L
.30801 .00679 L
.30815 .00679 L
.30828 .00679 L
.30842 .00679 L
.30867 .00679 L
.30894 .00678 L
.30924 .00677 L
.30986 .00675 L
.31052 .00673 L
.31172 .00665 L
.31427 .00644 L
.3167 .00621 L
.31785 .00611 L
.31849 .00606 L
.31909 .00603 L
.31965 .006 L
.31994 .00598 L
.32025 .00597 L
.32051 .00596 L
.32065 .00596 L
.3208 .00596 L
.32094 .00596 L
.32106 .00595 L
.32119 .00595 L
.32131 .00595 L
.32144 .00595 L
.32159 .00595 L
.32173 .00595 L
.32186 .00596 L
Mistroke
.322 .00596 L
.32216 .00596 L
.32247 .00597 L
.32278 .00598 L
.32311 .006 L
.32371 .00604 L
.32431 .00609 L
.32487 .00615 L
.32593 .00629 L
.32835 .00674 L
.33057 .00727 L
.33303 .00789 L
.33423 .00817 L
.33532 .00838 L
.33587 .00847 L
.33646 .00856 L
.33702 .00862 L
.33753 .00866 L
.33784 .00868 L
.33798 .00869 L
.33812 .00869 L
.33828 .0087 L
.33843 .0087 L
.33859 .0087 L
.33867 .0087 L
.33876 .0087 L
.33891 .0087 L
.33905 .0087 L
.33917 .0087 L
.33931 .00869 L
.33946 .00869 L
.33961 .00868 L
.3399 .00866 L
.3402 .00863 L
.34049 .0086 L
.34103 .00853 L
.34225 .00831 L
.34331 .00805 L
.34446 .0077 L
.34703 .00683 L
.34824 .00645 L
.34936 .00616 L
.34992 .00606 L
.35021 .00602 L
.35052 .00598 L
.35069 .00597 L
.35084 .00596 L
.35092 .00596 L
.35101 .00595 L
.3511 .00595 L
Mistroke
.35118 .00595 L
.35132 .00595 L
.35147 .00596 L
.35155 .00596 L
.35163 .00597 L
.35179 .00598 L
.35193 .00599 L
.35206 .00601 L
.35237 .00606 L
.35266 .00613 L
.35298 .00622 L
.35355 .00644 L
.35407 .0067 L
.35466 .00707 L
.35528 .00757 L
.35639 .00873 L
.35758 .01041 L
.35889 .01283 L
.36124 .01873 L
.36376 .02737 L
.36873 .05037 L
.37353 .07553 L
.37595 .08696 L
.37814 .0954 L
.37936 .09905 L
.38051 .10167 L
.38115 .10275 L
.38173 .1035 L
.38203 .10381 L
.38237 .10406 L
.38254 .10416 L
.38272 .10425 L
.3828 .10428 L
.38289 .10431 L
.38304 .10434 L
.38318 .10436 L
.38333 .10436 L
.38347 .10435 L
.3836 .10433 L
.38374 .10429 L
.3839 .10423 L
.38406 .10415 L
.38421 .10406 L
.38452 .10382 L
.38481 .10353 L
.38545 .10269 L
.386 .10174 L
.38659 .10048 L
.38767 .09759 L
.39007 .08862 L
Mistroke
.39261 .0761 L
.39718 .04981 L
.39973 .03565 L
.40205 .02446 L
.4032 .01977 L
.40446 .01541 L
.40563 .01207 L
.40674 .00962 L
.40787 .00777 L
.40847 .00707 L
.40878 .00677 L
.40912 .0065 L
.40943 .00631 L
.40976 .00614 L
.40995 .00608 L
.41011 .00603 L
.41019 .00601 L
.41028 .00599 L
.41044 .00597 L
.41058 .00596 L
.41067 .00595 L
.41074 .00595 L
.4109 .00596 L
.41099 .00597 L
.41107 .00598 L
.41121 .006 L
.41137 .00604 L
.41165 .00612 L
.41196 .00625 L
.41229 .00642 L
.4129 .00683 L
.41402 .00785 L
.41655 .01096 L
.41891 .01412 L
.42016 .01562 L
.4215 .01695 L
.42216 .01747 L
.42279 .01787 L
.42335 .01815 L
.42367 .01828 L
.42396 .01837 L
.42409 .0184 L
.42424 .01843 L
.42436 .01845 L
.42449 .01847 L
.42464 .01849 L
.4248 .0185 L
.42494 .01851 L
.42508 .0185 L
.42523 .0185 L
Mistroke
.42538 .01849 L
.42551 .01847 L
.42565 .01845 L
.42597 .01838 L
.42626 .0183 L
.42683 .01807 L
.42743 .01775 L
.42852 .01696 L
.42983 .01572 L
.43101 .01439 L
.4335 .01129 L
.43581 .0086 L
.43696 .00751 L
.43804 .00672 L
.43864 .00639 L
.43896 .00626 L
.43929 .00614 L
.43957 .00606 L
.43972 .00603 L
.43988 .006 L
.44003 .00598 L
.44016 .00597 L
.4403 .00596 L
.44042 .00595 L
.44056 .00595 L
.44071 .00596 L
.44085 .00597 L
.44098 .00599 L
.44112 .00601 L
.44128 .00604 L
.44159 .00612 L
.4419 .00623 L
.44223 .00637 L
.44283 .0067 L
.44399 .00759 L
.44505 .00868 L
.44988 .01553 L
.45209 .01863 L
.45335 .02003 L
.45397 .02059 L
.45453 .02101 L
.45481 .02119 L
.45512 .02135 L
.45539 .02147 L
.45568 .02158 L
.456 .02167 L
.45617 .0217 L
.45625 .02171 L
.45633 .02172 L
.45648 .02173 L
Mistroke
.45662 .02174 L
.45677 .02174 L
.45685 .02173 L
.45693 .02173 L
.45709 .02171 L
.45718 .02169 L
.45726 .02168 L
.45757 .0216 L
.45785 .0215 L
.45816 .02136 L
.45878 .02098 L
.45947 .02042 L
.46069 .01906 L
.46181 .01748 L
.46433 .01316 L
.46553 .01101 L
.46681 .00891 L
.46744 .00802 L
.46802 .0073 L
.46855 .00677 L
.46884 .00652 L
.46912 .00632 L
.46939 .00617 L
.46953 .0061 L
.46968 .00605 L
.46984 .006 L
.46999 .00597 L
.47014 .00596 L
.47027 .00595 L
.4704 .00596 L
.47052 .00598 L
.47065 .00601 L
.47079 .00605 L
.47093 .00611 L
.47107 .00618 L
.47134 .00635 L
.47162 .0066 L
.47193 .00694 L
.47248 .00773 L
.47313 .009 L
.47373 .0105 L
.47493 .01456 L
.47605 .0196 L
.47855 .03578 L
.48082 .05601 L
.48331 .0836 L
.48818 .14684 L
.49286 .20402 L
.49409 .21595 L
.49542 .22686 L
Mistroke
.49668 .23488 L
.49728 .2379 L
.49784 .24015 L
.49846 .24205 L
.49878 .2428 L
.49895 .24311 L
.49912 .2434 L
.49928 .24361 L
.49942 .24376 L
.49956 .24389 L
.4997 .24397 L
.49985 .24403 L
.5 .24405 L
.50016 .24403 L
.50024 .244 L
.50033 .24396 L
.50047 .24386 L
.50062 .24373 L
.50088 .24339 L
.50117 .24289 L
.50148 .2422 L
.50205 .24053 L
.50257 .23853 L
.50385 .23184 L
.50505 .22333 L
.50729 .20243 L
.51194 .14531 L
.51688 .08132 L
.51934 .05439 L
.52164 .03427 L
.52385 .02013 L
.52509 .01445 L
.52622 .01063 L
.5268 .00915 L
.52742 .0079 L
.52794 .0071 L
.52822 .00676 L
.52852 .00647 L
.52868 .00635 L
.52885 .00623 L
.529 .00614 L
.52914 .00608 L
.52931 .00602 L
.5294 .00599 L
.52949 .00598 L
.52965 .00596 L
.52973 .00595 L
.52981 .00595 L
.52996 .00596 L
.53009 .00599 L
Mistroke
.53024 .00602 L
.5304 .00608 L
.53073 .00623 L
.53103 .00642 L
.5316 .00691 L
.53222 .00758 L
.53332 .00911 L
.53582 .01343 L
.53823 .01755 L
.5395 .0193 L
.54019 .02009 L
.54084 .02069 L
.54142 .02112 L
.54172 .02129 L
.54204 .02145 L
.54223 .02153 L
.5424 .02159 L
.54256 .02163 L
.54273 .02168 L
.54289 .0217 L
.54304 .02172 L
.54319 .02173 L
.54328 .02174 L
.54336 .02174 L
.5435 .02173 L
.54364 .02172 L
.54378 .02171 L
.5439 .02169 L
.54405 .02166 L
.5442 .02162 L
.54449 .02152 L
.54479 .0214 L
.54511 .02123 L
.54568 .02086 L
.54682 .01986 L
.54804 .01846 L
.55025 .01533 L
.55264 .01172 L
.5539 .00997 L
.55523 .00836 L
.55642 .00724 L
.55709 .00675 L
.55773 .00639 L
.55801 .00626 L
.55831 .00615 L
.55856 .00608 L
.55884 .00602 L
.55899 .00599 L
.55914 .00597 L
.55922 .00597 L
Mistroke
.5593 .00596 L
.55946 .00595 L
.5596 .00595 L
.55976 .00596 L
.55989 .00597 L
.56004 .00599 L
.56019 .00601 L
.56033 .00604 L
.56066 .00612 L
.56095 .00622 L
.56126 .00635 L
.56238 .007 L
.56306 .00753 L
.5637 .0081 L
.5649 .00936 L
.56955 .01504 L
.57079 .01634 L
.57195 .01733 L
.57259 .01776 L
.57316 .01807 L
.57347 .0182 L
.5738 .01832 L
.57399 .01837 L
.57416 .01841 L
.57432 .01844 L
.57449 .01847 L
.57463 .01849 L
.57478 .0185 L
.57492 .0185 L
.57505 .01851 L
.5752 .0185 L
.57534 .01849 L
.57542 .01848 L
.5755 .01847 L
.57565 .01845 L
.57593 .01839 L
.57622 .01831 L
.57674 .01811 L
.57736 .01779 L
.57793 .01741 L
.57923 .01627 L
.58168 .01335 L
.58391 .01033 L
.58502 .00892 L
.58621 .00761 L
.58689 .007 L
.58751 .00655 L
.5878 .00637 L
.58813 .00621 L
.58843 .0061 L
Mistroke
.58871 .00602 L
.58886 .00599 L
.58901 .00597 L
.58914 .00596 L
.58928 .00595 L
.58943 .00596 L
.58951 .00596 L
.58959 .00597 L
.58975 .006 L
.58989 .00603 L
.59016 .00611 L
.59042 .00622 L
.59071 .00639 L
.59098 .00658 L
.5916 .00714 L
.59218 .00784 L
.59348 .01005 L
.59475 .01309 L
.59595 .01675 L
.5982 .02555 L
.60284 .04995 L
.60778 .0782 L
.61025 .08998 L
.61145 .09467 L
.61254 .09822 L
.61309 .0997 L
.61368 .10108 L
.61419 .10209 L
.61475 .10299 L
.61535 .1037 L
.61568 .10398 L
.61583 .10408 L
.61598 .10417 L
.61612 .10424 L
.61627 .10429 L
.6164 .10433 L
.61653 .10435 L
.61668 .10436 L
.61682 .10436 L
.61697 .10434 L
.61712 .1043 L
.61726 .10426 L
.61741 .10419 L
.61768 .10403 L
.618 .10378 L
.61829 .10348 L
.61893 .10264 L
.61953 .10158 L
.6206 .09916 L
.62176 .09575 L
Mistroke
.62413 .08662 L
.62666 .07457 L
.63168 .04824 L
.6342 .036 L
.63652 .02632 L
.63878 .01867 L
.63994 .01552 L
.64117 .0127 L
.64241 .01043 L
.64354 .00882 L
.64409 .00819 L
.64468 .00761 L
.64575 .0068 L
.64634 .00649 L
.64689 .00626 L
.6472 .00617 L
.64749 .00609 L
.64781 .00603 L
.64797 .00601 L
.64814 .00599 L
.64822 .00598 L
.64831 .00597 L
.64839 .00597 L
.64847 .00596 L
.64863 .00595 L
.64878 .00595 L
.64886 .00595 L
.64894 .00595 L
.6491 .00596 L
.6492 .00596 L
.64929 .00597 L
.64946 .00598 L
.64976 .00601 L
.65004 .00605 L
.65068 .00617 L
.65196 .00651 L
.65316 .0069 L
.65541 .00766 L
.65659 .00802 L
.65783 .00833 L
.6589 .00852 L
.6595 .0086 L
.66006 .00866 L
.66038 .00868 L
.66067 .00869 L
.66082 .0087 L
.66091 .0087 L
.66099 .0087 L
.66108 .0087 L
.66115 .0087 L
Mistroke
.66124 .0087 L
.66133 .0087 L
.66147 .0087 L
.66163 .0087 L
.66177 .0087 L
.6619 .00869 L
.66222 .00868 L
.66253 .00866 L
.66309 .00861 L
.6636 .00855 L
.66477 .00837 L
.66704 .00788 L
.66953 .00725 L
.67208 .00665 L
.67329 .00642 L
.6744 .00624 L
.67556 .0061 L
.67617 .00605 L
.67648 .00602 L
.67681 .006 L
.67711 .00599 L
.67737 .00597 L
.67763 .00597 L
.6779 .00596 L
.67804 .00596 L
.6782 .00595 L
.67834 .00595 L
.67847 .00595 L
.67863 .00595 L
.67871 .00595 L
.67879 .00595 L
.67894 .00595 L
.6791 .00596 L
.67924 .00596 L
.6794 .00596 L
.67969 .00597 L
.68023 .00599 L
.68087 .00602 L
.68146 .00606 L
.68367 .00625 L
.68594 .00646 L
.68713 .00657 L
.68839 .00666 L
.68903 .0067 L
.68974 .00674 L
.69038 .00676 L
.69067 .00677 L
.69098 .00678 L
.69114 .00678 L
.69129 .00679 L
Mistroke
.69144 .00679 L
.69159 .00679 L
.69173 .00679 L
.69187 .00679 L
.692 .00679 L
.69215 .00679 L
.69231 .00679 L
.69239 .00679 L
.69247 .00679 L
.69256 .00679 L
.69264 .00679 L
.69281 .00679 L
.69296 .00679 L
.69312 .00679 L
.69341 .00678 L
.694 .00676 L
.69456 .00674 L
.69581 .00668 L
.69835 .0065 L
.7006 .00632 L
.70304 .00613 L
.70437 .00606 L
.70495 .00603 L
.7056 .006 L
.70617 .00598 L
.70678 .00597 L
.70711 .00596 L
.70742 .00596 L
.70771 .00595 L
.70786 .00595 L
.70794 .00595 L
.70803 .00595 L
.70818 .00595 L
.70833 .00595 L
.70849 .00595 L
.70858 .00595 L
.70866 .00595 L
.70881 .00595 L
.70895 .00595 L
.70926 .00596 L
.70954 .00596 L
.70984 .00597 L
.71039 .00598 L
.71102 .006 L
.7117 .00602 L
.71291 .00607 L
.71532 .00618 L
.71659 .00624 L
.71793 .0063 L
.71913 .00634 L
Mistroke
.71977 .00636 L
.72045 .00638 L
.72077 .00638 L
.72108 .00639 L
.72137 .00639 L
.72165 .0064 L
.72191 .0064 L
.72204 .0064 L
.72218 .0064 L
.72233 .0064 L
.72248 .0064 L
.72263 .0064 L
.72276 .0064 L
.72292 .0064 L
.72306 .0064 L
.7232 .0064 L
.72334 .0064 L
.72365 .00639 L
.72395 .00639 L
.72448 .00638 L
.72505 .00637 L
.72624 .00634 L
.72755 .00629 L
.73009 .00618 L
.73241 .00608 L
.73356 .00604 L
.73482 .006 L
.73539 .00598 L
.736 .00597 L
.7363 .00597 L
.73658 .00596 L
.73683 .00596 L
.7371 .00596 L
.73725 .00596 L
.73739 .00595 L
.73753 .00595 L
.73766 .00595 L
.73781 .00595 L
.73789 .00595 L
.73798 .00595 L
.73812 .00595 L
.73828 .00595 L
.73836 .00595 L
.73844 .00595 L
.73859 .00595 L
.73875 .00595 L
.73892 .00596 L
.73924 .00596 L
.73953 .00596 L
.74009 .00597 L
Mistroke
.7406 .00598 L
.74176 .006 L
.74408 .00608 L
.74659 .00617 L
.74899 .00625 L
.7502 .00627 L
.75089 .00629 L
.75122 .00629 L
.75153 .00629 L
.75182 .0063 L
.75213 .0063 L
.75229 .0063 L
.75244 .0063 L
.75253 .0063 L
.75261 .0063 L
.7527 .0063 L
.75279 .0063 L
.75293 .0063 L
.75308 .0063 L
.75323 .0063 L
.75336 .0063 L
.75351 .0063 L
.75368 .0063 L
.75398 .00629 L
.75425 .00629 L
.75455 .00629 L
.75515 .00628 L
.75622 .00626 L
.75744 .00623 L
.75856 .00619 L
.76108 .0061 L
.76229 .00606 L
.76357 .00602 L
.76468 .00599 L
.76526 .00598 L
.76588 .00597 L
.76619 .00596 L
.76653 .00596 L
.76684 .00596 L
.76712 .00595 L
.76728 .00595 L
.76742 .00595 L
.76755 .00595 L
.76769 .00595 L
.76784 .00595 L
.76792 .00595 L
.768 .00595 L
.76816 .00595 L
.7683 .00595 L
.76842 .00595 L
Mistroke
.76856 .00595 L
.76885 .00596 L
.76911 .00596 L
.76935 .00596 L
.7699 .00597 L
.77049 .00598 L
.77173 .00601 L
.77307 .00605 L
.7755 .00614 L
.77785 .00623 L
.7791 .00627 L
.77978 .00629 L
.78044 .00631 L
.78105 .00632 L
.78138 .00633 L
.78172 .00633 L
.78201 .00633 L
.78233 .00634 L
.78247 .00634 L
.78262 .00634 L
.78276 .00634 L
.7829 .00634 L
.78303 .00634 L
.78317 .00634 L
.78331 .00634 L
.78343 .00634 L
.78357 .00634 L
.78373 .00634 L
.78388 .00634 L
.78401 .00634 L
.78429 .00634 L
.78459 .00633 L
.78519 .00632 L
.78582 .00631 L
.78642 .0063 L
.78753 .00627 L
.79005 .00617 L
.79247 .00607 L
.79352 .00603 L
.79466 .006 L
.7953 .00598 L
.79588 .00597 L
.79644 .00596 L
.79672 .00596 L
.79687 .00596 L
.79703 .00595 L
.79712 .00595 L
.7972 .00595 L
.79728 .00595 L
.79736 .00595 L
Mistroke
.79751 .00595 L
.79766 .00595 L
.79775 .00595 L
.79783 .00595 L
.79792 .00595 L
.79801 .00595 L
.79817 .00595 L
.79834 .00596 L
.79865 .00596 L
.79897 .00596 L
.79956 .00597 L
.80021 .00599 L
.8008 .00601 L
.80214 .00606 L
.80457 .0062 L
.80702 .00637 L
.80932 .00652 L
.81045 .00659 L
.81152 .00663 L
.81212 .00665 L
.81243 .00666 L
.81276 .00667 L
.81307 .00667 L
.81335 .00668 L
.81348 .00668 L
.81363 .00668 L
.81376 .00668 L
.81389 .00668 L
.81404 .00668 L
.81419 .00668 L
.81432 .00668 L
.81446 .00668 L
.81461 .00668 L
.81478 .00668 L
.81507 .00667 L
.81534 .00667 L
.81559 .00666 L
.81617 .00664 L
.81679 .00662 L
.81737 .00659 L
.81866 .00652 L
.82114 .00632 L
.82345 .00613 L
.82459 .00605 L
.82515 .00602 L
.82567 .00599 L
.82627 .00597 L
.82659 .00596 L
.82674 .00596 L
.82692 .00596 L
Mistroke
.82707 .00595 L
.8272 .00595 L
.82736 .00595 L
.8275 .00595 L
.82764 .00595 L
.82776 .00595 L
.8279 .00596 L
.82805 .00596 L
.82834 .00597 L
.82864 .00598 L
.82889 .00599 L
.82918 .00601 L
.82981 .00605 L
.83039 .00611 L
.83146 .00626 L
.8326 .00647 L
.83383 .00678 L
.83517 .0072 L
.8376 .00819 L
.84253 .01091 L
.84728 .01379 L
.84967 .01503 L
.85185 .01591 L
.85305 .01627 L
.8536 .0164 L
.85419 .01651 L
.85452 .01656 L
.85481 .0166 L
.85512 .01664 L
.85539 .01666 L
.85555 .01667 L
.85572 .01668 L
.85589 .01669 L
.85598 .01669 L
.85607 .01669 L
.85615 .0167 L
.85623 .0167 L
.85638 .0167 L
.85654 .0167 L
.85663 .01669 L
.85671 .01669 L
.85685 .01669 L
.857 .01668 L
.85727 .01667 L
.85756 .01664 L
.85787 .01661 L
.85844 .01653 L
.85896 .01644 L
.86016 .01615 L
.86145 .01573 L
Mistroke
.86402 .0146 L
.86646 .01329 L
.8711 .01054 L
.87345 .00924 L
.87604 .00802 L
.87849 .00711 L
.87959 .0068 L
.88079 .00652 L
.88205 .00629 L
.88263 .00621 L
.88324 .00613 L
.88385 .00607 L
.8844 .00603 L
.88495 .006 L
.88548 .00598 L
.88577 .00597 L
.88604 .00596 L
.8862 .00596 L
.88634 .00596 L
.88642 .00596 L
.8865 .00595 L
.88666 .00595 L
.88681 .00595 L
.88689 .00595 L
.88697 .00595 L
.88706 .00595 L
.88713 .00595 L
.88722 .00595 L
.88731 .00595 L
.88745 .00596 L
.88759 .00596 L
.8879 .00596 L
.8882 .00597 L
.88847 .00598 L
.88898 .00599 L
.89013 .00604 L
.89235 .00616 L
.89361 .00622 L
.89478 .00628 L
.89604 .00633 L
.89719 .00637 L
.89782 .00638 L
.89841 .00639 L
.89874 .0064 L
.89904 .0064 L
.8992 .0064 L
.89929 .0064 L
.89937 .0064 L
.89946 .0064 L
.89956 .0064 L
Mistroke
.89964 .0064 L
.89973 .0064 L
.89988 .0064 L
.89996 .0064 L
.90004 .0064 L
.90018 .0064 L
.90033 .0064 L
.90059 .0064 L
.90088 .0064 L
.90118 .00639 L
.90151 .00639 L
.90211 .00637 L
.90327 .00634 L
.90432 .00631 L
.90912 .00612 L
.91036 .00607 L
.91154 .00603 L
.91268 .006 L
.91373 .00598 L
.91428 .00597 L
.91487 .00596 L
.91516 .00596 L
.91546 .00596 L
.91564 .00596 L
.91579 .00595 L
.91594 .00595 L
.91611 .00595 L
.91626 .00595 L
.91635 .00595 L
.91643 .00595 L
.91657 .00595 L
.91673 .00595 L
.91681 .00595 L
.9169 .00595 L
.917 .00595 L
.91708 .00595 L
.91724 .00595 L
.91741 .00595 L
.91774 .00596 L
.91805 .00596 L
.91864 .00596 L
.91981 .00598 L
.92105 .006 L
.92328 .00604 L
.92567 .00608 L
.92672 .00609 L
.92786 .00611 L
.92848 .00611 L
.92876 .00611 L
.92906 .00612 L
Mistroke
.92933 .00612 L
.92962 .00612 L
.92977 .00612 L
.92992 .00612 L
.93006 .00612 L
.93021 .00612 L
.93036 .00612 L
.93051 .00612 L
.93059 .00612 L
.93067 .00612 L
.93083 .00612 L
.93098 .00612 L
.93113 .00612 L
.93141 .00612 L
.93173 .00612 L
.93204 .00612 L
.93273 .00611 L
.93333 .00611 L
.93398 .0061 L
.93517 .00608 L
.9374 .00605 L
.93998 .00601 L
.94231 .00598 L
.94348 .00596 L
.94413 .00596 L
.94474 .00596 L
.94527 .00595 L
.94557 .00595 L
.94571 .00595 L
.94585 .00595 L
.94601 .00595 L
.94615 .00595 L
.94629 .00595 L
.94642 .00595 L
.94656 .00595 L
.94672 .00595 L
.94679 .00595 L
.94688 .00595 L
.94703 .00595 L
.94732 .00595 L
.94763 .00595 L
.94818 .00596 L
.94885 .00596 L
.94945 .00596 L
.95199 .00599 L
.9545 .00602 L
.9556 .00603 L
.95682 .00604 L
.95798 .00605 L
.95854 .00605 L
Mistroke
.95907 .00605 L
.95936 .00605 L
.95967 .00606 L
.95996 .00606 L
.96022 .00606 L
.96038 .00606 L
.96053 .00606 L
.96067 .00606 L
.96082 .00606 L
.9609 .00606 L
.96098 .00606 L
.96107 .00606 L
.96115 .00606 L
.9613 .00606 L
.96146 .00606 L
.9616 .00606 L
.96177 .00606 L
.9621 .00606 L
.96269 .00605 L
.96338 .00605 L
.96402 .00605 L
.96645 .00603 L
.97113 .00598 L
.97234 .00597 L
.97299 .00596 L
.97367 .00596 L
.97433 .00596 L
.97493 .00595 L
.97522 .00595 L
.97537 .00595 L
.97553 .00595 L
.97567 .00595 L
.97582 .00595 L
.97596 .00595 L
.9761 .00595 L
.97619 .00595 L
Mfstroke
0 0 m
1 0 L
1 .25 L
0 .25 L
closepath
clip
newpath
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.0238095 0.280952 0.97619 0.519048 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.119048 0.00595238 0.238095 [
[.02381 -0.00655 -9.34375 -15.6875 ]
[.02381 -0.00655 9.34375 0 ]
[.2619 -0.00655 -9.34375 -15.6875 ]
[.2619 -0.00655 9.34375 0 ]
[.7381 -0.00655 -5 -15.6875 ]
[.7381 -0.00655 5 0 ]
[.97619 -0.00655 -5 -15.6875 ]
[.97619 -0.00655 5 0 ]
[.5 .2625 -97.0625 0 ]
[.5 .2625 97.0625 15.6875 ]
[ 0 0 0 0 ]
[ 1 .25 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
.02381 .00595 m
.02381 .0122 L
s
gsave
.02381 -0.00655 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(-) show
71.687500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%+ font Math1-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%+ Math1-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
%%+ font Math1-Italic
grestore
.2619 .00595 m
.2619 .0122 L
s
gsave
.2619 -0.00655 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(-) show
71.687500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%+ font Math1-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%+ Math1-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
%%+ font Math1-Italic
grestore
.7381 .00595 m
.7381 .0122 L
s
gsave
.7381 -0.00655 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.97619 .00595 m
.97619 .0122 L
s
gsave
.97619 -0.00655 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.125 Mabswid
.08333 .00595 m
.08333 .0097 L
s
.14286 .00595 m
.14286 .0097 L
s
.20238 .00595 m
.20238 .0097 L
s
.32143 .00595 m
.32143 .0097 L
s
.38095 .00595 m
.38095 .0097 L
s
.44048 .00595 m
.44048 .0097 L
s
.55952 .00595 m
.55952 .0097 L
s
.61905 .00595 m
.61905 .0097 L
s
.67857 .00595 m
.67857 .0097 L
s
.79762 .00595 m
.79762 .0097 L
s
.85714 .00595 m
.85714 .0097 L
s
.91667 .00595 m
.91667 .0097 L
s
.25 Mabswid
0 .00595 m
1 .00595 L
s
.5 0 m
.5 .25 L
s
gsave
.5 .2625 -158.063 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(single) show
91.562500 15.812500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(-) show
99.687500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(slit) show
117.562500 15.812500 moveto
(diffraction) show
172.375000 15.812500 moveto
(pattern) show
207.125000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(,) show
213.125000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(a) show
%%IncludeResource: font Math2
%%IncludeFont: Math2
/Math2 findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
219.125000 15.812500 moveto
(\\220) show
222.937500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(d) show
229.625000 15.812500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(=) show
238.125000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.5) show
253.125000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%+ font Math1-Italic
%%+ font Math2
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%+ Math1-Italic
%%+ Math2
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
%%+ font Math1-Italic
%%+ font Math2
grestore
.5 Mabswid
.02381 .00595 m
.02395 .00595 L
.02408 .00595 L
.02423 .00595 L
.02437 .00595 L
.02462 .00595 L
.02489 .00595 L
.02519 .00595 L
.02551 .00596 L
.02608 .00596 L
.02667 .00596 L
.02722 .00596 L
.02847 .00598 L
.02975 .00599 L
.03092 .00601 L
.03355 .00606 L
.03584 .00611 L
.03834 .00619 L
.04324 .00637 L
.04796 .00661 L
.05298 .00692 L
.05781 .00727 L
.06246 .00765 L
.0674 .00811 L
.07216 .0086 L
.07674 .00911 L
.08162 .00968 L
.08631 .01026 L
.0913 .0109 L
.09611 .01153 L
.10073 .01214 L
.10565 .01279 L
.11038 .01341 L
.11494 .01398 L
.11978 .01456 L
.12445 .01509 L
.12941 .0156 L
.13419 .01604 L
.13879 .0164 L
.14368 .01672 L
.14839 .01696 L
.15096 .01705 L
.15214 .01709 L
.15339 .01712 L
.15454 .01714 L
.15581 .01716 L
.15636 .01717 L
.15695 .01718 L
.15752 .01718 L
.15803 .01718 L
Mistroke
.15833 .01719 L
.15861 .01719 L
.15876 .01719 L
.15892 .01719 L
.15908 .01719 L
.15922 .01719 L
.15935 .01719 L
.15948 .01719 L
.15961 .01719 L
.15976 .01719 L
.15991 .01719 L
.16005 .01719 L
.16032 .01719 L
.16063 .01719 L
.16095 .01718 L
.16152 .01718 L
.16214 .01717 L
.16283 .01717 L
.16405 .01715 L
.16516 .01712 L
.16768 .01705 L
.16997 .01697 L
.17247 .01685 L
.17736 .01655 L
.18208 .01617 L
.18709 .01568 L
.19192 .01512 L
.19656 .01451 L
.20151 .0138 L
.20626 .01307 L
.21084 .01232 L
.21571 .0115 L
.2204 .0107 L
.22538 .00985 L
.23019 .00906 L
.2348 .00833 L
.23972 .00763 L
.24445 .00704 L
.249 .00658 L
.25153 .00637 L
.25384 .00621 L
.25611 .00609 L
.25726 .00604 L
.2579 .00602 L
.25851 .006 L
.2591 .00598 L
.25974 .00597 L
.26033 .00596 L
.26059 .00596 L
.26087 .00596 L
Mistroke
.26103 .00596 L
.26117 .00595 L
.26132 .00595 L
.26145 .00595 L
.26161 .00595 L
.26169 .00595 L
.26178 .00595 L
.26187 .00595 L
.26195 .00595 L
.26209 .00595 L
.26225 .00595 L
.26233 .00595 L
.26242 .00595 L
.26272 .00596 L
.26289 .00596 L
.26307 .00596 L
.26339 .00596 L
.26403 .00597 L
.26462 .00598 L
.26596 .00602 L
.26713 .00607 L
.26838 .00614 L
.27073 .0063 L
.27294 .00651 L
.27761 .00712 L
.28003 .00754 L
.28257 .00805 L
.28717 .00919 L
.29206 .0107 L
.29678 .01248 L
.30178 .01473 L
.30661 .01727 L
.31125 .02005 L
.31619 .0234 L
.32094 .027 L
.32551 .03083 L
.33038 .03528 L
.33506 .03994 L
.34005 .04529 L
.34484 .05081 L
.34946 .05646 L
.35437 .06282 L
.3591 .06926 L
.36412 .07643 L
.36896 .08362 L
.37362 .09078 L
.37857 .09861 L
.38334 .10635 L
.38793 .11392 L
.39282 .1221 L
Mistroke
.39752 .13004 L
.40251 .13852 L
.40733 .14667 L
.41196 .15447 L
.41689 .16266 L
.42163 .1704 L
.42619 .17768 L
.43105 .18519 L
.43572 .19216 L
.44069 .19925 L
.44548 .20571 L
.45009 .21155 L
.45499 .21734 L
.45971 .22244 L
.46472 .22734 L
.46955 .23151 L
.4742 .23499 L
.47914 .2381 L
.4839 .24049 L
.48848 .24222 L
.49083 .24289 L
.49335 .24344 L
.49458 .24364 L
.49591 .24382 L
.49652 .24388 L
.49717 .24394 L
.49778 .24398 L
.49834 .24401 L
.49862 .24402 L
.49889 .24403 L
.49917 .24404 L
.49931 .24404 L
.49947 .24404 L
.49963 .24405 L
.49972 .24405 L
.4998 .24405 L
.49995 .24405 L
.50011 .24405 L
.50025 .24405 L
.50038 .24405 L
.50053 .24404 L
.50069 .24404 L
.50097 .24403 L
.50128 .24403 L
.50184 .244 L
.50238 .24397 L
.50289 .24393 L
.50417 .24381 L
.50537 .24365 L
.50762 .24325 L
Mistroke
.51004 .24266 L
.51227 .24197 L
.51704 .24006 L
.52162 .23766 L
.5265 .2345 L
.5312 .2309 L
.53619 .22649 L
.541 .22171 L
.54563 .21663 L
.55055 .21077 L
.55529 .20469 L
.55985 .19849 L
.5647 .19154 L
.56937 .18455 L
.57434 .17684 L
.57912 .16918 L
.58373 .16165 L
.58862 .15349 L
.59334 .14555 L
.59835 .13706 L
.60317 .12887 L
.60782 .12104 L
.61276 .11278 L
.61751 .10495 L
.62209 .09756 L
.62696 .08988 L
.63164 .08271 L
.63663 .07534 L
.64143 .06854 L
.64604 .06227 L
.65096 .05594 L
.65569 .05018 L
.66023 .04497 L
.66508 .03979 L
.66974 .03517 L
.67469 .03065 L
.67947 .02668 L
.68406 .02322 L
.68895 .01993 L
.69365 .01712 L
.69865 .01452 L
.70346 .01238 L
.7081 .01065 L
.71303 .00913 L
.71777 .00798 L
.72234 .00713 L
.72488 .00677 L
.7272 .0065 L
.7296 .00628 L
.73188 .00612 L
Mistroke
.73303 .00606 L
.73426 .00602 L
.73486 .006 L
.73543 .00598 L
.73597 .00597 L
.73648 .00596 L
.73678 .00596 L
.73704 .00596 L
.73734 .00595 L
.73751 .00595 L
.73766 .00595 L
.73781 .00595 L
.73797 .00595 L
.73811 .00595 L
.73826 .00595 L
.73834 .00595 L
.73842 .00595 L
.7385 .00595 L
.73859 .00595 L
.73874 .00595 L
.7389 .00596 L
.7392 .00596 L
.73946 .00596 L
.74006 .00597 L
.74061 .00598 L
.74113 .00599 L
.74231 .00602 L
.74355 .00607 L
.74577 .00618 L
.74813 .00634 L
.75034 .00652 L
.75503 .00698 L
.76001 .0076 L
.7648 .00828 L
.76941 .00899 L
.77432 .0098 L
.77905 .0106 L
.78407 .01146 L
.78891 .01228 L
.79357 .01304 L
.79852 .01381 L
.80329 .01449 L
.80788 .0151 L
.81276 .01566 L
.81746 .01613 L
.82245 .01654 L
.82496 .0167 L
.82727 .01684 L
.82966 .01695 L
.8319 .01704 L
Mistroke
.83415 .01711 L
.83535 .01713 L
.83664 .01716 L
.83786 .01717 L
.83855 .01718 L
.83888 .01718 L
.83919 .01718 L
.83948 .01719 L
.83964 .01719 L
.8398 .01719 L
.83994 .01719 L
.84009 .01719 L
.84023 .01719 L
.84036 .01719 L
.84044 .01719 L
.84052 .01719 L
.84067 .01719 L
.84082 .01719 L
.84096 .01719 L
.84111 .01719 L
.84128 .01719 L
.84161 .01719 L
.84189 .01718 L
.8422 .01718 L
.84282 .01718 L
.8434 .01717 L
.84394 .01717 L
.84516 .01715 L
.84645 .01712 L
.84903 .01705 L
.85146 .01696 L
.85609 .01674 L
.86102 .01642 L
.86577 .01604 L
.87034 .01562 L
.8752 .01513 L
.87988 .0146 L
.88486 .01401 L
.88965 .0134 L
.89426 .0128 L
.89916 .01216 L
.90389 .01153 L
.9089 .01087 L
.91374 .01026 L
.91839 .00968 L
.92334 .0091 L
.92811 .00857 L
.93269 .00811 L
.93757 .00765 L
.94226 .00726 L
Mistroke
.94726 .0069 L
.95206 .00661 L
.95669 .00638 L
.96161 .00619 L
.96406 .00611 L
.96635 .00606 L
.96854 .00602 L
.97091 .00598 L
.97206 .00597 L
.97333 .00596 L
.97396 .00596 L
.97432 .00596 L
.97465 .00595 L
.97495 .00595 L
.97512 .00595 L
.97529 .00595 L
.97543 .00595 L
.97559 .00595 L
.97574 .00595 L
.97587 .00595 L
.97596 .00595 L
.97604 .00595 L
.97619 .00595 L
Mfstroke
0 0 m
1 0 L
1 .25 L
0 .25 L
closepath
clip
newpath
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.0238095 0.542857 0.97619 0.780952 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.119048 0.00595238 0.238095 [
[.02381 -0.00655 -9.34375 -15.6875 ]
[.02381 -0.00655 9.34375 0 ]
[.2619 -0.00655 -9.34375 -15.6875 ]
[.2619 -0.00655 9.34375 0 ]
[.7381 -0.00655 -5 -15.6875 ]
[.7381 -0.00655 5 0 ]
[.97619 -0.00655 -5 -15.6875 ]
[.97619 -0.00655 5 0 ]
[.5 .2625 -71.7188 0 ]
[.5 .2625 71.7188 15.6875 ]
[ 0 0 0 0 ]
[ 1 .25 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
.02381 .00595 m
.02381 .0122 L
s
gsave
.02381 -0.00655 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(-) show
71.687500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%+ font Math1-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%+ Math1-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
%%+ font Math1-Italic
grestore
.2619 .00595 m
.2619 .0122 L
s
gsave
.2619 -0.00655 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(-) show
71.687500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%+ font Math1-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%+ Math1-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
%%+ font Math1-Italic
grestore
.7381 .00595 m
.7381 .0122 L
s
gsave
.7381 -0.00655 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.97619 .00595 m
.97619 .0122 L
s
gsave
.97619 -0.00655 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.125 Mabswid
.08333 .00595 m
.08333 .0097 L
s
.14286 .00595 m
.14286 .0097 L
s
.20238 .00595 m
.20238 .0097 L
s
.32143 .00595 m
.32143 .0097 L
s
.38095 .00595 m
.38095 .0097 L
s
.44048 .00595 m
.44048 .0097 L
s
.55952 .00595 m
.55952 .0097 L
s
.61905 .00595 m
.61905 .0097 L
s
.67857 .00595 m
.67857 .0097 L
s
.79762 .00595 m
.79762 .0097 L
s
.85714 .00595 m
.85714 .0097 L
s
.91667 .00595 m
.91667 .0097 L
s
.25 Mabswid
0 .00595 m
1 .00595 L
s
.5 0 m
.5 .25 L
s
gsave
.5 .2625 -132.719 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4) show
69.000000 15.812500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(-) show
77.125000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(beam) show
106.000000 15.812500 moveto
(interference) show
167.875000 15.812500 moveto
(pattern) show
202.437500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%+ font Math1-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%+ Math1-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
%%+ font Math1-Italic
grestore
.5 Mabswid
.02381 .24405 m
.02395 .24403 L
.02408 .24399 L
.02423 .2439 L
.02437 .24379 L
.02462 .2435 L
.02489 .24308 L
.02519 .24247 L
.02551 .24166 L
.02608 .23981 L
.02722 .23458 L
.02847 .22657 L
.03109 .2031 L
.03355 .1748 L
.03823 .11361 L
.0406 .08362 L
.0432 .055 L
.04568 .03331 L
.04679 .02568 L
.04799 .01891 L
.04926 .01337 L
.04985 .01138 L
.05046 .00965 L
.05102 .00839 L
.05163 .00732 L
.0519 .00696 L
.05219 .00663 L
.05243 .00641 L
.05271 .00621 L
.05287 .00612 L
.05302 .00606 L
.05317 .00601 L
.05326 .00598 L
.05334 .00597 L
.05348 .00596 L
.05363 .00595 L
.05379 .00597 L
.05393 .006 L
.05402 .00602 L
.05411 .00604 L
.05427 .00611 L
.05458 .00627 L
.05494 .00653 L
.05527 .00682 L
.05645 .00826 L
.05769 .0103 L
.06226 .01896 L
.06347 .0208 L
.06461 .02216 L
.06524 .02273 L
Mistroke
.06557 .02297 L
.06593 .02319 L
.06625 .02335 L
.06655 .02346 L
.06669 .0235 L
.06685 .02353 L
.06699 .02356 L
.06713 .02358 L
.06728 .02359 L
.06743 .02359 L
.06756 .02358 L
.0677 .02357 L
.06785 .02355 L
.06801 .02352 L
.06831 .02344 L
.06858 .02334 L
.06883 .02322 L
.0694 .02287 L
.07003 .02236 L
.07061 .0218 L
.0719 .02021 L
.07668 .01235 L
.07792 .01039 L
.0791 .00877 L
.08024 .0075 L
.08129 .00664 L
.08184 .00632 L
.08212 .0062 L
.08242 .00609 L
.08258 .00605 L
.08273 .00601 L
.08281 .006 L
.0829 .00598 L
.08306 .00596 L
.08321 .00595 L
.08337 .00595 L
.0835 .00596 L
.08366 .00597 L
.08381 .00599 L
.08398 .00602 L
.08428 .0061 L
.08461 .00622 L
.08496 .00639 L
.08559 .00679 L
.08618 .00727 L
.08745 .00862 L
.08865 .01025 L
.09089 .01388 L
.09309 .01765 L
.09435 .01961 L
Mistroke
.09552 .02119 L
.09668 .0224 L
.09733 .02291 L
.09792 .02326 L
.09824 .0234 L
.09854 .02349 L
.09871 .02353 L
.09887 .02356 L
.09896 .02357 L
.09905 .02358 L
.09914 .02359 L
.09922 .02359 L
.09937 .02359 L
.09951 .02358 L
.09966 .02356 L
.09974 .02355 L
.09982 .02353 L
.09998 .0235 L
.10015 .02345 L
.10045 .02333 L
.10074 .02319 L
.10104 .02301 L
.10159 .0226 L
.10284 .02127 L
.10416 .01935 L
.10538 .01724 L
.10785 .01241 L
.10906 .01015 L
.11016 .00835 L
.11071 .0076 L
.1113 .00692 L
.11157 .00666 L
.11186 .00642 L
.11213 .00624 L
.11238 .00612 L
.11253 .00606 L
.11266 .00601 L
.11281 .00598 L
.11297 .00596 L
.1131 .00595 L
.11325 .00596 L
.11339 .00598 L
.11351 .00601 L
.11367 .00607 L
.11382 .00613 L
.11415 .00634 L
.11445 .0066 L
.11476 .00694 L
.11536 .00784 L
.1159 .00893 L
Mistroke
.11713 .01239 L
.11827 .01694 L
.11934 .02248 L
.12175 .03934 L
.12397 .06015 L
.12872 .11727 L
.13328 .17683 L
.13582 .20561 L
.13814 .22613 L
.13928 .23364 L
.13989 .23683 L
.14053 .23961 L
.14115 .24165 L
.14172 .24297 L
.14199 .24342 L
.14213 .24361 L
.14228 .24377 L
.14241 .24388 L
.14256 .24397 L
.14269 .24402 L
.14281 .24405 L
.14296 .24404 L
.14304 .24402 L
.14313 .24399 L
.14328 .2439 L
.14342 .24379 L
.14358 .24361 L
.14373 .24341 L
.14407 .24283 L
.14435 .24221 L
.14465 .24138 L
.14527 .23928 L
.14642 .23367 L
.1475 .22666 L
.14997 .20487 L
.15221 .17956 L
.15703 .11679 L
.16167 .06089 L
.16407 .0385 L
.16529 .0293 L
.1666 .02118 L
.16784 .01524 L
.16896 .01117 L
.1695 .00966 L
.17009 .00833 L
.17065 .00736 L
.17117 .0067 L
.17149 .0064 L
.17178 .00619 L
.17194 .00611 L
Mistroke
.17211 .00604 L
.17219 .00601 L
.17228 .00599 L
.17237 .00597 L
.17245 .00596 L
.1726 .00595 L
.17274 .00596 L
.1729 .00598 L
.17298 .00599 L
.17306 .00602 L
.17321 .00606 L
.17337 .00613 L
.17365 .00628 L
.17392 .00648 L
.17422 .00673 L
.17482 .00737 L
.17589 .00887 L
.17846 .01359 L
.18079 .01806 L
.18196 .01999 L
.18322 .02169 L
.1838 .0223 L
.18441 .02282 L
.18493 .02316 L
.18523 .02331 L
.1855 .02343 L
.18566 .02347 L
.18582 .02352 L
.18595 .02354 L
.18611 .02357 L
.18619 .02358 L
.18627 .02358 L
.18635 .02359 L
.18644 .02359 L
.1866 .02358 L
.18668 .02358 L
.18676 .02357 L
.18692 .02355 L
.18706 .02352 L
.18734 .02344 L
.18763 .02333 L
.18796 .02318 L
.18852 .02282 L
.18912 .02233 L
.19019 .02119 L
.19255 .01777 L
.19476 .014 L
.19699 .01036 L
.19817 .00873 L
.19943 .00736 L
Mistroke
.20013 .00678 L
.20075 .00639 L
.20106 .00624 L
.20139 .00612 L
.20154 .00607 L
.2017 .00603 L
.20184 .006 L
.20198 .00598 L
.20213 .00596 L
.20227 .00595 L
.20242 .00595 L
.2025 .00595 L
.20258 .00596 L
.20271 .00597 L
.20285 .00599 L
.203 .00602 L
.20314 .00605 L
.20346 .00614 L
.20375 .00626 L
.20441 .00662 L
.20498 .00705 L
.20559 .00761 L
.20669 .00886 L
.20919 .01261 L
.21159 .01673 L
.21285 .01879 L
.21419 .02071 L
.21539 .02208 L
.21606 .02269 L
.21639 .02292 L
.2167 .02312 L
.21698 .02326 L
.21728 .02339 L
.21756 .02348 L
.21768 .02352 L
.21782 .02354 L
.21797 .02357 L
.21812 .02358 L
.21826 .02359 L
.2184 .02359 L
.21854 .02358 L
.2187 .02356 L
.21878 .02355 L
.21886 .02353 L
.21901 .0235 L
.21932 .0234 L
.21965 .02326 L
.22024 .0229 L
.22088 .02238 L
.22158 .02164 L
Mistroke
.22283 .01995 L
.22401 .01799 L
.22624 .0137 L
.22868 .00917 L
.22932 .00818 L
.23 .00729 L
.23033 .00693 L
.23065 .00664 L
.23092 .00641 L
.23123 .00622 L
.23152 .00608 L
.23168 .00602 L
.23183 .00598 L
.23197 .00596 L
.2321 .00595 L
.23225 .00596 L
.23239 .00597 L
.23254 .00601 L
.23262 .00603 L
.2327 .00606 L
.23299 .0062 L
.23316 .00631 L
.23333 .00645 L
.23365 .00676 L
.23425 .00758 L
.2348 .00861 L
.23541 .01007 L
.23607 .01202 L
.2374 .01732 L
.23861 .02375 L
.24092 .04036 L
.24343 .06445 L
.24836 .12512 L
.25311 .18625 L
.2555 .21182 L
.25768 .22957 L
.25825 .23318 L
.25888 .23654 L
.25948 .2392 L
.26002 .24111 L
.26032 .24198 L
.26065 .24275 L
.26083 .2431 L
.261 .24337 L
.26117 .2436 L
.26133 .24377 L
.26146 .24388 L
.26161 .24397 L
.26169 .24401 L
.26177 .24403 L
Mistroke
.26192 .24405 L
.26207 .24402 L
.26221 .24397 L
.26237 .24387 L
.26254 .24372 L
.26283 .24335 L
.26313 .24281 L
.26367 .24148 L
.26425 .23951 L
.26489 .23675 L
.26603 .23025 L
.26709 .22252 L
.26956 .1991 L
.27181 .1728 L
.27663 .10956 L
.28127 .05519 L
.28367 .03408 L
.2849 .0256 L
.28621 .01827 L
.28745 .01304 L
.28858 .0096 L
.28912 .00838 L
.28971 .00735 L
.28998 .00698 L
.29027 .00665 L
.29051 .00642 L
.29079 .00622 L
.29095 .00613 L
.2911 .00606 L
.29126 .00601 L
.29135 .00599 L
.29143 .00597 L
.29158 .00595 L
.29174 .00595 L
.29189 .00597 L
.29203 .006 L
.29211 .00602 L
.29219 .00604 L
.29234 .0061 L
.29268 .00628 L
.29304 .00654 L
.29337 .00683 L
.29462 .00838 L
.29581 .01034 L
.30042 .01905 L
.30164 .0209 L
.30279 .02224 L
.30342 .0228 L
.30376 .02303 L
.30411 .02324 L
Mistroke
.30441 .02337 L
.30457 .02343 L
.30474 .02348 L
.3049 .02352 L
.30504 .02355 L
.30518 .02357 L
.30532 .02358 L
.30545 .02359 L
.3056 .02359 L
.30574 .02358 L
.30587 .02356 L
.30601 .02354 L
.30617 .02351 L
.30648 .02341 L
.30679 .02328 L
.30712 .02311 L
.30772 .02271 L
.30888 .0216 L
.30994 .02028 L
.31476 .01238 L
.31697 .00906 L
.31822 .0076 L
.31884 .00704 L
.3194 .00663 L
.31968 .00645 L
.31999 .00629 L
.32029 .00617 L
.32055 .00608 L
.3207 .00604 L
.32087 .006 L
.32096 .00599 L
.32104 .00598 L
.32112 .00597 L
.3212 .00596 L
.32134 .00595 L
.32149 .00595 L
.32164 .00596 L
.3218 .00597 L
.32188 .00599 L
.32197 .006 L
.32212 .00603 L
.32248 .00613 L
.32279 .00626 L
.32312 .00642 L
.32375 .00683 L
.32433 .00732 L
.32552 .00859 L
.32677 .01029 L
.329 .01392 L
.33142 .01802 L
Mistroke
.33361 .02117 L
.33419 .02184 L
.33483 .02245 L
.33537 .02288 L
.33598 .02324 L
.33631 .02339 L
.33662 .02349 L
.33679 .02353 L
.33694 .02356 L
.33703 .02357 L
.33712 .02358 L
.33721 .02358 L
.33729 .02359 L
.33744 .02359 L
.33757 .02358 L
.33772 .02357 L
.3378 .02355 L
.33788 .02354 L
.33818 .02346 L
.33834 .02341 L
.3385 .02335 L
.33914 .02301 L
.33974 .02255 L
.34087 .02136 L
.34208 .01965 L
.34339 .01739 L
.34574 .01282 L
.34698 .01046 L
.34832 .00825 L
.34898 .00738 L
.3496 .00672 L
.34987 .00649 L
.35016 .00628 L
.35033 .00619 L
.35048 .00611 L
.35062 .00606 L
.35077 .00601 L
.35092 .00598 L
.35107 .00596 L
.35121 .00595 L
.35134 .00596 L
.3515 .00598 L
.35164 .00602 L
.3518 .00608 L
.35196 .00615 L
.35223 .00633 L
.35249 .00655 L
.35307 .00723 L
.35363 .00816 L
.35423 .00948 L
Mistroke
.35532 .01273 L
.35662 .01814 L
.35781 .02466 L
.36009 .04138 L
.36259 .06565 L
.36748 .12609 L
.37219 .18666 L
.37459 .21231 L
.37586 .22328 L
.3772 .23258 L
.37839 .23867 L
.37907 .24113 L
.3794 .24205 L
.37971 .24277 L
.37999 .24328 L
.38014 .24349 L
.38029 .24369 L
.38044 .24383 L
.38057 .24393 L
.3807 .24399 L
.38082 .24403 L
.38097 .24405 L
.38112 .24402 L
.38126 .24397 L
.3814 .24388 L
.38156 .24375 L
.38172 .24356 L
.38202 .2431 L
.38233 .24247 L
.38266 .24163 L
.38325 .23969 L
.38389 .23698 L
.38459 .23327 L
.38702 .21501 L
.38936 .19078 L
.39188 .15978 L
.39685 .09467 L
.39933 .06554 L
.40163 .04295 L
.40386 .02614 L
.40511 .01904 L
.40624 .01401 L
.40682 .01191 L
.40744 .01007 L
.40798 .00877 L
.40856 .00767 L
.40888 .00717 L
.40917 .0068 L
.40947 .00649 L
.40975 .00627 L
Mistroke
.40991 .00618 L
.41008 .00609 L
.41025 .00603 L
.41034 .006 L
.41043 .00598 L
.41051 .00597 L
.41059 .00596 L
.41073 .00595 L
.41089 .00596 L
.41098 .00598 L
.41107 .00599 L
.41122 .00603 L
.41138 .0061 L
.41167 .00624 L
.41199 .00645 L
.41233 .00675 L
.41353 .00818 L
.41469 .01004 L
.41577 .01205 L
.41801 .01644 L
.42045 .02058 L
.42109 .02143 L
.42178 .02219 L
.42237 .02272 L
.42301 .02315 L
.42335 .02333 L
.4235 .02339 L
.42366 .02345 L
.42381 .02349 L
.42396 .02353 L
.42405 .02354 L
.42412 .02356 L
.42428 .02358 L
.42442 .02359 L
.42456 .02359 L
.42471 .02358 L
.42487 .02357 L
.42501 .02355 L
.42514 .02352 L
.42543 .02344 L
.42572 .02333 L
.426 .02321 L
.42662 .02282 L
.42722 .02232 L
.42787 .02167 L
.4301 .01864 L
.4325 .0146 L
.43511 .01033 L
.43633 .00866 L
.437 .00789 L
Mistroke
.43762 .00727 L
.4382 .0068 L
.43872 .00646 L
.439 .00631 L
.43931 .00618 L
.43947 .00612 L
.43963 .00607 L
.43978 .00603 L
.43993 .006 L
.44007 .00598 L
.44023 .00596 L
.44036 .00595 L
.44049 .00595 L
.44064 .00596 L
.44078 .00597 L
.44086 .00598 L
.44094 .00599 L
.44109 .00602 L
.44137 .00608 L
.44166 .00618 L
.44218 .00643 L
.44279 .00683 L
.44336 .0073 L
.44466 .00871 L
.44711 .01232 L
.44934 .01614 L
.45054 .01815 L
.45182 .02009 L
.45293 .0215 L
.45413 .02266 L
.45445 .0229 L
.45478 .02311 L
.45509 .02327 L
.45537 .02339 L
.45564 .02348 L
.4558 .02352 L
.45593 .02355 L
.45608 .02357 L
.45617 .02358 L
.45625 .02358 L
.45639 .02359 L
.45655 .02359 L
.45668 .02358 L
.4568 .02356 L
.45694 .02354 L
.45708 .02351 L
.45732 .02344 L
.45759 .02333 L
.45818 .02301 L
.45874 .02259 L
Mistroke
.46001 .02124 L
.4612 .01952 L
.46345 .01547 L
.46566 .01115 L
.46691 .00895 L
.46752 .00803 L
.46809 .00731 L
.46868 .00669 L
.46895 .00646 L
.46924 .00626 L
.4694 .00617 L
.46955 .0061 L
.4697 .00605 L
.46984 .00601 L
.46991 .00599 L
.47 .00597 L
.47008 .00596 L
.47017 .00595 L
.47025 .00595 L
.47033 .00596 L
.47048 .00597 L
.47057 .00599 L
.47065 .00601 L
.47081 .00606 L
.47098 .00614 L
.47117 .00625 L
.4715 .00651 L
.47181 .00684 L
.47239 .00766 L
.47302 .00887 L
.47362 .01037 L
.47418 .01207 L
.47544 .01709 L
.47667 .02358 L
.47799 .03228 L
.48057 .05458 L
.483 .08117 L
.48765 .14095 L
.49 .17158 L
.49259 .20174 L
.49377 .21355 L
.49505 .22438 L
.49624 .23256 L
.49735 .23827 L
.49794 .24055 L
.49857 .24236 L
.4989 .24304 L
.49921 .24352 L
.49938 .24373 L
.49954 .24387 L
Mistroke
.49963 .24393 L
.49972 .24398 L
.49982 .24402 L
.49991 .24404 L
.50006 .24404 L
.50014 .24403 L
.50022 .24401 L
.50037 .24394 L
.50052 .24382 L
.50066 .24369 L
.50079 .24354 L
.50108 .24309 L
.50139 .24244 L
.50168 .24171 L
.50233 .23957 L
.50291 .23712 L
.50352 .23395 L
.50462 .22683 L
.50712 .20476 L
.51214 .14378 L
.51698 .08146 L
.51919 .05696 L
.52163 .03516 L
.52279 .0269 L
.52403 .01964 L
.52535 .01376 L
.52658 .00985 L
.52721 .00838 L
.5278 .00735 L
.52811 .00693 L
.52845 .00656 L
.52862 .00641 L
.52881 .00627 L
.52898 .00616 L
.52914 .00609 L
.52928 .00603 L
.52943 .00599 L
.52952 .00597 L
.5296 .00596 L
.52976 .00595 L
.52984 .00595 L
.52992 .00596 L
.53007 .00598 L
.53015 .006 L
.53023 .00602 L
.53041 .00609 L
.5307 .00623 L
.53097 .0064 L
.53157 .00693 L
.53214 .00759 L
Mistroke
.53276 .00844 L
.53386 .01026 L
.53636 .01509 L
.53878 .01949 L
.54014 .02142 L
.54079 .02215 L
.54139 .0227 L
.54193 .02308 L
.54221 .02324 L
.54251 .02338 L
.54269 .02344 L
.54284 .02349 L
.54299 .02352 L
.54307 .02354 L
.54316 .02355 L
.54331 .02357 L
.54344 .02359 L
.5436 .02359 L
.54374 .02358 L
.5439 .02357 L
.54398 .02356 L
.54406 .02355 L
.54435 .02348 L
.54452 .02343 L
.5447 .02337 L
.54502 .02322 L
.54565 .02283 L
.54623 .02236 L
.5474 .0211 L
.5485 .01963 L
.55099 .01557 L
.55324 .01176 L
.55442 .00994 L
.55568 .00829 L
.55631 .00761 L
.55701 .00698 L
.55765 .00653 L
.55824 .00622 L
.55854 .00611 L
.5587 .00606 L
.55885 .00603 L
.55899 .006 L
.55912 .00598 L
.55926 .00596 L
.55941 .00595 L
.55949 .00595 L
.55957 .00595 L
.55972 .00596 L
.5598 .00597 L
.55988 .00597 L
Mistroke
.56006 .006 L
.56021 .00603 L
.56035 .00607 L
.56067 .00617 L
.56097 .0063 L
.56125 .00644 L
.56186 .00684 L
.56296 .00784 L
.56425 .00941 L
.56546 .0112 L
.57012 .01898 L
.57136 .02075 L
.57253 .02208 L
.57311 .02261 L
.57374 .02306 L
.5741 .02325 L
.57443 .0234 L
.57474 .02349 L
.57491 .02353 L
.57507 .02356 L
.57521 .02358 L
.57536 .02359 L
.57549 .02359 L
.57563 .02358 L
.57578 .02357 L
.57586 .02356 L
.57594 .02355 L
.57609 .02351 L
.57623 .02348 L
.5765 .02339 L
.57679 .02326 L
.57732 .02295 L
.5779 .02249 L
.57853 .02186 L
.57981 .0202 L
.58216 .01612 L
.58468 .01119 L
.5859 .00904 L
.58655 .00806 L
.58724 .00719 L
.58784 .0066 L
.58818 .00634 L
.58833 .00624 L
.5885 .00615 L
.58864 .00609 L
.58879 .00603 L
.58893 .00599 L
.58906 .00597 L
.58921 .00595 L
.58935 .00595 L
Mistroke
.58943 .00596 L
.58951 .00597 L
.58966 .006 L
.58981 .00605 L
.58995 .0061 L
.59025 .00628 L
.59053 .0065 L
.5908 .00677 L
.5914 .0076 L
.59206 .00885 L
.59269 .01044 L
.59337 .01257 L
.59459 .01756 L
.59583 .02424 L
.59718 .03331 L
.59961 .0545 L
.60456 .11262 L
.60932 .17495 L
.61172 .20258 L
.6139 .22285 L
.61511 .23145 L
.61625 .23764 L
.61688 .24018 L
.61746 .24197 L
.61776 .24269 L
.6181 .2433 L
.61827 .24354 L
.61845 .24375 L
.61853 .24383 L
.61862 .2439 L
.61878 .24399 L
.61893 .24404 L
.61901 .24405 L
.61909 .24405 L
.61923 .24402 L
.61938 .24396 L
.61951 .24387 L
.61964 .24376 L
.61993 .24341 L
.62024 .24288 L
.62052 .24225 L
.62116 .24038 L
.6217 .23826 L
.62229 .23543 L
.62337 .22892 L
.62575 .20897 L
.62828 .181 L
.63331 .1157 L
.63583 .08379 L
.63815 .05794 L
Mistroke
.64056 .0361 L
.6428 .02113 L
.6439 .01577 L
.64508 .0114 L
.64575 .00953 L
.64638 .00814 L
.64695 .0072 L
.64727 .00679 L
.64757 .00649 L
.64786 .00626 L
.64813 .00611 L
.64828 .00605 L
.64836 .00602 L
.64844 .006 L
.64859 .00597 L
.64874 .00595 L
.64888 .00595 L
.64903 .00597 L
.64918 .006 L
.64931 .00603 L
.64956 .00613 L
.64983 .00628 L
.65012 .00648 L
.65044 .00676 L
.65101 .00737 L
.65232 .00924 L
.6547 .01369 L
.6573 .01862 L
.65851 .02053 L
.65917 .02141 L
.65979 .02211 L
.66037 .02264 L
.66089 .02303 L
.66117 .02319 L
.66147 .02334 L
.66178 .02345 L
.66195 .0235 L
.6621 .02354 L
.66226 .02356 L
.66233 .02357 L
.66242 .02358 L
.66257 .02359 L
.66271 .02359 L
.6628 .02358 L
.66288 .02358 L
.66303 .02356 L
.66311 .02355 L
.66319 .02353 L
.66337 .02349 L
.66368 .02339 L
Mistroke
.66395 .02328 L
.66457 .02292 L
.66512 .02249 L
.66572 .02193 L
.66681 .02065 L
.67176 .01262 L
.67295 .01071 L
.67422 .00891 L
.67543 .00754 L
.67596 .00706 L
.67652 .00664 L
.6768 .00647 L
.67711 .0063 L
.67739 .00618 L
.67765 .00609 L
.6778 .00605 L
.67796 .00601 L
.67812 .00599 L
.6782 .00597 L
.67829 .00597 L
.67844 .00596 L
.67857 .00595 L
.67873 .00596 L
.67888 .00597 L
.67901 .00598 L
.67915 .00601 L
.67944 .00608 L
.6797 .00616 L
.67995 .00627 L
.68051 .00657 L
.68111 .007 L
.68233 .0082 L
.68366 .00993 L
.68609 .01382 L
.68867 .0182 L
.68988 .02004 L
.691 .02147 L
.69217 .02262 L
.69279 .02306 L
.6931 .02323 L
.69344 .02338 L
.69359 .02344 L
.69374 .02348 L
.69387 .02352 L
.69401 .02354 L
.69417 .02357 L
.69431 .02358 L
.69446 .02359 L
.69454 .02359 L
.69463 .02359 L
Mistroke
.69477 .02358 L
.6949 .02356 L
.69503 .02354 L
.69516 .02351 L
.69545 .02342 L
.69573 .02331 L
.69634 .02297 L
.697 .02244 L
.69821 .02111 L
.69937 .0194 L
.70045 .01754 L
.70291 .01276 L
.70412 .01047 L
.70521 .00863 L
.70576 .00785 L
.70635 .00712 L
.7069 .00658 L
.70715 .00638 L
.70742 .00621 L
.70757 .00614 L
.7077 .00608 L
.70784 .00603 L
.70797 .006 L
.70811 .00597 L
.70826 .00595 L
.70841 .00595 L
.70849 .00596 L
.70857 .00597 L
.70873 .006 L
.70887 .00605 L
.70903 .00612 L
.7092 .00621 L
.70948 .00641 L
.70979 .00671 L
.71042 .00755 L
.71102 .00866 L
.71155 .00993 L
.71214 .01163 L
.71344 .01666 L
.71466 .02291 L
.71714 .04056 L
.71938 .06192 L
.72422 .12073 L
.72887 .18118 L
.73124 .20744 L
.73249 .21905 L
.73383 .2293 L
.73449 .23346 L
.73512 .23678 L
.73573 .23944 L
Mistroke
.73629 .24137 L
.73683 .24272 L
.7371 .24323 L
.7374 .24364 L
.73755 .2438 L
.73769 .24391 L
.73785 .244 L
.73793 .24402 L
.73801 .24404 L
.73816 .24404 L
.73832 .24401 L
.73846 .24394 L
.73859 .24384 L
.73874 .2437 L
.73887 .24355 L
.73918 .24307 L
.73948 .24246 L
.7398 .24164 L
.74038 .23974 L
.74091 .23754 L
.7422 .23039 L
.74342 .2214 L
.74582 .19839 L
.74804 .17225 L
.75277 .11021 L
.75732 .05663 L
.75965 .03572 L
.76085 .02712 L
.76217 .01946 L
.76342 .01387 L
.76458 .01009 L
.76522 .00855 L
.76581 .00749 L
.76613 .00702 L
.76648 .00662 L
.76682 .00633 L
.76697 .00622 L
.76712 .00614 L
.76726 .00607 L
.76742 .00602 L
.76756 .00598 L
.76769 .00596 L
.76784 .00595 L
.76798 .00596 L
.76806 .00597 L
.76814 .00598 L
.76829 .00601 L
.76842 .00606 L
.76857 .00611 L
.76886 .00627 L
Mistroke
.76913 .00645 L
.76938 .00666 L
.77058 .00805 L
.77187 .01011 L
.77426 .0147 L
.77552 .01714 L
.77686 .01947 L
.77805 .02119 L
.77868 .02194 L
.77936 .02259 L
.77998 .02305 L
.78025 .02321 L
.78055 .02335 L
.78083 .02346 L
.7811 .02353 L
.78124 .02355 L
.78137 .02357 L
.78153 .02358 L
.78167 .02359 L
.78181 .02359 L
.78196 .02358 L
.7821 .02356 L
.78223 .02353 L
.78238 .0235 L
.78254 .02345 L
.78284 .02334 L
.78316 .0232 L
.78349 .023 L
.78408 .02257 L
.78525 .0214 L
.78631 .02003 L
.79115 .01205 L
.79337 .00878 L
.79398 .00806 L
.79463 .00739 L
.7952 .00691 L
.79581 .00649 L
.79615 .00631 L
.79646 .00618 L
.79676 .00608 L
.79691 .00604 L
.79707 .006 L
.79722 .00598 L
.79736 .00596 L
.79751 .00595 L
.79766 .00595 L
.7978 .00596 L
.79793 .00597 L
.79807 .00599 L
.79822 .00601 L
Mistroke
.79839 .00605 L
.79855 .0061 L
.79891 .00623 L
.79922 .00637 L
.79955 .00657 L
.80076 .00754 L
.8019 .00883 L
.80313 .01054 L
.80534 .01417 L
.80773 .01822 L
.80899 .02012 L
.81033 .02178 L
.8109 .02234 L
.81151 .02284 L
.81186 .02307 L
.81219 .02325 L
.8125 .02339 L
.81267 .02345 L
.81282 .02349 L
.81295 .02352 L
.8131 .02355 L
.81325 .02357 L
.81339 .02358 L
.81353 .02359 L
.81367 .02359 L
.81379 .02358 L
.81392 .02356 L
.81408 .02354 L
.81422 .02351 L
.81455 .02341 L
.81485 .02328 L
.81513 .02313 L
.81574 .02271 L
.81629 .02222 L
.81752 .02075 L
.81974 .01709 L
.82226 .01217 L
.82337 .01011 L
.82458 .00816 L
.82514 .00741 L
.82575 .00676 L
.82604 .0065 L
.82632 .0063 L
.82657 .00616 L
.8267 .0061 L
.82684 .00605 L
.82699 .006 L
.82713 .00597 L
.82721 .00596 L
.82729 .00595 L
Mistroke
.82744 .00595 L
.8276 .00597 L
.82775 .006 L
.82784 .00602 L
.82791 .00605 L
.82808 .00612 L
.82837 .00629 L
.82868 .00654 L
.82923 .0072 L
.82982 .00817 L
.83038 .00937 L
.83163 .01316 L
.83286 .01838 L
.83417 .02574 L
.83673 .0453 L
.83915 .06979 L
.84377 .12742 L
.84868 .19017 L
.85112 .21543 L
.85232 .22534 L
.85341 .23274 L
.85448 .23822 L
.85501 .24032 L
.8556 .24208 L
.85591 .24279 L
.85619 .2433 L
.85635 .24352 L
.85652 .24372 L
.85668 .24387 L
.85682 .24396 L
.85696 .24402 L
.85709 .24404 L
.85723 .24404 L
.85738 .244 L
.85753 .24392 L
.85768 .24381 L
.85782 .24367 L
.85796 .24349 L
.85827 .243 L
.8586 .24229 L
.85921 .24054 L
.85979 .23828 L
.86033 .23572 L
.86156 .22826 L
.86286 .21808 L
.86744 .16776 L
.87232 .10374 L
.87701 .05031 L
.87824 .0394 L
.87958 .02929 L
Mistroke
.88073 .02206 L
.882 .01574 L
.88259 .01337 L
.88323 .01122 L
.88378 .00968 L
.88439 .00832 L
.88503 .00723 L
.88538 .00678 L
.88571 .00645 L
.88602 .00623 L
.88616 .00614 L
.8863 .00608 L
.88645 .00602 L
.88653 .006 L
.88662 .00598 L
.88676 .00596 L
.88692 .00595 L
.88708 .00596 L
.88722 .00599 L
.88738 .00603 L
.88755 .00609 L
.88786 .00624 L
.88815 .00643 L
.88872 .00694 L
.88927 .00757 L
.89048 .00935 L
.89179 .01171 L
.89415 .01634 L
.8954 .01863 L
.89674 .02072 L
.8974 .02157 L
.89803 .02225 L
.89864 .02278 L
.89891 .02298 L
.89921 .02316 L
.89947 .02329 L
.89974 .02341 L
.8999 .02346 L
.90004 .0235 L
.90017 .02353 L
.90031 .02356 L
.90046 .02358 L
.90059 .02359 L
.90067 .02359 L
.90075 .02359 L
.90089 .02358 L
.90105 .02357 L
.90119 .02355 L
.90134 .02352 L
.90151 .02348 L
Mistroke
.90182 .02338 L
.90211 .02325 L
.90265 .02293 L
.90324 .02248 L
.90387 .02187 L
.90609 .01896 L
.90842 .01508 L
.911 .01078 L
.91217 .0091 L
.91344 .00762 L
.91402 .00709 L
.91464 .00663 L
.91516 .00633 L
.91546 .00619 L
.91573 .0061 L
.91589 .00605 L
.91605 .00602 L
.91619 .00599 L
.91634 .00597 L
.91643 .00596 L
.91651 .00596 L
.9166 .00595 L
.91668 .00595 L
.91683 .00596 L
.91692 .00596 L
.917 .00597 L
.91716 .00599 L
.9173 .00602 L
.91758 .00609 L
.91788 .00619 L
.9182 .00634 L
.91877 .00667 L
.91937 .00713 L
.92044 .00821 L
.92281 .01153 L
.92502 .01527 L
.92727 .019 L
.92846 .02069 L
.92972 .02213 L
.93035 .02269 L
.93069 .02293 L
.93105 .02315 L
.93138 .02332 L
.93168 .02344 L
.93198 .02352 L
.93212 .02355 L
.93228 .02357 L
.93236 .02358 L
.93245 .02358 L
.9326 .02359 L
Mistroke
.93275 .02359 L
.93289 .02357 L
.93304 .02355 L
.93313 .02354 L
.9332 .02352 L
.93354 .02343 L
.93383 .02331 L
.9341 .02317 L
.93471 .02277 L
.93529 .02227 L
.9359 .0216 L
.937 .02012 L
.93951 .01572 L
.94192 .011 L
.94319 .00881 L
.94389 .0078 L
.94453 .00702 L
.94511 .00649 L
.94544 .00626 L
.94574 .00611 L
.9459 .00604 L
.94606 .006 L
.94614 .00598 L
.94623 .00597 L
.94633 .00596 L
.94641 .00595 L
.9465 .00595 L
.94658 .00596 L
.94672 .00598 L
.94688 .00602 L
.94697 .00605 L
.94706 .00609 L
.94736 .00625 L
.94764 .00647 L
.94818 .00706 L
.94848 .00749 L
.94875 .00795 L
.94937 .00924 L
.95055 .01268 L
.95164 .01713 L
.95284 .02346 L
.95413 .03195 L
.9567 .05395 L
.95912 .08035 L
.96374 .13971 L
.96608 .17019 L
.96866 .2005 L
.96984 .21234 L
.9711 .22333 L
.9722 .23111 L
Mistroke
.9734 .23764 L
.9741 .24045 L
.97476 .24235 L
.97512 .2431 L
.97545 .24359 L
.97563 .24379 L
.97573 .24387 L
.97583 .24394 L
.97592 .24399 L
.97602 .24402 L
.97611 .24404 L
.97619 .24405 L
Mfstroke
0 0 m
1 0 L
1 .25 L
0 .25 L
closepath
clip
newpath
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{381.25, 305},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40008k0001bB000`400?l00000o`00003oo`3ooooo
0?ooocd0oooo003o0?ooool0oooo?@3oool00?l0ooooo`3ooolm0?ooo`00o`3ooolB0?ooo`030=nL
B0000000M;oo00<0oooo00@0_g@00000000007@0_oooo`3ooolQ0?ooo`00o`3ooolC0?ooo`090;md
0000M;l0oooo0?ooo`3ookl0M00004RLg`3ookl0M7Bo0?l0oooo8@3oool00?l0oooo503oool01P2o
M000M;oo0?ooo`3ookl0M00004RLgol0oooo8`3oool00?l0oooo5@3oool01@2oM000M;oo0?oog`2L
B00007Bo0?l0oooo8`3oool00?l0oooo5P3oool01@2oM4P0W=oo0;md0000B9`0gooo0?l0oooo8P3o
ool00?l0oooo5P3oool01@3ooml0W4Q809bLM00004P0W=oo0?l0oooo8P3oool00?l0oooo5`3oool0
103ooml0W4P000000018W=oo0?ooob80oooo003o0?oooaP0oooo00@0oono07@00000B9`0goooo`3o
oolQ0?ooo`00o`3ooolI0?ooo`030=nLB00007@0_ooo0?l0oooo8@3oool00?l0oooo6@3oool00`3o
_g@0001809cOo`0L0?ooo`050=nLB000000000000000001d_ol0303oool01@3OW4P0000000000000
0000M;oo0>h0oooo003o0?oooaT0oooo00<0okmd0000001d_ol0703oool00`3OW4P0001d0;ooo`0@
0?ooo`030?oo_`1d0000M;oo0>h0oooo003o0?oooaL0oooo00D0oooO0=oOo`3o_g@0000007Boo`0L
0?ooo`030=nLB000B9`0gooo0100oooo00<0oono07@0001d_ol0kP3oool00?l0oooo5`3oool01@3o
oml0W4P000000000B9`0gooo01`0oooo00<0gia80018W03Oool0403oool00`3ookl0M00007Boo`3^
0?ooo`00o`3ooolO0?ooo`040?oo_`1dM;l0oooo0;md00<0000000`0M;oo0?ooo`3oool0oooo0;md
00000000001809cOo`3oool0gia80018W03Oool20?ooo`070;md0000000007Bo0?ooo`3OW4P004RL
0=ooo`020?ooo`050?oog`2LB00000000000001d_ol02@3oool00`3ookl0M00007Boo`3^0?ooo`00
o`3ooolP0?ooo`030;md00000000M;oo0080oooo00<0okmd0000001d_ol00P3oool0203OW4P004RL
0=oog`2LM9`0gooo0?noM00007@0_ooo0P3oool01`3OW4P004RL0=noM01d_ol0gia80018W03Oool0
0P3oool01P3ookl0M01809cOo`3oool0_g@0001d_`P0oooo00<0oono07@0001d_ol0kP3oool00?l0
oooo803oool00`3OW4P0B9cO0?ooo`030?ooo`070;md0000B9`0gooo0?ooo`3o_g@0001809cOo`02
0?ooo`030?oo_`1d0000B9cO0080oooo00L0omnL04P0B02Lgol0oooo0=nLB000B9`0gooo0080oooo
00<0oono07@0B02Lgol00P3oool00`2oM00007Bo0?ooo`060?ooo`030?oo_`1d0000M;oo0>h0oooo
003o0?ooob00oooo00<0omnL09boo`3oool00`3oool01`2oM000001d0;ooo`3oool0oono07@00018
W=l00P3oool0303ooml0W4P00000M02oool0oooo0?oog`2LB000B9cO0?ooo`3OW4P004RL0=ooo`80
oooo00<0oooO09a80018W=l00P3oool00`3ooml0W4P004RLg`060?ooo`030?oo_`1d0000M;oo0>h0
oooo003o0?ooo`l0oooo00<0oono07@000000000100000000`0005L0E`0000000005000000030000
M02oool0oooo00@0oooo00@0omnL04P00000B9`0gooo0P3oool00`2oM00004RL0=ooo`020?ooo`03
0=nLB000B4P0B7Bo0080oooo00H0_g@0001d_`3oool0gia80018W03Oool30?ooo`030=nLB00007@0
_ooo0080oooo00<0omnL04P0001d_ol01@3oool00`3ookl0M00007Boo`3^0?ooo`00o`3ooolT0?oo
o`040?oOW0180000001809cOo`<0oooo00<0gia80000B02Lgol00P3oool02`3o_g@0001d0;noM018
W=l0oooo0?noM00007@0_ooo0=nLB000B9`0gooo00<0oooo00<0oono07@00018W=l00`3oool00`3O
W4P0001d0;ooo`040?ooo`030?oo_`1d0000M;oo0>h0oooo003o0?ooob<0oooo00@0okmd00000000
07@0_ooo103oool00`3ogi`0B00004RLg`020?ooo`0;0?oog`2LB000M;oo0=nLB000M;l0omnL04P0
B02Lgol0gia80018W03Oool0103oool00`3OW4P00000000000040000000304RLg`3oool0oooo0080
oooo00<0oono07@0001d_ol0kP3oool00?l0oooo8P3oool0103o_g@000000018W03Oool40?ooo`0A
0?noM000000000000018W03Oool0oono07@000000000B9cO0?ooo`3o_g@000000000B02Lgol0gia8
0018W03Oool0103oool00`3ooml0W4P0001d_`030?ooo`030=nLB00007@0_ooo00<0oooo00<0oono
07@0001d_ol0kP3oool00?l0oooo8P3oool00`2oM00004RL0=ooo`020?ooo`030?oog`2LM9`0gooo
00h0oooo00<0gia80018W03Oool01@3oool00`3ogi`0B00007Boo`020?ooo`030?oog`2LB000B9cO
00<0oooo00<0oono07@0001d_ol0kP3oool00?l0oooo8P3oool00`2oM000B9cO0?ooo`020?ooo`03
0?oog`2LB7@0_ooo00h0oooo00<0gia80018W03Oool01P3oool00`3o_g@0001d0;ooo`020?ooo`03
0?noM00007@0_ooo0080oooo00<0oono07@0001d_ol0kP3oool00?l0oooo8P3oool0203o_g@00018
09cOo`3oool0oooO09a800000000M;oo0P3oool0103ookl0M0000018W03Oool80?ooo`030=nLB000
B9`0gooo00L0oooo00L0gia80000M02oool0oooo0?oog`2LB4P0W=oo0080oooo00<0oono07@0001d
_ol0kP3oool00?l0oooo8`3oool01`3o_g@000000000000004P0W=oo0?oo_`2LW=l03P3oool00`3O
W4P0001d0;ooo`080?ooo`060?noM00004P0W=oo0?oog`2LB000M;oo0P3oool00`3ookl0M00007Bo
o`3^0?ooo`00o`3ooolB0?ooo`080?oo_`1d000000000000M02oool0gia800000018W=lN0?ooo`05
0=nLB000000000000000001d_ol01`3oool01P3ookl0M0000000000007@0_ooo0=nLB0<0000000<0
M;oo0?ooo`3oool0k03oool00?l0oooo4P3oool02@3OW4P004RL0=ooo`3ookl0M01d0;noM00007@0
W9ad09cOo`3o0?ooob80oooo003o0?oooa80oooo00P0_g@0001d_`3oool0oooo0?oo_`1dB4P0B018
09cOool0oooo8`3oool00?l0oooo4P3oool00`3o_g@004RL0=ooo`020?ooo`030?noM0000000B9cO
0?l0oooo8`3oool00?l0oooo4P3oool00`3ookl0M00007Boo`020?ooo`030?oog`2LB00007Bo0?l0
oooo8`3oool00?l0oooo4`3oool00`3o_g@0001d0;ooo`020?ooo`030=nLB00007@0_ooo0?l0oooo
8P3oool00?l0oooo503oool01`3OW4P0001d0;ooo`3oool0_g@00000B02Lgol0o`3ooolR0?ooo`00
o`3ooolE0?ooo`060?oOW0180000000004RL_`2LB000B9cOo`3ooolR0?ooo`00o`3ooolI0?ooo`03
0;md0000M;l0oooo0?l0oooo8@3oool00?l0oooo6@3oool00`3o_g@0001d0;ooo`3o0?ooob40oooo
003o0?oooaT0oooo00<0omnL04P0B02Lgol0o`3ooolQ0?ooo`00o`3ooolI0?ooo`030?oog`2LB000
B9cO0?l0oooo8@3oool00?l0oooo603oool0103ookl0M00000000000M;oo0?ooob40oooo003o0?oo
ool0oooo?@3oool00?l0ooooo`3ooolm0?ooo`00o`3ooooo0?ooocd0oooo003o0?ooool0oooo?@3o
ool00?l0ooooo`3ooolm0?ooo`00o`3ooooo0?ooocd0oooo00100?ooool00000g@00000O0?ooo`00
@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400oooo00<0
00000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0
oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o0?oo
omT0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`03
0000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3oool00`000000oooo
0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3o
ool00400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?oo
o`030000003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000
oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0
o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3o
ool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<00000
0?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo
01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00
@03oool00`000000oooo0?ooo`1m0?ooo`030?oog`2LB000B9cO03D0oooo00<0gia8000000000000
0`0000000`18W=l0oooo0?ooo`1U0?ooo`030=nLB0000000000000<0000000<0B9cO0?ooo`3oool0
=@3oool00`3ooml0W4P004RLg`1d0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003o
ool0oooo07h0oooo00<0gia80018W03Oool0=@3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02o
ool0oooo06H0oooo00<0okmd001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`0f0?ooo`030=nLB000
B9`0gooo07<0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0OP3oool0
0`3ogi`0B01809cOo`0f0?ooo`030?noM000M;l0oooo06/0oooo00<0okmd001d_`3oool0>P3oool0
0`3ogi`0B01809cOo`1c0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo
07X0oooo00<0okmd000000000000100000000`00B9`0gooo0?ooo`0d0?ooo`030?noM000M;l0oooo
06/0oooo00<0okmd001d_`3oool0=@3oool00`3o_g@0000000000004000000030018W03Oool0oooo
0700oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0K`3oool00`3o_g@0
000000000006000000050000M02oool0oooO09a8B02Lgol00P3oool00`2oM00007Bo0?ooo`0X0?oo
o`030?noM0000000000000H0000000<0001d0;ooo`3oool00`3oool00`3o_g@0B9cO0?ooo`1[0?oo
o`030?noM018W=l0oooo03@0oooo00<0oooO09a8B02Lgol00P3oool00`2oM00007Bo0?ooo`1b0?oo
o`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo07/0oooo00L0oono07@0M02o
ool0oooo0?noM000B9`0gooo03P0oooo00<0gia804RLg`3oool0J`3oool00`3OW4P0B9cO0?ooo`0d
0?ooo`070?oo_`1d07@0_ooo0?ooo`3o_g@004RL0=ooo`1b0?ooo`030000003oool0oooo01d0oooo
00100?ooo`030000003oool0oooo07`0oooo00H0omnL04Q8W03Oool0omnL04P0B02Lgoli0?ooo`03
0;md001d_ol0oooo06/0oooo00<0_g@007Boo`3oool0=03oool01P3ogi`0B4RL0=ooo`3ogi`0B018
09cOog80oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0O@3oool01@3o
_g@007Bo0?oog`2LB000B9cO03T0oooo00<0omnL04P0M02oool0J`3oool00`3ogi`0B01d0;ooo`0e
0?ooo`050?noM000M;l0oooO09a80018W=l0LP3oool00`000000oooo0?ooo`0M0?ooo`00@03oool0
0`000000oooo0?ooo`1n0?ooo`040=nLB018W=l0_g@0001d_cT0oooo00<0oooO09a8001d_ol0J`3o
ool00`3ooml0W4P007Boo`0f0?ooo`040=nLB018W=l0_g@0001d_g80oooo00<000000?ooo`3oool0
7@3oool00400oooo00<000000?ooo`3oool0O`3oool0103OW4P0M7Ad0000M02ooolh0?ooo`030?oo
_`1d0000B9cO06/0oooo00<0oono07@00018W=l0=`3oool0103OW4P0M7Ad0000M02oooma0?ooo`03
0000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0800oooo00<0_g@00000B02Lgol0
=03oool01`3o_g@004RL0=ooo`3oool0gia80000B02Lgol0I`3oool01`3o_g@004RL0=ooo`3oool0
gia80000B02Lgol0>03oool00`2oM000001809cOo`1a0?ooo`030000003oool0oooo01d0oooo0010
0?ooo`030000003oool0oooo0800oooo00<0oooO09a80018W=l0=@3oool01@3ogi`0B00000000000
0000M;oo06T0oooo00D0omnL04P000000000000007Boo`0i0?ooo`030?oog`2LB000B9cO0740oooo
00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003o
ool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M
0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400
oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000
003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?oo
o`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0k@3oool0
0`000000oooo0?ooo`3X0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo
0>d0oooo00<000000?ooo`3oool0j03oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000
oooo0?ooo`1c0?ooooH00000K`3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo
0?ooo`1h0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0203oool01P000000oooo
0?ooo`3oool000000?ooo`8000002P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool01P3oool2000000@0oooo00<000000?ooo`3oool00`3oool01000
0000oooo0?ooo`0000030?ooo`040000003oool00000000000@0oooo00<000000?ooo`0000001@3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0000000<0oooo100000030?ooo`80000000<0
oooo0000000000001@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00P0oooo00H0
00000?ooo`3oool0oooo0000003oool2000000/0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`1c0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo09<0oooo
1P00000b0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool00P0000000`3oool00000
0?ooo`030?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`040000003oool0oooo000000<0oooo00@000000?ooo`3oool0oooo10000003
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0;`3oool600000900oooo00<00000
0?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0b`3oool00`000000oooo0?ooo`080?oo
o`040000003oool0oooo0?ooo`8000001P3oool300000080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`040000003oool0oooo0?ooo`<000001@3oool2000000@0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`360?ooo`030000003oool0oooo01d0oooo0010
0?ooo`030000003oool0oooo0</0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0:
0?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool0aP3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3<0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo0<H0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3o
ool0c03oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`370?ooo`030000003oool0oooo01d0oooo00100?ooo`030000
003oool0oooo0<`0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0@0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0@0?ooo`030000
003oool0oooo00D0oooo00<000000?ooo`3oool0a`3oool00`000000oooo0?ooo`0M0?ooo`00@03o
ool00`000000oooo0?ooo`3<0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3oool07@3o
ool00400oooo00<000000?ooo`3oool0c@3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`370?ooo`030000003oool0
oooo01d0oooo00100?ooo`030000003oool0oooo0<d0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`0B0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`0B
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0a`3oool00`000000oooo0?ooo`0M
0?ooo`00@03oool00`000000oooo0?ooo`3=0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool04P3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool04P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3oool07@3oool0
0400oooo00<000000?ooo`3oool0c@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0180oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`380?ooo`030000003oool0oooo01d0oooo00100?oo
o`030000003oool0oooo0<d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0B0?oo
o`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`0B0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`00
0000oooo0?ooo`3=0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool04`3oool01@00
0000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?oo
o`3oool0cP3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01<0oooo00D000000?oo
o`3oool0oooo000000030?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`380?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo
0<h0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0C0?ooo`050000003oool0oooo
0?ooo`0000000`3oool00`000000oooo0?ooo`0D0?ooo`050000003oool0oooo0?ooo`000000b`3o
ool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3>0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool04`3oool01@000000oooo0?ooo`3oool0000000<0oooo00<0
00000?ooo`3oool0503oool01@000000oooo0?ooo`3oool000000</0oooo00<000000?ooo`3oool0
7@3oool00400oooo00<000000?ooo`3oool0cP3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo01<0oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0oooo01@0oooo
00D000000?ooo`3oool0oooo0000003;0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000
003oool0oooo0<l0oooo00@000000?ooo`3oool000005P3oool01@000000oooo0?ooo`3oool00000
00<0oooo00<000000?ooo`3oool0503oool01@000000oooo0?ooo`3oool000000</0oooo00<00000
0?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0c`3oool010000000oooo0?ooo`00000F
0?ooo`050000003oool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`0D0?ooo`050000003o
ool0oooo0?ooo`000000b`3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?oo
o`3?0?ooo`040000003oool0oooo000001H0oooo00D000000?ooo`3oool0oooo000000030?ooo`03
0000003oool0oooo01D0oooo00<000000?ooo`000000c03oool00`000000oooo0?ooo`0M0?ooo`00
@03oool00`000000oooo0?ooo`3?0?ooo`040000003oool0oooo000001H0oooo00D000000?ooo`3o
ool0oooo000000030?ooo`030000003oool0oooo01D0oooo00<000000?ooo`000000c03oool00`00
0000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3@0?ooo`8000005`3oool01@000000
oooo0?ooo`3oool0000000<0oooo00<000000?ooo`3oool05@3oool300000<`0oooo00<000000?oo
o`3oool07@3oool00400oooo00<000000?ooo`3oool0d03oool2000001L0oooo00D000000?ooo`3o
ool0oooo000000030?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0b`3oool00`00
0000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3Y0?ooo`050000003oool0oooo0?oo
o`0000000`3oool00`000000oooo0?ooo`3T0?ooo`030000003oool0oooo01d0oooo00100?ooo`03
0000003oool0oooo0>T0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo
0>D0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0jP3oool010000000
oooo0?ooo`0000020?ooo`030000003oool0oooo0>D0oooo00<000000?ooo`3oool07@3oool00400
oooo00<000000?ooo`3oool0jP3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo
0>D0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0jP3oool010000000
oooo0?ooo`0000020?ooo`030000003oool0oooo0>D0oooo00<000000?ooo`3oool07@3oool00400
oooo00<000000?ooo`3oool0jP3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo
0>D0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0jP3oool010000000
oooo0?ooo`0000020?ooo`030000003oool0oooo0>D0oooo00<000000?ooo`3oool07@3oool00400
oooo00<000000?ooo`3oool0jP3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo
0>D0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0jP3oool010000000
oooo0?ooo`0000020?ooo`030000003oool0oooo0>D0oooo00<000000?ooo`3oool07@3oool00400
oooo00<000000?ooo`3oool0jP3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo
0>D0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0jP3oool010000000
oooo0?ooo`0000020?ooo`030000003oool0oooo0>D0oooo00<000000?ooo`3oool07@3oool00400
oooo00<000000?ooo`3oool0jP3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo
0>D0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0jP3oool010000000
oooo0?ooo`0000020?ooo`030000003oool0oooo0>D0oooo00<000000?ooo`3oool07@3oool00400
oooo00<000000?ooo`3oool0j`3oool00`000000oooo000000020?ooo`030000003oool0oooo0>D0
oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0j`3oool00`000000oooo
000000020?ooo`030000003oool0oooo0>D0oooo00<000000?ooo`3oool07@3oool00400oooo00<0
00000?ooo`3oool0j`3oool00`000000oooo000000020?ooo`030000003oool0oooo0>D0oooo00<0
00000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0j`3oool01@000000oooo0000003o
ool000000>P0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0j`3oool0
1@000000oooo0000003oool000000>P0oooo00<000000?ooo`3oool07@3oool00400oooo00<00000
0?ooo`3oool0j`3oool01@000000oooo0000003oool000000>P0oooo00<000000?ooo`3oool07@3o
ool00400oooo00<000000?ooo`3oool0j`3oool01@000000oooo0000003oool000000>P0oooo00<0
00000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0k03oool2000000030?ooo`000000
oooo0>L0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0k03oool20000
00030?ooo`000000oooo0>L0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3o
ool0k03oool2000000030?ooo`000000oooo0>L0oooo00<000000?ooo`3oool07@3oool00400oooo
00<000000?ooo`3oool0k03oool2000000030?ooo`000000oooo0>L0oooo00<000000?ooo`3oool0
7@3oool00400oooo00<000000?ooo`3oool0k03oool2000000030?ooo`000000oooo0>L0oooo00<0
00000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0k03oool2000000030?ooo`000000
oooo0>L0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0k03oool20000
00030?ooo`000000oooo0>L0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3o
ool0k03oool2000000030?ooo`000000oooo0>L0oooo00<000000?ooo`3oool07@3oool00400oooo
00<000000?ooo`3oool0k03oool2000000030?ooo`000000oooo0>L0oooo00<000000?ooo`3oool0
7@3oool00400oooo00<000000?ooo`3oool0k03oool300000>T0oooo00<000000?ooo`3oool07@3o
ool00400oooo00<000000?ooo`3oool0k03oool300000>T0oooo00<000000?ooo`3oool07@3oool0
0400oooo00<000000?ooo`3oool0k03oool300000>T0oooo00<000000?ooo`3oool07@3oool00400
oooo00<000000?ooo`3oool0k@3oool200000>T0oooo00<000000?ooo`3oool07@3oool00400oooo
00<000000?ooo`3oool0k@3oool200000>T0oooo00<000000?ooo`3oool07@3oool00400oooo00<0
00000?ooo`3oool0k@3oool00`000000oooo0?ooo`3X0?ooo`030000003oool0oooo01d0oooo0010
0?ooo`030000003oool0oooo0>d0oooo00<000000?ooo`3oool0j03oool00`000000oooo0?ooo`0M
0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400
oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000
003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?oo
o`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0d@3oool0
2@3ookl0M00000000018W=l0oooo0?oo_`1d0000000004RLg`0d0?ooo`030?oog`2LB00000000080
000000<0M;oo0?ooo`3oool0`P3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo
0?ooo`3A0?ooo`0:0?oo_`1d04P0W=oo0;md0000M;l0oono07@0B02Lgol0_g@0001d_cD0oooo00<0
okmd0000M02oool0a@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3D
0?ooo`030?oo_`1d0000M;oo0080oooo00<0oono07@0001d_ol0=03oool00`3ookl0M00007Boo`35
0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0=D0oooo00<0gia80018
W03Oool00P3oool00`3OW4P004RL0=ooo`0c0?ooo`030?oog`2LB000B9cO0<D0oooo00<000000?oo
o`3oool07@3oool00400oooo00<000000?ooo`3oool0V@3oool00`3ooml0W4P004RLg`0>0?ooo`03
0?oo_`1d000000000080000000L0B9cO0?ooo`3oool0okmd0000000007@0_ooo0080oooo00T0_g@0
0000000004P0W=oo0?ooo`3o_g@000000000B02Lgol01@3oool0203ookl0M0000000000007@0_ooo
0=nLB0000000B9cO0P3oool0202oM00000000000B02Lgol0oooo0?noM00004P0W=oo0P3oool00`3o
_g@0001809cOo`020?ooo`0;0?oOW018000000000000001d_ol0gia800000000B9`0gooo0?ooo`3O
W4P00P000000200004P0W=oo0?ooo`3oool0okmd0000000004P0W=oo0P3oool0102oM00000000000
B02Lgol20?ooo`050?noM000000000000000B02Lgol00`3oool00`3OW4P004RL0=ooo`020?ooo`03
0;md0000000007Bo00D0oooo00<0gia80000000000000P0000001P1d_ol0oooo0?ooo`3oool0omnL
04P00080000000h0M;oo0=nLB000000004RL0=ooo`3o_g@000000000B02Lgol0oooo0?noM0000000
001809cOo`80oooo00D0gia8000000000000000007Boo`020?ooo`030;md0000M;l0oooo00@0oooo
00<0gia80018W03Oool00P3oool00`2oM0000000001d_`2D0?ooo`030000003oool0oooo01d0oooo
00100?ooo`030000003oool0oooo09X0oooo00<0gia80018W03Oool03@3oool0303ooml0W4P007Bo
o`3oool0omnL04P0001d_ol0oooo0?oOW01807@0_ono09bLg`80oooo00X0gia80018W03Ooml0W7BL
0=ooo`3o_g@0001809cOg`2LM9`0gooo103oool04P3OW4P004RL0=ooo`3ookl0M01d0;noM00007@0
W9ad09cOo`3oool0gia80018W03Ooml0W7BL0=ooo`3ookl0M00004RLg`80oooo00<0oono07@00018
W=l00P3oool08`2oM00004RL0=ooo`3oool0_g@007AdM00007@0_mnL09boo`3o_g@0000007Boo`3o
ool0oooO09a80018W=l0oooo0?noM00004P0W=oO09adW03Oool0oooo0=nLB000B9`0gooO09adW03O
ool0omnL04P0B02Lgol0oooo0?oOW018000007Bo0080oooo00<0okmd0000M02oool00P3oool0103O
W4P004RL0=noM01d_ol40?ooo`070?noM00004P0W=oo0?ooo`3ooml0W4P0001d_`020?ooo`0F0;md
0000B9`0gooo0?ooo`2oM000M7Ad0000M02ogi`0W;oo0?noM00004P0W=oO09adW03Oool0okmd0000
B02Lgml0W7BL0=ooo`3ogi`0B00007Boo`80oooo00H0_g@0001d_`3oool0okmd0000M02oool40?oo
o`030?noM00007@0_ooo0080oooo00@0gia80018W03O_g@0M;ooT`3oool00`000000oooo0?ooo`0M
0?ooo`00@03oool00`000000oooo0?ooo`2J0?ooo`030?oOW01804P0W=oo00h0oooo0100_gAd0;oo
o`3oool0oono07@0001d_ol0oooo0?oo_`1d0000M;oo0=nLW02oool0oooo0?noM00004P0W=oo0P3o
ool00`3ookl0M00007Boo`060?ooo`080;md0000M;l0oooo0?ooo`3ookl0M4Q804P0B02Lgol20?oo
o`030?noM00004P0W=oo00<0oooo00<0_g@00018W03Oool00P3oool01P2oM00004RL0=ooo`3oool0
_g@0001d_`<0oooo00L0_g@00000001d_ol0oooo0=nLB000B9`0gooo00<0oooo00D0gmoO0?ooo`3o
okl0M00007Boo`030?ooo`030?noM00004P0W=oo0080oooo00<0gia80018W03Oool00P3oool01`3o
gi`0B00004RLg`3oool0oono07@00018W=l00P3oool00`3ogi`0B01809cOo`050?ooo`030?oo_`1d
0000M;oo0080oooo00H0oooO09a80018W=l0oooo0;md0000M;l30?ooo`070;md00000000M;oo0?oo
o`3ookl0M00007Boo`020?ooo`030?oo_`1d0000M;oo0080oooo00<0gia80018W03Oool00`3oool0
1@3OW9`0_ooo0?oOW0180000M;oo00@0oooo00<0oono07@00018W=l00P3oool00`3ogi`0B01809cO
o`2D0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo09H0oooo00<0okmd
000000000000100000000`00B9`0gooo0?ooo`0>0?ooo`030=nLB00007@0_ooo0080oooo00<0_g@0
04RLg`3oool00P3oool00`3ookl0M00004RLg`030?ooo`030;md0000M;l0oooo00D0oooo00<0okmd
0018W03Oool00P3oool00`3o_g@0000004RLg`020?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia8
0000M02oool00P3oool01`3OW4P0001d0;ooo`3oool0okmd0000M02oool00P3oool01`3ookl0M000
04RLg`3oool0okmd0000M02oool01P3oool00`2oM00007Bo0?ooo`020?ooo`030?oo_`1d0000B9cO
0080oooo00<0okmd0000M02oool00`3oool02`3OW4P0001d0;ooo`3ooml0W4P00000M02oool0oooo
0?oog`2LB000B9cO00D0oooo00<0oooO09a80018W=l00`3oool01P3ogi`0B01809cOo`3o_g@0001d
0;ooo`80oooo00<0oono07@00018W=l00P3oool00`2oM00007Bo0?ooo`020?ooo`050;md0000M;l0
oooo0?ooo`3o_g@00`0000002018W=l0oooo0?ooo`3oool0oooO09a80000B9`0gooo0`3oool0203o
oml0W4P00000M02oool0oooo0?oog`2LB000B9cOU03oool00`000000oooo0?ooo`0M0?ooo`00@03o
ool00`000000oooo0?ooo`2F0?ooo`030?oog`2LB4P0W=oo0080oooo00<0_g@0001d_`3oool00`3o
ool00`3o_g@0000000000006000000060000M02oool0oooo0?noM00004P0W=oo0`3oool00`3OW4P0
04RL0=ooo`030?ooo`030;md0000B9`0gooo0080oooo00<0gia80018W03Oool01@3oool00`3ookl0
M00007Boo`020?ooo`030?oog`2LB00007Bo00<0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00007Boo`020?ooo`070?oOW0180000M;oo0?ooo`3ookl0M00007Boo`030?ooo`060;md0000B9`0
gooo0?oo_`1d0000M;oo1P3oool00`3OW4P004RL0=ooo`030?ooo`070;md0000B9`0gooo0?ooo`3o
okl0M00007Boo`030?ooo`070?oOW01804P0W=oo0?ooo`3OW4P004Q804Qd_`020?ooo`030;md0000
M;l0oooo00D0oooo00<0gia80000M02oool00P3oool01P3ooml0W4P004RLg`3ookl0M00007Boo`<0
oooo00L0_g@00018W03Oool0oooo0=nLB000B9`0gooo0080oooo00/0gia80018W03Oool0oooo0?oo
_`1d04P0W=oo0?ooo`3OW4P0000004RLg`020?ooo`030;md0018M7@0M9cO00@0oooo00<0gia80018
B018M;l00P3oool00`2oM00007Bo0?ooo`2C0?ooo`030000003oool0oooo01d0oooo00100?ooo`03
0000003oool0oooo09L0oooo00L0oono07@0M02oool0oooo0?noM000B9`0gooo00h0oooo00X0omnL
04P0B02Lgol0oooo0=nLM02Lgol0oooo0?oOW01807@0_ooo0`3oool00`3OW4P0001809cOo`020?oo
o`030?oOW01804P0W=oo00H0oooo00<0okmd0000M02oool00P3oool00`3OW4P0001d0;ooo`020?oo
o`030=nLB00004P0W=oo0080oooo00<0oooO09a80018W=l00P3oool00`3ooml0W4P004RLg`020?oo
o`030?noM00007@0_ooo0080oooo00l0gia80000M02oool0oooo0=nLB00007@0_ooo0?ooo`3ookl0
M00007Boo`3oool0omnL04P0B02Lgol00`3oool00`3OW4P0001809cOo`020?ooo`030=nLB00007@0
_ooo0080oooo00`0oono07@0001d_ol0oooo0?noM00007@0_kmd04RLg`3oool0okmd0000M02oool5
0?ooo`040?noM000000004RL0=ooo`80oooo00H0gia8001d_`3oool0okmd0000M02oool20?ooo`07
0=nLB00007@0_ooo0?ooo`3ogi`0B01809cOo`020?ooo`030?oOW01804P0W=oo0080oooo00`0okmd
0018W03Oool0oooo0?oo_`1d0000B9cO0?ooo`3o_g@004RL0;nLB01d_ol30?ooo`080?noM00007@0
_kmd04RLg`3oool0okmd0000M02ooonC0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000
003oool0oooo09P0oooo00H0omnL04Q8W03Oool0omnL04P0B02Lgol>0?ooo`0:0?oOW01804P0W=oo
0?oog`2LB000M;oo0?ooo`3ookl0M00007Boo`<0oooo00<0omnL04P00018W=l00P3oool00`3ooml0
W4P004RLg`070?ooo`070=nLB00007@0_ooo0?ooo`2oM000001809cOo`020?ooo`030?oOW0180000
B9cO00<0oooo00<0_g@00018W03Oool00P3oool00`2oM00004RL0=ooo`020?ooo`070=nLB00007@0
_ooo0?ooo`2oM000001809cOo`020?ooo`0:0;md000007@0_ooo0?ooo`3OW4P0B9cO0?ooo`3ooml0
W4P004RLg`<0oooo00<0omnL04P00018W=l00`3oool0402oM000001d0;ooo`3oool0okmd0000M02o
ool0oooo0?oog`2LB000M;oo0=nLB000M;l0omnL04P0B02Lgol50?ooo`080?oo_`1d0000M9bL04P0
M02oool0oooO09a80000M;l20?ooo`0;0=nLB00007@0_ooo0?ooo`2oM000001809cOo`3oool0oooO
09a80018W=l00P3oool00`3ooml0W4P004RLg`030?ooo`0F0=nLB00007@0_ooo0?ooo`3OW4P004RL
0=ooo`3ookl0M01809cOo`2oM000M7Ad0018W03Oool0oooO09a8001d_ol0gia8001d_`3ogi`0B018
09cOoi<0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0V@3oool01@3o
_g@007Bo0?oog`2LB000B9cO00l0oooo00D0_g@0000000000000000004RLg`020?ooo`050;md0018
W=l0oooo0?ooo`3o_g@00P0000001000B9`0gooo0?oo_`1d00030000000307Boo`3oool0oooo00@0
oooo00P0omnL04P000000000B9bo09a80018W=l0oooo0?noM080000000@004RL0=ooo`3ookl0M000
0`0000000`1d_kl0M000000000020000000907Boo`3oool0oooo0?oOW0180000000004RLW01d0000
B9cO00<0oooo00L0omnL04P000000000001809cOo`3ookl0M00000<0000000H0M;oo0?noM0000000
00000018W03Oool30?ooo`0?0?noM000000000000000001d_ol0oooo0?oo_`1d0000000004RLg`3o
ool0okmd0000000004P0W=oo00@0oooo00T0omnL04P000000000B9cO0?oo_`1d000000000000M02o
ool00`3oool0203ogi`0B00000000018W9`0M00004RLg`3ookl0M0000`0000000`1d_kl0M0000000
00020000000507Boo`3oool0oooo0?oOW01800000P0000004@1d_ol0omnL04P000000000M;oo0?oo
o`2oM000001d0;oo_`1d0000000004RLg`3oool0okmd0000000004P0W=oo09<0oooo00<000000?oo
o`3oool07@3oool00400oooo00<000000?ooo`3oool0VP3oool0103OW4P0B9cO0;md0000M;lF0?oo
o`030=nLB000B9`0gooo00P0oooo00<0okmd0000M02oool02`3oool00`2oM00007Bo0?ooo`070?oo
o`030?noM00004P0W=oo0080oooo00<0okmd0000B02Lgol04`3oool00`3o_g@0001d0;ooo`0`0?oo
o`030?noM00007@0_ooo0080oooo00<0okmd0000M02oool0Z`3oool00`000000oooo0?ooo`0M0?oo
o`00@03oool00`000000oooo0?ooo`2K0?ooo`040=nLB01dM7@0001d0;oooaD0oooo00<0omnL04P0
M02oool02@3oool00`3OW7@0W=oo0?ooo`0:0?ooo`030?noM00007@0_ooo00L0oooo00<0oooO09a8
0000M;l00P3oool00`3ooml0W4P0001d_`0D0?ooo`030=nLM02Lgol0oooo0300oooo00<0giad09cO
o`3oool00P3oool00`3OW7@0W=oo0?ooo`2Z0?ooo`030000003oool0oooo01d0oooo00100?ooo`03
0000003oool0oooo09`0oooo00<0_g@00000B02Lgol05@3oool00`3ooml0W4P007Boo`030?ooo`04
0?oo_`1d000004RL0=ooo`l0oooo00<0omnL04P0B02Lgol00P3oool0103ookl0M0000018W03Oool2
0?ooo`030?noM00007@0_ooo0080oooo00<0okmd0000M02oool0603oool0103ookl0M0000018W03O
oooL0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo09`0oooo00<0oooO
09a80018W=l05P3oool00`3OW4P0B9cO0?ooo`0E0?ooo`030?oog`2LB000B9cO00T0oooo00X0gia8
0018W03Ooml0W4P004RLg`3OW4P004RL0=oog`2LB000B9cOm@3oool00`000000oooo0?ooo`0M0?oo
o`00@03oool00`000000oooo0?ooo`2c0?ooo`040?oog`2LB0000000001d_aD0oooo00@0oono07@0
0000000007Bo2P3oool02@3OW4P000000000B02Lgol0oooo0=nLB0000000001809cOo`3e0?ooo`03
0000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3oool00`000000oooo
0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3o
ool00400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?oo
o`030000003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000
oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0
o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3o
ool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<00000
0?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo
01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00
@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400oooo00<0
00000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0
oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o0?oo
omT0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`03
0000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3oool00`000000oooo
0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3o
ool00400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?oo
o`030000003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000
oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0
o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3o
ool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1j0?ooo`030?oog`2LB000
B9cO03D0oooo00<0gia80000000000000`0000000`18W=l0oooo0?ooo`1U0?ooo`030=nLB0000000
000000<0000000<0B9cO0?ooo`3oool0=03oool00`3ooml0W4P004RLg`1h0?ooo`030000003oool0
oooo01d0oooo00100?ooo`030000003oool0oooo07/0oooo00<0gia80018W03Oool0=@3oool00`3o
_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo06H0oooo00<0okmd001d_`3oool00P3oool00`2o
M7@0_ooo0?ooo`0e0?ooo`030=nLB000B9`0gooo07L0oooo00<000000?ooo`3oool07@3oool00400
oooo00<000000?ooo`3oool0N`3oool00`3ogi`0B01809cOo`0f0?ooo`030?noM000M;l0oooo06/0
oooo00<0okmd001d_`3oool0>@3oool00`3ogi`0B01809cOo`1g0?ooo`030000003oool0oooo01d0
oooo00100?ooo`030000003oool0oooo07L0oooo00<0okmd000000000000100000000`00B9`0gooo
0?ooo`0d0?ooo`030?noM000M;l0oooo06/0oooo00<0okmd001d_`3oool0=03oool00`3o_g@00000
00000004000000030018W03Oool0oooo07@0oooo00<000000?ooo`3oool07@3oool00400oooo00<0
00000?ooo`3oool0K03oool00`3o_g@0000000000006000000050000M02oool0oooO09a8B02Lgol0
0P3oool00`2oM00007Bo0?ooo`0X0?ooo`030?noM0000000000000H0000000<0001d0;ooo`3oool0
0`3oool00`3o_g@0B9cO0?ooo`1[0?ooo`030?noM018W=l0oooo03<0oooo00<0oooO09a8B02Lgol0
0P3oool00`2oM00007Bo0?ooo`1f0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003o
ool0oooo07P0oooo00L0oono07@0M02oool0oooo0?noM000B9`0gooo03P0oooo00<0gia804RLg`3o
ool0J`3oool00`3OW4P0B9cO0?ooo`0c0?ooo`070?oo_`1d07@0_ooo0?ooo`3o_g@004RL0=ooo`1f
0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo07T0oooo00H0omnL04Q8
W03Oool0omnL04P0B02Lgoli0?ooo`030;md001d_ol0oooo06/0oooo00<0_g@007Boo`3oool0<`3o
ool01P3ogi`0B4RL0=ooo`3ogi`0B01809cOogH0oooo00<000000?ooo`3oool07@3oool00400oooo
00<000000?ooo`3oool0NP3oool01@3o_g@007Bo0?oog`2LB000B9cO03T0oooo00<0omnL04P0M02o
ool0J`3oool00`3ogi`0B01d0;ooo`0d0?ooo`050?noM000M;l0oooO09a80018W=l0MP3oool00`00
0000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1k0?ooo`040=nLB018W=l0_g@0001d
_cT0oooo00<0oooO09a8001d_ol0J`3oool00`3ooml0W4P007Boo`0e0?ooo`040=nLB018W=l0_g@0
001d_gH0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0O03oool0103O
W4P0M7Ad0000M02ooolh0?ooo`030?oo_`1d0000B9cO06/0oooo00<0oono07@00018W=l0=P3oool0
103OW4P0M7Ad0000M02ooome0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0
oooo07d0oooo00<0_g@00000B02Lgol0=03oool01`3o_g@004RL0=ooo`3oool0gia80000B02Lgol0
I`3oool01`3o_g@004RL0=ooo`3oool0gia80000B02Lgol0=`3oool00`2oM000001809cOo`1e0?oo
o`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo07d0oooo00<0oooO09a80018
W=l0=@3oool01@3ogi`0B000000000000000M;oo06T0oooo00D0omnL04P000000000000007Boo`0h
0?ooo`030?oog`2LB000B9cO07D0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?oo
o`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0
oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo
00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003o
ool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M
0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400
oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000
003oool0oooo0>X0oooo00<000000?ooo`3oool0j`3oool00`000000oooo0?ooo`0M0?ooo`00@03o
ool00`000000oooo0?ooo`1_0?ooooL00000LP3oool00`000000oooo0?ooo`0M0?ooo`00@03oool0
0`000000oooo0?ooo`1e0?ooo`030000003oool0oooo00H0oooo2P00000:0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool01`0000050?ooo`030000003oool0oooo0080oooo0`000006
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo00@0oooo100000030?ooo`030000003oool0oooo00<0oooo1`000002
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0203oool:000000P0oooo00<00000
0?ooo`3oool0MP3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1e0?oo
o`030000003oool0oooo0100oooo2000000=0?ooo`L00000303oool00`000000oooo0?ooo`050?oo
o`800000<03oool00`000000oooo0?ooo`0^0?ooo`8000001`3oool00`000000oooo0?ooo`0:0?oo
o`L000003@3oool800000180oooo00<000000?ooo`3oool0MP3oool00`000000oooo0?ooo`0M0?oo
o`00@03oool00`000000oooo0?ooo`2@0?ooo`d000007@3oool2000002h0oooo00<000000?ooo`3o
ool0;@3oool00`000000oooo0?ooo`0K0?ooo`d00000T`3oool00`000000oooo0?ooo`0M0?ooo`00
@03oool00`000000oooo0?ooo`2l0?ooo`800000;03oool00`000000oooo0?ooo`0Z0?ooo`<00000
_P3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`2n0?ooo`030000003o
ool0oooo02T0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`2o0?ooo`030000003o
ool0oooo01d0oooo00100?ooo`030000003oool0oooo0;l0oooo00<000000?ooo`3oool0:03oool0
0`000000oooo0?ooo`0X0?ooo`030000003oool0oooo0<00oooo00<000000?ooo`3oool07@3oool0
0400oooo00<000000?ooo`3oool0`03oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo
02L0oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000
oooo0?ooo`310?ooo`8000009`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0<80
oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0`P3oool00`000000oooo
0?ooo`0U0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0``3oool00`000000oooo
0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`330?ooo`030000003oool0oooo02@0oooo00<0
00000?ooo`3oool0903oool00`000000oooo0?ooo`340?ooo`030000003oool0oooo01d0oooo0010
0?ooo`030000003oool0oooo0<@0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0S
0?ooo`030000003oool0oooo0<D0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?oo
o`3oool0a03oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0280oooo00<000000?oo
o`3oool0aP3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`350?ooo`03
0000003oool0oooo0280oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`370?ooo`03
0000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3oool0
8@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3oool0
7@3oool00400oooo00<000000?ooo`3oool0aP3oool00`000000oooo0?ooo`0Q0?ooo`030000003o
ool0oooo0200oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`0M0?ooo`00@03oool0
0`000000oooo0?ooo`370?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool07`3oool0
0`000000oooo0?ooo`390?ooo`030000003oool0oooo01d0oooo000`0?ooo`030=nLW02oool0oooo
00d0oooo00<000000?ooo`3oool0b03oool200000200oooo00<000000?ooo`3oool07`3oool00`00
0000oooo0?ooo`390?ooo`030000003oool0oooo01d0oooo000`0?ooo`030;mdM02oool0oooo00d0
oooo00<000000?ooo`3oool0b@3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01h0
oooo00<000000?ooo`3oool0bP3oool00`000000oooo0?ooo`0M0?ooo`00;P3oool0103ookl0M000
0000M02oool>0?ooo`030000003oool0oooo0<X0oooo00<000000?ooo`3oool07@3oool00`000000
oooo0?ooo`0M0?ooo`030000003oool0oooo0</0oooo00<000000?ooo`3oool07@3oool002@0oooo
00<0gibL0;ooo`3oool00`3oool00`3ookl0M00000000003000000030000M02oool0oooo00d0oooo
00<000000?ooo`3oool0bP3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo01`0oooo
00<000000?ooo`3oool0c03oool00`000000oooo0?ooo`0M0?ooo`00903oool0103OW4P0M;oo0?oo
_`1d00050000000507Boo`3oool0oooo0;mdM02oool03P3oool00`000000oooo0?ooo`3;0?ooo`03
0000003oool0oooo01`0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`3<0?ooo`03
0000003oool0oooo01d0oooo000T0?ooo`030=nLB000000000000080000000<0M;oo0?ooo`3oool0
503oool00`000000oooo0?ooo`3;0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0
6`3oool00`000000oooo0?ooo`3=0?ooo`030000003oool0oooo01d0oooo000T0?ooo`030=nLB000
B9`0gooo01T0oooo00<000000?ooo`3oool0c03oool00`000000oooo0?ooo`0K0?ooo`030000003o
ool0oooo01/0oooo00<000000?ooo`3oool0c@3oool00`000000oooo0?ooo`0M0?ooo`00903oool0
0`3OW9`0_ooo0?ooo`0I0?ooo`030000003oool0oooo0<`0oooo00<000000?ooo`3oool06`3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0<h0oooo00<000000?ooo`3oool07@3oool0
0380oooo00@0oooO09a8000007@0_ooo2P3oool00`000000oooo0?ooo`3=0?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`3?0?ooo`030000003oool0
oooo01d0oooo000_0?ooo`050?noM000000000000000B02Lgol0303oool00`000000oooo0?ooo`3>
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`3?
0?ooo`030000003oool0oooo01d0oooo000/0?ooo`040;md000000000000001d_a00oooo00<00000
0?ooo`3oool0c`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<00000
0?ooo`3oool0d03oool00`000000oooo0?ooo`0M0?ooo`00:03oool01@3ogi`0B000000000000000
M;oo01<0oooo00<000000?ooo`3oool0c`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0
oooo01P0oooo00<000000?ooo`3oool0d03oool00`000000oooo0?ooo`0M0?ooo`009@3oool01@3O
W4P0000000000000B9`0gooo01H0oooo00<000000?ooo`3oool0d03oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0d@3oool00`000000oooo0?ooo`0M
0?ooo`008`3oool00`3ogi`0B00004RLg`0:0?ooo`030=nLW02oool0oooo00d0oooo00<000000?oo
o`3oool0d03oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01H0oooo00<000000?oo
o`3oool0dP3oool00`000000oooo0?ooo`0M0?ooo`00<03oool00`2oM7@0_ooo0?ooo`0=0?ooo`03
0000003oool0oooo0=40oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0E0?ooo`03
0000003oool0oooo0=<0oooo00<000000?ooo`3oool07@3oool002h0oooo00@0oono07@0000007@0
_ooo3P3oool00`000000oooo0?ooo`3A0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3o
ool05@3oool00`000000oooo0?ooo`3C0?ooo`030000003oool0oooo01d0oooo000T0?ooo`030=nL
W02oool0oooo00<0oooo00<0oono07@0000000000`0000000`0007@0_ooo0?ooo`0=0?ooo`030000
003oool0oooo0=80oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0D0?ooo`030000
003oool0oooo0=@0oooo00<000000?ooo`3oool07@3oool002@0oooo00@0gia807Boo`3ookl0M000
1@0000001@1d_ol0oooo0?ooo`2oM7@0_ooo00h0oooo00<000000?ooo`3oool0dP3oool00`000000
oooo0?ooo`0E0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0e03oool00`000000
oooo0?ooo`0M0?ooo`00903oool00`3OW4P00000000000020000000307Boo`3oool0oooo01@0oooo
00<000000?ooo`3oool0d`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo01<0oooo
00<000000?ooo`3oool0e@3oool00`000000oooo0?ooo`0M0?ooo`00903oool00`3OW4P004RL0=oo
o`0I0?ooo`030000003oool0oooo0=<0oooo00<000000?ooo`3oool0503oool00`000000oooo0?oo
o`0C0?ooo`030000003oool0oooo0=D0oooo00<000000?ooo`3oool07@3oool002@0oooo00<0gibL
0;ooo`3oool06@3oool00`000000oooo0?ooo`3D0?ooo`030000003oool0oooo01<0oooo00<00000
0?ooo`3oool04P3oool00`000000oooo0?ooo`3F0?ooo`030000003oool0oooo01d0oooo000^0?oo
o`030?noM000000000000080000000<0M;oo0?ooo`3oool02P3oool00`000000oooo0?ooo`3E0?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`3F0?oo
o`030000003oool0oooo01d0oooo000/0?ooo`040?oOW0180000000004RLg`<0oooo00<0_gAd0;oo
o`3oool02P3oool00`000000oooo0?ooo`3F0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`3G0?ooo`030000003oool0oooo01d0oooo000[0?ooo`03
0?oOW01807@0_ooo00D0oooo00<0_gAd0;ooo`3oool02P3oool00`000000oooo0?ooo`3G0?ooo`03
0000003oool0oooo0100oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`3H0?ooo`03
0000003oool0oooo01d0oooo000[0?ooo`030=nLW02oool0oooo00@0oooo00<0_g@004RLg`3oool0
2`3oool00`000000oooo0?ooo`3G0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0
3`3oool00`000000oooo0?ooo`3I0?ooo`030000003oool0oooo01d0oooo000[0?ooo`080=nLM02L
gol0oooo0?ooo`3ookl0M00000000018W=l=0?ooo`030000003oool0oooo0=P0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0=T0oooo00<000000?oo
o`3oool07@3oool002/0oooo00<0oooO09a8000000000P0000000`00B9`0gooo0?ooo`0=0?ooo`03
0000003oool0oooo0=P0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo0=X0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0
f@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
fP3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3I0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`3K0?ooo`030000003o
ool0oooo01d0oooo00100?ooo`030000003oool0oooo0=X0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0=/0oooo00<000000?ooo`3oool07@3oool0
0400oooo00<000000?ooo`3oool0fP3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool0g03oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000
oooo0?ooo`3K0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`3M0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0=`0oooo
0P00000<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0g@3oool00`000000oooo
0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3M0?ooo`030000003oool0oooo00X0oooo00<0
00000?ooo`3oool02@3oool200000>00oooo00<000000?ooo`3oool07@3oool00400oooo00<00000
0?ooo`3oool0gP3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool0h03oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3O0?oo
o`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`3P0?oo
o`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0>00oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0>40oooo00<000000?ooo`3o
ool07@3oool00400oooo00<000000?ooo`3oool0h@3oool00`000000oooo0?ooo`060?ooo`030000
003oool0oooo00H0oooo00<000000?ooo`3oool0hP3oool00`000000oooo0?ooo`0M0?ooo`00@03o
ool00`000000oooo0?ooo`3R0?ooo`8000001P3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo0><0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0i03oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0i03oool0
0`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3U0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00P3oool200000>L0oooo00<000000?ooo`3oool07@3oool00400
oooo00<000000?ooo`3oool0iP3oool900000>T0oooo00<000000?ooo`3oool07@3oool00400oooo
00<000000?ooo`3oool0j@3oool300000>`0oooo00<000000?ooo`3oool07@3oool00400oooo00<0
00000?ooo`3oool0jP3oool00`000000oooo0?ooo`3[0?ooo`030000003oool0oooo01d0oooo0010
0?ooo`030000003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`00
0000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3o
ool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooo
f@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<0
00000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0L@3oool00`3o_g@0000000000002
000000030018W03Oool0oooo03l0oooo00T0oono07@000000000B9cO0?ooo`3ookl0M00000000018
W=l0?03oool00`3ooml0W4P0000000020000000307Boo`3oool0oooo0400oooo00<0oooO09a8B02L
gol0T03oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1`0?ooo`030?no
M0000000B9cO00<0oooo00<0omnL04P0B02Lgol0?`3oool02P3ookl0M01809cOo`2oM00007Bo0?oo
_`1d04P0W=oo0;md0000M;lm0?ooo`030?noM00007@0_ooo04@0oooo00<0gia804RLg`3oool0S`3o
ool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1`0?ooo`030?noM00007@0
_ooo00@0oooo00<0gia8001d_`3oool0@@3oool00`3ookl0M00007Boo`020?ooo`030?oo_`1d0000
M;oo03`0oooo00<0oono07@0001d_ol0<@3oool00`3ookl0M01d0;ooo`0@0?ooo`030?oOW018M;l0
oooo08l0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0L@3oool0202o
M000001d0;ooo`3oool0oooO09a800000000B9cO@`3oool00`3OW4P004RL0=ooo`020?ooo`030=nL
B000B9`0gooo03/0oooo00<0oooO09a80018W=l0<P3oool00`3ogi`0B01d0;ooo`0?0?ooo`030?oo
_`1d07@0_ooo08l0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0FP3o
ool00`3ookl0M000000000020000000<04RLg`3oool0oooo0?ooo`2oM00000000000B02Lgol0oooo
0=nLB000B9`0gooo0P3oool00`2oM0000000001d_`020?ooo`060?oo_`1d000000000000000007@0
_ooo0P3oool0103o_g@000000000M02oool20?ooo`050=nLB000000000000000001d_ol0303oool0
0`3ookl0M000000000020000000704RLg`3oool0oooo0?noM0000000001d0;ooo`020?ooo`090;md
00000000001809cOo`3oool0okmd0000000004P0W=oo00D0oooo00P0oono07@000000000001d0;oo
o`3OW4P0000004RLg`80oooo00P0_g@00000000004P0W=oo0?ooo`3o_g@0001809cOo`80oooo00<0
okmd0000B02Lgol00P3oool00`2oM00007Bo0?ooo`040?ooo`0;0?oOW018000000000000001d_ol0
gia800000000B9`0gooo0?ooo`3OW4P00P000000200004P0W=oo0?ooo`3oool0okmd0000000004P0
W=oo0P3oool0102oM00000000000B02Lgol20?ooo`050?noM000000000000000B02Lgol00`3oool0
0`3OW4P004RL0=ooo`020?ooo`030;md0000000007Bo00H0oooo00<0gia80000000000000P000000
1P1d_ol0oooo0?ooo`3oool0omnL04P00080000000h0M;oo0=nLB000000004RL0=ooo`3o_g@00000
0000B02Lgol0oooo0?noM0000000001809cOo`80oooo00D0gia8000000000000000007Boo`020?oo
o`030;md0000M;l0oooo00@0oooo00<0gia80018W03Oool00P3oool00`2oM0000000001d_`020?oo
o`030;md001d_ol0oooo00D0oooo00`0omnL04P000000000000007Boo`3OW4P000000018W03Oool0
oooo0;md001d_ol20?ooo`080?oo_`1d000000000000M02oool0gia800000018W=lA0?ooo`040;md
00000000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL
0=ooof80oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0FP3oool01`3o
oml0W4P007Boo`3oool0omnL04P0001d_ol00P3oool0203OW4P004RL0=oog`2LM9`0gooo0?noM000
07@0_ooo0P3oool0203OW4P004RL0=noM01d_ol0oooo0?noM00007@0_ooo1@3oool0203ogi`0B01d
0;oo_`2LW=l0oooo0?oOW0180000M;oo0P3oool00`2oM00007Bo0?ooo`0:0?ooo`0<0?oog`2LB000
M;oo0?ooo`3ogi`0B00007Boo`3oool0omnL04P0M02ookl0W9cO0P3oool02P3OW4P004RL0=oog`2L
M9`0gooo0?noM00004P0W=oO09adW03Oool40?ooo`0B0=nLB000B9`0gooo0?oo_`1d07@0_kmd0000
M02LW7@0W=oo0?ooo`3OW4P004RL0=oog`2LM9`0gooo0?oo_`1d0000B9cO0P3oool00`3ookl0M000
04RLg`020?ooo`030?noM00007@0_ooo00@0oooo02<0_g@00018W03Oool0oooo0;md001dM7@0001d
0;oOW02L_ol0okmd0000001d_ol0oooo0?oog`2LB000B9cO0?ooo`3o_g@0001809cOg`2LM9`0gooo
0?ooo`3OW4P004RL0=oog`2LM9`0gooo0?oOW01804P0W=oo0?ooo`3ogi`0B000001d_`020?ooo`03
0?noM00007@0_ooo0080oooo00@0gia80018W03O_g@0M;oo1@3oool01`3o_g@0001809cOo`3oool0
oooO09a80000M;l00P3oool05P2oM00004RL0=ooo`3oool0_g@007AdM00007@0_mnL09boo`3o_g@0
001809cOg`2LM9`0gooo0?noM00004P0W=oO09adW03Oool0omnL04P0001d_ol20?ooo`060;md0000
M;l0oooo0?noM00007@0_ooo103oool00`3o_g@0001d0;ooo`020?ooo`070=nLB000B9`0gkmd07Bo
o`3ogi`0B00007Boo`060?ooo`0<0;md0000B9`0gooo0?ooo`2oM000M7Ad0000M02ogi`0W;oo0?oo
o`3o_g@007Bo0P3oool02@3OW4P004RL0=ooo`3ookl0M01d0;noM00007@0W9ad09cOo`0?0?ooo`06
0=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000
M;oo0?ooo`3ogi`0B7BoHP3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?oo
o`1K0?ooo`060;mdM02oool0oooo0?oo_`1d0000M;oo0P3oool00`3o_g@0001809cOo`020?ooo`03
0?oo_`1d0000B9cO0080oooo00<0omnL04P0B02Lgol00`3oool00`2oM000M;oo0?ooo`040?ooo`08
0?oo_`1d0000M;oo0=nLW02oool0gia80018W03Oool30?ooo`030=nLW02oool0oooo00X0oooo0100
_gAd0;ooo`3oool0oono07@0001d_ol0oooo0?oo_`1d0000M;oo0=nLW02oool0oooo0?noM00004P0
W=oo0P3oool00`3ookl0M00007Boo`060?ooo`080;md0000M;l0oooo0?ooo`3ookl0M4Q804P0B02L
gol20?ooo`030?noM00004P0W=oo00<0oooo00<0_g@00018W03Oool00P3oool01`2oM00004RL0=oo
o`3oool0omnL04P0001d_ol0103oool00`2oM00007Bo0?ooo`020?ooo`070;md00000000M;oo0?oo
o`3OW4P004RL0=ooo`030?ooo`050=oOg`3oool0oono07@0001d_ol00`3oool00`3o_g@0001809cO
o`020?ooo`030=nLB000B9`0gooo0080oooo00L0omnL04P00018W=l0oooo0?oo_`1d0000B9cO0080
oooo00<0omnL04P0B02Lgol01P3oool00`3ookl0M00007Boo`020?ooo`060?oog`2LB000B9cO0?oo
o`2oM00007Bo0`3oool01`2oM000000007Boo`3oool0oono07@0001d_ol00P3oool00`3ookl0M000
07Boo`020?ooo`030=nLB000B9`0gooo00<0oooo00D0gibL0;ooo`3ogi`0B00007Boo`040?ooo`03
0?oo_`1d0000B9cO0080oooo00<0omnL04P0B02Lgol02P3oool00`2oM00007Bo0?ooo`020?ooo`03
0;md00000000M;oo0080oooo00H0oono07A8W03Oool0oooo0;md0000M;l20?ooo`040?oo_`1dB4P0
B01809cOo`<0oooo00<0oono07@0000000001`0000001@1d_ol0oooo0?oog`2LB000B9cO0080oooo
00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`1P0?ooo`030000003oool0oooo01d0oooo
00100?ooo`030000003oool0oooo05h0oooo00<0gia80000M02oool00P3oool00`3ookl0M00004RL
g`020?ooo`080?oog`2LB000001d0;ooo`3oool0oooO09a80018W=l20?ooo`030?oo_`1d00000000
0080000000H0M;oo0?ooo`3oool0oooo0;md0018W=l20?ooo`050?noM0000000000000000018W=l0
4@3oool00`3OW4P0001d0;ooo`020?ooo`030;md0018W=l0oooo0080oooo00<0oono07@00018W=l0
0`3oool00`2oM00007Bo0?ooo`050?ooo`030?noM000B9`0gooo0080oooo00<0okmd00000018W=l0
0P3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo0080oooo00P0gia80000M02oool0
oooo0?oog`2LB00004RL0=ooo`<0oooo00<0okmd0000M02oool00P3oool01`3ookl0M00004RLg`3o
ool0okmd0000M02oool01P3oool00`2oM00007Bo0?ooo`020?ooo`030?oo_`1d0000B9cO0080oooo
00<0okmd0000M02oool00`3oool02`3OW4P0001d0;ooo`3ooml0W4P00000M02oool0oooo0?oog`2L
B000B9cO00H0oooo00<0oooO09a80018W=l00`3oool01P3ogi`0B01809cOo`3o_g@0001d0;ooo`80
oooo00<0oono07@00018W=l00P3oool00`2oM00007Bo0?ooo`020?ooo`050;md0000M;l0oooo0?oo
o`3o_g@00`0000002018W=l0oooo0?ooo`3oool0oooO09a80000B9`0gooo0`3oool0203ooml0W4P0
0000M02oool0oooo0?oog`2LB000B9cO2P3oool00`3o_g@0001d0;ooo`020?ooo`030?oo_`1d0000
B9cO0080oooo00L0oooO09a8B02Lgol0oooo0?noM000B9`0gooo0080oooo00<0okmd00000018W=l0
3`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02o
ool0H03oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1M0?ooo`030?no
M00004P0W=oo00@0oooo00<0_g@00018W03Oool00P3oool00`3OW4P004Q804Qd_`020?ooo`0;0;md
0000M;l0oooo0?ooo`2oM00004RL0=ooo`3oool0omnL04P00018W=l00P3oool02`3OW4P004RL0=oo
o`3oool0oono07@0B02Lgol0oooo0=nLB0000000B9cO0080oooo00<0okmd0000000000001P000000
1P0007@0_ooo0?ooo`3o_g@0001809cOo`<0oooo00<0gia80018W03Oool00`3oool00`2oM00004RL
0=ooo`020?ooo`030=nLB000B9`0gooo00D0oooo00<0oono07@0001d_ol00P3oool00`3ooml0W4P0
001d_`030?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P0001d_ol00P3oool00`3ogi`0B000
07Boo`020?ooo`030;md0018M7@0M9cO00<0oooo00<0oono07@0001d_ol00`3oool01P2oM00004RL
0=ooo`3ookl0M00007Boo`H0oooo00<0gia80018W03Oool00`3oool01`2oM00004RL0=ooo`3oool0
oono07@0001d_ol00`3oool01`3ogi`0B01809cOo`3oool0gia80018B018M;l00P3oool00`2oM000
07Bo0?ooo`060?ooo`030=nLB00007@0_ooo0080oooo00H0oooO09a80018W=l0oono07@0001d_ol3
0?ooo`070;md0000B9`0gooo0?ooo`3OW4P004RL0=ooo`020?ooo`0;0=nLB000B9`0gooo0?ooo`3o
okl0M01809cOo`3oool0gia800000018W=l00P3oool00`2oM000B7Ad07BLg`040?ooo`030=nLB000
B4P0B7Bo0080oooo00<0_g@0001d_`3oool02@3oool00`3ookl0M00007Boo`030?ooo`030;md0000
B9`0gooo0080oooo00H0gia804RLg`3oool0oono07@0001d_ol20?ooo`030?oog`2LB00007Bo00l0
oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo0600
oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0G03oool01P3ogi`0B018
09cOo`3oool0giad09cOo`80oooo00<0gia80000B02Lgol00P3oool02`3o_g@0001d0;noM018W=l0
oooo0?noM00007@0_ooo0?ooo`2oM00007Bo00<0oooo00L0gia80000M02oool0oooo0?oOW01807@0
_ooo0080oooo00L0okmd0018W03Oool0oooo0?oo_`1d0000B9cO00`0oooo00X0omnL04P0B02Lgol0
oooo0=nLM02Lgol0oooo0?oOW01807@0_ooo0`3oool00`3OW4P0001809cOo`020?ooo`030?oOW018
04P0W=oo00H0oooo00<0okmd0000M02oool00P3oool00`3OW4P0001d0;ooo`020?ooo`030=nLB000
04P0W=oo0080oooo00<0oooO09a80018W=l00P3oool00`3ooml0W4P004RLg`020?ooo`040?noM000
B9`0_ia807Boo`<0oooo00<0okmd0000M02oool00P3oool03`3OW4P0001d0;ooo`3oool0gia80000
M02oool0oooo0?oo_`1d0000M;oo0?ooo`3ogi`0B01809cOo`030?ooo`030=nLB00004P0W=oo0080
oooo00<0gia80000M02oool00P3oool0303ookl0M00007Boo`3oool0okmd0000M02o_g@0B9cO0?oo
o`3o_g@0001d0;ooo`H0oooo00@0okmd00000000B9`0gooo0P3oool01P3OW4P007Bo0?ooo`3o_g@0
001d0;ooo`80oooo00L0gia80000M02oool0oooo0?oOW01804P0W=oo0080oooo00<0omnL04P0B02L
gol00P3oool0303o_g@004RL0=ooo`3oool0oono07@00018W=l0oooo0?noM000B9`0_ia807Boo`<0
oooo00P0okmd0000M02o_g@0B9cO0?ooo`3o_g@0001d0;ooo`X0oooo00<0okmd0000M02oool00P3o
ool00`3OW4P0001d0;ooo`020?ooo`070?oOW018B9`0gooo0?ooo`3o_g@0001d0;ooo`020?ooo`03
0=nLB00007@0_ooo00l0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`0:0?ooo`03
0?oo_`1d0000B9cO0600oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0
G03oool01P3ogi`0B01809cOo`3ooml0W4P007Boo`80oooo00<0omnL04P00018W=l00P3oool0303o
oml0W4P007Boo`3OW4P007Bo0?oOW01804P0W=oo0?ooo`3ogi`0B00004RLg`80oooo00L0gia80000
M02oool0oooo0?oo_`1d0000M;oo00<0oooo00L0gia80000M02oool0oooo0=nLB000B9`0gooo00/0
oooo00X0omnL04P0B02Lgol0oooO09a8001d_ol0oooo0?oo_`1d0000M;oo0`3oool00`3ogi`0B000
04RLg`020?ooo`030?oog`2LB000B9cO00L0oooo00L0gia80000M02oool0oooo0;md000004P0W=oo
0080oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`020?ooo`0C0;md0000B9`0gooo
0?ooo`3ookl0M01809cOo`2oM000M7Ad0018W03Oool0oooo0=nLB00007@0_ooo0?ooo`2oM0000018
09cOo`020?ooo`0:0;md000007@0_ooo0?ooo`3OW4P0B9cO0?ooo`3ooml0W4P004RLg`<0oooo00<0
omnL04P00018W=l00`3oool0402oM000001d0;ooo`3oool0okmd0000M02oool0oooo0?oog`2LB000
M;oo0=nLB000M;l0omnL04P0B02Lgol60?ooo`080?oo_`1d0000M9bL04P0M02oool0oooO09a80000
M;l20?ooo`0;0=nLB00007@0_ooo0?ooo`2oM000001809cOo`3oool0oooO09a80018W=l00P3oool0
0`3ooml0W4P004RLg`030?ooo`0F0=nLB00007@0_ooo0?ooo`3OW4P004RL0=ooo`3ookl0M01809cO
o`2oM000M7Ad0018W03Oool0oooO09a8001d_ol0gia8001d_`3ogi`0B01809cOo`/0oooo00L0gia8
0000M02oool0oooo0;md000004P0W=oo0080oooo00<0oooO09a8M02oool00P3oool01`3OW4P0001d
0;ooo`3oool0_g@00000B02Lgol0103oool00`2oM00000000000000600000005001d_`3oool0gia8
0000B02Lgol00P3oool00`3ooml0W4P0001d_`0:0?ooo`030?noM0000000M;oo0600oooo00<00000
0?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0G@3oool01`2oM0000000000000000000
B9cO0?ooo`3o_g@00P0000002`00B9`0gooo0?oo_`1d0000000004RLg`3oool0okmd0000000004P0
W=oo0080oooo00<0oono07@000000000100000001018W=l0oooo0;md0018W=l40?ooo`050?oOW018
000000000000001d_ol03@3oool01@2oM0000000000000000000B9cO0080oooo00D0_g@004RLg`3o
ool0oooo0?noM002000000040018W03Oool0oono07@000<0000000<0M;oo0?ooo`3oool0103oool0
203ogi`0B00000000018W;l0W4P004RLg`3oool0okmd0P0000001000B9`0gooo0?oo_`1d00030000
000307Bo_`1d000000000080000000P0M9bL04P000000000M;oo0?ooo`2oM000001d0;ooo`80oooo
00H0omnL04P000000000B9bL07@00018W=l30?ooo`070?oOW018000000000000B02Lgol0oono07@0
00030000000607Boo`3o_g@0000000000000B9`0gooo0`3oool03`3o_g@00000000000000000M;oo
0?ooo`3ookl0M00000000018W=l0oooo0?noM0000000001809cOo`050?ooo`090?oOW01800000000
04RLg`3ookl0M0000000000007@0_ooo00<0oooo00P0omnL04P000000000B9bL07@00018W=l0oono
07@000<0000000<0M;no07@0000000000P0000001@1d_ol0oooo0?ooo`3ogi`0B000008000000140
M;oo0?oOW0180000000007Boo`3oool0_g@00000M02ookl0M00000000018W=l0oooo0?noM0000000
001809cOo`0<0?ooo`060?oOW0180000000004RLW01d0000B9cO0`3oool00`2oM000M;oo0?ooo`02
0?ooo`060?oOW0180000000004RL_`2LB000B9cO3`3oool00`3ogi`0B00004RLg`030?ooo`030;md
0000B9`0gooo00H0oooo00H0omoO0=ooo`3o_g@000000000M02ooomP0?ooo`030000003oool0oooo
01d0oooo00100?ooo`030000003oool0oooo07`0oooo00<0gia80018W03Oool0703oool00`3OW4P0
04RL0=ooo`080?ooo`030?noM00007@0_ooo00/0oooo00<0_g@0001d_`3oool01`3oool00`3o_g@0
001809cOo`020?ooo`030?noM00004P0W=oo01X0oooo00<0okmd0000M02oool0<@3oool00`3o_g@0
001d0;ooo`020?ooo`030?noM00007@0_ooo02d0oooo00<0okmd001d_`3oool01P3oool00`2oM000
07Bo0?ooo`0>0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool01@3o
okl0M0000000000004P0W=oo0640oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?oo
o`3oool0O03oool00`3ogi`0B01d0;ooo`0L0?ooo`030?oOW01807@0_ooo00T0oooo00<0giad09cO
o`3oool02P3oool00`3o_g@0001d0;ooo`070?ooo`030?oog`2LB00007Bo0080oooo00<0oooO09a8
0000M;l06`3oool00`3OW7@0W=oo0?ooo`0a0?ooo`030=nLM02Lgol0oooo0080oooo00<0giad09cO
o`3oool0;03oool00`3ookl0M01d0;ooo`060?ooo`030?noM00007@0_ooo00l0oooo00<0omnL04P0
001d_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo0680oooo00<000000?oo
o`3oool07@3oool00400oooo00<000000?ooo`3oool0I@3oool0103ookl0M0000018W03OoolC0?oo
o`030?oog`2LB000M;oo01`0oooo00<0oooO09a8001d_ol00`3oool0103ookl0M0000018W03Oool?
0?ooo`030?oOW01804P0W=oo0080oooo00@0oono07@00000B9`0gooo0P3oool00`3o_g@0001d0;oo
o`020?ooo`030?noM00007@0_ooo01l0oooo00@0oono07@00000B9`0goooH03oool00`2oM000M;oo
0?ooo`050?ooo`030?oOW01804P0W=oo0100oooo00H0gia80000M02oool0oooO09a80018W=l80?oo
o`030?oOW01dW=l0oooo0680oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3o
ool0O@3oool00`3OW4P0B9cO0?ooo`0L0?ooo`030=nLB018W=l0oooo01D0oooo00<0oooO09a80018
W=l02@3oool02P3OW4P004RL0=oog`2LB000B9cO0=nLB000B9`0gooO09a80018W=n00?ooo`030=nL
B018W=l0oooo00D0oooo00<0oooO09a80018W=l04@3oool0103OW4P0000000000000M;l:0?ooo`05
0;md00000000000000000018W=l0G`3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000
oooo0?ooo`1k0?ooo`040?oog`2LB0000000001d_a/0oooo00@0oooO09a80000000007Bo5@3oool0
103ookl0M00000000000M;l:0?ooo`090=nLB0000000001809cOo`3oool0gia80000000004P0W=oo
0800oooo00<0omnL04Q8W03Oool0103oool0103ookl0M00000000000M;n30?ooo`030000003oool0
oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?oo
o`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400oooo
00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003o
ool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o
0?ooomT0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0o`3ooooI0?oo
o`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3oool00`000000
oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool0
7@3oool00400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo0010
0?ooo`030000003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`00
0000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3o
ool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooo
f@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<0
00000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0
oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?oo
o`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400oooo
00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003o
ool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o
0?ooomT0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0o`3ooooI0?oo
o`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3oool00`000000
oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1m0?ooo`030?oog`2LB000B9cO03D0oooo
00<0gia80000000000000`0000000`18W=l0oooo0?ooo`1U0?ooo`030=nLB0000000000000<00000
00<0B9cO0?ooo`3oool0=@3oool00`3ooml0W4P004RLg`1d0?ooo`030000003oool0oooo01d0oooo
00100?ooo`030000003oool0oooo07h0oooo00<0gia80018W03Oool0=@3oool00`3o_g@007Bo0?oo
o`020?ooo`030;mdM02oool0oooo06H0oooo00<0okmd001d_`3oool00P3oool00`2oM7@0_ooo0?oo
o`0f0?ooo`030=nLB000B9`0gooo07<0oooo00<000000?ooo`3oool07@3oool00400oooo00<00000
0?ooo`3oool0OP3oool00`3ogi`0B01809cOo`0f0?ooo`030?noM000M;l0oooo06/0oooo00<0okmd
001d_`3oool0>P3oool00`3ogi`0B01809cOo`1c0?ooo`030000003oool0oooo01d0oooo00100?oo
o`030000003oool0oooo07X0oooo00<0okmd000000000000100000000`00B9`0gooo0?ooo`0d0?oo
o`030?noM000M;l0oooo06/0oooo00<0okmd001d_`3oool0=@3oool00`3o_g@00000000000040000
00030018W03Oool0oooo0700oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3o
ool0K`3oool00`3o_g@0000000000006000000050000M02oool0oooO09a8B02Lgol00P3oool00`2o
M00007Bo0?ooo`0X0?ooo`030?noM0000000000000H0000000<0001d0;ooo`3oool00`3oool00`3o
_g@0B9cO0?ooo`1[0?ooo`030?noM018W=l0oooo03@0oooo00<0oooO09a8B02Lgol00P3oool00`2o
M00007Bo0?ooo`1b0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo07/0
oooo00L0oono07@0M02oool0oooo0?noM000B9`0gooo03P0oooo00<0gia804RLg`3oool0J`3oool0
0`3OW4P0B9cO0?ooo`0d0?ooo`070?oo_`1d07@0_ooo0?ooo`3o_g@004RL0=ooo`1b0?ooo`030000
003oool0oooo01d0oooo00100?ooo`030000003oool0oooo07`0oooo00H0omnL04Q8W03Oool0omnL
04P0B02Lgoli0?ooo`030;md001d_ol0oooo06/0oooo00<0_g@007Boo`3oool0=03oool01P3ogi`0
B4RL0=ooo`3ogi`0B01809cOog80oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?oo
o`3oool0O@3oool01@3o_g@007Bo0?oog`2LB000B9cO03T0oooo00<0omnL04P0M02oool0J`3oool0
0`3ogi`0B01d0;ooo`0e0?ooo`050?noM000M;l0oooO09a80018W=l0LP3oool00`000000oooo0?oo
o`0M0?ooo`00@03oool00`000000oooo0?ooo`1n0?ooo`040=nLB018W=l0_g@0001d_cT0oooo00<0
oooO09a8001d_ol0J`3oool00`3ooml0W4P007Boo`0f0?ooo`040=nLB018W=l0_g@0001d_g80oooo
00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0O`3oool0103OW4P0M7Ad0000
M02ooolh0?ooo`030?oo_`1d0000B9cO06/0oooo00<0oono07@00018W=l0=`3oool0103OW4P0M7Ad
0000M02oooma0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0800oooo
00<0_g@00000B02Lgol0=03oool01`3o_g@004RL0=ooo`3oool0gia80000B02Lgol0I`3oool01`3o
_g@004RL0=ooo`3oool0gia80000B02Lgol0>03oool00`2oM000001809cOo`1a0?ooo`030000003o
ool0oooo01d0oooo00100?ooo`030000003oool0oooo0800oooo00<0oooO09a80018W=l0=@3oool0
1@3ogi`0B000000000000000M;oo06T0oooo00D0omnL04P000000000000007Boo`0i0?ooo`030?oo
g`2LB000B9cO0740oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0o`3o
oooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3oool0
0`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?oo
o`3oool07@3oool00400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0
oooo00100?ooo`030000003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03o
ool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400oooo00<00000
0?ooo`3oool0k@3oool00`000000oooo0?ooo`3X0?ooo`030000003oool0oooo01d0oooo00100?oo
o`030000003oool0oooo0800oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?oo
o`8000003@3oool00`000000oooo0?ooo`040?ooo`8000001P3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00@0oooo0P0000060?ooo`030000003oool0oooo00/0oooo0P0000060?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0<0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool2000000d0oooo00<000000?oo
o`3oool0103oool2000000H0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`0M0?oo
o`00@03oool00`000000oooo0?ooo`1c0?ooooH00000K`3oool00`000000oooo0?ooo`0M0?ooo`00
@03oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`000000
0`3oool2000000030?ooo`000000000000<0oooo00<000000?ooo`000000303oool00`000000oooo
000000040?ooo`040000003oool0oooo000000<0oooo00@000000?ooo`00000000001@3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0000000@0oooo00@000000?ooo`3oool000000`3oool2
000000030?ooo`000000oooo00X0oooo100000030?ooo`040000003oool0oooo000000@0oooo00<0
00000?ooo`0000001@3oool00`000000oooo0?ooo`040?ooo`030000003oool0000000<0oooo0P00
00000`3oool000000?ooo`030?ooo`030000003oool0000000`0oooo00<000000?ooo`0000000`3o
ool2000000030?ooo`000000000000<0oooo00<000000?ooo`0000001P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0000000@0oooo00@000000?ooo`3oool000000`3oool2000000030?oo
o`000000oooo00X0oooo00<000000?ooo`000000103oool010000000oooo0?ooo`0000040?ooo`03
0000003oool0000000D0oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`0M0?ooo`00
@03oool00`000000oooo0?ooo`1n0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3o
ool00P3oool010000000oooo0?ooo`0000030?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`0000000`3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo00000080oooo00<00000
0?ooo`3oool02@3oool00`000000oooo000000030?ooo`030000003oool0oooo0080oooo00@00000
0?ooo`3oool000000P3oool00`000000oooo0?ooo`090?ooo`040000003oool0oooo000000<0oooo
00D000000?ooo`3oool0oooo000000020?ooo`040000003oool0oooo000000D0oooo00<000000?oo
o`3oool00`3oool010000000oooo0?ooo`0000030?ooo`040000003oool0oooo000000<0oooo00@0
00000?ooo`3oool00000303oool00`000000oooo000000020?ooo`030000003oool0oooo0080oooo
00@000000?ooo`3oool000000`3oool00`000000oooo0?ooo`090?ooo`040000003oool0oooo0000
0080oooo00D000000?ooo`3oool0oooo000000030?ooo`040000003oool0oooo000000/0oooo00@0
00000?ooo`3oool000000P3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo00000080
oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo
0?ooo`1n0?ooo`070000003oool0oooo0?ooo`000000oooo000000050?ooo`030000003oool00000
00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`03000000D0oooo00<000000?oo
o`0000000`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00@00000103oool40000
0080oooo00<000000?ooo`3oool02@3oool01`000000oooo0?ooo`3oool000000?ooo`000000103o
ool400000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`070000003o
ool0oooo0?ooo`000000oooo000000040?ooo`80000000D0oooo0000003oool0oooo0000000;0?oo
o`070000003oool0oooo0?ooo`000000oooo000000050?ooo`030000003oool0000000<0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`04000000@0oooo00<000000?ooo`0000000`3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00@00000103oool00`000000oooo0000
00030?ooo`030000003oool0oooo07T0oooo00<000000?ooo`3oool07@3oool00400oooo00<00000
0?ooo`3oool0OP3oool010000000oooo0?ooo`3oool3000000D0oooo0`0000030?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool00P3oool2000000H0oooo0P0000030?ooo`030000003o
ool0oooo00T0oooo00@000000?ooo`3oool0oooo0P0000060?ooo`8000000`3oool00`000000oooo
0?ooo`090?ooo`040000003oool0oooo0?ooo`8000001P3oool2000000@0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo0?ooo`<000001@3oool20000
00@0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`020?ooo`8000001@3oool30000
00<0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`020?ooo`8000001P3oool20000
00<0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`020?ooo`8000001P3oool20000
00<0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000
oooo0?ooo`1n0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`090?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`080?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0N@3oool00`000000
oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1n0?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool0N03oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?oo
o`1n0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0N03oool00`000000oooo0?oo
o`0M0?ooo`00@03oool00`000000oooo0?ooo`1n0?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool01`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool0403oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool0403oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool0N03oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1m0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0@0?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo0140oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0
oooo0100oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo0100oooo00<000000?ooo`3oool0N03oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`00
0000oooo0?ooo`1m0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0403oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo0140oooo00D000000?ooo`3oool0oooo00000004
0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0N03oool00`000000oooo0?ooo`0M
0?ooo`00@03oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0140oooo00D000000?oo
o`3oool0oooo000000040?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0N03oool0
0`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo
0140oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo0140oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool04@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool0N03oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1m0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo0140oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0
oooo00H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo0100oooo00<000000?ooo`3oool0N03oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`00
0000oooo0?ooo`1m0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo0140oooo00D000000?ooo`3oool0oooo00000004
0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0N03oool00`000000oooo0?ooo`0M
0?ooo`00@03oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0140oooo00D000000?oo
o`3oool0oooo000000030?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0M`3oool0
0`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
00H0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo
0140oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool04@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0140oooo00<00000
0?ooo`3oool0M`3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1m0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo0140oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0
oooo0180oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0
oooo00H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`060?ooo`030000003oool0
oooo0180oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`00
0000oooo0?ooo`1m0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`0B0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool04P3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo0180oooo00D000000?ooo`3oool0oooo00000003
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`06
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0M
0?ooo`00@03oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool04P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0180oooo00D000000?oo
o`3oool0oooo000000030?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0M`3oool0
0`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
0180oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0oooo01<0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool04`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool0M`3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1m0?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0C0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo0180oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0
oooo01<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo0180oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`00
0000oooo0?ooo`1m0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`0C0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo0180oooo00D000000?ooo`3oool0oooo00000003
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0B
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0M
0?ooo`00@03oool00`000000oooo0?ooo`1l0?ooo`030000003oool0oooo01<0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0503oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0180oooo00D000000?oo
o`3oool0oooo000000030?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0M`3oool0
0`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1l0?ooo`030000003oool0oooo
01<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0180oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0oooo01<0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool04`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool0M`3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1l0?oo
o`030000003oool0oooo01<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0C0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo0180oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0
oooo01<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0C0?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo0180oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`00
0000oooo0?ooo`1l0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`0D0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo01<0oooo00@000000?ooo`3oool000000P3oool0
0`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3oool07@3oool0
0400oooo00<000000?ooo`3oool0O03oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0C0?ooo`040000003oool0oooo
00000080oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo01<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0D0?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`1f0?ooo`030000003oool0
oooo01d0oooo00100?ooo`030000003oool0oooo07`0oooo00<000000?ooo`3oool04`3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0D0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool04`3oool01000
0000oooo0?ooo`0000020?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0503oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0MP3oool0
0`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1l0?ooo`030000003oool0oooo
01<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01@0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo01@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0
oooo07H0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0O03oool00`00
0000oooo0?ooo`0C0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0D0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool0503oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0503oool0
0`000000oooo0?ooo`1f0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo
07`0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
01@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool0503oool010000000oooo0?ooo`0000020?ooo`030000003oool0
oooo01@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0C0?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo01@0oooo00<000000?ooo`3oool0MP3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`00
0000oooo0?ooo`1l0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`0D0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo01@0oooo00@000000?ooo`3oool000000P3oool0
0`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0D0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3oool07@3oool0
0400oooo00<000000?ooo`3oool0O03oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
01D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0D0?ooo`040000003oool0oooo
00000080oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo01<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`1f0?ooo`030000003oool0
oooo01d0oooo00100?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0503oool01000
0000oooo0?ooo`0000020?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0MP3oool0
0`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo
01D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01@0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo01D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0
oooo07H0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0N`3oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0D0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`1f0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo
07/0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
01H0oooo00D000000?ooo`3oool0oooo0000000H0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool05@3oool00`000000oooo000000020?ooo`030000003oool0oooo01D0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool05@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01H0oooo00<00000
0?ooo`3oool0M@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1k0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0F0?oo
o`050000003oool0oooo0?ooo`000000603oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo01D0oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo01H0oooo00D000000?ooo`3oool0oooo0000000H0?ooo`030000003oool0oooo07D0oooo00<0
00000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0N`3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool05P3oool01@000000oooo0?ooo`3o
ool0000001P0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0E0?ooo`030000003o
ool000000080oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo01D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0F0?ooo`050000003o
ool0oooo0?ooo`000000603oool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo01d0oooo
00100?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo01H0oooo00D000000?ooo`3oool0oooo0000000H0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool05@3oool00`000000oooo000000020?ooo`030000
003oool0oooo01D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0E0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool05P3oool01@000000oooo0?ooo`3oool0000001P0
oooo00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo
0?ooo`1k0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0F0?ooo`050000003oool0oooo0?ooo`000000603oool01@000000oooo0?ooo`3oool00000
01P0oooo00D000000?ooo`000000oooo0000000H0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool05@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01H0oooo00D00000
0?ooo`3oool0oooo0000000H0?ooo`030000003oool0oooo07D0oooo00<000000?ooo`3oool07@3o
ool00400oooo00<000000?ooo`3oool0N`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool05P3oool01@000000oooo0?ooo`3oool0000001P0oooo00D0
00000?ooo`3oool0oooo0000000H0?ooo`050000003oool000000?ooo`000000603oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0F0?ooo`050000003oool0oooo0?ooo`000000603oool00`000000oooo0?ooo`1e0?oo
o`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3o
ool05@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01H0oooo00D000000?ooo`3o
ool0oooo0000000H0?ooo`050000003oool0oooo0?ooo`000000603oool01@000000oooo0000003o
ool0000001P0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0E0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool05P3oool01@000000oooo0?ooo`3oool0000001P0oooo
00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?oo
o`1k0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0F0?ooo`050000003oool0oooo0?ooo`000000603oool01@000000oooo0?ooo`3oool0000001P0
oooo00D000000?ooo`000000oooo0000000H0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool05@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01H0oooo00D000000?oo
o`3oool0oooo0000000H0?ooo`030000003oool0oooo07D0oooo00<000000?ooo`3oool07@3oool0
0400oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`0F0?ooo`050000003oool0oooo
0?ooo`0000006@3oool01@000000oooo0?ooo`3oool0000001P0oooo00D000000?ooo`3oool0oooo
0000000H0?ooo`050000003oool000000?ooo`0000006@3oool01@000000oooo0?ooo`3oool00000
01L0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0F0?ooo`050000003oool0oooo
0?ooo`000000603oool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo01d0oooo00100?oo
o`030000003oool0oooo07X0oooo00<000000?ooo`3oool05`3oool010000000oooo0?ooo`00000I
0?ooo`050000003oool0oooo0?ooo`000000603oool01@000000oooo0?ooo`3oool0000001P0oooo
00D000000?ooo`000000oooo0000000I0?ooo`040000003oool0oooo000001P0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0F0?ooo`040000003oool0oooo000001T0oooo00<00000
0?ooo`3oool0M@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1j0?oo
o`030000003oool0oooo01L0oooo00@000000?ooo`3oool000006@3oool01@000000oooo0?ooo`3o
ool0000001P0oooo00D000000?ooo`3oool0oooo0000000H0?ooo`050000003oool000000?ooo`00
00006@3oool010000000oooo0?ooo`00000I0?ooo`050000003oool0oooo0?ooo`000000603oool0
10000000oooo0?ooo`00000I0?ooo`030000003oool0oooo07D0oooo00<000000?ooo`3oool07@3o
ool00400oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`0G0?ooo`040000003oool0
oooo000001T0oooo00D000000?ooo`3oool0oooo0000000I0?ooo`040000003oool0oooo000001P0
oooo00D000000?ooo`000000oooo0000000I0?ooo`040000003oool0oooo000001T0oooo00D00000
0?ooo`3oool0oooo0000000H0?ooo`040000003oool0oooo000001T0oooo00<000000?ooo`3oool0
M@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1j0?ooo`030000003o
ool0oooo01L0oooo00@000000?ooo`3oool000006@3oool01@000000oooo0?ooo`3oool0000001T0
oooo00@000000?ooo`3oool00000603oool01@000000oooo0000003oool0000001T0oooo00@00000
0?ooo`3oool000006@3oool01@000000oooo0?ooo`3oool0000001P0oooo00@000000?ooo`3oool0
00006P3oool00`000000oooo0?ooo`1d0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000
003oool0oooo07X0oooo00<000000?ooo`3oool05`3oool010000000oooo0?ooo`00000I0?ooo`04
0000003oool0oooo000001X0oooo00@000000?ooo`3oool000006@3oool2000000030?ooo`000000
oooo01P0oooo00@000000?ooo`3oool000006@3oool010000000oooo0?ooo`00000I0?ooo`040000
003oool0oooo000001X0oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`0M0?ooo`00
@03oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo01L0oooo00@000000?ooo`3oool0
00006@3oool010000000oooo0?ooo`00000J0?ooo`040000003oool0oooo000001T0oooo0`00000J
0?ooo`040000003oool0oooo000001T0oooo00@000000?ooo`3oool000006@3oool010000000oooo
0?ooo`00000J0?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3oool07@3oool00400oooo
00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0000001T0oooo
00@000000?ooo`3oool000006P3oool010000000oooo0?ooo`00000I0?ooo`<000006P3oool01000
0000oooo0?ooo`00000I0?ooo`040000003oool0oooo000001T0oooo00@000000?ooo`3oool00000
6P3oool00`000000oooo0?ooo`1d0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003o
ool0oooo07X0oooo00<000000?ooo`3oool0603oool00`000000oooo0000000J0?ooo`030000003o
ool0000001X0oooo00<000000?ooo`0000006P3oool3000001X0oooo00@000000?ooo`3oool00000
6@3oool010000000oooo0?ooo`00000J0?ooo`030000003oool0000001X0oooo00<000000?ooo`3o
ool0M03oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1j0?ooo`030000
003oool0oooo01P0oooo00<000000?ooo`0000006P3oool00`000000oooo0000000J0?ooo`030000
003oool0000001X0oooo0`00000J0?ooo`040000003oool0oooo000001X0oooo00<000000?ooo`00
00006P3oool00`000000oooo0000000J0?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3o
ool07@3oool00400oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0000001X0oooo00<000000?ooo`0000006P3oool00`000000oooo0000000J0?ooo`<00000
6`3oool2000001/0oooo00<000000?ooo`0000006P3oool00`000000oooo0000000J0?ooo`030000
003oool0oooo07@0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0N@3o
ool00`000000oooo0?ooo`0I0?ooo`8000006`3oool00`000000oooo0000000J0?ooo`030000003o
ool0000001X0oooo0`00000K0?ooo`8000006`3oool2000001/0oooo00<000000?ooo`0000006P3o
ool00`000000oooo0?ooo`1d0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0
oooo07T0oooo00<000000?ooo`3oool06@3oool2000001/0oooo0P00000K0?ooo`<000006`3oool2
000001/0oooo0P00000K0?ooo`8000006`3oool3000001/0oooo00<000000?ooo`3oool0L`3oool0
0`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`1h0?ooo`8000006`3oool20000
01`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`8000006`3oool20000
01/0oooo0P00000L0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0L`3oool00`00
0000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3]0?ooo`030000003oool0oooo0>P0
oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000
003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3oool00`000000oooo0?oo
o`0M0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool0
0400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`03
0000003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo
0?ooo`3T0?ooo`040?oo_`1d0000000004RLgcL0oooo00<0oooO09a8000000000P0000000`1d_ol0
oooo0?ooo`2a0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0>@0oooo
00D0oono07@0B02Lgol0_g@0001d_`0h0?ooo`030?noM00007@0_ooo0;@0oooo00<000000?ooo`3o
ool07@3oool00400oooo00<000000?ooo`3oool0i`3oool00`3ookl0M00007Boo`0g0?ooo`030?oo
_`1d0000M;oo0;@0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0j03o
ool00`3OW4P004RL0=ooo`0f0?ooo`030?oog`2LB000B9cO0;@0oooo00<000000?ooo`3oool07@3o
ool00400oooo00<000000?ooo`3oool0R03oool00`3ooml0W4P004RLg`0@0?ooo`050=nLB0000000
00000000B02Lgol0103oool01@3OW4P00000000000000000M;oo0080oooo01D0omnL04P000000000
000007Boo`3OW4P000000018W03Oool0gia80018W03Oool0oooo0?noM00007@0_ooo0?ooo`3ooml0
W4P00000001d_ol01P3oool0202oM00000000000B02Lgol0oooo0=nLB000B9`0gooo0P3oool0202o
M0000000001d_`3oool0okmd0000000004P0W=oo0P3oool01@3OW4P00000000000000000M;oo0080
oooo00<0_g@0001d_`3oool0103oool00`3o_g@0001809cOo`030?ooo`050=nLB000000000000000
001d_ol00P3oool00`2oM00007Bo0?ooo`050?ooo`050=nLB000000000000000001d_ol00P3oool0
0`3OW4P004RL0=ooo`020?ooo`030;md0000000007Bo0080oooo00D0gia8000000000000001809cO
o`030?ooo`050=nLB000000000000000001d_ol01`3oool00`3OW4P00000000000020000000607Bo
o`3oool0oooo0?ooo`3ogi`0B0000P0000003P1d_ol0gia800000000B9`0gooo0?noM00000000018
09cOo`3oool0okmd0000000004P0W=oo0P3oool01@3OW4P00000000000000000M;oo0080oooo00<0
_g@0001d_`3oool0103oool00`3OW4P004RL0=ooo`020?ooo`030;md0000000007Bo08<0oooo00<0
00000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0R@3oool00`3OW4P004RL0=ooo`0>
0?ooo`070?oOW01807@0_ooo0?ooo`3ookl0M00004RLg`020?ooo`030?oOW0180000M;oo0080oooo
01<0_g@0001d_`3oool0_g@00018W03Oool0oooo0;md001dM7@0001d0;oOW02L_ol0okmd0000B02L
gol0oooo0?oo_`1d0000M;oo0080oooo00@0_g@0001d_`2oM7@0_ooo1@3oool0203OW4P004RL0=oo
g`2LM9`0gooo0?noM00007@0_ooo0P3oool0303OW4P004RL0=noM01d_ol0okmd0000B02Lgml0W7BL
0=ooo`3ogi`0B00007Boo`80oooo00H0_g@0001d_`3oool0okmd0000M02oool40?ooo`030?oo_`1d
0000B9cO0080oooo00<0omnL04P0001d_ol00P3oool01P2oM00007Bo0?ooo`3o_g@0001d0;ooo`@0
oooo00<0omnL04P0001d_ol00P3oool01P2oM00007Bo0?ooo`3o_g@0001d0;ooo`80oooo00l0gia8
0018W03O_g@0M;oo0?noM0000000M;oo0?ooo`3ooml0W4P004RLg`3oool0omnL04P0001d_ol00P3o
ool00`2oM00007Bo0?ooo`050?ooo`070?noM00004P0W=oo0?ooo`3ooml0W4P0001d_`020?ooo`0F
0;md0000B9`0gooo0?ooo`2oM000M7Ad0000M02ogi`0W;oo0?noM00004P0W=oO09adW03Oool0okmd
0000B02Lgml0W7BL0=ooo`3ogi`0B00007Boo`80oooo00H0_g@0001d_`3oool0okmd0000M02oool4
0?ooo`030?noM00007@0_ooo0080oooo00@0gia80018W03O_g@0M;ooPP3oool00`000000oooo0?oo
o`0M0?ooo`00@03oool00`000000oooo0?ooo`290?ooo`030?oOW01804P0W=oo00h0oooo00<0oono
07@0001d_ol00P3oool01`3ookl0M00007Boo`3oool0gia80018W03Oool00`3oool0103OW9`0_ooo
0;md0000M;l30?ooo`0;0;md00000000M;oo0?ooo`3ookl0M00004RLg`3oool0oooO09a80000M;l0
0P3oool00`3OW4P004RL0=ooo`060?ooo`030?noM00004P0W=oo0080oooo00<0oono07@00018W=l0
0P3oool01`3ogi`0B01809cOo`3oool0oono07@0001d_ol00P3oool00`3OW4P004RL0=ooo`030?oo
o`050=nLW02oool0omnL04P0001d_ol01@3oool01`2oM00004RL0=ooo`3oool0gia80018W03Oool0
0`3oool01@3OW9`0_ooo0?oOW0180000M;oo00@0oooo00<0gia80018W03Oool00`3oool01@3OW9`0
_ooo0?oo_`1d0000B9cO0080oooo00L0omnL04P0B02Lgol0oooo0=nLB000B9`0gooo00<0oooo00D0
gmoO0?ooo`3OW4P004RL0=ooo`030?ooo`030=nLW02oool0oooo00@0oooo00<0oono07@0001d_ol0
0P3oool01P3ooml0W4P004RLg`3oool0_g@0001d_`<0oooo00L0_g@00000001d_ol0oooo0?oo_`1d
0000M;oo0080oooo00<0oono07@0001d_ol00P3oool00`3OW4P004RL0=ooo`030?ooo`050=nLW02o
ool0omnL04P0001d_ol0103oool00`3ookl0M00004RLg`020?ooo`030?oOW01804P0W=oo08<0oooo
00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0Q@3oool00`3o_g@000000000
0004000000030018W03Oool0oooo00`0oooo00<0_g@004RLg`3oool00P3oool0103OW4P004RL0=oo
o`3o_g@30000000704RLg`3oool0oooo0?ooo`3o_g@0001d0;ooo`020?ooo`0@0?oo_`1d0000B9cO
0?ooo`3ooml0W4P00018W03Oool0oooo0=nLB0000000M;oo0?ooo`3ogi`0B01809cOo`H0oooo00<0
oono07@00018W=l00P3oool0203ooml0W4P00000M02oool0oooo0?oog`2LB000B9cO0P3oool01@2o
M00007Bo0?ooo`3oool0okmd00<0000000P0B9cO0?ooo`3oool0oooo0?oog`2LB00004RL0=ooo`@0
oooo00D0gia80000M02oool0oooo0?noM0030000000804RLg`3oool0oooo0?ooo`3ooml0W4P00018
W03Oool30?ooo`050?noM0000000000000000018W=l00`3oool0303ooml0W4P00000M02oool0oooo
0?oog`2LB000B9cO0?ooo`3o_g@0001d0;ooo`D0oooo00D0okmd000000000000000004RLg`080?oo
o`030?oog`2LB000B9cO00<0oooo00H0omnL04P0B02Lgol0okmd0000M02oool20?ooo`030?oo_`1d
0000B9cO0080oooo00<0_g@0001d_`3oool00P3oool01@2oM00007Bo0?ooo`3oool0okmd00<00000
00P0B9cO0?ooo`3oool0oooo0?oog`2LB00004RL0=ooo`<0oooo00P0oooO09a8000007@0_ooo0?oo
o`3ooml0W4P004RLgh<0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0
Q@3oool00`3ooml0W4Q809cOo`020?ooo`030;md0000M;l0oooo00<0oooo00<0okmd000000000000
1P0000001@0007@0_ooo0=nLB000B9`0gooo0080oooo00h0omnL04P0B02Lgol0oono07@0B02Lgol0
oooo0=nLB0000000B9cO0?ooo`3ookl0M00007Boo`<0oooo00l0_g@00018W03Oool0oooo0=nLB000
B4P0M9cO0?ooo`3o_g@0001d07AdM02oool0oooO09a80018W=l01`3oool00`2oM00004RL0=ooo`02
0?ooo`030=nLB000B4P0B7Bo0080oooo00l0_g@0001d_`3oool0oooo0=nLB000B9`0gooo0?ooo`3o
okl0M01809cOo`3oool0gia800000018W=l00P3oool00`2oM000B7Ad07BLg`040?ooo`0;0?oOW018
0000M;oo0?ooo`3ookl0M01809cOo`3oool0gia800000018W=l00P3oool00`2oM000B7Ad07BLg`03
0?ooo`070?oo_`1d04P0W=oo0?ooo`3OW4P0000004RLg`020?ooo`030=nLB000B4P0B7Bo0080oooo
00H0_g@0001d_`3oool0oono07@0001d_ol50?ooo`070?oo_`1d04P0W=oo0?ooo`3OW4P0000004RL
g`070?ooo`030=nLB00007@0_ooo0080oooo00H0oooO09a80018W=l0oono07@0001d_ol30?ooo`07
0;md0000B9`0gooo0?ooo`3OW4P004RL0=ooo`020?ooo`0;0=nLB000B9`0gooo0?ooo`3ookl0M018
09cOo`3oool0gia800000018W=l00P3oool00`2oM000B7Ad07BLg`040?ooo`030=nLB000B4P0B7Bo
0080oooo00<0_g@0001d_`3oool0PP3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000
oooo0?ooo`260?ooo`070?oo_`1d07@0_ooo0?ooo`3o_g@004RL0=ooo`0>0?ooo`030?oOW01804P0
M;oO0080oooo00l0oono07@0001d_ol0oooo0?noM000B9`0gooo0?ooo`3ookl0M00004RLg`3oool0
okmd0000M02oool00P3oool0403OW4P0001d0;ooo`3oool0okmd0000M02LW7@0W=oo0?oo_`1d0000
M;no07Ad_`3oool0_g@00018W03Oool60?ooo`030=nLB00004P0W=oo0080oooo00`0okmd0000M02o
_g@0B9cO0?ooo`3o_g@0001d0;ooo`3oool0omnL04P0B02Lgol20?ooo`0<0?noM000B9`0gooo0?oo
o`3ookl0M00004RLg`3oool0okmd0018W02oW4P0M;oo0`3oool00`3ooml0W4P004RLg`020?ooo`0<
0?noM000B9`0gooo0?ooo`3ookl0M00004RLg`3oool0okmd0018W02oW4P0M;oo0`3oool0803o_g@0
04RL0=ooo`3oool0oono07@00018W=l0oooo0?noM00007@0_kmd04RLg`3oool0okmd0000M02oool0
oooo0=nLB00007@0_ooo0?ooo`3ookl0M00007Boo`3oool0okmd0018W03Oool0oooo0?oo_`1d0000
B9cO1P3oool0103o_g@000000018W03Oool20?ooo`060=nLB000M;l0oooo0?noM00007@0_ooo0P3o
ool01`3OW4P0001d0;ooo`3oool0omnL04P0B02Lgol00P3oool00`3ogi`0B01809cOo`020?ooo`0<
0?noM000B9`0gooo0?ooo`3ookl0M00004RLg`3oool0okmd0018W02oW4P0M;oo0`3oool0203o_g@0
001d0;noM018W=l0oooo0?noM00007@0_oooPP3oool00`000000oooo0?ooo`0M0?ooo`00@03oool0
0`000000oooo0?ooo`270?ooo`060?oOW018B9`0gooo0?oOW01804P0W=oo3P3oool0203ookl0M000
04Q8B018W=l0oooo0?noM00004P0W=oo0P3oool0703OW4P0001d0;ooo`3oool0gia80018W03Oool0
oooo0=nLB00007@0_ooo0?ooo`2oM000001809cOo`3oool0oono07@0B02Lgml0W4Q809cOg`2LB000
B9cO0?noM000M;l0okmd0000M02oool60?ooo`030?oOW0180000B9cO0080oooo00`0oooO09a8001d
_ol0gia8001d_`3ogi`0B01809cOo`3oool0oooO09a80018W=l30?ooo`0B0=nLB00007@0_ooo0?oo
o`3OW4P004RL0=ooo`3ookl0M01809cOo`2oM000M7Ad0018W03Oool0oooo0;md0000B9`0gooo0P3o
ool07P3OW4P0001d0;ooo`3oool0gia80018W03Oool0oono07@0B02Lgol0_g@007AdM000B9`0gooo
0?ooo`3OW4P0001d0;ooo`3oool0gia80018W03Oool0oooO09a8001d_ol0gia8001d_`3ogi`0B018
09cOo`80oooo00H0_g@00000M02oool0oooo0=nLB018W=l20?ooo`070=nLB00007@0_ooo0?ooo`3O
W4P004RL0=ooo`050?ooo`080?oo_`1d0000M9bL04P0M02oool0oooO09a80000M;l20?ooo`0;0=nL
B00007@0_ooo0?ooo`2oM000001809cOo`3oool0oooO09a80018W=l00P3oool00`3ooml0W4P004RL
g`030?ooo`0F0=nLB00007@0_ooo0?ooo`3OW4P004RL0=ooo`3ookl0M01809cOo`2oM000M7Ad0018
W03Oool0oooO09a8001d_ol0gia8001d_`3ogi`0B01809cOoh80oooo00<000000?ooo`3oool07@3o
ool00400oooo00<000000?ooo`3oool0R03oool01@3o_g@007Bo0?oog`2LB000B9cO00l0oooo00<0
_g@0001d_`3OW4P00P0000000`18W=l0oooo0?ooo`020?ooo`050?oOW018000000000000001d_ol0
0`3oool04`3ogi`0B00000000018W9`0M00004RLg`3ookl0M00000000018W=l0oooO09a800000000
B9cO0?ooo`3o_g@000000018W03Oool01@3oool0403o_g@0000000000000B9`0gooo0?oo_`1d0000
000004RLg`3oool0okmd0000000004P0W=oo0?oo_`1d00030000000507Boo`3oool0oooo0?oOW018
00000P0000002P1d_ol0omnL04P000000000M;oo0?ooo`2oM000001d0;oo_`1d00030000000507Bo
o`3oool0oooo0?oOW01800000P0000002@1d_ol0omnL04P000000000M;oo0?ooo`2oM000001d0;oo
o`020?ooo`0>0?oOW018000000000000001d_ol0oono07@000000000B9cO0?ooo`3o_g@000000000
B02Lgol30?ooo`050?oOW018000000000000B02Lgol00`3oool01@3ogi`0B000000000000000M;oo
00D0oooo00T0omnL04P000000000B9cO0?oo_`1d000000000000M02oool00`3oool0203ogi`0B000
00000018W9`0M00004RLg`3ookl0M0000`0000000`1d_kl0M000000000020000000507Boo`3oool0
oooo0?oOW01800000P0000004@1d_ol0omnL04P000000000M;oo0?ooo`2oM000001d0;oo_`1d0000
000004RLg`3oool0okmd0000000004P0W=oo0880oooo00<000000?ooo`3oool07@3oool00400oooo
00<000000?ooo`3oool0R@3oool0103OW4P0B9cO0;md0000M;l?0?ooo`030=nLB000B9`0gooo03L0
oooo00<0okmd0000M02oool04@3oool00`3o_g@0001809cOo`190?ooo`030?noM00007@0_ooo0080
oooo00<0okmd0000M02oool0VP3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo
0?ooo`2:0?ooo`040=nLB01dM7@0001d0;ooo`h0oooo00<0omnL04P0B02Lgol0>03oool00`3OW7@0
W=oo0?ooo`0@0?ooo`030?oog`2LB00007Bo04X0oooo00<0giad09cOo`3oool00P3oool00`3OW7@0
W=oo0?ooo`2I0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo08/0oooo
00<0_g@00000B02Lgol03P3oool00`3ookl0M00004RLg`0Y0?ooo`040?oo_`1d000004RL0=oooal0
oooo00<0okmd0000M02oool0jP3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo
0?ooo`2;0?ooo`030?oog`2LB000B9cO00l0oooo00<0gia8001d_`3oool0C03oool01@3OW4P004RL
0=oog`2LB000B9cO0>L0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0
V`3oool01@3ooml0W4P00000000007@0_ooo04d0oooo00@0gia80000000004P0W=ooi`3oool00`00
0000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3o
ool07@3oool00400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo
00100?ooo`030000003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool0
0`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400oooo00<000000?oo
o`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0
oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo
00<000000?ooo`3oool07@3oool00400oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003o
ool0oooo01d0oooo00100?ooo`030000003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M
0?ooo`00@03oool00`000000oooo0?ooo`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400
oooo00<000000?ooo`3oool0o`3ooooI0?ooo`030000003oool0oooo01d0oooo00100?ooo`030000
003oool0oooo0?l0oooof@3oool00`000000oooo0?ooo`0M0?ooo`00@03oool00`000000oooo0?oo
o`3o0?ooomT0oooo00<000000?ooo`3oool07@3oool00400ooooo`00003M000001l0oooo003o0?oo
ool0oooo?@3oool00?l0ooooo`3ooolm0?ooo`00o`3ooooo0?ooocd0oooo003o0?ooool0oooo?@3o
ool00?l0ooooo`3ooolm0?ooo`00o`3ooooo0?ooocd0oooo003o0?ooool0oooo?@3oool00?l0oooo
o`3ooolm0?ooo`00f@3oool00`3ooml0W4P0000000020000000307Boo`3oool0oooo04H0oooo00T0
oono07@000000000B9cO0?ooo`3ookl0M00000000018W=l0o`3oool<0?ooo`00f`3oool00`3o_g@0
001d0;ooo`190?ooo`0:0?oo_`1d04P0W=oo0;md0000M;l0oono07@0B02Lgol0_g@0001d_ol0oooo
2`3oool00=/0oooo00<0oono07@0001d_ol0C03oool00`3ookl0M00007Boo`020?ooo`030?oo_`1d
0000M;oo0?l0oooo2P3oool00=/0oooo00<0oooO09a80018W=l0C@3oool00`3OW4P004RL0=ooo`02
0?ooo`030=nLB000B9`0gooo0?l0oooo2@3oool00;<0oooo00X0gia8000000000000000007Boo`3o
ool0gia807Boo`3oool0gia8100000005018W=l0oooo0?oOW018000000000018W03Oool0_g@00000
0000M;l0oooo0?noM0000000001d0;ooo`3oool0okmd0000000004P0W=oo0P3oool01P2oM0000000
0000B02Lgol0oooo0=nLB0@0000000P0M;oo0?ooo`3oool0oooo0?noM0000000001d0;ooo`80oooo
00D0gia8000000000000000007Boo`0<0?ooo`030?oo_`1d000000000080000000L0B9cO0?ooo`3o
ool0okmd0000000007@0_ooo0080oooo00T0_g@00000000004P0W=oo0?ooo`3o_g@000000000B02L
gol0103oool00`2oM000000000000005000000030018W03Oool0oooo00@0oooo00P0_g@000000000
04P0W=oo0?ooo`3o_g@0001809cOo`80oooo00<0okmd0000B02Lgol00P3oool00`2oM00007Bo0?oo
o`040?ooo`0;0?oOW018000000000000001d_ol0gia800000000B9`0gooo0?ooo`3OW4P00P000000
200004P0W=oo0?ooo`3oool0okmd0000000004P0W=oo0P3oool0102oM00000000000B02Lgol20?oo
o`050?noM000000000000000B02Lgol00`3oool00`3OW4P004RL0=ooo`020?ooo`030;md00000000
07Bo00H0oooo00H0gia800000000000000000000B02Lgol60?ooo`0B0?oOW018000000000000001d
_ol0gia800000000B9`0gooo0?noM0000000001809cOo`3oool0okmd0000000004P0W=oo0P3oool0
1@3OW4P00000000000000000M;oo0080oooo00<0_g@0001d_`3oool0103oool00`3OW4P004RL0=oo
o`020?ooo`030;md0000000007Bo09H0oooo002d0?ooo`030?oog`2LB7@0_ooo0080oooo00P0_g@0
0000M02oool0oooo0?oog`2LB000001809cOo`80oooo0200omnL04P0B02Lgol0omnL04Qd_`2oM000
07Bo0?noW02Lgol0omnL04P0M02ookl0W9cO0?ooo`3o_g@0001809cOg`2LM9`0gooo0?ooo`3OW4P0
04RL0=oog`2LM9`0gooo0?noM00004P0W=oo0?ooo`3ooml0W4P0001d_`80oooo00P0omnL04P0M02o
okl0W9cO0?ooo`3ogi`0B00007Boo`80oooo00<0_g@0001d_`3oool02P3oool0303ooml0W4P007Bo
o`3oool0omnL04P0001d_ol0oooo0?oOW01807@0_ono09bLg`80oooo00X0gia80018W03Ooml0W7BL
0=ooo`3o_g@0001809cOg`2LM9`0gooo103oool0103ookl0M0000018W03Oool30?ooo`030;md0000
000007Bo00@0oooo00P0gia80018W03Ooml0W7BL0=ooo`3ookl0M00004RLg`80oooo00<0oono07@0
0018W=l00P3oool00`3o_g@0001d0;ooo`040?ooo`0S0;md0000B9`0gooo0?ooo`2oM000M7Ad0000
M02ogi`0W;oo0?noM0000000M;oo0?ooo`3ooml0W4P004RLg`3oool0okmd0000B02Lgml0W7BL0=oo
o`3oool0gia80018W03Ooml0W7BL0=ooo`3ogi`0B01809cOo`3oool0omnL04P00000M;l00P3oool0
0`3o_g@0001d0;ooo`020?ooo`040=nLB000B9`0gkmd07Boo`H0oooo00@0oooO09a80000B9`0gooo
1`3oool05P2oM00004RL0=ooo`3oool0_g@007AdM00007@0_mnL09boo`3o_g@0001809cOg`2LM9`0
gooo0?noM00004P0W=oO09adW03Oool0omnL04P0001d_ol20?ooo`060;md0000M;l0oooo0?noM000
07@0_ooo103oool00`3o_g@0001d0;ooo`020?ooo`040=nLB000B9`0gkmd07BooiD0oooo002e0?oo
o`070=nLM02Lgol0oooo0?ooo`2oM00004Q804Qd_`020?ooo`030?noM00004P0W=oo0080oooo0140
oooO09a80018W=l0oooo0?noM01dM7@004RL0=ooo`3oool0oono07@0001d_ol0gibL0;ooo`3ookl0
M00007Boo`030?ooo`030?noM00004P0W=oo0080oooo00<0oono07@0001d_ol00P3oool0303ooml0
W4P004RLg`3oool0oono07@0001d_ol0gibL0;ooo`3OW4P004RL0=ooo`<0oooo00<0gibL0;ooo`3o
ool02P3oool0402oM7@0_ooo0?ooo`3ookl0M00007Boo`3oool0oono07@0001d_ol0gibL0;ooo`3o
ool0okmd0000B02Lgol20?ooo`030?oo_`1d0000M;oo00L0oooo00<0gia80000B02Lgol0103oool0
103ookl0M0000018W03Oool20?ooo`030?noM00004P0W=oo00<0oooo00<0_g@00018W03Oool00P3o
ool01`2oM00004RL0=ooo`3oool0omnL04P0001d_ol0103oool00`2oM00007Bo0?ooo`020?ooo`07
0;md00000000M;oo0?ooo`3OW4P004RL0=ooo`030?ooo`050=oOg`3oool0oono07@0001d_ol00`3o
ool00`3o_g@0001809cOo`020?ooo`030=nLB000B9`0gooo0080oooo00L0omnL04P00018W=l0oooo
0?oo_`1d0000B9cO0080oooo00<0omnL04P0B02Lgol0203oool00`3o_g@0001809cOo`070?ooo`03
0;md0000M;l0oooo0080oooo00L0_g@00000001d_ol0oooo0?oo_`1d0000M;oo0080oooo00<0oono
07@0001d_ol00P3oool00`3OW4P004RL0=ooo`030?ooo`050=nLW02oool0omnL04P0001d_ol0103o
ool00`3ookl0M00004RLg`020?ooo`030?oOW01804P0W=oo09H0oooo002e0?ooo`0<0?oOW01dW=l0
oooo0?oog`2LB00007Bo0?noM018W=l0oooo0?oo_`1d0000B9cO0`3oool01`2oM00004RL0=ooo`3o
ool0_g@00000B02Lgol00P3oool00`2oM000B9cO0?ooo`020?ooo`030;md0000M;l0oooo0080oooo
00<0oono07@00018W=l00P3oool00`3ooml0W4P004RLg`030?ooo`060?oOW01804P0W=oo0?ooo`2o
M000B9cO0P3oool01@3o_g@00000000000000000B9cO0140oooo00<0gia80000M02oool00P3oool0
0`2oM000B9cO0?ooo`020?ooo`030?oo_`1d0000B9cO00<0oooo00<0_g@0001d_`3oool01P3oool0
0`3ogi`0B00007Boo`050?ooo`030?oOW018000007Bo0080oooo00<0oono07@00018W=l00`3oool0
0`3OW4P0001d0;ooo`020?ooo`080=nLB00007@0_ooo0?ooo`3ooml0W4P00018W03Oool30?ooo`03
0?noM00007@0_ooo0080oooo00L0oono07@00018W=l0oooo0?noM00007@0_ooo00H0oooo00<0_g@0
001d_`3oool00P3oool00`3ookl0M00004RLg`020?ooo`030?noM00007@0_ooo00<0oooo00/0gia8
0000M02oool0oooO09a8000007@0_ooo0?ooo`3ooml0W4P004RLg`080?ooo`030?oo_`1d0000B9cO
00L0oooo00<0okmd0000M02oool00P3oool00`3ookl0M00004RLg`020?ooo`030;md0000M;l0oooo
0080oooo00D0_g@0001d_`3oool0oooo0?noM0030000000804RLg`3oool0oooo0?ooo`3ooml0W4P0
0018W03Oool30?ooo`080?oog`2LB000001d0;ooo`3oool0oooO09a80018W=nF0?ooo`00]@3oool0
3@3ooml0W7BL0=ooo`3ookl0M00004RLg`3oool0giad09cOo`3oool0_g@00018W03Oool00P3oool0
1`3o_g@0001d0;ooo`3oool0oono07@0001d_ol00P3oool00`3OW4P004RL0=ooo`020?ooo`030=nL
B000B9`0gooo00<0oooo00<0_g@00018W03Oool00P3oool00`3OW4P0001d0;ooo`020?ooo`0?0?oo
g`2LB000B9cO0?ooo`3OW4P004RL0=ooo`3oool0oono07@0B02Lgol0oooo0=nLB0000000B9cO0080
oooo00<0okmd0000000000001P0000001P0007@0_ooo0?ooo`3o_g@0001809cOo`<0oooo00<0gia8
0018W03Oool00`3oool00`2oM00004RL0=ooo`020?ooo`030=nLB000B9`0gooo00H0oooo00<0oono
07@00000M;l01P3oool00`2oM000001809cOo`020?ooo`030;md0000B9`0gooo0080oooo00<0omnL
04P0001d_ol00P3oool00`3ogi`0B00007Boo`020?ooo`030;md0018M7@0M9cO00<0oooo00<0oono
07@0001d_ol00`3oool01P2oM00004RL0=ooo`3ookl0M00007Boo`H0oooo00<0gia80018W03Oool0
0`3oool01`2oM00004RL0=ooo`3oool0oono07@0001d_ol00`3oool01`3ogi`0B01809cOo`3oool0
gia80018B018M;l00P3oool00`2oM00007Bo0?ooo`070?ooo`030?oog`2LB00007Bo00L0oooo00<0
oono07@0001d_ol00`3oool01`2oM00004RL0=ooo`3oool0gia80018W03Oool00P3oool02`3OW4P0
04RL0=ooo`3oool0oono07@0B02Lgol0oooo0=nLB0000000B9cO0080oooo00<0_g@004QdM01dW=l0
103oool00`3OW4P004Q804Qd_`020?ooo`030;md0000M;l0oooo09D0oooo002f0?ooo`0A0=nLM02L
gol0omnL04P00018W=l0oooo0?oog`2LM9`0gooo0=nLB00004P0W=oo0?ooo`3ogml0gono07@0001d
_ol00P3oool01`2oM00007Bo0?ooo`3oool0omnL04P0M02oool00P3oool00`3ogi`0B01809cOo`03
0?ooo`030=nLB00004P0W=oo0080oooo00@0okmd00000000B9`0gooo0P3oool01P3OW4P007Bo0?oo
o`3ogi`0B01d0;ooo`80oooo00L0okmd0018W03Oool0oooo0?oo_`1d0000B9cO00`0oooo00X0omnL
04P0B02Lgol0oooo0=nLM02Lgol0oooo0?oOW01807@0_ooo0`3oool00`3OW4P0001809cOo`020?oo
o`030?oOW01804P0W=oo00L0oooo00<0_g@00018W03Oool01@3oool00`3o_g@0000004RLg`020?oo
o`030=nLB00004P0W=oo0080oooo00<0oooO09a80018W=l00P3oool00`3ooml0W4P004RLg`020?oo
o`040?noM000B9`0_ia807Boo`<0oooo00<0okmd0000M02oool00P3oool03`3OW4P0001d0;ooo`3o
ool0gia80000M02oool0oooo0?oo_`1d0000M;oo0?ooo`3ogi`0B01809cOo`030?ooo`030=nLB000
04P0W=oo0080oooo00<0gia80000M02oool00P3oool0303ookl0M00007Boo`3oool0okmd0000M02o
_g@0B9cO0?ooo`3o_g@0001d0;ooo`P0oooo00<0gia80000B018B0000`0000001`18W=l0oooo0?oo
o`3oool0okmd0000M02oool00P3oool01`3OW4P0001d0;ooo`3oool0omnL04P0B02Lgol00P3oool0
0`3ogi`0B01809cOo`020?ooo`0<0?noM000B9`0gooo0?ooo`3ookl0M00004RLg`3oool0okmd0018
W02oW4P0M;oo0`3oool0203o_g@0001d0;noM018W=l0oooo0?noM00007@0_oooU@3oool00;H0oooo
00D0omnL07BLg`3ogi`0B00007Boo`020?ooo`0:0?oOW01dW=l0omnL04P00018W=l0oooo0?oOW018
M;l0_g@0001d_`80oooo00L0okmd0018W03Oool0oooo0?oo_`1d0000M;oo0080oooo00<0oooO09a8
0018W=l00`3oool00`3ogi`0B00004RLg`020?ooo`0<0?oo_`1d0000M9bL04P0M02oool0oooO09a8
0000M;l0oooo0?oo_`1d0000M;oo0`3oool01`3OW4P0001d0;ooo`3oool0gia80018W03Oool02`3o
ool02P3ogi`0B01809cOo`3ooml0W4P007Boo`3oool0oono07@0001d_ol30?ooo`030?oOW0180000
B9cO0080oooo00<0oooO09a80018W=l01`3oool00`3o_g@0001809cOo`050?ooo`030?oOW0180000
07Bo0080oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`020?ooo`0C0;md0000B9`0
gooo0?ooo`3ookl0M01809cOo`2oM000M7Ad0018W03Oool0oooo0=nLB00007@0_ooo0?ooo`2oM000
001809cOo`020?ooo`0:0;md000007@0_ooo0?ooo`3OW4P0B9cO0?ooo`3ooml0W4P004RLg`<0oooo
00<0omnL04P00018W=l00`3oool0402oM000001d0;ooo`3oool0okmd0000M02oool0oooo0?oog`2L
B000M;oo0=nLB000M;l0omnL04P0B02Lgol80?ooo`030?noM00004P0W=oo0080oooo00@0omnL04P0
000007@0_ooo0P3oool02`3OW4P0001d0;ooo`3oool0_g@00000B02Lgol0oooo0?oog`2LB000B9cO
0080oooo00<0oooO09a80018W=l00`3oool05P3OW4P0001d0;ooo`3oool0gia80018W03Oool0oono
07@0B02Lgol0_g@007AdM000B9`0gooo0?oog`2LB000M;oo0=nLB000M;l0omnL04P0B02LgonE0?oo
o`00]P3oool01@3ooml0W7BL0=noM00004P0W=oo00<0oooo00T0gia807Bog`2LB00004RL0=ooo`3o
ool0okmd00000000M;l00P3oool00`3ogi`0B01809cOo`020?ooo`050;md0018W=l0oooo0?oo_`1d
00000`0000003`1d_ol0okmd00000000000004RL0=ooo`3ogi`0B00000000018W=l0oono07@00000
0000001d0;ooo`020?ooo`030;md0018W=l0oooo00<0oooo00D0omnL04P000000000000007Boo`0=
0?ooo`050;md00000000000000000018W=l00P3oool01@2oM000B9cO0?ooo`3oool0okmd00800000
00@004RL0=ooo`3ookl0M0000`0000000`1d_ol0oooo0?ooo`030?ooo`030?oOW0180000M;oo00D0
oooo00D0omnL04P00000M;l0oooo0?noM002000000040018W03Oool0oono07@000<0000000<0M;no
07@0000000000P000000201dW9`0B0000000001d_ol0oooo0;md000007@0_ooo0P3oool01P3ogi`0
B00000000018W9`0M00004RLg`<0oooo00L0omnL04P000000000001809cOo`3ookl0M00000<00000
00H0M;oo0?noM000000000000018W03Oool30?ooo`0?0?noM000000000000000001d_ol0oooo0?oo
_`1d0000000004RLg`3oool0okmd0000000004P0W=oo00P0oooo00<0oono07@00018W=l00`3oool0
0`3ogi`0B000001d_`030?ooo`080?oOW0180000000004RLW01d0000B9cO0?oo_`1d000300000003
07Bo_`1d000000000080000000D0M;oo0?ooo`3oool0omnL04P000020000000A07Boo`3ogi`0B000
0000001d_ol0oooo0;md000007@0_ono07@000000000B9cO0?ooo`3o_g@000000000B02Lgol0U@3o
ool00;L0oooo00@0giad07AdB00004P0W=oo0`3oool01@3ooml0W4Qd09bLB00004P0W=oo00/0oooo
00<0gia80018W03Oool00P3oool00`3o_g@0001d0;ooo`0A0?ooo`030=nLB000B9`0gooo01`0oooo
00<0gia80018W03Oool0203oool00`3o_g@0001d0;ooo`060?ooo`030?oog`2LB00007Bo00D0oooo
00<0okmd00000018W=l0203oool00`3o_g@0001809cOo`020?ooo`030?noM00004P0W=oo01X0oooo
00<0okmd0000M02oool08@3oool00`2oM00004RL0=ooo`030?ooo`030;md000007@0_ooo00X0oooo
00<0okmd0000M02oool00P3oool00`3o_g@0001d0;ooo`2]0?ooo`00]`3oool0103ogi`0B0000000
M02oool40?ooo`040?oo_`1dB4P0B00004RLg`/0oooo00<0omnL04P0M02oool00`3oool00`3OW7@0
W=oo0?ooo`0@0?ooo`030?oOW01807@0_ooo01`0oooo00<0omnL04P0M02oool02@3oool00`3OW7@0
W=oo0?ooo`060?ooo`030;md0000B9`0gooo00<0oooo00@0oooO09a8000007@0_ooo203oool00`3o
oml0W4P0001d_`020?ooo`030?oog`2LB00007Bo01/0oooo00<0giad09cOo`3oool0803oool00`3O
W4P0001d0;ooo`030?ooo`030;md000007@0_ooo00/0oooo00<0giad09cOo`3oool00P3oool00`3O
W7@0W=oo0?ooo`2/0?ooo`00]`3oool0103ooml0W4P00018W03Oool50?ooo`040?noM0000000001d
0;ooo`X0oooo00<0oooO09a8001d_ol0203oool0103ookl0M0000018W03Oool:0?ooo`030?oog`2L
B000M;oo01`0oooo00<0oooO09a8001d_ol00`3oool0103ookl0M0000018W03Oool;0?ooo`030=nL
B00004P0W=oo0080oooo00@0oono07@000000000M;oo103oool0103ookl0M0000018W03Oool20?oo
o`030?noM00007@0_ooo0080oooo00<0okmd0000M02oool07`3oool0103ookl0M0000018W03OoolJ
0?ooo`030?noM0000000M;oo0080oooo00<0okmd00000000M;l0`03oool00;H0oooo00D0okmd0000
0000000004RL0=ooo`060?ooo`040;md000000000000001d_`X0oooo00<0gia804RLg`3oool05P3o
ool00`3OW4P0B9cO0?ooo`0L0?ooo`030=nLB018W=l0oooo0100oooo00<0_g@00000000000001000
00000`00B9`0gooo0?ooo`0;0?ooo`0:0=nLB000B9`0gooO09a80018W=l0gia80018W03Ooml0W4P0
04RLgcT0oooo00<0_g@0000000000000100000000`00M;l0oooo0?ooo`2o0?ooo`00c@3oool0103o
oml0W4P000000000M;lE0?ooo`040?oog`2LB0000000001d_a/0oooo00@0oooO09a80000000007Bo
9`3oool02@3OW4P000000000B02Lgol0oooo0=nLB0000000001809cOo`3o0?ooo`<0oooo003o0?oo
ool0oooo?@3oool00?l0ooooo`3ooolm0?ooo`00o`3ooooo0?ooocd0oooo003o0?ooool0oooo?@3o
ool00001\
\>"],
  ImageRangeCache->{{{0, 570.875}, {456.5, 0}} -> {-0.166566, -0.123893, \
0.00331556, 0.00331556}, {{75.3125, 527.75}, {400.438, 287.313}} -> {-10.427, \
-1.41399, 0.0513158, 0.025658}, {{75.3125, 527.75}, {276, 162.875}} -> \
{-10.3156, -3.54254, 0.0513158, 0.025658}, {{75.3125, 527.75}, {151.563, \
38.5}} -> {-10.4369, -5.67641, 0.0513644, 0.0256823}}],

Cell["\<\
and here is an example of an envelope figure with logarithmic vertical scale.\
\
\>", "Text"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .5 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.119048 0.5 0.166667 [
[.02381 -0.0125 -9.34375 -15.6875 ]
[.02381 -0.0125 9.34375 0 ]
[.2619 -0.0125 -9.34375 -15.6875 ]
[.2619 -0.0125 9.34375 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.7381 -0.0125 -5 -15.6875 ]
[.7381 -0.0125 5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -23.75 -43.0625 ]
[.5 -0.0125 23.75 -15.6875 ]
[-0.0125 .1165 -31 -7.84375 ]
[-0.0125 .1165 0 7.84375 ]
[-0.0125 .16667 -25 -7.84375 ]
[-0.0125 .16667 0 7.84375 ]
[-0.0125 .28316 -25 -7.84375 ]
[-0.0125 .28316 0 7.84375 ]
[-0.0125 .33333 -19 -7.84375 ]
[-0.0125 .33333 0 7.84375 ]
[-0.0125 .44983 -19 -7.84375 ]
[-0.0125 .44983 0 7.84375 ]
[-0.0125 .5 -10 -7.84375 ]
[-0.0125 .5 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .25 -46.6875 -11.0938 ]
[-0.01875 .25 -31 11.0938 ]
[ 0 .5 .125 0 ]
[1.0125 0 0 -7.84375 ]
[1.0125 0 31 7.84375 ]
[1.0125 .1165 0 -7.84375 ]
[1.0125 .1165 31 7.84375 ]
[1.0125 .16667 0 -7.84375 ]
[1.0125 .16667 25 7.84375 ]
[1.0125 .28316 0 -7.84375 ]
[1.0125 .28316 25 7.84375 ]
[1.0125 .33333 0 -7.84375 ]
[1.0125 .33333 19 7.84375 ]
[1.0125 .44983 0 -7.84375 ]
[1.0125 .44983 19 7.84375 ]
[ 1 0 .125 0 ]
[.5 .5125 -98.875 0 ]
[.5 .5125 98.875 15.6875 ]
[ -0.0005 0 0 0 ]
[ 1.0005 .5 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(-) show
71.687500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%+ font Math1-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%+ Math1-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
%%+ font Math1-Italic
grestore
.2619 0 m
.2619 .00625 L
s
gsave
.2619 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(-) show
71.687500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%+ font Math1-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%+ Math1-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
%%+ font Math1-Italic
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.7381 0 m
.7381 .00625 L
s
gsave
.7381 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.125 Mabswid
.08333 0 m
.08333 .00375 L
s
.14286 0 m
.14286 .00375 L
s
.20238 0 m
.20238 .00375 L
s
.32143 0 m
.32143 .00375 L
s
.38095 0 m
.38095 .00375 L
s
.44048 0 m
.44048 .00375 L
s
.55952 0 m
.55952 .00375 L
s
.61905 0 m
.61905 .00375 L
s
.67857 0 m
.67857 .00375 L
s
.79762 0 m
.79762 .00375 L
s
.85714 0 m
.85714 .00375 L
s
.91667 0 m
.91667 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -84.75 -47.0625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 35.375 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 31.375000 moveto
 460.000000 31.375000 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 31.375000 lineto
closepath
clip newpath
63.000000 21.625000 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
65.937500 13.375000 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(d) show
%%IncludeResource: font Math1
%%IncludeFont: Math1
/Math1 findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
64.625000 21.625000 moveto
(\\200) show
66.125000 21.625000 moveto
(\\200) show
67.625000 21.625000 moveto
(\\200) show
69.125000 21.625000 moveto
(\\200) show
70.625000 21.625000 moveto
(\\200) show
72.125000 21.625000 moveto
(\\200) show
72.500000 21.625000 moveto
(\\200) show
66.062500 30.312500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(l) show
75.625000 21.625000 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Sin) show
%%IncludeResource: font Math2
%%IncludeFont: Math2
/Math2 findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
90.937500 21.625000 moveto
(@) show
94.875000 21.625000 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(q) show
%%IncludeResource: font Math2
%%IncludeFont: Math2
/Math2 findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
102.562500 21.625000 moveto
(D) show
106.500000 21.625000 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math1
%%+ font Times-Italic
%%+ font Math1-Italic
%%+ font Math2
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math1
%%+ Times-Italic
%%+ Math1-Italic
%%+ Math2
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math1
%%+ font Times-Italic
%%+ font Math1-Italic
%%+ font Math2
grestore
0 .1165 m
.00625 .1165 L
s
gsave
-0.0125 .1165 -92 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.005) show
90.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
0 .16667 m
.00625 .16667 L
s
gsave
-0.0125 .16667 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.01) show
84.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
0 .28316 m
.00625 .28316 L
s
gsave
-0.0125 .28316 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.05) show
84.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
0 .33333 m
.00625 .33333 L
s
gsave
-0.0125 .33333 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.1) show
78.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
0 .44983 m
.00625 .44983 L
s
gsave
-0.0125 .44983 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.5) show
78.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
0 .5 m
.00625 .5 L
s
gsave
-0.0125 .5 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
69.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.001 w
0 .05017 m
.00375 .05017 L
s
0 .07952 m
.00375 .07952 L
s
0 .10034 m
.00375 .10034 L
s
0 .12969 m
.00375 .12969 L
s
0 .14085 m
.00375 .14085 L
s
0 .15051 m
.00375 .15051 L
s
0 .15904 m
.00375 .15904 L
s
0 .21684 m
.00375 .21684 L
s
0 .24619 m
.00375 .24619 L
s
0 .26701 m
.00375 .26701 L
s
0 .29636 m
.00375 .29636 L
s
0 .30752 m
.00375 .30752 L
s
0 .31718 m
.00375 .31718 L
s
0 .32571 m
.00375 .32571 L
s
0 .3835 m
.00375 .3835 L
s
0 .41285 m
.00375 .41285 L
s
0 .43368 m
.00375 .43368 L
s
0 .46303 m
.00375 .46303 L
s
0 .47418 m
.00375 .47418 L
s
0 .48385 m
.00375 .48385 L
s
0 .49237 m
.00375 .49237 L
s
.25 Mabswid
0 0 m
0 .5 L
s
gsave
-0.01875 .25 -110.938 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
72.0938 11.8438 translate
90 rotate
-72.0938 -11.8438 translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.187500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(I) show
%%IncludeResource: font Math2
%%IncludeFont: Math2
/Math2 findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
68.187500 15.812500 moveto
(\\220) show
72.187500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(I) show
76.125000 17.750000 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 8.500000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0) show
81.187500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%+ font Math2
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%+ Math2
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
%%+ font Math2
grestore
.02381 .49375 m
.02381 .5 L
s
.2619 .49375 m
.2619 .5 L
s
.5 .49375 m
.5 .5 L
s
.7381 .49375 m
.7381 .5 L
s
.97619 .49375 m
.97619 .5 L
s
.125 Mabswid
.08333 .49625 m
.08333 .5 L
s
.14286 .49625 m
.14286 .5 L
s
.20238 .49625 m
.20238 .5 L
s
.32143 .49625 m
.32143 .5 L
s
.38095 .49625 m
.38095 .5 L
s
.44048 .49625 m
.44048 .5 L
s
.55952 .49625 m
.55952 .5 L
s
.61905 .49625 m
.61905 .5 L
s
.67857 .49625 m
.67857 .5 L
s
.79762 .49625 m
.79762 .5 L
s
.85714 .49625 m
.85714 .5 L
s
.91667 .49625 m
.91667 .5 L
s
.25 Mabswid
0 .5 m
1 .5 L
s
.99375 0 m
1 0 L
s
gsave
1.0125 0 -61 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.001) show
90.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.99375 .1165 m
1 .1165 L
s
gsave
1.0125 .1165 -61 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.005) show
90.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.99375 .16667 m
1 .16667 L
s
gsave
1.0125 .16667 -61 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.01) show
84.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.99375 .28316 m
1 .28316 L
s
gsave
1.0125 .28316 -61 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.05) show
84.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.99375 .33333 m
1 .33333 L
s
gsave
1.0125 .33333 -61 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.1) show
78.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.99375 .44983 m
1 .44983 L
s
gsave
1.0125 .44983 -61 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.5) show
78.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
grestore
.001 w
.99625 .05017 m
1 .05017 L
s
.99625 .07952 m
1 .07952 L
s
.99625 .10034 m
1 .10034 L
s
.99625 .12969 m
1 .12969 L
s
.99625 .14085 m
1 .14085 L
s
.99625 .15051 m
1 .15051 L
s
.99625 .15904 m
1 .15904 L
s
.99625 .21684 m
1 .21684 L
s
.99625 .24619 m
1 .24619 L
s
.99625 .26701 m
1 .26701 L
s
.99625 .29636 m
1 .29636 L
s
.99625 .30752 m
1 .30752 L
s
.99625 .31718 m
1 .31718 L
s
.99625 .32571 m
1 .32571 L
s
.99625 .3835 m
1 .3835 L
s
.99625 .41285 m
1 .41285 L
s
.99625 .43368 m
1 .43368 L
s
.99625 .46303 m
1 .46303 L
s
.99625 .47418 m
1 .47418 L
s
.99625 .48385 m
1 .48385 L
s
.99625 .49237 m
1 .49237 L
s
.25 Mabswid
1 0 m
1 .5 L
s
gsave
.5 .5125 -159.875 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 19.687500 moveto
 460.000000 19.687500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 19.687500 lineto
closepath
clip newpath
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4) show
69.000000 15.812500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(-) show
77.125000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(slit) show
95.062500 15.812500 moveto
(Diffraction) show
154.250000 15.812500 moveto
(Pattern) show
193.125000 15.812500 moveto
(with) show
216.750000 15.812500 moveto
(a) show
%%IncludeResource: font Math2
%%IncludeFont: Math2
/Math2 findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
222.750000 15.812500 moveto
(\\220) show
226.562500 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(d) show
233.250000 15.812500 moveto
%%IncludeResource: font Math1-Italic
%%IncludeFont: Math1-Italic
/Math1-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(=) show
241.750000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.5) show
256.750000 15.812500 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-Italic
%%+ font Math1-Italic
%%+ font Math2
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-Italic
%%+ Math1-Italic
%%+ Math2
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-Italic
%%+ font Math1-Italic
%%+ font Math2
grestore
0 0 m
1 0 L
1 .5 L
0 .5 L
closepath
clip
newpath
.5 Mabswid
[ .02 .02 ] 0 setdash
.03852 0 m
.04295 .03909 L
.04785 .07277 L
.05257 .09918 L
.05758 .12275 L
.06241 .14221 L
.06706 .15859 L
.07201 .17397 L
.07677 .18715 L
.08135 .19852 L
.08623 .20943 L
.09092 .21888 L
.09591 .22794 L
.10071 .2358 L
.10533 .24261 L
.11025 .24911 L
.11499 .25469 L
.11954 .25947 L
.12439 .26395 L
.12906 .26768 L
.13402 .27106 L
.1388 .27374 L
.14339 .2758 L
.14596 .27673 L
.14828 .27743 L
.15073 .27803 L
.15299 .27846 L
.15412 .27862 L
.15531 .27876 L
.1565 .27887 L
.15718 .27891 L
.15751 .27893 L
.15781 .27894 L
.15813 .27895 L
.15842 .27896 L
.15873 .27897 L
.15891 .27897 L
.15898 .27897 L
.15907 .27897 L
.15922 .27897 L
.15937 .27898 L
.15951 .27898 L
.15964 .27898 L
.1598 .27897 L
.15994 .27897 L
.16011 .27897 L
.16026 .27897 L
.16053 .27896 L
.16082 .27895 L
.16133 .27893 L
Mistroke
.16189 .2789 L
.1625 .27886 L
.16372 .27874 L
.16484 .27861 L
.16735 .27818 L
.16964 .27764 L
.17214 .27689 L
.17704 .2749 L
.18175 .27231 L
.18676 .26879 L
.19159 .26457 L
.19624 .25968 L
.20118 .25349 L
.20593 .24643 L
.21051 .23849 L
.21538 .22859 L
.22007 .2174 L
.22505 .20331 L
.22986 .18706 L
.23448 .1682 L
.23939 .14336 L
.24412 .11266 L
.24867 .07308 L
.251 .04663 L
.25352 .01035 L
Mfstroke
.25352 .01035 m
.25404 0 L
s
.26931 0 m
.27047 .02364 L
.27269 .05826 L
.27743 .1135 L
.28198 .15306 L
.28684 .18672 L
.2915 .21369 L
.29647 .23827 L
.30125 .25896 L
.30585 .27671 L
.31074 .29374 L
.31545 .30863 L
.32046 .3231 L
.32528 .33592 L
.32993 .34735 L
.33486 .35865 L
.33962 .36879 L
.34419 .37792 L
.34906 .38704 L
.35374 .39529 L
.35872 .40352 L
.36352 .41098 L
.36813 .41774 L
.37305 .42453 L
.37777 .43068 L
.3828 .43684 L
.38764 .44243 L
.3923 .44749 L
.39725 .45257 L
.40202 .45715 L
.40661 .46131 L
.41149 .46545 L
.41619 .46919 L
.42119 .4729 L
.426 .47622 L
.43064 .47919 L
.43556 .48212 L
.44031 .48471 L
.44487 .487 L
.44973 .48922 L
.4544 .49116 L
.45937 .493 L
.46416 .49456 L
.46876 .49588 L
.47366 .49707 L
.47838 .49803 L
.4834 .49884 L
.48591 .49917 L
.48823 .49942 L
.49063 .49963 L
Mistroke
.49287 .49979 L
.49514 .4999 L
.49635 .49994 L
.49701 .49996 L
.49763 .49998 L
.49823 .49999 L
.49853 .49999 L
.49886 .49999 L
.49919 .5 L
.49938 .5 L
.49947 .5 L
.49955 .5 L
.4997 .5 L
.49978 .5 L
.49986 .5 L
.49995 .5 L
.50004 .5 L
.5002 .5 L
.50034 .5 L
.50049 .5 L
.50065 .5 L
.5008 .5 L
.50107 .5 L
.50136 .49999 L
.50202 .49998 L
.50262 .49997 L
.50381 .49994 L
.50491 .4999 L
.50742 .49977 L
.50989 .49959 L
.51214 .49938 L
.51698 .49879 L
.52163 .49803 L
.52658 .49702 L
.53135 .49585 L
.53593 .49454 L
.54081 .49293 L
.54551 .49119 L
.5505 .48912 L
.55531 .48691 L
.55994 .48458 L
.56486 .48187 L
.5696 .47904 L
.57416 .47611 L
.57901 .47275 L
.58369 .46928 L
.58865 .46534 L
.59344 .46127 L
.59804 .4571 L
.60293 .45239 L
Mistroke
.60765 .44756 L
.61266 .4421 L
.61748 .43651 L
.62213 .43081 L
.62707 .42438 L
.63182 .4178 L
.6364 .4111 L
.64127 .40354 L
.64595 .39581 L
.65094 .38706 L
.65574 .37806 L
.66035 .36885 L
.66527 .35836 L
.67 .34753 L
.67455 .33635 L
.67939 .32352 L
.68405 .31013 L
.68901 .29457 L
.69378 .27807 L
.69837 .26051 L
.70326 .23954 L
.70796 .21653 L
.71296 .18801 L
.71778 .1549 L
.72241 .11509 L
.72476 .0904 L
.72734 .05785 L
.72979 .01906 L
Mfstroke
.72979 .01906 m
.7307 0 L
s
.74595 0 m
.74638 .00852 L
.74869 .04267 L
.7512 .07173 L
.75613 .11454 L
.76089 .14492 L
.76546 .16789 L
.77032 .18773 L
.77501 .2035 L
.77999 .21754 L
.78478 .22895 L
.78939 .23831 L
.7943 .24681 L
.79903 .25377 L
.80405 .26001 L
.80889 .26503 L
.81355 .26903 L
.8185 .27247 L
.82327 .27504 L
.82786 .27688 L
.83022 .2776 L
.83274 .27819 L
.83397 .27842 L
.83531 .27863 L
.83657 .27877 L
.83713 .27883 L
.83774 .27888 L
.83834 .27891 L
.8389 .27894 L
.83917 .27895 L
.83946 .27896 L
.83974 .27897 L
.83987 .27897 L
.83999 .27897 L
.84015 .27897 L
.84023 .27897 L
.84032 .27898 L
.84046 .27898 L
.84054 .27898 L
.84062 .27898 L
.84077 .27898 L
.8409 .27897 L
.84105 .27897 L
.84119 .27897 L
.84151 .27896 L
.84181 .27896 L
.84213 .27894 L
.84248 .27893 L
.84361 .27886 L
.84424 .2788 L
Mistroke
.84482 .27875 L
.84703 .27845 L
.84931 .27802 L
.8518 .27741 L
.85669 .27577 L
.8614 .27364 L
.8664 .2708 L
.87122 .26748 L
.87586 .26374 L
.88079 .25914 L
.88554 .25411 L
.8901 .24866 L
.89496 .24219 L
.89964 .23524 L
.90462 .22703 L
.90941 .21825 L
.91402 .2089 L
.91893 .19787 L
.92365 .18604 L
.92867 .17198 L
.9335 .1567 L
.93816 .14006 L
.9431 .11975 L
.94787 .09693 L
.95245 .07092 L
.95733 .03694 L
Mfstroke
.95733 .03694 m
.96143 0 L
s
[ ] 0 setdash
.09274 0 m
.09323 .00652 L
.09439 .01873 L
.09545 .02777 L
.0967 .03594 L
.09735 .03925 L
.09804 .04202 L
.09864 .04383 L
.09893 .04451 L
.0992 .04503 L
.09936 .04527 L
.09953 .04549 L
.09967 .04565 L
.09982 .04578 L
.0999 .04583 L
.09998 .04588 L
.10008 .04592 L
.10016 .04594 L
.10025 .04595 L
.10033 .04595 L
.10048 .04592 L
.10063 .04586 L
.10071 .04581 L
.10079 .04575 L
.10093 .04562 L
.10108 .04544 L
.10135 .04503 L
.10163 .04447 L
.10227 .04267 L
.10286 .04032 L
.10342 .03748 L
.10393 .0343 L
.10508 .02492 L
.10628 .01123 L
s
.10628 .01123 m
.107 0 L
s
.11725 0 m
.11732 .003 L
.1185 .04273 L
.11975 .07651 L
.12212 .12627 L
.12434 .16171 L
.12659 .19041 L
.12903 .21529 L
.1316 .23579 L
.13403 .25077 L
.13642 .26184 L
.13762 .26615 L
.13895 .26995 L
.14017 .27259 L
.14129 .2743 L
.14187 .27493 L
.14221 .27521 L
.14251 .27542 L
.14268 .27551 L
.14284 .27558 L
.14299 .27563 L
.14314 .27568 L
.14322 .2757 L
.1433 .27571 L
.14346 .27573 L
.14354 .27574 L
.14364 .27574 L
.14372 .27574 L
.1438 .27573 L
.14395 .27571 L
.14408 .27569 L
.14422 .27565 L
.14438 .27559 L
.1447 .27543 L
.145 .27524 L
.14553 .27477 L
.1461 .27411 L
.1473 .27214 L
.1486 .26908 L
.14984 .26529 L
.15097 .26106 L
.15349 .24865 L
.1558 .23341 L
.15831 .21193 L
.1607 .1855 L
.16324 .14906 L
.16559 .10376 L
.1678 .04438 L
s
.1678 .04438 m
.16898 0 L
s
.17729 0 m
.17735 .00175 L
.17859 .02752 L
.17993 .04756 L
.18119 .06111 L
.18236 .07002 L
.18292 .07327 L
.18353 .07607 L
.1841 .07809 L
.18462 .07944 L
.18494 .08004 L
.18523 .08044 L
.1854 .08061 L
.18556 .08072 L
.18564 .08077 L
.18573 .08081 L
.18582 .08083 L
.1859 .08084 L
.18605 .08083 L
.18618 .0808 L
.18633 .08073 L
.18649 .08061 L
.18666 .08045 L
.18682 .08025 L
.18711 .07979 L
.18767 .07855 L
.18828 .07664 L
.18892 .07402 L
.18952 .07099 L
.19059 .06412 L
.19174 .0545 L
.19296 .04156 L
.19429 .02361 L
.19554 .00213 L
s
.19554 .00213 m
.19564 0 L
s
.21047 0 m
.21068 .00285 L
.21184 .0147 L
.21248 .01989 L
.21307 .02401 L
.21363 .02723 L
.21424 .03008 L
.21479 .03211 L
.21529 .03352 L
.21559 .03416 L
.21586 .03461 L
.21601 .03481 L
.21617 .03499 L
.21633 .03512 L
.21647 .0352 L
.2166 .03525 L
.21672 .03528 L
.21686 .03527 L
.217 .03524 L
.21714 .03517 L
.2173 .03507 L
.21742 .03495 L
.21757 .03479 L
.21789 .0343 L
.21819 .0337 L
.21886 .03186 L
.21948 .02951 L
.22006 .0267 L
.22123 .01925 L
.22247 .0085 L
s
.22247 .0085 m
.22324 0 L
s
.24149 0 m
.24186 .00404 L
.24302 .01341 L
.24407 .01971 L
.24465 .02233 L
.24527 .02449 L
.24579 .02585 L
.24609 .02642 L
.24638 .02683 L
.24652 .02699 L
.24668 .02713 L
.24676 .02718 L
.24685 .02723 L
.247 .02728 L
.24716 .0273 L
.24731 .02728 L
.24744 .02723 L
.24758 .02715 L
.24774 .02702 L
.24791 .02684 L
.24826 .02632 L
.24858 .02566 L
.24888 .0249 L
.24948 .02293 L
.25013 .02008 L
.25133 .01291 L
.25248 .00324 L
s
.25248 .00324 m
.25278 0 L
s
.27015 0 m
.27039 .00351 L
.27169 .01852 L
.27292 .02923 L
.2741 .03686 L
.2747 .03977 L
.27535 .04229 L
.27591 .04389 L
.27622 .04457 L
.2765 .04505 L
.27677 .04539 L
.27692 .04553 L
.27706 .04563 L
.27718 .04569 L
.27731 .04573 L
.27745 .04574 L
.27758 .04572 L
.27773 .04567 L
.27781 .04562 L
.27789 .04557 L
.27804 .04545 L
.27818 .0453 L
.27849 .04484 L
.27883 .04417 L
.27945 .04244 L
.28002 .04027 L
.28111 .03449 L
.28226 .02585 L
.2834 .01425 L
s
.2834 .01425 m
.28446 0 L
s
.29778 0 m
.29906 .02504 L
.30022 .04298 L
.30149 .05863 L
.30274 .07059 L
.3039 .0791 L
.30513 .08558 L
.30582 .08814 L
.30612 .08901 L
.30645 .08982 L
.30673 .09037 L
.30703 .09082 L
.30719 .09101 L
.30735 .09115 L
.3075 .09124 L
.30765 .0913 L
.30778 .09133 L
.30792 .09132 L
.30805 .09129 L
.30818 .09123 L
.30832 .09114 L
.30847 .09099 L
.30875 .09063 L
.30906 .0901 L
.30939 .08934 L
.31006 .08721 L
.31068 .08452 L
.31126 .08133 L
.31252 .07197 L
.31371 .05953 L
.31478 .04472 L
.31595 .02351 L
s
.31595 .02351 m
.31691 0 L
s
.32529 0 m
.32587 .02269 L
.3271 .05846 L
.32823 .0842 L
.32954 .10842 L
.33075 .12651 L
.33305 .15212 L
.33427 .16174 L
.33494 .16603 L
.33556 .16934 L
.3361 .17176 L
.33669 .1739 L
.33728 .17551 L
.33761 .17616 L
.33775 .17639 L
.33791 .17661 L
.33806 .17679 L
.3382 .17691 L
.33833 .17701 L
.33847 .17708 L
.33859 .17712 L
.33871 .17713 L
.33884 .17712 L
.33898 .17707 L
.33912 .17699 L
.33928 .17686 L
.33943 .1767 L
.33957 .17652 L
.33984 .17608 L
.34012 .17548 L
.3407 .17376 L
.34124 .17159 L
.34184 .16853 L
.34246 .16447 L
.34351 .15542 L
.34466 .1417 L
.3459 .12115 L
.34702 .09492 L
.34765 .0759 L
.34832 .04988 L
.34896 .01764 L
s
.34896 .01764 m
.34921 0 L
s
.35291 0 m
.35309 .01781 L
.35371 .06153 L
.35429 .09448 L
.35546 .14563 L
.3567 .18734 L
.35892 .24397 L
.36127 .28874 L
.36347 .32186 L
.36814 .37331 L
.37055 .39263 L
.3731 .40866 L
.37548 .4201 L
.3777 .4279 L
.37896 .43118 L
.37959 .43248 L
.38016 .4335 L
.38067 .43427 L
.38124 .43495 L
.3818 .43547 L
.38211 .43569 L
.3824 .43584 L
.38256 .43591 L
.38272 .43596 L
.38287 .436 L
.38301 .43603 L
.38309 .43604 L
.38317 .43604 L
.38325 .43605 L
.38334 .43605 L
.38349 .43604 L
.38358 .43603 L
.38366 .43601 L
.38381 .43598 L
.38396 .43593 L
.38424 .43581 L
.38453 .43564 L
.38485 .4354 L
.38542 .43485 L
.38601 .43407 L
.38709 .43218 L
.38822 .42948 L
.38944 .42572 L
.39165 .41656 L
.39407 .40262 L
.39661 .3828 L
.399 .35824 L
.40122 .32853 L
.40249 .30744 L
.40369 .28393 L
Mistroke
.40593 .22579 L
.40711 .18304 L
.40775 .1534 L
.40835 .11936 L
.40886 .08282 L
.40941 .03049 L
Mfstroke
.40941 .03049 m
.40965 0 L
s
.41182 0 m
.41208 .0291 L
.41237 .05673 L
.41291 .09521 L
.41353 .12871 L
.41411 .1534 L
.41542 .19498 L
.41668 .22282 L
.41787 .24255 L
.41894 .25618 L
.4201 .26762 L
.42074 .27261 L
.42135 .27655 L
.4219 .27954 L
.42248 .28212 L
.42311 .2843 L
.42369 .28572 L
.42402 .28628 L
.42432 .28666 L
.42449 .28681 L
.42465 .28691 L
.42474 .28695 L
.42483 .28698 L
.42492 .287 L
.42501 .287 L
.42515 .28698 L
.4253 .28693 L
.42543 .28686 L
.42557 .28675 L
.42572 .2866 L
.42588 .2864 L
.42617 .28595 L
.42674 .28469 L
.42726 .2831 L
.42787 .28069 L
.42854 .27739 L
.42975 .26947 L
.43089 .25946 L
.43215 .24519 L
.43346 .2258 L
.43468 .20228 L
.43594 .1705 L
.43656 .15077 L
.43713 .1291 L
.43769 .10354 L
.4382 .07516 L
.43877 .03461 L
.43908 .00603 L
s
.43908 .00603 m
.43913 0 L
s
.44181 0 m
.44236 .05022 L
.44295 .08984 L
.44403 .14108 L
.44527 .18274 L
.44661 .21582 L
.44788 .23968 L
.44904 .25708 L
.45014 .27027 L
.45131 .28158 L
.4526 .29106 L
.4538 .29742 L
.45444 .29986 L
.45503 .30158 L
.45534 .30226 L
.45568 .30285 L
.45585 .30309 L
.45603 .30329 L
.45611 .30336 L
.4562 .30343 L
.45636 .30352 L
.45651 .30357 L
.45665 .30359 L
.45681 .30357 L
.45696 .30351 L
.4571 .30343 L
.45723 .30333 L
.45752 .30301 L
.45781 .30255 L
.45814 .30188 L
.45879 .3 L
.45942 .2975 L
.46001 .29445 L
.46113 .28681 L
.46234 .27538 L
.46365 .25817 L
.4649 .23573 L
.46603 .20794 L
.46666 .18805 L
.46733 .16128 L
.46797 .12849 L
.46856 .08783 L
.46883 .06364 L
.46912 .03107 L
s
.46912 .03107 m
.46934 0 L
s
.4711 0 m
.47116 .01173 L
.47134 .0384 L
.4715 .05914 L
.47212 .1192 L
.47274 .16236 L
.47333 .19453 L
.4745 .24456 L
.47575 .28468 L
.47797 .33822 L
.48034 .37993 L
.48255 .41001 L
.48479 .43431 L
.48723 .45524 L
.48979 .47213 L
.49222 .4841 L
.4946 .49244 L
.4958 .49545 L
.49645 .49675 L
.49713 .49788 L
.49771 .49865 L
.49802 .49899 L
.49835 .4993 L
.49865 .49953 L
.49892 .4997 L
.4992 .49984 L
.49933 .49989 L
.49946 .49993 L
.49961 .49996 L
.49976 .49999 L
.49992 .5 L
.50001 .5 L
.50009 .5 L
.50023 .49999 L
.50038 .49996 L
.50053 .49993 L
.50067 .49989 L
.501 .49974 L
.5013 .49957 L
.50165 .4993 L
.50197 .499 L
.50315 .49744 L
.50375 .49639 L
.5044 .49501 L
.50662 .48856 L
.50786 .48379 L
.50898 .47866 L
.5115 .46418 L
.5138 .44698 L
.51631 .4231 L
Mistroke
.51869 .394 L
.52123 .3537 L
.52236 .33134 L
.52358 .30297 L
.52578 .23378 L
.52693 .18122 L
.52757 .14199 L
.52815 .09525 L
.52848 .06066 L
.52866 .03875 L
.52884 .01153 L
Mfstroke
.52884 .01153 m
.52891 0 L
s
.53067 0 m
.53074 .01357 L
.53102 .04833 L
.53132 .0776 L
.53185 .11786 L
.53249 .15317 L
.53307 .17805 L
.53368 .19907 L
.53433 .21781 L
.53552 .24419 L
.53666 .26282 L
.53792 .2781 L
.53855 .28412 L
.53922 .28954 L
.53986 .29372 L
.54045 .29688 L
.54104 .29939 L
.54167 .30139 L
.54203 .30224 L
.54237 .30285 L
.54267 .30324 L
.54284 .30339 L
.543 .30349 L
.54314 .30355 L
.54329 .30358 L
.54343 .30358 L
.54356 .30355 L
.54371 .30348 L
.5438 .30343 L
.54388 .30337 L
.54403 .30323 L
.54418 .30306 L
.54446 .30264 L
.54478 .30202 L
.54542 .30032 L
.54599 .2983 L
.5466 .29553 L
.54785 .28807 L
.54901 .27879 L
.55008 .26785 L
.55135 .25165 L
.55255 .23235 L
.55362 .21094 L
.55479 .18111 L
.55596 .14156 L
.5566 .11346 L
.5572 .0806 L
.55776 .04087 L
s
.55776 .04087 m
.55818 0 L
s
.56087 0 m
.56092 .00568 L
.56125 .03626 L
.56185 .07842 L
.5624 .10806 L
.56291 .13085 L
.56408 .17106 L
.56527 .20119 L
.56652 .22535 L
.56758 .24158 L
.56875 .25572 L
.56999 .26738 L
.57112 .27539 L
.57174 .27887 L
.57242 .28191 L
.57306 .28412 L
.57336 .28495 L
.57364 .28559 L
.57391 .28609 L
.57421 .28652 L
.57437 .28669 L
.57452 .28682 L
.57466 .28691 L
.57481 .28697 L
.57496 .287 L
.57511 .28699 L
.57526 .28695 L
.57539 .28689 L
.57551 .28681 L
.57564 .2867 L
.57591 .28639 L
.57624 .28586 L
.57653 .28523 L
.5772 .28332 L
.57782 .28089 L
.5784 .27799 L
.57957 .27027 L
.58081 .25896 L
.58187 .24625 L
.58303 .22814 L
.58425 .20297 L
.58537 .1718 L
.58599 .14962 L
.58666 .11926 L
.5873 .08144 L
.58789 .03243 L
s
.58789 .03243 m
.58817 0 L
s
.59035 0 m
.59037 .003 L
.59064 .03625 L
.59094 .0658 L
.59154 .11211 L
.59261 .17077 L
.59385 .21873 L
.59497 .25214 L
.5975 .30729 L
.59982 .34334 L
.60232 .37268 L
.60472 .39392 L
.60726 .4108 L
.6084 .41677 L
.60962 .4222 L
.61076 .42642 L
.61183 .42961 L
.61305 .43246 L
.61369 .43362 L
.61438 .4346 L
.6147 .43497 L
.61504 .4353 L
.61534 .43555 L
.61565 .43575 L
.61594 .43589 L
.6161 .43595 L
.61625 .43599 L
.61639 .43602 L
.61652 .43604 L
.61667 .43605 L
.6168 .43605 L
.61694 .43603 L
.61709 .43601 L
.61723 .43598 L
.61736 .43594 L
.61766 .43581 L
.61797 .43563 L
.61856 .43515 L
.61921 .43442 L
.62036 .43258 L
.62143 .43024 L
.62271 .42669 L
.62391 .42253 L
.62615 .41263 L
.62857 .39863 L
.63079 .38238 L
.63325 .36013 L
.63555 .33448 L
.6379 .3021 L
.64013 .26362 L
Mistroke
.6413 .23913 L
.64258 .20788 L
.64373 .17399 L
.64482 .13484 L
.64595 .08161 L
.64655 .04464 L
.64686 .02181 L
Mfstroke
.64686 .02181 m
.64711 0 L
s
.65078 0 m
.65098 .01379 L
.65156 .04473 L
.6521 .06704 L
.65334 .10426 L
.65395 .1182 L
.65463 .13089 L
.65587 .14856 L
.65699 .16009 L
.65761 .16497 L
.65828 .16921 L
.65892 .17232 L
.65951 .17446 L
.65978 .17523 L
.66007 .1759 L
.66039 .17646 L
.66053 .17665 L
.66067 .17682 L
.66082 .17695 L
.66094 .17703 L
.66107 .17709 L
.66119 .17712 L
.66134 .17713 L
.66147 .1771 L
.66161 .17704 L
.66176 .17694 L
.66192 .17681 L
.66206 .17665 L
.66238 .17618 L
.66267 .17561 L
.66295 .17494 L
.66357 .17302 L
.66425 .17025 L
.6655 .1633 L
.66663 .15487 L
.66793 .14241 L
.66916 .12773 L
.67034 .1104 L
.67159 .08799 L
.67276 .06203 L
.67382 .03269 L
s
.67382 .03269 m
.67474 0 L
s
.68308 0 m
.68414 .02545 L
.68539 .04746 L
.6867 .06428 L
.68729 .0702 L
.68793 .07569 L
.68911 .08345 L
.68976 .08649 L
.69036 .08863 L
.69065 .08944 L
.69093 .09006 L
.69117 .09052 L
.69144 .09089 L
.69158 .09104 L
.69173 .09117 L
.69186 .09125 L
.69199 .0913 L
.69214 .09133 L
.69228 .09132 L
.69244 .09127 L
.69259 .09119 L
.69274 .09107 L
.6929 .0909 L
.6932 .09049 L
.69349 .08996 L
.69375 .08936 L
.69435 .08758 L
.69498 .08508 L
.69604 .07945 L
.6972 .07107 L
.69844 .05941 L
.69956 .04598 L
.70087 .02626 L
.70208 .00315 L
s
.70208 .00315 m
.70221 0 L
s
.71555 0 m
.71622 .00969 L
.71746 .02327 L
.71858 .03245 L
.71921 .0364 L
.71988 .03985 L
.72052 .04235 L
.72083 .0433 L
.72111 .04403 L
.72138 .04461 L
.72167 .0451 L
.72183 .04531 L
.72198 .04547 L
.72212 .04558 L
.72228 .04567 L
.72242 .04572 L
.72257 .04574 L
.72272 .04572 L
.72285 .04568 L
.72297 .04561 L
.7231 .04551 L
.72324 .04538 L
.72337 .04523 L
.72369 .04474 L
.724 .04412 L
.72457 .04255 L
.72518 .04028 L
.72586 .03709 L
.72699 .02992 L
.72824 .01919 L
.72955 .00437 L
s
.72955 .00437 m
.72985 0 L
s
.7472 0 m
.74729 .00102 L
.74849 .01153 L
.74915 .01612 L
.74978 .01962 L
.75038 .02239 L
.75093 .02435 L
.75123 .02518 L
.75155 .02593 L
.75188 .02654 L
.75204 .02678 L
.7522 .02695 L
.75235 .02709 L
.7525 .0272 L
.7526 .02724 L
.75268 .02727 L
.75276 .02729 L
.75284 .0273 L
.75292 .02729 L
.75301 .02728 L
.75311 .02725 L
.7532 .02721 L
.75335 .0271 L
.75352 .02694 L
.75383 .02654 L
.75416 .02594 L
.75474 .02445 L
.75535 .02233 L
.75589 .01987 L
.75712 .0124 L
.75826 .00287 L
s
.75826 .00287 m
.75853 0 L
s
.77678 0 m
.77718 .00495 L
.77841 .01646 L
.77958 .02469 L
.78017 .02789 L
.78083 .03073 L
.78138 .0326 L
.78169 .03341 L
.78198 .03405 L
.78224 .03452 L
.78239 .03473 L
.78254 .03491 L
.78268 .03504 L
.7828 .03514 L
.78293 .03521 L
.78305 .03526 L
.78319 .03528 L
.78334 .03527 L
.78349 .03522 L
.78362 .03516 L
.78377 .03504 L
.78393 .03488 L
.78423 .03447 L
.78455 .03388 L
.78488 .03308 L
.78548 .03119 L
.78603 .02891 L
.78654 .02629 L
.78771 .01846 L
.78896 .00684 L
s
.78896 .00684 m
.78951 0 L
s
.80435 0 m
.80562 .02214 L
.80684 .03915 L
.8081 .05296 L
.80929 .06321 L
.81036 .0703 L
.81152 .0759 L
.81217 .07809 L
.81248 .07893 L
.81277 .07957 L
.81303 .08003 L
.81331 .08042 L
.81346 .08057 L
.8136 .08068 L
.81374 .08077 L
.8139 .08082 L
.81398 .08084 L
.81407 .08084 L
.81423 .08082 L
.81431 .08079 L
.8144 .08075 L
.81448 .0807 L
.81458 .08062 L
.81474 .08046 L
.81492 .08025 L
.81523 .07975 L
.81552 .07911 L
.8158 .07838 L
.81643 .07623 L
.81704 .07345 L
.8177 .06962 L
.8189 .06024 L
.82007 .04757 L
.82115 .03204 L
.82232 .00946 L
s
.82232 .00946 m
.82268 0 L
s
.83101 0 m
.8311 .00378 L
.83171 .02759 L
.83281 .06319 L
.83531 .12282 L
.8379 .16659 L
.84033 .19767 L
.84275 .22169 L
.84498 .239 L
.84725 .25273 L
.84855 .25902 L
.84974 .26382 L
.85087 .26758 L
.8521 .27084 L
.85277 .27227 L
.85339 .27337 L
.85398 .27422 L
.85429 .27459 L
.85462 .27492 L
.85491 .27517 L
.85523 .27539 L
.85551 .27555 L
.85578 .27565 L
.85592 .27569 L
.85606 .27572 L
.85618 .27573 L
.85631 .27574 L
.85645 .27574 L
.8566 .27573 L
.85673 .27571 L
.85687 .27568 L
.85704 .27563 L
.8572 .27556 L
.85749 .27541 L
.85781 .2752 L
.85816 .2749 L
.85878 .27421 L
.85936 .27339 L
.86064 .27093 L
.86184 .26781 L
.86408 .25981 L
.86651 .24779 L
.86874 .23335 L
.87102 .21482 L
.87351 .18926 L
.87587 .15872 L
.8784 .1166 L
.87963 .09107 L
.88096 .058 L
s
.88096 .058 m
.88264 0 L
s
.89298 0 m
.89306 .00151 L
.89421 .01737 L
.89484 .02419 L
.89544 .0296 L
.896 .03383 L
.8966 .03761 L
.89711 .0402 L
.89766 .04241 L
.89796 .04341 L
.89824 .04418 L
.89853 .04481 L
.8988 .04527 L
.89896 .04549 L
.8991 .04565 L
.89927 .04579 L
.89936 .04585 L
.89944 .04589 L
.89959 .04594 L
.89973 .04595 L
.89989 .04593 L
.89997 .0459 L
.90005 .04586 L
.90013 .04581 L
.90022 .04575 L
.90038 .0456 L
.90068 .04521 L
.90104 .04457 L
.90136 .04382 L
.90196 .04202 L
.90259 .03952 L
.90372 .03348 L
.90494 .02466 L
.90609 .01397 L
.90714 .00194 L
s
.90714 .00194 m
.90727 0 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{376, 188},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40008d00016R000`400?l00000o`00003oo`3ooooo
0?ooocH0oooo003o0?ooool0oooo=P3oool00?l0ooooo`3ooolf0?ooo`00o`3oool;0?ooo`030=nL
B0000000M;oo00<0oooo00@0_g@00000000007@0_oooo`3ooolQ0?ooo`00o`3oool<0?ooo`090;md
0000M;l0oooo0?ooo`3ookl0M00004RLg`3ookl0M7Bo0?l0oooo8@3oool00?l0oooo3@3oool01P2o
M000M;oo0?ooo`3ookl0M00004RLgol0oooo8`3oool00?l0oooo3P3oool01@2oM000M;oo0?oog`2L
B00007Bo0?l0oooo8`3oool00?l0oooo3`3oool01@2oM4P0W=oo0;md0000B9`0gooo0?l0oooo8P3o
ool00?l0oooo3`3oool01@3ooml0W4Q809bLM00004P0W=oo0?l0oooo8P3oool00?l0oooo403oool0
103ooml0W4P000000018W=oo0?ooob80oooo003o0?oooa40oooo00@0oono07@00000B9`0goooo`3o
oolQ0?ooo`00o`3ooolB0?ooo`030=nLB00007@0_ooo0?l0oooo8@3oool00?l0oooo4P3oool00`3o
_g@0001809cOo`0L0?ooo`050=nLB000000000000000001d_ol0303oool01@3OW4P0000000000000
0000M;oo0>h0oooo003o0?oooa80oooo00<0okmd0000001d_ol0703oool00`3OW4P0001d0;ooo`0@
0?ooo`030?oo_`1d0000M;oo0>h0oooo003o0?oooa00oooo00D0oooO0=oOo`3o_g@0000007Boo`0L
0?ooo`030=nLB000B9`0gooo0100oooo00<0oono07@0001d_ol0kP3oool00?l0oooo403oool01@3o
oml0W4P000000000B9`0gooo01`0oooo00<0gia80018W03Oool0403oool00`3ookl0M00007Boo`3^
0?ooo`00o`3ooolH0?ooo`040?oo_`1dM;l0oooo0;md00<0000000`0M;oo0?ooo`3oool0oooo0;md
00000000001809cOo`3oool0gia80018W03Oool20?ooo`070;md0000000007Bo0?ooo`3OW4P004RL
0=ooo`020?ooo`050?oog`2LB00000000000001d_ol02@3oool00`3ookl0M00007Boo`3^0?ooo`00
o`3ooolI0?ooo`030;md00000000M;oo0080oooo00<0okmd0000001d_ol00P3oool0203OW4P004RL
0=oog`2LM9`0gooo0?noM00007@0_ooo0P3oool01`3OW4P004RL0=noM01d_ol0gia80018W03Oool0
0P3oool01P3ookl0M01809cOo`3oool0_g@0001d_`P0oooo00<0oono07@0001d_ol0kP3oool00?l0
oooo6@3oool00`3OW4P0B9cO0?ooo`030?ooo`070;md0000B9`0gooo0?ooo`3o_g@0001809cOo`02
0?ooo`030?oo_`1d0000B9cO0080oooo00L0omnL04P0B02Lgol0oooo0=nLB000B9`0gooo0080oooo
00<0oono07@0B02Lgol00P3oool00`2oM00007Bo0?ooo`060?ooo`030?oo_`1d0000M;oo0>h0oooo
003o0?oooaT0oooo00<0omnL09boo`3oool00`3oool01`2oM000001d0;ooo`3oool0oono07@00018
W=l00P3oool0303ooml0W4P00000M02oool0oooo0?oog`2LB000B9cO0?ooo`3OW4P004RL0=ooo`80
oooo00<0oooO09a80018W=l00P3oool00`3ooml0W4P004RLg`060?ooo`030?oo_`1d0000M;oo0>h0
oooo003o0?ooo`P0oooo00<0oono07@0000000001P0000000`0005L0E`0000000003000000030000
M02oool0oooo00@0oooo00@0omnL04P00000B9`0gooo0P3oool00`2oM00004RL0=ooo`020?ooo`03
0=nLB000B4P0B7Bo0080oooo00H0_g@0001d_`3oool0gia80018W03Oool30?ooo`030=nLB00007@0
_ooo0080oooo00<0omnL04P0001d_ol01@3oool00`3ookl0M00007Boo`3^0?ooo`00o`3ooolM0?oo
o`040?oOW0180000001809cOo`<0oooo00<0gia80000B02Lgol00P3oool02`3o_g@0001d0;noM018
W=l0oooo0?noM00007@0_ooo0=nLB000B9`0gooo00<0oooo00<0oono07@00018W=l00`3oool00`3O
W4P0001d0;ooo`040?ooo`030?oo_`1d0000M;oo0>h0oooo003o0?oooa`0oooo00@0okmd00000000
07@0_ooo103oool00`3ogi`0B00004RLg`020?ooo`0;0?oog`2LB000M;oo0=nLB000M;l0omnL04P0
B02Lgol0gia80018W03Oool0103oool00`3OW4P00000000000040000000304RLg`3oool0oooo0080
oooo00<0oono07@0001d_ol0kP3oool00?l0oooo6`3oool0103o_g@000000018W03Oool40?ooo`0A
0?noM000000000000018W03Oool0oono07@000000000B9cO0?ooo`3o_g@000000000B02Lgol0gia8
0018W03Oool0103oool00`3ooml0W4P0001d_`030?ooo`030=nLB00007@0_ooo00<0oooo00<0oono
07@0001d_ol0kP3oool00?l0oooo6`3oool00`2oM00004RL0=ooo`020?ooo`030?oog`2LM9`0gooo
00h0oooo00<0gia80018W03Oool01@3oool00`3ogi`0B00007Boo`020?ooo`030?oog`2LB000B9cO
00<0oooo00<0oono07@0001d_ol0kP3oool00?l0oooo6`3oool00`2oM000B9cO0?ooo`020?ooo`03
0?oog`2LB7@0_ooo00h0oooo00<0gia80018W03Oool01P3oool00`3o_g@0001d0;ooo`020?ooo`03
0?noM00007@0_ooo0080oooo00<0oono07@0001d_ol0kP3oool00?l0oooo6`3oool0203o_g@00018
09cOo`3oool0oooO09a800000000M;oo0P3oool0103ookl0M0000018W03Oool80?ooo`030=nLB000
B9`0gooo00L0oooo00L0gia80000M02oool0oooo0?oog`2LB4P0W=oo0080oooo00<0oono07@0001d
_ol0kP3oool00?l0oooo703oool01`3o_g@000000000000004P0W=oo0?oo_`2LW=l03P3oool00`3O
W4P0001d0;ooo`080?ooo`060?noM00004P0W=oo0?oog`2LB000M;oo0P3oool00`3ookl0M00007Bo
o`3^0?ooo`00o`3oool;0?ooo`080?oo_`1d000000000000M02oool0gia800000018W=lN0?ooo`05
0=nLB000000000000000001d_ol01`3oool01P3ookl0M0000000000007@0_ooo0=nLB0<0000000<0
M;oo0?ooo`3oool0k03oool00?l0oooo2`3oool02@3OW4P004RL0=ooo`3ookl0M01d0;noM00007@0
W9ad09cOo`3o0?ooob80oooo003o0?ooo`/0oooo00P0_g@0001d_`3oool0oooo0?oo_`1dB4P0B018
09cOool0oooo8`3oool00?l0oooo2`3oool00`3o_g@004RL0=ooo`020?ooo`030?noM0000000B9cO
0?l0oooo8`3oool00?l0oooo2`3oool00`3ookl0M00007Boo`020?ooo`030?oog`2LB00007Bo0?l0
oooo8`3oool00?l0oooo303oool00`3o_g@0001d0;ooo`020?ooo`030=nLB00007@0_ooo0?l0oooo
8P3oool00?l0oooo3@3oool01`3OW4P0001d0;ooo`3oool0_g@00000B02Lgol0o`3ooolR0?ooo`00
o`3oool>0?ooo`060?oOW0180000000004RL_`2LB000B9cOo`3ooolR0?ooo`00o`3ooolB0?ooo`03
0;md0000M;l0oooo0?l0oooo8@3oool00?l0oooo4P3oool00`3o_g@0001d0;ooo`3o0?ooob40oooo
003o0?oooa80oooo00<0omnL04P0B02Lgol0o`3ooolQ0?ooo`00o`3ooolB0?ooo`030?oog`2LB000
B9cO0?l0oooo8@3oool00?l0oooo4@3oool0103ookl0M00000000000M;oo0?ooob40oooo003o0?oo
ool0oooo=P3oool00?l0ooooo`3ooolf0?ooo`00o`3ooooo0?ooocH0oooo003o0?ooool0oooo=P3o
ool00?l0ooooo`3ooolf0?ooo`00o`3ooooo0?ooocH0oooo003o0?ooool0oooo=P3oool007`0oooo
00<0oooO09a80018W=l0D@3oool00`3OW4P00000000000030000000304RLg`3oool0oooo04X0oooo
00@0_g@0000000000000M;ooDP3oool00`3OW4P00000000000030000000304RLg`3oool0oooo0540
oooo00<0oooO09a80018W=l0GP3oool007d0oooo00<0gia80018W03Oool0D@3oool00`3o_g@007Bo
0?ooo`020?ooo`030;mdM02oool0oooo04T0oooo00H0gia8001d_`3oool0omnL04P00018W=mB0?oo
o`030?noM000M;l0oooo0080oooo00<0_gAd0;ooo`3oool0DP3oool00`3OW4P004RL0=ooo`1M0?oo
o`00O@3oool00`3ogi`0B01809cOo`1B0?ooo`030?noM000M;l0oooo04`0oooo00<0oooO09a80018
W=l00P3oool00`3o_g@0001809cOo`1B0?ooo`030?noM000M;l0oooo05H0oooo00<0omnL04P0B02L
gol0G@3oool007T0oooo00<0okmd000000000000100000000`00B9`0gooo0?ooo`1@0?ooo`030?no
M000M;l0oooo04/0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03OoomB0?ooo`03
0?noM000M;l0oooo0540oooo00<0okmd000000000000100000000`00B9`0gooo0?ooo`1J0?ooo`00
KP3oool00`3o_g@0000000000006000000050000M02oool0oooO09a8B02Lgol00P3oool00`2oM000
07Bo0?ooo`140?ooo`030?noM0000000000000H0000000<0001d0;ooo`3oool00`3oool00`3o_g@0
B9cO0?ooo`1:0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol0D`3oool00`3o_g@0
B9cO0?ooo`1@0?ooo`030?oog`2LB4P0W=oo0080oooo00<0_g@0001d_`3oool0G03oool007X0oooo
00L0oono07@0M02oool0oooo0?noM000B9`0gooo05@0oooo00<0gia804RLg`3oool0BP3oool00`2o
M00004RL0=ooo`020?ooo`030?oOW0180000B9cO05@0oooo00<0gia804RLg`3oool0D03oool01`3o
okl0M01d0;ooo`3oool0okmd0018W03Oool0G03oool007/0oooo00H0omnL04Q8W03Oool0omnL04P0
B02LgomE0?ooo`030;md001d_ol0oooo04T0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0
001d_`1E0?ooo`030;md001d_ol0oooo0500oooo00H0omnL04Q8W03Oool0omnL04P0B02LgomL0?oo
o`00O03oool01@3o_g@007Bo0?oog`2LB000B9cO05D0oooo00<0omnL04P0M02oool0B@3oool00`3o
gi`0B00004RLg`030?ooo`030;md0000B9`0gooo05@0oooo00<0omnL04P0M02oool0D@3oool01@3o
_g@007Bo0?oog`2LB000B9cO05`0oooo001m0?ooo`040=nLB018W=l0_g@0001d_eD0oooo00<0oooO
09a8001d_ol0B@3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo05@0oooo00<0
oooO09a8001d_ol0DP3oool0103OW4P0B9cO0;md0000M;mL0?ooo`00OP3oool0103OW4P0M7Ad0000
M02ooomD0?ooo`030?oo_`1d0000B9cO04X0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo
0?ooo`1D0?ooo`030?oo_`1d0000B9cO05<0oooo00@0gia807AdM00007@0_oooF`3oool007l0oooo
00<0_g@00000B02Lgol0D03oool01`3o_g@004RL0=ooo`3oool0gia80000B02Lgol0B`3oool01P3O
W4P0001d0;ooo`3ooml0W4P004RLge40oooo00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo05@0
oooo00<0_g@00000B02Lgol0F`3oool007l0oooo00<0oooO09a80018W=l0D@3oool01@3ogi`0B000
000000000000M;oo04d0oooo00@0gia800000000000007BoD`3oool01@3ogi`0B000000000000000
M;oo05D0oooo00<0oooO09a80018W=l0F`3oool00?l0ooooo`3ooolf0?ooo`00o`3ooooo0?ooocH0
oooo003o0?ooool0oooo=P3oool00?l0ooooo`3ooolf0?ooo`00o`3ooooo0?ooocH0oooo003o0?oo
onP0oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`040?ooo`040;md00000000
000007Boo`D0oooo00@0_g@0000000000000M;oo0`3oool00`3ookl0M000000000020000000304RL
g`3oool0oooo02L0oooo003o0?ooonL0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`03
0?oOW0180000M;oo00<0oooo00H0gia8001d_`3oool0omnL04P00018W=l30?ooo`060=nLB000M;l0
oooo0?oOW0180000B9cO103oool00`3o_g@0000007Boo`0Z0?ooo`00o`3ooooV0?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol01P3oool00`3ooml0W4P004RLg`020?ooo`070?noM000
04P0W=oo0?ooo`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00<0oooo00<0oooO09a80000
M;l0:P3oool00?l0ooooiP3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`D0
oooo00<0oooO09a80018W=l00P3oool01`3ooml0W4P00018W03Oool0oooO09a80018W=l00P3oool0
103ooml0W4P00018W03Oool30?ooo`030=nLB000B9`0gooo02T0oooo003o0?ooonH0oooo00<0oooO
09a80000M;l00`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2LB00007Bo00<0oooo00H0gia8
0000B02Lgol0oooO09a80000M;l30?ooo`030=nLB00004P0W=oo00<0oooo00<0omnL04P0B02Lgol0
:@3oool006h0ooooo`00001b000000L0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RL
g`060?ooo`030;md0000B9`0gooo0080oooo00L0omnL04P00018W=l0oooo0;md0000B9`0gooo0080
oooo00<0omnL04P00018W=l00`3oool00`3ookl0M00007Boo`0Y0?ooo`00KP3oool00`000000oooo
0?ooo`060?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03P3oool010000000oooo
0?ooo`0000050?ooo`040000003oool0oooo000000T0oooo00<000000?ooo`3oool01P3oool01000
0000oooo0?ooo`0000060?ooo`040000003oool0oooo00000080oooo00D000000?ooo`3oool0oooo
000000060?ooo`040000003oool0oooo0?ooo`80000000<0oooo0000003oool00P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool200000003
0?ooo`000000oooo00L0oooo0P0000090?ooo`030000003oool0oooo00P0oooo0P0000090?ooo`80
00002@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool01`3oool2000000T0oooo0P00000:0?ooo`030000003oool0oooo00L0oooo0P0000080?ooo`04
0000003oool00000000000H0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?oo
o`80000000@0oooo0000003oool0oooo0P0000030?ooo`030000003oool0oooo00@0oooo00D00000
0?ooo`3oool0oooo000000020?ooo`040000003oool0oooo000000D0oooo00D000000?ooo`3oool0
oooo000000080?ooo`030000003oool0oooo00H0oooo00D000000?ooo`3oool0oooo000000040?oo
o`050000003oool0oooo0?ooo`0000005@3oool00`000000oooo0?ooo`060?ooo`030000003oool0
oooo00D0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`060?ooo`030=nLB00004P0
W=oo0080oooo00L0oooO09a80000M;l0oooo0=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l0
103oool00`2oM00007Bo0?ooo`0X0?ooo`00KP3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`020?ooo`050000003o
ool0oooo0?ooo`0000004P3oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo00<0
oooo00D000000?ooo`3oool0oooo000000060?ooo`050000003oool0oooo0?ooo`0000000P3oool0
1@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`040?ooo`040000003oool0oooo000000P0oooo0P00000D0?ooo`8000002@3oool2000000T0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo
0P0000090?ooo`800000503oool2000000P0oooo00@000000?ooo`3oool000001P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0080oooo0P000000103oool000000?ooo`3oool20000
00<0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000D0oooo00<00000
0?ooo`3oool00`3oool01@000000oooo0?ooo`3oool000000140oooo00D000000?ooo`3oool0oooo
000000040?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`050?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool01@3oool00`3ogi`0B000
04RLg`030?ooo`060;md0000B9`0gooo0?oOW0180000B9cO0`3oool00`2oM00004RL0=ooo`030?oo
o`030=nLB000B9`0gooo02P0oooo001^0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000004P3oool01000
0000oooo0?ooo`0000060?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo0000
00060?ooo`050000003oool0oooo0?ooo`0000001P3oool01@000000oooo0?ooo`3oool0000000D0
oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`0000070?ooo`030000003oool00000
01@0oooo0P0000090?ooo`8000002@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo
00L0oooo0P0000090?ooo`800000503oool2000000P0oooo00@000000?ooo`3oool000001P3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo0P0000040?ooo`8000000P3oool0
0`000000oooo0?ooo`050?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?oo
o`030?ooo`050000003oool0oooo0?ooo`0000004@3oool01@000000oooo0?ooo`3oool0000000@0
oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`050?ooo`040?oog`2LB00004RL0=oo
o`80oooo00<0gia80018W03Oool01@3oool0103ooml0W4P00018W03Oool20?ooo`070=nLB000B9`0
gooo0?oog`2LB00004RL0=ooo`020?ooo`030=nLB000B9`0gooo00<0oooo00<0omnL04P0B02Lgol0
:03oool006h0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo0080oooo00D000000?ooo`3oool0oooo0000000B0?ooo`040000003oool0oooo000000H0
oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool0000000L0oooo00@000000?oo
o`3oool000001@3oool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`050000003oool0oooo0?ooo`0000001`3oool00`000000oooo0000000D0?ooo`8000002@3o
ool2000000T0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`070?ooo`8000002@3o
ool2000001@0oooo0P0000080?ooo`040000003oool0oooo000000H0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo0P0000020?ooo`030000
003oool0oooo00D0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo00<0
oooo00D000000?ooo`3oool0oooo0000000A0?ooo`030000003oool0oooo0080oooo00D000000?oo
o`3oool0oooo0000000R0?ooo`030000003oool0oooo00H0oooo00<0omnL04P0001d_ol00P3oool0
0`3OW4P007Bo0?ooo`060?ooo`030?oOW0180000M;oo0080oooo00L0gia8001d_`3oool0oooo0?oO
W0180000M;oo0080oooo00<0gia8001d_`3oool00`3oool00`3ookl0M00007Boo`0X0?ooo`00KP3o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0140oooo00D000000?ooo`3oool0oooo
000000030?ooo`030000003oool0oooo0100oooo00@000000?ooo`3oool000001P3oool00`000000
oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000001`3oool010000000oooo0?ooo`000005
0?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00D00000
0?ooo`3oool0oooo000000070?ooo`030000003oool0000001@0oooo0P0000090?ooo`8000002@3o
ool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00L0oooo0P0000090?ooo`800000503o
ool2000000P0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`8000000P3oool00`000000oooo
0?ooo`050?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`030?ooo`05
0000003oool0oooo0?ooo`0000004@3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo
0?ooo`0000008P3oool00`000000oooo0?ooo`070?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO
203oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`<0oooo00H0gia80000M02oool0oooO09a80018
W=l50?ooo`030;md0000M;l0oooo02L0oooo001^0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool04@3oool01@000000oooo0?ooo`3oool0000000<0oooo00<000000?ooo`3oool0403o
ool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool00`3oool01@000000oooo
0?ooo`3oool0000000L0oooo00@000000?ooo`00000000001@3oool2000000<0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000001`3oool0
0`000000oooo0000000D0?ooo`8000002@3oool2000000T0oooo00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`070?ooo`8000002@3oool2000001@0oooo0P0000080?ooo`050000003oool0
oooo0?ooo`0000001@3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`000000
1P3oool010000000oooo0?ooo`0000070?ooo`050000003oool0oooo0?ooo`0000001@3oool00`00
0000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000004@3oool00`000000oooo0?ooo`02
0?ooo`050000003oool0oooo0?ooo`0000004`3oool00`000000oooo0?ooo`0<0?ooo`030000003o
ool0oooo00P0oooo00@0gia800000000000007Bo2P3oool0103OW4P0000000000000M;l50?ooo`04
0=nLB00000000000001d_`@0oooo00D0oono07@000000000001d0;ooo`0W0?ooo`00KP3oool00`00
0000oooo0?ooo`0P0?ooo`050000003oool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`0@
0?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`030?ooo`050000003o
ool0oooo0?ooo`0000001`3oool010000000oooo0000000000060?ooo`050000003oool0oooo0?oo
o`0000001@3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000001`3oool0
0`000000oooo0000000D0?ooo`800000203oool00`000000oooo000000090?ooo`030000003oool0
oooo01<0oooo00<000000?ooo`3oool01`3oool2000000T0oooo0P00000D0?ooo`800000203oool0
1@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool00`3oool01@000000oooo0?oo
o`3oool0000000H0oooo0P0000000`3oool000000?ooo`060?ooo`050000003oool0oooo0?ooo`00
00001@3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000004@3oool00`00
0000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000004`3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo05<0oooo001^0?ooo`030000003oool0oooo0200oooo00D000000?oo
o`3oool0oooo000000030?ooo`030000003oool0oooo0100oooo00D000000?ooo`3oool0oooo0000
00050?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool00000203oool010000000oooo
0000000000060?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`030?oo
o`050000003oool0oooo0?ooo`0000001`3oool00`000000oooo0000000D0?ooo`800000203oool0
0`000000oooo000000090?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool01`3oool2
000000T0oooo0P00000D0?ooo`800000203oool01@000000oooo0?ooo`3oool0000000D0oooo00<0
00000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool0000000H0oooo0P0000000`3oool00000
0?ooo`070?ooo`030000003oool0000000H0oooo00<000000?ooo`3oool00`3oool01@000000oooo
0?ooo`3oool000000140oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool00000
01<0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1C0?ooo`00KP3oool00`000000
oooo0?ooo`0P0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`0?0?oo
o`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`040?ooo`030000003oool0
000000P0oooo100000060?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?oo
o`030?ooo`050000003oool0oooo0?ooo`0000001`3oool00`000000oooo0000000D0?ooo`800000
203oool00`000000oooo000000090?ooo`030000003oool0oooo0180oooo0P0000090?ooo`800000
2@3oool2000001@0oooo0P0000080?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo008000001`3oool3000000P0oooo00<000000?ooo`00
00001P3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000004@3oool00`00
0000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000004`3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo05<0oooo001^0?ooo`030000003oool0oooo0200oooo00D000000?oo
o`3oool0oooo000000040?ooo`030000003oool0oooo00l0oooo00D000000?ooo`3oool0oooo0000
00050?ooo`030000003oool0oooo00@0oooo00<000000?ooo`0000002@3oool00`000000oooo0000
00060?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`030?ooo`050000
003oool0oooo0?ooo`0000001`3oool00`000000oooo0000000D0?ooo`800000203oool00`000000
oooo000000090?ooo`030000003oool0oooo0180oooo0P0000090?ooo`8000002@3oool2000001@0
oooo0P0000080?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00800000203oool00`000000oooo0?ooo`070?ooo`030000003oool00000
00H0oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool000000140oooo00<00000
0?ooo`3oool00P3oool01@000000oooo0?ooo`3oool000000180oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`1C0?ooo`00KP3oool00`000000oooo0?ooo`0P0?ooo`040000003oool0
oooo000000D0oooo00<000000?ooo`3oool03`3oool01@000000oooo0?ooo`3oool0000000D0oooo
00<000000?ooo`3oool0103oool00`000000oooo0000000;0?ooo`030000003oool0oooo00@0oooo
00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3o
ool0oooo000000070?ooo`030000003oool0000001@0oooo0P0000080?ooo`030000003oool00000
00T0oooo00<000000?ooo`3oool04P3oool2000000T0oooo0P0000090?ooo`800000503oool00`00
0000oooo000000070?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo008000004P3oool00`000000oooo000000060?ooo`030000003oool0
oooo00<0oooo00D000000?ooo`3oool0oooo0000000A0?ooo`030000003oool0oooo0080oooo00@0
00000?ooo`3oool000004`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo05<0oooo
001^0?ooo`030000003oool0oooo0240oooo00<000000?ooo`0000001@3oool00`000000oooo0?oo
o`0?0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`050?ooo`800000
303oool00`000000oooo0?ooo`040?ooo`8000000P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00D000000?ooo`3oool0oooo000000070?ooo`030000003oool0000001@0
oooo0P0000080?ooo`030000003oool0000000T0oooo00<000000?ooo`3oool04P3oool2000000T0
oooo0P0000090?ooo`800000503oool00`000000oooo000000070?ooo`050000003oool0oooo0?oo
o`0000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo008000004P3oool30000
00L0oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool000000140oooo00<00000
0?ooo`3oool00P3oool010000000oooo0?ooo`00000C0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool0D`3oool006h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0000000D0oooo00<000000?ooo`3oool03`3oool01@000000oooo0?ooo`3o
ool0000000@0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0B0?ooo`030000003o
ool0000000L0oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000000L0oooo
00<000000?ooo`000000503oool2000000P0oooo00<000000?ooo`0000002@3oool00`000000oooo
0?ooo`0B0?ooo`8000002@3oool2000000T0oooo0P00000D0?ooo`030000003oool0000000L0oooo
00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3o
ool00000503oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0080oooo00D000000?oo
o`3oool0oooo0000000A0?ooo`030000003oool0oooo0080oooo00@000000?ooo`3oool000004`3o
ool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo05<0oooo001^0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool04@3oool00`000000oooo000000050?ooo`030000003oool0
oooo00l0oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo01/0oooo00<0
00000?ooo`0000001`3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`000000
1`3oool00`000000oooo0000000D0?ooo`800000203oool00`000000oooo000000090?ooo`030000
003oool0oooo0180oooo0P0000090?ooo`8000002@3oool2000001@0oooo00<000000?ooo`000000
1`3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0000000M0?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo0000000A0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`0000008`3oool00`000000oooo0?ooo`1C0?oo
o`00KP3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0140oooo00<000000?ooo`00
00001@3oool00`000000oooo0?ooo`0?0?ooo`050000003oool0oooo0?ooo`000000103oool00`00
0000oooo0?ooo`0K0?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool00P3oool01@00
0000oooo0?ooo`3oool0000000L0oooo00<000000?ooo`000000503oool2000000P0oooo00<00000
0?ooo`0000002@3oool00`000000oooo0?ooo`0B0?ooo`8000002@3oool2000000T0oooo0P00000D
0?ooo`030000003oool0000000L0oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003o
ool0oooo00D0oooo0P00000N0?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo
0000000A0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`0000008`3oool00`000000oooo
0?ooo`1C0?ooo`00KP3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0140oooo0`00
00050?ooo`030000003oool0oooo00l0oooo00D000000?ooo`3oool0oooo000000040?ooo`030000
003oool0oooo01`0oooo0P0000070?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0
oooo000000070?ooo`030000003oool0000001@0oooo0P0000080?ooo`030000003oool0000000T0
oooo00<000000?ooo`3oool04P3oool2000000T0oooo0P0000090?ooo`800000503oool00`000000
oooo000000070?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`050?oo
o`800000203oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0080oooo00D000000?oo
o`3oool0oooo0000000A0?ooo`030000003oool0oooo00<0oooo0P00000T0?ooo`030000003oool0
oooo05<0oooo001^0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool04P3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo00l0oooo00D000000?ooo`3oool0oooo00000004
0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3o
ool0000000L0oooo00<000000?ooo`000000503oool2000000P0oooo00<000000?ooo`000000203o
ool2000001@0oooo0P0000090?ooo`8000002@3oool2000001@0oooo00<000000?ooo`0000001P3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo
0000000A0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool08P3oool00`000000oooo
0?ooo`1C0?ooo`00KP3oool3000000h0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`0?0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`0U0?ooo`030000
003oool0oooo0080oooo00D000000?ooo`3oool0oooo000000070?ooo`030000003oool0000001@0
oooo0P0000080?ooo`030000003oool0000000P0oooo0P00000D0?ooo`8000002@3oool2000000T0
oooo0P00000D0?ooo`030000003oool0000000H0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool04`3oool00`000000
oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000004@3oool00`000000oooo0?ooo`0W0?oo
o`<00000E@3oool006h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0H0?ooo`03
0000003oool0oooo00l0oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool0203oool01@000000oooo0?ooo`3oool0000000@0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0000000D0?ooo`800000203oool00`000000oooo00000008
0?ooo`800000503oool2000000T0oooo0P0000090?ooo`800000503oool00`000000oooo00000006
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`02
0?ooo`050000003oool0oooo0?ooo`0000004@3oool00`000000oooo0?ooo`0Y0?ooo`030000003o
ool0oooo05<0oooo001^0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0603oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo0000
000D0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0203oool01@000000oooo0?oo
o`3oool0000000@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0000000D0?ooo`800000
203oool00`000000oooo000000080?ooo`800000503oool2000000T0oooo0P0000090?ooo`800000
503oool00`000000oooo000000060?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
4P3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0D`3oool006h0oooo
00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo
00D000000?ooo`3oool0oooo0000000D0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool0203oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0000000C0?ooo`030000003oool0000000P0oooo00<000000?ooo`000000203oool20000
01@0oooo0P0000090?ooo`030000003oool0000000P0oooo0P00000D0?ooo`030000003oool00000
00H0oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool0000000T0oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`0B0?ooo`050000003oool0oooo0?ooo`000000103o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`1C0?ooo`00KP3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`00000E0?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool0203oool01@000000oooo0?ooo`3oool0000000@0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0000000C0?ooo`030000003oool0000000P0oooo
00<000000?ooo`000000203oool2000001@0oooo0P0000090?ooo`030000003oool0000000P0oooo
0P00000D0?ooo`030000003oool0000000H0oooo00<000000?ooo`3oool00`3oool01@000000oooo
0?ooo`3oool0000000T0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0C0?ooo`04
0000003oool0oooo000000@0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool0D`3oool006h0oooo00<000000?ooo`3o
ool0:@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00@000000?ooo`3o
ool000005@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00P0oooo00D000000?oo
o`3oool0oooo000000040?ooo`030000003oool0oooo00D0oooo00<000000?ooo`0000004`3oool0
0`000000oooo000000080?ooo`030000003oool0000000P0oooo0P00000D0?ooo`8000002@3oool0
0`000000oooo000000080?ooo`800000503oool00`000000oooo000000060?ooo`030000003oool0
oooo00<0oooo00D000000?ooo`3oool0oooo000000090?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool04`3oool010000000oooo0?ooo`0000040?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo05<0oooo
001^0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`020?ooo`040000003oool0oooo000001D0oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`070?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`050?ooo`03
0000003oool0000001<0oooo00<000000?ooo`000000203oool00`000000oooo000000080?ooo`80
0000503oool2000000T0oooo00<000000?ooo`000000203oool2000001@0oooo00<000000?ooo`00
00001P3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo000000X0oooo00<000000?oo
o`3oool0703oool010000000oooo0?ooo`0000040?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo05<0oooo001^0?oo
o`030000003oool0oooo02T0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?oo
o`040000003oool0oooo000001D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`07
0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`050?ooo`030000003o
ool0000001<0oooo00<000000?ooo`000000203oool00`000000oooo000000080?ooo`800000503o
ool2000000T0oooo00<000000?ooo`000000203oool2000001@0oooo00<000000?ooo`0000001P3o
ool00`000000oooo0?ooo`030?ooo`040000003oool0oooo000000X0oooo00<000000?ooo`3oool0
703oool010000000oooo0?ooo`0000040?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool05`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo05<0oooo001^0?ooo`030000
003oool0oooo02T0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`040000
003oool0oooo000001D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`05
0000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`050?ooo`030000003oool00000
01<0oooo00<000000?ooo`000000203oool00`000000oooo000000080?ooo`800000503oool20000
00T0oooo00<000000?ooo`000000203oool2000001@0oooo00<000000?ooo`0000001P3oool00`00
0000oooo0?ooo`030?ooo`040000003oool0oooo000000X0oooo00<000000?ooo`3oool0703oool0
0`000000oooo000000050?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo05<0oooo001^0?ooo`030000003oool0oooo
02T0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo
000001D0oooo00<000000?ooo`3oool0403oool01@000000oooo0?ooo`3oool0000000D0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0000000C0?ooo`030000003oool0000000P0oooo00<0
00000?ooo`000000203oool2000001@0oooo0P0000090?ooo`030000003oool0000000P0oooo0P00
000D0?ooo`030000003oool0000000H0oooo00<000000?ooo`3oool00`3oool010000000oooo0?oo
o`00000Y0?ooo`030000003oool0000000D0oooo00<000000?ooo`3oool03P3oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0D`3oool006h0oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool0103oool2000002X0oooo00@000000?ooo`3oool000001@3oool00`000000oooo
0?ooo`040?ooo`030000003oool0000001<0oooo00<000000?ooo`000000203oool00`000000oooo
000000080?ooo`800000503oool2000000T0oooo00<000000?ooo`000000203oool2000001@0oooo
00<000000?ooo`0000001P3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo000002T0
oooo0`0000050?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0:P3oool00`000000
oooo0?ooo`1C0?ooo`00KP3oool3000000l0oooo00<000000?ooo`3oool05`3oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo00@0oooo0P00000Z0?ooo`040000003oool0oooo000000D0
oooo00<000000?ooo`3oool0103oool00`000000oooo0000000C0?ooo`030000003oool0000000P0
oooo00<000000?ooo`000000203oool2000001@0oooo0P0000090?ooo`030000003oool0000000P0
oooo0P00000D0?ooo`030000003oool0000000H0oooo00<000000?ooo`3oool00`3oool010000000
oooo0?ooo`00000Z0?ooo`8000001@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
02P0oooo0`00001E0?ooo`00KP3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01L0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0`0?ooo`040000003oool0oooo0000
00D0oooo00<000000?ooo`3oool0103oool00`000000oooo0000000C0?ooo`030000003oool00000
00P0oooo00<000000?ooo`000000203oool2000001@0oooo0P0000090?ooo`030000003oool00000
00P0oooo0P00000D0?ooo`030000003oool0000000H0oooo00<000000?ooo`3oool00`3oool01000
0000oooo0?ooo`00000D0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo05<0oooo001^0?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
0300oooo00<000000?ooo`0000001P3oool00`000000oooo0?ooo`040?ooo`030000003oool00000
01<0oooo00<000000?ooo`000000203oool00`000000oooo000000080?ooo`800000503oool20000
00T0oooo00<000000?ooo`000000203oool2000001@0oooo00<000000?ooo`0000001P3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0000001@0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0D`3oool006h0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool0<@3oool2000000H0oooo00<000000?ooo`3oool0103oool00`000000
oooo0000000C0?ooo`030000003oool0000000P0oooo00<000000?ooo`000000203oool2000001@0
oooo0P0000090?ooo`030000003oool0000000P0oooo0P00000D0?ooo`030000003oool0000000H0
oooo00<000000?ooo`3oool0103oool2000001D0oooo00<000000?ooo`3oool06P3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0D`3oool006h0oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool0<@3oool2000000H0oooo00<000000?ooo`3oool0103oool00`000000oooo
0000000C0?ooo`030000003oool0000000P0oooo00<000000?ooo`000000203oool00`000000oooo
0000000C0?ooo`8000002@3oool00`000000oooo000000080?ooo`030000003oool0000001<0oooo
00<000000?ooo`0000001P3oool00`000000oooo0?ooo`040?ooo`8000005@3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0:P3oool00`000000oooo
0?ooo`1C0?ooo`00KP3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01H0oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`040?ooo`030000003oool0000001<0oooo00<0
00000?ooo`000000203oool00`000000oooo000000080?ooo`030000003oool0000001<0oooo0P00
00090?ooo`030000003oool0000000P0oooo00<000000?ooo`0000004`3oool00`000000oooo0000
00060?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo05<0oooo001^0?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01/0oooo00<00000
0?ooo`3oool01`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`0000004`3oool00`000000oooo000000080?ooo`030000003oool0000000P0oooo00<00000
0?ooo`0000004`3oool2000000T0oooo00<000000?ooo`000000203oool00`000000oooo0000000C
0?ooo`030000003oool0000000H0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0F
0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0D`3oool006h0oooo00<000000?oo
o`3oool0:@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01/0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`0000004`3oool00`000000oooo000000080?ooo`030000003oool0000000P0oooo00<000000?oo
o`0000004`3oool00`000000oooo000000080?ooo`030000003oool0000000P0oooo00<000000?oo
o`0000004`3oool00`000000oooo000000050?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool05P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo05<0oooo001^0?ooo`<0
0000:@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`00
00004`3oool00`000000oooo000000080?ooo`030000003oool0000000P0oooo00<000000?ooo`00
00004`3oool00`000000oooo000000080?ooo`030000003oool0000000P0oooo00<000000?ooo`00
00004`3oool00`000000oooo000000050?ooo`030000003oool0oooo0140oooo00<000000?ooo`3o
ool02@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3o
ool05P3oool00`000000oooo0?ooo`0?0?ooo`<00000E@3oool00400oooo00@0_g@0000000000000
M;oo0`3oool00`3ogi`0B00007Boo`040?ooo`040;md00000000000007Boo`D0oooo00@0_g@00000
00000000M;oo0`3oool01P3ogi`0B00000000000000004RL0=ooo`X0oooo00<000000?ooo`3oool0
:P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`000000
4`3oool00`000000oooo000000080?ooo`030000003oool0000000P0oooo00<000000?ooo`000000
4`3oool00`000000oooo000000080?ooo`030000003oool0000000P0oooo00<000000?ooo`000000
4`3oool00`000000oooo000000050?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
5@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00H0oooo00@0_g@0000000000000
M;oo0`3oool00`3ogi`0B00007Boo`040?ooo`040;md00000000000007Boo`D0oooo00@0_g@00000
00000000M;oo0`3oool01P3ogi`0B00000000000000004RL0=ooobT0oooo000o0?ooo`060=nLB000
M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`030?ooo`060=nLB000M;l0oooo0?oO
W0180000B9cO0`3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00H0okmd0000001d_ol0
oooo0?oOW018M;l:0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0403oool00`00
0000oooo0?ooo`040?ooo`030000003oool0000001<0oooo00<000000?ooo`000000203oool00`00
0000oooo000000080?ooo`030000003oool0000001<0oooo00<000000?ooo`000000203oool00`00
0000oooo000000080?ooo`030000003oool0000001<0oooo00<000000?ooo`0000001@3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool01@3oool01P3O
W4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00`3oool01P3OW4P007Bo0?oo
o`3ogi`0B00004RLg`<0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`060?noM0000000
M;oo0?ooo`3ogi`0B7Bo:@3oool003h0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cO
o`060?ooo`030?oog`2LB000B9cO0080oooo00L0okmd0000B02Lgol0oooo0?oog`2LB000B9cO0080
oooo00<0okmd0000B02Lgol01P3oool00`3o_g@0B9cO0?ooo`080?ooo`030000003oool0oooo02X0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo0000
01<0oooo00<000000?ooo`000000203oool00`000000oooo000000080?ooo`030000003oool00000
01<0oooo00<000000?ooo`000000203oool00`000000oooo000000080?ooo`030000003oool00000
0180oooo00@000000?ooo`3oool000001@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0
oooo02H0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo0180oooo00<000000?ooo`3oool0103oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0
W=oo00H0oooo00<0oooO09a80018W=l00P3oool01`3o_g@0001809cOo`3oool0oooO09a80018W=l0
0P3oool00`3o_g@0001809cOo`060?ooo`030?noM018W=l0oooo02L0oooo000n0?ooo`030?oog`2L
B000B9cO0080oooo00@0oooO09a80000B9`0gooo1@3oool00`3ooml0W4P004RLg`020?ooo`070?oo
g`2LB00004RL0=ooo`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`D0oooo00<0oono
07@0M02oool0203oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool06P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool00`3oool010000000oooo0?ooo`00000C0?ooo`030000003oool0000000P0oooo00<0
00000?ooo`000000203oool00`000000oooo0000000C0?ooo`030000003oool0000000P0oooo00<0
00000?ooo`000000203oool00`000000oooo0000000B0?ooo`040000003oool0oooo000000D0oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00@0oooo
00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool50?ooo`030?oog`2LB000B9cO0080
oooo00L0oooO09a80000B9`0gooo0?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo1@3o
ool00`3ookl0M01d0;ooo`0W0?ooo`00?P3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0
W=oo00D0oooo00<0oooO09a80000M;l00`3oool01P3OW4P0001809cOo`3ooml0W4P0001d_`<0oooo
00<0gia80000B02Lgol01@3oool00`3ooml0W4P007Boo`080?ooo`@000004@3oool00`000000oooo
0?ooo`0E0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0L0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool000004`3oool00`000000
oooo000000080?ooo`030000003oool0000000P0oooo00<000000?ooo`0000004`3oool00`000000
oooo000000080?ooo`030000003oool0000000P0oooo00<000000?ooo`0000004P3oool010000000
oooo0?ooo`0000050?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`0W0?ooo`@000001P3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo
00D0oooo00<0oooO09a80000M;l00`3oool01P3OW4P0001809cOo`3ooml0W4P0001d_`<0oooo00<0
gia80000B02Lgol01@3oool00`3ooml0W4P007Boo`0W0?ooo`00?`3oool00`2oM00004RL0=ooo`02
0?ooo`030?oOW0180000B9cO00H0oooo00<0_g@00018W03Oool00P3oool01`3ogi`0B00004RLg`3o
ool0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`050?ooo`030?oo_`1d0000B9cO00P0oooo
00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo000001<0
oooo00<000000?ooo`000000203oool00`000000oooo000000080?ooo`030000003oool0000001<0
oooo00<000000?ooo`000000203oool00`000000oooo000000080?ooo`030000003oool000000180
oooo00@000000?ooo`3oool000001@3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`050?ooo`030;md0000B9`0gooo
0080oooo00<0omnL04P00018W=l01P3oool00`2oM00004RL0=ooo`020?ooo`070?oOW0180000B9cO
0?ooo`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00D0oooo00<0oono07@00018W=l09`3o
ool003l0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`060?ooo`030=nLB00004P0
W=oo0080oooo00L0oooO09a80000M;l0oooo0=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l0
1@3oool00`3o_g@0000007Boo`080?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0
5@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0
0P3oool01@000000oooo0?ooo`3oool000000180oooo00<000000?ooo`000000203oool00`000000
oooo000000080?ooo`030000003oool0000001<0oooo00<000000?ooo`000000203oool00`000000
oooo000000080?ooo`030000003oool000000180oooo00@000000?ooo`3oool000001@3oool00`00
0000oooo0?ooo`0j0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0:P3oool00`00
0000oooo0?ooo`050?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l01P3oool00`3O
W4P0001809cOo`020?ooo`070?oog`2LB00007Bo0?ooo`3OW4P0001809cOo`020?ooo`030?oog`2L
B00007Bo00D0oooo00<0okmd0000001d_ol09`3oool003l0oooo00<0omnL04P00018W=l00`3oool0
0`2oM00004RL0=ooo`050?ooo`030?oOW0180000B9cO00<0oooo00H0_g@00018W03Oool0omnL04P0
0018W=l30?ooo`0:0;md0000B9`0gooo0?ooo`3ogml0gooo0?noM0000000001d0;ooo`P0oooo00<0
00000?ooo`3oool04P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool0>P3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`000000
4P3oool00`000000oooo000000080?ooo`030000003oool0000000P0oooo00<000000?ooo`000000
4`3oool00`000000oooo000000080?ooo`030000003oool0000000P0oooo00<000000?ooo`000000
4P3oool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3o
ool01@3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00D0oooo00<0omnL04P00018
W=l00`3oool01P2oM00004RL0=ooo`3ogi`0B00004RLg`<0oooo00X0_g@00018W03Oool0oooo0?oO
g`3Oool0okmd0000000007@0_ooo9`3oool003l0oooo00@0oooO09a80000B9`0gooo0P3oool00`3O
W4P004RL0=ooo`050?ooo`040?oog`2LB00004RL0=ooo`80oooo00L0gia80018W03Oool0oooO09a8
0000B9`0gooo0080oooo00T0gia80018W03Oool0oooo0?oo_`1d000000000000B02Lgol02@3oool0
0`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0303oool0
0`000000oooo0?ooo`0k0?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo0000
000B0?ooo`030000003oool0000000P0oooo00<000000?ooo`000000203oool00`000000oooo0000
000C0?ooo`030000003oool0000000P0oooo00<000000?ooo`000000203oool00`000000oooo0000
000B0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`0M0?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0Z0?ooo`030000
003oool0oooo00D0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`050?ooo`04
0?oog`2LB00004RL0=ooo`80oooo00L0gia80018W03Oool0oooO09a80000B9`0gooo0080oooo00T0
gia80018W03Oool0oooo0?oo_`1d000000000000B02Lgol0:03oool00400oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`060?ooo`030?oOW0180000M;oo0080oooo00L0gia8001d_`3o
ool0oooo0?oOW0180000M;oo0080oooo00<0gia8001d_`3oool00P3oool00`3OW9`0_ooo0?ooo`0:
0?ooo`<000004P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0>`3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000004P3o
ool00`000000oooo000000080?ooo`030000003oool0000000P0oooo00<000000?ooo`0000004`3o
ool00`000000oooo000000070?ooo`040000003oool0oooo000000P0oooo00<000000?ooo`000000
4P3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool07P3oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0:03oool3000000P0
oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`060?ooo`030?oOW0180000M;oo0080
oooo00L0gia8001d_`3oool0oooo0?oOW0180000M;oo0080oooo00<0gia8001d_`3oool00P3oool0
0`3OW9`0_ooo0?ooo`0Y0?ooo`00@@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`P0oooo00H0
gia80000M02oool0oooO09a80018W=l30?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO0`3oool0
0`3ogi`0M9cO0?ooo`0:0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0503oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000004P3oool00`000000oooo0000
00080?ooo`030000003oool0000000P0oooo00<000000?ooo`0000004`3oool00`000000oooo0000
00070?ooo`040000003oool0oooo000000P0oooo00<000000?ooo`0000004P3oool01@000000oooo
0?ooo`3oool0000000@0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0D0?ooo`03
0000003oool0oooo01<0oooo00<000000?ooo`3oool01`3oool01P3OW4P0001d0;ooo`3ooml0W4P0
04RLg`P0oooo00H0gia80000M02oool0oooO09a80018W=l30?ooo`060=nLB00007@0_ooo0?oog`2L
B000B9cO0`3oool00`3ogi`0M9cO0?ooo`0Y0?ooo`00@P3oool0103OW4P0000000000000M;l:0?oo
o`040=nLB00000000000001d_`D0oooo00@0gia800000000000007Bo1@3oool01@2oM00000000000
00000000B9cO00L0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0D0?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo0100oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool000000180
oooo00<000000?ooo`000000203oool00`000000oooo000000080?ooo`030000003oool0000001<0
oooo00<000000?ooo`0000001`3oool010000000oooo0?ooo`0000080?ooo`030000003oool00000
0180oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool0503oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00P0oooo00@0gia8
00000000000007Bo2P3oool0103OW4P0000000000000M;l50?ooo`040=nLB00000000000001d_`D0
oooo00D0_g@0000000000000000004RLg`0V0?ooo`00KP3oool00`000000oooo0?ooo`0Z0?ooo`03
0000003oool0oooo00`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo0100oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool00000
0180oooo00<000000?ooo`000000203oool00`000000oooo000000080?ooo`030000003oool00000
01<0oooo00<000000?ooo`0000001`3oool010000000oooo0?ooo`0000080?ooo`030000003oool0
00000180oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo0100oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool04`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo05<0oooo001^
0?ooo`<00000:P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01T0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0080oooo00D00000
0?ooo`3oool0oooo0000000B0?ooo`030000003oool0000000P0oooo00<000000?ooo`000000203o
ool00`000000oooo0000000C0?ooo`030000003oool0000000L0oooo00@000000?ooo`3oool00000
203oool00`000000oooo0000000B0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000
oooo0?ooo`0@0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0603oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool04P3oool3000005D0
oooo001^0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool04@3oool01@000000oooo
0?ooo`3oool0000000<0oooo00<000000?ooo`3oool0403oool00`000000oooo000000080?ooo`03
0000003oool0000000P0oooo00<000000?ooo`0000004`3oool00`000000oooo000000070?ooo`04
0000003oool0oooo000000P0oooo00<000000?ooo`0000004P3oool01@000000oooo0?ooo`3oool0
000000@0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo01P0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0D0?ooo`030000003oool0
oooo01@0oooo00<000000?ooo`3oool0D`3oool006h0oooo00<000000?ooo`3oool0:`3oool00`00
0000oooo0?ooo`0;0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`0@0?ooo`050000003oool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0000000P0oooo00<000000?ooo`000000203oool00`000000oooo0000000C
0?ooo`030000003oool0000000L0oooo00@000000?ooo`3oool00000203oool00`000000oooo0000
000B0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`0?0?ooo`030000
003oool0oooo02P0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0C0?ooo`030000
003oool0oooo01D0oooo00<000000?ooo`3oool0D`3oool006h0oooo00<000000?ooo`3oool0:`3o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`0@0?ooo`050000003oool0oooo0?ooo`0000000`3oool00`000000oooo
0?ooo`0@0?ooo`030000003oool0000000P0oooo00<000000?ooo`000000203oool00`000000oooo
0000000C0?ooo`030000003oool0000000L0oooo00@000000?ooo`3oool00000203oool00`000000
oooo0000000B0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo02T0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0C0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool0D`3oool006h0oooo0`00000[0?ooo`03
0000003oool0oooo00/0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo0100oooo00@000000?ooo`3oool00000103oool00`000000oooo0?ooo`0@0?oo
o`030000003oool0000000P0oooo00<000000?ooo`000000203oool00`000000oooo0000000C0?oo
o`030000003oool0000000L0oooo00@000000?ooo`3oool00000203oool00`000000oooo0000000B
0?ooo`050000003oool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo02T0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0C0?ooo`030000003o
ool0oooo01<0oooo0`00001E0?ooo`00B@3oool0102oM00000000000001d_ol30?ooo`030?oOW018
0000M;oo00@0oooo00@0_g@0000000000000M;oo0`3oool00`3ookl0M000000000020000000304RL
g`3oool0oooo00P0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0C0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo0100oooo00@000000?ooo`3oool00000103oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0000000P0oooo00<000000?ooo`000000203oool00`000000oooo0000000C0?ooo`03
0000003oool0000000L0oooo00@000000?ooo`3oool00000203oool00`000000oooo0000000B0?oo
o`050000003oool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0
oooo02T0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0
oooo00H0oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`040?ooo`040;md0000
0000000007Boo`<0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`0`0?ooo`00B03o
ool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00`3oool01P3OW4P0
07Bo0?ooo`3ogi`0B00004RLg`@0oooo00<0okmd0000001d_ol02`3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0l
0?ooo`040000003oool0oooo000000@0oooo00<000000?ooo`3oool0403oool00`000000oooo0000
00080?ooo`030000003oool0000000P0oooo00<000000?ooo`0000004`3oool00`000000oooo0000
00070?ooo`040000003oool0oooo000000P0oooo00<000000?ooo`0000004P3oool00`000000oooo
0?ooo`020?ooo`040000003oool0oooo000003h0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`0[0?ooo`030000003oool0oooo00D0oooo00H0gia8001d_`3oool0omnL04P00018W=l2
0?ooo`030?oOW0180000M;oo00<0oooo00H0gia8001d_`3oool0omnL04P00018W=l40?ooo`030?no
M0000000M;oo03<0oooo00170?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol01P3o
ool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00<0oooo00<0oooO09a80000M;l02`3o
ool3000001H0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo03`0oooo00@000000?ooo`3oool00000103oool00`000000oooo0?ooo`0@0?ooo`030000
003oool0000000P0oooo00<000000?ooo`000000203oool00`000000oooo0000000C0?ooo`030000
003oool0000000L0oooo00@000000?ooo`3oool00000203oool00`000000oooo0000000B0?ooo`03
0000003oool0oooo0080oooo00@000000?ooo`3oool00000?P3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo02T0oooo0`0000060?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000
B02Lgol01P3oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00<0oooo00<0oooO09a8
0000M;l0<`3oool004L0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool50?oo
o`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo0`3oool00`3OW4P004RL0=ooo`0:
0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo03`0oooo00@000000?ooo`3oool00000103oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0000000L0oooo00@000000?ooo`3oool00000203oool00`000000oooo
0000000C0?ooo`030000003oool0000000L0oooo00@000000?ooo`3oool000001`3oool010000000
oooo0?ooo`00000B0?ooo`030000003oool0oooo0080oooo00@000000?ooo`3oool000008`3oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0:`3oool0
0`000000oooo0?ooo`040?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo1@3o
ool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`<0oooo00<0gia80018W03Oool0
<P3oool004L0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2L
B00007Bo00<0oooo00<0gia80000B02Lgol00`3oool00`3ogi`0B01809cOo`0:0?ooo`030000003o
ool0oooo01H0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo03`0oooo00@000000?ooo`3oool00000103oool00`000000oooo0?ooo`0@0?ooo`030000
003oool0000000L0oooo00D000000?ooo`3oool0oooo000000070?ooo`030000003oool0000001<0
oooo00<000000?ooo`0000001`3oool010000000oooo0?ooo`0000070?ooo`040000003oool0oooo
00000180oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`00000S0?ooo`030000003o
ool0oooo01P0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0[0?ooo`030000003o
ool0oooo00@0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2L
B00007Bo00<0oooo00<0gia80000B02Lgol00`3oool00`3ogi`0B01809cOo`0b0?ooo`00B03oool0
0`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00H0oooo00<0_g@00018W03Oool00P3oool0
0`3ogi`0B00004RLg`030?ooo`030?oo_`1d0000M;oo00X0oooo1000000F0?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0m0?ooo`040000003oool0
oooo000000@0oooo00<000000?ooo`3oool0403oool00`000000oooo000000070?ooo`050000003o
ool0oooo0?ooo`0000001`3oool00`000000oooo0000000C0?ooo`030000003oool0000000L0oooo
00@000000?ooo`3oool000001`3oool010000000oooo0?ooo`00000B0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`000000903oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool0:03oool4000000L0oooo00<0_g@00018W03Oool00P3oool00`3o
gi`0B00004RLg`060?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l00`3oool00`3o
okl0M00007Boo`0b0?ooo`00B03oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00H0
oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`040?ooo`030;md0000M;l0oooo00T0
oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool0?P3oool00`000000oooo000000040?ooo`030000003oool0oooo0100
oooo00<000000?ooo`0000001`3oool01@000000oooo0?ooo`3oool0000000L0oooo00<000000?oo
o`0000004`3oool00`000000oooo000000070?ooo`040000003oool0oooo000000L0oooo00@00000
0?ooo`3oool000004P3oool00`000000oooo0?ooo`020?ooo`030000003oool0000002@0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0100oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`050?ooo`030=nLB00004P0W=oo0080oooo00<0
oooO09a80000M;l01P3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00@0oooo00<0
_g@0001d_`3oool0<@3oool004P0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`05
0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool00`3oool00`3OW4P004RL0=ooo`09
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0:
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0000000@0oooo00<000000?ooo`3oool0403oool00`000000oooo00000007
0?ooo`050000003oool0oooo0?ooo`0000001`3oool00`000000oooo0000000C0?ooo`030000003o
ool0000000L0oooo00@000000?ooo`3oool000001`3oool010000000oooo0?ooo`00000B0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`0000004P3oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0oooo01L0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0oooo01P0oooo00<000000?ooo`3oool01@3oool00`3ogi`0B00004RLg`030?ooo`03
0;md0000B9`0gooo00D0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`030?ooo`03
0=nLB000B9`0gooo0340oooo00180?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03O
ool01@3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00<0oooo00<0omnL04P0
B02Lgol02@3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo0P0000050?oo
o`030000003oool0oooo0100oooo00@000000?ooo`3oool000001P3oool01@000000oooo0?ooo`3o
ool0000000L0oooo00<000000?ooo`0000004`3oool00`000000oooo000000070?ooo`040000003o
ool0oooo000000L0oooo00@000000?ooo`3oool000004P3oool00`000000oooo0?ooo`030?ooo`80
00004P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3o
ool01@3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00D0oooo00@0oooO09a8
0000B9`0gooo0P3oool00`3OW4P004RL0=ooo`030?ooo`030?oOW01804P0W=oo0340oooo00190?oo
o`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01P3oool00`3ogi`0B00007Boo`020?oo
o`030=nLB000M;l0oooo00<0oooo00<0oono07@0001d_ol02@3oool00`000000oooo0?ooo`0/0?oo
o`030000003oool0oooo00T0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0@0?oo
o`030000003oool0oooo0100oooo0P0000050?ooo`030000003oool0oooo0100oooo00@000000?oo
o`3oool000001P3oool01@000000oooo0?ooo`3oool0000000L0oooo00<000000?ooo`0000004`3o
ool00`000000oooo000000070?ooo`040000003oool0oooo000000L0oooo00@000000?ooo`3oool0
00004P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0140oooo00<000000?ooo`3o
ool0403oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00H0oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`060?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3o
ool00`3oool00`3ookl0M00007Boo`0a0?ooo`00BP3oool01P3OW4P0001d0;ooo`3ooml0W4P004RL
g`P0oooo00H0gia80000M02oool0oooO09a80018W=l50?ooo`030;md0000M;l0oooo00P0oooo00<0
00000?ooo`3oool0;03oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool04P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0100oooo00@0
00000?ooo`3oool000001P3oool01@000000oooo0?ooo`3oool0000000L0oooo00<000000?ooo`00
00004`3oool00`000000oooo000000070?ooo`040000003oool0oooo000000L0oooo00@000000?oo
o`3oool000004P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool06@3oool00`000000oooo0?ooo`070?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO
203oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`D0oooo00<0_g@0001d_`3oool0<03oool004/0
oooo00@0gia800000000000007Bo2P3oool0103OW4P0000000000000M;l40?ooo`050?oo_`1d0000
00000000M02oool0203oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01H0oooo00<0
00000?ooo`3oool0403oool010000000oooo0?ooo`0000060?ooo`050000003oool0oooo0?ooo`00
00001`3oool00`000000oooo0000000C0?ooo`030000003oool0000000L0oooo00@000000?ooo`3o
ool000001`3oool010000000oooo0?ooo`00000B0?ooo`030000003oool0oooo01H0oooo00<00000
0?ooo`3oool04P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`3oool0403oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00P0oooo00@0gia8
00000000000007Bo2P3oool0103OW4P0000000000000M;l40?ooo`050?oo_`1d000000000000M02o
ool0<03oool006h0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`090?ooo`030000
003oool0oooo01L0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0F0?ooo`030000
003oool0oooo0100oooo00@000000?ooo`3oool000001P3oool01@000000oooo0?ooo`3oool00000
00L0oooo00<000000?ooo`0000004`3oool00`000000oooo000000070?ooo`040000003oool0oooo
000000L0oooo00@000000?ooo`3oool000004P3oool00`000000oooo0?ooo`0F0?ooo`030000003o
ool0oooo02`0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool0D`3oool006h0oooo00<000000?ooo`3oool0;03oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0100oooo00@000000?ooo`3oool000001P3o
ool01@000000oooo0?ooo`3oool0000000L0oooo00<000000?ooo`0000004`3oool00`000000oooo
000000070?ooo`050000003oool0oooo0?ooo`0000001P3oool010000000oooo0?ooo`00000B0?oo
o`030000003oool0oooo01H0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`090?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`1C0?oo
o`00KP3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool02@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3o
ool05P3oool00`000000oooo0?ooo`0@0?ooo`040000003oool0oooo000000H0oooo00D000000?oo
o`3oool0oooo000000070?ooo`030000003oool0000001<0oooo00<000000?ooo`0000001`3oool0
1@000000oooo0?ooo`3oool0000000H0oooo00@000000?ooo`3oool000004@3oool00`000000oooo
0?ooo`0G0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`0/0?ooo`030000003oool0oooo05<0oooo001^0?ooo`030000003oool0oooo01X0oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo04D0oooo00<0
00000?ooo`3oool0403oool010000000oooo0?ooo`0000060?ooo`050000003oool0oooo0?ooo`00
00001`3oool00`000000oooo0000000C0?ooo`030000003oool0000000L0oooo00D000000?ooo`3o
ool0oooo000000060?ooo`040000003oool0oooo00000140oooo00<000000?ooo`3oool0AP3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0D`3oool0
06h0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0@0?ooo`040000003oool0oooo
000000H0oooo00D000000?ooo`3oool0oooo000000070?ooo`030000003oool0000001<0oooo00<0
00000?ooo`0000001`3oool01@000000oooo0?ooo`3oool0000000H0oooo00@000000?ooo`3oool0
00004@3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool0;03oool00`000000oooo0?ooo`1C0?ooo`00KP3oool00`000000oooo0?ooo`0K0?ooo`030000
003oool0oooo00h0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`160?ooo`030000
003oool0oooo0100oooo00@000000?ooo`3oool000001P3oool01@000000oooo0?ooo`3oool00000
00H0oooo00@000000?ooo`3oool000004`3oool00`000000oooo000000070?ooo`050000003oool0
oooo0?ooo`0000001P3oool010000000oooo0?ooo`00000A0?ooo`030000003oool0oooo04H0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo05<0oooo
001^0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0403oool010000000oooo0?oo
o`0000060?ooo`050000003oool0oooo0?ooo`0000001P3oool010000000oooo0?ooo`00000C0?oo
o`030000003oool0000000L0oooo00D000000?ooo`3oool0oooo000000060?ooo`040000003oool0
oooo00000140oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0E0?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`1C0?ooo`00KP3oool0
0`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`160?ooo`030000003oool0oooo00l0oooo00D000000?ooo`3oool0oooo0000
00060?ooo`050000003oool0oooo0?ooo`0000001P3oool010000000oooo0?ooo`00000C0?ooo`03
0000003oool0000000L0oooo00D000000?ooo`3oool0oooo000000060?ooo`040000003oool0oooo
00000140oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`1C0?ooo`00KP3oool00`00
0000oooo0?ooo`0L0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`160?ooo`030000003oool0oooo00l0oooo00D000000?ooo`3oool0oooo00000006
0?ooo`050000003oool0oooo0?ooo`0000001P3oool010000000oooo0?ooo`00000B0?ooo`040000
003oool0oooo000000L0oooo00D000000?ooo`3oool0oooo000000060?ooo`040000003oool0oooo
00000140oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0
oooo05<0oooo001^0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0;P3oool00`00
0000oooo0?ooo`0?0?ooo`050000003oool0oooo0?ooo`0000001P3oool01@000000oooo0?ooo`3o
ool0000000H0oooo00@000000?ooo`3oool000004P3oool01@000000oooo0?ooo`3oool0000000H0
oooo00D000000?ooo`3oool0oooo000000060?ooo`040000003oool0oooo00000140oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01D0oooo00<00000
0?ooo`3oool01`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01h0oooo00<00000
0?ooo`3oool0D`3oool00380oooo00<0gibL0;ooo`3oool0>@3oool00`000000oooo0?ooo`0]0?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0F0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool03`3oool01@000000oooo0?ooo`3oool0
000000H0oooo00D000000?ooo`3oool0oooo000000060?ooo`040000003oool0oooo00000180oooo
00D000000?ooo`3oool0oooo000000060?ooo`050000003oool0oooo0?ooo`0000001P3oool01000
0000oooo0?ooo`00000A0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0303oool0
0`000000oooo0?ooo`0N0?ooo`030000003oool0oooo05<0oooo000b0?ooo`030;mdM02oool0oooo
03T0oooo0`00000]0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`0G0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool03`3oool01@00
0000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`04
0?ooo`040000003oool0oooo00000180oooo00D000000?ooo`3oool0oooo000000060?ooo`050000
003oool0oooo0?ooo`0000001@3oool01@000000oooo0?ooo`3oool000000140oooo00<000000?oo
o`3oool05@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01@0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo01d0oooo0`00001E0?oo
o`00<03oool0103ookl0M0000000M02ooolj0?ooo`030000003oool0oooo02d0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`050000003oool0oooo0?ooo`0000001@3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000004P3o
ool01@000000oooo0?ooo`3oool0000000H0oooo00D000000?ooo`3oool0oooo000000050?ooo`05
0000003oool0oooo0?ooo`0000004@3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo
01P0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`1C0?ooo`009P3oool00`3OW9`0
_ooo0?ooo`030?ooo`030?oo_`1d0000000000<0000000<0001d0;ooo`3oool0>@3oool00`000000
oooo0?ooo`0]0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool04`3oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool03P3oool01@000000
oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?oo
o`050000003oool0oooo0?ooo`0000004@3oool01@000000oooo0?ooo`3oool0000000D0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`000000
4@3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0
503oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
7`3oool00`000000oooo0?ooo`1C0?ooo`009P3oool0103OW4P0M;oo0?oo_`1d00050000000507Bo
o`3oool0oooo0;mdM02oool0>P3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01D0
oooo00<000000?ooo`3oool03P3oool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000004@3oool0
1@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`050000003oool0oooo0?ooo`0000004@3oool00`000000oooo0?ooo`0D0?ooo`030000
003oool0oooo0300oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo0200oooo00<000000?ooo`3oool0D`3oool002H0oooo00<0gia80000000000000P00
00000`1d_ol0oooo0?ooo`100?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool02`3o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo00D000000?ooo`3oool0oooo
000000050?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0103oool01@000000oooo
0?ooo`3oool000000140oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool000000140oooo00<00000
0?ooo`3oool0503oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo05<0oooo000V0?oo
o`030=nLB000B9`0gooo04D0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0K0?oo
o`030000003oool0oooo01@0oooo00<000000?ooo`3oool03P3oool01@000000oooo0?ooo`3oool0
000000D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`050000003oool0
oooo0?ooo`0000004@3oool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000004@3oool00`000000
oooo0?ooo`0D0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool01P3oool00`000000
oooo0?ooo`0^0?ooo`030000003oool0oooo05<0oooo000V0?ooo`030=nLW02oool0oooo04D0oooo
00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00h0oooo
00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool01@000000oooo0?ooo`3oool000000140oooo00D000000?ooo`3oool0oooo00000005
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`3o
ool000000140oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0a0?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`1C0?ooo`00KP3oool0
0`000000oooo0?ooo`0R0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`180?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
1@000000oooo0?ooo`3oool000000140oooo00D000000?ooo`3oool0oooo000000050?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool000000140
oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo02h0
oooo00<000000?ooo`3oool0D`3oool003@0oooo00@0oooO09a8000007@0_ooo=P3oool00`000000
oooo0?ooo`0S0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`180?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool01@000000
oooo0?ooo`3oool000000140oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`1C0?ooo`00<@3oool01@3o
_g@000000000000004P0W=oo03P0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo0000000A0?ooo`050000003o
ool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo03@0oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo02h0oooo
00<000000?ooo`3oool0D`3oool002h0oooo00@0_g@000000000000007Bo?03oool00`000000oooo
0?ooo`0T0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01@3oool00`000000oooo
0?ooo`180?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool01@000000oooo
0?ooo`3oool000000140oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo
05<0oooo000Z0?ooo`050?oOW018000000000000001d_ol0?`3oool3000002@0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo04P0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000004@3oool0
1@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`0e0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo02<0oooo0`00001E0?ooo`009`3oool01@3OW4P0000000000000
B9`0gooo0480oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo0100oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00D000000?ooo`3oool0oooo0000000A0?ooo`050000003oool0oooo0?ooo`00
00001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3o
ool03`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool09P3oool00`000000oooo0?ooo`1C0?ooo`009@3oool00`3ogi`0B00004RLg`0:0?ooo`030=nL
W02oool0oooo03T0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo00l0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0C0?ooo`030000
003oool0oooo00h0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool000000140
oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo00l0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo02L0oooo00<000000?oo
o`3oool0D`3oool00380oooo00<0_gAd0;ooo`3oool0>@3oool00`000000oooo0?ooo`0^0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0Q0?ooo`03
0000003oool0oooo01<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`05
0000003oool0oooo0?ooo`0000004@3oool01@000000oooo0?ooo`3oool0000000D0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0240oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool09`3oool00`000000oooo0?ooo`1C0?ooo`00<03oool0103ookl0M0000000M02ooolj
0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo0000000A0?ooo`050000003o
ool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01<0oooo
00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`1C0?ooo`009P3oool00`3OW9`0_ooo0?oo
o`030?ooo`030?oo_`1d0000000000<0000000<0001d0;ooo`3oool0>@3oool00`000000oooo0?oo
o`0Y0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo0000000A0?ooo`050000
003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01<0
oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`1C0?ooo`009P3oool0103OW4P0M;oo
0?oo_`1d00050000000507Boo`3oool0oooo0;mdM02oool0>P3oool00`000000oooo0?ooo`0Z0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0<0?oo
o`030000003oool0oooo02D0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo0000000A0?ooo`050000003oool0
oooo0?ooo`000000103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool0=03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo02l0oooo00<0
00000?ooo`3oool0D`3oool002H0oooo00<0gia80000000000000P0000000`1d_ol0oooo0?ooo`10
0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3o
ool0000003L0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00<0oooo00@000000?ooo`3oool000001`3oool00`000000oooo0?ooo`020?ooo`050000
003oool0oooo0?ooo`0000004@3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?oo
o`3oool01@3oool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool04P3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0<03oool00`000000oooo0?ooo`1C0?ooo`009P3oool00`3OW4P004RL0=ooo`150?oo
o`<00000;03oool200000080oooo00D000000?ooo`3oool0oooo0000000h0?ooo`030000003oool0
oooo0180oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`030?ooo`040000003oool0
oooo000000L0oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo00l0oooo
00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo00D0oooo00@000000?ooo`3o
ool000001@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0180oooo00<000000?oo
o`3oool0=P3oool01@000000oooo0?ooo`3oool000000300oooo0`00001E0?ooo`009P3oool00`3O
W9`0_ooo0?ooo`0P0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol0103oool0
102oM00000000000001d_ol30?ooo`060?oOW0180000000000000000B9`0gooo2P3oool00`000000
oooo0?ooo`0^0?ooo`030000003oool0000000<0oooo00<000000?ooo`3oool0=P3oool00`000000
oooo0?ooo`0B0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool00`3oool010000000
oooo0?ooo`0000070?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`0?
0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`050?ooo`040000003o
ool0oooo000000D0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0B0?ooo`030000
003oool0oooo02P0oooo00<000000?ooo`3oool02`3oool01P000000oooo0?ooo`000000oooo0000
0340oooo00<000000?ooo`3oool01P3oool0102oM00000000000001d_ol30?ooo`030?oOW0180000
M;oo00@0oooo00@0_g@0000000000000M;oo0`3oool01P3ogi`0B00000000000000004RL0=oooc80
oooo000`0?ooo`030?noM000000000000080000000<0M;oo0?ooo`3oool0403oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00`3oool01P3OW4P007Bo0?ooo`3ogi`0
B00004RLg`80oooo00H0okmd0000001d_ol0oooo0?oOW018M;l:0?ooo`030000003oool0oooo02l0
oooo00D000000?ooo`000000oooo0000001>0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool00`3oool010000000oooo0?ooo`0000070?ooo`050000003oool0oooo0?ooo`000000103o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo
000000070?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0n0?ooo`800000303oool010000000oooo00000000000b0?ooo`030000003oool0oooo
00D0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo00<0oooo00H0
gia8001d_`3oool0omnL04P00018W=l20?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo<P3oool0
02h0oooo00@0omnL04P000000000B9cO0`3oool00`2oM7@0_ooo0?ooo`0?0?ooo`030?oog`2LB000
B9cO0080oooo00<0okmd0000B02Lgol01P3oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0
W=oo00H0oooo00<0okmd04RLg`3oool0203oool00`000000oooo0?ooo`0a0?ooo`<00000203oool3
000004<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`030?ooo`040000003oool0
oooo000000L0oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000000L0oooo00@000000?ooo`3o
ool000001@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0400oooo0`0000090?oo
o`<00000<`3oool00`000000oooo0?ooo`040?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000
B02Lgol01P3oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00H0oooo00<0okmd04RL
g`3oool0<03oool002d0oooo00<0omnL04P0M02oool01@3oool00`2oM7@0_ooo0?ooo`0?0?ooo`03
0?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo1@3oool00`3ooml0W4P004RLg`020?oo
o`040?oog`2LB00004RL0=ooo`D0oooo00<0oono07@0M02oool0203oool00`000000oooo0?ooo`0f
0?ooo`L00000A@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00<0oooo00@00000
0?ooo`3oool000001`3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool0
3`3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000001`3oool010000000
oooo0?ooo`0000050?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0@`3oool20000
00D0oooo0`00000e0?ooo`030000003oool0oooo00@0oooo00<0oooO09a80018W=l00P3oool0103o
oml0W4P00018W03Oool50?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo1@3o
ool00`3ookl0M01d0;ooo`0`0?ooo`00;@3oool00`3OW9`0_ooo0?ooo`040?ooo`030;md0018W=l0
oooo0100oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2LB000
07Bo00<0oooo00<0gia80000B02Lgol01@3oool00`3ooml0W4P007Boo`080?ooo`030000003oool0
oooo0880oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`030?ooo`030000003oool0
000000P0oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000000P0oooo00<000000?ooo`000000
1P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0880oooo00<000000?ooo`3oool0
103oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00D0oooo00<0oooO09a80000M;l0
0`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2LB000M;oo0300oooo000]0?ooo`080=nLM02L
gol0oooo0?ooo`3ookl0M00000000018W=lC0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P0
0018W=l01P3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00D0oooo00<0oono07@0
0018W=l0203oool400000840oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?oo
o`800000203oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`020?ooo`040000003oool0oooo000000T0oooo0P0000070?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool0O`3oool4000000L0oooo00<0_g@00018W03Oool00P3o
ool00`3ogi`0B00004RLg`060?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l01@3o
ool00`3ookl0M00004RLg`0`0?ooo`00;@3oool00`3ooml0W4P000000002000000030018W03Oool0
oooo01<0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`060?ooo`030=nLB00004P0
W=oo0080oooo00<0oooO09a80000M;l01@3oool00`3o_g@0000007Boo`080?ooo`030000003oool0
oooo0880oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`800000203oool0
1@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`020?ooo`040000003oool0oooo000000T0oooo0P0000070?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool0PP3oool00`000000oooo0?ooo`050?ooo`030=nLB00004P0W=oo0080oooo
00<0oooO09a80000M;l01P3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00D0oooo
00<0okmd0000001d_ol0<03oool004P0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=oo
o`050?ooo`030?oOW0180000B9cO00<0oooo00X0_g@00018W03Oool0oooo0?oOg`3Oool0okmd0000
000007@0_ooo203oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0?0?ooo`040000003oool0oooo000000@0oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo00000180oooo
00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo06l0oooo
00<000000?ooo`3oool01@3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00D0oooo
00<0omnL04P00018W=l00`3oool02P2oM00004RL0=ooo`3oool0omoO0=ooo`3o_g@000000000M02o
ool`0?ooo`00B03oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00D0oooo00@0
oooO09a80000B9`0gooo0P3oool02@3OW4P004RL0=ooo`3oool0oono07@000000000001809cOo`09
0?ooo`030000003oool0oooo06l0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`0000001@3oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo0080oooo00@000000?ooo`3oool000004P3oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0L03oool00`000000oooo0?oo
o`050?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01@3oool0103ooml0W4P0
0018W03Oool20?ooo`090=nLB000B9`0gooo0?ooo`3ookl0M0000000000004P0W=oo0340oooo0019
0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01P3oool00`3ogi`0B00007Boo`02
0?ooo`030=nLB000M;l0oooo0080oooo00<0gibL0;ooo`3oool02P3oool00`000000oooo0?ooo`1_
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0000000D0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool000000180oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool01P3oool00`3ogi`0B00007Boo`02
0?ooo`030=nLB000M;l0oooo00H0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`02
0?ooo`030=nLW02oool0oooo0380oooo001:0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO203o
ool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`<0oooo00<0omnL07BLg`3oool02P3oool3000006l0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0100
oooo00<000000?ooo`0000001@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`0000004P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool0KP3oool3000000T0oooo00H0gia80000M02oool0oooO09a80018W=l8
0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO0`3oool00`3ogi`0M9cO0?ooo`0b0?ooo`00B`3o
ool0103OW4P0000000000000M;l:0?ooo`040=nLB00000000000001d_`D0oooo00D0_g@000000000
0000000004RLg`070?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool03`3oool00`00
0000oooo0?ooo`0<0?ooo`030000003oool0oooo0100oooo00<000000?ooo`0000001@3oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`0000004P3oool00`00
0000oooo0?ooo`0<0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0L03oool00`00
0000oooo0?ooo`080?ooo`040=nLB00000000000001d_`X0oooo00@0gia800000000000007Bo1@3o
ool01@2oM0000000000000000000B9cO02l0oooo001^0?ooo`030000003oool0oooo0700oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0140oooo0P00
00050?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00`3oool2000001<0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0740oooo00<0
00000?ooo`3oool0D`3oool006h0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo0740oooo00<000000?ooo`3oool0D`3oool006h0oooo0`00001a0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0<0?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool0K`3oool3000005D0oooo001^0?ooo`03
0000003oool0oooo08<0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0H0?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0<0?ooo`03
0000003oool0oooo0880oooo00<000000?ooo`3oool0D`3oool006h0oooo00<000000?ooo`3oool0
P`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
PP3oool00`000000oooo0?ooo`1C0?ooo`00KP3oool3000008<0oooo00<000000?ooo`3oool02`3o
ool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool06@3o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0800oooo0`00001E0?ooo`00DP3oool0
102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo0080oooo00<0oono07@0000000000P00
00000`18W=l0oooo0?ooo`080?ooo`030000003oool0oooo08<0oooo00<000000?ooo`3oool02`3o
ool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool06@3o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo08<0oooo00<000000?ooo`3oool01P3o
ool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo0080oooo00<0oono07@000000000
0P0000000`18W=l0oooo0?ooo`0i0?ooo`00D@3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80
oooo00<0omnL04P0001d_ol0103oool00`3o_g@0000007Boo`0;0?ooo`030000003oool0oooo08<0
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo08<0
oooo00<000000?ooo`3oool01@3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL
04P0001d_ol0103oool00`3o_g@0000007Boo`0l0?ooo`00D03oool00`3ooml0W4P004RLg`020?oo
o`030?noM00004P0W=oo00P0oooo00<0oooO09a80000M;l02`3oool3000008<0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0840oooo0`0000060?oo
o`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol0203oool00`3ooml0W4P0001d_`0l0?oo
o`00D03oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`P0oooo00<0gia80018
W03Oool02P3oool00`000000oooo0?ooo`1d0?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool0M03oool00`000000oooo0?ooo`040?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a8
0000B9`0gooo203oool00`3OW4P004RL0=ooo`0k0?ooo`00D03oool00`3ooml0W4P0001d_`030?oo
o`030=nLB00004P0W=oo00P0oooo00<0omnL04P0B02Lgol02P3oool00`000000oooo0?ooo`1d0?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0I0?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`040?oo
o`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol0203oool00`3ogi`0B01809cOo`0k0?oo
o`00D@3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00P0oooo00<0oono07@0001d
_ol02P3oool4000007@0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1b0?ooo`@0
00001`3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00P0oooo00<0oono07@0001d
_ol0>`3oool00540oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`090?ooo`030;md
0000M;l0oooo00T0oooo00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0=0?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo07D0oooo00<000000?ooo`3oool01@3oool00`3OW4P0001809cOo`020?ooo`030?oo
g`2LB00007Bo00T0oooo00<0_g@0001d_`3oool0>P3oool00540oooo00<0omnL04P00018W=l00`3o
ool00`2oM00004RL0=ooo`080?ooo`030=nLB000B9`0gooo00T0oooo00<000000?ooo`3oool0M@3o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool06@3o
ool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo07D0oooo00<000000?ooo`3oool01@3o
ool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00P0oooo00<0gia80018W03Oool0>P3o
ool00540oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`080?ooo`030?oOW018
04P0W=oo00T0oooo00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo07H0oooo00<000000?ooo`3oool01@3oool0103ooml0W4P00018W03Oool20?ooo`030=nL
B000B9`0gooo00P0oooo00<0omnL04P0B02Lgol0>P3oool00580oooo00<0omnL04P0001d_ol00P3o
ool00`3OW4P007Bo0?ooo`080?ooo`030?oo_`1d0000M;oo00T0oooo00<000000?ooo`3oool0MP3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3oool01P3o
ool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00P0oooo00<0oono07@0001d_ol0>P3o
ool005<0oooo00H0gia80000M02oool0oooO09a80018W=l:0?ooo`030;md0000M;l0oooo00P0oooo
00<000000?ooo`3oool0MP3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo07H0oooo
00<000000?ooo`3oool01`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`X0oooo00<0_g@0001d
_`3oool0>@3oool005@0oooo00@0gia800000000000007Bo2@3oool01@3ookl0M0000000000007@0
_ooo00P0oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`080?ooo`040=nLB0000000
0000001d_`T0oooo00D0oono07@000000000001d0;ooo`0i0?ooo`00KP3oool00`000000oooo0?oo
o`240?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`240?ooo`030000003oool0oooo05<0oooo001^0?ooo`030000003oool0oooo08@0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool06P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo08@0oooo00<00000
0?ooo`3oool0D`3oool006h0oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`090?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0J0?oo
o`030000003oool0oooo00T0oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`1C0?oo
o`00KP3oool00`000000oooo0?ooo`240?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool06`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`240?ooo`030000003oool0oooo05<0oooo001^0?ooo`030000
003oool0oooo08@0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0K0?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`080?ooo`030000
003oool0oooo00P0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`1C0?ooo`00KP3o
ool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool06`3o
ool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0NP3o
ool00`000000oooo0?ooo`1C0?ooo`00KP3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`1C0?ooo`00KP3oool00`00
0000oooo0?ooo`1j0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`0L0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool06`3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0N`3oool00`00
0000oooo0?ooo`1C0?ooo`00KP3oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool0N`3oool00`000000oooo0?ooo`1C0?ooo`00KP3oool00`000000oooo
0?ooo`1k0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`0L0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0703oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0O03oool00`000000oooo
0?ooo`1C0?ooo`00KP3oool00`000000oooo0?ooo`1l0?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool0703oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool0O03oool00`000000oooo0?ooo`1C0?ooo`00KP3oool00`000000oooo0?ooo`1m
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0L
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0O@3oool00`000000oooo0?ooo`1C
0?ooo`00KP3oool3000008D0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0L0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo08<0oooo0`00001E0?ooo`00KP3oool00`000000oooo0?ooo`250?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo01`0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`250?ooo`03
0000003oool0oooo05<0oooo001^0?ooo`030000003oool0oooo08D0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
703oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3oool0
D`3oool006h0oooo00<000000?ooo`3oool0Q@3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo01d0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0L0?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`1C0?ooo`00KP3oool0
0`000000oooo0?ooo`250?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool07P3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`260?ooo`030000003oool0oooo05<0oooo001^0?ooo`030000003oool0oooo
08D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool0P03oool00`000000oooo0?ooo`1C0?ooo`00KP3oool00`000000
oooo0?ooo`200?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01P3oool00`000000
oooo0?ooo`0N0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0703oool00`000000
oooo0?ooo`060?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0P@3oool00`000000
oooo0?ooo`1C0?ooo`00KP3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`1C0?ooo`00KP3oool00`000000oooo0?oo
o`220?ooo`050000003oool0oooo0?ooo`0000001`3oool00`000000oooo0?ooo`0N0?ooo`030000
003oool0oooo00X0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0PP3oool00`000000oooo0?ooo`1C0?ooo`00KP3o
ool00`000000oooo0?ooo`220?ooo`050000003oool0oooo0?ooo`0000001`3oool00`000000oooo
0?ooo`0N0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool07P3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0PP3oool00`000000oooo
0?ooo`1C0?ooo`00KP3oool3000008<0oooo00@000000?ooo`3oool000001P3oool00`000000oooo
0?ooo`0O0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool07P3oool00`000000oooo
0?ooo`040?ooo`050000003oool0oooo0?ooo`000000P`3oool3000005D0oooo001^0?ooo`030000
003oool0oooo08<0oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`0O0?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`040?ooo`04
0000003oool0oooo000008H0oooo00<000000?ooo`3oool0D`3oool006h0oooo00<000000?ooo`3o
ool0Q03oool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo01l0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo0080oooo00D000000?oo
o`3oool0oooo000000260?ooo`030000003oool0oooo05<0oooo001^0?ooo`030000003oool0oooo
08L0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
08P0oooo00<000000?ooo`3oool0D`3oool006h0oooo00<000000?ooo`3oool0Q`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool02@3oool00`000000
oooo0?ooo`0O0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0R03oool00`000000
oooo0?ooo`1C0?ooo`00KP3oool00`000000oooo0?ooo`280?ooo`050000003oool0oooo0?ooo`00
00008P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0200oooo00D000000?ooo`3o
ool0oooo0000002:0?ooo`030000003oool0oooo05<0oooo001^0?ooo`030000003oool0oooo08P0
oooo00D000000?ooo`3oool0oooo0000000R0?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool0803oool01@000000oooo0?ooo`3oool0000008X0oooo00<000000?ooo`3oool0D`3oool0
06h0oooo00<000000?ooo`3oool0R03oool010000000oooo0?ooo`00000S0?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`02000008X0oooo00<00000
0?ooo`3oool0D`3oool006h0oooo0`0000290?ooo`030000003oool0000002<0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`0P0?ooo`@00000R@3oool3000005D0oooo001B0?ooo`04
0;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B000000000000000
04RL0=ooo`X0oooo00<000000?ooo`3oool0R@3oool3000002<0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`0R0?ooo`800000S03oool00`000000oooo0?ooo`060?ooo`040;md0000
0000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=oo
oc/0oooo001A0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`02
0?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo2P3oool00`000000oooo0?ooo`2;0?ooo`030000
003oool0oooo0240oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0R0?ooo`030000
003oool0oooo08/0oooo00<000000?ooo`3oool01@3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RL
g`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0omnL04Qd_c/0oooo001@
0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`08
0?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0S03oool00`000000oooo0?ooo`04
0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`0i
0?ooo`00D03oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono
07@0M02oool0203oool00`000000oooo0?ooo`2<0?ooo`030000003oool0oooo0200oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo08d0oooo00<00000
0?ooo`3oool0103oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0
oono07@0M02oool0>@3oool00500oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0:
0?ooo`030?oog`2LB000M;oo00P0oooo00<000000?ooo`3oool0S@3oool00`000000oooo0?ooo`0P
0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`2=
0?ooo`030000003oool0oooo00@0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0:
0?ooo`030?oog`2LB000M;oo03T0oooo001A0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P0
0018W=l02P3oool00`3ookl0M00004RLg`080?ooo`@00000[`3oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo0:d0oooo100000070?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P0
0018W=l02P3oool00`3ookl0M00004RLg`0i0?ooo`00D@3oool00`3OW4P0001809cOo`020?ooo`03
0?oog`2LB00007Bo00X0oooo00<0okmd0000001d_ol0203oool00`000000oooo0?ooo`2`0?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`3oool0/03oool00`000000oooo0?ooo`050?ooo`03
0=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`0i0?ooo`00
D@3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00H0oooo00H0omoO0=ooo`3o_g@0
00000000M02oool80?ooo`030000003oool0oooo0;00oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`2a0?ooo`030000003oool0oooo00D0oooo00<0omnL04P00018W=l00`3oool00`2o
M00004RL0=ooo`060?ooo`060?oOg`3Oool0okmd0000000007@0_ooo>@3oool00540oooo00@0oooO
09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`050?oo_`1d000000000000B02Lgol0
2@3oool00`000000oooo0?ooo`2`0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
/@3oool00`000000oooo0?ooo`050?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03O
ool01P3oool01@3ookl0M0000000000004P0W=oo03X0oooo001B0?ooo`030?oOW0180000M;oo0080
oooo00<0gia8001d_`3oool01`3oool00`3OW9`0_ooo0?ooo`0:0?ooo`030000003oool0oooo09<0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0;40
oooo00<000000?ooo`3oool01P3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0
oooo00<0gibL0;ooo`3oool0>`3oool005<0oooo00H0gia80000M02oool0oooO09a80018W=l80?oo
o`030?oOW01dW=l0oooo00X0oooo0`00002C0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0980oooo0`0000090?oo
o`060=nLB00007@0_ooo0?oog`2LB000B9cO203oool00`3ogi`0M9cO0?ooo`0k0?ooo`00E03oool0
103OW4P0000000000000M;l:0?ooo`050;md00000000000000000018W=l01`3oool00`000000oooo
0?ooo`2D0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool01@3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo09D0oooo00<000000?ooo`3oool0203oool0103OW4P00000
00000000M;l:0?ooo`050;md00000000000000000018W=l0>03oool006h0oooo00<000000?ooo`3o
ool0U@3oool2000001X0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0H0?ooo`03
0000003oool0oooo09H0oooo00<000000?ooo`3oool0D`3oool006h0oooo00<000000?ooo`3oool0
U`3oool2000001P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0G0?ooo`800000
V@3oool00`000000oooo0?ooo`1C0?ooo`00KP3oool3000009P0oooo00<000000?ooo`3oool05P3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0U`3o
ool3000005D0oooo001^0?ooo`030000003oool0oooo0;40oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`0E0?ooo`030000003oool0oooo09X0oooo00<000000?ooo`3oool0D`3oool0
06h0oooo00<000000?ooo`3oool0/P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0;80oooo00<000000?ooo`3oool0D`3oool006h0oooo0`00002b0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0/03oool3000005D0oooo001N0?ooo`030?oo_`1d0000000000800000
00<0B9cO0?ooo`3oool0203oool00`000000oooo0?ooo`2O0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool0D`3oool00600oooo00<0okmd0000001d_ol02`3oool00`000000oooo0?oo
o`2P0?ooo`<000003`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00d0oooo0`00
002S0?ooo`030000003oool0oooo05<0oooo001P0?ooo`030?oog`2LB00007Bo00/0oooo00<00000
0?ooo`3oool0X`3oool2000000h0oooo00D000000?ooo`3oool0oooo0000000>0?ooo`030000003o
ool0oooo0:@0oooo00<000000?ooo`3oool0D`3oool00640oooo00<0gia80018W03Oool02P3oool3
000000H0oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`150?ooo`030000003oool0
oooo00/0oooo00D000000?ooo`000000oooo0000000;0?ooo`<00000AP3oool00`000000oooo0?oo
o`1E0?ooo`030000003oool0oooo00@0oooo0`00001E0?ooo`00H@3oool00`3ogi`0B01809cOo`0:
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0C
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0C
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0A
0?ooo`D000004`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01<0oooo00<00000
0?ooo`3oool04P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01<0oooo00<00000
0?ooo`3oool04`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool0D`3oool00640oooo00<0oono07@0001d_ol02P3ooooo000007800000E@3oool00680
oooo00<0_g@0001d_`3oool0o`3oooo@0?ooo`00HP3oool00`3OW4P004RL0=ooo`3o0?ooom00oooo
001R0?ooo`030?oOW01804P0W=oo0?l0ooood03oool00680oooo00<0oono07@0001d_ol0o`3oooo@
0?ooo`00H`3oool00`2oM00007Bo0?ooo`3o0?oooll0oooo001Q0?ooo`050?oo_`1d000000000000
M02oool0o`3oooo?0?ooo`00o`3ooooo0?ooocH0oooo003D0?ooo`090?oo_`1d0000000004RLg`3o
ool0oono07@000000000B9cO0:H0oooo00<0oooO09a8B02Lgol0[P3oool00=@0oooo00X0oono07@0
B02Lgol0_g@0001d_`3ookl0M01809cOo`2oM00007BoYP3oool00`3OW4P0B9cO0?ooo`2]0?ooo`00
e`3oool00`3ookl0M00007Boo`020?ooo`030?oo_`1d0000M;oo0:D0oooo00<0omnL04Qd_`3oool0
[@3oool00=P0oooo00<0gia80018W03Oool00P3oool00`3OW4P004RL0=ooo`2T0?ooo`030?oo_`1d
07@0_ooo0:d0oooo002H0?ooo`030?oog`2LB000B9cO00h0oooo00<0oono07@0000000000P000000
1`18W=l0oooo0?ooo`3o_g@000000000M02oool00P3oool02@2oM00000000000B02Lgol0oooo0?no
M0000000001809cOo`040?ooo`030;md00000000000000D0000000<004RL0=ooo`3oool0103oool0
202oM00000000000B02Lgol0oooo0?noM00004P0W=oo0P3oool00`3o_g@0001809cOo`020?ooo`03
0;md0000M;l0oooo00@0oooo00/0omnL04P000000000000007Boo`3OW4P000000018W03Oool0oooo
0=nLB002000000080000B02Lgol0oooo0?ooo`3o_g@000000000B02Lgol20?ooo`040;md00000000
001809cOo`80oooo00D0okmd000000000000001809cOo`030?ooo`030=nLB000B9`0gooo0080oooo
00<0_g@000000000M;l01P3oool01P3OW4P00000000000000000001809cOo`H0oooo0180omnL04P0
00000000000007Boo`3OW4P000000018W03Oool0okmd0000000004P0W=oo0?ooo`3o_g@000000000
B02Lgol20?ooo`050=nLB000000000000000001d_ol00P3oool00`2oM00007Bo0?ooo`040?ooo`03
0=nLB000B9`0gooo0080oooo00<0_g@000000000M;l0203oool01`2oM7@0_ooo0?ooo`3oool0omnL
04Q8W03Oool0103oool03@2oM00000000000B02Lgol0oooo0?noM0000000001809cOo`3oool0gia8
0018W03Oool00P3oool00`2oM0000000001d_`040?ooo`0<0?oOW018000000000000001d_ol0gia8
00000000B9`0gooo0?ooo`2oM000M;oo0`3oool0203ookl0M0000000000007@0_ooo0=nLB0000000
B9cO4@3oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo0080oooo00H0omnL04P0
0000000000000018W03Ooomo0?ooo`00V@3oool00`3OW4P004RL0=ooo`0=0?ooo`0<0?oog`2LB000
M;oo0?ooo`3ogi`0B00007Boo`3oool0omnL04P0M02ookl0W9cO0P3oool02P3OW4P004RL0=oog`2L
M9`0gooo0?noM00004P0W=oO09adW03Oool40?ooo`040?oo_`1d000004RL0=ooo`<0oooo00<0_g@0
00000000M;l0103oool0203OW4P004RL0=oog`2LM9`0gooo0?oo_`1d0000B9cO0P3oool00`3ookl0
M00004RLg`020?ooo`030?noM00007@0_ooo00@0oooo02<0_g@00018W03Oool0oooo0;md001dM7@0
001d0;oOW02L_ol0okmd0000001d_ol0oooo0?oog`2LB000B9cO0?ooo`3o_g@0001809cOg`2LM9`0
gooo0?ooo`3OW4P004RL0=oog`2LM9`0gooo0?oOW01804P0W=oo0?ooo`3ogi`0B000001d_`020?oo
o`030?noM00007@0_ooo0080oooo00@0gia80018W03O_g@0M;oo1P3oool0103ooml0W4P00018W03O
ool70?ooo`0F0;md0000B9`0gooo0?ooo`2oM000M7Ad0000M02ogi`0W;oo0?noM00004P0W=oO09ad
W03Oool0okmd0000B02Lgml0W7BL0=ooo`3ogi`0B00007Boo`80oooo00H0_g@0001d_`3oool0okmd
0000M02oool40?ooo`030?noM00007@0_ooo0080oooo00@0gia80018W03O_g@0M;oo1`3oool0202o
M000001d0;ooo`3oool0okmd00000000B9`0gooo0`3oool03@3OW4P004RL0=oog`2LM9`0gooo0?no
M00004P0W=oO09adW03Oool0okmd0000M02oool00P3oool0103OW4P004RL0=noM01d_ol30?ooo`0<
0;md0000B9`0gooo0?ooo`2oM000M7Ad0000M02ogi`0W;oo0?ooo`3o_g@007Bo0`3oool02@3OW4P0
04RL0=ooo`3ookl0M01d0;noM00007@0W9ad09cOo`0?0?ooo`060=nLB000M;l0oooo0?oOW0180000
B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7BoO`3oool0
09T0oooo00<0omnL04P0B02Lgol03P3oool0402oM7@0_ooo0?ooo`3ookl0M00007Boo`3oool0oono
07@0001d_ol0gibL0;ooo`3oool0okmd0000B02Lgol20?ooo`030?oo_`1d0000M;oo00L0oooo00<0
gia80000B02Lgol0103oool0103ookl0M0000018W03Oool20?ooo`030?noM00004P0W=oo00<0oooo
00<0_g@00018W03Oool00P3oool01`2oM00004RL0=ooo`3oool0omnL04P0001d_ol0103oool00`2o
M00007Bo0?ooo`020?ooo`070;md00000000M;oo0?ooo`3OW4P004RL0=ooo`030?ooo`050=oOg`3o
ool0oono07@0001d_ol00`3oool00`3o_g@0001809cOo`020?ooo`030=nLB000B9`0gooo0080oooo
00L0omnL04P00018W=l0oooo0?oo_`1d0000B9cO0080oooo00<0omnL04P0B02Lgol0203oool00`3o
_g@0001809cOo`070?ooo`030;md0000M;l0oooo0080oooo00L0_g@00000001d_ol0oooo0?oo_`1d
0000M;oo0080oooo00<0oono07@0001d_ol00P3oool00`3OW4P004RL0=ooo`030?ooo`050=nLW02o
ool0omnL04P0001d_ol0103oool00`3ookl0M00004RLg`020?ooo`030?oOW01804P0W=oo00L0oooo
00T0oooO09a80000M7@0W9cO0?ooo`3o_g@0001d0;oOW01dW=l00`3oool00`3o_g@0001809cOo`02
0?ooo`030?oo_`1d0000M;oo0080oooo00<0oono07@0001d_ol00P3oool00`3ogi`0B01809cOo`04
0?ooo`030;md0000M;l0oooo0080oooo00<0_g@00000001d_ol00P3oool00`3ookl0M4RL0=ooo`02
0?ooo`080;md0000M;l0oooo0?ooo`3ookl0M4Q804P0B02Lgol30?ooo`030?oo_`1d0000000000L0
000000D0M;oo0?ooo`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00/0oooo00<0okmd04RL
g`3oool0O@3oool009D0oooo00<0okmd000000000000100000000`00B9`0gooo0?ooo`0>0?ooo`03
0=nLB00007@0_ooo0080oooo00<0_g@004RLg`3oool00P3oool00`3ookl0M00004RLg`030?ooo`03
0;md0000M;l0oooo00H0oooo00<0omnL04P0001d_ol01@3oool00`3ogi`0B000001d_`020?ooo`03
0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool00P3oool0203OW4P0001d0;ooo`3oool0oooO
09a80000B9`0gooo0`3oool00`3o_g@0001d0;ooo`020?ooo`070?oo_`1d0000B9cO0?ooo`3o_g@0
001d0;ooo`060?ooo`030;md0000M;l0oooo0080oooo00<0oono07@00018W=l00P3oool00`3o_g@0
001d0;ooo`030?ooo`0;0=nLB00007@0_ooo0?oog`2LB000001d0;ooo`3oool0oooO09a80018W=l0
203oool00`3ookl0M00004RLg`070?ooo`030?noM00007@0_ooo0080oooo00<0oono07@00018W=l0
0P3oool00`2oM00007Bo0?ooo`020?ooo`050;md0000M;l0oooo0?ooo`3o_g@00`0000002018W=l0
oooo0?ooo`3oool0oooO09a80000B9`0gooo0`3oool0203ooml0W4P00000M02oool0oooo0?oog`2L
B000B9cO1`3oool02P3ooml0W4P004RLg`3OW7@0W=oo0=nLB00007@0_ooo0=nLM02Lgol20?ooo`03
0?oo_`1d0000B9cO00<0oooo00`0_g@0001d_`3oool0oooo0?oog`2LB000001d0;ooo`3oool0oooO
09a80018W=l40?ooo`030?noM00007@0_ooo0080oooo00<0oono07@00018W=l00P3oool00`3ooml0
W4Q809cOo`020?ooo`030?noM000B9`0gooo0080oooo00<0okmd00000018W=l03`3oool00`3ooml0
W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02oool0O@3oool009D0
oooo00<0oooO09a8B02Lgol00P3oool00`2oM00007Bo0?ooo`030?ooo`030?noM0000000000000H0
000000H0001d0;ooo`3oool0okmd0000B02Lgol30?ooo`030=nLB000B9`0gooo00<0oooo00<0_g@0
0018W03Oool00P3oool00`3OW4P004RL0=ooo`060?ooo`030?oo_`1d000007Bo00H0oooo00<0_g@0
0000B02Lgol00P3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000M;oo0080oooo00<0omnL
04P0001d_ol00P3oool00`2oM000B7Ad07BLg`030?ooo`030?oo_`1d0000M;oo00<0oooo00H0_g@0
0018W03Oool0oono07@0001d_ol60?ooo`030=nLB000B9`0gooo00<0oooo00L0_g@00018W03Oool0
oooo0?oo_`1d0000M;oo00<0oooo00L0omnL04P0B02Lgol0oooo0=nLB000B4P0B7Bo0080oooo00<0
_g@0001d_`3oool01`3oool00`3ooml0W4P0001d_`070?ooo`030?oo_`1d0000M;oo00<0oooo00L0
_g@00018W03Oool0oooo0=nLB000B9`0gooo0080oooo00/0gia80018W03Oool0oooo0?oo_`1d04P0
W=oo0?ooo`3OW4P0000004RLg`020?ooo`030;md0018M7@0M9cO00@0oooo00<0gia80018B018M;l0
0P3oool00`2oM00007Bo0?ooo`060?ooo`0:0?oo_`1d0000B9cO0?oog`2LW;l0gia80000M02oool0
oooO09bL_`<0oooo00<0_g@00018W03Oool00P3oool00`3OW4P004RL0=ooo`020?ooo`030=nLB000
B4P0B7Bo0080oooo00<0_g@0001d_`3oool00`3oool00`3ookl0M00007Boo`030?ooo`030;md0000
B9`0gooo0080oooo00L0gia804RLg`3oool0oooo0?oo_`1d0000M;oo0080oooo00<0oooO09a80000
M;l03`3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00X0oooo00<0oooO09a8001d
_ol0O@3oool009H0oooo00L0oono07@0M02oool0oooo0?noM000B9`0gooo00h0oooo00X0omnL04P0
B02Lgol0oooo0=nLM02Lgol0oooo0?oOW01807@0_ooo0`3oool00`3OW4P0001809cOo`020?ooo`03
0?oOW01804P0W=oo00L0oooo00<0_g@00018W03Oool01@3oool00`3o_g@0000004RLg`020?ooo`03
0=nLB00004P0W=oo0080oooo00<0oooO09a80018W=l00P3oool00`3ooml0W4P004RLg`020?ooo`04
0?noM000B9`0_ia807Boo`<0oooo00<0okmd0000M02oool00P3oool03`3OW4P0001d0;ooo`3oool0
gia80000M02oool0oooo0?oo_`1d0000M;oo0?ooo`3ogi`0B01809cOo`030?ooo`030=nLB00004P0
W=oo0080oooo00<0gia80000M02oool00P3oool0303ookl0M00007Boo`3oool0okmd0000M02o_g@0
B9cO0?ooo`3o_g@0001d0;ooo`P0oooo00<0gia80000B018B0000`0000001`18W=l0oooo0?ooo`3o
ool0okmd0000M02oool00P3oool01`3OW4P0001d0;ooo`3oool0omnL04P0B02Lgol00P3oool00`3o
gi`0B01809cOo`020?ooo`0<0?noM000B9`0gooo0?ooo`3ookl0M00004RLg`3oool0okmd0018W02o
W4P0M;oo0`3oool0203o_g@0001d0;noM018W=l0oooo0?noM00007@0_ooo1P3oool02`3ookl0M000
07Boo`3oool0okmd00000000B9`0gooo0?ooo`3o_g@0M;oo0080oooo00<0gia80000B02Lgol00P3o
ool00`3ogi`0B01809cOo`020?ooo`080?noM00007@0_kmd04RLg`3oool0okmd0000M02oool40?oo
o`030?noM00007@0_ooo0080oooo00<0gia80000M02oool00P3oool00`3ogi`0B4RL0=ooo`020?oo
o`030?noM00007@0_ooo0080oooo00<0gia80000M02oool03`3oool00`2oM00004RL0=ooo`020?oo
o`030?oOW0180000B9cO00X0oooo00<0oono07@00018W=l0O@3oool009L0oooo00H0omnL04Q8W03O
ool0omnL04P0B02Lgol>0?ooo`0:0?oOW01804P0W=oo0?oog`2LB000M;oo0?ooo`3ookl0M00007Bo
o`<0oooo00<0omnL04P00018W=l00P3oool00`3ooml0W4P004RLg`070?ooo`030?noM00004P0W=oo
00D0oooo00<0omnL04P00000M;l00P3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo
0080oooo01<0_g@00018W03Oool0oooo0?oo_`1d04P0W=oo0;md001dM7@004RL0=ooo`3oool0gia8
0000M02oool0oooo0;md000004P0W=oo0080oooo00X0_g@00000M02oool0oooo0=nLB018W=l0oooo
0?oog`2LB000B9cO0`3oool00`3ogi`0B00004RLg`030?ooo`0@0;md000007@0_ooo0?ooo`3o_g@0
001d0;ooo`3oool0oooO09a8001d_ol0gia8001d_`3ogi`0B01809cOo`P0oooo00<0okmd0000B02L
gol00P3oool0103ogi`0B0000000M02oool20?ooo`0;0=nLB00007@0_ooo0?ooo`2oM000001809cO
o`3oool0oooO09a80018W=l00P3oool00`3ooml0W4P004RLg`030?ooo`0F0=nLB00007@0_ooo0?oo
o`3OW4P004RL0=ooo`3ookl0M01809cOo`2oM000M7Ad0018W03Oool0oooO09a8001d_ol0gia8001d
_`3ogi`0B01809cOo`H0oooo00<0okmd0000B02Lgol00P3oool01P2oM00004RL0=ooo`3oool0omnL
04Qd_`80oooo00<0omnL04P00018W=l00P3oool00`3ooml0W4P004RLg`020?ooo`080?oo_`1d04P0
W=oo0=nLB000M;l0omnL04P0B02Lgol50?ooo`070=nLB00007@0_ooo0?ooo`2oM000001809cOo`02
0?ooo`030?oog`2LB7@0_ooo00<0oooo00L0gia80000M02oool0oooo0;md000004P0W=oo00@0oooo
00<0_g@00000000000001P0000001@00M;l0oooo0=nLB00004P0W=oo0080oooo00<0oooO09a80000
M;l02P3oool00`3o_g@0000007Boo`1m0?ooo`00V03oool01@3o_g@007Bo0?oog`2LB000B9cO00l0
oooo00D0_g@0000000000000000004RLg`020?ooo`050;md0018W=l0oooo0?ooo`3o_g@00P000000
1000B9`0gooo0?oo_`1d00030000000307Boo`3oool0oooo00<0oooo00<0omnL04P0001d_ol01@3o
ool01@3ogi`0B000001d_`3oool0okmd0080000000@004RL0=ooo`3ookl0M0000`0000000`1d_kl0
M000000000020000000807BLW0180000000007Boo`3oool0_g@00000M02oool20?ooo`060?oOW018
0000000004RLW01d0000B9cO0`3oool01`3ogi`0B0000000000004P0W=oo0?oo_`1d00000`000000
1P1d_ol0okmd00000000000004RL0=ooo`<0oooo00l0okmd000000000000000007Boo`3oool0oono
07@000000000B9cO0?ooo`3o_g@000000000B02Lgol0203oool00`3ookl0M00004RLg`030?ooo`03
0?oOW018000007Bo00<0oooo00P0omnL04P000000000B9bL07@00018W=l0oono07@000<0000000<0
M;no07@0000000000P0000001@1d_ol0oooo0?ooo`3ogi`0B000008000000140M;oo0?oOW0180000
000007Boo`3oool0_g@00000M02ookl0M00000000018W=l0oooo0?noM0000000001809cOo`040?oo
o`050?oog`2LB00000000000M02oool00P3oool0203ogi`0B7Bo0?ooo`3ookl0M000001d_`3oool0
okmd0P0000001000B9`0gooo0?oo_`1d00030000000907Boo`3ooml0W4P004RLg`3oool0okmd0000
000004P0W=oo00H0oooo00H0omnL04P000000000B9bL07@00018W=l30?ooo`030;md001d_ol0oooo
00<0oooo00H0omnL04P000000000B9bo09a80018W=l?0?ooo`030?oOW0180000B9cO00<0oooo00<0
_g@00018W03Oool01P3oool01P3ogml0gooo0?noM0000000001d0;ooogd0oooo002I0?ooo`040=nL
B018W=l0_g@0001d_aH0oooo00<0gia80018W03Oool0203oool00`3o_g@0001d0;ooo`060?ooo`03
0?oog`2LB00007Bo00D0oooo00<0okmd00000018W=l0203oool00`3o_g@0001809cOo`020?ooo`03
0?noM00004P0W=oo01X0oooo00<0okmd0000M02oool08@3oool00`2oM00004RL0=ooo`030?ooo`03
0;md000007@0_ooo00X0oooo00<0okmd0000M02oool00P3oool00`3o_g@0001d0;ooo`0a0?ooo`03
0?noM00007@0_ooo0080oooo00<0gia8001d_`3oool04`3oool00`3o_g@007Bo0?ooo`070?ooo`03
0;md0000M;l0oooo00h0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?oo
o`050?oo_`1d000000000000B02Lgol0OP3oool009X0oooo00@0gia807AdM00007@0_ooo5@3oool0
0`3ogi`0B01d0;ooo`090?ooo`030=nLM02Lgol0oooo00H0oooo00<0_g@00018W03Oool00`3oool0
103ooml0W4P00000M02oool80?ooo`030?oog`2LB00007Bo0080oooo00<0oooO09a80000M;l06`3o
ool00`3OW7@0W=oo0?ooo`0P0?ooo`030=nLB00007@0_ooo00<0oooo00<0_g@00000M02oool02`3o
ool00`3OW7@0W=oo0?ooo`020?ooo`030=nLM02Lgol0oooo0340oooo00L0giad09cOo`3oool0oooo
0?noM00007@0_ooo01<0oooo00<0oono07@0M02oool01`3oool00`3o_g@0001d0;ooo`0?0?ooo`03
0?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01`3oool00`3OW9`0_ooo0?ooo`1o0?ooo`00
V`3oool00`2oM000001809cOo`0E0?ooo`030?oog`2LB000M;oo00<0oooo00@0oono07@00000B9`0
gooo2`3oool00`3OW4P0001809cOo`020?ooo`040?oo_`1d0000000007Boo`@0oooo00@0oono07@0
0000B9`0gooo0P3oool00`3o_g@0001d0;ooo`020?ooo`030?noM00007@0_ooo01l0oooo00@0oono
07@00000B9`0gooo6P3oool00`3o_g@0000007Boo`020?ooo`030?noM000000007Bo03l0oooo00@0
oono07@00000B9`0gooo1P3oool00`3ookl0M01809cOo`0D0?ooo`030;md001d_ol0oooo00H0oooo
00<0omnL04P0B02Lgol0403oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`P0oooo00<0omnL07BL
g`3oool0O`3oool009/0oooo00<0oooO09a80018W=l05P3oool00`3OW4P0B9cO0?ooo`0@0?ooo`03
0;md00000000000000@0000000<004RL0=ooo`3oool02`3oool02P3OW4P004RL0=oog`2LB000B9cO
0=nLB000B9`0gooO09a80018W=li0?ooo`030;md00000000000000@0000000<007Bo0?ooo`3oool0
B@3oool00`2oM000M;oo0?ooo`0C0?ooo`030=nLB018W=l0oooo00H0oooo00<0oooO09a80018W=l0
4@3oool0103OW4P0000000000000M;l:0?ooo`050;md00000000000000000018W=l0O03oool00;80
oooo00@0oooO09a80000000007Bo9`3oool02@3OW4P000000000B02Lgol0oooo0=nLB00000000018
09cOo`2:0?ooo`040?oo_`1d00000000001d_a@0oooo00<0omnL04Q8W03Oool01@3oool0103ookl0
M00000000000M;nP0?ooo`00o`3ooooo0?ooocH0oooo003o0?ooool0oooo=P3oool00?l0ooooo`3o
oolf0?ooo`00o`3ooooo0?ooocH0oooo0000\
\>"],
  ImageRangeCache->{{{0, 563}, {281, 0}} -> {-6.73168, -4.12995, 0.0342971, \
0.0244979}}]
}, Closed]],

Cell[CellGroupData[{

Cell["Hydrogenic wave functions", "ExerciseTitle"],

Cell["Wave functions for the hydrogen atom take the form", "ExerciseText"],

Cell[BoxData[
    \(TraditionalForm\`\(\[Psi]\_\(n\[InvisibleComma]\[ScriptL]\
\[InvisibleComma]\ m\)\)(
        r, \[Theta], \[Phi]) \[Proportional] \[ExponentialE]\^\(\(-\[Kappa]\)\
\ r\)\ \((\[Kappa]\ r)\)\^\[ScriptL]\ \(\(L\_\(n - \[ScriptL] - 
                1\)\%\(2\ \[ScriptL] + 1\)\)(
          2\ \[Kappa]\ r)\)\ \(\(Y\_\[ScriptL]\%m\)(\[Theta], \[Phi])\)\)], \
"DisplayFormula",
  Evaluatable->False,
  TextAlignment->-0.5],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`n\)]],
  " is the principal (radial) quantum number, ",
  Cell[BoxData[
      \(TraditionalForm\`\[ScriptL]\)]],
  " is the orbital angular momentum in units of ",
  Cell[BoxData[
      \(TraditionalForm\`\[HBar]\)]],
  ", ",
  Cell[BoxData[
      \(TraditionalForm\`m\)]],
  " is the magnetic (azimuthal) quantum number, and ",
  Cell[BoxData[
      \(TraditionalForm\`\[HBar]\ \[Kappa] = \@\(2\ \[Mu]\ E\_b\)\)]],
  "is the wave number for positive binding energy ",
  Cell[BoxData[
      \(TraditionalForm\`E\_b\)]],
  ".  The radial dependence involves generalized Laguerre functions and the \
angular dependence spherical harmonics.  Note that we are not interested in \
the normalization of these functions at this time.  The probability density \
is given by the absolute square of the wave function, namely ",
  Cell[BoxData[
      \(TraditionalForm\`\[LeftBracketingBar]\[Psi]\[RightBracketingBar]\^2\)]\
],
  ".  Write a function that plots the probability density in the ",
  Cell[BoxData[
      \(TraditionalForm\`x\[VeryThinSpace]y\)]],
  " plane (",
  Cell[BoxData[
      \(TraditionalForm\`\[Theta] = \[Pi]\/2\)]],
  ") given values of the quantum numbers ",
  Cell[BoxData[
      \(TraditionalForm\`n\[VeryThinSpace]\[ScriptL]\[VeryThinSpace]m\)]],
  "\[VeryThinSpace] using ",
  StyleBox["Plot3D", "InlineInput",
    FontWeight->"Bold"],
  ".   You will need to express the Cartesian coordinates in terms of \
spherical coordinates.  Write a similar function that uses ",
  StyleBox["DensityPlot", "InlineInput",
    FontWeight->"Bold"],
  ".  Show several cases of each."
}], "ExerciseText",
  CellDingbat->None]
}, Closed]],

Cell[CellGroupData[{

Cell["Sampling points", "ExerciseTitle"],

Cell[TextData[{
  "Write a function which displays the points sampled by ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " 's adaptive algorithm for ",
  StyleBox["Plot", "InlineInput",
    FontWeight->"Bold"],
  " as vertical bars between the x-axis and the points themselves."
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Curve shown by tangents", "ExerciseTitle"],

Cell["\<\
Gass suggests the following exercise.  Write a routine which plots a function \
not as a curve but as a set of short line segments tangent to the curve at \
equally spaced points.\
\>", "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Compare lengths of circular arc and chord", "ExerciseTitle"],

Cell["\<\
Draw a figure that can be used to derive the difference in length between a \
circular arc and a chord connecting a pair of points.  Your figure should \
include the arc, the chord, and radial lines connecting the center of \
curvature with the two points and with the center of the arc.  Label each \
part of the figure appropriately.\
\>", "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Potential and field for uniform sphere", "ExerciseTitle"],

Cell[TextData[{
  "The electric field of a uniformly charged sphere or the gravitational \
field of a sphere of uniform density both have the same shape.  Let ",
  Cell[BoxData[
      \(TraditionalForm\`\[Phi] = 3\/2 - r\^2\/2\)]],
  " for ",
  Cell[BoxData[
      \(TraditionalForm\`r \[LessEqual] 1\)]],
  " or ",
  Cell[BoxData[
      \(TraditionalForm\`\[Phi] = r\^\(-1\)\)]],
  " for ",
  Cell[BoxData[
      \(TraditionalForm\`r \[GreaterEqual] 1\)]],
  " represent the normalized potential, where ",
  Cell[BoxData[
      \(TraditionalForm\`r = \@\(x\^2 + y\^2 + z\^2\)\)]],
  " is the distance from the center in units of the radius of the sphere.  \
The field if then ",
  Cell[BoxData[
      \(TraditionalForm\`f\&\[RightVector] = \(-\[Del]\&\[RightVector] \[Phi]\
\)\)]],
  ".\na) Use ",
  StyleBox["ContourPlot", "InlineInput",
    FontWeight->"Bold"],
  " to display equipotential lines in the ",
  StyleBox["xy",
    FontSlant->"Italic"],
  " plane (two-dimensions).\nb) Use ",
  StyleBox["PlotGradientField", "InlineInput",
    FontWeight->"Bold"],
  " to display the force field in the ",
  StyleBox["xy",
    FontSlant->"Italic"],
  " plane.  Combine this plot with the equipotentials from the preceding \
exercise.\nc) Use ",
  StyleBox["ContourPlot3D", "InlineInput",
    FontWeight->"Bold"],
  " and ",
  StyleBox["PlotGradientField3D", "InlineInput",
    FontWeight->"Bold"],
  " to display equipotential surfaces and the vector field within a \
representative volume of space. "
}], "ExerciseText"]
}, Closed]]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.2 for Microsoft Windows",
ScreenRectangle->{{0, 1400}, {0, 952}},
WindowToolbars->{"RulerBar", "EditBar"},
WindowSize->{1392, 912},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
PrintingCopies->1,
PrintingPageRange->{Automatic, Automatic},
ShowSelection->True,
ShowCellLabel->False,
Magnification->1.5,
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions", "Subtitle"],
  
  Cell["\<\
Modify the definitions below to change the default appearance of all cells in \
a given style.  Make modifications to any definition using commands in the \
Format menu.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Section"],
  
  Cell[StyleData[All, "Working"],
    PageWidth->WindowWidth,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Presentation"],
    PageWidth->WindowWidth,
    ScriptMinSize->12,
    FontSize->16],
  
  Cell[StyleData[All, "Condensed"],
    PageWidth->PaperWidth,
    CellBracketOptions->{"Margins"->{1, 1},
    "Widths"->{0, 5}},
    ScriptMinSize->7,
    FontSize->10],
  
  Cell[StyleData[All, "Printout"],
    PageWidth->PaperWidth,
    ScriptMinSize->7,
    FontSize->10,
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Section"],
  
  Cell["\<\
The options defined for the style below will be used at the Notebook level.\
\>", "Text"],
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"]}, {Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PageFooters->{{Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}, {
        Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}},
    PageHeaderLines->{True, True},
    PageFooterLines->{True, True},
    PrintingOptions->{"FirstPageHeader"->False,
    "FacingPages"->True},
    CellFrameLabelMargins->6,
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headings", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    ShowCellBracket->True,
    CellMargins->{{0, 0}, {0, 0}},
    CellGroupingRules->"TitleGrouping",
    PageBreakBelow->False,
    LineSpacing->{1, 0},
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontSize->34,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Title", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 0},
    FontSize->44],
  
  Cell[StyleData["Title", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24,
    FontTracking->"Plain",
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    CellMargins->{{0, 0}, {0, 0}},
    PageBreakBelow->False,
    LineSpacing->{1, 3},
    ParagraphIndent->-96,
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->18,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Subtitle", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 6},
    ParagraphIndent->-157,
    FontSize->30],
  
  Cell[StyleData["Subtitle", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-78,
    FontSize->14],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-85,
    FontSize->16,
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    ShowCellBracket->True,
    CellMargins->{{10, 4}, {30, 10}},
    PageBreakBelow->False,
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Presentation"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {40, 20}},
    LineSpacing->{1, 0},
    FontSize->24],
  
  Cell[StyleData["Subsubtitle", "Condensed"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {12, 8}},
    FontSize->12],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    ShowCellBracket->True,
    CellMargins->{{9, 10}, {50, 10}},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellFrame->{{0, 0}, {0, 1}},
    CellDingbat->None,
    CellMargins->{{12, Inherited}, {4, 24}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->6,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Section", "Presentation"],
    CellMargins->{{10, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["Section", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["Section", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellMargins->{{12, Inherited}, {8, 20}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsection", "Presentation"],
    CellMargins->{{11, 10}, {8, 32}},
    LineSpacing->{1, 0},
    FontSize->22],
  
  Cell[StyleData["Subsection", "Condensed"],
    CellMargins->{{8, Inherited}, {2, 12}},
    FontSize->12],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{9, 0}, {4, 40}},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{25, Inherited}, {8, 12}},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsubsection", "Presentation"],
    CellMargins->{{29, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Subsubsection", "Condensed"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Body Text", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Text",
    FontFamily->"Times"],
  
  Cell[StyleData["Text", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Text", "Condensed"],
    CellMargins->{{9, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{9, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallText"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 6},
    CounterIncrements->"SmallText",
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["SmallText", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["SmallText", "Condensed"],
    CellMargins->{{8, 10}, {2, 2}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["SmallText", "Printout"],
    CellMargins->{{9, 0}, {4, 4}}]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Input/Output", "Section"],
  
  Cell["\<\
The cells in this section define styles used for input and output to the \
kernel.  Be careful when modifying, renaming, or removing these styles, \
because the front end associates special meanings with these style names.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    CellFrame->0.5,
    CellMargins->{{25, 10}, {8, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{5, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold",
    Background->RGBColor[1, 0.700008, 0.4]],
  
  Cell[StyleData["Input", "Presentation"],
    CellMargins->{{25, Inherited}, {10, 10}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Input", "Condensed"],
    CellMargins->{{25, 0}, {6, 6}}],
  
  Cell[StyleData["Input", "Printout"],
    CellFrame->0.5,
    CellMargins->{{25, 0}, {6, 6}},
    Background->None]
  }, Closed]],
  
  Cell[StyleData["InlineInput"],
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    CellFrame->True,
    CellMargins->{{25, 10}, {8, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelMargins->{{3, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Output",
    Background->RGBColor[1, 0.900008, 0.900008]],
  
  Cell[StyleData["Output", "Presentation"],
    CellMargins->{{25, Inherited}, {12, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Output", "Condensed"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4}],
  
  Cell[StyleData["Output", "Printout"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontColor->RGBColor[1, 0, 0]],
  
  Cell[StyleData["Message", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Message", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Message", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Print",
    StyleMenuListing->None],
  
  Cell[StyleData["Print", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Print", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Print", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Graphics",
    StyleMenuListing->None],
  
  Cell[StyleData["Graphics", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Graphics", "Condensed"],
    CellMargins->{{52, Inherited}, {Inherited, Inherited}},
    ImageSize->{175, 175}],
  
  Cell[StyleData["Graphics", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    ImageSize->{250, 250}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->11,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["CellLabel", "Presentation"],
    FontSize->12],
  
  Cell[StyleData["CellLabel", "Condensed"],
    FontSize->8],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontSize->8,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Unique Styles", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Author"],
    CellMargins->{{10, 4}, {2, 10}},
    FontSize->12,
    FontSlant->"Italic"],
  
  Cell[StyleData["Author", "Presentation"],
    CellMargins->{{12, 10}, {2, 12}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Author", "Condensed"],
    CellMargins->{{8, 10}, {1, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4},
    FontSize->9],
  
  Cell[StyleData["Author", "Printout"],
    CellMargins->{{9, 0}, {4, 12}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Copyright"],
    CellMargins->{{10, 10}, {40, 2}},
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["Copyright", "Presentation"],
    CellMargins->{{12, 10}, {50, 2}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["Copyright", "Condensed"],
    CellMargins->{{8, 10}, {12, 1}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["Copyright", "Printout"],
    CellMargins->{{9, 0}, {72, 4}},
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Abstract"],
    CellMargins->{{45, 75}, {Inherited, 30}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Abstract", "Printout"],
    CellMargins->{{36, 67}, {Inherited, 50}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Example"],
    CellMargins->{{12, 10}, {5, 12}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Example",
    FontFamily->"Times",
    FontWeight->"Bold"],
  
  Cell[StyleData["Example", "Presentation"],
    CellMargins->{{18, 10}, {8, 20}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["Example", "Condensed"],
    CellMargins->{{8, 10}, {4, 8}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4}],
  
  Cell[StyleData["Example", "Printout"],
    CellMargins->{{9, 0}, {4, 10}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Definition"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Definition",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Definition", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Definition", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Definition", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.6]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Theorem"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Theorem",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Theorem", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Theorem", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Theorem", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.4]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Caution"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Caution",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0.996109, 0, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Caution", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Caution", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Caution", "Printout"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Advice"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Advice",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0, 1, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Advice", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Advice", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Advice", "Printout"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Exercise"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->"NormalGrouping",
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 0.996109]],
  
  Cell[StyleData["Exercise", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Exercise", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Exercise", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseMain"],
    CellFrame->{{6, 0}, {0, 1}},
    CellMargins->{{12, Inherited}, {4, 20}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.571389, 0.19675, 0.570504],
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseMain", "Presentation"],
    CellMargins->{{18, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["ExerciseMain", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["ExerciseMain", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    CellFrameColor->GrayLevel[0.500008],
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseTitle"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseTitle", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["ExerciseTitle", "Condensed"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["ExerciseTitle", "Printout"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseText"],
    CellMargins->{{32, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"ExcersiceText",
    FontFamily->"Times"],
  
  Cell[StyleData["ExerciseText", "Presentation"],
    CellMargins->{{32, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["ExerciseText", "Condensed"],
    CellMargins->{{32, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["ExerciseText", "Printout"],
    CellMargins->{{32, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Problem"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Problem", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Problem", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Problem", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    CellHorizontalScrolling->True,
    DefaultFormatType->DefaultInputFormatType,
    ScriptLevel->0,
    SingleLetterItalics->True,
    UnderoverscriptBoxOptions->{LimitsPositioning->True}],
  
  Cell[StyleData["DisplayFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChemicalFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    DefaultFormatType->DefaultInputFormatType,
    AutoSpacing->False,
    ScriptLevel->1,
    ScriptBaselineShifts->{0.6, Automatic},
    SingleLetterItalics->False,
    ZeroWidthTimes->True],
  
  Cell[StyleData["ChemicalFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Program"],
    CellMargins->{{12, 10}, {Inherited, 6}},
    FontFamily->"Courier"],
  
  Cell[StyleData["Program", "Presentation"],
    CellMargins->{{13, 30}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Condensed"],
    CellMargins->{{8, 10}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Printout"],
    CellMargins->{{9, 0}, {Inherited, 4}},
    FontSize->9.5]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Automatic Numbering", "Section"],
  
  Cell["\<\
The following styles are useful for numbered equations, figures, etc.  They \
automatically give the cell a FrameLabel containing a reference to a \
particular counter, and also increment that counter.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedEquation"],
    CellMargins->{{62, 10}, {Inherited, Inherited}},
    CellFrameLabels->{{None, Cell[ 
            TextData[ {"(", 
              CounterBox[ "NumberedEquation"], ")"}]]}, {None, None}},
    DefaultFormatType->DefaultInputFormatType,
    CounterIncrements->"NumberedEquation",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedEquation", "Presentation"],
    CellMargins->{{74, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Condensed"],
    CellMargins->{{52, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Printout"],
    CellMargins->{{54, 0}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedFigure"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Figure ", 
              CounterBox[ "NumberedFigure"]}]], None}},
    CounterIncrements->"NumberedFigure",
    ImageMargins->{{43, Inherited}, {Inherited, 0}},
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedFigure", "Presentation"]],
  
  Cell[StyleData["NumberedFigure", "Condensed"]],
  
  Cell[StyleData["NumberedFigure", "Printout"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedTable"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Table ", 
              CounterBox[ "NumberedTable"]}]], None}},
    TextAlignment->Center,
    CounterIncrements->"NumberedTable",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedTable", "Presentation"]],
  
  Cell[StyleData["NumberedTable", "Condensed"]],
  
  Cell[StyleData["NumberedTable", "Printout"]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headers and Footers", "Section"],
  
  Cell[StyleData["Header"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["Footer"],
    CellMargins->{{0, 0}, {0, 4}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["PageNumber"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Bold"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlink Styles", "Section"],
  
  Cell["\<\
The cells below define styles useful for making hypertext ButtonBoxes.  The \
\"Hyperlink\" style is for links within the same Notebook, or between \
Notebooks.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonFrame->"None",
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Presentation"]],
  
  Cell[StyleData["Hyperlink", "Condensed"]],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell["\<\
The following styles are for linking automatically to the on-line help \
system.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Presentation"]],
  
  Cell[StyleData["MainBookLink", "Condensed"]],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnsLink", "Presentation"]],
  
  Cell[StyleData["AddOnsLink", "Condensed"]],
  
  Cell[StyleData["AddOnLink", "Printout"],
    FontWeight->"Bold",
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuideLink", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Presentation"]],
  
  Cell[StyleData["RefGuideLink", "Condensed"]],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Presentation"]],
  
  Cell[StyleData["GettingStartedLink", "Condensed"]],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Presentation"]],
  
  Cell[StyleData["OtherInformationLink", "Condensed"]],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]]
  }, Closed]]
  }, Open  ]]
  }]
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "Info3350813126-5278694"->{
    Cell[1979940, 40927, 361, 9, 35, "Print",
      CellTags->"Info3350813126-5278694"]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"Info3350813126-5278694", 7331544, 189986}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 80, 2, 81, "Title"],
Cell[1859, 57, 455, 15, 85, "Author"],

Cell[CellGroupData[{
Cell[2339, 76, 31, 0, 83, "Section"],
Cell[2373, 78, 1228, 22, 171, "Text"],
Cell[3604, 102, 1070, 22, 146, "Text"],
Cell[4677, 126, 5121, 127, 313, "Text"],
Cell[9801, 255, 1036, 24, 147, "Text"],
Cell[10840, 281, 2503, 70, 312, "Text"],
Cell[13346, 353, 834, 16, 120, "Text"],
Cell[14183, 371, 222, 4, 67, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[14442, 380, 33, 0, 51, "Section"],
Cell[14478, 382, 223, 4, 125, "Input"],
Cell[14704, 388, 57, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[14786, 392, 44, 1, 73, "Input"],
Cell[14833, 395, 9776, 242, 150, "Print"]
}, Closed]],
Cell[24624, 640, 52, 0, 41, "Text"],
Cell[24679, 642, 55, 1, 73, "Input"],
Cell[24737, 645, 73, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[24835, 649, 44, 1, 73, "Input"],
Cell[24882, 652, 37442, 932, 1070, "Print"]
}, Closed]],
Cell[62339, 1587, 104, 3, 41, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[62480, 1595, 54, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[62559, 1599, 34, 0, 68, "Subsection"],
Cell[62596, 1601, 908, 28, 95, "Text"],

Cell[CellGroupData[{
Cell[63529, 1633, 65, 1, 73, "Input"],
Cell[63597, 1636, 36549, 915, 279, 13729, 629, "GraphicsData", "PostScript", \
"Graphics"],
Cell[100149, 2553, 130, 3, 59, "Output"]
}, Closed]],
Cell[100294, 2559, 1036, 18, 147, "Text"],
Cell[101333, 2579, 130, 5, 75, "Exercise"],
Cell[101466, 2586, 115, 3, 58, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[101618, 2594, 41, 0, 68, "Subsection"],
Cell[101662, 2596, 349, 6, 67, "Text"],

Cell[CellGroupData[{
Cell[102036, 2606, 71, 1, 73, "Input"],
Cell[102110, 2609, 345, 5, 35, "Message"],
Cell[102458, 2616, 345, 5, 35, "Message"],
Cell[102806, 2623, 345, 5, 35, "Message"],
Cell[103154, 2630, 328, 5, 35, "Message"],
Cell[103485, 2637, 33246, 797, 279, 11917, 529, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[136746, 3437, 732, 15, 119, "Text"],

Cell[CellGroupData[{
Cell[137503, 3456, 75, 1, 73, "Input"],
Cell[137581, 3459, 40068, 919, 279, 11942, 567, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[177686, 4383, 108, 2, 99, "Input"],
Cell[177797, 4387, 40068, 919, 279, 11942, 567, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[217880, 5309, 198, 5, 42, "Text"],
Cell[218081, 5316, 84, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[218190, 5321, 66, 1, 73, "Input"],
Cell[218259, 5324, 35835, 870, 279, 12039, 572, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[254109, 6197, 360, 10, 42, "Text"],
Cell[254472, 6209, 209, 4, 73, "Input"],
Cell[254684, 6215, 390, 8, 68, "Text"],

Cell[CellGroupData[{
Cell[255099, 6227, 61, 1, 73, "Input"],
Cell[255163, 6230, 301, 5, 35, "Message"],
Cell[255467, 6237, 858, 13, 86, "Message"],
Cell[256328, 6252, 301, 5, 35, "Message"],
Cell[256632, 6259, 910, 13, 86, "Message"],
Cell[257545, 6274, 301, 5, 35, "Message"],
Cell[257849, 6281, 332, 5, 35, "Message"],
Cell[258184, 6288, 858, 13, 86, "Message"],
Cell[259045, 6303, 334, 5, 35, "Message"],
Cell[259382, 6310, 326, 4, 35, "Message"],
Cell[259711, 6316, 323, 4, 35, "Message"],
Cell[260037, 6322, 322, 4, 35, "Message"],
Cell[260362, 6328, 328, 5, 35, "Message"],
Cell[260693, 6335, 33246, 797, 279, 11917, 529, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[293954, 7135, 605, 16, 95, "Text"],

Cell[CellGroupData[{
Cell[294584, 7155, 72, 1, 73, "Input"],
Cell[294659, 7158, 39556, 893, 279, 11213, 539, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[334230, 8054, 273, 5, 67, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[334540, 8064, 45, 0, 68, "Subsection"],
Cell[334588, 8066, 174, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[334787, 8073, 120, 2, 73, "Input"],
Cell[334910, 8077, 44874, 1156, 279, 17133, 809, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[379821, 9238, 128, 3, 73, "Input"],
Cell[379952, 9243, 140722, 3286, 279, 33694, 1960, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[520723, 12535, 43, 0, 68, "Subsection"],

Cell[CellGroupData[{
Cell[520791, 12539, 50, 0, 55, "Subsubsection"],
Cell[520844, 12541, 514, 10, 94, "Text"],

Cell[CellGroupData[{
Cell[521383, 12555, 46, 1, 73, "Input"],
Cell[521432, 12558, 1003, 17, 205, "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[522484, 12581, 47, 0, 55, "Subsubsection"],
Cell[522534, 12583, 544, 16, 69, "Text"],

Cell[CellGroupData[{
Cell[523103, 12603, 189, 3, 99, "Input"],
Cell[523295, 12608, 77687, 1771, 279, 21873, 1077, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[600997, 14382, 1014, 26, 123, "Text"],

Cell[CellGroupData[{
Cell[602036, 14412, 190, 3, 73, "Input"],
Cell[602229, 14417, 85292, 2094, 279, 26680, 1366, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[687570, 16517, 35, 0, 55, "Subsubsection"],
Cell[687608, 16519, 269, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[687902, 16529, 248, 4, 99, "Input"],
Cell[688153, 16535, 93830, 2200, 279, 25851, 1356, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[782032, 18741, 35, 0, 55, "Subsubsection"],
Cell[782070, 18743, 320, 5, 67, "Text"],

Cell[CellGroupData[{
Cell[782415, 18752, 78, 1, 73, "Input"],
Cell[782496, 18755, 50700, 1171, 279, 15173, 728, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[833211, 19929, 303, 8, 68, "Text"],

Cell[CellGroupData[{
Cell[833539, 19941, 131, 2, 73, "Input"],
Cell[833673, 19945, 57074, 1201, 279, 14232, 668, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[890762, 21149, 674, 16, 95, "Text"],

Cell[CellGroupData[{
Cell[891461, 21169, 141, 3, 73, "Input"],
Cell[891605, 21174, 57074, 1201, 279, 14232, 668, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[948728, 22381, 58, 0, 55, "Subsubsection"],
Cell[948789, 22383, 1353, 30, 175, "Text"],
Cell[950145, 22415, 305, 7, 68, "Text"],

Cell[CellGroupData[{
Cell[950475, 22426, 137, 2, 73, "Input"],
Cell[950615, 22430, 62507, 1559, 279, 19887, 1028, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[1013137, 23992, 287, 5, 67, "Text"],

Cell[CellGroupData[{
Cell[1013449, 24001, 90, 1, 74, "Input"],
Cell[1013542, 24004, 362826, 5392, 279, 22178, 1182, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[1376383, 29399, 308, 7, 68, "Text"],

Cell[CellGroupData[{
Cell[1376716, 29410, 123, 2, 74, "Input"],
Cell[1376842, 29414, 451480, 7917, 279, 50151, 2958, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[1828337, 37334, 259, 5, 67, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[1828633, 37344, 44, 0, 55, "Subsubsection"],
Cell[1828680, 37346, 261, 5, 67, "Text"],

Cell[CellGroupData[{
Cell[1828966, 37355, 50, 1, 73, "Input"],
Cell[1829019, 37358, 10111, 149, 2234, "Output"]
}, Closed]],
Cell[1839145, 37510, 757, 20, 96, "Text"],
Cell[1839905, 37532, 85, 1, 73, "Input"],
Cell[1839993, 37535, 72, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[1840090, 37540, 90, 1, 73, "Input"],
Cell[1840183, 37543, 43842, 1199, 279, 19602, 895, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[1884040, 38745, 96, 2, 41, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[1884173, 38752, 37, 0, 55, "Subsubsection"],
Cell[1884213, 38754, 385, 6, 67, "Text"],

Cell[CellGroupData[{
Cell[1884623, 38764, 108, 2, 86, "Input"],
Cell[1884734, 38768, 39360, 966, 279, 13020, 636, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[1924109, 39737, 233, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[1924367, 39747, 129, 2, 86, "Input"],
Cell[1924499, 39751, 55065, 1161, 444, 12889, 636, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[1979579, 40915, 284, 5, 67, "Text"],

Cell[CellGroupData[{
Cell[1979888, 40924, 49, 1, 73, "Input"],
Cell[1979940, 40927, 361, 9, 35, "Print",
  CellTags->"Info3350813126-5278694"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[1980350, 40942, 36, 0, 55, "Subsubsection"],
Cell[1980389, 40944, 1571, 43, 147, "Text"],
Cell[1981963, 40989, 348, 6, 125, "Input"],
Cell[1982314, 40997, 72, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[1982411, 41002, 141, 2, 73, "Input"],
Cell[1982555, 41006, 52641, 1264, 279, 15569, 802, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[2035211, 42273, 324, 10, 67, "Text"],
Cell[2035538, 42285, 406, 13, 42, "Text"],
Cell[2035947, 42300, 444, 13, 46, "Text"],
Cell[2036394, 42315, 492, 14, 42, "Text"],
Cell[2036889, 42331, 632, 19, 46, "Text"],
Cell[2037524, 42352, 335, 9, 42, "Text"],
Cell[2037862, 42363, 299, 9, 42, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[2038198, 42377, 36, 0, 55, "Subsubsection"],
Cell[2038237, 42379, 649, 13, 94, "Text"],
Cell[2038889, 42394, 115, 2, 73, "Input"],
Cell[2039007, 42398, 170, 5, 42, "Text"],
Cell[2039180, 42405, 280, 5, 99, "Input"],
Cell[2039463, 42412, 117, 2, 73, "Input"],

Cell[CellGroupData[{
Cell[2039605, 42418, 47, 1, 73, "Input"],
Cell[2039655, 42421, 93823, 2201, 279, 25722, 1356, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[2133493, 44625, 138, 5, 42, "Text"],
Cell[2133634, 44632, 449, 12, 68, "Text"],
Cell[2134086, 44646, 238, 6, 42, "Text"],
Cell[2134327, 44654, 224, 6, 42, "Text"],
Cell[2134554, 44662, 218, 6, 42, "Text"],
Cell[2134775, 44670, 474, 17, 42, "Text"],
Cell[2135252, 44689, 692, 26, 42, "Text"],
Cell[2135947, 44717, 195, 3, 73, "Input"],

Cell[CellGroupData[{
Cell[2136167, 44724, 47, 1, 73, "Input"],
Cell[2136217, 44727, 94652, 2202, 279, 26494, 1356, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[2230884, 46932, 115, 2, 73, "Input"],
Cell[2231002, 46936, 959, 31, 95, "Text"],
Cell[2231964, 46969, 225, 7, 58, "Exercise"],

Cell[CellGroupData[{
Cell[2232214, 46980, 487, 8, 151, "Input"],
Cell[2232704, 46990, 97427, 2239, 279, 26249, 1356, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[2330146, 49232, 359, 11, 42, "Text"],
Cell[2330508, 49245, 730, 13, 203, "Input"],
Cell[2331241, 49260, 54, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[2331320, 49264, 89, 1, 73, "Input"],
Cell[2331412, 49267, 2082, 50, 40, "Print"]
}, Closed]],
Cell[2333509, 49320, 39, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[2333573, 49324, 207, 3, 73, "Input"],
Cell[2333783, 49329, 102377, 2301, 279, 26180, 1356, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[2436221, 51637, 48, 0, 68, "Subsection"],
Cell[2436272, 51639, 364, 7, 68, "Text"],
Cell[2436639, 51648, 464, 11, 81, "Text"],
Cell[2437106, 51661, 113, 2, 106, "Input"],

Cell[CellGroupData[{
Cell[2437244, 51667, 152, 2, 73, "Input"],
Cell[2437399, 51671, 50249, 1099, 279, 13197, 637, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[2487663, 52773, 536, 7, 93, "Text"],
Cell[2488202, 52782, 294, 6, 41, "Text"],
Cell[2488499, 52790, 215, 5, 41, "Text"],
Cell[2488717, 52797, 218, 5, 41, "Text"],
Cell[2488938, 52804, 257, 5, 41, "Text"],
Cell[2489198, 52811, 116, 3, 41, "Text"],
Cell[2489317, 52816, 78, 1, 73, "Input"],
Cell[2489398, 52819, 60, 0, 41, "Text"],
Cell[2489461, 52821, 84, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[2489570, 52826, 59, 1, 73, "Input"],
Cell[2489632, 52829, 68, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[2489737, 52835, 59, 1, 73, "Input"],
Cell[2489799, 52838, 68, 1, 59, "Output"]
}, Closed]],
Cell[2489882, 52842, 227, 5, 65, "Exercise"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[2490158, 52853, 48, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[2490231, 52857, 51, 0, 68, "Subsection"],
Cell[2490285, 52859, 1057, 23, 148, "Text"],
Cell[2491345, 52884, 96, 2, 41, "Text"],
Cell[2491444, 52888, 502, 11, 177, "Input"],
Cell[2491949, 52901, 181, 5, 42, "Text"],

Cell[CellGroupData[{
Cell[2492155, 52910, 93, 1, 73, "Input"],
Cell[2492251, 52913, 41848, 1012, 279, 14032, 664, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[2534148, 53931, 52, 0, 68, "Subsection"],
Cell[2534203, 53933, 664, 16, 96, "Text"],

Cell[CellGroupData[{
Cell[2534892, 53953, 122, 2, 73, "Input"],
Cell[2535017, 53957, 67250, 1456, 279, 16552, 826, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[2602282, 55416, 459, 12, 69, "Text"],

Cell[CellGroupData[{
Cell[2602766, 55432, 445, 8, 177, "Input"],
Cell[2603214, 55442, 84734, 2006, 279, 25352, 1268, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[2687963, 57451, 263, 8, 42, "Text"],
Cell[2688229, 57461, 209, 4, 83, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[2688475, 57470, 47, 0, 68, "Subsection"],
Cell[2688525, 57472, 743, 19, 95, "Text"],

Cell[CellGroupData[{
Cell[2689293, 57495, 68, 1, 73, "Input"],
Cell[2689364, 57498, 41954, 1031, 279, 14409, 687, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[2731333, 58532, 59, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[2731417, 58536, 86, 1, 73, "Input"],
Cell[2731506, 58539, 85472, 2532, 146, 42085, 1992, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[2816993, 61074, 157, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[2817175, 61081, 92, 1, 73, "Input"],
Cell[2817270, 61084, 83763, 2511, 394, 42084, 1992, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[2901048, 63598, 424, 14, 42, "Text"],
Cell[2901475, 63614, 89, 2, 58, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[2901601, 63621, 49, 0, 68, "Subsection"],
Cell[2901653, 63623, 392, 6, 93, "Text"],
Cell[2902048, 63631, 795, 16, 127, "Text"],
Cell[2902846, 63649, 192, 3, 107, "Input"],

Cell[CellGroupData[{
Cell[2903063, 63656, 69, 1, 73, "Input"],
Cell[2903135, 63659, 38602, 876, 279, 11517, 537, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[2941752, 64538, 172, 3, 73, "Input"],

Cell[CellGroupData[{
Cell[2941949, 64545, 117, 2, 73, "Input"],
Cell[2942069, 64549, 56840, 1360, 279, 17467, 869, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[2998970, 65916, 34, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[2999029, 65920, 38, 0, 68, "Subsection"],
Cell[2999070, 65922, 352, 9, 68, "Text"],
Cell[2999425, 65933, 155, 3, 103, "Input"],
Cell[2999583, 65938, 143, 3, 96, "Input"],
Cell[2999729, 65943, 492, 14, 68, "Text"],

Cell[CellGroupData[{
Cell[3000246, 65961, 134, 3, 73, "Input"],
Cell[3000383, 65966, 36060, 877, 279, 12756, 585, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[3036458, 66846, 75, 0, 41, "Text"],
Cell[3036536, 66848, 143, 3, 96, "Input"],

Cell[CellGroupData[{
Cell[3036704, 66855, 128, 2, 73, "Input"],
Cell[3036835, 66859, 45848, 1075, 279, 14723, 686, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[3082698, 67937, 663, 16, 68, "Text"],
Cell[3083364, 67955, 164, 3, 96, "Input"],

Cell[CellGroupData[{
Cell[3083553, 67962, 179, 3, 73, "Input"],
Cell[3083735, 67967, 50990, 1314, 279, 17367, 894, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[3134740, 69284, 605, 13, 94, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[3135382, 69302, 63, 0, 68, "Subsection"],
Cell[3135448, 69304, 1683, 44, 148, "Text"],
Cell[3137134, 69350, 202, 7, 42, "Text"],
Cell[3137339, 69359, 350, 9, 42, "Text"],
Cell[3137692, 69370, 324, 8, 42, "Text"],
Cell[3138019, 69380, 302, 8, 42, "Text"],
Cell[3138324, 69390, 399, 11, 42, "Text"],
Cell[3138726, 69403, 1203, 36, 96, "Text"],
Cell[3139932, 69441, 568, 12, 94, "Text"],
Cell[3140503, 69455, 415, 7, 133, "Input"],
Cell[3140921, 69464, 151, 2, 150, "Input"],

Cell[CellGroupData[{
Cell[3141097, 69470, 324, 6, 154, "Input"],
Cell[3141424, 69478, 71223, 1782, 279, 28535, 1250, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[3212662, 71263, 120, 3, 41, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[3212831, 71272, 52, 0, 51, "Section"],
Cell[3212886, 71274, 796, 21, 95, "Text"],
Cell[3213685, 71297, 9050, 236, 390, "Text"],
Cell[3222738, 71535, 221, 4, 67, "Text"],

Cell[CellGroupData[{
Cell[3222984, 71543, 384, 8, 125, "Input"],
Cell[3223371, 71553, 71073, 1688, 279, 21597, 1073, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[3294459, 73244, 241, 5, 41, "Text"],

Cell[CellGroupData[{
Cell[3294725, 73253, 315, 6, 100, "Input"],
Cell[3295043, 73261, 126130, 2204, 279, 18765, 874, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[3421188, 75468, 422, 11, 141, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[3421647, 75484, 55, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[3421727, 75488, 28, 0, 68, "Subsection"],
Cell[3421758, 75490, 499, 14, 68, "Text"],

Cell[CellGroupData[{
Cell[3422282, 75508, 328, 6, 99, "Input"],
Cell[3422613, 75516, 392052, 8185, 363, 221212, 6071, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[3814680, 83704, 663, 20, 94, "Text"],
Cell[3815346, 83726, 59, 0, 41, "Text"],
Cell[3815408, 83728, 386, 9, 68, "Text"],
Cell[3815797, 83739, 230, 5, 41, "Text"],
Cell[3816030, 83746, 218, 5, 41, "Text"],
Cell[3816251, 83753, 144, 2, 41, "Text"],
Cell[3816398, 83757, 235, 7, 41, "Text"],
Cell[3816636, 83766, 309, 8, 42, "Text"],

Cell[CellGroupData[{
Cell[3816970, 83778, 211, 4, 73, "Input"],
Cell[3817184, 83784, 333827, 7488, 238, 219811, 6076, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[4151026, 91275, 114, 3, 58, "Exercise"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[4151189, 91284, 35, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[4151249, 91288, 58, 0, 68, "Subsection"],
Cell[4151310, 91290, 745, 16, 95, "Text"],
Cell[4152058, 91308, 104, 2, 73, "Input"],
Cell[4152165, 91312, 162, 3, 73, "Input"],

Cell[CellGroupData[{
Cell[4152352, 91319, 269, 5, 123, "Input"],
Cell[4152624, 91326, 228495, 6373, 356, 68252, 4389, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[4381134, 97702, 246, 7, 58, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[4381417, 97714, 34, 0, 68, "Subsection"],
Cell[4381454, 97716, 933, 27, 95, "Text"],

Cell[CellGroupData[{
Cell[4382412, 97747, 149, 2, 74, "Input"],
Cell[4382564, 97751, 32931, 849, 233, 12281, 591, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[4415510, 98603, 198, 5, 42, "Text"],
Cell[4415711, 98610, 350, 11, 58, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[4416098, 98626, 60, 0, 68, "Subsection"],
Cell[4416161, 98628, 758, 24, 108, "Text"],
Cell[4416922, 98654, 445, 9, 67, "Text"],

Cell[CellGroupData[{
Cell[4417392, 98667, 219, 3, 99, "Input"],
Cell[4417614, 98672, 24629, 530, 444, 3213, 261, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[4442292, 99208, 37, 0, 68, "Subsection"],
Cell[4442332, 99210, 735, 19, 95, "Text"],
Cell[4443070, 99231, 207, 5, 42, "Text"],
Cell[4443280, 99238, 328, 7, 99, "Input"],
Cell[4443611, 99247, 119, 2, 73, "Input"],

Cell[CellGroupData[{
Cell[4443755, 99253, 74, 1, 73, "Input"],
Cell[4443832, 99256, 449498, 21086, 215, 383284, 20264, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[4893391, 120349, 32, 0, 51, "Section"],
Cell[4893426, 120351, 266, 5, 67, "Text"],
Cell[4893695, 120358, 675, 23, 119, "Text"],
Cell[4894373, 120383, 890, 28, 69, "Text"],
Cell[4895266, 120413, 113, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[4895404, 120420, 203, 4, 74, "Input"],
Cell[4895610, 120426, 637870, 9051, 444, 81305, 2175, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[5533517, 129482, 224, 4, 74, "Input"],
Cell[5533744, 129488, 1058143, 40871, 444, 538850, 34455, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[6591902, 170362, 275, 7, 68, "Text"],
Cell[6592180, 170371, 155, 3, 94, "Input"],
Cell[6592338, 170376, 115, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[6592478, 170383, 76, 1, 73, "Input"],
Cell[6592557, 170386, 106, 2, 80, "Output"]
}, Closed]],
Cell[6592678, 170391, 186, 5, 42, "Text"],

Cell[CellGroupData[{
Cell[6592889, 170400, 586, 11, 125, "Input"],
Cell[6593478, 170413, 67565, 1575, 444, 17797, 956, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[6661058, 171991, 265, 6, 42, "Text"],
Cell[6661326, 171999, 550, 17, 69, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[6661913, 172021, 36, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[6661974, 172025, 46, 0, 68, "Subsection"],
Cell[6662023, 172027, 1002, 23, 122, "Text"],
Cell[6663028, 172052, 621, 13, 151, "Input"],

Cell[CellGroupData[{
Cell[6663674, 172069, 355, 5, 125, "Input"],
Cell[6664032, 172076, 68931, 1834, 213, 26602, 1307, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[6732978, 173913, 629, 13, 95, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[6733644, 173931, 47, 0, 68, "Subsection"],
Cell[6733694, 173933, 1510, 43, 123, "Text"],
Cell[6735207, 173978, 576, 16, 68, "Text"],

Cell[CellGroupData[{
Cell[6735808, 173998, 330, 6, 125, "Input"],
Cell[6736141, 174006, 63206, 1606, 279, 22636, 1101, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[6799408, 175619, 42, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[6799475, 175623, 32, 0, 68, "Subsection"],
Cell[6799510, 175625, 750, 21, 95, "Text"],
Cell[6800263, 175648, 130, 3, 41, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[6800430, 175656, 48, 0, 68, "Subsection"],
Cell[6800481, 175658, 1478, 28, 200, "Text"],

Cell[CellGroupData[{
Cell[6801984, 175690, 1987, 35, 1187, "Input"],
Cell[6803974, 175727, 34479, 750, 284, 11701, 464, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[6838502, 176483, 42, 0, 68, "Subsection"],
Cell[6838547, 176485, 1251, 31, 147, "Text"],
Cell[6839801, 176518, 580, 16, 68, "Text"],
Cell[6840384, 176536, 61, 1, 73, "Input"],
Cell[6840448, 176539, 832, 16, 203, "Input"],
Cell[6841283, 176557, 2051, 36, 359, "Input"],
Cell[6843337, 176595, 335, 7, 67, "Text"],

Cell[CellGroupData[{
Cell[6843697, 176606, 919, 15, 229, "Input"],
Cell[6844619, 176623, 75443, 1455, 250, 16886, 728, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[6920077, 178081, 231, 6, 41, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[6920357, 178093, 33, 0, 54, "ExerciseMain"],

Cell[CellGroupData[{
Cell[6920415, 178097, 66, 0, 58, "ExerciseTitle"],
Cell[6920484, 178099, 608, 21, 67, "ExerciseText"],
Cell[6921095, 178122, 206, 5, 77, "DisplayFormula"],
Cell[6921304, 178129, 235, 6, 41, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[6921576, 178140, 41, 0, 58, "ExerciseTitle"],
Cell[6921620, 178142, 309, 9, 41, "ExerciseText"],
Cell[6921932, 178153, 239, 5, 90, "DisplayFormula",
  Evaluatable->False],
Cell[6922174, 178160, 444, 17, 42, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[6922655, 178182, 50, 0, 58, "ExerciseTitle"],
Cell[6922708, 178184, 348, 11, 41, "ExerciseText"],
Cell[6923059, 178197, 201, 5, 41, "DisplayFormula",
  Evaluatable->False],
Cell[6923263, 178204, 126, 5, 41, "ExerciseText"],
Cell[6923392, 178211, 239, 5, 90, "DisplayFormula",
  Evaluatable->False],
Cell[6923634, 178218, 148, 6, 41, "ExerciseText"],
Cell[6923785, 178226, 234, 5, 86, "DisplayFormula",
  Evaluatable->False],
Cell[6924022, 178233, 86, 1, 41, "ExerciseText"],
Cell[6924111, 178236, 723, 24, 67, "ExerciseText"],
Cell[6924837, 178262, 224, 5, 41, "ExerciseText"],
Cell[6925064, 178269, 421, 14, 42, "ExerciseText"],
Cell[6925488, 178285, 252, 5, 41, "ExerciseText"],
Cell[6925743, 178292, 185, 4, 41, "ExerciseText"],
Cell[6925931, 178298, 391, 8, 67, "ExerciseText"],
Cell[6926325, 178308, 581, 16, 67, "ExerciseText"],
Cell[6926909, 178326, 96, 2, 41, "ExerciseText"],
Cell[6927008, 178330, 195699, 6092, 470, 82598, 4691, "GraphicsData", \
"PostScript", "Graphics"],
Cell[7122710, 184424, 103, 3, 41, "Text"],
Cell[7122816, 184429, 169618, 4180, 294, 51413, 2716, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[7292471, 188614, 50, 0, 58, "ExerciseTitle"],
Cell[7292524, 188616, 74, 0, 41, "ExerciseText"],
Cell[7292601, 188618, 435, 9, 45, "DisplayFormula",
  Evaluatable->False],
Cell[7293039, 188629, 1699, 44, 151, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[7294775, 188678, 40, 0, 58, "ExerciseTitle"],
Cell[7294818, 188680, 307, 8, 42, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[7295162, 188693, 48, 0, 58, "ExerciseTitle"],
Cell[7295213, 188695, 211, 4, 41, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[7295461, 188704, 66, 0, 58, "ExerciseTitle"],
Cell[7295530, 188706, 367, 6, 67, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[7295934, 188717, 63, 0, 58, "ExerciseTitle"],
Cell[7296000, 188719, 1520, 43, 182, "ExerciseText"]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

